PRAGMA foreign_keys=OFF;
BEGIN TRANSACTION;
CREATE TABLE evaluations(boardid integer, depth integer, score text, pv text, foreign key(boardid) references boards(id));
INSERT INTO evaluations VALUES(57766,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(57766,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(57766,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(57766,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(57766,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(57766,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57766,36,'cp 25','e2e4 e7e5 g1f3 b8c6 f1b5 g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 d7d5 f1d3 g7g6 b1d2 e8g7 h2h3 c7c6 d2f3 a7a5 a2a4 c8f5 d3f5 g7f5 e1d2 d8d7 a1e1 a8e8 e1e8 d7e8 f4c7 f6d8 d2f4 f7f6 c7d8 e8d8');
INSERT INTO evaluations VALUES(57752,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(57752,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(57752,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(57752,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(57752,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(57752,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(57752,36,'cp 25','e7e5 g1f3 b8c6 f1b5 g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 c1f4 d7d5 f1d3 g7g6 b1d2 e8g7 d2f3 c8f5 d3f5 g7f5 e1d2 c7c6 a1e1 a7a5 g2g4 f5d6 f4d6 d8d6 h2h4 a5a4 h4h5 a4a3 b2b3 f6e7 h5h6');
INSERT INTO evaluations VALUES(57162,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(57162,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(57162,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(57162,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(57162,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(57162,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(57162,36,'cp 20','g1f3 b8c6 f1b5 g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6f5 c1f4 d7d6 h2h3 c8e6 b1d2 f5h4 f1d3 h4g6 f4g3 d8d7 e1e2 a8e8 e2f3 c7c6 d3c2 f6e7 a1e1 e6d5');
INSERT INTO evaluations VALUES(57155,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(57155,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(57155,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(57155,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(57155,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(57155,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(57155,36,'cp 22','b8c6 f1b5 g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 d7d5 f1d3 g7g6 b1d2 e8g7 d2f3 a7a5 f4h6 c7c6 e1e3 d8b6 a1e1 c8e6 h2h3 b6b2 e1b1 b2c3 b1b7 c3a1 g1h2 a1c3');
INSERT INTO evaluations VALUES(22807,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(22807,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(22807,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(22807,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(22807,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(22807,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22807,36,'cp 29','f1b5 g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 c6e5 e1e5 f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 d8e7 e1e7 f6e7 c3c4 e7f6 b1c3 d7d6 c1e3 f6c3 b2c3 b7b6 e3d4 a7a5 a1e1 g8f8 f2f3 e8f6 g1f2 f6d7');
INSERT INTO evaluations VALUES(22644,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(22644,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(22644,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(22644,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(22644,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(22644,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(22644,36,'cp 22','g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 d7d6 c1e3 c7c6 d5c6 b7c6 b1d2 d6d5 d2f3 c8g4 f3d4 g4d7 e3f4 e8c7 e1d2 c7e6 f4e3 h7h6 d4e6 d7e6 e3d4 a7a5 a1e1 f6d4');
INSERT INTO evaluations VALUES(16315,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(16315,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(16315,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(16315,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(16315,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(16315,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(16315,36,'cp 20','e1g1 f6e4 f1e1 e4d6 f3e5 c6e5 e1e5 f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 d8e7 e1e7 f6e7 c1e3 d7d6 f2f3 c8d7 g1f2 h7h5 c3c4 e7f6 b1c3 a7a5 f1d3 a5a4 a1c1 g7g6 g2g3 e8g7 c3b5 g7e8 c1c2 c7c6 d5c6 b7c6');
INSERT INTO evaluations VALUES(16313,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(16313,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(16313,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(16313,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(16313,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(16313,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(16313,36,'cp 24','f6e4 f1e1 e4d6 f3e5 c6e5 e1e5 f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 c7c6 c1e3 d7d6 d5c6 b7c6 b1d2 d6d5 e3f4 c8f5 d2f3 e8c7 f3d4 c7e6 f4e5 e6d4 c3d4 f6e5 e1e5 g7g6 a1c1 d8b6');
INSERT INTO evaluations VALUES(16757,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16757,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(16757,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(16757,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(16757,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(16757,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(16757,36,'cp 28','f1e1 e4d6 f3e5 c6e5 e1e5 f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 d7d6 c1e3 h7h6 b1d2 c8f5 d2b3 f6g5 c3c4 g5e3 e1e3 d8g5 a1e1 g5e3 e1e3 g8f8 b3d4 f5d7 f2f3 a7a5 e3b3 b7b6 h2h3');
INSERT INTO evaluations VALUES(16758,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(16758,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(16758,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(16758,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(16758,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(16758,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(16758,36,'cp 22','e4d6 f3e5 c6e5 e1e5 f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 d7d6 c1e3 c7c6 d5c6 b7c6 b1d2 c8f5 d2b3 d6d5 e1d2 a7a5 a1e1 e8d6 e3d4 f6g5 d2d1 h7h6 b3c5 a8b8 b2b3 g5f6 f1d3 f6d4 c3d4 b8b4 d1c2 f5d3 c2d3');
INSERT INTO evaluations VALUES(16876,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16876,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(16876,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(16876,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(16876,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(16876,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(16876,36,'cp 22','f3e5 c6e5 e1e5 f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 c7c6 c1e3 h7h6 e1d2 d7d6 d5c6 b7c6 c3c4 d6d5 c4d5 d8d5 d2e2 c8f5 b1c3 d5e6 a1d1 e8c7 e2c4 e6c4 f1c4 f6c3 b2c3 f5e6 c4e6 c7e6');
INSERT INTO evaluations VALUES(16874,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(16874,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(16874,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(16874,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(16874,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(16874,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(16874,36,'cp 25','f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 c7c6 c1e3 c6d5 e1d2 e8c7 b1a3 d7d6 a1e1 c8d7 a3c2 d7c6 g2g3 d8d7 f1g2 a7a5 e3d4 f6d4 d2d4 h7h6 e1d1 c7e6 d4d2 a8e8 g2d5');
INSERT INTO evaluations VALUES(10965,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(10965,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(10965,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(10965,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(10965,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(10965,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(10965,36,'cp 20','b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 c1f4 d7d5 a2a4 a7a5 f1d3 g7g6 b1d2 e8g7 d2f3 c8f5 d3f5 g7f5 e1d1 c7c6 d1b3 d8d7 a1e1 f5g7 g2g3 h7h5 f4e5 f6e7 b3c2 a8e8 e5g7');
INSERT INTO evaluations VALUES(10966,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(10966,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(10966,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(10966,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(10966,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(10966,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(10966,36,'cp 18','c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 h7h6 e1e4 d8e7 e4d3 e7e5 b1a3 d7d6 c1d2 c8d7 a3c4 e5f5 d3f5 d7f5 f2f3 a7a5 c4e3 f5d7 a2a4 h6h5 b2b3 g7g6 g2g3 f6e5 g1f2 e8f6 a1d1 h5h4 d1e1 h4g3 h2g3');
INSERT INTO evaluations VALUES(57751,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(57751,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(57751,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(57751,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(57751,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(57751,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(57751,36,'cp 29','g8f6 c2c4 e7e6 g1f3 d7d5 b1c3 f8b4 d1a4 b8c6 e2e3 e8g8 c1d2 d5c4 f1c4 b4d6 e1g1 a7a6 a2a3 e6e5 d4d5 c6e7 a4c2 c8g4 f3g5 g4f5 e3e4 f5g6 d2e3 c7c6 d5c6 e7c6 f1d1 c6d4 c2d2 b7b5 c4a2 d8e7 c3e2 g6h5 e2d4 e5d4 d2d4 h5d1');
INSERT INTO evaluations VALUES(50342,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(50342,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50342,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(50342,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(50342,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50342,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(50342,36,'cp 20','c2c4 e7e6 g1f3 d7d5 b1c3 f8b4 d1a4 b8c6 e2e3 e8g8 c1d2 d5c4 f1c4 b4d6 e1g1 e6e5 d4d5 c6e7 a4c2 c8g4 f3g5 g8h8 a1c1 h7h6 g5e4 g4f5 a2a3 a7a5 f2f3 f6e4 c3e4 e7g8 d2e1 g8f6 e1h4 g7g5 h4e1 f5g6 c4d3 f6d5 e4d6 d8d6 e1g3');
INSERT INTO evaluations VALUES(50326,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(50326,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(50326,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(50326,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(50326,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(50326,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(50326,36,'cp 18','e7e6 g1f3 d7d5 b1c3 f8b4 c4d5 e6d5 c1g5 e8g8 e2e3 h7h6 g5h4 c8f5 f1e2 b8d7 d1b3 b4c3 b3c3 c7c6 f3d2 f6e4 h4d8 e4c3 b2c3 f8d8 e1g1 d8c8 f1c1 c6c5 a2a4 d7f6 h2h3 c5d4 c3d4 b7b6 e2b5 c8c1 a1c1 a8c8 c1c8 f5c8 f2f3 c8d7 b5d7 f6d7 e3e4 d7f6 e4e5 f6e8');
INSERT INTO evaluations VALUES(49397,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(49397,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(49397,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(49397,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(49397,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(49397,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(49397,36,'cp 22','g1f3 d7d5 b1c3 f8b4 d1a4 b8c6 e2e3 e8g8 a4c2 f8e8 c1d2 h7h6 a2a3 b4f8 c4d5 e6d5 f1d3 c8g4 e1g1 f8d6 g1h1 g4h5 f3g1 c6e7 c3b5 h5g6 d3g6 e7g6 f2f3 c7c6 b5d6 d8d6 a1c1 a8c8 g1e2 b7b6 e2g3 d6e6 f1e1 c6c5 d4c5 b6c5 d2c3 f6d7 g3f5');
INSERT INTO evaluations VALUES(49393,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(49393,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(49393,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(49393,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(49393,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(49393,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(49393,36,'cp 14','d7d5 b1c3 f8b4 d1a4 b8c6 e2e3 e8g8 a4c2 f8e8 c1d2 b4f8 a2a3 h7h6 c4d5 e6d5 f1d3 c8g4 e1g1 f8d6 g1h1 g4h5 f3g1 c6e7 c3b5 h5g6 d3g6 f7g6 a1c1 d8d7 b5d6 c7d6 g1e2 a8c8 c2b3 g8h7 f2f3 e7f5 b3d3 a7a6 e2f4 c8c1 f1c1');
INSERT INTO evaluations VALUES(48267,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(48267,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(48267,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(48267,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(48267,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(48267,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(48267,36,'cp 23','b1c3 f8b4 d1a4 b8c6 e2e3 e8g8 a4c2 f8e8 c1d2 b4f8 a2a3 h7h6 c4d5 e6d5 f1d3 c6b8 e1g1 c7c6 c3e2 f8d6 e2g3 b8d7 f1c1 a7a6 b2b4 d7f8 a3a4 c8d7 d2c3 d8e7 f3e5 f6g4 e5g4 d7g4 b4b5 a6b5 a4b5 a8a1 c1a1 c6b5 d3b5 e8c8 b5d3');
INSERT INTO evaluations VALUES(48255,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(48255,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(48255,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48255,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48255,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(48255,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(48255,36,'cp 8','f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 d8e7 f3d2 e7e6 f2f3 c7c5 h4f6 e6f6 d4c5 f6c3 b2c3 f8c8 d2b3 b8d7 e1d2 d7c5 b3d4 f5d7 a1b1 c8c7 f1b5 d7b5 d4b5 c7c8 h1c1 a7a6 b5d4 c8c7');
INSERT INTO evaluations VALUES(42045,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(42045,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(42045,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(42045,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42045,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42045,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42045,36,'cp 19','c1g5 h7h6 g5f6 d8f6 e2e3 c7c5 c4d5 e6d5 d1b3 e8g8 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7c5 d1d5 b7b6 c3c4 c8b7 d5d4 a8c8 f1e2 f8d8 e1g1 b7f3 g2f3 d8d4 e3d4 c5e6 f1d1 e6f4 e2f1 g8f8 d4d5 f4g6 d1d4 g6e5 f1e2 c8d8 a2a3 f8e7 f3f4 e5g6 e2d3');
INSERT INTO evaluations VALUES(41991,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(41991,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41991,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(41991,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(41991,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41991,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41991,36,'cp 17','h7h6 g5f6 d8f6 e2e3 c7c5 c4d5 e6d5 d1b3 e8g8 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 c8d7 d4b3 f8c8 f2f3 d7a4 d1d4 a4b3 a2b3 c8c5 e1d2 a8c8 f1d3 c5a5 h2h4 g8f8 h1b1 b7b6 d3f5 a5a2 f5c2 h6h5 d4a4 a2a4 b3a4 f6d7 c2f5');
INSERT INTO evaluations VALUES(41642,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41642,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41642,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41642,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(41642,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41642,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(41642,36,'cp 19','g5f6 d8f6 e2e3 c7c5 c4d5 e6d5 d1b3 e8g8 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 c8d7 f2f3 f8c8 d4b3 c8c7 f1e2 d7a4 d1d4 a4b3 a2b3 c7c5 e1d2 a8c8 e2d3 c5a5 h1c1 g8f8 h2h4 a5a3 d3c2 h6h5 c1d1 a3a2 d4f4 b7b6 f4f5 a7a5');
INSERT INTO evaluations VALUES(41634,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41634,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(41634,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(41634,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41634,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41634,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(41634,36,'cp 21','d8f6 d1b3 c7c5 c4d5 e6d5 e2e3 e8g8 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 c8d7 f2f3 f8c8 d4b3 d7a4 f1e2 a4b3 a2b3 c8c5 e1d2 a7a5 d1a1 a8c8 h1c1 g8f8 b3b4 a5b4 c3b4 c5c1 a1c1 c8c1 d2c1 f8e7 c1b2 f6e8 b2b3 e7d6 b3c3 e8c7 e2d3');
INSERT INTO evaluations VALUES(33408,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(33408,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(33408,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(33408,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(33408,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(33408,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(33408,36,'cp 15','d1b3 c7c5 c4d5 e6d5 e2e3 e8g8 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 c8d7 f2f3 f8c8 d4b3 d7a4 d1d4 a4b3 a2b3 c8c5 e1d2 a8c8 f1d3 c5a5 h2h4 g8f8 g2g4 f6d7 h1c1 d7e5 d3e2 e5c6 d4a4 a5a4 b3a4 c6a5 c1b1 c8c5');
INSERT INTO evaluations VALUES(52450,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52450,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(52450,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(52450,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52450,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52450,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(52450,36,'cp 33','g1f3 b8c6 d2d4 c5d4 f3d4 g8f6 b1c3 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 f1e2 e8g8 e1g1 b8d7 c1e3 a7a6 b5a3 f7f5 f2f3 f5f4 e3f2 b7b6 a3c4 a6a5 g1h1 d7c5 e2d3 g8h8 b2b3 c8a6 f2c5 b6c5 c4a3 a6b7 d3c4 d8d7 a3b5');
INSERT INTO evaluations VALUES(52444,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(52444,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(52444,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(52444,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(52444,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52444,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(52444,36,'cp 30','b8c6 b1c3 e7e5 f1c4 d7d6 d2d3 f8e7 f3d2 e7g5 h2h4 g5h6 d1h5 g7g6 h5d1 g8f6 d2f1 h6c1 d1c1 h7h5 c1g5 c6d4 f1e3 f6g4 g5d8 e8d8 c3d1 g4e3 d1e3 c8e6 c2c3 d4c6 e1e2 d8d7 a2a4 a8f8 g2g3 f7f5 a4a5 c6e7 f2f3 e6c4 e3c4 e7c6 e2e3 f5f4 g3f4 f8f4');
INSERT INTO evaluations VALUES(19462,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(19462,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(19462,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(19462,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(19462,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(19462,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(19462,36,'cp 30','b1c3 e7e5 f1c4 f8e7 d2d3 d7d6 f3d2 g8f6 d2f1 c8g4 d1d2 f6d7 c3d5 e8g8 f1e3 g4e6 d5e7 d8e7 b2b3 d7b6 e1g1 b6c4 d3c4 f7f5 e4f5 e6f5 c1b2 f5e6 e3d5 e7d7 f2f4 c6d4 f4e5 d6e5 c2c3 d4f5 b2a3 b7b6 a1e1 e6d5 d2d5 d7d5');
INSERT INTO evaluations VALUES(19434,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(19434,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(19434,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(19434,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(19434,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(19434,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19434,36,'cp 29','e7e5 f1c4 f8e7 d2d3 d7d6 a2a4 g8f6 c1g5 a7a6 g5f6 e7f6 c3d5 e8g8 c2c3 a8b8 d1d2 c8g4 d2e3 g4e6 e1g1 h7h6 f1b1 f6e7 h2h3 b7b5 a4b5 a6b5 c4b3 d8c8 e3e2 g8h8 d5e7 c6e7 b3e6 c8e6 a1a7 f8c8 b1a1');
INSERT INTO evaluations VALUES(19002,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(19002,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(19002,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(19002,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(19002,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(19002,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19002,36,'cp 42','f1c4 f8e7 d2d3 d7d6 a2a4 a7a6 f3d2 g8f6 d2f1 c8g4 d1d2 g4e6 f1e3 a8b8 d2d1 e8g8 e1g1 f6d7 e3d5 e7g5 c1e3 c6d4 e3g5 d8g5 a4a5 g8h8 d5e3 d7f6 c3d5 f6d5 c4d5 e6d5 e3d5 f7f5 c2c3 d4c6 f2f3 f5f4 d1b3 g5g6 b3b6 f8d8');
INSERT INTO evaluations VALUES(18963,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(18963,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(18963,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18963,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(18963,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(18963,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(18963,36,'cp 35','f8e7 d2d3 d7d6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 d8c7 c2c3 e8g8 a4a5 a8b8 e1g1 c8e6 d5e6 f7e6 d1e2 f8f7 b2b4 c5b4 c3b4 c7d7 f1c1 f6d8 e2a2 h7h6 c1c2 b8c8 c2c4 c8c4 a2c4 g8h7');
INSERT INTO evaluations VALUES(5243,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(5243,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(5243,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(5243,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(5243,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(5243,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(5243,36,'cp 41','d2d3 d7d6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 d8c7 c2c3 e8g8 e1g1 c8e6 d1b3 e6g4 f3d2 f6g5 h2h3 g4e2 f1e1 e2d3 d2c4 g8h8 a1d1 d3c4 b3c4 g7g6 g2g3 g5h6 h3h4 f7f5 e4f5 g6f5 c4b3 e5e4 b3b7 c7b7');
INSERT INTO evaluations VALUES(5244,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(5244,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(5244,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(5244,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(5244,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(5244,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(5244,36,'cp 21','d7d6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 e8g8 c2c3 c8e6 d1b3 e6g4 f3d2 f6g5 d2c4 a8b8 e1g1 d8e7 a4a5 g4e6 d5e6 f7e6 b3d1 f8f7 g2g3 b8f8 g1g2 e7d7 d1g4 g5d8 c4d2 h7h6 h2h4 d6d5 e4d5 e6d5 g4d7 f7d7');
INSERT INTO evaluations VALUES(55064,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55064,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(55064,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(55064,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(55064,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(55064,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(55064,36,'cp 31','c2c4 e7e6 b1c3 g8f6 c4d5 e6d5 c1g5 f8e7 e2e3 e8g8 f1d3 h7h6 g5h4 c7c6 d1c2 b8d7 g1f3 f8e8 e1g1 f6e4 d3e4 e7h4 e4h7 g8h8 f3h4 d8h4 h7d3 d7f6 a1e1 f6g4 h2h3 g4f6 a2a3 f6h5 c3e2 h5f6 b2b4 f6e4 b4b5 c8h3 g2h3 h4h3 b5c6 b7c6 d3e4 e8e4');
INSERT INTO evaluations VALUES(55044,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(55044,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55044,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55044,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(55044,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(55044,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(55044,36,'cp 26','e7e6 b1c3 g8f6 c4d5 e6d5 c1g5 c7c6 e2e3 h7h6 g5h4 f8e7 f1d3 e8g8 d1c2 b8d7 g1f3 f8e8 e1g1 f6e4 d3e4 e7h4 e4h7 g8h8 f3h4 d8h4 h7d3 d7f6 a1e1 f6g4 h2h3 g4f6 a2a3 h4g5 f2f4 g5h4 c2f2 h4f2 g1f2 h6h5 e3e4 f6e4 c3e4 d5e4 e1e4 e8e4 d3e4 c8e6 f1e1');
INSERT INTO evaluations VALUES(55818,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55818,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(55818,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(55818,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(55818,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55818,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(55818,36,'cp 26','b1c3 g8f6 c4d5 e6d5 c1g5 c7c6 e2e3 h7h6 g5h4 f8e7 d1c2 b8d7 g1f3 e8g8 f1d3 f8e8 e1g1 f6e4 h4e7 d8e7 f1e1 d7f6 f3e5 f6d7 d3e4 d5e4 e5d7 c8d7 a1c1 a8d8 c3e2 d7g4 e2g3 g7g6 h2h3 g4c8 a2a3 g8g7 c2c5 e7c5 c1c5');
INSERT INTO evaluations VALUES(55814,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(55814,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(55814,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(55814,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(55814,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55814,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(55814,36,'cp 30','g8f6 c4d5 e6d5 c1g5 c7c6 e2e3 h7h6 g5h4 c8f5 d1f3 f5e6 h4f6 d8f6 f3f6 g7f6 f1d3 a7a5 g1e2 h6h5 h2h4 f8d6 g2g3 b8d7 e2f4 d7b6 c3e2 a5a4 a1c1 b6c4 f4e6 d6b4 e1f1 c4d2 f1g1 f7e6 g1g2 a4a3 b2b3');
INSERT INTO evaluations VALUES(48258,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48258,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(48258,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(48258,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(48258,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(48258,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(48258,36,'cp 30','c4d5 e6d5 c1g5 c7c6 e2e3 h7h6 g5h4 c8f5 d1f3 f5e6 h4f6 d8f6 f3f6 g7f6 f1d3 b8d7 g1e2 h6h5 h2h4 f8d6 f2f3 a7a5 e1f2 f6f5 g2g3 e8e7 e2f4 d7f8 f4g2 f8g6 c3e2 e7d8 a2a3 d6e7 d3c2 b7b5 c2d3 d8c7 h1c1');
INSERT INTO evaluations VALUES(48179,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(48179,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(48179,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(48179,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(48179,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(48179,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48179,36,'cp 32','e6d5 c1g5 c7c6 e2e3 h7h6 g5h4 c8f5 d1f3 f5e6 h4f6 d8f6 f3f6 g7f6 f1d3 b8d7 g1e2 d7b6 e2g3 b6c4 b2b3 f8b4 g3e2 c4d6 a2a3 b4c3 e2c3 h8g8 g2g3 e8e7 f2f3 a7a5 e1f2 h6h5 c3a4 h5h4 a1c1 g8h8 a4b6 a8b8 d3b1 h4g3 h2g3');
INSERT INTO evaluations VALUES(48662,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48662,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(48662,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(48662,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48662,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(48662,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48662,36,'cp 28','c1g5 c7c6 e2e3 h7h6 g5h4 c8f5 d1f3 f5e6 h4f6 d8f6 f3f6 g7f6 f1d3 b8d7 g1e2 h6h5 h2h4 f8d6 g2g3 a7a5 a1c1 d7b6 f2f3 f6f5 e2f4 a5a4 e1f2 e8d7 c3e2 b6c8 c1c2 c8e7 a2a3 d7c7 h1b1 c7d8 b1c1 e6d7 c1d1 e7g6 f4g2');
INSERT INTO evaluations VALUES(48622,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(48622,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(48622,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(48622,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48622,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(48622,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(48622,36,'cp 21','c7c6 e2e3 h7h6 g5h4 c8f5 d1f3 f5e6 h4f6 d8f6 f3f6 g7f6 g1e2 b8d7 e2g3 d7b6 f1d3 b6c4 b2b3 f8b4 g3e2 c4b6 f2f3 b4d6 e1f2 f6f5 g2g3 a7a5 a2a3 e8e7 a3a4 h8g8 e2f4 d6f4 g3f4 b6c8 c3e2 c8d6 e2g3');
INSERT INTO evaluations VALUES(45752,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(45752,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45752,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45752,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45752,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(45752,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(45752,36,'cp 26','e2e3 h7h6 g5h4 c8f5 d1f3 f5e6 h4f6 d8f6 f3f6 g7f6 g1e2 b8d7 e2g3 d7b6 f1d3 b6c4 b2b3 f8b4 g3e2 c4b6 a2a3 b4d6 e2g3 h6h5 g3f5 e6f5 d3f5 h5h4 e1g1 a7a5 e3e4 a8d8 e4d5 b6d5 c3d5 c6d5 f1b1 e8g8 b3b4 f8e8 b4b5');
INSERT INTO evaluations VALUES(45749,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(45749,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(45749,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(45749,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(45749,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(45749,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(45749,36,'cp 35','h7h6 g5h4 c8f5 d1f3 f5e6 h4f6 d8f6 f3f6 g7f6 g1e2 a7a5 e2f4 f8d6 f1d3 a5a4 c3e2 e8e7 h2h3 f6f5 a2a3 b8d7 a1c1 d7b6 g2g3 h8g8 h3h4 b6d7 f4h5 g8b8 e2f4 d7f6 h5f6 e7f6 f4h5 f6e7 e1f1 e6d7 f1g2');
INSERT INTO evaluations VALUES(45550,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(45550,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(45550,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45550,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45550,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(45550,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(45550,36,'cp 36','g5h4 f8e7 d1c2 e8g8 f1d3 f8e8 g1e2 b8d7 e1g1 a7a5 a2a3 d7f8 h2h3 f8e6 a1c1 f6h5 h4e7 d8e7 c3a4 a8b8 d3h7 g8h8 h7f5 e6g5 f5c8 b8c8 a4c5 c8a8 c2b3 e8b8 c5d3 g5e4 b3c2 h8g8 d3e5 h5f6');
INSERT INTO evaluations VALUES(45552,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(45552,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45552,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(45552,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(45552,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(45552,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45552,36,'cp 27','f8e7 d1c2 e8g8 f1d3 f8e8 h2h3 b8d7 g1e2 d7f8 h4g3 f8e6 e1g1 a7a5 e2f4 e6f4 g3f4 c8e6 a1c1 a8c8 f4e5 f6d7 e5g3 e7f8 f1d1 d7f6 f2f3 b7b5 g1h1 a5a4 a2a3 f8d6 g3d6 d8d6 c2f2 e6d7 d3b1 f6h5');
INSERT INTO evaluations VALUES(57044,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(57044,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(57044,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(57044,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(57044,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(57044,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(57044,36,'cp 33','d2d4 d7d5 e4e5 c7c5 c2c3 b8c6 g1f3 c8d7 f1e2 c5d4 c3d4 g8e7 e1g1 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 e2d3 f5e3 f2e3 c6b4 d3b1 b4c6 h2h4 a8c8 f1f2 e8g8 a2a3 b7b5 a4c3 b5b4 a3b4 c6b4 f3g5 g7g6 d1g4 a7b6');
INSERT INTO evaluations VALUES(57029,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(57029,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(57029,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(57029,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57029,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(57029,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(57029,36,'cp 16','d7d5 b1c3 g8f6 c1g5 d5e4 c3e4 b8d7 e4f6 d7f6 f1d3 c7c5 d4c5 f8c5 g1f3 d8c7 e1g1 c5d6 f1e1 c8d7 d1e2 h7h6 g5f6 g7f6 d3b5 d7b5 e2b5 e8f8 e1e4 a8d8 e4c4 c7b6 a2a4 f8e7 b5b6 a7b6 c4h4 d6c5 g2g3 d8d5 g1g2 h6h5');
INSERT INTO evaluations VALUES(55882,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55882,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(55882,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(55882,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(55882,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55882,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(55882,36,'cp 32','b1c3 g8f6 e4e5 f6d7 f2f4 c7c5 g1f3 b8c6 c1e3 c5d4 f3d4 d8b6 d1d2 b6b2 a1b1 b2a3 f1b5 c6d4 e3d4 a7a6 b5d7 c8d7 b1b3 a3e7 b3b7 e7h4 g2g3 h4d8 d4b6 d8c8 b7c7 c8d8 d2d4 f8a3 f4f5 e8g8 e1g1 a8c8 c7b7 d8e8 f5f6 g7f6 f1f6 a3e7 d4g4 g8h8 f6f3 d7c6');
INSERT INTO evaluations VALUES(55874,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(55874,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(55874,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55874,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(55874,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55874,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(55874,36,'cp 14','g8f6 c1g5 d5e4 c3e4 b8d7 g1f3 h7h6 e4f6 d7f6 g5h4 c7c5 f1d3 c5d4 f3d4 f8c5 d4f3 e8g8 e1g1 b7b6 d1e2 c8b7 a1d1 d8c7 f3e5 f8d8 h4g3 c7e7 c2c3 a8c8 g3h4 e7c7 f1e1 b7a8 g1h1 d8d5 h4g3 c8d8 b2b4 c5f8 b4b5 c7c8 f2f3');
INSERT INTO evaluations VALUES(48377,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(48377,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(48377,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(48377,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(48377,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(48377,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(48377,36,'cp 24','c1g5 d5e4 c3e4 b8d7 g1f3 h7h6 e4f6 d7f6 g5e3 f8d6 f1d3 e8g8 d1e2 b7b6 e1c1 f6d5 e3d2 d5f4 d3h7 g8h8 d2f4 d6f4 c1b1 c8b7 h7e4 b7e4 e2e4 d8f6 c2c3 a8d8 g2g3 f4d6 d1d2 f8e8 h1d1 f6f5 d1e1 f5b5 d2e2 h8g8 h2h4 a7a5 f3e5');
INSERT INTO evaluations VALUES(48220,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(48220,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(48220,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(48220,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48220,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(48220,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48220,36,'cp 19','d5e4 c3e4 b8d7 g1f3 h7h6 e4f6 d7f6 g5e3 f8d6 f1d3 e8g8 d1e2 b7b6 e1c1 f6d5 e3d2 d5f4 d3h7 g8h8 d2f4 d6f4 c1b1 c8b7 h7e4 b7e4 e2e4 d8f6 g2g3 f4d6 d1d2 a8d8 c2c3 f8e8 h1d1 f6f5 e4f5 e6f5 b1c2 e8e4 f3h4 g7g6 f2f4 d8e8 h4f3 f7f6');
INSERT INTO evaluations VALUES(48454,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(48454,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(48454,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(48454,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(48454,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(48454,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(48454,36,'cp 22','c3e4 f8e7 g5f6 e7f6 g1f3 e8g8 d1d2 b8d7 e1c1 b7b6 f1d3 c8b7 h2h4 a8c8 d2f4 f6e7 f3e5 d7f6 e4f6 e7f6 d3e4 b7e4 f4e4 f6e5 e4e5 d8d6 h1e1 f8d8 e1e3 d6e5 e3e5 g8f8 c2c3 c7c6 h4h5');
INSERT INTO evaluations VALUES(48452,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(48452,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(48452,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48452,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48452,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48452,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48452,36,'cp 12','f8e7 g5f6 e7f6 g1f3 e8g8 d1d2 b8d7 e4f6 d7f6 f3e5 b7b6 e1c1 c8b7 f1d3 d8e7 h1e1 a8d8 d2e3 f6d7 g2g3 d7e5 d4e5 d8d7 d3b5 d7d5 b5c4 d5d1 e1d1 h7h6 a2a3 f8d8 d1d8 e7d8 e3d2 d8e7 c4d3 a7a5 d3e2 b7c6');
INSERT INTO evaluations VALUES(41028,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(41028,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41028,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(41028,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(41028,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(41028,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(41028,36,'cp 16','g5f6 e7f6 g1f3 b8d7 f1d3 c7c5 e4f6 d7f6 d4c5 d8a5 d1d2 a5c5 e1c1 e8g8 d2g5 c5g5 f3g5 b7b6 g5e4 c8b7 e4f6 g7f6 d3e2 f8c8 d1d7 b7c6 d7d4 g8f8 h1g1 c8c7 c2c3 a8c8 e2a6 c8b8 a6d3 f8e7 g2g3 c6d5 a2a3 b8h8 g1d1 h7h5 d4h4');
INSERT INTO evaluations VALUES(40905,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40905,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40905,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(40905,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40905,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(40905,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40905,36,'cp 25','e7f6 g1f3 b8d7 f1d3 c7c5 e4f6 d7f6 d4c5 d8a5 d1d2 a5c5 e1c1 e8g8 d2g5 c5g5 f3g5 b7b6 g5e4 c8b7 e4f6 g7f6 d1d2 f6f5 h1d1 g8g7 d3e2 b7c6 g2g3 a8c8 c2c4 g7f6 d2d6 f8g8 d1d2 f6e7 b2b4 a7a5 a2a3');
INSERT INTO evaluations VALUES(41977,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41977,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(41977,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(41977,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(41977,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41977,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(41977,36,'cp 27','g1f3 e8g8 f1c4 a7a6 e1g1 b8d7 d1e2 c7c5 f1d1 c5d4 e4f6 d8f6 e2e4 a8a7 e4d4 f6d4 d1d4 b7b5 c4f1 a7c7 c2c4 b5c4 a1c1 c8b7 d4c4 c7c4 c1c4 f8d8 b2b4 d7b6 c4c1 g8f8 f3e5 d8c8 c1c8 b7c8 f1d3 h7h6 a2a3 f8e7 f2f3 c8b7');
INSERT INTO evaluations VALUES(41976,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41976,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(41976,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(41976,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41976,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41976,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(41976,36,'cp 21','e8g8 f1c4 a7a6 d1e2 b8d7 e1g1 c7c5 f1d1 c5d4 e4f6 d8f6 d1d4 b7b5 c4b3 d7c5 a1d1 c8b7 f3e5 a8d8 h2h4 d8d4 d1d4 f6e7 c2c3 c5b3 a2b3 f8d8 e2d2 d8d4 d2d4 f7f6 e5d3 g8f7 d4b6 e7d7 d3c5 d7d1 g1h2 b7d5 c5a6 d1b3');
INSERT INTO evaluations VALUES(37016,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(37016,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(37016,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(37016,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(37016,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(37016,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(37016,36,'cp 13','f1c4 a7a6 d1e2 b7b5 c4d3 c8b7 e1g1 b8d7 c2c3 f6e7 f1d1 b7d5 h2h3 g7g6 a2a4 c7c6 a4b5 a6b5 a1a8 d8a8 h3h4 a8d8 h4h5 f7f5 e4g3 d7f6 f3e5 g8g7 c3c4 b5c4 d3c4 e7d6 c4d5 c6d5 d1c1 f6e4 g3e4 f5e4');
INSERT INTO evaluations VALUES(22708,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(22708,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(22708,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(22708,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(22708,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(22708,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(22708,36,'cp 4','g8f6 d2d3 f8c5 b1c3 h7h6 e1g1 e8g8 a2a3 a7a6 c1e3 d7d6 e3c5 d6c5 c3d5 b7b5 d5f6 d8f6 c4d5 f8e8 f3d2 a8b8 d1f3 f6f3 d2f3 b8b6 a3a4 c6b4 a4b5 b4c2 a1c1 c2b4 c1c5 b6b5 c5b5 a6b5 f1c1 c8e6 d5c6 b4c6 c1c6 e6g4 c6c7 g4f3 g2f3 e8d8 c7c5 d8d3 c5e5');
INSERT INTO evaluations VALUES(16403,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16403,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(16403,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(16403,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(16403,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(16403,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(16403,36,'cp 20','d2d3 f8c5 c2c3 a7a6 e1g1 d7d6 a2a4 e8g8 h2h3 h7h6 c4b3 c8e6 b3e6 f7e6 b2b4 c5a7 c1e3 f6h5 e3a7 a8a7 g2g3 a7a8 b1d2 h5f6 d1e2 d8d7 b4b5 c6e7 d3d4 e5d4 c3d4 a6b5 a4b5 c7c6 b5b6 a8a1 f1a1 e7g6 e4e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(16395,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(16395,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(16395,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(16395,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(16395,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(16395,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(16395,36,'cp 14','f8c5 e1g1 a7a6 b1c3 d7d6 c3d5 f6d5 c4d5 e8g8 c2c3 d8f6 a2a4 c5a7 h2h3 c8e6 d5e6 f7e6 b2b4 h7h6 d1e2 f8f7 c1e3 a7e3 f2e3 a8f8 f3d2 f6g5 f1f7 f8f7 a1f1 f7f1 g1f1 g5g6 f1e1 g6g5 e2f3');
INSERT INTO evaluations VALUES(10119,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(10119,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(10119,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(10119,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10119,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(10119,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(10119,36,'cp 22','c2c3 a7a5 e1g1 d7d6 f1e1 e8g8 h2h3 h7h6 c4b5 c5a7 a2a4 c6e7 d3d4 e7g6 b5d3 f8e8 b1d2 c7c6 d2f1 e5d4 c3d4 d6d5 e4e5 f6e4 d3e4 d5e4 e1e4 f7f5 e4e2 c8e6 c1d2 g6h4 f3h4 d8h4 d2a5 h4d4 a5c3 d4d1 a1d1');
INSERT INTO evaluations VALUES(10111,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(10111,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(10111,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10111,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(10111,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10111,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(10111,36,'cp 0','a7a5 e1g1 d7d6 f1e1 e8g8 c4b3 c5a7 h2h3 h7h6 a2a4 f8e8 b1a3 c8e6 a3c4 f6h5 d3d4 e5d4 c3d4 d6d5 e4d5 e6d5 c1e3 h5f6 f3e5 a7b8 e3f4 b8a7');
INSERT INTO evaluations VALUES(55082,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(55082,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(55082,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(55082,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(55082,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(55082,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(55082,36,'cp 75','e4d5 g8f6 d2d4 d8d5 b1c3 d5a5 g1f3 c8f5 f3e5 c7c6 f1c4 e7e6 g2g4 f5g6 h2h4 b8d7 e5d7 f6d7 h4h5 g6e4 e1g1 e4d5 c3d5 c6d5 c4d3 f8d6 d1f3 a5b6 c2c3 d6c7 g1g2 b6d6 f1h1 h7h6 a2a4 e8c8 b2b4 c8b8 c1e3 a7a6 b4b5');
INSERT INTO evaluations VALUES(55019,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(55019,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(55019,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(55019,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(55019,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(55019,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(55019,36,'cp 49','d8d5 b1c3 d5a5 d2d4 g8f6 g1f3 c7c6 f1c4 c8f5 f3e5 e7e6 g2g4 f5g6 h2h4 b8d7 e5d7 f6d7 h4h5 g6e4 e1g1 e4d5 c3d5 c6d5 c4d3 f8d6 d1f3 g7g6 h5h6 e8g8 c2c4 d5c4 d3c4 a5d8 g1g2 d7b6 c4b3 a8c8 c1e3 b6c4 b3c4 c8c4 f1c1 c4c1 a1c1');
INSERT INTO evaluations VALUES(34603,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(34603,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(34603,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(34603,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(34603,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(34603,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(34603,36,'cp 64','b1c3 d5a5 d2d4 g8f6 g1f3 c8f5 f3e5 c7c6 f1c4 e7e6 g2g4 f5g6 h2h4 b8d7 e5d7 f6d7 h4h5 g6e4 e1g1 e4d5 c3d5 c6d5 c4d3 f8d6 d1f3 g7g6 h5h6 e8g8 c2c3 a8e8 g1g2 a5d8 a2a4 e6e5 f3d5 d7f6 d5b7 f6g4 b7f3 f7f5 d4e5');
INSERT INTO evaluations VALUES(34595,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(34595,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(34595,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(34595,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(34595,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(34595,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(34595,36,'cp 63','d5a5 g1f3 g8f6 d2d4 c8f5 f3e5 c7c6 c1f4 b8d7 e5c4 a5d8 c4e3 d8a5 a2a3 g7g6 b2b4 a5d8 e3f5 g6f5 f1d3 e7e6 e1g1 f8g7 c3e2 a7a5 c2c4 e8g8 d1c2 f8e8 f4d2 a5b4 a3b4 d8e7 h2h3 d7f8 d2c3 f8g6 f1b1 f6h5 b4b5 a8a1 b1a1 c6b5 c4b5');
INSERT INTO evaluations VALUES(34687,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(34687,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(34687,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(34687,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(34687,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(34687,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(34687,36,'cp 76','g1f3 c7c6 d2d4 g8f6 c1f4 c8g4 h2h3 g4f3 d1f3 e7e6 f1e2 b8d7 e1g1 f8e7 c3d1 a5d5 f3d3 e8g8 e2f3 d5a5 d1e3 a8d8 c2c3 f6h5 f4h2 a5g5 d3c2 h5f4 g1h1 d7f6 a2a4 f6d5 c2b3 d8d7 e3c4 f4g6 a1e1 g6h4 c4e5 h4f3');
INSERT INTO evaluations VALUES(34681,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(34681,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(34681,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(34681,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(34681,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(34681,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(34681,36,'cp 86','g8f6 d2d4 c7c6 c1f4 c8g4 h2h3 g4f3 d1f3 e7e6 a2a3 f6d5 f4d2 b8d7 c3e4 a5b6 e1c1 d5f6 e4f6 d7f6 d2e3 f6d5 h3h4 f8e7 g2g4 a8c8 g4g5 c6c5 d4c5 e7c5 e3c5 c8c5 f1d3 b6c7 c1b1 e8g8 f3e4 g7g6 h4h5 d5f4 h5g6 h7g6');
INSERT INTO evaluations VALUES(33739,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(33739,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(33739,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(33739,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(33739,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(33739,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(33739,36,'cp 60','d2d4 c7c6 c1f4 c8f5 f3e5 b8d7 e5c4 a5d8 c4e3 f5e6 a2a4 g7g6 a4a5 f6h5 f4g5 f8g7 f1e2 h7h6 g5h4 g7f6 h4f6 d7f6 e2h5 f6h5 e1g1 e8g8 f1e1 d8c7 a5a6 b7b6 d4d5 a8d8 d1f3 c6d5 c3d5 c7c5 a1d1 e6d5 d1d5 c5c7 d5d8 f8d8 e3d5');
INSERT INTO evaluations VALUES(33731,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(33731,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(33731,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(33731,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(33731,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(33731,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(33731,36,'cp 49','c8f5 c1d2 e7e6 f3e5 a5b6 d1f3 c7c6 e1c1 b8d7 f1d3 f5d3 f3d3 a8d8 d3g3 b6d4 e5d7 d4d7 d2g5 f8d6 g3h4 d7e7 c3e4 d6e5 c2c3 h7h6 d1d8 e8d8 f2f4 e5c7 e4f6 e7f8 f6h7 h6g5 h4g5 f7f6 h7f8 f6g5 f8e6 d8d7 f4f5 h8h2 h1h2 c7h2 e6g7 d7e7 g7h5 h2e5 c1c2');
INSERT INTO evaluations VALUES(32630,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(32630,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(32630,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(32630,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(32630,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(32630,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(32630,36,'cp 57','f3e5 c7c6 c1f4 b8d7 e5c4 a5d8 c4e3 f5e6 a2a4 g7g6 a4a5 f6h5 f4g5 h7h6 g5h4 f8g7 f1e2 g7f6 h4f6 d7f6 e2f3 a8c8 e1g1 e8g8 c3a4 b7b6 c2c4 h5g7 a5b6 a7b6 d4d5 c6d5 c4d5 e6d7 d1e2 b6b5 a4c3 b5b4 c3e4 g7f5 f1d1 d8c7 g2g3 f6e4 f3e4');
INSERT INTO evaluations VALUES(57750,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(57750,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(57750,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(57750,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57750,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(57750,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(57750,36,'cp 0','e7e5 g2g3 g8f6 f1g2 c7c6 d2d4 e5e4 d4d5 f8b4 c1d2 d8e7 b1c3 e8g8 g1h3 c6d5 c4d5 d7d6 c3e4 c8h3 e4f6 e7f6 g2h3 f6b2 a1c1 b8a6 c1c2 b4d2 d1d2 b2e5 e1g1 a6c5 f1c1 b7b6 e2e3 g7g6 a2a3 f8e8 d2d1 e5e4 c2c4');
INSERT INTO evaluations VALUES(57130,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(57130,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(57130,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(57130,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(57130,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(57130,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(57130,36,'cp 11','b1c3 g8f6 g1f3 b8c6 e2e4 f8b4 d2d3 d7d6 a2a3 b4c5 b2b4 c5b6 c1e3 c8g4 f1e2 g4f3 e2f3 e8g8 e1g1 c6d4 c3a4 d4f3 d1f3 b6d4 a1c1 f6d7 f3e2 d4e3 e2e3 h7h6 c4c5 a7a5 b4b5 d6c5 a4c5 d7c5 c1c5 d8e7 d3d4 e5d4 e3d4 f8e8 f1d1');
INSERT INTO evaluations VALUES(57124,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(57124,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(57124,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(57124,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(57124,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(57124,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(57124,36,'cp 22','g8f6 g1f3 b8c6 e2e4 f8b4 d2d3 d7d6 a2a3 b4c5 b2b4 c5b6 c1e3 c8g4 f1e2 e8g8 e1g1 g4f3 e2f3 c6d4 c3a4 d4f3 d1f3 b6e3 f2e3 f8e8 a4c3 e8e6 a3a4 c7c6 a4a5 a7a6 h2h3 d8e7 f3e2 f6d7 f1f5 e6f6 a1f1 h7h6 g1h2 a8f8 b4b5 f6f5');
INSERT INTO evaluations VALUES(49560,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(49560,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(49560,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(49560,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(49560,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(49560,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49560,36,'cp 9','g1f3 b8c6 e2e4 f8b4 d2d3 d7d6 a2a3 b4c5 b2b4 c5b6 c1e3 c8g4 f1e2 g4f3 e2f3 e8g8 e1g1 c6d4 c3a4 d4f3 d1f3 b6e3 f3e3 f8e8 f2f3 h7h5 a4c3 c7c6 d3d4 d8b6 f1d1 h5h4 h2h3 f6h5 g1h2 h5g3 b4b5 e5d4 e3d4 b6d4 d1d4');
INSERT INTO evaluations VALUES(49557,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(49557,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(49557,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(49557,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49557,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(49557,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(49557,36,'cp 10','b8c6 e2e4 f8b4 d2d3 d7d6 a2a3 b4c5 b2b4 c5b6 c1e3 c8g4 f1e2 g4f3 e2f3 e8g8 e1g1 c6d4 c3a4 d4f3 d1f3 b6e3 f3e3 f8e8 f2f3 h7h5 e3f2 c7c6 a1d1 d8e7 d3d4 f6d7 a4c3 d7b6 f2a2 e5d4 d1d4 a7a5 b4b5 a5a4 a2f2');
INSERT INTO evaluations VALUES(16449,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(16449,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(16449,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(16449,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16449,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(16449,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(16449,36,'cp 16','e2e4 f8b4 d2d3 d7d6 a2a3 b4c5 b2b4 c5b6 c1e3 c8g4 f1e2 g4f3 e2f3 e8g8 e1g1 c6d4 c3a4 d4f3 d1f3 b6e3 f3e3 f8e8 f2f3 f6h5 g2g3 h5f6 a4c3 c7c6 d3d4 d8e7 g1g2 e8c8 f1d1 a7a5 c4c5 d6c5 b4c5 e5d4 e3d4 c8c7');
INSERT INTO evaluations VALUES(16428,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(16428,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(16428,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(16428,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16428,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16428,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(16428,36,'cp 13','f8c5 f3e5 c6e5 d2d4 c5b4 d4e5 f6e4 d1d4 e4c3 b2c3 b4e7 f1d3 e8g8 e1g1 d7d6 c1f4 d6e5 f4e5 f8e8 f1e1 d8d4 c3d4 c7c6 c4c5 c8e6 a1b1 b7b5 c5b6 a7b6 d3e4 e6a2 e4c6 a2b1 e1b1 e8c8 c6a8 c8a8 g2g3 a8d8 b1b6 f7f6 b6b7 f6e5 b7e7 e5d4');
INSERT INTO evaluations VALUES(10156,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10156,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10156,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10156,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(10156,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(10156,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(10156,36,'cp 12','f3e5 c6e5 d2d4 c5b4 d4e5 f6e4 d1f3 d8h4 g2g3 e4g5 f3d5 h4e4 e1d1 b4c3 b2c3 e8g8 f2f4 c7c6 d5e4 g5e4 c1e3 d7d5 d1c2 c8f5 c2b2 f7f6 c4d5 c6d5 e5f6 f8f6 e3d4 f6c6 a1c1 e4c5 h1g1 f5e4 g3g4 c5e6 d4e5 a8c8 g1g3 g8f8 b2a1 a7a6 f4f5 e6c5');
INSERT INTO evaluations VALUES(10044,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(10044,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(10044,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10044,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10044,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(10044,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10044,36,'cp 13','c6e5 d2d4 c5b4 d4e5 f6e4 d1f3 d8h4 g2g3 e4g5 f3d5 h4e4 e1d1 b4c3 b2c3 e8g8 f2f4 c7c6 d5e4 g5e4 c1e3 d7d5 d1c2 c8f5 c2b2 f7f6 e5f6 f8e8 c4d5 e4f6 e3d4 f6d5 f1c4 b7b5 c4d5 c6d5 h1e1 a7a5 g3g4 f5d3 e1e5 e8e5 d4e5 a8a6');
INSERT INTO evaluations VALUES(55611,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(55611,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(55611,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(55611,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(55611,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(55611,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(55611,36,'cp 63','d2d4 e5d4 f3d4 g8f6 b1c3 f8e7 c1f4 e8g8 d1d2 b8c6 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1d3 c8e6 h1e1 a8d8 f4h4 h7h6 c3e4 f8e8 e4f6 e7f6 h4f6 g7f6 d3e4 c7c6 a2a3 e6g4 f2f3 g4e6 d1d8 e8d8 e1d1 d8d1 c1d1 f6f5 e4d3');
INSERT INTO evaluations VALUES(55586,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(55586,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(55586,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(55586,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(55586,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(55586,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(55586,36,'cp 56','e5d4 f3d4 f8e7 c1f4 b8c6 b1c3 g8f6 d1d2 e8g8 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1d3 c8e6 h1e1 a8e8 a2a3 c7c6 f2f3 e7c5 e1e5 e6d5 c3d5 e8e5 f4e5 f6d5 d3e4 c5e3 c1b1 f8d8 b1a1 e3c5 h2h4');
INSERT INTO evaluations VALUES(55685,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(55685,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(55685,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(55685,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(55685,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(55685,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(55685,36,'cp 63','f3d4 g8f6 b1c3 f8e7 c1f4 b8c6 d1d2 c6d4 d2d4 e8g8 e1c1 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 f4f3 c7c6 f3e2 a8d8 g2g3 e6e5 d1d8 f8d8 f2f3 d8e8 h1e1 a6a5 a2a3 e7e6 a3a4 g8h8 e2f2');
INSERT INTO evaluations VALUES(55672,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(55672,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(55672,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(55672,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(55672,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(55672,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(55672,36,'cp 57','g8f6 b1c3 f8e7 c1f4 b8c6 d1d2 c6d4 d2d4 e8g8 e1c1 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 f4f3 c7c6 f3e2 a8d8 f2f4 f6d5 c3d5 d8d5 d1d5 c6d5 g2g3 e7d6 h1e1 f8f6 c1b1 d6c6 h2h4 h7h6 a2a3 a6a5 e2e5 a5a4 g3g4');
INSERT INTO evaluations VALUES(47942,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(47942,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(47942,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(47942,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(47942,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(47942,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(47942,36,'cp 64','b1c3 f8e7 c1f4 b8c6 d1d2 c6d4 d2d4 a7a6 e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 f4c4 a8d8 c4e2 b7b5 g2g3 e6e5 f2f3 e7e6 a2a3 d8d1 h1d1 f8e8 c3e4 e6c6 b2b3 a6a5 e4f6 g7f6 f3f4 e5f4 e2g4 g8f8 d1d7 e8e1 c1b2');
INSERT INTO evaluations VALUES(47939,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47939,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(47939,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(47939,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(47939,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(47939,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(47939,36,'cp 59','f8e7 c1f4 b8c6 d1d2 c6d4 d2d4 a7a6 e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 f4f3 c7c6 f3e2 a8d8 g2g3 f6d5 c3e4 a6a5 h1e1 e6e5 a2a3 b7b5 f2f3 a5a4 c2c4 b5b4 c4d5 c6d5 a3b4 d5e4 d1d8 e7d8 e2c4 g8h8 c4e4');
INSERT INTO evaluations VALUES(40841,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(40841,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(40841,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(40841,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(40841,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(40841,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(40841,36,'cp 63','c1f4 b8c6 d1d2 c6d4 d2d4 a7a6 e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 f4f3 c7c6 f3e2 a8d8 d1d8 f8d8 h1e1 d8e8 a2a3 h7h6 f2f4 e7c5 g2g3 a6a5 e2d2 b7b5 c3e4 f6e4 e1e4 c5d5 d2d5 c6d5');
INSERT INTO evaluations VALUES(40845,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(40845,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(40845,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(40845,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(40845,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(40845,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(40845,36,'cp 65','b8c6 d1d2 c6d4 d2d4 a7a6 e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 f4f3 c7c6 f3e2 a8d8 d1d8 f8d8 h1e1 g8f7 g2g3 h7h6 c3b1 e6e5 b1d2 d8e8 e2c4 f7g6 f2f3 e7d6 a2a3 d6d5 c4e2 b7b5 h2h4 e8d8 e2e3 g6f7');
INSERT INTO evaluations VALUES(57761,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(57761,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(57761,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(57761,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(57761,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(57761,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(57761,36,'cp 14','d7d5 d2d4 g8f6 c2c4 e7e6 b1c3 d5c4 e2e3 a7a6 a2a4 c7c5 f1c4 c5d4 e3d4 b8c6 e1g1 f8b4 c1e3 e8g8 a1c1 c8d7 d4d5 f6d5 c3d5 e6d5 d1d5 d8e7 f1d1 d7e6 d5e4 a8e8 c4e6 e7e6 e4e6');
INSERT INTO evaluations VALUES(55140,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55140,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(55140,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55140,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(55140,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(55140,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(55140,36,'cp 26','d2d4 g8f6 c2c4 e7e6 b1c3 d5c4 e2e3 a7a6 a2a4 c7c5 f1c4 c5d4 e3d4 f8e7 e1g1 b8c6 d4d5 f6d5 c3d5 e6d5 c4d5 e8g8 c1e3 c8f5 d1b3 c6b4 f1d1 d8a5 f3d4 f5g6 d5f3 a8d8 h2h3 e7f6 a1c1 b4d5 e3d2 a5b6 f3d5 b6b3 d5b3');
INSERT INTO evaluations VALUES(55057,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(55057,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(55057,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(55057,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(55057,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55057,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(55057,36,'cp 27','e7e6 c2c4 g8f6 b1c3 c7c6 e2e3 b8d7 d1c2 f8d6 f1d3 d5c4 d3c4 c6c5 d4c5 d6c5 e1g1 a7a6 c4d3 b7b6 c3e4 c8b7 f1d1 d8e7 b2b3 f6e4 d3e4 d7f6 e4b7 e7b7 c1b2 e8g8 a1c1 a8c8 c2e2 f8d8 d1d8 c8d8');
INSERT INTO evaluations VALUES(55861,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(55861,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(55861,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(55861,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(55861,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(55861,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(55861,36,'cp 15','c2c4 g8f6 b1c3 c7c6 e2e3 b8d7 d1c2 f8d6 f1d3 e8g8 e1g1 d5c4 d3c4 b7b5 c4d3 c8b7 e3e4 e6e5 d4e5 d7e5 f3e5 d6e5 h2h3 b5b4 c3a4 e5d4 e4e5 d4e5 a4c5 b7c8 c1e3 f6d5 d3h7 g8h8 e3d2 e5c7 f1e1 d8h4 h7f5 c7b6 f5c8 a8c8 c5d3 a7a5');
INSERT INTO evaluations VALUES(55816,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(55816,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55816,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(55816,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(55816,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(55816,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(55816,36,'cp 8','g8f6 b1c3 f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 c7c6 f3d2 f6e4 h4d8 e4c3 b2c3 f8d8 d2b3 a8c8 f2f3 c6c5 e1f2 c5d4 c3d4 c8c2 h1c1 d8c8 c1c2 c8c2 a1c1 c2a2 c1c7 g8h7 c7b7 h7g6 h2h4 f5d3 b7d7 d3e2');
INSERT INTO evaluations VALUES(52385,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(52385,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(52385,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(52385,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(52385,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(52385,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(52385,36,'cp 38','b1c3 d7d5 d2d4 d5e4 c3e4 c8f5 e4g3 f5g6 g1f3 b8d7 h2h4 h7h6 h4h5 g6h7 f1d3 h7d3 d1d3 e7e6 c1f4 f8e7 e1c1 g8f6 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e1 d8d5 f3e5 d5e4 e1e4 f6e4 f4e3 f8d8 g2g4 a8c8 f2f3 e4f6 c2c4 f6d7 f3f4 b7b5 c4c5 d7e5 d4e5');
INSERT INTO evaluations VALUES(52376,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(52376,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52376,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52376,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52376,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(52376,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(52376,36,'cp 33','d7d5 d2d4 d5e4 c3e4 c8f5 e4g3 f5g6 g1f3 b8d7 h2h4 h7h6 f1d3 g6d3 d1d3 e7e6 c1f4 g8f6 e1c1 f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e1 d8d5 f3e5 d5e4 e1e4 f6e4 f4e3 f8d8 e5d3 e4f6 c2c3 b7b6 b1c2 a8c8 f2f3 h6h5 e3g5 f6d5 g5d2 f7f6');
INSERT INTO evaluations VALUES(52798,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52798,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(52798,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(52798,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52798,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(52798,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(52798,36,'cp 34','d2d4 d5e4 c3e4 c8f5 e4g3 f5g6 h2h4 h7h6 g1f3 b8d7 h4h5 g6h7 f1d3 h7d3 d1d3 e7e6 c1f4 f8b4 c2c3 b4e7 g3e4 d7f6 e4d2 f6d5 f4e5 g8f6 c3c4 d5b6 e1c1 b6d7 d2e4 e8g8 c1b1 d7e5 f3e5 f6e4 d3e4 e7d6 g2g4 d6e5 e4e5');
INSERT INTO evaluations VALUES(52784,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52784,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52784,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(52784,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(52784,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52784,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(52784,36,'cp 25','d5e4 c3e4 c8f5 e4g3 f5g6 g1f3 b8d7 h2h4 h7h6 f1d3 g6d3 d1d3 e7e6 c1f4 f8e7 e1c1 g8f6 f3e5 d7e5 d4e5 d8d3 d1d3 f6d5 f4d2 e8g8 c2c4 d5b6 b2b3 a8d8 d3d8 f8d8 c1c2 b6d7 f2f4 d7c5 h4h5 c5d3 g3e4 d3c5 e4c5 e7c5');
INSERT INTO evaluations VALUES(52900,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(52900,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(52900,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(52900,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52900,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(52900,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(52900,36,'cp 25','c3e4 c8f5 e4g3 f5g6 h2h4 h7h6 g1f3 b8d7 f1d3 g6d3 d1d3 g8f6 c1f4 e7e6 e1c1 f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e2 d8d5 f3e5 d5e4 e2e4 f6e4 f4e3 f8d8 c2c3 c6c5 d4c5 e7c5 e3c5 e4c5 b1c2 c5e4 f2f3 e4g3 h1e1');
INSERT INTO evaluations VALUES(52896,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52896,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(52896,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(52896,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(52896,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(52896,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(52896,36,'cp 45','g8f6 e4f6 e7f6 c2c3 f8d6 f1d3 e8g8 g1e2 b8d7 d1c2 h7h6 h2h3 f8e8 e1g1 d7f8 c1e3 c8e6 e2f4 d6f4 e3f4 d8d7 f1e1 a7a5 f4e3 a5a4 b2b3 f8g6 b3a4 g6h4 d3f1 e6d5 c3c4 d5e4 c2b2 a8a4 b2b3 a4a5 b3c3 d7c7 f2f3');
INSERT INTO evaluations VALUES(45037,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(45037,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(45037,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(45037,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(45037,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(45037,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(45037,36,'cp 49','e4f6 e7f6 c2c3 f8d6 f1d3 e8g8 d1c2 h7h6 g1e2 f8e8 e1g1 b8d7 c1e3 d7f8 h2h3 c8e6 e2f4 d6f4 e3f4 a7a5 b2b3 d8d7 f1e1 b7b5 a1d1 f6f5 d3f1 f8g6 c2d2 g6f4 d2f4 a8d8 c3c4 b5c4 b3c4 d7d6 f4e3 f5f4 e3e4 g7g6');
INSERT INTO evaluations VALUES(44848,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(44848,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(44848,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(44848,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(44848,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(44848,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(44848,36,'cp 30','e7f6 c2c3 f8d6 f1d3 e8g8 d1c2 h7h6 g1e2 f8e8 e1g1 b8d7 h2h3 d7f8 c1d2 c8e6 e2f4 d6f4 d2f4 a7a5 b2b3 d8d7 a1d1 a5a4 c3c4 g8h8 b3b4 d7d4 f4e3 d4e5 f1e1 e5h5 f2f4 f6f5 e3d4 f8d7 c4c5 d7f6 d4f6');
INSERT INTO evaluations VALUES(45829,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(45829,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(45829,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(45829,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(45829,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(45829,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(45829,36,'cp 38','c2c3 f8d6 f1d3 e8g8 d1c2 f8e8 g1e2 h7h6 c1e3 b8d7 e1g1 d7f8 h2h3 c8e6 e2f4 d6f4 e3f4 d8d7 b2b3 a7a5 a1d1 f6f5 f4h2 b7b5 f1e1 b5b4 c3c4 d7d4 d3e4 d4b6 e4f5 e6f5 c2f5 a8d8 h2g3 e8e1 d1e1 f8e6 g1h2 a5a4 b3a4');
INSERT INTO evaluations VALUES(45827,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(45827,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(45827,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(45827,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(45827,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(45827,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(45827,36,'cp 31','f8d6 f1d3 e8g8 g1e2 f8e8 d1c2 h7h6 e1g1 b8d7 h2h3 d7f8 c1e3 c8e6 e2f4 d6f4 e3f4 d8d7 f1e1 a7a5 a2a4 f8g6 f4g3 g6e7 f2f3 e7d5 e1e2 d5b6 b2b3 b6d5 a1e1 a8d8 d3h7 g8h8 h7e4 h8g8 g1h2 f6f5 e4d3 b7b5 a4b5 c6b5');
INSERT INTO evaluations VALUES(40529,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(40529,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(40529,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(40529,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(40529,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(40529,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(40529,36,'cp 46','f1d3 e8g8 g1e2 b8d7 d1c2 h7h6 e1g1 f8e8 h2h3 d7f8 c1e3 c8e6 e2f4 d6f4 e3f4 d8d7 b2b3 f6f5 a1d1 f8g6 f4h2 a8d8 g1h1 b7b5 f1e1 d7e7 f2f3 e7f6 h2g1 g6f4 c3c4 f4d3 c2d3 a7a6 g1f2 f6g5 d3c3 g5f6');
INSERT INTO evaluations VALUES(40530,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(40530,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(40530,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(40530,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(40530,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(40530,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(40530,36,'cp 47','e8g8 g1e2 b8d7 d1c2 h7h6 e1g1 f8e8 c1e3 d7f8 h2h3 c8e6 e2f4 d6f4 e3f4 d8d7 f1e1 a7a5 a2a4 f8g6 f4g3 g6e7 f2f3 e7d5 b2b3 a8d8 c2d2 d5b6 d3c2 c6c5 d2d3 g7g6 d3b5 d7b5 a4b5 c5d4 g3c7 d4d3 c7d8');
INSERT INTO evaluations VALUES(36128,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(36128,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(36128,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(36128,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(36128,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(36128,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(36128,36,'cp 36','g1e2 f8e8 d1c2 h7h6 e1g1 b8d7 c1e3 d7f8 h2h3 c8e6 e2f4 d6f4 e3f4 d8d7 b2b3 a7a5 f1e1 b7b5 a1d1 f8g6 f4h2 g6h4 e1e4 h4f5 a2a4 e8d8 e4e1 b5b4 d4d5 d7d5 h2c7 d8d7 d3c4 d5c5 d1d7 e6d7 c3b4 c5b4');
INSERT INTO evaluations VALUES(53074,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(53074,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(53074,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(53074,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(53074,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(53074,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(53074,36,'cp 20','d2d4 c5d4 f3d4 g8f6 b1c3 a7a6 f2f3 e7e5 d4b3 c8e6 c1e3 h7h5 f1e2 b8d7 a2a4 f8e7 e1g1 h5h4 a4a5 a8c8 d1d2 e8g8 f1d1 d8c7 a1c1 f6h5 c3d5 e6d5 d2d5 h5f4 d5d2 f4e2 d2e2 d7c5 e2c4 h4h3 b3c5 h3g2 g1g2 d6c5 c4a4');
INSERT INTO evaluations VALUES(53049,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(53049,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(53049,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53049,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(53049,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53049,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(53049,36,'cp 24','c5d4 f3d4 g8f6 b1c3 a7a6 c1e3 e7e5 d4b3 c8e6 f1e2 f8e7 f2f4 e5f4 e3f4 b8c6 d1d2 d6d5 e4d5 f6d5 c3d5 d8d5 d2d5 e6d5 e1c1 a8d8 f4c7 d8d7 c7b6 c6e5 b3c5 d5g2 h1g1 g2c6 d1d7 e5d7 c5d7 c6d7 g1g7 h7h5 g7g3 d7c6');
INSERT INTO evaluations VALUES(53137,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(53137,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53137,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53137,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(53137,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(53137,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53137,36,'cp 25','f3d4 g8f6 b1c3 a7a6 f1d3 g7g6 c1e3 f6g4 e3g5 h7h6 g5c1 f8g7 d3e2 g4f6 c1e3 b8c6 e1g1 e8g8 d1d2 c6d4 e3d4 b7b5 a2a4 c8d7 a4b5 a6b5 a1a8 d8a8 f2f3 b5b4 c3d5 f6d5 d4g7 g8g7');
INSERT INTO evaluations VALUES(53129,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(53129,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53129,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53129,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53129,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53129,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53129,36,'cp 19','g8f6 b1c3 a7a6 c1e3 e7e5 d4b3 f8e7 h2h3 h7h6 d1d2 c8e6 e1c1 b7b5 g2g4 b8d7 f2f4 b5b4 c3d5 f6e4 d2b4 e6d5 d1d5 e4g3 h1g1 g3f1 g1f1 e8g8 c1b1 a8b8 b4a5 d8c8 d5d1 c8c6 g4g5 h6g5 f4g5 b8c8 a5d2 d6d5 d2d5');
INSERT INTO evaluations VALUES(45250,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(45250,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(45250,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(45250,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(45250,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(45250,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45250,36,'cp 21','b1c3 a7a6 c1e3 e7e5 d4f3 f8e7 f1c4 e8g8 e1g1 b8c6 d1e2 c6a5 c4d3 c8d7 h2h3 a8c8 a1d1 a5c6 a2a3 b7b5 f1e1 h7h6 f3h4 c6a5 e2f3 a5c4 d3c4 c8c4 h4f5 d7f5 e4f5 d8c7 g2g4 d6d5 c3d5 f6d5 f3d5 c4c2');
INSERT INTO evaluations VALUES(45247,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(45247,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(45247,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45247,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45247,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(45247,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(45247,36,'cp 22','a7a6 c1e3 e7e5 d4f3 f8e7 f1c4 e8g8 c4b3 b8c6 d1e2 c6a5 e3g5 a5b3 a2b3 c8e6 g5f6 e7f6 e1g1 a8c8 f1d1 d8c7 f3e1 c7c6 e1d3 f6d8 d3b4 c6d7 b4d5 e6g4 f2f3 g4e6 d5e3 d8b6 g1h1 b6e3 e2e3');
INSERT INTO evaluations VALUES(43410,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43410,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(43410,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43410,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(43410,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(43410,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(43410,36,'cp 24','c1e3 e7e5 d4f3 f8e7 f1c4 c8e6 c4b3 e8g8 e1g1 b7b5 a2a4 b5b4 c3d5 f6d5 b3d5 e6d5 d1d5 b8d7 d5d3 d7f6 e3g5 f6h5 g5d2 a6a5 g2g3 h5f6 c2c3 b4c3 d2c3 d8c7 f3d2 f8c8 a1c1 c7b7 f1d1 b7c6 b2b3 h7h6 c3b2');
INSERT INTO evaluations VALUES(43405,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43405,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43405,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(43405,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(43405,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43405,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43405,36,'cp 13','e7e5 d4b3 c8e6 f1e2 f8e7 d1d3 b8c6 e1g1 a8c8 d3d2 e8g8 e2f3 c6b8 a2a4 b8d7 a4a5 f6g4 f3g4 e6g4 b3c1 g4e6 c1e2 d7f6 f1d1 c8c6 e2g3 d8c8 d2e2 f8d8 e3b6 d8d7 g3f1 c6c4 f1e3 f6e4 c3d5');
INSERT INTO evaluations VALUES(43449,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(43449,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(43449,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(43449,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(43449,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43449,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43449,36,'cp 20','d4b3 c8e6 f2f3 h7h5 f1e2 b8d7 f3f4 f6g4 e2g4 d8h4 e3f2 h4g4 f4f5 e6b3 a2b3 g4d1 a1d1 a8c8 e1e2 g7g6 f5g6 f7g6 h1f1 e8f7 e2d3 f7e6 c3d5 f8h6 c2c3 h8f8 c3c4 d7f6 d1a1 f6d5 e4d5 e6d7 f2b6 h5h4 f1f8 c8f8 b3b4 h6f4 h2h3');
INSERT INTO evaluations VALUES(43459,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(43459,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43459,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43459,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(43459,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(43459,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43459,36,'cp 9','c8e6 f2f3 h7h5 d1d2 b8d7 c3d5 e6d5 e4d5 g7g6 e1c1 d8c7 c1b1 f8g7 f1e2 e8g8 g2g4 f8c8 d1c1 a6a5 g4g5 f6e8 a2a4 d7b6 e3b6 c7b6 e2b5 e8c7 c2c4 c7a6 h2h4 a6c5 b3c5 c8c5 d2e1 a8f8 e1e4 c5c7 e4e2');
INSERT INTO evaluations VALUES(28543,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(28543,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(28543,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28543,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28543,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(28543,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(28543,36,'cp 25','f2f3 f8e7 d1d2 h7h5 c3d5 f6d5 e4d5 e6f5 f1e2 a6a5 a2a4 b8d7 e1g1 e8g8 f3f4 e5f4 e3f4 d7f6 e2b5 f5d7 d2d3 d7b5 d3b5 f8e8 h2h3 d8d7 b3a5 d7b5 a4b5 f6d5 f4d2 b7b6 a5c6 a8a1 c6e7 d5e7 f1a1 e7f5');
INSERT INTO evaluations VALUES(28545,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(28545,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(28545,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(28545,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(28545,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(28545,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(28545,36,'cp 24','f8e7 d1d2 h7h5 c3d5 f6d5 e4d5 e6f5 f1e2 a6a5 a2a4 b8d7 e1g1 e8g8 f3f4 e5f4 e3f4 e7f6 e2h5 f6b2 a1b1 b2f6 b3d4 f6d4 d2d4 d8h4 h5e2 a8c8 b1b7 c8c2 e2d1 c2a2 f4e3 h4e4 f1f2 a2a3 d4e4 f5e4 b7d7 a3e3 d7d6 e4d3 f2d2 e3e1 g1f2');
INSERT INTO evaluations VALUES(28184,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(28184,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(28184,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28184,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(28184,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(28184,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(28184,36,'cp 17','d1d2 e8g8 e1c1 b8d7 g2g4 b7b5 g4g5 f6h5 c3d5 e6d5 e4d5 f7f6 h2h4 h5g3 h1g1 g3f1 d1f1 a6a5 c1b1 a5a4 b3c1 d7b6 a2a3 b6c4 d2d3 f6f5 c1a2 f5f4 e3f2 d8d7 f1d1 e7d8 a2b4 d7f5 d3f5 f8f5 b4c6 g8f8 g1e1 d8a5 e1e2 g7g6 b1a2 a5b6 f2b6 c4b6');
INSERT INTO evaluations VALUES(28185,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(28185,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(28185,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(28185,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(28185,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(28185,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(28185,36,'cp 32','b8d7 g2g4 e8g8 g4g5 f6h5 e1c1 b7b5 c1b1 d7b6 b3a5 d8c7 c3d5 b6d5 e4d5 e6f5 h2h4 h5g3 h1g1 g3f1 d1f1 f5g6 g1h1 g6c2 d2c2 c7a5 f1c1 f7f5 h4h5 f5f4 e3d2 a5a4 c2a4 b5a4 g5g6 e7f6 c1c6 h7g6 h5g6');
INSERT INTO evaluations VALUES(23939,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(23939,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(23939,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23939,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(23939,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(23939,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(23939,36,'cp 12','g2g4 e8g8 e1c1 b7b5 g4g5 f6h5 c3d5 e6d5 e4d5 f7f6 g5f6 e7f6 c1b1 d7b6 b3a5 h5f4 a5c6 d8c7 f1d3 b6d5 e3f4 d5f4 d3e4 d6d5 e4d5 f4d5 d2d5 g8h8 c6b4 a6a5 d5c6 c7c8 c6c8 a8c8 b4d3 h8g8 d1e1 c8d8 d3e5 d8d2 a2a3 f8c8 c2c3 c8c5 e5g4');
INSERT INTO evaluations VALUES(57119,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57119,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57119,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57119,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(57119,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(57119,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57119,36,'cp 12','g8f6 b1c3 b8c6 d2d3 c6a5 g1e2 f8c5 e1g1 c7c6 e2g3 e8g8 a2a3 a5c4 d3c4 c5d4 d1d3 d7d6 c1e3 d4e3 d3e3 c8e6 b2b3 d8b6 e3b6 a7b6 f1d1 f8d8 f2f3 g8f8 d1d2 f8e7 a3a4 g7g6 a1d1 h7h5 d2d3 h5h4 g3f1 f6h5 g2g3 h4g3');
INSERT INTO evaluations VALUES(49538,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49538,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(49538,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(49538,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(49538,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(49538,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(49538,36,'cp 14','b1c3 f8c5 g1f3 d7d6 d2d4 e5d4 f3d4 e8g8 e1g1 c8g4 f2f3 b8c6 c1e3 g4e6 c4e6 f7e6 d1d2 c6d4 e3d4 c5d4 d2d4 d6d5 a1d1 c7c6 f1e1 d8c7 e4d5 e6d5 c3e2 c7b6 e2f4 f8e8 e1e8 a8e8 g1f2 g7g5 d4b6 a7b6 f4d3 e8e7 a2a4');
INSERT INTO evaluations VALUES(49530,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(49530,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49530,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(49530,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49530,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49530,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(49530,36,'cp 12','f8c5 g1f3 d7d6 d2d4 e5d4 f3d4 e8g8 e1g1 c8g4 f2f3 b8c6 c1e3 g4e6 c4b5 c5d4 e3d4 c6d4 d1d4 c7c5 d4d2 c5c4 d2d4 a7a6 b5c4 e6c4 d4c4 d8b6 g1h1 b6b2 c3a4 b2b5 c4b3 a8c8 a1c1 b5b3 c2b3 d6d5 c1c8 f8c8 a4b6 c8c2 b6d5');
INSERT INTO evaluations VALUES(42548,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42548,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42548,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42548,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(42548,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(42548,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(42548,36,'cp 0','g1f3 d7d6 d2d4 e5d4 f3d4 e8g8 e1g1 f8e8 c1g5 b8d7 a2a4 h7h6 g5f6 d8f6 d4b3 d7e5 c4e2 e5g6 b3c5 d6c5 c3d5 f6d8 d5c3');
INSERT INTO evaluations VALUES(42544,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42544,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42544,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(42544,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(42544,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42544,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(42544,36,'cp 0','d7d6 d2d4 e5d4 f3d4 e8g8 c1g5 f8e8 e1g1 b8d7 f2f4 d6d5 e4d5 d7b6 c4b3 h7h6 g5h4 a7a5 a2a4 c7c6 g1h1 b6d5 c3d5 c6d5 d4f3 a8a6 d1d3 e8e3 d3b5 d8c7 f3e5 a6b6 b5a5 c5b4 a5a7 b4c5');
INSERT INTO evaluations VALUES(41855,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41855,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41855,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41855,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(41855,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41855,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41855,36,'cp 0','e1g1 e8g8 d2d4 e5d4 f3d4 b8d7 d4b3 c5b6 a2a4 a7a5 c1g5 f8e8 g1h1 h7h6 g5h4 d7e5 h4f6 d8f6 c3d5 f6h4 d5b6 e5g4 h2h3 c7b6 h1g1 g4f6 f2f3 c8h3 d1d6 h3e6 c4e6 f7e6 a1d1 f6h5 d6h2 h4g5 f3f4 g5g4 f1f3');
INSERT INTO evaluations VALUES(41854,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(41854,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41854,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41854,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41854,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41854,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41854,36,'cp 0','e8g8 d2d4 e5d4 f3d4 f8e8 c1g5 b8d7 d4b3 c5b6 a2a4 a7a5 g1h1 h7h6 g5h4 d7e5 h4f6 d8f6 c3d5 f6h4 d5b6 e5g4 h2h3 c7b6 h1g1 g4e5 c4d5 c8h3 g2h3 h4h3 f2f4 h3g3 g1h1 g3h3');
INSERT INTO evaluations VALUES(36990,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36990,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36990,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36990,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36990,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36990,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36990,36,'cp 0','d2d4 e5d4 f3d4 b8d7 a2a4 d7e5 c4b3 e5c6 d4c6 b7c6 a4a5 d8e7 d1d3 a8b8 h2h3 c8e6 b3a4 b8b4 b2b3 e6d7 c1e3 f6e4 c3e4 c5e3 d3e3 b4e4 e3a7 d7h3 g2h3 e7g5 g1h1 g5h4 h1h2 h4f4 h2g2 f4g5');
INSERT INTO evaluations VALUES(36998,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36998,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36998,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36998,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36998,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36998,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36998,36,'cp 0','e5d4 f3d4 f8e8 c1g5 b8d7 d4b3 c5b6 a2a4 a7a5 g1h1 h7h6 g5h4 d7e5 h4f6 d8f6 c3d5 f6h4 d5b6 e5g4 h2h3 c7b6 h1g1 g4e5 c4d5 c8h3 g2h3 h4h3 f2f4 h3g3 g1h1 g3h3');
INSERT INTO evaluations VALUES(49649,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(49649,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(49649,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(49649,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(49649,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(49649,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(49649,36,'cp 30','f3e5 d7d6 e5f3 f6e4 d2d4 d6d5 f1d3 f8e7 e1g1 c8f5 f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 f5d3 d1d3 a8e8 c1f4 e7d6 c3c4 d5a5 f4e3 a5a3 d3f5 b7b6 c4c5 b6c5 d4c5 d6e5 f3e5 c6e5 e3f4 g7g6 f5c2 e5d3 e1e3 e8e3 f4e3 a3b2 c2b2 d3b2 a1b1');
INSERT INTO evaluations VALUES(49491,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(49491,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(49491,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(49491,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(49491,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(49491,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(49491,36,'cp 48','d7d6 e5f3 f6e4 d2d4 d6d5 f1d3 f8b4 c2c3 b4d6 e1g1 e8g8 c3c4 c7c6 f1e1 f8e8 d1b3 b8a6 c4d5 c6d5 b1c3 c8e6 a2a3 h7h6 c1e3 e8e7 c3b5 a6b8 b5d6 e4d6 f3e5 e7c7 a1c1 b8d7 c1c7 d8c7 e1c1 c7d8 b3b4 d7b6 e3f4 e6f5 b2b3 f5d3 e5d3');
INSERT INTO evaluations VALUES(47769,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47769,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(47769,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(47769,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(47769,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(47769,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(47769,36,'cp 30','e5f3 f6e4 d2d4 d6d5 f1d3 f8d6 e1g1 e8g8 c2c4 c7c6 f1e1 c8f5 d1b3 d8d7 b1c3 e4c3 d3f5 d7f5 b2c3 b7b6 f3h4 f5h5 g2g3 b8d7 c4d5 c6d5 h4g2 a8d8 c1f4 d7f6 f4d6 d8d6 e1e5 h5g4 g2f4 g4d7 a1e1 f8e8 h2h4 h7h6');
INSERT INTO evaluations VALUES(47957,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(47957,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(47957,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(47957,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(47957,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47957,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(47957,36,'cp 33','f6e4 d2d4 d6d5 f1d3 f8d6 e1g1 e8g8 f1e1 c8f5 c2c4 c7c6 d1b3 d8d7 a2a3 f8e8 b1c3 b8a6 c4d5 e4c3 e1e8 a8e8 b3c3 f5d3 c3d3 c6d5 d3b3 b7b6 c1d2 h7h6 a1c1 a6c7 b3c2 c7e6 g2g3 g7g5 c2f5 d7b5 g1g2 e6g7 f5f6 e8e6 f6d8 b5e8 c1c8 g8f8');
INSERT INTO evaluations VALUES(48020,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(48020,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(48020,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(48020,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(48020,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(48020,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48020,36,'cp 35','d2d4 d6d5 f1d3 f8d6 e1g1 e8g8 c2c4 c7c6 d1c2 h7h6 f1e1 c8e6 c4d5 c6d5 d3e4 d5e4 c2e4 b8c6 b1c3 c6b4 a2a3 b4d5 c1d2 f8e8 h2h3 a8c8 a1c1 d6f8 f3e5 d5c3 d2c3 e6d5 e4g4 d5e6 g4h5 d8d5 f2f4 f7f6 f4f5 f6e5 f5e6 d5e6');
INSERT INTO evaluations VALUES(48012,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(48012,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(48012,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(48012,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(48012,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48012,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(48012,36,'cp 16','d6d5 f1d3 f8d6 e1g1 e8g8 f1e1 c8f5 c2c4 c7c6 d1b3 d8d7 b1c3 e4c3 d3f5 d7f5 b2c3 b7b6 c4d5 c6d5 b3b5 f5d7 a2a4 d7b5 a4b5 a7a5 b5a6 b8a6 c1a3 d6a3 a1a3 a6c7 a3b3 b6b5 g2g3 f7f6 f3h4 g7g6 h4g2 f8e8 e1e8 c7e8 g2e3 e8c7 g1g2 g8f7 g2f3 a8a2 h2h4');
INSERT INTO evaluations VALUES(49036,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(49036,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(49036,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(49036,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(49036,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(49036,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(49036,36,'cp 22','f1d3 f8d6 e1g1 e8g8 c2c4 c7c6 d1c2 h7h6 f1e1 c8e6 c4d5 c6d5 d3e4 d5e4 c2e4 b8c6 b1c3 f8e8 d4d5 e6d7 e4d3 c6e5 f3e5 d6e5 c1d2 d8b6 h2h3 a8d8 a1b1 e5c7 e1e8 d8e8 c3e4 d7f5 b1e1 e8e4 e1e4 f5e4 d3e4 b6b2 d2f4 c7f4 e4f4 g7g5 f4c7 b2a2 d5d6 a2a1 g1h2 a1e5 g2g3');
INSERT INTO evaluations VALUES(49035,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(49035,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(49035,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(49035,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(49035,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(49035,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(49035,36,'cp 25','f8d6 e1g1 e8g8 c2c4 c7c6 d1c2 h7h6 f1e1 c8e6 c4d5 c6d5 d3e4 d5e4 c2e4 b8c6 b1c3 c6b4 a2a3 b4d5 c1d2 a8c8 h2h3 d6b8 f3e5 f8e8 a1c1 d5c3 c1c3 c8c3 d2c3 e6d5 e4g4 d5e6 g4d1 e6d5 e1e3 b8c7 d1h5 d8d6 c3b4 d6d8 b4d2 d8f6');
INSERT INTO evaluations VALUES(41749,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41749,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(41749,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(41749,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(41749,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(41749,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41749,36,'cp 28','e1g1 e8g8 c2c4 c7c6 d1c2 h7h6 f1e1 c8e6 c4d5 c6d5 d3e4 d5e4 c2e4 b8c6 b1c3 c6b4 a2a3 b4d5 c1d2 f8e8 h2h3 a8c8 a1c1 d6f8 f3e5 d5c3 d2c3 d8d5 e4d3 f8d6 c1d1 a7a6 f2f4 f7f6 f4f5 e6f7 e5f7 d5f7 e1e8 c8e8 d4d5 f7c7');
INSERT INTO evaluations VALUES(41748,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(41748,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41748,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(41748,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41748,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(41748,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41748,36,'cp 27','e8g8 c2c4 c7c6 d1c2 h7h6 f1e1 c8e6 c4d5 c6d5 d3e4 d5e4 c2e4 e6d5 e4g4 d8d7 g4d7 b8d7 b1c3 d5f3 g2f3 a8c8 c1d2 f8d8 c3b5 d6b8 a1c1 c8c1 d2c1 f7f5 b2b3 d7f6 e1e7 a7a6 b5c3 d8d7 e7d7 f6d7 c3d5 g8f7 c1d2 f7e6 d5c3');
INSERT INTO evaluations VALUES(36957,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(36957,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36957,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(36957,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36957,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36957,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(36957,36,'cp 24','c2c4 c7c6 d1c2 h7h6 f1e1 c8e6 c4d5 c6d5 d3e4 d5e4 c2e4 e6d5 e4g4 d8d7 g4d7 b8d7 b1c3 d5f3 g2f3 a8c8 c1d2 f8d8 a1c1 d7f8 c3e4 c8c1 e1c1 d6b8 c1c4 f8g6 d2e3 b7b6 e4g3 g6e7 g3e2 f7f6 e2c3 g8f7 b2b3 f7e6 f3f4 e7f5');
INSERT INTO evaluations VALUES(36956,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36956,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36956,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36956,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36956,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36956,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(36956,36,'cp 21','c7c6 d1c2 h7h6 f1e1 c8e6 c4d5 c6d5 d3e4 d5e4 c2e4 e6d5 e4g4 d8d7 g4d7 b8d7 b1c3 d5f3 g2f3 a8c8 c1d2 f8d8 c3e4 d7f8 a1c1 c8c1 e1c1 d6b8 c1c4 f8g6 d2e3 f7f6 e4c3 g8f7 c4a4 g6h4 f3f4 h4f5 a4b4 d8d7 d4d5');
INSERT INTO evaluations VALUES(36135,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(36135,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36135,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(36135,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(36135,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36135,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(36135,36,'cp 28','f1e1 c8f5 f3g5 f8e8 d3e4 d5e4 b1c3 b8a6 g5e4 f5e4 c3e4 d6h2 g1h2 d8h4 h2g1 e8e4 c1e3 a6c7 d1b3 b7b6 f2f3 e4e6 e3f2 h4g5 e1e6 c7e6 b3e3 g5e3 f2e3 a8d8 a1d1 g8f8 g1f2 f7f5 g2g3 f8f7 f2e2 h7h5 d4d5 c6d5 d1d5 d8d5');
INSERT INTO evaluations VALUES(22766,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(22766,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(22766,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(22766,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(22766,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(22766,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(22766,36,'cp 12','e5d4 f3d4 g8f6 d4c6 b7c6 b1c3 f8b4 f1d3 e8g8 e1g1 d7d5 e4d5 c6d5 h2h3 c7c6 d1f3 f8e8 c1g5 h7h6 g5h4 a8b8 f1e1 e8e1 a1e1 c8e6 a2a3 b4e7 b2b4 a7a5 b4a5 d8a5 d3f5 a5a3 f5e6 f7e6 e1e6 b8f8 f3e2 e7d6 h4f6 f8f6 e6f6');
INSERT INTO evaluations VALUES(22944,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(22944,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(22944,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(22944,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(22944,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(22944,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(22944,36,'cp 28','f3d4 g8f6 d4c6 b7c6 e4e5 d8e7 d1e2 f6d5 g2g3 f7f6 c2c4 c8a6 b2b3 f6e5 f1g2 d5f6 e1g1 e5e4 c1a3 e7e6 a3f8 h8f8 b1d2 e8c8 a1e1 f8e8 g2e4 d7d5 e4f3 e6e2 e1e2 e8e2 f3e2 d5c4 d2c4 a6c4 e2c4 d8d2 a2a3 c8d7 f1e1 c6c5 e1e2 d2e2 c4e2 g7g5 e2c4 d7d6 g1g2 g5g4');
INSERT INTO evaluations VALUES(22913,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(22913,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(22913,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(22913,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(22913,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(22913,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(22913,36,'cp 5','g8f6 d4c6 b7c6 f1d3 d7d5 b1c3 f8b4 e4d5 c6d5 h2h3 e8g8 e1g1 c7c6 d1f3 h7h6 c1f4 a7a5 a2a3 b4d6 f1e1 c8e6 a1d1 d6f4 f3f4 d8b8 f4b8 a8b8 c3a4 f8e8 b2b3 g7g6 a4c5 g8g7 f2f3 f6d7 c5d7 e6d7');
INSERT INTO evaluations VALUES(16678,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16678,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16678,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(16678,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(16678,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(16678,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(16678,36,'cp 12','d4c6 b7c6 f1d3 d7d5 b1c3 f8b4 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 f8e8 c1g5 h7h6 g5h4 b4d6 a1e1 c8e6 e1e2 a8b8 b2b3 b8b4 h4g3 d8c7 g3d6 c7d6 f1e1 b4b8 c3a4 c6c5 d3f5 b8c8 a4b2 e6f5 e2e8 c8e8 e1e8 f6e8 f3f5 g7g6');
INSERT INTO evaluations VALUES(16233,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16233,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16233,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16233,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16233,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16233,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16233,36,'cp 0','b7c6 f1d3 d7d5 b1c3 f8b4 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 f8e8 c1f4 a7a5 c3a4 c8e6 b2b3 f6e4 d3e4 d5e4 f3g3 d8f6 a1d1 h7h5 f4e3 a8c8 a4b6 c8b8 b6a4');
INSERT INTO evaluations VALUES(11942,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(11942,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(11942,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(11942,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(11942,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(11942,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(11942,36,'cp 4','f1d3 d7d5 b1c3 f8b4 e4d5 e8g8 e1g1 c6d5 h2h3 f8e8 d1f3 c7c6 c1g5 b4d6 a1e1 c8e6 c3e2 h7h6 g5f6 d8f6 f3f6 g7f6 e2g3 g8f8 b2b3 a7a5 g3f5 d6b4 e1e2 h6h5 g2g4 e6f5 e2e8 a8e8 d3f5 e8e5 g1g2 h5g4 h3g4');
INSERT INTO evaluations VALUES(11940,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11940,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11940,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(11940,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(11940,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11940,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11940,36,'cp 3','d7d5 b1c3 f8b4 e4d5 e8g8 e1g1 c6d5 h2h3 f8e8 d1f3 c7c6 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 b2b3 d8c7 f4d6 c7d6 f1e1 h7h6 e1e2 e6d7 a4b6 e8e2 f3e2 a8e8 e2d2 d6b4 b6d7 f6d7 c2c3 b4b6 d1e1 g8f8 e1e8 f8e8 d2e2 e8d8 d3f5 g7g6 f5d7 d8d7');
INSERT INTO evaluations VALUES(11828,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(11828,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11828,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11828,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11828,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(11828,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(11828,36,'cp 5','b1c3 f8b4 e4d5 c6d5 e1g1 e8g8 h2h3 c7c6 d1f3 f8e8 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 b2b3 d8c7 f4d6 c7d6 f1e1 h7h6 e1e3 e6d7 f3g3 d6g3 e3g3 g8f8 g1f1 a8b8 a2a3 d7c8 d1e1 g7g6 g3f3 e8e1 f1e1 f8e7 f3f4');
INSERT INTO evaluations VALUES(55047,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(55047,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55047,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(55047,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55047,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55047,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(55047,36,'cp 8','c7c5 e2e3 b8c6 g1f3 g8f6 b1d2 d8b6 d4c5 b6b2 a1b1 b2c3 f1b5 e7e6 e1g1 f8e7 d2b3 e8g8 d1d3 f6e4 b5c6 b7c6 f3e5 e7f6 f2f3 f6e5 f3e4 f8d8 f1d1 f7f6 a2a3 a7a5 d3c3 e5c3 b3d4 c3d4 e3d4 d5e4 f4d6 c8a6 b1b6 d8c8 d1e1 a6c4 e1e4 c4d5 e4e1 a5a4');
INSERT INTO evaluations VALUES(53252,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53252,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53252,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(53252,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53252,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(53252,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(53252,36,'cp 0','e2e3 b8c6 f1b5 c5d4 e3d4 g8f6 c2c3 c8f5 g1f3 e7e6 d1a4 d8b6 f3e5 a8c8 b1d2 f8e7 a4b3 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(53250,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53250,36,'cp 0','b8c6 f1b5 c5d4 e3d4 g8f6 c2c3 c8f5 b1d2 e7e6 g1f3 d8b6 d1b3 f8e7 b5e2 e8g8 f3e5 h7h6 b3b6 a7b6 h2h4 f5h7 e2b5 c6a7 b5e2');
INSERT INTO evaluations VALUES(19777,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19777,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19777,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19777,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19777,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(19777,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19777,36,'cp 0','f1b5 c5d4 e3d4 g8f6 c2c3 d8b6 d1b3 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(57026,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(57026,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(57026,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(57026,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(57026,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(57026,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(57026,36,'cp 37','e2e4 d7d5 b1c3 g8f6 c1g5 d5e4 c3e4 f8e7 g5f6 e7f6 g1f3 e8g8 d1d2 b7b6 e1c1 c8b7 d2e3 f6e7 f3e5 b8c6 f1b5 c6b4 a2a3 c7c6 b5c4 b4d5 e3g3 d8c7 h1e1 f7f6 e5d3 g8h8 g3c7 d5c7');
INSERT INTO evaluations VALUES(22779,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(22779,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22779,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(22779,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22779,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22779,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22779,36,'cp 5','g8f6 d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 c1f4 f8e8 d1f3 b4d6 b2b3 c8e6 c3e2 a7a5 f1e1 d8b8 f4d6 b8d6 f3f4 d6f4 e2f4 g8f8 f2f3 g7g5 f4e2 c6c5 e2c3 a8b8 a1d1 c5c4');
INSERT INTO evaluations VALUES(16507,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(16507,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(16507,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(16507,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(16507,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(16507,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(16507,36,'cp 12','d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 c6d5 e1g1 e8g8 h2h3 c7c6 d1f3 f8e8 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 b2b3 d8c7 f4d6 c7d6 d1e1 e6d7 e1e8 a8e8 f3g3 d6g3 f2g3 d7e6 a4c5 f6d7 c5d7 e6d7 g1f2 h7h6');
INSERT INTO evaluations VALUES(16482,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(16482,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(16482,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(16482,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(16482,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16482,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(16482,36,'cp 4','e5d4 f3d4 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 a7a5 d1f3 c7c6 c3a4 c8e6 c1f4 f8e8 a1d1 f6d7 a2a3 b4f8 c2c4 d7b6 a4c3 d8f6 c4d5 b6d5 f3e4 g7g6 f4e5 f6g5 h3h4 g5g4 c3d5 c6d5 e4d4 g4d4 e5d4 a5a4');
INSERT INTO evaluations VALUES(16691,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(16691,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(16691,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(16691,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(16691,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(16691,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(16691,36,'cp 5','f3d4 f8b4 d4c6 b7c6 f1d3 e8g8 e1g1 d7d5 e4d5 c6d5 d1f3 c8g4 f3g3 c7c5 c1g5 b4c3 b2c3 c5c4 g5f6 d8f6 g3g4 c4d3 c2d3 f6c3 g4f3 a8c8 g2g3 a7a5 f1e1 c3d4 f3e3 d4e3 e1e3 c8c3 a1b1 c3c2 a2a4 c2a2 g1g2 a2a4 b1b5 a4a1');
INSERT INTO evaluations VALUES(16673,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(16673,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(16673,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(16673,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(16673,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(16673,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(16673,36,'cp 0','f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 f8e8 d1f3 c7c6 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 b2b3 d8c7 f4d6 c7d6 f1e1 h7h6 e1e2 a8d8 d1e1 g8f8 e2e5 d8b8 e5e2');
INSERT INTO evaluations VALUES(10473,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10473,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10473,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10473,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10473,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10473,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10473,36,'cp 0','d4c6 b7c6 f1d3 e8g8 e1g1 d7d5 e4d5 c6d5 h2h3 f8e8 d1f3 c7c6 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 b2b3 d8c7 f4d6 c7d6 f1e1 h7h6 e1e2 a8d8 d1e1 g8f8 a4b2 d6a3 b2a4');
INSERT INTO evaluations VALUES(9961,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(9961,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(9961,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9961,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9961,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9961,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9961,36,'cp 0','b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 f8e8 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 f1e1 d8c7 f4d6 c7d6 b2b3 h7h6 e1e2 a8d8 d1e1 g8f8 a4b2 d6a3 b2a4');
INSERT INTO evaluations VALUES(7616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7616,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7616,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7616,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7616,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7616,36,'cp 0','f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 f8e8 d1f3 c7c6 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 b2b3 d8c7 f4d6 c7d6 f1e1 h7h6 e1e2 a8d8 d1e1 g8f8 a4b2 d6a3 b2a4');
INSERT INTO evaluations VALUES(7618,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7618,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7618,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7618,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7618,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7618,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7618,36,'cp 0','d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 f8e8 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 b2b3 d8c7 f4d6 c7d6 f1e1 h7h6 e1e2 a8d8 d1e1 g8f8 a4b2 d6b4 b2a4');
INSERT INTO evaluations VALUES(7550,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7550,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(7550,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(7550,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(7550,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(7550,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(7550,36,'cp 8','e4d5 e8g8 e1g1 c6d5 c1g5 c7c6 d1f3 b4d6 h2h3 f8e8 f1e1 c8e6 c3e2 d6e5 a1b1 d8d6 b2b3 a7a5 g5f4 e5f4 e2f4 e6d7 c2c4 d5d4 f4h5 f6h5 f3h5 g7g6 e1e8 d7e8 h5g5 c6c5 b1e1 g8g7 g5e5 d6e5 e1e5 a8c8 f2f4 g7f6 g1f2 a5a4 b3a4 e8d7 a2a3 h7h6');
INSERT INTO evaluations VALUES(7548,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7548,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(7548,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(7548,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(7548,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(7548,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(7548,36,'cp 5','e8g8 e1g1 c6d5 c1g5 c7c6 d1f3 b4d6 h2h3 h7h6 g5f4 f8e8 f1e1 c8e6 c3e2 d6f4 e2f4 e6d7 e1e8 d8e8 a1d1 e8e5 c2c4 d5c4 d3c4 e5b2 c4b3 a8e8 f3d3 b2e5 d3c4 f6d5 f4d5 c6d5 c4d5 e5d5 d1d5 d7e6 d5a5 e6b3 a2b3 e8e1 g1h2 e1e2 f2f4 e2e3 a5a7 e3b3');
INSERT INTO evaluations VALUES(6437,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(6437,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(6437,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(6437,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6437,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(6437,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(6437,36,'cp 11','e1g1 g8f6 b1c3 e8g8 d2d3 d7d6 c3a4 c6a5 a4c5 a5c4 d3c4 d6c5 f3e5 f6e4 f1e1 d8d1 e1d1 e4d6 b2b3 f8d8 e5d3 b7b6 f2f3 c8f5 g2g4 f5g6 g1g2 a7a5 h2h4 d6e8 c1b2 a5a4 h4h5 g6d3 d1d3 d8d3 c2d3 f7f6 d3d4 c5d4 b2d4 c7c5');
INSERT INTO evaluations VALUES(6436,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(6436,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(6436,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6436,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(6436,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(6436,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(6436,36,'cp 6','g8f6 b1c3 e8g8 d2d3 d7d6 c3a4 c6a5 a4c5 a5c4 d3c4 d6c5 f3e5 f6e4 f1e1 d8d1 e1d1 e4d6 b2b3 f8d8 c1f4 c8f5 e5d3 b7b6 f2f3 f7f6 g2g4 f5g6 h2h4 d6e8 d3e1 h7h5 a2a4 h5g4 f3g4 a7a5 a1c1 g6e4 g4g5 e4f5 h4h5 d8d1 c1d1 f6g5');
INSERT INTO evaluations VALUES(10128,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(10128,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(10128,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(10128,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10128,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10128,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(10128,36,'cp 9','b1c3 e8g8 d2d3 d7d6 c3a4 c6a5 a4c5 a5c4 d3c4 d6c5 f3e5 f6e4 f1e1 d8d1 e1d1 e4d6 b2b3 f7f6 e5d3 b7b6 c1f4 f8d8 f2f3 g7g5 f4d6 d8d6 d3f2 c8f5 d1d6 c7d6 a1d1 a8e8 f2g4 f5g4 f3g4 e8e6 h2h3 g8f7 g1f2 f7e7 d1d5 e6e5 d5e5 d6e5 f2e3 e7e6 e3f3 e6f7 f3e4 f7e6');
INSERT INTO evaluations VALUES(10088,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(10088,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(10088,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(10088,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(10088,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(10088,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(10088,36,'cp 12','e8g8 d2d3 d7d6 c3a4 c6a5 a4c5 a5c4 c5b3 c4b6 b3d2 b6d7 f1e1 f8e8 d2f1 b7b5 f1g3 c7c5 h2h3 h7h6 c2c3 a7a5 d3d4 a5a4 d4d5 d7b6 d1e2 c8d7 f3d2 g8h7 d2f1 f6g8 f1e3 g8e7 c1d2 e7g6 b2b3 g6f4 e2f3 d8f6');
INSERT INTO evaluations VALUES(4317,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(4317,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(4317,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(4317,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(4317,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4317,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4317,36,'cp 8','d2d3 d7d6 c3a4 c6a5 a4c5 a5c4 c5b3 c4b6 f1e1 b6d7 b3d2 f8e8 d2f1 h7h6 f1g3 c7c5 c1e3 d7f8 h2h3 f8g6 a2a4 b7b6 c2c4 c8e6 d1d2 a7a5 f3h2 f6d7 b2b3 g8h7 a1d1 d7b8 f2f4 e5f4 e3f4 g6f4 d2f4 b8c6');
INSERT INTO evaluations VALUES(4322,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4322,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(4322,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4322,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(4322,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4322,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(4322,36,'cp 9','d7d6 c3a4 c6a5 a4c5 a5c4 c5b3 c4b6 b3d2 b6d7 f1e1 f8e8 d2f1 c7c5 f1g3 h7h6 c1e3 d7f8 h2h3 f8g6 c2c4 c8e6 d1d2 a7a5 g1h2 g8h7 f3g1 a8b8 a2a4 f6g8 g1e2 g8e7 e1f1 e7c6 f2f4 e5f4 e2f4 d8d7 g3h5 g6f4 h5f4 c6e5');
INSERT INTO evaluations VALUES(3821,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(3821,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(3821,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(3821,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(3821,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(3821,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(3821,36,'cp 16','c3a4 c6a5 a4c5 a5c4 c5b3 c4b6 b3d2 b6d7 f1e1 f8e8 d2f1 c7c5 f1g3 d7f8 h2h3 h7h6 c1e3 f8g6 c2c4 g8h7 d1d2 a7a5 e1f1 f6g8 f3h2 c8e6 f2f4 g6f4 e3f4 e5f4 f1f4 d8g5 g3f1 g5g6 f4f3 g8e7 f1e3');
INSERT INTO evaluations VALUES(3868,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(3868,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3868,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(3868,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(3868,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(3868,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3868,36,'cp 0','c6a5 a4c5 a5c4 c5b3 c4b6 b3d2 b6d7 f1e1 f8e8 c2c4 a7a6 b2b4 b7b5 a2a4 c7c5 a4b5 c5b4 c1b2 c8b7 d3d4 a6b5 c4b5 a8a1 d1a1 d8a8 h2h3 h7h6 a1a8 b7a8 d2c4 f6e4 d4e5 d6e5 f3e5 e4c3 b5b6 d7e5 e1e5 e8e5 c4e5 f7f6 e5d3');
INSERT INTO evaluations VALUES(52416,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(52416,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(52416,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(52416,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52416,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(52416,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(52416,36,'cp -4','g8f6 b1c3 e7e6 g1f3 d7d5 e4d5 e6d5 c4b5 b8c6 d2d4 f8e7 e1g1 e8g8 d4c5 e7c5 c1g5 c8e6 g5h4 h7h6 f1e1 a7a6 b5c6 b7c6 d1d3 f8e8 f3d4 e6d7 e1e8 d7e8 a1e1 a6a5 h2h3 a5a4 d4f3 e8d7 f3e5 g7g5');
INSERT INTO evaluations VALUES(44521,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(44521,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44521,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(44521,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(44521,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(44521,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44521,36,'cp 0','e4e5 d7d5 c4b5 c8d7 b5e2 f6g8 c2c3 b8c6 d2d4 c5d4 c3d4 d8b6 g1f3 d7g4 b1d2 e7e6 h2h3 g4h5 a2a3 a8c8 d1a4 b6a5 a4a5 c6a5 b2b4 a5c4 e1g1 g8e7 b4b5 h5g6 a3a4 b7b6 a1a2 h7h5 f3h4 g6h7 h4f3');
INSERT INTO evaluations VALUES(44509,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44509,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44509,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44509,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44509,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44509,36,'cp 0','d7d5 c4b5 c8d7 b5e2 f6g8 c2c3 b8c6 d2d4 c5d4 c3d4 d8b6 g1f3 d7g4 b1d2 e7e6 h2h3 g4h5 a2a3 a8c8 e1g1 g8e7 d1a4 b6a5 a4d1');
INSERT INTO evaluations VALUES(45405,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45405,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(45405,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45405,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45405,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45405,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(45405,36,'cp 0','c4b5 c8d7 b5e2 f6g8 c2c3 b8c6 g1f3 d7g4 d2d4 c5d4 f3d4 g4e2 d1e2 c6d4 c3d4 e7e6 e1g1 g8e7 b1d2 e7f5 d2f3 h7h5 c1g5 f8e7 g5e7 d8e7 e2b5 e7d7 b5c5 d7e7');
INSERT INTO evaluations VALUES(45291,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45291,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45291,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45291,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45291,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45291,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(45291,36,'cp 0','c8d7 b5e2 f6g8 c2c3 b8c6 d2d4 c5d4 c3d4 d8b6 g1f3 d7g4 b1d2 e7e6 h2h3 g4h5 a2a3 a8c8 e1g1 g8e7 d1a4 b6a5 a4d1');
INSERT INTO evaluations VALUES(57145,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(57145,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(57145,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(57145,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(57145,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(57145,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(57145,36,'cp -4','e5d4 g1f3 f8b4 b1d2 b8c6 f1d3 g8f6 e1g1 e8g8 e4e5 f6g4 a2a3 b4e7 f1e1 d7d6 e5d6 c7d6 d2b3 e7h4 e1f1 h4f6 d3e2 f8e8 h2h3 g4e5 f3d4 c6d4 b3d4 e5c6 d4c6 b7c6 e2d3 a7a5 a1b1 d6d5 c1e3 a5a4 f1e1 c8a6 d3a6 a8a6 c2c3 a6a8 d1d3 h7h6 e3d4 f6d4 e1e8 d8e8 c3d4');
INSERT INTO evaluations VALUES(57311,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(57311,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(57311,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(57311,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(57311,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(57311,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(57311,36,'cp -9','g1f3 f8b4 b1d2 b8c6 f1d3 g8f6 e4e5 f6g4 e1g1 e8g8 a2a3 b4e7 f1e1 d7d6 e5d6 c7d6 d2b3 e7h4 e1f1 g4e5 b3d4 e5d3 d1d3 c6e5 d3b3 h4f6 c1f4 f8e8 a1d1 b7b6 f1e1 c8g4 c2c3 d8d7 b3d5 d7a4 f4g3 a8d8 h2h3 e5f3 d4f3 g4e6 d5d2 e6b3 d1a1 e8e1 a1e1 b3e6');
INSERT INTO evaluations VALUES(57310,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57310,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(57310,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(57310,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(57310,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(57310,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(57310,36,'cp -15','f8b4 b1d2 b8c6 f1d3 d7d6 e1g1 b4c5 d2b3 c5b6 d3b5 a7a6 b5c6 b7c6 b3d4 g8e7 f1e1 a6a5 c2c4 e8g8 d1d3 d8d7 c1e3 c6c5 d4e2 e7g6 e2c3 a5a4 c3d5 b6a5 e3d2 a5d2 d3d2 c8b7 h2h3 f7f6 a1d1 d7f7 d2c3 b7c6 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(39404,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(39404,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(39404,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(39404,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(39404,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(39404,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(39404,36,'cp -4','b1d2 b8c6 f1d3 d7d6 e1g1 b4c5 d2b3 c5b6 d3b5 a7a6 b5c6 b7c6 b3d4 g8e7 f1e1 a6a5 c2c4 e8g8 d1d3 d8d7 c1e3 f7f6 a1d1 a5a4 d4f5 b6a5 f5e7 d7e7 e3d2 a5d2 f3d2 e7e5 b2b4 a4b3 a2b3 c6c5 d2f3 e5h5 d3d5 h5d5 c4d5 g7g5');
INSERT INTO evaluations VALUES(57038,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(57038,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(57038,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(57038,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(57038,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(57038,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(57038,36,'cp 11','d7d5 e4d5 e6d5 d2d4 f8d6 f1d3 g8e7 e1g1 e8g8 c1g5 b8c6 c2c3 c8g4 b1d2 d8d7 f1e1 e7g6 h2h3 g4f5 d3f5 d7f5 d2f1 f5d7 h3h4 f7f6 g5d2 h7h5 f3h2 d7f5 f1g3 d6g3 f2g3');
INSERT INTO evaluations VALUES(55906,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55906,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(55906,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(55906,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(55906,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(55906,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(55906,36,'cp 12','e4d5 e6d5 d2d4 f8d6 c2c4 g8f6 c4c5 d6e7 d1a4 c8d7 a4c2 e8g8 c1e3 f8e8 f1e2 g7g6 e1g1 d7f5 e2d3 f5d3 c2d3 b8c6 a2a3 f6e4 b1c3 e7f6 f1e1 d8d7 h2h3 e4c3 b2c3 b7b6 a1d1 a8b8 c5b6 c7b6 e3f4 b8c8 e1e8 d7e8 d1e1 e8d7 d3b5');
INSERT INTO evaluations VALUES(55779,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(55779,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(55779,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(55779,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55779,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(55779,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(55779,36,'cp 13','e6d5 d2d4 f8d6 c2c4 g8f6 c4c5 d6e7 f1d3 e8g8 e1g1 b7b6 c5b6 a7b6 b1c3 c8g4 h2h3 g4h5 f1e1 c7c6 a2a4 b8a6 d3a6 a8a6 d1d3 b6b5 c1f4 h5f3 d3f3 b5a4 c3a4 f6e4 a4c3 a6a1 e1a1 e4c3 b2c3 d8d7 a1e1 e7f6 g2g4');
INSERT INTO evaluations VALUES(56286,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(56286,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(56286,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(56286,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(56286,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(56286,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(56286,36,'cp 15','d2d4 f8d6 c2c4 c7c6 f1d3 g8e7 e1g1 h7h6 f1e1 e8g8 b1c3 d5c4 d3c4 b8a6 c4a6 b7a6 f3e5 c8e6 c1f4 a8c8 d1d3 f8e8 h2h3 e7d5 f4g3 c6c5 c3e4 d5b4 d3d1 c5d4 e4d6 d8d6 e5f7 d6d5 f7d6 b4c2 d6c8 c2e1 d1e1');
INSERT INTO evaluations VALUES(56208,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(56208,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(56208,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(56208,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(56208,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(56208,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(56208,36,'cp 14','g8f6 f1d3 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 c8e6 f1e1 b8d7 b1d2 c7c6 c2c3 d8b6 d1c2 a8e8 d2f1 e6g4 h4g3 d6g3 h2g3 c6c5 f1e3 g4f3 d4c5 d7c5 g2f3 c5d3 c2d3 b6b2 e1b1 b2a3 b1b7 e8c8 b7b3 a3c5 g1g2 f8e8 e3f5 c5c7 f5e3');
INSERT INTO evaluations VALUES(48683,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(48683,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(48683,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(48683,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(48683,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(48683,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(48683,36,'cp 26','f1d3 c7c5 c2c3 f8e7 d4c5 e7c5 e1g1 e8g8 c1f4 c8g4 b1d2 c5b6 d1c2 b8d7 a1e1 g4h5 d3f5 h5g6 e1e2 f8e8 e2e8 d8e8 f1e1 e8d8 f5g6 h7g6 g2g3 d7c5 d2b3 c5e6 f4e3 b6e3 e1e3 a7a5 g1g2 d8d7 b3d4');
INSERT INTO evaluations VALUES(48670,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48670,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48670,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(48670,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(48670,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(48670,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(48670,36,'cp 18','c7c5 c2c3 f8d6 d4c5 d6c5 e1g1 e8g8 c1f4 c8g4 b1d2 b8c6 d1c2 a8c8 f1e1 g4h5 d3f5 h5g6 d2b3 c5b6 a1d1 f8e8 e1e8 d8e8 d1e1 e8f8 f4e3 b6e3 e1e3 c8e8 e3e8 f8e8 h2h3 g6f5 c2f5 e8e6 f3d4 e6f5 d4f5 f6e4 a2a4 h7h6');
INSERT INTO evaluations VALUES(46433,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46433,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46433,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(46433,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46433,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46433,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(46433,36,'cp 20','c2c3 f8e7 d4c5 e7c5 e1g1 e8g8 c1f4 c8g4 b1d2 b8c6 h2h3 g4h5 d1b3 d5d4 c3d4 c6d4 f3d4 c5d4 g2g4 h5g6 d3g6 h7g6 a1d1 d8b6 d2c4 b6b3 a2b3 d4c5 f4d6 c5d6 c4d6 b7b6 f1e1 f8d8 e1e7 a7a5 g1g2 d8d7 e7d7 f6d7 d6b7 d7c5 b7c5 b6c5');
INSERT INTO evaluations VALUES(46431,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(46431,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46431,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46431,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46431,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46431,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(46431,36,'cp 19','c5c4 d3c2 c8g4 c1g5 f8e7 b1d2 e8g8 d2f1 b8c6 f1e3 g4h5 e1g1 f8e8 f1e1 f6g4 g5f4 g4e3 e1e3 d8d7 d1d2 e7f8 a1e1 f7f6 a2a4 e8e3 f2e3 h5g6 e3e4 d5e4 c2e4 c6e7 e4g6 e7g6 f4g3 a8e8 e1e8 d7e8 d2c2');
INSERT INTO evaluations VALUES(52374,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52374,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(52374,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(52374,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(52374,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(52374,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(52374,36,'cp 28','d7d5 b1d2 d5e4 d2e4 c8f5 e4g3 f5g6 g1f3 b8d7 h2h4 h7h6 f1d3 g6d3 d1d3 e7e6 c1f4 d8a5 f4d2 a5c7 e1c1 g8f6 g3e4 a7a6 g2g3 f6e4 d3e4 d7f6 e4e2 a8d8 h1g1 c6c5 d4c5 c7c5 d2c3 d8d1 g1d1 f8e7 f3e5 c5c7 e2c4 c7c4 e5c4 e8g8 c3f6 e7f6 d1d6 f8e8 d6b6');
INSERT INTO evaluations VALUES(52790,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(52790,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52790,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52790,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(52790,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(52790,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(52790,36,'cp 31','b1c3 d5e4 c3e4 c8f5 e4g3 f5g6 g1f3 b8d7 h2h4 h7h6 f1d3 g6d3 d1d3 e7e6 c1f4 d8a5 f4d2 f8b4 c2c3 b4d6 c3c4 a5c7 g3e4 d6e7 e1g1 g8f6 g2g3 f6e4 d3e4 d7f6 e4e2 e8g8 f1d1 c6c5 d2f4 e7d6 f4d6 c7d6 d4c5 d6c5 f3e5 f8d8 b2b3 c5c7');
INSERT INTO evaluations VALUES(55227,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(55227,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(55227,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(55227,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(55227,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(55227,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(55227,36,'cp 33','g1f3 g8f6 e2e3 e7e6 f1c4 a7a6 e1g1 c7c5 d4c5 d8d1 f1d1 f8c5 c4f1 c8d7 b1c3 e8g8 f3e5 c5e7 g2g3 f8c8 f1g2 c8c7 e3e4 d7e8 c1f4 c7c8 a1c1 f6d7 e5d7 b8d7 c3a4 c8c1 d1c1');
INSERT INTO evaluations VALUES(55225,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(55225,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(55225,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55225,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(55225,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(55225,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(55225,36,'cp 36','a7a6 e2e3 e7e6 f1c4 g8f6 e1g1 c7c5 b2b3 c5d4 f3d4 f8d6 b1d2 e8g8 c4d3 c8d7 d1e2 b8c6 d4c6 d7c6 d2c4 d8c7 c4d6 c7d6 f1d1 d6e7 a2a4 f8d8 c1a3 e7e8 d3c4 d8d1 a1d1 h7h6 a3b2');
INSERT INTO evaluations VALUES(50857,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50857,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(50857,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(50857,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(50857,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(50857,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50857,36,'cp 36','e2e3 e7e6 f1c4 g8f6 e1g1 c7c5 d4c5 d8d1 f1d1 f8c5 c4e2 e8g8 c1d2 b7b6 d2e1 c8b7 f3e5 b8c6 e5c4 f8d8 b1d2 b6b5 d2b3 c5e7 d1d8 a8d8 c4a5 c6a5 b3a5 b7a8 a1c1 f6d5 a5c6 a8c6 c1c6 e7b4 g1f1 b4e1 f1e1 d5b4 c6b6 b4a2 b6a6');
INSERT INTO evaluations VALUES(57760,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57760,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57760,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57760,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57760,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57760,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57760,36,'cp 0','g8f6 g1f3 e7e6 c2c4 f8e7 d2d4 e8g8 f1d3 d7d5 e1g1 d5c4 d3c4 a7a6 b2b3 b7b5 c4e2 c8b7 a2a4 b5b4 c1b2 c7c5 d4c5 b8d7 c5c6 b7c6 b1d2 d8c7 a1c1 c7b7 d2c4 c6d5 c4a5 b7b6 a5c4');
INSERT INTO evaluations VALUES(50381,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(50381,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(50381,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(50381,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(50381,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(50381,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(50381,36,'cp 10','g1f3 e7e6 b2b3 d7d5 c1b2 b8d7 d2d4 b7b6 f1d3 c8b7 e1g1 f8d6 f3e5 e8g8 d1e2 f6e4 c2c4 c7c5 c4d5 e6d5 b1d2 d7f6 d3e4 d5e4 d2c4 b7a6 e2e1 c5d4 b2d4 d8c7 h2h3 d6c5 a1d1 c5d4 d1d4');
INSERT INTO evaluations VALUES(50382,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50382,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50382,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(50382,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(50382,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50382,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(50382,36,'cp 10','d7d5 d2d4 e7e6 b2b3 b7b6 f1b5 c7c6 b5d3 f8e7 c1b2 e8g8 e1g1 c8b7 c2c4 c6c5 d4c5 d5c4 d3c4 e7c5 a2a3 a7a6 b3b4 d8d1 f1d1 c5e7 c4e2 f8c8 b1d2 b8d7 d2c4 b7d5 a1c1 b6b5 c4d2 c8c1 d1c1 a6a5 e2b5 a5b4 a3b4');
INSERT INTO evaluations VALUES(47279,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47279,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(47279,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47279,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(47279,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47279,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(47279,36,'cp 0','d2d4 e7e6 f1d3 b7b6 e1g1 f8d6 b2b3 e8g8 c2c4 c8b7 c1b2 b8d7 b1d2 f6e4 a1c1 a8c8 c4d5 e6d5 f3e5 d8e7 d2f3 d7b8 f3d2');
INSERT INTO evaluations VALUES(47198,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47198,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47198,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47198,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47198,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47198,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47198,36,'cp 9','e7e6 b2b3 b7b6 f1b5 c7c6 b5d3 c8b7 e1g1 c6c5 b1d2 b8d7 d1e2 f8e7 c1b2 e8g8 f1d1 c5d4 e3d4 a8c8 a1c1 e7b4 c2c3 b4d6 f3e5 f8e8 d1e1 d8e7 a2a4 c8c7 a4a5 b6a5 c1a1 e8c8 a1a5 d7e5');
INSERT INTO evaluations VALUES(48353,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(48353,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(48353,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(48353,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48353,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(48353,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48353,36,'cp 4','b2b3 b7b6 c1b2 f8d6 f1b5 c7c6 b5d3 c8b7 b1d2 c6c5 e1g1 e8g8 d4c5 d6c5 a2a3 b8d7 c2c4 d5c4 d2c4 b6b5 c4e5 d7e5 b2e5 b5b4 a3b4 c5b4 a1c1 a7a5 d1e2 a8c8 f1d1 d8e7 c1c8 f8c8 e5b2 b4c3 d1c1 c3b2');
INSERT INTO evaluations VALUES(48311,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(48311,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48311,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(48311,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(48311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48311,36,'cp 0','f8b4 c2c3 b4d6 c1b2 b8d7 f1d3 b7b6 c3c4 c8b7 b1d2 e8g8 e1g1 c7c5 c4d5 f6d5 a1c1 a8c8 d2c4 d6e7 d3b1 c5d4 f3d4 c8c5 b2a3 c5c8');
INSERT INTO evaluations VALUES(57170,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(57170,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(57170,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(57170,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(57170,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(57170,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(57170,36,'cp -61','e5f4 g1f3 d7d6 f1c4 h7h6 h2h4 g8f6 b1c3 b8c6 d2d4 f6h5 c3d5 h5g3 h1h2 c6e7 d5f4 g3e4 d1e2 d6d5 c4d5 e7d5 e2e4 d8e7 e4e7 d5e7 e1f2 c8f5 c2c3 h6h5 h2h1 c7c6 f3g5 f7f6 g5e6 e8f7 h1e1 f5e6 f4e6 e7f5 e6f8 a8f8 g2g3 h8g8 b2b3 g7g5 h4g5 f6g5');
INSERT INTO evaluations VALUES(57323,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(57323,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(57323,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(57323,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(57323,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(57323,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(57323,36,'cp -67','g1f3 d7d6 f1c4 h7h6 h2h4 g8f6 b1c3 b8c6 d2d4 f6h5 c3d5 h5g3 h1h2 c6e7 d5f4 g3e4 d1e2 d6d5 c4d5 e7d5 e2e4 d8e7 e4e7 d5e7 h2h1 c8f5 c2c3 h6h5 e1f2 f7f6 b2b3 e8f7 c1a3 c7c6 h1e1 e7d5 f4d5 f8a3 d5e3 f5e6 e3c4 a3e7 e1e2 h8e8');
INSERT INTO evaluations VALUES(57322,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(57322,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(57322,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(57322,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(57322,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(57322,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(57322,36,'cp -61','d7d6 f1c4 h7h6 h2h4 g8f6 b1c3 b8c6 d2d4 f6h5 c3d5 h5g3 h1h2 c6e7 d5f4 g3e4 d1e2 d6d5 c4d5 e7d5 e2e4 d8e7 e4e7 d5e7 h2h1 h6h5 e1f2 f7f6 b2b3 e8f7 c1a3 c7c6 h1e1 e7d5 a3f8 d5f4 f8b4 c8e6 a2a4 h8d8 a4a5 d8d7 b4d2 f4d5 c2c4 d5e7 g2g3');
INSERT INTO evaluations VALUES(55695,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(55695,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(55695,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55695,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(55695,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(55695,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(55695,36,'cp -49','f1c4 h7h6 h2h4 g8f6 b1c3 b8c6 d2d4 f6h5 c3d5 h5g3 h1h2 c6e7 d5f4 g3e4 d1e2 e4f6 c1d2 d6d5 c4d3 c8g4 e2e3 d8d6 e1c1 f6e4 c2c4 g4f3 c4c5 d6f6 g2f3 e4d2 f4d5 f6f3 d5c7 e8d8 e3e5 e7c6 e5e8 d8c7 e8a8 f3d3 h2d2 d3e4 d4d5');
INSERT INTO evaluations VALUES(55660,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(55660,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(55660,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(55660,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(55660,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(55660,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(55660,36,'cp -57','h7h6 h2h4 b8c6 d2d4 g8f6 b1c3 f6h5 c3d5 h5g3 h1h2 c6e7 d5f4 g3e4 d1e2 e4f6 c1d2 d6d5 c4d3 c8g4 e1c1 d8d7 d1e1 e8c8 e2f2 d7a4 f3e5 a4a2 d2c3 c8b8 e5f7 f6e4 d3e4 d5e4 f7d8 e7f5 c1d2 e4e3 e1e3 f5e3 f2e3 f8d6 d8c6 b7c6 h2h1 a2f7 f4d3 f7d5');
INSERT INTO evaluations VALUES(55425,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(55425,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(55425,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(55425,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(55425,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(55425,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(55425,36,'cp -62','h2h4 b8c6 d2d4 g8f6 b1c3 f6h5 c3d5 h5g3 h1h2 c6e7 d5c3 g7g5 d1d3 e7g6 h4g5 h6g5 h2h8 g6h8 c1d2 h8g6 e1c1 c7c6 c4b3 f8g7 d3c4 c8e6 d4d5 c6d5 c3d5 e8f8 c1b1 g6e5 f3e5 g7e5 d2c3 a8c8 c4d3 e5c3 d5c3 e6b3 a2b3 d8f6 c3b5');
INSERT INTO evaluations VALUES(54981,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(54981,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(54981,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(54981,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(54981,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(54981,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(54981,36,'cp 67','d2d4 g8f6 b1c3 e7e5 g1f3 b8d7 f1c4 f8e7 a2a4 e8g8 h2h3 e5d4 d1d4 d7c5 e1g1 c8e6 f1e1 c7c6 c1e3 e6c4 d4c4 c5e6 f3d4 e6d4 e3d4 f8e8 a1d1 d8c7 c4d3 f6d7 b2b4 d7e5 d3g3 e5g6');
INSERT INTO evaluations VALUES(54971,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(54971,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(54971,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(54971,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(54971,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(54971,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(54971,36,'cp 58','g8f6 b1c3 e7e5 g1f3 b8d7 a2a4 f8e7 f1c4 e8g8 e1g1 e5d4 d1d4 d7c5 h2h3 c8e6 f1e1 c7c6 c1e3 f8e8 a1d1 e7f8 f3g5 h7h6 g5e6 c5e6 d4d3 d8c7 b2b4 a8d8 b4b5 e6c5 e3c5 d6c5 d3e2 d8d1 e1d1');
INSERT INTO evaluations VALUES(47029,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(47029,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(47029,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(47029,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(47029,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(47029,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(47029,36,'cp 55','b1c3 e7e5 g1f3 b8d7 f1c4 f8e7 e1g1 e8g8 f1e1 c7c6 a2a4 f8e8 h2h3 e5d4 d1d4 d7f8 c1f4 c8e6 a4a5 e6c4 d4c4 f8e6 f4g3 f6h5 g3h2 d8c7 f3d4 h5f4 h2f4 e6f4 a5a6 a8c8 a6b7 c7b7 c4b3 b7b6 d4f5 e7f8 e1d1 b6b3 c2b3');
INSERT INTO evaluations VALUES(47025,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47025,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(47025,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(47025,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(47025,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(47025,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(47025,36,'cp 59','e7e5 g1f3 b8d7 f1c4 f8e7 e1g1 e8g8 f1e1 c7c6 a2a4 f8e8 h2h3 e5d4 f3d4 e7f8 c4a2 h7h6 c1f4 d7c5 d1f3 c5e6 a2e6 c8e6 a1d1 d8b6 e1e3 d6d5 e4e5 f6h7 c3e2 b6b2 e3b3 b2a2 b3b7 a2a4 d4e6 e8e6');
INSERT INTO evaluations VALUES(47847,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47847,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(47847,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(47847,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(47847,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(47847,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(47847,36,'cp 51','g1f3 b8d7 f1c4 f8e7 e1g1 e8g8 f1e1 e5d4 f3d4 d7e5 c4f1 c7c6 h2h3 e5g6 a2a4 f8e8 g2g3 c8d7 f1g2 e7f8 c1e3 g6e5 b2b3 d6d5 f2f4 e5g6 e4e5 f8b4 d1d3 b4c3 d3c3 f6e4 g2e4 d5e4 e3d2 d7h3 e1e4 g6e7 e5e6 h3e6');
INSERT INTO evaluations VALUES(47846,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(47846,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(47846,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(47846,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(47846,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(47846,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(47846,36,'cp 51','b8d7 h2h3 c7c6 a2a4 f8e7 f1c4 e5d4 d1d4 d7c5 e1g1 e8g8 f1d1 c8e6 c1f4 a7a5 f4g3 e6c4 d4c4 b7b5 a4b5 c6b5 c3b5 c5e4 g3f4 d8d7 c4e2 f8e8 c2c4 h7h6 f3d2 e7f8 e2f3 e4g5 f3d3 a5a4 b5c3 d6d5 f4g5 h6g5 c3d5 f6d5');
INSERT INTO evaluations VALUES(15920,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(15920,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(15920,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(15920,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(15920,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(15920,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(15920,36,'cp 59','a2a4 f8e7 f1c4 e8g8 e1g1 e5d4 f3d4 f8e8 f1e1 d7e5 c4a2 h7h6 c1f4 e7f8 f4g3 e5c6 f2f3 a7a6 d1d2 c6d4 d2d4 c8e6 a2e6 e8e6 c3e2 c7c6 c2c4 f6d7 e1d1 a6a5 g3f2 d8c7 d4d2 d7c5 d2c3 a8d8 c3c2 e6e8 a1b1 c7d7 h2h3 f8e7');
INSERT INTO evaluations VALUES(19427,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(19427,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19427,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19427,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19427,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19427,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(19427,36,'cp 26','c5d4 f3d4 g8f6 b1c3 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 b2b4 b8d7 a4a5 a7a6 b5a3 b7b6 a5b6 d8b6 c2c3 d7f6 a3c4 b6b7 f1e2 e8g8 c1g5 b7d5 d1d5 f6d5 g5e7 d5e7 e1c1 c8b7 d1d6 b7g2 h1g1');
INSERT INTO evaluations VALUES(19556,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(19556,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(19556,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(19556,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19556,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(19556,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(19556,36,'cp 28','f3d4 g8f6 b1c3 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 b2b4 a7a6 b5a3 b7b6 f1e2 b8d7 e1g1 e8g8 c1e3 f7f5 f2f3 c8b7 g1h1 a6a5 c2c3 a5b4 c3b4 d7f6 e2c4 g8h8 a3b5 a8c8 c4b3 b7a6 e3f2 a6b5 a4b5 g7g6 a1a7');
INSERT INTO evaluations VALUES(19540,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19540,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19540,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(19540,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19540,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(19540,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(19540,36,'cp 27','g8f6 b1c3 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 f1e2 b8d7 c1d2 a7a6 b5a3 e8g8 a4a5 d7f6 d2b4 c8f5 a3c4 f5e4 f2f3 e4d5 c4b6 d5e6 b6a8 d8a8 e1g1 a8d8 g1h1 d6d5 d1d2 d5d4 b4e7 d8e7 e2d3 h7h6 f1d1 f6d5');
INSERT INTO evaluations VALUES(12881,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12881,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(12881,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12881,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(12881,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(12881,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(12881,36,'cp 25','b1c3 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 f1e2 e8g8 e1g1 b8d7 c1d2 a7a6 b5a3 a6a5 e2b5 f7f5 a3c4 d8c7 g1h1 f5f4 f2f3 d7c5 d1e2 c8d7 b5d7 c5d7 d2e1 d7b6 c4b6 c7b6 a1b1 b6e3 e2b5 e3c5 b5c5 d6c5 c2c4 b7b6 e1f2 e7f6 f1e1');
INSERT INTO evaluations VALUES(12879,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(12879,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(12879,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(12879,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(12879,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(12879,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12879,36,'cp 30','e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 c2c4 a7a6 b5c3 f8e7 f1d3 e8g8 e1g1 f7f5 f2f3 b8d7 g1h1 a6a5 c1e3 b7b6 a2a3 d7c5 d3c2 c8a6 b2b4 c5d7 c2d3 a8c8 d1e2 e7f6 a1c1 f6g5 c3a4 g8h8 h2h3 g5e3 e2e3 d8h4 e3f2 h4d4 f2d4 e5d4');
INSERT INTO evaluations VALUES(12587,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(12587,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(12587,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(12587,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(12587,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(12587,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(12587,36,'cp 29','d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 c2c4 a7a6 b5c3 f8e7 f1d3 f7f5 f2f3 e8g8 e1g1 c8d7 d1b3 d7c8 g1h1 b8d7 b3d1 e7g5 a2a3 g5c1 a1c1 d7f6 d1c2 g7g6 b2b4 c8d7 c2d2 d8b6 c1b1 g8g7 a3a4 f8e8 a4a5 b6d4 c3e2 d4h4');
INSERT INTO evaluations VALUES(12515,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(12515,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(12515,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12515,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12515,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12515,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(12515,36,'cp 9','d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 f1e2 b8d7 e1g1 e8g8 c1d2 f7f5 a4a5 a7a6 b5a3 e5e4 a3c4 d7e5 c4b6 a8b8 f2f4 e4f3 e2f3 e7f6 d2c3 g7g5 d1d2 f5f4 a1e1 d8c7 e1e5 f6e5 c3e5 d6e5 b2b4 c7g7 f3e4 c8f5 e4f5 f8f5 d5d6 b8d8 c2c4');
INSERT INTO evaluations VALUES(13839,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13839,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(13839,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13839,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(13839,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(13839,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13839,36,'cp 17','c3d5 f6d5 e4d5 c6b8 c2c4 a7a6 b5c3 f8e7 f1d3 f7f5 f2f3 e8g8 e1g1 b8d7 g1h1 e7g5 b2b4 a6a5 c1g5 d8g5 c3b5 a5b4 b5d6 d7c5 d3c2 c8d7 d1e1 g5e7 d6b5 d7b5 c4b5 c5d7 d5d6 e7d6 a1d1 d6c7 c2b3 g8h8 e1b4 d7c5 b5b6');
INSERT INTO evaluations VALUES(13831,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(13831,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13831,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(13831,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(13831,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(13831,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13831,36,'cp 34','f6d5 e4d5 c6b8 a2a4 b8d7 f1e2 f8e7 e1g1 e8g8 c1d2 a7a6 b5a3 a6a5 e2b5 f7f5 a3c4 d8c7 g1h1 f5f4 f2f3 d7f6 b2b3 b7b6 d2c3 f8d8 f1f2 c8d7 b5d7 f6d7 f2f1 d7c5 d1e2 d8e8 c3b2 e7h4 b2a3 e8e7');
INSERT INTO evaluations VALUES(13900,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(13900,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(13900,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(13900,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(13900,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(13900,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(13900,36,'cp 30','e4d5 c6b8 a2a4 b8d7 f1e2 f8e7 e1g1 a7a6 b5a3 e8g8 c1d2 a6a5 e2b5 f7f5 a3c4 d8c7 b2b4 a5b4 d2b4 d7c5 d1e1 c8d7 b4a5 c7b8 c4b6 d7b5 a4b5 a8a7 a5b4 a7a1 e1a1 f8e8 b4c5 d6c5 b6c4 e7d6 g2g3 e5e4 a1a5 f5f4 a5b6 d6e5 b6c5 e8c8 c5b4 b8c7 c4e5 c7e5');
INSERT INTO evaluations VALUES(13899,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(13899,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(13899,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13899,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(13899,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(13899,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13899,36,'cp 20','c6b8 a2a4 f8e7 f1e2 b8d7 e1g1 e8g8 c1d2 a7a6 b5a3 a6a5 g1h1 f7f5 e2b5 d7c5 a3c4 f5f4 f2f3 d8c7 d2e1 c8f5 e1f2 e7f6 c2c3 g7g5 b2b4 a5b4 c3b4 c5d3 f2b6 c7e7 c4d2 a8c8 a4a5 d3b4 d2e4 b4c2 a1c1 f5e4 f3e4 c2d4');
INSERT INTO evaluations VALUES(45959,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45959,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(45959,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(45959,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(45959,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45959,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(45959,36,'cp 25','a2a4 f8e7 f1e2 e8g8 e1g1 b8d7 c1d2 a7a6 b5a3 a6a5 g1h1 d7b6 a3b5 b6d5 d2a5 d8a5 d1d5 a8b8 d5d3 c8e6 f2f4 b8c8 c2c3 e5f4 d3f3 d6d5 f3f4 a5d8 e2d3 g7g6 a4a5 e6d7 b5d4 d8c7 f4c7 c8c7');
INSERT INTO evaluations VALUES(55962,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55962,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55962,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55962,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(55962,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(55962,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(55962,36,'cp 33','c7c5 c2c3 c8d7 g1f3 b8c6 f1e2 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 a4c3 f5e3 f2e3 e8g8 e2d3 b7b5 b2b3 b5b4 c3e2 f8c8 h2h4 d7b5 e2f4 c8c3 c1c3 b4c3 d1c2 b5d3 c2d3');
INSERT INTO evaluations VALUES(54299,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(54299,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54299,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(54299,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(54299,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(54299,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(54299,36,'cp 25','c2c3 c8d7 g1f3 b8c6 f1e2 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 a4c3 b7b5 e2d3 f5e3 f2e3 a5c4 d1e2 b5b4 c3d1 e8g8 d1f2 d7b5 f2h3 a8c8 g1h1 g7g6 e2f2 c4a3 h3f4 c8c1 f1c1 b5d3 f4d3');
INSERT INTO evaluations VALUES(54291,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(54291,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54291,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54291,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(54291,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(54291,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(54291,36,'cp 11','c8d7 f1e2 g8e7 g1f3 c5d4 c3d4 b8c6 b1c3 e7f5 e1g1 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 e2d3 f5e3 f2e3 a5c4 d3c4 d5c4 a4c3 e8g8 f3d2 b7b5 c3e4 d7c6 d1g4 g8h8 e4g3 f7f5 e5f6 e7f6 d2e4 c6e4 g3e4 e6e5');
INSERT INTO evaluations VALUES(30897,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(30897,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(30897,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(30897,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(30897,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(30897,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(30897,36,'cp 28','g1f3 b8c6 f1e2 c5d4 c3d4 g8e7 e1g1 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 b7b5 a4c3 b5b4 c3a4 c6a5 a1c1 d7b5 b2b3 b5a4 b3a4 f8e7 e2d3 e8g8 d1c2 g7g6 d3f5 e6f5 e3h6 f8e8 f1e1 a5c4 h2h4 a8c8 h4h5 a7b6 e5e6 f7e6 c2d3 a6a5 f3g5 e7g5 h6g5 g6h5');
INSERT INTO evaluations VALUES(30896,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(30896,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(30896,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30896,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(30896,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(30896,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(30896,36,'cp 32','b8c6 f1e2 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1g5 f8e7 g5e7 d8e7 d1d2 e8g8 a1c1 a8c8 e2d1 f5h4 f3h4 e7h4 f2f4 h4d8 f1f2 d8b6 d1a4 f7f6 e5f6 f8f6 h2h3 c8c7 a4c6 c7c6 g1h2 c6c7 d2e3 c7c4 c1d1 c4b4 d1d2 a6a5');
INSERT INTO evaluations VALUES(25948,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(25948,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(25948,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(25948,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(25948,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(25948,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(25948,36,'cp 39','f1e2 c5d4 c3d4 g8e7 e1g1 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 e2d3 f5e3 f2e3 a5c4 d3c4 d5c4 a4c3 e8g8 c3e4 d7c6 f3d2 b7b5 f1f4 c6d5 d1h5 f7f5 e5f6 e7f6 c1f1 d5e4 d2e4 f6d4 f4f8 a8f8 e3d4 a7d4 e4f2 d4b2 h5c5 b2a2 c5e5 a2a3 e5e6 g8h8 f1e1');
INSERT INTO evaluations VALUES(25947,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(25947,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(25947,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(25947,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(25947,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(25947,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(25947,36,'cp 13','g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 e2d3 f5e3 f2e3 d7b5 h2h4 a5c4 d3c4 b5c4 f1f2 e8g8 f2c2 a6a5 b2b3 c4b5 c2c7 e7d8 c7c3 b7b6 a4b2 a7a6 b2d3 d8e7 a2a4 b5d3 d1d3 a6d3 c3d3 a8c8 d3c3');
INSERT INTO evaluations VALUES(24452,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(24452,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(24452,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(24452,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(24452,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(24452,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24452,36,'cp 43','e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 b7b5 a4c3 b5b4 c3a4 c6a5 a1c1 d7b5 e2d3 f5e3 f2e3 f8e7 b2b3 e8g8 d1c2 h7h6 h2h4 b5d3 c2d3 f8c8 c1c2 a7b7 f1c1 c8c6 a4c5 e7c5 d4c5 b7e7 d3d2');
INSERT INTO evaluations VALUES(24451,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24451,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(24451,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24451,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(24451,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(24451,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(24451,36,'cp 41','c5d4 c3d4 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 e2d3 f5e3 f2e3 d7b5 d1c2 h7h6 g2g4 a5c4 a4c3 c4e3 c3b5 e3c2 b5a7 c2b4 a7c8 b4d3 c1c7 e8d8 c7e7 a8c8 e7b7 c8c7 b7b8 c7c8 b8b6 d8c7 b6b3 d3f4 f1c1 c7d8 c1c8 d8c8 g1f2 g7g5');
INSERT INTO evaluations VALUES(17889,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(17889,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17889,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(17889,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17889,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(17889,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(17889,36,'cp 41','b5a4 f8e7 e1g1 g8f6 f1e1 b7b5 a4b3 e8g8 c2c3 d7d5 e4d5 f6d5 f3e5 c6e5 e1e5 c7c6 e5e1 e7d6 d2d3 c8f5 d1f3 d8f6 b1d2 f6g6 b3d1 f5d3 d2e4 d3e4 f3e4 a6a5 d1f3 a5a4 a2a3 d6c5 e4g6 h7g6 c1g5 f7f6 g5h4 f8e8 h4g3 e8e1 a1e1 c5f8 e1d1');
INSERT INTO evaluations VALUES(17914,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(17914,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(17914,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(17914,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(17914,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(17914,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(17914,36,'cp 19','g8f6 e1g1 f8e7 f1e1 b7b5 a4b3 e8g8 a2a4 c8b7 d2d3 d7d6 c1d2 b5b4 c2c3 a8b8 c3b4 b7c8 b4b5 a6b5 a4b5 b8b5 b1a3 b5b8 h2h3 c8e6 b3e6 f7e6 d1c2 c6b4 c2b3 d8d7 d2b4 c7c5 a3c4 b8b4 b3a2 f8b8 a2a7 b4b7 a7a6');
INSERT INTO evaluations VALUES(11541,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(11541,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(11541,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(11541,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(11541,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(11541,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11541,36,'cp 28','e1g1 f8e7 f1e1 b7b5 a4b3 e8g8 a2a4 c8b7 d2d3 d7d6 c1d2 b5b4 c2c3 a8b8 c3b4 b7c8 b4b5 a6b5 a4b5 b8b5 b1a3 b5b8 h2h3 c8d7 d2c3 f6h5 b3c4 h5f4 a3c2 h7h6 b2b4 e7f6 c3d2 f4e6 a1a3 c6e7 c2e3 g8h8 a3a6 e7c6');
INSERT INTO evaluations VALUES(11540,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(11540,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(11540,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(11540,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11540,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(11540,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(11540,36,'cp 28','f8e7 f1e1 b7b5 a4b3 e8g8 a2a4 c8b7 d2d3 d7d6 c1d2 b5b4 c2c3 a8b8 c3b4 b7c8 b4b5 a6b5 a4b5 b8b5 b1a3 b5b8 h2h3 c8d7 a3c4 h7h6 b3a4 c6b4 d2c3 f8e8 a4d7 d8d7 d1d2 c7c5 c3b4 b8b4 a1a5 d7e6 e1a1 e8b8 d2d1 b4b7');
INSERT INTO evaluations VALUES(7373,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(7373,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(7373,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(7373,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(7373,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(7373,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(7373,36,'cp 18','f1e1 b7b5 a4b3 e8g8 a2a4 b5b4 a4a5 d7d6 c2c3 a8b8 d2d3 c8e6 b3e6 f7e6 b1d2 b8b5 d3d4 b4c3 b2c3 e5d4 f3d4 c6d4 c3d4 c7c5 d4c5 b5c5 c1a3 c5e5 d2c4 e5e4 c4d6 e4e1 d1e1 d8d7 d6c4 e7a3 a1a3');
INSERT INTO evaluations VALUES(7374,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(7374,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(7374,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7374,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(7374,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(7374,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(7374,36,'cp 24','b7b5 a4b3 e8g8 a2a4 b5b4 d2d3 d7d6 a4a5 c8e6 b3e6 f7e6 b1d2 a8b8 c2c3 d8e8 d3d4 b4c3 b2c3 e5d4 f3d4 e8d7 d1e2 c6d4 c3d4 d7b5 c1a3 b5e2 e1e2 b8b5 e4e5 f6d5 d2c4 f8c8 e5d6 d5f4 e2e3 c7d6 a3d6 c8c4 d6f4 c4d4');
INSERT INTO evaluations VALUES(7500,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(7500,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(7500,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(7500,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(7500,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(7500,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(7500,36,'cp 30','a4b3 e8g8 h2h3 c8b7 d2d3 d7d6 a2a3 h7h6 b1c3 c6b8 c1d2 c7c5 b3a2 b8c6 a3a4 b5b4 c3d5 f6d5 a2d5 a8b8 c2c3 e7f6 f3h2 f6g5 d2g5 d8g5 h2g4 c6e7 d5b7 b8b7 g4e3 b4c3 b2c3 e7g6 a4a5');
INSERT INTO evaluations VALUES(7489,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(7489,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(7489,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(7489,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(7489,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(7489,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(7489,36,'cp 32','e8g8 c2c3 d7d6 h2h3 c8b7 d2d4 f8e8 a2a3 e7f8 b1d2 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 c2d3 f6h5 d3f1 h5f4 d1c2 a8c8 a1d1 d8b6 g2g3 f4e6 f1g2 e5d4 f3d4 d6d5 d4e6 e8e6 c3c4 c8e8 d2f3');
INSERT INTO evaluations VALUES(2807,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(2807,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(2807,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(2807,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(2807,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(2807,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(2807,36,'cp 35','c2c3 d7d6 h2h3 f8e8 d2d4 c8b7 b1d2 e7f8 a2a3 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 a8c8 a3a4 c7c6 d2f1 d8c7 f1g3 g7g6 d1d2 f8g7 c3c4 b5c4 b3c4 e5d4 b2d4 d7e5 f3e5 d6e5 d4c3 e8d8 d2e3 c6c5');
INSERT INTO evaluations VALUES(2812,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(2812,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(2812,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(2812,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(2812,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(2812,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(2812,36,'cp 16','d7d5 e4d5 f6d5 f3e5 c6e5 e1e5 c7c6 e5e1 e7d6 d2d3 c8f5 d1f3 d8h4 g2g3 h4h3 c1e3 f5d3 b3d5 c6d5 b1d2 h3f5 f3f5 d3f5 a2a3 f8e8 d2f3 f5e4 f3d4 d6e5 e1d1 h7h6 d1d2 f7f6 a3a4 e5d4 d2d4 g8f7 h2h4 g7g5 a4b5 a6b5');
INSERT INTO evaluations VALUES(2707,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(2707,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(2707,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(2707,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(2707,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(2707,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(2707,36,'cp 22','e4d5 f6d5 f3e5 c6e5 e1e5 c7c6 e5e1 e7d6 d2d3 c8f5 d1f3 d8h4 g2g3 h4h3 c1e3 f5d3 b3d5 c6d5 b1d2 h3f5 f3f5 d3f5 a2a3 a8c8 d2b3 f5e4 b3d4 d6c5 g1f1 e4d3 f1g2 c5d4 e3d4 f8e8 g3g4 f7f6 f2f3 g8f7 e1d1 d3c4 d1d2 a6a5 h2h4 a5a4');
INSERT INTO evaluations VALUES(2706,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(2706,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(2706,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(2706,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(2706,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(2706,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(2706,36,'cp 34','f6d5 f3e5 c6e5 e1e5 c7c6 e5e1 e7d6 d2d3 c8f5 d1f3 d8h4 g2g3 h4h3 c1e3 f5d3 b1d2 d3f5 b3d5 c6d5 f3d5 f8d8 d5g2 b5b4 g2h3 f5h3 d2c4 b4c3 b2c3 d6c7 a2a4 h3e6 c4b6 a8b8 a4a5 c7e5 a1c1 d8d3 b6a4 b8c8 c3c4 e6c4');
INSERT INTO evaluations VALUES(2713,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(2713,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(2713,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2713,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(2713,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(2713,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2713,36,'cp 37','f3e5 c6e5 e1e5 c7c6 e5e1 e7d6 d2d3 c8f5 d1f3 d8h4 g2g3 h4h3 b1d2 a8e8 d2e4 f5g4 f3g2 h3g2 g1g2 f7f5 h2h3 g4h5 c1f4 d6f4 g3f4 f5e4 d3e4 h5f3 g2f3 f8f4 f3g3 f4e4 e1e4 e8e4 f2f3 e4e5 c3c4 b5c4 b3c4 a6a5 a1c1 g8f8 c4d3 e5g5 g3f2 d5b4 d3h7 g5h5 a2a3 h5h7 a3b4 a5b4 f2g3 h7h6 c1c4 h6g6 g3f2 g6h6 h3h4 h6g6 c4b4');
INSERT INTO evaluations VALUES(2711,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(2711,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(2711,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(2711,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(2711,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(2711,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(2711,36,'cp 23','c6e5 e1e5 c7c6 d2d3 e7d6 e5e1 c8f5 d1f3 d8h4 g2g3 h4h3 c1e3 f5d3 b1d2 d3f5 b3d5 c6d5 f3d5 f8d8 d5g2 b5b4 e3b6 h3g2 g1g2 d8d7 d2c4 b4c3 b2c3 f7f6 b6d4 d6c7 a2a4 g8f7 a4a5 a8b8 f2f3 f5e6 c4b6 d7d6 h2h4 h7h6 h4h5');
INSERT INTO evaluations VALUES(2749,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(2749,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(2749,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(2749,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(2749,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(2749,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(2749,36,'cp 25','e1e5 c7c6 e5e1 e7d6 d2d3 c8f5 d1f3 d8h4 g2g3 h4h3 c1e3 f5d3 b1d2 d3f5 b3d5 c6d5 f3d5 f8d8 d5g2 b5b4 e3b6 h3g2 g1g2 d8d7 d2c4 b4c3 b2c3 f7f6 b6d4 d6c7 e1e3 f5g6 a2a4 g6f7 c4b6 c7b6 d4b6 d7d2 a4a5 a8c8 b6d4 h7h6 g2g1 c8e8 a1e1 e8e3 e1e3 d2c2');
INSERT INTO evaluations VALUES(2748,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(2748,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(2748,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(2748,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(2748,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(2748,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(2748,36,'cp 20','c7c6 e5e1 e7d6 d2d3 c8f5 d1f3 d8h4 g2g3 h4h3 c1e3 f5d3 b1d2 d3f5 b3d5 c6d5 f3d5 f8d8 d5g2 b5b4 d2c4 b4c3 g2h3 f5h3 b2c3 a8c8 c4d6 d8d6 e3d4 d6e6 e1e6 h3e6 g1g2 e6d5 g2h3 f7f6 g3g4 c8e8 a2a4 e8e2 h3g3 e2b2 a4a5 b2a2 a1a2');
INSERT INTO evaluations VALUES(2822,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(2822,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(2822,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2822,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(2822,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(2822,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(2822,36,'cp 24','d2d3 e7d6 e5e1 c8f5 d1f3 d8h4 g2g3 h4h3 c1e3 f5d3 b1d2 d3f5 b3d5 c6d5 f3d5 f8d8 d5g2 b5b4 g2h3 f5h3 d2c4 b4c3 b2c3 a8c8 c4d6 d8d6 e3d4 d6e6 f2f3 e6e1 a1e1 h3e6 a2a4 c8b8 g3g4 e6d7 e1e7 d7c6 g1f2 h7h6 e7a7 c6a4 a7a6 a4b5');
INSERT INTO evaluations VALUES(55543,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(55543,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55543,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(55543,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55543,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(55543,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(55543,36,'cp 16','f8e7 d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 c1f4 c7c6 a2a3 d6d5 e4d5 c6d5 c4a2 e7c5 e1g1 b8c6 d4f3 d5d4 c3e2 c8f5 f1e1 d4d3 c2d3 d8d3 d1d3 f5d3 f4g5 f6g4 g5h4 a8e8 e2f4 d3b5 b2b4 c5d4 a1d1 e8e1 d1e1 d4c3 e1d1 c3f6 f4d5');
INSERT INTO evaluations VALUES(38226,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38226,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38226,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(38226,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(38226,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(38226,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(38226,36,'cp 21','d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 c1f4 c7c6 a2a3 d6d5 e4d5 c6d5 c4a2 b8c6 e1g1 e7c5 d4f3 d5d4 c3e2 c8f5 e2g3 f5g6 f1e1 d4d3 c2d3 g6d3 f4e3 c5e3 e1e3 d3g6 d1b3 d8b6 b3c3 a8e8 a1e1 e8e3 e1e3');
INSERT INTO evaluations VALUES(38227,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38227,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38227,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(38227,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(38227,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(38227,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(38227,36,'cp 32','e5d4 f3d4 g8f6 b1c3 e8g8 c1f4 c7c6 a2a3 f8e8 e1g1 e7f8 f1e1 b8d7 f4g3 d7c5 f2f3 a7a5 d1d2 c8d7 a3a4 d6d5 e4d5 e8e1 g3e1 f6d5 c4d5 c6d5 e1f2 c5e6 c3d5 d7a4 d4e6 f7e6 a1a4 d8d5 d2d5 e6d5 f2d4 a8c8 a4a5');
INSERT INTO evaluations VALUES(38249,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(38249,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38249,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38249,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(38249,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38249,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(38249,36,'cp 42','f3d4 g8f6 b1c3 e8g8 c1f4 c7c6 a2a3 b8d7 e1g1 d6d5 e4d5 d7b6 d1d3 b6c4 d3c4 f6d5 c3d5 c6d5 c4d3 e7f6 f1e1 c8d7 f4e5 f8e8 d4f3 f6e5 f3e5 d7c6 a1d1 d8b6 d3f3 b6c7 f3c3 c7b6 h2h3 a8d8 d1d2 c6b5 c3d4 b6d4 d2d4');
INSERT INTO evaluations VALUES(38244,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38244,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(38244,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(38244,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(38244,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(38244,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38244,36,'cp 26','g8f6 b1c3 e8g8 c1f4 c7c6 a2a3 f8e8 e1g1 e7f8 f1e1 b8d7 h2h3 d7c5 d1f3 c5e4 c3e4 e8e4 e1e4 f6e4 f3e4 d6d5 c4d5 d8d5 e4d5 c6d5 a1d1 c8e6 f2f3 a8c8 f4e3 a7a5 c2c3 f8e7 g1f2 g7g5 g2g3 h7h6 f2g2 a5a4');
INSERT INTO evaluations VALUES(40826,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40826,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(40826,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(40826,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(40826,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(40826,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(40826,36,'cp 30','b1c3 e8g8 c1f4 c7c6 a2a3 d6d5 e4d5 c6d5 c4a2 e7c5 e1g1 b8c6 d4f3 c8e6 b2b4 c5d6 f4g5 d6e7 f1e1 h7h6 g5h4 a8c8 d1d3 f6g4 h4e7 c6e7 h2h3 e6f5 d3d2 g4f6 a1d1 a7a6 f3d4 f5g6 a2b3 d8c7 c3e2 f8e8 d2f4 e7f5 d4f5 e8e2 f4c7 e2e1 d1e1 c8c7 f5e7 g8h7 e7d5 f6d5 b3d5 g6c2 d5f3 c2f5 h3h4');
INSERT INTO evaluations VALUES(40824,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40824,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(40824,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40824,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(40824,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(40824,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40824,36,'cp 40','e8g8 c1f4 c7c6 a2a3 d6d5 e4d5 c6d5 c4a2 b8c6 e1g1 e7c5 d4f3 c8e6 b2b4 c5d6 f4g5 d6e7 f1e1 h7h6 g5h4 a8c8 d1d3 f6g4 h4e7 c6e7 a1d1 e6f5 d3d2 g4f6 f3d4 f5g6 a2b3 f8e8 h2h3 g6h5 f2f3 h5g6 a3a4 d8c7 c3e2 e7c6');
INSERT INTO evaluations VALUES(36246,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36246,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(36246,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(36246,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(36246,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(36246,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(36246,36,'cp 33','e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 e7f6 d1c2 f6d4 c3d4 b8c6 e4c6 b7c6 c2c6 c8e6 c6c3 e6d5 c1f4 c7c6 a1e1 d8d7 b2b3 e8e1 f1e1 a8e8 f2f3 e8e1 c3e1 h7h6 e1d1 f7f6 f4e3 d7e8 e3f2 g8h8 g1h1 d5e6 d1d2 e6g8 h2h3 e8e6 d2d3');
INSERT INTO evaluations VALUES(36245,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36245,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36245,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36245,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36245,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36245,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36245,36,'cp 12','f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 f1e1 d7f6 e4f3 c7c5 d4b5 c8d7 c3c4 a7a6 b5c3 d8c7 c1g5 a8d8 e1e7 e8e7 g5f6 g7f6 c3d5 d7e6 d5c7 d8d1 a1d1 e7c7 d1d8 g8g7 f3d5 b7b5 b2b3 b5c4 b3c4 f6f5 g1f1 g7f6 f1e2 f6e7 d8a8 e6d5 c4d5 e7d6 a8a6 d6d5 e2d3 c7b7 d3c3 f5f4 a6f6 d5e5 f6h6');
INSERT INTO evaluations VALUES(36277,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36277,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(36277,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(36277,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36277,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36277,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(36277,36,'cp 20','c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 f1e1 d7f6 e4f3 c7c5 d4b5 c8d7 a2a4 a7a6 b5a3 e7f8 a3c4 d7e6 d1d8 a8d8 c4b6 f6d7 b6d7 d8d7 c1e3 h7h6 a4a5 g7g5 g2g3 g8g7 a1d1 e8d8 d1d7 d8d7 g1g2 g5g4');
INSERT INTO evaluations VALUES(36276,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(36276,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36276,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36276,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36276,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36276,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(36276,36,'cp 25','d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 f1e1 d7f6 e4f3 c7c5 d4b5 c8d7 c3c4 a7a6 b5c3 d8c7 g2g3 a8d8 c1f4 c7c8 a2a4 h7h6 h2h4 b7b6 d1b3 d7c6 f3c6 c8c6 a1d1 f6h5 f4e5 d8d1 b3d1 e7f6 c3d5 f6e5 d1h5 e5f6 d5f6 g7f6 e1e8 c6e8 h5g4 g8f8 b2b3 e8e6 g4f3 f8g7 g1g2 e6d7 f3d5');
INSERT INTO evaluations VALUES(36421,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(36421,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(36421,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36421,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(36421,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36421,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36421,36,'cp 20','c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 c1f4 d7c5 e4c2 c5e6 d4e6 c8e6 c2b3 e6b3 d1b3 d8c8 f1e1 c7c6 g2g3 b7b6 b3c4 e7c5 c4d3 h7h6 a1d1 a7a5 g1g2 e8e1 d1e1 a5a4 h2h3 c5f8 c3c4 c8d8 e1d1 d8e7 f4d6 e7d6 d3d6');
INSERT INTO evaluations VALUES(52752,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(52752,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52752,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(52752,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52752,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52752,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52752,36,'cp 38','g1f3 g8f6 b1c3 e7e6 c1g5 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 b8c6 g5f4 c8d7 g2g4 e6e5 f3e5 c6e5 f4e5 d7g4 c4e2 g4e2 c3e2 c5e7 d1c1 f6d7 e5c7 f8c8 f1d1 d7f6 a2a4');
INSERT INTO evaluations VALUES(52749,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52749,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52749,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(52749,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(52749,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52749,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52749,36,'cp 20','g8f6 b1c3 e7e6 c1g5 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 h7h6 g5h4 c6c5 d4c5 d8d1 a1d1 e7c5 f3e5 c5b4 e1g1 b4c3 b2c3 f6d7 h4g3 d7e5 g3e5 b7b6 e5d6 f8e8 c4b5 c8d7 b5e2 d7c6 f2f3 b8d7 e3e4 d7c5 d1d2 e8c8 f1d1 c6a4 d1c1 c5b7 d6b4 e6e5 d2d5 b7c5');
INSERT INTO evaluations VALUES(44913,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(44913,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44913,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(44913,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(44913,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44913,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(44913,36,'cp 25','b1c3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 a7a6 h4f6 g7f6 g2g4 b8c6 c4d3 c5e7 d3e4 a8b8 c3a4 b7b5 a4c5 e7c5 e4c6 c5b6 b2b3 b5b4 h2h3 f8d8 f3d4');
INSERT INTO evaluations VALUES(44901,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(44901,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(44901,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(44901,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44901,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(44901,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(44901,36,'cp 28','e7e6 c1g5 f8e7 e2e3 h7h6 g5h4 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 c5b4 d1c1 c8d7 c4d3 b8c6 h1d1 f8d8 c3e4 b4e7 e4c5 d7e8 a2a3 d8c8 b2b4 b7b6 c5a6 e7d6 h4g3 d6g3 h2g3 f6d5');
INSERT INTO evaluations VALUES(45635,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(45635,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45635,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(45635,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(45635,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(45635,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45635,36,'cp 25','c1g5 h7h6 g5f6 d8f6 e2e3 b8d7 f1d3 d5c4 d3c4 g7g6 e1g1 f8g7 e3e4 e6e5 d4d5 d7b6 f3d2 f6g5 f1e1 e8g8 a2a4 f8d8 a4a5 c6d5 d2f3 g5e7 c4d5 b6d5 c3d5 e7d6 d1b3 c8e6 e1d1 e6d5 d1d5 d6c7 a1d1 d8d5 d1d5 c7c1 d5d1');
INSERT INTO evaluations VALUES(45610,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(45610,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(45610,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(45610,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(45610,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(45610,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(45610,36,'cp 30','h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 c6c5 d4c5 d5c4 d3c4 d8d1 a1d1 e7c5 e1e2 c5b4 d1c1 c8d7 f3e5 f8c8 c3b5 a7a6 b5d4 b4d6 f2f4 d6e5 f4e5 f6g4 d4f3 b8c6 h4g3 c6e7 h2h3 c8c4 c1c4 d7b5 h1c1');
INSERT INTO evaluations VALUES(45540,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45540,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(45540,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(45540,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45540,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(45540,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45540,36,'cp 28','g5f6 d8f6 e2e3 f8d6 f1d3 e8g8 d1c2 b8d7 e1g1 d5c4 d3c4 f6e7 a2a3 d7f6 e3e4 e6e5 d4e5 d6e5 a1e1 e5c3 c2c3 f8d8 f3h4 f6g4 c3g3 e7e5 e1d1 d8e8 f1e1 e5g3 h2g3 g4e5 c4e2 g7g6 f2f4 e5g4 e2f3 c8e6');
INSERT INTO evaluations VALUES(45538,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(45538,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45538,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(45538,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45538,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(45538,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45538,36,'cp 24','d8f6 e2e3 f8d6 f1d3 d5c4 d3c4 b8d7 a2a3 e8g8 e1g1 f6e7 d1c2 d7f6 e3e4 e6e5 d4e5 d6e5 a1e1 e5c3 c2c3 f8d8 c4d3 c8e6 f3d4 e6g4 d4b3 b7b6 d3b1 c6c5 e4e5 f6d5 c3g3 g4e6 f2f4 f7f5 e5f6 e7f6 b1d3 g8h8 g1h1');
INSERT INTO evaluations VALUES(33616,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(33616,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(33616,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(33616,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(33616,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(33616,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(33616,36,'cp 13','e2e3 b8d7 f1d3 d5c4 d3c4 f8d6 e1g1 e8g8 d1c2 f6e7 a2a3 d7f6 c3e2 f8e8 h2h3 e6e5 d4e5 d6e5 f3e5 e7e5 c2b3 e5c7 c4d3 c8e6 b3c3 a7a5 f1d1 a8d8 a1c1 a5a4 d3b1 e6b3 d1d8 e8d8 e2d4');
INSERT INTO evaluations VALUES(33615,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(33615,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(33615,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(33615,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(33615,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33615,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(33615,36,'cp 13','b8d7 f1d3 d5c4 d3c4 f8d6 e1g1 e8g8 a2a3 f6e7 d1c2 d7f6 e3e4 e6e5 d4e5 d6e5 a1e1 e5c3 c2c3 f8d8 e1e3 c8e6 f3d4 e6c4 c3c4 d8d7 e4e5 f6g4 d4c6 e7e6 c4e6 f7e6 e3g3 g4f2 f1f2 d7d1 f2f1 d1f1 g1f1 b7c6 g3c3 a8b8 c3c6 b8b2 c6e6 b2b1 f1f2 b1b2 f2f3 b2b3 f3e4 b3a3 e6e7 a3a2 g2g3 a2h2');
INSERT INTO evaluations VALUES(47150,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(47150,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(47150,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(47150,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(47150,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(47150,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(47150,36,'cp 105','c4d5 c7c6 d5c6 b8c6 g1f3 e7e5 d4e5 d8d1 e1d1 f6g4 d1e1 f8c5 e2e3 g4e5 f3e5 c6e5 c1d2 c8f5 f1b5 e8e7 d2c3 f7f6 b1d2 h8d8 d2b3 c5b6 b3d4 b6d4 c3d4 a8c8 d4c3 f5e6 f2f3 e6c4 b5c4 e5c4 e3e4 b7b5 a1c1 a7a5 a2a3 e7f7 e1e2 b5b4 a3b4 a5b4 c3b4 c4b2 c1c8');
INSERT INTO evaluations VALUES(47099,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(47099,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(47099,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(47099,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(47099,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(47099,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(47099,36,'cp 89','c7c6 d5c6 b8c6 g1f3 e7e5 d4e5 d8d1 e1d1 f6g4 d1e1 f8c5 e2e3 g4e5 f3e5 c6e5 c1d2 c8f5 d2c3 e5d3 f1d3 f5d3 b1d2 c5b6 d2b3 d3c4 h2h4 h7h5 b3d4 e8g8 h1g1 f7f6 f2f3 f6f5 g2g3 b6d4 c3d4 a8c8 d4c3 f8d8 e1f2 d8d3 g1d1 c8d8 d1d3 d8d3 a1c1 c4b5 a2a3 b5c6 c3b4 d3b3');
INSERT INTO evaluations VALUES(44869,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(44869,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(44869,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(44869,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(44869,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(44869,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(44869,36,'cp 98','d5c6 b8c6 g1f3 e7e5 d4e5 d8d1 e1d1 f6g4 d1e1 f8c5 e2e3 g4e5 f3e5 c6e5 c1d2 c8f5 d2c3 e5d3 f1d3 f5d3 b1d2 c5b6 d2b3 d3c4 h2h4 f7f5 h4h5 e8g8 g2g3 a8c8 a1d1 b6d8 d1d2 d8f6 c3f6 f8f6 h1h4 f6a6 d2c2 b7b5 h4f4 a6a2 f4f5');
INSERT INTO evaluations VALUES(44841,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(44841,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(44841,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(44841,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(44841,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(44841,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(44841,36,'cp 93','b8c6 g1f3 e7e5 d4e5 d8d1 e1d1 f6g4 d1e1 f8c5 e2e3 g4e5 f3e5 c6e5 c1d2 c8f5 d2c3 e5d3 f1d3 f5d3 b1d2 c5b6 h2h4 f7f6 d2b3 e8f7 c3d4 b6c7 a1c1 c7d6 b3c5 d6c5 d4c5 a8c8 f2f3 b7b6 e1d2 d3a6 c5d4 c8c1 h1c1 h8c8 c1h1 c8d8 a2a4 f7g6 h1a1 h7h5 a4a5 a6b5 a5a6 b5c6 a1c1');
INSERT INTO evaluations VALUES(13191,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(13191,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(13191,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(13191,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(13191,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(13191,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(13191,36,'cp 91','g1f3 e7e5 d4e5 d8d1 e1d1 f6g4 d1e1 f8c5 e2e3 g4e5 f3e5 c6e5 c1d2 c8f5 d2c3 e5d3 f1d3 f5d3 b1d2 c5b6 h2h4 e8g8 h4h5 f7f5 g2g3 f8d8 h1h4 a8c8 a1d1 d3c2 d1c1 c2d3 h4b4 d8d7 b4f4 h7h6 c1d1 d3c2 d1a1 d7d8 a2a4 b6c5 d2f3 c2e4 f3h4 c5d6 h4f5 d6f4');
INSERT INTO evaluations VALUES(13189,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(13189,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(13189,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(13189,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(13189,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(13189,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(13189,36,'cp 97','e7e5 d4e5 d8d1 e1d1 f6g4 d1e1 f8c5 e2e3 g4e5 f3e5 c6e5 c1d2 c8f5 d2c3 e5d3 f1d3 f5d3 b1d2 c5b6 h2h4 h7h5 d2b3 a8c8 e1d2 d3e4 f2f3 e4d5 b3d4 f7f6 d4f5 c8c7 a2a4 c7d7 d2e2 e8f7 a4a5 b6c7 h1c1 h8d8 c3d4');
INSERT INTO evaluations VALUES(57673,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(57673,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(57673,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(57673,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(57673,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(57673,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(57673,36,'cp 68','d2d4 f8g7 b1c3 d7d6 c1e3 a7a6 d1d2 b7b5 f2f3 b8d7 h2h4 h7h5 g1h3 c8b7 h3g5 c7c5 d4c5 d7c5 e3d4 g7d4 d2d4 e7e5 d4d2 b5b4 c3d1 f7f6 g5h3 d6d5 d2b4 d8e7 d1f2 d5e4 f3e4 c5e4 b4a4 e7d7 a4d7 e8d7 e1c1 e4d6 c1b1');
INSERT INTO evaluations VALUES(57663,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(57663,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(57663,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(57663,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(57663,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(57663,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(57663,36,'cp 57','f8g7 b1c3 d7d5 e4d5 g8f6 f1c4 c7c6 d5c6 b8c6 c1e3 f6g4 g1f3 e8g8 c4b3 g4e3 f2e3 e7e5 d4d5 e5e4 d5c6 e4f3 d1d8 f8d8 g2f3 b7c6 a1d1 d8d1 e1d1 g8f8 c3a4 h7h5 d1e2 g6g5 h1d1 g5g4 d1d6 g7e5 d6c6 e5h2');
INSERT INTO evaluations VALUES(39607,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(39607,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(39607,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(39607,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(39607,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(39607,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(39607,36,'cp 69','b1c3 d7d5 c3d5 c7c6 d5e3 d8d4 d1d4 g7d4 f2f3 d4g7 a2a4 g8f6 c2c3 e8g8 g1e2 f8d8 e3c4 c6c5 e2f4 b7b6 g2g4 b8c6 h2h4 e7e5 f4g2 h7h5 g4g5 f6e8 e1f2 c8e6 g2e3 e8d6 f1e2 d6c4 e2c4 c6a5 c4d5 a8b8 h1d1 a5b3');
INSERT INTO evaluations VALUES(39603,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(39603,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(39603,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(39603,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(39603,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(39603,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(39603,36,'cp 62','c7c6 c1e3 d7d6 f1e2 d8b6 b2b3 g8f6 g1f3 e8g8 e1g1 b6c7 a2a4 f6g4 e3c1 b8d7 h2h3 g4f6 c1e3 e7e5 d4e5 d7e5 d1d2 e5f3 e2f3 c7a5 a1d1 f8e8 c3d5 a5d2 d5f6 g7f6 d1d2 f6e5 f1d1 c8e6 c2c4 e8d8 a4a5 d8e8 d2d6 e5d6');
INSERT INTO evaluations VALUES(37737,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(37737,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(37737,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(37737,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(37737,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37737,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(37737,36,'cp 75','f2f4 d7d5 e4e5 h7h5 g1f3 c8f5 c1e3 g7f8 h2h3 e7e6 f1e2 b8d7 d1d2 d8a5 a2a3 f8e7 e3f2 a7a6 e1g1 g8h6 f1c1 a5d8 c3d1 a6a5 a3a4 e8g8 e2d3 c6c5 c2c3 f5e4 d2e2 e4d3 e2d3 c5d4 c3d4 h6f5');
INSERT INTO evaluations VALUES(37744,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(37744,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(37744,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(37744,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(37744,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(37744,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(37744,36,'cp 73','h7h5 e4e5 d7d5 h2h3 c8f5 g1f3 e7e6 c1e3 b8d7 f1e2 g7f8 f3g5 d8b6 a1b1 b6c7 e1g1 g8h6 b2b4 f8e7 a2a4 a7a5 b4b5 e7b4 c3a2 c6b5 a2b4 a5b4 e2b5 e8g8 b1b4 d7b6 c2c3 c7c3 e3d2 c3c7 d1e2 f8c8 f1c1 c7c1 d2c1');
INSERT INTO evaluations VALUES(23450,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(23450,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(23450,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(23450,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(23450,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(23450,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(23450,36,'cp 40','d2d4 d7d5 e4e5 c8f5 c2c3 e7e6 f1d3 f5d3 d1d3 d8d7 g1h3 g8e7 e1g1 e7g6 h3f4 g6f4 c1f4 c6e7 b1d2 b7b6 a2a4 a7a5 d2f3 e7f5 h2h4 h7h5 g2g3 c7c5 f1c1 a8b8 c1e1 f8e7 g1g2 e8g8 d3e2 f8c8 f4g5 e7g5');
INSERT INTO evaluations VALUES(23441,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(23441,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(23441,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(23441,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(23441,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(23441,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(23441,36,'cp 27','d7d5 e4e5 c8f5 c2c3 e7e6 f1d3 f5d3 d1d3 d8d7 g1e2 f7f6 f2f4 e8c8 e1g1 g7g5 b2b4 g5f4 c1f4 f6e5 d4e5 g8e7 b4b5 c6a5 b1d2 e7g6 f4g5 d8e8 g5f6 f8g7 d3e3 b7b6 c3c4 g7f6 e5f6 d5c4 a1d1 d7b5 e2d4 b5e5 d2e4 e8d8 f6f7 d8d5 d4f3 e5f4 e3f4 g6f4 d1d5 e6d5 f3e5');
INSERT INTO evaluations VALUES(21045,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(21045,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(21045,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(21045,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(21045,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(21045,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(21045,36,'cp 48','e4e5 c8f5 c2c3 e7e6 f1d3 f5d3 d1d3 d8d7 g1e2 f7f6 f2f4 g7g5 e5f6 g5g4 f4f5 g8f6 f5e6 d7e6 e1g1 f8d6 b1a3 d6a3 b2a3 e8c8 c1g5 d8f8 e2f4 e6d7 g5f6 f8f6 f4h5 f6f1 a1f1 d7e7 h5f6 e7a3 f6g4 a7a5 g4e5 c6e5 d4e5');
INSERT INTO evaluations VALUES(21086,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(21086,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(21086,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(21086,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(21086,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(21086,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(21086,36,'cp 39','c8f5 c2c3 e7e6 f1d3 f5d3 d1d3 d8d7 g1e2 f7f6 f2f4 e8c8 b2b4 g7g5 e1g1 g5f4 c1f4 f6e5 d4e5 g8e7 b4b5 c6a5 b1d2 e7g6 f4g5 f8e7 g5e7 d7e7 d3e3 h8f8 a2a4 b7b6 e2d4 d8e8 a1e1 e7h4 d2f3 h4g4 e3g5 g4f4 g2g3 f4g5 f3g5 f8f1 g1f1 a5c4');
INSERT INTO evaluations VALUES(26654,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(26654,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(26654,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(26654,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(26654,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(26654,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26654,36,'cp 35','c2c3 e7e6 f1d3 f5d3 d1d3 d8d7 g1e2 f7f6 f2f4 e8c8 b2b4 g7g5 e1g1 g5f4 c1f4 f6e5 d4e5 g8e7 b4b5 c6a5 b1d2 e7g6 f4g5 f8e7 g5f6 e7f6 e5f6 e6e5 c3c4 d5c4 d3d7 d8d7 d2e4 b7b6 a1c1 h8f8 e4g5 h7h6 g5e4 a5b7 c1c4');
INSERT INTO evaluations VALUES(26651,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(26651,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(26651,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(26651,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(26651,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26651,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26651,36,'cp 23','e7e6 f1d3 f5d3 d1d3 d8d7 g1e2 e8c8 f2f4 f7f6 b2b4 g7g5 e1g1 g5f4 c1f4 f6e5 d4e5 g8e7 b4b5 c6a5 b1d2 e7g6 f4g5 f8e7 g5e7 d7e7 d3e3 h8f8 d2b3 f8f1 a1f1 a5b3 a2b3 d8f8 f1a1 e7g7 a1a7 c8b8 a7a1 g7e5 e3e5 g6e5');
INSERT INTO evaluations VALUES(26849,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(26849,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(26849,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(26849,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(26849,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(26849,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(26849,36,'cp 55','f1e2 f7f6 f2f4 g7g5 e2h5 e8d7 f4g5 f6e5 g1f3 h7h6 c3c4 d5c4 f3e5 c6e5 d4e5 f5d3 h5e2 d7c8 b1c3 h6g5 e2d3 c4d3 d1g4 g8h6 g4e6 d8d7 e6c4 d7g4 c4g4 h6g4 c1g5 c8d7 e1d2 g4e5 h2h4 f8h6 g5h6 h8h6 a1e1 h6e6 c3d5 a8g8 d5f4 e5c4 d2d3 e6f6 d3c4 f6f4 c4d3 f4f2 h4h5 f2b2 e1g1 g8g3 d3e4 b2g2 g1g2 g3g2');
INSERT INTO evaluations VALUES(55056,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55056,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(55056,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55056,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(55056,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55056,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55056,36,'cp 6','e7e6 b2b3 c7c5 g1f3 g8f6 f1d3 b7b6 e1g1 c8b7 c1b2 c5d4 e3d4 f8d6 b1d2 e8g8 d1e2 b8c6 a2a3 a8c8 f3e5 c6e7 c2c3 e7f5 a3a4 d6e7 b3b4 f5d6 f2f3 f6h5 g2g3 a7a5 b4a5 b6a5 f3f4 h5f6 b2a3 f6e4');
INSERT INTO evaluations VALUES(55856,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55856,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(55856,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(55856,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(55856,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55856,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(55856,36,'cp 12','g1f3 g8f6 f1d3 b7b6 e1g1 c8b7 b2b3 f8e7 c2c4 d5c4 b3c4 c7c5 c1b2 b8c6 b1d2 e8g8 a1c1 c5d4 e3d4 a8c8 f1e1 f8e8 d2f1 e7f8 f1e3 c8c7 d3b1 c6e7 h2h3 e7g6 a2a3 d8e7 f3d2 g6f4');
INSERT INTO evaluations VALUES(55859,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(55859,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55859,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55859,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55859,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55859,36,'cp 0','g8f6 b2b3 f8d6 b1d2 b8d7 c2c4 b7b6 f1d3 e8g8 e1g1 c8b7 c1b2 c7c5 c4d5 f6d5 a1c1 a8c8 d2c4 d6c7 d1e2 c5d4 b2d4 d8e7 f1d1 d7c5 d3c2 f7f5 a2a3 c5e4 c2b1 e4c5 b1c2');
INSERT INTO evaluations VALUES(55052,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(55052,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55052,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(55052,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(55052,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55052,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(55052,36,'cp -11','g8f6 c1f4 c7c5 e2e4 c5d4 d1d4 b8c6 f1b5 c8d7 b5c6 d7c6 e4d5 f6d5 e1c1 e7e6 g1f3 d5f4 d4f4 d8f6 f4e3 f6h6 e3h6 g7h6 f3e5 c6g2 h1g1 g2c6 e5c6 b7c6 c3e4 f8e7 e4d6 e8f8 c2c4 a8d8 d6b7 d8d1 c1d1 c6c5 d1e2 e7f6 b7c5 f6b2 g1d1 h8g8 d1d8 f8g7 d8g8 g7g8 c5b3 b2f6 b3a5 g8f8 c4c5');
INSERT INTO evaluations VALUES(47178,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47178,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47178,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47178,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47178,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47178,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47178,36,'cp -6','c1f4 e7e6 c3b5 f8b4 c2c3 b4a5 e2e3 e8g8 g1f3 a7a6 b5a3 b8d7 h2h3 c7c6 f1e2 a5c7 e1g1 f8e8 f4c7 d8c7 c3c4 e6e5 d4e5 d7e5 c4d5 f6d5 a3c2 c8f5 c2d4 f5e4 a1c1 a8d8 d1b3 e5f3 e2f3 d5f6 f1d1 g7g6 a2a3 c7e7 b3c3 f6d5 c3d2 d5f6');
INSERT INTO evaluations VALUES(47152,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47152,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47152,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47152,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47152,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(47152,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47152,36,'cp 0','e7e6 c3b5 b8a6 e2e3 f8e7 g1f3 e8g8 h2h4 c7c6 b5c3 d8a5 f3d2 a5b6 c3a4 b6a5');
INSERT INTO evaluations VALUES(48280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48280,36,'cp 0','c3b5 b8a6 e2e3 f8e7 g1f3 e8g8 h2h4 c7c6 b5c3 d8a5 f3d2 c6c5 g2g4 c5d4 d2b3 a5b6 e3d4 e6e5 d4e5 f6g4 d1d2 e7b4 f2f3 g4e5 e1c1 e5c4 f1c4 d5c4 f4e3 b6c7 e3f4');
INSERT INTO evaluations VALUES(48160,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48160,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48160,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(48160,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48160,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48160,36,'cp -2','f8b4 c2c3 b4a5 e2e3 a7a6 b5a3 e8g8 g1f3 d8e7 a3b1 f6h5 b1d2 h5f4 e3f4 f7f6 g2g3 a5b6 f1g2 c7c5 d4c5 e7c5 e1g1 b8c6 d2b3 c5d6 c3c4 f8d8 a1c1 d5c4 d1d6 d8d6 c1c4 a8b8 f1c1 g8f7 b3c5 c8d7 c5d7 d6d7 b2b4 a6a5 a2a3 a5b4 a3b4');
INSERT INTO evaluations VALUES(57151,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(57151,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(57151,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(57151,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(57151,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(57151,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(57151,36,'cp 8','b8c6 g1f3 g8f6 d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 c6d5 e1g1 e8g8 h2h3 c7c6 d1f3 b4d6 c1f4 c8e6 b2b3 f8e8 c3e2 a7a5 f1e1 d8b8 f4d6 b8d6 f3f4 d6f4 e2f4 g8f8 f2f3 f8e7 c2c4 g7g5 f4e2 e7d6 a1d1 a8d8 e2d4 d5c4');
INSERT INTO evaluations VALUES(22783,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(22783,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(22783,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(22783,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(22783,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(22783,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(22783,36,'cp 12','g1f3 g8f6 f1b5 c6d4 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d2c3 f8c5 d1e2 f6e6 e2e6 f7e6 f2f3 e8g8 c1d2 d7d6 b5d3 e6e5 e1c1 c8e6 a2a4 a7a5 h1e1 e6f5 e1e2 a8e8 h2h4 h7h6 h4h5 f5h7 d1e1 c7c6 d3h7 g8h7 e2e4 h7g8');
INSERT INTO evaluations VALUES(52435,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52435,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(52435,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(52435,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52435,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(52435,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(52435,36,'cp 0','c5d4 g1f3 e7e5 c2c3 b8c6 c3d4 e5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 e2f4 b4d6 f4d5 c6d5 c1e3 c8e6 e3d4 a7a5 a1c1 f8b8 c1c6 d6e7 f1d1 a5a4 d4e5 b8b4 e5c3 b4b8');
INSERT INTO evaluations VALUES(52529,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52529,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52529,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52529,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52529,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(52529,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52529,36,'cp 7','g1f3 e7e5 c2c3 b8c6 c3d4 e5d4 f3d4 f8b4 b1c3 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 f8d8 f1d1 b4e7 d3e4 a8b8 e2d4 c8d7 e4d5 c6d5 b2b3 a7a5 c1f4 b8c8 a1c1 e7c5 a2a4 f7f6 f4e3 g8f7 d4b5 c5e3 c1c8 d7c8 f2e3 c8e6');
INSERT INTO evaluations VALUES(52528,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(52528,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52528,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52528,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(52528,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(52528,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(52528,36,'cp 4','e7e5 c2c3 b8c6 c3d4 e5d4 f3d4 f8b4 b1c3 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 e2f4 a7a5 d3e4 f8d8 e4f3 c8f5 f4d5 c6d5 f1d1 b4e7 c1e3 e7f6 e3d4 f5e6 d4f6 g7f6');
INSERT INTO evaluations VALUES(52181,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(52181,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52181,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(52181,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(52181,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(52181,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(52181,36,'cp 12','c2c3 b8c6 c3d4 e5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4d6 e2d4 c8d7 f1d1 f8b8 d4f3 d6e7 f3e5 d7e8 d3e4 e7f6 e5c6 e8c6 e4d5 c6d5 d1d5 f6b2 a1b1 b2c1 b1c1 b8b7 a2a4 g7g6 a4a5 g8g7 c1c2 a8b8 d5d6');
INSERT INTO evaluations VALUES(52179,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(52179,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(52179,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(52179,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52179,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(52179,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(52179,36,'cp 4','b8c6 c3d4 e5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 f1d1 b4e7 e2f4 f8d8 d3e4 c8g4 d1e1 e7b4 e1f1 b4d6 h2h3 d5f4 c1f4 d6f4 h3g4 f4e5 a1b1 a8b8 b2b4 d8d2 e4c6 d2a2 f1d1 a2b2 b1b2 e5b2 d1d7');
INSERT INTO evaluations VALUES(19064,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19064,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(19064,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19064,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(19064,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(19064,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(19064,36,'cp 9','c3d4 e5d4 f3d4 f8b4 b1c3 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4e7 e2f4 d5f4 c1f4 c8e6 f4e5 f8d8 d3a6 f7f6 e5f4 e7d6 f4d6 d8d6 f1c1 a8d8 a6f1 a7a5 b2b3 g8f7 c1c5 d8a8 a1c1 a5a4 c5c6 d6c6 c1c6');
INSERT INTO evaluations VALUES(19062,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19062,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(19062,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(19062,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(19062,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(19062,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19062,36,'cp 5','e5d4 f3d4 f8b4 b1c3 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 f8d8 f1d1 b4e7 c1d2 a7a5 d3e4 e7f6 d2c3 d5c3 d1d8 f6d8 b2c3 d8f6 e2d4 c8d7 e4c6 d7c6 d4c6 f6c3 a1c1 c3d2 c1d1 d2b4 g1f1 g7g6 c6b4 a5b4');
INSERT INTO evaluations VALUES(19092,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(19092,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19092,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(19092,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19092,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19092,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19092,36,'cp 5','f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 f8d8 f1d1 b4e7 c1d2 a7a5 d3e4 e7f6 d2c3 d5c3 d1d8 f6d8 b2c3 d8f6 e2d4 c8a6 a1b1 a6c4 a2a3 g7g6 b1b6 a8c8 d4c6 c4e6 b6a6 g8g7 c6d4 c8c3 d4e6 f7e6');
INSERT INTO evaluations VALUES(19090,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(19090,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19090,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19090,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19090,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(19090,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(19090,36,'cp 7','g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4d6 e2d4 c8d7 d4f3 f8b8 f1d1 d7e8 b2b3 a7a5 d3c4 a5a4 c1b2 h7h6 b2d4 f7f6 b3a4 a8a4 c4b3 a4a6 f3d2 e8f7 d2e4 d6f8 d4c5 d5f4 c5f8');
INSERT INTO evaluations VALUES(12614,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(12614,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(12614,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12614,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(12614,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(12614,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(12614,36,'cp 9','b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 f1d1 b4d6 b2b3 a7a5 c1d2 d6a3 a1b1 h7h6 e2d4 c8g4 d3e2 g4e2 d4e2 d5b4 d2b4 a3b4 e2d4 c6c5 d4c6 a5a4 g1f1 f8e8 c6b4 c5b4 b3a4 a8a4');
INSERT INTO evaluations VALUES(12613,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(12613,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(12613,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(12613,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(12613,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(12613,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(12613,36,'cp 7','f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 a7a5 f1d1 b4d6 d3c2 a5a4 e2c3 d5c3 d1d6 c3e2 g1h2 e2c1 a1c1 c8e6 b2b3 a4b3 c2b3 e6b3 a2b3 a8a2 h2g1 a2b2 d6d3 g7g6 c1c6 f8b8 g1h2 b2f2');
INSERT INTO evaluations VALUES(7878,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(7878,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(7878,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7878,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(7878,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(7878,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(7878,36,'cp 2','d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 a7a5 f1d1 b4d6 d3e4 f8e8 e4d5 c6d5 c1e3 c8e6 a1c1 e8b8 b2b3 a5a4 c1c6 d6f8 b3a4 a8a4 e2f4 e6f5 f4d5 a4a2 d5b6 a2a6 b6d7 a6c6 d7b8');
INSERT INTO evaluations VALUES(7818,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7818,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(7818,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7818,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(7818,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(7818,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(7818,36,'cp 0','b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4d6 e2d4 c8d7 d4f3 f8b8 f1d1 a7a5 d3c4 d7e8 b2b3 a5a4 c1b2 h7h6 b2e5 d6e5 f3e5 d5b6 c4e2 b6d5');
INSERT INTO evaluations VALUES(7671,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(7671,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7671,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7671,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7671,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(7671,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(7671,36,'cp 8','f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 a7a5 f1d1 b4d6 a2a3 f8e8 a1b1 a5a4 c1d2 d6e5 e2c3 c8e6 d3c2 d5c3 d2c3 e6a2 b1a1 e5c3 a1a2 c3f6 b2b4 a4b3 c2b3 a8b8 b3c2 b8a8 a3a4 e8b8 a4a5');
INSERT INTO evaluations VALUES(49613,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(49613,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(49613,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(49613,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(49613,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(49613,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(49613,36,'cp 0','b8c6 f1b5 c6d4 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d2c3 f8c5 d1e2 f6e6 e2e6 f7e6 f2f3 e8g8 b5d3 d7d6 c1d2 e6e5 e1c1 c8e6 a2a4 a7a5 h1e1 h7h6 b2b3 a8e8 e1e2 e6f5 d3c4 f5e6 c4b5 c7c6 b5d3 e6f5 d1e1 b7b6 h2h4 f5d3 c2d3 f8f5 d2e3 c5e3 e2e3 g8f7 c3c4');
INSERT INTO evaluations VALUES(52440,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52440,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52440,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52440,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(52440,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(52440,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52440,36,'cp 24','a7a6 g1f3 b7b5 d2d4 c5d4 f3d4 e7e6 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 f1e1 f8c5 d1h5 c5d4 a2a4 b5a4 c3a4 g8f6 h5g5 d7d6 c1e3 d4e3 e1e3 e8g8 e3g3 g7g6 a4c3 a6a5 h2h4 f6d7 g5h6 d8f6 h4h5 a5a4 a1d1 f6g7');
INSERT INTO evaluations VALUES(50605,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50605,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(50605,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(50605,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(50605,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(50605,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(50605,36,'cp 41','g1f3 e7e6 d2d4 c5d4 f3d4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 f1e1 g8e7 d3f1 e7g6 c1e3 d8b8 a2a4 f8d6 g2g3 d6b4 a4b5 a6b5 a1a8 c6a8 d1d4 b4c3 d4c3 e8g8 e3d4 f8c8 c3d2 a8c6 d2b4 g6e5 f1g2 h7h6 e1a1 d7d6 a1d1 f7f6 d4c3 e5f7 h2h4 e6e5');
INSERT INTO evaluations VALUES(50603,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(50603,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(50603,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(50603,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(50603,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(50603,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(50603,36,'cp 43','b7b5 d2d4 c5d4 f3d4 e7e6 f1d3 c8b7 e1g1 d8b6 c1e3 f8c5 c3e2 b8c6 c2c3 b5b4 d4c6 b7c6 e3c5 b6c5 a2a3 b4c3 b2b4 c5e5 f2f4 e5b8 e2c3 g8f6 a1c1 e8g8 d1e2 f8c8 h2h3 g7g6 e4e5 f6d5 c3e4 b8b6 g1h2 c6b5 e4d6 c8c1 f1c1');
INSERT INTO evaluations VALUES(51243,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(51243,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(51243,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(51243,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(51243,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(51243,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(51243,36,'cp 50','d2d4 c5d4 f3d4 e7e6 f1d3 c8b7 e1g1 d8b6 d4f3 b6c7 f1e1 f8c5 c1e3 g8e7 d1d2 e8g8 a2a4 c5e3 e1e3 b5a4 a1a4 f8d8 d3f1 d7d5 e4d5 e7d5 c3d5 b7d5 e3c3 c7b6 d2d4 b6b7 b2b3 b8d7 f3d2 h7h6 h2h3 a6a5 d2c4 b7c7');
INSERT INTO evaluations VALUES(51240,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(51240,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(51240,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(51240,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(51240,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(51240,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(51240,36,'cp 43','c5d4 f3d4 e7e6 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 f1e1 d8b8 c3d5 f8d6 h2h3 g8e7 d5e7 d6e7 c2c3 b8e5 c1e3 e8g8 d1e2 e5f6 e3d4 f6g5 a1d1 e7c5 d4c5 g5c5 e2d2 a6a5 a2a3 a8c8 d3f1 f7f5 b2b4 c5e5 e4f5 e5f5');
INSERT INTO evaluations VALUES(51994,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(51994,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(51994,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(51994,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(51994,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(51994,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(51994,36,'cp 36','b1c3 a7a6 d2d4 c5d4 f3d4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 f1e1 d8b8 c3d5 f8d6 h2h3 g8e7 d5e7 d6e7 c2c3 b8e5 c1e3 e8g8 d1e2 e5f6 e3d4 f6g5 g2g3 e7c5 h3h4 g5e7 e2d2 h7h6 a2a3 a8c8 a1d1 c8c7 d3f1 d7d6 d2e3');
INSERT INTO evaluations VALUES(51979,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(51979,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(51979,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(51979,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(51979,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(51979,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(51979,36,'cp 31','b8c6 d2d4 c5d4 f3d4 g8f6 f1e2 d8c7 c1e3 a7a6 e1g1 f8e7 f2f4 d7d6 d1e1 c6d4 e3d4 e8g8 e1g3 b7b5 a2a3 c8b7 a1d1 a8d8 e2f3 f6e8 g3f2 d8c8 f1e1 e8f6 e4e5 f6e8 f3b7 c7b7 b2b4 d6e5 f4e5 f7f5 e5f6 e7f6 c3e4 f6d4');
INSERT INTO evaluations VALUES(18811,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(18811,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(18811,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(18811,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(18811,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(18811,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(18811,36,'cp 24','d2d4 c5d4 f3d4 g8f6 f1e2 f8b4 e1g1 b4c3 b2c3 f6e4 d4c6 d7c6 d1d8 e8d8 e2f3 e4f6 f1e1 h8e8 c3c4 e6e5 c1b2 e5e4 b2f6 g7f6 f3e4 f6f5 e4d3 c8d7 e1e8 d8e8 f2f4 d7e6 a1e1 e8f8 a2a4 a8d8 g1f2 b7b6 a4a5 c6c5 a5b6 a7b6 e1b1 d8d4 f2e3 e6c4 d3f5 c4d5 b1b6 d5g2 f5h7');
INSERT INTO evaluations VALUES(18799,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(18799,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(18799,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(18799,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(18799,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(18799,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(18799,36,'cp 26','c5d4 f3d4 a7a6 c1e3 g8f6 a2a3 f8e7 f2f4 d7d6 h2h3 e8g8 d1f3 c6d4 e3d4 b7b5 f1d3 f6d7 e1c1 c8b7 h1f1 a8c8 f3g3 e6e5 f4e5 e7h4 g3f3 d7e5 d4e5 d6e5 c1b1 d8f6 f3f6 h4f6 c3d5');
INSERT INTO evaluations VALUES(18921,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(18921,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(18921,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(18921,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18921,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(18921,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(18921,36,'cp 44','f3d4 g8f6 d4c6 b7c6 e4e5 f6d5 c3e4 c8b7 f1e2 c6c5 e1g1 d8c7 e4d6 f8d6 e5d6 c7c6 f2f3 c5c4 d1d4 e8g8 e2c4 c6d6 b2b3 f8c8 a2a4 d6b6 f1d1 a7a5 h2h4 b6d4 d1d4 d5b4 d4d7 b4c2 a1b1 b7a6 c4a6 a8a6 d7b7');
INSERT INTO evaluations VALUES(18906,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(18906,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(18906,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(18906,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(18906,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(18906,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(18906,36,'cp 32','g8f6 d4c6 b7c6 e4e5 f6d5 c3e4 c8b7 f1e2 c6c5 e1g1 d8c7 e4d6 f8d6 e5d6 c7c6 f2f3 c6b6 c2c4 d5b4 c1e3 a7a5 d1d2 e8g8 f1e1 a8e8 e2f1 b4c6 f3f4 c6d4 e3d4 c5d4 a1d1 b6d6 d2d4 d6d4 d1d4 d7d5 e1c1 d5c4 f1c4 e8c8 d4d7 b7c6 d7a7 a5a4 b2b4 a4b3');
INSERT INTO evaluations VALUES(12478,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(12478,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12478,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(12478,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12478,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(12478,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(12478,36,'cp 22','d4c6 b7c6 e4e5 f6d5 c3e4 d8c7 f2f4 c7b6 c2c4 d5e3 d1d3 f8b4 c1d2 e8g8 a1c1 a7a5 d2b4 b6b4 d3c3 f7f6 a2a3 b4c3 c1c3 e3f5 e5f6 a8b8 b2b3 d7d5 f6g7 f5g7 e4d6 f8f4 c4d5 e6d5 c3c6 c8g4 h2h3 g4e6 f1b5 g7f5 d6f5 e6f5 g2g4 f5g4');
INSERT INTO evaluations VALUES(12293,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(12293,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12293,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(12293,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(12293,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(12293,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(12293,36,'cp 35','b7c6 e4e5 f6d5 c3e4 c8b7 f1e2 c6c5 e1g1 d8c7 e4d6 f8d6 e5d6 c7c6 f2f3 c5c4 d1d4 e8g8 e2c4 c6d6 b2b3 f8c8 f1d1 d6b6 d4b6 d5b6 c4d3 a7a5 c1d2 b7a6 c2c4 d7d5 d2a5 d5c4 d3e4 b6d5 e4d5 e6d5 d1d5 a6b7 d5b5 b7c6 b5c5 c6d7 c5c8 a8c8 a5d2 c4c3 d2c1 f7f6 g1f2 h7h5 c1e3 h5h4 a1c1 g8f7');
INSERT INTO evaluations VALUES(11977,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(11977,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(11977,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(11977,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(11977,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(11977,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(11977,36,'cp 31','e4e5 f6d5 c3e4 c8b7 f1e2 c6c5 e1g1 d8c7 e4d6 f8d6 e5d6 c7c6 f2f3 c5c4 d1d4 e8g8 e2c4 c6d6 b2b3 f8c8 c4d3 d6b6 d4b6 a7b6 c2c4 b7a6 c1d2 d5f6 f1c1 d7d5 d3f1 d5c4 b3c4 f6d7 a2a4 h7h6 a4a5 b6a5 d2a5 a6b7 g1f2 b7c6');
INSERT INTO evaluations VALUES(11972,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(11972,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(11972,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(11972,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(11972,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(11972,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(11972,36,'cp 42','f6d5 c3e4 c8b7 f1e2 d8a5 c2c3 c6c5 e1g1 a5c7 e4d6 f8d6 e5d6 c7c6 f2f3 c6b6 c3c4 d5b4 c1e3 a7a5 d1d2 e8g8 f1e1 a5a4 e2f1 f7f6 h2h3 a8a5 e1b1 f8a8 b1d1 b4c6 a1b1 c6e5 d2f2 b7c6 g1h1 b6a7 d1d2 a8b8 b2b4 a4b3 a2b3 g8h8 d2d1 e5g6');
INSERT INTO evaluations VALUES(11969,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(11969,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(11969,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(11969,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(11969,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(11969,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(11969,36,'cp 53','c3e4 c8b7 f1e2 c6c5 e1g1 d8c7 e4d6 f8d6 e5d6 c7c6 f2f3 c6b6 c2c4 d5b4 c1e3 a7a5 f1e1 e8g8 d1d2 a5a4 e2f1 f7f6 h2h3 b4c6 e1d1 c6e5 d2f2 f8c8 a1b1 b6c6 e3f4 e5f7 b2b4 a4b3 a2b3 a8a5 f1d3 c8a8 b3b4 c5b4 b1b4 a5a2 d3e4 a2f2 e4c6');
INSERT INTO evaluations VALUES(11968,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(11968,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11968,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11968,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11968,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11968,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11968,36,'cp 37','c6c5 f1e2 c8b7 e1g1 d8c7 e4d6 f8d6 e5d6 c7c6 f2f3 c6b6 c2c4 d5b4 c1e3 a7a5 f1e1 e8g8 e2f1 a5a4 d1d2 b4c6 a1d1 a8a5 d1b1 b7a8 e1d1 f7f6 h2h3 b6a7 f3f4 c6b4 a2a3 b4c6 d2f2 e6e5 f1d3 c6d4 f4e5 f6e5 f2h4 g7g6');
INSERT INTO evaluations VALUES(16383,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(16383,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(16383,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(16383,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(16383,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(16383,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(16383,36,'cp -11','f6e4 c3e4 d7d5 c4d3 d5e4 d3e4 f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 f8e8 e1c1 c8e6 h1e1 e6a2 b2b3 d6a3 c1d2 d8d4 f3d4 a3b4 c2c3 b4c5 d1a1 c5d4 e3d4 e8e1 d2e1 a2b3 a1a7 a8a7 d4a7 f7f6 e1d2 b3c4 g2g3 g8f7 a7c5 f7e6 c5f8 g7g6 d2e3 e6d5 g3g4 c4b5 h2h4 d5c4 f8g7 c4c3 g7f6 c3c4');
INSERT INTO evaluations VALUES(16825,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(16825,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16825,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16825,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16825,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(16825,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16825,36,'cp 0','c3e4 d7d5 c4d3 d5e4 d3e4 f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 c6c5 d4c3 d8d7 e1c1 d7f5 f3d2 f8e8 h1e1 c8b7 d2c4 d6h2 g2g3 b7f3 d1d3 h7h5 e3f4 e8e1 c3e1 f3e4 d3e3 e4c2 e3e8 a8e8 e1e8 g8h7 c4e3 f5e4 e8f7 e4g6 f7c7 c2b1 c7c5 b1a2 e3d1 a2b3 c1d2');
INSERT INTO evaluations VALUES(16820,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16820,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(16820,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16820,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(16820,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16820,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(16820,36,'cp -2','d7d5 c4d3 d5e4 d3e4 f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 f8e8 e1c1 c8e6 h1e1 e6a2 b2b3 d6a3 c1d2 d8d4 f3d4 a3b4 c2c3 b4c5 d1a1 c5d4 e3d4 e8e1 d2e1 a2b3 a1a7 a8a7 d4a7 f7f6 e1e2 b3c4 e2e3 g8f7 a7b8 c6c5 b8c7 f7e6 g2g4 e6d5 h2h4 c4b5 g4g5 d5c6');
INSERT INTO evaluations VALUES(15376,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(15376,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(15376,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(15376,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(15376,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(15376,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(15376,36,'cp -5','c4d3 d5e4 d3e4 f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 f8e8 e1c1 c8e6 h1e1 e6a2 b2b3 d6a3 c1d2 d8d4 f3d4 a3b4 c2c3 b4c5 d1a1 c5d4 e3d4 e8e1 d2e1 a2b3 a1a7 a8a7 d4a7 b3c4 e1d2 f7f6 d2e3 g8f7 e3d4 c4f1 g2g3 f7e6 a7b8 e6f5 f2f3 f5e6 b8c7 e6f5 c7b8 g7g5 b8d6 h7h5');
INSERT INTO evaluations VALUES(15385,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15385,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(15385,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(15385,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(15385,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(15385,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15385,36,'cp -7','d5e4 d3e4 f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 c6c5 d4c3 d8e8 e1c1 e8b5 b2b3 b5b4 c3b4 c5b4 f3d2 c8b7 d2c4 b7g2 h1g1 g2e4 c4d6 c7d6 d1d4 d6d5 d4b4 f8c8 c2c4 a7a5 b4a4 d5c4 a4c4 c8c4 b3c4 a8b8 c1d2 b8b2 d2c3 b2a2');
INSERT INTO evaluations VALUES(15400,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15400,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(15400,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(15400,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(15400,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15400,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15400,36,'cp -1','d3e4 f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 c6c5 d4c3 d8e8 e1c1 e8b5 b2b3 f8e8 a2a4 b5b4 c3b4 c5b4 f3d2 c8g4 d1e1 h7h6 f2f3 g4e6 d2e4 d6e5 e4c5 a7a5 e3h6 e6b3 c2b3 g7h6 c1c2 f7f5 c5d3 e5d6 g2g3 g8f7');
INSERT INTO evaluations VALUES(15396,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(15396,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(15396,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(15396,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15396,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(15396,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(15396,36,'cp -2','f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 f8e8 e1c1 c8e6 h1e1 e6a2 b2b3 d6a3 c1d2 d8d4 f3d4 a3b4 c2c3 b4c5 d1a1 c5d4 e3d4 e8e1 d2e1 a2b3 a1a7 a8a7 d4a7 f7f6 e1e2 b3c4 e2e3 g8f7 e3d4 c4f1 g2g3 f7e6 a7b8 e6d6 b8a7 f1e2 a7c5 d6e6 d4e3 e2f1 c5f8 g7g5 g3g4 e6d5 f8g7');
INSERT INTO evaluations VALUES(9214,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(9214,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(9214,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(9214,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(9214,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(9214,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(9214,36,'cp -8','d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 f8e8 e1c1 c8e6 h1e1 e6a2 b2b3 d6a3 c1d2 d8d4 f3d4 a3b4 c2c3 b4c5 d1a1 c5d4 e3d4 e8e1 d2e1 a2b3 a1a7 a8a7 d4a7 f7f6 e1e2 b3d5 g2g3 g8f7 a7c5 f7e6 e2e3 e6f5 c5f8 g7g6 f8e7 f5g4 e7f6 g4h3 f6e5 h3h2 e5c7');
INSERT INTO evaluations VALUES(22519,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(22519,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(22519,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(22519,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(22519,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(22519,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(22519,36,'cp 81','d2d4 e5d4 e1g1 g8f6 e4e5 d7d5 c4b5 f6e4 f3d4 c8d7 b5c6 b7c6 f2f3 e4g5 f3f4 g5e4 b1c3 f8c5 c3e4 d5e4 c1e3 e8g8 d1e2 c5b6 d4b3 d8e7 e3b6 a7b6 e2e4 f7f6 a2a4 f8e8 f1e1 f6e5 f4e5 c6c5 a4a5 e7f7 a5a6 d7f5 e4e2 f7g6 a6a7 f5c2');
INSERT INTO evaluations VALUES(22521,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(22521,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(22521,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(22521,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(22521,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(22521,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(22521,36,'cp 68','e5d4 e1g1 g8f6 e4e5 d7d5 c4b5 f6e4 f3d4 c8d7 b5c6 b7c6 f2f3 e4c5 f3f4 c5e4 b1c3 f8c5 c3e4 d5e4 c1e3 e8g8 d1e2 c5b6 a2a4 a7a5 d4b3 f7f5 h2h3 d8e7 f1d1 d7e6 e3b6 c7b6 b3d4 a8d8 e2f2 c6c5 d4c6 d8d1 a1d1 e7e8 d1d6 e6d5 c6a5 e8a4 d6d5 a4a5 c2c3');
INSERT INTO evaluations VALUES(22542,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(22542,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(22542,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(22542,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(22542,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(22542,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(22542,36,'cp 67','f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 e4e5 d7d5 e5f6 d8f6 c4d3 d5d4 d1e2 e8f8 e1g1 d4c3 b2c3 b4d6 a1b1 c8f5 c1e3 h6h5 c3c4 f5d3 c2d3 c6c5 b1b5 a7a6 b5a5 f6c3 e2d2 c3d2 e3d2 h5h4 d2e3 h8h5 d3d4 h4h3 g2g3 h5f5 d4c5');
INSERT INTO evaluations VALUES(22538,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(22538,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(22538,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(22538,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(22538,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(22538,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(22538,36,'cp 55','g8f6 b1c3 f8b4 d4c6 b7c6 e1g1 d7d6 e4e5 d6e5 d1d8 e8d8 c4f7 c8f5 f7b3 a7a5 c3a4 d8c8 c1e3 c8b7 c2c3 b4e7 a1d1 h8d8 a4c5 e7c5 e3c5 f5d3 f1e1 f6d7 c5e3 e5e4 f2f3 a5a4 b3e6 d7f6 e3d4 a4a3 b2b3 d8e8 e6f7 e8e7');
INSERT INTO evaluations VALUES(16184,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(16184,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(16184,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(16184,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(16184,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(16184,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(16184,36,'cp 62','b1c3 f8b4 d4c6 b7c6 e1g1 d7d6 e4e5 d6e5 d1d8 e8d8 c4f7 c8f5 f7b3 a7a5 c3a4 d8c8 c1e3 c8b7 a2a3 b4d6 f1d1 h8d8 h2h3 f6d7 d1d2 d7b6 a1d1 b6a4 b3a4 f5e6 d2d3 d6e7 d3d8 e7d8 a4b3 e6b3 c2b3 d8f6 d1d3 h6h5 g1f1 a5a4 b3a4 a8a4');
INSERT INTO evaluations VALUES(16181,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(16181,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(16181,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(16181,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(16181,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(16181,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(16181,36,'cp 57','f8b4 d4c6 b7c6 e1g1 d7d6 e4e5 d6e5 d1d8 e8d8 c4f7 c8f5 f7b3 a7a5 c3a4 d8c8 c1e3 h8d8 f1d1 c8b7 c2c3 d8d1 a1d1 b4d6 h2h3 f6d7 d1d2 g7g5 g1f1 d7f6 a4b6 f6e4 b6a8 e4d2 e3d2 f5d3 f1e1 b7a8 b3d1 d3e4 g2g3 e4d3 d1e2 d3b1 a2a3 a8b7 d2e3 a5a4 e2f3');
INSERT INTO evaluations VALUES(9870,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(9870,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(9870,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(9870,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(9870,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(9870,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(9870,36,'cp 56','d4c6 b7c6 e1g1 d7d6 e4e5 d6e5 d1d8 e8d8 c4f7 c8f5 f7b3 a7a5 c3a4 d8c8 c1e3 f6d5 a2a3 b4d6 e3c5 c8b7 a1d1 h8d8 f2f3 d5f6 h2h3 h6h5 c5e3 e5e4 f3e4 f5e4 d1e1 e4d5 b3d5 c6d5 a4c5 d6c5 e3c5 f6d7 c5d4');
INSERT INTO evaluations VALUES(9722,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(9722,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(9722,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(9722,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(9722,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(9722,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(9722,36,'cp 72','b7c6 e1g1 d7d6 e4e5 d6e5 d1d8 e8d8 c4f7 c8f5 f7b3 a7a5 c3a4 d8c8 c1e3 f6d5 e3c5 c8b7 f2f3 b4c5 a4c5 b7b6 c5a4 b6a7 f1f2 h8e8 a1e1 e8e7 f2d2 d5f6 d2d4 e5e4 d4c4 a7b7 h2h3 a8e8 c4c5 g7g6 g2g4 f5e6 b3e6 e7e6 c5a5 f6d7 f3f4 e8f8 f4f5 g6f5 a5f5 f8f5 g4f5');
INSERT INTO evaluations VALUES(7572,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(7572,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(7572,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(7572,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(7572,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(7572,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(7572,36,'cp 54','e1g1 d7d6 e4e5 d6e5 d1d8 e8d8 c4f7 c8f5 f7b3 a7a5 c3a4 d8c8 f1d1 c8b7 c1e3 f6g4 e3c5 g4f6 f2f3 a8d8 h2h3 h8e8 d1d8 e8d8 c5b4 a5b4 a1e1 d8d2 c2c3 b4c3 a4c3 f6d7 e1e2 d2e2 c3e2 d7c5 b3c4 b7b6 e2g3 f5h7');
INSERT INTO evaluations VALUES(50352,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(50352,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(50352,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(50352,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(50352,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(50352,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(50352,36,'cp 57','e4e5 f6d5 d2d4 d7d6 c2c4 d5b6 f2f4 d6e5 f4e5 c8f5 b1c3 e7e6 g1f3 b8c6 c1e3 f8e7 f1e2 f7f6 e1g1 e8g8 e5f6 e7f6 d1d2 d8e7 e3f2 a8d8 a1d1 e6e5 d4d5 e5e4 f3d4 c6d4 f2d4 f6d4 d2d4 b6d7 d4a7 d7c5 b2b4 c5d3 a7b7 g8h8 c3b5 e7b4 b7c7 d8c8 c7g3 c8c4');
INSERT INTO evaluations VALUES(50309,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(50309,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(50309,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(50309,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(50309,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(50309,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(50309,36,'cp 57','f6d5 d2d4 d7d6 g1f3 d6e5 f3e5 c7c6 f1e2 c8f5 e1g1 b8d7 e5f3 e7e6 c2c4 d5f6 c1f4 f6h5 f4g5 h5f6 e2d3 h7h6 g5f4 f5d3 d1d3 f6h5 f4e3 f8d6 b1c3 e8g8 a1d1 h5f6 h2h3 d8c7 a2a3 a7a6 f1e1 b7b5 c4c5 d6f4 e3f4');
INSERT INTO evaluations VALUES(50409,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(50409,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(50409,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(50409,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(50409,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(50409,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(50409,36,'cp 59','d2d4 d7d6 g1f3 d6e5 f3e5 c7c6 f1e2 c8f5 e1g1 b8d7 e5f3 e7e6 c2c4 d5f6 c1f4 f6h5 f4g5 f8e7 g5e7 d8e7 f1e1 e8g8 d1b3 f8b8 b3e3 f5b1 a1b1 h5f6 h2h3 h7h6 a2a3 b7b5 f3e5 d7e5 d4e5 f6d7 c4b5 c6b5');
INSERT INTO evaluations VALUES(50405,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(50405,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(50405,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(50405,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50405,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(50405,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(50405,36,'cp 59','d7d6 c2c4 d5b6 e5d6 e7d6 g1f3 f8e7 f1e2 e8g8 e1g1 c7c5 b1c3 c8g4 c1e3 b8c6 d4d5 g4f3 e2f3 c6e5 b2b3 f7f5 f3h5 g7g6 h5e2 e7f6 d1d2 b6d7 f2f4 e5f7 a1e1 a7a6 c3a4 f8e8 e2f3 f6h4 g2g3 h4f6 d2d3');
INSERT INTO evaluations VALUES(47070,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(47070,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(47070,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(47070,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(47070,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47070,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(47070,36,'cp 64','c2c4 d5b6 e5d6 e7d6 g1f3 f8e7 f1e2 e8g8 e1g1 c8f5 b1c3 e7f6 h2h3 b8c6 c1f4 a7a5 d1d2 a5a4 b2b3 f5g6 a1e1 a4b3 a2b3 b6d7 e2d1 h7h6 c3d5 c6e7 d5f6 d7f6 f3h4 f8e8 d1f3 g6h7 d4d5 e7f5 h4f5 h7f5 e1e8 d8e8 f1e1 e8d7');
INSERT INTO evaluations VALUES(47067,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(47067,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(47067,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(47067,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(47067,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(47067,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(47067,36,'cp 57','d5b6 e5d6 e7d6 f1e2 f8e7 g1f3 e8g8 e1g1 c8e6 d1c2 e6g4 c1e3 c7c6 b1d2 g4h5 e2d3 h5g6 d3g6 h7g6 a1e1 f8e8 e1e2 b8a6 f1e1 a6c7 h2h3 b6d7 c2b3 d8c8 e3g5 e7f8 e2e8 c7e8 g5e7 f8e7 e1e7 g8f8 e7e3 e8f6 f3g5 f8g8 d2f3');
INSERT INTO evaluations VALUES(46926,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(46926,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(46926,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(46926,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(46926,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(46926,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(46926,36,'cp 63','e5d6 e7d6 g1f3 f8e7 f1e2 e8g8 e1g1 f8e8 h2h3 c8f5 c1f4 e7f6 b1c3 b8c6 b2b3 h7h6 d1d2 a7a5 a1e1 a5a4 d4d5 c6e7 c3b5 f5e4 f3d4 a4b3 a2b3 b6d7 e2f3 e4g6 b3b4 d7e5 c4c5 d6c5 f4e5 f6e5');
INSERT INTO evaluations VALUES(46915,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46915,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(46915,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(46915,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(46915,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(46915,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(46915,36,'cp 62','e7d6 f1e2 f8e7 g1f3 e8g8 e1g1 c7c6 f1e1 f8e8 c1f4 c8g4 b1c3 e7f6 d1b3 g4f3 e2f3 f6d4 f4d6 b8d7 c3e4 d7f6 e4f6 d8f6 c4c5 b6d7 b3b7 d7c5 b7c6 c5e6 a1d1 d4b2 c6a8 e8a8 f3a8 b2d4 g1h1 f6f2');
INSERT INTO evaluations VALUES(47735,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(47735,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(47735,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(47735,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(47735,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(47735,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(47735,36,'cp 66','h2h3 c8f5 f1d3 f5d3 d1d3 f8e7 g1f3 e8g8 e1g1 b8c6 c1f4 e7f6 b2b3 d6d5 c4c5 b6c8 b1c3 c8e7 a1b1 d8d7 b3b4 e7g6 f4g3 a7a6 f1e1 f8e8 e1e8 a8e8 b4b5 a6b5 c3b5 f6d8 a2a4 c6a5 d3c3 a5c4 b5a3 d7a4 a3c4 a4c4 c3c4');
INSERT INTO evaluations VALUES(57110,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(57110,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(57110,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(57110,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(57110,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(57110,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(57110,36,'cp -44','b8c6 f1c4 g7g6 h5d1 f8g7 b1c3 c6a5 d2d3 d7d6 c1e3 g8f6 a2a4 a5c4 d3c4 f6g4 e3c1 f7f5 f2f3 g4f6 d1d3 c8e6 c1g5 h7h6 g5h4 d8d7 g1e2 f5e4 c3e4 g6g5 h4f2 g5g4 e2c3 g4f3 e4f6 g7f6 d3f3 d7f7 b2b3 e8c8 e1g1 c8b8');
INSERT INTO evaluations VALUES(22696,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(22696,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(22696,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(22696,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(22696,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(22696,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(22696,36,'cp -39','f1c4 g7g6 h5f3 g8f6 g1e2 d7d6 b1c3 f8g7 d2d3 e8g8 a2a3 h7h6 f3g3 c6e7 e1g1 c7c6 f2f4 d6d5 c4b3 e5f4 g3f4 c8e6 c1e3 g6g5 f4g3 f6g4 e3d4 f7f5 e4d5 f5f4 g3f3 c6d5 a1e1 g4e5 d4e5 g7e5 g1h1 e6f7');
INSERT INTO evaluations VALUES(22689,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(22689,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(22689,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(22689,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(22689,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(22689,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22689,36,'cp -29','g7g6 h5f3 g8f6 g1e2 f8g7 b1c3 d7d6 d2d3 e8g8 a2a3 h7h6 e1g1 g8h7 h2h3 c8e6 c1d2 a7a6 a1d1 f6d7 c4d5 c6d4 e2d4 e5d4 c3e2 e6d5 e4d5 d8f6 f3g4 d7b6 c2c3 d4c3 e2c3 f8e8 d1e1 e8e1 f1e1 a8d8 g4b4 f6d4 b4d4 g7d4');
INSERT INTO evaluations VALUES(22448,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(22448,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(22448,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(22448,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(22448,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(22448,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(22448,36,'cp -20','h5f3 g8f6 g1e2 f8g7 b1c3 e8g8 d2d3 d7d6 a2a3 h7h6 h2h3 c8e6 h3h4 e6c4 d3c4 f6d7 f3d3 d7c5 d3d1 c5e6 c1e3 c6d4 d1d2 a7a6 h4h5 g6g5 e1g1 f7f5 e4f5 d4f5 e2g3 e6f4 g3f5 f8f5 a1d1 f5f7 e3f4 g5f4');
INSERT INTO evaluations VALUES(22460,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(22460,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(22460,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(22460,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(22460,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(22460,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(22460,36,'cp -47','g8f6 g1e2 f8g7 d2d3 d7d6 b1c3 e8g8 a2a3 h7h6 h2h3 c8e6 c3d5 e6d5 e4d5 e5e4 f3e3 c6e5 c4a2 e4d3 c2d3 f8e8 e1g1 d8d7 f2f4 e5c4 e3f3 c4b6 f4f5 g6g5 e2c3 b6a4 a2b3 b7b5 a1b1 a8b8 b3c2 a7a5 c1d2 a4b6 c3e4 f6e4 d3e4 b6c4');
INSERT INTO evaluations VALUES(16060,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(16060,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(16060,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(16060,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(16060,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(16060,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(16060,36,'cp -35','g1e2 f8g7 d2d3 d7d6 b1c3 e8g8 a2a3 h7h6 h2h3 c8e6 c3d5 e6d5 e4d5 e5e4 f3e3 c6e5 c4a2 e4d3 c2d3 f8e8 e1g1 d8d7 d3d4 e5c6 e3d3 c6e7 e2c3 a7a6 c1d2 h6h5 a1e1 a8b8 a2b3 b7b5 c3e4 e7f5 e4g5 g7h6 g1h1 e8e1 f1e1 g8g7 b3a2 h6g5');
INSERT INTO evaluations VALUES(16061,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(16061,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16061,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(16061,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(16061,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(16061,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(16061,36,'cp -38','f8g7 b1c3 e8g8 e1g1 d7d6 d2d3 f6h5 a2a3 c8e6 c4e6 f7e6 f3g4 d8d7 f2f3 f8f7 c3d1 a8f8 d1f2 h5f4 c1f4 e5f4 c2c3 c6e5 g4h3 c7c5 a3a4 d7e7 b2b3 e5c6 a1d1 f8e8 h3g4 c6e5 g4h3 d6d5 d3d4 c5d4 c3d4 e5c6');
INSERT INTO evaluations VALUES(9697,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(9697,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(9697,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(9697,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(9697,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(9697,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(9697,36,'cp -33','b1c3 d7d6 d2d3 e8g8 e1g1 f6h5 a2a3 c8e6 c4e6 f7e6 f3g4 d8d7 f2f3 h5f6 g4h4 a7a5 c1g5 f6h5 h4g4 a8b8 a3a4 c6b4 a1c1 f8f7 g1h1 b8f8 e2g1 d6d5 g5d2 h5f4 g1h3 f4h3 g4h3 b4c6 d2g5 b7b6 c3b5 c6b4 b5c3 c7c5');
INSERT INTO evaluations VALUES(9684,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(9684,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(9684,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(9684,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(9684,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(9684,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(9684,36,'cp -37','e8g8 d2d3 d7d6 e1g1 f6h5 a2a3 c8e6 c4e6 f7e6 f3g4 d8d7 f2f3 a7a5 c1g5 a5a4 a1e1 f8f7 g1h1 c6a5 e2g3 h5f4 g3e2 a8f8 e2g1 a5c6 g2g3 f4h5 c3e2 b7b5 g5e3 c6d4 e3d4 e5d4 g1h3 c7c5 h3g5 f7f6 e1d1 h7h6');
INSERT INTO evaluations VALUES(50337,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(50337,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(50337,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(50337,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(50337,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(50337,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(50337,36,'cp 22','d7d5 c2c4 e7e6 b1c3 f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 d8e7 f3d2 f8c8 a1c1 a7a5 f1e2 c7c6 e1g1 b7b5 a2a3 b8d7 f1e1 b5b4 a3b4 a5b4 c3b3 c6c5 h4f6 d7f6 d4c5 c8c5 d2f3 e7c7 c1d1 c7b6 f3d4 f5d7 e2f3 d7a4');
INSERT INTO evaluations VALUES(47202,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47202,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(47202,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(47202,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(47202,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(47202,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(47202,36,'cp 19','c2c4 e7e6 b1c3 f8b4 d1a4 b8c6 e2e3 e8g8 c1d2 d5c4 f1c4 b4d6 a4c2 e6e5 f3e5 c6e5 d4e5 d6e5 f2f4 e5c3 d2c3 f6g4 c3d4 c7c5 d4c5 f8e8 e1g1 g4e3 c5e3 e8e3 f1e1 e3e7 a1d1 e7e1 d1e1 c8d7 c2b3 d7c6 c4f7 g8h8 e1d1 d8f6 f7d5 f6f4 d5c6 b7c6 b3e6 a8f8 e6e7 f4f2 g1h1 f2f7');
INSERT INTO evaluations VALUES(47145,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(47145,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(47145,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47145,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(47145,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47145,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(47145,36,'cp 13','e7e6 b1c3 f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 c7c6 f3d2 f6e4 h4d8 e4c3 b2c3 f8d8 a2a4 a8c8 d2b3 c6c5 e1d2 c5c4 b3c1 d7f6 e2f3 f5e4 c1e2 b7b6 h1c1 d8e8 f3e4 f6e4 d2e1 c8b8');
INSERT INTO evaluations VALUES(57763,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(57763,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(57763,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(57763,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(57763,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(57763,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57763,36,'cp 7','e7e5 c2c4 g8f6 f1g2 f8c5 b1c3 c7c6 e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 b8c6 e1g1 e8g8 c1e3 c8g4 d1d3 g4e2 d3e2 c6d4 e2d2 d4e6 e3b6 d8b6 g2d5 a8d8 a1d1 f8e8 h2h4 h7h6 d2c2 e6d4');
INSERT INTO evaluations VALUES(57233,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(57233,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(57233,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(57233,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(57233,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(57233,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(57233,36,'cp 15','c2c4 g8f6 f1g2 f8c5 b1c3 c7c6 e2e4 d7d6 g1e2 c8g4 e1g1 b8d7 h2h3 g4e2 c3e2 d8b6 b2b3 d7f8 d2d4 c5d4 e2d4 e5d4 c1f4 e8c8 e4e5 d6e5 f4e5 b6c5 e5f6 g7f6 d1f3 f6f5 b3b4 c5c4 f3f5 c4e6 f5c5 d4d3');
INSERT INTO evaluations VALUES(57129,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(57129,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(57129,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(57129,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(57129,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(57129,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(57129,36,'cp 14','g8f6 f1g2 f8c5 d2d3 e8g8 g1f3 b8c6 e1g1 d7d5 c4d5 f6d5 a2a3 c5e7 b2b4 e7f6 e2e4 d5b6 c1b2 c8g4 d1c2 g4f3 g2f3 a7a5 b4b5 c6d4 b2d4 d8d4 b1d2 f8d8 f3e2 d4d6 d2f3 a8c8 a3a4 c7c6 b5c6 c8c6 c2b3 d8c8 f1b1 c8c7 g1g2 b6d7');
INSERT INTO evaluations VALUES(49572,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(49572,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(49572,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(49572,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(49572,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(49572,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(49572,36,'cp 12','f1g2 f8c5 d2d3 d7d6 g1f3 a7a5 e1g1 c7c6 b1c3 e8g8 d3d4 e5d4 f3d4 a5a4 e2e3 a4a3 d1c2 f8e8 h2h3 a3b2 c1b2 b8d7 f1d1 d7e5 c3b1 d8b6 b1d2 c5a3 b2a3 a8a3 d2b1 a3a6 a2a4 h7h6 b1d2 c8d7 d1b1 b6a7');
INSERT INTO evaluations VALUES(49573,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(49573,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(49573,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(49573,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(49573,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(49573,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(49573,36,'cp 14','f8c5 d2d3 e8g8 b1c3 f8e8 g1f3 d7d5 c4d5 f6d5 e1g1 d5f6 c1g5 b8d7 c3e4 c5f8 a1c1 c7c6 g2h3 h7h6 g5f6 d7f6 h3c8 a8c8 d1b3 c8b8 e4f6 d8f6 f3d2 f6e7 c1c2 b8d8 b3a4 d8a8 f1c1 a7a5 a2a3 e8d8 a4g4 d8d5 g4f3 a5a4 d2c4');
INSERT INTO evaluations VALUES(42604,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(42604,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42604,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42604,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(42604,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(42604,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(42604,36,'cp 20','d2d3 e8g8 b1c3 f8e8 g1f3 d7d5 c3d5 f6d5 c4d5 d8d5 e1g1 b8c6 a2a3 c5b6 b2b4 a7a5 f3d2 d5d8 c1b2 a5b4 a3b4 a8a1 d1a1 c8g4 f1e1 c6d4 b2d4 b6d4 a1a3 c7c6 e2e3 d4b6 d2c4 h7h5 h2h4 b6c7 c4d2 c7d6 d2e4 d6f8 e4c5 d8b6 a3c3 f8c5');
INSERT INTO evaluations VALUES(42600,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(42600,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(42600,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(42600,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(42600,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42600,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(42600,36,'cp 14','e8g8 b1c3 b8c6 a2a3 a7a5 e2e3 c5a7 d1c2 d7d6 g1f3 h7h6 b2b3 c6e7 e1g1 c7c6 b3b4 e7g6 c1d2 a7b6 f1b1 a5b4 a3b4 a8a1 b1a1 c8e6 c3e4 f6e4 d3e4 d8e7 c2b3 e7d7 d2c3 e6g4 b3b2 f8d8 h2h4 g4h3 f3h2 g6f8 g2h3');
INSERT INTO evaluations VALUES(50327,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(50327,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(50327,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(50327,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(50327,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50327,36,'cp 0','d7d5 e2e3 c7c5 g1f3 b8c6 b1d2 d8b6 d4c5 b6b2 a1b1 b2c3 f1d3 g7g6 e1g1 f8g7 f3d4 a7a6 d2b3 e7e5 d4e2 c3b4 a2a3 b4a3 b1a1 a3b2 a1b1');
INSERT INTO evaluations VALUES(47158,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47158,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47158,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47158,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47158,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47158,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47158,36,'cp 17','e2e3 e7e6 g1f3 c7c5 b1d2 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 f6d5 b1c1 c5e7 f1c4 d5f6 e1g1 b8c6 d1c2 c8d7 f1d1 a8c8 c2b1 c6b8 a2a3 f8d8 c4e2 h7h6 c1c8 d8c8 d1c1 c8c1 b1c1 d7c6');
INSERT INTO evaluations VALUES(47154,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(47154,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(47154,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(47154,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47154,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47154,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47154,36,'cp 12','e7e6 g1f3 c7c5 b1d2 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 c5d5 f1c4 d5d7 e1g1 d7e7 b1c1 f8d8 d1c2 b7b6 c4e2 c8b7 c2a4 f6d5 f1d1 b8d7 d2e4 h7h6 h2h3 d7f6 e4f6 e7f6 a4d4 f6d4 f3d4 d5f6 d4f3 d8d1 e2d1 b7f3 g2f3');
INSERT INTO evaluations VALUES(48287,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48287,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(48287,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(48287,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(48287,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(48287,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(48287,36,'cp 20','g1f3 c7c5 b1d2 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 f6d5 b1c1 c5e7 f1c4 b8c6 e1g1 f8d8 d1c2 d5b6 c4d3 h7h6 a2a3 c8d7 f1d1 a8c8 c2b1 d7e8 h2h3 a7a5 d3e2 c8c7 d2b3 d8d1 e2d1 b6d5 d1e2');
INSERT INTO evaluations VALUES(48289,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(48289,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(48289,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48289,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48289,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(48289,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(48289,36,'cp 12','c7c5 b1d2 f8d6 d4c5 d6f4 e3f4 a7a5 f3e5 d8c7 f1d3 b8d7 d1e2 d7c5 e1g1 e8g8 f1e1 c5d3 c2d3 c7c2 d2b3 c2e2 e1e2 f6d7 e2c2 d7e5 f4e5 c8d7 f2f4 f8c8 c2c8 a8c8 a1c1 c8c1 b3c1');
INSERT INTO evaluations VALUES(46207,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(46207,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46207,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46207,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46207,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(46207,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(46207,36,'cp 15','b1d2 d8b6 a1b1 f8d6 d4c5 b6c5 b2b4 c5c7 f4d6 c7d6 f1e2 e8g8 c2c4 d6e7 e1g1 b7b6 d1b3 b8d7 a2a3 c8b7 b1c1 d5c4 d2c4 b7d5 f3d4 f8c8 b3b2 a7a5 b4a5 b6a5 d4b5 d5c4 e2c4 c8c5 f1d1 a8c8 c4e2 c5c1 d1c1');
INSERT INTO evaluations VALUES(46208,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(46208,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(46208,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46208,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(46208,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(46208,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(46208,36,'cp 11','d8b6 a1b1 f8d6 d4c5 b6c5 b2b4 c5c7 f4d6 c7d6 c2c4 d5c4 f1c4 e8g8 e1g1 b7b6 c4e2 d6e7 a2a3 c8b7 b1c1 b8d7 c1c7 f8b8 d1a1 a7a5 b4b5 f6d5 c7c1 b8c8 f3d4 h7h6 f1d1 d7c5 d2c4 c8c7 e2f3 a8c8');
INSERT INTO evaluations VALUES(57012,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(57012,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(57012,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(57012,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(57012,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(57012,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(57012,36,'cp 14','g8f6 g1f3 d7d5 b1c3 f8e7 c1f4 e8g8 e2e3 b7b6 f1d3 c8a6 c4d5 f6d5 c3d5 d8d5 e1g1 d5b7 a1c1 c7c5 b2b4 a6d3 d1d3 f8c8 f3e5 b8d7 d4c5 d7e5 f4e5 b6c5 b4c5 c8c5 c1c5 e7c5 f1b1 c5b6 e5c3 a8c8');
INSERT INTO evaluations VALUES(16369,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16369,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16369,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(16369,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16369,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16369,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16369,36,'cp 18','d7d5 e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5d3 f6d5 g5f3 f8d6 e1g1 d5f4 f1e1 f4d3 c2d3 e8g8 f3e5 c6c5 b1c3 c8b7 b2b3 a5c6 e5c6 b7c6 d1h5 f7f5 c1b2 f8f6 g2g3 d6f8 h5h4 d8d3 h4c4 d3c4 b3c4 f6d6 c3d5 c6d5 c4d5 d6d5 b2c3 g8f7 a1b1 a8d8');
INSERT INTO evaluations VALUES(15082,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15082,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(15082,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15082,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(15082,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15082,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(15082,36,'cp 0','e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5d3 f6d5 g5f3 f8d6 b1c3 d5f4 e1g1 f4d3 c2d3 e8g8 f1e1 c6c5 f3e5 c8b7 b2b3 f8e8 e5c4 a5c4 e1e8 d8e8 d3c4 e8e5 g2g3 e5f5 c1b2 d6e5 d1b1 f5f3 c3d5 a8d8 b2e5 d8d5 c4d5 b7d5 g1f1 f3h1 f1e2 h1f3');
INSERT INTO evaluations VALUES(14981,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14981,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14981,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14981,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14981,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14981,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14981,36,'cp 0','c6a5 c4b5 c7c6 d5c6 b7c6 b5d3 f6d5 g5f3 f8d6 b1c3 d5f4 e1g1 f4d3 c2d3 e8g8 f1e1 c6c5 f3e5 c8b7 b2b3 f8e8 e5c4 a5c4 e1e8 d8e8 d3c4 e8e5 g2g3 e5f5 c1b2 d6e5 d1b1 f5f3 c3d5 a8d8 b2e5 d8d5 c4d5 b7d5 g1f1 f3h1 f1e2 h1f3');
INSERT INTO evaluations VALUES(15837,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15837,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15837,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15837,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15837,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15837,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15837,36,'cp 0','c4b5 c7c6 d5c6 b7c6 b5d3 f6d5 g5f3 f8d6 e1g1 d5f4 f1e1 f4d3 c2d3 e8g8 b1c3 c6c5 f3e5 c8b7 b2b3 f8e8 e5c4 a5c4 e1e8 d8e8 d3c4 e8e5 g2g3 e5f5 c1b2 d6e5 d1b1 f5f3 c3d5 a8d8 b2e5 d8d5 c4d5 b7d5 g1f1 f3h1 f1e2 h1f3');
INSERT INTO evaluations VALUES(15844,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15844,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15844,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15844,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15844,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15844,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15844,36,'cp 0','c7c6 d5c6 b7c6 b5d3 f6d5 g5f3 f8d6 e1g1 d5f4 b1c3 f4d3 c2d3 e8g8 f1e1 c6c5 f3e5 c8b7 b2b3 f8e8 e5c4 e8e1 d1e1 a5c4 d3c4 d6h2 g1h2 d8h4 h2g1 b7g2 g1g2 h4g4 g2h2 g4h4');
INSERT INTO evaluations VALUES(14004,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14004,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14004,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14004,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14004,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14004,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14004,36,'cp 0','d5c6 b7c6 b5d3 f6d5 g5f3 f8d6 e1g1 d5f4 b1c3 f4d3 c2d3 e8g8 f1e1 c6c5 f3e5 c8b7 b2b3 f8e8 e5c4 e8e1 d1e1 a5c4 d3c4 d6h2 g1h2 d8h4 h2g1 b7g2 g1g2 h4g4 g2h2 g4h4');
INSERT INTO evaluations VALUES(13449,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13449,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13449,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13449,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(13449,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13449,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13449,36,'cp 0','b7c6 b5d3 f6d5 g5f3 f8d6 e1g1 d5f4 b1c3 f4d3 c2d3 e8g8 f1e1 c6c5 f3e5 c8b7 b2b3 f8e8 e5c4 e8e1 d1e1 a5c4 d3c4 d6h2 g1h2 d8h4 h2g1 b7g2 g1g2 h4g4 g2h2 g4h4');
INSERT INTO evaluations VALUES(12075,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12075,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12075,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(12075,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12075,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12075,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(12075,36,'cp 0','b5d3 f6d5 g5f3 f8d6 e1g1 d5f4 b1c3 f4d3 c2d3 e8g8 b2b3 f8e8 f1e1 f7f6 c3e4 d6f8 c1b2 c6c5 b2a3 a5b7 b3b4 d8d3 e1e3 d3d5 b4c5 c8e6 a1c1 b7a5 e4c3 d5d8 d2d3 a5c6 c3e4 d8d5 e4d6 e8d8 h2h3 h7h6 c1c4 c6a5 c4a4 a5c6 a4c4');
INSERT INTO evaluations VALUES(12071,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12071,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12071,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(12071,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12071,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12071,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12071,36,'cp 0','f6d5 g5f3 f8d6 e1g1 d5f4 b1c3 f4d3 c2d3 e8g8 b2b3 f8e8 f1e1 f7f6 c1b2 c6c5 c3e4 d6f8 b2a3 a5b7 b3b4 d8d3 e1e3 d3d5 b4c5 c8e6 a1c1 b7a5 d2d3 a5c6 e4c3 d5d8 c3e4');
INSERT INTO evaluations VALUES(12078,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12078,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12078,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12078,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12078,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12078,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12078,36,'cp 0','g5f3 f8d6 e1g1 d5f4 b1c3 f4d3 c2d3 e8g8 b2b3 f8e8 f1e1 f7f6 c3e4 d6f8 c1b2 c6c5 b2a3 a5b7 b3b4 d8d3 e1e3 d3d5 b4c5 c8e6 e4c3 d5c6 c3e4');
INSERT INTO evaluations VALUES(21858,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(21858,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(21858,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(21858,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(21858,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(21858,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(21858,36,'cp 83','b1c3 g8f6 d2d4 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 b2b3 d8d7 c1b2 f8e8 d1d3 d7e6 a1e1 a7a5 e1e2 e7f8 h2h3 f6d7 d3f3 f7f6 b2c1 e6f7 c1e3 d7b6 e3f2 d6d5 f1d1 a5a4 f2b6 c7b6 c3a4 e8e4');
INSERT INTO evaluations VALUES(21853,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(21853,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(21853,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(21853,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(21853,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(21853,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(21853,36,'cp 91','g8f6 d2d4 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 b2b3 f8e8 d1f3 e7f8 c1d2 a7a5 a1e1 f6d7 a2a4 d8e7 g1h1 d7c5 e1e2 a8d8 h2h3 h7h6 f1e1 c5e6 d2e3 e7f6 c3d1 e6d4 e3d4 f6d4 d1f2 d4b4');
INSERT INTO evaluations VALUES(15424,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(15424,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(15424,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(15424,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(15424,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(15424,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(15424,36,'cp 82','d2d4 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 a7a5 d1f3 e8g8 c1e3 d8d7 h2h3 f8b8 a1b1 d7e6 b2b3 a5a4 f1e1 e7f8 b1d1 a4b3 a2b3 f6d7 e3f2 b8e8 g2g4 a8b8 g1g2 f7f6 h3h4 e6f7 h4h5 h7h6 d1d3 g8h8 e1d1 f7e7 f2d4 e7f7');
INSERT INTO evaluations VALUES(15418,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(15418,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(15418,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(15418,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(15418,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(15418,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(15418,36,'cp 78','e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 h2h3 f6d7 c1d2 e7f6 a1d1 f6d4 g1h2 f8e8 d2e1 d8f6 b2b3 d7c5 a2a3 h7h6 h2h1 d4c3 e1c3 f6e7 e4e5 d6e5 f4e5 e7e6 d1e1 e6d5 f3g4 c5e6 e1e2 d5b5 c3b2 b5d5');
INSERT INTO evaluations VALUES(15446,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(15446,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(15446,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(15446,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(15446,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(15446,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(15446,36,'cp 74','f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 d8d7 h2h3 f8b8 e4e5 f6d5 c3d5 c6d5 f3d5 c7c6 d5d4 b8b4 c2c4 d7e6 e5d6 e6d6 d4d6 e7d6 b2b3 a5a4 f1d1 d6f8 g1f2 g7g6 e3d4 c6c5 d4c3 b4b6 d1d3 f8g7 a1d1 g7c3 d3c3 a4b3');
INSERT INTO evaluations VALUES(15442,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(15442,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(15442,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(15442,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(15442,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(15442,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(15442,36,'cp 64','c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 c1e3 e8g8 f2f4 d8b8 a1b1 b8b4 f1e1 f6d7 e3f2 f8d8 d1d3 a7a5 a2a3 b4b7 b2b3 d7f8 d3f3 e7f6 b1d1 d8e8 h2h3 a8d8 d1d3 f6c3 d3c3 c6c5 f2h4 f7f6 c3e3 g8h8 f3g4 b7c8 g4g3 c8d7 g1h1 f8e6');
INSERT INTO evaluations VALUES(25528,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(25528,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(25528,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(25528,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(25528,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(25528,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(25528,36,'cp 75','e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 a7a5 d1f3 e8g8 c1e3 d8d7 h2h3 f8b8 b2b3 d6d5 e4e5 e7b4 c3e2 f6e4 a1d1 f7f5 g2g4 b8e8 a2a4 e8e7 e2g3 g7g6 g1g2 a8f8 c2c4 f5g4 f3g4 e4g3 g4d7 e7d7 g2g3 g8g7 g3g4');
INSERT INTO evaluations VALUES(25527,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(25527,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(25527,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(25527,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(25527,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(25527,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(25527,36,'cp 72','f8e7 d4c6 d7c6 b5c6 b7c6 b2b3 e8g8 f2f4 a7a5 d1f3 d8c8 h2h3 c8e6 c1e3 a5a4 f1e1 a4b3 c2b3 f8e8 e3f2 f6d7 a1d1 e7f8 e1e2 g7g6 f2d4 e6e7 e2e1 f8g7 d4g7 g8g7 e1e2 d7f6 f3f2 g7g8 g1h2 f6h5 e4e5 d6e5 e2e5');
INSERT INTO evaluations VALUES(24212,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(24212,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(24212,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(24212,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(24212,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(24212,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(24212,36,'cp 72','d4c6 d7c6 b5c6 b7c6 f2f4 a7a5 d1f3 e8g8 b2b3 d8c8 h2h3 c8e6 c1e3 a5a4 a1e1 a4b3 c2b3 d6d5 e4e5 f6e4 c3e4 d5e4 f3e2 f8d8 e1c1 a8a3 g1h2 h7h6 c1c4 d8a8 e2c2 a3a2 c2e4 a8a3 f4f5 e6d5 e4d5 c6d5 c4c7 a3b3');
INSERT INTO evaluations VALUES(24183,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(24183,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(24183,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(24183,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(24183,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(24183,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(24183,36,'cp 79','d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 b2b3 d8d7 h2h3 d7e6 c1e3 a5a4 a1e1 a4b3 c2b3 f6d7 e1e2 f8e8 e3f2 e7f8 g1h1 f7f6 f1d1 e6f7 e2d2 d7b6 f2b6 c7b6 c3e2 b6b5 e2g3 g7g6 f4f5 g8h8 h3h4 b5b4 h1h2');
INSERT INTO evaluations VALUES(50191,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(50191,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(50191,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(50191,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50191,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(50191,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(50191,36,'cp 47','b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 f8g7 g1f3 c7c5 c1e3 d8a5 d1d2 e8g8 a1c1 c5d4 c3d4 a5d2 f3d2 b8c6 d2b3 c8d7 d4d5 c6e5 f1e2 f8c8 e1g1 f7f5 f2f4 e5g4 e2g4 f5g4 c1c8 a8c8 b3c5 d7e8 c5b7 c8c2 f1c1 c2a2 c1c8');
INSERT INTO evaluations VALUES(50186,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(50186,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(50186,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50186,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(50186,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(50186,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(50186,36,'cp 39','d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 f8g7 g1f3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 e1e2 a8d8 d4d5 e7e6 b1b5 b7b6 h2h3 g4f3 e2f3 e6d5 b5d5 a5c6 d3b5 c6d4 e3d4 g7d4 h1d1 d4f6 a2a4 d8c8 e4e5 f6e7 f3e4 f8d8 d5d7');
INSERT INTO evaluations VALUES(46756,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(46756,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46756,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(46756,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(46756,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46756,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(46756,36,'cp 39','c4d5 f6d5 e2e4 d5c3 b2c3 f8g7 g1f3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 f1d3 c8g4 d2a5 c6a5 e1e2 e7e6 e4e5 a8d8 h1c1 a5c6 c1c4 d8d7 d3e4 f7f6 h2h3 g4f5 e4f5 g6f5 a2a4 a7a6 b1b6 f8d8 c4c5 g7f8 c5c1 f8a3 c1c2 f6e5 f3e5 c6d4 e3d4 d7d4');
INSERT INTO evaluations VALUES(46738,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(46738,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(46738,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(46738,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(46738,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(46738,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(46738,36,'cp 19','f6d5 e2e4 d5c3 b2c3 f8g7 g1f3 c7c5 c1e3 c8g4 a1c1 e8g8 f1e2 d8a5 e1g1 a5a2 c1b1 c5d4 c3d4 b8c6 d4d5 g4f3 e2f3 c6e5 b1b7 g7f6 f3e2 e5c4 e2c4 a2c4 d1b1 a7a5 f1c1 c4e2 c1e1 e2h5 e3f4 f6e5 f4g3 e5g3 h2g3 h5e5 b1a1 e5a1');
INSERT INTO evaluations VALUES(46838,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(46838,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(46838,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(46838,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(46838,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(46838,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(46838,36,'cp 20','e2e4 d5c3 b2c3 f8g7 f1c4 c7c5 g1e2 b8c6 c1e3 e8g8 e1g1 b7b6 d1d2 c8b7 a1c1 a8c8 f2f3 c5d4 c3d4 d8d6 f1d1 c6a5 c4d3 c8c1 d1c1 f8d8 h2h3 e7e6 c1d1 d8c8 e3h6 d6e7 h6g7 g8g7 g1h1 a5c4 d3c4 c8c4 d4d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(46842,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(46842,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(46842,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(46842,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46842,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(46842,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(46842,36,'cp 18','d5c3 b2c3 f8g7 c1e3 c7c5 a1c1 d8a5 d1d2 c5d4 c3d4 a5d2 e1d2 e8g8 g1f3 b8c6 d4d5 f8d8 d2e1 c6b4 e3d2 b4a6 d2g5 a6b4 f1c4 b7b5 c4b5 b4a2 c1c7 c8g4 g5e7 d8c8 c7c8 a8c8 e1d2 c8b8 b5a6 a2b4 e7d6 b4a6 d6b8 a6b8');
INSERT INTO evaluations VALUES(46856,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(46856,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(46856,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46856,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(46856,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46856,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46856,36,'cp 23','b2c3 f8g7 f1c4 c7c5 g1e2 e8g8 c1e3 b8c6 e1g1 b7b6 a1c1 c8b7 f2f3 e7e6 d1d2 a8c8 h2h4 d8e7 e3g5 e7d7 f1d1 c6a5 c4d3 c5d4 c3d4 c8c1 d1c1 g7d4 e2d4 d7d4 g5e3 d4d6 c1d1 a5c6 h4h5 f8d8 d2c1 d6e7 h5h6 f7f6 f3f4 f6f5 d3c4 f5e4');
INSERT INTO evaluations VALUES(46855,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46855,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(46855,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(46855,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(46855,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46855,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46855,36,'cp 15','f8g7 g1f3 c7c5 c1e3 c8g4 a1c1 e8g8 f1e2 d8a5 e1g1 a5a2 c1b1 c5d4 c3d4 b8c6 d4d5 g4f3 g2f3 c6d4 e3d4 g7d4 d1d4 a2e2 g1g2 e2a6 f1c1 a8c8 b1a1 a6b6 d4b6 a7b6 c1c8 f8c8 a1b1 c8c7 b1b6 g8f8 g2g3 f7f6 f3f4 f8f7 f2f3 f6f5 e4e5 f7e8');
INSERT INTO evaluations VALUES(41567,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(41567,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(41567,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(41567,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(41567,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(41567,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(41567,36,'cp 22','g1f3 c7c5 c1e3 c8g4 d1a4 b8d7 f1e2 e8g8 e1g1 d7b6 a4d1 a8c8 a1b1 g4f3 g2f3 c5d4 c3d4 b6c4 e2c4 c8c4 d1d3 c4a4 d3c2 a4a5 d4d5 d8d7 c2b3 b7b6 e3d2 a5c5 f1c1 f8c8 g1g2 c5c1 b1c1 c8c1 d2c1 g7e5 h2h4 e5d4 c1e3 d4e3 b3e3');
INSERT INTO evaluations VALUES(41566,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41566,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41566,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41566,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(41566,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(41566,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41566,36,'cp 23','c7c5 c1e3 c8g4 d1a4 b8d7 f1e2 e8g8 e1g1 d7b6 a4d1 c5d4 c3d4 a8c8 a1b1 g4f3 g2f3 b6c4 e2c4 c8c4 d1d3 c4a4 d4d5 d8d7 f1c1 d7h3 e3f4 g7h6 d3e3 h6f4 e3f4 f8c8 c1c8 h3c8 b1c1 c8h3 f4b8 g8g7 b8e5 g7h6 e5g3 h3g3 f2g3');
INSERT INTO evaluations VALUES(40435,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(40435,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40435,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(40435,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(40435,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(40435,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(40435,36,'cp 9','c1e3 c8g4 a1c1 e8g8 f1e2 d8a5 e1g1 a5a2 c1b1 c5d4 c3d4 b8c6 d4d5 c6e5 b1b7 a7a5 b7e7 a5a4 e3d4 e5f3 e2f3 g4f3 g2f3 g7d4 d1d4 a4a3 d5d6 a2b2 d4b2 a3b2 g1g2 f8b8 f1b1 g8f8 e7c7 a8a1 b1b2 b8b2 c7c8 f8g7 d6d7 a1d1 d7d8q d1d8 c8d8 b2b4 g2g3 h7h5 d8d5 b4c4 h2h3 c4a4 h3h4 a4c4 d5a5');
INSERT INTO evaluations VALUES(40432,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40432,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40432,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(40432,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40432,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(40432,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40432,36,'cp 10','c8g4 d1a4 b8d7 f1e2 e8g8 a1d1 a7a6 h2h3 g4f3 e2f3 b7b5 a4b3 a8b8 e1g1 d8a5 h3h4 h7h5 e4e5 e7e6 e3d2 a5b6 f1e1 b5b4 d4c5 d7c5 b3b4 c5d3 b4b6 b8b6 d2e3 d3e1 e3b6 e1f3 g2f3 f8c8 b6d4 g7f8 a2a4 c8c4 d1b1 f8c5 b1b8 g8g7 d4c5 c4c5');
INSERT INTO evaluations VALUES(57662,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(57662,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(57662,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(57662,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(57662,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(57662,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(57662,36,'cp 66','e2e4 d7d6 b1c3 g8f6 f2f4 f8g7 g1f3 c7c5 d4c5 d8a5 f1d3 a5c5 d1e2 e8g8 c1e3 c5a5 e1g1 c8g4 g1h1 b8c6 a1d1 e7e6 h2h3 g4f3 e2f3 a8c8 a2a4 a7a6 e3d2 f6d7 f4f5 d7e5 f3g3 e5d3 c2d3');
INSERT INTO evaluations VALUES(57220,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(57220,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(57220,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(57220,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(57220,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(57220,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(57220,36,'cp 12','d2d4 e5d4 e3d4 d7d5 g1f3 f8d6 f1d3 g8e7 e1g1 e8g8 c1g5 h7h6 g5h4 c7c6 c2c4 d5c4 d3c4 d8c7 b1c3 e7f5 h4g3 b8d7 d1d3 d7f6 f1e1 d6g3 h2g3 f5d6 c4b3 c8f5 d3d2 d6e4 d2d1 e4d6 f3e5 f8e8 a2a4 f5e6 b3e6');
INSERT INTO evaluations VALUES(57139,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(57139,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(57139,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(57139,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(57139,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(57139,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(57139,36,'cp 2','e5d4 e3d4 d7d5 g1f3 f8d6 f1d3 b8c6 e1g1 g8e7 c2c3 c8f5 d3f5 e7f5 f1e1 c6e7 d1b3 b7b6 c1g5 f7f6 g5e3 e8g8 c3c4 g8h8 b1c3 d5c4 b3c4 d8d7 e3d2 a8d8 h2h3 c7c6 g2g4 f5h6 c3e4 h6f7 a1c1 d6c7 e4g3 f7d6 c4e6 d7e6 e1e6 f8e8');
INSERT INTO evaluations VALUES(57305,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(57305,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(57305,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(57305,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(57305,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(57305,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(57305,36,'cp 10','e3d4 d7d5 g1f3 f8d6 f1d3 b8c6 e1g1 g8e7 c2c3 c8f5 d3f5 e7f5 d1b3 b7b6 f1e1 c6e7 c1g5 f7f6 g5e3 e8g8 c3c4 g8h8 b1c3 d5c4 b3c4 d8d7 a1d1 a8d8 h2h3 e7g6 e3d2 f8e8 e1e8 d8e8 c3b5 c7c6 b5c3 f5e7 d1e1 e8d8 d4d5 c6d5 c3d5');
INSERT INTO evaluations VALUES(57286,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(57286,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(57286,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57286,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(57286,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(57286,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(57286,36,'cp 9','d7d5 g1f3 f8d6 f1d3 b8c6 e1g1 g8e7 c2c3 c8f5 d3f5 e7f5 d1b3 b7b6 f1e1 c6e7 c1g5 f7f6 g5e3 e8g8 c3c4 g8h8 b1c3 d5c4 b3c4 d8d7 a1d1 a8d8 h2h3 c7c6 e3d2 d6b8 d2c1 g7g5 a2a4 h7h5 e1e6 f5g7 e6e1');
INSERT INTO evaluations VALUES(56217,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(56217,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56217,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(56217,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(56217,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(56217,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(56217,36,'cp 20','g1f3 f8d6 c2c4 c7c6 f1d3 d5c4 d3c4 g8f6 e1g1 e8g8 b1c3 c8f5 h2h3 h7h6 f1e1 b8d7 f3e5 d7b6 c4b3 b6d5 d1f3 f5e6 c1d2 d5c7 e5c4 c7d5 a1d1 d5c3 b2c3 e6d5 f3d3 b7b5 c4e5 d6e5 e1e5 d5b3 a2b3 f8e8 d1e1 e8e5 e1e5');
INSERT INTO evaluations VALUES(52380,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52380,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(52380,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(52380,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(52380,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52380,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(52380,36,'cp 23','d7d5 b1c3 c8g4 d2d4 e7e6 f1d3 g8f6 e1g1 f6e4 c3e4 d5e4 d3e4 g4h5 h2h3 b8d7 g2g4 h5g6 e4g6 h7g6 g1g2 f8d6 c2c4 e8g8 d1d3 d7f6 c1g5 b7b5 b2b3 b5c4 b3c4 c6c5 a1b1 d8c7 g5f6 g7f6 d4d5 e6d5 d3d5 a8b8 f3d2 g8g7 d2e4 f8d8 b1b8');
INSERT INTO evaluations VALUES(52810,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52810,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(52810,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(52810,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52810,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(52810,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52810,36,'cp 28','b1c3 c8g4 h2h3 g4f3 d1f3 e7e6 f1e2 g8f6 e1g1 f8c5 d2d3 b8d7 f3g3 e8g8 c1f4 g8h8 e2f3 f8e8 e4e5 f6g8 c3e2 f7f6 e5f6 d8f6 f3g4 g8h6 f4h6 f6h6 d3d4 c5b6 a2a4 d7f6 a4a5 b6d8 c2c3 f6e4 g3e3 d8c7 e3h6 g7h6 g4h5 e8f8 a5a6');
INSERT INTO evaluations VALUES(52795,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52795,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(52795,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52795,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52795,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(52795,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52795,36,'cp 40','c8g4 h2h3 g4f3 d1f3 e7e6 f1e2 g8f6 d2d3 b8d7 e1g1 f8c5 f3g3 e8g8 c1f4 g8h8 e4e5 f6g8 e2g4 g8e7 c3a4 c5b4 c2c3 b4a5 d3d4 a5c7 g3e3 b7b5 a4c5 d7c5 d4c5 h8g8 a1e1 a7a5 f4g5 h7h6 g5h4 b5b4');
INSERT INTO evaluations VALUES(31126,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(31126,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(31126,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(31126,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(31126,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31126,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(31126,36,'cp 22','h2h3 g4f3 d1f3 e7e6 f1e2 g8f6 d2d3 b8d7 e1g1 f8c5 f3f4 c5b6 f4g3 e8g8 c1f4 g8h8 a1d1 b6d4 f4c7 d8c8 c7d6 f8g8 e4d5 e6d5 e2f3 g8e8 d1e1 c8d8 d6c7 d8c8 e1e8 f6e8 c7f4 d4f6 f3g4 c8d8 c3e2 d7f8 g3f3 h8g8 f1b1 h7h6 f4g3 f8g6 c2c3');
INSERT INTO evaluations VALUES(31125,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(31125,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(31125,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(31125,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31125,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(31125,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(31125,36,'cp 25','g4f3 d1f3 g8f6 f1e2 e7e6 d2d3 b8d7 f3g3 h7h5 e1g1 h5h4 g3f4 d8b8 f4b8 a8b8 c1d2 f8d6 a2a4 a7a5 e4d5 e6d5 e2d1 b8a8 f1e1 e8d8 c3e2 d6e5 e2c1 d8c7 c2c3 e5d6 c1b3 d7e5 d1c2 f6h5 a1d1 e5g6 d3d4 h5f4');
INSERT INTO evaluations VALUES(31134,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31134,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(31134,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31134,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(31134,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31134,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31134,36,'cp 26','d1f3 e7e6 f1e2 g8f6 d2d3 b8d7 f3g3 h7h5 e1g1 h5h4 g3f4 d8b8 e4d5 b8f4 c1f4 e6d5 a1e1 e8c8 e2d1 f8c5 f4g5 c8c7 a2a3 a7a5 g5f4 c5d6 f4d2 d8a8 c3e2 d6e5 c2c3 h8e8 e2d4 e5d4 c3d4 e8e1 f1e1');
INSERT INTO evaluations VALUES(31133,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(31133,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(31133,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(31133,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(31133,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(31133,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(31133,36,'cp 34','e7e6 f1e2 g8f6 d2d3 b8d7 f3g3 h7h5 e1g1 h5h4 g3f4 d8b8 f4b8 a8b8 c1d2 f8d6 a2a4 a7a5 e4d5 e6d5 e2f3 d6b4 d2f4 b8d8 c3e2 d7f8 e2d4 f8e6 f4e5 e6d4 e5d4 b4d6 f1e1 e8d7 d4f6 g7f6 f3g4 d7c7 d3d4 d6b4 c2c3 b4d6 g1h1');
INSERT INTO evaluations VALUES(31474,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(31474,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(31474,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(31474,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(31474,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(31474,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(31474,36,'cp 22','f1e2 g8f6 d2d3 b8d7 f3g3 h7h5 e1g1 h5h4 g3f4 d8b8 f4b8 a8b8 c1d2 d5e4 c3e4 f6e4 d3e4 d7c5 e4e5 b8d8 d2e3 a7a5 f1d1 f8e7 e3d2 a5a4 e2f3 c5d7 d2a5 d8a8 a5c3 d7b6 c3d4 b6d7 b2b4 a4b3 a2b3 e8g8 f3e2 f8d8 f2f4 a8a1 d1a1');
INSERT INTO evaluations VALUES(51504,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(51504,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(51504,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(51504,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(51504,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(51504,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(51504,36,'cp 107','d2d4 e7e6 f1d3 c8b7 g1f3 d7d5 e4d5 d8d5 c2c4 d5d8 b1c3 g8f6 d1e2 f8e7 e1g1 e8g8 f1e1 b8d7 c1f4 f6h5 f4d2 h5f6 a1d1 f8e8 d2g5 c7c5 d4d5 e6d5 c4d5 h7h6 g5h4 e7d6 e2c2 e8e1 d1e1 g7g5 h4g3 d6g3');
INSERT INTO evaluations VALUES(51495,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(51495,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(51495,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(51495,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(51495,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(51495,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(51495,36,'cp 93','c8b7 f1d3 d7d6 g1f3 b8d7 e1g1 g7g6 a2a4 a7a6 f1e1 f8g7 a4a5 g8f6 a5b6 c7b6 c2c4 e7e5 d4d5 a6a5 b1c3 d7c5 d3c2 e8g8 c1g5 h7h6 g5e3 f6g4 e3d2 g4f6 b2b3 b7a6 d1c1 g8h7 h2h3 a6c8 b3b4 c5a6 b4a5 b6a5 c3b5 a6c5');
INSERT INTO evaluations VALUES(30812,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(30812,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(30812,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(30812,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(30812,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(30812,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(30812,36,'cp 109','f1d3 d7d6 g1f3 b8d7 e1g1 g8f6 f1e1 e7e5 d4d5 f8e7 c2c4 h7h6 b1c3 f6h7 c1e3 e8g8 d1e2 b7c8 d3c2 a7a5 a2a3 g8h8 g2g3 d7f6 g1g2 f6g8 f3d2 f7f5 e4f5 c8f5 c2f5 f8f5 h2h4 f5f7 a1b1 d8d7');
INSERT INTO evaluations VALUES(30809,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(30809,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(30809,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(30809,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(30809,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(30809,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(30809,36,'cp 104','d7d6 g1f3 b8d7 e1g1 e7e5 f1e1 f8e7 d4d5 g8f6 c2c4 b7c8 b1c3 a7a5 a2a3 h7h6 d3c2 e8g8 c1e3 f6h7 d1e2 g8h8 h2h3 d7f6 g1h2 f6g8 e1f1 c8d7 a1d1 a5a4 h2h1 g7g6 f3g1 e7g5 f2f4 e5f4 e3f4 g5f4 f1f4');
INSERT INTO evaluations VALUES(30536,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(30536,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(30536,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(30536,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(30536,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(30536,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(30536,36,'cp 97','g1f3 b8d7 e1g1 e7e5 f1e1 f8e7 d4d5 a7a5 a2a3 g8f6 c2c4 b7c8 b1c3 h7h6 d3c2 e8g8 c1e3 f6h7 d1e2 d7f6 c2a4 c8d7 a4b5 d7b5 c4b5 f6d7 a1c1 h7g5 f3d2 g8h8 d2c4 f7f5 e4f5 f8f5 h2h3 d7f6 e2c2');
INSERT INTO evaluations VALUES(30535,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(30535,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(30535,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(30535,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(30535,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(30535,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(30535,36,'cp 86','b8d7 a2a4 g7g6 a4a5 a7a6 e1g1 f8g7 a5b6 c7b6 c2c4 g8f6 b1c3 e7e5 d4d5 a6a5 c3a4 b7c8 a1a3 e8g8 f1e1 d7c5 a4c5 b6c5 h2h3 f6h5 c1d2 h5f4 d3c2 g7h6 g1h2 g8g7 d2c3 c8d7 c2a4 d7c8 f3g1 f4h5 g2g3 g7g8 a4c6 a8a7 d1e2 f7f5 e4f5 f8f5');
INSERT INTO evaluations VALUES(19375,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(19375,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(19375,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19375,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19375,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19375,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19375,36,'cp 0','e7e6 e1g1 d7d5 e4d5 e6d5 c4b3 c8e6 f1e1 g8f6 d2d4 h7h6 c1f4 f8d6 f4d6 d8d6 d4c5 d6c5 b1c3 e8g8 c3a4 c5d6 a4c3');
INSERT INTO evaluations VALUES(18746,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18746,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18746,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18746,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18746,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18746,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18746,36,'cp 0','e1g1 d7d5 e4d5 e6d5 c4b3 c8e6 f1e1 g8f6 d2d4 c5c4 b3a4 f8d6 c1g5 e8g8 b1d2 c6e7 c2c3 e7g6 a4c2 f8e8 c2a4');
INSERT INTO evaluations VALUES(18745,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18745,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18745,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18745,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18745,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18745,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18745,36,'cp 0','d7d5 e4d5 e6d5 c4b3 c8e6 f1e1 g8f6 d2d4 c5c4 b3a4 f8d6 c1g5 e8g8 b1d2 c6e7 c2c3 e7g6 a4c2 f8e8 c2a4');
INSERT INTO evaluations VALUES(20244,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20244,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20244,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20244,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20244,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20244,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20244,36,'cp 0','e4d5 e6d5 c4b3 c8e6 f1e1 g8f6 d2d4 c5c4 b3a4 f8d6 c1g5 e8g8 b1d2 c6e7 c2c3 e7g6 a4c2 f8e8 c2a4');
INSERT INTO evaluations VALUES(20222,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20222,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20222,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20222,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20222,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20222,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20222,36,'cp 0','e6d5 c4b3 c8e6 f1e1 g8f6 d2d4 c5c4 b3a4 f8d6 c1g5 e8g8 b1d2 c6e7 c2c3 e7g6 a4c2 f8e8 c2a4');
INSERT INTO evaluations VALUES(20504,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20504,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20504,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20504,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20504,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20504,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20504,36,'cp 0','c4b3 c8e6 f1e1 g8f6 d2d4 c5c4 b3a4 f8d6 c1g5 e8g8 b1d2 c6e7 c2c3 e7g6 a4c2 f8e8 c2a4');
INSERT INTO evaluations VALUES(20528,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20528,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20528,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20528,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20528,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20528,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20528,36,'cp 0','c8e6 f1e1 g8f6 d2d4 c5c4 b3a4 f8d6 c1g5 e8g8 b1d2 c6e7 c2c3 e7g6 a4c2 f8e8 c2a4');
INSERT INTO evaluations VALUES(57756,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(57756,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(57756,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(57756,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(57756,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(57756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57756,36,'cp 0','e7e5 c1b2 b8c6 e2e3 g8f6 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a1e1 b6c7 g2g3 g7g6 c2c4 d8f6 b2a3 c8d7 e1e6 d7e6 f1e1 a8d8 a3c5 b7b6 c5a3 c7d6 a3b2');
INSERT INTO evaluations VALUES(57192,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(57192,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(57192,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(57192,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(57192,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(57192,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57192,36,'cp 0','c1b2 b8c6 e2e3 g8f6 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a1e1 a7a5 a2a4 g7g6 b2c1 d8h4 c2c3 b6c7 g2g3 h4f6 c1a3 f6f3 d3e2 f3f6');
INSERT INTO evaluations VALUES(57193,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(57193,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57193,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57193,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57193,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(57193,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57193,36,'cp 0','b8c6 e2e3 g8f6 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a1e1 b6c7 a2a4 a7a5 b2a3 d8f6 g2g3 g7g6 c2c3 f6f3 d3e2 f3f6');
INSERT INTO evaluations VALUES(22850,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22850,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22850,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22850,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22850,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22850,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22850,36,'cp 0','e2e3 d7d5 f1b5 f8d6 f2f4 f7f6 d1h5 g7g6 h5h4 e5f4 b1c3 g8e7 e1c1 e8g8 e3f4 c6d4 b5d3 d4e6 g1h3 c7c6 c3e2 e6c5 h3f2 c5d3 f2d3 e7f5 h4f2 d8b6 f2f3 f5h4 f3g3 h4f5');
INSERT INTO evaluations VALUES(22843,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22843,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(22843,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(22843,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22843,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22843,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(22843,36,'cp 0','d7d5 f1b5 f8d6 f2f4 f7f6 d1h5 g7g6 h5h4 g8e7 h4f6 h8f8 f6h4 e5f4 g1f3 f4e3 d2e3 a7a6 b5c6 e7c6 h4h7 c8f5 b1c3 d5d4 e3d4 f8f7 h7h8 f7f8 h8g7 f8f7 g7h8');
INSERT INTO evaluations VALUES(21731,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21731,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21731,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21731,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21731,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21731,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21731,36,'cp 0','f1b5 f8d6 f2f4 f7f6 d1h5 g7g6 h5h4 g8e7 h4f6 h8f8 f6h4 e5f4 g1f3 f4e3 d2e3 a7a6 b5c6 e7c6 h4h7 c8f5 b1c3 d5d4 e3d4 f8f7 h7h8 f7f8');
INSERT INTO evaluations VALUES(21537,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21537,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21537,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21537,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21537,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21537,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21537,36,'cp 0','f8d6 f2f4 f7f6 d1h5 g7g6 h5h4 g8e7 h4f6 h8f8 f6h4 e5f4 g1f3 f4e3 d2e3 a7a6 b5c6 e7c6 h4h7 c8f5 b1c3 d5d4 e3d4 f8f7 h7g8 f7f8 g8g7 f8f7 g7g8');
INSERT INTO evaluations VALUES(5898,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5898,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5898,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5898,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5898,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5898,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5898,36,'cp 0','f2f4 f7f6 d1h5 g7g6 h5h4 g8e7 h4f6 h8f8 f6h4 e5f4 g1f3 f4e3 d2e3 a7a6 b5c6 e7c6 h4h7 c8f5 b1c3 d5d4 e3d4 f8f7 h7h8 f7f8 h8g7 f8f7');
INSERT INTO evaluations VALUES(5897,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(5897,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5897,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(5897,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5897,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5897,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(5897,36,'cp -11','f7f6 d1h5 g7g6 h5h4 e5f4 b1c3 g8e7 h4f6 h8f8 f6h4 f4e3 d2e3 e7f5 h4d8 e8d8 c3d5 c6b4 d5b4 f5e3 e1d2 e3g2 b4d3 c7c6 b2g7 c6b5 g7f8 d6f8 g1f3 c8h3 d2e2 a8c8 h1d1 c8c2 d1d2 c2d2 f3d2 h3f5 d2f3 f8h6 a1g1 g2f4 d3f4 h6f4 h2h4 d8e7 h4h5 e7d6 h5g6 f5g6 f3h4 g6f7 g1g7');
INSERT INTO evaluations VALUES(6107,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(6107,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(6107,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6107,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6107,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6107,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6107,36,'cp 0','d1h5 g7g6 h5h4 g8e7 h4f6 h8f8 f6h4 e5f4 g1f3 f4e3 d2e3 a7a6 b5c6 e7c6 h4h7 c8f5 b1c3 d5d4 e3d4 f8f7 h7h8 f7f8');
INSERT INTO evaluations VALUES(21025,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(21025,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(21025,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(21025,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(21025,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(21025,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21025,36,'cp 62','c2c4 e7e6 b1c3 g8f6 e2e3 f8d6 c4c5 d6e7 a2a3 f6e4 d1c2 f7f5 b2b4 a7a6 c3e2 c8d7 h2h4 b7b6 c1b2 e8g8 e2f4 e7f6 f1e2 d8e7 a1c1 d7e8 f4d3 b6b5 a3a4 e8h5 a4b5 a6b5');
INSERT INTO evaluations VALUES(20976,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(20976,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(20976,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(20976,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(20976,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(20976,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(20976,36,'cp 59','e7e6 e2e3 g8f6 b1c3 c8d7 b2b3 f8d6 c4c5 d6e7 a2a3 a7a5 a1b1 f6e4 c3e4 d5e4 f3d2 f7f5 b3b4 a5b4 a3b4 e7f6 b4b5 c6e7 d1b3 e7d5 d2c4 e8g8 f1e2 g8h8 e1g1 d8e8 g1h1 c7c6 b5c6 b7c6 c4d6 e8e7 h1g1 d7e8 c1a3');
INSERT INTO evaluations VALUES(21383,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(21383,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(21383,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(21383,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(21383,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(21383,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(21383,36,'cp 62','e2e3 g8f6 a2a3 d5c4 b1d2 e6e5 d4e5 f6g4 d1a4 c8d7 e5e6 d7e6 f1c4 e6d7 a4b3 g4e5 c4e2 f8d6 e1g1 d8e7 b3b7 a8b8 b7a6 e8g8 a6a4 e5f3 d2f3 e7f6 a4c2 f8e8 e2d3 h7h5 h2h3 c6e5 f3e5 f6e5 f2f4 e5f6 g1h1 d7b5 d3b5 b8b5');
INSERT INTO evaluations VALUES(21382,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(21382,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(21382,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(21382,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(21382,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(21382,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(21382,36,'cp 55','g8f6 a2a3 d5c4 b1d2 b7b5 a3a4 b5a4 f1c4 c6b4 e1g1 c8d7 f3e5 f8e7 e5d7 f6d7 d1a4 e8g8 c4e2 a7a5 e2f3 a8a7 d2b3 c7c6 f3c6 b4c6 a4c6 d8a8 c6a8 f8a8 c1d2 a5a4 b3c1 a4a3 b2a3 a7a3 a1a3 a8a3 f1d1 h7h6 g2g3 e6e5 h2h3 e5d4 e3d4 a3a1 g1g2 d7b6');
INSERT INTO evaluations VALUES(14783,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14783,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(14783,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(14783,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(14783,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(14783,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(14783,36,'cp 62','b1d2 f8d6 f1d3 e8g8 e1g1 f8e8 h2h3 e6e5 c4d5 f6d5 d3e4 d5f6 e4c6 b7c6 d4e5 d6e5 f3e5 e8e5 f1e1 c8a6 b2b3 d8d6 c1b2 e5d5 b2c3 f6e4 d2e4 d5d1 e4d6 d1e1 c3e1 c7d6 a1d1 d6d5 d1c1 a8c8 e1b4 f7f6 f2f3 a6b5 a2a4 b5d3 c1c5 c8c7 b4d2 g8f7');
INSERT INTO evaluations VALUES(14781,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14781,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(14781,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(14781,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(14781,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(14781,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(14781,36,'cp 60','a7a5 a2a3 f6e4 c4d5 e6d5 d2e4 d5e4 f3d2 f7f5 f1c4 c8d7 d2b1 d8g5 g2g3 c6d8 b1c3 d8f7 h2h4 g5d8 d1b3 f7d6 c4e2 f8e7 c3a4 d7a4 b3a4 d8d7 a4b3 a5a4 b3c2 e8g8 c1d2 g8h8 e1c1 d6c8 d2c3 b7b5 d4d5 e7d6 c1b1');
INSERT INTO evaluations VALUES(20983,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(20983,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(20983,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(20983,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(20983,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(20983,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(20983,36,'cp 43','e2e3 g8f6 g1f3 c8g4 c2c4 e7e6 b1c3 f8b4 a1c1 e8g8 f1e2 g4f3 g2f3 b4c3 c1c3 c6e7 h2h4 c7c6 c4c5 g7g6 e2d3 e7f5 f4g5 g8g7 h4h5 h7h6 g5f6 d8f6 h5g6 f7g6 d3f5 f6f5 c3b3 a8b8 f3f4 f5e4');
INSERT INTO evaluations VALUES(20980,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(20980,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(20980,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(20980,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(20980,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(20980,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(20980,36,'cp 26','g8f6 g1f3 c8g4 c2c4 e7e6 b1c3 f8b4 a1c1 e8g8 f1e2 g4f3 g2f3 d5c4 e2c4 d8e7 d1c2 e6e5 d4e5 c6e5 c4e2 c7c6 f4g3 e5g6 c1d1 b4c3 b2c3 a8d8 e1g1 d8d1 f1d1 f8d8 d1d4 d8d7 g1g2 f6d5 c3c4 d5f6 d4d7 f6d7');
INSERT INTO evaluations VALUES(14227,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(14227,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(14227,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14227,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(14227,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14227,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14227,36,'cp 45','g1f3 c8g4 c2c4 e7e6 b1c3 f8b4 a1c1 e8g8 f1e2 d5c4 e2c4 c6a5 c4d3 g4f3 g2f3 f6d5 f4g3 c7c5 d4c5 d5c3 b2c3 b4c5 e1e2 b7b6 d1a4 d8e7 h2h4 c5d6 f3f4 a8c8 h4h5 h7h6 g3h4 e7b7 a4e4 b7e4 d3e4 a5c4 e4d3 c4b2 d3b1 b2c4');
INSERT INTO evaluations VALUES(14229,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(14229,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(14229,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(14229,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(14229,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(14229,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(14229,36,'cp 44','c8g4 c2c4 e7e6 b1c3 f8b4 a1c1 e8g8 f1e2 d5c4 e2c4 c6a5 c4d3 c7c5 d4c5 b4c5 e1g1 a5c6 d3b5 a8c8 h2h3 d8d1 f1d1 g4f3 g2f3 c5e7 c3e4 f6e4 f3e4 g7g5 f4g3 f8d8 d1d8 e7d8 g3e5 f7f6 e5c3 c6e5 c1d1 d8b6 g1g2 a7a6 b5e2 g8g7 f2f4');
INSERT INTO evaluations VALUES(24925,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(24925,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(24925,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(24925,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(24925,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24925,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(24925,36,'cp 33','c2c4 e7e6 a2a3 f8d6 f4g5 h7h6 g5h4 d5c4 f1c4 e6e5 d4d5 c6e7 h4f6 g7f6 h2h3 g4d7 b1c3 f6f5 d1c2 a8b8 c4b5 e8f8 b5d7 d8d7 e3e4 h8g8 g2g3 b7b5 f3h4 f5f4 g3f4 e5f4 h4f3 b5b4 a3b4 b8b4');
INSERT INTO evaluations VALUES(24888,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(24888,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(24888,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(24888,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(24888,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(24888,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(24888,36,'cp 41','e7e6 b1c3 f8b4 a1c1 e8g8 f1e2 d5c4 e2c4 c6a5 c4d3 c7c5 d4c5 b4c5 e1g1 a8c8 d3e2 c5b4 d1a4 b4c3 c1c3 c8c3 b2c3 a5c6 h2h3 g4f5 f1d1 d8b6 a4a3 h7h6 c3c4 f8d8 d1d8 b6d8 a3b2 b7b6 f3d4 c6d4 e3d4 f5e4');
INSERT INTO evaluations VALUES(25192,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(25192,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(25192,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25192,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(25192,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25192,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(25192,36,'cp 45','a2a3 f8d6 f4g5 h7h6 g5h4 d6e7 h2h3 g4f5 b1c3 e8g8 a1c1 f6e4 h4e7 c6e7 f1d3 c7c6 e1g1 a7a5 f3e5 e4c3 b2c3 f5d3 d1d3 d5c4 e5c4 a8c8 d3b1 d8c7 c4b6 c8d8 f1d1 e7c8 b6c8 d8c8 b1b3 c7e7 c3c4');
INSERT INTO evaluations VALUES(52849,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(52849,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(52849,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52849,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52849,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(52849,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52849,36,'cp 28','c6c5 g1f3 c5d4 d1d4 b8c6 d4f4 e7e6 f1d3 c6b4 b1c3 b4d3 c2d3 g8e7 h2h4 h7h5 e1g1 e7f5 c3e2 f8e7 e2d4 f5h4 f3h4 e7h4 d4f3 h4e7 f4g3 e8f8 c1g5 c8d7 g5e7 d8e7 a1c1 f8g8 c1c7 e7d8 c7b7 d7c6 b7b3 d5d4 g3f4 c6f3 f4f3');
INSERT INTO evaluations VALUES(53332,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(53332,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(53332,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(53332,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(53332,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(53332,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(53332,36,'cp 49','g1f3 c5d4 f3d4 b8c6 d4c6 b7c6 f1d3 e7e6 e1g1 g8e7 b1d2 a7a5 d1e2 e7g6 d2f3 f8c5 c1e3 c5e7 a1d1 d8c7 e3c1 c6c5 c2c4 d5d4 h2h4 c8b7 h4h5 g6f8 f3d2 f7f5 e5f6 g7f6 d3e4 h8g8 e4b7 c7b7 d2e4 e8c8 c1f4');
INSERT INTO evaluations VALUES(53330,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(53330,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53330,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53330,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(53330,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(53330,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(53330,36,'cp 27','c5d4 d1d4 b8c6 d4f4 e7e6 b1c3 c6b4 f1d3 b4d3 c2d3 c8d7 e1g1 g8e7 h2h4 e7f5 h4h5 a8c8 c1d2 h7h6 c3e2 f8e7 a1c1 e8g8 c1c8 d8c8 f1c1 c8b8 d2b4 d7b5 b4e7 f5e7 f4d2 f8c8 c1c8 b8c8 e2d4 c8d7 g2g4 a7a6 g1g2 e7c6 d4c6 d7c6 g2g3');
INSERT INTO evaluations VALUES(53447,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(53447,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(53447,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(53447,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(53447,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(53447,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(53447,36,'cp 40','d1d4 b8c6 d4f4 e7e6 f1d3 c6b4 b1c3 b4d3 c2d3 g8e7 h2h4 h7h5 e1g1 e7f5 c3e2 f8e7 e2d4 f5h4 f3h4 e7h4 d4f3 h4e7 f4g3 e8f8 c1g5 e7g5 f3g5 h8h6 a1c1 f8g8 d3d4 h6g6 f2f4 c8d7 g3h3 g6h6 h3e3 a8c8 c1c8 d8c8 f1c1 c8b8 c1c3 f7f6 e5f6');
INSERT INTO evaluations VALUES(53442,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(53442,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(53442,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(53442,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(53442,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(53442,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(53442,36,'cp 32','b8c6 d4f4 e7e6 f1d3 c6b4 e1g1 b4d3 c2d3 g8e7 h2h4 e7f5 h4h5 c8d7 b1c3 f8e7 c3e2 h7h6 e2g3 f5h4 f3h4 e7h4 f4g4 h4g5 g3e2 e8g8 c1g5 d8g5 g4g5 h6g5 e2d4 a7a5 f2f3 f7f6 f1e1 g8f7 g1f2 a8a6 e5f6 g7f6 a1c1 a6b6 c1c7 b6b2 f2e3');
INSERT INTO evaluations VALUES(57154,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(57154,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(57154,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(57154,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(57154,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(57154,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(57154,36,'cp -17','b8c6 f1e2 g8f6 c2c4 f8c5 g1f3 a7a5 h2h3 d7d6 b1c3 c8d7 e1g1 e8g8 a2a3 f8e8 d1c2 h7h6 c1e3 c6d4 e3d4 c5d4 f3d4 e5d4 c3d5 f6d5 c4d5 c7c6 d5c6 b7c6 f2f4 c6c5 a1b1 a5a4 e2g4 d7b5 f1f2 d8b6');
INSERT INTO evaluations VALUES(22802,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(22802,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(22802,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(22802,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(22802,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(22802,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(22802,36,'cp -15','a2a3 d7d5 e4d5 d8d5 b1c3 d5d6 f1e2 b7b6 f2f4 c8b7 e2f3 g8e7 f4e5 d6e5 d1e2 e8c8 e2e5 c6e5 f3b7 c8b7 g1e2 e7c6 c1f4 f7f5 e1c1 e5f7 h1f1 g7g6 f4d2 f8d6 g2g3 h8e8 c1b1');
INSERT INTO evaluations VALUES(22827,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(22827,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(22827,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(22827,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(22827,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(22827,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(22827,36,'cp -1','f8c5 g1f3 g8f6 b2b4 c5b6 c2c3 d7d6 f1e2 c6e7 b1d2 e7g6 e1g1 e8g8 a3a4 c7c6 a4a5 b6c7 d2b3 h7h6 d1c2 c8e6 c3c4 d6d5 b3c5 e6c8 a5a6 b7b6 c5b7 d8e7 c1d2 c8e6 b4b5 d5e4 d3e4 c6c5 f1e1 g6f4');
INSERT INTO evaluations VALUES(6514,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6514,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(6514,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6514,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6514,36,'cp -12','g1f3 d7d6 f1e2 a7a5 c2c3 g8f6 d3d4 c5a7 c1e3 e8g8 b1d2 f8e8 d4e5 c6e5 e3a7 e5f3 e2f3 a8a7 d1a4 a7a8 e1g1 f6d7 a4d4 b7b6 f3d1 c8a6 f1e1 d7e5 d1a4 e8e6 a4c2 d8g5 d4e3 g5e3 e1e3 e6f6 f2f3 a8d8 a1e1');
INSERT INTO evaluations VALUES(6513,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6513,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6513,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(6513,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6513,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(6513,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(6513,36,'cp -11','d7d6 c2c3 f7f5 e4f5 g8f6 d3d4 c5b6 d4e5 c6e5 f3e5 d8e7 b1d2 e7e5 d1e2 f6g4 e2e5 g4e5 f5f6 g7f6 d2c4 c8e6 c4b6 a7b6 c1f4 e8f7 e1c1 h7h5 h2h4 a8a4 g2g3 e6d7 h1g1 d7c6 f4e5 f6e5 f1e2 a4a8 d1d2 f7e6 f2f4');
INSERT INTO evaluations VALUES(51926,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(51926,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(51926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51926,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51926,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51926,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51926,36,'cp 0','g1f3 g8f6 b1c3 b8c6 d2d4 f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 d2b3 e6e5 c1e3 f8e7 e1c1 d8c6 h2h4 h7h6 e3d2 c8d7 c4d5 e8c8 d2c3 h8e8 a2a3 a7a6 d5f7 e8f8 f7d5');
INSERT INTO evaluations VALUES(51922,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51922,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51922,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51922,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51922,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51922,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51922,36,'cp 0','g8f6 b1c3 b8c6 d2d4 f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 d2b3 e6e5 c1e3 f8e7 e1c1 d8c6 h2h4 h7h6 e3d2 c8d7 c4d5 e8c8 d2c3 h8e8 d5f7 e8f8 f7d5');
INSERT INTO evaluations VALUES(44185,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44185,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44185,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44185,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44185,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44185,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44185,36,'cp 7','b1c3 d7d6 d2d4 c5d4 f3d4 a7a6 a2a4 b8c6 c1e3 f8e7 e1g1 e8g8 d1e2 c8d7 d4c6 d7c6 a4a5 f6d7 f2f3 e7f6 e2d2 f6e5 c4e2 a8c8 f1d1 d8c7 f3f4 e5c3 b2c3 c6e4 c3c4 d7f6 e3b6 c7c6 d2d6 e4c2 d6c6 c8c6');
INSERT INTO evaluations VALUES(44177,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44177,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44177,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44177,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44177,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44177,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44177,36,'cp 0','d7d6 d2d4 c5d4 f3d4 a7a6 a2a4 b8c6 c1e3 f8e7 e1g1 e8g8 h2h3 c8d7 d1e2 a8c8 c4d3 c6b4 a4a5 d8c7 d4f3 d7c6 f3d4');
INSERT INTO evaluations VALUES(45985,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45985,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45985,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45985,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(45985,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(45985,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(45985,36,'cp 17','d2d4 c5d4 f3d4 a7a6 a2a4 f8e7 c1e3 e8g8 e1g1 b8c6 d4c6 b7c6 a4a5 c8b7 f1e1 d6d5 c4f1 c6c5 e4d5 e6d5 c3a4 d8c7 a4b6 a8e8 e3g5 c7c6 c2c3 c5c4 d1a4 c6a4 a1a4 b7c6 a4a1 h7h6 g5h4 e7d8 e1e8');
INSERT INTO evaluations VALUES(46011,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46011,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(46011,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(46011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46011,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46011,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46011,36,'cp 1','c5d4 f3d4 a7a6 a2a4 f8e7 e1g1 b8c6 c1e3 e8g8 d4c6 b7c6 a4a5 d6d5 e3b6 d8d6 c4d3 a8b8 d1e2 f6d7 c3a4 e7d8 b6d8 f8d8 f1d1 h7h6 e2e1 c6c5 e1e3 c8b7 e4d5 e6d5 d3f5 b7c6 f5d7 c6d7 e3c5 d7a4 c5d6 d8d6 a1a4 b8b2');
INSERT INTO evaluations VALUES(46043,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46043,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46043,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46043,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46043,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(46043,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46043,36,'cp 0','f3d4 a7a6 c4b3 b7b5 e1g1 f8e7 d1f3 d8b6 c1e3 b6b7 f3g3 b5b4 c3a4 b8d7 c2c3 e8g8 e3h6 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(46040,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46040,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46040,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46040,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46040,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46040,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46040,36,'cp 0','a7a6 c4b3 b7b5 c1g5 f8e7 d1f3 d8b6 e1c1 e8g8 g5e3 b6b7 g2g4 b5b4 g4g5 b4c3 g5f6 e7f6 f3f6 g7f6 h1g1 g8h8 e3h6 c3b2 c1b2 f8e8 d4f5 e6f5 h6g7 h8g8 g7h6');
INSERT INTO evaluations VALUES(43522,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43522,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(43522,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43522,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43522,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43522,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43522,36,'cp 0','c4b3 b7b5 a2a3 f8e7 c1e3 e8g8 d1e2 c8b7 f2f3 b8c6 e1c1 d8c7 c1b1 b5b4 a3b4 c6b4 h2h4 d6d5 e3g5 d5e4 f3e4 c7e5 d4f3 e5c7');
INSERT INTO evaluations VALUES(43532,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43532,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43532,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43532,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43532,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43532,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43532,36,'cp 0','b7b5 a2a3 f8e7 c1e3 e8g8 d1e2 c8b7 f2f3 b8c6 e1c1 d8c7 c1b1 b5b4 a3b4 c6b4 h2h4 d6d5 e3g5 d5e4 f3e4 c7e5 d4f3 e5c7');
INSERT INTO evaluations VALUES(43752,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43752,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43752,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43752,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43752,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43752,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43752,36,'cp 0','c1e3 f8e7 d1e2 e8g8 a2a3 c8b7 f2f3 b8c6 e1c1 d8c7 c1b1 b5b4 a3b4 c6b4 h2h4 d6d5 e3g5 d5e4 f3e4 c7e5 d4f3 e5c7');
INSERT INTO evaluations VALUES(43750,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(43750,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(43750,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(43750,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(43750,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43750,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(43750,36,'cp 0','f8e7 a2a3 e8g8 f2f3 c8d7 g2g4 b8c6 h2h4 c6a5 d1d2 a5b3 c2b3 a8c8 b3b4 e6e5 d4f5 d7f5 g4f5 a6a5 b4a5 d8a5 e1g1 b5b4 a3b4 a5b4 a1a4 b4b3 a4a3 b3b4');
INSERT INTO evaluations VALUES(55769,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(55769,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(55769,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55769,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(55769,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(55769,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(55769,36,'cp 0','e6d5 g1f3 f8d6 f1d3 g8e7 e1g1 e8g8 c2c4 d5c4 d3c4 b8c6 h2h3 e7f5 c1e3 d8f6 b1c3 f8e8 d1d2 c8d7 a1e1 h7h6 a2a3 a8d8 c4d3 f5h4 f3h4 f6h4 c3b5 d7h3 b5d6 h3g2 g1g2 h4g4 g2h1 g4h4 h1g1 h4g4 g1h1');
INSERT INTO evaluations VALUES(52434,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(52434,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(52434,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(52434,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(52434,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(52434,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(52434,36,'cp 84','d4d5 e7e6 b1c3 d7d6 e2e4 g8f6 f1c4 f8e7 g1e2 e6d5 c3d5 e8g8 e1g1 f6d5 c4d5 b8a6 e2c3 a6b4 c1f4 c8e6 a2a3 b4d5 c3d5 e6d5 d1d5 f8e8 a1d1 d8b6 b2b3 b6a6 a3a4 c5c4 h2h3 h7h6 f1e1 a8d8 f4g3 c4b3 c2b3 a6b6 d1d2 d8c8');
INSERT INTO evaluations VALUES(52494,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(52494,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(52494,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(52494,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(52494,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(52494,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(52494,36,'cp 77','g8f6 b1c3 e7e6 e2e4 d7d6 f1c4 f8e7 g1e2 e8g8 e1g1 e6d5 c3d5 f6d5 c4d5 b8a6 e2f4 a6c7 a2a4 c7d5 f4d5 f8e8 f1e1 c8e6 c1f4 e7f6 c2c3 f6e5 f4e5 d6e5 c3c4 e6d5 c4d5 d8d6 g2g3 b7b6 d1d3 a8d8 a1d1 e8f8 b2b3 h7h6 d3c3 f8e8');
INSERT INTO evaluations VALUES(44594,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(44594,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(44594,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(44594,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(44594,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(44594,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(44594,36,'cp 85','c2c4 e7e6 b1c3 e6d5 c4d5 d7d6 e2e4 g7g6 h2h3 f8g7 g1f3 e8g8 c1e3 f8e8 f3d2 a7a6 a2a4 f6e4 c3e4 f7f5 f1e2 f5e4 d2c4 g7f8 d1b3 b8d7 e3f4 d8f6 b3g3 d7e5 c4e5 d6e5 f4g5 f6f7 a1d1 c8f5 e1g1 h7h6 g5d2 f7d5 f1e1');
INSERT INTO evaluations VALUES(44590,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(44590,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(44590,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(44590,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(44590,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(44590,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(44590,36,'cp 63','g7g6 b1c3 f8g7 e2e4 e8g8 g1f3 e7e6 c1e3 e6d5 c4d5 d7d6 h2h3 f8e8 f3d2 f6e4 d2e4 f7f5 f1b5 e8e4 c3e4 f5e4 e1g1 g7b2 a1b1 b2g7 d1d2 a7a6 e3g5 d8f8 b5e2 b7b5 a2a4 b5b4 f2f3 g7c3 f3e4 c3d2 f1f8 g8f8 g5d2');
INSERT INTO evaluations VALUES(44430,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(44430,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(44430,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(44430,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(44430,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(44430,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(44430,36,'cp 71','b1c3 d7d6 e2e4 f8g7 f1d3 e8g8 h2h3 e7e6 g1f3 e6d5 e4d5 f8e8 c1e3 a7a6 e1g1 b8d7 d1d2 b7b5 c4b5 d7b6 b5a6 c8a6 d3a6 a8a6 b2b3 d8a8 a1d1 f6h5 e3h6 g7c3 d2c3 b6d5 c3c1 a6a2 d1e1 a2a7 e1e8 a8e8 c1d2 e8a8');
INSERT INTO evaluations VALUES(44428,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(44428,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(44428,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(44428,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(44428,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(44428,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(44428,36,'cp 77','d7d6 h2h3 f8g7 e2e4 e7e6 g1f3 e6d5 e4d5 e8g8 f1d3 f8e8 c1e3 b8d7 e1g1 a7a6 d1d2 b7b5 c4b5 d7b6 b5a6 c8a6 d3a6 a8a6 b2b3 f6e4 c3e4 g7a1 f1a1 e8e4 d2d3 e4e3 d3a6 e3f3 g2f3 b6d5 g1h2 d8f6 a1e1 g8g7 a6f1 f6f4 h2h1 f4f3 f1g2 f3f5 a2a4 d5b4 e1e3 b4d5');
INSERT INTO evaluations VALUES(44779,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(44779,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(44779,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(44779,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(44779,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(44779,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(44779,36,'cp 90','e2e4 f8g7 g1f3 e8g8 h2h3 e7e6 f1d3 e6d5 e4d5 f8e8 c1e3 e8e3 f2e3 b8d7 d1d2 d8e7 e1g1 f6h5 g2g4 h5g3 f1f2 h7h5 g1h2 h5h4 c3e2 b7b5 c4b5 c5c4 d3c2 g7b2 a1e1 g3e2 d2e2 b2e5 h2g2 c8b7 e2c4 e5g3 c4e4 e7e4 c2e4 g3f2');
INSERT INTO evaluations VALUES(44777,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(44777,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(44777,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(44777,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(44777,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(44777,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(44777,36,'cp 78','f8g7 g1f3 e8g8 h2h3 e7e6 f1d3 e6d5 e4d5 f8e8 c1e3 b8d7 e1g1 a7a6 a2a4 d7e5 f3e5 e8e5 e3f4 e5e8 d1d2 f6h5 f4h2 f7f5 f1e1 e8e1 a1e1 d8f8 a4a5 c8d7 b2b4 c5b4 c3e2 f5f4 d2b4 a8b8 b4b6 b8e8 b6b7 f8e7 c4c5');
INSERT INTO evaluations VALUES(40365,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(40365,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(40365,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(40365,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(40365,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(40365,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(40365,36,'cp 86','g1f3 e7e6 h2h3 e8g8 f1d3 e6d5 e4d5 f8e8 c1e3 g7h6 e1g1 h6e3 f2e3 d8e7 e3e4 b8d7 d1d2 d7e5 f3e5 e7e5 f1f3 e8f8 d2f4 e5f4 f3f4 f6e8 g2g4 g8g7 a1f1 c8d7 g1f2 f7f6 f2e3 e8c7 e3d2 a8e8 f1b1 a7a6 a2a4 e8b8 f4f2 b7b5 b2b3 b5c4 d3c4');
INSERT INTO evaluations VALUES(40362,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(40362,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(40362,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(40362,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(40362,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(40362,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(40362,36,'cp 80','e8g8 h2h3 e7e6 f1d3 e6d5 e4d5 f8e8 c1e3 g7h6 e1g1 h6e3 f2e3 d8e7 e3e4 b8d7 d1d2 d7e5 f3e5 e7e5 f1f3 e8f8 d2f4 e5f4 f3f4 g8g7 a1f1 f6e8 g2g4 c8d7 h3h4 h7h6 g4g5 f7f6 f1f2 a8b8 g1g2 h6g5 h4g5 a7a6 a2a4 f8h8 g5f6 g7f7 f4f3 h8h5 e4e5 h5e5 d3g6 f7g6 f6f7 e5g5 g2f1');
INSERT INTO evaluations VALUES(35972,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(35972,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(35972,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(35972,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(35972,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(35972,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(35972,36,'cp 69','h2h3 e7e6 f1d3 e6d5 e4d5 f8e8 c1e3 g7h6 e1g1 h6e3 f2e3 d8e7 e3e4 b8d7 d1d2 d7e5 f3e5 e7e5 f1f3 e8f8 d2f4 e5f4 f3f4 g8g7 a1f1 f6e8 g2g4 h7h6 g1f2 f7f6 f2g3 e8c7 f4f3 b7b6 h3h4 c8d7 f1f2 a8e8 c3d1 e8e5 d1e3 c7e8 b2b3 e8c7 h4h5 g6g5 f2f1 a7a6');
INSERT INTO evaluations VALUES(35970,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(35970,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(35970,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(35970,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(35970,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(35970,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(35970,36,'cp 68','e7e6 f1d3 e6d5 e4d5 f8e8 c1e3 g7h6 e1g1 h6e3 f2e3 d8e7 e3e4 b8d7 d1d2 d7e5 f3e5 e7e5 f1f3 e8f8 d2f4 f6e8 a1f1 f7f6 f4d2 c8d7 b2b3 e8c7 c3e2 b7b6 g1h1 a8e8 e2f4 e8e7 d3c2 c7e8 a2a4 a7a6 c2b1 d7c8 d2f2 e5d4 g2g4 d4f2 f3f2');
INSERT INTO evaluations VALUES(36086,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(36086,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(36086,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(36086,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(36086,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(36086,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(36086,36,'cp 70','f1d3 e6d5 e4d5 f8e8 c1e3 g7h6 d1d2 h6e3 f2e3 d8e7 e3e4 b8d7 e1g1 d7e5 f3e5 e7e5 f1f3 e8f8 d2f4 g8g7 a1f1 e5f4 f3f4 f6e8 g2g4 h7h6 g1f2 c8d7 f2e3 f7f6 e3d2 e8c7 f4f2 a8b8 b2b3 b8e8 c3e2 e8e5 a2a3 b7b5 c4b5 d7b5 a3a4 b5d3');
INSERT INTO evaluations VALUES(36089,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(36089,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(36089,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(36089,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(36089,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(36089,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(36089,36,'cp 59','e6d5 e4d5 f8e8 c1e3 g7h6 e1g1 h6e3 f2e3 d8e7 e3e4 b8d7 d1d2 d7e5 f3e5 e7e5 f1f3 e8f8 d2f4 f6e8 g2g4 c8d7 a1f1 f7f6 f4d2 a7a6 a2a4 a8b8 d2h2 e8c7 a4a5 b7b5 a5b6 e5h2 g1h2 b8b6 c3d1 g8g7 g4g5 c7e8 h2g3 b6b7 f1f2');
INSERT INTO evaluations VALUES(36079,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(36079,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(36079,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(36079,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(36079,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(36079,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(36079,36,'cp 68','e4d5 f8e8 c1e3 g7h6 e1g1 h6e3 f2e3 d8e7 d1c2 b8d7 a1e1 e8f8 g2g4 f6e8 f1f2 d7e5 f3e5 e7e5 f2f3 b7b6 c2h2 e5h2 g1h2 a8b8 e1f1 b8b7 h2g3 b7e7 h3h4 g8g7 e3e4 h7h6 d3c2 c8d7 f1f2 a7a6 a2a4 e8c7 c3d1 b6b5 b2b3 b5c4 b3c4 c7e8');
INSERT INTO evaluations VALUES(54970,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(54970,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(54970,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(54970,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(54970,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(54970,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54970,36,'cp 62','b1c3 g8f6 e2e4 e7e5 g1f3 b8d7 f1c4 f8e7 e1g1 h7h6 a2a4 e8g8 h2h3 f8e8 f1e1 c7c6 a4a5 e5d4 f3d4 e7f8 c1f4 d7c5 d1f3 a7a6 c4f1 d6d5 e4d5 e8e1 a1e1 c6d5 f4e5 c5e4 d4b3 c8e6 e1d1 f8d6 e5d6 d8d6');
INSERT INTO evaluations VALUES(54964,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(54964,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(54964,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(54964,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(54964,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(54964,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(54964,36,'cp 50','g8f6 e2e4 e7e5 g1f3 b8d7 h2h3 c7c6 a2a4 f8e7 f1c4 e8g8 e1g1 f8e8 a4a5 d8c7 f1e1 e5d4 f3d4 e7f8 c1f4 a8b8 f4g3 b7b5 a5b6 a7b6 b2b4 b6b5 c4b3 h7h6 f2f4 c8b7 d4f5 b8a8 a1a8 b7a8 g1h1');
INSERT INTO evaluations VALUES(47010,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(47010,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(47010,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(47010,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(47010,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(47010,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(47010,36,'cp 52','e2e4 e7e5 g1f3 b8d7 h2h3 c7c6 a2a4 f8e7 f1c4 e8g8 e1g1 f8e8 a4a5 d8c7 f1e1 e5d4 f3d4 a8b8 c1f4 e7f8 f4g3 h7h6 d1e2 d7c5 e2f3 c5d7 c4b3 d7c5 g3h4 c5b3 c2b3 f6d7 h4g3 d7e5 f3e3 c8d7 f2f4 e5g6');
INSERT INTO evaluations VALUES(51974,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(51974,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(51974,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51974,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(51974,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(51974,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(51974,36,'cp 21','c5d4 f3d4 b8c6 b1c3 g8f6 a2a3 d7d5 f1b5 c8d7 e4d5 e6d5 e1g1 f8e7 f1e1 e8g8 c1e3 f8e8 h2h3 a7a6 b5a4 c6a5 a4b3 a5b3 c2b3 e7d6 a1c1 d7c6 b3b4 d6b8 c3e2 c6d7 e3f4 b8f4 e2f4');
INSERT INTO evaluations VALUES(52081,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52081,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(52081,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(52081,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(52081,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(52081,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(52081,36,'cp 39','f3d4 b8c6 b1c3 a7a6 c1e3 d8c7 d1f3 b7b5 f3g3 c7g3 h2g3 c6d4 e3d4 c8b7 f2f3 g8e7 a2a3 a8c8 e1c1 e7c6 d4b6 h7h5 a3a4 b5a4 c3a4 c6b4 d1d2 d7d5 e4d5 b4d5 b6d4 f7f6 a4b6 d5b6 d4b6 c8b8 f1e2 f8b4 c2c3 b7c6');
INSERT INTO evaluations VALUES(52059,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(52059,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(52059,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(52059,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(52059,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(52059,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(52059,36,'cp 43','g8f6 b1c3 b8c6 d4c6 b7c6 e4e5 f6d5 c3e4 d8c7 f2f4 c7b6 c2c4 d5e3 d1d3 f8b4 c1d2 e8g8 e4c3 e3f1 h1f1 f8d8 c3a4 b6b7 e1c1 d7d5 e5d6 b4d6 d3g3 b7e7 d2a5 d8d7 f1f2 e6e5 a5c3 f7f6 f4e5 d6e5 d1d7 e5g3 d7e7 g3f2 c1d2');
INSERT INTO evaluations VALUES(44292,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(44292,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(44292,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(44292,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(44292,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(44292,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(44292,36,'cp 38','b1c3 b8c6 d4c6 b7c6 e4e5 f6d5 c3e4 d8c7 f2f4 c7b6 c2c4 d5e3 d1d3 f8b4 c1d2 e8g8 e4c3 e3f1 h1f1 f7f6 e1c1 a8b8 c1c2 b4e7 b2b3 c6c5 h2h4 f6e5 f4e5 c8b7 d3d7 b7g2 f1f8 e7f8 d2e3 g2h3 c2b2 h3g4 d1d2 b8b7 d7d8 b6c6 d2f2 b7f7 f2f7');
INSERT INTO evaluations VALUES(44289,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(44289,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44289,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(44289,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44289,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(44289,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(44289,36,'cp 26','b8c6 d4c6 b7c6 e4e5 f6d5 c3e4 d8c7 f2f4 c7b6 c2c4 d5e3 d1d3 f8b4 c1d2 e8g8 e4c3 e3f1 h1f1 f7f6 e1c1 a8b8 c1b1 d7d5 e5d6 b4d6 b2b3 f8d8 d3e2 e6e5 g2g4 c8e6 f4f5 e6f7 d2c1 d6e7 d1d8 b8d8 c1e3 b6b4 c3e4 b4a3 g4g5 f6g5 e3g5');
INSERT INTO evaluations VALUES(57143,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(57143,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(57143,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(57143,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(57143,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(57143,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(57143,36,'cp 163','d4e5 b8c6 g1f3 d7d5 c1f4 f7f6 b1c3 f8b4 e2e3 g7g5 f4g3 g5g4 f3d4 c6e5 d1d2 c7c6 d4b3 h7h5 a2a3 b4f8 e1c1 h5h4 g3f4 e5g6 f1d3 g6f4 e3f4 f8d6 h1e1 e8f7 c3e4 d6c7 g2g3 g8e7 e4f6 e7g6 d3g6 f7g6 f6d5 c6d5 b3d4 d8f6 d4b5 c7b6 d2d5 b6f2 b5d6 f2e1');
INSERT INTO evaluations VALUES(57107,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(57107,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(57107,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(57107,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(57107,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(57107,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(57107,36,'cp 166','b8c6 g1f3 f7f6 e2e4 f6e5 f1c4 g8f6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8e7 f1e1 e8g8 e1e5 d4f5 g5e4 e7f6 e4f6 d8f6 e5e4 f5d6 e4f4 c8f5 b1c3 f6g6 h2h4 h7h5 c1e3 f5c2 d1d2 c2f5 a1e1 d6e4 d2d1 e4d6 c3e2 d6e4 d1d4 a8e8 d4a7 f5h3 g2g3 f8f7 f4f7');
INSERT INTO evaluations VALUES(22684,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(22684,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(22684,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(22684,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(22684,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(22684,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(22684,36,'cp 181','g1f3 d7d5 c1f4 f7f6 b1c3 f8b4 e2e3 g7g5 f4g3 g5g4 f3d4 b4c3 b2c3 c6e5 f2f4 g4f3 d4f3 c8g4 a1b1 g8e7 f1e2 e8g8 b1b7 d8c8 b7b3 e5c4 g3f4 c8e6 f3d4 g4e2 d4e6 e2d1 e1d1 f8f7 e6c7 a8d8 c7e6 d8e8 e6d4 e7g6 d4f5 f7d7 h1f1 g8f7 g2g3');
INSERT INTO evaluations VALUES(22682,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(22682,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(22682,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(22682,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(22682,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(22682,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(22682,36,'cp 133','d7d5 c1f4 f7f6 b1c3 f8b4 e2e3 g7g5 f4g3 g5g4 f3d4 c6e5 d1d2 c7c5 d4b3 c5c4 b3d4 d8a5 e1c1 b4c3 d2c3 a5c3 b2c3 c8d7 e3e4 d5e4 d1e1 e8c8 e1e4 g8e7 g3e5 f6e5 e4e5 e7c6 d4c6 d7c6 f1e2 c8b8 e2g4 c6g2 h1g1 g2c6 f2f4 d8g8 g1g3 g8f8 g4f5 h8g8 f5e6 g8g6 g3g6 h7g6');
INSERT INTO evaluations VALUES(21654,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(21654,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(21654,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(21654,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(21654,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(21654,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(21654,36,'cp 173','c1f4 f7f6 b1c3 f8b4 e2e3 g7g5 f4g3 g5g4 f3d4 c6e5 f2f4 e5f7 f4f5 h7h5 d4b5 b4c3 b2c3 f7e5 d1d5 d8d5 b5c7 e8f7 c7d5 c8f5 g3e5 f6e5 f1d3 g8e7 e1g1 e7d5 f1f5 f7e6 a1f1 a8d8 e3e4 d5f4 d3c4 e6d6 g2g3 d6c5 g3f4 c5c4 f5e5 h8e8 e5h5 e8e4 f4f5 c4c3 f5f6 c3c2 h5h7 b7b5 h7c7 c2b2 c7a7 d8f8 f6f7 b5b4');
INSERT INTO evaluations VALUES(21644,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(21644,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(21644,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(21644,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(21644,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(21644,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(21644,36,'cp 162','f7f6 b1c3 f8b4 e2e3 g7g5 f4g3 g5g4 f3d4 c6e5 f2f4 e5f7 f4f5 d8e7 d1d2 c7c5 d4b5 d5d4 b5c7 e8f8 e1c1 d4c3 b2c3 b4a5 c7a8 f8g7 e3e4 g8h6 g3f4 h8d8 d2e3 d8d1 c1d1 b7b5 h2h3 c8b7 h3g4 h6g8 f1b5 b7a8 b5c4 e7e4 e3e4 a8e4 c4f7 e4g2 h1h2 g2f3 d1d2 g7f7 h2h7 f7e8 g4g5 f6g5 f4g5 a5b6 c3c4 f3e4 h7g7 e8f8 f5f6 a7a6 d2e3');
INSERT INTO evaluations VALUES(49620,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(49620,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(49620,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(49620,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(49620,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(49620,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(49620,36,'cp 7','g1f3 b8c6 f1b5 c6d4 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d1e2 f6e6 d2c3 f8c5 e2e6 f7e6 f2f3 e8g8 b5d3 b7b6 c1d2 c8b7 e1c1 h7h6 h2h4 c5f2 d1f1 f2g3 h1h3 g3f4 f1d1 f4d2 d1d2 a8d8 h4h5 f8f4 b2b3 e6e5 a2a4');
INSERT INTO evaluations VALUES(49391,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(49391,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(49391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49391,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49391,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49391,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(49391,36,'cp 16','f8b4 g1f3 b7b6 e2e3 c7c5 f1d3 d7d5 e1g1 e8g8 d4c5 d5c4 d3c4 d8d1 f1d1 b4c5 b2b3 b8c6 c1b2 c8b7 c4e2 f8d8 a1c1 d8d1 c1d1 c6b4 a2a3 b4d5 c3d5 f6d5 d1c1 c5e7 f3e5 f7f6 e5c6 e7c5 b3b4 b7c6 b4c5 b6b5 e2f3');
INSERT INTO evaluations VALUES(42481,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42481,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(42481,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(42481,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(42481,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42481,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(42481,36,'cp 9','g1f3 e8g8 e2e3 d7d5 f1d3 d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 d1e2 c8b7 c1g5 b4c3 b2c3 b8d7 c4b3 d8c7 c3c4 h7h6 g5h4 c7f4 h4g3 b7f3 e2f3 f4f3 g2f3 a8c8 f1e1 f6h5 g1f1 h5g3 h2g3 f8d8 a1d1 d8e8 f3f4 d7f6 e1e3 g8f8 b3a4');
INSERT INTO evaluations VALUES(42477,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42477,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42477,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42477,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(42477,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(42477,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42477,36,'cp 5','e8g8 e2e3 d7d5 f1d3 d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 c1g5 b8d7 f3e5 c8b7 e5d7 d8d7 g5f6 g7f6 d4d5 b4c3 b2c3 b7d5 d1g4 g8h8 g4d4 h8g7 f1d1 f8g8 d1d3 g7f8 g2g3 d7c6 c4d5 c6d5 d4d5 e6d5 d3d5 f8g7 a1d1 g8e8 d5h5 a8c8 d1d4 c8c3');
INSERT INTO evaluations VALUES(37177,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(37177,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37177,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37177,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(37177,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(37177,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37177,36,'cp 27','e2e3 d7d5 f1d3 b7b6 e1g1 c7c5 d4c5 d5c4 d3c4 d8d1 f1d1 b4c5 a2a3 c5e7 f3d4 c8b7 f2f3 f8d8 c4e2 b8c6 d4c6 b7c6 e3e4 e7c5 g1f1 d8d1 c3d1 a8d8 c1f4 f6h5 f4e3 c5e3 d1e3 h5f4 a1c1 d8c8 e2a6');
INSERT INTO evaluations VALUES(37175,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(37175,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37175,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(37175,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(37175,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(37175,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(37175,36,'cp 4','d7d5 c1d2 b7b6 c4d5 e6d5 f1d3 c7c5 d4c5 b6c5 e1g1 b8c6 a2a3 b4a5 d3b5 c6e7 b2b4 c5b4 a3b4 a5b4 c3e4 d5e4 d2b4 f6d5 b4e7 d5e7 f3d4 a7a5 d1c2 e7f5 b5c6 a8a6 f1d1 f5d4 d1d4 d8c7 c6e4 c7c2 e4c2 c8e6');
INSERT INTO evaluations VALUES(37043,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(37043,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37043,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(37043,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37043,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37043,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37043,36,'cp 14','f1d3 d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 c1g5 c8b7 d1e2 b4c3 b2c3 b8d7 c4b3 d8c7 c3c4 h7h6 g5h4 c7f4 h4g3 b7f3 e2f3 f4f3 g2f3 a8c8 a1c1 f8d8 f1d1 f6h5 g3d6 g7g5 c4c5 h5f4 g1h1 f4g6 h2h3 d7f6 a2a3 g6f4 d6f4');
INSERT INTO evaluations VALUES(37049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37049,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(37049,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37049,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(37049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37049,36,'cp 9','d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 d1e2 c8b7 f1d1 h7h6 c1f4 b8c6 d4d5 f6d5 c3d5 e6d5 c4d5 d8f6 f4e3 f8d8 a1c1 b4f8 h2h3 d8d6 a2a3 a8d8 d5e4 d6d1 c1d1 d8d1 e2d1 c6a5 e4b7 a5b7 f3d4 a7a5 d1c2 b7c5 b2b4 a5b4 a3b4 c5e6 c2c6 f8b4');
INSERT INTO evaluations VALUES(37084,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37084,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37084,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37084,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37084,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37084,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37084,36,'cp 0','d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 d1e2 c8b7 f1d1 h7h6 c1e3 b4c3 b2c3 d8c7 c4d3 b8d7 e3d2 a8c8 a1c1 f8e8 d1e1 f6d5 e2e4 d5f6');
INSERT INTO evaluations VALUES(37082,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37082,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37082,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37082,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37082,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37082,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37082,36,'cp 0','c7c5 e1g1 c5d4 e3d4 b7b6 d1e2 h7h6 c1f4 c8b7 f1d1 b8c6 d4d5 f6d5 c3d5 e6d5 c4d5 d8f6 f4e3 f8d8 a2a3 b4f8 a1c1 d8d6 d5e4 d6e6 e4d5');
INSERT INTO evaluations VALUES(49533,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(49533,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(49533,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(49533,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(49533,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(49533,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(49533,36,'cp -65','f6e4 f3e5 d7d5 c4e2 f8d6 e5f3 e8g8 e1g1 f8e8 d2d4 c8f5 c2c4 d5c4 e2c4 b8d7 b1c3 d7b6 c4d3 d8f6 c3e4 f5e4 d3e4 e8e4 c1d2 f6f5 h2h3 a8e8 d1c2 g7g6 c2b3 f5e6 b3e6 e4e6 f1e1 f7f6 e1e6 e8e6 a1c1 g8f7 g1f1 c7c6 c1e1 b6d5 a2a3 e6e1 f1e1');
INSERT INTO evaluations VALUES(49889,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(49889,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(49889,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(49889,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(49889,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(49889,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(49889,36,'cp -56','f3e5 d7d5 c4e2 f8d6 e5f3 e8g8 e1g1 f8e8 d2d4 c7c6 c2c4 c8e6 d1b3 d5c4 e2c4 e6c4 b3c4 b8d7 b1c3 d7f6 c4d3 h7h6 f1e1 e4c3 b2c3 e8e1 f3e1 b7b5 c3c4 a7a5 c1e3 b5b4 e1f3 a5a4 a1b1 c6c5');
INSERT INTO evaluations VALUES(49822,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(49822,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(49822,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(49822,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(49822,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(49822,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(49822,36,'cp -66','d7d5 c4e2 f8d6 e5f3 e8g8 e1g1 f8e8 d2d4 b8c6 c2c4 d5c4 e2c4 c8g4 c1e3 d8f6 b1c3 f6g6 a1c1 h7h6 c4d3 e4c3 c1c3 g6f6 h2h3 g4f3 d1f3 f6f3 g2f3 c6e7 c3b3 b7b6 f1c1 a8c8 g1f1 c8d8 d3b5 e8f8 b5d3 e7d5 d3e4 d5f4 e3f4');
INSERT INTO evaluations VALUES(49045,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(49045,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(49045,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(49045,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(49045,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(49045,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(49045,36,'cp -54','c4e2 f8d6 e5f3 e8g8 e1g1 h7h6 d2d3 e4f6 c2c4 f8e8 b1c3 b8c6 h2h3 d5d4 c3b5 d6c5 c1f4 c5b6 f1e1 c8f5 e2f1 e8e1 d1e1 a7a6 b5a3 a6a5 e1d2 f6d7 a1e1 d7c5 a3b5 c6b4 b5d4 c5d3 f1d3 b4d3');
INSERT INTO evaluations VALUES(10089,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(10089,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(10089,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10089,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10089,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(10089,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10089,36,'cp 15','d2d3 a7a6 c3d5 f6d5 c4d5 e8g8 c2c3 d7d6 e1g1 d8f6 h2h3 c5a7 d5b3 c8e6 c1g5 f6g6 g1h2 e6b3 d1b3 a8b8 a2a4 g6e6 b3e6 f7e6 b2b4 h7h6 g5e3 b8a8 e3a7 a8a7 h2g1 a7a8 a1d1 f8f7 a4a5 a8f8 f1e1 c6e7 d3d4');
INSERT INTO evaluations VALUES(10095,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10095,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10095,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(10095,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(10095,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(10095,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10095,36,'cp 7','a7a6 c3d5 f6d5 c4d5 d7d6 c2c3 c5a7 e1g1 e8g8 h2h3 d8f6 a2a4 c8e6 d5e6 f6e6 c1e3 a7e3 f2e3 c6e7 f1f2 h7h6 f3h4 a8d8 d1f3 d6d5 e4d5 e6d5 d3d4 e5d4 e3d4 d5g5 h4f5 e7f5 f3f5 g5f5 f2f5 f7f6');
INSERT INTO evaluations VALUES(7303,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(7303,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(7303,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(7303,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(7303,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(7303,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(7303,36,'cp 19','c3d5 f6d5 c4d5 d7d6 c2c3 d8f6 e1g1 e8g8 a2a4 c5a7 d5b3 a6a5 h2h3 c8e6 c1g5 f6g6 g1h2 a7b6 g5e3 b6e3 f2e3 g6h6 b3e6 h6e6 d1e2 c6b8 f1f2 b8d7 a1f1 e6e7 d3d4 a8e8 f3d2 e5d4 e3d4 d7f6');
INSERT INTO evaluations VALUES(55167,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55167,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55167,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(55167,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(55167,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(55167,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55167,36,'cp 0','c8f5 c2c3 c7c5 d2d4 e7e6 f1b5 b8c6 g1e2 d8b6 b5a4 f7f6 e1g1 f6e5 d4e5 e8c8 a4c6 b6c6 a2a4 c6c7 b1a3 c7e5 b2b4 a7a6 b4b5 e5c7 e2g3 f5g6 d1e2 a6a5 e2e6 c7d7 e6b6 d7c7');
INSERT INTO evaluations VALUES(31949,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31949,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31949,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31949,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31949,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31949,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31949,36,'cp 0','d2d4 e7e6 c2c3 c7c5 f1b5 b8c6 g1e2 d8b6 b5a4 f7f6 e1g1 f6e5 d4e5 e8c8 a4c6 b6c6 a2a4 c6c7 b1a3 c7e5 b2b4 a7a6 b4b5 e5c7 e2g3 f5g6 d1e2 a6a5 e2e6 c7d7 e6b6 d7c7');
INSERT INTO evaluations VALUES(31945,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31945,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31945,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31945,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31945,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31945,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31945,36,'cp 0','e7e6 c2c3 c7c5 f1b5 b8c6 g1e2 d8b6 b5a4 f7f6 e1g1 f6e5 d4e5 e8c8 a4c6 b6c6 a2a4 c6c7 b1a3 c7e5 b2b4 a7a6 b4b5 e5c7 e2g3 f5g6 d1e2 a6a5 e2e6 c7d7 e6b6 d7c7');
INSERT INTO evaluations VALUES(32357,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32357,36,'cp -1','c2c3 c7c5 f1b5 b8d7 g1e2 a8c8 e1g1 a7a6 b5d3 g8e7 d3f5 e7f5 b1d2 f8e7 d2f3 e8g8 d1d3 c5d4 c3d4 d8c7 e2f4 c7c2 g2g4 f5h4 f3h4 e7h4 c1e3 c2d3 f4d3 c8c4 f1c1 f8c8 c1c4 c8c4 a1c1 h4d8 f2f3 c4c1 d3c1 f7f6 e5f6 d8f6');
INSERT INTO evaluations VALUES(32353,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32353,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(32353,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(32353,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(32353,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(32353,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(32353,36,'cp 0','c7c5 f1b5 b8d7 g1e2 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 g8e7 b1c3 e8g8 e2f4 a8c8 e1g1 a7a6 b5d7 d8d7 a1c1 c8c7 f2f3 f8c8 g2g4 f5g6 h2h4 h7h6 h4h5 g6h7 g1g2 c7c4 g2g3 d7d8 f1h1 d8b6 f4e2 b6b4 c1e1 e7c6 e1c1');
INSERT INTO evaluations VALUES(31669,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(31669,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31669,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31669,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31669,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31669,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31669,36,'cp -5','f1b5 b8d7 g1e2 a8c8 b5a4 c5d4 e2d4 f5g6 h2h4 h7h5 b1d2 g8e7 d2f3 e7c6 c1g5 f8e7 g5e7 d8e7 d4c6 b7c6 a4c2 g6e4 c2e4 d5e4 f3g5 d7e5 e1g1 f7f5 f2f3 e4f3 g2f3 e5g6 d1b3 e8g8 a1e1 f8f6 g5e6 e7f7 b3c4');
INSERT INTO evaluations VALUES(57753,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(57753,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(57753,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(57753,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(57753,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(57753,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(57753,36,'cp -38','g8f6 g1f3 c7c5 g2g3 d7d5 f1g2 b8c6 f3e5 c6e5 f4e5 f6d7 c2c4 d7e5 c4d5 g7g6 d2d3 f8g7 b1c3 e8g8 c1f4 b7b5 e1g1 c8a6 d1d2 a8c8 a1c1 a6b7 e2e4 b5b4 c3d1 d8a5 b2b3 b7a6 d1b2 a5a2');
INSERT INTO evaluations VALUES(50359,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(50359,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(50359,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(50359,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(50359,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(50359,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(50359,36,'cp -21','g2g3 d7d5 g1f3 e7e6 e2e3 c7c5 b2b3 b8c6 a2a4 f8e7 f1g2 e8g8 e1g1 b7b6 d1e2 d8c7 c1b2 c8b7 f3e5 a7a6 d2d3 b6b5 b1d2 f6d7 e5d7 c7d7 d2f3 d5d4 e3d4 c5d4 f3e5 c6e5 g2b7 d7b7 f4e5 b7d5 a4b5 a6b5');
INSERT INTO evaluations VALUES(53009,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(53009,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53009,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(53009,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53009,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53009,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53009,36,'cp 0','g8f6 b1c3 a7a6 d2d4 c5d4 f3d4 e7e6 a2a4 f8e7 e1g1 b8c6 c1e3 e8g8 h2h3 c8d7 d1e2 a8c8 c4d3 c6b4 f1d1 e6e5 d4f3 d7e6 f3g5 e6d7');
INSERT INTO evaluations VALUES(45140,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(45140,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(45140,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45140,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45140,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45140,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(45140,36,'cp 20','b1c3 a7a6 d2d4 c5d4 f3d4 e7e6 c4b3 b7b5 c1e3 f8e7 a2a3 e8g8 f2f3 c8b7 d1e2 b8c6 e1c1 d8c7 c1b1 f8b8 g2g4 c6d4 e3d4 b7c6 e4e5 f6e8 e5d6 e8d6 h1e1 d6c4 f3f4 c7b7 c3e4 b8d8 b1a1 c6e4 e2e4 b7e4 e1e4 a8c8 b3c4 c8c4 c2c3 d8d5 g4g5 c4c8 a1b1 g7g6');
INSERT INTO evaluations VALUES(45126,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(45126,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45126,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(45126,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(45126,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45126,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45126,36,'cp 2','a7a6 d2d4 c5d4 f3d4 e7e6 e1g1 f8e7 a2a4 b8c6 c1e3 e8g8 d4c6 b7c6 a4a5 c8b7 f1e1 d6d5 e3b6 d8b8 c4d3 f8e8 c3a4 d5e4 d3f1 h7h6 h2h3 e7d6 a4c5 d6c5 b6c5 e8d8 d1e2 f6d7 e2e3 d7c5 e3c5 d8d5 c5e3');
INSERT INTO evaluations VALUES(43373,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43373,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(43373,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43373,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(43373,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43373,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43373,36,'cp 10','a2a4 e7e6 d2d4 c5d4 f3d4 f8e7 c1e3 b8c6 e1g1 e8g8 h2h3 d8c7 d1e2 c8d7 f2f4 f6e4 c3e4 d6d5 f4f5 e6f5 c4d5 f5e4 e2h5 c7g3 e3f2 g3g6 h5g6 h7g6 d5e4 e7f6 f1d1 f8e8 e4d5 a8c8 c2c3 c6d4 f2d4 f6d4 d1d4 d7c6 c3c4 e8e2 b2b4 c6d5 c4d5');
INSERT INTO evaluations VALUES(43393,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43393,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(43393,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(43393,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43393,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(43393,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(43393,36,'cp 9','e7e6 d2d4 d6d5 e4d5 e6d5 c4a2 c5c4 c1g5 f8e7 e1g1 e8g8 b2b3 b8c6 b3c4 d5c4 a2c4 c8g4 h2h3 g4f3 d1f3 c6d4 f3d3 a8c8 a1d1 d8c7 f1e1 e7d8 g5f6 d8f6 c3d5 c7c4 d5f6 g7f6 d3d4 c4d4 d1d4 b7b5 d4g4 g8h8 a4b5 a6b5 g4b4 c8c5 e1e7 h8g7 e7d7 h7h6 b4g4 g7h8 d7b7 c5c2 b7b5');
INSERT INTO evaluations VALUES(43502,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(43502,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43502,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(43502,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43502,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43502,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(43502,36,'cp 0','d2d4 c5d4 f3d4 f8e7 e1g1 b8c6 h2h3 c8d7 c1e3 e8g8 c4d3 d8c7 f1e1 c6a5 d4f3 a5c6');
INSERT INTO evaluations VALUES(43505,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43505,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(43505,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43505,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(43505,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(43505,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(43505,36,'cp 7','d6d5 e4d5 e6d5 c4a2 c5c4 c1g5 f8e7 e1g1 e8g8 b2b3 b8c6 b3c4 d5c4 a2c4 c8g4 h2h3 g4f3 d1f3 c6d4 f3d3 a8c8 a1d1 d8c7 f1e1 e7d8 g5f6 d8f6 c3d5 c7c4 d5f6 g7f6 d3d4 c4d4 d1d4 c8c2 e1e7 b7b5 e7b7 b5a4 d4g4 g8h8 g4a4 c2c6 a4g4 h7h6 b7e7 h8h7 g1h2 c6b6 g4c4 a6a5');
INSERT INTO evaluations VALUES(23146,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(23146,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(23146,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(23146,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(23146,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(23146,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(23146,36,'cp 66','d8f6 d4e3 f8b4 c2c3 b4e7 f1d3 f6h6 e3h6 g8h6 f2f3 d7d6 b1a3 c7c6 c1f4 f7f5 f4h6 f5e4 h6g7 h8g8 d3e4 g8g7 g2g3 d6d5 e4d3 c8h3 f3f4 h7h5 a3c2 h5h4 c2d4 h4g3 h2g3 g7g3 e1c1 e8f7 d3f5 h3f5 d4f5 g3f3 h1h7 f7g6 h7e7 g6f5 e7b7 f3f4');
INSERT INTO evaluations VALUES(1969,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(1969,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(1969,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(1969,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(1969,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(1969,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(1969,36,'cp 71','d4e3 f8b4 c2c3 b4e7 f1d3 f6h6 e3h6 g8h6 f2f3 d7d6 b1a3 h6g8 a3b5 e7d8 c1e3 a7a6 b5d4 h7h5 h2h4 g8f6 e1f2 e8g8 g2g3 c7c6 d3c2 f8e8 e3f4 d6d5 e4e5 f6d7 e5e6 f7e6 c2g6 e8e7 g6h5');
INSERT INTO evaluations VALUES(1970,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(1970,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(1970,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(1970,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(1970,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(1970,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(1970,36,'cp 51','f8b4 c2c3 b4e7 f1d3 f6h6 e3h6 g8h6 c1h6 g7h6 a2a4 h6h5 b1a3 a7a5 a3b5 e8d8 d3c4 c7c6 b5d4 h8f8 e1c1 d8c7 f2f4 e7c5 h1f1 d7d6 f4f5 c8d7 f1f4 c5d4 c3d4 d6d5 e4d5 c6d5 c4d5 d7a4 d1e1 a4c6');
INSERT INTO evaluations VALUES(47114,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(47114,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(47114,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(47114,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(47114,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(47114,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(47114,36,'cp 71','d2d4 d8d5 b1c3 d5a5 g1f3 c8f5 f3e5 c7c6 c1f4 b8d7 e5c4 a5d8 c4e3 f5e6 a2a4 g7g6 a4a5 f8g7 f1e2 b7b5 a5b6 a7b6 e2f3 e8g8 e1g1 a8a1 d1a1 d8c8 a1a3 f8e8 f1a1 b6b5 h2h3 c8b7 a3a7 b7a7 a1a7 f6h5 f3c6 h5f4 c6d7 g7d4 c3b5 d4a7 d7e8');
INSERT INTO evaluations VALUES(47100,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(47100,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(47100,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(47100,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(47100,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(47100,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(47100,36,'cp 71','d8d5 b1c3 d5a5 g1f3 c8f5 f3e5 c7c6 h2h4 b8d7 e5c4 a5c7 d1f3 f5g4 c1f4 g4f3 f4c7 f3d5 c3d5 c6d5 c4e3 e7e6 f1d3 a8c8 c7f4 f8e7 h4h5 h7h6 a2a4 d7b6 a4a5 b6c4 d3c4 d5c4 c2c3 e8g8 a1a4 f6d5 f4h2 f8e8 e1g1 d5e3 f2e3 e7g5');
INSERT INTO evaluations VALUES(33685,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(33685,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(33685,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(33685,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(33685,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(33685,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(33685,36,'cp 62','b1c3 d5a5 g1f3 c8f5 f3e5 c7c6 c1f4 b8d7 e5c4 a5d8 c4e3 e7e6 e3f5 e6f5 d4d5 f8b4 d1e2 d8e7 e2e7 e8e7 d5c6 b4c3 b2c3 b7c6 e1c1 f6e4 f1d3 e4c3 d1e1 e7f6 f4d2 c3a4 g2g4 g7g6 g4g5 f6g7 e1e7 d7b6 h1e1 b6d5 d3c4 d5e7 e1e7 h8d8 e7f7 g7h8 c4e6');
INSERT INTO evaluations VALUES(33681,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(33681,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(33681,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(33681,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(33681,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(33681,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(33681,36,'cp 73','d5d8 g1f3 e7e6 f1c4 a7a6 a2a4 f8e7 e1g1 b8c6 c1f4 e8g8 f1e1 c6b4 d1e2 b7b6 a1d1 c8b7 f3e5 d8e8 c4b3 e7d6 h2h3 g8h8 f4g5 b4d5 c3d5 e6d5 g5h4 f6e4 f2f3 f7f6 f3e4 f6e5 d4e5 e8e5');
INSERT INTO evaluations VALUES(47366,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(47366,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(47366,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(47366,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(47366,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(47366,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(47366,36,'cp 77','g1f3 e7e6 f1c4 b8c6 e1g1 f8e7 a2a3 e8g8 f1e1 b7b6 d4d5 c6a5 c4a2 f6d5 c3d5 e6d5 d1d5 d8d5 a2d5 c7c6 e1e7 c6d5 c1d2 c8e6 f3d4 f8c8 d2a5 b6a5 a3a4 g8f8 e7b7 c8b8 b7b5 a7a6 b5b8 a8b8 b2b3 b8c8 f2f3 f8e7 g1f2 e7d6 f2e3 g7g5 g2g3 g5g4 f3g4 e6g4 e3d2');
INSERT INTO evaluations VALUES(47360,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(47360,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(47360,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(47360,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(47360,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(47360,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(47360,36,'cp 71','e7e6 f1c4 f8e7 e1g1 b8c6 a2a3 e8g8 f1e1 b7b6 d4d5 c6a5 c4a2 f6d5 c3d5 e6d5 d1d5 d8d5 a2d5 c7c6 e1e7 c6d5 c1d2 c8e6 d2a5 b6a5 f3d4 f8e8 e7e8 a8e8 a3a4 g8f8 f2f3 f8e7 g1f2 e7d6 b2b3 g7g5 h2h4 g5h4 a1h1 h4h3 g2g4 d6e5');
INSERT INTO evaluations VALUES(48508,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(48508,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(48508,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(48508,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(48508,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(48508,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(48508,36,'cp 73','c1g5 f8e7 f1d3 e8g8 d1e2 b8c6 e1c1 c6b4 h1e1 b4d3 e2d3 f8e8 a2a3 a7a5 d4d5 h7h6 g5f6 e7f6 c3e4 f6e7 d3c4 e7d6 e4d6 c7d6 d5e6 e8e6 c4d4 e6f6 d4d5 c8g4 d5b7 a8b8 b7e4 g4f3 g2f3 d8b6');
INSERT INTO evaluations VALUES(48448,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(48448,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(48448,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(48448,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(48448,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(48448,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(48448,36,'cp 76','f8e7 f1d3 b8c6 a2a3 f6d5 g5d2 e7f6 c3e2 d5e7 d2f4 b7b6 c2c3 c8b7 d1c2 e7g6 d3e4 g6f4 e2f4 d8d6 g2g3 e8c8 e1c1 c6a5 h1e1 b7e4 c2e4 d6c6 e4c6 a5c6 c1c2 a7a5 h2h4 a5a4 h4h5 c6a5 g3g4 h8e8 f4d3 a5c4 g4g5 f6e7 f3e5 c4e5 e1e5 c8b7');
INSERT INTO evaluations VALUES(41022,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(41022,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(41022,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(41022,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(41022,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(41022,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(41022,36,'cp 92','f1d3 e8g8 d1e2 b8c6 e1c1 c6b4 h1e1 b4d3 e2d3 f8e8 a2a3 a7a5 d4d5 h7h6 d3c4 e7d6 g5h4 e6d5 c3d5 e8e1 d5f6 g7f6 f3e1 a8a6 h4g3 g8g7 c4c3 d8e7 g3d6 a6d6 c3g3 g7h7 d1d6 c7d6 g3d3 h7g7 e1f3 c8g4 d3e3 e7e3 f2e3 f6f5 c1d2 g4f3 g2f3 g7g6 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(41025,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(41025,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(41025,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(41025,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(41025,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(41025,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(41025,36,'cp 79','b8c6 a2a3 f6d5 g5d2 e7f6 c3e2 d5e7 d2f4 b7b6 c2c3 c8b7 d1c2 e7g6 d3e4 g6f4 e2f4 d8d6 g2g3 e8c8 e1c1 c6a5 h1e1 b7e4 c2e4 d6c6 e4c6 a5c6 c1c2 h7h5 h2h4 a7a5 b2b4 a5a4 e1e4 c6e7 f3e5 f6e5 e4e5 g7g6 b4b5 e7f5 d4d5');
INSERT INTO evaluations VALUES(55224,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(55224,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(55224,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55224,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(55224,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(55224,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(55224,36,'cp 21','e7e6 g1f3 a7a6 f1c4 g8f6 e1g1 c7c5 d4c5 f8c5 d1d8 e8d8 c4e2 d8e7 f1d1 c8d7 b1c3 h8c8 f3e5 b7b5 a2a4 b5b4 c3b1 d7e8 b1d2 b8d7 e5c4 a6a5 d2b3 d7b6 e2f3 f6d5 c4a5 e8a4 c1d2 a8a5 b3a5');
INSERT INTO evaluations VALUES(56020,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(56020,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(56020,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56020,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(56020,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(56020,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(56020,36,'cp 30','f1c4 g8f6 g1f3 c7c5 e1g1 a7a6 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 f1d1 c8d7 b2b3 d7b5 c1b2 b8d7 h2h3 h8c8 a1c1 e7f8 g2g4 h7h6 c4f1 b5f1 g1f1 f6d5 d2c4 f8g8 h3h4 b7b5 f3e5 d7e5 c4e5 c5e7 e5c6 e7f6 b2f6');
INSERT INTO evaluations VALUES(56007,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(56007,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(56007,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(56007,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(56007,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(56007,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(56007,36,'cp 28','c7c5 g1f3 a7a6 e1g1 g8f6 b2b3 c5d4 f3d4 f8d6 c1b2 e8g8 b1d2 e6e5 d4f3 b8c6 c4e2 d6c7 d1c2 d8e7 a1c1 h7h6 f1d1 f8e8 h2h3 c8e6 a2a3 a8c8 e2a6 c6b4 a3b4 b7a6 d2c4 f6d5 c2d2 e7b4 d2b4 d5b4 f3e5');
INSERT INTO evaluations VALUES(54091,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54091,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54091,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(54091,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(54091,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(54091,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(54091,36,'cp 34','g1f3 a7a6 e1g1 g8f6 d4c5 f8c5 d1d8 e8d8 c4e2 d8e7 b2b3 b8d7 f3d4 d7b6 c1b2 c8d7 e2f3 b6d5 f1c1 a8c8 b1d2 c8c7 c1c5 c7c5 b2a3 h8c8 e3e4 d5f4 e4e5 f6d5 d2e4 b7b6 e4c5 b6c5 a1d1 d5b4 d1c1 c5d4');
INSERT INTO evaluations VALUES(54092,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54092,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(54092,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54092,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54092,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(54092,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54092,36,'cp 31','g8f6 e1g1 a7a6 b1d2 b8d7 b2b3 b7b6 c1b2 c8b7 a1c1 f8e7 d4c5 d7c5 b3b4 c5d7 b4b5 e8g8 a2a4 d7c5 b2d4 f6d7 d1c2 e7f6 f1d1 d8e7 b5a6 c5a6 c4b5 a8c8 c2b1 c8c1 b1c1 f8c8 c1b1 a6c5 d2c4 f6d4 d1d4');
INSERT INTO evaluations VALUES(46117,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(46117,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(46117,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(46117,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(46117,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(46117,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(46117,36,'cp 25','e1g1 a7a6 b2b3 c5d4 f3d4 f8d6 c4e2 e6e5 d4f3 e8g8 b1d2 b8c6 c1b2 d6c7 d1c2 d8e7 a1c1 h7h6 f1d1 f8d8 a2a3 c8e6 b3b4 e6d5 h2h3 a8c8 c2b1 e7e6 b4b5 a6b5 e2b5 e5e4 f3d4 c6d4 b2d4 c7e5 d4e5 c8c1 d1c1 e6e5 d2c4 e5g5');
INSERT INTO evaluations VALUES(46116,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46116,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(46116,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(46116,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(46116,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(46116,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(46116,36,'cp 30','a7a6 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 b2b3 b8d7 c4e2 b7b6 f1d1 h8d8 a2a3 c8b7 b3b4 c5d6 c1b2 b6b5 d2b3 a8c8 b3a5 b7d5 f3d4 d7b6 a1c1 e7d7 f2f3 c8c1 d1c1 d8c8 c1c8 d7c8 d4c6 d5c6 a5c6 c8c7 c6e5 b6d5 e5f7');
INSERT INTO evaluations VALUES(43559,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(43559,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(43559,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43559,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(43559,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43559,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43559,36,'cp 27','d4c5 f8c5 d1d8 e8d8 c4e2 d8e7 b2b3 b8d7 f3d4 d7e5 c1b2 c8d7 a2a3 h8c8 b1d2 c5d6 a1c1 c8c1 f1c1 a8c8 c1c8 d7c8 f2f4 e5d7 e2f3 g7g6 d2c4 e6e5 f4e5 d7e5 c4e5 d6e5 d4f5 e7e6 b2e5 e6e5 f5e7 c8e6 f3b7 e6b3 b7a6');
INSERT INTO evaluations VALUES(52747,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52747,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52747,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(52747,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(52747,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(52747,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52747,36,'cp 25','g8f6 g1f3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 c4e2 b4e7 a2a3 a7a6 b5d6 b7b5 d1c1 e7d6 c1c6 d6b8 h4f6 g7f6 g2g3 c8b7 c6b6 b7d5 f1c1 f8d8 g1f1 f6f5 f3d4 b8e5');
INSERT INTO evaluations VALUES(44905,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44905,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(44905,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(44905,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44905,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(44905,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(44905,36,'cp 21','g1f3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 b5d6 f8d8 d6e4 d8d1 e4f6 g7f6 f1d1 b4e7 f3d2 c6e5 c4e2 e5g6 h4g3 c8d7 d2f3 d7c6 f3d4 c6d5 d1c1');
INSERT INTO evaluations VALUES(19378,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(19378,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19378,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19378,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19378,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19378,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(19378,36,'cp 6','g1f3 e7e6 e1g1 d7d5 e4d5 e6d5 c4b3 g8f6 d2d4 c8e6 c2c3 h7h6 f1e1 d8b6 h2h3 f8e7 d4c5 e7c5 c1e3 e8g8 e3c5 b6c5 b1d2 a8d8 d1c2 b7b5 a1d1 f8e8 d2f1 d5d4 f3d4 c6d4 d1d4 d8d4 c3d4 c5c2 b3c2 e6a2 e1e8 f6e8 f1e3 e8f6');
INSERT INTO evaluations VALUES(22712,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(22712,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(22712,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(22712,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(22712,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(22712,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22712,36,'cp 16','g1f3 f8c5 e1g1 g8f6 d2d3 d7d6 c2c3 a7a6 h2h3 e8g8 b1d2 c8e6 f1e1 h7h6 b2b4 c5a7 c4b3 e6b3 d1b3 f8e8 a2a4 d6d5 e4d5 d8d5 c3c4 d5d7 c1b2 a8d8 c4c5 d7d3 b4b5 d3b3 d2b3 c6b4 e1e5 b4d3 e5e8 d8e8 b2d4 f6d5 a1d1');
INSERT INTO evaluations VALUES(50324,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(50324,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(50324,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50324,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(50324,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(50324,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(50324,36,'cp 23','g1f3 e7e6 b1c3 d7d5 d2d4 c7c6 c1g5 f8e7 e2e3 b8d7 a2a4 e8g8 f1e2 b7b6 e1g1 d5c4 e2c4 c8b7 d1e2 a7a6 g5h4 b6b5 c4d3 b5b4 h4f6 g7f6 c3e4 f6f5 e4c5 d7c5 d4c5 e7c5 e3e4 d8f6 e4f5 e6f5 a1e1');
INSERT INTO evaluations VALUES(50322,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(50322,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50322,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50322,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(50322,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(50322,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50322,36,'cp 19','c7c6 d2d4 d7d5 b1c3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 c4e2 a7a6 b5d4 c8d7 d4c6 d7c6 f3e5 a8c8 h4f6 g7f6 e5c6 c8c6 a2a3 b4a5 b2b4 a5b6 d1d7 c6c7 f1d1 f8c8 d7c7 b6c7');
INSERT INTO evaluations VALUES(44468,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(44468,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(44468,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(44468,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(44468,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44468,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(44468,36,'cp 33','d2d4 d7d5 b1c3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 f3e5 f6d7 e5d7 c8d7 c3e4 f8c8 b2b3 d7c6 e4d6 c8c7 h4g3 c7d7 d6b5 d7d1 f1d1 b8d7 b5d4 d7b6 d4c6 b7c6 d1c1 b6c4 c1c4');
INSERT INTO evaluations VALUES(44475,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(44475,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(44475,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(44475,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44475,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(44475,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44475,36,'cp 27','d7d5 b1c3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 h4g3 c8d7 a2a3 b4e7 c4e2 f8d8 b5d6 e7d6 g3d6 d7e8 h2h3 f6e4 d6c7 d8d1 f1d1 a8c8 c7h2 g7g5 f3d2 e4d2 d1d2 c6a5 f2f3 a5b3');
INSERT INTO evaluations VALUES(52441,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(52441,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52441,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52441,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(52441,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(52441,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(52441,36,'cp 2','g8f6 e4e5 f6d5 d2d4 c5d4 c3d4 d7d6 g1f3 b8c6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 b5c6 d7c6 e1g1 e7e6 e5c6 b7c6 c1e3 f8e7 b1c3 e8g8 a1c1 d8d7 d1f3 h7h6 b2b3 a7a5 f1d1 a5a4 c3a4 b6a4 b3a4 a8a4 c1c6 f8a8 d4d5 e6d5');
INSERT INTO evaluations VALUES(44566,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44566,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(44566,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44566,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(44566,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(44566,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44566,36,'cp 5','e4e5 f6d5 d2d4 c5d4 g1f3 b8c6 c3d4 d7d6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5d3 a8c8 c1e3 c6b4 d3e4 f8e7 d1g4 e8g8 e3h6 e7f6 f1d1 b6d5 a1c1 f8d8 c3d5 b4d5 c1c5 g8h8 h6g5');
INSERT INTO evaluations VALUES(44513,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(44513,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44513,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(44513,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44513,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44513,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44513,36,'cp 7','f6d5 g1f3 b8c6 f1c4 d5b6 c4b3 c5c4 b3c2 d7d5 e5d6 d8d6 e1g1 c8g4 f1e1 c6e5 e1e5 g4f3 d1e1 f3c6 b2b3 g7g6 c1a3 d6d7 c2e4 a8c8 e1e2 f7f6 e5a5 a7a6 a3c5 b6d5 b1a3 d5f4 e2c4 d7d2 e4c6 c8c6 c5e3 c6c4');
INSERT INTO evaluations VALUES(44665,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44665,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44665,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(44665,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44665,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(44665,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(44665,36,'cp 15','g1f3 b8c6 f1c4 d5b6 c4b3 d7d6 e5d6 c5c4 b3c2 e7e5 e1g1 f8d6 a2a4 e8g8 d2d3 c8f5 d3c4 e5e4 b1d2 e4f3 c2f5 f3g2 f1e1 d8h4 d2f3 h4h5 f5h7 h5h7 d1d6 a8e8 c1f4 h7h3 f4g3 h3g4 d6f4 g4f4 g3f4 b6c4 e1e8 f8e8');
INSERT INTO evaluations VALUES(44664,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(44664,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44664,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(44664,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44664,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44664,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44664,36,'cp 0','d7d6 d2d4 c5d4 c3d4 b8c6 f1c4 d6e5 d4e5 d5b4 e1g1 d8d1 f1d1 c8f5 b1c3 b4c2 a1b1 c2b4');
INSERT INTO evaluations VALUES(45275,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45275,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45275,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45275,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45275,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45275,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(45275,36,'cp 0','d2d4 c5d4 c3d4 b8c6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5c6 d7c6 c1e3 b6d5 d1g4 g7g6 a1c1 d5e3 f2e3 f7f5 g4f4 c6d7 c3e2 f8d6 f4h6 d6f8');
INSERT INTO evaluations VALUES(45273,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45273,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45273,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45273,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45273,36,'cp 0','c5d4 f1c4 d6e5 f3e5 e7e6 d1d4 d8b6 e5f3 b8c6 d4e4 f8e7 e1g1 e8g8 f1e1 f8d8 c4b3 b6c7 a2a4 d5f6 e4h4 f6d5');
INSERT INTO evaluations VALUES(55595,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55595,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(55595,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(55595,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(55595,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(55595,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(55595,36,'cp 31','c7c5 f1c4 f8e7 d2d3 b8c6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 e8g8 c2c3 c8d7 e1g1 d7c6 d1b3 g7g6 f1b1 g8g7 h2h3 a8c8 d5c6 b7c6 b3c2 d8d7 b2b4 c5b4 b1b4 a7a5 b4b6 f6d8 b6a6 c8b8 a1b1 b8b1 c2b1');
INSERT INTO evaluations VALUES(53895,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(53895,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(53895,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(53895,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(53895,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(53895,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(53895,36,'cp 25','f1c4 f8e7 d2d3 b8c6 e1g1 g8f6 c1g5 a7a6 a2a4 e8g8 g5f6 e7f6 c3d5 g7g6 c2c3 a8b8 d5f6 d8f6 f3e1 b7b5 a4b5 a6b5 c4a2 f6g5 e1c2 h7h5 c2e3 c6d8 a2d5 d8e6 d5e6 c8e6 a1a6 f8d8 d1f3 b5b4 f1a1 b4c3 b2c3 h5h4');
INSERT INTO evaluations VALUES(53872,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53872,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(53872,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(53872,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(53872,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(53872,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(53872,36,'cp 34','b8c6 d2d3 f8e7 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 a8b8 a4a5 c8e6 c2c4 e8g8 e1g1 a7a6 d5e6 f7e6 d1e2 h7h6 f1b1 b7b5 a5b6 d8b6 a1a2 b6b7 b1a1 a6a5 a2a5 b7b2 e2b2 b8b2 a5a6 f8d8 a6a7 d8b8 h2h3');
INSERT INTO evaluations VALUES(20593,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(20593,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(20593,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(20593,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(20593,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(20593,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(20593,36,'cp 36','d2d3 f8e7 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 a8b8 c2c3 e8g8 h2h3 c8d7 e1g1 d7c6 d1b3 g7g6 b3a2 d8d7 f1c1 c6d5 a2d5 d7c6 d5c6 b7c6 b2b4 c5b4 c3b4 f8c8 a1b1 f6d8 g2g3 g8g7 g1g2 f7f5 b4b5 c6b5 c1c8 b8c8 a4b5');
INSERT INTO evaluations VALUES(20594,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(20594,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(20594,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(20594,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(20594,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(20594,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(20594,36,'cp 17','f8e7 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 e8g8 a4a5 a8b8 e1g1 c8e6 d5e6 f7e6 d1e2 b8c8 f1b1 d8d7 e2d1 f6d8 c2c4 h7h6 d1a4 d7e7 b2b4 b7b6 b4c5 c8c5 a5a6 d6d5 f3e5 d5c4 e5c4 d8c7 b1f1 g8h7');
INSERT INTO evaluations VALUES(23148,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(23148,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(23148,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(23148,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(23148,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(23148,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(23148,36,'cp 76','d1d4 d8f6 d4c4 c7c6 b1c3 b7b5 c4d3 b5b4 c3a4 f6e6 c1e3 g8f6 f2f3 f8e7 a4c5 e6e5 d3d4 e5d4 e3d4 e8g8 e1c1 d7d5 e4d5 f6d5 f1c4 a7a5 h1e1 f8e8 e1e5 c8e6 c5e6 f7e6 g2g3 a5a4 e5e4 a4a3 b2a3 a8a3 f3f4 g8f7 c4b3 e7f6 d4b2 f6b2 c1b2 g7g6 e4c4');
INSERT INTO evaluations VALUES(16307,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16307,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16307,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(16307,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16307,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16307,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(16307,36,'cp 0','c6d4 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d1e2 f6e6 e2e6 f7e6 d2c3 f8e7 c1e3 a7a6 b5d3 d7d6 d3e4 a8b8 e3a7 b8a8 a7e3');
INSERT INTO evaluations VALUES(16895,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(16895,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(16895,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(16895,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(16895,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16895,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(16895,36,'cp 6','f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d1e2 f6e6 d2c3 f8c5 e2e6 f7e6 f2f3 e8g8 b5d3 d7d5 c1d2 e6e5 e1c1 c8f5 h1e1 c5d6 d2e3 f5d3 c2d3 e5e4 f3e4 d5e4 g2g3 e4d3 d1d3 a8e8 d3d5 e8e6 c1d2 c7c6 d5a5 a7a6 d2d3 f8f3 e1e2 d6c7 a5g5 g8f8');
INSERT INTO evaluations VALUES(16891,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(16891,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16891,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(16891,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16891,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(16891,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(16891,36,'cp 1','e5d4 e4e5 d4c3 e5f6 d8f6 d1e2 f6e6 d2c3 f8c5 e2e6 f7e6 f2f3 e8g8 b5d3 d7d5 c1g5 c8d7 c3c4 a7a6 g5h4 b7b5 c4d5 e6d5 e1d2 c5b6 c2c3 f8e8 a1e1 g7g6 h4g3 e8e1 h1e1 g8f7 e1e2 a6a5 b2b3 a5a4 b3b4 a8e8 e2e8 d7e8');
INSERT INTO evaluations VALUES(16903,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(16903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16903,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(16903,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(16903,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(16903,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(16903,36,'cp 0','e4e5 d4c3 e5f6 d8f6 d2c3 f8c5 d1e2 f6e6 e2e6 f7e6 f2f3 a7a5 a2a4 e6e5 c1d2 c7c6 b5d3 d7d6 e1c1 c8e6 h1e1 e8d7 e1e2 h8f8 h2h4 h7h6 h4h5 f8f6 d1e1 a8f8 e1h1 f8a8');
INSERT INTO evaluations VALUES(16885,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16885,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16885,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16885,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16885,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16885,36,'cp 0','d4c3 e5f6 d8f6 d1e2 f6e6 d2c3 f8c5 c1f4 c7c6 b5d3 d7d5 h2h4 e8g8 f2f3 e6f6 f4g5 f6e6');
INSERT INTO evaluations VALUES(16887,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16887,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16887,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16887,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16887,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16887,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16887,36,'cp 0','e5f6 d8f6 d1e2 f6e6 d2c3 f8c5 c1f4 c7c6 b5d3 d7d5 a2a4 a7a5 e2e6 c8e6 h2h4 h7h5 e1g1 e8d7 f4e5 h8g8 d3h7 g8h8');
INSERT INTO evaluations VALUES(16881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16881,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16881,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16881,36,'cp 0','d8f6 d1e2 f6e6 d2c3 f8c5 c1f4 c7c6 b5d3 d7d5 a2a4 a7a5 e2e6 c8e6 h2h4 h7h5 e1g1 e8d7 f4e5 h8g8 d3h7 g8h8');
INSERT INTO evaluations VALUES(6427,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(6427,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(6427,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6427,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(6427,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(6427,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6427,36,'cp 12','g8f6 d2d3 a7a5 e1g1 d7d6 f1e1 e8g8 h2h3 c5a7 c4b3 h7h6 b1d2 c8e6 d2f1 e6b3 d1b3 a5a4 b3c2 d6d5 c1e3 d5d4 c3d4 c6d4 f3d4 a7d4 e3d4 d8d4 c2c7 d4b2 f1e3 f8c8 c7e7 c8e8 e7c5 e8c8 e1b1 c8c5 b1b2');
INSERT INTO evaluations VALUES(10103,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(10103,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10103,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(10103,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(10103,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10103,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(10103,36,'cp 11','d2d3 a7a6 a2a4 e8g8 e1g1 d7d6 a4a5 h7h6 c1e3 c5e3 f2e3 c6e7 b1d2 e7g6 h2h3 f8e8 d3d4 c8e6 d4d5 e6d7 d1b3 a8b8 c4d3 b7b5 a5b6 b8b6 b3c2 c7c6 c3c4 d8e7 f1f2 c6d5');
INSERT INTO evaluations VALUES(47143,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47143,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(47143,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47143,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(47143,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(47143,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47143,36,'cp 15','d5c4 g1f3 a7a6 e2e4 b7b5 e4e5 f6d5 a2a4 e7e6 a4b5 d5b6 f1e2 f8e7 e1g1 e8g8 c1e3 c8b7 b5a6 a8a6 d1c2 a6a1 f1a1 b8a6 c2d1 a6b4 f3d2 b4d5 c3d5 b7d5 d1c1 d8b8 d2c4 b6c4 e2c4 d5c4 c1c4 b8b2 a1c1 f8c8 c4c6 h7h6 h2h3');
INSERT INTO evaluations VALUES(47351,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47351,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47351,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47351,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(47351,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47351,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(47351,36,'cp 14','g1f3 a7a6 e2e4 b7b5 e4e5 f6d5 a2a4 e7e6 a4b5 d5b6 f1e2 f8e7 e1g1 e8g8 c1e3 a6b5 a1a8 b6a8 c3b5 a8b6 d1c2 c8b7 b5a3 b8c6 a3c4 c6b4 c2d1 b6c4 e2c4 b4d5 d1b3 d5e3 f2e3 d8a8 b3a2 a8a2 c4a2 c7c5 a2c4 b7f3 g2f3 c5d4 e3d4 f8d8 f1d1 d8d7');
INSERT INTO evaluations VALUES(47350,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(47350,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(47350,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(47350,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47350,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(47350,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47350,36,'cp 19','a7a6 e2e4 b7b5 e4e5 f6d5 a2a4 e7e6 a4b5 d5b6 f1e2 f8e7 e1g1 e8g8 c1e3 a6b5 a1a8 b6a8 c3b5 a8b6 d1c2 c8b7 b5a3 b8c6 a3c4 c6b4 c2d1 d8d7 c4b6 c7b6 d1d2 f8a8 f1c1 b7e4 h2h3 h7h6 f3e1 d7b7 g1h2 e4d5 e2f3 a8a4 f3d5 b4d5');
INSERT INTO evaluations VALUES(43606,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43606,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(43606,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(43606,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(43606,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(43606,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(43606,36,'cp 25','e2e4 b7b5 e4e5 f6d5 a2a4 e7e6 a4b5 d5b6 f1e2 f8e7 e1g1 e8g8 c1e3 d8d7 b2b3 c4b3 a1b1 a6b5 e2b5 d7d8 b1b3 c8a6 f3d2 a6b5 b3b5 b8c6 d2e4 g8h8 d4d5 e6d5 c3d5 b6c4 d5e7 d8e7 d1d5 e7e6 d5e6 f7e6 e4g5 c6e5 e3c5');
INSERT INTO evaluations VALUES(23446,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(23446,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(23446,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(23446,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(23446,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23446,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(23446,36,'cp 25','e7e5 f1b5 g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 d7d5 f1d3 g7g6 b1d2 e8g7 d2f3 c8f5 d3f5 g7f5 e1d2 d8d7 f3e5 f6e5 d4e5 f5g7 f4g5 c7c6 g5f6 d7f5');
INSERT INTO evaluations VALUES(52373,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(52373,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(52373,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(52373,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(52373,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(52373,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(52373,36,'cp 43','e2e4 d7d5 b1d2 d5e4 d2e4 c8f5 e4g3 f5g6 h2h4 h7h6 g1f3 b8d7 h4h5 g6h7 f1d3 h7d3 d1d3 e7e6 c1f4 d8a5 f4d2 a5c7 e1c1 g8f6 g3e4 a8d8 e4f6 d7f6 f3e5 a7a6 d2f4 f6d5 f4g3 f8d6 c2c4 d5b4 d3f3 c6c5 a2a3 c5d4 d1d4 e8g8 d4d6 c7d6 a3b4 d6b4 h1d1 d8d1 f3d1 b4a5 c1b1 f8d8');
INSERT INTO evaluations VALUES(43185,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(43185,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(43185,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(43185,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(43185,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(43185,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(43185,36,'cp 75','e2e4 d7d6 f1e2 e8g8 c1e3 b8a6 h2h3 e7e5 d4d5 f6e8 g2g4 f7f5 e4f5 g6f5 g4f5 c8f5 g1f3 a6c5 h1g1 g8h8 d1d2 e8f6 f3g5 f6e4 g5e4 c5e4 c3e4 f5e4 e1c1 e4g6 e2d3 g6d3 d2d3 d8f6 g1g4 g7h6 d1g1 h6e3 f2e3 f8g8');
INSERT INTO evaluations VALUES(43194,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(43194,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(43194,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(43194,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(43194,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(43194,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(43194,36,'cp 53','e8g8 f1e2 d7d6 c1e3 b8a6 g1f3 e7e5 e1g1 d8e7 d1c2 e5d4 f3d4 c7c6 f1e1 f8e8 f2f3 d6d5 c4d5 a6b4 c2d2 b4d5 e4d5 e7e3 d2e3 e8e3 d5c6 b7c6 d4c6 c8d7 e2b5 e3e1 a1e1 d7c6 b5c6 a8c8 c6b5 c8c7 e1c1 g7f8 c3e2 c7c1 e2c1 a7a5 c1d3 f8d6');
INSERT INTO evaluations VALUES(37414,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(37414,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(37414,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(37414,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(37414,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(37414,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(37414,36,'cp 64','f1e2 d7d6 g1f3 e7e5 e1g1 e5d4 f3d4 f8e8 f2f3 b8c6 c1e3 f6h5 d1d2 h5f4 d4c6 f4e2 c3e2 b7c6 e3d4 g7d4 e2d4 c8d7 b2b3 a7a5 d4e2 a5a4 b3b4 a4a3 a1b1 d8e7 f1e1 d7e6 b1c1 a8a4 e2d4 c6c5 d4c6 e7d7');
INSERT INTO evaluations VALUES(37413,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(37413,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(37413,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(37413,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(37413,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(37413,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(37413,36,'cp 61','d7d6 g1f3 e7e5 e1g1 e5d4 f3d4 f8e8 f2f3 b8c6 c1e3 f6h5 d1d2 h5f4 f1f2 c6d4 e3d4 g7d4 d2d4 d8g5 c3d5 f4d5 c4d5 c8d7 e2d3 g5e5 d4e5 e8e5 f2c2 c7c5 d5c6 d7c6 a1d1 e5c5 g1f2 g8f8 d1c1 c5c2 c1c2 f8e7 c2c3 a8c8');
INSERT INTO evaluations VALUES(36575,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(36575,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(36575,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(36575,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(36575,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(36575,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(36575,36,'cp 64','g1f3 e7e5 e1g1 b8a6 f1e1 c8g4 c1e3 e5d4 f3d4 g4e2 d1e2 a6c5 f2f3 c7c6 a1d1 f6d7 d4b3 d8e7 b3c5 d7c5 e2d2 a8d8 e3g5 g7f6 g5f4 f8e8 c3e2 f6e5 f4h6 e7h4 g2g3 h4h3 b2b3 f7f6 h6e3 c5e6 a2a4 e6c5 d2c2 a7a5');
INSERT INTO evaluations VALUES(36565,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(36565,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36565,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(36565,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(36565,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(36565,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(36565,36,'cp 48','e7e5 e1g1 b8a6 f1e1 c8g4 c1e3 g4f3 e2f3 e5d4 e3d4 f8e8 d1d2 a6c5 a1d1 c5e6 d4e3 h7h5 h2h3 f6d7 f3e2 d7c5 b2b4 g7c3 d2c3 c5e4 c3c2 e4f6 e2f3 c7c6 c2d2 d6d5 c4d5 f6d5 f3d5 c6d5 d2d5 d8d5 d1d5 e6f8 d5d6 a7a5 b4b5 a8d8 e1d1 d8d6 d1d6');
INSERT INTO evaluations VALUES(36867,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(36867,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(36867,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(36867,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(36867,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(36867,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(36867,36,'cp 71','e1g1 b8a6 c1e3 d8e7 d1c2 e5d4 f3d4 c7c6 a1d1 f8e8 f2f3 d6d5 c4d5 f6d5 c3d5 c6d5 e2a6 b7a6 d4c6 e7b7 d1d5 c8e6 d5c5 b7b2 c2b2 g7b2 c5c2 a8c8 c2b2 c8c6 f1d1 h7h5 h2h3 a6a5 d1d4 c6c4 d4c4 e6c4 e3a7 e8c8 a2a4 f7f5 a7e3 c4f7 e3g5 f5e4');
INSERT INTO evaluations VALUES(36866,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(36866,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(36866,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(36866,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(36866,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(36866,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(36866,36,'cp 45','b8a6 c1e3 d8e7 d4d5 f6g4 e3g5 f7f6 g5h4 g6g5 h4g3 f6f5 e4f5 c8f5 f1e1 g4h6 f3d2 a6c5 b2b4 c5d3 e2d3 f5d3 a1c1 d3g6 d2e4 h6f5 c4c5 h7h6 f2f3 a7a6 a2a4 a8d8 d1b3 f5d4 b3c4 e7f7 h2h3 g6f5 b4b5 f5e4 f3e4 d6c5 b5a6 b7a6 c4c5 d4b3');
INSERT INTO evaluations VALUES(3559,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(3559,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(3559,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(3559,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(3559,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(3559,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(3559,36,'cp 47','c1e3 d8e7 d4d5 f6g4 e3g5 f7f6 g5h4 g6g5 h4g3 f6f5 e4f5 c8f5 f1e1 g4h6 f3d2 a6c5 b2b4 c5d3 e2d3 f5d3 f2f3 e7d7 d2e4 d3e4 e1e4 f8f7 h2h3 a8f8 g3f2 h6f5 e4g4 f5d4 c3e4 h7h6 d1d3 d7f5 a2a4 f5g6');
INSERT INTO evaluations VALUES(17880,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(17880,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17880,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17880,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17880,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17880,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17880,36,'cp 0','d7c6 b1c3 f7f6 d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 h2h4 e8c8 f2f3 h7h5 d4e2 d7e6 e1f2 g8e7 e2f4 e6f7 c1e3 e7g6 f4g6 f7g6 a1d1 f8d6 c3e2 g6f7 b2b3 a6a5 a2a4 h8e8 c2c4 d6e7 e2f4 b7b6 f4d3 f7g6 d3f4');
INSERT INTO evaluations VALUES(17819,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17819,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17819,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17819,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17819,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17819,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17819,36,'cp 0','b1c3 f7f6 d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 h2h4 e8c8 f2f3 h7h5 d4e2 d7e6 e1f2 e6f7 c1e3 f8d6 a1d1 g8e7 e3f4 d6c5 f4e3');
INSERT INTO evaluations VALUES(17813,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17813,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17813,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17813,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17813,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17813,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17813,36,'cp 0','f7f6 d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 h2h4 h7h5 d4e2 d7e6 f2f3 e8c8 e1f2 e6f7 c1e3 f8d6 a1d1 g8e7 e3f4 d6c5 f4e3');
INSERT INTO evaluations VALUES(17835,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17835,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17835,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17835,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17835,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17835,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17835,36,'cp 0','d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 h2h4 h7h5 d4e2 d7e6 c1e3 e8c8 f2f3 f8d6 e1f2 g8e7 e3f4 d6b4 a2a3 b4c5 f4e3 c5d6 e3f4');
INSERT INTO evaluations VALUES(17831,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(17831,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(17831,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17831,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17831,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17831,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17831,36,'cp 0','e5d4 d1d4 d8d4 f3d4 c8d7 h2h4 h7h5 d4e2 d7e6 c1e3 e8c8 f2f3 f8d6 e1f2 b7b6 e3f4 d6c5 f4e3');
INSERT INTO evaluations VALUES(55826,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(55826,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55826,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(55826,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55826,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(55826,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(55826,36,'cp 16','e2e3 c7c5 g1f3 g8f6 b1d2 d8b6 a1b1 f8d6 d4c5 b6c5 b2b4 c5c7 f4d6 c7d6 c2c4 e8g8 d1b3 a7a5 a2a3 d6e7 f1e2 a5b4 a3b4 b7b6 e1g1 c8b7 b3b2 b8d7 h2h3 d5c4 d2c4 h7h6 b1d1 b7d5 d1d4 f8c8 f1d1');
INSERT INTO evaluations VALUES(55823,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(55823,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55823,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(55823,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55823,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(55823,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(55823,36,'cp 13','g8f6 g1f3 c7c5 b1d2 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 f6d5 b1c1 c5e7 f1c4 d5b6 c4d3 f8d8 d1c2 h7h6 e1g1 c8d7 h2h3 b8c6 a2a3 a8c8 c2b1 c6b8 c1c8 d8c8 f1c1 d7c6 d3e4 c6e4 c1c8 b6c8 b1e4 c8d6');
INSERT INTO evaluations VALUES(57301,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(57301,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(57301,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(57301,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(57301,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(57301,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(57301,36,'cp -4','b8c6 d4e3 g8f6 c1d2 g7g6 b1c3 f8g7 e1c1 e8g8 f1c4 f8e8 g1f3 d7d6 h1e1 h7h6 h2h3 a7a6 c3d5 e8e4 e3b3 e4c4 b3c4 c8e6 e1e6 f7e6 d5b4 d6d5 b4c6 d5c4 c6d8 a8d8 d1e1 f6d5 e1e6 d8d6 e6e4 b7b5 c2c3 g8f7 d2e3 g6g5 c1c2 d5e3 e4e3 c7c6');
INSERT INTO evaluations VALUES(22933,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(22933,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(22933,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(22933,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(22933,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(22933,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22933,36,'cp -8','d4e3 g8f6 c1d2 g7g6 b1c3 f8g7 g1f3 e8g8 f1c4 f8e8 e1g1 h7h6 a1d1 d7d6 f1e1 a7a6 h2h3 b7b5 c4d5 c8b7 a2a3 f6d5 c3d5 c6d4 f3d4 b7d5 f2f3 f7f5 d2c3 f5e4 d4b5 a6b5 d1d5 e4f3 e3e8 d8e8 e1e8 a8e8 c3g7 g8g7 g1f2 e8e2 f2f3 e2c2 d5b5');
INSERT INTO evaluations VALUES(22954,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22954,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22954,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22954,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(22954,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(22954,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(22954,36,'cp 0','g8f6 c1d2 f8e7 b1c3 d7d5 c3d5 f6d5 e4d5 d8d5 g1e2 c8g4 h2h3 g4e2 f1e2 c6d4 e2d3 e8c8 e1g1 h8e8 d2c3 e7c5 e3g3 d4e2 d3e2 e8e2 a1e1 e2e1 f1e1 d5f5 e1e5 d8d1 g1h2 c5f2 e5f5 f2g1 h2h1 g1f2');
INSERT INTO evaluations VALUES(16709,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16709,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16709,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16709,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(16709,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16709,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16709,36,'cp -11','c1d2 f8e7 b1c3 d7d5 c3d5 f6d5 e4d5 d8d5 g1e2 c8g4 h2h3 g4e2 f1e2 c6d4 e2d3 e8c8 e1g1 h8e8 d2c3 e7f6 e3g3 c8b8 f1e1 a7a5 e1e8 d8e8 a1d1 h7h5 g3f4 e8d8 c3d4 d5d4 f4d4 d8d4 b2b3 h5h4 d1e1 a5a4 d3c4 d4d2 b3a4 d2c2 c4f7');
INSERT INTO evaluations VALUES(16710,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(16710,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(16710,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(16710,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(16710,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16710,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16710,36,'cp -9','f8e7 b1c3 d7d5 c3d5 f6d5 e4d5 d8d5 g1e2 c8f5 e2c3 d5e6 e3e6 f7e6 e1c1 e8c8 f1c4 c6d4 d2f4 a7a6 c4d3 h8f8 f4e3 f5d3 d1d3 d4f5 d3d8 c8d8 e3d2 d8d7 c3e4 b7b6 f2f3 f5h4 h1g1 e6e5 d2e3 h7h5 c2c3 f8f5 g1d1 d7e6');
INSERT INTO evaluations VALUES(50331,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50331,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(50331,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50331,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50331,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(50331,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(50331,36,'cp 0','d7d5 c1f4 e7e6 c3b5 b8a6 e2e3 f8e7 g1f3 e8g8 h2h4 c7c6 b5c3 d8b6 a2a3 c6c5 a1b1 c8d7 f1d3 c5c4 d3e2 a6c7 f3e5 d7e8 g2g4 f6d7 e5d7 e8d7 g4g5 c7e8 e2g4 a8d8 d1f3 e7d6 f4d6 e8d6 f3g2 f7f5 g5f6 f8f6');
INSERT INTO evaluations VALUES(31169,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(31169,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(31169,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(31169,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(31169,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(31169,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(31169,36,'cp 48','f1e2 c6c5 g1f3 e7e6 c1e3 c5d4 f3d4 g8e7 c2c4 a7a6 d1a4 b8c6 b1c3 d5c4 e1c1 d8a5 a4a5 c6a5 d4f5 e7f5 e3b6 a5c6 f2f4 f8b4 c3e4 h7h5 c1b1 a8c8 e2c4 b4a5 e4d6 f5d6 e5d6 e8d7 b6f2 h8h6 f4f5 h6f6 f5e6 f7e6 f2e3 c8f8 c4d3 a5b4 d3e4 h5h4 a2a3 b4d6 e3c5');
INSERT INTO evaluations VALUES(31166,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(31166,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(31166,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(31166,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(31166,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(31166,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(31166,36,'cp 25','c6c5 g1f3 e7e6 e1g1 b8c6 c2c3 c5d4 c3d4 g8e7 b1c3 f5g4 f3h4 g4e2 c3e2 a8c8 g2g3 d8b6 a1b1 c6a5 h4f3 a5c4 b2b3 e7c6 h2h4 c4a3 b1b2 h7h6 h4h5 f8e7 b2d2 e8g8 g1g2 a3b5 c1b2');
INSERT INTO evaluations VALUES(47195,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(47195,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(47195,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(47195,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47195,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47195,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47195,36,'cp 11','g1f3 e7e6 b2b3 f8d6 b1d2 b8d7 c1b2 b7b6 f1d3 c8b7 f3e5 e8g8 d1e2 f6e4 d2e4 d5e4 d3a6 b7a6 e2a6 d6b4 c2c3 b4d6 e5c4 d8c8 a6c8 f8c8 a2a4 f7f6 c4d6 c7d6 c3c4 a7a6 b2a3');
INSERT INTO evaluations VALUES(57589,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(57589,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(57589,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(57589,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(57589,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(57589,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(57589,36,'cp 64','c1g5 h7h6 g5h4 g7g5 e2e3 f8g7 h4g3 c7c5 c2c3 d7d5 f1b5 c8d7 a2a4 c5d4 c3d4 g8f6 h2h4 d7b5 a4b5 e7e6 b1c3 b8d7 g1f3 g5g4 f3e5 f6h5 g3h2 d7e5 h2e5 e8g8 g2g3 g7e5 d4e5 d8c7 d1d4 c7b6 d4b6 a7b6 a1a8 f8a8 e1g1 g8g7 f1c1');
INSERT INTO evaluations VALUES(57612,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(57612,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(57612,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(57612,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(57612,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(57612,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(57612,36,'cp 47','h7h6 g5h4 g7g5 e2e3 f8g7 h4g3 c7c5 c2c3 d7d5 g1f3 d8b6 d4c5 b6b2 b1d2 b8c6 a1b1 b2c3 f1b5 g5g4 f3h4 g8f6 e1g1 e8f7 d1e2 h8f8 h2h3 g4h3 g2h3 c3c5 f1c1 c5a3 d2f3 f6e4 b5c6 b7c6 c1c6 e4g3 f2g3 f7g8 h4g6 f8e8');
INSERT INTO evaluations VALUES(57548,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(57548,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(57548,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(57548,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(57548,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(57548,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(57548,36,'cp 69','g5h4 g7g5 e2e3 f8g7 h4g3 c7c5 c2c3 d7d5 h2h4 g5g4 f1b5 c8d7 a2a4 g8f6 g1e2 d7b5 a4b5 b8d7 e1g1 e8g8 e2f4 d8b6 b1a3 a8c8 d4c5 d7c5 f4d5 f6d5 d1d5 e7e6 d5a2 c5e4 a3c2 c8c5 c2d4 g7d4 c3d4 c5b5 g3e5 g4g3 e5g3 e4g3 f2g3 b5b2 a2a7 b6a7');
INSERT INTO evaluations VALUES(57547,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(57547,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(57547,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(57547,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(57547,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(57547,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(57547,36,'cp 50','g7g5 h4g3 f8g7 e2e3 c7c5 c2c3 d7d5 h2h4 g5g4 g1e2 g8f6 e2f4 e8g8 b1d2 b8c6 f1e2 e7e6 e1g1 a7a6 d1b3 f8e8 a2a4 c8d7 f1d1 b7b6 a1c1 c5c4 b3a2 b6b5 b2b3 c6a5 b3c4 b5c4 f2f3 g4f3 g2f3 g8h7 g1h1 a5c6 d1g1 e8g8');
INSERT INTO evaluations VALUES(57636,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(57636,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(57636,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(57636,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(57636,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(57636,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(57636,36,'cp 64','h4g3 f8g7 e2e3 c7c5 c2c3 d8b6 b1d2 d7d5 d4c5 b6b2 g1f3 b8c6 a1b1 b2c3 f1b5 g5g4 f3h4 g8f6 e1g1 e8g8 h4g6 f8e8 d2b3 g8h7 b1c1 c3b4 b5c6 b7c6 g6e5 c8a6 f1e1 b4a4 b3d4 a4d1 e1d1 f6e4 e5c6 a8c8 c6b4 a6c4 c5c6 e4g3 h2g3');
INSERT INTO evaluations VALUES(57637,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(57637,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(57637,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(57637,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(57637,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(57637,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(57637,36,'cp 49','f8g7 e2e3 c7c5 c2c3 d8b6 b1d2 g8f6 h2h4 d7d5 d1c2 g5g4 g1e2 f6h5 g3f4 e8g8 g2g3 c8d7 f1g2 c5d4 e3d4 d7b5 g2d5 e7e6 d5b3 b5e2 e1e2 h5f4 g3f4 b6d6 a1e1 b8c6 e2d1 a8e8 d1c1 c6a5 b3a4 e8d8 c2d3 g8h8 c1b1 d6f4 a4c2');
INSERT INTO evaluations VALUES(19350,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(19350,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19350,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(19350,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(19350,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(19350,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(19350,36,'cp 37','e7e6 e1g1 g8e7 f1e1 c6d4 f3d4 c5d4 d2d3 a7a6 b5a4 d8c7 b1d2 b7b5 a4b3 c8b7 a2a4 e7g6 a4b5 a6b5 a1a8 b7a8 d2f3 f8c5 h2h4 h7h5 c1d2 a8c6 d1e2 f7f6 e1a1 g6f4 e2e1 e8g8 g2g3 f4g6 d2b4 c6b7 b4c5');
INSERT INTO evaluations VALUES(18695,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(18695,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(18695,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18695,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(18695,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(18695,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18695,36,'cp 45','e1g1 g8e7 f1e1 c6d4 f3d4 c5d4 c2c3 e7c6 d2d3 f8c5 b1d2 e8g8 d2b3 c5b6 b5c6 d7c6 b3d4 b6d4 c3d4 d8d4 d1c2 d4f6 c1e3 e6e5 c2c3 f8d8 a2a4 h7h5 a1d1 h5h4 d3d4 e5d4 e3d4 f6g6 h2h3 c8e6 a4a5 f7f6 d4c5 d8d1 e1d1');
INSERT INTO evaluations VALUES(18693,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(18693,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(18693,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(18693,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18693,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(18693,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(18693,36,'cp 31','g8e7 f1e1 c6d4 f3d4 c5d4 c2c3 a7a6 b5f1 e7c6 b2b4 f8d6 c1b2 d4c3 d2c3 d8c7 g2g3 b7b6 b1d2 e8g8 a2a4 d6e7 d2c4 d7d6 a4a5 b6a5 c4a5 c6a5 a1a5 c8b7 d1d4 e7f6 d4d2 h7h6 h2h4 f6e5 f2f4 e5f6 g1h2 e6e5 f4f5 a8c8 c3c4 f8d8 e1e3 c8b8 d2f2 d8c8 b2c3');
INSERT INTO evaluations VALUES(12676,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(12676,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(12676,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(12676,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(12676,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(12676,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(12676,36,'cp 39','f1e1 a7a6 b5f1 b7b6 d2d4 c6d4 f3d4 c5d4 d1d4 e7c6 d4d1 d8c7 c1e3 c8b7 b1d2 c6e5 d1e2 f8c5 e3c5 c7c5 c2c4 e8g8 d2b3 c5c7 a1c1 f7f5 e4f5 f8f5 b3d4 f5g5 f2f4 e5g4 g2g3 g5g6 f1g2 a8e8 g2b7 c7b7 c1d1 g4f6 d4f3 e6e5');
INSERT INTO evaluations VALUES(50385,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(50385,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(50385,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(50385,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50385,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50385,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(50385,36,'cp 27','d2d4 d7d5 c2c4 e7e6 b1c3 f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 c7c6 f3d2 f6e4 h4d8 e4c3 b2c3 f8d8 f2f3 b7b5 a2a4 d8b8 a4b5 c6b5 e3e4 f5e6 e1f2 a7a6 h1b1 d7b6 e2d3 b6a4 e4d5 e6d5 c3c4 b5c4 b1b8 a8b8 d3c4 d5c4 d2c4');
INSERT INTO evaluations VALUES(49532,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(49532,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49532,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(49532,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49532,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49532,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(49532,36,'cp -13','c7c6 c4b3 d7d5 g1f3 a7a5 a2a4 f8b4 c2c3 b4d6 e1g1 e8g8 f1e1 d5e4 d3e4 d8c7 h2h3 b8d7 b1a3 h7h6 a3c4 d6e7 d1e2 f8e8 b3c2 b7b6 c1d2 c8a6 b2b4 a5b4 c3b4 b6b5 a4b5 a6b5 a1b1 c6c5 b4c5 b5a6 d2b4 d7c5');
INSERT INTO evaluations VALUES(44080,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44080,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44080,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44080,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44080,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44080,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44080,36,'cp 0','g1f3 d7d5 c4b3 a7a5 a2a4 f8b4 c2c3 b4d6 e1g1 e8g8 f1e1 d5e4 d3e4 d8c7 b1a3 b8d7 c1g5 f8e8 g5h4 d6f8 f3g5 e8e7 g5f3');
INSERT INTO evaluations VALUES(44077,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44077,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44077,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44077,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44077,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44077,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44077,36,'cp 0','d7d5 c4b3 a7a5 a2a4 f8b4 c2c3 b4d6 e1g1 e8g8 f1e1 d5e4 d3e4 d8c7 c1g5 b8d7 b1a3 f8e8 g5h4 h7h6 a3c4 d6c5 f3d2 c5e7 d2f3');
INSERT INTO evaluations VALUES(45804,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45804,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45804,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45804,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45804,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45804,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45804,36,'cp 0','c4b3 a7a5 a2a4 f8b4 c2c3 b4d6 e1g1 e8g8 f1e1 d5e4 d3e4 d8c7 b1a3 b8d7 c1g5 f8e8 g5h4 h7h6 a3c4 d6c5 f3d2 c5e7 d2f3');
INSERT INTO evaluations VALUES(52520,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(52520,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(52520,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52520,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(52520,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(52520,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(52520,36,'cp -14','b8c6 d4d1 g8f6 b1c3 e7e6 f1d3 f8e7 g1f3 d7d6 e1g1 a7a6 a2a4 e8g8 h2h3 d8c7 f1e1 c8d7 c1e3 f8e8 d1d2 a8c8 e1d1 e8d8 f3d4 c6e5 d2e2 d8e8 d4f3 e5c4 d3c4 c7c4 e2c4 c8c4 e4e5');
INSERT INTO evaluations VALUES(19552,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(19552,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(19552,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(19552,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(19552,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(19552,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19552,36,'cp 0','d4a4 e7e6 f1d3 f8c5 g1f3 g8e7 e1g1 e8g8 b1d2 e7g6 c2c3 d8c7 a4d1 a7a6 a2a4 f8d8 d2b3 c5b6 d1e2 d7d6 c1e3 c8d7 f1d1 h7h6 b3d2 c6e5 f3e5 d6e5 a4a5 b6e3 e2e3 d7c6 d2c4 c6b5 c4b6');
INSERT INTO evaluations VALUES(19575,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19575,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19575,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(19575,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19575,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(19575,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(19575,36,'cp -11','e7e6 g1f3 f8c5 f1d3 g8e7 e1g1 e7g6 b1d2 e8g8 c2c3 d8c7 a4d1 a7a6 d2b3 c5a7 b3d4 d7d6 d4c6 b7c6 h2h4 e6e5 h4h5 g6e7 h5h6 e7g6 h6g7 g8g7 f3h2 f7f5 e4f5 c8f5 d3f5 f8f5 c1e3 a8f8 e3a7 c7a7 d1d6 f5f2 g1h1');
INSERT INTO evaluations VALUES(18942,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18942,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18942,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18942,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(18942,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18942,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18942,36,'cp 0','f1d3 f8c5 g1f3 g8e7 e1g1 e7g6 b1d2 e8g8 c2c3 d8c7 a4d1 f8d8 d2b3 c5b6 b3d4 d7d6 c1g5 c6e7 f1e1 h7h6 g5e3 c8d7 d3c2 g6e5 f3d2 e5g6');
INSERT INTO evaluations VALUES(21951,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(21951,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(21951,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(21951,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(21951,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(21951,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(21951,36,'cp 93','d4d5 c6b8 c1e3 g8f6 f1d3 g7g6 c2c4 f8g7 b1c3 e8g8 e1g1 f6g4 e3d2 f7f5 b2b4 g4f6 d1c2 f6e4 c3e4 f5e4 d3e4 b8d7 f3g5 d7f6 f1e1 f6e4 g5e4 c8f5 f2f3 d8d7 h2h3 b7b6 c2d3 f5e4 e1e4 a7a5 b4a5 b6a5');
INSERT INTO evaluations VALUES(21875,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(21875,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(21875,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(21875,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(21875,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(21875,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(21875,36,'cp 97','c6b8 c1e3 f8e7 f1d3 c8g4 b1d2 g8f6 h2h3 g4h5 b2b4 a7a5 a2a3 c7c6 c2c4 b8d7 e1g1 e8g8 d1c2 h5g6 f3h4 f6h5 h4g6 h7g6 d5c6 b7c6 d2b3 e7g5 b3a5 g5e3 f2e3 d8g5 f1f3 c6c5 g2g4 a8a6 c2e2 c5b4 a3b4 f8c8');
INSERT INTO evaluations VALUES(55500,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(55500,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(55500,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(55500,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(55500,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(55500,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(55500,36,'cp 87','c1e3 g8f6 f1d3 g7g6 c2c4 f8g7 b1c3 e8g8 a2a3 a7a5 h2h3 b8d7 f3d2 d7c5 d3e2 f6d7 g2g4 g7f6 b2b4 a5b4 a3b4 a8a1 d1a1 c5a6 a1b1 f6g5 d2f1 c7c5 e3g5 d8g5 h3h4 g5e7 b4b5 a6b4 f1e3 d7b6 g4g5 f7f5 e4f5');
INSERT INTO evaluations VALUES(19438,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(19438,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(19438,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(19438,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(19438,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(19438,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(19438,36,'cp 39','g1f3 e7e5 f1c4 d7d6 d2d3 h7h6 h2h3 g7g6 f3h2 f8g7 e1g1 g8e7 h2g4 c8e6 c4e6 f7e6 a1b1 g6g5 a2a3 e8g8 b2b4 b7b6 g4h2 e7g6 b4c5 b6c5 g2g3 c6d4 c3b5 d8d7 b5d4 c5d4 c1d2 d6d5 c2c3 d5e4 d1h5 d7f7 d3e4 a8d8 c3d4');
INSERT INTO evaluations VALUES(55304,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(55304,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(55304,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(55304,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(55304,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(55304,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(55304,36,'cp 69','g1f3 g8f6 d2d4 e7e6 f1c4 a7a6 a2a4 f8d6 e1g1 h7h6 f1e1 e8g8 f3e5 c8d7 d1e2 b8c6 e1d1 c6b4 c1f4 b4d5 f4d2 d8e7 a4a5 d5c3 d2c3 a8d8 c4d3 d7e8 d1e1 c7c5 d4c5 d6c5');
INSERT INTO evaluations VALUES(55297,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(55297,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(55297,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(55297,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(55297,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(55297,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(55297,36,'cp 66','g8f6 d2d4 e7e6 f1c4 a7a6 a2a4 b8c6 e1g1 f8e7 f1e1 e8g8 c1f4 c6b4 d1e2 b7b6 a1d1 c8b7 c4b3 e7d6 f4g5 g8h8 h2h3 a8b8 g5h4 b8a8 e2e3 a8c8 f3e5 c8b8 e3e2 d8e8 a4a5 b6a5 c3a4 d6e5 d4e5 e8c6');
INSERT INTO evaluations VALUES(47415,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(47415,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(47415,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(47415,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(47415,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(47415,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(47415,36,'cp 71','d2d4 e7e6 f1c4 a7a6 a2a4 f8e7 e1g1 b8c6 c1f4 e8g8 f1e1 c6b4 d1e2 b7b6 a1d1 c8b7 c4b3 b4d5 c3d5 f6d5 f4g3 e7d6 f3e5 d8g5 a4a5 b6a5 c2c3 d5f6 d1a1 f6e4 a1a5 c7c5 e5d7 d6g3 h2g3 g5d8 d7f8 d8a5 f8d7 a8d8 b3c2 d8d7');
INSERT INTO evaluations VALUES(39606,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(39606,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(39606,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(39606,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(39606,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(39606,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(39606,36,'cp 56','d7d6 f1d3 b8c6 c2c3 e7e5 h2h3 g8e7 e1g1 e8g8 d4e5 c6e5 f3e5 d6e5 d1e2 e7c6 f1d1 d8e7 a2a4 c6d8 b1d2 d8e6 d2f3 c8d7 d3c4 f8d8 c1e3 e6f4 e3f4 e5f4 e4e5 d7c6 d1e1 a7a6 e5e6 f7e6 a1d1 d8d1 c4e6 g8h8 e1d1');
INSERT INTO evaluations VALUES(38423,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(38423,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(38423,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(38423,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38423,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(38423,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(38423,36,'cp 51','f1d3 b8c6 c2c3 e7e5 h2h3 g8e7 e1g1 e8g8 d4e5 c6e5 f3e5 d6e5 d1e2 e7c6 a2a4 d8e7 c1e3 c6d8 a4a5 d8e6 b1d2 e6f4 e3f4 e5f4 d2f3 c8e6 a5a6 b7b6 d3c4 f8e8 f1e1 e7c5 c4e6 e8e6 a1d1 a8e8 e4e5 g7e5');
INSERT INTO evaluations VALUES(38415,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(38415,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(38415,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(38415,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(38415,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(38415,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(38415,36,'cp 49','b8c6 c2c3 e7e5 h2h3 g8e7 e1g1 e8g8 d4e5 c6e5 f3e5 d6e5 d1e2 e7c6 a2a4 d8e7 b1d2 c8e6 a4a5 a7a6 d2f3 h7h6 b2b4 f8d8 c1e3 f7f5 f1d1 e7f7 e3c5 g8h7 d3c2 e6c4 e2e1 f5f4 g1h2 g6g5 d1d8 a8d8 f3d2 c4e6 f2f3 f7d7');
INSERT INTO evaluations VALUES(57757,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(57757,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(57757,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(57757,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(57757,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(57757,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57757,36,'cp 0','d7d5 d2d4 g8f6 c1f4 e7e6 c3b5 f8b4 c2c3 b4a5 g1f3 e8g8 e2e3 a7a6 b5a3 b8d7 a3b1 f6h5 f4g5 f7f6 g5h4 g7g5 f3d2 h5g7 h4g3 f6f5 d2f3 g7h5 b1d2 h5g3 h2g3 g5g4 f3g1 c7c6 f1d3 d8e7 g1e2 d7f6 d1c2 c8d7 e1c1 a5c7 c3c4');
INSERT INTO evaluations VALUES(55115,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55115,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(55115,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(55115,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(55115,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55115,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55115,36,'cp 0','d2d4 g8f6 c1f4 e7e6 c3b5 f8b4 c2c3 b4a5 e2e3 a7a6 b5a3 e8g8 g1f3 b8d7 f1e2 c7c6 e1g1 d8e7 d1c2 f6h5 f3g5 h5f6');
INSERT INTO evaluations VALUES(35127,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(35127,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(35127,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(35127,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(35127,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(35127,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(35127,36,'cp 144','b1c3 b8c6 d2d4 f8b4 c1d2 c6d4 f3d4 e5d4 c3b5 b4d2 d1d2 f6b6 e1c1 g8e7 d2f4 d7d6 e4e5 e8g8 e5d6 c7d6 f4d4 d6d5 d4b6 a7b6 a2a3 f8d8 d1d4 g7g6 f1e2 g8g7 h1d1 c8e6 b5c3 e7c6 d4d2 a8a5 e2f3 c6e7 g2g3 h7h5 c3e2 g7f6 f3g2 h5h4 e2c3 h4g3 h2g3 a5c5');
INSERT INTO evaluations VALUES(35117,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(35117,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(35117,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(35117,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(35117,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(35117,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(35117,36,'cp 132','b8c6 d2d4 f8b4 c1d2 c6d4 f3d4 e5d4 c3b5 b4d2 d1d2 f6b6 e1c1 g8e7 d2f4 d7d6 e4e5 e8g8 e5d6 c7d6 f4d4 d6d5 d4b6 a7b6 a2a3 f8d8 f1e2 c8d7 b5d4 d7c6 h1e1 g8f8 e2f3 d8d6 e1e3 e7g6 g2g3 a8e8 e3e8 f8e8 d1e1 e8f8 f3g4 c6d7 g4d7 d6d7 c2c3 g6e7 e1d1 e7c6 d1d2');
INSERT INTO evaluations VALUES(1868,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(1868,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(1868,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(1868,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(1868,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(1868,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(1868,36,'cp 146','d2d4 f8b4 c1d2 c6d4 f3d4 e5d4 c3b5 b4d2 d1d2 f6b6 e1c1 g8e7 d2f4 d7d6 e4e5 e8g8 e5d6 c7d6 f4d4 d6d5 d4b6 a7b6 a2a3 c8d7 f1e2 f8d8 d1d2 d7b5 e2b5 d8d6 h1d1 g8f8 b5c4 d6f6 c4b3 b6b5 d2e2 f6d6 g2g3 h7h5 e2d2 d6f6 b3d5 e7d5 d2d5');
INSERT INTO evaluations VALUES(52456,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(52456,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52456,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(52456,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(52456,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(52456,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(52456,36,'cp -23','e7e6 g1f3 d7d5 b1c3 b8c6 f1b5 g8e7 e4d5 e7d5 b5c6 b7c6 f3e5 f8d6 e1g1 c8a6 d2d3 d5c3 b2c3 c5c4 d1f3 e8g8 f3c6 d8b6 c6b6 a7b6 c1e3 d6e5 f4e5 c4d3 c2d3 a6d3 f1d1 d3c2 d1d2 c2a4 e3b6 f8c8 b6d4 a4c6 h2h4 h7h5 a2a4 c6a4 d2e2 a4c6 a1a8');
INSERT INTO evaluations VALUES(52013,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(52013,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(52013,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52013,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(52013,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(52013,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(52013,36,'cp -10','b1c3 b8c6 g1f3 d7d5 f1b5 g8e7 e4d5 e7d5 b5c6 b7c6 f3e5 f8d6 e1g1 c8a6 d2d3 d5c3 b2c3 c5c4 d1f3 e8g8 f3c6 d8b6 c6b6 a7b6 c1e3 d6e5 f4e5 c4d3 c2d3 a6d3 f1d1 d3c4 a2a4 c4b3 d1d6 b3a4 e3b6 f8c8 d6d2 h7h5 d2d4 a4c6 a1a8 c8a8 d4d8 g8h7 d8a8 c6a8 c3c4 g7g5 b6c7 h7g6 c7d6');
INSERT INTO evaluations VALUES(52009,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(52009,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(52009,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(52009,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52009,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52009,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(52009,36,'cp -15','b8c6 g1f3 d7d5 f1b5 g8e7 d2d4 c5d4 f3d4 d5e4 c3e4 d8d5 d1d3 c8d7 d4c6 e7c6 a2a3 f7f5 e4f2 d5d3 f2d3 a7a6 b5a4 b7b5 a4b3 a8c8 c1e3 c6a5 e1g1 f8d6 a1c1 e8f7 f1e1 a5b3 c2b3 d7c6 e3c5 d6c7 c5d4 c6e4 d3e5 c7e5 d4e5 h8d8 b3b4 c8c1');
INSERT INTO evaluations VALUES(18850,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(18850,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(18850,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(18850,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(18850,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(18850,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(18850,36,'cp -10','g1f3 d7d5 f1b5 g8e7 e4d5 e7d5 b5c6 b7c6 f3e5 f8d6 e1g1 c8a6 d2d3 d5c3 b2c3 c5c4 d1f3 a8c8 e5c4 a6c4 d3c4 e8g8 c1e3 d8a5 f3e4 d6c5 e3d4 f8d8 h2h3 c5d4 c3d4 a5c3 a1d1 c3c4 f4f5 e6f5 e4f5 c4d5 f5f2 h7h6 c2c3 d8e8 d1e1 c8d8 e1e8 d8e8 f1e1 g8f8 a2a4 d5c4 e1e8 f8e8 f2e3 e8d7 e3f2');
INSERT INTO evaluations VALUES(18849,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(18849,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(18849,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(18849,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(18849,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(18849,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(18849,36,'cp -20','d7d5 f1b5 g8e7 d2d4 a7a6 b5c6 e7c6 c1e3 d5e4 c3e4 c5d4 f3d4 c8d7 d4f3 d8c7 d1d2 c6b4 e4c3 d7c6 f3e5 f8e7 e1c1 e8g8 e5c6 c7c6 d2f2 a8d8 a2a3 b4d5 c3d5 d8d5 d1d5 c6d5 h1d1 d5c6 e3b6 f8c8 c1b1 f7f6 b6d4 g8f7 f2e2');
INSERT INTO evaluations VALUES(20304,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(20304,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(20304,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(20304,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(20304,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(20304,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(20304,36,'cp -12','f1b5 g8e7 d2d4 a7a6 b5c6 e7c6 e4d5 e6d5 d4c5 c8g4 c3d5 g4e6 d5b6 d8d1 e1d1 a8d8 d1e2 f8c5 c1e3 c6d4 e2f2 c5b6 h1d1 d8c8 f3d4 b6d4 d1d4 c8c2 d4d2 c2d2 e3d2 e8g8 a2a4 f8d8 d2e3 f7f6 a1c1 d8d3 c1c3 d3c3 b2c3 e6d7 a4a5 h7h5 h2h3 g8f7 e3c5 f7g6 g2g4 h5g4');
INSERT INTO evaluations VALUES(20188,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(20188,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(20188,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(20188,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(20188,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(20188,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(20188,36,'cp -9','g8e7 e4d5 e7d5 b5c6 b7c6 f3e5 f8d6 e1g1 e8g8 d2d3 d5c3 b2c3 c8a6 d1f3 c5c4 f3c6 d8b6 c6b6 a7b6 c1e3 d6e5 f4e5 c4d3 c2d3 a6d3 f1d1 d3c2 d1d2 c2a4 e3b6 f8c8 b6d4 a4c6 a2a3 h7h6 d2b2 a8a5 b2b4 c8a8 a3a4 h6h5 c3c4 g8h7 h2h3 a5a4 a1a4 c6a4 c4c5 a4c6');
INSERT INTO evaluations VALUES(12942,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(12942,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12942,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(12942,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(12942,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(12942,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(12942,36,'cp -6','e4d5 e7d5 b5c6 b7c6 f3e5 f8d6 e1g1 e8g8 d2d3 d5c3 b2c3 c8a6 d1f3 c5c4 f3c6 d8b6 c6b6 a7b6 c1e3 d6e5 f4e5 c4d3 c2d3 f8c8 e3b6 a6d3 f1d1 d3c2 d1d2 c2a4 b6d4 a4c6 a2a4 a8a4 a1a4 c6a4 d4b6 h7h5 d2d8 c8d8 b6d8 g8h7 d8e7 h7g6 e7f8 f7f5 e5f6 g7f6 c3c4 g6f5 f8e7 e6e5 g1f2 a4d7');
INSERT INTO evaluations VALUES(52426,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52426,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(52426,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52426,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52426,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52426,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52426,36,'cp 20','b1c3 b8c6 g1f3 e7e5 g2g3 g7g6 f1g2 f8g7 e1g1 g8e7 a1b1 d7d6 a2a3 a7a5 d2d3 e8g8 f3e1 c8e6 c1g5 f7f6 g5e3 f6f5 e3g5 d8d7 e1c2 d6d5 c3a4 d7d6 a4b6 a8a6 c4d5 e7d5 b6d5 e6d5 g2d5 d6d5 b2b4 c5b4 a3b4 a5b4 c2b4 c6b4 b1b4 b7b5 d1b1');
INSERT INTO evaluations VALUES(52421,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(52421,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(52421,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52421,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(52421,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52421,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52421,36,'cp 14','b8c6 g1f3 e7e5 e2e3 g8f6 f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e2c4 b7b6 e3e4 c6a5 c4d5 c8b7 c1e3 f8e7 b2b3 e8c8 g1f1 c8c7 d5b7 d8d1 a1d1 a5b7 f1e2 h8d8 d1d8 c7d8 f3d2 b7d6 f2f3 d6b5 e2d3 b5c7');
INSERT INTO evaluations VALUES(19401,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(19401,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19401,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19401,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19401,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(19401,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19401,36,'cp 19','g1f3 e7e5 e2e3 g8f6 f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e2c4 b7b6 c4d5 c8b7 e3e4 c6a5 g1f1 e8c8 d5b7 a5b7 c1e3 d8d1 a1d1 f8e7 f3d2 c8c7 a2a4 c7c6 f1e2 a7a6 d1a1 h8d8 b2b3 b6b5 g2g3');
INSERT INTO evaluations VALUES(19399,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(19399,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(19399,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19399,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(19399,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(19399,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(19399,36,'cp 19','e7e5 e2e3 g8f6 f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e2c4 b7b6 c4d5 c8b7 e3e4 c6a5 g1f1 f8e7 f1e2 b7d5 d1d5 e8f7 b2b3 h8d8 c1e3 f7e6 a1d1 c5c4 f3d2 c4b3 d5d8 a8d8 a2b3 d8c8 e2d3 c8d8 d3c2 d8c8 g2g3 e7b4');
INSERT INTO evaluations VALUES(18987,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(18987,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(18987,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(18987,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(18987,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(18987,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18987,36,'cp 17','g2g3 g7g6 a2a3 f8g7 d2d3 g8e7 f1g2 e8g8 e1g1 d7d6 a1b1 a7a5 c1g5 f7f6 g5e3 c8e6 f3e1 f6f5 e3g5 d8d7 e1c2 d6d5 c3a4 d7d6 a4b6 a8a6 b6d5 e7d5 c4d5 e6d5 g2d5 d6d5 b2b4 c5b4 a3b4 a5b4 c2b4 c6b4 b1b4 b7b5 e2e4 f5e4 d3e4 d5d1 f1d1 a6a2 d1d2 a2d2');
INSERT INTO evaluations VALUES(18988,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(18988,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(18988,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(18988,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18988,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(18988,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(18988,36,'cp 20','g7g6 d2d3 f8g7 a2a3 g8e7 f1g2 e8g8 e1g1 d7d6 a1b1 a7a5 f3e1 c8e6 c1g5 f7f6 g5d2 f6f5 c3d5 e6f7 b2b4 c5b4 a3b4 e7d5 c4d5 c6b4 d2b4 a5b4 b1b4 d8d7 d1b3 f8b8 b4b6 d7a4 b3b1 a4a1 e1c2 a1b1 f1b1 a8a2 c2e3 a2e2 e3c4');
INSERT INTO evaluations VALUES(6479,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(6479,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(6479,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(6479,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(6479,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(6479,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(6479,36,'cp 65','f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 f1d3 g8f6 e1g1 e8g8 c1g5 d8e8 d1e1 h7h6 g5d2 d7d5 e4d5 c8d7 g1h1 c7c6 d5c6 d7c6 f2f3 e8b8 e1h4 b7b5 a2a3 a7a5 f1e1 f8e8 a1d1 b5b4 a3b4 a5b4 c3e2 e5b2 h4b4 b2e5 b4c5 c6d7 e2d4 a8a2 f3f4 e5d4 e1e8 b8e8 c5d4');
INSERT INTO evaluations VALUES(6400,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(6400,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(6400,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(6400,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(6400,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(6400,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(6400,36,'cp 65','c6e5 d2d4 c5d6 d4e5 d6e5 f1d3 g8f6 e1g1 d7d6 c3e2 f6d7 c2c3 b7b6 c1e3 c8b7 f2f3 e5f6 e2f4 e8g8 d1d2 f6g5 b2b3 h7h6 c3c4 d7e5 d3c2 a7a5 f4d5 b7d5 c4d5 g5e3 d2e3 a5a4 b3a4 d8f6 f3f4 e5c4 e3d3 c4b2 d3e2 f6d4 g1h1 b2a4');
INSERT INTO evaluations VALUES(6844,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(6844,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(6844,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(6844,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(6844,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(6844,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(6844,36,'cp 59','d2d4 c5d6 d4e5 d6e5 f1d3 g8f6 e1g1 e8g8 c1g5 d8e8 f1e1 d7d6 d1d2 h7h6 g5h4 g7g5 h4g3 c7c6 c3e2 f6h5 g3e5 d6e5 d2c3 e8e7 a1d1 c8g4 f2f3 g4d7 b2b4 a8d8 d3c4 e7f6 a2a4 d7e6 c4e6 f6e6 a4a5 h5f4 e2f4 g5f4');
INSERT INTO evaluations VALUES(6843,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(6843,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(6843,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(6843,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(6843,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(6843,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(6843,36,'cp 53','c5d6 d4e5 d6e5 f1d3 g8f6 e1g1 e8g8 c1g5 d8e8 d1e1 h7h6 g5h4 f6h5 e1d2 g7g5 h4g3 d7d6 c3d5 h5g3 h2g3 e8d8 c2c3 c7c6 d5e3 g5g4 d3e2 h6h5 e3c4 g8g7 a1d1 d8e7 c4d6 f8d8 d6c8 d8d2 c8e7 d2e2 f2f4 g4f3 e7f5 g7f8 g2f3 e2b2 d1d7 b2a2 d7b7 e5c3');
INSERT INTO evaluations VALUES(6799,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(6799,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(6799,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(6799,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(6799,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(6799,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(6799,36,'cp 53','d4e5 d6e5 f1d3 g8f6 e1g1 e8g8 c1g5 d8e8 f1e1 d7d6 d1d2 h7h6 g5h4 g7g5 h4g3 f6h5 g3e5 e8e5 a2a4 c7c6 a1d1 c8g4 f2f3 g4e6 b2b3 f8e8 c3e2 a8d8 d2b4 b7b6 d3c4 h5f4 e2f4 e5f4 c4e6 e8e6 a4a5 b6b5 b4c3 c6c5 g2g3');
INSERT INTO evaluations VALUES(6798,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(6798,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(6798,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(6798,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(6798,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(6798,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(6798,36,'cp 47','d6e5 f1d3 g8f6 e1g1 e8g8 c1g5 d8e8 f1e1 d7d6 d1d2 f6g4 f2f3 f7f6 g5f4 e5f4 d2f4 g4e5 f4d2 e8g6 d3e2 c8h3 g2g3 g8h8 c3d5 g6f7 d2c3 g7g5 a1d1 h3e6 f3f4 g5f4 d5f4 e6g4 e2g4 e5g4 c3b3 a8e8 a2a4 f7b3 c2b3 g4e5');
INSERT INTO evaluations VALUES(6883,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(6883,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(6883,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(6883,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(6883,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(6883,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(6883,36,'cp 68','f1d3 g8f6 e1g1 e8g8 c1g5 d8e8 f1e1 d7d6 d1d2 f6g4 f2f3 f7f6 g5f4 e5f4 d2f4 g4e5 f4d2 c8e6 d3e2 e8c6 a1d1 c6c5 d2d4 c5d4 d1d4 g7g5 g1f2 g8f7 g2g3 e5g6 f2e3 f7e7 c3b5 e7d7 f3f4 g5f4 g3f4 f6f5 e4e5 f8g8 e5d6 c7c5');
INSERT INTO evaluations VALUES(20335,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(20335,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(20335,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(20335,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(20335,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(20335,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(20335,36,'cp 24','g1f3 c8d7 a2a3 c5c4 g2g3 g8e7 b1d2 c6a5 a1b1 h7h6 f1g2 d8b6 h2h4 e8c8 e1g1 e7f5 f1e1 f8e7 f3h2 a5b3 d2b3 d7a4 d1h5 g7g6 h5e2 b6b3 h2g4 b3c2 c1d2 g6g5 h4h5 c8b8 g1h2 b8a8 g4e3 c2d3 e2d3 c4d3 e3f5 e6f5');
INSERT INTO evaluations VALUES(20332,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(20332,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(20332,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(20332,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(20332,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20332,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(20332,36,'cp 29','c8d7 f1e2 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 e2d3 f5e3 f2e3 d7b5 a4c3 b5d3 d1d3 a7b6 c1c2 e8g8 f1c1 a5c6 a2a3 a8c8 c3a4 b6a5 d3b3 b7b5 a4c5 e7c5 c2c5 c6e7 h2h3 a5b6 b3c2 c8c5 c2c5');
INSERT INTO evaluations VALUES(50346,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50346,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50346,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50346,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(50346,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(50346,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50346,36,'cp 0','e7e5 g1f3 b8c6 d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 f8e8 c1g5 h7h6 g5h4 b4d6 a1e1 c8e6 a2a3 a7a5 e1e2 d6e7 h4g3 e7d6');
INSERT INTO evaluations VALUES(55223,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55223,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(55223,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55223,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55223,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(55223,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55223,36,'cp 0','a7a6 e2e3 b7b5 a2a4 b5b4 f1c4 e7e6 c3e2 g8f6 g1f3 c8b7 e1g1 b8d7 e2g3 f8d6 d1e2 e8g8 e3e4 d6g3 h2g3 f6e4 c1f4 h7h6 a4a5 d7f6 f1d1 f6g4 d1f1');
INSERT INTO evaluations VALUES(50855,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50855,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50855,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50855,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50855,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50855,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50855,36,'cp 0','e2e3 b7b5 a2a4 b5b4 f1c4 e7e6 c3e2 c8b7 g1f3 g8f6 e1g1 f8d6 e2g3 b8d7 d1e2 e8g8 e3e4 d6g3 h2g3 f6e4 c1f4 h7h6 a4a5 d7f6 f1d1 f6g4 d1f1');
INSERT INTO evaluations VALUES(50853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50853,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(50853,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(50853,36,'cp -6','b7b5 a2a4 b5b4 f1c4 e7e6 c3e2 g8f6 g1f3 c8b7 e1g1 b8d7 e2g3 c7c5 d1e2 f8d6 e3e4 c5d4 f3d4 e8g8 f1d1 d8b8 a4a5 d7e5 c4b3 f8d8 c1e3 f6e4 g3e4 b7e4 d4e6 f7e6 e3b6 b8b7 b3e6 g8h8 b6d8 a8d8 d1d4');
INSERT INTO evaluations VALUES(51118,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51118,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51118,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51118,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51118,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51118,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51118,36,'cp 0','a2a4 b5b4 f1c4 e7e6 c3e2 c8b7 g1f3 g8f6 e1g1 f8d6 e2g3 b8d7 d1e2 c7c5 f1d1 e8g8 e3e4 c5d4 f3d4 d8b8 a4a5 d7e5 c1g5 e5c4 g5f6 c4e5 g3h5 g7f6 e2e3 e5g4 h5f6 g4f6 e3g5 g8h8 g5f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(51123,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51123,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51123,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51123,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51123,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51123,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51123,36,'cp 0','b5b4 f1c4 c8b7 g1f3 e7e6 c3e2 g8f6 b2b3 c7c5 c1b2 b8d7 e1g1 f8d6 e2g3 e8g8 d4c5 d7c5 b2f6 g7f6 d1d4 d8e7 f1d1 a8d8 d4h4 g8h8 a1c1 f8g8 h4h5 g8g6 d1d4 f6f5 c1d1 g6g8 d4h4 g8g7 h4d4 g7g8');
INSERT INTO evaluations VALUES(21908,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(21908,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(21908,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(21908,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(21908,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(21908,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(21908,36,'cp 50','e1g1 f8e7 d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 c1f4 c6d4 d1d4 c8e6 a1d1 f6h5 f4e3 e7f6 d4d3 e6c4 d3c4 f8e8 f2f3 f6e5 e3d4 e5d4 c4d4 h5f4 g1h1 a7a6 d4b4 a6a5 b4d4 f4e6 d4d2 a5a4 c3d5 h7h6 f1e1 d8d7 c2c4 b7b5 d2c3 g8h8');
INSERT INTO evaluations VALUES(21906,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(21906,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(21906,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(21906,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(21906,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(21906,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(21906,36,'cp 50','f8e7 d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 a4a5 d7c6 c4d3 f8e8 h2h3 a8d8 b2b4 f6h5 b4b5 c6d7 f4h2 a6b5 d3b5 c7c6 b5e2 h5f6 f2f4 g7g6 f4f5 g6f5 d4e3 d6d5 e4f5 e6f5');
INSERT INTO evaluations VALUES(5640,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(5640,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(5640,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(5640,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(5640,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5640,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(5640,36,'cp 55','d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 a2a4 c6d4 d1d4 c8e6 f1d1 f6g4 c3d5 e7h4 g2g3 c7c6 d5f4 e6c4 d4c4 g4e5 c4b3 d8d7 f2f3 h4g5 g1g2 f8e8 a4a5 a8b8 h2h4 g5d8 f4d3 d8f6 c1e3 a7a6 e3b6 e5d3 d1d3 d6d5 a1d1 d7e7 b6f2 h7h6');
INSERT INTO evaluations VALUES(5646,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(5646,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(5646,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(5646,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(5646,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5646,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(5646,36,'cp 50','e5d4 f3d4 c6d4 d1d4 g8f6 b1c3 e8g8 a2a4 c8e6 f1d1 d8c8 c4d3 f8d8 h2h3 c7c6 c1f4 d6d5 e4d5 f6d5 d4e4 d5f6 e4f3 f6d5 c3d5 e6d5 f3g3 e7f6 c2c3 d8e8 a4a5 c8e6 f4e3 e6e5 g3e5 f6e5 d1d2 e5d6 a1d1');
INSERT INTO evaluations VALUES(5661,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(5661,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(5661,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(5661,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(5661,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(5661,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(5661,36,'cp 37','f3d4 g8f6 b1c3 e8g8 c1f4 c6e5 c4b3 c8g4 f2f3 g4d7 a2a4 a7a6 f1e1 c7c5 d4f5 d7f5 e4f5 d8c8 c3d5 f6d5 b3d5 c8f5 f4g3 a8b8 d5e4 f5f6 a4a5 f8e8 c2c3 h7h5 h2h4 c5c4 d1d5 e5d3 e4d3 c4d3 d5d3 b8d8');
INSERT INTO evaluations VALUES(5653,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(5653,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(5653,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(5653,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(5653,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(5653,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(5653,36,'cp 44','g8f6 b1c3 e8g8 a2a4 f8e8 h2h3 h7h6 d4c6 b7c6 d1f3 e7f8 c1e3 c8e6 c4d3 a7a5 f1d1 e6c8 d1e1 a8b8 b2b3 f6d7 d3f1 d8f6 f3f6 d7f6 e3d2 b8a8 a1d1 g7g5 f2f3 f6d7 c3e2 d7e5 f3f4 g5f4 e2f4 f8g7');
INSERT INTO evaluations VALUES(8643,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(8643,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(8643,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(8643,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(8643,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(8643,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(8643,36,'cp 40','b1c3 e8g8 a2a4 c6d4 d1d4 c8e6 a4a5 d8d7 f1d1 e6c4 d4c4 c7c6 c1f4 f8d8 e4e5 f6e8 e5d6 e8d6 c4e2 d7e6 e2e6 f7e6 f4d6 e7d6 c3e4 d6e7 d1d8 a8d8 g1f1 g8f8 c2c3 h7h6 f2f3 e7f6 f1e2 f6e7 h2h3 d8d5 b2b4');
INSERT INTO evaluations VALUES(8641,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(8641,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(8641,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(8641,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(8641,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(8641,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(8641,36,'cp 37','e8g8 a2a4 c6d4 d1d4 c8e6 a4a5 d8d7 c1f4 e6c4 d4c4 c7c6 a1d1 f8d8 f1e1 d7c8 f4e3 e7f8 e3g5 c8e6 c4d4 f8e7 g5f4 f6h5 f4c1 h7h6 h2h3 e7g5 c3e2 e6e5 d4b4 d8d7 c1g5 e5g5 b4d2 a8e8 d2g5 h6g5');
INSERT INTO evaluations VALUES(3412,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(3412,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(3412,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(3412,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(3412,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(3412,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(3412,36,'cp 46','c1f4 c6e5 c4b3 c8g4 f2f3 g4d7 a2a4 c7c5 d4e2 c5c4 f4e5 d6e5 b3c4 a8c8 b2b3 d8c7 g1h1 f8d8 c3d5 f6d5 e4d5 a7a6 e2c3 b7b5 a4b5 a6b5 c3b5 d7b5 c4b5 c7c2 b5c4 c2c3 d1e2 e7d6 g2g3 c8a8 a1d1 g7g6 d1d3 c3b4 f1d1 g8g7 h1g2 a8a3 e2d2');
INSERT INTO evaluations VALUES(6432,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(6432,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(6432,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(6432,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(6432,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(6432,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(6432,36,'cp 9','g8f6 b1c3 d7d6 c3a4 c5b6 e1g1 c6a5 c4b3 h7h6 c2c3 a5b3 a2b3 c8g4 h2h3 g4h5 a4b6 a7b6 a1a8 d8a8 g2g4 h5g6 g1h2 e8g8 f1g1 a8d8 c3c4 f6h7 g4g5 h7g5 f3g5 h6g5 c1g5 d8d7 g5e3 b6b5 d1g4 d7g4');
INSERT INTO evaluations VALUES(55073,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(55073,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(55073,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(55073,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(55073,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(55073,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(55073,36,'cp -55','d5d4 c3e2 e7e5 e2g3 c8e6 c2c3 b8c6 f1b5 g8e7 d2d3 d4c3 b2c3 d8d6 g1f3 a7a6 b5a4 e8c8 a4c2 f7f6 h2h3 g7g5 g3e2 h7h5 d3d4 g5g4 h3g4 e6g4 d4d5 c6a5 d1d3 e7g6 f3d2 h5h4 d2b3 a5b3 a2b3 c7c6 c1e3 h4h3 g2h3 c6d5');
INSERT INTO evaluations VALUES(55259,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(55259,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(55259,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(55259,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(55259,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(55259,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55259,36,'cp -34','c3e2 e7e5 c2c3 c7c5 e2g3 h7h5 h2h4 b8c6 f1c4 c8g4 g1f3 f8d6 g3f5 g8h6 f5d6 d8d6 d2d3 e8g8 d1d2 g8h7 f3h2 g4e6 c4e6 d6e6 d2g5 e6g6 c1d2 h6g4 g5g6 h7g6 h2f3 f7f6 a2a4 a8b8 c3d4 c5d4 a1c1 f8h8');
INSERT INTO evaluations VALUES(55272,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(55272,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(55272,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(55272,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(55272,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(55272,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(55272,36,'cp -36','e7e5 c2c3 c7c5 e2g3 h7h5 h2h4 b8c6 f1c4 c8g4 g1f3 f8d6 g3f5 g8h6 f5g3 d8c7 d2d3 a7a6 d1a4 g7g6 c4d5 e8f8 d5c6 c7c6 a4c6 b7c6 e1d1 f8g7 d1c2 a6a5 a2a4 g4e6 c1h6 g7h6 c3c4 h6g7 a1b1 e6g4 f3d2 h8b8 f2f3 g4d7 b2b3');
INSERT INTO evaluations VALUES(56599,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(56599,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(56599,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(56599,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(56599,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(56599,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(56599,36,'cp -28','c2c3 c7c5 e2g3 h7h5 h2h4 b8c6 f1c4 c8g4 g1f3 f8d6 g3f5 g8h6 f5g3 f7f6 d2d3 g7g6 g3e2 a7a5 e1g1 g4f3 g2f3 d8d7 g1g2 h6f7 c1d2 e8g8 a2a3 g8h8 c3d4 c5d4 d1a4 f7d8 b2b4 f6f5 e4f5');
INSERT INTO evaluations VALUES(56579,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(56579,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(56579,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(56579,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(56579,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(56579,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(56579,36,'cp -36','c7c5 e2g3 h7h5 h2h4 c8g4 g1f3 b8c6 f1c4 f8d6 g3f5 g8h6 f5d6 d8d6 d2d3 e8g8 e1g1 a7a5 a2a4 g8h8 d1d2 a8d8 f3h2 g4e6 c4e6 d6e6 d2e2 h6g4 f2f3 g4h2 g1h2 f7f6 c3c4 g7g6 c1h6 f8f7 f3f4 f7h7 f4f5 g6f5 e4f5 e6e8 h6d2 e8d7 e2f3 b7b6');
INSERT INTO evaluations VALUES(52715,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52715,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52715,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(52715,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(52715,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(52715,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52715,36,'cp 19','c6d5 d2d4 b8c6 c2c3 c8g4 h2h3 g4f5 d1b3 d8d7 f1b5 e7e6 f3e5 d7c7 b3a4 g8e7 c1f4 c7c8 b1d2 a7a6 b5e2 e7g6 e5g6 f5g6 a4d1 c8d7 e1g1 f8d6 f4d6 d7d6 a2a4 d6c7 f1e1 e8g8 d2f3 h7h6 e2d3 g6d3 d1d3 c6a5 e1e2 f8c8 f3e5 a5c6 a1e1 c6e5');
INSERT INTO evaluations VALUES(53426,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53426,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(53426,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(53426,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53426,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53426,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(53426,36,'cp 28','d2d4 b8c6 c2c3 c8g4 h2h3 g4h5 d1b3 d8d7 f3e5 c6e5 d4e5 a7a6 c3c4 d5c4 f1c4 e7e6 e1g1 b7b5 c4d3 a8d8 d3e4 f8c5 a2a4 g8e7 a4b5 d7b5 b3c2 c5d4 b1c3 b5c4 f1e1 d4e5 e4b7 h5g6 b7a6 c4a6 a1a6 g6c2 e1e5 e7g6 e5c5 e8g8 c1g5 f7f6');
INSERT INTO evaluations VALUES(53384,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53384,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53384,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53384,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53384,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53384,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(53384,36,'cp 12','b8c6 c2c3 c8g4 d1b3 d8d7 f1e2 e7e6 c1f4 f8d6 f4d6 g4f3 e2f3 d7d6 b1d2 a8b8 a2a4 g8f6 e1g1 c6a5 b3b5 a5c6 f1e1 e8g8 f3e2 b8c8 g2g3 b7b6 e2d3 c8c7 d2f3 f8c8 g1g2 h7h6');
INSERT INTO evaluations VALUES(19866,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19866,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19866,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(19866,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(19866,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(19866,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(19866,36,'cp 21','c2c3 a7a6 f1d3 c8g4 c1f4 e7e6 b1d2 g8f6 h2h3 g4h5 d1e2 f8d6 f4d6 d8d6 e1g1 d6f4 a1e1 a8c8 e2e3 f4e3 f2e3 h5g6 d3g6 h7g6 e3e4 f6e4 d2e4 d5e4 e1e4 h8h5 h3h4 h5a5 a2a3 c6e7 e4e2 a5b5 f1e1 e7d5 f3g5 c8c7 g2g3');
INSERT INTO evaluations VALUES(19854,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19854,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19854,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(19854,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(19854,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(19854,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(19854,36,'cp 29','a7a6 f1d3 c8g4 c1f4 e7e6 b1d2 g8f6 h2h3 g4h5 e1g1 f8d6 f4d6 d8d6 d1e2 e8g8 e2e3 d6c7 f1e1 a8c8 f3h4 f8e8 d2f3 f6d7 e3g5 h5f3 h4f3 h7h6 g5d2 c6a5 g2g3 a5c4 d2e2 c7b6 d3c4 d5c4 a2a4');
INSERT INTO evaluations VALUES(57308,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(57308,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(57308,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(57308,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(57308,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(57308,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(57308,36,'cp -50','d4c3 b1c3 f8c5 f1c4 d7d6 g1f3 g8f6 e4e5 d6e5 d1d8 e8d8 f3e5 h8e8 c1f4 c8e6 e1c1 d8c8 c4e6 f7e6 f4g3 e8e7 h1e1 c5d6 f2f3 d6e5 e1e5 b8c6 e5e3 b7b6 d1e1 c6d8 c3e4 f6d5 e3e2 h7h6 e4c3 d5c3 b2c3 e7d7 h2h4 a7a5 a2a4 c8b7 h4h5 a8a6 g3e5');
INSERT INTO evaluations VALUES(57317,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(57317,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(57317,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(57317,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(57317,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(57317,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(57317,36,'cp -29','b1c3 f8c5 f1c4 d7d6 g1f3 g8f6 e4e5 d6e5 d1d8 e8d8 f3e5 h8e8 c1f4 c8e6 e1c1 d8c8 c4e6 f7e6 f4g3 c5d6 h1e1 e8e7 g3h4 a7a5 f2f4 e7e8 f4f5 e6f5 e5f7 e8e1 d1e1 c8d7 f7d6 c7d6 h4f6 g7f6 e1f1 b8c6 f1f5 d7e6 f5h5 a8h8 c3b5 c6e5 b5d4 e6d7');
INSERT INTO evaluations VALUES(57313,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(57313,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(57313,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(57313,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(57313,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(57313,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(57313,36,'cp -45','f8c5 f1c4 d7d6 g1f3 g8f6 e4e5 d6e5 d1d8 e8d8 f3e5 h8e8 c1f4 c8e6 e1c1 d8c8 c4e6 f7e6 f4g3 a7a5 h1e1 f6h5 g3h4 a8a6 c3b5 a5a4 e1e4 h5f6 e4c4 b7b6 e5d3 b8d7 b5c3 c5d6 h4g3 d6g3 h2g3 f6d5 c4a4 a6a4 c3a4 e6e5 d1e1 e5e4 d3f4 d7f6 f4d5 f6d5 a4c3 d5c3 b2c3');
INSERT INTO evaluations VALUES(39303,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(39303,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(39303,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(39303,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(39303,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(39303,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(39303,36,'cp -47','f1c4 d7d6 g1f3 g8f6 e4e5 d6e5 d1d8 e8d8 f3e5 h8e8 c1f4 c8e6 e1c1 d8c8 c4e6 f7e6 f4g3 a7a5 h1e1 f6h5 g3h4 a8a6 c3b5 a5a4 c1b1 a6a5 b5c3 c5b4 e5c4 a5a6 e1e5 h5f6 c3b5 b8c6 e5e3 e6e5 h4f6 g7f6 e3h3 e8h8 h3g3 b4f8 g3f3 f8g7 a2a3 h8d8 d1d8 c6d8 c4e3 d8c6 e3d5');
INSERT INTO evaluations VALUES(39284,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(39284,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(39284,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(39284,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(39284,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(39284,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(39284,36,'cp -38','d7d6 g1f3 g8f6 e4e5 d6e5 d1d8 e8d8 f3e5 h8e8 c1f4 c8e6 e1c1 d8c8 c4e6 f7e6 f4g3 c5d6 h1e1 e8e7 g3h4 a7a5 f2f4 a8a6 a2a3 e7e8 c1b1 b8d7 e1e2 a5a4 e5d3 a6a5 d1e1 f6d5 c3d5 a5d5 e2e6 e8e6 e1e6 d7b6 b1c2 c8d7 e6e1 d5f5 g2g4');
INSERT INTO evaluations VALUES(39257,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(39257,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(39257,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(39257,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(39257,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(39257,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(39257,36,'cp 216','f3e5 b8c6 e5f3 d7d5 e4d5 d8d5 f1e2 c6d4 f3d4 d5d4 e1g1 g8f6 c2c3 d4h4 d2d4 c5d6 g2g3 h4h3 f1e1 e8g8 e2f1 h3f5 b1d2 c8e6 d1f3 f5f3 d2f3 f8e8 f3e5 c7c5 d4c5 d6c5 e5d3 c5b6 f1g2 e6d5 e1e8 a8e8 g2d5');
INSERT INTO evaluations VALUES(39189,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(39189,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(39189,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(39189,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(39189,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(39189,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(39189,36,'cp 195','b8c6 e5c6 d7c6 c2c3 d8h4 d1f3 h4h6 b2b4 c5e7 b1a3 h6g6 d2d4 c8g4 f3f4 g4e6 f2f3 e7d6 e4e5 d6e7 f4d2 f7f6 f1d3 e6f5 d3f5 g6f5 a3c4 f6e5 c4e5 a7a5 b4a5 e7d6 e1g1 g8e7 h2h3 e8g8 a2a4 d6e5 d4e5');
INSERT INTO evaluations VALUES(55071,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(55071,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(55071,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(55071,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(55071,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(55071,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(55071,36,'cp -75','d5e4 b1c3 g8f6 f2f3 e4f3 g1f3 c7c6 f1d3 c8g4 c1e3 e7e6 h2h3 g4h5 d1e2 f8b4 e1g1 b8d7 e2f2 h5g6 d3g6 h7g6 f3g5 d7b6 e3f4 b4e7 a1d1 b6c4 b2b3 c4d6 g2g4 d8d7 f4e5 e8c8 f2g3 d8f8 g5f3 f6d5 c3d5 e6d5 e5g7');
INSERT INTO evaluations VALUES(55248,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(55248,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(55248,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(55248,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(55248,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(55248,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(55248,36,'cp -68','b1c3 g8f6 f2f3 e4f3 g1f3 c7c6 h2h3 c8f5 f1d3 f5d3 d1d3 b8d7 c1d2 g7g6 e1c1 f8g7 h1e1 e8g8 g2g4 b7b5 g4g5 f6h5 c1b1 d7b6 h3h4 a7a5 c3e4 d8d5 d2c1 a8e8 d3f1 b6c4 f3e5 g7e5 d4e5 d5e5 f1e2 c4d6');
INSERT INTO evaluations VALUES(55243,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(55243,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(55243,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(55243,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(55243,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(55243,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(55243,36,'cp -75','g8f6 f2f3 e4f3 g1f3 c7c6 h2h3 c8f5 f1d3 f5d3 d1d3 e7e6 c1d2 b7b5 f3e5 f8d6 e1g1 e8g8 a1e1 b5b4 c3a4 d6e5 d4e5 d8d3 c2d3 f6d5 a4c5 a7a5 e1c1 f8c8 h3h4 c8c7 g2g4 h7h6 g1g2 a5a4 g4g5 h6g5 h4g5 b8d7');
INSERT INTO evaluations VALUES(47380,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(47380,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(47380,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(47380,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(47380,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(47380,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(47380,36,'cp -62','f2f3 e4f3 g1f3 c7c6 h2h3 c8f5 f1d3 f5d3 d1d3 e7e6 c1d2 f8b4 e1c1 b4c3 d3c3 e8g8 c1b1 f6e4 c3e3 e4d2 e3d2 d8d5 d2e2 b8d7 g2g4 d7b6 h3h4 d5b5 e2e1 a7a5 a2a3 b6a4 b2b3 a4b6 f3e5 b6d5');
INSERT INTO evaluations VALUES(47377,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(47377,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(47377,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(47377,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(47377,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(47377,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(47377,36,'cp -85','e4f3 g1f3 g7g6 c1g5 f8g7 d1d2 e8g8 e1c1 c7c5 d4c5 d8a5 f1b5 a7a6 b5a4 a5c5 h1e1 b8c6 a4c6 b7c6 g5h6 c8g4 h6g7 g8g7 d2d4 c5d4 d1d4 g4f3 g2f3 e7e6 d4c4 f8c8 c3e4 c8c7 e4c5 f6d5 b2b3 a6a5 c1b2 h7h6 h2h4 g7f6 c4d4 f6e7 c2c4 d5f6 e1e5 c7c8 e5e1');
INSERT INTO evaluations VALUES(47363,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(47363,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(47363,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(47363,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(47363,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(47363,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(47363,36,'cp -64','g1f3 g7g6 c1g5 f8g7 d1d2 e8g8 e1c1 c7c6 h2h3 c8e6 g5h6 b8d7 h6g7 g8g7 d1e1 d8a5 a2a3 a8e8 e1e6 f7e6 f3g5 e6e5 g5e6 g7h8 f1c4 e5d4 d2d4 a5e5 d4e5 d7e5 e6f8 e5c4 f8e6 h8g8 g2g4 h7h6 h1e1 g8f7 g4g5 h6g5 e6g5 f7g7 g5e6 g7h6');
INSERT INTO evaluations VALUES(47361,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(47361,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(47361,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(47361,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(47361,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(47361,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(47361,36,'cp -82','g7g6 c1g5 f8g7 d1d2 e8g8 e1c1 c7c5 f1c4 c5d4 f3d4 d8a5 h2h3 b8c6 d4c6 b7c6 g5h6 g7h6 d2h6 c8a6 c4a6 a5a6 h6g5 a8c8 c1b1 a6b7 h1e1 f8d8 g2g4 d8d1 e1d1 c8e8 b2b3 b7c7 g5c5 h7h5 g4h5 f6h5 d1e1 e7e6');
INSERT INTO evaluations VALUES(22786,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(22786,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(22786,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(22786,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(22786,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(22786,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(22786,36,'cp -6','d7d5 d1a4 f7f6 d2d3 g8e7 b2b4 a7a6 b4b5 c6a7 b5a6 a7c6 a4c2 a8a6 b1d2 d5d4 c3d4 e5d4 h2h4 e7g6 h4h5 g6e5 f1e2 f8b4 h5h6 g7g5 f3e5 c6e5 a1b1 d8e7 e1g1 e8d8 d2f3 c7c5 a2a3 e5f3 e2f3 b4a5 f3h5 g5g4 f2f4');
INSERT INTO evaluations VALUES(21710,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(21710,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(21710,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(21710,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(21710,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(21710,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(21710,36,'cp -5','d1a4 d8d6 d2d3 g8f6 f1e2 f8e7 b2b4 e8g8 b1d2 a7a6 a4c2 b7b5 a2a4 c8e6 e1g1 f6d7 c1b2 f7f5 h2h3 d7b6 a4b5 a6b5 a1a8 f8a8 e4d5 e6d5 c3c4 c6b4 c2b3 b5c4 d3c4 d5f3 d2f3');
INSERT INTO evaluations VALUES(21770,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(21770,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(21770,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(21770,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(21770,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(21770,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(21770,36,'cp -16','d8d6 f1b5 g8f6 e4d5 f6d5 e1g1 c8d7 d2d4 e5d4 f3d4 d5b6 a4b3 c6d4 b5d7 d6d7 c3d4 e8c8 b1c3 d7e6 b3e6 f7e6 f1d1 b6d5 g2g3 f8e7 c3e4 h7h6 a2a3 b7b6 d1e1 h8f8 c1d2 g7g5 a1c1 e7f6 e4f6 f8f6 e1e5 d8f8 e5e2 g5g4');
INSERT INTO evaluations VALUES(57435,30,'cp 247',NULL);
INSERT INTO evaluations VALUES(57435,31,'cp 247',NULL);
INSERT INTO evaluations VALUES(57435,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(57435,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(57435,34,'cp 247',NULL);
INSERT INTO evaluations VALUES(57435,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(57435,36,'cp 258','f3e5 d8e7 e5f3 d7d5 d2d3 d5e4 d3e4 e7e4 f1e2 b8c6 b1c3 e4e6 e1g1 c8d7 f3d4 c6d4 e2h5 g7g6 f1e1 e8c8 e1e6 d4e6 h5g4 f6f5 g4f3 g8e7 c1e3 e7c6 c3b5 c8b8 f3c6 d7c6 e3a7 b8a8 d1e2 c6b5 e2e6 f8d6 a7e3 f5f4 e3d4 h8e8 e6b3 b5c6 f2f3 e8e2 b3c4 d8e8 a2a4 g6g5');
INSERT INTO evaluations VALUES(57388,30,'cp 232',NULL);
INSERT INTO evaluations VALUES(57388,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(57388,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(57388,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(57388,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(57388,35,'cp 248',NULL);
INSERT INTO evaluations VALUES(57388,36,'cp 227','g8e7 e5f3 d7d5 e4d5 d8d5 c2c4 d5f7 d2d4 c8g4 f1e2 g4f3 e2f3 b8c6 d4d5 e8c8 e1g1 c6e5 f3e2 e7f5 b1c3 c8b8 a1b1 f8b4 a2a3 b4c3 b2c3 h8e8 a3a4 f5d6 d1b3 e5d7 c1a3 b7b6 a3d6 c7d6 b3d1 d7c5 a4a5 b8b7 f1e1 f7c7 a5b6 a7b6 d1c2 e8e5 c2h7 d8e8 h7c2');
INSERT INTO evaluations VALUES(6421,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(6421,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(6421,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(6421,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6421,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(6421,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6421,36,'cp 15','g8f6 d2d3 h7h6 h2h3 d7d6 c3a4 c5b6 c2c3 c8d7 a2a3 c6e7 e1g1 e7g6 a4b6 a7b6 c4b3 e8g8 f1e1 f8e8 b3c2 d6d5 e4d5 f6d5 d3d4 e5d4 d1d4 e8e1 f3e1 d5f6 c1f4 c7c5 d4d6 g6f4 d6f4 d7c6 a1d1 f6d5 f4f5 d8g5 f5g5 h6g5 g2g3');
INSERT INTO evaluations VALUES(47155,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(47155,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47155,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(47155,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47155,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47155,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47155,36,'cp 15','c7c5 e2e3 b8c6 b1d2 e7e6 c2c3 f8e7 f3e5 d8b6 a1b1 c6e5 f4e5 c8d7 a2a4 c5c4 b2b3 c4b3 d1b3 b6b3 b1b3 b7b6 f1b5 d7b5 a4b5 e8g8 e1e2 f6d7 h1a1 d7e5 d4e5 f8c8 f2f4 c8c7 d2f3 a8c8 e2d3 g7g6 a1a6');
INSERT INTO evaluations VALUES(45362,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45362,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45362,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45362,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45362,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45362,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45362,36,'cp 0','e2e3 b8c6 b1d2 d8b6 d4c5 b6b2 c2c4 e7e6 a1c1 b2a2 f3d4 e6e5 c1a1 a2b2 a1b1 b2a2');
INSERT INTO evaluations VALUES(45361,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45361,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45361,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45361,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45361,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45361,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45361,36,'cp 0','b8c6 b1d2 d8b6 d4c5 b6b2 c2c4 e7e6 a1c1 b2a2 f3d4 e6e5 c1a1 a2b2 a1b1 b2a2');
INSERT INTO evaluations VALUES(13095,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13095,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13095,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13095,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13095,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13095,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13095,36,'cp 0','b1d2 d8b6 d4c5 b6b2 c2c4 e7e6 a1c1 b2a2 f3d4 e6e5 c1a1 a2b2 a1b1 b2a2');
INSERT INTO evaluations VALUES(13096,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13096,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13096,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13096,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13096,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13096,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13096,36,'cp 0','d8b6 d4c5 b6b2 c2c4 e7e6 a1c1 b2a2 f3d4 e6e5 c1a1 a2b2 a1b1 b2a2');
INSERT INTO evaluations VALUES(54976,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54976,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(54976,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(54976,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(54976,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(54976,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54976,36,'cp 28','c7c5 d2d4 c5d4 f3d4 g8f6 b1c3 a7a6 h2h3 e7e5 d4b3 c8e6 d1f3 b8d7 f1d3 a8c8 c1e3 f8e7 a2a4 h7h6 a4a5 e8g8 e1g1 d8c7 f1d1 f8d8 d3e2 d7c5 b3c5 d6c5 d1d8 c8d8 b2b3 d8d4 f3g3 g8f8 c3d5 f6d5 e3d4 c5d4 e4d5 e6d5 e2d3 e7b4');
INSERT INTO evaluations VALUES(55228,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(55228,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(55228,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(55228,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(55228,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(55228,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(55228,36,'cp 18','e7e5 g1f3 e5d4 f1c4 b8c6 e1g1 c8e6 c4b5 f8c5 b2b4 c5b6 a2a4 a7a6 b5c6 b7c6 a4a5 b6a7 c1b2 g8f6 b2d4 f6e4 d4a7 d8d1 f1d1 a8a7 f3d4 c6c5 d4c6 a7a8 f2f3 e4f6 b4c5 f6d7 d1c1 e8g8 b1d2 f8e8 a1b1 d7f6 g1f2 e6d5 c6b4 g8f8 g2g4 e8d8 d2b3 d5b3 b1b3 d8d2 f2g3 f6d5 b4d5');
INSERT INTO evaluations VALUES(56557,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(56557,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(56557,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(56557,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(56557,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(56557,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(56557,36,'cp 31','g1f3 e5d4 f1c4 b8c6 e1g1 c8e6 c4e6 f7e6 d1b3 d8d7 b3b7 a8b8 b7a6 f8d6 b1d2 h7h6 d2b3 g8f6 b3d4 c6d4 f3d4 e8g8 d4f3 f6e4 a6e2 e4f6 b2b3 b8b4 c1b2 b4e4 e2c2 d7b5 a1d1 b5f5 b2f6 f8f6 d1d2 e6e5 c2d3 e4b4 d3f5 f6f5 f1d1 e5e4');
INSERT INTO evaluations VALUES(56556,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(56556,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(56556,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(56556,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(56556,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(56556,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(56556,36,'cp 40','e5d4 f1c4 b8c6 e1g1 c8e6 c4e6 f7e6 d1b3 d8d7 b3b7 a8b8 b7a6 f8d6 b1d2 h7h6 d2b3 g8f6 b3d4 c6d4 f3d4 e8g8 d4f3 f6e4 a6e2 e4c5 b2b3 d7b5 e2b5 b8b5 f1d1 a7a5 f3d4 b5b7 c1a3 f8e8 a3c5 d6c5 a1c1 c5a3 c1c3 e8d8 d1d3 e6e5 d4c6 d8d3 c3d3 e5e4 d3d7');
INSERT INTO evaluations VALUES(56657,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(56657,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(56657,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(56657,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(56657,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(56657,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(56657,36,'cp 44','f1c4 b8c6 e1g1 c8e6 c4e6 f7e6 d1b3 d8d7 b3b7 a8b8 b7a6 f8d6 b1d2 h7h6 d2b3 g8f6 b3d4 c6d4 f3d4 e8g8 d4f3 f6e4 a6e2 e4f6 b2b3 a7a5 f3e5 d6e5 e2e5 b8b5 e5e2 a5a4 b3a4 b5a5 f1e1 a5a4 e2e6 d7e6 e1e6 c7c5 c1b2 a4e4 e6b6 e4e2 b2f6 f8f6 b6f6 g7f6');
INSERT INTO evaluations VALUES(56645,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(56645,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(56645,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(56645,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(56645,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(56645,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(56645,36,'cp 24','b8c6 e1g1 c8e6 c4e6 f7e6 b1d2 d8d7 d2c4 g8f6 c4e5 c6e5 f3e5 d7b5 c1f4 f8d6 a2a4 b5a6 d1d4 e8g8 b2b4 f6g4 g2g3 d6e5 f4e5 a8d8 d4b2 g4e5 b2e5 a6d6 e5d6 d8d6 f1c1 d6d4 c1c7 f8f7 a1c1 f7c7 c1c7 d4b4 a4a5 b4b5 a5a6 b5b1 g1g2 b7a6 c7a7 b1e1 a7a8 g8f7');
INSERT INTO evaluations VALUES(21837,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(21837,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21837,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(21837,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(21837,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21837,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(21837,36,'cp 44','e1g1 c8e6 c4e6 f7e6 d1b3 d8d7 b3b7 a8b8 b7a6 f8d6 b1d2 h7h6 d2c4 g8e7 c1d2 e8g8 f3e1 e7g6 a1d1 c6b4 d2b4 b8b4 b2b3 e6e5 e1d3 b4b8 f2f3 d7e7 d1c1 g8h7 g1h1 b8a8 a6a5 f8f7 c1c2 g6f4 d3f4 f7f4 c4b2 c7c5 b2d3 f4f8 a5a6 f8f6 a6c4 a8f8');
INSERT INTO evaluations VALUES(23472,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(23472,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(23472,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23472,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(23472,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(23472,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(23472,36,'cp 57','d2d4 d7d5 c2c4 e7e6 b1c3 g8f6 e2e3 f8e7 a2a3 e8g8 b2b4 d5c4 f1c4 e7d6 e1g1 e6e5 b4b5 c6a5 c4f7 f8f7 d4e5 d6e7 e5f6 d8d1 c3d1 g7f6 f3d4 c8d7 a3a4 c7c5 b5c6 a5c6 d4c6 d7c6 c1a3 f7g7 g2g3 a8d8 a3e7 g7e7 d1c3 d8d2');
INSERT INTO evaluations VALUES(23438,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(23438,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(23438,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(23438,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(23438,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(23438,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(23438,36,'cp 51','d7d5 c2c4 e7e6 b1c3 g8f6 c4d5 e6d5 c1g5 f8e7 e2e3 f6h5 g5e7 c6e7 f1d3 e8g8 e1g1 h5f6 b2b4 c8f5 a2a4 c7c6 h2h3 d8d6 d1c2 f5d3 c2d3 f8c8 a4a5 h7h6 f1c1 d6b4 c1b1 b4d6 b1b7 c6c5 d4c5 c8c5');
INSERT INTO evaluations VALUES(52306,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(52306,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(52306,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(52306,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(52306,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(52306,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(52306,36,'cp 57','d2d4 c5d4 f3d4 b8c6 c2c4 g8f6 b1c3 d7d6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 a7a5 e1g1 a5a4 e3d4 c8d7 b2b4 a4b3 a2b3 d7e6 f2f4 a8a1 f1a1 e6g4 e2f1 g4d7 h2h3 d7c6 c3d5 f6d7 d4g7 g8g7 b3b4 c6d5 d3d5');
INSERT INTO evaluations VALUES(52293,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(52293,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(52293,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(52293,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(52293,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(52293,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(52293,36,'cp 64','c5d4 f3d4 f8g7 c2c4 d7d6 b1c3 d8a5 c1e3 b8c6 f1e2 c6d4 e3d4 g7d4 d1d4 g8f6 e1g1 e8g8 b2b3 c8e6 f2f4 a8c8 c3b5 a5b6 e2f3 b6d4 b5d4 b7b6 f1e1 c8c5 h2h3 f8c8 a2a4 c8c7 d4b5 c7d7 a1d1 a7a6 b5d4 d7c7 e4e5 d6e5 d4e6 f7e6 f4e5 f6d7');
INSERT INTO evaluations VALUES(52338,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(52338,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(52338,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52338,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(52338,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(52338,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(52338,36,'cp 64','f3d4 b8c6 c2c4 d7d6 f1e2 g8f6 b1c3 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 c8e6 e3d4 d8a5 e1g1 f6d7 d4g7 g8g7 b2b3 d7f6 h2h3 a8c8 f2f4 a7a6 a1c1 c8c7 d3g3 b7b5 f4f5 e6d7 g1h1 b5c4 e2c4 a5e5 g3e5 d6e5 c4a6 f8a8');
INSERT INTO evaluations VALUES(52329,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(52329,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(52329,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(52329,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(52329,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(52329,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(52329,36,'cp 53','b8c6 c2c4 d7d6 b1c3 c6d4 d1d4 g8f6 f1e2 f8g7 c1e3 a7a5 e1g1 e8g8 d4d2 a5a4 e3d4 d8a5 b2b4 a5b4 a1b1 b4a5 e4e5 f6e8 d2e3 e8c7 e5d6 e7d6 d4g7 g8g7 e2f3 f8e8 e3d2 a4a3 f1c1 a5e5 c1e1 e5c5 c3e4 c5a5 d2d6 a8a6 d6d3 a5e5 d3e3 a6c6 e3a3');
INSERT INTO evaluations VALUES(19319,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(19319,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(19319,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(19319,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(19319,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(19319,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(19319,36,'cp 50','c2c4 d7d6 f1e2 g8f6 b1c3 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 c8d7 e1g1 a7a6 a2a4 a6a5 d3d2 d7c6 f2f3 f6d7 b2b3 d7c5 a1b1 b7b6 e3g5 c5e6 g5h6 e6c5 h6g7 g8g7 d2e3 f7f6 f3f4 d8c7 h2h4 f6f5 e4f5 f8f5 c3b5');
INSERT INTO evaluations VALUES(19310,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(19310,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(19310,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(19310,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(19310,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(19310,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(19310,36,'cp 48','g8f6 b1c3 c6d4 d1d4 d7d6 f1e2 f8g7 e1g1 e8g8 d4d3 c8d7 c1e3 a7a6 a2a4 a6a5 a1d1 d7c6 d3c2 f6d7 b2b3 b7b6 h2h4 d7c5 h4h5 g7c3 c2c3 c5e4 c3b2 e7e5 e2f3 f8e8 h5h6 e8e6 f3e4 c6e4 f2f4 d8e7 b2d2 a8e8 e3b6 f7f5 f4e5 d6e5');
INSERT INTO evaluations VALUES(12735,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(12735,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(12735,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(12735,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(12735,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(12735,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(12735,36,'cp 62','b1c3 d7d6 f1e2 c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 c8e6 c1d2 a7a6 b2b3 b7b5 c4b5 a6b5 c3b5 d6d5 e4e5 f6e4 d2f4 g6g5 f4e3 g7e5 e3d4 d8b8 d4e5 b8e5 d3e3 g5g4 a2a4 f8c8 e2d3 f7f5 e3d4 e5d4 b5d4 e6d7 f2f3 c8c3 d3e4 f5e4 f3e4');
INSERT INTO evaluations VALUES(12734,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(12734,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(12734,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(12734,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(12734,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(12734,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(12734,36,'cp 47','d7d6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 a7a6 e1g1 c8d7 a2a4 a6a5 a1d1 d7c6 b2b3 f6d7 d3c2 b7b6 e3g5 d7c5 c3d5 a8a7 f2f4 c6d5 e4d5 f7f5 g1h1 d8e8 d1e1 h7h6 g5h4 g7f6 h4f2 h6h5 h2h3 a7d7 e2f3 e8f7 f2e3 f8c8');
INSERT INTO evaluations VALUES(13034,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(13034,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(13034,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(13034,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(13034,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(13034,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(13034,36,'cp 62','f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 a7a6 e1g1 c8d7 a2a4 a6a5 a1d1 d7c6 d3c2 f6d7 b2b3 b7b6 h2h4 d7c5 h4h5 g7c3 c2c3 c5e4 c3b2 e7e5 e2f3 f7f5 f3e4 c6e4 f2f3 e4b7 b2f2 f5f4 e3b6 d8d7 f1e1 f8e8 f2h4 e8e6 h4g5 d7c6 b6f2 a8e8 d1d2 c6c8');
INSERT INTO evaluations VALUES(13033,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(13033,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(13033,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(13033,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(13033,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13033,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(13033,36,'cp 42','c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 a7a6 c1e3 c8d7 a2a4 a6a5 a1d1 d7c6 d3c2 f6d7 b2b3 b7b6 e3g5 d7c5 c3d5 a8a7 f2f4 c6d5 e4d5 f7f5 e2f3 h7h6 g5h4 h6h5 h4g5 g7f6 h2h4 a7b7 g2g3 b7b8 f1e1 g8h7 e1e2 d8c7 g1h2');
INSERT INTO evaluations VALUES(55406,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(55406,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(55406,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(55406,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(55406,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(55406,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(55406,36,'cp 72','d2d4 e5d4 f3d4 g8f6 b1c3 f8e7 d1e2 e8g8 c1f4 b8c6 e1c1 c6d4 d1d4 f6d7 e4e5 d6e5 f4e5 e7f6 e5f6 d8f6 h1d1 d7b6 c4b3 c7c5 d4d2 c8f5 b3d5 f8e8 e2b5 a7a6 b5a5 a8b8 f2f3 e8c8 g2g3 b6d5 c3d5 f6e6 a5c3 b7b5 g3g4 f5g6 f3f4 g8h8');
INSERT INTO evaluations VALUES(55408,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(55408,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(55408,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(55408,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(55408,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(55408,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(55408,36,'cp 62','e5d4 f3d4 g8f6 b1c3 f8e7 c1f4 e8g8 d1e2 c7c6 e1c1 f8e8 f2f3 b7b5 c4b3 a7a5 a2a4 b5b4 c3b1 e7f8 g2g4 b8d7 g4g5 f6h5 f4e3 c8a6 e2f2 h6g5 d4f5 h5f4 e3f4 g5f4 f5d6 f8d6 d1d6');
INSERT INTO evaluations VALUES(55431,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(55431,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(55431,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(55431,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(55431,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(55431,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(55431,36,'cp 98','f3d4 g8f6 b1c3 f8e7 d1e2 e8g8 c1f4 b8c6 e1c1 c6d4 d1d4 f6d7 e4e5 d6e5 f4e5 e7f6 e5f6 d8f6 h1d1 d7b6 c4b3 c7c5 d4d2 c8f5 e2b5 a8c8 b3d5 a7a6 b5b3 b6d5 c3d5 f6e5 b3b7 c8b8 f2f4 e5f4 b7f7 f8f7 d5f4 f5g4 d2d8 b8d8 d1d8 g8h7 f4d3 c5c4 d8d4 c4d3 d4g4 f7f1 c1d2 d3c2 d2c2');
INSERT INTO evaluations VALUES(55426,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(55426,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(55426,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(55426,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(55426,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(55426,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(55426,36,'cp 87','g8f6 b1c3 f8e7 d1e2 e8g8 c1f4 b8c6 e1c1 c6d4 d1d4 f6d7 e4e5 d6e5 f4e5 e7f6 e5f6 d8f6 h1d1 d7b6 c4b3 c7c5 d4d2 f6g5 f2f3 c8f5 e2e5 c5c4 h2h4 g5g6 b3a4 b6a4 c3a4 f8e8 e5b5 b7b6 a4c3 a8c8 b5a6 g6c6 a2a3 f5e6 d2d4 c6a8 a6b5 c8c5 b5b4 a8b8 g2g4');
INSERT INTO evaluations VALUES(47644,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(47644,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(47644,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(47644,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(47644,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(47644,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(47644,36,'cp 81','b1c3 f8e7 d1e2 e8g8 c1f4 b8c6 e1c1 c6d4 d1d4 f6d7 e4e5 d6e5 f4e5 e7f6 e5f6 d8f6 h1d1 d7b6 c4b3 c7c5 d4d2 f6g5 f2f3 c8f5 a2a4 c5c4 b3c4 f8e8 e2f1 b6c4 f1c4 a8d8 c4f1 a7a6 f1f2 b7b5 h2h4 d8d2 d1d2 g5f4 g2g3 f4c7 a4b5 a6b5 c3b5 e8a8 b5a3 h6h5 c1b1');
INSERT INTO evaluations VALUES(47642,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(47642,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(47642,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(47642,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(47642,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(47642,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(47642,36,'cp 91','f8e7 d1e2 e8g8 c1f4 f6d7 e1c1 e7f6 f4e3 b8c6 e2d2 c6d4 e3d4 f6d4 d2d4 d8g5 c1b1 d7b6 c4e2 c8d7 g2g4 a8d8 h2h4 g5c5 d4c5 d6c5 b1c1 d7c6 f2f3 c5c4 g4g5 f8e8 g5h6 g7h6 a2a4 d8d1 h1d1 b6a4 c3a4 c6a4 e2c4');
INSERT INTO evaluations VALUES(40699,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(40699,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(40699,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(40699,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(40699,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(40699,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(40699,36,'cp 113','d1e2 e8g8 c1f4 a7a6 e1c1 b7b5 c4d5 f6d5 c3d5 e7f6 e2h5 f8e8 h2h4 e8e4 h5f3 e4d4 d5f6 d8f6 d1d4 f6d4 f3a8 d4f4 c1b1 c8f5 a8b8 g8h7 b8c7 f4f2 h1c1 d6d5 c7d6 f2h4 d6d5 h4f6 b2b3 f6g6 g2g3 g6f6 c1e1 f5g6 d5d2 a6a5 d2a5 f6f2 a5c3 b5b4 c3b4 g6c2 b1b2 c2g6 b2a3 f2g3');
INSERT INTO evaluations VALUES(54297,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(54297,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(54297,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(54297,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(54297,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(54297,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(54297,36,'cp -16','c5d4 f3d4 g8e7 f1d3 b8c6 d4c6 e7c6 d1e2 d8c7 f2f4 f8c5 c2c3 c8d7 b1d2 a7a5 d2b3 c5b6 c1e3 a5a4 b3d4 b6d4 c3d4 c6b4 d3b1 c7a5 e1f2 d7b5 e2d2 b4d3 b1d3 a5d2 e3d2 b5d3 a1c1 e8d7 c1c3 d3f5 h1c1 h8c8 c3c8 a8c8 c1c8 d7c8 f2f3 c8d7 d2b4 d7c6');
INSERT INTO evaluations VALUES(54436,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(54436,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(54436,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(54436,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(54436,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(54436,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(54436,36,'cp 0','f3d4 b8c6 d4c6 b7c6 b1d2 d8b6 f1d3 c8a6 e1g1 g8e7 d2f3 c6c5 b2b4 a6d3 b4c5 d3c2 d1c2 b6a6 f3d4 e7c6 d4c6 a6c6 c1e3 f8e7 f1c1 a8b8 a1b1 b8b1 c2b1 e8g8 g2g3 f8c8 c1c3 a7a5 b1c2 a5a4 h2h4 h7h6 h4h5');
INSERT INTO evaluations VALUES(54419,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54419,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(54419,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(54419,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(54419,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(54419,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54419,36,'cp -9','g8e7 f1d3 b8c6 d4c6 e7c6 d1e2 d8c7 f2f4 c6d4 e2f2 f8c5 c1e3 c7b6 e1g1 d4e2 f2e2 c5e3 g1h1 c8d7 b1d2 g7g6 c2c3 e3d2 e2d2 d7b5 d3b5 b6b5 a2a4 b5d7 f1f3 a7a6 a4a5 d7b5 b2b4 e8g8 g2g4 d5d4 c3d4 f8d8 f3d3');
INSERT INTO evaluations VALUES(44693,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(44693,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(44693,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44693,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(44693,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(44693,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(44693,36,'cp 0','f1d3 b8c6 d4c6 e7c6 d1e2 d8c7 f2f4 c6d4 e2f2 f8c5 c1e3 c7b6 e1g1 c8d7 g1h1 b6b2 b1d2 b7b6 d2b3 d4b3 a2b3 c5e3 f2e3 b2c3 e3g3 g7g6 h2h4 e8g8 h1h2 g8h8 g3g5 f7f5 e5f6 c3f6 a1a6 d7c8 a6a1');
INSERT INTO evaluations VALUES(52418,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52418,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(52418,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(52418,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52418,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(52418,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(52418,36,'cp 0','b8c6 g1e2 g8f6 b1c3 e7e6 d2d4 c5d4 e2d4 f8b4 d4c6 b7c6 f1d3 e6e5 c1e3 e8g8 e1g1 b4c3 b2c3 d7d6 c4c5 f6g4 e3c1 d6c5 d3c4 d8h4 h2h3 g4f6 f2f3 c8h3 g2h3 h4g3 g1h1 f6e4 f3e4 g3h3 h1g1 h3g3 g1h1 g3h4 h1g1 h4g3');
INSERT INTO evaluations VALUES(19391,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(19391,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19391,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(19391,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(19391,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19391,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(19391,36,'cp 5','g1e2 g8f6 b1c3 e7e6 d2d4 d7d5 c4d5 e6d5 c1g5 c5d4 e2d4 f8e7 f1b5 c8d7 b5c6 b7c6 e4d5 f6d5 d1e2 d5c3 b2c3 h7h6 e2e7 d8e7 g5e7 e8e7 e1c1 h8e8 c3c4 a8d8 h1e1 e7f6 e1e8 d7e8 d4b3 d8d1 c1d1 f6e5 d1c2 e5d6 g2g3 c6c5 c2d3 e8d7 f2f4 d7f5 d3c3 h6h5');
INSERT INTO evaluations VALUES(19392,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19392,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19392,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(19392,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19392,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(19392,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19392,36,'cp 10','g8f6 b1c3 e7e6 d2d4 c5d4 e2d4 f8b4 d4c6 b7c6 f1d3 e6e5 e1g1 e8g8 h2h3 b4c3 b2c3 d7d6 c1e3 c8e6 d1d2 d8c7 f2f4 e5f4 e3f4 f6d7 f1d1 d7e5 d3f1 c7b6 c4c5 b6c5 f4e3 c5a5 d2d6 f8d8 d6c5 a5c5 d1d8 a8d8 e3c5 d8d2 c5a7 d2a2 a1a2 e6a2 g1f2 a2c4');
INSERT INTO evaluations VALUES(6394,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(6394,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(6394,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(6394,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(6394,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(6394,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(6394,36,'cp 64','c2c3 g8f6 d2d4 e5d4 e4e5 f6d5 c3d4 c5b4 b1d2 e8g8 e1g1 b4e7 a2a3 d7d6 d1c2 d5f4 b5c6 b7c6 d2c4 f4g6 c4a5 c6c5 a5c6 d8e8 c6e7 e8e7 e5d6 c7d6 d4c5 d6c5 f1e1 e7b7 c1e3 c8g4 f3d2 f8c8 f2f3 g4e6 d2e4 c5c4 a1d1 h7h6');
INSERT INTO evaluations VALUES(6389,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(6389,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(6389,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(6389,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(6389,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(6389,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(6389,36,'cp 48','g8f6 d2d4 e5d4 e4e5 f6d5 c3d4 c5b4 b1d2 e8g8 e1g1 b4e7 a2a3 d7d6 d1c2 c8e6 d2e4 d8d7 h2h3 a7a6 b5c4 h7h6 f1e1 d5b6 c4f1 e6d5 c1f4 f7f5 e4d2 a8d8 a1d1 g8h8 b2b4 d5g8 b4b5 a6b5 f1b5 d8a8 d2c4 b6c4');
INSERT INTO evaluations VALUES(10025,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(10025,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(10025,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(10025,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(10025,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(10025,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(10025,36,'cp 36','d2d4 e5d4 e4e5 f6d5 c3d4 c5b4 c1d2 b4d2 b1d2 d7d6 e1g1 e8g8 f1e1 d5e7 d1c2 c8f5 c2c3 e7d5 c3b3 f5e6 d2e4 h7h6 a1d1 a7a5 a2a3 a5a4 b3d3 e6g4 g2g3 d6e5 b5c6 b7c6 d4e5 d8e7 h2h3');
INSERT INTO evaluations VALUES(10019,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(10019,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(10019,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(10019,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(10019,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(10019,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(10019,36,'cp 57','e5d4 e4e5 f6e4 e1g1 d7d5 c3d4 c5b6 h2h3 e8g8 b1c3 c8f5 c1e3 c6e7 b5d3 f7f6 d1c2 e4c3 b2c3 f5d3 c2d3 e7g6 a1b1 c7c6 e5f6 d8f6 a2a4 a8b8 f1e1 b6c7 e3g5 f6f5 d3f5 f8f5 e1e3 g6f4 b1e1');
INSERT INTO evaluations VALUES(10041,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(10041,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(10041,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(10041,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(10041,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(10041,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(10041,36,'cp 45','e4e5 f6d5 e1g1 c5e7 c3d4 d7d6 e5d6 d8d6 b1c3 e8g8 f1e1 c8e6 a2a3 d5b6 b5c6 b7c6 c1g5 e7g5 f3g5 e6f5 e1e5 f5g6 g5e4 d6d8 e4c5 a8b8 e5e2 b6d7 c3a4 d7c5 a4c5 d8d5 b2b4 b8d8 a1c1 f8e8 e2e8 d8e8');
INSERT INTO evaluations VALUES(10016,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(10016,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(10016,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(10016,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(10016,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(10016,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(10016,36,'cp 40','f6d5 c3d4 c5b4 b1d2 e8g8 e1g1 b4e7 a2a3 d7d6 d1c2 c8e6 d2e4 h7h6 c1d2 d8d7 h2h3 c6b4 c2a4 b4c6 a1c1 d5b6 a4c2 a7a6 b5e2 f8d8 f1e1 e6f5 e5d6 e7d6 e2d3 f5g6 d2c3 d6f8 c1d1 a8c8 c2c1 b6d5 d3c4 d7f5 c4b3 d5f4 e4c5');
INSERT INTO evaluations VALUES(23440,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(23440,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(23440,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(23440,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(23440,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(23440,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(23440,36,'cp 75','d4d5 c6e5 e2e4 e7e6 b1c3 g8f6 f2f4 e5g6 d5e6 d7e6 d1d8 e8d8 c1d2 f8b4 e1c1 d8e8 g1f3 b4c3 d2c3 g6f4 c3e5 f4g6 e5c7 c8d7 f1d3 f6g4 d1d2 f7f6 h1d1 a8c8 c7d6 g4e3 d1e1 e3g4 e4e5 g4e5 d6e5 g6e5 f3e5 f6e5 e1e5 e8e7');
INSERT INTO evaluations VALUES(23420,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(23420,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(23420,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(23420,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(23420,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(23420,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(23420,36,'cp 78','c6e5 e2e4 e7e6 b1c3 g8f6 d1d4 e5g6 h2h4 e6e5 d4d3 h7h5 c1e3 f8e7 g2g3 a7a6 g1h3 b7b5 a2a4 b5b4 c3d1 d7d6 a4a5 e8g8 h3g5 f6g4 e3d2 f7f5 f2f3 g4h6 f1h3 e7g5 d2g5 d8e8 g5h6 f5e4 f3e4 c8h3 h1h3 g7h6 h3h1 e8f7');
INSERT INTO evaluations VALUES(23487,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(23487,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(23487,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(23487,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(23487,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(23487,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(23487,36,'cp 71','b1c3 e7e6 e2e4 g8f6 f2f4 e5g6 g1f3 f8b4 f1d3 e8g8 e1g1 d7d6 d5e6 f7e6 d1e1 c8d7 a2a3 b4c5 g1h1 d8e8 e1e2 e6e5 f4f5 g6f4 c1f4 e5f4 a1e1 g8h8 e4e5 d6e5 f3e5 c5d6 f1f4 e8e5 e2e5 d6e5 e1e5 f8e8 e5e2 e8e2 c3e2 a8e8 d3c4 b7b5');
INSERT INTO evaluations VALUES(57013,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(57013,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57013,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(57013,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57013,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(57013,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57013,36,'cp 0','c7c5 e2e3 d8b6 b1c3 c5d4 e3d4 b6b2 c3b5 f8b4 e1e2 b8c6 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(51958,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51958,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51958,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51958,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51958,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51958,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51958,36,'cp 0','e2e3 d8b6 b1c3 c5d4 e3d4 b6b2 c3b5 f8b4 e1e2 b8c6 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(51956,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51956,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51956,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51956,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51956,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51956,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51956,36,'cp 0','d8b6 b1c3 c5d4 e3d4 b6b2 c3b5 f8b4 e1e2 b8c6 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(22767,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(22767,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(22767,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(22767,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(22767,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(22767,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(22767,36,'cp 86','d4d5 c6e7 g1e2 d7d6 e2g3 g7g6 f1d3 g8f6 c2c4 f8g7 b1c3 e8g8 h2h4 h7h5 c1g5 f6h7 g5e3 c7c6 a2a4 c8d7 f2f3 c6d5 c4d5 f7f5 e4f5 e7f5 d3f5 d7f5 d1b3 h7f6 e3g5 d8b6 b3b6 a7b6 g3f5 g6f5');
INSERT INTO evaluations VALUES(22658,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(22658,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(22658,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(22658,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(22658,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(22658,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(22658,36,'cp 68','c6e7 g1e2 d7d6 e2g3 g7g6 f1d3 f8g7 c2c4 g8f6 h2h4 h7h5 b1c3 e8g8 f2f3 c7c6 a2a4 f6d7 d3e2 d7c5 a4a5 f7f5 c1g5 f5f4 g3f1 g7f6 b2b4 f6g5 h4g5 c5a6 c4c5 a6b4 d1b3 d6c5 d5d6 g8g7 d6e7 d8e7 e2c4 e7g5 f1d2 g5g2');
INSERT INTO evaluations VALUES(23375,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(23375,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(23375,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(23375,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(23375,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(23375,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(23375,36,'cp 70','g1e2 d7d6 e2g3 g7g6 f1d3 g8f6 c2c4 f8g7 b1c3 e8g8 f2f3 f6d7 c1e3 f7f5 e3f2 d7f6 d1c2 f5e4 c3e4 c7c6 e4f6 g7f6 g3e4 e7f5 d5c6 b7c6 e4f6 d8f6 h2h4 f5g7 c2d2 c8f5 d3f5 g7f5 h4h5 a8d8 h5g6 f6g6 e1c1 f5g3');
INSERT INTO evaluations VALUES(50336,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(50336,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50336,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(50336,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(50336,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(50336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50336,36,'cp 0','b7b6 g1f3 c8b7 f1d3 d7d5 e1g1 e7e6 c2c4 f8d6 b2b3 e8g8 c4d5 e6d5 c1a3 d6a3 b1a3 d8e7 a3b5 c7c5 a1c1 b8c6 b5c3 a8c8 f1e1 f8d8 d3f5 c8b8 f5d3 b8c8');
INSERT INTO evaluations VALUES(43877,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43877,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43877,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43877,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43877,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43877,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(43877,36,'cp 5','f1d3 d7d5 g1f3 c8b7 c2c4 e7e6 b1c3 d5c4 d3c4 b8d7 e1g1 f8d6 f1e1 c7c5 e3e4 c5d4 d1d4 d6c5 d4d3 f6g4 e1f1 g4e5 f3e5 d7e5 c4b5 e5c6 d3g3 e8g8 f1d1 d8e7 c1h6 f7f6 h6e3 c5e3 g3e3 a8d8 f2f3 g8h8');
INSERT INTO evaluations VALUES(53390,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(53390,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53390,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(53390,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53390,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53390,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(53390,36,'cp 13','c2c3 b8c6 g1f3 c8g4 h2h3 g4h5 c1f4 e7e6 b1d2 f8d6 f4d6 d8d6 f1d3 g8f6 e1g1 e8g8 f1e1 a8c8 d1b3 d6c7 f3e5 c6e5 e1e5 c7d6 e5e3 c8c7 d2f3 h5f3 e3f3 f6d7 b3c2 g7g6 c2d2 g8g7 a1e1 f8c8');
INSERT INTO evaluations VALUES(53380,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53380,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53380,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(53380,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(53380,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53380,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53380,36,'cp 16','b8c6 g1f3 c8g4 h2h3 g4h5 d1b3 d8d7 f3e5 c6e5 d4e5 a7a6 c3c4 d5c4 f1c4 e7e6 e1g1 b7b5 c4d3 a8d8 d3e4 f8c5 c1g5 g8e7 b3g3 e8g8 b1d2 h5g6 e4g6 h7g6 d2e4 e7f5 e4c5 f5g3 c5d7 g3e2 g1h2 d8d7 f1d1 e2d4 a1c1 f7f6 e5f6 g7f6 g5e3 e6e5 e3d4 d7d4 d1d4');
INSERT INTO evaluations VALUES(19855,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(19855,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(19855,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(19855,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(19855,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(19855,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(19855,36,'cp 13','f1d3 g8f6 g1f3 c8g4 h2h3 g4h5 c1f4 e7e6 b1d2 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 d6c7 d1e2 c7b6 a1b1 c6e7 e2e3 h5g6 f3e5 g6d3 e3d3 f8c8 e5g4 f6e8 h3h4 e7f5 h4h5 h7h6 d2f3 e8d6 g4e5 d6e4 e1e2 f5d6 b1e1 c8c7 d3e3');
INSERT INTO evaluations VALUES(19856,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19856,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19856,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(19856,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(19856,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(19856,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(19856,36,'cp 15','g8f6 c1f4 c8g4 d1b3 d8d7 b1d2 e7e6 g1f3 f8d6 f4d6 d7d6 e1g1 e8g8 f1e1 g4f5 d3f5 e6f5 f3e5 d6c7 e5c6 b7c6 b3c2 g7g6 d2b3 f6d7 b3c1 c6c5 d4c5 d7c5 c1b3 f8e8 b3c5 c7c5 c2d2 e8e4 a1d1 c5c4 e1e4 d5e4 d2d4 c4a2');
INSERT INTO evaluations VALUES(13162,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(13162,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(13162,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13162,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(13162,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13162,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(13162,36,'cp 19','c1f4 c8g4 d1b3 c6a5 b3a4 a5c6 b1d2 e7e6 d3b5 d8b6 g1f3 g4f3 d2f3 a7a6 b5c6 b6c6 a4c2 a6a5 a2a4 h7h6 f3e5 c6a6 c2e2 f8d6 f2f3 a6e2 e1e2 d6e5 f4e5 f6d7 e5f4 a8c8 h2h4 f7f6 h1c1 e8f7');
INSERT INTO evaluations VALUES(13137,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(13137,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13137,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13137,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(13137,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13137,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13137,36,'cp 18','c8g4 d1b3 c6a5 b3a4 a5c6 b1d2 e7e6 h2h3 g4h5 g1f3 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 a7a6 f3e5 f8c8 a2a3 b7b5 a4b3 a8b8 e5c6 d6c6 b3c2 h5g6 d2f3 g6d3 c2d3 a6a5 f3e5 c6b7 d3e3 f6e4 e5d3 h7h6 e1e2 c8c6 a1e1 b8c8 h3h4 b7b6 g2g4 b6c7 d3e5');
INSERT INTO evaluations VALUES(24632,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(24632,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24632,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24632,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(24632,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(24632,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24632,36,'cp 18','d1b3 c6a5 b3a4 a5c6 b1d2 e7e6 h2h3 g4h5 g1f3 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 f8c8 f3e5 a7a6 a2a3 b7b5 a4b3 a8b8 e5c6 d6c6 b3c2 h5g6 d2f3 g6d3 c2d3 f6e4 f3e5 c6c7 d3e3 e4d6 e5d3 a6a5 e1e2 h7h6 h3h4 d6f5 e3h3 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(24628,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(24628,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(24628,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(24628,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(24628,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24628,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24628,36,'cp 24','c6a5 b3a4 a5c6 b1d2 e7e6 g1f3 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 c6e7 f3h4 e7g6 h4g6 h7g6 a4c2 f8c8 a2a4 c8c7 h2h3 g4f5 d2f3 a7a5 f3e5 f5d3 c2d3 d6a6 h3h4 a6d3 e5d3 f6e4 g2g3 a8c8 g1g2 g8f8 e1e3 e4d6 g3g4 d6c4');
INSERT INTO evaluations VALUES(52526,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(52526,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(52526,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(52526,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(52526,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(52526,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(52526,36,'cp -29','d4c3 b1c3 d7d6 g1f3 b8c6 f1c4 a7a6 e1g1 e7e6 d1e2 f8e7 f1d1 c8d7 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 a8c8 h2h3 e8g8 a2a3 d8e8 b2b4 d7e6 b4b5 a6b5 e2b5 c6d8 a3a4 e8b5 a4b5 e6d5 d1d5 f7f5 e4f5 f8f5');
INSERT INTO evaluations VALUES(52535,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(52535,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(52535,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(52535,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52535,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(52535,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(52535,36,'cp -15','b1c3 b8c6 g1f3 e7e6 f1e2 d7d6 e1g1 g8f6 d1b3 f8e7 f1d1 e6e5 c1g5 e8g8 g5f6 e7f6 b3a3 f6e7 c3d5 c8g4 d5e7 d8e7 a3d6 e7d6 d1d6 g4f3 e2f3 f8d8 d6d5 g7g6 g1f1 d8d5 e4d5 c6d4 a1e1 d4f3 g2f3 f7f6 e1c1 a8d8 c1c7 d8d5');
INSERT INTO evaluations VALUES(52530,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(52530,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(52530,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(52530,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(52530,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(52530,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(52530,36,'cp -22','e7e6 g1f3 b8c6 f1e2 g8f6 e1g1 d7d6 d1b3 f8e7 f1d1 e6e5 c1e3 e8g8 a1c1 h7h6 e3c5 a7a6 b3a3 c8e6 c5d6 e7d6 d1d6 d8e7 c3d5 e6d5 e4d5 c6d4 f3d4 e5d4 e2f1 f8e8 c1d1 f6e4 d6b6 e7a3 b2a3 e4c3 d1d4 e8e1 f2f4 a8e8');
INSERT INTO evaluations VALUES(52083,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(52083,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(52083,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(52083,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(52083,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(52083,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(52083,36,'cp -13','g1f3 b8c6 f1e2 d7d6 e1g1 g8f6 d1b3 f8e7 f1d1 e6e5 c1e3 e8g8 a1c1 a8b8 c3d5 c8e6 b3a4 h7h6 h2h3 f8e8 e2b5 a7a6 b5c6 b7c6 d5f6 e7f6 a4c6 b8b2 c6a6 d8a8 a6a8 e8a8 d1d6 b2a2 d6c6 g8h7 f3d2 a2a4 c6c2 h7g6 c2c6');
INSERT INTO evaluations VALUES(52082,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(52082,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(52082,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(52082,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(52082,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(52082,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(52082,36,'cp -32','b8c6 f1e2 d7d6 e1g1 g8f6 d1b3 f8e7 f1d1 e6e5 c1g5 e8g8 g5f6 e7f6 a1c1 h7h6 h2h3 a8b8 b3a3 f6e7 c3d5 e7g5 c1c2 c8e6 b2b4 b8c8 b4b5 c6e7 c2c8 e7c8 f3g5 d8g5 d1d3 f8d8 e2g4 e6g4 h3g4 g5g4');
INSERT INTO evaluations VALUES(18927,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(18927,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(18927,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(18927,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(18927,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(18927,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(18927,36,'cp -13','f1e2 d7d6 e1g1 g8f6 d1b3 f8e7 f1d1 e6e5 c1e3 e8g8 a1c1 h7h6 h2h3 a8b8 c3d5 c8e6 b3a4 f8e8 e2b5 a7a6 b5c6 b7c6 d5f6 e7f6 a4c6 b8b2 c6a6 b2a2 a6d6 d8a5 d6b6 a5a4 b6c6 e8a8 c6a4 a2a4 f3d2 g8h7 c1c6 h7g6 d1c1 h6h5 d2c4 e6c4 c1c4');
INSERT INTO evaluations VALUES(41292,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(41292,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(41292,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(41292,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(41292,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(41292,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(41292,36,'cp 63','g1f3 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 b8d7 a1b1 g8h8 a2a3 f6g8 b2b4 b7b6 c3b5 f7f5 d1c2 g7h6 c1b2 g8e7 b1d1 d7f6 b5c3 f5e4 c3e4 e7f5 c4c5 f6e4 f3e4 h6g7 e4f5 f8f5 b2c1 a5b4 a3b4 d6c5 b4c5 b6c5 c2c5');
INSERT INTO evaluations VALUES(41286,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(41286,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(41286,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(41286,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(41286,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(41286,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(41286,36,'cp 50','e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 b8d7 a1b1 h7h5 c1g5 d8e8 a2a3 f6h7 g5c1 g7f6 c3b5 e8d8 b2b4 f6g5 c1b2 b7b6 d1c2 h5h4 f1d1 g5f4 f3e2 a5b4 a3b4 d7f6 b1a1 h7g5 a1a8 d8a8 b5c3 a8c8 e2f1 f6h5 g1h1');
INSERT INTO evaluations VALUES(36567,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(36567,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(36567,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(36567,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(36567,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(36567,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(36567,36,'cp 57','f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 b8d7 a1b1 h7h5 c1g5 d8c8 a2a3 f6h7 g5c1 g7f6 b2b4 f6g5 c1b2 c8d8 c3b5 b7b6 d1c2 g5h6 f3e2 h6f4 b1a1 h5h4 g1h1 d7f6 c4c5 a5b4 a3b4 a8a1 f1a1 h7g5 e2d3 b6c5 b4c5 d6c5 a1a7 d8b8 a7c7');
INSERT INTO evaluations VALUES(15302,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(15302,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(15302,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(15302,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(15302,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(15302,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(15302,36,'cp 88','d2d4 c6d4 c2c3 b7b5 c4d3 h7h6 g5f7 e8f7 c3d4 f8b4 c1d2 b4d2 b1d2 d5f4 d3e4 d8d4 e1g1 a8b8 d1c2 c8g4 h2h3 g4e2 d2f3 e2f3 e4f3 h8d8 c2c7 f7g8 a1e1 d8e8 b2b3 d4b6 c7b6 b8b6 e1e3 b6d6 f1c1 d6d2 a2a3 d2a2 c1c5 a2a1 g1h2 a1a3 c5b5 a3a2 h2g3 a2d2 b5e5 e8e5 e3e5');
INSERT INTO evaluations VALUES(15303,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(15303,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(15303,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(15303,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(15303,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(15303,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(15303,36,'cp 85','c6d4 c2c3 b7b5 c4d3 h7h6 g5f7 e8f7 c3d4 f8b4 c1d2 b4d2 b1d2 d5f4 d3e4 d8d4 e1g1 a8b8 d1c2 c8g4 h2h3 g4e2 d2f3 e2f3 e4f3 h8d8 a1d1 f4d3 f3h5 f7g8 h5g6 e5e4 c2c7 d4b6 c7f7 g8h8 g6e4 d8f8 f7g6 b6g6 e4g6 d3b2 d1d7 a7a5 h3h4 f8d8 d7a7 a5a4 h4h5 b5b4 f1e1 b4b3 a2b3 a4b3');
INSERT INTO evaluations VALUES(15847,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(15847,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(15847,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(15847,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(15847,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(15847,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(15847,36,'cp 77','c2c3 b7b5 c4d3 h7h6 g5f7 e8f7 c3d4 f8b4 c1d2 b4d2 b1d2 d5f4 d3e4 d8d4 e1g1 a8b8 d1c2 c8g4 h2h3 g4e2 d2f3 e2f3 e4f3 h8d8 a1d1 f4d3 f3h5 f7g8 h5g6 e5e4 c2c7 d4c4 c7g3 c4e6 f2f3 d3f4 g3f4 e6g6 f3e4 d8f8 f4e3 f8f1 d1f1 a7a5 e4e5 g6e6 a2a3 e6b6 e3b6 b8b6 f1c1 b5b4 c1c8 g8f7 c8c7 f7g6 a3b4 a5b4 g1h1');
INSERT INTO evaluations VALUES(15845,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(15845,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(15845,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(15845,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(15845,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(15845,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(15845,36,'cp 57','b7b5 c4d3 h7h6 g5f7 e8f7 c3d4 f8b4 c1d2 d5f4 d3e4 b4d2 b1d2 d8d4 e1g1 a8b8 d1c2 c8g4 h2h3 g4e2 d2f3 e2f3 e4f3 h8d8 c2c7 f7g8 a1d1 f4d3 d1d2 d4c5 c7c5 d3c5 f1d1 d8d2 d1d2 g8f8 f3h5 c5a4 d2e2 a7a6 g2g3 b8b6 h3h4 e5e4 h5g4 b6d6 g4f5 e4e3 f2e3');
INSERT INTO evaluations VALUES(57022,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(57022,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(57022,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(57022,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(57022,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(57022,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(57022,36,'cp 15','d7d5 c2c4 g8f6 b1c3 f8b4 c4d5 e6d5 c1g5 e8g8 e2e3 h7h6 g5h4 c8f5 d1b3 b4c3 b2c3 b8d7 f1e2 c7c5 d4c5 d7c5 b3b4 a8c8 e1g1 g7g5 h4g3 f8e8 a1d1 c5e4 f3d4 f5d7 b4b7 e4c3 d1c1 c3e2 d4e2 d8b6 b7b6 a7b6 e2c3 d7c6 a2a3 f6e4 c3e4 d5e4');
INSERT INTO evaluations VALUES(52705,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52705,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52705,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52705,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(52705,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52705,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52705,36,'cp 16','c6d5 c2c3 b8c6 f1d3 g8f6 g1f3 c8g4 h2h3 g4h5 c1f4 e7e6 e1g1 f8d6 f4d6 d8d6 b1d2 e8g8 d1e1 d6f4 e1e3 f4e3 f2e3 h5g6 d3g6 h7g6 f1e1 a8c8 e3e4 d5e4 d2e4 b7b6 e4d6 c8d8 d6c4 c6e7 c4e5 e7f5 g1f2 f5d6 a2a3 f6e4 f2g1');
INSERT INTO evaluations VALUES(53012,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(53012,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(53012,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53012,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(53012,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(53012,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(53012,36,'cp 0','g1f3 g8f6 b1c3 e7e6 d2d4 c5d4 f3d4 a7a6 a2a4 f8e7 e1g1 e8g8 c1e3 b8c6 c4b3 d8c7 d1e2 c8d7 g1h1 a8c8 a1d1 c6a5 b3a2 a5c4 e3c1 c7c5 d4b3 c5c7 b3d4');
INSERT INTO evaluations VALUES(20981,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(20981,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(20981,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(20981,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(20981,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(20981,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(20981,36,'cp 32','g8f6 e2e3 c8g4 c2c4 e7e6 b1c3 f8b4 a1c1 e8g8 f1e2 g4f3 e2f3 d5c4 f4g5 c6a5 d1a4 c7c5 e1g1 d8e8 d4c5 b4c3 a4e8 f8e8 c1c3 a8c8 g5f6 g7f6 b2b4 c4b3 a2b3 a5c6 b3b4 a7a6 f1a1 g8g7 f3c6 c8c6 c3c2 e8d8 g1f1 f6f5 f1e2 d8d5');
INSERT INTO evaluations VALUES(14232,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(14232,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(14232,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14232,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14232,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14232,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(14232,36,'cp 41','c2c4 d5c4 b1c3 f6d5 e2e3 e7e6 f1c4 f8b4 e1g1 d5f4 e3f4 e8g8 c4d3 d8f6 c3e2 b4a5 d1c2 g7g6 d3e4 f8d8 g2g3 c8d7 f3e5 a5b6 f1d1 c6b4 c2b3 b4d5 e4d5 e6d5 b3d5 d7c6 d5b3 c6d5 b3e3 c7c6 e2c3 f6f5 c3d5 d8d5 g1h1');
INSERT INTO evaluations VALUES(14204,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(14204,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(14204,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14204,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(14204,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(14204,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(14204,36,'cp 27','d5c4 b1c3 a7a6 e2e4 b7b5 d4d5 e7e5 f4g3 c6a5 f3e5 f8c5 f1e2 e8g8 e1g1 d8e7 f1e1 c5d6 e5f3 f6e4 c3e4 e7e4 g3d6 c7d6 d1d2 a5b7 b2b3 e4g6 b3c4 b5c4 f3d4 c8d7 e2c4 b7c5 d2f4 a8e8 d4c6 c5d3 c4d3 g6d3 a1c1 e8e1 c1e1 d3d5');
INSERT INTO evaluations VALUES(14377,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14377,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(14377,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(14377,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(14377,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(14377,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(14377,36,'cp 13','e2e3 e7e5 f4e5 f8b4 f3d2 b7b5 e5f6 d8f6 f1e2 c8f5 e1g1 e8g8 a2a3 b4d2 b1d2 a8c8 a1c1 c8d8 b2b3 c6d4 e3d4 f6d4 d2c4 d4f6 d1e1 b5c4 e2c4 f5d3 c4d3 d8d3 c1c7 d3b3 c7a7 g7g6 e1c1 f6d4 a7c7 b3d3 h2h3 f8d8 c1a1 d4a1 f1a1 d3d1 a1d1 d8d1 g1h2 d1d2 a3a4 d2f2 a4a5 f2a2');
INSERT INTO evaluations VALUES(14376,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(14376,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(14376,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(14376,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(14376,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14376,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14376,36,'cp 21','e7e5 f4e5 f8b4 f3d2 b7b5 e5f6 d8f6 a2a3 b4d2 b1d2 e8g8 f1e2 c8f5 e1g1 a8c8 a1c1 c8d8 b2b3 c6d4 e3d4 f6d4 d2c4 d4f6 c4d2 f6b2 e2b5 d8d2 d1f3 d2c2 c1c2 f5c2 b5c4 b2a3 f3c3 c2f5 c3e5 f5e6 c4e6 f7e6 e5e6 g8h8 e6c4 a7a6 c4c3 a3d6 f1e1 h7h6 g2g3 d6b6');
INSERT INTO evaluations VALUES(53055,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(53055,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53055,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(53055,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(53055,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53055,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(53055,36,'cp 32','e7e5 f1c4 f8e7 d2d3 b8c6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 e8g8 c2c3 a8b8 e1g1 c8d7 b2b4 c5b4 c3b4 a7a6 d1b3 d7g4 f3d2 f6g5 d2b1 d8e7 b1c3 g4e6 d5c4 f8c8 c4e6 f7e6 c3e2 h7h6 b4b5 a6b5 b3b5');
INSERT INTO evaluations VALUES(57034,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(57034,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(57034,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(57034,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(57034,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(57034,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(57034,36,'cp 27','c7c5 g1f3 a7a6 d2d4 c5d4 f3d4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 g8e7 a2a4 b5b4 c3a2 a6a5 d3b5 d8c7 c1d2 e7g6 c2c3 b4c3 a2c3 f8d6 g2g3 e8g8 a1c1 c7b7 f1d1 d6b4 h2h4 h7h6 b5c6 d7c6 h4h5 g6e5 d2f4 f7f6 f4e5 f6e5');
INSERT INTO evaluations VALUES(51981,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(51981,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(51981,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(51981,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(51981,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(51981,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(51981,36,'cp 29','g1e2 a7a6 d2d4 c5d4 e2d4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 g8e7 c1e3 b5b4 c3b1 e7g6 b1d2 f8e7 f2f4 e8g8 d2c4 d7d5 e4d5 e6d5 c4e5 g6e5 f4e5 c6b5 d3b5 a6b5 e2b5 d8c7 b5d5 a8a5 d5d3 a5e5 a1d1 f8e8 g1h1');
INSERT INTO evaluations VALUES(51982,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51982,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(51982,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(51982,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(51982,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(51982,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(51982,36,'cp 26','a7a6 d2d4 c5d4 e2d4 b7b5 f1d3 c8b7 d1e2 b5b4 c3a4 g8f6 e1g1 b8c6 d4b3 c6e5 a4c5 b7c6 c1f4 e5d3 c2d3 f8c5 b3c5 d7d6 c5b3 e6e5 f4g5 h7h6 g5f6 d8f6 b3a5 c6b5 e2d2 e8g8 a5c4 f6e6 d2b4 d6d5 e4d5 e6d5 c4e3');
INSERT INTO evaluations VALUES(57749,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(57749,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(57749,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(57749,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(57749,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(57749,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(57749,36,'cp -29','e7e5 c1b2 f8b4 b2e5 g8f6 c2c3 b4e7 e5f6 e7f6 d2d4 d7d5 e2e3 c7c5 b1d2 e8g8 f1d3 f6e7 d4c5 b8d7 d2b3 e7c5 d1c2 g7g6 b3c5 d7c5 g1e2 b7b6 e1g1 c8b7 e2f4 d8c7 d3e2 f8d8 c2b2 c5e4');
INSERT INTO evaluations VALUES(57115,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(57115,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(57115,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(57115,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(57115,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(57115,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(57115,36,'cp -27','c1b2 f8b4 b2e5 g8f6 e2e3 d7d5 c2c4 e8g8 g1f3 c7c5 c4d5 f6d5 d1c2 b8c6 e5b2 f8e8 f1d3 h7h6 e1g1 c8g4 d3e4 a8c8 h2h3 g4e6 e4h7 g8h8 h7f5 e6f5 c2f5 h8g8 a2a3 b4a5 f1c1 b7b6 f5g4 d5f6 g4f5 f6e4 f5g4 e4g5 f3g5 d8g5 g4g5 h6g5 b1c3');
INSERT INTO evaluations VALUES(57116,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(57116,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(57116,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(57116,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(57116,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(57116,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(57116,36,'cp -25','f8b4 b2e5 g8f6 e2e3 e8g8 c2c3 b4e7 d2d4 d7d5 e5f6 e7f6 f1d3 c7c5 g1e2 f6e7 e1g1 c5c4 d3c2 b8d7 b1d2 d7f6 a2a4 e7d6 e2g3 d8a5 g3h5 d6e7 h5f6 e7f6 d2b1 f8e8 d1f3 c8d7 f3f4 a8d8 h2h4 b7b5 h4h5');
INSERT INTO evaluations VALUES(39349,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(39349,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(39349,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(39349,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(39349,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(39349,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(39349,36,'cp -37','b2e5 g8f6 e2e3 d7d5 e5f6 d8f6 c2c3 b4d6 d2d4 e8g8 g1f3 f8d8 f1e2 c7c5 d4c5 d6c5 f3d4 b8c6 b1d2 b7b6 e1g1 h7h6 a2a4 c6a5 h2h3 c8b7 d2b3 a5c4 a4a5 a8c8 g1h1 g7g6 h1g1 c8c7 a5b6 c5b6');
INSERT INTO evaluations VALUES(39315,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(39315,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(39315,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(39315,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(39315,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(39315,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(39315,36,'cp -36','g8f6 e2e3 d7d5 e5f6 d8f6 c2c3 b4d6 d1b3 c7c6 d2d4 b7b5 a2a4 b5a4 b3a4 c8f5 g1e2 e8g8 e2g3 f5e6 f1d3 c6c5 d4c5 d6c5 g3h5 f6e7 h5f4 b8d7 b1d2 a7a5 d2b3 c5d6 e1g1 d7e5 d3b5 f8c8 a4d4 e5g4 h2h3 g4f6 f4e6 f7e6');
INSERT INTO evaluations VALUES(42706,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(42706,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(42706,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(42706,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(42706,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(42706,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(42706,36,'cp -20','c2c3 b4e7 e5f6 e7f6 d2d4 c7c5 e2e3 d7d5 b1d2 b8c6 g2g3 e8g8 f1g2 f8e8 g1e2 c5d4 c3d4 f6d4 e2d4 c6d4 e1g1 d4c6 d2b3 c8e6 b3c5 d8e7 a1c1 a8d8 d1a4 d8d6 a4b3 e6c8 g2d5 c6a5 b3d3 b7b6 c5b3 e8d8 e3e4 a5c6 d3c3 c8b7 f1e1 h7h6 b3d4 c6d4 c3d4 g8h8 e4e5');
INSERT INTO evaluations VALUES(42703,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(42703,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(42703,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(42703,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(42703,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(42703,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(42703,36,'cp -30','b4e7 e5f6 e7f6 e2e3 c7c5 d2d4 d7d5 b1d2 e8g8 f1d3 c5c4 d3c2 b8c6 g1e2 a8b8 a2a4 a7a6 e1g1 b7b5 e2f4 f6g5 a4b5 a6b5 d1f3 g5f4 f3f4 b8b6 a1a8 b5b4 c2f5 c6e7 f5c8 e7c8 c3b4 b6b4 d2f3 d8b6 h2h3 b6b7 f1a1 b4b6 g1h1 c8d6');
INSERT INTO evaluations VALUES(57759,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(57759,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(57759,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(57759,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(57759,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(57759,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(57759,36,'cp -12','d7d5 e2e4 d5e4 d3e4 d8d1 e1d1 b8c6 g1f3 g8f6 f1d3 e7e5 b1d2 f6d7 d2b3 d7c5 b3c5 f8c5 d1e2 f7f6 d3b5 c8d7 c2c3 e8c8 b2b4 c5b6 c1e3 b6e3 e2e3 a7a5 b5c6 d7c6 b4a5 c6b5 h1d1 h8e8 a1c1 e8g8 d1d8 g8d8 f3d2 b5a6');
INSERT INTO evaluations VALUES(55127,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(55127,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(55127,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(55127,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(55127,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(55127,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(55127,36,'cp 0','e2e4 d5e4 d3e4 d8d1 e1d1 b8c6 g1f3 g8f6 b1d2 e7e6 f1b5 c8d7 b5d3 e8c8 d1e2 d7e8 a2a4 f6d7 d2b3 c6b4 c1e3 a7a6 h1c1 h7h6 h2h4 f8e7 a4a5 b4c6 g2g3 e7f6 a1b1 g7g5 h4g5 h6g5 e3g5 f6g5 f3g5 d7e5 f2f3');
INSERT INTO evaluations VALUES(55076,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55076,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(55076,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(55076,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(55076,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(55076,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(55076,36,'cp -8','d5e4 d3e4 d8d1 e1d1 b8c6 g1f3 g8f6 b1d2 h8g8 c2c3 g7g5 f3d4 c6d4 c3d4 c8e6 f2f3 e8c8 d4d5 e6d7 d2b3 g5g4 c1e3 b7b6 a1c1 e7e6 d5e6 f7e6 e3f4 f6e8 d1c2 e8d6 f1a6 c8b8 h1d1 g4f3 g2f3 f8g7 f4d6 c7d6 d1d6 g7e5');
INSERT INTO evaluations VALUES(55287,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55287,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(55287,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55287,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55287,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(55287,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55287,36,'cp -6','d3e4 d8d1 e1d1 b8c6 g1f3 e7e6 c1f4 f8d6 f4g3 d6g3 h2g3 a7a6 b1d2 g8f6 f1d3 e8e7 d2c4 b7b6 f3e5 c6e5 c4e5 f6d7 e5d7 c8d7 d1e2 a6a5 e2e3 h7h6 f2f3 e6e5 d3c4 a8d8 a1d1 f7f6 b2b3 d7e6 d1d8 h8d8 c4e6');
INSERT INTO evaluations VALUES(55279,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(55279,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(55279,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(55279,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(55279,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(55279,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(55279,36,'cp -2','d8d1 e1d1 b8c6 g1f3 e7e6 c1f4 f8d6 f4g3 d6g3 h2g3 g8f6 b1d2 a7a6 f1d3 b7b6 d1e2 h7h5 h1h4 g7g6 d2c4 c8b7 c4e3 e8e7 a1h1 a8d8 g3g4 c6d4 f3d4 d8d4 f2f3 d4a4 e4e5 f6d7 e3c4 b7d5 g4h5 h8h5 h4h5 g6h5 c4e3 d7e5');
INSERT INTO evaluations VALUES(34657,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(34657,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(34657,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(34657,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(34657,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(34657,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34657,36,'cp -11','e1d1 b8c6 g1f3 g8f6 f1d3 e7e5 b1d2 f6d7 a2a4 a7a5 d3b5 f7f6 d2b3 d7c5 b3c5 f8c5 d1e2 e8f7 b5c4 f7e7 c2c3 c6d8 f3e1 c8e6 c4e6 d8e6 f2f3 h7h5 e1d3 h8d8 g2g3 d8d6 d3c5 e6c5 c1e3 c5d3 h1b1 a8d8 b2b4');
INSERT INTO evaluations VALUES(34656,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(34656,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(34656,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(34656,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(34656,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(34656,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(34656,36,'cp -8','b8c6 g1f3 g8f6 f1d3 e7e5 b1d2 f6d7 a2a4 a7a5 c2c3 d7c5 d3c4 h7h5 f3e1 h5h4 f2f3 f7f6 d2b3 c8e6 c4e6 c5e6 d1c2 b7b6 e1d3 e8c8 c1e3 g7g6 b3d2 f8h6 e3h6 h8h6 d2c4 e6f4 d3f4 e5f4 h2h3 g6g5 a1d1');
INSERT INTO evaluations VALUES(1637,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(1637,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(1637,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(1637,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(1637,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(1637,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(1637,36,'cp -8','g1f3 g8f6 f1d3 e7e5 b1d2 f6d7 a2a4 a7a5 c2c3 d7c5 d3c4 f7f6 f3e1 c6d8 e1d3 d8f7 f2f3 h7h5 d1c2 h5h4 h2h3 c5d3 c4d3 f8c5 d2c4 c8d7 c1e3 c5e3 c4e3 f7d6 e3c4 e8e7 c4d6 e7d6');
INSERT INTO evaluations VALUES(1636,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(1636,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(1636,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(1636,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(1636,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(1636,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(1636,36,'cp -9','g8f6 f1d3 e7e5 b1d2 f6d7 a2a4 a7a5 c2c3 f7f6 f3e1 d7c5 d3c4 c6d8 f2f3 h7h5 h2h4 d8f7 d1c2 f7d6 c4e2 c8e6 e1d3 c5d3 e2d3 e8c8 d2f1 b7b6 f1e3 c8b7 b2b3 g7g6 b3b4 f6f5 b4a5 f5e4 f3e4');
INSERT INTO evaluations VALUES(47174,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(47174,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(47174,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(47174,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(47174,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(47174,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(47174,36,'cp -22','c7c5 d4c5 b8c6 e2e3 e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3a4 c5d6 c2c3 e8g8 h2h3 c8d7 a2a3 f8e8 f1e1 a7a6 c3d4 c6d4 a4c3 d7c6 c1g5 d6c5 g5h4 h7h6 f3d4 c5d4 d1e2 a8c8 a1c1 d8d6 e1d1 b7b5 d3b1 b5b4');
INSERT INTO evaluations VALUES(45371,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(45371,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(45371,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(45371,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(45371,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(45371,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(45371,36,'cp -21','d4c5 b8c6 e2e3 e7e6 f1d3 f8c5 e1g1 e8g8 e3e4 d5d4 c3e2 e6e5 a2a3 a7a5 h2h3 h7h6 c1d2 a5a4 c2c3 d4c3 d2c3 d8d6 d3c2 f8d8 d1d6 c5d6 a1d1 c8e6 e2g3 g7g6 d1d2 d8d7 f1d1 d6c7 g3e2 d7d2 d1d2');
INSERT INTO evaluations VALUES(45301,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(45301,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(45301,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(45301,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(45301,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(45301,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(45301,36,'cp -20','b8c6 e2e3 e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 c2c3 d4c3 e2c3 c8e6 c1g5 h7h6 g5h4 a8c8 a1c1 g7g5 h4g3 f6h5 f3e5 h5g3 e5c6 c8c6 h2g3 c6d6 c3d5 e6d5 e4d5 d6d5 d1h5 g8g7 d3c4 d5e5 h5f3 d8f6 f3f6 g7f6');
INSERT INTO evaluations VALUES(13075,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(13075,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(13075,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(13075,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(13075,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(13075,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(13075,36,'cp -16','e2e3 e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3a4 c5d6 c2c3 c8g4 d3b5 e8g8 b5c6 b7c6 h2h3 g4f3 d1f3 h7h6 b2b3 f8e8 c1d2 e8e6 a4b2 d8e7 f1e1 f6d7 b2d3 a8e8 a1c1 c6c5 c3d4 e5d4 d2f4 d6f4 f3f4 g8f8 f4f3 e6e4 f3e4 e7e4');
INSERT INTO evaluations VALUES(13074,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(13074,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(13074,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(13074,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(13074,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(13074,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(13074,36,'cp -19','e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3a4 c5d6 c2c3 c8g4 d3b5 e8g8 b5c6 b7c6 h2h3 g4f3 d1f3 h7h6 b2b3 f8e8 a4b2 e8e6 c1d2 f6d7 b2d3 c6c5 a1c1 a7a5 c3d4 e5d4 d2f4 d6f4 f3f4 d8e7 f4c7 e6e4');
INSERT INTO evaluations VALUES(55969,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55969,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55969,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55969,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55969,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(55969,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55969,36,'cp 0','c7c5 c2c3 d5d4 g2g3 b8c6 d2d3 d4c3 b2c3 b7b6 f1g2 d8d7 b1d2 c8a6 d2c4 a8d8 c1f4 a6c4 d3c4 g8e7 d1d7 d8d7 f3d2 c6a5 e1c1 e7g6 d2b3 g6f4 g3f4 a5b3 a2b3 d7c7 c1d2 f8e7 h2h4 h7h5 d1g1 g7g6 d2e3 e8g8 g1d1 f8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(54316,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(54316,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(54316,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(54316,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(54316,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(54316,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(54316,36,'cp 0','c2c3 d5d4 b1a3 b8c6 f1d3 f7f5 e5f6 g8f6 e1g1 f8d6 c3d4 e8g8 d1e2 c5d4 a3c4 d6c7 c4e5 c6e5 f3e5 f6d5 e2h5 d5f6');
INSERT INTO evaluations VALUES(54311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54311,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54311,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54311,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(54311,36,'cp 12','d5d4 b1a3 b8c6 f1d3 f7f5 e5f6 g8f6 e1g1 f8d6 b2b4 c5b4 a3b5 e8g8 b5d6 d8d6 c3b4 c6b4 c1a3 a7a5 f1e1 c8d7 a1b1 d7c6 d1b3 g8h8 f3e5 d6d5 e5c6 b7c6 a3b4 a5b4 b3b4 d5a2 b4d4 f8d8 d4c3 a8a3 d3c4 a2b1 e1b1 a3c3 d2c3 f6e4 g2g3 e4c3 b1b6 c3b5 b6c6 d8d1 g1g2 b5d4');
INSERT INTO evaluations VALUES(54087,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(54087,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(54087,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54087,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(54087,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(54087,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(54087,36,'cp 0','b1a3 b8c6 f1d3 g8e7 e1g1 e7g6 f1e1 a7a6 g2g3 b7b5 d3e4 c8d7 a3c2 a8c8 c3d4 c5d4 c2d4 c6d4 f3d4 g6e5 d1e2 e5c6 d4c6 d7c6 e4c6 c8c6 e2e4 c6c4 d2d4 f8e7 c1e3 e8g8 e1d1 d8d5 e4d5 e6d5 b2b3 c4c2 a2a4 b5b4 a1c1 f8c8 c1c2');
INSERT INTO evaluations VALUES(48199,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(48199,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(48199,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(48199,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(48199,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(48199,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(48199,36,'cp 11','f8e7 e2e3 e8g8 g1f3 h7h6 g5f6 e7f6 a1c1 c7c6 f1d3 b8d7 e1g1 d5c4 d3c4 e6e5 c3e4 e5d4 e4f6 d7f6 f3d4 c6c5 d4b5 c8d7 b2b3 d7b5 c4b5 d8a5 d1e2 a8d8 b5d3 d8d7 e3e4 f8d8 d3c4 a5d2 e4e5 d2e2 c4e2');
INSERT INTO evaluations VALUES(40921,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40921,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(40921,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40921,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40921,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40921,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(40921,36,'cp 10','g1f3 e8g8 e2e3 h7h6 g5f6 e7f6 a1c1 c7c6 f1d3 b8d7 e1g1 d5c4 d3c4 e6e5 d4d5 d7b6 b2b3 c6d5 c4d5 b6d5 d1d5 d8d5 c3d5 f8d8 f1d1 c8d7 g1f1 g8f8 e3e4 a8c8 f1e2 c8c1 d1c1 d7c6 e2e3 c6d5 e4d5 f8e7 c1c7 d8d7 c7d7 e7d7 e3e4 d7d6');
INSERT INTO evaluations VALUES(40920,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(40920,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(40920,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40920,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(40920,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40920,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40920,36,'cp 0','h7h6 g5f6 e7f6 e2e3 e8g8 a1c1 c7c6 f1d3 d5c4 d3c4 b8d7 e1g1 c6c5 c3e4 c5d4 e4f6 d7f6 d1d4 b7b6 c4e2 c8b7 d4e5 a8c8 f1d1 d8e7 f3d4 e7b4 e5b5 b4e7');
INSERT INTO evaluations VALUES(40709,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40709,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(40709,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40709,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(40709,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40709,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40709,36,'cp 9','g5f6 e7f6 e2e3 e8g8 a1c1 c7c6 f1d3 b8d7 e1g1 d5c4 d3c4 e6e5 c3e4 e5d4 e4f6 d7f6 f3d4 c6c5 d4b5 c8d7 f1e1 d7b5 c4b5 d8e7 d1c2 b7b6 b5c4 a8d8 c1d1 d8d1 e1d1 f8d8 d1d8 e7d8 c4b3 d8e7 h2h3 e7e5 b3c4 e5e4 c4d3');
INSERT INTO evaluations VALUES(40707,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(40707,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(40707,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(40707,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(40707,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40707,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(40707,36,'cp 3','e7f6 e2e3 e8g8 a1c1 c7c6 f1e2 d5c4 e2c4 b8d7 e1g1 c6c5 d1e2 c5d4 e3d4 d7b6 c4d3 g7g6 f1d1 c8d7 d3e4 a8b8 e2e3 f6g7 h2h4 b6c4 e3e2 c4d6 h4h5 g6h5 e4b1 b8c8 d4d5 e6e5 c3e4 c8c1 d1c1 d6e4 e2e4 f7f5');
INSERT INTO evaluations VALUES(41640,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41640,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41640,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(41640,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(41640,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(41640,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41640,36,'cp 16','e2e3 e8g8 a1c1 c7c6 f1d3 b8d7 e1g1 d5c4 d3c4 c6c5 d4c5 d7c5 b2b4 d8d1 f1d1 c5d7 c4e2 f8d8 c3e4 f6e7 a2a3 g8f8 g2g3 a7a6 h2h4 b7b5 c1c7 d7b6 d1d8 e7d8 c7c1 c8b7 e4c5 a8c8 c1d1 b7f3 e2f3 b6c4 c5a6 c4a3 a6c5 a3c4');
INSERT INTO evaluations VALUES(41639,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(41639,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(41639,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(41639,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41639,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41639,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41639,36,'cp 7','e8g8 a1c1 c7c6 f1d3 d5c4 d3c4 b8d7 e1g1 c6c5 d4c5 d7c5 b2b4 c5d7 d1b3 d7e5 c4e2 b7b6 f3e5 f6e5 e2f3 a8b8 f1d1 d8e7 b3a4 e7h4 h2h3 c8b7 f3b7 b8b7 f2f4 e5c3 c1c3 h4f6 a4c6 b7b8 b4b5 f8d8');
INSERT INTO evaluations VALUES(36910,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36910,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36910,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36910,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36910,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(36910,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(36910,36,'cp 8','a1c1 c7c6 f1d3 d5c4 d3c4 b8d7 e1g1 c6c5 d4c5 d7c5 b2b4 d8d1 f1d1 c5d7 c4e2 f6e7 b4b5 d7c5 f3e5 f8d8 e2f3 d8d1 c1d1 f7f5 c3e2 e7f6 e5c4 a8b8 b5b6 a7a6 g2g3 c8d7 e2f4 d7b5 c4d6 b8d8 d6b5 d8d1 f3d1');
INSERT INTO evaluations VALUES(56398,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(56398,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(56398,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(56398,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(56398,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(56398,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(56398,36,'cp 127','e4d5 e5e4 d1e2 g8f6 d2d3 d8d5 b1d2 b8c6 d3e4 d5e6 h2h3 f8c5 e4e5 f6d7 d2b3 c5b6 c1f4 a7a5 c2c3 e8g8 e2e4 a5a4 f1c4 e6g6 e4g6 h7g6 b3d2 d7c5 e1g1 c8e6 f4e3 f8d8 e3c5 e6c4 d2c4 b6c5 f1d1 f7f6 e5f6 d8d1 a1d1 g7f6 d1d7');
INSERT INTO evaluations VALUES(56167,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(56167,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(56167,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(56167,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(56167,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(56167,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(56167,36,'cp 118','d8d5 b1c3 d5c5 d2d4 e5d4 f3d4 g8f6 d1e2 f8e7 c1e3 e8g8 d4b5 c5c6 e1c1 a7a6 b5a7 c6e8 a7c8 e8c8 g2g3 f8e8 h2h4 c8f5 f1g2 b8c6 c3d5 f6d5 g2d5 e7c5 e2c4 c5e3 f2e3 c6e5 c4f4 f5f4 e3f4 e5c6 d5c6 b7c6 d1d7 a8c8 b2b4 e8e2 h1d1 h7h6 a2a3 g8f8 h4h5');
INSERT INTO evaluations VALUES(34947,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(34947,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(34947,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(34947,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(34947,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(34947,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(34947,36,'cp 155','b1c3 d5c5 d2d4 e5d4 f3d4 g8f6 f1e2 c7c6 c1e3 c5e5 d1d2 f8b4 e3f4 e5c5 d4b3 c5e7 e1g1 e8g8 e2d3 f8d8 f1e1 e7f8 a2a3 b4e7 d2e3 d8e8 e3g3 b8d7 g3f3 d7c5 b3c5 e7c5 e1e8 f8e8');
INSERT INTO evaluations VALUES(34938,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(34938,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(34938,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(34938,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(34938,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(34938,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(34938,36,'cp 123','d5c5 d2d4 e5d4 f3d4 g8f6 d1e2 f8e7 c1e3 e8g8 d4b5 c5c6 e1c1 a7a6 b5a7 c6e8 a7c8 e8c8 g2g3 b8c6 f1g2 f8e8 e3f4 e7d6 e2d2 d6b4 a2a3 b4a5 d2d3 a5b6 c3d5 f6d5 g2d5 c8d7 c1b1 a8d8 f4g5 d8b8 d5g2 d7d3 d1d3 c6e5');
INSERT INTO evaluations VALUES(34887,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(34887,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(34887,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(34887,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(34887,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(34887,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(34887,36,'cp 150','d2d4 e5d4 f3d4 g8f6 f1e2 f8e7 c1e3 c5a5 d4b3 a5f5 e2d3 f5e5 d1d2 c8e6 e1c1 e6b3 c2b3 b8c6 e3f4 e5a5 c3b5 a5d2 d1d2 e8g8 h1d1 f6h5 f4e3 h5f6 b5c7 a8c8 c7b5 f6d5 d3f5 d5e3 f2e3 c8b8 f5e4 e7g5 c1b1 g5e3 d2d7 c6e5 d7b7 e5g4 h2h3 g4f2 d1e1 f2e4 b7b8 f8b8');
INSERT INTO evaluations VALUES(34886,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(34886,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(34886,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(34886,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(34886,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(34886,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(34886,36,'cp 122','e5d4 f3d4 g8f6 f1e2 f8e7 c1e3 c5a5 d4b3 a5f5 e2d3 f5e5 d1d2 c8e6 e1c1 e6b3 e3f4 e5a5 c2b3 b8c6 c3b5 a5d2 d1d2 e8g8 h1d1 f6h5 f4e3 h5f6 b5c7 a8c8 c7b5 f6d5 d3f5 d5e3 f2e3 c8b8 f5e4 e7g5 c1b1 g5e3 d2d7 c6e5 d7b7 g7g6 a2a4 b8b7 e4b7');
INSERT INTO evaluations VALUES(34890,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(34890,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(34890,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(34890,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(34890,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(34890,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(34890,36,'cp 134','f3d4 g8f6 f1e2 f8e7 c1e3 c5a5 d4b3 a5f5 e2d3 f5e5 d1d2 c8e6 e1c1 e6b3 e3f4 e5a5 c2b3 b8c6 c3b5 a5d2 d1d2 e8g8 h1d1 f6h5 f4e3 h5f6 b5c7 a8c8 c7b5 f6d5 d3f5 d5e3 f2e3 c8b8 d2d7 e7c5 f5e4 c6e5 d7b7 b8b7 e4b7 f8b8 b7a6 c5e3 c1c2 g7g6 b5c3 h7h5 c3d5 e3g5 b3b4 g8g7 c2b3');
INSERT INTO evaluations VALUES(34889,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(34889,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(34889,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(34889,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(34889,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(34889,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(34889,36,'cp 153','g8f6 f1e2 c7c6 c1e3 c5a5 d4b3 a5c7 d1d2 f8e7 e1c1 e8g8 h1e1 c8f5 e3f4 c7c8 f4g5 f8d8 e2d3 f5e6 d2e2 b8d7 b3d4 d7c5 d3f5 h7h6 d4e6 c5e6 d1d8 c8d8 g5d2 d8d6 f5e6 d6e6 e2e6 f7e6 e1e6 g8f7 e6e2 a8e8 a2a3 b7b5 h2h3');
INSERT INTO evaluations VALUES(19061,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(19061,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(19061,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(19061,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(19061,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(19061,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(19061,36,'cp 57','d4b5 a7a6 b5d6 f8d6 d1d6 d8e7 d6d1 g8f6 b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 c2c4 e8g8 f1e2 b8d7 e1g1 f8b8 a2a4 a6a5 e2d3 e5e4 d3e2 d7c5 a1a3 e7e5 d1d2 b8e8 g5e3 e5e7 e3d4 f6d7 f2f3 e8f8 f3e4 c5e4 d2c2 f7f5 a3e3 d7c5 e2d3 e7g5 f1e1 h7h6 d3e4 f5e4');
INSERT INTO evaluations VALUES(18953,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(18953,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(18953,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(18953,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(18953,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(18953,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(18953,36,'cp 38','d7d6 c2c4 g7g6 b1c3 a7a6 b5a3 c8e6 f1d3 g8f6 a3c2 a8c8 e1g1 f8g7 c1g5 h7h6 g5h4 c6a5 b2b3 b7b5 c2e3 b5c4 b3c4 a5c4 d3c4 e6c4 d1a4 d8d7 e3c4 d7a4 c4d6 e8d7 c3a4 d7d6 f2f3 c8c2 a1b1 h8c8 a4b6 c8c6 f1d1 d6e6 b6d5 g6g5 h4e1 g7f8 b1b8 f8c5 g1f1');
INSERT INTO evaluations VALUES(20586,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(20586,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(20586,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(20586,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(20586,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(20586,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(20586,36,'cp 54','c2c4 g7g6 f1d3 a7a6 b5c3 g8f6 e1g1 f8g7 c1g5 h7h6 g5e3 f6g4 e3d2 e8g8 c3d5 d8h4 h2h3 g4f6 b1c3 f6d5 c3d5 c6d4 d2e3 c8e6 a1c1 a8c8 f1e1 e6d5 c4d5 b7b5 c1c3 b5b4 c3c1 g8h7 c1c8 f8c8 e3d4 e5d4 d3f1');
INSERT INTO evaluations VALUES(20583,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(20583,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(20583,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(20583,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(20583,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(20583,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(20583,36,'cp 38','g7g6 f1d3 a7a6 b5c3 g8f6 e1g1 f8g7 c1g5 c6d4 h2h3 e8g8 b1d2 c8e6 c3d5 e6d7 f1e1 d4e6 g5e3 f6d5 c4d5 e6d4 a1c1 a8c8 c1c3 f8e8 d3f1 h7h5 d2c4 g7f8 a2a4 b7b5 a4b5 a6b5 c4d2 d8b6 d2b3 f8g7 d1d3 f7f5 b3d4 e5d4');
INSERT INTO evaluations VALUES(57010,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(57010,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(57010,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(57010,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(57010,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(57010,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(57010,36,'cp 29','b1c3 d7d5 d2d4 g8f6 g1f3 f8e7 c1f4 e8g8 e2e3 c7c5 d4c5 e7c5 c4d5 e6d5 f1e2 b8c6 e1g1 c8e6 c3a4 c5e7 f3d4 c6d4 d1d4 f6e4 f4e5 f7f6 e5f4 d8a5 e2d3 a8c8 d3e4 d5e4 d4e4 e6d5 e4e7 a5a4 e7d6');
INSERT INTO evaluations VALUES(57005,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(57005,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(57005,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(57005,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(57005,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(57005,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(57005,36,'cp 20','d7d5 d2d4 g8f6 c4d5 e6d5 c1g5 f8e7 e2e3 h7h6 g5h4 c7c6 f1d3 e8g8 d1c2 b8d7 g1f3 f8e8 e1g1 f6e4 d3e4 e7h4 e4h7 g8h8 f3h4 d8h4 h7d3 d7f6 a1e1 f6g4 h2h3 g4f6 c3e2 a7a5 e2f4 h8g8 a2a3 c8d7 e1c1 e8e7 f1e1 f6e4 d3e4 e7e4 f2f3 e4e8');
INSERT INTO evaluations VALUES(55802,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(55802,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(55802,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(55802,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(55802,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(55802,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(55802,36,'cp 19','d2d4 g8f6 c4d5 e6d5 c1g5 f8e7 e2e3 e8g8 f1d3 h7h6 g5h4 c7c6 d1c2 b8d7 g1f3 f8e8 e1g1 f6e4 d3e4 e7h4 e4h7 g8h8 f3h4 d8h4 h7d3 d7f6 a1e1 f6g4 h2h3 g4f6 c3e2 h8g8 e2f4 a7a5 e1c1 f6h5 f4e2 h5f6 f1e1 h4g5 e2f4 f6h5 f4h5 g5h5');
INSERT INTO evaluations VALUES(57652,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(57652,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(57652,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(57652,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(57652,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(57652,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57652,36,'cp 34','g8f6 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 f8g7 f1c4 c7c5 g1e2 e8g8 c1e3 b8c6 e1g1 b7b6 a1c1 e7e6 c4b5 c8b7 d4c5 c6e5 d1d8 f8d8 c5b6 a7b6 f2f4 e5g4 e3b6 d8d6 b6c5 d6d2 h2h3 g4f6 e4e5 f6e4 c5e3 d2a2 b5c4 a2a3 f1d1 g7f8 d1d7 b7a6 d7c7 a6c4 c7c4 a3a2 e2d4');
INSERT INTO evaluations VALUES(55133,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(55133,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(55133,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(55133,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55133,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(55133,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55133,36,'cp 0','d2d4 e7e6 g1f3 g8f6 b2b3 b7b6 c2c4 c8b7 b1c3 b8d7 c1b2 f8d6 f1d3 e8g8 e1g1 f8e8 a2a4 a7a6 a4a5 d8e7 f3d2 d6b4 a5b6 c7b6 c3a2 b4d6 a2c3');
INSERT INTO evaluations VALUES(21851,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(21851,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(21851,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(21851,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(21851,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(21851,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(21851,36,'cp 77','e5d4 f3d4 c8d7 b1c3 g8f6 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 d8c8 h2h3 c8e6 a1d1 a5a4 f1e1 f8b8 e3f2 e7f8 b2b3 a4b3 c2b3 f6d7 c3e2 c6c5 e2c3 h7h6 f2h4 f7f6 h4g3 g8h8 g1h1 c7c6 f3d3');
INSERT INTO evaluations VALUES(21866,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(21866,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(21866,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(21866,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(21866,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(21866,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(21866,36,'cp 75','f3d4 c8d7 b1c3 g8f6 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 d8d7 h2h3 a5a4 a1d1 d7e6 f1e1 f8d8 e3d4 e7f8 g2g4 h7h6 g1g2 c6c5 d4f2 a8b8 f4f5 e6e8 b2b3 a4b3 a2b3 c5c4 c3d5');
INSERT INTO evaluations VALUES(21863,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(21863,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21863,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(21863,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(21863,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(21863,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(21863,36,'cp 70','c8d7 b1c3 g8f6 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 d8d7 h2h3 f8b8 e4e5 f6d5 c3d5 c6d5 f3d5 c7c6 d5d4 b8b4 d4d3 b4b2 f4f5 e7f8 a2a4 b2b7 a1d1 d6d5 e3d4 a8e8 d4c3 f8b4 f5f6 g7g6 d3e3');
INSERT INTO evaluations VALUES(27223,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(27223,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(27223,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(27223,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(27223,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(27223,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(27223,36,'cp 74','b1c3 g8f6 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 b2b3 d8d7 c1b2 d7g4 f3d3 g4g6 a1e1 d6d5 f4f5 g6h5 g1h1 h5h4 g2g3 h4h3 f1f4 d5e4 c3e4 f8d8 d3e2 f6e4 f4e4 e7d6 e2g4 h3g4');
INSERT INTO evaluations VALUES(27221,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(27221,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(27221,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(27221,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(27221,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(27221,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27221,36,'cp 73','g8f6 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 a7a5 h2h3 e8g8 d1f3 f6d7 c1d2 a5a4 a1d1 f8e8 d2e1 d8c8 b2b3 a4b3 a2b3 d7f8 c3e2 e7f6 e2g3 c8e6 g1h2 c6c5 e1c3 e6e7 d1e1 f6c3 f3c3 f8e6 g3f5 e7f6 c3f6 g7f6 f5h6 g8f8');
INSERT INTO evaluations VALUES(22795,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(22795,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(22795,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(22795,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(22795,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(22795,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(22795,36,'cp -15','g8f6 c2c4 f8c5 h2h3 h7h6 f1e2 d7d6 a2a3 a7a5 c1e3 f6h5 e3d2 c6e7 f3h4 h5f6 b1c3 e7g6 h4g6 f7g6 e1g1 e8g8 d2e3 c5e3 f2e3 c8d7 d1d2 c7c6 d3d4 g6g5 d4e5 d6e5 a1d1 d7e6 d2d6 d8d6 d1d6 e6f7 f1d1 a5a4');
INSERT INTO evaluations VALUES(16549,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(16549,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(16549,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(16549,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(16549,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(16549,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(16549,36,'cp -15','c2c4 f8c5 h2h3 d7d6 b1c3 e8g8 f1e2 a7a5 e1g1 f8e8 c1e3 c6d4 d1d2 c7c6 e2d1 h7h6 a2a3 d4f3 d1f3 c8e6 f3e2 c5d4 a1e1 f6d7 e2d1 d4e3 e1e3 d8g5 g1h1 d7c5 c3e2 a5a4 f2f4 e5f4 e2f4 g5e5 d2f2 c5d7');
INSERT INTO evaluations VALUES(16433,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16433,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(16433,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(16433,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(16433,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(16433,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(16433,36,'cp -7','f8c5 h2h3 d7d6 b1c3 e8g8 f1e2 f8e8 e1g1 a7a6 a1b1 c8d7 c1g5 h7h6 b2b4 c5a7 g5e3 c6d4 a2a4 c7c6 a4a5 b7b5 a5b6 a7b6 c3a4 d4f3 e2f3 b6e3 f2e3 c6c5 a4c3 c5b4 b1b4 a8b8 d1b1 a6a5 b4b8 d8b8 b1a2');
INSERT INTO evaluations VALUES(10168,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(10168,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(10168,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(10168,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(10168,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(10168,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(10168,36,'cp -6','h2h3 d7d6 b1c3 h7h6 c1e3 e8g8 f1e2 c6d4 e1g1 a7a6 e3d4 e5d4 c3d5 f6d5 c4d5 c7c6 d5c6 b7c6 f3d2 a6a5 e2g4 c8a6 d1c2 a5a4 e4e5 c5b4 e5d6 b4d2 c2d2 d8d6 g4f3 a8b8 b2b3 a4a3 a1e1 f8c8 e1c1 c6c5 f1e1 a6b5');
INSERT INTO evaluations VALUES(10166,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(10166,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(10166,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(10166,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(10166,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(10166,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10166,36,'cp -6','e8g8 f1e2 d7d6 b1c3 a7a6 e1g1 a8b8 a2a3 c6d4 f3d4 c5d4 c1e3 b7b5 c4b5 d4e3 f2e3 a6b5 c3d5 f6d5 e4d5 d8g5 f1f3 c8d7 d1c1 f8c8 e3e4 g5e7 c1e3 c7c5 d5c6 c8c6 a1f1 d7e6 d3d4 e5d4 e3d4 c6c2 e2d1 c2c1 d4d2 c1b1 g1h2 g7g6 d1e2');
INSERT INTO evaluations VALUES(4365,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4365,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4365,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4365,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(4365,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(4365,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4365,36,'cp 0','f1e2 d7d6 e1g1 c6d4 c1e3 d4f3 e2f3 c5e3 f2e3 c7c6 a2a4 c8e6 b1c3 a7a5 b2b3 f6d7 g1h2 d8g5 d1e2 a8e8 a1b1 f7f5 e4f5 e6f5 f3e4 f5e6 e4f3');
INSERT INTO evaluations VALUES(51494,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(51494,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(51494,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(51494,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(51494,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(51494,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(51494,36,'cp 113','e2e4 c8b7 f1d3 d7d5 e4e5 d8d7 c2c3 e7e6 g1e2 g8e7 e1g1 b7a6 d3a6 b8a6 e2f4 g7g6 b1d2 c7c5 d2f3 a6b8 f4d3 c5d4 c3d4 b8c6 c1g5 e7f5 g5f6 f8g7 f6g7 f5g7 d1d2 a8c8 h2h3 h7h5 a1c1 g7f5');
INSERT INTO evaluations VALUES(57669,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(57669,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(57669,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(57669,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(57669,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(57669,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(57669,36,'cp 65','f8g7 d2d4 d7d6 f1d3 b8c6 c2c3 e7e5 h2h3 g8e7 e1g1 e8g8 d4e5 c6e5 f3e5 d6e5 d1e2 a7a5 f1d1 e7c6 a2a4 d8e7 d3c4 f8d8 c1d2 d8d7 c4d5 c6d8 d2e3 d8e6 b1d2 h7h6 e2f3 e7f6 f3f6 g7f6 d2c4 e6f4 g1f1 f4d5');
INSERT INTO evaluations VALUES(39622,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(39622,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(39622,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(39622,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(39622,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(39622,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(39622,36,'cp 64','d2d4 d7d6 c1e3 g8f6 b1c3 a7a6 a2a4 b8c6 f1e2 f6g4 e3f4 e8g8 d4d5 e7e5 f4g5 c6e7 h2h3 g4f6 d1d2 f6h5 g2g4 h5f4 g5f4 e5f4 d2f4 f7f5 f4e3 f5e4 e3e4 c8d7 e1g1 g8h8 e4e3 c7c6 e2c4 b7b5 a4b5 c6b5 a1d1');
INSERT INTO evaluations VALUES(47866,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(47866,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(47866,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(47866,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(47866,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(47866,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(47866,36,'cp 59','d2d4 b8d7 f1c4 f8e7 a2a4 c7c6 e1g1 e8g8 f1e1 h7h6 h2h3 f8e8 a4a5 e5d4 f3d4 d7c5 c1f4 e7f8 d1f3 c5e6 c4e6 c8e6 a1d1 e6d7 a5a6 b7a6 e4e5 d6e5 f4e5 f6d5 c3d5 c6d5 f3d5 d7e6 d5b7 e6d5 b7a6 d8b6 a6b6 a7b6 d4f3 d5c6 e5c3 e8e1 f3e1');
INSERT INTO evaluations VALUES(23108,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(23108,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(23108,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(23108,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(23108,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(23108,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(23108,36,'cp 165','f3d4 e5d4 e1g1 g8f6 f1e1 d7d6 c2c3 f8e7 c3d4 d6d5 e4d5 f6d5 b1c3 c7c6 d1b3 e8g8 c3d5 c6d5 c4d5 a7a5 d5f3 e7f6 d4d5 f8e8 e1e8 d8e8 b3e3 e8b5 a2a4 b5b4 e3e8 b4f8 e8f8 g8f8 f3e4 c8d7 g2g3 a8c8 d5d6 d7e6 e4b7 c8d8 b7e4 d8d6 d2d3');
INSERT INTO evaluations VALUES(23102,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(23102,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(23102,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(23102,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(23102,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(23102,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(23102,36,'cp 150','e5d4 e1g1 g8f6 f1e1 d7d6 c2c3 f8e7 c3d4 d6d5 e4d5 f6d5 d1b3 c7c6 b1c3 e8g8 c3d5 c6d5 c4d5 a7a5 b3c4 a5a4 b2b4 c8f5 d5b7 a8b8 b7e4 b8b4 c4c3 f5e4 e1e4 b4b8 c1b2 e7f6 a1e1 d8d5 c3a3 d5b5 b2c3 f8e8 e4e8 b8e8 e1c1 h7h5 a3c5 b5d3 c5h5 f6d4 c3d4');
INSERT INTO evaluations VALUES(23139,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(23139,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(23139,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(23139,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(23139,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(23139,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(23139,36,'cp 151','e1g1 g8f6 f1e1 d7d6 c2c3 f8e7 c3d4 d6d5 e4d5 f6d5 d1b3 c7c6 b1c3 e8g8 c3d5 c6d5 c4d5 a7a5 b3c4 a5a4 b2b4 c8f5 a2a3 d8d7 d5e4 e7f6 c1b2 a8c8 c4d3 f5e4 e1e4 f8e8 a1e1 e8e4 d3e4 c8d8 g2g3 h7h5 g1g2 h5h4 g3h4 f6d4 b2c3 b7b5 e4f3 d4c3 d2c3 d8e8 e1e8 d7e8 h2h3 e8e1 f3c6');
INSERT INTO evaluations VALUES(23137,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(23137,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(23137,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(23137,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(23137,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(23137,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(23137,36,'cp 162','g8f6 f1e1 d7d6 c2c3 f8e7 c3d4 d6d5 e4d5 f6d5 b1c3 c7c6 d1b3 e8g8 c3d5 c6d5 c4d5 a7a5 b3c4 a5a4 b2b4 c8f5 a2a3 e7f6 c1b2 d8d7 d5e4 f5e4 e1e4 b7b5 c4d3 a8e8 e4e3 e8c8 b2c3 d7d5 a1e1 g7g6 d3e4 c8d8 e4d5 d8d5 e3d3 f8c8 g1f1 c8c4 h2h3 g8f8 e1e4 h7h6');
INSERT INTO evaluations VALUES(17029,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(17029,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(17029,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(17029,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(17029,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(17029,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(17029,36,'cp 164','f1e1 d7d6 c2c3 f8e7 c3d4 d6d5 e4d5 f6d5 b1c3 c7c6 d1b3 e8g8 c3d5 c6d5 c4d5 a7a5 b3c4 a5a4 b2b4 c8f5 a2a3 d8d7 d5e4 f8e8 c1b2 e7f6 e4f5 d7f5 e1e3 a8c8 c4d3 f5d5 a1e1 e8d8 d3e4 d5e4 e3e4 g8f8 b2c3 c8c4 g1f1 f6d4');
INSERT INTO evaluations VALUES(50221,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(50221,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(50221,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(50221,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50221,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(50221,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(50221,36,'cp 27','c2c4 f8g7 b1c3 d7d5 d1b3 d5c4 b3c4 e8g8 e2e4 a7a6 c4b3 b7b5 e4e5 f6d7 e5e6 f7e6 c1e3 d7f6 f1e2 d8d6 e1g1 b8c6 a1c1 c6a5 b3c2 c8b7 f1e1 a5c4 e3g5 c7c5 d4c5 d6c5 c3a4 c5c6 e2c4 b5c4 c2c4 c6c4 c1c4 b7d5 c4c5 d5a2');
INSERT INTO evaluations VALUES(50188,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(50188,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(50188,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(50188,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50188,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50188,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(50188,36,'cp 29','f8g7 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 e8g8 a1c1 f8d8 d4d5 e7e6 e3g5 f7f6 g5f4 b8c6 f1e2 g6g5 f4e3 e6d5 e4d5 b7b6 e1g1 c8e6 c3c4 e6f5 d2b2 c6b4 a2a3 b4d3 e2d3 f5d3 f1d1 d3g6 h2h3 a5a6 b2b3');
INSERT INTO evaluations VALUES(43189,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(43189,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(43189,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43189,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(43189,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(43189,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(43189,36,'cp 34','b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 e8g8 a1c1 f8d8 d4d5 e7e6 e3g5 f7f6 g5e3 b8c6 h2h4 e6d5 e4d5 c6e7 c3c4 a5d2 f3d2 d8e8 e1d1 b7b6 e3f4 e7f5 f1d3 f5d4 h4h5 c8f5 d3f5 d4f5 d2f3 a8d8 h5h6 g7f8 f4c7');
INSERT INTO evaluations VALUES(43181,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(43181,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(43181,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(43181,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(43181,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(43181,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(43181,36,'cp 21','d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 e1g1 b7b6 e3g5 f7f6 g5d2 a8d8 d2a5 b6a5 f1d1 d8b8 d1c1 b8d8 d4d5 f6f5 h2h3 f5e4 d3e4 g4f5 c1e1 d8d7 e4f5 f8f5');
INSERT INTO evaluations VALUES(41447,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(41447,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(41447,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41447,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(41447,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(41447,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(41447,36,'cp 36','c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 e1g1 b7b6 e3g5 e7e6 d4d5 e6d5 e4d5 f8c8 d3a6 c8c2 h2h3 g4f3 g2f3 g7f8 f1d1 f8d6 a2a4 c2a2 a6b5 a5b7 b5c6');
INSERT INTO evaluations VALUES(41427,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(41427,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(41427,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41427,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(41427,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41427,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41427,36,'cp 16','f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 h2h3 g4f3 g2f3 f8d8 d4d5 e7e6 e3g5 f7f6 g5d2 b7b6 d2a5 b6a5 d3c4 a8c8 c4b3 e6d5 b3d5 g8h8 e1f1 g7f8 h3h4 c8b8 f1g2 h8g7 b1b8 d8b8 h1c1 f8d6');
INSERT INTO evaluations VALUES(41548,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41548,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41548,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(41548,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(41548,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(41548,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41548,36,'cp 29','e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 e1g1 b7b6 f1d1 f8d8 d4d5 e7e6 h2h3 g4f3 g2f3 e6d5 e4d5 d8d5 d3e4 d5d1 b1d1 a8f8 d1d7 a5c4 e4d5 c4e3 f2e3 a7a5 f3f4 h7h6 g1g2 g6g5 g2f3 a5a4 f4g5 h6g5');
INSERT INTO evaluations VALUES(41555,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(41555,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41555,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41555,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41555,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41555,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(41555,36,'cp 24','d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 e1e2 a8d8 d4d5 b7b6 h2h3 g4c8 h1c1 e7e6 d5e6 c8e6 c1c2 d8c8 b1c1 c8c2 c1c2 f8c8 c2c8 e6c8 f3d4 g8f8 f2f4 f8e7 g2g4 c8d7 e3f2 a5c6 f2h4 g7f6 d4c6 d7c6 h4f6 e7f6 e2e3 h7h6 h3h4 c6d7 g4g5 h6g5 h4g5 f6e6 e3d4');
INSERT INTO evaluations VALUES(41569,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(41569,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(41569,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(41569,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(41569,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(41569,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41569,36,'cp 40','b2c3 c7c5 h2h3 e8g8 f1e2 c5d4 c3d4 b8c6 c1e3 d8a5 e3d2 a5a3 d4d5 c6e5 e1g1 e5f3 e2f3 g7a1 d1a1 f7f6 f1b1 f8f7 d2b4 a3a6 e4e5 f6e5 a1e5 a6f6 e5e3 c8h3 d5d6 e7e5 b4c3 a8e8 b1d1 h3d7 f3d5 d7c6 d5f7 f6f7 d1d2 f7f5 f2f3 g6g5 e3a7 g8f7 d6d7 e8d8');
INSERT INTO evaluations VALUES(6564,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(6564,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(6564,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(6564,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(6564,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6564,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(6564,36,'cp 12','d4c6 d8f6 d1f3 b7c6 b1d2 a7a5 a2a4 d7d6 d2b3 c5b6 c1d2 f6f3 g2f3 g8e7 d2e3 b6e3 f2e3 f7f5 h1g1 f5e4 g1g7 e4f3 b3d4 c8d7 e1c1 c6c5 d4f3 e7f5 g7g2 d7c6 f1b5 c6b5 a4b5 f5e3');
INSERT INTO evaluations VALUES(6353,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(6353,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(6353,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(6353,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(6353,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6353,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6353,36,'cp 9','d8f6 d1f3 b7c6 b1c3 f6f3 g2f3 d7d6 c1d2 g8f6 c3a4 c5b6 d2c3 e8e7 b2b3 h8g8 a4b2 a7a5 b2a4 b6a7 c3a5 a7f2 e1f2 a8a5 a4b2 c8e6 a2a4 g7g5 b2c4 e6c4 f1c4 g8a8 f2e3 g5g4 h2h4 g4f3 e3f3 a5h5 a1g1 f6d7 f3e3');
INSERT INTO evaluations VALUES(238,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(238,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(238,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(238,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(238,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(238,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(238,36,'cp 13','d1f3 b7c6 b1c3 f6f3 g2f3 d7d6 h2h4 a7a5 h4h5 g8f6 c1d2 e8g8 h5h6 g7g6 d2e3 c5e3 f2e3 f8e8 f1d3 e8e5 b2b3 g8f8 e1e2 e5g5 e2f2 g5c5 c3a4 c5e5 a2a3 c8d7 a4c3 e5h5 h1h5 f6h5 a1g1 h5f6 c3e2 f8e7 a3a4 c6c5 e2f4');
INSERT INTO evaluations VALUES(236,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(236,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(236,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(236,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(236,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(236,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(236,36,'cp 4','b7c6 b1c3 f6f3 g2f3 d7d6 h2h4 a7a5 c3a4 c5d4 c1e3 d4e3 f2e3 g8f6 b2b3 e8e7 e1d2 h7h6 f1d3 c8d7 a1g1 g7g6 a4c3 h8g8 f3f4 d7e6 g1g2 h6h5 h1e1 e6h3 g2g3 h3g4 d2c1 g8b8 e1g1 b8g8');
INSERT INTO evaluations VALUES(104,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(104,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(104,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(104,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(104,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(104,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(104,36,'cp 5','b1c3 c5d4 f3g3 g8e7 c1d2 e8g8 f1d3 e7g6 e1g1 f6e5 c3a4 d7d6 a1d1 a7a5 b2b3 f7f5 e4f5 c8f5 d2c3 d4c3 a4c3 e5c5 c3e4 c5b6 f1e1 a5a4 e4g5 f5d3 d1d3 a8e8 d3e3 e8e3 g3e3 a4b3 c2b3 b6e3 e1e3');
INSERT INTO evaluations VALUES(52443,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(52443,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52443,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(52443,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(52443,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(52443,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(52443,36,'cp -18','b8c6 g2g3 g8f6 c2c4 e7e6 f2f4 d7d5 e4e5 f6g8 f1g2 h7h5 b1c3 g8e7 g1f3 e7f5 e1g1 f8e7 f1e1 a7a6 b2b3 a8b8 c1b2 e8g8 f3g5 e7g5 f4g5 d8g5 c4d5 e6d5 c3d5 h5h4 d1c1 g5g6 d5f4 g6g4 g2h3 g4g5 h3f5 c8f5 f4d5 g5c1 a1c1 h4g3 h2g3 f5e6');
INSERT INTO evaluations VALUES(19457,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(19457,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(19457,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(19457,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(19457,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(19457,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(19457,36,'cp -17','b1c3 g8f6 c3d5 e7e6 d5f6 d8f6 g2g3 d7d5 f1g2 f8e7 g1f3 d5e4 d3e4 e6e5 c2c3 f6g6 d1e2 e8g8 f3d2 c8g4 f2f3 g4e6 d2c4 f7f5 f3f4 f5e4 e1g1 b7b5 f4f5 e6f5 c4e3 f5d7 e2b5 g8h8 e3d5 a8d8 f1f8 e7f8 c1e3');
INSERT INTO evaluations VALUES(19440,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(19440,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(19440,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19440,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(19440,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19440,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(19440,36,'cp -18','g8f6 c3d5 e7e6 d5f6 d8f6 g1f3 h7h6 d1e2 b7b6 g2g3 c5c4 d3d4 c6d4 f3d4 f6d4 c2c3 d4c5 e2c4 c8b7 c4e2 a8c8 c1e3 c5a5 f2f3 f8c5 a2a4 c5e3 e2e3 a5c5 e3c5 c8c5 e1f2 e8e7 f1e2 d7d5 e4d5 b7d5 h1d1 a7a5 d1d4 g7g5 a1d1');
INSERT INTO evaluations VALUES(12828,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(12828,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(12828,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(12828,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(12828,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(12828,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(12828,36,'cp -16','c3d5 e7e6 d5f6 d8f6 g1f3 h7h6 d1e2 b7b6 g2g3 c5c4 d3d4 c6d4 f3d4 f6d4 c2c3 d4c5 e2c4 c8b7 c4a4 b7c6 a4c2 a7a5 c1e3 c5e5 f1g2 f8c5 e3f4 e5h5 e1g1 e6e5 f4d2 e8g8 a1e1 f7f6 d2e3 c5e3 e1e3 h5g6');
INSERT INTO evaluations VALUES(56996,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56996,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(56996,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(56996,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(56996,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(56996,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(56996,36,'cp -31','c7c5 g1f3 b8c6 c2c3 d7d6 e5d6 f8d6 d2d4 c5d4 f3d4 c8d7 d4f5 d6f8 f5d6 f8d6 d1d6 g8e7 f1d3 e6e5 b1d2 e8g8 d3c2 d8c8 d2e4 d7e6 c1g5 f8d8 d6a3 f7f6 g5e3 c8c7 e1g1 e7d5 a1d1 h7h6 f1e1 c6a5 e3c1 f6f5 e4g3 a5c4 a3b3 f5f4 g3f5');
INSERT INTO evaluations VALUES(51909,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(51909,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(51909,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(51909,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(51909,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(51909,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(51909,36,'cp -21','c2c3 d7d6 g1f3 b8c6 e5d6 f8d6 d2d4 c5d4 f3d4 c8d7 d4f5 d6f8 f5d6 f8d6 d1d6 g8e7 f1d3 e8g8 e1g1 e6e5 f1e1 d7f5 d6d8 f8d8 d3b5 a8c8 b1d2 f7f6 d2b3 b7b6 f2f4 e7g6 b5c6 c8c6 f4e5 f6e5 c1g5 d8d5 g5e3 g6f4 a1d1 d5d1 e1d1 f5e6');
INSERT INTO evaluations VALUES(51907,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(51907,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(51907,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(51907,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(51907,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(51907,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(51907,36,'cp -27','d7d6 g1f3 d6e5 f3e5 b8d7 d2d4 c5d4 f1b5 g8f6 d1d4 f8e7 e5d7 c8d7 b5d7 d8d7 d4d7 e8d7 e1e2 e6e5 c1g5 h8d8 c3c4 f6e4 h1d1 d7e6 d1d8 a8d8 g5e7 e6e7 b2b3 d8d4 f2f3 e4c5 b1c3 a7a6 a1c1 f7f5 c1d1 g7g5 c3d5 e7e6 d1d4 e5d4 d5b6 e6f6 e2d2 g5g4 f3g4');
INSERT INTO evaluations VALUES(50319,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50319,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(50319,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(50319,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50319,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(50319,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(50319,36,'cp 6','e7e5 g2g3 f8b4 f1g2 c7c6 g1f3 d7d6 e1g1 e8g8 d1c2 b4c3 c2c3 a7a5 d2d3 f8e8 b2b3 d6d5 c1a3 c8g4 c4d5 g4f3 g2f3 c6d5 f1c1 b8c6 f3g2 h7h6 c3b2 d8d7 c1c5 e5e4 a1d1 e4d3 e2d3 d5d4 c5b5');
INSERT INTO evaluations VALUES(21018,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(21018,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(21018,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(21018,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(21018,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21018,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(21018,36,'cp 52','g1f3 g8f6 b1d2 e7e6 b2b3 a7a5 c2c4 a5a4 b3a4 c8d7 a2a3 c6e7 a4a5 a8a5 f1d3 e7c8 e1g1 f8e7 f3e5 d7a4 d1f3 e8g8 c1b2 c8b6 c4d5 e6d5 b2c3 a5a8 f1c1 f8e8 c3b4 b6d7 f3g3 g7g6 g3f4 d7e5 d4e5');
INSERT INTO evaluations VALUES(21021,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(21021,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(21021,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(21021,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(21021,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(21021,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(21021,36,'cp 51','c8g4 c2c4 e7e6 a2a3 f8e7 b2b4 g8f6 c1b2 e8g8 c4c5 f6e4 f1e2 a7a6 e1g1 b7b6 b1d2 b6c5 d4c5 e4d2 d1d2 e7f6 f1d1 g4f3 e2f3 c6e5 f3e2 c7c6 a1c1 d8e7 d2c2 e5d7 e3e4 g7g6 b2c3 f6c3 c2c3 e7f6 c3f6 d7f6 e4e5');
INSERT INTO evaluations VALUES(26617,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(26617,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(26617,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(26617,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(26617,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(26617,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(26617,36,'cp 57','c2c4 e7e6 a2a3 g8f6 h2h3 g4f5 b1c3 f8e7 f1d3 f5g6 d3e2 e8g8 f3h4 c6a5 h4g6 h7g6 c4c5 b7b6 e1g1 b6c5 d4c5 a5c4 e2c4 d5c4 d1a4 e7c5 a4c4 c5d6 b2b4 f6d5 c3d5 e6d5 c4c3 d6e7 c1b2 e7f6 c3c2 f6b2 c2b2 c7c6');
INSERT INTO evaluations VALUES(26591,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(26591,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(26591,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(26591,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(26591,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(26591,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(26591,36,'cp 55','e7e6 h2h3 g4h5 a2a3 g8f6 b1d2 f8e7 f1e2 e8g8 b2b4 d5c4 e2c4 a8c8 c1b2 a7a5 b4b5 c6a7 e1g1 c7c6 b5c6 a7c6 d1b3 d8d7 f1c1 f8d8 c4d3 c8c7 d3b5 d8c8 f3e5 d7d8 b5c6 b7c6 b3a4 c6c5 a4a5 c8a8 a5c3 f6d5 c3b3 a8b8 b3a2');
INSERT INTO evaluations VALUES(26831,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(26831,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(26831,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(26831,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(26831,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(26831,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(26831,36,'cp 54','a2a3 g8f6 h2h3 g4h5 b1d2 f8e7 f1e2 e8g8 b2b4 d5c4 e2c4 a8c8 e1g1 a7a5 b4b5 c6b8 c1b2 c7c5 b5c6 b8c6 d1b3 d8d7 c4b5 d7d5 f1c1 d5b3 d2b3 h5f3 g2f3 c6a7 b5d3 a5a4 b3c5 e7c5 d4c5 f6d7 b2d4 a7c6 f3f4 c6d4 e3d4');
INSERT INTO evaluations VALUES(19442,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19442,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19442,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19442,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(19442,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19442,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19442,36,'cp 6','g8f6 e4e5 f6d5 d2d4 c5d4 f1c4 d5b6 c4b3 d7d5 e5d6 d8d6 f3d4 c6d4 c3d4 c8e6 b1c3 e6b3 a2b3 g7g6 e1g1 f8g7 d1f3 e8g8 f3b7 f8b8 b7e4 d6d4 e4d4 g7d4 f1d1 b8d8 g1f1 g8g7 f1e2 a7a6 h2h3 d8d7 g2g4 h7h6 a1a5 e7e6 c1d2');
INSERT INTO evaluations VALUES(12829,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12829,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12829,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(12829,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(12829,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(12829,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(12829,36,'cp 12','e4e5 f6d5 f1c4 d5b6 c4b3 c5c4 b3c2 d7d5 e5d6 d8d6 b1a3 c6e5 f3e5 d6e5 d1e2 e5e2 e1e2 e7e5 b2b3 f8a3 c1a3 c8e6 a3c5 e8c8 g2g3 d8d7 h1d1 h8d8 c5b6 a7b6 d2d3 c4d3 d1d3 d7d3 c2d3 h7h6 e2e3 c8c7 a1d1 g7g6 f2f3 d8d6 d1d2');
INSERT INTO evaluations VALUES(12771,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12771,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(12771,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12771,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(12771,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12771,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(12771,36,'cp 19','f6d5 f1c4 d5b6 c4b3 d7d5 e5d6 d8d6 e1g1 c8e6 b3e6 d6e6 a2a4 b6d5 a4a5 e6d7 a5a6 b7b6 d2d4 c5d4 f3d4 e7e6 d4c6 d7c6 d1g4 d5f6 g4e2 f8e7 b1d2 e8g8 d2f3 a8d8 c1f4 d8d5 f1d1 f8d8 d1d5 c6d5 f4e5 d5e4 e2e4 f6e4 e5d4 e4c5');
INSERT INTO evaluations VALUES(12916,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(12916,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(12916,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12916,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(12916,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(12916,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(12916,36,'cp 24','f1c4 d5b6 c4b3 d7d5 e5d6 c5c4 b3c2 d8d6 e1g1 c8g4 h2h3 g4h5 f1e1 e7e6 b2b3 f8e7 d1e2 c4b3 a2b3 d6c7 c1a3 e8g8 a3e7 c7e7 c2e4 f8d8 d2d4 e7c7 e2b2 b6d5 b1d2 h5g6 a1d1 c7f4 e4g6 h7g6 d2e4 f4c7');
INSERT INTO evaluations VALUES(12914,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(12914,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12914,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(12914,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(12914,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(12914,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(12914,36,'cp 14','d5b6 c4b3 c5c4 b3c2 d7d5 e5d6 d8d6 b1a3 c6e5 f3e5 d6e5 d1e2 e5e2 e1e2 e7e5 b2b3 f8a3 c1a3 c8e6 a3c5 e8c8 c5b6 a7b6 b3c4 e6c4 d2d3 c4d5 h1b1 c8c7 a2a4 d8d6 a4a5 b6a5 a1a5 f7f6 g2g3 h8d8 h2h4 d5c6 b1a1 d6d5 e2e3 d5d7');
INSERT INTO evaluations VALUES(12753,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(12753,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12753,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12753,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(12753,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(12753,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(12753,36,'cp 20','c4b3 d7d5 e5d6 d8d6 e1g1 c8e6 b3e6 d6e6 a2a4 e6d7 a4a5 b6d5 a5a6 b7b6 d2d4 c5d4 f3d4 e7e6 d4c6 d7c6 d1g4 a8c8 b1d2 d5f6 g4e2 f8c5 d2f3 e8g8 f1e1 f6d5 c1d2 h7h6 a1d1 c6c7 h2h4 c8d8 f3e5 c5e7');
INSERT INTO evaluations VALUES(12754,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(12754,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12754,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(12754,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(12754,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(12754,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(12754,36,'cp 7','d7d6 e5d6 d8d6 e1g1 c5c4 b3c2 c8g4 f1e1 e7e6 b1a3 a8d8 b2b3 c4b3 a2b3 b6d5 a3c4 d6c7 c1b2 f8e7 d2d4 e8g8 c2e4 g4h5 d1c2 h5g6 e4g6 h7g6 c4e5 c6e5 f3e5 b7b5 c3c4 e7d6 e5f3 a7a5 g2g3 d8b8 c2e2 b5c4 b3c4 d5f6 b2c3');
INSERT INTO evaluations VALUES(57157,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(57157,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(57157,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(57157,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(57157,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(57157,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(57157,36,'cp -42','g8f6 f1c4 c7c6 b1c3 b7b5 c4b3 a7a5 a2a3 f8c5 g1e2 d7d6 e1g1 c8g4 f3g3 b8d7 h2h3 g4h5 g3h4 h7h6 g2g4 h5g6 d2d3 c5a7 e2g3 d7c5 c1g5 c5b3 c2b3 a7b6 g1g2 e8d7 g5f6 d8f6 h4f6 g7f6');
INSERT INTO evaluations VALUES(49660,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(49660,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(49660,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(49660,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(49660,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(49660,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(49660,36,'cp -13','f1c4 c7c6 b1c3 b7b5 c4b3 a7a5 a2a3 f8c5 d2d3 d7d6 g1e2 a5a4 b3a2 c8e6 c1e3 c5e3 f3e3 e6a2 a1a2 b8d7 e1g1 d8b6 e3b6 d7b6 f2f4 e8e7 a2a1 h7h5 h2h3 g7g6 c3a2 f6d7 f4e5 d6e5 f1f2 f7f6 a1f1 h8f8 e2c3 a8c8 a2b4 h5h4');
INSERT INTO evaluations VALUES(49535,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(49535,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(49535,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(49535,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(49535,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(49535,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(49535,36,'cp -38','c7c6 b1c3 b7b5 c4b3 a7a5 a2a3 f8c5 g1e2 d7d6 d2d3 c8g4 f3g3 d8e7 e1g1 b8d7 g1h1 g4e6 b3e6 f7e6 f2f4 e8g8 g3h3 h7h5 c1e3 c5e3 h3e3 f6g4 e3g3 e5f4 e2f4 e7f7 h2h3 e6e5 f4e2 f7f1 a1f1 f8f1 e2g1 g4f2 h1h2 d7f6');
INSERT INTO evaluations VALUES(44085,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(44085,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(44085,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(44085,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(44085,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(44085,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(44085,36,'cp -18','b1c3 b7b5 c4b3 a7a5 a2a3 f8c5 d2d3 e8g8 g1e2 d7d6 e1g1 c8g4 f3g3 b8d7 g1h1 g4e6 b3e6 f7e6 g3h3 d8e7 e2g3 a8e8 c3e2 e7f7 f2f3 d6d5 c1d2 a5a4 a1e1 h7h6 f3f4 e5f4 e2f4 d7e5 f4h5 f7g6 h5f6 f8f6 e4d5 f6f1 e1f1 c6d5');
INSERT INTO evaluations VALUES(44072,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(44072,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(44072,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(44072,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(44072,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(44072,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(44072,36,'cp -33','b7b5 c4b3 a7a5 a2a3 f8c5 d2d3 e8g8 g1e2 d7d6 e1g1 c8g4 f3g3 g4e6 b3e6 f7e6 g1h1 d8e7 g3h3 b8d7 e2g3 a5a4 f2f4 e5f4 c1f4 c5d4 g3e2 d4b6 f1f3 f8f7 a1f1 a8f8 c3d1 b6c5 g2g4 e6e5 f4e3 d6d5 g4g5 f6e8 e3c5 d7c5 d1e3 f7f3 f1f3 f8f3 h3f3');
INSERT INTO evaluations VALUES(55093,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(55093,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(55093,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(55093,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(55093,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(55093,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(55093,36,'cp -20','e2e3 g8f6 g1f3 c7c5 b2b3 g7g6 f1e2 f8g7 c1b2 e8g8 e1g1 b7b6 f3e5 e7e6 c2c4 c8b7 e2f3 f6e4 d2d3 e4d6 d1e2 f7f6 e5g4 d6f5 b1c3 f5h4 c4d5 e6d5 a1d1 h7h5 g4f2 h4f3 e2f3 f6f5 f3g3 g8h7 d3d4 c5d4 e3d4 b8c6 c3b5');
INSERT INTO evaluations VALUES(55089,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55089,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(55089,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(55089,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(55089,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(55089,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(55089,36,'cp -45','g7g6 g1f3 f8g7 d2d4 c7c5 f1b5 c8d7 b5d3 g8h6 e1g1 e8g8 a2a4 b8c6 d4c5 d8a5 d3b5 a7a6 c1d2 a5c7 b5c6 c7c6 d2c3 c6c5 c3d4 g7d4 e3d4 c5b4 d1c1 f7f6 f1e1 h6f5 c2c3 b4d6 a4a5 g8f7 b1a3 f8c8 a3c2 d7b5 c2e3 d6f4 e3d5 f4c1 a1c1');
INSERT INTO evaluations VALUES(49427,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(49427,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(49427,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(49427,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(49427,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(49427,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(49427,36,'cp 33','c2c4 d7d5 b1c3 f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 g7g5 h4g3 f6e4 f3d2 e4g3 h2g3 c7c6 a2a3 b4f8 d1c2 f8g7 e2e3 e8g8 f1d3 b8d7 e1g1 d7f6 b2b4 c8g4 a1c1 f8e8 d2f3 a8c8 d3f5 c8c7 f5g4 f6g4 c2f5 g4f6 f3e5 h6h5');
INSERT INTO evaluations VALUES(55824,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(55824,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(55824,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(55824,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(55824,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(55824,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(55824,36,'cp 12','g8f6 e2e3 c7c5 b1d2 d8b6 a1b1 f8d6 d4c5 b6c5 b2b4 c5c7 f4d6 c7d6 c2c4 e8g8 f1e2 d6e7 e1g1 b7b6 d1b3 b8d7 f1c1 d5c4 d2c4 c8b7 f3d4 f8c8 a2a4 b7e4 b1a1 a7a6 b3b2 e4b7 h2h3 h7h6 c4d2 f6d5 b4b5 c8c1 a1c1 d7c5 b5a6 b7a6');
INSERT INTO evaluations VALUES(48291,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(48291,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(48291,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(48291,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(48291,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(48291,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(48291,36,'cp 17','e2e3 f8d6 f4d6 d8d6 c2c4 c7c5 c4d5 f6d5 f1e2 c5d4 d1d4 e8g8 b1c3 b8c6 d4c4 d6b4 c4b4 c6b4 c3d5 b4d5 a1c1 c8d7 e1g1 f8c8 g2g3 g8f8 f3e5 d7e8 e2f3 f8e7 f3d5 e6d5 e5f3 e8d7 f3d4 c8c4 b2b3 c4c1 f1c1');
INSERT INTO evaluations VALUES(22899,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(22899,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22899,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(22899,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22899,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(22899,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(22899,36,'cp -25','f8b4 c2c3 d4c3 b2c3 b4a5 e1g1 a5b6 e4e5 g8e7 c1a3 e8g8 f1e1 h7h6 c4b3 f8e8 b1d2 e7f5 d2c4 d7d5 e5d6 e8e1 d1e1 c7d6 a1d1 c8e6 c4d6 e6b3 a2b3 f5d6 d1d6 d8c7 g2g3 c6a5 e1e5 a8d8 f3h4 a5c6 e5d5 d8d6 a3d6 c7d8 h4f5 d8f6 h2h4 f6c3');
INSERT INTO evaluations VALUES(6696,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(6696,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(6696,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(6696,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(6696,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(6696,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(6696,36,'cp -6','c2c3 d4c3 e1g1 g8f6 e4e5 d7d5 e5f6 d5c4 d1d8 c6d8 f6g7 h8g8 b1c3 b4c3 b2c3 f7f6 f3d4 g8g7 f1e1 e8f7 c1f4 d8e6 d4e6 c8e6 f4c7 a8g8 g2g3 g7g5 a1d1 b7b6 d1d2 g5f5 e1e3 f5c5 c7f4 c5d5 d2d5 e6d5 f2f3 g8d8 g1f2 d5c6 h2h4 d8d1 g3g4 d1d2 e3e2 d2e2 f2e2');
INSERT INTO evaluations VALUES(6695,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(6695,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6695,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(6695,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6695,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(6695,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(6695,36,'cp -11','d4c3 e1g1 g8f6 e4e5 d7d5 e5f6 d5c4 d1d8 c6d8 f6g7 h8g8 b1c3 b4c3 b2c3 f7f6 f3d4 g8g7 f1e1 e8f7 c1f4 d8e6 d4e6 c8e6 f4c7 a8g8 g2g3 g7g5 e1e4 g5d5 a1e1 g8e8 f2f3 d5d3 e1e3 d3e3 e4e3 e8c8 c7f4 b7b5 g1f2 c8d8 f4c7 d8d1 c7a5 d1d5 f3f4 h7h5 a5b4');
INSERT INTO evaluations VALUES(6685,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(6685,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(6685,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(6685,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(6685,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(6685,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(6685,36,'cp -5','e1g1 g8f6 e4e5 d7d5 e5f6 d5c4 d1d8 c6d8 f6g7 h8g8 b1c3 b4c3 b2c3 f7f6 f1e1 e8f7 f3d4 g8g7 c1f4 d8e6 d4e6 c8e6 f4c7 a8g8 g2g3 g7g5 e1e4 g5d5 a1e1 g8e8 f2f3 d5d3 e1e3 f6f5 e3d3 c4d3 e4d4 e8c8 c7a5 e6a2');
INSERT INTO evaluations VALUES(6684,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(6684,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(6684,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(6684,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(6684,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(6684,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(6684,36,'cp -12','g8f6 e4e5 d7d5 e5f6 d5c4 d1d8 c6d8 f6g7 h8g8 b1c3 b4c3 b2c3 f7f6 f3d4 g8g7 c1f4 d8e6 f1e1 e8f7 d4e6 c8e6 f4c7 a8g8 g2g3 g7g5 e1e4 g5d5 a1e1 g8e8 f2f3 e6d7 g1f2 e8e4 f3e4 d5c5 c7d6 c5b5 e1e2 f7g6 d6b8 a7a5 e2d2 d7c6 f2e3 b5b1 e3d4');
INSERT INTO evaluations VALUES(1477,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(1477,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(1477,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(1477,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(1477,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(1477,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(1477,36,'cp 25','a2a3 c5c4 b1d2 c6a5 f1e2 c8d7 a1b1 g8e7 e1g1 h7h6 f1e1 e8c8 d2f1 a5b3 f1g3 d7a4 c1e3 b6c7 f3d2 e7c6 g3h5 c8b8 e2f1 b8a8 d2b3 c4b3 d1g4 g7g6 h5f6 d8c8 h2h4 h6h5 g4f3 c6a5 f1e2 f8h6 e3g5 h6g5 h4g5 a5c4');
INSERT INTO evaluations VALUES(1487,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(1487,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(1487,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(1487,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(1487,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(1487,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(1487,36,'cp 22','c5c4 b1d2 c6a5 d1c2 c8d7 f1e2 h7h6 e1g1 g8e7 a1b1 e7c8 f1e1 b6c7 d2f1 c8b6 e2d1 e8c8 f1g3 d7a4 c2e2 a4b3 c1e3 f8e7 h2h3 c8b8 f3h2 b8a8 h2g4 d8f8 g3h5 g7g6 h5f6 h6h5 g4h2 e7f6 e5f6 a5c6 e3g5 b3d1 e2d1');
INSERT INTO evaluations VALUES(1496,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(1496,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(1496,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(1496,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(1496,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(1496,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(1496,36,'cp 30','f1e2 c6a5 b1d2 c8d7 a1b1 g8e7 h2h4 e8c8 h4h5 c8b8 d2f1 h7h6 c1e3 b6c7 f1g3 e7c8 e1f1 c8b6 d1d2 b8a8 f1g1 f8e7 e2d1 d7a4 g3e2 a4b3 b1a1 b3d1 d2d1 a5b3 a1b1 d8c8 f3d2 b3a5');
INSERT INTO evaluations VALUES(1495,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(1495,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(1495,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(1495,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(1495,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(1495,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(1495,36,'cp 37','g8e7 b1d2 c6a5 h2h4 c8d7 a1b1 e8c8 h4h5 h7h6 g2g3 c8b8 e1f1 b8a8 f1g2 e7c8 d2f1 f8e7 c1f4 b6c7 f1e3 c8b6 d1e1 d8g8 e2d1 d7b5 e1e2 b5a4 d1a4 b6a4 e3g4 c7d7 f4e3 g8c8 f3h4 d7e8 e2d2 a5b3');
INSERT INTO evaluations VALUES(55144,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(55144,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(55144,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(55144,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(55144,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(55144,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(55144,36,'cp 24','g1f3 g7g6 f1g2 f8g7 d2d4 c7c6 c2c4 g8f6 c4d5 c6d5 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 e8g8 e1g1 d8b6 a1c1 f6g4 e5g7 g8g7 c1c2 a8c8 d2f4 g4f6 f4e5 h7h5 f1c1 c8c4 e2e3 c4c6 h2h3 f8c8 a2a3 h5h4');
INSERT INTO evaluations VALUES(55141,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(55141,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(55141,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(55141,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(55141,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55141,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(55141,36,'cp 26','g7g6 d2d4 c7c6 c2c4 g8f6 c4d5 c6d5 b1c3 b8c6 f1g2 f8g7 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 e8g8 e1g1 a7a5 f1c1 d8e7 h2h4 f8c8 e2e4 d5e4 c3e4 c8c1 a1c1 f6e4 g2e4 e7b4 d2b4 a5b4 e5g7 g8g7 e4b7 a8a2 c1c7');
INSERT INTO evaluations VALUES(54873,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(54873,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(54873,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54873,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(54873,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(54873,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54873,36,'cp 17','f1g2 f8g7 d2d4 c7c6 c2c4 g8f6 c4d5 c6d5 f3e5 b8c6 b1c3 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 a8c8 e1g1 e8g8 f1c1 d8b6 e2e3 f6h5 g3g4 f7f6 e5g3 h5g3 h2g3 f8d8 a1b1 g7f8 e3e4 d5e4 c3e4 f8e7 c1c8 d8c8 g4g5 f6g5 e4g5');
INSERT INTO evaluations VALUES(54874,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(54874,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(54874,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(54874,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54874,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(54874,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54874,36,'cp 13','f8g7 d2d4 c7c6 c2c4 g8f6 c4d5 c6d5 f3e5 b8c6 e1g1 c8d7 c1f4 e8g8 b1c3 e7e6 d1d2 c6e5 f4e5 a8c8 h2h4 f6g4 e5f4 g4f6 a1c1 f8e8 f4e5 a7a5 a2a3 d8b6 g2f3 h7h5 e2e3 c8c4 c1b1 e8c8 f1c1 c4c6 e5f6');
INSERT INTO evaluations VALUES(38477,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(38477,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38477,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(38477,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(38477,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(38477,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38477,36,'cp 23','d2d4 g8f6 c2c4 c7c6 c4d5 c6d5 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 e8g8 e1g1 a8c8 h2h4 h7h5 f1c1 d8b6 e2e3 c8c6 c3e2 f8c8 c1c6 d7c6 a1c1 b6d8 b2b3 f6e4 d2b2 c6b5 e2f4 c8c1 b2c1 e4d6 e5g7 g8g7 c1b2 b7b6 b2d2 b5c6');
INSERT INTO evaluations VALUES(38469,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(38469,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(38469,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(38469,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(38469,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(38469,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(38469,36,'cp 11','c7c6 c2c4 g8f6 c4d5 c6d5 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 d8e7 e1g1 e8g8 f1c1 f8c8 h2h4 a7a5 g2f3 h7h5 e2e4 d5e4 c3e4 f6e4 f3e4 d7c6 e4c6 c8c6 c1c6 b7c6 d2c3 g7e5 d4e5 a8b8 a1c1 e7b4');
INSERT INTO evaluations VALUES(53058,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(53058,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(53058,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(53058,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(53058,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(53058,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(53058,36,'cp 31','g1e2 g8f6 d2d4 c5d4 e2d4 a7a6 h2h3 e7e5 d4e2 h7h5 g2g3 f8e7 f1g2 b7b5 c3d5 b8d7 e2c3 c8b7 e1g1 f6d5 c3d5 h5h4 c2c3 b7d5 d1d5 e8g8 g3g4 d7b6 d5d2 b6a4 f1d1 a4c5 d2d5 a8c8 c1e3 e7g5 e3c5 c8c5 d5d6 d8d6 d1d6');
INSERT INTO evaluations VALUES(53059,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(53059,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(53059,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53059,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53059,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(53059,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(53059,36,'cp 17','g8f6 d2d4 c5d4 e2d4 a7a6 c1e3 e7e5 d4b3 c8e6 f1e2 f8e7 c3d5 b8d7 d1d3 e8g8 e1g1 e6d5 e4d5 f6e8 a1d1 e7g5 e3g5 d8g5 b3a5 a8b8 b2b4 f7f5 d3d2 g5d2 d1d2 g8f7 f2f3 g7g6 f1e1 b7b5 a5c6 b8a8 c2c4 e8c7');
INSERT INTO evaluations VALUES(45193,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(45193,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45193,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45193,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(45193,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(45193,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(45193,36,'cp 36','d2d4 c5d4 e2d4 a7a6 c1e3 e7e5 d4b3 c8e6 f2f3 h7h5 f1e2 f8e7 c3d5 e6d5 e4d5 b8d7 d1d2 d8c7 c2c4 a6a5 a2a4 e8g8 e1g1 h5h4 a1d1 d7b6 b3a5 b6d5 c4d5 c7a5 d2a5 a8a5 e2b5 a5a8 g2g4 h4g3 h2g3 f8c8 g3g4');
INSERT INTO evaluations VALUES(31833,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31833,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(31833,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(31833,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(31833,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(31833,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31833,36,'cp 18','e2e3 g8f6 g1f3 e7e6 c2c4 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 b8c6 e1g1 e8g8 b5c6 b7c6 h2h3 a8c8 d1e2 c6c5 f1c1 c5d4 f3d4 h7h6 c1c8 d8c8 e2b5 b6d5 d4c6 c8d7 a1c1');
INSERT INTO evaluations VALUES(31830,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31830,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31830,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31830,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31830,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31830,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(31830,36,'cp 13','g8f6 g1f3 e7e6 f1d3 f5g6 e1g1 f8d6 f4g5 e8g8 d3g6 h7g6 b1d2 c7c5 d4c5 d6c5 c2c4 b8c6 d1e2 c5e7 f1d1 d8b6 c4d5 f6d5 d2c4 b6c7 g2g3 e7g5 f3g5 a8d8 h2h4 d5f6 d1d8 f8d8 a1c1 c7e7 a2a3 e6e5 c1d1 e5e4');
INSERT INTO evaluations VALUES(29539,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(29539,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29539,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29539,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(29539,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(29539,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(29539,36,'cp 22','g1f3 e7e6 c2c4 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 e1g1 e8g8 d1b3 b8c6 c4b5 b6d5 b5c6 b7c6 f1c1 a7a5 c1c6 a8b8 b3a4 b8b2 d2c4 b2c2 a4a5 d8a5 c4a5 d5c3 c6d6 c3e2 g1h1 g6e4 f3e5 f7f6 e5c4 e2c3 a5c6');
INSERT INTO evaluations VALUES(29541,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(29541,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(29541,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29541,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(29541,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(29541,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(29541,36,'cp 16','e7e6 c2c4 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 b8d7 e1g1 e8g8 d1b3 d7f6 f1c1 h7h6 h2h3 a7a5 c1e1 a8c8 a1c1 c8a8 b5f1 a8b8 b3b5 b6d5 d2c4 d5b4 b5a5 d8a5 c4a5 b4a2 c1c7 a2b4');
INSERT INTO evaluations VALUES(30097,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(30097,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30097,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30097,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(30097,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(30097,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(30097,36,'cp 12','c2c4 f8d6 f4d6 c7d6 d1b3 d5c4 f1c4 d8e7 c4b5 b8d7 e1g1 e8g8 b1d2 a8c8 f1c1 d7b6 a2a4 c8c1 a1c1 f8c8 c1c8 b6c8 f3h4 g7g6 h4f5 g6f5 b3c3 e7d8 b5d3 f6d5 c3b3 d8b6 g1f1 b6b4 f1e2 c8b6 b3b4 d5b4 d2b1 d6d5 b1c3 b6c4');
INSERT INTO evaluations VALUES(30065,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(30065,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(30065,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(30065,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(30065,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(30065,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(30065,36,'cp 8','f8d6 f4d6 c7d6 d1b3 d5c4 f1c4 d8e7 c4b5 b8d7 e1g1 e8g8 f1c1 f8c8 b1d2 d7b6 a2a4 c8c1 a1c1 a8c8 c1c8 b6c8 f3h4 g7g6 h4f5 g6f5 b5e2 e7c7 b3c4 c7c4 d2c4 d6d5 c4e5 g8f8 a4a5 f6e8 f2f3 c8d6');
INSERT INTO evaluations VALUES(23011,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(23011,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(23011,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(23011,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(23011,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(23011,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(23011,36,'cp 83','f3d4 e5d4 e1g1 f8c5 b5c4 d7d6 d2d3 c7c6 a2a4 g8f6 c4a2 a7a5 b1d2 c8e6 a2e6 f7e6 d2b3 c5b6 c1g5 h7h6 g5f6 d8f6 f2f4 h6h5 d1d2 e8g8 a1e1 f8f7 g2g3 e6e5 f4e5 f6e5 f1f7 g8f7 g1g2');
INSERT INTO evaluations VALUES(23007,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(23007,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(23007,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(23007,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(23007,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(23007,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(23007,36,'cp 79','e5d4 e1g1 c7c6 b5c4 g8f6 d2d3 a7a5 f1e1 d7d6 c2c3 b7b5 c4b3 d4c3 b1c3 f8e7 d3d4 e8g8 h2h3 c8e6 d4d5 c6d5 e4d5 e6d7 c1e3 a5a4 b3c2 a4a3 b2b4 a8c8 d1d3 g7g6 c2b3 f8e8 e3d4 e7f8 a1d1');
INSERT INTO evaluations VALUES(23020,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(23020,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(23020,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(23020,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(23020,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(23020,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(23020,36,'cp 83','e1g1 f8c5 b5c4 d7d6 d2d3 g8e7 d1h5 e8g8 c1g5 h7h6 b1d2 c8e6 c4e6 f7e6 g5e7 d8e7 g2g3 e6e5 a2a3 f8f6 h5g4 e7e6 g4e6 f6e6 g1g2 a7a5 f2f4 e5f4 f1f4 a5a4 a1f1 a8e8 h2h4 d6d5 f4f5 e6e5 g2h3 d5e4 d2e4 c5d6 h4h5 e5e4 d3e4 e8e4 f5a5 e4e3 a5a4 d6g3');
INSERT INTO evaluations VALUES(23019,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(23019,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(23019,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(23019,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(23019,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(23019,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(23019,36,'cp 65','f8c5 b5c4 d7d6 d2d3 g8e7 c1g5 e8g8 d1h5 h7h6 b1d2 c8e6 c4e6 f7e6 g5e7 d8e7 a2a3 e6e5 g2g3 f8f6 h5g4 e7e6 g4e6 f6e6 b2b4 c5b6 a1b1 a8f8 a3a4 c7c6 b4b5 d6d5 b5c6 b7c6 g1g2 e6f6 e4d5 c6d5 b1b5 e5e4 b5d5 e4e3 d2e4 f6c6 f1c1 e3e2 a4a5 b6c7');
INSERT INTO evaluations VALUES(6832,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(6832,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(6832,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(6832,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(6832,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(6832,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(6832,36,'cp 70','b5c4 d7d6 d2d3 g8e7 c1g5 e8g8 d1h5 h7h6 b1d2 c8e6 c4e6 f7e6 g5e7 d8e7 a2a3 e6e5 g2g3 f8f6 f2f4 e5f4 f1f4 e7e5 h5f3 a8f8 f4f6 f8f6 f3g4 d6d5 a1e1 a7a5 e1e2 d5e4 e2e4 e5f5 g4f5 f6f5 g1g2 a5a4 d2f3 f5d5 h2h4 g8f7 g3g4 b7b6 g4g5');
INSERT INTO evaluations VALUES(49637,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(49637,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(49637,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(49637,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(49637,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49637,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(49637,36,'cp -19','b8c6 c2c4 f8c5 h2h3 d7d6 b1c3 e8g8 f1e2 c8d7 e1g1 a7a5 a2a3 f8e8 c1d2 h7h6 a1b1 f6h7 g1h2 f7f5 e4f5 d7f5 c3e4 a5a4 e4c3 c6d4 f3d4 e5d4 c3e4 f5e4 d3e4 e8e4 b2b4 a4b3 e2f3 e4e5 f3b7 a8a3 b1b3 a3b3 d1b3 h7f8');
INSERT INTO evaluations VALUES(57121,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(57121,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(57121,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(57121,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(57121,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(57121,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(57121,36,'cp -54','f8c5 g1f3 b8c6 d2d3 d7d6 c1e3 g8e7 a2a3 e8g8 b2b4 c5d4 a1a2 d4b6 f1e2 f7f5 e1g1 g8h8 b1d2 e7g6 d2b3 b6e3 f2e3 d8e7 e4f5 c8f5 b4b5 c6b8 a3a4 b8d7 a4a5 a8e8 e3e4 f5e6 g2g3 b7b6 f1f2 e6h3 a5b6 a7b6 d1d2 h7h6 d2e3');
INSERT INTO evaluations VALUES(39212,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(39212,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(39212,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(39212,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(39212,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(39212,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(39212,36,'cp -30','d2d3 b8c6 c1e3 d7d6 g1f3 g8e7 a2a3 a7a5 d3d4 e5d4 f3d4 e8g8 b1c3 c8e6 d4e6 f7e6 d1d2 e6e5 e1c1 c6d4 c1b1 e7c6 g2g3 d4b3 d2d3 g8h8 f1h3 c5e3 f2e3 f8f2 h1f1 b3c5 f1f2 c5d3 d1d3 d8g5 f2f7');
INSERT INTO evaluations VALUES(39210,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(39210,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(39210,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(39210,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(39210,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(39210,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(39210,36,'cp -36','d7d6 g1f3 b8c6 c1e3 g8e7 a2a3 a7a5 d3d4 e5d4 f3d4 c8e6 b1c3 c6d4 e3d4 e7c6 d4c5 d6c5 d1d8 a8d8 f2f4 e8e7 a1d1 d8d1 e1d1 h8d8 d1c1 g7g6 c3d5 e6d5 e4d5 c6d4 f1d3 e7d6 h1e1 d8b8 e1e5 h7h5 h2h4 b7b5 c4b5 d4b5 d3b5 b8b5 e5e8 d6d5 e8e7 f7f5 c1c2 b5b6 e7c7 d5e4');
INSERT INTO evaluations VALUES(38709,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(38709,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(38709,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(38709,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(38709,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(38709,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(38709,36,'cp -36','c1e3 b8c6 g1f3 g8e7 d3d4 e5d4 f3d4 c5b4 b1c3 e8g8 f1e2 f7f5 e4f5 c6d4 e3d4 e7f5 e1g1 d8h4 a2a3 b4a5 d4e5 c8e6 e5g3 h4f6 c3d5 f6b2 a1b1 b2d4 d5e7 g8h8 e7f5 d4d1 f1d1 e6f5 b1b7 f5c2 d1c1 c2e4 b7b3 a5b6 g1f1 a7a5 f2f3 a5a4 b3b6 c7b6 g3d6 f8f7 c1d1 e4c2');
INSERT INTO evaluations VALUES(51975,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(51975,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(51975,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(51975,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(51975,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(51975,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(51975,36,'cp 92','d4d5 g8f6 b1c3 d7d6 d5e6 c8e6 g1e2 b8c6 e2f4 f8e7 g2g3 e8g8 f1g2 e6g4 f2f3 g4e6 f4e6 f7e6 f3f4 d8c7 e1g1 a8d8 f4f5 c7d7 f5e6 d7e6 c1f4 g8h8 a2a4 e6c4 c3d5 f6d5 e4d5 c6e5 b2b3 c4c3');
INSERT INTO evaluations VALUES(52027,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(52027,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(52027,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(52027,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(52027,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(52027,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(52027,36,'cp 70','g8f6 b1c3 d7d6 f1c4 f8e7 g1e2 e6d5 c3d5 e8g8 e1g1 f6d5 c4d5 b8a6 e2f4 a6c7 f1e1 f8e8 a2a4 c7d5 f4d5 c8d7 c1f4 d7c6 d1d3 e7f6 c2c3 h7h6 f4g3 c6d5 d3d5 d8b6 e1e2 b6a6 e2d2 f6g5 f2f4 g5f6 a1e1 a6a4 d5b7 a8b8 b7d5 b8d8 d5d3 a4c6');
INSERT INTO evaluations VALUES(44264,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(44264,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(44264,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(44264,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(44264,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(44264,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(44264,36,'cp 67','b1c3 d7d6 f1c4 f8e7 g1e2 e6d5 c3d5 f6d5 c4d5 e8g8 e1g1 b8a6 e2c3 a6b4 c1f4 f8e8 d5b3 c8e6 d1d2 e6b3 c2b3 d8d7 a1d1 a8d8 a2a3 b4c6 c3d5 f7f5 f2f3 e7f8 d2d3 d7f7 b3b4 f5e4 f3e4 f7g6 b4b5');
INSERT INTO evaluations VALUES(44263,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(44263,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(44263,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(44263,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(44263,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(44263,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(44263,36,'cp 70','d7d6 f1c4 f8e7 g1e2 e6d5 c3d5 f6d5 c4d5 e8g8 e1g1 b8a6 e2f4 a6c7 c2c3 f8e8 f2f3 c7d5 f4d5 c8e6 c1f4 e7f8 a2a4 f7f5 f4g3 d8d7 e4f5 e6f5 f1f2 d7f7 a4a5 a8d8 f2d2 e8e6 c3c4 d8e8 b2b3 e6g6 d5f4');
INSERT INTO evaluations VALUES(46033,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(46033,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(46033,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(46033,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(46033,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(46033,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(46033,36,'cp 63','d5e6 f7e6 g1f3 b8c6 c1f4 f8e7 e4e5 d6e5 f3e5 d8d4 e5c6 b7c6 f4g3 f6d5 f1e2 d5c3 d1d4 c5d4 b2c3 e8g8 c3d4 e7b4 e1f1 f8d8 a1d1 c6c5 h2h4 d8d4 d1d4 c5d4 g3e5 c8d7 e2d3 d7a4 f1e2 b4c3 h1b1 a4c6 h4h5 c6d5 f2f3 d5c6 b1b8 a8b8 e5b8');
INSERT INTO evaluations VALUES(18055,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(18055,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(18055,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(18055,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(18055,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(18055,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(18055,36,'cp 37','g8f6 e1g1 c8b7 d2d3 f8c5 b1c3 e8g8 a2a4 c6a5 a4b5 a5b3 c2b3 a6b5 a1a8 b7a8 f3e5 d7d5 c1g5 c5e7 c3b5 h7h6 g5h4 d5e4 d3e4 g7g5 h4g3 f6e4 e5c4 d8b8 b5c3 f8d8 d1c2 b8b7 c3e4 b7e4 c2e4 a8e4 g3c7 d8d3 c7e5 d3b3 e5c3 e4d3 f1e1 e7b4 c4d2 b4c3');
INSERT INTO evaluations VALUES(11699,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11699,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(11699,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11699,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(11699,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11699,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11699,36,'cp 29','e1g1 c8b7 d2d3 f8c5 a2a4 e8g8 h2h3 d7d6 b1c3 b5b4 c3d5 c6a5 d5f6 d8f6 b3a2 b7c8 c1e3 c8e6 e3c5 d6c5 a2e6 f6e6 f3h4 a5c6 h4f5 g7g6 f5e3 a8d8 e3c4 e6f6 d1g4 f6e7 f2f4 b4b3 c2c3 d8d3');
INSERT INTO evaluations VALUES(11698,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11698,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(11698,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(11698,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(11698,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11698,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(11698,36,'cp 24','c8b7 b1c3 f8e7 d2d3 e8g8 f1e1 d7d6 a2a3 f8e8 c3d5 f6d5 b3d5 d8c8 a3a4 c6d4 f3d4 b7d5 e4d5 e5d4 c1d2 c8b7 a4b5 a6b5 a1a8 b7a8 d1g4 e7f6 e1e8 a8e8 g4e4 e8d8 h2h3 d8d7 e4f3 h7h6');
INSERT INTO evaluations VALUES(24231,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(24231,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24231,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(24231,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(24231,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(24231,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(24231,36,'cp 37','b1c3 f8d6 d2d3 c6a5 d3d4 e5d4 d1d4 e8g8 c1g5 a5b3 a2b3 d6e7 e4e5 f6h5 g5e7 d8e7 c3d5 b7d5 d4d5 h5f4 d5e4 f4g6 f1e1 f8e8 e4d4 e7e6 a1d1 a8d8 h2h4 e6f5 d4c3 h7h6 g2g3 d7d6 e5d6 e8e1 d1e1 c7d6 f3d4 f5d7 e1e3 d8e8 e3e8 d7e8 c3c7');
INSERT INTO evaluations VALUES(24233,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(24233,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24233,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24233,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(24233,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(24233,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(24233,36,'cp 31','f8e7 d2d3 e8g8 f1e1 d7d6 a2a3 f8e8 c3d5 f6d5 b3d5 d8c8 a3a4 h7h6 c2c3 c6d8 d3d4 e5d4 f3d4 e7f6 a4b5 f6d4 c3d4 a6b5 d5b7 c8b7 a1a8 b7a8 d4d5 a8a2 d1d3 a2c4 d3c4 b5c4 f2f3 c7c6 g1f1 c6d5 e4d5 e8e1 f1e1 d8b7 e1d1 b7c5');
INSERT INTO evaluations VALUES(46705,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(46705,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(46705,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(46705,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(46705,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(46705,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(46705,36,'cp 65','c1e3 c7c6 g1f3 f8g7 h2h3 e8g8 a2a4 b8d7 f1e2 e7e5 d4e5 d7e5 d1d2 d8a5 f3d4 e5d7 e1g1 f8e8 f1e1 a5c7 f2f3 a7a6 a4a5 d6d5 e4d5 f6d5 c3d5 c6d5 c2c3 d7f6 e3f4 c7d8 e1d1 c8d7 e2f1 b7b5 a5b6 d8b6');
INSERT INTO evaluations VALUES(46699,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(46699,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(46699,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(46699,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(46699,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(46699,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(46699,36,'cp 76','c7c6 h2h3 f8g7 g1f3 e8g8 a2a4 d8c7 d1d2 d6d5 e3f4 c7b6 e4e5 f6e4 d2c1 e4c3 b2c3 c6c5 a4a5 b6c7 f1e2 b8c6 e1g1 c6a5 e5e6 c7d8 e6f7 f8f7 c1e3 d8f8 f4h2 c5d4 c3d4 a5c4 e2c4 d5c4 f1e1 a7a5');
INSERT INTO evaluations VALUES(44724,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(44724,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(44724,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(44724,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(44724,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(44724,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(44724,36,'cp 62','a2a4 f8g7 h2h3 e8g8 g1f3 b8d7 f1e2 e7e5 d4e5 d7e5 e1g1 e5f3 e2f3 c8e6 d1d2 d8c7 a4a5 f8d8 e3g5 d8d7 g5f4 f6e8 f1e1 g7e5 d2e3 e5f4 e3f4 d6d5 f4c7 e8c7 e4d5 c7d5 c3e4 a8e8 a1d1 d7e7 e4d6');
INSERT INTO evaluations VALUES(52484,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(52484,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(52484,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52484,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(52484,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(52484,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(52484,36,'cp 24','e2e4 e7e6 b1c3 b8c6 d2d4 c5d4 f3d4 g8f6 d4b5 d7d6 c1f4 e6e5 f4g5 a7a6 b5a3 b7b5 g5f6 g7f6 c3d5 f8g7 f1d3 c6e7 d5e7 d8e7 e1g1 e8g8 d1f3 f6f5 c2c4 d6d5 c4d5 f5e4 d3e4 a8b8 a1d1 b8b6 d5d6 b6d6 d1d6 e7d6 f1d1 d6b6 a3c2');
INSERT INTO evaluations VALUES(57234,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(57234,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(57234,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(57234,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(57234,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(57234,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(57234,36,'cp -20','d7d5 d2d3 g8f6 g1f3 f8d6 e2e4 d5e4 d3e4 e8g8 d1e2 b8d7 e1g1 d7c5 f3h4 g7g6 f1d1 d8e7 b1c3 c5e6 h4f3 c7c6 a2a4 a7a5 h2h3 b7b6 c1e3 c8a6 e2e1 d6c7 g2f1 a6f1 g1f1 f8e8 f3g5 e6d4 e3d4 e5d4 d1d4');
INSERT INTO evaluations VALUES(56487,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(56487,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(56487,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(56487,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(56487,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(56487,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(56487,36,'cp -10','d2d4 e5d4 g1f3 f8b4 b1d2 g8f6 f3d4 e8g8 e1g1 b4e7 e2e4 c8g4 f2f3 d5e4 d2e4 g4d7 c2c3 f8e8 c1e3 b8c6 d4c6 d7c6 d1d8 a8d8 e4f6 e7f6 g1f2 a7a5 f1e1 a5a4 a2a3 c6d5 a1d1 f6e7 f3f4 d5g2 f2g2 h7h5 e3g1 d8d1 e1d1 e8d8 g1d4');
INSERT INTO evaluations VALUES(56370,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(56370,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56370,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(56370,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(56370,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56370,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(56370,36,'cp -25','e5d4 g1f3 f8b4 b1d2 g8f6 f3d4 b4c5 d4b3 c5b6 a2a4 a7a5 e1g1 c8e6 e2e4 e8g8 e4e5 f6d7 d2f3 c7c5 f3g5 d7e5 g5e6 f7e6 d1e2 b8c6 c1f4 c5c4 f4e5 c4b3 c2c4 c6e5 e2e5 d8f6 e5f6 f8f6 c4d5 a8f8 d5e6 f6f2 f1f2 f8f2 g2d5 f2b2 g1h1');
INSERT INTO evaluations VALUES(22836,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(22836,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22836,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(22836,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(22836,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(22836,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(22836,36,'cp 12','g1f3 d7d5 e4d5 d8d5 f4e5 c8g4 b1c3 d5a5 f1e2 f8c5 d2d4 e8c8 c1d2 g4f3 d4c5 f3g2 h1g1 g2h3 e2g4 h3g4 d1g4 c8b8 e1c1 a5c5 g4g7 g8e7 g7f7 c6e5 f7f6 e7g6 d2h6 d8d1 g1d1 b7b6 f6e6 c5c4 e6c4 e5c4 d1d7 c4e5 d7g7 e5g4 h6f4 g6f4 g7g4 h8f8 g4g7 h7h5 g7g5 h5h4 c1d2 b8b7');
INSERT INTO evaluations VALUES(22834,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(22834,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(22834,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(22834,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(22834,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(22834,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22834,36,'cp 13','d7d5 e4d5 d8d5 f4e5 c8g4 b1c3 d5a5 f1e2 f8c5 d2d4 e8c8 c1d2 g4f3 d4c5 f3g2 h1g1 g2h3 e2g4 h3g4 d1g4 c8b8 e1c1 a5c5 g4g7 g8e7 g7f7 c6e5 f7f6 e7g6 d2h6 d8d1 g1d1 b7b6 f6e6 c5c6 e6c6 e5c6 d1f1 h8e8 h6d2 c6d4 f1f7 e8e7 f7e7 g6e7 c3e4 d4f3 d2f4 e7d5 f4g3 b8b7');
INSERT INTO evaluations VALUES(21723,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21723,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(21723,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(21723,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(21723,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(21723,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(21723,36,'cp 12','e4d5 d8d5 f4e5 c8g4 b1c3 d5a5 f1e2 f8c5 d2d4 e8c8 c1d2 g4f3 d4c5 f3g2 h1g1 g2h3 e2g4 h3g4 d1g4 c8b8 g4g7 a5c5 e1c1 g8e7 g7f7 c6e5 f7f4 d8f8 f4e3 c5e3 d2e3 e7g6 g1f1 e5g4 e3g1 b7b6 c3b5 f8f1 d1f1 h8e8 b5d4 b8b7');
INSERT INTO evaluations VALUES(21573,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21573,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(21573,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(21573,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(21573,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(21573,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21573,36,'cp 14','d8d5 f4e5 c8g4 b1c3 d5a5 f1e2 f8c5 d2d4 e8c8 c1d2 g4f3 d4c5 f3g2 h1g1 g2h3 e2g4 h3g4 d1g4 c8b8 g4g7 a5c5 e1c1 g8e7 g7f7 c6e5 f7f4 d8d2 f4d2 e5f3 c3e4 c5e5 d2c3 e5c3 e4c3 f3g1 d1g1 e7f5 g1g5 h8f8 g5h5 f8f7 c3e4 h7h6 c2c4 b7b6 c4c5 f7e7 e4g3 f5g3 h2g3 b8b7 h5h6 b6c5');
INSERT INTO evaluations VALUES(1720,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(1720,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(1720,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(1720,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(1720,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(1720,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(1720,36,'cp 11','f4e5 c8g4 b1c3 d5a5 f1e2 f8c5 d2d4 e8c8 c1d2 g4f3 d4c5 f3g2 h1g1 g2h3 e2g4 h3g4 d1g4 c8b8 g4g7 a5c5 e1c1 g8e7 g7f7 c6e5 f7f4 e5c4 g1g5 c5c6 f4e4 c4d2 e4c6 e7c6 d1d2 d8d2 c1d2 h8d8 d2c1 b8c8 c3d5 b7b6 g5h5 d8d7 h2h3 d7f7 c2c3 c6e7 d5e7 f7e7 c1d2 a7a5 a2a4 e7f7 d2d3 f7f2');
INSERT INTO evaluations VALUES(20078,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(20078,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(20078,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(20078,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(20078,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(20078,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(20078,36,'cp 57','c2c4 g7g6 f1e2 c6d4 d1d4 g8f6 c1e3 f8g7 e1g1 e8g8 d4d3 c8g4 b1c3 g4e2 d3e2 f6d7 a1c1 d8a5 f1d1 g7c3 c1c3 f8c8 d1c1 c8c7 h2h3 a8c8 g1h1 a7a6 e2d2 a5b4 a2a3 b4a4 d2e2 d7e5 f2f4 e5c4 f4f5');
INSERT INTO evaluations VALUES(20067,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(20067,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(20067,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(20067,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(20067,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(20067,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(20067,36,'cp 62','g8f6 b1c3 g7g6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 a7a5 e1g1 a5a4 e3d4 d8a5 f2f4 c8g4 a1c1 g4e2 d3e2 f6d7 d4g7 g8g7 c3b5 f8c8 g1h1 c8c5 f1d1 a8c8 b2b3 a5b4 c1b1 a4b3 a2b3 c5b5 c4b5 d7f6');
INSERT INTO evaluations VALUES(13362,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(13362,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(13362,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(13362,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(13362,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(13362,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(13362,36,'cp 68','b1c3 g7g6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 c8g4 f2f3 g4e6 e1g1 a7a6 g1h1 d8a5 e3d2 f6d7 b2b3 d7e5 d3c2 e5c6 c3d5 a5a3 d2c3 g7c3 d5c3 c6d4 c2d3 d4e2 c3e2 b7b5 c4b5 a6b5 e2d4 b5b4 d4e6 f7e6 d3d2 a8c8');
INSERT INTO evaluations VALUES(13360,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(13360,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(13360,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(13360,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(13360,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(13360,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(13360,36,'cp 56','g7g6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 c8e6 e1g1 f6d7 b2b3 d8a5 e3d4 g7d4 d3d4 a8c8 g1h1 d7f6 a1c1 a5c5 d4d2 b7b5 c4b5 c5b4 f2f3 c8c5 d2b2 f8c8 c3a4 c5c1 f1c1 c8c1 b2c1 b4d4 a4c3 g8f8 c1e1 d6d5 e4d5 e6d5 c3d5 f6d5 e2c4');
INSERT INTO evaluations VALUES(18058,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(18058,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(18058,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(18058,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(18058,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(18058,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(18058,36,'cp 49','a4b3 g8f6 e1g1 f8c5 a2a4 c8b7 d2d3 e8g8 b1c3 c6a5 a4b5 a5b3 c2b3 a6b5 a1a8 b7a8 f3e5 b5b4 c3a4 c5e7 e5f3 c7c5 c1e3 d7d6 f3d2 f8e8 f1e1 f6d7 d3d4 c5d4 e3d4 e7g5 d2c4 e8e4 e1e4 a8e4 c4d6 e4c6 a4c5 d7c5 d4c5 g5f6 c5b4');
INSERT INTO evaluations VALUES(18910,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(18910,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(18910,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(18910,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18910,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(18910,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(18910,36,'cp 37','b1c3 g8f6 d4c6 b7c6 e4e5 f6d5 c3e4 c8b7 f1e2 c6c5 e1g1 d8c7 e4d6 f8d6 e5d6 c7b6 c2c4 d5b4 a2a3 b4c6 b2b4 c6d4 c1f4 a7a5 e2d3 a5b4 d1g4 b4a3 f1b1 b6c6 b1b7 c6b7 g4g7 e8c8 g7e5 b7a7 f4e3 d4b3 a1a2 a7a5 g2g3 f7f6 e5f6 h8f8 f6e7 d8e8 e7g7');
INSERT INTO evaluations VALUES(34553,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(34553,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(34553,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(34553,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(34553,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(34553,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(34553,36,'cp 114','f1e2 e6b6 g1f3 c8g4 e1g1 g8f6 d2d4 e7e6 c1e3 b6a5 h2h3 g4f3 e2f3 c7c6 c3e2 b8d7 c2c4 f8e7 a2a3 a5c7 b2b4 e8g8 d1b3 a7a5 a1b1 a5b4 a3b4 f8d8 g2g3 d7f8 b4b5 f8g6 f1d1 a8a3 b3c2 d8a8 b5c6 b7c6 g1g2');
INSERT INTO evaluations VALUES(34554,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(34554,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(34554,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(34554,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(34554,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(34554,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(34554,36,'cp 111','e6b6 g1f3 g8f6 d2d4 c8g4 e1g1 e7e6 c1e3 b6a5 h2h3 g4f3 e2f3 c7c6 c3e2 f8e7 c2c4 e8g8 a2a3 a5c7 e2c3 b8d7 d1e2 f8e8 a1d1 a8d8 f1e1 e7f8 e3g5 h7h6 g5h4 f8e7 c4c5 d7f8 h4g3 c7c8 b2b4 f6d5 c3e4');
INSERT INTO evaluations VALUES(34614,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(34614,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(34614,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(34614,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(34614,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(34614,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34614,36,'cp 105','f1e2 c7c6 g1f3 e5c7 d2d4 c8f5 e1g1 b8d7 d4d5 e7e5 d5e6 f5e6 f3g5 e8c8 e2g4 e6g4 d1g4 g8h6 g4h3 c7d6 f1e1 f8e7 g5f3 h8e8 c3e4 d6e6 h3e6 f7e6 c1h6 g7h6 g2g3 c8c7 g1g2 d7f6 c2c3 d8d5 e4d2 e7d6 a1d1 d6c5 d2c4');
INSERT INTO evaluations VALUES(34615,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(34615,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(34615,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(34615,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(34615,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(34615,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(34615,36,'cp 78','c8g4 d2d4 g4e2 g1e2 e5a5 b2b4 a5h5 d1d3 e7e6 a2a3 c7c6 e2f4 h5f5 d3f5 e6f5 e1g1 f8d6 f1e1 g8e7 c3a4 b8d7 f4h5 h8g8 a4c5 e8c8 c1g5 f7f6 g5d2 d6c5 b4c5 e7d5 c2c4 d5c7 d2f4 g7g6 f4c7 c8c7 h5f4 d8e8 g2g3 c7c8 f2f3 g6g5 f4h5 b7b6 c5b6');
INSERT INTO evaluations VALUES(32866,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(32866,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(32866,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(32866,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(32866,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(32866,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(32866,36,'cp 108','d2d4 g4e2 g1e2 e5a5 b2b4 a5h5 d1d3 e7e6 a2a3 c7c6 e2f4 h5f5 d3f5 e6f5 c1b2 a7a5 e1g1 a5b4 a3b4 a8a1 f1a1 f8b4 a1a7 b4d6 c3e2 g8e7 a7b7 e8g8 b2c1 f8d8 f4d3 d8d7 b7b3 d7d8 c2c4 b8d7 g1f1 d7f8 c1a3 d6a3 b3a3');
INSERT INTO evaluations VALUES(32858,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(32858,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(32858,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(32858,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(32858,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(32858,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(32858,36,'cp 113','g4e2 g1e2 e5a5 b2b4 a5h5 d1d3 e7e6 a2a3 b8d7 e1g1 f8d6 c1f4 d6f4 e2f4 h5f5 d3f5 e6f5 f1e1 e8d8 c3d5 g8f6 d5e3 g7g6 e3c4 h8e8 f2f3 b7b6 g1f1 a7a6 c4e5 d7e5 d4e5 f6d7 e5e6 f7e6 f4e6 d8c8 e6g5 e8e1 a1e1 h7h6 g5f7 c8b7 f7h6 a6a5 h6f7 a5b4');
INSERT INTO evaluations VALUES(32861,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(32861,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(32861,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(32861,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(32861,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(32861,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(32861,36,'cp 116','g1e2 e5a5 b2b4 a5f5 e1g1 g8f6 a1b1 e7e6 e2g3 f5g4 d1g4 f6g4 d4d5 f8d6 c3b5 g4f6 c2c4 e8g8 b5d6 c7d6 f1d1 e6d5 c4d5 f8e8 c1b2 b8d7 g3f5 e8e4 b2d4 f6e8 f5e3 e8c7 a2a4 d7b6 e3f5 b6d5 d4g7 e4b4 b1b4 d5b4 g7h6 c7e6 f5d6');
INSERT INTO evaluations VALUES(32860,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(32860,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(32860,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(32860,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(32860,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(32860,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(32860,36,'cp 93','e5a5 b2b4 a5f5 e1g1 g8f6 a1b1 e7e6 e2g3 f5g4 d1g4 f6g4 d4d5 f8d6 f1e1 e8g8 c3b5 e6d5 h2h3 g4e5 g3f5 b8c6 b5d6 c7d6 b4b5 c6a5 c1a3 a5c4 f5e7 g8h8 a3d6 c4d6 e1e5 f8e8 e5d5 e8e7 d5d6 a8c8 b1d1 g7g6 d6d7 c8c7 d7e7 c7e7 c2c4 e7c7 d1c1 c7c5 g2g4 h8g8 a2a3 a7a5 b5a6 b7a6 g1g2');
INSERT INTO evaluations VALUES(32904,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(32904,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(32904,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(32904,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(32904,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(32904,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(32904,36,'cp 118','b2b4 a5h5 d1d3 e7e6 a2a3 b8c6 b4b5 c6d8 c1f4 a8c8 f4e5 f7f6 e5g3 f8d6 e1g1 g8e7 c3e4 d8f7 c2c4 e8g8 f1e1 e7g6 e4d6 c7d6 a1c1 f8e8 a3a4 e6e5 d4e5 d6e5 c4c5 e8d8 d3e4 c8c7 f2f3 g6e7 e4b4 h7h6 g3f2 h5g6 e2g3 g8h8');
INSERT INTO evaluations VALUES(39576,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(39576,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(39576,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(39576,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(39576,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(39576,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(39576,36,'cp 63','e2e4 g8f6 b1c3 d7d6 g1f3 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 a2a3 b8d7 c1e3 f6e8 b2b4 d8e7 h2h3 g4f3 e2f3 f7f5 a1b1 e7h4 b4a5 f5f4 e3d2 a8a5 b1b7 d7c5 b7b1 a5a3 c3b5 a3a8 d1c2 h7h5 b1a1 a8b8 a1a7 g7f6 b5c7 e8c7');
INSERT INTO evaluations VALUES(39577,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(39577,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(39577,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(39577,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(39577,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(39577,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(39577,36,'cp 62','g8f6 b1c3 e8g8 f1e2 d7d6 g1f3 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 f6e8 a2a3 b8d7 a1b1 f7f5 b2b4 b7b6 c1d2 g8h8 c3b5 d8h4 d1c1 h4d8 b4a5 b6a5 d2g5 g7f6 g5e3 f5f4 e3a7 f6e7 f3g4 e8f6 g4d7 f6d7 c1c2');
INSERT INTO evaluations VALUES(43197,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(43197,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(43197,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(43197,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(43197,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(43197,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(43197,36,'cp 69','b1c3 e8g8 g1f3 d7d6 f1e2 e7e5 e1g1 b8a6 f1e1 e5d4 f3d4 a6c5 e2f1 f6g4 h2h3 d8f6 h3g4 f6d4 d1f3 c8d7 c1e3 d4f6 f3f6 g7f6 c3d5 f6b2 a1b1 b2e5 d5c7 a8c8 c7d5 g8h8 f2f3 b7b6 f1e2 f8e8 b1c1 d7a4 g1f2 f7f6 e1h1 h8g7 g2g3 g6g5 c1b1 h7h6');
INSERT INTO evaluations VALUES(49581,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(49581,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(49581,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(49581,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(49581,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(49581,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(49581,36,'cp 68','d4e5 f6g4 e2e3 g4e5 f2f4 e5c6 g1e2 f8c5 b1c3 e8g8 e2g3 f8e8 f1e2 d7d6 e1g1 b8d7 c3e4 d7f6 e4c5 d6c5 c1d2 c8g4 e2g4 f6g4 d1g4 d8d2 g3f5 d2b2 a1b1 b2f6 b1b7 c6e7 f5g3 e7g6 b7c7 e8e3 f4f5 g6f8 c7c5 a8e8 c5d5 f6c3 g4d4 c3d4 d5d4 e3a3 c4c5 a3a2');
INSERT INTO evaluations VALUES(49500,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(49500,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(49500,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(49500,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(49500,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(49500,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(49500,36,'cp 72','f6g4 e2e3 g4e5 f2f4 e5c6 b1c3 g7g6 d1d2 f8g7 b2b3 d7d6 c1b2 e8g8 g1e2 a7a5 e2g3 b8d7 f1e2 d7c5 e1g1 c6b4 a1d1 c8d7 f4f5 g7e5 c3b5 d7f5 g3f5 g6f5 b2e5 d6e5 f1f5 d8d2 d1d2 c5e4 d2d7 b4a2 b5c7 a8d8 d7d8 f8d8 f5e5');
INSERT INTO evaluations VALUES(49850,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(49850,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(49850,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(49850,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(49850,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(49850,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(49850,36,'cp 73','e2e3 g4e5 f2f4 e5c6 b1c3 d7d6 b2b3 g7g6 c1b2 f8g7 d1d2 e8g8 g1e2 a7a5 e2g3 b8d7 f1e2 d7c5 e1g1 c6b4 a1d1 c8d7 c3e4 c5e4 g3e4 g7b2 d2b2 f7f5 e4c3 b4a6 e2f3 d7c6 c3d5 a6b4 a2a3 b4d5 c4d5 c6b5 f1e1 d8e7 e3e4 f5e4 e1e4 e7f6 b2f6 f8f6 d1c1');
INSERT INTO evaluations VALUES(49847,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(49847,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(49847,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(49847,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(49847,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(49847,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(49847,36,'cp 71','g4e5 f2f4 e5c6 b1c3 g7g6 b2b3 f8g7 d1d2 d7d6 c1b2 e8g8 g1e2 a7a5 e2g3 b8a6 f1e2 a6c5 e1g1 c6b4 f4f5 g7e5 f1f3 d8h4 c3d5 e5b2 d2b2 b4d5 c4d5 c8d7 a1f1 a8e8 b2c3 h4b4 c3c1 b4h4 f3f4 h4g5 e3e4 c7c6 c1c3 c6d5 e4d5 g5e7 f5g6 f7g6 f4f8 e8f8 f1f8 g8f8 c3a5');
INSERT INTO evaluations VALUES(49876,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(49876,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(49876,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(49876,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(49876,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(49876,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(49876,36,'cp 81','f2f4 e5c6 b1c3 d7d6 b2b3 g7g6 c1b2 f8g7 d1d2 e8g8 g1e2 c8g4 e2g3 b8d7 f1d3 a7a5 d3b1 a5a4 c3a4 g7b2 d2b2 d8e7 e1f2 f8e8 h1e1 h7h5 h2h3 h5h4 g3f1 g4e6 a4c3 e7f6 a2a3 c6e5');
INSERT INTO evaluations VALUES(49873,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(49873,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(49873,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(49873,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(49873,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(49873,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(49873,36,'cp 56','e5c6 b1c3 g7g6 d1d2 f8g7 b2b3 e8g8 c1b2 d7d6 g1e2 c8g4 e2g3 b8d7 f1d3 a7a5 e1g1 d7c5 d3e4 h7h5 a1e1 a5a4 c3a4 c5a4 b3a4 h5h4 g3e2 g4e2 d2e2 a8a4 f4f5 g7b2 e2b2 a4c4 e4d3 c4b4 b2c2 c6e5 f5g6 f7g6 a2a3 b4g4 d3e2');
INSERT INTO evaluations VALUES(55588,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(55588,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(55588,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(55588,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(55588,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(55588,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(55588,36,'cp 53','g1f3 e5d4 f3d4 g8f6 b1c3 f8e7 c1f4 b8c6 d1d2 e8g8 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 h1e1 a8d8 f4e3 d8d1 c3d1 f8e8 g2g3 e7d6 a2a3 c7c5 f2f3 b7b5 d1c3 d6c6 h2h4 b5b4 c3e4 f6e4 e3e4 c6b5 a3b4 c5b4');
INSERT INTO evaluations VALUES(55042,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(55042,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(55042,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(55042,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(55042,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(55042,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(55042,36,'cp 81','c4d5 d8d5 b1c3 d5d6 d2d4 e7e5 d4e5 d6d1 e1d1 c8e6 c3b5 b8a6 g1f3 e8c8 c1d2 a6c5 b5a7 c8b8 a7b5 c5e4 d1e1 e4d2 f3d2 g8e7 e2e3 e7c6 d2f3 f8b4 b5c3 e6d5 f1b5 d5f3 g2f3 c6e5 f3f4 b4c3 b2c3 e5d3 e1e2 d3c5 f2f3 c7c6 b5c4 f7f6 a1d1 b8c7 c4b3 d8d1 h1d1 h8a8');
INSERT INTO evaluations VALUES(55018,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(55018,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(55018,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(55018,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(55018,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(55018,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(55018,36,'cp 77','d8d5 b1c3 d5d6 d2d4 e7e5 d4e5 d6d1 e1d1 c8e6 c3b5 b8a6 c1d2 e8c8 g1f3 a6c5 b5a7 c8b8 a7b5 c5e4 d1e1 e4d2 f3d2 f8b4 b5c3 g8e7 e2e3 e7c6 d2f3 e6d5 f1b5 d5f3 g2f3 c6e5 f3f4 b4c3 b2c3 e5d3 e1e2 d3c5 f2f3 f7f6 b5c4 c7c6 a1d1 b8c7 d1d8 h8d8 h1d1 d8d1 e2d1 c7d6 d1d2 c5a4 c4b3');
INSERT INTO evaluations VALUES(34602,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(34602,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(34602,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(34602,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(34602,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(34602,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(34602,36,'cp 76','b1c3 d5d6 d2d4 e7e5 d4e5 d6d1 c3d1 f8b4 d1c3 b8c6 g1f3 c8g4 c1d2 g8e7 a2a3 b4c3 d2c3 e7d5 c3d2 e8c8 d2g5 d8d7 e1c1 d5b6 d1d7 b6d7 g5f4 h8e8 h2h4 d7c5 h4h5 c5e4 h1h4 g4f5 g2g4 f5d7 h5h6 g7h6 f4e3 c6e5');
INSERT INTO evaluations VALUES(34594,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(34594,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(34594,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(34594,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(34594,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(34594,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(34594,36,'cp 62','d5d6 d2d4 e7e5 d4e5 d6d1 e1d1 c8e6 c3b5 b8a6 c1d2 e8c8 g1f3 a6c5 b5a7 c8b8 a7b5 c5e4 d1e1 e4d2 f3d2 f7f6 b5c3 f6e5 a2a3 g8f6 e2e3 h7h6 f1e2 f8e7 d2f3 f6d7 b2b4 e6f7 f3d2 d7f6 h2h3 c7c6 a1c1');
INSERT INTO evaluations VALUES(34543,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34543,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(34543,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(34543,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(34543,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(34543,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(34543,36,'cp 89','d2d4 e7e5 d4e5 d6d1 e1d1 c8e6 c3b5 b8a6 g1f3 e8c8 c1d2 a6c5 d1e1 c5e4 b5a7 c8b8 a7b5 e4d2 f3d2 f8b4 b5c3 g8e7 e2e3 e7c6 d2f3 e6d5 f1b5 d5f3 g2f3 c6e5 e1e2 g7g5 a1d1 c7c6 b5a4 h7h6 d1d8 h8d8 a4b3 b4c3 b2c3 b8c7 h1g1 f7f6 f3f4 g5f4 e3f4 e5d3 e2f3 d3c5 b3c2 d8a8');
INSERT INTO evaluations VALUES(34537,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(34537,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(34537,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(34537,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34537,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(34537,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(34537,36,'cp 63','e7e5 d4e5 d6d1 c3d1 f8b4 d1c3 b8c6 g1f3 g8e7 c1d2 c8g4 a2a3 b4c3 d2c3 e7d5 f3d4 d5c3 d4c6 b7c6 b2c3 a8b8 f2f3 g4e6 g2g3 e8e7 f1g2 b8b2 e1f2 e6c4 h1e1 h8b8 f3f4 c6c5 g2e4 g7g6 a1b1 a7a5 f2e3 a5a4 e3f2 b8b3 b1b2 b3b2');
INSERT INTO evaluations VALUES(47104,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(47104,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47104,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(47104,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(47104,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(47104,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47104,36,'cp 13','f6d5 f1c4 e7e6 g1f3 f8e7 e1g1 e8g8 f1e1 c7c5 c4d5 e6d5 d2d4 b8c6 d4c5 d5d4 c3e4 c8f5 c1f4 f5e4 e1e4 e7c5 a2a3 a7a5 d1d3 d8b6 a1b1 a5a4 h2h3 h7h6 f3d2 a8e8 d3g3 c6e7 b2b4 a4b3 g3b3 b6b3 d2b3 c5a3 b3d4 e7c6 e4e8');
INSERT INTO evaluations VALUES(47457,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47457,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(47457,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47457,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(47457,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(47457,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47457,36,'cp 23','f1c4 e7e6 g1f3 f8e7 e1g1 e8g8 f1e1 c7c5 c4d5 e6d5 d2d4 b8c6 d4c5 d5d4 c3e4 c8f5 c1f4 f5e4 e1e4 e7c5 a2a3 a7a5 d1d3 d8b6 a1b1 a5a4 h2h4 h7h6 f4d2 a8e8 b2b4 a4b3 b1b3 b6a6 d3a6 b7a6 e4e8 f8e8 g1f1 c5d6 a3a4');
INSERT INTO evaluations VALUES(47434,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(47434,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(47434,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(47434,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47434,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47434,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47434,36,'cp 11','e7e6 g1f3 d5c3 b2c3 c7c5 e1g1 f8d6 d2d4 e8g8 d4d5 e6d5 d1d5 b8c6 f3g5 d6e7 f1e1 d8d5 c4d5 h7h6 d5c6 e7g5 c1g5 b7c6 g5e7 f8e8 e7c5 e8e1 a1e1 c8f5 e1e7 a7a5 h2h4 f5c2 f2f3 c2d3 e7b7 a5a4 a2a3 a8a5 c5b4 a5d5 b7c7 d3b5 g2g4 h6h5 g1f2 d5d2 f2g3 h5g4 f3g4 d2e2 c7c8 g8h7');
INSERT INTO evaluations VALUES(48078,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(48078,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48078,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(48078,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(48078,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(48078,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48078,36,'cp 17','g1f3 d5c3 b2c3 c7c5 e1g1 f8d6 d2d4 e8g8 d4d5 e6d5 d1d5 b8c6 f3g5 d6e7 f1e1 d8d5 c4d5 h7h6 d5c6 e7g5 c1g5 b7c6 g5f4 f7f6 e1e2 c8f5 a1e1 f8d8 f2f3 f5g6 h2h4 d8e8 g2g4 g6f7 e2e8 a8e8 e1e8 f7e8 f4e3 e8g6 e3c5 g6c2 c5a7');
INSERT INTO evaluations VALUES(48076,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(48076,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(48076,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(48076,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(48076,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(48076,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(48076,36,'cp 20','d5c3 b2c3 c7c5 e1g1 f8d6 d2d4 e8g8 d4d5 e6d5 d1d5 b8c6 f3g5 d6e7 f1e1 d8d5 c4d5 h7h6 d5c6 e7g5 c1g5 b7c6 g5e7 f8e8 e7c5 e8e1 a1e1 c8f5 e1e2 a8b8 f2f3 a7a5 g1f2 f5g6 a2a3 f7f6 e2e6 b8b2 f2g3 b2c2 c5d4 c2a2');
INSERT INTO evaluations VALUES(47144,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(47144,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(47144,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47144,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47144,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47144,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47144,36,'cp 0','e7e6 g1f3 b7b6 f1d3 f8d6 b2b3 e8g8 e1g1 c8b7 c1b2 d5c4 d3c4 b8d7 b1c3 a7a6 a2a4 d6b4 c3a2 b4d6');
INSERT INTO evaluations VALUES(48263,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48263,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48263,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48263,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48263,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(48263,36,'cp 0','g1f3 b7b6 b1c3 c8b7 c4d5 e6d5 f1b5 c7c6 b5d3 f8d6 e3e4 f6e4 c3e4 d5e4 d3e4 h7h6 e1g1 b8d7 e4c2 e8g8 d1d3 d7f6 f3e5 f8e8 d3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(48264,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48264,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48264,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48264,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48264,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(48264,36,'cp 0','b7b6 f1d3 c8b7 e1g1 d5c4 d3c4 a7a6 b2b3 f8e7 c1b2 b8d7 b1d2 e8g8 a1c1 c7c5 a2a4 c5d4 f3d4 e7b4 d4c2 b4e7');
INSERT INTO evaluations VALUES(43797,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(43797,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43797,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(43797,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(43797,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43797,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43797,36,'cp 2','b1c3 c8b7 f1d3 d5c4 d3c4 a7a6 e1g1 b6b5 c4b3 c7c5 e3e4 c5d4 f3d4 b8d7 c1e3 a8c8 d1e2 f8c5 f1d1 d8e7 e4e5 d7e5 d4b5 e8g8 e3c5 c8c5 b5d6 e5g6 e2e3 c5e5 e3a7 b7c6 a7e7 g6e7 d1d2 e5c5 a1d1 h7h6 f2f3 e7g6 c3e4 c6e4 d6e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(43795,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43795,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43795,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43795,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43795,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43795,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(43795,36,'cp 0','f8d6 b2b3 e8g8 f1d3 c8b7 c1b2 b8d7 e1g1 a7a6 d1c2 d5c4 b3c4 c7c5 c3e4 b7e4 d3e4 f6e4 c2e4 c5d4 e4d4 d7f6 a1d1 d6c5 d4f4 d8e7 g2g4 f6d7 d1d7 e7d7 b2g7 g8g7 f4g5 g7h8 g5f6 h8g8 f6g5');
INSERT INTO evaluations VALUES(55843,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(55843,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(55843,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(55843,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(55843,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(55843,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(55843,36,'cp 27','e2e4 g8f6 e4e5 f6d7 f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 c3e2 d8b6 d1c1 f7f6 c2c3 f8e7 e5f6 d7f6 d4c5 e7c5 e3c5 b6c5 e2d4 c6d4 f3d4 e8g8 f1d3 c8d7 e1g1 a8e8 c1e3 c5b6 e3e2 d7c6 a2a4 f6e4 a4a5 b6c7 d3e4 d5e4 d4c6 c7c6');
INSERT INTO evaluations VALUES(55815,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55815,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55815,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55815,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55815,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55815,36,'cp 0','g8f6 g1f3 f8e7 f1e2 d5c4 e1g1 e8g8 e2c4 a7a6 b2b3 b8d7 c1b2 c7c5 a2a4 b7b6 b1d2 c5d4 f3d4 c8b7 a1c1 e7b4 d4c2 b4e7');
INSERT INTO evaluations VALUES(39205,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(39205,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(39205,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(39205,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(39205,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(39205,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(39205,36,'cp 21','g1f3 b8c6 c2c3 g8f6 d2d3 a7a5 e1g1 d7d6 f1e1 e8g8 h2h3 h7h6 c4b3 c5a7 a2a4 c8e6 b1a3 f8e8 a3c4 d8e7 c1e3 a7e3 c4e3 e7d7 d1c2 e6b3 c2b3 c6e7 f3h4 d7c6 a1d1 a8d8 e3c4 b7b6 b3c2 e7g6 h4g6');
INSERT INTO evaluations VALUES(39202,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(39202,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(39202,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(39202,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(39202,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(39202,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(39202,36,'cp 6','b8c6 d2d3 d7d6 c2c3 g8f6 e1g1 e8g8 f1e1 a7a6 c4b3 h7h6 b1d2 c8e6 h2h3 e6b3 d2b3 c5a7 c1e3 a7e3 e1e3 c6e7 d3d4 e7g6 d4e5 d6e5 d1d8 f8d8 b3a5 a8b8 g2g3 f6e8 e3e1 f7f6 g1g2 e8d6 a1d1 g8f7 a2a4 g6f8');
INSERT INTO evaluations VALUES(16126,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(16126,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(16126,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(16126,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(16126,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(16126,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(16126,36,'cp 66','d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 e1g1 d7d6 e4e5 d6e5 d1d8 e8d8 c4f7 c8f5 f7b3 d8c8 a2a3 b4d6 c1e3 c8b7 a1d1 a7a5 h2h3 f6d7 c3a4 h8f8 f1e1 f8d8 e1e2 d7b6 e2d2 b6a4 b3a4 f5e6 a4b3 e6b3 c2b3 a5a4 b3b4 d8f8 g2g3 a8b8 g1g2 b8d8 f2f3');
INSERT INTO evaluations VALUES(16154,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(16154,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(16154,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(16154,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(16154,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(16154,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(16154,36,'cp 51','e5d4 f3d4 f8b4 d4c6 b7c6 e4e5 d7d5 e5f6 d8f6 c4d3 d5d4 d1e2 e8f8 e1g1 d4c3 b2c3 b4d6 f1e1 c8f5 c1e3 f5d3 e2d3 f8g8 g2g3 f6g6 d3g6 f7g6 a1b1 g8f7 b1b7 h8e8 e1d1 e8b8 d1b1 b8b7 b1b7 a7a5 c3c4 f7e6 h2h4');
INSERT INTO evaluations VALUES(16186,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(16186,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(16186,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(16186,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(16186,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(16186,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(16186,36,'cp 52','f3d4 f8b4 d4c6 b7c6 e4e5 d7d5 e5f6 d8f6 c4d3 d5d4 d1e2 e8f8 e1g1 d4c3 b2c3 b4d6 f1e1 c8f5 c1e3 f5d3 e2d3 f8g8 g2g3 f6g6 a1c1 h6h5 d3g6 f7g6 c1d1 g8f7 c3c4 h8e8 c4c5 d6e5 g1g2 e5f6 d1d7 e8e7 d7e7 f6e7 e1b1 f7e6 g2f3 a8f8 f3e2');
INSERT INTO evaluations VALUES(55738,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(55738,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(55738,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(55738,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(55738,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(55738,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(55738,36,'cp 95','g1f3 g8f6 f1d3 c7c5 c2c3 b8d7 e1g1 f8e7 f1e1 e8g8 h2h3 a7a6 a2a4 f8e8 a4a5 c5d4 c3d4 b7b5 a5b6 c8b7 b1c3 d8b6 c1e3 b6d8 d1e2 d8b8 f3d2 e8c8 e3f4 e6e5 f4e3 d6d5 d4e5 b8e5 d2f3 e5c7 e4d5 f6d5 d3a6 d5c3 b2c3');
INSERT INTO evaluations VALUES(55737,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(55737,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(55737,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(55737,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(55737,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(55737,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(55737,36,'cp 103','g8f6 f1d3 c7c5 c2c3 b8d7 e1g1 f8e7 f1e1 e8g8 h2h3 a7a6 a2a4 f8e8 b1a3 c5d4 c3d4 h7h6 a3b1 d7f8 c1e3 c8d7 b1c3 b7b5 a4b5 a6b5 c3b5 a8a1 d1a1 f6e4 e1c1 d8b6 b5c7 e8c8 c7e6 c8c1 a1c1 f8e6 d3e4 e7f6 c1c3 e6g5 f3g5 h6g5 e4d5');
INSERT INTO evaluations VALUES(48059,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(48059,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(48059,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(48059,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(48059,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(48059,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(48059,36,'cp 104','f1d3 c7c5 c2c3 b8d7 e1g1 f8e7 f1e1 e8g8 h2h3 a7a6 a2a4 d8c7 b1d2 b7b6 d4d5 e6d5 e4d5 f8e8 c3c4 g7g6 a4a5 b6b5 b2b3 e7f8 e1e8 f6e8 c1b2 f8g7 b2g7 e8g7 d2e4 f7f5 e4g5 d7f6 c4b5 a6b5 d3b5');
INSERT INTO evaluations VALUES(48057,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(48057,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(48057,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(48057,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(48057,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(48057,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(48057,36,'cp 103','c7c5 c2c3 b8d7 e1g1 f8e7 f1e1 e8g8 h2h3 a7a6 a2a4 d8c7 b1d2 b7b6 d4d5 e6d5 e4d5 f8e8 c3c4 g7g6 d2f1 c8b7 f1e3 e7f8 c1d2 b6b5 c4b5 a6b5 d3b5 f6d5 a4a5 d5f6 d2c3 f8g7 e3c4 d6d5 a5a6 e8e1 d1e1 d5c4');
INSERT INTO evaluations VALUES(57263,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(57263,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(57263,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(57263,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(57263,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(57263,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(57263,36,'cp -59','c3b2 c1b2 g8f6 e4e5 f8b4 b2c3 d8e7 d1e2 f6g8 g1f3 g8h6 a2a3 b4c3 b1c3 e8g8 e1g1 b8c6 c3d5 e7d8 h2h3 d7d6 f1e1 c8e6 a1d1 h6f5 d5f4 e6c4 e2c4 f5h4 f3h4 d8h4 e5d6 c7d6 d1d6 a8d8 g2g3 h4g5 d6d5 g5f6 h3h4 h7h6 c4e4 g8h8 g1g2 d8d5 f4d5 f6d6');
INSERT INTO evaluations VALUES(57267,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(57267,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(57267,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(57267,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(57267,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(57267,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(57267,36,'cp -45','c1b2 g8f6 b1c3 d7d5 c3d5 f6d5 c4d5 f8b4 e1f1 d8e7 a2a3 b4c5 d1f3 c7c6 b2g7 h8g8 g7f6 e7f8 d5b3 c8g4 f3c3 b8d7 f6h4 c5e7 h4e7 f8e7 f2f3 g4e6 b3e6 e7e6 g1e2 e8c8 f1f2 f7f5 e4f5 e6f5 h1d1 d7e5 e2g3 e5d3 f2g1 f5c5 c3c5 d3c5 a1c1 b7b6 d1e1 d8d2');
INSERT INTO evaluations VALUES(57264,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(57264,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(57264,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(57264,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(57264,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(57264,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(57264,36,'cp -62','g8f6 e4e5 f8b4 b2c3 d8e7 d1e2 f6g8 g1f3 g8h6 e1g1 e8g8 c3b4 e7b4 a2a3 b4e7 b1c3 b8c6 c3d5 e7d8 f1e1 d7d6 h2h3 c8e6 a1d1 h6f5 d5f4 e6c4 e2c4 f5h4 f3h4 d8h4 e5d6 c7d6 d1d6 a8d8 g2g3 h4g5 d6d5 g5f6 c4e4 a7a6 h3h4 h7h6 g1g2 d8d5 f4d5 f6d6 d5b4 c6b4 e4b4 d6c6 b4e4 c6e4 e1e4 f8a8 e4e7 b7b5 e7d7 g8f8 d7b7 g7g6 b7b6 f8g7 h4h5 g6h5');
INSERT INTO evaluations VALUES(49729,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(49729,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(49729,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(49729,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(49729,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(49729,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(49729,36,'cp -60','e4e5 f8b4 b2c3 d8e7 d1e2 f6g8 a2a3 b4c3 b1c3 g8h6 g1f3 e8g8 e1g1 b8c6 f1e1 d7d6 c3d5 e7d8 e5d6 c7d6 d5e7 g8h8 a1b1 c8g4 c4d5 g4f3 d5f3 c6d4 e2e4 d4f3 e4f3 b7b6 f3b7 h6g8 b1c1 d6d5 e7d5 d8d6 c1c7 a8d8 e1d1 d6a3 b7a7');
INSERT INTO evaluations VALUES(49498,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(49498,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(49498,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(49498,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(49498,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(49498,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(49498,36,'cp -66','f8b4 b2c3 d8e7 d1e2 f6g8 g1f3 g8h6 e1g1 e8g8 a2a3 b4c3 b1c3 b8c6 f1e1 d7d6 c3d5 e7d8 a1d1 f8e8 e2b2 a8b8 h2h3 c8e6 e5d6 c7d6 b2d2 h6f5 d2f4 b8c8 c4d3 d8d7 f3g5 c6e5 g5e6 f7e6 d3f5 e6f5 d5e3 g7g6 e3d5 d7f7 d5e3 b7b6 f4b4 f5f4 e3g4 e5g4 e1e8 f7e8 h3g4');
INSERT INTO evaluations VALUES(42724,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(42724,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(42724,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(42724,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(42724,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(42724,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(42724,36,'cp -66','b2c3 d8e7 d1e2 f6g8 g1f3 g8h6 e1g1 e8g8 a2a3 b4c3 b1c3 b8c6 c3d5 e7d8 f1e1 d7d6 e5d6 c7d6 d5e7 c6e7 e2e7 a8b8 c4d5 c8g4 e7e3 g4f3 d5f3 d8a5 e3d3 f8e8 e1d1 g7g6 d3d6 h6f5 d6f4 b8d8 f4b4 d8d1 a1d1 a5b4 a3b4 e8e7 g2g3');
INSERT INTO evaluations VALUES(52358,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(52358,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(52358,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(52358,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(52358,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(52358,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(52358,36,'cp -43','d7d5 e4d5 c6d5 c4b3 b8c6 d2d4 g8f6 h2h3 c8f5 c2c3 e7e6 c1f4 f8e7 g1f3 e8g8 e1g1 b7b5 a2a3 d8b6 b1d2 a7a5 d1e2 f6d7 f1e1 f8c8 d2f1 d7f6 e1c1 a5a4 b3c2 f5c2 c1c2 h7h6 f3e5 c6a5 e5d3 a5c4 a1e1');
INSERT INTO evaluations VALUES(52730,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(52730,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(52730,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(52730,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(52730,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(52730,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(52730,36,'cp -32','e4d5 c6d5 c4b3 g8f6 h2h3 b8c6 d2d4 c8f5 c2c3 e7e6 c1f4 f8d6 f4d6 d8d6 g1f3 b7b5 b3c2 f5c2 d1c2 e8g8 e1g1 b5b4 a2a3 a7a5 a3b4 a5b4 b1d2 h7h6 f1c1 c6e7 d2b3 b4c3 b2c3 f6d7 b3d2 a8a1 c1a1 f8c8 c3c4 d5c4 d2c4');
INSERT INTO evaluations VALUES(52702,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(52702,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(52702,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(52702,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(52702,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(52702,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(52702,36,'cp -46','c6d5 c4b3 c8f5 d2d4 e7e6 g1f3 f8d6 e1g1 h7h6 c2c4 d5c4 b3c4 g8f6 b1c3 e8g8 d4d5 e6d5 c3d5 b8c6 c1e3 a8c8 c4b3 f8e8 f1e1 d6f8 f3d4 f5d7 d4f3 f6d5 d1d5 d7e6 d5d8 c8d8 b3e6 e8e6');
INSERT INTO evaluations VALUES(53352,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(53352,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(53352,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(53352,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(53352,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(53352,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(53352,36,'cp -31','c4b5 c8d7 a2a4 a7a6 b5d7 b8d7 d2d4 e7e6 g1e2 g8e7 c2c3 e7g6 c1f4 b7b5 a4b5 a6b5 a1a8 d8a8 f4g3 f8e7 h2h4 h7h5 e2f4 g6f4 g3f4 a8a1 d1c2 e8g8 e1g1 e7h4 c2e2 f8a8 e2b5');
INSERT INTO evaluations VALUES(53150,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(53150,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(53150,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(53150,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(53150,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(53150,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(53150,36,'cp -40','c8d7 b5d7 b8d7 g1f3 e7e6 e1g1 d8c7 d2d4 h7h6 a2a4 a7a6 b1d2 a8c8 c2c3 f8d6 g2g3 g8e7 f3e1 e8g8 d2b3 d7b6 e1d3 b6c4 b3d2 f8e8 f1e1 e7g6 d1f3 c4a5 h2h4 g6e7 d2f1 a5b3 a1b1 c7c6 c1f4 d6f4 d3f4');
INSERT INTO evaluations VALUES(30936,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(30936,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(30936,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(30936,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(30936,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(30936,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(30936,36,'cp -44','b5d7 b8d7 g1f3 e7e6 e1g1 g8f6 f1e1 f8d6 d2d4 d8c7 c1g5 b7b5 b1d2 e8g8 c2c3 a7a5 g5f6 d7f6 d1e2 h7h6 g2g3 c7c6 e1c1 f8e8 a2a3 f6d7 c1e1 e8f8 g1g2 f8c8 h2h4 d7b6 a1c1 a8b8 f3e5 d6e5 e2e5');
INSERT INTO evaluations VALUES(30870,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(30870,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(30870,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(30870,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(30870,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(30870,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(30870,36,'cp -38','b8d7 d2d4 e7e6 g1f3 h7h6 e1g1 d8c7 d1e2 a8c8 g2g3 f8d6 b1c3 d6e7 c3b5 c7a5 a2a4 a7a6 c1d2 a5b6 b5c3 g8f6 f1c1 e8g8 c3d1 e7d6 d1e3 f6e4 c2c4 e4d2 e2d2 d6b4 d2e2 d5c4 c1c4 c8c4 e2c4');
INSERT INTO evaluations VALUES(26219,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(26219,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(26219,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(26219,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(26219,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(26219,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(26219,36,'cp -37','g1f3 e7e6 d2d4 f8d6 e1g1 g8e7 c1g5 e8g8 g5e7 d6e7 c2c3 b7b5 a2a3 d8c7 b1d2 a7a5 d1e2 c7c6 g2g3 e7d6 f1e1 h7h6 d2b3 f8c8 h2h4 d7b6 f3e5 d6e5 d4e5 c6c4 b3d4 b5b4 a3b4 a5b4 a1a8 b6a8 e1c1 a8b6 b2b3 c4e2 d4e2');
INSERT INTO evaluations VALUES(26216,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(26216,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(26216,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(26216,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(26216,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(26216,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(26216,36,'cp -41','g8f6 e1g1 e7e6 f1e1 d8c7 d2d4 f8d6 c2c3 e8g8 c1g5 b7b5 b1d2 a7a5 a2a3 h7h6 g5f6 d7f6 g2g3 c7c6 d1e2 f6d7 e1d1 a8b8 h2h4 f8e8 d1e1 e8c8 d2b3 c6c4 e2d1 c4c7 b3c1 b5b4 c3b4 a5b4 a3a4 b4b3 c1d3 c7c2 e1e2 c2d1 a1d1 d7b6 a4a5');
INSERT INTO evaluations VALUES(24586,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(24586,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(24586,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(24586,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(24586,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(24586,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(24586,36,'cp -31','c2c3 e7e6 d2d4 d8c7 c1g5 f6e4 b1d2 e4g5 f3g5 f8d6 a2a4 e8g8 g2g3 h7h6 g5f3 a7a5 d1b3 f8c8 e1g1 c7b6 b3b6 d7b6 h2h4 a8b8 f3e1 g7g5 e1f3 g5g4 f3e1 h6h5 e1d3 f7f6 f1c1 g8f7 d2b3 b6c4');
INSERT INTO evaluations VALUES(19951,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(19951,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(19951,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(19951,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(19951,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(19951,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(19951,36,'cp 37','d2d4 c5d4 f3d4 g8f6 b1c3 e7e5 d4e2 f8e7 e1g1 c8e6 c4b3 e8g8 c1g5 h7h6 g5f6 e7f6 e2g3 g7g6 d1d3 a8c8 a1d1 d8b6 g3e2 c6d4 e2d4 e5d4 b3e6 f7e6 c3b5 e6e5 d3b3 g8h7 c2c3 f6g7 d1d2 h6h5 c3d4 a7a6 b5c3 b6b3 a2b3 e5d4');
INSERT INTO evaluations VALUES(19962,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(19962,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(19962,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(19962,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(19962,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(19962,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(19962,36,'cp 25','c5d4 f3d4 g8f6 b1c3 e7e5 d4f3 c8e6 d1e2 a8c8 c4b3 h7h6 e1g1 f8e7 c3d5 e8g8 d5f6 e7f6 c1e3 c6a5 b3e6 f7e6 a1d1 d6d5 e3a7 b7b6 e4d5 e6d5 e2b5 a5c4 d1d5 d8c7 b2b3 c7a7 b3c4 a7a2 f3e5 f6e5 d5e5 a2c4');
INSERT INTO evaluations VALUES(20064,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(20064,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(20064,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(20064,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(20064,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(20064,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(20064,36,'cp 44','f3d4 g8f6 b1c3 e7e5 d4f3 c8e6 d1e2 a8c8 c4b3 h7h6 e1g1 f8e7 c3d5 e8g8 c2c3 c6a5 d5f6 e7f6 b3e6 f7e6 f1d1 d8c7 b2b3 f8d8 c1b2 d6d5 d1d2 a5c6 a1d1 c7f7 g2g3 a7a6 g1g2 d8d7 h2h3 c8d8 h3h4 b7b5 a2a4 c6a5');
INSERT INTO evaluations VALUES(20060,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(20060,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(20060,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(20060,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(20060,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(20060,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(20060,36,'cp 38','g8f6 b1c3 e7e6 c1e3 d6d5 e4d5 e6d5 c4b3 f8b4 e1g1 h7h6 c3e2 e8g8 h2h3 b4a5 e2f4 a5b6 c2c3 f8e8 a1c1 e8e4 f1e1 c6d4 c3d4 e4e8 d1f3 c8e6 a2a3 d8d6 f4e6 e8e6 e3f4 d6d7 e1e6 d7e6 c1d1 a8e8 f4e5');
INSERT INTO evaluations VALUES(13354,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(13354,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(13354,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(13354,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(13354,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(13354,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(13354,36,'cp 49','b1c3 e7e5 d4f5 c8f5 e4f5 d8a5 c1d2 a5c5 c4b3 d6d5 d2g5 e8c8 e1g1 f8e7 g5f6 e7f6 c3d5 c6d4 c2c4 c8b8 d1d3 c5c8 a1d1 c8f5 d3f5 d4f5 g2g3 f5d4 g1g2 d8d6 f1e1 h8d8 b3a4 d8c8 b2b3 d4c2 e1e4 c2d4 h2h4 d6d8 g2h1 a7a6 h4h5');
INSERT INTO evaluations VALUES(13352,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(13352,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(13352,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(13352,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(13352,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(13352,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(13352,36,'cp 27','d8b6 d4b3 e7e6 c1f4 c6e5 c4e2 c8d7 d1d2 b6c7 c3b5 d7b5 e2b5 e5d7 f2f3 a7a6 b5d7 f6d7 e1c1 d7e5 d2a5 a8c8 a5c7 c8c7 c2c3 g7g6 c1c2 h7h5 a2a4 e5c4 b3d2 c4e5 f4e5 d6e5 d2f1 f8h6 d1d6');
INSERT INTO evaluations VALUES(806,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(806,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(806,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(806,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(806,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(806,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(806,36,'cp 42','d4b3 e7e6 c1f4 c6e5 c4e2 c8d7 d1d2 a7a6 f4g5 b6c7 g5f6 g7f6 a2a4 e8c8 f2f4 e5c6 e1c1 h7h5 c1b1 c8b8 h1e1 c6a5 e2f3 a5b3 c2b3 c7c5 c3e2 d6d5 e4d5 e6d5 d2c3 d7f5 b1a1 c5c3 b2c3 h5h4 d1d5');
INSERT INTO evaluations VALUES(49645,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(49645,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(49645,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49645,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(49645,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(49645,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49645,36,'cp 7','f2f4 e5f4 f1e2 f8c5 c1f4 b8c6 b1c3 d7d5 e4e5 f6d7 g1f3 c6d4 c3d5 e8g8 b2b4 d4f3 e2f3 c5d4 c2c3 d4e5 e1g1 c7c6 d3d4 e5f4 d5f4 a7a5 a2a3 a5b4 a3b4 a8a1 d1a1 d7f6 a1a5 d8e8 a5e5 c8g4 e5e8 f8e8 f3g4 f6g4 d4d5 c6d5 f4d5 g8f8 f1a1 e8d8 c3c4 g4e5 d5b6 d8d4 c4c5');
INSERT INTO evaluations VALUES(49677,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49677,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(49677,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49677,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(49677,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(49677,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(49677,36,'cp -14','b8c6 f4e5 c6e5 g1f3 e5g6 f1e2 d7d5 e4e5 f6d7 d3d4 f8e7 e1g1 e8g8 b1c3 c7c6 e2d3 f7f6 e5e6 d7b6 f1e1 e7d6 d3g6 h7g6 f3h4 f8e8 d1g4 e8e6 e1e6 d8d7 c1f4 d6f4 a1e1 f4e5 d4e5 d7e6 g4e6 c8e6 e5f6 a8e8 f6g7 g8g7 h4f3 e6f7 g1f2 c6c5 c3b5 e8e1');
INSERT INTO evaluations VALUES(16596,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(16596,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(16596,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(16596,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(16596,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16596,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16596,36,'cp 0','f4e5 c6e5 g1f3 e5g6 f1e2 d7d5 e4e5 f6d7 d3d4 f8e7 e1g1 e8g8 e2d3 f7f6 e5e6 d7b6 d1e2 f8e8 d3g6 h7g6 f3h4 d8d6 h4g6 d6e6 e2h5 e7d6 c1g5 e6g4 h5h8 g8f7 f1f6 g7f6 h8f6 f7g8 f6h8 g8f7');
INSERT INTO evaluations VALUES(16357,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16357,36,'cp 0','c6e5 g1f3 e5g6 f1e2 d7d5 e4e5 f6d7 d3d4 f8e7 e1g1 e8g8 e2d3 f7f6 e5e6 d7b6 d1e2 f8e8 d3g6 h7g6 f3h4 d8d6 h4g6 d6e6 e2h5 e7d6 h5h8 g8f7 h8h5');
INSERT INTO evaluations VALUES(16963,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16963,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16963,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16963,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16963,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16963,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16963,36,'cp 0','g1f3 e5g6 f1e2 d7d5 e4e5 f6d7 d3d4 f8e7 e1g1 e8g8 e2d3 f7f6 e5e6 d7b6 d1e2 f8e8 d3g6 h7g6 f3h4 d8d6 h4g6 d6e6 e2h5 e7d6 h5h8 g8f7 h8h5');
INSERT INTO evaluations VALUES(17176,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(17176,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17176,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(17176,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(17176,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(17176,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(17176,36,'cp 60','b1c3 a7a6 b5c4 d7d6 d2d4 c6d4 f3d4 e5d4 d1d4 c8e6 c3d5 c7c6 d5e7 d8e7 c4e6 f7e6 c1e3 e7f7 e1c1 f8e7 d4b6 a8b8 f2f4 e8g8 c1b1 d6d5 h1e1 e7d8 b6b4 d8f6 g2g3 d5e4 b4e4 b8d8 e3b6 d8d5 c2c4 d5d1 e1d1 f8e8 a2a3 f7h5');
INSERT INTO evaluations VALUES(17172,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(17172,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(17172,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(17172,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(17172,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(17172,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(17172,36,'cp 55','a7a6 b5c4 d7d6 d2d4 e5d4 f3d4 c6d4 d1d4 c8e6 c4e6 f7e6 d4b4 d8c8 c3e2 e7c6 b4b3 f8e7 c1e3 c8d7 e2f4 c6d8 h2h4 e8g8 g2g3 f8f7 a1d1 d7b5 c2c3 e6e5 f4d5 d8e6 f2f4 e5f4 g3f4 c7c6 b3b5');
INSERT INTO evaluations VALUES(11584,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(11584,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(11584,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(11584,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(11584,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(11584,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(11584,36,'cp 47','b5c4 d7d6 d2d4 c6d4 f3d4 e5d4 d1d4 c8e6 c4e6 f7e6 d4b4 d8c8 b4b3 e7c6 e1g1 f8e7 c3e2 e7f6 c2c3 e8g8 e2f4 c6d8 c1e3 c8d7 f4d3 d7b5 f1d1 b5b3 a2b3 a6a5 f2f3 d8f7 d3f4 f8e8 g1f2 g7g5 f4h5 f6h8 h2h3 a8a6 f2e2 b7b5 e3d4');
INSERT INTO evaluations VALUES(11589,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11589,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(11589,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11589,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(11589,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(11589,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(11589,36,'cp 37','d7d6 d2d4 e5d4 f3d4 c6d4 d1d4 c8e6 c4e6 f7e6 d4b4 d8c8 b4b3 e7c6 c3e2 f8e7 e1g1 e8g8 e2f4 c6d8 c1e3 c8d7 a2a4 e7f6 h2h3 d7f7 a4a5 a8b8 f4d3 e6e5 b3f7 g8f7 d3b4 d8e6 b4d5 h7h6 c2c3 f6g5 e3g5 h6g5');
INSERT INTO evaluations VALUES(22518,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(22518,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(22518,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(22518,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(22518,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(22518,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(22518,36,'cp 52','d7d6 d2d4 f8e7 d4d5 c6b8 b1c3 g8f6 a2a4 e8g8 d1e2 a7a5 h2h3 b8a6 c1e3 c8d7 a1d1 f6h7 g1h2 g8h8 g2g4 a6b4 b2b3 h8g8 f1g1 a8c8 g1g3 h7g5 e2d2 g5f3 g3f3 e7g5 e3g5 d8g5 c4b5 d7b5 c3b5 g5d2 d1d2 b4a6 h2g2 a6c5 b5c3 c8e8 f3f5');
INSERT INTO evaluations VALUES(21287,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(21287,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(21287,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(21287,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(21287,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(21287,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(21287,36,'cp 79','d2d4 f8e7 d4d5 c6b8 c4d3 g8f6 c2c4 c8g4 h2h3 g4h5 c1e3 e8g8 b1c3 b8d7 g1h2 h5f3 d1f3 f6h7 e3d2 e7g5 d2e1 g7g6 g2g3 h6h5 h3h4 g5h6 f3e2 d7c5 d3c2 h7f6 f2f3 a7a5 h2g2 a5a4 e1f2 f6d7 c3b5');
INSERT INTO evaluations VALUES(21296,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(21296,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(21296,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(21296,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(21296,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(21296,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(21296,36,'cp 63','f8e7 d4d5 c6b8 c4d3 g8f6 c2c4 c8g4 h2h3 g4h5 c1e3 e8g8 b1c3 b8d7 g2g4 h5g6 d1e2 c7c6 f3h4 f6h7 h4f5 g6f5 e4f5 e7g5 f1d1 d7f6 d5c6 b7c6 d3e4 g5e3 e2e3 a8b8 e4c6 d8c7 c6f3 c7c4 d1d6 b8b2 e3e5 f8e8');
INSERT INTO evaluations VALUES(49405,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(49405,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(49405,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(49405,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(49405,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49405,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(49405,36,'cp 13','e2e3 d7d5 g1f3 c7c5 b1d2 d8b6 a1b1 f8d6 d4c5 b6c5 b2b4 c5c7 f4d6 c7d6 c2c4 e8g8 f1e2 b7b6 e1g1 b8d7 d1b3 c8b7 f1c1 d5c4 d2c4 d6e7 b3b2 f8c8 a2a3 h7h6 c4d2 f6d5 c1c8 a8c8 b1c1 c8c1 b2c1 e7d6 f3d4 d6c7 c1c7 d5c7');
INSERT INTO evaluations VALUES(49402,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(49402,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(49402,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(49402,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(49402,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(49402,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49402,36,'cp 0','c7c5 g1f3 d7d5 b1d2 f8e7 d4c5 e8g8 f1e2 e7c5 c2c4 b8d7 e1g1 b7b6 a1c1 c8b7 a2a3 a8c8 h2h3 c5e7 f4h2 d5c4 d2c4 a7a6 f3d4 b7d5 b2b4 h7h6 c4d6 c8c1 d1c1 d7b8 d6c4 b8d7 c4d6');
INSERT INTO evaluations VALUES(44222,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44222,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44222,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44222,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(44222,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44222,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44222,36,'cp 4','g1f3 d7d5 c2c3 f8d6 f1b5 b8c6 b5c6 b7c6 f4d6 d8d6 d1a4 c5d4 c3d4 e8g8 b1d2 c6c5 e1g1 c5d4 a4d4 c8d7 f1c1 f8b8 c1c2 a7a5 a1c1 d7e8 h2h3 a5a4 d2c4 d6a6 c4e5 f6e4 f3d2 e4d2 d4d2 h7h6 e5c6 e8c6 c2c6 a6b5 c1c2 b5b4 d2b4');
INSERT INTO evaluations VALUES(44223,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44223,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(44223,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44223,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(44223,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(44223,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44223,36,'cp 2','d7d5 c2c3 f8d6 f1b5 b8c6 b5c6 b7c6 f4d6 d8d6 d1a4 e8g8 b1d2 c5d4 c3d4 c6c5 a4a3 d6a6 a3a6 c8a6 d4c5 a8b8 d2b3 f8c8 a1c1 a6c4 f3d2 c4b5 b3d4 f6d7 d4b5 b8b5 d2b3 g8f8 c1c2 f8e7 e1d2 d7c5 c2c5 b5c5 b3c5 c8c5 h1c1 c5c1 d2c1 e6e5 b2b3');
INSERT INTO evaluations VALUES(50198,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(50198,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(50198,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(50198,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50198,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50198,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50198,36,'cp 11','c2c3 d7d6 b1d2 b8d7 g1f3 f6h5 f4g5 h7h6 g5h4 c7c5 e2e3 f8g7 f1d3 c5d4 c3d4 g6g5 h4g3 d7f6 e1g1 e8g8 e3e4 c8g4 a1c1 a8c8 c1c8 d8c8 d1b3 h5g3 h2g3 g4e6 b3a3 a7a6 a3a5 f6g4 a2a4 e6d7');
INSERT INTO evaluations VALUES(50194,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(50194,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(50194,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50194,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50194,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50194,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(50194,36,'cp 0','d7d6 b1d2 f8g7 e2e4 b8d7 g1f3 e7e5 d4e5 d6e5 f3e5 f6h5 e5d3 h5f4 d3f4 e8g8 g2g3 d7c5 d1f3 c7c6 f1e2 c5a4 a1b1 f8e8 f3e3 d8b6 e3f3');
INSERT INTO evaluations VALUES(30808,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(30808,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(30808,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(30808,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(30808,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(30808,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(30808,36,'cp 88','e7e6 a2a3 g8f6 f1d3 d7d6 g1f3 b8d7 c1f4 f8e7 h2h4 c7c5 e4e5 d6e5 d4e5 f6d5 c3d5 b7d5 d1e2 d8c7 a1d1 h7h6 e1g1 c5c4 d3e4 d5e4 e2e4 e8g8 f3d4 d7c5 e4e2 f8d8 e2g4 g8h7 d4b5 c7c6 b5d6 e7d6 e5d6 f7f5 g4h5 h7g8');
INSERT INTO evaluations VALUES(30670,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(30670,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(30670,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(30670,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(30670,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(30670,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(30670,36,'cp 61','a2a3 d7d5 e4d5 e6d5 d1f3 b8c6 c1e3 d8d7 f1d3 g7g6 g1e2 e8c8 h2h4 f8g7 h4h5 g8h6 e1c1 d7g4 f3g4 h6g4 d1f1 d8e8 h5h6 g7f6 c3d5 c6d4 d5f6 d4e2 d3e2 g4f6 f1e1 b7g2 h1g1 g2b7 e3d4 e8e6 e2g4 f6g4 e1e6 f7e6 d4h8 g4h6 g1g3 h6f5');
INSERT INTO evaluations VALUES(30684,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(30684,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(30684,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(30684,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(30684,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(30684,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(30684,36,'cp 62','d7d5 e4d5 e6d5 h2h4 b8c6 g1e2 g8f6 d1d3 d8e7 g2g3 e8c8 c1f4 c8b8 e1c1 a7a6 f1g2 h7h6 c1b1 c6a5 f4c1 a5c4 d1e1 e7d7 e2f4 c7c6 e1e2 b7c8 g2h3 d7c7 h3c8 c7c8 h1f1 f8d6 f1e1 h8e8 f2f3 e8e2');
INSERT INTO evaluations VALUES(57076,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(57076,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(57076,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(57076,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(57076,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(57076,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(57076,36,'cp 30','g2g3 d7d5 f1g2 c7c5 c2c4 d5d4 e2e3 b8c6 d2d3 g8f6 e1g1 f8d6 e3d4 c5d4 c1g5 e8g8 b1d2 h7h6 g5f6 d8f6 d2e4 f6d8 e4d6 d8d6 a1c1 a7a5 c4c5 d6c7 f3d2 c8d7 d2c4 c6e7 d1g4 c7c5 g2b7 a8a7 b7f3 a7c7 g4f4 f8c8 g1h1');
INSERT INTO evaluations VALUES(57077,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(57077,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(57077,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(57077,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(57077,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(57077,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(57077,36,'cp 25','d7d5 f1g2 g8f6 d2d4 b7b5 a2a4 b5b4 e1g1 f8e7 f3e5 e8g8 c2c4 b4c3 b1c3 a7a5 f1e1 f6d7 e5d3 c8b7 e2e4 d5e4 c3e4 b8c6 e4c3 a8b8 d4d5 c6b4 d5e6 f7e6 g2b7 b8b7 c3e4 b4d3 d1d3 d7c5 d3d8 f8d8');
INSERT INTO evaluations VALUES(55951,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55951,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55951,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(55951,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(55951,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55951,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(55951,36,'cp 31','f1g2 c7c5 e1g1 g8f6 c2c4 d5d4 d2d3 b8c6 e2e3 f8d6 e3d4 c5d4 c1g5 h7h6 g5f6 d8f6 b1d2 f6e7 d2e4 e8g8 e4d6 e7d6 f1e1 a7a5 d1e2 c8d7 f3e5 d7e8 a1b1 a5a4 f2f4 a8a6 a2a3 a6a5 e5f3 d6c7 e2f2 c7d6 b2b4 a4b3 b1b3 a5a3');
INSERT INTO evaluations VALUES(55952,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(55952,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(55952,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(55952,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(55952,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(55952,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(55952,36,'cp 32','c7c5 e1g1 g8f6 c2c4 d5d4 d2d3 b8c6 e2e3 f8d6 e3d4 c5d4 c1g5 e8g8 b1d2 h7h6 g5f6 d8f6 d2e4 f6e7 e4d6 e7d6 f1e1 c8d7 d1e2 a7a5 f3e5 d7e8 a1b1 a5a4 f2f4 d6c7 b2b4 a4b3 b1b3 a8a7 e5f3 e8d7');
INSERT INTO evaluations VALUES(54273,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(54273,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(54273,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54273,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(54273,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(54273,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(54273,36,'cp 35','c2c4 d5d4 e2e3 b8c6 d2d3 f8d6 e1g1 g8f6 e3d4 c5d4 c1g5 h7h6 g5f6 d8f6 b1d2 e8g8 d2e4 f6e7 e4d6 e7d6 a1c1 c8d7 c4c5 d6c7 f1e1 a7a5 b2b3 a8b8 d1e2 f8d8 a2a3 d7e8 c1c4 b7b6 f3d2 c6e7 e2h5 b8c8 d2e4 b6c5 e4c5');
INSERT INTO evaluations VALUES(54168,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(54168,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54168,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54168,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(54168,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54168,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(54168,36,'cp 27','d5d4 e2e3 b8c6 e1g1 g8f6 d2d3 f8d6 e3d4 c5d4 c1g5 e8g8 b1d2 h7h6 g5f6 d8f6 d2e4 f6e7 e4d6 e7d6 f1e1 a7a5 d1e2 a5a4 f3d2 a8a5 b2b3 a4a3 g2c6 b7c6 d2e4 d6b4 e2d2 b4d2 e4d2 c6c5 f2f4 f7f6 h2h4 c8d7');
INSERT INTO evaluations VALUES(49521,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49521,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(49521,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(49521,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49521,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49521,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(49521,36,'cp 0','h5e5 f8e7 e5f4 b8c6 b1c3 d7d5 e4e5 f6g4 g1f3 c6b4 f1b5 c7c6 b5a4 f7f6 d2d4 e8g8 f4g3 f6e5 d4e5 h7h5 h2h4 c8f5 e1g1 b4c2 a4c2 f5c2 c1g5 c2f5 f1e1 d8c7 g5e7 c7e7 e1e2 e7c7 a1e1 a8e8 f3d4 c7b6 e1d1 b6c7');
INSERT INTO evaluations VALUES(49511,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49511,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(49511,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49511,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(49511,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49511,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(49511,36,'cp -12','f8e7 e5f4 b8c6 e4e5 f6h5 f4f3 g7g6 b1c3 e8g8 f3d1 d7d6 e5d6 e7d6 d2d3 c6d4 f1e2 h5f4 c1f4 d6f4 g1f3 d4e2 c3e2 f4h6 e1g1 b7b6 d3d4 c8b7 d1d3 d8f6 a1d1 h6g7 c2c4 a8e8 e2c3 b7f3 d3f3 f6f3 g2f3 e8d8 c3d5 f8e8 d5c7');
INSERT INTO evaluations VALUES(43021,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(43021,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(43021,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(43021,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43021,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43021,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(43021,36,'cp 0','e5f4 b8c6 e4e5 f6h5 f4f3 g7g6 b1c3 e8g8 f3d1 d7d6 e5d6 e7d6 d2d3 c6d4 f1e2 d4e2 g1e2 b7b6 c1h6 f8e8 d1d2 c8b7 e1g1 d8h4 h2h3 f7f6 c3b5 g6g5 b5d6 c7d6 f2f3 e8e2 d2e2 h5f4 e2e7 f4h3 g1h2 h3f4 h2g1 f4h3');
INSERT INTO evaluations VALUES(43048,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(43048,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(43048,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(43048,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(43048,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(43048,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43048,36,'cp 0','b8c6 g1f3 e8g8 d2d4 d7d5 e4e5 f6h5 f4d2 f7f6 e5f6 f8f6 d2d1 c8g4 f1e2 d8d6 b1c3 a8f8 e1g1 h5f4 c1f4 f6f4 d1d3 c6b4 d3d1');
INSERT INTO evaluations VALUES(57002,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(57002,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(57002,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(57002,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(57002,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(57002,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(57002,36,'cp -35','d7d5 e4d5 e6d5 c4e2 g8f6 g1f3 f8d6 d2d4 e8g8 c1g5 f8e8 e1g1 c8g4 b1d2 b8d7 c2c3 g4f5 f1e1 c7c6 f3h4 f5e6 h4f3 d8c7 g5h4 f6h5 d2f1 e6f5 f1e3 f5e4 h4g3 h5g3 h2g3 a7a5 e2d3 d7f6 d3e4');
INSERT INTO evaluations VALUES(55794,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(55794,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(55794,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(55794,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(55794,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(55794,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(55794,36,'cp -33','e4d5 e6d5 c4e2 g8f6 g1f3 h7h6 d2d4 f8d6 c2c4 d5c4 b1d2 c7c6 d2c4 d6c7 e1g1 e8g8 c4e5 c8e6 d1c2 b8d7 c1f4 f8e8 f1e1 c7a5 e1d1 a5b6 f4g3 e6d5 e2c4 f6e4 d1e1 d7e5 g3e5 d5c4 c2c4 e4d6 c4b3');
INSERT INTO evaluations VALUES(55766,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(55766,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(55766,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(55766,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(55766,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(55766,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(55766,36,'cp -33','e6d5 c4e2 f8d6 g1f3 g8f6 d2d4 h7h6 c2c4 d5c4 b1d2 c7c6 d2c4 d6c7 e1g1 c8e6 f1e1 e8g8 e2d3 f8e8 c1d2 b8d7 a1c1 d7b6 b2b3 e6d5 c4e5 b6d7 d2f4 d7f8 d3c4 c7a5 f4d2 f8e6 a2a3 a5d2 d1d2');
INSERT INTO evaluations VALUES(56177,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(56177,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(56177,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(56177,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(56177,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(56177,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(56177,36,'cp -34','c4e2 h7h6 d2d4 g8f6 g1f3 f8d6 c2c4 d5c4 b1d2 c7c6 d2c4 d6c7 e1g1 e8g8 a2a4 c8e6 d1b3 e6d5 b3b7 d8e7 b7a8 c6c5 a8a7 c7h2 f3h2 e7a7 d4c5 a7c5 b2b4 c5e7 c4e3 b8c6 f1e1 c6d4 c1a3 d4e2 e1e2 d5b7 b4b5 e7e5 a3b2 e5h5');
INSERT INTO evaluations VALUES(56296,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(56296,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(56296,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(56296,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(56296,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(56296,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(56296,36,'cp -43','h7h6 d2d4 f8d6 g1f3 g8f6 c2c4 d5c4 b1d2 e8g8 d2c4 c8e6 e1g1 c7c6 c4d6 d8d6 h2h3 f8e8 e2d3 b8d7 f1e1 e6d5 e1e8 a8e8 c1e3 a7a6 a2a3 d7f8 a1c1 d5e4 d3e4 f6e4 d1b3 e8e7 c1e1 f8g6 e3c1 d6f6 b3c2 f6f5 e1e2 e4d6 c2f5 d6f5 e2e7 f5e7');
INSERT INTO evaluations VALUES(21958,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(21958,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(21958,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(21958,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(21958,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(21958,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(21958,36,'cp 87','d2d4 g8f6 f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 f8e8 c1e3 d8b8 a1b1 a7a5 e4e5 f6d5 e3d4 d5c3 d4c3 b8b5 b1d1 b5a4 a2a3 a8d8 g1h1 e7f8 d1d2 d6e5 f4e5 a4c4 c3a5 d8d2 a5d2');
INSERT INTO evaluations VALUES(21944,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(21944,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(21944,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(21944,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(21944,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(21944,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(21944,36,'cp 79','f8e7 f1b5 e5d4 f3d4 c8d7 d4c6 b7c6 b5d3 e7f6 e1g1 g8e7 c1e3 f6c3 b2c3 e8g8 f2f4 f7f5 d3c4 g8h8 e4e5 e7d5 c4d5 c6d5 e5d6 d7b5 d6c7 d8c7 f1f3 a8e8 a2a4 b5c4 f3g3 f8f7 d1h5 e8e6 e3d4 c4e2 h5h4 h7h6 h2h3 h8g8 g1h2 c7d7');
INSERT INTO evaluations VALUES(10118,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(10118,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10118,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(10118,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10118,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(10118,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10118,36,'cp 2','d7d6 c2c3 a7a5 c4b3 h7h6 f1e1 e8g8 h2h3 c8e6 b3a4 c5a7 d3d4 e6d7 a4c2 f8e8 a2a4 e5d4 c3d4 c6b4 b1c3 b4c2 d1c2 c7c6 c1f4 d7e6 a1d1 d8d7 e4e5 f6d5 c3d5 e6d5 e5d6 d5f3 g2f3 d7h3 c2c3 e8e1 d1e1 h3f5 f4g3 f5d5');
INSERT INTO evaluations VALUES(9352,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(9352,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(9352,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9352,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(9352,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9352,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(9352,36,'cp 19','c2c3 a7a5 f1e1 e8g8 h2h3 h7h6 c4b3 f8e8 b1d2 c8e6 b3a4 f6d7 d2f1 d6d5 c1e3 a8b8 e4d5 e6d5 a4b3 d5b3 d1b3 c5f8 a1d1 d7f6 f1g3 d8d7 g3e4 f8e7 e4f6 e7f6 d3d4 e5d4 c3d4 c6b4 d4d5 d7f5 e1e2 b8a8 a2a3');
INSERT INTO evaluations VALUES(9345,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(9345,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9345,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9345,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9345,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9345,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(9345,36,'cp 0','a7a5 f1e1 e8g8 h2h3 h7h6 c4b5 c5a7 a2a4 c6e7 d3d4 e7g6 b1d2 c7c6 b5d3 f8e8 d2f1 d6d5 f3e5 g6e5 d4e5 e8e5 c1f4 f6e4 d3e4 d8f6 e4h7 g8h7 f4e5 f6f2 g1h2 f2g1 h2g3 g1f2');
INSERT INTO evaluations VALUES(54961,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(54961,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(54961,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(54961,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(54961,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54961,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54961,36,'cp 21','e7e5 e2e3 g8f6 b1c3 b8d7 f1e2 g7g6 g1f3 f8g7 e1g1 e8g8 d4e5 d6e5 e3e4 c7c6 b2b4 d8e7 a2a3 f6h5 g2g3 f8d8 d1b3 d7f8 c1g5 h5f6 f1d1 c8h3 g5e3 f6g4 d1d8 a8d8 e3c5 e7f6');
INSERT INTO evaluations VALUES(55569,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(55569,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(55569,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(55569,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(55569,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(55569,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(55569,36,'cp 39','d4d5 f7f5 e2e4 f5e4 g1e2 f8e7 e2g3 g8f6 b1c3 e8g8 c3e4 f6e4 g3e4 c8f5 f1d3 b8d7 d1c2 g8h8 e1g1 d8e8 c1e3 e8g6 a1d1 d7f6 e4g3 f5d3 c2d3 g6d3 d1d3 a7a6 f2f3 h8g8 d3c3 a8e8 g3f5 g7g6 f5e7 e8e7 c3b3 b7b6 e3g5');
INSERT INTO evaluations VALUES(55498,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(55498,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(55498,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55498,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(55498,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(55498,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(55498,36,'cp 24','f8e7 b1c3 e7g5 e2e3 g5h6 f1d3 d8e7 e3e4 h6c1 a1c1 a7a5 g1e2 b8d7 d3c2 g8f6 d1d2 d7c5 e1g1 e8g8 b2b3 f6h5 f2f3 g7g6 g2g4 h5f6 d2h6 g8h8 c1e1 c8d7 h2h3 f6g8 h6d2 e7h4 g1g2 h8g7');
INSERT INTO evaluations VALUES(31162,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(31162,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(31162,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31162,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(31162,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(31162,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31162,36,'cp 34','e7e6 f1e2 a7a6 e1g1 c6c5 c1e3 c5d4 f3d4 g8e7 c2c4 b8c6 c4d5 e7d5 d4c6 b7c6 d1a4 d8d7 b1a3 f8e7 a3c4 e8g8 f1d1 d7b7 e3d2 f8d8 a4b3 b7b3 a2b3 d5b4 d2c3 d8d1 a1d1 b4d5 c4a5 a8c8 d1c1 d5c3 c1c3 e7b4 c3c6 c8c6 a5c6');
INSERT INTO evaluations VALUES(31535,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(31535,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(31535,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(31535,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(31535,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(31535,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31535,36,'cp 29','f1e2 a7a6 c1e3 d8b6 b2b3 c6c5 c2c4 b8c6 c4d5 e6d5 e1g1 c5d4 f3d4 c6d4 e3d4 f8c5 d4c5 b6c5 b1d2 g8e7 d2f3 e8g8 d1d2 c5b6 f3d4 f5e4 d2e3 f8e8 a1c1 h7h6 f1d1 a8d8 a2a3 e7g6');
INSERT INTO evaluations VALUES(31534,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(31534,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(31534,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(31534,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(31534,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(31534,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(31534,36,'cp 22','c6c5 e1g1 b8c6 c2c3 a7a6 a2a3 c5d4 c3d4 g8e7 c1e3 f5g4 b1d2 e7f5 h2h3 f5e3 f2e3 g4f5 f3e1 f5g6 e2d3 f8e7 d3g6 h7g6 e1d3 e8g8 d2b3 a8c8 a1c1 b7b6 d1g4 c6a5 b3a5 c8c1 f1c1 b6a5 g4g3 d8b6');
INSERT INTO evaluations VALUES(31674,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31674,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31674,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(31674,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31674,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(31674,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(31674,36,'cp 28','e1g1 b8c6 c2c3 c5d4 c3d4 g8e7 b1c3 e7c8 e2d3 f5d3 d1d3 f8e7 g2g3 c6b4 d3d2 e8g8 g1g2 c8b6 b2b3 b4c6 d2d3 c6b4 d3e2 a8c8 c1d2 a7a6 h2h4 h7h6 a2a4 b4c6 e2e3 b6d7 c3e2 e7b4 e2f4 d8b6 d2b4');
INSERT INTO evaluations VALUES(31673,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(31673,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31673,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31673,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(31673,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31673,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31673,36,'cp 17','b8c6 c2c3 a7a6 a2a3 c5d4 c3d4 g8e7 c1e3 f5g4 b1d2 e7f5 h2h3 f5e3 f2e3 g4f5 b2b4 a6a5 b4b5 c6b8 f3e1 f8e7 a3a4 b8d7 e2d3 f5d3 e1d3 e8g8 d1b3 e7g5 f1f3 a8c8 a1f1 g5e7 b3d1 f7f6 d3f4 d8b6 e5f6 d7f6 d1b3');
INSERT INTO evaluations VALUES(26414,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(26414,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(26414,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26414,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(26414,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26414,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26414,36,'cp 18','c2c3 a7a6 c1e3 d8b6 f3h4 f5e4 b1d2 f8e7 d4c5 e7c5 e3c5 b6c5 d1a4 h7h6 a1c1 g8e7 d2e4 d5e4 a4e4 c5e5 e4c2 e5c7 h4f3 e8g8 f1d1 f8d8 g2g3 a8c8 d1d8 c8d8 c1d1 d8d1 c2d1 e7d5 a2a3 d5f6 b2b4 a6a5 d1d2 b7b6 c3c4 f6d7 h2h4 d7e5 f3e5 c6e5');
INSERT INTO evaluations VALUES(26411,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(26411,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26411,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(26411,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(26411,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26411,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26411,36,'cp 24','a7a6 c1e3 d8b6 f3h4 f5e4 b1d2 f8e7 d4c5 e7c5 e3c5 b6c5 d1a4 g8e7 d2e4 d5e4 a4e4 c5e5 e4c2 e5c7 g2g3 e8g8 h4f3 h7h6 a1d1 a8c8 f1e1 e7d5 d1d2 f8d8 e1d1 d5f6 d2d8 c8d8 a2a3 d8d1 c2d1 g7g6 d1d2 g8g7 d2e3 c6a5 b2b4 a5c4 e2c4');
INSERT INTO evaluations VALUES(31196,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(31196,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(31196,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(31196,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(31196,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(31196,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(31196,36,'cp 34','e4g3 f5g6 g1f3 b8d7 h2h4 h7h6 f1d3 g6d3 d1d3 e7e6 c1f4 g8f6 e1c1 f8e7 c1b1 e8g8 f3e5 d7e5 f4e5 f6g4 d3e2 g4e5 d4e5 d8b6 d1d3 a8d8 d3b3 b6a5 b3b7 d8b8 b7b8 f8b8 b2b3 b8d8 f2f4 d8d2 e2f3 a5d5 g3e4');
INSERT INTO evaluations VALUES(31191,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(31191,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(31191,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31191,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31191,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31191,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(31191,36,'cp 32','f5g6 h2h4 h7h6 g1f3 b8d7 c1f4 e7e6 f1d3 g6d3 d1d3 d8a5 c2c3 g8f6 e1g1 f8e7 f1e1 e8g8 a2a4 f8e8 f3e5 a8d8 b2b4 a5c7 d3f3 f6d5 g3h5 d5f4 f3f4 d7f6 a4a5 e8f8 h5f6 e7f6 g2g3 b7b6 e1e3 f6e5 f4e5 c7e5 e3e5 d8d7 h4h5 f8d8 a5b6 a7b6 g1f1');
INSERT INTO evaluations VALUES(31044,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(31044,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(31044,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31044,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(31044,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31044,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(31044,36,'cp 35','g1f3 b8d7 h2h4 h7h6 c1f4 e7e6 f1d3 g6d3 d1d3 d8a5 c2c3 g8f6 e1g1 f8e7 f1e1 e8g8 a2a4 f8e8 f3e5 a8d8 b2b4 a5c7 d3f3 f6d5 g3h5 d5f4 f3f4 d7f6 a4a5 e8f8 h5f6 e7f6 g2g3 b7b6 e1e3 f6e5 f4e5 c7e5 e3e5 d8d7 h4h5 f8d8');
INSERT INTO evaluations VALUES(31042,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(31042,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(31042,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31042,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(31042,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(31042,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(31042,36,'cp 34','b8d7 h2h4 h7h6 f1d3 g6d3 d1d3 e7e6 c1f4 d8a5 c2c3 f8e7 e1g1 g8f6 f1e1 e8g8 a2a4 f8e8 f3e5 a8d8 b2b4 a5c7 d3f3 f6d5 g3h5 d5f4 f3f4 d7f6 a4a5 e8f8 h5f6 e7f6 e1e4 f6e5 f4e5 c7d6 e5d6 d8d6 g1f1 a7a6 f1e2 d6d5 a1d1 d5d8 g2g3 d8d5');
INSERT INTO evaluations VALUES(26190,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(26190,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(26190,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(26190,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(26190,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(26190,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26190,36,'cp 26','h2h4 h7h6 c1f4 e7e6 f1d3 g6d3 d1d3 d8a5 c2c3 g8f6 e1g1 f8e7 f1e1 e8g8 a2a4 f6d5 d3d2 d5f4 d2f4 d7f6 g3f5 a8e8 f5e7 e8e7 f3e5 f8d8 b2b4 a5c7 a4a5 f6d7 e5d3 c7f4 d3f4 e6e5 a5a6 b7b6 f4d3 d8e8 d4e5 d7e5 d3e5 e7e5 e1e5 e8e5 g1f1 e5e4 h4h5');
INSERT INTO evaluations VALUES(26187,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26187,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(26187,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(26187,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(26187,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26187,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(26187,36,'cp 40','h7h6 f1d3 g6d3 d1d3 e7e6 c1f4 d8a5 f4d2 a5c7 e1c1 g8f6 g3e4 f6e4 d3e4 d7f6 e4e2 a8d8 h1g1 f6g4 c1b1 f8e7 c2c3 c6c5 d4c5 e7c5 f3d4 g4f6 d4b3 c5e7 g2g4 h6h5 g4h5 h8h5 g1g7 h5h4 d1g1 c7c4 e2c4 h4c4 b3d4');
INSERT INTO evaluations VALUES(26178,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(26178,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(26178,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(26178,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(26178,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(26178,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26178,36,'cp 30','f1d3 g6d3 d1d3 e7e6 c1f4 d8a5 c2c3 g8f6 e1g1 f8e7 f1e1 e8g8 a2a4 c6c5 g3f5 f8e8 f5e7 e8e7 f4g3 e7e8 d4c5 a5c5 a1d1 c5a5 f3e5 d7e5 g3e5 e8d8 d3b5 a5b6 d1d8 a8d8 b5b6 a7b6 e5d4 f6d7 g2g3 d7c5');
INSERT INTO evaluations VALUES(26177,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(26177,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(26177,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26177,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(26177,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(26177,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(26177,36,'cp 28','g6d3 d1d3 e7e6 c1d2 g8f6 e1c1 f8e7 g3e4 e8g8 h1g1 f6e4 d3e4 d7f6 e4e2 d8d5 c2c4 d5e4 g1e1 e4e2 e1e2 b7b5 b2b3 f8e8 d1e1 b5c4 b3c4 a8b8 c1c2 b8b7 d2c3 e8c8 a2a4 e7b4 f3e5 b4c3 c2c3');
INSERT INTO evaluations VALUES(26182,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26182,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(26182,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(26182,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26182,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(26182,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(26182,36,'cp 29','d1d3 e7e6 c1f4 d8a5 c2c3 g8f6 e1g1 f8e7 f1e1 e8g8 a2a4 f6d5 d3d2 d5f4 d2f4 a5d8 g3h5 d7f6 h5f6 e7f6 a4a5 a8c8 b2b4 d8e7 g2g3 b7b6 a5b6 a7b6 f4e3 f8d8 f3d2 c6c5 d4c5 b6c5 b4b5 c5c4 a1a7 d8d3 a7e7');
INSERT INTO evaluations VALUES(26181,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26181,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(26181,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(26181,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(26181,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(26181,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(26181,36,'cp 25','e7e6 c1f4 d8a5 f4d2 f8b4 a2a3 b4d2 f3d2 g8f6 e1c1 e8g8 g3e4 a5d5 c2c4 d5f5 f2f3 f6e4 d3e4 f5e4 d2e4 f8d8 d1d2 g8f8 h1d1 f8e7 a3a4 b7b6 c1c2 a7a5 g2g4 d7f6 e4f6 e7f6 b2b3 d8d7 c4c5 b6c5 d4c5 d7d2 d1d2');
INSERT INTO evaluations VALUES(26087,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(26087,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(26087,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26087,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(26087,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(26087,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(26087,36,'cp 31','c1f4 d8a5 f4d2 f8b4 g3e4 g8f6 e4d6 e8e7 d6b7 b4d2 f3d2 a5b4 d3b3 b4d4 e1c1 a8b8 b3a3 c6c5 b7a5 h8c8 d2c4 d4f4 a3e3 f4e3 c4e3 b8b4 e3c4 d7b6 b2b3 f6e4 h1f1 e4c3 d1e1 b6c4 a5c4 b4b8 a2a3 c3d5 e1e4 d5f6 e4e2 f6d5 f2f4 d5b6 e2e4 f7f5');
INSERT INTO evaluations VALUES(55892,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55892,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(55892,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(55892,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(55892,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55892,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(55892,36,'cp 25','d2d4 g8f6 e4e5 f6d7 f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 c5d4 e3d4 c6d4 d2d4 d8b6 d4b6 d7b6 f1d3 c8d7 e1d2 b6a4 c3a4 d7a4 h1c1 a8c8 g2g3 e8e7 c2c4 d5c4 c1c4 c8c4 d3c4 h7h5 d2e3 h5h4 a1c1 h4g3 h2g3 a4c6 c4d3 h8h2');
INSERT INTO evaluations VALUES(16522,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(16522,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(16522,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(16522,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(16522,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(16522,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(16522,36,'cp -32','d7d5 e4d5 f6d5 g2g3 f8e7 c1d2 e8g8 f1g2 c8e6 e1g1 f7f6 f1e1 d8d7 d1e2 d5c3 d2c3 e6d5 a2a3 f8e8 a1d1 a8d8 b2b4 e7f8 f3d2 d5g2 g1g2 b7b6 e2e4 c6e7 c3b2 e7f5 d2c4 f8e7 b2c3 c7c5 c4e3 f5e3 e1e3');
INSERT INTO evaluations VALUES(15137,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(15137,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(15137,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(15137,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(15137,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(15137,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(15137,36,'cp -16','e4d5 f6d5 g2g3 f8e7 c1d2 c8e6 f1g2 e8g8 e1g1 f7f6 c3d5 e6d5 c2c3 a7a5 f1e1 a5a4 d1c2 d8d7 d3d4 e5d4 f3d4 d5g2 g1g2 c6d4 c3d4 d7d5 c2e4 d5e4 e1e4 e7d6 d4d5 f8e8 a1e1 e8e4 e1e4 b7b5 e4e1 d6e5 e1c1 e5b2 c1c7');
INSERT INTO evaluations VALUES(15006,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(15006,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(15006,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(15006,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(15006,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(15006,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(15006,36,'cp -23','f6d5 g2g3 f8e7 c1d2 e8g8 f1g2 c8e6 e1g1 f7f6 f1e1 d8d7 c3d5 e6d5 c2c3 a8d8 b2b4 a7a6 a2a4 b7b5 a4b5 a6b5 d1b1 e7d6 d3d4 d7f7 d4e5 c6e5 b1f5 e5f3 g2f3 d5f3 f5f3 d6b4 c3b4 d8d2 a1c1 f7d7 f3b3 f8f7 e1d1');
INSERT INTO evaluations VALUES(15349,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(15349,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(15349,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(15349,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(15349,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(15349,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(15349,36,'cp -16','g2g3 f8e7 c1d2 e8g8 f1g2 c8e6 e1g1 d8d7 f1e1 f7f6 c3d5 e6d5 c2c3 a8d8 b2b4 a7a6 a2a4 d5f3 d1f3 d7d3 e1d1 d3f3 g2f3 f6f5 f3e2 d8d6 b4b5 c6a5 b5a6 b7a6 d2e3 d6c6 d1d5 a5c4 e2c4 c6c4 d5e5 e7d6 e5a5 c4c6 a1b1 f5f4 g3f4 d6f4 e3f4 f8f4 b1b8 g8f7');
INSERT INTO evaluations VALUES(49832,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(49832,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(49832,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(49832,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(49832,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(49832,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(49832,36,'cp 86','d1e2 d8e7 e2e4 d7d6 d2d4 d6e5 d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5g6 h2h4 a7a6 b5c3 h8e8 f1c4 e8e7 h4h5 g6f4 g2g3 c8e6 h5h6 g7g6 c4e6 f4e6 c3e4 e7d7 d2c1 d8e7 e4g5 a8h8 h1h4 e7f6 g5e6 f7e6 h4e4 f6g5 e4e6 g5h6');
INSERT INTO evaluations VALUES(49833,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(49833,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(49833,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(49833,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(49833,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(49833,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(49833,36,'cp 82','d8e7 e2e4 d7d6 d2d4 d6e5 d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5g6 h2h4 c8d7 h4h5 g6f4 h1h4 f4e6 b5d4 h8e8 h5h6 e6d4 e1e8 d8e8 h6g7 e8e7 h4d4 a8g8 f1d3 g8g7 g2g3 h7h5 d4d5 d7g4 d5a5 a7a6 d2e3 g7g8 a2a4 c7c6 b2b4 g8e8 a5c5 b7b6 c5c6');
INSERT INTO evaluations VALUES(34070,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(34070,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(34070,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(34070,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(34070,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(34070,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(34070,36,'cp 67','e2e4 d7d6 d2d4 d6e5 d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5d7 f1c4 h8f8 d2c1 d7b6 e1e4 c8d7 h1d1 a7a6 b5c3 c7c6 f2f3 b6c4 e4c4 d8e7 c3a4 d7e6 c4h4 h7h6 a4c5 b7b6 c5e6 f7e6 h4b4 b6b5 b4g4 f8g8 g4d4 a8d8 d4d8 g8d8 d1d8 e7d8 c1d2 d8e7 d2e3 e6e5 g2g3 e7e6 e3e4 a6a5');
INSERT INTO evaluations VALUES(34068,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34068,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(34068,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(34068,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(34068,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(34068,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(34068,36,'cp 66','d7d6 d2d4 d6e5 d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5d7 f1c4 h8f8 b5d4 d7b6 c4d3 c8d7 e1e5 f8e8 h1e1 e8e5 e1e5 g7g6 b2b3 a7a5 f2f4 c7c6 c2c4 d7e8 d4f3 b6d7 e5e1 h7h6 d2c3 d8c7 g2g4 d7c5 d3c2 e8d7 f4f5 a8f8 e1e7');
INSERT INTO evaluations VALUES(33806,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(33806,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(33806,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(33806,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(33806,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(33806,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(33806,36,'cp 56','d2d4 d6e5 d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5d7 f1c4 h8f8 b5d4 d7b6 c4d3 c8d7 e1e5 f8e8 h1e1 e8e5 e1e5 g7g6 b2b3 a7a5 a2a3 c7c6 c2c4 d7e8 f2f4 b6d7 e5e3 d8c7 d3c2 a8d8 d2c3 d7f6 e3e7 d8d7 e7e1 b7b6 f4f5 c7d6 d4f3');
INSERT INTO evaluations VALUES(33804,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(33804,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(33804,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(33804,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(33804,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(33804,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(33804,36,'cp 59','d6e5 d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5g6 h2h4 c8d7 h4h5 g6f4 h1h4 f4e6 b5d4 h8e8 h5h6 g7g6 d4f3 e8e7 f1c4 d8e8 f3e5 e8f8 h4e4 a8e8 e5d7 e7d7 d2c3 d7d6 b2b4 f8e7 a2a4 e8d8 a4a5 a7a6 e4e3 e7f6 e3f3 f6e7 e1e4');
INSERT INTO evaluations VALUES(33852,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(33852,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(33852,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(33852,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(33852,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(33852,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(33852,36,'cp 74','d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5d7 f1c4 h8f8 b5d4 d7b6 c4d3 c8d7 e1e5 f8e8 h1e1 g7g6 b2b4 e8e5 e1e5 c7c6 c2c4 f7f6 e5e4 a7a5 b4b5 f6f5 e4e5 c6b5 c4c5 b6c4 d3c4 b5c4 d2c3 d7c6 f2f3 d8d7 c3c4 a8b8 e5e6 b8d8 d4c6 b7c6');
INSERT INTO evaluations VALUES(33849,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(33849,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(33849,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(33849,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(33849,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(33849,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(33849,36,'cp 53','b8c6 b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5g6 h2h4 c8d7 h4h5 g6f4 h1h4 f4e6 b5d4 h8e8 d4f3 d8e7 h5h6 g7g6 f1c4 a8d8 f3g5 f7f6 g5e6 d7e6 d2c3 e7f7 c4e6 e8e6 e1e6 f7e6 h4e4 e6f7 e4d4 d8d4 c3d4 f7e7 a2a4 b7b6 b2b4 e7d6 f2f3 c7c6 a4a5 g6g5 a5b6 a7b6 c2c4');
INSERT INTO evaluations VALUES(908,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(908,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(908,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(908,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(908,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(908,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(908,36,'cp 67','b1c3 e7e5 e4e5 c6e5 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 a1e1 e5g6 h2h4 c8d7 h4h5 g6f4 h1h4 f4e6 b5d4 h8e8 d4f3 d8e7 h5h6 g7g6 f1c4 a8d8 d2c3 b7b5 c4d3 c7c5 f3g5 c5c4 d3f1 f7f6 g5e6 d7e6 h4e4 d8d6 b2b4 e8d8 f2f4 e7f7 g2g4 f6f5 g4f5 e6f5 e4e7 f7f6 e7h7');
INSERT INTO evaluations VALUES(32094,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(32094,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(32094,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(32094,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(32094,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(32094,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(32094,36,'cp 109','h2h3 g4h5 b1c3 g8f6 a2a3 c7c6 d2d3 h5g6 c4a2 b8a6 f3h4 f8e7 h4g6 h7g6 c1e3 a6c7 d3d4 a7a6 d1d3 c7b5 e1c1 e8g8 d4e5 d6e5 c3a4 d8d3 d1d3 a8d8 h1d1 b5d4 c2c3 d4e2 c1c2 e2f4 d3d8 f8d8 d1d8 e7d8 e3f4 e5f4 a4c5 b7b6 c5a6');
INSERT INTO evaluations VALUES(32092,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(32092,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(32092,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(32092,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(32092,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(32092,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(32092,36,'cp 85','g4h5 d2d3 b8d7 b1c3 c7c6 a2a4 g8f6 g2g4 h5g6 d1e2 d7c5 c1d2 d8c7 h3h4 h7h5 g4g5 f6d7 d3d4 c5e4 c3e4 d6d5 c4d3 e8c8 e1c1 d5e4 d3e4 g6e4 e2e4 e5d4 e4d4 d7b6 d4e3 f8d6 h1e1 c7d7 e3c3 h8e8 c3g7 b6a4 g7f6');
INSERT INTO evaluations VALUES(32084,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(32084,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(32084,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(32084,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(32084,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(32084,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(32084,36,'cp 95','b1c3 c7c6 d1e2 b8d7 a2a4 g8f6 d2d3 d7c5 g2g4 h5g6 c1d2 c5e6 e1c1 d8c7 c4e6 f7e6 d3d4 e5d4 f3d4 e6e5 d4e6 c7e7 e6f8 h8f8 h3h4 e8c8 h4h5 g6f7 h5h6 g7g6 d2g5 f7e6 f2f3 e7c7 e2f2 c8b8 h1g1 b8a8');
INSERT INTO evaluations VALUES(32080,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(32080,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(32080,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(32080,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(32080,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(32080,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(32080,36,'cp 77','c7c6 d1e2 b8d7 a2a4 h5g6 d2d3 g8f6 f3h4 f8e7 g2g4 d7c5 c1e3 d8a5 e3d2 a5d8 e2f3 c5e6 h4g6 h7g6 c4e6 f7e6 h3h4 d6d5 e1c1 f6d7 g4g5 d7c5 e4d5 c6d5 f3g3 d5d4 c3e4 c5a4 g3e5 d8b6 b2b3 e8g8 h4h5 g6h5');
INSERT INTO evaluations VALUES(31590,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(31590,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(31590,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(31590,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(31590,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(31590,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(31590,36,'cp 80','d1e2 b8d7 a2a4 h5g6 d2d3 g8f6 f3h4 f8e7 c1e3 d8a5 e3d2 a5c7 g2g4 d7c5 h4f5 g6f5 g4f5 g7g6 f2f4 f6h5 e2f2 h5f4 e1c1 h8f8 d3d4 c5d7 d2f4 e5f4 f2f4 e8c8 c4b3 g6g5 f4g3 c8b8 c1b1 b8a8 d4d5 d7e5 d5c6 b7c6');
INSERT INTO evaluations VALUES(52365,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(52365,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(52365,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(52365,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(52365,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52365,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(52365,36,'cp 36','d7d5 b1c3 g8f6 g1f3 e7e6 c1g5 f8e7 e2e3 b8d7 a2a4 e8g8 f1e2 b7b6 e1g1 a7a5 c4d5 f6d5 g5e7 d8e7 c3d5 e6d5 d1b3 c8b7 f1c1 f8c8 c1c2 e7b4 b3a2 c8c7 a1c1 h7h6 b2b3 a8c8 a2b1 b4e7 b1b2 c6c5 e2b5');
INSERT INTO evaluations VALUES(16311,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(16311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16311,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(16311,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(16311,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16311,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(16311,36,'cp 6','f8c5 c2c3 e8g8 e1g1 d7d6 f1e1 c5b6 h2h3 c6e7 d3d4 c7c6 b5d3 e7g6 b1d2 c8e6 a2a4 a7a5 d2f1 f6h5 d3c2 h5f4 c1e3 b6c7 f1g3 d8f6 g3f5 f8e8 e3d2 a8d8 d2e3');
INSERT INTO evaluations VALUES(10030,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(10030,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10030,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10030,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(10030,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10030,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(10030,36,'cp 8','c2c3 e8g8 e1g1 d7d6 h2h3 c6e7 d3d4 c5b6 f1e1 d6d5 f3e5 f6e4 b1d2 e4d6 b5d3 f7f6 e5f3 c8f5 d3f1 c7c6 f3h4 f5c8 d2b3 g7g5 h4f3 e7g6 b3c5 f8e8 a2a4 b6c7 b2b4 b7b6 c5a6 c8a6 f1a6 e8e1 f3e1 d6e4 a6d3 f6f5');
INSERT INTO evaluations VALUES(10026,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(10026,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10026,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(10026,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(10026,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(10026,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10026,36,'cp 4','e8g8 e1g1 d7d6 h2h3 c5b6 f1e1 c6e7 d3d4 d6d5 b1d2 e5d4 e4e5 f6e4 f3d4 e7f5 d2e4 d5e4 d4f5 c8f5 d1d8 a8d8 a2a4 a7a6 b5f1 h7h6 a4a5 b6a7 g2g4 f5h7 f1c4 c7c6 c1f4 g7g5 f4e3 a7e3 f2e3 d8d2 e1e2 d2e2 c4e2 f8e8 a1d1 e8e5');
INSERT INTO evaluations VALUES(4271,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(4271,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(4271,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4271,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(4271,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4271,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4271,36,'cp 2','e1g1 d7d6 h2h3 c5b6 b1d2 c6e7 d2c4 c8e6 c4b6 a7b6 b5a4 h7h6 f1e1 c7c5 a2a3 e7g6 a4c2 d6d5 e4d5 d8d5 d3d4 e5d4 c3d4 f8d8 d4c5 d5d1 e1d1 d8d1 c2d1 b6c5 c1e3 b7b6 a3a4 f6d5 e3d2 g8f8 a4a5 b6a5 a1a5 a8a5 d2a5 f7f6 g1h2 g6f4 d1c2');
INSERT INTO evaluations VALUES(4270,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(4270,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(4270,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(4270,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(4270,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4270,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4270,36,'cp 9','d7d6 h2h3 c5b6 f1e1 c6e7 d3d4 d6d5 f3e5 f6e4 b1d2 e4d6 b5d3 c8f5 d2b3 f5d3 e5d3 e7g6 g2g3 d8d7 a2a4 a7a5 h3h4 f8e8 e1e8 a8e8 h4h5 g6f8 d1f3 d6e4 g1g2 f8e6 d3e5 d7e7 c1e3 c7c6 e5d3 h7h6 a1e1 e4f6 f3f5 b6c7');
INSERT INTO evaluations VALUES(3783,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(3783,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(3783,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(3783,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(3783,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3783,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(3783,36,'cp 24','h2h3 c5b6 f1e1 c6e7 d3d4 d6d5 f3e5 f6e4 b1d2 e4d6 b5d3 f7f6 e5f3 c8f5 d2b3 f5d3 d1d3 f8e8 c1f4 d6e4 b3d2 e7g6 f4h2 e4d2 e1e8 d8e8 f3d2 e8f7 a2a4 a7a5 d3b5 a8e8 h2g3 f7e6 d2f3 e6c6 b5c6 b7c6 g1f1 g6f8');
INSERT INTO evaluations VALUES(3782,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(3782,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(3782,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3782,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(3782,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3782,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3782,36,'cp 0','a7a6 b5a4 c5a7 b1d2 c6e7 f1e1 c7c5 d2f1 e7g6 a4c2 h7h6 a2a4 c8e6 d3d4 c5d4 c3d4 f6h5 c1e3 h5f4 d4e5 d6e5 e3a7 a8a7 d1d8 f8d8 f1e3 a7a8 a1d1 h6h5 f3g5 d8d1 e1d1 f4e2 g1f1 e2d4 g5e6 f7e6 g2g3 g6e7 h3h4 e7c6 c2d3 b7b5 d3e2');
INSERT INTO evaluations VALUES(55883,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(55883,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(55883,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(55883,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(55883,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55883,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55883,36,'cp 11','c7c5 g1f3 c5d4 e4d5 d8d5 f1c4 d5d7 e1g1 b8c6 d2b3 g8f6 b3d4 c6d4 d1d4 d7d4 f3d4 f8c5 f1d1 e8g8 c1f4 b7b6 d4b3 c5e7 a2a4 c8b7 c4e2 f8c8 c2c3 a7a5 f4e5 g8f8 b3d4 f6e4 e5f4 e4c5 e2b5 b7e4 d4f3 e4c2 d1d4 c2b3');
INSERT INTO evaluations VALUES(54241,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54241,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54241,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(54241,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(54241,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(54241,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54241,36,'cp 13','g1f3 c5d4 e4d5 d8d5 f1c4 d5d7 e1g1 b8c6 d2b3 g8f6 f3d4 c6d4 d1d4 d7d4 b3d4 f8c5 f1d1 e8g8 c1f4 b7b6 d4b3 c5e7 a2a4 c8b7 c4e2 f8c8 c2c3 a7a5 b3d4 c8d8 f4e5 a8c8 d4b5 f6e4 g1f1 e4c5 b5d6');
INSERT INTO evaluations VALUES(54239,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(54239,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(54239,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54239,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(54239,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54239,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54239,36,'cp 25','c5d4 e4d5 d8d5 f1c4 d5d7 e1g1 b8c6 d2b3 g8f6 f3d4 c6d4 b3d4 a7a6 f1e1 f8c5 c2c3 b7b5 c4e2 e8g8 c1e3 d7c7 a2a4 e6e5 d4c2 f8d8 d1c1 c5e3 c2e3 a8b8 c3c4 b5c4 e2c4 h7h6 c4f1 c7c1 e1c1 f6d5 c1d1 c8e6');
INSERT INTO evaluations VALUES(54396,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(54396,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54396,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(54396,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(54396,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(54396,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54396,36,'cp 23','e4d5 d8d5 f1c4 d5d7 e1g1 g8f6 d2b3 b8c6 f3d4 c6d4 b3d4 d7c7 d1e2 a7a6 c1g5 f8e7 f1e1 e8g8 c4b3 b7b5 c2c3 h7h6 g5h4 f8e8 h4g3 c7b6 a2a4 c8d7 a4a5 b6b7 b3c2 a8c8 g3f4 d7c6 d4c6 b7c6 a1d1 e8d8');
INSERT INTO evaluations VALUES(54330,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(54330,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54330,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(54330,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(54330,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(54330,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54330,36,'cp 17','d8d5 f1c4 d5d7 e1g1 g8f6 d2b3 b8c6 b3d4 c6d4 f3d4 a7a6 c4b3 f8c5 c2c3 b7b5 c1g5 f6e4 g5h4 c8b7 f1e1 g7g5 h4g3 e4g3 h2g3 e8c8 d1h5 d7c7 a2a4 b5b4 h5g5 b4c3 b2c3 c8b8 a1d1 h7h5 d4f3 d8d1 e1d1 h8c8 b3c2 b8a7');
INSERT INTO evaluations VALUES(34491,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(34491,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(34491,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(34491,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(34491,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(34491,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(34491,36,'cp 19','f1c4 d5d7 e1g1 b8c6 d2b3 a7a6 b3d4 c6d4 d1d4 d7d4 f3d4 c8d7 f1d1 f8c5 a2a4 g8f6 c2c3 e8g8 c4e2 e6e5 d4b3 c5b6 c1g5 d7e6 b3d2 a8c8 g5f6 g7f6 d2e4 f6f5 e4f6 g8g7 f6d7 e6d7 d1d7 c8c7 a1d1 f8c8 d7c7 c8c7 g1f1 e5e4 d1d8');
INSERT INTO evaluations VALUES(34486,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(34486,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(34486,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(34486,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(34486,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(34486,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(34486,36,'cp 21','d5d7 e1g1 b8c6 d2b3 a7a6 b3d4 c6d4 d1d4 d7d4 f3d4 c8d7 f1d1 f8c5 a2a4 g8f6 c1f4 e8c8 c2c3 h8e8 d1e1 c5a7 h2h3 a7d4 c3d4 f6d5 c4d5 e6d5 e1c1 d7c6 f2f3 c8d7 g1f2 f7f6 b2b4 g7g5 f4h2 d7e7 c1c5 e7f7 b4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(47809,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(47809,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(47809,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(47809,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(47809,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(47809,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(47809,36,'cp 26','b1c3 f8e7 d2d4 e5d4 f3d4 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 e7d6 f1e1 b8d7 d4b5 d6e5 h2h3 a7a6 b5d4 d7f6 e4c2 h7h6 c1e3 c7c5 d4b3 d8d1 a1d1 b7b6 b3d2 e5c7 c2a4 e8d8 a4c6 a8b8 d2e4 c8e6 e4f6 g7f6 c6e4 e6a2 d1d8 c7d8 e4f3');
INSERT INTO evaluations VALUES(47796,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(47796,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(47796,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(47796,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(47796,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47796,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(47796,36,'cp 26','f8e7 d2d4 e5d4 f3d4 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 c2c3 d7f6 e4c2 f8e8 h2h3 a7a6 f1e1 c7c5 d4f5 c8f5 d1d8 a8d8 c2f5 c5c4 g1f1 b7b5 g2g3 e7c5 e1e8 f6e8 c1f4 g7g6 f5c2 e8f6 a2a4 f6d5 f4d2 c5b6 a1d1 d8e8 d2g5');
INSERT INTO evaluations VALUES(40763,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(40763,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40763,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(40763,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(40763,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40763,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40763,36,'cp 31','d2d4 e5d4 f3d4 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 f1e1 d7f6 e4f3 c7c5 d4b5 c8d7 c3c4 a7a6 b5c3 d8c7 c1g5 a8d8 e1e7 e8e7 g5f6 g7f6 c3d5 d7e6 d5e7 c7e7 d1e2 e7d7 b2b3 b7b5 c4b5 a6b5 a1d1 d7d1 e2d1 d8d1 f3d1 g8f8 d1c2 f6f5 g2g3');
INSERT INTO evaluations VALUES(40781,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(40781,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(40781,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40781,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40781,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(40781,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40781,36,'cp 29','e5d4 f3d4 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 e4c2 d7f6 h2h3 c7c5 d4f5 c8f5 c2f5 d8b6 d1f3 c5c4 f5c2 e7c5 b2b4 c4b3 a2b3 a7a5 c1f4 b6c6 f3c6 b7c6 f1e1 e8e1 a1e1 f6d5 f4d2 g7g6 g2g4 f7f6');
INSERT INTO evaluations VALUES(40831,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40831,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(40831,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40831,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40831,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(40831,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40831,36,'cp 31','f3d4 e8g8 c1f4 c7c6 a2a3 d6d5 e4d5 c6d5 c4a2 b8c6 e1g1 e7c5 d4f3 d5d4 c3e2 c8f5 f1e1 d4d3 c2d3 d8d3 d1d3 f5d3 a1d1 a8d8 b2b4 c5b6 e2c3 d3g6 h2h3 g6f5 b4b5 c6d4 f3d4 b6d4 c3d5 f6d5 d1d4 d5f4');
INSERT INTO evaluations VALUES(16529,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16529,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16529,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16529,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16529,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(16529,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(16529,36,'cp -3','d2d4 f6e4 f1d3 d7d5 f3e5 f8d6 e1g1 d6e5 d4e5 c6e5 d3e4 d5e4 d1a4 e5c6 a4e4 c8e6 b1a3 d8e7 a3b5 a7a6 b5d4 c6d4 e4d4 e8g8 f1e1 f8e8 c1f4 a8d8 d4e3 f7f6 e3g3 c7c6 a2a4 e7f7 a4a5 d8d5 h2h4 d5b5 b2b4 b5d5 e1e3 e8d8 a1e1 d5d1 e1d1');
INSERT INTO evaluations VALUES(16484,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16484,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16484,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16484,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16484,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(16484,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16484,36,'cp -7','f6e4 f1d3 d7d5 f3e5 f8d6 e1g1 d6e5 d4e5 c6e5 d3e4 d5e4 d1a4 e5c6 a4e4 c8e6 b1a3 d8e7 a3b5 a7a6 b5d4 c6d4 e4d4 e8g8 d4e4 f8e8 c1e3 c7c6 f1e1 a8d8 f2f3 d8d5 a2a4 e8d8 e4f4 e7d6 f4d6 d8d6 a4a5 h7h5 g1f2 d5d3 e3d4 f7f6 d4c5 d3d2 e1e2 d2e2 f2e2 e6c4 e2f2');
INSERT INTO evaluations VALUES(16847,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16847,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(16847,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(16847,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16847,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16847,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16847,36,'cp 0','f1d3 d7d5 f3e5 f8d6 e1g1 d6e5 d4e5 c6e5 d3e4 d5e4 d1a4 e5c6 a4e4 c8e6 b1a3 a7a6 a3c2 d8d5 e4d5 e6d5 f1e1 d5e6 e1d1 e6d7 d1e1');
INSERT INTO evaluations VALUES(56202,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(56202,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(56202,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(56202,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(56202,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(56202,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(56202,36,'cp 0','g8f6 c1f4 f8d6 f4d6 d8d6 d1d3 e8g8 e1c1 f8e8 d3g3 d6d8 g1e2 c7c6 e2f4 c8f5 f1d3 f5d3 f4d3 d8b6 d3e5 b8d7 h1e1 b6c7 e5d3 c7b6');
INSERT INTO evaluations VALUES(48664,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48664,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48664,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48664,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48664,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48664,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48664,36,'cp 0','c1f4 f8e7 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(48643,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48643,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48643,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48643,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48643,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48643,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48643,36,'cp -3','f8d6 f4d6 d8d6 d1d3 e8g8 e1c1 f8e8 d3g3 d6g3 h2g3 c8f5 g1h3 h7h5 h3f4 c7c6 f2f3 g7g6 f1d3 f5d3 f4d3 b8d7 d1e1 e8e1 h1e1 a8e8 c1d2 e8e1 d2e1 f6e8 c3d1 g8g7 d1e3 e8d6 a2a4 a7a5 e1e2 g7f6 b2b3 d7f8 c2c3 f8e6 e2f2');
INSERT INTO evaluations VALUES(21857,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(21857,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(21857,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(21857,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(21857,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21857,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(21857,36,'cp 50','a7a6 b5a4 b7b5 a4b3 c6a5 d2d4 e5d4 f3d4 c8b7 c1d2 a5b3 d4b3 g8f6 b3a5 d8c8 f1e1 f8e7 e4e5 d6e5 e1e5 b7d5 d2g5 d5e6 b1c3 h7h6 g5h4 e8g8 a5c6 e7d6 h4f6 g7f6 e5e3 c8d7 c6d4 d6e5 d4f3 d7d1 a1d1 e5d6 f3d4 d6e5 d4c6 e5c3 e3c3 f8e8');
INSERT INTO evaluations VALUES(17782,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17782,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(17782,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17782,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17782,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(17782,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(17782,36,'cp 36','b5a4 g8f6 f1e1 b7b5 a4b3 f8e7 c2c3 e8g8 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 a2a4 h7h6 b3c2 e5d4 c3d4 c6b4 c2b1 g7g6 d1b3 c7c5 d4d5 f8g7 a1a3 f6d7 d2f1 b5a4 b3a4 a6a5 f1e3 b7a6 a4d1 d8b6');
INSERT INTO evaluations VALUES(17807,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(17807,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17807,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(17807,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(17807,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17807,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(17807,36,'cp 26','g8f6 c2c3 f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 d2c4 d6e5 c4e5 a6a5 e5d3 h7h6 g5h4 d8c8 d3c5 e6d5 f2f3 g7g5 h4g3 b7b6 g3e5 h8e8 e5f6 e8e1 a1e1 b6c5 e1e5 d5a2 e5c5 c8b7 g1f2 a2e6 c5b5 b7c6');
INSERT INTO evaluations VALUES(11433,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11433,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11433,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(11433,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(11433,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(11433,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11433,36,'cp 36','c2c3 f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 e5d3 d8c8 g5f6 g7f6 d2e4 h8d8 e4d6 d8d6 d3c5 a6a5 e1e4 b7b6 c5e6 f7e6 e4h4 d6d7 h4h6 a5a4 h6f6 c8b7 a1b1 a4a3 g2g4 a8a4 h2h3 a4c4 b2a3 c4c3 g1g2 c3a3 f6e6');
INSERT INTO evaluations VALUES(11425,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(11425,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(11425,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(11425,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(11425,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(11425,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11425,36,'cp 37','f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 d2c4 d6e5 c4e5 a6a5 e5d3 d8c8 d3c5 h8e8 c5e6 e8e6 g5f6 g7f6 g1f1 a8a6 a1d1 a6d6 d1d6 e6d6 f1e2 d6e6 e2d2 c8d7 e1e6 f7e6 d2d3 d7c6 d3d4 c6b5 h2h4 h7h5 d4e3 b5c4');
INSERT INTO evaluations VALUES(11437,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(11437,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(11437,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(11437,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(11437,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(11437,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(11437,36,'cp 26','d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 e5d3 h7h6 g5f6 g7f6 d2e4 d8e7 e4d6 e7d6 d3f4 a8d8 b2b3 f6f5 a1d1 d6e7 g1f1 c7c6 f1e2 a6a5 d1d3 e7f6 e1d1 d8d3 d1d3 a5a4 f4e6 f7e6 b3a4 h8a8 d3d4 c6c5 d4h4 f6g6 g2g4');
INSERT INTO evaluations VALUES(11436,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11436,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(11436,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(11436,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(11436,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(11436,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11436,36,'cp 34','c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 e5d3 h7h6 g5f6 g7f6 d2e4 d8e7 e4d6 e7d6 d3f4 a8d8 b2b3 f6f5 c3c4 d6e7 e1e5 e7f6 a1e1 d8d6 c4c5 d6c6 f4d5 e6d5 e5d5 c6e6 e1e6 f7e6 d5d7 h8c8 g1f1 e6e5 f1e2 a6a5 e2d3');
INSERT INTO evaluations VALUES(41360,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(41360,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(41360,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(41360,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(41360,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(41360,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(41360,36,'cp 43','c1e3 e8g8 d1d2 a7a6 e1c1 b7b5 f1d3 c7c5 d4c5 b5b4 c3d5 f6d5 e4d5 d6c5 e3c5 d8d5 d2b4 d5a2 b4a3 g7h6 f3d2 a2a3 c5a3 a8a7 h1e1 b8c6 c1b1 h6g7 d3e4 a7c7 d2b3 c6e5 e4d5 h7h6 h2h3 f8d8 b3c5 h6h5 d5b3 e5c6 d1d5 a6a5 f2f3 d8d5 b3d5 c6d4 c2c3');
INSERT INTO evaluations VALUES(41356,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(41356,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(41356,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(41356,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(41356,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(41356,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(41356,36,'cp 41','e8g8 d1d2 a7a6 f1d3 b7b5 e3h6 b8c6 c3d5 g7h6 d2h6 f6g4 h6d2 e7e6 d5e3 g4e3 d2e3 e6e5 d4d5 c6b4 a2a3 b4d3 c2d3 c8g4 f3d2 a8c8 a1c1 g4d7 e1g1 c7c6 d5c6 d7c6 d2f3 d8e7 h2h3 f8e8 d3d4 e5d4 f3d4 c6b7 c1c8 e8c8 f1e1 b5b4 a3b4 c8c4 b2b3 c4b4');
INSERT INTO evaluations VALUES(36654,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(36654,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36654,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(36654,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(36654,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(36654,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(36654,36,'cp 66','d1d2 a7a6 e3h6 b7b5 f1d3 b8c6 c3d5 g7h6 d2h6 f6g4 h6d2 e7e6 d5e3 g4e3 d2e3 e6e5 d4d5 c6b4 a2a3 b4d3 c2d3 c8g4 a1c1 f7f5 h2h4 g4f3 g2f3 d8e7 h4h5 g6g5 h5h6 f5e4 d3e4 f8f4 e1e2 g8h8 c1c3 e7f7 h1c1 f7h5 c3c7 a8f8 c1c3');
INSERT INTO evaluations VALUES(36655,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(36655,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(36655,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(36655,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(36655,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(36655,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(36655,36,'cp 34','a7a6 e3h6 b7b5 h6g7 g8g7 f1d3 b8c6 e4e5 d6e5 d4e5 f6g4 e1c1 b5b4 h2h3 b4c3 d2c3 g4f2 e5e6 f7f6 c3c6 a8b8 d3b5 c8b7 c6c5 f2e4 c5e3 e4d6 b5d3 c7c5 e3c5 b8c8 c5e3 d8c7 f3d4 d6c4 d3c4 c7c4 c1b1 f8d8 a2a3 b7d5 b2b3 c4c3 e3c3 c8c3');
INSERT INTO evaluations VALUES(19998,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(19998,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(19998,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(19998,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(19998,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(19998,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(19998,36,'cp 37','d2d4 c5d4 f3d4 e7e5 d4f5 c8f5 e4f5 d8a5 d1d5 a5d5 c3d5 e8c8 f1c4 c6d4 c4d3 g8e7 d5e7 f8e7 c1e3 d6d5 e3d4 e5d4 g2g3 c8c7 e1f1 e7f6 f1g2 h7h5 h2h4 d8c8 a2a4 c7d6 a4a5 c8c5 a1a4 h8c8 a4b4 c8c7 h1a1 c7d7 g2f1 d7c7 f1e2 f6e7 b4d4 e7f6');
INSERT INTO evaluations VALUES(19987,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(19987,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(19987,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(19987,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(19987,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(19987,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(19987,36,'cp 54','c5d4 f3d4 g8f6 c1g5 e7e6 d1d2 a7a6 e1c1 c8d7 f2f3 f8e7 c1b1 b7b5 g5e3 h7h5 d4c6 d7c6 f1d3 f6d7 h2h4 d7e5 e3g5 d8c7 c3e2 a8c8 e2d4 c6b7 d3e2 e8g8 d2e3 f8d8 g2g4 h5g4 f3g4 d8d7 g5e7 d7e7');
INSERT INTO evaluations VALUES(20085,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(20085,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(20085,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(20085,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(20085,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(20085,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(20085,36,'cp 51','f3d4 g8f6 c1g5 e7e6 d1d2 f8e7 e1c1 e8g8 f2f4 c6d4 d2d4 h7h6 g5f6 e7f6 d4e3 e6e5 g2g3 d8a5 a2a3 e5f4 g3f4 f6c3 e3c3 a5c3 b2c3 c8g4 d1d6 g4f3 h1g1 f3e4 f1d3 e4d3 c2d3 g8h8 c3c4 f8d8 d6d8 a8d8 c1d2 g7g6 h2h4 h8g7 h4h5');
INSERT INTO evaluations VALUES(20073,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(20073,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(20073,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(20073,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(20073,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(20073,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(20073,36,'cp 56','g8f6 c1g5 a7a6 d1d3 e7e6 e1c1 c8d7 f2f4 h7h6 g5h4 f8e7 d4f3 a8c8 h4f6 e7f6 d3d6 f6c3 b2c3 d8c7 f1e2 c6b8 d6c7 c8c7 c3c4 d7c6 e2d3 b8d7 g2g3 d7c5 d1e1 e8e7 e1e3 h8d8 h1d1 c5d7 f4f5');
INSERT INTO evaluations VALUES(13370,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(13370,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(13370,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(13370,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(13370,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(13370,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(13370,36,'cp 57','c1g5 e7e6 d1d2 c6d4 d2d4 f8e7 e1c1 e8g8 f2f4 d8a5 e4e5 d6e5 d4e5 a5e5 f4e5 f6d5 g5e7 d5e7 f1d3 b7b6 c3b5 c8b7 d1d2 b7c6 b5d6 e7c8 b2b4 c8d6 e5d6 f8d8 d3a6 a8b8 c2c4 f7f5 h1d1 g8f7 b4b5 c6b7 a6b7 b8b7 c1b1 h7h6 b1b2');
INSERT INTO evaluations VALUES(13348,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(13348,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(13348,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(13348,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(13348,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(13348,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(13348,36,'cp 41','e7e6 d1d2 f8e7 e1c1 e8g8 h2h4 c6d4 d2d4 c8d7 f2f3 d7c6 c1b1 d8c7 g2g4 f8c8 f1d3 b7b5 h4h5 b5b4 c3e2 a7a5 h5h6 g7g6 d3c4 e6e5 d4d3 a8b8 b2b3 a5a4 g5f6 e7f6 d3d6 f6g5 d6c7 c8c7 e2c1 c6b5 c4b5 b8b5 c1d3');
INSERT INTO evaluations VALUES(13974,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(13974,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(13974,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(13974,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(13974,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(13974,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(13974,36,'cp 47','d1d2 c6d4 d2d4 f8e7 e1c1 e8g8 h2h4 c8d7 f2f3 d7c6 c1b1 d8c7 c3b5 c6b5 f1b5 h7h6 g5c1 d6d5 e4e5 f6d7 b5d7 c7d7 d4g4 g8h7 f3f4 a8c8 g4f3 d7c6 c2c3 b7b5 g2g4 b5b4 f3d3 h7g8 c3b4 e7b4 a2a3 b4a5 f4f5');
INSERT INTO evaluations VALUES(13975,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(13975,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(13975,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(13975,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(13975,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(13975,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(13975,36,'cp 47','f8e7 e1c1 e8g8 h2h4 a7a6 g5f6 e7f6 d4c6 b7c6 d2d6 d8b6 d6g3 a8b8 e4e5 b6b2 c1d2 f6e7 d1b1 b2b1 c3b1 b8b1 g3h3 e7b4 d2e3 b4c5 e3f3 b1b2 f1d3 c5d4 h4h5 h7h6 f3e2 c6c5 h3h4 c8b7 h1d1 d4e5 h4e7 e5f6 e7c5 b7g2 f2f3');
INSERT INTO evaluations VALUES(8189,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(8189,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(8189,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(8189,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(8189,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(8189,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(8189,36,'cp 42','e1c1 e8g8 f2f4 c6d4 d2d4 d8a5 e4e5 d6e5 d4e5 a5e5 f4e5 f6d5 g5e7 d5e7 f1d3 b7b6 d3e4 a8b8 h1e1 c8b7 d1d7 b7e4 e1e4 e7c6 c3b5 f8d8 d7c7 b8c8 c7c8 d8c8 e4c4 c6e7 c4c8 e7c8 c1d2 g8f8 d2d3 f8e7 a2a4 e7d7 c2c3 a7a6 b5d4 c8e7 d3c4');
INSERT INTO evaluations VALUES(8188,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(8188,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(8188,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(8188,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(8188,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(8188,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8188,36,'cp 43','e8g8 h2h4 a7a6 c1b1 c6d4 d2d4 d8c7 g5f6 e7f6 d4d6 c7d6 d1d6 b7b5 a2a4 f6e5 d6d1 c8b7 f2f3 b5a4 c3a4 b7c6 a4b6 a8b8 b6c4 e5c7 h4h5 h7h6 d1d3 a6a5 d3c3 f8c8 f1d3 b8b4 b1a2 c8b8 b2b3 c6e8 c4b2 c7f4 c3c4 a5a4 c4b4 b8b4 b3a4 e8a4 b2a4');
INSERT INTO evaluations VALUES(47446,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(47446,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(47446,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(47446,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(47446,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(47446,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(47446,36,'cp 104','g1f3 c8f5 b1d2 d5f6 g2g3 b8c6 f1g2 e7e6 e1g1 f8e7 f3e5 c6d4 e2e3 d4c2 e3e4 c2a1 e4f5 c7c6 f5e6 f7e6 d1e2 a1c2 d2b3 c2b4 f1d1 b4d5 b3d4 e8g8 d4e6 d8e8 e6f8 e7f8 d1e1 f8d6 e2c4 d6e5 c1f4 f6d7 g2d5 c6d5 c4d5 e8f7 d5f7 g8f7 f4e5 d7e5 e1e5 a8d8 g1f1 d8d1 e5e1');
INSERT INTO evaluations VALUES(47444,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(47444,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(47444,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(47444,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(47444,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(47444,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(47444,36,'cp 101','c8f5 b1d2 d5b6 e2e4 f5g6 h2h4 h7h6 f1d3 g6h5 d1c2 b8d7 a2a4 a7a5 d2b3 e7e5 f3e5 d7e5 d4e5 f8b4 c1d2 d8e7 e1g1 e8g8 d2b4 e7b4 b3c5 h5g4 c2c3 b4c3 b2c3 b6d7 c5d7 g4d7 f1d1 f8d8 d3c4 d7e8 c4b3 e8c6 f2f3');
INSERT INTO evaluations VALUES(29802,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(29802,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(29802,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(29802,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(29802,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(29802,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(29802,36,'cp 119','b1d2 d5b6 e2e4 f5g6 h2h4 h7h6 f1d3 b8d7 a2a4 a7a5 h4h5 g6h7 e1g1 e7e6 d2b3 f8b4 c1d2 d8e7 d1e1 e8g8 b3a5 c7c5 d2b4 c5b4 a5b3 b6a4 e1e2 e6e5 d4d5 d7f6 d3b5 a4c5 b3c5 e7c5 a1c1 c5e7 f3d2 a8a2 c1b1 g8h8 f1d1 f8a8');
INSERT INTO evaluations VALUES(29801,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(29801,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(29801,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(29801,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(29801,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(29801,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(29801,36,'cp 124','d5f6 g2g3 b8c6 f1g2 e7e6 e1g1 f8e7 f3e5 c6d4 e2e3 d4c2 e3e4 c2a1 e4f5 c7c6 f5e6 f7e6 d1e2 a1c2 d2b3 c2b4 f1d1 b4d5 b3d4 e8g8 d4e6 d8e8 e6f8 e7f8 d1e1 f8d6 e2c4 d6e5 c1f4 f6d7 g2d5 c6d5 c4d5 e8f7 d5f7 g8f7 f4e5 d7e5 e1e5 a8c8 e5f5 f7e6 f5b5 b7b6 b2b3 c8c3');
INSERT INTO evaluations VALUES(20342,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(20342,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(20342,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(20342,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(20342,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(20342,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(20342,36,'cp 29','c2c3 c8d7 f1e2 c5d4 c3d4 g8e7 e1g1 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 b7b5 a4c3 b5b4 c3a4 c6a5 a1c1 d7b5 b2b3 b5a4 b3a4 f8e7 e2d3 e8g8 d3f5 e6f5 d1c2 g7g6 f1e1 a7d7 c2c7 d7d8 e1d1 d8c7 c1c7');
INSERT INTO evaluations VALUES(51500,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51500,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(51500,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51500,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(51500,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(51500,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(51500,36,'cp 59','c8b7 b1c3 e7e6 d2d4 f8b4 f1d3 g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 d2f3 b4c3 b2c3 d8c7 f1e1 c6e5 f3e5 c7e5 a1b1 a8a2 c3c4 b7c6 f2f4 e5c5 c1e3 c5e7 d1f3 d7d6 b1b6 e7c7 b6b1 c7e7 e3f2 f6d7 b1a1 f8a8 a1a2 a8a2 f3g3 e6e5 f4f5 h7h6 h2h3 a2a8 f2e3');
INSERT INTO evaluations VALUES(30824,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(30824,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(30824,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(30824,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(30824,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(30824,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(30824,36,'cp 80','b1c3 e7e6 d2d4 f8b4 f1d3 g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 b4c3 b2c3 e8g8 d2f3 d8c7 f1e1 c6e5 f3e5 c7e5 a1b1 a8a2 c3c4 b7c6 f2f4 e5c5 c1e3 c5e7 d1f3 d7d6 b1b6 e7c7 b6b1 f8a8 e3d4 e6e5 f4e5 d6e5 d4e3 c6a4 b1c1 f6d7');
INSERT INTO evaluations VALUES(30815,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(30815,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(30815,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(30815,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(30815,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(30815,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(30815,36,'cp 79','e7e6 d2d4 f8b4 f1d3 g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 d2f3 b4c3 b2c3 d8c7 f1e1 h7h6 a1b1 a8a2 c1e3 d7d5 e3b6 c7c8 e4d5 f6d5 b1b3 d5b6 b3b6 c8a8 c3c4 c6e7 f3h4 b7c6 b6b1 f8b8 b1b8 a8b8 h4f3 c6f3 d1f3 b8b4 f3e4 g7g6 h2h3 b4b2 e1d1 e7f5');
INSERT INTO evaluations VALUES(30689,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(30689,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(30689,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(30689,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(30689,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(30689,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(30689,36,'cp 82','d2d4 f8b4 f1d3 g8f6 f3d2 e8g8 e1g1 d7d6 a2a3 b4c3 b2c3 c7c5 f1e1 c5d4 c3d4 b7a6 c2c4 b8c6 c1b2 a8c8 d3f1 c6a5 a1c1 h7h6 d4d5 e6e5 a3a4 f8e8 b2a3 a5b7 h2h3 f6d7 a3b4 g8h8 e1e3 b7c5 e3g3 e8g8 c1c3 d7f6');
INSERT INTO evaluations VALUES(30667,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(30667,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(30667,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(30667,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(30667,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(30667,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(30667,36,'cp 77','f8b4 f1d3 g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 b4c3 b2c3 e8g8 d2f3 h7h6 f1e1 d7d5 e4d5 f6d5 c1d2 f8e8 c3c4 d5f6 d2c3 f6d7 h2h3 d7c5 e1e3 c5a4 d1e1 d8c7 a1b1 a4c3 e1c3 c6e7 c3b2 e7f5 d3f5 e6f5 e3e8 a8e8 b2b6');
INSERT INTO evaluations VALUES(27946,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(27946,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(27946,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(27946,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(27946,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(27946,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(27946,36,'cp 62','f1d3 g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 d2f3 b4c3 b2c3 h7h6 f1e1 d8c7 a1b1 a8a2 c1e3 b6b5 d1d2 f6g4 e3f4 d7d6 c3c4 e6e5 c4b5 e5f4 d3c4 a2a5 b5c6 b7c6 d2f4 g4e5 c4b3 e5f3 f4f3 a5e5 f3d3 f8e8 f2f3 d6d5');
INSERT INTO evaluations VALUES(27950,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(27950,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(27950,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(27950,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(27950,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(27950,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27950,36,'cp 77','g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 d2f3 b4c3 b2c3 h7h6 f1e1 d7d6 a2a3 d8c7 a1b1 f8c8 c1e3 f6d7 f3d4 d7c5 d4b5 c7e7 d3f1 c8d8 f2f3 c6e5 c3c4 b7c6 b1a1 a8a4 b5c3 a4a6 d1c1 d8c8 c1d2 c8a8 a1d1 a6a3 c3b5 c6b5');
INSERT INTO evaluations VALUES(28253,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(28253,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(28253,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(28253,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(28253,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(28253,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(28253,36,'cp 72','f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 d2f3 b4c3 b2c3 d8c7 f1e1 h7h6 a1b1 a8a2 c1e3 d7d5 e3b6 c7c8 e4d5 f6d5 b1b3 d5b6 b3b6 c6e7 c3c4 c8a8 f3h4 b7c6 h2h3 a2a5 b6b1 a5g5 d1d2 g5h5 g2g3 h5a5 d2c3 f8d8');
INSERT INTO evaluations VALUES(57754,30,'cp -161',NULL);
INSERT INTO evaluations VALUES(57754,31,'cp -152',NULL);
INSERT INTO evaluations VALUES(57754,32,'cp -154',NULL);
INSERT INTO evaluations VALUES(57754,33,'cp -176',NULL);
INSERT INTO evaluations VALUES(57754,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(57754,35,'cp -158',NULL);
INSERT INTO evaluations VALUES(57754,36,'cp -146','d7d5 h2h3 h7h5 g4h5 e7e5 f1g2 b8c6 d2d3 c8e6 b1c3 g8f6 c1g5 d8d7 h5h6 d5d4 c3e4 f6e4 g2e4 f7f6 h6g7 f8g7 e4g6 e6f7 g6f7 d7f7 g5d2 f6f5 c2c4 e8c8 d1b3 e5e4 e1c1 g7h6 h3h4 h6d2 d1d2 f7g6 d2d1 c8b8 c1b1 b8a8 f2f4 g6g2');
INSERT INTO evaluations VALUES(55098,30,'cp -153',NULL);
INSERT INTO evaluations VALUES(55098,31,'cp -166',NULL);
INSERT INTO evaluations VALUES(55098,32,'cp -156',NULL);
INSERT INTO evaluations VALUES(55098,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(55098,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(55098,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(55098,36,'cp -165','e2e3 b8c6 b1c3 e7e5 d2d4 c8e6 d4e5 c6e5 h2h3 h7h5 g1f3 e5f3 d1f3 h5g4 h3g4 e6g4 f3g2 h8h1 g2h1 g8f6 c1d2 d8d6 h1g2 e8c8 f2f3 g4d7 e1c1 g7g6 c3b5 d7b5 f1b5 d6e6 c1b1 a7a6 b5f1 f6d7 d2c1 f8e7 f1e2 d7c5 a2a3 e7d6 d1h1 c7c6 h1h7 d8e8');
INSERT INTO evaluations VALUES(55095,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(55095,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(55095,32,'cp -174',NULL);
INSERT INTO evaluations VALUES(55095,33,'cp -178',NULL);
INSERT INTO evaluations VALUES(55095,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(55095,35,'cp -159',NULL);
INSERT INTO evaluations VALUES(55095,36,'cp -151','b8c6 b1c3 e7e5 d2d4 c8e6 d4e5 c6e5 f1e2 g8f6 h2h3 f8d6 f2f4 e5c6 g1f3 h7h5 g4g5 f6e4 f3d4 c6d4 d1d4 e4g3 h1g1 g3e2 c3e2 d8e7 c1d2 e6h3 e1c1 h3g4 d1e1 c7c6 d4g7 e8c8 g7c3 c6c5 c3d3 e7e6 d2c3 h8e8 c3f6 d8d7 c1b1 e6e3 d3d5');
INSERT INTO evaluations VALUES(57758,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57758,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57758,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(57758,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(57758,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57758,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57758,36,'cp 0','e7e5 d2d4 e5e4 c3c4 c7c6 b1c3 d7d5 c1f4 f8b4 c4d5 b4c3 b2c3 c6d5 e2e3 b8c6 c3c4 g8e7 d1d2 e8g8 c4c5 b7b6 a1c1 h7h6 f4d6 c6a5 f1b5 c8d7 b5a6 d7c8');
INSERT INTO evaluations VALUES(57205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57205,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57205,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57205,36,'cp 0','d2d4 e5e4 c3c4 c7c6 b1c3 d7d5 c1f4 f8b4 c4d5 b4c3 b2c3 c6d5 e2e3 b8c6 c3c4 g8e7 d1d2 e8g8 c4c5 b7b6 a1c1 h7h6 f4d6 c6a5 f1b5 c8d7 b5a6 d7c8');
INSERT INTO evaluations VALUES(57138,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57138,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57138,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57138,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57138,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57138,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57138,36,'cp 0','e5e4 c3c4 c7c6 b1c3 d7d5 c1f4 f8d6 f4d6 d8d6 e2e3 g8e7 c4d5 c6d5 g1e2 b8d7 d1b3 e8g8 c3b5 d6b6 b5c3');
INSERT INTO evaluations VALUES(57290,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57290,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57290,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57290,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57290,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57290,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57290,36,'cp 0','c3c4 c7c6 b1c3 d7d5 c1f4 f8b4 c4d5 b4c3 b2c3 c6d5 e2e3 b8c6 c3c4 g8e7 c4c5 e8g8 a1c1 b7b6 d1d2 h7h6 f1b5 c6a5 f4d6 c8d7 b5a6 d7c8');
INSERT INTO evaluations VALUES(44541,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(44541,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(44541,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(44541,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(44541,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(44541,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(44541,36,'cp 84','d4d5 b7b5 c4b5 a7a6 e2e3 a6b5 f1b5 d8a5 b1c3 c8b7 c1d2 a5b6 a2a4 f6d5 g1f3 e7e6 e1g1 d5b4 e3e4 f8e7 d2f4 e8g8 f4b8 d7d6 b8d6 e7d6 d1e2 f8d8 c3b1 b6c7 b1a3 b4c6 g2g3 c6d4 f3d4 c5d4 f1d1 c7c5 a3c4 d6e7 g1g2 f7f5 f2f3 f5e4 f3e4');
INSERT INTO evaluations VALUES(56357,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(56357,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(56357,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(56357,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(56357,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(56357,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(56357,36,'cp 80','d4e5 g8e7 c4d5 d8d5 d1d5 e7d5 a2a3 b8c6 g1f3 h7h6 h2h4 g7g6 e2e4 d5b6 c1f4 f8g7 b1d2 a7a6 a1c1 e8g8 f1e2 c8g4 f4e3 b6d7 f3d4 c6e5 c1c7 f8c8 c7c8 a8c8 f2f3 c8c1 e2d1 g4e6 d4e6 f7e6 e1e2 e5c6 h4h5 g6g5 b2b3 c1a1');
INSERT INTO evaluations VALUES(56305,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(56305,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(56305,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(56305,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(56305,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(56305,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(56305,36,'cp 97','g8e7 c4d5 d8d5 c1d2 d5e5 b1c3 e7d5 g1f3 e5d6 e2e4 d5c3 d2c3 d6d1 a1d1 a7a5 a2a3 b8d7 f1c4 f8e7 f3d4 d7b6 d4b5 e7d8 c4e2 e8g8 c3e5 c7c6 b5c7 d8c7 e5c7 b6a4 d1d2 c8e6 f2f3 f8c8 c7d6 c8d8 e1f2 a4b2 h1b1 b2c4 e2c4 e6c4 b1b7 c4b5 h2h3 h7h6 b7c7 d8e8 f2e3 a8c8');
INSERT INTO evaluations VALUES(52989,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(52989,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(52989,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(52989,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(52989,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52989,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52989,36,'cp 24','b8d7 e1g1 a7a6 b5d3 g8f6 c2c3 e7e6 d3c2 b7b5 d2d4 c8b7 f1e1 f8e7 b1d2 d8c7 d2f1 e8g8 f1g3 f8c8 c2d3 c5c4 d3b1 a6a5 a2a3 e7f8 h2h3 a5a4 b1c2 e6e5 c1e3 c8e8 d1d2 d6d5 e4d5 f6d5 f3e5 d7e5 d4e5 e8e5 e3d4 e5e1 a1e1');
INSERT INTO evaluations VALUES(18603,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(18603,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(18603,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(18603,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(18603,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(18603,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(18603,36,'cp 11','d2d4 c5d4 d1d4 a7a6 b5e2 g8f6 d4e3 e7e5 e1g1 d7c5 b1c3 f8e7 f1d1 c8d7 a2a4 f6g4 e3d2 g4f6 e2c4 e8g8 d2e2 b7b5 a4b5 a6b5 a1a8 d8a8 c4b5 d7b5 c3b5 c5e4 f3g5 f8d8 b5c7 a8b7 g5e4 f6e4 c7d5 e4f6 c2c4 f6d5 d1d5 d8c8 d5d1 c8e8 h2h3');
INSERT INTO evaluations VALUES(18598,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(18598,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18598,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(18598,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18598,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(18598,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(18598,36,'cp 11','c5d4 f3d4 g8f6 e1g1 a7a6 b5d3 d7c5 b1c3 g7g6 f1e1 f8g7 d3f1 e8g8 d4b3 c5b3 a2b3 c8g4 d1d3 g4d7 c1e3 b7b5 f2f3 d7c6 d3d2 f6d7 e3d4 d7e5 c3d5 c6d5 e4d5');
INSERT INTO evaluations VALUES(18596,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18596,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(18596,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18596,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(18596,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(18596,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(18596,36,'cp 20','d1d4 a7a6 b5e2 g8f6 e1g1 e7e5 d4e3 d7c5 f3d2 c5e6 c2c4 e6d4 e2d1 c8e6 b2b3 b7b5 b1c3 f8e7 d2f3 d4f3 e3f3 b5c4 b3c4 a8c8 d1a4 f6d7 a4b3 e8g8 f3e2 d7c5 f1d1 g8h8 a1b1 f7f5 c1a3 f5f4 a3c5 c8c5 c3d5 d8d7');
INSERT INTO evaluations VALUES(18595,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18595,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(18595,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(18595,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(18595,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(18595,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18595,36,'cp 15','a7a6 b5e2 g8f6 e1g1 e7e5 d4e3 d7c5 b1c3 f8e7 a2a4 f6g4 e3d2 g4f6 f1d1 c8d7 e2c4 e8g8 d2e2 b7b5 a4b5 a6b5 a1a8 d8a8 c4b5 d7b5 c3b5 c5e4 f3d2 e4d2 c1d2 a8c6 c2c4 d6d5 d2c3 c6c4 e2c4 d5c4 c3e5 f6e4 d1e1 e4f6 b5c3 f8d8 e5f4 e7c5');
INSERT INTO evaluations VALUES(14264,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(14264,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(14264,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14264,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(14264,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14264,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(14264,36,'cp 15','c1f4 c8f5 e2e3 e7e6 f1b5 f8d6 f3e5 e8g8 b5c6 b7c6 e5c6 d8e8 c6e5 a8b8 a1b1 d6b4 e1g1 b4c3 b2c3 b8b1 d1b1 f6e4 f1e1 c7c5 f2f3 e4c3 b1b3 c5d4 e3d4 c3b5 b3b4 f7f6 a2a4 b5c7 e5d3 c7a6 b4b5 e8c8 f4d6 f8d8 d6c5 f5d3 c2d3 a6c5 d4c5 e6e5');
INSERT INTO evaluations VALUES(14216,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(14216,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14216,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(14216,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(14216,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(14216,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(14216,36,'cp 23','f6h5 f4d2 h5f6 e2e3 e7e6 h2h3 c8d7 f1d3 c6b4 e1g1 c7c5 e3e4 c5d4 f3d4 d8b6 d4f3 d5e4 d3e4 d7c6 d2e3 b6a5 e4c6 b4c6 d1e2 f8e7 c3b5 e8g8 b5d4 a8c8 f1e1 c6d4 e3d4 a5c7 c2c3 f6d5 f3e5');
INSERT INTO evaluations VALUES(55145,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55145,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(55145,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(55145,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(55145,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(55145,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(55145,36,'cp -35','e7e5 d2d4 e5d4 d1d4 g8f6 g1f3 f8e7 e1g1 e8g8 c2c4 c7c5 d4d3 d5d4 b2b4 c5b4 f3d4 a7a5 a2a3 b8a6 c1b2 f8e8 d3d1 e7c5 e2e3 c8g4 d1c1 d8d7 a3b4 a6b4 b1c3 b4d3 c1c2 d3b2 c2b2');
INSERT INTO evaluations VALUES(44793,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(44793,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(44793,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(44793,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(44793,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(44793,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(44793,36,'cp 69','c1e3 b8c6 f2f3 h7h5 f1c4 c8d7 d1d2 c6d4 e3d4 f8h6 d2d3 e8g8 e1g1 h6f4 a1d1 a8c8 c4b3 a7a5 a2a4 e7e6 d4e3 f4e5 g1h1 d8c7 e3g5 d7c6 d3d2 f6d7 g5e7 f8e8 e7d6 e5d6 d2d6 c7d6 d1d6 d7c5 f1d1 g8g7 h2h3 h5h4');
INSERT INTO evaluations VALUES(44789,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(44789,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(44789,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(44789,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(44789,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(44789,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(44789,36,'cp 77','b8c6 f2f3 h7h5 f1c4 c8d7 d1d2 c6d4 e3d4 f8h6 d2f2 d8a5 a1d1 e7e5 d4e3 h6e3 f2e3 a8c8 c4b3 e8e7 h2h4 a5b6 e3b6 a7b6 d1d2 b6b5 a2a3 c8c5 e1d1 d7e6 d1c1 e6c4 c1b1 h8d8 h1e1 c5c8 e1d1 c8c5 c3a2 e7e6 b3c4 b5c4 a2c3 d8d7');
INSERT INTO evaluations VALUES(13036,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(13036,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(13036,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(13036,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(13036,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(13036,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(13036,36,'cp 87','f2f3 h7h5 f1c4 c8d7 d1d2 c6d4 e3d4 f8h6 d2f2 d8a5 a1d1 e8g8 e1g1 h6g7 a2a3 d7c6 g1h1 a8c8 f2h4 c6a4 c4d3 a4c6 h2h3 b7b5 h4g5 f6h7 g5e3 g7d4 e3d4 h7f6 f1e1 a5b6 d4b6 a7b6 c3b5');
INSERT INTO evaluations VALUES(13038,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(13038,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(13038,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(13038,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(13038,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(13038,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(13038,36,'cp 59','f8g7 d1d2 e8g8 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8b6 c3a4 b6a5 b2b3 c8f5 g2g4 f5e6 d4e5 a5b4 c2c4 d5f6 h2h4 f6d7 e5c3 b4c3 a4c3 d7c5 f1d3 a7a5 c3e4 c5e4 d3e4 a5a4 c1c2 a4b3 a2b3 a8a2 c2c3 c6c5');
INSERT INTO evaluations VALUES(8339,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(8339,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(8339,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(8339,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(8339,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(8339,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(8339,36,'cp 62','d1d2 e8g8 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6b5 g5e3 c1b1 a8d8 b5a5 d6d5 a5c7 d8d7 c7c3 d7d8 b2b3 a7a5 b1b2 e3b6 d3e4 d8c8 e4d5 c8c3');
INSERT INTO evaluations VALUES(8340,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(8340,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(8340,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(8340,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(8340,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(8340,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(8340,36,'cp 40','e8g8 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6b5 g5e3 c1b1 a8d8 b5a5 e6f5 d2e2 e3g5 a5c3 f5d3 d1d3 d6d3 c2d3 e7e6 c3c4 g5f5 e2e4 f5g5 c4c2 g5d5 c2c3 d5d3 c3d3 d8d3 b1c2 d3d5 e4c4');
INSERT INTO evaluations VALUES(3256,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(3256,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(3256,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(3256,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(3256,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(3256,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(3256,36,'cp 43','e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6a3 a8c8 g2g3 g5e3 d3e2 e3a3 b2a3 d6d2 d1d2 a7a5 c1b2 g8g7 d2d4 c8b8 b2c3 b8c8 e2c4 e6f5 c3b3 h7h5 d4d2 c8c5 a3a4 h5h4 g3h4 g7f6');
INSERT INTO evaluations VALUES(3255,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(3255,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(3255,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(3255,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(3255,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(3255,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(3255,36,'cp 54','d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6a3 a8c8 g2g3 g5e3 d3e2 d6d2 d1d2 e3a3 b2a3 a7a5 d2d4 e6a2 c1b2 a2e6 e2b5 c8c5 c2c4 g8g7 b2c3 g7f6 d4d2 c5c8 d2d1 h7h5 f3f4 e6f5 d1e1 c8b8 h2h4');
INSERT INTO evaluations VALUES(3153,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(3153,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(3153,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(3153,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(3153,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(3153,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(3153,36,'cp 54','e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6a3 a8c8 g2g3 g5e3 d3e2 e3a3 b2a3 d6d2 d1d2 a7a5 d2d4 e6a2 c1b2 a2e6 e2b5 c8c5 c2c4 g8g7 b2c3 g7f6 f3f4 h7h5 d4d8 g6g5 f4g5 c5g5 d8a8 h5h4 g3h4');
INSERT INTO evaluations VALUES(3147,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(3147,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(3147,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(3147,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(3147,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(3147,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(3147,36,'cp 53','f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6a3 a8c8 g2g3 g5e3 d3e2 e3a3 b2a3 d6c6 d2d8 g8g7 d8c8 c6c8 d1d4 c8c5 f3f4 g6g5 f4g5 c5g5 d4a4 a7a5 c1d2 e6a2 a4g4 g5g4 e2g4 g7f6 d2d3 e7e5');
INSERT INTO evaluations VALUES(3166,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(3166,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(3166,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(3166,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(3166,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(3166,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(3166,36,'cp 54','d4c6 b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6a3 a8c8 g2g3 g5e3 d3e2 e3a3 b2a3 d6c6 d2d8 g8g7 d8c8 e6c8 a3a4 c8e6 e2b5 c6c3 d1d3 c3d3 b5d3 e6a2 c2c4 a2b3 a4a5 b3a4 f3f4 a4c6 c1b2 g7f6 b2c3 f6e6');
INSERT INTO evaluations VALUES(3143,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(3143,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(3143,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(3143,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(3143,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(3143,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(3143,36,'cp 60','b7c6 e3d4 g7d4 d2d4 d8a5 c3d5 c6d5 d4d5 a5c7 d5c4 c7a5 f1d3 c8e6 c4a6 a5g5 d1d2 f8d8 h1d1 d8d6 a6a3 a8c8 g2g3 g5e3 d3e2 e3a3 b2a3 d6c6 d2d8 c8d8 d1d8 g8g7 e2d3 g7f6 c1b2 c6b6 b2c3 e6f5 d3f5 g6f5 f3f4 b6c6 c3b2 c6b6 b2c1 b6a6 d8d7 h7h5 d7b7 a6a3 c1b2 a3f3 b7a7 h5h4');
INSERT INTO evaluations VALUES(2904,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(2904,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(2904,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(2904,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(2904,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(2904,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(2904,36,'cp 55','e3d4 g7d4 d2d4 d8b6 c3a4 b6a5 b2b3 a5c7 f1c4 f8d8 g2g4 c8e6 a4c5 d5f4 c5e6 f4e6 d4c3 c7f4 c1b2 e6c7 d1d8 a8d8 h1e1 e7e6 c3e5 f4e5 e1e5 c7d5 g4g5 g8f8 c4d3 c6c5 d3e4 f8e7 e4d5 e7d6 e5e4 e6d5 e4f4 d8d7 f4f6 d6e5 b2c3 h7h5 c3d3 c5c4 b3c4 d5c4 d3c4 d7d2 h2h4 d2c2 c4b3 c2d2 f6f7 d2d1 f7a7');
INSERT INTO evaluations VALUES(55626,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(55626,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(55626,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(55626,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(55626,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(55626,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(55626,36,'cp 23','d2d3 c7c5 g1f3 b8c6 f1e2 e5f4 c1f4 f8e7 e1g1 g8f6 c2c4 f6g4 b1c3 e8g8 c3d5 g4e5 f3e5 d6e5 f4e3 e7d6 e2g4 c8e6 g4f5 g8h8 d5c3 b7b6 d1g4 d8d7 a1d1 e6f5 g4f5 a8d8 g2g3 f7f6 c3d5 d6e7 f5d7 d8d7 a2a4 c6b4');
INSERT INTO evaluations VALUES(55624,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(55624,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(55624,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(55624,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(55624,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(55624,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(55624,36,'cp 0','b8c6 g1f3 g7g6 f4e5 c6e5 b1c3 e5f3 d1f3 f8g7 c1e3 g8f6 h2h3 c7c6 f1e2 c8e6 e1g1 d6d5 e3g5 d8b6 g1h1 d5e4 f3f4 f6d7 f4d6 g7f8 d6g3 f8g7');
INSERT INTO evaluations VALUES(21990,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(21990,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(21990,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(21990,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21990,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21990,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(21990,36,'cp 25','g1f3 g7g6 f4e5 c6e5 b1c3 e5f3 d1f3 f8g7 f1e2 c8e6 e1g1 g8e7 c3d5 e6d5 e4d5 e8g8 c2c3 d8e8 e2d1 e8b5 d1b3 a8e8 b3c4 b5a4 g2g4 a7a5 c1d2 a4c2 f3g2 c7c6 a1e1 e7d5 c4d5 c6d5 e1e8 f8e8 g2d5');
INSERT INTO evaluations VALUES(21989,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(21989,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(21989,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(21989,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(21989,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(21989,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(21989,36,'cp 12','g7g6 f4e5 c6e5 f1e2 f8g7 b1c3 g8e7 c1f4 c7c5 d1d2 e7c6 f4g5 f7f6 g5e3 e5f3 e2f3 e8g8 e1g1 c8e6 c3e2 c6e5 e2f4 e5f3 f1f3 d8d7 a1f1 a8e8 b2b3 b7b6 c2c4 f6f5 e4f5 f8f5 d3d4 f5f7 h2h3 e6f5 d4c5 d6c5 f4d5 f5e6 d5f6 g7f6 f3f6 d7d2 e3d2 f7f6 f1f6 e6f5 f6d6');
INSERT INTO evaluations VALUES(32120,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(32120,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(32120,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(32120,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(32120,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(32120,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(32120,36,'cp 128','d4e5 b8c6 e5d6 f8d6 h2h3 g4h5 b1c3 d8d7 d1d3 e8c8 c1g5 g8f6 e1c1 h8e8 f1e2 d7e6 d3b5 a7a6 b5f5 d6b4 d1d8 c6d8 f5e6 f7e6 g5f6 g7f6 f3d2 h5f7 a2a3 b4c3 b2c3 e6e5 g2g3 f7e6 e2g4 c8b8 g4e6 d8e6 h3h4 h7h5');
INSERT INTO evaluations VALUES(32074,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(32074,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(32074,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(32074,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(32074,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(32074,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(32074,36,'cp 137','b8c6 e5d6 f8d6 h2h3 g4h5 b1c3 d8d7 d1d3 e8c8 c1g5 f7f6 g5e3 h5f3 g2f3 c6e5 d3e2 d6b4 a1d1 d7d1 e2d1 d8d1 e1d1 e5f3 c3d5 b4d6 h3h4 f3e5 f1h3 c8b8 d1e2 e5c4 e3c1 g8e7 d5e7 d6e7 h4h5 g7g5 h3e6 c4e5 f2f4 g5f4 c1f4 e7d6 h1g1 e5c6 f4e3 a7a6 e2f2');
INSERT INTO evaluations VALUES(26939,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(26939,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(26939,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(26939,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(26939,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(26939,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(26939,36,'cp 127','e5d6 f8d6 b1c3 d8d7 c1e3 e8c8 d1d3 f7f5 e1c1 f5f4 e3d4 d7f7 c3d5 g8e7 d4c3 h8e8 d3b5 e7d5 d1d5 f7e7 b5a4 e7e4 a4e4 e8e4 d5g5 g4f3 g2f3 e4e7 g5g7 e7g7 c3g7 c6e5 f1h3 c8b8 h3f5 e5f3 f5h7 d6c5 g7h6 c5f2');
INSERT INTO evaluations VALUES(26892,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(26892,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(26892,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(26892,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(26892,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(26892,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(26892,36,'cp 122','f8d6 b1c3 d8d7 c1e3 e8c8 d1d3 f7f5 e1c1 f5f4 e3d4 d7f7 f1e2 g4f3 d3f3 g8f6 d4f6 f7f6 c3d5 f6e5 c1b1 c6d4 f3d3 d4e2 d3e2 c7c6 d5c3 d6b4 e2g4 c8b8 c3e2 e5e4 g4f4 e4f4 e2f4 d8d1 h1d1 h8e8 c2c3 g7g5 f4h5 b4a5 b1c2 a5c7');
INSERT INTO evaluations VALUES(23847,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(23847,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(23847,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(23847,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(23847,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(23847,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(23847,36,'cp 134','f1b5 d8e7 b5c6 b7c6 e1g1 a8d8 b1d2 g8f6 h2h3 g4f3 d1f3 e8g8 f1e1 f8e8 f3f5 h7h6 g2g3 e7e6 g1f1 g7g5 a2a4 f6e4 f5e6 e4d2 c1d2 e8e6 e1e6 f7e6 f1e2 c6c5 a1h1 d6e7 d2a5 d8d4 b2b3 e7d6 h3h4 g5g4 a5c3 d4d5 h4h5 d5f5 h1h4 f5g5 a4a5 a7a6');
INSERT INTO evaluations VALUES(10153,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(10153,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(10153,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10153,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10153,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10153,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10153,36,'cp 0','e5d4 c3d4 c5b4 c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e4f6 f1e1 e7d8 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(10294,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10294,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10294,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10294,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10294,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10294,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10294,36,'cp 0','c3d4 c5b4 c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(10289,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10289,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10289,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10289,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10289,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10289,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10289,36,'cp 0','c5b4 c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(10505,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10505,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10505,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10505,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10505,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10505,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10505,36,'cp 0','c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(10503,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10503,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10503,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10503,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10503,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10503,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10503,36,'cp 0','f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(10693,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10693,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10693,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10693,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10693,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10693,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10693,36,'cp 0','d2b4 c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(10687,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10687,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10687,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10687,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10687,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10687,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10687,36,'cp 0','c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(10817,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10817,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10817,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10817,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10817,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10817,36,'cp 0','c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(9653,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9653,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9653,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9653,36,'cp 0','e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(4694,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4694,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4694,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4694,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4694,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4694,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4694,36,'cp 0','d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(4693,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4693,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4693,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4693,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4693,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4693,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4693,36,'cp 0','f7e8 b3b4 d8e7 b4e7 e8e7 e1g1 e7d8 f1e1 e4f6 f3g5 h8f8 b1c3 d7d6 f2f3 c8f5 g5e4 f6e4 f3e4 f5g6 e4e5 d6e5 e1e5 d8d7 e5d5 d7c6 d5c5 c6d7');
INSERT INTO evaluations VALUES(48254,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(48254,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(48254,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48254,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48254,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48254,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48254,36,'cp 0','b7b6 g1f3 c8b7 c4d5 e6d5 f1b5 c7c6 b5d3 f8d6 e1g1 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 e4c2 b8d7 d1d3 d7f6 f3e5 f8e8 c1h6 g7h6 e5f7 g8f7 d3g6 f7f8 g6h6 f8f7 h6g6 f7f8');
INSERT INTO evaluations VALUES(43794,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43794,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43794,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43794,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43794,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43794,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43794,36,'cp 12','g1f3 f8d6 f1d3 c8b7 e1g1 e8g8 b2b3 d5c4 b3c4 c7c5 f1e1 b8d7 a2a4 a8c8 d3f1 f6e4 c3e4 b7e4 c1b2 d7f6 f3d2 c5d4 d2e4 f6e4 d1d4 e4f6 g2g3 d8c7 e1d1 d6c5 d4f4 c7f4 g3f4 f8d8');
INSERT INTO evaluations VALUES(55077,30,'cp -131',NULL);
INSERT INTO evaluations VALUES(55077,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(55077,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(55077,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(55077,34,'cp -147',NULL);
INSERT INTO evaluations VALUES(55077,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(55077,36,'cp -134','d5e4 f3g5 c8f5 b1c3 g8f6 d2d3 e4d3 d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 e4f6 g7f6 d3f3 b8c6 c1e3 h8g8 h2h3 e8c8 a1d1 c6e5 f3e4 d7c6 d1d8 c8d8 e4c6 e5c6 g2g4 h6h5 e1e2 g8h8 h1d1 d8c8 c3e4 h5g4 h3g4 c6e5 f2f3 b7b6 e4f2 e7e6 e3d4 f8d6 f3f4 e5d7 e2f3 h8h2 d1h1 h2h1 f2h1 c8d8');
INSERT INTO evaluations VALUES(55288,30,'cp -126',NULL);
INSERT INTO evaluations VALUES(55288,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(55288,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(55288,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(55288,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(55288,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(55288,36,'cp -125','f3g5 c8f5 b1c3 g8f6 d2d3 e4d3 d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 e4f6 g7f6 d3f3 b8c6 c1e3 h8g8 h2h3 e8c8 a1d1 c6e5 f3e4 d7c6 d1d8 c8d8 g2g4 h6h5 e4c6 e5c6 e1e2 g8h8 h1d1 d8c8 c3e4 h5g4 h3g4 h8h4 f2f3 h4h2 e4f2 e7e6 e3f4 h2h8 c2c3 b7b6 f2e4 f8g7 e4f2 c6e7 d1h1 h8h1 f2h1 f6f5');
INSERT INTO evaluations VALUES(55179,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(55179,31,'cp -130',NULL);
INSERT INTO evaluations VALUES(55179,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(55179,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(55179,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(55179,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(55179,36,'cp -125','c8f5 b1c3 g8f6 d2d3 e4d3 d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 e4f6 g7f6 d3f3 b8c6 c1e3 h8g8 h2h3 e8c8 a1d1 c6e5 f3e4 d7c6 d1d8 c8d8 g2g4 h6h5 e4c6 e5c6 e1e2 g8h8 h1g1 h5g4 h3g4 h8h4 c3d5 f8h6 f2f4 h4h2 e2f3 e7e6 d5f6 h6g7 f6e4 h2c2 g1h1 g7b2');
INSERT INTO evaluations VALUES(31987,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(31987,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(31987,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(31987,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(31987,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(31987,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(31987,36,'cp -133','b1c3 g8f6 d2d3 e4d3 d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 e4f6 g7f6 d3f3 b8c6 c1e3 h8g8 h2h3 e8c8 a1d1 c6e5 f3e4 d7c6 e4c6 e5c6 g2g4 f8g7 d1d8 g8d8 e1e2 e7e6 h1d1 d8e8 c3b5 a7a6 b5d4 c6e7 f2f4 f6f5 g4f5 g7d4 e3d4 e7f5 d4e5 e8g8');
INSERT INTO evaluations VALUES(31985,30,'cp -126',NULL);
INSERT INTO evaluations VALUES(31985,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(31985,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(31985,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(31985,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(31985,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(31985,36,'cp -122','g8f6 d2d3 e4d3 d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 d3b5 d7b5 e4f6 g7f6 c3b5 b8a6 c1e3 c7c6 b5c3 a6c7 e1c1 f6f5 d1d2 h8g8 g2g3 f8g7 h1d1 g7c3 b2c3 c7d5 c1b2 h6h5 e3a7 b7b5 a7e3 e7e6 a2a3 f7f6 d2d4 g8g4');
INSERT INTO evaluations VALUES(29728,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(29728,31,'cp -102',NULL);
INSERT INTO evaluations VALUES(29728,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(29728,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(29728,34,'cp -140',NULL);
INSERT INTO evaluations VALUES(29728,35,'cp -143',NULL);
INSERT INTO evaluations VALUES(29728,36,'cp -133','d2d3 e4d3 d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 e4f6 g7f6 d3f3 b8c6 c1e3 h8g8 a1d1 d7g4 f3g3 a8c8 a2a3 e7e6 c3b5 g4g3 h2g3 a7a6 b5c3 c6e5 e3d4 g8g6 h1h4 c8d8 c3e2 e5c6 d4e3 d8d1 e1d1 c6e7 e2f4 g6g8 f4h5 e7d5 e3d2 f6f5');
INSERT INTO evaluations VALUES(29730,30,'cp -135',NULL);
INSERT INTO evaluations VALUES(29730,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(29730,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(29730,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(29730,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(29730,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(29730,36,'cp -120','e4d3 d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 e4f6 g7f6 d3f3 b8c6 c1e3 h8g8 h2h3 e8c8 a1d1 c6e5 f3e4 d7c6 d1d8 c8d8 e4c6 e5c6 g2g4 h6h5 e1e2 g8h8 h1d1 d8c8 c3e4 h5g4 h3g4 h8h4 f2f3 f8h6 e4f2 h6e3 e2e3 h4h2 d1h1 h2h1 f2h1 c8d8 c2c3 e7e6 h1g3 c6e7 c3c4 e7g6 e3e4 d8e8 g3h5 e8e7 f3f4 a7a6 f4f5');
INSERT INTO evaluations VALUES(29732,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(29732,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(29732,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(29732,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(29732,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(29732,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(29732,36,'cp -134','d1f3 d8c8 f1d3 f5d3 f3d3 h7h6 g5e4 c8d7 e4f6 g7f6 d3f3 b8c6 c1e3 h8g8 h2h3 e8c8 a1d1 c6e5 f3e4 d7c6 d1d8 c8d8 g2g4 h6h5 e4c6 e5c6 e1e2 g8h8 h1d1 d8c8 c3e4 h5g4 h3g4 h8h4 f2f3 e7e6 e4f6 h4h2 e3f2 f8g7 g4g5 g7f6 g5f6 h2h6 f2d4 c6d4 d1d4 h6f6 d4g4 f6g6 g4g6 f7g6 e2e3 g6g5 e3e4 c8d7 e4e5 b7b5');
INSERT INTO evaluations VALUES(16227,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(16227,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(16227,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(16227,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(16227,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(16227,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(16227,36,'cp -20','d7c6 d2d3 f8d6 h2h3 h7h6 b1d2 c8e6 d1e2 d8e7 b2b3 e8c8 c1b2 f6d7 e1c1 h8e8 d3d4 e5d4 f3d4 d6a3 d4e6 a3b2 c1b2 e7e6 h1e1 d7c5 e4e5 f7f6 f2f4 f6e5 f4e5 b7b6 d2f3 d8d1 e1d1 a7a5 d1d4 c8b7 a2a4 e6e7');
INSERT INTO evaluations VALUES(15698,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(15698,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(15698,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(15698,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(15698,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15698,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(15698,36,'cp -15','d2d3 f8d6 a2a4 c8e6 b1a3 h7h6 d1e2 a7a5 f3d2 f6d7 a3c4 d8e7 c4d6 e7d6 b2b3 e8c8 c1a3 c6c5 e1c1 b7b6 a3b2 f7f6 d2c4 d6e7 c4e3 d7b8 g2g3 b8c6 f2f4 c8b7 f4e5 f6e5 b2c3 c6d4 e2f2 e7d7 c1b2');
INSERT INTO evaluations VALUES(15696,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(15696,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(15696,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(15696,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(15696,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15696,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(15696,36,'cp -13','f8d6 e1g1 c8e6 b1d2 f6d7 d2b3 d8e7 f3g5 e8c8 g5e6 e7e6 d1e1 g7g6 f2f4 f7f5 e4f5 e6f5 f4e5 f5e5 e1e5 d6e5 c1g5 d8e8 d3d4 e5d6 a1e1 c6c5 g5e3 c5d4 e3d4 e8e1 f1e1 h8d8 b3d2 d6e5 e1e4 e5f6 d4f6');
INSERT INTO evaluations VALUES(9522,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9522,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9522,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(9522,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(9522,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9522,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9522,36,'cp -13','b1d2 c8e6 e1g1 d8e7 b2b3 e8c8 d2c4 e6c4 b3c4 f6d7 c2c3 b7b6 a1b1 c6c5 d1b3 c8b7 a2a4 a7a5 b3b5 g7g6 c1e3 c7c6 b5b2 d6c7 f1e1 h8e8 g2g3 f7f6 b1d1 b7c8 g1g2 h7h5 h2h3 e8h8 b2e2 g6g5 e1h1');
INSERT INTO evaluations VALUES(19568,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(19568,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(19568,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(19568,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19568,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(19568,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(19568,36,'cp -20','g8f6 b1c3 e7e6 f1d3 f8e7 g1f3 d7d6 e1g1 e8g8 h2h3 a7a6 a2a4 c8d7 d1e2 d8c7 c1d2 c6b4 f1d1 b4d3 c2d3 b7b5 a4b5 a6b5 c3a2 c7b7 d2c3 e6e5 g1h1 f8c8 a2b4 d7e6 e2d2 h7h6');
INSERT INTO evaluations VALUES(12902,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(12902,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(12902,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(12902,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(12902,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(12902,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(12902,36,'cp -4','b1c3 e7e6 g1f3 a7a6 f1d3 d7d6 a2a4 f8e7 e1g1 e8g8 h2h3 c8d7 c1d2 d8c7 f1e1 f8e8 d1e2 c6b4 a4a5 b4d3 c2d3 b7b5 a5b6 c7b6 d2e3 b6b7 e1c1 h7h6 c3a4 d7b5 a4c3 e8c8 c3b5 c8c1 a1c1 b7b5 c1c7 e7d8');
INSERT INTO evaluations VALUES(12891,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(12891,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(12891,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(12891,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(12891,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(12891,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(12891,36,'cp -16','e7e6 g1f3 a7a6 f1d3 d7d6 a2a4 f8e7 h2h3 e8g8 e1g1 d8c7 c1e3 c8d7 f1e1 f8e8 f3d2 d6d5 e4d5 e6d5 e3g5 d7e6 c3e2 a8d8 d2f3 f6e4 g5e3 e6d7 e2f4 d7f5 c2c3 g7g5 f4e2 f5g6 e2d4 c6a5 d3e4 d5e4 f3g5');
INSERT INTO evaluations VALUES(12501,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12501,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(12501,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12501,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12501,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12501,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12501,36,'cp 0','g1f3 a7a6 f1d3 d7d6 a2a4 f8e7 h2h3 e8g8 c1e3 f8e8 e1g1 h7h6 f1e1 c8d7 c3e2 d8c7 e2g3 a8c8 c2c3 c6e5 f3e5 d6e5 g3h5 e8d8 h5f6 e7f6 d1c2 f6g5 e3g5 h6g5 c2e2 g8f8 c3c4 d7c6 b2b4 c7d7');
INSERT INTO evaluations VALUES(12500,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(12500,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(12500,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(12500,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(12500,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(12500,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(12500,36,'cp -18','a7a6 f1d3 d7d6 a2a4 f8e7 h2h3 e8g8 c1e3 d8c7 e1g1 c8d7 f3d4 c6b4 f1e1 b7b6 d1f3 f8c8 e1c1 c7b7 e4e5 b7f3 d4f3 d6e5 f3e5 e7c5 e3c5 c8c5 e5c4 b4d3 c2d3 c5c6 a4a5 b6b5 c4e5 c6c8 f2f4 h7h6');
INSERT INTO evaluations VALUES(11100,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(11100,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(11100,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(11100,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(11100,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(11100,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(11100,36,'cp -5','f1d3 d7d6 a2a4 f8e7 h2h3 e8g8 c1e3 b7b6 e1g1 c6b4 f3d4 e6e5 d4e2 d6d5 c3d5 f6d5 e4d5 b4d5 e2c3 d5e3 f2e3 f7f5 d3c4 g8h8 d1d8 e7d8 a1d1 g7g6 c3d5 c8e6 d5c7 e6c4 c7a8 c4f1 g1f1 b6b5 a4b5 a6b5 f1e2 d8e7');
INSERT INTO evaluations VALUES(57051,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(57051,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(57051,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(57051,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(57051,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(57051,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(57051,36,'cp -20','d7d5 b1c3 d5d4 c3e2 c7c5 e2g3 h7h5 f1b5 b8d7 e4e5 h5h4 g3e4 a7a6 b5e2 d7b6 b2b3 c8d7 c2c4 d4c3 d2c3 g8h6 c3c4 h6f5 e2f3 d8c7 g1h3 e8c8 e1g1 d7c6 d1e1 f5d4 c1b2 f7f6 a1d1 f8e7 b2d4 c5d4 h3f2 f6e5 f4e5 c7e5 f2d3');
INSERT INTO evaluations VALUES(55923,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(55923,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(55923,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(55923,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(55923,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55923,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(55923,36,'cp -23','b1c3 d5d4 c3e2 c7c5 e2g3 h7h5 f1b5 c8d7 b5d7 d8d7 d2d3 h5h4 g3f1 f8d6 g1h3 b8c6 d1e2 e8c8 f1d2 f7f6 e1g1 d6c7 a2a3 c8b8 a1b1 e6e5 b2b4 e5f4 h3f4 c5b4 a3b4 g8e7 d2c4 c6e5 c4e5 c7e5 g1h1 g7g5 f4h5 h4h3 h5f6 h3g2 e2g2');
INSERT INTO evaluations VALUES(55920,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(55920,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(55920,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(55920,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(55920,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(55920,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(55920,36,'cp -38','d5d4 c3e2 c7c5 e2g3 h7h5 f1b5 c8d7 b5d7 d8d7 d2d3 h5h4 g3f1 f8d6 g1h3 b8c6 f1d2 e8c8 e1g1 f7f6 d1e2 d6c7 a1b1 g8e7 a2a3 e6e5 f4f5 c8b8 h3f2 e7c8 d2f3 c8d6 c1d2 d7f7 c2c4 d4c3 b2c3 b7b6 h2h3 f7a2');
INSERT INTO evaluations VALUES(56064,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(56064,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(56064,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(56064,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(56064,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(56064,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(56064,36,'cp -20','c3e2 c7c5 e2g3 h7h5 f1b5 c8d7 b5d7 d8d7 d2d3 h5h4 g3f1 f8d6 g1h3 f7f6 f1d2 b8c6 e1g1 e8c8 a2a3 g8e7 d1e2 d6c7 d2f3 e6e5 f4e5 c6e5 f3e5 c7e5 c1f4 e7g6 f4e5 g6e5 h3f4 e5g4 f4g6 h8e8 g6h4 g4e3 f1f3 c5c4 g1h1');
INSERT INTO evaluations VALUES(56065,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56065,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(56065,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(56065,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(56065,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(56065,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(56065,36,'cp -21','c7c5 e2g3 h7h5 f1b5 c8d7 b5d7 d8d7 d2d3 h5h4 g3f1 f8d6 g1h3 b8c6 f1d2 e8c8 e1g1 f7f6 a2a3 g8e7 d1e2 d6c7 a1b1 e6e5 f4f5 c8b8 h3f2 e7c8 h2h3 c8b6 b2b3 b6c8 b1a1 a7a6 d2f3 c8d6 f2g4 b7b5 g1h1 d8c8 e2f2');
INSERT INTO evaluations VALUES(18731,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18731,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(18731,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(18731,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(18731,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18731,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(18731,36,'cp 0','g8f6 d2d4 d7d5 e4d5 e6d5 c4b5 f8e7 e1g1 e8g8 d4c5 e7c5 c1g5 c8e6 d1d3 a8c8 a1e1 h7h6 g5h4 e6g4 f3e5 g4e6');
INSERT INTO evaluations VALUES(12352,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12352,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(12352,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12352,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(12352,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12352,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(12352,36,'cp 0','d2d4 f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 f2f3 e4e3 c4b5 e8f7 d2c4 a7a6 b5a4 f8c5 c1e3 c5e3 c4e3 b7b5 a4b3 f7f6 e1c1 d8f7 d1d2 h8d8 d2d8 f7d8 h1d1 f6e7 c2c3 h7h6 d1e1 e7f6 e1d1');
INSERT INTO evaluations VALUES(12390,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12390,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12390,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12390,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12390,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(12390,36,'cp 0','f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 f2f3 e4e3 c4b5 e8f7 d2c4 a7a6 b5a4 f8c5 c1e3 c5e3 c4e3 b7b5 a4b3 f7f6 e1f2 d8f7 h2h4 h8d8 h1e1 h7h6 a2a4 c8d7 a4b5 a6b5 a1a8 d8a8 e1d1 d7c6 c2c3 a8d8 d1d8 f7d8 f3f4 d8b7 g2g3 b7c5 b3c2 c6e4 c2e4 c5e4 f2f3 e4d2 f3e2 d2e4');
INSERT INTO evaluations VALUES(22511,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(22511,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(22511,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(22511,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(22511,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(22511,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(22511,36,'cp 14','f8c5 c3a4 d7d6 e1g1 c5b6 a4b6 a7b6 d2d4 c8g4 h2h3 g4f3 d1f3 d8f6 f3f6 g8f6 d4e5 d6e5 f2f3 e8e7 c1e3 c6d4 e3d4 e5d4 e4e5 f6d7 f3f4 d7c5 b2b3 a8a3 f4f5 f7f6 e5e6 c7c6 f1f4 b6b5 c4d3 c5d3');
INSERT INTO evaluations VALUES(6211,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(6211,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(6211,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(6211,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(6211,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6211,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(6211,36,'cp 25','c3d5 g8f6 c2c3 e8g8 d2d3 d7d6 b2b4 c5b6 a2a4 a7a6 d5b6 c7b6 e1g1 d6d5 e4d5 f6d5 f1e1 f8e8 c1d2 d5f6 h2h3 c8f5 d1b3 d8d7 d2e3 f5d3 c4d3 d7d3 e3b6 d3d5 c3c4 d5d6 b4b5 c6d4 b6d4 e5d4 e1e8 a8e8 b3d3 e8e4 a1c1 d6b6 c1e1');
INSERT INTO evaluations VALUES(6254,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(6254,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(6254,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(6254,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6254,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6254,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(6254,36,'cp 27','g8f6 c2c3 a7a5 d2d4 e5d4 c3d4 c5a7 d5c3 e8g8 d4d5 c6e7 e1g1 d7d6 h2h3 c7c6 d5c6 b7c6 f1e1 d6d5 e4d5 f6d5 c3d5 e7d5 c4d5 d8d5 d1d5 c6d5 c1e3 c8d7 a1d1 a7e3 e1e3 f8b8 d1d2 a5a4 f3d4 b8e8 e3e8 a8e8 f2f3 g8f8');
INSERT INTO evaluations VALUES(9826,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(9826,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(9826,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(9826,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(9826,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(9826,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(9826,36,'cp 25','c2c3 e8g8 d2d3 c6a5 b2b4 a5c4 b4c5 c4a5 e1g1 d7d6 c3c4 b7b6 c1d2 a5b7 c5c6 b7c5 d2e3 f6d5 e4d5 c8g4 h2h3 g4f3 d1f3 c5d3 e3h6 g7h6 f3d3 f7f5 f2f4 e5e4 d3d4 d8f6 d4f6 f8f6 g1f2 g8f7 a2a4 a8g8 a4a5');
INSERT INTO evaluations VALUES(55768,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(55768,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(55768,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55768,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(55768,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(55768,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55768,36,'cp 15','e6d5 b1c3 c7c6 c1f4 g8f6 e2e3 c8f5 a2a3 f8d6 g1e2 e8g8 f4d6 d8d6 e2g3 f5g6 h2h4 h7h6 h4h5 g6h7 f1d3 b8d7 d3h7 f6h7 g3f5 d6e6 d1c2 h7g5 g2g4 d7f6 e1c1 g8h8 f2f4 g5e4 c3e4 f6e4 c1b1 a8e8 b1a1 f7f6 h1g1 e4d6 f5d6 e6d6 d1e1 h8g8');
INSERT INTO evaluations VALUES(56213,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(56213,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(56213,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56213,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56213,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56213,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56213,36,'cp 0','b1c3 c7c6 c1f4 g8f6 e2e3 c8f5 a2a3 f8d6 f4d6 d8d6 d1b3 d6e7 b3b4 e7d7 g1f3 a7a5 b4b3 e8g8 f3e5 d7e7 f1d3 f5d3 e5d3 f6e8 c3a4 e7c7 b3b6 c7e7');
INSERT INTO evaluations VALUES(56201,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56201,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56201,36,'cp 6','c7c6 c1f4 g8f6 e2e3 c8f5 f1d3 f5d3 d1d3 f8b4 a2a3 b4e7 h2h3 b8d7 g1f3 e8g8 e1g1 d7b6 d3c2 f8e8 f1e1 h7h6 f4g3 e7d6 f3e5 b6c4 f2f4 d6e7 g1h1 c4d6 g3h4 a7a5 a1c1 f6e4 c3e4 d6e4');
INSERT INTO evaluations VALUES(53718,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(53718,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53718,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(53718,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53718,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(53718,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(53718,36,'cp 10','c1f4 g8f6 e2e3 c8f5 f1d3 f5d3 d1d3 b8d7 h2h3 f8b4 g1e2 f6h5 f4h2 d7b6 a2a3 b4d6 h2d6 d8d6 b2b4 a7a6 e1g1 e8g8 a3a4 d6b4 f1b1 b4a5 b1b3 h5f6 a1b1 f6d7 e2g3 a8e8 e3e4 g7g6 e4e5 f7f6');
INSERT INTO evaluations VALUES(53713,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(53713,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(53713,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(53713,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(53713,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(53713,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(53713,36,'cp 4','g8f6 e2e3 c8f5 f1d3 f5d3 d1d3 f8e7 g1e2 e8g8 h2h3 e7d6 e1g1 f8e8 f1e1 d6f8 d3c2 b8d7 e2c1 f6e4 a2a4 a7a5 c1d3 d7b6 c2b3 e4d2 b3d1 d2e4 c3e2 f8d6 b2b3 d6f4 d3f4 e4d6 a1c1 b6d7 e2c3');
INSERT INTO evaluations VALUES(45763,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(45763,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(45763,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(45763,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45763,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(45763,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(45763,36,'cp 12','e2e3 c8f5 f1d3 f5d3 d1d3 f8e7 g1e2 e8g8 e2g3 b8a6 e1g1 a6c7 g3f5 f8e8 f5e7 d8e7 h2h3 c7e6 f4h2 a8d8 a2a4 a7a5 d3d2 g7g6 f1c1 f6h5 c3e2 d8a8 c1c2 h5g7 e2c1 h7h5 c1d3 h5h4 b2b3 g7f5');
INSERT INTO evaluations VALUES(45761,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(45761,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45761,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(45761,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(45761,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(45761,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45761,36,'cp 0','c8f5 f1d3 f5d3 d1d3 b8d7 h2h3 f8b4 g1e2 e8g8 a2a3 b4c3 e2c3 f8e8 e1g1 f6e4 f1c1 e4c3 c1c3 a7a5 a3a4 d7b6 c3c2 e8e6 d3b3 a8a6 b3c3 d8e7 b2b3 a6a8 c3c5 e7d8 c5c3');
INSERT INTO evaluations VALUES(22145,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(22145,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(22145,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(22145,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(22145,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(22145,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(22145,36,'cp 125','b1c3 g8f6 f2f3 f8e7 c1f4 e8g8 e1c1 c8e6 c1b1 f6d7 d4d2 e7f6 h2h4 f6e5 f4e5 d6e5 f1c4 e6c4 d2d7 d8d7 d1d7 a8c8 h1d1 f7f6 b2b3 c4f7 b1b2 f7e8 d7d2 c7c6 c3a4 f8f7 d2d8 c8d8 d1d8 g8f8 a4c5 b7b6');
INSERT INTO evaluations VALUES(22140,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(22140,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(22140,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(22140,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(22140,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(22140,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(22140,36,'cp 122','g8f6 c1f4 f8e7 e1c1 e8g8 f2f3 c8e6 c1b1 f6d7 h2h4 e7f6 d4d2 f6e5 f4g5 f7f6 g5e3 c7c6 f3f4 e5c3 d2c3 d8e7 f1e2 f6f5 c3a3 f5e4 a3d6 f8e8 g2g4 d7b6 d6e7 e8e7 e3c5 e7d7 f4f5 e6f7 b1c1 d7d5 c5d4 a8d8 d4e3 d5d1 h1d1 d8d1 e2d1 f7a2 b2b3 b6d5 e3a7 d5c3');
INSERT INTO evaluations VALUES(15768,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(15768,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(15768,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(15768,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(15768,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(15768,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(15768,36,'cp 138','c1f4 f8e7 e1c1 e8g8 f2f3 a7a6 g2g4 b7b5 c1b1 c8e6 c3d5 f8e8 d5e7 d8e7 f4c1 a8d8 d4f2 d6d5 g4g5 f6h5 f1e2 g7g6 h1e1 d8d7 b2b3 c7c5 f3f4 d5e4 d1d7 e7d7 e1d1 d7c8 e2h5 g6h5 c1b2');
INSERT INTO evaluations VALUES(15774,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(15774,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(15774,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(15774,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(15774,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(15774,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(15774,36,'cp 122','f8e7 f2f3 c8e6 e1c1 f6d7 d4g7 e7f6 g7h6 f6c3 b2c3 d8e7 h6g7 e8c8 h2h4 d7e5 f4g5 e7f8 g7f6 d8e8 g5h6 f8e7 f6e7 e8e7 h6g7 h8g8 g7e5 d6e5 g2g4 f7f6 a2a4 a7a5 f1e2 h7h5 g4h5 g8h8 c1b2 c7c6 h5h6 h8h6 h4h5');
INSERT INTO evaluations VALUES(8699,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(8699,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(8699,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(8699,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(8699,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(8699,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(8699,36,'cp 112','e1c1 e8g8 f2f3 c8e6 c1b1 f6d7 d4d2 e7f6 h2h4 f6e5 f4g5 f7f6 g5e3 c7c6 f3f4 e5c3 d2c3 d8e7 h4h5 e6g4 d1e1 h7h6 c3b3 g8h8 b3b7 e7e4 f1d3 e4d5 b2b3 f8c8 b7a6 g4f5 d3f5 d5f5 a6e2 d7c5 g2g4 f5d5 e3c5 d5c5 g4g5 f6g5');
INSERT INTO evaluations VALUES(8698,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(8698,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(8698,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(8698,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(8698,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(8698,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(8698,36,'cp 118','e8g8 f2f3 a7a6 c1b1 b7b5 g2g4 c8e6 c3d5 f6d5 e4d5 c7c5 d5c6 d8c7 f1d3 c7c6 d3e4 d6d5 e4d5 e6d5 d4d5 c6b6 f4c1 a8c8 d5f5 e7h4 g4g5 b6g6 f5g6 f7g6 f3f4 f8d8 c2c3 h7h5 h1f1 c8b8 d1d8 b8d8 b1c2 d8e8 c2d3 e8d8 d3e4');
INSERT INTO evaluations VALUES(3447,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(3447,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(3447,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(3447,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(3447,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(3447,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(3447,36,'cp 127','c1b1 c8e6 f2f3 f6d7 h2h4 e7f6 d4d2 f6e5 f4g5 f7f6 g5e3 c7c6 f3f4 e5c3 d2c3 d8e7 f1e2 f6f5 c3a3 f5e4 h4h5 a8e8 d1d6 f8f6 h1d1 d7b6 h5h6 b6d5 e2c4 d5e3 d6e6 f6e6 a3e3 g8h8 h6g7 h8g7 c4e6 e7e6 e3a7 g7g8 a2a3 e4e3 a7b7');
INSERT INTO evaluations VALUES(21466,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(21466,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(21466,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(21466,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(21466,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(21466,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(21466,36,'cp 65','g1f3 f7f6 f1b5 c8d7 b1c3 a7a6 b5c6 d7c6 c3e2 f6e5 f3e5 f8d6 e2f4 d8f6 f4h5 f6e7 c1f4 g8f6 e1g1 e8g8 h5f6 g7f6 e5c6 b7c6 f4d6 e7d6 a1c1 e6e5 c2c4 a8e8 c4d5 c6d5 c1c3 f8f7 d1d2 e5d4 d2d4 d6e5 d4a7 e5e2 c3g3 g8h8 a7d4 e2c4 d4c4 d5c4');
INSERT INTO evaluations VALUES(21464,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(21464,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(21464,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(21464,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(21464,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(21464,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(21464,36,'cp 83','f7f6 f1b5 c8d7 b1c3 d8e7 e1g1 e7f7 b5c6 d7c6 a2a4 g8e7 b2b4 a7a6 a1b1 e7g6 b4b5 c6d7 h2h4 h7h5 b5a6 b7a6 c3e2 d7a4 e2f4 f8e7 f1e1 e8g8 e5f6 e7f6 f4e6 f8e8 f3g5 f6g5 e6g5 e8e1 d1e1 f7f5 e1e2 a4c2 b1b7 g6h4');
INSERT INTO evaluations VALUES(22238,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(22238,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(22238,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(22238,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(22238,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(22238,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(22238,36,'cp 90','c2c3 f6e5 f3e5 c6e5 d4e5 g8e7 f1d3 c7c5 e1g1 d8c7 c1g5 c5c4 d3c2 e7c6 b1d2 c6e5 f1e1 f8d6 g5f4 e5f7 c2a4 e8e7 f4d6 c7d6 b2b3 h8d8 d1h5 d6f4 d2f3 f4h6 h5g4 e6e5 g4g3 h6g6 g3g6 h7g6 f3e5 f7e5 e1e5 e7f6 a1e1 c8g4 b3c4 d5c4 e5e7 d8d2 e7b7');
INSERT INTO evaluations VALUES(22235,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(22235,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(22235,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(22235,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(22235,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(22235,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(22235,36,'cp 92','c8d7 b2b4 a7a6 f1d3 f6e5 d4e5 g8h6 d1e2 f8e7 c1h6 g7h6 b1d2 e8g8 e1g1 d8e8 a2a4 e8f7 g1h1 f7g7 b4b5 c6a5 b5a6 b7a6 d3a6 g8h8 a6b5 f8g8 f3e1 c7c6 b5d3 h6h5 f2f4 h5h4 e1f3 c6c5 f3g5 e7g5 f4g5 a5c6 d2f3 c5c4 d3c2 d7e8 e2e3 e8h5 f3h4 g7g5');
INSERT INTO evaluations VALUES(53061,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(53061,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53061,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53061,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(53061,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53061,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53061,36,'cp 0','g8f6 f1e2 e7e5 e2b5 b8c6 e1g1 d8b6 b1a3 a7a6 b5c4 f8e7 f1e1 e8g8 a3c2 c6a5 c4f1 a5c6');
INSERT INTO evaluations VALUES(45197,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45197,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45197,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45197,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45197,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45197,36,'cp 0','f1e2 e7e5 e2b5 b8c6 e1g1 d8b6 b1a3 a7a6 b5c4 f8e7 f1e1 e8g8 a3c2 c6a5 c4f1 a5c6');
INSERT INTO evaluations VALUES(45198,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45198,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45198,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45198,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45198,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45198,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45198,36,'cp 0','e7e5 e2b5 b8c6 e1g1 d8b6 b1a3 a7a6 b5c4 f8e7 f1e1 e8g8 a3c2 c6a5 c4f1 a5c6');
INSERT INTO evaluations VALUES(19447,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(19447,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(19447,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(19447,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(19447,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(19447,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(19447,36,'cp 20','d2d4 d7d5 d4c5 d5e4 d1d8 e8d8 c1e3 e7e5 f2f3 f7f5 f3e4 f5f4 e3f2 g8f6 b1d2 f6g4 g1h3 g4f2 h3f2 f8c5 f1b5 c5f2 e1f2 d8c7 d2f3 h8e8 g2g3 h7h6 h2h4 a7a6 b5c4 c8e6 c4b3 e6b3 a2b3 a8d8 g3f4 e5f4 h1g1 e8e4');
INSERT INTO evaluations VALUES(19426,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(19426,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19426,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(19426,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(19426,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(19426,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(19426,36,'cp 20','d7d5 e4d5 d8d5 g1f3 e7e6 b1a3 g8f6 c1e3 c5d4 a3b5 d5d7 b5d4 d7c7 d1d2 f6d5 e1c1 f8e7 e3g5 c6d4 f3d4 a7a6 g5e7 d5e7 f1d3 e8g8 h1e1 b7b5 d2e3 e7d5 e3g3 c8b7 g3c7 d5c7 d3e4 b7e4 e1e4 c7d5 d4c6 a8c8 c6b4 d5b4 e4b4 c8a8');
INSERT INTO evaluations VALUES(19793,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(19793,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19793,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(19793,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19793,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(19793,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19793,36,'cp 22','e4d5 d8d5 g1f3 e7e6 f1e2 g8f6 e1g1 c5d4 c3d4 f8e7 b1c3 d5d6 c3b5 d6d8 c1f4 f6d5 f4g3 e8g8 a1c1 c8d7 b5c3 a8c8 c3d5 e6d5 d1b3 d8b6 b3b6 a7b6 g3f4 f8d8 c1c3 d7f5 f1c1 e7f6 f4e3 f5e4 a2a3 c8a8 c3b3 e4f3');
INSERT INTO evaluations VALUES(50310,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(50310,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(50310,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(50310,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(50310,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(50310,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(50310,36,'cp -14','d7d5 e2e3 c7c5 g5f6 g7f6 d4c5 e7e6 g1f3 f8c5 f1e2 b8c6 e1g1 e8g8 c2c4 d5d4 e3d4 c6d4 b1c3 f6f5 f3d4 c5d4 d1d3 d8f6 a1d1 d4c3 b2c3 e6e5 d3f3 a8b8 f3e3 c8e6 f2f4 e5e4 c4c5 g8g7 a2a3 b8d8 d1d8 f6d8');
INSERT INTO evaluations VALUES(47122,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(47122,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(47122,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47122,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47122,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47122,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47122,36,'cp 0','e2e3 c7c5 g5f6 g7f6 d4c5 e7e6 g1f3 f8c5 f1e2 b8c6 e1g1 e8g8 c2c4 d5d4 e3d4 c6d4 b1c3 b7b6 c3e4 e6e5 e4g3 c8e6 e2d3 e6g4 d3e2');
INSERT INTO evaluations VALUES(47118,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47118,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47118,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47118,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47118,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47118,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47118,36,'cp 0','c7c5 g5f6 g7f6 d4c5 e7e6 g1f3 f8c5 f1e2 b8c6 e1g1 e8g8 c2c4 d5d4 e3d4 c6d4 b1c3 f6f5 f3d4 c5d4 d1d3 d4f6 d3g3 g8h8 a1d1 c8d7 g3f4 d8e7 c3e4 f5e4 d1d7 e7d7 f4f6 h8g8 f1d1 d7a4 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(55108,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55108,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55108,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(55108,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55108,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55108,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55108,36,'cp -5','c1b2 c8f5 g1f3 e7e6 g2g3 g8f6 d2d3 h7h6 f1g2 f8e7 b1d2 e8g8 f3e5 f5h7 e1g1 c7c6 e2e4 a7a5 a2a4 e7b4 g3g4 b8d7 e5d7 f6d7 g1h1 d7f6 f2f4 d5e4 d2e4 h7e4 g2e4 f6e4 d3e4 d8d1 f1d1 f8d8 f4f5 e6f5 g4f5');
INSERT INTO evaluations VALUES(55109,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55109,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(55109,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55109,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55109,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55109,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(55109,36,'cp -11','c8f5 e2e3 e7e6 g1f3 c7c6 d2d4 g8f6 f1d3 f8d6 e1g1 e8g8 d3f5 e6f5 b2a3 d6a3 b1a3 f8e8 c2c4 b8d7 d1c2 f6e4 a1c1 a8c8 c2b2 d7f6 a3b1 d8e7 b1c3 e4c3 b2c3 g7g6 a2a4 c8c7 f3e5 g8g7 c3b2 e8c8 e5d3 h7h5');
INSERT INTO evaluations VALUES(31879,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31879,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(31879,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31879,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31879,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(31879,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(31879,36,'cp 0','e2e3 e7e6 g1f3 c7c6 d2d4 f8d6 f1d3 g8f6 e1g1 e8g8 d3f5 e6f5 b2a3 f8e8 a3d6 d8d6 d1c1 f6e4 c2c4 b8d7 b1c3 e4c3 c1c3 d7f6 f3e5 f6g4 e5f3');
INSERT INTO evaluations VALUES(31876,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31876,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31876,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31876,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31876,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31876,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31876,36,'cp 0','e7e6 g1f3 c7c6 d2d4 g8f6 f1d3 f8b4 c2c3 b4d6 d3f5 e6f5 e1g1 d8e7 d1c1 h7h5 c3c4 a7a5 b2a3 d6b4 a3b2');
INSERT INTO evaluations VALUES(32329,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32329,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32329,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32329,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32329,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32329,36,'cp 0','g1f3 c7c6 d2d4 g8f6 f1d3 f8b4 c2c3 b4d6 d3f5 e6f5 e1g1 d8e7 d1c1 h7h5 c3c4 a7a5 b2a3 d6b4 a3b2');
INSERT INTO evaluations VALUES(32330,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32330,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32330,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32330,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32330,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32330,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32330,36,'cp 0','c7c6 d2d4 f8d6 f1d3 g8f6 d3f5 e6f5 e1g1 e8g8 b2a3 d6a3 b1a3 d8d6 d1c1 b8d7 a3b1 f6e4 c2c4 f8e8 b1c3 e4c3 c1c3 d7f6 f3e5 f6g4 e5f3 g4f6');
INSERT INTO evaluations VALUES(53573,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(53573,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(53573,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(53573,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(53573,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53573,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(53573,36,'cp 44','e2e4 d5e4 c3e4 f8b4 c1d2 d8d4 d2b4 d4e4 f1e2 b8a6 b4a5 c8d7 g1f3 g8e7 d1d6 b7b6 e1c1 a8d8 e2d3 e4g4 f3e5 e7c8 d6d7 d8d7 e5g4 b6a5 g4e5 d7c7 d3e4 a6b8 c4c5 f7f6 e5c4 b8a6 c4a5 a6c5 e4c6 e8e7 c1b1 c8b6 b2b4 c5a4 d1c1 h8c8 h1e1 e7f7');
INSERT INTO evaluations VALUES(53585,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(53585,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(53585,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(53585,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(53585,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53585,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(53585,36,'cp 35','d5e4 c3e4 c6c5 e4c5 b8c6 c1e3 e6e5 d4e5 d8a5 d1d2 c6b4 c5e4 c8f5 a1d1 b4c2 e1e2 a5e5 d2c2 f5e4 c2c3 e5c3 b2c3 g8h6 g1f3 h6f5 f3d4 f5e3 e2e3 e4c6 d4c6 b7c6 e3d2 e8c8 f1d3 f8c5 f2f3 c8c7 h2h4 h8e8 d2c2 h7h6 h1e1 c5d6 d3e4 d6c5 d1d8 e8d8');
INSERT INTO evaluations VALUES(53653,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(53653,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53653,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53653,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(53653,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(53653,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(53653,36,'cp 41','c3e4 c6c5 e4c5 b8c6 c5b3 a7a5 a2a4 f8b4 c1d2 c6d4 d2b4 d4b3 d1b3 d8b6 g1f3 a5b4 f1e2 g8f6 f3e5 e8g8 e5d3 c8d7 b3b4 b6b4 d3b4 d7a4 f2f3 f8d8 e1f2 g8f8 f2e3 a4c6 b4c6 b7c6 b2b4 d8b8 a1a8 b8a8 f3f4 a8a4 h1b1 f8e7 e2f3 e7d6');
INSERT INTO evaluations VALUES(53652,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(53652,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(53652,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(53652,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(53652,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53652,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(53652,36,'cp 39','c6c5 e4c5 b8c6 c5b3 a7a5 a2a4 f8b4 c1d2 c6d4 d2b4 d4b3 d1b3 d8b6 g1f3 a5b4 f1e2 g8f6 f3e5 e8g8 e5d3 c8d7 b3b4 b6b4 d3b4 d7a4 e1d2 f6e4 d2e3 e4d6 e2d3 f8d8 h1b1 g8f8 b2b3 a4e8 a1a8 d8a8 f2f3 h7h6 e3d4 b7b6 b1e1 a8a5 d4c3 g7g5 c3b2 a5a8 g2g3 e8d7 d3e4 d6e4 e1e4');
INSERT INTO evaluations VALUES(31859,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(31859,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(31859,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(31859,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(31859,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(31859,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(31859,36,'cp 31','c2c4 e7e6 b1c3 g8f6 f3h4 f5g6 d1b3 b8c6 e2e3 f8b4 c4d5 e6d5 c1d2 e8g8 h4g6 h7g6 f1e2 a7a5 a2a3 b4c3 d2c3 d8d7 b3b5 f8b8 e1g1 d7e6 f1c1 a5a4 c3e1 f6e4 c1c2 e4d6 b5d3 d6c4 c2c4 d5c4 d3c4 e6c4 e2c4');
INSERT INTO evaluations VALUES(31826,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31826,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31826,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31826,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31826,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(31826,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31826,36,'cp 27','e7e6 b1c3 g8f6 f3h4 f5g6 h4g6 h7g6 d1b3 b8c6 e2e3 f8b4 c1d2 d8d7 e1c1 b4c3 d2c3 a7a5 c1b1 a5a4 b3c2 e8g8 a2a3 c6a5 c3a5 a8a5 h2h4 f8b8 f1d3 d5c4 c2c4 d7e7 d1c1 c7c6 e3e4 e6e5 c4b4 e7c7 b1a1 e5d4 b4d4');
INSERT INTO evaluations VALUES(32284,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32284,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(32284,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(32284,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(32284,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(32284,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(32284,36,'cp 28','d1b3 b8c6 a2a3 a7a6 e2e3 g8f6 c1d2 a8b8 d2c3 f8d6 b1d2 e8g8 c4c5 d6e7 f1e2 b7b5 c5b6 b8b6 b3a4 a6a5 e1g1 d8a8 f3h4 f5g4 f2f3 g4h5 g2g4 f6d7 h4g2 h5g6 f3f4 h7h6 f4f5 e6f5 g4f5 g6h7 g2f4 d7f6 e2d3');
INSERT INTO evaluations VALUES(32277,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32277,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(32277,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(32277,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(32277,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(32277,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(32277,36,'cp 24','b8c6 a2a3 a7a6 e2e3 g8f6 b1c3 d5c4 f1c4 f8d6 f3h4 e8g8 h4f5 e6f5 c4e2 a8b8 e1g1 c6e7 c1d2 c7c6 e2f3 f8e8 f1e1 e7d5 g2g3 d5c3 d2c3 f6e4 e1d1 g7g6 g1g2 h7h5 h2h4 d8c7 c3e1 g8g7 a1c1 b8d8');
INSERT INTO evaluations VALUES(57152,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(57152,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(57152,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(57152,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(57152,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(57152,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(57152,36,'cp -29','d7d5 d2d3 g8f6 g1f3 d5e4 f3e5 b8d7 e5d7 d8d7 c1g5 d7c6 d3e4 c6e4 d1e2 f8e7 e2e4 f6e4 g5e7 e8e7 b1d2 e4d2 e1d2 h8d8 f1d3 c8f5 a1e1 e7f6 e1e3 d8d6 e3f3 g7g6 h1e1 a8d8 e1e3 h7h5 d2c2 d6d3 e3d3 d8d3 f3d3 h5h4 b2b4 f6e5 g2g3 f5d3 c2d3 h4g3 h2g3 b7b5 c3c4 b5c4 d3c4');
INSERT INTO evaluations VALUES(56389,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(56389,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(56389,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(56389,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(56389,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(56389,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(56389,36,'cp -23','d2d3 g8f6 g1f3 f8d6 c1g5 c8e6 b1d2 b8d7 d3d4 h7h6 e4d5 e6d5 g5f6 d8f6 f1c4 d5f3 d1f3 e5d4 c3d4 e8g8 f3f6 d7f6 e1g1 a8d8 d2f3 c7c6 f1e1 f8e8 e1e8 f6e8 g2g3 d6c7 g1g2 g8f8 a1d1 g7g6 c4b3 e8d6 h2h4 f8g7');
INSERT INTO evaluations VALUES(56390,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(56390,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(56390,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(56390,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(56390,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56390,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(56390,36,'cp -32','d5e4 d3e4 d8d1 e1d1 f7f5 g1f3 g8f6 f3e5 b8d7 e5d7 c8d7 c1e3 f6g4 b1d2 g4e3 f2e3 e8c8 d1c2 f8c5 a1e1 f5e4 d2e4 d7f5 f1d3 d8d3 e4c5 d3e3 c2d2 e3e1 h1e1 h8d8 d2c1 b7b6 c5b3 f5g6 b3d4 d8f8 d4f3 g6f7 b2b3 f7d5 e1e7 d5f3 g2f3');
INSERT INTO evaluations VALUES(47223,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47223,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(47223,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(47223,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(47223,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(47223,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(47223,36,'cp 27','e4e5 f6d7 d2d4 c7c5 f2f4 e7e6 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 c6d4 e3d4 c5d4 d2d4 d8b6 c3e2 d7c5 e1c1 c8d7 e2c3 e8c8 h2h4 h7h5 f1e2 c8b8 b2b4 c5a4 c3a4 b6d4 d1d4 d7a4 c1d2 a4b5 e2b5 a6b5');
INSERT INTO evaluations VALUES(47293,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(47293,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(47293,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(47293,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47293,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(47293,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47293,36,'cp 32','f6d7 d2d4 c7c5 f2f4 b8c6 g1f3 e7e6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 e8g8 d2f2 c6d4 e3d4 c5d4 f2d4 d7b8 f1d3 b8c6 d4f2 b7b5 c3e2 f7f6 f2h4 g7g6 e5f6 d8f6 h4f6 f8f6 h1f1 e6e5 f4e5 f6f1 e1f1 c6e5 a1e1 e5d3 c2d3');
INSERT INTO evaluations VALUES(49181,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(49181,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(49181,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49181,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(49181,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(49181,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(49181,36,'cp 28','d2d4 e7e6 f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 a8b8 d2f2 c6d4 e3d4 c5d4 f2d4 d8b6 d4b6 d7b6 e1c1 e8e7 d1d4 b6d7 f1d3 h7h5 h1e1 b7b5 f4f5 c8b7 c1d2 b8c8 e1e3 d7c5 g2g3 c5d3 e3d3 e6f5');
INSERT INTO evaluations VALUES(49176,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(49176,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(49176,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(49176,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49176,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(49176,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49176,36,'cp 37','c7c5 f2f4 b8c6 g1f3 e7e6 c1e3 a7a6 d1d2 d8a5 f1e2 b7b5 e1g1 b5b4 c3d1 f8e7 c2c3 b4c3 d1c3 e8g8 a2a3 c5d4 f3d4 c6d4 e3d4 e7c5 b2b4 c5d4 d2d4 a5b6 d4b6 d7b6 a3a4 d5d4 c3e4 b6d5 a1d1 a8b8 d1d4 b8b4 d4b4 d5b4');
INSERT INTO evaluations VALUES(44022,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(44022,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(44022,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(44022,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44022,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(44022,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44022,36,'cp 39','f2f4 e7e6 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 a8b8 d2f2 c6d4 e3d4 c5d4 f2d4 d8b6 d4b6 d7b6 e1d2 c8d7 f1d3 e8e7 h1f1 d7c6 c3e2 b6c4 d3c4 d5c4 g2g3 h7h5 e2d4 c6g2 f1f2 g2d5 d2c3 g7g6 a3a4 b8c8');
INSERT INTO evaluations VALUES(16137,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(16137,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(16137,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(16137,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(16137,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(16137,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(16137,36,'cp 24','c2c3 f8c5 b2b4 c5b6 a2a4 a7a6 e1g1 e8g8 b1d2 d7d6 f1e1 c8e6 h2h3 d8d7 c4b3 e6b3 d2b3 f8e8 d1c2 d6d5 a4a5 b6a7 c1e3 a7e3 e1e3 b7b6 e3e1 d7d6 f3d2 d6d7 e4d5 f6d5 d2c4 g8h8 g1h1 f7f6 h1g1 h8g8 f2f3 b6a5 c4a5 c6a5');
INSERT INTO evaluations VALUES(16133,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(16133,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(16133,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(16133,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(16133,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(16133,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(16133,36,'cp 13','f8c5 b2b4 c5b6 a2a4 a7a6 e1g1 d7d6 b1d2 e8g8 h2h3 c8e6 c4e6 f7e6 d2c4 b6a7 c1e3 a7e3 f2e3 c6e7 d1c2 e7g6 d3d4 f6d7 a4a5 d8e7 a1d1 a8c8 d4e5 g6e5 c4e5 d7e5 f3e5 d6e5 c2e2 f8f1 d1f1 c8f8 f1d1 g8h7 g1h2 f8f7 d1d2 e7h4 e2c4');
INSERT INTO evaluations VALUES(9768,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(9768,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(9768,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(9768,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(9768,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(9768,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(9768,36,'cp 23','b2b4 c5b6 a2a4 a7a6 b1d2 e8g8 e1g1 d7d6 f1e1 c8e6 c4b3 f8e8 b3e6 e8e6 d2c4 d6d5 e4d5 d8d5 c1e3 e5e4 c4b6 c7b6 d3e4 d5d1 a1d1 f6e4 f3d4 e6e8 d4c6 b7c6 d1c1 a6a5 b4a5 b6a5 h2h3 e4d6 c3c4 f7f6 g2g3 d6e4 g1g2');
INSERT INTO evaluations VALUES(9738,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(9738,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(9738,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(9738,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(9738,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(9738,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9738,36,'cp 24','c5b6 a2a4 a7a6 e1g1 d7d6 b1d2 b6a7 f1e1 c8e6 c4b3 f6g4 e1f1 e8g8 h2h3 g4f6 f1e1 e6b3 d2b3 c6e7 d1c2 f8e8 c1e3 a7e3 e1e3 e7g6 a4a5 b7b6 e3e1 d8d7 g1h2 d6d5 c3c4 d7c6 g2g3 e8d8 h2g2 b6b5 b3c5 b5c4 d3c4 d5e4 c5e4');
INSERT INTO evaluations VALUES(9708,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(9708,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(9708,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(9708,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(9708,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(9708,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(9708,36,'cp 18','a2a4 a7a6 a4a5 b6a7 e1g1 d7d6 d1b3 e8g8 c1e3 a7e3 f2e3 c6e7 b1d2 e7g6 d3d4 b7b5 a5b6 c7b6 b4b5 a6a5 c4d3 f8e8 c3c4 e8e7 h2h3 c8b7 b3c2 e5d4 e3d4 f6h5 g1h2 h5f4 c2c3 f4d3 c3d3 e7e8 a1e1 a5a4');
INSERT INTO evaluations VALUES(9716,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(9716,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9716,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(9716,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(9716,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(9716,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9716,36,'cp 22','a7a6 e1g1 d7d6 f1e1 c8e6 b1d2 e8g8 h2h3 b6a7 c4b3 f8e8 b3e6 e8e6 d1c2 d8d7 d2f1 d6d5 c1e3 d5d4 c3d4 c6d4 e3d4 a7d4 f3d4 d7d4 c2c7 d4b4 e1b1 e6e7 c7c2 b4d6 f1e3 a8d8 e3f5 e7c7 f5d6 c7c2');
INSERT INTO evaluations VALUES(7278,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(7278,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(7278,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(7278,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(7278,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(7278,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(7278,36,'cp 29','e1g1 d7d6 b1d2 b6a7 h2h3 e8g8 f1e1 c8e6 c4b3 f8e8 b3e6 e8e6 d1c2 d8d7 d2f1 d6d5 c1d2 a8d8 a1d1 d5e4 d3e4 e6d6 b4b5 a6b5 a4b5 c6e7 c2a4 a7c5 f3e5 d7e6 e5f3 d6d3 e4e5 f6d7 a4g4 e7g6 g4e6 f7e6');
INSERT INTO evaluations VALUES(14510,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14510,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(14510,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14510,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(14510,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(14510,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(14510,36,'cp 65','e8f7 d1f3 f7e6 b1c3 c6b4 e1g1 c7c6 d2d4 d8f6 f3d1 e6e7 f1e1 f6g6 a2a3 d5c3 b2c3 b4d5 e1e5 e7d8 c4d3 g6f7 c3c4 d5f4 c4c5 c8d7 c1f4 f7f4 c2c3 f8e7 d1b3 a8b8 g2g3 f4f3 e5e3 f3f6 a1e1 h8e8 b3a4 b8a8 d3h7 f6f7 a4a5 d8c8 h7g8 f7g8 e3e7 e8e7 e1e7 g8d5 e7g7');
INSERT INTO evaluations VALUES(2428,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(2428,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(2428,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(2428,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(2428,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(2428,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(2428,36,'cp 66','d1f3 f7e6 b1c3 c6b4 e1g1 c7c6 d2d4 d8f6 f3e2 e6e7 c3e4 f6g6 a2a3 c8f5 f2f3 b4a6 d4e5 e7e8 c4d5 c6d5 e4d6 e8d8 c1f4 a8c8 d6b7 d8e7 e5e6 g6e6 f4d6 e7f6 e2a6 f8d6 a6d6 c8c2 d6e6 f5e6 b2b4 d5d4 b7c5 e6f5 g2g4 f5g6 a1e1 h8e8 c5d7 f6f7 d7e5 f7f8 e5g6 h7g6 e1e8 f8e8 f1d1 c2c3 g1f2 c3a3 d1d4');
INSERT INTO evaluations VALUES(2427,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(2427,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(2427,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(2427,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(2427,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(2427,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(2427,36,'cp 60','f7e6 b1c3 c6b4 e1g1 c7c6 d2d4 d8f6 f3d1 e6e7 f1e1 f6g6 e1e5 e7d8 a2a3 d5c3 b2c3 b4d5 c4d3 g6f7 c3c4 d5f4 c4c5 c8d7 c1f4 f7f4 d1e1 g7g6 e5e4 f4f6 e1b4 a8c8 a1e1 c8c7 b4a5 a7a6 e1b1 d8c8 a5b6 c8b8 d4d5 f8c5 b6c5 c6d5 c5d5 b8a7 d5d4 f6d4');
INSERT INTO evaluations VALUES(2444,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(2444,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(2444,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(2444,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(2444,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(2444,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(2444,36,'cp 78','b1c3 c6b4 e1g1 c7c6 d2d4 d8f6 f3e2 e6e7 c3e4 f6g6 a2a3 c8f5 f2f3 b4a6 d4e5 a6c7 g1h1 b7b5 c4d3 h7h5 c2c4 b5c4 d3c4 e7d8 b2b4 a7a6 c1d2 g6e6 e4g5 e6g6 a1c1 d8c8 c4d5 c7d5 b4b5 c6c5 g5e4 g6b6 e4c5 f8c5 d2b4 c8b7 b4c5 b6b5 e2d2 a8d8 a3a4 b5b3');
INSERT INTO evaluations VALUES(2443,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(2443,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(2443,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(2443,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(2443,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(2443,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(2443,36,'cp 92','c6b4 e1g1 c7c6 d2d4 d8f6 f3d1 e6e7 f1e1 f6g6 a2a3 d5c3 b2c3 b4d5 e1e5 e7d8 c4d3 g6f7 c3c4 d5f4 c4c5 c8d7 d3f5 f4g6 c1g5 d8c7 f5g6 h7g6 d1d2 b7b6 a3a4 a8e8 c5b6 a7b6 a4a5 b6b5 a5a6 c7b8 d2a5 b8a8 a5c7 e8b8 a1e1 f8b4 e5e7 b4e7 e1e7 h8e8 c7d7 e8e7 d7c6 a8a7 c6c5 a7a6 g5e7 f7c4');
INSERT INTO evaluations VALUES(2456,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(2456,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(2456,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(2456,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(2456,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(2456,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(2456,36,'cp 119','e1g1 c7c6 d2d4 d8f6 f3d1 e6e7 f1e1 h7h6 e1e5 e7d8 c3e4 f6g6 a2a3 c8f5 e4g3 f5c2 d1f3 c2d3 c4b3 b4c2 b3d5 f8d6 d5f7 c2d4 f3e3 g6f7 e3d4 d8c7 e5e3 d3g6 e3f3 f7e7 c1d2 h8e8 d4a4 e7d7 a1d1 c6c5 a4c4 a8d8 d2f4 d7c6 b2b4 d6f4');
INSERT INTO evaluations VALUES(2457,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(2457,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(2457,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(2457,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(2457,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(2457,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(2457,36,'cp 96','c7c6 d2d4 d8f6 f3d1 e6e7 f1e1 f6g6 a2a3 d5c3 e1e5 e7d8 b2c3 b4d5 c4d3 g6f7 c3c4 d5f4 c4c5 g7g5 c1f4 f7f4 e5e4 f4f6 d1d2 c8d7 a1e1 f8g7 e4e7 a7a5 c2c3 h8f8 d2b2 f6e7 b2b6 d8e8 e1e7 e8e7 h2h3 a8d8 d3h7 f8f7 h7d3 g7f6 b6b7 e7f8 d3e4 f7e7 e4c6');
INSERT INTO evaluations VALUES(2401,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(2401,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(2401,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(2401,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(2401,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(2401,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(2401,36,'cp 116','d2d4 d8f6 f3d1 e6e7 f1e1 h7h6 e1e5 e7d8 c3e4 f6g6 a2a3 c8f5 e4g3 f5c2 d1f3 b4d3 e5f5 d8c7 c4d3 c2d3 f3d3 a8e8 c1d2 f8d6 d3f3 h8f8 a1e1 e8e1 d2e1 f8f5 f3f5 g6f5 g3f5 g7g6 f5d6 c7d6 g1f1 d6e6 e1d2 h6h5 a3a4 e6f5 f2f3 g6g5 a4a5 g5g4 f1e2 b7b6');
INSERT INTO evaluations VALUES(2403,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(2403,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(2403,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(2403,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(2403,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(2403,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(2403,36,'cp 102','d8f6 f3d1 e6e7 f1e1 h7h6 e1e5 e7d8 c3e4 f6g6 a2a3 c8f5 e4g3 f5c2 d1f3 b4d3 e5f5 f8d6 c4d3 c2d3 f3d3 d8c7 c1d2 a8e8 a1d1 a7a6 h2h3 c7b8 f5f3 g6d3 f3d3 h8f8 d1e1 e8e1 d2e1 g7g5 g3e4 f8e8 f2f3 b8c7 g1f2 b7b6 g2g3 a6a5 d3d1 c7d7 e1d2 a5a4 d1e1 e8b8 e4d6 d7d6');
INSERT INTO evaluations VALUES(6416,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(6416,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(6416,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(6416,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(6416,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(6416,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(6416,36,'cp -20','c5b4 c2c3 b4a5 d2d4 d7d6 d1b3 d8d7 b1d2 e5d4 c3d4 g8f6 e1g1 e8g8 e4e5 a5d2 e5f6 c6a5 b3c2 a5c4 c1d2 c4d2 c2d2 d7f5 f6g7 f8e8 d4d5 b7b6 f1c1 e8e7 c1e1 e7e1 a1e1 c8d7 h2h3 a8c8 f3d4 f5d5 d2f4 c7c5 e1e7 c5d4 e7d7 c8e8');
INSERT INTO evaluations VALUES(6618,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6618,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6618,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6618,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6618,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6618,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(6618,36,'cp -10','c2c3 b4a5 d2d4 d7d6 d1b3 d8d7 e1g1 a5b6 b1d2 c6a5 b3b1 g8e7 d4e5 e8g8 e5d6 c7d6 a2a4 e7g6 c4a2 d7c6 c1a3 g6f4 a3b4 f4e2 g1h1 e2c3 b4c3 c6c3 e4e5 b6d4 f3d4 c3d2 d4f3 d2h6 e5d6 h6d6 f3g5 d6g6 b1b5 b7b6 a2d5');
INSERT INTO evaluations VALUES(6616,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(6616,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(6616,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(6616,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(6616,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(6616,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(6616,36,'cp -37','b4a5 d2d4 d7d6 d1b3 d8d7 b1d2 a5b6 e1g1 c6a5 b3c2 f7f6 a2a4 g8e7 c4a2 d7c6 c2d3 c8g4 c1a3 e7g6 g2g3 e5d4 c3d4 c6a4 a3b4 a4b4 f1b1 g6e5 d3e3 a5c4 d2c4 e5f3 g1g2 b6d4 b1b4 d4e3 c4e3 f3e5 e3g4 e5g4 b4b7 e8d7 a1c1 h8c8 a2b3 a7a5 b3a4 d7d8');
INSERT INTO evaluations VALUES(6728,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(6728,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(6728,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(6728,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(6728,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(6728,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(6728,36,'cp -5','d2d4 d7d6 d1b3 d8d7 b1d2 e5d4 c3d4 g8f6 e1g1 e8g8 e4e5 a5d2 e5f6 c6a5 b3c2 a5c4 c1d2 c4d2 c2d2 d7f5 f6g7 f8e8 d4d5 f5h5 a1c1 c8g4 f3g5 h7h6 f2f3 h5g5 d2g5 h6g5 f3g4 a8c8 f1f5 e8e5 f5e5 d6e5 d5d6 c7c6 c1c5 c8d8 c5e5 d8d6 e5g5 d6g6 g5g6 f7g6 g1f2 g8g7 f2f3 g7h7 f3e4 c6c5 g4g5 b7b5');
INSERT INTO evaluations VALUES(6731,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(6731,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(6731,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(6731,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(6731,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(6731,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(6731,36,'cp -32','d7d6 d1b3 d8d7 b1d2 a5b6 e1g1 c6a5 b3d1 g8e7 d4e5 e7g6 e5d6 d7d6 c4d5 e8g8 d2b3 g6f4 c3c4 f4d5 c4d5 c7c6 b3a5 b6a5 d1a4 d6b4 a4b4 a5b4 d5c6 b7c6 c1e3 f8e8 f1c1 a7a5 a2a3 b4f8 e3c5 f8c5 c1c5 e8e4 c5c6 h7h6 a1c1 c8e6 c6c7');
INSERT INTO evaluations VALUES(6080,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(6080,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(6080,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(6080,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(6080,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(6080,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(6080,36,'cp -25','d1b3 d8d7 e1g1 a5b6 c4b5 a7a6 b5a4 b6a7 b3b2 b7b5 a4c2 g8f6 a2a4 a8b8 a4b5 b8b5 b2a2 e5d4 c2a4 e8g8 a4b5 a6b5 c3d4 a7d4 f3d4 c6d4 f2f3 b5b4 c1e3 d4e6 f1d1 d7b5 b1d2 d6d5 a2a5 c8d7 d2b3 d5e4 f3e4 b5c4 b3d4 c4c3 e3f2');
INSERT INTO evaluations VALUES(55040,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55040,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55040,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(55040,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(55040,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(55040,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(55040,36,'cp -12','d5d4 b2b4 c7c5 b4b5 g7g6 g2g3 b7b6 f1g2 c8b7 a2a4 f8g7 e1g1 b8d7 c1b2 g8h6 e2e3 h6f5 e3d4 a8b8 a1a2 b7f3 d1f3 f5d4 f3d1 e8g8 b2c3 e7e6 a4a5 h7h5 g2e4 h5h4 d2d3 h4h3 a5b6 a7b6 a2a7 d7f6 b1d2 f6e4 d2e4');
INSERT INTO evaluations VALUES(55207,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55207,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55207,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(55207,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(55207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55207,36,'cp 0','b2b4 c7c5 b4b5 g7g6 g2g3 a7a6 f1g2 g8h6 a2a4 f8g7 d2d3 h6f5 b1a3 b8d7 f3d2 a8a7 d2e4 e8g8 c1d2 a6a5 b5b6 a7a6 a1b1 a6b6 a3b5 b6e6 b5a7 b7b6 e4g5 e6f6 g5e4');
INSERT INTO evaluations VALUES(55184,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55184,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55184,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55184,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55184,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55184,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55184,36,'cp 0','c7c5 b4b5 a7a6 g2g3 g7g6 f1g2 g8h6 d2d3 h6f5 a2a4 f8g7 b1a3 e8g8 f3d2 a8a7 d2e4 b8d7 c1d2 a6a5 b5b6 a7a6 a1b1 a6b6 a3b5 b6e6 b5a7 b7b6 e4g5 e6f6 g5e4');
INSERT INTO evaluations VALUES(6603,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(6603,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(6603,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(6603,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(6603,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(6603,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(6603,36,'cp 73','c3d5 g8f6 d5b4 c6b4 d2d4 d7d6 c2c3 b4c6 f1d3 e8g8 e1g1 c8g4 d4d5 c6e7 h2h3 g4h5 c3c4 h7h6 g1h2 g7g5 g2g4 h5g6 d1e2 c7c6 a2a4 g8g7 f1e1 f8e8 b2b3 c6d5 c4d5 a8c8 a4a5');
INSERT INTO evaluations VALUES(6584,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(6584,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(6584,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(6584,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(6584,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6584,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(6584,36,'cp 72','g8f6 d5b4 c6b4 d2d4 e5d4 e4e5 f6e4 f3d4 d8e7 c2c3 b4c6 f1d3 e4c5 e1g1 c5d3 d1d3 e8g8 d3g3 c6d4 c3d4 d7d6 c1h6 f7f6 f1e1 d6e5 d4e5 f8e8 h2h4 f6f5 e5e6 e7f6 h6g5 f6g6 g3c7 c8e6 c7b7 a8b8 b7a7 h7h6 g5f4 b8b2 a7a3 b2c2 a3a4');
INSERT INTO evaluations VALUES(10323,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(10323,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(10323,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(10323,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(10323,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(10323,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(10323,36,'cp 63','d5b4 c6b4 d2d4 d7d6 a2a3 b4c6 d4e5 f6e4 d1d5 c8f5 f1d3 c6e7 d5b5 c7c6 b5b4 d6d5 e1g1 a7a5 b4d4 c6c5 d4e3 c5c4 d3e2 e8g8 f3d4 f5g6 f2f4 d8d7 g2g4 f7f5 e5f6 f8f6 f4f5 g6f7 b2b3 a8e8 c1b2 e7c6 b3c4 e4d6 e3f3 d6c4 e2c4 d5c4 d4c6');
INSERT INTO evaluations VALUES(10345,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(10345,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(10345,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(10345,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(10345,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(10345,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(10345,36,'cp 65','c6b4 d2d4 d7d6 a2a3 b4c6 d4e5 f6e4 d1d5 c8f5 f1d3 c6e7 d5d4 d6d5 e1g1 c7c5 d4e3 c5c4 d3e2 e8g8 f3d4 a8c8 g2g4 f5d7 f2f3 e4c5 c1d2 c5e6 d4e6 d7e6 d2c3 e7c6 a1d1 d8d7 d1d2 c8e8 f1d1 d7c7 f3f4 d5d4');
INSERT INTO evaluations VALUES(10788,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(10788,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(10788,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(10788,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(10788,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(10788,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(10788,36,'cp 63','d2d4 d7d6 a2a3 b4c6 d4e5 f6e4 d1d5 c8f5 f1d3 c6e7 d5d4 e4c5 d3b5 c7c6 b5c4 c5e6 c4e6 f5e6 d4d6 d8d6 e5d6 e7f5 c1f4 e6d5 e1c1 e8c8 h1e1 d5f3 g2f3 f7f6 c2c4 g7g5 f4e3 f5d6 b2b3 d6f7 e3a7 f7e5 e1e3 d8d1 c1d1');
INSERT INTO evaluations VALUES(54831,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(54831,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(54831,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(54831,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(54831,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(54831,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(54831,36,'cp 64','b1c3 f8g7 c1e3 a7a6 g1f3 g8f6 d1d2 e8g8 e3h6 c7c5 h6g7 g8g7 e1c1 b8c6 d4c5 d6c5 d2f4 d8a5 e4e5 f6g4 f1c4 g4f2 f3g5 f7f6 g5f7 f8f7 f4f2 f7f8 e5f6 e7f6 d1d5 c8f5 d5c5 a5c7 c3d5 c7d6 d5f4 f5g4 c4d5 a8d8 h1e1 c6e5');
INSERT INTO evaluations VALUES(54825,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(54825,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(54825,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(54825,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(54825,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(54825,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(54825,36,'cp 68','g8f6 c1e3 c7c6 g1f3 f8g7 a2a4 f6g4 e3g5 h7h6 g5c1 e7e5 d4e5 g4e5 f3e5 d6e5 f1c4 d8d1 c3d1 a7a5 f2f3 b8d7 b2b3 g7f8 d1b2 f8c5 c1d2 f7f6 g2g4 d7b6 h2h4 b6c4 b2c4 b7b6 h4h5 g6g5 d2e3 c5e3 c4e3 c8e6 e3f5');
INSERT INTO evaluations VALUES(38794,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38794,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38794,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38794,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(38794,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(38794,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(38794,36,'cp 31','e4e5 c7c5 a2a3 b4c3 b2c3 g8e7 f1d3 c8d7 d1g4 c5c4 d3e2 e7f5 h2h4 b8c6 g4f4 h7h6 h4h5 f7f6 g1f3 e8g8 f4h2 d8a5 c1d2 f6e5 f3e5 c6e5 h2e5 a5b6 e1g1 f5d6 a1b1 b6c6 f2f3 d6f7 e5h2 a8e8 d2f4 e6e5 f4e5 f7e5 d4e5');
INSERT INTO evaluations VALUES(38847,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(38847,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(38847,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(38847,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38847,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(38847,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(38847,36,'cp 35','g8e7 a2a3 b4c3 b2c3 c7c5 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 d1c1 c6a5 e1g1 c5c4 d3e2 d7a4 f3h4 e8c8 g2g3 c8b8 g1h2 d8g8 h4g2 b8a8 g2e3 c7d8 e3g4 d8b6 a1b1 b6c7 g4e3 c7d7 e2f3 g8f8 b1b4');
INSERT INTO evaluations VALUES(41153,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(41153,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(41153,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(41153,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(41153,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(41153,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(41153,36,'cp 50','a2a3 b4c3 b2c3 c7c5 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 d1c1 c6a5 e1g1 a5c4 d2f4 e8c8 a3a4 c8b8 f4g3 g7g6 d4c5 c7c5 f3d4 b8a8 a4a5 e7f5 d4f5 g6f5 g3h4 d8c8 h4f6 h8h7 c1f4 c5c7 a5a6 b7b6');
INSERT INTO evaluations VALUES(41154,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(41154,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(41154,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(41154,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(41154,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(41154,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(41154,36,'cp 43','b4c3 b2c3 c7c5 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 d1c1 c6a5 e1g1 a5c4 d2f4 e8c8 a3a4 c8b8 d4c5 c7c5 a4a5 a7a6 f1e1 b8a8 f3d2 d7b5 d3c4 b5c4 f4e3 c5c6 e3b6 d8c8 d2f3 e7f5 f3d4 f5d4 c3d4 c4b5 c2c3 g7g6 h5g6 f7g6 b6c5 h6h5 c1g5 h5h4 g5g6');
INSERT INTO evaluations VALUES(41157,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(41157,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(41157,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41157,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(41157,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(41157,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(41157,36,'cp 57','b2c3 c7c5 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 d1c1 c6a5 e1g1 a5c4 d2f4 e8c8 a3a4 d8g8 f4g3 g7g6 g3h4 g6g5 h4g3 c8b8 f1e1 g8c8 d3c4 d5c4 f3d2 e7f5 d4d5 f5g3 d5d6 c7a5 f2g3 a5c3 d2f3 d7c6 a1a3 c3a5 c2c3 b8a8 f3d2');
INSERT INTO evaluations VALUES(41156,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(41156,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(41156,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(41156,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(41156,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41156,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41156,36,'cp 50','c7c5 h2h4 d8c7 a1b1 b8c6 g1f3 h7h6 f1e2 b7b6 h4h5 c6b8 e2b5 c8d7 b5d3 d7a4 f3h4 c5c4 d3e2 c7d7 g2g4 b8a6 c1d2 e8c8 d1c1 a6c7 h1h3 g7g5 h4g2 d8f8 f2f4 f7f6 e5f6 e7g8 f4g5 h6g5 d2g5 g8f6 g2h4 f6e4 h4g6 a4c2 g6h8');
INSERT INTO evaluations VALUES(40258,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(40258,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(40258,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(40258,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(40258,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(40258,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(40258,36,'cp 60','h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 d1c1 c6a5 e1g1 c5c4 d3e2 d7a4 f3h4 e8c8 g2g3 d8g8 h4g2 c8b8 g2f4 b8a8 a1b1 a4d7 e2g4 c7d8 g4h3 d8c7 g1h2 g8b8 a3a4 a5c6 b1a1 b8g8 a4a5 g7g5 h5g6 e7g6 a5a6');
INSERT INTO evaluations VALUES(55053,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(55053,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55053,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(55053,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55053,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55053,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55053,36,'cp 0','e7e6 c1f4 g8f6 e2e3 f8d6 f4g3 e8g8 b1d2 c7c5 g1f3 d8c7 d4c5 d6g3 h2g3 c7c5 f1d3 h7h6 e3e4 b8d7 d1e2 c5c7 e1g1 d5e4 d2e4 b7b6 e4f6 d7f6 e2e5 c7e7 f1e1 c8b7 a1d1 a8d8 d3c2 d8d5 e5e2 f8d8 f3e5 d5d1 e1d1 e7c7 a2a4 d8d1 c2d1');
INSERT INTO evaluations VALUES(55849,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55849,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55849,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55849,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55849,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55849,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55849,36,'cp 10','g1f3 g8f6 c1f4 f8d6 f4g3 b7b6 e2e3 c8b7 b1d2 e8g8 f3e5 c7c5 f1d3 d8c7 f2f4 f6e4 e1g1 e4d2 d1d2 b8c6 d2e2 c6e7 e2g4 g7g6 g3h4 e7f5 h4f6 d6e7 f6e7 c7e7 g4h3 f7f6 e5f3 c5c4 d3b1 f5d6 g2g4 d6f7 h3g3 g8h8 b1c2');
INSERT INTO evaluations VALUES(55846,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(55846,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(55846,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(55846,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(55846,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55846,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(55846,36,'cp 0','g8f6 c1f4 f8d6 e2e3 e8g8 f4g3 b7b6 b1d2 c7c5 f1d3 c8b7 d1e2 f6e4 f3e5 d8c7 e2h5 b8d7 e5d7 c7d7 g3d6 d7d6 h5h4 h7h6 d2e4 d5e4 d3e4 g7g5 h4g4 f7f5 g4h5 b7e4 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(48335,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(48335,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48335,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(48335,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48335,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(48335,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(48335,36,'cp 0','c1f4 f8d6 e2e3 e8g8 f1d3 c7c5 e1g1 b8c6 f4g3 d6g3 h2g3 d8e7 b1d2 e6e5 d4e5 c6e5 f3e5 e7e5 d1f3 c8e6 a2a4 g7g6 a4a5 a8d8 f1e1 h7h5 f3f4 e5f4 g3f4 g8g7 d2f3 f8e8 f3e5 h5h4 f2f3 f6d7 d3b5 d7e5 b5e8 e5d3');
INSERT INTO evaluations VALUES(48282,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(48282,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48282,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48282,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48282,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(48282,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(48282,36,'cp 0','f8d6 e2e3 e8g8 f4g3 b7b6 b1d2 c8b7 f1d3 c7c5 d1e2 d8c7 e3e4 d5e4 d2e4 f6e4 d3e4 b7e4 e2e4 b8d7 e1g1 c5d4 f3d4 d7f6 e4e2 d6g3 h2g3 f8d8 a1d1 c7c5 d4b3 c5e7 b3d4');
INSERT INTO evaluations VALUES(41788,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41788,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41788,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41788,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41788,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41788,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41788,36,'cp 2','f4g3 c7c5 e2e3 e8g8 b1d2 b7b6 f1d3 c8b7 d1e2 f6e4 f3e5 e4d2 e2d2 b8c6 d2e2 d8c7 e5c6 b7c6 e2h5 f7f5 a2a4 a7a5 h5e2 g7g6 g3d6 c7d6 f2f4 c5c4 d3c2 b6b5 g2g4 b5b4 h2h4 b4b3 c2d1 f5g4 e2g4');
INSERT INTO evaluations VALUES(49559,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(49559,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(49559,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49559,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(49559,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49559,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49559,36,'cp 13','f8b4 f1g2 e8g8 e2e4 d7d6 g1e2 b8c6 e1g1 c8g4 h2h3 g4e2 c3e2 b4c5 d2d3 c6d4 c1d2 a7a5 d2c3 d4e2 d1e2 h7h5 h3h4 c5b4 e2d2 b4c3 d2c3 c7c5 c3d2 a5a4 g2h3 d8a5 d2a5 a8a5 a2a3 a5a8 g1g2');
INSERT INTO evaluations VALUES(42771,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(42771,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42771,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(42771,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42771,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(42771,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(42771,36,'cp 17','f1g2 e8g8 e2e4 d7d6 g1e2 b8c6 d2d3 a7a6 e1g1 b4c5 a2a3 c6d4 h2h3 c7c6 b2b4 c5a7 a3a4 c8e6 c1e3 d8d7 g1h2 d6d5 c4c5 d5e4 c3e4 f6e4 d3e4 a8d8 e2d4 e5d4 e3d2 a7b8 f1e1 f8e8 f2f4 f7f6 d1c2 d4d3 c2b2 d7d4 b2d4 d8d4');
INSERT INTO evaluations VALUES(42772,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(42772,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42772,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42772,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(42772,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42772,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(42772,36,'cp 5','e8g8 e2e4 b4c3 b2c3 c7c6 g1e2 d7d5 e4d5 c6d5 d2d4 e5d4 c4d5 f6d5 d1d4 c8e6 e1g1 b8c6 d4c5 d8b6 c5a3 f8d8 e2f4 d5f4 c1f4 d8d7 a1b1 b6d8 a3a4 h7h6 f1d1 d7d1 b1d1 d8b6 h2h4 a8c8 c3c4 c8d8 d1d8 b6d8 g1h2 c6a5 a4b4 e6c4 f4e3');
INSERT INTO evaluations VALUES(37273,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(37273,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37273,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(37273,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(37273,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37273,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(37273,36,'cp 12','e2e4 b4c3 b2c3 c7c6 g1e2 d7d5 e4d5 c6d5 c4d5 f6d5 d2d4 e5d4 d1d4 c8e6 e1g1 b8c6 d4c5 d8b6 c5a3 f8d8 e2f4 d5f4 c1f4 d8d7 a1b1 b6d8 a3a4 h7h6 h2h4 a8c8 f1d1 d7d1 b1d1 d8e7 c3c4 c8d8 d1d8 e7d8 g1h2 g7g5 h4g5 h6g5 f4e3 c6e5 a4a7 e5g4 h2g1 d8d1 g2f1');
INSERT INTO evaluations VALUES(57199,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(57199,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(57199,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(57199,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(57199,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(57199,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(57199,36,'cp 37','g1f3 b8c6 d2d4 e5d4 f3d4 g8f6 c1g5 f8b4 d4c6 b7c6 d1d4 d8e7 g5f6 g7f6 e2e3 b4c5 d4d2 d7d5 c3e2 a8b8 c2c3 h7h5 e2d4 c8d7 f1d3 c5d6 d3f5 c6c5 f5d7 e8d7 d4f5 e7e6 f5d6 c7d6 e1c1 d7c6 d2d3 f6f5 d3a6 b8b6 a6a7 h8b8 d1d3 h5h4');
INSERT INTO evaluations VALUES(57197,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(57197,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(57197,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(57197,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(57197,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(57197,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(57197,36,'cp 25','b8c6 d2d4 e5d4 f3d4 g8f6 c1g5 f8b4 d4c6 b4c3 b2c3 b7c6 d1d4 h7h6 g5f6 d8f6 d4f6 g7f6 g2g3 c8a6 e2e3 a6f1 h1f1 h6h5 h2h4 e8e7 e1c1 c6c5 d1d5 d7d6 c1d2 a8b8 d5f5 b8b5 e3e4 b5b2 f1a1 a7a5 c3c4 b2b4 d2c3 b4a4');
INSERT INTO evaluations VALUES(22855,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(22855,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(22855,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(22855,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(22855,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(22855,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(22855,36,'cp 38','d2d4 e5d4 f3d4 g8f6 c1g5 f8b4 d1d3 e8g8 a2a3 b4a5 d4b3 c6e5 d3g3 a5c3 g3c3 f8e8 f2f3 c7c6 e1c1 d7d5 g5h4 a7a5 h4g3 e5g6 e2e4 a5a4 b3d2 d8b6 h2h4 h7h6 h4h5 g6f8 g3d6 f8d7 f1d3 c6c5 e4e5 d5d4 d2c4 d4c3 c4b6');
INSERT INTO evaluations VALUES(22754,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(22754,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(22754,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(22754,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(22754,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(22754,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(22754,36,'cp 26','e5d4 f3d4 g8f6 c1g5 f8b4 d1d3 e8g8 a2a3 b4a5 d4b3 c6e5 d3g3 a5c3 g3c3 f8e8 f2f3 d7d6 e1c1 a7a5 e2e3 a5a4 b3d4 c7c6 h2h4 h7h6 g5f4 f6d5 c3d2 e5g6 g2g4 d8e7 h4h5 g6f4 e3f4 e7f6 f4f5');
INSERT INTO evaluations VALUES(22936,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(22936,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(22936,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(22936,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(22936,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(22936,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(22936,36,'cp 34','f3d4 g8f6 c1g5 f8b4 d1d3 e8g8 a2a3 b4a5 d4b3 c6e5 d3g3 a5c3 g3c3 f8e8 f2f3 c7c6 e1c1 d7d5 g5h4 a7a5 h4g3 d8c7 e2e4 d5e4 b3d2 c7e7 c3e5 e7e5 g3e5 e8e5 d2e4 c8e6 e4f6 g7f6 f3f4 e5c5 f1d3 b7b5 h1e1 g8g7 g2g3 b5b4 e1e4 f6f5 e4d4 b4a3 b2a3');
INSERT INTO evaluations VALUES(22911,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(22911,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(22911,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22911,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(22911,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(22911,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(22911,36,'cp 27','g8f6 c1g5 f8b4 d4c6 b4c3 b2c3 b7c6 d1d4 h7h6 g5f6 d8f6 d4f6 g7f6 g2g3 h8g8 e1c1 g8g5 d1d4 g5a5 a2a4 e8f8 d4f4 a8b8 e2e3 d7d6 f1g2 c8d7 f4f6 a5a4 f6h6 f8g7 h6h5 g7g6 h5h4 a4h4 g3h4 b8h8');
INSERT INTO evaluations VALUES(16672,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16672,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(16672,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(16672,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(16672,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(16672,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(16672,36,'cp 29','c1g5 f8b4 d4c6 b4c3 b2c3 b7c6 d1d4 h7h6 g5f6 d8f6 d4f6 g7f6 g2g3 h8g8 e1c1 g8g5 d1d4 g5a5 a2a4 c6c5 d4f4 c8b7 h1g1 b7c6 f1g2 c6g2 g1g2 a8b8 g2g1 b8b6 g1d1 e8e7 c1d2 b6d6 d2e1 d6a6 d1a1 a6b6 e2e3 a7a6 h2h4');
INSERT INTO evaluations VALUES(45187,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(45187,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(45187,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(45187,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(45187,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45187,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(45187,36,'cp 20','d2d4 c5d4 f3d4 a7a6 f1e2 e7e5 d4b3 f8e7 c1e3 c8e6 c3d5 b8d7 d1d3 e8g8 e1g1 e6d5 e4d5 f6e8 a2a4 f7f5 f2f3 e7g5 e3f2 a6a5 b3d2 d7c5 f2c5 g5d2 d3d2 d6c5 e2b5 e8d6 a1d1 c5c4');
INSERT INTO evaluations VALUES(45174,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45174,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(45174,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45174,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45174,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(45174,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(45174,36,'cp 8','c5d4 f3d4 a7a6 f1d3 g7g6 e1g1 f8g7 f2f3 e8g8 c1e3 b8c6 d4c6 b7c6 a1b1 a8b8 d1d2 d8c7 b2b3 f8d8 c3a4 d6d5 d2e1 f6h5 a4c5 a6a5 g1h1 h5f4 e1d2 f4e6 b1d1 e6c5 e3c5 d5e4');
INSERT INTO evaluations VALUES(45259,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(45259,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(45259,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(45259,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45259,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(45259,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(45259,36,'cp 23','f3d4 a7a6 c1e3 e7e5 d4b3 c8e6 f1e2 f8e7 d1d3 e8g8 e1g1 b8c6 a2a3 a8c8 c3d5 e6d5 e4d5 e5e4 d3d2 c6e5 a1c1 e5c4 e2c4 c8c4 d2e2 d8c8 e3d4 f8e8 c2c3 f6d5 b3d2 c4c6 d2e4 d5f4 e2f3 f4e6 d4e3 d6d5 e4g3 c8d8');
INSERT INTO evaluations VALUES(55234,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(55234,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(55234,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(55234,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(55234,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(55234,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(55234,36,'cp 64','g8f6 g1f3 e7e6 c1g5 f8e7 f1d3 e8g8 d1e2 b8c6 e1c1 c6b4 h1e1 b4d3 e2d3 b7b6 d4d5 c8b7 d5e6 d8d3 d1d3 f7e6 e1e6 e7d6 g5f6 f8f6 e6f6 g7f6 g2g3 d6c5 c3d5 a8f8 d5c7 c5f2 c7d5 g8f7 d3d2 f2c5 b2b4 b7d5 d2d5 c5b4 d5d7 f7g8 d7a7 f8c8 f3d4');
INSERT INTO evaluations VALUES(6973,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(6973,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(6973,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(6973,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(6973,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6973,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(6973,36,'cp 51','d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 e1g1 e7b4 d4c6 d7c6 c4d3 f8e8 c1f4 f6d7 a2a3 b4d6 f4d6 c7d6 h2h3 d7c5 b2b4 c5e6 d1d2 a7a5 b4b5 e6d4 f1d1 c8d7 d3f1 d4b5 f1b5 c6b5 d2d6');
INSERT INTO evaluations VALUES(6976,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(6976,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(6976,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(6976,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(6976,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(6976,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(6976,36,'cp 45','e5d4 f3d4 g8f6 b1c3 e8g8 e1g1 e7b4 c3d5 f6d5 c4d5 b4c5 c1e3 c6d4 e3d4 d7d6 a2a4 c5d4 d1d4 f8e8 a4a5 c7c6 d5c4 c8e6 c4d3 c6c5 d4c3 d8c7 f1e1 a8d8 d3b5 e8e7 a1d1 d6d5 e4d5 d8d5 b5f1 d5d1 e1d1 h7h6');
INSERT INTO evaluations VALUES(6990,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(6990,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(6990,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(6990,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(6990,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(6990,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(6990,36,'cp 37','e1g1 g8f6 f3d4 e8g8 b1c3 e7b4 c3d5 f6d5 c4d5 b4c5 c1e3 c6d4 e3d4 d8e7 f1e1 d7d6 d1d2 c5d4 d2d4 f8e8 a1d1 a7a5 a2a4 e7e5 e1e3 e5d4 d1d4 g8f8 e3c3 c7c6 d5c4 d6d5 e4d5 c6d5 d4d5 e8e1 c4f1 c8e6 d5d3 b7b6 c3c7 a8e8 f2f3');
INSERT INTO evaluations VALUES(6989,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(6989,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(6989,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6989,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(6989,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(6989,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(6989,36,'cp 36','g8f6 f3d4 e8g8 b1c3 e7b4 c3d5 f6d5 c4d5 b4c5 c1e3 c6d4 e3d4 d7d6 a2a4 d8e7 f1e1 c7c6 d5c4 c8e6 c4e6 e7e6 d1d3 a8d8 a1d1 c5d4 d3d4 a7a6 a4a5 h7h6 f2f4 f7f5 h2h3 e6g6 e1e2 f8e8 e4f5 g6f5');
INSERT INTO evaluations VALUES(10944,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(10944,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(10944,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(10944,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(10944,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(10944,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(10944,36,'cp 36','f3d4 e8g8 b1c3 e7b4 c3d5 f6d5 c4d5 b4c5 c1e3 c6d4 e3d4 d7d6 a2a4 d8e7 f1e1 c7c6 d5b3 c8e6 b3e6 e7e6 d1d3 c5d4 d3d4 f8e8 f2f4 a7a6 a1d1 a8d8 a4a5 h7h6 d4b4 e6e7 h2h3 f7f6 b4c3 e7c7 b2b4 c7f7 e1e3 f6f5 e4f5 e8e3 c3e3');
INSERT INTO evaluations VALUES(10938,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(10938,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(10938,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(10938,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(10938,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(10938,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(10938,36,'cp 39','e8g8 b1c3 e7b4 c1g5 h7h6 g5f6 d8f6 c3d5 f6d4 d1d4 c6d4 d5b4 d7d6 a1d1 d4c6 b4d5 c8e6 a2a3 c6e5 c4e2 e6d5 d1d5 f8e8 f2f3 g8f8 g1f2 a7a6 f1d1 b7b5 g2g3 e5d7 b2b4 a8b8 c2c4 d7b6 d5d4 b6a4 c4b5 a6b5 d4d5');
INSERT INTO evaluations VALUES(46997,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(46997,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(46997,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(46997,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(46997,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(46997,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(46997,36,'cp 76','g1f3 g7g6 b1c3 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 b8d7 a1b1 f6e8 a2a3 b7b6 b2b4 f7f5 c1d2 d8h4 d1c1 h7h6 c1c2 h6h5 f3e2 g8h7 e4f5 g6f5 b4a5 a8a5 f2f4 e5e4 d2e1 h4f6 e2h5 a5a3 c3b5 a3a8 g1h1');
INSERT INTO evaluations VALUES(46995,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(46995,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(46995,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(46995,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(46995,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(46995,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(46995,36,'cp 57','g7g6 b1c3 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 b8a6 f1e1 e5d4 f3d4 a6c5 e2f1 f6g4 h2h3 d8f6 h3g4 f6d4 d1f3 c8d7 c1e3 d4f6 f3f6 g7f6 c3d5 f6b2 a1b1 b2e5 f2f4 e5g7 e3c5 d6c5 b1b7 a8b8 b7c7 g7d4 g1h2 d7g4 e4e5 g4e6 d5f6 g8g7');
INSERT INTO evaluations VALUES(46669,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(46669,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(46669,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(46669,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(46669,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(46669,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(46669,36,'cp 67','b1c3 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 b8a6 c1e3 d8e7 d1c2 a6b4 c2d2 c7c5 d4d5 f6g4 e3g5 f7f6 g5h4 b4a6 a2a3 g4h6 h2h3 b7b6 f3h2 a6c7 e2g4 c8a6 b2b3 h6f7 g1h1 h7h5 g4e2 g7h6 d2c2 g8h7 h4g3 h6g5');
INSERT INTO evaluations VALUES(46664,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(46664,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(46664,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(46664,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(46664,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(46664,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(46664,36,'cp 58','f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 e5d4 f3d4 f8e8 f2f3 c7c6 g1h1 b8d7 c1g5 h7h6 g5h4 d7e5 d1b3 g6g5 h4f2 a7a6 a1d1 f6h5 c4c5 h5f4 f2e3 d8f6 c5d6 f6d6 e3f4 g5f4 c3a4 a8b8 d4f5 d6e6 b3c2 e6f6 f5g7 f6g7 a4c5 g7f6');
INSERT INTO evaluations VALUES(41273,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(41273,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(41273,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(41273,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(41273,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(41273,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(41273,36,'cp 66','e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 b8d7 a1b1 h7h5 c1g5 d8e8 g5e3 f6h7 h3h4 h7f6 g2g3 g8h7 e3g5 d7c5 b2b3 e8d7 d1e2 b7b6 b1d1 f6g4 f3g4 h5g4 g5e3 g7h6');
INSERT INTO evaluations VALUES(57579,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(57579,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(57579,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(57579,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(57579,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(57579,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(57579,36,'cp 46','e7e6 g2g3 g8f6 g1f3 f8e7 b1c3 e8g8 f1g2 d7d5 e1g1 c7c6 d1c2 f6e4 c1f4 g7g5 f4e3 b8d7 a1d1 e4d6 b2b3 b7b6 f3e5 c8b7 f2f4 d7e5 d4e5 d6f7 c4d5 c6d5 c3b5 d8d7 b5d4 a7a5 d1c1 f8c8 c2b2 b7a6 c1c8 a8c8 f1c1 c8c1 b2c1 e7c5 c1d2 g8h8 h2h4');
INSERT INTO evaluations VALUES(57349,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(57349,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(57349,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(57349,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(57349,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(57349,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(57349,36,'cp 64','g2g3 g8f6 g1f3 f8b4 c1d2 b4e7 b1c3 e8g8 f1g2 d7d5 e1g1 c7c6 d1c2 f6e4 d2f4 g7g5 f4e3 b8d7 a1d1 e4d6 c4c5 d6f7 b2b4 e7f6 b4b5 a7a6 a2a4 f6g7 h2h4 h7h6 b5c6 b7c6 a4a5 a8b8 h4g5 h6g5 d1b1 b8b1 f1b1 d8a5 e3g5 f7g5 f3g5 g7d4');
INSERT INTO evaluations VALUES(57348,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(57348,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(57348,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(57348,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(57348,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(57348,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(57348,36,'cp 44','g8f6 g1f3 f8e7 b1c3 e8g8 f1g2 d7d5 e1g1 b8d7 d1c2 c7c6 f3g5 d7b8 a2a3 a7a5 g5f3 f6e4 c1f4 b8d7 a1d1 g7g5 f4e3 e4d6 b2b3 e7f6 c3a4 f8f7 f3e1 b7b5 c4b5 c6b5 a4c5 d8e7 e1d3 d7c5 d4c5 d6e4');
INSERT INTO evaluations VALUES(49941,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(49941,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(49941,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(49941,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(49941,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(49941,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(49941,36,'cp 58','f1g2 f8e7 g1h3 e8g8 b2b3 d7d6 e1g1 c7c6 c1b2 d8c7 e2e3 a7a5 a2a3 e6e5 b1c3 h7h6 a1c1 f8e8 c4c5 e5e4 d4d5 d6c5 f2f3 e4f3 d1f3 b8d7 f3f5 d7e5 f5b1 e5g4 c3d1 e7d6 h3f4 a5a4 b3b4 c6d5 f4d5 f6d5 g2d5 c8e6');
INSERT INTO evaluations VALUES(49942,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(49942,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(49942,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(49942,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(49942,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(49942,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(49942,36,'cp 53','d7d5 g1h3 f8e7 e1g1 e8g8 d1c2 b8c6 e2e3 c6b4 c2e2 d5c4 b1d2 b4d5 d2c4 b7b6 h3f4 f6e4 f1d1 d8e8 a2a3 c8a6 c1d2 d5f4 e3f4 e8c6 b2b3 f8d8 d2e3 c6e8 a1c1 a6b7 c4d2 b7d5 d2e4 f5e4');
INSERT INTO evaluations VALUES(49116,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(49116,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(49116,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(49116,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(49116,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(49116,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(49116,36,'cp 64','g1h3 f8e7 e1g1 e8g8 b1d2 a7a5 b2b3 c7c5 c1a3 c5d4 a3e7 d8e7 c4d5 f6d5 g2d5 e6d5 d2f3 b8c6 d1d2 c8e6 h3f4 e6f7 f3d4 g7g5 f4g2 e7f6 d4c6 b7c6 a1c1 a5a4 b3b4 f5f4 c1c5 f7e8 b4b5 c6b5 d2d5 e8f7 d5g5 f6g5 c5g5 g8h8 g5b5 f7a2 g2f4 h8g8 f2f3');
INSERT INTO evaluations VALUES(16655,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16655,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(16655,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(16655,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(16655,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16655,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(16655,36,'cp 0','e4e5 d7d5 c4b5 f6e4 f3d4 c8d7 b5c6 b7c6 e1g1 f8e7 f2f3 e4c5 f3f4 e8g8 f4f5 c5e4 b1c3 e7c5 c3a4 c5e7');
INSERT INTO evaluations VALUES(16349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16349,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16349,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(16349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16349,36,'cp 0','d7d5 c4b5 f6e4 f3d4 c8d7 b5c6 b7c6 e1g1 f8c5 f2f3 e4g5 f3f4 g5e4 c2c3 c5b6 c1e3 c6c5 d4b3 c5c4 e3b6 a7b6 d1d5 c4b3 d5e4 a8a2 b1d2 e8g8 d2b3 a2b2 b3d4 c7c5 e5e6 d7e6 d4e6 f7e6 e4e6 g8h8 a1d1 b2d2 d1d2 d8d2 e6e7 h8g8 e7e6');
INSERT INTO evaluations VALUES(15068,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(15068,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(15068,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(15068,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(15068,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(15068,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15068,36,'cp 0','c4b5 f6e4 f3d4 c8d7 b5c6 b7c6 e1g1 f8e7 f2f3 e4c5 f3f4 e8g8 f4f5 c5e4 b1c3 e7c5 c3a4 c5e7');
INSERT INTO evaluations VALUES(14946,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(14946,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14946,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(14946,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14946,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(14946,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(14946,36,'cp -10','f6e4 f3d4 c8d7 b5c6 b7c6 e1g1 f8c5 f2f3 e4g5 f3f4 g5e4 c1e3 d8b8 b2b3 c5b6 b1d2 c6c5 e5e6 e8g8 d2e4 c5d4 e3d4 d7e6 e4c5 e6f5 d1h5 b8c8 a1e1 f5e4 c2c3 f8e8 h5g5 f7f6 g5g3 g8h8 g3f2 a7a5 f4f5 b6c5 d4c5 c8d7');
INSERT INTO evaluations VALUES(15254,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15254,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15254,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15254,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15254,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15254,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(15254,36,'cp 0','f3d4 c8d7 b5c6 b7c6 e1g1 f8c5 f2f3 e4g5 f3f4 g5e4 c1e3 d8b8 b2b3 e8g8 d1d3 c5b4 a2a3 b4e7 d4f3 d7f5 f3d4');
INSERT INTO evaluations VALUES(15252,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15252,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15252,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15252,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15252,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15252,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15252,36,'cp 0','c8d7 b5c6 b7c6 e1g1 f8c5 f2f3 e4g5 f3f4 g5e4 c1e3 d8b8 b2b3 e8g8 d1d3 c5b4 a2a3 b4e7 d4f3 d7f5 f3d4');
INSERT INTO evaluations VALUES(25485,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25485,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25485,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25485,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25485,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25485,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25485,36,'cp 0','b5c6 b7c6 e1g1 f8c5 f2f3 e4g5 f3f4 g5e4 c1e3 d8b8 b2b3 e8g8 d1d3 c5b4 a2a3 b4e7 d4f3 d7f5 f3d4');
INSERT INTO evaluations VALUES(25452,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25452,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25452,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25452,36,'cp 0','b7c6 e1g1 f8c5 f2f3 e4g5 f3f4 g5e4 c1e3 d8b8 b2b3 e8g8 d1d3 c5b4 a2a3 b4e7 d4f3 d7f5 f3d4');
INSERT INTO evaluations VALUES(24378,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24378,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24378,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24378,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24378,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24378,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24378,36,'cp 0','e1g1 f8c5 f2f3 e4g5 f3f4 g5e4 c1e3 d8b8 b2b3 e8g8 d1d3 c5b4 a2a3 b4e7 d4f3 d7f5 f3d4');
INSERT INTO evaluations VALUES(24377,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24377,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24377,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24377,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24377,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24377,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(24377,36,'cp 0','f8c5 f2f3 e4g5 f3f4 g5e4 c1e3 d8b8 b2b3 c5b6 b1d2 c6c5 e5e6 e8g8 d4f5 d7e6 f5e7 g8h8 d2e4 d5e4 f4f5 b8e8 f5e6 e8e7 e6f7 c5c4 d1e1 f8f7 f1f7 e7f7 e3b6 c7b6 e1e4 a8e8 e4c6 c4b3 c2b3 f7e7 h2h3 e7e2 g1h1 e2h5 c6c3 h5e2');
INSERT INTO evaluations VALUES(23985,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23985,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23985,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23985,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23985,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23985,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(23985,36,'cp 0','f2f3 e4g5 c1e3 g5e6 f3f4 e8g8 f4f5 e6d4 e3d4 c5d4 d1d4 d8g5 h2h4 g5g3 f1f3 g3e1 f3f1');
INSERT INTO evaluations VALUES(23984,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23984,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23984,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23984,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23984,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23984,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23984,36,'cp 0','e4g5 c1e3 g5e6 f3f4 e8g8 f4f5 e6d4 e3d4 c5d4 d1d4 d8g5 h2h4 g5g3 f1f3 g3e1 f3f1');
INSERT INTO evaluations VALUES(23982,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23982,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23982,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23982,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23982,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23982,36,'cp 0','c1e3 g5e6 f3f4 e8g8 f4f5 e6d4 e3d4 c5d4 d1d4 d8g5 h2h4 g5g3 f1f3 g3e1 f3f1');
INSERT INTO evaluations VALUES(51919,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(51919,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(51919,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(51919,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(51919,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(51919,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(51919,36,'cp -28','a7a6 a2a4 g8f6 d2d3 b8c6 g1f3 d7d5 c4b3 d5e4 c3e4 f8e7 e1g1 e8g8 c2c3 b7b6 e4f6 e7f6 d1e2 a8a7 c1f4 a7d7 f1e1 c8b7 a1d1 f8e8 f4g3 h7h6 b3c2 g7g6 h2h3 c6e7 f3e5 f6e5 g3e5 e7f5 g1h1 f7f6 e5h2 h6h5 f2f3');
INSERT INTO evaluations VALUES(50483,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(50483,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(50483,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(50483,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(50483,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(50483,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(50483,36,'cp -16','a2a4 g8f6 d2d3 d7d5 c4b3 d5e4 c3e4 b8c6 g1f3 f8e7 c2c3 e8g8 e4f6 e7f6 e1g1 b7b6 d1e2 a8a7 c1f4 a7d7 f1e1 c8b7 a1d1 f8e8 f3d2 g7g6 h2h3 f6g7 b3c2 h7h6 f4h2 h6h5 d2e4 f7f5 e4d2 e6e5 f2f4 d8c7 d2f3');
INSERT INTO evaluations VALUES(50539,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(50539,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(50539,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(50539,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(50539,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(50539,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(50539,36,'cp -33','g8f6 d2d3 d7d5 c4b3 d5e4 c3e4 b8c6 g1f3 f8e7 c2c3 e8g8 e4f6 e7f6 d1e2 b7b6 c1f4 c8b7 e1g1 c6e7 f1d1 e7f5 d3d4 c5d4 f3d4 f5d4 c3d4 a8c8 e2e3 g7g6 h2h3 f6g7 f4e5 b7d5 e5g7 g8g7 e3e5 f7f6 e5g3 f8f7 d1e1 f7c7 a1d1 d8d7 b3d5 d7d5');
INSERT INTO evaluations VALUES(43316,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(43316,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(43316,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(43316,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(43316,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(43316,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(43316,36,'cp -19','d2d3 d7d5 c4b3 d5e4 c3e4 b8c6 g1f3 f8e7 e1g1 e8g8 c2c3 b7b6 d1e2 f6d5 f1d1 d8c7 d3d4 f8d8 d4c5 b6c5 b3c2 h7h6 e4g3 e7f8 d1e1 f7f5 c1d2 f8d6 e2c4 c7f7 a1d1 a6a5 c4b5 c8b7 b5e2 d8e8');
INSERT INTO evaluations VALUES(52748,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52748,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52748,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(52748,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(52748,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(52748,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(52748,36,'cp 11','c8f5 f1d3 f5g6 g1f3 g8f6 e1g1 b8d7 d3g6 h7g6 c4d5 c6d5 d1b3 a8b8 c1d2 a7a6 f1c1 e7e6 h2h3 d8b6 b3c2 f6e4 b1c3 b8c8 c2d3 e4d2 d3d2 b6d8 c3a4 c8c1 a1c1 f8e7 c1c2 g6g5 a4c5 d7c5 d4c5');
INSERT INTO evaluations VALUES(31070,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(31070,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31070,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(31070,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(31070,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31070,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31070,36,'cp 23','f1d3 f5g6 g1f3 e7e6 e1g1 g8f6 d1c2 b8d7 b1d2 g6d3 c2d3 f8e7 f1d1 e8g8 e3e4 d5e4 d2e4 c6c5 b2b3 c5d4 f3d4 d7c5 e4c5 e7c5 c1b2 e6e5 d4f5 d8d3 d1d3 f6e4 b2e5 e4f2 d3d7 f8d8 d7d8 a8d8 e5g7 a7a6 g1f1');
INSERT INTO evaluations VALUES(31069,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31069,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(31069,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(31069,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31069,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31069,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(31069,36,'cp 17','f5g6 g1f3 g8f6 e1g1 e7e6 d1c2 d5c4 c2c4 b8d7 f1d1 f8d6 b1d2 e6e5 d4e5 d6e5 c4c2 e8g8 a1b1 g6d3 c2d3 e5c7 d3c2 d7e5 f3e5 c7e5 d2f3 d8e7 c1d2 f8d8 h2h3 h7h6 b1c1 e5c7 d2c3 e7e6 a2a3 d8d1 c1d1');
INSERT INTO evaluations VALUES(20997,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(20997,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20997,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(20997,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(20997,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(20997,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(20997,36,'cp 8','c8f5 c1f4 e7e6 e2e3 f8d6 f4g3 g8f6 f1d3 c6e7 d1e2 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 f5e4 e2e4 d8d7 e1g1 b7b6 c2c3 f8e8 e4e2 h7h6 f1e1 e7f5 a1d1 a7a5 g3e5 d6e5 f3e5 d7a4 a2a3 a8d8 h2h3 d8d5 c3c4');
INSERT INTO evaluations VALUES(26557,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(26557,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(26557,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(26557,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(26557,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26557,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(26557,36,'cp 0','c1f4 e7e6 e2e3 f8d6 f4g3 g8f6 f1d3 c6e7 d1e2 e8g8 e3e4 f6e4 c3e4 d5e4 d3e4 d8d7 e1g1 f5e4 e2e4 a7a5 c2c3 a5a4 e4e2 d7c6 a2a3 c6a6 e2c2 h7h6 f1e1 a6c4 c2e4 c4b5 e4c2');
INSERT INTO evaluations VALUES(26530,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26530,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26530,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26530,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26530,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26530,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26530,36,'cp 0','e7e6 e2e3 f8d6 f4g3 g8f6 f1d3 c6e7 d1e2 e8g8 e3e4 f6e4 c3e4 f5e4 d3e4 d5e4 e2e4 a7a5 c2c3 d8d7 e1g1 d7c6 e4e2 h7h6 f1e1 b7b5 a2a3 f8e8 a1d1 e7d5 f3d2 d5e7 d2f3');
INSERT INTO evaluations VALUES(26805,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26805,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26805,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(26805,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(26805,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26805,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26805,36,'cp 0','e2e3 f8d6 f4g3 g8f6 f1d3 c6e7 d1e2 e8g8 e3e4 f6e4 c3e4 f5e4 d3e4 d5e4 e2e4 d8d7 e1g1 a7a5 c2c3 d7b5 e4c2 a5a4 f1e1 b5c4 f3d2 c4c6 d2f3');
INSERT INTO evaluations VALUES(26804,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26804,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26804,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26804,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26804,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26804,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26804,36,'cp 0','f8d6 f4g3 g8f6 f1d3 c6e7 d1e2 e8g8 e3e4 f6e4 c3e4 d5e4 d3e4 f5e4 e2e4 d8d7 e1g1 a7a5 c2c3 d7b5 e4c2 a5a4 f1e1 b5c4 f3d2 c4c6 d2f3');
INSERT INTO evaluations VALUES(22966,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(22966,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(22966,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(22966,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(22966,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(22966,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(22966,36,'cp 35','d2d4 g7g5 d4d5 g5g4 d5c6 g4f3 d1f3 d8h4 g2g3 f4g3 h2g3 h4f6 f3b3 f6g6 c6b7 g6e4 c1e3 e4b7 h1h4 b7b3 a2b3 c8b7 e3d4 f8d6 e1f2 f7f6 b1c3 d6e5 a1a5 e8c8 a5e5 f6e5 d4e5 d8f8 h4f4 g8e7 e5h8 f8h8 c3b5 e7c6 f1e2 c8b8 b5d4 c6e5 f4f5 d7d6');
INSERT INTO evaluations VALUES(22923,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(22923,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(22923,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(22923,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(22923,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(22923,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(22923,36,'cp 12','g7g5 d4d5 g5g4 d5c6 g4f3 d1f3 d8h4 g2g3 f4g3 h2g3 h4f6 f3b3 f6g6 c6b7 g6e4 c1e3 e4b7 h1h4 f8e7 h4f4 b7b3 a2b3 c8b7 b1a3 h7h5 a3b5 e7d8 f1c4 d7d5 c4d3 a7a6 b5d4 d8g5 f4f3 g5e3 f3e3 e8f8 e1d2 h8h6 a1a5 a8e8 a5c5 c7c6 c5a5 h6f6 e3e8 f8e8 d2e3 g8e7 d3a6 b7a6 a5a6 e8d7 a6a7 d7d6');
INSERT INTO evaluations VALUES(22492,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(22492,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22492,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22492,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(22492,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(22492,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(22492,36,'cp 12','d4d5 g5g4 d5c6 g4f3 d1f3 d8h4 g2g3 f4g3 h2g3 h4f6 f3b3 f6g6 c6b7 g6e4 c1e3 e4b7 h1h4 f8e7 h4f4 b7b3 a2b3 c8b7 b1a3 h7h5 a3b5 e7d8 f4b4 b7c6 a1a7 g8f6 a7a8 c6a8 b5c3 f6g4 e3f4 d7d6 b4b8 a8c6 f4g5 f7f6 g5h4 c6f3 b3b4 e8d7 f1h3 c7c6 b8b7 d7c8 b7g7 c8b8 g7d7 d8b6 d7d6 b6f2 e1d2');
INSERT INTO evaluations VALUES(52855,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(52855,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(52855,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(52855,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(52855,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(52855,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(52855,36,'cp 36','c8g4 f1e2 e7e6 e1g1 b8d7 d2d4 g8e7 b1d2 c6c5 c2c4 e7c6 c4d5 e6d5 a2a3 f8e7 d4c5 e7c5 h2h3 g4e6 d2b3 c5b6 c1f4 e8g8 a1c1 a8c8 f1e1 f7f6 e5f6 d7f6 e2d3 f6e4 f4e3 b6e3 e1e3 g8h8 d3e4 d5e4 d1d8 f8d8 b3c5');
INSERT INTO evaluations VALUES(31149,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(31149,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(31149,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31149,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(31149,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31149,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31149,36,'cp 45','f1e2 c6c5 c2c4 d5c4 b1a3 b8c6 a3c4 g8h6 e1g1 h6f5 e5e6 f7e6 b2b4 g7g6 b4c5 f8g7 a1b1 a8b8 c4e3 g4f3 e3f5 e6f5 e2f3 e8g8 c1b2 e7e5 d1b3 g8h8 d2d3 d8e7 f1e1 f8d8 b2a1 c6d4 a1d4 d8d4 c5c6 b7b6 e1c1 b8c8 c1c4 d4d8 a2a4 e7d6 g2g3 c8c7');
INSERT INTO evaluations VALUES(31150,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(31150,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31150,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(31150,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(31150,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31150,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31150,36,'cp 12','e7e6 e1g1 g8e7 d2d4 c6c5 d4c5 b8c6 f3e1 h7h5 e1d3 e7g6 h2h3 g4f5 b1d2 f5d3 c2d3 f8c5 d2b3 d8b6 b3c5 b6c5 c1e3 d5d4 a1c1 c5e5 e3d2 e8g8 e2h5 e5f5 d1f3 f5b5 h5g6 f7g6 f3e2 b5b2 c1b1 b2a2 b1b7 a2d5 f1e1 a8e8 h3h4 f8f5 e2e4 d5e4 e1e4 f5c5');
INSERT INTO evaluations VALUES(31503,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(31503,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31503,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(31503,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(31503,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31503,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31503,36,'cp 14','e1g1 g8e7 d2d4 c6c5 d4c5 b8c6 f3e1 h7h5 e1d3 e7g6 b1d2 d8c7 f1e1 g4e2 d1e2 f8c5 d3c5 c6d4 e2d1 c7c5 c2c3 d4f5 d2f3 c5b6 d1a4 b6c6 a4a3 g6h4 f3h4 f5h4 c1g5 h4f5 a1c1 b7b5 g5e3 a7a6 b2b3 h5h4 h2h3 a8c8 e3d4 h8h5 e1d1 c8c7');
INSERT INTO evaluations VALUES(22734,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(22734,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(22734,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(22734,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(22734,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(22734,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(22734,36,'cp 32','g2g3 g8f6 f1g2 f8c5 g1f3 d7d6 e1g1 e8g8 e2e3 c8f5 d2d3 a7a5 h2h3 h7h6 a2a3 f5h7 f3e1 d8d7 g1h2 f8e8 b2b3 a8d8 c1b2 c6e7 d1c2 c7c6 c3e4 f6e4 d3e4 f7f6 a1d1 e7c8 e1d3 c5b6 b3b4 a5b4');
INSERT INTO evaluations VALUES(22733,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(22733,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(22733,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(22733,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(22733,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(22733,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(22733,36,'cp 21','g8f6 f1g2 f8c5 g1f3 d7d6 e1g1 e8g8 e2e3 c8f5 d2d3 d8d7 d3d4 c5b6 d4e5 d6e5 d1d7 f5d7 b2b3 f8e8 a2a3 e5e4 f3d2 d7f5 b3b4 a7a5 a1b1 a8d8 c4c5 b6a7 d2c4 a5b4 a3b4 a7b8 b4b5 c6e5 c4e5 e8e5 c1a3 h7h5 f1d1');
INSERT INTO evaluations VALUES(16451,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(16451,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(16451,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(16451,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(16451,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(16451,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(16451,36,'cp 24','f1g2 f8c5 d2d3 e8g8 a2a3 d7d6 e2e3 c6e7 b2b4 c5b6 a3a4 c7c6 g1f3 c8e6 e1g1 b6c7 b4b5 d6d5 c1a3 h7h6 a4a5 f8e8 a5a6 b7a6 b5c6 e7c6 d1a4 d8d7 c4d5 f6d5 c3e4 a6a5 a1c1 d5b4 e4c5');
INSERT INTO evaluations VALUES(16452,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(16452,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(16452,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16452,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(16452,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(16452,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(16452,36,'cp 25','f8c5 g1f3 e8g8 e1g1 d7d6 e2e3 c8f5 d2d3 a7a5 h2h3 h7h6 a2a3 f5h7 b2b3 d6d5 c3d5 f6d5 c4d5 d8d5 f3g5 d5d3 g5h7 d3d1 f1d1 g8h7 d1d5 c5d6 c1b2 a8a6 d5b5 a6b6 b5b6 c7b6 a1d1 f8d8 g2c6 b7c6 b2e5 d6e7 d1d8 e7d8');
INSERT INTO evaluations VALUES(10180,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(10180,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(10180,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(10180,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(10180,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(10180,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(10180,36,'cp 36','g1f3 d7d6 e1g1 e8g8 e2e3 c8f5 d2d3 d8d7 d3d4 c5b6 d4e5 d6e5 d1d7 f6d7 b2b3 f8e8 c1a3 a8d8 c3d5 e5e4 f3d4 c6d4 e3d4 b6d4 d5e7 e8e7 a3e7 d8e8 a1d1 c7c5 e7g5 d7e5 g5e3 e5d3 b3b4 f5g4 e3d4 g4d1 f1d1 c5d4 g2e4 e8e4 d1d3 e4e2 d3d4');
INSERT INTO evaluations VALUES(10179,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(10179,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(10179,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10179,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(10179,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(10179,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(10179,36,'cp 28','e8g8 e1g1 d7d6 e2e3 c8f5 d2d3 d8c8 b2b3 f5g4 c1b2 c8d7 d1c2 a7a5 a2a3 c5b6 a1e1 a8e8 f3g5 c6e7 f2f3 g4f5 g5e4 f6e4 d3e4 f5e6 f3f4 f7f6 f4f5 e6f7 h2h3 c7c6 g1h2 e8c8 a3a4 d6d5 e4d5 c6d5 b2a3 d5c4 b3c4 f7c4');
INSERT INTO evaluations VALUES(4371,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(4371,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(4371,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(4371,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(4371,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(4371,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(4371,36,'cp 30','e1g1 d7d6 e2e3 c8f5 d2d3 d8c8 b2b3 f8e8 a2a3 c5b6 c3a4 f5h3 c1b2 c8d7 f1e1 h7h6 g2h1 a8d8 a4b6 a7b6 e3e4 c6e7 d3d4 e7g6 d1c2 d7g4 a3a4 f6d7 e1e3 d7f6 b3b4 c7c5 b4b5 f6d7 d4e5');
INSERT INTO evaluations VALUES(4370,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(4370,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4370,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(4370,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(4370,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(4370,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(4370,36,'cp 18','d7d6 e2e3 c8f5 d2d3 d8c8 b2b3 f5g4 c1b2 a7a5 d1d2 h7h6 a1c1 g4h3 c3d5 f6d5 c4d5 h3g2 g1g2 c6b4 b2e5 b4d5 e5b2 f7f5 e3e4 d5e7 f3h4 c8e6 d2c2 c5b6 f2f4 a5a4 b3a4 e6a2 c1a1 a2f7 g2h1 f5e4 d3e4');
INSERT INTO evaluations VALUES(3855,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(3855,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(3855,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(3855,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(3855,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(3855,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3855,36,'cp 36','e2e3 f8e8 d2d3 h7h6 b2b3 c8f5 h2h3 c6b4 e3e4 f5d7 a2a3 b4c6 b3b4 c5b6 c3e2 a7a5 a1b1 a5b4 a3b4 a8a2 b4b5 c6e7 e2c3 a2a8 c3a4 b6a5 g1h2 c7c6 c1e3 c6b5 c4b5 a8c8 d3d4 f6e4 d4e5');
INSERT INTO evaluations VALUES(3854,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(3854,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(3854,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(3854,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(3854,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3854,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3854,36,'cp 20','c8f5 d2d3 h7h6 h2h3 a7a5 a2a3 c5a7 d1c2 f5h7 a1b1 c6b8 b2b4 b8d7 c1b2 c7c6 b1a1 f8e8 f1d1 a7b6 f3d2 h7g6 a1c1 a5b4 a3b4 d7f8 d2e4 f6e4 d3e4 d8e7 c2d2 g6h5 d1f1 f8e6 c1a1 a8a1');
INSERT INTO evaluations VALUES(55546,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(55546,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(55546,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(55546,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(55546,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(55546,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(55546,36,'cp 31','g1f3 g8f6 b1c3 f8e7 d2d4 e5d4 f3d4 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 e4c2 d7f6 h2h3 a7a6 a2a4 h7h6 d1f3 d8d5 f3d5 f6d5 c2e4 d5b6 f1e1 e7d6 a4a5 b6d7 d4f5 d6f8 c1f4 d7f6');
INSERT INTO evaluations VALUES(53979,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53979,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(53979,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(53979,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(53979,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(53979,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53979,36,'cp 25','e1g1 g8f6 f1e1 b8c6 c4f1 e6e5 c2c3 f8e7 d2d4 c5d4 c3d4 c8g4 c1e3 d6d5 d4e5 f6e4 b1c3 e4c3 b2c3 e8g8 f1e2 a8c8 h2h3 g4f3 e2f3 c6e5 f3d5 c8c3 d5e4 e7f6 d1d8 f8d8 e3a7 c3a3 a7b6 d8d2 e4b7 a3a2 a1a2 d2a2 b6e3 e5c4');
INSERT INTO evaluations VALUES(53978,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(53978,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53978,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(53978,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(53978,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(53978,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53978,36,'cp 9','b8c6 f1e1 g8f6 c4f1 e6e5 c2c3 f8e7 d2d4 c5d4 c3d4 c8g4 c1e3 d6d5 d4e5 f6e4 b1c3 e4c3 b2c3 e8g8 f1e2 a8c8 h2h3 g4f3 e2f3 c6e5 f3d5 c8c3 d5b7 e7c5 e3c5 c3c5 d1b3 d8c7 b7a6 g7g6 a1d1 e5c4 b3g3 c4b6 g3c7 c5c7 d1c1 c7c1 e1c1 f8d8 a6f1 d8d2');
INSERT INTO evaluations VALUES(20676,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20676,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(20676,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(20676,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20676,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(20676,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(20676,36,'cp 23','f1e1 g8f6 c4b3 c6a5 d2d4 c5d4 d1d4 a7a6 b1c3 f8e7 e4e5 a5c6 d4f4 c6e5 f3e5 d6e5 f4e5 d8d6 e5d6 e7d6 c3d5 c8d7 d5b6 a8d8 b6d7 d8d7 c2c3 d6c5 g2g3 e8g8 b3c2 f8d8 g1g2 h7h6 c1f4 f6g4 a1d1 d7d1');
INSERT INTO evaluations VALUES(57651,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(57651,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(57651,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(57651,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(57651,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(57651,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(57651,36,'cp 37','g1f3 f8g7 d2d4 g8f6 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 e1e2 e7e6 h1d1 a8d8 e4e5 a5c6 d3e4 d8d7 a2a3 f8d8 e4c6 b7c6 e3g5 d8e8 b1c1 e8c8 h2h3 g4f5');
INSERT INTO evaluations VALUES(57649,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(57649,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(57649,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(57649,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(57649,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(57649,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57649,36,'cp 36','f8g7 d2d4 g8f6 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 f1d3 c8g4 d2a5 c6a5 e1g1 b7b6 f1d1 f8d8 d3a6 g4c8 a6e2 f7f5 e4e5 c8b7 e3g5 g8f8 a2a4 a8c8 g5e3 h7h6 h2h4 b7d5 d1c1');
INSERT INTO evaluations VALUES(39567,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(39567,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39567,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(39567,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(39567,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(39567,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(39567,36,'cp 38','d2d4 g8f6 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 f1d3 c8g4 d2a5 c6a5 e1g1 b7b6 f1d1 f8d8 d3a6 g4c8 a6e2 f7f5 e4e5 c8b7 b1c1 b7d5 a2a4 d8c8 e2a6 c8c1 d1c1');
INSERT INTO evaluations VALUES(39574,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(39574,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(39574,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(39574,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(39574,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(39574,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(39574,36,'cp 43','g8f6 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 a1b1 e8g8 f1e2 c5d4 c3d4 d8a5 c1d2 a5a2 e1g1 b7b6 f1e1 c8a6 h2h4 b8c6 d2e3 h7h5 d4d5 f8d8 b1c1 a8c8 e3g5 b6b5 d1c2 c6b4 c2a2 b4a2 c1c8 d8c8 g5e7 g7f8 e7f6 f8b4 e1a1 c8c1 a1c1 a2c1 e2f1 c1b3 e4e5');
INSERT INTO evaluations VALUES(52408,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52408,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52408,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52408,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(52408,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(52408,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52408,36,'cp 3','e7e6 b1c3 f8c5 c3e4 g8f6 e4c5 d8a5 c1d2 a5c5 e2e3 d7d5 c2c4 e8g8 c4d5 f8d8 g1f3 f6e4 a1c1 c5d5 f1c4 d5d6 d1c2 e4d2 c2d2 d6d2 f3d2 b8c6 d2f3 h7h6 e1e2 c8d7 c1c3 g7g5 h2h3 d7e8 h1c1 g8f8 c4d3 c6b4 d3b1');
INSERT INTO evaluations VALUES(51899,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51899,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51899,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51899,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(51899,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(51899,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(51899,36,'cp 9','e2e4 f8c5 f1d3 b8c6 g1f3 g8e7 e1g1 e7g6 b1c3 b7b6 f1e1 e8g8 e4e5 c8b7 a2a3 a7a6 c3e4 d8c7 e4c5 b6c5 h2h4 c6d4 f3d4 c5d4 d3g6 f7g6 d1d4 c7c6 d4g4 c6c2 c1e3 b7d5 g4d4 a8c8 a1c1 c2f5 f2f3 h7h6 b2b4 g6g5 h4g5 h6g5');
INSERT INTO evaluations VALUES(51893,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51893,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(51893,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(51893,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51893,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51893,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51893,36,'cp 2','f8c5 f1d3 g8e7 g1f3 b8c6 e1g1 e7g6 c2c3 e8g8 b1d2 c5e7 d2b3 b7b6 b3d4 c8b7 f1e1 a8c8 c1d2 d8c7 a1c1 f8d8 d1e2 g6f4 d2f4 c7f4 c1d1 c6d4 f3d4 e7c5 d3a6 b7a6 e2a6 f4b8 a6e2 d7d5 d4f3 d5e4 d1d8 c8d8 e2e4 c5e7');
INSERT INTO evaluations VALUES(37681,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37681,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(37681,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(37681,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(37681,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37681,36,'cp 1','f1d3 b8c6 g1f3 g8e7 e1g1 e8g8 f1e1 e7g6 c2c3 b7b6 c1e3 d7d6 f3d4 c8b7 b1d2 h7h6 d4c6 b7c6 d1e2 d8c7 d2f3 e6e5 h2h4 c5e3 e2e3 c7d7 g2g3 d7e6 c3c4 g6e7 a1d1 a8d8 d1d2');
INSERT INTO evaluations VALUES(47042,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47042,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(47042,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(47042,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(47042,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(47042,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(47042,36,'cp 21','b1c3 c7c5 d2d4 c5d4 f3d4 a7a6 c1e3 e7e5 d4b3 f8e7 h2h3 h7h6 d1d2 c8e6 e1c1 b8d7 g2g4 b7b5 f2f4 b5b4 c3d5 f6e4 d2b4 e6d5 d1d5 e4g3 h1g1 g3f1 g1f1 e8g8 b4a5 d8b8 f4e5 d7e5 b3d4 e5c4 a5c3 c4e3 c3e3 e7g5');
INSERT INTO evaluations VALUES(47032,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47032,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(47032,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(47032,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47032,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(47032,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(47032,36,'cp 23','c7c5 d2d4 c5d4 f3d4 a7a6 f2f3 e7e5 d4b3 c8e6 c1e3 h7h5 f1e2 b8d7 a2a4 f8e7 a4a5 h5h4 e1g1 h4h3 g2g3 e8g8 c3d5 f6d5 e4d5 e6f5 d1d2 a8c8 c2c3 f5g6 c3c4 f8e8 a1c1 d7f6 g3g4 f6h7 c4c5 d6c5');
INSERT INTO evaluations VALUES(22514,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(22514,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(22514,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22514,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(22514,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(22514,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(22514,36,'cp 25','g8f6 d2d3 f8c5 b2b4 c5b6 a2a4 a7a6 e1g1 e8g8 f1e1 d7d6 b1d2 c8e6 h2h3 f6h5 c4a2 d8d7 d2c4 b6a7 f3g5 h5f6 g5e6 f7e6 c1e3 a7e3 e1e3 b7b5 c4d2 c6e7 a2b3 e7g6 a4b5 a6b5 a1a8 f8a8 d3d4 e5d4');
INSERT INTO evaluations VALUES(16130,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(16130,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(16130,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(16130,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(16130,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(16130,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(16130,36,'cp 17','d2d3 f8c5 e1g1 d7d6 b2b4 c5b6 f1e1 e8g8 b1d2 a7a5 b4b5 f6g4 e1e2 c6e7 d3d4 e7g6 e2e1 f8e8 d1c2 g4f6 c4f1 c7c6 b5c6 b7c6 h2h3 c8a6 f1a6 a8a6 d4e5 g6e5 f3e5 e8e5 c3c4 b6d4 a1b1 a6b6 d2f3 b6b1 c2b1');
INSERT INTO evaluations VALUES(5400,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(5400,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(5400,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(5400,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(5400,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(5400,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5400,36,'cp 21','c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8e8 a1d1 b5c4 d4c5 a8d8 e1e7 e8e7 c5e7 d8d5 d1d5 c4d5 e7f6 h6f4 f6d4 a7a6 g2g3 f4c7 f2f4 f7f5 g1f2 g8f7 f2e2 d5c4 e2e3 c4b3 a2b3 f7e6 e3d2 e6d5 d2d3 c7d6 d4c3 d6c5 h2h3 c5f2');
INSERT INTO evaluations VALUES(5399,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(5399,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(5399,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(5399,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(5399,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(5399,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(5399,36,'cp 21','g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8e8 b3c5 b5c4 b2b3 c4d5 c5a6 h6d2 e1e5 d5c6 a6c7 d2f4 e5c5 f4c7 c5c6 e8d8 d4e3 c7b6 e3b6 a7b6 c6b6');
INSERT INTO evaluations VALUES(8050,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(8050,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8050,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(8050,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(8050,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(8050,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(8050,36,'cp 24','f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8e8 b3c5 b5c4 b2b3 c4d5 c5a6 h6d2 e1d1 d2a5 b3b4 d5c4 b4a5 c4a6 d4e3 e8b8 d1d7 b8b7 a1d1 b7d7');
INSERT INTO evaluations VALUES(8034,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(8034,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8034,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(8034,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(8034,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(8034,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(8034,36,'cp 30','e8g8 c4b3 d7d5 e4d5 c6a5 d1f3 a5b3 a2b3 c8g4 f3g3 g4h5 d5d6 f6g4 h2h3 g4e3 f2e3 a7a6 g3f2 e7e5 d4e2 h5e2 f2e2 d8d6 e3e4 f7f5 c3d5 g7h6 e1g1 f5e4 e2e4 d6c5 g1h2 g8h8 c2c3 a8d8 f1f8 c5f8 a1e1 h6g7 b3b4 f8f5 c3c4 f5e4 e1e4 h8g8 h2g3 g8f7 g3f2 f7e6 b2b3 g7f6 f2f3 h7h5 e4e1 d8d7');
INSERT INTO evaluations VALUES(3059,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(3059,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3059,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(3059,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(3059,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(3059,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3059,36,'cp 23','c4b3 d7d5 e4d5 c6a5 e1g1 a5b3 d4b3 b7b6 d5d6 c8b7 f1e1 e7d6 h2h3 d8c7 d1d2 f8e8 a2a4 f6d7 c3b5 c7c6 f2f3 d7e5 d2f2 a7a6 b5c3 e5c4 e3d4 g7d4 b3d4 c6c5 b2b3 e8e1 a1e1 c4e5 c3e2 a8c8 g1h1 e5c6 f2h4 g8g7 d4c6 b7c6 e2d4');
INSERT INTO evaluations VALUES(3071,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(3071,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(3071,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(3071,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(3071,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(3071,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(3071,36,'cp 12','d7d5 e4d5 c6a5 d1f3 a5b3 a2b3 c8g4 f3g3 g4h5 d5d6 f6g4 h2h3 g4e3 f2e3 a7a6 g3f2 e7e5 d4e2 h5e2 f2e2 d8d6 e3e4 f7f5 c3d5 g7h6 e1g1 f5e4 e2e4 d6c5 g1h2 g8h8 f1f8 a8f8 a1e1 h6f4 d5f4 e5f4 e4e7 c5e7 e1e7 b7b5 h2g1 f8c8 c2c3 c8d8 g1f2 d8d2 e7e2 d2d3 e2e8 h8g7 e8e7 g7f6 e7h7 d3d2 f2f3 d2b2');
INSERT INTO evaluations VALUES(16402,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(16402,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(16402,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(16402,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(16402,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(16402,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(16402,36,'cp -39','f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 f8c5 c2c3 c5e7 d2d4 c6c5 c1g5 c5d4 c3d4 e8g8 b1c3 c8e6 f1e1 f8e8 h2h3 h7h6 g5h4 f6d5 h4e7 d5e7 d1d2 c7c6 a1d1 d8d6 c3e4 d6c7 e5g4 e7d5 e4c5 e6g4 h3g4 d5f6 f2f3 a8d8 e1e8 d8e8 d1e1 e8e1 d2e1 c7f4 c5b7 f4d4 e1f2 d4c4 b2b3');
INSERT INTO evaluations VALUES(16834,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(16834,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(16834,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(16834,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(16834,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(16834,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(16834,36,'cp -34','c4d5 e4f6 d5c6 d7c6 f3e5 f8c5 c2c3 c5e7 d2d4 c6c5 d4c5 d8d1 f1d1 e7c5 e5d3 c5e7 c1g5 c8f5 d3f4 e8g8 g5f6 e7f6 f4d5 f6d8 b1d2 f8e8 d5e3 f5c8 d2f3 a7a5 a2a4 h7h5 f3d4 c7c6 h2h3 d8c7 e3c4 g7g6 d4f3 g8g7');
INSERT INTO evaluations VALUES(10380,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(10380,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(10380,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(10380,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(10380,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(10380,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(10380,36,'cp 28','c3d5 e8g8 c2c3 b4e7 d5f6 e7f6 a2a4 d7d6 e1g1 d8e7 a4a5 c8e6 d2d3 a7a6 c1e3 e7d7 c4e6 d7e6 d1c2 f6e7 d3d4 e5d4 f3d4 c6d4 e3d4 a8d8 c3c4 c7c5 d4e3 f7f5 e4f5 f8f5 a1e1 f5e5 e3d2 e7f6 e1e5 f6e5 f1e1 e6f6 d2c3 e5c3 b2c3');
INSERT INTO evaluations VALUES(10324,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(10324,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(10324,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(10324,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(10324,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(10324,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(10324,36,'cp 31','e8g8 c2c3 b4e7 d2d3 d7d6 e1g1 f6d5 c4d5 a7a5 a2a4 e7f6 f1e1 h7h6 h2h3 c6e7 d5b3 e7g6 d3d4 c8e6 c1e3 d8e7 b3c2 g6h4 f3h4 f6h4 d1d2 e7f6 b2b4 h4g5 b4b5 e6c4 a1b1 b7b6 e3g5 f6g5 e1e3 a8c8 c2b3 c4b3 b1b3 g5f6 e3f3');
INSERT INTO evaluations VALUES(4414,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(4414,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(4414,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(4414,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(4414,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(4414,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(4414,36,'cp 26','c2c3 b4e7 d2d3 d7d6 e1g1 f6d5 c4d5 g8h8 d3d4 e5d4 c3d4 c6b4 d5b3 d6d5 f1e1 d5e4 e1e4 c8f5 e4e5 d8d7 a2a3 b4c6 e5e1 e7f6 d4d5 c6e7 c1g5 f6g5 f3g5 a8d8 d1d4 e7d5 a1d1 c7c6 d4a7 h7h6 g5f3 f8e8 e1e8 d8e8 b3d5 c6d5 h2h3 f5e4 f3d4');
INSERT INTO evaluations VALUES(55090,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55090,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(55090,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(55090,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(55090,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(55090,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(55090,36,'cp -25','c7c5 e2e3 g8f6 b2b3 g7g6 f1e2 f8g7 c1b2 e8g8 e1g1 e7e6 f3e5 b7b6 c2c4 c8b7 e2f3 f6e4 d1e2 f7f6 e5g4 f6f5 b2g7 g8g7 g4e5 b8d7 e5d7 d8d7 d2d3 e4f6 b1a3 a8b8 e2b2 f8d8 a1d1 d7e7 h2h3 g7g8 f1e1 d8d7 a3b5 d5c4 f3b7 b8b7');
INSERT INTO evaluations VALUES(53297,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(53297,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(53297,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(53297,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(53297,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(53297,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(53297,36,'cp -13','d2d3 g8f6 g2g3 e7e6 f1g2 f8e7 e1g1 b8c6 e2e4 d5e4 d3e4 f6e4 d1d8 c6d8 c1e3 f7f6 f3e1 e4d6 e3c5 d6c4 c5d4 d8c6 g2c6 b7c6 e1d3 c8a6 d4f2 e8f7 f1d1 h8d8 b2b3 c4b6 b1c3 a6d3 c2d3 b6d5 c3d5 d8d5 a1c1 c6c5 f2e3 a8d8 g1f2 d5d3 d1d3 d8d3');
INSERT INTO evaluations VALUES(53294,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(53294,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(53294,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(53294,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(53294,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(53294,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(53294,36,'cp -40','g8f6 g2g3 e7e6 f1g2 f8e7 e1g1 e8g8 a2a4 b8c6 e2e3 a7a6 b2b3 a8b8 d1e2 d5d4 f3e1 d8c7 e3e4 b7b5 b1d2 f6d7 h2h4 d7b6 h4h5 b5a4 h5h6 g7h6 b3a4 a6a5 e2h5 g8h8 g1h2 f8g8 d2f3 f7f6 h5h6 c6b4');
INSERT INTO evaluations VALUES(19485,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19485,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(19485,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(19485,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19485,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19485,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19485,36,'cp 0','g1f3 g8f6 b1c3 e7e6 f1b5 c6d4 b5d3 d4c6');
INSERT INTO evaluations VALUES(19484,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(19484,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(19484,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19484,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19484,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19484,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19484,36,'cp 3','e7e6 f1b5 g8e7 e1g1 a7a6 b5e2 d7d5 d2d3 e7g6 g2g3 f8e7 c2c3 b7b5 a2a4 a8b8 a4b5 a6b5 d1e1 g6f8 b1a3 f8d7 f4f5 b5b4 a3c2 e8g8 f5e6 f7e6 e4d5 e6d5 c1f4 b8b6 a1a8 b4c3 b2c3 d7f6 d3d4 e7d6 f3e5');
INSERT INTO evaluations VALUES(18855,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(18855,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18855,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18855,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(18855,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(18855,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18855,36,'cp 16','f1b5 g8e7 e1g1 a7a6 b5e2 d7d5 d2d3 e7g6 g2g3 f8e7 c2c3 b7b5 c1e3 d5d4 e3f2 e6e5 f4f5 g6f8 a2a4 a8b8 b1d2 f8d7 g3g4 e8g8 a4b5 a6b5 g1h1 d4c3 b2c3 b5b4 c3b4 c5b4 d2b3 c8b7 f1g1 d8c7 g4g5 b8c8');
INSERT INTO evaluations VALUES(18697,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(18697,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(18697,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18697,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18697,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18697,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18697,36,'cp 0','c6d4 f3d4 c5d4 e1g1 a7a6 b5e2 f8c5 d2d3 g8e7 b1d2 e8g8 g1h1 b7b6 c2c3 e7c6 d1e1 d7d5 d2b3 a6a5 e1g3 f8e8 e2f3 c8a6 b3c5 b6c5 f4f5 a6d3 c1h6 d8f6 h6g5 f6e5 g5f4 e5f6');
INSERT INTO evaluations VALUES(34676,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(34676,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(34676,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(34676,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34676,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(34676,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(34676,36,'cp 62','g8f6 g1f3 c8f5 f3e5 c7c6 h2h4 b8d7 e5c4 a5c7 d1f3 f5g4 c1f4 g4f3 f4c7 f3d5 c3d5 c6d5 c4e3 a8c8 c7f4 e7e6 f1d3 h7h5 c2c3 f6g4 e3f1 g4f6 g2g3 f6e4 f1d2 d7f6 d3b5 e8d8 d2f3 f8d6 f3e5 d6e5 d4e5 f6g4 f2f3 a7a6 b5d3 g4f2');
INSERT INTO evaluations VALUES(33734,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(33734,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(33734,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(33734,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(33734,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(33734,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(33734,36,'cp 76','g1f3 c8f5 f3e5 c7c6 c1f4 b8d7 e5c4 a5d8 c4e3 f5e6 a2a4 g7g6 a4a5 f6h5 f4g5 h7h6 g5h4 f8g7 f1e2 d7f6 e1g1 h5f4 e2f3 a8c8 d1d2 e8g8 f1d1 b7b5 a5b6 a7b6 h2h3 f6h5 c3e2 f4e2 d2e2 h5f4 e2a6 b6b5 h4g3 f4d5 e3d5 e6d5 f3d5 c6d5 c2c3');
INSERT INTO evaluations VALUES(16387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16387,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16387,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(16387,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16387,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(16387,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(16387,36,'cp 0','d2d3 c6a5 c4b3 a5b3 a2b3 f8b4 g1e2 d7d5 e1g1 d5e4 d3e4 e8g8 d1d8 f8d8 c1e3 a7a5 f1d1 d8d1 a1d1 c8e6 f2f3 h7h6 g1f2 b4e7 e2c1 f6d7 c3b5 a8c8 b5a7 c8f8 a7b5');
INSERT INTO evaluations VALUES(16389,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16389,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16389,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16389,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16389,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16389,36,'cp 0','c6a5 c4b3 a5b3 a2b3 f8b4 g1f3 d7d6 h2h3 e8g8 e1g1 f8e8 c1d2 h7h6 c3e2 b4d2 d1d2 f6h5 b3b4 c8d7 a1a5 d8f6 d2e3 a7a6 f1e1 f6e7 c2c4 e7e6 f3h4 e6f6 h4f3');
INSERT INTO evaluations VALUES(17051,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17051,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17051,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17051,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17051,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17051,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17051,36,'cp 0','c4b3 f8b4 g1f3 a5b3 a2b3 d7d6 h2h3 e8g8 e1g1 f8e8 c1d2 h7h6 c3e2 b4d2 d1d2 f6h5 b3b4 c8d7 a1a5 d8f6 d2e3 a7a6 f1a1 f6e6 f3h4 e6f6');
INSERT INTO evaluations VALUES(17057,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17057,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17057,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(17057,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17057,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17057,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17057,36,'cp 0','f8b4 g1f3 a5b3 a2b3 d7d6 h2h3 e8g8 e1g1 f8e8 c1d2 a7a6 c3e2 b4d2 d1d2 f6h5 b3b4 c8d7 f1e1 d8f6 d2e3 h7h6 g2g4 h5f4 e2f4 e5f4 e3d4 f6g6 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(35103,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(35103,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(35103,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(35103,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(35103,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(35103,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(35103,36,'cp 122','d7d6 d2d4 e5d4 f3d4 b8c6 d4c6 b7c6 e1g1 g8h6 b1c3 f8e7 h2h3 e8g8 f1e1 g8h8 c3e2 c8e6 c4d3 a8e8 c2c3 e6d7 e2f4 h6g8 d3c2 e7d8 f4h5 f6h4 e4e5 f7f5 c1f4 g8f6 e5f6 e8e1 d1e1 h4h5 f6g7 h8g7 c2b3 f8e8 e1d2 d8f6');
INSERT INTO evaluations VALUES(34776,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(34776,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(34776,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(34776,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(34776,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(34776,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(34776,36,'cp 128','d2d4 e5d4 f3d4 b8c6 d4c6 b7c6 e1g1 f8e7 f1e1 g8h6 h2h3 e8g8 b1c3 c8d7 e4e5 f6g6 c3e2 e7g5 e2g3 g5c1 d1c1 a8e8 c4d3 f7f5 e5d6 c7d6 c1f4 g6f6 f4a4 h6f7 a4a7 d7e6 a7c7 e8c8 c7b7 f7e5 b7b4 c6c5 b4d2 e5d3 d2d3');
INSERT INTO evaluations VALUES(34778,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(34778,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(34778,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(34778,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(34778,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(34778,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(34778,36,'cp 125','e5d4 f3d4 b8c6 d4c6 b7c6 b1c3 f8e7 h2h3 f6e5 c1e3 g8f6 e1g1 e8g8 f2f4 e5h5 d1d2 a7a5 c4e2 h5h4 a1e1 f8e8 e2d3 e7f8 f1f3 f6d7 g2g4 h4e7 g4g5 a5a4 f3g3 c8a6 d3a6 a8a6 g1g2 a6a8 h3h4 e7e6 e3d4');
INSERT INTO evaluations VALUES(47352,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(47352,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(47352,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(47352,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47352,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(47352,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(47352,36,'cp 30','g1f3 c7c5 f1c4 e7e6 e1g1 a7a6 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 b2b3 b8d7 c4e2 b7b6 f3e1 a6a5 c1b2 h8d8 d2c4 b6b5 c4d2 c8a6 e1f3 c5b4 a2a3 b4d2 f3d2 d7c5 f1d1 b5b4 e2a6 a8a6 a3b4 a5b4 a1a6 c5a6');
INSERT INTO evaluations VALUES(47353,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(47353,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(47353,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(47353,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(47353,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47353,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47353,36,'cp 19','e7e6 f1c4 c7c5 e1g1 a7a6 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 b2b3 b8d7 c4e2 b7b6 f3e1 a6a5 c1b2 h8d8 d2c4 b6b5 c4d2 c8a6 e1f3 e7f8 f1c1 c5e7 a2a3 f6d5 f3d4 b5b4 e2a6 a8a6 a3b4 e7b4 b2a3 e6e5');
INSERT INTO evaluations VALUES(48489,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48489,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(48489,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(48489,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(48489,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(48489,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(48489,36,'cp 36','f1c4 c7c5 e1g1 a7a6 d4c5 d8d1 f1d1 f8c5 b1d2 e8g8 c4e2 f8d8 f3e5 b7b5 a2a4 b5a4 a1a4 c8b7 d2b3 d8d1 e2d1 c5d6 e5c4 d6c7 d1e2 h7h6 c1d2 b8d7 f2f3 d7b6 c4b6 c7b6 b3a5 b7d5 a4a3 b6c7 b2b4 c7d6');
INSERT INTO evaluations VALUES(48470,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(48470,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(48470,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(48470,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(48470,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(48470,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48470,36,'cp 18','a7a6 e1g1 c7c5 b2b3 b8d7 c1b2 f8e7 d4c5 e7c5 b1d2 b7b5 c4d3 c8b7 a2a4 b5a4 b3a4 e8g8 a1b1 h7h6 d2c4 b7d5 d1e2 d5c4 d3c4 d8a5 b1a1 f8c8 h2h3 c5f8 f1d1 d7b6 c4d3 b6d5 a1c1 a5a4 f3e5 c8c1 d1c1');
INSERT INTO evaluations VALUES(43672,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(43672,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43672,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(43672,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(43672,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(43672,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43672,36,'cp 22','e1g1 c7c5 b2b3 b8d7 c1b2 f8e7 d4c5 e7c5 c4d3 b7b6 a2a3 c8b7 b3b4 c5d6 b1d2 e8g8 d1b1 d6e7 d2e4 f6e4 d3e4 b7e4 b1e4 e7f6 b2d4 a6a5 a1c1 a5b4 a3b4 h7h6 c1c6 f6d4 e4d4 a8c8 f1c1 c8c6');
INSERT INTO evaluations VALUES(43671,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(43671,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(43671,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43671,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(43671,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(43671,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(43671,36,'cp 21','c7c5 b2b3 b8d7 c1b2 f8e7 d4c5 e7c5 b1d2 b7b5 c4d3 e8g8 d1c2 c5e7 a2a4 b5a4 b3a4 c8b7 f1d1 h7h6 d2c4 b7f3 g2f3 a8c8 a1c1 f6d5 d3e4 d5b6 c2b3 b6c4 c1c4 d8b6 b3c3 e7f6 c3c2 c8c4 c2c4 f6b2 d1d7 b2f6');
INSERT INTO evaluations VALUES(34015,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(34015,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(34015,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(34015,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(34015,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(34015,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(34015,36,'cp 90','d2d4 b8c6 b1c3 c6e5 d4e5 e7e5 f2f4 e5e6 d1d4 d7d5 f4f5 e6b6 d4b6 a7b6 e4e5 d5d4 c3b5 f6d5 f1c4 c7c6 c4d5 a8a5 d5c4 c6b5 c4d3 b5b4 c1f4 c8d7 e1f2 f8e7 h1e1 e8g8 g2g4 e7h4 f4g3 h4g5 d3c4 g7g6 a1d1 g6f5 d1d4 d7e6 g4f5 e6f5 f2g2 f5c2 d4d7 c2f5 d7b7');
INSERT INTO evaluations VALUES(34014,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34014,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(34014,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(34014,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(34014,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(34014,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(34014,36,'cp 74','d7d6 e5f3 e7e4 f1e2 c8f5 b1a3 e4e7 e1g1 e7d7 a3c4 f8e7 c4e3 f5g6 f3h4 c7c6 g2g4 f6d5 h4g6 d5e3 c1e3 h7g6 d4d5 e8g8 g4g5 f8e8 f1e1 e7f8 e2f3 b8a6 c2c4 a6c5 f3g2 a8d8 b2b4 c5e4 a1c1 c6c5 b4b5 a7a6 b5a6 b7a6 d1f3');
INSERT INTO evaluations VALUES(33774,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(33774,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(33774,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(33774,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(33774,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(33774,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(33774,36,'cp 98','e5f3 e7e4 f1e2 c8f5 b1a3 e4e6 f3h4 f5g6 e1g1 f8e7 f1e1 e6d7 e2f3 b8c6 a3c4 e8g8 d4d5 c6e5 c4e5 d6e5 e1e5 e7d6 e5e1 a8e8 c1e3 e8e5 c2c4 f8e8 a1c1 g6h5 h2h3 d6c5 d1b3 c5e3 e1e3 e5e3 f2e3 h5f3 h4f3 b7b6 b3d3 h7h6 b2b4 a7a5 a2a3 a5b4 a3b4 d7e7');
INSERT INTO evaluations VALUES(33793,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(33793,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(33793,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(33793,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(33793,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(33793,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(33793,36,'cp 99','e7e4 f1e2 c8f5 b1a3 e4e7 e1g1 e7d7 c2c4 c7c6 d4d5 f8e7 f3h4 f5e4 f2f3 e4g6 g2g4 e7d8 g1g2 d8b6 c1g5 e8g8 g5f6 g7f6 e2d3 f8e8 h4f5 g6f5 d3f5 d7c7 d1c2 b8a6 f5h7 g8g7 h7e4 a6c5 a1e1 e8e5 f3f4 e5e4 e1e4 c5e4 c2e4 c6d5 c4d5');
INSERT INTO evaluations VALUES(33872,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(33872,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(33872,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(33872,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(33872,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(33872,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(33872,36,'cp 98','f1e2 c8f5 b1a3 e4e7 e1g1 e7d7 c2c4 c7c6 d4d5 f8e7 f3h4 f5e4 f2f3 e4g6 g2g4 e7d8 g1g2 d8b6 c1g5 e8g8 g5f6 g7f6 e2d3 f8e8 h4f5 g6f5 d3f5 d7e7 f1e1 e7e1 d1e1 e8e1 a1e1 b8a6 e1e7 c6d5 c4d5 a6c7 g2g3 c7d5 e7b7 b6d4 a3c4 d5b6 b2b3 a8e8 c4b6 a7b6 f5e4 e8e5 b7d7 e5c5 a2a4 d4e5 g3f2 c5c3 e4d5');
INSERT INTO evaluations VALUES(33870,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(33870,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(33870,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(33870,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(33870,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(33870,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(33870,36,'cp 92','c8f5 b1a3 e4e7 e1g1 e7d7 c2c4 c7c6 d4d5 f8e7 f3h4 f5e4 f2f3 e4g6 g2g4 e7d8 g1g2 d8b6 c1g5 e8g8 g5f6 g7f6 e2d3 f8e8 h4f5 g6f5 d3f5 d7e7 a3c2 e7e2 d1e2 e8e2 g2g3 b8a6 d5c6 b7c6 b2b4 c6c5 a2a3 a6c7 f1d1 a8e8 f5d3 e2e5 d3e4 c7e6 d1d5 e6g5 d5e5 e8e5 a1e1 g5e6 e4d5 e5e1 c2e1');
INSERT INTO evaluations VALUES(32688,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(32688,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(32688,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(32688,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(32688,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(32688,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(32688,36,'cp 93','b1a3 e4e7 e1g1 e7d7 c2c4 c7c6 d4d5 f8e7 f3h4 f5e4 f2f3 e4g6 g2g4 e7d8 g1g2 d8b6 c1g5 e8g8 g5f6 g7f6 e2d3 f8e8 h4f5 g6f5 d3f5 d7e7 f1e1 e7e1 d1e1 e8e1 a1e1 b8a6 e1e7 a8b8 h2h4 a6b4 e7d7 c6d5 h4h5 b6c5 h5h6 b4c6 c4d5 c6e5 d7c7 c5b6 c7c8 b8c8 f5c8 e5d3 c8b7 d3b2 b7c6');
INSERT INTO evaluations VALUES(55888,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55888,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55888,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(55888,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(55888,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55888,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(55888,36,'cp -10','d5d4 c3e2 c7c5 c2c3 g8f6 e4e5 f6d7 c3d4 c5d4 e2d4 d7e5 f1b5 e5c6 e1g1 c8d7 d4e2 f8d6 d2d4 c6e7 b5c4 d7c6 e2c3 h7h6 h2h3 e8g8 f1e1 d6c7 f3e5 c7e5 d4e5 d8d1 e1d1 b8d7 f2f4 d7b6 c4b3 b6d5 c1d2 d5c3 d2c3 e7d5');
INSERT INTO evaluations VALUES(56057,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(56057,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56057,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(56057,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(56057,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56057,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56057,36,'cp 0','c3e2 c7c5 c2c3 g8f6 c3d4 c5d4 e4e5 f6d7 e2d4 d7e5 f1b5 e5c6 e1g1 f8e7 d4c6 b8c6 f3e5 c8d7 b5c6 d7c6 e5c6 b7c6 d2d4 e8g8 c1e3 d8d5 d1a4 f8b8 b2b3 b8b4 a4a6 b4b6 a6a4');
INSERT INTO evaluations VALUES(56062,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(56062,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(56062,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(56062,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(56062,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(56062,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(56062,36,'cp -13','c7c5 c2c3 g8f6 c3d4 c5d4 e2d4 f6e4 b2b3 e4f6 c1b2 f8d6 d4b5 e8g8 b5d6 d8d6 f1c4 b7b6 b2f6 g7f6 e1g1 c8b7 d2d4 f8d8 a1c1 b8d7 d4d5 b7d5 c4d5 d6d5 d1d5 e6d5 f3d4 d7c5 f1d1 d8d7 g1f1 a8e8 c1c3 c5e6 c3d3 e8c8 h2h3 e6d4 d3d4 c8c2 d4d5 d7d5');
INSERT INTO evaluations VALUES(54111,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54111,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54111,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(54111,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54111,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(54111,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(54111,36,'cp -4','c2c3 g8f6 c3d4 c5d4 e2d4 f6e4 b2b3 c8d7 f1d3 e4f6 d1e2 b8c6 c1b2 f8e7 e1g1 e8g8 d4c6 d7c6 f3e5 d8d5 e5c6 d5c6 a1c1 c6b6 g2g3 a8d8 b2e5 f6d5 d3e4 e7f6 d2d4 f6e5 d4e5 b6a5 e2c2 g7g6 f1d1 d8d7 e4d5 d7d5 d1d5 a5d5 c2c7 d5d2 c1c4 d2a2 c7b7 a2b1 g1g2');
INSERT INTO evaluations VALUES(54108,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54108,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(54108,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(54108,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(54108,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(54108,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54108,36,'cp -3','g8f6 c3d4 c5d4 e4e5 f6d7 e2d4 d7e5 f1b5 e5c6 e1g1 c8d7 d4b3 f8e7 d2d4 c6b4 b5c4 d7a4 a2a3 b4d5 d1e2 a4b3 c4b3 b8c6 e2e4 e8g8 b3c2 d5f6 e4d3 a8c8 b2b4 d8d5 c1b2 f8d8 c2b3 d5e4 d3d2 e4f5 f1e1 a7a6 a1c1 f6d5 b3c2 f5f6 c2b3');
INSERT INTO evaluations VALUES(16425,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(16425,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16425,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(16425,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(16425,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16425,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16425,36,'cp 0','e5d4 e4e5 d7d5 c4b5 f6e4 f3d4 f8c5 c1e3 c8d7 b5c6 b7c6 b1d2 e4d2 d1d2 c5b6 e1c1 e8g8 e3g5 d8c8 g5f6 h7h6 d2d3 g7f6 e5f6 g8h8 d3e3 h8h7 e3d3');
INSERT INTO evaluations VALUES(52704,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52704,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(52704,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(52704,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52704,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(52704,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(52704,36,'cp 9','c6d5 c1f4 b8c6 e2e3 g8f6 f1b5 e7e6 b1c3 f8d6 g1e2 d6e7 e1g1 f6h5 f4e5 e8g8 h2h3 c8d7 b5d3 a8c8 g2g4 h5f6 e5f4 c6a5 e2g3 a5c4 d1e2 e7d6 f4d6 c4d6 f1c1 c8c7 g4g5 f6e8 h3h4 g7g6 a1b1');
INSERT INTO evaluations VALUES(53388,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53388,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(53388,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(53388,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(53388,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53388,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(53388,36,'cp 26','c1f4 b8c6 e2e3 g8f6 f1b5 e7e6 b1c3 f8d6 f4g3 e8g8 g1f3 d6g3 h2g3 c8d7 b5d3 h7h6 a2a3 d8e7 f3e5 d7e8 f2f4 c6e5 f4e5 f6h7 d1g4 h7g5 e1g1 a8d8 c3e2 d8c8 g4f4 f7f6 a1c1 c8c1 f1c1 f6e5 f4e5');
INSERT INTO evaluations VALUES(53375,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(53375,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53375,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53375,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53375,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53375,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(53375,36,'cp 16','g8f6 e2e3 b8c6 f1b5 e7e6 b1c3 f8d6 g1e2 d6e7 e1g1 f6h5 f4e5 e8g8 h2h3 c8d7 a1c1 a8c8 b5d3 h5f6 e5f4 c6a5 e2g3 a5c4 d1e2 e7d6 f4g5 h7h6 g5f6 d8f6 d3c4 d5c4 c3e4 f6e7 e4d6 e7d6 f1d1 b7b5 g3e4 d6e7');
INSERT INTO evaluations VALUES(45431,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45431,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(45431,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(45431,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(45431,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(45431,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(45431,36,'cp 14','e2e3 b8c6 f1b5 e7e6 b1c3 f8d6 g1h3 d6e7 a1c1 e8g8 e1g1 c8d7 b5d3 a8c8 d1e2 a7a6 h3g5 h7h6 g5f3 f6h5 f4e5 f7f6 e5g3 h5g3 h2g3 f6f5 c3a4 e7d6 a4c5 c8c7 a2a3 d8e7 c5d7 e7d7 f1d1 d7e7 b2b4 g7g5 b4b5');
INSERT INTO evaluations VALUES(45428,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(45428,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(45428,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(45428,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45428,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(45428,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45428,36,'cp 13','b8c6 f1b5 e7e6 b1c3 f8d6 g1e2 e8g8 e1g1 d6e7 e2c1 c6a5 c1b3 c8d7 b3a5 d8a5 b5d3 f8c8 a2a3 a5d8 h2h3 c8c6 f2f3 a8c8 g1h1 d7e8 d1d2 h7h6 a1c1 e7d6 f4d6 d8d6 c3e2 c6c1 f1c1 c8c1 e2c1 d6b6 d2c3 e8b5 h1g1 a7a6');
INSERT INTO evaluations VALUES(13139,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13139,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13139,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13139,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(13139,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(13139,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13139,36,'cp 23','f1b5 e7e6 b1c3 f8d6 f4g3 d6g3 h2g3 c8d7 g1e2 d8e7 e2f4 h7h6 a1c1 c6a5 b5d3 e8g8 f2f3 a8c8 e1f2 a5c4 d1e2 c4d6 e2d2 c8c6 c3e2 c6c1 h1c1 f8c8 c1c8 d7c8');
INSERT INTO evaluations VALUES(57037,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(57037,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(57037,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(57037,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(57037,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(57037,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(57037,36,'cp -22','d7d5 d1e2 g8f6 g1f3 f8e7 g2g3 e8g8 f1g2 b7b5 e1g1 c7c5 c1f4 b8c6 f3e5 c6e5 f4e5 c8b7 e4d5 b7d5 a2a4 d5g2 g1g2 d8d5 g2g1 f6d7 e5f4 b5b4 b1d2 e7f6 a1b1 a8c8 e2e4 d5e4 d3e4 c5c4 f1e1 f6d4 f4e3 d4e3 e1e3');
INSERT INTO evaluations VALUES(55904,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55904,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55904,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55904,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55904,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(55904,36,'cp 0','d1e2 g8f6 g1f3 f8e7 g2g3 e8g8 f1g2 b7b5 e1g1 c7c5 a2a4 b5b4 c2c4 b4c3 b2c3 b8c6 b1a3 h7h6 f1d1 a7a6 c1f4 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(55903,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55903,36,'cp 0','g8f6 g1f3 f8e7 g2g3 e8g8 f1g2 b7b5 e1g1 c7c5 a2a4 b5b4 c2c4 b4c3 b2c3 b8c6 b1a3 h7h6 f1d1 a7a6 c1f4 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(16144,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(16144,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(16144,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(16144,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(16144,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(16144,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(16144,36,'cp 50','d2d4 e5d4 e4e5 d7d5 c4b5 f6d7 b5c6 b7c6 f3d4 d7b6 d4c6 d8d7 c6d4 c7c5 e5e6 d7b7 e6f7 e8f7 d4e2 f8e7 a2a4 a7a5 b1c3 h8f8 e2f4 d5d4 c3b5 f7g8 c2c3 d4c3 b2c3 c8f5 f1e1 a8d8 d1h5 f8f7 f4g6 f5g6 h5g6 e7f8 g6h5 b6d5 c3c4 d5b4 h5g6 f7f6');
INSERT INTO evaluations VALUES(16156,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(16156,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(16156,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(16156,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(16156,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(16156,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(16156,36,'cp 39','e5d4 e4e5 d7d5 c4b5 f6d7 f1e1 d7c5 f3d4 c8d7 d4c6 b7c6 b5f1 f8e7 c1e3 e8g8 c2c4 d7e6 c4d5 c6d5 b1c3 c7c6 c3e2 e6d7 e2g3 c5e6 f2f4 e7b4 e1e2 b4c5 e3c5 e6c5 d1d2 d8b6 e2f2 a7a5 f4f5 b6b4 b2b3 b4d2 f2d2 a5a4');
INSERT INTO evaluations VALUES(16188,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(16188,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(16188,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(16188,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(16188,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(16188,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(16188,36,'cp 62','e4e5 d7d5 c4b5 f6d7 b5c6 b7c6 f3d4 d7b6 d4c6 d8d7 c6a5 f8e7 a2a3 e8g8 f1e1 f8e8 b1c3 d5d4 c3e4 d7d5 b2b4 c8f5 e4g3 f5g6 f2f4 e7h4 c1d2 a8d8 a1c1 h4g3 h2g3 g6e4 c2c4 d4c3 d2c3 e4g2 d1d5 g2d5 c3d4 d5e6 a5c6 d8d5 c1c5 d5c5 d4c5');
INSERT INTO evaluations VALUES(16110,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(16110,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(16110,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(16110,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(16110,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(16110,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(16110,36,'cp 56','d7d5 c4b5 f6d7 f1e1 d7c5 f3d4 c8d7 b5c6 b7c6 f2f4 c5e6 c1e3 e6d4 e3d4 f8e7 b1c3 e8g8 d1d3 c6c5 d4f2 d5d4 c3e4 a8b8 b2b3 d8c8 h2h3 b8b6 g2g4 d7b5 d3f3 f7f6 e5f6 e7f6 f4f5 f6e5 f2g3 e5g3 f3g3 c8d7 g3f3 b5c6 e4c5');
INSERT INTO evaluations VALUES(51804,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(51804,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(51804,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(51804,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(51804,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(51804,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(51804,36,'cp 102','g1f3 d7d5 b1d2 g8f6 e4e5 f6d7 f1d3 c6c5 c2c3 b8c6 e1g1 a7a5 d3c2 c5d4 c3d4 b7b5 f1e1 d7b6 d2f1 a5a4 f1g3 b6c4 g3h5 c6b4 c2b1 h7h6 d1e2 c8d7 c1e3 d8c7 e1c1 c7b6 h2h3 a8c8 c1c3 b4c6 b1d3 c6b4 f3e1 b4d3 e1d3 b5b4');
INSERT INTO evaluations VALUES(51803,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(51803,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(51803,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(51803,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(51803,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(51803,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(51803,36,'cp 104','d7d5 b1d2 g8f6 f1d3 c6c5 e4e5 f6d7 c2c3 b8c6 e1g1 a7a5 d3c2 c5d4 c3d4 b7b5 f1e1 a5a4 d2f1 d7b6 f1g3 h7h6 g3h5 c8d7 c1f4 c6b4 c2b1 b6c4 d1e2 d8b6 e1c1 e8d8 f4d2 c4d2 e2d2 a4a3 b2b3 a8c8 b1d3 b4c6 h2h4 d8c7 d2f4 d7e8');
INSERT INTO evaluations VALUES(53612,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(53612,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(53612,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(53612,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(53612,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(53612,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(53612,36,'cp 96','b1d2 g8f6 e4e5 f6d7 f1d3 c6c5 c2c3 b8c6 e1g1 a7a5 d3c2 c5d4 c3d4 b7b5 f1e1 d7b6 d2f1 a5a4 f1g3 c6b4 c2b1 h7h6 a2a3 b4c6 c1d2 b6c4 d2c3 f8e7 f3d2 g7g6 d2c4 b5c4 b1c2 c8d7 a1b1 c6a5 c2a4 d7a4 d1a4 d8d7 a4d1 h6h5 e1e3 a5b3 g3e2');
INSERT INTO evaluations VALUES(53613,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(53613,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(53613,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(53613,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(53613,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(53613,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(53613,36,'cp 95','b8d7 e4e5 c6c5 c2c3 g8e7 f1d3 a7a5 e1g1 e7c6 d3c2 c5d4 c3d4 b7b5 f1e1 d7b6 d2f1 f8b4 c1g5 d8c7 e1e2 h7h6 g5c1 a5a4 f1g3 b4f8 g3h5 b6c4 a1b1 c6b4 a2a3 b4c6 c1d2 c7b6 d2c3 c8d7 c2d3 e8c8 e2c2 c8b8 d1e2 d8c8 b1c1 b8a8 h2h4 f8e7 h5g7 h6h5');
INSERT INTO evaluations VALUES(50561,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(50561,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(50561,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(50561,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(50561,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(50561,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(50561,36,'cp 66','f1d3 d7d5 e4d5 d8d5 e1g1 c8d7 c1e3 b8c6 b1c3 d5e5 d4f3 e5c7 d1d2 g8f6 a1d1 f8e7 e3f4 c7a5 f1e1 e8g8 a2a3 a8c8 d2c1 e7a3 c3e4 a3e7 e4d6 c6b4 d6c8 f8c8 f3e5 d7e8 c1e3 a5c7 e3g3 b4d3');
INSERT INTO evaluations VALUES(50557,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(50557,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(50557,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(50557,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50557,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(50557,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(50557,36,'cp 55','f8c5 d4b3 c5a7 d1e2 g8e7 c1e3 e8g8 e3a7 a8a7 e2e3 b8c6 b1c3 d7d5 e4d5 e7d5 c3d5 e6d5 e1g1 f8e8 e3f4 c8e6 f1d1 h7h6 c2c3 a7a8 d1d2 a8c8 d3f1 d8c7 f4c7 c8c7 f2f3 e8c8 g1f2 g8f8 a2a3 c6e5');
INSERT INTO evaluations VALUES(56045,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(56045,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(56045,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(56045,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(56045,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(56045,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(56045,36,'cp 46','b8d7 c2c3 d7f6 e4f6 g8f6 g1f3 a7a6 f3e5 f8d6 f1e2 f6d5 e1g1 e8g8 d1d3 b7b6 e5c4 c8b7 c4d6 d8d6 d3f3 b7c6 f1e1 f8e8 c1d2 h7h6 a1d1 a6a5 f3g3 a8d8 f2f3 f7f6 b2b3 d6g3 h2g3 d8d7');
INSERT INTO evaluations VALUES(22398,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(22398,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(22398,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(22398,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(22398,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(22398,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(22398,36,'cp 43','g1f3 g8f6 e4f6 d7f6 c2c3 a7a6 c1f4 c7c5 f1e2 c5d4 d1d4 d8d4 f3d4 c8d7 e1c1 e8c8 e2f3 f8c5 f4e5 h7h5 h1e1 c5b6 h2h3 h5h4 b2b3 f6e8 e5h2 b6c7 h2c7 e8c7 c3c4 f7f6 f3g4 f6f5 g4f3 h8g8 c1b2 g7g5 a2a4 g5g4 h3g4 f5g4');
INSERT INTO evaluations VALUES(22397,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(22397,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(22397,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(22397,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(22397,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(22397,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(22397,36,'cp 53','g8f6 e4f6 d7f6 c2c3 a7a6 c1f4 c7c5 f1e2 c5d4 d1d4 d8d4 f3d4 c8d7 e1c1 e8c8 e2f3 f8c5 d1d2 c5b6 h1e1 b6c7 f4e5 h7h6 g2g3 c7e5 e1e5 g7g5 c3c4 g5g4 f3g2 h6h5 d4b3 c8b8 e5g5 h5h4');
INSERT INTO evaluations VALUES(15974,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(15974,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(15974,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(15974,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(15974,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(15974,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(15974,36,'cp 36','e4f6 d7f6 c2c3 c7c5 c1e3 d8c7 f1b5 c8d7 b5d7 f6d7 d4d5 e6d5 d1d5 d7f6 d5c4 f8d6 e3g5 c7e7 e1f1 e8g8 a1e1 e7d8 g2g3 h7h6 g5f6 d8f6 f1g2 f8d8 e1e4 a7a6 h2h4 b7b5 c4e2 f6f5 c3c4 d6f8 b2b3 f5d7 h1e1 d7b7 h4h5 b5c4 b3c4 a8c8 f3e5 c8c7');
INSERT INTO evaluations VALUES(15932,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(15932,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(15932,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(15932,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(15932,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(15932,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(15932,36,'cp 30','d7f6 c2c3 c7c5 f3e5 a7a6 c1e3 d8c7 d1a4 f6d7 f1b5 c5d4 b5d7 c8d7 a4d4 d7b5 a2a4 f8d6 e5f7 e8f7 a4b5 a6b5 e1e2 c7c6 h1d1 a8a1 d1a1 b5b4 a1d1 h8d8 d4b6 c6c4 e2e1 d6c7 b6b7 d8d1 e1d1 b4c3 b2b3 c4d5 b7d5 e6d5 d1c2 c7e5 e3c5 e5h2 c2c3 h2e5 c3d3 e5f6 b3b4 f7e6');
INSERT INTO evaluations VALUES(15783,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(15783,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(15783,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(15783,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(15783,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(15783,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(15783,36,'cp 46','c2c3 c7c5 c1e3 d8c7 f1b5 c8d7 b5d7 f6d7 d4d5 e6d5 d1d5 d7f6 d5c4 f8d6 e3g5 c7e7 e1f1 e8g8 a1e1 e7d8 g2g3 a8b8 a2a4 a7a6 f1g2 b7b5 a4b5 a6b5 c4d3 h7h6 h2h4 b8b6 e1d1 d6e7 d3d8 f8d8 d1d8 e7d8 h1d1 d8e7 g5f6 b6f6 d1d5 f6e6 f3e5 b5b4 f2f4 b4c3 b2c3 f7f6');
INSERT INTO evaluations VALUES(17040,30,'cp 220',NULL);
INSERT INTO evaluations VALUES(17040,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(17040,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(17040,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(17040,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(17040,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(17040,36,'cp 214','e4e5 c7c5 d4e3 f6d5 e3f3 d5c7 f1c4 d7d5 e5d6 c7e6 c4b5 c8d7 f3b7 d7b5 b7b5 d8d7 b1a3 f8d6 c1e3 a8b8 b5d7 e8d7 e1c1 d7c6 a3c4 d6c7 g2g3 h7h5 h2h4 h8e8 d1d3 b8d8 d3a3 c6b5 b2b3 a7a5 c1b2 d8a8 a3a4 a8d8 h1e1');
INSERT INTO evaluations VALUES(16944,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(16944,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(16944,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(16944,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(16944,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(16944,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(16944,36,'cp 217','c7c5 d4e3 f6d5 e3f3 d5c7 f1c4 d7d5 e5d6 c7e6 c4b5 c8d7 f3b7 d7b5 b7b5 d8d7 b1a3 f8d6 c1e3 d6e5 b5d7 e8d7 e1c1 d7c6 g2g3 e5c7 c2c4 a8d8 a3b5 c7b8 h1e1 h7h5 h2h4 a7a6 b5c3 b8e5 c3d5 h8e8 b2b3 a6a5 d1d3 e5d4 c1d2 a5a4 b3a4 d8a8 e3d4');
INSERT INTO evaluations VALUES(12638,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(12638,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(12638,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(12638,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(12638,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(12638,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(12638,36,'cp 224','d4e3 f6d5 e3f3 d5c7 f1c4 d7d5 e5d6 c7e6 c4b5 c8d7 f3b7 d7b5 b7b5 d8d7 b1a3 f8d6 c1e3 d6e5 b5d7 e8d7 e1c1 d7c6 g2g3 e5c7 c2c4 a8d8 a3b5 c7b8 h1e1 h7h5 h2h4 a7a6 b5c3 b8e5 c3d5 h8e8 d1d3 d8b8 b2b3 e6d4 e3d4 e5d4 e1e8');
INSERT INTO evaluations VALUES(12641,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(12641,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(12641,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(12641,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(12641,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(12641,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(12641,36,'cp 224','f6d5 e3f3 d5c7 f1c4 d7d5 e5d6 c7e6 c4b5 c8d7 f3b7 d7b5 b7b5 d8d7 b1a3 f8d6 c1e3 d6e5 b5d7 e8d7 e1c1 d7c6 g2g3 h8e8 d1d3 e5c7 a3c4 f7f5 b2b3 a8d8 d3d8 e6d8 c1d1 e8f8 h1e1 f5f4 e3d2 f4g3 f2g3 f8f2 e1e2 f2f1 d2e1 f1h1 d1d2 h7h5 d2d3 c6d7 e1c3 g7g6 a2a4 d8e6');
INSERT INTO evaluations VALUES(12663,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(12663,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(12663,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(12663,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(12663,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(12663,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(12663,36,'cp 212','e3f3 d5c7 f1c4 d7d5 e5d6 c7e6 c4b5 c8d7 f3b7 d7b5 b7b5 d8d7 b1a3 f8d6 c1e3 d6e5 b5d7 e8d7 e1c1 d7c6 g2g3 e5c7 a3c4 a8d8 h1e1 h8e8 c2c3 d8d1 c1d1 h7h5 e3d2 h5h4 d1c2 e8h8 e1e2 g7g6 b2b3 c6d7 c4e5 c7e5 e2e5 f7f5 c2d3 h4g3 h2g3 d7d6');
INSERT INTO evaluations VALUES(39112,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(39112,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(39112,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(39112,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(39112,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(39112,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(39112,36,'cp 117','d2d4 b8c6 c2c3 e5d4 c3d4 d6b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c7c6 f1e1 d8c7 d1h5 e5c3 b2c3 e8g8 d5c6 b7c6 b5c4 e7g6 h5c5 c7b6 c1e3 b6c5 e3c5 f8d8 a1d1 c8e6 c4e6 f7e6 d1d8 a8d8 g2g3 d8d5 c5a7 d5a5');
INSERT INTO evaluations VALUES(39105,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(39105,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(39105,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(39105,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(39105,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(39105,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(39105,36,'cp 120','b8c6 c2c3 e5d4 c3d4 d6b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c8d7 b5d3 e5c3 b2c3 e7d5 f1e1 d7e6 c3c4 d5f6 a1b1 d8d6 b1b7 e8g8 d3f1 f8b8 d1f3 b8b7 f3b7 a8e8 b7f3 e6d7 c1f4 d6b6 e1e8 f6e8 f3d5 d7c6 d5e5 b6b7 h2h4 h7h6 e5e7 b7b6 h4h5 b6b1 f4c7 e8c7 e7c7');
INSERT INTO evaluations VALUES(6776,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(6776,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(6776,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(6776,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(6776,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(6776,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(6776,36,'cp 105','c2c3 e5d4 c3d4 d6b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c8d7 b5d3 e5c3 b2c3 e7d5 c3c4 d5f6 f1e1 d7e6 a1b1 d8d6 b1b7 e8g8 d3f1 f8b8 b7b8 a8b8 d1f3 e6g4 f3e3 a7a6 c1d2 g4d7 d2b4 d6c6 e3d3 h7h6 h2h3 b8e8 e1e8 f6e8 c4c5 d7c8 b4c3 c8e6 d3a6 c6c5 a6a5 c5a5 c3a5 c7c6 a5b6 e6b3');
INSERT INTO evaluations VALUES(6774,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(6774,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(6774,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(6774,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(6774,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(6774,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(6774,36,'cp 123','e5d4 c3d4 d6b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c8d7 b5d3 e5c3 b2c3 e7d5 c3c4 d5f6 f1e1 d7e6 a1b1 e8g8 b1b7 d8d6 d3f1 f8e8 h2h3 f6e4 d1h5 e4d2 c1d2 d6d2 e1d1 d2c3 h5c5 g7g6 b7a7 a8a7 c5a7 e8c8 a7c5 c8a8');
INSERT INTO evaluations VALUES(21337,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(21337,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21337,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(21337,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(21337,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(21337,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(21337,36,'cp 33','b7c6 d2d4 e5d4 d1d4 g8f6 c1f4 c6c5 d4d3 f8e7 b1c3 e8g8 e1c1 c8e6 e4e5 d6e5 f3e5 d8d3 e5d3 c5c4 d3e5 e7b4 f4d2 b4d6 h1e1 f6d5 e5c6 d5c3 d2c3 f8e8 g2g3 f7f6 a2a3 e6d7 e1e8 a8e8 c6a5 d7g4 d1d4 g4e6 d4e4 f6f5 e4e2');
INSERT INTO evaluations VALUES(18274,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(18274,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18274,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(18274,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(18274,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(18274,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(18274,36,'cp 31','d2d4 e5d4 d1d4 g8f6 b1c3 f8e7 c1g5 e8g8 e1c1 c6c5 d4d3 c8b7 h1e1 h7h6 g5h4 f8e8 e4e5 f6g4 h4e7 e8e7 c3d5 e7e6 d3c4 g4e5 f3e5 d6e5 c4c5 b7d5 c5d5 d8g5 e1e3 a8e8 d5e4 f7f5 e4c4 g8h7 g2g3 f5f4 g3f4 g5f4 c4f4 e5f4 e3e6 e8e6');
INSERT INTO evaluations VALUES(18268,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(18268,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(18268,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18268,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(18268,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(18268,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(18268,36,'cp 8','e5d4 d1d4 g8f6 b1c3 f8e7 e4e5 f6g4 c1f4 e8g8 e1c1 d6e5 f3e5 d8d4 d1d4 e7f6 h1e1 f8e8 f2f3 g4e5 d4e4 c8f5 e4e3 e8e6 f4e5 a8e8 e5f4 f6c3 b2c3 e6e3 f4e3 h7h5 e3f2 e8e1 f2e1 f7f6 c3c4 f5e6 a2a4 e6c4 e1f2 c4f7 f2a7');
INSERT INTO evaluations VALUES(18278,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(18278,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(18278,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(18278,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18278,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(18278,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(18278,36,'cp 9','d1d4 g8f6 b1c3 f8e7 c1g5 h7h6 g5h4 c6c5 d4d3 e8g8 e1c1 c8b7 h1e1 f6d7 h4e7 d8e7 g2g4 a8e8 f3d2 e7h4 d3g3 h4g3 h2g3 d7e5 f2f3 b7c6 b2b3 a7a5 a2a4 e8e7 e1e3 f7f6 c3d5 c6d5 e4d5 f8e8 d1h1 g8f7 c1d1 e8h8 d1e2 h8b8 d2c4 e5c4');
INSERT INTO evaluations VALUES(18277,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18277,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18277,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(18277,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18277,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(18277,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(18277,36,'cp 23','g8f6 d4a4 c8b7 b1c3 f6d7 c1g5 f8e7 g5e7 d8e7 a4a5 d7b6 e1c1 e8g8 a5f5 c6c5 h1e1 e7e6 g2g4 a7a5 f5e6 f7e6 f3g5 f8e8 f2f4 h7h6 g5h3 b7c6 g4g5 a5a4 e4e5 d6d5 g5h6 g7h6 h3f2');
INSERT INTO evaluations VALUES(11875,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(11875,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(11875,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(11875,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(11875,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(11875,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(11875,36,'cp 6','d4a4 c8d7 b1c3 f8e7 e1g1 e8g8 a4c4 f8e8 e4e5 f6d5 c3d5 c6d5 c4d5 d7e6 d5a5 d6e5 f3e5 e7f6 f1e1 c7c5 c1d2 d8a5 d2a5 e6f5 g2g4 f6e5 g4f5 e5b2 e1e8 a8e8 a1b1 b2d4 c2c3 d4f6 b1b7 h7h5 b7a7 e8e5 a7a8 g8h7 g1g2');
INSERT INTO evaluations VALUES(9757,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(9757,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9757,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(9757,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9757,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(9757,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(9757,36,'cp 6','e1g1 e8g8 h2h3 d7d6 c3a4 c5b6 a4b6 a7b6 a2a4 c8d7 c1e3 c6e7 c4b3 e7g6 f1e1 c7c5 f3h2 d7e6 b3e6 f7e6 b2b4 f8f7 d1e2 d8d7 b4c5 b6c5 a4a5 g6e7 e3d2 e7c6 h2f1 d6d5 e4d5 e6d5');
INSERT INTO evaluations VALUES(9756,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(9756,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(9756,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(9756,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(9756,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9756,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9756,36,'cp 3','e8g8 h2h3 d7d6 c3a4 c5b6 c2c3 c6e7 a4b6 a7b6 c4b3 e7g6 f1e1 f8e8 b3c2 d6d5 d3d4 f6e4 f3e5 c8f5 e5g6 f5g6 c1f4 b6b5 a2a4 b5a4 a1a4 a8a4 c2a4 c7c6 a4c2 d8f6 f4g3 e4d6 c2g6 f6g6 g3d6 e8e1 d1e1 g6d6 e1e8 d6f8 e8f8 g8f8 f2f3 g7g6 h3h4');
INSERT INTO evaluations VALUES(4171,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4171,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4171,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4171,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4171,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(4171,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4171,36,'cp 14','h2h3 d7d6 c3a4 c5b6 a4b6 a7b6 a2a4 c8d7 c2c3 c6e7 c1e3 e7g6 f1e1 f8e8 f3h2 d7c8 d1b3 d8e7 h2f1 f6h5 d3d4 h5f4 e3f4 g6f4 f1e3 c8e6 d4e5 d6e5 c4e6 f4e6 a1d1 e8d8 d1d8 a8d8 e1d1 d8d1 b3d1 e6c5 d1d5 g8h7 e3g4 c5a4 g4e5 a4b2');
INSERT INTO evaluations VALUES(4167,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(4167,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4167,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4167,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4167,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4167,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4167,36,'cp 0','a7a6 c1e3 d7d6 e3c5 d6c5 a2a4 f8e8 c3d5 c8e6 d5e3 e6c4 e3c4 b7b6 g2g3 d8d7 g1h2 a8d8 b2b3 e8e6 d1e2 g8h7 c4e3 g7g6 a1e1 h7g8 e1d1 e6d6 f1e1 d6e6');
INSERT INTO evaluations VALUES(2643,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2643,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2643,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2643,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(2643,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(2643,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(2643,36,'cp 10','c3d5 d7d6 c2c3 c5a7 d5f6 d8f6 c4b3 c6e7 d3d4 e7g6 f1e1 f8e8 c1e3 c7c6 a2a4 a6a5 b3c2 c8e6 g1h2 d6d5 d4e5 g6e5 e3a7 e5f3 d1f3 f6f3 g2f3 a8a7 e4d5 c6d5 f3f4 a7a8 h2g3 g7g6 a1d1 a8a6 f4f5 g6f5 b2b3');
INSERT INTO evaluations VALUES(17037,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(17037,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(17037,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(17037,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(17037,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(17037,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(17037,36,'cp 118','d7d6 c1f4 f8e7 f2f3 e8g8 e1c1 c8e6 c1b1 f6d7 d4d2 e7f6 h2h4 f6e5 f4g5 f7f6 g5e3 c7c6 f3f4 e5c3 d2c3 d8e7 f1e2 f6f5 c3a3 f5e4 a3d6 f8e8 g2g4 d7b6 d6e7 e8e7 e3c5 e7e8 h1e1 b6d5 f4f5 e6f7 b1c1 a8d8 e2f1 b7b6 c5g1 b6b5 a2a3 d8d7 h4h5 h7h6 c2c4 b5c4 f1c4');
INSERT INTO evaluations VALUES(55625,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(55625,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(55625,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(55625,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(55625,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(55625,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(55625,36,'cp -55','e5f4 d2d4 g7g5 b1c3 f8g7 g2g3 b8c6 d4d5 c6e5 g3f4 g5f4 c1f4 c8g4 f1e2 e5f3 e2f3 d8f6 e1g1 g4h3 f3g2 h3g2 g1g2 f6g6 g2h3 g7c3 b2c3 e8c8 d1g4 d8d7 f1g1 f7f6 a1e1 g6e8 h3g2 b7b6 g4e6 e8e6 d5e6 d7e7 g2f3 e7e6 g1g7 e6e8 g7f7 g8e7 f7f6');
INSERT INTO evaluations VALUES(50609,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(50609,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(50609,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(50609,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50609,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(50609,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(50609,36,'cp 59','c2c4 b8c6 d2d4 c5d4 f3d4 e7e5 d4f5 g8f6 b1c3 d7d6 f5e3 g7g6 e3d5 f8g7 c1e3 f6d5 c3d5 e8g8 h2h4 c6d4 h4h5 c8e6 f1d3 a8c8 a2a4 e6d5 c4d5 d8a5 e3d2 a5b6 e1g1 b6b3 d2e3 b3d1 a1d1 g6h5 e3d4 e5d4 f2f4 h5h4 a4a5');
INSERT INTO evaluations VALUES(50596,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(50596,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(50596,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(50596,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(50596,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(50596,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(50596,36,'cp 65','b8c6 d2d4 c5d4 f3d4 g8f6 b1c3 e7e5 d4f5 d7d6 f5e3 g7g6 e3d5 f8g7 c1e3 c6d4 d5f6 g7f6 f1d3 c8e6 e1g1 e8g8 b2b3 d8a5 e3d2 a5d8 c3e2 d4c6 a1c1 f6g7 g1h1 f7f5 e4f5 g6f5 e2c3 c6b4 d3e2 a8c8 f2f4 d6d5 c4d5 b4d5 c3d5 d8d5 c1c8 f8c8');
INSERT INTO evaluations VALUES(17493,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(17493,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(17493,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(17493,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(17493,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(17493,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(17493,36,'cp 72','d2d4 c5d4 f3d4 e7e5 d4f5 d7d6 f5e3 g8f6 b1c3 c8e6 f1d3 g7g6 e3d5 e6d5 c3d5 f6d5 c4d5 c6d4 c1e3 f8g7 e1g1 e8g8 a1c1 f7f5 e3d4 f5e4 d3e4 e5d4 d1g4 d8e7 g4e6 f8f7 e4d3 a8e8 g2g3 g7h6 c1e1 e7e6 e1e6 e8e6 d5e6 f7e7 f1e1 h6g7');
INSERT INTO evaluations VALUES(21089,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(21089,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(21089,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(21089,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(21089,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21089,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(21089,36,'cp 61','d2d4 c8f5 c2c3 e7e6 f1e2 f7f6 f2f4 g7g5 e2h5 e8d7 f4g5 f6e5 g1f3 h7h6 c3c4 d5c4 f3e5 c6e5 d4e5 f5d3 b1c3 h6g5 h5g6 d7c8 g6d3 c4d3 d1g4 g8h6 g4e6 d8d7 e6d7 c8d7 c1g5 h6g4 a1d1 g4e5 g5f6 h8h5 f6e5 h5e5 e1d2 d7c6 d2d3 f8c5 g2g3 a8e8 g3g4 b7b5');
INSERT INTO evaluations VALUES(48656,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(48656,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48656,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(48656,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(48656,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48656,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(48656,36,'cp -12','f8d6 c3b5 d6b4 c2c3 b4a5 f1d3 e8g8 e1g1 c7c6 b5a3 c8g4 c1g5 a5c7 g5h4 b8d7 a3c2 d8b8 h4g3 f8e8 c2e3 g4h5 f1e1 b8d8 d1b3 a8b8 f3h4 h5g6 e3f5 f6e4 b3c2 g6f5 h4f5 d7f6 f5e3 c7g3 h2g3 g7g6 g3g4');
INSERT INTO evaluations VALUES(41703,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(41703,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41703,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41703,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(41703,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41703,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41703,36,'cp 0','c3b5 e8g8 b5d6 d8d6 h2h3 c8f5 f1e2 h7h6 e1g1 b8c6 c2c3 c6e7 f3e5 e7g6 g2g4 f5d7 e2d3 f6e4 f2f3 f7f6 e5g6 d6g3 g1h1 g3h3 h1g1 h3g3 g1h1');
INSERT INTO evaluations VALUES(10022,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(10022,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10022,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(10022,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(10022,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(10022,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10022,36,'cp 17','e1g1 e8g8 b5c6 d7c6 h2h3 f6d7 d2d3 f8e8 c1e3 c5d6 f3d2 d7f8 f2f4 e5f4 e3f4 f8g6 f4h2 d6h2 g1h2 d8d6 h2h1 c8e6 d2f3 f7f6 d1d2 c6c5 a1e1 e8f8 d2f2 b7b6 b2b3 a8e8 c3e2 e6d7 e2g3 g8h8 g3h5 d7e6 a2a4 e6g8 h5g3 d6d7');
INSERT INTO evaluations VALUES(10021,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(10021,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(10021,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10021,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10021,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10021,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10021,36,'cp 10','e8g8 f3e5 c6e5 d2d4 c5d6 f2f4 e5c6 e4e5 d6e7 e5f6 e7f6 d4d5 c6b4 d5d6 c7d6 c1e3 d6d5 e3f2 d7d6 a2a3 f6c3 b2c3 d8a5 a3b4 a5b5 a1a5 b5c6 d1d3 b7b6 a5d5 c8b7 d5g5 f7f6 g5g3 a8e8 f1d1 c6d7 h2h3 d6d5 f4f5 g8h8');
INSERT INTO evaluations VALUES(4263,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(4263,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(4263,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4263,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4263,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(4263,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(4263,36,'cp 19','b5c6 d7c6 h2h3 f8e8 d2d3 f6d7 c1e3 c5d6 f3d2 d7f8 f2f4 e5f4 e3f4 c8e6 d1f3 f8g6 f4d6 c7d6 f3f2 e8f8 a1e1 f7f5 e4f5 e6f5 d2f3 d6d5 f2c5 h7h6 a2a4 d8b6 c5b6 a7b6 f3d4 f5d7 b2b4 f8f1 g1f1 c6c5 d4f3 c5b4 c3d5');
INSERT INTO evaluations VALUES(4249,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(4249,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4249,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(4249,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(4249,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(4249,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(4249,36,'cp 9','d7c6 h2h3 f8e8 d2d3 c5d6 a2a4 h7h6 f3h2 c6c5 d1f3 c8e6 b2b3 a7a6 f1e1 g8h7 c3d5 f6g8 h2g4 e6d5 e4d5 d8d7 g4e3 e8f8 f3f5 d7f5 e3f5 a8e8 c1d2 g8e7 f5e7 e8e7 a4a5 f7f5 g1f1 h7g8 a1a4 e7e8 f1e2 h6h5 a4h4');
INSERT INTO evaluations VALUES(3906,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(3906,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(3906,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(3906,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3906,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3906,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3906,36,'cp 9','d2d3 f8e8 h2h3 f6d7 c1e3 c5d6 f3d2 d7f8 f2f4 e5f4 e3f4 f8g6 f4d6 c7d6 d1f3 c8e6 f3f2 f7f6 a2a4 b7b6 f1e1 d8d7 d2f3 d6d5 a4a5 c6c5 e4d5 e6d5 c3d5 d7d5 f3d2 a8d8 b2b3 g6e5 a5b6 a7b6 d2c4 e5c4 b3c4 e8e1 a1e1');
INSERT INTO evaluations VALUES(3907,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(3907,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(3907,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(3907,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(3907,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3907,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3907,36,'cp 12','f8e8 h2h3 f6d7 c1e3 c5d6 f3d2 d7f8 f2f4 e5f4 e3f4 f8g6 f4d6 c7d6 d1f3 e8f8 f3e3 c8e6 a2a4 d8a5 c3e2 a5b4 e2d4 a8e8 d2f3 h7h6 e3e1 b4c5 e1f2 f7f5 e4f5 e6f5 a1e1 e8e1 f1e1');
INSERT INTO evaluations VALUES(56019,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(56019,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(56019,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(56019,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(56019,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(56019,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(56019,36,'cp 58','e2e4 c7c5 d4d5 e6d5 d1d5 d8d5 e4d5 g8f6 c1f4 f6h5 f4e3 c8f5 e1c1 b8d7 g1f3 e8c8 d5d6 h5f6 f1c4 f6g4 c4f7 g7g6 h1e1 d7e5 f3e5 g4e5 f7d5 f8d6 d5e4 f5e4 c3e4 d6c7 e4c5 b7b6 f2f4 e5c4 c5e6 d8d1 c1d1 c4b2 d1c2 b2c4 e6c7 c4e3 e1e3 c8c7 e3e7 c7d6 e7a7 h8f8 c2d3 f8f4');
INSERT INTO evaluations VALUES(56027,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(56027,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(56027,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(56027,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(56027,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(56027,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(56027,36,'cp 69','c7c5 d4d5 a7a6 f1c4 b7b5 c4b3 c5c4 b3c2 e6e5 b2b3 d8c7 b3c4 c7c4 c1b2 b8d7 c2d3 c4b4 a1b1 g8f6 g1f3 f8d6 a2a4 d7c5 e1g1 e8g8 a4b5 c5d3 d1d3 a6b5 c3b5 c8a6 b2e5 a6b5 d3b5 b4b5 b1b5 f6e4 e5d4 a8b8 f1b1 f7f6 g2g3');
INSERT INTO evaluations VALUES(54100,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(54100,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(54100,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(54100,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(54100,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(54100,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(54100,36,'cp 75','d4d5 a7a6 f1c4 b7b5 c4b3 c5c4 b3c2 e6e5 b2b3 d8c7 b3c4 c7c4 c1b2 g8f6 c2d3 c4b4 a1b1 f8d6 g1f3 b8d7 a2a4 d7c5 e1g1 e8g8 a4b5 c5d3 d1d3 a6b5 c3b5 b4e4 d3e4 f6e4 b5d6 e4d6 f3e5 f8e8 e5c6 c8a6 f1c1 a6d3 b1a1 a8a1 c1a1 d3e4 a1d1');
INSERT INTO evaluations VALUES(54125,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(54125,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(54125,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(54125,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(54125,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(54125,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(54125,36,'cp 64','g8f6 f1c4 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 a2a4 e7d6 f1d1 c8g4 h2h3 g4f3 g2f3 b8d7 a4a5 f8b8 g1f1 b7b5 a5b6 d7b6 c4e2 d6e5 d5d6 b8d8 c3e4 f6e4 f3e4 d8d6 d1d6 e5d6 f2f4 b6d7 a1a6 a8a6 e2a6 d6e7 a6b5 d7f8 f1e2 f8e6 b5c4 e7f6');
INSERT INTO evaluations VALUES(46132,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(46132,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(46132,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(46132,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(46132,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46132,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(46132,36,'cp 61','f1c4 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 d5d6 e7d8 c4d5 a8a7 f1d1 b8d7 f3e5 d7e5 f4e5 c8g4 f2f3 g4e6 a1c1 e6d5 c3d5 f6d5 d1d5 f7f6 e5g3 a7d7 g3f2 c5c4 b2b3 c4b3 a2b3 g8f7 c1c6 f7e6 d5d3 a6a5 f2c5 f8e8 c6c8 a5a4 b3a4 b5a4');
INSERT INTO evaluations VALUES(46127,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(46127,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(46127,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(46127,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(46127,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46127,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(46127,36,'cp 62','e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 d5d6 e7d8 c4d5 a8a7 f1d1 b8d7 f3e5 d7e5 f4e5 f6d5 c3d5 a7d7 d5c7 c5c4 a2a4 b5a4 d1c1 d8f6 f2f4 f6d8 c1c4 f7f6 e5c3 d7d6 c3b4 d6d3 b4f8 g8f8 a1a4 d3d1 g1f2 d1d2 f2e3');
INSERT INTO evaluations VALUES(46412,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(46412,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(46412,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(46412,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(46412,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46412,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(46412,36,'cp 63','e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 f1e1 e7d8 c4d3 c5c4 d3f1 b8d7 f4d6 f8e8 a2a4 d8a5 a4b5 e8e1 a1e1 a6b5 e1a1 c8b7 c3b5 b7d5 f3d4 f6e8 d6f4 d7c5 d4f5 e8f6 f5e7 g8h8 b5d6 c5d3 e7d5 f6d5 d6c4 d5f4 a1a5 a8a5');
INSERT INTO evaluations VALUES(46404,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(46404,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(46404,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(46404,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(46404,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(46404,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(46404,36,'cp 62','a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e7e2 c4e2 e8g8 f3d2 d6e5 d2c4 e5d4 a2a4 b8d7 f1d1 d7e5 c4e5 d4e5 c1e3 e5d6 a4a5 c8d7 f2f3 h7h5 e3f2 a8b8 b2b3 f8e8 e2c4 d6e5 a1c1 e5f4 c1c2 f4c7 f2g3 c7g3 h2g3 b7b5 a5b6');
INSERT INTO evaluations VALUES(15481,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(15481,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(15481,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(15481,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(15481,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(15481,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(15481,36,'cp 151','f3g5 d6d5 e4d5 c6a5 c4b5 c8d7 d1e2 f8e7 b5d7 d8d7 e1g1 e8g8 e2e5 a8e8 g5e4 a5c4 e4f6 e7f6 e5f4 c4d6 a1b1 d6f5 f4g4 g7g6 g4d1 h7h5 d2d3 f6e5 a2a4 e5c3 b2c3 d7d5 b1b5 c7c5 c3c4 d5c6');
INSERT INTO evaluations VALUES(15472,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(15472,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(15472,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(15472,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(15472,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(15472,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(15472,36,'cp 150','d6d5 e4d5 c6a5 c4b5 c8d7 d1e2 f8e7 g5f3 e8g8 b5d7 d8d7 f3e5 d7d6 d2d4 f6d5 c3e4 d6e6 e1g1 c7c5 e5f3 c5d4 f3d4 e6g6 e4g3 f8e8 e2b5 a5c6 d4f5 a8d8 b5b7 e7c5 b7b5 c5b6 b5b3 d5e7 f5h4 g6f6 h4f3 c6e5 f3e5 f6e5 c1e3 b6e3 f2e3');
INSERT INTO evaluations VALUES(15080,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(15080,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(15080,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(15080,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(15080,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(15080,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(15080,36,'cp 164','e4d5 c6a5 c4b5 c8d7 d1e2 f8e7 g5f3 e8g8 b5d7 d8d7 f3e5 d7d6 d2d4 f6d5 c3e4 d6e6 e1g1 c7c5 e5f3 a5c6 d4c5 f7f5 e4g3 e6e2 g3e2 e7c5 e2f4 d5f4 c1f4 a8e8 f1e1 e8e4 f4d2 f8d8 g1f1 c6d4 f3d4 e4d4 d2e3 d4c4 e3c5 c4c5 c2c3 d8d2 e1e2');
INSERT INTO evaluations VALUES(14979,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(14979,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(14979,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(14979,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(14979,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(14979,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(14979,36,'cp 157','c6a5 c4b5 c8d7 d1e2 f8e7 g5f3 e8g8 b5d7 d8d7 f3e5 d7d6 d2d4 f6d5 c1d2 a5c6 c3b5 d6d8 e1c1 a7a6 e5c6 b7c6 b5c3 d5c3 d2c3 f8e8 e2f3 d8d5 f3d5 c6d5 h1e1 f7f6 g2g3 g8f7 c3a5 e7d6 e1e8 a8e8 d1e1 e8e1 a5e1 g7g5 a2a3 h7h5 e1b4 d6b4 a3b4 g5g4 b2b3 c7c6 c1d2');
INSERT INTO evaluations VALUES(15834,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(15834,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(15834,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(15834,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(15834,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(15834,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(15834,36,'cp 162','c4b5 c8d7 d1e2 f8e7 g5f3 e8g8 b5d7 d8d7 f3e5 d7d6 d2d4 f6d5 c1d2 a5c6 c3b5 d6d8 e1g1 a7a6 e5c6 b7c6 b5c3 e7f6 e2c4 d5c3 d2c3 d8d5 c4d5 c6d5 a1d1 f6e7 c3a5 f7f6 g2g3 e7d6 b2b3 a8e8 f1e1 g8f7 g1g2 e8e1 d1e1');
INSERT INTO evaluations VALUES(47003,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(47003,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(47003,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(47003,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(47003,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47003,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47003,36,'cp 22','g1f3 c7c5 d4c5 d8a5 b1c3 e7e5 f4d2 a5c5 e2e4 f8e7 f1d3 c8e6 e1g1 e8g8 d2e3 c5c8 f3g5 e6d7 f2f4 f6g4 c3d5 c8d8 e3d2 e5f4 d5e7 d8e7 d2f4 h7h6 g5f3 b8c6 d1d2 f8e8 a1e1 a8d8 h2h3 g4e5 d3e2 e5f3 f1f3 e7e4 f4h6');
INSERT INTO evaluations VALUES(47001,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(47001,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(47001,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(47001,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(47001,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(47001,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(47001,36,'cp 16','c7c5 d4c5 d8a5 b1c3 e7e5 f4d2 a5c5 e2e4 f8e7 f1d3 e8g8 e1g1 c8e6 d2e3 c5c8 f3g5 e6d7 f2f4 f6g4 c3d5 c8d8 d5e7 d8e7 e3d2 e5f4 d2f4 b8c6 d1d2 h7h6 g5f3 f8e8 a1e1 a7a6 b2b3 b7b5 h2h3 g4e5 e1d1 e5d3 d2d3 a8d8 c2c3 c6e5 d3d6 e5f3 g2f3');
INSERT INTO evaluations VALUES(22956,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(22956,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(22956,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(22956,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(22956,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(22956,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(22956,36,'cp -83','g8f6 d1e2 f8c5 c1e3 c5e3 e2e3 e8g8 b1d2 d7d5 e1c1 f8e8 f2f3 d8e7 g1e2 d5e4 d2e4 f6e4 f3e4 e7e4 e3e4 e8e4 e2c3 e4e5 f1b5 c8e6 b5c6 b7c6 h1e1 e5f5 b2b3 a7a5 d1d2 g8f8 g2g3 a8e8 e1e4 h7h6 d2e2 g7g5');
INSERT INTO evaluations VALUES(16717,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(16717,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(16717,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(16717,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(16717,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(16717,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(16717,36,'cp -89','f1d3 d7d5 e4d5 d8d5 g1f3 f8b4 b1c3 c8g4 d1e2 e8f8 e1g1 d5h5 e2d1 c6d4 d3e2 d4e2 d1e2 a8e8 e2b5 b4c3 b5h5 g4h5 b2c3 h5f3 g2f3 f8e7 c1g5 e7d7 g5f6 g7f6 a1d1 d7c6 f1e1 e8e5 e1e4 e5e4 f3e4 h8e8 d1d4 e8e5 g1f1 e5a5 a2a4 a5c5 d4d3 c5c4 d3f3 c6d7 f3f6 d7e7 f6h6 c4a4 h6h7');
INSERT INTO evaluations VALUES(16703,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(16703,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(16703,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(16703,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(16703,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(16703,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(16703,36,'cp -94','d7d5 e4d5 d8d5 g1f3 c8g4 b1c3 f8b4 d1e2 e8f8 e1g1 d5h5 e2d1 c6d4 d3e2 d4e2 d1e2 a8e8 e2b5 b4c3 b5h5 g4h5 b2c3 h5f3 g2f3 f8e7 c1g5 e7d7 g5f6 g7f6 a1d1 d7c6 f1e1 e8e5 e1e4 e5e4 f3e4 h8e8 f2f3 e8e5 d1d8 f6f5 d8h8 f5e4 f3e4 h7h5 g1f1 e5c5 f1e2 c5c3 h8h6 c6d7');
INSERT INTO evaluations VALUES(15059,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(15059,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(15059,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(15059,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(15059,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(15059,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(15059,36,'cp -75','e4d5 d8d5 g1f3 f8b4 c2c3 b4d6 d1e2 c8e6 b1d2 d5h5 d2e4 f6e4 e2e4 e8c8 c1e3 h8e8 e4a4 d6f4 e1c1 f4e3 f2e3 h5c5 h1e1 g7g6 a4b5 c5e7 b5a4 c8b8 a2a3 e6c8 a4f4 f7f6 e3e4 h7h5 d3c2 g6g5 f4e3 d8d1 e1d1 h5h4 f3d4 c6d4');
INSERT INTO evaluations VALUES(14978,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(14978,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(14978,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(14978,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(14978,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(14978,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(14978,36,'cp -91','d8d5 g1f3 f8b4 b1c3 c8g4 d1e2 e8f8 e1g1 d5h5 e2d1 c6d4 d3e2 d4e2 d1e2 a8e8 e2b5 b4c3 b5h5 g4h5 b2c3 h5f3 g2f3 f8e7 a1b1 b7b6 f1d1 e8d8 d1e1 e7d7 c1g5 h8e8 g5f6 g7f6 g1f1 e8e1 f1e1 d8e8 e1d1 a7a5 d1d2 e8e5 c3c4 d7c6 d2c3 e5e2');
INSERT INTO evaluations VALUES(49591,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(49591,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(49591,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49591,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(49591,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(49591,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(49591,36,'cp 34','f6e4 f3e5 d7d5 f1d3 b8d7 e5d7 c8d7 e1g1 f8d6 b1c3 e4c3 b2c3 d8h4 f1e1 d7e6 g2g3 h4h3 c3c4 d5c4 d3c4 e8g8 c4e6 f7e6 a1b1 a8e8 b1b5 f8f5 b5f5 h3f5 c2c4 b7b6 d1e2 a7a5 g1g2 a5a4 h2h4 h7h5 c1b2 g8f7 e1d1 a4a3');
INSERT INTO evaluations VALUES(49902,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(49902,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(49902,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(49902,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(49902,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(49902,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(49902,36,'cp 28','f3e5 d7d5 f1d3 b8d7 e5d7 c8d7 e1g1 f8d6 b1c3 e4c3 b2c3 d8h4 g2g3 h4h3 f1e1 d7e6 c3c4 d5c4 d3c4 e8g8 c4e6 f7e6 a1b1 b7b6 d1e2 a8e8 b1b5 f8f5 b5f5 h3f5 c2c4 e8e7 c1b2 h7h5 h2h4 f5h3 e2f3 h3f5 g1g2 f5f3 g2f3 g8f7 e1e2 c7c5');
INSERT INTO evaluations VALUES(49825,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(49825,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(49825,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(49825,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49825,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(49825,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(49825,36,'cp 23','d7d5 f1d3 b8d7 e5d7 c8d7 e1g1 f8d6 b1c3 e4c3 b2c3 d8h4 g2g3 h4h3 f1e1 d7e6 c3c4 d5c4 d3c4 e8g8 c4e6 f7e6 a1b1 b7b6 b1b5 f8f5 d1e2 a8e8 b5f5 h3f5 c2c4 e8e7 h2h4 h7h6 c1b2 a7a6 a2a4 e7f7 b2c1 f5f3 c1e3');
INSERT INTO evaluations VALUES(49050,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(49050,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(49050,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(49050,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49050,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(49050,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(49050,36,'cp 34','f1d3 b8d7 e5d7 c8d7 e1g1 f8d6 b1c3 e4c3 b2c3 d8h4 f1e1 d7e6 g2g3 h4h3 c3c4 d5c4 d3c4 e8g8 c4e6 f7e6 a1b1 b7b6 b1b5 f8f5 d1e2 a8e8 a2a4 c7c6 b5f5 h3f5 c2c4 e8e7 e1d1 c6c5 d4c5 d6c5 c1e3 h7h6 d1d8 g8h7 e3d4 c5d4 d8d4');
INSERT INTO evaluations VALUES(49048,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(49048,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(49048,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(49048,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(49048,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(49048,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(49048,36,'cp 23','b8d7 e5d7 c8d7 e1g1 f8d6 b1c3 e4c3 b2c3 d8h4 g2g3 h4h3 c3c4 d5c4 f1e1 d7e6 d3c4 e8g8 c4e6 f7e6 a1b1 b7b6 b1b5 f8f5 d1e2 a8e8 a2a4 h7h6 b5f5 h3f5 c2c4 e8e7 c1b2 a7a6 c4c5 b6c5 e2a6 f5d5 d4c5 d6c5 a4a5');
INSERT INTO evaluations VALUES(55610,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55610,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(55610,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55610,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(55610,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55610,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(55610,36,'cp 17','g8f6 b1c3 c7c5 f1c4 f8e7 d2d3 b8c6 c1g5 a8b8 e1g1 e8g8 a2a4 a7a6 g5d2 c8e6 f1e1 c6d4 a4a5 h7h6 f3d4 c5d4 c3d5 f6d5 e4d5 e6f5 f2f4 e7h4 e1e2 f8e8 f4e5 d6e5 d2b4 b7b5 a5b6 d8b6 b4a5');
INSERT INTO evaluations VALUES(47894,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(47894,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(47894,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47894,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(47894,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47894,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47894,36,'cp 31','b1c3 c7c5 f1c4 f8e7 d2d3 b8c6 c1g5 e8g8 g5f6 e7f6 c3d5 f6g5 c2c3 g5h6 e1g1 c8e6 d1e2 a8c8 f1d1 g8h8 a2a3 a7a6 d5e3 h6e3 e2e3 c6e7 c4e6 f7e6 b2b4 c5b4 c3b4 e7c6 a1c1 h7h6 d3d4 e5d4 f3d4 c6d4 e3d4 c8c1');
INSERT INTO evaluations VALUES(47865,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(47865,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(47865,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47865,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47865,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(47865,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(47865,36,'cp 17','c7c5 f1c4 f8e7 d2d3 b8c6 c1g5 c8e6 e1g1 e8g8 g5f6 e7f6 c3d5 f6g5 c2c3 g5h6 d1e2 a8c8 f1d1 c6e7 d5e7 d8e7 a2a3 c8c7 b2b4 a7a6 c4e6 e7e6 b4c5 c7c5 c3c4 b7b5 c4b5 c5b5 a3a4 b5b4 d3d4 e6c4 e2c4 b4c4 a4a5 f8c8');
INSERT INTO evaluations VALUES(45911,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(45911,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(45911,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(45911,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45911,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45911,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45911,36,'cp 24','f1c4 f8e7 d2d3 b8c6 f3h2 e8g8 h2g4 f6d7 a2a4 d7b6 c4a2 g8h8 g4e3 e7g5 e1g1 c8e6 a2d5 e6d5 c3d5 b6d5 e3d5 g5c1 a1c1 f7f5 e4f5 f8f5 d1g4 d8d7 d5e3 f5f7 g4d7 f7d7 c2c3 h7h6 c1d1 d6d5 a4a5 h8h7');
INSERT INTO evaluations VALUES(45898,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(45898,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(45898,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45898,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45898,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(45898,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(45898,36,'cp 25','f8e7 d2d3 e8g8 c1g5 b8c6 g5f6 e7f6 c3d5 f6g5 c2c3 g5h6 f3h2 c6e7 e1g1 c8e6 d5e7 d8e7 d1b3 a8b8 c4e6 e7e6 b3e6 f7e6 a2a4 a7a5 h2f3 b7b5 a4b5 b8b5 c3c4 b5b3 a1a5 f8b8 f1d1 b3b2 a5a6 b2b6 a6b6 b8b6 d1a1 g8f7 g2g3');
INSERT INTO evaluations VALUES(52412,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(52412,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(52412,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(52412,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(52412,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(52412,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(52412,36,'cp -38','e7e6 h5e2 g7g6 f2f4 f8g7 g1f3 g8e7 c2c3 e8g8 d2d3 d7d6 a2a4 b8c6 e2f2 b7b6 f1e2 d6d5 e1g1 d5d4 c1d2 c6a5 a1a3 d4c3 d2c3 e7c6 b1d2 c8b7 h2h3 f7f6 d3d4 c6d4 f3d4 c5d4 f2d4 d8d4 c3d4 a8d8');
INSERT INTO evaluations VALUES(51914,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(51914,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(51914,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(51914,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(51914,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(51914,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(51914,36,'cp -37','h5e2 b8c6 g2g3 g8e7 d2d3 g7g6 f1g2 f8g7 g1f3 e8g8 c2c3 b7b6 b1a3 c8a6 h2h4 h7h6 e1g1 d8c7 e4e5 d7d6 e5d6 c7d6 a3c4 a6c4 d3c4 a8d8 c1e3 e6e5 f1d1 d6c7 d1d8 f8d8 a1d1 f7f5 d1d8 c6d8 e3c1 e5e4 f3d2');
INSERT INTO evaluations VALUES(42733,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(42733,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(42733,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(42733,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(42733,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(42733,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(42733,36,'cp 59','f3e5 e8g8 f1e2 f8e8 e5d3 b4c3 d2c3 f6e4 e1g1 d7d6 f2f3 e4g5 d3f4 a7a5 f4h5 b8d7 f1e1 a5a4 e2f1 g5e6 c1e3 d8h4 h5g3 d7e5 b2b4 b7b6 d1d5 c8d7 a1d1 h4f6 d5d2 e5g6 e3f2 a8d8 b4b5 h7h6');
INSERT INTO evaluations VALUES(42710,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(42710,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(42710,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(42710,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(42710,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(42710,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(42710,36,'cp 46','e8g8 f1e2 b4c3 d2c3 f8e8 e5d3 f6e4 e1g1 d7d6 f2f3 e4g5 d3f4 a7a5 f4h5 h7h6 b2b3 a5a4 e2d3 a4b3 c2b3 g5e6 d1d2 d6d5 d3c2 d8g5 d2g5 h6g5 f1e1 e8d8 a2a4 d5d4 h5g3 d4c3');
INSERT INTO evaluations VALUES(37258,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(37258,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(37258,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(37258,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(37258,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(37258,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(37258,36,'cp 41','f1e2 b4c3 d2c3 f6e4 e5d3 d7d6 e1g1 f8e8 f2f3 e4g5 d3f4 a7a5 a2a4 h7h6 f4h5 b8d7 f1e1 d7f6 h5g3 b7b6 b2b3 c8b7 d1d2 f6d7 c1b2 d7c5 b3b4 c5d7 a1d1 b7c6 b4b5 c6b7 c3c4 g5e6 f3f4 d7c5 f4f5 e6g5 g3h5 f7f6 e2f1 e8e1 d1e1 c5a4 b2d4');
INSERT INTO evaluations VALUES(39572,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(39572,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(39572,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(39572,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(39572,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(39572,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(39572,36,'cp 64','g8f6 e2e4 d7d6 f1e2 b8a6 h2h4 e7e5 d4d5 a6c5 d1c2 e8g8 c1e3 h7h5 f2f3 c7c6 c2d2 a7a5 g1h3 c6d5 c4d5 c8d7 h3f2 a5a4 a1c1 d8a5 c3d1 f8c8 d2a5 a8a5 f2d3 f6e4 f3e4 c5e4 c1c8 d7c8 e3b6 a5d5 d3f2 e4c5 d1c3 d5d4 e1g1 g7h6');
INSERT INTO evaluations VALUES(22617,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(22617,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(22617,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(22617,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(22617,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(22617,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(22617,36,'cp 79','g1f3 d7d5 e4e5 f7f6 f1b5 c8d7 e1g1 d8e7 f1e1 c6e5 f3e5 f6e5 b5d7 e7d7 d1h5 g7g6 h5e5 f8g7 e5e3 g8e7 c2c3 e7f5 e3e6 d7e6 e1e6 e8d7 e6e1 h8e8 e1e8 a8e8 g1f1 b7b6 g2g3 c7c5 d4c5 b6c5 c1f4 g7f6 h2h4 a7a5 b1d2 a5a4 a1c1 f5d6 b2b3 a4b3 a2b3 d6b5 c3c4 b5c3 c4d5 c3d5');
INSERT INTO evaluations VALUES(22616,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22616,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(22616,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(22616,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(22616,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(22616,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(22616,36,'cp 90','d7d5 b1d2 g8f6 e4e5 f6d7 d2b3 f8e7 c1f4 e8g8 c2c3 a7a5 a2a4 f7f6 e5f6 g7f6 f1b5 e7d6 f4g3 f8f7 e1g1 c6e7 b5d3 f7g7 f1e1 d7f8 d1d2 g8h8 f3h4 b7b6 d3c2 e7g6 h4g6 f8g6 b3c1 d8g8 c2g6 g7g6 c1d3 g6g7 d2e3 c8d7 g1h1 a8e8 g3d6 c7d6 d3f4 e6e5 f4h5');
INSERT INTO evaluations VALUES(21442,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(21442,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21442,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(21442,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(21442,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(21442,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(21442,36,'cp 104','b1d2 g8f6 e4e5 f6d7 d2b3 f8e7 c1f4 a7a5 a2a4 e8g8 f1b5 f7f6 e5f6 d7f6 e1g1 f6e4 f4g3 e7d6 c2c3 c8d7 f1e1 d7e8 b3d2 e4g3 h2g3 d8f6 e1e3 e8h5 d1e2 h5g4 a1e1 a8e8 d2b3 f6h6 b3c5 d6c5 d4c5 g4f3 e3f3');
INSERT INTO evaluations VALUES(21443,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(21443,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(21443,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(21443,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(21443,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(21443,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(21443,36,'cp 79','g8f6 e4e5 f6d7 d2b3 f8e7 c2c3 e8g8 c1f4 a7a5 a2a4 f7f6 e5f6 g7f6 f1b5 f8f7 e1g1 e7d6 f4h6 c6e7 f3h4 d7f8 d1h5 f8g6 b5d3 g6h4 h5h4 g8h8 a1e1 e7g8 h6f4 d6f4 h4f4 f7g7 b3c1 e6e5 d4e5 f6e5 e1e5 c8h3 g2g3 h3f1 d3f1 d8d7 f4d4 a8e8 e5d5');
INSERT INTO evaluations VALUES(14873,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(14873,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(14873,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(14873,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(14873,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(14873,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(14873,36,'cp 90','e4e5 f6d7 d2b3 f8e7 c1f4 e8g8 c2c3 a7a5 a2a4 f7f6 e5f6 g7f6 f1b5 g8h8 f4g3 f8g8 e1g1 e7d6 d1d2 d7f8 b3c1 c6e7 b5d3 f8g6 h2h4 g8g7 c1e2 e6e5 d3g6 e7g6 d4e5 g6e5 d2d5 e5f3 d5f3 c8g4 f3d3 g4e2 d3e2 d6g3 f2g3');
INSERT INTO evaluations VALUES(14888,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(14888,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(14888,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(14888,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(14888,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(14888,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(14888,36,'cp 96','f6d7 d2b3 f8e7 c1f4 e8g8 c2c3 a7a5 f1b5 a5a4 b3d2 f7f5 e5f6 e7f6 e1g1 e6e5 b5c6 b7c6 d4e5 f6e7 f3d4 d7b8 d4e2 c6c5 f4g3 b8c6 h2h3 d5d4 f1e1 a4a3 b2a3 d4d3 e2f4 a8a3 d2b1 a3a6 d1d3 d8d3 f4d3 c5c4 d3f4 e7g5 e1e4 c6e7 e4c4');
INSERT INTO evaluations VALUES(15894,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(15894,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(15894,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(15894,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(15894,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(15894,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(15894,36,'cp 70','d2b3 f8e7 c1f4 e8g8 c2c3 a7a5 a2a4 f7f6 e5f6 g7f6 f1b5 f8f7 e1g1 e7d6 d1d2 d7f8 f4g3 c6e7 b3c1 f8g6 b5d3 d6g3 h2g3 e6e5 d4e5 f6e5 c3c4 c7c6 c4d5 c6d5 f1e1 d8d6 f3g5 f7g7 g5e4 d6c6 e4c3 c8e6');
INSERT INTO evaluations VALUES(47448,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(47448,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(47448,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(47448,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(47448,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(47448,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(47448,36,'cp 82','d5f6 f1d3 e7e5 d4e5 f6g4 d3b5 c7c6 d1d8 e8d8 b5e2 g4e5 b1c3 f8c5 h2h3 b7b5 c1d2 c5b6 f2f4 e5d7 e4e5 a7a5 c3e4 d7c5 e4c5 b6c5 g1f3 d8e7 a2a4 b5b4 e1c1 c8e6 f4f5 e6f5 d2g5 f7f6 e5f6 g7f6 g5f4 f5e6 f3d4 c5d4 d1d4 h8d8 h1d1 b8d7 f4d6 e7f7');
INSERT INTO evaluations VALUES(47374,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(47374,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(47374,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(47374,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(47374,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47374,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(47374,36,'cp 84','f1d3 e7e5 d4e5 f6g4 d3b5 c7c6 d1d8 e8d8 b5e2 g4e5 b1c3 f8c5 c1d2 e5g4 g1h3 f7f6 c3a4 c5d6 f2f3 g4e5 f3f4 e5f7 h3f2 b8d7 e1g1 d7b6 a4c3 c8e6 f1d1 d8e7 b2b3 a8d8 f2d3 g7g5 d2e3 g5f4 d3f4 f7e5 a1c1 h8e8 f4e6 e7e6');
INSERT INTO evaluations VALUES(16399,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(16399,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(16399,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(16399,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16399,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(16399,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(16399,36,'cp 21','g1f3 f8c5 e1g1 d7d6 c2c3 a7a6 a2a4 e8g8 a4a5 h7h6 c1e3 c5e3 f2e3 c6e7 b1d2 e7g6 d3d4 c8d7 c4d3 d8e7 d1b3 a8b8 h2h3 e5d4 e3d4 g6f4 b3c2 f6h5 a1e1 f8e8 e4e5 f4d3 c2d3 h5f4 d3c2 b8a8 d2c4');
INSERT INTO evaluations VALUES(42560,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(42560,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(42560,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42560,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42560,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(42560,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(42560,36,'cp 12','g1f3 e8g8 b1c3 d8e7 e1g1 d7d6 c3a4 c5b6 h2h3 c8e6 c4e6 e7e6 c2c4 b8c6 f3h4 c6e7 a4b6 a7b6 a2a4 f6d7 b2b4 f7f5 h4f5 e7f5 e4f5 f8f5 c1e3 f5f7 d1d2 a8f8 g1h2 e6f6 a1e1 f6h4 d3d4');
INSERT INTO evaluations VALUES(42556,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(42556,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42556,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(42556,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(42556,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(42556,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42556,36,'cp 17','e8g8 e1g1 d7d6 c2c3 c5b6 f1e1 h7h6 b1d2 c7c6 d3d4 e5d4 c3d4 d6d5 c4b3 d5e4 d2e4 c8f5 e4f6 d8f6 c1d2 b8d7 d2c3 f5g4 d4d5 f6g6 d5c6 b7c6 b3c2 g4f3 c2g6 f3d1 g6f5 d1f3 g2f3 d7f6 a1d1 f8e8 g1f1 e8d8 d1d8 a8d8 f5e4 f6d5 c3d2');
INSERT INTO evaluations VALUES(37234,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(37234,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(37234,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(37234,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(37234,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(37234,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(37234,36,'cp 13','c1g5 c7c6 e1g1 d7d6 c2c3 b7b5 d3d4 c5b6 c4d3 c8g4 a2a4 h7h6 g5h4 b8d7 a4b5 c6b5 d3b5 g7g5 h4g3 f6e4 b5d7 d8d7 d4e5 d6e5 d1d7 g4d7 f3e5 d7e6 f1e1 a8d8 e5f3 f7f5 f3d4 f8e8 b1a3 b6d4 c3d4 d8d4');
INSERT INTO evaluations VALUES(37214,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(37214,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37214,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(37214,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(37214,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(37214,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(37214,36,'cp 20','d7d6 c2c3 h7h6 g5h4 d8e7 e1g1 g7g5 h4g3 f6h5 b2b4 c5b6 a2a4 a7a5 b4b5 c8g4 c4d5 c7c6 d5a2 h5g3 h2g3 b8d7 b1a3 b6a7 a3c4 d6d5 e4d5 c6d5 b5b6 d7b6 c4b6 a7b6 a2d5 a8c8 d1b3 b6c5 a1b1 b7b6 b3c4 e7d6 c4g4');
INSERT INTO evaluations VALUES(36984,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36984,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36984,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(36984,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(36984,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36984,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36984,36,'cp 25','c2c3 c7c6 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 d3d4 c5b6 d4e5 d6e5 d1d8 f8d8 f3e5 d8e8 e5g6 f6e4 g5f4 h7g6 b1d2 e4d2 f4d2 b8d7 a1d1 d7e5 c4e2 a8d8 d2f4 d8d1 f1d1 f7f6 g1g2 g6g5 f4h2 g8f7 a2a4 e8d8 d1d8 b6d8 b2b4 a7a5 f2f4');
INSERT INTO evaluations VALUES(51424,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(51424,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(51424,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(51424,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(51424,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(51424,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(51424,36,'cp 86','g1f3 c8b7 f1d3 c7c5 c2c3 g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 b8c6 h2h4 h7h6 h4h5 c5d4 c3d4 c6b4 d3b5 a7a6 b5a4 b7c6 a4c6 b4c6 c1e3 f8e7 f1c1 c6a5 b1d2 e8g8 d2f1 a5c4 b2b3 c4e3 f1e3 f8e8');
INSERT INTO evaluations VALUES(51422,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(51422,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(51422,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(51422,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(51422,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(51422,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(51422,36,'cp 104','c8b7 f1d3 c7c5 d4d5 e6d5 e4d5 b7d5 b1c3 d5e6 c1f4 a7a6 d1e2 g8f6 e1c1 b6b5 f3e5 f8e7 g2g4 b5b4 c3e4 f6e4 d3e4 a8a7 c1b1 e8g8 h2h4 d7d6 e5c4 a7d7 e4d5 e6d5 d1d5 b8c6 g4g5 b4b3 a2b3 c6b4 d5d2 f8e8 e2g4 d6d5 c4e5 d7d6 h4h5 d5d4');
INSERT INTO evaluations VALUES(30678,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(30678,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(30678,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(30678,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(30678,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(30678,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(30678,36,'cp 111','f1d3 c7c5 d4d5 e6d5 e4d5 b7d5 b1c3 d5e6 c1f4 a7a6 d1e2 g8f6 e1c1 b6b5 f3e5 f8e7 g2g4 d7d6 g4g5 f6d7 e5f7 e6f7 f4d6 b8c6 e2e4 a8c8 h1e1 e8f8 d6e7 c6e7 d3f1 e7c6 h2h4 c6d4 f1h3 b5b4 c3e2 d8c7 e2d4 c5d4 c1b1 c8e8 e4d4 d7e5 f2f4 e5c6 e1e8 f8e8 d4d3');
INSERT INTO evaluations VALUES(30674,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(30674,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(30674,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(30674,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(30674,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(30674,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(30674,36,'cp 88','c7c5 c2c3 g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 f8e7 a2a4 b8c6 b1a3 a7a6 c1f4 d7f8 f4e3 f8g6 a1e1 c5d4 c3d4 c6b4 d3b1 b7c6 e2d1 d8d7 b2b3 b6b5 a4b5 c6b5 a3b5 d7b5 d1d2 a6a5 e1c1 b5b7 h2h4');
INSERT INTO evaluations VALUES(30490,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(30490,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(30490,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(30490,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(30490,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(30490,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(30490,36,'cp 108','d4d5 e6d5 e4d5 b7d5 b1c3 d5e6 c1f4 b8c6 d3a6 c6b4 a6b7 a8b8 f4b8 d8b8 b7e4 g8f6 a2a3 f6e4 c3e4 b4d5 e1g1 f8e7 e4g5 e7g5 f3g5 b8e5 g5e6 f7e6 c2c3 e8g8 g2g3 h7h5 d1d3 f8f5 a1d1 e5c7 c3c4 d5f6 f1e1 h5h4 f2f4 b6b5 c4b5 c5c4 d3d6 c7d6 d1d6 f5b5');
INSERT INTO evaluations VALUES(17811,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(17811,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(17811,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(17811,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(17811,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(17811,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(17811,36,'cp -87','d8e7 e5f3 e7e4 d1e2 e4e2 e1e2 g8e7 b1c3 h7h5 d2d3 e7g6 a2a4 a6a5 c1d2 f7f6 a1e1 e8f7 e2d1 f8b4 c3e4 c8g4 c2c3 g6h4 h2h3 g4f3 g2f3 b4d6 f3f4 a8d8 d1c2 h4g6 e4d6 d8d6 d2e3 g6h4 b2b4 h8d8 e3c5 d6d3 b4a5');
INSERT INTO evaluations VALUES(1173,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(1173,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(1173,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(1173,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(1173,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(1173,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(1173,36,'cp -78','e5f3 e7e4 d1e2 e4e2 e1e2 b7b6 d2d3 c8b7 a2a4 c6c5 a4a5 b6b5 h1e1 e8c8 f3e5 d8e8 e2f1 e8e7 c2c4 f7f6 e5f3 b5c4 d3c4 b7f3 g2f3 e7e1 f1e1 g8e7 c1e3 e7c6 e1f1 f8d6 b1d2 h8e8 d2e4 f6f5 e4d6 c7d6 h2h4 e8e6 h4h5 c6e5 b2b3 e5f3');
INSERT INTO evaluations VALUES(1174,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(1174,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(1174,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(1174,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(1174,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(1174,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(1174,36,'cp -67','e7e4 d1e2 e4e2 e1e2 b7b6 d2d3 c8b7 a2a4 c6c5 c1e3 e8c8 b1d2 d8e8 a4a5 b6b5 h1e1 g8f6 e2f1 f6d5 d2b3 d5e3 e1e3 e8e3 f2e3 f8d6 f1f2 b7d5 b3d2 f7f6 g2g3 c8b7 c2c3 d5e6 e3e4 h8d8 f2e2 e6h3 e2e3 d8e8 a1c1 h3e6 c1a1 g7g5 a1c1 g5g4');
INSERT INTO evaluations VALUES(1214,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(1214,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(1214,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(1214,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(1214,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(1214,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(1214,36,'cp -68','d1e2 e4e2 e1e2 b7b6 d2d3 c8b7 a2a4 c6c5 c1e3 e8c8 b1d2 g8f6 a4a5 b6b5 c2c4 b5c4 d2c4 f8e7 c4e5 b7d5 h1c1 h8e8 e2f1 e7d6 e5c4 d5f3 c4d6 c7d6 g2f3 f6d5 d3d4 d5e3 f2e3 e8e3 d4c5 e3f3 f1g1 c8b7 c1f1 f3f1 a1f1 d6c5 f1f7 b7c6 f7g7 d8d2');
INSERT INTO evaluations VALUES(1215,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(1215,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(1215,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(1215,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(1215,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(1215,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(1215,36,'cp -85','e4e2 e1e2 b7b6 d2d3 c8b7 a2a4 c6c5 c1e3 g8f6 h1e1 f8d6 b1c3 e8c8 e2f1 f6g4 a4a5 h8e8 a5b6 c7b6 e3d2 e8e1 a1e1 b7f3 g2f3 g4e5 e1e3 e5g6 c3d5 c8b7 c2c4 b7c6 d2c3 g6f4 d5f4 d6f4 e3e7 d8d7 e7d7 c6d7 h2h4 f7f6 h4h5');
INSERT INTO evaluations VALUES(1219,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(1219,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(1219,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(1219,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(1219,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(1219,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(1219,36,'cp -68','e1e2 b7b6 a2a4 c6c5 d2d3 c8b7 a4a5 b6b5 h1e1 e8c8 f3e5 d8e8 e2f1 e8e7 c2c4 f7f6 e5f3 b5c4 d3c4 b7f3 g2f3 e7e1 f1e1 f8d6 c1e3 g8e7 b1c3 e7c6 f3f4 h8d8 c3d5 c6d4 e1d1 c7c6 d5b6 c8c7 a1a3 d4e6 f4f5 d6h2 d1c2 e6d4 e3d4 c5d4');
INSERT INTO evaluations VALUES(1218,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(1218,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(1218,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(1218,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(1218,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(1218,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(1218,36,'cp -63','b7b6 a2a4 c6c5 d2d3 c8b7 a4a5 b6b5 h1e1 e8c8 f3e5 d8e8 e2f1 e8e7 c2c4 f7f6 e5f3 b5c4 d3c4 b7f3 g2f3 e7e1 f1e1 f8d6 b1c3 g8e7 c1e3 e7c6 f3f4 c6d4 a1d1 h8e8 e1d2 d4b3 d2d3 e8d8 d3c2 b3a5 b2b3 a5c6 c3e4 c6d4 e3d4 c5d4 e4d6 d8d6 b3b4 c8b7 c2d3');
INSERT INTO evaluations VALUES(57513,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(57513,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(57513,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(57513,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(57513,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(57513,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(57513,36,'cp 169','f3e5 g8f6 f1c4 d8e7 d2d4 b8c6 e1g1 f5e4 c1f4 d7d5 c4b3 c8e6 c2c4 d5c4 b3a4 e6d7 e5c6 d7c6 a4c6 b7c6 b1d2 e7e6 f4c7 f8b4 a1c1 b4d2 d1d2 e8g8 d2e2 a8c8 c7e5 c8e8 f1e1 f6g4 e5g3 e4e3 f2f3 g4f2 c1c4 e6d5 c4c1 g8h8 g3f2 e3f2 e2f2 e8e1 c1e1');
INSERT INTO evaluations VALUES(57465,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(57465,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(57465,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(57465,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(57465,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(57465,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(57465,36,'cp 161','d8f6 d2d4 d7d6 e5c4 f5e4 b1c3 f6g6 f2f3 b8c6 d4d5 c6e5 c3e4 g8f6 e4f6 g7f6 d1e2 c8d7 c1e3 e8c8 e2d2 d8e8 e1c1 e5c4 f1c4 c8b8 c1b1 f6f5 h1e1 f8g7 c4f1 b7b6 f1a6 h7h5 e3g5 e8e1 d1e1 g7f6 g5f6 g6f6 c2c3 f5f4 e1e4 d7f5');
INSERT INTO evaluations VALUES(35229,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(35229,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(35229,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(35229,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(35229,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(35229,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(35229,36,'cp 159','d2d4 d7d6 e5c4 f5e4 b1c3 f6g6 f2f3 f8e7 c3e4 d6d5 c4e5 g6b6 e4g3 g8f6 c2c3 e8g8 e5d3 c7c5 d3c5 e7c5 d4c5 b6c5 f1e2 b8c6 d1d3 f8e8 c1g5 c8e6 d3b5 c5e7 g5f6 e7f6 e1g1 e6f7 a1e1 c6e5 b5b7 e8b8 b7c7 b8b2 e2b5 b2b5 e1e5');
INSERT INTO evaluations VALUES(35228,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(35228,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(35228,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(35228,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(35228,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(35228,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(35228,36,'cp 189','d7d6 e5c4 f5e4 b1c3 f6g6 f2f3 f8e7 c3e4 d6d5 c4e5 g6b6 e4c3 g8f6 f1e2 e8g8 c3a4 b6a5 c2c3 c7c6 e1g1 b8d7 e5d3 e7d6 c1f4 d6f4 d3f4 b7b5 a4c5 d7c5 d4c5 b5b4 d1d4 b4c3 b2c3 c8a6 e2a6 a5a6 a2a4 a8b8 a1e1 f8e8 h2h3 a6b7 e1e6 e8e6 f4e6 b7d7 d4f4 b8e8 e6d4 g8h8');
INSERT INTO evaluations VALUES(34987,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(34987,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(34987,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(34987,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(34987,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(34987,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(34987,36,'cp 175','e5c4 f5e4 b1c3 f6g6 f2f3 f8e7 c3e4 d6d5 c4e5 g6b6 e4g3 g8f6 c2c3 c7c5 d1a4 b8c6 f1b5 e8g8 b5c6 b7c6 a4c6 a8b8 c6b6 b8b6 b2b3 c5d4 c3d4 e7b4 c1d2 b4a3 a1b1 f6d7 e5d3 b6e6 e1f2 e6c6 d2b4 c6c2 g3e2 f8e8 h1e1 c2a2 b4a3 c8a6 d3f4 a2a3 f4d5 a3a2');
INSERT INTO evaluations VALUES(34988,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(34988,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(34988,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(34988,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(34988,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(34988,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(34988,36,'cp 200','f5e4 b1c3 f6g6 f2f3 f8e7 c3e4 d6d5 c4e5 g6b6 e4c3 g8f6 f1e2 e8g8 c3a4 b6a5 c2c3 c7c6 e1g1 b8d7 e5d3 a5c7 c1f4 c7d8 d1d2 f6h5 f4e3 e7d6 a4c5 d7c5 d4c5 d6c7 e3g5 h5f6 f1e1 b7b6 c5b6 a7b6 g5f4 c8f5 e2f1 f5d3 f4c7 d8c7 f1d3 g8h8 a2a3 f6d7 e1e7');
INSERT INTO evaluations VALUES(34991,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(34991,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(34991,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(34991,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(34991,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(34991,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(34991,36,'cp 187','b1c3 f6g6 f2f3 f8e7 c3e4 d6d5 c4e5 g6b6 e4c3 g8f6 f1e2 e8g8 c3a4 b6a5 c2c3 c7c6 e1g1 a5c7 c3c4 c8e6 c1f4 c7d8 a4c5 e7c5 d4c5 b8d7 e5d7 f6d7 f4d6 d5c4 d1c2 d8f6 d6f8 f6f8 f1e1 f8c5 g1h1 d7f6 e2f1 e6f7 a1d1 b7b5 f1d3 c5g5 d3f5 a8e8');
INSERT INTO evaluations VALUES(34990,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(34990,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(34990,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(34990,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(34990,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(34990,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(34990,36,'cp 174','f6g6 f2f3 f8e7 c3e4 d6d5 c4e5 g6b6 e4c3 g8f6 f1e2 e8g8 c3a4 b6a5 c2c3 c7c6 e1g1 a5c7 f3f4 b7b5 a4c5 e7c5 d4c5 b8d7 e5d7 c8d7 c1e3 d7f5 e3d4 f8e8 d4f6 g7f6 e2f3 c7e7 d1d4 e7e3 d4e3 e8e3 g1f2 a8e8 f1e1 e3e1 a1e1 e8e1 f2e1');
INSERT INTO evaluations VALUES(34993,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(34993,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(34993,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(34993,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(34993,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(34993,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(34993,36,'cp 201','f2f3 f8e7 c3e4 d6d5 c4e5 g6b6 e4g3 g8f6 c2c3 e8g8 e5d3 c7c5 d3c5 e7c5 d4c5 b6c5 d1d4 f8e8 e1f2 c5f8 f2g1 b8c6 d4f2 f8d6 c1e3 c8e6 g3e2 e6f7 h2h4 h7h6 h1h3 f7e6 h3h2 e6d7 a1d1 d6e7 e3c5 e7f7 e2d4 a7a6 g2g4 h6h5 g4g5 f6h7 f1d3 c6d4 c5d4 h7f8');
INSERT INTO evaluations VALUES(14295,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(14295,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(14295,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14295,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(14295,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(14295,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(14295,36,'cp 50','c2c4 e7e6 a2a3 d5c4 b1d2 e6e5 d4e5 f6g4 d1a4 c8d7 e5e6 d7e6 f1c4 e6d7 e1g1 f8d6 a4c2 e8g8 f1d1 d8e7 h2h3 g4e5 c4e2 e5f3 d2f3 e7f6 c1d2 d7f5 c2b3 c6e5 d2c3 f5e4 c3e5 d6e5 f3e5 f6e5 b3b5 e5b5 e2b5');
INSERT INTO evaluations VALUES(14210,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(14210,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(14210,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(14210,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(14210,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(14210,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14210,36,'cp 57','e7e6 a2a3 d5c4 b1d2 e6e5 d4e5 f6g4 d1a4 c8d7 e5e6 d7e6 f1c4 e6d7 a4b3 g4e5 c4e2 d7e6 b3b7 a8b8 b7a6 b8b6 a6a4 e5d3 e2d3 d8d3 f3e5 d3d5 e5c6 d5c6 a4c6 b6c6 b2b4 f8e7 c1b2 c6c2 b2d4 c7c5 b4c5 e7c5 e1d1 e6f5 e3e4 c2d2 d1d2 c5d4 e4f5 d4a1');
INSERT INTO evaluations VALUES(21001,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(21001,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(21001,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(21001,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(21001,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21001,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(21001,36,'cp 38','c1f4 a7a6 e2e3 g8f6 g1f3 c8g4 h2h3 g4f3 d1f3 e7e6 g2g4 f8d6 e1c1 d6f4 f3f4 d8e7 f1g2 e8c8 h1e1 h7h6 a2a3 c8b8 f4g3 e7d6 g3d6 d8d6 c3a4 g7g5 f2f4 c6e7 a4c5 e7g6 c5d3 f6d7 c1b1 d6c6 c2c3');
INSERT INTO evaluations VALUES(20978,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(20978,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(20978,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(20978,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(20978,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(20978,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(20978,36,'cp 21','a7a6 e2e3 c8f5 f1d3 f5d3 d1d3 e7e6 g1f3 f8d6 f4g3 g8f6 e1g1 h7h6 a2a3 c6e7 e3e4 f6e4 c3e4 d5e4 d3e4 b7b6 f1e1 e8g8 g3h4 d8e8 h4e7 d6e7 d4d5 e7f6 d5e6 e8e6 e4e6 f7e6 c2c3 g8f7 e1e3 a8d8 g2g3 d8d6 a1e1 f8d8 g1g2 g7g5 h2h3 f6g7');
INSERT INTO evaluations VALUES(17709,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(17709,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(17709,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(17709,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(17709,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(17709,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(17709,36,'cp 20','e2e4 d5e4 d4d5 e7e5 f4g3 c6e7 g3e5 e7g6 d1d4 g6e5 d4e5 d8e7 e5e7 g8e7 d5d6 c7d6 c3e4 e7c6 e1c1 d6d5 d1d5 c8e6 d5d2 e6a2 g1e2 a8d8 d2d8 e8d8 e2c3 a2e6 e4g5 f8c5 g5e6 f7e6 c3e4 c5a7 c1b1 d8e7 c2c3 b7b5 f2f3 h8d8 f1e2 h7h6 g2g3');
INSERT INTO evaluations VALUES(16127,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(16127,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(16127,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(16127,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(16127,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(16127,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16127,36,'cp 11','f8c5 e1g1 e8g8 h2h3 d7d6 c3a4 c5b6 a4b6 a7b6 a2a4 c8d7 c2c3 c6e7 c1d2 e7g6 f1e1 f8e8 d1b3 d7e6 c4e6 e8e6 c3c4 f6d7 b3c2 g6h4 f3h4 d8h4 d2e3 e6g6 g1h2 d7c5 c2d1 c5e6 g2g3 h4f6 b2b4 a8d8 e1f1 g8h8 f2f3 e6d4');
INSERT INTO evaluations VALUES(6393,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(6393,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(6393,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(6393,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(6393,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(6393,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(6393,36,'cp 36','d7d6 c2c3 a7a6 b5a4 c5a7 d2d4 b7b5 a4b3 g8f6 c1e3 e8g8 b1d2 c8g4 h2h3 g4h5 a1c1 f8e8 f1e1 h5f3 d2f3 c6a5 d1c2 a5b3 a2b3 h7h6 c1d1 d8e7 b3b4 a7b6 d4e5 d6e5 f3h4 a8d8 h4f5 e7e6 b2b3 b6e3 e1e3 d8d1 c2d1 e6b6 d1e2');
INSERT INTO evaluations VALUES(5949,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(5949,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(5949,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(5949,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(5949,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5949,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(5949,36,'cp 39','c2c3 a7a6 b5a4 c5a7 d2d4 b7b5 a4b3 g8f6 c1e3 e8g8 b1d2 c8g4 h2h3 g4h5 a1c1 f8e8 f1e1 h5f3 d2f3 c6a5 e3g5 h7h6 g5h4 a7b6 d4e5 d6e5 d1d8 a8d8 h4f6 g7f6 c1d1 a5b7 f3h4 d8d1 e1d1 b7d6 b3d5 g8f8 g2g3 a6a5 g1g2 e8d8');
INSERT INTO evaluations VALUES(5944,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(5944,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(5944,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(5944,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(5944,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5944,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(5944,36,'cp 31','a7a6 b5c6 b7c6 d2d4 e5d4 c3d4 c5b6 b1c3 g8e7 f1e1 e8g8 h2h3 e7g6 c3a4 h7h6 a4b6 c7b6 d4d5 c6d5 d1d5 c8e6 d5d3 g6e5 f3e5 d6e5 d3g3 d8f6 c1d2 f8e8 f2f3 e6d7 e1d1 e8e6 d2e3 f6e7 g3f2 e6g6 g1h2 f7f5 e3b6 f5e4 b6c5 e7e6 f3e4 d7c6');
INSERT INTO evaluations VALUES(4759,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(4759,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(4759,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(4759,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(4759,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(4759,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(4759,36,'cp 43','b5a4 b7b5 a4c2 c5a7 a2a4 b5b4 d2d4 b4c3 d4e5 c6e5 h2h3 c8e6 b1c3 g8f6 b2b4 e8g8 a1b1 h7h6 f1e1 f6d7 c1e3 a7e3 e1e3 e5c4 e3e1 c7c5 b4c5 d7c5 f3d4 c4a3 b1c1 d8f6 e1e3 a8e8 d4f5 a3c2 d1c2');
INSERT INTO evaluations VALUES(4768,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(4768,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(4768,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4768,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(4768,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(4768,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(4768,36,'cp 44','b7b5 a4c2 c5a7 a2a4 b5b4 d2d4 b4c3 d4e5 c6e5 h2h3 g8f6 b1c3 e8g8 a1b1 c8e6 b2b4 h7h6 c1f4 f8e8 f4g3 f6h5 g3h2 h5f6 f1e1 f6d7 h2g3 a8b8 f3e5 d7e5 c3d5 c7c6 d5e3 a6a5 g3e5 d6e5 d1d8 b8d8 b4a5 a7c5 a5a6 d8a8 c2d3');
INSERT INTO evaluations VALUES(4906,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(4906,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(4906,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(4906,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(4906,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(4906,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(4906,36,'cp 41','a4c2 c5a7 a2a4 b5b4 d2d4 b4c3 d4e5 c6e5 h2h3 g8f6 b1c3 e8g8 a1b1 c8e6 c1f4 f6d7 f4g3 a8b8 b2b4 e6c4 f1e1 e5f3 d1f3 d7e5 f3f4 f8e8 c3d5 c7c5 b4c5 c4d5 e4d5 a7c5 g1h2 g7g6 b1b8 d8b8 e1b1 b8a7 f4d2 g8g7');
INSERT INTO evaluations VALUES(4904,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(4904,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(4904,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(4904,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(4904,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(4904,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(4904,36,'cp 40','c5a7 a2a4 b5b4 d2d4 b4c3 d4e5 c6e5 h2h3 g8f6 b1c3 e8g8 a1b1 h7h6 f1e1 c8e6 b2b4 a8b8 c1f4 f6d7 f4g3 e5f3 d1f3 d7e5 f3d1 f8e8 c2d3 a7d4 c3e2 e5d3 d1d3 d4f6 b4b5 a6b5 a4b5 f6e5 f2f4 e5f6 e2d4 f6d4 d3d4 e6d7 d4d3 g8h8 e1d1');
INSERT INTO evaluations VALUES(47096,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(47096,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(47096,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(47096,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47096,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47096,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47096,36,'cp 9','c7c6 d2d4 c6d5 b1c3 b8c6 g1f3 c8g4 c1g5 e7e6 c4d5 e6d5 f1e2 f8e7 h2h3 g4h5 e1g1 e8g8 f1e1 h7h6 g5h4 f8e8 f3e5 h5e2 e1e2 f6e4 h4e7 e8e7 e5c6 b7c6 d1e1 d8b6 a1d1 a8c8 f2f3 e4c3 b2c3 e7e2 e1e2 g8h8 e2e7');
INSERT INTO evaluations VALUES(44862,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(44862,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(44862,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44862,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(44862,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(44862,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44862,36,'cp 6','b1c3 c6d5 d2d4 b8c6 g1f3 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 e1g1 e7e6 f1d1 f8e7 a1c1 e8g8 g2g3 d8a5 b2b3 f8d8 c3a4 a5d5 f3e2 c8b8 a4c3 d5d7 c3e4 b8b4 e2c2 h7h6 c2c6 b4d4 d1d4 d7d4 g1g2 d4b6 c6b6 a7b6');
INSERT INTO evaluations VALUES(44861,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44861,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44861,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44861,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44861,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(44861,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44861,36,'cp 6','c6d5 c4d5 f6d5 g1f3 b8c6 f1b5 e7e6 d2d4 f8d6 e1g1 e8g8 b5d3 h7h6 f1e1 d5f6 c3e4 f6e4 d3e4 c8d7 c1d2 c6e7 e4b7 a8b8 b7e4 b8b2 d1c1 b2b5 a2a4 b5b3 f3e5 d6e5 d4e5 d7c6 e4c6 e7c6 d2h6 c6b4 c1g5 d8g5 h6g5 b4d3 e1e3');
INSERT INTO evaluations VALUES(45416,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45416,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45416,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45416,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(45416,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45416,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45416,36,'cp 0','d2d4 b8c6 g1f3 c8g4 c1g5 d8a5 g5d2 g4f3 g2f3 d5c4 f1c4 a5b4 b2b3 c6d4 a1c1 e7e6 c3b5 b4c5 d2e3 c5b4 e3d2');
INSERT INTO evaluations VALUES(45418,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(45418,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45418,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(45418,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(45418,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(45418,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(45418,36,'cp 9','b8c6 g1f3 c8g4 c1g5 f6e4 g5e3 e7e6 c4d5 e6d5 d1b3 f8b4 f3e5 d8a5 a2a3 b4c3 b2c3 c6e5 d4e5 a5c3 b3c3 e4c3 a1c1 c3a4 c1c7 e8g8 f2f3 f8c8 c7b7 g4e6 e1d2 a4c5 e3c5 c8c5 b7b5 a8c8 f1d3 c5c3 b5a5 c3c7 f3f4 g7g6 h1b1');
INSERT INTO evaluations VALUES(13125,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13125,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13125,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13125,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13125,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13125,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13125,36,'cp 14','g1f3 c8g4 c1g5 d8a5 g5f6 e7f6 h2h3 g4f3 d1f3 d5c4 f3e3 f8e7 f1c4 e8g8 e1g1 a8d8 a1d1 e7b4 d1d3 g7g6 f1d1 g8g7 e3f3 b4c3 b2c3 h7h5 c4b3 d8d7 f3f4 f8d8 g2g3 c6e7 d3f3 a5f5 f4f5 e7f5 g3g4 f5h4 f3g3 h5g4');
INSERT INTO evaluations VALUES(13124,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(13124,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13124,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13124,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(13124,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13124,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13124,36,'cp 14','c8g4 c1g5 d8a5 g5f6 e7f6 h2h3 g4f3 d1f3 d5c4 f3e3 f8e7 f1c4 e8g8 e1g1 a8d8 a1d1 e7b4 d1d3 g7g6 f1d1 g8g7 e3f3 b4c3 b2c3 h7h5 f3f4 f8e8 g2g3 d8d7 d3f3 f6f5 f3e3 e8e4 e3e4 f5e4 f4e4 a5c3 d4d5 c3e5 e4e5 c6e5 c4b5 d7c7 d5d6');
INSERT INTO evaluations VALUES(24627,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(24627,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(24627,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(24627,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(24627,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24627,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(24627,36,'cp 9','c1g5 d8a5 g5f6 e7f6 h2h3 g4f3 d1f3 d5c4 f3e3 f8e7 f1c4 e8g8 e1g1 a8d8 a1d1 e7b4 d1d3 g7g6 f1d1 g8g7 e3f3 d8d7 c3d5 f6f5 a2a3 b4d6 g2g4 c6e7 b2b4 a5d8 d5e7 d7e7 g4f5 e7c7 c4b3 d8g5 f3g4 g5f5 g4f5 g6f5 d3f3 g7f6 g1f1 f8d8 d1e1 d8c8 e1d1');
INSERT INTO evaluations VALUES(52411,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(52411,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(52411,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(52411,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(52411,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(52411,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(52411,36,'cp -68','b8c6 f2f4 d7d6 g1f3 g7g6 f1b5 c8d7 d1e2 g8h6 e5d6 h6f5 e1g1 e7e6 d2d3 f8g7 c2c3 f5d6 b5a4 e8g8 a4c2 d8c7 b1d2 b7b6 f3e5 d7c8 d2f3 c8b7 c1d2 d6f5 a1e1 a8e8 g2g4 f5d6 e2g2 b6b5 c2b3 c6e7 g2f2');
INSERT INTO evaluations VALUES(19365,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(19365,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(19365,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(19365,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(19365,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(19365,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(19365,36,'cp -62','f2f4 d7d5 g1f3 g8h6 d2d3 e7e6 b1c3 d8b6 c3e2 c8d7 c2c3 f8e7 e2g3 f7f6 e5f6 g7f6 f1e2 e8c8 e1g1 h6g4 h2h3 g4h6 g1h2 d8g8 d3d4 c5d4 f3d4 c6d4 c3d4 c8b8 c1d2 e7d6 g3h5 b6d4 g2g4');
INSERT INTO evaluations VALUES(19363,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(19363,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(19363,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(19363,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(19363,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(19363,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(19363,36,'cp -58','d7d5 g1f3 d8b6 c2c3 c8f5 f3h4 f5d7 h4f3 e7e6 b1a3 g8e7 d1b3 b6d8 d2d3 e7f5 f1e2 h7h5 e1g1 f8e7 b3d1 d5d4 f3e1 d8a5 a3c4 a5a6 e1c2 b7b5 c4d2 d4c3 b2c3 a8d8 d1e1 e8g8 d2e4 b5b4 d3d4 a6a5 d4c5');
INSERT INTO evaluations VALUES(19825,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(19825,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(19825,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(19825,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(19825,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(19825,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(19825,36,'cp -48','g1f3 c8g4 f1b5 d8b6 h2h3 g4f3 b5c6 b6c6 d1f3 e7e6 d2d3 g8e7 e1g1 e8c8 b1c3 h7h5 c3e2 e7f5 e2g3 g7g6 g3f5 g6f5 c1e3 f8e7 c2c3 h8g8 e3f2 g8g7 g1h1 h5h4 d3d4 d8g8 f1g1 c6b5 a2a4 b5b2 d4c5 g7g6');
INSERT INTO evaluations VALUES(19823,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(19823,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(19823,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(19823,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(19823,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(19823,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(19823,36,'cp -46','g8h6 f1b5 c8d7 b1a3 e7e6 c2c3 d8b6 b5e2 f7f6 d1b3 f8e7 b3b6 a7b6 a3c2 e8g8 e5f6 g7f6 d2d4 h6f5 c2e3 f5d6 f4f5 c5d4 f5e6 d7e6 f3d4 c6d4 c3d4 a8a4 e1g1 f8c8 f1d1 f6f5 b2b3 a4a8 a2a4 d6e4 g2g3');
INSERT INTO evaluations VALUES(54967,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(54967,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(54967,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(54967,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(54967,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(54967,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(54967,36,'cp 62','g8f6 c2c4 g7g6 b1c3 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 c1e3 b8a6 a2a3 b7b6 f3d2 g4e2 d1e2 d8d7 f1b1 a5a4 e2d1 a6c5 b2b4 a4b3 e3c5 b6c5 b1b3 a8a7 a3a4 f8a8 b3a3 h7h5 d2f3 a7a6 a4a5 g7h6 d1d3 d7e7 h2h3 f6h7 d3c2 g8g7');
INSERT INTO evaluations VALUES(47019,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(47019,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(47019,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(47019,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(47019,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(47019,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(47019,36,'cp 62','c2c4 g7g6 b1c3 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 a2a3 b8a6 c1e3 d8e8 f3e1 g4e2 d1e2 b7b6 e1d3 a6c5 e3c5 b6c5 b2b4 c5b4 a3b4 a5b4 d3b4 h7h5 b4c6 e8d7 c3b5 g8h7 e2f3 f6g4 a1a8 f8a8 b5c3');
INSERT INTO evaluations VALUES(21992,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(21992,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(21992,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(21992,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(21992,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21992,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21992,36,'cp 36','d2d4 e5d4 f3d4 d6d5 b1c3 d5e4 d4c6 d8d1 e1d1 b7c6 c3e4 f8e7 f1d3 c8g4 d1e1 e8c8 e1f2 g8f6 h1e1 g4e6 e4f6 e7f6 a1b1 g7g6 b2b3 a7a5 c1b2 f6b2 b1b2 d8d4 e1e4 h8d8 b2b1 c8b7 e4d4 d8d4 f2e3 c6c5 b1f1 a5a4 c2c3 d4d8 b3a4 e6a2');
INSERT INTO evaluations VALUES(21952,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(21952,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(21952,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21952,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21952,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21952,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(21952,36,'cp 24','e5d4 f3d4 d6d5 b1c3 d5e4 d4c6 d8d1 e1d1 b7c6 c3e4 f8e7 f1d3 c8g4 d1e1 e8c8 e1f2 g8f6 h1e1 h8e8 e4f6 e7f6 e1e8 d8e8 c2c3 e8d8 d3a6 c8d7 h2h3 g4e6 c1e3 d8b8 a1d1 d7e8 d1d2 e6a2 a6d3 g7g6 e3a7 b8d8 a7e3 h7h5 f4f5 g6f5 d3f5 d8d2 e3d2 a2d5 f2g3 h5h4 g3h2');
INSERT INTO evaluations VALUES(22040,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(22040,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(22040,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(22040,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(22040,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(22040,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(22040,36,'cp 33','f3d4 d6d5 b1c3 d5e4 d4c6 d8d1 e1d1 b7c6 c3e4 f8e7 f1d3 c8g4 d1e1 e8c8 e1f2 g8f6 h1e1 h8e8 e4f6 e7f6 e1e8 d8e8 a1b1 f6d4 f2f1 g7g6 c1d2 g4e6 b2b3 c6c5 h2h3 c8d7 b1e1 h7h5 e1e4 f7f5 e4e2 e6d5 g2g3 e8e2 f1e2 d7e6 d2e3 a7a5 e3d4');
INSERT INTO evaluations VALUES(22029,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(22029,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(22029,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(22029,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(22029,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(22029,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(22029,36,'cp 20','d6d5 b1c3 d5e4 d4c6 d8d1 e1d1 b7c6 c3e4 f8e7 f1d3 c8g4 d1e1 e8c8 e1f2 g8f6 c1e3 f6e4 d3e4 c8b7 h2h3 g4e6 a1d1 g7g6 e4f3 h7h5 h1e1 e7f6 a2a3 a7a5 e3c5 a5a4 c2c3 e6b3 d1d8 h8d8 g2g3 b3d5 e1e2 d5f3 f2f3 d8d5 c5e7 d5d3 f3e4');
INSERT INTO evaluations VALUES(52777,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(52777,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52777,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(52777,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(52777,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52777,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52777,36,'cp 28','c1f4 c8f5 b1d2 e7e6 e2e3 f8d6 f3e5 f7f6 e5d3 g8e7 f4g3 d6g3 h2g3 b8d7 c2c4 e6e5 f1e2 d8b6 c4d5 c6d5 d2b3 e8g8 d1d2 f8c8 d3c5 d7c5 b3c5 b6c6 a1c1 b7b6 c5d3 c6d6 e1g1 f5d3 d2d3 e5e4 d3b5');
INSERT INTO evaluations VALUES(52757,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(52757,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52757,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(52757,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(52757,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52757,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(52757,36,'cp 19','c8f5 e2e3 e7e6 b1d2 f8d6 f3e5 f7f6 e5d3 g8e7 f4g3 d6g3 h2g3 b8d7 f1e2 e6e5 c2c4 d8b6 c4d5 c6d5 d2b3 b6d6 g3g4 f5g6 e1g1 e8g8 a1c1 a7a5 b3c5 d7c5 c1c5 b7b6 c5c3 g6f7 d4e5 f6e5 d1a4 f7e8 a4a3');
INSERT INTO evaluations VALUES(31086,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(31086,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(31086,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31086,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31086,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31086,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31086,36,'cp 17','e2e3 e7e6 c2c4 f8d6 f4d6 d8d6 f1d3 f5d3 d1d3 g8f6 d3b3 d6c7 e1g1 e8g8 h2h3 h7h6 f1c1 b8d7 b1c3 f8c8 a2a4 a7a5 c4d5 f6d5 f3e1 d5b4 c3a2 b4a2 a1a2 c8d8 e1d3 e6e5 d4e5 d7e5');
INSERT INTO evaluations VALUES(31085,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(31085,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31085,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(31085,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(31085,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(31085,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(31085,36,'cp 15','e7e6 c2c4 f8d6 f4d6 d8d6 b1c3 g8f6 d1b3 d6e7 f1e2 a7a5 a2a3 a5a4 b3d1 d5c4 e2c4 b8d7 e1g1 e8g8 c4d3 f5d3 d1d3 e6e5 d3c2 e5d4 f3d4 d7c5 a1d1 f8d8 d4f3 f6d5 c3d5 c6d5 d1d4 c5b3 f1d1 a8c8 d4d5 c8c2');
INSERT INTO evaluations VALUES(31456,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31456,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(31456,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(31456,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31456,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(31456,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31456,36,'cp 15','c2c4 f8d6 f4d6 d8d6 f1e2 g8f6 e1g1 a7a5 b1c3 e8g8 f3h4 f5e4 c4d5 c6d5 a1c1 b8c6 a2a3 h7h6 h4f3 f8d8 h2h3 d6e7 e2b5 c6a7 b5d3 a7c6 d1e2 e4d3 e2d3 f6e8 c3a4 e8d6 c1c2 d6c4 d3e2');
INSERT INTO evaluations VALUES(31439,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31439,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(31439,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31439,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31439,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(31439,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(31439,36,'cp 21','f8d6 f4d6 d8d6 f1d3 f5d3 d1d3 g8f6 e1g1 e8g8 d3b3 d6c7 f1c1 b8d7 b1d2 a7a5 a2a4 h7h6 c1c2 f8d8 h2h3 d8c8 a1c1 c8e8 c1d1 e8d8 c2c3 a8c8 c4d5 f6d5 c3c1 b7b6 d2c4 d7f6 f3e5 d5b4 e5d3 c6c5 d3b4');
INSERT INTO evaluations VALUES(46700,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(46700,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(46700,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(46700,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(46700,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(46700,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(46700,36,'cp 42','f8g7 c1e3 e8g8 d1d2 c7c5 d4c5 d6c5 d2d8 f8d8 e3c5 b8c6 f1b5 f6d7 c5a3 d7b6 f3d2 c8d7 e1c1 g7c3 b2c3 g8f8 d2b3 c6e5 b5e2 d7e6 f2f4 d8d1 h1d1 e5c4 a3c5 b6a4 c5d4 a8c8 d1e1 c4a3 e2d3 a4c3 f4f5 e6c4 d4c5 c4d3');
INSERT INTO evaluations VALUES(55974,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55974,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55974,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(55974,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(55974,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(55974,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(55974,36,'cp 19','d2d4 c7c5 c2c3 b8c6 g1f3 c8d7 f1e2 c5d4 c3d4 g8e7 e1g1 e7f5 b1c3 a7a6 c1e3 f5e3 f2e3 f8e7 a1c1 a8c8 e2d3 c6a5 d1e2 e8g8 e2f2 g8h8 c3e2 d7b5 d3b5 a6b5 e2f4 h8g8 h2h4 c8c1 f1c1 b5b4');
INSERT INTO evaluations VALUES(21881,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(21881,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(21881,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(21881,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(21881,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(21881,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(21881,36,'cp 42','c6e5 b1c3 g8f6 c1f4 e5g6 f4g3 f8e7 d1d2 e8g8 e1c1 h7h6 a2a3 a7a6 h2h4 f8e8 f1e2 c8d7 c1b1 f6h5 g3h2 h5f6 h1g1 f6g4 h2g3 e7f6 f3d4 g6h4 f2f3 g4e5 c3d5 b7b5 g3h4 f6h4 f3f4 e5c4 e2c4');
INSERT INTO evaluations VALUES(22112,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(22112,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(22112,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(22112,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(22112,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(22112,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(22112,36,'cp 48','c2c4 e5f3 d1f3 g8f6 b1c3 f8e7 f1e2 e8g8 c1f4 c8g4 f3d3 g4e2 d3e2 f6d7 e1g1 e7f6 c3d5 f8e8 a1d1 d7b6 e2c2 b6d5 d1d5 f6e5 f4e3 d8e7 a2a4 e7e6 h2h3 c7c6 d5d2 h7h6 b2b3 e5f6 e3d4 f6e5 f2f4 e5f6 d4f6 e6f6');
INSERT INTO evaluations VALUES(21898,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(21898,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(21898,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(21898,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(21898,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(21898,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(21898,36,'cp 36','c6a5 c4b5 c7c6 b5d3 c8e6 e1g1 f8e7 c3e2 g8f6 c2c3 c6c5 d3b5 e6d7 b5d7 d8d7 e2g3 a5c6 d2d4 e5d4 c3d4 d6d5 e4d5 d7d5 d4c5 d5c5 c1e3 c5d5 d1a4 e8g8 f1d1 d5a5 a4a5 c6a5 g3f5 a5c6 f5e7 c6e7 h2h3 f8d8 e3c5');
INSERT INTO evaluations VALUES(22154,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(22154,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(22154,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(22154,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(22154,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(22154,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(22154,36,'cp 39','c4b5 c7c6 b5d3 c6c5 e1g1 g8f6 c3d5 f8e7 c2c3 e8g8 d5e7 d8e7 d3c2 c8g4 d2d3 d6d5 h2h3 g4h5 d1e2 a8d8 e4d5 d8d5 e2e3 a5c6 f3h4 e7d7 f1e1 h5g6 h4g6 h7g6 c2b3 d5d3 e3c5 f8d8 c1g5 d7f5 g5h4 g6g5 h4g3 f6e4 c5b5 e4g3 f2g3 f5g6 b5b7 d3g3 g1h2');
INSERT INTO evaluations VALUES(57021,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(57021,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57021,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57021,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(57021,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57021,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57021,36,'cp 0','g8f6 g1f3 b7b6 b2b3 c8b7 c1b2 d7d5 f1d3 f8d6 c2c4 d5c4 d3c4 a7a6 e1g1 e8g8 c4e2 b8d7 b1d2 c7c5 a1c1 d8e7 d2c4 d6c7 c1c2 f8c8 c4d2 c7d6 d2c4');
INSERT INTO evaluations VALUES(49424,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49424,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49424,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49424,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49424,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49424,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(49424,36,'cp 15','f1d3 d7d5 g1f3 b7b6 b2b3 c7c5 c1b2 c8b7 b1d2 b8c6 e1g1 f8e7 d4c5 e7c5 c2c4 d5d4 e3d4 c6d4 b3b4 d4f3 d2f3 c5e7 a2a3 e8g8 f3e5 d8c7 d1e2 a7a5 b2c3 a5b4 a3b4 a8a1 f1a1 f8d8 e2b2 e7d6');
INSERT INTO evaluations VALUES(49420,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(49420,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(49420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49420,36,'cp 0','c7c5 g1f3 d7d5 b2b3 b7b6 c1b2 c8b7 b1d2 f8e7 e1g1 b8c6 a2a3 e8g8 d1e2 a7a5 c2c4 c5d4 e3d4 a8c8 f1c1 f8e8 c1c2 e7d6 a1c1 d6f4 c1e1 f4d6');
INSERT INTO evaluations VALUES(44234,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44234,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44234,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44234,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44234,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44234,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44234,36,'cp 0','g1f3 d7d5 b2b3 b7b6 c1b2 b8c6 b1d2 f8e7 a2a3 c8b7 e1g1 e8g8 d1e2 a7a5 c2c4 c5d4 e3d4 a8c8 f1c1 f8e8 c1c2 e7d6 a1c1 d6f4 c1e1 f4d6');
INSERT INTO evaluations VALUES(44235,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44235,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44235,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44235,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44235,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44235,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44235,36,'cp 0','d7d5 b2b3 b7b6 e1g1 b8c6 d4c5 f8c5 c1b2 e8g8 a2a3 c8b7 b3b4 c5d6 b1d2 d8c7 b2f6 g7f6 f3d4 c6d4 d1g4 g8h8 g4h4 d6h2 g1h1 f6f5 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(46245,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46245,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46245,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46245,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46245,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46245,36,'cp 0','b2b3 b7b6 e1g1 c8b7 c1b2 b8c6 d4c5 f8c5 a2a3 e8g8 b3b4 c5d6 b1d2 d8c7 b2f6 g7f6 f3d4 c7e7 d3h7 g8h7 d1h5 h7g7 h5g4 g7h6 g4h3 h6g7');
INSERT INTO evaluations VALUES(46222,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46222,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46222,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46222,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46222,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46222,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46222,36,'cp 0','b7b6 e1g1 c8b7 c1b2 b8c6 d4c5 f8c5 a2a3 e8g8 b3b4 c5d6 b1d2 d8c7 b2f6 g7f6 f3d4 c6d4 d1g4 g8h8 g4h4 d6h2 g1h1 f6f5 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(1844,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(1844,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(1844,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(1844,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(1844,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(1844,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(1844,36,'cp 136','b1c3 g8e7 e1g1 a7a6 b5a4 g7g5 a4c6 d7c6 d2d4 c8g4 d4e5 f6g6 h2h3 a8d8 d1e2 g4f3 e2f3 f8g7 f3e2 g7e5 f2f4 g5f4 c1f4 g6g7 f4e5 g7e5 f1f2 e8g8 a1f1 e7g6 f2f5 e5d4 g1h1 d8e8 e2f3 e8e7 a2a3 d4g7 f3e3 e7e5 f5f3 g6h4 f3f2 f7f5');
INSERT INTO evaluations VALUES(1840,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(1840,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(1840,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(1840,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(1840,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(1840,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(1840,36,'cp 112','g8e7 e1g1 c6d4 f3d4 e5d4 c3e2 c7c6 b5a4 d7d5 d2d3 d5e4 d3e4 c8g4 f2f3 g4d7 c2c3 d4c3 b2c3 e7g6 c1e3 f8d6 f3f4 f6e7 a4c2 d6c7 a1b1 c7b6 d1d2 e8g8 e2g3 g8h8 e4e5 a8d8 b1b6 a7b6 e3b6 d7e6 d2f2 e6a2 b6c5 e7c7 c5f8 g6f8');
INSERT INTO evaluations VALUES(1065,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(1065,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(1065,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(1065,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(1065,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(1065,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(1065,36,'cp 129','e1g1 a7a6 b5a4 g7g5 a4c6 d7c6 d2d4 c8g4 d4e5 f6g6 h2h3 g4f3 d1f3 f8g7 f3g3 h7h6 f2f4 g5f4 c1f4 g6g3 f4g3 e7g6 e5e6 f7e6 g3c7 g7d4 g1h1 a8c8 c7h2 b7b5 a1d1 d4e5 h2g1 c6c5 g1e3 h6h5 d1d2 c5c4 h1g1 h8g8 c3e2 c8d8 b2b3 d8d2 e3d2 c4b3 a2b3');
INSERT INTO evaluations VALUES(1064,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(1064,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(1064,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(1064,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(1064,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(1064,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(1064,36,'cp 94','a7a6 b5a4 g7g5 a4c6 d7c6 d2d4 c8g4 d4e5 f6g6 h2h3 a8d8 d1e2 g4f3 e2f3 f8g7 e5e6 f7e6 f3g3 h8g8 g3c7 d8d7 c7g3 g7c3 g3c3 g6e4 f1e1 e4f5 b2b3 e7d5 c3g3 g8g6 e1e5 f5c2 e5g5 e8f7 g3h4 f7g8 g5g6 h7g6 c1g5 c2f5 a1e1 e6e5 h4g4 f5g4 h3g4 g8g7 e1e5');
INSERT INTO evaluations VALUES(54230,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(54230,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(54230,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(54230,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(54230,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(54230,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(54230,36,'cp 147','e4d5 e6d5 f1b5 b8c6 d4c5 c8e6 c1e3 g8e7 g1f3 e7f5 e3g5 f8e7 d1d2 h7h6 g5e7 d8e7 g2g4 f5h4 f3h4 e7h4 e1c1 e8g8 f2f3 a8c8 h1e1 f8d8 b5c6 b7c6 c3e2 h4h2 b2b3 e6d7 e2d4 h2d2 c1d2 d8e8 d2c3 a7a5');
INSERT INTO evaluations VALUES(54128,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(54128,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(54128,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(54128,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(54128,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(54128,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(54128,36,'cp 139','e6d5 f1b5 b8c6 d4c5 c8e6 c1e3 g8e7 d1d2 e7f5 g1f3 f5e3 d2e3 f8e7 f3d4 d8c8 a1d1 e8g8 d4e6 f7e6 c3a4 e7f6 e1g1 c8d7 f1e1 a8e8 b2b4 d7f7 c2c4 h7h6 a2a3 g8h8 g2g3 f7g6 b5c6 b7c6 e3e2 h6h5 b4b5 d5c4 b5c6 c4c3 c6c7 c3c2');
INSERT INTO evaluations VALUES(54516,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(54516,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(54516,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(54516,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(54516,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(54516,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(54516,36,'cp 173','f1b5 b8c6 d4c5 d5d4 d1e2 c8e6 c3e4 g8f6 c1g5 f8e7 g5f6 g7f6 g1f3 f6f5 e4d6 e7d6 c5d6 d8d6 e1c1 e8c8 b5c6 d6c6 f3d4 c6g2 d4e6 f7e6 e2e6 c8b8 e6f5 d8d1 h1d1 g2h2 b2b3 h7h5 f2f4 b8a8 f5e4 h5h4 d1d7 h8b8 f4f5 h2f2 d7h7');
INSERT INTO evaluations VALUES(54462,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(54462,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(54462,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(54462,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(54462,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(54462,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(54462,36,'cp 136','b8c6 d4c5 c8e6 c1e3 g8h6 g1f3 h6f5 e3g5 f8e7 d1d2 e8g8 e1c1 f7f6 g5f4 e7c5 h1e1 e6f7 c3d5 c6d4 f3d4 f5d4 d5e7 g8h8 f4e3 d8e7 e3d4 e7d6 a2a4 f7g6 b5c4 c5d4 d2d4 d6h2 g2g3 h2h5 b2b3 h5a5 c1b2 a8d8 d4c3 a5b6 d1d8 f8d8 c3e3 b6e3 e1e3 d8d2');
INSERT INTO evaluations VALUES(20459,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(20459,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(20459,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(20459,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(20459,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(20459,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(20459,36,'cp 160','d4c5 c8e6 c1e3 g8e7 g1f3 e7f5 e3g5 f8e7 d1d2 f7f6 g5f4 e7c5 e1c1 e8g8 h1e1 e6f7 c3d5 c6d4 d5e7 f5e7 f3d4 e7g6 d4e6 d8d2 f4d2 c5f2 e1e2 f7e6 e2e6 g6e5 b5a4 g8h8 c1b1 a8d8 c2c3 f2b6 b2b4 g7g5 d2c1 d8d3 d1d3 e5d3 e6e7 d3c1 b1c1 b6g1 e7b7');
INSERT INTO evaluations VALUES(20456,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(20456,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(20456,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(20456,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(20456,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(20456,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(20456,36,'cp 137','c8e6 c1e3 g8h6 g1f3 h6f5 e3g5 f8e7 d1d2 e8g8 e1c1 f7f6 g5f4 e7c5 h1e1 e6f7 c3d5 c6d4 d5e7 f5e7 f3d4 e7g6 d4e6 d8d2 f4d2 c5f2 e1e2 f7e6 e2e6 g6e5 b5a4 f8d8 d1f1 f2c5 d2f4 b7b6 f4e5 f6e5 f1e1 d8f8 c2c3 f8f6 e6e5');
INSERT INTO evaluations VALUES(22515,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(22515,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(22515,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(22515,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(22515,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(22515,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(22515,36,'cp 20','g8f6 c2c3 f8c5 b2b4 c5b6 a2a4 a7a6 e1g1 e8g8 b1d2 d7d6 h2h3 c8e6 f1e1 d8d7 c4b3 e6b3 d2b3 a8e8 b3d2 c6e7 d2c4 b6a7 c1e3 a7e3 c4e3 e7g6 c3c4 f6h5 e3f5 h5f4 d3d4 g6e7 f5g3 e7c6 b4b5 c6d4 f3d4 e5d4 d1d4 f4e6 d4c3');
INSERT INTO evaluations VALUES(55001,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(55001,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(55001,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(55001,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(55001,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(55001,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(55001,36,'cp 71','d2d4 g8f6 c2c4 g7g6 b1c3 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 c1e3 b8a6 a2a3 b7b6 a1b1 a6c5 f3d2 g4e2 d1e2 a5a4 b2b4 a4b3 e3c5 b6c5 b1b3 f6h5 d2f3 g7h6 g2g3 d8c8 f1a1 f7f5 a1b1 h5f6 g1g2 f6e4 c3e4 f5e4 e2e4');
INSERT INTO evaluations VALUES(19615,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(19615,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(19615,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(19615,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(19615,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(19615,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(19615,36,'cp 94','e7e6 b1c3 a7a6 c1f4 b7b5 e1c1 c8b7 f2f3 a8c8 c1b1 g8f6 d4a7 b7c6 f1d3 f8b4 c3e2 e8g8 e2d4 c8a8 d4c6 d7c6 a7b7 d8e7 b7e7 b4e7 d3e2 a8d8 d1d8 f8d8 c2c4 f6d7 c4b5 c6b5 h1c1 e7c5 h2h4 c5b6 f4d6');
INSERT INTO evaluations VALUES(19148,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(19148,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(19148,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(19148,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(19148,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(19148,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(19148,36,'cp 116','b1c3 a7a6 c1f4 b7b5 f1e2 c8b7 e1c1 g8f6 e2f3 f8e7 h1e1 e8g8 e4e5 b7f3 g2f3 f6h5 f4e3 a8c8 d4a7 e7g5 a7d7 g5f4 e3f4 h5f4 d7d8 c8d8 a2a3 f8e8 b2b4 f4g6 d1d6 d8d6 e5d6 e8d8 e1d1');
INSERT INTO evaluations VALUES(19146,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(19146,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(19146,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(19146,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(19146,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(19146,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(19146,36,'cp 86','a7a6 c1f4 b7b5 e1c1 c8b7 f1e2 g8f6 e2f3 f8e7 h1e1 e8g8 e4e5 b7f3 g2f3 f6h5 f4e3 d8a5 c1b1 a8c8 e1g1 b5b4 c3e4 a5c7 d1d2 f7f5 e5f6 h5f6 e4f6 e7f6 d4d7 c7d7 d2d7 a6a5 g1d1 a5a4 d7a7 a4a3 b2a3 b4a3 a7a3 f6e5 a3a7 e5h2 a2a4 h2e5 a4a5 c8c4 b1c1 f8c8 d1d2');
INSERT INTO evaluations VALUES(17468,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(17468,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(17468,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(17468,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(17468,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(17468,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(17468,36,'cp 105','c1f4 b7b5 e1c1 c8b7 f1e2 g8f6 e2f3 f8e7 a2a3 e8g8 e4e5 b7f3 g2f3 f6h5 f4e3 d8a5 c3e4 b5b4 a3b4 e7b4 c1b1 a8b8 h1g1 b4e7 b2b3 f8c8 d4a7 a5c7 a7c7 c8c7 b1b2 b8b5 f3f4 f7f6 e4d6 b5b8 d1d3 a6a5 d3d1 f6e5 f4e5');
INSERT INTO evaluations VALUES(51111,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51111,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(51111,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(51111,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(51111,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(51111,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(51111,36,'cp 63','d2d4 e7e6 f1d3 d7d5 b1c3 c7c5 d4c5 f8c5 e4d5 e6d5 c3d5 c5f2 e1f2 d8d5 d1f3 d5d4 c1e3 d4h4 f3g3 h4g3 f2g3 b8c6 g1f3 g8f6 h1e1 e8g8 c2c4 c6b4 d3b1 f8e8 e3d2 e8e1 d2e1 b4c6 g3f2 h7h6 h2h3 c8e6 b2b3 b7b5 b1d3 b5c4 b3c4');
INSERT INTO evaluations VALUES(51107,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(51107,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(51107,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(51107,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(51107,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(51107,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(51107,36,'cp 51','e7e6 f1d3 d7d5 b1c3 c7c5 d4c5 f8c5 e4d5 e6d5 c3d5 c5f2 e1f2 d8d5 g1f3 g8f6 h1e1 c8e6 f2g1 b8c6 c2c4 d5d6 d3f1 d6d1 e1d1 e8g8 h2h3 a6a5 c1e3 f8e8 d1e1 a5a4 f3d4 c6d4 e3d4 e8d8 a1d1 a8c8 b2b3 h7h6 g1h2 a4b3 a2b3 c8a8 d4g1 e6f5 b3b4 d8d1');
INSERT INTO evaluations VALUES(51038,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(51038,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(51038,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(51038,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(51038,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(51038,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(51038,36,'cp 51','f1d3 d7d5 b1c3 c7c5 d4c5 f8c5 e4d5 e6d5 c3d5 c5f2 e1f2 d8d5 d1f3 d5d4 c1e3 d4h4 f3g3 h4g3 h2g3 b8c6 a1e1 c8e6 e3d2 g8f6 d2c3 e8c8 a2a4 h7h5 g1f3 d8d5 h1h4 h8g8 f2g1 d5c5 c3f6 g7f6 b2b4 c5d5 d3e4 d5d7 b4b5 c6e5 f3e5 f6e5 h4h5 a6b5 a4b5 g8g3 h5e5 c8c7 e4f3');
INSERT INTO evaluations VALUES(34599,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(34599,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(34599,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(34599,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34599,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(34599,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(34599,36,'cp 16','c8g4 b1c3 d5e6 d1e2 e6e2 f1e2 b8c6 c3d5 e8c8 d5e3 g4h5 f3e5 c6e5 e2h5 e7e6 h5e2 f8c5 e3c4 e5c4 e2c4 g8f6 c2c3 c5e7 d2d4 c7c5 c1e3 f6d5 e1c1 d5e3 f2e3 e7g5 c1c2 g5e3 h1f1 c5d4 f1f7 d4c3 c4e6 c8b8');
INSERT INTO evaluations VALUES(32850,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(32850,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32850,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(32850,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(32850,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(32850,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(32850,36,'cp 27','b1c3 d5e6 d1e2 b8c6 h2h3 e6e2 f1e2 g4h5 g2g4 h5g6 e2b5 e7e6 b5c6 b7c6 d2d3 h7h5 g4g5 f8d6 c1e3 g8e7 f3h4 a7a5 a2a4 g6h7 e1c1 e8g8 d1e1 c6c5 h4f3 e7d5 e3d2 d5b4 b2b3 f8b8 h1g1 b4c6 c3e4 e6e5 d2c3 c6d4');
INSERT INTO evaluations VALUES(32849,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(32849,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(32849,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32849,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(32849,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(32849,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(32849,36,'cp 36','d5e6 d1e2 e6e2 f1e2 b8c6 h2h3 g4h5 g2g4 h5g6 e2b5 e7e6 b5c6 b7c6 d2d3 h7h5 g4g5 f8d6 c3e4 g8e7 b2b3 a7a5 a2a4 e7d5 c1b2 d6b4 e1f1 e8g8 f3e5 b4d6 e5c6 a8a6 c6e5 d5b4 e5g6 f7g6 f1g2 b4c2 a1c1 c2b4 e4d6 c7d6');
INSERT INTO evaluations VALUES(32835,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(32835,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(32835,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(32835,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(32835,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(32835,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(32835,36,'cp 33','d1e2 b8c6 h2h3 e6e2 f1e2 g4h5 g2g4 h5g6 e2b5 e7e6 b5c6 b7c6 d2d3 f8d6 c1d2 g8e7 e1c1 f7f6 d1e1 e8f7 f3d4 h7h5 f2f4 c6c5 d4f3 e7c6 b2b3 h5g4 h3g4 a8b8 f3h4 g6h7 c1b2 c6d4 g4g5 e6e5 f4e5 d6e5 h1f1 e5d6 g5f6');
INSERT INTO evaluations VALUES(16408,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16408,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16408,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16408,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16408,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16408,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16408,36,'cp 0','g1e2 d7d6 b1c3 c8g4 f3e3 c6b4 c4b3 g4e6 b3a4 e6d7 a4b3');
INSERT INTO evaluations VALUES(16409,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16409,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16409,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16409,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16409,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16409,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16409,36,'cp 0','d7d6 b1c3 c8g4 f3e3 c6b4 c4b3 g4e6 b3a4 e6d7 a4b3');
INSERT INTO evaluations VALUES(15507,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15507,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15507,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15507,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15507,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15507,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15507,36,'cp 0','b1c3 c8g4 f3e3 c6b4 c4b3 g4e6 b3a4 e6d7 a4b3');
INSERT INTO evaluations VALUES(51322,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(51322,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(51322,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(51322,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(51322,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(51322,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(51322,36,'cp 83','a2a4 b5b4 d1f3 c7c6 f1c4 e7e6 b1d2 g8f6 a4a5 a7a6 d2b3 c8b7 g1h3 f8e7 e1g1 e8g8 h3f4 b8d7 f4d3 d8c7 b3c5 d7c5 d4c5 f6d7 e3e4 d7c5 c1f4 c7d8 f1d1 c5d3 d1d3 d8e8 a1d1 c6c5 f4e5 g8h8 d3d7 b7c6 d7c7 c6b5 c4b5 a6b5 d1d7 a8a5 b2b3 e8a8 e5b2 a5a1 b2a1 a8a1 d7d1');
INSERT INTO evaluations VALUES(51333,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(51333,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(51333,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(51333,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(51333,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(51333,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(51333,36,'cp 75','b5b4 d1f3 c7c6 f1c4 e7e6 a4a5 g8f6 b1d2 a7a6 g1h3 a8a7 e1g1 c6c5 h3f4 c8b7 f3h3 c5d4 e3d4 f8d6 d2f3 e8g8 f3e5 b7e4 c1e3 a7c7 f1c1 d8e7 c4b3 e4f5 g2g4 f5e4 g4g5 f6d5 f4d5 e4d5 b3d5 e6d5 h3f5 d6e5 f5e5');
INSERT INTO evaluations VALUES(51337,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(51337,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(51337,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(51337,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(51337,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(51337,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(51337,36,'cp 84','d1f3 c7c6 f1c4 e7e6 g1e2 c8b7 a4a5 a7a6 e3e4 g8f6 e4e5 f6d5 b1d2 c6c5 f3g3 b8c6 d4c5 d5e7 d2e4 e7f5 g3f4 c6a5 e4d6 f8d6 e5d6 a5c4 f4c4 b7g2 h1g1 g2c6 c4b4 e8g8 b4g4 d8d7 e2d4 g7g6 b2b3 f5d4 g4d4 e6e5 d4c4 f8b8');
INSERT INTO evaluations VALUES(47332,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(47332,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(47332,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(47332,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47332,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(47332,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(47332,36,'cp 26','a7a6 g1f3 c7c5 e1g1 e7e6 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 b2b3 b8d7 c4e2 b7b6 a2a3 c8b7 c1b2 h8c8 b3b4 c5d6 d2c4 d6c7 a3a4 f6d5 b4b5 a6b5 a4b5 d7c5 f3d4 a8a1 f1a1 c8a8 b2a3 d5c3 e2f3 b7f3 g2f3 e7e8 a1c1 c3a4 d4c6 c5d3 c1c2 a4c5 a3c5 d3c5');
INSERT INTO evaluations VALUES(43602,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43602,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(43602,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43602,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(43602,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(43602,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(43602,36,'cp 37','g1f3 c7c5 e1g1 e7e6 b2b3 c5d4 f3d4 f8d6 b1d2 e8g8 c4e2 e6e5 d4f3 b8c6 c1b2 d6c7 d1c2 d8e7 a1c1 f8d8 f1d1 c8g4 a2a3 g4h5 e2d3 d8d7 d3f5 h5g6 f5g6 h7g6 b3b4 a8d8 f3g5 c6a7 g2g3 a6a5 b4a5 c7a5 d2c4 d7d1 c1d1 a7c6 c4a5 d8d1 c2d1');
INSERT INTO evaluations VALUES(43603,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43603,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(43603,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(43603,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(43603,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(43603,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(43603,36,'cp 29','e7e6 e1g1 c7c5 b2b3 b8d7 c1b2 f8e7 d4c5 e7c5 b1d2 e8g8 c4d3 b7b5 d2e4 c5e7 d1b1 h7h6 f1d1 d8c7 a2a4 f6e4 d3e4 c8b7 e4h7 g8h8 a4b5 a6b5 a1a8 b7a8 h7d3 b5b4 b1a1 e7f6 d1c1 c7b7 b2f6 d7f6');
INSERT INTO evaluations VALUES(15090,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(15090,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(15090,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(15090,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(15090,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(15090,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(15090,36,'cp 52','e4d5 f6d5 e1g1 f8e7 f1e1 f7f6 c2c3 c8g4 h2h3 g4h5 d3d4 e5d4 c3d4 d8d7 b1c3 h5f7 a2a3 e8c8 d1c2 a7a6 c4d3 c8b8 c1d2 g7g6 a1c1 h8e8 c3e4 g6g5 d3a6 b7a6 c2c6 d7c6 c1c6 b8b7 c6c2 b7b8 e4c5 e7c5 e1e8 d8e8 d4c5 b8b7 f3d4 f7g6 c2c1 h7h5 c5c6 b7b6');
INSERT INTO evaluations VALUES(14988,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14988,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(14988,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14988,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14988,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(14988,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(14988,36,'cp 41','f6d5 e1g1 f8e7 f1e1 f7f6 c2c3 c8g4 h2h3 g4h5 d3d4 e5d4 c3d4 d8d7 b1c3 h5f7 a2a3 a7a6 c4d3 e8c8 d1c2 c8b8 c1d2 g7g6 a1c1 d5b6 b2b4 h8e8 c3e4 f7d5 d2c3 b6c4 a3a4 f6f5 e4c5 e7c5 d4c5 e8e1 f3e1 c6e5 d3e2 d7e7 b4b5 a6b5 a4b5 e7c5');
INSERT INTO evaluations VALUES(15318,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(15318,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(15318,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(15318,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(15318,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(15318,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(15318,36,'cp 56','e1g1 d5b6 c4b5 f8d6 f3e5 d6e5 b5c6 b7c6 f1e1 f7f6 d3d4 e8g8 d4e5 d8d1 e1d1 f6e5 d1e1 f8e8 b1a3 c8f5 e1e2 a7a5 c1g5 h7h6 g5d2 a5a4 a1e1 e5e4 f2f3 f5g6 f3e4 g6f7 e2e3 a8d8 d2c3 f7a2 h2h3 e8e7 a3b1 e7f7 e3g3 a2e6');
INSERT INTO evaluations VALUES(15317,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(15317,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(15317,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(15317,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(15317,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(15317,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(15317,36,'cp 58','f8e7 f1e1 f7f6 c2c3 c8g4 h2h3 g4h5 d3d4 e5d4 c3d4 h5f7 b1c3 d8d7 d1c2 e8c8 a2a3 a7a6 c4d3 c8b8 c1d2 g7g6 a1c1 d5b6 b2b4 h8e8 c3e4 f7d5 d2f4 b6c8 c2b2 g6g5 f4g3 h7h5 b4b5 a6b5 d3b5 g5g4 h3g4 h5g4 f3d2 c8a7 e4c3 e7d6 e1e8 d8e8 c3d5 d6g3');
INSERT INTO evaluations VALUES(8534,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(8534,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(8534,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(8534,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(8534,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(8534,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(8534,36,'cp 69','f1e1 f7f6 c2c3 c8g4 h2h3 g4h5 d3d4 e5d4 c3d4 d8d7 b1c3 h5f7 d1c2 e8c8 a2a3 a7a6 c4d3 c8b8 c1d2 g7g6 a1c1 d5b6 d2f4 h8e8 c3e4 f7d5 b2b4 d5e4 d3e4 b6d5 f4d2 e7f8 c2b3 f6f5 e4d3 e8e1 c1e1 d5b6 d2g5');
INSERT INTO evaluations VALUES(8535,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(8535,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(8535,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(8535,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(8535,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(8535,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(8535,36,'cp 50','f7f6 c2c3 c8g4 h2h3 g4h5 d3d4 e5d4 c3d4 e8g8 b1c3 h5f7 c3d5 f7d5 d1b3 d5c4 b3c4 f8f7 c1f4 c6b4 f4c7 d8d5 c4d5 b4d5 c7a5 e7b4 a5b4 d5b4 e1c1 f7d7 c1c4 b4c6 a1e1 a8d8 e1e4 g7g5 g2g4 c6e7 h3h4 h7h6 g1g2 e7d5 h4h5 d5f4 g2g3 g8f7 b2b4 a7a6 a2a4 f4d5 a4a5 b7b5 a5b6');
INSERT INTO evaluations VALUES(57660,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(57660,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(57660,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(57660,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(57660,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(57660,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(57660,36,'cp 41','g8f6 c2c4 f8g7 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 d2a5 c6a5 f1d3 c8g4 e1g1 b7b6 f1d1 f8d8 d3a6 g4c8 a6e2 c8b7 d4d5 a8c8 f3d4 a5c4 e3f4 c4d6 f2f3 e7e5 d5e6 d6e4 d4b5 e4c3 d1d8 c8d8 b5c3 g7d4 g1h1 d4c3');
INSERT INTO evaluations VALUES(42616,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(42616,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(42616,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(42616,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(42616,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(42616,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(42616,36,'cp 133','f3e5 c5d4 e5g4 f6g4 d1g4 e8g8 g4g3 f8e8 d2d3 d7d5 c1g5 d8d7 f1e2 d4c3 b2c3 d7c6 c3c4 d5e4 e1g1 c8f5 d3d4 c6g6 g5f4 g6g3 f2g3 b8c6 c2c3 a8c8 f4c7 f5g4 e2g4 c8c7 c4c5 b7b6 c5b6 a7b6 g4e2 c6a5 e2b5 e8d8');
INSERT INTO evaluations VALUES(42530,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(42530,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(42530,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(42530,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(42530,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(42530,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(42530,36,'cp 138','c5d4 e5f3 d4c3 d2c3 f6e4 f1d3 e4f6 d1e2 d8e7 c1e3 e8g8 e1c1 d7d6 h2h3 f8e8 h1e1 c8d7 e2f1 b8c6 b2b3 d6d5 c1b2 e7d6 g2g3 c6e5 f3e5 e8e5 f1g2 a8e8 e3f4 e5e1 f4d6 e1d1 d6c7 f6e4 g2f3 d1e1 d3e4 e1e4');
INSERT INTO evaluations VALUES(55060,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55060,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(55060,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(55060,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(55060,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(55060,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55060,36,'cp 0','c7c5 c2c4 d5c4 g1f3 g8f6 d1a4 d8d7 a4c4 c5d4 c4d4 b8c6 d4d7 c8d7 f1g2 h7h6 e1g1 e7e5 f1d1 f8c5 b1c3 e8c8 c1d2 d7e6 c3a4 c5d6 a4c3');
INSERT INTO evaluations VALUES(53274,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(53274,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(53274,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53274,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53274,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(53274,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53274,36,'cp 7','c2c4 d5c4 g1f3 g8f6 d1a4 b8d7 f1g2 e7e6 a4c4 a7a6 c4b3 c5d4 f3d4 f8c5 d4c2 c5e7 e1g1 d7c5 b3e3 c8d7 b1c3 d8c7 e3f4 e7d6 f4d4 d6e5 c1f4 e5d4 f4c7 d4c3 b2c3 e8g8 f1c1 c5a4 c7a5 f8c8 c3c4');
INSERT INTO evaluations VALUES(48427,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(48427,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(48427,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(48427,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(48427,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(48427,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(48427,36,'cp 23','f6d7 f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 a8b8 d2f2 d8b6 e1c1 c6d4 e3d4 c5d4 d1d4 b6c5 d4d2 c5f2 d2f2 e8g8 f1d3 b7b5 b2b4 d7b6 c3e2 c8d7 c1d2 f8c8 e2d4 b6c4 d3c4 c8c4 c2c3 b8a8 g2g3 a6a5 h1b1 f7f6 e5f6 g7f6 d4b3');
INSERT INTO evaluations VALUES(49159,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(49159,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(49159,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(49159,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(49159,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(49159,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(49159,36,'cp 33','f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 d1d2 b7b5 a2a3 d8a5 d4c5 f8c5 e3c5 d7c5 f3d4 a5c7 b2b4 c6d4 d2d4 c5e4 c3e4 d5e4 c2c4 e8g8 c4c5 f8d8 d4c3 f7f5 f1e2 c8b7 e1g1 h7h6 a1d1 b7d5 h2h3 c7e7 g1h2 a8c8 g2g4 f5g4 e2g4 a6a5 c3g3 a5b4');
INSERT INTO evaluations VALUES(49154,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(49154,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(49154,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(49154,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(49154,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(49154,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(49154,36,'cp 25','c7c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 a8b8 d2f2 d8b6 e1c1 c6d4 e3d4 c5d4 f2d4 b6d4 d1d4 h7h5 f1d3 b7b6 h2h4 c8b7 h1e1 g7g6 d4b4 b8c8 b4b3 e8e7 c3e2 d7c5 b3c3 c5a4 c3c8 h8c8 g2g3 e7d7');
INSERT INTO evaluations VALUES(43987,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(43987,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43987,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43987,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(43987,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(43987,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(43987,36,'cp 31','g1f3 b8c6 c1e3 a7a6 d1d2 b7b5 a2a3 d8a5 d4c5 f8c5 e3c5 d7c5 f3d4 a5b6 d4c6 b6c6 h2h4 e8g8 h4h5 h7h6 e1c1 c8b7 h1h3 c6c7 g2g4 a8d8 d2d4 c5e4 c1b1 f7f6 e5f6 g7f6 h3d3 c7f4 c3e4 d5e4 d4d8 e4d3 d8d3 g8h8 b1a1');
INSERT INTO evaluations VALUES(43986,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43986,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43986,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(43986,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(43986,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43986,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(43986,36,'cp 33','b8c6 c1e3 a7a6 d1d2 b7b5 a2a3 d8a5 d4c5 f8c5 e3c5 d7c5 f3d4 c6d4 d2d4 c5a4 b2b4 a5c7 c3a4 b5a4 c2c4 d5c4 a1c1 e8g8 c1c4 c7e7 d4b6 f8d8 c4c7 d8d7 c7d7 e7d7 f1e2 c8b7 e1g1 d7c6 b6c6 b7c6 f1d1 c6b5');
INSERT INTO evaluations VALUES(12111,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(12111,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(12111,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(12111,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(12111,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(12111,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(12111,36,'cp 37','c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 e8g8 d2f2 c6d4 e3d4 d8c7 f1d3 c7b6 d4c5 d7c5 e1g1 c8d7 f4f5 d7c6 f5f6 c5d7 f2b6 d7b6 f6g7 g8g7 f1f4 b6d7 a1e1 f8h8 e1e3 g7f8 c3e2 f8e7 f4h4 h7h6 e2d4 a8c8 d3e2 c8g8 h4h5 a6a5 h2h4 a5a4');
INSERT INTO evaluations VALUES(12110,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(12110,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(12110,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(12110,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(12110,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(12110,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(12110,36,'cp 27','a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 e8g8 d2f2 c5d4 e3d4 c6d4 f2d4 d7b8 f1d3 b8c6 d4e3 b7b5 e1g1 b5b4 c3a4 a6a5 a4c5 d8b6 e3f2 a8b8 h2h3 f7f6 a3b4 f6e5 b4a5 c6a5 b2b4 a5c6 b4b5 c6e7 c5a6 b6f2 g1f2 c8a6 b5a6 e5e4 d3e2 f8f4 f2e3 f4f1 e2f1 e7f5 e3f4');
INSERT INTO evaluations VALUES(11052,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(11052,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(11052,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(11052,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(11052,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(11052,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(11052,36,'cp 33','a2a3 c5d4 f3d4 f8c5 d1d2 e8g8 d2f2 c6d4 e3d4 c5d4 f2d4 d7b8 f1d3 b8c6 d4e3 b7b5 c3e2 b5b4 a3a4 a6a5 b2b3 c8a6 e1c1 a6d3 d1d3 d8c7 h1d1 c7a7 g2g4 a7e3 d3e3 a8c8 h2h4 h7h6 h4h5 c8c7 d1h1 c7e7 e3h3 f8c8 c1b1');
INSERT INTO evaluations VALUES(52364,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52364,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(52364,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52364,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(52364,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52364,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52364,36,'cp 35','b1c3 g8f6 g1f3 d7d5 d2d4 e7e6 c1g5 b8d7 e2e3 f8e7 f1d3 h7h6 g5h4 d5c4 d3c4 b7b5 c4d3 c8b7 e1g1 e8g8 a1c1 a8c8 a2a3 a7a6 b2b4 a6a5 c1b1 a5b4 a3b4 f6d5 h4e7 d8e7 c3d5 e6d5 d1b3 d7b6 f1c1');
INSERT INTO evaluations VALUES(52360,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(52360,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52360,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52360,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(52360,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(52360,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52360,36,'cp 26','d7d5 d2d4 g8f6 g1f3 e7e6 c1g5 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 h7h6 g5h4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 a2a3 b4e7 c4e2 a7a6 b5d6 f8d8 d6c4 b7b5 d1d8 e7d8 c4d6 d8e7 f1c1 c8d7 h4g3 a8d8 h2h4 f6d5 f3e5 c6e5');
INSERT INTO evaluations VALUES(52738,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(52738,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52738,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(52738,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(52738,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52738,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52738,36,'cp 22','d2d4 g8f6 g1f3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 a2a3 b4e7 c4e2 a7a6 b5d4 c8d7 d4f5 f8d8 f5e7 c6e7 f3d4 g8f8 d1c1 e7c6 f1d1 c6d4 d1d4 d7c6');
INSERT INTO evaluations VALUES(47116,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(47116,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(47116,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(47116,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(47116,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(47116,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(47116,36,'cp -21','h7h6 g5f6 e7f6 e2e3 c7c6 f1d3 g7g6 g1f3 f8d6 e1g1 e8g8 c3e2 g8g7 c2c4 d5c4 d3c4 b8d7 f1e1 a7a5 e2c3 b7b5 c4d3 c8b7 a2a4 b5a4 d1a4 d8e7 a4c2 d6b4 e3e4 f8e8 a1c1 d7f8 d3c4 f8e6 h2h3 h6h5 c4e6 e7e6 e4e5');
INSERT INTO evaluations VALUES(46879,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46879,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46879,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(46879,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(46879,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(46879,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(46879,36,'cp -9','g5f6 e7f6 e2e3 g7g6 f1d3 c7c6 g1f3 f8d6 c3e2 e8g8 e1g1 g8g7 h2h3 b8d7 b2b3 b7b6 c2c4 c8b7 c4d5 c6d5 a1c1 f6f5 c1c2 d8e7 d1a1 f8c8 c2c8 b7c8 f1c1 c8b7 g2g3 a7a6 g1g2 d7f6 f3e5 d6e5 d4e5 d5d4 f2f3 e7e5 a1d4');
INSERT INTO evaluations VALUES(46876,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46876,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(46876,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(46876,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(46876,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(46876,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(46876,36,'cp -18','e7f6 e2e3 c7c6 g1f3 f8d6 f1d3 e8g8 e1g1 g7g6 c3e2 f8e8 c2c4 d5c4 d3c4 b8d7 h2h3 d7b6 c4b3 b6d5 f1e1 c8d7 e2c3 d5c3 b2c3 f6f5 a2a4 d8c7 d1c2 e8e7 e1d1 a8e8 a4a5 g6g5 f3d2 g8h8 d2c4 f5f4 c4d6 c7d6');
INSERT INTO evaluations VALUES(47725,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(47725,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(47725,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(47725,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(47725,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(47725,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(47725,36,'cp -12','e2e3 g7g6 f1d3 c7c6 c3e2 f8d6 g1f3 e8g8 e1g1 g8g7 f1e1 b8d7 c2c4 d5c4 d3c4 a7a5 a2a3 f8e8 e2c3 f6f5 e3e4 f5e4 c3e4 d7f6 e4d6 d8d6 e1e8 f6e8 d4d5 c6c5 d1a4 e8f6 a1d1 c8d7 a4c2 d7g4 c2b3 g4f3 g2f3');
INSERT INTO evaluations VALUES(47723,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(47723,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(47723,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(47723,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(47723,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(47723,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(47723,36,'cp -18','c7c6 f1d3 g7g6 g1f3 f8d6 e1g1 e8g8 c3e2 g8g7 c2c4 d5c4 d3c4 b8d7 f1e1 a7a5 a2a3 f6f5 a1c1 d7f6 e2f4 d8e7 f4d3 f6e4 c4a2 h6h5 g2g3 c8d7 h2h4 f8d8 f3e5 d7e8 e5c4 d6c7 d1b3 a8b8 c4e5 f7f6');
INSERT INTO evaluations VALUES(23233,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(23233,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(23233,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(23233,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(23233,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(23233,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(23233,36,'cp 209','f1c4 c6d4 f3d4 e5d4 d1d4 g8e7 b1c3 e7c6 d4e3 f8d6 c1d2 b7b6 e1c1 c8b7 e3h3 c6e5 c4b3 d8e7 c3b5 e8c8 b5a7 c8b8 a7b5 d6c5 f2f4 e5f7 h1e1 f7d6 b5d6 c5d6 b3d5 b7d5 e4d5 e7f7 h3b3 h8e8 d2e3 e8e4 g2g3 b8b7 c2c4 d8e8 a2a4 e8a8 c4c5 a8a4');
INSERT INTO evaluations VALUES(23224,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(23224,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(23224,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(23224,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(23224,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(23224,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(23224,36,'cp 182','e5d4 e1g1 g8e7 c2c3 d7d5 e4d5 e7d5 f3d4 c6e5 d1e2 f8e7 c4b3 c7c6 f2f4 e5g6 f4f5 g6e5 e2h5 e5f7 b1d2 e8g8 d2e4 f7g5 f1e1 g8h8 g1h1 c6c5 c1g5 f6g5 d4e6 c8e6 f5e6 d5f6 e4f6 f8f6 h5e2 d8f8 h1g1 c5c4 b3c4 f6f4 c4d3 f8f6 a1d1 a8f8');
INSERT INTO evaluations VALUES(23252,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(23252,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(23252,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(23252,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(23252,34,'cp 223',NULL);
INSERT INTO evaluations VALUES(23252,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(23252,36,'cp 237','e1g1 c6a5 c4g8 h8g8 f3d4 a5c6 c1e3 c6d4 e3d4 d7d6 b1c3 g7g6 f2f4 c7c6 d1f3 f8e7 a1e1 c8d7 b2b4 e8f7 f4f5 f7g7 f5g6 h7g6 c3e2 g8h8 e2f4 d8g8 a2a4 a8f8 f3g3 g7h7 a4a5 g8f7 c2c3 d7e6 f4e6 f7e6');
INSERT INTO evaluations VALUES(6221,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(6221,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(6221,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(6221,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(6221,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6221,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(6221,36,'cp 39','c2c3 g8f6 b2b4 c5b6 a2a4 a7a6 e1g1 e8g8 b1d2 d7d6 h2h3 c8e6 f1e1 f6h5 c4a2 d8f6 d2c4 b6a7 c4e3 a7e3 c1e3 h5f4 e3f4 f6f4 e1e3 f4f6 d1c2 e6a2 c2a2 f6e6 a2c2 f7f5 e4f5 e6f5 e3e4 f8f7 a1e1 a8f8 c2b3 c6e7 d3d4 e5d4');
INSERT INTO evaluations VALUES(6217,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(6217,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(6217,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(6217,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(6217,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6217,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(6217,36,'cp 20','c5b6 a2a4 g8f6 e1g1 e8g8 b2b4 a7a6 b1d2 d7d6 h2h3 b6a7 f1e1 c8e6 c4b3 f8e8 d2f1 e6b3 d1b3 d6d5 b3c2 d8d6 c1e3 d5d4 c3d4 a7d4 e3d4 c6d4 f3d4 d6d4 f1e3 a8d8 e1d1 d4d7 a1b1 e8e6 e3f5 e6c6 c2b3 d7e6 b3e6 c6e6 g1f1 c7c6');
INSERT INTO evaluations VALUES(20977,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(20977,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(20977,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(20977,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(20977,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(20977,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(20977,36,'cp 55','g1f3 e7e6 b1c3 g8f6 e2e3 f8e7 a2a3 e8g8 b2b4 c6b8 c4c5 c7c6 f1d3 a7a5 c1b2 b7b6 c3a4 b6c5 b4c5 c8a6 d3a6 a8a6 e1g1 b8d7 d1c2 d8c7 b2c3 f8b8 f1b1 a6a8 a4b2 f6e4 b2d3 e7f6 h2h3 e4c3 c2c3');
INSERT INTO evaluations VALUES(57213,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(57213,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57213,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57213,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57213,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57213,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(57213,36,'cp 0','c2c4 g8f6 b1c3 b8c6 g2g3 d7d5 c4d5 f6d5 g1f3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 c1e3 e8g8 a1c1 f8d8 a2a3 a7a5 f3d2 a5a4 e3c5 e7g5 c5e3');
INSERT INTO evaluations VALUES(57125,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57125,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57125,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57125,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57125,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57125,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57125,36,'cp 0','b8c6 g1f3 g8f6 b1c3 d7d5 c4d5 f6d5 g2g3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 c1e3 e8g8 a1c1 f8d8 a2a3 a7a5 f3d2 a5a4 d1c2 e6d5 g2d5 d7d5 d2e4 h7h6 e4c3 d5a5 c3e4');
INSERT INTO evaluations VALUES(22737,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22737,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22737,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22737,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22737,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22737,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22737,36,'cp 12','g1f3 g8f6 g2g3 d7d5 c4d5 f6d5 b1c3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 c1e3 e8g8 a1c1 f8d8 a2a3 a7a5 d1c2 a5a4 f3d2 e6d5 g2d5 d7d5 d2e4 h7h6 e4c5 e7c5 c2c5 f7f6 c1c2 d5c5 c2c5 a8a5 f1c1 d8d5 c5c4 d5b5 c1c2 b5b3 g1g2 g8f7 f2f4 a5b5 f4e5');
INSERT INTO evaluations VALUES(22736,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(22736,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22736,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(22736,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(22736,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(22736,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22736,36,'cp 6','g8f6 b1c3 f8e7 g2g3 d7d5 c3d5 f6d5 c4d5 d8d5 f1g2 c8e6 e1g1 d5d7 b2b4 e7b4 f3g5 e6f5 a1b1 h7h6 g5f3 a7a5 a2a3 b4d6 b1b7 e8g8 c1b2 a5a4 d1a1 a8e8 b2c3 f5h3 g2h3 d7h3 b7b5 c6d4 c3d4 e5d4 f3d4');
INSERT INTO evaluations VALUES(16457,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(16457,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(16457,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(16457,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(16457,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(16457,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(16457,36,'cp 6','b1c3 d7d5 c4d5 f6d5 g2g3 f8e7 c3d5 d8d5 f1g2 e8g8 e1g1 d5d6 d1c2 c8e6 a2a3 a7a5 c1e3 a5a4 a1c1 d6d7 f3d2 f8b8 f1e1 h7h6 d2e4 e6d5 e3c5 e7g5 e4g5 h6g5 g2d5 d7d5 c2c4 d5d7 c1c2 b7b5 c4e4');
INSERT INTO evaluations VALUES(16453,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(16453,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(16453,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16453,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(16453,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16453,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(16453,36,'cp 7','d7d5 c4d5 f6d5 g2g3 f8e7 c3d5 d8d5 f1g2 e8g8 e1g1 d5d6 a2a3 a7a5 c1g5 c6d4 g5e7 d6e7 f3d4 e5d4 f1e1 e7d6 d1d2 c7c5 e2e3 d4e3 e1e3 a8a7 d3d4 b7b6 d4d5 a7d7 a1d1 d7d8 g2f3 f8e8 e3e8 d8e8 d2d3 h7h6 d3c3 c8h3 f3g2');
INSERT INTO evaluations VALUES(15099,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(15099,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(15099,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15099,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(15099,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15099,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(15099,36,'cp 2','c4d5 f6d5 g2g3 f8e7 c3d5 d8d5 f1g2 a7a5 e1g1 a5a4 c1e3 d5b5 b2b3 e8g8 d1c2 a4b3 a2b3 c8e6 f3d2 f7f5 g2c6 b7c6 a1a8 f8a8 f1c1 e6d5 d2c4 e7d6 e3d2 a8b8 b3b4 e5e4 c4d6 c7d6 d3e4 d5e4 c2c4 b5c4 c1c4 c6c5 b4c5 b8b1 c4c1 b1c1 d2c1 d6c5');
INSERT INTO evaluations VALUES(15005,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15005,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(15005,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(15005,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15005,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15005,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15005,36,'cp 0','f6d5 g2g3 f8e7 c3d5 d8d5 f1g2 a7a5 e1g1 a5a4 c1e3 d5b5 b2b3 e8g8 b3a4 b5a4 d1a4 a8a4 f1c1 e7f6 f3d2 e5e4 a1b1 e4d3 e2d3 a4a2 d2e4 f6d4 e3d4 c6d4 c1c7 b7b6 e4c3 a2d2 g2e4 c8h3 e4g2');
INSERT INTO evaluations VALUES(15348,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15348,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(15348,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15348,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15348,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(15348,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(15348,36,'cp 4','g2g3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 a2a3 f7f6 d1a4 e8g8 c1e3 a8d8 f1c1 c6d4 a4d7 d8d7 f3d4 e5d4 e3d2 c7c6 g1f1 g7g5 d2b4 e7b4 a3b4 a7a6 h2h4 g5h4 g3h4 g8g7 a1a5 d7e7 g2f3 f8g8 f1e1 g7f7');
INSERT INTO evaluations VALUES(15351,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(15351,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(15351,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(15351,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(15351,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(15351,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15351,36,'cp 0','f8b4 c1d2 d5c3 d2c3 b4c3 b2c3 d8f6 f3d2 e8g8 f1g2 c8d7 d1a4 a8d8 e1g1 b7b6 a4c4 h7h6 c4a6 d7g4 a6a4 g4d7');
INSERT INTO evaluations VALUES(9182,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9182,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9182,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(9182,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(9182,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(9182,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9182,36,'cp 0','c1d2 d5c3 d2c3 b4c3 b2c3 d8f6 d1c2 c8d7 f1g2 a8d8 e1g1 e8g8 f3d2 b7b6 d2c4 h7h6 f1d1 c6e7 a2a4 a7a5 e2e4 e7c6 d1d2 d7e6 c4e3 d8d7 a1d1 f8d8 c2b2 g8h8 g2f1 h8g8');
INSERT INTO evaluations VALUES(9181,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9181,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9181,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9181,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9181,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9181,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9181,36,'cp 0','d5c3 b2c3 b4d6 f1g2 e8g8 e1g1 h7h6 d1c2 a8b8 d3d4 d8f6 d4d5 c6d8 f3h4 b7b6 c3c4 d8b7 d2c3 c8d7 e2e4 f6g5 c3d2 g5e7 a1e1 b8e8 g1h1 b7c5 d2e3 d7c8 e3d2');
INSERT INTO evaluations VALUES(55177,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55177,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55177,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55177,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55177,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55177,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(55177,36,'cp 6','g1f3 b8c6 d2d3 c8g4 f1e2 e7e6 e1g1 g8e7 a2a3 e7g6 f3d4 g4e2 d4e2 c6e5 d3d4 e5d7 c2c4 f8d6 h2h3 e8g8 d1c2 c7c6 c1e3 d6c7 b1d2 d8e7 f1e1 a8d8 a1d1 h7h6 d2f3 f7f5 g2g3 d7f6 e3f4 g6f4 e2f4 c7f4 g3f4 f6h5');
INSERT INTO evaluations VALUES(55174,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55174,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(55174,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(55174,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55174,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(55174,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(55174,36,'cp 0','b8c6 f1b5 e7e6 c2c3 d4d3 d1a4 d8d5 a4c4 a7a6 c4d5 e6d5 b5d3 c8g4 e1g1 e8c8 f1e1 d8e8 d3e2 g4f3 e2f3 c6e5 g1f1 g8f6 f3e2 h7h5 d2d3 f8d6 h2h3 e5g6 b1d2 g6f4 d2f3 f4e2 e1e2 f6d7 c1e3 d7e5 f3e5 e8e5 e3d4 e5e2 f1e2 h8e8 e2f1 d6f8 b2b4 c8d7 a2a4');
INSERT INTO evaluations VALUES(21097,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(21097,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21097,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(21097,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(21097,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21097,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21097,36,'cp 2','f1b5 e7e6 b5c6 b7c6 d2d3 g8e7 c1g5 c8a6 d1e2 d8d5 h2h4 e7f5 b1d2 h7h6 g5f4 c6c5 h4h5 e8c8 b2b3 f5e7 d2c4 d5c6 f4d2 e7d5 e1d1 f7f5 e5f6 g7f6 f3h4 h8g8 h4g6 d8e8 g6f8 g8f8 d2h6 f8g8');
INSERT INTO evaluations VALUES(20944,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(20944,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20944,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20944,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20944,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20944,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20944,36,'cp 0','e7e6 b5c6 b7c6 d2d3 g8e7 c1g5 c8a6 e1g1 c6c5 b1d2 a6b7 d2e4 h7h6 g5e7 f8e7 f3d2 d8d5 f2f4 h6h5 f1f2 e7h4 f2e2 h4e7');
INSERT INTO evaluations VALUES(53379,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(53379,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53379,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(53379,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(53379,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53379,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53379,36,'cp 0','b8c6 g1f3 a7a6 c3e2 c8f5 c2c3 e7e6 e2g3 f5g6 c1f4 g8f6 f1d3 f8d6 f4d6 d8d6 e1g1 a8c8 d3g6 h7g6 f1e1 d6f4 h2h4 f6e4 g3f1 h8h4 g2g3 h4g4 d1e2 e8f8 f3h2 g4g5 h2f3');
INSERT INTO evaluations VALUES(19853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19853,36,'cp 0','g1f3 c8g4 h2h3 g4f3 d1f3 g8f6 c1e3 a8c8 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(19851,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19851,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19851,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19851,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19851,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19851,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19851,36,'cp 0','c8g4 h2h3 g4f3 d1f3 g8f6 c1e3 e7e6 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(26297,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26297,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26297,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26297,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26297,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26297,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26297,36,'cp 0','h2h3 g4f3 d1f3 e7e6 c1e3 a8c8 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(26294,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26294,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26294,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26294,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26294,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26294,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26294,36,'cp -9','g4f3 d1f3 e7e6 c1e3 d8b6 e1c1 a8c8 c1b1 g8f6 f3g3 c6e7 e3c1 e7f5 g3d3 h7h5 d3b5 b6b5 f1b5 e8d8 h1f1 h5h4 c3e2 f5d6 b5d3 b7b5 c2c3 d8d7 d3c2 g7g6 c2d3 d7e8 d1e1 a7a6 e2f4 f6e4');
INSERT INTO evaluations VALUES(53271,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(53271,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(53271,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(53271,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(53271,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(53271,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(53271,36,'cp 44','c2c4 d5c4 d4d5 e7e6 e2e4 e6d5 e4d5 g8f6 f1c4 f8d6 d1e2 d8e7 b1c3 e7e2 c4e2 b8d7 c1e3 e8g8 e1c1 f8d8 h2h3 d7b6 a2a4 a7a5 h1e1 c8d7 b2b3 h7h6 g2g4 d8e8 c1b2 a8c8 e2b5 e8d8 b5c4 d8e8 e3d2 b6c4 b3c4 e8e1 d1e1 c8e8 e1e8 f6e8 c3e4');
INSERT INTO evaluations VALUES(53246,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(53246,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(53246,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(53246,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(53246,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(53246,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(53246,36,'cp 51','d5c4 e2e4 c5d4 d1d4 d8d4 f3d4 c8d7 f1c4 b8c6 c1e3 e7e6 b1d2 c6d4 e3d4 g8e7 a1c1 e7c6 d4e3 f8e7 f2f4 e8g8 e1f2 f8d8 d2f3 d7e8 a2a3 g8f8 e4e5 a8c8 c4e2 f7f6 b2b4 f6e5 f4e5 e8g6 h1d1 d8d1 c1d1 f8e8 b4b5 c6a5 e3a7');
INSERT INTO evaluations VALUES(53202,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(53202,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(53202,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(53202,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(53202,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(53202,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(53202,36,'cp 36','e2e4 c5d4 d1d4 d8d4 f3d4 a7a6 f1c4 e7e5 d4f3 b8c6 e1g1 g8f6 f3g5 c6d8 b1c3 f8b4 c4b3 c8d7 g5f3 d8c6 c3d5 f6d5 b3d5 b4c5 c1e3 c5e3 f2e3 f7f6 f3d2 e8c8 a1c1 c8b8 d2b3 c6e7 b3c5 d7c6 d5c6 e7c6 c5a6 b7a6 c1c6');
INSERT INTO evaluations VALUES(53203,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(53203,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(53203,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53203,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(53203,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(53203,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(53203,36,'cp 23','c5d4 d1d4 d8d4 f3d4 a7a6 f1c4 e7e5 d4f3 b8c6 e1g1 g8f6 f3g5 c6d8 b1c3 f8c5 c1e3 c5e3 f2e3 c8e6 c3d5 e6d5 e4d5 b7b5 c4b3 e8g8 a1c1 d8b7 g5f3 f8e8 c1c7 b7a5 d5d6 a5b3 a2b3 e8d8 f1d1 f6e4 f3e5 d8d6 d1d4 d6d4 e3d4 a8d8 e5c6 d8e8 b3b4 e4f6');
INSERT INTO evaluations VALUES(53461,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(53461,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(53461,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(53461,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(53461,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(53461,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(53461,36,'cp 31','d1d4 d8d4 f3d4 a7a6 f1c4 e7e5 d4c2 b8c6 b1c3 b7b5 c4b3 g8f6 c1e3 c8d7 e1g1 c6a5 f2f3 f8e7 f1c1 a5b3 a2b3 e8g8 b3b4 f8c8 e3c5 e7c5 b4c5 c8c5 c2b4 c5c8 b4d3 h7h6 d3e5 d7e8 e5d3 a6a5 g1f2 a5a4 c3e2 c8c1 a1c1');
INSERT INTO evaluations VALUES(44909,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44909,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(44909,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(44909,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44909,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(44909,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(44909,36,'cp 23','g1f3 c8f5 f3h4 f5e6 b1d2 g7g5 h4f3 g5g4 c4d5 c6d5 f3e5 b8d7 d2b3 f8g7 e5d7 e6d7 b3c5 f6e4 c5e4 d5e4 c1d2 a8c8 d1b3 e8g8 f1e2 c8c7 b3a3 h7h5 a3a7 e7e5 d2c3 e5d4 c3d4 d7e6 e1g1 g7d4 a7d4 d8d4 e3d4');
INSERT INTO evaluations VALUES(44910,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44910,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(44910,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44910,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44910,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(44910,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(44910,36,'cp 10','c8f5 f1d3 f5g6 e1g1 e7e6 d3g6 h7g6 b1d2 b8d7 d1c2 f8b4 f1d1 e8g8 h2h3 a7a5 e3e4 b4d2 f3d2 d5e4 d2e4 f6e4 c2e4 a5a4 c1f4 f8e8 e4e3 d8a5 b2b3 a4b3 e3b3 b7b5 c4b5 c6b5 d1b1 d7f6 b3b5 a5b5 b1b5 f6d5 f4g3 a8a4 b5b8 e8b8 g3b8 a4d4 b8e5 d4b4');
INSERT INTO evaluations VALUES(28969,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(28969,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(28969,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(28969,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(28969,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(28969,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(28969,36,'cp 12','f1d3 f5g6 e1g1 e7e6 d3g6 h7g6 b1d2 f8d6 h2h3 e8g8 d1c2 b8d7 e3e4 f6e4 d2e4 d5e4 c2e4 f8e8 c1f4 d6f4 e4f4 d8b8 f4h4 b8d8 h4e4 d8c7 f1e1 a8d8 b2b4 a7a5 a2a3 d8a8 g2g3 b7b5 a1c1 a5b4 a3b4 d7f6 e4e5 c7b7');
INSERT INTO evaluations VALUES(28966,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(28966,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(28966,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(28966,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(28966,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(28966,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(28966,36,'cp 12','f5g6 e1g1 e7e6 d1c2 b8d7 d3g6 h7g6 b1d2 c6c5 c4d5 f6d5 d4c5 f8c5 d2b3 c5b6 e3e4 d8c7 c2c7 d5c7 a2a4 c7a6 a4a5 d7c5 b3c5 b6c5 c1d2 f7f6 a1a4 e8f7 a4c4 c5e7 f1c1 e6e5 g1f1 g6g5 h2h3 h8e8 f1e2 a8d8');
INSERT INTO evaluations VALUES(44572,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(44572,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(44572,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(44572,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(44572,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(44572,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(44572,36,'cp 57','e4e5 f6d5 b1c3 e7e6 c3d5 e6d5 d2d4 d7d6 f1b5 b8c6 c2c3 f8e7 e5d6 d8d6 d4c5 d6c5 b5d3 c8g4 c1e3 c5a5 d1b3 e8c8 f3d4 c6d4 e3d4 e7c5 e1g1 c5d4 c3d4 a5b6 b3b6 a7b6 a2a4 g4e6 f1e1 c8b8');
INSERT INTO evaluations VALUES(44514,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(44514,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(44514,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(44514,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(44514,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(44514,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(44514,36,'cp 67','f6d5 b1c3 e7e6 c3d5 e6d5 d2d4 b8c6 d4c5 f8c5 d1d5 d7d6 e5d6 d8b6 d5e4 c8e6 e4h4 c5d6 f1e2 e6f5 c2c3 e8g8 e1g1 a8e8 h4c4 f5e4 a2a4 c6a5 c4a2 d6b8 c1e3 b6g6 a1d1 e4c2 d1d2 c2b3 a2b1 b3a4 b1g6 h7g6 f3d4 a5b3 d4b3 a4b3 d2d7');
INSERT INTO evaluations VALUES(44666,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(44666,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(44666,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(44666,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(44666,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(44666,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(44666,36,'cp 73','b1c3 e7e6 c3d5 e6d5 d2d4 b8c6 d4c5 f8c5 d1d5 d7d6 e5d6 d8b6 d5e4 c8e6 e4h4 c5d6 f1e2 e6f5 e1g1 e8g8 c2c3 a8e8 h4c4 f5g6 g2g3 h7h6 e2d3 g6h5 c4d5 h5g4 c1e3 b6b2 f3d4 d6e5 d5e4 f7f5 e4d5 g8h8 d5c5 b2b6 c5b6 a7b6 d4c6 b7c6 e3b6 e5c3');
INSERT INTO evaluations VALUES(44663,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(44663,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(44663,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(44663,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(44663,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(44663,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(44663,36,'cp 51','e7e6 c3d5 e6d5 d2d4 d7d6 f1b5 b8c6 e1g1 f8e7 c2c4 a7a6 e5d6 d8d6 d4c5 d6c5 c1e3 c5d6 c4c5 d6d8 b5a4 e8g8 h2h3 c8f5 f3d4 d8d7 d4f5 d7f5 b2b4 a8d8 a4c2 f5f6 a1b1 g7g6 a2a4 f8e8 c2d3 e7f8 b4b5 a6b5 b1b5');
INSERT INTO evaluations VALUES(44142,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(44142,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(44142,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(44142,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(44142,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(44142,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(44142,36,'cp 64','c3d5 e6d5 d2d4 b8c6 d4c5 f8c5 d1d5 d7d6 e5d6 d8b6 d5e4 c8e6 e4h4 c5d6 c2c3 d6e7 h4f4 e8g8 f1e2 a8d8 e1g1 f8e8 f4a4 e6d5 a4c2 e7f6 c1f4 h7h6 f1d1 d5e4 c2b3 d8d1 e2d1 b6a6 f4e3 c6a5 b3b4 a5c4 e3d4 f6d4 c3d4 e4d5 d1b3');
INSERT INTO evaluations VALUES(44132,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(44132,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(44132,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(44132,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(44132,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(44132,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(44132,36,'cp 40','e6d5 d2d4 b8c6 d4c5 f8c5 d1d5 d7d6 e5d6 d8b6 d5e4 c8e6 e4h4 c5d6 f1e2 e6f5 c2c3 e8g8 h4g5 f5g6 g5b5 b6b5 e2b5 f8e8 c1e3 d6c5 f3d4 c5d4 c3d4 a7a6 b5c6 b7c6 e1d2 a6a5 h1c1 g6e4 g2g3 a5a4 b2b4 a4b3 a2b3 f7f6 a1a8 e8a8 h2h4 a8a2 d2c3 e4d5 c3b4 g8f7');
INSERT INTO evaluations VALUES(44361,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(44361,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(44361,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(44361,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44361,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(44361,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(44361,36,'cp 52','d2d4 d7d6 f1b5 b8c6 d1e2 f8e7 e5d6 d8d6 d4c5 d6c5 c1e3 c5d6 c2c3 e8g8 a1d1 c8f5 b5d3 f5g6 d3g6 h7g6 e1g1 a8d8 d1d2 d8d7 f1d1 d6e6 d1e1 f8d8 g2g3 e7f6 f3d4 f6d4 e3d4 d8e8 e2e6 f7e6 d4e3 a7a6');
INSERT INTO evaluations VALUES(50209,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(50209,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(50209,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(50209,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(50209,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(50209,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(50209,36,'cp 65','e2e4 d7d6 c1e3 c7c6 g1f3 f8g7 h2h3 e8g8 a2a4 b8d7 f1e2 e7e5 d4e5 d7e5 e1g1 e5f3 e2f3 c8e6 d1d2 d8c7 e3d4 a8d8 a4a5 b7b5 a5b6 a7b6 f1d1 f6d7 d4g7 g8g7 d2d4 d7f6 b2b4 d6d5 e4d5 e6d5 f3d5 c6d5 b4b5');
INSERT INTO evaluations VALUES(50232,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(50232,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(50232,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(50232,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50232,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(50232,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(50232,36,'cp 73','d7d6 c1e3 c7c6 g1f3 f8g7 h2h3 e8g8 a2a4 b8d7 f1e2 e7e5 d4e5 d7e5 e1g1 d8c7 d1d2 e5f3 e2f3 c8e6 a4a5 b7b5 a5b6 a7b6 e3f4 a8a1 f1a1 f6e8 f4h6 g7h6 d2h6 e8f6 h6d2 g8g7 d2d4 f8d8 a1d1 d8d7 d4e3 d7d8 c3e2 d6d5');
INSERT INTO evaluations VALUES(34544,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34544,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(34544,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(34544,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(34544,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(34544,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34544,36,'cp 70','g1f3 c7c6 d2d4 g8f6 f1d3 c8g4 c1e3 g4h5 h2h3 d6c7 d1e2 e7e6 e1c1 f6d5 c3d5 c6d5 g2g4 h5g6 f3e5 g6d3 e5d3 f8d6 c1b1 b8c6 c2c3 e8c8 e3c1 c8b8 f2f4 a7a6 e2f3 g7g6 f4f5 g6f5 g4f5 h8g8 f5e6 f7e6 b1a1');
INSERT INTO evaluations VALUES(34541,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(34541,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(34541,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(34541,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(34541,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(34541,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(34541,36,'cp 72','g8f6 d2d4 c7c6 f3e5 g7g6 c1f4 f6h5 f4e3 h5f6 f1d3 b8d7 f2f4 f8g7 e1g1 e8g8 a2a4 d7b6 g1h1 b6d5 c3d5 c6d5 a4a5 f6e4 c2c4 f7f6 e5f3 c8f5 c4c5 d6d7 a1c1 f5g4 h1g1 e7e6 b2b4 a7a6 f1e1 a8c8');
INSERT INTO evaluations VALUES(33661,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(33661,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(33661,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(33661,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(33661,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(33661,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(33661,36,'cp 85','d2d4 c7c6 f1e2 c8g4 h2h3 g4f3 e2f3 b8d7 e1g1 e7e6 c3e2 d6c7 c2c4 f8e7 c1f4 e7d6 f4e3 e8g8 c4c5 d6e7 e3f4 c7c8 f1e1 f8d8 d1c2 d7f8 a1d1 d8d7 f4g3 c8d8 h3h4 f6d5 h4h5 a7a6 e2c3 d5b4 c2b3 d7d4 d1d4 d8d4 e1d1');
INSERT INTO evaluations VALUES(33657,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(33657,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(33657,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(33657,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(33657,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(33657,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(33657,36,'cp 70','c7c6 f1d3 c8g4 c1e3 g4h5 h2h3 d6c7 d1e2 e7e6 e1c1 f6d5 c3d5 c6d5 g2g4 h5g6 f3e5 g6d3 e5d3 b8c6 c1b1 e8c8 h3h4 f8d6 h4h5 c8b8 f2f4 d8c8 c2c3 c6a5 d3e5 b8a8 e3c1 a7a6 g4g5 a5c4 d1e1 d6e5 f4e5 c7b6 h5h6');
INSERT INTO evaluations VALUES(33603,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(33603,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(33603,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(33603,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(33603,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(33603,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(33603,36,'cp 77','f1d3 c8g4 c1e3 g4h5 h2h3 e7e6 d1e2 d6c7 e1c1 f6d5 c3d5 c6d5 g2g4 h5g6 f3e5 g6d3 e5d3 f8d6 h3h4 b8c6 c1b1 e8c8 c2c3 h8e8 f2f4 c6a5 h4h5 c8b8 e3c1 a7a6 g4g5 d8c8 h1h3 b8a8 d1e1 a5c4 d3e5 a6a5 e1h1 a8b8 e5c4');
INSERT INTO evaluations VALUES(47356,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(47356,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(47356,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(47356,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(47356,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(47356,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(47356,36,'cp -34','e7e5 g1f3 e5d4 d1d4 d8d4 f3d4 f8c5 c1e3 f6g4 c3d5 g4e3 f2e3 e8d8 f1c4 b8d7 d5f4 d7e5 a1d1 d8e8 c4b3 c7c6 f4d3 e5d3 d1d3 g7g6 e4e5 h8f8 h2h3 f7f6 e5f6 f8f6 e1e2 e8e7 g2g4 c8d7 h1d1 f6d6 d4f3 a8f8 a2a4 d6d3 d1d3 a7a5 f3g5 h7h6 g5f7');
INSERT INTO evaluations VALUES(48917,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(48917,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(48917,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(48917,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(48917,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(48917,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(48917,36,'cp -38','g1f3 e5d4 d1d4 d8d4 f3d4 f8c5 c1e3 f6g4 c3d5 g4e3 f2e3 e8d8 f1c4 b8c6 e1c1 c6e5 d5f4 d8e7 f4d5 e7d7 c4e2 c7c6 b2b4 c6d5 b4c5 d5e4 d4f3 e5d3 e2d3 e4d3 f3d4 d7c7 h1f1 h8f8 d4b5 c7c6 b5d4 c6c5 d1d3 b7b5 d3c3 c5b6 c3c6 b6b7 c6d6 c8e6 d4e6 f7e6 d6d7 b7b6 f1f8 a8f8 d7g7 f8f2 c1b1 a7a5 g7h7 f2g2 h2h4 a5a4 h7e7 a4a3 e7e6 b6a5 e6e7 g2b2 b1a1 b5b4 e7e5 a5b6 e5e6 b6b5 e6e5 b5c4');
INSERT INTO evaluations VALUES(48916,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(48916,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(48916,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(48916,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(48916,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(48916,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(48916,36,'cp -54','e5d4 d1d4 d8d4 f3d4 f8c5 c1e3 f6g4 c3d5 g4e3 f2e3 e8d8 f1c4 b8d7 d5f4 d7e5 e1c1 d8e7 f4d5 e7d7 c4e2 c7c6 b2b4 c6d5 b4c5 d5e4 d4f3 e5d3 e2d3 e4d3 f3d4 d7c7 h1f1 f7f6 d4b5 c7c6 b5d4 c6c5 d1d3 c5d6 d4f5 d6e5 f5g7 h8g8 g7h5 c8f5 d3d2 a8d8 d2f2 f5d3 f1d1 d8d6 h5f4 d3e4 d1d6 e5d6 f2d2 d6e7 h2h3');
INSERT INTO evaluations VALUES(48951,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(48951,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(48951,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(48951,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(48951,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(48951,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48951,36,'cp -36','d1d4 d8d4 f3d4 f8c5 c1e3 f6g4 c3d5 g4e3 f2e3 e8d8 e1c1 b8d7 f1c4 d7e5 d5f4 d8e7 f4d5 e7d7 c4e2 c7c6 b2b4 c6d5 b4c5 d5e4 d4f3 e5d3 e2d3 e4d3 f3d4 d7c7 h1f1 h8f8 d4b5 c7c6 b5d4 c6c5 d1d3 b7b5 d3c3 c5b6 c3c6 b6b7 c6d6 c8e6 d4e6 f7e6 d6d7 b7b6 f1f8 a8f8 d7g7 f8f2 c1b1 a7a5 h2h4 a5a4 g7h7 f2g2 h4h5 b6a5 h7a7 a5b4 h5h6 g2h2 h6h7 h2h4');
INSERT INTO evaluations VALUES(49416,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(49416,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(49416,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(49416,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(49416,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(49416,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49416,36,'cp 38','e2e4 d7d5 e4e5 f6d7 f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 c3e2 d8b6 d1c1 f8e7 c2c3 f7f6 a2a3 a6a5 e5f6 e7f6 c1c2 e8g8 e1c1 f6e7 g2g4 a5a4 h1g1 b6b3 f1h3 b3c2 c1c2 d7b6 d4c5 b6c4 e3f2 c6a5');
INSERT INTO evaluations VALUES(49440,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(49440,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(49440,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(49440,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(49440,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(49440,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(49440,36,'cp 27','d7d5 e4e5 f6d7 f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 a8b8 d2f2 c6d4 e3d4 c5d4 f2d4 d8b6 d4b6 d7b6 e1c1 c8d7 f1d3 b8c8 c3e2 b6a4 h2h4 h7h5 h1h3 e8e7 h3e3 g7g6 g2g3 a4c5 c1d2 c8c7 d1b1 h8c8 e2d4 a6a5 c2c3');
INSERT INTO evaluations VALUES(55840,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(55840,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(55840,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55840,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(55840,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(55840,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55840,36,'cp -20','g8f6 c1f4 c7c5 e2e3 a7a6 d4c5 f8c5 f1d3 b8d7 e3e4 e6e5 f4g5 d8b6 e1g1 d5e4 c3e4 f6e4 d3e4 e8g8 g5h4 h7h6 d1d3 f8e8 a1b1 c5f8 b2b4 a6a5 a2a3 g7g5 h4g3 d7f6 f1d1 b6c7 c2c4 f6e4 d3e4 a5b4 a3b4 f8g7 c4c5 c7c6');
INSERT INTO evaluations VALUES(48321,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(48321,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(48321,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(48321,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48321,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48321,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(48321,36,'cp -21','c1f4 a7a6 e2e3 b8d7 f3d2 b7b5 h2h4 h7h5 f1e2 c7c5 e1g1 c8b7 c3b1 b7c6 d2f3 g7g6 f3g5 c5c4 b1d2 f8g7 f4d6 d7b6 d6a3 a6a5 c2c3 d8c7 d2f3 f6e4 d1c2 g7f6 g5e4 d5e4 f3g5 e8c8 f2f3 e4f3 e2f3');
INSERT INTO evaluations VALUES(48278,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48278,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(48278,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(48278,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(48278,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48278,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(48278,36,'cp -12','f8b4 d1d3 f6e4 f3d2 e4c3 b2c3 b4d6 d3g3 e8g8 e2e4 d6f4 g3f4 c7c5 d4c5 b8d7 d2b3 e6e5 f4f3 d5e4 f3e4 d7f6 e4e3 c8f5 f1d3 f5d3 e3d3 b7b6 e1g1 d8c7 f1e1 b6c5 d3c4 f8e8 a1d1 g7g6 c4c5 c7c5 b3c5 a8c8 c5d7 f6d7 d1d7 c8c3 h2h3 c3c2 d7a7 e8d8 e1e5 d8d1 g1h2 c2f2 h2g3');
INSERT INTO evaluations VALUES(42006,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(42006,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(42006,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42006,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(42006,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(42006,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(42006,36,'cp -7','d1d3 f6e4 f3d2 e4c3 b2c3 b4d6 d3g3 e8g8 e2e4 d6f4 g3f4 c7c5 d4c5 b8d7 d2b3 e6e5 f4e3 d5e4 e3e4 d7f6 e4e3 c8f5 f1d3 f5d3 e3d3 b7b6 e1g1 d8c7 f1e1 b6c5 d3c4 a8c8 b3d2 f8e8 a1d1 g7g6 h2h3 g8g7 d2e4 f6e4 e1e4 f7f5 c4b5');
INSERT INTO evaluations VALUES(19315,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(19315,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(19315,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(19315,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(19315,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(19315,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(19315,36,'cp 35','f8g7 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8d8 e1e7 a7a6 a2a4 b5e8 d4c3 d8d5 h2h4 a8b8 b3d4 h6g7 d4f3 g7c3 b2c3 d5c5 e7a7 e8c6 f3d4 c6b7 a1c1 b8d8 a7b7 d8d4 c3d4');
INSERT INTO evaluations VALUES(30923,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(30923,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30923,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(30923,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(30923,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(30923,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(30923,36,'cp 36','b5d7 d8d7 c2c4 b8c6 d2d4 e7e5 d4c5 d6c5 d1d7 e8d7 c1e3 f7f6 b1c3 c6d4 e1c1 a8c8 b2b4 c5b4 c3e2 c8c4 c1b1 g8h6 e2d4 e5d4 f3d4 h6f7 d4b3 d7e6 f2f3 b7b6 d1d5 f7e5 d5d8 c4c3 e3d4 e5c6 d8a8 c6d4 b3d4 e6e5');
INSERT INTO evaluations VALUES(30864,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(30864,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(30864,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(30864,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(30864,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(30864,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(30864,36,'cp 20','b8d7 e1g1 g8f6 d1e2 a7a6 d2d4 c5d4 f3d4 d8a5 c2c4 a5e5 c1e3 e5e4 b1c3 e4g6 f2f4 g6h5 e2h5 f6h5 g2g4 h5f6 g4g5 f6g4 e3d2 e7e5 d4f3 e5f4 a1e1 g4e5 d2f4 e8c8 f4e5 d6e5 f3e5 d7e5 e1e5 f8d6 e5e2 f7f6 g5f6 g7f6 g1g2 h8e8 c3e4 d6e5 e4f6 e5f6 e2e8 d8e8 f1f6');
INSERT INTO evaluations VALUES(26207,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26207,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(26207,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26207,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(26207,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(26207,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26207,36,'cp 23','c2c4 e7e6 b1c3 a8c8 e1g1 a7a6 d2d4 c5d4 d1d4 g8f6 b2b3 f8e7 d4e3 e8g8 c1d2 d8c7 a1c1 b7b6 f3d4 f8e8 h2h3 e7f8 f1d1 c7b8 d1e1 d7c5 e3e2 d6d5 c4d5 e6d5 e4e5 b6b5 d2f4 b5b4 c3a4 c5a4');
INSERT INTO evaluations VALUES(26200,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(26200,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(26200,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(26200,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(26200,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(26200,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(26200,36,'cp 24','e7e6 b1c3 a7a6 e1g1 a8c8 d2d4 c5d4 d1d4 g8f6 b2b3 f8e7 c1d2 d8c7 d4e3 e8g8 h2h3 f8e8 e3e2 b7b6 f1d1 c7b7 a1c1 d7c5 d1e1 h7h6 c1d1 c5d7 f3d4 e7f8 e2e3 d7c5 e3f3 c5d7 f3e2 b7a7 e2e3 a7b7');
INSERT INTO evaluations VALUES(44900,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44900,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44900,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(44900,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44900,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44900,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(44900,36,'cp 18','b8d7 g1f3 e7e6 d1c2 f8d6 f1d3 e8g8 e1g1 d5c4 d3c4 a7a6 e3e4 e6e5 f1d1 e5d4 f3d4 d8c7 h2h3 f8e8 a2a4 d7c5 f2f3 c8e6 c4f1 a8d8 c3e2 e6c8 c1e3 c5e6 d4f5 d6f8 d1d8 e6d8 e2d4 c6c5 e3f2 d8c6 d4c6 c8f5 c6d4 f5d7 d4b3 e8c8 a4a5 c5c4 b3d4');
INSERT INTO evaluations VALUES(12235,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(12235,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(12235,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(12235,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(12235,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(12235,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(12235,36,'cp 24','g1f3 e7e6 d1c2 f8d6 f1d3 d5c4 d3c4 e8g8 e1g1 a7a6 e3e4 e6e5 c1e3 b7b5 c4e2 f8e8 h2h3 e5d4 f3d4 d8c7 a1c1 c8b7 d4f5 d6c5 f1d1 c5e3 f5e3 d7c5 e2f3 a8d8 g2g3 f6d7 f3g2 d7e5 c3b5 d8d1 e3d1 a6b5 c2c5');
INSERT INTO evaluations VALUES(12236,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(12236,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12236,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(12236,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(12236,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(12236,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12236,36,'cp 25','e7e6 d1c2 f8d6 f1d3 e8g8 e1g1 d5c4 d3c4 a7a6 a2a4 c6c5 f1d1 c5d4 e3d4 d8c7 c2e2 b7b6 h2h3 c8b7 c1e3 d6f4 e3f4 c7f4 d4d5 e6d5 c3d5 f6d5 c4d5 b7d5 d1d5 f8e8 e2c2 a8c8 c2b3 d7f6 d5d1 f4c7 a1c1 c7c1 d1c1 c8c1 g1h2 e8d8 b3b6');
INSERT INTO evaluations VALUES(12159,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(12159,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12159,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12159,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(12159,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(12159,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12159,36,'cp 21','d1c2 f8d6 f1d3 e8g8 e1g1 d5c4 d3c4 a7a6 a2a4 c6c5 f1d1 c5d4 e3d4 d8c7 c2e2 b7b6 d4d5 d7e5 f3e5 d6e5 c1g5 e5h2 g1h1 h2e5 a1c1 c7b8 c4a2 e6d5 g5f6 e5f6 c3d5 b8e5 e2e5 f6e5 d5b6 a8b8 b6c8 b8c8 c1c8 f8c8 g2g3 e5b2');
INSERT INTO evaluations VALUES(12160,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12160,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(12160,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12160,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(12160,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(12160,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12160,36,'cp 9','f8d6 f1d3 e8g8 e1g1 d5c4 d3c4 a7a6 a2a4 c6c5 f1d1 c5d4 e3d4 d8c7 c2e2 b7b6 d4d5 d7e5 f3e5 d6e5 c1g5 e5h2 g1h1 h2e5 a1c1 c7b8 g5f6 e5f6 d5e6 c8e6 c4e6 f7e6 e2e6 g8h8 c3d5 f6b2 c1c2 b8e5 e6e5 b2e5 d5b6 a8d8 b6d7');
INSERT INTO evaluations VALUES(7750,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(7750,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(7750,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(7750,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(7750,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(7750,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(7750,36,'cp 21','f1d3 e8g8 e1g1 d5c4 d3c4 a7a6 e3e4 e6e5 f1d1 d8c7 h2h3 b7b5 c4b3 c8b7 d4e5 d7e5 f3d4 f8d8 c1g5 c6c5 g5f6 g7f6 d4f5 c5c4 d1d6 d8d6 c2c1 g8h8 b3c2 e5g6 f5d6 c7d6 c1e3 a8g8 a1d1 d6e5 e3b6 g6f4 b6b7 g8g2 g1f1');
INSERT INTO evaluations VALUES(19616,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(19616,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(19616,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(19616,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(19616,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(19616,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(19616,36,'cp 80','d1d4 e7e6 b1c3 a7a6 f1e2 d8c7 c1e3 b7b5 e1c1 c8b7 d1d2 g8f6 e4e5 f6d5 c3d5 b7d5 h1d1 f8e7 e2f3 d5f3 g2f3 a8d8 f3f4 e8g8 d4a7 c7c4 c1b1 g7g6 a2a3 a6a5 f4f5 e6f5 e3b6 d8b8 b6a5 f8c8 a7d7 e7c5 a5c7 b8a8');
INSERT INTO evaluations VALUES(38808,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38808,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(38808,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(38808,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38808,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38808,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(38808,36,'cp 37','d1a4 b8c6 e2e3 g8f6 c1d2 e8g8 a2a3 b4c3 d2c3 f6e4 a4c2 c6e7 g1f3 b7b6 f1d3 e4c3 b2c3 h7h6 c4d5 e6d5 e1g1 d8d6 c3c4 d5c4 d3c4 c7c5 f1d1 c8g4 c2e4 g4f5 e4b7 f8b8 b7a6 b8d8 a6a4 c5d4 d1d4');
INSERT INTO evaluations VALUES(38846,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(38846,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(38846,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38846,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38846,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(38846,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(38846,36,'cp 32','b8c6 e2e3 g8f6 c1d2 e8g8 a2a3 b4c3 d2c3 f6e4 a4c2 b7b6 g1f3 c6e7 f1d3 e4c3 b2c3 h7h6 c4d5 e6d5 e1g1 d8d6 c3c4 d5c4 d3c4 c7c5 f1d1 c8g4 c4e2 a8c8 d4c5 d6c5 c2a4 c5c2 a4c2 c8c2 e2a6 c2c7 d1c1 c7c1 a1c1');
INSERT INTO evaluations VALUES(5830,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(5830,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(5830,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(5830,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(5830,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(5830,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(5830,36,'cp 32','e2e3 g8f6 c1d2 e8g8 a2a3 b4c3 d2c3 c6e7 a1c1 b7b6 g1f3 f6e4 f1e2 a7a5 e1g1 c8a6 f1e1 h7h6 f3e5 d5c4 e2c4 a6c4 e5c4 a8c8 c4e5 e7d5 e1d1 d8g5 c3e1 c7c5 e5d7 d5e3 f2e3 g5e3 g1h1 c5d4 c1c8 f8c8 a4d4');
INSERT INTO evaluations VALUES(43397,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(43397,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(43397,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(43397,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(43397,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(43397,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43397,36,'cp 0','e7e6 d1d2 b8d7 e1c1 b7b5 c3d5 e6d5 e4d5 d7e5 f2f4 e5g6 f1d3 c8d7 h1e1 g6e7 d4f5 d8b6 g5f6 g7f6 f5e7 f8e7 d2e2 b6d8 e2h5 d8b6');
INSERT INTO evaluations VALUES(43511,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43511,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43511,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43511,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43511,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(43511,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43511,36,'cp 0','d1f3 b8d7 e1c1 d8c7 f1d3 f8e7 f3g3 b7b5 h1e1 h7h6 g5d2 b5b4 c3e2 a8b8 e2f4 d7c5 g3g7 h8g8 g7h6 f6g4 h6h5 g4f6 h5h6');
INSERT INTO evaluations VALUES(43508,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43508,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43508,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43508,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43508,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43508,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43508,36,'cp 0','b8d7 e1c1 d8c7 f1d3 b7b5 h1e1 f8e7 f3g3 h7h6 g5d2 b5b4 c3e2 a8b8 e2f4 d7c5 g3g7 h8g8 g7h6 f6g4 h6h7 g4f6');
INSERT INTO evaluations VALUES(52481,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52481,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52481,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52481,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52481,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52481,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52481,36,'cp 20','c2c4 g7g6 b1c3 g8f6 d2d4 c5d4 e3d4 f8g7 d4d5 e8g8 g1f3 e7e6 f1e2 e6d5 c4d5 d7d6 c1f4 f8e8 e1g1 a7a6 f1e1 b8d7 f4d6 d7b6 d6e5 f6d5 e5g7 g8g7 c3d5 d8d5 d1d5 b6d5 e2c4 c8e6 f3g5 d5f4 c4e6 f4e6 g5e6 e8e6 e1e6 f7e6 a1c1 g7g8 g2g3');
INSERT INTO evaluations VALUES(52423,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52423,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52423,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52423,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(52423,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52423,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52423,36,'cp 12','g8f6 b1c3 e7e6 g1f3 d7d5 d2d4 a7a6 c4d5 e6d5 f1e2 b8c6 e1g1 f8d6 d4c5 d6c5 a2a3 e8g8 b2b4 d5d4 e3d4 c5d4 f3d4 c6d4 c1e3 d4e2 d1e2 f8e8 f2f3 d8e7 f1e1 c8d7 e2d2 d7c6 e3f2 a8d8 d2g5 h7h6 e1e7');
INSERT INTO evaluations VALUES(44532,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(44532,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44532,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(44532,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(44532,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44532,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44532,36,'cp 5','b1c3 d7d5 c4d5 f6d5 g1f3 e7e6 d2d4 b8c6 a2a3 c5d4 e3d4 f8e7 f1d3 e8g8 e1g1 e7f6 d3e4 b7b6 c3d5 e6d5 e4b1 g7g6 h2h3 f8e8 c1e3 c8a6 f1e1 c6a5 b2b3 a8c8 b1d3 a6d3 d1d3 d8d7 a1c1 c8c1 e1c1 a5c6 h3h4 c6e7');
INSERT INTO evaluations VALUES(44527,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(44527,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(44527,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(44527,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44527,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(44527,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44527,36,'cp 4','d7d5 c4d5 f6d5 d2d4 b8c6 g1f3 c5d4 e3d4 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 c3d5 d7d5 b5e2 f7f6 e1g1 e7f7 c1e3 f8e7 a1c1 a8c8 g1g2 d5b7 e2d3 c8d8 d3e4 d8b8 f1d1 b7e4 f3e4 b8b4');
INSERT INTO evaluations VALUES(52758,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(52758,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(52758,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52758,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(52758,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(52758,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52758,36,'cp 36','e2e3 c8f5 c2c4 e7e6 d1b3 b7b6 c4d5 e6d5 g1f3 f8d6 f4d6 d8d6 f1d3 f5d3 b3d3 g8f6 e1g1 a7a5 b1c3 e8g8 b2b3 h7h6 a1c1 f8e8 h2h3 e8c8 f1d1 b8d7 d3f5 c8c7 c1c2 a8e8 c3e2 g7g6 f5f4 d6f4 e2f4 g8g7 d1c1 e8c8');
INSERT INTO evaluations VALUES(52756,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(52756,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(52756,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(52756,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(52756,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(52756,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(52756,36,'cp 31','c8f5 c2c4 e7e6 d1b3 b7b6 g1f3 f8d6 f4d6 d8d6 f1e2 g8e7 e1g1 b8d7 c4d5 e6d5 h2h3 a7a5 f1c1 e8g8 e2d3 a5a4 b3d1 f5d3 d1d3 a8a5 b1c3 h7h6 c1c2 c6c5 d4c5 b6c5 a1d1 c5c4 d3d4 a4a3 b2b3 c4b3');
INSERT INTO evaluations VALUES(31083,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(31083,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(31083,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(31083,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31083,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(31083,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31083,36,'cp 37','c2c4 e7e6 d1b3 b7b6 g1f3 f8d6 f4e5 d6e5 f3e5 g8e7 c4d5 e6d5 f1d3 e8g8 d3f5 e7f5 e1g1 d8d6 e5f3 b8d7 h2h3 h7h6 b3d3 d6e6 b1c3 f8c8 f1c1 d7f6 c1c2 f5e7 c3e2 e7g6 a1c1 c8c7 b2b3 a8c8 d3a6 f6e4 a2a4 e4g5 f3g5');
INSERT INTO evaluations VALUES(31060,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(31060,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(31060,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(31060,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(31060,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(31060,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(31060,36,'cp 43','e7e6 d1b3 b7b6 g1f3 f8d6 f4e5 d6e5 f3e5 g8e7 c4d5 e6d5 f1d3 d8d6 e1g1 b8d7 d3f5 e7f5 e5f3 e8g8 b1c3 h7h6 h2h3 f5e7 a1c1 f8c8 f1e1 c8e8 e1d1 a7a5 f3e1 d6b4 b3c2 b4d6 c3e2 e8c8 e1d3 d7f6 c2b3 c6c5 d4c5 b6c5 b3a3');
INSERT INTO evaluations VALUES(31438,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(31438,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(31438,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(31438,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(31438,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(31438,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(31438,36,'cp 38','d1b3 b7b6 g1f3 f8d6 f4e5 d6e5 f3e5 g8e7 c4d5 e6d5 f1d3 e8g8 e1g1 d8d6 d3f5 e7f5 e5f3 b8d7 b1c3 h7h6 a1c1 f8d8 h2h3 d7f6 b3a4 c6c5 d4c5 b6c5 f1d1 d6e6 c1c2 d8c8 c3e2 a7a5 e2f4 e6d6 b2b3 f5e7 a4a3 g8h8 d1c1 f6e4');
INSERT INTO evaluations VALUES(17818,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(17818,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17818,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(17818,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(17818,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17818,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(17818,36,'cp -5','d8f6 d2d4 e5d4 c1g5 f6d6 f3d4 f8e7 g5e7 g8e7 b1c3 c8d7 d4b3 d6d1 a1d1 b7b6 f2f3 c6c5 f1f2 e8c8 f2d2 e7c6 g1f2 f7f5 b3c1 h8e8 b2b3 f5e4 c3e4 d7f5 d2d8 e8d8 d1d8 c8d8 f2e3 c6b4 c2c3 b4d5 e3d2 f5e4 f3e4 d5f6 d2d3 f6d7 h2h3');
INSERT INTO evaluations VALUES(1186,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1186,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1186,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(1186,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1186,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1186,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(1186,36,'cp -2','d2d4 e5d4 c1g5 f6d6 f3d4 f8e7 g5e7 g8e7 d4b3 d6d1 f1d1 b7b6 b3d2 c8d7 a2a3 c6c5 b2b4 c5b4 a3b4 e7c6 b4b5 c6b4 c2c3 b4d3 b5a6 d3b2 d1c1 b2d3');
INSERT INTO evaluations VALUES(31913,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(31913,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(31913,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(31913,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(31913,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(31913,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(31913,36,'cp 96','f3e5 g4c8 g2g3 c7c6 c2c4 e7e6 f1g2 f7f6 e5f3 f6f5 d1c2 f8d6 e1g1 g8f6 f3e5 e8g8 b1d2 b8d7 d2f3 f6e4 e5d3 a7a5 c1f4 d6e7 f3e5 g7g5 f4c1 d7e5 d3e5 e7f6 b2b3 d8e7 f2f3 e4d6 c1a3 f6g7 e2e3 f8e8');
INSERT INTO evaluations VALUES(31942,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(31942,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(31942,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(31942,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(31942,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(31942,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(31942,36,'cp 78','g4c8 c2c4 e7e6 b1c3 g8f6 g2g3 c7c6 f1g2 f6d7 c1f4 d7e5 d4e5 f8c5 e1g1 e8g8 a1c1 b8d7 c4d5 e6d5 e2e4 d5d4 c3a4 b7b6 e5e6 f7e6 a4c5 b6c5 f4d6 c8a6 f1e1 f8f7 e4e5 d8a5 g2c6 a8d8 c6e4 a5b5 d1b3 c5c4 b3b5 a6b5 c1d1 c4c3 d1d4 c3b2');
INSERT INTO evaluations VALUES(57402,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(57402,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(57402,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(57402,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(57402,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(57402,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(57402,36,'cp 181','b8c6 d2d4 c6d4 f3d4 e5d4 d1d4 g8e7 b1c3 e7c6 d4e3 f8d6 c1d2 b7b6 e1c1 c8b7 a2a3 d8e7 f2f4 e8c8 e3d3 h8e8 h1e1 c8b8 c1b1 e7f8 h2h4 d6c5 g2g4 c6a5 c4d5 b7d5 c3d5 c7c6 d2a5 c6d5 a5c3 d5e4 e1e4 e8e4');
INSERT INTO evaluations VALUES(23219,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(23219,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(23219,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(23219,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(23219,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(23219,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(23219,36,'cp 202','d2d4 e5d4 e1g1 f8c5 c2c3 d4d3 b2b4 c5b6 b1a3 g8e7 c4b3 d7d6 d1d3 e7g6 a3c4 c6e5 f3e5 g6e5 c4e5 f6e5 a2a4 a7a6 d3c4 d8e7 c1e3 c8d7 e3b6 c7b6 f1d1 h8f8 c4d5 e8c8 d5d6 e7d6 d1d6 d7c6');
INSERT INTO evaluations VALUES(48022,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(48022,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(48022,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(48022,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(48022,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(48022,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(48022,36,'cp 7','d8e7 b1c3 e4c3 d2c3 e7e2 f1e2 b8c6 c1e3 c8d7 f3d4 c6d4 c3d4 h7h5 e2f3 e8c8 h2h4 d7g4 f3d5 g4e6 d5e6 f7e6 e3g5 d8e8 e1g1 c8d7 a1e1 f8e7 g5e7 e8e7 e1e2 h8f8 f2f4 b7b6 g2g3 c7c5 c2c3 f8a8 g1g2 a8b8 f1e1 c5d4');
INSERT INTO evaluations VALUES(33823,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33823,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(33823,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33823,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(33823,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(33823,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(33823,36,'cp 12','b1c3 e4c3 d2c3 e7e2 f1e2 c8d7 c1e3 b8c6 f3d4 a7a6 d4c6 d7c6 f2f3 b7b5 e1f2 f8e7 a2a3 c6d7 h1e1 e8g8 a1d1 f8e8 c3c4 a8b8 c4c5 d7f5 c5d6 e7d6 e2d3 f5d3 d1d3 d6e5 c2c3 e5d6 h2h3 a6a5 f3f4 a5a4 e1d1 h7h5');
INSERT INTO evaluations VALUES(33818,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33818,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(33818,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(33818,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(33818,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(33818,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(33818,36,'cp 8','e4c3 d2c3 e7e2 f1e2 b8c6 c1e3 c8d7 f3d4 c6d4 c3d4 h7h5 e2f3 e8c8 h2h4 d7g4 f3d5 g4e6 d5e6 f7e6 e3g5 d8e8 e1g1 c8d7 f2f4 f8e7 g5e7 e8e7 g2g3 b7b6 g1g2 h8c8 a1d1 c7c5 c2c3 a7a5 f1e1 b6b5 g2f3 c5d4 d1d4');
INSERT INTO evaluations VALUES(33825,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(33825,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(33825,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(33825,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(33825,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(33825,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33825,36,'cp 4','d2c3 e7e2 f1e2 f8e7 c1e3 c8d7 f3d4 b8c6 e1c1 e8c8 d4c6 d7c6 h1e1 h8e8 g2g3 f7f5 e2c4 e7f6 e3d4 f6e5 f2f4 e5d4 c3d4 c6e4 e1e3 d6d5 c4e2 c8d7 d1e1 c7c6 g3g4 e8e7 g4g5 d8e8 h2h4 b7b6');
INSERT INTO evaluations VALUES(33824,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(33824,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33824,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33824,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33824,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33824,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33824,36,'cp 0','e7e2 f1e2 f8e7 c1e3 b8c6 f3d4 c8d7 e1c1 e8c8 d4c6 d7c6 h1e1 h8e8 g2g3 f7f5 b2b4 a7a6 e3d4 g7g6 h2h4 e7f8 d4f6 f8e7 f6d4');
INSERT INTO evaluations VALUES(2042,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(2042,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(2042,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(2042,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(2042,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(2042,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(2042,36,'cp 181','c1g5 e7e6 b1c3 f8b4 d1d3 f7f6 e5f6 g8f6 g2g3 e8g8 f1g2 h7h6 g5f6 e6f6 e1g1 b4c3 d3c3 f6c3 b2c3 c6a5 f3d4 c7c6 e2e4 d7d6 f2f4 a5c4 f1f2 c4b6 a2a4 a7a5 f2f1 b6c4 f1e1 c8d7 a1b1 b7b5 a4b5 c6b5 d4b5 a8b8 b5d4');
INSERT INTO evaluations VALUES(2033,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(2033,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(2033,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(2033,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(2033,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(2033,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(2033,36,'cp 192','e7e6 g5f4 d7d6 e5d6 e6f6 d1c1 f8d6 f4d6 f6d6 c1d2 d6e7 b1c3 g8f6 e1c1 e8g8 d2f4 a7a6 h1g1 c8g4 e2e3 b7b5 h2h3 g4e6 f1d3 a8e8 f3g5 c6b4 a2a3 b4a2 c3a2 e6a2 d1d2 h7h6 g5f3 c7c5 g1d1 a2e6 d3e2 b5b4 a3b4 c5b4 f3d4 a6a5 d4e6');
INSERT INTO evaluations VALUES(56049,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(56049,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(56049,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(56049,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(56049,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(56049,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(56049,36,'cp 57','c3e4 b8d7 g1f3 g8f6 e4f6 d7f6 c2c3 a7a6 f3e5 f8d6 e5c4 d6e7 f1e2 f6d5 c4e5 e7d6 e1g1 e8g8 e2f3 c7c5 e5c4 b7b5 f3d5 e6d5 c4d6 d8d6 d4c5 d6c5 c1e3 c5c6 d1f3 f8e8 f1d1 c8b7 f3g3 e8e4 e3d4 f7f6 f2f3 e4e2 d1e1 a8e8 e1e2 e8e2');
INSERT INTO evaluations VALUES(11502,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(11502,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(11502,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11502,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(11502,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(11502,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(11502,36,'cp 43','b5c6 d7c6 f3e5 f6e4 c3e4 d8d5 e1g1 d5e5 d2d4 e5d5 f1e1 c8e6 e4g5 e8c8 g5e6 f7e6 d1g4 d5d4 g4e6 d8d7 c1g5 f8d6 e6e8 d7d8 e8e4 d4e4 e1e4 d8e8 a1e1 e8e4 e1e4 c8d7 g2g3 c6c5 g5d2 b7b5 g1g2 g7g6 f2f4 c5c4 g2f3 h7h5 d2c3');
INSERT INTO evaluations VALUES(11493,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(11493,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11493,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(11493,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(11493,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(11493,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(11493,36,'cp 43','d7c6 f3e5 f6e4 c3e4 d8d5 e1g1 d5e5 d2d4 e5d5 e4g5 c8e6 g5e6 f7e6 f1e1 e8c8 d1g4 d5d4 g4e6 d8d7 c1g5 f8d6 e6e8 d7d8 e8e4 d4e4 e1e4 d8f8 g2g3 f8f5 g5e3 f5e5 e4e5 d6e5 c2c3 h8e8 a1e1 e5f6 f2f4 a6a5 g3g4 g7g6 g1g2 a5a4 g4g5 f6g7');
INSERT INTO evaluations VALUES(11447,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(11447,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(11447,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(11447,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(11447,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(11447,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(11447,36,'cp 44','f3e5 f6e4 c3e4 d8d4 e1g1 d4e5 d2d4 e5d5 e4g5 c8e6 g5e6 f7e6 f1e1 e8c8 d1g4 d5d4 g4e6 d8d7 c1g5 f8d6 e6e8 d7d8 e8e4 d4e4 e1e4 d8e8 a1e1 c8d7 g1f1 e8f8 g5d2 f8f5 e1e2 g7g6 c2c4 h8f8 g2g3 f5f3 d2h6 f8f7 e2d2 b7b6 f1g2 d7c8 h2h3 a6a5 b2b3 c6c5 g3g4 f3c3 f2f4');
INSERT INTO evaluations VALUES(11438,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(11438,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(11438,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11438,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(11438,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11438,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11438,36,'cp 41','f6e4 c3e4 d8d5 e1g1 d5e5 d2d4 e5d5 e4g5 c8e6 f1e1 e8c8 g5e6 f7e6 d1g4 d5d4 g4e6 d8d7 c1g5 f8d6 e6e8 d7d8 e8e4 d4e4 e1e4 d8f8 h2h4 f8f5 g2g4 f5c5 c2c4 b7b5 b2b3 b5c4 b3c4 c5a5 g5f4 d6f4 e4f4 g7g6 g1g2 h8e8 g2g3 h7h5');
INSERT INTO evaluations VALUES(11464,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(11464,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(11464,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(11464,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(11464,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(11464,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11464,36,'cp 45','c3e4 d8d5 e1g1 d5e5 d2d4 e5d5 e4g5 c8e6 g5e6 f7e6 f1e1 e8c8 d1g4 d5d4 g4e6 d4d7 e6d7 d8d7 e1e8 d7d8 e8d8 c8d8 c1e3 f8d6 f2f4 h8f8 g2g3 b7b5 g1g2 d8d7 b2b3 c6c5 e3d2 g7g6 c2c4 b5c4 b3c4 a6a5 d2c3 a5a4 a1d1 a4a3 g2f3');
INSERT INTO evaluations VALUES(11463,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(11463,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(11463,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(11463,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(11463,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(11463,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11463,36,'cp 38','d8d5 e1g1 d5e5 d2d4 e5d5 e4g5 c8e6 f1e1 e8c8 g5e6 f7e6 d1g4 d5d4 g4e6 d4d7 e6d7 d8d7 e1e8 d7d8 e8d8 c8d8 c1e3 f8d6 f2f4 d8d7 g2g3 h8f8 b2b3 b7b5 g1g2 b5b4 e3d4 g7g6 c2c4 b4c3 d4c3 f8f5 g2f3 f5c5 a1c1 g6g5 h2h3 g5f4 g3f4 c5h5 c1h1 c6c5');
INSERT INTO evaluations VALUES(42564,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(42564,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(42564,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(42564,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(42564,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(42564,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(42564,36,'cp 208','f3e5 e8g8 d2d4 c5b6 e5f7 d8e7 f7e5 g8h8 e1g1 d7d6 e5f3 c8g4 c2c3 e7e4 c4d3 e4d5 b1d2 b8d7 h2h3 g4h5 a2a4 c7c5 a4a5 b6c7 c3c4 d5g8 d4d5 d7e5 d3e2 e5f3 e2f3 h5g6 f1e1 g8f7 a5a6 b7b6 d2f1 f6e4 f1g3 e4g3 f2g3 h8g8 c1d2');
INSERT INTO evaluations VALUES(42527,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(42527,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(42527,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(42527,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(42527,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(42527,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(42527,36,'cp 195','e8g8 d2d4 c5b6 e5f7 d8e7 f7e5 g8h8 e1g1 d7d6 e5f3 c8g4 c2c3 e7e4 c4d3 e4d5 b1d2 g4h5 h2h3 a7a5 a2a4 b8d7 f1e1 a8d8 d3c4 d5c6 c4e2 c6d5 d2f1 f6e4 c1e3 d7f6 f3d2 e4d2 f1d2 h7h6 e2h5 f6h5 d2f3 d8e8 d1d3');
INSERT INTO evaluations VALUES(37208,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(37208,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(37208,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(37208,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(37208,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(37208,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(37208,36,'cp 195','d2d4 c5e7 d1e2 c7c6 c4b3 d7d5 b1c3 f6e4 c3e4 d5e4 e2e4 b8d7 c1f4 d7f6 e4d3 e7d6 e1g1 f6h5 f4e3 a7a5 a2a4 d6e5 d4e5 d8d3 c2d3 g7g6 f2f4 c8e6 b3e6 f7e6 g2g4 h5g7 e3c5 f8f7 a1a3 b7b5 g1g2 f7d7 f1d1 d7f7 g2g3 h7h5 c5d6 h5g4');
INSERT INTO evaluations VALUES(37209,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(37209,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(37209,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(37209,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(37209,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(37209,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(37209,36,'cp 182','c5e7 d1e2 d7d5 e4d5 f6d5 e1g1 d5b6 c4b3 b8c6 e5c6 b7c6 f1e1 e7d6 b1d2 a7a5 a2a4 d8f6 c2c3 f6g6 d2f3 c8g4 f3e5 g4e2 e5g6 f8e8 g6e5 d6e5 d4e5 e8e5 c1f4 e5e7 f4c7 e7c7 e1e2 g8f8 g1f1 a8d8 a1e1 c7d7 g2g3 d7d2 e2d2');
INSERT INTO evaluations VALUES(31855,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31855,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31855,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31855,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31855,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(31855,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31855,36,'cp 19','g1f3 e7e6 c2c4 c7c6 f1d3 f5g6 d1c2 g8f6 e1g1 b8d7 b1d2 g6d3 c2d3 f8e7 b2b3 e8g8 e3e4 d5e4 d2e4 d8c7 e4f6 d7f6 g2g3 h7h6 d3e2 f8e8 c1b2 a8d8 f1e1 e7b4 e1d1 c7a5 f3e5 c6c5 d4c5 b4c5 d1d8 e8d8');
INSERT INTO evaluations VALUES(31857,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(31857,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31857,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31857,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(31857,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(31857,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31857,36,'cp 19','c7c6 c2c4 e7e6 f1d3 f5g6 e1g1 b8d7 d1c2 g8f6 d3g6 h7g6 b1d2 c6c5 c4d5 f6d5 d4c5 f8c5 d2b3 c5b6 c2e2 d8c7 a2a4 a7a5 g2g3 d5b4 c1d2 c7c2 b3d4 b6d4 f3d4 c2d3 d2b4 d3e2 d4e2 a5b4');
INSERT INTO evaluations VALUES(31092,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(31092,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31092,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31092,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31092,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31092,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31092,36,'cp 22','f1d3 e7e6 c2c4 f5g6 e1g1 g8f6 d1c2 b8d7 b1d2 c6c5 d3g6 h7g6 c4d5 f6d5 d4c5 a8c8 d2e4 d7c5 c1d2 c5e4 c2e4 f8c5 a1c1 e8g8 e4a4 a7a6 c1c2 c5e7 f1c1 c8c2 a4c2 d8d7 g2g3 f8d8 h2h4 b7b6 e3e4 d5b4 d2b4');
INSERT INTO evaluations VALUES(31091,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31091,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(31091,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31091,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31091,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31091,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31091,36,'cp 13','e7e6 c2c4 f5g6 e1g1 b8d7 d1c2 g8f6 b1d2 c6c5 d3g6 h7g6 c4d5 f6d5 d4c5 a8c8 d2e4 d7c5 c1d2 c5e4 c2e4 f8c5 e4a4 d8d7 a4b3 e8g8 f1d1 d7c6 a1c1 c6a6 h2h3 f8d8 d2e1 c5e7 a2a3 c8c1 d1c1 d5b6 c1c2 e7f6 c2c7 d8d7 c7d7');
INSERT INTO evaluations VALUES(46993,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(46993,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(46993,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(46993,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(46993,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(46993,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(46993,36,'cp 60','g7g6 h2h3 f8g7 e2e4 e7e5 d4d5 b8a6 c1e3 e8g8 f1e2 f6e8 g2g4 f7f5 g4f5 g6f5 e4f5 c8f5 g1f3 g8h8 f3g5 a6c5 d1d2 h7h6 b2b4 h6g5 b4c5 d6c5 e3g5 d8d7 e1c1 e8d6 d1g1 f5h7 g5e3 d7e7 h3h4 f8f4 h4h5');
INSERT INTO evaluations VALUES(46666,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46666,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(46666,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(46666,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(46666,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46666,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(46666,36,'cp 62','e2e4 f8g7 f1e2 e8g8 g1f3 e7e5 e1g1 c8g4 d4d5 a7a5 a2a3 b8a6 c1e3 b7b6 g1h1 g4d7 f3d2 f6e8 f2f3 f7f5 e4f5 g6f5 f1g1 f5f4 e3f2 e8f6 d1c2 d8e7 a1b1 a6c5 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(46673,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(46673,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(46673,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(46673,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(46673,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(46673,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(46673,36,'cp 59','f8g7 f1e2 e8g8 h2h3 e7e5 d4d5 b8a6 c1g5 h7h6 g5e3 f6d7 a2a3 f7f5 b2b4 f5f4 e3c1 b7b6 g1f3 a6b8 e1g1 d7f6 c4c5 b6c5 b4c5 b8a6 c5d6 c7d6 a3a4 a6c5 f3d2 a8b8 c1a3 g8h7 a1b1 c8d7 b1b8 d8b8');
INSERT INTO evaluations VALUES(45634,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(45634,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(45634,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(45634,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(45634,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(45634,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(45634,36,'cp 18','b8d7 d1c2 f8d6 f1d3 d5c4 d3c4 e6e5 e1g1 e8g8 b2b3 d8e7 a2a4 e5e4 f3g5 d6h2 g1h2 f6g4 h2g1 e7g5 c2e4 d7b6 c4d3 g5h6 e4f4 h6f4 e3f4 f8d8 a4a5 b6d5 c3d5 d8d5 a5a6 b7a6 f1d1 g4f6 c1a3 d5d8 a3e7 d8e8 d1e1 c8e6 e7f6 g7f6 a1a6 e6b3 e1e8 a8e8');
INSERT INTO evaluations VALUES(49552,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(49552,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(49552,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(49552,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(49552,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(49552,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(49552,36,'cp -13','b1c3 f8c5 h2h3 e8g8 g1f3 b8c6 d2d3 d7d6 f1e2 a7a6 e1g1 c8d7 a2a3 c6d4 b2b4 c5b6 c1e3 a6a5 e3d4 e5d4 c3d5 f6d5 e4d5 f8e8 d1c2 h7h6 c4c5 d6c5 b4c5 b6c5 c2c5 e8e2 a1c1 e2a2 f3d4 a2d2 f1e1 d2d3 c5c7 d8c7');
INSERT INTO evaluations VALUES(49547,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49547,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(49547,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(49547,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(49547,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(49547,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(49547,36,'cp -26','f8c5 g2g3 a7a6 d2d3 d7d6 h2h3 b8c6 f1g2 b7b5 g1e2 a8b8 e1g1 e8g8 a1b1 c6d4 c1e3 c7c6 b2b4 d4e2 d1e2 c5e3 e2e3 c8e6 c4b5 a6b5 a2a4 b5a4 c3a4 d8d7 a4b6 d7c7 b6c4 b8b5 g1h2 h7h6 f1c1');
INSERT INTO evaluations VALUES(42580,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(42580,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(42580,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(42580,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(42580,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(42580,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(42580,36,'cp -10','h2h3 d7d6 d2d3 b8c6 g1f3 a7a6 f1e2 c8d7 e1g1 e8g8 a2a3 c6d4 f3d4 c5d4 a3a4 c7c6 c1e3 d4e3 f2e3 b7b5 a4b5 a6b5 a1a8 d8a8 c4b5 c6b5 d3d4 a8e8 e2f3 b5b4 c3a2 d6d5 a2b4 d5e4 f3e2 d7b5 d4e5 b5e2 d1e2');
INSERT INTO evaluations VALUES(42579,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(42579,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(42579,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(42579,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(42579,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(42579,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(42579,36,'cp -20','e8g8 g1f3 b8c6 f1e2 c6d4 d2d3 d4f3 e2f3 c7c6 e1g1 d7d6 c1e3 c5e3 f2e3 b7b5 c4b5 c6b5 d3d4 b5b4 d4e5 d6e5 d1d8 f8d8 c3d5 c8b7 d5f6 g7f6 f1d1 g8g7 d1d8 a8d8 a1c1 d8c8 c1c8 b7c8 g1f1 g7f8 a2a3 b4a3 b2a3 f8e7 f1e2 c8d7 e2e1');
INSERT INTO evaluations VALUES(37242,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(37242,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(37242,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(37242,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(37242,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(37242,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(37242,36,'cp -9','g1f3 f8e8 d2d3 d7d6 f1e2 b8c6 e1g1 a7a5 c1e3 c6d4 d1d2 c7c6 e2d1 h7h6 a2a3 d4f3 d1f3 c8e6 a1e1 a8b8 e1b1 f6d7 g1h2 d8c7 f3e2 c5d4 f2f4 d4e3 d2e3 e5f4 e3f4 d7e5 f4g3 c7e7 b1d1 b8d8 d3d4 e5c4');
INSERT INTO evaluations VALUES(53247,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(53247,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(53247,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(53247,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(53247,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(53247,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(53247,36,'cp 70','c4d5 d8d5 g1f3 c5d4 b1c3 d5a5 f3d4 g8f6 d4b3 a5h5 e2e4 h5d1 e1d1 e7e5 f1b5 c8d7 b5d7 b8d7 f2f3 a8c8 d1e2 f8b4 c1d2 d7c5 b3c5 b4c5 g2g4 h7h6 a1c1 e8e7 c1c2 c5b6 h1c1 e7e6 c3a4 c8c2 c1c2 h8a8 d2b4 b6d4 a4c5 d4c5 c2c5');
INSERT INTO evaluations VALUES(53228,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(53228,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(53228,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(53228,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(53228,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(53228,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(53228,36,'cp 76','d8d5 g1f3 c5d4 b1c3 d5a5 f3d4 g8f6 d4b3 a5c7 c3b5 c7e5 e2e4 e7e6 d1d4 e5d4 b3d4 e8d8 b5c3 c8d7 f2f3 f8d6 c1e3 a7a6 f1e2 d8e7 e1g1 e6e5 d4b3 h8c8 a1c1 d7e6 b3a5 b7b5 a2a3 b8c6 a5b7 c6d4 b7d6 d4e2 c3e2 c8c1 f1c1 e7d6');
INSERT INTO evaluations VALUES(34359,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(34359,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(34359,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34359,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(34359,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(34359,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(34359,36,'cp 76','g1f3 c5d4 b1c3 d5a5 f3d4 g8f6 d4b3 a5c7 c3b5 c7c6 e2e4 e7e5 c1e3 f8b4 b3d2 e8g8 f2f3 b4c5 a1c1 c5e3 c1c6 b8c6 d1b3 e3c5 b3c2 c5e7 b5c7 c6d4 c2b1 a8b8 f1d3 f8d8 e1g1 d4e6 c7d5 f6d5 e4d5 d8d5 d3h7 g8h8 d2c4 d5c5 b1e4 c5c4 e4c4 h8h7 c4e4 h7g8 e4e5');
INSERT INTO evaluations VALUES(34358,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(34358,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(34358,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(34358,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(34358,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(34358,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(34358,36,'cp 56','c5d4 b1c3 d5a5 f3d4 g8f6 d4b3 a5h5 e2e4 h5d1 e1d1 e7e5 f1c4 b8d7 c3b5 d7b6 c4d3 e8d8 d1e2 b6a4 b3a5 a7a6 b5c3 a4c3 b2c3 d8c7 a5c4 f8c5 f2f3 h8e8 a2a4 c8d7 c1e3 c5e3 e2e3 d7c6 a4a5 f6d7 h1d1 e8e6 e3e2 d7c5 c4e3 c5d3 d1d3');
INSERT INTO evaluations VALUES(34379,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(34379,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(34379,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(34379,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34379,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(34379,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(34379,36,'cp 64','b1c3 d5a5 f3d4 g8f6 d4b3 a5h5 c1d2 e7e5 e2e3 c8g4 f1e2 g4e2 d1e2 b8c6 a1c1 a8d8 e2h5 f6h5 g2g4 h5f6 f2f3 f6d5 c3e4 h7h5 g4g5 f7f5 g5f6 g7f6 b3c5 d8b8 e1e2 h5h4 a2a3 e8f7 b2b4 f8e7 c5b7 c6b4 b7d6 e7d6 e4d6 f7e6 a3b4 e6d6');
INSERT INTO evaluations VALUES(34375,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(34375,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(34375,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(34375,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(34375,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(34375,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(34375,36,'cp 46','d5a5 f3d4 g8f6 d4b3 a5h5 e2e4 h5d1 e1d1 e7e5 f1b5 c8d7 b5d7 b8d7 f2f3 f8b4 c1d2 e8e7 d1e2 h8c8 a1c1 e7e6 h1d1 b4c3 d2c3 d7c5 b3c5 c8c5 d1d2 a8c8 c1d1 f6e8 g2g3 g7g5 a2a4 f7f6 h2h4 g5h4 g3h4 b7b6');
INSERT INTO evaluations VALUES(19927,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(19927,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(19927,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(19927,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(19927,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(19927,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(19927,36,'cp 58','e1g1 c8d7 f1e1 a7a6 b5c6 d7c6 d2d4 c5d4 f3d4 a8c8 c2c4 e7e6 b1c3 f8e7 c1f4 g8f6 c3d5 e8g8 d4c6 c8c6 d5e7 d8e7 b2b3 b7b5 c4b5 a6b5 d1d3 e7b7 f4d6 f8d8 e4e5 f6e8 e1d1 e8d6 e5d6 c6d6 d3d6 d8d6 d1d6 h7h5 d6d1 h5h4 h2h3 g7g5 a2a4 b5a4 b3a4 b7b4');
INSERT INTO evaluations VALUES(19925,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(19925,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(19925,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(19925,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(19925,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(19925,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19925,36,'cp 28','c8d7 f1e1 a7a6 b5c6 d7c6 d2d4 c5d4 f3d4 a8c8 c2c4 e7e6 b1c3 g8f6 d1d3 f8e7 c1f4 c6d7 d4e2 d8c7 b2b3 e8g8 a1c1 f8d8 e1d1 d7e8 h2h3 c7a5 f4d2 a5h5 d2e3 h5a5 e2d4 h7h6 g1h1 e8d7 f2f3 d6d5 c4d5');
INSERT INTO evaluations VALUES(26055,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(26055,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(26055,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(26055,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26055,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(26055,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(26055,36,'cp 38','f1e1 a7a6 b5f1 d7g4 c2c3 d6d5 e4d5 d8d5 f1e2 g8f6 d2d3 e7e6 b1a3 f8e7 a3c4 d5d8 h2h3 g4h5 g2g4 h5g6 c1f4 f6d5 f4g3 f7f6 f3h4 g6f7 e2f3 e8g8 a2a4 g7g6 a4a5 e6e5 h4g2 g8h8 g4g5 d8c7 f3d5 f7d5 g5f6 e7f6 c4b6 a8d8 b6d5');
INSERT INTO evaluations VALUES(26057,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(26057,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26057,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(26057,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(26057,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(26057,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(26057,36,'cp 51','a7a6 b5c6 d7c6 d2d4 c5d4 f3d4 a8c8 c2c4 e7e6 b1c3 g8f6 c3d5 c6d7 d5f6 d8f6 b2b3 f8e7 c1b2 f6g6 b2a3 c8d8 d1f3 g6f6 f3d3 e8g8 a1d1 h7h5 h2h3 f8e8 d3g3 e6e5 d4e2 b7b5 c4b5 d7b5 e2c3 h5h4 g3g4 b5d7 c3d5 d7g4 d5f6 e7f6 h3g4');
INSERT INTO evaluations VALUES(37400,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(37400,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(37400,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(37400,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(37400,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37400,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(37400,36,'cp 63','e2e4 d7d6 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 c1e3 b8a6 a2a3 d8e8 h2h3 g4d7 b2b3 f6h5 b3b4 h5f4 c4c5 a5b4 a3b4 a6b4 d1d2 f4e2 d2e2 e8e7 c5d6 e7d6 c3b5 d7b5 e2b5 b7b6 a1c1 f8d8 b5c4 b4a6 h3h4');
INSERT INTO evaluations VALUES(37411,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(37411,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(37411,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(37411,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(37411,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37411,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(37411,36,'cp 51','d7d6 f1e2 e7e5 e1g1 e5d4 f3d4 f8e8 f2f3 c7c6 g1h1 b8d7 c1e3 a7a6 e3f2 d6d5 c4d5 c6d5 e4d5 d7b6 d1b3 f6d5 c3d5 g7d4 f2d4 b6d5 e2c4 c8e6 f1e1 d5f4 c4e6 e8e6 e1e6 f4e6 d4c3 d8c7 a1e1 b7b5 a2a4 b5a4 b3a4 c7f4 a4c2');
INSERT INTO evaluations VALUES(49546,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(49546,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(49546,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(49546,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(49546,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(49546,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(49546,36,'cp -30','e5d4 g1f3 f8b4 c2c3 d4c3 b2c3 d7d5 e4d5 b4e7 e1g1 e8g8 c4b3 c8g4 h2h3 g4h5 f1e1 f8e8 c1b2 e7c5 e1e8 d8e8 b1d2 b8d7 d1f1 h5g6 c3c4 c5d6 a1e1 e8c8 b3d1 h7h6 d2b3 c8f8 f3d4 a7a6 d1c2 g6c2 d4c2 a8e8 f1d3 e8e1 c2e1');
INSERT INTO evaluations VALUES(49738,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(49738,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(49738,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(49738,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(49738,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(49738,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(49738,36,'cp -28','g1f3 f8b4 c2c3 d4c3 b2c3 d7d5 e4d5 b4e7 e1g1 e8g8 c4b3 c8g4 h2h3 g4h5 f1e1 f8e8 c1b2 e7c5 e1e8 d8e8 b1d2 b8d7 d1f1 e8f8 c3c4 f8d6 b3c2 c7c6 d5c6 d6c6 f3d4 c5d4 b2d4 b7b6 a1c1 a8e8 c1e1 e8d8 e1e7 d7c5 d2b3 h5g6 b3c5 b6c5');
INSERT INTO evaluations VALUES(49737,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(49737,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(49737,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49737,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(49737,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(49737,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(49737,36,'cp -22','f8b4 c2c3 d4c3 b2c3 d7d5 e4d5 b4e7 e1g1 e8g8 c4b3 c8g4 h2h3 g4h5 f1e1 f8e8 c1b2 e7c5 e1e8 d8e8 b1d2 b8d7 d1f1 h5g6 c3c4 e8f8 b3a4 c5d6 f1d1 d7c5 a4c2 f8e7 d2b3 a8e8 b3c5 d6c5 c2g6 h7g6 b2d4 b7b6 d4c5 e7c5 d1d4 c5d6 a1c1 f6d7 c1c2 d6f6');
INSERT INTO evaluations VALUES(42833,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(42833,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(42833,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(42833,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(42833,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(42833,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(42833,36,'cp -25','c2c3 d4c3 b2c3 d7d5 e4d5 b4e7 e1g1 e8g8 c4b3 c8g4 c1b2 b8d7 b1d2 g4h5 c3c4 a8c8 f1e1 c7c6 d2f1 e7b4 e1e3 b4c5 b2d4 c6d5 c4d5 d8b6 d4c5 d7c5 h2h3 b6d6 f1g3 h5g6 f3e5 c5b3 a2b3 d6d5 d1d5 f6d5');
INSERT INTO evaluations VALUES(42832,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(42832,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(42832,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(42832,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(42832,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(42832,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(42832,36,'cp -35','d4c3 b2c3 b4c5 e1g1 d7d5 e4d5 e8g8 c1a3 d8d6 a3c5 d6c5 d1d4 c5d4 f3d4 f8d8 f1d1 c7c6 d5c6 b8c6 c4b3 c6d4 c3d4 c8e6 b1c3 e6b3 a2b3 a8c8 d1d3 a7a6 a1d1 f6d5 c3d5 d8d5 g2g3 g8f8 g1g2 c8c2 d1e1 d5b5 e1e5');
INSERT INTO evaluations VALUES(42827,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(42827,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(42827,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(42827,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(42827,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(42827,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(42827,36,'cp -25','b2c3 d7d5 e4d5 b4e7 c4b3 e8g8 e1g1 c8g4 c1b2 b8d7 b1d2 f8e8 c3c4 c7c6 d5c6 b7c6 b3c2 e7b4 a1b1 d8a5 h2h3 g4h5 d1c1 a8d8 a2a3 b4f8 b2c3 a5a6 c3b4 d7c5 f1e1 e8e1 c1e1 h5f3 d2f3 a6c4 b1c1 c4d5 e1e3 d8e8 e3c3 c5e6 b4f8 e6f4 g1h2 e8f8');
INSERT INTO evaluations VALUES(42826,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(42826,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(42826,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(42826,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(42826,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(42826,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(42826,36,'cp -21','d7d5 e4d5 b4e7 c4b3 c8g4 h2h3 g4h5 e1g1 e8g8 f1e1 e7c5 c1f4 c5d6 f4e3 b8d7 b1d2 f8e8 e3d4 e8e1 d1e1 d8f8 e1f1 a8e8 a1e1 a7a6 c3c4 e8e1 f1e1 h5g6 e1c1 b7b6 b3c2 f8c8 d4b2 c7c6 c2g6');
INSERT INTO evaluations VALUES(55493,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55493,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55493,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55493,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55493,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55493,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55493,36,'cp 0','c7c6 b5e2 g8f6 d2d3 g7g6 c2c4 f8g7 b1c3 e8g8 e1g1 f8e8 c1e3 b8a6 d1c2 f6g4 e3g5 g4f6');
INSERT INTO evaluations VALUES(53798,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53798,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53798,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(53798,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53798,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(53798,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53798,36,'cp 0','b5e2 g8f6 d2d3 f8e7 e1g1 e8g8 c2c4 c6c5 b1c3 b8c6 f3e1 c8e6 e1c2 f6d7 e2g4 d7f6 g4f3 c6d4 g2g3 a7a6 a2a4 a8b8 a1b1 d4f3 d1f3 e6g4 f3g2 d8d7 f1e1 g4h3 g2f3 h3g4 f3g2');
INSERT INTO evaluations VALUES(53836,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53836,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53836,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53836,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53836,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53836,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53836,36,'cp 6','g8f6 d2d3 f8e7 c2c4 e8g8 e1g1 c6c5 b1c3 b8c6 a1b1 c8d7 a2a3 a7a5 c1d2 h7h6 h2h3 f6h7 d2e3 a8b8 f3d2 e7g5 e2g4 d7e6 d2f3 e6g4 h3g4 d8d7 f3g5 h7g5 c3d5 g5e6 b2b4 c5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(57653,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(57653,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(57653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57653,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57653,36,'cp 0','f8g7 g1f3 g8f6 b1c3 d7d5 e2e3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(39582,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39582,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39582,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39582,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(39582,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39582,36,'cp 0','b1c3 d7d5 e2e3 g8f6 g1f3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(39578,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39578,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39578,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39578,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39578,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39578,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39578,36,'cp 0','d7d5 c3b5 b8a6 e2e3 g8f6 b5c3 a6b8 h2h4 h7h5 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(38454,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38454,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38454,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38454,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38454,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38454,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38454,36,'cp 0','e2e3 g8f6 h2h4 h7h5 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(38452,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38452,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38452,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38452,36,'cp 0','g8f6 h2h4 h7h5 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(41457,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41457,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41457,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41457,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41457,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41457,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41457,36,'cp 0','h2h4 h7h5 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(41456,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41456,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41456,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41456,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41456,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41456,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41456,36,'cp 0','h7h5 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(55048,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(55048,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(55048,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(55048,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(55048,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(55048,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(55048,36,'cp -70','g8f6 g1f3 c8f5 c2c4 e7e6 b1c3 f8e7 c4d5 e6d5 e2e3 e8g8 a2a3 f6e4 f1d3 e4c3 b2c3 f5d3 d1d3 b8d7 c3c4 c7c5 e1g1 d7f6 c1b2 a8c8 d4c5 e7c5 c4d5 d8d5 f1d1 f8d8 b2d4 f6e4 d3b5 b7b6 h2h3 e4g3 d4c5 d5d1 a1d1 d8d1 g1h2 g3f1 h2g1');
INSERT INTO evaluations VALUES(47163,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(47163,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(47163,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(47163,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(47163,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(47163,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(47163,36,'cp -65','e2e3 c8f5 g1f3 e7e6 f1d3 f8e7 b1c3 e8g8 e1g1 c7c5 d4c5 e7c5 d3f5 e6f5 d1d3 d8d7 a2a3 a7a6 b2b4 c5a7 b4b5 f8d8 a3a4 a7c5 c1a3 c5a3 a1a3 a6b5 c3b5 b8c6 f1b1 g7g6 b5d4 c6d4 d3d4 f6e4 a3b3 a8a4 b3b7 d7c6 b7b8 a4d4 b8d8 g8g7');
INSERT INTO evaluations VALUES(47161,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(47161,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(47161,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(47161,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(47161,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(47161,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(47161,36,'cp -69','c8f5 g1f3 e7e6 f1d3 f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 f6e4 e1g1 b8d7 c1a3 e7a3 b1a3 e8g8 c2c4 d7f6 a1c1 f8e8 f3e5 c7c6 a3b1 a7a5 a2a4 c8d8 h2h3 g7g6 d3c2 h7h5 b1c3 e4c3 c2c3 f6e4 c3b2 h5h4 g1h2 g8g7 e5d3');
INSERT INTO evaluations VALUES(29551,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(29551,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(29551,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(29551,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(29551,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(29551,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(29551,36,'cp -64','g1f3 e7e6 f1d3 f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 f6e4 c1a3 e7a3 b1a3 b8d7 c2c4 d7f6 e1g1 e8g8 f3e5 g7g6 a1c1 c7c6 a3b1 a7a5 a2a4 f8e8 b1c3 e4c3 d3c3 e8e7 c3b2 c8c7 h2h3 a8e8 f1e1 c7b6 c1c2 f6e4 g1h2');
INSERT INTO evaluations VALUES(29552,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(29552,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(29552,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(29552,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(29552,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(29552,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(29552,36,'cp -69','e7e6 f1d3 f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 e8g8 c2c4 c7c6 e1g1 f8e8 c1a3 e7a3 b1a3 c8e6 a1e1 f6e4 a3b1 b8a6 d3e2 f7f6 a2a3 a6c7 f3d2 e4d6 c4c5 d6e4 d2e4 f5e4 b1c3 g8h8 g1h1 f6f5 e2c2 e8e7');
INSERT INTO evaluations VALUES(30109,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(30109,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(30109,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(30109,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(30109,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(30109,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(30109,36,'cp -58','f1d3 f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 e8g8 c2c4 c7c6 e1g1 f8e8 c1a3 e7a3 b1a3 c8e6 f1e1 f6e4 a1c1 g7g6 a3b1 a7a5 f3e5 b8d7 h2h3 d7e5 f4e5 f7f6 e5f6 e4f6 b1d2 e8e7 c4d5 e6d5 c1c5 d5d8 a2a4 f6d5 d2f3');
INSERT INTO evaluations VALUES(30107,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(30107,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(30107,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(30107,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(30107,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(30107,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(30107,36,'cp -64','f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 e8g8 c2c4 f8e8 e1g1 c7c6 c1a3 e7a3 b1a3 c8e6 f1e1 f6e4 a1c1 b8d7 a3b1 e8e7 a2a4 a7a6 b1d2 a8e8 h2h3 d7f6 f3e5 e4d2 d3d2 f6e4 d2b2 e4d6 c4c5 d6e4 b3b4 f7f6 e5d3 g7g5');
INSERT INTO evaluations VALUES(57421,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(57421,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(57421,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(57421,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(57421,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(57421,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(57421,36,'cp 185','d7d5 e4d5 e5e4 f3d2 f6f5 c2c4 g8f6 b1c3 f8e7 f1e2 e8g8 d2b3 b7b6 e1g1 a7a5 c1f4 e7d6 f4d6 c7d6 d1d2 b8a6 a1e1 d8e7 b3a1 a6c7 a1c2 b6b5 c4b5 c7d5 f2f3 c8e6 c3d5 f6d5 a2a4 d5b6 d4d5');
INSERT INTO evaluations VALUES(56825,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(56825,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(56825,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(56825,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(56825,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(56825,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(56825,36,'cp 188','e4d5 e5e4 f3d2 f6f5 c2c4 g8f6 f1e2 f8e7 e1g1 e8g8 b1c3 b7b6 d2b3 a7a5 c1f4 b8d7 a1c1 c8a6 f1e1 f6e8 f2f3 a5a4 c3a4 e7d6 d1d2 d6f4 d2f4 e8d6 c4c5 a6e2 c5d6 c7d6 e1e2 a8a4 f4d6 a4a8 g1h1 d7f6 d6d8 a8d8');
INSERT INTO evaluations VALUES(56809,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(56809,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(56809,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(56809,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(56809,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(56809,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(56809,36,'cp 180','e5e4 f3d2 f6f5 c2c4 g8f6 b1c3 f8e7 d2b3 e8g8 f1e2 a7a5 e1g1 b7b6 c1f4 e7d6 f4d6 c7d6 d1d2 b8a6 a1e1 d8e7 b3a1 c8d7 a2a3 a6c7 a1c2 b6b5 c4b5 f6d5 c2e3 d7e6 c3d5 c7d5 f2f3 d5f6 d4d5 e6d7 f3f4 g8h8');
INSERT INTO evaluations VALUES(55530,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55530,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(55530,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(55530,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(55530,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(55530,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(55530,36,'cp -25','f1c4 g7g6 h5d1 f8g7 d2d3 g8e7 g1f3 b8c6 c2c3 e8g8 a2a4 d8e8 b2b4 g8h8 e1g1 a7a6 b1d2 f7f5 h2h3 h7h6 c4a2 c8d7 f1e1 g6g5 e4f5 d7f5 d2e4 e8g6 e1e3 d6d5 e4g3 e5e4 b4b5 a6b5 d3e4 d5e4 a4b5');
INSERT INTO evaluations VALUES(55524,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(55524,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(55524,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(55524,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55524,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(55524,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(55524,36,'cp -25','g7g6 h5d1 g8f6 d2d3 f8g7 g1f3 c7c6 e1g1 e8g8 f1e1 d6d5 c4b3 d5e4 f3e5 d8c7 d3d4 c6c5 c2c3 b8c6 e5c6 b7c6 h2h3 c8a6 c1e3 a8d8 b1d2 c5d4 c3d4 f8e8 d2c4 f6d5 e3g5 a6c4 b3c4 d8d7 d1d2 f7f5 g5h6 g7h6');
INSERT INTO evaluations VALUES(55320,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(55320,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(55320,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(55320,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(55320,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(55320,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(55320,36,'cp -27','h5d1 g8f6 d2d3 f8g7 g1f3 c7c6 e1g1 e8g8 f1e1 b7b5 c4b3 a7a5 a2a4 b8a6 c2c3 a6c5 b3c2 b5b4 d3d4 e5d4 c3d4 c5e6 b1d2 c8a6 c2b3 d6d5 e4e5 f6d7 h2h4 a6b7 h4h5 c6c5 h5g6 h7g6 d2f1 e6d4 f3d4 c5d4 d1d4');
INSERT INTO evaluations VALUES(43209,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(43209,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43209,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43209,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(43209,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(43209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43209,36,'cp 0','f1e2 d7d6 h2h3 c7c5 g1f3 b8c6 e1g1 d8b6 b1a3 b6b2 a3b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(43208,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43208,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43208,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43208,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43208,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43208,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43208,36,'cp 0','d7d6 g1f3 c7c5 c2c3 e8g8 d4c5 d6c5 d1d8 f8d8 b1d2 c8e6 h2h4 h7h6 a2a4 b8c6 e3e4 f6h5 f4h2 h5f6');
INSERT INTO evaluations VALUES(41321,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41321,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(41321,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41321,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41321,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(41321,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41321,36,'cp 0','h2h3 c7c5 g1f3 b8c6 c2c3 d8b6 d1c1 c5d4 e3d4 e7e5 d4e5 f6d5 e5e6 c8e6 f4g3 e8g8 e1g1 d5f6 b1a3 f8e8 g3f4 f6d5 f4g3');
INSERT INTO evaluations VALUES(47173,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(47173,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(47173,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(47173,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(47173,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(47173,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(47173,36,'cp -21','e7e6 f1d3 c7c5 d4c5 b8c6 g1f3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 c1e3 c5e3 f2e3 a8c8 a1c1 d8d6 d3b1 d6b8 d1e2 c6e7 c1d1 f8d8 e2b5 e7c6 d1d8 c8d8 f1d1 d8d1 c3d1 b8c7 d1c3 g8h8');
INSERT INTO evaluations VALUES(48316,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(48316,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(48316,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(48316,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48316,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(48316,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(48316,36,'cp -37','f1d3 c7c5 g1f3 c5c4 d3e2 b8c6 e1g1 a7a6 a2a4 f8d6 b2b3 c4b3 c2b3 d8e7 e2d3 c8d7 c1b2 c6b4 b2a3 e8g8 f3e5 f8c8 c3a2 b4c6 a3d6 e7d6 e5d7 f6d7 d1d2 c8c7 f1c1 e6e5 d4e5 d7e5 d3e2 a8d8 b3b4 d5d4 e3d4 d6d4 d2d4');
INSERT INTO evaluations VALUES(48327,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(48327,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(48327,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(48327,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(48327,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48327,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(48327,36,'cp -27','c7c5 g1f3 c5c4 d3e2 a7a6 a2a4 b8c6 e1g1 f8d6 b2b3 c4b3 c2b3 d8e7 e2d3 c8d7 c1b2 c6b4 b2a3 e8g8 f3e5 f8c8 d1d2 c8c7 f1c1 e7d8 c3a2 c7c1 d2c1 d6e5 a2b4 e5c7 c1b2 d8b8 h2h3 b7b5 a4b5 a6b5 d3e2 c7d6');
INSERT INTO evaluations VALUES(52424,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(52424,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(52424,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52424,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(52424,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(52424,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(52424,36,'cp 30','b8c6 b1c3 e7e5 g2g3 g7g6 d2d3 f8g7 f1g2 g8e7 a2a3 d7d6 a1b1 a7a5 e1g1 e8g8 f3e1 c8e6 c1g5 f7f6 g5d2 d6d5 c3a4 b7b6 b2b4 c5b4 a3b4 a5b4 d2b4 c6b4 b1b4 a8b8 d1b1 e5e4 b4b6 f6f5 c4d5 e7d5 b6b8 d8b8 b1b8');
INSERT INTO evaluations VALUES(19407,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(19407,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(19407,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(19407,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(19407,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19407,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19407,36,'cp 17','b1c3 e7e5 g2g3 g7g6 f1g2 d7d6 e1g1 f8g7 a2a3 a7a5 d2d3 g8e7 f3e1 c8e6 e1c2 e8g8 c2e3 c6d4 a1b1 a8b8 b2b4 c5b4 a3b4 b7b5 c3b5 d4b5 c4b5 b8b5 c1d2 a5b4 d1a4 d8d7 b1b4 f8b8 f1a1 b5b4 d2b4 d7a4');
INSERT INTO evaluations VALUES(53572,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(53572,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(53572,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(53572,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53572,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(53572,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(53572,36,'cp 20','g8f6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 f3e5 b4c3 b2c3 f6d7 h4g3 d7e5 g3e5 b7b6 e5d6 f8e8 c4b5 c8d7 b5e2 d7c6 d1d2 b8d7 f2f3 d7c5 f1d1 c6a4 d1b1 c5b7 d6a3 a8c8 c3c4');
INSERT INTO evaluations VALUES(6563,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(6563,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(6563,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(6563,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(6563,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6563,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(6563,36,'cp 15','d8f6 c2c3 d7d6 f1e2 f6g6 e1g1 g8f6 f2f3 d6d5 d1d2 d5e4 d4c6 c5e3 d2e3 b7c6 b1d2 e8g8 f3e4 f8e8 a1e1 c8e6 e4e5 e6d5 e2f3 f6d7 e3d4 a8d8 c3c4 d5f3 d2f3 c6c5 d4e3 g6e6 b2b3 f7f6 h2h3 a7a6 e3c3 f6e5');
INSERT INTO evaluations VALUES(296,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(296,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(296,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(296,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(296,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(296,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(296,36,'cp 25','c2c3 d7d6 b1a3 c6d4 e3d4 c5d4 d1d4 c8d7 a3b5 e8d8 f2f3 a7a6 d4f6 g8f6 b5d4 d7e6 a2a4 a6a5 g2g4 f6d7 f1d3 c7c6 h2h4 d8c7 e1f2 h7h6 d3c2 g7g6 h4h5 g6g5 f2g3 h8e8 b2b4 d7e5 a1d1 e5d7 d1d2 a5b4 c3b4');
INSERT INTO evaluations VALUES(294,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(294,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(294,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(294,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(294,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(294,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(294,36,'cp 8','d7d6 b1a3 c5a3 b2a3 f6e5 f1d3 g8f6 f2f4 e5a5 e1g1 f6g4 e3d2 a5c5 g1h1 c6d4 c3d4 c5d4 d1f3 g4f2 f3f2 d4d3 a1c1 c7c5 c1c3 d3d4 c3e3 d4b2 e3b3 b2d4 f2d4 c5d4 b3d3 f7f5 d3d4 e8g8 e4f5 c8f5 d4d6 f8e8 f1e1');
INSERT INTO evaluations VALUES(193,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(193,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(193,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(193,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(193,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(193,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(193,36,'cp 19','b1a3 c6d4 e3d4 c5d4 d1d4 c8d7 a3c2 f6d4 c2d4 a7a5 a2a4 g8e7 f1c4 e7c6 f2f3 c6d4 c3d4 f7f5 e4f5 d7f5 e1f2 e8c8 a1e1 d8e8 e1e8 h8e8 g2g4 f5d7 b2b3 h7h6 h2h4 c7c6 c4d3 c6c5 d4d5 c8c7 h4h5 e8e5 d3c4');
INSERT INTO evaluations VALUES(33761,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33761,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33761,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33761,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(33761,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(33761,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33761,36,'cp 0','d2d4 e7e5 g1f3 b8c6 c1e3 e5d4 f3d4 c8d7 d4c6 d5c6 d1d4 f8e7 e1c1 e7f6 d4c4 e8c8 h1g1 c6e6 c4e6 d7e6 d1d8 h8d8 a2a3 a7a5 g2g3 d8e8 f1b5 e6d7 b5c4 d7e6');
INSERT INTO evaluations VALUES(33752,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(33752,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33752,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33752,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(33752,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(33752,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33752,36,'cp 0','e7e5 g1f3 b8c6 c1e3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 d8e8 f1b5 f5d7 a2a4 c6e5 f3e5 e8e5 b5d7 c8d7 h1d1 d7c8 e3a7 b7b6 a4a5 c8b7 f2f4 e5e6 a7b6 c7b6 a5a6 b7a8 d1d8 a8a7 d8d7 a7a8');
INSERT INTO evaluations VALUES(33933,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33933,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33933,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33933,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33933,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33933,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33933,36,'cp 1','g1f3 b8c6 c1e3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 d8e8 f1b5 f5d7 a2a4 c6e5 f3e5 e8e5 e3d4 e5g5 g2g3 a7a6 b5d7 c8d7 h1d1 d7c6 d1e1 g5d5 c2c3 f8c5 d4c5 c6c5 e1e2 h8d8 c1c2 g7g5 f2f3 c5c6 a1f1 h7h5 f3f4 d5f5 b2b4 g5f4');
INSERT INTO evaluations VALUES(33932,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33932,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33932,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33932,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33932,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33932,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33932,36,'cp 0','b8c6 c1e3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 d8e8 f1b5 f5d7 a2a4 c6e5 f3e5 e8e5 b5d7 c8d7 h1d1 d7c8 c2c3 f8c5 e3c5 e5c5 c1c2 c5h5 h2h3 h5e5 d1d2 h8d8 g2g3 d8d2 c2d2 e5g5 h3h4 g5f5 d2e3 g7g6 a4a5 f5e5 e3d3 e5d5 d3e3');
INSERT INTO evaluations VALUES(981,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(981,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(981,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(981,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(981,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(981,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(981,36,'cp 0','c1e3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 d8e8 f1b5 f5d7 a2a4 c6e5 f3e5 e8e5 e3d4 e5g5 g2g3 a7a6 b5d7 c8d7 h1d1 d7c6 c2c3 g5h5 h2h4 f8c5 c1c2 h8e8 b2b4 c5d4 d1d4 e8e2 d4d2 e2d2 c2d2 h5d5 d2e3 d5e5 e3d3 e5d5');
INSERT INTO evaluations VALUES(16489,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(16489,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(16489,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(16489,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(16489,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(16489,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(16489,36,'cp 106','d4d5 c6b8 f3e5 d8e7 e5d3 e7e4 f1e2 e4g2 e2f3 g2h3 d1e2 f8e7 d5d6 c7d6 d3f4 h3h4 b1c3 e8g8 c1d2 d6d5 h1g1 b8c6 e1c1 d7d6 f4d5 f6d5 c3d5 e7d8 d2c3 d8g5 c1b1 c8f5 e2f1 a8e8 f3e2 f5h3 f1e1 c6e7 d5c7 e8c8 c7b5 e7f5 b5a7 c8c5 a7b5');
INSERT INTO evaluations VALUES(16338,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(16338,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(16338,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(16338,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(16338,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(16338,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(16338,36,'cp 104','c6b8 f3e5 d8e7 e5d3 e7e4 f1e2 e4g2 e2f3 g2h3 d1e2 f8e7 d5d6 c7d6 d3f4 h3h4 b1c3 b8c6 c1d2 e8g8 h1g1 d6d5 e1c1 d7d6 c3d5 f6d5 f4d5 e7d8 d2c3 d8g5 c1b1 c8f5 g1g3 a8e8 e2b5 h7h6 h2h3 e8c8 b5b7 c6e5 b7a7 e5f3 g3f3 h4e4 a7d4 e4d4 d1d4 g7g6 a2a3 c8e8');
INSERT INTO evaluations VALUES(22710,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(22710,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(22710,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22710,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22710,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22710,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(22710,36,'cp 0','d8f6 c2c3 d7d6 d2d3 c8e6 c4e6 f6e6 g1e2 g8f6 c1g5 f8e7 h2h4 h7h5 b1d2 f6d7 g5e3 d7f6');
INSERT INTO evaluations VALUES(1864,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1864,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1864,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1864,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1864,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1864,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1864,36,'cp 0','c2c3 d7d6 d2d3 c8e6 c4e6 f6e6 g1e2 g8f6 c1g5 f8e7 h2h4 h7h5 b1d2 f6d7 g5e3 d7f6');
INSERT INTO evaluations VALUES(1859,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1859,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(1859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1859,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1859,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1859,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1859,36,'cp 0','d7d6 d2d3 f8e7 b1d2 a7a5 f3e2 f6g6 g1f3 a5a4 d2f1 g8f6 h2h3 c8e6 f3h4 g6h5 h4f3');
INSERT INTO evaluations VALUES(31160,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31160,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(31160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31160,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31160,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31160,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(31160,36,'cp 7','f5d3 d1d3 e7e6 g1f3 c6c5 c2c3 b8d7 h2h4 a8c8 e1g1 g8e7 b1a3 e7c6 g2g3 f8e7 a3c2 d8b6 g1g2 c5d4 c3d4 h7h6 f1d1 e8g8 d3e2 a7a5 c2e1 c6b4 a2a3 b6a6 e2a6 b4a6 a1b1 d7b6 g2f1 a5a4 c1e3');
INSERT INTO evaluations VALUES(31145,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(31145,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31145,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31145,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(31145,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31145,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31145,36,'cp 5','c2d3 e7e6 b1c3 g8e7 g1f3 e7f5 h2h4 h7h5 c1g5 d8b6 c3e2 c6c5 d4c5 f8c5 d3d4 c5b4 e1f1 b8c6 g2g3 b4e7 g5e7 c6e7 d1d2 a7a5 a1c1 e8g8 f1g2 f8c8 c1c3 c8c3 b2c3 a8c8 e2f4 g7g6 h1c1 b6c6 f3g5 b7b5 g5f3 b5b4 f4e2 c6b7');
INSERT INTO evaluations VALUES(31143,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(31143,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31143,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(31143,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31143,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31143,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31143,36,'cp 0','e7e6 b1c3 g8e7 g1f3 e7f5 h2h4 h7h5 c1g5 d8b6 c3e2 c6c5 d4c5 f8c5 d3d4 c5b4 e1f1 b8c6 a1c1 b4e7 g5e7 c6e7 d1d2 e8g8 g2g3 f8c8 f1g2 a7a5 h1d1 g7g6 c1c3 c8c4 d1c1 a8c8 c3b3 b6c7 b3c3');
INSERT INTO evaluations VALUES(31492,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31492,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31492,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31492,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31492,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31492,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(31492,36,'cp 0','b1c3 g8e7 g1f3 e7f5 e1g1 b8d7 g2g3 h7h6 h2h3 f8e7 c3e2 a7a5 c1d2 d7f8 h3h4 h6h5 a1c1 f8g6 a2a3 e8g8 f1e1 d8e8 d1b3 e8d7 b3d1');
INSERT INTO evaluations VALUES(57158,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(57158,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(57158,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(57158,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(57158,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(57158,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(57158,36,'cp -47','d7d5 e4d5 d8d5 g1f3 c8g4 f1e2 b8c6 b1c3 d5a5 d2d3 f8b4 c1d2 e8c8 h2h3 g4f5 f3h4 f5e6 e2g4 e6d7 a2a3 b4c3 d2c3 a5c5 e1g1 g8e7 g4d7 d8d7 d1g4 g7g6 a1e1 f7f6 g4c4 c5c4 d3c4 g6g5 h4g2 c6d4 c3d4 d7d4 g2e3 d4d7 b2b4 h7h5');
INSERT INTO evaluations VALUES(56403,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(56403,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(56403,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(56403,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(56403,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(56403,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(56403,36,'cp -42','e4d5 d8d5 g1f3 c8g4 f1e2 b8c6 b1c3 d5a5 d2d3 e8c8 c1d2 f8b4 h2h3 g4f5 f3h4 f5e6 e2g4 e6d7 g4d7 d8d7 e1g1 g8e7 d1g4 g7g6 a1e1 f7f6 g4c4 e7d5 h4f3 h8d8 h3h4 b4c3 d2c3 d5c3 b2c3 d7e7 e1e3 f6f5 f1e1 a5d5 c4d5 d8d5');
INSERT INTO evaluations VALUES(56171,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(56171,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(56171,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(56171,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(56171,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(56171,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(56171,36,'cp -39','d8d5 g1f3 c8g4 f1e2 b8c6 b1c3 d5a5 d2d3 e8c8 c1d2 a5b6 e1g1 f7f5 a1b1 h7h5 d2e3 b6b4 f3g5 g4e2 d1e2 b4e7 b2b4 e7d7 h2h4 g8f6 e2f3 g7g6 a2a3 f8h6 c3e2 e5e4 d3e4 c6e5 f3g2 h6g5 e3g5 f6e4 g5d8');
INSERT INTO evaluations VALUES(34952,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(34952,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(34952,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(34952,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(34952,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(34952,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(34952,36,'cp -40','g1f3 c8g4 f1e2 b8c6 b1c3 d5a5 d2d3 e8c8 c1d2 f8b4 h2h3 g4f5 f3h4 f5e6 e2g4 e6d7 e1g1 g8e7 g4d7 d8d7 d1g4 g7g6 a1e1 f7f6 g4c4 e7d5 h4f3 h8d8 e1e2 d5c3 d2c3 b4c3 b2c3 c6e7 c4e6 d8f8 e6b3 e7d5 c3c4 d5c3 e2e1');
INSERT INTO evaluations VALUES(53405,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(53405,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(53405,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(53405,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(53405,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(53405,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(53405,36,'cp -35','g8f6 h2h3 b8c6 d2d4 d8a5 c2c3 c8f5 g1e2 e7e6 e1g1 f8e7 c1e3 h7h6 b1d2 e8g8 f1e1 f8c8 d2f3 b7b5 a2a3 a5b6 f3e5 c6a5 e2g3 f5h7 g3h5 b6b7 h5f6 e7f6 b3c2 h7c2 d1c2 f6e5 d4e5 a5c4 e3d4 a7a5');
INSERT INTO evaluations VALUES(45471,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(45471,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(45471,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(45471,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(45471,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(45471,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(45471,36,'cp -25','h2h3 b8c6 d2d4 c8f5 c2c3 e7e6 c1f4 a8c8 b1d2 f8d6 f4d6 d8d6 b3a4 f5d3 d2b3 d3c4 a4c6 c8c6 g1e2 d6c7 e1g1 a7a5 b3d2 c6b6 d2c4 c7c4 b2b3 c4a6 e2f4 b6c6 d1d2 e8g8 f4d3 a6b6 a1e1 b6c7');
INSERT INTO evaluations VALUES(45470,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(45470,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(45470,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(45470,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(45470,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(45470,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(45470,36,'cp -37','b8c6 d2d4 c8f5 c2c3 e7e6 c1f4 f8d6 f4d6 d8d6 g1f3 e8g8 e1g1 b7b5 b3c2 f5c2 d1c2 f6h5 b1d2 f7f6 f1e1 e6e5 a2a4 b5a4 c3c4 c6b4 c2f5 h5f4 c4d5 d6d5 a1a4 a7a5 e1a1 f4e2 g1h2 e2d4 f3d4 d5d4 d2b3 d4d5 a4a5 a8a5 b3a5');
INSERT INTO evaluations VALUES(22704,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22704,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(22704,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(22704,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(22704,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22704,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(22704,36,'cp 7','g8f6 d2d3 c6a5 c4b3 a5b3 a2b3 f8b4 g1f3 d7d6 h2h3 e8g8 e1g1 f8e8 c1d2 c8e6 f1e1 a7a6 c3e2 b4d2 d1d2 d6d5 e2c3 d5d4 c3e2 f6d7 f3g5 g8h8 f2f4 f7f6 g5f3 d8e7 c2c3 d4c3 d2c3 e6f7 d3d4 e5d4');
INSERT INTO evaluations VALUES(10033,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(10033,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(10033,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(10033,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(10033,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(10033,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(10033,36,'cp 64','f3e5 f6e4 d1e2 c6e5 d2d4 e5g6 d4c5 d8e7 b2b4 e8g8 c2c3 f8e8 f2f3 e4f6 e2e7 e8e7 f1d1 b7b6 c5b6 a7b6 a2a4 c7c6 b5f1 d7d5 c3c4 d5c4 f1c4 c8e6 c4f1 e6b3 d1d6 e7e1 b1c3 b3c4 c1b2 a8e8 g1f2 e1a1 b2a1 c4f1');
INSERT INTO evaluations VALUES(10015,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(10015,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(10015,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(10015,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(10015,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(10015,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(10015,36,'cp 47','f6e4 d1e1 c6e5 d2d4 c5e7 d4e5 e4c5 b1c3 c7c6 b2b4 c6b5 b4c5 e8g8 e1e3 b7b6 c5b6 a7b6 c3b5 c8b7 a2a4 d8c8 c1a3 e7a3 a1a3 c8c2 e3b6 c2c6 b6c6 b7c6 f1a1 f8b8 f2f4 a8a5 b5d4 b8b4 d4c6 d7c6 g2g3 g8f8 a1a2 f8e7 g1f2 f7f6 e5f6 e7f6 f2f3');
INSERT INTO evaluations VALUES(18805,30,'cp 264',NULL);
INSERT INTO evaluations VALUES(18805,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(18805,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(18805,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(18805,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(18805,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(18805,36,'cp 232','d4d5 c6a5 d5d6 b7b5 c1e3 c8b7 b1c3 a7a6 e3c5 g8f6 e4e5 f6d5 a2a4 d5c3 b2c3 d8c8 d1d4 b7f3 g2f3 g7g6 f3f4 f8g7 h2h4 e8g8 h4h5 b5a4 h1h3 a5c6 d4e3 c8b7 f1d3 a8b8 c5a3 b7b6 h5g6 b6e3 f2e3 h7g6 d3a6');
INSERT INTO evaluations VALUES(18865,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(18865,31,'cp 257',NULL);
INSERT INTO evaluations VALUES(18865,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(18865,33,'cp 250',NULL);
INSERT INTO evaluations VALUES(18865,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(18865,35,'cp 254',NULL);
INSERT INTO evaluations VALUES(18865,36,'cp 250','c6a5 d5d6 a7a6 e4e5 b7b5 c1e3 g8h6 a2a4 b5b4 b1d2 c8b7 e3c5 f7f6 c5b4 a5c6 b4c3 c6e5 f3e5 f6e5 d2c4 h6f7 d1d2 g7g6 c4e5 f8h6 f2f4 f7e5 c3e5 e8g8 f1d3 h6g7 e5g7 g8g7 h1f1 d8f6 e1c1 b7d5 c2c4 a8c8 b2b3 h7h5 c1c2 h5h4 d2e3 d5g2');
INSERT INTO evaluations VALUES(22628,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(22628,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(22628,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(22628,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(22628,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(22628,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(22628,36,'cp 75','d2d4 d7d5 b1d2 g8f6 e4e5 f6d7 d2b3 f8e7 c1f4 e8g8 c2c3 a7a5 d1c2 f7f6 f1d3 f6e5 f4e5 c6e5 d3h7 g8h8 f3e5 d7e5 d4e5 a5a4 b3d4 a4a3 b2a3 c7c5 d4f3 c8d7 e1g1 a8a3 c2b2 a3a8 h7d3 f8f3 g2f3 c5c4 d3c2');
INSERT INTO evaluations VALUES(50195,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50195,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50195,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50195,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50195,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50195,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50195,36,'cp 0','f8g7 f1e2 c7c5 g1f3 b8c6 c2c3 d8b6 b1a3 b6b2 a3b5 f6d5 a1c1 b2a2 c1a1 a2b2 a1b1 b2a2');
INSERT INTO evaluations VALUES(31827,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(31827,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(31827,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(31827,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(31827,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(31827,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(31827,36,'cp 110','c4d5 g8f6 g1f3 f6d5 b1d2 d5b6 e2e4 f5g6 h2h4 g6h5 a2a3 e7e5 d4e5 f8e7 f1e2 e8g8 d1c2 c7c5 f3g1 h5e2 g1e2 b8c6 d2f3 d8d7 c1e3 f8d8 e1g1 d7d3 c2d3 d8d3 a1c1 d3b3 e3c5 e7c5 c1c5 b6a4 c5c1 a4b2 e5e6 f7e6 e2f4 b3a3');
INSERT INTO evaluations VALUES(31813,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(31813,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(31813,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(31813,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(31813,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(31813,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(31813,36,'cp 109','g8f6 g1f3 f6d5 b1d2 d5b6 e2e4 f5g6 h2h4 h7h6 f1d3 b8d7 h4h5 g6h7 e1g1 e7e6 a2a4 a7a5 d2b3 f8b4 c1d2 d8e7 d1e1 e8g8 b3a5 c7c5 d2b4 c5b4 a5c4 b6c4 d3c4 d7f6 c4d3 f6h5 e1e3 h5f6 f1e1 f8c8 e1c1 g7g5 c1c8 a8c8 f3e5 f6d7 e5d7 e7d7');
INSERT INTO evaluations VALUES(29495,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(29495,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(29495,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(29495,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(29495,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(29495,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(29495,36,'cp 121','g1f3 f5b1 a1b1 d8d5 a2a3 b8d7 g2g3 e7e5 f1g2 e5e4 f3d2 d5d4 b2b4 c7c6 c1b2 d4d6 d2e4 d6d1 b1d1 f6e4 g2e4 a7a5 e1g1 d7f6 b2f6 g7f6 b4b5 f6f5 e4f5 f8a3 b5c6 b7c6 f5d7 e8f8 d7c6 a8c8 c6f3 f8g7 d1d5 c8c5 d5d7 h8b8 f1d1');
INSERT INTO evaluations VALUES(29493,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(29493,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(29493,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(29493,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(29493,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(29493,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(29493,36,'cp 104','f6d5 b1d2 d5b6 e2e4 f5g6 h2h4 h7h6 f1d3 b8d7 d2b3 e7e6 h4h5 g6h7 c1d2 f8e7 d1e2 e8g8 a1c1 c7c6 d3b1 g8h8 e1g1 d8c7 f1d1 a8e8 b3c5 d7c5 d4c5 b6d7 b2b4 a7a5 a2a3 a5b4 a3b4 d7f6 d2c3 f6h5 f3e5 h5f4 e2e3');
INSERT INTO evaluations VALUES(51952,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(51952,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(51952,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(51952,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(51952,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(51952,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(51952,36,'cp 89','d4d5 e6d5 c4d5 d7d6 g1f3 g7g6 e2e4 f8g7 b1c3 g8f6 f3d2 e8g8 f1e2 b8d7 e1g1 f8e8 a2a4 g6g5 f1e1 h7h6 d1c2 d7e5 d2f1 e5g6 f1e3 g6e5 c1d2 f6g4 e3f1 e5g6 f1g3 g6f4 c3d1 h6h5 e2g4 c8g4 d1e3 g7d4 h2h3 g4d7');
INSERT INTO evaluations VALUES(52026,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(52026,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(52026,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(52026,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(52026,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(52026,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(52026,36,'cp 87','e6d5 c4d5 g7g6 e2e4 d7d6 g1f3 f8g7 b1c3 g8f6 f3d2 e8g8 f1e2 b8d7 e1g1 f8e8 a2a4 g6g5 f1e1 h7h6 d1c2 d7e5 d2f1 e5g6 f1g3 g6f4 e2b5 e8e7 c1d2 f6g4 b5e2 g4e5 c3d1 f4e2 e1e2 c8g4 e2e3 a7a6 d2c3 g4d7');
INSERT INTO evaluations VALUES(52169,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(52169,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(52169,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(52169,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(52169,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(52169,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(52169,36,'cp 86','c4d5 g7g6 e2e4 d7d6 f1e2 f8g7 g1f3 c8g4 e1g1 g8f6 c1f4 e8g8 b1d2 d8e7 h2h3 g4f3 e2f3 b8d7 a2a4 f6e8 f3e2 g7d4 a1b1 a8c8 e2d3 d7e5 d2b3 e5d3 d1d3 d4e5 f4h6 e5g7 h6g7 e8g7 b3d2 f7f5 e4f5 f8f5 b1e1 e7h4');
INSERT INTO evaluations VALUES(52159,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(52159,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(52159,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(52159,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(52159,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(52159,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(52159,36,'cp 84','g7g6 b1c3 d7d6 e2e4 f8g7 g1f3 g8f6 f1e2 c8g4 e1g1 e8g8 c1f4 a7a6 a2a4 g4f3 e2f3 d8e7 e4e5 d6e5 d5d6 e7e6 f4g5 b8d7 f3b7 a8a7 b7d5 f6d5 d1d5 e6d5 c3d5 h7h6 g5e3 a7b7 a1c1 b7b2 f1d1 f8b8 e3c5 d7c5');
INSERT INTO evaluations VALUES(52433,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(52433,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(52433,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(52433,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(52433,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(52433,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(52433,36,'cp -5','c5d4 d1d4 b8c6 d4a4 g8f6 e2e4 d7d6 f1e2 e7e6 e1g1 d8c7 c2c3 f8e7 c1f4 e8g8 f1d1 c8d7 a4c2 c6e5 f3d4 d6d5 e4d5 f6d5 f4g3 f7f5 g3e5 c7e5 b1d2 e5c7 e2c4 e7d6 h2h3 g8h8 c4d5 e6d5');
INSERT INTO evaluations VALUES(52525,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(52525,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(52525,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(52525,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(52525,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(52525,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(52525,36,'cp -8','f3d4 d7d5 c2c4 d5c4 e2e3 e7e5 d4f3 d8d1 e1d1 b8c6 f1c4 c8f5 d1e2 a8c8 c4b5 f8d6 b1d2 d6b8 d2c4 f7f6 c4a5 f5d7 a5c6 d7c6 b5c6 c8c6 c1d2 g8e7 h1c1 c6c1 a1c1 e8d7 e3e4 e7c6 d2e3 b8d6 h2h3 h7h5 a2a3 h8c8 b2b4 a7a6 f3d2 c6d4 e2d1');
INSERT INTO evaluations VALUES(52513,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(52513,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(52513,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(52513,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(52513,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52513,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(52513,36,'cp -27','d7d5 e2e4 d5e4 b1c3 a7a6 f1c4 g8f6 c1g5 d8c7 d1e2 b8d7 e1c1 h7h6 g5f6 d7f6 c3d5 f6d5 c4d5 e7e6 d5e4 f8e7 g2g3 c8d7 f2f4 e8g8 c2c3 d7a4 e4c2 a4c2 c1c2 e7f6 c2b1 f8d8 a2a3 g7g6 d4f3 f6g7 f3e5 g7e5 f4e5');
INSERT INTO evaluations VALUES(53374,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(53374,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(53374,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(53374,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(53374,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(53374,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(53374,36,'cp -29','c2c4 e7e5 d4f3 d5d4 e2e3 b8c6 e3d4 e5d4 f1d3 d8e7 d3e2 c8g4 e1g1 e8c8 e2d3 e7c7 b1a3 f8a3 b2a3 g8f6 h2h3 g4h5 a1b1 h8e8 c1g5 h5g6 d1b3 d8d7 f1e1 f6e4 f3d2 f7f5 d2f3 g6h5 b3c2 h7h6 g5d2 d7d8 c4c5 h5f3 g2f3 e4d2 d3f5 c8b8 c2d2 e8e1 d2e1');
INSERT INTO evaluations VALUES(53356,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(53356,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(53356,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(53356,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(53356,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(53356,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(53356,36,'cp -12','e7e5 d4c2 d5d4 e2e3 b8c6 f1d3 g8f6 e3d4 e5d4 e1g1 c8g4 d1e1 f8e7 h2h3 g4e6 b1d2 f6d7 d3e4 d7e5 b2b3 d4d3 c2e3 e8g8 c1b2 e5g6 d2f3 g6f4 f3e5 e7f6 e4c6 b7c6 e5c6 d8d7 b2f6 g7f6 e3d5 e6d5 c6e7 g8g7 e7d5 f4d5 c4d5 d7d5 a1d1 a8d8 e1c3 d3d2');
INSERT INTO evaluations VALUES(49403,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(49403,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(49403,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(49403,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49403,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(49403,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49403,36,'cp 6','c7c5 e2e3 d7d5 b1d2 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 c5d5 f1c4 d5d6 e1g1 f8d8 d1c2 b8d7 d2e4 d6e7 c4e2 b7b6 f1d1 c8b7 e4f6 e7f6 c2c7 d7c5 d1d8 f6d8 c7d8 a8d8 b1c1 b7d5 a2a3 d5b7 f3d4');
INSERT INTO evaluations VALUES(44224,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44224,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44224,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44224,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(44224,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44224,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(44224,36,'cp 21','e2e3 d7d5 b1d2 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 f6d5 b1c1 c5e7 f1c4 d5f6 e1g1 f8d8 d1c2 b7b6 c4b5 c8b7 f1d1 b8a6 f3d4 a6b4 c2b1 a7a6 b5e2 a8c8 a2a3 c8c1 d1c1 b4d5');
INSERT INTO evaluations VALUES(9773,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(9773,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(9773,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(9773,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(9773,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(9773,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9773,36,'cp 17','a2a4 d7d6 a4a5 e8g8 c2c3 a7a6 h2h3 c5a7 f1e1 c8e6 b1d2 d8d7 c4b3 e6b3 d1b3 f8e8 d2f1 d6d5 f1g3 a8b8 g3f5 d7e6 b3c2 b8d8 c1e3 a7e3 f5e3 d5e4 d3e4 c6e7 a1d1 e7g6 c3c4 d8d1 e1d1 e6c6 e3d5 f6d5');
INSERT INTO evaluations VALUES(9815,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(9815,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(9815,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(9815,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9815,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9815,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(9815,36,'cp 0','e8g8 c1e3 c5b6 b1c3 d7d6 c3d5 f6d5 c4d5 d8f6 h2h3 c8e6 d5e6 f7e6 c2c3 f8f7 e3b6 a7b6 b2b4 a8f8 f3h2 c6b8 d1g4 f6e7 a1a2 e7f6');
INSERT INTO evaluations VALUES(14814,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(14814,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(14814,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(14814,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(14814,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(14814,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(14814,36,'cp 50','c2c4 f8d6 f4g3 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 a1c1 h7h6 e1g1 c8g4 g3d6 d8d6 d1c2 c7c6 f3e5 g4d7 b2b4 d7e6 c2b2 f7f6 e5f3 e6f5 d3a6 a8b8 a6e2 f8c8 h2h3 f5e4 a2a4 e7f5 b4b5 c6b5 a4b5 b8a8');
INSERT INTO evaluations VALUES(14772,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(14772,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14772,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14772,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(14772,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(14772,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14772,36,'cp 50','f8d6 f4g3 e8g8 b1c3 b7b6 c4d5 f6d5 c3e4 c8b7 f1d3 c6b4 d3b1 f7f5 e4d6 c7d6 e1g1 d8d7 a2a3 b4c6 f1e1 d5f6 f3d2 c6e7 f2f3 f6h5 g3f2 d7b5 b2b3 b5d7 b1d3 e7g6 a1c1 a8c8 c1c8 f8c8');
INSERT INTO evaluations VALUES(9224,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(9224,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(9224,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(9224,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(9224,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(9224,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(9224,36,'cp 47','f4g3 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 e1g1 c8f5 a1c1 f5d3 d1d3 a7a5 c1c2 c7c6 f1c1 d8d7 g3d6 d7d6 h2h4 a5a4 d3c3 f8c8 f3e1 h7h5 e1d3 c8c7 c3b4 d6b4 d3b4 a8c8 b4d3 f7f6 d3f4 g7g6 b2b3 a4b3 a2b3 c8a8 f4d3');
INSERT INTO evaluations VALUES(9226,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(9226,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(9226,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(9226,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(9226,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(9226,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(9226,36,'cp 40','e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 e1g1 g7g6 f1e1 c7c6 d1a4 d6g3 h2g3 d8d6 e3e4 d5e4 d3e4 c8d7 f3e5 f8c8 g3g4 d7e6 a1c1 c8c7 e5c6 e7c6 c1c6 c7c6 e4c6 a8d8 d4d5 e6d5 c6d5 d6d5 a4a7');
INSERT INTO evaluations VALUES(55137,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55137,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55137,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55137,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55137,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(55137,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(55137,36,'cp 0','g8f6 d2d4 e7e6 b2b3 f8d6 c2c4 b7b6 b1c3 e8g8 f1d3 c8b7 e1g1 d5c4 b3c4 c7c5 c1b2 c5d4 e3d4 b8d7 f1e1 f8e8 d1e2 a7a6 a1d1 d8c7 f3e5 a8c8 f2f4 d7f8 e2f2 f8g6 a2a3 e8d8 g1h1 f6h5 b2c1 h5f6 c1b2');
INSERT INTO evaluations VALUES(26538,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(26538,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(26538,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(26538,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(26538,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(26538,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(26538,36,'cp 51','c2c4 g8f6 b1c3 e7e6 c4c5 f8e7 f1b5 e8g8 g1f3 f6e4 c3e4 f5e4 e1g1 e7f6 f3d2 e4f5 f4g3 a7a6 b5a4 d8e7 d2b3 c6b4 b3c1 b4c6 c1d3 f5d3 d1d3 f6h4 g3h4 e7h4 a2a3 g7g6 b2b4 c6e7 g2g3 h4e4 d3e4 d5e4 f1d1 c7c6');
INSERT INTO evaluations VALUES(26521,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(26521,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(26521,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(26521,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(26521,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(26521,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(26521,36,'cp 36','g8f6 b1c3 e7e6 a2a3 h7h6 g1f3 f8d6 f4d6 c7d6 c4d5 f6d5 c3d5 e6d5 f1d3 f5e4 d3b1 d8f6 e1g1 e4f3 d1f3 f6f3 g2f3 g7g6 f3f4 c6e7 b1d3 e8d7 g1g2 d7e6 g2f3 a8c8 h2h3 c8c7 f1h1 h8c8 h3h4 h6h5 f3e2 a7a6 b2b4 e7f5 e2d2');
INSERT INTO evaluations VALUES(24798,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24798,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(24798,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(24798,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(24798,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(24798,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(24798,36,'cp 57','b1c3 e7e6 c4c5 f6e4 c3e4 f5e4 g1f3 f8e7 f1b5 e8g8 e1g1 a7a5 h2h3 e7f6 f3d2 e4g6 d1b3 d8e7 b5c6 b7c6 b3c3 e7d7 d2b3 a5a4 b3a5 g6f5 b2b4 a4b3 a2b3 h7h5 a1a4 g7g5 f4e5 f6e5 d4e5 g5g4 h3h4 d7e7 a5c6 e7h4 c6d4');
INSERT INTO evaluations VALUES(24796,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(24796,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(24796,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(24796,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(24796,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(24796,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(24796,36,'cp 33','e7e6 a2a3 h7h6 g1f3 f8d6 f4d6 c7d6 c4d5 f6d5 c3d5 e6d5 f1d3 f5e4 e1g1 e8g8 d3b1 d8b6 f3d2 e4b1 a1b1 a8c8 d2b3 c6e7 b3c1 c8c4 h2h4 f8c8 c1d3 c4c2 d3b4 c2c4 d1g4 a7a5 b4d3 b6b3 d3f4 c4c7 f4e2 c7c2 g4d7 c2e2 d7e7');
INSERT INTO evaluations VALUES(25135,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(25135,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(25135,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(25135,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(25135,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(25135,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(25135,36,'cp 51','c4c5 f8e7 f1b5 f6e4 g1f3 e8g8 c3e4 f5e4 e1g1 a7a5 f3d2 e4g6 d1b3 e7g5 f4g5 d8g5 b3c3 f8b8 h2h3 g5e7 f1c1 h7h6 b2b3 f7f6 a2a3 e6e5 b3b4 a5b4 a3b4 e5d4 e3d4 g8h7 c1e1 e7d7 b5c6 d7c6');
INSERT INTO evaluations VALUES(43398,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(43398,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43398,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43398,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43398,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43398,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43398,36,'cp 0','e7e6 a2a4 f8e7 c1e3 e8g8 e1g1 b8c6 d4c6 b7c6 a4a5 c8b7 f1e1 c6c5 d1e2 f6d7 f2f4 d8c7 e3f2 e7f6 e2d2 f8d8 h2h3 h7h6 d2d3 f6e7 f2g3 e7f6');
INSERT INTO evaluations VALUES(38409,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(38409,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(38409,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(38409,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(38409,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(38409,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(38409,36,'cp 59','g8f6 c1e3 a7a6 a2a4 b8c6 f1e2 e8g8 e1g1 f6g4 e3g5 h7h6 g5h4 g4f6 h2h3 f6h5 d4d5 c6b8 a4a5 g6g5 f3h2 h5f4 h4g3 e7e5 a1a3 f4e2 d1e2 f7f5 e4f5 c8f5 h2g4 b8d7 f2f3 d7f6 g4f6 d8f6 c3e4 f6f7 f1d1 a8e8 g3f2 f5e4 f3e4 h6h5 a3g3 f7f4');
INSERT INTO evaluations VALUES(18733,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18733,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18733,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(18733,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18733,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18733,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(18733,36,'cp 13','g1f3 g8f6 d2d4 f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 f2f3 e4f3 d2f3 f8c5 c1f4 d8f7 e1c1 e8e7 h1e1 h8d8 d1d8 e7d8 c4e6 c8e6 e1e6 d8d7 e6e2 c5e7 c2c3 a8e8 c1c2 g7g5 e2d2 d7e6 f4e3 a7a6 d2e2 e6d7 e3b6 f7d6 f3d4 e7f6 e2e8 d7e8 c2d3 e8e7 c3c4');
INSERT INTO evaluations VALUES(56933,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(56933,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(56933,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(56933,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(56933,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(56933,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(56933,36,'cp -25','h2h4 g5g4 f3e5 g8f6 e5g4 b8c6 b1c3 c6d4 g4f6 d8f6 d2d3 c7c6 c3e2 d4e6 d1d2 d7d5 e2f4 d5e4 d3e4 f8h6 g2g3 h8g8 e4e5 f6e5 d2e2 h6f4 c1f4 e5e2 f1e2 e6f4 g3f4 c8f5 e1c1 g8g3 e2d3 f5d3 d1d3 g3d3 c2d3 e8e7 h1e1 e7f6 e1g1 a8d8 c1c2 a7a6 g1e1 h7h5');
INSERT INTO evaluations VALUES(56930,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(56930,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(56930,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(56930,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(56930,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(56930,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(56930,36,'cp -30','g5g4 f3e5 d7d6 e5g4 b8c6 b1c3 f8e7 d2d4 e7h4 g4f2 d8g5 f1b5 c8d7 b5c6 h4f2 e1f2 d7c6 d1f3 g5g6 f3f4 e8c8 d4d5 c6d7 f4g5 g8f6 g5g6 h7g6 c1e3 f6g4 f2f3 f7f5 h1h8 d8h8 e3f4 b7b5 e4f5 b5b4 c3e4 d7f5 a1e1 c8b7 c2c3 b4c3 e4c3 g4f6 f4g5');
INSERT INTO evaluations VALUES(56939,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56939,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(56939,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(56939,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(56939,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(56939,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(56939,36,'cp -34','f3e5 d7d6 e5g4 b8c6 g4f2 d8f6 c2c3 c8d7 d2d4 h7h5 f2d3 f8h6 d1f3 d7g4 f3f2 f6e7 d3f4 e7e4 f1e2 e8c8 e1g1 g4e2 f2e2 d8e8 e2e4 e8e4 b1a3 c6e7 f4h3 f7f6 c1h6 g8h6 f1e1 e4e1 a1e1 e7f5 h3f4 f5h4 f4d5 h4g6 d5f6 h6g4 f6g4 h5g4 a3c2 g6f4 g1f2');
INSERT INTO evaluations VALUES(56923,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(56923,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(56923,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(56923,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(56923,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(56923,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(56923,36,'cp -35','d7d6 e5g4 b8c6 g4f2 d8f6 c2c3 h7h5 d2d4 c8d7 d1f3 f8h6 f2d3 d7g4 f3f2 f6e7 d3f4 e7e4 f1e2 g4e2 f2e2 e4e2 f4e2 h6c1 e2c1 e8c8 e1g1 c6e7 b1d2 g8h6 c1d3 e7f5 f1f4 d8e8 d2e4 e8e6 a1e1 d6d5 e4c5 e6e1 d3e1 b7b6 c5d3 h8e8 e1c2 e8e2 f4f2 e2f2 g1f2 f5h4');
INSERT INTO evaluations VALUES(55342,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55342,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(55342,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(55342,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(55342,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(55342,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(55342,36,'cp -45','e5g4 b8c6 g4f2 d8f6 c2c3 h7h5 d2d4 c8d7 f2d3 f8h6 d1f3 d7g4 f3f2 f6e7 d3f4 e7e4 c1e3 h6f4 f2f4 e4c2 b1d2 e8c8 f1b5 g8e7 e1g1 e7d5 f4f2 c2b2 a1b1 b2c2 b1c1 c2g6 e3g5 f7f6 c3c4 f6g5 c4d5 c6e7 f2f7 g6f7 f1f7 e7d5 h4g5 h8f8 g5g6 f8f7');
INSERT INTO evaluations VALUES(55361,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(55361,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(55361,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(55361,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(55361,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(55361,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(55361,36,'cp -38','f8e7 d2d4 e7h4 g4f2 h4g3 d1f3 d8f6 b1c3 f6d4 c1f4 g3f4 f3f4 c8e6 a1d1 d4f6 f1b5 b8d7 f4e3 g8e7 h1h5 e8c8 e3a7 d7b8 b5a4 d6d5 e4d5 e7d5 a7d4 f6d4 d1d4 d5c3 b2c3 e6a2 f2e4 d8g8 e1f2 a2e6 e4g5 g8g6 d4f4 h8g8 g5e6 g6g2 f2f1 f7e6 h5h7 b8a6 a4d7 c8b8 d7e6');
INSERT INTO evaluations VALUES(19558,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(19558,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(19558,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(19558,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(19558,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19558,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(19558,36,'cp -22','g1f3 e7e6 f1e2 g8f6 e1g1 d7d6 d1b3 f8e7 f1d1 e6e5 c1g5 e8g8 g5f6 e7f6 a1c1 a8b8 b3a3 f6e7 c3d5 e7g5 f3g5 d8g5 a3d6 c8e6 d6a3 f8d8 e2c4 h7h6 a3e3 b8c8 e3g5 h6g5 f2f3 g8f8 d5e3 e6c4 e3c4 f7f6 c4e3 d8d1 c1d1 c6d4 g1f2 f8f7 h2h3 g7g6');
INSERT INTO evaluations VALUES(19557,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(19557,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(19557,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19557,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(19557,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(19557,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(19557,36,'cp -15','e7e6 f1e2 f8e7 e1g1 d7d6 d1b3 g8f6 f1d1 e6e5 c1g5 e8g8 g5f6 e7f6 a1c1 h7h6 c3d5 f6g5 c1c3 a8b8 g2g3 c8g4 h2h4 g5e7 b3a3 f8e8 e2b5 e7f8 b5c6 b7c6 c3c6 d8d7 d1c1 e8c8 c6c8 b8c8 c1c8 d7c8 a3b3 g4e6 b3d1 c8c5 h4h5 a7a6 g1g2 g8h8');
INSERT INTO evaluations VALUES(18064,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(18064,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(18064,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(18064,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(18064,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(18064,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(18064,36,'cp 50','b1c3 d8f6 d2d4 e5d4 f3d4 f8b4 e1g1 b4c3 b2c3 g8e7 c1a3 d7d6 f1e1 e8g8 e4e5 d6e5 d1e2 c8d7 e2e5 f6e5 e1e5 e7g6 e5c5 f8e8 f2f3 g6f4 a1d1 f7f6 g1f2 e8e5 c5e5 f6e5 d4b3 d7f5 b3a5 f5c2 d1d2 c2f5 g2g3 f4g6 g3g4 f5e6 f2g3 h7h6 a5c6 a8e8');
INSERT INTO evaluations VALUES(18061,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(18061,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(18061,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(18061,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(18061,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(18061,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(18061,36,'cp 39','d8f6 d2d4 e5d4 f3d4 f8b4 e1g1 b4c3 b2c3 g8e7 c1a3 d7d6 f1e1 e8g8 e4e5 d6e5 d1e2 c8d7 e2e5 f6e5 e1e5 e7g6 e5c5 f8e8 f2f3 e8e5 a1b1 e5c5 a3c5 g6f4 b1b7 f4e6 d4e6 d7e6 a2a3 a8d8 c5d4 c6c5 d4e3 h7h6 g1f2 d8d1 b7c7 d1a1');
INSERT INTO evaluations VALUES(19372,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(19372,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(19372,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(19372,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(19372,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(19372,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19372,36,'cp 0','a7a6 c4d5 c6d4 g1e2 e7e6 d5b3 b7b5 e2d4 c5d4 c3e2 c8b7 d2d3 f8c5 a2a4 g8e7 e1g1 e8g8 a4b5 a6b5 a1a8 d8a8 f2f3 f7f5 c1g5 e7g6 e2f4 a8b8 f4g6 h7g6 d1e1 c5d6 e1h4 f8f7 g1h1 b7c6 h1g1');
INSERT INTO evaluations VALUES(17478,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17478,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(17478,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17478,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17478,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17478,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(17478,36,'cp 0','c4d5 c6d4 g1e2 e7e6 d5b3 b7b5 e2d4 c5d4 c3e2 c8b7 d2d3 f8c5 a2a4 g8e7 a4b5 a6b5 a1a8 d8a8 e1g1 e8g8 g1h1 f7f5 f2f3 e7c6 e4f5 f8f5 e2g3 f5f7 d1e1 c6e7 g3e4 a8a7 e4g5 f7f6 g5e4');
INSERT INTO evaluations VALUES(57156,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(57156,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(57156,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(57156,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(57156,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(57156,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(57156,36,'cp -100','d7d5 e4d5 d8d5 b1c3 d5e6 f3f4 e5f4 f1e2 f8d6 g1f3 g8f6 e1g1 e8g8 d2d4 e6e7 c3b5 b8c6 e2d3 f8e8 f1e1 c8e6 c2c3 h7h6 d1d2 f6d5 a2a3 a7a6 b5d6 e7d6 c3c4 d5e3 d4d5 a8d8 d2f2 e6g4 c1e3 f4e3 e1e3 e8e3 f2e3 g4f3 e3f3 c6e5 f3f5');
INSERT INTO evaluations VALUES(56400,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(56400,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(56400,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(56400,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(56400,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(56400,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(56400,36,'cp -107','e4d5 d8d5 b1c3 d5e6 f1b5 c8d7 b5d7 b8d7 d2d3 g8e7 g1e2 e8c8 e1g1 e7f5 d1e1 f8e7 g1h1 h7h5 c3e4 h5h4 h2h3 e6c6 e1c3 c6c3 b2c3 d7f6 c3c4 f6h5 f3f4 f5d4 e2d4 e5d4 h1g1 d8e8 e4g5 h5g3 g5f7');
INSERT INTO evaluations VALUES(56168,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(56168,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(56168,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(56168,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(56168,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(56168,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(56168,36,'cp -77','d8d5 b1c3 d5e6 f3f4 e5f4 f1e2 f8d6 d2d4 g8f6 g1f3 e8g8 e1g1 e6e7 c3b5 b8c6 c2c3 h7h6 e2c4 a7a6 f1e1 e7d8 b5d6 d8d6 d1d2 f6h5 d2f2 c8g4 f3e5 c6e5 d4e5 d6c6 h2h3 g4f5 c4e2 c6g6 e2h5 g6h5 c1f4 f8e8 e1e2 f5e6 e2e3 a8d8 a1e1');
INSERT INTO evaluations VALUES(34950,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(34950,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(34950,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(34950,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(34950,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(34950,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(34950,36,'cp -76','b1c3 d5e6 f3f4 e5f4 f1e2 f8d6 d2d4 g8f6 g1f3 e8g8 e1g1 e6e7 c3b5 b8c6 c2c3 h7h6 e2c4 g7g5 f1e1 e7d8 h2h4 f6g4 h4g5 h6g5 b5d6 c7d6 g2g3 d8f6 g3f4 g5f4 d1d2 g4e3 e1e3 f6g7 d2g2 f4e3 c1e3 c8e6 d4d5 e6g4 d5c6 g4f3 g2g7 g8g7 e3d4 g7h6 c6c7 a8c8 g1f2 c8c7 c4b3');
INSERT INTO evaluations VALUES(34939,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(34939,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(34939,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(34939,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(34939,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(34939,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(34939,36,'cp -76','d5e6 f3f4 e5f4 f1e2 f8d6 d2d4 g8f6 g1f3 e6e7 c3b5 e8g8 e1g1 b8c6 c2c3 h7h6 e2c4 g7g5 f1e1 e7d8 h2h4 f6e8 h4g5 h6g5 g2g3 d8f6 c4d3 c8g4 d1e2 g4f5 d3f5 f6f5 g3f4 d6f4 c1f4 f5f4 d4d5 c6e7 e2g2 e7d5 g2g5 f4g5 f3g5 e8f6 g5f3 f8e8 e1e8 f6e8 a1e1 e8f6');
INSERT INTO evaluations VALUES(47432,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47432,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47432,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(47432,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(47432,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(47432,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47432,36,'cp 0','d8d5 d2d4 b8c6 c1e3 e7e5 g1f3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 d8e8 f1b5 f5d7 a2a4 c6e5 f3e5 e8e5 b5d7 c8d7 h1d1 d7c8 e3a7 b7b6 a4a5 c8b7 a7b6 c7b6 a5a6 b7a7 d1d7 a7a8 d7d8 a8a7');
INSERT INTO evaluations VALUES(56698,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(56698,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(56698,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(56698,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(56698,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(56698,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(56698,36,'cp 225','f1c4 b8c6 e1g1 h7h5 b2b4 a7a6 a2a4 g7g5 b1c3 h5h4 c1e3 g5g4 f3h4 h8h4 c4g8 h4h8 g8b3 f6f5 d4e5 d8h4 e3f4 f8g7 f4g3 h4h6 e4f5 g7e5 f1e1 c8f5 b4b5 e8c8 e1e5 c6e5 b5a6 b7a6 d1e2 c8b7 c3d5 e5c6 a1e1 c6e5 g3e5 d6e5 e2e5');
INSERT INTO evaluations VALUES(56692,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(56692,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(56692,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(56692,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(56692,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(56692,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(56692,36,'cp 226','b8c6 e1g1 h7h5 b2b4 g7g5 b4b5 c6a5 c4e2 f8h6 c1d2 b7b6 d2a5 b6a5 d4e5 g5g4 f3d4 f6e5 d4c6 d8f6 b1c3 c8e6 c3d5 e6d5 d1d5 f6f7 d5d3 g8f6 f2f3 f7g7 f3g4 h5g4 f1f5 e8g8 a1f1 g8h8 g2g3 f6d7 g1g2 f8f5 f1f5 a5a4 d3c4 a8f8');
INSERT INTO evaluations VALUES(22309,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(22309,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(22309,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(22309,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(22309,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(22309,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(22309,36,'cp 221','e1g1 c6a5 c4e2 a5c6 b1c3 c8g4 c1e3 g4f3 e2f3 e5d4 e3d4 c6d4 d1d4 c7c6 a1d1 g8h6 c3e2 d8d7 e2f4 h6f7 f4h5 c6c5 d4e3 e8c8 h5f4 c8b8 f4d5 f7e5 e3b3 d7c8 f3h5 g7g6 h5e2 g6g5 a2a4 c5c4 b3c3 g5g4');
INSERT INTO evaluations VALUES(22308,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(22308,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(22308,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(22308,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(22308,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(22308,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(22308,36,'cp 210','g7g6 h2h3 f8h6 b1c3 h6c1 a1c1 e8f8 d1d2 f8g7 c1d1 c6d4 f3d4 e5d4 d2d4 g8e7 f2f4 e7c6 d4e3 h8e8 a2a3 g7h8 c3d5 c8e6 e3c3 e6d5 c4d5 d8e7 b2b4 c6d8 a3a4 a7a6 d1d3 c7c6 d5b3 d8e6 c3d2 e6g7 d3d6 e7e4 d6f6 e4e3 d2e3 e8e3');
INSERT INTO evaluations VALUES(57126,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(57126,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(57126,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(57126,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57126,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(57126,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(57126,36,'cp 7','g8f6 b1c3 b8c6 g1f3 f8b4 d1b3 b4c3 b3c3 d8e7 d2d4 e5e4 f3g1 d7d5 b2b3 e8g8 g1e2 e7d8 e2f4 c6e7 h2h3 f8e8 a2a4 c7c6 c1a3 a7a5 f1e2 e7g6 f4g6 h7g6 c3d2 b7b6 g2g4 f6h7 e1c1 d8h4 d2e1 d5c4 e2c4 c8e6 c1b2 e6c4');
INSERT INTO evaluations VALUES(49568,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49568,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49568,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(49568,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(49568,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(49568,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49568,36,'cp 9','b1c3 f8b4 g1e2 d7d5 c4d5 f6d5 d1b3 d5c3 b2c3 b4d6 d2d4 e8g8 e2g3 b8d7 f1d3 d7f6 e1g1 c7c5 d4c5 d6c5 c3c4 c8e6 f1d1 d8c7 g3e4 f6e4 d3e4 a8d8 e4d5 e6g4 d1e1 g4e6 d5e6 f7e6 c1a3 c5a3 b3a3 c7c4');
INSERT INTO evaluations VALUES(49556,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(49556,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49556,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49556,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(49556,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(49556,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(49556,36,'cp 4','f8b4 g1e2 d7d5 c4d5 f6d5 a2a3 d5c3 e2c3 b4d6 b2b4 e8g8 c1b2 a7a5 b4a5 a8a5 f1e2 d6e7 d1c2 c8e6 e1g1 h7h6 f1c1 c7c6 c3e4 b8d7 a3a4 d7c5 e4c5 a5c5 b2c3 e7d6 e2f3 d8d7 a4a5 f8a8 c2b1 d6f8');
INSERT INTO evaluations VALUES(42768,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(42768,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(42768,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42768,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(42768,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(42768,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42768,36,'cp 0','g1e2 e8g8 a2a3 b4c3 e2c3 d7d5 c4d5 f6d5 d1c2 c8e6 f1d3 d5c3 b2c3 h7h6 a3a4 d8d7 e1g1 c7c5 c1a3 f8d8 d3h7 g8h8 a3c5 d7d2 c2d2 d8d2 h7e4 b8d7 c5a3 a8b8 a3c1 d2e2 c1a3');
INSERT INTO evaluations VALUES(42769,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42769,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42769,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42769,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42769,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42769,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42769,36,'cp 0','e8g8 a2a3 b4c3 e2c3 d7d5 c4d5 f6d5 d1c2 c8e6 f1d3 d5c3 d2c3 d8h4 e3e4 a7a5 e1g1 a5a4 c2e2 b8c6 f2f4 f8d8 c1e3 e6g4 e2c2 g4e6 c2e2');
INSERT INTO evaluations VALUES(38628,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38628,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(38628,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(38628,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(38628,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(38628,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(38628,36,'cp 17','c2c4 g8f6 c4c5 d6e7 d1a4 c8d7 a4c2 e8g8 b1c3 f8e8 f1e2 d8c8 e1g1 d7f5 e2d3 f5d3 c2d3 c7c6 c1f4 b8d7 b2b4 f6e4 f1e1 e4c3 d3c3 d7f6 c3b3 f6e4 a2a4 e7f6 f4e5 c8f5 e5f6 f5f6 e1e3 e8e6 a1e1 a8e8 b4b5');
INSERT INTO evaluations VALUES(38612,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38612,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(38612,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38612,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(38612,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38612,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(38612,36,'cp 0','g8f6 c4c5 d6e7 f1d3 b7b6 c1e3 e8g8 e1g1 b6c5 d4c5 f6g4 e3f4 g4f6');
INSERT INTO evaluations VALUES(41689,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(41689,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41689,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(41689,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(41689,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(41689,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41689,36,'cp 13','c4c5 d6e7 d1a4 c8d7 a4c2 e8g8 c1e3 b8c6 b1c3 d8c8 f1b5 d7f5 c2d1 a7a6 b5c6 b7c6 e1g1 f8e8 f1e1 f6d7 h2h3 e7f6 d1d2 h7h6 e3f4 c8b7 b2b3 f5g6 g1h1 g6h5 e1e8 a8e8 a1e1 e8e1 f3e1 d7f8 f4e5');
INSERT INTO evaluations VALUES(41679,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41679,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41679,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41679,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41679,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41679,36,'cp 0','d6e7 d1a4 c8d7 a4c2 e8g8 c1e3 b8c6 b1c3 d8c8 h2h3 d7f5 c2d1 b7b6 d1a4 f5d7 f1b5 c8b7 b5a6 b7b8 a6b5');
INSERT INTO evaluations VALUES(41056,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41056,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41056,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41056,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41056,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41056,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41056,36,'cp 18','d1a4 c8d7 a4c2 e8g8 c1e3 f8e8 f1e2 g7g6 b1c3 c7c6 e1g1 d7f5 e2d3 f5d3 c2d3 b8d7 b2b4 a7a6 a2a4 f6e4 h2h3 e7f6 a1b1 d7f8 b1b3 e4g5 e3g5 f6g5 b4b5 a6b5 a4b5 g5f4 b5c6 b7c6 b3b6 d8c7 f1b1 f8e6 c3e2');
INSERT INTO evaluations VALUES(32131,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(32131,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(32131,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(32131,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(32131,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(32131,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(32131,36,'cp 78','d2d4 b8d7 c1e3 f8e7 h2h3 g4f3 d1f3 e7f6 d4d5 f6g5 e3g5 d8g5 h3h4 g5h6 g2g4 h6f4 f3f4 e5f4 f1b5 h7h6 b5d7 e8d7 f2f3 h6h5 g4g5 g8e7 c3e2 e7g6 b2b3 a8e8 e1f2 c7c6 a1d1 a7a6 a2a4 c6d5 d1d5');
INSERT INTO evaluations VALUES(32115,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(32115,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(32115,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(32115,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(32115,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(32115,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(32115,36,'cp 124','b8d7 c1e3 f8e7 h2h3 g4f3 d1f3 e7f6 e1c1 e5d4 e3d4 f6d4 d1d4 d8g5 c1b1 g8f6 h3h4 g5c5 f3e3 e8g8 f2f4 c5b6 f1e2 d7c5 e3d2 c5e6 d4d3 e6c5 d3e3 h7h5 e2f3 a8e8 h1e1 e8e6 e4e5 d6e5 f4e5 f6g4 f3g4 h5g4');
INSERT INTO evaluations VALUES(27201,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(27201,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(27201,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(27201,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(27201,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(27201,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(27201,36,'cp 110','b5c6 d7c6 b1c3 e5d4 f3d4 g8f6 d1e2 f8e7 f2f3 f6d7 c1e3 e7h4 g2g3 h4f6 e1c1 d8e7 h1e1 f6d4 e3d4 e8g8 g3g4 b7b5 b2b3 d7c5 c3b5 c5b3 a2b3 e7g5 c1b2 g5b5 e2b5 c6b5 h2h4 f8e8 d4c3 b5c6 h4h5 h7h6');
INSERT INTO evaluations VALUES(27160,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(27160,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(27160,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(27160,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(27160,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(27160,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(27160,36,'cp 108','d7c6 b1c3 e5d4 f3d4 f8e7 d4c6 b7c6 b2b3 g8f6 c1b2 e8g8 d1d3 f6d7 f2f4 d7c5 d3f3 a7a5 a2a4 d8d7 e1c1 f7f5 h1e1 f5e4 c3e4 c5e4 f3e4 a8e8 g2g3 d6d5 e4d3 e7b4 e1e8 f8e8 b2c3 d7d6 c3e5 d6e7 d3a6 e7d7 c1b2');
INSERT INTO evaluations VALUES(26769,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(26769,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(26769,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(26769,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(26769,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(26769,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(26769,36,'cp 94','b1c3 e5d4 f3d4 g7g6 c1e3 f8g7 d1d2 g8f6 f2f3 f6d7 e1c1 d7b6 e3g5 d8d7 b2b3 h7h6 g5e3 d7e7 c3b5 e7d7 a2a4 a7a6 b5c3 d7e7 c1b1 b6d7 h1e1 e8c8 c3a2 e7e5 a2b4 d7b8 c2c3 e5h2 d4c6 b8c6 b4c6 b7c6 e3d4 g7d4 c3d4 c8d7');
INSERT INTO evaluations VALUES(26768,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(26768,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(26768,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(26768,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(26768,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(26768,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(26768,36,'cp 109','e5d4 f3d4 g7g6 c1e3 f8g7 d1d2 g8f6 f2f3 f6d7 e3g5 g7f6 h2h4 h7h6 g5f6 d8f6 e1c1 e8c8 d2e2 d7e5 g2g3 c6d7 c3d5 f6g7 f3f4 e5g4 d4c6 d7c6 d5e7 c8d7 e7c6 d7c6 e2g4 c6b6 g4e2 c7c6 d1d3 b6c7 h1d1 h8e8 e4e5 d6e5 d3d8 e8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(26781,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(26781,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(26781,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(26781,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(26781,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(26781,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(26781,36,'cp 100','f3d4 g7g6 c1e3 f8g7 d1d2 g8f6 f2f3 f6d7 e3g5 g7f6 h2h4 h7h6 g5f6 d8f6 a2a4 d7c5 b2b4 c5e6 a1d1 e6d4 d2d4 f6d4 d1d4 c6d7 e1f2 e8c8 a4a5 d7e6 h4h5 g6g5 g2g4 c7c5 b4c5 d6c5 d4d8 h8d8 f2e3 d8d7 c3b5 a7a6 b5c3');
INSERT INTO evaluations VALUES(26775,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(26775,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(26775,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(26775,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(26775,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(26775,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(26775,36,'cp 85','g7g6 c1e3 f8g7 d1d2 g8f6 f2f3 f6d7 e3g5 g7f6 h2h4 h7h6 g5f6 d8f6 e1c1 e8c8 d2e2 c8b8 d4c6 b7c6 e2a6 d7b6 a2a4 f6e6 b2b3 e6c8 a6d3 b6d7 c3e2 h6h5 d3c3 c8b7 d1d4 c6c5 d4d2 h8e8 e2f4 b7b4 c3b4 c5b4 d2d5 e8e5 d5e5 d7e5 f4d5 a7a5');
INSERT INTO evaluations VALUES(50087,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(50087,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(50087,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(50087,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(50087,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(50087,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(50087,36,'cp 53','g2g3 e7e6 f1g2 d7d5 g1h3 f8d6 e1g1 e8g8 b2b3 a7a5 h3f4 c7c6 c1a3 d5c4 b3c4 e6e5 c4c5 d6c7 d1b3 g8h8 d4e5 c7e5 b1c3 e5f4 g3f4 d8e7 a1d1 c8e6 b3b2 b8a6 d1d6 e7f7 f1a1 h7h6 a1b1 a6b4 a3b4 a5b4 b2b4');
INSERT INTO evaluations VALUES(50086,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(50086,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(50086,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(50086,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(50086,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(50086,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(50086,36,'cp 50','e7e6 f1g2 d7d5 g1f3 f6e4 e1g1 f8e7 d1c2 b8d7 b1c3 c7c6 b2b3 a7a5 f3d2 e4d2 c1d2 e8g8 c3a4 e7d6 a4b2 d8e7 a1c1 d6a3 c1d1 b7b6 d2c1 c8a6 b2d3 a3d6 c1f4 d6a3 d3e5 d7e5 f4e5 f8c8 e2e3');
INSERT INTO evaluations VALUES(55973,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(55973,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(55973,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55973,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(55973,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(55973,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(55973,36,'cp 17','d2d4 c7c5 c2c3 c8d7 a2a3 b8c6 g1f3 d8a5 b1d2 c5d4 d2b3 a5b6 b3d4 c6d4 d1d4 b6d4 c3d4 g8e7 c1d2 d7a4 f1d3 h7h6 e1e2 a7a6 h2h4 e7c6 a1c1 f8e7 g2g4 e8d7 h4h5 a4b5 e2e3 b5d3');
INSERT INTO evaluations VALUES(51972,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(51972,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(51972,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(51972,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(51972,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(51972,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(51972,36,'cp 9','d7d5 e4e5 c8d7 f1e2 b8c6 g1f3 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1e3 f5e3 f2e3 f8e7 e2d3 e8g8 a1c1 a8c8 d1d2 f7f6 e5f6 e7f6 d3b1 d7e8 c3e2 d8e7 e2f4 c8d8 d2c2 g7g6 c2c5 f8f7 c5e7 f7e7');
INSERT INTO evaluations VALUES(54234,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(54234,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(54234,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54234,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(54234,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(54234,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54234,36,'cp 18','e4d5 e6d5 g1f3 g8f6 f1d3 c5c4 d3c2 c8g4 e1g1 f8e7 b2b3 c4b3 a2b3 e8g8 f1e1 f8e8 c1f4 e7d6 e1e8 f6e8 f4g3 g7g6 b1d2 d6g3 h2g3 d8c7 d1e1 b8d7 c3c4 e8f6 c2d3 a8e8 d3e2 a7a6 d2f1 g8g7 f1e3');
INSERT INTO evaluations VALUES(54129,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54129,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(54129,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54129,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(54129,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(54129,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(54129,36,'cp 25','e6d5 f1b5 c8d7 b5d7 b8d7 g1f3 g8f6 e1g1 f8e7 d4c5 d7c5 c1e3 e8g8 b1d2 f8e8 h2h3 a7a6 d1c2 d8c7 e3d4 c5d7 d2b3 e7d6 f1e1 h7h6 a1d1 e8e1 d1e1 a8e8 e1e8 f6e8 g2g3 d7e5 f3e5 d6e5 d4e5 c7e5 b3d4 e5e1 g1g2');
INSERT INTO evaluations VALUES(54518,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(54518,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54518,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54518,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(54518,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54518,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(54518,36,'cp 23','g1f3 g8f6 f1b5 c8d7 b5d7 b8d7 e1g1 f8e7 d4c5 d7c5 c1e3 e8g8 b1d2 f8e8 h2h3 h7h6 d1c2 e7f8 f1e1 a7a5 d2b3 b7b6 a1d1 d8c7 e3d4 c5e4 b3d2 f8c5 e1e2 a8d8 d1e1 c7c6 a2a4 c5d6 d4f6 e4f6');
INSERT INTO evaluations VALUES(54517,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(54517,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(54517,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(54517,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(54517,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54517,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54517,36,'cp 22','g8f6 f1b5 c8d7 b5d7 b8d7 e1g1 f8e7 d4c5 d7c5 c1e3 e8g8 h2h3 f8e8 d1c2 c5e6 b1d2 a7a5 f1e1 a5a4 a2a3 d8c7 f3d4 e7f8 a1d1 e6c5 d4f5 c7c6 d2f3 h7h6 f5d4 c6d7 c2f5 a8c8 f5d7 f6d7 g1f1 d7b6 d4b5 b6c4 e3c5 e8e1 f3e1 f8c5');
INSERT INTO evaluations VALUES(46430,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46430,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46430,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46430,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(46430,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(46430,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(46430,36,'cp 25','f1d3 f8e7 d4c5 e7c5 e1g1 e8g8 c1f4 c8g4 b1d2 b8c6 d1c2 a8c8 h2h3 g4h5 c2b3 d8b6 f4g5 h5f3 d2f3 f6e4 b3b6 c5b6 g5f4 g7g6 d3c2 f8d8 a1e1 e4c5 f4g5 d8e8 g5e3 c5e6 c2b3 d5d4 b3e6 d4e3 e6c8 e3f2 f1f2 b6f2 g1f2');
INSERT INTO evaluations VALUES(19555,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(19555,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(19555,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(19555,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(19555,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(19555,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(19555,36,'cp 25','c3d4 d7d5 e4d5 d8d5 g1f3 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 f1e2 e8g8 e1g1 f8e8 f1e1 c8f5 h2h3 d5c6 d1d6 e7d5 d6c6 b7c6 c3d4 d5f4 e2g4 g7g6 g4e2 f4d3 e1d1 h7h5 e2d3 e4d3 d4c3 c6c5 f2f3 a8b8 a1c1');
INSERT INTO evaluations VALUES(19546,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(19546,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(19546,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(19546,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(19546,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(19546,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(19546,36,'cp 27','d7d5 e4d5 d8d5 g1f3 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 f1e2 e8g8 e1g1 f8e8 f1e1 c8d7 e2g4 d7c6 d1e2 d5c5 a1d1 a8d8 d1d8 e8d8 e1d1 d8d1 e2d1 c6d5 a2a3 e7g6 d1d4 c5d4 c3d4 a7a6 g2g3 g6e7 g1f1 d5e6 g4d1');
INSERT INTO evaluations VALUES(19869,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(19869,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(19869,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(19869,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(19869,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(19869,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(19869,36,'cp 23','e4d5 d8d5 g1f3 e7e5 b1c3 f8b4 f1e2 e5d4 e1g1 d5d8 c3b5 g8e7 f3d4 e8g8 c1f4 c6d4 d1d4 d8d4 b5d4 e7f5 d4f5 c8f5 e2f3 a8d8 f3b7 b4d2 f4d2 d8d2 b2b3 f8d8 b7f3 d2b2 f1c1 g7g6 c1c7 d8d2 c7a7 g8g7 a2a4 b2b3 a4a5 d2a2');
INSERT INTO evaluations VALUES(19832,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(19832,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19832,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(19832,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19832,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(19832,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19832,36,'cp 32','d8d5 g1f3 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 f1e2 e8g8 d1a4 d5c6 a4c6 b7c6 e1g1 e7d5 f1c1 d5c3 c1c3 c8e6 c3c6 f8c8 e2b5 a8b8 c6c8 b8c8 b2b4 g8f8 a2a4 f8e7 a1d1 g7g6 g2g3 c8c3 d1d4 c3b3 d4e4 b3b1 g1g2 e6d5');
INSERT INTO evaluations VALUES(1406,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(1406,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(1406,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(1406,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(1406,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(1406,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(1406,36,'cp 42','g1f3 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 f1e2 e8g8 d1a4 d5c6 a4c6 b7c6 e1g1 a7a5 e2d1 e7d5 f1e1 e4e3 f2f3 c8e6 c3d4 a5a4 a2a3 f8d8 d4e3 d5e3 e1e3 g7g6 e3e2 d8d4 d1c2 c6c5 a1c1 a8b8 c2e4 e6c4');
INSERT INTO evaluations VALUES(1405,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(1405,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(1405,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(1405,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(1405,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(1405,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(1405,36,'cp 41','e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1a4 c8d7 a4a3 d5e6 a3b4 e6c6 a1d1 a7a5 b4d6 e7f5 d6c6 d7c6 f1e2 e8g8 e1g1 b7b5 a2a3 f8b8 e2g4 b5b4 c3d2 f5e7 d2e3 b4a3 b2a3 e7g6 e5e6 g6e5 e6f7 g8f7 g4e2 c6b5 e2b5');
INSERT INTO evaluations VALUES(1564,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(1564,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(1564,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(1564,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(1564,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(1564,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(1564,36,'cp 31','b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1a4 c8d7 a4a3 d5e6 a3b4 e7d5 b4b7 d7c6 f1b5 a8c8 b5c6 c8c6 b7a7 e8g8 a7d4 f8c8 a1d1 c6c5 b2b4 d5c3 d4d8 e6e8 d8e8 c8e8 b4c5 c3d1 e1d1 e8e5 d1d2 e5d5 d2e3 d5c5 e3e4 c5c2 h1b1 h7h6 b1b8 g8h7');
INSERT INTO evaluations VALUES(18679,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(18679,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(18679,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18679,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18679,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(18679,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(18679,36,'cp 0','b7c6 f1d3 f8c5 b1c3 g8f6 e1g1 a7a5 d1e2 e8g8 c3a4 c5a7 c1e3 d7d5 e3a7 a8a7 f1e1 f8e8 a1d1 h7h6 c2c3 a7a8 b2b3 c8b7 a4c5 d8b6 c5a4');
INSERT INTO evaluations VALUES(18405,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18405,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18405,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18405,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18405,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18405,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18405,36,'cp 0','f1d3 f8c5 b1c3 g8f6 e1g1 a7a5 d1e2 e8g8 c3a4 c5a7 c1e3 d7d5 e3a7 a8a7 f1e1 f8e8 a1d1 h7h6 c2c3 a7a8 b2b3 c8b7 a4c5 d8b6 c5a4');
INSERT INTO evaluations VALUES(18402,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18402,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18402,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18402,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18402,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18402,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18402,36,'cp 0','f8c5 b1c3 g8f6 e1g1 a7a5 d1e2 e8g8 c3a4 c5a7 c1e3 d7d5 e3a7 a8a7 f1e1 f8e8 a1d1 h7h6 c2c3 a7a8 b2b3 c8b7 a4c5 d8b6 c5a4');
INSERT INTO evaluations VALUES(39539,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(39539,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(39539,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(39539,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(39539,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(39539,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(39539,36,'cp -17','d7d6 e5d6 c7d6 g1f3 g8f6 f1e2 b8c6 c2c3 e8g8 d4d5 c6e5 e1g1 e7e6 d5e6 c8e6 h2h3 a7a6 c1f4 f8e8 b1d2 d8b6 d1c2 e6f5 c2b3 b6c7 f1e1 h7h6 a2a4 f5d3 b3d1 a8c8 f3d4 d3e2 e1e2');
INSERT INTO evaluations VALUES(38342,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(38342,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(38342,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(38342,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(38342,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(38342,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(38342,36,'cp -24','e5d6 d8d6 g1f3 c7c5 d4c5 d6c5 b1c3 a7a6 c3e4 c5c7 c2c3 b8c6 f1d3 c8e6 e1g1 g8f6 e4f6 e7f6 d1a4 e8g8 c1e3 c6e5 f3e5 f6e5 a4b4 b7b5 f1d1 f8d8 b4h4 f7f5 e3g5 d8d7 d3e2 d7d1 e2d1 a8e8 a2a4 e6c4 a4b5 a6b5 d1c2');
INSERT INTO evaluations VALUES(38331,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(38331,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(38331,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(38331,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(38331,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(38331,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(38331,36,'cp -24','d8d6 g1f3 c7c5 d4c5 d6c5 b1c3 a7a6 c3e4 c5c7 c2c3 b8c6 f1d3 c8e6 e1g1 g8f6 e4f6 e7f6 d1a4 e8g8 f1d1 b7b5 a4h4 a8d8 c1e3 f8e8 d3f1 d8d1 a1d1 c6e7 h4g3 c7b7 f3d4 e6a2 b2b3 e7d5 e3d2 f6f5 g3d6 g7e5');
INSERT INTO evaluations VALUES(33058,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(33058,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(33058,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(33058,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(33058,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(33058,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(33058,36,'cp -21','g1f3 c7c5 d4c5 d6c5 b1c3 a7a6 f1e2 g8f6 e1g1 c5a5 d1d4 b8c6 d4a4 c8e6 a4a5 c6a5 f1d1 e8g8 f3d4 e6d7 c1g5 f8d8 a1c1 e7e6 d4b3 a5b3 a2b3 d7c6 d1d8 a8d8 g5e3 d8c8 b3b4 b7b5 c1d1 f6d5 c3d5 c6d5');
INSERT INTO evaluations VALUES(33057,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(33057,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(33057,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(33057,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(33057,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(33057,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(33057,36,'cp -24','c7c5 d4c5 d6c5 b1c3 g7c3 b2c3 c5c3 c1d2 c3c5 f1d3 g8f6 e1g1 e8g8 f1e1 b8c6 a1b1 b7b6 d1c1 f6g4 e1e2 c8f5 h2h3 f5d3 c2d3 g4f6 c1c5 b6c5 b1c1 a8c8 d2e3 f6d7 e3c5 c6b4 c5e3 c8c1 e3c1 b4d5 e2c2 f7f6 g1f1 f8b8');
INSERT INTO evaluations VALUES(55020,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55020,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(55020,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55020,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(55020,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(55020,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(55020,36,'cp -11','f7f6 g5d2 b8c6 c2c4 c6d4 e2e3 d4c6 d1h5 g7g6 h5d5 c8d7 f1e2 f6f5 d5d3 g8f6 b1c3 e7e5 a1d1 e5e4 d3c2 c6b4 c2b3 a7a5 c3a4 d7a4 b3a4 d8d7 a4d7 f6d7 a2a3 b4a6 g1h3 f8g7 b2b4 a5b4 a3b4 c7c5 b4c5 a6c5 e1g1 g7f6 h3f4 e8g8');
INSERT INTO evaluations VALUES(54699,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(54699,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(54699,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(54699,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(54699,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(54699,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(54699,36,'cp -16','g5d2 b8c6 c2c4 c6d4 d2c3 e7e5 e2e3 d4f5 c4d5 g8e7 d1a4 c8d7 a4b3 c7c6 d5c6 d7c6 b1a3 d8b6 f1b5 c6b5 b3b5 b6c6 e1c1 a8c8 c1b1 c6b5 a3b5 e7c6 c3e1 h7h5 h2h4 f8e7 g1e2 c8d8 d1d8 e8d8 e3e4 a7a6 b5c3 f5d4 f2f3 d8d7 e2g3 g7g6 e1f2 h8d8');
INSERT INTO evaluations VALUES(10065,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(10065,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(10065,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(10065,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(10065,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(10065,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(10065,36,'cp -51','e8g8 a2a4 c6e7 c4b3 e7g6 g5f3 c7c6 e1g1 c5b6 c1e3 d7d5 e3b6 a7b6 e4d5 f6d5 f1e1 c8g4 b3d5 d8d5 b1d2 f7f6 h2h3 g4e6 d2e4 c6c5 f3d2 d5c6 e4c3 g6h4 f2f3 h4f5 d2c4 c6c7 d1e2 f5d4 e2f2 a8d8 c3e4');
INSERT INTO evaluations VALUES(4304,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(4304,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(4304,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(4304,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(4304,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(4304,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(4304,36,'cp -46','a2a4 c6e7 c4b3 c7c6 g5f3 e7g6 e1g1 c5b6 b1c3 d7d6 h2h3 h7h6 c1e3 b6a5 f3d2 f8e8 d3d4 e5d4 e3d4 a5c3 d4c3 f6e4 d2e4 e8e4 f1e1 e4e1 d1e1 d6d5 e1e3 c8e6 a4a5 d8g5 c3d2 g5e5 e3e5 g6e5 d2e3 f7f6 f2f3 a7a6 e3d4 a8e8 g1f2 e6f7 a1e1 g7g5');
INSERT INTO evaluations VALUES(50404,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(50404,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(50404,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(50404,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50404,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(50404,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(50404,36,'cp 49','d5b6 d2d4 d7d6 e5d6 e7d6 b1c3 f8e7 f1e2 e8g8 g1f3 c8e6 b2b3 d6d5 c4c5 b6d7 b3b4 c7c6 e1g1 e7f6 c1f4 a7a5 a2a3 a5b4 a3b4 a8a1 d1a1 e6g4 f1d1 f8e8 h2h3 g4f3 e2f3 b8a6 b4b5 a6c7 f3g4 c7e6 f4e3');
INSERT INTO evaluations VALUES(50301,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(50301,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(50301,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(50301,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(50301,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(50301,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(50301,36,'cp 72','d2d4 d7d6 e5d6 e7d6 b1c3 f8e7 f1d3 e8g8 g1f3 c8g4 e1g1 b8c6 c1e3 e7f6 b2b3 f8e8 d3e4 g4h5 h2h3 h5g6 e4g6 h7g6 g2g4 d8d7 f1e1 d6d5 c4c5 b6c8 b3b4 a7a6 a2a3 c8a7 g4g5 f6d8 g1g2 a7b5');
INSERT INTO evaluations VALUES(50302,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(50302,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(50302,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(50302,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(50302,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(50302,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(50302,36,'cp 64','d7d6 e5d6 e7d6 g1f3 f8e7 h2h3 c8f5 f1e2 e8g8 e1g1 b8c6 b1c3 e7f6 c1f4 f8e8 b2b3 a7a5 d1d2 h7h6 a1e1 a5a4 d4d5 c6e5 f3d4 f5g6 f4e3 e5d7 e2d3 g6d3 d2d3 d7e5 d3c2 a4b3 a2b3 e5g6 c2d2 f6d4 e3d4 b6d7 f2f4');
INSERT INTO evaluations VALUES(15435,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(15435,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(15435,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(15435,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(15435,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(15435,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(15435,36,'cp 64','d2d4 c8d7 b1c3 f8e7 d4d5 c6b8 a2a4 e8g8 b2b3 d7g4 b5e2 b8d7 a4a5 a7a6 f3d2 g4e2 d1e2 g7g6 d2c4 f6h5 c1h6 f8e8 f2f3 e7g5 h6g5 d8g5 e2d2 g5d2 c4d2 f7f5 a1a4 b7b5 a5b6 c7b6 a4c4 a8c8 c4c6 d7b8 c6c8 e8c8 d2c4 h5f4 e4f5');
INSERT INTO evaluations VALUES(15420,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15420,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(15420,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(15420,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(15420,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(15420,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(15420,36,'cp 59','c8d7 b1c3 f8e7 d4d5 c6b8 a2a4 e8g8 b2b3 d7g4 f1e1 a7a6 b5e2 b8d7 a4a5 g4h5 f3d2 h5e2 d1e2 g7g6 c3a4 f6h5 d2f3 h5g7 c2c4 f7f5 b3b4 d7f6 a4c3 f6e4 c3e4 f5e4 e2e4 g7f5 c1d2 e7f6 a1d1 g8h8 e4d3 d8d7');
INSERT INTO evaluations VALUES(25499,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(25499,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(25499,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(25499,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(25499,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(25499,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(25499,36,'cp 73','b1c3 f8e7 b5c6 d7c6 f1e1 h7h6 h2h3 e5d4 f3d4 e8g8 d4c6 b7c6 f2f4 a7a5 d1f3 f8e8 c1d2 a5a4 a1d1 f6d7 d2e3 d8b8 e4e5 d7f8 f3c6 e7h4 e1f1 b8b2 c3d5 d6e5 f4f5 b2a2 f5f6 g7g6 e3h6 a4a3 h6c1');
INSERT INTO evaluations VALUES(25495,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(25495,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(25495,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(25495,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25495,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(25495,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(25495,36,'cp 52','f8e7 b5c6 d7c6 f1e1 e5d4 f3d4 e8g8 d4f5 f6d7 c3d5 e7f6 d5f6 d8f6 c1d2 f8e8 d2c3 f6g6 f2f3 d7e5 f5e3 a8d8 d1e2 f7f6 b2b3 g6f7 a1d1 e5g6 e2f2 a7a6 f2g3 g6e7 c3b2 b7b5 c2c4 b5c4 b3c4 e7g6 e3d5 g8h8 d1c1 d8b8 b2c3 g6e5 c3e5 d6e5 h2h4 c6d5 e4d5');
INSERT INTO evaluations VALUES(24190,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(24190,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(24190,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(24190,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(24190,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(24190,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(24190,36,'cp 51','d4d5 c6b8 a2a4 e8g8 d1e2 h7h6 c1e3 f6g4 e3d2 g4f6 f1d1 a7a5 b5d7 b8d7 d2e3 f6g4 e3c1 g4f6 b2b3 f6h7 c1e3 f7f5 e4f5 f8f5 d1f1 f5f8 e2b5 b7b6 b5c4 a8c8 c3b5 e7g5 b5a7 c8a8 f3g5 h6g5 a7b5 a8c8 f2f3 d7f6 h2h3');
INSERT INTO evaluations VALUES(31831,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(31831,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31831,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31831,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(31831,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(31831,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(31831,36,'cp 9','e7e6 e2e3 f8d6 f4d6 c7d6 c2c4 d5c4 f1c4 g8f6 d1b3 d8e7 c4b5 b8d7 e1g1 e8g8 f1c1 f8c8 b1d2 h7h6 b5e2 d7b6 h2h3 c8c1 a1c1 a8c8 c1c3 c8c7 b3a3 b6c8 a3a5 b7b6 a5a3 c7c3 a3c3 e7d7 g1h1 c8e7 c3a3 d7c7');
INSERT INTO evaluations VALUES(32298,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32298,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(32298,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(32298,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(32298,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(32298,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(32298,36,'cp 17','e2e3 f8d6 f4d6 c7d6 c2c4 d5c4 f1c4 g8f6 d1b3 d8e7 c4b5 b8d7 b1c3 e8g8 e1g1 f6e4 f1c1 d7f6 c3e4 f5e4 f3d2 f8c8 b5f1 e4d5 b3a3 e7d8 c1c3 c8c6 a1c1 c6c3 c1c3 a7a6 d2c4 d5c4 f1c4 d6d5 c4e2');
INSERT INTO evaluations VALUES(32296,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(32296,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(32296,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(32296,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(32296,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(32296,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(32296,36,'cp 12','f8d6 f4d6 c7d6 c2c4 d5c4 f1c4 g8f6 d1b3 d8e7 c4b5 b8d7 e1g1 e8g8 b1c3 f8c8 f1c1 c8c7 f3h4 f5g6 h4g6 h7g6 c3e2 a8c8 c1c7 c8c7 b5d3 d7b6 g2g3 d6d5 a1c1 c7c1 e2c1 f6e4 b3c2 e7b4 c1b3 b6c4 d3e4 d5e4 b3c5');
INSERT INTO evaluations VALUES(28144,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(28144,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(28144,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(28144,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28144,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(28144,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(28144,36,'cp 17','f4d6 c7d6 c2c4 d5c4 f1c4 g8f6 d1b3 d8e7 c4b5 b8d7 e1g1 e8g8 b1c3 f8c8 f1c1 c8c7 f3h4 f5g6 h4g6 h7g6 c3e2 a8c8 b5d3 d6d5 c1c7 c8c7 a1c1 c7c1 e2c1 d7b6 g2g3 f6e4 c1e2 e7c7 d3e4 d5e4 e2c3 c7c6 g1g2 g8h8 h2h3 b6d5 c3d5 c6d5 b3d5 e6d5');
INSERT INTO evaluations VALUES(28134,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(28134,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(28134,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(28134,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28134,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(28134,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28134,36,'cp 8','c7d6 c2c4 d5c4 f1c4 g8f6 d1b3 d8c7 e1g1 e8g8 b1c3 b8c6 c4e2 d6d5 b3a3 f5g4 f1c1 f8c8 f3e1 g4e2 c3e2 c7b6 e1d3 h7h6 c1c2 c6a5 a1c1 c8c2 c1c2 a5c4 a3b3 f6e4 h2h3 a8c8 b3b6 c4b6 c2c8 b6c8 f2f3 e4d6 d3c5');
INSERT INTO evaluations VALUES(50555,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(50555,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(50555,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(50555,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(50555,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(50555,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(50555,36,'cp 28','b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 g8f6 a2a3 d7d6 a3a4 b5a4 c3a4 f8e7 f1d1 d8b8 a4c3 a6a5 d3b5 b8b7 b5c6 b7c6 e2b5 c6b5 c3b5 e8g8 b5d6 e7d6 d1d6 f8c8 e4e5 f6d5 c2c3 a5a4 h2h3 h7h6 f2f3 a4a3 a1a3 a8a3 b2a3 c8c3');
INSERT INTO evaluations VALUES(51186,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51186,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(51186,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(51186,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(51186,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(51186,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(51186,36,'cp 23','f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 g8e7 a2a4 b5b4 c3a2 a6a5 d3b5 d8c7 c1d2 e7g6 c2c3 b4c3 a2c3 f8d6 g2g3 e8g8 f1c1 c7b7 b5c6 d7c6 c3d1 c6c5 d2c3 f8d8 h2h4 h7h6 e4e5 d6c7 h4h5 g6e7 d1e3 g8h8');
INSERT INTO evaluations VALUES(51187,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51187,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(51187,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(51187,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(51187,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(51187,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(51187,36,'cp 36','c8b7 e1g1 b5b4 c3a4 b8c6 d4b3 d7d6 d1e2 f8e7 c1e3 g8f6 f2f4 e8g8 a1d1 d8c7 c2c4 b4c3 a4c3 f8c8 h2h3 h7h6 d1c1 c7d8 f1d1 f6d7 d3c4 e7f8 b3d2 a8b8 d2f3 c6a5 c4d3 b7a8 a2a3 a5b3');
INSERT INTO evaluations VALUES(30323,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(30323,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(30323,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(30323,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(30323,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(30323,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(30323,36,'cp 31','e1g1 b8c6 d4c6 b7c6 f1e1 g8e7 d3f1 e7g6 c1e3 f7f6 c3d5 e8f7 c2c4 b5c4 d5b6 f8b4 e1e2 a8b8 b6c4 c6e4 e3a7 b8b7 a7b6 b7b6 c4b6 d7d5 b6c4 g6f4 e2e4 d5e4 d1a4 f4d5 a4a6 d8b8 a2a4 f6f5 a4a5 b4c5 a6c6 h8c8 c6d7 f7f6 a5a6');
INSERT INTO evaluations VALUES(30322,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(30322,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(30322,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(30322,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30322,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(30322,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(30322,36,'cp 40','b8c6 d4c6 b7c6 f1e1 g8e7 d3f1 e7g6 c1e3 f7f6 c3d5 e8f7 c2c4 b5c4 d5b6 f8b4 b6a8 b4e1 a8b6 e1b4 b6c4 c6e4 e3b6 d8e8 d1a4 e8e7 a2a3 b4d6 a4a6 d6f4 b2b4 h8a8 a6b5 a8b8 b5h5 d7d5 b6c5 e7c7 a1e1 f7g8 e1e4 d5e4 g2g3 g6e5');
INSERT INTO evaluations VALUES(56412,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(56412,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(56412,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(56412,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(56412,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(56412,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(56412,36,'cp -30','e4d5 e5f4 g1f3 g8f6 c2c4 c7c6 d2d4 f8b4 b1c3 c6d5 f1e2 b8c6 e1g1 c8e6 d1b3 d5c4 e2c4 e6c4 b3c4 e8g8 c1f4 a8c8 c4b5 d8b6 b5b6 a7b6 c3b5 f6d5 f4g3 f8d8 a2a3 b4f8 a1b1 f7f6 f3d2 d8e8 f1e1 g8f7 e1e8 c8e8 d2c4');
INSERT INTO evaluations VALUES(56159,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(56159,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(56159,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(56159,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(56159,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(56159,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(56159,36,'cp -18','e5f4 g1f3 g8f6 c2c4 c7c6 d2d4 f8b4 b1c3 c6d5 f1e2 b8c6 e1g1 c8e6 d1b3 d5c4 e2c4 e6c4 b3c4 e8g8 c1f4 a8c8 c4b5 b4a5 g1h1 a5c3 b2c3 c6a5 b5d3 a5c4 a1e1 b7b5 f3e5 f6d5 d3g3 d5f4 f1f4 f7f6');
INSERT INTO evaluations VALUES(56149,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(56149,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56149,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(56149,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(56149,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(56149,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(56149,36,'cp -11','g1f3 g8f6 c2c4 c7c6 d2d4 f8b4 b1c3 c6d5 f1d3 d8e7 f3e5 c8g4 d1a4 g4d7 a4d1 d5c4 d3c4 b8c6 c1f4 c6e5 f4e5 a8c8 d1e2 b4c3 b2c3 f6g4 e5d6 e7e3 e2e3 g4e3 c4b3 d7e6 e1f2 c8c3 h1e1 e8d7 d6b4 c3b3 a2b3 e3c2 a1a7 c2e1 f2e1 e6b3 a7b7 d7c6 b7e7');
INSERT INTO evaluations VALUES(56148,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(56148,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(56148,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(56148,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(56148,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(56148,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(56148,36,'cp -48','g8f6 c2c4 c7c6 d2d4 f8b4 b1c3 c6d5 f1d3 d8e7 f3e5 c8g4 d1a4 g4d7 a4d1 b8c6 e1g1 c6e5 d4e5 e7e5 g1h1 b4c3 c1f4 e5e6 b2c3 e8g8 f4g5 d5c4 d3f5 e6c6 g5f6 d7f5 f1f5 g7f6 d1g4 g8h8 g4h4 f8g8 h4f6 c6f6 f5f6 g8f8 h2h3 h8g7 f6f5 a8c8 a1f1 c8c7 f5g5 g7h6');
INSERT INTO evaluations VALUES(48580,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(48580,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(48580,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(48580,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(48580,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(48580,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(48580,36,'cp -25','c2c4 c7c6 d2d4 f8b4 b1c3 c6d5 f1e2 b8c6 c1f4 d5c4 e2c4 d8e7 f4e5 c6e5 f3e5 e8g8 e1g1 c8e6 c4e6 e7e6 d1a4 b4e7 a1e1 e6d6 e1d1 d6b6 e5c4 b6b4 a4b4 e7b4 d1d3 f8d8 f1f5 b4f8 c4e5 f8d6 e5c4 a7a6 c4b6 a8b8 c3d5 f6d5 f5d5 g7g6');
INSERT INTO evaluations VALUES(48566,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(48566,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(48566,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(48566,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(48566,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(48566,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(48566,36,'cp -36','c7c6 d2d4 f8b4 b1c3 c6d5 f1e2 b8c6 c1f4 d5c4 e2c4 d8e7 f4e5 c6e5 f3e5 e8g8 e1g1 c8e6 c4e6 e7e6 d1a4 e6b6 a1d1 b4e7 e5c4 b6b4 a4b4 e7b4 c4e3 a8c8 c3d5 f6d5 e3d5 b4d6 f1e1 f7f6 d5e7 d6e7 e1e7 f8f7 e7e2 f7d7 d4d5 g8f7 e2d2 b7b6 d5d6 c8c5 g1f2 c5a5 a2a3');
INSERT INTO evaluations VALUES(45725,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(45725,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(45725,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(45725,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(45725,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45725,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(45725,36,'cp -12','d2d4 f8b4 b1c3 c6d5 f1d3 d8e7 f3e5 b8c6 e1g1 d5c4 d3c4 e8g8 e5c6 b7c6 c1f4 c8e6 f1e1 a8d8 c4e6 f7e6 d1b3 d8d4 e1e6 e7f7 c3e2 d4e4 e6e4 f7b3 a2b3 f6e4 f4e3 f8d8 e2d4 c6c5 d4c6 d8d3 c6b4 d3e3 b4a6 e3e2 a1c1 e2b2 a6c5 e4d6 c1e1 g8f7 c5e4 d6e4 e1e4');
INSERT INTO evaluations VALUES(49444,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(49444,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(49444,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(49444,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(49444,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(49444,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(49444,36,'cp 122','e4e5 f6d5 c2c4 d5e7 b1c3 d7d6 e5d6 c7d6 g1f3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 d4e5 d6e5 d2d8 f8d8 f3e5 b8c6 e5c6 b7c6 g4f5 d8e8 e1d2 c6b5 f1g2 c8f5 g2a8 e8a8 c4b5 a8d8 d2e1 d8d4 f4d2 d4e4 e1f1 g7b2 a1b1 b2f6 b1b4 e4b4 d2b4 f5d3 f1g2 d3b5 f2f3 b5c4');
INSERT INTO evaluations VALUES(49349,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(49349,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(49349,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(49349,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(49349,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(49349,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(49349,36,'cp 96','f6d5 c2c4 d5e7 b1c3 d7d6 e5d6 c7d6 g1f3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 d4e5 d6e5 d2d8 f8d8 f3e5 b8c6 e5c6 b7c6 g4f5 d8e8 e1d2 c6b5 f1g2 c8f5 g2a8 e8a8 c4b5 g7b2 a1d1 a8c8 d2e2 f5e6 f4h6 e6a2 h4h5 a2c4 e2e3 c8e8 e3f4 c4b5 f4g3 b5e2 d1d2 b2c3 d2c2 c3b4 h5g6 f7g6');
INSERT INTO evaluations VALUES(49326,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(49326,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(49326,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(49326,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(49326,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(49326,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(49326,36,'cp 123','c2c4 d5e7 b1c3 d7d6 e5d6 c7d6 g1f3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 e1c1 a7a6 b5c3 h6h5 g2g4 h5g4 f3h2 b8c6 h2g4 e6e5 d4e5 d6e5 d2d8 f8d8 d1d8 c6d8 f4e5 d8c6 e5g7 g8g7 h4h5 c8e6 h5g6 f7g6 c3e4 a8f8 e4g5 e6g8 f1d3 f5d4 d3e4 c6b4 c1b1 b7b5 h1d1');
INSERT INTO evaluations VALUES(49323,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(49323,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(49323,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(49323,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(49323,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(49323,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(49323,36,'cp 112','d5e7 g1f3 d7d6 e5d6 c7d6 b1c3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 e1c1 h6h5 g2g4 h5g4 f3h2 b8c6 h2g4 a7a6 b5c3 e6e5 d4e5 d6e5 d2d8 f8d8 d1d8 c6d8 f4e5 d8c6 e5g7 g8g7 h4h5 c8e6 h5g6 f7g6 c3e4 f5g3 f2g3 e6g4 f1g2 c6e5 e4f2 a8c8 b2b3 b7b5 g2b7 c8c7 b7a6 b5c4 a6c4 g4d7 h1h4 e5c4 b3c4');
INSERT INTO evaluations VALUES(50048,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(50048,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(50048,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(50048,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(50048,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(50048,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(50048,36,'cp 115','b1c3 d7d6 e5d6 c7d6 g1f3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 e1c1 a7a6 b5c3 h6h5 g2g4 h5g4 f3h2 b8c6 h2g4 e6e5 d4e5 d6e5 d2d8 f8d8 d1d8 c6d8 f4e5 d8c6 e5g7 g8g7 h4h5 f5g3 f2g3 c8g4 h5h6 g7h8 f1g2 a8d8 c3d5 c6e5 d5f6 g4e6 b2b3 b7b5 c4b5 a6b5 h1d1 e5d3 c1b1 e6f5 g2e4');
INSERT INTO evaluations VALUES(16125,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(16125,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16125,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16125,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(16125,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16125,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(16125,36,'cp 7','f8c5 d2d3 e8g8 a2a3 a7a6 c1e3 d7d6 c3d5 f6d5 c4d5 c6e7 d5a2 c5e3 f2e3 e7g6 f3d2 d8g5 d1e2 c8g4 e2f2 g5e7 h2h3 g4e6 a2e6 f7e6 f2e2 g8h7 d2f3 g6h4 f3h4 e7h4 f1f8 a8f8 a1f1 f8f1 g1f1 h7g6 b2b4 h4f6 f1g1 b7b6');
INSERT INTO evaluations VALUES(9752,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(9752,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(9752,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(9752,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(9752,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(9752,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(9752,36,'cp 5','f1e1 f6g4 e1f1 g4f6 d2d3 e8g8 a2a3 d7d6 c3a4 c5b6 h2h3 c6e7 a4b6 a7b6 f1e1 c8e6 c4e6 f7e6 c2c3 e7g6 d3d4 f6d7 d4e5 d7e5 f3e5 g6e5 f2f4 e5d7 c1e3 f8f7 d1b3 e6e5 f4e5 d7e5 e3d4 d8e7 d4e5 e7e5 e1f1 a8f8 f1f7 f8f7 a1f1');
INSERT INTO evaluations VALUES(9753,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(9753,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(9753,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9753,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9753,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(9753,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(9753,36,'cp 4','a7a5 c3d5 e8g8 c2c3 d7d6 d2d4 c5a7 h2h3 f8e8 d4e5 c6e5 f3e5 d6e5 c1e3 f6e4 e3a7 e4d6 a7c5 d6c4 b2b3 c4b6 d5e7 g8h8 d1d8 e8d8 e1e5 a5a4 e7c8 a8c8 c5b6 c7b6 e5e7 c8c3 b3a4 h8g8 e7b7 c3c2 b7b6 d8d2 a1f1 c2a2 b6a6 d2d4 a4a5');
INSERT INTO evaluations VALUES(53378,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53378,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(53378,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53378,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(53378,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53378,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(53378,36,'cp 0','g8f6 c1f4 b8c6 e2e3 c8f5 f1b5 e7e6 g1f3 f6d7 e1g1 f8e7 a1c1 a8c8 a2a3 g7g5 f4e5 e8g8 e5g3 a7a6 b5d3 f5d3 d1d3 d7b6 c1c2 b6c4 h2h3 b7b5 f1c1 f7f5 d3d1 f5f4 e3f4 g5f4 g3h2 d8d7 c3e2');
INSERT INTO evaluations VALUES(45443,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(45443,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45443,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45443,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45443,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(45443,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(45443,36,'cp 23','c1f4 b8c6 e2e3 c8f5 d1b3 c6a5 f1b5 f5d7 b3c2 e7e6 b5d3 a8c8 g1f3 f8b4 e1g1 e8g8 a2a3 b4c3 b2c3 f6e4 d3e4 d5e4 c2e4 c8c3 f4d6 f8e8 d6b4 c3c8 b4a5 d8a5 e4b7 d7c6 b7b4 a5h5 f3e5 c6g2 g1g2 f7f6 b4a5 h5g5 g2h1 f6e5 a1b1 e8f8 a5e5 g5e5 d4e5');
INSERT INTO evaluations VALUES(45424,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(45424,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(45424,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(45424,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45424,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(45424,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(45424,36,'cp 14','b8c6 e2e3 c8f5 d1b3 c6a5 b3a4 f5d7 a4c2 e7e6 g1f3 a8c8 f1d3 b7b5 a2a3 f8e7 e1g1 e8g8 f3e5 a7a6 f1c1 h7h6 h2h3 d8b6 f4g3 a5c4 c2e2 d7c6 a3a4 c4e5 g3e5 b5a4 d3a6 c8a8 a6b5 c6b5 e2b5 b6b5 c3b5 f6e4 f2f3 e4d2 e5d6 e7d6 b5d6 f8d8 d6b7 d2b3 b7d8');
INSERT INTO evaluations VALUES(13132,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(13132,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13132,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(13132,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(13132,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(13132,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(13132,36,'cp 25','e2e3 c8f5 d1b3 c6a5 f1b5 f5d7 b3c2 e7e6 b5d3 a8c8 g1f3 b7b5 a2a3 f8e7 h2h3 a7a6 e1g1 e8g8 f1c1 h7h6 f3e5 d8b6 b2b3 f8d8 e5d7 f6d7 c2b2 d7f6 c1c2 c8c6 a1c1 d8c8 a3a4 b5b4 c3b1 b6b7 b1d2 f6d7 c2c6 c8c6 c1c6 b7c6 b2c2');
INSERT INTO evaluations VALUES(13129,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(13129,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(13129,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(13129,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(13129,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(13129,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13129,36,'cp 23','c8f5 d1b3 c6a5 b3a4 f5d7 a4c2 a8c8 f1d3 b7b5 a2a3 e7e6 g1f3 f8e7 h2h3 e8g8 c2e2 a5c4 e1g1 c4a3 f3e5 b5b4 e5d7 f6d7 c3b5 a3b5 d3b5 a7a5 b5d7 d8d7 a1a5 e7d6 f4d6 d7d6 a5a7 c8c7 a7c7 d6c7 e2b5 f8b8 b5c5 c7c5 d4c5');
INSERT INTO evaluations VALUES(24618,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24618,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24618,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(24618,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(24618,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24618,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(24618,36,'cp 28','d1b3 c6a5 b3a4 f5d7 a4c2 e7e6 f1d3 a8c8 g1f3 b7b5 a2a3 f8e7 e1g1 a7a6 h2h3 e8g8 f1c1 h7h6 f3e5 d8b6 f4g3 a5c4 c2e2 d7e8 a3a4 f6d7 e5f3 d7f6 a4b5 a6b5 g3h4 b6b7 b2b3 c4d6 e2d1 e8d7 f3e5 b5b4 e5d7 b7d7 c3a4');
INSERT INTO evaluations VALUES(50386,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(50386,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(50386,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(50386,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(50386,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(50386,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(50386,36,'cp 20','c7c5 f1g2 b8c6 c2c4 d7d5 c4d5 f6d5 d2d4 c5d4 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 b1c3 c8f5 g2c6 b7c6 f2f3 e7e5 e2e4 f5e6 c1e3 f8e7 d1e2 g7g5 h1c1 g5g4 c3d1 e6d7 d1f2 h7h5 c1c3 f7f5 a1d1 f5e4 f2e4 g4f3 e2f2 a8b8 b2b3 h8f8 e3a7 b8b7');
INSERT INTO evaluations VALUES(44586,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(44586,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44586,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44586,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(44586,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44586,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44586,36,'cp 9','c2c4 d7d5 c4d5 f6d5 d2d4 c5d4 f1g2 b8c6 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 c1e3 e7e5 g2c6 b7c6 f2f3 f8e7 b1d2 c8e6 d1e1 e8g8 e1f2 a7a5 h1c1 f8b8 b2b3 a5a4 b3a4 a8a4 c1c6 b8b2 d2e4 a4a2 a1a2 b2a2 e3g5 e7a3 g5c1 a3c1 c6c1');
INSERT INTO evaluations VALUES(44535,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(44535,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(44535,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44535,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(44535,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44535,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(44535,36,'cp 10','d7d5 c4d5 f6d5 d2d4 c5d4 f1g2 b8c6 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 b1c3 c8f5 c1d2 e7e5 d1e1 f5e6 g2c6 b7c6 f2f3 f8e7 e1f2 f7f5 b2b3 a8c8 a1c1 e8g8 c3a4 c6c5 a4b2 e5e4 h1d1 e4f3 e2f3 f8d8 d2c3 d8d1 c1d1 c8c6 b2c4 c6a6 d1d2 e6c4 b3c4 g8f7');
INSERT INTO evaluations VALUES(45343,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(45343,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(45343,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(45343,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45343,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(45343,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45343,36,'cp 3','c4d5 f6d5 f1g2 b8c6 d2d4 c5d4 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 b1c3 c8f5 g2c6 b7c6 c1e3 e7e5 d1e1 f8b4 f2f3 b4c3 b2c3 f7f6 e1f2 e8f7 g3g4 f5e6 h1b1 h8b8 a2a4 a7a5 e3c5 b8b3 b1b3 e6b3 c5b6 a8b8 b6a5 b8a8 a5b6 a8a4 a1a4');
INSERT INTO evaluations VALUES(21854,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(21854,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(21854,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21854,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(21854,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(21854,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21854,36,'cp 49','a7a6 b5a4 g8f6 e1g1 f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 d8c8 d2b3 f6d7 b3d4 d7e5 d4e6 f7e6 e1e5 c8d7 a1e1 a8e8 e1d1 f8d6 e5e4 h8f8 g5f4 d7c6 f4e3 b7b6 d1d3 c6b7 h2h3 g7g6 g1f1');
INSERT INTO evaluations VALUES(17779,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(17779,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(17779,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(17779,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(17779,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(17779,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(17779,36,'cp 47','b5a4 g8f6 e1g1 f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 d2c4 d6e5 c4e5 h8e8 e5d3 d8d7 a1d1 a8d8 d3c5 d7c8 d1d8 e8d8 c5e6 f7e6 g1f1 d8d6 g5f4 d6b6 e1e2 b6c6 f4e5 c8d7 g2g3 a6a5 b2b3 a5a4 c3c4 a4b3 a2b3');
INSERT INTO evaluations VALUES(17804,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(17804,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(17804,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(17804,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(17804,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(17804,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(17804,36,'cp 29','g8f6 e1g1 f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6e8 c1g5 h7h6 g5f6 g7f6 e5d3 e8b5 a2a4 b5d3 e1d1 f8c5 d1d3 d8e7 b1d2 h8d8 a1e1 e7f8 d3d8 a8d8 d2b3 d8d5 b3c5 d5c5 e1d1 c5c4 d1d4 c4d4 c3d4 f8e7 g1f1 f6f5 f1e2 e7d6 e2d3 d6e6');
INSERT INTO evaluations VALUES(11426,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(11426,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(11426,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11426,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(11426,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(11426,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11426,36,'cp 43','e1g1 f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6e8 c1g5 h7h6 g5f6 g7f6 e5g4 f8g7 b1d2 e8d7 g4e3 h8e8 a1d1 a6a5 d2b3 b7b6 d1d2 d8c8 e1d1 d7a4 d2d4 a4b5 c3c4 b5c6 d4g4 g7f8 b3d4 c6e4 d4b5 f8c5 g4f4 f6f5 g2g3 c5e3 f2e3 a5a4 f4h4 e8d8 d1d8 c8d8 h4h6');
INSERT INTO evaluations VALUES(53094,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53094,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(53094,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53094,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(53094,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53094,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53094,36,'cp 21','b1c3 b8c6 f1b5 c8d7 g1f3 g7g6 e1g1 g8f6 d2d3 f8g7 b5c4 e8g8 a2a4 a7a6 c1e3 e7e6 h2h3 d8c7 d1e1 c6e7 c4b3 d7c6 a4a5 f6d7 e1f2 a8e8 e3d2 h7h6 g2g4 d6d5 b3a4 c6a4 c3a4 d5e4 d3e4');
INSERT INTO evaluations VALUES(53089,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53089,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(53089,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(53089,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53089,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53089,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53089,36,'cp 0','g7g6 f1b5 c8d7 g1f3 f8g7 e1g1 d7b5 c3b5 g8f6 d2d3 e8g8 h2h3 b8c6 c1e3 d8b6 c2c4 a7a6 b5c3 b6b2 c3a4 b2a3 e3c1 a3b4 c1d2 b4a3');
INSERT INTO evaluations VALUES(52627,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52627,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52627,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52627,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52627,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52627,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52627,36,'cp 0','f1b5 c8d7 g1f3 f8g7 e1g1 d7b5 c3b5 g8f6 d2d3 e8g8 h2h3 b8c6 c1e3 d8b6 c2c4 a7a6 b5c3 b6b2 c3a4 b2a3 e3c1 a3b4 c1d2 b4a3');
INSERT INTO evaluations VALUES(39597,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(39597,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(39597,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(39597,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(39597,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(39597,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(39597,36,'cp 67','e2e4 d7d6 c1e3 g8f6 b1c3 e8g8 d1d2 a7a6 e1c1 b7b5 f1d3 c7c5 d4c5 b5b4 c3d5 f6d5 e4d5 d6c5 e3c5 d8d5 d2b4 d5a2 b4a3 g7h6 f3d2 a2a3 c5a3 a8a7 h2h4 b8c6 h4h5 g6g5 d2b3 c6e5 d3e4 a7c7 a3c5 g5g4 c1b1 e5c4 e4d5 c8e6 h1e1 c4d2 b3d2');
INSERT INTO evaluations VALUES(55078,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(55078,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(55078,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(55078,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(55078,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(55078,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(55078,36,'cp -90','e7e5 e4d5 d8d5 b1c3 d5e6 f1d3 b8c6 b2b3 c8d7 c1b2 e8c8 d1e2 f7f5 e1c1 g8f6 d3c4 e6e8 e2f2 a7a6 a2a4 c8b8 g1e2 c6b4 d2d4 e5e4 d4d5 f8d6 e2d4 e4e3 f2e2 e8e7 c1b1 f5f4 d4e6 d6e5 g2g3 e5c3 b2c3 b4d5 c4d5 d7e6 d5e6 e7e6 g3f4 f6d5');
INSERT INTO evaluations VALUES(54156,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54156,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54156,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(54156,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(54156,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54156,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54156,36,'cp 17','e4d5 e6d5 c4b5 b8c6 d2d4 g8f6 e1g1 c8e6 b1c3 f8d6 d4c5 d6c5 c1g5 e8g8 d1d3 a8c8 a1e1 c6b4 d3d2 a7a6 b5a4 h7h6 g5h4 b4c6 a4b3 c5e7 e1d1 c6a5 f1e1 a5b3 a2b3 f8e8 h2h3 f6e4 c3e4 d5e4 e1e4 d8d2 d1d2');
INSERT INTO evaluations VALUES(54121,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54121,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54121,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(54121,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54121,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54121,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54121,36,'cp 0','e6d5 c4b5 b8c6 e1g1 f8d6 d2d4 g8e7 d4c5 d6c5 b1d2 c8g4 h2h3 g4h5 d2b3 c5d6 f1e1 e8g8 c2c3 f8e8 b5e2 d6c7 b3d4 h7h6 c1e3 h5g6 d1d2 e7f5 d4f5 g6f5 a1d1 f5e4 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(54500,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54500,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54500,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(54500,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(54500,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54500,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(54500,36,'cp 8','c4b5 b8c6 d2d4 f8d6 d4c5 d6c5 e1g1 g8e7 b1d2 e8g8 d2b3 c5d6 h2h3 a7a6 b5d3 c6b4 b3d4 b4d3 d1d3 d8c7 c1e3 f8e8 a1d1 c8d7 f1e1 a8c8 d3b3 h7h6 e3d2 d6c5 d2c3 e7g6 e1e8 c8e8 c3d2 d7c6 d2e3 c5d6 c2c3 g6e5 f3e5 e8e5 d4c6 b7c6');
INSERT INTO evaluations VALUES(54472,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(54472,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54472,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(54472,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(54472,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(54472,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(54472,36,'cp 14','b8c6 d2d4 g8f6 e1g1 f8d6 d4c5 d6c5 b1c3 e8g8 c1g5 c8g4 h2h3 g4h5 b5a4 a8c8 g5f6 d8f6 d1d5 h5f3 d5f3 f6f3 g2f3 c6d4 g1g2 f8d8 a1d1 g8f8 a4b3 d4b3 c2b3 c5d4 f1e1 d4f6 f3f4 g7g6 d1d8 c8d8 e1d1 d8d1 c3d1 f8e7 g2f3 e7e6 f3e4 f6g7 d1c3 a7a6 b3b4 g7f6');
INSERT INTO evaluations VALUES(20482,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(20482,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(20482,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(20482,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(20482,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(20482,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(20482,36,'cp 8','d2d4 f8d6 d4c5 d6c5 b1c3 g8e7 c1e3 c5e3 f2e3 e8g8 d1d2 c8e6 e1c1 d8c7 b5d3 a7a6 c1b1 a8d8 d2f2 e7g6 h1f1 h7h6 c3e2 g6e5 f3e5 c6e5 h2h3 c7b6 e2d4 f8e8 g2g4 f7f6 f2f4 e6d7 f4g3 e5c4 d3c4 d5c4 g4g5');
INSERT INTO evaluations VALUES(20463,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(20463,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20463,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20463,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20463,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(20463,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(20463,36,'cp 0','f8d6 d4c5 d6c5 b1c3 g8e7 c1e3 c5e3 f2e3 e8g8 d1d2 c8e6 e1c1 d8c7 b5d3 a7a6 c1b1 a8d8 d2f2 e7g6 a2a3 h7h6 f3d4 g6e5 f2f4 c7e7 d4c6 e5c6 c3e2 e6d7 e2c3');
INSERT INTO evaluations VALUES(43211,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43211,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43211,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43211,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(43211,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43211,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43211,36,'cp 5','d7d6 f1e2 c7c5 c2c3 e8g8 d4c5 d6c5 e1g1 f6h5 f4e5 b8c6 e5g7 g8g7 d1c1 h5f6 c3c4 e7e6 b1c3 d8e7 b2b3 b7b6 c1b2 c8b7 a1d1 a8d8 h2h3 g7g8 f3h2 h7h6 b2a1 c6b4 d1d8 f8d8 e2f3 b7f3 h2f3');
INSERT INTO evaluations VALUES(41326,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(41326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41326,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(41326,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41326,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41326,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41326,36,'cp 12','f1e2 c7c5 c2c3 e8g8 d4c5 d6c5 e1g1 f6h5 d1d8 f8d8 f4c7 d8d7 c7e5 b8c6 e5g7 g8g7 e2b5 d7d6 b5c6 d6c6 c3c4 h5f6 b1c3 c8f5 f1d1 c6a6 f3e1 a6e6 h2h3 h7h5 c3b5 f6e8 e1f3 b7b6 b5c3 e8d6 b2b3 f5c2 d1d2 c2f5 a1d1 d6e4');
INSERT INTO evaluations VALUES(41325,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(41325,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41325,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41325,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41325,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(41325,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41325,36,'cp 14','f6h5 f4g5 h7h6 g5h4 g6g5 f3d2 h5f4 e3f4 g5h4 c2c3 e8g8 d2f3 e7e5 d4e5 d6e5 d1d8 f8d8 f3e5 b8c6 e5c6 b7c6 b1d2 d8e8 d2c4 a8b8 a1d1 h4h3 g2h3 c8h3 d1d2 e8e4 c4e5 g7e5 f4e5 e4e5 h1g1 g8f8 e1d1 b8e8 b2b3 h3f5 g1g3 f5b1 f2f4 e5e4 e2f3 e4f4');
INSERT INTO evaluations VALUES(39218,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(39218,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(39218,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(39218,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(39218,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(39218,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(39218,36,'cp 66','e2e3 c5b4 c3d5 b4e7 g1f3 d7d6 d2d4 e5e4 f3d2 f7f5 b2b4 g8f6 d5e7 d8e7 f1e2 a7a5 b4a5 a8a5 e1g1 e8g8 c1b2 a5a8 a2a4 b8c6 d2b3 b7b6 d4d5 c6e5 b2c3 e5g6 d1d2 c8d7 a4a5 b6a5 a1a5');
INSERT INTO evaluations VALUES(39215,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(39215,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(39215,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(39215,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(39215,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(39215,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(39215,36,'cp 62','c5b4 g1e2 g8f6 a2a3 b4e7 d2d4 e5d4 e3d4 d7d5 c3d5 f6d5 c4d5 d8d5 e2c3 d5d8 f1d3 c7c6 e1g1 c8e6 d3e4 e8g8 c1f4 e7d6 f4e3 h7h6 d4d5 c6d5 e4d5 e6d5 d1d5 b8c6 a1d1 d6e5 d5d7 d8d7 d1d7 f8d8 d7b7');
INSERT INTO evaluations VALUES(50217,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(50217,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50217,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(50217,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(50217,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(50217,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(50217,36,'cp 32','g1f3 f8g7 c2c4 e8g8 f1e2 c7c5 d4d5 d7d6 b1c3 e7e6 e1g1 e6d5 c4d5 c8g4 h2h3 g4f3 e2f3 a7a6 a2a4 b8d7 d1c2 c5c4 f3e2 a8c8 c1d2 f8e8 f1d1 d7c5 e2c4 f6e4 d2e1 e4c3 e1c3 g7c3 c2c3 c5e4 c3c2 d8h4 a4a5');
INSERT INTO evaluations VALUES(50218,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(50218,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(50218,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(50218,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(50218,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(50218,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(50218,36,'cp 16','f8g7 c2c4 e8g8 f1e2 c7c5 d4d5 e7e6 b1c3 d7d6 e3e4 e6d5 c4d5 c8g4 e1g1 d8e7 c1g5 h7h6 g5f6 g7f6 f3d2 g4e2 d1e2 b8d7 f2f4 f6d4 g1h1 a8e8 a1e1 a7a6 a2a4 f7f5 e4f5 e7e2 c3e2 f8f5 e2d4 e8e1 d4f5 e1f1 d2f1 g6f5 f1g3');
INSERT INTO evaluations VALUES(43246,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(43246,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(43246,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(43246,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43246,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(43246,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(43246,36,'cp 22','c2c4 e8g8 f1e2 c7c5 d4d5 e7e6 b1c3 d7d6 f3d2 b8a6 e3e4 a6c7 a2a4 f8e8 e1g1 b7b6 a1a3 c7a6 f2f3 a6b4 d2b1 d8e7 c1g5 h7h6 g5e3 a7a6 d1d2 g8h7 e3f2 c8b7 a4a5 e6d5 c3d5 f6d5 c4d5 b6a5');
INSERT INTO evaluations VALUES(43188,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(43188,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(43188,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(43188,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(43188,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(43188,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43188,36,'cp 25','e8g8 b1c3 d7d5 c4d5 f6d5 f1c4 d5c3 b2c3 c7c5 e1g1 d8c7 d1e2 b8c6 c1a3 c7a5 a3b2 c8g4 h2h3 g4f3 e2f3 a8c8 f3e2 e7e6 f1d1 c6e7 e3e4 f8d8 a2a4 c5d4 c3d4 a5b4 c4b5 e7c6 a1b1 g7d4 b2d4 c6d4 b1b4');
INSERT INTO evaluations VALUES(37407,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(37407,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(37407,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(37407,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(37407,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37407,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37407,36,'cp 15','f1e2 c7c5 d4c5 b8a6 e1g1 a6c5 b2b4 c5e4 c1b2 b7b6 b1c3 e4c3 b2c3 c8b7 a1c1 d7d6 d1b3 d8d7 f1d1 f8c8 a2a3 a7a5 b3b2 a5b4 a3b4 a8a4 c1a1 c8a8 h2h3 d7c8 f3d2 a4a1 d1a1 a8a1 b2a1 c8a8 a1a8 b7a8 f2f4 f6e4 d2e4 a8e4 c3g7 g8g7 g1f2 d6d5 c4c5 b6c5 b4c5');
INSERT INTO evaluations VALUES(37406,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37406,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(37406,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(37406,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(37406,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(37406,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(37406,36,'cp 9','c7c5 d4c5 b8a6 e1g1 a6c5 b2b4 c5e4 c1b2 b7b6 b1d2 e4d2 f3d2 c8b7 d1b3 a7a5 a2a3 d8c7 f1d1 f8c8 h2h3 d7d6 a1c1 a5b4 a3b4 c7c6 e2f3 c6d7 f3b7 d7b7 d2f3 b7a6 c1a1 a6b7 b2d4 a8a1 d1a1 f6d7 d4g7 g8g7 a1c1 g7g8 f3d4');
INSERT INTO evaluations VALUES(6388,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(6388,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6388,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(6388,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6388,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6388,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6388,36,'cp 0','c6d4 e1g1 c7c6 b5c4 d4f3 d1f3 g8f6 c4b3 e8g8 d2d3 d7d6 c1e3 c5b6 h2h3 a7a5 f1e1 f6d7 e1d1 b6d4 a2a3 d7c5 b3a2 g8h8 e3d4 e5d4 c3e2 f7f5 e4f5 f8f5 f3g3 f5g5 g3f3');
INSERT INTO evaluations VALUES(6826,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6826,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6826,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6826,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6826,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6826,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6826,36,'cp 0','e1g1 c7c6 b5c4 d4f3 d1f3 g8f6 d2d3 a7a5 f3g3 d8e7 a2a4 d7d6 g3g7 h8g8 g7h6 g8g6 h6h4 g6g4 h4h6');
INSERT INTO evaluations VALUES(52974,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(52974,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(52974,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52974,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(52974,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(52974,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(52974,36,'cp 93','c2c4 d6d5 e4d5 c6d5 g1f3 g8f6 b1c3 b8c6 c4d5 f6d5 f1b5 c8g4 d1a4 d8d6 f3e5 g4d7 e5d7 d6d7 e1g1 e7e6 c3d5 d7d5 c1e3 a7a6 b5e2 f8e7 e2f3 d5b5 a4b5 a6b5 d4d5 e6d5 f3d5 e8g8 f1d1 e7f6 g2g3 f8d8 a1c1 f6b2 c1b1 c6e7 d5b7 d8d1 b1d1');
INSERT INTO evaluations VALUES(52962,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(52962,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(52962,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(52962,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(52962,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(52962,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(52962,36,'cp 87','d6d5 e4d5 c6d5 g1f3 g8f6 b1c3 b8c6 c1g5 e7e6 c4c5 h7h6 g5f6 g7f6 f1b5 f8g7 e1g1 e8g8 f1e1 a7a6 b5a4 d8c7 a1c1 a8b8 h2h3 f8d8 d1d2 c6e7 b2b4 b7b6 c3e2 e6e5 f3h4 e5e4 e2f4 g8h7 g2g3 f6f5 h4g2 e7g6 g1h1 g6f4');
INSERT INTO evaluations VALUES(4341,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(4341,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(4341,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(4341,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(4341,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(4341,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4341,36,'cp 11','c2c3 d7d6 h2h3 a7a6 a2a4 h7h6 a4a5 c5a7 c1e3 d6d5 e4d5 f6d5 e3a7 a8a7 f1e1 d5f4 b1d2 b7b5 a5b6 c7b6 f3e5 c6e5 e1e5 f4d3 c4d3 d8d3 e5e3 d3b5 d1b3 c8d7 b3b5 d7b5 a1e1 f8d8 d2b3 a6a5 b3d4 b5d7 f2f4 g8f8 e3e4 f7f5 d4e6 d7e6 e4e6');
INSERT INTO evaluations VALUES(4333,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4333,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(4333,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(4333,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(4333,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(4333,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4333,36,'cp 1','d7d6 h2h3 a7a6 b1d2 h7h6 b2b4 c5a7 c4b3 c6e7 b3c2 c8e6 d3d4 e7g6 f1e1 c7c6 d2f1 a6a5 a2a3 f8e8 c1e3 a7b6 f1g3 b6c7 d4e5 d6e5 a3a4 d8d1 e1d1 f6d7 d1b1 d7b6 b4a5 b6c4 b1b7 c4e3 f2e3 c7a5');
INSERT INTO evaluations VALUES(3830,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(3830,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(3830,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3830,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(3830,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(3830,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(3830,36,'cp 11','f1e1 a7a6 c4b3 h7h6 b1d2 c8e6 d2f1 d6d5 e4d5 f6d5 h2h3 f8e8 f1g3 c5d6 c1d2 b7b6 c3c4 d5f4 b3a4 e6d7 d2f4 e5f4 e1e8 d7e8 g3e4 d6c5 d3d4 c6d4 e4c5 b6c5 a4e8 d4f3 d1f3 d8e8');
INSERT INTO evaluations VALUES(3831,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(3831,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(3831,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(3831,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3831,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3831,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3831,36,'cp 13','a7a6 b2b4 c5a7 a2a4 h7h6 h2h3 c8e6 c4e6 f7e6 c1e3 a7e3 e1e3 d8e8 b1d2 f6h5 g2g3 e8g6 g1h2 g6f6 d1e2 f6f7 a1f1 h5f6 a4a5 a8e8 d3d4 e5d4 f3d4 f7d7 f1d1 e6e5 e2c4 d7f7 d4c6 f7c4 d2c4 b7c6');
INSERT INTO evaluations VALUES(52768,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(52768,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(52768,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(52768,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(52768,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(52768,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(52768,36,'cp 37','e2e4 d5e4 c3e4 c8f5 e4g3 f5g6 h2h4 h7h6 g1f3 b8d7 h4h5 g6h7 f1d3 h7d3 d1d3 e7e6 c1f4 g8f6 e1c1 f8e7 g3e4 f6e4 d3e4 e8g8 c1b1 d7f6 e4e1 d8d5 f3e5 d5e4 e1e4 f6e4 f4e3 f8d8 c2c3 e4f6 a2a4 f6d5 b1c2 a8c8');
INSERT INTO evaluations VALUES(57003,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(57003,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(57003,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(57003,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(57003,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(57003,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(57003,36,'cp 0','d7d5 e4d5 e6d5 d2d4 f8b4 b1c3 g8f6 g1f3 e8g8 f1e2 c8e6 c4c5 b7b6 c5b6 a7b6 e1g1 h7h6 c1f4 b4c3 b2c3 f6e4 d1b3 d8f6 f4e5 f6d8 e5f4');
INSERT INTO evaluations VALUES(55797,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(55797,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55797,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55797,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55797,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(55797,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(55797,36,'cp 0','e4d5 e6d5 d2d4 f8b4 b1c3 g8f6 g1f3 b8c6 f1e2 d5c4 e2c4 e8g8 e1g1 c8g4 c1e3 d8d6 a1c1 a8e8 c4e2 g4h5 h2h3 b4c3 b2c3 f6d5 e3d2 h7h6 f1e1 e8e7 e2f1 e7e1 d2e1 f8e8 e1d2 d5f6 d2e3 f6d5');
INSERT INTO evaluations VALUES(55767,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55767,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55767,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55767,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55767,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55767,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(55767,36,'cp 0','e6d5 d2d4 f8b4 b1c3 g8f6 g1f3 e8g8 f1e2 b8c6 e1g1 d5c4 e2c4 c8g4 c1e3 d8d6 a1c1 a8e8 h2h3 g4h5 c3b5 d6d7 b5c3');
INSERT INTO evaluations VALUES(56181,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56181,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56181,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56181,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56181,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56181,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56181,36,'cp 0','d2d4 g8f6 b1c3 b8c6 g1f3 f8b4 f1e2 d5c4 e2c4 c8g4 e1g1 e8g8 c1e3 d8d6 a1c1 a8e8 h2h3 g4h5 c3b5 d6d7 b5c3');
INSERT INTO evaluations VALUES(56182,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56182,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56182,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56182,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(56182,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(56182,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(56182,36,'cp -2','f8b4 b1c3 g8f6 g1f3 b8c6 f1e2 d5c4 e2c4 c8g4 e1g1 e8g8 c1e3 g4h5 h2h3 f8e8 a2a3 b4c3 b2c3 d8d6 a3a4 h7h6 a4a5 a8d8 c4e2 b7b6 a5b6 a7b6 f1e1 c6e7 f3e5 h5e2 d1e2 f6d7 e5f3 e7d5 e2d3 c7c5 c3c4 d5e3 e1e3');
INSERT INTO evaluations VALUES(38942,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(38942,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(38942,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(38942,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(38942,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(38942,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38942,36,'cp -4','b1c3 g8f6 g1f3 b8c6 f1e2 e8g8 e1g1 d5c4 e2c4 c8g4 c1e3 g4h5 a2a3 b4a5 h2h3 d8d7 c4e2 a8d8 c3a4 a5b6 a4b6 a7b6 f3e5 c6e5 e2h5 f6h5 d4e5 d7f5 d1f3 f5f3 g2f3 f7f6 e5e6 f8e8 a1c1 c7c5 b2b4 e8e6 b4c5 b6c5');
INSERT INTO evaluations VALUES(38940,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(38940,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(38940,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(38940,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(38940,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(38940,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(38940,36,'cp -11','g8f6 g1f3 e8g8 f1e2 c8e6 c4c5 f6e4 d1b3 b4c3 b2c3 b7b6 e1g1 f8e8 c5b6 a7b6 c1f4 b8c6 e2d3 c6a5 b3c2 e6f5 f1e1 a5c4 a2a4 f5g6 d3c4 d5c4 f3e5 e4d6 c2d2 g6e4 f2f3 e4b7 e5g4 e8e1 d2e1 b7c6 f4d6 c7d6 g4e3 a8a4 a1a4');
INSERT INTO evaluations VALUES(42251,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42251,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(42251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42251,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(42251,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(42251,36,'cp -8','g1f3 e8g8 f1e2 d5c4 e2c4 b8c6 e1g1 c8g4 c1e3 d8d6 a1c1 a8d8 c3a4 d6d7 h2h3 g4f3 d1f3 c6a5 a4c5 b4c5 d4c5 a5c4 c1c4 d7d5 f3d5 f6d5 e3d2 f7f6 c4d4 c7c6 f1e1 f8e8 e1e8 d8e8 g1f1');
INSERT INTO evaluations VALUES(42249,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42249,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42249,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42249,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42249,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42249,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(42249,36,'cp -29','e8g8 f1d3 c8g4 e1g1 b8c6 c1e3 d5c4 d3c4 d8d6 a1c1 a8e8 c3b5 d6d7 b5c3 a7a6 h2h3 g4h5 a2a3 b4c3 c1c3 h7h6 c4b3 e8d8 c3c5 d7d6 g2g4 h5g6 d1c1 f6e4 c5d5 d6f6 f3e5 d8d5 e5g6 c6d4 e3d4 d5d4 g6f8 g8f8 c1c7');
INSERT INTO evaluations VALUES(37117,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(37117,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(37117,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(37117,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(37117,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(37117,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(37117,36,'cp -25','f1e2 d5c4 e2c4 b8c6 c1e3 c8g4 e1g1 d8d6 a1c1 a8e8 h2h3 g4h5 c4e2 a7a6 d1b3 b4c3 c1c3 f6d5 c3c1 c6a5 b3d3 h7h6 f3e5 h5e2 d3e2 f7f6 e5g6 f8f7 e2d3 a5c6 e3d2 c6e7 g6h4 d6d7 c1e1 e8d8 e1e4 c7c6 f1e1');
INSERT INTO evaluations VALUES(37116,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(37116,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37116,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(37116,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(37116,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(37116,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(37116,36,'cp -13','d5c4 e2c4 c8g4 e1g1 b8c6 c1g5 h7h6 g5h4 b4a5 a1c1 g4f3 d1f3 d8d4 h4f6 d4f6 f3f6 g7f6 c3d5 g8g7 f1d1 a8d8 g1f1 c6e5 c4e2 f8e8 g2g3 c7c6 d5e3 g7g6 d1d8 e8d8 c1d1 f6f5 d1d8 a5d8 f2f4 d8b6 e3c4 e5c4 e2c4 b6d4 b2b3 b7b5 c4e2 g6f6 e2d3 f6e6 h2h3 e6d6 d3f5');
INSERT INTO evaluations VALUES(37124,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(37124,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(37124,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(37124,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(37124,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(37124,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(37124,36,'cp -14','e2c4 c8g4 e1g1 b8c6 c1g5 h7h6 g5h4 b4a5 a1c1 f8e8 h2h3 g4f3 d1f3 d8d4 h4f6 d4f6 f3f6 g7f6 c4b5 e8e6 b5c6 e6c6 c3d5 c6c1 f1c1 g8g7 b2b4 a8d8 b4a5 d8d5 a5a6 b7a6 c1c7 d5a5 c7a7 h6h5 g1h2 h5h4 g2g3 a5a2 g3h4 a2f2 h2g3 f2a2 h4h5 a2a3 g3g4 a3a1 h5h6 g7h6 a7f7 h6g6');
INSERT INTO evaluations VALUES(53063,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(53063,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(53063,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(53063,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(53063,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(53063,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(53063,36,'cp 87','d2d4 g8f6 f1d3 c5d4 c3d4 e7e5 g1f3 e5d4 f3d4 g7g6 d3c4 f8g7 b1c3 e8g8 e1g1 a7a6 c1e3 b8c6 f2f3 c6a5 c4b3 f8e8 d1d2 b7b5 f1d1 c8b7 a1c1 a8c8 h2h3 f6h5 b3d5 a5c4 d5c4 c8c4');
INSERT INTO evaluations VALUES(53048,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(53048,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(53048,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(53048,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(53048,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(53048,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(53048,36,'cp 69','g8f6 f1d3 c5d4 c3d4 e7e5 g1e2 e5d4 e2d4 b8c6 d4c6 b7c6 d1c2 f8e7 e1g1 e8g8 c2c6 c8e6 b1c3 d6d5 f1d1 e6g4 d1d2 d5d4 c3d5 a8c8 d5f6 e7f6 c6b5 g7g6 h2h3 g4d7 b5d5 d8e7 d2e2 c8c5 d5b7 f6e5 b7a6 e7h4 a6a7');
INSERT INTO evaluations VALUES(45177,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(45177,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(45177,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(45177,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(45177,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(45177,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(45177,36,'cp 67','f1d3 c5d4 c3d4 e7e5 g1e2 f8e7 e1g1 e5d4 e2d4 b8c6 d3b5 c8d7 b1c3 e8g8 c1e3 a7a6 d4c6 b7c6 b5e2 d7e6 a1c1 h7h6 h2h3 d8b8 b2b3 d6d5 e4d5 f6d5 c3d5 e6d5 e2f3 f8d8 f3d5 c6d5 d1d2 e7f6 f1d1');
INSERT INTO evaluations VALUES(45178,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(45178,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(45178,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(45178,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(45178,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(45178,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(45178,36,'cp 62','c5d4 c3d4 e7e5 g1e2 e5d4 e2d4 g7g6 b1c3 f8g7 d3e2 e8g8 d4b5 f6e8 c3d5 b8d7 e1g1 a7a6 b5c3 b7b5 c1f4 c8b7 a1c1 a8c8 a2a3 d7b6 d5b6 d8b6 e2f3 g7c3 b2c3 b6a5 d1b3 a5c7 f1d1');
INSERT INTO evaluations VALUES(50196,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50196,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50196,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50196,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50196,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50196,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50196,36,'cp 0','f8g7 e2e3 d7d6 f1e2 c7c5 b1c3 c5d4 e3d4 e8g8 d1d2 a7a6 f4h6 b7b5 a2a4 b5b4 c3d1 f6e4 d2e3 e4f6');
INSERT INTO evaluations VALUES(43213,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43213,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43213,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43213,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43213,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43213,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43213,36,'cp 0','b1c3 d7d5 e2e3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(43200,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43200,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43200,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43200,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43200,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43200,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43200,36,'cp 0','d7d5 e2e3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(41460,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41460,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41460,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41460,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41460,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41460,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41460,36,'cp 0','e2e3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(41458,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41458,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41458,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41458,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41458,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41458,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41458,36,'cp 0','e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(36741,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36741,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36741,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36741,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36741,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36741,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36741,36,'cp 0','c3b5 b8a6 b5c3 a6b8');
INSERT INTO evaluations VALUES(21903,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(21903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21903,36,'cp 0','c6a5 c4d5 g8f6 e1g1 f8e7 b2b4 a5c6 d5c6 b7c6 a2a4 e8g8 b4b5 c8e6 c2c4 f6d7 b1c3 f7f5 b5c6 d7b8 c3d5 b8c6 a1b1 g8h8 a4a5 f5e4 d3e4 e6g4 d1a4 g4d7 a4a3 d7g4');
INSERT INTO evaluations VALUES(22155,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22155,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22155,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22155,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(22155,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(22155,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22155,36,'cp 0','c4d5 g8f6 e1g1 f8e7 b2b4 a5c6 d5c6 b7c6 a2a4 e8g8 f1e1 d8e8 b1d2 a7a5 c2c3 f6h5 d3d4 f7f5 d4e5 e8g6 e5d6 h5f4 g2g3 c7d6 e4f5 c8f5 e1e7 f4h3 g1g2 h3f4');
INSERT INTO evaluations VALUES(21913,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(21913,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(21913,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(21913,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(21913,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(21913,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21913,36,'cp 54','g8f6 d4d5 c6b8 b1c3 f8e7 e1g1 e8g8 a2a4 a7a5 h2h3 b8a6 c1e3 a6b4 f3d2 f6d7 c3b5 g8h8 c2c3 b4a6 b5a7 e7g5 e3g5 d8g5 a7c8 f8c8 b2b4 a5b4 c4a6 b7a6 c3b4 c8b8 a1b1 d7f6 g1h1 g5g6 d2c4');
INSERT INTO evaluations VALUES(15524,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(15524,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(15524,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(15524,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(15524,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(15524,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(15524,36,'cp 57','d4d5 c6b8 b1c3 b8d7 e1g1 f8e7 a2a4 e8g8 d1e2 a7a5 c1e3 f6g4 e3d2 d7b6 c4b3 c7c6 f1d1 c6d5 c3d5 b6d5 b3d5 h7h6 d2c3 d8c7 d5b3 b7b6 f3d2 g4f6 f2f3 f6h5 g2g3 c8e6 b3e6 f7e6 e2c4 c7d7 g1g2 h5f6 c4e2');
INSERT INTO evaluations VALUES(15453,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(15453,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(15453,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(15453,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(15453,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(15453,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(15453,36,'cp 63','c6b8 b1c3 f8e7 e1g1 e8g8 a2a4 a7a5 h2h3 b8a6 c1e3 a6b4 f3d2 f6d7 c3b5 h7h6 f1e1 e7g5 e3g5 h6g5 a1a3 d7c5 a3g3 c8d7 c2c3 b4a6 b2b3 g7g6 b5a3 g8g7 c4f1 f7f6 a3c4 b7b6 c4e3 d8e7 f1a6 a8a6');
INSERT INTO evaluations VALUES(5975,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(5975,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(5975,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(5975,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(5975,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(5975,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(5975,36,'cp 47','d2d4 e5d4 c3d4 c5b6 h2h3 g8f6 e1g1 e8g8 f1e1 f6e4 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 c8f5 e4e1 b6a5 c1g5 f7f6 d1b3 f8f7 g5d2 a5c3 b2c3 d8d5 c3c4 d5d8 b3b7 d8c8 b7b5 f5d7 b5c5 e7f5 a4a5 c8b7 c5a3 a8e8 e1e8 d7e8 a5a6 b7c6 a1c1');
INSERT INTO evaluations VALUES(5994,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(5994,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(5994,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(5994,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(5994,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5994,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(5994,36,'cp 55','e5d4 c3d4 c5b6 h2h3 g8f6 e1g1 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 c8f5 e4e1 b6a5 c1g5 f7f6 d1b3 f8f7 g5f4 a5c3 b2c3 d8d5 c3c4 d5c5 b3b7 c5c6 b7c7 c6c7 f4c7 e7c6 c7d6 f7d7 c4c5 g8f7 a4a5 a8d8 f3h4 f5e6 a5a6 e6b3 h4f5');
INSERT INTO evaluations VALUES(6042,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(6042,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(6042,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(6042,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(6042,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6042,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(6042,36,'cp 75','c3d4 c5b6 e1g1 g8f6 h2h3 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 a7a5 c1g5 f7f6 d1b3 g8h8 g5f4 e7g6 f4h2 f6f5 e4c4 f5f4 a1d1 h7h6 g1h1 d8f6 c3b5 f6f7 b3c3 c8d7 b5c7 a8c8 c7b5 c8c4 c3c4 b6f2 c4c7 f7e7 c7a5');
INSERT INTO evaluations VALUES(6039,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(6039,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(6039,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(6039,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(6039,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(6039,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(6039,36,'cp 69','c5b6 e1g1 g8f6 h2h3 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 h7h6 a4a5 b6c5 c1f4 e7f5 b2b4 c5d6 d1d2 c8d7 a1e1 f8e8 f3e5 a7a6 d2e2 d6e5 f4e5 f5d6 e4e3 d8h4 e5d6 e8e3 e2e3 c7d6');
INSERT INTO evaluations VALUES(5781,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(5781,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(5781,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(5781,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(5781,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(5781,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(5781,36,'cp 52','h2h3 g8f6 e1g1 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 a7a5 c1g5 f7f6 d5d6 c7d6 g5e3 b6e3 e4e3 g8h8 e3d3 a8a6 f3d4 d6d5 a1c1 a6b6 c3b5 e7c6 c1c5 f8e8 d3c3 h7h6 d4c6 b7c6 c5c6 c8a6 c6c7 a6b5 a4b5 b6b5');
INSERT INTO evaluations VALUES(5778,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(5778,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(5778,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(5778,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(5778,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(5778,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(5778,36,'cp 42','g8f6 e1g1 e8g8 f1e1 f6e4 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 a7a5 c1g5 f7f6 d1b3 g8h8 g5f4 e7g6 f4h2 f6f5 e4c4 f5f4 a1d1 f8f5 g1h1 h7h6 c4e4 b6f2 c3b5 f5f6 f3d4 f2d4 b5d4 c8f5 e4e1 d8d7 b3b7 a8d8 d4c6 f5c2 c6d8 c2d1 d8c6');
INSERT INTO evaluations VALUES(19165,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(19165,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(19165,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(19165,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(19165,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(19165,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(19165,36,'cp 189','d4e3 d7d6 e3g3 g8f6 b1c3 c8d7 f1e2 d8e7 e2f3 h7h6 h2h4 d6d5 e4e5 d7c6 c1f4 e7e6 e1c1 f6e4 f3e4 d5e4 h1e1 e6f5 c1b1 g7g5 h4g5 h6g5 g3g5 f5g5 f4g5 f8e7 g5e7 e8e7 c3e4 a8g8 g2g3 h8h2 e4c5 h2f2 e5e6 f7f5 c5b7 g8g3');
INSERT INTO evaluations VALUES(19168,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(19168,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(19168,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(19168,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(19168,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(19168,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(19168,36,'cp 196','d7d6 b1c3 g8f6 e3g3 c8d7 f1e2 d8e7 e2f3 h7h6 h2h4 d6d5 e4e5 d7c6 c1f4 e7e6 f3e2 f6d7 e2g4 f7f5 g4h5 e8d8 c3e2 d5d4 c2c3 d4c3 g3c3 f8e7 a1d1 d8c8 h5f3 h8d8 f3c6 b7c6 f4g3 d7b6 e2f4 d8d1 e1d1 e6d7 d1e2 c8b7 h1d1');
INSERT INTO evaluations VALUES(20755,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(20755,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(20755,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(20755,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(20755,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(20755,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(20755,36,'cp 194','b1c3 g8f6 e3g3 c8d7 c1d2 d8e7 e1c1 e8c8 f1d3 d7c6 h1e1 e7d7 d2g5 h7h6 g5f6 g7f6 d3c4 h6h5 c4d5 c8b8 c1b1 d7g4 d5f7 f6f5 f7d5 g4g3 h2g3 f5e4 c3e4 b8c7 d5c6 c7c6 f2f4 f8g7 c2c3 h8g8 e1e2 b7b5 b1c2 g7h8 e4g5 h8f6');
INSERT INTO evaluations VALUES(20751,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(20751,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(20751,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(20751,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(20751,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(20751,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(20751,36,'cp 199','g8f6 e3g3 c8d7 c1d2 d8e7 e1c1 e8c8 f1d3 d7c6 h1e1 e7d7 d3c4 h7h6 a2a3 g7g5 h2h4 g5h4 g3h4 f8e7 h4h1 f6g4 h1f1 c8b8 c4d5 h6h5 c1b1 b8a8 f2f3 g4e5 f1e2 e7f6 e1h1 e5g6 h1f1 f6d4 d5c6 b7c6 f3f4');
INSERT INTO evaluations VALUES(14011,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(14011,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(14011,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(14011,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(14011,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(14011,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(14011,36,'cp 217','e3g3 c8e6 f1e2 f8e7 c1f4 d8a5 e1c1 e8g8 a2a3 b7b5 f4h6 g7g6 e4e5 d6e5 g3e5 f8d8 d1d8 a8d8 e2b5 f6h5 h1d1 d8d1 c1d1 a5b6 d1c1 e7d6 e5e3 h5f6 b5e2 d6h2 g2g3 b6c6 e2f3 c6b6 c3d5 f6d5 f3d5 b6a6 d5e6');
INSERT INTO evaluations VALUES(39238,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(39238,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(39238,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(39238,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(39238,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(39238,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(39238,36,'cp 36','g1f3 d7d6 d2d4 e5d4 f3d4 g8f6 c1f4 e8g8 d1d2 b8c6 d4c6 b7c6 f1d3 f6g4 e1g1 a7a5 a1d1 d8e7 c3a4 c5a7 d1e1 f7f5 h2h3 g4e5 e4f5 c8f5 a4c3 f5d3 c2d3 a7d4 f4e3 e7h4 e3d4 h4d4');
INSERT INTO evaluations VALUES(39233,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(39233,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(39233,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(39233,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(39233,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(39233,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(39233,36,'cp 31','d7d6 d2d4 e5d4 f3d4 g8f6 c1f4 e8g8 d1d2 d6d5 e4d5 f8e8 f1e2 f6d5 e1c1 d5f4 d2f4 c5d6 f4f3 b8d7 c3e4 d7f6 e4f6 d8f6 f3f6 g7f6 e2f3 g8f8 a2a4 a8b8 d1d3 c8d7 b2b3 e8e5 h1d1 a7a5 h2h3 h7h5 c1b2 d6c5 d4e2 c5d6 e2f4 h5h4 d3d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(38725,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(38725,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(38725,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(38725,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(38725,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(38725,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38725,36,'cp 26','d2d4 e5d4 f3d4 g8f6 c1f4 e8g8 d1d2 d6d5 e4d5 f8e8 f1e2 f6d5 e1c1 d5f4 d2f4 c5d6 f4f3 b8d7 c3e4 d7f6 e4f6 d8f6 f3f6 g7f6 e2f3 a7a5 h1e1 e8e1 d1e1 g8f8 g2g3 c7c6 c2c3 f6f5 f3e2 a5a4 e1d1 d6c5 f2f4 f8g7 a2a3 g7f6 e2c4 c5b6');
INSERT INTO evaluations VALUES(38721,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38721,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38721,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(38721,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(38721,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38721,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(38721,36,'cp 17','e5d4 f3d4 g8f6 c1f4 e8g8 d1d2 d6d5 e4d5 f8e8 f1e2 f6d5 f4g5 d8d7 e1c1 b8c6 d4c6 d5c3 d2c3 d7c6 e2f3 c6b6 h1e1 e8e1 d1e1 c8e6 g5e3 c5e3 e1e3 a8d8 b2b3 c7c5 g2g3 b6c7 f3e2 h7h6 c1b2 b7b6 a2a3 d8d4 e2b5 c7d6 f2f4');
INSERT INTO evaluations VALUES(38748,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(38748,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38748,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38748,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(38748,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(38748,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(38748,36,'cp 25','f3d4 g8f6 c1f4 e8g8 d1d2 d6d5 e4d5 f8e8 f1e2 f6d5 e1c1 d5f4 d2f4 c5d6 f4f3 b8d7 c3e4 d7f6 e4f6 d8f6 f3f6 g7f6 e2f3 a7a5 a2a4 a8b8 c2c3 c8d7 d4b5 d6f4 c1c2 d7f5 c2b3 f4e5 b5d4 f5g6 h2h4 h7h5 b3c4 e8d8 h1e1 c7c6 c4c5 b7b6 c5c6 g8g7');
INSERT INTO evaluations VALUES(38745,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(38745,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38745,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38745,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(38745,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(38745,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(38745,36,'cp 26','g8f6 c1f4 e8g8 d1d2 d6d5 e4d5 f8e8 f1e2 f6d5 e1c1 d5f4 d2f4 c5d6 f4f3 b8d7 c3e4 d7f6 e4f6 d8f6 f3f6 g7f6 e2f3 g8f8 h2h3 d6e5 a2a3 a8b8 g2g3 e8d8 c2c3 e5d4 d1d4 d8d4 c3d4 c7c6 c1d2 c8e6 f3e4 b8d8 d2e3 e6d5 e4d5 d8d5 b2b3 f6f5');
INSERT INTO evaluations VALUES(41905,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41905,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(41905,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(41905,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(41905,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41905,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41905,36,'cp 21','c1f4 e8g8 d1d2 d6d5 e4d5 f8e8 f1e2 f6d5 e1c1 d5f4 d2f4 c5d6 f4f3 b8d7 c3e4 d7f6 e4f6 d8f6 f3f6 g7f6 e2f3 d6f4 c1b1 c7c6 c2c3 f4c7 a2a4 a7a5 d1d2 e8e5 g2g4 f6f5 g4f5 g8g7 h1d1 c7b6 f3g4 b6d4 d2d4 c8f5 b1c1 a8b8 f2f3 f5g6 d4d7');
INSERT INTO evaluations VALUES(20820,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(20820,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(20820,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(20820,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(20820,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(20820,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(20820,36,'cp 99','c2c4 g7g6 b1c3 b7b6 f1e2 f8g7 h2h3 g8f6 g2g4 h7h5 f3h2 c8d7 c1e3 c7c6 a2a4 a8c8 f2f3 e8g8 d1d2 g8h8 h1g1 a7a6 h2f1 h8g8 e1f2 g8h8 a1d1 c6d5 c4d5 b6b5 a4b5 a6b5 c3b5 d7b5 e2b5 c8b8 g4g5 b8b5');
INSERT INTO evaluations VALUES(20812,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(20812,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(20812,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(20812,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(20812,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(20812,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(20812,36,'cp 101','g7g6 b1c3 f8g7 c4c5 g8f6 c5d6 c7d6 f1c4 e8g8 e1g1 f6e8 a2a4 f7f5 f1e1 g8h8 b2b3 f5e4 f3g5 e8f6 c3e4 f6e4 g5e4 c8f5 e4g3 e7g8 h2h3 g7h6 c1h6 g8h6 d1d2 d8h4 a4a5 f5d7 a5a6 b7b5 c4f1 h6f5 g3e4 f5d4');
INSERT INTO evaluations VALUES(17175,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(17175,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(17175,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17175,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17175,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(17175,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(17175,36,'cp 33','e7g6 d2d4 e5d4 f3d4 f8c5 d4f5 e8g8 c1e3 c5e3 f5e3 d7d6 f2f4 c6e7 f4f5 g6e5 b1c3 g8h8 d1d4 a7a6 b5e2 e7g8 c3d5 c7c6 d5f4 c6c5 d4d2 b7b5 a2a4 c8b7 f1d1 g8f6 d2d6 d8d6 d1d6 b7e4 a4b5 a6b5 a1a8 f8a8 e2b5 f6g4 e3g4 e5g4 d6a6 a8a6 b5a6 e4c2 f4d3');
INSERT INTO evaluations VALUES(16728,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(16728,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(16728,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(16728,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(16728,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(16728,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(16728,36,'cp 53','d2d4 e5d4 f3d4 f8c5 d4f5 e8g8 c1e3 c5e3 f5e3 d7d6 b1c3 c6e7 f2f4 c7c6 b5c4 c8e6 f4f5 d8b6 f5e6 b6e3 g1h1 f7e6 c4e6 g8h8 d1d6 a8d8 d6g3 e3g3 h2g3 g6e5 h1g1 f8f1 a1f1 g7g5 f1d1 d8d1 c3d1 h8g7 d1e3 g7f6 e6g4 e5f7 g4f5 e7f5 e3f5 f7e5 g1f2 e5g4 f2f3');
INSERT INTO evaluations VALUES(53343,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(53343,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(53343,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(53343,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(53343,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(53343,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(53343,36,'cp -3','c2c3 c8f5 d2d4 e7e6 f1b5 b8c6 g1f3 f5g6 b1d2 c5d4 f3d4 g8e7 h2h4 h7h5 b5a4 a7a6 d2f3 b7b5 d4c6 e7c6 a4c2 g6c2 d1c2 f8e7 c1f4 a8c8 a2a3 a6a5 e1g1 e8g8 f4g5 b5b4 a3b4 a5b4 c2e2 b4c3 b2c3 e7g5 f3g5');
INSERT INTO evaluations VALUES(53338,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(53338,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53338,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(53338,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53338,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(53338,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53338,36,'cp 0','c8f5 d2d4 e7e6 f1b5 b8c6 g1e2 d8b6 b5a4 c5d4 a4c6 b7c6 c3d4 g8e7 b1c3 b6a6 e1g1 e7g6 c1e3 f8e7 c3a4 e8g8 b2b3 e7a3 e3c1 a3e7');
INSERT INTO evaluations VALUES(31295,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31295,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31295,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31295,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31295,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31295,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31295,36,'cp 0','f1b5 b8c6 d2d4 e7e6 g1e2 d8b6 b5a4 c5d4 e1g1 f5g4 c3d4 g4e2 d1e2 b6d4 b1c3 d4b6 c3b5 a7a6 c1e3 b6a5 e3d2 a5b6');
INSERT INTO evaluations VALUES(51984,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(51984,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(51984,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51984,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(51984,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(51984,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(51984,36,'cp 22','d7d5 e4d5 d8d5 f1e2 b8c6 e1g1 g8f6 d2d4 c5d4 f3d4 c6d4 c3d4 d5d6 e2f3 f8e7 b1c3 e8g8 g2g3 f8d8 c1e3 d6b4 d1b3 b4b3 a2b3 a7a5 f1c1 a8a6 c3a4 f6d5 e3d2 e7f6 a4c5 a6a7 f3d5 d8d5 a1a5 a7a5 d2a5 f6d4 c5e4');
INSERT INTO evaluations VALUES(54250,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(54250,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54250,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54250,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(54250,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(54250,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(54250,36,'cp 22','e4d5 d8d5 d2d4 b8c6 c1e3 c5d4 c3d4 g8f6 b1c3 f8b4 f1d3 e8g8 e1g1 d5h5 h2h3 b4d6 f1e1 h7h6 a1c1 c8d7 d1b3 c6a5 b3c2 a5c6 a2a3 a8c8 c2d2 c6e7 f3e5 d7c6 e5c6 c8c6 d2e2 f8d8 e2h5');
INSERT INTO evaluations VALUES(54144,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(54144,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54144,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54144,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(54144,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(54144,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54144,36,'cp 15','d8d5 f1e2 b8c6 e1g1 g8f6 d2d4 c5d4 f3d4 c8d7 c1e3 f8e7 e2f3 d5e5 b1d2 e8g8 d4c6 d7c6 f3c6 b7c6 e3d4 e5f5 d2c4 f8d8 d1e2 f6d5 a2a4 a7a5 a1e1 d5f4 e2f3 c6c5 d4e5 f4d3 f3f5 e6f5 e1d1 d3e5 d1d8 a8d8');
INSERT INTO evaluations VALUES(34485,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(34485,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(34485,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(34485,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(34485,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(34485,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(34485,36,'cp 24','f1e2 g8f6 e1g1 b8c6 d2d4 c5d4 f3d4 c8d7 c1e3 f8e7 e2f3 d5e5 b1d2 e8g8 d4c6 d7c6 f3c6 b7c6 e3d4 e5b5 d1c2 f8d8 c3c4 b5f5 c2f5 e6f5 d4c3 f6e4 d2e4 f5e4 g2g4 e7f6 c3f6 g7f6 a1d1 d8d3 d1d3 e4d3 f1d1 a8d8');
INSERT INTO evaluations VALUES(50175,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(50175,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(50175,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(50175,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(50175,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(50175,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(50175,36,'cp 49','e2e4 e7e5 g1f3 b8c6 d2d4 e5d4 f3d4 f8g7 c1e3 f6g4 d1g4 c6d4 g4d1 c7c5 f1d3 d7d6 d1d2 e8g8 e1g1 c8e6 a1d1 a7a6 b2b3 d8d7 c3e2 b7b5 e3d4 c5d4 c4b5 a6b5 f2f4 f7f5 e4f5 e6f5 d3f5 d7f5 h2h3 f5e4 e2d4 e4d4 d2d4 g7d4 d1d4 a8a2 d4d6');
INSERT INTO evaluations VALUES(50167,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(50167,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(50167,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(50167,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(50167,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(50167,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50167,36,'cp 37','e7e5 g1f3 b8c6 d2d4 e5d4 f3d4 f8g7 c1e3 f6g4 d1g4 c6d4 g4d1 c7c5 f1d3 d7d6 h2h4 h7h6 d1d2 a7a5 e3d4 c5d4 c3d5 a5a4 g2g3 e8g8 h4h5 g6g5 e1f1 c8e6 f1g2 g7e5 f2f4 g5f4 g3f4 e5g7 a1f1 e6d5 e4d5 g8h8 f4f5 d8g5 d2g5 h6g5');
INSERT INTO evaluations VALUES(49208,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(49208,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(49208,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(49208,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(49208,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(49208,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(49208,36,'cp 71','g1f3 b8c6 d2d4 e5d4 f3d4 f8g7 d4c6 d7c6 d1d8 e8d8 f1e2 h8e8 c1g5 h7h6 g5h4 c8d7 f2f4 d8c8 e4e5 f6g4 h2h3 g4e3 e1f2 e3f5 h4f6 g7f6 e5f6 e8e6 g2g4 f5d4 g4g5 d4e2 c3e2 b7b5 h3h4 b5c4 a1e1 a8b8 h4h5 g6h5 g5h6 e6f6 h1h5 c8b7 h6h7 b8h8 e1g1 d7f5 g1g7 f5d3 f2f3');
INSERT INTO evaluations VALUES(50492,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(50492,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(50492,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50492,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(50492,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(50492,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(50492,36,'cp -1','e1g1 b7b5 c4e2 c8b7 a2a4 b5b4 e4e5 b8c6 c2c3 g8e7 d2d4 c5d4 c3d4 d7d6 b1d2 d6e5 d4e5 c6a5 d2c4 a5c4 e2c4 d8d1 f1d1 e7d5 f3d2 f8e7 d2b3 d5b6 c4e2 b7d5 b3a5 b6d7 e2c4 d5c4 a5c4 a8c8 c4d6 e7d6 d1d6');
INSERT INTO evaluations VALUES(50490,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50490,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50490,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50490,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50490,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50490,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50490,36,'cp 0','b7b5 c4e2 c8b7 e4e5 d7d6 a2a4 b5b4 d2d4 b8c6 c2c3 d6e5 d4e5 d8d1 f1d1 c6a5 c1e3 a5b3 a1a2 b7d5 a4a5 g8e7 c3b4 c5b4 a2a4 e7f5 e3b6 d5c6 a4a2 c6d5');
INSERT INTO evaluations VALUES(51199,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(51199,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(51199,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(51199,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(51199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51199,36,'cp 9','c4e2 c8b7 d2d3 f8e7 f1e1 b8c6 c2c3 g8f6 e4e5 f6d5 c3c4 d5b6 b1c3 c6d4 c4b5 d4e2 d1e2 b6d5 c3d5 b7d5 b5a6 a8a6 d3d4 d8a8 d4c5 e7c5 c1d2 e8g8 a2a3 c5e7 a1c1 f8b8 d2e3 a8b7 c1c2 d5e4 c2c3 b7b2');
INSERT INTO evaluations VALUES(51218,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(51218,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(51218,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51218,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(51218,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(51218,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51218,36,'cp 0','c8b7 d2d3 b8c6 f1e1 g8f6 c2c3 d7d6 c1e3 f8e7 b1d2 e8g8 e2f1 c6e5 f3e5 d6e5 a2a4 d8c7 d1f3 b7c6 a4b5 a6b5 f3g3 g8h8 d2f3 e7d6 g3h3 h8g8 h3g3');
INSERT INTO evaluations VALUES(30336,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(30336,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30336,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(30336,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(30336,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30336,36,'cp 0','d2d3 b8c6 f1e1 g8f6 c2c3 d8c7 e2f1 f8e7 d3d4 e8g8 c1g5 h7h6 g5h4 d7d5 e4d5 f6d5 h4g3 c7b6 d4c5 e7c5 b1d2 f8d8 d1e2 c6e7 d2e4 d5e3 e4c5 e3f1 c5b7 f1g3 h2g3 b6b7 f3e5 e7d5 a2a4 d5f6 e2f3 d8d5 a4b5 a6b5 a1a8 b7a8 e1d1 d5d1 f3d1 a8e4 e5d3 e4c4');
INSERT INTO evaluations VALUES(30335,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30335,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30335,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30335,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30335,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(30335,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30335,36,'cp 0','b8c6 f1e1 g8f6 c2c3 f8e7 e4e5 f6d5 c3c4 d5b6 b1c3 c6d4 c4b5 d4e2 d1e2 a6b5 c3b5 e8g8 b5c3 b6d5 c3e4 d8b6 b2b3 g8h8 c1d2 f7f5 e5f6 e7f6 e4f6 g7f6 f3h4 f8g8 a2a4 g8g7 d2h6 g7g8');
INSERT INTO evaluations VALUES(11539,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(11539,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(11539,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(11539,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(11539,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(11539,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11539,36,'cp 0','f8c5 c2c3 b7b5 a4b3 d7d6 c1g5 h7h6 g5h4 c8e6 e1g1 g7g5 b3e6 f7e6 h4g3 e8g8 a2a4 b5b4 d3d4 e5d4 c3d4 c5b6 d1d3 g5g4 a4a5 b6a7 e4e5 g4f3 d3g6 g8h8 g3h4 d6e5 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(7333,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7333,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7333,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7333,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(7333,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7333,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7333,36,'cp 15','c2c3 b7b5 a4b3 d7d6 c1g5 h7h6 g5h4 c8e6 b1d2 c5b6 a2a4 g7g5 h4g3 e8g8 e1g1 e6b3 d1b3 a8b8 b3d1 d8d7 h2h3 b8e8 f1e1 c6e7 d2f1 e7g6 f1e3 b6e3 e1e3 g5g4 h3g4 f6g4 e3e2 f7f5 e4f5 f8f5 a4b5 a6b5 d3d4 e5e4 f3h4 g6h4 g3h4 e4e3 f2e3');
INSERT INTO evaluations VALUES(7330,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(7330,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(7330,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7330,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(7330,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(7330,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7330,36,'cp 0','d7d6 e1g1 c5a7 h2h3 e8g8 f1e1 f6h5 d3d4 d8f6 a4c6 b7c6 b1d2 h7h6 d2c4 f8e8 b2b4 e5d4 c3d4 f6g6 f3h4 g6f6');
INSERT INTO evaluations VALUES(49379,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(49379,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(49379,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(49379,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(49379,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(49379,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(49379,36,'cp 23','g1f3 d7d5 d2d4 c7c5 c4d5 c5d4 d1d4 e6d5 e2e4 b8c6 f1b5 d5e4 d4d8 e8d8 f3g5 c8e6 e1g1 f8b4 g5e4 f6e4 c3e4 d8e7 c1e3 h8d8 b5c6 b7c6 a2a3 b4d6 e4d6 e7d6 e3d2 c6c5 d2c3 f7f6 b2b4 d6c6 b4c5 d8d7 f1e1');
INSERT INTO evaluations VALUES(49377,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(49377,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(49377,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(49377,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(49377,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(49377,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(49377,36,'cp 19','d7d5 d2d4 f8b4 c4d5 e6d5 c1g5 e8g8 e2e3 h7h6 g5h4 c8f5 d1b3 b4c3 b3c3 c7c6 f3d2 b8d7 f1e2 f6e4 h4d8 e4c3 b2c3 f8d8 f2f3 a8c8 e1f2 c6c5 h1c1 c8c7 e3e4 f5e6 e4d5 e6d5 c3c4');
INSERT INTO evaluations VALUES(48237,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(48237,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(48237,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(48237,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(48237,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(48237,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(48237,36,'cp 16','d2d4 f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 c7c5 d4c5 b8d7 d1d4 d8a5 a2a3 b4c3 b2c3 e8g8 d4b4 a5c5 h4f6 c5b4 a3b4 d7f6 f3d4 b7b6 e2e3 a7a5 f2f3 c8d7 b4a5 b6a5 f1d3 a5a4 a1a2 f6e8 e1d2 e8d6 h2h4 f8c8 h1a1 d6c4 d3c4 c8c4 h4h5');
INSERT INTO evaluations VALUES(52720,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(52720,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(52720,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(52720,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(52720,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(52720,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(52720,36,'cp 112','d5c6 b8c6 b1c3 g8f6 g1f3 e7e5 f1b5 f8d6 d2d3 h7h6 e1g1 e8g8 h2h3 a7a6 b5c6 b7c6 f1e1 f8e8 c3e4 f6e4 d3e4 a6a5 c1d2 d8e7 d1e2 a5a4 d2c3 c8a6 e2d2 a6b5 e1d1 d6c7 d2d7 e7d7 d1d7 a8c8 f3h4 c6c5 d7d2 a4a3 b2b3 c5c4 h4f5 c4b3 c2b3 c7b6 c3b4 c8d8');
INSERT INTO evaluations VALUES(52692,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(52692,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(52692,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(52692,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(52692,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(52692,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(52692,36,'cp 92','b8c6 f1c4 g8f6 g1f3 c8g4 h2h3 g4h5 e1g1 d8c7 c4e2 e7e5 d2d3 a8d8 b1c3 f8e7 f3h4 h5e2 c3e2 e8g8 h4f5 f8e8 f1e1 f6d5 e2c3 e7f8 c1d2 d5c3 d2c3 g7g6 d1g4 c7d7 f5e3 f7f5 g4h4 f5f4');
INSERT INTO evaluations VALUES(19901,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(19901,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(19901,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(19901,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(19901,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(19901,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(19901,36,'cp 107','b1c3 g8f6 f1b5 e7e5 g1f3 f8d6 d2d3 h7h6 e1g1 e8g8 h2h3 c6a5 c3e4 f6e4 d3e4 c8e6 d1e2 d8c7 f3h4 f8d8 h4f5 d6f8 f5e3 a8c8 f1d1 d8d1 e2d1 f8c5 b5d3 c8d8 d1e1 a5c6 c1d2 c6d4 e3d5 e6d5 e4d5 d8d5');
INSERT INTO evaluations VALUES(19898,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(19898,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(19898,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(19898,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(19898,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(19898,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(19898,36,'cp 91','g8f6 g1f3 e7e5 f1b5 f8d6 d2d3 h7h6 h2h3 e8g8 e1g1 c6a5 c3e4 f6e4 d3e4 c8e6 b5d3 d8c7 d1e2 f8d8 f1d1 a7a6 b2b3 a5c6 c2c3 c6e7 c3c4 e7c6 c1e3 d6a3 a1b1 a8c8 f3e1 a3b4 e1c2 b4c3 e2f3 c7a5 a2a4 a5c7 f3g3 c6b4 c2b4 c3b4');
INSERT INTO evaluations VALUES(13194,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(13194,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(13194,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(13194,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(13194,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(13194,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(13194,36,'cp 87','g1f3 e7e5 f1b5 f8d6 d2d3 e8g8 b5c6 b7c6 c1g5 a8b8 d1c1 c8g4 f3d2 g4e6 d2e4 d6e7 e4f6 g7f6 g5h4 g8h8 e1g1 f8g8 g1h1 d8f8 b2b3 g8g4 h4g3 f8g7 c1e3 b8g8 c3e2 h7h5 f2f3 g4g6 f1g1 c6c5 a1e1 h5h4 g3h4 g6g2 h4g3 g2g1 e1g1 g7h7 g3f2 g8g6 e3e4 h7h3 e2g3');
INSERT INTO evaluations VALUES(13193,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(13193,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(13193,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(13193,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(13193,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(13193,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(13193,36,'cp 86','e7e5 f1b5 f8d6 d2d3 e8g8 b5c6 b7c6 c1g5 a8b8 d1c1 c8g4 f3d2 g4e6 a1b1 d8c7 g5f6 g7f6 e1g1 e5e4 g1h1 e4d3 c2d3 g8h8 c3e4 d6h2 g2g3 h2g3 e4g3 e6d5 d2e4 f8g8 c1e3 g8g6 b2b3 b8g8 e3f3');
INSERT INTO evaluations VALUES(13815,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(13815,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(13815,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(13815,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(13815,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(13815,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(13815,36,'cp 88','f1c4 e5e4 f3g5 c8g4 c4e2 g4f5 d2d3 e4d3 e2d3 f5d3 c2d3 f8b4 d1e2 c6e7 g5e4 b4c3 b2c3 f6e4 d3e4 e8g8 c1e3 d8c7 a1c1 f7f5 e1g1 f5f4 e3d4 f4f3 g2f3 e7g6 g1h1 c7f4 f1d1 f4f3 e2f3 f8f3 h1g2 f3f7 g2f1 g6f4 d4c5 a8e8 c5a7 e8e4');
INSERT INTO evaluations VALUES(17041,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(17041,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(17041,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(17041,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(17041,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(17041,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(17041,36,'cp 145','d1d4 d7d6 f2f3 f8e7 c1f4 a7a6 e1c1 e8g8 c1b1 b7b5 g2g4 c7c5 d4f2 b5b4 c3e2 a6a5 e2g3 c8e6 g3f5 d6d5 e4d5 e6f5 g4f5 e7d6 f4e3 d8e7 f1b5 a5a4 e3g5 e7e5 f2g3 a4a3 g3e5 d6e5 d5d6 f8d8 g5f6 g7f6 b5c6 a8b8 d1g1 g8h8');
INSERT INTO evaluations VALUES(51487,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51487,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(51487,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(51487,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(51487,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(51487,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(51487,36,'cp 55','e7e6 e2e4 f8b4 b1c3 c8b7 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 f1d3 d7d6 g1h3 b8c6 e1g1 g8f6 f3f4 e6e5 d4d5 c6b8 c4c5 b6c5 h3f2 b8d7 c3c4 b7c8 c1d2 a7a5 d1e2 a5a4 a1b1 h7h5 f4f5 h5h4 g3g4 f6h7 h2h3 h7g5 a2a3 f7f6 g1h1 c8a6 f1g1');
INSERT INTO evaluations VALUES(51408,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(51408,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(51408,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(51408,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(51408,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(51408,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51408,36,'cp 52','e2e4 f8b4 b1c3 c8b7 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 f1d3 d7d6 g1h3 b8c6 e1g1 e8c8 d1e2 g8f6 f3f4 e7d7 h3f2 c6a5 f1e1 h7h5 a1b1 c8b8 d4d5 c7c6 e4e5 f6g4 e5d6 d7d6 f4f5 e6e5 f2g4 h5g4 e2e5 d6e5 e1e5 c6d5 c4d5 d8d5 e5d5 b7d5 c1f4 b8a8 b1e1 d5a2');
INSERT INTO evaluations VALUES(51410,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(51410,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(51410,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(51410,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(51410,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(51410,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51410,36,'cp 30','f8b4 c1d2 b4d2 d1d2 d7d5 f1d3 c8a6 c4d5 a6d3 d2d3 e6d5 e4e5 g8e7 f2f4 c7c5 b1c3 b8c6 g1f3 e8g8 e1g1 f7f6 e5f6 f8f6 a1e1 f6f4 d4c5 b6c5 d3e3 f4f6 e3c5 d8d6 c5d6 f6d6 e1e2 a8d8 e2d2 d5d4 c3e4 d6d5 a2a3 h7h6');
INSERT INTO evaluations VALUES(48201,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48201,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(48201,36,'cp 0','d5c4 d1a4 b8d7 e2e3 a7a6 f1c4 f8e7 e1g1 e8g8 a4c2 c7c5 f1d1 h7h6 g5f6 d7f6 d4c5 d8c7 c4e2 c8d7 b1d2 f8c8 b2b4 b7b6 a1c1 b6c5 b4c5 e7c5 d2b3 c5b6 c2c7 c8c7 c1c7 b6c7 g2g3 g8f8 b3c5 d7b5 f3d4 b5e2 d4e2 f8e7 e2d4 f6d5 d1c1 c7b6 d4c6 e7d6 c5b7 d6d7');
INSERT INTO evaluations VALUES(48445,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48445,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48445,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(48445,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(48445,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48445,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(48445,36,'cp -9','d1a4 b8d7 e2e3 f8e7 f1c4 e8g8 a4c2 a7a6 e1g1 c7c5 f1d1 h7h6 g5f6 d7f6 d4c5 d8c7 c4e2 c8d7 b1d2 f8c8 b2b4 b7b6 a1c1 b6c5 b4c5 c7c5 c2c5 e7c5 d2b3 c5e7 c1c8 d7c8 d1c1 e7a3 c1c2 c8b7 f3d2 a3f8 e2f3 f6d5 c2c1 a8c8 c1c8 b7c8');
INSERT INTO evaluations VALUES(57755,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(57755,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(57755,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(57755,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(57755,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(57755,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(57755,36,'cp -72','e7e5 c2c4 g8f6 d2d3 h7h6 g2g3 d7d5 c4d5 f6d5 g1f3 b8c6 f1g2 f8c5 e1g1 e8g8 b1c3 d5f6 a2a3 a7a5 c1d2 d8e7 a1c1 f8d8 d1c2 c8g4 e2e3 e7d7 c3e4 f6e4 d3e4 c5d6 f1d1 a5a4 d2c3 d7e7 d1d5 g4e6 d5b5 e6b3');
INSERT INTO evaluations VALUES(57187,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(57187,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(57187,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(57187,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(57187,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(57187,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(57187,36,'cp -38','c2c4 g8f6 b1c3 f8b4 e2e3 b4c3 b2c3 e5e4 f2f3 d7d6 f3e4 f6e4 d1f3 f7f5 d2d3 e4f6 f1e2 b8d7 f3g3 e8g8 g1h3 d7e5 e1g1 e5g4 e2f3 a8b8 h3f4 c8d7 a2a4 h7h6 a4a5 g8h7 a5a6 b7b6 f4d5 f8e8');
INSERT INTO evaluations VALUES(19937,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(19937,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(19937,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(19937,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(19937,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(19937,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(19937,36,'cp 24','e7e5 d2d3 f8e7 f3d2 g8f6 d2f1 c8g4 f2f3 g4e6 f1e3 e8g8 e1g1 a7a6 a2a4 c6b4 c1d2 d6d5 e4d5 f6d5 c3d5 b4d5 e3d5 e6d5 d2c3 e7f6 d1e2 f8e8 a4a5 d5c6 e2f2 d8d6 a1e1 a8d8 e1e2 c6d5 h2h4 d5c4 d3c4 d6c7 h4h5 h7h6 f2g3');
INSERT INTO evaluations VALUES(4323,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(4323,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(4323,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4323,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(4323,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4323,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4323,36,'cp 14','c1g5 h7h6 g5h4 c5e7 h2h3 d7d6 a2a4 c6b4 e1g1 c8e6 a4a5 d8d7 c4e6 f7e6 d3d4 e5d4 f3d4 a8e8 h4g3 a7a6 d1e2 b4c6 d4c6 d7c6 e4e5 f6d5 c3d5 e6d5 e2d3 d6e5 g3e5 e7f6 e5f6 f8f6 a1d1 e8e5 f2f4 c6c5 g1h1 e5e4 d3d5 c5d5 d1d5 e4f4');
INSERT INTO evaluations VALUES(4279,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(4279,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4279,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4279,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4279,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(4279,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4279,36,'cp 0','h7h6 g5h4 c5e7 a2a4 d7d6 h2h3 f6e4 h4e7 e4c3 e7d8 c3d1 d8c7 d1b2 c4b5 c8d7 c7d6 f8e8 e1d2 a7a6 b5c6 d7c6 f3e5 c6g2 h1g1 g2h3 a1b1 a8d8 d6c7 d8c8 c7d6');
INSERT INTO evaluations VALUES(4141,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4141,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4141,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(4141,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(4141,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4141,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4141,36,'cp 7','g5h4 c5e7 h2h3 d7d6 a2a4 c6b4 e1g1 c8e6 f3d2 a7a6 h4g3 d8d7 c3e2 f6h5 c2c3 b4c6 b2b4 e7g5 c4e6 f7e6 g3h2 g5f4 e2g3 f4d2 d1d2 h5f4 d3d4 e5d4 c3d4 c6e7 f1d1 f8f7 f2f3 a8f8');
INSERT INTO evaluations VALUES(4157,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(4157,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(4157,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4157,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(4157,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(4157,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4157,36,'cp 12','c5e7 e1g1 d7d6 a2a4 c6b4 d3d4 f6h5 h4e7 d8e7 f3e5 d6e5 d1h5 e5d4 c3d5 b4d5 e4d5 c8d7 h5f3 e7c5 b2b3 a8e8 f3f4 c7c6 a1d1 c6d5 f4d4 c5d4 d1d4 d5c4 d4d7 e8e2 f1c1 f8e8 g1f1 c4b3 c2b3 e8e7 f2f4 e2e1 c1e1 e7d7 f4f5 d7d4 h2h3 d4b4');
INSERT INTO evaluations VALUES(4546,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4546,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4546,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4546,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(4546,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(4546,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4546,36,'cp 12','a2a4 d7d6 e1g1 g7g5 h4g3 c8g4 h2h3 g4h5 g3h2 g5g4 h3g4 h5g4 c3e2 g8h8 d1d2 f6h7 f3e1 e7g5 f2f4 e5f4 h2f4 c6e5 e2g3 f8g8 f4g5 d8g5 d2g5 h7g5 c4d5 g4e6 g3f5 e6f5 f1f5 c7c6 d5b3 a8d8 d3d4 e5g4');
INSERT INTO evaluations VALUES(22828,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(22828,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(22828,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(22828,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(22828,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(22828,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(22828,36,'cp -12','g8f6 b1c3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 c8g4 c1e3 h7h6 h2h3 g4h5 c2c4 d8f6 c4c5 d6f4 a1b1 a8d8 b5c6 b7c6 f1e1 a7a5 e3f4 f6f4 b1b7 f8e8 e1e8 d8e8 d1d2 f4d2 f3d2 e8d8 b7c7 d8d4 d2f3');
INSERT INTO evaluations VALUES(16590,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(16590,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(16590,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16590,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16590,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16590,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16590,36,'cp -1','b1c3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 e1g1 e8g8 d2d4 e5d4 c3d4 h7h6 b5d3 d8f6 c1e3 f8e8 h2h3 c8f5 c2c4 a8d8 d3f5 f6f5 d1a4 f5d7 a4c2 d6f8 f1d1 b7b6 e3f4 c6e7 f4h2 e7g6 a3a4 c7c6 f3e5 g6e5 d4e5 d7e7 h2f4 d8d1 a1d1 e7a3');
INSERT INTO evaluations VALUES(16581,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(16581,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(16581,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(16581,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16581,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(16581,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(16581,36,'cp 0','d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 e1g1 e8g8 d2d4 e5d4 c3d4 c8g4 c1e3 c6e7 h2h3 g4h5 f1e1 e7d5 e3d2 d8f6 g2g4 h5g6 c2c4 d5f4 c4c5 f4h3 g1g2 h3f4 g2g3 f4h5 g3g2');
INSERT INTO evaluations VALUES(15163,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15163,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(15163,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15163,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15163,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(15163,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(15163,36,'cp 0','e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 e8g8 e1g1 h7h6 b5d3 d8f6 c1e3 c8f5 c2c4 b7b6 h2h3 a8b8 f1e1 f8e8 d3f5 f6f5 d1b3 b6b5 c4c5 d6f8 a1d1 e8d8 b3a2 f5d5 a2d5 d8d5 e3f4 b8c8 f4g3 c6e7 g1h1');
INSERT INTO evaluations VALUES(15015,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15015,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15015,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15015,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15015,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(15015,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15015,36,'cp 0','f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 e8g8 e1g1 c8g4 c1e3 c6e7 h2h3 g4h5 f1e1 e7d5 e3d2 d8f6 g2g4 h5g6 c2c4 a7a6 b5a4 d5f4 c4c5 f4h3 g1g2 h3f4 g2g3 f4e2 g3g2 e2f4');
INSERT INTO evaluations VALUES(15364,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(15364,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(15364,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(15364,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(15364,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(15364,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(15364,36,'cp 0','f1b5 d5c3 b2c3 f8d6 e1g1 e8g8 d2d4 e5d4 c3d4 c8g4 c1e3 g4h5 b5d3 d8f6 a1b1 a8b8 b1b5 h5f3 d1f3 f6f3 g2f3 a7a6 b5b3 b7b5 f1b1 c6a5 b3b2 a5c6');
INSERT INTO evaluations VALUES(15248,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15248,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15248,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15248,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15248,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15248,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15248,36,'cp -1','d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 e8g8 e1g1 c8g4 c1e3 g4h5 b5d3 d8f6 a1b1 a8b8 b1b5 h5f3 d1f3 f6f3 g2f3 a7a6 b5b3 b7b5 f1a1 g7g6 g1g2 f8d8 f3f4 g8g7 g2f3 f7f5 c2c3 c6e7 a1b1 c7c6 c3c4 b5c4 b3b8 d8b8 b1b8 d6b8 d3c4');
INSERT INTO evaluations VALUES(14261,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14261,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14261,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(14261,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(14261,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14261,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14261,36,'cp 0','c8g4 f1b5 d8d6 h2h3 g4f3 d1f3 e8c8 b5a4 c8b8 c3b5 d6d7 b5c3');
INSERT INTO evaluations VALUES(24935,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24935,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24935,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24935,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24935,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24935,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24935,36,'cp 0','f1b5 d8d6 h2h3 g4f3 d1f3 e8c8 b5a4 c8b8 c3b5 d6b4 b5c3');
INSERT INTO evaluations VALUES(24781,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24781,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24781,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24781,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24781,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24781,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24781,36,'cp 0','d8d6 h2h3 g4f3 d1f3 e8c8 b5a4 c8b8 c3b5 d6b4 b5c3');
INSERT INTO evaluations VALUES(10068,30,'cp 247',NULL);
INSERT INTO evaluations VALUES(10068,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(10068,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(10068,33,'cp 263',NULL);
INSERT INTO evaluations VALUES(10068,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(10068,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(10068,36,'cp 150','c4f7 e8e7 f7c4 d8e8 b1c3 h7h6 g5f3 d7d6 e1g1 e7d8 c4e2 c6d4 f3d4 e5d4 c3d5 e8e4 d5f6 g7f6 d2d3 e4f5 e2f3 c7c6 f1e1 h6h5 c2c3 d8c7 e1e7 c8d7 f3e4 f5g4 d1g4 h5g4 e4f5 a8d8 c1f4 c7c8 e7d7 d8d7 f5g4 d6d5 g2g3 a7a5 a1b1 b7b5 f4d2 c8d8 g4d7 d8d7 g1g2 a5a4 h2h4 d4c3 d2c3 d5d4');
INSERT INTO evaluations VALUES(9646,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(9646,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(9646,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(9646,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(9646,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(9646,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(9646,36,'cp 123','e8e7 f7c4 d8e8 b1c3 c6d4 c3a4 d7d5 a4c5 d5c4 c2c3 d4c6 d2d4 e5d4 c3d4 h7h6 g5f3 b7b6 d4d5 b6c5 d5c6 e8c6 e1g1 h8e8 f3e5 c6e4 f1e1 e4d5 c1e3 c8e6 d1d5 e6d5 e3c5 e7d8 c5d4 d8c8 f2f3 c8b7 g1f2 e8e6 a1c1 a8e8 e1e3 c7c5 d4c3 e8b8 a2a4 e6e7 c1e1 b7a8 e5g6 e7e3');
INSERT INTO evaluations VALUES(4698,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(4698,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(4698,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(4698,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(4698,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(4698,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(4698,36,'cp 136','f7c4 d8e8 b1c3 c6d4 d2d3 d7d6 c3e2 b7b5 e2d4 b5c4 d4f5 c8f5 e4f5 g7g6 g5e4 g6f5 e4c5 e8c6 d1f3 c6c5 f3f5 c4d3 c1g5 a8f8 e1g1 h7h6 g5e3 c5c4 c2d3 c4e6 f5e6 e7e6 a1e1 h6h5 d3d4 e5e4 b2b3 e6d5 h2h3 h8g8 f2f3 e4f3 f1f3 f6e4 f3f8 g8f8 e1f1');
INSERT INTO evaluations VALUES(49784,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(49784,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(49784,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(49784,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(49784,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(49784,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(49784,36,'cp -47','b1c3 d7d5 e4d5 f6d5 f1b5 c7c6 d1e2 c8e6 c3d5 d8d5 b5a4 d5d8 a4b3 d8f6 d2d4 f8d6 b3e6 f6e6 e2e6 f7e6 e1g1 h7h6 c1d2 b8a6 a1e1 a6c7 c2c4 g7g5 h2h4 h8g8 h4g5 h6g5 g2g4 e8c8 g1g2 d6e7 f1h1 e7f6 d2c3 c6c5 d4c5 f6c3 b2c3 d8d3 h1h5 d3c3 h5g5 g8f8 g5h5');
INSERT INTO evaluations VALUES(49780,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(49780,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(49780,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(49780,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(49780,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(49780,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(49780,36,'cp -52','d7d5 e4d5 f6d5 f1b5 c7c6 d1e2 c8e6 c3d5 d8d5 b5a4 d5d8 a4b3 d8f6 b3e6 f7e6 d2d4 f8d6 c1d2 b8d7 e1c1 e8g8 f3g5 a8e8 g5e4 f6e7 e2c4 f8f5 g2g3 d6b8 d1e1 d7f6 g3f4 b8f4 d2f4 f5f4 e4c5 e7f7 e1e6 f6d5 e6e8 f7e8 c5b7 f4e4 h1d1');
INSERT INTO evaluations VALUES(48706,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(48706,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(48706,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(48706,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(48706,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(48706,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(48706,36,'cp -46','e4d5 f6d5 f1b5 c7c6 d1e2 c8e6 c3d5 d8d5 b5a4 b8a6 a4b3 d5d7 d2d4 f8d6 f3g5 a6c7 e1g1 e8g8 c1f4 d6f4 f1f4 e6b3 a2b3 f7f6 g5f3 f8e8 f4e4 e8e4 e2e4 a8e8 e4d3 c7d5 a1e1 e8e1 f3e1 d7g4 c2c3 g8f7 b3b4 h7h5 g2g3 d5f4 d3c4 f7g6 c4f1 f4h3 g1g2 h3g5 f1d3 g6h6 h2h4');
INSERT INTO evaluations VALUES(48579,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(48579,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(48579,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(48579,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(48579,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(48579,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48579,36,'cp -36','f6d5 f1b5 c7c6 d1e2 c8e6 c3d5 d8d5 b5a4 b8a6 a4b3 d5d7 d2d4 f8d6 e1g1 e8c8 b3e6 f7e6 e2e4 d8f8 f1e1 f8f5 e4e6 g7g5 c1d2 a6c7 e6d7 c8d7 c2c4 g5g4 f3e5 d6e5 e1e5 f5e5 d4e5 h8f8 d2a5 b7b6 a1d1 d7c8 a5c3 f8d8 d1d8 c8d8 h2h3 g4g3 g1f1 d8e7 f1e2 e7e6 e2f3 e6f5 b2b4 c7e6');
INSERT INTO evaluations VALUES(48988,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(48988,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(48988,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(48988,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(48988,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(48988,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(48988,36,'cp -55','f1b5 c7c6 d1e2 c8e6 c3d5 d8d5 b5a4 d5d8 a4b3 d8f6 d2d4 b8d7 b3e6 f7e6 c1d2 f8d6 e1c1 e8g8 f3g5 a8e8 g5e4 f6e7 e2c4 f8f5 g2g3 d6c7 d1e1 d7f6 g3f4 f6d5 e4g3 f5f8 f4f5 e7f7 e1e6 e8e6 f5e6 f7e6 c4d3 d5f6 b2b3 f6d5 h1e1 e6g4 c2c4 d5f4');
INSERT INTO evaluations VALUES(50323,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(50323,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(50323,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50323,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(50323,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(50323,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(50323,36,'cp 14','e7e5 f1g2 f8c5 d2d3 e8g8 g1f3 b8c6 e1g1 d7d5 c4d5 f6d5 a2a3 c5b6 b1d2 c8g4 d2c4 d8e7 c1d2 f8d8 h2h3 g4h5 f1e1 d5f6 d1c1 e5e4 d3e4 f6e4 d2e3 e4c5 e3c5 b6c5 b2b4 h5f3 g2f3');
INSERT INTO evaluations VALUES(55774,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55774,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(55774,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(55774,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(55774,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(55774,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(55774,36,'cp -17','e6d5 d2d4 g8f6 g1f3 f8d6 c3b5 d6b4 c2c3 b4a5 f1d3 e8g8 e1g1 c7c6 b5a3 a5c7 f1e1 c8g4 a3c2 g4h5 c1g5 d8d6 g2g3 h7h6 g5f4 d6d8 f4c7 d8c7 d3e2 b8d7 c2e3 f8e8 f3h4 h5e2 d1e2 d7b6 e2c2 a7a6 h4f3');
INSERT INTO evaluations VALUES(56273,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(56273,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56273,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(56273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56273,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56273,36,'cp 0','d2d4 g8f6 c1f4 f8e7 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(47354,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(47354,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(47354,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(47354,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(47354,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(47354,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(47354,36,'cp 31','e2e3 a7a6 f1c4 e7e6 e1g1 c7c5 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 c4e2 c8d7 d2b3 c5d6 b3a5 a8a7 a5c4 d7b5 c1d2 b5c4 e2c4 h8c8 f1c1 a7a8 g2g3 b8c6 a2a3 f6e4 d2e1 f7f5 f3d2 c6e5 c4e2 c8c1 a1c1 e4d2 e1d2');
INSERT INTO evaluations VALUES(54954,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(54954,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(54954,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(54954,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54954,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54954,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54954,36,'cp 0','g8f6 d2d3 e7e6 d1e2 c7c5 g1f3 f8e7 a2a4 b8c6 e1g1 e8g8 f1e1 c8d7 c2c3 h7h6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(46980,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46980,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46980,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46980,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46980,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46980,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46980,36,'cp 0','d2d3 e7e6 d1e2 f8e7 g1f3 c7c5 e1g1 b8c6 a2a4 e8g8 f1e1 c8d7 c2c3 h7h6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(46977,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46977,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46977,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46977,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46977,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(46977,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46977,36,'cp 0','e7e6 d1e2 f8e7 g1f3 e8g8 e1g1 b7b6 c2c3 b8d7 c1f4 c8b7 b1d2 f6h5 f4e3 h5f6 c4b3 c7c5 e3f4 f6h5 f4e3');
INSERT INTO evaluations VALUES(6557,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6557,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(6557,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6557,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6557,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(6557,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6557,36,'cp 0','c2c3 g8f6 c3d4 c5b4 c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 f7f8 b3b4 d8e7 b4e7 f8e7 e1g1 h8e8 f1e1 e7f8 b1a3 e4f6 a3b5 e8e1 a1e1 f6d5 f3g5 a7a6 g5h7 f8f7 b5c3 d5c3 b2c3 f7g6 h7f8 g6f7');
INSERT INTO evaluations VALUES(6553,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6553,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6553,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6553,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(6553,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6553,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6553,36,'cp 0','g8f6 c3d4 c5b4 c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 f7e8 b3b4 d8e7 b4e7 e8e7 e1g1 e7d8 f1e1 e4f6 b1c3 d7d6 f3g5 h8f8 g5e4 f6e4 e1e4 c8f5 e4e3 a7a5 f2f3 a8a6 a1e1 a6b6 b2b3 f5g6 g1f2 b6b4 e1d1 g6f5 c3d5 b4b5 d5c3');
INSERT INTO evaluations VALUES(57694,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(57694,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(57694,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(57694,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(57694,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(57694,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(57694,36,'cp 61','e2e4 d7d6 d2d4 g8f6 f1d3 f8g7 e1g1 e8g8 f1e1 c7c5 d4c5 d6c5 e4e5 f6d5 d3e4 d5c7 b1d2 b8c6 e4c6 b7c6 c2c3 c8f5 d1a4 c7e6 d2b3 d8d5 c1e3 f5d3 a1d1 f8b8 h2h3 c5c4 b3a5 d5b5 a4b5 b8b5 a5c6');
INSERT INTO evaluations VALUES(33213,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(33213,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(33213,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(33213,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(33213,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(33213,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(33213,36,'cp 166','e1g1 g8e7 b1c3 h7h6 d2d4 c5d4 f3d4 e5d4 c3b5 b8a6 f1e1 d7d6 b5d4 e8g8 a2a4 c8e6 c4f1 c7c5 d4b5 d6d5 e4d5 e7d5 b5d6 a6b4 c2c3 b4c6 d6b7 f6e7 d1d5 e6d5 e1e7 c6e7 b7c5 f8d8 b2b4 e7f5 c1f4 f5d6 a1d1 d6e4 d1d5 d8d5 c5e4 d5d1 f2f3 d1a1 g1f2 a1a4 g2g3 a4a2 f2e3 a2h2');
INSERT INTO evaluations VALUES(33212,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(33212,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(33212,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(33212,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(33212,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(33212,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(33212,36,'cp 161','g8e7 b1c3 h7h6 d2d4 c5d4 f3d4 e5d4 c3b5 b8a6 f1e1 d7d6 b5d4 e8g8 a1b1 e7g6 b2b3 a6c5 c1b2 g6e5 d4b5 f6e7 c4f1 a7a6 b5c3 e7h4 d1d2 h4g5 d2e3 g5e3 e1e3 c8e6 f2f4 e5g4 e3g3 f7f5 e4f5 e6f5 c3d5 c5e4 d5e7 g8h8 e7f5 e4g3 b2g7 h8g8 g7f8');
INSERT INTO evaluations VALUES(33479,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(33479,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(33479,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(33479,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(33479,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(33479,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(33479,36,'cp 145','b1c3 e8g8 c3a4 f6c6 b2b3 d7d6 a4c5 c6c5 a2a4 c8g4 c2c3 b8d7 h2h3 g4f3 d1f3 a7a6 a4a5 d7f6 f1e1 f8e8 c1a3 c5a7 d2d3 a7b8 a3c1 e7g6 b3b4 b8d8 c1g5 h7h6 g5f6 d8f6 f3f6 g7f6 g2g3 c7c6 a1d1 g6e7 c4b3 a8d8 d3d4 d6d5 d4e5');
INSERT INTO evaluations VALUES(19389,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(19389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19389,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(19389,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19389,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(19389,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(19389,36,'cp -5','e7e5 d2d3 g7g6 b1c3 d7d6 g2g3 g8e7 f1g2 f8g7 e1g1 e8g8 a1b1 c8e6 a2a3 a7a5 c1d2 a8b8 d1c1 b7b6 c3b5 f7f6 b2b4 a5b4 a3b4 d8d7 d2c3 h7h6 c1d2 f6f5 e4f5 e6f5 f3h4 f5g4 b4c5 b6c5 g2e4 g4h3 f1e1 g6g5 h4g2 d6d5 c4d5 e7d5');
INSERT INTO evaluations VALUES(18984,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18984,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(18984,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(18984,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(18984,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18984,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18984,36,'cp 0','b1c3 g7g6 a2a3 d7d6 b2b4 f8g7 f1e2 b7b6 e1g1 g8e7 d2d3 e8g8 b4c5 b6c5 c1d2 c6d4 a1b1 c8d7 h2h3 d4e6 f3g5 e6f4 e2g4 e7c6 g4d7 d8d7 d2f4 e5f4 c3d5 f8b8 b1b8 a8b8 g5f3 b8b2 d5f4 b2a2 d1b3 a2b2');
INSERT INTO evaluations VALUES(18978,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(18978,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18978,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(18978,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(18978,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(18978,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18978,36,'cp 0','d7d6 g2g3 g7g6 a1b1 f8g7 d2d3 g8e7 f1g2 e8g8 a2a3 a7a5 e1g1 c8e6 f3e1 f7f5 e4f5 g6f5 c1g5 d8d7 e1c2 f5f4 c3d5 a8e8 d5b6 d7d8 b6d5');
INSERT INTO evaluations VALUES(20603,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20603,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20603,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20603,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20603,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20603,36,'cp 0','g2g3 g7g6 a1b1 f8g7 d2d3 c8e6 a2a3 a7a5 f3g5 e6c8 f1g2 g8e7 e1g1 e8g8 g5f3 c8e6 f3e1 f7f5 e4f5 g6f5 c1g5 d8d7 e1c2 f5f4 c3d5 a8e8 d5b6 d7d8 b6d5');
INSERT INTO evaluations VALUES(56525,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(56525,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(56525,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(56525,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(56525,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(56525,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(56525,36,'cp 84','g1f3 b8c6 b1d2 g8e7 d2b3 e7g6 b3d4 g6e5 f3e5 c6e5 e2e3 f8c5 c1d2 c5d4 e3d4 d8d4 d2c3 d4e4 d1e2 e5d3 e1d2 e4e2 d2e2 d3f4 e2f3 f4e6 h2h4 f7f6 f1d3 e8f7 f3g3 e6c5 d3c2 c8e6 h1d1 a7a5 h4h5 g7g6 d1d4 e6f5 c2f5 g6f5 d4d5 b7b6 d5f5 c5e4 g3f4 e4c3 b2c3 h8g8 a1g1');
INSERT INTO evaluations VALUES(56524,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(56524,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(56524,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(56524,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(56524,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(56524,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(56524,36,'cp 80','b8c6 b1d2 g8e7 d2b3 e7g6 b3d4 g6e5 f3e5 c6e5 e2e3 f8c5 c1d2 c5d4 e3d4 d8d4 d2c3 d4e4 d1e2 e5d3 e1d2 e4e2 d2e2 d3f4 e2e3 f4e6 f1d3 f7f6 d3c2 e8f7 f2f3 a7a5 h2h4 h7h5 e3f2 e6c5 h1d1 c8e6 d1d4 b7b6 a1d1 a8e8 c3d2 e8e7 d2f4 h8e8 d1e1 g7g6 b2b3 e6f5 e1e7 e8e7');
INSERT INTO evaluations VALUES(21779,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(21779,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(21779,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(21779,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(21779,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(21779,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21779,36,'cp 106','a2a3 g7g6 c1g5 d8d7 g5f4 f7f6 e5e6 d7d8 b2b4 c8e6 b4b5 c6a5 f3d4 e6c4 e2e3 c4f1 e1f1 f8d6 f4d6 d8d6 d1a4 c7c5 b5c6 a5c6 b1d2 a8d8 d4c6 b7c6 d2f3 d6d5 h2h4 g8e7 h4h5 g6g5 f1g1 d8d7 h1h3 d5b5 a4g4 b5f5 g4b4 c6c5 b4c3 e8g8 e3e4');
INSERT INTO evaluations VALUES(21781,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(21781,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(21781,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(21781,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(21781,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(21781,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(21781,36,'cp 93','g7g6 c1g5 d8d7 g5f4 f7f6 e5e6 d7d8 b2b4 c8e6 b4b5 c6a5 f3d4 e6c4 e2e3 c4f1 e1f1 f8d6 f4d6 d8d6 d1a4 c7c5 b5c6 a5c6 b1d2 a8d8 d4c6 b7c6 d2f3 d6d3 f1g1 g8e7 g2g3 d3b5 a4c2 b5d3 c2a2 d3d5 g1g2 d5a2 a1a2 e8f7 h1c1 d8d3 f3d4 h8d8 c1c2 h7h5 d4c6 e7c6 c2c6');
INSERT INTO evaluations VALUES(20009,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(20009,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(20009,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(20009,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(20009,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(20009,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(20009,36,'cp 87','d2d4 c5d4 c3d4 g8f6 d4d5 c6e5 f1e2 e5f3 e2f3 g7g6 c1e3 f8g7 e1g1 e8g8 b1c3 f6d7 h2h3 d7c5 a1c1 a7a5 f1e1 g7e5 e3d4 c8d7 d4e5 d6e5 d1d2 d8b6 c1c2 a8c8 e1c1 f7f5 e4f5 d7f5 f3g4 f5g4 h3g4 b6b4 c1e1 b4d4 d2e2 d4d3');
INSERT INTO evaluations VALUES(19988,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(19988,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(19988,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(19988,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(19988,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(19988,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(19988,36,'cp 79','c5d4 c3d4 g8f6 d4d5 c6e5 f1e2 e5f3 e2f3 g7g6 e1g1 f8g7 b1c3 e8g8 h2h3 a7a6 c1e3 f6d7 a1c1 b7b5 e3d4 g7d4 d1d4 c8b7 b2b3 d8a5 f3g4 a5b6 d4d2 d7f6 f1e1 f6g4 h3g4 f8c8 g4g5 c8c7 c3e2 a8c8 c1c7 c8c7 e2d4 c7c5 b3b4 c5c8');
INSERT INTO evaluations VALUES(20086,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(20086,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(20086,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(20086,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(20086,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(20086,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(20086,36,'cp 91','c3d4 g8f6 b1c3 d6d5 e4e5 f6e4 f1e2 e4c3 b2c3 e7e6 e2d3 f8e7 e1g1 c8d7 f3e1 h7h5 a1b1 c6a5 d1f3 a8c8 f3g3 g7g6 c1d2 h5h4 g3g4 a5c4 d3c4 d5c4 b1b7 d7c6 b7b1 d8d5 d2g5 h4h3 g5e7 e8e7 g4f4 e7e8');
INSERT INTO evaluations VALUES(20081,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(20081,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(20081,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(20081,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(20081,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(20081,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(20081,36,'cp 76','g8f6 b1c3 d6d5 e4e5 f6e4 f1e2 e7e6 e1g1 e4c3 b2c3 c8d7 e2d3 f8e7 f3d2 c6a5 d2b3 a8c8 a2a4 a5b3 d1b3 b7b6 c1d2 e8g8 f1e1 h7h6 b3d1 f7f5 d1e2 d7e8 e1b1 g8h8 d3a6 c8b8 b1b2 e8g6 g2g3 e7g5 f2f4 g5e7 h2h4 f8g8');
INSERT INTO evaluations VALUES(13385,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(13385,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(13385,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(13385,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(13385,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(13385,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(13385,36,'cp 79','b1c3 d6d5 e4e5 f6e4 f1e2 e7e6 e1g1 e4c3 b2c3 c8d7 e2d3 f8e7 f3d2 c6a5 d2b3 a8c8 a2a4 a5b3 d1b3 b7b6 b3c2 h7h6 c1d2 e8g8 c2d1 f7f5 d1e2 d7e8 h2h3 g8h8 f1b1 e7g5 f2f4 g5e7 g1h2 e8g6 b1b2 d8c7 a4a5 b6a5 b2b5 a5a4 a1a4 g6e8');
INSERT INTO evaluations VALUES(13381,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(13381,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(13381,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(13381,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(13381,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(13381,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(13381,36,'cp 70','d6d5 e4e5 f6e4 f1e2 e7e6 e1g1 e4c3 b2c3 c8d7 e2d3 f8e7 f3d2 c6a5 d2b3 a8c8 a2a4 a5b3 d1b3 b7b6 c1d2 e8g8 b3d1 f7f5 d1e2 g8h8 f1b1 d7e8 d3a6 c8b8 h2h3 h7h6 a6d3 b8c8 g2g3 g7g5 g3g4 f5f4 f2f3 f8f7');
INSERT INTO evaluations VALUES(55501,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(55501,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(55501,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(55501,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(55501,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(55501,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(55501,36,'cp 13','f8e7 f1d3 e7g5 c1d2 b8d7 g1e2 g5d2 d1d2 g8e7 b1c3 d7c5 e1g1 a7a5 b2b3 e8g8 a1e1 c8d7 e2g3 c5d3 c2d3 f7f5 e4f5 e7f5 g3f5 f8f5 d3d4 e5d4 d2d4 c7c5 d5c6 d7c6 f2f3 d8f6 e1d1 a8e8 d4f6 f5f6');
INSERT INTO evaluations VALUES(9334,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(9334,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(9334,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(9334,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9334,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9334,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9334,36,'cp 13','c3a4 c5b6 a4b6 a7b6 a2a4 e8g8 e1g1 h7h6 f1e1 c8e6 c4e6 f7e6 c2c3 d8e8 b2b4 e8g6 h2h3 f8f7 c1e3 a8f8 g1h1 g6h5 f3h2 h5d1 e1d1 d6d5 f2f3 c6e7 a4a5 b6a5 b4a5 f8a8 d1b1 e7c6 b1b7');
INSERT INTO evaluations VALUES(9418,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9418,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(9418,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(9418,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9418,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(9418,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(9418,36,'cp 9','c5b6 c2c3 h7h6 e1g1 e8g8 a4b6 a7b6 a2a4 c6e7 h2h3 e7g6 f1e1 f8e8 c1e3 d6d5 e4d5 f6d5 d3d4 e5d4 e3d4 c8e6 d1c2 d5f4 c4e6 f4e6 b2b3 e6d4 e1e8 d8e8 c3d4 e8c6 a1c1 c6c2 c1c2 c7c6 h3h4');
INSERT INTO evaluations VALUES(26542,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(26542,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26542,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(26542,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(26542,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(26542,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(26542,36,'cp 54','c2c4 g8f6 e2e3 e7e6 b1c3 f8b4 a1c1 e8g8 a2a3 b4c3 c1c3 c6e7 f1e2 f6e4 c3c1 e7g6 f4g3 c7c6 f3d2 d8a5 f2f3 e4g3 h2g3 e6e5 d4e5 g6e5 c4d5 f8e8 c1c3 c6d5 d2b3 a5b6 b3d4 f5g6 d1b3 e5c4 b3b6 a7b6 e2c4 d5c4');
INSERT INTO evaluations VALUES(26522,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(26522,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(26522,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(26522,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(26522,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(26522,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(26522,36,'cp 50','g8f6 b1c3 e7e6 e2e3 f8b4 a1c1 e8g8 a2a3 b4c3 c1c3 c6e7 f3d2 a8c8 f1e2 c7c5 d4c5 c8c5 b2b4 c5c6 e1g1 d5c4 c3c4 f5d3 c4c6 e7c6 d2c4 d3c4 e2c4 d8d1 f1d1 f8c8 c4f1 f6d5 f4g3 f7f5 d1c1 g8f7 f1d3 c8d8 d3e2 g7g5 b4b5 c6a5 g3e5 a5b3 e2h5 f7e7');
INSERT INTO evaluations VALUES(24800,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(24800,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(24800,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(24800,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(24800,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(24800,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(24800,36,'cp 41','e2e3 e7e6 c4c5 f6e4 h2h3 f8e7 f1b5 e8g8 e1g1 e7f6 a2a3 a7a6 b5c6 b7c6 f3e1 f6g5 f4h2 d8e7 d1a4 a8b8 b2b4 a6a5 a4c6 a5b4 a3b4 b8b4 e1f3 b4b2 c6c7 e7c7 h2c7 g5f6 f3e5 f6e5');
INSERT INTO evaluations VALUES(24799,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(24799,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(24799,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(24799,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(24799,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(24799,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(24799,36,'cp 52','e7e6 c4c5 f6e4 h2h3 f8e7 f1b5 e8g8 e1g1 e7f6 b1d2 e4d2 d1d2 a7a5 d2c3 g7g5 f4h2 d8e7 b5d3 g5g4 d3f5 g4f3 f5c2 e6e5 d4e5 f6e5 c3d3 e5h2 g1h2 e7e5 g2g3 f7f5 a1d1 a8d8 d3b3 g8g7 a2a3 d5d4 b3b7 f8f6 b7b3 e5c5 e3d4');
INSERT INTO evaluations VALUES(25139,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(25139,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(25139,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25139,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(25139,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(25139,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25139,36,'cp 43','a2a3 f8d6 f4g5 e8g8 b1c3 d6e7 a1c1 h7h6 g5f4 e7d6 f4g3 d6g3 h2g3 a7a5 f1d3 c6e7 e1g1 c7c6 f3e5 f8e8 d3c2 f5c2 d1c2 e7f5 g3g4 f5d6 c4c5 d6e4 c3e4 f6e4 e5c4 d8e7');
INSERT INTO evaluations VALUES(25140,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(25140,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(25140,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(25140,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25140,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(25140,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(25140,36,'cp 33','f8d6 f4g5 d6e7 b1c3 e8g8 c4d5 e6d5 f1d3 f6e4 g5e7 c6e7 d3e4 f5e4 c3e4 d5e4 f3d2 f8e8 d2e4 e7f5 e4c3 d8d4 e1g1 d4d1 f1d1 a8d8 b2b4 h7h5 a1c1 g8f8 g1f1 c7c6 d1d8 e8d8 c3e4 f8e7 f1e2 d8d5');
INSERT INTO evaluations VALUES(23423,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(23423,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(23423,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(23423,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(23423,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(23423,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(23423,36,'cp -21','e7e6 d2d4 d7d5 e4d5 e6d5 c4b3 g8f6 g1e2 c6a5 e1g1 f8d6 f1e1 e8g8 c1f4 f8e8 f4d6 d8d6 h2h3 c8f5 b1d2 e8e7 e2g3 e7e1 d1e1 a8e8 e1d1 f5g6 c2c3 a7a6 d2f3 c7c6 f3e5 c6c5 d1d2 f6e4 g3e4 g6e4 f2f3 a5b3 a2b3');
INSERT INTO evaluations VALUES(22582,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(22582,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(22582,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(22582,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(22582,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(22582,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(22582,36,'cp -21','c4b3 d7d5 e4d5 e6d5 g1e2 c6a5 e1g1 g8f6 d2d4 f8d6 c1f4 e8g8 f4d6 d8d6 h2h3 f8e8 f1e1 a5b3 a2b3 c8f5 b1d2 h7h6 c2c3 a7a5 e2g3 e8e1 d1e1 f5g6 e1c1 a8a6 c1d1 d6d7 g3e2 g6h5 d2f1 d7b5 g2g4 h5g6 e2f4 b5b6 f1e3 c7c6 g1g2 g6h7');
INSERT INTO evaluations VALUES(34433,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(34433,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(34433,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(34433,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(34433,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(34433,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(34433,36,'cp 36','a2a3 b8c6 g1f3 g8h6 f1d3 c5d4 c3d4 c8d7 d3c2 h6f5 c2f5 e6f5 b1c3 d7e6 b2b4 f8e7 e1g1 h7h6 h2h4 a8c8 c3e2 e8g8 e2f4 a7a6 c1e3 c6a7 f4d3 c8c4 d3b2 c4c8 d1d2 a7b5 b2a4 b6a7 a4c5 b7b6 a3a4 b6c5 a4b5 c5b4 a1a6');
INSERT INTO evaluations VALUES(34439,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(34439,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(34439,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(34439,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34439,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(34439,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(34439,36,'cp 27','b8c6 g1f3 g8h6 f1d3 c5d4 c3d4 h6f5 d3f5 e6f5 b1c3 c8e6 b2b4 h7h6 h2h4 f8e7 e1g1 e8g8 d1d3 a7a6 c1d2 c6a7 f1b1 f8c8 a3a4 b6d8 h4h5 d8d7 b4b5 a6b5 c3b5 b7b6 b5a7 d7a7 a4a5 b6a5 a1a5 a7c7 a5a8 c8a8 d3b5 c7c8 b5b7');
INSERT INTO evaluations VALUES(1488,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(1488,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(1488,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(1488,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(1488,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(1488,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(1488,36,'cp 21','g1f3 g8h6 f1d3 c5d4 c3d4 c8d7 d3c2 h6f5 c2f5 e6f5 e1g1 d7e6 b1c3 c6a5 c3e2 h7h6 f3d2 f8e7 b2b4 a5c6 d2b3 a7a5 b3a5 c6a5 b4a5 b6a5 c1d2 a5a4 d1a4 a8a4 d2b4 e7b4 a3b4 a4a1 f1a1 e8d7 h2h4 g7g5 h4g5 h6g5');
INSERT INTO evaluations VALUES(57595,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(57595,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(57595,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(57595,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(57595,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(57595,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(57595,36,'cp 201','e4f5 g8f6 d2d4 d7d5 f1d3 c7c5 c2c3 c5c4 d3c2 e7e6 f5e6 f8d6 d1e2 d8e7 g1f3 c8e6 e1g1 e8g8 f1e1 f8e8 f3e5 b8c6 f2f4 e6g4 e2f2 e7e6 b1d2 g4f5 c2f5 e6f5 d2f1 f6e4 f2f3 e8f8 e5c4 d6f4 c1f4 f5f4 f3f4 f8f4 c4e3 c6e7 e1e2 a8f8 f1g3 f4f2 a1e1 f2e2 e1e2 e4g3 h2g3 h7h5 a2a3 f8e8 g3g4 h5g4');
INSERT INTO evaluations VALUES(57571,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(57571,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(57571,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(57571,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(57571,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(57571,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(57571,36,'cp 206','g8f6 d2d4 d7d5 f1d3 c7c5 c2c3 c5c4 d3c2 e7e6 d1e2 f8d6 f5e6 d8e7 g1f3 e8g8 e1g1 c8e6 f1e1 f8e8 f3e5 e6g4 e2e3 e7e6 b1d2 g4f5 c2f5 e6f5 f2f4 b8c6 d2f1 b7b5 e3f3 f5h5 g2g4 h5h4 e1e2 d6e5 f4e5 h4g4 f3g4 f6g4 e2g2 c6e5 d4e5 g4e5 c1e3 b5b4 c3b4 e5c6 a1e1 c6b4');
INSERT INTO evaluations VALUES(50074,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(50074,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(50074,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(50074,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(50074,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(50074,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(50074,36,'cp 184','d2d4 d7d5 f1d3 c7c5 c2c3 c5c4 d3c2 e7e6 f5e6 f8d6 g1f3 e8g8 d1e2 d8e7 e1g1 c8e6 f3g5 e6d7 e2e7 d6e7 f1e1 f8e8 g5f3 e7d6 e1e8 d7e8 c1g5 b7b5 a2a3 b8d7 b1d2 e8f7 g5f6 d7f6 f3e5 a8e8 f2f4 a7a5 d2f1 f7g6 c2g6 h7g6 f1e3 g6g5 f4g5 f6e4');
INSERT INTO evaluations VALUES(50071,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(50071,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(50071,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(50071,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(50071,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(50071,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(50071,36,'cp 178','d7d5 f1d3 c7c5 c2c3 c5c4 d3c2 e7e6 d1e2 f8d6 g1f3 d8e7 f5e6 e8g8 e1g1 c8e6 f3g5 e6d7 e2e7 d6e7 f1e1 f8e8 c1f4 f6h5 f4d2 g7g6 h2h3 b8c6 g2g4 h5g7 b1a3 e7d6 a3b5 d6b8 e1e8 g7e8 a1e1 a7a6 b5a3 b8d6 c2d1 e8f6 a3c2 a8e8 g5f3 e8e1 c2e1 d6b8 e1g2 f6e4 d2e3');
INSERT INTO evaluations VALUES(46585,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(46585,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(46585,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(46585,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(46585,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(46585,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(46585,36,'cp 191','f1d3 c7c5 c2c3 c5c4 d3c2 e7e6 d1e2 f8d6 g1f3 e8g8 f5e6 d8e7 e1g1 c8e6 f3g5 e6d7 e2e7 d6e7 f1e1 f8e8 g5f3 e7d6 e1e8 d7e8 f3h4 b7b5 c1g5 b8d7 h4f5 d6f8 f5e3 d7b6 a2a3 f6e4 g5f4 e8f7 f2f3 e4f6 b1d2 f7g6 c2g6 h7g6 d2f1 b6a4 a1b1');
INSERT INTO evaluations VALUES(46583,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(46583,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(46583,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(46583,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(46583,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(46583,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(46583,36,'cp 183','c7c5 c2c3 c5c4 d3c2 e7e6 d1e2 f8d6 f5e6 d8e7 g1f3 c8e6 e1g1 e8g8 f1e1 f8e8 f3e5 e6g4 e2e3 b8c6 f2f4 e7e6 b1d2 g4f5 c2f5 e6f5 d2f1 a8d8 f1g3 f5c8 c1d2 c6e7 e3f3 e7g6 e1e2 g6h4 f3f1 h4f5 a1e1 f5g3 h2g3 f6e4 f1f3 e8f8 g3g4 d8e8 f3h3');
INSERT INTO evaluations VALUES(55600,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55600,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(55600,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55600,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(55600,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(55600,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55600,36,'cp 3','g8f6 d1c2 d6d5 d2d4 f6e4 f3e5 f8d6 f1d3 c7c5 d3e4 d5e4 c2e4 e8g8 e1g1 f8e8 e4d5 c8e6 d5b7 b8d7 c1f4 a8b8 b7a6 b8b6 a6a7 b6b2 e5d3 d6f4 d3f4 e6c4 f1d1 b2f2 b1d2 f2f4 d2c4 e8e2 d1d2 e2d2 c4d2 c5d4 c3d4 d7b6 d2f3 f4d4 a1e1 d4d7 a7a5 h7h6 a5b4');
INSERT INTO evaluations VALUES(47876,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(47876,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(47876,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(47876,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(47876,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(47876,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47876,36,'cp 25','d1c2 d6d5 d2d4 f6e4 f3e5 f8d6 f1d3 e8g8 e1g1 b8d7 c1f4 f8e8 b1d2 d6e5 d4e5 e4d2 d3h7 g8h8 f4d2 d7e5 h7d3 c8d7 a1e1 d8f6 d2e3 e5d3 c2d3 f6d6 e1d1 c7c6 c3c4 d7e6 e3d4 d6f4 d3c3 f7f6 c3b4 f4f5 b4b7 d5c4 f1e1');
INSERT INTO evaluations VALUES(47878,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(47878,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(47878,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(47878,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(47878,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47878,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47878,36,'cp 16','d6d5 d2d4 f6e4 f3e5 f8d6 f1d3 e8g8 e1g1 b8d7 c1f4 f8e8 b1d2 d6e5 d4e5 e4d2 d3h7 g8h8 f4d2 d7e5 h7d3 c8d7 d2e3 d8f6 f1e1 h8g8 f2f3 e5d3 c2d3 c7c6 d3d2 a7a5 a2a4 f6g6 h2h4 e8e6 e3f2 a8e8 c3c4 g6f6 c4c5 e6e1 a1e1 e8e1 f2e1 d7e6 e1f2 f6g6');
INSERT INTO evaluations VALUES(10220,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(10220,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10220,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(10220,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10220,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(10220,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(10220,36,'cp 19','c3a4 c5b4 c2c3 b4d6 c3c4 a7a6 a2a3 e8g8 f1e2 b7b5 a4c3 c6d4 c4b5 d4b5 e1g1 f8e8 d1c2 c7c5 c3b5 a6b5 c1e3 c8a6 f1c1 d6f8 c2d1 d7d6 b2b4 f6g4 e3g5 g4f6 h2h3 d8e7 g5e3 a8c8 d1b3 e7d7 e3g5 d7e7 a3a4 b5a4');
INSERT INTO evaluations VALUES(17475,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17475,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(17475,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17475,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(17475,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(17475,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(17475,36,'cp 79','b5c6 b7c6 e1g1 e7e6 b2b3 g8e7 c1a3 e7g6 d2d4 c5d4 a3f8 h8f8 d1d4 f7f6 c2c4 e6e5 d4e3 e8f7 f1d1 d7d6 b1c3 c8e6 d1d2 d8c7 a1d1 a8d8 f3e1 g6f4 e1d3 f4g6 d3b2 g6f4 f2f3 h7h5 g1h1 g7g5 h2h3 f7g7 b2d3 f4g6');
INSERT INTO evaluations VALUES(17472,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17472,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(17472,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17472,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(17472,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(17472,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17472,36,'cp 71','b7c6 e1g1 e7e6 b2b3 g8e7 c1a3 e7g6 d2d4 c5d4 a3f8 h8f8 d1d4 f7f6 f1d1 d7d5 e4d5 c6d5 c2c4 c8b7 d4e3 d8e7 h2h4 h7h5 c4d5 e6d5 e3d4 e8f7 b1c3 f8d8 a1c1 a8c8 d1e1 e7d6 c3e2 c8c1 e1c1 d8c8 d4a7 c8c1 e2c1');
INSERT INTO evaluations VALUES(18186,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(18186,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(18186,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(18186,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(18186,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(18186,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(18186,36,'cp 68','e1g1 e7e6 b2b3 g8e7 c1a3 e7g6 d2d4 c5d4 a3f8 h8f8 d1d4 f7f6 c2c4 e6e5 d4e3 d7d6 f1d1 e8f7 b1c3 c8e6 d1d2 d8c7 a1d1 f8d8 h2h3 h7h6 g1h2 f7g8 f3g1 e6f7 g1e2 g6e7 e2g3 g8h8 g3f5 e7f5 e4f5 a8b8 h2h1 a6a5 f2f3 d6d5 c4d5 c6d5 c3d5 f7d5');
INSERT INTO evaluations VALUES(18185,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(18185,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(18185,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(18185,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(18185,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(18185,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(18185,36,'cp 63','e7e6 b2b3 g8e7 c1a3 e7g6 d2d4 c5d4 a3f8 h8f8 d1d4 f7f6 f1d1 d7d5 e4d5 c6d5 c2c4 c8b7 d4e3 d8e7 h2h4 h7h5 c4d5 e6d5 e3d3 g6e5 d3e2 e8f7 b1c3 e5g6 e2e7 g6e7 f3d4 a8c8 a1c1 f8d8 c3e2 c8c1 d1c1');
INSERT INTO evaluations VALUES(18148,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(18148,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(18148,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(18148,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(18148,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(18148,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(18148,36,'cp 76','b2b3 g8e7 c1a3 e7g6 d2d4 c5d4 a3f8 h8f8 d1d4 f7f6 c2c4 e6e5 d4e3 d7d6 b1c3 e8f7 a1d1 c8e6 d1d2 d8c7 f1d1 f8d8 h2h3 h7h6 f3e1 f7g8 e3g3 g6e7 e1c2 d8d7 c3a4 a8f8 f2f4 d7d8 f4f5 e6f7 g3c3 f7h5 g2g4 h5f7 c2e3 g8h8 c3c2 a6a5');
INSERT INTO evaluations VALUES(56263,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(56263,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(56263,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(56263,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(56263,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(56263,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(56263,36,'cp -34','g8f6 d2d4 a7a5 a2a4 f8d6 g1f3 h7h6 e1g1 e8g8 f1e1 f8e8 e1e8 d8e8 b1a3 c8g4 c1e3 c7c6 c2c4 g4e6 c4d5 f6d5 a3c4 d6c7 e3d2 e8d8 a1c1 b8d7 d1c2 d8f6 c1e1 f6f5 c2f5 e6f5 c4e3 d5e3 e1e3 g8f8 d4d5 d7c5 d5c6 c5b3 e3b3');
INSERT INTO evaluations VALUES(48717,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(48717,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(48717,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(48717,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(48717,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(48717,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(48717,36,'cp -22','d2d4 a7a5 a2a4 f8d6 g1f3 e8g8 e1g1 c7c6 c2c4 d5c4 b3c4 b8a6 b1c3 a6c7 c1g5 c8e6 d1d3 f8e8 g5h4 h7h6 c4e6 c7e6 f1e1 d6b4 h2h3 e6f4 d3f5 f4d5 e1e8 d8e8 f5c2 e8f8 h4g3 a8d8 a1e1 d8e8 e1e8 f8e8 g3e5 b4f8 e5f6');
INSERT INTO evaluations VALUES(48652,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(48652,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(48652,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(48652,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(48652,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(48652,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(48652,36,'cp -36','f8d6 g1f3 e8g8 e1g1 h7h6 c2c4 d5c4 b3c4 b8d7 b1c3 f8e8 f1e1 d7b6 c4b3 c8g4 e1e8 d8e8 d1d3 g4e6 b3e6 e8e6 c1d2 c7c6 a1e1 e6c4 d3c2 b6d5 b2b3 c4a6 a2a4 d5c7 f3e5 a8d8 e5c4 d6b4 c3e4 f6d5 d2b4');
INSERT INTO evaluations VALUES(41696,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(41696,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(41696,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(41696,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(41696,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(41696,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(41696,36,'cp -25','g1f3 e8g8 e1g1 h7h6 c2c4 d5c4 b3c4 b8d7 c4b3 f8e8 h2h3 c7c6 b1c3 a7a5 c1d2 d7f8 d4d5 a5a4 b3c2 a4a3 b2a3 c6d5 c3b5 d6a3 b5a3 a8a3 f1e1 e8e1 d2e1 d8e7 f3d4 c8d7 a1b1 f8e6 d4e6 e7e6');
INSERT INTO evaluations VALUES(41695,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(41695,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(41695,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(41695,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(41695,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(41695,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(41695,36,'cp -31','a7a5 a2a4 e8g8 e1g1 h7h6 f1e1 c7c6 f3e5 f8e8 c1f4 d8c7 c2c4 d5c4 b3c4 c8e6 b1d2 b8d7 f4g3 d7b6 c4e6 e8e6 e5d3 e6e1 d1e1 d6g3 h2g3 g7g6 d2f3 g8g7 e1d2 b6d5 a1e1 c7b6 d3c5 b6b4 d2b4 a5b4');
INSERT INTO evaluations VALUES(52270,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(52270,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52270,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(52270,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(52270,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(52270,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(52270,36,'cp 31','f8g7 c2c3 e7e6 d2d4 c5d4 e1g1 g8e7 c4b3 d7d5 e4d5 e7d5 c1g5 d8d7 c3d4 e8g8 f1e1 b8c6 b1c3 h7h6 b3d5 e6d5 g5f4 d7d8 d1d2 g6g5 f4g3 c8g4 f3e5 c6e5 d4e5 d5d4 c3b5 d8b6 b5d6 a8d8 h2h4 f7f6 h4g5');
INSERT INTO evaluations VALUES(37768,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(37768,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(37768,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(37768,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(37768,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(37768,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(37768,36,'cp 28','c2c3 e7e6 d2d4 c5d4 e1g1 g8e7 c4b3 d7d5 e4d5 e7d5 c1g5 d8a5 c3d4 e8g8 b1c3 d5c3 b2c3 b7b6 h2h4 b8d7 d4d5 d7c5 g5e7 f8e8 e7c5 a5c5 f3g5 e8d8 d1f3 d8f8 a1d1 c5c3 d5e6 c8e6 f3c3 g7c3 b3e6 f7e6 d1d7 a8e8 d7h7 c3g7');
INSERT INTO evaluations VALUES(37762,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(37762,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(37762,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(37762,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(37762,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(37762,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(37762,36,'cp 20','e7e6 d2d4 c5d4 e1g1 g8e7 c4b3 d7d5 e4d5 e7d5 c1g5 d8a5 f3d4 e8g8 g5h4 a5b6 b1d2 b8c6 d2c4 b6c5 d4c6 c5c6 h4g3 b7b5 c4e5 c6b6 f1e1 d5f6 a2a4 a7a5 d1d2 b5a4 b3a4 c8b7 e5c4 b6a6 c4d6 f8d8 g3e5 f6g4');
INSERT INTO evaluations VALUES(37619,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(37619,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(37619,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(37619,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(37619,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(37619,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(37619,36,'cp 32','d2d4 c5d4 e1g1 g8e7 c4b3 d7d5 e4d5 e7d5 c1g5 d8a5 f3d4 e8g8 g5h4 a5b6 f1e1 b8c6 d4f3 c6e7 h4g3 a7a5 a2a4 e7f5 g3e5 c8d7 b1d2 g7e5 f3e5 d7c6 b3d5 e6d5 d1b3 b6b3 d2b3 f7f6 e5c6 b7c6 b3c5 f8b8 a1b1 d5d4 g2g3 d4c3 b2c3 b8b1 e1b1 f5d6 b1b6 a8c8');
INSERT INTO evaluations VALUES(37628,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(37628,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(37628,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(37628,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(37628,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37628,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(37628,36,'cp 24','c5d4 c4b3 d7d5 e4d5 e6d5 f3d4 b8c6 c1e3 g8f6 e1g1 e8g8 h2h3 f8e8 b1a3 c8d7 a3c2 a7a6 f1e1 a8c8 e3g5 d8b6 e1e8 c8e8 d4f3 d7e6 g5e3 b6c7 c2d4 e6d7 d4c6 d7c6 a2a4 a6a5 e3d4 f6e4 d4g7 g8g7 b3d5');
INSERT INTO evaluations VALUES(37634,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(37634,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37634,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(37634,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(37634,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(37634,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(37634,36,'cp 34','c4b3 d7d5 e4d5 e6d5 f3d4 b8c6 c1e3 g8f6 e1g1 e8g8 h2h3 a7a6 f1e1 c8d7 b1a3 f8e8 a3c2 a8c8 e3g5 e8e1 d1e1 d8b6 a1d1 c8e8 g5e3 b6c7 e1d2 c6a5 e3f4 a5b3 a2b3 c7b6 f4e3 b6d6 c2b4 f6e4 d2c1 h7h5 d4f3 d7c6 e3d4 e4c5 b4c6 b7c6');
INSERT INTO evaluations VALUES(54829,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(54829,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(54829,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(54829,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(54829,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(54829,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(54829,36,'cp 61','f8g7 f1d3 b8c6 c2c3 e7e5 h2h3 h7h6 a2a4 g8e7 e1g1 e8g8 d4e5 d6e5 b2b4 g6g5 c1e3 e7g6 f1e1 c8e6 d3f1 g6f4 d1c2 d8f6 e3f4 g5f4 b1d2 c6e7 f1c4 e6d7 a4a5 a8d8 a1d1 b7b5 a5b6 c7b6 b4b5');
INSERT INTO evaluations VALUES(15552,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(15552,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(15552,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(15552,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(15552,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(15552,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(15552,36,'cp 76','f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 d8d7 h2h3 f8b8 b2b3 d6d5 e4e5 e7b4 e3d4 b4c3 d4c3 f6e4 a2a4 d7e7 f3d3 f7f6 e5f6 g7f6 c3b2 g8h8 g2g4 b8g8 a1d1 a8e8 g1h2 e7g7');
INSERT INTO evaluations VALUES(54962,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54962,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(54962,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(54962,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54962,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(54962,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(54962,36,'cp 17','g8f6 g1f3 c7c5 c2c3 b8d7 h2h3 g7g6 e2e3 f8g7 f1e2 b7b6 b1d2 e8g8 e1g1 f8e8 f4h2 c8b7 a2a4 a8c8 d1b3 a7a6 e3e4 c5d4 c3d4 e7e5 e2c4 d8e7 d4e5 d7e5 h2e5 d6e5 b3b6 f6e4 c4a6 c8d8 d2e4 b7e4');
INSERT INTO evaluations VALUES(19236,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(19236,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(19236,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(19236,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(19236,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(19236,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(19236,36,'cp 51','c2c3 e7e6 e1g1 g8e7 d2d4 c5d4 f3d4 f8g7 d4c6 b7c6 d1d6 c8b7 f1e1 e7c8 d6g3 e8g8 e4e5 d7d6 c1f4 d6e5 f4e5 g7e5 g3e5 d8d6 e5g5 f8d8 c4f1 c6c5 b1a3 d6e7 g5e3 c8b6 a1d1 a8c8 b2b3 d8d1 e1d1 b6d5 e3e5 b7c6 a3c4 e7c7 d1c1 c7e5 c4e5 g8g7 c3c4');
INSERT INTO evaluations VALUES(19230,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(19230,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(19230,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(19230,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(19230,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(19230,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(19230,36,'cp 41','e7e6 c4b5 d7d5 e4e5 c8d7 b5c6 d7c6 d2d4 c5d4 f3d4 f8g7 d1e2 d8c7 f2f4 g8h6 e1g1 c6d7 b1d2 e8g8 d2f3 f7f6 a2a4 a8e8 f1e1 a7a6 c1e3 h6g4 e3d2 g4h6 a4a5 h6f5 e5f6 g7f6 d4f5 e6f5 d2e3 d7c6 e2d2 e8e4 g2g3 c7d7 e3d4 f6d4 f3d4 f8e8 d4f3 e4e1 a1e1 e8e1 d2e1');
INSERT INTO evaluations VALUES(18619,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(18619,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(18619,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(18619,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(18619,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(18619,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(18619,36,'cp 48','c4b5 d7d5 e4e5 c8d7 b5c6 d7c6 d2d4 c5d4 f3d4 d8c7 d1e2 f8g7 f2f4 g8h6 b1d2 c6d7 d2f3 e8g8 a2a4 a8e8 e1g1 f7f6 f1e1 h6f5 d4f5 e6f5 e2f2 f6e5 f3e5 a7a6 c1e3 g7e5 e3b6 c7c6 e1e5 e8e5 f4e5 f5f4 h2h3 c6e6 b6d4 g6g5 a4a5');
INSERT INTO evaluations VALUES(35052,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(35052,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(35052,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(35052,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(35052,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(35052,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(35052,36,'cp 170','f1d3 h4d8 g1f3 a7a6 c2c4 c7c5 d4d5 d7d6 b1c3 e6e5 h2h3 b8d7 c1e3 f8e7 a2a4 g8f6 g2g3 h7h5 f3h4 f6h7 h4f5 e7f8 d1d2 g7g6 f5h4 f8e7 a4a5 e7h4 g3h4 d8h4 e1c1 b7b6 d3c2 b6a5 c2a4');
INSERT INTO evaluations VALUES(35047,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(35047,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(35047,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(35047,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(35047,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(35047,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(35047,36,'cp 160','h4d8 b1c3 a7a6 d1e2 d7d6 g1f3 b8d7 e4e5 g8e7 e1g1 d6d5 f1e1 h7h6 c3a4 e7c6 a2a3 b7b5 a4c3 c8b7 e2d1 c6a5 c3e2 c7c5 e2f4 f8e7 c2c3 c5c4 d3c2 d8c7 h2h4 e8c8 h4h5 d7b6 f3h2 c8b8 c1e3 d8g8 h2g4 c7d7 f2f3 b8c7 d1e2');
INSERT INTO evaluations VALUES(37863,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(37863,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(37863,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(37863,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(37863,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(37863,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(37863,36,'cp 51','c2c4 b8c6 c1e3 d8b6 d4b3 b6d8 d1d2 g8f6 f2f3 d7d6 f1e2 c8e6 b1c3 e8g8 e1g1 f6d7 f3f4 d7c5 b3d4 c6d4 e3d4 d8b6 a1d1 g7d4 d2d4 f7f6 b2b3 c5d7 d4b6 a7b6 h2h4 d7c5 g2g4 a8a3 g1f2 h7h6 f2f3 e6f7 f1h1 g8g7 g4g5 f6g5');
INSERT INTO evaluations VALUES(37853,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(37853,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(37853,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(37853,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(37853,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(37853,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(37853,36,'cp 57','b8c6 c1e3 d7d6 b1c3 c6d4 e3d4 g7d4 d1d4 g8f6 f1e2 e8g8 h2h3 c8d7 e1c1 d8a5 g2g4 a8c8 c1b1 d7c6 f2f3 a7a6 d4d2 a5e5 h1e1 f6d7 h3h4 d7c5 h4h5 e5f6 b1a1 c5e6 c3d5 c6d5 e4d5 e6f4 e2f1');
INSERT INTO evaluations VALUES(5394,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(5394,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(5394,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(5394,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(5394,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(5394,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(5394,36,'cp 59','c1e3 d7d6 b1c3 g7d4 e3d4 c6d4 d1d4 g8f6 f1e2 e8g8 e1g1 c8d7 a2a3 d8b6 d4b6 a7b6 a3a4 h7h5 b2b3 h5h4 f2f3 g8g7 g1f2 g6g5 f2e3 d7e6 f3f4 g5f4 f1f4 f8h8 a1f1 f6d7 c3d5 e6d5 e4d5 d7e5 e3d4 a8c8 f1e1 f7f6');
INSERT INTO evaluations VALUES(5393,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(5393,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(5393,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(5393,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(5393,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(5393,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(5393,36,'cp 67','d7d6 f1e2 g7d4 e3d4 c6d4 d1d4 g8f6 b1c3 e8g8 e1g1 c8d7 a2a3 d8b6 d4b6 a7b6 b2b3 h7h5 h2h4 f8d8 a3a4 g8g7 f2f3 d7e6 g1f2 f6d7 f2e3 f7f6 f3f4 d8g8 f1b1 g7f7 b1h1 e6g4 e2g4 h5g4 h4h5 g6h5 h1h5 g8h8');
INSERT INTO evaluations VALUES(5533,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(5533,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(5533,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(5533,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(5533,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(5533,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(5533,36,'cp 58','b1c3 d8a5 f1e2 g7d4 e3d4 c6d4 d1d4 g8f6 e1g1 e8g8 b2b3 c8d7 h2h3 a7a6 c3d5 f6d5 e4d5 a8c8 a2a4 f8e8 f1d1 h7h5 d4b2 a5b4 a4a5 c8c5 b2d2 b4d2 d1d2 e7e5 d5e6 e8e6 f2f3 d7c6 g1f2 c5e5 b3b4 e5e3 a1c1 h5h4 c1d1 g8g7');
INSERT INTO evaluations VALUES(5531,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(5531,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(5531,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(5531,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(5531,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(5531,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(5531,36,'cp 58','c6d4 e3d4 g7d4 d1d4 g8f6 f1e2 d8a5 e1g1 c8e6 f1e1 e8g8 b2b3 a5e5 d4e3 a8c8 f2f4 e5c5 e2f3 c5e3 e1e3 c8c5 e4e5 d6e5 e3e5 c5e5 f4e5 f6d7 f3b7 d7e5 c3b5 f8b8 b7e4 a7a6 b5d4 e6d7 d4f3 e5c6 c4c5 e7e6');
INSERT INTO evaluations VALUES(18761,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18761,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(18761,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18761,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18761,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(18761,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18761,36,'cp 0','c5d4 f3d4 f8c5 d4b3 c5e7 f2f4 g8f6 e4e5 f6e4 b1d2 d7d5 c4d3 f7f5 d2f3 a7a5 a2a4 d8b6 d1e2 c8d7 c1e3 b6b4 e3d2 b4b6 d2e3');
INSERT INTO evaluations VALUES(18891,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18891,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18891,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18891,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18891,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(18891,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18891,36,'cp 0','f3d4 f8c5 d4b3 c5e7 b1c3 g8f6 c1e3 e7b4 e1g1 b4c3 b2c3 f6e4 d1g4 e8g8 g4e4 d7d5 c4d5 d8d5 e4d5 e6d5 a2a4 c8f5 f1c1 f8e8 h2h3 h7h6 a4a5 e8e7 b3d4 c6d4 e3d4 f5d7 c1e1 e7e1 a1e1 a7a6 e1e7 d7c6 e7c7 a8e8 g1f1 c6b5 f1g1');
INSERT INTO evaluations VALUES(18885,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18885,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18885,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(18885,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18885,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18885,36,'cp 0','f8c5 d4b3 c5e7 e1g1 g8f6 b1c3 e8g8 c4e2 b7b6 f2f4 c8b7 e2f3 d7d6 c1e3 a7a6 a2a4 d8c7 d1d2 a8d8 a1d1 f6d7 h2h3 f8e8 g1h1 d7c5 b3d4 e7f6 d4e2 c6b4 e2d4');
INSERT INTO evaluations VALUES(5176,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5176,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5176,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5176,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5176,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(5176,36,'cp 8','d4b3 c5e7 b1c3 g8f6 e1g1 e8g8 c1e3 h7h6 c4d3 d7d5 e4d5 f6d5 c3d5 d8d5 c2c3 c6e5 d3c2 e5c4 d1d5 e6d5 e3c1 c8g4 b3d4 e7f6 a2a4 g4d7 f1d1 f8e8 a4a5 f6d4 c3d4 a8c8 c2d3 a7a6 b2b3 c4d6 c1f4 d6e4');
INSERT INTO evaluations VALUES(35114,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(35114,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(35114,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(35114,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(35114,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(35114,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(35114,36,'cp 131','e5d4 c2c3 d4c3 b1c3 f8b4 e4e5 f6c6 f1e2 g8e7 e1g1 a7a6 c1g5 c6e6 c3e4 h7h6 a2a3 b4a5 e4f6 e8f8 g5h4 b8c6 f6h5 d7d5 h5f4 e6d7 h4g3 a5b6 h2h3 f8g8 e2d3 d7e8 b2b4 c8f5 f1e1 c6d8 d3f5 e7f5 f4d5 d8e6 d1c2 f5g3');
INSERT INTO evaluations VALUES(35151,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(35151,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(35151,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(35151,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(35151,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(35151,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(35151,36,'cp 124','c2c3 d4c3 b1c3 f8b4 e4e5 b4c3 b2c3 f6c6 f1e2 b7b6 e1g1 c8b7 f1e1 g8e7 c1a3 e7d5 e2f1 c6c3 a3d6 c7d6 e5d6 e8f8 a1c1 c3a5 f1c4 d5c3 d1d2 c3e4 d2a5 b6a5 c4d5 b7c6 d5e4 g7g6 e4d3 f8g7 f3d4 c6b7 c1c7 b8c6 c7d7 b7c8 d7f7 g7f7 d4c6 h8e8 c6e5 f7g7 d3b5 e8d8 d6d7');
INSERT INTO evaluations VALUES(16136,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16136,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(16136,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(16136,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(16136,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(16136,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(16136,36,'cp 26','f8c5 c2c3 c5b6 b1d2 e8g8 a2a4 d7d6 b2b4 a7a5 b4b5 c6e7 c4b3 c7c6 c1a3 e7g6 d2c4 b6c7 f1e1 c6b5 a4b5 c8e6 c4e3 a5a4 b3e6 f7e6 e3c4 d8d7 d3d4 f8f7 b5b6 c7b8 a3d6 b8d6 d4e5 d6e7 e5f6 d7d1 e1d1 f7f6');
INSERT INTO evaluations VALUES(57150,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(57150,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(57150,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(57150,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(57150,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(57150,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(57150,36,'cp -83','g8f6 b1c3 f8c5 f1e2 e8g8 d2d3 c7c6 g1f3 d7d5 c1b2 f8e8 e1g1 d5d4 c3b1 c5d6 h2h3 a7a6 a2a4 c6c5 b2c1 h7h6 f3h2 d6c7 h2g4 f6h7 g4h2 b8c6 g1h1 h7f6 h2g4 c8e6 g4f6 d8f6 e2g4 e6g4 d1g4 b7b5 b1d2 c6b4 g4d1 f6e6 d2f3 f7f6 a4b5');
INSERT INTO evaluations VALUES(49610,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(49610,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(49610,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(49610,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(49610,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(49610,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(49610,36,'cp -78','b1c3 f8c5 g1f3 e8g8 f1e2 d7d5 d2d3 c7c6 e1g1 f8e8 c1b2 d5d4 c3b1 c5d6 h2h3 h7h6 a2a4 a7a6 f3h2 d6c7 h2g4 f6h7 b1d2 h7f8 g4h2 f8e6 b2c1 b8d7 e2g4 d7f6 g4f5 g7g6 f5e6 c8e6 g1h1 g8g7 d2c4 e8h8');
INSERT INTO evaluations VALUES(49612,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(49612,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(49612,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(49612,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(49612,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(49612,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(49612,36,'cp -87','f8c5 g1f3 e8g8 f1e2 d7d5 d2d3 c7c6 e1g1 f8e8 h2h3 b8d7 a2a3 c5b6 f1e1 d7f8 e2f1 h7h6 b3b4 b6c7 a3a4 f8g6 a4a5 c7d6 a1b1 b7b6 c1d2 c8d7 b4b5 b6a5 b5c6 d7c6 c3d5 f6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(42613,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(42613,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(42613,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(42613,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(42613,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(42613,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(42613,36,'cp -92','g1f3 e8g8 f1e2 d7d5 d2d3 c7c6 e1g1 f8e8 h2h3 b8d7 a2a4 h7h6 a4a5 c5d6 a5a6 b7a6 f1e1 a6a5 c1d2 d7f8 e4d5 c6d5 c3b5 d6b4 c2c3 b4c5 d3d4 e5d4 b5d4 f6e4 e2d3 c8d7 d1c2 e4d2 c2d2 d8c7 d3f5 a8d8 f5d7 f8d7 e1e8 d8e8');
INSERT INTO evaluations VALUES(57422,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(57422,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(57422,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(57422,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(57422,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(57422,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(57422,36,'cp 214','f1c4 b8c6 g1f3 c6d4 f3d4 e5d4 d1d4 g8e7 b1c3 e7c6 d4e3 f8d6 c1d2 b7b6 f2f4 c8b7 e1c1 c6a5 c4d5 d8e7 e4e5 f6e5 d5b7 a5b7 c3d5 e7f7 f4e5 d6c5 e3b3 e8g8 h1f1 f7e6 d5c7 e6b3 a2b3 f8f1 d1f1 a8f8 f1d1 f8f2 b3b4 c5d4 d2c3 d4c3');
INSERT INTO evaluations VALUES(57406,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(57406,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(57406,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(57406,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(57406,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(57406,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(57406,36,'cp 180','b8c6 g1f3 c6d4 f3d4 e5d4 d1d4 g8e7 b1c3 e7c6 d4e3 f8d6 c1d2 b7b6 e3h3 c8b7 e1c1 c6e5 c4b3 d8e7 c3b5 e8c8 b5a7 c8b8 a7b5 b7e4 h1e1 e4c6 b5d4 e7f8 f2f4 d6c5 d4c6 e5c6 d2c3 c5b4 d1d7 b4c3 b2c3 f8a3 c1b1 c6a5 h3d3 a5c6 d7d8 h8d8 d3h7 a3c5');
INSERT INTO evaluations VALUES(5983,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(5983,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(5983,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(5983,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(5983,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(5983,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(5983,36,'cp 32','c2c3 c5b6 b2b4 g8f6 a2a4 a7a6 d2d3 h7h6 h2h3 e8g8 b1d2 c8e6 f1e1 b6a7 c4b3 e6b3 d1b3 f8e8 b3c2 d8d7 d2f1 d6d5 c1d2 a8d8 b4b5 c6b8 e4d5 d7d5 c3c4 d5d6 a1d1 a6b5 c4b5 d6d3 c2d3 d8d3 f3e5');
INSERT INTO evaluations VALUES(5974,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(5974,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(5974,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(5974,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(5974,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(5974,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(5974,36,'cp 18','g8f6 b2b4 c5b6 a2a4 a7a6 d2d3 e8g8 h2h3 c6e7 c4b3 e7g6 a4a5 b6a7 c1e3 a7e3 f2e3 c7c6 d1c2 h7h6 b1d2 d6d5 e4d5 c6d5 e3e4 c8e6 d3d4 f6e4 d2e4 d5e4 b3e6 e4f3 c2g6 f7e6 g6e6 g8h8 e6e5 d8h4 f1f3 f8f3 g2f3 h4h3 e5f4 a8e8');
INSERT INTO evaluations VALUES(9339,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(9339,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(9339,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(9339,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9339,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(9339,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9339,36,'cp 18','b2b4 c5b6 a2a4 a7a6 d2d3 e8g8 c4b3 h7h6 f1e1 c8e6 b1d2 e6b3 d1b3 f8e8 c1b2 b6a7 d3d4 d8d7 h2h3 f6h5 a1d1 h5f4 g1h2 b7b5 d2f1 b5a4 b3a4 e5d4 c3d4 d6d5 f1g3 d5e4 g3e4 d7d5 e4c3 d5c4 e1e4 e8e4 c3e4');
INSERT INTO evaluations VALUES(22810,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(22810,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(22810,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(22810,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(22810,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(22810,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(22810,36,'cp -19','g8f6 b1c3 f8c5 f1g2 d7d6 e1g1 e8g8 d2d3 a7a5 h2h3 a5a4 g1h2 c8e6 f3g5 h7h6 g5e6 f7e6 a2a3 d8e7 c3e2 d6d5 c1d2 f8f7 d1e1 a8f8 a1d1 c6d4 e2d4 c5d4 c2c3 d4c5 e1e2 c7c6 d1e1 f6d7 f2f4 e7d8 e4d5 c6d5 c3c4 e5f4');
INSERT INTO evaluations VALUES(16571,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(16571,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(16571,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(16571,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(16571,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(16571,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(16571,36,'cp -11','b1c3 f8c5 f1g2 d7d6 e1g1 a7a5 d2d3 h7h6 c3d5 e8g8 c1e3 c8e6 d3d4 e6d5 e4d5 c6d4 f3d4 e5d4 e3d4 c5d4 d1d4 d8d7 a1e1 f8e8 d4c3 b7b6 b2b3 h6h5 a2a4 a8d8 h2h3 g8f8 h3h4 f8g8 c3f3 e8e5 c2c4 d8e8');
INSERT INTO evaluations VALUES(16510,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16510,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16510,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(16510,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(16510,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(16510,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16510,36,'cp -14','f8c5 f1g2 d7d6 d2d3 a7a5 e1g1 h7h6 h2h3 c8e6 g1h2 a5a4 c3b5 e8g8 c2c4 f6d7 b5c3 f7f5 e4f5 e6f5 c1e3 f5h7 c3a4 c5e3 f2e3 c6b4 e3e4 b4d3 d1d3 a8a4 c4c5 a4a6 a1d1 g8h8 c5d6 c7d6 f3h4 d7f6 a2a3 d6d5 d3b5 a6b6 b5c5 b6c6 c5b5 h7e4 g2e4 f6e4');
INSERT INTO evaluations VALUES(10211,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(10211,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10211,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10211,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10211,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10211,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10211,36,'cp 0','f1g2 d7d6 e1g1 a7a5 d2d3 c8e6 f3h4 e6g4 g2f3 g4h3 f3g2');
INSERT INTO evaluations VALUES(10212,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(10212,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(10212,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(10212,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(10212,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(10212,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10212,36,'cp -13','d7d6 d2d3 a7a5 e1g1 e8g8 h2h3 c8e6 c3a4 c5a7 c2c4 c6d4 a4c3 c7c6 f3d4 a7d4 c3e2 d4a7 b2b3 b7b5 g1h2 b5c4 b3c4 a8b8 f2f4 a7c5 c1d2 e5f4 e2f4 c5d4 a1b1 b8b1 d1b1 f6d7 f4e6');
INSERT INTO evaluations VALUES(9394,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(9394,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9394,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(9394,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(9394,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(9394,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(9394,36,'cp -16','e1g1 e8g8 d2d3 a7a5 h2h3 c8e6 c3a4 c5a7 c2c4 c6d4 a4c3 c7c6 f3d4 a7d4 c3e2 d4a7 b2b3 b7b5 g1h2 a5a4 a1b1 a4b3 a2b3 b5c4 b3c4 a7c5 d1c2 d8c7 c1b2 f8b8 d3d4 e5d4 b2d4 f6d7 d4c5 d7c5 e2f4 b8b1 f1b1');
INSERT INTO evaluations VALUES(9393,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(9393,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(9393,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(9393,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(9393,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(9393,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(9393,36,'cp -20','a7a5 d2d3 c8e6 c3e2 e8g8 h2h3 h7h6 c2c4 a5a4 g1h2 f6d7 e2c3 a4a3 b2b3 f7f5 e4f5 e6f5 c1e3 c5e3 f2e3 f5h7 d1d2 c6b4 c3e4 c7c5 f3g1 d8e7 g1e2 d6d5 c4d5 b4d5 e2c3 d5c3 e4c3 d7f6 a1e1 g8h8 c3d5 f6d5 g2d5 a8e8');
INSERT INTO evaluations VALUES(52446,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(52446,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(52446,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(52446,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(52446,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(52446,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(52446,36,'cp -53','b8c6 b1c3 e7e6 f3g3 a7a6 f1e2 d7d5 d2d3 g8f6 g1f3 h7h6 e1g1 g7g6 f3d2 f8e7 e4e5 f6d7 f2f4 c6d4 e2d1 d4f5 g3f2 b7b5 c3e2 h6h5 a2a4 c8b7 e2g3 f5g7 a4b5 h5h4 g3e2 a6b5 a1a8 d8a8');
INSERT INTO evaluations VALUES(19466,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(19466,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(19466,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(19466,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(19466,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(19466,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(19466,36,'cp -45','b1c3 g8f6 c3d5 g7g6 d5f6 e7f6 f1c4 c6e5 f3b3 e5c4 b3c4 d7d6 b2b3 f8g7 c1b2 e8g8 g1e2 c8e6 c4c3 d6d5 e4d5 e6d5 e1g1 a8c8 d2d3 b7b5 f1e1 h7h5 c3d2 f8e8 c2c4 d5c6 a1d1 g8h7 c4b5 c6b5 e2f4 e8e1 d1e1 c5c4 b3c4 b5c4');
INSERT INTO evaluations VALUES(19436,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(19436,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(19436,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(19436,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(19436,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(19436,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(19436,36,'cp -75','e7e6 f3g3 g8f6 f1b5 c6d4 b5d3 h7h5 h2h4 a7a6 e4e5 f6g4 f2f3 c5c4 f3g4 c4d3 g3d3 d4c6 g4h5 d8a5 d3e2 a5e5 g1f3 e5e2 e1e2 b7b5 d2d3 c8b7 c1f4 a8c8 c3e4 e6e5 f4d2 f7f5 e4g3 f5f4 g3e4 d7d5 e4g5 f8d6 a1e1 h8h5 e2d1 e8d7 e1e2 b5b4 a2a3 b4a3 b2a3 a6a5 h1e1');
INSERT INTO evaluations VALUES(34209,30,'cp 360',NULL);
INSERT INTO evaluations VALUES(34209,31,'cp 331',NULL);
INSERT INTO evaluations VALUES(34209,32,'cp 354',NULL);
INSERT INTO evaluations VALUES(34209,33,'cp 354',NULL);
INSERT INTO evaluations VALUES(34209,34,'cp 371',NULL);
INSERT INTO evaluations VALUES(34209,35,'cp 378',NULL);
INSERT INTO evaluations VALUES(34209,36,'cp 363','d2d4 e5d4 c1g5 f6d6 f3d4 f8e7 g5e3 g8f6 f2f4 d6b4 a2a3 b4a5 d4b3 a5c7 e4e5 f6d5 c3d5 c6d5 d1d5 d7d6 e5d6 c7d6 d5d6 e7d6 e1c1 d6c7 b3d4 e8g8 d4b5 b8a6 f1d3 f8e8 b5c7 a6c7 e3f2 b7b6 h1e1 c8g4 e1e8 a8e8 d1e1 e8e1 f2e1 g4d7 c1d2 d7c6 g2g3 c7e8 b2b3 e8f6 f4f5 c6f3');
INSERT INTO evaluations VALUES(34207,30,'cp 358',NULL);
INSERT INTO evaluations VALUES(34207,31,'cp 386',NULL);
INSERT INTO evaluations VALUES(34207,32,'cp 375',NULL);
INSERT INTO evaluations VALUES(34207,33,'cp 358',NULL);
INSERT INTO evaluations VALUES(34207,34,'cp 358',NULL);
INSERT INTO evaluations VALUES(34207,35,'cp 357',NULL);
INSERT INTO evaluations VALUES(34207,36,'cp 372','e5d4 c1g5 f6d6 f3d4 f8e7 g5e3 g8f6 f2f4 d6c7 e4e5 f6d5 c3d5 c6d5 d4f5 b8c6 f5g7 e8f8 g7f5 d7d6 f5e7 c6e7 f1d3 e7f5 d3f5 c8f5 d1d5 d6e5 e1g1 a8d8 d5e5 c7e5 f4e5 f5e6 e3a7 f8e8 a7e3 h8g8 b2b3 g8g4 h2h3 g4e4 a1e1 d8d5 c2c4 d5d3 e3f4 e4e1 f1e1 e8d7 f4e3');
INSERT INTO evaluations VALUES(34216,30,'cp 383',NULL);
INSERT INTO evaluations VALUES(34216,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(34216,32,'cp 412',NULL);
INSERT INTO evaluations VALUES(34216,33,'cp 408',NULL);
INSERT INTO evaluations VALUES(34216,34,'cp 371',NULL);
INSERT INTO evaluations VALUES(34216,35,'cp 389',NULL);
INSERT INTO evaluations VALUES(34216,36,'cp 388','c1g5 f6d6 f3d4 d6b4 a2a3 b4a5 d1d2 d7d6 e1c1 f8e7 g5e7 g8e7 f2f4 b8d7 f1c4 e8g8 d4b3 a5c7 d2d6 c7d6 d1d6 d7f6 c4e2 b7b6 h1d1 c6c5 b3d2 f6e8 d6d3 e8f6 d2c4 c8g4 e2f3 g4f3 g2f3 f6h5 f4f5 h5f4 d3d7 e7c6 c3b5 c6d4 b5d4 c5d4');
INSERT INTO evaluations VALUES(34213,30,'cp 406',NULL);
INSERT INTO evaluations VALUES(34213,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(34213,32,'cp 388',NULL);
INSERT INTO evaluations VALUES(34213,33,'cp 385',NULL);
INSERT INTO evaluations VALUES(34213,34,'cp 375',NULL);
INSERT INTO evaluations VALUES(34213,35,'cp 393',NULL);
INSERT INTO evaluations VALUES(34213,36,'cp 376','f6d6 f3d4 d6b4 a2a3 b4a5 d1d2 d7d6 e1c1 f8e7 g5e7 g8e7 f1c4 b8d7 f2f4 e8g8 d4b3 a5c7 d2d6 c7d6 d1d6 d7f6 h1e1 b7b6 d6d2 c6c5 c3b5 c8b7 e4e5 f6d5 g2g3 b7c6 a3a4 a8d8 b5d6 d5c7 a4a5 c6d5 e1d1 d5e6 c4e6 f7e6 d6e4 e7c6 d2d8 f8d8');
INSERT INTO evaluations VALUES(16538,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(16538,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(16538,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(16538,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(16538,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(16538,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(16538,36,'cp -16','d7d5 b1d2 a7a5 f1e2 f8d6 e1g1 e8g8 e4d5 f6d5 f1e1 f8e8 d2c4 h7h6 h2h3 d8f6 c4d6 f6d6 e2f1 c8f5 c1d2 a8d8 a2a3 d5b6 e1e3 a5a4 b2b4 a4b3 d1b3 b6d5 e3e1 e5e4 f3h4 f5h7 d3e4 h7e4 a1d1 g7g5 h4f3 e4f3 e1e8 d8e8 g2f3 c6e5');
INSERT INTO evaluations VALUES(15147,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(15147,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(15147,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(15147,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(15147,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(15147,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(15147,36,'cp -12','b1d2 f8d6 f1e2 h7h6 e1g1 e8g8 f1e1 a7a5 e2f1 f8e8 e4d5 f6d5 h2h3 c8f5 d2c4 a5a4 c1d2 d5b6 c4d6 d8d6 d2e3 b6d5 e3d2 e8d8 d1e2 d5f6 f3e5 d6e5 e2e5 c6e5 e1e5 f5d3 d2e3 d3f1 g1f1 f6d5 e3c5 b7b6 c3c4 d5f4 c5e3 f4d3 e5e7 c7c5');
INSERT INTO evaluations VALUES(15146,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(15146,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(15146,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(15146,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(15146,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(15146,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(15146,36,'cp -23','f8d6 f1e2 e8g8 e1g1 f8e8 f1e1 a7a5 e2f1 c8e6 e4d5 e6d5 b2b3 h7h6 d2c4 d5c4 b3c4 e5e4 d3e4 f6e4 d1c2 e4c5 c1e3 d8f6 a1d1 c6e5 f3e5 f6e5 g2g3 e5e4 c2e4 c5e4 e3d4 e4c5 e1e8 a8e8 d4e3 a5a4 d1d5 b7b6 g1g2 a4a3');
INSERT INTO evaluations VALUES(31144,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31144,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(31144,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(31144,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31144,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(31144,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31144,36,'cp 0','e7e6 c2c3 c6c5 g1e2 b8c6 e1g1 d8b6 b1d2 c5d4 c3d4 c6b4 d3b3 b4c6');
INSERT INTO evaluations VALUES(31495,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(31495,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(31495,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31495,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31495,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31495,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(31495,36,'cp 13','g1f3 c6c5 c2c3 b8d7 d3e2 g8e7 d4c5 d7c5 e1g1 a7a6 f1d1 d8c7 h2h4 e7c6 h4h5 f8e7 c1f4 h7h6 b1d2 e8g8 a1c1 f8d8 f3d4 a8c8 c1c2 e7f8 d4c6 c7c6 f4e3 c5e4 d2e4 d5e4 d1d4 d8d4 e3d4 f8e7 e2g4 g8h8 b2b3 c8d8');
INSERT INTO evaluations VALUES(31493,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(31493,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31493,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(31493,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(31493,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(31493,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(31493,36,'cp 10','c6c5 c2c3 b8d7 e1g1 g8e7 d4c5 d7c5 d3e2 a7a6 f1d1 d8c7 h2h4 e7c6 b1d2 f8e7 d2b3 e8g8 b3c5 e7c5 d1d3 h7h6 c1f4 f7f6 a1e1 c5a7 f3d4 a7d4 c3d4 c7f7 e2d2 f6f5 d3g3 g8h7 g3b3 f7d7 a2a3 b7b5 b3g3');
INSERT INTO evaluations VALUES(31653,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(31653,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(31653,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31653,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(31653,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(31653,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(31653,36,'cp 17','c2c3 b8d7 e1g1 g8e7 d4c5 d7c5 d3e2 e7c6 f1d1 d8b6 b1a3 c5e4 a3c2 f8e7 c2d4 e8g8 c1e3 b6a5 d4c6 b7c6 f3e1 a5a4 e1d3 a4c4 d1e1 a7a5 f2f3 e4c5 d3c5 e7c5 e2c4 c5e3 g1f1 d5c4 e1e3 h7h6 f1e2 f8b8');
INSERT INTO evaluations VALUES(31649,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31649,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(31649,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31649,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31649,36,'cp 13','b8d7 e1g1 g8e7 d4c5 d7c5 d3e2 a7a6 c1e3 e7c6 f1d1 d8c7 b1d2 a8d8 e3d4 f8e7 d2b3 c6d4 f3d4 e8g8 g2g3 h7h6 g1g2 d8c8 h2h4 c7b6 b3c5 e7c5 d1d3 c5d4 d3d4 f8d8 g2g1 b6c5 a1d1 c5b5 e2b5 a6b5');
INSERT INTO evaluations VALUES(26145,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(26145,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(26145,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(26145,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(26145,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(26145,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26145,36,'cp 0','e1g1 g8e7 d4c5 d7c5 d3e2 e7c6 f1d1 d8b6 c1e3 b6a6 e2a6 c5a6 b1d2 a6c5 c3c4 e8c8 d2b3 c5b3 a2b3 d5c4 d1d8 c6d8 a1a7 c4b3 a7a8 c8d7 f3d2 f8b4 d2b3 h8e8 a8b8 d7c7 b8a8');
INSERT INTO evaluations VALUES(26146,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26146,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26146,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26146,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26146,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26146,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26146,36,'cp 0','g8e7 d4c5 d7c5 d3e2 e7c6 f1d1 d8b6 c1e3 b6a6 e2a6 c5a6 b1d2 a6c5 c3c4 e8c8 d2b3 c5b3 a2b3 d5c4 d1d8 c6d8 a1a7 c4b3 a7a8 c8d7 f3d2 f8b4 d2b3 h8e8 a8a7 d7c7 a7a8');
INSERT INTO evaluations VALUES(24539,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24539,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24539,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24539,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24539,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24539,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24539,36,'cp 0','d4c5 d7c5 d3e2 e7c6 f1d1 d8b6 c1e3 b6a6 e2a6 c5a6 b1d2 a6c5 c3c4 e8c8 d2b3 c5b3 a2b3 d5c4 d1d8 c6d8 a1a7 c4b3 a7a8 c8d7 f3d2 f8b4 d2b3 h8e8 a8a7 d7c7 a7a8');
INSERT INTO evaluations VALUES(55400,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(55400,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(55400,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(55400,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(55400,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(55400,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(55400,36,'cp 50','c7c5 c3d5 g8f6 c2c3 f6d5 c4d5 f8e7 d2d4 e5d4 c3d4 b8c6 e1g1 e8g8 c1e3 d8b6 a1b1 c6b4 d5c4 c8g4 d4d5 e7g5 e3g5 h6g5 d1e2 b6d8 a2a3 g4f3 g2f3 b4a6 b2b4 c5b4 c4a6 b7a6 b1b4 d8f6 g1g2 a8c8 b4a4 c8c3 a4a6 f8c8 a6a7');
INSERT INTO evaluations VALUES(53510,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(53510,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(53510,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(53510,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(53510,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(53510,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(53510,36,'cp 53','c3d5 g8f6 c2c3 f6d5 c4d5 d8f6 d2d3 f8e7 e1g1 b8d7 a2a4 d7b6 d5b3 c8e6 a4a5 b6d7 f3d2 e8g8 d2c4 f6g6 c4e3 e7g5 b3e6 f7e6 d1b3 g5e3 c1e3 a8b8 b3a4 f8f7 b2b4 g8h7 f2f3 b8c8 g1h1 c8c7 a1b1 c7c8 f1c1 c8f8 c1e1 h7h8 c3c4');
INSERT INTO evaluations VALUES(31346,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(31346,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(31346,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(31346,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31346,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(31346,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(31346,36,'cp 35','h2h3 g4f3 d1f3 e7e6 c1e3 b8c6 b1d2 f8d6 e1c1 d8f6 c2c3 f6f3 d2f3 a7a6 a2a4 a8c8 d1e1 g8f6 g2g4 h7h5 g4g5 f6e4 c1b1 f7f5 g5f6 g7f6 h1g1 e8f7 h3h4 c6e7 f1d3 e7g6 b1c2 g6e7 f3d2 e4d2 e3d2 e7f5');
INSERT INTO evaluations VALUES(31343,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(31343,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(31343,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(31343,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(31343,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(31343,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(31343,36,'cp 35','g4f3 d1f3 e7e6 c1e3 b8c6 b1d2 f8d6 e1c1 d8f6 c2c3 f6f3 d2f3 a7a6 a2a4 f7f6 b2b4 h7h5 f1d3 g8e7 c1b2 e8f7 d1e1 e7f5 e3c1 g7g5 g2g3 a8e8 e1e2 d6c7 h1e1 g5g4 h3g4 h5g4 d3f5 e6f5 e2e8 h8e8 e1e8 f7e8');
INSERT INTO evaluations VALUES(31353,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(31353,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(31353,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31353,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(31353,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(31353,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(31353,36,'cp 35','d1f3 e7e6 c1e3 b8c6 b1d2 f8d6 e1c1 d8f6 c2c3 f6f3 d2f3 a7a6 a2a4 f7f6 f1d3 e8f7 b2b4 g8e7 g2g3 h7h5 c1b2 g7g5 d1e1 a8g8 e3d2 h5h4 g3h4 g5h4 h1g1 g8g1 e1g1 e7f5 d3f5 e6f5 b2c2 c6d8 f3e1 d8e6 b4b5');
INSERT INTO evaluations VALUES(31352,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31352,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31352,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(31352,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(31352,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(31352,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31352,36,'cp 39','e7e6 f1d3 g8f6 f3e2 f8d6 b1d2 d8c7 e1g1 b8c6 d2f3 c6b4 d3b5 b4c6 c2c3 f6e4 f3g5 e4f6 b5d3 e8g8 g5f3 a8e8 c1g5 f6d7 a1d1 h7h6 g5c1 a7a6 f1e1 f7f5 c3c4 d7f6 c4c5 d6f4 c1f4 c7f4 d3c2 f6e4 c2a4 e4g5');
INSERT INTO evaluations VALUES(31725,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(31725,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(31725,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(31725,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(31725,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(31725,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31725,36,'cp 33','f1d3 g8f6 b1d2 f8d6 f3e2 e8g8 e1g1 b8c6 c2c3 e6e5 d4e5 d6e5 d2f3 f8e8 e2c2 h7h6 c1e3 d8c7 c2a4 e8e7 a1d1 a8e8 f3e5 c6e5 d3b5 e5c6 a4c2 a7a6 b5a4 b7b5 a4b3 c6a5 e3d4 a5b3 c2b3 e7e2 d4f6 g7f6 b3d5');
INSERT INTO evaluations VALUES(23443,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(23443,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23443,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23443,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(23443,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(23443,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(23443,36,'cp 5','e7e5 g1f3 g8f6 f1b5 c6d4 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d2c3 f8c5 d1e2 f6e6 e1g1 e6e2 b5e2 d7d5 c1f4 c7c6 e2d3 e8g8 f1e1 c8e6 a2a4 a7a5 h2h3 f8e8 e1e2 e6d7 e2e8 d7e8 f4c7 h7h6 g2g4 g8f8 g1g2 f7f6 h3h4 e8f7');
INSERT INTO evaluations VALUES(5969,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5969,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(5969,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(5969,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5969,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(5969,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(5969,36,'cp 8','c3a4 d8f6 e1g1 g8e7 c2c3 h7h6 b2b4 c5b6 a4b6 a7b6 a2a4 e8g8 d2d4 e7g6 h2h3 e5d4 c3d4 c6b4 d1d2 b4c6 a1a3 c8d7 f1e1 f8d8 c1b2 c6e7 d4d5 g6e5 f3e5 d6e5 d2c2 e7g6 c2b3 g8h7 e1d1 g6f4 c4f1 d8c8 b3c2');
INSERT INTO evaluations VALUES(6033,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(6033,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(6033,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6033,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(6033,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6033,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6033,36,'cp 0','d8f6 c2c3 g8e7 b2b4 c5b6 e1g1 h7h6 a4b6 a7b6 a2a4 e7g6 h2h3 g6f4 d2d4 f4h3 g2h3 c8h3 g1h1 h3f1 c4f1 e5d4 c3d4 c6b4 a1b1 b4c6 d4d5 c6e7 f1b5 c7c6 b5e2 e8g8 c1b2 f6f4 d1d4 f7f6 b2c1 f4g4 b1b6 g8h8 c1d2 g4h5 h1g1 h5g4 g1h1');
INSERT INTO evaluations VALUES(53006,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53006,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53006,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53006,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(53006,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(53006,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(53006,36,'cp 0','a7a6 a2a4 b8c6 g1f3 g8f6 d2d4 c5d4 f3d4 e7e6 c1e3 f8e7 h2h3 c8d7 c4d3 a8c8 e1g1 e8g8 f2f4 c6d4 e3d4 d7c6 d1e2 d8a5 g1h1 a5b4 e2e3 f6d7 c3e2 e6e5 f4e5 d6e5 d4c3 b4c5 e3g3 c5d6 a4a5 d6g6 c3e5 d7e5 g3e5 e7f6');
INSERT INTO evaluations VALUES(50647,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(50647,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(50647,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(50647,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50647,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50647,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50647,36,'cp 3','a2a4 g8f6 g1e2 b8c6 d2d4 c5d4 e2d4 e7e6 e1g1 f8e7 c1e3 e8g8 h2h3 c8d7 c4d3 d8c7 d1e2 c6b4 f2f4 e6e5 f4e5 d6e5 d4f5 e7c5 g1h1 d7f5 f1f5 b4d3 e2d3 a8d8 d3c4 d8c8 e3c5 c7c5 c4e2 g8h8 a1f1 h7h6 e2f3');
INSERT INTO evaluations VALUES(50674,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(50674,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50674,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50674,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50674,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50674,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50674,36,'cp 14','e7e6 g1e2 f8e7 d2d4 c5d4 e2d4 g8f6 e1g1 d8c7 d1e2 e8g8 a4a5 c8d7 c1e3 b8c6 d4c6 d7c6 e3b6 c7c8 h2h3 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 c6b5 e4d3 b5d3 e2d3 c8c6 f1d1 a8c8 c2c3 e7c5 b6c5 c6c5 d3d4 c5c7 d4d6 c7d6 d1d6 f8d8');
INSERT INTO evaluations VALUES(50749,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50749,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50749,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(50749,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(50749,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(50749,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50749,36,'cp 7','g1e2 g8f6 d2d4 c5d4 e2d4 f8e7 e1g1 b8c6 c1e3 e8g8 h2h3 c8d7 f2f4 a8c8 c4d3 c6d4 e3d4 d7c6 d1e2 f6d7 g1h1 g7g6 a4a5 e7f6 e2f2 f6h4 f2e3 e6e5 f4e5 d7e5 e3e2 h4g5 b2b4 e5d3 e2d3 f8e8 c3d5 c6d5 e4d5 f7f5 b4b5');
INSERT INTO evaluations VALUES(57189,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(57189,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(57189,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(57189,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(57189,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(57189,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(57189,36,'cp -29','d7d5 c1b2 f8d6 c2c4 c7c6 b1c3 g8f6 c4d5 c6d5 c3b5 e8g8 g1f3 b8c6 b5d6 d8d6 d2d4 e5e4 f3e5 c8e6 e5c6 d6c6 f1e2 c6d7 e1g1 f8e8 a1c1 e6g4 e2g4 f6g4 h2h3 g4f6 d1d2 e8e6 c1c2 f6e8 b2a3 e6g6 g1h1 a8d8 f1a1 h7h6 a1c1');
INSERT INTO evaluations VALUES(56425,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(56425,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(56425,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(56425,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(56425,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(56425,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(56425,36,'cp -25','c1b2 f8d6 c2c4 c7c6 b1c3 g8f6 c4d5 c6d5 c3b5 e8g8 g1f3 b8c6 b5d6 d8d6 d2d4 d6b4 d1d2 b4b6 d2d1 f6e4 f1e2 b6a5 e1f1 f7f6 h2h3 c8f5 d1e1 a5e1 a1e1 e5d4 f3d4 c6d4 b2d4 a8c8 e2d1 f5g6 h3h4 h7h5 g2g3 a7a6 f1g2 b7b5 d1f3 c8c2 e1e2');
INSERT INTO evaluations VALUES(56426,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(56426,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(56426,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(56426,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(56426,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(56426,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(56426,36,'cp -26','f8d6 c2c4 c7c6 b1c3 g8f6 c4d5 c6d5 c3b5 e8g8 g1e2 b8c6 b5d6 d8d6 e2g3 c8g4 f1e2 g4e2 d1e2 f6e4 e1g1 e4g3 h2g3 a8d8 a1c1 a7a6 f2f4 e5e4 g3g4 f8e8 g4g5 d6g6 e2f2 h7h6 g5h6 g6h6 f2g3 d5d4 f4f5 d4e3 d2e3 d8d2 f1f2 e8d8');
INSERT INTO evaluations VALUES(38640,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(38640,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(38640,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(38640,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(38640,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(38640,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(38640,36,'cp -25','c2c4 c7c6 b1c3 g8e7 f1d3 b8a6 c4d5 a6b4 d3b1 c6d5 g1e2 c8f5 b1f5 e7f5 d2d3 d5d4 e3d4 f5d4 e2d4 e5d4 c3e4 e8g8 e1g1 b4c6 b2c1 d8d7 e4d6 d7d6 d1f3 f8e8 c1d2 c6e5 f3b7 e5d3 a1d1 h7h6 b7f3 d6g6 d2f4 d3f4 f3f4 e8e4 f4f3 a8e8');
INSERT INTO evaluations VALUES(12786,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(12786,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(12786,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(12786,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(12786,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(12786,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12786,36,'cp 0','e4e5 d7d5 c4b5 f6d7 e1g1 d7b8 d2d4 a7a6 b5c6 b8c6 c2c4 c5d4 f3d4 d5c4 d4c6 d8d1 f1d1 b7c6 c1e3 g7g6 b1d2 c4c3 b2c3 f8g7 e3c5 c8e6 d2f3 a8d8 d1b1 e6d5 b1b7 d8d7 b7b8 d7d8 b8b6');
INSERT INTO evaluations VALUES(12766,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(12766,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12766,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(12766,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(12766,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(12766,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(12766,36,'cp 0','d7d5 c4b5 f6d7 e1g1 d7b8 d2d4 a7a6 b5c6 b8c6 c2c4 c5d4 f3d4 d5c4 d4c6 d8d1 f1d1 b7c6 c1e3 c8e6 b1d2 g7g6 a1c1 f8g7 d2c4 e6c4 c1c4 g7e5 b2b3 e5d6 c4c6 e8d7 c6c5 d7e6 c5c4 h8c8 c4e4 e6f5 e4a4 f7f6 g1f1 c8c6 d1d4 f5e6 a4a5 a8c8 a5a4');
INSERT INTO evaluations VALUES(32213,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(32213,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(32213,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(32213,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(32213,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(32213,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(32213,36,'cp 60','c4e6 f7e6 c2c3 g8f6 e1g1 b8c6 d1b3 d8c8 f1e1 f8e7 d2d4 e8g8 d4e5 c6e5 f3e5 d6e5 b3c4 c8d7 a2a4 a7a6 b1d2 b7b5 c4e2 d7c6 d2f3 e7d6 h2h3 h7h6 e2c2 f8f7 a4b5 a6b5 a1a8 c6a8 c1d2 a8b7 c3c4 b7c6 e1c1 c6c4');
INSERT INTO evaluations VALUES(32201,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(32201,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(32201,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(32201,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(32201,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(32201,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(32201,36,'cp 54','f7e6 c2c3 g8f6 e1g1 b8c6 d1b3 d8c8 f1e1 f8e7 d2d4 e8g8 d4e5 c6e5 f3e5 d6e5 b1d2 a7a5 d2f3 a5a4 b3c4 e7d6 c1g5 c8e8 g5f6 f8f6 e1e3 b7b6 c4e2 e8h5 h2h3 d6c5 e3d3 a8f8 a1d1 c5d6 d1d2 g7g5 f3h2 h5e2 d2e2 b6b5 h2g4');
INSERT INTO evaluations VALUES(32446,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(32446,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(32446,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(32446,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(32446,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(32446,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(32446,36,'cp 65','c2c3 g8f6 d2d4 b8c6 d1b3 d8c8 e1g1 f8e7 f1e1 e8g8 d4e5 c6e5 f3e5 d6e5 b3c4 c8d7 b1d2 b7b5 c4e2 d7c6 d2f3 e7d6 c1g5 f6d7 a1c1 h7h6 g5e3 c6e4 e2b5 d7f6 e3d4 a8b8 b5a4 e4f5 d4e5 b8b2 e5d6 c7d6 a4a3 f8b8 c3c4 d6d5 a3a7 d5c4 c1c4');
INSERT INTO evaluations VALUES(32443,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32443,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(32443,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(32443,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(32443,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(32443,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(32443,36,'cp 49','g8f6 d1b3 d8c8 e1g1 f8e7 f1e1 b8c6 d2d4 e8g8 d4e5 c6e5 f3e5 d6e5 h2h3 a7a5 a2a4 c8d7 b1d2 e7c5 d2f3 d7c6 b3c2 c5d6 c1g5 h7h6 g5f6 f8f6 e1e2 b7b6 f3e1 c6e8 e1d3 f6f8 g1h2 e8e7 a1d1 a8d8 d1d2 d6c5');
INSERT INTO evaluations VALUES(20422,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(20422,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(20422,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(20422,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(20422,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(20422,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(20422,36,'cp 53','g8e7 b1c3 e7f5 f1b5 c8d7 e1g1 a8c8 b5a4 d8b6 c3e2 f8e7 a2a3 c6a5 a4c2 d7b5 f1e1 e8g8 e2c3 b5c4 c3a4 b6d8 h2h3 a5c6 c2f5 e6f5 b2b3 c4b5 a4c3 b5a6 c3e2 a6e2 d1e2 d8d7 e2d3 c8c7 c1f4 d7e6 b3b4 f8c8 a1c1 h7h6');
INSERT INTO evaluations VALUES(12977,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(12977,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(12977,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(12977,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(12977,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(12977,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(12977,36,'cp 41','b1c3 e7f5 f1b5 c8d7 e1g1 d8b6 b5a4 f8e7 c3e2 c6b4 a4d7 e8d7 c1g5 e7g5 f3g5 a8f8 g5f3 f8c8 a1b1 c8c7 a2a3 b4c6 b2b4 h7h5 d1d3 c6e7 f3g5 c7c4 g5e4 a7a6 e4d2 c4c8 d2b3 b6b5 d3d2 b7b6 e2c3 b5c6 f1c1 d7e8 h2h3 c6d7');
INSERT INTO evaluations VALUES(12974,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12974,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(12974,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12974,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12974,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(12974,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12974,36,'cp 33','e7f5 f1b5 c8d7 e1g1 a8c8 b5a4 d8b6 c3e2 b6a6 a4c2 c6b4 c2f5 e6f5 e2c3 d7e6 a2a3 b4d3 d1e2 c8c6 g2g3 f5f4 f1d1 d3c1 e2a6 c6a6 d1c1 f4g3 h2g3 f8e7 f3e1 e8g8 e1d3 a6b6 c3a4 b6b3 d3f4 e7g5 f4e6 f7e6 c1c7 g5d8 c7c3 b3c3 a4c3');
INSERT INTO evaluations VALUES(13479,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(13479,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(13479,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(13479,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(13479,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(13479,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(13479,36,'cp 34','f1b5 c8d7 e1g1 d8b6 b5a4 a8c8 c3e2 b6a6 a4c2 c6b4 c2f5 e6f5 e2c3 d7e6 f3e1 f8e7 c1e3 e8g8 a2a3 b4c6 e1d3 c6a5 d3f4 f8d8 d1f3 a6c6 a1c1 a5c4 f3g3 g8h8 h2h4 c6d7 f4d3 h7h6 c3e2 c4e3 g3e3 e7h4');
INSERT INTO evaluations VALUES(13456,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(13456,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(13456,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(13456,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(13456,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(13456,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(13456,36,'cp 52','c8d7 e1g1 d8b6 b5a4 a8c8 c3e2 f8e7 a2a3 c6a5 a4c2 d7b5 f1e1 b5e2 e1e2 e8g8 b2b4 a5c6 c2f5 e6f5 d1d3 a7a5 b4a5 c6a5 c1d2 a5c4 d3f5 e7a3 f5d3 a3e7 d2g5 e7g5 f3g5 b6h6 d3g3 h6g6 e5e6');
INSERT INTO evaluations VALUES(57133,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(57133,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(57133,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(57133,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(57133,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(57133,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(57133,36,'cp -12','e5d4 d1d4 b8c6 d4d1 g8f6 b1c3 f8b4 c1d2 e8g8 g1f3 f8e8 e2e3 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 b7b6 a1d1 c8b7 f1e2 d7d6 e1g1 c6e5 f3e5 e8e5 d1d4 d8f6 h2h4 a7a5 d4f4 f6e7 f1d1 h7h6 a2a3 g7g6');
INSERT INTO evaluations VALUES(57282,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(57282,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57282,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57282,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57282,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57282,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57282,36,'cp 8','d1d4 b8c6 d4d1 g8f6 b1c3 f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 d7d6 d1c2 b4c3 d2c3 a7a5 a2a3 f6e4 f1d3 d8e7 e1g1 e4c3 c2c3 c6e5 f3e5 e7e5 a1c1 b7b6 f1d1 c8b7 d3f1 g7g6 d1d4 e8e7 c1d1 a8e8 c3d2 b7e4');
INSERT INTO evaluations VALUES(57280,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(57280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57280,36,'cp 0','b8c6 d4d1 g8f6 b1c3 f8b4 c1d2 e8g8 g1f3 b7b6 e2e3 b4c3 d2c3 f6e4 d1c2 f8e8 f1d3 e4c3 c2c3 c8b7 e1g1 d8e7 f1d1 c6e5 f3e5 e7e5 c3c2 g7g6 d3f1 d7d6 d1d2 b7e4 f1d3 e4b7');
INSERT INTO evaluations VALUES(22907,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22907,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22907,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22907,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22907,36,'cp 0','d4d1 g8f6 b1c3 f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 b7b6 f1e2 c8b7 e1g1 a7a5 f1d1 e4c3 c2c3 d7d6 e2f1 e8e6 d1d2 d8e7 a1d1 c6e5 f3e5 e6e5 d2d4 h7h6 d1d2 e7f6 c3c2 f6e7');
INSERT INTO evaluations VALUES(22903,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22903,36,'cp 0','g8f6 b1c3 f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 b7b6 f1e2 d7d6 e1g1 c8b7 a1d1 c6e5 f3e5 e8e5 d1d4 e5g5 g2g3 d8f6 c3c2 g5e5 f1d1 a8e8 d4f4 f6e6 e2g4 e6g6 g4f5 g6h5 f5g4');
INSERT INTO evaluations VALUES(16661,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(16661,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16661,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16661,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16661,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(16661,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16661,36,'cp 0','b1c3 f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 b7b6 f1e2 d7d6 e1g1 c8b7 f1d1 c6e5 f3e5 e8e5 d1d4 a7a5 c3c2 d8f6 a1d1 a8e8 e2f1 h7h6 b2b3 b7e4 f1d3 e4b7');
INSERT INTO evaluations VALUES(16659,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(16659,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(16659,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16659,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16659,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16659,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16659,36,'cp -8','f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 b7b6 f1e2 d7d6 e1g1 c8b7 f1d1 c6e5 f3e5 e8e5 d1d4 a7a5 a1d1 e5g5 g2g3 d8e8 d4f4 a5a4 h2h4 g5e5 c3c2 a8a5 b2b4 a4b3 a2b3 e8e7 e2f1 g7g6 f1g2');
INSERT INTO evaluations VALUES(10509,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10509,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10509,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10509,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10509,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10509,36,'cp 0','c1d2 e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 b7b6 f1d3 d8e7 e1g1 c8b7 f1d1 c6e5 f3e5 e7e5 c3c2 g7g6 d3f1 b7e4 f1d3');
INSERT INTO evaluations VALUES(10508,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10508,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10508,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10508,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10508,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10508,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10508,36,'cp 0','e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 b7b6 f1e2 c8b7 e1g1 d7d6 f1d1 c6e5 f3e5 e8e5 d1d4 d8f6 d4f4 f6e6 e2g4 e6e7 a1d1 a8e8 g4e2 a7a5 c3c2 e7g5 g2g3 h7h6 b2b3 b7e4 e2d3 e4c6 d3f1 c6e4 f1d3');
INSERT INTO evaluations VALUES(4477,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4477,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4477,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4477,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4477,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4477,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4477,36,'cp 0','e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 b7b6 f1d3 c8b7 e1g1 g7g6 a1d1 d8e7 e3e4 e7b4 c3f6 b4e7');
INSERT INTO evaluations VALUES(33822,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33822,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(33822,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(33822,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33822,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(33822,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33822,36,'cp 3','e4f6 b1c3 e7e2 f1e2 g7g6 f3d4 f8g7 e2f3 b8d7 g2g3 c7c6 a2a4 e8g8 h2h3 a7a5 c1e3 f6e8 f3g2 d6d5 c3e2 e8d6 g3g4 f8e8 e1d2 d7c5 c2c3 c8d7 h1e1 b7b6 b2b3 d6b7 c3c4 d5c4 b3c4 c5a6 g2c6 d7c6 d4c6 g7a1 e1a1 b7c5');
INSERT INTO evaluations VALUES(33798,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(33798,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(33798,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(33798,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(33798,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(33798,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(33798,36,'cp 15','b1c3 e7e2 f1e2 g7g6 f3d4 f8g7 e2f3 b8d7 g2g3 c7c6 e1g1 e8g8 d4b3 a7a5 a2a4 f8e8 c1f4 d6d5 f1e1 g7f8 c3e2 d7c5 b3c5 f8c5 g1g2 h7h5 h2h3 g8g7 f4d2 c8d7 a1d1 d7e6 d1c1 c5b6 d2c3 e6d7');
INSERT INTO evaluations VALUES(33796,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(33796,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(33796,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(33796,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(33796,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33796,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(33796,36,'cp 5','e7e2 f1e2 g7g6 f3d4 f8g7 e2f3 b8d7 g2g3 e8g8 e1g1 d7c5 d4b3 c8h3 f1d1 c5b3 a2b3 c7c6 b3b4 a7a6 d3d4 a8c8 c1f4 f8d8 c3e4 f6e4 f3e4 d8e8 e4f3 d6d5 c2c3 e8e6 f4e3 h7h5 d1e1 c8e8 f3g2 h3g2 g1g2 h5h4 g3g4 g8h8');
INSERT INTO evaluations VALUES(33874,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33874,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(33874,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(33874,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(33874,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(33874,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(33874,36,'cp 8','f1e2 g7g6 f3d4 f8g7 e2f3 b8d7 g2g3 e8g8 e1g1 d7c5 d4b3 c8h3 f1e1 c5b3 a2b3 c7c6 c1f4 d6d5 f3g2 h3f5 b3b4 a7a6 f4e5 f8e8 g1f1 h7h5 h2h3 f6d7 e5g7 g8g7 c3e2 d7f8 e2d4 f8e6 d4e6 f5e6 f2f4 a8d8 f1f2 d5d4 b2b3');
INSERT INTO evaluations VALUES(33873,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33873,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33873,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33873,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33873,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(33873,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(33873,36,'cp 0','g7g6 f3d4 f8g7 e2f3 b8d7 g2g3 e8g8 h2h3 a7a5 a2a4 c7c6 c1f4 d6d5 c3e2 f8e8 f4e3 d7c5 g3g4 f6d7 e1d2 c5e6 a1e1 e6c5');
INSERT INTO evaluations VALUES(49848,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(49848,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(49848,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(49848,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(49848,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(49848,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(49848,36,'cp 63','f8c5 e2e3 b8c6 a2a3 a7a5 b1c3 e8g8 c3e4 c5a7 b2b4 d8e7 c4c5 g4e5 c1b2 e5g6 e4g3 d7d6 c5d6 c7d6 f1b5 g6e5 e1g1 a5b4 b5c6 b7c6 a3b4 f7f6 f3d4 d6d5 d1c2 e7b7 h2h3 a7d4 b2d4 a8a1 f1a1 b7b4');
INSERT INTO evaluations VALUES(42865,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(42865,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(42865,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(42865,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(42865,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(42865,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(42865,36,'cp 72','e2e3 b8c6 b1c3 e8g8 f1e2 c6e5 f3e5 g4e5 f2f4 e5c6 a2a3 f8e8 d1d3 a7a5 c1d2 d7d6 e1g1 c6e7 g1h1 e7f5 e3e4 f5d4 a1e1 d4e2 c3e2 c8d7 f4f5 d7c6 e2f4 a5a4 f1f3 d8h4 f4d5 c6d5 c4d5 f7f6 h2h3 e8e7 f3f4 h4h5 f4g4 a8e8 d3g3');
INSERT INTO evaluations VALUES(42864,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(42864,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(42864,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(42864,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(42864,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(42864,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(42864,36,'cp 74','b8c6 a2a3 a7a5 b1c3 e8g8 c3e4 c5a7 b2b4 f8e8 c1b2 d7d6 h2h3 g4e5 b4b5 c8f5 e4g3 d8f6 d1d2 e5f3 g2f3 a7d4 e1c1 d4b2 d2b2 f6b2 c1b2 c6e7 g3f5 e7f5 f1d3 e8e5 f3f4 e5c5 d3e4 a8b8 d1d5 c5d5 c4d5 f5e7 b2c3 g8f8 c3d4 b8c8 h3h4 c7c6 d5c6 b7c6');
INSERT INTO evaluations VALUES(10601,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(10601,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(10601,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(10601,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(10601,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(10601,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(10601,36,'cp 81','a2a3 a7a5 b1c3 e8g8 c3e4 b7b6 d1d5 d8e7 c1d2 c8b7 d2c3 c5b4 a1d1 b4c3 e4c3 g4e5 f3e5 e7e5 f1d3 e5f6 e1g1 d7d6 d3e4 c6d8 d5d2 b7e4 c3e4 f6g6 e4c3 f7f5 c3d5 g6f7 b2b4 d8c6 h2h3 a5b4 a3b4 a8c8 d2c2 c6e7 d1d4 f8e8 c2d2 e7d5 c4d5');
INSERT INTO evaluations VALUES(10602,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(10602,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(10602,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(10602,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(10602,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(10602,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(10602,36,'cp 72','a7a5 b1c3 e8g8 c3e4 c5a7 b2b4 f8e8 c1b2 d7d6 f1e2 c6e5 c4c5 g4f6 f3e5 d6e5 d1d8 e8d8 e4g3 d8e8 e1g1 c7c6 f1d1 a7b8 b4b5 a5a4 a1c1 c8e6 e2f3 e6b3 d1d2 e5e4 g3e4 f6e4 f3e4 b8h2 g1h2 e8e4');
INSERT INTO evaluations VALUES(7282,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(7282,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(7282,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(7282,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7282,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(7282,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(7282,36,'cp 65','b1c3 e8g8 c3e4 b7b6 d1d5 d8e7 c1d2 c8b7 d2c3 c5b4 a1d1 b4c3 e4c3 g4e5 f3e5 e7e5 f1d3 d7d6 d3e4 e5d5 e4d5 a8b8 f2f3 c6e5 b2b3 g7g6 h2h4 b7d5 d1d5 e5d7 e1f2 d7c5 h1b1 f8e8 g2g4 g8f8 h4h5 e8e5 g4g5 e5d5 c3d5');
INSERT INTO evaluations VALUES(39038,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(39038,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(39038,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(39038,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(39038,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(39038,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(39038,36,'cp 110','b1c3 d7d6 f1d3 a7a6 c1g5 g8e7 d1d2 h7h6 g5e3 b8d7 a2a4 b7b6 e1g1 c7c5 a1d1 g6g5 d4c5 b6c5 d3c4 a8b8 c4b3 d8c7 d2d6 c7d6 d1d6 c5c4 b3c4 b8b2 e3d4 e8g8 f1c1 e7g6 d4g7 g8g7 f3d4 b2b4 c4a6 c8a6 d6d7 g7g8 d7d6 b4c4 d6a6 c4d4');
INSERT INTO evaluations VALUES(39030,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(39030,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(39030,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(39030,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(39030,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(39030,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(39030,36,'cp 111','a7a6 f1d3 d7d6 e4e5 h7h6 h2h4 g8e7 d1e2 b8c6 c1f4 d6e5 d4e5 e7d5 f4g3 d5c3 b2c3 d8e7 e1g1 b7b6 h4h5 g6g5 a1d1 c8b7 f3d4 a8d8 d3a6 c6d4 c3d4 b7a6 e2a6 e8g8 c2c3 d8a8 a6e2 a8a3 c3c4 f8a8 d4d5 e7c5 d5e6 f7e6 e2g4 c5e7');
INSERT INTO evaluations VALUES(40954,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40954,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40954,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(40954,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(40954,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40954,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(40954,36,'cp 30','c1f4 e8g8 e2e3 b7b6 f1d3 c7c5 e1g1 b8c6 c4d5 f6d5 c3d5 e6d5 d3b5 c8b7 b2b3 c6b4 a2a3 b4c6 b5d3 g7g6 f1e1 a8c8 d4c5 b6c5 h2h4 e7h4 f3h4 d8h4 a1c1 h4e7 d1g4 c6a5 f4h6 a5b3 h6f8 c8f8');
INSERT INTO evaluations VALUES(40946,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40946,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40946,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(40946,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40946,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40946,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40946,36,'cp 12','e8g8 e2e3 b8d7 c4c5 c7c6 f1d3 b7b6 b2b4 a7a5 a2a3 c8a6 e1g1 d8c8 h2h3 c8b7 c3a4 a6d3 d1d3 b6b5 a4c3 a8a6 f3d2 f8a8 d2b3 e7d8 a1b1 a5b4 a3b4 a6a3 d3c2 h7h6 b1a1 a3a1 b3a1 d8c7 f4c7 b7c7 a1b3');
INSERT INTO evaluations VALUES(36314,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36314,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36314,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36314,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36314,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36314,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36314,36,'cp 27','e2e3 b8d7 c4c5 c7c6 b2b4 b7b6 a2a3 a7a5 h2h3 c8a6 f1a6 a8a6 e1g1 d8c8 d1c2 c8a8 a1b1 a5b4 a3b4 a8b7 b1a1 f8a8 a1a6 b7a6 f1b1 a6b7 c3a4 b6b5 a4c3 a8a7 f3d2 f6e8 c2d1 b7a8 d2b3 a7a6 b1c1 e7d8 d1c2');
INSERT INTO evaluations VALUES(36312,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36312,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(36312,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(36312,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36312,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36312,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36312,36,'cp 17','b8d7 c4c5 f6h5 a2a3 a7a5 f1d3 c7c6 b2b4 h5f4 e3f4 b7b6 e1g1 c8a6 d3a6 a8a6 d1e2 d8a8 b4b5 c6b5 e2b5 a6a7 f3e5 d7e5 c5b6 e5c4 b6a7 a8a7 f4f5 a7d4 b5d7 c4d2 d7e7 d2f1 f5e6 f1h2 e6f7 f8f7 e7d8 f7f8 d8d5 d4d5 c3d5 h2g4 f2f3 g4f6 a1d1 f8e8 d5f6 g7f6 d1c1 e8e6 g1f2 g8f7');
INSERT INTO evaluations VALUES(3990,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(3990,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(3990,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(3990,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(3990,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(3990,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3990,36,'cp 34','c4c5 c7c6 h2h3 a7a5 a2a3 b7b6 b2b4 c8a6 f1a6 a8a6 e1g1 d8c8 d1c2 c8b7 f1b1 a5b4 a3b4 f8a8 a1a6 b7a6 g2g4 h7h6 g1g2 a6b7 b1b2 d7f8 f3e5 f6d7 e5d3 f8g6 f4g3 b6b5 f2f3 a8a3 b2a2 a3a2 c2a2');
INSERT INTO evaluations VALUES(55513,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55513,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(55513,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(55513,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(55513,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55513,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(55513,36,'cp 28','d6e5 d1d8 e8d8 f1c4 f7f6 f2f4 b8d7 c1e3 f8b4 b1d2 e5f4 e3f4 g8e7 e1c1 e7g6 f4g3 d7e5 g1f3 d8e7 f3e5 g6e5 g3e5 f6e5 h1f1 c8e6 c4e6 e7e6 c2c3 b4e7 c1c2 a8f8 a2a4 f8f1 d1f1 h8f8 b2b4 f8f1 d2f1');
INSERT INTO evaluations VALUES(56614,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(56614,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(56614,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(56614,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(56614,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(56614,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(56614,36,'cp 37','d1d8 e8d8 f1c4 f7f6 f2f4 b8d7 g1f3 f8d6 f4f5 a7a5 c1e3 d6c5 e3g1 c7c6 a2a4 c5g1 h1g1 d7c5 b1c3 g8e7 e1c1 c8d7 f3e1 d8c7 e1d3 c5d3 d1d3 a8d8 g1d1 d7c8 d3h3 d8d1 c3d1 c8d7 d1e3 b7b5 c4b3 b5a4 b3a4 e7c8 e3c4');
INSERT INTO evaluations VALUES(35407,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(35407,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(35407,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(35407,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(35407,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(35407,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(35407,36,'cp 37','e8d8 f1c4 f7f6 f2f4 b8d7 g1f3 f8d6 f4f5 d6c5 a2a3 a7a5 b1c3 c7c6 a3a4 d7b6 c4b3 c8d7 c1d2 g8h6 e1c1 d8c7 f3e1 c5d4 d2h6 g7h6 d1d3 h6h5 d3g3 b6c8 c3e2 h5h4 g3h3 d4b6 h3h4 c8d6 e2c3 b6e3 c1b1');
INSERT INTO evaluations VALUES(35460,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(35460,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(35460,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35460,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(35460,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(35460,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(35460,36,'cp 33','f1c4 f7f6 f2f4 b8d7 c1e3 g8h6 h2h3 f8d6 f4f5 d6b4 e1f2 b4c5 g1f3 c5e3 f2e3 d7b6 c4e2 h6f7 h1d1 c8d7 c2c4 c7c5 b2b3 d8e7 b1c3 d7c6 g2g4 b6d7 h3h4 h7h6 f3e1 h8g8 d1d3 a7a5 a1d1 a5a4 c3a4 c6a4 b3a4');
INSERT INTO evaluations VALUES(35454,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(35454,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(35454,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35454,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(35454,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35454,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(35454,36,'cp 45','f7f6 f2f4 b8d7 g1f3 f8d6 f4f5 d6c5 e1e2 g8e7 b1c3 d7b6 h1d1 d8e8 c4d3 c8d7 a2a4 a7a5 g2g4 e7c8 g4g5 c8d6 b2b3 e8e7 h2h4 h7h6 g5g6 c7c6 c1e3 c5e3 e2e3 a8d8 f3d2 d7e8 d1g1 h6h5 a1d1 d8d7 d2c4 b6c4 d3c4 d6c4 b3c4 d7d1');
INSERT INTO evaluations VALUES(35469,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(35469,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(35469,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(35469,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(35469,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(35469,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(35469,36,'cp 32','f2f4 b8d7 g1f3 f8d6 f4f5 d6c5 b1c3 d7b6 c4b3 a7a5 a2a3 g8e7 c1d2 c8d7 e1c1 e7c8 h1e1 c8d6 d2e3 c5e3 e1e3 d8e7 h2h4 d7c6 g2g4 a8d8 g4g5 h7h6 g5g6 a5a4 b3e6 c6d7 e6a2 d7c6 h4h5 d6c8');
INSERT INTO evaluations VALUES(55598,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(55598,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(55598,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(55598,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(55598,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(55598,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(55598,36,'cp 66','d2d4 e5d4 d1d4 b8c6 d4d3 g8f6 c1e3 c8e6 e1c1 f6g4 f2f4 g7g6 g1f3 f8h6 g2g3 e8g8 e3g1 g4f6 h2h4 h6g7 f3d4 e6g4 f1e2 h7h5 e2g4 f6g4 d4c6 b7c6 g1d4 g7d4 d3d4 d8f6 d4d2 a8b8 a2a3 c6c5 h1e1 f8e8 b2b3 f6g7');
INSERT INTO evaluations VALUES(55583,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(55583,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(55583,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(55583,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(55583,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(55583,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(55583,36,'cp 37','e5d4 d1d4 b8c6 d4d3 g8f6 c1g5 h7h6 g5e3 f8e7 e1c1 e8g8 f2f4 a8b8 g1f3 f8e8 h2h3 e7f8 e4e5 f6d7 g2g4 d6e5 f4e5 b7b5 g4g5 c8b7 g5h6 c6e5 f3e5 d7e5 d3f5 d8h4 f1b5 b7h1 b5e8 b8e8 d1h1 e5c4 f5f4 h4h6 f4h6 g7h6 h1g1 f8g7 e3a7');
INSERT INTO evaluations VALUES(55682,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(55682,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(55682,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(55682,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(55682,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(55682,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(55682,36,'cp 58','d1d4 b8c6 d4d3 g7g6 c1e3 f8g7 e1c1 c8e6 f2f3 a7a6 d3d2 d8d7 g1e2 g8e7 e2f4 e8c8 f4e6 f7e6 c1b1 c8b8 h2h4 d7e8 a2a3 d6d5 e4d5 e6d5 e3f2 e8f7 c3a4 h7h5 a4c5 f7f6 c2c3 e7f5 f1d3 c6e5 d2c2 e5d3');
INSERT INTO evaluations VALUES(55677,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(55677,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(55677,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(55677,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(55677,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(55677,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(55677,36,'cp 49','b8c6 d4d3 g8f6 c1f4 c6e5 d3e3 c7c6 e1c1 f8e7 h2h3 d8b6 a2a4 c8e6 g1e2 f6h5 e3b6 a7b6 f4h2 g7g5 e2d4 h5f6 g2g3 e5d7 f1g2 h7h5 f2f4 g5g4 h3g4 e6g4 d1e1 h5h4 g3h4 h8h4 d4f3 h4h5 f3g5 h5h4 h2g1 h4h1 g2h1 f6h5 g1e3 e7g5 f4g5 e8c8');
INSERT INTO evaluations VALUES(22035,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(22035,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(22035,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(22035,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(22035,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(22035,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(22035,36,'cp 74','d4d3 g8f6 c1e3 c8e6 e1c1 f6g4 f2f4 g7g6 g1f3 f8h6 c1b1 e8g8 h2h3 g4e3 d3e3 d8f6 g2g3 a8e8 a2a3 h6g7 g3g4 c6a5 h1g1 e6c4 g4g5 f6e7 f3d2 e7e6 h3h4 c4f1 g1f1 a5c4 e3g3 c4d2 d1d2 g7c3 g3c3 e6e4 c3c7');
INSERT INTO evaluations VALUES(52883,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52883,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(52883,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(52883,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(52883,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(52883,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52883,36,'cp 34','e2e3 b7b5 a2a4 b5b4 c3e4 d8d5 e4d2 c4c3 b2c3 b4c3 d2b1 d5a5 d1c2 b8a6 c2c3 a6b4 c1d2 e7e6 g1f3 g8f6 f1c4 f8d6 e1g1 e8g8 c3b2 a8b8 b1a3 f6e4 d2b4 b8b4 b2c2 e4g5 f3g5 a5g5 c4d3 c6c5 a3b5 c8b7 e3e4 b4d4 b5d4 c5d4 a1c1 f8c8');
INSERT INTO evaluations VALUES(52881,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(52881,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52881,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(52881,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(52881,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(52881,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(52881,36,'cp 31','b7b5 a2a4 b5b4 c3e4 d8d5 e4d2 c4c3 b2c3 b4c3 d2b1 d5a5 d1c2 b8a6 c2c3 a6b4 c1d2 e7e6 g1f3 g8f6 f1c4 f8e7 e1g1 e8g8 f1c1 a8b8 d2e1 c6c5 b1d2 c8b7 f3e5 c5d4 c3d4 f8c8 d2b3 a5b6 a4a5 b6d4 b3d4 b7e4 c4e2 c8c1 a1c1 e7d6 d4c6 b4c6 e5c6');
INSERT INTO evaluations VALUES(51646,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(51646,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(51646,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(51646,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(51646,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(51646,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(51646,36,'cp 40','a2a4 b5b4 c3e4 d8d5 e4d2 c4c3 b2c3 b4c3 d2b1 d5a5 d1c2 b8a6 c2c3 a6b4 c1d2 e7e6 g1f3 g8f6 f1c4 f8e7 e1g1 e8g8 f1c1 a8b8 c3b3 a5d8 d2c3 c6c5 d4c5 e7c5 c1d1 d8e7 c3e5 b8b7 b1d2 b4c6 e5f6 g7f6 b3c3 c5b4 c3b2 b4d2 b2d2 f8d8 d2c3 d8d1 a1d1');
INSERT INTO evaluations VALUES(51657,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(51657,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51657,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(51657,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(51657,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(51657,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(51657,36,'cp 34','b5b4 c3e4 d8d5 e4d2 c4c3 b2c3 b4c3 d2b1 d5a5 d1c2 b8a6 c2c3 a6b4 c1d2 e7e6 g1f3 g8f6 f1c4 f8e7 e1g1 e8g8 f1c1 a8b8 d2e1 c6c5 b1d2 c5d4 c3d4 b4c6 d4f4 c8b7 d2e4 a5h5 e4g3 h5g4 f4g4 f6g4 a4a5 b7a8 a5a6 g4f6 a1b1 b8b1 c1b1');
INSERT INTO evaluations VALUES(51667,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51667,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51667,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(51667,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(51667,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51667,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(51667,36,'cp 46','c3e4 d8d5 e4d2 c4c3 b2c3 b4c3 d2b1 d5a5 d1c2 b8a6 c2c3 a6b4 c1d2 e7e6 g1f3 g8f6 f1c4 f8e7 e1g1 e8g8 f1c1 a8b8 d2e1 c8b7 b1d2 a5f5 d2b3 c6c5 d4c5 b7f3 b3d4 f5c5 d4f3 f8c8 a4a5 f6e4 c3b3 b4d5 b3d3 e4d6 c4d5 c5d5 d3d5 e6d5 c1c8 b8c8 a1d1');
INSERT INTO evaluations VALUES(45458,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(45458,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45458,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(45458,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(45458,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45458,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(45458,36,'cp 33','c2c3 b8c6 c1f4 a7a6 f1d3 c8g4 h2h3 g4h5 d1e2 e7e6 b1d2 f8d6 f4d6 d8d6 e2e3 d6c7 e1g1 e8g8 f1e1 f8e8 f3e5 a8c8 d2f3 c6e5 f3e5 f6d7 e5d7 c7d7 a2a4 d7c7 a4a5 h5g6 d3g6 h7g6 h3h4 f7f6 e1e2 e6e5 d4e5 e8e5 e3d2 e5e2');
INSERT INTO evaluations VALUES(45446,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(45446,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(45446,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45446,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(45446,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(45446,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(45446,36,'cp 21','c8f5 c1f4 b8c6 d1b3 d8d7 b1d2 e7e6 f3e5 d7c8 f1d3 f6h5 d3f5 h5f4 g2g3 c6e5 d4e5 f4g6 f5g6 h7g6 e1g1 a7a5 a1c1 a5a4 b3d1 c8d7 g1g2 f8e7 d2f3 d7b5 d1d2 b5c4 b2b4 c4g4 h2h4 a8c8 f1h1 c8c4 c1e1 a4a3 f3d4 h8h4 h1h4');
INSERT INTO evaluations VALUES(29139,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29139,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(29139,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29139,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(29139,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(29139,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29139,36,'cp 24','c1f4 b8c6 d1b3 d8d7 f3e5 d7c8 b1d2 e7e6 f1d3 f5d3 e5d3 f6h5 f4e5 f7f6 e5g3 h5g3 h2g3 f8d6 e1g1 c8d7 f1e1 e8c8 b3a4 h8e8 d3c5 d6c5 d4c5 c8b8 b2b4 c6e5 a4b3 h7h5 a2a4 g7g5 b4b5 h5h4 g3h4 g5h4 c5c6');
INSERT INTO evaluations VALUES(52389,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52389,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(52389,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52389,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52389,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(52389,36,'cp 0','d7d5 d2d3 d5e4 d3e4 d8d1 e1d1 g8f6 e4e5 f6d5 g1f3 f7f6 c2c4 d5c7 c1e3 g7g6 b1d2 f8g7 e5f6 g7f6 d1c1 c8f5 f3d4 f6d4 e3d4 h8f8 d4e3 c7e6 f1e2 h7h5 b2b4 c6c5 b4b5 b8d7 h1e1 e6d4 e3d4 c5d4 c1b2 a8d8 a1d1 d7c5');
INSERT INTO evaluations VALUES(52822,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(52822,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(52822,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(52822,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(52822,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(52822,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52822,36,'cp 0','d2d3 d5e4 d3e4 d8d1 e1d1 g8f6 e4e5 f6d5 g1f3 c8f5 f1d3 f5d3 c2d3 b8a6 b1c3 a6c7 c1d2 e7e6 d1e2 f8e7 g2g4 h7h5 h2h3 a7a5 c3e4 a5a4 a1g1 g7g6 g1c1 c7b5 g4g5 e8c8 c1c4 d5b6 c4c1');
INSERT INTO evaluations VALUES(52820,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(52820,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52820,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52820,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52820,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52820,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(52820,36,'cp 0','d5e4 d3e4 d8d1 e1d1 g8f6 e4e5 f6d5 g1f3 c8f5 f1d3 f5d3 c2d3 b8a6 b1c3 a6c7 d1e2 e7e6 c1d2 f8e7 a1c1 h7h5 g2g3 f7f6 a2a4 f6f5 h2h4 e8d7 a4a5 b7b6 c3a4 h8g8 d3d4 a8b8 c1c2 b6b5 a4c3 d5c3 b2c3 g7g6 h1a1 c7d5 a1c1');
INSERT INTO evaluations VALUES(52931,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(52931,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(52931,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(52931,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(52931,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(52931,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52931,36,'cp 13','d3e4 d8d1 e1d1 g8f6 e4e5 f6d5 g1f3 c8f5 f1d3 f5d3 c2d3 b8a6 b1c3 a6c7 d1e2 e7e6 c1d2 a7a5 a2a4 e8c8 a1c1 f7f5 c3d5 d8d5 h2h4 h7h5 d2e3 f8e7 h1d1 c8b8 g2g3 h8d8 e3b6 d8d7 d3d4 b8c8 b2b3 c8b8 b6c7 b8c7 d1g1 b7b6');
INSERT INTO evaluations VALUES(57111,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(57111,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(57111,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(57111,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(57111,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(57111,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(57111,36,'cp -111','b8c6 f1b5 g8f6 h5d1 c6e7 b5e2 c7c6 c2c4 d7d5 d2d4 e5e4 b1c3 h7h5 f2f3 e7f5 c4d5 f6d5 c3d5 d8d5 d1b3 d5b3 a2b3 f8b4 e1d1 f5d6 c1d2 b4d2 d1d2 e8g8 f3e4 d6e4 d2e1 f8e8 g1f3 a7a5 e2c4 e4f6 f3e5 c8e6 c4e6 e8e6');
INSERT INTO evaluations VALUES(22697,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(22697,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(22697,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(22697,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(22697,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(22697,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(22697,36,'cp -81','f1b5 g8f6 h5d1 c6e7 b5e2 d7d5 d2d4 e5e4 c2c4 c7c6 c1d2 c8e6 d1b3 d5c4 e2c4 e6c4 b3c4 e7d5 b1c3 f8d6 c3d5 f6d5 c4c2 d8h4 g1e2 e8g8 e2c3 f8e8 c3d5 c6d5 c2b3 h4h5 h2h3 e8e6 a1c1 e6h6 b3b7 a8b8 b7a7 h6g6');
INSERT INTO evaluations VALUES(22639,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(22639,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(22639,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(22639,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(22639,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(22639,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(22639,36,'cp -111','g8f6 h5d1 d7d5 d2d4 e5d4 e3d4 f8d6 g1f3 e8g8 e1g1 c8g4 b5e2 f8e8 b1c3 d8d7 c3b5 d6f8 h2h3 g4h5 c1e3 h7h6 b5c3 f8b4 c3a4 b7b6 a2a3 b4d6 a4c3 a7a6 f1e1 e8e6 f3h4 h5e2 c3e2 c6e7 e3f4 d6f4 e2f4 e6e4');
INSERT INTO evaluations VALUES(16305,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(16305,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(16305,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(16305,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(16305,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(16305,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(16305,36,'cp -96','h5d1 d7d5 d2d4 e5d4 e3d4 f8d6 g1f3 e8g8 e1g1 c8g4 c2c3 c6e7 f1e1 e7g6 h2h3 g4h5 b5e2 f8e8 b1d2 g6f4 e2f1 d8d7 d1b3 e8e1 f3e1 d7f5 b3c2 f4e2 f1e2 f5c2 e1c2 h5e2 d2f1 c7c6 c1g5 f6e4 g5e3 e2d3 c2e1 d3c4 f1d2 e4d2 e3d2 c4b5');
INSERT INTO evaluations VALUES(14220,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14220,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(14220,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(14220,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(14220,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(14220,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(14220,36,'cp 12','c8f5 f1d3 f5d3 d1d3 e7e6 g1f3 f8d6 f4g3 e8g8 g3h4 d6e7 b1d2 c6b8 h4f6 e7f6 e3e4 b8d7 e4d5 e6d5 e1g1 c7c6 f1e1 d8b6 g2g3 a7a5 c3c4 d5c4 d2c4 b6c7 a1d1 a8e8 a2a4 g7g6 g1g2 f6g7 d4d5 e8e1');
INSERT INTO evaluations VALUES(24816,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(24816,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24816,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(24816,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24816,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(24816,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24816,36,'cp 20','f1d3 f5d3 d1d3 e7e6 g1f3 f8d6 f4g3 e8g8 g3h4 a7a5 e1g1 d6e7 h4f6 e7f6 f1d1 a5a4 c3c4 d8d7 d3e2 d5c4 b1a3 c6a5 a3c4 a5c4 e2c4 f8d8 g2g3 c7c6 g1g2 d7d5 d1c1 e6e5 c4d5 d8d5 d4e5 f6e5 f3e5 d5e5');
INSERT INTO evaluations VALUES(24822,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24822,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(24822,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(24822,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24822,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(24822,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(24822,36,'cp 18','f5d3 d1d3 e7e6 g1f3 f8d6 f4g3 e8g8 g3h4 d6e7 b1d2 c6b8 d3b5 b7b6 e1g1 c7c6 b5e2 b8d7 a1d1 f8e8 h2h3 d8c7 h4g3 e7d6 g3d6 c7d6 f1e1 b6b5 e3e4 d5e4 d2e4 f6e4 e2e4 d6d5');
INSERT INTO evaluations VALUES(24898,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24898,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24898,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(24898,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24898,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24898,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(24898,36,'cp 15','d1d3 e7e6 g1f3 f8d6 f4g3 e8g8 g3h4 d6e7 b1d2 c6b8 d3b5 b7b6 e1g1 c7c6 b5e2 c6c5 a1c1 b8d7 h2h3 h7h6 f3e5 a8c8 f2f4 f6e4 h4e7 d8e7 d2e4 d5e4 c1d1 f8d8 a2a4 g8h8 d1d2 c5d4 d2d4 d7e5 f4e5');
INSERT INTO evaluations VALUES(24897,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24897,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24897,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24897,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(24897,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24897,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24897,36,'cp 15','e7e6 g1f3 f8d6 f4g3 e8g8 g3h4 a8c8 b1d2 c6b8 e1g1 d6e7 h4f6 e7f6 e3e4 c7c6 g2g3 f6e7 f1e1 b8d7 g1g2 d8c7 a2a3 h7h6 b2b4 a7a5 d3e2 d5e4 d2e4 c8d8 e4c5 b7b6 c5d7 c7d7 f3e5 d7c7 b4a5 b6a5');
INSERT INTO evaluations VALUES(25199,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(25199,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(25199,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(25199,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(25199,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(25199,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(25199,36,'cp 14','g1f3 f8d6 f4g3 e8g8 b1d2 c6e7 e3e4 d5e4 d2e4 f6e4 d3e4 d8d7 e1g1 d7b5 e4c2 a7a5 a1d1 b7b6 f1e1 h7h6 h2h3 b5c6 a2a3 f8e8 c2e2 a8d8 g3h4 d8d7 h4e7 d7e7 e2c2 c6a8 f3e5 d6e5 e1e5 e7d7');
INSERT INTO evaluations VALUES(25200,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25200,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(25200,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(25200,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(25200,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(25200,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(25200,36,'cp 13','f8d6 f4g3 e8g8 g3h4 d6e7 b1d2 c6b8 h4f6 e7f6 e3e4 b8d7 e4d5 e6d5 e1g1 c7c6 f1e1 d8c7 e1e2 f8e8 a1e1 e8e2 e1e2 d7f8 g2g3 c7d7 d2b3 b7b6 g1g2 a8e8 e2e8 d7e8 b3c1 h7h6 h2h4 e8d7 h4h5 c6c5 d4c5');
INSERT INTO evaluations VALUES(24114,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(24114,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(24114,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24114,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(24114,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(24114,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(24114,36,'cp 11','f4g3 e8g8 g3h4 d6e7 b1d2 c6b8 h4f6 e7f6 e3e4 b8d7 e4d5 e6d5 e1g1 c7c6 f1e1 d8c7 g2g3 f8e8 a2a4 g7g6 g1g2 f6g7 a4a5 e8e1 f3e1 a8e8 d3f3 d7f6 e1d3 h7h5 h2h4 f6g4 a5a6 b7b6 a1e1 e8e7 e1e2 e7e2 f3e2 g4f6');
INSERT INTO evaluations VALUES(24118,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24118,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(24118,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(24118,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24118,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24118,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24118,36,'cp 15','e8g8 g3h4 d6e7 b1d2 c6b8 h4f6 e7f6 e1g1 b8d7 e3e4 c7c6 f1d1 d8c7 c3c4 d5e4 d2e4 f6e7 g2g3 a8d8 d3e2 h7h6 a2a3 f8e8 e4c3 d7f6 b2b4 a7a5 a1b1 b7b5 c4b5 a5b4 a3b4 c6b5 d1c1 c7b6 e2b5 b6b5 c3b5 f6d5');
INSERT INTO evaluations VALUES(21400,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(21400,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21400,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(21400,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(21400,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(21400,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(21400,36,'cp 46','f8d6 f4g3 g8f6 c2c4 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 a1c1 c8f5 e1g1 f5d3 d1d3 a7a5 c1c2 c7c6 f1c1 f8e8 g3d6 d8d6 d3b3 e8b8 h2h3 h7h6 b3c3 b8e8 f3e1 g7g5 e1f3 g8g7 f3e5 a8c8');
INSERT INTO evaluations VALUES(5921,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(5921,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(5921,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(5921,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(5921,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(5921,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(5921,36,'cp 61','f4g3 g8f6 c2c4 e8g8 b1c3 f6e4 d1c2 e4c3 b2c3 h7h6 c4d5 e6d5 f1d3 c8e6 e1g1 d8d7 f1e1 f8e8 a2a4 b7b6 a1d1 c6e7 c3c4 d5c4 d3c4 e6c4 c2c4 d7c6 d1c1 c6c4 c1c4 d6g3 h2g3 c7c6 e1c1 a8c8 e3e4 c6c5 d4c5');
INSERT INTO evaluations VALUES(5924,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(5924,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(5924,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(5924,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(5924,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5924,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(5924,36,'cp 58','g8f6 c2c4 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6b4 d3b1 g7g6 a2a3 b4c6 e1g1 c6e7 b2b4 c8g4 h2h3 g4f3 d1f3 a7a5 b4b5 d6g3 f3g3 d8d6 g3d6 c7d6 b1d3 a8c8 f1c1 c8c1 a1c1 f8c8 c1c8 e7c8 g2g4 h7h6 g1g2 c8e7 g2g3 f7f5 h3h4 f5g4');
INSERT INTO evaluations VALUES(9259,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(9259,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(9259,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(9259,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(9259,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(9259,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(9259,36,'cp 54','c2c4 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 e1g1 c8f5 a1c1 f5d3 d1d3 a7a5 c1c3 c7c6 f1c1 h7h6 g3d6 d8d6 d3c2 f8c8 h2h4 h6h5 f3e1 g7g6 g2g3 d6e6 e1d3 c8e8 d3f4 e6d6 a2a3 e8c8 f4d3 g8g7 b2b4 d6f6 b4a5 a8a5 d3b4');
INSERT INTO evaluations VALUES(47204,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47204,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47204,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47204,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47204,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(47204,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(47204,36,'cp 5','c7c5 f1g2 c5d4 e1g1 e7e6 f3d4 e6e5 d4f3 b8c6 c2c4 d5d4 e2e3 f8e7 e3d4 e5d4 c1f4 e8g8 f3e5 d8b6 b1d2 c6e5 f4e5 f6g4 d2f3 g4e5 f3e5 c8f5 g2d5 b6b2 d1f3 g7g6 g3g4 f5e6 d5e6 f7e6 f3e4 b2a3 e5g6 h7g6 e4g6 g8h8');
INSERT INTO evaluations VALUES(45386,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(45386,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(45386,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45386,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45386,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45386,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(45386,36,'cp 0','f1g2 c5d4 e1g1 h7h6 b2b3 e7e6 c1b2 f8c5 f3d4 e8g8 c2c4 e6e5 d4f3 e5e4 f3d4 b8c6 d4c6 b7c6 b1c3 e4e3 f2f4 d5c4 c3a4 c5e7 d1d8 f8d8 g2c6 a8b8 b2e5 b8b4 e5c3 b4b8');
INSERT INTO evaluations VALUES(45387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45387,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45387,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45387,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45387,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(45387,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45387,36,'cp 8','c5d4 e1g1 h7h6 b2b3 e7e6 c1b2 f8c5 f3d4 e8g8 c2c4 e6e5 d4f3 e5e4 f3d4 b8c6 c4d5 d8d5 d4c6 d5c6 b1c3 c8g4 h2h3 g4f5 g3g4 f5g6 c3a4 c5e7 a1c1 c6e6 a4c5 e6b6 c5d7 f6d7 d1d7 e7d6 d7a4 f7f5 g4f5 g6f5 a4c4 g8h8');
INSERT INTO evaluations VALUES(45465,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45465,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45465,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45465,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45465,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(45465,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(45465,36,'cp 9','e1g1 h7h6 b2b3 e7e6 c1b2 f8c5 b2d4 c5d4 d1d4 b8c6 d4b2 e8g8 b1a3 d8e7 c2c4 b7b6 c4d5 f6d5 a3c4 c8b7 a1c1 b6b5 c4e5 c6e5 f3e5 b5b4 f1d1 f8c8 g2d5 b7d5 c1c8 a8c8 d1c1 c8c7 f2f3 e7d6 e5c4 d6c6 g1g2 c7c8');
INSERT INTO evaluations VALUES(57580,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(57580,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(57580,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(57580,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(57580,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(57580,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(57580,36,'cp 56','g8f6 e2e3 e7e6 f1e2 b8c6 g1f3 d7d6 h2h3 g7g6 c2c4 f6e4 b1d2 e6e5 d4e5 d6e5 f4e5 c6e5 f3e5 d8f6 e5d3 c8d7 e2f3 d7c6 d2b3 e8c8 b3d4 c6d7 d1c2 c7c5 d4e2 d7c6 e2f4 g6g5 f4d5 c6d5 c4d5 d8d5 e1c1 c8b8');
INSERT INTO evaluations VALUES(50092,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(50092,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(50092,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(50092,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(50092,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(50092,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(50092,36,'cp 55','e2e3 e7e6 f1e2 f8b4 c2c3 b4e7 c3c4 e8g8 a2a3 d7d6 b1c3 b8c6 f4g3 e6e5 g1f3 e5d4 f3d4 c6d4 d1d4 c8e6 e1g1 f6d7 d4d2 e7f6 a1d1 a7a5 b2b4 a5b4 a3b4 d7b6 d2c2 d8e7 c3b5 e7f7 c4c5 e6b3');
INSERT INTO evaluations VALUES(50089,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(50089,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(50089,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50089,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(50089,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(50089,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(50089,36,'cp 51','e7e6 f1e2 b8c6 g1f3 d7d6 e1g1 h7h6 c2c4 g7g5 f4g3 f8g7 b1c3 e8g8 a1c1 a7a6 h2h3 g8h8 g3h2 b7b6 d4d5 c6e7 d5e6 c8e6 f3d4 e6g8 d1c2 d8d7 f1d1 a8e8 b2b4 e7c6 d4c6 d7c6 c3d5 c6d7 d5f6 f8f6');
INSERT INTO evaluations VALUES(49950,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(49950,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(49950,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(49950,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(49950,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(49950,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(49950,36,'cp 67','f1e2 f8e7 c2c4 e8g8 a2a3 d7d6 b1c3 b8c6 f4g3 e6e5 g1f3 e5d4 f3d4 c6d4 d1d4 c8e6 a1c1 a7a5 e1g1 f6d7 b2b4 e7f6 d4d2 f6c3 c1c3 a5b4 a3b4 d8f6 e2f3 a8a2 c3c2 a2c2 d2c2 b7b6 f1d1 d7e5 f3d5 g8h8 h2h3 h7h6 g1h1 e5d7 g3h2 e6d5 c4d5');
INSERT INTO evaluations VALUES(55283,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(55283,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(55283,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(55283,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(55283,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(55283,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55283,36,'cp 14','c3e4 g8f6 e4f6 e7f6 d2d4 f8d6 f1e2 e8g8 g1f3 c7c5 c1e3 c5d4 f3d4 b8c6 c2c3 c8d7 d4c6 d7c6 e2f3 d8c7 f3c6 c7c6 e1g1 b7b6 d1f3 c6f3 g2f3 f8d8 a2a4 g8f8 f1d1 d6e7 b2b4 f8e8 d1d8 a8d8 a4a5 d8d3 a5b6 a7b6 e3b6');
INSERT INTO evaluations VALUES(55275,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55275,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55275,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55275,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55275,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55275,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55275,36,'cp 0','b8c6 g1f3 g8f6 e4f6 e7f6 f1e2 f8d6 d2d4 e8g8 e1g1 f8e8 c2c4 c8g4 c1e3 f6f5 d4d5 f5f4 e3d2 e8e2 d1e2 c6d4 e2e4 g4f3 g2f3 d8g5 g1h1 g5h5 f1e1 d4f3 d2f4 h7h6 b2b3 a8d8 e1e3 f3d2 e4d4 d2f3 d4e4');
INSERT INTO evaluations VALUES(21148,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21148,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21148,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21148,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21148,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21148,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21148,36,'cp 0','g1f3 g8f6 e4f6 e7f6 d2d4 f8d6 f1e2 e8g8 e1g1 f8e8 c2c4 c8g4 c1e3 f6f5 d4d5 f5f4 e3d2 e8e2 d1e2 c6d4 e2e4 g4f3 g2f3 d8g5 g1h1 g5h5 e4d4 h5f3 h1g1 d6e5 d4e5 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(21147,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21147,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21147,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21147,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21147,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21147,36,'cp 0','g8f6 e4f6 e7f6 d2d4 f8d6 f1e2 e8g8 c2c4 f8e8 e1g1 c8g4 c1e3 f6f5 d4d5 f5f4 e3d2 e8e2 d1e2 c6d4 e2e4 g4f3 g2f3 d8g5 g1h1 g5h5 e4d4 h5f3 h1g1 d6e5 d4e5 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(18801,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(18801,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18801,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18801,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18801,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(18801,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18801,36,'cp 13','d7d5 e4e5 c8d7 f1e2 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1e3 f5e3 f2e3 f8e7 e2d3 e8g8 a1c1 a8c8 d1e1 f7f6 e5f6 f8f6 e1h4 f6h6 h4f2 d7e8 c3e2 h6f6 e2f4 h7h6 h2h3 e7d6 f2g3 c6b4');
INSERT INTO evaluations VALUES(20285,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(20285,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(20285,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(20285,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(20285,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(20285,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(20285,36,'cp 20','e4e5 c8d7 f1e2 g8e7 e1g1 e7g6 c1e3 d8b6 b1a3 f8e7 a3c2 e8g8 a1b1 c5d4 c3d4 a8c8 b2b4 a7a5 h2h4 g6h4 f3h4 e7h4 b4b5 c6d8 a2a4 f7f5 e2d3 h4e7 g2g3 d8f7 f2f4 g8h8');
INSERT INTO evaluations VALUES(29543,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(29543,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(29543,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(29543,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(29543,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(29543,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(29543,36,'cp 13','c2c4 e7e6 e2e3 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 b8d7 e1g1 e8g8 d1b3 d7f6 f1e1 a7a5 a1c1 a5a4 b3a3 f6e4 b2b3 a4b3 a3b3 e4d2 f3d2 d6d5 c1c3 b6c8 e1c1 c8d6 a2a4 d8a5 d2f3 f8c8 c3c8 a8c8');
INSERT INTO evaluations VALUES(29517,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29517,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29517,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(29517,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29517,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(29517,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(29517,36,'cp 17','e7e6 e2e3 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 b8c6 e1g1 e8g8 a1c1 a8c8 d1b3 c8c7 b5c6 b7c6 b3a3 d8e7 e3e4 f8c8 h2h3 c6c5 d4c5 c7c5 c1c5 c8c5 e4e5 d6e5 d2b3 c5c7 a3e7 c7e7 f3e5 e7c7');
INSERT INTO evaluations VALUES(30066,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(30066,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(30066,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(30066,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(30066,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(30066,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(30066,36,'cp 23','e2e3 f5b1 a1b1 f8b4 e1e2 d5c4 d1a4 b8c6 f3e5 f6d5 e5c6 d8d7 b1c1 b7c6 c1c4 a8b8 a4c6 d5f4 e3f4 b4d6 b2b3 d6f4 c6d7 e8d7 c4a4 f4d6 a4a5 b8b4 e2e3 h8a8 f1b5 c7c6 b5a4 b4b6 h1c1 d7c7 d4d5 e6d5 a5d5 a8e8 e3f3 e8e5 d5d4 g7g6 g2g3');
INSERT INTO evaluations VALUES(18822,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(18822,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(18822,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(18822,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18822,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(18822,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(18822,36,'cp 45','d2d4 d7d5 e4e5 c8d7 f1e2 g8e7 e1g1 e7g6 c1e3 d8b6 b1a3 c5d4 c3d4 g6e7 a3b5 e7f5 b5c3 f5e3 f2e3 f8e7 e2d3 c6b4 d3b1 b4c6 c3a4 b6d8 a2a3 c6a5 a4c5 d7b5 b1d3 b5d3 d1d3');
INSERT INTO evaluations VALUES(46196,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46196,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46196,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(46196,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46196,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(46196,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(46196,36,'cp 22','d8b6 d1c2 f8d6 f4g3 e8g8 b1d2 a7a5 f1d3 b8d7 e1g1 b6c7 c3c4 c5d4 f3d4 d6g3 h2g3 b7b6 c4d5 c7c2 d3c2 f6d5 f1d1 c8a6 c2b3 d5f6 b3c4 a6c4 d2c4 f8c8 c4d6 c8c5 a1c1 g8f8 d4c6 c5c1 d1c1 f6d5 a2a3 g7g6');
INSERT INTO evaluations VALUES(33618,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33618,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(33618,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(33618,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(33618,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(33618,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(33618,36,'cp 14','d1c2 f8d6 f4g3 e8g8 b1d2 a7a5 f1e2 a5a4 e3e4 f6e4 d2e4 d5e4 g3d6 b6d6 c2e4 b8d7 a2a3 c5d4 e4d4 d6d4 f3d4 d7c5 f2f4 c8d7 e1c1 g7g6 h1e1 h7h5 e2d3 a8c8 g2g3 c5d3 d1d3 f8d8 e1d1 d7e8 d3d2 d8d5 d4c2 d5d2 d1d2 g8g7 c2d4 e8c6');
INSERT INTO evaluations VALUES(33619,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(33619,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(33619,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(33619,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(33619,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33619,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(33619,36,'cp 9','f8d6 f4g3 e8g8 b1d2 b8d7 a1c1 b6c7 c3c4 b7b6 c4d5 e6d5 b2b3 c8b7 f1d3 a8c8 e1g1 d6g3 h2g3 f8e8 f1e1 c7b8 c2b2 g7g6 d4c5 b6c5 g3g4 h7h6 d3e2 c8c7 b2a3 a7a5');
INSERT INTO evaluations VALUES(18740,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(18740,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(18740,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(18740,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(18740,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(18740,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(18740,36,'cp -39','d7d5 e4d5 e6d5 c4b3 d8e7 d1e2 e7e2 e1e2 c8e6 h1e1 h7h6 b3a4 g8e7 d3d4 e7f5 d4c5 f8c5 e2f1 e8g8 b1c3 a8c8 c3e2 g7g5 c2c3 f8d8 a4c2 d8e8 c2a4 c5b6 h2h3 c8d8 c1d2 d5d4 e2d4 f5d4 c3d4');
INSERT INTO evaluations VALUES(20243,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(20243,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(20243,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(20243,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(20243,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(20243,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(20243,36,'cp -24','e4d5 e6d5 c4b3 d8e7 d1e2 e7e2 e1e2 c8e6 h1e1 h7h6 b3a4 g8e7 d3d4 c5c4 e2f1 a7a6 c2c3 b7b5 a4c2 e8d7 b2b3 e7g6 b3c4 b5c4 b1d2 f8d6 f1g1 h8b8 g2g3 e6g4 g1g2 d6c7 h2h3 g4e6 d2f1 g6e7 f1e3 c7a5 c1d2 b8b2 e1e2 a8b8 g3g4');
INSERT INTO evaluations VALUES(20221,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(20221,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(20221,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(20221,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(20221,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(20221,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(20221,36,'cp -34','e6d5 c4b3 g8f6 e1g1 f8e7 h2h3 b7b5 c2c3 c8f5 f1e1 e8g8 d3d4 c5c4 b3c2 f5c2 d1c2 h7h6 c1f4 b5b4 b1d2 f8e8 f4e5 f6d7 e5g3 a7a5 a2a3 b4b3 c2f5 d7b6 d2f1 e7f8 e1e8 d8e8 f1e3 e8e6 f5h5 g7g6 h5g4 e6g4 h3g4');
INSERT INTO evaluations VALUES(20503,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(20503,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(20503,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(20503,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(20503,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(20503,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(20503,36,'cp -37','c4b3 d8e7 d1e2 e7e2 e1e2 c8e6 h1e1 h7h6 b3a4 g8e7 d3d4 e7f5 d4c5 f8c5 e2f1 e8c8 b1d2 c5b6 d2b3 h8e8 c1d2 g7g5 c2c3 e6d7 a1d1 e8e1 d1e1 f7f6 a4c6 b7c6 f3d4 f5d6 d4e6 d8e8 e6c5 d6c4 c5d7 e8e1 f1e1 c8d7');
INSERT INTO evaluations VALUES(20526,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(20526,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(20526,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(20526,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(20526,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(20526,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(20526,36,'cp -36','g8f6 e1g1 f8e7 h2h3 e8g8 b1c3 d5d4 c3e2 c6a5 c2c3 a5b3 a2b3 b7b6 c3d4 c8b7 e2g3 a7a5 g3f5 c5d4 f5e7 d8e7 f3d4 f8e8 d4f3 a8d8 c1g5 e7d6 d3d4 h7h6 g5h4 d6e6 a1c1 e6f5 f1e1 b7f3 e1e8 d8e8 d1f3 f5f3');
INSERT INTO evaluations VALUES(13760,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(13760,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(13760,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(13760,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(13760,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(13760,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(13760,36,'cp -27','e1g1 f8e7 c2c4 e8g8 c4d5 f6d5 d3d4 c5d4 f3d4 e7f6 d4c6 b7c6 b1d2 f8e8 d2c4 c8e6 c1d2 d5b6 c4b6 a7b6 d1c2 c6c5 d2c3 f6c3 c2c3 e6b3 c3b3 d8f6 a1e1 h7h6 h2h3 e8e1 f1e1 f6c6 a2a3 b6b5 e1d1 a8e8 b3c2 c5c4');
INSERT INTO evaluations VALUES(13759,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(13759,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(13759,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(13759,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(13759,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(13759,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(13759,36,'cp -28','f8e7 c2c4 e8g8 c4d5 f6d5 d3d4 c5d4 f3d4 e7f6 d4c6 b7c6 b1d2 f8e8 d2c4 c8e6 c1d2 d5b6 c4b6 a7b6 d1c2 c6c5 d2c3 f6c3 c2c3 e6b3 c3b3 d8f6 a1e1 g7g6 h2h3 e8e1 f1e1 h7h5 a2a4 a8d8 b3c2 d8d5 b2b3 h5h4 c2e2');
INSERT INTO evaluations VALUES(8120,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(8120,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(8120,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(8120,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(8120,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(8120,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(8120,36,'cp -24','c2c4 e8g8 c4d5 f6d5 d3d4 c5d4 f3d4 e7f6 d4c6 b7c6 b1d2 f8e8 d2c4 c8e6 c1d2 d5b6 c4b6 a7b6 d1c2 c6c5 f1d1 e6b3 c2b3 d8c7 d2c3 c7c6 c3f6 c6f6 d1e1 g7g6 h2h3 e8e1 a1e1 a8d8 b3c2 h7h5 a2a4 f6d4 b2b3 d4f6 c2e2 g8g7 e1d1 d8d4 d1d4 f6d4');
INSERT INTO evaluations VALUES(57643,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(57643,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(57643,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(57643,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(57643,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(57643,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(57643,36,'cp 17','c7c5 c2c3 f8g7 g1f3 e7e6 d2d4 c5d4 e1g1 g8e7 f3d4 e8g8 f1e1 d7d5 e4d5 e7d5 b1d2 b8c6 d2f3 h7h6 c4b3 d8b6 b3d5 e6d5 h2h3 c8d7 d1b3 b6b3 a2b3 c6d4 f3d4 g7d4 c3d4 g6g5 e1e7 d7c6');
INSERT INTO evaluations VALUES(52272,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52272,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(52272,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(52272,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(52272,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(52272,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(52272,36,'cp 42','c2c3 f8g7 g1f3 e7e6 d2d4 c5d4 e1g1 g8e7 f3d4 e8g8 f1e1 d7d5 e4d5 e7d5 b1d2 h7h6 d2e4 b8c6 d4c6 b7c6 c4f1 d8c7 c1d2 f7f5 e4c5 e6e5 b2b4 e5e4 d1b3 a7a5 a2a4 f8d8 h2h3 a5b4 c3b4 g7a1 e1a1 g8h7 a1c1');
INSERT INTO evaluations VALUES(52268,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(52268,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(52268,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(52268,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(52268,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(52268,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52268,36,'cp 31','f8g7 g1f3 e7e6 d2d4 c5d4 e1g1 g8e7 f3d4 e8g8 f1e1 d7d5 e4d5 e7d5 b1d2 b8c6 d4c6 b7c6 d2e4 h7h6 c4f1 d8c7 d1a4 a7a5 c3c4 d5e7 c1d2 e7f5 d2c3 e6e5 a1d1 f5d4 c4c5 c8f5 f1c4 f8d8 c3d4 f5e4 e1e4 e5d4 g2g3 g7f6 c4d3 a8b8 a4c2 g8g7 d1e1');
INSERT INTO evaluations VALUES(37763,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(37763,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(37763,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(37763,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(37763,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(37763,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(37763,36,'cp 25','g1f3 e7e6 d2d4 c5d4 e1g1 g8e7 f3d4 e8g8 f1e1 d7d5 e4d5 e7d5 b1d2 h7h6 d2e4 b8c6 d4c6 b7c6 c4f1 d8c7 c1d2 c6c5 b2b4 c5b4 c3b4 c7e7 a1c1 d5b4 d2e3 f8d8 d1b3 b4d5 e3c5 e7h4 g2g3 h4h5 f1g2 c8a6 c1d1');
INSERT INTO evaluations VALUES(22724,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(22724,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(22724,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(22724,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(22724,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(22724,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(22724,36,'cp -48','f8c5 d2d3 d7d6 c1e3 g8e7 d3d4 c5b4 b1c3 e5d4 f3d4 e8g8 f1e2 f7f5 e4f5 c6d4 d1d4 e7f5 d4d5 g8h8 e3d2 d8e7 d5d3 b4c5 e1g1 f5d4 a1e1 c8f5 d3g3 d4c2 e1d1 e7f7 e2f3 f5d3 f3d5 f7g6 g3g6 h7g6 d5e4 d3f1 d1f1 c2d4 e4g6 c7c6 c3e4 d4e2 g1h1 e2f4 d2f4 f8f4 g2g3');
INSERT INTO evaluations VALUES(6453,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(6453,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(6453,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(6453,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(6453,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(6453,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(6453,36,'cp -39','d2d3 d7d6 c1e3 g8e7 d3d4 c5b4 b1c3 e5d4 f3d4 e8g8 f1e2 f7f5 e4f5 c6d4 d1d4 e7f5 d4d5 g8h8 e3d2 d8e7 d5d3 b4c5 e1g1 f5d4 a1e1 c8f5 d3g3 d4c2 e1d1 e7f7 e2f3 f5d3 f3h5 f7c4 b2b3 c4d4 c3e2 d4e5 d2c3 e5g3 h2g3 d3a6 d1c1 a6e2 h5e2 c2d4 e2d1 b7b5 c3b2 b5b4');
INSERT INTO evaluations VALUES(6451,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(6451,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(6451,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(6451,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(6451,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(6451,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(6451,36,'cp -46','d7d6 c1e3 g8e7 d3d4 c5b4 b1c3 e5d4 f3d4 e8g8 f1e2 f7f5 e4f5 c6d4 d1d4 e7f5 d4d5 g8h8 e3d2 d8e7 d5d3 b4c5 e1g1 f5d4 a1e1 c8f5 d3g3 d4c2 e1c1 e7f7 c3d5 c2d4 f1e1 c7c6 d5f4 d4e2 e1e2 a8e8 e2e8 f8e8 b2b3 h7h6 c1e1 c5d4 h2h3 e8e1 d2e1 f5b1');
INSERT INTO evaluations VALUES(5998,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(5998,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(5998,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(5998,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(5998,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(5998,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(5998,36,'cp -30','c1e3 g8e7 d3d4 e5d4 f3d4 c5b4 b1c3 e8g8 f1e2 f7f5 e4f5 c6d4 d1d4 e7f5 d4d5 f8f7 e1g1 c7c6 d5d3 f5e3 d3e3 d8e7 e3d2 e7h4 a1d1 c8f5 a2a3 b4a5 d2d4 h4d4 d1d4 f7e7 e2f3 f5e6 f1d1 a8e8 h2h3 e6f7 d4e4 a5c3 e4e7 e8e7 b2c3 f7c4');
INSERT INTO evaluations VALUES(57073,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(57073,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(57073,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57073,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(57073,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(57073,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(57073,36,'cp 9','b2b3 g8f6 g1f3 d7d5 c1b2 f8d6 c2c4 e8g8 b1c3 b8d7 f1e2 f8e8 e1g1 b7b6 a2a4 c7c6 b2a3 c6c5 a4a5 b6a5 c4d5 e6d5 d2d4 c8b7 c3a4 f6e4 a4c5 d7c5 d4c5 d6c5 a3c5 e4c5 f3d4 c5e6 d4e6 e8e6 e2f3');
INSERT INTO evaluations VALUES(57058,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(57058,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(57058,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57058,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(57058,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57058,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57058,36,'cp 5','c7c5 c1b2 g8f6 g1f3 f8e7 f1e2 d7d5 d2d4 c5d4 f3d4 e8g8 e1g1 b8c6 d4c6 b7c6 c2c4 c8b7 b1d2 e7d6 a1c1 d8e7 d2f3 f6e4 d1c2 a8d8 e2d3 c6c5 b2e5 d6e5 f3e5 e4f6 f1e1 d5d4 a2a3 a7a5 e3d4 d8d4 d3f1');
INSERT INTO evaluations VALUES(52016,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52016,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(52016,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52016,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52016,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52016,36,'cp 7','g1f3 g8f6 c1b2 f8e7 f1e2 d7d5 d2d4 c5d4 f3d4 e8g8 e1g1 b8c6 d4f3 a8b8 c2c4 d5c4 e2c4 d8d1 f1d1 c8d7 c4e2 f8c8 b1d2 d7e8 a1c1 b7b5 f3e5 a7a6 e5c6 c8c6 c1c6 e8c6 g1f1 b8d8 d2f3 f6e4');
INSERT INTO evaluations VALUES(11710,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(11710,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(11710,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(11710,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(11710,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(11710,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11710,36,'cp 37','a4b3 c8b7 d2d3 f8e7 c2c4 d7d6 c4b5 a6b5 b1c3 b5b4 c3d5 e8g8 c1d2 f6d5 b3d5 d8b8 f1e1 c6d8 d3d4 c7c5 d4c5 d6c5 d1b3 b7d5 e4d5 e7d6 a2a3 b4a3 b3b8 a8b8 b2a3 f7f6 a3a4 d8f7 a4a5 b8a8 e1b1');
INSERT INTO evaluations VALUES(20975,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(20975,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(20975,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(20975,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(20975,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(20975,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(20975,36,'cp 45','e7e6 b1c3 g8f6 a2a3 f8d6 c4c5 d6e7 g1f3 a7a5 f1b5 e8g8 e1g1 f6e4 a1b1 c8d7 d1c2 f7f5 b5d3 a5a4 c3a4 c6d4 e3d4 d7a4 b2b3 a4e8 f3e5 e7f6 f2f3 e4d6 c2c3 d6f7 f3f4 f7e5 f4e5 f6e7 b3b4 c7c6 d3e2 e8g6 b4b5 d8c7');
INSERT INTO evaluations VALUES(21381,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(21381,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(21381,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(21381,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(21381,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(21381,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(21381,36,'cp 51','g1f3 g8f6 b1c3 f8d6 c4c5 d6e7 a2a3 a7a5 f1d3 e8g8 e1g1 b7b6 c5b6 c7b6 d1e2 e7d6 f1d1 c8b7 c1d2 c6a7 d2e1 d8e7 f3d2 b7c6 d1c1 b6b5 c3b5 c6b5 d3b5 a7b5 e2b5 f8b8 b5c6 a8a7 d2f3 b8b2');
INSERT INTO evaluations VALUES(6348,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(6348,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(6348,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(6348,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(6348,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(6348,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(6348,36,'cp -17','d7c6 d2d3 f7f6 e1g1 b7b6 c2c3 c5d6 d3d4 g8e7 b1d2 c8e6 d1e2 e7g6 d2c4 h7h6 c4d6 c7d6 f3h4 g6h4 e2h5 e6f7 h5h4 e8g8 f1e1 d8d7 c1e3 e5d4 e3d4 c6c5 d4e3 f8e8 f2f3 d6d5 a1d1 d7c6 e4d5 f7d5 e3f4 d5e6 h2h3');
INSERT INTO evaluations VALUES(6089,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(6089,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(6089,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(6089,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(6089,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(6089,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(6089,36,'cp -14','d2d3 f7f6 e1g1 b7b6 c2c3 g8e7 d3d4 c5d6 b1d2 e7g6 d4e5 d6e5 d2c4 c8a6 f3e5 f6e5 d1e2 e8g8 c1e3 d8f6 a2a4 a8d8 f1d1 f6f7 b2b3 h7h6 h2h3 c6c5 d1d8 f8d8 a4a5 b6b5 c4d2 g6f4 e3f4');
INSERT INTO evaluations VALUES(6086,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(6086,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(6086,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(6086,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(6086,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(6086,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(6086,36,'cp -14','f7f6 e1g1 a7a5 c1e3 c5e3 f2e3 g8h6 a2a3 h6f7 f3h4 a5a4 b1c3 e8g8 h4f5 g8h8 d1f3 c8e6 a1d1 c6c5 f3f2 c7c6 c3e2 f7d6 h2h3 d8c7 e2g3 g7g6 f5d6 c7d6 g3e2 b7b5 f2h4 d6e7 f1f2 b5b4');
INSERT INTO evaluations VALUES(47856,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(47856,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(47856,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(47856,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(47856,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(47856,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(47856,36,'cp 85','d4e5 f6e4 d1d5 e4c5 c1g5 d8d7 e5d6 f8d6 b1c3 c7c6 d5d2 d6e7 e1c1 d7d2 d1d2 c8e6 f1e2 a7a5 h1e1 b8d7 e2c4 d7f6 d2e2 h7h6 g5h4 e8g8 c4e6 c5e6 f3d4 e6d4 e2e7 g7g5 h4g3 f8e8 f2f3 b7b5 a2a4 b5b4 e7e8 f6e8 c3b1 d4e6 b1d2 e8f6 g3f2 f6d5 g2g3');
INSERT INTO evaluations VALUES(47776,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(47776,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(47776,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(47776,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(47776,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(47776,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(47776,36,'cp 79','f6e4 d1d5 e4c5 c1g5 d8d7 e5d6 f8d6 b1c3 c7c6 d5d2 d6e7 e1c1 d7d2 d1d2 c8e6 f1e2 b8a6 g5e3 a6c7 h1d1 c7d5 c3d5 e6d5 f3d4 d5e4 e2g4 h7h5 g4h3 a8d8 d4f5 d8d2 d1d2 e4f5 h3f5 g7g6 f5c8 b7b6 e3c5 e7c5 c8d7 e8e7 d7c6 h8c8 c6e4 e7f8');
INSERT INTO evaluations VALUES(48000,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(48000,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(48000,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(48000,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(48000,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(48000,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(48000,36,'cp 73','d1d5 e4c5 c1g5 d8d7 e5d6 f8d6 b1c3 c7c6 d5d2 d6e7 e1c1 d7d2 d1d2 c8e6 f1e2 b8a6 g5e3 a6c7 h1d1 c5d7 a2a3 d7f6 f3d4 e8g8 d4e6 c7e6 e2f3 h7h6 g2g3 f8e8 h2h3 g8f8 c3a4 a8c8 b2b4 a7a5 a4b6 c8c7 c2c3 a5b4 a3b4');
INSERT INTO evaluations VALUES(52425,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(52425,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(52425,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52425,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52425,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(52425,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(52425,36,'cp 1','g7g6 f1g2 f8g7 e2e3 e7e6 b1c3 b8c6 g1e2 g8e7 e1g1 e8g8 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 c3d5 e6d5 c1e3 c8g4 f1e1 f8e8 h2h3 g4f5 a1c1 a8c8 e2f4 c6d4 c1c8 d8c8 e3d4 e8e1 d1e1 g7d4 f4d5 c8c5 d5e7 g8g7 e7f5 c5f5 g2b7');
INSERT INTO evaluations VALUES(52287,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(52287,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52287,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52287,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52287,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(52287,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(52287,36,'cp 12','f1g2 b8c6 g1f3 f8g7 b1c3 d7d6 e1g1 c8g4 d2d3 d8d7 a1b1 g4h3 g2h3 d7h3 c3d5 h3d7 c1d2 g8f6 b2b4 f6d5 c4d5 c6b4 d2b4 c5b4 b1b4 b7b6 d1a4 d7a4 b4a4 e8d7 f1c1 h8c8 c1c8 d7c8 h2h4 g7b2 a4c4 c8b7');
INSERT INTO evaluations VALUES(52288,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(52288,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(52288,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(52288,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(52288,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(52288,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(52288,36,'cp 0','b8c6 b1c3 f8g7 e2e3 e7e6 g1e2 g8e7 e1g1 e8g8 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 f1e1 f8e8 h2h4 h7h6 c3d5 e6d5 c1e3 c8g4 a1c1 d8a5 d1d3 a8d8 a2a3 g4f5 d3d1 f5g4');
INSERT INTO evaluations VALUES(19263,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(19263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19263,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19263,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(19263,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(19263,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(19263,36,'cp 5','e2e3 f8g7 b1c3 e7e6 g1e2 g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 e1g1 e8g8 h2h4 d5c3 b2c3 e6e5 c1g5 d8c7 f1e1 h7h6 g5e3 c8g4 d1a4 g4d7 a4a3 f8e8 a1d1 a8c8 d4e5 c6e5 e2f4 c7c3 a3c3 c8c3 g2b7 c3c2');
INSERT INTO evaluations VALUES(19271,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(19271,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(19271,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(19271,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(19271,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(19271,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(19271,36,'cp 33','d2d4 c5d4 f3d4 f8g7 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8d8 e1e7 a7a5 d4c3 d8d5 a2a4 b5d7 e7e5 d5e5 c3e5 h6g7 e5g7 g8g7 b3c5 d7f5 a1c1 a8d8 h2h3 d8d2 b2b3');
INSERT INTO evaluations VALUES(19266,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(19266,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19266,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(19266,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(19266,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(19266,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(19266,36,'cp 37','c5d4 f3d4 f8g7 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8d8 e1e7 a7a6 d4c3 d8d5 e7e5 d5e5 c3e5 a8c8 h2h3 h6f8 e5c3 b5c4 c3f6 c8a8 a1c1 f8g7 f6g7 c4b3 g7f6 b3a2');
INSERT INTO evaluations VALUES(19323,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(19323,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(19323,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(19323,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(19323,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(19323,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(19323,36,'cp 19','f3d4 f8g7 c1e3 g8f6 f1c4 e8g8 c4b3 d7d6 f2f3 c8d7 d1e2 c6a5 e2d2 a8b8 h2h4 h7h5 e3h6 b7b5 h6g7 g8g7 c3d5 a5b3 a2b3 e7e5 d5f6 d8f6 d4e2 f8d8 e2c3 d7e6 a1a7 b8c8 a7a5 b5b4 c3d5 e6d5');
INSERT INTO evaluations VALUES(44877,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(44877,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(44877,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(44877,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(44877,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(44877,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(44877,36,'cp -13','d5c4 e2e3 b7b5 a2a4 d8b6 f1e2 c8b7 g1f3 e7e6 e1g1 b8d7 a4a5 b6c7 b2b3 b5b4 g5f4 f8d6 a5a6 b7c8 c3a2 d6f4 e3f4 c4b3 d1b3 a8b8 f4f5 e6f5 a2b4 c7d6 b3e3 f6e4 b4d3 d7f6 d3e5 f6d5 e3a3 d6a3 a1a3 d5b4 f1b1 c8e6 e5d3 b4a6 b1b8 a6b8 a3a7');
INSERT INTO evaluations VALUES(44999,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(44999,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(44999,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(44999,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(44999,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(44999,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(44999,36,'cp -25','e2e3 b7b5 a2a4 d8b6 f1e2 c8b7 g1f3 e7e6 f3e5 h7h6 g5h4 b8d7 h4f6 d7f6 e2h5 g7g6 d1f3 f8e7 c3e4 e8c8 e4f6 c6c5 f3f4 c5d4 e1g1 e7f6 f4f6 g6h5 a4b5 h8g8 g2g3 d4e3 e5c4 b6d4 f6d4 d8d4 c4e3 c8b8 f1d1 d4b4 d1d7 b4b2 d7f7 h5h4 g1f1 h4g3');
INSERT INTO evaluations VALUES(44997,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(44997,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(44997,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(44997,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(44997,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(44997,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(44997,36,'cp -35','b7b5 a2a4 d8b6 f1e2 c8b7 g1f3 e7e6 e1g1 b8d7 a4b5 c6b5 b2b3 f8b4 c3a2 a7a5 g5f6 d7f6 b3c4 b5c4 a2b4 a5b4 a1a8 b7a8 f3e5 e8g8 e2c4 f8b8 d1a4 a8d5 c4d5 f6d5 e5d7 b6b5 f1a1 b5a4 a1a4 b8c8 d7c5 h7h6');
INSERT INTO evaluations VALUES(43933,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(43933,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(43933,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(43933,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43933,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(43933,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(43933,36,'cp -26','a2a4 d8b6 f1e2 c8b7 g1f3 e7e6 e1g1 b8d7 a4a5 b6c7 b2b3 b5b4 a5a6 b7c8 g5f4 f8d6 c3a2 d6f4 e3f4 c4b3 d1b3 a8b8 f4f5 e6f5 a2b4 c7d6 b3e3 f6e4 b4d3 e8g8 f1c1 f8e8 e2f1 d6c7 d4d5 c6c5 f3d2 e4f6 e3f3 b8b6 d2c4 b6a6 d5d6 c7d8 f3f5 d7b6 f5c5 b6c4 c5c4 a6d6');
INSERT INTO evaluations VALUES(21704,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(21704,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(21704,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(21704,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(21704,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(21704,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(21704,36,'cp 101','f3e5 d5e4 f1b5 c8d7 b5c6 d7c6 b1c3 f8b4 e1g1 g8f6 c1g5 b4c3 b2c3 d8d6 e5c6 d6c6 g5f6 c6f6 d1g4 e8g8 g4e4 b7b6 a1e1 h7h6 h2h3 a8d8 e1e3 f6d6 e4d3 f8e8 c3c4 d6c6 f1d1 e8e3 f2e3 c6d6 e3e4 c7c5 d4d5');
INSERT INTO evaluations VALUES(21633,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(21633,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(21633,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(21633,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(21633,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(21633,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(21633,36,'cp 117','d5e4 f1b5 c8d7 b5c6 d7c6 e1g1 f8d6 e5c6 b7c6 d1e2 g8f6 b1d2 e8g8 d2e4 f8e8 e4f6 d8f6 e2d3 h7h5 h2h3 c6c5 c1e3 c5d4 e3d4 d6e5 a1d1 a8d8 d4e5 f6e5 d3d8 e8d8 d1d8 g8h7 c2c3 a7a5 d8d7 f7f6 f1d1 a5a4 a2a3 e5e2 d7d2 e2e6 h3h4 e6c4 g2g3 c4e6 d2d8 e6a2 d1d2 a2b1 g1h2 b1f5 d8d4 f5b5 c3c4');
INSERT INTO evaluations VALUES(21772,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(21772,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(21772,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(21772,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(21772,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(21772,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(21772,36,'cp 127','f1b5 c8d7 b5c6 d7c6 e1g1 f8d6 e5c6 b7c6 d1e2 g8f6 b1d2 e8g8 d2e4 f8e8 e4f6 d8f6 e2d3 h7h5 c1e3 e8e6 a1e1 a8e8 h2h3 c6c5 c2c3 f6h4 e3d2 c5d4 e1e6 e8e6 c3d4 h4e4 d3e4 e6e4 d2c3 f7f6 f1e1 e4e1 c3e1 g8f7 g1f1 f7e6 f1e2 e6d5 e2d3 c7c6 e1a5 d6e7 a5d2 f6f5 d2f4 e7f6 f4e3 g7g6');
INSERT INTO evaluations VALUES(21540,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(21540,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(21540,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(21540,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(21540,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(21540,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(21540,36,'cp 112','c8d7 b5c6 d7c6 e1g1 f8d6 e5c6 b7c6 d1e2 g8f6 b1c3 e8g8 c3e4 f8e8 e4f6 d8f6 e2d3 h7h5 c1e3 e8e6 h2h3 a8e8 a1e1 c6c5 c2c3 f6h4 e3d2 c5d4 e1e6 e8e6 c3d4 h4e4 d3e4 e6e4 d2c3 f7f6 f2f3 e4e6 f1c1 g8f7 g1f1 d6f4 c1e1 e6a6 a2a3 f4g3 e1e4 h5h4 d4d5 a6d6');
INSERT INTO evaluations VALUES(23199,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(23199,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(23199,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(23199,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(23199,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(23199,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(23199,36,'cp 121','d2d4 e5d4 f3d4 f8b4 c2c3 c6d4 b5c4 b4a5 d1d4 g8e7 d4d1 a5b6 a2a4 c7c6 e1g1 d7d5 e4d5 c6d5 c4b3 e8g8 f1e1 g8h8 b1a3 c8f5 c1e3 b6e3 e1e3 f5e4 a3b5 a7a6 b5d4 f8e8 d1d2 e7c6 a1e1 c6e5 h2h4 e4g6 d2d1 g6f7 b3c2 d8c7 b2b4 a8d8 h4h5 h7h6');
INSERT INTO evaluations VALUES(23194,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(23194,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(23194,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(23194,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(23194,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(23194,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(23194,36,'cp 107','e5d4 f3d4 f8b4 b1c3 g8e7 e1g1 e8g8 c3d5 e7d5 e4d5 c6d4 d1d4 b4e7 c1d2 a7a6 b5c4 d7d6 d2c3 f8f7 c4e2 c7c5 d5c6 b7c6 f1e1 d6d5 e2h5 g7g6 h5f3 e7d6 h2h4 d6c7 h4h5 c8f5 d4c5 d8d6 c5d6 c7d6 a1d1 d6c5');
INSERT INTO evaluations VALUES(23202,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(23202,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(23202,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(23202,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(23202,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(23202,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(23202,36,'cp 105','f3d4 f8b4 c2c3 c6d4 d1d4 b4a5 d4d1 g8e7 b5c4 a5b6 a2a4 c7c6 e1g1 d7d5 e4d5 c6d5 c4b3 e8g8 f1e1 g8h8 a4a5 b6c7 b1d2 f8e8 d1h5 g7g6 h5f3 d8d6 d2f1 c8f5 c1e3 f5e4 f3g3 d6d8 e3f4 c7a5 f1d2 e7f5 g3h3 a5b6 d2e4 e8e4 e1e4 d5e4');
INSERT INTO evaluations VALUES(23201,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(23201,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(23201,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(23201,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(23201,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(23201,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(23201,36,'cp 106','f8b4 b1c3 g8e7 d1h5 e7g6 d4f5 b4c3 b2c3 e8g8 f2f4 d7d6 e1g1 d8e8 h5f3 c8d7 b5d3 e8f7 f5e3 a8e8 c1b2 c6b8 c3c4 b8a6 b2a3 g8h8 f3f2 d7c6 e3d5 b7b6 a1e1 a6c5 a3c5 b6c5 f2g3 f7d7 d5c3 c6b7 a2a4 f6f5 e4e5 g6e7');
INSERT INTO evaluations VALUES(27215,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(27215,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(27215,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(27215,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(27215,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(27215,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27215,36,'cp 54','d2d4 e5d4 f3d4 g7g6 b1c3 f8g7 d4c6 b7c6 b5d3 g8f6 h2h3 a8b8 a1b1 f6g4 d1d2 g4e5 d3a6 c6c5 b2b3 e8g8 c1b2 e5c6 c3e2 g7b2 b1b2 d8e7 e2c3 e7e5 b2b1 e5d4 f1d1 d4d2 d1d2 c6b4 a6f1 g8g7 f2f3 f7f5');
INSERT INTO evaluations VALUES(27200,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(27200,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(27200,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(27200,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(27200,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(27200,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(27200,36,'cp 63','e5d4 b5c6 b7c6 d1d4 c6c5 d4d3 g8e7 b1c3 e7g6 h2h4 g6h4 f3h4 d8h4 c3d5 h4d8 f1e1 f8e7 c1f4 e8g8 e4e5 e7g5 f4g5 d8g5 e5d6 c7d6 d5e7 g8h8 d3d6 d7h3 d6d5 g5d5 e7d5 a8d8 g2h3 d8d5 a1d1 d5d4 c2c3 d4d8 d1d8 f8d8 e1e7 h7h6 e7f7 d8d2 f7b7 a7a5 g1g2 a5a4 g2f3 d2d3 f3g4');
INSERT INTO evaluations VALUES(27229,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(27229,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(27229,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(27229,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(27229,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(27229,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27229,36,'cp 61','b5c6 b7c6 d1d4 g8e7 b1c3 e7g6 e4e5 c6c5 d4e4 f8e7 e5d6 c7d6 h2h4 e8g8 h4h5 g6h4 c3d5 h4f3 e4f3 h7h6 b2b3 f8e8 c1f4 e7f6 a1d1 d7c6 d5f6 d8f6 f3c6 f6f4 d1d6 f4f5 c2c4 f5h5 c6b7 h5f5 d6d5 f5e4 f1d1 a7a5 b7c6 e4c2 d5d2 c2f5 c6d5 f5d5 c4d5 a5a4 b3a4 a8a4');
INSERT INTO evaluations VALUES(27166,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(27166,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(27166,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(27166,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(27166,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(27166,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(27166,36,'cp 44','b7c6 d1d4 g8e7 b1c3 e7g6 e4e5 c6c5 d4e4 f8e7 e5d6 c7d6 h2h4 h7h5 c3d5 e8g8 c1g5 e7g5 f3g5 d8e8 e4d3 e8e5 f1e1 e5f5 a1d1 f5d3 d1d3 a8e8 e1e8 f8e8 d5e3 g6h4 d3d6 d7g4 e3g4 h5g4 g1f1 h4f5 d6d5 g7g6 d5d7 f7f6 g5f7 a7a5 f7d6 f5d6 d7d6 g8f7 d6a6');
INSERT INTO evaluations VALUES(49628,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(49628,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(49628,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(49628,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(49628,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49628,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(49628,36,'cp -25','g1f3 f6e4 d2d3 e4c5 f3e5 d8e7 d3d4 b8c6 c1e3 c6e5 d4c5 e7e6 f1e2 b7b6 c5b6 a7b6 e1g1 c8b7 a2a4 f8e7 e3d4 e8g8 d4e5 e6e5 e2f3 d7d5 f1e1 e5f6 b1a3 g7g6 a3c2 c7c6 g2g3 h7h5 h2h4 g8g7 f3g2 e7d6 d1d2 a8a6 g2f1 a6a7');
INSERT INTO evaluations VALUES(49625,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(49625,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(49625,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(49625,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(49625,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(49625,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(49625,36,'cp -25','f6e4 d2d3 e4c5 f3e5 d8e7 d3d4 b8c6 c1e3 c6e5 d4c5 e7e6 f1e2 b7b6 c5b6 a7b6 e1g1 c8b7 a2a4 f8e7 e3d4 e8g8 d4e5 e6e5 e2f3 d7d5 f1e1 e5f6 b1a3 a8e8 g2g3 c7c6 a3c2 g7g6 c2d4 h7h5 h2h4 e7d6 f3g2 e8e1 d1e1 f8a8 e1d2 g8g7');
INSERT INTO evaluations VALUES(49906,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(49906,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(49906,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(49906,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(49906,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(49906,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(49906,36,'cp -26','d2d3 e4c5 f3e5 d8e7 d3d4 b8c6 c1e3 c6e5 d4c5 e7e6 f1e2 b7b6 c5b6 a7b6 e1g1 c8b7 a2a4 f8e7 e3d4 e8g8 d4e5 e6e5 e2f3 d7d5 f1e1 e5f6 b1a3 c7c6 a3c2 g7g6 g2g3 h7h5 h2h4 e7d6 f3g2 g8g7 c2d4 a8a7 b2b4 f8a8 f2f4 d6e7 g1h1 b7c8 b4b5 c8g4 d1c2 e7c5 b5c6 c5d4 c3d4');
INSERT INTO evaluations VALUES(21824,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(21824,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(21824,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(21824,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(21824,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(21824,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(21824,36,'cp 72','d1d8 e8d8 b1c3 f8b4 c1d2 f7f6 e1c1 c8d7 c3a4 b4d6 d2e3 g8e7 a2a3 h7h5 a4c3 a7a6 f3d2 b7b5 d2b3 d8e8 b3c5 d7c8 h2h4 c6d4 b2b4 d6c5 b4c5 d4e6 a3a4 b5a4 c3a4 a8b8 f2f3 e7c6 f1c4 e8e7 a4c3');
INSERT INTO evaluations VALUES(2263,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(2263,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(2263,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(2263,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(2263,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(2263,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(2263,36,'cp 51','e8d8 b1c3 f8d6 c1e3 g8f6 e1c1 a7a6 f3d2 b7b5 d2b3 d8e7 f1e2 h8d8 c3d5 f6d5 e4d5 c6b8 g2g4 b8d7 h1g1 e7f8 h2h4 d8e8 b3a5 f7f5 g4f5 d7f6 e2d3 e5e4 d3e2 e8e5 c2c4 e5f5 g1g5 f5g5 h4g5 f6g4 e2g4 c8g4 c4c5 g4d1 c5d6 d1f3 d6c7 a8c8');
INSERT INTO evaluations VALUES(2319,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(2319,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(2319,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(2319,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(2319,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(2319,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(2319,36,'cp 69','b1c3 f8d6 c1g5 d8e8 f3d2 c8e6 e1c1 g8f6 f2f3 a7a6 f1c4 b7b5 c4e6 f7e6 d2b3 f6d7 c1b1 d7b6 h2h4 b5b4 c3e2 a6a5 g5e3 b6c4 e3f2 a5a4 b3d2 c6a5 e2c1 c4d2 d1d2 a5c4 d2e2 e8f7 c1d3 h8d8 h1c1 h7h5 b2b3 c4a3 b1b2 a3b5');
INSERT INTO evaluations VALUES(2318,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(2318,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(2318,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(2318,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(2318,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(2318,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(2318,36,'cp 64','f8d6 c1e3 g8f6 h2h3 a7a6 e1c1 d8e8 g2g3 c8e6 a2a3 b7b5 f3d2 f6d7 f2f4 f7f6 a3a4 c6d4 a4b5 a6b5 c3b5 d4b5 f1b5 e5f4 e3f4 e8e7 f4d6 c7d6 b5d7 e6d7 d2b1 a8a1 c1d2 h8b8 b2b3 b8a8 b1c3 a1d1 h1d1 d7h3 d1h1 h3e6 h1h7 e7f7');
INSERT INTO evaluations VALUES(55821,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(55821,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55821,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55821,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(55821,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(55821,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55821,36,'cp -8','g8f6 c3b5 f8b4 c2c3 b4a5 e2e3 a7a6 b5a3 b8d7 g1f3 e8g8 h2h3 c7c6 f1e2 f8e8 e1g1 a5c7 f4c7 d8c7 c3c4 e6e5 d4e5 d7e5 c4d5 f6d5 a3c2 c8f5 c2d4 f5e4 d1b3 a8d8 a1c1 e5f3 e2f3 d5f6 f1d1 g7g6 f3e4 f6e4 b3c2');
INSERT INTO evaluations VALUES(41242,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(41242,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(41242,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(41242,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(41242,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(41242,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41242,36,'cp 28','e8g8 c4b3 c8g4 h2h3 g4f3 d1f3 b8c6 c1e3 e7e5 d4e5 d6e5 c3d5 f6d5 b3d5 c6d4 e3d4 e5d4 e1g1 c7c6 d5c4 d4d3 a1d1 g7b2 d1d3 d8f6 a2a4 a8d8 f1b1 b7b6 d3d8 f8d8 c4f7 g8f7 b1b2 f6f3 g2f3 d8d2 a4a5 f7e6 a5b6 a7b6 b2b6 d2c2');
INSERT INTO evaluations VALUES(36507,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36507,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(36507,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36507,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(36507,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36507,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36507,36,'cp 39','c4b3 c8g4 c1e3 b8c6 h2h3 g4f3 d1f3 e7e5 d4e5 d6e5 c3d5 f6d5 b3d5 c6d4 e3d4 e5d4 e1g1 a8b8 g2g3 c7c6 d5b3 d8c7 h3h4 b7b5 h4h5 a7a5 a2a4 d4d3 f3d3 g7b2 a1b1 f8d8 d3f3 b5a4 b3a4 c7e5 h5h6 e5c3 f3f4 c3f6 f4f6 b2f6 f2f4');
INSERT INTO evaluations VALUES(46173,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46173,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46173,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(46173,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(46173,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(46173,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(46173,36,'cp 19','b8c6 a2a3 d5c4 f1c4 a7a6 c4a2 b7b5 e1g1 c8b7 d1e2 f8d6 d4c5 d6c5 b2b4 c5d6 c1b2 e8g8 f3d2 a8c8 a1d1 d8c7 f2f4 c6e7 a2b1 e7f5 c3e4 f6e4 d2e4 b7e4 b1e4 d6e7 g2g4 f5d6 d1c1 c7b8 e4d3 c8c1 f1c1 f8c8 g4g5 c8c1 b2c1 d6f5 c1b2 b8b6');
INSERT INTO evaluations VALUES(13551,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(13551,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(13551,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13551,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(13551,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(13551,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(13551,36,'cp 13','a2a3 d5c4 f1c4 a7a6 e1g1 b7b5 c4a2 c8b7 d1e2 c5d4 f1d1 f8d6 e3d4 e8g8 c1g5 h7h6 g5f6 d8f6 c3e4 f6e7 e4d6 e7d6 d4d5 e6d5 d1d5 d6e7 e2c2 a8d8 a1d1 d8d5 a2d5 c6a5 d5b7 e7b7 b2b4 a5c4 a3a4 b7c7 a4b5 a6b5');
INSERT INTO evaluations VALUES(13559,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(13559,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(13559,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13559,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(13559,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13559,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13559,36,'cp 0','d5c4 f1c4 a7a6 e1g1 b7b5 c4a2 f8e7 d1c2 c5d4 e3d4 c8b7 f1d1 c6a5 f3e5 e8g8 h2h3 a8c8 c2e2 b7d5 c3d5 f6d5 c1d2 a5c6 a1c1 d8d6 e2e4 d5f6 e4e1 f6d5');
INSERT INTO evaluations VALUES(21027,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21027,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(21027,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21027,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(21027,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21027,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21027,36,'cp 36','c8f5 f1g2 c6b4 b1a3 e7e6 e1g1 h7h6 c1f4 f8d6 f4d6 c7d6 f3h4 f5h7 c2c4 g8f6 d1a4 b4c6 c4d5 e6d5 a3b5 e8g8 b5c3 g7g5 h4f3 d8b6 a4b5 b6b5 c3b5 a7a6 b5c3 f8c8 f1d1 b7b5 a1c1 g8g7 h2h4');
INSERT INTO evaluations VALUES(26570,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(26570,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(26570,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(26570,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(26570,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(26570,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(26570,36,'cp 40','f1g2 c6b4 b1a3 e7e6 e1g1 b4c6 a3b1 c6b4 f3e1 f8e7 c2c3 b4c6 e1d3 g8f6 f1e1 e8g8 b1d2 f8e8 e2e4 f5e4 d2e4 f6e4 g2e4 d5e4 e1e4 d8d5 d1e2 e7d6 c1f4 c6e7 h2h4 a8d8 h4h5 h7h6 a2a3 e7f5 a1e1 c7c5 d3c5 d6c5 d4c5 d5c5');
INSERT INTO evaluations VALUES(26571,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26571,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(26571,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(26571,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(26571,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26571,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26571,36,'cp 26','c6b4 b1a3 e7e6 e1g1 b4c6 a3b1 c6b4 f3e1 f8e7 c2c3 b4c6 b1d2 g8f6 e1d3 a7a5 f1e1 h7h5 d1b3 d8d7 d2f3 a5a4 b3c2 f6g4 h2h3 g4f6 c1g5 f6e4 g5e7 d7e7 h3h4 f7f6 c3c4 e8g8 c4d5 e6d5 a1c1 g7g5 h4g5');
INSERT INTO evaluations VALUES(26706,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(26706,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(26706,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(26706,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(26706,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(26706,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(26706,36,'cp 49','b1a3 e7e6 f3h4 f5g4 c2c3 b4c6 d1b3 a8b8 e2e4 d5e4 g2e4 g4h3 a3c4 f8e7 e4f3 g8f6 h4g2 h3g4 f3g4 f6g4 a2a4 a7a6 a4a5 d8d5 e1g1 d5b5 b3a4 e8g8 g2f4 g4f6 f4d3 e7d6 a4b5 a6b5 c4e3 f8d8 c1d2 h7h6');
INSERT INTO evaluations VALUES(26707,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(26707,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(26707,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(26707,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(26707,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(26707,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(26707,36,'cp 32','e7e6 f3h4 f5g4 c2c3 b4c6 d1b3 a8b8 e2e4 d5e4 g2e4 g4h3 a3c4 f8e7 e4f3 g8f6 b3c2 d8d7 h4g2 e8g8 e1g1 h3g4 f3g4 f6g4 a2a4 a7a5 c2e2 g4f6 c1g5 f6d5 g5e7 d5e7 a1d1 b8d8 c4e3 b7b6 h2h4 e7d5 e3d5 d7d5 g2f4');
INSERT INTO evaluations VALUES(20196,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(20196,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(20196,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(20196,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(20196,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(20196,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(20196,36,'cp -31','c5d4 f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 b1d2 g8e7 c2c3 e7f5 d2f3 f8c5 b2b4 c5e7 a2a4 c6c5 c1d2 a7a6 d1e2 e8g8 e1g1 h7h6 f1b1 d8d7 a4a5 c5b4 c3b4 d7a7 b4b5 a6b5 b1b5 f8c8 g2g3 e7c5 g1g2 f5d4 f3d4');
INSERT INTO evaluations VALUES(20168,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(20168,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(20168,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(20168,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(20168,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(20168,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(20168,36,'cp -27','f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 b1d2 g8e7 c2c3 e7f5 d2f3 f8c5 b2b4 c5b6 a2a4 a7a5 e1g1 e8g8 c1d2 d8c7 g2g3 f8b8 g1g2 h7h6 h2h4 b8d8 d1e2 d5d4 f1c1 d4c3 d2c3 d8d5 b4a5 b6a5 c3a5 d5a5 c1c4 c6c5 e2e4 f5e7 a1c1 e7d5 e4c2 c7b7 c4c5 a5c5');
INSERT INTO evaluations VALUES(20162,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(20162,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(20162,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(20162,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(20162,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(20162,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(20162,36,'cp -43','c8d7 d4c6 d7c6 b5c6 b7c6 b1d2 g8h6 d2f3 h6f5 c2c3 f8c5 b2b4 c5b6 d1d3 a7a5 a2a4 e8g8 e1g1 h7h6 c1d2 d8e7 d3e2 a8a7 f1b1 f8a8 e2d1 e7c7 g2g3 a7b7 d1e2 b6a7 b1c1 a5b4 c3b4 c6c5 b4c5 a7c5 g1g2 b7b2 a4a5 a8b8 a5a6');
INSERT INTO evaluations VALUES(25914,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(25914,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(25914,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(25914,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(25914,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(25914,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(25914,36,'cp -41','d4c6 d7c6 b5c6 b7c6 b1d2 g8h6 d2f3 h6f5 c2c3 f8c5 b2b4 c5b6 a2a4 a7a5 e1g1 e8g8 d1d3 h7h6 c1d2 d8c7 g2g3 f8d8 f1c1 c6c5 b4b5 c5c4 d3e2 c7e7 a1b1 a8c8 g3g4 f5h4 d2e3 b6e3 e2e3 h4f3 e3f3 e7c7 b5b6 c7e5');
INSERT INTO evaluations VALUES(25907,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(25907,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(25907,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(25907,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(25907,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(25907,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(25907,36,'cp -37','d7c6 b5c6 b7c6 b1d2 g8h6 d2f3 h6f5 c2c3 f8c5 b2b4 c5b6 a2a4 a7a5 e1g1 e8g8 d1d3 h7h6 c1d2 d8e7 g2g3 f8b8 d3e2 a5b4 c3b4 f5d4 f3d4 b6d4 a1a2 c6c5 b4c5 e7c5 a4a5 b8b2 a2b2 d4b2 d2h6 b2c3 h6d2 d5d4 a5a6 c5a5 d2c3');
INSERT INTO evaluations VALUES(22895,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(22895,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(22895,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(22895,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(22895,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(22895,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(22895,36,'cp -135','g7g5 d2d4 g5g4 e1g1 g4f3 d1f3 d8e7 f3f2 d7d6 c1f4 f8g7 c2c3 c8d7 b1a3 g8h6 f2g3 h8g8 f4h6 g7d4 c3d4 g8g3 c4f7 e8d8 h2g3 c6d4 f7h5 d7e8 f1f8 e7f8 h6f8 e8h5 f8g7 d4e6 g7c3 d8d7 a3c4 e6c5 e4e5 h5f7 c4d2 d6d5 a1f1 f7g6 f1f6 c5e6 d2b3 c7c5 f6e6 d7e6');
INSERT INTO evaluations VALUES(22489,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(22489,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(22489,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(22489,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(22489,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(22489,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(22489,36,'cp -110','d2d4 g5g4 e1g1 g4f3 d1f3 d8e7 f3f2 f8g7 c2c3 d7d6 c1f4 c8d7 b1d2 g7h6 f4g3 f7f6 a2a4 h6d2 f2d2 e8c8 b2b4 h7h5 a4a5 h5h4 g3f4 h4h3 b4b5 e7e4 f1e1 e4g2 d2g2 h3g2 b5c6 d7c6 c4e6 c6d7 e6d5 g8h6 a5a6 b7b6 d5b7 c8b8 e1e7 h6f5 e7f7 f5h4 f4g3 d7c8 b7d5 d8g8 g3h4 h8h4 f7f6');
INSERT INTO evaluations VALUES(22490,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(22490,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(22490,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(22490,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(22490,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(22490,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(22490,36,'cp -101','g5g4 e1g1 g4f3 d1f3 d8e7 f3f2 d7d6 c1f4 f8g7 c2c3 c8d7 b1d2 g7h6 f4g3 f7f6 b2b4 h6d2 f2d2 e8c8 c4d3 h7h5 g3h4 e7g7 a2a4 c6e7 b4b5 g7h6 d2h6 h8h6 a4a5 d8e8 c3c4 e7g6 h4e1 g6f8 e1d2 h6g6 e4e5 g6g7 e5f6 g7f7 a5a6 f7f6 a6b7 c8b7 c4c5 f6f1 a1f1 d6c5 d4c5 b7b8');
INSERT INTO evaluations VALUES(21119,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(21119,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(21119,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(21119,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(21119,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(21119,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(21119,36,'cp 86','g1f3 g8f6 e2e4 c8g4 d4d5 c6e5 c1e3 c7c6 d5c6 d8d1 a1d1 e5c6 f1c4 e7e6 d1c1 g4f3 g2f3 f6d7 c4b5 f8c5 b5c6 b7c6 e1e2 c5e3 f2e3 e8e7 c3a4 a8c8 h1g1 h8g8 f3f4 g7g6 b2b3 f7f6 a4c5 g6g5 c5d7 e7d7 e4e5 g5f4 e5f6 g8g1 c1g1 f4e3 e2e3');
INSERT INTO evaluations VALUES(21118,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(21118,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(21118,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(21118,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(21118,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(21118,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21118,36,'cp 84','g8f6 e2e4 c8g4 d4d5 c6e5 c1e3 c7c6 d5c6 e5c6 f1c4 d8d1 a1d1 e7e6 d1c1 g4f3 g2f3 f6d7 c4e2 f8e7 e1g1 e8g8 f1d1 f8d8 a2a3 d7c5 c3b5 d8d1 c1d1 e6e5 e2c4 c5e6 c4e6 f7e6 b5c7 a8d8 d1d8 e7d8 c7e6');
INSERT INTO evaluations VALUES(14380,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(14380,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(14380,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(14380,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(14380,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(14380,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(14380,36,'cp 61','e2e3 e7e5 d4d5 e5e4 d5c6 d8d1 e1d1 e4f3 g2f3 b7c6 f1c4 f6d7 h2h4 d7b6 c4d3 a7a5 h4h5 f8e7 d1c2 c8e6 e3e4 e8c8 c1e3 g7g5 d3a6 c8b8 a1d1 d8d1 h1d1 h7h6 a6d3 b6d7 f3f4 g5f4 e3f4 e7c5 f4g3 h8g8 d3a6 b8a7 a6e2 a7b7 b2b3');
INSERT INTO evaluations VALUES(14379,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14379,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(14379,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14379,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(14379,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(14379,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14379,36,'cp 51','e7e5 d4d5 e5e4 d5c6 d8d1 e1d1 e4f3 g2f3 b7c6 f1c4 f6d7 e3e4 d7e5 c4e2 f7f5 h1g1 e5g6 c1e3 f8d6 h2h4 c8e6 h4h5 g6f4 g1g7 d6e5 g7g1 e8c8 d1c2 f4e2 c3e2 f5e4 f3e4 e6c4 e2c3 c4d3 c2b3 h8e8 a1d1 e5f6 e3g5 f6g5 g1g5 d3e4 d1d8 c8d8 c3e4');
INSERT INTO evaluations VALUES(19452,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(19452,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(19452,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(19452,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19452,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(19452,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(19452,36,'cp -22','g7g6 b1c3 d7d6 g2g3 f8g7 f1g2 g8f6 e1g1 e8g8 a2a4 a8b8 c3d5 f6d5 e4d5 c6b4 f3e1 e7e6 d5e6 c8e6 c2c3 b4c6 e1f3 h7h6 c1e3 c6e7 d3d4 e7d5 e3d2 b7b6 f1e1 b8c8 c3c4 d5b4 d1b3 f8e8 d4d5 e6f5 d2b4 c5b4');
INSERT INTO evaluations VALUES(19281,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19281,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(19281,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(19281,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(19281,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(19281,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(19281,36,'cp -5','c2c3 g8f6 e4e5 f6d5 d1b3 e7e6 d3d4 f8g7 c1g5 d8b6 b3b6 a7b6 f1c4 c5d4 c4d5 e6d5 g5f6 g7f6 e5f6 d7d6 c3d4 c8g4 b1c3 g4f3 g2f3 c6d4 e1c1 d4f3 d1d5 e8c8 c3b5 c8b8 h2h3 d8c8 c1b1 c8c5 h1d1');
INSERT INTO evaluations VALUES(19278,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(19278,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(19278,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(19278,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(19278,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(19278,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(19278,36,'cp -10','g8f6 f1e2 f8g7 e1g1 e8g8 f1e1 e7e5 c3c4 d7d6 b1c3 f6e8 a2a3 e8c7 c3d5 b7b6 b2b4 c7e6 a1b1 f7f5 e4f5 g6f5 e2f1 c6e7 g2g3 e7d5 c4d5 e6c7 b4c5 b6c5 f1g2 f5f4 f3d2 g8h8 c1b2 c8f5 d2c4 d8f6 g3g4 f5g6');
INSERT INTO evaluations VALUES(9329,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9329,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(9329,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(9329,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(9329,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(9329,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(9329,36,'cp 7','c3a4 c5b6 d2d3 c6a5 c4b3 c7c6 a4b6 a7b6 c1d2 e8g8 h2h3 h7h6 f1e1 c8d7 d2c3 a5b3 a2b3 a8a1 d1a1 f8e8 a1d1 c6c5 f3h2 d7c6 h2f1 b6b5 b3b4 c5c4 f1g3 d6d5 e4d5 d8d5 f2f3 c4d3 c2d3 f6h7 d3d4 e5d4 e1e8 c6e8 c3d4');
INSERT INTO evaluations VALUES(9419,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(9419,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9419,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(9419,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(9419,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(9419,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9419,36,'cp 0','c5b6 d2d3 c6a5 c4d5 c7c6 a4b6 a7b6 d5f7 e8f7 b2b4 h8f8 f1e1 f7g8 h2h3 c6c5 b4a5 a8a5 c2c3 d8c7 c1g5 a5a3 d1c1 a3a6 c1d1');
INSERT INTO evaluations VALUES(42041,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(42041,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(42041,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(42041,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(42041,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(42041,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(42041,36,'cp 19','d1a4 b8c6 c1d2 e8g8 a2a3 b4c3 d2c3 c6e7 g1f3 b7b6 a4c2 a7a5 c4d5 e6d5 f1d3 c8a6 d3a6 a8a6 e1g1 f6e4 f1c1 d8a8 f3d2 f8c8 a1b1 e4d2 c2d2 h7h6 h2h3 c7c5 d4c5 b6c5 b2b4 d5d4 e3d4 a5b4 a3b4 c5c4');
INSERT INTO evaluations VALUES(42118,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(42118,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(42118,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(42118,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(42118,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42118,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(42118,36,'cp 25','b8c6 c1d2 e8g8 a2a3 b4c3 d2c3 c6e7 g1f3 a7a5 f1d3 d5c4 d3c4 e7d5 a4c2 d5c3 c2c3 b7b6 c4b5 c8b7 b5c6 f6d5 c3c2 b7c6 c2c6 d5e7 c6a4 a8c8 e1g1 c7c5 f1d1 d8d6 h2h3 e7d5 a1c1 d6c7 d4c5 b6c5');
INSERT INTO evaluations VALUES(9018,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(9018,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(9018,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(9018,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(9018,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(9018,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(9018,36,'cp 25','c1d2 e8g8 a2a3 b4c3 d2c3 f6e4 a4c2 b7b6 c4d5 e6d5 f1d3 c8f5 g1e2 c6e7 e1g1 a8c8 c3e1 e4d6 e1b4 c7c6 b4d6 f5d3 d6e7 d3c2 e7d8 f8d8 f1c1 c2a4 f2f3 a4b5 g1f2 c6c5 h2h4 b5e2 f2e2 c5d4 e3d4 f7f6 h4h5 g8f7 e2f2 c8c1 a1c1');
INSERT INTO evaluations VALUES(9017,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(9017,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(9017,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(9017,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(9017,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(9017,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(9017,36,'cp 15','e8g8 a2a3 b4c3 d2c3 c6e7 g1f3 b7b6 f1e2 a7a5 c4d5 e6d5 e1g1 c8d7 a4b3 a5a4 b3c2 d7f5 c2c1 e7g6 e2b5 f5d7 b5d3 f6e4 c1c2 d7f5 f1e1 f8e8 a1c1 c7c5 f3d2 c5c4 d3e4 f5e4 d2e4 e8e4 f2f3');
INSERT INTO evaluations VALUES(39550,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(39550,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(39550,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(39550,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(39550,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(39550,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(39550,36,'cp 27','c7c5 c2c3 e7e6 d2d4 c5d4 e1g1 g8e7 c4b3 d7d5 e4d5 e7d5 c1g5 d8d7 c3d4 h7h6 g5d2 e8g8 b1c3 d5e7 d2f4 b8c6 f1e1 d7d8 h2h3 c6d4 f3d4 d8d4 d1d4 g7d4 a1d1 e7f5 f4h6 d4g7 h6g5 f5d4 g5e7 f8e8');
INSERT INTO evaluations VALUES(11537,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11537,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11537,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11537,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11537,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11537,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11537,36,'cp 0','f8c5 e1g1 e8g8 a4c6 d7c6 a2a4 f8e8 d2d3 c5d6 h2h3 c8e6 c3e2 h7h6 e2g3 c6c5 b2b3 b7b5 c1b2 f6d7 d1e2 d8f6 f1e1 e8b8 e1f1');
INSERT INTO evaluations VALUES(7326,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7326,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7326,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7326,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7326,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7326,36,'cp 0','e1g1 e8g8 a4c6 d7c6 a2a4 c5d6 h2h3 h7h6 d2d3 c8e6 c3e2 c6c5 e2g3 f8e8 b2b3 b7b5 c1b2 f6d7 d1e2 d8f6 f3d2 f6g5 d2f3');
INSERT INTO evaluations VALUES(7325,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7325,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7325,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7325,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7325,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7325,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7325,36,'cp 0','e8g8 a4c6 d7c6 d2d3 c5d6 h2h3 f8e8 a2a4 c8e6 c3e2 h7h6 e2g3 c6c5 b2b3 b7b5 c1b2 f6d7 d1e2 d8f6 f1e1 e8b8 b2c3 b8e8');
INSERT INTO evaluations VALUES(2678,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2678,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2678,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2678,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2678,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2678,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2678,36,'cp 0','a4c6 d7c6 d2d3 c5d6 h2h3 f8e8 a2a4 c8e6 c3e2 h7h6 e2g3 c6c5 b2b3 b7b5 c1b2 f6d7 d1e2 d8f6 f1e1 a8b8 b2c3 b8a8');
INSERT INTO evaluations VALUES(51837,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(51837,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(51837,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(51837,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(51837,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51837,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51837,36,'cp 38','b1c3 f8c5 d1g4 g7g6 d2d3 d7d6 g4e2 g8f6 a2a4 d8e7 g1f3 c8e6 c4e6 e7e6 a4a5 e8g8 c1d2 b8a6 e1g1 f8e8 c3a4 c5b4 f3g5 e6g4 f2f3 g4h5 g5h3 b4d2 e2d2 d6d5 h3f2 g8g7 a4c3 a6c7 c3e2 c7e6 b2b3 h5h4 g2g3');
INSERT INTO evaluations VALUES(51832,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(51832,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(51832,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(51832,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(51832,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(51832,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(51832,36,'cp 37','f8c5 d1g4 g7g6 d2d3 d7d6 g4e2 g8f6 a2a4 d8e7 a4a5 c8e6 c4e6 e7e6 c1d2 a7a6 c3a4 c5a7 g1f3 e8g8 d2h6 f8e8 e1g1 b8d7 b2b4 f6h5 f1e1 d7f8 e2d2 e6f6 h6g5 f6g7 a4b6 a7b6 a5b6 f8e6 g5h6 g7f6 h6e3 e6f4');
INSERT INTO evaluations VALUES(37654,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(37654,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(37654,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(37654,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(37654,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(37654,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(37654,36,'cp 44','d1g4 g7g6 d2d3 d7d6 g4e2 g8f6 a2a4 d8e7 g1f3 c8e6 c4e6 e7e6 a4a5 e8g8 c1d2 b8a6 e1g1 f8e8 c3a4 e6d7 a4c5 a6c5 f3g5 f6h5 b2b4 c5e6 g5e6 e8e6 f2f3 d6d5 e2f2 a7a6 a1e1 a8e8 g1h1 h5g7 d3d4 e5d4 f2d4 f7f6');
INSERT INTO evaluations VALUES(49452,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(49452,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(49452,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(49452,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(49452,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(49452,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(49452,36,'cp 141','e4e5 f6d5 c2c4 d5e7 d2d4 d7d6 e5d6 c7d6 b1c3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 d4e5 d6e5 e1c1 d8d2 f4d2 f5e7 g4g5 h6h5 b5c7 c8g4 f1e2 b8c6 c7a8 f8a8 h1e1 e5e4 f3h2 g4e6 h2f1 c6d4 f1g3 d4e2 e1e2 e6g4 e2e4 g4d1 c1d1 e7c6 b2b3 a8d8 d1c1 f7f5 g5f6 g7f6 d2e3 d8c8');
INSERT INTO evaluations VALUES(49352,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(49352,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(49352,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(49352,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(49352,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(49352,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(49352,36,'cp 90','f6d5 c2c4 d5e7 d2d4 d7d6 e5d6 c7d6 b1c3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 d4e5 d6e5 d2d8 f8d8 f3e5 b8c6 e5c6 b7c6 g4f5 d8e8 e1d2 c6b5 f1g2 c8f5 g2a8 e8a8 c4b5 g7b2 a1d1 a8c8 f2f3 c8c5 a2a4 c5c4 f4h6 c4a4 d2e3 a4b4 h4h5 b2f6 h5g6 f7g6 d1d5 f5e6 d5d6 g8f7 d6a6 b4b5 a6a7 f7g8 h1g1');
INSERT INTO evaluations VALUES(49328,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(49328,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(49328,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(49328,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(49328,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(49328,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(49328,36,'cp 91','c2c4 d5e7 d2d4 d7d6 e5d6 c7d6 h2h4 h7h6 b1c3 g7g6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 d4e5 d6e5 e1c1 d8d2 f4d2 f5e7 g4g5 h6h5 b5c7 c8g4 f1e2 b8c6 c7a8 f8a8 h1e1 e5e4 f3h2 g4e6 h2f1 e7f5 f1g3 f5g3 f2g3 c6d4 e2f1 e6g4 e1e4 g4d1 c1d1 d4f5 f1g2 g7b2 e4e1 a8c8 g2d5');
INSERT INTO evaluations VALUES(49322,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(49322,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(49322,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(49322,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(49322,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(49322,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(49322,36,'cp 98','d5e7 d2d4 d7d6 e5d6 c7d6 h2h4 h7h6 b1c3 g7g6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 g4f5 e5f4 f1d3 d6d5 c4c5 b7b6 f5g6 b6c5 g6f7 f8f7 h1g1 a7a6 b5c3 c5d4 f3d4 b8c6 d4c6 d8e8 d2e2 e8c6 g1g6 c6c5 e1c1 g8f8 e2h5 g7c3 b2c3 c5c3 d3c2 c8f5 h5h6 f8e8 g6c6 c3a1 c1d2');
INSERT INTO evaluations VALUES(53608,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(53608,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(53608,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(53608,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(53608,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(53608,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(53608,36,'cp 65','f1d3 d5e4 c3e4 d8d4 g1f3 d4d8 d1e2 g8f6 e4f6 g7f6 c1f4 d8e7 e1g1 e6e5 f4g3 c8e6 a2a3 b8d7 b2b4 d7b6 f3h4 e7d7 f2f4 e5f4 f1f4 f8d6 h4f5 d6f4 g3f4 b6d5 f4g3 h8g8 e2f2 g8g3 f5g3 a7a5 b4a5 a8a5 a1e1 a5a3 g3h5 a3a2 h5f6 d5f6 f2f6');
INSERT INTO evaluations VALUES(53609,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(53609,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(53609,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(53609,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(53609,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(53609,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(53609,36,'cp 62','d5e4 c3e4 d8d4 g1f3 d4d8 d1e2 g8f6 e4f6 g7f6 c1f4 d8e7 e1g1 e6e5 f4g3 c8g4 h2h3 g4h5 d3f5 h5g6 f3h4 b8d7 f5d7 e7d7 f2f4 d7d4 g1h2 d4e4 e2f2 e4c2 f4f5 c2f2 g3f2 e8c8 f5g6 h7g6 a1d1 f8e7 h4f3 e7d6 g2g3 e5e4 f3d2 f6f5 d2c4 d6c7 f2c5');
INSERT INTO evaluations VALUES(19216,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(19216,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(19216,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(19216,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(19216,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(19216,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19216,36,'cp 50','e1g1 f8g7 b5c6 b7c6 f1e1 g8h6 c2c3 e8g8 d2d4 c5d4 c3d4 d7d5 e4e5 f7f6 e5f6 e7f6 h2h3 h6f5 b1c3 f8e8 c1d2 e8e1 d1e1 d8e7 e1c1 c8d7 c3a4 g7f8 c1c2 f5d6 a4c5 e7f7 a2a3 d6e4 d2b4 a7a5 c2a4 e4c5 b4c5 f8c5 d4c5');
INSERT INTO evaluations VALUES(19215,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(19215,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(19215,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(19215,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(19215,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(19215,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(19215,36,'cp 38','f8g7 b5c6 b7c6 f1e1 g8h6 c2c3 e8g8 d2d4 c5d4 c3d4 d7d5 e4e5 f7f6 e5f6 e7f6 b1d2 f8e8 d2b3 h6f5 c1d2 f5d6 d1c2 c8d7 e1e8 d8e8 a1e1 e8f7 h2h3 g7f8 b3c5 a8b8 d2f4 b8e8 e1e8 d7e8 h3h4 d6c4 b2b3 c4d6 c2c1 e8d7 f4h6 f8h6 c1h6');
INSERT INTO evaluations VALUES(5273,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(5273,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(5273,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(5273,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(5273,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(5273,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(5273,36,'cp 47','f1e1 g8f6 b5c6 d7c6 d2d3 e8g8 h2h3 f6e8 e4e5 e8c7 a2a4 c8e6 b1c3 c5c4 d3d4 h7h6 c3e4 e6d5 f3h4 a7a5 c2c3 b7b5 e4c5 c7e6 c5e6 d5e6 f2f4 d8d7 c1e3 b5b4 d1e2 f8b8 f4f5 e6f5 h4f5 g6f5 e2c4 b4c3 b2c3');
INSERT INTO evaluations VALUES(5274,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(5274,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(5274,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(5274,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(5274,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(5274,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5274,36,'cp 28','g8f6 b5c6 d7c6 d2d3 e8g8 h2h3 f6e8 e4e5 e8c7 c1e3 c7e6 d1d2 b7b6 e3h6 e6d4 f3d4 c5d4 a2a4 c6c5 d2f4 c8b7 b1d2 b7d5 d2e4 d8d7 a4a5 a8e8 a5b6 a7b6 f4g5 d7e6 e4g3 e8a8 h6g7 g8g7 a1a8 f8a8 g3h5 g7g8 h5f4');
INSERT INTO evaluations VALUES(49624,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(49624,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(49624,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(49624,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(49624,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(49624,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(49624,36,'cp -28','f8b4 c1d2 d7d5 e4d5 f6d5 g2g3 b8c6 f1g2 d5c3 b2c3 b4a5 g1f3 e8g8 e1g1 c8g4 h2h3 g4e6 f3g5 e6d5 g2d5 d8d5 c3c4 d5d8 d1h5 h7h6 d2a5 c6a5 g5f3 f7f6 f3h4 a5c6 h4f5 d8d7 h5g4 g8h8 a1b1 b7b6 b1b5 a8e8 a2a4 d7f7 b5d5');
INSERT INTO evaluations VALUES(42742,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(42742,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(42742,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(42742,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(42742,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(42742,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(42742,36,'cp -9','c1d2 b8c6 c3d5 b4d2 d1d2 d7d6 g2g3 f6d5 e4d5 c6d4 c2c3 d4f5 g1e2 e8g8 f1g2 c8d7 c3c4 c7c5 b2b4 b7b6 e1g1 h7h6 a1b1 a8b8 h2h3 d8g5 d2g5 h6g5 a2a3 b8c8 f1e1 f5d4 b4c5 d4e2 e1e2 c8c5 g2f1');
INSERT INTO evaluations VALUES(42743,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(42743,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(42743,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(42743,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(42743,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(42743,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(42743,36,'cp -33','d7d5 e4d5 e8g8 f1e2 b4c3 d2c3 f8e8 g1f3 d8d6 e1g1 f6d5 f1e1 b8c6 c3d2 c8f5 c2c3 a7a5 a2a3 d5f6 d2g5 h7h6 g5f6 d6f6 d1b3 b7b6 a1d1 e8d8 b3a4 d8d6 d3d4 e5d4 f3d4 c6d4 d1d4 d6d4 a4d4 f6d4 c3d4 f5d7 e2f3');
INSERT INTO evaluations VALUES(10063,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(10063,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(10063,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(10063,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(10063,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(10063,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(10063,36,'cp -67','e8g8 d2d3 c6a5 c4b3 h7h6 g5f3 a5b3 a2b3 d7d5 e1g1 f8e8 h2h3 c5f8 c1d2 d5e4 c3e4 b7b5 f1e1 c8b7 d2c3 f6d7 b3b4 f7f5 e4g3 d8f6 d3d4 e5e4 d4d5 f6f7 f3h4 g7g6 f2f3 e4e3 d1d4 f8g7 d4g7 f7g7 c3g7 g8g7 g3e2 d7f6 e2f4 f6d5 f4g6 d5b4');
INSERT INTO evaluations VALUES(4296,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(4296,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(4296,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(4296,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(4296,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(4296,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(4296,36,'cp -57','d2d3 c6a5 c4b3 a5b3 a2b3 h7h6 g5f3 d7d5 e1g1 f8e8 h2h3 c5f8 c1d2 d5e4 c3e4 b7b5 f1e1 c8b7 b3b4 f6e4 d3e4 a7a6 d1e2 c7c5 b4c5 f8c5 d2a5 d8e7 a1d1 a8c8 c2c3 e7f6 d1d7 b7c6 d7d1 f6f4 e2d3 f7f6 b2b4 c5a7 c3c4 b5c4 d3c4 g8h8');
INSERT INTO evaluations VALUES(4297,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(4297,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(4297,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(4297,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(4297,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(4297,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(4297,36,'cp -52','c6a5 c4b3 a5b3 a2b3 h7h6 g5f3 d7d5 e1g1 f8e8 h2h3 c5f8 c1d2 b7b6 e4d5 f6d5 f1e1 d5c3 d2c3 f7f6 f3h4 c8e6 h4g6 f8d6 d3d4 e5d4 d1d4 a7a5 g6f4 e6f7 a1d1 d8d7 d4d3 e8e1 c3e1 d7e8 d3f3 a5a4 e1c3 a4b3 c2b3 d6f4 f3f4');
INSERT INTO evaluations VALUES(4539,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(4539,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(4539,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(4539,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(4539,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(4539,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(4539,36,'cp -62','c4b3 h7h6 g5f3 a5b3 a2b3 d7d5 e1g1 f8e8 h2h3 c5f8 c1d2 b7b6 e4d5 f6d5 f1e1 d5c3 b2c3 f8d6 c3c4 a7a5 d2c3 f7f6 d3d4 e5e4 c4c5 d6f4 f3d2 d8d5 d1e2 e4e3 e2c4 c8b7 c4d5 b7d5 d2f1 e3e2 f1g3 f4g3 f2g3 e8e4 c3d2 a8e8 g1f2 g8f7 g3g4 e4d4');
INSERT INTO evaluations VALUES(23056,30,'cp 489',NULL);
INSERT INTO evaluations VALUES(23056,31,'cp 485',NULL);
INSERT INTO evaluations VALUES(23056,32,'cp 501',NULL);
INSERT INTO evaluations VALUES(23056,33,'cp 501',NULL);
INSERT INTO evaluations VALUES(23056,34,'cp 500',NULL);
INSERT INTO evaluations VALUES(23056,35,'cp 502',NULL);
INSERT INTO evaluations VALUES(23056,36,'cp 490','c2c3 d4c6 d5c6 b7c6 g1f3 d7d6 f1c4 g8f6 e1g1 h7h6 c4d3 g7g6 c3c4 f8g7 b1c3 a7a5 g1h1 e8g8 f3g1 c8e6 d1c2 f6h5 c1e3 h5f4 f2f3 f4d3 c2d3 f7f5 e3f2 a5a4 b2b4 f5e4 d3e4 e6f5 e4c6 e5e4 c3e4 g7a1');
INSERT INTO evaluations VALUES(23053,30,'cp 491',NULL);
INSERT INTO evaluations VALUES(23053,31,'cp 496',NULL);
INSERT INTO evaluations VALUES(23053,32,'cp 483',NULL);
INSERT INTO evaluations VALUES(23053,33,'cp 480',NULL);
INSERT INTO evaluations VALUES(23053,34,'cp 498',NULL);
INSERT INTO evaluations VALUES(23053,35,'cp 494',NULL);
INSERT INTO evaluations VALUES(23053,36,'cp 493','d4c6 d5c6 b7c6 g1f3 d7d6 f1c4 g8f6 e1g1 f8e7 f1e1 a7a5 b1d2 a5a4 c4d3 e8g8 d3c2 a4a3 b2b4 f8e8 c2a4 c8d7 c1a3 c6c5 a4d7 d8d7 a3b2 c5b4 c3b4 c7c5 b2c3 c5b4 c3b4 d7b7 b4c3');
INSERT INTO evaluations VALUES(19376,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(19376,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(19376,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(19376,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(19376,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(19376,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(19376,36,'cp -98','e7e6 c4b3 b7b5 c2c3 c5c4 b3c2 g8f6 f3e2 d8a5 g1f3 b5b4 b2b3 c4b3 c2b3 b4c3 b1c3 c8a6 b3c4 a6c4 e2c4 f8e7 e1g1 e8g8 d2d4 f8c8 c1d2 d7d5 c4d3 d5e4 c3e4 a5d5 e4c3 d5d7 a1c1 c8d8 d2e3 c6b4 d3d2 b4d5 c3d5 d7d5');
INSERT INTO evaluations VALUES(18748,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(18748,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(18748,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(18748,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(18748,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(18748,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(18748,36,'cp -73','g1e2 c6e5 f3c3 e5c4 c3c4 g8f6 e4e5 b7b5 c4b3 f6d5 b1c3 d5c3 b3c3 c8b7 c3g3 d7d6 d2d4 c5d4 e2d4 d6e5 g3e5 a7a6 e1g1 d8d5 e5d5 b7d5 f1d1 a8c8 c2c3 f7f6 a2a4 b5b4 c3b4 f8b4 d4f5 e8f7 f5e3 h8d8 e3d5 d8d5 c1e3 c8c2 d1d5');
INSERT INTO evaluations VALUES(18749,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(18749,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(18749,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(18749,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(18749,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(18749,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(18749,36,'cp -97','c6e5 f3c3 e5c4 c3c4 g8f6 d2d3 b7b6 e1g1 f8d6 c4c3 e8g8 c3e1 f6h5 e2g3 h5g3 h2g3 f7f5 a2a4 c8b7 b1c3 a7a6 c1e3 d8e8 a1d1 e8e7 e1d2 d6c7 e3f4 d7d6 d2e3 b7c6 f1e1 a8e8 e3d2 e7d7');
INSERT INTO evaluations VALUES(19129,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(19129,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(19129,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(19129,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(19129,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(19129,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(19129,36,'cp -96','f3c3 e5c4 c3c4 g8f6 e1g1 b7b5 c4b5 f6e4 d2d3 a7a6 b5b3 e4f6 b1d2 f8e7 d2c4 e8g8 b3b6 a6a5 b6d8 f8d8 c1d2 a5a4 c4b6 a8a6 b6c8 d8c8 e2c3 c8a8 c3b5 f6e8 c2c4 f7f6 f1c1 g7g5 a1b1 g5g4 c1d1 g8f7 d3d4 c5d4 b5d4');
INSERT INTO evaluations VALUES(22555,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(22555,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(22555,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(22555,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(22555,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(22555,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(22555,36,'cp 139','g8e7 c1d2 d7d5 e4d5 e7f5 d4d3 f5d6 b1c3 f8e7 c4b5 d6b5 d3b5 c8d7 b5d3 e8g8 e1c1 a7a6 h1e1 f8e8 d3g3 e7h4 e1e8 d7e8 g3f3 d8f6 f3f6 h4f6 c3e4 f6e7 d2f4 a8c8 d1e1 e8d7 c2c4 g7g5 f4e5 f7f5 e4c3');
INSERT INTO evaluations VALUES(53829,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(53829,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(53829,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(53829,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(53829,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(53829,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(53829,36,'cp 70','d2d4 b8d7 g2g4 h7h6 h1g1 g8f6 c1e3 f8e7 a2a3 d8a5 f3d2 d7f8 d1e2 a5c7 h2h4 e5d4 e3d4 f8e6 d4e3 d6d5 e4d5 f6d5 c3d5 c6d5 d2f3 c8d7 e1c1 e8c8 c1b1 e7c5 g4g5 h6g5 f3g5 d5d4 e3d2 e6g5 g1g5 g7g6 h4h5 h8e8 e2f3');
INSERT INTO evaluations VALUES(53825,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(53825,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(53825,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(53825,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(53825,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(53825,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(53825,36,'cp 51','b8d7 f1c4 f8e7 a2a4 g8f6 h2h3 h7h6 e1g1 e8g8 f1e1 f8e8 a4a5 e5d4 f3d4 e7f8 c1f4 d7c5 d1f3 a7a6 c4f1 d6d5 e4e5 f6e4 c3e4 c5e4 c2c3 c6c5 d4e2 e4g5 f3e3 c8d7 f4h2 g5e4 e2g3');
INSERT INTO evaluations VALUES(18540,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(18540,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(18540,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(18540,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(18540,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(18540,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(18540,36,'cp 64','f1c4 f8e7 a2a4 g8f6 e1g1 h7h6 h2h3 e8g8 f1e1 e5d4 f3d4 f8e8 c1f4 e7f8 a4a5 d7e5 c4b3 e5d7 d1f3 d7c5 b3a2 c5e6 a2e6 c8e6 a5a6 b7a6 a1a6 e6c8 a6a1 d8b6 a1d1 c8d7 b2b3 c6c5 d4e2 e8e6');
INSERT INTO evaluations VALUES(53577,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(53577,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(53577,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(53577,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(53577,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(53577,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(53577,36,'cp 33','e2e3 g8f6 b2b3 f8d6 f1d3 b7b6 e1g1 e8g8 b1d2 c8b7 c1b2 c6c5 d4c5 b6c5 d1c2 b8c6 a2a3 h7h6 a1d1 d8e7 c4d5 e6d5 f3h4 d6e5 h4f5 e7e6 d2f3 e5b2 c2b2 f8d8 f5g3 f6d7 b3b4 c5c4 d3c2 c6e5 f3e5 d7e5');
INSERT INTO evaluations VALUES(53575,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(53575,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53575,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53575,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(53575,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(53575,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(53575,36,'cp 30','g8f6 b2b3 b7b6 c1b2 f8d6 b1d2 c8b7 f1d3 c6c5 c4d5 e6d5 e1g1 e8g8 f1e1 f8e8 a1c1 b8d7 d1e2 d6f8 f3e5 d8e7 e1d1 a8c8 e5d7 f6d7 h2h3 g7g6 d4c5 d7c5 d2f3 e8d8 c1c2 f8g7 f3d4 c5d3 c2c8 d8c8 e2d3');
INSERT INTO evaluations VALUES(45646,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(45646,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45646,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(45646,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45646,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(45646,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45646,36,'cp 29','f1d3 d5c4 d3c4 a7a6 e1g1 c6c5 d4c5 f8c5 d1d8 e8d8 c4e2 d8e7 a2a3 b7b5 b1d2 c8b7 d2b3 c5d6 b3a5 b7d5 f3d4 d6c5 d4b3 c5d6 c1d2 b8c6 a5c6 d5c6 f2f3 c6d5 b3a5 h8c8 a1c1 d6e5 e3e4 c8c1 f1c1 e5b2');
INSERT INTO evaluations VALUES(45643,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(45643,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45643,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45643,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(45643,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(45643,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(45643,36,'cp 28','d5c4 d3c4 c6c5 e1g1 a7a6 c4e2 c5d4 f3d4 f8d6 b2b3 e6e5 d4f3 b8c6 b1d2 e8g8 d2c4 d6c7 c1b2 e5e4 f3d4 c6d4 d1d4 c8e6 d4c3 b7b5 f1d1 d8b8 c4e5 c7e5 c3e5 b8e5 b2e5 f8c8 a1c1 c8c1 d1c1 f6d5 e5d4 d5b4 a2a3 b4d3 c1c7');
INSERT INTO evaluations VALUES(57017,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(57017,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(57017,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(57017,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(57017,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(57017,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(57017,36,'cp 22','g8f6 e2e4 d7d5 e4e5 f6d7 f2f4 c7c5 g1f3 c5d4 c3b5 d8a5 c1d2 a5b6 b5d4 f8e7 a1b1 b8c6 c2c3 d7c5 d2e3 c8d7 f1e2 e8g8 e1g1 c6d4 e3d4 b6c7 b2b4 c5e4 b1c1 a7a5 e2d3 f8c8 b4b5 e4c5 a2a4 c5d3 d1d3');
INSERT INTO evaluations VALUES(54960,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(54960,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(54960,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(54960,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(54960,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(54960,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(54960,36,'cp 49','b1c3 e7e5 g2g3 f7f5 g1f3 f8e7 f1g2 g8f6 b2b4 e8g8 d2d3 a7a6 e1g1 d8e8 a2a4 b8c6 a1b1 a6a5 b4b5 c6b4 c1a3 c7c5 b5c6 b7c6 a3b4 a5b4 b1b4 e7d8 b4b1 d8a5 d1c2 f5f4 h2h3 f4g3 f2g3 c8d7 e2e3 h7h6 f3h4 d7e6 h4f5 g8h8');
INSERT INTO evaluations VALUES(54956,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(54956,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(54956,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(54956,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(54956,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(54956,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(54956,36,'cp 40','e7e5 g2g3 f8e7 g1f3 f7f5 f1g2 g8f6 e1g1 e8g8 d2d4 e5e4 f3g5 h7h6 g5h3 c7c6 d4d5 b8d7 b2b3 a7a5 f2f3 a5a4 c3a4 c6d5 h3f4 d5c4 f3e4 f5e4 a4c3 d7e5 b3c4 g8h8 a2a4 c8f5 c1e3 e5c4 e3d4');
INSERT INTO evaluations VALUES(55559,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(55559,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(55559,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(55559,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55559,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(55559,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(55559,36,'cp 43','g1f3 f7f5 g2g3 f8e7 f1g2 g8f6 b2b4 e8g8 d2d3 a7a6 e1g1 g8h8 a2a4 b8c6 d1b3 d8e8 c1b2 c6d8 c3d5 f6d5 c4d5 c8d7 f3d2 d8f7 e2e3 a8b8 b2c3 e7f6 f1c1 b8c8 d2c4 c8a8 b3c2 e8e7 c4a5');
INSERT INTO evaluations VALUES(55557,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55557,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(55557,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55557,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55557,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55557,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(55557,36,'cp 28','f7f5 g2g3 f8e7 f1g2 g8f6 e1g1 e8g8 b2b4 e5e4 f3e1 d6d5 c3d5 f6d5 c4d5 d8d5 c1b2 b8c6 d2d3 c8e6 d1b1 c6d4 b2d4 d5d4 d3e4 f5e4 e1c2 d4b6 a2a4 e7f6 a1a3 a7a5 g2e4 c7c6 b4a5 b6b1 f1b1 f6e7');
INSERT INTO evaluations VALUES(56740,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(56740,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(56740,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(56740,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(56740,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(56740,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(56740,36,'cp 37','g2g3 g8f6 d2d3 b8c6 f1g2 f8e7 e1g1 e8g8 e2e3 d8e8 f3e1 a7a5 b2b3 c6b4 c1b2 c7c6 f2f4 b4a6 e1f3 e5e4 d3e4 f6e4 c3a4 e8d8 a1c1 e7f6 b2f6 e4f6 d1d4 d8e7 d4d2 a6c5 a4c5 d6c5 f3e5 f8d8 d2b2');
INSERT INTO evaluations VALUES(51921,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(51921,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(51921,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(51921,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(51921,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(51921,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(51921,36,'cp -23','d7d5 e4d5 e6d5 c4b3 f8e7 g1e2 g8f6 b1c3 e8g8 e1g1 c8e6 e2f4 b8c6 f4e6 f7e6 c3e2 e7d6 c2c3 d8c7 e2g3 b7b5 c1g5 a8e8 d3d4 g8h8 d4c5 d6c5 d1e2 c5b6 g5f6 g7f6 g1h1 f6f5 a1e1 b5b4 g3h5 b4c3 b2c3');
INSERT INTO evaluations VALUES(54154,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(54154,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(54154,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(54154,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(54154,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(54154,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(54154,36,'cp -13','e4d5 e6d5 c4b3 g8f6 b1c3 f8e7 g1e2 e8g8 e1g1 b8c6 c1g5 f6h5 g5e7 c6e7 d1d2 b7b6 a1e1 c8b7 e2f4 h5f6 d2e2 e7g6 e2f3 g6f4 f3f4 f8e8 h2h3 d8d7 c3e2 a7a5 a2a4 h7h6 e2g3 e8e1 f1e1');
INSERT INTO evaluations VALUES(54119,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(54119,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(54119,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(54119,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(54119,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(54119,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(54119,36,'cp -24','e6d5 c4b3 f8e7 g1e2 g8f6 b1c3 e8g8 e1g1 b8c6 c1g5 f6h5 g5e7 c6e7 d1d2 h5f6 e2f4 b7b6 a1e1 c8b7 d2e3 e7g6 e3f3 g6f4 f3f4 f8e8 a2a3 d8d7 b3a4 b7c6 a4c6 d7c6 d3d4 h7h6 f4d2 a8d8 e1e8 d8e8 f1d1 e8d8 f2f3');
INSERT INTO evaluations VALUES(54499,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(54499,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(54499,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(54499,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54499,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(54499,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(54499,36,'cp -16','c4b3 b8c6 b1c3 g8f6 c1g5 f8e7 g1e2 e8g8 e1g1 f6h5 g5e7 c6e7 e2g3 h5f6 g3h5 f6h5 d1h5 c8e6 c3e2 a7a5 a2a4 d8d6 f1e1 a8e8 e2c3 d6d7 h5e5 e7g6 e5g3 e8e7 a1d1 f8d8 d3d4 c5c4 b3a2 h7h6 a2b1 d8e8');
INSERT INTO evaluations VALUES(54537,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(54537,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(54537,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(54537,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(54537,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(54537,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(54537,36,'cp -23','g8f6 b1c3 f8e7 g1e2 e8g8 e1g1 b8c6 c1g5 f6h5 g5e7 c6e7 d1d2 h5f6 e2f4 b7b6 a1e1 c8b7 d2e2 e7g6 e2f3 g6f4 f3f4 f8e8 c3e2 d8b8 b3a4 e8e6 b2b4 h7h6 c2c3 b7c6 f4b8 a8b8 a4c6 e6c6 e1b1 c6c8 b4c5 b6c5 h2h3 b8b6 e2g3 g8f8');
INSERT INTO evaluations VALUES(46445,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(46445,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(46445,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(46445,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(46445,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(46445,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(46445,36,'cp -19','b1c3 f8e7 g1e2 e8g8 e1g1 b8c6 c1g5 f6h5 g5e7 c6e7 e2g3 h5f6 g3h5 f6h5 d1h5 c8e6 c3e2 c5c4 d3c4 d5c4 b3a4 e7g6 a1d1 d8f6 h5b5 b7b6 c2c3 a8d8 a4c2 g6f4 e2f4 f6f4 b5c6 g7g6 c6e4 f4e4 c2e4 g8g7 e4c6 d8c8 c6b7 c8c7 b7d5 e6d5 d1d5');
INSERT INTO evaluations VALUES(46448,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(46448,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(46448,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(46448,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(46448,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(46448,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(46448,36,'cp -25','b8c6 c1g5 f8e7 g1e2 e8g8 e1g1 f6h5 g5e7 c6e7 d1d2 h5f6 e2f4 b7b6 f1e1 c8b7 d2e2 e7g6 e2f3 g6f4 f3f4 f8e8 c3e2 d8b8 f4b8 a8b8 f2f3 d5d4 a2a4 g8f8 b3c4 b7c6 b2b3 h7h6 e2g3 f6d5 e1e8 b8e8');
INSERT INTO evaluations VALUES(19481,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(19481,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(19481,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(19481,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(19481,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(19481,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(19481,36,'cp -20','g7g6 g1f3 f8g7 a2a4 a7a6 f1e2 e7e6 e4e5 d7d5 d2d4 c5d4 f3d4 c8d7 c1e3 f7f6 e5f6 g8f6 e1g1 e8g8 d4c6 d7c6 e3d4 d8c7 a4a5 a8d8 d4e5 c7e7 d1d4 f6e4 e2g4 f8e8 a1d1 g7e5 f4e5 e7g5');
INSERT INTO evaluations VALUES(19294,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(19294,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(19294,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(19294,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(19294,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19294,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(19294,36,'cp -6','a2a4 f8g7 g1f3 a7a6 d2d3 d7d6 f1e2 e7e6 e1g1 g8e7 d1e1 a8b8 c3d1 b7b5 a4b5 a6b5 c2c3 b5b4 c1e3 e8g8 d3d4 b4c3 b2c3 c5d4 c3d4 b8b4 g1h1 c8b7 e2d3 b4b3 d3c2 b3b6 e1d2 d6d5 e4e5 b7a6 f1e1');
INSERT INTO evaluations VALUES(43914,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(43914,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(43914,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(43914,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(43914,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(43914,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(43914,36,'cp 96','c4b5 a7a6 e2e3 c8b7 b1c3 a6b5 f1b5 d8a5 c1d2 a5b6 a2a4 f6d5 g1f3 e7e6 e1g1 f8e7 e3e4 d5b4 f3e5 b8c6 e5c4 b6d8 d1g4 e8g8 d2h6 e7f6 f1d1 d7d5 b5c6 b4c6 e4d5 c6d4 d5e6 f7e6 h6f4 b7a6 b2b3 d8e8 f4d6 f6e7 c3e4 a6c4 d6e7 e8e7 b3c4 g8h8 g1h1');
INSERT INTO evaluations VALUES(43909,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(43909,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(43909,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(43909,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(43909,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(43909,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(43909,36,'cp 81','a7a6 e2e3 e7e6 d5e6 f7e6 b1c3 f8e7 g1f3 e8g8 a2a4 d7d5 b2b3 g8h8 f1e2 a6b5 c3b5 b8c6 c1b2 d8a5 b2c3 a5b6 f3e5 f6e4 e5c6 e4c3 b5c3 b6c6 e1g1 c6d6 c3b5 d6e5 d1d2 e5b8 h2h3 c8a6 a1c1 b8e5 f2f4');
INSERT INTO evaluations VALUES(43718,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(43718,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(43718,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(43718,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(43718,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(43718,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(43718,36,'cp 75','e2e3 e7e6 d5e6 f7e6 b1c3 f8e7 a2a4 e8g8 b2b3 d7d5 g1f3 g8h8 f1d3 a6b5 c3b5 b8c6 c1b2 f6d7 e1g1 e7f6 b2f6 d8f6 d3e2 c8b7 b5d6 b7a6 e2a6 a8a6 d6b5 a6a8 h2h3 c6b4 f3d2 f6g6 b5c7 a8a7');
INSERT INTO evaluations VALUES(43717,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(43717,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(43717,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(43717,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(43717,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(43717,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(43717,36,'cp 88','e7e6 d5e6 f7e6 b1c3 f8e7 a2a4 d7d5 g1f3 e8g8 b2b3 c8b7 f1e2 d8a5 c1d2 a5d8 f3g5 b7c8 d1c2 a6b5 c3b5 h7h6 g5f3 f6e4 e1g1 b8c6 a1d1 c8d7 d2e1 a8c8 c2b1 d7e8 e2d3 e4d6 d1c1 e6e5 c1c5 e5e4');
INSERT INTO evaluations VALUES(10902,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(10902,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(10902,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(10902,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(10902,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(10902,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(10902,36,'cp 37','e1g1 e8g8 f1e1 d7d6 a2a4 h7h6 c2c3 a7a6 b1d2 c6a5 c4a2 c7c5 d3d4 d8c7 b2b3 b7b5 c1a3 b5a4 b3a4 a8b8 d2f1 f8e8 d4d5 g7g6 d1c2 c8d7 f3d2 f6h5 g2g3 e7d8 f1e3 h5f6 h2h3');
INSERT INTO evaluations VALUES(10901,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(10901,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(10901,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(10901,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(10901,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(10901,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(10901,36,'cp 13','e8g8 f1e1 d7d6 a2a4 c8e6 b1c3 e6g4 c1e3 c6b4 c4b3 c7c5 h2h3 g4e6 b3c4 h7h6 e3d2 d8d7 a4a5 b7b6 f3h2 b4c6 c3d5 e6d5 e4d5 c6a5 d2a5 b6a5 a1a5 d7b7 a5b5 b7c7 h2f1 a7a5 f1e3 a5a4 c2c3');
INSERT INTO evaluations VALUES(4586,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(4586,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(4586,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(4586,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(4586,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(4586,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(4586,36,'cp 25','a2a4 d7d6 c2c3 a7a5 b1a3 h7h6 f1e1 c8e6 h2h3 e6c4 a3c4 f8e8 c4e3 e7f8 d1b3 d8d7 c1d2 c6d8 b3c2 d8e6 e3c4 b7b6 d2e3 d7c6 b2b4 a5b4 c3b4 b6b5 a4b5 c6b5 a1a5 a8a5 b4a5 e8a8 e1a1 b5c6 c2a4 c6a6 g2g3 d6d5 e4d5 f6d5 c4e5 d5e3');
INSERT INTO evaluations VALUES(48469,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(48469,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48469,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(48469,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(48469,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(48469,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(48469,36,'cp 30','g1f3 c7c5 e1g1 a7a6 b2b3 b7b5 c4e2 b8d7 a2a4 b5b4 b1d2 c8b7 d2c4 f8e7 c1b2 e8g8 a1c1 a8c8 f3e5 c5d4 e5d7 f6d7 d1d4 e7f6 d4a7 d8c7 b2f6 d7f6 f1d1 f6e4 e2d3 b7d5 a7c7 c8c7 c4b6 c7c3 b6d5 e6d5 c1c3 e4c3');
INSERT INTO evaluations VALUES(14212,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(14212,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(14212,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(14212,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(14212,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(14212,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(14212,36,'cp 164','c4d5 f6d5 e2e4 d5b6 d4d5 c6b8 a2a4 a7a5 c1e3 e7e6 b1c3 f8b4 f1b5 c8d7 d5e6 f7e6 f3e5 d8f6 e5d7 b6d7 e1g1 c7c6 b5c4 b8a6 a1c1 a6c7 c3a2 b4e7 c4e2 f6b2 e3d4 b2a3 e2h5 e8d8 a2c3 d8c8 d4g7 h8g8 g7d4 g8d8 d1c2 d7f6 d4b6 f6h5 b6c7 d8d7 c2e2 d7c7');
INSERT INTO evaluations VALUES(14185,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(14185,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(14185,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(14185,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(14185,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(14185,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(14185,36,'cp 161','f6d5 e2e4 d5b6 d4d5 c6b8 a2a4 a7a5 b1c3 e7e6 c1e3 f8b4 f1b5 c8d7 d5e6 f7e6 f3e5 d8f6 e5d7 b6d7 e1g1 c7c6 b5c4 b8a6 c4a6 b7a6 h2h3 f6g6 a1c1 a8d8 d1c2 e8g8 c3e2 d7e5 e2f4 g6f6 c2e2 g7g5 f4h5 f6g6 f1d1 d8d1 c1d1');
INSERT INTO evaluations VALUES(14424,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(14424,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(14424,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(14424,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(14424,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(14424,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(14424,36,'cp 160','e2e4 d5b6 d4d5 c6b8 a2a4 a7a5 c1e3 e7e6 b1c3 f8b4 f1b5 c8d7 d5e6 f7e6 f3e5 d8f6 e5d7 b6d7 e1g1 c7c6 b5c4 b8a6 c4a6 b7a6 h2h3 f6g6 d1c2 e8g8 c3e2 c6c5 e2f4 g6f6 f4d3 a8c8 a1d1 f6e7 b2b3 c5c4 b3c4 d7c5 f2f4 e7c7 d3e5');
INSERT INTO evaluations VALUES(14427,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(14427,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(14427,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(14427,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(14427,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(14427,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(14427,36,'cp 163','d5b6 d4d5 c6b8 a2a4 a7a5 c1e3 e7e6 b1c3 f8b4 f1b5 c8d7 d5e6 f7e6 f3e5 d8f6 e5d7 b6d7 e1g1 c7c6 b5c4 b8a6 c4a6 b7a6 h2h3 f6g6 d1c2 e8g8 c3e2 g6f7 a1c1 c6c5 e2f4 d7e5 b2b3 h7h6 f1d1 g8h7 f4d3 e5d3 d1d3 f7g6 c2c4 a8d8 c1d1 d8d3 c4d3 f8f3 d3c4');
INSERT INTO evaluations VALUES(14155,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(14155,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(14155,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(14155,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(14155,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(14155,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(14155,36,'cp 167','d4d5 c6b8 a2a4 a7a5 c1e3 e7e6 b1c3 f8b4 f1b5 c8d7 d5e6 f7e6 f3e5 d8f6 e5d7 b6d7 e1g1 c7c6 b5c4 b8a6 c4a6 b7a6 h2h3 f6g6 a1c1 a8d8 d1c2 e8g8 c3e2 d7e5 e2f4 g6f6 c2e2 b4d6 f4h5 f6h4 f2f4 e5g6 e4e5 d6b4 g1h2 c6c5 c1c4 h4e7 h5g3 g6h4 e2c2 h4f5 g3f5 e6f5');
INSERT INTO evaluations VALUES(14149,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(14149,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(14149,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(14149,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(14149,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(14149,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(14149,36,'cp 186','c6b8 a2a4 a7a5 b1c3 e7e6 c1e3 f8b4 f1b5 c8d7 d5e6 f7e6 f3e5 d8f6 e5d7 b6d7 e1g1 c7c6 b5c4 b8a6 c4a6 b7a6 a1c1 f6g6 c3e2 g6e4 e2g3 e4g6 h2h4 a8d8 h4h5 g6f7 h5h6 e8g8 h6g7 f8e8 d1h5 d7f6 h5h4 c6c5 e3c5 f7g7 c5b4 a5b4 c1c4 a6a5 f1c1 f6d5');
INSERT INTO evaluations VALUES(56204,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(56204,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56204,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(56204,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(56204,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(56204,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(56204,36,'cp 10','c7c5 g1f3 c5c4 d3e2 f8e7 b2b3 c4b3 a2b3 g8f6 e1g1 e8g8 c1f4 b8c6 c2c3 c8f5 f1e1 f8e8 b1d2 a7a5 d2f1 h7h6 f1e3 f5g6 h2h3 e7d6 f4d6 d8d6 e2f1 g6e4 d1e2 a8b8 f3d2 b7b5 d2e4 f6e4');
INSERT INTO evaluations VALUES(54521,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(54521,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(54521,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(54521,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(54521,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54521,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54521,36,'cp 12','g1f3 c5c4 d3e2 g8f6 b2b3 c4b3 a2b3 f8e7 e2d3 b8c6 e1g1 e8g8 f1e1 c8g4 c2c3 a7a5 c1f4 f8e8 b1d2 f6h5 f4e3 h5f6 e3g5 h7h6 g5h4 f6h5 h2h3 g4e6 h4e7 d8e7 d2f1 h5f4 f1g3 e7f6 e1e3 e6d7 d3b5 e8e7 b5c6 b7c6 e3e7 f6e7');
INSERT INTO evaluations VALUES(54520,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(54520,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(54520,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54520,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(54520,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(54520,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(54520,36,'cp 7','c5c4 d3e2 g8f6 b2b3 c4b3 a2b3 f8e7 e1g1 e8g8 f1e1 b8c6 e2d3 c8g4 c2c3 d8c7 c1g5 f8e8 h2h3 g4h5 b1d2 h5g6 d3g6 h7g6 g5f6 e7f6 e1e8 a8e8 d2f1 c6e7 a1a7 c7c3 a7b7 e7f5 b7b5 e8a8 b5d5 a8a1 d5c5 c3b2 c5c2 a1d1 c2b2 f5d4');
INSERT INTO evaluations VALUES(1194,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(1194,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(1194,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(1194,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(1194,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(1194,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(1194,36,'cp -53','e5g4 d4e4 g4e3 g8e7 b1c3 e4e6 d2d3 b7b6 h2h4 c8b7 d1g4 e6g4 e3g4 h7h5 g4e3 e7g6 c1d2 c6c5 e1c1 f7f6 f2f3 e8f7 e3c4 a8d8 d1e1 g6e7 c4e3 e7c6 e3f5 b7a8 f5g3 d8d7 c3e4 f8e7 g3f5 c6d4 f5e7 d7e7');
INSERT INTO evaluations VALUES(1202,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(1202,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(1202,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(1202,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(1202,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(1202,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(1202,36,'cp -72','d4e4 g4e3 e4g6 d2d3 c8d7 b1c3 e8c8 d1f3 c6c5 c1d2 g8e7 e1c1 e7c6 h2h4 h7h5 d1e1 c6d4 f3g3 g6g3 f2g3 f7f6 h1f1 d7e6 e3c4 e6f7 f1f2 b7b6 c1b1 h8g8 c4e5 f7e8 e5c4 c8b7 c3e2 e8g6 c4e3 d4e2 e1e2 d8e8 b2b3 b6b5 e3d5');
INSERT INTO evaluations VALUES(1209,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(1209,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(1209,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(1209,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(1209,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(1209,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(1209,36,'cp -72','g4e3 e4g6 d2d3 c8d7 d1f3 e8c8 c1d2 c6c5 b1c3 g8e7 h2h4 h7h5 e1c1 e7c6 f3g3 g6g3 f2g3 f7f6 d1e1 c6d4 h1f1 b7b6 e3c4 d7g4 d2f4 g4e6 c3e4 d4c6 e4c3 e6f7 a2a4 c8b7 b2b3 f8e7 e1e4 f6f5 e4e1 e7f6 c1d2');
INSERT INTO evaluations VALUES(1212,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(1212,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(1212,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(1212,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(1212,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(1212,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(1212,36,'cp -70','e4g6 b1c3 c8d7 d1f3 e8c8 d2d3 g8e7 h2h4 h7h5 c1d2 c6c5 e1c1 e7c6 d1e1 c6d4 f3g3 g6g3 f2g3 f7f6 h1f1 b7b6 e3c4 b6b5 c4e3 d7e6 a2a3 e6f7 e3f5 c5c4 f5d4 d8d4 e1e4 d4e4 c3e4 c4d3 c2d3 c8d7 c1c2');
INSERT INTO evaluations VALUES(47108,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(47108,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(47108,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(47108,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(47108,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47108,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(47108,36,'cp 76','d8d5 b1c3 d5a5 d2d4 c7c6 c1f4 c8g4 h2h3 g4f3 d1f3 e7e6 f1e2 b8d7 e1g1 f8e7 c3d1 e8g8 d1e3 a5b6 c2c3 b6b2 a1b1 b2c3 b1b7 c3a3 f3c6 d7b6 e2f3 a8c8 c6b5 e7d6 f4e5 h7h6 f1d1 d6e5 d4e5 f6d5 e3d5 b6d5 f3d5');
INSERT INTO evaluations VALUES(33692,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(33692,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(33692,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(33692,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(33692,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(33692,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(33692,36,'cp 86','b1c3 d5a5 d2d4 c7c6 c1f4 c8f5 f3e5 b8d7 e5c4 a5d8 c4e3 f5e6 a2a4 g7g6 a4a5 f6h5 f4g5 h7h6 g5h4 f8g7 f1e2 d7f6 e1g1 h5f4 e2f3 a8c8 f1e1 e8g8 c3a4 b7b6 a5b6 a7b6 c2c4 g8h8 h2h3 f6h5 d4d5 c6d5 c4d5 e6d7 a4c3');
INSERT INTO evaluations VALUES(33687,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(33687,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(33687,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(33687,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(33687,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(33687,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(33687,36,'cp 59','d5a5 d2d4 c8f5 f3e5 c7c6 c1f4 b8d7 e5c4 a5d8 c4e3 f5e6 a2a4 g7g6 a4a5 f6h5 f4g5 h7h6 g5h4 f8g7 f1e2 d7f6 e1g1 h5f4 e2f3 a8c8 d1d2 e8g8 f1d1 f4d5 c3d5 e6d5 d2e2 d8d7 e3d5 c6d5 c2c3 e7e6 h4g3 d7c6 a5a6 b7b5 h2h3 f6e4');
INSERT INTO evaluations VALUES(50348,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50348,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50348,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50348,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50348,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(50348,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(50348,36,'cp 0','e7e5 f2f4 e5f4 c1f4 b8c6 f1e2 h7h6 f4g3 f8c5 c2c3 c5b6 b1d2 d8e7 d3d4 f6e4 d2e4 e7e4 g1f3 d7d6 e1g1 e8g8 f3h4 g7g5 e2d3 e4g4 d3e2');
INSERT INTO evaluations VALUES(53268,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(53268,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53268,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(53268,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(53268,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53268,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53268,36,'cp 21','c2c4 g8f6 b1c3 e7e6 g1f3 b8c6 c4d5 e6d5 f1b5 c5d4 f3d4 c8d7 e1g1 f8d6 d1b3 a7a6 b5e2 e8g8 c3d5 f6d5 b3d5 c6d4 e3d4 d7c6 d5h5 g7g6 h5h3 f8e8 e2f3 d8c8 f3c6 c8h3 g2h3 b7c6 c1e3 f7f5 b2b3 f5f4 e3d2 f4f3');
INSERT INTO evaluations VALUES(53245,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(53245,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(53245,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(53245,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(53245,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(53245,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(53245,36,'cp 0','g8f6 c4d5 f6d5 d4c5 e7e6 e3e4 d5f6 d1d8 e8d8 b1c3 f8c5 g1f3 c8d7 a2a3 f6g4 h2h3 g4f2 h1h2 d7c6 f3e5 c6e8 e5f3');
INSERT INTO evaluations VALUES(45349,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(45349,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(45349,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(45349,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(45349,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45349,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45349,36,'cp 0','c4d5 f6d5 b1c3 b8c6 f1c4 e7e6 g1f3 c5d4 e3d4 f8b4 d1d3 c6a5 c4b5 c8d7 e1g1 d7c6 c1g5 f7f6 g5d2 e8g8 b5c6 a5c6 a1c1 d5c7 d2f4 c7d5');
INSERT INTO evaluations VALUES(45325,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(45325,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45325,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(45325,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45325,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(45325,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(45325,36,'cp 8','f6d5 d4c5 e7e6 g1f3 f8c5 f1c4 e8g8 e1g1 b8c6 a2a3 c5d6 b1d2 d6e7 d1c2 c8d7 b2b4 a8c8 c2b3 a7a5 c4d5 e6d5 b4b5 c6b8 a3a4 d7e6 f3d4 e7b4 d2f3 b8d7 b3a2 c8c4 c1d2 d8e7 d2b4 c4b4 f1b1 f8c8 b1b4 e7b4 a1b1 b4c4 a2a1 d7c5 b5b6 c5d3 d4e6');
INSERT INTO evaluations VALUES(55405,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(55405,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(55405,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(55405,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(55405,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(55405,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(55405,36,'cp 62','b8c6 d2d4 f8e7 d4d5 c6b8 c4d3 g8f6 c1e3 f6g4 e3d2 g4f6 c2c4 c8g4 h2h3 g4h5 d2e3 f6h7 b1d2 h7g5 d3e2 h5g6 d1c2 e8g8 b2b4 b8d7 c4c5 f7f5 e3g5 e7g5 e2d3 g5d2 f3d2 f5f4 f1e1 g8h8 f2f3 a7a5 c5c6 b7c6 d5c6 d7b6 b4b5');
INSERT INTO evaluations VALUES(48783,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(48783,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(48783,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(48783,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(48783,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(48783,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(48783,36,'cp 53','e4d5 f6d5 g1f3 b8c6 e1g1 f8e7 f1e1 f7f6 h2h3 c8f5 d3d4 d5b6 c4b3 a7a5 a2a4 c6d4 f3d4 d8d4 d1f3 f5d7 f3b7 d4d6 b7f3 d6c6 f3h5 g7g6 h5e2 d7e6 b3e6 c6e6 b1c3 e8g8 c3b5 a8c8 c1d2 c7c6 b5c3 b6c4 c3e4 f8d8 a1d1 c4d2 d1d2 d8d2 e4d2');
INSERT INTO evaluations VALUES(48589,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(48589,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(48589,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(48589,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(48589,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(48589,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48589,36,'cp 41','f6d5 g1f3 b8c6 e1g1 f8e7 f1e1 f7f6 h2h3 c8f5 b1c3 f5e6 c4b3 d8d7 d3d4 d5c3 b2c3 e8c8 d1d3 c8b8 c1e3 g7g5 d4e5 d7d3 c2d3 e6b3 a2b3 d8d3 f3d4 f6e5 d4c6 b7c6 e3a7 b8b7 a7e3 h8b8 e1c1 b7c8 a1b1 b8a8 c1c2 e5e4 c3c4 d3d7 g1f1 e7f6');
INSERT INTO evaluations VALUES(49006,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(49006,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(49006,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(49006,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(49006,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(49006,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(49006,36,'cp 56','g1f3 b8c6 e1g1 f8e7 f1e1 f7f6 d3d4 d5b6 c4b3 c6d4 f3d4 d8d4 d1d4 e5d4 c1f4 c8f5 f4c7 e8d7 b1d2 e7c5 d2e4 c5b4 c2c3 d4c3 b2c3 d7c7 c3b4 f5e4 e1e4 a8e8 a1c1 c7b8 e4g4 e8e7 c1d1 b8c7 g4g3 c7b8 g1f1 h8c8 g3d3 e7e4 d3d4 e4d4 d1d4 b8c7 f1e2');
INSERT INTO evaluations VALUES(49002,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(49002,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(49002,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(49002,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(49002,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(49002,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(49002,36,'cp 35','b8c6 e1g1 f8e7 f1e1 f7f6 d3d4 d5b6 c4b5 e8g8 b5c6 b7c6 d4e5 d8d1 e1d1 f8e8 c1f4 c8f5 b1d2 f5c2 d1c1 c2d3 c1c6 b6d5 f4g3 f6f5 d2b3 g7g5 a1d1 f5f4 d1d3 d5b4 d3c3 a8d8 g1f1 b4c6 g3f4 g5f4 c3c6 d8d1 f1e2 d1b1 c6c2');
INSERT INTO evaluations VALUES(21607,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(21607,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(21607,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(21607,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(21607,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(21607,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(21607,36,'cp 16','b1c3 g8f6 f1b5 f8b4 c1g5 e8g8 e1g1 b4c3 b2c3 h7h6 g5h4 d8d6 f1e1 c8g4 d1d3 g4f3 d3f3 f6e4 h4g3 d6d7 c3c4 e4g3 h2g3 a8d8 c2c3 a7a6 b5c6 d7c6 c4d5 c6d5 e1e5 d5f3 g2f3 f8e8 e5c5 d8d7 g1f1 b7b6 c5c6 a6a5 a1e1 e8e1 f1e1 d7e7 e1d1 a5a4');
INSERT INTO evaluations VALUES(21593,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21593,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(21593,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(21593,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(21593,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(21593,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(21593,36,'cp 12','g8f6 f1b5 f8b4 e1g1 e8g8 b5c6 b7c6 f3e5 c8a6 f1e1 f8e8 c1g5 d8d6 e1e3 f6d7 g5f4 d7e5 f4e5 d6d7 e3g3 b4f8 h2h3 g8h8 d1g4 d7g4 h3g4 f8d6 e5d6 c7d6 g4g5 h8g8 b2b3 a6c8 g3e3 e8e3 f2e3 c8f5 a1c1 a8e8 g1f2');
INSERT INTO evaluations VALUES(15041,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(15041,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(15041,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(15041,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(15041,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(15041,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(15041,36,'cp 20','f1b5 f8b4 e1g1 e8g8 c3e2 c6e7 c2c3 b4d6 e2g3 e7g6 f1e1 c7c6 b5d3 c8d7 d1c2 g6f4 d3f5 d8c7 f5d7 f6d7 c1f4 d6f4 g3f5 f8e8 g2g3 f4d6 f5d6 c7d6 h2h4 h7h6 g1g2 a7a5 a2a4 d7f6 e1e5 d6d7 a1e1 f6e4 e5e8 a8e8 f3e5 d7c7 h4h5 e4f6');
INSERT INTO evaluations VALUES(14940,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(14940,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(14940,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(14940,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(14940,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(14940,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(14940,36,'cp 22','f8b4 e1g1 e8g8 c3e2 b4d6 c2c3 c6e7 e2g3 c7c6 b5d3 e7g6 f1e1 f8e8 c1g5 e8e1 d1e1 h7h6 g5f6 d8f6 e1e8 g6f8 d3f5 f6e7 e8e7 d6e7 f5c8 a8c8 g3f5 c8e8 a1e1 f8e6 h2h4 e7f8 h4h5 g8h7 g2g3 g7g6 h5g6 f7g6 f5h4 f8d6 f3e5 d6e5 e1e5 h7g7 f2f4 g7f6');
INSERT INTO evaluations VALUES(5979,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(5979,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(5979,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5979,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(5979,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5979,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(5979,36,'cp 17','e1g1 g8f6 c2c3 a7a5 h2h3 e8g8 f1e1 c5a7 c4b5 h7h6 b1d2 c6e7 d3d4 e7g6 b5d3 f8e8 d3c2 b7b5 a2a4 c7c6 d2f1 b5b4 c3b4 a5b4 d4e5 g6e5 f3e5 d6e5 c1e3 d8e7 e3a7 a8a7 f1e3 c8e6 e3f5 e7c5');
INSERT INTO evaluations VALUES(5978,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(5978,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5978,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(5978,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(5978,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5978,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(5978,36,'cp 9','g8f6 c2c3 a7a5 f1e1 e8g8 h2h3 h7h6 b1d2 c8e6 c4b5 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 b5c4 e7g6 a2a4 e6c4 e3c4 a8d8 b2b4 b7b6 b4a5 b6a5 a1b1 a7a6 d1c2 d6d5 e4d5 f6d5');
INSERT INTO evaluations VALUES(34589,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(34589,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(34589,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(34589,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(34589,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(34589,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(34589,36,'cp 11','b8c6 g1f3 c8g4 f1e2 e8c8 c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 e2g4 d8d7 g4c4 g8f6 a1d1 d7d1 e1d1 f8d6 d1c1 e5e6 c4e6 f7e6 f2f3 h8f8 e3d2 c8d7 b2b3 h7h5 h2h3 b7b6 h1e1 f6d5 c3e4 d5f4 d2f4 f8f4 c1d2 f4f5 d2e3 h5h4 e1d1');
INSERT INTO evaluations VALUES(1609,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(1609,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(1609,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1609,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(1609,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(1609,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(1609,36,'cp 14','g1f3 c8g4 f1e2 e8c8 c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 e2c4 g8f6 a1d1 d8d1 e1d1 f8d6 d1c1 e5e6 c4e6 f7e6 e3d2 h8f8 f2f3 h7h5 b2b3 c8d7 c1d1 e6e5 h1e1 d7e6 c3e4 a7a6 a2a4 b7b5 c2c4 f6e4 e1e4 b5a4 b3a4 g7g6 d2c3 f8f5');
INSERT INTO evaluations VALUES(1606,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(1606,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(1606,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(1606,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(1606,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(1606,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(1606,36,'cp 12','c8g4 f1e2 e8c8 c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 e2c4 g8f6 a1d1 d8d1 e1d1 f8d6 d1c1 e5e6 c4e6 f7e6 f2f3 h8f8 e3d2 h7h5 c1d1 c8d7 d1e2 e6e5 d2e3 a7a6 h2h3 d7e6 h1d1 d6b4 d1d3 b4e7 a2a4 c7c6 d3d1 f8d8 d1d8 e7d8 c3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(27620,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(27620,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(27620,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(27620,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(27620,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(27620,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(27620,36,'cp 18','f1e2 e8c8 c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 e2c4 g8f6 a1d1 d8d1 e1d1 f8d6 d1c1 e5e6 c4e6 f7e6 f2f3 c8d7 h1d1 a7a6 h2h3 b7b5 e3d2 h7h5 d1e1 e6e5 c3e4 d7e6 b2b3 h8d8 c2c3 f6e4 e1e4 d6e7 d2e3 d8d7 a2a4 c7c6 c3c4 h5h4 c1c2 b5a4 b3a4');
INSERT INTO evaluations VALUES(27619,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(27619,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(27619,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(27619,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(27619,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27619,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(27619,36,'cp 15','e8c8 c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 e2c4 g8f6 a1d1 d8d1 e1d1 f8d6 d1c1 e5e6 c4e6 f7e6 f2f3 c8d7 e3d2 h7h5 h2h3 b7b6 h1e1 h8f8 c1d1 f6d5 e1e4 a7a6 d1e2 d5b4 a2a3 b4d5 e4h4 g7g6 h4a4 a6a5 c3e4 d6e5 c2c3 d5f4 d2f4 f8f4');
INSERT INTO evaluations VALUES(34,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(34,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(34,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(34,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(34,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(34,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(34,36,'cp 11','c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 a1d1 d8d1 e1d1 g8f6 e2c4 f8d6 d1c1 e5e6 c4e6 f7e6 f2f3 c8d7 h2h4 h7h5 e3g5 a7a6 h1e1 b7b5 e1e2 f6d5 c3e4 h8f8 c2c3 c7c5 b2b3 c5c4 b3c4 b5c4 e4d6 d7d6 e2e4 d5c3 e4c4');
INSERT INTO evaluations VALUES(34431,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(34431,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(34431,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(34431,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34431,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(34431,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(34431,36,'cp 28','b8c6 a2a3 g8h6 f1d3 c8d7 d3c2 c5d4 c3d4 h6f5 c2f5 e6f5 b1c3 d7e6 b2b4 h7h6 e1g1 g7g5 d1d3 f8e7 c3a4 b6a6 d3a6 b7a6 c1b2 a8b8 a1c1 e8d7 a4c5 e7c5 c1c5 b8b5 f1c1 b5c5 c1c5 f5f4 f3e1 h8c8 g2g3 h6h5 g3f4 g5f4 e1d3 f4f3 d3f4 c8g8 g1h1');
INSERT INTO evaluations VALUES(19566,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(19566,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(19566,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(19566,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(19566,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(19566,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(19566,36,'cp -15','g8f6 g1f3 d7d5 e4d5 f6d5 e3b3 e7e6 a2a3 f8c5 f1d3 e8g8 e1g1 h7h6 c1d2 e6e5 b1c3 d5b6 b3b5 c5d6 d2e3 c8g4 d3e2 d8c7 f1d1 a8d8 h2h3 g4e6 d1d2 d6e7 a1d1 d8d2 d1d2 e7d8 c3a4 b6d7 a4c5 d7c5 e3c5 a7a6 b5a4 d8e7 c5e7 c6e7');
INSERT INTO evaluations VALUES(12899,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(12899,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12899,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12899,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(12899,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12899,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(12899,36,'cp 0','g1f3 d7d5 e4d5 f6d5 e3b3 e7e6 a2a3 f8e7 c2c4 d5f6 b1c3 e8g8 f1e2 b7b6 c1f4 c6d4 f3d4 d8d4 f4g3 d4c5 e1g1 c8b7 a1d1 c5c6 e2f3 c6c8 f3e2');
INSERT INTO evaluations VALUES(12900,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12900,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(12900,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12900,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12900,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12900,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12900,36,'cp -1','d7d5 e4d5 f6d5 e3b3 e7e6 a2a3 f8c5 f1d3 d8b6 b3b6 a7b6 e1g1 d5b4 c1e3 b4d3 c2d3 e8g8 d3d4 c5e7 b1c3 c6a5 f3d2 c8d7 d4d5 b6b5 f1d1 b5b4 a3b4 e7b4 d2e4 a5b3 a1a8 f8a8 d5e6 d7e6 c3b5 a8a1 d1a1 b3a1 e3c5 b4c5 e4c5');
INSERT INTO evaluations VALUES(32294,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32294,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(32294,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(32294,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32294,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(32294,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(32294,36,'cp 18','g1f3 f8d6 f4d6 c7d6 c2c4 d5c4 f1c4 g8f6 d1b3 d8e7 c4b5 b8d7 e1g1 e8g8 b1c3 f6e4 f1c1 d7f6 c3e4 f5e4 f3d2 e4d5 b3c3 f6e4 d2e4 d5e4 c3c7 e7c7 c1c7 f8c8 a1c1 e4c6 c7c8 a8c8 b5c6 c8c6 c1c6 b7c6 g2g3 g7g6 g1g2 g8g7 g2f3 a7a5');
INSERT INTO evaluations VALUES(47000,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47000,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47000,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47000,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47000,36,'cp 0','g7g6 g1f3 f8g7 h2h3 c7c5 f1e2 b8c6 e1g1 d8b6 b1a3 c8e6 c2c4 b6b2 a3b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(46684,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46684,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46684,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46684,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46684,36,'cp 0','f1e2 f8g7 g1f3 c7c5 h2h3 b8c6 e1g1 f6d5 f4g3 c5d4 e3d4 d8b6 c2c4 d5f6 b1c3 b6b2 c3b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(46683,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46683,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46683,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46683,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46683,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46683,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46683,36,'cp 0','f8g7 g1f3 c7c5 h2h3 b8c6 e1g1 f6d5 f4g3 c5d4 e3d4 d8b6 c2c4 d5f6 b1c3 b6b2 c3b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(57585,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(57585,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(57585,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(57585,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(57585,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(57585,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(57585,36,'cp 52','g8f6 g2g3 e7e6 c2c4 f8e7 b1c3 e8g8 f1g2 d7d5 e1g1 c7c6 d1c2 f6e4 f3e5 b8d7 e5d3 a7a5 b2b3 b7b6 c1b2 e4c3 c2c3 d7f6 f1d1 c8a6 a1c1 e7d6 d3e5 a8c8 e2e3 b6b5 c4c5 d6e5 d4e5 f6e4');
INSERT INTO evaluations VALUES(50104,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(50104,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(50104,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(50104,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(50104,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(50104,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(50104,36,'cp 57','g2g3 e7e6 c2c4 f8e7 f1g2 d7d5 d1c2 c7c6 b1c3 f6e4 c1f4 b8d7 a1d1 e7f6 e1g1 g7g5 f4e5 d7e5 f3e5 e8g8 c3e4 d5e4 f2f3 c6c5 f3e4 c5d4 e5d3 d8c7 e4e5 f6e5 d3e5 c7e5 c2d2 e5e3 d2e3 d4e3 d1d6 f5f4');
INSERT INTO evaluations VALUES(50105,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(50105,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(50105,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(50105,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(50105,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(50105,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(50105,36,'cp 51','f6e4 c2c4 e7e6 d1c2 f8e7 f1g2 c7c6 c1f4 d7d5 b1c3 b8d7 a1d1 e8g8 e1g1 a7a5 f3e1 e7g5 f4g5 d8g5 b2b3 g5e7 d1c1 b7b6 e2e3 c8a6 c3e4 f5e4 f2f3 e4f3 f1f3 c6c5 f3f8 a8f8 c4d5 e6d5 g2d5 g8h8');
INSERT INTO evaluations VALUES(21398,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(21398,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21398,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(21398,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(21398,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(21398,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(21398,36,'cp 42','c2c4 f8d6 f4d6 d8d6 b1c3 g8f6 g1f3 e8g8 a2a3 a7a6 a1c1 b7b6 h2h3 c6e7 c4d5 e6d5 f1d3 c8f5 d3f5 e7f5 e1g1 f8c8 d1d3 f5e7 f1d1 h7h6 c1c2 c7c6 c3a4 c8b8 d1c1 a6a5 f3e1 f6d7 d3e2 b8c8 a4c3');
INSERT INTO evaluations VALUES(21375,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(21375,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(21375,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(21375,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(21375,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(21375,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(21375,36,'cp 46','f8d6 f4g3 g8f6 b1c3 e8g8 g1f3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 e1g1 c8f5 a1c1 f5d3 d1d3 a7a5 c1c2 c7c6 f1c1 f8e8 f3g5 g7g6 g5f3 d8d7 a2a3 d7e6 d3b3 d6g3 h2g3 a8b8 b3a4 b8c8 f3e5 c6c5 b2b3 g8g7 e5d3');
INSERT INTO evaluations VALUES(47285,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(47285,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47285,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47285,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(47285,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(47285,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(47285,36,'cp 15','g7g6 d2d4 f8g7 c2c4 c7c6 c4d5 c6d5 f3e5 b8c6 b1c3 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 e8g8 e1g1 d8e7 f1c1 f8c8 e2e4 d5e4 c3e4 c8c1 a1c1 f6d5 e4c3 d7c6 g2d5 e6d5 e5g7 g8g7 c1e1 e7f6 e1e5 f6f3 e5e7 a8c8 d2e3 f3e3 f2e3 g7f8');
INSERT INTO evaluations VALUES(46812,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46812,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(46812,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(46812,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(46812,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46812,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(46812,36,'cp 15','c2c4 c7c6 e1g1 f8g7 d2d4 e8g8 c4d5 c6d5 b1c3 f6e4 f3d2 e4d2 c1d2 e7e6 d2e3 b8c6 a1c1 d8f6 h2h4 c8d7 f1e1 h7h6 d1d2 c6d4 e3h6 g7h6 d2h6 d4f5 h6f4 d7c6 e2e4 d5e4 g2e4 f8c8 e4g2 c6g2 g1g2 c8d8 c1d1 a8c8');
INSERT INTO evaluations VALUES(13235,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(13235,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(13235,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(13235,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(13235,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(13235,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(13235,36,'cp 48','d2d4 c5d4 f3d4 e7e6 c1e3 a7a6 c4b3 f8e7 d1e2 c6d4 e3d4 b7b5 e1c1 e8g8 f2f4 d8c7 c1b1 b5b4 e4e5 d6e5 d4e5 c7b7 e5f6 e7f6 c3e4 f6e7 h2h4 a6a5 h4h5 a5a4 b3c4 b7c7 h5h6 g7g6 h1f1 a4a3 g2g4');
INSERT INTO evaluations VALUES(13270,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(13270,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(13270,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(13270,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(13270,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(13270,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(13270,36,'cp 34','c5d4 f3d4 e7e6 c1e3 f8e7 d1e2 c6d4 e3d4 a7a6 c4b3 b7b5 e1c1 e8g8 a2a3 d8c7 c1b1 a8b8 e4e5 f6d7 e5d6 e7d6 c3e4 d6e5 d4e5 c7e5 h1e1 d7c5 e4c5 e5c5 f2f3 c8b7 d1d3 f8d8 e1d1 d8d3 e2d3 b8c8 c2c3 b7c6 d3d4 c5d4 d1d4 g7g6 c3c4 b5c4 b3c4');
INSERT INTO evaluations VALUES(13357,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(13357,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(13357,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(13357,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(13357,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(13357,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(13357,36,'cp 46','f3d4 e7e5 d4e2 f8e7 e1g1 c8e6 c4b3 e8g8 c1g5 d8d7 d1d3 c6b4 d3d2 h7h6 g5f6 e7f6 a1d1 f8d8 b3e6 f7e6 c3d5 e6d5 d2b4 d5d4 c2c3 d6d5 b4b3 d7f7 c3d4 e5d4 e4e5 f6e5 e2d4 g8h8 d1d3 f7f6 f1d1 a8c8 g2g3 e5d4 d3d4');
INSERT INTO evaluations VALUES(53021,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(53021,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(53021,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(53021,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53021,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(53021,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53021,36,'cp 13','g1e2 e7e5 d2d3 b8c6 g2g3 a7a6 b1c3 a8b8 f1g2 b7b5 e1g1 f8e7 b2b3 g8f6 c3d5 e8g8 c1d2 b5c4 b3c4 f6d5 e4d5 c6d4 e2d4 e5d4 g2e4 f7f5 e4g2 e7g5 f2f4 g5f6 d1c2 c8d7 g2f3 d8c7 a1b1 b8b1 f1b1 f8b8 h2h3 g7g6 g1h1 b8b1 c2b1 f6e7');
INSERT INTO evaluations VALUES(6488,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(6488,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(6488,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(6488,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(6488,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(6488,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(6488,36,'cp 116','d2d4 e5d4 c3d4 c5b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c8d7 b5d3 e5c3 b2c3 e7d5 f1e1 d7e6 c3c4 d5f6 a1b1 e8g8 b1b7 d8d6 d3f1 f8b8 b7b8 a8b8 d1f3 e6g4 f3e3 a7a6 c1d2 g4d7 d2b4 d6c6 e3d4 b8e8 e1d1 h7h6 h2h3 e8e4 d4d2 d7f5 b4c3');
INSERT INTO evaluations VALUES(6472,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(6472,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(6472,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(6472,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(6472,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(6472,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(6472,36,'cp 115','e5d4 c3d4 c5b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c8d7 b5d3 e5c3 b2c3 e7d5 c3c4 d5f6 f1e1 d7e6 a1b1 e8g8 b1b7 d8d6 d3f1 f8b8 b7b8 a8b8 d1f3 e6g4 f3e3 a7a6 c1d2 g4e6 h2h3 d6c6 d2b4 h7h6 e3g3 f6e4 g3f4 e4d6 e1e5 d6c4 e5c5');
INSERT INTO evaluations VALUES(6568,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(6568,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(6568,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(6568,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(6568,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(6568,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(6568,36,'cp 105','c3d4 c5b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c8d7 b5d3 e5c3 b2c3 e8g8 c3c4 d7f5 d3e2 e7g6 d1d4 f8e8 f1e1 b7b6 c1b2 f7f6 c4c5 b6c5 d4c5 d8d6 b2d4 d6c5 d4c5 a8d8 c5a7 g6f4 e2c4 e8e1 a1e1 f4d5 a7c5 h7h6');
INSERT INTO evaluations VALUES(6565,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(6565,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(6565,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(6565,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(6565,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(6565,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(6565,36,'cp 108','c5b4 b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 d4d5 c6e5 f3e5 d6e5 e2b5 c8d7 b5d3 e5c3 b2c3 e8g8 c3c4 d7f5 d3e2 e7g6 c1e3 f8e8 f1e1 b7b6 a3a4 a7a5 a1c1 h7h6 h2h3 d8d6 c4c5 b6c5 e3c5 d6f6 c5e3 g6h4 c1c6 f6e5 e2f1 f5e4');
INSERT INTO evaluations VALUES(6660,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(6660,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(6660,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(6660,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(6660,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(6660,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(6660,36,'cp 107','b1c3 d7d5 e4d5 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 e2d3 e8g8 d4d5 c6b8 f1e1 h7h6 f3e5 b8d7 e5c4 d7f6 c4d6 c7d6 d1b3 c8f5 d3f1 e7g6 h2h3 b7b6 b3b4 f8e8 e1e8 f6e8 c1e3 a8c8 b4d4 e8f6 a1e1 g6e5 b2b3');
INSERT INTO evaluations VALUES(6657,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(6657,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(6657,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(6657,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(6657,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(6657,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(6657,36,'cp 116','d7d5 e4d5 d8d5 f1e2 g8e7 a2a3 b4c3 b2c3 c6a5 e1g1 e8g8 c1f4 c7c6 f1e1 e7g6 f4g3 c8e6 h2h4 h7h6 d1c2 a5c4 a3a4 e6g4 e2d3 f8e8 f3h2 g4e6 h2f1 d5d7 f1e3 b7b5 a4b5 c6b5 e3c4 e6c4 e1e8 a8e8 d3c4 b5c4 c2a4 d7a4 a1a4 e8e1 g1h2');
INSERT INTO evaluations VALUES(52366,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52366,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52366,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52366,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52366,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(52366,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(52366,36,'cp 15','d8b6 b2b3 g8f6 g1f3 d7d6 h2h3 c8f5 g2g4 f5g6 f1g2 e7e6 c2c4 h7h6 b1c3 b8d7 e1g1 f8e7 f4g3 e8g8 e2e3 d6d5 a2a3 b6a6 d1e2 f6e4 c3e4 g6e4 f1c1 f8c8 f3e5 e4g2 e5d7 g2e4 f2f3 e4h7 h3h4 d5c4 b3c4 c6c5');
INSERT INTO evaluations VALUES(52439,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(52439,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(52439,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(52439,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(52439,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(52439,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(52439,36,'cp -20','g7g6 c1b2 g8f6 b1c3 f8g7 c3d5 e8g8 d5f6 e7f6 f1c4 f8e8 c4d5 b8a6 g1e2 a6b4 e1g1 b4d5 e4d5 b7b5 c2c4 d7d6 d2d3 c8g4 f2f3 g4d7 d1d2 a7a5 a1e1 a5a4 e2g3 e8e1 f1e1 a4b3 a2b3 f6f5 b2g7 g8g7 g3e2 d8f6 e1d1 a8a1 d1a1 f6a1 d2c1');
INSERT INTO evaluations VALUES(52710,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(52710,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(52710,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(52710,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(52710,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(52710,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52710,36,'cp 0','c6d5 g1f3 b8c6 d2d4 a7a6 h2h3 c8f5 f1d3 f5d3 d1d3 e7e6 c3e2 g8f6 e1g1 f8e7 c2c3 e8g8 c1g5 d8b6 a1b1 b6c7 g5f4 c7b6');
INSERT INTO evaluations VALUES(53415,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53415,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53415,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53415,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(53415,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(53415,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53415,36,'cp 0','g1f3 b8c6 d2d4 a7a6 h2h3 c8f5 f1d3 f5d3 d1d3 e7e6 c3e2 g8f6 e1g1 f8e7 c2c3 e8g8 c1f4 d8b6 a1b1 a6a5 a2a4 b6a6 d3a6 a8a6 e2c1 f6e4 c1d3 g7g5 f4e3 f8c8 f1e1 h7h5 h3h4 g5g4 f3g5 e4d6 d3f4 e7g5 h4g5');
INSERT INTO evaluations VALUES(53411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53411,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(53411,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(53411,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53411,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53411,36,'cp 0','b8c6 d2d4 c8g4 h2h3 g4f3 d1f3 e7e6 c1e3 g8f6 f1d3 a8c8 a2a3 f8d6 e1c1 c6a5 c3b5 d6b8 b5c3 b8d6');
INSERT INTO evaluations VALUES(19876,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(19876,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19876,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19876,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19876,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19876,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19876,36,'cp -12','f1b5 c8g4 h2h3 g4f3 d1f3 e7e6 c3e2 g8f6 c2c3 a7a6 b5a4 b7b5 a4c2 c6e5 f3g3 e5g6 h3h4 f8d6 g3h3 g6f4 e2f4 d6f4 h3f3 d8c7 d2d4 f4c1 a1c1 a8b8 c2d3 e8g8 e1g1 b5b4 c3b4 c7b6 a2a3 b6d4 d3a6 d4b2 c1c7');
INSERT INTO evaluations VALUES(19717,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(19717,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(19717,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(19717,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19717,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(19717,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(19717,36,'cp -9','c8g4 h2h3 g4f3 d1f3 e7e6 c3e2 a7a6 b5c6 b7c6 e1g1 g8f6 d2d4 f8d6 c1g5 e8g8 b2b3 a6a5 f1d1 h7h6 g5h4 a5a4 c2c3 a4b3 a2b3 d6e7 h4f6 e7f6 e2f4 a8a1 d1a1 e6e5 f4h5 e5d4 h5f6 d8f6 f3f6 g7f6 c3d4 g8g7 g2g3 f8b8 g1g2 b8b3');
INSERT INTO evaluations VALUES(26247,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(26247,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(26247,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(26247,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(26247,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(26247,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(26247,36,'cp -12','h2h3 g4f3 d1f3 e7e6 c3e2 a7a6 b5c6 b7c6 e1g1 g8f6 d2d4 f8d6 c1g5 e8g8 b2b3 h7h6 g5h4 d6e7 f1d1 a6a5 e2c3 f8e8 c3a4 f6e4 h4e7 d8e7 f3f4 a8d8 f2f3 e4f6 d1e1 f6d7 f4e3 d8c8 a1d1 c6c5 c2c4 d5c4 b3c4 c5d4');
INSERT INTO evaluations VALUES(46174,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(46174,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(46174,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(46174,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(46174,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(46174,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(46174,36,'cp 34','c4d5 c5d4 d1d4 e6d5 e2e4 b8c6 f1b5 d5e4 d4d8 e8d8 f3g5 c8e6 e1g1 f8b4 c3e4 f6e4 g5e4 d8e7 c1g5 f7f6 g5e3 a7a6 b5e2 e7f7 a2a3 b4e7 a1c1 h8d8 e4c5 e7c5 c1c5 d8d7 f1d1 a8d8 d1d7 d8d7 f2f3 c6e7 c5c2 e6f5 c2d2 d7d2 e3d2 e7d5 f3f4 b7b6 g1f2 f7e6 f2f3 f5c2 e2c4');
INSERT INTO evaluations VALUES(46134,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46134,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(46134,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(46134,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46134,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46134,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46134,36,'cp 30','c5d4 d1d4 e6d5 e2e4 b8c6 f1b5 d5e4 d4d8 e8d8 f3g5 c8e6 e1g1 f8b4 c3e4 f6e4 g5e4 d8e7 c1g5 f7f6 g5e3 h8d8 a2a3 c6d4 b5a4 b4a5 e4c5 b7b5 a4d1 a5b6 c5e6 e7e6 a1c1 a8c8 c1c8 d8c8 e3d2 g7g6 f1e1 e6f7 g1f1 d4c2 e1e2 c2d4 e2e4 c8e8 e4e8 f7e8 d2e3 e8d7 h2h4 d4c6 e3b6 a7b6 d1c2 c6e5 c2e4');
INSERT INTO evaluations VALUES(46290,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46290,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46290,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(46290,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46290,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(46290,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(46290,36,'cp 20','d1d4 e6d5 e2e4 b8c6 f1b5 d5e4 d4d8 e8d8 f3g5 c8e6 e1g1 f8b4 c3e4 f6e4 g5e4 d8e7 c1g5 f7f6 g5e3 h8d8 a2a3 c6d4 b5a4 b4a5 e4c5 b7b5 a4d1 a5b6 c5e6 e7e6 a3a4 b5a4 a1a4 e6f7 g2g3 d4f5 e3c1 g7g6 g1g2 f7g7 f1e1 a8c8 d1b3 d8d4 a4d4 b6d4 e1d1 c8c7 b3a4 d4b6 c1d2 b6d4 d1c1 c7c1');
INSERT INTO evaluations VALUES(49425,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49425,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49425,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(49425,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(49425,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(49425,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49425,36,'cp 0','d7d5 c2c4 f8e7 b1c3 e8g8 b2b3 c7c5 d4c5 e7c5 c4d5 c5b4 d1c2 f6d5 c1d2 d5c3 d2c3 b4c3 c2c3 b7b6 f1e2 c8b7 e1g1 b8d7 a1c1 a8c8 c3b2 d8e7 f1d1 f8d8 b3b4 d7f6 d1d8 e7d8 a2a3 c8c1 b2c1 d8d6 c1b2 f6d7 b2d4 d6d5 e2c4 d5d4 f3d4 h7h6');
INSERT INTO evaluations VALUES(6500,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(6500,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(6500,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(6500,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(6500,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(6500,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(6500,36,'cp 74','f3e5 c5f2 e1f2 c6e5 d3d4 e5c6 c2c3 g8f6 f1d3 d7d5 e4d5 c6e7 h1e1 e8g8 f2g1 e7d5 b1a3 c8g4 d1b3 g4h5 c1g5 h5g6 d3e2 g6e4 e2f1 h7h6 g5h4 g7g5 h4g3 d5f4 a3c4 e4d5 b3c2 f4h5 c4e3 h5g3 h2g3 f8e8');
INSERT INTO evaluations VALUES(6755,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(6755,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(6755,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(6755,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(6755,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(6755,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(6755,36,'cp 76','e1g1 a7a6 b5a4 b7b5 a4b3 c8b7 d2d4 g8f6 f1e1 e8g8 c1g5 f8e8 c2c3 d6f8 a2a4 h7h6 a4b5 a6b5 a1a8 b7a8 g5h4 g7g5 h4g3 d7d6 d4e5 d6e5 b1d2 d8d6 h2h3 d6c5 d1e2 c5b6 g3h2 a8b7 e1d1 f8c5 b3c2 g5g4 h3g4');
INSERT INTO evaluations VALUES(6754,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(6754,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(6754,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(6754,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(6754,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(6754,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(6754,36,'cp 96','a7a6 b5a4 b7b5 a4b3 g8f6 d2d4 c8b7 f1e1 e8g8 c1g5 d6e7 d4d5 c6a5 f3e5 d7d6 g5f6 e7f6 e5d3 c7c5 d5c6 b7c6 b1c3 f6c3 b2c3 c6b7 d3b4 a8c8 d1d4 d8c7 a1d1 a5c4 d4a7 a6a5 b3c4 c7c4 a7a5 c4c3 a5b5 c8c4 b4d5 b7d5 b5d5 c3c2 d1d2');
INSERT INTO evaluations VALUES(4890,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(4890,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(4890,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(4890,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(4890,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(4890,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(4890,36,'cp 88','b5a4 b7b5 a4b3 c8b7 d2d4 g8f6 f1e1 e8g8 c1g5 d6e7 d4d5 c6a5 f3e5 d7d6 g5f6 e7f6 e5d3 c7c6 d5c6 b7c6 b1c3 f6c3 b2c3 d8f6 d1d2 c6b7 f2f3 a8c8 a1d1 f6c3 d2c3 c8c3 d3f4 f8d8 e4e5 c3c6 f4d5 g8f8 d5b4 c6b6 e5e6 d8e8 b4d5 b7d5 b3d5 f7e6 d5e6');
INSERT INTO evaluations VALUES(4893,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(4893,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(4893,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(4893,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(4893,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(4893,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(4893,36,'cp 82','b7b5 a4b3 g8f6 d2d4 c8b7 f1e1 e8g8 c1g5 d6e7 d4d5 c6a5 f3e5 d7d6 g5f6 e7f6 e5d3 c7c5 d5c6 b7c6 b1c3 f6c3 b2c3 d8f6 d1d2 a8c8 f2f3 c6b7 d3f2 f6c3 d2d6 a5c4 d6d3 c3f6 d3e2 h7h6 a2a4 f8d8 e1d1 d8d1 a1d1 f6g5 g2g3 g5c5 a4b5 a6b5');
INSERT INTO evaluations VALUES(4927,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(4927,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(4927,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(4927,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(4927,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(4927,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(4927,36,'cp 108','a4b3 c8b7 f1e1 g8f6 d2d4 e8g8 c1g5 d6e7 d4d5 c6a5 f3e5 d7d6 g5f6 e7f6 e5d3 c7c5 d5c6 b7c6 b1c3 f6c3 b2c3 d8f6 d1d2 c6b7 f2f3 f8c8 a1d1 f6c3 d2c3 c8c3 d3f4 c3c6 d1d4 a8c8 e1d1 a5c4 c2c3 h7h6 f4e2 c4e3 d1d3 e3c4 b3c4 c6c4 d4d6');
INSERT INTO evaluations VALUES(4922,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(4922,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(4922,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(4922,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(4922,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(4922,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(4922,36,'cp 105','g8f6 d2d4 c8b7 f1e1 e8g8 c1g5 d6e7 d4d5 c6a5 f3e5 d7d6 g5f6 e7f6 e5d3 c7c5 d5c6 b7c6 b1c3 f6c3 b2c3 a8c8 d3b4 c6b7 d1d4 d8c7 e1e3 a5c4 b3c4 c7c4 d4d6 a6a5 b4d5 b7d5 d6d5 g7g6 h2h4 c8c5 d5d6 f8c8 e4e5 c5c6 d6d7 c4e6 d7d2 b5b4 c3b4 c6c2');
INSERT INTO evaluations VALUES(7474,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(7474,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(7474,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(7474,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(7474,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(7474,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(7474,36,'cp 89','d2d4 c8b7 f1e1 e8g8 c1g5 d6e7 d4d5 c6a5 f3e5 d7d6 g5f6 e7f6 e5d3 c7c5 d5c6 b7c6 b1c3 f6c3 b2c3 a8c8 d3b4 c6b7 d1d4 d8c7 e1e3 a5c4 b3c4 c7c4 d4d6 a6a5 b4d5 b7d5 d6d5 g7g6 h2h4 c8c5 d5d6 f8e8 f2f3 c5h5 d6f4 e8d8 a1e1');
INSERT INTO evaluations VALUES(53357,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53357,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(53357,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(53357,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53357,36,'cp 0','g8f6 b1c3 b8c6 g1f3 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 e1g1 e7e6 f1d1 f8e7 a1c1 e8g8 b2b3 d8a5 c3a4 f8d8 c1c4 d8d6 a4c3 c8d8 g2g3 a5b6 c3a4 b6a5');
INSERT INTO evaluations VALUES(45420,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45420,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45420,36,'cp 5','b1c3 b8c6 g1f3 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 a1d1 e7e6 f3g3 h7h5 h2h4 d8c7 g3e5 c7e5 d4e5 c8b8 d1d2 f8b4 e1e2 b4c3 b2c3 b8b5 f2f4 e8g8 h1d1 b5c5 d2d3 f8b8 g2g3 c5a5 d1d2 a5a3 c3c4 a3a4 d3d7 a4c4');
INSERT INTO evaluations VALUES(52371,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52371,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52371,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52371,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(52371,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52371,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52371,36,'cp 28','g8f6 c2c4 d7d5 b1c3 e7e6 e2e3 b8d7 f1d3 d5c4 d3c4 b7b5 c4d3 c8b7 e1g1 a7a6 e3e4 c6c5 d4d5 c5c4 d5e6 c4d3 e6d7 d8d7 f1e1 f8e7 e4e5 f6g4 h2h3 g4h6 c1h6 g7h6 c3e4 e8c8 a2a4 e7b4 a4b5 b4e1 e4d6 c8b8 d1e1 a6b5 f3d4');
INSERT INTO evaluations VALUES(44484,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44484,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44484,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(44484,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(44484,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(44484,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(44484,36,'cp 31','c2c4 d7d5 b1c3 e7e6 c1g5 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5e7 g5h4 a7a6 c4e2 b7b5 a2a4 b5b4 c3e4 c8b7 f3d2 b8d7 e2f3 b7d5 e4f6 d7f6 f3d5 f6d5 h4e7 d5e7 d1c1 h7h6 f1d1 f8c8');
INSERT INTO evaluations VALUES(9996,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(9996,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(9996,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(9996,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(9996,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9996,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9996,36,'cp 10','e1g1 e8g8 d2d3 b4c3 b2c3 d7d6 f1e1 a7a6 b5c6 b7c6 h2h3 h7h6 f3h2 f6d7 c1e3 f8e8 h2f1 d7f8 f2f4 e5f4 e3f4 f8g6 f4g3 d8g5 d1f3 a8b8 c3c4 b8b2 f3f2 g6e5 f1e3 c8e6 g3f4 g5h5 f2d2 f7f5 e4f5 e6f5 f4h2');
INSERT INTO evaluations VALUES(9995,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9995,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(9995,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9995,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(9995,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(9995,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9995,36,'cp 10','e8g8 d2d3 b4c3 b2c3 d7d6 f1e1 a7a6 b5c6 b7c6 h2h3 h7h6 d3d4 f6d7 a1b1 a6a5 c3c4 f8e8 b1b3 a5a4 b3c3 c6c5 d4e5 d6e5 f3h2 d7f8 d1d5 a8a6 d5c5 a6d6 c5a3 d6d1 e1d1 d8d1 h2f1 e8d8 c1e3 d1e2 f1g3 d8d1 g1h2 e2e1');
INSERT INTO evaluations VALUES(4254,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(4254,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(4254,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4254,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(4254,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(4254,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4254,36,'cp 13','d2d3 b4c3 b2c3 d7d6 f3d2 c8g4 f2f3 g4e6 d1e1 f6h5 f1f2 e6d7 d2f1 c6e7 a2a4 a7a6 b5d7 d8d7 a4a5 f7f5 e4f5 e7f5 f1e3 a8b8 e3f5 d7f5 a1b1 h7h6 c3c4 h5f6 c4c5 f8f7 c5d6 c7d6 c1e3');
INSERT INTO evaluations VALUES(4255,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(4255,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(4255,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(4255,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(4255,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(4255,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(4255,36,'cp 8','b4c3 b2c3 d7d6 f1e1 a7a6 b5c6 b7c6 h2h3 h7h6 d3d4 f6d7 c3c4 f8e8 c1b2 d8f6 c4c5 d7f8 a1b1 a6a5 c5d6 c7d6 b2c3 f8g6 a2a4 g6f4 c3d2 f6g6 d2f4 e5f4 d1d3 c6c5 f3d2 g6f6 b1b5 f6d4 d3d4 c5d4');
INSERT INTO evaluations VALUES(56396,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(56396,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(56396,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(56396,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(56396,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(56396,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(56396,36,'cp 9','e4d5 d8d5 b1c3 f8b4 g1f3 g8f6 f1e2 e8g8 e1g1 b4c3 b2c3 b8c6 f1e1 c8f5 f3h4 f5d7 e2f3 d5d6 c1b2 f8e8 a2a4 h7h6 a4a5 b7b5 a5b6 a7b6 a1a8 e8a8 c3c4 a8e8 f3c6 d7c6 h4f5 d6d7 f5g3 d7d6');
INSERT INTO evaluations VALUES(56165,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(56165,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(56165,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(56165,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(56165,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56165,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(56165,36,'cp 6','d8d5 b1c3 f8b4 g1e2 b8c6 a2a3 b4c3 e2c3 d5e6 f1e2 c8d7 c1e3 g8e7 d1d2 e8g8 f2f4 e5f4 e3f4 c6d4 e1c1 e7d5 e2f3 d4f3 g2f3 d5c3 d2c3 a8c8 h1e1 e6d5 e1e7 f7f6 c3c4 d5c4 d3c4 d7c6 f4c7 f8f7 d1d8 c8d8 c7d8 f7e7 d8e7');
INSERT INTO evaluations VALUES(34945,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(34945,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(34945,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(34945,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(34945,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34945,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(34945,36,'cp 12','b1c3 f8b4 g1e2 d5e6 a2a3 b4c3 e2c3 b8c6 c3b5 e6d7 f1e2 g8e7 f2f4 a7a6 b5c3 d7d4 f4e5 e7g6 d1d2 e8g8 d2e3 c8g4 e2g4 d4g4 e3e4 c6e5 e4g4 e5g4 e1g1 a8e8 h2h3 g4e5 g1f2 f7f6 c1e3 g8f7 c3e4 e5c6 a1e1');
INSERT INTO evaluations VALUES(34941,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(34941,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(34941,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(34941,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(34941,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(34941,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34941,36,'cp 0','f8b4 g1e2 d5e6 a2a3 b4c3 e2c3 b8c6 f1e2 c8d7 c1e3 g8e7 d1d2 e8g8 e1c1 c6d4 f2f4 d4e2 d2e2 e7d5 e2d2 e5f4 e3f4 d5c3 d2c3 e6a2 b2b4 f8e8 d1e1 a2d5 g2g3 c7c6 a3a4 h7h6 a4a5 d5g2 e1g1 g2d5');
INSERT INTO evaluations VALUES(33135,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(33135,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(33135,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(33135,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(33135,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(33135,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(33135,36,'cp 15','g1f3 b8c6 f1e2 g8f6 e1g1 d5d8 c3e4 b4e7 c2c3 f6e4 d3e4 d8d1 f1d1 f7f6 a2a4 a7a5 c1e3 e8f7 f3d2 c8e6 h2h4 h7h5 g1f1 b7b6 f2f3 h8d8 e2b5 c6a7 b5c4 a7c8 c4e6 f7e6 d2c4 c8d6 c4d6 e7d6 f1e2 d6c5 d1d8 a8d8 e3c5 b6c5 g2g4 d8b8');
INSERT INTO evaluations VALUES(55603,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(55603,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55603,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55603,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55603,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55603,36,'cp 0','c7c5 f1e2 f8e7 e1g1 g8f6 c2c4 b8c6 a2a3 e8g8 b1c3 c8d7 a1b1 a7a5 b2b3 f6e8 c3d5 e8c7 d5b6 a8b8 c1d2 c7e6 b6d5 e6c7');
INSERT INTO evaluations VALUES(53899,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53899,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53899,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53899,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53899,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53899,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53899,36,'cp 0','f1e2 g8f6 c2c4 b8c6 e1g1 f8e7 b1c3 e8g8 a2a3 c8d7 a1b1 a7a5 c1d2 h7h6 b2b3 f6e8 c3d5 e8f6 d5c3');
INSERT INTO evaluations VALUES(45633,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45633,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45633,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(45633,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(45633,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(45633,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(45633,36,'cp 18','d1c2 f8d6 f1d3 e8g8 g1f3 d5c4 d3c4 b8d7 e1g1 a7a6 f1d1 b7b5 c4d3 d8c7 c3e4 f6e4 d3e4 d7f6 e4d3 c6c5 d4c5 c7c5 a2a4 c8b7 c2c5 d6c5 c1d2 f8d8 d2a5 d8b8 a1c1 c5d6 a4b5 a6b5 a5c3 b5b4 c3d4 b8c8 f3e5 c8c1 d1c1 b7d5');
INSERT INTO evaluations VALUES(56029,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(56029,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(56029,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(56029,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(56029,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(56029,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(56029,36,'cp 58','f1c4 c7c5 d4d5 g8f6 b1c3 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 d5d6 e7d8 c4d5 a8a7 a1d1 b8d7 d5c6 c8b7 c6b7 a7b7 f1e1 d8a5 e1e7 a5d8 e7e2 c5c4 g2g3 h7h6 f3e5 g7g5 e5d7 f6d7 f4e3 f8e8 e3d4 e8e2 c3e2 d8f6 d4f6 d7f6');
INSERT INTO evaluations VALUES(56008,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(56008,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(56008,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(56008,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(56008,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(56008,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(56008,36,'cp 58','c7c5 d4d5 g8f6 b1c3 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 d5d6 e7d8 c4d5 a8a7 a1d1 c5c4 f4e3 a7d7 e3c5 f6d5 c3d5 b8c6 a2a4 f7f6 d5c3 f8e8 a4b5 a6b5 c3b5 c6e5 f1e1 e5f3 g2f3 e8e1 d1e1 c8b7 b5d4 g8f7 f3f4 g7g6');
INSERT INTO evaluations VALUES(54094,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(54094,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(54094,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(54094,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(54094,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(54094,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54094,36,'cp 50','d4d5 g8f6 b1c3 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 d5d6 e7d8 c4d5 a8a7 f1d1 b8d7 f4e3 d8a5 d5b3 a5c3 b2c3 c8b7 c3c4 a7a8 f3e1 f8e8 e1d3 a8c8 a1c1 b5b4 d1e1 f6e4 b3a4');
INSERT INTO evaluations VALUES(54115,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(54115,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(54115,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(54115,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(54115,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(54115,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(54115,36,'cp 54','g8f6 b1c3 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 d5d6 e7d8 c4d5 a8a7 f1d1 b8d7 f3e5 d7e5 f4e5 f8e8 d1e1 c8e6 d5c6 e6d7 c6f3 b5b4 c3d5 f6d5 f3d5 g8f8 a1c1 d7b5 e1e4 c5c4 c1e1 f7f6 e5f4 e8e4 e1e4 c4c3 f4e3 c3b2');
INSERT INTO evaluations VALUES(23062,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(23062,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(23062,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(23062,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(23062,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(23062,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(23062,36,'cp -84','d8g5 e1g1 g5e5 c2c3 d4e6 d2d4 e5e4 f1e1 e4h4 g2g3 h4d8 d4d5 g8f6 b1d2 f8c5 d5e6 f7e6 b2b4 c5e7 a2a4 a7a5 b4a5 a8a5 d2b3 a5a8 c1f4 e8g8 a4a5 d7d5 c4f1 e7d6 f4d6 d8d6 d1d4 f6e4 f2f4 b7b6 f1g2 b6a5 g2e4 d5e4 a1a5 c8b7 a5a8 b7a8 d4d6 c7d6 b3d4 a8d5');
INSERT INTO evaluations VALUES(1977,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(1977,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(1977,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(1977,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(1977,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(1977,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(1977,36,'cp -68','c4f7 e8d8 e1g1 g5e5 c2c3 d4e6 d2d3 g7g6 c1e3 g8h6 f7e6 e5e6 c3c4 h6f7 b1c3 b7b6 d3d4 h8g8 d4d5 e6e8 e3d4 f8c5 d4f6 c5e7 d5d6 f7d6 f6e7 d8e7 c3d5 e7d8 d5f6 e8f7 f6g8 f7g8 f1e1 d6e8 e4e5 g8e6 d1d4 a7a5 d4h4 e6e7 h4d4 c7c5 d4d5 e8c7 d5d6 e7d6');
INSERT INTO evaluations VALUES(1778,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(1778,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(1778,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(1778,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(1778,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(1778,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(1778,36,'cp -82','e8d8 e1g1 g5e5 c2c3 d4e6 d2d3 g7g6 b1d2 d7d6 d3d4 e5g7 f7e6 c8e6 d4d5 e6d7 d2c4 g8f6 f1e1 f6g4 f2f4 g6g5 h2h3 g4h6 c1e3 h8g8 e3d4 g7e7 f4f5 g5g4 f5f6 e7f7 h3h4 g4g3 c4a5 d8c8 d1b3 b7b6 a5c6 h6g4 b3b5 a7a5 b2b4 g4f6 b4a5');
INSERT INTO evaluations VALUES(2333,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(2333,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(2333,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(2333,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(2333,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(2333,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(2333,36,'cp -42','e1g1 g5e5 c2c3 d4e6 d2d3 g7g5 f7e6 d7e6 f2f4 e5g7 b1a3 f8e7 a3c4 a7a5 c4e5 g8h6 f4g5 h8g8 d3d4 e7g5 c1g5 g7g5 f1f2 c8d7 d1b3 d8c8 a1f1 a8a6 e5d7 c8d7 d4d5 e6d5 b3b5 d7d8 b5b7 a6b6 b7a7 g5e3 a7a5 e3e4 f1d1 d8e7 a5d5 e4d5 d1d5');
INSERT INTO evaluations VALUES(2332,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(2332,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(2332,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(2332,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(2332,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(2332,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(2332,36,'cp -93','g5e5 c2c3 d4e6 d2d3 g7g5 f7e6 d7e6 f2f4 e5g7 f4g5 h7h6 d1h5 f8c5 d3d4 h6g5 c1g5 c5e7 g5e7 g8e7 h5f7 h8h7 f7f6 g7f6 f1f6 h7g7 b1d2 e7g6 f6f3 c8d7 f3g3 d8e7 g1f2 g7g8 f2e3 g6h4 a1g1 b7b6 d2f3 g8g3 h2g3 h4g6 g3g4 a8f8 g1f1 g6f4 g2g3');
INSERT INTO evaluations VALUES(2334,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(2334,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(2334,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(2334,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(2334,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(2334,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(2334,36,'cp -74','c2c3 d4e6 d2d3 g7g5 f7e6 d7e6 f2f4 e5g7 f4g5 h7h6 d1h5 f8c5 d3d4 h6g5 c1g5 c5e7 g5e7 g8e7 h5f7 g7h6 f7f4 h6f4 f1f4 e6e5 f4f7 a7a5 b1d2 e5d4 c3d4 c8d7 a1f1 a8a6 f7f8 h8f8 f1f8 d7e8 d2f3 e7g6 f8g8 a6b6 b2b3 a5a4 b3a4 g6f4 g2g3 b6b1 g1f2 b1b2 f2e3 f4g2 e3d3 b2a2 f3e5 a2a4 g8f8 a4a3 d3e2');
INSERT INTO evaluations VALUES(52359,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(52359,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(52359,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52359,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52359,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52359,36,'cp 2','d7d5 e4d5 c6d5 c4d5 g8f6 g1f3 f6d5 b1c3 b8c6 f1c4 d5b6 c4b5 g7g6 b5c6 b7c6 e1g1 f8g7 d2d4 e8g8 h2h3 c8e6 f1e1 a8c8 c1f4 c6c5 f4e5 g7e5 e1e5 c5d4 d1d4 b6c4 e5e2 d8d4 f3d4 c4b2 e2b2 c8c3 b2b7 c3d3');
INSERT INTO evaluations VALUES(52733,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52733,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(52733,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(52733,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(52733,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(52733,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52733,36,'cp 5','e4d5 g8f6 b1c3 c6d5 c4d5 f6d5 g1f3 b8c6 f1c4 d5b6 c4b5 g7g6 e1g1 f8g7 d2d4 e8g8 b5c6 b7c6 h2h3 c8e6 f1e1 a8c8 c1g5 h7h6 g5f4 c6c5 f4e5 g7e5 e1e5 c5d4 f3d4 b6c4 e5e2 c4b2 e2b2 c8c3 d4e6 d8d1 a1d1 f7e6 b2b7 c3c2 b7e7 f8f2 d1d8 f2f8 d8f8 g8f8 e7a7 c2e2 a7a8 f8g7');
INSERT INTO evaluations VALUES(52703,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(52703,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(52703,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(52703,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52703,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(52703,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52703,36,'cp 0','c6d5 c4d5 g8f6 b1c3 f6d5 g1f3 b8c6 f1b5 e7e6 e1g1 f8d6 d2d4 e8g8 b5d3 c8d7 f1e1 d8b6 c3d5 e6d5 c1e3 c6b4 d3b1 b4c6 b1d3');
INSERT INTO evaluations VALUES(53355,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53355,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53355,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(53355,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(53355,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53355,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(53355,36,'cp 21','c4d5 g8f6 g1f3 f6d5 b1c3 b8c6 f1c4 d5b6 c4b5 g7g6 e1g1 f8g7 d2d4 e8g8 b5c6 b7c6 h2h3 d8c7 c1e3 c8e6 f1e1 b6d5 d1c1 d5e3 c1e3 f8d8 b2b3 a8c8 a1d1 g7f6 f3e5 f6g7 c3a4 e6d5 d1c1 e7e6 e1d1 c7e7 c1c2 e7d6 d1c1');
INSERT INTO evaluations VALUES(53348,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(53348,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(53348,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(53348,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(53348,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53348,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(53348,36,'cp 12','g8f6 g1f3 f6d5 b1c3 b8c6 f1c4 d5b6 c4b5 g7g6 e1g1 f8g7 d2d4 e8g8 b5c6 b7c6 h2h3 c8e6 f1e1 a8c8 d1e2 f8e8 c1f4 e6c4 e2c2 c4d5 c3d5 d8d5 c2e2 d5b5 a1c1 b6d5 f4g5 b5e2 e1e2 e7e6 g5d2 f7f6 c1c5 g7f8 c5a5');
INSERT INTO evaluations VALUES(45410,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45410,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(45410,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45410,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45410,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(45410,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45410,36,'cp 8','b1c3 f6d5 g1f3 b8c6 f1c4 d5b6 c4b5 g7g6 d2d4 f8g7 e1g1 e8g8 b5c6 b7c6 h2h3 c8e6 f1e1 a8c8 c1g5 h7h6 g5e3 b6c4 d1e2 c4e3 e2e3 g8h7 a1d1 d8b6 c3a4 b6d8 a4c5 e6d5 e3e7 d5f3 g2f3 d8d5 g1g2 f8e8 e7e8 c8e8 e1e8');
INSERT INTO evaluations VALUES(45408,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(45408,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45408,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(45408,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(45408,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45408,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45408,36,'cp 3','f6d5 g1f3 b8c6 d2d4 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 b5e2 f7f6 e1g1 d5c3 b2c3 e7f7 c1e3 f8e7 f1b1 h8c8 b1b6 c8c3 b6a6 a8a6 e2a6 d7d5 g1g2 e7c5 a6b5 c5e3 b5d5 e6d5 f2e3');
INSERT INTO evaluations VALUES(45514,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45514,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(45514,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45514,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45514,36,'cp 7','g1f3 b8c6 d2d4 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 b5c4 d5c3 b2c3 d7d5 c4d5 e6d5 c1e3 e7e6 e1g1 a8c8 f1b1 f8c5 e3c5 c8c5 b1b6 e6d7 b6a6 h8c8 a6a7 c8c7 a7c7 d7c7 a4a5 c5c3 g1g2 d5d4 a5a6 d4d3 a6a7');
INSERT INTO evaluations VALUES(45513,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(45513,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45513,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45513,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45513,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45513,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45513,36,'cp 4','b8c6 d2d4 c8f5 f1b5 e7e6 d1a4 d8d6 f3e5 d5b6 b5c6 b7c6 a4c6 d6c6 e5c6 f7f6 e1g1 e8d7 c6a5 f8b4 a5b7 h8c8 a2a3 b4c3 b7c5 d7e8 b2c3 b6d7 c1e3 d7c5 d4c5 c8d8 c3c4 e6e5 f1d1 d8c8 f2f3 f5e6 d1d6 e6c4 a1c1');
INSERT INTO evaluations VALUES(20044,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(20044,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(20044,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(20044,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(20044,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(20044,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(20044,36,'cp 9','f1b5 c8d7 b1c3 g7g6 e1g1 f8g7 d2d3 g8f6 a2a4 e8g8 h2h3 e7e6 c1e3 a7a6 b5c4 c6e7 c4b3 d7c6 a4a5 b7b6 a5b6 d8b6 d1e1 b6c7 g2g4 f8b8 e1h4 c5c4 b3c4 d6d5 e3d4 d5c4 d4f6 b8b2 f6g7');
INSERT INTO evaluations VALUES(19929,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19929,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(19929,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19929,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(19929,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(19929,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19929,36,'cp 14','c8d7 e1g1 g7g6 b1c3 f8g7 d2d3 g8f6 a2a4 e8g8 h2h3 e7e6 c1e3 a7a6 b5c4 c6e7 c4b3 d7c6 a4a5 b7b6 a5b6 a6a5 b6b7 a8b8 d1e1 b8b7 e4e5 f6e8 e5d6 d8d6 c3e4 c6e4 d3e4 g7b2');
INSERT INTO evaluations VALUES(26061,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(26061,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26061,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(26061,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26061,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(26061,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(26061,36,'cp 0','d2d3 a7a6 b5a4 e7e6 e1g1 f8e7 c2c4 e7f6 b1c3 g8e7 h2h3 e8g8 c1e3 a8b8 f1f2 f8e8 a1b1 e7g6 b1a1');
INSERT INTO evaluations VALUES(46721,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(46721,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(46721,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(46721,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(46721,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(46721,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(46721,36,'cp 48','d2d4 f8g7 c1g5 e8g8 d1d2 d6d5 e4d5 f6d5 g5h6 d5c3 h6g7 g8g7 d2c3 d8d6 f1c4 b8d7 e1c1 d7b6 h1e1 c8g4 c4b3 a8d8 e1e4 g4f3 c3f3 a7a5 a2a4 c7c6 g2g4 e7e6 h2h4 h7h6 c2c3 b6d5 e4e5 f7f6 e5e4 g6g5 h4g5 f6g5');
INSERT INTO evaluations VALUES(14219,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(14219,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(14219,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(14219,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(14219,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(14219,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(14219,36,'cp 104','c3b5 e7e5 f4e5 c6e5 d4e5 a7a6 b5c3 d5d4 e5f6 d4c3 d1d8 e8d8 b2c3 g7f6 e2e3 f8a3 g1f3 b7b5 a1b1 c7c6 f1d3 d8c7 f3d4 c8d7 h2h4 h7h5 e1g1 a8d8 g2g3 d7e6 d3e2 a3c5 b1a1 c5a3 f1b1 e6d5 g1f1 d5e4 b1b3 a3c5 e2f3 c5d4 f3e4 d4e5 f2f4 e5d6 e4d3 h8g8 f1f2 d6c5');
INSERT INTO evaluations VALUES(14175,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(14175,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(14175,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(14175,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(14175,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(14175,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(14175,36,'cp 83','e7e5 f4e5 c6e5 d4e5 a7a6 b5c3 d5d4 e5f6 d4c3 d1d8 e8d8 b2c3 g7f6 e2e3 f8a3 a1b1 b7b5 f1d3 c7c6 g1e2 d8c7 e2d4 c8d7 h2h4 h7h5 e1g1 a3d6 d3e4 h8e8 e4f3 e8e5 b1d1 a8d8 d1d2 e5c5 f1d1 d6e7 g2g3 d7e6');
INSERT INTO evaluations VALUES(14963,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(14963,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(14963,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(14963,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(14963,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(14963,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(14963,36,'cp 90','f4e5 c6e5 d4e5 a7a6 b5c3 d5d4 e5f6 d4c3 d1d8 e8d8 b2c3 g7f6 e2e3 f8a3 g1f3 b7b5 a1b1 c7c6 f1d3 d8c7 f3d4 c8e6 d3e4 c7b6 e1g1 a8d8 g2g3 h7h5 h2h4 d8d6 f1d1 b6c7 e4f3 e6c4 d4f5 d6d1 b1d1 c4e6 f5d4 a3b2 d4e6 f7e6 c3c4 f6f5 g1g2 b5c4 f3e2');
INSERT INTO evaluations VALUES(14961,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(14961,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(14961,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(14961,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(14961,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(14961,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(14961,36,'cp 85','c6e5 d4e5 a7a6 b5c3 d5d4 e5f6 d4c3 d1d8 e8d8 b2c3 g7f6 e2e3 f8a3 g1f3 b7b5 a1d1 d8e7 f3d4 c8e6 f1d3 a3b2 d4c6 e7f8 e1d2 h7h5 d1b1 b2a3 c3c4 b5c4 d3e2 h8g8 e2f3 g8g5 c6d4 a8d8 d2e2 a3d6 h1d1 g5c5 g2g3 e6d7 b1b4 a6a5 b4b7 f8g7');
INSERT INTO evaluations VALUES(15787,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(15787,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(15787,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(15787,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(15787,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(15787,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(15787,36,'cp 87','d4e5 a7a6 b5c3 d5d4 e5f6 d4c3 d1d8 e8d8 b2c3 g7f6 e2e3 f8a3 a1d1 d8e7 f1c4 c8f5 c4d3 f5e6 g1e2 f6f5 e2d4 e7f6 e1g1 a8d8 g2g3 h7h5 h2h4 d8d6 e3e4 f5e4 d3e4 b7b6 d4e6 f6e6 e4f3 d6d1 f1d1 a3d6 c3c4 e6e7 f3d5');
INSERT INTO evaluations VALUES(15786,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(15786,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(15786,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(15786,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(15786,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(15786,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(15786,36,'cp 79','a7a6 b5c3 d5d4 e5f6 d4c3 d1d8 e8d8 b2c3 g7f6 e2e3 f8a3 g1f3 c7c6 a1b1 b7b5 f1d3 d8c7 f3d4 c8d7 h2h4 h7h5 e1g1 a3d6 b1d1 c6c5 d4f5 d6e5 c3c4 d7e6 f5e7 h8e8 e7d5 e6d5 c4d5 c5c4 d3e2 e8h8 d5d6 e5d6 d1d5 d6e5');
INSERT INTO evaluations VALUES(53092,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53092,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53092,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53092,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53092,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(53092,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53092,36,'cp 0','g7g6 b1c3 f8g7 f1b5 c8d7 e1g1 d7b5 c3b5 d8d7 a2a4 b8c6 d2d3 g8f6 d1e1 e8g8 c1d2 a8e8 h2h3 e7e5 f4e5 d6e5 e1h4 f6h5 h4g4 e8e6 g4h4 e6e8');
INSERT INTO evaluations VALUES(52628,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(52628,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52628,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(52628,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52628,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52628,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(52628,36,'cp 13','f1b5 c8d7 b5d7 d8d7 c2c3 g8f6 d1e2 b8c6 e1g1 e7e6 g1h1 f8g7 d2d4 c5d4 c3d4 e8g8 b1c3 a8c8 c1d2 c6e7 a1c1 a7a6 a2a3 d6d5 e4e5 f6e4 c3e4 c8c1 f1c1 d5e4 e2e4 f8c8 c1c8 d7c8 e4e2 c8d7');
INSERT INTO evaluations VALUES(54421,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(54421,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(54421,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(54421,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54421,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(54421,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(54421,36,'cp 35','b8c6 b1c3 g8e7 f1b5 d8b6 g1f3 c8d7 e1g1 e7f5 b5a4 f8e7 c3e2 h7h6 h2h3 b6a6 e2c3 c6b4 a4b3 a6d3 a2a3 d3d1 f1d1 b4c6 b3c2 f5h4 f3h4 e7h4 c3b5 e8g8 b5d6 h4e7 d6b7 f8b8 b7d6 e7d6 e5d6 c6a5 g2g4 a5c4 b2b3 c4d6');
INSERT INTO evaluations VALUES(20424,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(20424,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(20424,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(20424,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(20424,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(20424,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(20424,36,'cp 46','b1c3 g8e7 g1f3 e7f5 f1b5 d8b6 e1g1 c8d7 b5a4 f8e7 c3e2 h7h6 a2a3 c6a5 a4c2 f5h4 f3h4 e7h4 a1b1 h4e7 b2b4 a5c4 d1d3 a8c8 f1e1 a7a6 d3d1 b6c7 c2d3 e7g5 h2h3 g5c1 b1c1 c7b6 c1c3 d7b5 d3b1');
INSERT INTO evaluations VALUES(20420,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(20420,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(20420,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(20420,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(20420,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(20420,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(20420,36,'cp 42','g8e7 g1f3 e7f5 f1b5 c8d7 e1g1 d8b6 b5a4 h7h6 c3e2 f8e7 a2a3 c6a5 a4c2 g7g5 h2h3 e8c8 e2c3 d7e8 c2f5 e6f5 d1c2 c8b8 c2f5 b6e6 f5d3 h8g8 c3e2 d8c8 a3a4 e6b6 b2b4 b6b4 c1d2 b4c4 d3h7');
INSERT INTO evaluations VALUES(55745,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(55745,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(55745,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(55745,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(55745,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(55745,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(55745,36,'cp 104','d2d4 d6d5 b1d2 g8f6 e4e5 f6d7 c2c3 c7c5 f1d3 b8c6 e1g1 a7a5 d3c2 c5d4 c3d4 b7b5 f1e1 c6b4 c2b1 a5a4 d2f1 h7h6 c1d2 d7b6 a2a3 b4c6 d2c3 b6c4 b1c2 f8e7 f1e3 c8d7 g2g3 b5b4 a3b4 c6b4 c2a4 e8g8 g1g2 d7a4 a1a4 a8a4 d1a4');
INSERT INTO evaluations VALUES(40919,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40919,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(40919,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(40919,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40919,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(40919,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(40919,36,'cp 7','h7h6 g5f6 e7f6 g1f3 e8g8 a1c1 c7c6 f1d3 b8d7 e1g1 d5c4 d3c4 c6c5 c3e4 c5d4 e4f6 d7f6 d1d4 d8d4 f3d4 f8d8 f2f3 c8d7 f1d1 g8f8 g1f2 f8e7 e3e4 a8c8 d4e2 e6e5 c4b3 c8c1 d1c1 d7e6 b3e6 e7e6 f2e3 d8d7');
INSERT INTO evaluations VALUES(40708,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(40708,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40708,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40708,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40708,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(40708,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(40708,36,'cp 14','g5f6 e7f6 g1f3 e8g8 a1c1 c7c6 f1e2 d5c4 e2c4 b8d7 e1g1 c6c5 c3e4 c5d4 e4f6 d7f6 d1d4 d8d4 f3d4 f8d8 c1c2 c8d7 c4e2 a8c8 f1c1 c8c2 c1c2 g8f8 f2f4 f6d5 g1f2 d7e8 g2g3 g7g5 c2c1 f8e7 e3e4 d5b4 c1c7 d8d7');
INSERT INTO evaluations VALUES(40706,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40706,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(40706,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(40706,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40706,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40706,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(40706,36,'cp 0','e7f6 g1f3 e8g8 a1c1 c7c6 f1d3 b8d7 e1g1 d5c4 d3c4 e6e5 c3e4 e5d4 e4f6 d7f6 f3d4 d8e7 d1c2 f8d8 f1d1 c8g4 d1e1 g4d7 e1d1');
INSERT INTO evaluations VALUES(41638,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(41638,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(41638,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(41638,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41638,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41638,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(41638,36,'cp 11','g1f3 e8g8 c4d5 e6d5 a1c1 f6e7 f1d3 c7c6 e1g1 b8d7 d1c2 a7a5 c1e1 e7d6 a2a3 d8c7 e1c1 c7d8 f1e1 d7f6 f3e5 f6g4 f2f4 d6b8 d3f5 c8f5 c2f5 g4e5 f4e5 f7f6 e5e6 d8d6 g2g3 f8e8 e3e4 d6e6 f5e6 e8e6 e4d5 e6e1 c1e1 b8a7');
INSERT INTO evaluations VALUES(14777,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(14777,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(14777,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(14777,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(14777,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(14777,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(14777,36,'cp 61','a2a3 f6e4 d1c2 e4c3 c2c3 a7a5 c3c2 d8f6 e2e3 f8d6 c4c5 d6e7 f1d3 e6e5 d4e5 c6e5 f3e5 f6e5 c1d2 e5h5 d2c3 c7c6 e1g1 e7c5 c3g7 h8g8 c2c5 g8g7 f2f3 h5e5 f1f2 c8h3 f3f4 e5h5 f4f5 h5g5 g2g3 a5a4 a1c1');
INSERT INTO evaluations VALUES(14788,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14788,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(14788,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14788,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(14788,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(14788,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(14788,36,'cp 56','f8e7 e2e3 e8g8 b2b4 c6b8 c4c5 c7c6 f1e2 b8d7 e1g1 d8c7 c1b2 b7b6 c3a4 a7a5 d1c2 c8a6 e2a6 a8a6 c5b6 d7b6 b4b5 a6a7 b5c6 b6c4 f3e5 a7a6 f1c1 f6e4 c2e2 e4d6 a4c5 a6b6 b2c3 f8b8');
INSERT INTO evaluations VALUES(8483,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(8483,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(8483,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(8483,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(8483,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(8483,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(8483,36,'cp 58','e2e3 e8g8 b2b4 c6b8 c4c5 c7c6 f1e2 b8d7 e1g1 d8c7 c1b2 b7b6 c3a4 b6c5 a4c5 a7a5 d1c2 d7c5 b4c5 c8a6 e2a6 a8a6 b2c3 a6a8 g2g3 f6d7 a1b1 e6e5 f3e5 d7e5 d4e5 e7c5 a3a4 c5b4 c3b4 a5b4 b1b4 f8c8 e5e6 f7e6 f1b1 c6c5 b4b7 c7c6 b1b6');
INSERT INTO evaluations VALUES(8482,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(8482,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(8482,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(8482,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(8482,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(8482,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(8482,36,'cp 45','e8g8 b2b4 c6b8 c4c5 c7c6 f1e2 d8c7 e1g1 b8d7 c1b2 b7b6 c3a4 b6c5 a4c5 a7a5 d1c2 d7c5 b4c5 c8a6 e2a6 a8a6 b2c3 f6d7 a1b1 a6a8 g2g3 e6e5 f3e5 d7e5 d4e5 e7c5 a3a4 f8e8 b1b3 c5b4 c3b4 a5b4 b3b4 e8e5');
INSERT INTO evaluations VALUES(3327,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(3327,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(3327,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(3327,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(3327,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(3327,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(3327,36,'cp 66','b2b4 d5c4 f1c4 e7d6 e1g1 e6e5 h2h3 c8f5 c1b2 f8e8 d1b3 d8e7 a1d1 a7a6 c3d5 f6d5 d4e5 c6e5 b2e5 d5b6 e5d6 c7d6 c4d5 f5d7 d1c1 b6d5 b3d5 d7e6 d5d2 a8c8 f1d1 d6d5 f3d4 c8c4 c1c4 d5c4');
INSERT INTO evaluations VALUES(6426,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(6426,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6426,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6426,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6426,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(6426,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(6426,36,'cp 4','d2d4 e5d4 c3d4 c5b6 e1g1 c6a5 c4d3 g8e7 b1c3 e8g8 h2h3 a5c6 d4d5 c6d4 d5d6 c7d6 c1f4 d4f3 d1f3 b6c5 f3g3 b7b6 a1c1 e7g6 f4d6 c5d6 g3d6 c8b7 c3b5 b7c6 f2f4 d8e7 d6e7 g6e7 b5d6 e7c8 d6b5');
INSERT INTO evaluations VALUES(6445,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6445,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6445,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6445,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(6445,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6445,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(6445,36,'cp 5','e5d4 e1g1 c5b6 c3d4 d7d6 h2h3 c6a5 c4d3 g8e7 b1c3 e8g8 f1e1 h7h6 c3a4 c7c6 a4b6 a7b6 a2a4 f8e8 f3h4 c8d7 d1h5 e7d5 c1h6 g7h6 e4d5 d8g5 h5g5 h6g5 e1e8 d7e8 h4f5 e8d7 f5d6 c6d5 a1b1 a5c4 d3c4 d5c4 d6c4 a8a4 c4d6 a4d4');
INSERT INTO evaluations VALUES(6552,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(6552,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(6552,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(6552,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(6552,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(6552,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(6552,36,'cp 13','c3d4 c5b6 e1g1 c6a5 c4d3 g8e7 b1c3 e8g8 f1e1 d7d6 h2h3 h7h6 c3a4 c7c6 a4b6 a7b6 a2a4 f8e8 f3h4 c8d7 d1h5 e7d5 c1h6 g7h6 e4d5 d8g5 e1e8 a8e8 h5g5 h6g5 h4f3 g5g4 h3g4 d7g4 a1b1 g4f3 g2f3 c6d5 b1b6 e8a8 b6d6 a5c6 d3b5 c6d4 d6d5 d4b5 d5g5 g8f8');
INSERT INTO evaluations VALUES(6547,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(6547,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(6547,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6547,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(6547,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(6547,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6547,36,'cp 0','c5b6 e1g1 c6a5 c4d3 g8e7 f1e1 d7d6 h2h3 e8g8 b1c3 h7h6 a1b1 f8e8 c1e3 e7g6 d1c2 c8e6 c3a4 d6d5 e3h6 g7h6 e4d5 e6d5 d3g6 d5f3 g6f7 g8f7 c2h7 f7f8 h7h8 f8f7');
INSERT INTO evaluations VALUES(6335,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6335,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6335,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6335,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6335,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(6335,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6335,36,'cp 8','e1g1 c6a5 c4d3 g8e7 b1c3 e8g8 f1e1 d7d6 h2h3 a5c6 d3c2 e7g6 c3d5 c6e7 d5b6 a7b6 d4d5 c7c5 a2a3 c8d7 a1b1 b6b5 c2d3 c5c4 d3c2 f7f5 c1g5 d8c7 f3d4 c7c5 e4f5 e7f5 d4e6 d7e6 d5e6 a8a3');
INSERT INTO evaluations VALUES(6334,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6334,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6334,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6334,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6334,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(6334,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6334,36,'cp 0','c6a5 c4d3 g8e7 b1c3 d7d6 h2h3 e8g8 f1e1 a5c6 d3c2 e7g6 c3d5 c6e7 d5b6 a7b6 d4d5 c7c5 a2a3 c8d7 a1b1 b6b5 c2d3 c5c4 d3c2 f7f5 c1b2 f5e4 e1e4 f8f5 e4d4 f5h5 f3e1 h5e5 e1f3');
INSERT INTO evaluations VALUES(28959,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(28959,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(28959,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(28959,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(28959,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(28959,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(28959,36,'cp 77','d1b3 d8b6 c4c5 b6b3 a2b3 g7g6 e2e3 f8g7 f3h4 f5c8 b3b4 f6e4 f2f4 b8a6 f1a6 b7a6 h4f3 a8b8 a1a4 h7h5 c3e4 d5e4 f3g5 f7f5 h2h4 e8d7 e1d2 g7f6 d2c3 f6g5 h4g5 d7e6 c3c4 e6d7 c1d2 d7e6 d2c3 h8d8');
INSERT INTO evaluations VALUES(28954,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(28954,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(28954,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(28954,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(28954,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(28954,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(28954,36,'cp 56','d8b6 c4c5 b6b3 a2b3 g7g6 e2e3 f8g7 f3h4 f5c8 b3b4 f6e4 f2f4 b8a6 f1a6 b7a6 h4f3 h7h5 c3e4 d5e4 f3g5 f7f5 h2h4 a8b8 a1a4 g7f6 e1d2 b8b5 d2c2 f6g5 f4g5 e8f7 c1d2 f7e6 d2e1 f5f4 e3f4 e6d5 e1c3 h8f8 g2g3 a6a5 b4a5 c8f5 c2d2 f8b8');
INSERT INTO evaluations VALUES(32568,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(32568,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(32568,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(32568,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(32568,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(32568,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(32568,36,'cp 70','c4c5 b6b3 a2b3 g7g6 e2e3 f8g7 b3b4 f6e4 f3h4 f5c8 f2f4 b8a6 f1a6 b7a6 h4f3 a8b8 a1a4 h7h5 c3e4 d5e4 f3g5 f7f5 h2h4 g7f6 c1d2 b8b5 e1d1 h8g8 d1c2 e8d7 h1a1 c8b7 d2c3 g8f8 a1f1 f8a8 f1d1 a8b8 b2b3 b8f8 d1f1 f8b8 f1a1 b8f8 c3e1');
INSERT INTO evaluations VALUES(21402,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(21402,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(21402,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(21402,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(21402,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(21402,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(21402,36,'cp 51','e2e3 g8f6 c2c4 f8d6 f4g3 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 e1g1 c8f5 g3d6 d8d6 d3f5 e7f5 d1a4 f5e7 f1c1 f8c8 c1c2 h7h6 a1c1 c7c6 f3e1 c8c7 e1d3 a8c8 h2h3 a7a5 a2a3 e7f5 b2b4 a5b4');
INSERT INTO evaluations VALUES(55062,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(55062,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(55062,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(55062,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(55062,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(55062,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55062,36,'cp 0','g8f6 g1f3 e7e6 c1f4 f8d6 e2e3 d6f4 e3f4 e8g8 c2c3 b7b6 f1d3 c7c5 e1g1 c8a6 d3a6 b8a6 b1d2 a6c7 d1c2 a8c8 c2a4 c8a8');
INSERT INTO evaluations VALUES(47208,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47208,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47208,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(47208,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47208,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47208,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(47208,36,'cp -1','g1f3 c7c5 e2e3 e7e6 f1e2 b7b6 c2c4 d5c4 e1g1 c8b7 d4c5 d8d1 f1d1 f8c5 e2c4 b8d7 b2b3 a8c8 c1b2 c8c7 b1d2 e8g8 a1c1 f8c8 c4e2 c5b4 c1c4 f6d5 f3d4 a7a6 a2a3 b4e7 b3b4 b6b5 c4c7');
INSERT INTO evaluations VALUES(47201,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47201,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47201,36,'cp 0','e7e6 c1f4 f8d6 e2e3 d6f4 e3f4 e8g8 c2c3 d8d6 d1c1 b7b6 c1e3 c7c5 f1b5 c8a6 a2a4 a6b7 b5d3 b7a6');
INSERT INTO evaluations VALUES(48357,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48357,36,'cp 0','c1f4 f8d6 e2e3 d6f4 e3f4 e8g8 c2c3 d8d6 d1c1 b7b6 c1e3 c7c5 f1b5 c8a6 a2a4 a6b7 b5d3 b7a6');
INSERT INTO evaluations VALUES(48290,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48290,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48290,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48290,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48290,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48290,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48290,36,'cp 0','f8d6 e2e3 d6f4 e3f4 e8g8 c2c3 d8d6 d1c1 b7b6 c1e3 c7c5 f1b5 c8a6 a2a4 a6b7 b5d3 b7a6');
INSERT INTO evaluations VALUES(41795,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(41795,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41795,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41795,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41795,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41795,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41795,36,'cp 0','e2e3 d6f4 e3f4 e8g8 c2c3 b7b6 b1d2 c7c5 f3e5 d8d6 a2a3 a7a5 g2g3 b8c6 f1b5 c8b7 d2f3 f6e4 e1g1 c6e7 f1e1 d6c7 e5g4 e7f5 a3a4 b7a6 b5a6 a8a6 f3g5 e4d6 g5f3 d6e4');
INSERT INTO evaluations VALUES(47349,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(47349,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47349,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(47349,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(47349,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(47349,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(47349,36,'cp 12','a7a6 f1c4 b7b5 c4e2 e7e6 e2f3 a8a7 c3e4 f6e4 f3e4 c7c5 d4c5 d8d1 e1d1 f8c5 g1f3 f7f5 e4c2 c5e7 a2a4 b5a4 c2a4 c8d7 d1e2 e7f6 f3d4 e8f7 c1d2 f6d4 e3d4 d7a4 a1a4 b8c6 h1c1 c6e7 g2g3 h8b8 b2b4 a7d7 a4a6 e7d5 c1c6 d5b4 d2b4 b8b4 c6e6 b4d4');
INSERT INTO evaluations VALUES(43605,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(43605,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(43605,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(43605,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(43605,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(43605,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43605,36,'cp 6','f1c4 b7b5 c4e2 e7e6 e2f3 a8a7 c3e4 f6e4 f3e4 b8d7 e4c6 c8b7 d1c2 b7c6 c2c6 d8a8 c6a8 a7a8 a2a4 b5a4 a1a4 c7c5 g1f3 c5d4 f3d4 f8c5 d4f3 e8g8 e1e2 f8b8 h1d1 b8b7 a4a2 c5f8 c1d2 e6e5 d1a1 f7f6 a2a6 a8a6 a1a6 b7b2');
INSERT INTO evaluations VALUES(43601,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(43601,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(43601,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(43601,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(43601,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43601,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43601,36,'cp 8','e7e6 a2a4 c7c5 g1f3 c5d4 e3d4 f8e7 e1g1 e8g8 d1e2 b8c6 f1d1 c6b4 f3e5 b7b6 e2f3 a8a7 d4d5 f6d5 c3d5 e6d5 c4d5 b4d5 e5c6 d8d6 f3d5 d6d5 d1d5 a7c7 c6e7 c7e7 c1d2 c8e6 d5d6 e7d7 d6d7 e6d7 d2e3 b6b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(43669,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(43669,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43669,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(43669,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(43669,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(43669,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(43669,36,'cp 12','a2a4 c7c5 g1f3 b8c6 e1g1 c5d4 e3d4 f8e7 d4d5 e6d5 c3d5 f6d5 c4d5 e8g8 d1b3 c6a5 b3a2 a5c6 f1d1 c8e6 c1f4 e6d5 d1d5 d8c8 a2b3 f8d8 d5d8 c6d8 a1e1 d8e6 h2h3 b7b5 f3e5 b5a4 b3a4 c8b7 e5c6 e7f6 f4e5 a8c8');
INSERT INTO evaluations VALUES(30792,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(30792,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(30792,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(30792,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(30792,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(30792,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(30792,36,'cp 33','b1c3 e7e6 f2f3 f8b4 e2e4 b4c3 b2c3 d8h4 g2g3 h4e7 f1d3 d7d6 h2h4 b8c6 e1f2 c6a5 d1a4 b7c6 a4c2 c6b7 f2g2 g8f6 g1h3 c7c5 h3f2 h7h6 h4h5 e8g8 c2e2 f6d7 h1h4 c5d4 c3d4 e6e5 c1e3 e5d4 e3d4 d7e5 d4c3');
INSERT INTO evaluations VALUES(30789,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(30789,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(30789,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(30789,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(30789,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(30789,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30789,36,'cp 27','e7e6 e2e4 f8b4 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 g1h3 d7d6 d1a4 e7d7 a4d7 b8d7 h3f4 g7g5 f4h5 g5g4 c1g5 g4f3 f1d3 f7f6 g5h4 g8h6 e1g1 e8g8 a1e1 a8e8 f1f3 f6f5 h4g5 h6f7 g5c1 f5e4 d3e4 b7e4 e1e4 f7h8 g1g2 f8f3 g2f3 g8f7 c1g5 e8f8 h2h4 e6e5');
INSERT INTO evaluations VALUES(30644,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(30644,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(30644,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(30644,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30644,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(30644,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(30644,36,'cp 38','e2e4 f8b4 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 h2h4 d7d6 d1a4 e7d7 a4d7 b8d7 g1e2 g8e7 e2f4 h7h5 e1f2 e7c6 c1b2 c6a5 f4g2 b7a6 g2e3 c7c5 f1d3 a8c8 a1c1 d7f6 h1d1 g7g6 d1h1 a5c6 c1d1 c6a5 b2a3 f6d7 d1d2');
INSERT INTO evaluations VALUES(30650,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(30650,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(30650,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(30650,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(30650,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(30650,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(30650,36,'cp 27','f8b4 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 h2h4 d7d6 g1e2 g8f6 d1a4 e7d7 a4d7 f6d7 e2f4 b8c6 e1f2 c6a5 f4g2 b7a6 g2e3 h7h5 c1a3 c7c5 f1d3 a8c8 a1c1 g7g6 h1e1 e8e7 e1d1 h8d8 a3b2 d8h8 d1f1 a5c6 a2a4 c6a5 d3e2 c8c7 b2a3 c7c8');
INSERT INTO evaluations VALUES(27977,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(27977,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(27977,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(27977,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(27977,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(27977,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(27977,36,'cp 38','f2f3 d8h4 g2g3 b4c3 b2c3 h4e7 g1h3 d7d6 f1d3 b8c6 e1g1 e8c8 f3f4 e7d7 h3g5 c6a5 d4d5 g8f6 a1b1 h8e8 f1e1 h7h6 g5f3 c7c5 b1b2 d7c7 b2e2 f6d7 e2d2 b7a6 d1e2 a6b7 e2f2 c8b8 f3h4 b7a6 f2f1 e6d5 c4d5 a6d3');
INSERT INTO evaluations VALUES(27975,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(27975,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(27975,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(27975,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(27975,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(27975,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(27975,36,'cp 39','d8h4 g2g3 b4c3 b2c3 h4e7 d1a4 f7f5 e4e5 e7f7 e1f2 g8e7 h2h4 e8g8 g1h3 f8d8 h3f4 h7h6 c1a3 c7c5 d4d5 e7g6 f4g6 f7g6 h4h5 g6f7 a1d1 d7d6 a3c1 b8d7 e5d6 e6e5 a4c2 d8f8 h1h4 f7f6 f1h3 f6f7 f2g1 b7a6 c2a4');
INSERT INTO evaluations VALUES(50308,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(50308,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(50308,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(50308,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(50308,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(50308,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(50308,36,'cp -39','e7e6 c2c4 e6d5 c4d5 f8b4 b1c3 e8g8 g1h3 d7d6 h3f4 f8e8 g2g3 f6e4 c1d2 b4c3 b2c3 g7g5 f4d3 c8f5 h2h4 e4d2 d1d2 f5e4 h1h2 g5g4 f1g2 b8d7 g2e4 e8e4 a1b1 d7b6 d2g5 d8g5 h4g5 b6d5 b1b7 a8e8 e2e3 e4c4 b7a7 c4c3');
INSERT INTO evaluations VALUES(49347,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(49347,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(49347,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(49347,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49347,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(49347,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(49347,36,'cp -19','d5e6 f7e6 g1f3 b8c6 c2c4 e6e5 b1c3 f8b4 c1d2 d7d6 c3d5 a7a5 a2a3 b4d2 f3d2 e8g8 e2e3 c8e6 d5c3 e6f5 f1e2 g8h8 e1g1 e5e4 d2b3 c6e5 b3d4 f5g4 b2b3 d8e7 c3d5 g4e2 d1e2 e7f7 e2c2 f6d5 c4d5 f7d5');
INSERT INTO evaluations VALUES(49301,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(49301,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(49301,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(49301,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(49301,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(49301,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(49301,36,'cp -41','f7e6 c2c4 b8c6 b1c3 d7d5 e2e3 c6e5 g1f3 e5f3 d1f3 f8d6 f1d3 c7c6 f3e2 e8g8 e3e4 d8c7 c1e3 d6b4 c4d5 e6d5 e4d5 f6d5 e3d2 d5f4 d2f4 c7f4 e1g1 f4h6 a1d1 c8e6 d3e4');
INSERT INTO evaluations VALUES(49978,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(49978,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(49978,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(49978,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49978,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49978,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(49978,36,'cp -32','g1f3 b8c6 c2c4 e6e5 b1c3 f8b4 c1d2 d7d6 e2e3 e5e4 f3d4 c6d4 e3d4 b4c3 b2c3 e8g8 f1e2 d8e8 e1g1 e8g6 g1h1 c8g4 h2h3 g4e2 d1e2 f8f7 d2e3 a8f8 a2a4 c7c6 a4a5 d6d5 c4d5 f6d5 c3c4 d5b4 d4d5 c6d5');
INSERT INTO evaluations VALUES(11697,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(11697,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(11697,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11697,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11697,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(11697,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(11697,36,'cp 0','f8c5 a2a4 a8b8 b1c3 h7h6 a4b5 a6b5 e1g1 e8g8 c3e2 d7d6 c2c3 c5b6 e2g3 c8e6 b3e6 f7e6 d3d4 e5d4 c3d4 d8e8 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(7449,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7449,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7449,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7449,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(7449,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(7449,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7449,36,'cp 10','c1g5 e8g8 e1g1 d7d6 c2c3 h7h6 g5h4 c8e6 b1d2 e6b3 a2b3 c5b6 b3b4 f8e8 f1e1 g7g5 h4g3 d8d7 d2f1 c6e7 h2h3 d7e6 g3h2 g5g4 h3g4 f6g4 h2g3 f7f5 f3h2 h6h5 e4f5 e6f5 h2g4 f5g4 d1g4 h5g4 g3h4 e7g6 h4g5 g8g7');
INSERT INTO evaluations VALUES(7444,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7444,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7444,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(7444,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7444,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7444,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7444,36,'cp 0','d7d6 c2c3 e8g8 g5h4 h7h6 b1d2 c8e6 e1g1 e6b3 a2b3 c5b6 f1e1 g7g5 h4g3 d8d7 d2f1 a6a5 f1e3 c6e7 d3d4 e5d4 f3d4 f6e4 e3g4 b6d4 c3d4 f7f5 g4h6 g8g7 h6f5 d7f5 f2f3 e4g3 e1e7 g7f6 d1e1 f5d5 h2g3 d5d4 g1f1 d4d3 f1f2 d3d4');
INSERT INTO evaluations VALUES(57427,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(57427,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(57427,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(57427,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(57427,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(57427,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(57427,36,'cp 123','b8c6 f1c4 f8c5 a2a4 d7d6 e1g1 a7a6 h2h3 d8d7 d2d3 g7g5 c3d5 h7h5 c2c3 c6e7 d3d4 c5a7 d4e5 d6e5 d1b3 e7d5 c4d5 g8h6 d5b7 a8b8 b7c8 b8b3 c8d7 e8d7 f3d2 b3b8 d2c4 h6f7 b2b4 f7d6 f1d1 d7e6 c4d6 c7d6 c1a3 b8c8 b4b5 h8d8');
INSERT INTO evaluations VALUES(23237,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(23237,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(23237,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(23237,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(23237,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(23237,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(23237,36,'cp 152','f1c4 f8c5 h2h3 d7d6 a2a4 d8d7 e1g1 a7a6 d2d3 g7g5 c3d5 h7h5 c2c3 c5a7 d3d4 g5g4 f3h4 g4h3 g2g3 h3h2 g1h1 c6e7 d5e3 d7c6 d4d5 c6d7 c4e2 e8d8 e2h5 d7h3 h5f3 g8h6 f3g2 h3d7 c3c4 h6g4 a4a5 a7e3 c1e3 g4e3 f2e3 d7g4 d1g4 c8g4 f1f6 g4e2 f6f7 e2c4');
INSERT INTO evaluations VALUES(23213,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(23213,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(23213,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(23213,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(23213,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(23213,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(23213,36,'cp 126','f8c5 e1g1 d7d6 h2h3 a7a6 a2a4 d8d7 d2d3 g7g5 c3d5 h7h5 c2c3 c5a7 d3d4 c6e7 d5e3 e7g6 d4e5 d6e5 d1b3 g8e7 f1d1 d7c6 e3d5 e8f8 d5f6 c6f6 c1g5 f6c6 d1d8 f8g7 d8h8 g7h8 a1d1 c8h3 c4d5 e7d5 e4d5 c6b6 b3b6 a7b6 g2h3 a8f8 g1g2 h8g7 d5d6 c7d6 d1d6 b6c5 d6d7 f8f7 d7d8');
INSERT INTO evaluations VALUES(6921,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(6921,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(6921,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(6921,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(6921,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(6921,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(6921,36,'cp 162','f3h4 g7g6 e1g1 d7d6 c3d5 a7a6 a2a4 c6e7 c2c3 e7d5 d2d4 c5b6 c4d5 g8e7 d5c4 e7c6 c1e3 d8e7 f1e1 a6a5 h2h3 c6d8 h4f3 c8e6 c4b5 e8f8 f3d2 f8g7 d2c4 e6c4 b5c4 h8f8 d1b3 g7h8 c4d5 a8b8 a1d1 d8c6 g1h1 b8e8 f2f4 e5d4 c3d4 f6f5 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(22028,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(22028,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(22028,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(22028,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(22028,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(22028,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(22028,36,'cp 76','b1c3 g8f6 f1b5 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 b2b3 d8d7 c1b2 d7g4 f3d3 g4g6 a1e1 f8e8 g1h1 e7f8 h2h3 e8e7 e1e2 a8e8 h1h2 a5a4 b3a4 f6d7 d3f3 g6e6 a4a5 e8a8 e4e5 d6e5 f4f5');
INSERT INTO evaluations VALUES(22025,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(22025,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(22025,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(22025,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(22025,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(22025,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(22025,36,'cp 62','g8f6 f1b5 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 d8d7 h2h3 a7a5 c1e3 f8b8 e4e5 f6d5 c3d5 c6d5 f3d5 d7e6 d5e6 f7e6 e5d6 c7d6 e3d4 g8f7 a1d1 a5a4 f1f2 h7h5 f2e2 a8a6 b2b3 e7f6 g1f2 a6c6 f2f3 a4b3 c2b3 d6d5 f4f5 e6f5 d4f6 f7f6 d1d5 b8b4 e2e5');
INSERT INTO evaluations VALUES(15641,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(15641,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(15641,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(15641,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(15641,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(15641,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(15641,36,'cp 85','d4c6 b7c6 f1d3 a8b8 e1g1 g7g6 b2b3 f8g7 c1g5 h7h6 g5d2 e8g8 f2f4 f6h5 h2h3 d8h4 d1e1 g6g5 e1h4 g5h4 a1e1 h5g3 f1f3 f8e8 e4e5 d6e5 f4e5 e8e5 c3e2 g3e2 e1e2 e5e2 d3e2 g7f8 d2f4 f8c5 g1f1 c5d6 f4h6');
INSERT INTO evaluations VALUES(14702,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(14702,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(14702,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14702,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(14702,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(14702,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(14702,36,'cp 65','b7c6 f1d3 a8b8 e1g1 g7g6 b2b3 f8g7 c1g5 h7h6 g5d2 e8g8 f2f4 f6h5 h2h3 d8h4 d1e1 h4e1 a1e1 f8e8 g1h2 h5f6 c3a4 f6d7 d2a5 d7b6 a4c3 h6h5 c3b1 c8d7 b1d2 c6c5 e1e2 e8e7 f1e1 b8e8 d2f3 d7c6 g2g4 h5g4 h3g4 d6d5 e4d5 e7e2 e1e2 e8e2 d3e2');
INSERT INTO evaluations VALUES(11877,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(11877,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(11877,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(11877,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(11877,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(11877,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(11877,36,'cp 70','f1d3 a8b8 e1g1 g7g6 b2b3 f8g7 c1b2 e8g8 d1d2 f8e8 a1e1 f6h5 c3d1 g7b2 d1b2 d8f6 b2c4 h5f4 e1e3 f4d3 c2d3 f6d4 d2a5 b8b7 h2h3 c8d7 e4e5 d6d5 c4d2 e8c8 f1e1 c6c5 a5a3 d7e6 e3g3 h7h6 d2f3 d4b4 a3c1 g8g7 h3h4');
INSERT INTO evaluations VALUES(11878,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(11878,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(11878,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(11878,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(11878,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(11878,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(11878,36,'cp 64','a8b8 e1g1 g7g6 d1d2 f8g7 b2b3 e8g8 c1b2 f8e8 a1e1 f6h5 c3d1 g7b2 d1b2 d8f6 b2c4 h5f4 e1e3 f4d3 c2d3 c8e6 d2a5 e8c8 a5a6 c6c5 e4e5 f6d8 f1c1 d8d7 e5d6 e6c4 d3c4 d7d6 a6d6 c7d6 c1d1 c8c6 e3d3 b8d8 g2g4 c6a6 a2a4 a6b6 g1g2 g8f8 g4g5 d8e8');
INSERT INTO evaluations VALUES(47611,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47611,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(47611,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(47611,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(47611,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(47611,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(47611,36,'cp 62','d2d4 b8d7 a2a4 c7c6 e1g1 f8e7 h2h3 e8g8 a4a5 f8e8 f1e1 e5d4 f3d4 e7f8 c1f4 a7a6 c4f1 d7c5 d1f3 c5e6 d4e6 c8e6 f4e3 d6d5 e3b6 d8b8 a1d1 f6e4 c3e4 d5e4 f3e4 e6d5 e4d4 e8e1 d1e1 b8d6 g2g3 f8e7 c2c3 e7f6 d4d1 d6d7 c3c4 d5e6 b2b4 d7d1 e1d1');
INSERT INTO evaluations VALUES(47630,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(47630,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(47630,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(47630,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(47630,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(47630,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(47630,36,'cp 56','b8d7 a2a4 c7c6 e1g1 f8e7 h2h3 e8g8 f1e1 f8e8 a4a5 d8c7 c4f1 d7f8 d4d5 c8d7 a5a6 c6d5 e4d5 e8b8 a6b7 c7b7 f1a6 b7c7 f3d2 b8b4 d1e2 f8g6 d2c4 g6h4 b2b3 h4f5 e2d1 b4b8 c1b2 f5h4 c4e3 h4f5 d1d2 f5e3 d2e3 e7f8 a6f1');
INSERT INTO evaluations VALUES(15885,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15885,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(15885,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(15885,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(15885,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(15885,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(15885,36,'cp 66','a2a4 c7c6 e1g1 f8e7 a4a5 e8g8 h2h3 d8c7 f1e1 f8e8 c4f1 e7f8 d4d5 c6d5 c3d5 f6d5 e4d5 f7f5 c2c4 d7f6 b2b4 c7f7 c1b2 a7a6 d1b3 c8d7 b2e5 e8e5 f3e5 d6e5 e1e5 f6e4 c4c5 g8h8 a1e1 f7f6 f2f4 a8e8 e5e8 d7e8 d5d6 e8c6');
INSERT INTO evaluations VALUES(18736,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(18736,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(18736,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(18736,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(18736,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(18736,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(18736,36,'cp -32','d7d5 e4d5 e6d5 c4b5 f8d6 d2d4 c5d4 f3d4 g8e7 e1g1 e8g8 f1e1 d8b6 b5d3 c8e6 d4f3 e7g6 d1c2 a8c8 c1e3 b6c7 h2h3 c6b4 c2d1 b4d3 d1d3 f7f6 f3d4 e6d7 d3d1 a7a6 d4b3 g6e5 d1d5 g8h8');
INSERT INTO evaluations VALUES(20241,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(20241,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(20241,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(20241,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(20241,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(20241,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(20241,36,'cp -26','e4d5 e6d5 c4b5 f8d6 d2d4 c5d4 f3d4 g8e7 b1d2 a7a6 b5a4 b7b5 d4c6 e7c6 a4c2 d8e7 d1e2 c6e5 e1g1 e8g8 h2h3 f8e8 d2b3 c8d7 b3d4 e7f6 c2b3 e5c4 e2h5 d6e5 c1g5 f6d6 b3c4 b5c4 g5e3 a8b8 b2b4 c4b3 a2b3 e5d4 e3d4 b8b3 f1d1 h7h6 d4e3 d6g6 h5g6 f7g6 d1e1');
INSERT INTO evaluations VALUES(20219,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(20219,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(20219,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(20219,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(20219,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(20219,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(20219,36,'cp -28','e6d5 c4b5 f8d6 d2d4 c5d4 f3d4 g8e7 b1d2 a7a6 b5a4 b7b5 d4c6 e7c6 a4c2 d8e7 d1e2 c6e5 h2h3 e8g8 e1g1 f8e8 d2b3 c8d7 b3d4 e7f6 c2b3 e5c4 e2h5 d6e5 c1g5 f6d6 b3c4 b5c4 f1d1 d6b6 b2b4 c4b3 a2b3 e5d4 d1d4 b6b3 g5e3 b3c3');
INSERT INTO evaluations VALUES(20499,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(20499,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(20499,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(20499,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(20499,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(20499,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(20499,36,'cp -19','c4b5 f8d6 d2d4 c5d4 f3d4 g8e7 b1d2 a7a6 b5a4 b7b5 d4c6 e7c6 a4c2 d8e7 d1e2 c6e5 h2h3 e8g8 e1g1 f8e8 d2b3 c8d7 c1f4 e7f6 f4g3 g7g6 e2d1 e5c4 d1d5 d7e6 d5d4 f6d4 b3d4 d6g3 f2g3 c4b2 d4e6 e8e6 c2b3 e6e7 f1f6 g8g7 a1f1 b2c4 b3c4 b5c4');
INSERT INTO evaluations VALUES(20475,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(20475,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(20475,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(20475,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(20475,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(20475,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(20475,36,'cp -12','f8d6 d2d4 c5d4 f3d4 g8e7 b1d2 a7a6 b5a4 b7b5 d4c6 e7c6 a4c2 d8e7 d1e2 c6e5 e1g1 e8g8 h2h3 b5b4 f2f4 e5g6 e2e7 d6e7 c3b4 e7b4 d2f3 b4c5 g1h1 g6e7 f1d1 c8d7 c1d2 d7b5 a1c1 b5e2 d1e1 e2f3 g2f3 c5d6 f4f5 a6a5');
INSERT INTO evaluations VALUES(15497,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(15497,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(15497,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(15497,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(15497,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(15497,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(15497,36,'cp 106','f3g5 d6d5 e4d5 b7b5 c4b5 d8d5 b5c6 d5c6 d1f3 c6f3 g5f3 e5e4 d3e4 f6e4 b1d2 c8b7 e1g1 e8c8 d2e4 b7e4 c1e3 d8d5 f3d4 a7a5 a1d1 a5a4 c2c4 d5d7 f2f3 e4g6 d4c6 d7d1 f1d1 c8b7 c6e5 f8d6 e3f4 g6c2 d1c1 c2f5 c4c5 d6e5 c5c6 b7b6 f4e5 f7f6 e5g3 h8d8');
INSERT INTO evaluations VALUES(15474,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(15474,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(15474,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(15474,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(15474,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(15474,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(15474,36,'cp 108','d6d5 e4d5 b7b5 c4b5 d8d5 b5c6 d5c6 d1f3 c8b7 e1g1 h7h6 f3c6 b7c6 g5f3 c6f3 g2f3 f8d6 f1e1 e8g8 g1f1 f8b8 b1d2 f6d5 d2c4 d5b4 e1e2 a7a5 a2a4 f7f6 c1e3 h6h5 a1d1 b4c6 c2c3 b8d8 h2h4 c6e7 d3d4 e5d4 d1d4 e7f5 d4d5 f5h4 d5h5');
INSERT INTO evaluations VALUES(15081,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(15081,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(15081,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(15081,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(15081,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(15081,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(15081,36,'cp 120','e4d5 b7b5 c4b5 d8d5 b5c6 d5c6 d1f3 c6f3 g5f3 e5e4 d3e4 f6e4 e1g1 c8b7 b1d2 e4d2 c1d2 f7f6 f3d4 e8c8 d2e3 d8d5 f1e1 f8d6 c2c4 d5e5 a2a3 h8e8 e1d1 e5e4 a1c1 d6e5 b2b4 e4g4 g2g3 e5d4 e3d4 e8d8 f2f4 a7a6 d4e3 d8e8 c1c3 h7h5 g1f2 g7g5');
INSERT INTO evaluations VALUES(14980,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(14980,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(14980,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(14980,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(14980,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(14980,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(14980,36,'cp 106','b7b5 c4b5 d8d5 b5c6 d5c6 d1f3 c6f3 g5f3 c8b7 e1g1 b7f3 g2f3 f6d5 f1e1 f7f6 f3f4 d5b4 e1e2 b4c6 c1e3 e8f7 f4e5 c6e5 b1d2 g7g5 e3d4 e5g6 a1e1 f8b4 c2c3 h8d8 d4f6 f7f6 d2e4 f6f5 e4g3 f5g4 c3b4 d8d3 a2a3 g6f4 e2e7 a8f8');
INSERT INTO evaluations VALUES(47286,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47286,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47286,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(47286,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47286,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(47286,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47286,36,'cp 26','g1f3 g7g6 d2d4 f8g7 c2c4 c7c6 c4d5 c6d5 b1c3 b8c6 e1g1 e8g8 f3e5 c8f5 e5c6 b7c6 c1f4 f6d7 d1d2 e7e5 d4e5 d7e5 b2b3 f8e8 a1c1 h7h5 c3a4 h5h4 f4g5 g7f6 g3h4 d8d6 h2h3 e5d7 h4h5 g6h5 g5f6 d6f6 a4c5 g8h7 f1d1');
INSERT INTO evaluations VALUES(51329,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(51329,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(51329,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(51329,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(51329,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(51329,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(51329,36,'cp 87','a2a4 c7c6 a4b5 c6b5 b1c3 a7a6 c3b5 a6b5 a1a8 c8b7 a8a1 e7e6 g1e2 b7e4 e2c3 f8b4 f1e2 g8f6 e1g1 b4c3 b2c3 e8g8 c1g5 b8d7 e2f3 d8b6 f1e1 e4f3 d1f3 f6d5 g5d2 h7h6 a1a3 d7f6 h2h3 f8c8 e1b1 c8b8 b1c1 g8h7 c1d1 b8c8 d2e1 h7g8 d1b1 b5b4 c3b4 b6d4 b1a1 c4c3 a3a8');
INSERT INTO evaluations VALUES(51335,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(51335,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(51335,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(51335,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(51335,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(51335,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(51335,36,'cp 48','c7c6 a4b5 c6b5 b1c3 d8b6 c3d5 b6b7 c1f4 e7e5 f4e5 b8d7 e5f4 g8f6 d5c7 e8d8 c7a8 b7e4 g1e2 e4a8 f2f3 d7b6 e2c3 b6d5 f4d2 d5c3 d2c3 f8d6 c3a5 d8e7 b2b3 c4b3 d1b3 c8e6 b3b2 a7a6 f1d3 f6d5 a5d2 a8b7 e1g1 h8c8 f1e1 h7h6 b2a2 d5b4 d2b4 d6b4 e1e4 e7f8 a2a6 b7a6 a1a6');
INSERT INTO evaluations VALUES(51665,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(51665,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(51665,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(51665,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(51665,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(51665,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51665,36,'cp 43','a4b5 c6b5 b1c3 d8b6 c3d5 b6b7 c1f4 e7e5 f4e5 b8d7 e5f4 g8f6 d5c7 e8d8 c7a8 b7e4 g1e2 e4a8 e2c3 f8b4 f2f3 a7a5 f1e2 a8c6 e1g1 h8e8 f1e1 d7b6 f4d2 d8c7 f3f4 c6d7 e2f3 c8b7 d4d5 e8e1 d2e1 b6d5 c3d5 f6d5 e1b4 a5b4 d1d4 d7e6 g1h1 e6f6 d4c5 f6c6 c5f8');
INSERT INTO evaluations VALUES(51639,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(51639,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51639,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(51639,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(51639,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(51639,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51639,36,'cp 40','c6b5 b1c3 d8b6 b2b3 e7e6 b3c4 f8b4 c1d2 b5c4 f1c4 g8e7 g1f3 e8g8 e1g1 a7a5 c3a4 b6d6 g2g3 b4d2 d1d2 b8d7 d2d3 c8b7 f1b1 f8b8 f3g5 e7g6 e4e5 d6e7 g5e4 b7e4 d3e4 b8b1 a1b1 a8b8 b1b8 d7b8 a4c5 e7c7 c4b5 g6e7 e4a8 g8h8 g1h1 h7h6');
INSERT INTO evaluations VALUES(51680,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51680,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51680,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51680,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(51680,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(51680,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51680,36,'cp 15','b1c3 d8b6 b2b3 e7e6 b3c4 f8b4 c1d2 b5c4 f1c4 g8e7 g1e2 e8g8 e1g1 a7a5 c3a4 b6c6 c4d3 b8d7 d2f4 e6e5 f4g3 c6d6 a1c1 c8a6 d3a6 d6a6 d4e5 f8d8 e5e6 d7c5 e6f7 g8f8 a4c5 d8d1 c5a6 d1c1 e2c1 a8a6 c1d3 e7c6 f1c1 a5a4 d3b4');
INSERT INTO evaluations VALUES(51679,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51679,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(51679,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(51679,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(51679,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(51679,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(51679,36,'cp 14','d8b6 b2b3 e7e6 b3c4 f8b4 c1d2 b5c4 f1c4 g8e7 g1f3 e8g8 e1g1 a7a5 f1e1 b8d7 d4d5 e6d5 c3d5 e7d5 c4d5 a8b8 d2f4 b4e1 f3e1 d7f6 f4b8 b6b8 a1a5 b8b6 a5a4 f8d8 e1c2 f6d5 a4d4 c8g4 d1g4 d5f6 d4d8 b6d8 g4e2 h7h6 f2f3 d8a5');
INSERT INTO evaluations VALUES(22556,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(22556,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(22556,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(22556,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(22556,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(22556,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(22556,36,'cp 144','d1d4 d7d6 e1g1 c8e6 b1c3 g8f6 f1e1 d8d7 c1f4 f8e7 e4e5 d6e5 d4e5 e8g8 c4e6 f7e6 e5e6 d7e6 e1e6 e7b4 f4d2 a8d8 d2e1 d8d7 e6e2 f8d8 a2a3 b4c5 g1f1 f6g4 h2h3 g4f6 a3a4 a7a5 b2b3 c5e7 g2g3 e7b4 c3e4 f6d5 e1d2 b4d2 e2d2');
INSERT INTO evaluations VALUES(10365,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(10365,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(10365,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(10365,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(10365,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(10365,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(10365,36,'cp -28','f1e2 d7d5 e4d5 f6d5 c1d2 a7a5 e1g1 e8g8 a2a3 d5c3 d2c3 b4c3 b2c3 c8f5 f1e1 f8e8 f3d2 b7b6 d1b1 d8d6 e2f3 a8d8 b1b5 c6a7 b5b2 a7c8 a1d1 h7h6 d2c4 d6f6 c4e3 f5d7 c3c4 c8e7 e3d5 e7d5 c4d5 f6d6 c2c4 g8h8');
INSERT INTO evaluations VALUES(10364,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(10364,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(10364,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(10364,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(10364,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(10364,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(10364,36,'cp -31','d7d5 e4d5 f6d5 c1d2 d5c3 d2c3 b4c3 b2c3 e8g8 e1g1 f7f5 f1e1 f8e8 a2a4 d8f6 a4a5 c8d7 d3d4 e5e4 f3e5 c6e5 d4e5 f6e7 f2f4 d7e6 d1d4 e8d8 d4e3 h7h6 a5a6 b7b6 c3c4 e6f7 a1b1 a8b8 g2g3 g8h8');
INSERT INTO evaluations VALUES(9121,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(9121,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(9121,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(9121,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(9121,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(9121,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(9121,36,'cp -26','e4d5 f6d5 c1d2 d5c3 d2c3 b4c3 b2c3 f7f5 e1g1 e8g8 f1e1 f8e8 a2a4 d8f6 a4a5 c8d7 d1c1 a8d8 c1g5 f6g5 f3g5 h7h6 g5f3 g8f7 f3d2 d7e6 e1b1 b7b6 a5b6 a7b6 b1e1 f7f6 d2f1 g7g6 e2f3 e6d5 f3d5 d8d5 f1e3 d5d8');
INSERT INTO evaluations VALUES(53050,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(53050,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(53050,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(53050,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(53050,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(53050,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(53050,36,'cp 160','d4c5 d8a5 b1c3 a5c5 c1e3 c5a5 d1d5 a5d5 c3d5 e8d8 d5c3 b8c6 e1c1 g7g6 f2f4 c8g4 f1e2 g4e2 g1e2 d8c7 e2d4 g8f6 h1e1 a8d8 c3d5 f6d5 d4b5 c7c8 e4d5 a7a6 b5c3 c6b8 e3d4 h8g8 g2g4 b8d7 d1d3 c8b8 h2h3 d8e8 d3e3 e8c8 e3e4 f7f5 e4e6 f8g7 d4g7 g8g7');
INSERT INTO evaluations VALUES(52992,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(52992,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(52992,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(52992,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(52992,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(52992,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(52992,36,'cp 152','d8a5 b1c3 a5c5 c1e3 c5a5 d1d5 a5d5 c3d5 e8d8 f2f4 e7e6 d5c3 a7a6 a2a4 g8f6 g1f3 c8d7 e4e5 d6e5 f3e5 d8e8 e1c1 f8b4 e5c4 b8c6 e3b6 b4e7 f1e2 c6b4 e2f3 d7c6 f3c6 b4c6 f4f5 e6f5 h1e1 e8f8 c3d5 f6d5 d1d5 g7g6');
INSERT INTO evaluations VALUES(34344,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(34344,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(34344,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(34344,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(34344,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(34344,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(34344,36,'cp 166','b1c3 a5c5 c1e3 c5a5 d1d5 a5d5 c3d5 e8d8 f2f4 e7e6 d5c3 a7a6 a2a4 c8d7 g1f3 g8f6 e4e5 d6e5 f3e5 d8e8 e1c1 f8b4 e5c4 b8c6 e3b6 a8c8 g2g4 c6b8 g4g5 d7c6 h1g1 f6d5 b6d4 d5f4 d4g7 h8g8 g7e5 b4c5 c4d6 c5d6 e5d6 f4g6 f1g2 b8d7 g2c6 c8c6 c3e4 g6e5 e4f6 d7f6 d6e5');
INSERT INTO evaluations VALUES(34343,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(34343,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(34343,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(34343,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(34343,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(34343,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(34343,36,'cp 147','a5c5 c1e3 c5a5 d1d5 a5d5 c3d5 e8d8 f2f4 g8f6 d5c3 g7g5 f4f5 h7h6 f1c4 h8h7 h2h4 g5g4 e1c1 b8c6 c1b1 c8d7 g1e2 c6e5 c4b3 f8g7 e3d4 d8e8 e2f4 d7c6 d4e5 d6e5 f4d5 f6d5 b3d5 h7h8 d1d3 g7f6 g2g3 c6d5 d3d5 a8d8 h1d1 d8d5');
INSERT INTO evaluations VALUES(34342,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(34342,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(34342,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(34342,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(34342,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(34342,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(34342,36,'cp 154','c1e3 c5a5 d1d5 a5d5 c3d5 e8d8 f2f4 g8f6 d5c3 f6g4 e3d2 b8c6 h2h3 g4f6 d2e3 e7e6 e1c1 a7a6 e4e5 f6d5 c3d5 e6d5 g1e2 d6e5 f4e5 d8e8 e2f4 d5d4 e3d4 c6d4 d1d4 f8c5 d4d3 b7b5 f1e2 c8f5 d3g3 a8d8 h1f1 g7g6 e2d3 f5e6 a2a3 e8e7 g3f3 c5d4 f4e6 f7e6 c1b1 d4e5 f3f7 e7d6');
INSERT INTO evaluations VALUES(34341,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(34341,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(34341,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(34341,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(34341,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(34341,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(34341,36,'cp 163','c5a5 d1d5 a5d5 c3d5 e8d8 d5c3 g8f6 f2f4 f6g4 e3d2 b8c6 h2h3 g4f6 e1c1 e7e6 d2e3 a7a6 e4e5 f6d5 c3d5 e6d5 g1e2 d6e5 f4e5 d8e8 e2f4 d5d4 e3d4 c6d4 d1d4 f8c5 d4d3 c8f5 d3g3 a8d8 f1d3 g7g5 f4h5 f5d3 c2d3 c5d4 h1f1 e8e7 c1b1 d4e5 g3g5 d8d5 f1e1 e7d6');
INSERT INTO evaluations VALUES(54528,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(54528,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(54528,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(54528,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(54528,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(54528,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(54528,36,'cp 98','f1b5 b8c6 e1g1 g8f6 f1e1 f8e7 f3e5 d8c7 b5c6 b7c6 c1f4 c7b6 d4c5 b6c5 b1c3 c5b6 c3a4 b6b5 c2c4 b5b7 a1c1 c8e6 c4d5 f6d5 f4d2 a8c8 e5d3 b7b8 a4c5 e7c5 d3c5 e8g8 c5e6 f7e6 d1e2 c8e8 c1c2');
INSERT INTO evaluations VALUES(54463,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(54463,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(54463,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(54463,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(54463,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(54463,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(54463,36,'cp 104','b8c6 e1g1 g8f6 f1e1 c8e6 c1g5 f8d6 d4c5 d6c5 g5f6 g7f6 d1d2 e8g8 b5d3 g8g7 b1c3 f8g8 a1d1 g7h8 a2a3 d8f8 g2g3 f8g7 b2b4 c5b6 d2f4 a8c8 c3b5 a7a6 b5d4 c6d4 f3d4 e6g4 d3f5 g4d1 f5c8 g7g5 f4g5 f6g5 c8b7 d1g4');
INSERT INTO evaluations VALUES(20464,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(20464,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(20464,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(20464,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(20464,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(20464,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(20464,36,'cp 85','e1g1 g8f6 f1e1 c8e6 c1g5 f8d6 d4c5 d6c5 g5f6 g7f6 d1d2 e8g8 d2h6 c6e7 b5d3 e7g6 b1c3 f8e8 h6d2 a7a6 a1d1 c5a7 h2h3 d8b6 e1f1 e8e7 c3e2 b6d6 c2c3 a8e8 e2d4 e6d7 d3f5 d7a4 b2b3 a4d7 g2g3 g6e5 d2f4 e5f3 f4f3 a7d4 d1d4 d7f5 f3f5 e7e5 f5d3');
INSERT INTO evaluations VALUES(20462,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(20462,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(20462,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(20462,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(20462,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(20462,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(20462,36,'cp 88','g8f6 f1e1 c8e6 c1g5 f8d6 d4c5 d6c5 g5f6 g7f6 d1d2 e8g8 d2h6 c6e7 b5d3 e7g6 b1c3 f8e8 h2h3 d8b6 h6d2 c5d6 c3e2 b6b2 a1b1 b2a3 e2d4 d6f4 d2d1 a3e7 g2g3 f4e5 d4e6 f7e6 d3g6 h7g6 f3e5 f6e5 e1e5 e7f6 d1d4 b7b6 b1d1 e8e7 d1e1');
INSERT INTO evaluations VALUES(13707,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(13707,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(13707,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(13707,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(13707,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(13707,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(13707,36,'cp 89','f1e1 c8e6 c1g5 f8e7 d4c5 e7c5 g5f6 g7f6 d1d2 e8g8 d2h6 c6e7 b5d3 e7g6 b1c3 f8e8 h6d2 a7a6 a1d1 d8c7 h2h3 g6f4 c3e2 f4e2 e1e2 g8g7 d1e1 e6d7 f3h4 e8e2 e1e2 c7g3 h4f5 d7f5 d3f5 a8d8 d2d3 g3d3 f5d3 c5d6 g1f1 d6e5 c2c3 g7g8 g2g3 h7h5 f1g2 h5h4');
INSERT INTO evaluations VALUES(54176,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(54176,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(54176,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(54176,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(54176,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(54176,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(54176,36,'cp 40','c4d5 e6d5 g1f3 b8c6 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 d1d4 b6d4 e3d4 f3d5 d4f6 d5c4 c1c4 g7f6 c4c7 b7b6 g2g3 f8d8 a2a3 d8d2 f1e1 a7a5 c7c6 d2b2 e1e7 a8f8 c6f6 b2b1 g1g2 g8g7 f6f5 g7g6 f5f3 b1b2');
INSERT INTO evaluations VALUES(54127,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(54127,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(54127,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(54127,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(54127,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(54127,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54127,36,'cp 28','e6d5 g1f3 b8c6 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 e3d4 f3d1 d4b6 d1g4 b6e3 g4e6 f1d1 f8c8 c4e6 f7e6 g1f1 f6d5 f1e2 b7b5 c1c8 a8c8 d1c1 c8c1 e3c1 g8f7 c1e3 a7a6');
INSERT INTO evaluations VALUES(54514,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(54514,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(54514,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(54514,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54514,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(54514,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(54514,36,'cp 34','g1f3 b8c6 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 e3d4 f3d1 d4b6 d1g4 b6e3 g4e6 f1d1 f8c8 c4e6 f7e6 g1f1 f6d5 f1e2 b7b5 e3d4 a7a5 a2a3 g8f7 g2g3 b5b4 a3b4 d5b4 d4e5 c8c1 d1c1 b4d5');
INSERT INTO evaluations VALUES(54512,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(54512,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(54512,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(54512,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(54512,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(54512,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(54512,36,'cp 39','b8c6 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 d1d4 b6d4 e3d4 f3d5 d4f6 d5c4 c1c4 g7f6 c4c7 b7b5 f1c1 f8d8 g1f1 a7a6 c1c2 g8g7 b2b3 d8d1 f1e2 d1h1 h2h3 g7g6 c7d7 h7h5 e2f3 a8e8 c2c7');
INSERT INTO evaluations VALUES(20514,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(20514,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(20514,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(20514,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(20514,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(20514,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(20514,36,'cp 34','d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 d1d4 b6d4 e3d4 f3d5 d4f6 d5c4 c1c4 g7f6 h2h3 f8d8 c4c7 b7b5 f1e1 d8d2 e1e7 a8f8 c7a7 d2b2 e7b7 g8g7 a2a4 b5b4 g1h2 b4b3 a4a5 f6f5 f2f3 f5f4 a5a6 b2a2 b7b3');
INSERT INTO evaluations VALUES(20489,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(20489,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(20489,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(20489,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(20489,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(20489,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(20489,36,'cp 45','d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 d1d4 b6d4 e3d4 f3d5 d4f6 d5c4 c1c4 g7f6 h2h3 f8d8 c4c7 b7b5 f1e1 d8d2 e1e7 a8f8 c7a7 d2b2 e7b7 g8g7 a2a4 b5b4 a4a5 h7h5 a5a6 g7g6 b7b6 b2a2 a7b7 f8d8');
INSERT INTO evaluations VALUES(44562,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(44562,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44562,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(44562,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(44562,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(44562,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44562,36,'cp 8','d7d6 d2d4 c5d4 f3d4 a7a6 f1e2 e7e5 d4b3 f8e7 c1e3 c8e6 e1g1 e8g8 d1d3 b8d7 a2a4 d8c7 a4a5 d7c5 b3c5 d6c5 e2f3 a8d8 d3e2 c7c6 h2h3 h7h6 f1b1 c6c7 e2e1 g8h8 b1d1 d8d1 e1d1 f8d8 d1e1 c7c6 a1d1 c6c7 d1d8 e7d8');
INSERT INTO evaluations VALUES(52432,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(52432,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52432,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(52432,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52432,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52432,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(52432,36,'cp 2','d7d5 c2c4 g8f6 c4d5 f6d5 d4c5 e7e6 a2a3 f8c5 g1f3 e8g8 f1c4 b8d7 e1g1 d5f6 d1e2 b7b6 b2b4 c5e7 f1d1 a7a5 b4a5 a8a5 b1c3 c8b7 c4b5 d8c8 c1b2 d7c5 a3a4 c8a8 a1c1 c5b3 c1b1 b3c5 d1c1 c5b3');
INSERT INTO evaluations VALUES(21911,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21911,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(21911,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(21911,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21911,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(21911,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(21911,36,'cp 24','c8g4 h2h3 g4f3 d1f3 c6d4 f3d3 d8h4 g2g3 h4e7 c2c3 d4c6 d3e2 e5f4 d2d4 g8f6 e1g1 e7e4 e2e4 f6e4 f1e1 f7f5 c1f4 c6e7 c4e6 g7g5 f4e3 e4g3 e3g5 g3e4 g5e7 f8e7 e6f5 e4g3 f5g4 h7h5 g4e6 e8d8 b1d2 h8f8 e1e3 h5h4 d2f3 c7c6 a1e1 e7f6 g1g2 d6d5 f3e5 d8c7 e5g6');
INSERT INTO evaluations VALUES(26983,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(26983,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(26983,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(26983,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(26983,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26983,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(26983,36,'cp 10','h2h3 g4f3 d1f3 c6d4 f3d3 g8f6 e1g1 a7a6 f4e5 d6e5 c2c3 d4e6 d3e2 f8c5 g1h1 d8d7 d2d3 e8g8 a2a4 b7b5 c4e6 d7e6 c1g5 f6d7 b1d2 h7h6 g5h4 c5e7 d2f3 c7c5 h4e7 e6e7 d3d4 e5d4 c3d4 c5d4 a4b5 d7c5 b5a6 d4d3 e2e3 a8a6 a1a6 c5a6');
INSERT INTO evaluations VALUES(26981,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(26981,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(26981,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(26981,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(26981,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(26981,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26981,36,'cp 13','g4f3 d1f3 c6d4 f3d3 g8f6 e1g1 a7a6 c2c3 d4e6 f4e5 d6e5 d3e2 f8c5 g1h1 e8g8 d2d3 d8d7 c4e6 d7e6 b1d2 a8d8 d2c4 b7b5 c4e3 e6d6 f1d1 c5e3 c1e3 f8e8 b2b4 h7h6 a2a3 d6c6 a1c1 f6d7 h1g1 d7f8 c3c4 f8e6 c4b5');
INSERT INTO evaluations VALUES(54973,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54973,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(54973,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(54973,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(54973,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54973,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54973,36,'cp 16','c7c5 g1f3 a7a6 d2d4 c5d4 f3d4 g8f6 c1e3 e7e5 d4b3 c8e6 d1f3 g7g6 h2h3 b7b5 g2g4 f8g7 g4g5 f6h5 e1c1 e8g8 c1b1 b8c6 h3h4 d8c7 c3d5 e6d5 d1d5 a8b8 c2c3 f7f5 e4f5 f8f5 f3e4 c6e7 f1h3 h5f4 h3f5 e7d5 e3f4 d5f4 f5g4');
INSERT INTO evaluations VALUES(26540,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26540,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(26540,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26540,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(26540,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(26540,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(26540,36,'cp 27','e7e6 a2a3 f8d6 f4g3 g8f6 c2c4 e8g8 b1c3 c6e7 a1c1 c7c6 f1e2 h7h6 f3d2 f5h7 g3d6 d8d6 e1g1 e7f5 b2b4 a7a5 d1b3 a5b4 a3b4 d6e7 b3b2 d5c4 d2c4 f5d6 c4b6 a8d8 b4b5 d6b5 c3b5 c6b5 e2b5');
INSERT INTO evaluations VALUES(26815,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(26815,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(26815,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(26815,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(26815,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(26815,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(26815,36,'cp 34','a2a3 g8f6 c2c4 f8d6 f4g3 c6e7 b1d2 e8g8 f1e2 b7b6 b2b4 c7c5 d4c5 b6c5 d1b3 f6e4 g3d6 d8d6 d2e4 f5e4 e1g1 c5b4 a3b4 d5c4 b3c4 e4d5 c4c3 f8c8 c3a3 d5f3 e2f3 a8b8 f1b1 c8c7 a3a4 e7d5 b4b5 d5c3');
INSERT INTO evaluations VALUES(26820,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26820,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(26820,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(26820,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(26820,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(26820,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(26820,36,'cp 40','f8d6 f4g3 g8f6 c2c4 c6e7 b1d2 e8g8 f1e2 b7b6 b2b4 a7a5 c4c5 d6g3 h2g3 a5b4 a3b4 d8b8 e1g1 a8a1 d1a1 b6c5 b4c5 b8b7 f3e5 f8a8 a1c3 e7c6 g3g4 f6e4 d2e4 f5e4 e5c6 b7c6 g4g5 c6a4 e2d3 a4a3 c3a3 a8a3 d3e4');
INSERT INTO evaluations VALUES(23851,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(23851,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23851,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(23851,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(23851,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(23851,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(23851,36,'cp 24','f4g3 g8f6 c2c4 c6e7 b1c3 c7c6 f1e2 e8g8 g3d6 d8d6 e1g1 h7h6 h2h3 d5c4 e2c4 e7d5 d1b3 d5c3 b3c3 c6c5 a1c1 c5d4 c3d4 d6e7 c4e2 f8d8 d4c3 f5e4 f1d1 d8d1 e2d1 b7b6 d1e2 a8d8 f3e5 e4b7');
INSERT INTO evaluations VALUES(50601,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(50601,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50601,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(50601,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50601,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50601,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(50601,36,'cp 15','c5d4 f1e2 b8c6 f3d4 e7e5 d4b3 g8f6 b1c3 f8b4 e1g1 b4c3 b2c3 e8g8 d1d3 d7d5 e4d5 d8d5 c3c4 d5d8 c1a3 f8e8 f1e1 c8g4 f2f3 g4h5 d3d8 e8d8 b3c5 e5e4 c5b7 d8d2 a3c1 d2c2 e2d1 c6d4 b7d6 e4f3 d1c2 d4c2');
INSERT INTO evaluations VALUES(50626,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(50626,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(50626,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(50626,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(50626,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(50626,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50626,36,'cp 29','f1e2 b8c6 f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 b1d2 e7e5 a2a4 a6a5 f2f3 f6d7 d2c4 f7f6 h2h4 b7b6 h4h5 d8c7 c4e3 d7c5 e2c4 h7h6 c1d2 c5b7 g2g4 f8b4 c2c3 b4c5 e3f5 b7d6 c4d3 c8f5 g4f5 a8d8 e1e2');
INSERT INTO evaluations VALUES(52774,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52774,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52774,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(52774,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(52774,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(52774,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52774,36,'cp 24','g1f3 c8f5 f1d3 e7e6 e1g1 f5g6 c2c4 g8f6 d1c2 b8d7 d3g6 h7g6 b1d2 c6c5 c4d5 f6d5 d4c5 f8c5 d2b3 c5b6 a2a4 d8c7 c2e2 g6g5 a4a5 g5g4 a5b6 g4f3 b6c7 f3e2 f1e1 d5c7 e1e2 d7e5 b3d4 a7a6 e3e4 e8c8 c1e3 e5d3');
INSERT INTO evaluations VALUES(52775,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(52775,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(52775,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(52775,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52775,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(52775,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(52775,36,'cp 12','c8f5 c2c4 e7e6 f1d3 f8b4 b1c3 f5d3 d1d3 g8f6 c1d2 b4a5 e1g1 e8g8 a2a3 b8d7 b2b4 a5c7 e3e4 f6e4 c3e4 d5e4 d3e4 a7a5 g2g3 h7h6 a1b1 a5b4 d2b4 d7f6 e4e2 c7d6 b4d6 d8d6 b1b7 c6c5 d4c5 d6c5 b7b5 c5a3 f1b1 f8c8 g1g2 a3e7');
INSERT INTO evaluations VALUES(41793,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41793,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(41793,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(41793,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41793,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(41793,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(41793,36,'cp 14','f4d6 c7d6 c2c4 b8c6 b1c3 c6e7 d1b3 e8g8 f1d3 c8d7 e1g1 d7c6 f3d2 d5c4 d2c4 b7b5 c4d2 a8b8 f1c1 a7a6 d2e4 f6e4 c3e4 d8b6 b3c2 c6e4 d3e4 b8c8 c2e2 g7g6 a2a3 b6a5 e4d3 c8c1 a1c1 f8c8 c1c8 e7c8');
INSERT INTO evaluations VALUES(41675,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(41675,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(41675,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41675,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(41675,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41675,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(41675,36,'cp 4','c7d6 c2c4 b8c6 b1c3 c6e7 f3d2 e8g8 f1e2 d5c4 d2c4 e7d5 d1b3 d5c3 b2c3 b7b6 c4d2 c8b7 e1g1 f6e4 d2e4 b7e4 f1c1 d8g5 e2f1 f8c8 a2a4 d6d5 a4a5 a8b8 b3b5 g5d8 a5b6 a7b6 c3c4 d5c4 f1c4 b8a8 a1a8 c8a8 b5e5 e4d5 c4d5 d8d5 e5d5 e6d5 f2f3 h7h6 c1c7 a8a1 g1f2');
INSERT INTO evaluations VALUES(40583,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40583,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(40583,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(40583,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(40583,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40583,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40583,36,'cp 14','c2c4 b8c6 c4d5 f6d5 f1c4 c6e7 e3e4 d5f6 c4d3 e8g8 e1g1 c8d7 b1c3 d8b6 d1e2 a8c8 a1d1 c8c7 g2g3 h7h6 f1e1 d7c6 a2a3 f8d8 h2h4 a7a6 d3b1 d8c8 g1g2 c6d7 e4e5 f6d5 c3d5 e7d5');
INSERT INTO evaluations VALUES(22775,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(22775,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(22775,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(22775,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(22775,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(22775,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(22775,36,'cp -37','f7f5 d2d4 e5d4 e4f5 f8b4 c1d2 b4c5 a2a3 d7d6 f1c4 d8e7 c4e2 c8f5 e1g1 g8f6 f1e1 c6e5 b3b4 c5b6 a3a4 a7a6 d2f4 e8g8 a4a5 b6a7 f4e5 d6e5 e2c4 g8h8 e1e5 e7d7 h2h3 h7h6 b1d2 c7c6 c4d3 f6d5 d3f5 f8f5 e5f5');
INSERT INTO evaluations VALUES(42480,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(42480,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42480,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(42480,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(42480,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(42480,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(42480,36,'cp -2','c7c5 e2e3 c5d4 e3d4 e8g8 f1d3 b8c6 g1f3 d7d5 c4c5 b7b6 a2a3 b4c3 d2c3 f6e4 d1c2 b6c5 d3e4 d5e4 c2e4 c5d4 f3d4 c6d4 e4d4 d8d4 c3d4 f8d8 d4c3 c8a6 a1d1 d8d1 e1d1 f7f6 f2f3 a8d8 d1c2 a6c4 b2b3 c4d5 c2b2 e6e5 h1c1 d8c8');
INSERT INTO evaluations VALUES(39922,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39922,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39922,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39922,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39922,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39922,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(39922,36,'cp 2','g1f3 c5d4 f3d4 e8g8 e2e3 b8c6 a2a3 b4e7 f1e2 d7d5 c4d5 c6d4 e3d4 f6d5 e1g1 d5b6 d2f4 c8d7 e2f3 d7c6 f3c6 b7c6 a1c1 a8c8 f1e1 e7f6 f4e3 b6c4 c3e4 c4b2 e4f6 d8f6 d1c2 c8b8 g2g3 f6f3 c2c6 f3c6 c1c6');
INSERT INTO evaluations VALUES(39918,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(39918,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(39918,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(39918,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(39918,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(39918,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39918,36,'cp -5','c5d4 f3d4 e8g8 a2a3 b4e7 e2e3 b8c6 f1e2 d7d5 c4d5 c6d4 e3d4 f6d5 e1g1 d5c3 d2c3 e7f6 e2f3 a8b8 a1c1 b7b6 d4d5 f6c3 c1c3 e6d5 f3d5 c8b7 d5f3 d8f6 f3b7 b8b7 d1f3 f6f3 c3f3 b7d7 f3c3 d7d2 b2b4 a7a5 g2g3 f8d8 f1c1');
INSERT INTO evaluations VALUES(50083,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(50083,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(50083,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(50083,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(50083,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(50083,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(50083,36,'cp 50','e7e6 a2a3 g7g6 g1f3 f8g7 b2b4 a7a5 c1b2 b7b6 e2e3 e8g8 f1e2 d8e7 e1g1 d7d6 d1b3 b8d7 a1c1 c7c5 f1d1 a5b4 a3b4 c8b7 c1a1 f6e4 c3e4 b7e4 f3d2 e4c6 e2f3 c6f3 d2f3 g6g5 d4c5 a8a1 d1a1 g7b2 b3b2 d6c5 b4c5 b6c5');
INSERT INTO evaluations VALUES(49935,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(49935,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(49935,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(49935,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(49935,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(49935,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(49935,36,'cp 52','a2a3 g7g6 g1f3 d7d6 b2b4 f8g7 e2e3 b7b6 f1e2 c8b7 c1b2 e8g8 d1b3 d8e7 e1g1 b8d7 a3a4 f5f4 b2c1 f4e3 c1e3 f6h5 a4a5 h5f4 e3f4 f8f4 c3b5 a8f8 a5b6 a7b6 a1a7 b7f3 e2f3 f4f3 g2f3 c7c6 b5c3 g7d4');
INSERT INTO evaluations VALUES(49943,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(49943,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(49943,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(49943,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(49943,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(49943,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(49943,36,'cp 54','g7g6 g1f3 f8g7 b2b4 d7d6 e2e3 e8g8 f1e2 d8e7 c1b2 b7b6 e1g1 c8b7 d4d5 f6e4 f3d4 c7c5 d5c6 b8c6 c3e4 c6d4 b2d4 b7e4 d4g7 g8g7 d1b3 e4b7 a3a4 a7a5 f1d1 b7c6 b4a5 b6a5 b3a3 f8d8 a1b1 d8b8 b1b8 a8b8 a3c3 e6e5 c3a5');
INSERT INTO evaluations VALUES(55964,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(55964,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(55964,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(55964,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(55964,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(55964,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(55964,36,'cp -33','c7c5 g1f3 b8c6 c2c3 g8e7 f1d3 f7f5 e1g1 d5d4 d3c4 c6a5 c4b5 c8d7 b5e2 e7d5 d2d3 f8e7 c3c4 d5c7 c1d2 a5c6 g2g3 e8g8 b1a3 a7a6 h2h4 b7b5 f3e1 b5b4 a3c2 a6a5 e2f3 a5a4 d1e2 h7h6 b2b3 d7e8 h4h5 d8d7');
INSERT INTO evaluations VALUES(54305,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(54305,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(54305,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(54305,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(54305,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(54305,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(54305,36,'cp -30','c2c3 b8c6 g1f3 g8e7 f1d3 f7f5 e1g1 a7a6 d3e2 d5d4 d2d3 b7b6 b1a3 e7d5 c3c4 d5c7 a3c2 c8b7 f3e1 d8d7 e2f3 f8e7 c1d2 b6b5 d1e2 e8g8 b2b3 a8b8 g2g3 c6a7 e2g2 b7f3 e1f3 a7c6 a1c1 b5c4 b3c4');
INSERT INTO evaluations VALUES(54302,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(54302,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(54302,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(54302,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(54302,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(54302,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(54302,36,'cp -23','g8e7 f1d3 b8c6 g1f3 f7f5 e1g1 b7b6 b1a3 a7a6 a3c2 d5d4 c3d4 d8d7 a2a4 c8b7 d3c4 c6d4 c2d4 c5d4 d2d3 e7d5 c1d2 a8c8 g1h1 b6b5 a4b5 a6b5 c4d5 b7d5 f3d4 f8c5 d4f3 e8g8 d1e2 h7h6 d2e3 c5e7 f1c1 c8c1 a1c1');
INSERT INTO evaluations VALUES(56378,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(56378,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(56378,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(56378,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(56378,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(56378,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(56378,36,'cp 200','d4e5 g8e7 f2f4 b8c6 b1c3 d5e4 d1d8 c6d8 c1d2 c8f5 e1c1 d8e6 d1e1 e4e3 d2e3 e8c8 g1f3 e7g6 g2g3 f8b4 e3d2 h7h5 h2h3 g6e7 a2a3 b4c5 h1h2 a7a6 d2e3 c5e3 e1e3 c7c6 h2d2 d8d2 f3d2 g7g5 f4g5 e6g5 h3h4 g5e6 d2c4 c8c7 c4d6 f5g6 f1h3 h8d8 h3e6 f7e6');
INSERT INTO evaluations VALUES(56309,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(56309,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(56309,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(56309,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(56309,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(56309,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(56309,36,'cp 189','g8e7 f2f4 b8c6 b1c3 d5e4 d1d8 c6d8 c3e4 e7d5 c1d2 d5b4 d2b4 f8b4 c2c3 b4e7 g1f3 d8e6 f4f5 e6c5 e4g3 c8d7 e1c1 e8c8 f1c4 d7c6 c1c2 f7f6 e5e6 d8d1 h1d1 c5e4 c4d5 e4g3 h2g3 c6e8 d1h1 c7c6 d5e4 h7h5 f3d4 c8c7 g3g4');
INSERT INTO evaluations VALUES(53635,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(53635,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(53635,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(53635,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(53635,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(53635,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(53635,36,'cp 59','c6c5 c2c3 b8c6 f1d3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f8e7 b1c3 e8g8 d3f5 e6f5 d1b3 g7g5 c3d5 f5f4 e3d2 c8e6 d5e7 c6e7 b3b7 g5g4 f3e1 e7f5 d2c3 e6d5 b7a6 a8c8 a6d3 d8g5 d3d2 f7f6 a2a3 f4f3 d2g5 f6g5 g2g3 d5c6 e1d3 c6b5');
INSERT INTO evaluations VALUES(54298,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(54298,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(54298,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(54298,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(54298,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54298,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(54298,36,'cp 59','c2c3 b8c6 f1d3 g8e7 b1a3 c5d4 a3b5 e7f5 d3f5 e6f5 b5d4 f8c5 c1e3 c5d4 e3d4 e8g8 d1d2 f8e8 e1g1 h7h6 h2h4 c8e6 f1e1 a7a6 d2f4 a8c8 e1d1 b7b5 a1c1 d8e7 f4e3 b5b4 c3b4 c6b4 a2a3 b4c2 e3d2 c2d4 f3d4');
INSERT INTO evaluations VALUES(54290,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(54290,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(54290,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(54290,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(54290,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(54290,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(54290,36,'cp 77','b8c6 f1d3 g8e7 h2h4 c5d4 c3d4 e7f5 d3f5 e6f5 b1c3 f8b4 e1f1 b4c3 b2c3 c8e6 c1f4 h7h6 h1h3 c6e7 f3e1 e8g8 e1d3 g8h7 h4h5 e7c6 f4c1 c6a5 d3f4 d8d7 f1g1 a5c4 a2a4 f8b8 a4a5 b7b6 a5a6 b8c8 h3g3 h7h8 g1h2 c8c7 d1e2 b6b5 f4d3');
INSERT INTO evaluations VALUES(20333,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(20333,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(20333,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(20333,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(20333,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(20333,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(20333,36,'cp 62','f1d3 g8e7 b1a3 c5d4 a3b5 e7f5 d3f5 e6f5 b5d4 f8c5 c1e3 c5d4 e3d4 e8g8 e1g1 f8e8 d1d2 h7h6 h2h4 c8e6 d2f4 a7a5 h4h5 a5a4 a1d1 c6a5 f4g3 a5c4 b2b3 c4a5 d1b1 d8c7 f1d1 a4b3 a2b3 a5c6 d1e1 e8c8 d4e3 g8h8 f3h4');
INSERT INTO evaluations VALUES(20336,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(20336,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(20336,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(20336,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(20336,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(20336,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(20336,36,'cp 71','g8e7 h2h4 c5d4 c3d4 e7f5 d3f5 e6f5 b1c3 f8b4 c1f4 d8a5 f4d2 e8g8 e1g1 b4c3 d2c3 a5d8 d1d2 c8e6 f1c1 h7h6 d2f4 a7a5 a2a3 a5a4 f3e1 g8h7 e1d3 f8g8 f4g3 g7g5 h4g5 d8g5 g3g5 h6g5 f2f4 a8d8 f4g5');
INSERT INTO evaluations VALUES(53132,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53132,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53132,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53132,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53132,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53132,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53132,36,'cp 0','g8f6 f1c4 e7e6 e1g1 f8e7 e4e5 d6e5 d4e5 b8d7 e5g3 e8g8 b1c3 a7a6 f1d1 b7b5 c4d3 c8b7 c1h6 f6h5 g3h3 h5f6');
INSERT INTO evaluations VALUES(45258,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45258,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(45258,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45258,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45258,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(45258,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(45258,36,'cp 17','f1e2 g7g6 e1g1 f8g7 d4e3 b8c6 b1c3 e8g8 c3d5 f6g4 e3a3 g4f6 d5f6 g7f6 f1d1 c8g4 h2h3 g4f3 a3f3 c6d4 f3d3 d8c7 c2c3 d4e2 d3e2 f8c8 a2a4 a7a6 a4a5 c7c4 e2c2 a8b8 c1e3 c4e6 f2f4 e6c4 e3d4 f6d4 d1d4');
INSERT INTO evaluations VALUES(12822,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12822,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(12822,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12822,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12822,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(12822,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12822,36,'cp 41','f1b5 d8c7 d2d3 d7d6 e1g1 e7e6 c1g5 f8e7 e4e5 d6e5 b5c6 c7c6 f3e5 c6d6 e5c4 d6c7 d1f3 e8g8 f1e1 c8d7 c4e5 d7e8 g5f4 c7c8 h2h3 e8c6 e5c6 c8c6 f3c6 b7c6 f4g5 h7h6 g5h4 g7g5 h4g3 f6d5 c3e4 f8d8 g3e5 d5b4 e1e2');
INSERT INTO evaluations VALUES(12761,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(12761,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(12761,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(12761,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(12761,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(12761,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(12761,36,'cp 36','d8c7 e1g1 c6d4 f1e1 a7a6 b5c4 e7e6 e4e5 f6g8 d2d3 b7b5 c4b3 g8e7 c3e4 e7f5 f3d4 c5d4 c1f4 c8b7 a1c1 f8e7 c2c4 b7c6 b3c2 e8g8 d1g4 c7b7 b2b3 a6a5 e4d6 e7d6 e5d6');
INSERT INTO evaluations VALUES(1142,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(1142,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(1142,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(1142,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(1142,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(1142,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(1142,36,'cp 59','e1g1 c6d4 f1e1 a7a6 b5f1 e7e5 c3d5 f6d5 e4d5 f8d6 b2b3 e8g8 c1b2 b7b5 c2c3 d4f3 d1f3 f7f5 c3c4 e5e4 f3h3 d6e5 b2e5 c7e5 h3g3 e5g3 h2g3 b5c4 b3c4 a8b8 a1b1 b8b1 e1b1 g8f7 f2f3 d7d6 b1b6 f7e7 b6b8 h7h5 b8a8 e4f3 g2f3');
INSERT INTO evaluations VALUES(56249,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(56249,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(56249,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(56249,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(56249,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(56249,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(56249,36,'cp -27','e4d5 g8f6 c2c4 c7c6 d2d4 f8b4 b1c3 c6d5 f1e2 b8c6 c1f4 d5c4 e2c4 d8e7 f4e5 c6e5 f3e5 e8g8 e1g1 b4d6 f1e1 d6e5 e1e5 e7c7 c4b3 f6g4 c3d5 c7d8 e5e1 d8d6 g2g3 g8h8 d5e3 g4e3 e1e3 c8d7 d1d3 h8g8 a1f1 h7h6 e3f3 d7e6 b3e6');
INSERT INTO evaluations VALUES(47188,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(47188,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(47188,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(47188,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(47188,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(47188,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(47188,36,'cp -12','c7c5 d4c5 e7e6 g1f3 f8c5 e1g1 e8g8 c2c4 d5c4 d3c4 a7a6 d1d8 f8d8 c1d2 b7b6 f1c1 c8b7 d2e1 b8d7 b1d2 d8c8 c1c2 c8c7 a1c1 a8c8 c4d3 b6b5 a2a3 c5d6 c2c7 c8c7 c1c7 d6c7 d3e2 e6e5 b2b4 e5e4 f3d4 c7d6');
INSERT INTO evaluations VALUES(45381,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(45381,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(45381,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(45381,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(45381,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(45381,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45381,36,'cp -25','d4c5 e7e6 g1f3 f8c5 b2b3 b8c6 c1b2 d8e7 c2c4 d5c4 d3c4 e6e5 b1d2 e8g8 d1b1 f8d8 a2a3 c8g4 e1g1 g4h5 d2e4 h5g6 e4f6 e7f6 b1a2 f6e7 b3b4 c5d6 g2g3 g6h5 c4d5 e5e4 f3d2 d6e5 b2e5 e7e5 d5c6 b7c6');
INSERT INTO evaluations VALUES(32078,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(32078,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(32078,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(32078,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(32078,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(32078,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(32078,36,'cp 196','d1f3 d6e5 f3b3 b8d7 b3b7 g8f6 b7c6 f8d6 b1d2 e8g8 d2c4 f8e8 c1e3 d6b4 c2c3 e8e6 c6a4 b4c5 a4c2 a7a5 f1e2 c5e3 c4e3 d8e7 e2c4 e6b6 e1g1 g7g6 b2b3 g8g7 a1d1 h7h5 h2h3 b6d6 d1d6 e7d6 f1d1 d6c5 c2d2 c5e7 b3b4 d7b6 c4d5 a8d8');
INSERT INTO evaluations VALUES(32077,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(32077,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(32077,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(32077,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(32077,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(32077,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(32077,36,'cp 175','d6e5 f3b3 b8d7 b3b7 g8f6 b7c6 f8d6 b1d2 e8g8 d2c4 f8e8 c1e3 d6b4 c2c3 e8e6 c6a4 b4c5 a4c2 d8e7 b2b4 c5e3 c4e3 a7a5 a2a3 d7b6 f1e2 g7g6 e1g1 e6d6 e2b5 g8g7 f1b1 h7h5 h2h3 h5h4 b1d1 d6d1 e3d1 e7d6 d1e3 d6e6 b5f1 e6c6 f1d3 f6h5 a1d1 a5b4 c3b4 c6c2 e3c2');
INSERT INTO evaluations VALUES(32419,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(32419,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(32419,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(32419,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(32419,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(32419,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(32419,36,'cp 199','f3b3 b8d7 b3b7 f8c5 b7b3 g8f6 b3f3 e8g8 a2a4 a7a5 f1b5 f6e8 f3h3 d7f6 b1d2 e8d6 b5d3 d8e7 e1g1 h7h6 h3f3 f6d7 f3e2 c5b6 c2c3 f8d8 d3c2 d7c5 g1h1 e7f6 b2b3 a8c8 c1a3 c5e6 g2g3 e6c5 a3c5 b6c5');
INSERT INTO evaluations VALUES(32416,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(32416,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(32416,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(32416,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(32416,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(32416,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(32416,36,'cp 184','b8d7 b3b7 g8f6 b7c6 f8d6 b1d2 e8g8 d2c4 f8e8 c1e3 d6b4 c2c3 e8e6 c6a4 b4c5 e1c1 d8e7 f2f3 c5e3 c4e3 e7c5 e3c4 d7b6 a4b4 c5f2 b4b3 f2c5 c1b1 a7a5 d1d2 e6c6 c4a3 c6d6 d2d6 c5d6 f1e2 d6c5 h1d1 c5f2 e2f1 g7g6 b3c2 f2c5 h2h3 f6h5 b1a1 h5f4 f1c4');
INSERT INTO evaluations VALUES(22130,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(22130,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(22130,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(22130,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(22130,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(22130,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(22130,36,'cp 71','g8f6 b1c3 f8e7 c1f4 e8g8 e1c1 c8e6 c1b1 f6d7 c4e2 e7f6 d4e3 d8e7 h2h4 f6e5 g2g3 d7b6 b2b3 a8d8 a2a4 f7f6 b1b2 a7a5 f4e5 d6e5 f2f4 e5f4 g3f4 g8h8 d1e1 e6g8 h4h5 h7h6 e2b5 b6c8 e4e5 f6e5 f4e5');
INSERT INTO evaluations VALUES(15757,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(15757,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(15757,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(15757,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(15757,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(15757,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(15757,36,'cp 77','c1f4 f8e7 b1c3 e8g8 e1c1 c8e6 c1b1 f6d7 c4e2 e7f6 d4e3 d8e7 h2h4 f6e5 g2g3 d7b6 b2b3 a8d8 f4e5 d6e5 f2f4 e5f4 g3f4 d8d1 c3d1 g8h8 a2a4 f7f6 d1c3 e7b4 b1b2 f8d8 h4h5 h7h6 h1g1 e6c4 a4a5 b6a4 c3a4 c4e2 a4c5 b4d4 e3d4 d8d4 b2c3 d4d1 g1d1 e2d1');
INSERT INTO evaluations VALUES(57646,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(57646,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(57646,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(57646,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(57646,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(57646,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(57646,36,'cp 12','c7c5 e2e3 f8g7 d2d4 c5d4 e3d4 g8f6 d4d5 e8g8 f1e2 e7e6 g1f3 e6d5 c4d5 d7d6 c1f4 b8d7 e1g1 d7c5 f3d2 f8e8 d2c4 f6e4 c3b5 g7f8 a2a4 c8d7 f4e3 d7b5 a4b5 a7a5 b5a6 a8a6 a1a6 c5a6 d1b3 a6c5 b3b5 d8d7 b5d7 c5d7 f1b1 f8g7 g2g3');
INSERT INTO evaluations VALUES(52283,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(52283,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52283,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(52283,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(52283,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(52283,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52283,36,'cp 18','e2e3 f8g7 d2d4 c5d4 e3d4 g8f6 d4d5 e8g8 g1f3 e7e6 f1e2 e6d5 c1f4 d7d6 e1g1 f8e8 c4d5 a7a6 f1e1 b8d7 f4d6 d7b6 d6e5 f6d5 c3d5 d8d5 d1d5 b6d5 e2c4 c8e6 f3g5 d5c7 g5e6 c7e6 e5g7 g8g7 c4e6');
INSERT INTO evaluations VALUES(52280,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(52280,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52280,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(52280,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52280,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52280,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52280,36,'cp 11','g8f6 d2d4 c5d4 e3d4 f8g7 d4d5 e8g8 f1e2 e7e6 c1g5 d7d6 g1f3 h7h6 g5e3 e6d5 c4d5 b8d7 a1c1 d7c5 e1g1 f8e8 f1e1 a7a6 h2h3 c8d7 e2f1 a8c8 d1d2 g8h7 e3d4 c5e4 c3e4 f6e4 d2e3 c8c1 e1c1 h7g8 d4g7 g8g7');
INSERT INTO evaluations VALUES(31180,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(31180,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(31180,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(31180,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(31180,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(31180,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(31180,36,'cp 42','d2d4 e7e6 f1e2 c6c5 c1e3 c5d4 f3d4 g8e7 e1g1 b8c6 e2b5 a7a6 b5c6 b7c6 c2c4 d5c4 d1a4 d8d7 a4c4 c6c5 d4f5 e7f5 b1c3 d7c6 a1c1 f8e7 c4a4 c6a4 c3a4 f5e3 f2e3 e8g8 f1d1 f8d8 g1f1 g7g5 h2h3 h7h6 d1d8 a8d8 a4c5');
INSERT INTO evaluations VALUES(57179,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(57179,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(57179,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(57179,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(57179,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(57179,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(57179,36,'cp -87','f1g2 b8c6 c2c4 d7d6 h2h3 h7h5 g4h5 d8g5 g2f3 c6d4 b1c3 g8e7 h3h4 d4f3 g1f3 g5h5 c3b5 e5e4 f3d4 e8d8 d2d3 h5g4 d3e4 g4g2 e1d2 g2e4 f2f3 e4g6 h1g1 g6h7 e2e4 h7h4 d2c2 h4h2 c2b3 h8h3 b5c3 c8d7 c1e3 e7c6');
INSERT INTO evaluations VALUES(57180,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(57180,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(57180,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(57180,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(57180,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(57180,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(57180,36,'cp -89','b8c6 c2c4 h7h5 g4h5 d8h4 d2d3 b7b6 b1c3 c8b7 g1f3 h4h5 a2a3 e8c8 h2h4 d8e8 e2e4 f7f5 c1e3 c8b8 f3g5 f5f4 d1h5 h8h5 e3d2 c6d4 a1d1 f8e7 g5f3 d4f3 g2f3 h5h4 h1h4 e7h4 e1e2 g8e7 c3d5 e7d5 c4d5 c7c6 d2c3 c6d5');
INSERT INTO evaluations VALUES(22839,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(22839,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(22839,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(22839,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(22839,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(22839,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(22839,36,'cp -98','c2c4 h7h5 g4h5 d8h4 d2d3 b7b6 b1c3 c8b7 a2a3 e8c8 g1f3 h4h5 h2h4 c8b8 e2e4 f7f5 f3g5 c6d4 c1e3 h5e8 d1a4 g8f6 e1c1 b7c6 c3b5 d4b5 c4b5 c6b7 f2f3 f6h5 c1b1 h5g3 h1h3 g3e2 a4c2 e2f4 e3f4 e5f4 d3d4 f8e7 d4d5');
INSERT INTO evaluations VALUES(22727,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(22727,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(22727,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(22727,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(22727,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(22727,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(22727,36,'cp -118','h7h5 g4h5 d8h4 d2d3 b7b6 b1c3 c8b7 c1d2 e8c8 d1a4 a7a5 a2a3 g8f6 g1f3 h4h5 f3g5 c6d4 g2b7 c8b7 h2h4 f8e7 e1c1 f6g4 d1f1 f7f6 g5e4 f6f5 e4g5 f5f4 g5e4 g4f6 e2e3 f6e4 c3e4 d4e6 a4c2 d7d5 c4d5 d8d5 e4c3 d5d8 c3e4 f4e3 d2e3 e6d4 e3d4');
INSERT INTO evaluations VALUES(22707,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(22707,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22707,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(22707,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(22707,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(22707,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(22707,36,'cp 8','f8c5 g1f3 g8f6 b1c3 a7a6 c3d5 f6d5 c4d5 d7d6 e1g1 e8g8 c2c3 d8f6 b2b4 c5a7 d5b3 h7h6 c1e3 c6e7 d3d4 c8g4 a1c1 e5d4 c3d4 c7c6 e4e5 d6e5 d4e5 f6g6 b3c2 g6h5 d1d3 e7g6 e3a7');
INSERT INTO evaluations VALUES(6434,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(6434,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(6434,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(6434,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6434,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(6434,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6434,36,'cp 15','g1f3 a7a6 e1g1 g8f6 c2c3 d7d6 b1d2 c8e6 h2h3 e8g8 f1e1 h7h6 c4b3 e6b3 d2b3 c5a7 c1e3 a7e3 e1e3 f8e8 b3d2 c6e7 d3d4 e7g6 a2a4 e5d4 c3d4 c7c5 d4c5 d6c5 d1b3 g6f4 b3b7 a8b8 b7a6 b8b2 d2c4 b2e2 c4d6 f6e4');
INSERT INTO evaluations VALUES(57216,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(57216,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(57216,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(57216,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(57216,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(57216,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(57216,36,'cp -32','d7d5 d3d4 e5d4 e3d4 g8f6 g1f3 f8d6 c2c4 c7c6 f1e2 e8g8 e1g1 d5c4 e2c4 c8f5 b1c3 b8d7 c1g5 h7h6 g5h4 d8b6 d1d2 f6e4 c3e4 f5e4 h4g3 d6g3 h2g3 d7f6 a1d1 e4d5 f3e5 a8d8 f1e1 d5c4 e5c4 b6a6 d2c2 f6d5');
INSERT INTO evaluations VALUES(56466,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(56466,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(56466,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(56466,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(56466,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(56466,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(56466,36,'cp -34','d3d4 b8c6 g1f3 e5e4 f3d2 f7f5 c2c4 g8f6 f1e2 c8e6 e1g1 f8d6 b1c3 d5c4 f2f4 e4f3 e2f3 d6b4 d1a4 a8b8 f3c6 b7c6 d2c4 e8g8 c4e5 f6g4 e5g4 f5g4 f1f8 b4f8 a4c6 f8d6 c1d2 b8b2 c3e4 h7h6 d2e1 b2b6 c6c3 e6f5 e4d6 c7d6 e1g3 d8e7');
INSERT INTO evaluations VALUES(56366,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(56366,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(56366,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(56366,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(56366,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(56366,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(56366,36,'cp -28','b8c6 g1f3 e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 c8e6 a2a3 a7a6 c4d5 f6d5 f1c4 d8d7 e1g1 f8e7 c3d5 e6d5 d1c2 h7h5 b2b4 e7f6 c4d5 d7d5 c1b2 c6e7 a3a4 c7c6 b4b5 c6b5 a4b5 a8c8 c2b3 d5b3 d2b3 a6b5 b3c5');
INSERT INTO evaluations VALUES(47034,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(47034,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(47034,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(47034,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(47034,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47034,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(47034,36,'cp 71','d2d4 e7e5 g1f3 b8d7 f1c4 f8e7 e1g1 e8g8 h2h3 e5d4 d1d4 d7c5 c1e3 c8e6 a1d1 c7c6 f1e1 e6c4 d4c4 b7b5 c4e2 c5e4 c3e4 f6e4 c2c4 e7f6 c4b5 c6b5 e2b5 f8e8 e3d4 f6d4 d1d4 e4f6 e1e8 d8e8 b5a6 d6d5 b2b3 e8e7 b3b4 g7g6 b4b5 a8e8 a2a4');
INSERT INTO evaluations VALUES(57227,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(57227,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(57227,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(57227,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(57227,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(57227,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(57227,36,'cp 169','f3e5 d7d5 d2d4 g8f6 g2g3 c7c5 f1g2 c5d4 d1d4 b8c6 e5c6 b7c6 c2c4 f8e7 e1g1 e8g8 b1c3 c8e6 c4d5 f6d5 c3d5 c6d5 c1f4 e7f6 f4e5 f6e5 d4e5 a8c8 b2b3 d8a5 f1d1 c8c2 g2d5 a5b5 a2a4 b5e2 e5e2 c2e2 b3b4 e2b2 b4b5 e6d5');
INSERT INTO evaluations VALUES(57102,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(57102,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(57102,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(57102,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(57102,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(57102,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(57102,36,'cp 165','g8f6 d2d4 d7d5 g2g3 c7c5 f1g2 c5d4 d1d4 f8e7 e1g1 e8g8 c2c4 b8c6 e5c6 b7c6 b1c3 c8e6 c1f4 d5c4 d4d8 a8d8 g2c6 f6d5 c6d5 e6d5 f4c7 d8d7 a1d1 d7c7 c3d5 c7b7 d5e7 b7e7 d1d2 f8c8 f1c1 g7g6 c1c3 a7a5 f2f3 e7b7');
INSERT INTO evaluations VALUES(49495,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(49495,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(49495,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(49495,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(49495,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(49495,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(49495,36,'cp 179','d2d4 f8e7 c1g5 d7d5 e2e3 h7h6 g5h4 b8d7 e5d7 c8d7 b1c3 c7c6 f1d3 e8g8 e1g1 f8e8 c3e2 f6e4 h4e7 d8e7 b2b3 a8d8 d3e4 e7e4 e2g3 e4g6 c2c4 d5c4 b3c4 b7b5 c4b5 c6b5 a1c1 d8c8');
INSERT INTO evaluations VALUES(10888,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(10888,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(10888,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(10888,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(10888,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(10888,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(10888,36,'cp 28','d2d3 e8g8 e1g1 d7d6 a2a4 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 c1e3 d6d5 c4a2 d5e4 c3e4 f6d5 e3d4 e5d4 d1f3 d8d7 a1e1 e7b4 e1e2 a8e8 a2b3 a7a5 h3h4 e8e7 e4g3 g7g6 b3d5 d7d5 f3d5 c6d5 e2e7 b4e7 h4h5 f7f5');
INSERT INTO evaluations VALUES(10889,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(10889,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(10889,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(10889,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(10889,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(10889,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(10889,36,'cp 20','d7d6 a2a4 c6b4 h2h3 c8e6 e1g1 e8g8 a4a5 d8d7 f1e1 h7h6 c1d2 e6c4 d3c4 f8e8 c3d5 b4d5 c4d5 b7b5 b2b3 c7c6 c2c4 a8b8 d1e2 b5c4 b3c4 e8c8 e1c1 a7a6 e2d3 c6d5 c4d5 c8c1 a1c1 d7b5 d3c4 b5c4');
INSERT INTO evaluations VALUES(8585,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(8585,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(8585,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(8585,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8585,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(8585,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(8585,36,'cp 25','a2a4 c6b4 h2h3 c8e6 a4a5 d8d7 e1g1 e8g8 c1d2 e6c4 d3c4 c7c6 c3e2 b4a6 e2g3 a6c5 d1e2 c5e6 f1d1 d7c7 g3f5 f8e8 f3h2 g7g6 f5e7 c7e7 b2b4 h7h5 d1e1 e6f4 e2f3 f4e6 a1d1 b7b6 h2f1 a8d8 a5b6 a7b6');
INSERT INTO evaluations VALUES(52431,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(52431,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52431,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52431,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52431,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52431,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52431,36,'cp -1','g8f6 e2e3 e7e6 b1d2 d7d5 g1f3 b8d7 f1d3 b7b6 e1g1 c8b7 d1e2 f8e7 b2b3 e8g8 f1d1 a8c8 c1b2 c8c7 a2a4 c5d4 c3d4 e7b4 a1c1 c7c1 d1c1 d8a8 f3e5 d7e5 d4e5 f6e4 d2f3 e4c5 d3c2 a7a5');
INSERT INTO evaluations VALUES(44550,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(44550,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44550,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(44550,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(44550,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(44550,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(44550,36,'cp 12','g1f3 e7e6 c1f4 f8e7 e2e3 f6h5 d4c5 h5f4 e3f4 e7c5 f1d3 d7d5 e1g1 b8c6 b1d2 e8g8 d2b3 c5e7 g2g3 d8c7 f1e1 g7g6 g1g2 c8d7 d1d2 a8c8 a1d1 e7f6 d3b5 h7h5 b5c6 d7c6 b3d4 g8g7 d2c2 f8h8 h2h4');
INSERT INTO evaluations VALUES(44549,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44549,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(44549,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44549,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(44549,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44549,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44549,36,'cp 0','d7d5 d4c5 e7e6 c1e3 d8c7 b2b4 a7a5 a2a3 f8e7 h2h3 b8d7 b1d2 e8g8 f3d4 f8d8 d4b5 c7c6 b5d4');
INSERT INTO evaluations VALUES(45378,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45378,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45378,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45378,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45378,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45378,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45378,36,'cp 0','d4c5 e7e6 c1e3 d8c7 b2b4 a7a5 a2a3 f8e7 h2h3 b8d7 b1d2 e8g8 f3d4 f8d8 d4b5 c7c6 b5d4');
INSERT INTO evaluations VALUES(19077,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19077,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19077,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19077,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19077,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19077,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19077,36,'cp 0','g8f6 f1d3 f8c5 a2a3 e8g8 c2c4 d7d6 h2h3 c8e6 e1g1 c6d4 b1c3 a8c8 c1d2 d4f3 d1f3 c5d4 f3e2 f6d7 a1c1 d7c5 d3c2 c5d7');
INSERT INTO evaluations VALUES(12603,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12603,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12603,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12603,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12603,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12603,36,'cp 0','f1d3 f8c5 a2a3 e8g8 c2c4 d7d6 h2h3 c8e6 e1g1 c6d4 b1c3 a8c8 c1d2 d4f3 d1f3 c5d4 f3e2 f6d7 a1c1 d7c5 d3c2 c5d7');
INSERT INTO evaluations VALUES(12599,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12599,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12599,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12599,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12599,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12599,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12599,36,'cp 0','f8c5 a2a3 e8g8 c2c4 d7d6 h2h3 c8e6 b1c3 c6d4 e1g1 a8c8 c1d2 f6d7 c3d5 d4f3 d1f3 c5d4 a1b1 d7c5 d3c2 c5d7');
INSERT INTO evaluations VALUES(7856,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7856,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7856,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7856,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7856,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7856,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7856,36,'cp 0','a2a3 e8g8 c2c4 d7d6 h2h3 c8e6 b1c3 c6d4 e1g1 a8c8 c1d2 f6d7 c3d5 d4f3 d1f3 c5d4 a1b1 d7c5 d3c2 c5d7');
INSERT INTO evaluations VALUES(47284,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(47284,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(47284,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(47284,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(47284,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47284,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47284,36,'cp 26','f1g2 g7g6 c2c4 d5c4 e1g1 f8g7 d1a4 f6d7 b1c3 e8g8 a4c4 d7b6 c4h4 e7e5 d2d3 d8h4 f3h4 c7c6 c1d2 c8e6 b2b4 b8d7 h4f3 h7h6 b4b5 c6b5 c3b5 e5e4 d3e4 g7a1 f1a1 h6h5 f3d4 f8c8 d4e6 f7e6 e4e5');
INSERT INTO evaluations VALUES(22815,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22815,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22815,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22815,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22815,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22815,36,'cp -8','f1c4 d8f6 c2c3 d7d6 d2d3 c8e6 b1d2 g8e7 b2b4 d6d5 e4d5 f6f3 g1f3 e7d5 d2e4 f7f6 e1g1 e8c8 c1d2 e6g8 f1e1 d5b6 c4g8 h8g8 b4b5 c6e7 d3d4 e5d4 f3d4 e7d5 g1f1 b6c4 a1d1 c4b2 d1a1 d5b6 d2e3 b6c4 a2a4 b2d3 e1d1');
INSERT INTO evaluations VALUES(54228,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(54228,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(54228,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(54228,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(54228,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(54228,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(54228,36,'cp 61','c5d4 f3d4 d5e4 c3e4 a7a6 f1d3 f8e7 e1g1 g8f6 f1e1 e8g8 c2c3 b8d7 d1f3 f8e8 d3c2 f6e4 f3e4 d7f6 e4h4 g7g6 c2b3 c8d7 c1h6 f6h5 h4e4 d8c7 e4f3 e7f8 h6g5 f8d6 g2g3 a8c8 a1d1 b7b5 d1d2 b5b4 e1d1 b4c3 b2c3 c7c3 f3c3');
INSERT INTO evaluations VALUES(54394,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(54394,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(54394,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(54394,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(54394,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(54394,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(54394,36,'cp 60','f3d4 d5e4 c3e4 a7a6 f1d3 f8e7 e1g1 g8f6 c2c3 b8d7 f1e1 e8g8 e4f6 d7f6 c1f4 f6d5 f4g3 e7d6 d4f3 d6g3 h2g3 h7h6 f3e5 c8d7 c3c4 d5f6 d1d2 d7c6 d2e3 d8e7 a1d1 f8d8 e3b6 a8c8 b2b3 d8d6 d3e2 f6e4 e2f3 c8d8 d1d6');
INSERT INTO evaluations VALUES(54356,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(54356,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(54356,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(54356,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(54356,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(54356,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(54356,36,'cp 55','d5e4 c3e4 a7a6 f1d3 f8e7 e1g1 g8f6 c2c3 b8d7 f1e1 e8g8 e4f6 d7f6 c1f4 f6d5 f4g3 e7d6 d4f3 d6g3 h2g3 c8d7 f3e5 d7c6 d1e2 d8b6 a1c1 d5f6 a2a3 a8d8 b2b4 c6a4 c3c4 b6c7 e2b2 f6d7 e5f3 h7h6 c4c5 f8e8 d3c2 a4b5 c2b3 b7b6 c5b6');
INSERT INTO evaluations VALUES(54454,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(54454,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(54454,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(54454,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(54454,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(54454,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(54454,36,'cp 62','c3e4 a7a6 f1d3 f8e7 e1g1 g8f6 c2c3 e8g8 d1f3 b8d7 f1e1 f8e8 d3c2 f6e4 f3e4 d7f6 e4h4 g7g6 c2b3 c8d7 c1h6 f6h5 h4e4 d8c7 e4f3 e7d6 g2g3 a8c8 h2h4 b7b5 h6e3 d6e5 a1d1 e5g7 g3g4 h5f6 e3g5 d7c6 d4c6 c7c6 f3c6');
INSERT INTO evaluations VALUES(54453,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(54453,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(54453,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(54453,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(54453,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(54453,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(54453,36,'cp 48','a7a6 f1d3 f8e7 e1g1 g8f6 c2c3 b8d7 f1e1 e8g8 d1f3 f8e8 d3c2 f6e4 f3e4 d7f6 e4h4 g7g6 c2b3 f6h5 h4e4 d8c7 d4f3 h5f6 e4e2 b7b5 f3e5 a8b8 c1g5 f6h5 g5h6 e7f8 h6f8 e8f8 a1d1 h5f6 e2f3 g8g7 f3g3 c8b7 e5d7 c7g3 f2g3 f6d7 d1d7 b7c6');
INSERT INTO evaluations VALUES(57054,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(57054,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(57054,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(57054,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(57054,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(57054,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(57054,36,'cp -24','b1c3 d7d5 e2e4 d5e4 c3e4 g8f6 e4f6 d8f6 g2g3 h7h5 d2d4 c7c5 g1f3 b8c6 c1e3 c5d4 e3d4 c6d4 d1d4 f6d4 f3d4 f8c5 d4f3 c8d7 f3e5 e8c8 e5d7 d8d7 a1d1 h5h4 f1g2 d7e7 c2c3 e6e5 f4e5 e7e5 e1d2 c5d6 d2c1 c8c7 g2f3 h4g3 h2g3 h8h1');
INSERT INTO evaluations VALUES(6485,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(6485,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6485,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(6485,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(6485,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6485,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6485,36,'cp 0','g8f6 f3e5 c6e5 d3d4 c5d6 d4e5 d6e5 f1d3 e8g8 e1g1 f8e8 c1g5 c7c6 g1h1 h7h6 g5h4 e5c3 b2c3 g7g5 h4g3 f6e4 f2f4 d7d5 d1h5 c8f5 f4g5 f5g6 h5h6 d8g5 h6g5 e4g5 g3f4 g5e4 c3c4 d5c4 d3c4 a8d8 c4d3 e4c5 d3g6 f7g6 h2h3 d8d7 h1g1 c5e4');
INSERT INTO evaluations VALUES(54238,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(54238,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(54238,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(54238,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(54238,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(54238,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(54238,36,'cp 95','f1b5 b8c6 e4d5 e6d5 e1g1 g8f6 f1e1 c8e6 c1g5 f8d6 d4c5 d6c5 g5f6 g7f6 d1d2 e8g8 d2h6 c6e7 b5d3 e7g6 b1c3 f8e8 h6d2 a7a6 h2h3 d8c7 c3e2 e8e7 a1d1 e6d7 e2c3 d7e6 d3g6 h7g6 c3d5 e6d5 d2d5 g8g7 e1e7 c7e7 d5d2 e7c7 d2e2 a8c8');
INSERT INTO evaluations VALUES(54038,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(54038,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(54038,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(54038,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(54038,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(54038,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(54038,36,'cp 82','c8d7 b5d7 d8d7 e4d5 c5d4 d1d4 e6d5 b1c3 g8f6 e1g1 b8c6 d4d3 f8e7 c1g5 c6b4 d3d2 f6e4 c3e4 d5e4 d2d7 e8d7 a1d1 d7e8 g5e7 e8e7 d1d4 e4f3 d4b4 h8b8 g2f3 b7b6 f1e1 e7f8 b4d4 b8d8 e1d1 d8d4 d1d4 a8c8 c2c3 f8e7 g1f1 c8c5 f1e2 e7e6 d4h4 h7h6 h4d4 g7g6 d4h4 c5a5 a2a4 e6e7 h4d4 a5f5');
INSERT INTO evaluations VALUES(30883,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(30883,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(30883,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(30883,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(30883,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(30883,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(30883,36,'cp 82','b5d7 d8d7 e4d5 c5d4 d1d4 e6d5 b1c3 g8f6 e1g1 b8c6 d4d3 f8e7 c1g5 c6b4 d3d2 f6e4 c3e4 d5e4 d2d7 e8d7 a1d1 d7e8 g5e7 e8e7 d1d4 e4f3 d4b4 h8b8 g2f3 b7b6 b4e4 e7f8 f1d1 b8d8 e4d4 d8d4 d1d4 a8c8 c2c3 f8e7 g1f1 c8c5 a2a4 g7g6 h2h4 e7e6 f1e2 e6e7 e2d3 c5f5');
INSERT INTO evaluations VALUES(30855,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(30855,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(30855,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(30855,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(30855,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(30855,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(30855,36,'cp 80','d8d7 e4d5 c5d4 d1d4 e6d5 b1c3 g8f6 e1g1 b8c6 d4d3 f8e7 c1g5 c6b4 d3d2 f6e4 c3e4 d5e4 d2d7 e8d7 f1d1 d7e8 g5e7 e8e7 d1d4 e4f3 d4b4 h8b8 g2f3 b7b6 a2a4 b8d8 a4a5 a8b8 a5b6 a7b6 c2c3 b8b7 g1g2 e7f8 a1a6 d8d6 b4f4');
INSERT INTO evaluations VALUES(13568,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(13568,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(13568,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(13568,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(13568,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13568,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(13568,36,'cp 16','b1d2 f8d6 f4g3 e8g8 f1d3 g7g6 e1g1 b7b6 g3h4 c8b7 f1e1 d6e7 d4c5 b6c5 e3e4 d5e4 d3e4 f6e4 h4e7 c6e7 d2e4 d8c7 d1e2 a8b8 b2b3 b8d8 h2h3 b7e4 e2e4 d8d5 a1d1 f8d8 e4e2 h7h6 g2g3 g8g7 h3h4 d5d1 e1d1 d8d1 e2d1 e7c6 d1e2');
INSERT INTO evaluations VALUES(13566,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(13566,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(13566,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13566,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13566,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13566,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(13566,36,'cp 8','f8d6 f4g3 e8g8 f1d3 g7g6 g3h4 d6e7 e1g1 b7b6 h4g5 d8c7 f3e5 f6d7 e5d7 c8d7 g5e7 c6e7 f1e1 d7c6 d2f3 g8g7 a2a4 a7a5 h2h4 e7f5 d3f5 g6f5 h4h5 f7f6 d4c5 b6c5 b2b4 c5b4 c3b4 a5b4 h5h6 g7h8');
INSERT INTO evaluations VALUES(8387,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(8387,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(8387,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(8387,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(8387,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(8387,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(8387,36,'cp 29','f4g3 e8g8 f1d3 g7g6 g3h4 d6e7 e1g1 b7b6 f3e5 c8b7 h4g5 c6e5 d4e5 f6d7 g5e7 d8e7 f2f4 b7c6 d1e1 b6b5 d2f3 b5b4 h2h4 a7a5 e1g3 g8h8 d3c2 f8g8 g3h3 g8g7 a1e1 a8b8 f1f2 a5a4 c3b4 b8b4 b2b3 a4b3 a2b3 d5d4 f3d2 d7b6 e3e4 c6b5 d2f3');
INSERT INTO evaluations VALUES(8392,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(8392,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8392,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(8392,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(8392,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(8392,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(8392,36,'cp 24','e8g8 f1d3 g7g6 e1g1 b7b6 f3e5 c8b7 d1f3 f6e8 g3f4 f7f6 e5c6 b7c6 f3g3 d6f4 e3f4 e8g7 g3e3 c5c4 d3c2 b6b5 a2a3 a7a5 g2g4 d8d6 b2b3 d6c7 f1e1 b5b4 b3c4 d5c4 a3b4 a5b4 c3b4 c6d5');
INSERT INTO evaluations VALUES(3281,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(3281,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(3281,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(3281,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(3281,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3281,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(3281,36,'cp 26','f1d3 g7g6 g3h4 e6e5 f3e5 c6e5 d4e5 d6e5 e1g1 d8b6 d1b3 g8g7 d2f3 b6b3 a2b3 e5c7 b3b4 c5b4 c3b4 c8d7 f3d4 c7d6 b4b5 f6e4 f1d1 d6e5 f2f3 e4c5 d3f1 c5e6 d4c2 e6c7 h4g3 e5g3 h2g3 d7b5 f1b5 c7b5 d1d5 f8c8 d5b5');
INSERT INTO evaluations VALUES(3284,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(3284,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(3284,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(3284,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(3284,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3284,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3284,36,'cp 16','a7a5 a2a3 d8e7 e3e4 f6e4 d2e4 d5e4 d3e4 c5d4 c3d4 e6e5 d4e5 c6e5 e1g1 e5f3 d1f3 a8a6 f1e1 c8e6 a1d1 d6g3 h2g3 a6b6 f3d3 e7f6 e4h7 g8h8 h7e4 b6b2 d3e3 h8g8 d1b1 f8d8 e4b7 b2b1 e1b1 g7g6 b7e4 f6d4');
INSERT INTO evaluations VALUES(21081,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(21081,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(21081,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(21081,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(21081,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21081,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(21081,36,'cp 40','d2d4 c8f5 c2c4 e7e6 f1g2 c6b4 e1g1 b4c2 c4d5 c2a1 e2e4 f5e4 d5e6 f7e6 b1c3 e4f3 d1f3 f8d6 f3h5 e8f8 h5e2 c7c6 e2e6 d8e7 e6f5 g8f6 c1e3 g7g6 f5b1 f8g7 d4d5 d6e5 f1e1 e7d7 e1d1 d7g4 b1a1 h8d8 a2a3 c6d5 c3d5 f6d5 g2d5');
INSERT INTO evaluations VALUES(52012,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(52012,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(52012,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(52012,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(52012,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(52012,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(52012,36,'cp -10','d7d5 b1c3 d5d4 c3b1 f8d6 g2g3 b8c6 f1c4 g8e7 a2a4 c6a5 b1a3 e8g8 d2d3 c8d7 c4b5 e7c6 a3c4 a5c4 b5c4 a7a6 e1g1 c6a5 c4a2 b7b5 a4b5 a6b5 c1d2 c5c4 d1e2 a8c8 b2b4 a5c6 d3c4 c6b4 d2b4');
INSERT INTO evaluations VALUES(54262,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(54262,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(54262,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(54262,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54262,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(54262,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54262,36,'cp 0','f1b5 b8d7 e4e5 g8e7 e1g1 a7a6 b5e2 e7f5 e2d3 f5h6 d3e2');
INSERT INTO evaluations VALUES(54044,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54044,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54044,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54044,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54044,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54044,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54044,36,'cp 0','b8d7 e4e5 g8e7 e1g1 a7a6 b5e2 e7f5 e2d3 f5h6 d3e2');
INSERT INTO evaluations VALUES(55678,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(55678,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55678,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55678,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(55678,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(55678,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(55678,36,'cp 27','g8f6 b1c3 b8c6 f1b5 f8e7 e4e5 f6g4 e5d6 d8d6 d4d6 e7d6 c3e4 e8g8 e4d6 c7d6 e1g1 g4e5 f3e5 d6e5 c2c3 c8e6 f2f4 a7a6 b5e2 f7f6 c1e3 e5f4 e3f4 a8d8 f1e1 g8f7 b2b3 d8d7 e1d1 f8d8 d1d7 d8d7 a1d1 d7d1 e2d1 c6e5 f4e3 e6f5 d1e2 f7e7');
INSERT INTO evaluations VALUES(47952,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47952,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(47952,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(47952,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(47952,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(47952,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(47952,36,'cp 30','c1f4 b8c6 f1b5 f8e7 b1c3 e8g8 d4d2 c8d7 b5d3 f8e8 e1c1 a7a6 h1e1 d7g4 c1b1 c6e5 d3e2 e5g6 f4g3 b7b5 h2h3 g4h5 a2a3 e7f8 f3d4 c7c5 d4f5 b5b4 a3b4 c5b4 c3d5 f6e4 d2b4 a8b8 b4d4 e4g3 f5g3 h5e2 e1e2 g6e7 d5e7 e8e7 e2e7 f8e7');
INSERT INTO evaluations VALUES(49776,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(49776,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(49776,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(49776,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(49776,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(49776,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(49776,36,'cp 73','e4e5 f6e4 d1d4 d7d5 e5d6 e4d6 d4f4 b8c6 b1c3 c8f5 f1b5 d8e7 c1e3 d6b5 c3b5 e7b4 f4b4 f8b4 c2c3 b4d6 b5d6 c7d6 e1c1 f5e6 b2b3 d6d5 f3g1 e8c8 g1e2 h8e8 d1d2 f7f6 e2g3 b7b6 c1b2 d8d7 a2a4 e6f7 h1d1 c6e5 h2h4 e5g6 h4h5 g6h4 d1g1 e8e5');
INSERT INTO evaluations VALUES(49504,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(49504,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(49504,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(49504,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(49504,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(49504,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(49504,36,'cp 71','f6e4 d1d4 d7d5 e5d6 e4d6 b1c3 b8c6 d4f4 c8f5 f1b5 d8e7 c1e3 d6b5 c3b5 e7b4 f4b4 f8b4 c2c3 b4d6 b5d6 c7d6 e1c1 f5e6 b2b3 d6d5 f3g1 e8c8 g1e2 h8e8 e2g3 f7f6 d1d2 b7b6 g3h5 e8e7 h2h4 d8d7 h5g3 e7e8 f2f3 e6f7 e3f2 e8e5 g3e2 d7d8 c1b2 c8b7 f2g3 e5e8');
INSERT INTO evaluations VALUES(49856,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(49856,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(49856,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(49856,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(49856,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(49856,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(49856,36,'cp 58','d1d4 d7d5 e5d6 e4d6 b1c3 b8c6 d4f4 c8f5 f1b5 d8e7 c1e3 d6b5 c3b5 e7b4 f4b4 f8b4 c2c3 b4d6 b5d6 c7d6 e1c1 f5e6 b2b3 d6d5 f3g1 f7f6 h2h4 e8c8 g1e2 h8e8 e2g3 b7b6 e3d4 d8d7 h1e1 d7e7 h4h5 e6g4 e1e7 c6e7 f2f3 e7f5 g3f5 g4f5 c1d2 f5d7 d1h1 e8e6 g2g3 d7c6 h1a1 g7g5');
INSERT INTO evaluations VALUES(49853,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(49853,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(49853,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(49853,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(49853,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(49853,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(49853,36,'cp 58','d7d5 e5d6 e4d6 b1c3 b8c6 d4f4 c8f5 f1b5 d8e7 c1e3 d6b5 c3b5 e7b4 f4b4 f8b4 c2c3 b4d6 b5d6 c7d6 e1c1 f5e6 b2b3 d6d5 f3g1 e8c8 g1e2 h8e8 d1d2 f7f6 c1b2 b7b6 h2h4 e6g4 e2g3 c6e7 f2f3 e7f5 e3f4 f5g3 f4g3 g4e6 a2a4 c8b7 g3f2 e6f7 a4a5 d8d7 a5b6');
INSERT INTO evaluations VALUES(49060,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(49060,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(49060,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(49060,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(49060,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(49060,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(49060,36,'cp 69','e5d6 e4d6 b1c3 b8c6 d4f4 c8f5 f1b5 d8e7 c1e3 d6b5 c3b5 e7b4 f4b4 f8b4 c2c3 b4d6 b5d6 c7d6 e1c1 f5e6 b2b3 d6d5 f3g1 e8c8 g1e2 f7f6 f2f3 h8e8 e3f2 e6f7 d1d2 b7b6 h2h4 e8e5 h1d1 g7g5 d1h1 f7g6 h4h5 g6f7 a2a4 e5e6 c1b2 d8e8 e2d4');
INSERT INTO evaluations VALUES(47748,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(47748,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(47748,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(47748,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(47748,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(47748,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(47748,36,'cp 63','e4d6 b1c3 b8c6 d4f4 c8f5 f1b5 d8e7 c1e3 d6b5 c3b5 e7b4 f4b4 f8b4 c2c3 b4d6 b5d6 c7d6 e1c1 f5e6 b2b3 d6d5 f3g1 e8c8 g1e2 h8e8 d1d2 f7f6 h2h3 b7b6 c1b2 e6f5 h3h4 f5g4 e2g3 c6e7 a2a4 d8d7 h4h5 e7f5 g3f5 g4f5 a4a5 c8b7 a5b6 a7b6 h1d1 b7c6 d1a1 d7e7');
INSERT INTO evaluations VALUES(47750,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(47750,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(47750,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(47750,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(47750,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(47750,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(47750,36,'cp 56','b1c3 b8c6 d4f4 c8f5 f1b5 d8e7 c1e3 d6b5 c3b5 e7b4 f4b4 f8b4 c2c3 b4d6 b5d6 c7d6 e1c1 f5e6 b2b3 d6d5 f3g1 e8c8 g1e2 h8e8 d1d2 f7f6 h2h4 b7b6 h4h5 e6f7 e2g3 d8d7 c1c2 c8b7 c2b2 h7h6 b2c1 e8d8 c1c2 d8e8 c2b2 f7e6 b2c1 e6f7 h1h4');
INSERT INTO evaluations VALUES(17897,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(17897,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(17897,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(17897,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(17897,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(17897,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(17897,36,'cp 77','d2d4 e5d4 e1g1 d7d6 f3d4 c8d7 a2a4 g8f6 d4c6 d7c6 b1c3 f8e7 f2f3 f6d7 c1e3 e7g5 e3d4 g5f4 c3e2 f4e5 b2b4 b7b5 c4a2 e8g8 a4a5 c6b7 d4e5 d7e5 d1d2 c7c5 c2c3 d8c7 a1d1 a8d8 e2g3 c5c4 g3f5 d6d5 d2g5 e5g6');
INSERT INTO evaluations VALUES(17899,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(17899,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(17899,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(17899,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(17899,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(17899,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(17899,36,'cp 75','e5d4 e1g1 d7d6 f3d4 c8d7 b1c3 g8f6 f1e1 f8e7 a2a4 c6d4 d1d4 e8g8 c1f4 d7c6 b2b4 f6d7 c4b3 e7f6 d4d2 f8e8 f2f3 b7b5 a4a5 h7h6 a1d1 d7e5 f4e3 a8b8 c3d5 c6d5 d2d5 d8e7 g1f1 f6h4 e1e2 b8d8 f3f4 e5c4 b3c4 b5c4');
INSERT INTO evaluations VALUES(17917,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(17917,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(17917,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(17917,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(17917,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(17917,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(17917,36,'cp 71','e1g1 d7d6 f3d4 c8d7 b1c3 g8f6 d4c6 d7c6 a2a4 f8e7 f2f3 e8g8 c1e3 f8e8 d1d2 a8b8 a4a5 h7h6 e3f2 f6d7 b2b4 e7h4 f2a7 b8a8 a7d4 h4g5 d2f2 g5f6 a1d1 b7b6 a5b6 d7b6 c4b3 f6d4 d1d4');
INSERT INTO evaluations VALUES(24828,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(24828,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(24828,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(24828,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(24828,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(24828,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(24828,36,'cp 83','f1b5 f5d7 c2c4 e7e6 b1c3 f8b4 a1c1 e8g8 f4g5 c6e7 g5f6 g7f6 b5d7 d8d7 c4d5 b4c3 c1c3 e7d5 c3c5 c7c6 e1g1 f6f5 f3e5 d7e7 d1h5 g8g7 a2a4 a8d8 a4a5 a7a6 b2b3 f8g8 g2g3 f7f6 e5d3 g7h8 f1e1 g8g7');
INSERT INTO evaluations VALUES(24777,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(24777,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(24777,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(24777,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(24777,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(24777,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(24777,36,'cp 77','f5d7 c2c4 e7e6 e1g1 f8d6 c4c5 d6f4 e3f4 e8g8 b1c3 a7a6 b5c6 d7c6 a1c1 g8h8 f1e1 f6g8 b2b4 g8e7 a2a4 b7b5 a4b5 c6b5 d1d2 h8g8 h2h3 e7f5 c1a1 f5h4 f3e5 f7f6 e5d3 d8d7 c3b5 a6b5 a1a8 f8a8 d2e2');
INSERT INTO evaluations VALUES(57768,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(57768,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(57768,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(57768,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(57768,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(57768,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(57768,36,'cp -7','c7c5 e2e3 b8c6 d2d4 c5d4 e3d4 d7d5 c2c3 g8f6 c1f4 c8f5 g1f3 e7e6 b1d2 h7h6 f1e2 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 d6c7 a3a4 f6e8 d2f1 e8d6 e2d3 f5d3 d1d3 c6e7 d3e2 f8e8 f1e3 e7c6 h2h4 a8c8 h4h5 d6e4');
INSERT INTO evaluations VALUES(52491,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(52491,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(52491,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(52491,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(52491,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(52491,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(52491,36,'cp -8','e2e3 b8c6 d2d4 c5d4 e3d4 d7d5 c2c3 c8f5 c1f4 e7e6 g1f3 g8f6 b1d2 f8d6 f4d6 d8d6 f1b5 e8g8 e1g1 f8c8 a3a4 a7a6 b5e2 h7h6 d2b3 d6c7 f1e1 c8e8 h2h3 a8c8 e2f1 f5g6 b3c5 f6e4 c5e4 d5e4 f3d2 f7f5 d2c4 f5f4');
INSERT INTO evaluations VALUES(52490,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(52490,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(52490,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(52490,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(52490,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(52490,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(52490,36,'cp -12','b8c6 d2d4 c5d4 e3d4 d7d5 f1b5 c8f5 g1f3 e7e6 c1f4 g8e7 e1g1 e7g6 f4g3 h7h5 h2h3 f8d6 g3d6 d8d6 b1c3 e8g8 c3e2 f5e4 f3g5 f7f6 g5e4 d5e4 f2f3 f6f5 b5c6 d6c6 f3e4 f5e4 f1f8 a8f8 e2c3 h5h4 d1h5');
INSERT INTO evaluations VALUES(21122,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(21122,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(21122,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(21122,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(21122,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(21122,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(21122,36,'cp 86','d4d5 c6b8 e2e4 e7e6 f1c4 g8f6 b1c3 e6d5 c3d5 f6d5 e1g1 f8e7 c4d5 e8g8 d1c2 c7c6 d5b3 b8a6 c1e3 d8a5 f3d4 a6b4 c2c3 c6c5 d4e2 b4c6 f1d1 a5c3 e2c3 b7b6 c3d5 c8e6 f2f4 f8d8 e3f2 e6d5 b3d5 a8c8 a1c1 c6d4 f2d4 c5d4 c1c8');
INSERT INTO evaluations VALUES(20954,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(20954,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(20954,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(20954,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(20954,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(20954,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(20954,36,'cp 89','c6b8 e2e4 e7e6 f1c4 g8f6 d1a4 c8d7 a4b3 e6d5 e4d5 d8e7 e1d1 e7b4 h1e1 e8d8 b3c2 d7g4 c1d2 b4c5 d2e3 c5d6 b1d2 b8d7 h2h3 g4h5 a1c1 h5g6 c2b3 d6b4 f3h4 f8d6 b3b4 d6b4 h4g6 h7g6 e3g5 b4d6 e1e2 a7a6 a2a3 h8h5 d2f3 h5h8 d1c2 h8h5 h3h4');
INSERT INTO evaluations VALUES(20194,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(20194,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(20194,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(20194,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(20194,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(20194,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(20194,36,'cp -29','d8a5 b5c6 b7c6 d4c5 a5c5 g1f3 g8e7 e1g1 c5b6 b2b4 e7f5 b1a3 a7a5 a3c2 a5b4 c3b4 c8a6 f1e1 a6b5 g2g4 f5e7 c1e3 b6b7 c2d4 b5a4 d1d2 e7g6 a2a3 h7h5 h2h3 h5g4 h3g4 c6c5 b4c5 f8c5 a1b1 b7d7 g1g2 c5d4 e3d4 e8g8 e1c1 a4b5 d4e3 d5d4 d2d4 d7d4 f3d4 b5d3 b1b4 g6e5 d4c6 e5c6 c1c6');
INSERT INTO evaluations VALUES(1561,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(1561,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(1561,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(1561,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(1561,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(1561,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(1561,36,'cp -29','b5c6 b7c6 d4c5 a5c5 g1f3 c5b6 e1g1 g8e7 b2b4 a7a5 c1e3 b6b8 e3c5 e7g6 c5f8 g6f8 d1d4 f8d7 b1d2 c8a6 f1e1 e8g8 d2b3 a5b4 c3b4 a6c4 b3c5 d7c5 d4c5 b8b5 f3d4 b5c5 b4c5 a8a5 d4c6 a5c5 c6d4 f8c8 a2a4 c8a8 e1c1 h7h6 c1b1 c5a5 b1b4 g7g6 h2h4');
INSERT INTO evaluations VALUES(1531,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(1531,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(1531,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(1531,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(1531,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(1531,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(1531,36,'cp -46','b7c6 d4c5 a5c5 g1f3 c5b6 e1g1 g8e7 b2b4 a7a5 c1e3 b6b8 e3c5 e7g6 c5f8 g6f8 d1d4 c8a6 f1e1 f8d7 a2a4 e8g8 b1a3 c6c5 b4c5 f8c8 e1b1 b8a7 c5c6 c8c6 d4a7 a8a7 f3d4 c6c8 a3b5 a6b5 a4b5 d7e5 b5b6 a7b7 a1a5 e5c4 a5a6 h7h6 b1b4 e6e5 d4f5');
INSERT INTO evaluations VALUES(1317,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(1317,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(1317,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(1317,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(1317,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(1317,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(1317,36,'cp -40','d4c5 a5c5 g1f3 g8e7 e1g1 c5b6 b2b4 a7a5 c1e3 b6b8 e3c5 e7g6 c5f8 g6f8 a2a3 f8d7 d1d4 c8a6 f1c1 e8g8 b1d2 a6e2 d4e3 e2f3 d2f3 f8c8 a1b1 b8b7 h2h4 h7h6 f3d4 b7a6 h4h5 c6c5 d4b3 a5b4 a3b4 a6c4 b4c5 d7c5 b3c5 c4c5');
INSERT INTO evaluations VALUES(9654,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(9654,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(9654,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(9654,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(9654,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(9654,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(9654,36,'cp 55','c5f2 e1f1 d8e7 f7h8 d7d5 e4d5 c6d4 d5d6 e7d6 c2c3 c8g4 d1a4 b7b5 c4b5 d4b5 f1f2 g4d7 a4d1 d6b6 f2e1 b5d6 d1f3 e5e4 f3e3 b6a6 b2b3 e8c8 c1a3 d7g4 a3d6 d8d6 c3c4 d6d3 e3f4 e4e3 h2h3 e3d2 b1d2 a6e6 e1f1 e6e2 f1g1 d3d2 f4g3 e2e4 a1f1 g4e2 f1e1 e4d4 g3f2 d4f2 g1f2 e2c4 f2g1 c4d5 h1h2 c8b7');
INSERT INTO evaluations VALUES(9657,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(9657,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(9657,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(9657,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(9657,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(9657,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(9657,36,'cp 40','e1f1 d8e7 f7h8 d7d5 e4d5 c6d4 d5d6 e7d6 c2c3 c8g4 d1a4 b7b5 c4b5 d4b5 f1f2 g4d7 a4d1 d6b6 f2e1 b5d6 d1f3 e5e4 f3e3 b6a6 b2b3 e8c8 c3c4 d8h8 b1c3 h8e8 h2h3 d6f5 e3c5 e4e3 h1f1 e3d2 e1d2 f5g3 d2c2 g3f1 c1f4 a6b6 c5b6 a7b6 a1f1 d7c6 f1g1 f6e4 c3b5 g7g5 f4h2 c6b5 c4b5 e4f2 c2d2 f2e4 d2d3 e4c5 d3d4 e8e2 h2e5 c8b7 d4d5 e2d2 d5c4');
INSERT INTO evaluations VALUES(9656,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(9656,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(9656,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(9656,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(9656,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(9656,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(9656,36,'cp 45','d8e7 f7h8 d7d5 e4d5 c6d4 d5d6 e7d6 c2c3 c8g4 d1a4 b7b5 c4b5 d4b5 f1f2 g4d7 a4d1 d6b6 f2e1 b5d6 d1f3 e5e4 f3e3 b6a6 h2h3 e8c8 c3c4 d8h8 b2b3 d6f5 e3c5 e4e3 b1c3 h8e8 h1f1 e3d2 e1d2 f5g3 d2c2 g3f1 c1f4 d7c6 a1f1 f6e4 c3e4 c6e4 c2c3 a6b6 c5b6 a7b6 f1f2 e8f8');
INSERT INTO evaluations VALUES(582,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(582,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(582,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(582,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(582,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(582,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(582,36,'cp 44','f7h8 d7d5 e4d5 c6d4 d5d6 e7d6 c2c3 c8g4 d1a4 b7b5 c4b5 d4b5 f1f2 g4d7 a4d1 d6b6 f2e1 b5d6 d1f3 e5e4 f3e3 b6a6 b2b3 e8c8 c3c4 d8h8 b1c3 h8e8 h2h3 d6f5 e3c5 e4e3 h1f1 e3d2 e1d2 f5g3 d2c2 g3f1 c1f4 a6c6 c5c6 d7c6 a1f1 c6g2 f1g1 g2e4 c2b2');
INSERT INTO evaluations VALUES(415,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(415,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(415,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(415,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(415,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(415,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(415,36,'cp 42','d7d5 e4d5 c6d4 d5d6 e7d6 h8f7 d6c5 d2d3 c8g4 d1d2 f2h4 b1c3 b7b5 b2b4 c5e7 c4b3 f6d5 b3d5 e7f6 f1g1 g4e2 d5f3 e2f3 d2e3 e8f7 g2f3 d4f3 g1g2 f6g6 g2f3 g6h5 f3g2 h5g4 e3g3 h4g3 h2g3 f7g8 c1d2 g4b4 a2a4 b4c5 a4b5 e5e4 c3e4 c5c2 a1a3 c2b2 a3c3 b2b5 c3c7 b5d3 h1e1 a8d8 c7a7 d3d5 d2c3 d8d7 a7d7 d5d7 e1e3 d7d5 g2g1 d5b3 g1h2 b3a2 e4d2');
INSERT INTO evaluations VALUES(406,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(406,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(406,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(406,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(406,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(406,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(406,36,'cp 54','e4d5 c6d4 d5d6 e7d6 h8f7 d6c5 d2d3 c8g4 d1d2 f2h4 b1c3 b7b5 b2b4 c5e7 c4b3 f6d5 b3d5 e7f6 f1g1 g4e2 d5f3 e2f3 d2e3 e8f7 g2f3 d4f3 g1g2 f6g6 g2f3 g6h5 f3g2 h5g4 e3g3 h4g3 h2g3 g4b4 c1d2 f7g8 a2a4 b4c5 a4b5 e5e4 h1e1 e4d3 c2d3 c7c6 b5c6 c5c6 c3e4 c6d5 a1c1 a7a5 c1c7 a8f8 c7c5 d5d3 c5a5 h7h6 d2f4 d3d4 a5c5 f8a8 f4e5 d4a4 e1e3 a8f8');
INSERT INTO evaluations VALUES(405,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(405,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(405,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(405,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(405,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(405,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(405,36,'cp 101','c6d4 d5d6 e7d6 h8f7 d6c5 d2d3 f2h4 b2b4 c5e7 b1d2 b7b5 c4b3 e5e4 d3e4 d4b3 a2b3 e7f7 d2f3 f6e4 d1d3 f7f5 g2g4 f5g4 h1g1 g4f5 c1g5 h4g5 a1e1 g5f6 d3e4 f5e4 e1e4 e8d8 f3d4 a7a5 f1e2 a5b4 d4c6 d8d7 c6e7 c7c6 e7c8 d7c8 e4b4 c8c7 b4e4 h7h6 e2f3 a8a2');
INSERT INTO evaluations VALUES(408,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(408,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(408,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(408,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(408,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(408,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(408,36,'cp 96','d5d6 e7d6 h8f7 d6c5 d2d3 f2h4 b2b4 c5e7 b1d2 b7b5 c4b3 e5e4 d3e4 d4b3 a2b3 e7f7 d2f3 f6e4 d1d3 f7f5 g2g4 f5g4 h1g1 g4f5 c1g5 h4g5 a1e1 g5f6 d3e4 f5e4 e1e4 e8d8 f1e2 c8f5 g1d1 d8c8 e4e8 c8b7 e8a8 b7a8 e2d2 a8b7 d1e1 b7b6 e1e8 c7c5 e8b8 b6c6 b4c5 c6c5 c2c3');
INSERT INTO evaluations VALUES(6225,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(6225,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(6225,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(6225,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(6225,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(6225,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(6225,36,'cp 102','c2c3 d7d6 d2d4 e5d4 c3d4 c5b6 b1c3 g8e7 h2h3 e8g8 f1e1 c6a5 c4d3 a5c6 d3c2 f7f5 c2b3 g8h8 e4e5 f5f4 e5d6 c7d6 f3h4 c8f5 c1f4 c6d4 f4e3 d4b3 d1b3 d6d5 e3b6 a7b6 a1d1 f5h7 h4f3 h7g8 b3b5 e7c6 c3d5');
INSERT INTO evaluations VALUES(6215,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(6215,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(6215,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(6215,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(6215,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(6215,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(6215,36,'cp 124','d7d6 d2d4 e5d4 c3d4 c5b6 h2h3 g8e7 b1c3 e8g8 f1e1 e7g6 c1e3 c8e6 d4d5 b6e3 e1e3 c6e5 f3e5 g6e5 c4f1 e6d7 f2f4 e5g6 d1f3 d8f6 g2g3 b7b5 f3f2 b5b4 c3b5 c7c6 d5c6 d7c6 a1d1 f8d8 b5d4 a8c8 f2d2 a7a5 a2a3 b4a3 d4c6 c8c6 e4e5');
INSERT INTO evaluations VALUES(5734,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(5734,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(5734,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(5734,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(5734,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(5734,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(5734,36,'cp 106','d2d4 e5d4 c3d4 c5b6 h2h3 g8e7 b1c3 e8g8 f1e1 e7g6 c1e3 c8e6 d4d5 b6e3 e1e3 c6e5 f3e5 g6e5 c4f1 e6d7 f2f4 e5g6 d1f3 d8f6 g2g3 f6d4 f3f2 a8e8 e3e1 d4b4 a2a3 b4b3 h3h4 a7a6 f1g2 d7g4 e1e3 b3b6 a1e1 b6d4 g1h2 f7f5 e4f5 g4f5');
INSERT INTO evaluations VALUES(5749,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(5749,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(5749,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(5749,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(5749,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(5749,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(5749,36,'cp 106','e5d4 c3d4 c5b6 b1c3 g8e7 h2h3 e8g8 f1e1 g8h8 c4b3 c8d7 c1e3 f7f5 d4d5 c6e5 f3e5 d6e5 d5d6 c7d6 d1d6 b6e3 f2e3 d7c6 d6e5 f5e4 a1d1 d8e8 e1f1 f8f1 d1f1 e7g8 e5e8 a8e8 f1f7 g8f6 g1f2 e8d8 f2e1 a7a5 f7c7 d8d7 c7c8 h8h7 b3c2 g7g6 c3e2 f6d5');
INSERT INTO evaluations VALUES(5752,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(5752,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(5752,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(5752,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(5752,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(5752,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(5752,36,'cp 114','c3d4 c5b6 h2h3 g8e7 f1e1 e8g8 b1c3 g8h8 c1f4 e7g6 c3d5 c6e7 d5b6 a7b6 f4d2 d6d5 e4d5 e7d5 f3e5 c7c6 d1h5 h8h7 a2a3 d8f6 c4d5 c6d5 d2b4 g6f4 h5f3 f8e8 g1h2 c8e6 b4d2 f4g6 f3f6 g7f6 e5d3 a8c8 e1e3 e6f5 d3b4 e8e3 d2e3 g6e7 g2g4 f5e4 h2g3');
INSERT INTO evaluations VALUES(5751,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(5751,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(5751,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(5751,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(5751,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(5751,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(5751,36,'cp 106','c5b6 h2h3 g8e7 b1c3 e8g8 f1e1 g8h8 c1f4 e7g6 c3d5 c6e7 d5b6 a7b6 f4d2 d6d5 e4d5 e7d5 f3e5 c7c6 d1h5 h8h7 a2a3 d8f6 c4d5 c6d5 d2b4 g6f4 h5f3 f8e8 g1h2 c8e6 b4d6 f4g6 f3f6 g7f6 e5d3 a8c8 a1c1 c8c4 c1c4 d5c4 d3b4 g6h4 d4d5 h4g2 h2g2');
INSERT INTO evaluations VALUES(5723,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(5723,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(5723,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(5723,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(5723,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(5723,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(5723,36,'cp 110','b1c3 g8e7 h2h3 e8g8 f1e1 e7g6 c1e3 c8e6 d4d5 b6e3 e1e3 c6e5 f3e5 g6e5 c4f1 e6d7 f2f4 e5g6 d1f3 a7a6 a1d1 b7b5 b2b3 d8f6 g2g3 b5b4 c3e2 c7c5 d5c6 d7c6 f1g2 a8c8 g1h2 f8d8 f3f2 a6a5 d1d2 c6b5 e2d4 b5d7');
INSERT INTO evaluations VALUES(5720,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(5720,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(5720,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(5720,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(5720,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(5720,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(5720,36,'cp 95','g8e7 h2h3 e8g8 f1e1 e7g6 c1e3 c8e6 d4d5 b6e3 e1e3 c6e5 f3e5 g6e5 c4f1 e6d7 f2f4 e5g6 d1f3 b7b5 f3f2 a7a6 g2g3 f7f5 f1g2 b5b4 c3e2 d8f6 a1c1 a8c8 e4f5 d7f5 e3b3 a6a5 a2a3 b4a3 b2a3 h6h5 b3c3 c8e8 e2d4 h5h4 g3g4 f5d7');
INSERT INTO evaluations VALUES(49676,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(49676,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(49676,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(49676,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(49676,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(49676,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(49676,36,'cp -30','d7d5 f4e5 f6e4 g1f3 f8c5 d2d4 c5b4 c1d2 c7c5 f1d3 e4d2 d1d2 c5d4 f3d4 e8g8 d2f4 b8c6 d4c6 b4c3 b2c3 b7c6 e1g1 a8b8 f1f3 c6c5 c3c4 d5c4 d3c4 c8e6 a2a3 b8b6 a1e1 h7h6 f3g3 e6c4 f4c4 d8c7 g3d3 f8e8 d3d5 b6c6 h2h3 e8e6 c4c3 c5c4 c3e3 e6g6 g1h1');
INSERT INTO evaluations VALUES(48850,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(48850,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(48850,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(48850,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(48850,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(48850,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(48850,36,'cp -22','f4e5 f6e4 g1f3 f8c5 d2d4 c5b4 c1d2 c7c5 f1d3 e4d2 d1d2 c5d4 f3d4 b8c6 d2f4 e8g8 d4c6 b4c3 b2c3 b7c6 e1g1 a8b8 f1f3 c6c5 c3c4 d5c4 d3c4 c8e6 a2a3 d8c7 a1e1 b8d8 f3g3 g7g6 c2c3 f8e8 g3f3 e6c4 f4c4 e8e6 h2h3 d8d2');
INSERT INTO evaluations VALUES(48769,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(48769,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(48769,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48769,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48769,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48769,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(48769,36,'cp -25','f6e4 g1f3 f8c5 d2d4 c5b4 c1d2 c7c5 f1d3 e4d2 d1d2 c5d4 f3d4 e8g8 d2f4 b8c6 d4c6 b4c3 b2c3 b7c6 e1g1 a8b8 a1e1 c8e6 e1e3 g8h8 e3g3 c6c5 h2h3 b8b6 g1h2 c5c4 d3f5 d8c7 g3g5 h7h6 g5h5 h8g8 f4g3 e6f5 h5f5 b6e6 h3h4 c7e7 f1d1 e7d7 g3f4 a7a5');
INSERT INTO evaluations VALUES(49064,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(49064,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(49064,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49064,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(49064,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(49064,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(49064,36,'cp -26','g1f3 f8c5 d2d4 c5b4 c1d2 c7c5 f1d3 e4d2 d1d2 c5d4 f3d4 e8g8 d2f4 b8c6 d4c6 b4c3 b2c3 b7c6 e1g1 d8e7 f1f3 a8b8 f3g3 b8b6 h2h4 g8h8 a1f1 c6c5 g3g5 h7h6 g5h5 h8g8 g1h2 c5c4 d3f5 c8f5 h5f5 b6e6 h2g1 a7a5 f1f3 f7f6 e5f6 e6f6 f5f6 f8f6');
INSERT INTO evaluations VALUES(49061,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(49061,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(49061,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49061,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(49061,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(49061,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(49061,36,'cp -12','f8c5 d2d4 c5b4 c1d2 c7c5 f1d3 e4d2 d1d2 b8c6 d2f4 c5d4 f3d4 e8g8 d4c6 b4c3 b2c3 b7c6 e1g1 a8b8 a1e1 c8e6 e1e3 c6c5 e3g3 g8h8 h2h4 d8e7 g3g5 c5c4 d3f5 h7h6 g5h5 e6f5 h5f5 b8b2 f5f7 f8f7 f4f7 e7e5 h4h5 b2c2 f7g6 e5e3 g1h2 c2f2 f1f2 e3f2 g6e8 h8h7 e8g6 h7g8 g6e6 f2f7');
INSERT INTO evaluations VALUES(47215,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(47215,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(47215,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(47215,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(47215,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(47215,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(47215,36,'cp -80','d5e4 f2f3 e4f3 g1f3 c7c6 f1d3 c8g4 c1e3 e7e6 e1g1 g4h5 d1e1 h5g6 f3e5 g6d3 e5d3 f8e7 e1g3 e8g8 f1f3 b8d7 a1f1 f6e8 e3f4 d8b6 g3f2 e8f6 g1h1 g8h8 f3h3 a8d8 d3e5 d7e5 f4e5');
INSERT INTO evaluations VALUES(49708,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(49708,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(49708,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(49708,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(49708,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(49708,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(49708,36,'cp 10','c2c4 f8c5 b1c3 c7c6 e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 b8c6 c1e3 e8g8 e1g1 c8g4 h2h3 g4e2 d1e2 c6d4 e3d4 b6d4 a1d1 d8f6 b2b4 a8d8 e2g4 d4c3 d1d5 d8d5 g2d5 g7g6 f1d1 f8d8 b4b5 b7b6 d5g2 d8d1 g4d1 g8g7');
INSERT INTO evaluations VALUES(57769,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(57769,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(57769,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(57769,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(57769,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(57769,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(57769,36,'cp -21','e7e5 e2e4 g8f6 b1c3 f8b4 c3d5 f6d5 e4d5 e8g8 c2c3 b4c5 b2b4 c5d6 a4a5 c7c6 f1c4 b8a6 g1e2 a6c7 d2d4 c7d5 c4d5 c6d5 e1g1 d8c7 d4e5 d6e5 h2h3 d7d6 c1e3 c7c4 d1d2 f8e8 f1e1 c8f5 a1c1 c4d3 e2f4 d3d2 e3d2 d5d4 c3d4');
INSERT INTO evaluations VALUES(57256,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(57256,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(57256,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(57256,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(57256,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(57256,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(57256,36,'cp -12','e2e4 g8f6 b1c3 f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 f1e2 b4f8 c3e4 e8e4 e1g1 d7d5 e2f3 e4e8 f1e1 c8f5 e1e8 d8e8 d3e1 c7c6 d2d4 b8d7 a4a5 f8d6 g2g3 d7f6 e1g2 e8d7 c1f4 d6f4 g2f4 a8e8 c2c3 g7g6 d1d2 h7h5 a5a6 b7b6 a1e1 f6e4');
INSERT INTO evaluations VALUES(57250,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(57250,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(57250,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(57250,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(57250,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(57250,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(57250,36,'cp -15','g8f6 b1c3 f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 f1e2 b4f8 c3e4 e8e4 e1g1 d7d5 e2f3 e4e8 f1e1 c8f5 e1e8 d8e8 d3e1 c7c6 a4a5 f8d6 g2g3 b8d7 d2d4 d7f6 e1g2 e8d7 c1f4 d6f4 g2f4 a8e8 h2h4 f6e4 c2c3 h7h6 a5a6 b7b6 f4g2 e4d6');
INSERT INTO evaluations VALUES(49714,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(49714,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(49714,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(49714,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(49714,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(49714,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(49714,36,'cp -20','b1c3 f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 f1e2 b4f8 c3e4 e8e4 e1g1 d7d5 e2f3 e4e8 f1e1 c8f5 e1e8 d8e8 d3e1 c7c6 d2d4 b8d7 a4a5 e8e6 c1f4 a8e8 c2c3 h7h6 e1d3 d7f6 h2h3 f5d3 d1d3 f8d6 f4d6 e6d6 g2g3 g7g6 d3d2 g8g7 a1e1 e8e1 d2e1 b7b6');
INSERT INTO evaluations VALUES(49712,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49712,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(49712,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(49712,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49712,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(49712,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(49712,36,'cp -12','f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 c3e4 e8e4 f1e2 b4f8 e1g1 d7d5 e2f3 e4e8 f1e1 c8f5 e1e8 d8e8 d3e1 c7c6 d2d4 f8d6 g2g3 b8d7 e1g2 d7f6 c1f4 d6f4 g2f4 e8e7 c2c3 h7h6 d1d2 a8e8 f4d3 f5d3 d2d3 e7d7 g1g2 g7g5 d3d2 g8g7 a4a5 a7a6');
INSERT INTO evaluations VALUES(47187,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(47187,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47187,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47187,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47187,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(47187,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(47187,36,'cp 13','g1f3 e7e6 c1f4 f6h5 f4g5 f8e7 g5e7 d8e7 e2e3 h5f6 c3c4 e8g8 b1c3 f8d8 c4d5 e6d5 f1d3 c7c5 d4c5 e7c5 e1g1 b8c6 a1c1 c8g4 h2h3 g4f3 d1f3 c5e7 c1d1 e7b4 f3e2 d5d4 c3e4 f6e4 d3e4 d4e3 e2e3');
INSERT INTO evaluations VALUES(47182,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(47182,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(47182,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47182,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47182,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47182,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(47182,36,'cp -4','c8f5 d1b3 d8c8 c1f4 c7c5 b1d2 c5c4 b3d1 h7h6 a2a4 e7e6 b2b3 c4b3 d1b3 b8c6 e2e3 f8e7 f1e2 e8g8 e1g1 c6a5 b3b2 b7b6 f1c1 c8b7 c3c4 a5c4 d2c4 d5c4 c1c4 a8c8 f3d2 c8c4 d2c4 f8c8 a4a5 f6d5 f4g3 e7b4 h2h3 b6b5 c4d6');
INSERT INTO evaluations VALUES(29574,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29574,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29574,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29574,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29574,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29574,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29574,36,'cp 0','d1b3 d8c8 c1f4 c7c5 b1d2 c5c4 b3d1 b8c6 a2a4 a7a6 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(47119,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47119,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47119,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47119,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47119,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(47119,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(47119,36,'cp -18','c7c5 e2e3 b8c6 d4c5 f6e4 g5h4 g7g6 b1d2 e4c5 c2c3 a7a5 a2a4 f8g7 f1e2 e8g8 e1g1 h7h6 f3d4 g6g5 h4g3 e7e5 d4b5 c8f5 f2f4 d8d7 f4e5 c6e5 b5d4 f5g6 e2b5 e5c6 d4b3 c5e6 d2f3 e6c7 g3c7 d7c7 d1d5 c7e7');
INSERT INTO evaluations VALUES(45336,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(45336,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(45336,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(45336,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(45336,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45336,36,'cp 0','g5f6 g7f6 d4c5 e7e5 e2e4 d5e4 d1d8 e8d8 f3d2 f6f5 b1c3 f8c5 e1c1 c5d4 c3b5 b8c6 d2b3 d8e7 b5d4 e5d4 b3d4 c8e6 f2f3 a8d8 d4c6 b7c6 d1d8 h8d8 f1e2 e4e3 f3f4 e6d5 h1e1 d5g2 e2h5 d8d4 e1e3 e7d6 e3d3 d6c5 d3c3 c5d6');
INSERT INTO evaluations VALUES(55014,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(55014,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(55014,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(55014,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(55014,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(55014,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(55014,36,'cp -110','e7e5 c2c3 c7c6 d5b4 g8f6 g1f3 f8d6 d2d3 e8g8 a2a3 c6c5 c3d4 c5d4 c1g5 c8e6 b4d5 b8d7 d5f6 d7f6 f1e2 h7h6 g5h4 d8e7 e1g1 a7a5 h4f6 e7f6 f3d2 f6g6 a1c1 e6h3 e2f3 b7b5 d2b3 b5b4 g1h1 h3e6 a3b4 d6b4 b3c5 f8c8 c5e6 g6e6 c1c8 a8c8');
INSERT INTO evaluations VALUES(56129,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(56129,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(56129,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(56129,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(56129,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(56129,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(56129,36,'cp -107','c2c3 c7c6 d5b4 g8f6 g1f3 d4c3 b2c3 f6e4 d1e2 e4f6 f3e5 f8e7 b4c2 e8g8 c2e3 f6d5 e3d5 c6d5 e2d1 b8c6 e5c6 b7c6 f1e2 d8a5 e2d3 f8e8 e1g1 e7d6 f1e1 e8e1 d1e1 c8d7 e1d1 a8e8 d3f1 c6c5 g2g3 c5c4 d1f3');
INSERT INTO evaluations VALUES(56125,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(56125,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(56125,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(56125,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(56125,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(56125,35,'cp -121',NULL);
INSERT INTO evaluations VALUES(56125,36,'cp -101','c7c6 d5b4 g8f6 g1f3 f8d6 d2d3 e8g8 a2a3 d4c3 b2c3 b8d7 f1e2 d7c5 e1g1 f8e8 b4c2 c5b3 a1b1 b3c1 d1c1 b7b5 c2e3 f6g4 c3c4 a8b8 c4c5 d6c7 c1c3 a7a5 e2d1 d8e7 d1b3 e8d8 h2h3 g4f6 f1c1 h7h6 c1e1');
INSERT INTO evaluations VALUES(23430,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(23430,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(23430,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(23430,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(23430,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(23430,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(23430,36,'cp 67','d2d4 g8f6 g1f3 e7e6 a2a3 d7d5 b1c3 f8e7 e2e3 e8g8 d1c2 b7b6 c4d5 e6d5 f1d3 c6b8 e3e4 d5e4 c3e4 c8b7 e4f6 e7f6 c1e3 h7h6 e1c1 b8c6 f3e5 d8d6 f2f4 c6e7 h1e1 e7d5 e3d2 a8d8 g2g4 f6h4 e1e2 c7c5 d4c5 d6c5');
INSERT INTO evaluations VALUES(23431,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(23431,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(23431,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(23431,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(23431,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(23431,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(23431,36,'cp 51','e7e6 e2e3 d7d5 a2a3 g8f6 b1c3 f8e7 g1f3 e8g8 b2b4 c6b8 c4c5 c7c6 f1d3 a7a5 c1b2 b7b6 c3a4 b6c5 b4c5 c8a6 d3a6 a8a6 b2c3 b8d7 e1g1 d8c7 d1c2 a6a8 f1b1 f8b8 h2h3 e7d8 a4b2 c7c8 b1c1 d8c7');
INSERT INTO evaluations VALUES(22596,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(22596,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(22596,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(22596,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(22596,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(22596,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(22596,36,'cp 64','g1f3 d7d5 e2e3 g8f6 b1c3 f8e7 a2a3 e8g8 d1c2 h7h6 b2b4 c6b8 c4c5 b7b6 f1e2 a7a5 c1b2 b6c5 b4c5 c8a6 e2a6 a8a6 e1g1 b8d7 f1b1 a6a8 c3a4 f6e4 b2c3 d8c8 c3e1 c8a6 c5c6 d7b6 a4b6 c7b6 a3a4 e7d6 f3e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(22595,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(22595,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(22595,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(22595,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(22595,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(22595,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(22595,36,'cp 59','g8f6 a2a3 d7d5 b1c3 f8e7 c1f4 f6e4 e2e3 e4c3 b2c3 e8g8 c4d5 e6d5 f1d3 e7d6 f4d6 d8d6 e1g1 c6a5 d1c2 g7g6 h2h4 c7c5 d4c5 d6c5 h4h5 a5c4 h5g6 f7g6 d3c4 d5c4 f3d4 c8d7 c2e2 a7a6 a3a4 a8c8 f1d1 c8e8 e3e4 b7b5 a4b5 d7b5');
INSERT INTO evaluations VALUES(16274,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(16274,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(16274,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(16274,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(16274,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(16274,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(16274,36,'cp 64','a2a3 d7d6 e2e3 g7g6 f1e2 f8g7 b1c3 e8g8 e1g1 a7a5 a1b1 e6e5 d4d5 c6e7 e3e4 f6h5 f1e1 f7f5 e4f5 c8f5 f3g5 f5b1 g5e6 d8d7 e6f8 a8f8 c3b1 d7f5 c1e3 f5f7 e2h5 g6h5 b1c3 e7f5 c3e4 f7g6 e3d2 b7b6 h2h3 g8h8');
INSERT INTO evaluations VALUES(16277,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(16277,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(16277,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(16277,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(16277,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(16277,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(16277,36,'cp 54','g7g6 e2e3 f8g7 b2b4 d7d6 f1e2 e6e5 b4b5 c6b8 e1g1 e5d4 f3d4 e8g8 c1b2 a7a6 b1d2 b8d7 a3a4 d7c5 h2h3 d8e7 a4a5 f8e8 d4b3 c8f5 e2f3 a8d8 a1a3 h7h6 b2f6 e7f6 b3c5 d6c5 f3b7 d8d3 a3b3 f6d6 b3d3 f5d3 b5a6');
INSERT INTO evaluations VALUES(48176,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(48176,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(48176,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(48176,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(48176,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(48176,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(48176,36,'cp 48','d5e6 c8e6 f1e2 b8c6 g1f3 f8c5 e1g1 e6f5 d2d3 e8g8 c1e3 c5e3 f2e3 d8e7 e3e4 f5g4 b1c3 g4f3 e2f3 a8d8 c3d5 e7e5 d1d2 f6d7 a1d1 c6d4 f3g4 c7c6 d5e3 e5g5 g4d7 d8d7 d2f2 f8d8 h2h3 a7a5 b2b3 d4e6 g1h2');
INSERT INTO evaluations VALUES(48063,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(48063,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(48063,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(48063,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(48063,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(48063,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(48063,36,'cp 51','c8e6 f1e2 b8c6 g1f3 f8c5 e1g1 e8g8 d2d3 f8e8 b1c3 e6f5 c1f4 c6d4 f3d4 c5d4 a1c1 c7c6 d1d2 f6d7 c3e4 f5e4 d3e4 d7c5 c1d1 d4f6 d2c1 d8e7 f2f3 a8d8 f4e3 c5e6 f3f4 f6d4 e3d4 d8d4 c1e3 e7c5 d1d4 c5d4 e3d4 e6d4');
INSERT INTO evaluations VALUES(30034,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(30034,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(30034,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(30034,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(30034,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(30034,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(30034,36,'cp 51','f1e2 f8c5 g1f3 b8c6 e1g1 e6f5 d2d3 e8g8 b1c3 f8e8 c1f4 c6d4 f3d4 c5d4 a1c1 c7c6 d1d2 f6d7 c3e4 d7e5 c1d1 f5g6 f4e3 d4e3 d2e3 d8a5 e3d2 a5d2 d1d2 e5c4 e4f6 g7f6 d3c4 a8d8 f1d1 d8d2 d1d2 g8f8 g1f1 g6f5 a2a3 a7a5 d2d6 f8e7 c4c5 f5e6 f2f4 f6f5 e2d3');
INSERT INTO evaluations VALUES(21191,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(21191,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(21191,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(21191,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(21191,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(21191,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(21191,36,'cp 93','d2d4 g8f6 b1c3 e7e5 f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 b2b3 a7a5 d1f3 d8d7 c1b2 d7g4 f3d3 f8e8 a1e1 g4g6 g1h1 e7f8 h2h3 f6d7 d3f3 d7c5 b2a3 c5d7 a3c1 d7c5 e1e2 a5a4 b3b4 c5d7 f1e1 d7f6 e4e5 d6e5');
INSERT INTO evaluations VALUES(21185,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(21185,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(21185,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(21185,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(21185,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(21185,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(21185,36,'cp 75','g8f6 b1c3 a7a6 d4d5 c6b8 a2a4 e7e6 h2h3 e6d5 e4d5 f8e7 g2g3 e8g8 f1g2 b8d7 e1g1 d7e5 a4a5 e5f3 d1f3 c8d7 f1e1 h7h6 c3a2 f6h7 a2b4 e7f6 c2c3 f6g5 c1f4 g5f4 f3f4 f8e8');
INSERT INTO evaluations VALUES(14487,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(14487,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(14487,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(14487,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(14487,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(14487,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(14487,36,'cp 80','b1c3 e7e5 f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 d8d7 d1f3 d7g4 f3d3 e8g8 c1e3 a7a5 b2b3 f8e8 a1e1 g4g6 e3d4 f6d7 e1e2 e7f6 h2h3 h7h5 f1f3 h5h4 d4f6 d7f6 e4e5 d6e5 d3g6 f7g6 f4e5');
INSERT INTO evaluations VALUES(14483,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(14483,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(14483,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(14483,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(14483,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(14483,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(14483,36,'cp 75','e7e5 f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 a7a5 d1f3 e8g8 c1e3 d8d7 h2h3 a5a4 a1d1 d7e6 f1e1 f8d8 e3d4 e7f8 b2b3 a4b3 a2b3 a8a5 g1h1 d8e8 e4e5 d6e5 d4e5 f6d5 c3e4 f7f6 e5b2 e6f5 c2c4 a5a2 c4d5 a2b2');
INSERT INTO evaluations VALUES(17261,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17261,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(17261,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(17261,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(17261,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(17261,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17261,36,'cp 76','d4d5 c6b8 c2c4 e7e6 a2a3 f8c5 b1c3 d7d6 c1g5 e6d5 c4d5 e8g8 d1c2 b8d7 e2e3 d8e7 f1e2 f8e8 a1d1 d7e5 f3d2 h7h6 g5h4 e5g6 h4g3 c8g4 e2g4 f6g4 e1g1 h6h5 h2h3 h5h4 g3d6 e7d6 h3g4 h4h3 f1e1 h3g2');
INSERT INTO evaluations VALUES(17224,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(17224,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(17224,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(17224,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(17224,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(17224,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(17224,36,'cp 68','c6b8 b1c3 d7d6 e2e4 g7g6 c1e3 c7c6 h2h3 f8g7 d1d2 c6d5 e4d5 b8d7 f1e2 a7a6 a2a4 e8g8 e1g1 b7b6 e2c4 d8c7 b2b3 d7c5 e3d4 c8d7 d2e2 f8e8 a1d1 e7e5 d5e6 d7e6 b3b4 c5d7 c4e6 e8e6');
INSERT INTO evaluations VALUES(45134,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45134,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(45134,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(45134,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(45134,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(45134,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45134,36,'cp 0','e7e6 a2a4 b8c6 e1g1 f8e7 c4b3 c6a5 b3a2 a7a6 c2c3 e8g8 c1f4 c8d7 b1d2 a5c6 a2c4 c6a5');
INSERT INTO evaluations VALUES(45995,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45995,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45995,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45995,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45995,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45995,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45995,36,'cp 0','e1g1 f8e7 a2a4 b8c6 c4b3 c6a5 b3a2 a7a6 c1f4 b7b5 b1d2 e8g8 c2c3 a5c6 h2h3 h7h6 f4g3 d8b6 a2b3 c8d7 d1e2 f8d8 f1e1 d7e8 d2f1 a8b8 b3c2 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(45994,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45994,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45994,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45994,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45994,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45994,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45994,36,'cp 0','f8e7 a2a4 b8c6 c4b3 c6a5 b3a2 a7a6 c1f4 e8g8 b1d2 a5c6 c2c3 c8d7 h2h3 b7b5 f1e1 h7h6 f4g3 d8b6 a2b1 f8d8 b1c2 d7e8 d1e2 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(40180,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40180,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40180,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40180,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40180,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40180,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40180,36,'cp 0','a2a4 b8c6 c4b3 c6a5 b3a2 c8d7 c2c3 a7a6 c1f4 e8g8 b1d2 a5c6 h2h3 b7b5 a2b3 d8c7 f4g3 f8b8 b3c2 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(19007,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(19007,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(19007,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(19007,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19007,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(19007,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19007,36,'cp 12','f1c4 g8f6 d2d3 f8e7 e1g1 e8g8 f1e1 d7d6 a2a4 c8e6 b1a3 h7h6 c1d2 a7a6 c4e6 f7e6 d1e2 f6d7 b2b4 c5b4 c3b4 d6d5 a3c2 e7d6 b4b5 a6b5 a4b5 a8a1 e1a1 c6e7 d3d4 d8c7 d4e5 d6e5 f3e5 c7e5 d2b4 f8f7 a1a7');
INSERT INTO evaluations VALUES(18966,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(18966,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(18966,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18966,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18966,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(18966,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(18966,36,'cp 29','g8f6 f3g5 d7d5 e4d5 f6d5 d2d4 c5d4 g5f7 e8f7 d1f3 d8f6 c4d5 c8e6 d5e6 f7e6 f3e2 e6d7 e1g1 d7c7 b1d2 a8d8 d2e4 f6e6 c1g5 f8e7 g5e7 e6e7 a1d1 c7b8 h2h3 a7a6 a2a3 d4c3 e4c3 d8d4 b2b4 h8d8 f1e1 d4d1 e1d1 d8d1 e2d1 c6d4 d1d2 h7h5 f2f3 h5h4 c3e4 b8c7');
INSERT INTO evaluations VALUES(12528,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(12528,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(12528,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(12528,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12528,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(12528,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(12528,36,'cp 22','f3g5 d7d5 e4d5 f6d5 d2d4 c5d4 g5f7 e8f7 d1f3 f7e6 a2a4 f8e7 f3e4 g7g6 c3d4 c6b4 b1c3 e6f7 c4d5 b4d5 e4d5 d8d5 c3d5 e7d6 d4e5 d6e5 e1g1 c8f5 h2h3 h8e8 d5e3 f5d3 f1d1 a8d8 e3g4 e5g7 c1g5 d8d7 g4h6 f7f8 d1d2 e8e5 h3h4 e5e2 a1d1 e2d2 d1d2 f8e8 g1h2 a7a5 h6g4 e8f7 g5e3');
INSERT INTO evaluations VALUES(12521,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12521,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12521,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12521,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(12521,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(12521,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12521,36,'cp 11','d7d5 e4d5 f6d5 d2d4 c5d4 d1b3 c8e6 g5e6 f7e6 b3b7 a8c8 e1g1 f8e7 b7a6 e8g8 b1d2 d8b6 d2e4 c6a5 c4b5 b6a6 b5a6 c8c6 a6e2 d4c3 b2c3 g8f7 a1b1 d5c3 e4c3 c6c3 c1e3 f8c8 e2a6 c8c7 f1c1 c3c2 c1c2 c7c2 e3a7 c2a2');
INSERT INTO evaluations VALUES(40394,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(40394,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(40394,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(40394,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(40394,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(40394,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(40394,36,'cp 71','f2f3 b8c6 d1d2 e8g8 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 f1e2 c8e6 g4h5 f6h5 d4g7 g8g7 f3f4 h5g3 h1g1 g3e2 d2e2 a8c8 e1c1 f8h8 f4f5 e6a2 e2e3 f7f6 g1g6 g7f8 h4h5 a2f7 h5h6 a5a1 c1d2 a1b2 d1h1 b2b4 g6g3 f8e8 h1b1 b4c5 e3c5 c8c5 b1b7 h8h6');
INSERT INTO evaluations VALUES(40396,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(40396,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(40396,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(40396,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(40396,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(40396,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(40396,36,'cp 55','e8g8 d1d2 b8c6 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4h5 f6h5 d4g7 g8g7 e1c1 f8h8 f3f4 c8g4 d1e1 a8d8 f4f5 h5f6 f1d3 h8h5 h1g1 d8h8 f5g6 f7g6 e4e5 d6e5 d2g2 h8c8 g2b7 c8c7 b7g2 h5h4 g2g3 h4h3 g3e5 a5e5 e1e5 h3h5 e5e3 g7f7 c3e4 f6e4 e3e4 g4f5');
INSERT INTO evaluations VALUES(36014,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(36014,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(36014,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(36014,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(36014,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(36014,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(36014,36,'cp 60','d1d2 b8c6 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4h5 f6h5 d4g7 g8g7 e1c1 f8h8 f3f4 c8g4 d1e1 a8d8 f1c4 h5f6 c4b3 d6d5 e4d5 e7e6 d2d4 e6d5 e1e7 a5b6 d4b6 a7b6 e7b7 d8d7 b7d7 g4d7 b3d5 f6d5 c3d5 d7c6 c2c4 b6b5 h1h3 b5c4 d5e3 c6b5 f4f5');
INSERT INTO evaluations VALUES(36015,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(36015,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(36015,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(36015,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(36015,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(36015,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(36015,36,'cp 79','b8c6 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4h5 f6h5 d4g7 g8g7 e1c1 a5e5 h1g1 c8e6 c3e2 f8h8 g1g5 e5h2 f3f4 h2h4 f4f5 e6c4 e2d4 c4f1 d1f1 h5g3 d4e6 f7e6 d2c3 g7f8 g5g3 g6f5 e4f5 e6f5 g3g1 f8e8 c3c7 a8d8 c7b7 h8f8 b7a7 h4c4 f1e1 d8d7 a7b6 c4a2 b6b8 d7d8 b8b5 d8d7');
INSERT INTO evaluations VALUES(51985,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(51985,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(51985,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(51985,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(51985,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51985,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(51985,36,'cp 32','d2d4 d7d5 e4e5 b8c6 g1f3 c8d7 f1e2 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 a2a3 f8e7 b2b4 e8g8 c1b2 h7h6 g1h1 b7b5 e2d3 f5d4 f3d4 c6d4 c3d5 d4c6 d5f4 a8c8 d1f3 e7g5 f4h3 d7e8 h3g5 h6g5 g2g3');
INSERT INTO evaluations VALUES(48383,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(48383,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(48383,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(48383,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(48383,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(48383,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(48383,36,'cp 60','e4e5 f6d7 c2c3 c7c5 f1d3 c5d4 c3d4 b8c6 g1f3 d8b6 e1g1 a7a5 f1e1 f8e7 h2h3 a5a4 a2a3 c6d4 f3d4 b6d4 d2f3 d4b6 c1e3 d7c5 a1c1 b6a5 c1c3 c5d3 d1d3 c8d7 e1c1 a5d8 h3h4 d7c6 h4h5 h7h6 d3d4 e8g8 d4f4 e7g5 f3g5 d8g5 f4g5 h6g5 e3g5');
INSERT INTO evaluations VALUES(48431,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(48431,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(48431,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(48431,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(48431,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(48431,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(48431,36,'cp 40','f6d7 c2c3 c7c5 f1d3 c5d4 c3d4 b8c6 g1f3 d8b6 e1g1 a7a5 f1e1 a5a4 h2h3 f8e7 a1b1 c6b4 d2f1 d7f8 c1g5 b4d3 d1d3 e7g5 f3g5 f8g6 d3a3 b6d4 g5f3 d4b6 f1g3 h7h5 b1c1 c8d7 c1c2 h5h4 g3e2 f7f6 e2c1 e8f7 c1d3 a8c8 d3c5 b6a7 c2c1 b7b6 c5d3 c8c1 d3c1 g6e5 f3e5 f6e5 c1d3');
INSERT INTO evaluations VALUES(49170,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(49170,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(49170,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(49170,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(49170,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(49170,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(49170,36,'cp 46','g1f3 c7c5 c2c3 b8c6 f1d3 c5d4 c3d4 d8b6 e1g1 a7a5 h2h3 c6d4 f3d4 b6d4 d2f3 d4b6 a2a3 a5a4 d3c2 d7c5 d1d4 c8d7 c1e3 f8e7 f1c1 e8g8 f3g5 g7g6 h3h4 b6d8 d4g4 h7h5 g4g3 d5d4 e3d4 e7g5 h4g5 c5b3 c2b3 a4b3 c1c3 d7c6 a1d1 a8c8 g3e3');
INSERT INTO evaluations VALUES(49167,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(49167,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(49167,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(49167,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(49167,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(49167,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(49167,36,'cp 40','c7c5 c2c3 d8b6 f1d3 c5d4 c3d4 b8c6 e1g1 a7a5 f1e1 f8e7 h2h3 a5a4 a1b1 c6b4 d2f1 d7f8 f1g3 f8g6 h3h4 c8d7 d3g6 f7g6 a2a3 b4c6 b2b4 a4b3 b1b3 b6a7 h4h5 g6h5 g3h5 e8g8 c1e3 c6a5 b3b1 d7e8 h5f4 e8d7 f3d2 a5c6 d2f1 e7a3');
INSERT INTO evaluations VALUES(44009,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(44009,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(44009,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(44009,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(44009,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(44009,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(44009,36,'cp 41','c2c3 d8b6 f1d3 b8c6 e1g1 c5d4 c3d4 a7a5 h2h3 f8e7 f1e1 a5a4 a2a3 c6d4 f3d4 b6d4 d2f3 d4b6 c1e3 d7c5 a1c1 c8d7 d3b1 b6b5 h3h4 a8c8 e1e2 e8g8 e2c2 b7b6 f3d4 b5a6 d1h5 g7g6 h5h6 c5d3 c2c8 f8c8 c1c8 d7c8 d4c6 e7f8 h6g5 a6b7 b1d3 b7c6 g5f6 c6c7 h4h5');
INSERT INTO evaluations VALUES(43998,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(43998,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(43998,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(43998,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43998,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(43998,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(43998,36,'cp 37','c5d4 c3d4 b8c6 f1d3 a7a5 e1g1 d8b6 f1e1 f8e7 h2h3 a5a4 e1e2 b6a7 a2a3 c6d4 f3d4 a7d4 d2f3 d4a7 c1d2 d7c5 d3c2 c5b3 c2b3 a4b3 a1c1 c8d7 d1b3 d7a4 b3e3 a7a6 e2e1 e8g8 d2b4 e7b4 a3b4 f8c8 f3d4 a4d7 e3f4 c8c1 e1c1');
INSERT INTO evaluations VALUES(50522,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(50522,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(50522,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(50522,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(50522,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(50522,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(50522,36,'cp 38','d2d4 c5d4 f3d4 b8c6 d4c6 b7c6 f1d3 d7d5 e1g1 g8f6 d1f3 f8e7 f3g3 f6h5 g3g4 g7g6 c1h6 d5d4 c3a4 e6e5 g4f3 e7g5 d3c4 a8a7 h6g5 d8g5 f3a3 g5e7 a3c5 e7c5 a4c5 e8e7 g2g3 f7f6 f2f4 e7d6 c5d3 e5f4 g3f4 a7e7 e4e5 f6e5 f4e5 d6c7 a1c1 h5g7 c1e1 h8e8 f1f6 g7f5');
INSERT INTO evaluations VALUES(50514,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(50514,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(50514,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(50514,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(50514,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(50514,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(50514,36,'cp 36','c5d4 f3d4 d8c7 f1d3 b8c6 d4c6 b7c6 d1e2 c8b7 e1g1 d7d5 c1g5 f8d6 f2f4 g8e7 e4e5 d6c5 g1h1 e7g6 e2g4 c5e7 a1e1 e7g5 g4g5 e8g8 f1f3 c6c5 f3h3 d5d4 c3e4 b7e4 d3e4 c7a5 g5g3 a8b8 b2b3 c5c4 b3c4');
INSERT INTO evaluations VALUES(50567,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50567,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(50567,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(50567,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(50567,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(50567,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(50567,36,'cp 45','f3d4 b8c6 d4c6 b7c6 f1d3 d7d5 e1g1 g8f6 d1f3 f8e7 f3g3 e8g8 c1h6 f6e8 h6f4 e7h4 g3f3 h4g5 e4e5 c6c5 b2b3 g5f4 f3f4 d8c7 f1e1 g7g6 f4h4 a8b8 h4g5 e8g7 a1d1 c8d7 g2g4 c7a5 c3e2 d7b5 c2c4 d5c4 b3c4 b5c6 e2g3');
INSERT INTO evaluations VALUES(51813,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(51813,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(51813,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(51813,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(51813,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(51813,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(51813,36,'cp 111','d2d4 d7d5 b1d2 g8f6 f1d3 c6c5 e4e5 f6d7 c2c3 a7a5 e1g1 b8c6 f1e1 c5d4 c3d4 d8b6 a2a4 c6d4 f3d4 b6d4 d2f3 d4b6 c1e3 f8c5 a1c1 c5e3 e1e3 d7c5 d3b5 c8d7 b5d7 c5d7 e3b3 b6d8 b3b7 a8c8 c1c8 d8c8 b7a7 e8g8 h2h3 h7h6 a7a5 c8c7 b2b4 f8c8 a5b5 c7c4 d1d4 c4d4 f3d4 d7e5 f2f3');
INSERT INTO evaluations VALUES(19386,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(19386,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(19386,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19386,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(19386,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(19386,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(19386,36,'cp -1','g7g6 d2d3 f8g7 g2g3 a7a6 a2a4 e7e6 f1g2 d7d6 g1e2 g8e7 e1g1 e8g8 a1b1 c8d7 c1g5 a8b8 b2b3 b7b6 g5e3 d8c7 d3d4 e6e5 d4d5 c6d4 e3g5 f7f6 g5e3 f6f5 d1d2 f5e4 c3e4 e7f5 d2d3 d4e2 d3e2 f5d4 e2d1 d7f5 g1h1 b6b5');
INSERT INTO evaluations VALUES(19248,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(19248,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(19248,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(19248,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19248,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(19248,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19248,36,'cp 0','d2d3 f8g7 g2g3 a7a6 a2a4 e7e6 g1e2 d7d6 f1g2 g8e7 e1g1 e8g8 a1b1 c8d7 c1g5 d8b6 b2b3 a8e8 g5e3 c6d4 e3f4 g7e5 f4h6 e5g7');
INSERT INTO evaluations VALUES(19249,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19249,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19249,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19249,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19249,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19249,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(19249,36,'cp -15','d7d6 g2g3 f8g7 f1g2 a7a6 g1e2 a8b8 a2a4 c8g4 e1g1 g4e2 c3e2 g8f6 d3d4 c5d4 e2d4 f6d7 c1e3 e8g8 b2b3 c6d4 e3d4 d8a5 a1b1 g7d4 d1d4 f8c8 g2h3 d7e5 f2f4 e5c6 d4d5 c8d8 d5a5 c6a5 g1f2 e7e6');
INSERT INTO evaluations VALUES(19663,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19663,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19663,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19663,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(19663,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19663,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19663,36,'cp 0','g2g3 a7a6 g1e2 a8b8 f1g2 f8g7 c1e3 g8f6 a2a4 f6g4 e3c1 e7e6 e1g1 e8g8 a1b1 b7b6 b2b3 g4f6 c1e3 f6g4');
INSERT INTO evaluations VALUES(4991,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(4991,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(4991,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(4991,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(4991,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(4991,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(4991,36,'cp 41','f1d3 d7d6 e1g1 g8f6 b1c3 a7a5 c1g5 h7h6 g5h4 g7g5 h4g3 h6h5 h2h3 f6d7 g1h1 g5g4 h3h4 d7e5 d1d2 d8f6 a1e1 a5a4 c3e2 c5b6 b2b3 a4b3 a2b3 e8f8 b3b4 f8g7 e2f4 f6h6 h1g1 a8a2 d3e2 h6f6');
INSERT INTO evaluations VALUES(4988,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(4988,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(4988,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(4988,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(4988,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(4988,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(4988,36,'cp 34','d8h4 d1e2 g8f6 c1e3 c5b6 h2h3 d7d5 e3b6 a7b6 e4d5 e8d8 e1g1 h8e8 e2d2 f6d5 b1c3 h4f4 f1d1 f4d2 d1d2 c8b7 d3c4 e8e5 a1d1 d8e8 c3e2 e8f8 c4b3 e5e7 a2a4 c6c5 b3d5 b7d5 e2c3 c7c6 c3d5 c6d5 d2d5');
INSERT INTO evaluations VALUES(57168,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(57168,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(57168,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(57168,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(57168,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(57168,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(57168,36,'cp -10','g8f6 b1c3 b8c6 g1f3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 c8g4 c1e3 c6e7 b5d3 e7d5 e3d2 d6f4 d3e4 c7c6 f1e1 f4d2 d1d2 d5f6 f3e5 f6e4 e1e4 g4f5 e4e2 f7f6 e5c4 f5g4 f2f3 g4h5 c4e3 h5f7 c2c4 d8d7 d4d5 a8d8 a1d1 c6d5 c4d5 f8e8');
INSERT INTO evaluations VALUES(49675,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(49675,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(49675,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(49675,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(49675,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(49675,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49675,36,'cp -15','b1c3 b8c6 g1f3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 h7h6 b5d3 d8f6 h2h3 c8f5 d3f5 f6f5 c2c4 f8e8 c1e3 a8d8 d1a4 f5d7 a4c2 b7b6 f1d1 d6f8 d4d5 c6a5 a3a4 c7c6 e3d2 a5b7 d2f4 f8d6 f4d6 d7d6');
INSERT INTO evaluations VALUES(49672,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(49672,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(49672,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49672,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(49672,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(49672,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(49672,36,'cp -6','b8c6 g1f3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 e8g8 e1g1 h7h6 b5d3 d8f6 c1e3 f8e8 h2h3 c8f5 d3f5 f6f5 c2c4 d6f8 d1a4 f5d7 a4c2 b7b6 f1d1 c6e7 e3d2 a7a5 a1b1 d7f5 c2f5 e7f5 g2g4 f5e7 d1e1 e7g6');
INSERT INTO evaluations VALUES(16583,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(16583,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(16583,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(16583,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(16583,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16583,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16583,36,'cp 0','g1f3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 e8g8 e1g1 h7h6 b5d3 d8f6 c1e3 c8g4 h2h3 g4h5 d3e4 f6e6 e4d3');
INSERT INTO evaluations VALUES(31068,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(31068,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(31068,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(31068,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(31068,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(31068,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(31068,36,'cp 152','d1b3 d8b6 b3b6 a7b6 c4d5 g8f6 d5c6 b8c6 e2e3 c6b4 e1d2 f6e4 c3e4 f5e4 f1b5 e8d8 f2f3 e4g6 h2h4 a8a2 a1a2 b4a2 g1e2 f7f6 e2c3 a2c3 b2c3 e7e6 h4h5 g6e8 b5c4 e8d7 d2e2 d8c7 c1d2 f8e7 h1a1 f6f5 a1b1 e7d6 e3e4 f5e4 f3e4 e6e5 c4d5 d7c6 h5h6 g7h6');
INSERT INTO evaluations VALUES(31064,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(31064,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(31064,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(31064,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(31064,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(31064,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(31064,36,'cp 149','d8b6 b3b6 a7b6 c4d5 g8f6 d5c6 b8c6 e2e3 c6b4 f1b5 e8d8 e1d1 f5c2 d1d2 f6e4 c3e4 c2e4 f2f3 e4g6 h2h4 a8a2 a1a2 b4a2 g1e2 f7f6 e2c3 a2c3 b2c3 e7e6 h4h5 g6e8 b5d3 d8c7 d2e2 h7h6 c1d2 f8d6 g2g4 e6e5 h1a1 c7b8 c3c4 e8f7 d2c3 h8d8 d4d5 d6c5 a1b1');
INSERT INTO evaluations VALUES(53071,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53071,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53071,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(53071,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(53071,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53071,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(53071,36,'cp 0','c2c3 g8f6 g2g3 b8c6 f1g2 e7e5 c3c4 g7g6 b1c3 a8b8 g1e2 f8g7 e1g1 e8g8 c1e3 f6g4 e3d2 g4f6 a1b1 a7a6 a2a4 c6b4 d2g5 h7h6 g5e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(53064,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53064,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53064,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53064,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53064,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53064,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53064,36,'cp 0','g8f6 g2g3 e7e5 c3c4 b8c6 f1g2 a8b8 b1c3 g7g6 g1e2 f8g7 e1g1 e8g8 c1e3 b7b6 a1b1 a7a6 a2a4 g8h8 h2h3 c8b7 d1d2 f6g8 f2f4 f7f5 c3d5 g8f6 b1e1 c6d4 e2d4 e5d4 e3f2 b7c6 b2b3 f6d5 c4d5 c6d7 g1h1 b6b5');
INSERT INTO evaluations VALUES(45203,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45203,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45203,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(45203,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(45203,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45203,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(45203,36,'cp 0','g2g3 e7e5 c3c4 g7g6 f1g2 f8g7 g1e2 b8c6 b1c3 a8b8 a2a4 e8g8 e1g1 f6e8 c1e3 e8c7 d1d2 a7a6 c3d5 b7b5 a4b5 a6b5 e3g5 f7f6 g5e3 b5c4 d3c4 g8h8 d5c7 d8c7 e2c3 c6d4 c3d5 c7f7 e3d4 e5d4');
INSERT INTO evaluations VALUES(51431,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(51431,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(51431,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(51431,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(51431,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(51431,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(51431,36,'cp 111','d2d4 c8b7 f1d3 c7c5 c2c3 g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 f8e7 a2a3 a7a5 a3a4 b7a6 d3b5 e8g8 c1f4 a6b5 a4b5 f8e8 g2g3 d7f8 h2h4 b8d7 h4h5 h7h6 b1a3 f7f6 a3c2 f6e5 f4e5 d7e5 f3e5 e7d6 f2f4 d6e5 d4e5 f8d7 c2e3 a5a4 e2g4 g8h8 c3c4 d5c4');
INSERT INTO evaluations VALUES(48819,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(48819,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(48819,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(48819,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(48819,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(48819,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(48819,36,'cp 109','e4d5 f8d6 f1b5 c7c6 d5c6 b8c6 d2d3 e8g8 b5c6 b7c6 c1g5 a8b8 d1c1 c8g4 f3d2 g4e6 d2e4 d6e7 e4f6 g7f6 g5h4 g8h8 e1g1 f8g8 g1h1 d8d4 c3e4 b8b2 h4f6 e7f6 e4f6 g8g6 f6e4 b2a2 e4g3 e6d5 f2f3 a2a1 c1a1 g6g8 f1b1 d4a1 b1a1');
INSERT INTO evaluations VALUES(48601,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(48601,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(48601,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(48601,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(48601,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(48601,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(48601,36,'cp 92','f8d6 f1b5 c7c6 d5c6 b8c6 d2d3 h7h6 e1g1 e8g8 c3e4 f6e4 d3e4 c8e6 c1e3 d8c7 c2c3 f8d8 f3d2 c6a5 b5e2 d6c5 e3c5 c7c5 d1c2 a8c8 a1d1 a5c6 c2a4 c6a5 f1e1 a7a6 h2h3 c5c7 d2f1 d8d1 e1d1 a5c4 e2c4 c7c4 a4c4 c8c4 f1d2 c4c6 c3c4 b7b5 c4b5 a6b5');
INSERT INTO evaluations VALUES(39254,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39254,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(39254,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(39254,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39254,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39254,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(39254,36,'cp 0','g1f3 d7d6 b1c3 a7a5 c1e3 c5e3 f2e3 g8f6 d3d4 d8e7 f1d3 e8g8 d1d2 c7c6 e1g1 b8d7 d2e1 b7b5 a2a3 f8e8 h2h3 c8a6 e1h4 d7f8 h4f2 f8d7');
INSERT INTO evaluations VALUES(39246,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(39246,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(39246,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(39246,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(39246,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39246,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39246,36,'cp 0','d7d6 b1c3 a7a5 c1e3 c5e3 f2e3 g8f6 d3d4 d8e7 f1d3 e8g8 h2h3 b8d7 e1g1 d6d5 f3e5 d7e5 d4e5 e7e5 c3d5 f6d5 e4d5 c8h3 g2h3 e5g3 g1h1 g3h3 h1g1 h3g3 g1h1');
INSERT INTO evaluations VALUES(38731,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38731,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38731,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38731,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38731,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(38731,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38731,36,'cp 6','b1c3 a7a5 c1e3 c5e3 f2e3 g8f6 d3d4 d8e7 f1d3 e8g8 h2h3 b8d7 e1g1 d6d5 f3d2 d5e4 d2e4 f6e4 d3e4 d7f6 e4d3 e5d4 e3d4 f8d8 d1f3 d8d4 a1e1 c8e6 f3b7 a8d8 a2a3 e7c5 g1h1 d4d7 b7f3 f6h5 c3e4');
INSERT INTO evaluations VALUES(38726,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(38726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38726,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(38726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38726,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(38726,36,'cp 0','a7a5 c1e3 c5e3 f2e3 g8f6 d3d4 d8e7 f1d3 e8g8 h2h3 b8d7 e1g1 d6d5 f3d2 d5e4 d2e4 f6e4 d3e4 d7b6 e4d3 f7f5 d4e5 e7e5 d1f3 c7c6 e3e4 f5f4 f3f2 f4f3 f2h4 e5d4 g1h2 d4e5');
INSERT INTO evaluations VALUES(10054,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10054,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(10054,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(10054,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10054,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10054,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10054,36,'cp -12','h7h6 g5e3 c5b6 e1g1 e8g8 b1c3 d7d6 a2a4 c8e6 c4e6 f7e6 c3e2 d8e8 c2c3 b6e3 f2e3 c6e7 a4a5 e8d7 e2g3 f8f7 d1b3 f6g4 a1e1 b7b6 h2h3 g4f6 d3d4 e7g6 d4e5 d6e5 e1a1 a8f8 a5b6');
INSERT INTO evaluations VALUES(9735,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(9735,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(9735,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(9735,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(9735,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(9735,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9735,36,'cp 0','g5e3 c5b6 e1g1 e8g8 b1c3 d7d6 a2a4 c8e6 e3b6 a7b6 f1e1 e6c4 d3c4 f6d7 c3d5 c6e7 b2b4 e7d5 c4d5 f7f5 e1e3 d7f6 f3d2 f6g4 e3h3 g4f6');
INSERT INTO evaluations VALUES(54192,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54192,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(54192,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54192,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(54192,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54192,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(54192,36,'cp 15','b1d2 d8b6 g1f3 g8f6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 f6d5 b1c1 c5e7 f1c4 b8c6 d1c2 d5b6 c4d3 h7h6 a2a3 c8d7 e1g1 f8d8 f1d1 d7e8 d2e4 a8c8 c2b1 a7a5 d3e2 a5a4 h2h3 c6a5 c1c8 b6c8 d1d8 e7d8 b1c2 e8c6');
INSERT INTO evaluations VALUES(45438,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45438,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45438,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45438,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45438,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45438,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45438,36,'cp -6','c8g4 h2h3 g4f3 d1f3 b8c6 c1e3 e7e6 f1d3 f8d6 c3e2 c6b4 e2f4 b4d3 f4d3 a8c8 c2c3 f6e4 e3f4 d8f6 f3e3 d6f4 d3f4 e8g8 e1g1 e4d6 g2g3 a7a5 f1e1 h7h6 g1g2 f8e8 h3h4 c8c7 h4h5 e8c8');
INSERT INTO evaluations VALUES(29168,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29168,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29168,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29168,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(29168,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(29168,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29168,36,'cp -3','h2h3 g4f3 d1f3 b8c6 c1e3 e7e6 f1d3 f8d6 c3e2 c6b4 e2c1 a8c8 c2c3 b4d3 c1d3 f6e4 e3f4 d8f6 f3e3 d6f4 d3f4 a7a5 a2a4 e8g8 e1g1 f6h6 f1e1 b7b6 e1e2 c8c7 e3f3 h6f6 g2g3 f6f5 g1g2 g7g5 f4d3 f5f3 g2f3 h7h5 f3g2 f8c8');
INSERT INTO evaluations VALUES(29165,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29165,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(29165,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29165,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(29165,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(29165,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29165,36,'cp 0','g4f3 d1f3 b8c6 c1e3 e7e6 f1d3 f8d6 c3e2 e8g8 c2c3 e6e5 e3g5 e5e4 g5f6 d8f6 f3f6 g7f6 d3c2 f6f5 f2f3 c6e7 e1c1 b7b5 d1f1 b5b4 f3e4 f5e4 f1f6 a8d8 h1f1 b4c3 b2c3 g8g7 c1d2 e7g8 f6f5 g8e7');
INSERT INTO evaluations VALUES(29181,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29181,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29181,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29181,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29181,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29181,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29181,36,'cp 0','d1f3 b8c6 c1e3 e7e6 f1d3 f8d6 c3e2 c6b4 e2c1 b4d3 c1d3 f6e4 e3f4 h7h6 f4d6 e4d6 c2c3 d8f6 f3g4 f6g5 g4f3');
INSERT INTO evaluations VALUES(29180,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29180,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29180,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29180,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29180,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29180,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29180,36,'cp 0','b8c6 c1e3 e7e6 f1d3 f8d6 c3e2 c6b4 e2c1 b4d3 c1d3 f6e4 e3f4 h7h6 f4d6 e4d6 c2c3 d8f6 f3g4 f6g5 g4f3');
INSERT INTO evaluations VALUES(19920,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(19920,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19920,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(19920,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19920,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(19920,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19920,36,'cp 19','e7e5 d2d3 a7a6 b5c4 b7b5 c4b3 c8e6 e1g1 g8f6 c3d5 e6d5 b3d5 f6d5 e4d5 c6e7 c2c4 g7g6 a1b1 f8g7 b2b4 c5b4 b1b4 e8g8 d1b3 b5c4 d3c4 e7f5 c1g5 d8c8 f1b1 a6a5 b4b7 a5a4 b3d3 h7h6 g5e3 f5e3 f2e3 f7f5 e3e4 f5e4 d3e4 c8f5 e4f5');
INSERT INTO evaluations VALUES(20574,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(20574,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(20574,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(20574,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(20574,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(20574,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(20574,36,'cp 27','d2d3 a7a6 b5c4 h7h6 e1g1 g7g6 a2a3 f8g7 c3d5 g8e7 b2b4 c5b4 d5b4 c8d7 b4c6 d7c6 c4b3 b7b5 a3a4 b5a4 b3a4 c6a4 a1a4 d8d7 c2c4 e8g8 a4a1 e7c6 c1e3 a6a5 d1a4 d7c7 f1b1 f8b8 b1b8 a8b8 h2h3');
INSERT INTO evaluations VALUES(20575,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(20575,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(20575,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(20575,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(20575,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(20575,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(20575,36,'cp 15','a7a6 b5c4 b7b5 c4b3 c8e6 e1g1 g8f6 c3d5 f8e7 c2c3 e8g8 f1e1 a8b8 d5e7 d8e7 c1g5 e6b3 a2b3 b8a8 h2h3 a6a5 g5h4 f8c8 f3h2 e7e6 h2f1 c8b8 f1e3 b5b4 h4f6 e6f6 e3d5 f6d8 f2f4 c6e7 d5e7 d8e7');
INSERT INTO evaluations VALUES(22732,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(22732,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22732,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(22732,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(22732,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22732,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(22732,36,'cp 7','g8f6 e2e4 f8b4 d2d3 d7d6 f1e2 e8g8 a2a3 b4c3 b2c3 h7h6 e1g1 f6h7 a3a4 g8h8 d3d4 f7f5 e4f5 c8f5 a4a5 h7f6 f1e1 e5e4 f3h4 f5h7 f2f4 d8e7 f4f5 b7b6 a5b6 a7b6 c1f4 a8a1 d1a1 c6d8 f4g3 d8f7 h4g6 h7g6 f5g6');
INSERT INTO evaluations VALUES(53284,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(53284,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53284,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53284,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(53284,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53284,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53284,36,'cp 20','e4d5 d8d5 d2d4 g8f6 g1f3 e7e6 f1e2 c5d4 d1d4 f8e7 d4d5 f6d5 e1g1 b8c6 c3c4 d5b4 b1c3 e8g8 a1b1 f8d8 a2a3 b4c2 f1d1 c6d4 f3d4 c2d4 e2d3 c8d7 b2b4 d7e8 c1e3 f7f5 c3d5 e6d5 e3d4 d5c4 d3c4 e8f7 c4f1 b7b6 d4c3 a8c8 d1d8 c8d8 b1c1 d8c8');
INSERT INTO evaluations VALUES(53232,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53232,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53232,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(53232,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(53232,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(53232,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(53232,36,'cp 6','d8d5 g1f3 g8f6 d2d4 e7e6 f1e2 c5d4 d1d4 b8c6 d4d5 f6d5 e1g1 b7b6 f1d1 f8e7 a2a3 c6a5 b1d2 c8b7 e2b5 b7c6 b5c6 a5c6 c3c4 d5f6 b2b4 e8g8 c1b2 f8d8 g1f1 a8c8 a1c1 f6e8 d2e4 d8d1 c1d1 c6b8 d1c1 f7f6 b2d4 e6e5 d4e3 b8d7 f3d2 g8f7');
INSERT INTO evaluations VALUES(34368,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(34368,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(34368,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(34368,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(34368,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(34368,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(34368,36,'cp 22','g1f3 e7e6 d2d4 g8f6 f1e2 c5d4 d1d4 b8c6 d4d5 f6d5 e1g1 b7b6 c3c4 d5b4 b1c3 b4c2 a1b1 c2d4 f3d4 c6d4 e2d3 c8b7 c1e3 e8c8 f1d1 f8e7 b2b4 d4c6 c4c5 c6e5 d3e2 d8d1 b1d1 b6c5 b4c5 h8d8 d1b1 a7a6 c3a4 b7e4 e2a6 c8c7 b1a1 e4c6 a4c3');
INSERT INTO evaluations VALUES(34367,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(34367,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(34367,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(34367,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(34367,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(34367,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(34367,36,'cp 19','g8f6 d2d4 e7e6 b1a3 b8c6 c1e3 c5d4 a3b5 d5d8 b5d4 f6d5 d4c6 b7c6 d1d4 d8b6 e1c1 a8b8 d1d2 b6d4 f3d4 c8d7 d4c2 d5e3 c2e3 e8e7 g2g3 g7g6 f2f4 f8g7 f1c4 h8d8 h1d1 d7e8 d2d8 b8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(33609,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33609,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(33609,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(33609,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(33609,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(33609,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(33609,36,'cp 10','d2d4 e7e6 f1e2 c5d4 c3d4 d5d6 e1g1 b8c6 b1c3 f8e7 c3b5 d6d8 c1f4 f6d5 f4g3 e8g8 a1c1 c8d7 b5c3 a8c8 c3d5 e6d5 d1b3 d8b6 b3b6 a7b6 g3f4 h7h6 h2h3 d7e6 c1c3 g7g5 f4e3 c6a5 f1c1 a5c4 c3b3 c4e3 c1c8 f8c8 f2e3 g8g7');
INSERT INTO evaluations VALUES(33607,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(33607,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(33607,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(33607,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(33607,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(33607,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(33607,36,'cp 15','e7e6 b1a3 b8c6 c1e3 c5d4 a3b5 d5d8 b5d4 d8c7 d1a4 f6d5 e1c1 f8c5 d4b5 c5e3 f2e3 c7e7 e3e4 d5b6 a4b3 e8g8 f1e2 c8d7 b3a3 e7a3 b5a3 c6e7 e2b5 d7c6 b5c6 e7c6 a3b5 f8d8 b2b3 g8f8 d1d8 a8d8 h1d1 d8d1 c1d1 b6c8 d1e2 f8e7 b5d4 c6d4 c3d4 f7f5 e2d3 f5e4 d3e4 e7f6 a2a4 c8b6 f3e5 b6d5');
INSERT INTO evaluations VALUES(33623,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(33623,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(33623,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(33623,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(33623,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33623,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(33623,36,'cp 13','f1e2 c5d4 c3d4 d5d6 e1g1 b8c6 b1c3 f8e7 c3b5 d6d8 c1f4 f6d5 f4g3 e8g8 a1c1 c8d7 b5c3 a8c8 e2d3 e7f6 f1e1 c6e7 g3d6 d5c3 b2c3 d7c6 d6b4 f8e8 c3c4 e7g6 d3f1 h7h6 a2a3 c6f3 d1f3 d8d4 f3b7');
INSERT INTO evaluations VALUES(30767,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30767,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(30767,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(30767,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(30767,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(30767,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(30767,36,'cp 33','g8f6 b1c3 c7c6 f1e2 c6d5 e4d5 g7g6 h2h4 h7h5 c1g5 f8g7 d1d2 a7a6 g5f6 g7f6 e1c1 d7d6 g2g4 h5g4 e2g4 b7c8 g4f3 c8f5 h4h5 b8d7 g1e2 d8c7 e2d4 c7c5 d2e3 f6d4 d1d4 a8c8 h5g6 h8h1 f3h1 f5g6 h1e4 g6e4 e3e4 d7e5');
INSERT INTO evaluations VALUES(28784,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(28784,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(28784,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(28784,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(28784,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28784,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(28784,36,'cp 34','b1c3 c7c6 f1e2 c6d5 e4d5 e7e6 e2f3 d8c8 g1e2 f8e7 b2b3 e8g8 c1b2 b8a6 d1d2 a8b8 e1g1 a6b4 e2f4 c8c7 f1d1 f6d5 f4d5 b4d5 f3d5 b7d5 c3d5 e6d5 a1c1 f8e8 d2d5 e7c5 d1d2 e8e6 c1d1 b8e8 g2g3 c7c6 d5c6 d7c6 g1f1 c5b4 c2c3 b4f8 c3c4 f8c5');
INSERT INTO evaluations VALUES(28781,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(28781,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(28781,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(28781,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(28781,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28781,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(28781,36,'cp 33','c7c6 c1g5 c6d5 g5f6 e7f6 c3d5 f8c5 d1f3 e8g8 e1c1 b8c6 g1e2 f8e8 h2h4 a8c8 c1b1 c6e5 f3g3 f6f5 e4f5 b7d5 d1d5 e5c4 e2c3 e8e1 d5d1 c4d2 b1c1 e1d1 c3d1 d2f1 h1f1 c5e7 g3f3 e7h4 d1c3 h4g5 c1b1 d8c7 f3e4 g8f8 c3d5 c7c4');
INSERT INTO evaluations VALUES(28626,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(28626,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(28626,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(28626,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(28626,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(28626,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(28626,36,'cp 18','f1e2 b8a6 e2f3 e7e6 g1e2 f8d6 c1g5 d8c7 d1d2 d6e5 g5f4 e5f4 e2f4 c6d5 e4d5 c7e5 f4e2 f6d5 c3d5 b7d5 f3d5 e5d5 d2d5 e6d5 e1c1 a6c7 e2c3 e8c8 c3d5 c7d5 d1d5 h8e8 c1d2 g7g6 h1d1 c8c7 g2g3 e8e6 d5d3 d7d5 a2a3 e6e5 d3f3 f7f5 d2d3 d8c8 d3c3');
INSERT INTO evaluations VALUES(56022,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(56022,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(56022,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(56022,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(56022,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(56022,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(56022,36,'cp 31','e2e3 c7c5 f1c4 a7a6 e1g1 g8f6 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 f1d1 c8d7 b2b3 d7b5 c1b2 b8d7 h2h3 h8c8 g2g4 h7h6 a1c1 e7f8 c4f1 c5e7 d2c4 b5c4 f1c4 c8c7 c4d3 a8c8 g1f1 c7c1 d1c1 c8c1 b2c1 f6d5');
INSERT INTO evaluations VALUES(56021,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(56021,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(56021,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56021,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(56021,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(56021,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(56021,36,'cp 29','g8f6 f1c4 c7c5 e1g1 a7a6 d4c5 f8c5 d1d8 e8d8 b2b3 b8d7 c1b2 b7b6 a2a3 c8b7 b1d2 d8e7 c4e2 b6b5 a3a4 h8b8 b2c3 b5b4 c3b2 b8c8 a1c1 f6d5 d2c4 a6a5 f1d1 f7f6 c1c2 d7b6 d1c1 d5c3 b2c3 b4c3 c4b6 c5b6 c2c3');
INSERT INTO evaluations VALUES(17269,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(17269,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(17269,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(17269,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(17269,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17269,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(17269,36,'cp 1','e7e5 f1b5 c6d4 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d1e2 f8e7 d2c3 e8g8 c1e3 c7c6 b5d3 d7d5 e1g1 f8e8 e3d4 f6h6 h2h3 c8e6 e2e5 e7f6 e5c7 e8e7 c7g3 f6d4 c3d4 h6f6 c2c3 e6f5 d3f5 f6f5 a1e1 f5f6 e1e5 a8e8 f1e1 e7e5 d4e5 f6e6');
INSERT INTO evaluations VALUES(53908,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(53908,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(53908,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(53908,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(53908,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(53908,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(53908,36,'cp 106','d4e2 f8e7 b1c3 b8c6 f2f3 g8f6 e2f4 a7a6 c1e3 c8e6 d1d2 d8a5 f1e2 a8d8 a1d1 a5c7 e1g1 e8g8 g1h1 b7b5 a2a4 b5a4 e2a6 c6b4 a6b5 e6d7 b5c4 f8e8 d1a1 c7b7 c3a4 d7b5 b2b3 b5c4 b3c4 e7f8');
INSERT INTO evaluations VALUES(53913,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(53913,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(53913,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(53913,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(53913,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(53913,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(53913,36,'cp 94','g8f6 b1c3 a7a6 g2g3 b8d7 a2a4 a8b8 e2f4 f8e7 a4a5 e8g8 f1g2 d7e5 e1g1 e5c6 c3d5 f6d5 f4d5 c8e6 f1e1 e6d5 e4d5 c6e5 g2f1 f8e8 c1d2 e7f6 b2b3 g7g6 g1g2 f6g7 d2f4 h7h5 a1a4 h5h4 e1e3 h4g3 h2g3 e8e7 f4g5');
INSERT INTO evaluations VALUES(38360,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(38360,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38360,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(38360,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(38360,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38360,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(38360,36,'cp 35','g8f6 d1e2 e8g8 e1g1 c8g4 e4e5 d6e5 d4e5 f6d5 b1d2 d5b6 c4d3 b8c6 h2h3 c6d4 e2e4 d4f3 d2f3 g4f3 e4f3 g7e5 f3b7 e7e6 b7e4 e5g7 a2a4 a7a5 f1d1 d8e7 c2c3 f8d8 c1f4 b6d5 f4g3 e7c5 d1d2 g7h6 d2c2 h6g7 e4f3 h7h5 a1d1 c5b6 c2d2 g7h6');
INSERT INTO evaluations VALUES(41249,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(41249,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(41249,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41249,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(41249,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(41249,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(41249,36,'cp 42','d1e2 e8g8 e4e5 d6e5 d4e5 f6d5 e1g1 c8g4 b1d2 d5b6 c4d3 b8c6 h2h3 g4e6 f1e1 b6d7 d3c4 d7e5 c4e6 f7e6 f3e5 c6e5 d2c4 e5c4 e2e6 f8f7 e6c4 d8d4 c4e2 a8f8 c1e3 d4b2 a1b1 b2a2 b1b7 a2d5 b7c7 g7d4 e3d4 d5d4 e1f1 a7a5 c7e7 f7e7 e2e7 d4c4 e7d6 c4c2');
INSERT INTO evaluations VALUES(41247,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(41247,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(41247,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41247,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(41247,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(41247,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(41247,36,'cp 25','e8g8 e1g1 c8g4 e4e5 d6e5 d4e5 f6d5 b1d2 d5b6 h2h3 b6c4 e2c4 g4f5 f3d4 b8c6 d4c6 b7c6 f1e1 d8d5 c4c3 a7a5 d2f3 f5e4 c1g5 e4f3 c3f3 d5f3 g2f3 e7e6 a1d1 a5a4 d1d7 f8b8 b2b3 a8a5 f3f4 a4b3 a2b3 a5c5 d7c7 c5c2 g1g2');
INSERT INTO evaluations VALUES(43409,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(43409,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43409,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43409,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43409,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43409,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43409,36,'cp 18','e7e5 d4b3 f8e7 c1e3 c8e6 c3d5 f6d5 e4d5 e6f5 e1g1 e8g8 d1d2 f5g6 b3a5 b7b6 a5c4 b8d7 a2a4 a6a5 f2f3 h7h6 e2d3 g6d3 d2d3 d8c7 b2b3 d7c5 d3e2 a8e8 a1d1 f7f5 f1e1 g8h8 g2g3 e7g5 e3f2 g5e7 f3f4 e5f4 g3f4 e7f6 e2h5');
INSERT INTO evaluations VALUES(43453,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43453,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(43453,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(43453,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(43453,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(43453,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43453,36,'cp 27','d4b3 f8e7 c1e3 c8e6 e1g1 e8g8 d1d3 b8c6 a2a3 f8e8 a1c1 e7f8 e3g5 h7h6 g5f6 d8f6 c3d5 f6d8 f1d1 a8c8 d5e3 f8e7 e2g4 e6g4 e3g4 a6a5 b3d2 e7g5 h2h3 a5a4 c2c3 c6e7 g4e3 g5e3 d3e3 d6d5');
INSERT INTO evaluations VALUES(43462,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43462,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43462,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(43462,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(43462,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(43462,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43462,36,'cp 12','f8e7 e1g1 e8g8 e2f3 c8e6 a2a4 d8c7 c1e3 b7b6 h2h3 h7h6 f1e1 b8d7 b3d2 f8b8 d1e2 b6b5 e1c1 b5b4 c3d5 e6d5 e4d5 d7c5 e2c4 f6d7 d2e4 d7b6 c4b4 c5e4 b4e4 b6c4 b2b3 c4e3 f2e3 e7g5 c1d1 a6a5');
INSERT INTO evaluations VALUES(39677,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(39677,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39677,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(39677,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(39677,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(39677,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(39677,36,'cp 11','e1g1 e8g8 d1d3 c8e6 c1e3 b8d7 a2a4 a8c8 f1d1 d8c7 e2f3 h7h6 a4a5 d7c5 b3c5 d6c5 c3a4 f8d8 d3c3 d8d1 a1d1 f6d7 h2h3 c7c6 c3a3 b7b5 a5b6 d7b6 a4b6 c6b6 f3g4 e6g4 h3g4 c8b8 b2b3 b6e6 e3c5 e7c5 a3c5 b8c8 c5d6 c8c2 d6e6 f7e6');
INSERT INTO evaluations VALUES(39676,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(39676,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(39676,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(39676,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(39676,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(39676,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(39676,36,'cp 12','e8g8 c1e3 c8e6 d1d3 b8d7 a2a4 a8c8 a4a5 d8c7 f1d1 f8d8 e2f3 d7c5 b3c5 d6c5 d3e2 h7h6 d1c1 d8d6 h2h3 c8e8 c3a4 f6d7 c1b1 e8c8 b1d1 d6d1 e2d1 c8d8 d1e1 c7c6 e3d2 c6c7 e1e2 c5c4 g1h2 c7c6 g2g3 d7f6 f3g2');
INSERT INTO evaluations VALUES(35563,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(35563,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(35563,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(35563,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(35563,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(35563,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(35563,36,'cp 17','c1e3 c8e6 d1d2 a6a5 d2d3 b8c6 a2a3 f6e4 d3e4 d6d5 c3d5 d8d5 e4d5 e6d5 f1d1 d5b3 c2b3 c6d4 e3d4 e5d4 g2g3 g7g6 a1c1 f8c8 c1c4 e7f6 g1g2 c8c6 c4c6 b7c6 g2g1 a8c8 e2c4 c8d8 g1g2');
INSERT INTO evaluations VALUES(35559,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(35559,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(35559,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(35559,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(35559,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(35559,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(35559,36,'cp 9','c8e6 d1d2 a6a5 d2d3 a5a4 b3d2 d6d5 c3d5 f6d5 e4d5 d8d5 d3d5 e6d5 d2c4 b8d7 f1d1 d5e6 d1d2 f8c8 a2a3 g8f8 a1d1 e6c4 e2c4 c8c4 d2d7 c4c2 d7b7 a8d8 d1d8 e7d8 g2g3 d8a5 h2h4 a5d2 e3d2 c2d2 g1f1 g7g6 b7b5 f7f6 b5b4 f6f5 b4a4 d2b2 f1g2 b2a2 a4a7 f5f4 a7h7');
INSERT INTO evaluations VALUES(27744,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(27744,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(27744,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(27744,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(27744,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(27744,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(27744,36,'cp 19','d1d3 b8d7 a2a4 a8c8 a4a5 d8c7 f1d1 h7h6 e2f3 d7c5 b3c5 d6c5 c3a4 c7a5 a4c5 a5b4 c5e6 f7e6 a1b1 f8d8 d3e2 d8d1 e2d1 c8d8 d1e1 b4e1 b1e1 d8c8 c2c3 e7c5 e3d2 c8d8 e1d1 a6a5 g1f1 g8f7 f1e2 g7g5 h2h3 b7b6 d2e1 d8d1 e2d1 h6h5 b2b4 a5b4 c3b4 c5d4');
INSERT INTO evaluations VALUES(48268,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(48268,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(48268,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48268,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(48268,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(48268,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(48268,36,'cp 20','f8b4 c1d2 b4e7 f1g2 c7c6 e1g1 e8g8 d1c2 b7b6 d2f4 c8b7 f1d1 b8d7 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 d8c7 e2e3 a8d8 a1c1 d7f6 c4e5 c6c5 d4c5 b7g2 g1g2 b6c5 b2b3 f6d5 g2g1 h7h6 e5c4 d5c3 c2c3 d8d1 c1d1 f8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(42059,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(42059,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42059,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(42059,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42059,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42059,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42059,36,'cp 18','c1d2 b4e7 f1g2 c7c6 d1c2 b7b6 d2f4 c8b7 e1g1 b8d7 f1d1 e8g8 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 d8c7 e2e3 a8d8 a1c1 c6c5 d4d5 e6d5 c3d5 b7d5 d1d5 b6b5 c4e5 d7e5 d5e5 a7a6 a2a4 e7d6 e5d5 d6e7 c1d1 d8d5 d1d5 f8d8 d5d8 c7d8 a4b5 a6b5');
INSERT INTO evaluations VALUES(42058,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(42058,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42058,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42058,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(42058,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42058,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(42058,36,'cp 7','b4e7 f1g2 e8g8 d1c2 c7c6 e1g1 b7b6 d2f4 c8b7 f1d1 b8d7 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 d8c7 e2e3 a8c8 a2a4 a7a6 c3e4 f7f5 e4g3 b6b5 c4e5 d7e5 f4e5 c6c5 d4d5 b7d5 g2d5 e6d5 d1d5 c7c6 a1d1 c5c4 a4b5 a6b5 g3e2 b5b4');
INSERT INTO evaluations VALUES(40963,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40963,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40963,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(40963,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(40963,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(40963,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40963,36,'cp 10','f1g2 b8d7 e1g1 e8g8 d1c2 c7c6 f1d1 b7b6 b2b3 h7h6 d2e1 c8b7 b1c3 a8c8 e2e4 d5e4 f3e5 c8c7 c3e4 f6e4 g2e4 d8b8 c2e2 d7e5 d4e5 f8d8 d1d8 b8d8 e1c3 c7d7 e2f3 d8c8 a2a4 b6b5 a4a5 b5c4 b3c4 c6c5 e4b7 c8b7 f3b7');
INSERT INTO evaluations VALUES(40964,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(40964,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40964,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40964,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40964,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(40964,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(40964,36,'cp 7','e8g8 d1c2 c7c6 e1g1 b8d7 d2f4 b7b6 c4d5 c6d5 f1c1 c8b7 f4c7 d8e8 c2b3 f6e4 b1c3 d7f6 a2a4 e4c3 b3c3 a8c8 c3e3 f6e4 c7f4 h7h6 h2h4 e7d6 f4d6 e4d6 b2b3 e8e7 f3e5 c8c1 a1c1 f8c8 e3f4 d6f5 c1c8 b7c8 e2e3 e7b4 e5c6 b4e1 g1h2');
INSERT INTO evaluations VALUES(36334,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36334,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36334,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(36334,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36334,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36334,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36334,36,'cp 11','d1c2 b8d7 e1g1 c7c6 d2f4 b7b6 c4d5 c6d5 f1c1 c8b7 f4c7 d8e8 c2b3 f6e4 b1c3 d7f6 f3e5 a8c8 c3b5 e4d2 b3d3 d2c4 e5c4 e8b5 c4d6 b5d3 e2d3 e7d6 c7d6 c8c1 a1c1 f8a8 g1f1 f6e8 d6f4 h7h6 h2h4 a8c8 c1c8 b7c8 f1e2 c8d7 g3g4 f7f6 f4b8 a7a5 f2f4 g8f7 f4f5');
INSERT INTO evaluations VALUES(47803,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(47803,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(47803,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(47803,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(47803,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(47803,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(47803,36,'cp 15','f8e7 e1g1 e8g8 f1e1 c7c5 b1c3 b8c6 c1g5 f6e8 g5d2 c6a5 c4d5 a5c6 a2a3 e8c7 d5c4 c8g4 h2h3 g4f3 d1f3 b7b5 c4d5 c6d4 f3d1 a8b8 d5a2 e7g5 c3d5 g5d2 d1d2 a7a5 d5c7 d8c7 c2c3 d4e6 a2e6 f7e6 b2b4 a5a4 a1d1 c7f7 b4c5 d6c5');
INSERT INTO evaluations VALUES(40773,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(40773,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40773,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40773,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(40773,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(40773,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40773,36,'cp 30','e1g1 e8g8 b1c3 c8e6 a2a4 d8d7 h2h3 b8c6 c1d2 e6c4 d3c4 a7a5 d2g5 f6e8 g5e3 h7h6 c3d5 e8f6 f3d2 e7d8 b2b3 c6b4 c2c3 b4d5 c4d5 f6e4 d2e4 f7f5 e4g3 f5f4 d1e2 f4e3 f2e3 d8g5 g3e4 f8f1 a1f1 a8f8 e2d3 g5h4 c3c4 f8f1 g1f1 d7f7 f1g1 f7g6 g1h2');
INSERT INTO evaluations VALUES(40772,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40772,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40772,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40772,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40772,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40772,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40772,36,'cp 18','e8g8 a2a4 c7c5 f1e1 b8c6 c2c3 h7h6 b1a3 c8e6 c1d2 e6c4 a3c4 f8e8 d1b3 d8d7 c4e3 a8d8 c3c4 e8f8 a4a5 f6h7 d2c3 h7g5 b3d1 g5e6 h2h4 d8b8 g2g3 g7g6 f3h2 f7f5 e4f5 g6f5 d1h5 f8f7 h5h6');
INSERT INTO evaluations VALUES(36229,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36229,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(36229,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(36229,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36229,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36229,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36229,36,'cp 25','f1e1 c7c5 b1c3 b8c6 a2a4 c8e6 c1g5 c6b4 h2h3 a7a6 c3b1 f6d7 g5d2 d8c7 a4a5 h7h6 d2c3 b4c6 b1d2 g8h7 d2f1 d7f6 f1e3 g7g6 c4e6 f7e6 e3c4 f8f7 c3d2 a8f8 c2c3 f6h5');
INSERT INTO evaluations VALUES(22941,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(22941,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(22941,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(22941,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(22941,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(22941,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(22941,36,'cp -12','d4c3 b1c3 f8c5 f1c4 d7d6 e1g1 g8e7 c3a4 e7g6 a4c5 d6c5 c1e3 d8d1 f1d1 b7b6 c4b5 c8b7 b2b4 a7a6 b5c6 b7c6 b4c5 b6b5 f3d4 c6e4 f2f3 e4d5 d4b5 a6b5 d1d5 g6e7 d5d2 c7c6 a2a4 e8g8 d2a2 b5b4 a4a5 e7d5 e3d2 a8a6 h2h3 f8b8 g1f2 f7f6 a2a4');
INSERT INTO evaluations VALUES(22952,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(22952,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(22952,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(22952,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(22952,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(22952,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(22952,36,'cp -23','b1c3 f8c5 f1c4 d7d6 e1g1 g8e7 c3a4 e7g6 a4c5 d6c5 c1e3 d8d1 f1d1 b7b6 c4d5 c8b7 a2a3 c6d8 b2b4 b7d5 e4d5 c5b4 a3b4 d8b7 h2h4 b7d6 h4h5 g6e7 e3f4 f7f6 f3d4 e8d7 f4d6 c7d6 b4b5 e7d5 d4f5 g7g6 d1d5 g6f5 d5f5 d7e6 g2g4 h8g8 a1e1 e6f7 e1e4 g8g5 f5g5 f6g5 g1g2 a8d8 f2f4 f7f6');
INSERT INTO evaluations VALUES(22946,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(22946,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(22946,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(22946,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(22946,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(22946,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(22946,36,'cp -30','f8b4 f1c4 d7d6 f3g5 g8h6 e1g1 e8g8 c3d5 c6e5 c4e2 a7a5 d5b4 a5b4 h2h3 g8h8 f2f4 e5c6 f1e1 f7f6 g5f3 c8e6 c1e3 d8e7 e3f2 h6f7 a2a4 b4a3 a1a3 a8a3 b2a3 g7g5 f4g5 f7g5 f3g5 f6g5 d1d2 c6e5 f2e3 h7h6 e3d4 b7b6 d2c3 e7g7 d4e5 d6e5');
INSERT INTO evaluations VALUES(6669,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(6669,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(6669,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(6669,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(6669,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(6669,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(6669,36,'cp -37','f1c4 d7d6 f3g5 g8h6 e1g1 e8g8 c3d5 c6e5 c4e2 a7a5 d5b4 a5b4 h2h3 g8h8 f2f4 e5c6 f1e1 f7f6 g5f3 d8e7 e2b5 h6f7 a2a4 b4a3 b2a3 c8d7 a3a4 c6e5 f4e5 d7b5 c1b2 b5c6 e5f6 g7f6 f3d4 f7e5 g1h1 f8g8 d4f5 e7e6 f5d4 e6f7 d4c6 b7c6');
INSERT INTO evaluations VALUES(6681,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(6681,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(6681,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(6681,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(6681,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(6681,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(6681,36,'cp -23','d7d6 f3g5 g8h6 e1g1 e8g8 f2f4 g8h8 a2a3 b4c3 b2c3 d8e7 f1e1 c8g4 g5f3 f7f5 h2h3 g4h5 c4d5 a8e8 d1d3 e7d7 c1d2 h5f3 d3f3 h6g8 e4f5 d7f5 d2e3 f5d3 e3f2 e8e1 a1e1 c6e5 f3d1 d3f5 e1f1 e5c6 d5c6 b7c6 d1a4 f5d7 a4a7 g8f6 f1e1');
INSERT INTO evaluations VALUES(6066,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(6066,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(6066,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(6066,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(6066,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(6066,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(6066,36,'cp -18','f3g5 g8h6 e1g1 e8g8 f2f4 g8h8 a2a3 b4c3 b2c3 d8e7 f1e1 c8g4 g5f3 c6a5 c4d5 h6g8 c1d2 c7c6 d5a2 b7b5 a3a4 g4f3 d1f3 b5a4 c3c4 a5b3 a2b3 a4b3 f3b3 f8e8 d2c3 g8f6 b3b2 a7a5 e4e5 d6e5 e1e5 e7d7 e5e8 f6e8 a1a5 d7d1 g1f2 a8a5 c3a5');
INSERT INTO evaluations VALUES(54194,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(54194,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(54194,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54194,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(54194,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54194,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(54194,36,'cp 10','e2e3 g8f6 b1d2 d8b6 a1b1 f8d6 d4c5 b6c5 b2b4 c5c7 f4d6 c7d6 c2c4 e8g8 f1e2 b7b6 e1g1 d6e7 d1b3 b8d7 f1c1 d5c4 c1c4 a7a5 b4b5 c8b7 c4c7 f8b8 e3e4 e7d6 c7c4 b8c8 b1c1 d6e7 c4c8 a8c8 c1c8 b7c8 e2f1 c8b7 e4e5');
INSERT INTO evaluations VALUES(54193,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(54193,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(54193,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(54193,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54193,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54193,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54193,36,'cp 9','g8f6 b1d2 f8e7 c2c3 e8g8 f3e5 d8b6 d1c2 b8c6 f1e2 a7a5 e1g1 a5a4 b2b3 h7h6 f1b1 a4a3 b1d1 c8d7 e5d7 f6d7 d2f3 d7f6 e2d3 f8d8 h2h3 e7d6 f4d6 d8d6 d4c5 b6c5 b3b4 c5a7 e3e4 d6d8 b4b5 c6a5 c2e2 f6e4');
INSERT INTO evaluations VALUES(57224,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(57224,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(57224,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(57224,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(57224,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(57224,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(57224,36,'cp -23','e5e4 f3d4 c7c5 d4e2 d7d5 d2d3 e4d3 c2d3 b8c6 g2g3 f8d6 f1g2 g8f6 e1g1 e8g8 d3d4 c5d4 e3d4 h7h6 b1c3 f8e8 c1e3 c8e6 e3f4 e6g4 h2h3 g4h5 f4d6 d8d6 f1e1 h5e2 e1e2 e8e2 c3e2 d6b4 b2b3 a8d8 a1c1 b4a3 e2c3');
INSERT INTO evaluations VALUES(57328,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(57328,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(57328,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(57328,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(57328,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(57328,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(57328,36,'cp -22','f3d4 d7d5 d2d3 g8f6 d3e4 d5e4 h2h3 c8d7 b1c3 h7h5 f1c4 a7a6 a2a3 f8d6 b2b4 b8c6 c1b2 c6e5 c4b3 d8e7 d1e2 a8c8 e1g1 e8g8 c3d5 f6d5 b3d5 c7c5 b4c5 c8c5 c2c4 b7b5 d4b3 c5d5 c4d5 e7g5 b2e5 g5e5 f2f4 e4f3 e2f3 e5h2 g1f2 g8h8');
INSERT INTO evaluations VALUES(57284,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(57284,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(57284,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(57284,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(57284,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(57284,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(57284,36,'cp -28','d7d5 d2d3 g8f6 d3e4 d5e4 h2h3 c8d7 b1c3 h7h5 a2a3 a7a6 b2b4 f8d6 c1b2 b8c6 d4c6 d7c6 b4b5 a6b5 c3b5 c6b5 f1b5 c7c6 b5c4 d8e7 e1g1 a8a5 a1b1 e8g8 b2f6 g7f6 b1b7 d6h2 g1h2 e7b7 d1d6 g8g7 f1d1 f8e8 d6f4 b7e7 c4e2 a5g5 a3a4');
INSERT INTO evaluations VALUES(56196,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(56196,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(56196,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(56196,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(56196,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(56196,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(56196,36,'cp -16','d2d3 c7c5 d4b5 e4d3 f1d3 g8f6 e3e4 a7a6 b5c3 d5d4 c3e2 f6g4 f2f3 g4e5 f3f4 e5d3 c2d3 d8h4 g2g3 h4d8 f4f5 b8c6 e1g1 f8d6 e2f4 e8g8 f4d5 c6e5 c1f4 b7b5 f4e5 d6e5 b1d2 c8b7 d1g4 b7d5 e4d5 f7f6 g4f3');
INSERT INTO evaluations VALUES(6212,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(6212,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(6212,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6212,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(6212,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(6212,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(6212,36,'cp 7','g8f6 e1g1 e8g8 h2h3 a7a6 c1e3 d7d6 e3c5 d6c5 a2a4 d8d6 f3d2 b7b6 c3d5 a8b8 d5e3 f8d8 d2f3 b6b5 a4b5 a6b5 c4b3 c6d4 f3d4 c5d4 e3d5 f6d5 b3d5 c7c6 d5b3 c8e6 b3e6 d6e6 d1f3 b8a8 f3g3 e6f6 a1a8 d8a8 f2f4 e5f4 g3f4 f6f4 f1f4');
INSERT INTO evaluations VALUES(53136,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(53136,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(53136,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(53136,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(53136,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(53136,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(53136,36,'cp 59','c3d4 d6d5 e4d5 g8f6 f1b5 b8d7 g1f3 a7a6 b5d7 d8d7 f3e5 d7d8 b1c3 f6d5 e1g1 g7g6 f1e1 f8g7 h2h4 e8g8 h4h5 f8e8 d1f3 c8e6 h5g6 h7g6 e5g6 d5c3 e1e6 f7e6 b2c3 d8d5 f3e2 a8d8 g6h4 e8f8 e2g4 f8f7 c1g5 d5c6 a1c1 c6d7');
INSERT INTO evaluations VALUES(53130,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(53130,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(53130,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(53130,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(53130,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(53130,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(53130,36,'cp 61','d6d5 e4d5 g8f6 f1b5 b8d7 b1c3 a7a6 b5d7 d8d7 g1f3 f6d5 f3e5 d7d8 e1g1 g7g6 d1b3 c8e6 b3b7 d8c8 b7c8 a8c8 c3d5 e6d5 c1e3 f8g7 f1d1 e8g8 b2b3 f8d8 a1c1 h7h6 f2f3 g6g5 g1f2 f7f5 g2g3 d5b7 c1c8 d8c8 d1c1 c8c1 e3c1 g7e5 d4e5 g8f7 c1d2 b7c6 f3f4 g5f4 d2f4 h6h5 f2e3');
INSERT INTO evaluations VALUES(57666,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(57666,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(57666,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(57666,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(57666,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(57666,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(57666,36,'cp 34','c7c5 d2d4 c5d4 d1d4 g8f6 c1e3 b8c6 d4d2 f8g7 f2f3 e8g8 e1c1 d8a5 c1b1 d7d6 g1e2 b7b5 e2f4 a8b8 c3d5 a5d2 d5f6 g7f6 d1d2 f6g7 f1e2 f7f5 e4f5 f8f5 e2d3 f5e5 h1e1 e5e3 e1e3 g7h6 f4d5 h6e3 d5e3 c8e6 b2b3 g8f7 h2h4');
INSERT INTO evaluations VALUES(52299,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(52299,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(52299,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52299,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(52299,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(52299,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(52299,36,'cp 44','d2d4 c5d4 d1d4 g8f6 c1e3 f8g7 d4d2 d7d6 f2f3 b8c6 g1e2 d8a5 e1c1 b7b5 e2d4 c6d4 e3d4 b5b4 d4f6 b4c3 d2c3 a5c3 f6c3 g7c3 b2c3 c8e6 c1d2 e8g8 a2a4 f8c8 d1b1 c8c7 f1d3 h7h5 b1b5 e6c4 b5a5 h5h4 d3c4 c7c4 h1b1 a8c8 a5a7');
INSERT INTO evaluations VALUES(52291,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52291,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52291,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(52291,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(52291,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(52291,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52291,36,'cp 41','c5d4 d1d4 g8f6 g1f3 b8c6 d4a4 d7d6 e4e5 d6e5 f3e5 d8d4 e5c6 d4a4 c3a4 b7c6 c1d2 c8f5 f1a6 a8b8 e1c1 f6d5 a2a3 e7e6 a6e2 h7h5 h1e1 f8e7 c2c4 d5b6 a4b6 a7b6 d2e3 f7f6 e2f3 c6c5 h2h3 e6e5 d1d2 e8f7 e1d1 b8d8 d2d8 h8d8 d1d8');
INSERT INTO evaluations VALUES(52335,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(52335,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(52335,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(52335,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52335,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(52335,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(52335,36,'cp 62','d1d4 g8f6 c1e3 f8g7 d4d2 d7d6 f2f3 b8c6 g1e2 e8g8 e1c1 d8a5 c1b1 b7b5 e3h6 a8b8 h6g7 g8g7 e2c1 c8e6 g2g4 c6e5 g4g5 f6h5 f3f4 e5c4 d2d4 g7g8 c1b3 a5c7 f4f5 g6f5 e4f5 e6f5 c3d5 c7d7 f1e2 h5g7 h2h4 c4e5 d4e3 d7e6 e3a7 e5c6');
INSERT INTO evaluations VALUES(52333,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(52333,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(52333,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(52333,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52333,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(52333,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52333,36,'cp 45','g8f6 g1f3 b8c6 d4a4 d7d6 e4e5 d6e5 f3e5 c8d7 e5d7 f6d7 f1b5 f8g7 e1g1 e8g8 f1d1 d8c8 c3d5 d7b6 d5b6 a7b6 a4c4 c6a7 c4c8 f8c8 b5d3 b6b5 c2c3 c8d8 c1e3 b5b4 c3b4 g7b2 a1b1 b2d4 a2a4 d4e3 f2e3 a7c6 a4a5 g8f8 d3e2 d8d1 e2d1 e7e6 d1f3 f8e7 g1f2 e7d6 f2e2 a8c8 b1c1 c8c7 c1d1 d6e7');
INSERT INTO evaluations VALUES(44446,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(44446,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(44446,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44446,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(44446,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(44446,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(44446,36,'cp 42','g1f3 b8c6 d4a4 d7d6 e4e5 d6e5 f3e5 c8d7 e5d7 d8d7 c1e3 f8g7 a1d1 d7c8 f1e2 e8g8 e1g1 e7e6 a4a3 f8d8 d1d8 c8d8 f1d1 d8c7 a3b3 h7h5 h2h3 a7a6 a2a4 a8b8 g2g3 b8c8 e2f3 f6d7 f3g2 c6a5 b3b4 a5c6 b4d6 g7c3 d6c7 c8c7 b2c3 c6e5 e3f4 c7c3 g2b7 c3c2');
INSERT INTO evaluations VALUES(44445,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(44445,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44445,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(44445,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(44445,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(44445,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(44445,36,'cp 36','b8c6 d4a4 d7d6 e4e5 d6e5 f3e5 c8d7 e5d7 d8d7 c1e3 f8g7 a1d1 d7g4 a4b3 e8g8 h2h3 g4b4 f1e2 e7e6 e1g1 f8d8 b3b4 c6b4 d1d8 a8d8 e3a7 b4c2 f1d1 f6d5 c3d5 d8d5 d1d5 e6d5 e2f3 g7b2 f3d5 b7b5 d5b3 c2d4 b3d1 d4c6 a7b6');
INSERT INTO evaluations VALUES(12741,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(12741,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12741,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(12741,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(12741,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(12741,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(12741,36,'cp 41','d4a4 d7d6 e4e5 d6e5 f3e5 c8d7 e5d7 d8d7 c1e3 f8g7 a1d1 d7g4 a4b3 e8g8 h2h3 g4b4 f1b5 a8c8 b5c6 b4b3 c2b3 c8c6 e1e2 a7a6 d1d3 h7h6 h1d1 g8h7 b3b4 h6h5 d3d8 h5h4 d8f8 g7f8 a2a4 f8h6 e3c5 c6e6 e2d3 f6d7');
INSERT INTO evaluations VALUES(12744,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(12744,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(12744,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12744,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(12744,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(12744,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(12744,36,'cp 28','d7d6 e4e5 d6e5 f3e5 c8d7 e5d7 d8d7 c1e3 f8g7 a1d1 d7g4 a4b3 e8g8 h2h3 g4b4 f1e2 a8d8 e1g1 b7b6 d1d8 f8d8 f1a1 b4h4 a1d1 d8d1 e2d1 h4b4 b3b4 c6b4 a2a3 b4d5 c3d5 f6d5 e3c1 f7f5 c2c3 g8f7 g1f1 f7e6 d1c2 e6d6 g2g4 e7e6 f1e2 b6b5 c1d2 d5b6 e2d3 f5g4 h3g4 b6c4 d2c1 d6e7 c2b3');
INSERT INTO evaluations VALUES(42482,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(42482,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42482,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42482,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42482,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42482,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42482,36,'cp 0','e8g8 g1f3 d7d5 e2e3 c7c5 a2a3 b4c3 b2c3 c5d4 c3d4 b7b6 a3a4 c8a6 c1a3 f8e8 f1e2 a6c4 e2c4 d5c4 c2c4 d8d5 a1c1 d5b7 e1g1 b8d7 c4c6 b7a6 c6c4');
INSERT INTO evaluations VALUES(37180,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37180,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37180,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37180,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37180,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37180,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37180,36,'cp 0','a2a3 b4c3 c2c3 b7b6 c1g5 h7h6 g5h4 c7c5 g1f3 c8b7 e2e3 c5d4 f3d4 b8c6 d4c6 b7c6 f2f3 a8c8 f1e2 f6e4 h4d8 e4c3 d8b6 c3e2 b6a5 c6a4 e1e2 c8c4 a1c1 f8c8 c1c4 c8c4 a5c3 f7f6 e3e4 a4b3 h1c1 g8f7 e2e3 c4c5 e3f2');
INSERT INTO evaluations VALUES(37185,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37185,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37185,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37185,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37185,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37185,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37185,36,'cp 0','b4c3 c2c3 b7b6 c1g5 h7h6 g5h4 c8b7 e2e3 d7d5 c4c5 b6c5 h4f6 d8f6 c3c5 f8c8 g1f3 b8d7 c5b5 d7b6 a1c1 a7a5 b5c5 b6d7 c5b5');
INSERT INTO evaluations VALUES(37202,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37202,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37202,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37202,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37202,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37202,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37202,36,'cp 0','c2c3 b7b6 c1g5 h7h6 g5h4 c8b7 e2e3 d7d5 c4c5 b6c5 h4f6 d8f6 c3c5 f8c8 g1f3 b8d7 c5b5 d7b6 b5c5');
INSERT INTO evaluations VALUES(37199,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37199,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37199,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37199,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37199,36,'cp 0','b7b6 c1g5 h7h6 g5h4 c8b7 e2e3 d7d5 h4f6 d8f6 c4c5 b6c5 c3c5 f8c8 g1f3 b8d7 c5b5 d7b6 b5c5');
INSERT INTO evaluations VALUES(35660,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35660,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35660,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35660,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35660,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35660,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35660,36,'cp 0','c1g5 h7h6 g5h4 c8b7 e2e3 d7d5 h4f6 d8f6 c4c5 b6c5 c3c5 f8c8 g1f3 b8d7 c5b5 d7b6 b5c5');
INSERT INTO evaluations VALUES(1879,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(1879,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(1879,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(1879,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(1879,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(1879,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(1879,36,'cp 19','d4b5 f8c5 d1f3 c5b6 f3g3 d7d6 b1c3 g8e7 c1g5 f6g6 g5e3 b6a5 a2a3 a7a6 b2b4 a6b5 b4a5 a8a5 f1b5 f7f5 g3g6 e7g6 e3d2 e8d8 a3a4 a5a8 d2g5 c6e7 e1g1 c8d7 f1b1 f5e4 c3e4 h7h6 g5e7 g6e7 b5d7 d8d7 b1b7 e7c6');
INSERT INTO evaluations VALUES(1845,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(1845,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(1845,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1845,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(1845,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(1845,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(1845,36,'cp 9','f8c5 d1e2 c5b6 b1c3 g8e7 c1e3 b6a5 e2f3 f6g6 f3g3 d7d6 a2a3 a7a6 b2b4 a6b5 b4a5 b5b4 a3b4 c6b4 e1d2 b4c6 f1b5 g6g3 h2g3 e8d8 c3d5 e7d5 e4d5 c6a5 e3d4 f7f6 d4c3 b7b6 h1h6 c8b7 h6f6 b7d5 f6f5');
INSERT INTO evaluations VALUES(46180,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(46180,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(46180,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46180,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46180,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46180,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(46180,36,'cp 22','c4d5 e6d5 f1e2 b8c6 e1g1 c5d4 f3d4 f8d6 b2b3 e8g8 c1b2 f8e8 b1d2 a7a6 d4c6 b7c6 d1c2 c8d7 a1c1 a6a5 f1d1 d8e7 e2d3 d7g4 d1e1 d6b4 b2d4 g4h5 d4c5 e7e6 c5b4 a5b4 c2c6 a8a2 d2f3 h5f3 g2f3');
INSERT INTO evaluations VALUES(46135,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(46135,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46135,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46135,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(46135,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(46135,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(46135,36,'cp 24','e6d5 f1e2 b8c6 e1g1 c5d4 f3d4 f8d6 b2b3 e8g8 c1b2 f8e8 b1d2 c6d4 b2d4 c8f5 d2f3 f5e4 e2d3 d8e7 a1c1 e7e6 d1e2 a7a6 f1d1 h7h6 h2h3 a8d8 d4c5 d8c8 c5d6 e4f3 e2f3 c8c1 d1c1 e6d6');
INSERT INTO evaluations VALUES(46434,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(46434,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(46434,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(46434,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46434,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46434,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46434,36,'cp 24','f1e2 b8c6 e1g1 c5d4 f3d4 f8d6 b2b3 e8g8 c1b2 f8e8 b1d2 c6d4 b2d4 c8f5 d2f3 f5e4 e2d3 d8e7 a1c1 a7a6 d1e2 a8d8 f1d1 e7e6 c1c2 e6f5 d4c5 d6c5 c2c5 e4d3 d1d3 d8c8 c5c8 e8c8 f3d4 f5e5 d3d1');
INSERT INTO evaluations VALUES(54179,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(54179,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(54179,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(54179,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(54179,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(54179,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(54179,36,'cp 50','c4d5 e6d5 b1c3 g8f6 c1g5 c8e6 g5f6 d8f6 e2e4 d5e4 c3e4 f6f4 f1b5 b8c6 d1e2 e8c8 b5c6 b7c6 e1g1 e6d5 f1e1 c8b8 e4c5 f8c5 d4c5 h8e8 e2e8 d8e8 e1e8 b8b7 e8e7 b7c8 e7e3 f4b4 b2b3 b4c5 h2h3 c5c2 f3e5 c2d2 e5c4 d5c4');
INSERT INTO evaluations VALUES(54132,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(54132,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(54132,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(54132,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(54132,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(54132,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(54132,36,'cp 40','e6d5 c1g5 f8e7 g5e7 d8e7 d4c5 e7c5 b1c3 g8f6 e2e3 b8c6 f1e2 e8g8 e1g1 f8d8 d1b3 f6e4 f1c1 e4c3 c1c3 c5e7 a1d1 d5d4 f3d4 c6d4 d1d4 d8d4 e3d4 e7e2 c3e3 c8e6 e3e2 e6b3 a2b3 g8f8 f2f3 a8d8 e2e4 f7f5 e4f4 f8e7 g1f2 a7a6 g2g4 f5g4 f3g4 a6a5');
INSERT INTO evaluations VALUES(54525,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(54525,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(54525,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(54525,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(54525,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(54525,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54525,36,'cp 53','b1c3 g8f6 c1g5 c5d4 f3d4 f8e7 e2e3 b8c6 f1b5 c8d7 e1g1 e8g8 g5f6 e7f6 c3d5 f6d4 e3d4 d7e6 b5c6 e6d5 c6d5 d8d5 d1a4 b7b5 a4b4 a8d8 f1d1 g7g6 b4c5 a7a6 c5d5 d8d5 d1d2 h7h5 g2g3 f8d8 a1d1 h5h4');
INSERT INTO evaluations VALUES(18916,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(18916,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(18916,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(18916,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(18916,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(18916,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18916,36,'cp 32','d7d5 e4e5 g8e7 b1c3 e7f5 f1b5 d8b6 e1g1 c8d7 b5a4 a8c8 c3e2 b6a6 a4c2 c6b4 c2f5 e6f5 e2c3 d7e6 a2a3 b4d3 d1d2 f8e7 f1d1 d3c1 a1c1 e8g8 g2g3 h7h6 h2h4 c8c4 f3e1 f5f4 e1g2 f4g3 f2g3 f7f6 g2e3 f6e5');
INSERT INTO evaluations VALUES(20395,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(20395,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(20395,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(20395,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(20395,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(20395,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(20395,36,'cp 34','e4e5 g8e7 b1c3 e7f5 f1b5 c8d7 e1g1 d8b6 g2g4 f5e7 c1e3 h7h5 g4g5 g7g6 b5d3 c6b4 d3b1 b4c6 c3a4 b6d8 f1e1 c6a5 a4c5 e7c6 b2b3 a8c8 h2h3 f8e7 g1g2 e8g8 c5d3 e7a3 d1d2 c6e7 f3h4 d7b5 d3f4 a5c6 e1g1');
INSERT INTO evaluations VALUES(23121,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(23121,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(23121,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(23121,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(23121,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(23121,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(23121,36,'cp 76','f3d4 e5d4 d1d4 d8f6 d4e3 f8b4 c2c3 b4e7 f1d3 f6h6 e3h6 g8h6 c1h6 g7h6 a2a4 h6h5 b1a3 h5h4 a3b5 e8d8 e1g1 a7a5 e4e5 f7f6 a1e1 f6e5 e1e5 c7c6 b5d4 e7f6 e5h5 d7d5 d3h7 d8c7 f1e1 f6d4 c3d4 h4h3 g2g3 c8g4 h5h6');
INSERT INTO evaluations VALUES(23116,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(23116,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(23116,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23116,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(23116,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(23116,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(23116,36,'cp 75','e5d4 d1d4 d8f6 d4e3 f8b4 c2c3 b4a5 b1d2 d7d6 d2c4 a5b6 c4b6 a7b6 f1d3 g8e7 e1g1 e8g8 c1d2 f8e8 b2b3 c8e6 c3c4 e7c6 f1c1 f6d4 e3d4 c6d4 d2c3 d4c6 d3c2 f7f6 c1e1 g8f7 a2a4 h7h5 f2f3 e6d7');
INSERT INTO evaluations VALUES(23153,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(23153,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(23153,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(23153,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(23153,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(23153,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(23153,36,'cp 87','d1d4 d8f6 d4c4 c7c6 b1c3 b7b5 c4d3 b5b4 c3a4 f6e6 f2f3 d7d5 c1e3 g8f6 a4c5 f8c5 e3c5 d5e4 f3e4 e6e4 d3e4 f6e4 c5b4 a7a5 b4a3 c8e6 f1d3 f7f5 h1f1 e8c8 e1c1 h8e8 g2g4 g7g6 d3a6 c8b8 d1d8 e8d8 a6d3 b8c7 b2b3 e4c3 g4f5 e6f5 d3f5 g6f5 f1f5');
INSERT INTO evaluations VALUES(48344,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48344,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(48344,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48344,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(48344,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(48344,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48344,36,'cp 0','c7c5 c2c4 d5c4 d3c4 c5d4 e3d4 h7h6 b1c3 f8d6 e1g1 e8g8 d1e2 b8c6 f1d1 c6e7 f3e5 b7b6 e2f3 e7d5 f3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(50686,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(50686,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(50686,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(50686,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(50686,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(50686,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(50686,36,'cp 77','c2c4 e7e6 c1e3 g8f6 b1c3 f8e7 f1e2 b8c6 e1g1 c8d7 d4c6 d7c6 d1d4 e8g8 b2b4 f6d7 d4d3 b7b6 e3d4 d8b8 d3g3 e7f6 a1d1 f8d8 f2f4 b8c7 g1h1 a8c8 f4f5 e6f5 d4f6 d7f6 f1f5 f6e4');
INSERT INTO evaluations VALUES(50681,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(50681,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(50681,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(50681,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(50681,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(50681,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(50681,36,'cp 68','e7e6 b1c3 f8e7 f1e2 g8f6 e1g1 e8g8 c1e3 b8c6 h2h3 c8d7 a1c1 c6d4 d1d4 d8b8 a2a4 a6a5 d4d2 d7c6 e2d3 f6d7 d3c2 e7f6 c1d1 f8d8 f2f4 d7c5 e3d4 f6d4 d2d4 b8c7 f4f5 e6f5 f1f5 d8d7 b2b3 a8e8 d1f1 c7b6 d4f2');
INSERT INTO evaluations VALUES(50762,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(50762,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(50762,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(50762,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(50762,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(50762,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(50762,36,'cp 76','f1e2 f8e7 b1c3 g8f6 c1e3 e8g8 e1g1 b7b6 f2f4 c8b7 e2f3 b8d7 d1e2 a8c8 g2g4 e6e5 d4f5 e5f4 e3f4 d7e5 f4e5 d6e5 a1d1 d8c7 g4g5 f6e8 h2h4 e8d6 f5e3 b6b5 c3d5 b7d5 e3d5 c7b7 c4c5 c8c5 b2b3 a6a5 f3g4 e7d8 g1h2 g7g6 e2e3 c5c2 h2h3 c2a2 f1f6 d8f6 d5f6 g8h8 d1d6');
INSERT INTO evaluations VALUES(45359,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(45359,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45359,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45359,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45359,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45359,36,'cp 0','g1f3 b8c6 f1e2 c5d4 e3d4 d8b6 b1c3 c8g4 c3a4 b6d8 a4c5 e7e6 e1g1 a8c8 c5b3 f8e7 c2c3 e8g8 a2a4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(19348,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(19348,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(19348,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(19348,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(19348,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(19348,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19348,36,'cp 0','c6d4 g1f3 a7a6 b5d3 e7e6 e1g1 d4c6 d3e2 c6d4 e2d3');
INSERT INTO evaluations VALUES(19585,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19585,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19585,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19585,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19585,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19585,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19585,36,'cp 0','g1f3 a7a6 b5d3 e7e6 e1g1 d4c6 d3e2 c6d4');
INSERT INTO evaluations VALUES(19584,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(19584,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19584,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19584,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19584,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19584,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19584,36,'cp 0','a7a6 b5d3 g7g6 f3d4 c5d4 c3e2 d7d5 c2c3 d4c3 d1a4 c8d7 a4d4 d5e4 d3e4 c3b2 c1b2 g8f6 e4b7 a8b8 b7a6 d8b6 d4b6 b8b6 b2f6 b6f6 a1b1 f6a6 b1b8 d7c8 b8c8 e8d7 c8c2 f8g7 e2c1 h8b8 e1e2 b8b2 c2b2 g7b2 c1d3 b2d4 d3c1');
INSERT INTO evaluations VALUES(23473,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(23473,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(23473,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23473,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(23473,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(23473,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(23473,36,'cp 8','e7e5 c2c4 e5e4 f3h4 g8f6 f1g2 d7d5 c4d5 d8d5 b1c3 d5d4 d2d3 e4d3 c1e3 d4d8 e2d3 c6b4 e1g1 f8e7 c3e4 c7c6 a2a3 b4d5 e3c5 e8g8 c5e7 d8e7 f1e1 d5c7 a1c1 c7b5 e4c5 e7c7 h4f3 c8g4 c5e4 f6e4 d3e4 f8d8 d1c2');
INSERT INTO evaluations VALUES(22875,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(22875,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(22875,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(22875,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(22875,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(22875,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22875,36,'cp 12','c2c4 e5e4 f3h4 g8f6 f1g2 d7d5 c4d5 d8d5 b1c3 d5d4 d2d3 e4d3 e1g1 f8e7 h4f3 d4d8 e2d3 e8g8 a2a3 c8e6 d3d4 f6d5 f1e1 f8e8 d1b3 a8b8 b3c2 d8d7 c1g5 e7g5 f3g5 e6f5 c2d2 d5c3 b2c3 h7h6 g5f3 e8e1 a1e1 f5e6 f3e5 c6e5 e1e5');
INSERT INTO evaluations VALUES(22741,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(22741,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(22741,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(22741,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(22741,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(22741,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(22741,36,'cp 22','e5e4 f3h4 g8f6 f1g2 d7d5 c4d5 d8d5 b1c3 d5d4 d2d3 e4d3 e1g1 f8e7 h4f3 d4d8 e2d3 e8g8 d3d4 f8e8 f1e1 h7h6 a2a3 c8e6 c1e3 e7f8 f3e5 c6e5 d4e5 d8d1 a1d1 f6g4 g2b7 a8b8 b7c6 g4e5 c6e8 e5f3 g1h1 f3e1 e8a4 e1f3 h1g2');
INSERT INTO evaluations VALUES(55030,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(55030,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(55030,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(55030,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(55030,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(55030,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(55030,36,'cp -20','g1f3 g8f6 e2e3 e7e5 f3e5 f8b4 c1b2 e8g8 f1e2 c8f5 e1g1 b8d7 e5f3 f8e8 d2d3 c7c5 b1d2 b7b5 a2a4 b4d2 f3d2 b5b4 a1c1 a7a5 c2c4 d7e5 d2f3 e5d7 c4d5 f6d5 f3d2 f5e6 e2f3 a8c8');
INSERT INTO evaluations VALUES(55029,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(55029,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(55029,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(55029,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(55029,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(55029,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(55029,36,'cp -32','g8f6 e2e3 e7e5 f3e5 f8b4 f1e2 e8g8 c1b2 c8f5 e1g1 b8d7 e5d7 f6d7 b1c3 c7c6 e2g4 f5g6 c3e2 b4d6 e2f4 b7b5 a2a4 b5b4 a4a5 d7e5 f4g6 h7g6 g4e2 a8b8 b2e5 d6e5 d2d4 e5f6 g2g3 f8e8 e2d3');
INSERT INTO evaluations VALUES(48275,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48275,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48275,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48275,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48275,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48275,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48275,36,'cp 0','c4d5 e6d5 b1c3 c7c6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(48188,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48188,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48188,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48188,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48188,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48188,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48188,36,'cp 0','e6d5 b1c3 c7c6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(48698,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48698,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48698,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48698,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48698,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48698,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48698,36,'cp 0','b1c3 c7c6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(48694,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48694,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48694,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48694,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48694,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48694,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48694,36,'cp 0','c7c6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(45782,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45782,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45782,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45782,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45782,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45782,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45782,36,'cp 0','c1f4 f6h5 f4c1 h5f6');
INSERT INTO evaluations VALUES(50365,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(50365,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(50365,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(50365,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50365,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(50365,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(50365,36,'cp 9','g1f3 d7d5 c1b2 c7c5 e2e3 e7e6 c2c4 b8c6 c4d5 e6d5 f1b5 c8g4 d2d4 a8c8 d4c5 f8c5 e1g1 e8g8 b1c3 f8e8 h2h3 g4h5 c3e2 h5f3 g2f3 c5d6 f3f4 a7a6 b5c6 b7c6 g1h2 d6f8 e2g3 f6e4 d1g4 e4g3 g4g3');
INSERT INTO evaluations VALUES(50363,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50363,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50363,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(50363,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50363,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50363,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50363,36,'cp 0','d7d5 c1b2 e7e6 c2c4 f8e7 d2d4 c7c5 e2e3 e8g8 f1d3 b8c6 e1g1 c5d4 e3d4 b7b6 f1e1 c8b7 b1c3 c6a5 f3e5 a8c8 a1c1 a5c6 e5f3 c6a5');
INSERT INTO evaluations VALUES(47251,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47251,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47251,36,'cp 0','c1b2 e7e6 c2c4 f8e7 d2d4 c7c5 e2e3 c5d4 e3d4 e8g8 f1d3 b8c6 e1g1 b7b6 f1e1 c8b7 b1c3 c6a5 f3e5 a5c6');
INSERT INTO evaluations VALUES(47252,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47252,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47252,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47252,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47252,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47252,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(47252,36,'cp 0','e7e6 c2c4 f8e7 e2e3 e8g8 f1e2 c7c5 e1g1 b7b6 d2d4 c8b7 b1d2 b8d7 a1c1 a8c8 c4d5 f6d5 a2a3 e7f6 d2e4 f6e7 e4d2');
INSERT INTO evaluations VALUES(48409,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(48409,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48409,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48409,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48409,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48409,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(48409,36,'cp 0','c2c4 f8e7 e2e3 e8g8 f1e2 b7b6 e1g1 c8b7 d2d4 d5c4 e2c4 b8d7 d1e2 a7a6 a2a4 c7c5 d4c5 e7c5 f1d1 d8e7 b1d2 f8d8 c4d3 c5d6 h2h3 h7h6 a1c1 d7c5 d3c2 a8c8 d2c4 d6c7 c4d2');
INSERT INTO evaluations VALUES(48233,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48233,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48233,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48233,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48233,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48233,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48233,36,'cp 0','f8e7 e2e3 e8g8 d2d4 c7c5 f1d3 c5d4 e3d4 b7b6 e1g1 c8b7 b1c3 b8c6 f1e1 c6b4 d3f1 a8c8 a2a3 b4c6 f1d3 c6a5 f3e5 a5c6');
INSERT INTO evaluations VALUES(9767,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(9767,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(9767,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(9767,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(9767,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9767,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9767,36,'cp 6','d7d6 b1d2 e8g8 f1e1 a7a6 c4b3 c5a7 h2h3 c8e6 d2f1 d6d5 c1e3 a7e3 f1e3 d5e4 d3e4 d8d1 a1d1 e6b3 a2b3 f6e4 e3c4 e4d6 c4e5 c6e5 e1e5 f8e8 d1d5 a8c8 g1f1 f7f6 e5e1 e8e1 f3e1 g8f7');
INSERT INTO evaluations VALUES(8844,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(8844,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(8844,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(8844,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(8844,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(8844,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(8844,36,'cp 23','f1e1 e8g8 b1d2 a7a6 d2f1 c8e6 c4e6 f7e6 f1g3 d8e8 h2h3 a6a5 c1e3 c5e3 e1e3 a5a4 d1e2 c6e7 d3d4 e7g6 d4e5 d6e5 c3c4 e8c6 e2c2 b7b6 a1d1 a8d8 d1e1 f6d7 b2b4 a4b3 a2b3 d7c5 b3b4 c6a4 c2a4 c5a4');
INSERT INTO evaluations VALUES(8845,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(8845,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8845,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(8845,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(8845,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(8845,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(8845,36,'cp 8','a7a6 b1d2 e8g8 d2f1 c8e6 c4e6 f7e6 f1g3 d8e8 h2h3 f6d7 c1e3 c5e3 e1e3 c6e7 d1b3 d7c5 b3c2 e7g6 d3d4 c5d7 d4e5 d7e5 f3e5 g6e5 g3e2 e5g6 c2b3 a8b8 a1f1 g6f4 e2f4 f8f4 f2f3 e8g6 g1h2 g6f7 e3d3 f4f6 d3d4 e6e5 b3f7 g8f7');
INSERT INTO evaluations VALUES(36296,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(36296,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(36296,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36296,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(36296,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36296,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(36296,36,'cp 0','h7h6 g5f6 e7f6 a1c1 c7c6 f1e2 d5c4 e2c4 b8d7 e1g1 c6c5 c3e4 c5d4 e4f6 d7f6 d1d4 d8d4 f3d4 c8d7 c4e2 a8c8 f2f4 f8d8 e2f3 b7b6 g1f2 g7g5 g2g3 c8c5 d4b3 c5c8');
INSERT INTO evaluations VALUES(36196,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36196,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36196,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36196,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(36196,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36196,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36196,36,'cp 13','g5h4 b7b6 f1e2 d5c4 e2c4 b8d7 d1e2 c8b7 e1g1 f6d5 c3d5 e6d5 h4e7 d8e7 c4d3 c7c5 b2b3 d7f6 f1c1 f8c8 c1c2 a7a5 d3b5 c8c7 a1c1 f6e4 h2h3 a8c8 a2a4 g7g6 e2e1 b7c6 b5a6 c6b7 a6e2 g8g7 d4c5 c7c5 c2c5 c8c5');
INSERT INTO evaluations VALUES(36199,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(36199,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36199,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(36199,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36199,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(36199,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(36199,36,'cp 7','b7b6 f1e2 d5c4 e2c4 b8d7 d1e2 c8b7 e1g1 f6d5 c3d5 e6d5 h4e7 d8e7 c4d3 c7c5 b2b3 f8c8 f1c1 a7a5 d3b5 b7c6 a2a4 c6b5 e2b5 g7g6 c1c2 e7d6 a1c1 g8g7 h2h3 c8c7 b5d3 d7f6 d4c5 b6c5 f3d4');
INSERT INTO evaluations VALUES(35627,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(35627,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(35627,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35627,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(35627,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(35627,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(35627,36,'cp 18','f1e2 d5c4 e2c4 b8d7 d1e2 c8b7 e1g1 f6d5 c3d5 e6d5 h4e7 d8e7 c4d3 c7c5 b2b3 f8c8 a1c1 a7a5 f1d1 d7f6 a2a4 c8c7 c1c2 f6e4 h2h3 a8c8 d4c5 c7c5 c2c5 b6c5 e2e1 c5c4 d3e4 e7e4 e1c3 e4e6 b3c4');
INSERT INTO evaluations VALUES(17271,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(17271,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(17271,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(17271,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(17271,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(17271,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17271,36,'cp 75','d2d4 e7e6 a2a3 a7a6 g1f3 d7d6 h2h3 f8e7 f1d3 b7b5 e1g1 e8g8 b2b3 c8b7 f1e1 a8c8 c1d2 f8e8 c3e2 c6b8 e2g3 c7c5 c2c3 b8d7 a3a4 c5c4 b3c4 b5c4 d3c2 d8c7 f3h2 b7c6 h2g4 f6g4 d1g4 d7f6 g4e2 c7b6 e2c4 d6d5');
INSERT INTO evaluations VALUES(17265,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(17265,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(17265,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(17265,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(17265,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(17265,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(17265,36,'cp 75','e7e6 a2a3 a7a6 c1e3 d7d6 g1f3 f8e7 f1d3 b7b5 e1g1 e8g8 b2b4 c8b7 d1e2 c6b8 a3a4 d6d5 e4d5 e6d5 a4b5 e7b4 e3d2 a6b5 c3b5 f8e8 f3e5 b4d2 e2d2 c7c6 a1a8 b7a8 f1a1 a8b7 b5c3 b8d7 a1b1 d8c7 e5d7 f6d7 c3e2');
INSERT INTO evaluations VALUES(16294,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(16294,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(16294,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(16294,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(16294,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(16294,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(16294,36,'cp 80','a2a3 a7a6 g1f3 d7d6 h2h3 f8e7 f1d3 b7b5 e1g1 e8g8 b2b3 c6a5 b3b4 a5c4 a3a4 c8d7 d1e2 c7c6 d3c4 b5c4 a4a5 a8b8 e2c4 d6d5 c4a6 e7b4 c1d2 f6e4 c3e4 d5e4 d2b4 b8b4 f3e5 b4d4 e5c6 d7c6 a6c6');
INSERT INTO evaluations VALUES(44476,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44476,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44476,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44476,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(44476,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(44476,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(44476,36,'cp 26','b1c3 d7d5 g1f3 e7e6 c1g5 h7h6 g5h4 b8d7 e2e3 f8e7 f1e2 e8g8 e1g1 b7b6 a2a4 a7a6 c4d5 e6d5 a1c1 f8e8 h4g3 f6h5 f3e5 d7e5 e2h5 e5d7 c3e2 d7f6 h5f3 c8d7 d1c2 a8c8 g3e5 c6c5');
INSERT INTO evaluations VALUES(44473,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(44473,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(44473,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44473,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44473,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(44473,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(44473,36,'cp 31','d7d5 g1f3 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 b8d7 f1e2 e8g8 e1g1 b7b6 c4d5 f6d5 c3d5 e6d5 h4e7 d8e7 d1c2 c6c5 f1c1 c8b7 b2b3 f8e8 c2b2 d7f8 c1c2 f8e6 a1c1 a8c8 b2a3 a7a6 h2h3 e8d8 e2d3 c8c7 a3b2');
INSERT INTO evaluations VALUES(31824,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(31824,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(31824,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(31824,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(31824,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(31824,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(31824,36,'cp 38','e7e6 g1f3 g8f6 f3h4 f5g6 d1b3 b8c6 e2e3 f8b4 c4d5 d8d5 b3d5 f6d5 c1d2 e8c8 f1b5 g6h5 b5c6 b7c6 c3d5 b4d2 e1d2 d8d5 h1c1 e6e5 h4f5 d5d7 c1c6 h5g6 e3e4 c8b7 a1c1 g6f5 e4f5 h8d8 b2b4 f7f6 d2e3');
INSERT INTO evaluations VALUES(32281,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(32281,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(32281,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(32281,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(32281,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(32281,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(32281,36,'cp 28','g1f3 g8f6 f3h4 f5g6 d1b3 b8c6 e2e3 f8b4 h4g6 h7g6 c1d2 d8d7 c4d5 e6d5 f1d3 g6g5 e1c1 e8c8 c3a4 b4d2 d1d2 c6a5 b3b4 b7b6 d2c2 d7d6 b4d6 d8d6 c1b1 c7c6 h2h3 g5g4 h3h4 a5c4 h4h5 h8h5 h1h5 f6h5 d3c4 d5c4 c2c4 h5f6 a4c3');
INSERT INTO evaluations VALUES(32280,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32280,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(32280,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(32280,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(32280,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(32280,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(32280,36,'cp 20','g8f6 d1b3 b8c6 c4d5 e6d5 e2e4 d5e4 f3e5 f5e6 b3b7 c6d4 f1b5 d4b5 b7b5 e6d7 b5c4 d8e7 e5d7 e7d7 c1g5 d7e6 c4a4 e6d7 a4c2 d7d3 a1c1 e8c8 c2a4 e4e3 g5e3 f6d5 c3d5 d8d5 a4c4 d3c4 c1c4 a7a5 b2b3 f8d6 g2g3 h8e8 e1g1 c8d7 f1c1 e8b8 c1e1');
INSERT INTO evaluations VALUES(30069,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(30069,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(30069,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(30069,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30069,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(30069,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(30069,36,'cp 38','f3h4 f5g6 d1b3 b8c6 c4d5 e6d5 e2e3 f8b4 c1d2 e8g8 h4g6 h7g6 a2a3 b4c3 d2c3 a8b8 b3a4 f8e8 f1d3 b8a8 a4b3 a8b8 b3c2 c6e7 e1g1 d8d7 a1e1 e7f5 e1e2 c7c6 f1e1 f6e4 c3b4 e4f6 f2f3 b8a8 b4c5 e8c8 c2a4 b7b6 c5b4 a7a5 b4c3 b6b5');
INSERT INTO evaluations VALUES(53010,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(53010,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(53010,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(53010,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(53010,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(53010,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(53010,36,'cp -53','g8f6 g1e2 b8c6 e1g1 g7g6 h2h3 f8g7 d2d3 a8b8 b1c3 a7a6 a2a4 e8g8 c1e3 f6e8 f3g3 c6d4 c4b3 e8c7 e3d4 c5d4 c3b1 c7e6 f2f4 e6c5 b1d2 c8d7 f4f5 g6f5 e4f5 g8h8 g3h4 g7f6 h4f2 c5b3 d2b3 f8g8 b3d2');
INSERT INTO evaluations VALUES(45145,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(45145,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(45145,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(45145,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(45145,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(45145,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(45145,36,'cp -37','g1e2 b8c6 e1g1 g7g6 d2d3 f8g7 b1c3 e8g8 a2a3 a7a6 c4a2 a8b8 f3g3 f6h5 g3h4 b7b5 c3d5 e7e6 c1g5 f7f6 g5e3 g8h8 d5c3 f6f5 h4d8 f8d8 h2h3 a6a5 g2g4 h5f6 e4f5 e6f5 g4g5 f6d7 f2f4 b5b4 a3b4 b8b4');
INSERT INTO evaluations VALUES(45146,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(45146,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(45146,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(45146,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(45146,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(45146,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(45146,36,'cp -47','b8c6 d2d3 g7g6 e1g1 f8g7 f3f4 a7a6 a2a3 e8g8 b1c3 b7b5 c4a2 a8b8 f4h4 f6h5 f2f3 a6a5 c3d5 e7e6 c1g5 d8d7 d5f6 g7f6 g5f6 h5f6 h4f6 d7d8 f6c3 d8b6 c3d2 c8a6 g1h1 c5c4 f3f4 c4d3 c2d3 b5b4 a1c1 b4b3 a2b3 b6b3 c1c6 b3d3 c6d6');
INSERT INTO evaluations VALUES(10500,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(10500,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(10500,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(10500,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(10500,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(10500,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(10500,36,'cp -25','f6e4 e1g1 b4c3 d4d5 c3f6 f1e1 c6e7 e1e4 e8g8 d5d6 c7d6 c1g5 d6d5 c4d5 e7d5 d1d5 f6g5 f3g5 h7h6 g5f3 d7d6 a2a3 d8f6 d5d4 f6g6 f3h4 g6g5 d4d6 f8d8 d6b4 b7b6 h4f3 g5f6 e4e3 c8b7 a1e1 a8b8 b4c3 f6c3 e3c3');
INSERT INTO evaluations VALUES(10692,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(10692,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(10692,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(10692,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(10692,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(10692,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(10692,36,'cp -20','e1g1 b4c3 d4d5 c3f6 f1e1 e8g8 e1e4 c6e7 d5d6 c7d6 c1g5 f6g5 f3g5 d6d5 c4d5 e7d5 d1d5 h7h6 g5h3 d7d6 h3f4 d8g5 e4d4 g5d5 f4d5 c8e6 d5c3 f8d8 a1d1 g8f8 d4d6 f8e7 d6d2 d8d2 d1d2 a8d8 d2d8 e7d8 f2f3 d8e7 g1f2 e7d6 b2b4 d6e5 f2e3 g7g5 g2g3');
INSERT INTO evaluations VALUES(10691,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(10691,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(10691,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(10691,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(10691,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(10691,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(10691,36,'cp -44','b4c3 d4d5 c3f6 f1e1 c6e7 e1e4 d7d6 e4f4 a7a6 f4f6 g7f6 d1d4 e7g6 c4d3 h8g8 c1f4 g6f4 d4f4 e8f8 d3h7 g8g7 f4h6 d8d7 h7d3 d7g4 g2g3 c8d7 h6f6 g4h5 a1e1 h5d5 f6e7 f8g8 d3e4 d5b5 f3d4 b5b2 e7d7 b2d4 e4b7 a8f8 d7c7 d4d2 c7c1 d2c1 e1c1 f8b8 b7d5 b8b2 c1c7 g8h7 d5f7');
INSERT INTO evaluations VALUES(10701,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(10701,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(10701,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(10701,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(10701,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(10701,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10701,36,'cp -21','d4d5 c6e5 d1e2 e8g8 b2c3 e5c4 e2c4 e4d6 c4f4 f8e8 c1a3 e8e4 f4g3 d6f5 g3g5 d8g5 f3g5 e4e5 f2f4 e5e3 a1e1 h7h6 e1e3 f5e3 f1e1 h6g5 e1e3 g5f4 e3e8 g8h7 d5d6 c7d6 a3d6 b7b5 d6f4 c8b7 e8a8 b7a8 a2a3 h7g6 g2g4 a7a6 g1f2 g6f6 f2g3 a8d5 f4e3 f6e6 e3d4 f7f6 h2h4 d5c4 d4c5 e6f7 h4h5 f7e6 c5d4 e6d5');
INSERT INTO evaluations VALUES(10655,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(10655,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(10655,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(10655,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(10655,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(10655,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(10655,36,'cp -31','c6e5 d1e2 e8g8 b2c3 e5c4 e2c4 e4d6 c4f4 f8e8 c1a3 e8e4 f4g3 d6f5 g3g5 d8g5 f3g5 e4e5 f2f4 e5e3 f1e1 h7h6 e1e3 f5e3 a1e1 h6g5 e1e3 g5f4 e3e8 g8h7 d5d6 c7d6 a3d6 g7g5 d6e7 g5g4 e7d6 b7b5 d6f4 c8b7 e8b8 a8b8 f4b8 a7a5 g1f2 h7g6 g2g3 b7d5');
INSERT INTO evaluations VALUES(10812,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(10812,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(10812,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(10812,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(10812,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(10812,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(10812,36,'cp -21','d1e2 e8g8 b2c3 e5c4 e2e4 c4d6 e4f4 f8e8 c1a3 e8e4 f4g3 d6f5 g3g5 d8g5 f3g5 e4e5 f2f4 e5e3 f1e1 h7h6 e1e3 f5e3 a1e1 h6g5 e1e3 g5f4 e3e8 g8h7 d5d6 c7d6 a3d6 b7b5 d6f4 c8b7 e8a8 b7a8 g2g4 a8d5 a2a3 d5e6 h2h3 h7g6 g1f2 a7a5 f4e3 f7f6 f2g3 g6f7 h3h4 e6c4 g4g5 f7g6 g5f6 g7f6');
INSERT INTO evaluations VALUES(4335,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(4335,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4335,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(4335,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(4335,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(4335,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4335,36,'cp 32','e1g1 d7d6 c1g5 h7h6 g5h4 a7a5 b1d2 g7g5 h4g3 f6h7 g1h1 g8h8 d3d4 e5d4 d2b3 d4c3 b3c5 c3b2 a1b1 d6c5 d1b3 d8f6 g3c7 a5a4 b3b2 f6b2 b1b2 f8e8 f3d2 h8g8 f2f4 a4a3 b2b3 c6a5 c7a5 a8a5 f4g5 h7g5 b3b6');
INSERT INTO evaluations VALUES(54433,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(54433,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(54433,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(54433,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(54433,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54433,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(54433,36,'cp 36','c3d4 b8c6 b1c3 g8e7 g1f3 e7f5 f1b5 c8d7 e1g1 d8b6 b5a4 f8e7 c3e2 c6b4 a4d7 e8d7 c1g5 e7g5 f3g5 a8f8 a1b1 d7c8 d1d2 c8b8 g5f3 h7h6 a2a3 b4c6 b2b4 g7g5 h2h3 h8g8 g2g4 f5e7 d2e3 h6h5 f3g5 h5g4 h3g4 f7f5 e5f6 f8f6 f2f3');
INSERT INTO evaluations VALUES(48272,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(48272,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(48272,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48272,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48272,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(48272,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(48272,36,'cp -13','d5c4 e2e3 a7a6 f1c4 b7b5 c4e2 c7c5 e2f3 a8a7 c3e4 f6e4 f3e4 b8d7 d4c5 f8c5 b2b4 c5e7 c1b2 e7f6 b2d4 c8b7 e4f3 f6d4 d1d4 d8f6 d4f6 g7f6 f3b7 a7b7 a1c1 d7b6 g1f3 e8e7 e1e2 b6c4 a3a4 h8c8 f3d2 b7c7 d2c4 c7c4 a4b5 a6b5');
INSERT INTO evaluations VALUES(48495,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(48495,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(48495,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(48495,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(48495,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(48495,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48495,36,'cp -8','e2e3 a7a6 f1c4 b7b5 c4e2 c7c5 e2f3 a8a7 c3e4 f6e4 f3e4 b8d7 d4c5 d7c5 d1d8 e8d8 e4c2 f8d6 g1e2 c5a4 c2a4 b5a4 c1d2 c8d7 d2c3 f7f6 f2f4 d8e7 e1f2 a7c7 h1d1 d7b5 d1d2 h8c8 a1d1 d6c5 c3a5 c7b7 e2d4 b5e8 a5c3');
INSERT INTO evaluations VALUES(48493,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(48493,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(48493,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(48493,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(48493,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(48493,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(48493,36,'cp -14','a7a6 f1c4 b7b5 c4e2 c7c5 e2f3 a8a7 c3e4 f6e4 f3e4 b8d7 d4c5 d7c5 d1d8 e8d8 e4c2 c5a4 c2a4 b5a4 c1d2 c8d7 f2f4 a7b7 d2c3 f8c5 e1f2 d8e7 a1c1 h8c8 g1e2 f7f6 g2g4 e7e8 g4g5 f6g5 f4g5 c5e7 c3g7 e7g5 c1c8 d7c8 g7c3 g5e7 h1c1');
INSERT INTO evaluations VALUES(53636,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(53636,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(53636,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(53636,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(53636,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(53636,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(53636,36,'cp 64','g1f3 c6c5 c2c3 b8c6 f1d3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f5e3 f2e3 f8e7 b1c3 c8d7 g1h1 e8g8 e3e4 c6b4 e4d5 b4d5 c3d5 e6d5 d1b3 d7e6 b3b7 a7a5 b7b3 g7g6 b3c2 d8b6 c2d2 a5a4 h2h4 b6b4 d2e3 b4b2 h4h5 b2a3');
INSERT INTO evaluations VALUES(47373,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47373,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(47373,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(47373,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(47373,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(47373,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(47373,36,'cp 62','e7e5 g1f3 e5d4 d1d4 d8d4 f3d4 f8c5 d4b5 b8a6 c1f4 e8g8 f4c7 c8e6 c7d6 c5d6 b5d6 a6b4 e1e2 a8d8 d6b7 d8b8 b7c5 f8c8 c5e6 b4c2 a1b1 f7e6 e2d1 c8c3 b2c3 b8b1 d1c2 b1a1 c2b2 a1e1 f2f3 g8f8 b2c2 f6d7 h2h4 e1a1 c2b2 a1d1 b2b3 d7c5 b3c4 c5d3 c4b5 d1b1 b5c6');
INSERT INTO evaluations VALUES(48918,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(48918,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(48918,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(48918,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(48918,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(48918,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(48918,36,'cp 64','g1f3 e5d4 f3d4 f8c5 c1e3 d8e7 f2f3 e8g8 d1d2 c5b6 e1c1 b8c6 d4c6 b7c6 f1c4 h7h6 c1b1 a8b8 d2e2 b6e3 e2e3 f6d7 d1d4 b8b4 c4e2 a7a5 c3a4 c6c5 d4c4 b4c4 e2c4 d7b6 c4b5 b6a4 b5a4 c5c4 h1c1');
INSERT INTO evaluations VALUES(35284,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(35284,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(35284,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(35284,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(35284,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(35284,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(35284,36,'cp 104','f1c4 g8f6 b1c3 c7c6 e1g1 d7d6 d2d4 c8g4 d4d5 e7c7 h2h3 c6d5 c4d5 g4e6 c3b5 c7d7 c2c4 e6d5 e4d5 f8e7 c1e3 a7a6 b5c3 d7c7 f3d2 e8g8 b2b4 b8d7 d1b3 f6e8 g2g4 f7f5 g4f5 e8f6 c3e4 f6e4 d2e4');
INSERT INTO evaluations VALUES(35277,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(35277,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(35277,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(35277,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(35277,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(35277,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(35277,36,'cp 105','g8f6 e1g1 d7d6 d2d4 c8g4 b1d2 c7c6 f1e1 e7c7 c2c3 f8e7 h2h3 g4h5 g2g4 h5g6 f3h4 b8d7 g4g5 f6h5 d2f3 e8c8 h4g6 f7g6 b2b4 h7h6 f3h4 e7g5 h4g6 g5c1 a1c1 h5f4 g6h8 d8h8 d1g4 c8b8 g4g7 c7d8 g1h2');
INSERT INTO evaluations VALUES(34043,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(34043,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(34043,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(34043,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(34043,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(34043,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(34043,36,'cp 95','e1g1 d7d6 d2d4 b8c6 b1c3 c8g4 c4b5 a7a6 b5c6 b7c6 d1d3 g4f3 d3f3 e7e6 d4e5 d6e5 b2b3 f8b4 c3a4 e8g8 c1e3 f8d8 a1d1 a6a5 h2h3 h7h6 g1h2 e6e7 g2g3 b4a3 h2g2 d8d6 d1d3 d6d3 c2d3 a8d8 f1d1 e7e6 a4c5 a3c5 e3c5 f6d7 c5a3');
INSERT INTO evaluations VALUES(29669,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(29669,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(29669,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(29669,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(29669,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(29669,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(29669,36,'cp 44','c2c4 c7c6 d1b3 d8b6 f3e5 g4f5 c4d5 b6b3 a2b3 f6d5 f1d3 f5d3 e5d3 b8a6 c1d2 e7e6 a1a4 d5b6 a4a1 b6d5 e1e2 a6b4 d3b4 d5b4 g2g4 h7h5 g4g5 b4d5 b1a3 f8a3 b2a3 e8c8 h1c1 h8e8 f2f3 d5c7 h2h4 d8d7 d2c3 b7b6 a3a4');
INSERT INTO evaluations VALUES(29631,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(29631,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29631,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(29631,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(29631,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(29631,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(29631,36,'cp 38','c7c6 d1b3 d8b6 f3e5 g4f5 c4d5 b6b3 a2b3 f6d5 f1d3 f5g6 b1c3 g6d3 e5d3 e7e6 c1d2 d5b4 d3b4 f8b4 e1e2 a7a6 c3e4 b4e7 b3b4 b8d7 d2c3 e8c8 e4c5 h7h5 h2h4 h8e8 c5d3 e7d6 h1d1 d7f6 f2f3 f6d5 c3d2 d6e7 g2g3 d5c7 d2c3 g7g6');
INSERT INTO evaluations VALUES(29023,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(29023,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(29023,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(29023,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(29023,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(29023,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(29023,36,'cp 40','h2h3 g4f5 f1d3 f5d3 d1d3 b8d7 e1g1 e7e6 b1d2 h7h6 e3e4 d5e4 d2e4 f6e4 d3e4 f8d6 c1d2 e8g8 f1e1 a7a5 a2a4 f8e8 a1d1 d6b4 d2b4 a5b4 b2b3 d8c7 d1d3 a8d8 e1d1 d8a8 e4e2 a8d8 d3e3 c6c5 d4d5 d7f6 d5e6');
INSERT INTO evaluations VALUES(29020,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(29020,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(29020,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(29020,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(29020,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(29020,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(29020,36,'cp 27','g4f5 f1d3 f5d3 d1d3 e7e6 e1g1 f8e7 b1c3 e8g8 f1d1 d8c7 e3e4 d5e4 c3e4 b8d7 e4f6 d7f6 d3e2 f8e8 d1d3 c6c5 b2b3 c5d4 c1b2 e7c5 f3d4 a8d8 a1d1 a7a6 d4f3 d8d3 d1d3 e8d8 b2e5 c7b6 f3e1 d8d3 e1d3');
INSERT INTO evaluations VALUES(51522,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(51522,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(51522,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(51522,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(51522,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(51522,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(51522,36,'cp 84','e2e4 c8b7 b1c3 e7e6 d2d4 g8f6 e4e5 f6d5 c3d5 b7d5 c1f4 f8e7 f1d3 b8c6 c2c3 f7f6 e5f6 e7f6 e1g1 e8g8 f4g3 c6e7 f1e1 e7f5 d3e4 d5e4 e1e4 f5g3 h2g3 d8e7 a2a4 e7f7 a4a5 c7c5 a5b6 a7b6');
INSERT INTO evaluations VALUES(45444,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45444,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45444,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45444,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45444,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45444,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45444,36,'cp 0','g1f3 c8g4 h2h3 g4f3 d1f3 b8c6 c1e3 e7e6 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(52447,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(52447,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(52447,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(52447,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(52447,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(52447,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(52447,36,'cp -16','d7d5 d2d3 g8f6 e4e5 f6g4 f2f4 b8c6 f1g2 g4h6 c2c3 h6f5 b1a3 h7h5 a3c2 e7e6 d1e2 b7b6 g1f3 c8b7 c2e3 c6e7 e1g1 a7a5 f1d1 g7g6 d3d4 b7a6 e2f2 a8c8 g2h3 c5d4 c3d4 f5e3 c1e3 e7f5 h3f5 e6f5 h2h3 f8e7');
INSERT INTO evaluations VALUES(53289,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(53289,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(53289,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(53289,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(53289,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(53289,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53289,36,'cp 0','d2d3 g8f6 e4d5 f6d5 f1g2 b8c6 g1f3 g7g6 e1g1 f8g7 f1e1 e8g8 b1d2 b7b6 c2c3 c8b7 h2h4 d8c7 h4h5 a8d8 d2e4 c7d7 a2a4 g6h5 f3h4 d7g4 g2f3 g4d7 f3g2');
INSERT INTO evaluations VALUES(46947,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(46947,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(46947,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(46947,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(46947,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(46947,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(46947,36,'cp 73','g1f3 c8g4 f1e2 g4f3 e2f3 e7e5 d4d5 f8c5 e1g1 e8g8 d1e2 b8d7 a2a4 a7a5 b1d2 c7c6 d5c6 b7c6 d2b3 c5d4 f1d1 c6c5 e2b5 b6c8 b3d4 c5d4 f3g4 d7f6 g4c8 d8c8 b5e5 f8e8 e5d4 e8e4 d4d3 c8e8 c1e3 e4a4 a1c1 a4b4 e3c5 b4b8 d3g3 e8b5 c1c3 a5a4');
INSERT INTO evaluations VALUES(46946,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(46946,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(46946,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(46946,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(46946,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(46946,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(46946,36,'cp 70','c8g4 b1c3 e7e5 d4d5 b8d7 h2h3 g4h5 f1e2 f8c5 e1g1 e8g8 d1d3 h7h6 c1e3 h5f3 e2f3 c5e3 d3e3 b6c4 e3e2 c4d6 a1c1 a7a6 f1d1 g8h8 b2b3 d8e7 c3a4 d6b5 e2e3 f8e8 f3e2 e8d8 e2b5 a6b5 a4c5 d7c5 e3c5 e7c5 c1c5');
INSERT INTO evaluations VALUES(29434,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(29434,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(29434,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(29434,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(29434,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(29434,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(29434,36,'cp 65','f1e2 g4f3 e2f3 e7e5 d4d5 f8c5 b1d2 c7c6 e1g1 e8g8 d2b3 c5d6 a2a4 d8e7 a4a5 b6d7 a5a6 b7b6 g2g3 d7f6 f3g2 h7h6 c1d2 c6d5 e4d5 f8c8 d1e2 b8d7 f1c1 c8e8 c1c2 a8c8 c2c6 e5e4 b3d4 c8c6 d5c6 d7e5');
INSERT INTO evaluations VALUES(29433,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(29433,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(29433,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(29433,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(29433,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(29433,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(29433,36,'cp 57','g4f3 e2f3 e7e5 d4d5 f8c5 b1d2 c7c6 e1g1 e8g8 d2b3 c5d6 a2a4 d8e7 a4a5 b6d7 a5a6 b7b6 f1e1 f8d8 c1d2 d7f6 d1e2 d8e8 e2c4 c6d5 e4d5 e7f8 c4a4 b8d7 a1c1 a8c8 f3e2 c8c1 e1c1 d7c5 a4b5 c5b3 b5b3 d6c5');
INSERT INTO evaluations VALUES(29523,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(29523,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(29523,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(29523,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(29523,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(29523,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(29523,36,'cp 69','d1b3 d5c4 b3b7 b8d7 g1f3 a8b8 b7a6 f6e4 g2g4 f5g6 f1g2 e4c3 b2c3 g6e4 e1g1 e7e6 f3d2 e4g2 g1g2 h7h5 d2c4 h5g4 c1f4 d7b6 a6b5 d8d7 b5d7 e8d7 c4e5 d7e8 a1b1 f8d6 f4g3 d6e5 g3e5 e8d7 e5g7 h8e8 g2g3 f7f5 g7e5 e8g8 e2e4 f5e4');
INSERT INTO evaluations VALUES(29519,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(29519,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(29519,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(29519,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(29519,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(29519,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(29519,36,'cp 65','d5c4 b3b7 b8d7 b7c6 a8b8 g1f3 e7e6 g2g3 f8e7 f1g2 e8g8 e1g1 f5c2 f3e1 c2g6 c6c4 c7c5 d4c5 d7c5 e1f3 d8a5 f3e5 f8c8 c4f4 b8b4 f4f3 g6h5 e5c6 h5f3 c6a5 f3g2 g1g2 e7d8 a5b3 c5b3 a2b3 a7a5 a1a3 h7h6 f1d1 d8b6 c3a2 f6g4 f2f4 g4f2 a2b4 f2d1');
INSERT INTO evaluations VALUES(15300,30,'cp -239',NULL);
INSERT INTO evaluations VALUES(15300,31,'cp -218',NULL);
INSERT INTO evaluations VALUES(15300,32,'cp -237',NULL);
INSERT INTO evaluations VALUES(15300,33,'cp -235',NULL);
INSERT INTO evaluations VALUES(15300,34,'cp -230',NULL);
INSERT INTO evaluations VALUES(15300,35,'cp -232',NULL);
INSERT INTO evaluations VALUES(15300,36,'cp -250','d8g5 c4d5 c6d4 f3f7 e8d8 e1g1 c8g4 d2d3 d4e2 g1h1 e2c1 d5f3 c1e2 h2h3 g4f3 f7f3 e2f4 f3b7 a8c8 b1c3 f8d6 a1e1 c7c6 f1g1 g5e7 b7a6 e7h4 a6c4 h8f8 g1f1 h4g5 f1g1 g5h6 e1e3 h6e6 c4a6 f4d5 e3e2 d5c3 b2c3 c8c7 g1e1 e6d5 e1b1 c6c5 h1g1 d8e7 e2e1 d5c6 a6c4');
INSERT INTO evaluations VALUES(971,30,'cp -241',NULL);
INSERT INTO evaluations VALUES(971,31,'cp -241',NULL);
INSERT INTO evaluations VALUES(971,32,'cp -213',NULL);
INSERT INTO evaluations VALUES(971,33,'cp -243',NULL);
INSERT INTO evaluations VALUES(971,34,'cp -228',NULL);
INSERT INTO evaluations VALUES(971,35,'cp -223',NULL);
INSERT INTO evaluations VALUES(971,36,'cp -238','c4d5 c6d4 f3f7 e8d8 e1g1 c8g4 d2d3 d4e2 g1h1 e2c1 d5f3 c1e2 h2h3 g4f3 f7f3 e2f4 f3b7 a8c8 b1c3 c7c6 f1g1 g5e7 b7a6 c8c7 a1d1 e7c5 d3d4 e5d4 d1d4 c5d4 g1d1 d4d1 c3d1 f4d5 c2c4 d5b6 d1c3 f8b4 c3e4 b6d7 a6a4 h8e8 f2f3 b4f8 a4a5 d7f6 c4c5 f6e4 f3e4 d8c8 g2g3 e8e4 h3h4 f8e7 b2b4 e4e1 h1h2');
INSERT INTO evaluations VALUES(966,30,'cp -254',NULL);
INSERT INTO evaluations VALUES(966,31,'cp -236',NULL);
INSERT INTO evaluations VALUES(966,32,'cp -213',NULL);
INSERT INTO evaluations VALUES(966,33,'cp -258',NULL);
INSERT INTO evaluations VALUES(966,34,'cp -252',NULL);
INSERT INTO evaluations VALUES(966,35,'cp -250',NULL);
INSERT INTO evaluations VALUES(966,36,'cp -239','c6d4 f3f7 e8d8 e1g1 c8g4 d5b7 a8b8 f7d5 f8d6 b1c3 g5g6 f2f4 h8f8 b7c6 b8b6 c6a4 c7c6 d5e4 d6c5 g1h1 d4e2 f1f3 g6e4 c3e4 g4f3 g2f3 c5e7 b2b3 f8f4 c1b2 e2d4 b2d4 e5d4 a1g1 g7g6 h1g2 d8c7 g1e1 b6b8 c2c3 e7h4 e1e2 d4c3 e4c3 b8f8 e2e6 f4f3 a4c6 f3f5 g2h3 h4g5 c3e4 a7a6 e4g3 f5c5');
INSERT INTO evaluations VALUES(1024,30,'cp -215',NULL);
INSERT INTO evaluations VALUES(1024,31,'cp -258',NULL);
INSERT INTO evaluations VALUES(1024,32,'cp -250',NULL);
INSERT INTO evaluations VALUES(1024,33,'cp -251',NULL);
INSERT INTO evaluations VALUES(1024,34,'cp -231',NULL);
INSERT INTO evaluations VALUES(1024,35,'cp -235',NULL);
INSERT INTO evaluations VALUES(1024,36,'cp -224','f3f7 e8d8 e1g1 c8g4 d5b7 a8b8 f7d5 f8d6 b1c3 g5g6 f2f4 h8f8 b7c6 d4c2 f4e5 f8f1 g1f1 g4e6 d5f3 e6c4 f1g1 d6c5 g1h1 c2a1 d2d4 c5e7 h2h3 g6f7 f3e4 c4e6 e4h7 b8b4 h1h2 g7g6 h7h6 b4d4 c3b5 a1c2 c1g5 e7g5 h6g5 d8c8 b5d4 c2d4 c6e4 d4f5 g5f6 f5e7 a2a3 c8d7 f6f7 e6f7 h2g3 c7c5');
INSERT INTO evaluations VALUES(53113,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(53113,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(53113,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(53113,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(53113,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(53113,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(53113,36,'cp 90','e2e4 g8f6 b1c3 e7e6 f1c4 f8e7 g1e2 e6d5 c3d5 f6d5 c4d5 b8a6 e2c3 a6c7 c1f4 e8g8 e1g1 c8e6 a2a4 d8d7 d1d2 e7f6 d5e6 c7e6 f4g3 f6c3 b2c3 d7c6 f2f3 a8d8 f1d1 d8d7 h2h3 f8e8 a4a5 e8d8 c3c4 e6d4');
INSERT INTO evaluations VALUES(53110,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(53110,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(53110,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(53110,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(53110,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(53110,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(53110,36,'cp 95','g8f6 f1d3 g7g6 c2c4 e7e6 b1c3 f8g7 h2h3 e6d5 e4d5 e8g8 g1f3 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1c2 e4f6 a1e1 b8d7 e3f4 d8c7 g2g4 b7b5 b2b3 b5c4 b3c4 d7b6 g1g2 f6d7 e1e2 d7e5 f3e5 g7e5 f4h6 c7e7');
INSERT INTO evaluations VALUES(45231,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(45231,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(45231,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(45231,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(45231,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(45231,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(45231,36,'cp 99','f1d3 g7g6 c2c4 f8g7 b1c3 e8g8 h2h3 e7e6 g1f3 e6d5 e4d5 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1c2 e4f6 e3d2 b8a6 a2a3 a6c7 b2b3 d8d7 a1b1 b7b5 d2a5 e8e7 f1e1 e7e1 b1e1 b5c4 b3c4 a8b8 g2g3 f6e8 g1g2 b8b2 c2c1 d7a4 a5c7 e8c7');
INSERT INTO evaluations VALUES(45229,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(45229,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(45229,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(45229,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(45229,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(45229,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(45229,36,'cp 76','g7g6 c2c4 f8g7 h2h3 e7e6 g1f3 e6d5 e4d5 d8e7 d3e2 c8f5 e1g1 e8g8 f1e1 f5b1 a1b1 b8a6 d1b3 e7d7 c1f4 a8e8 a2a3 e8e4 f4d2 f8e8 e2d3 e4e1 d2e1 b7b6 b3c2 a6c7 e1c3 b6b5 b2b3 e8b8 d3f1 d7d8 c3a5 b5c4 f1c4 f6d7 b1e1');
INSERT INTO evaluations VALUES(50593,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50593,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50593,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50593,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50593,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50593,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50593,36,'cp 0','d2d3 e7e6 c4b3 d7d5 e4d5 e6d5 b1c3 c8e6 d1f3 b8c6 g1e2 g8f6 c1g5 f8e7 e1c1 c6e5 f3g3 e5g6 g3f3');
INSERT INTO evaluations VALUES(50589,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50589,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50589,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50589,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50589,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50589,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50589,36,'cp 0','e7e6 c4b3 d7d5 e4d5 e6d5 g1e2 b8c6 e1g1 f8e7 b1c3 g8f6 c1g5 f6h5 g5c1');
INSERT INTO evaluations VALUES(50489,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(50489,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(50489,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(50489,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(50489,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(50489,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(50489,36,'cp 8','c4b3 d7d5 c2c3 d5e4 d3e4 d8d1 b3d1 g8f6 f2f3 b7b5 a2a4 c8b7 c1e3 f8e7 g1h3 e8g8 d1e2 c5c4 a4b5 a6b5 a1a8 b7a8 b1d2 f6d7 b2b3 c4b3 d2b3 f8c8 e3d2 e7f6 e2b5 f6c3 e1e2 a8c6 b5c6 c8c6 h1c1 c3f6 h3f2 c6c1 d2c1 g8f8 c1e3 f6d8 f2d3 b8c6 b3c5 f8e7 c5d7');
INSERT INTO evaluations VALUES(39563,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(39563,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(39563,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(39563,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(39563,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(39563,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(39563,36,'cp 59','d2d4 d7d6 e2e4 g8f6 g1f3 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 c1e3 b8a6 a2a3 b7b6 a1b1 a6c5 f3d2 g4e2 d1e2 a5a4 b2b4 a4b3 e3c5 b6c5 b1b3 f6h5 d2f3 g7h6 a3a4 h5f4 e2c2 f7f5 b3a3 g6g5 e4f5 g5g4 f3e1 d8g5 a4a5 g5f5 c2f5 f8f5');
INSERT INTO evaluations VALUES(17279,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(17279,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(17279,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(17279,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(17279,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(17279,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(17279,36,'cp 99','e4e5 f6d5 c2c4 d5b6 d2d4 d7d6 e5e6 f7e6 b1c3 e6e5 d4d5 c6d4 f3d4 e5d4 d1d4 e7e5 d4e4 f8e7 f1d3 b6d7 e4e2 e8g8 c1e3 b7b6 d3c2 a7a5 a2a3 d7c5 e1g1 c8f5 c2f5 f8f5 e3c5 b6c5 e2d3 d8d7 c3e4 f5f8 a1e1 d7f5 e1e2 f5g6 f1e1 a8b8 d3c2 g6h5');
INSERT INTO evaluations VALUES(17226,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(17226,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(17226,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(17226,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(17226,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(17226,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(17226,36,'cp 112','f6d5 c2c4 d5b6 d2d4 d7d6 e5e6 f7e6 c1e3 e6e5 d4d5 c6d4 f3d4 e5d4 e3d4 e7e5 d5e6 d8g5 b1d2 c8e6 f1e2 g5f4 e2h5 e6f7 d4e3 f7h5 d1h5 f4f7 h5a5 f7d7 e1g1 f8e7 e3b6 c7b6 a5d5 e8c8 d2e4 h8f8 a1d1 f8f5 d5d3 g7g6 f1e1 c8b8 e4c3 e7f6 b2b3 d7g7');
INSERT INTO evaluations VALUES(17303,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(17303,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(17303,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(17303,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(17303,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(17303,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(17303,36,'cp 90','d2d4 d7d6 c2c4 d5b6 e5e6 f7e6 b1c3 g7g6 c1e3 f8g7 h2h4 e6e5 d4d5 c6d4 f3d4 e5d4 e3d4 g7d4 d1d4 e7e5 d5e6 e8g8 e1c1 c8e6 f2f3 d8f6 h4h5 f6g5 d1d2 g6h5 c3e4 g5h6 f1d3 d6d5 c4d5 b6d5 e4c5 e6f7 c5b7 d5f4 d3c4 f4e6 d4e5 a8b8 b7c5 e6c5');
INSERT INTO evaluations VALUES(17301,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(17301,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(17301,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(17301,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(17301,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(17301,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(17301,36,'cp 74','d7d6 c2c4 d5b6 e5e6 f7e6 b1c3 g7g6 c1e3 f8g7 h2h4 e6e5 d4d5 c6d4 f3d4 e5d4 e3d4 g7d4 d1d4 e7e5 d5e6 e8g8 h4h5 d8f6 d4f6 f8f6 b2b3 c8e6 f2f3 d6d5 c4d5 e6d5 c3d5 b6d5 f1c4 a8e8 e1f2 f6d6 h5h6 g8f8 a1d1 e8e5 d1d4 d5e3 d4d6 c7d6');
INSERT INTO evaluations VALUES(14492,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(14492,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(14492,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(14492,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(14492,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(14492,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(14492,36,'cp 86','c2c4 d5b6 b1c3 c8g4 e5d6 e7d6 c1e3 g7g6 h2h3 g4f3 d1f3 f8g7 e1c1 d8e7 h3h4 e8c8 h4h5 f7f5 c3d5 b6d5 c4d5 c6b8 g2g4 b8d7 g4f5 h8f8 f1d3 g6f5 h5h6 g7h8 d3f5 e7f7 h1h5 c8b8 f3h3 d7f6 h5h4 f7d5 c1b1 f8e8 f5d3 d5g8 h3f5 f6d5 f5h7 g8h7 d3h7');
INSERT INTO evaluations VALUES(14491,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(14491,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(14491,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(14491,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(14491,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(14491,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(14491,36,'cp 100','d5b6 b1c3 c8g4 e5d6 e7d6 c1e3 g7g6 h2h3 g4f3 d1f3 f8g7 a1d1 e8g8 f1e2 c6e7 c4c5 b6d7 c5d6 c7d6 f3b7 e7f5 e1g1 a8b8 b7a7 b8a8 a7b7 a8b8 b7e4 b8b2 e3c1 b2b4 c3b5 d7c5 e4f3 b4d4 b5d4 f5d4 d1d4 g7d4');
INSERT INTO evaluations VALUES(14159,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(14159,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(14159,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(14159,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(14159,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(14159,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(14159,36,'cp 70','e5e6 f7e6 c1e3 g7g6 h2h4 f8g7 b1c3 e6e5 d4d5 c6d4 f3d4 e5d4 e3d4 g7d4 d1d4 e7e5 d5e6 e8g8 h4h5 d8f6 d4f6 f8f6 f2f3 c8e6 h5g6 h7g6 e1f2 g8g7 a1e1 a8d8 h1h4 e6g8 h4d4 c7c5 d4e4 d6d5 e4e7 f6f7 c4d5 b6d5 c3d5 d8d5 e7f7 g8f7 e1e7 g7f6 e7b7 d5d2 f2e3 d2d1 f1a6 f7a2 b7a7 a2d5');
INSERT INTO evaluations VALUES(17350,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(17350,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(17350,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(17350,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(17350,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(17350,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(17350,36,'cp 32','d2d4 c5d4 f3d4 d8c7 d4c6 g8f6 d1e2 d7c6 a2a4 f8b4 e1g1 e6e5 f2f3 e8g8 c1e3 f6h5 f1d1 h5f4 e2f2 c8e6 c4f1 a6a5 g2g3 f4h5 c3b1 h5f6 b1a3 h7h6 a3c4 f6d7 c4b6 d7b6 e3b6 c7e7 c2c3 b4d6 d1d2 f8e8 a1d1 d6c7 b6e3 e7f6');
INSERT INTO evaluations VALUES(17374,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(17374,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(17374,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(17374,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(17374,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(17374,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(17374,36,'cp 17','c5d4 f3d4 d8c7 d4c6 g8f6 d1e2 d7c6 a2a4 f8d6 h2h3 b7b6 e1g1 c8b7 f2f4 e6e5 f4f5 b6b5 c4d3 e8g8 c1g5 d6e7 c3d1 h7h6 g5h4 a8d8 c2c3 d8d6 d1f2 d6d7 f2g4 f6g4 h4e7 d7e7 e2g4 f7f6 d3c2 c7b6 g1h2 c6c5 a4b5 a6b5 g4g6 c5c4 a1d1');
INSERT INTO evaluations VALUES(17423,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(17423,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(17423,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(17423,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(17423,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(17423,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(17423,36,'cp 25','f3d4 d8c7 d4c6 g8f6 d1f3 c7c6 c4e2 f8b4 c1d2 d7d6 e1c1 e8g8 e4e5 c6f3 e2f3 d6e5 g2g4 a8b8 g4g5 f6d7 c3e4 b4d2 d1d2 f7f6 g5f6 d7f6 e4d6 f6d5 f3d1 b7b5 h1e1 b8b6 d6c8 f8c8 e1e5 g8f7 d1f3 f7f6 e5e4 g7g5');
INSERT INTO evaluations VALUES(17418,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(17418,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(17418,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(17418,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(17418,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(17418,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(17418,36,'cp 24','d8c7 d4c6 g8f6 d1e2 d7c6 e4e5 f6d5 e1g1 d5c3 b2c3 c6c5 e2f3 c8d7 f1d1 d7a4 c4d3 f8e7 d3e4 e8g8 d1e1 f8b8 h2h4 a4c6 e4c6 c7c6 f3c6 b7c6 g2g3 a6a5 e1e4 h7h6 g1g2 a5a4 e4c4 a4a3 h4h5 b8b7 c4e4 b7b5 c3c4');
INSERT INTO evaluations VALUES(1256,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(1256,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(1256,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(1256,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(1256,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(1256,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(1256,36,'cp 26','d4c6 g8f6 e4e5 c7c6 e5f6 c6c4 d1f3 d7d5 c1e3 c4h4 f6g7 f8g7 e1c1 e8g8 f3g3 h4g3 h2g3 c8d7 e3h6 a8c8 c3e2 c8c5 h6g7 g8g7 h1h5 h7h6 e2d4 b7b5 a2a3 f8e8 f2f3 c5c7 d4b3 e8h8 c2c3 c7c8 f3f4 c8c4 c1d2 h8c8 d1h1');
INSERT INTO evaluations VALUES(49158,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49158,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49158,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49158,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(49158,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49158,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49158,36,'cp 0','c7c5 c3e2 c5d4 e2d4 b8c6 c1f4 f8e7 h2h4 d8b6 c2c3 b6b2 d4b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(43994,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43994,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43994,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43994,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43994,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43994,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43994,36,'cp 0','c3e2 c5d4 e2d4 b8c6 c1f4 f8e7 h2h4 d8b6 c2c3 b6b2 d4b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(44007,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44007,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44007,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44007,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44007,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44007,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44007,36,'cp 0','c5d4 e2d4 b8c6 c1f4 f8e7 h2h4 d8b6 c2c3 b6b2 d4b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(55113,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(55113,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(55113,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(55113,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(55113,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(55113,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(55113,36,'cp -18','g8f6 d2d4 c7c5 c1f4 b8c6 e2e3 c5d4 f3d4 e7e6 f1e2 f8e7 d4c6 b7c6 e1g1 e8g8 c3a4 d8a5 c2c4 c8a6 d1c2 a8d8 f1d1 d5c4 e2c4 a6b5 b2b3 b5a4 b3a4 f6d5 a1c1 e7f6 c4d3 d5f4 e3f4');
INSERT INTO evaluations VALUES(47260,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(47260,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(47260,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(47260,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(47260,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(47260,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(47260,36,'cp -13','d2d4 c7c5 c1f4 b8c6 e2e3 c8g4 f1b5 e7e6 h2h3 g4f3 d1f3 c5d4 e3d4 a8c8 e1g1 d8b6 f3d3 a7a6 b5c6 c8c6 c3e2 f8e7 c2c3 e8g8 a2a4 a6a5 f1b1 f8c8 e2c1 c6c4 d3d1 e7d6 f4d6 b6d6 c1d3 f6d7 d1b3');
INSERT INTO evaluations VALUES(38410,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(38410,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(38410,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(38410,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(38410,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(38410,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(38410,36,'cp 62','c1e3 a7a6 h2h4 h7h5 g1f3 g8f6 f3g5 f6g4 e3f4 b8c6 f2f3 e7e5 d4e5 g4e5 d1d2 f7f6 g5h3 c6e7 e1c1 c8e6 f4e3 d8d7 c1b1 e6f7 f1e2 e8c8 c3d5 c8b8 h3f4 f6f5 d5e7 d7e7 f4d5 f7d5 d2d5 e5c6 d5b3');
INSERT INTO evaluations VALUES(38408,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(38408,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(38408,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(38408,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38408,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(38408,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(38408,36,'cp 51','a7a6 h2h4 h7h5 g1f3 g8f6 f3g5 f6g4 e3f4 b8c6 f2f3 e7e5 d4e5 g4e5 d1d5 c6e7 d5d2 b7b5 e1c1 a8b8 f1e2 b5b4 f4e5 g7e5 c3a4 e5g7 e2c4 e8g8 g2g4 d8e8 c4b3 c8d7 a4c5 a6a5 c5d7 e8d7 g4h5 a5a4 b3f7 f8f7 g5f7 g8f7');
INSERT INTO evaluations VALUES(37527,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(37527,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(37527,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(37527,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(37527,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(37527,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(37527,36,'cp 85','d1d2 b7b5 e1c1 c8b7 h2h4 h7h5 f2f3 b8d7 g1h3 c7c5 d4c5 d7c5 e3d4 g7f6 h3g5 d8c7 g2g3 a8b8 a2a3 b7a8 d2e3 f6d4 e3d4 g8f6 f1h3 e8g8 h1e1 f8d8 f3f4 e7e5 d4e3 a6a5 f4e5 d6e5 d1d8 b8d8 c3b5 c7b6');
INSERT INTO evaluations VALUES(37528,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(37528,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(37528,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(37528,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(37528,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(37528,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(37528,36,'cp 51','b7b5 f2f3 g8f6 e1c1 b8d7 h2h4 h7h5 g1h3 d7b6 h3g5 b5b4 c3e2 b6c4 d2d3 c4e3 d3e3 f6h7 e2f4 g7h6 g5h7 h8h7 f1c4 c7c6 e4e5 d6d5 c4e2 d8a5 c1b1 h6f4 e3f4 a5b6 g2g4 a6a5 f4e3 a5a4 d1e1 a4a3 b2b3 c8d7 g4h5 h7h5 f3f4 h5h8');
INSERT INTO evaluations VALUES(16344,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(16344,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(16344,32,'cp 230',NULL);
INSERT INTO evaluations VALUES(16344,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(16344,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(16344,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(16344,36,'cp 220','e5c6 b7c6 b1c3 f8b4 e4e5 d8e7 d1e2 f6d5 c3d5 c6d5 c2c3 b4a5 d2d4 e8g8 e2b5 a5b6 b5d5 a8b8 c1f4 c8b7 d5c4 d7d6 e1c1 d6e5 f4e5 b8d8 f1d3 b7d5 c4a4 c7c5 d4c5 b6c5 e5g3 e7g5 c1c2 d5g2 h1e1 g2f3 d1d2 d8d5 a4h4 g5h4 g3h4 a7a5 d3c4 d5f5 b2b3');
INSERT INTO evaluations VALUES(16234,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(16234,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(16234,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(16234,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(16234,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(16234,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(16234,36,'cp 214','b7c6 b1c3 f8b4 e4e5 d8e7 d1e2 f6d5 c3d5 c6d5 c2c3 b4a5 d2d4 c7c6 c1f4 e8g8 e2e3 a5c7 f1d3 a7a5 e1g1 c8a6 d3a6 a8a6 a1e1 e7e6 b2b3 a6a8 f4g3 f8e8 e3d2 e6g6 e1e3 d7d6 e5d6 e8e3 d2e3 c7d6 g3d6 g6d6 f1e1 h7h6');
INSERT INTO evaluations VALUES(11943,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(11943,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(11943,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(11943,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(11943,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(11943,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(11943,36,'cp 228','b1c3 f8b4 e4e5 d8e7 d1e2 f6d5 c3d5 c6d5 c2c3 b4a5 d2d4 c7c6 h2h4 a5c7 c1g5 e7e6 e1c1 d7d6 h4h5 h7h6 e5d6 c7d6 g5d2 e8g8 e2f3 a7a5 f1d3 e6g4 f3g4 c8g4 f2f3 g4e6 c1c2 a5a4 g2g4 f8b8 g4g5 h6g5 d2g5 f7f6 g5d2 g8f7');
INSERT INTO evaluations VALUES(19905,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(19905,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19905,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(19905,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19905,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19905,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19905,36,'cp 11','c8d7 c1e3 g7g6 b1c3 g8f6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 d1d2 d8a5 e1c1 f8g7 e3h6 e8g8 h6g7 g8g7 g2g4 f8b8 h2h4 h7h5 g4h5 f6h5 h1g1 a5b4 b2b3 a6a5 c1b2 a5a4 a2a3 b4c4 b3b4 c6c5 b4b5 a8a5 d2d3 b8b5 c3b5 a5b5 b2a1 c4d3 d1d3 h5f4 d3d2');
INSERT INTO evaluations VALUES(26028,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(26028,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(26028,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(26028,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(26028,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(26028,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(26028,36,'cp 16','c1e3 g7g6 b1c3 g8f6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 e1g1 d8a5 g1h1 f8g7 d1e1 e8g8 c3d5 a5d8 d5b6 a8b8 b6a4 c6c5 a1d1 c5c4 b2b4 d8d7 a4b6 d7c6 e1f2 f8e8 e3d4 f6d7 b6d7 c6d7 d4g7 g8g7 f2d4 f7f6 d4c4 d7b5 c4b5 b8b5 d1b1 e8c8 f1c1');
INSERT INTO evaluations VALUES(26026,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(26026,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(26026,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(26026,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(26026,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(26026,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26026,36,'cp 17','g7g6 b1c3 g8f6 f2f3 a7a6 b5e2 f8g7 d1d2 d8c7 e1c1 h7h5 c3d5 f6d5 e4d5 c6d4 e3d4 g7h6 d4e3 h6e3 d2e3 a8c8 e2d3 c7c5 d1e1 c5e3 e1e3 d7f5 d3f5 g6f5 h1e1 c8c7 e1e2 h5h4 a2a4 e8d8 c2c3 h8h5 f3f4 b7b5 a4b5 a6b5 b2b3');
INSERT INTO evaluations VALUES(29564,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(29564,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(29564,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(29564,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(29564,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29564,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29564,36,'cp 10','c1f4 e7e6 e2e3 c7c6 f1d3 f5g6 e1g1 f8e7 d1e2 f6h5 f4e5 e8g8 h2h3 h5f6 e5h2 d8b6 d3g6 h7g6 a2a4 f8c8 a4a5 b6a6 f1c1 a6e2 c3e2 c6c5 g2g4 b8c6 g4g5 f6e4 h3h4 c5d4 e2d4 e7d6 c2c3 a7a6 d4c6 b7c6');
INSERT INTO evaluations VALUES(29529,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29529,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(29529,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(29529,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29529,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(29529,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(29529,36,'cp 5','e7e6 e2e3 f8b4 f1d3 e8g8 e1g1 f5g6 c3e2 b4e7 c2c4 d5c4 d3c4 a7a6 a2a4 c7c5 d4c5 e7c5 a1c1 b8d7 d1b3 d8b6 f1d1 f8d8 e2d4 b6b3 c4b3 a8c8 f3e5 c5b4 c1c8 d8c8 e5g6 h7g6 g1f1 f6e4');
INSERT INTO evaluations VALUES(30081,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(30081,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30081,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(30081,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(30081,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(30081,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(30081,36,'cp 3','e2e3 f8d6 f3e5 b8d7 f1d3 e8g8 d3f5 e6f5 d1d3 g7g6 e5d7 d8d7 f4g5 d6e7 g2g3 f8c8 g5f6 e7f6 a1d1 a7a5 e1g1 g8g7 b2b3 f6e7 h2h4 c8e8 c3e2 c7c6 c2c4 a5a4 g1g2 a4b3 a2b3 a8a3 e2f4 e7d6 f4h3 a3a2');
INSERT INTO evaluations VALUES(30080,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30080,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30080,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30080,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30080,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(30080,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30080,36,'cp 0','f8d6 f3e5 b8d7 f1d3 e8g8 d3f5 e6f5 d1d3 g7g6 e5d7 d8d7 f4g5 d6e7 g2g3 a7a5 a2a4 c7c6 g5f6 e7f6 h2h4 f6e7 b2b3 f8b8 e1f1 e7d6 f1g2 h7h6 c3e2 b7b5 e2f4 d6c7 f4e2');
INSERT INTO evaluations VALUES(28449,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(28449,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28449,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(28449,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28449,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28449,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(28449,36,'cp 12','f3e5 b8d7 f1d3 e8g8 e1g1 d8e7 d3f5 e6f5 d1d3 d7e5 d4e5 d6e5 c3d5 f6d5 d3d5 e5f4 e3f4 e7b4 b2b3 g7g6 a1d1 c7c6 d5d2 a7a5 f1e1 b4d2 d1d2 f8d8 e1d1 d8d2 d1d2 g8f8 d2d7 a8e8 g1f1 e8e7 d7d8 e7e8 d8e8 f8e8 f1e2 e8e7 e2d3 e7d6 d3d4 b7b6 a2a4');
INSERT INTO evaluations VALUES(47026,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(47026,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(47026,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47026,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(47026,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47026,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47026,36,'cp 31','e7e5 g1f3 e5d4 f3d4 g7g6 c1g5 f8g7 e1g1 e8g8 h2h3 f8e8 b1c3 a7a5 f1e1 b8d7 d1d2 c7c6 b2b3 d7c5 g5h6 g7h6 d2h6 c8d7 h6f4 f6h5 f4d6 c5d3 c2d3 d7h3 d6d8 e8d8 d4e2 h3e6 a1d1 f7f5 c3a4 h5g7 a4c5 b7b6 c5e6 g7e6 d3d4');
INSERT INTO evaluations VALUES(47853,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(47853,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(47853,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(47853,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(47853,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(47853,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(47853,36,'cp 33','d4d5 b8d7 g1e2 f8e7 c2c4 d7c5 b1c3 c5d3 d1d3 a7a5 c1e3 f6g4 e3d2 g4f6 f2f3 f6d7 c3a4 e8g8 e2c3 d7b8 g2g4 e7h4 e1d1 h4f2 d1c2 c8d7 a1f1 f2d4 c2b1 b8a6 d2e3 d4e3 d3e3 d7a4 c3a4 b7b6 f1f2 d8h4 b2b3 a6c5');
INSERT INTO evaluations VALUES(47764,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(47764,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(47764,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47764,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47764,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(47764,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(47764,36,'cp 28','b8d7 c1e3 c7c6 c2c4 d7c5 e3c5 d6c5 b1c3 f8d6 g1f3 e8g8 a2a4 c8d7 f3d2 d6c7 e1g1 g7g6 f1e1 a7a5 d2f3 d8e7 d1d2 a8d8 a1d1 f6e8 d2h6 d7g4 d3e2 g4f3 e2f3 e8d6 b2b3 f7f5 h6e3 d6f7 e4f5 g6f5 f3h5 e5e4');
INSERT INTO evaluations VALUES(15903,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(15903,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(15903,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(15903,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(15903,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(15903,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(15903,36,'cp 37','c1e3 c7c6 c2c4 d7c5 e3c5 d6c5 g1f3 f8d6 b1c3 c8d7 a2a4 e8g8 e1g1 f6e8 d1e2 e8c7 a4a5 c7a6 f1d1 d8f6 e2e3 a6b4 d1d2 a8d8 h2h3 g7g6 d3e2 f6e7 e3h6 f7f6 a1d1 b4a6 h3h4 e7g7 h6g7 g8g7 h4h5');
INSERT INTO evaluations VALUES(19373,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(19373,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(19373,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(19373,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(19373,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(19373,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(19373,36,'cp -14','e7e6 d2d4 d7d5 e4d5 e6d5 c4b3 c5c4 b3c2 f8d6 g1f3 g8e7 e1g1 e8g8 f1e1 c8f5 c1g5 f5c2 d1c2 b7b5 a2a3 d8d7 g5h4 e7g6 h4g3 d6f4 b1d2 a7a6 d2f1 a8e8 h2h4 h7h5 e1e8 f8e8 a1e1 e8e6 g3f4 g6f4 e1e6');
INSERT INTO evaluations VALUES(18738,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(18738,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(18738,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(18738,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(18738,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(18738,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(18738,36,'cp -11','d2d4 d7d5 e4d5 e6d5 c4b3 c5c4 b3c2 f8d6 g1f3 g8e7 e1g1 e8g8 f1e1 b7b5 b1d2 c8f5 b2b3 d8d7 b3c4 b5c4 a1b1 f5c2 d1c2 e7g6 d2f1 f8b8 b1b8 a8b8 h2h4 h7h6 g2g3 c6e7 f1e3 d7h3 e3g2 e7f5 g2e3 f5e3 c1e3 g6e7 e3f4 e7f5 f4d6');
INSERT INTO evaluations VALUES(18760,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(18760,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(18760,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(18760,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(18760,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(18760,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(18760,36,'cp -22','d7d5 e4d5 e6d5 c4b3 c5c4 b3c2 f8d6 g1f3 g8e7 e1g1 b7b5 f1e1 e8g8 b1d2 c8f5 b2b3 h7h6 b3c4 b5c4 a1b1 d8d7 d2f1 a8e8 c1d2 f5c2 d1c2 e7f5 g2g3 g7g5 e1e8 f8e8 b1e1 e8e1 f3e1 c6e7 e1g2 e7g6 f1e3 f5e7 e3f1');
INSERT INTO evaluations VALUES(43411,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(43411,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43411,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(43411,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43411,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43411,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(43411,36,'cp 15','e7e5 d4e2 f8e7 h2h3 c8e6 e1g1 e8g8 f2f4 d8c7 e2g3 b8d7 g1h1 b7b5 c1e3 e5f4 e3f4 d7e5 c3e2 e5g6 f4g5 a8c8 d1d2 f6d7 g5e7 g6e7 g3h5 c7c5 e2g3 g8h8 a1d1 e7g6 b2b3 g6e5 g3f5 e6f5 e4f5 e5d3 d2d3');
INSERT INTO evaluations VALUES(43454,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(43454,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(43454,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(43454,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(43454,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43454,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(43454,36,'cp 13','d4e2 f8e7 e1g1 e8g8 h2h3 c8e6 f2f4 d8c7 e2g3 b8d7 g1h1 b7b5 a2a3 e5f4 c1f4 d7e5 c3e2 e5g6 f4g5 f6h5 g5d2 h5g3 e2g3 g6e5 d2c3 c7c5 g3f5 e6f5 f1f5 g7g6 f5f1 f8e8 d1d2 e7f8 a3a4 f8g7 d3e2 c5c6 a4b5 a6b5 a1a8 e8a8 f1d1');
INSERT INTO evaluations VALUES(43471,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(43471,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43471,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43471,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43471,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(43471,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43471,36,'cp 11','f8e7 e1g1 e8g8 h2h3 c8e6 f2f4 d8c7 g1h1 b8d7 e2g3 b7b5 a2a3 f8e8 f4f5 e6c4 c1g5 a8c8 g5f6 d7f6 d1f3 a6a5 c3b5 c7b7 a3a4 d6d5 a1d1 c4d3 c2d3 c8c2 g3h5 e8d8 h5f6 e7f6 b5c3 d5e4 d3e4 d8d1 f1d1 b7b2 c3d5 c2d2 f3g4 g8h8 d5f6 d2d1 g4d1');
INSERT INTO evaluations VALUES(39686,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(39686,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39686,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(39686,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(39686,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(39686,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(39686,36,'cp 27','h2h3 e8g8 e1g1 c8e6 f2f4 d8c7 g1h1 b7b5 e2g3 b8d7 d1f3 e5f4 c1f4 d7e5 f3f2 b5b4 c3e2 d6d5 e2d4 e7d6 a1c1 c7b6 g3f5 e6f5 d4f5 b6f2 f1f2 e5d3 c2d3 d6f4 f5e7 g8h8 f2f4 d5e4 d3e4 f8d8 e4e5 f6d5 e7d5 d8d5 c1c7 d5e5 f4b4 e5e1 h1h2');
INSERT INTO evaluations VALUES(49398,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49398,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49398,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49398,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49398,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49398,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49398,36,'cp 3','d7d5 e2e3 f8e7 g1f3 e8g8 b2b3 b7b6 c1b2 c7c5 c4d5 f6d5 d4c5 e7c5 d1c2 c5e7 b1c3 d5c3 b2c3 e7f6 a1c1 c8b7 f1e2 b8d7 e1g1 a8c8 c2b2 b7d5 b3b4 a7a5 c3f6 d8f6 b2d2 a5b4 a3b4 d5f3 e2f3 d7e5 f3e2 c8c1');
INSERT INTO evaluations VALUES(27204,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(27204,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(27204,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(27204,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(27204,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(27204,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(27204,36,'cp 83','d2d4 e5d4 f3d4 f8e7 d4c6 b7c6 b5d3 e7f6 e1g1 g8e7 f2f4 e8g8 d1e1 f8e8 c1e3 a8b8 e3a7 b8b2 e4e5 e7g6 e1c1 b2b7 e5f6 b7a7 f4f5 g6e5 f6g7 d8h4 c3e4 g8g7 a2a4 c6c5 f1f4 h4h6 a4a5 f7f6 a5a6 g7h8 c1d2 e5d3 c2d3 d7c6 e4g3 h6g5');
INSERT INTO evaluations VALUES(27198,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(27198,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(27198,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(27198,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(27198,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(27198,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27198,36,'cp 67','e5d4 f3d4 g8f6 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 a7a5 d1f3 e8g8 b2b3 a5a4 c3a4 d6d5 e4e5 f6e4 f4f5 f8e8 c1f4 e7g5 f5f6 g5f4 f3f4 e8e6 f6g7 d8e7 f1f3 e6e5 a1f1 f7f6 a4c3 e4c3 f3c3 e5e4 f4f6 e7f6 f1f6');
INSERT INTO evaluations VALUES(27227,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(27227,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(27227,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(27227,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(27227,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(27227,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(27227,36,'cp 66','f3d4 g8f6 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 d8d7 h2h3 f8b8 a1b1 d7e6 e3f2 e7f8 b2b3 a5a4 b1e1 a4b3 c2b3 f6d7 g1h1 b8d8 e1d1 d8e8 f2h4 d7c5 f1e1 c5d7 e1e2 f8e7 h4g3 a8d8 e2e1 e7f8 b3b4');
INSERT INTO evaluations VALUES(23017,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(23017,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23017,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(23017,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(23017,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(23017,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(23017,36,'cp 22','c7c6 b5a4 g8f6 c2c4 d4c3 b2c3 d7d6 e1g1 g7g6 b1d2 f8g7 d2f3 b7b5 a4b3 e8g8 a2a4 a7a5 c1a3 f8e8 f1e1 b5a4 b3a4 d8c7 h2h3 c8d7 a3c1 c6c5 c1f4 d7a4 a1a4 f6d7 d1a1 d7e5 f3e5 g7e5');
INSERT INTO evaluations VALUES(19098,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(19098,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(19098,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(19098,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(19098,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(19098,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(19098,36,'cp 28','b5a4 g8f6 c2c4 f8e7 b1d2 e8g8 e1g1 d7d6 d2f3 d8b6 a2a3 a7a5 h2h3 f6d7 c1f4 d7c5 a1b1 c5a4 d1a4 b6a7 e4e5 d6e5 f4e5 c8f5 e5d4 e7c5 d4c5 a7c5 f1e1 f5d3 b1d1 d3g6 d1d7 c5b6');
INSERT INTO evaluations VALUES(19116,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(19116,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19116,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(19116,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(19116,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(19116,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(19116,36,'cp 27','g8f6 c2c4 d4c3 b2c3 d7d6 a4c2 g7g6 h2h3 f8g7 d3d4 e8g8 e1g1 b7b5 c1g5 h7h6 g5h4 c8b7 a2a4 d8c7 b1d2 f6h5 g2g4 h5f6 h4g3 a7a6 a1b1 f6d7 d2f3 a8c8 f1e1 f8e8 d1d3 c6c5 a4b5 c5d4 b5a6 b7a8');
INSERT INTO evaluations VALUES(38704,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38704,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38704,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(38704,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(38704,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(38704,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(38704,36,'cp 39','c2c3 g8f6 d2d4 e5d4 c3d4 c5b4 b1d2 b4d2 f3d2 d6d5 e4d5 e8g8 e1g1 b8d7 d5d6 c7d6 h2h3 f8e8 c4d3 h7h6 d1f3 d8b6 d2b3 a7a5 c1e3 a5a4 b3d2 d6d5 b2b3 a4b3 d2b3 b6d8 f1c1 d7f8 e3f4 f6e4 a2a4 f8g6 f4g3 e4g5 f3d1 d8f6 b3c5');
INSERT INTO evaluations VALUES(38698,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(38698,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(38698,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38698,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(38698,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(38698,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(38698,36,'cp 43','g8f6 d2d4 e5d4 c3d4 c5b4 b1d2 b4d2 f3d2 e8g8 e1g1 d6d5 e4d5 b8d7 d5d6 c7d6 c4d3 f8e8 h2h3 h7h6 d1f3 d6d5 d2b1 d7f8 b1c3 f8e6 c1e3 d8d6 a1d1 c8d7 f3g3 d6g3 f2g3 d7c6 g3g4 a7a6 g1h2 a8d8 d1e1 f6e4 d3c2 e4c3');
INSERT INTO evaluations VALUES(41861,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(41861,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(41861,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(41861,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(41861,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(41861,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41861,36,'cp 47','d2d4 e5d4 c3d4 c5b4 b1d2 e8g8 e1g1 b8c6 d4d5 c6e5 f3e5 d6e5 d1c2 f6h5 c4e2 h5f4 d2c4 f4e2 c2e2 f7f6 c1e3 f6f5 f2f3 b4d6 a1c1 c8d7 h2h3 d7b5 e2c2 f5e4 f3e4 f8f1 c1f1 d8e8 b2b3 e8g6 a2a4 b5c4 c2c4 h7h6 a4a5');
INSERT INTO evaluations VALUES(47783,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(47783,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(47783,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(47783,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(47783,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(47783,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(47783,36,'cp -22','d6d5 e4d5 h7h6 g5f3 e5e4 f3e5 d8e7 d2d4 e4d3 f2f4 b8d7 e1g1 d7e5 f4e5 e7c5 g1h1 f6d5 c4d5 c5d5 b1c3 d5c6 c2d3 c8e6 c1e3 e8c8 a1c1 c6a6 d3d4 f8b4 a2a3 b4c3 c1c3 e6d5 h1g1 a6e6 d1c2 d8d7 f1c1 c7c6 h2h3 h8d8');
INSERT INTO evaluations VALUES(48775,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48775,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48775,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48775,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(48775,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48775,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48775,36,'cp -18','e4d5 h7h6 g5f3 e5e4 f3e5 d8e7 d2d4 e4d3 f2f4 f6g4 c4d3 g4e5 f4e5 e7e5 d1e2 f8d6 e2e5 d6e5 b1d2 e8g8 d2c4 f8e8 e1f2 e5d4 c1e3 d4e3 c4e3 b8d7 h1d1 d7f6 h2h3 a7a5 c2c4 g8f8 e3c2 c7c6 d5c6 b7c6 c2d4 c6c5 d4b5 e8e7 d1d2 c8e6 a1d1 a5a4 b5d6 e7d7 d6e4 f6e4 d3e4 d7d2 d1d2');
INSERT INTO evaluations VALUES(48585,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(48585,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(48585,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(48585,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(48585,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(48585,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(48585,36,'cp -14','h7h6 g5f3 e5e4 f3e5 d8e7 d2d4 e4d3 f2f4 f6g4 c4d3 g4e5 f4e5 e7e5 d1e2 f8d6 b1d2 e8g8 e2e5 d6e5 d2c4 f8e8 e1f2 e5d4 c1e3 d4e3 c4e3 b8d7 h1d1 d7f6 h2h3 a7a5 c2c4 c8d7 b2b3 a8d8 a1c1 g8f8 c1c3 c7c6 d3c2 c6d5 e3d5 f6d5 d1d5 d7c6 d5d1 d8d1 c2d1 f8e7 d1f3');
INSERT INTO evaluations VALUES(47705,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(47705,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(47705,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(47705,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(47705,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(47705,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(47705,36,'cp -8','g5f3 e5e4 f3e5 d8e7 d2d4 e4d3 f2f4 b8d7 e1g1 e7c5 g1h1 d7e5 f4e5 f6d5 c4d5 c5d5 b1c3 d5c6 c2d3 c8e6 c1e3 e8c8 a1c1 c6a6 d3d4 f8b4 a2a3 b4c3 c1c3 e6d5 h1g1 a6g6 d1e2 c8b8 b2b4 h8e8 b4b5 d8d7 e2f2 b7b6 a3a4');
INSERT INTO evaluations VALUES(53287,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(53287,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(53287,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(53287,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(53287,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(53287,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(53287,36,'cp 141','e4d5 g8f6 f1b5 c8d7 b5c4 b7b5 c4e2 f6d5 b1c3 e7e6 e1g1 f8e7 c3b5 e8g8 d2d4 a7a6 b5c3 d5c3 b2c3 d7a4 e2d3 b8d7 d1e2 d8c7 d4d5 c5c4 d3c4 d7b6 c4d3 e6d5 c1f4 c7f4 e2e7 a4b5 e7b4 f4c7 a2a4 b5d3 c2d3 b6d7 b4d4 f8c8 g2g3 c7c3 d4d5 c3c6 d5c6 c8c6 d3d4 d7f6');
INSERT INTO evaluations VALUES(53235,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(53235,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(53235,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(53235,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(53235,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(53235,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(53235,36,'cp 118','g8f6 f1b5 c8d7 b5c4 b7b5 c4e2 f6d5 b1c3 e7e6 e1g1 f8e7 c3b5 e8g8 c2c4 a7a6 b5c3 d5c3 d2c3 d8c7 c1g5 e7g5 f3g5 h7h6 g5e4 f8d8 d1d2 d7c6 d2e3 b8d7 b2b3 c6e4 e3e4 a6a5 g2g3 a8b8 f1d1 d7f6 e4c2 d8d1 a1d1 b8d8 d1d8 c7d8 e2f3 d8c7 c2e2 f6d7 e2d1 d7f6');
INSERT INTO evaluations VALUES(45333,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(45333,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(45333,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(45333,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(45333,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(45333,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(45333,36,'cp 128','f1b5 c8d7 b5c4 b7b5 c4e2 f6d5 b1c3 e7e6 e1g1 d5c3 d2c3 d8c7 f1e1 f8d6 d1d3 a7a6 e1d1 d6e7 a2a4 b5a4 d3e4 d7c6 e4e5 c7e5 f3e5 c6b5 c3c4 b5c6 e5c6 b8c6 e2f3 a8c8 a1a4 c6d4 f3e4 d4e2 g1f1 e2c1 d1c1 e7f6 c2c3 c8b8 a4a6 b8b2 a6a8 f6d8');
INSERT INTO evaluations VALUES(49595,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(49595,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(49595,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(49595,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(49595,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(49595,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(49595,36,'cp 169','d4e5 f6e4 d1e2 e4c5 b1c3 b7b6 c1e3 b8c6 e1c1 c8a6 e2e1 a6f1 e1f1 c5e6 f2f4 f8c5 f1e2 c5e3 e2e3 d8h4 c3d5 e8c8 g1f3 h4h5 h1e1 h8e8 g2g3 c8b7 e3f2 b7b8 a2a3 c6e7 d5c3 f7f6 f2f1 e7c6 e5f6 g7f6 c3d5 h5g6 f3h4 g6f7 h4f5 e6c5 f1c4 e8e1 d1e1 d8e8');
INSERT INTO evaluations VALUES(49505,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(49505,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(49505,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(49505,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(49505,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(49505,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(49505,36,'cp 191','f6e4 d1e2 e4c5 b1c3 f8e7 c1e3 d7d5 e5d6 d8d6 a1d1 d6a6 c3d5 a6e2 g1e2 b8a6 d5e7 e8e7 f2f3 h8d8 d1d8 e7d8 e1f2 d8e7 e2g3 b7b6 a2a3 g7g6 h2h4 c5e6 h4h5 a6c5 f1c4 c8a6 e3c5 e6c5 c4a6 c5a6');
INSERT INTO evaluations VALUES(49863,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(49863,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(49863,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(49863,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(49863,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(49863,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(49863,36,'cp 180','d1e2 e4c5 b1c3 f8e7 c1e3 d7d5 e5d6 d8d6 a1d1 d6a6 c3d5 a6e2 g1e2 b8a6 d5e7 e8e7 f2f3 h8d8 d1d8 e7d8 e1f2 d8e8 a2a3 c5e6 e2c3 c8d7 f1c4 a6c5 h1d1 d7c6 h2h4 h7h6 h4h5 c5a4 c3a4 c6a4 b2b3 a4c6 b3b4 a8d8 c4d3 c6a4 b4b5 e6d4');
INSERT INTO evaluations VALUES(49862,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(49862,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(49862,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(49862,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(49862,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(49862,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(49862,36,'cp 200','e4c5 b1c3 b7b5 c3b5 c8a6 e2d1 b8c6 g1f3 a8b8 a2a4 a6b7 f1c4 a7a6 b5c3 d7d6 d1e2 c6e5 f3e5 d6e5 e2e5 d8e7 e5e7 f8e7 e1g1 e8g8 c1e3 c5e6 a1d1 e7f6 d1d3 b7c6 b2b3 f8d8 c3d5 c6d5 d3d5 d8d5 c4d5 e6d4 d5e4');
INSERT INTO evaluations VALUES(39100,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(39100,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(39100,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(39100,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(39100,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(39100,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(39100,36,'cp 60','c7c5 d2d3 h7h6 f3g1 b8c6 b1c3 a7a6 a2a4 g8f6 g1e2 e8g8 e1g1 c6b4 e2g3 d6c7 g3h5 d7d6 f2f4 d6d5 c3d5 f6d5 e4d5 b4d5 f4f5 d5f4 h5f4 e5f4 d1g4 d8d4 g1h1 d4f6 c1f4 c8f5 g4f3 c7f4 f3f4 f5e6 f4f6 g7f6 f1f6 e6c4 d3c4 f8e8 a1f1 e8e2 f6f7 e2c2 f7b7 c2c4 b7e7');
INSERT INTO evaluations VALUES(52414,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(52414,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(52414,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(52414,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(52414,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(52414,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(52414,36,'cp -47','c5b4 a2a3 d7d5 e4d5 d8d5 g1f3 e7e6 d2d4 d5d8 f1d3 g8f6 e1g1 b8c6 a3b4 c6b4 d3e2 f8d6 b1a3 c8d7 c2c3 b4d5 d1b3 a7a5 f3e5 f6e4 c3c4 d5c3 e2f3 e8g8 f3e4 c3e4 b3b7 e4c5 b7b2 a8b8 a3b5 d7b5 c4b5');
INSERT INTO evaluations VALUES(52500,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(52500,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(52500,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(52500,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(52500,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(52500,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(52500,36,'cp -51','g1f3 d7d5 e4d5 g8f6 c2c4 b4c3 b1c3 f6d5 f1b5 c8d7 d1b3 d5c3 d2c3 e7e6 b5d7 b8d7 b3b7 f8c5 c1f4 e8g8 e1g1 d7b6 a1d1 b6d5 f4e5 d8b6 d1b1 b6b7 b1b7 a8c8 h2h3 f7f6 e5d4 f8f7 b7b5 c5d4 c3d4 c8c2 a2a3 d5f4 b5b3 f4e2 g1h2');
INSERT INTO evaluations VALUES(34617,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(34617,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(34617,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(34617,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(34617,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(34617,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(34617,36,'cp 45','e5e2 f1e2 g8f6 g1f3 a7a6 d2d4 c8f5 c1f4 c7c6 e1c1 b8d7 f3h4 f5e6 h1e1 g7g6 h4f3 f8g7 f3g5 e6d5 f2f3 b7b5 e2f1 e7e6 g5e4 e8g8 c1b1 h7h5 f4g5 f8c8 h2h3 b5b4 c3a4 a6a5 c2c3 b4c3 e4c3 c8c7 c3d5 f6d5');
INSERT INTO evaluations VALUES(34665,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(34665,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(34665,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(34665,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(34665,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(34665,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(34665,36,'cp 37','f1e2 g8f6 g1f3 a7a6 d2d4 c8f5 c1f4 c7c6 e1c1 b8d7 f3h4 f5e6 h1e1 g7g6 h4f3 f8g7 f3g5 e6d5 f2f3 e8g8 g2g4 b7b5 h2h4 d7b6 c1b1 a6a5 f4c1 e7e6 g5e4 d5c4 e2c4 b6c4 e4f6 g7f6 h4h5 c4d6 a2a4 g6h5');
INSERT INTO evaluations VALUES(34663,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(34663,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(34663,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(34663,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(34663,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34663,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(34663,36,'cp 27','g8f6 g1f3 a7a6 d2d4 c8f5 c1f4 c7c6 f3e5 e7e6 e5c4 b8d7 a1c1 f6d5 c3d5 e6d5 c4e3 f5e6 h2h4 h7h5 e1d2 d7f6 f2f3 e8d7 b2b3 f6e8 c1b1 f8b4 c2c3 b4d6 f4d6 e8d6 g2g4 d7c7');
INSERT INTO evaluations VALUES(33722,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(33722,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(33722,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(33722,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(33722,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(33722,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(33722,36,'cp 37','g1f3 a7a6 d2d4 c8f5 c1f4 c7c6 e1c1 b8d7 f3h4 f5e6 f4d2 b7b5 h1e1 d7b6 h4f3 f6g4 e1f1 e6c4 h2h3 g4f6 f3e5 c4e2 c3e2 a8c8 b2b3 c6c5 d4c5 c8c5 e5d3 c5d5 f2f3 e7e6 d2e3 f6d7');
INSERT INTO evaluations VALUES(33721,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(33721,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(33721,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(33721,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(33721,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(33721,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(33721,36,'cp 16','a7a6 f3e5 c8f5 f2f4 b8d7 d2d3 d7e5 f4e5 f6d7 c1f4 e7e6 e1c1 h7h5 e2f3 e8c8 h1e1 h5h4 h2h3 f8e7 e1e2 g7g5 f4d2 d8g8 a2a3 g8g7 d1h1 g5g4 h3g4 f5g4 f3g4 g7g4 h1f1');
INSERT INTO evaluations VALUES(53943,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(53943,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(53943,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(53943,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(53943,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(53943,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(53943,36,'cp 74','d4f3 g8f6 f1d3 g7g6 c2c4 b8c6 b1c3 f8g7 e1g1 e8g8 f1e1 f6d7 c1g5 f7f6 g5e3 d7c5 d3c2 c8g4 h2h3 g4f3 d1f3 c5e6 c3e2 c6b4 c2b1 d8e7 h3h4 h7h5 e2c3 b4c6 b1d3 a7a5 d3f1 g8h7 c3d5 e7f7 f3h3 c6d4 a1d1 g7h6 e3h6');
INSERT INTO evaluations VALUES(53949,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(53949,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(53949,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(53949,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(53949,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(53949,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(53949,36,'cp 57','g8f6 f1d3 b8c6 e1g1 g7g6 c2c4 f8g7 b1c3 e8g8 a1b1 f6e8 c3d5 c8g4 h2h3 g4f3 d1f3 e8c7 d5c7 c6d4 f3g4 d8c7 c1e3 h7h5 g4d1 d4e6 h3h4 g8h7 d1d2 b7b6 b1d1 a8d8 g2g3 d8c8 b2b3 c7e7 d3e2 f8d8 e2f3 e6c5 f3g2 c5e6 a2a4 d8d7');
INSERT INTO evaluations VALUES(45955,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(45955,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(45955,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(45955,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(45955,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(45955,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(45955,36,'cp 66','f1d3 b8c6 c2c4 g7g6 b1c3 f8g7 e1g1 e8g8 f1e1 f6d7 c3d5 d7c5 b2b4 c5e6 a1b1 a7a5 b4b5 c6b4 a2a4 b4d3 d1d3 e6c5 d3a3 c8e6 c1e3 a8c8 f3d2 f7f5 e1c1 b7b6 f2f3 g8h8 g1h1 f8f7 c1e1 g7f8 b1c1 f7d7 h2h3 e6d5 c4d5');
INSERT INTO evaluations VALUES(45954,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(45954,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(45954,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(45954,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(45954,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(45954,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(45954,36,'cp 74','b8c6 e1g1 g7g6 h2h3 f8g7 f1e1 e8g8 c2c4 f6d7 b1c3 d7c5 d3f1 c6d4 c1e3 c5e6 c3b5 d4b5 c4b5 b7b6 d1d5 a8b8 a1d1 c8b7 d5d6 d8d6 d1d6 b7e4 f3g5 e4f5 f1c4 b8d8 e1d1 d8d6 d1d6 e6g5 e3g5 f8c8 c4d5 h7h6 g5e3 g7f8 d6c6 g6g5 e3d2 f8g7');
INSERT INTO evaluations VALUES(13894,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(13894,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(13894,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(13894,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(13894,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(13894,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(13894,36,'cp 58','c2c4 g7g6 b1c3 f8g7 h2h3 e8g8 e1g1 f6d7 c3d5 d7c5 b2b4 c5e6 b4b5 c6d4 f3d4 e6d4 a2a4 c8e6 a4a5 d8d7 a1a3 e6d5 c4d5 f8c8 c1e3 h7h5 d1a4 d7d8 a4b4 c8c5 f1a1 g7f6 b5b6 a7b6 b4b6 d8b6 a5b6 a8a3 a1a3');
INSERT INTO evaluations VALUES(28957,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(28957,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(28957,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(28957,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28957,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(28957,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28957,36,'cp 22','e7e6 f3h4 f5g6 f1e2 b8d7 e1g1 f8d6 g2g3 e8g8 h4g6 h7g6 d1c2 a7a6 f1d1 d8e7 a2a3 f8e8 c1d2 d5c4 e2c4 b7b5 c4f1 a8c8 c3e4 f6e4 c2e4 e6e5 d4e5 d6e5 a1b1 c6c5 d2a5 d7f6 e4f3 e7e6 f1g2 c5c4 f3b7');
INSERT INTO evaluations VALUES(29241,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(29241,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29241,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(29241,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(29241,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(29241,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(29241,36,'cp 18','f3h4 f5g6 f1e2 b8d7 e1g1 f8d6 g2g3 e8g8 h4g6 h7g6 d1c2 d5c4 e2c4 b7b5 c4e2 a8c8 c3e4 f6e4 c2e4 b5b4 b2b3 d7f6 e4d3 c6c5 c1b2 c5d4 b2d4 d6c5 d4b2 d8d3 e2d3 f8d8 f1d1 c5b6 d1d2 f6d5 d3a6 d5c3 d2d8 c8d8');
INSERT INTO evaluations VALUES(29234,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(29234,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29234,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(29234,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(29234,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(29234,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(29234,36,'cp 8','f5g6 f1e2 b8d7 e1g1 f8d6 g2g3 a7a6 a2a4 e8g8 h4g6 h7g6 d1c2 f8e8 a4a5 d5c4 e2c4 d6b4 c2b3 c6c5 d4d5 e6d5 c3d5 d7e5 d5b4 c5b4 c4e2 d8d7 a1a4 e5c6 f1d1 d7f5 f2f3 a8d8 e3e4 d8d1 e2d1 f6e4 f3e4 f5e4');
INSERT INTO evaluations VALUES(29317,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(29317,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29317,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(29317,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29317,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(29317,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(29317,36,'cp 17','f1e2 a7a6 e1g1 b8d7 g2g3 f8d6 a2a4 e8g8 h4g6 h7g6 c1d2 a6a5 d1b3 d8c7 f1d1 d6e7 b3c2 e7b4 h2h4 d7b6 b2b3 c7e7 d2c1 d5c4 b3c4 e6e5 a1b1 f8e8 e2f1 b4c3 c2c3 e5d4 e3d4 b6a4 c3c2');
INSERT INTO evaluations VALUES(56033,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(56033,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(56033,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(56033,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(56033,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(56033,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(56033,36,'cp 76','g8f6 c1g5 f8e7 d1d2 b8d7 e1c1 a7a6 f1d3 c7c5 h2h4 c5d4 f3d4 d7c5 d3e2 c8d7 d2e3 d8a5 g2g4 h7h6 g5f4 c5a4 d4b3 a5b6 c3a4 b6e3 f4e3 d7a4 e2f3 f6d5 f3d5 e6d5 e3c5 e7c5 b3c5 a4c6 h1e1 e8f8 f2f4 a8e8 h4h5 e8e1 d1e1 g7g6 h5g6 f7g6');
INSERT INTO evaluations VALUES(18973,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(18973,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(18973,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(18973,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(18973,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(18973,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(18973,36,'cp 25','e1g1 f8e7 c2c3 g8f6 d2d4 c5d4 c3d4 e5d4 e4e5 f6g4 f1e1 d7d6 e5d6 d8d6 c1g5 g4f6 b1c3 e8g8 c3b5 d6d8 b5d4 c6d4 d1d4 d8d4 f3d4 e7c5 d4b3 c5b4 e1d1 b4e7 g5e3 c8e6 c4e6 f7e6 b3c5 e7c5 e3c5 f8c8 a1c1 g8f7 g2g3 f6d5 g1g2 b7b6 c5a3 c8c1');
INSERT INTO evaluations VALUES(18972,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(18972,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(18972,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(18972,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(18972,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(18972,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(18972,36,'cp 21','f8e7 d2d4 e5d4 c2c3 g8f6 c3d4 c5d4 e4e5 f6g4 f1e1 d7d6 e5d6 d8d6 c1g5 g4f6 b1c3 e8g8 c3b5 d6d8 b5d4 c6d4 d1d4 d8d4 f3d4 e7c5 d4b3 c5b4 e1d1 b4e7 a1c1 h7h6 g5d2 c8f5 d2e3 f8d8 d1d8 e7d8 b3d4 f5d7 d4f3 f6g4 e3d4 d7c6 b2b3 c6f3 g2f3 d8f6 d4a7 a8a7');
INSERT INTO evaluations VALUES(39552,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(39552,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(39552,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(39552,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(39552,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(39552,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(39552,36,'cp 25','g1f3 c7c5 c2c3 e7e6 d2d4 c5d4 e1g1 g8e7 c4b3 d7d5 e4d5 e7d5 c1g5 d8a5 f3d4 b8c6 d4c6 b7c6 g5h4 c8a6 f1e1 e8g8 d1f3 d5b6 b1d2 c6c5 h4f6 g7f6 f3f6 c5c4 b3c2 a8b8 b2b4 b6d5 f6d4 a5c7 d2e4 f8d8 e4f6 d5f6 d4f6 a6b7 a1d1 d8d1 c2d1');
INSERT INTO evaluations VALUES(51789,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51789,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(51789,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51789,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(51789,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(51789,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51789,36,'cp 55','e2e4 d7d5 b1c3 d5e4 c3e4 f8b4 c1d2 d8d4 d2b4 d4e4 f1e2 b8a6 b4a5 f7f6 g1f3 b7b6 f3d2 e4f4 e2h5 g7g6 h5f3 g8e7 a5c3 e8g8 d2e4 a6c5 g2g3 f4f5 d1e2 c8a6 e1g1 e7d5 e4c5 b6c5 f3g4 d5c3 b2c3 f5e5 e2e5 f6e5 f1e1 a6c4 e1e5 c4d5 g4e6 d5e6 e5e6 a8d8 e6c6 d8d2 f2f4 f8b8 c6c5 b8b2 c5c8 g8g7 c8c7 g7f8 c7h7');
INSERT INTO evaluations VALUES(51791,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(51791,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(51791,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(51791,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(51791,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(51791,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(51791,36,'cp 18','d7d5 c4d5 e6d5 e4e5 c6c5 g1f3 c5d4 f1b5 c8d7 f3d4 f8c5 c1e3 c5d4 b5d7 d8d7 e3d4 g8e7 b1c3 b8c6 e1g1 e8g8 f2f4 a8c8 a1c1 c6d4 d1d4 b7b6 c1d1 f8d8 d1d2 c8c4 d4f2 d5d4 c3e4 e7f5 f1d1');
INSERT INTO evaluations VALUES(53586,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(53586,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53586,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(53586,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(53586,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(53586,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(53586,36,'cp 35','b1c3 d5e4 c3e4 f8b4 c1d2 d8d4 d2b4 d4e4 f1e2 b8a6 b4a5 b7b6 d1d6 e6e5 a1d1 c8f5 g1f3 g8e7 a5c3 a6c5 f3e5 e8g8 f2f3 e4h4 g2g3 h4h3 d6e7 f7f6 b2b4 c5a4 c4c5 a8e8 e2c4 f5e6 c4e6 h3e6 e7e6 e8e6 d1d6 e6e8 e1g1 f6e5 c3a1 b6c5 b4c5 a4c5 d6c6 c5d3 c6a6 e8c8 a6a7 c8a8 a7d7 d3c5 d7d6 e5e4 d6c6 c5d3');
INSERT INTO evaluations VALUES(34556,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(34556,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(34556,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(34556,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(34556,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(34556,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(34556,36,'cp 20','b8c6 g1f3 e6e2 f1e2 c8f5 e2b5 e7e6 b5c6 b7c6 d2d3 f8d6 c1d2 h7h5 f3d4 g8e7 e1c1 f7f6 c3e4 e8f7 h2h3 g7g5 h1e1 h5h4 e1e2 h8d8 d1e1 e6e5 d4f5 e7f5 d2e3 a7a5 a2a4 f5e3 e2e3 c6c5 c2c3 d6e7');
INSERT INTO evaluations VALUES(1590,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(1590,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(1590,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(1590,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(1590,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(1590,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(1590,36,'cp 26','g1f3 e6e2 f1e2 c6b4 e1d1 c8f5 d2d3 b4c6 b2b4 a7a6 a1b1 e7e6 b4b5 a6b5 c3b5 e8d7 c1f4 f7f6 f4c7 a8a2 c7b6 f8b4 c2c3 b4a5 b6a5 c6a5 c3c4 g8e7 f3d4 f5g6 b5c3 a2a3 d1d2 e7c6 d4b5 a3b3 d2e3 h8d8 e2f3 d8a8 h1d1 d7e7 b1b3');
INSERT INTO evaluations VALUES(1586,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(1586,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(1586,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1586,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(1586,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(1586,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(1586,36,'cp -3','c6b4 e2e6 c8e6 f1b5 c7c6 b5a4 g8f6 a2a3 b4d5 a4b3 a7a5 d2d3 e6f5 a3a4 h7h5 e1g1 d5b4 c3a2 e7e6 c1d2 b4a2 b3a2 h5h4 h2h3 f6d5 a2d5 c6d5 a1b1 f8e7 b2b4 a5b4 d2b4 a8a4 b4c3 e7f6 c3f6 g7f6 b1b7 e8g8');
INSERT INTO evaluations VALUES(56302,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(56302,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(56302,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(56302,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(56302,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(56302,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(56302,36,'cp 82','f8d6 d2d4 d5e4 f1c4 d6e5 d1h5 d8e7 h5e5 e7e5 d4e5 b8c6 b1c3 c6e5 c4b3 c8d7 e1g1 e8c8 c3e4 f7f6 e4c5 d7f5 c5e6 f5e6 b3e6 c8b8 c1e3 g8e7 a1d1 h8e8 f2f4 d8d1 f1d1 e7c6 f4e5 e8e6 d1d7 b8c8 d7g7 c6e5 h2h3 h7h5 g7h7 e6c6');
INSERT INTO evaluations VALUES(38637,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(38637,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(38637,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(38637,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(38637,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(38637,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(38637,36,'cp 92','d2d4 d5e4 f1c4 d6e5 d1h5 d8e7 h5e5 e7e5 d4e5 b8c6 b1c3 c6e5 c4b3 c8d7 e1g1 e8c8 c3e4 f7f6 e4c5 d7f5 c5e6 f5e6 b3e6 c8b8 c1e3 g8e7 a1d1 e7c6 f2f4 d8e8 e6b3 e5g4 e3c1 e8e7 f1e1 e7e1 d1e1 c6d4 h2h3 g4h6 g1f2 c7c5 c1e3 d4b3 a2b3');
INSERT INTO evaluations VALUES(38636,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(38636,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(38636,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(38636,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(38636,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(38636,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(38636,36,'cp 87','d5e4 f1c4 d6e5 d1h5 d8e7 h5e5 e7e5 d4e5 b8c6 b1c3 c6e5 c4b3 c8d7 e1g1 e8c8 c3e4 f7f6 e4c5 d7f5 c5e6 f5e6 b3e6 c8b8 c1e3 g8e7 a1d1 b7b6 f2f4 e5c6 e3f2 d8d1 f1d1 h8d8 d1d8 c6d8 e6b3 c7c5 g2g4 b8c7 c2c3 f6f5 g4f5 e7f5 g1g2 c7d7 g2f3 f5d6 b3g8 h7h6 g8d5 d8e6 f2e3');
INSERT INTO evaluations VALUES(38642,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(38642,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(38642,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(38642,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(38642,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(38642,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(38642,36,'cp 82','f1c4 d6e5 d1h5 d8e7 h5e5 e7e5 d4e5 b8c6 b1c3 c6e5 c4b3 c8d7 c1f4 f7f6 c3e4 e8c8 e4c5 d7f5 e1g1 g8e7 c5e6 f5e6 b3e6 c8b8 a1d1 h8e8 f4e5 f6e5 e6d7 e8f8 f1e1 e7g6 d7g4 c7c6 g2g3 b8c7 d1d8 f8d8 h2h4 c7d6 g4f5 g6f8 a2a3 g7g6 f5h3 b7b6 g1g2');
INSERT INTO evaluations VALUES(40062,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(40062,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40062,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40062,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(40062,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40062,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(40062,36,'cp 21','h7h6 g5h4 e8g8 f1e2 d5c4 e2c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 d1c1 c8d7 f3e5 f8c8 c4e2 b4c3 c1c3 c8c3 b2c3 d7e8 f1b1 b8c6 e5c4 b7b6 c4d6 c6e7 e3e4 e7g6 h4g3 e6e5 f2f3 f6d7 d6c4 f7f6 c4d6 d7c5 b1d1 a8d8 d6e8');
INSERT INTO evaluations VALUES(40034,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(40034,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(40034,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40034,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40034,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(40034,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(40034,36,'cp 22','g5h4 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 c5b4 h4f6 g7f6 c3e4 b4e7 e4d6 f8d8 d6c8 d8c8 c4d3 b8c6 g2g4 c8c7 d1c1 a8c8 c1c2 c6b4 c2c7 c8c7 a2a3 b4d3 e2d3 e7d6 f3d4 a7a6 h2h3 d6e5 b2b4 g8g7 f2f4 e5d4 d3d4 c7c2 a3a4');
INSERT INTO evaluations VALUES(40037,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(40037,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(40037,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40037,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40037,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(40037,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(40037,36,'cp 18','e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 c5b4 h4f6 g7f6 c3e4 b4e7 e4d6 b8c6 g2g4 f8d8 d6c8 a8c8 c4d3 c8c7 d1c1 d8c8 c1c3 c6e5 c3c7 c8c7 f3e5 f6e5 d3e4 a7a5 f2f3 e7b4 h1b1 b7b6 e4d3 g8g7 b1d1 c7d7');
INSERT INTO evaluations VALUES(35722,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(35722,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(35722,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(35722,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(35722,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(35722,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(35722,36,'cp 24','f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 c5b4 h4f6 g7f6 c3e4 b4e7 e4d6 b8c6 g2g4 f8d8 d6c8 a8c8 d1d8 e7d8 c4d3 g8f8 f3d2 c8c7 f2f4 c6b4 d3e4 b4a6 h1d1 f8e7 a2a3 a6b8 e4d3 a7a6 d2e4 b8d7 b2b4 d7b6');
INSERT INTO evaluations VALUES(35724,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(35724,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(35724,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(35724,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(35724,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35724,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(35724,36,'cp 20','d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 b8c6 c4b5 c5e7 a2a3 g7g5 h4g3 c8d7 f3e5 c6e5 g3e5 f8d8 b5d7 d8d7 d1c1 g8h7 h2h3 a7a6 h1d1 a8d8 b2b4 h7g6 e5f6 e7f6 d1d7 d8d7 c3a4 b7b5 a4c5 d7c7 e2d2 f6b2 c1c2 b2a3 c5a6 c7a7 c2c6 a3b2');
INSERT INTO evaluations VALUES(35726,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(35726,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(35726,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(35726,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(35726,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(35726,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35726,36,'cp 19','d3c4 c6c5 d4c5 d8d1 a1d1 b8d7 c5c6 b7c6 e1g1 f8d8 e3e4 g8f8 f3d4 d7e5 c4e2 c6c5 d4b5 c8a6 h4g3 e5c6 d1d8 a8d8 b5a7 a6e2 a7c6 e2f1 c6d8 f1d3 e4e5 e7d8 e5f6 d8f6 g3d6 f6e7 d6e7 f8e7 f2f3 e7d6 g1f2 d3c4 f2e3 d6c6 g2g3 e6e5 b2b3 c4e6 c3e4 f7f6 e4c3');
INSERT INTO evaluations VALUES(35725,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(35725,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(35725,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(35725,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(35725,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35725,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(35725,36,'cp 21','c6c5 d4c5 d8d1 a1d1 b8d7 c5c6 b7c6 e1g1 f8d8 c4e2 c8b7 d1c1 a7a5 f1d1 a5a4 a2a3 d7c5 f3e5 d8d1 e2d1 c5b3 c1c2 c6c5 d1e2 g7g5 h4g3 f6e4 e2f3 f7f5 f3e4 b7e4 c3e4 f5e4 e5c6 g8f7 g3e5 a8c8 c6e7 f7e7');
INSERT INTO evaluations VALUES(44964,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(44964,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(44964,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(44964,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(44964,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(44964,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(44964,36,'cp 119','e4e5 f6d7 f1d3 c6c5 e5e6 d7f6 e6f7 e8f7 d4c5 b8c6 g1f3 e7e5 d3b5 c8g4 c1g5 h7h6 g5f6 g7f6 b5c6 b7c6 f3e5 f6e5 d1g4 h6h5 g4h3 f8c5 e1c1 d8f6 h1f1 f6e6 h3f3 e6f6 f3d3 f6g6 d3a6 h8b8 c1b1 e5e4 c3a4 c5f8 f2f3 e4e3 f1e1 b8e8 e1e2 f8h6 d1e1 f7g8 a4c5');
INSERT INTO evaluations VALUES(44992,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(44992,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(44992,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(44992,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(44992,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(44992,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(44992,36,'cp 114','f6d7 f1d3 c6c5 e5e6 d7f6 e6f7 e8f7 g1f3 b8c6 d4c5 e7e5 d3b5 c8g4 c1g5 h7h6 g5f6 g7f6 b5c6 b7c6 f3e5 f6e5 d1g4 h6h5 g4e2 d8f6 e1g1 a8e8 a1d1 a7a5 a2a3 e5e4 e2e3 h5h4 c3e2 h4h3 g2g3 f8h6 e3c3 f6c3 b2c3 h6f8 c3c4 d5c4 d1d7 e8e7 f1d1 e7d7 d1d7 f7e6 d7c7 f8c5 c7c6 e6d5');
INSERT INTO evaluations VALUES(44021,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(44021,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(44021,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(44021,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(44021,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(44021,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(44021,36,'cp 122','f1d3 c6c5 e5e6 d7f6 e6f7 e8f7 g1f3 b8c6 d4c5 e7e5 d3b5 c8g4 c1g5 h7h6 g5f6 g7f6 b5c6 b7c6 f3e5 f6e5 d1g4 h6h5 g4e2 d8f6 e1g1 a8e8 a1d1 a7a5 a2a3 e5e4 c3a4 f7g8 b2b4 a5b4 a3b4 h8h7 a4b6 h5h4 e2g4 h7g7 g4h3 e8d8 c2c3 g7f7 g1h1 d8e8 b6d7 f6e6 h3e6');
INSERT INTO evaluations VALUES(52403,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52403,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(52403,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(52403,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(52403,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(52403,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52403,36,'cp 33','b1c3 d7d5 e2e4 c8g4 d2d4 g8f6 h2h3 g4f3 d1f3 e7e6 e4d5 c6d5 f1d3 d8b6 c3e2 f8d6 c1g5 b8d7 e1c1 e8g8 c1b1 b6c6 h3h4 b7b5 g2g4 a7a5 g5c1 f6e4 f3e3 d6e7 f2f3 e4d6 g4g5 f8c8 e3f4 b5b4 h4h5 a5a4 g5g6 f7f5 g6h7 g8h8');
INSERT INTO evaluations VALUES(52397,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(52397,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(52397,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(52397,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(52397,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(52397,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(52397,36,'cp 35','d7d5 e2e4 d5e4 c3e4 g8f6 d1e2 f6e4 e2e4 b8d7 d2d4 d7f6 e4f4 d8d6 f1d3 d6f4 c1f4 c8g4 f3e5 g4h5 c2c3 f6d5 f4g3 e7e6 e1g1 a7a5 a2a4 f8e7 e5c4 e8g8 g3d6 e7d6 c4d6 b7b6 d6c4 h5g6 d3g6 h7g6 g2g3 a8d8 f1e1 f8e8');
INSERT INTO evaluations VALUES(52834,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(52834,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(52834,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52834,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52834,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(52834,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(52834,36,'cp 25','e2e4 d5e4 c3e4 g8f6 d1e2 f6e4 e2e4 b8d7 d2d4 d7f6 e4f4 d8d6 f1d3 d6f4 c1f4 c8g4 f3e5 g4h5 c2c3 f6d5 f4g3 f7f6 e5c4 e7e6 c4a5 e8c8 e1g1 h5g4 f1e1 g4f5 d3e4 g7g5 h2h4 g5h4 e4f5 e6f5 g3h4 h8g8 g2g3 f5f4');
INSERT INTO evaluations VALUES(51497,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(51497,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(51497,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(51497,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(51497,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(51497,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(51497,36,'cp 81','c8b7 d2d4 e7e6 f1d3 f8b4 g1h3 d7d5 e4e5 c7c5 e1g1 c5c4 d3e2 d8d7 h3f4 b8c6 e2f3 b4c3 b2c3 g7g6 a2a4 c6a5 c1a3 g8e7 a3b4 e7c6 b4d6 c6e7 d1d2 e8c8 f1b1 b7a6 f3e2 e7f5 d6b4');
INSERT INTO evaluations VALUES(30817,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(30817,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(30817,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(30817,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(30817,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(30817,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(30817,36,'cp 96','d2d4 e7e6 f1d3 g8f6 h2h4 d7d6 d1e2 a7a6 g1f3 b8d7 e4e5 d6e5 d4e5 f6d5 c3d5 b7d5 d3e4 c7c6 c2c4 d5e4 e2e4 a8c8 e1g1 d7c5 e4g4 d8d3 h4h5 d3f5 g4f5 e6f5 c1e3 f5f4 e3f4 c5d3 g2g3 d3b2 f1c1 f8c5 h5h6 g7h6 f4h6');
INSERT INTO evaluations VALUES(2035,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(2035,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(2035,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(2035,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(2035,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(2035,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(2035,36,'cp 200','d7d6 e5d6 e7f6 d1c1 f8d6 f4d6 f6d6 c1d2 d6e7 b1c3 g8f6 e1c1 e8g8 d2f4 a7a6 g2g3 b7b5 f1g2 b5b4 c3d5 f6d5 d1d5 c8e6 d5d3 a8d8 f3e5 c6e5 f4e5 c7c5 g2d5 d8e8 d5e6 f7e6 f2f4 c5c4 d3d4 c4c3 h1d1 f8f5 e5d6 e7d6 d4d6 f5h5 h2h4 h5a5 c1b1');
INSERT INTO evaluations VALUES(11993,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11993,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11993,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11993,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(11993,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(11993,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11993,36,'cp 0','d1f3 f8c5 f3g3 e8g8 c1h6 f6e8 f1c4 d7d6 h6g5 e8f6 e1c1 c5d4 c3e2 d4c5');
INSERT INTO evaluations VALUES(47362,30,'cp -192',NULL);
INSERT INTO evaluations VALUES(47362,31,'cp -196',NULL);
INSERT INTO evaluations VALUES(47362,32,'cp -207',NULL);
INSERT INTO evaluations VALUES(47362,33,'cp -199',NULL);
INSERT INTO evaluations VALUES(47362,34,'cp -199',NULL);
INSERT INTO evaluations VALUES(47362,35,'cp -198',NULL);
INSERT INTO evaluations VALUES(47362,36,'cp -208','d8d4 c1e3 d4h4 g2g3 h4g4 f3f2 g4f5 g1f3 f6g4 f2e2 f5e6 e3f4 e6e2 f1e2 c7c6 h2h3 g4f6 e1c1 b8d7 g3g4 d7b6 h1e1 h7h5 f3e5 c8e6 g4g5 f6d7 e5d7 e6d7 a2a4 d7h3 a4a5 b6d7 a5a6 b7a6 g5g6 f7g6 e2a6 d7c5 a6e2 a8d8 e2f3 d8d1 e1d1 h3e6 f3c6 e8f7');
INSERT INTO evaluations VALUES(33711,30,'cp -193',NULL);
INSERT INTO evaluations VALUES(33711,31,'cp -178',NULL);
INSERT INTO evaluations VALUES(33711,32,'cp -213',NULL);
INSERT INTO evaluations VALUES(33711,33,'cp -210',NULL);
INSERT INTO evaluations VALUES(33711,34,'cp -201',NULL);
INSERT INTO evaluations VALUES(33711,35,'cp -190',NULL);
INSERT INTO evaluations VALUES(33711,36,'cp -190','c1e3 d4h4 g2g3 h4g4 f3f2 g4f5 g1f3 f6g4 f2e2 g4e3 e2e3 e7e6 e1c1 f5c5 f1b5 c7c6 e3d3 f8e7 b5c4 c5a5 c1b1 b8d7 d3e2 e8g8 h2h4 a5c7 g3g4 b7b5 c4d3 d7c5 g4g5 c5d3 c2d3 e6e5 f3e5 c8f5 e5c6 c7c6 e2e7 a8e8');
INSERT INTO evaluations VALUES(33710,30,'cp -179',NULL);
INSERT INTO evaluations VALUES(33710,31,'cp -197',NULL);
INSERT INTO evaluations VALUES(33710,32,'cp -192',NULL);
INSERT INTO evaluations VALUES(33710,33,'cp -188',NULL);
INSERT INTO evaluations VALUES(33710,34,'cp -209',NULL);
INSERT INTO evaluations VALUES(33710,35,'cp -201',NULL);
INSERT INTO evaluations VALUES(33710,36,'cp -210','d4h4 g2g3 h4g4 f3f2 g4f5 f2d2 e7e6 e1c1 f8e7 g1e2 b8c6 e2d4 c6d4 e3d4 e8g8 d2e1 f8d8 c1b1 c8d7 g3g4 f5a5 g4g5 f6e8 h2h4 d7c6 h1h2 e6e5 e1e5 a5e5 d4e5 d8d1 c3d1 e7d6 e5d6 e8d6 f1d3 a7a5 d1e3');
INSERT INTO evaluations VALUES(51480,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(51480,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(51480,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(51480,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(51480,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51480,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(51480,36,'cp 12','e7e6 b1c3 c7c5 g1f3 a7a6 d2d4 c5d4 f3d4 c8b7 d1e2 d8c7 c4d3 f8d6 c1e3 g8f6 e1c1 d6e5 h1f1 b8c6 d4c6 b7c6 e3d2 e5d4 f2f4 d7d6 g2g4 b6b5 g4g5 f6d7 f4f5 e8g8 f5f6 b5b4 f6g7');
INSERT INTO evaluations VALUES(51395,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(51395,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(51395,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(51395,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(51395,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(51395,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(51395,36,'cp 19','b1c3 c8b7 g1f3 g8f6 e4e5 d7d5 c4f1 d5d4 e5f6 d4c3 f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 b8d7 f1e2 e8c8 h1e1 h8f8 d2f4 d7f6 d1d8 f8d8 g2g3 d8f8 a2a3 h7h6 f4e5 f8g8 f3d2 f6d7 e5g7 g8g7 f2f4 c8d8 e2f3 b7f3 d2f3 c7c5 f3e5 d7e5 e1e5 d8d7 c2c4 f7f5 c1c2 g7g8 b2b4 h6h5 b4c5');
INSERT INTO evaluations VALUES(51392,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51392,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(51392,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(51392,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(51392,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51392,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(51392,36,'cp 12','c8b7 g1f3 g8f6 e4e5 d7d5 c4f1 d5d4 e5f6 d4c3 f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 b8d7 f1b5 e8c8 f3g5 a7a6 b5e2 d8f8 h1e1 h7h6 g5h3 g7d4 g2g3 e6e5 f2f4 f7f6 e2d3 b7f3 d3a6 c8b8 a6e2 f3g2 c2c3 d4c5 b2b4 g2h3 b4c5 d7c5 f4e5');
INSERT INTO evaluations VALUES(30621,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30621,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(30621,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(30621,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(30621,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(30621,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(30621,36,'cp 27','g1f3 g8f6 e4e5 d7d5 c4f1 d5d4 e5f6 d4c3 f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 b8c6 d2f4 a8c8 a2a3 e8g8 h2h4 c6e7 h4h5 e7d5 f4e5 c8d8 f1c4 h7h6 c4d5 d8d5 d1d5 b7d5 e5c7 f8c8 c7f4 a7a5 h1d1 b6b5 f4d6 d5b3 d1d2 b3d5 f3e5');
INSERT INTO evaluations VALUES(30620,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(30620,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(30620,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(30620,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(30620,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(30620,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(30620,36,'cp 27','g8f6 e4e5 d7d5 c4f1 d5d4 e5f6 d4c3 f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 b8d7 f1b5 e8c8 f3g5 a7a6 b5e2 d8f8 f2f4 d7c5 h1e1 c5a4 c2c3 h7h6 g5f3 b6b5 d2e3 b7d5 a2a3 a4b6 g2g3 f8d8 f3e5 g7e5 f4e5 b6d7 e3d4 h6h5 e1f1 c7c5 d4e3 h5h4 g3g4 h4h3 f1f7');
INSERT INTO evaluations VALUES(28699,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(28699,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(28699,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(28699,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(28699,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(28699,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(28699,36,'cp 16','e4e5 d7d5 c4f1 d5d4 e5f6 d4c3 f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 b8d7 f1b5 e8c8 f3g5 a7a6 b5e2 d8f8 f2f4 d7c5 d2e3 b7g2 h1g1 g2c6 e3c5 b6c5 e2a6 c8b8 g1g3 g7d4 g3b3 b8a7 a6c4 h7h6 g5f3 d4f6 f3e5 f6e5 f4e5 h8g8 b3h3 g8g5 h3h6 g5e5 b2b3 c6d5');
INSERT INTO evaluations VALUES(16390,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(16390,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(16390,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(16390,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(16390,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(16390,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(16390,36,'cp -119','f6e4 e1g1 d7d5 c4b5 f8d6 d2d3 e4f6 f1e1 e5e4 c1g5 e8g8 d3e4 d5e4 f3d2 f8e8 b1a3 d6a3 b2a3 c8f5 d2b3 d8e7 b5c6 b7c6 d1d4 f5g6 a1d1 e8d8 d4c4 h7h6 d1d8 a8d8 g5h4 e7d6 h2h3 d8d7 b3d4 f6h5 c4c6 f7f5 c6d6 d7d6 d4b5 d6d2 b5c7 d2a2');
INSERT INTO evaluations VALUES(16829,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(16829,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(16829,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(16829,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(16829,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(16829,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(16829,36,'cp -122','e1g1 d7d5 c4b5 f8d6 d2d3 e4f6 f1e1 e8g8 b5c6 b7c6 f3e5 d6e5 e1e5 f6g4 e5g5 f7f5 h2h3 h7h6 g5h5 g4e5 c1f4 e5g6 f4e3 d5d4 c3d4 f5f4 e3d2 a8b8 d1c2 f4f3 c2c6 b8b6 c6c4 g8h7 b1c3 f3g2 c3e4 g6h4 e4g5 h7g6 h5h4 h6g5 h4e4 c8h3 a1e1 b6b2 d2e3 d8f6 c4c7 f8f7');
INSERT INTO evaluations VALUES(16828,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(16828,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(16828,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(16828,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(16828,34,'cp -113',NULL);
INSERT INTO evaluations VALUES(16828,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(16828,36,'cp -108','d7d5 c4b5 f8d6 d2d3 e4f6 f1e1 e8g8 b5c6 b7c6 f3e5 d6e5 e1e5 f6g4 e5e2 d8d6 g2g3 d6g6 b1d2 g6d3 e2e1 c8d7 d2f3 d3d1 e1d1 f8e8 c1d2 g4f6 d2f4 d7g4 g1g2 c6c5 f4c7 e8e2 h2h3 g4f3 g2f3 e2b2 c7e5 f6e4 d1d5 e4f2 d5c5 f2d3 c5d5 d3e5 d5e5 b2c2 e5e2 c2c3 f3g2');
INSERT INTO evaluations VALUES(38384,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(38384,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(38384,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(38384,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(38384,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(38384,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(38384,36,'cp 60','b1c3 g8f6 f1e2 e8g8 c1e3 b8a6 g1f3 e7e5 e1g1 d8e7 d1c2 c7c6 a1d1 e5d4 f3d4 f8e8 f2f3 d6d5 c4d5 c6d5 e2a6 b7a6 d4c6 e7e6 c3d5 f6d5 d1d5 c8b7 c6d4 e6b6 d5c5 a8c8 d4b3 c8c5 c2c5 e8c8 c5b6 a7b6 e3b6');
INSERT INTO evaluations VALUES(38380,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(38380,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(38380,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(38380,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(38380,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(38380,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(38380,36,'cp 62','g8f6 h2h3 b8d7 g1f3 e7e5 f1d3 e5d4 f3d4 e8g8 e1g1 f8e8 f1e1 d7c5 d3c2 a7a6 a2a4 c8d7 c1g5 h7h6 g5h4 c5e6 d4e6 d7e6 d1e2 g6g5 h4g3 g5g4 h3g4 f6g4 c3d5 e6d5 c4d5 d8g5 e2f3');
INSERT INTO evaluations VALUES(55722,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(55722,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(55722,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(55722,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(55722,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(55722,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(55722,36,'cp 68','g1f3 d6d5 g2g3 g8f6 f1g2 f8b4 b1c3 e8g8 e1g1 d5c4 f3e5 f6d5 d1c2 c7c5 f1d1 c5d4 d1d4 b8c6 e5c6 b7c6 d4c4 b4e7 a2a3 c8b7 c1f4 d5f4 c4f4 b7a6 b2b4 d8c7 a1c1 a8d8 c3e4 d8d7 h2h4 f8d8 e2e3 c7e5 e4c5 e7c5 b4c5');
INSERT INTO evaluations VALUES(55721,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(55721,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(55721,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(55721,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(55721,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(55721,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(55721,36,'cp 63','d6d5 g2g3 g8f6 f1g2 f8e7 e1g1 e8g8 b1c3 b8d7 c1f4 c7c6 b2b3 b7b6 d1d3 c6c5 e2e4 d5e4 c3e4 c8b7 e4d6 e7d6 f4d6 f8e8 a1d1 d8c8 d4c5 d7c5 d3e3 h7h6 d6c5 b6c5 f3e5 b7g2 g1g2 c8c7 d1d2 a7a5 f1d1 a5a4 e5d7 c7c6 e3f3 c6f3 g2f3 a4b3');
INSERT INTO evaluations VALUES(48784,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(48784,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(48784,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(48784,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(48784,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(48784,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(48784,36,'cp 12','e4d5 e5e4 f3d4 c7c6 b1c3 f8b4 a2a3 b4a5 d4b3 c6d5 c4b5 b8c6 b3a5 d8a5 d2d4 c8e6 c1g5 f6d7 g5e3 e8g8 d1d2 a8c8 e1g1 a5b6 f2f3 e4f3 f1f3 a7a6 b5d3 b6b2 a1b1 b2a3 f3g3 d7f6');
INSERT INTO evaluations VALUES(48590,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(48590,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(48590,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48590,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(48590,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48590,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48590,36,'cp 0','e5e4 f3e5 d8e7 d2d4 e4d3 f2f4 b8d7 e1g1 d7e5 f4e5 e7c5 g1h1 f6d5 c4d5 c5d5 b1c3 d5c6 c2d3 c8e6 d3d4 e8c8 c1g5 d8d7 a1c1 e6c4 d4d5 c6g6 g5f4 c4f1 d1f1 f8b4 c3b5 d7d5 c1c7 c8d8 h2h4 b4e7 c7b7 g6g4 e5e6 f7e6 b7b8 d8d7 b8b7');
INSERT INTO evaluations VALUES(48560,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48560,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48560,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48560,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48560,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48560,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48560,36,'cp 9','f3d4 c7c6 b1c3 c6d5 c3d5 f8c5 d4b5 e8g8 d5f6 d8f6 d2d4 e4d3 e1g1 b8c6 c4d3 c6e5 d1h5 e5d3 c2d3 c5b6 b5c3 f6f5 h5f5 c8f5 c3d5 f5d3 d5b6 a7b6 f1d1 d3c2 d1d4 f8d8 c1e3 b6b5 f2f3 d8d4 e3d4 a8d8 d4c3 d8d1 a1d1 c2d1 g1f2 f7f6 b2b4 f6f5 c3b2 g8f7');
INSERT INTO evaluations VALUES(38664,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(38664,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38664,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38664,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(38664,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(38664,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38664,36,'cp 30','f4d6 c7d6 c2c4 d5c4 f1c4 g8f6 b1c3 e8g8 g1e2 d6d5 c4d3 b8c6 e1g1 c6b4 d3b1 b4c6 a2a3 c8d7 b1d3 d8e7 d1e1 a8c8 f2f3 e6e5 a1d1 e5d4 e3d4 c8e8 d3b1 c6a5 e1f2 e7d6 f2g3 d6b6 e2f4 b6b2 f4d5');
INSERT INTO evaluations VALUES(38603,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(38603,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38603,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(38603,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(38603,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38603,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38603,36,'cp 21','c7d6 c2c4 d5c4 f1c4 g8f6 b1c3 e8g8 g1f3 a7a6 e1g1 b7b5 c4d3 c8b7 f3d2 e6e5 d2e4 e5d4 e3d4 f6e4 d3e4 b7e4 c3e4 b8c6 d1d2 a8c8 a1c1 d8b6 d4d5 c6e5 b2b3 a6a5 e4g3 c8c1 f1c1 f8e8 g3f5 a5a4 f5e3 a4b3 a2b3 h7h6 h2h3');
INSERT INTO evaluations VALUES(55106,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(55106,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55106,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55106,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55106,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55106,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55106,36,'cp 0','c8f5 c1b2 e7e6 g2g3 g8f6 f1g2 f8e7 d2d3 e8g8 e1g1 a7a5 b1d2 h7h6 f3e5 f5h7 a2a4 b8d7 e5d7 f6d7 e2e4 e7f6 b2f6 d7f6 d1e2 c7c6 e4e5 f6e8 e2e3 e8c7 f2f4 c7a6 c2c3 b7b6 a1d1 a6c5 d3d4 c5d3 h2h3 f8e8');
INSERT INTO evaluations VALUES(31877,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(31877,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31877,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(31877,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(31877,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31877,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31877,36,'cp 0','c1b2 e7e6 d2d3 c7c5 b1d2 b8c6 g2g3 f7f6 f1g2 e6e5 e2e4 f5g4 e4d5 d8d5 h2h3 g4e6 f3h4 d5d7 d1h5 e6f7 h5g4 f7e6');
INSERT INTO evaluations VALUES(31878,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(31878,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31878,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(31878,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(31878,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31878,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31878,36,'cp 0','e7e6 g2g3 g8f6 f1g2 f8e7 e1g1 e8g8 c2c4 b8d7 b1c3 c7c6 f3h4 f5e4 g2h3 h7h6 c3e4 d5e4 d2d3 g7g5 d1d2 e4d3 e2d3 g8h7 b2c1 h7g8');
INSERT INTO evaluations VALUES(32331,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(32331,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(32331,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(32331,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(32331,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(32331,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32331,36,'cp 0','c2c4 c7c6 e2e3 h7h6 d2d4 f8b4 b1d2 g8f6 a2a3 b4d6 f1e2 b8d7 e1g1 f6e4 b3b4 e8g8 d2e4 f5e4 c4c5 d6c7 a3a4 d8e7 f3d2 e4h7 d2f3 h7e4');
INSERT INTO evaluations VALUES(21120,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(21120,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(21120,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(21120,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(21120,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(21120,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(21120,36,'cp 81','f1c4 e7e5 g1f3 e5d4 e3d4 g8f6 e1g1 f8e7 b1c3 e8g8 h2h3 h7h6 c1f4 e7d6 f4e5 c8f5 f1e1 f6d7 d1c1 d7b6 c4b3 c6a5 c1f4 a5b3 a2b3 f5h7 e5d6 c7d6 e1e3 a7a6 a1e1 a8c8 f3h4 c8c7 h4f5 h7f5 f4f5 d6d5 f5e5 d8d7 e5g3 g8h8');
INSERT INTO evaluations VALUES(21104,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(21104,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(21104,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(21104,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(21104,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(21104,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(21104,36,'cp 62','e7e5 g1f3 f8b4 b1c3 e5d4 e3d4 g8f6 e1g1 e8g8 h2h3 h7h6 c1f4 b4d6 f4e5 c8f5 f1e1 f6d7 d1c1 d7b6 c4b3 c6a5 b3c2 f5c2 c1c2 f8e8 a1d1 a5c6 e1e4 d6e5 d4e5 d8e7 d1e1 a8d8 a2a3 e7e6 b2b4 a7a6 b4b5 a6b5 c3b5 b6d5 e4g4 g8h8 b5d4');
INSERT INTO evaluations VALUES(21788,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(21788,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(21788,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(21788,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(21788,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21788,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(21788,36,'cp 81','g1f3 f8b4 b1c3 e5d4 e3d4 g8f6 e1g1 e8g8 h2h3 h7h6 c1f4 b4d6 f4e5 c8f5 f1e1 f6d7 d1c1 d7b6 c4b3 c6a5 b3c2 f5c2 c1c2 f8e8 e1e4 d6f8 e4g4 b6d5 c2f5 d5c3 b2c3 a5c6 e5f4 c6e7 f5c2 e7g6 c2b3 e8e6 b3b7 g6f4 g4f4 f8d6 f4g4 e6e2 b7b3 d8e7 c3c4 c7c6 a1b1');
INSERT INTO evaluations VALUES(21789,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(21789,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(21789,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(21789,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(21789,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(21789,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(21789,36,'cp 48','f8b4 b1c3 e5d4 e3d4 g8f6 e1g1 e8g8 h2h3 h7h6 f1e1 b4d6 d1c2 c6a5 c4d3 c8e6 c1d2 a5c6 a2a3 f8e8 a1d1 c6e7 f3e5 c7c6 c3a4 b7b6 d2f4 f6d5 f4g3 d6e5 d4e5 d8c7 a4c3 d5c3 c2c3 e8d8 d3e4 d8d1 e1d1');
INSERT INTO evaluations VALUES(43164,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43164,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(43164,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43164,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(43164,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(43164,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(43164,36,'cp 9','c7c5 g1f3 d7d5 d2d4 e8g8 d4c5 d8a5 e1g1 d5c4 d1d4 b8c6 d4c4 c8e6 c4h4 a5c5 f1d1 f8d8 c1e3 c5c4 h4c4 e6c4 f3d4 f6d5 c3d5 c4d5 g2d5 d8d5 d4c6 b7c6 a1b1 d5d1 b1d1 g7b2 a2a4 a7a5 d1d7 b2a3 d7c7 f7f5 e3f4 a8d8 c7c6 g8f7 g1g2');
INSERT INTO evaluations VALUES(40003,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40003,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40003,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40003,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40003,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(40003,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(40003,36,'cp 29','e2e3 e8g8 g1e2 e7e6 b2b3 d7d5 e1g1 b8c6 d2d4 c5d4 e3d4 c8d7 a1b1 b7b6 c1a3 c6e7 c4d5 e6d5 e2f4 a8c8 f1e1 f8e8 b1c1 e7f5 e1e8 d8e8 a3b2 d7e6 c3d5 c8c1 d5f6 g7f6 b2c1 f5d4 g1h1 e6f5 c1b2 f6g7 h2h4 d4c6 b2g7 g8g7 d1a1 g7g8');
INSERT INTO evaluations VALUES(39999,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(39999,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(39999,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(39999,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(39999,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(39999,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(39999,36,'cp 15','e7e6 g1e2 d7d5 b2b3 e8g8 e1g1 b8c6 d2d4 c5d4 e3d4 c8d7 c1a3 c6e7 a3d6 d7c6 c4c5 a7a5 a2a3 f6e8 d6f4 b7b6 b3b4 a5b4 a3b4 a8a1 d1a1 b6c5 b4c5 d8d7 a1a6 e7f5 a6d3 e8f6 f4e5 f8a8 f1b1 f6g4 e5f4 e6e5 f4e5');
INSERT INTO evaluations VALUES(39361,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(39361,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39361,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(39361,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(39361,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(39361,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(39361,36,'cp 22','c3d5 b4e7 d2d4 e5d4 d1d4 g8f6 d5e7 d8e7 c1g5 b8c6 d4c3 d7d6 g1f3 c8e6 e2e3 h8g8 g5f6 e7f6 c3f6 g7f6 f3d4 c6e5 b2b3 a7a5 d4b5 e8d7 e1c1 a5a4 c1b2 a4b3 a2b3 a8a5 b5c3 f6f5 g2g3 g8a8 f1g2');
INSERT INTO evaluations VALUES(39314,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(39314,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39314,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(39314,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(39314,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(39314,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(39314,36,'cp 10','b4e7 d2d4 e5d4 d5e7 d8e7 d1d4 g8f6 c1g5 b8c6 d4c3 d7d6 g1f3 c6e5 g5h4 e8g8 e2e3 a7a5 e1c1 f8e8 f3e5 d6e5 f1d3 f6e4 d3e4 e7h4 f2f3 c7c6 c4c5 c8e6 a2a3 a5a4 e4c2 e5e4 c2e4 e6b3 d1d2 a8d8 d2d8 e8d8 e4c2 h4f2 h1d1');
INSERT INTO evaluations VALUES(56926,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(56926,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(56926,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(56926,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(56926,34,'cp -164',NULL);
INSERT INTO evaluations VALUES(56926,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(56926,36,'cp -145','g5g4 e1g1 g4f3 d1f3 d8f6 d2d3 b8c6 c1f4 d7d6 b1c3 f6g7 c3d5 e8d8 f4g3 c8e6 d5f4 e6c4 d3c4 g8f6 f4d5 f6d7 g3h4 f7f6 c4c5 d8c8 d5f6 d7e5 f3h3 c8b8 f6d5 g7d7 h3f5 a7a5 a1d1 a5a4 h4g3 d7e8 c2c3 f8h6 d5f6 e8f7 f6d5');
INSERT INTO evaluations VALUES(56937,30,'cp -116',NULL);
INSERT INTO evaluations VALUES(56937,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(56937,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(56937,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(56937,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(56937,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(56937,36,'cp -132','e1g1 g4f3 d1f3 d8f6 d2d3 b8c6 c1f4 d7d6 b1c3 g8e7 c3b5 e8d8 d3d4 e7g6 f3g3 g6f4 f1f4 f6g6 g3f2 f8e7 f4f7 c8d7 b5c3 d7e8 f7f3 g6g7 c3e2 d8c8 a1f1 c8b8 c2c3 a7a5 e2f4 a5a4 c4e6 e7g5 f4d5 a4a3 b2a3 e8g6 f2e2');
INSERT INTO evaluations VALUES(56936,30,'cp -146',NULL);
INSERT INTO evaluations VALUES(56936,31,'cp -151',NULL);
INSERT INTO evaluations VALUES(56936,32,'cp -158',NULL);
INSERT INTO evaluations VALUES(56936,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(56936,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(56936,35,'cp -169',NULL);
INSERT INTO evaluations VALUES(56936,36,'cp -148','g4f3 d1f3 d8f6 d2d3 b8c6 c1f4 d7d6 b1c3 f6g7 c3d5 e8d8 f4g3 c8e6 d5f4 e6c4 d3c4 g8f6 f4d5 f6d7 g3h4 f7f6 c4c5 f8e7 a1d1 c6e5 f3c3 h8g8 g2g3 d7c5 b2b4 c5e6 h4f6 e7f6 d5f6 e5d7 c3b3 d7f6 b3e6 g8f8 e4e5 f6d7 f1f8 d7f8 e6d5 f8d7 e5e6 d7f6 d5b7 a8c8');
INSERT INTO evaluations VALUES(56935,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(56935,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(56935,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(56935,33,'cp -151',NULL);
INSERT INTO evaluations VALUES(56935,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(56935,35,'cp -174',NULL);
INSERT INTO evaluations VALUES(56935,36,'cp -135','d1f3 d8f6 d2d3 b8c6 c1f4 d7d6 b1c3 g8e7 f3e3 f6g6 d3d4 f8g7 c3b5 e8g8 c2c3 c8d7 f1f3 c6a5 c4d3 d7b5 d3b5 d6d5 f4c7 a5c4 e3e2 g6e6 a1e1 a8c8 e4d5 e6e2 e1e2 c8c7 b5c4 e7c8 c4d3 f8d8 f3e3 c8d6 a2a4 a7a5 e3h3 h7h6 g2g4 d6e8');
INSERT INTO evaluations VALUES(56934,30,'cp -151',NULL);
INSERT INTO evaluations VALUES(56934,31,'cp -146',NULL);
INSERT INTO evaluations VALUES(56934,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(56934,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(56934,34,'cp -160',NULL);
INSERT INTO evaluations VALUES(56934,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(56934,36,'cp -163','d8f6 d2d3 b8c6 c1f4 d7d6 b1c3 g8e7 c3b5 e8d8 d3d4 e7g6 f3g3 g6f4 f1f4 f6g6 g3f2 f8e7 f4f7 c8d7 b5c3 d7e8 f7f3 g6g7 c3e2 h8f8 f3f8 g7f8 f2e3 f8g7 c2c3 d8c8 a1f1 a7a5 e2f4 c6d8 f4d5 e8c6 e3f4 c6d5 e4d5 c8b8 c4d3 a5a4 f4f5 e7g5 f5f8 g7f8 f1f8 b8a7 d3h7');
INSERT INTO evaluations VALUES(35001,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(35001,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(35001,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(35001,33,'cp -144',NULL);
INSERT INTO evaluations VALUES(35001,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(35001,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(35001,36,'cp -130','d2d3 b8c6 c1f4 d7d6 b1c3 f6g7 c3d5 e8d8 f4g3 c8e6 d5f4 d8d7 a1e1 e6c4 d3c4 a8d8 g3h4 f8e7 f4d5 d7c8 d5e7 g8e7 h4f6 g7f8 f3h3 c8b8 f6h8 f8h8 f1f7 h8b2 h3h7 e7c8 g1h1 b2b6 f7f5 c6e5 h2h3 b6c6 h7g7 c6c4 f5f8 d8f8 g7f8 c4c2');
INSERT INTO evaluations VALUES(35000,30,'cp -147',NULL);
INSERT INTO evaluations VALUES(35000,31,'cp -153',NULL);
INSERT INTO evaluations VALUES(35000,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(35000,33,'cp -178',NULL);
INSERT INTO evaluations VALUES(35000,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(35000,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(35000,36,'cp -152','b8c6 c1f4 d7d6 b1c3 f6g7 c3d5 e8d8 a1e1 c8e6 f4g3 c6e5 f3f2 f8e7 c4b3 h7h5 d3d4 h5h4 g3f4 e5g4 f2d2 h4h3 g2g3 e6d5 b3d5 g8f6 d5c4 f6h5 e4e5 h5f4 f1f4 e7g5 e5d6 f7f5 c4e6 g4h6 d6c7 d8c7 d2c3 c7b8 f4f2 g5f6 c3c5 g7c7 e6f5 c7c5 d4c5 h6f5 f2f5');
INSERT INTO evaluations VALUES(57201,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(57201,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(57201,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(57201,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(57201,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(57201,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(57201,36,'cp -40','d7d5 d2d4 b8c6 g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 c4d5 f6d5 b1c3 c8e6 f1c4 d8d7 c3d5 e6d5 c4d5 d7d5 d1b3 d5b3 d2b3 b7b6 c1d2 e8d7 e1e2 f8d6 d2c3 a8e8 b3d2 h7h5 a2a4 c6e7 h2h4 e7d5 g2g3 e8g8 d2c4 d7e6 c4d6 c7d6 c3d2 g8c8 h1c1 a7a6 d2e1');
INSERT INTO evaluations VALUES(56444,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(56444,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(56444,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(56444,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(56444,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(56444,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(56444,36,'cp -37','d2d4 b8c6 g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 c4d5 f6d5 f1c4 d8d7 c3d5 e6d5 c4d5 d7d5 d1b3 d5b3 d2b3 e8d7 c1d2 f8d6 e1e2 b7b6 d2c3 c6e7 f2f3 e4f3 g2f3 a8e8 e2d3 e7d5 c3d2 g7g5 a1f1 h8f8 e3e4 d5f4 d2f4 d6f4 h2h3 e8d8 a2a4 h7h5');
INSERT INTO evaluations VALUES(56364,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(56364,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(56364,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(56364,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(56364,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(56364,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(56364,36,'cp -30','b8c6 g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 a2a3 d8d7 f1e2 f8e7 b2b4 a7a6 e1g1 e8g8 c1b2 c6d8 d1b3 g8h8 a3a4 d8c6 c4d5 f6d5 e2c4 c6b4 b2a3 b7b5 a4b5 a6b5 c3b5 c7c6 b5c3 a8b8 c3e2 e7d6 a3b4 b8b4');
INSERT INTO evaluations VALUES(21681,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(21681,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(21681,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(21681,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(21681,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21681,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(21681,36,'cp -21','g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 a2a3 f8e7 b2b4 a7a6 d1b3 e8g8 c4d5 e6f7 f1c4 c6a7 e1g1 a7b5 c3b5 a6b5 c4b5 c7c6 b5c4 b7b5 d5d6 e7d6 c4f7 f8f7 b3e6 f6d5 c1b2 d5c7 e6b3 c7d5 f1c1 d8d7 a3a4 a8e8 a4b5 c6b5');
INSERT INTO evaluations VALUES(21682,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(21682,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(21682,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(21682,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(21682,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21682,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(21682,36,'cp -28','e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 a2a3 f8e7 c4d5 f6d5 f1c4 c6a5 c4d5 e6d5 c3d5 d8d5 d1c2 a5c6 c2c4 e8c8 b2b4 a7a6 c4d5 d8d5 e1e2 a6a5 d2b1 a5b4 a3b4 e7b4 a1a8 c6b8 c1a3 d5b5 h1c1 h8d8 a3b4 b5b4 b1c3 b4b2 e2e1');
INSERT INTO evaluations VALUES(21613,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(21613,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(21613,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(21613,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(21613,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(21613,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(21613,36,'cp -18','f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 a2a3 d8d7 f1e2 a7a6 e1g1 f8e7 d1c2 e8g8 b2b4 c6d8 c4d5 f6d5 d2c4 d8f7 c4e5 f7e5 d4e5 d5c3 c2c3 f8c8 f1d1 d7e8 c1d2 h7h5 d2e1 c7c6 e2c4 e8f7 c4e6 f7e6 a3a4');
INSERT INTO evaluations VALUES(55303,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(55303,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(55303,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(55303,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(55303,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(55303,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(55303,36,'cp 106','e2e4 c7c5 g1f3 a7a6 d2d4 c5d4 d1d4 d8d4 f3d4 e7e6 d4b3 c8d7 c1e3 d7c6 a1c1 b8d7 f1e2 a8c8 e1g1 g8f6 f1d1 h7h5 f2f3 f8e7 b3a5 e8g8 a2a3 d7e5 b2b4 h5h4 g2g3 h4g3 h2g3 f8d8 d1d8 e7d8 a5c6 c8c6');
INSERT INTO evaluations VALUES(55276,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(55276,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(55276,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(55276,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(55276,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(55276,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(55276,36,'cp 86','c7c5 g1f3 a7a6 d2d4 c5d4 d1d4 d8d4 f3d4 e7e6 f2f3 g8f6 d4b3 b8c6 c1e3 e6e5 f1c4 b7b5 c4d5 c8d7 a1c1 a8c8 e1e2 h7h5 h1d1 h5h4 a2a3 f8e7 b3c5 e7c5 e3c5 c6e7 c5b4 f6d5 c3d5 e7d5 e4d5 f7f6 c1c8 d7c8 d1c1 e8d7 c1c6 c8b7 c6e6 d7c8 d5d6 h8d8 e2f2 b7c6 b4a5');
INSERT INTO evaluations VALUES(14208,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14208,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(14208,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14208,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(14208,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14208,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(14208,36,'cp 63','g1f3 e7e6 c4d5 e6d5 c1g5 f8e7 e2e3 f6h5 g5e7 c6e7 f1d3 e8g8 d1c2 h5f6 h2h3 g7g6 f3e5 c8e6 e1g1 f6d7 e5f3 e7f5 a2a4 c7c6 b2b4 f8e8 f1c1 f5d6 c2b2 d8e7 a4a5 a7a6 c3a4 e6f5 d3e2 f5e6 a4c5 d7c5 b4c5');
INSERT INTO evaluations VALUES(14207,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(14207,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(14207,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(14207,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(14207,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(14207,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(14207,36,'cp 57','e7e6 a2a3 f6e4 d1c2 e4c3 c2c3 a7a5 c1f4 f8d6 f4g3 d5c4 c3c4 h7h5 g3h4 f7f6 e2e3 g7g5 h4g3 c6e7 f1d3 h5h4 g3d6 c7d6 h2h3 c8d7 e1g1 e7d5 a1c1 a8c8 c4b3 e8e7 d3e4 d7c6 b3d3 d8d7 f1e1 c6b5 d3b1 e7f7 c1c8 h8c8 e4d5 e6d5 f3g5 f6g5 b1h7 f7f8 h7h8 f8f7');
INSERT INTO evaluations VALUES(15298,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(15298,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(15298,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(15298,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(15298,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(15298,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(15298,36,'cp -25','f7f6 b1c3 f8b4 g5e4 e8g8 e1g1 c8e6 c1e3 b4c3 e4c3 d8d7 c3d5 e6d5 h2h3 a8d8 d1e2 h7h6 f1e1 f8e8 a1d1 d7e6 b2b3 d8d7 f2f3 d5c4 d3c4 e8d8 a2a4 g8f7 g1h1 a7a6 d1d7 d8d7 e1d1 c6d4 e3d4 d7d4 d1d4 e5d4 e2d3 e6e3');
INSERT INTO evaluations VALUES(15869,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(15869,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(15869,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(15869,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(15869,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(15869,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(15869,36,'cp -30','b1c3 f8b4 g5e4 c8e6 e1g1 e8g8 c1e3 b4c3 e4c3 d8d7 c3d5 e6d5 h2h3 a8d8 d1e2 f8e8 f1e1 h7h6 a1d1 a7a6 c4d5 d7d5 a2a3 d8d7 e2g4 d5e6 g4e6 e8e6 f2f4 e5f4 e3f4 e6e1 d1e1 g8f7 f4e3 g7g5 g2g3 h6h5 g1f2 c6d4 e3d4 d7d4 f2e2 h5h4 e2e3 d4a4 g3g4 a4f4 b2b4 f6f5 g4f5');
INSERT INTO evaluations VALUES(31322,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(31322,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(31322,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(31322,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(31322,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(31322,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(31322,36,'cp 27','f1b5 b8c6 f3e5 a8c8 d1e2 f5d7 e5c6 d7c6 b5d3 e7e6 b1d2 a7a6 c2c3 d8d7 d2f3 c6b5 f3e5 b5d3 e5d3 f8d6 e2g4 d6f8 c1f4 g8f6 g4e2 f8d6 f4d6 d7d6 e1g1 a6a5 g2g3 e8g8 g1g2 c8c7 e2f3 h7h6 f1e1 d6b6 a2a4 f8c8 e1e2 g8h8');
INSERT INTO evaluations VALUES(31269,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(31269,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(31269,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31269,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31269,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(31269,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(31269,36,'cp 31','b8d7 c2c4 d5c4 b1c3 g8f6 b5c4 e7e6 d4d5 e6d5 c4d5 d7c5 e1g1 f6d5 c3d5 f5e6 d5f4 f8d6 f3d4 d6f4 c1f4 e8g8 b2b4 c5a4 d4e6 d8d1 f1d1 f7e6 f4e5 f8d8 g1f1 a4b6 d1d8 a8d8 a1c1 d8c8 c1c8 b6c8 b4b5 c8e7 e5d4 a7a6 b5a6 b7a6');
INSERT INTO evaluations VALUES(57096,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(57096,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(57096,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(57096,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(57096,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(57096,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(57096,36,'cp -22','g8f6 a2a3 f8c5 c2c4 c7c6 g1f3 d7d6 d5c6 b8c6 e2e3 e8g8 b1c3 c8f5 f1e2 a7a5 b2b3 f6e4 c3e4 f5e4 c1b2 f7f5 e2d3 e4d3 d1d3 d8e7 e1g1 g8h8 d3d5 f5f4 a1e1 h7h6 e1e2 a8c8 e3f4 f8f4 e2e4 c8f8 e4f4 f8f4');
INSERT INTO evaluations VALUES(49483,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(49483,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(49483,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(49483,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(49483,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(49483,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(49483,36,'cp -25','a2a3 f8c5 c2c4 d7d6 b1c3 a7a5 e2e3 c7c6 g1e2 h7h5 e2g3 h5h4 g3e4 f6e4 c3e4 b8a6 f1e2 c8d7 h2h3 c6d5 c4d5 f7f5 e4c3 e8g8 e1g1 a6c7 c3a4 c5a7 b2b4 b7b5 a4c3 d8g5 b4a5 a7c5 a5a6 a8a6 a1b1 g5g6 c3b5 c7b5 e2b5');
INSERT INTO evaluations VALUES(57008,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(57008,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(57008,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57008,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(57008,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(57008,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(57008,36,'cp 16','d7d5 d2d4 g8f6 b1c3 f8b4 c4d5 e6d5 c1g5 e8g8 e2e3 h7h6 g5h4 c8f5 d1b3 b4c3 b3c3 c7c6 f3d2 b8d7 f1e2 f6e4 h4d8 e4c3 b2c3 f8d8 g2g4 f5h7 h2h4 c6c5 g4g5 a8c8 e2f3 c5d4 c3d4 d7b6 a2a4 h7d3 g5h6 g7h6 a4a5 b6c4');
INSERT INTO evaluations VALUES(55807,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55807,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(55807,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55807,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(55807,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(55807,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(55807,36,'cp 16','d2d4 g8f6 b1c3 f8b4 d1b3 c7c5 d4c5 b8a6 a2a3 b4c3 b3c3 a6c5 c1g5 c5e4 g5f6 d8f6 c3f6 e4f6 c4d5 f6d5 a1c1 c8d7 e2e4 d5f6 f3e5 a8c8 c1c8 d7c8 f2f3 e8e7 e1d2 f6d7 e5d3 h8d8 f1e2 d7e5 h1c1 e5d3 e2d3 e6e5 d2e3');
INSERT INTO evaluations VALUES(56269,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(56269,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(56269,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(56269,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(56269,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(56269,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(56269,36,'cp -39','d5d4 c3b1 c8e6 f1b5 c7c6 b5d3 g8f6 e1g1 f8d6 c2c3 c6c5 d3b5 b8c6 c3d4 c5d4 d2d3 h7h6 b1d2 e8g8 d2b3 e6b3 d1b3 a7a6 b5c6 b7c6 h2h3 c6c5 f3d2 a6a5 d2c4 d6c7 b3b7 f6d5 b7c6 d5b4 c6c5 d8d7 c5h5 b4d3 h5f3 d3b4 c1d2 f8e8');
INSERT INTO evaluations VALUES(56663,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(56663,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(56663,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(56663,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(56663,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(56663,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(56663,36,'cp -43','c3b1 c8e6 f1b5 c7c6 b5d3 g8f6 d1e2 f8e7 d3c4 e6c4 e2c4 c6c5 d2d3 b8c6 a2a4 f6d5 e1g1 d5b4 c1d2 d8d5 c4d5 b4d5 f1e1 f7f6 b1a3 g7g5 e1e4 h7h5 a1e1 e8c8 a3c4 e7d6 c4d6 d8d6 g1f1 h8d8 h2h3 c8d7 f1e2 d5c7');
INSERT INTO evaluations VALUES(11935,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(11935,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(11935,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(11935,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(11935,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11935,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(11935,36,'cp 43','e1g1 d7d6 f1e1 c6c5 c2c3 c8g4 d2d4 e5d4 c3d4 g4f3 d1f3 c5d4 c1f4 f8e7 e4e5 d6e5 f3c6 f6d7 f4e5 e8g8 e5d4 e7f6 d4f6 d7f6 c6c2 f8e8 b1d2 d8d5 b2b3 e8e1 a1e1 a8e8 e1e8 f6e8 h2h3 g7g6 d2f3 c7c5 c2c3 d5d1 g1h2 d1d6 g2g3 f7f6 h2h1 g8g7 c3c2');
INSERT INTO evaluations VALUES(11934,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(11934,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(11934,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11934,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(11934,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(11934,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11934,36,'cp 33','d7d6 f1e1 c8g4 h2h3 g4f3 d1f3 f8e7 b1a3 c6c5 a3c4 f6d7 c4a5 e7h4 g2g3 h4g5 h3h4 g5h6 c2c3 e8g8 b2b4 d8e8 a2a3 e8e6 b4c5 d7c5 d2d4 h6c1 a1c1 c5a4 c1b1 f7f5 e4f5 e6f5 f3f5 f8f5 d4e5');
INSERT INTO evaluations VALUES(11866,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11866,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(11866,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11866,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(11866,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11866,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11866,36,'cp 37','f1e1 c6c5 c2c3 c8g4 d2d4 g4f3 d1f3 c5d4 c3d4 e5d4 c1f4 f8e7 e4e5 d6e5 f3c6 f6d7 f4e5 e8g8 e5d4 e7f6 d4f6 d8f6 c6f6 d7f6 b1c3 f8d8 g1f1 g8f8 a1c1 c7c6 c3a4 d8d6 c1c5 a8c8 g2g3 h7h5 c5a5 c8c7 a5a6 d6d2 e1e2 d2e2 f1e2');
INSERT INTO evaluations VALUES(11867,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11867,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(11867,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(11867,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(11867,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(11867,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11867,36,'cp 30','c8g4 h2h3 g4f3 d1f3 c6c5 d2d3 f8e7 b1a3 f6d7 c1e3 d7f8 c2c3 f8e6 a1d1 e8g8 d3d4 e5d4 c3d4 c5d4 e3d4 e6d4 d1d4 e7f6 d4d2 d8e7 a3c4 a7a5 d2d5 a5a4 e4e5 e7e6 b2b3 a4b3 a2b3 f6e5 c4e5 d6e5');
INSERT INTO evaluations VALUES(18059,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(18059,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(18059,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(18059,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18059,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(18059,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(18059,36,'cp 16','d8e7 d2d4 d7d6 e5c6 e7e4 d1e2 e4e2 e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e7c6 e2f1 e8c8 d1e1 f7f5 b2f6 f8g7 f6g7 g8g7 b1c3 c8d7 a1d1 a6a5 a2a3 g7f7 h2h4 d8b8 d1d5 f7e7 e1e7 d7e7 d5b5 b8b5 c3b5 e7d7 f1e2 c6e5 f2f4 e5c6 b5c3 c6d4 e2d3 d4e6');
INSERT INTO evaluations VALUES(1295,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(1295,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(1295,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(1295,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(1295,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(1295,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(1295,36,'cp 21','d2d4 d7d6 e5c6 e7e4 d1e2 e4e2 e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 b1c3 e7c6 c3d5 e8d7 h1d1 g7g6 c1g5 f8e7 g5f6 e7f6 d5f6 d7e7 f6d5 e7d8 e2f1 a8b8 b2b3 h8e8 c2c4 e8e6 a1c1 c6e7 d5e3 d8d7 c4c5 b8b5 c1c2 h7h5 g2g3 e7c6 d1d2 a6a5 c5d6 e6d6');
INSERT INTO evaluations VALUES(1294,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(1294,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(1294,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1294,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(1294,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(1294,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(1294,36,'cp 7','d7d6 e5c6 e7e4 d1e2 e4e2 e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e7c6 e2f1 e8c8 d1e1 c6b4 e1e2 f8h6 a2a3 b4d5 e2e4 f7f5 e4h4 h6g7 b2g7 g8g7 c2c4 d5f6 b1c3 g7e7 a1e1 e7e1 f1e1 c7c6 e1d2 c8d7 h4h6 a6a5 h2h4 d8b8 d2c2 d7e6 f2f3 d6d5');
INSERT INTO evaluations VALUES(1277,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(1277,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(1277,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1277,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(1277,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(1277,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(1277,36,'cp 16','e5c6 e7e4 d1e2 e4e2 e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e8c8 e2f1 e7c6 d1e1 f7f5 b2f6 f8g7 f6g7 g8g7 b1c3 a6a5 a1d1 g7f7 h2h3 c8d7 c3a4 f7e7 a4c5 d7c8 e1e7 c6e7 c5a4 e7c6 c2c4 c8d7 d1d2 d8b8 f2f3 b8e8 a2a3 f5f4 a4c5 d7e7 c5d3');
INSERT INTO evaluations VALUES(1276,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(1276,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(1276,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(1276,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(1276,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(1276,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(1276,36,'cp 11','e7e4 d1e2 e4e2 e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e7c6 e2f1 e8c8 b2f6 d8e8 b1c3 c6e5 f6g5 e5g4 h2h3 f8g7 g5d2 g4f6 a1c1 e8e6 d2e3 g8e8 d1d3 h7h5 c1d1 f6d7 c3d5 d7f6 d5f6 g7f6 c2c4 f6b2 d1b1 b2f6 g2g3');
INSERT INTO evaluations VALUES(1281,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(1281,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(1281,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(1281,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(1281,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(1281,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(1281,36,'cp 8','d1e2 e4e2 e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e7c6 e2f1 e8c8 b2f6 d8e8 b1c3 c6e5 f6g5 f8g7 c3d5 e5g4 a1c1 h7h6 g5f4 g6g5 f4g3 f7f5 h2h3 g4f6 f2f3 f6d5 d1d5 f5f4 g3e1 e8e6 c2c4 e6e5 d5d3 h6h5 e1c3 e5f5 c3g7 g8g7 c1e1 c8d7');
INSERT INTO evaluations VALUES(1280,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(1280,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(1280,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(1280,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(1280,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(1280,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(1280,36,'cp 12','e4e2 e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 h1d1 e7c6 b1c3 e8c8 e2f1 d8e8 a1b1 g7g6 c3d5 f8g7 c1d2 e8e4 b2b3 h8e8 c2c4 c6e7 d5e7 e4e7 d2e3 e7e4 g2g3 e8e6 d1d5 h7h5 b1c1 e4e5 d5d3 g7f6 h2h4 e5f5 b3b4 e6e4 a2a4 c8b7 b4b5 a6b5');
INSERT INTO evaluations VALUES(1279,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(1279,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(1279,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(1279,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(1279,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(1279,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(1279,36,'cp 11','e1e2 c8b7 d4d5 b7c6 d5c6 g8e7 h1d1 g7g6 c1d2 e7c6 e2f1 e8c8 d2c3 h8g8 a2a3 f7f5 b2b4 f8g7 c3g7 g8g7 b1c3 g6g5 a1b1 c6e5 a3a4 g7f7 a4a5 c7c6 h2h3 h7h5 d1d4 d8d7 c3a4 c8c7 a4b6 d7d8 b6c4 e5c4 d4c4 f7e7 b1d1');
INSERT INTO evaluations VALUES(1278,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(1278,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(1278,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(1278,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(1278,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(1278,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(1278,36,'cp 7','c8b7 d4d5 b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e7c6 e2f1 e8c8 d1e1 c8d7 b2f6 f8g7 f6g7 g8g7 b1c3 f7f5 c3d5 a6a5 e1e2 d8f8 a1d1 c6b4 c2c3 b4d5 d1d5 f8a8 d5d4 g7f7 f2f3 c7c5 d4a4 f5f4 f1f2 f7f5 a4e4 g6g5 c3c4 f5f7 h2h4');
INSERT INTO evaluations VALUES(27479,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(27479,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(27479,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(27479,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(27479,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(27479,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(27479,36,'cp 10','d4d5 b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e7c6 e2f1 e8c8 d1e1 c8d7 b2f6 f8g7 f6g7 g8g7 b1c3 f7f5 a1d1 a6a5 c3d5 g7f7 c2c3 f5f4 f2f3 d8e8 e1e8 d7e8 c3c4 e8d7 c4c5 f7f5 a2a3 c6e7 d5e7 d7e7 c5d6 c7d6 d1d4 e7d7 f1e2 f5b5');
INSERT INTO evaluations VALUES(27478,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(27478,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(27478,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(27478,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(27478,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(27478,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(27478,36,'cp 5','b7c6 d5c6 g8e7 b2b3 g7g6 c1b2 h8g8 h1d1 e7c6 b2f6 e8d7 b1c3 a8e8 e2f1 e8e6 c3d5 c6e7 f6e7 f8e7 d1d4 e7f6 d5f6 e6f6 a1e1 a6a5 d4e4 f6e6 e4f4 e6e1 f1e1 f7f5 f4a4 g8a8 e1d2 c7c5 d2d3 d6d5 a2a3 d7c6 c2c4 d5c4 d3c4 g6g5 b3b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(51846,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(51846,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(51846,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(51846,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(51846,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(51846,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(51846,36,'cp 108','g1f3 d8a5 d1c2 d7d6 d2d4 c8g4 c1d2 b8d7 d4d5 f8e7 e2e4 a5c7 f1e2 g8f6 d2e3 e8g8 e1g1 c6d5 c4d5 f8c8 f1c1 c7d8 h2h3 g4f3 e2f3 a7a6 a2a4 f6e8 f3e2 e7g5 c2d2 g5e3 d2e3 d8e7 a4a5 g7g6 b2b4 d7f6');
INSERT INTO evaluations VALUES(51844,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(51844,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(51844,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(51844,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(51844,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(51844,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(51844,36,'cp 87','d8a5 d1c2 d7d6 d2d4 c8g4 c2b3 a5c7 d4e5 b8d7 e5d6 f8d6 g2g3 g8f6 f1g2 e8g8 c1e3 a8e8 e3d4 g4e6 f3g5 e6g4 a1d1 h7h6 g5f3 g4e6 d4f6 d7f6 f3d4 e6g4 e1g1 d6c5 d4f3 a7a5 d1d2 e8d8 d2d8 f8d8 e2e3 c7b6 b3b6 c5b6');
INSERT INTO evaluations VALUES(1476,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(1476,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1476,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(1476,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(1476,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(1476,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(1476,36,'cp 0','c5d4 c3d4 g8h6 e2d3 c8d7 d3c2 h6f5 c2f5 f8b4 c1d2 e6f5 d2b4 b6b4 b1c3 b4b2 c3d5 e8g8 e1g1 d7e6 d5f4 a8d8 d4d5 c6e5 a1b1 b2c3 f4e6 e5f3 d1f3 c3f3 g2f3 f7e6 d5e6 f8e8 b1b7 e8e6 b7a7 e6g6 g1h1 d8d3 a2a4 d3f3 a7a8 g8f7 a4a5 f3a3 f1b1 a3a4 b1b7 f7e6 a8a6 e6e5 a6g6 h7g6 b7g7 a4a5 g7g6');
INSERT INTO evaluations VALUES(1523,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(1523,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(1523,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(1523,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(1523,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(1523,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(1523,36,'cp 17','c3d4 g8h6 e2d3 h6f5 d3f5 e6f5 b1c3 c8e6 h2h4 f8e7 e1g1 h7h6 f1e1 a8c8 c3a4 b6b5 e1e2 c6b4 c1e3 b5d3 d1d3 b4d3 e2d2 d3b4 a2a3 b4c6 b2b4 b7b6 a1c1 e8d7 c1c3 c8c7 h4h5 h8c8 e3f4 c6d8 d2c2 c7c3 a4c3');
INSERT INTO evaluations VALUES(1510,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(1510,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(1510,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1510,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(1510,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(1510,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1510,36,'cp 10','g8h6 e2d3 c8d7 d3c2 h6f5 c2f5 e6f5 b1c3 d7e6 h2h4 f8e7 c3e2 h7h6 e1g1 a7a5 f1e1 a5a4 a1b1 e8g8 e2f4 f8c8 e1e2 b6d8 h4h5 e7g5 e2c2 g5f4 c1f4 d8d7 c2c1 c6d8 d1d2 c8c1 b1c1 a8c8 c1c8 d7c8 d2b4 c8c4 b4d6 d8c6');
INSERT INTO evaluations VALUES(826,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(826,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(826,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(826,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(826,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(826,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(826,36,'cp 12','e2d3 h6f5 d3f5 e6f5 b1c3 c8e6 h2h4 h7h6 c3e2 f8e7 e1g1 a7a5 f1e1 a5a4 a1b1 e8g8 e2f4 b6a6 c1d2 f8c8 b1c1 a4a3 b2b3 c6a7 c1c8 a8c8 d1b1 a6b6 f4e2 a7b5 h4h5 c8c6 b1d3 b5a7 e2f4 b6b5 d3b1 b5b6 d2e3 b6a6 e1e2');
INSERT INTO evaluations VALUES(10102,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(10102,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(10102,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(10102,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(10102,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(10102,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(10102,36,'cp -29','f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 e8g8 d2d4 c5d6 c1g5 h7h6 g5h4 d6e5 d4e5 d8d1 f1d1 f6h5 f2f3 g7g5 h4g3 f8e8 b1a3 c8e6 a3c2 c6c5 c2e3 e8d8 b2b3 a7a5 c3c4 a5a4 d1d8 a8d8 a1d1 d8d4 d1d4 c5d4 e3d5 c7c5 d5f6 h5f6 e5f6 a4b3 a2b3 e6f5 g3d6');
INSERT INTO evaluations VALUES(10626,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(10626,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(10626,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(10626,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(10626,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(10626,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(10626,36,'cp -39','c4d5 e4f6 d5c6 d7c6 f3e5 c5d6 d2d4 e8g8 c1g5 h7h6 g5h4 d6e5 d4e5 d8d1 f1d1 f6h5 f2f3 g7g5 h4f2 f8e8 c3c4 c8f5 b1c3 h5f4 g2g4 f5g6 d1d7 a8d8 a1d1 f4d3 d7d8 e8d8 g1g2 d8e8 b2b3 a7a5 f2g3 d3e5 h2h4 f7f6 g3e5 e8e5 d1d8 g8f7 h4h5 g6h7 d8c8 e5e7 g2f2 f6f5');
INSERT INTO evaluations VALUES(47175,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(47175,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(47175,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(47175,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(47175,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(47175,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(47175,36,'cp -63','c7c5 e2e4 d5e4 d4c5 d8d1 e1d1 e7e5 c3e4 f6e4 f3e4 f8c5 f1c4 e8e7 g1f3 f7f6 c2c3 a7a5 d1e2 b8d7 c4d5 a8b8 h1d1 b7b5 d5c6 h8d8 c1d2 b5b4 c3c4 c8a6 b2b3 b8c8 c6d7 d8d7 d2e3 d7d1 a1d1 a6b7 e3c5 c8c5');
INSERT INTO evaluations VALUES(45372,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(45372,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(45372,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(45372,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(45372,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(45372,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(45372,36,'cp -44','e2e4 b8c6 e4d5 f6d5 d4c5 d5c3 d1d8 c6d8 b2c3 c8d7 g1h3 a8c8 h3g5 c8c5 g5e4 c5c8 c1b2 d8e6 e1c1 h7h5 c3c4 h8h6 g2g3 h5h4 f1e2 d7c6 c1b1 c6e4 f3e4 e6c5 e4e5 h6b6 b1a1 e7e6 b2d4 b6a6 c2c3 h4g3 h2g3');
INSERT INTO evaluations VALUES(44258,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(44258,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(44258,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(44258,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(44258,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(44258,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(44258,36,'cp 88','g7g6 e2e4 d7d6 g1f3 f8g7 h2h3 e6d5 e4d5 e8g8 f1d3 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1b3 b8a6 d3e4 e8e4 a1e1 d8b6 b3a4 a8e8 b2b3 b6c7 e3d2 e4e1 f1e1 e8e1 f3e1 c7e7 e1f3 h7h6 d2f4 g6g5 f4d2 g8f8 a4b5 a6b8 g2g4 b7b6 g1g2 e7d7 a2a4 d7b5');
INSERT INTO evaluations VALUES(44023,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(44023,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(44023,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(44023,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(44023,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(44023,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(44023,36,'cp 89','e2e4 d7d6 f1d3 f8g7 h2h3 e6d5 e4d5 e8g8 g1f3 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1b3 b8d7 e3f4 d8c7 a1e1 e4f6 g2g4 e8e1 f1e1 a8e8 e1e8 f6e8 d3c2 g7e5 f4c1 a7a6 b3e3 b7b5 e3e2 b5c4 e2c4 c7a5 c2d3 e8c7 c1d2 a5b5 b2b3');
INSERT INTO evaluations VALUES(17433,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(17433,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(17433,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(17433,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(17433,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(17433,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(17433,36,'cp 32','c1e3 g8f6 a2a3 d8c7 f2f4 d7d5 e4e5 f6e4 c3e4 d5e4 d4c6 c7c6 f1e2 f8e7 d1d4 e8g8 d4b6 c8d7 e1c1 a8c8 c2c3 c6b6 e3b6 d7a4 d1d4 a4c6 h2h4 c6d5 d4d1 c8c6 b6e3 c6c7 h4h5 h7h6 g2g4 b7b5 g4g5 f7f5 e5f6 g7f6 g5g6 f6f5 e3b6');
INSERT INTO evaluations VALUES(17430,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(17430,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17430,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(17430,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(17430,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17430,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(17430,36,'cp 25','g8f6 a2a3 d8c7 f2f4 d7d5 e4e5 f6e4 c3e4 d5e4 d4c6 c7c6 f1e2 f8e7 d1d4 e8g8 d4b6 c8d7 a1d1 a8d8 b6c6 d7c6 d1d8 f8d8 e1f2 f7f6 e5f6 g7f6 h1d1 d8d1 e2d1 f6f5 g2g3 c6b5 c2c3 b5d3');
INSERT INTO evaluations VALUES(11092,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(11092,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(11092,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(11092,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(11092,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(11092,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(11092,36,'cp 36','a2a3 d8c7 f2f4 d7d5 e4e5 f6e4 c3e4 d5e4 d4c6 c7c6 d1d4 f8e7 f1e2 e8g8 d4b6 c8d7 a1d1 c6c8 c2c3 d7c6 h2h4 f8d8 h4h5 h7h6 e1f2 d8d1 h1d1 c8e8 e3c5 g8h7 g2g3 e7d8 b6b4 f7f5 e5f6 d8f6 c5d4');
INSERT INTO evaluations VALUES(11098,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11098,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(11098,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11098,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(11098,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11098,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(11098,36,'cp 30','d8c7 f2f4 d7d5 e4e5 f6e4 c3e4 d5e4 d4c6 c7c6 f1e2 f8e7 d1d4 e8g8 c2c4 c8d7 a1d1 a8d8 d4b6 c6c8 e1g1 h7h6 b2b3 g8h7 a3a4 f7f5 e5f6 e7f6 b6c5 d7c6 d1d8 f8d8 f1d1 c8c7 d1d8 c7d8 g2g3 d8e8 g1f2');
INSERT INTO evaluations VALUES(671,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(671,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(671,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(671,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(671,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(671,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(671,36,'cp 32','f2f4 d7d5 e4e5 f6e4 c3e4 d5e4 d4c6 c7c6 c2c4 f8e7 d1d4 c8d7 e1c1 c6c7 d4e4 d7a4 d1d2 e8g8 c1b1 f8d8 f1d3 g7g6 e4f3 d8d3 d2d3 a4c6 f3e2 c6e4 b1a1 e4d3 e2d3 c7c6 d3c2 e7c5 e3c5 c6c5 h1c1 c5e3 g2g3 a8d8 a1a2 h7h5 c1d1 d8d1 c2d1');
INSERT INTO evaluations VALUES(40063,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40063,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(40063,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(40063,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40063,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40063,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(40063,36,'cp 20','e2e3 h7h6 g5h4 e8g8 f1e2 d5c4 e2c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 b5d6 f8d8 d6e4 d8d1 e4f6 g7f6 f1d1 b4e7 c4b5 g8f8 f3d2 c6e5 d2e4 f6f5 e4f6 e5g6 f6d7 c8d7 h4e7 g6e7 d1d7 a8b8 f2f4 e7d5 g1f2 b8c8 d7b7 c8c2 f2f3 c2b2');
INSERT INTO evaluations VALUES(12774,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12774,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12774,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12774,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12774,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12774,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12774,36,'cp 0','e7e6 d2d4 f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 f2f3 e4e3 c4b5 e8f7 d2c4 a7a6 b5a4 f8c5 c1e3 c5e3 c4e3 b7b5 a4b3 f7f6 a2a4 c8b7 e1c1 d8f7 h1e1 h8d8 d1d8 f7d8 e1d1 f6e7 d1e1');
INSERT INTO evaluations VALUES(23203,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(23203,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(23203,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(23203,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(23203,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(23203,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(23203,36,'cp 146','c6e7 c2c4 d7d6 b1c3 g7g6 f1e2 f8g7 e1g1 g8h6 b2b4 e8g8 a2a4 h6f7 a4a5 f6f5 f3g5 f7g5 c1g5 h7h6 g5d2 f5e4 c3e4 e7f5 e2d3 f5d4 f2f3 c8f5 a5a6 b7b6 g2g4 f5d7 g1h1 g8h8 b4b5 d8e7 d2e3 a8d8 h1g2 d8e8 e4g3 e7f7 h2h4 d4f5 g4f5');
INSERT INTO evaluations VALUES(23361,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(23361,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(23361,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(23361,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(23361,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(23361,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(23361,36,'cp 144','c2c4 d7d6 b1c3 g7g6 f1e2 f8g7 e1g1 f6f5 c4c5 g8f6 e2b5 e8f8 b5d3 h7h6 c5d6 c7d6 f3d2 f8g8 a2a4 f5e4 d2e4 f6e4 d3e4 c8f5 c1e3 g8h7 a4a5 h8f8 d1b3 f5e4 c3e4 e7f5 f1c1 d8e7 b3d3 e7f7 a5a6 f5e3 f2e3');
INSERT INTO evaluations VALUES(23359,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(23359,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(23359,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(23359,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(23359,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(23359,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(23359,36,'cp 142','d7d6 b1c3 g7g6 f1e2 f8g7 e1g1 g8h6 a2a4 e8g8 a4a5 h6f7 b2b4 f6f5 f3g5 f7g5 c1g5 h7h6 g5d2 f5e4 g2g4 g6g5 a1a3 e7g6 c3e4 g6f4 f1e1 c8d7 e2f1 d8e8 h2h3 g8h8 d2f4 g5f4 g1h1 a7a6 f2f3 e8g6 f1d3 a8e8 c4c5');
INSERT INTO evaluations VALUES(20797,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(20797,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(20797,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(20797,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(20797,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(20797,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(20797,36,'cp 150','b1c3 g7g6 f1e2 f8g7 e1g1 g8h6 a2a4 h6f7 a4a5 e8g8 b2b4 f6f5 f3g5 f7g5 c1g5 h7h6 g5c1 f5e4 g2g4 g6g5 c3e4 e7g6 a1a3 g6f4 f2f3 c8d7 e2d3 a7a6 d1c2 d8e7 g1h1 g7f6 c1f4 e5f4 h1g2 a8e8 c4c5 f6d4 c5c6 b7c6 d5c6');
INSERT INTO evaluations VALUES(55819,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55819,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55819,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55819,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55819,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(55819,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(55819,36,'cp -5','d5c4 e2e3 b7b5 b2b3 c4b3 f1b5 b8d7 d1b3 a8b8 b3a4 g8f6 b5c6 c8b7 g1f3 b7c6 a4c6 b8b6 c6a4 d8c8 b1d2 f8e7 e1g1 e8g8 d2c4 b6a6 c4a5 c7c5 c1d2 c8a8 a1c1 f8b8 d4c5 d7c5 a4c4 e7f8 d2b4 a6a5 b4a5 c5b3 c4a4 b3c1 f1c1 a8d5');
INSERT INTO evaluations VALUES(56026,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(56026,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56026,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(56026,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(56026,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(56026,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(56026,36,'cp 0','e2e3 b7b5 b2b3 c4b3 f1b5 b8d7 d1b3 a8b8 b3a4 g8f6 b5c6 c8b7 g1f3 b7c6 a4c6 f8e7 b1d2 b8b6 c6a4 d8c8 e1g1 e8g8 d2c4 b6a6 c4a5 f6e4 f3d2 e4f6');
INSERT INTO evaluations VALUES(56025,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56025,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(56025,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(56025,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(56025,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(56025,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(56025,36,'cp -9','b7b5 b2b3 c4b3 f1b5 b8d7 d1b3 a8b8 b3a4 g8f6 b5c6 c8b7 g1f3 b7c6 a4c6 f8e7 e1g1 e8g8 a3a4 b8b6 c6c2 c7c5 c1a3 d8a8 f1c1 f8c8 d4c5 e7c5 b1d2 b6c6 a3c5 c6c5 c2b2 a8c6 c1c5 c6c5 h2h3 h7h6 d2b3 c5b4 a4a5');
INSERT INTO evaluations VALUES(17906,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(17906,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(17906,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(17906,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(17906,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(17906,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(17906,36,'cp 103','d2d4 e5d4 f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 f8d6 c1e3 d8e7 g2g4 c8e6 c3e2 f6d7 e1f2 h7h6 h2h4 g7g6 e2d4 c6c5 d4e6 f7e6 f1e2 a8f8 a2a4 d6f4 g4g5 a6a5 a1d1 f4e3 f2e3 b7b6 b2b3 h6g5 h4g5 d7e5 f3f4 e5c6 h1h8 f8h8');
INSERT INTO evaluations VALUES(17901,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(17901,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(17901,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(17901,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(17901,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(17901,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(17901,36,'cp 98','e5d4 f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 f8d6 c1e3 d8e7 g2g4 c8e6 c3e2 f6d7 e2d4 d7e5 e1f2 h7h5 a1d1 g7g5 e3g5 f7f6 g5c1 h5g4 d4e6 e7e6 f3f4 e5f7 c1e3 e6e7 h2h4 g4h3 h1h3 h8h3 f1h3 a8h8 d1h1 c6c5 b2b3 b7b6 f2f3 a6a5');
INSERT INTO evaluations VALUES(17920,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(17920,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(17920,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(17920,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(17920,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(17920,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(17920,36,'cp 113','f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 f8d6 c1e3 h7h5 e3d4 c6c5 d4e3 h5h4 f3f4 f6g4 e3d2 b7b5 e4e5 d6e7 f1e2 h4h3 e2f3 a8b8 g2h3 h8h3 f3g2 h3h8 h2h3 b5b4 h3g4 h8h1 g2h1 b4c3 d2c3 c8g4 e1f2 g4e6 h1e4 f7f5 e4f3 d8e8 a1h1 e8f7 f2g2');
INSERT INTO evaluations VALUES(17918,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(17918,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(17918,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(17918,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(17918,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(17918,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(17918,36,'cp 106','g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 f8d6 c1e3 h7h5 e3d4 c6c5 d4e3 h5h4 f1e2 b7b5 e1g1 c8e6 a1d1 d8c8 a2a4 b5b4 c3b1 c8b7 b1d2 a8d8 d2c4 d6e7 b2b3 g7g6 e3g5 h4h3 d1d8 e7d8 f1d1 f6d7 g5d8 h8d8 g2h3 e6h3');
INSERT INTO evaluations VALUES(34145,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(34145,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(34145,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(34145,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(34145,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34145,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(34145,36,'cp 31','f1d3 g8f6 d1e2 f8d6 f2f4 d6f4 c1f4 c7f4 h1f1 f4h2 e4e5 h2h4 g2g3 h4g3 e1d2 g3g5 d2d1 g5g4 e5f6 g4d4 f6g7 h8g8 c3e4 e8d8 c2c3 d4e5 f1f7 g8g7 f7g7 e5g7 d1c2 d7d5 e4c5 d8c7 a1h1 h7h6 e2f2 b8c6 h1g1 g7e5 g1g8 d5d4 c5b3');
INSERT INTO evaluations VALUES(34146,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(34146,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(34146,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(34146,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(34146,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34146,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(34146,36,'cp 26','g8f6 e1g1 f8e7 d1e2 d7d6 f2f4 e8g8 g1h1 b8d7 c1e3 d7c5 f1f3 c8d7 f3g3 c5d3 c2d3 b7b5 a2a3 c7b7 e4e5 f6d5 e2h5 g7g6 h5h6 f7f5 c3d5 b7d5 a1d1 f8f7 g3h3 a8c8 d1f1 e7f8');
INSERT INTO evaluations VALUES(33534,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(33534,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(33534,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(33534,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(33534,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(33534,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(33534,36,'cp 49','f2f4 f8b4 c1d2 b4c3 b2c3 d7d6 e1g1 e6e5 f4e5 d6e5 d1e1 e5d4 c3d4 b8c6 e4e5 f6d5 c2c4 d5e7 d2c3 e8g8 d4d5 e7g6 d5c6 c7c6 e1e4 c8e6 e4c6 b7c6 f1b1 a8b8 g1f2 h7h5 d3e2 g6f4 e2f1 h5h4 b1d1 f8d8 f2e3 d8d1 a1d1 f4g6 d1d6 b8b1 f1d3');
INSERT INTO evaluations VALUES(48318,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(48318,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(48318,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(48318,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48318,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(48318,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(48318,36,'cp -39','c7c5 d4c5 b8c6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 c2c3 d4c3 e2c3 h7h6 c1e3 c5e3 f2e3 c8e6 d1e2 a8c8 a1c1 d8d6 f1d1 c6e7 d3b1 d6b8 h2h3 b7b6 g1h2 f8d8 d1d8 c8d8 b1c2 e7g6');
INSERT INTO evaluations VALUES(46227,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(46227,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(46227,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(46227,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(46227,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(46227,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(46227,36,'cp -31','d4c5 b8c6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3a4 c5e7 c2c3 c8g4 d3b5 d8a5 b5c6 b7c6 c3d4 a8d8 c1d2 a5a6 d1c2 g4f3 g2f3 d8d4 a4c3 f6h5 d2e3 e8g8 c3e2 d4c4 c2b3 c4b4 b3c2 c6c5 f1d1 a6e6 b2b3 e6h3 e2g3 h5f4 e3f4 e5f4');
INSERT INTO evaluations VALUES(46098,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(46098,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(46098,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(46098,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(46098,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(46098,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(46098,36,'cp -26','b8c6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3a4 c5d6 c2c3 c8g4 d3b5 e8g8 b5c6 b7c6 h2h3 g4f3 d1f3 h7h6 b2b3 f8e8 c1d2 e8e6 a4b2 d8c7 f1e1 c6c5 c3d4 c5d4 a1c1 c7b7 b2c4 d6c5 a2a3 c5f8 b3b4 b7a6 c4a5 a8c8 c1c8');
INSERT INTO evaluations VALUES(55900,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(55900,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(55900,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(55900,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(55900,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(55900,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(55900,36,'cp -7','d5e4 d3e4 d8d1 e1d1 b8c6 c1f4 f8d6 f4g3 d6g3 h2g3 g8f6 b1d2 a7a6 f1d3 e8e7 d2c4 b7b6 f3e5 c6e5 c4e5 f6d7 e5d7 c8d7 a2a4 a6a5 a1a3 e6e5 a3c3 c7c5 b2b3 a8d8 d1e2 d7e6 d3c4 e6c4 c3c4 h7h5 e2e3 g7g5 e3e2 h8h6 c2c3 d8d6 b3b4 c5b4 c3b4 a5b4 c4b4');
INSERT INTO evaluations VALUES(56083,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(56083,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56083,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(56083,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(56083,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(56083,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(56083,36,'cp -12','d3e4 d8d1 e1d1 b8c6 c1f4 g8f6 b1d2 f8d6 f4g3 d6g3 h2g3 a7a6 f1d3 e8e7 d1e2 h7h6 d2c4 a6a5 c4e5 c6e5 f3e5 f6d7 e5c4 b7b6 a1d1 h8d8 b2b3 f7f6 g3g4 d7c5 e4e5 c8b7 e5f6 e7f6 f2f3 c5d3 d1d3 d8d3 e2d3 a8d8 d3e3 e6e5');
INSERT INTO evaluations VALUES(56072,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(56072,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56072,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(56072,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(56072,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(56072,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(56072,36,'cp -9','d8d1 e1d1 b8c6 c1f4 f8d6 f4g3 b7b6 b1d2 c8b7 f1b5 g8e7 c2c3 e8c8 d1c2 e7g6 a2a4 c6e5 h2h4 a7a6 f3e5 d6e5 g3e5 g6e5 b5e2 d8d7 f2f3 h8d8 a1d1 a6a5 e2b5 c7c6 b5e2 c8c7 d2c4 d7d1 h1d1 d8d1 c2d1');
INSERT INTO evaluations VALUES(34847,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(34847,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(34847,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(34847,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(34847,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(34847,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(34847,36,'cp -10','e1d1 b8c6 c1f4 f8d6 f4g3 b7b6 b1d2 c8b7 f1b5 e8c8 a2a4 g8e7 c2c3 a7a6 b5e2 e7g6 d1c2 c6e5 f3e5 g6e5 f2f3 g7g5 d2c4 e5c4 e2c4 h8g8 g3d6 c7d6 h2h3 h7h5 h1f1 c8c7 g2g3 h5h4 g3g4 g8h8 b2b3');
INSERT INTO evaluations VALUES(34846,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(34846,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(34846,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(34846,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(34846,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(34846,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(34846,36,'cp -13','b8c6 c1f4 f8d6 f4g3 b7b6 f1b5 c8b7 b1d2 e8c8 c2c3 a7a6 b5c6 b7c6 d1c2 g8h6 g3d6 c7d6 f3d4 c6b7 f2f3 e6e5 d4e2 d6d5 e4d5 b7d5 e2g3 d5e6 a1e1 h8e8 b2b3 f7f6 d2c4 e6c4 b3c4 c8c7 a2a4 h6f7');
INSERT INTO evaluations VALUES(1684,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(1684,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(1684,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(1684,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(1684,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(1684,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(1684,36,'cp 0','c1f4 f8d6 f4g3 d6g3 h2g3 g8f6 b1d2 a7a6 f1d3 h7h6 d1e2 e8e7 a2a4 b7b6 d2b3 c8b7 a4a5 h8d8 f3e1 c6e5 e1f3');
INSERT INTO evaluations VALUES(43258,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(43258,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(43258,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(43258,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(43258,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(43258,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(43258,36,'cp 40','d7d5 e4e5 f6e4 f1d3 e4c3 b2c3 c7c5 e1g1 b8c6 d3b5 d8c7 a2a4 e8g8 c1a3 b7b6 h2h3 c8f5 f1e1 f8d8 f3h4 f5d7 b5f1 a7a6 d1d2 c6a5 h4f3 d7f5 g2g4 f5d7 d2g5 d7a4 a3c5 b6c5 a1a4 h7h6 g5c1 e7e6 c1e3 g7f8 e1a1');
INSERT INTO evaluations VALUES(41509,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41509,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(41509,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(41509,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(41509,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(41509,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(41509,36,'cp 48','e4e5 f6e4 f1d3 e4c3 b2c3 c7c5 h2h3 c5c4 d3e2 f7f6 e1g1 e8g8 a2a4 b8c6 e5f6 e7f6 f1e1 f8e8 f3h2 c8e6 c1f4 d8d7 h2f1 a8d8 e2f3 c6e7 d1b1 e7f5 b1b5 f5h4 f3d1 h4f5 g2g4 f5d6 f4d6 d7d6 b5b7 f6f5 d1f3 e8f8 g4f5 e6f5 b7a7 g8h8 e1e7');
INSERT INTO evaluations VALUES(41517,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(41517,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(41517,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(41517,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(41517,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(41517,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(41517,36,'cp 41','f6e4 f1d3 e4c3 b2c3 c7c5 h2h3 c5c4 d3e2 f7f6 e1g1 e8g8 f1e1 b8c6 e5f6 e7f6 f3h2 f8e8 h2f1 c8e6 a2a4 d8d7 c1f4 e6f7 e2g4 e8e1 d1e1 f6f5 g4f3 a8e8 e1d2 c6d8 f1e3 d8e6 e3d5 e6d4 c3d4 f7d5 f3d5 d7d5 c2c3 e8e4 a1b1');
INSERT INTO evaluations VALUES(41560,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(41560,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(41560,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41560,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(41560,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41560,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(41560,36,'cp 34','f1d3 e4c3 b2c3 c7c5 h2h3 b8c6 e1g1 d8a5 c1d2 c5c4 d3e2 a5a4 d1c1 c8f5 e2d1 e8c8 f1e1 h7h6 f3h4 f5e6 g2g4 a4a5 a1b1 a5c7 h4g2 e6d7 h3h4 c7a5 c1b2 a5c7 e5e6 d7e6 g2f4 d8d6 e1e6 f7e6 f4g6 h8g8 h4h5 e6e5 d4e5 d6g6 h5g6');
INSERT INTO evaluations VALUES(42596,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42596,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42596,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(42596,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42596,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(42596,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(42596,36,'cp 10','c7c6 e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 e8g8 e1g1 b8c6 c1e3 c8g4 d1d3 g4e2 d3e2 c6d4 e3d4 b6d4 a1d1 d8f6 b2b4 a8d8 e2g4 d4c3 g2d5 g7g6 b4b5 c3a5 d5b7 d8d1 f1d1 a5b6 g4f3 f6f3 b7f3 f8d8 d1d8 b6d8 g1g2 d8b6 a2a3');
INSERT INTO evaluations VALUES(39875,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39875,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(39875,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(39875,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(39875,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(39875,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(39875,36,'cp 12','e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 e8g8 e1g1 b8c6 c1e3 c8g4 d1d3 g4e2 d3e2 b6d4 e3d4 c6d4 e2d2 d8f6 a1d1 d4f3 g2f3 f6f3 d2d5 f3f6 d5d4 f8d8 d4f6 g7f6 f1e1 b7b6 b2b4 a7a5 b4b5 a8c8 a2a4 d8d1 e1d1');
INSERT INTO evaluations VALUES(39873,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(39873,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(39873,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(39873,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(39873,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(39873,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(39873,36,'cp 10','c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 e8g8 e1g1 b8c6 c1e3 c8g4 d1d3 g4e2 d3e2 c6d4 e3d4 b6d4 a1d1 d8f6 b2b4 d4c3 b4b5 a8e8 e2d3 g7g6 d3d5 b7b6 d5c6 e8d8 d1d5 d8d5 g2d5 f8d8 c6f6 c3f6 f1d1 d8d6');
INSERT INTO evaluations VALUES(32089,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(32089,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(32089,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(32089,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(32089,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(32089,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(32089,36,'cp 52','g8f6 h2h3 g4f3 d1f3 b8c6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 e2g3 d8d7 c2c3 e8g8 b2b4 a5c4 d3c2 d6c7 e4d5 d7d5 f3e2 c4b6 d2d3 d5e6 c3c4 a7a5 b4a5 a8a5 a2a4 a5a8 c1d2 b6d7 f1b1 d7c5 d2b4 f6d7 e2e3 c7d6 a4a5 f7f5 b4c3');
INSERT INTO evaluations VALUES(29897,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(29897,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(29897,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(29897,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(29897,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(29897,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(29897,36,'cp 71','h2h3 g4f3 d1f3 b8c6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 a1b1 e8g8 e2g3 d8d7 b2b4 a5c4 b4b5 c4b6 f1e1 d7e6 b5c6 b7c6 g3f5 d6c7 e4d5 c6d5 f3g3 g7g6 f5d4 e6d7 g3f3 f6e4 d4b5 e4c5 d3f1 c5e6 b5c7 d7c7 c1b2 d5d4');
INSERT INTO evaluations VALUES(29895,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(29895,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(29895,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(29895,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(29895,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(29895,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(29895,36,'cp 62','g4f3 d1f3 b8c6 c3e2 c6a5 c4d3 g7g6 c2c3 f8g7 d3c2 e8g8 b2b4 a5c6 d2d3 c6b8 a2a4 a7a5 b4b5 c7c6 g2g4 b8d7 h3h4 d7c5 h4h5 c5e6 c1e3 c6b5 h5g6 h7g6 a4b5 d6d5 g4g5 f6h5 e4d5 e6g5 f3g2 h5f4 e2f4');
INSERT INTO evaluations VALUES(29916,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(29916,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(29916,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(29916,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(29916,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(29916,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(29916,36,'cp 72','d1f3 b8c6 c3e2 c6a5 c4d3 g7g6 c2c3 f8g7 b2b4 a5c6 d3c2 e8g8 e1g1 a7a5 b4b5 c6b8 d2d3 c7c6 a2a4 b8d7 a1b1 d8c7 c1e3 d6d5 g2g4 c7d6 f3g2 d5e4 d3e4 d7c5 f1d1 d6e7 e2g3 c5e6 g4g5 f6d7 b5c6 b7c6 b1b7 a8d8 c2b3 e6g5');
INSERT INTO evaluations VALUES(29913,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(29913,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(29913,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(29913,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(29913,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(29913,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(29913,36,'cp 43','b8c6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 a1b1 e8g8 e2g3 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 g3f5 f8e8 a2a4 d6f8 a4a5 a7a6 f5e3 a8d8 e3c4 d8d7 f1d1 h7h6 c1e3 d7d3 d1d3 e6c4 e3c5 f6e4');
INSERT INTO evaluations VALUES(25316,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(25316,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(25316,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(25316,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(25316,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(25316,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(25316,36,'cp 53','c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 e2g3 e8g8 a1b1 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 f8e8 b5c6 b7c6 c1g5 f6d7 f1c1 d6f8 c1c2 a7a5 g3f5 a5a4 b1c1 c6c5 g5d2 e8b8 f5e3 a4a3 f3d1 b8b7 c1b1');
INSERT INTO evaluations VALUES(35097,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35097,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(35097,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(35097,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(35097,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(35097,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(35097,36,'cp 25','b1c3 b8c6 f1c4 f8c5 g1f3 g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 f3d2 g6e7 f1f2 e8g8 a1f1 e6c4 d2c4 a8d8 h2h3 f7f6 c3d5 d8c8 a2a4 g8f7 a4a5 e7d5 e4d5');
INSERT INTO evaluations VALUES(35094,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(35094,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(35094,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(35094,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(35094,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(35094,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(35094,36,'cp 12','b8c6 f1c4 f8c5 g1f3 g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 c4d5 e8d7 f3d2 f7f6 a2a3 a7a5 f1f2 g6e7 a1f1 h8b8 d5e6 d7e6 a3a4 c6b4 c3b5 e6d7 d2c4 d6d5 e4d5 e7d5 b2b3 d5e7');
INSERT INTO evaluations VALUES(43827,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43827,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(43827,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(43827,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(43827,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43827,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43827,36,'cp 25','g2g3 c8a6 b2b3 f8b4 c1d2 b4e7 f1g2 c7c6 d2c3 d7d5 b1d2 b8d7 e1g1 e8g8 f1e1 c6c5 e2e4 d5e4 d2e4 f6e4 e1e4 a6b7 e4e2 e7f6 e2d2 d8e7 d1e2 a8c8 a1d1 f8d8 h2h4 h7h6 c3b2 c5d4 b2d4 e6e5 d4b2 e5e4');
INSERT INTO evaluations VALUES(43828,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(43828,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(43828,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43828,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(43828,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43828,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(43828,36,'cp 12','c8a6 b2b3 f8b4 c1d2 b4e7 f1g2 c7c6 d2c3 d7d5 b1d2 b8d7 e1g1 e8g8 f1e1 c6c5 c4d5 e6d5 c3b2 a6b7 a1c1 f8e8 e2e3 a8c8 d1e2 d7f8 h2h3 f8g6 e1d1 e7d6 a2a3 d8e7 d4c5 b6c5');
INSERT INTO evaluations VALUES(28603,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(28603,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(28603,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(28603,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(28603,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28603,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28603,36,'cp 21','b2b3 f8b4 c1d2 b4e7 f1g2 c7c6 d2c3 d7d5 b1d2 b8d7 e2e4 f6e4 d2e4 d5e4 f3e5 d7e5 d4e5 e8g8 g2e4 d8c7 e1g1 b6b5 d1e2 a8d8 f1d1 b5c4 b3c4 d8d1 a1d1 f8d8 d1d8 c7d8 e2d3 g7g6 d3d8 e7d8 e4d3 c6c5 f2f3 a6b7 g1f2 b7c6 f2e3 c6a4 f3f4 g8g7 d3e4 f7f5 e5f6 d8f6');
INSERT INTO evaluations VALUES(28602,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(28602,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(28602,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(28602,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(28602,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(28602,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(28602,36,'cp 21','d7d5 f1g2 f8b4 c1d2 b4e7 c4d5 e6d5 b1c3 e8g8 e1g1 b8d7 f3e5 a6b7 a1c1 f8e8 c1c2 h7h6 d2f4 d7f8 e5d3 f8e6 f4e5 f6e4 e2e3 e4c3 c2c3 c7c5 h2h4 e7f6 d1g4 d8e7 f1e1 a8d8 c3c1 b7a6 e1d1 a6b7 e5f6 e7f6 d3f4 c5d4 f4e6 e8e6');
INSERT INTO evaluations VALUES(22808,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(22808,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(22808,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(22808,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(22808,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(22808,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(22808,36,'cp -22','g8f6 b1c3 d7d5 e4d5 f6d5 e2b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 h7h6 f1e1 d8f6 c2c3 c8f5 a2a4 c6a5 c1a3 d6a3 a1a3 a8d8 h2h3 f5e6 d1e2 c7c5 e2e5 f6e5 e1e5 c5d4 f3d4 a5c4 b5c4 e6c4 f2f3 f8e8 e5e8 d8e8');
INSERT INTO evaluations VALUES(16569,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(16569,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16569,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(16569,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(16569,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(16569,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(16569,36,'cp -21','b1c3 d7d5 e4d5 f6d5 e2b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 h7h6 c1b2 c8g4 b5e2 f8e8 d4d5 g4f3 e2f3 c6e5 f3e2 d6c5 g2g3 d8f6 a1c1 e8e7 c2c4 e5f3 e2f3 f6b2 c1c2 b2f6 g1g2 a8d8 c2d2 b7b6 d1c2 a7a5 c2d3 d8e8 f1d1');
INSERT INTO evaluations VALUES(16508,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(16508,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(16508,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(16508,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(16508,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(16508,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(16508,36,'cp -24','d7d5 e4d5 f6d5 e2b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 h7h6 c2c3 d8f6 e1g1 e8g8 b5d3 f8e8 c1e3 c6e7 c3c4 c7c6 f1e1 b7b6 d1c2 c8f5 e3d2 f6g6 d3f5 e7f5 c2d3 c6c5 e1e8 a8e8 d4d5 f5d4 d3g6 f7g6 a1e1');
INSERT INTO evaluations VALUES(15131,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(15131,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(15131,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(15131,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(15131,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(15131,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(15131,36,'cp -16','e4d5 f6d5 e2b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 h7h6 f1e1 c8g4 c2c3 d8f6 h2h3 g4h5 a1b1 f8e8 e1e8 a8e8 c1e3 a7a6 b5e2 b7b5 c3c4 b5c4 e2c4 d6f4 e3f4 f6f4 c4a6 e8d8 d1c1 f4d6 d4d5 h5f3 d5c6 f3d5 b1b2 d5c6 a6f1 d6f6 f2f3 f6d4 g1h1');
INSERT INTO evaluations VALUES(15000,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(15000,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(15000,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(15000,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(15000,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(15000,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(15000,36,'cp -18','f6d5 e2b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 h7h6 c2c3 d8f6 f1e1 c8f5 b5d3 b7b6 d3e4 f5e4 e1e4 f6f5 d1e2 f5d7 c1f4 d6f4 e4f4 f8e8 e2c2 e8e7 h2h3 a8e8 a1d1 e7e2 d1d2 e2d2 c2d2 d7d5 f4g4 g8f8 g4g3 d5c4');
INSERT INTO evaluations VALUES(15343,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(15343,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(15343,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(15343,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(15343,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(15343,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(15343,36,'cp -18','e2b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 e8g8 c2c3 h7h6 e1g1 d8f6 f1e1 c8f5 b5d3 b7b6 c1e3 c6e7 a2a4 e7d5 d1c2 f5d3 c2d3 a7a5 e3d2 a8d8 g2g3 c7c5 g1g2 c5d4 c3d4 d6b4 a1c1 f6d6 e1e5 d6d7 d2b4 a5b4 d3b3');
INSERT INTO evaluations VALUES(10301,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(10301,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(10301,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10301,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(10301,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(10301,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(10301,36,'cp 29','c1e3 c5b6 d1d2 e8g8 e1c1 f8e8 f2f3 d7d5 e4d5 f6d5 e3g5 f7f6 f1c4 c6d4 c3d5 c8e6 g5f6 g7f6 d5b6 a7b6 d2d4 d8d4 c4e6 e8e6 d1d4 a8a2 c1b1 a2a5 d4d7 a5c5 h1d1 h7h5 c2c3 e6e2 d1d2 e2d2 d7d2 g8f7 d2d7 f7e6 d7h7 b6b5 b1a2 b5b4 c3b4 c5c2 a2b3 c2g2 h7c7 g2h2 c7b7 h5h4 b7h7 h4h3');
INSERT INTO evaluations VALUES(10300,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(10300,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(10300,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(10300,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(10300,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(10300,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(10300,36,'cp 27','c5b6 a2a3 d7d6 d1d2 f6g4 d4c6 b6e3 c6d8 e3d2 e1d2 e8d8 f2f3 g4f6 g2g4 c7c6 f1d3 c8e6 c3e2 d8c7 e2d4 a7a5 h2h4 h7h6 a1g1 e6d7 d3c4 h8f8 b2b3 a8e8 c4d3 e8e5 c2c4 g7g6 g1e1 f8e8 h1f1 e5e7 d4c2 e8d8 c2e3 d7e6');
INSERT INTO evaluations VALUES(9942,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(9942,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(9942,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(9942,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(9942,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(9942,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9942,36,'cp 31','a2a3 d7d6 d1d2 f6g4 d4c6 b6e3 c6d8 e3d2 e1d2 e8d8 f2f3 g4f6 g2g4 c7c6 h2h4 d8c7 f1d3 c8e6 c3e2 h7h6 a1g1 a7a5 e2d4 a8e8 b2b4 a5b4 a3b4 g7g6 g1a1 f6d7 g4g5 d7e5 f3f4 e5g4 b4b5 h6g5 h4g5 e6d7 b5c6 h8h1 a1h1 b7c6 h1h7 c6c5 d4f3 g4f2 h7f7 f2e4 d2e3');
INSERT INTO evaluations VALUES(57268,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(57268,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(57268,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(57268,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(57268,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(57268,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(57268,36,'cp -88','b8c6 d2d4 d8h4 e1f1 d7d6 g1f3 c8g4 f3h4 g4d1 b1c3 d1c2 c1f4 g8f6 a1c1 c6d4 f1f2 f6g4 f2g3 g4e5 h4f3 d4f3 c4b5 c7c6 g2f3 c6b5 c1c2 a8c8 h1c1 e8d7 c3b5 c8c2 c1c2 a7a6 b5c3 f8e7 c3d5 e7d8 f4e5 d6e5 c2d2 d8c7 g3g2 c7d6 d5e3 d7e6');
INSERT INTO evaluations VALUES(22896,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(22896,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(22896,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(22896,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(22896,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(22896,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(22896,36,'cp -91','d2d4 d8h4 e1f1 d7d6 g1f3 c8g4 f3h4 g4d1 b1c3 d1c2 c1f4 c6d4 a1c1 g8f6 h4f3 d4e6 f4e3 c2e4 c3e4 f6e4 c4e6 f7e6 c1c7 f8e7 c7b7 e7f6 f1e2 d6d5 h1c1 e8g8 c1c7 e4d6 b7a7 f8b8 a7a8 b8a8 a2a3 f6b2 c7e7 a8a3 e7e6 d6c4 e3c5 a3a5 e6e8 g8f7 e8e7 f7f6 e7c7');
INSERT INTO evaluations VALUES(22897,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(22897,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(22897,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(22897,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(22897,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(22897,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(22897,36,'cp -84','d8h4 e1f1 d7d6 g1f3 c8g4 f3h4 g4d1 b1c3 d1c2 c1f4 c6d4 a1c1 g8f6 h4f3 d4e6 f4e3 c2e4 c3e4 f6e4 c4e6 f7e6 c1c7 f8e7 c7b7 e7f6 f1e2 e8g8 h1c1 d6d5 e3d4 f6d4 f3d4 f8f2 e2e3 e6e5 d4e6 f2g2 c1c7 g2h2 c7g7 g8h8 b7a7 h2h3 e3e2 a8e8 a7e7 e8e7 g7e7 h7h5 e6d8 h3h2 e2d3 e4c5 d3c3 e5e4 d8f7 h8g7');
INSERT INTO evaluations VALUES(48466,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(48466,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48466,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(48466,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48466,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(48466,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48466,36,'cp 16','a7a6 a2a4 c7c5 g1f3 b8c6 e1g1 c5d4 e3d4 f8b4 d4d5 c6a5 c4a2 e8g8 d5e6 c8e6 a2e6 f7e6 d1e1 d8e8 c1f4 a8c8 f3g5 e6e5 f4e5 a5c4 e5f6 e8e1 f1e1 g7f6 g5e6 f8f7 e1e2 f7d7 a1d1 d7d1 c3d1');
INSERT INTO evaluations VALUES(10087,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10087,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10087,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10087,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10087,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10087,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10087,36,'cp 0','d7d6 a2a3 e8g8 d2d3 c6d4 e1g1 c7c6 c4a2 c5b6 f3d4 b6d4 c3e2 d4b6 e2g3 c8e6 d1e2 f6d7 a2e6 f7e6 e2g4 d8f6 c1e3 f6e7 e3g5 e7f7 g5e3');
INSERT INTO evaluations VALUES(9328,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9328,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9328,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9328,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9328,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9328,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9328,36,'cp 0','a2a3 e8g8 d2d3 c6d4 c1e3 c5b6 e1g1 d4f3 d1f3 c7c6 c3e2 c8e6 e2g3 h7h6 g3h5 f6h5 c4e6 f7e6 f3h5 b6e3 f2e3 d8b6 h5g4 b6e3 g1h2 a8e8 f1f8 g8f8 a1f1 f8g8 f1f6 e8e7 f6e6 e7e6 g4e6 g8h7 e6f5 h7g8 a3a4 e3d2 h3h4 d2c1 f5e6 g8h7 e6d6');
INSERT INTO evaluations VALUES(9357,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9357,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(9357,36,'cp 0','e8g8 d2d3 c6d4 c1e3 c5b6 e1g1 d4f3 d1f3 c7c6 c3e2 c8e6 e2g3 f6d7 c4e6 f7e6 f3g4 d8e7 e3g5 e7f7 g5e3');
INSERT INTO evaluations VALUES(3838,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(3838,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(3838,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3838,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3838,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3838,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3838,36,'cp 0','d2d3 a7a5 c1e3 c5e3 f2e3 c6e7 a3a4 c7c6 e1g1 e7g6 d3d4 c8e6 c4d3 e6d7 d1d2 d8e7 d3c4 d7e6 c4d3');
INSERT INTO evaluations VALUES(3840,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3840,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3840,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(3840,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3840,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(3840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3840,36,'cp 0','c6d4 e1g1 c7c6 c4a2 a7a5 c1e3 d4f3 d1f3 c5e3 f3e3 c8e6 a2e6 f7e6 a1d1 d8c7 c3e2 a8e8 f2f4 e5f4 f1f4 f6d7 f4f1 f8f1 d1f1 e8f8 f1f8 d7f8 d3d4 h7h6 e2f4 c7b6 e4e5 d6e5 e3e5 b6b2 f4e6 f8e6 e5e6 g8f8 e6f5 f8e8 f5e5 e8f7 e5f5');
INSERT INTO evaluations VALUES(3921,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3921,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3921,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3921,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3921,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3921,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3921,36,'cp 0','e1g1 c7c6 c4a2 a7a5 c1e3 d4f3 d1f3 c5e3 f3e3 c8e6 a2e6 f7e6 a1d1 d8c7 c3e2 a8e8 f2f4 e5f4 f1f4 f6d7 f4f1 a5a4 f1f8 e8f8 e3a7 d7c5 d1f1 f8f1 g1f1 c7e7 f1g1 g8f7 a7b8 e7f8 b8c7 f8e7');
INSERT INTO evaluations VALUES(19374,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(19374,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(19374,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(19374,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(19374,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(19374,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(19374,36,'cp -25','e7e6 c4b3 g8f6 b1c3 f8e7 g1f3 d7d6 e1g1 e8g8 f1e1 a7a6 a2a3 f6d7 c1e3 d7e5 f3d2 b7b5 f2f4 e5g6 a3a4 a8b8 a4b5 a6b5 d1f3 c6b4 c3a2 b4a2 a1a2 c8b7 a2a7 d8c7 e1a1 b8a8 a7a8 f8a8');
INSERT INTO evaluations VALUES(18742,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(18742,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(18742,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(18742,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(18742,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(18742,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(18742,36,'cp -24','c4b3 g8f6 c2c4 d7d5 c4d5 e6d5 b1c3 c8e6 c1g5 f8e7 g1f3 e8g8 e1g1 d5e4 d3e4 h7h6 b3e6 f7e6 g5f6 e7f6 d1b3 d8e7 c3e2 a8d8 e2f4 f8e8 a1d1 f6d4 f4e2 e7f7 e2d4 c5d4 f3e1 e6e5 b3f7 g8f7 e1d3 b7b5 a2a3 g7g5 h2h3 f7e6');
INSERT INTO evaluations VALUES(48385,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48385,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48385,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48385,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(48385,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48385,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48385,36,'cp 0','e4e5 f6d7 d2d4 c7c5 c3e2 c5d4 e2d4 b8c6 c1f4 f8e7 h2h4 d8b6 c2c3 b6b2 d4b5 a8b8 h4h5 e8g8 f1e2 d7c5 h5h6 g7g6 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(48432,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48432,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48432,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48432,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48432,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48432,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48432,36,'cp 0','f6d7 c3e2 c7c5 d2d4 c5d4 e2d4 b8c6 c1f4 f8e7 h2h4 d8b6 c2c3 b6b2 d4b5 a8b8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(49173,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49173,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49173,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49173,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49173,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49173,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49173,36,'cp 0','c3e2 c7c5 d2d4 c5d4 e2d4 b8c6 c1f4 f8e7 h2h4 d8b6 c2c3 b6b2 d4b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(54245,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(54245,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(54245,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(54245,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(54245,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(54245,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(54245,36,'cp 82','e4d5 e6d5 f1b5 b8c6 d2d4 g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 c8g4 b5c6 b7c6 f3e5 g4h5 e5d7 a8d8 d7c5 d8d5 d3c3 h6d2 c3d2 d5d2 a1c1 h5g6 f1d1 d2d1 c1d1 g6c2 d1d7 f8a8 f2f3 h7h5 g1f2 c2f5 d7c7 a7a5 c7c6 a8b8 b2b3 f5b1 a2a3 b1a2 c6c7 b8b5 a3a4');
INSERT INTO evaluations VALUES(54140,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(54140,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(54140,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(54140,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(54140,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(54140,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(54140,36,'cp 99','e6d5 f1b5 b8c6 e1g1 g8f6 d2d4 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6g6 b5d3 g6h6 d4c5 d6c5 f1e1 f8d8 c2c4 c8e6 d3e4 c6b4 a2a3 b4d5 e4d5 a7a5 d1d3 a8c8 a1d1 b7b5 d3c2 e6d5 c4d5 c5f8 c2b3 a5a4 b3d3 b5b4 a3b4 f8b4 e1e2');
INSERT INTO evaluations VALUES(54544,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(54544,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(54544,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(54544,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(54544,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(54544,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(54544,36,'cp 95','f1b5 b8c6 e1g1 g8f6 d2d4 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6g6 b5d3 g6h6 d4c5 d6c5 f1e1 f8d8 c2c4 c8e6 a2a3 b7b5 a1c1 b5c4 d3c4 h6h5 c4b3 c5f8 c1c6 e6d5 b3d5 d8d5 d1e2 d5d7 g2g3 h5d5 e1c1 a7a5 c6c8 a8c8 c1c8');
INSERT INTO evaluations VALUES(54468,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(54468,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(54468,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(54468,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(54468,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(54468,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(54468,36,'cp 97','b8c6 d2d4 g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6g6 b5d3 g6h6 d4c5 d6c5 f1e1 f8d8 d5e3 h6f4 d1e2 c8e6 c2c3 c6e5 f3e5 f4e5 d3c4 a8c8 c4e6 e5e6 g2g3 g7g6 e3g4 e6e2 e1e2 b7b5 a1e1 b5b4 c3c4 g8g7 g1g2 a7a5');
INSERT INTO evaluations VALUES(20472,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(20472,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(20472,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(20472,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(20472,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(20472,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(20472,36,'cp 88','d2d4 g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6g6 b5d3 g6h6 d4c5 d6c5 f1e1 f8d8 d5e3 g7g6 c2c3 h6f4 d1e2 d8e8 d3b5 c8g4 a1d1 c5e3 f2e3 f4c7 h2h3 g4f3 e2f3 a8d8 f3f4 c7e7 a2a3 g8g7 d1d8 e8d8 b5c6 b7c6 e3e4 a7a5 e4e5');
INSERT INTO evaluations VALUES(20458,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(20458,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(20458,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(20458,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(20458,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(20458,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(20458,36,'cp 83','g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6g6 b5d3 g6h6 d4c5 d6c5 f1e1 f8d8 d5e3 g7g6 c2c3 h6f4 d1e2 d8e8 d3b5 c8e6 a2a4 a7a6 b5c6 b7c6 a4a5 e6f5 b2b4 c5e3 f2e3 f4c7 e2c4 f5e4 f3d4 e4d5 c4c5 c7d7');
INSERT INTO evaluations VALUES(13703,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(13703,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(13703,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(13703,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(13703,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(13703,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(13703,36,'cp 100','e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6g6 b5d3 g6h6 d4c5 d6c5 f1e1 f8d8 d5e3 g7g6 c2c3 h6f4 d1e2 d8e8 d3b5 c8e6 g2g3 f4c7 e2c2 a8d8 b5f1 c6e5 f3e5 c7e5 f1g2 e5c7 e3d5 e6d5 e1e8 d8e8 g2d5 e8e5 d5e4 b7b5 a1d1 a7a6 e4g2');
INSERT INTO evaluations VALUES(13701,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(13701,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(13701,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(13701,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(13701,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(13701,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(13701,36,'cp 87','c8g4 c1g5 f8e7 d4c5 e8g8 h2h3 g4f3 d1f3 c6d4 f3d3 d4e6 g5f6 e7f6 c3d5 f6b2 a1b1 e6c5 d3f3 b2e5 f1e1 e5d6 b1d1 d8a5 b5c4 b7b5 c4b3 c5b3 a2b3 a8e8 e1e2 e8e2 f3e2 d6c5 g2g3 h7h6 h3h4 f8d8 g1g2 g8h8 e2e4 b5b4 e4f3');
INSERT INTO evaluations VALUES(54216,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54216,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54216,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(54216,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(54216,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(54216,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54216,36,'cp 21','c2c4 b8c6 c4d5 e6d5 f1e2 c5d4 f3d4 g8f6 e1g1 f8d6 b2b3 d8e7 d4b5 d6b8 c1b2 e8g8 b1d2 a7a6 b5d4 c6d4 b2d4 c8f5 d2f3 f5e4 a1c1 b8d6 e2d3 f8c8 d1e2 e7e6 c1c8 a8c8 d4b2 e6f5 f1d1 f5g6 g2g3 d6b4 d3e4 d5e4');
INSERT INTO evaluations VALUES(54178,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54178,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(54178,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(54178,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54178,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(54178,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(54178,36,'cp 13','b8c6 c4d5 e6d5 b1c3 g8f6 f1b5 c5d4 e3d4 f8d6 e1g1 e8g8 c1g5 c8g4 b5e2 d6e7 h2h3 g4h5 f1e1 h7h6 g5h4 f8e8 f3e5 h5e2 e1e2 f6e4 h4e7 e8e7 e5c6 b7c6 a1c1 e7e6 d1c2 d8e7 c3e4 d5e4 c2a4 a8d8 c1c6 e6c6 a4c6');
INSERT INTO evaluations VALUES(20255,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(20255,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(20255,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(20255,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20255,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(20255,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(20255,36,'cp 16','c4d5 e6d5 f1b5 c5d4 f3d4 c8d7 e1g1 f8d6 b2b3 g8f6 c1a3 d6a3 b1a3 a8c8 b5e2 e8g8 a3b5 h7h6 a1c1 d8a5 c1c2 c6b4 c2c8 f8c8 a2a3 b4c2 d4c2 d7b5 e2b5 a5b5 c2d4 b5c5 b3b4 c5c4 d4f5 c8c7 d1d4 c4d4 f5d4 c7c3');
INSERT INTO evaluations VALUES(20231,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(20231,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(20231,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(20231,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(20231,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(20231,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(20231,36,'cp 8','e6d5 f1e2 g8f6 e1g1 c5d4 f3d4 f8d6 b2b3 d8e7 c1b2 e7e5 g2g3 e8g8 d1d2 e5e7 f1c1 c8d7 b1c3 f8e8 e2f3 f6e4 c3e4 d5e4 d4c6 d7c6 f3e2 a8d8 d2c3 d6e5 c3e5 e7e5 b2e5 e8e5 c1d1 e5d5 d1d5');
INSERT INTO evaluations VALUES(20521,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(20521,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(20521,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(20521,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(20521,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(20521,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(20521,36,'cp 17','f1e2 c5d4 f3d4 f8d6 e1g1 g8f6 b2b3 d8e7 d4b5 d6c5 c1b2 e8g8 b2f6 e7f6 b1c3 f8d8 a1c1 f6e5 e2f3 c8e6 c3a4 c5e7 a4c5 e7c5 c1c5 a8c8 d1d2 b7b6 c5c1 g7g6 f1d1 a7a6 b5c3 d5d4 c3a4 e5b5 f3c6 c8c6 e3d4 c6c1 d1c1');
INSERT INTO evaluations VALUES(52973,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(52973,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(52973,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(52973,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(52973,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(52973,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(52973,36,'cp 79','g8f6 f1d3 c8g4 h2h3 g4f3 d1f3 g7g6 f3e2 b8d7 c2c4 a7a6 e1g1 f6h5 g2g3 e7e5 d4d5 f8e7 b1d2 e8g8 b2b4 a6a5 b4b5 c6b5 c4b5 d7c5 d3c2 e7g5 a1b1 g5h6 g1g2 a8c8 a2a4 h6d2 c1d2');
INSERT INTO evaluations VALUES(45105,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(45105,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(45105,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(45105,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(45105,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(45105,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(45105,36,'cp 64','f1d3 c8g4 h2h3 g4f3 d1f3 b8d7 f3e2 g7g6 c2c4 e7e5 d4d5 f8e7 g2g3 a7a6 e1g1 f6h5 b1d2 e7g5 b2b4 c6c5 a2a3 e8g8 g1g2 g5h6 a1b1 b7b6 f1e1 h5g7 c1b2 f7f5 d2f3 d8e7 b4b5 a6b5 c4b5 f5f4 g3g4');
INSERT INTO evaluations VALUES(45103,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(45103,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(45103,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(45103,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(45103,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(45103,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(45103,36,'cp 85','e7e5 e1g1 c8g4 d4e5 d6e5 h2h3 g4f3 d1f3 b8d7 b1d2 f8e7 a2a4 e8g8 d2c4 a7a5 b2b3 e7c5 f1d1 f8e8 c1g5 h7h6 g5h4 c5d4 a1c1 d8c7 d3e2 f6h7 c4e3 d7c5 e3f5 h7g5 h4g5 h6g5 f5d4 e5d4 d1d4 c5e6 d4d1 c7e5');
INSERT INTO evaluations VALUES(55608,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(55608,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(55608,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(55608,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(55608,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(55608,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(55608,36,'cp 29','f2f4 g7g6 g1f3 f8g7 f4e5 d6e5 f1e2 c7c5 e1g1 b8c6 c2c3 b7b6 b1d2 g8e7 d1e1 e8g8 d2c4 d8c7 c1e3 c8a6 a2a4 a6c4 d3c4 e7c8 e1h4 c8d6 f3g5 h7h6 g5f3 c7e7 h4e7 c6e7 f3d2 f7f5 a4a5 d6e4 d2e4 f5e4 a1d1 f8f1 e2f1');
INSERT INTO evaluations VALUES(33757,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(33757,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(33757,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(33757,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(33757,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(33757,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(33757,36,'cp 0','b8c6 d2d4 c8g4 f1e2 e8c8 e1g1 c6d4 f3d4 g4e2 d1e2 d5d4 c1g5 d4c5 h2h4 h7h6 g5e3 c5a5 e2f3 a5d5 f3g4 e7e6 f1d1 d5e5 d1d8 c8d8 e3d4 e5f5 g4f5 e6f5 a1d1 d8c8 d1e1 c8d7 e1d1 d7c6 d1e1');
INSERT INTO evaluations VALUES(901,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(901,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(901,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(901,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(901,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(901,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(901,36,'cp -18','d2d4 c8g4 f1e2 e8c8 e1g1 g4f3 e2f3 d5d4 d1e2 c6e5 f3e4 e7e6 c1g5 f7f6 a1d1 d4b6 d1d8 c8d8 g5e3 f8c5 e3c5 b6c5 e4b7 d8e7 b7e4 h8d8 c2c3 f6f5 e4c2 h7h6 h2h3 d8d6 f1d1 d6d1 e2d1 c5c4 c2b3 c4e4 b3c2 e4f4 g2g3 f4c4 c2b3 c4d3 b3c2 d3d1 c2d1');
INSERT INTO evaluations VALUES(895,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(895,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(895,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(895,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(895,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(895,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(895,36,'cp -22','c8g4 f1e2 e8c8 e1g1 c6d4 f3d4 g4e2 d1e2 d5d4 c1g5 d4c5 h2h4 h7h6 g5e3 c5c6 f1d1 e7e6 e3a7 f8e7 a7d4 d8d5 g2g3 h8g8 d4e3 g7g5 d1d5 c6d5 h4h5 d5e4 a1d1 e7c5 e2d3 e4d3 d1d3 c5e3 d3e3 g8d8 g1f1 d8d2 e3e2 d2d4 c2c3 d4a4 a2a3 c8d7 f1e1 b7b5 e1d2 d7e7 d2e3');
INSERT INTO evaluations VALUES(27423,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(27423,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(27423,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(27423,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(27423,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(27423,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(27423,36,'cp -13','f1e2 e8c8 e1g1 c6d4 f3d4 g4e2 d1e2 d5d4 c1g5 d4c5 h2h4 h7h6 g5e3 c5c6 a1d1 e7e6 e3a7 f8e7 a7d4 d8d5 d4c3 d5c5 g2g3 h8g8 f1e1 e7d6 e2e4 c6e4 e1e4 c8d7 a2a4 f7f6 d1d3 h6h5 g1f1 b7b6 b2b4 c5d5 d3d5 e6d5 e4d4 c7c6 b4b5 d6c5 b5c6 d7c6');
INSERT INTO evaluations VALUES(27422,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(27422,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(27422,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(27422,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(27422,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(27422,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(27422,36,'cp -10','e8c8 e1g1 c6d4 f3d4 g4e2 d1e2 d5d4 c1g5 d4c5 h2h4 h7h6 g5e3 c5c6 a1d1 e7e6 e3a7 f8e7 a7d4 d8d5 d4c3 d5c5 g2g3 h8g8 f1e1 g7g5 h4g5 h6g5 a2a4 e7d6 e2e4 c6e4 e1e4 e6e5 g3g4 f7f6 g1g2 g8h8 d1h1 h8h1 g2h1 c5c6 h1g2 c8d7');
INSERT INTO evaluations VALUES(19,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(19,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(19,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(19,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(19,36,'cp -5','e1g1 c6d4 f3d4 g4e2 d1e2 d5d4 c1g5 d4c5 h2h4 h7h6 g5e3 c5c6 a1d1 e7e6 e3a7 f8e7 a7d4 d8d5 d4c3 d5c5 g2g3 h8g8 f1e1 g7g5 h4g5 h6g5 e2e4 g8d8 d1d8 c8d8 e4c6 c5c6 a2a4 c6c4 a4a5 d8d7 g1f1 e7b4 e1d1 d7e7 c3b4 c4b4 b2b3 f7f5 d1a1 b4e4 a5a6 b7a6 a1a6 g5g4 a6a7');
INSERT INTO evaluations VALUES(6216,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(6216,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(6216,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(6216,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(6216,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(6216,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(6216,36,'cp 180','b2b4 c5d6 e1g1 a7a6 d2d4 g8f6 f1e1 e8g8 a2a3 b7b5 c4d3 f8e8 c1b2 a8b8 b1d2 a6a5 d4d5 c6e7 c3c4 a5b4 c4b5 e7g6 a3a4 d6c5 h2h3 d7d6 d2b3 c5b6 b2c1 c8d7 d3f1 f6h5 c1e3 g6f4 b3d2 d8f6 a4a5 b6e3 f2e3 f4g6 a5a6 h5g3 a6a7 b8a8');
INSERT INTO evaluations VALUES(6203,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(6203,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(6203,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(6203,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(6203,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(6203,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(6203,36,'cp 171','c5d6 e1g1 g8f6 d2d4 a7a6 f1e1 e8g8 c1b2 b7b5 c4d3 f8e8 a2a3 a8b8 b1d2 a6a5 b4a5 d6f8 f3e5 c6a5 e5f3 d7d6 a3a4 b5a4 a1a4 c7c5 a4a2 c8g4 h2h3 g4h5 d4d5 c5c4 d3c2 h5g6 f3d4 a5b3 d2b3 c4b3');
INSERT INTO evaluations VALUES(47767,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47767,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47767,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(47767,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(47767,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(47767,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(47767,36,'cp 28','f1d3 b8d7 c1e3 c7c6 c2c4 d7c5 e3c5 d6c5 g1f3 f8d6 b1c3 e8g8 a2a4 c8d7 a4a5 d7g4 a5a6 c6d5 c4d5 c5c4 d3c4 a8c8 c4e2 d8b6 e1g1 g4f3 e2f3 b7a6 d1a4 d6c5 a4a6 b6b2 c3b5 b2b4 b5a7 c8a8 a6b5 b4b5 a7b5 a8a1 f1a1');
INSERT INTO evaluations VALUES(50193,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50193,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50193,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50193,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(50193,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50193,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(50193,36,'cp 0','d7d5 e2e3 f8g7 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(46766,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46766,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46766,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(46766,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46766,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46766,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46766,36,'cp 8','g1f3 f8g7 e2e3 e8g8 f1e2 c7c5 e1g1 c5d4 e3d4 b8c6 f3e5 f6d7 c3d5 d7e5 d4e5 c8e6 c2c4 g7e5 f4e5 c6e5 d1d4 e5c6 d4c5 e6d5 c4d5 c6e5 f1c1 e5d7 c5b4 d7b6 e2f3 a7a5 b4e4 d8d6 h2h4 d6b4 c1c7 b4e4 f3e4');
INSERT INTO evaluations VALUES(46765,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46765,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46765,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46765,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46765,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46765,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46765,36,'cp 0','f8g7 e2e3 e8g8 f1e2 c7c5 e1g1 b7b6 c3b5 b8c6 b5c7 a8b8 c7b5');
INSERT INTO evaluations VALUES(14302,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(14302,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(14302,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(14302,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(14302,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(14302,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(14302,36,'cp 30','c8f5 e1g1 e7e6 c2c4 f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 c6e5 d4e5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c7c6 d5c4 a8d8 c1e3 b7b5 c4d3 f5d3 e2d3 d8d3 f1d1 d3d1 a1d1 f8d8 d1c1 d8c8 g1g2 f7f5 e5f6 e7f6 b2b3 g8f7 g2f3 f7e6 g3g4 f6e7 f3e4 c6c5 h2h4 e7h4 c1c5 c8c5 e3c5 g7g6 c5e3 h4e7');
INSERT INTO evaluations VALUES(24865,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24865,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(24865,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(24865,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(24865,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(24865,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24865,36,'cp 49','c2c4 e7e6 e1g1 f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 e6d5 c1g5 c6e5 d4e5 f6e4 g5e7 d8e7 d1d5 e4c3 b2c3 f8d8 d5b7 a8b8 b7a7 e7e5 c3c4 f5e6 c4c5 b8b5 a1d1 d8b8 c5c6 g7g6 a2a4 b5b6 e2e3 g8g7 h2h4 h7h6 d1d4 b6b1 f1b1');
INSERT INTO evaluations VALUES(42568,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(42568,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(42568,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42568,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42568,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42568,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42568,36,'cp 0','g1e2 d7d6 b1c3 c7c6 d2d3 b7b5 c4b3 c8e6 b3e6 f7e6 f3h3 d8e7 f2f4 e8g8 f4e5 d6e5 e2g1 b8d7 g1f3 a7a5 c3d1 a5a4 a2a3 h7h6 c1e3 c5e3 d1e3 c6c5 f3h4 g8h7 h4f3 h7g8');
INSERT INTO evaluations VALUES(42569,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(42569,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(42569,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42569,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42569,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42569,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42569,36,'cp 0','d7d6 b1c3 c7c6 d2d3 b7b5 c4b3 c8e6 b3e6 f7e6 f3h3 d8e7 f2f4 e8g8 f4e5 d6e5 e2g1 b8d7 g1f3 a7a5 c1d2 c5b6 a2a3 b6d4 c3d1 h7h6 h1f1 f8f7 h3g3 g8h7 g3h3');
INSERT INTO evaluations VALUES(48245,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48245,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48245,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48245,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(48245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48245,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48245,36,'cp 0','c7c5 g1f3 c5d4 f3d4 b8d7 d1a4 f8c5 d4b3 e8g8 b3c5 d7c5 a4a3 b7b6 c4d5 f6d5 c3d5 e6d5 e2e3 d5d4 a1d1 c8b7 f1c4 d8f6 e1g1 d4e3 f4e3 b7g2 g1g2 f6g6 g2h3 g6f5 h3g3 c5e4 g3g2 f5g4 g2h1 g4f3 h1g1 a8d8 d1d8 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(46167,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46167,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46167,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46167,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46167,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46167,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46167,36,'cp 0','g1f3 c5d4 f3d4 b8d7 d1a4 f8c5 d4b3 e8g8 b3c5 d7c5 a4a3 b7b6 c4d5 f6d5 c3d5 e6d5 e2e3 d5d4 a1d1 c8b7 f1c4 d8f6 e1g1 d4e3 f4e3 b7g2 g1g2 f6g6 g2h3 g6h5 h3g3 h5g6');
INSERT INTO evaluations VALUES(46166,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46166,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46166,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46166,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46166,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46166,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46166,36,'cp 0','c5d4 f3d4 b8d7 d1a4 f8c5 d4b3 e8g8 b3c5 d7c5 a4a3 b7b6 c4d5 f6d5 c3d5 e6d5 e2e3 c8f5 a1d1 d8f6 f1e2 c5e6 f4g3 a8c8 g3d6 f8d8 e1g1 e6c5 d6f4 h7h6 f1e1 c5e6 f4d6 e6c5');
INSERT INTO evaluations VALUES(56874,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(56874,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(56874,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(56874,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(56874,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(56874,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(56874,36,'cp 138','h2h4 d7d5 b1c3 d5e4 c3e4 b8c6 g1f3 g8f6 c1g5 f8e7 g5f6 e7f6 f1b5 c8d7 c2c3 h7h5 d1e2 e8g8 e1c1 f6g7 e4c5 a7a6 c5b7 c6d4 e2e4 f7f5 e4d4 g7d4 b7d8 d7b5 d1d4 a8d8 h1d1 d8d4');
INSERT INTO evaluations VALUES(57085,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(57085,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(57085,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(57085,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(57085,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(57085,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(57085,36,'cp -25','c7c6 b5e2 d7d5 g1f3 d5e4 f3e5 d8g5 e5g4 f7f5 d2d4 g5g6 g4e3 g8f6 g2g3 c8e6 e3g2 g6f7 g2f4 e6c4 e2c4 f7c4 c2c3 b8d7 b1d2 c4f7 d1b3 d7b6 a2a4 f8d6 f4g2 h7h6 a4a5 b6d5 d2c4 d6b8 c4e3 f5f4 e3d5 f6d5 c1f4 e8g8 e1c1 b8f4 g2f4 d5f4 b3f7 f8f7 g3f4');
INSERT INTO evaluations VALUES(51821,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(51821,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(51821,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(51821,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(51821,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(51821,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(51821,36,'cp -21','b5e2 g8f6 g1f3 f6e4 f3e5 d7d5 b1c3 e4c3 d2c3 f8d6 e5f3 e8g8 e1g1 c8e6 f3g5 e6f5 e2d3 d8f6 d1f3 f5d3 f3d3 f6g6 d3g6 h7g6 c1e3 f7f6 g5f3 f8e8 f1d1 d6c7 g2g3 g6g5 g1g2 g5g4 f3h4 g8f7 c3c4 d5c4 h4f5 c7e5 f5d6 e5d6 d1d6');
INSERT INTO evaluations VALUES(16547,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(16547,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(16547,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16547,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(16547,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(16547,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(16547,36,'cp -20','d7d5 b1d2 f8d6 e1g1 e8g8 f1e1 h7h6 c2c3 a7a5 e4d5 f6d5 d2c4 f8e8 h2h3 a5a4 c4d6 d8d6 e2f1 c8f5 c1d2 d6f6 b2b4 e5e4 d3e4 f5e4 b4b5 e4f3 e1e8 a8e8 d1f3 f6f3 g2f3 c6e5 g1g2 e5g6 g2g3 f7f5 h3h4 f5f4 g3h3 g6e5 f1e2');
INSERT INTO evaluations VALUES(15156,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(15156,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(15156,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(15156,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(15156,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(15156,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(15156,36,'cp -24','e4d5 f6d5 e1g1 f8d6 a2a3 a7a5 b1c3 d5c3 b2c3 e8g8 f1e1 c8f5 a1b1 a8b8 b1b5 h7h6 f3d2 b7b6 d2c4 f5e6 c4d6 d8d6 e2f3 f7f6 d1e2 e6d7 b5b1 g8h8 e2e4 f8e8 e4d5 d6d5 f3d5 b8d8 d5e4 c6e7');
INSERT INTO evaluations VALUES(15012,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(15012,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(15012,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(15012,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(15012,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(15012,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(15012,36,'cp -30','f6d5 e1g1 f8d6 f1e1 e8g8 h2h3 f8e8 e2f1 c8f5 c2c3 h7h6 b1d2 a7a5 d2c4 d8f6 c4d6 f6d6 c1d2 a8d8 d1b3 d5b6 d2e3 a5a4 b3d1 a4a3 b2a3 d6a3 e3b6 c7b6 d1c2 b6b5 a1b1 a3c5 c2b3 f5d3 f1d3 d8d3 b3b5 d3c3 b5c5 c3c5 b1b7 e5e4 b7b2 f7f5 a2a3');
INSERT INTO evaluations VALUES(15357,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(15357,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(15357,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(15357,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(15357,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(15357,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(15357,36,'cp -28','e1g1 f8d6 f1e1 e8g8 b1d2 f8e8 c2c3 h7h6 e2f1 c8f5 h2h3 a7a5 d2c4 d8f6 c4d6 f6d6 a2a3 f5g6 f3h4 g6h7 h4f3 a8d8 d1b3 b7b6 b3a4 h7f5 c1d2 d6f6 c3c4 d5f4 d2f4 e5f4 c4c5 e8e1 a1e1 b6c5 a4f4 f5d3');
INSERT INTO evaluations VALUES(15356,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(15356,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(15356,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(15356,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(15356,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(15356,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(15356,36,'cp -28','f8d6 f1e1 e8g8 c2c3 h7h6 b1d2 f8e8 e2f1 a7a5 h2h3 c8f5 d2c4 d8f6 c4d6 f6d6 c1d2 a8d8 a2a3 d5f6 d3d4 e5d4 e1e8 d8e8 f3d4 c6d4 c3d4 d6d4 d2c3 d4d1 a1d1 e8a8 f2f3 c7c6 d1d6 f6e8 d6d2 f5e6 g1f2 e8c7 f1d3');
INSERT INTO evaluations VALUES(9212,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(9212,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(9212,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(9212,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(9212,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(9212,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(9212,36,'cp -21','f1e1 e8g8 b1d2 h7h6 c2c3 f8e8 e2f1 a7a5 h2h3 c8f5 d2c4 d8f6 c4d6 f6d6 a2a3 a8d8 d3d4 e5e4 f3h2 c6e7 h2g4 b7b6 g4e5 c7c5 d1h5 d6e6 f1b5 e8f8 h5e2 f7f6 e5g4 c5d4 c3d4 f5g4 h3g4 f6f5 c1d2 e7g6 a1c1 g6f4 d2f4 d5f4 e2e3');
INSERT INTO evaluations VALUES(19364,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(19364,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(19364,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(19364,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(19364,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(19364,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(19364,36,'cp -83','d8c7 d2d4 c5d4 c1f4 g8f6 c2c3 f6d5 f4g3 d4c3 b1c3 d5c3 b2c3 e7e6 a1c1 f8e7 f1d3 d7d6 e5d6 e7d6 e1g1 d6g3 h2g3 c8d7 d1e2 e8g8 f1d1 a8d8 e2e4 g7g6 e4e3 g8g7 d3b5 e6e5 b5c6 d7c6 d1d8 f8d8 e3e5 c7e5 f3e5 d8d2 e5c6 b7c6 c1b1 d2a2 b1b7 g7f6 b7c7 a2a1 g1h2');
INSERT INTO evaluations VALUES(2131,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(2131,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(2131,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(2131,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(2131,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(2131,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(2131,36,'cp -90','d2d4 c5d4 c1f4 g8f6 c2c3 f6d5 f4g3 d4c3 b1c3 d5c3 b2c3 e7e6 f1d3 d7d6 e1g1 c8d7 f1e1 d6e5 f3e5 f8d6 d1g4 c6e5 g4g7 e8c8 g3e5 h8g8 e5d6 c7d6 g7d4 d6d4 c3d4 d7c6 g2g3 g8h8 d3e4 d8d4 e4c6 b7c6 e1e5 d4d5 e5e4 c8c7 e4a4 a7a5 a1c1 h8b8 c1c2');
INSERT INTO evaluations VALUES(2130,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(2130,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(2130,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(2130,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(2130,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(2130,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(2130,36,'cp -84','c5d4 c1f4 g8f6 c2c3 f6d5 f4g3 d4c3 b1c3 d5c3 b2c3 e7e6 a1c1 a7a6 f1d3 d7d6 e5d6 f8d6 g3d6 c7d6 e1g1 c8d7 c1b1 d6c7 d1b3 a8b8 b3c4 h7h6 c4c5 c7a5 c5a5 c6a5 d3a6 e8e7 a6b5 h8c8 f3d2 b7b6 b5d7 e7d7 f1d1 d7e8 d2e4 a5c4 e4d6 c4d6 d1d6');
INSERT INTO evaluations VALUES(2135,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(2135,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(2135,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(2135,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(2135,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(2135,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(2135,36,'cp -87','c1f4 g8f6 h2h3 f6d5 f4g3 c7b6 d1c1 d7d6 e5d6 e7e5 c2c4 d4c3 b1c3 d5c3 b2c3 f8d6 c1d2 d6a3 f1c4 b6b2 e1g1 b2d2 f3d2 e8g8 d2f3 e5e4 f3g5 c8f5 a1e1 a8c8 c4d5 h7h6 g5e4 f8d8 c3c4 c6e7 d5b7 c8c4 g3e5 e7g6 e5g3');
INSERT INTO evaluations VALUES(34032,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(34032,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(34032,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(34032,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(34032,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(34032,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(34032,36,'cp 178','e5e7 f8e7 e4e5 f6d5 b1c3 d5c3 d2c3 b8c6 g1f3 d7d5 c1f4 c8g4 f1e2 c6d8 f4e3 d8e6 h2h3 g4f3 e2f3 c7c6 f3e2 e7g5 e3g5 e6g5 e1c1 g5e4 h1f1 b7b5 f2f4 a7a5 d1e1 e8e7 e2d3 h8d8 d3e4 d5e4 e1e4');
INSERT INTO evaluations VALUES(33995,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(33995,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(33995,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(33995,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(33995,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(33995,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(33995,36,'cp 191','f8e7 b1c3 b8c6 f1c4 c6d4 c4b3 a7a5 g1f3 d4b3 a2b3 b7b6 c3b5 e7d8 d2d3 e8g8 c2c4 c7c6 b5c3 d7d5 e4d5 c6d5 c1e3 c8b7 e1d2 f6d7 c3d5 b7d5 c4d5 d8e7 h1c1 e7b4 d2e2 d7c5 f3d4 f8e8 c1c4 a8d8 d4c6 d8d5 c6b4 a5b4 c4b4 c5d3 b4b6');
INSERT INTO evaluations VALUES(33474,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(33474,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(33474,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(33474,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(33474,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(33474,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(33474,36,'cp 182','e4e5 f6d5 b1c3 d5c3 d2c3 b8c6 g1f3 d7d5 c1f4 e7c5 f4g3 c8g4 e1c1 c6e7 d1d2 c7c6 f3d4 e7f5 g3f4 f5d4 c3d4 c5e7 f1d3 h7h5 f2f3 g4e6 h2h4 b7b5 f4g5 a7a5 g2g3 g7g6 d2f2 b5b4 c1d2 a5a4 g5e7 e8e7 a2a3 b4a3 b2a3 h8b8 h1e1');
INSERT INTO evaluations VALUES(33470,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(33470,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(33470,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(33470,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(33470,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(33470,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(33470,36,'cp 173','f6d5 b1c3 d5c3 d2c3 b8c6 g1f3 d7d5 c1f4 c8f5 e1c1 e8c8 h2h4 f5g4 d1e1 h7h6 f1e2 d8e8 f4g3 f7f6 e2d1 h8f8 e5f6 e7f6 e1e8 f8e8 h1e1 e8e1 f3e1 g4e6 e1d3 b7b6 h4h5 c6a5 d1e2 c7c5 d3f4 e6f7');
INSERT INTO evaluations VALUES(33477,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(33477,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(33477,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(33477,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(33477,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(33477,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(33477,36,'cp 169','b1c3 d5c3 d2c3 b8c6 g1f3 d7d5 c1f4 c8g4 f1e2 h7h6 h2h3 g4f5 e2d3 f5e6 f4g3 g7g5 h3h4 e8c8 f3d4 e7c5 d4e6 f7e6 a1d1 d8f8 c3c4 c6b4 c4d5 b4d5 h4g5 h6g5 e1e2 h8h1 d1h1 d5f4 g3f4 f8f4 h1h8 c8d7 f2f3');
INSERT INTO evaluations VALUES(28041,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(28041,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(28041,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(28041,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(28041,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(28041,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(28041,36,'cp 126','h2h4 h7h6 c1f4 a7a6 d1e2 c7c5 d4d5 b6b5 e4e5 e7e6 a1d1 g8e7 d5e6 f7e6 c3e4 b7e4 e2e4 b8c6 c2c3 d8a5 f1d3 a5a2 e4e2 b5b4 d3b1 a2b3 c3c4 e8c8 e1g1 g6g5 h4g5 h6g5 f4g5 d8g8 g5f6 g7f6 e5f6 e7f5 b1e4');
INSERT INTO evaluations VALUES(16056,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(16056,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(16056,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(16056,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(16056,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(16056,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(16056,36,'cp -88','c6a5 c4b5 f8g7 d2d3 c7c6 b5a4 d7d5 c1g5 e8g8 a4c2 d5e4 d3e4 h7h6 g5f6 g7f6 b1d2 b7b5 f3e2 d8b6 g1f3 a5b7 e1g1 a7a5 b2b4 a5b4 c3b4 g8g7 a2a4 b5a4 c2a4 b6b4 e2c4 c6c5 f1b1 b4a5 c4d5 a5c7');
INSERT INTO evaluations VALUES(16088,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(16088,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(16088,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(16088,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(16088,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(16088,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(16088,36,'cp -94','c4b5 f8g7 d2d3 c7c6 b5a4 d7d5 c1g5 d5e4 d3e4 e8g8 a4c2 h7h6 g5f6 g7f6 b1d2 b7b5 f3e2 a5b7 g1f3 a7a5 e1g1 d8b6 b2b4 g8g7 a2a3 b7d8 f1c1 d8e6 g2g3 a5b4 a3b4 a8a1 c1a1 c6c5 h2h4 c5b4 c3b4 e6d4 f3d4 b6d4');
INSERT INTO evaluations VALUES(50338,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(50338,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(50338,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(50338,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(50338,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(50338,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(50338,36,'cp -66','d7d5 b1c3 c7c5 e2e4 b8c6 e4d5 f6d5 d4c5 d5c3 d1d8 c6d8 b2c3 c8d7 f3f4 a8c8 g1f3 e7e6 f1d3 f7f6 c1e3 f8c5 e3d4 b7b6 a2a4 c5d6 g2g3 d8c6 e1g1 e8e7 f1e1 c6d4 f3d4 d6c5 g1g2 h8d8 d3h7 c5d4 c3d4 d7c6 g2f2 d8d4');
INSERT INTO evaluations VALUES(47205,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(47205,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(47205,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(47205,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(47205,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(47205,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(47205,36,'cp -60','b1c3 c7c5 e2e4 c5d4 d1d4 b8c6 f1b5 c8d7 b5c6 d7c6 e4d5 f6d5 g1e2 f7f6 c1d2 d5c3 d4c3 d8d5 d2e3 e7e5 e1g1 d5b5 c3b3 b5b3 a2b3 a7a6 e3b6 e8f7 c2c4 h7h5 e2c3 g7g5 a1d1 a8c8 h2h3 f8b4 c3d5');
INSERT INTO evaluations VALUES(37860,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(37860,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(37860,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(37860,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(37860,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(37860,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(37860,36,'cp 38','b8c6 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8e8 a1d1 b5c4 d4c5 a8d8 e1e7 e8e7 c5e7 d8d5 d1d5 c4d5 e7f6 h6f8 f2f3 f8d6 g1f2 d6c7 b3c1 g8f8 b2b3 f8e8 c1e2 e8d7 f6d4 a7a6 g2g3 h7h5');
INSERT INTO evaluations VALUES(47357,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(47357,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(47357,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(47357,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(47357,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(47357,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(47357,36,'cp 47','e4e5 f6d5 f1c4 b8c6 g1e2 c8f5 b1c3 d5b6 c4b5 e7e6 c1e3 f8e7 e2g3 f5g6 h2h4 e7h4 g3h5 e8g8 h5f4 h4g5 f4g6 f7g6 d1g4 f8f5 g4h3 h7h6 e1g1 c6b4 f2f4 g5e7 a2a3 b4c2 a1d1 c2e3 h3e3 b6d5 c3d5 d8d5 b5d3 g8h7 d3f5 g6f5');
INSERT INTO evaluations VALUES(47304,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(47304,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(47304,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(47304,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(47304,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47304,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(47304,36,'cp 26','f6d5 f1c4 b8c6 g1e2 c8f5 b1c3 d5b6 c4b5 e7e6 c1e3 f8e7 e2g3 f5g6 h2h4 h7h5 g3e2 g6f5 a1c1 a7a6 b5c6 b7c6 e2f4 g7g6 g2g3 d8d7 e1g1 a8b8 d1e2 b6d5 c3d5 c6d5 e2a6 b8b2 a6a7 e7d8 c1c3 e8g8 f1c1 b2b4 a7c5 b4b2 c5a3 b2b6');
INSERT INTO evaluations VALUES(47471,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(47471,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(47471,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(47471,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(47471,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(47471,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(47471,36,'cp 46','f1c4 b8c6 g1e2 c8f5 b1c3 d5b6 c4b5 e7e6 c1e3 f8e7 e2g3 f5g6 h2h4 h7h5 g3e2 g6f5 a1c1 a7a6 b5c6 b7c6 e2f4 g7g6 g2g3 d8d7 e1g1 a8b8 b2b3 e7a3 d1d2 e8g8 c1d1 a3b4 f4d3 f5d3 d2d3 b4c3 d3c3 b6d5 c3d2 d5e3 d2e3 b8b4 e3c3 f8b8 c3d2 a6a5');
INSERT INTO evaluations VALUES(47470,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(47470,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(47470,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(47470,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(47470,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(47470,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47470,36,'cp 28','b8c6 g1e2 c8f5 b1c3 d5b6 c4b5 e7e6 c1e3 f8b4 a2a3 b4c3 e2c3 a7a6 b5c6 b7c6 e1g1 e8g8 d1e2 h7h6 a1c1 d8d7 h2h4 f8b8 h4h5 b6d5 c3a4 c6c5 a4c5 d7b5 e2d2 f5g4 f2f3 g4h5 b2b3 d5e3 d2e3 h5g6');
INSERT INTO evaluations VALUES(21379,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(21379,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(21379,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(21379,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(21379,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(21379,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21379,36,'cp 62','g8f6 a2a3 f6e4 d1c2 e4c3 c2c3 a7a5 c1f4 d5c4 c3c4 f8d6 f4g3 h7h5 e2e3 h5h4 g3e5 f7f6 e5d6 c7d6 h2h3 d6d5 c4c2 c8d7 f1d3 a8c8 c2e2 d8b6 e1g1 e8f7 a1c1 a5a4 c1c5 c6e7 f1c1 b6d6 c5c8 d7c8 d3b5 c8d7 b5d7 d6d7');
INSERT INTO evaluations VALUES(47321,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(47321,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(47321,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(47321,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(47321,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(47321,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(47321,36,'cp -53','f1c4 e7e6 b1c3 b7b6 c3e4 f6e4 g5e4 c8b7 d2d3 b8c6 e1g1 d8e7 c1g5 f7f6 d1h5 e7f7 h5f7 e8f7 g5d2 a8d8 d2c3 c6e7 f1e1 b7d5 a2a4 a7a5 c4b5 h7h5 b2b4 a5b4 c3b4 e7f5 b4f8 h8f8 c2c3 h5h4 a4a5 b6a5 a1a5 h4h3 g2g3 c7c6 b5c4 d5c4 d3c4 f5d6 e4d6 d8d6');
INSERT INTO evaluations VALUES(47318,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(47318,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(47318,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(47318,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(47318,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(47318,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(47318,36,'cp -48','e7e6 b1c3 b8c6 e1g1 f8e7 c3e4 f6e4 g5e4 e8g8 e4g3 f7f5 f1e1 g8h8 d1h5 c6d4 h5d1 f5f4 g3e4 c7c5 c2c3 d4c6 d2d3 c6e5 f2f3 b7b6 e4f2 e7d6 d1e2 e5c4 d3c4 e6e5 b2b3 c8b7 e1d1 d8c7');
INSERT INTO evaluations VALUES(48456,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(48456,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(48456,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(48456,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(48456,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(48456,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(48456,36,'cp -61','b1c3 b8c6 e1g1 f8e7 g5e4 f6e4 c3e4 f7f5 e4g3 e8g8 f1e1 g8h8 c4b5 c6d4 b5c4 e6e5 c2c3 d4c6 d2d3 a7a5 c4b5 f5f4 g3e4 c8f5 b5c6 b7c6 d3d4 a5a4 d1e2 e5d4 c1f4 d4d3 e2e3 d8d5 a1d1 a8d8 e4g5');
INSERT INTO evaluations VALUES(48455,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(48455,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(48455,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(48455,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(48455,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(48455,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48455,36,'cp -51','b8c6 e1g1 f8e7 g5e4 f6e4 c3e4 e8g8 e4g3 f7f5 f1e1 g8h8 c2c3 e6e5 a2a4 f5f4 g3e4 f4f3 g2f3 b7b6 d2d3 c6a5 c4a2 c7c5 c1e3 d8d7 g1h1 c8a6 e1g1 a8d8 d1f1 d7f5 f1g2 f5f3 g2f3 f8f3 e4g5 a6b7 g5f3 b7f3 g1g2 d8d3 h1g1 f3g2 g1g2');
INSERT INTO evaluations VALUES(16366,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(16366,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(16366,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(16366,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(16366,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(16366,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(16366,36,'cp -30','h7h6 g5f6 d8f6 f1e2 g7g6 b1c3 c6d4 c3d5 f6d6 e1g1 f8g7 d5e3 d4e2 d1e2 e8g8 f1e1 d6e7 e2d2 d7d6 d3d4 e5d4 f3d4 c8d7 f2f3 f8e8 c2c3 a7a5 e3d5 e7d8 a2a4 c7c6 d5e3 d8g5 a1d1 h6h5 d4c2 a8d8 d2f2');
INSERT INTO evaluations VALUES(16114,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(16114,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(16114,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(16114,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(16114,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(16114,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16114,36,'cp -36','g5f6 d8f6 g2g3 f8c5 f1h3 c6d4 b1d2 d4e6 e1g1 a7a5 c2c3 e8g8 a2a4 b7b6 a1b1 f6e7 d2c4 d7d6 h3g2 c8a6 b2b4 a5b4 c3b4 a6c4 b4c5 c4a6 c5b6 e6c5 d3d4 a6f1 d4c5 f1g2 g1g2 d6c5 b6b7 a8d8 f3d2');
INSERT INTO evaluations VALUES(16097,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(16097,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16097,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(16097,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(16097,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(16097,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(16097,36,'cp -35','d8f6 f1e2 g7g6 b1c3 f8g7 c3d5 f6d8 h2h4 h6h5 c2c3 c6e7 d5e3 d7d5 e1g1 e8g8 d1b3 c7c6 f1e1 d8d6 a1d1 g7h6 b3c2 h6e3 f2e3 c8g4 d3d4 g4f3 e2f3 a8d8 e1f1 d6e6 e4d5 e7d5 f3d5 d8d5 e3e4 d5d6');
INSERT INTO evaluations VALUES(1047,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(1047,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(1047,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(1047,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(1047,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(1047,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(1047,36,'cp -28','f1e2 g7g6 e1g1 f8g7 c2c3 h6h5 b1a3 d7d6 d1b3 e8g8 a3c4 c6d8 a1d1 f8e8 b3c2 d8e6 g2g3 e6g5 f3g5 f6g5 h2h4 g5d8 e2f3 a7a5 c4e3 g7h6 f1e1 a5a4 d3d4 c8e6 d4d5 e6d7 f3g2 d8f6 g1h2 b7b5 a2a3 e8b8 c2d2 f6e7 g2h3 d7h3 h2h3');
INSERT INTO evaluations VALUES(1046,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(1046,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(1046,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(1046,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(1046,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(1046,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(1046,36,'cp -45','g7g6 e1g1 f8g7 c2c3 d7d6 b1a3 c8d7 a3c4 g6g5 c4e3 c6e7 d3d4 e8c8 d4e5 d6e5 f3d2 h6h5 d2c4 g5g4 d1b3 c8b8 a1d1 g7h6 a2a4 a7a6 b3c2 e7g6 e3d5 f6e6 g2g3 h5h4 c4e3 h4g3 f2g3 h6e3 d5e3');
INSERT INTO evaluations VALUES(16741,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(16741,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(16741,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(16741,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(16741,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(16741,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(16741,36,'cp 140','b1c3 d7d6 h2h3 f8e7 d2d4 e5d4 f3d4 e8g8 e1g1 e7f6 d4f3 g8h8 c3d5 c8e6 f1e1 h6g8 a2a4 c6e5 f3e5 f6e5 f2f4 e5f6 d5f6 g8f6 c4d3 f8e8 a4a5 d6d5 e4e5 f6e4 a5a6 b7b6 d3e4 d5e4 d1d8 a8d8 c1e3 f7f5 e5f6 g7f6');
INSERT INTO evaluations VALUES(16735,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(16735,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(16735,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(16735,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(16735,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(16735,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(16735,36,'cp 133','d7d6 h2h3 f8e7 d2d4 e5d4 f3d4 e8g8 e1g1 e7f6 d4f3 g8h8 c3d5 c8e6 f1e1 h6g8 c4f1 e6d5 d1d5 a8b8 a2a4 g8e7 d5d1 e7g6 c2c3 f8e8 a4a5 h8g8 g2g3 b7b5 a5b6 a7b6');
INSERT INTO evaluations VALUES(15668,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(15668,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(15668,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(15668,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(15668,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(15668,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(15668,36,'cp 138','h2h3 f8e7 d2d4 e5d4 f3d4 e8g8 e1g1 e7f6 d4f3 g8h8 c3d5 h6g8 f1e1 c8e6 a2a4 c6e5 f3e5 d6e5 d1e2 f6g5 e1d1 g5c1 a1c1 d8g5 e2e3 g5h4 c4e2 f8c8 a4a5 g8f6 a5a6 b7b6 d5e7 c8e8 e7c6 h4e4 e3e4 f6e4');
INSERT INTO evaluations VALUES(21126,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(21126,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(21126,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(21126,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(21126,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(21126,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(21126,36,'cp 88','g1f3 c8g4 d4d5 c6e5 b1c3 g8f6 c1e3 c7c6 d5c6 e5f3 g2f3 d8d1 a1d1 g4f3 c3b5 f3d1 b5c7 e8d8 c7a8 b7c6 e1d1 d8c8 h1g1 g7g6 f1c4 c8b7 c4f7 f8g7 f2f3 f6h5 d1e2 h8a8 g1d1 b7c7 f7e6 a8d8 d1d8 c7d8 e6g4 h5f6 e3d4 f6g4 d4g7 g4f6 e2d3 d8e8 d3c4 f6d7 f3f4 a7a6 a2a3 e8f7 g7d4 f7e6 d4e3 d7f6');
INSERT INTO evaluations VALUES(21125,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(21125,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(21125,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(21125,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(21125,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(21125,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(21125,36,'cp 38','g8f6 b1c3 c8g4 d4d5 c6e5 c1e3 a7a6 h2h3 e5f3 g2f3 g4h5 f1c4 e7e5 d5e6 d8d1 a1d1 f7e6 e1e2 f8b4 e3d4 h5f7 h1g1 f6h5 g1g4 e8g8 e4e5 b7b5 c4d3 b4c3 b2c3 f7g6 d3g6 h7g6 d4e3 f8f5 d1d7 f5e5 d7c7 g6g5 c3c4 b5c4 c7c4 a8b8 e2f1 b8b2 a2a4 h5f4');
INSERT INTO evaluations VALUES(14385,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14385,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(14385,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(14385,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(14385,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(14385,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14385,36,'cp 54','b1c3 c8g4 d4d5 c6e5 c1e3 a7a6 h2h3 e5f3 g2f3 g4h5 f1c4 e7e5 d5e6 d8d1 a1d1 f7e6 c4e2 f8d6 f3f4 h5e2 e1e2 d6b4 e2f3 b4c3 b2c3 g7g6 h1g1 h8f8 h3h4 a8d8 d1d8 e8d8 g1g5 b7b6 a2a4 d8d7 c3c4 f6h5 g5e5 f8f7 c4c5 f7f8 c5b6 c7b6 a4a5 b6b5 e5c5 f8d8');
INSERT INTO evaluations VALUES(14383,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(14383,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(14383,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(14383,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14383,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(14383,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(14383,36,'cp 58','c8g4 d4d5 c6e5 c1e3 a7a6 h2h3 e5f3 g2f3 g4h5 f1c4 e7e5 d5e6 d8d1 a1d1 f7e6 e1e2 f8b4 e3d4 h5f7 h1g1 f6h5 a2a3 b4e7 d4e5 h8g8 c4a2 c7c6 f3f4 g7g5 f4g5 g8g5 f2f4 g5g1 d1g1 e7c5 g1g4 b7b5 c3d1 c5b6 a2b1 c6c5 d1f2 e8c8 g4g1 c5c4 b1c2');
INSERT INTO evaluations VALUES(25001,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(25001,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(25001,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(25001,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(25001,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(25001,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(25001,36,'cp 64','d4d5 c6e5 c1e3 a7a6 h2h3 e5f3 g2f3 g4h5 f1c4 e7e5 d5e6 d8d1 a1d1 f7e6 e1e2 f8b4 e3d4 h5f7 a2a3 b4e7 h1g1 e8g8 e4e5 f6d7 c3e4 f8d8 c4b3 b7b5 d1c1 c7c5 d4c5 e7c5 e4c5 d7e5 b3e6 d8e8 e6f7 g8f7 c5e4 e5c4 c1c3 a8c8 g1c1 c8d8 b2b4 f7g8');
INSERT INTO evaluations VALUES(24790,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(24790,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(24790,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(24790,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(24790,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(24790,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(24790,36,'cp 43','c6e5 c1e3 a7a6 h2h3 e5f3 g2f3 g4h5 f1c4 e7e5 d5e6 d8d1 a1d1 f7e6 e1e2 f8b4 e3d4 h5f7 h1g1 f6h5 a2a3 b4e7 d4e5 h8g8 f3f4 a8d8 d1d8 e7d8 g1d1 g7g5 f4f5 e8e7 c4e6 f7e6 f5e6 e7e6 e5h2 h5f6 e2f3 g8f8 c3e2 c7c5 f3e3 f6e4 e3e4 f8f2 e2c3 f2h2 d1d8 h2b2 c3a4 b2e2 e4f3 e2h2 a4c5 e6f5 c5e4 h2h3 e4g3 f5f6 d8f8 f6e5');
INSERT INTO evaluations VALUES(25034,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(25034,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(25034,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(25034,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(25034,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(25034,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25034,36,'cp 37','c1e3 a7a6 h2h3 e5f3 g2f3 g4h5 f1c4 e7e5 d5e6 d8d1 a1d1 f7e6 c4e2 f8d6 f3f4 h5e2 e1e2 d6b4 e2f3 b4c3 b2c3 g7g6 h3h4 h8f8 a2a4 f6h5 h1g1 a8d8 d1d8 e8d8 g1g5 d8d7 c3c4 f8b8 g5e5 h5f6 e3c5 b7b6 c5b4 c7c5 b4c3 b8d8 f4f5 g6f5 e4f5 e6f5');
INSERT INTO evaluations VALUES(14776,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(14776,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(14776,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(14776,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(14776,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(14776,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(14776,36,'cp 38','f8e7 a2a3 e8g8 b2b4 c6b8 c4c5 c7c6 f1d3 a7a5 c1b2 b7b6 c3a4 b6c5 b4c5 c8a6 d3a6 a8a6 e1g1 b8d7 b2c3 d8c7 d1c2 a6a8 a4b2 f6e4 b2d3 e7f6 a1b1 e4c3 c2c3 f8e8 f1c1 e6e5 f3e5 f6e5 d4e5 d7e5');
INSERT INTO evaluations VALUES(8478,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(8478,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(8478,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(8478,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(8478,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(8478,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(8478,36,'cp 65','a2a3 e8g8 b2b4 c6b8 c4c5 c7c6 f1e2 b8d7 e1g1 d8c7 c1b2 b7b6 c3a4 a7a5 d1c2 b6c5 a4c5 d7c5 b4c5 c8a6 e2a6 a8a6 b2c3 a6a8 f1c1 f8b8 g2g3 h7h6 a1b1 e7d8 c2a4 f6d7 h2h4 d7f6 b1b8 a8b8 f3e5');
INSERT INTO evaluations VALUES(54837,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(54837,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(54837,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(54837,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(54837,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(54837,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(54837,36,'cp 54','d2d4 g8f6 f1d3 f8g7 e1g1 b8c6 f1e1 e7e5 c2c3 c8g4 d4d5 c6e7 a2a4 a7a5 h2h3 g4f3 d1f3 e8g8 f3d1 f6d7 b2b4 f7f5 b1d2 g8h8 b4a5 a8a5 d2b3 a5a8 a4a5 d7f6 c3c4 f5e4 d3e4 f6e4 e1e4 e7f5 c1d2 g7h6 d2h6');
INSERT INTO evaluations VALUES(11587,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(11587,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(11587,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(11587,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(11587,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(11587,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(11587,36,'cp 72','b5c4 h7h6 d2d4 e7g6 a2a4 f8e7 c1e3 d7d6 d4d5 c6b8 a4a5 c8g4 h2h3 g4f3 d1f3 e7g5 e3d2 g5d2 b1d2 e8g8 f1e1 d8g5 a1d1 b8d7 f3e3 g5e3 e1e3 b7b6 b2b4 b6a5 b4a5 f8b8 c4f1 g6e7 e3e1');
INSERT INTO evaluations VALUES(11591,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(11591,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(11591,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(11591,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(11591,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(11591,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(11591,36,'cp 59','d7d6 f3g5 d6d5 e4d5 c6a5 b2b3 a5c4 b3c4 e7d5 g5f7 e8f7 c4d5 d8d5 b1c3 d5c6 d1h5 f7g8 h5e5 h7h5 e5d5 c6d5 c3d5 c8f5 d2d3 a8d8 d5c3 h5h4 h2h3 h8h5 f1e1 f8b4 c1b2 f5g6 e1e2 h5c5 c3d1 c5c6 a2a3 b4c5 g1f1 d8e8 e2e8 g6e8 b2e5 e8g6 f2f3 c5b6 d1c3');
INSERT INTO evaluations VALUES(44261,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(44261,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(44261,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(44261,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(44261,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(44261,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(44261,36,'cp 102','b1c3 g7g6 e2e4 d7d6 h2h3 f8g7 g1f3 e8g8 f1d3 e6d5 e4d5 f8e8 c1e3 b8d7 e1g1 a7a6 a2a4 d7e5 f3e5 e8e5 e3f4 e5e8 d1d2 d8c7 a1e1 c8d7 g2g4 e8e1 f1e1 a8e8 g1g2 e8e1 d2e1 f6e8 b2b3 g7e5 f4e5 d6e5 a4a5 f7f5 g4f5 d7f5 c3e4 f5e4 e1e4');
INSERT INTO evaluations VALUES(55675,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(55675,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(55675,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(55675,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(55675,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(55675,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(55675,36,'cp -56','g7g5 g2g3 f8g7 b1c3 b8c6 d4d5 c6e5 g3f4 g5f4 c1f4 c8g4 f1e2 e5f3 e2f3 d8f6 e1g1 f6f4 f3g4 f4h6 f1f3 g7e5 g4h3 g8f6 d1e2 h8g8 g1h1 a7a6 a1f1 f6g4 f1g1 h6h4 c3d1 e5d4 g1g4 g8g4 h3g4 h4g4 d1f2 g4h4 f2d3 e8c8 c2c3 d4b6 f3f4 h4e7 f4f1');
INSERT INTO evaluations VALUES(55352,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(55352,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(55352,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(55352,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(55352,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(55352,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(55352,36,'cp -65','b1c3 g5g4 c1f4 g4f3 d1f3 b8c6 f1b5 c8d7 e1c1 d8h4 c1b1 e8c8 d4d5 c6e5 f3e3 d7b5 c3b5 a7a6 f4g5 e5c4 b5a7 c8d7 g5d8 c4e3 d8h4 e3d1 h1d1 f8e7 h4f2 g8f6 d1f1 f6e4 f2e3 h8g8 g2g3 e4f6 c2c4 f6g4 e3g1 g4e5 b1c2 h7h5 g1d4 f7f6 c2c3 c7c5 d4e5 d6e5');
INSERT INTO evaluations VALUES(51094,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(51094,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(51094,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(51094,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(51094,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(51094,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(51094,36,'cp 110','d2d4 d7d6 b1c3 c7c6 g1f3 g7g6 a2a4 f8g7 f1c4 g8f6 e1g1 e8g8 c4b3 b8a6 h2h3 a6c7 f1e1 c8e6 d4d5 c6d5 c3d5 f6d5 e4d5 e6d7 c2c3 c7a6 c1e3 a6c5 e3c5 d6c5 f3e5 g7f6 e1e4 d8c7 d1e2 f6e5');
INSERT INTO evaluations VALUES(51091,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(51091,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(51091,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(51091,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(51091,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(51091,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(51091,36,'cp 99','d7d6 g1f3 g7g6 b1c3 c7c6 c1f4 f8g7 h2h3 g8f6 d1d2 e8g8 f4h6 b8a6 f1a6 a8a6 e1g1 b7b5 e4e5 f6d5 c3d5 c6d5 f1e1 c8e6 c2c3 e6f5 f3h4 d8d7 e5d6 e7d6 h6f4 f5e6 a2a3 a6c6 h4f3 e6f5');
INSERT INTO evaluations VALUES(31529,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(31529,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(31529,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31529,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(31529,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31529,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31529,36,'cp 0','f5d3 d1d3 c6c5 c2c3 b8d7 e1g1 g8e7 d4c5 d7c5 d3e2 a7a6 f1d1 d8c7 b1d2 e7c6 d2b3 f8e7 h2h4 e8g8 b3c5 e7c5 d1d3 c5a7 b2b3 a7b8 d3e3 b8a7');
INSERT INTO evaluations VALUES(31497,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31497,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31497,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31497,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31497,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31497,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31497,36,'cp 0','d1d3 c6c5 c2c3 b8d7 e1g1 g8e7 d4c5 d7c5 d3e2 a7a6 f1d1 d8c7 h2h4 e7c6 c1f4 f8e7 h4h5 h7h6 b1d2 e8g8 f3d4 c7b6 a1b1 a8c8 e2g4 g8h8 g4e2');
INSERT INTO evaluations VALUES(22022,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(22022,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(22022,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(22022,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(22022,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(22022,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(22022,36,'cp 55','f3d4 g8f6 b1c3 f8e7 e1g1 e8g8 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 a4a5 d7c6 c4d3 f8e8 f1e1 a8d8 b2b4 f6h5 b4b5 c6d7 f4e3 a6b5 d3b5 c7c6 b5e2 h5f6 d4d2 d6d5 e3b6 d5e4 b6d8 d7d2 d1d2 e7d8');
INSERT INTO evaluations VALUES(22017,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(22017,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(22017,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(22017,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(22017,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(22017,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(22017,36,'cp 38','f8e7 b1c3 g8f6 c1f4 e8g8 e1g1 c6e5 c4b3 c8g4 f2f3 g4d7 a2a4 c7c5 d4e2 c5c4 f4e5 d6e5 b3c4 a8c8 b2b3 d8c7 g1h1 f8d8 c3d5 f6d5 e4d5 g8h8 d1d3 f7f5 d3d2 a7a6 a4a5 d7b5 c4b5 a6b5 e2g3 c7c2 d2c2 c8c2 g3f5');
INSERT INTO evaluations VALUES(5654,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(5654,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(5654,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(5654,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(5654,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(5654,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(5654,36,'cp 43','e1g1 g8f6 b1c3 e8g8 c1f4 c6e5 c4b3 c8g4 f2f3 g4d7 a2a4 c7c5 d4e2 c5c4 f4e5 d6e5 b3c4 a8c8 b2b3 d8c7 g1h1 f8d8 c3d5 f6d5 e4d5 g8h8 e2g3 f7f5 d1d2 e7d6 c4d3 g7g6 g3e2 d8e8 c2c3 c7c5 d3c2 c5a5 b3b4 a5c7 f1d1 h8g7 a4a5 d7b5');
INSERT INTO evaluations VALUES(50366,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(50366,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(50366,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(50366,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50366,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(50366,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(50366,36,'cp -1','d7d5 g1f3 c8g4 e2e3 e7e6 h2h3 g4h5 d2d3 b8d7 b1d2 h7h6 e3e4 h5g6 d1e2 a7a5 e4d5 f6d5 a2a3 a5a4 b3b4 c7c5 c2c4 d5e7 b4b5 e7f5 g2g3 f5d4 f3d4 c5d4 b2d4 d7c5 e2e3 d8g5 f2f4 g5f5 d4c5 f8c5');
INSERT INTO evaluations VALUES(47256,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47256,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47256,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47256,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47256,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47256,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47256,36,'cp 12','g1f3 c8g4 e2e3 e7e6 h2h3 g4h5 d2d3 b8d7 b1d2 h7h6 e3e4 h5g6 d1e2 a7a5 e4d5 f6d5 a2a3 f8e7 b2g7 h8g8 g7b2 e7f6 d3d4 g6c2 a1c1 c2f5 g2g3 c7c6 d2c4 d8c7 f1g2 e8c8 e1g1 d5f4 e2e3 f5h3 g2h3 f4h3 g1h1');
INSERT INTO evaluations VALUES(20979,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20979,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20979,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(20979,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20979,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20979,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20979,36,'cp 0','f7f6 e2e4 e7e5 d4e5 f6e5 f4g3 d5e4 d1d8 e8d8 f1b5 g8f6 b1d2 f8d6 e1c1 c8e6 g1h3 h7h6 h1e1 e6g4 b5e2 g4e6');
INSERT INTO evaluations VALUES(20838,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20838,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20838,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20838,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20838,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20838,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20838,36,'cp 0','e2e4 e7e5 d4e5 f6e5 f4g3 d5e4 d1d8 e8d8 f1b5 g8f6 b1d2 f8d6 e1c1 c8e6 g1h3 h7h6 h1e1 e6g4 b5e2 g4e6');
INSERT INTO evaluations VALUES(50620,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(50620,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50620,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(50620,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(50620,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(50620,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(50620,36,'cp 18','g8f6 f1d3 b8c6 d4c6 d7c6 h2h3 e7e5 c1e3 f8b4 c2c3 b4d6 d1e2 d8e7 b1d2 c8e6 e1g1 e8g8 f1d1 h7h6 a2a4 b7b5 b2b4 f8d8 d1c1 e7c7 c1b1 a6a5 a4b5 c6b5 d3b5 c7c3 b4a5 a8a5 a1a5 c3a5 b5c4 e6c4 e2c4');
INSERT INTO evaluations VALUES(43337,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43337,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43337,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43337,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(43337,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(43337,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43337,36,'cp 18','f1d3 b8c6 d4c6 d7c6 a2a4 e7e5 e1g1 c8e6 b1d2 f8c5 a4a5 d8e7 d1f3 e8g8 d2c4 a8d8 c4e5 c5d4 e5c4 e6c4 d3c4 e7e4 f3e4 f6e4 g2g3 d4f6 c4d3 e4c5 d3e2 c5e4 a1a3 f8e8 e2f3 e4d2 c1d2 d8d2 a3b3 d2c2 b3b7 c2b2');
INSERT INTO evaluations VALUES(52455,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(52455,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52455,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(52455,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(52455,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(52455,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(52455,36,'cp 0','g7g6 g1f3 f8g7 h2h3 d7d6 f1b5 b8d7 e1g1 a7a6 b5d7 c8d7 d2d3 g8f6 c2c4 e8g8 b1c3 b7b5 f1e1 f6e8 e4e5 a8b8 b2b4 b8c8 b4c5 c8c5 c3d5 d7e6 d5f4 e6d7');
INSERT INTO evaluations VALUES(52313,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(52313,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(52313,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(52313,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(52313,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(52313,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(52313,36,'cp 0','g1f3 f8g7 h2h3 e7e6 d2d4 c5d4 c2c3 d7d5 e4d5 d8d5 c3d4 g8e7 b1c3 d5d8 f1c4 e8g8 e1g1 b8c6 c1f4 c6d4 f3d4 d8d4 d1d4 g7d4 c3b5 d4b2 a1b1 b2g7 b5c7 a8b8 c7a6 b8a8');
INSERT INTO evaluations VALUES(52312,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(52312,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52312,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52312,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52312,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(52312,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(52312,36,'cp -8','f8g7 h2h3 a7a6 d2d3 b8c6 c2c3 d7d5 f1e2 d5d4 c3d4 c5d4 a3a4 g8f6 b1a3 e8g8 e1g1 f6d7 c1d2 a6a5 d1b3 d7b6 f1c1 e7e5 a3c4 b6c4 b3c4 c8e6 c4b5 a8a6 b2b4 a6b6 b5c5 a5b4 a4a5 b6a6 d2b4 g7h6 f3e5 h6c1 a1c1');
INSERT INTO evaluations VALUES(37830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37830,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(37830,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(37830,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(37830,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(37830,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37830,36,'cp 0','h2h3 d7d6 f1b5 b8d7 e1g1 a7a6 b5d7 c8d7 d2d3 g8f6 c2c4 e8g8 b1c3 b7b5 f1e1 d7c6 d3d4 e7e6 d4d5 e6d5 c4d5 c6b7 d1c2 b5b4 c3a4 a6a5 c1f4 f8e8 a1d1 a8c8 c2c4 c8a8');
INSERT INTO evaluations VALUES(22740,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(22740,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22740,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(22740,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(22740,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(22740,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(22740,36,'cp 17','g2g3 e5e4 f3h4 g8f6 f1g2 d7d5 c4d5 d8d5 b1c3 d5d4 d2d3 e4d3 e1g1 f8e7 h4f3 d4d8 e2d3 e8g8 d3d4 c8e6 c1g5 c6b4 f1e1 c7c6 a2a3 b4d5 g5d2 d8c8 c3a4 h7h6 f3e5 e6h3 a1c1 f8e8 d1f3 h3g2 g1g2');
INSERT INTO evaluations VALUES(57196,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(57196,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(57196,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(57196,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(57196,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(57196,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(57196,36,'cp -7','d7d5 d1h5 b8c6 f1b5 d8d6 g1f3 e5e4 h5e5 c8e6 e5d6 f8d6 f3d4 g8e7 d2d3 e4d3 c2d3 e8c8 d4e6 f7e6 d3d4 e6e5 d4e5 d6e5 e1g1 g7g6 a1b1 a7a6 b5e2 e7f5 f1d1 d5d4 e3d4 c6d4 e2d3 h8e8 g1f1 d4c6 c3e4 e5g7 c1g5');
INSERT INTO evaluations VALUES(56435,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(56435,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56435,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(56435,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(56435,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(56435,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(56435,36,'cp -19','d1h5 b8c6 f1b5 d8d6 g1f3 e5e4 h5e5 c8e6 e5d6 f8d6 f3d4 g8e7 d2d3 e4d3 c2d3 e8c8 d4c6 e7c6 d3d4 h7h5 b5c6 b7c6 c1d2 h5h4 c3a4 h4h3 g2g3 e6g4 e1g1 g4e2 f1c1 g7g5 a4c5 g5g4 b2b4 e2f3 a2a4 h8e8 b4b5 d6c5 c1c5 c6b5 a4b5');
INSERT INTO evaluations VALUES(56327,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(56327,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(56327,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(56327,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(56327,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(56327,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(56327,36,'cp -23','b8c6 f1b5 d8d6 g1f3 e5e4 h5e5 c8e6 e5d6 f8d6 f3d4 g8e7 d4e6 f7e6 a2a3 a7a6 b5e2 b7b5 b2b3 h7h5 c1b2 e8d7 h2h4 g7g6 d2d3 e4d3 c2d3 c6a5 b3b4 a5b7 g2g4 c7c5 g4h5 g6h5 c3a2 h8h6 h1g1 h6g6 g1g6 e7g6 e2h5 g6h4 b2f6 h4f5 h5g4 d7c6');
INSERT INTO evaluations VALUES(21662,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(21662,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(21662,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(21662,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(21662,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(21662,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(21662,36,'cp -19','f1b5 d8d6 g1f3 e5e4 h5e5 c8e6 e5d6 f8d6 f3d4 g8e7 d2d3 e4d3 c2d3 e8c8 d4e6 f7e6 d3d4 e6e5 d4e5 d6e5 b5e2 e7f5 c1d2 d5d4 e2g4 g7g6 e3d4 d8d4 g4f5 g6f5 e1c1 h7h5 c3e2 d4g4 g2g3 c6d4 e2d4 g4d4 d2e3 d4c4 c1b1 c4e4 e3a7 b7b6');
INSERT INTO evaluations VALUES(21528,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(21528,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(21528,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(21528,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21528,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(21528,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(21528,36,'cp -12','d8d6 g1f3 e5e4 h5e5 c8e6 e5d6 f8d6 f3d4 g8e7 d2d3 e4d3 c2d3 e8c8 d4c6 e7c6 d3d4 c6e7 c1d2 c7c6 b5d3 h7h5 a1b1 a7a6 a2a4 h5h4 h2h3 e6f5 e1e2 d6b4 c3a2 f5d3 e2d3 b4d2 d3d2 e7f5 a2b4 a6a5 b4d3 f5d6 h1c1 c8c7 d2e2 d8a8 b2b4 a5b4 b1b4 d6c4');
INSERT INTO evaluations VALUES(22638,30,'cp -238',NULL);
INSERT INTO evaluations VALUES(22638,31,'cp -225',NULL);
INSERT INTO evaluations VALUES(22638,32,'cp -214',NULL);
INSERT INTO evaluations VALUES(22638,33,'cp -228',NULL);
INSERT INTO evaluations VALUES(22638,34,'cp -205',NULL);
INSERT INTO evaluations VALUES(22638,35,'cp -233',NULL);
INSERT INTO evaluations VALUES(22638,36,'cp -211','g8f6 h5d1 c6d4 b5a4 f6e4 d2d3 e4c5 c2c3 d4c6 a4c2 d7d5 d3d4 e5d4 c3d4 c5e6 g1f3 g7g6 b1c3 f8g7 c3e2 e8g8 e1g1 d8d6 f1e1 b7b6 h2h4 c6b4 h4h5 c8a6 a2a3 b4c2 d1c2 a6e2 c2e2 e6d4 f3d4 g7d4 c1h6 f8d8 a1d1 c7c5 h6g5 f7f6 g5h4');
INSERT INTO evaluations VALUES(16304,30,'cp -209',NULL);
INSERT INTO evaluations VALUES(16304,31,'cp -191',NULL);
INSERT INTO evaluations VALUES(16304,32,'cp -198',NULL);
INSERT INTO evaluations VALUES(16304,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(16304,34,'cp -199',NULL);
INSERT INTO evaluations VALUES(16304,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(16304,36,'cp -199','h5d1 c6d4 b5a4 f6e4 d2d3 e4c5 c2c3 d4c6 a4c2 d7d5 d3d4 e5d4 c3d4 c5e6 g1f3 g7g6 b1c3 f8g7 c3e2 e8g8 e1g1 b7b6 c2a4 c6e7 b2b4 c8a6 f1e1 a6e2 e1e2 a8c8 c1b2 c7c6 a1c1 b6b5 a4c2 a7a5 b4a5 d8a5 h2h4');
INSERT INTO evaluations VALUES(50682,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50682,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(50682,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(50682,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(50682,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50682,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(50682,36,'cp 10','g8f6 c1e3 e7e5 d4f3 f8e7 f1c4 e8g8 e1g1 b8c6 e3g5 c8e6 c4b3 f6d7 g5e7 d8e7 d1e2 d7c5 c3d5 e7d8 a1d1 a8c8 d1d2 c6e7 d5e7 d8e7 b3e6 f7e6 g2g3 b7b5 b2b3 c8c6 c2c3 c5d7 f1d1 f8c8');
INSERT INTO evaluations VALUES(52379,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52379,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52379,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(52379,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52379,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(52379,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52379,36,'cp 19','e7e5 g1f3 g8f6 b1d2 d8c7 d3d4 e5d4 f3d4 d7d6 g2g3 g7g6 f1g2 b8d7 b2b3 f8g7 e1g1 e8g8 c1b2 f8e8 c2c4 a7a5 h2h3 c7b6 d1c2 f6g4 d2f3 g4e5 f1d1 e5f3 d4f3 g7b2 c2b2 d7c5 d1d6 c5e4 d6d4 e4c5 a1d1 c8f5 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(51863,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(51863,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(51863,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51863,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(51863,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(51863,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51863,36,'cp 27','d3d4 d7d6 g1f3 g8f6 d4e5 f6e4 b1d2 e4d2 c1d2 b8d7 d1e2 d6e5 f3e5 d7e5 e2e5 d8e7 e5e7 f8e7 f1c4 e8g8 e1c1 e7f6 h1e1 c8f5 d2e3 a7a5 h2h3 b7b5 c4d3 f5e6 d3e4 f8c8 a2a3 f6e7 f2f3 h7h6 g2g4 e7h4 e1e2 a5a4 c2c3 h4f6 e3c5');
INSERT INTO evaluations VALUES(34827,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(34827,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(34827,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(34827,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(34827,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(34827,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(34827,36,'cp 91','c2c4 d5d8 b1c3 c7c5 g1f3 g8f6 c1e3 c5d4 f3d4 a7a6 g2g3 e6e5 d4e2 d8d1 a1d1 b8c6 f1g2 c8e6 b2b3 f8b4 e1g1 e8g8 c3d5 e6d5 c4d5 c6e7 d5d6 e7f5 e3d2 b4d2 d1d2 f8d8 g2b7 a8a7 b7g2 d8d6 d2d6 f5d6 f1d1 d6e8 e2c3 g8f8');
INSERT INTO evaluations VALUES(34821,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(34821,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(34821,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(34821,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34821,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(34821,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(34821,36,'cp 90','d5d8 b1c3 c7c5 g1f3 g8f6 c1e3 f8e7 d4c5 e8g8 a2a3 f6g4 e3f4 e7c5 f4g3 d8d1 a1d1 b8c6 f1e2 c5e7 h2h3 g4h6 b2b4 e7f6 c3e4 h6f5 g3c7 e6e5 c4c5 c8e6 b4b5 f8c8 g2g4 f5d4 e4f6 g7f6 b5c6 c8c7 f3d4 e5d4 c6b7 c7b7 e2f3');
INSERT INTO evaluations VALUES(56010,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(56010,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(56010,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(56010,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(56010,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(56010,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(56010,36,'cp 74','g1f3 g8f6 b1c3 c7c5 c1e3 c5d4 f3d4 a7a6 g2g3 e6e5 d4e2 d8c7 a1c1 f8c5 e3c5 c7c5 f1g2 b8c6 e1g1 e8g8 c3d5 f8d8 a2a3 c8g4 h2h3 g4e6 b2b4 c5d6 e2c3 e6d5 c3d5 f6d5 g2d5 d8d7 d1g4 a8d8 f1e1 d6g6 d5c6 g6g4 h3g4 b7c6 e1e5 g8f8');
INSERT INTO evaluations VALUES(56009,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(56009,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(56009,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(56009,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(56009,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(56009,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(56009,36,'cp 68','c7c5 b1c3 g8f6 c1e3 c5d4 f3d4 a7a6 g2g3 e6e5 d4f3 d8d1 a1d1 b8c6 a2a3 c8g4 f1e2 f8e7 c3d5 f6d5 c4d5 g4f3 e2f3 c6d4 e3d4 e5d4 d1d4 e7d6 d4d3 a8c8 d3b3 b7b5 e1g1 e8e7 a3a4 b5a4 f1e1 e7f6 b3b6 h8d8 b6a6 c8c2 a6a4 c2b2');
INSERT INTO evaluations VALUES(57487,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(57487,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(57487,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(57487,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(57487,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(57487,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(57487,36,'cp 169','g8f6 d2d4 e5d4 f3d4 d7d5 f1b5 c7c6 b5d3 f8d6 e1g1 e8g8 c2c4 d6c5 d4b3 c5b6 c1g5 b8d7 f1e1 a7a5 c4d5 a5a4 d3c4 c6d5 c4d5 g8h8 b3d4 b6d4 d1d4 h7h6 g5f6 d7f6 b1c3 c8f5 a1d1 f6d5 c3d5 d8g5 h2h4 g5g4 d4g4 f5g4 f2f3 g4h5 d5b6 a8a5 b6d7 f8e8 g1f2 a4a3 b2b3 e8e1 f2e1');
INSERT INTO evaluations VALUES(49989,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(49989,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(49989,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(49989,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(49989,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(49989,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(49989,36,'cp 178','d2d4 e5d4 f3d4 d7d5 f1b5 c7c6 b5d3 f8c5 d1e2 d8e7 e2e7 c5e7 e1g1 e8g8 f1e1 e7d6 c2c3 b8d7 c1g5 d7e5 d3c2 e5c4 b2b3 c4e5 b1d2 c8d7 h2h3 f8e8 g5h4 e5f7 g2g4 a7a5 a2a4 d6f4 d2f3 f6e4 c3c4 g7g5 f5g6');
INSERT INTO evaluations VALUES(33684,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(33684,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(33684,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(33684,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(33684,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(33684,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(33684,36,'cp 132','b1c3 d5d6 g1f3 c8f5 f3e5 b8d7 d1b3 d7e5 d4e5 d6e5 b3b7 f5e4 c3e4 e5e4 b7e4 f6e4 f2f3 e4f6 e2e4 a8b8 c1e3 b8b2 e1c1 b2b8 e3a7 b8b7 a7d4 f6d7 c1c2 e7e6 a2a4 f8d6 a4a5 e8g8 a5a6 b7b4 d4c3 b4b6 d1b1 b6b1');
INSERT INTO evaluations VALUES(33680,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(33680,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(33680,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(33680,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(33680,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(33680,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(33680,36,'cp 121','d5a5 c1d2 a5b6 d1c2 b8c6 g1f3 c6b4 c2b1 c8g4 a2a3 b4c6 e2e3 a7a6 f1e2 e7e6 b2b4 f8d6 h2h3 g4f5 b1b2 b6a7 e1g1 e8g8 a1c1 c6e7 f3e5 f8d8 e2d1 c7c6 d1f3 f5g6 c3a4 d6e5 d4e5 f6d7 d2c3');
INSERT INTO evaluations VALUES(33733,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(33733,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(33733,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(33733,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(33733,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(33733,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(33733,36,'cp 116','c1d2 e7e5 e2e4 c7c6 d4e5 a5e5 g1f3 e5c7 e4e5 f6d5 f1c4 c8e6 d1b3 c7b6 c3d5 c6d5 c4d5 b6b3 d5b3 e6b3 a2b3 b8c6 d2c3 f8e7 e1e2 e8d7 f3e1 h8d8 e1c2 d7e6 g2g4 f7f5 e5f6 e7f6 f2f4 h7h5 g4g5 f6c3 b2c3 e6f5 a1a4 d8d7 a4c4 a8e8 c2e3 f5g6 h1g1 g6f7');
INSERT INTO evaluations VALUES(33732,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(33732,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(33732,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(33732,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(33732,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(33732,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(33732,36,'cp 142','c7c6 e2e4 e7e5 d4e5 a5e5 g1f3 e5c7 e4e5 f6d5 f1d3 c8e6 e1g1 b8d7 f3g5 e8c8 g5e6 f7e6 c3d5 e6d5 f2f4 d7c5 d3c2 d5d4 c2f5 c8b8 d1f3 g7g6 f5h3 f8e7 b2b4 c5a4 a2a3 a4b6 f3b3 b6d5 g1h1 c7b6 f1f3 d5e3 a1e1 h8f8 h3e6 b6b5 d2e3 d4e3');
INSERT INTO evaluations VALUES(40924,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(40924,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40924,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(40924,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40924,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(40924,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(40924,36,'cp 18','c4d5 e6d5 e2e3 c7c6 f1d3 f6e4 g5e7 d8e7 d1c2 c8f5 e1g1 b8d7 b1c3 e8g8 a2a4 f8e8 f1c1 a8d8 a4a5 a7a6 c3a4 e7b4 d3e4 d5e4 a4c5 d7c5 c2c5 b4c5 c1c5 f5e6 f3d2 f7f5 h2h3 g8f7 a1c1 d8d5 d2c4 d5c5 c4d6 f7e7 d4c5');
INSERT INTO evaluations VALUES(40915,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(40915,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40915,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40915,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(40915,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40915,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40915,36,'cp 0','e6d5 e2e3 c7c6 f1d3 f6e4 g5e7 d8e7 b1c3 c8f5 f3e5 e4d6 d3f5 d6f5 d1g4 g7g6 g4f4 f7f6 e5d3 b8a6 e1g1 e8g8 h2h4 a8e8 h4h5 a6c7 a1c1 f5d6 f1e1 c7e6 f4f3');
INSERT INTO evaluations VALUES(41074,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(41074,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(41074,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(41074,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(41074,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(41074,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(41074,36,'cp 0','b1c3 h7h6 g5f6 e7f6 e2e3 e8g8 d1b3 c7c6 f1d3 d8b6 b3c2 c8e6 e1g1 b8d7 h2h3 f6e7 c3a4 b6c7 a1b1 a7a5 a2a3 f8e8 f1c1 e8c8 d3h7 g8f8 h7d3');
INSERT INTO evaluations VALUES(41062,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41062,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(41062,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(41062,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41062,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41062,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41062,36,'cp 0','h7h6 g5f6 e7f6 e2e3 e8g8 d1b3 c7c6 f1d3 b8d7 b3c2 f6e7 e1g1 e7d6 a1e1 d8c7 e1c1 c7d8');
INSERT INTO evaluations VALUES(40723,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(40723,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(40723,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40723,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40723,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(40723,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(40723,36,'cp 3','g5f6 e7f6 e2e3 e8g8 d1b3 c7c6 f1d3 d8b6 b3c2 c8e6 e1g1 b8d7 b2b4 f8c8 a1b1 a7a5 b4a5 b6a5 b1b7 c6c5 b7b5 a5a3 c2d2 c5d4 c3d5 d4e3 d2e3 e6d5 b5d5 d7c5 d3f5 a3e3 f2e3 c5e6 f5e6 f7e6 d5d1 a8a2');
INSERT INTO evaluations VALUES(21041,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(21041,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21041,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(21041,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(21041,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(21041,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(21041,36,'cp 21','d5e4 d4d5 c6b8 c3e4 c7c6 f1c4 c6d5 d1d5 d8d5 c4d5 e7e6 d5c4 g8f6 e4f6 g7f6 c1e3 h8g8 g2g3 b7b6 g1f3 b8d7 c4b5 a7a6 b5c6 a8a7 e3b6 a7b7 b6d4 b7c7 c6d7 c8d7 d4f6 c7c2 e1g1 f8g7 f6c3 g7c3 b2c3 d7c6 f3e5 c2c3');
INSERT INTO evaluations VALUES(21136,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(21136,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(21136,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(21136,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(21136,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21136,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21136,36,'cp 23','d4d5 c6b8 c3e4 c7c6 f1c4 c6d5 d1d5 d8d5 c4d5 e7e6 d5c4 c8d7 c1f4 d7c6 f2f3 c6e4 f3e4 b8d7 e1c1 a8c8 c4d3 g8f6 g1f3 a7a6 h2h3 d7c5 h1e1 c5d3 d1d3 h7h6 c2c3 f6d7 c1c2 d7c5 d3d4 f7f6 e4e5 f6f5 f4e3 f8e7 e1d1 e8g8 b2b4 c5a4');
INSERT INTO evaluations VALUES(20957,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(20957,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(20957,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(20957,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20957,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20957,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20957,36,'cp 19','c6b8 c3e4 c7c6 f1c4 c6d5 d1d5 d8d5 c4d5 e7e6 d5c4 g8f6 e4f6 g7f6 c1e3 h8g8 g2g3 c8d7 g1f3 d7c6 c4e2 b8d7 h1f1 a8c8 e1c1 a7a6 f1e1 c6e4 f3d4 f8c5 e2d3 e4d3 d1d3 c5e7 d4f3 e7c5 a2a4 g8g4 b2b3 h7h5 c1b2 c5e3 e1e3 d7c5');
INSERT INTO evaluations VALUES(29584,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(29584,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(29584,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29584,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29584,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(29584,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(29584,36,'cp 22','c2c4 c7c6 f1d3 f5g6 e1g1 e7e6 d1c2 g6d3 c2d3 f8e7 b1c3 e8g8 h2h3 b8d7 e3e4 d5e4 c3e4 d8c7 e4f6 d7f6 c1d2 a8d8 d3c2 f6d7 a1d1 f8e8 d2c3 e7f6 f1e1 c6c5 d4d5 f6c3 c2c3 e6d5 d1d5 d7f6 d5d8 e8d8');
INSERT INTO evaluations VALUES(29526,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(29526,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(29526,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(29526,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29526,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29526,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(29526,36,'cp 13','c7c6 f1d3 f5g6 e1g1 e7e6 d1c2 d5c4 c2c4 b8d7 f1d1 f8d6 b1d2 e6e5 d4e5 d6e5 c4c2 d8e7 d3g6 h7g6 f3e5 d7e5 d2e4 f6e4 c2e4 f7f5 e4c2 e7g5 g1h1 a8d8 h2h3 d8d1 c2d1 g5g4 c1d2 g4d1 a1d1');
INSERT INTO evaluations VALUES(40953,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(40953,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40953,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40953,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(40953,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(40953,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(40953,36,'cp 2','e8g8 b2b3 c7c5 d4c5 e7c5 c4d5 c5b4 d1c2 f6d5 c1d2 b4c3 d2c3 d5c3 c2c3 b7b6 f1e2 c8b7 a1c1 b8d7 e1g1 a8c8 c3b2 d8e7 b3b4 f8d8 a2a3 d7f6 c1c8 d8c8 f1c1 c8c1 b2c1 h7h6 h2h3 f6e4 c1c2 a7a5 e2d3 e4f6 b4a5 b6a5');
INSERT INTO evaluations VALUES(36321,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(36321,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(36321,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(36321,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(36321,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(36321,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36321,36,'cp 9','b2b3 c7c5 d4c5 e7c5 c1b2 b8c6 a2a3 d5c4 b3b4 c5e7 f1c4 d8d1 a1d1 a7a6 e1g1 b7b5 c4e2 c8b7 f3d2 a6a5 b4a5 b5b4 a3b4 e7b4 d2c4 b7a6 c4b6 b4c3 e2a6 a8a6 b2c3 c6a5 b6d7 f6d7 d1d7 a6a8 d7c7 a5b3');
INSERT INTO evaluations VALUES(36318,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36318,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(36318,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36318,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(36318,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36318,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36318,36,'cp 0','c7c5 d4c5 e7c5 c4d5 c5b4 d1c2 f6d5 c1d2 b4c3 d2c3 d5c3 c2c3 b7b6 f1e2 c8b7 e1g1 b8d7 c3b2 a8c8 a1c1 c8c1 f1c1 d8a8 b2a3 h7h6 h2h3 d7f6 b3b4 a7a6 a3b2 f8c8 c1c8 a8c8 f3e5 f6d5 e2f3 d5f6 f3e2');
INSERT INTO evaluations VALUES(8829,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(8829,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(8829,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8829,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(8829,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8829,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(8829,36,'cp 11','c3a4 c5b6 a4b6 a7b6 a2a4 e8g8 f1e1 d8e7 c1d2 c8e6 c4e6 e7e6 b2b4 f6d7 f3h4 c6e7 h2h3 f7f5 e4f5 e7f5 d1g4 e6f7 h4f5 f7f5 g4f5 f8f5 a4a5 f5f8 d2e3 b6a5 b4a5 b7b6 a5b6 d7b6 a1a8 f8a8 e3b6 c7b6 e1b1 a8a6 f2f3 g8f7 g1f2 f7e7 b1b5');
INSERT INTO evaluations VALUES(8870,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(8870,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(8870,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(8870,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8870,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(8870,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(8870,36,'cp 15','c5b6 a2a3 c8e6 c4e6 f7e6 c2c3 e8g8 a4b6 a7b6 h2h3 f8f7 f1e1 f6d7 b2b4 d7f8 a3a4 f8g6 c1d2 d8e7 d2e3 a8f8 f3h2 g6f4 e3f4 f7f4 a1a2 e7f7 e1f1 g8h8 g2g3 f4f6 g1g2 d6d5');
INSERT INTO evaluations VALUES(1899,30,'cp 239',NULL);
INSERT INTO evaluations VALUES(1899,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(1899,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(1899,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(1899,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(1899,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(1899,36,'cp 238','f4d2 b4e7 b1c3 c6e5 e2e4 g8f6 f3e5 e7e5 f2f4 e5e6 d1f3 d7d5 c3d5 f6d5 f1c4 e6b6 c4d5 c8e6 e1c1 e6d5 e4d5 e8c8 h1e1 h7h5 a2a3 f8c5 d2c3 b6g6 b2b4 c5d6 g2g3 g6g4 f3e3 b7b6 e3d3 h8e8 c1b2 e8e1 d1e1 h5h4 c3e5 d6e5 f4e5 g4d7 e1d1 d8e8 d3d4 h4g3');
INSERT INTO evaluations VALUES(1903,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(1903,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(1903,32,'cp 230',NULL);
INSERT INTO evaluations VALUES(1903,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(1903,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(1903,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(1903,36,'cp 226','b4e7 b1c3 c6e5 e2e4 g8f6 f3e5 e7e5 f2f4 e5e6 d1f3 d7d5 c3d5 f6d5 f1c4 e6b6 c4d5 c8e6 e1c1 e6d5 e4d5 e8c8 h1e1 h7h5 a2a3 f8c5 d2c3 b6g6 b2b4 c5d6 g2g3 c8b8 c3d4 h5h4 c1b2 h4g3 h2g3 b7b6 d1d3 d8e8 e1e8 h8e8 c2c4 e8e1 c4c5 d6e7 b2b3 e7f6 d4f6 g6f6');
INSERT INTO evaluations VALUES(6447,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(6447,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(6447,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(6447,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(6447,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(6447,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(6447,36,'cp -22','c6d4 c1e3 d8f6 e3d4 e5d4 e4e5 f6g6 e1g1 g8e7 f3d4 d7d5 e5d6 g6d6 d4f3 e8g8 d1d6 c5d6 b1d2 c8f5 f3d4 f5g6 c2c3 e7f5 d4f3 a8d8 c4b3 f8e8 f1e1 d6c5 a1d1 g6h5 d2e4 d8d1 b3d1 c5b6');
INSERT INTO evaluations VALUES(6852,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(6852,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(6852,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(6852,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(6852,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(6852,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(6852,36,'cp -17','c1e3 d8f6 e3d4 e5d4 e4e5 f6g6 e1g1 g8e7 f3d4 d7d5 e5d6 g6d6 d4f3 e8g8 d1d6 c5d6 b1d2 c8f5 f3d4 f5g6 f1e1 f8e8 c2c3 e7f5 d4f3 g8f8 g2g3 e8e1 a1e1 a8d8 d2b3 b7b6 a2a4 a7a5 h2h3 f7f6 g3g4 f5e7 b3d4 g6f7 d4e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(49359,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(49359,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(49359,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(49359,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(49359,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(49359,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(49359,36,'cp 28','e2e4 h7h6 g5f6 d8f6 b1c3 a7a6 g2g3 d7d5 f1g2 c7c6 g1e2 f6d8 e1g1 f8e7 f1e1 e8g8 e2f4 b7b5 a2a3 b8d7 h2h4 d5e4 e1e4 d8c7 d1e2 c7b6 f4h5 d7f6 e4f4 f6h5 e2h5 a6a5 a1d1 b5b4');
INSERT INTO evaluations VALUES(49362,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(49362,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49362,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(49362,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(49362,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(49362,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(49362,36,'cp 21','h7h6 g5f6 d8f6 b1c3 d7d6 d1d2 g7g5 h2h4 c8d7 c3b5 d7b5 f1b5 c7c6 b5e2 b8d7 e1c1 f8e7 d2e3 e8c8 c1b1 c8b8 g2g3 e6e5 c2c3 g5h4 h1h4 f6g6 h4h1 d6d5 e2d3 e5d4 c3d4 d5e4 d3e4 g6e6 g1f3 d7f6 f3e5 e7d6 b1a1 f6e4');
INSERT INTO evaluations VALUES(49233,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(49233,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(49233,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(49233,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(49233,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(49233,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(49233,36,'cp 21','g5f6 d8f6 b1c3 d7d6 d1d2 g7g5 h2h4 c8d7 c3b5 d7b5 f1b5 c7c6 b5e2 b8d7 e1c1 f8e7 d2e3 e8c8 c1b1 c8b8 c2c3 c6c5 g2g3 c5d4 c3d4 d7b6 g1f3 f6g7 d1c1 d8c8 a2a3 h8d8 c1c8 d8c8 e2d3 g5g4 f3d2 d6d5 e4e5 a7a6 h1c1 c8c1 b1c1 b8a8 c1b1');
INSERT INTO evaluations VALUES(49229,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(49229,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(49229,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(49229,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(49229,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(49229,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49229,36,'cp 12','d8f6 b1c3 d7d6 d1d2 g7g5 h2h4 c8d7 c3b5 d7b5 f1b5 c7c6 b5e2 b8d7 e1c1 f8e7 d2e3 e8c8 c1b1 c6c5 c2c3 c5d4 c3d4 c8b8 g1f3 d7b6 g2g3 f6g7 d1c1 d8c8 a2a3 g5h4 g3h4 c8c1 h1c1 h8c8 c1g1 g7f8 e2d3 a7a6 d4d5 b6d7 g1g3 e6d5 e4d5');
INSERT INTO evaluations VALUES(33959,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(33959,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(33959,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(33959,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(33959,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33959,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(33959,36,'cp 14','b1c3 f8b4 d1d3 d7d6 e1c1 b8d7 g2g3 e8g8 f2f4 b7b5 g1e2 a8b8 f1g2 f6e7 e4e5 b4c3 d3c3 d7b6 b2b3 d6e5 f4e5 e7a3 c1b1 c8b7 g2b7 b8b7 c3c5 a3c5 d4c5 b6d5 b1b2 b5b4 e2f4 c7c6 f4d5 e6d5 c2c4 d5c4 b3c4 b7e7');
INSERT INTO evaluations VALUES(34056,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(34056,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(34056,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(34056,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(34056,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(34056,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(34056,36,'cp 41','f6e4 f1e2 d7d5 e1g1 e7d8 d2d4 f8d6 c2c4 c7c6 f1e1 e8g8 b1c3 h7h6 c4d5 e4c3 b2c3 c6d5 e2d3 b8d7 f3e5 d7f6 c1f4 c8e6 e1e3 f6e4 d1c2 a8c8 d3e4 d5e4 c2e4 d6e5 e4e5 d8d5 h2h4 d5e5 e3e5 f8d8 e5e3 c8c6 a2a4 d8c8 a1a3 a7a6 f2f3 e6d5 f4g3 f7f6 h4h5 c6c4');
INSERT INTO evaluations VALUES(34077,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(34077,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(34077,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(34077,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(34077,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(34077,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(34077,36,'cp 34','f1e2 d7d5 e1g1 e7d8 d2d4 f8d6 c2c4 c7c6 f1e1 e8g8 e2d3 f8e8 d1b3 b8d7 c4d5 c6d5 b3d5 d7f6 d5b3 c8e6 b3d1 e6d5 b1d2 a8c8 d2e4 f6e4 c1e3 e4f6 h2h3 h7h6 a2a3 a7a6 a1c1 c8c1 e3c1 e8e1 f3e1 d6c7 e1f3 d5f3 d1f3 d8d4 g2g3 c7d6 g1g2');
INSERT INTO evaluations VALUES(34078,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(34078,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(34078,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(34078,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(34078,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(34078,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(34078,36,'cp 37','d7d5 e1g1 e7d8 d2d4 f8d6 c2c4 c7c6 c4d5 c6d5 b1c3 e4c3 b2c3 e8g8 e2d3 b8d7 f1e1 h7h6 f3e5 d7f6 c1f4 c8e6 e1e3 f6e4 d1c2 d6e5 f4e5 a8c8 c2b2 d8b6 b2b6 a7b6 c3c4 e4f6 e5f6 g7f6 c4d5 e6d5 d3f5 c8c4 e3h3 f8a8 h3h6 c4d4 h6f6');
INSERT INTO evaluations VALUES(33841,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(33841,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(33841,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(33841,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(33841,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(33841,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(33841,36,'cp 55','e1g1 e7d8 d2d4 f8d6 c2c4 c7c6 b1c3 e8g8 f1e1 f8e8 c4c5 d6c7 d1b3 e4c3 b2c3 h7h6 c3c4 b7b6 c4d5 c6d5 c1d2 b8c6 e2b5 e8e1 a1e1 c6e7 b5d3 c8g4 d3c2 e7c6 b3d3 g7g6 c2d1 a8b8 d1a4 g4f3 g2f3 d8f6 a4c6 f6c6');
INSERT INTO evaluations VALUES(33840,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(33840,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(33840,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33840,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(33840,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(33840,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(33840,36,'cp 56','e7d8 d2d4 f8d6 c2c4 c7c6 f1e1 e8g8 b1c3 f8e8 c4c5 d6c7 d1b3 e4c3 b2c3 h7h6 c3c4 b7b6 c4d5 c6d5 c1d2 e8e6 a1c1 b8c6 e2b5 c6e7 b3a3 a8b8 c5b6 a7b6 f3e5 c7e5 d4e5 d5d4 f2f4 e6g6 b5d3');
INSERT INTO evaluations VALUES(50368,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(50368,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(50368,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(50368,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(50368,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(50368,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(50368,36,'cp -37','d7d5 d2d4 c7c5 c1f4 b8c6 e2e3 c8g4 f1b5 e7e6 h2h3 g4f3 d1f3 c5d4 e3d4 d8b6 b5c6 b7c6 e1c1 f8e7 g2g4 e8g8 h3h4 c6c5 d4c5 b6c5 g4g5 f6d7 c1b1 a8c8 f3e3 a7a5 h4h5 c5c6 e3a7 e7c5 a7a5 c5f2');
INSERT INTO evaluations VALUES(47887,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(47887,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(47887,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(47887,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(47887,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(47887,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(47887,36,'cp 0','c2c4 c7c5 b1c3 b8c6 f1e2 f8e7 e1g1 e8g8 a2a3 a7a5 c1e3 c8e6 e3d2 h7h6 a1b1 e6d7 h2h3 f6h7 c3d5 h7f6');
INSERT INTO evaluations VALUES(47825,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47825,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47825,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47825,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47825,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47825,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47825,36,'cp 0','c7c5 b1c3 b8c6 a1b1 c8d7 h2h3 f8e7 f1e2 e8g8 e1g1 a8b8 a2a3 a7a5 c1e3 h7h6 e3d2 f6e8 c3d5 e8f6');
INSERT INTO evaluations VALUES(45907,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45907,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45907,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45907,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45907,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(45907,36,'cp 0','b1c3 b8c6 a2a3 f8e7 f1e2 e8g8 e1g1 a7a5 c1e3 c8d7 b2b3 f6g4 e3c1 g4f6');
INSERT INTO evaluations VALUES(45905,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45905,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45905,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45905,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45905,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45905,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45905,36,'cp 0','b8c6 f1e2 f8e7 a2a3 e8g8 e1g1 h7h6 a1b1 b7b6 f3e1 c6d4 f2f4 e5f4 c1f4 f6d7 e2g4 c8b7 e1f3 d4f3 d1f3 d7e5 f4e5 d6e5 f1f2 d8d4 c3b5 d4d8');
INSERT INTO evaluations VALUES(13858,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13858,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13858,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(13858,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(13858,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13858,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(13858,36,'cp 8','a2a3 f8e7 a1b1 a7a5 f1e2 e8g8 e1g1 h7h6 h2h3 c8e6 c1d2 b7b6 b2b3 g8h7 c3d5 e6d7 b3b4 a5b4 a3b4 f6d5 c4d5 c6a7 b1a1 d8c7 d1c2 a7b5 c2b3 f8b8 b3b2 a8a1 f1a1 e7f6 e2f1 h7g8 a1b1 b8a8 f3e1 f6g5 d2g5 h6g5');
INSERT INTO evaluations VALUES(47447,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(47447,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(47447,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(47447,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(47447,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(47447,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(47447,36,'cp 74','g1f3 e7e6 c2c4 d5f6 b1c3 c7c5 c1e3 f6g4 e3g5 f8e7 g5e7 d8e7 f1e2 c5d4 f3d4 g4f6 d4b5 e8g8 d1d6 b8c6 e1c1 c8d7 e2f3 e7d8 d6c5 a7a6 b5d6 b7b6 c5e3 d8c7 d1d2 a8d8 h1d1 c6a5 b2b3 d7c6 f3c6 a5c6 h2h3 h7h6 c1b2 f6d7 c3a4 c6e5');
INSERT INTO evaluations VALUES(47445,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(47445,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(47445,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(47445,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(47445,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(47445,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(47445,36,'cp 85','e7e6 c2c4 d5f6 b1c3 c7c5 c1e3 c5d4 f3d4 a7a6 g2g3 e6e5 d4e2 d8d1 a1d1 b8c6 f1g2 c8e6 c3d5 e6d5 c4d5 f8b4 e2c3 c6a5 d5d6 a5c4 e1g1 c4e3 f2e3 a8d8 g2b7 d8d6 d1d6 b4d6 b7a6 e8e7 a2a3 d6c5 f1e1 h8d8 b2b4 c5a7 a6c4 d8d2');
INSERT INTO evaluations VALUES(48092,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(48092,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(48092,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(48092,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(48092,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(48092,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(48092,36,'cp 74','c2c4 d5f6 b1c3 c7c5 c1e3 b8d7 f1e2 f8e7 e1g1 e8g8 d4c5 d7c5 d1c1 c5d7 f1d1 b7b6 e3f4 f6h5 f4d6 c8b7 c1e3 h5f6 h2h3 h7h6 a2a3 f8e8 d6e7 d8e7 c3b5 e8d8 b2b4 f6e4 a1c1 a7a6 b5c3 e4c3 c1c3');
INSERT INTO evaluations VALUES(48083,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(48083,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(48083,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(48083,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(48083,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(48083,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(48083,36,'cp 69','d5f6 b1c3 c7c5 c1e3 b8d7 f1e2 f8e7 e1g1 e8g8 d4c5 d7c5 d1c2 b7b6 a1d1 d8c7 b2b4 c5d7 c3b5 c7b8 c4c5 b6c5 b4c5 f6g4 e3d4 e6e5 d4a1 d7c5 h2h3 a7a6 h3g4 a6b5 f3e5 b8b6 a1d4 b6e6 e2b5 e6a2 c2c3 a2b3 e5c6 e7d6 b5e2 b3c3 d4c3 c5e4 e2f3');
INSERT INTO evaluations VALUES(48477,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(48477,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(48477,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(48477,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(48477,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(48477,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(48477,36,'cp 84','b1c3 c7c5 c1e3 b8d7 f1e2 f8e7 e1g1 e8g8 d4c5 d7c5 f3e5 d8c7 e3f4 c7b6 d1c2 f8d8 a1d1 c8d7 e5d7 d8d7 d1d7 c5d7 f1d1 a8d8 g2g3 d7f8 d1d8 b6d8 f4e3 b7b6 c2a4 a7a5 a4c6 f8d7 a2a3 e7c5 e3c5 b6c5 c6d6 d8b8');
INSERT INTO evaluations VALUES(48475,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(48475,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(48475,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(48475,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(48475,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(48475,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(48475,36,'cp 74','c7c5 c1e3 b8d7 f1e2 f8e7 e1g1 e8g8 d4c5 d7c5 f3e5 c5e4 c3e4 f6e4 d1c2 f7f5 e5d3 b7b6 f1d1 c8b7 d3c5 d8c8 c5b7 c8b7 e2f3 a8d8 h2h3 h7h6 b2b3 g8f7 a1c1 b7c6 f3e4 c6e4 c2e4 f5e4 g1f1 e7c5 e3c5 b6c5 d1d8');
INSERT INTO evaluations VALUES(48281,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(48281,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(48281,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(48281,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(48281,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(48281,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48281,36,'cp 0','f8d6 g1f3 e8g8 f4g3 b7b6 b1d2 c8b7 f1d3 c7c5 d1e2 f6e4 f3e5 d8c7 e2h5 b8d7 e5d7 c7d7 g3d6 d7d6 h5h4 h7h6 d2e4 d5e4 d3e4 g7g5 h4g4 f7f5 g4h5 b7e4 f2f3 e4d5 h5g6 g8h8 e3e4 f5e4 g6h6 h8g8 h6g6 g8h8 f3e4 d6f4 g6h6 h8g8 h6g6');
INSERT INTO evaluations VALUES(41785,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41785,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41785,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41785,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41785,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41785,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41785,36,'cp 0','g1f3 e8g8 f4g3 b7b6 b1d2 c8b7 f1d3 c7c5 d1e2 f6e4 f3e5 d8c7 e2h5 b8d7 e5d7 c7d7 g3d6 d7d6 h5h4 h7h6 d2e4 d5e4 d3e4 g7g5 h4g4 f7f5 g4h5 b7e4 f2f3 e4d5 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(41786,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41786,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41786,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41786,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(41786,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41786,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41786,36,'cp 0','e8g8 f1d3 c7c5 e1g1 b8c6 f4g3 d6g3 h2g3 b7b6 d1e2 c8b7 b1d2 f8e8 f3e5 c6e5 d4e5 f6e4 d2e4 d5e4 d3b5 e8f8 a1d1 d8g5 d1d7 b7c8 d7d2 c8b7');
INSERT INTO evaluations VALUES(19990,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(19990,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(19990,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(19990,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(19990,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(19990,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(19990,36,'cp 174','d4d5 c6b8 b1c3 g7g6 h2h3 g8f6 c1f4 f8g7 f1e2 b8d7 a2a4 a7a6 e1g1 e8g8 d1d2 f6e8 a4a5 b7b5 a5b6 d8b6 a1a2 e8c7 f1a1 c8b7 e2c4 b6b4 f4h6 g7c3 d2c3 b4c3 b2c3 f8c8 a1b1 c8b8 c4d3 c7b5 h6d2 f7f5 c3c4 b5d4 f3d4 c5d4');
INSERT INTO evaluations VALUES(20050,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(20050,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(20050,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(20050,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(20050,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(20050,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(20050,36,'cp 169','c6b8 b1c3 g7g6 h2h3 f8g7 a2a4 g8f6 c1f4 e8g8 f1e2 b7b6 e1g1 b8d7 d1d2 d7e5 f3e5 d6e5 f4g5 f6e8 a4a5 f7f6 g5e3 c8d7 a1a3 e8d6 f1a1 a8b8 a5b6 a7b6 a3a7 d6c8 a7a6 c8d6 f2f3 f6f5 e3f2 b8b7 b2b3 g7f6 c3a4 d7a4 a1a4');
INSERT INTO evaluations VALUES(53109,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(53109,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(53109,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(53109,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(53109,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(53109,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(53109,36,'cp 175','b1c3 g7g6 f1e2 f8g7 h2h3 g8f6 a2a4 e8g8 c1f4 b7b6 e1g1 c8a6 f3d2 d8c7 d2c4 f6h5 f4d2 g7c3 b2c3 h5f6 f2f3 b8d7 d2g5 f8b8 e2d3 a6c4 d3c4 a7a6 d1e2 c7b7 f3f4 b6b5 a4b5 a6b5 a1a8 b7a8 c4b5');
INSERT INTO evaluations VALUES(53328,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(53328,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53328,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(53328,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(53328,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53328,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(53328,36,'cp 35','b8c6 a2a3 a7a6 f1e2 c8f5 c1e3 e7e6 b1d2 c5d4 c3d4 g8e7 g1f3 h7h6 a1c1 f5h7 g2g4 a8c8 e1g1 c6a7 c1c8 e7c8 f3e1 a7c6 f2f4 d8b6 b2b4 h6h5 f4f5 c6e5 f5e6 f7e6 g4h5 c8d6 e1g2 e5d7 d2f3 d6f5 f3g5 f8e7 g5h7 h8h7 e3f2');
INSERT INTO evaluations VALUES(19818,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(19818,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19818,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(19818,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19818,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(19818,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(19818,36,'cp 26','a2a3 a7a6 f1e2 c8f5 c1e3 e7e6 b1d2 g8e7 d4c5 c6e5 g1f3 e5f3 d2f3 e7c6 e1g1 f8e7 b2b4 f5e4 d1d2 e8g8 a1d1 d8c7 f3d4 e7f6 e3g5 f6g5 d2g5 h7h6 g5g3 c7g3 h2g3 e6e5 d4c6 b7c6 c3c4 f8d8 f2f4 a6a5 c4d5 a5b4 a3b4');
INSERT INTO evaluations VALUES(21082,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(21082,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(21082,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(21082,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(21082,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(21082,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(21082,36,'cp -21','e7e5 d2d3 c8e6 e1g1 g8f6 b1d2 a7a5 b2b3 f8e7 c1b2 d8d6 e2e3 e8g8 a2a3 h7h6 d1e2 f8d8 h2h3 d8e8 e3e4 d5e4 d3e4 d6c5 c2c3 b7b5 f1c1 c5b6 b3b4 e8d8 g2f1 d8b8 c3c4 a5b4 a3b4 a8a1 b2a1');
INSERT INTO evaluations VALUES(21763,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(21763,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(21763,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(21763,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(21763,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(21763,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(21763,36,'cp -12','d2d3 a7a5 e1g1 g8f6 b1c3 h7h6 e2e4 d5e4 d3e4 d8d1 f1d1 c8e6 c3b5 a8c8 b2b3 f8c5 f3e1 e8g8 c1a3 c5a3 b5a3 f8d8 a3b5 f6e8 b5c3 g8f8 f2f4 e6g4 d1d8 c8d8 g1f2 e5f4 g3f4 c6d4 c3d5 c7c6 d5e3 g4h5 h2h3 e8f6');
INSERT INTO evaluations VALUES(21751,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(21751,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(21751,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(21751,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21751,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(21751,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(21751,36,'cp -14','g8f6 e1g1 a7a5 b1c3 h7h6 e2e4 d5e4 d3e4 d8d1 f1d1 c8e6 b2b3 f8c5 c3b5 a8c8 f3e1 c6d4 e1d3 c7c6 b5c7 c8c7 d3c5 d4e2 g1f1 e2c1 a1c1 e8e7 f2f3 h8a8 f1g1 g7g5 g2f1 g5g4 c5e6 e7e6 f3g4 f6g4');
INSERT INTO evaluations VALUES(15188,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(15188,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(15188,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(15188,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(15188,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15188,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(15188,36,'cp -18','e1g1 a7a5 b1c3 h7h6 e2e4 d5e4 d3e4 d8d1 f1d1 c8e6 f3e1 f8c5 c3b5 a8c8 e1d3 c5b6 b5c3 c6d4 d1d2 f6d7 c3d5 b6a7 b2b3 e6g4 c1b2 f7f6 d3e1 d4e6 b2a3 c8a8 h2h3 g4h5 g3g4 h5g6 a1d1 d7f8 c2c3 c7c6 d5e7 g6h7');
INSERT INTO evaluations VALUES(15187,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(15187,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(15187,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(15187,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(15187,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(15187,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(15187,36,'cp -25','a7a5 b1c3 h7h6 f1e1 f8d6 e2e4 d5e4 d3e4 e8g8 a2a3 d6c5 c1e3 c5e3 e1e3 c8e6 h2h3 d8d1 a1d1 f8d8 d1d8 a8d8 e3d3 d8d3 c2d3 f6e8 c3b5 g8f8 d3d4 e6c4 d4d5 c4b5 d5c6 b5c6 f3e5 c6b5 f2f4 f8e7 g1f2 e8d6 g2f3 b7b6 f2e3 f7f6 e5g6 e7d7');
INSERT INTO evaluations VALUES(52808,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(52808,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(52808,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(52808,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52808,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(52808,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(52808,36,'cp 25','g1f3 e7e6 e4e5 c6c5 c2c3 b8c6 d3d4 c8d7 f1e2 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 a7a6 c1f4 f8e7 g2g4 f5h4 f3h4 e7h4 f4e3 f7f6 f2f4 e8g8 a1c1 f6e5 d4e5 h4e7 g4g5 d5d4 e3d4 c6d4 d1d4 d7c6 d4d8 a8d8');
INSERT INTO evaluations VALUES(52805,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(52805,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(52805,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(52805,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(52805,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(52805,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(52805,36,'cp 31','e7e6 a2a4 d8c7 b1d2 g8f6 e4e5 f6d7 d3d4 c6c5 c2c3 b8c6 f1d3 c5d4 c3d4 c6b4 d3b1 b7b6 e1g1 c8a6 f1e1 b4d3 b1d3 a6d3 d2b1 d3g6 b1c3 a7a6 f3h4 f8e7 h4g6 h7g6 g2g3 d7b8 c1e3 b8c6 h2h4 c7b7 a1c1 a8c8 d1g4 c6a5 c3e2 a5c4 e2f4 e8g8 h4h5 g6g5 f4h3 c4e3 f2e3');
INSERT INTO evaluations VALUES(53623,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(53623,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(53623,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(53623,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(53623,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(53623,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(53623,36,'cp 24','a2a4 c6c5 g2g3 b8c6 d1e2 f8e7 f1g2 g8f6 e1g1 a7a6 c1f4 e8g8 f1e1 h7h6 b1a3 c8d7 h2h4 f6h5 f4d2 h5f6 c2c3 d8b6 d2f4 a8c8 e4e5 f6h7 e1b1 f7f6 a3c2 g7g5 e5f6 g5f4 f6e7 c6e7 d3d4 c5d4 c2d4');
INSERT INTO evaluations VALUES(9346,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(9346,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(9346,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9346,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(9346,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(9346,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(9346,36,'cp 11','e1g1 a7a6 h2h3 e8g8 f1e1 h7h6 a2a4 c8e6 b1d2 c5a7 c4b3 f8e8 d2c4 c6e7 c1e3 c7c5 e3d2 e7g6 a4a5 f6h5 f3e5 g6e5 d1h5 e5d3 e1e3 d3f4 h5f3 f4g6 f3g3 g6e5 c4e5 e6b3 e5f3 b3e6');
INSERT INTO evaluations VALUES(55328,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(55328,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(55328,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(55328,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(55328,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(55328,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(55328,36,'cp -47','g8f6 b1c3 f8g7 g1e2 e8g8 e1g1 c8g4 f3e3 c7c6 f2f3 g4c8 c4b3 d6d5 e4d5 f6d5 e3f2 c8e6 c3d5 e6d5 d2d3 a7a5 b3d5 c6d5 f3f4 b8c6 f4e5 c6e5 c1f4 a8c8 f4e5 g7e5 d3d4 e5b8 c2c3 f7f5 a2a4 f5f4');
INSERT INTO evaluations VALUES(47574,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(47574,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(47574,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(47574,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(47574,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(47574,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(47574,36,'cp -37','b1c3 f8g7 g1e2 e8g8 e1g1 a7a6 a2a3 c8g4 f3e3 b7b5 c4a2 c7c5 d2d3 b8c6 c3d5 g4e6 e2c3 c6d4 e3d2 f6h5 c3e2 g7h6 d2h6 d4e2 g1h1 e2c1 h6c1 a8c8 c1d2 g8h8 d5c3 e6a2 c3a2 f7f5 a2c3 d8h4 h1g1 f8f7 f2f3 h5f4');
INSERT INTO evaluations VALUES(47568,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(47568,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(47568,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(47568,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(47568,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(47568,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(47568,36,'cp -50','f8g7 g1e2 e8g8 e1g1 a7a6 a2a3 c8g4 f3e3 b7b5 c4a2 c7c5 d2d3 b8c6 c3d5 g4e6 e2c3 c6d4 e3d2 f6h5 c3e2 g7h6 d2h6 d4e2 g1h1 e2c1 h6c1 a8c8 c1d2 g8h8 f2f3 f7f5 d5c3 e6d7 a3a4 b5b4 c3d5 a6a5 a2c4 d7c6 c2c3 c8b8 f1b1 c6d5 c4d5');
INSERT INTO evaluations VALUES(22372,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(22372,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(22372,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(22372,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(22372,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(22372,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(22372,36,'cp 93','c2c4 g7g6 f1e2 f8g7 b1c3 g8f6 e1g1 e8g8 c1e3 d8e7 d1c2 e5d4 f3d4 c7c6 a1d1 f8e8 e3g5 d7c5 f2f3 a7a5 f1e1 a5a4 e2f1 e7f8 g5e3 f6h5 c2f2 c8d7 f2d2 g7e5 g2g3 h5f6 e3f2 f6h5 d4c2 h5f6 f1g2 h7h5 g1h1');
INSERT INTO evaluations VALUES(22368,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22368,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(22368,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(22368,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(22368,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(22368,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(22368,36,'cp 77','g7g6 f1e2 f8g7 b1c3 g8f6 d4e5 d6e5 e1g1 e8g8 d1c2 c7c6 b2b4 f6h5 g2g3 d8e7 a1b1 f8e8 f1d1 h5f6 c1e3 f6g4 e3g5 g4f6 h2h3 a7a5 a2a3 a5b4 a3b4 d7f8 e2f1 f8e6 g5e3 h7h6 c4c5 f6h7 h3h4');
INSERT INTO evaluations VALUES(48205,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(48205,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(48205,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(48205,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(48205,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48205,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(48205,36,'cp -21','f8e7 e2e3 e8g8 a2a3 c7c5 d4c5 e7c5 f1d3 h7h6 g5h4 c5e7 d1e2 a7a6 h2h3 b8c6 e1g1 b7b5 h4g3 c8b7 e3e4 c6d4 f3d4 d5e4 c3e4 d8d4 c2c3 d4d5 f2f3 a8d8 a1d1 f6e4 f3e4 d5c5 g1h2 d8d7 e2f3 c5b6 d3c2 d7d1 f1d1');
INSERT INTO evaluations VALUES(40929,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(40929,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(40929,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(40929,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(40929,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(40929,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(40929,36,'cp -18','g5f6 e7f6 e2e4 e8g8 f1e2 c7c5 e4d5 c5d4 c3e4 e6e5 e4f6 d8f6 e1g1 f8d8 f1e1 c8f5 c2c3 d4c3 b2c3 b8d7 c3c4 b7b6 f3d2 f5g6 d1b3 h7h6 a1d1 d8e8 b3g3 d7c5 h2h4 a8d8 h4h5 g6f5 d2b3 c5b7 e2g4 f5c8 d1c1 c8g4 g3g4 g8h8 e1e3 b7d6');
INSERT INTO evaluations VALUES(40903,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(40903,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(40903,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(40903,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(40903,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(40903,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(40903,36,'cp -21','e7f6 e2e4 e8g8 d1d2 c7c5 d4c5 d5e4 c3e4 f6b2 a1b1 b2f6 e4f6 d8f6 f1d3 b8d7 c5c6 b7c6 e1g1 d7c5 d3c4 f6e7 f3e5 c8a6 c4a6 c5a6 d2a5 a6c5 f1d1 a8c8 a5a3 c8c7 d1d4 f7f6 d4c4 c5b7 a3e7 c7e7 e5c6 e7c7 c6a5 c7c4 a5c4 f8d8 g2g3 d8d7 b1b5 d7c7 g1g2 g7g6 c4e3');
INSERT INTO evaluations VALUES(41968,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(41968,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(41968,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(41968,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(41968,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(41968,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(41968,36,'cp -28','e2e4 e8g8 d1d2 c7c5 d4c5 d5e4 c3e4 f6b2 a1b1 b2f6 e4f6 d8f6 f1d3 f6e7 e1g1 b8d7 c5c6 b7c6 d2a5 d7c5 d3c4 c5b7 a5e5 b7d6 e5c5 d6f5 c5e5 f8d8 f1e1 f5d6 c4d3 c8d7 h2h4 h7h6 c2c4 d7e8 c4c5 d6b5 e5e4 f7f5 e4e3');
INSERT INTO evaluations VALUES(41969,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(41969,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(41969,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(41969,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(41969,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(41969,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(41969,36,'cp -13','e8g8 d1d2 c7c5 d4c5 d5e4 c3e4 f6b2 a1b1 b2f6 e4f6 d8f6 f1d3 b8d7 c5c6 d7c5 e1g1 b7c6 d3c4 f6e7 f3e5 c8a6 c4a6 c5a6 d2a5 a6c5 a5a3 f8e8 e5c6 e7d6 b1b5 e8c8 a3c5 d6c6 c5c6 c8c6 f1c1 a7a6 b5a5 c6c4 g2g3 a8a7 c2c3 g8f8 g1g2 f8e7 h2h4 h7h6 h4h5 c4c6 c3c4 a7c7 a2a3 c6c4 c1c4 c7c4 a5a6');
INSERT INTO evaluations VALUES(37436,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37436,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(37436,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(37436,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(37436,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37436,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37436,36,'cp 12','f1e2 c7c5 e1g1 d7d6 h2h3 d8b6 d1c1 b8c6 c2c3 c8e6 b1d2 a8c8 d4c5 b6c5 e3e4 c5a5 a2a4 a5c7 f1e1 a7a6 e2f1 f8d8 c1c2 h7h6 a1c1 f6d7 c2d1 g6g5 f4g3 c7b6 b2b4 d7e5 a4a5 b6c7 f3d4 c6d4 c3d4 c7c1 d1c1 c8c1 e1c1');
INSERT INTO evaluations VALUES(37435,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(37435,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37435,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37435,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37435,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37435,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37435,36,'cp 18','c7c5 e1g1 d7d5 c2c3 b8c6 h2h3 b7b6 a2a4 c8b7 b1d2 f6d7 a4a5 c6a5 b2b4 c5b4 c3b4 a5c4 d2c4 d5c4 e2c4 a7a5 b4a5 a8a5 a1c1 d8a8 c4e2 e7e6 d1b3 b7c6 b3b4 f8c8 b4e7 a8b7 f3g5 d7f6 e7b7 c6b7 c1c8 b7c8');
INSERT INTO evaluations VALUES(16343,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(16343,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(16343,32,'cp -186',NULL);
INSERT INTO evaluations VALUES(16343,33,'cp -186',NULL);
INSERT INTO evaluations VALUES(16343,34,'cp -196',NULL);
INSERT INTO evaluations VALUES(16343,35,'cp -190',NULL);
INSERT INTO evaluations VALUES(16343,36,'cp -217','c6e5 d2d4 e5g6 e4e5 f6g8 f1c4 c7c6 d1f3 d7d5 e5d6 c8e6 d4d5 g6e5 f3e2 e5c4 d5e6 c4d6 c1f4 d8f6 e6f7 e8f7 f4e5 f6e6 e1c1 g8f6 h1e1 a8d8 h2h3 f7g8 e2f3 e6f7 e1e3 f6e8 e5f4 d8a8 c1b1 h7h6 e3d3 f7g6 c3e2 d6c4 d3d7 b7b5 e2c3 e8f6 f3c6 f6d7 c6a8 g8h7');
INSERT INTO evaluations VALUES(16961,30,'cp -206',NULL);
INSERT INTO evaluations VALUES(16961,31,'cp -185',NULL);
INSERT INTO evaluations VALUES(16961,32,'cp -226',NULL);
INSERT INTO evaluations VALUES(16961,33,'cp -208',NULL);
INSERT INTO evaluations VALUES(16961,34,'cp -163',NULL);
INSERT INTO evaluations VALUES(16961,35,'cp -222',NULL);
INSERT INTO evaluations VALUES(16961,36,'cp -196','d2d4 e5g6 e4e5 f6g8 f1c4 c7c6 d1e2 f8b4 c4f7 e8f7 e2c4 d7d5 e5d6 c8e6 c4b4 b7b6 b4a3 h7h6 c1d2 g8f6 h2h4 h8e8 e1c1 d8d7 h4h5 g6f8 f2f3 a8d8 d2f4 f7g8 f4e5 d7f7 c1b1 f6d7 b2b3 a7a5 h1e1 f7h5 a3a4 h5f7 e1e3 c6c5 f3f4 e6f5 b1b2 g8h7 d4c5 d7c5 a4c6');
INSERT INTO evaluations VALUES(16956,30,'cp -202',NULL);
INSERT INTO evaluations VALUES(16956,31,'cp -211',NULL);
INSERT INTO evaluations VALUES(16956,32,'cp -209',NULL);
INSERT INTO evaluations VALUES(16956,33,'cp -212',NULL);
INSERT INTO evaluations VALUES(16956,34,'cp -206',NULL);
INSERT INTO evaluations VALUES(16956,35,'cp -200',NULL);
INSERT INTO evaluations VALUES(16956,36,'cp -209','e5g6 e4e5 f6g8 d1f3 c7c6 f1c4 d7d5 e5d6 c8e6 d4d5 g6e5 f3e2 e5c4 d5e6 c4d6 c1f4 d8f6 e6f7 e8f7 f4e5 f6e6 e1c1 g8f6 h1e1 a8d8 h2h3 f7g8 e2f1 e6f7 d1d4 d8d7 g2g4 f6d5 f1d3 d6e8 c3d5 d7d5 f2f4 d5d4 d3d4 e8f6 e1d1 h7h6 e5f6 f7f6 d4f6 g7f6 d1d7 h6h5 d7b7');
INSERT INTO evaluations VALUES(17066,30,'cp -217',NULL);
INSERT INTO evaluations VALUES(17066,31,'cp -219',NULL);
INSERT INTO evaluations VALUES(17066,32,'cp -211',NULL);
INSERT INTO evaluations VALUES(17066,33,'cp -236',NULL);
INSERT INTO evaluations VALUES(17066,34,'cp -236',NULL);
INSERT INTO evaluations VALUES(17066,35,'cp -236',NULL);
INSERT INTO evaluations VALUES(17066,36,'cp -236','e4e5 f6g8 f1c4 c7c6 d1e2 f8b4 c4f7 e8f7 e2c4 d7d5 e5d6 c8e6 c4b4 b7b6 c3e4 h7h6 e1g1 g8f6 f2f3 h8e8 b2b3 f7g8 c1b2 e6d5 c2c4 f6e4 f3e4 d5e4 f1f2 d8d7 a1e1 e4f5 e1f1 e8f8 b4d2 a8e8 d4d5 c6d5 d2d5 e8e6');
INSERT INTO evaluations VALUES(17064,30,'cp -219',NULL);
INSERT INTO evaluations VALUES(17064,31,'cp -229',NULL);
INSERT INTO evaluations VALUES(17064,32,'cp -212',NULL);
INSERT INTO evaluations VALUES(17064,33,'cp -214',NULL);
INSERT INTO evaluations VALUES(17064,34,'cp -216',NULL);
INSERT INTO evaluations VALUES(17064,35,'cp -217',NULL);
INSERT INTO evaluations VALUES(17064,36,'cp -199','f6g8 f1c4 c7c6 d1e2 f8b4 e1g1 d7d5 e5d6 e8f8 c4f7 b4c3 f7g8 f8g8 e2c4 g8f8 b2c3 d8d6 f2f4 d6d5 c4d5 c6d5 f4f5 g6e7 f5f6 g7f6 f1f6 f8g7 f6d6 h8e8 c1h6 g7g8 h6g5 c8f5 a1f1 f5c2 d6d7 e7f5 g2g4 e8e4 h2h3 c2a4 d7b7 f5d6 b7c7 e4e6 c3c4 e6g6 g5f4 d5c4 f4e5 a8f8 f1f8 g8f8 c7a7');
INSERT INTO evaluations VALUES(22994,30,'cp -206',NULL);
INSERT INTO evaluations VALUES(22994,31,'cp -229',NULL);
INSERT INTO evaluations VALUES(22994,32,'cp -216',NULL);
INSERT INTO evaluations VALUES(22994,33,'cp -236',NULL);
INSERT INTO evaluations VALUES(22994,34,'cp -237',NULL);
INSERT INTO evaluations VALUES(22994,35,'cp -226',NULL);
INSERT INTO evaluations VALUES(22994,36,'cp -216','f1c4 c7c6 d1e2 f8b4 e1g1 d7d5 e5d6 e8f8 c4f7 b4c3 f7g8 f8g8 e2c4 g8f8 b2c3 d8d6 f2f4 d6d5 c4d5 c6d5 f4f5 g6e7 f5f6 g7f6 f1f6 f8g7 f6d6 h8e8 c1h6 g7g8 h6g5 e7g6 h2h3 c8f5 a1f1 f5e4 d6d7 e8f8 g5h6 f8f1 g1f1 b7b6 g2g4 e4c2 f1f2 c2d3 d7g7 g8h8 g7d7 d3e4 f2g3');
INSERT INTO evaluations VALUES(22992,30,'cp -219',NULL);
INSERT INTO evaluations VALUES(22992,31,'cp -214',NULL);
INSERT INTO evaluations VALUES(22992,32,'cp -216',NULL);
INSERT INTO evaluations VALUES(22992,33,'cp -243',NULL);
INSERT INTO evaluations VALUES(22992,34,'cp -231',NULL);
INSERT INTO evaluations VALUES(22992,35,'cp -220',NULL);
INSERT INTO evaluations VALUES(22992,36,'cp -217','c7c6 d1e2 f8b4 e1g1 d7d5 e5d6 e8f8 c4f7 b4c3 f7g8 f8g8 e2c4 g8f8 b2c3 d8d6 f2f4 c8f5 g2g4 f5g4 f4f5 g6e7 c4b3 d6d5 b3b7 a8e8 c1a3 g4h3 f1f2 d5e4 b7b3 h7h5 b3c4 h5h4 f5f6 e4g6 g1h1 g7f6 a1e1 h8h7 c4c6 f8g8 c6f6 g6f6 f2f6 h7f7 f6f7 g8f7 c3c4 e7c6 e1e8 f7e8 a3c5 h3f1 d4d5 c6e5 h1g1 f1c4 c5a7 c4d5 a7c5 e5f7 c5b6 d5a2');
INSERT INTO evaluations VALUES(45611,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(45611,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(45611,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(45611,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(45611,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(45611,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45611,36,'cp 34','e2e3 h7h6 g5h4 f8e7 g1f3 b8d7 f1d3 d5c4 d3c4 e8g8 a2a4 c6c5 e1g1 c5d4 d1d4 b7b6 a1d1 c8b7 e3e4 d7c5 d4e3 d8c7 c3b5 c7c6 e4e5 f6g4 e3e2 e7h4 f3h4 c6e4 h4f3 f8d8 b5d6 e4e2 c4e2 b7f3');
INSERT INTO evaluations VALUES(45609,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(45609,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45609,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(45609,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(45609,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(45609,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(45609,36,'cp 13','h7h6 g5h4 f8e7 g1f3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 c4e2 b4e7 h2h3 a7a6 b5d4 c8d7 d4f5 e6f5 h4f6 d7e6 f6e7 c6e7 a2a3 e7c6 d1c1 f8d8 c1c2 g7g5 f1c1 g5g4 h3g4 f5g4 f3e1 a8c8');
INSERT INTO evaluations VALUES(45539,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(45539,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(45539,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45539,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(45539,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45539,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(45539,36,'cp 33','g5h4 f8e7 g1f3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 d1c1 c8d7 f1d1 f8c8 c4f1 g7g5 h4g3 b4c3 c1c3 c8c3 b2c3 b8c6 h2h4 g5g4 f3d2 d7e8 f1e2 a8d8 d2c4 f6e4 d1d8 c6d8 e2g4 e4c3 c4d6 e8c6 a2a3 c3b5 d6b5');
INSERT INTO evaluations VALUES(45543,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(45543,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45543,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45543,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45543,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(45543,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(45543,36,'cp 17','f8e7 g1f3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 b5d6 f8d8 d6e4 b4e7 d1d8 e7d8 f1d1 d8e7 e4f6 g7f6 a2a3 g8f8 f3d2 c6e5 c4e2 e5g6 h4g3 c8d7 e2f3 d7a4 f3b7 a8d8');
INSERT INTO evaluations VALUES(40036,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40036,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40036,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(40036,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40036,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(40036,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40036,36,'cp 29','g1f3 e8g8 f1e2 d5c4 e2c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 c4e2 a7a6 b5d4 c8d7 d4c6 d7c6 f3e5 a8c8 e5c6 c8c6 h4f6 g7f6 a2a3 b4d6 e2f3 c6b6 b2b4 f8c8 d1d3 g8f8 f1d1 f8e7 f3b7 d6h2 g1h2 b6b7 f2f3 b7c7 d1d2 f6f5 d3d6');
INSERT INTO evaluations VALUES(16016,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(16016,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(16016,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(16016,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(16016,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(16016,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(16016,36,'cp 68','f2f4 e7e5 g1f3 e5d4 d1d4 c7c6 c1e3 d6d5 e4e5 f8c5 d4d2 f6e4 c3e4 d5e4 f3d4 d7b6 e1c1 e8g8 h2h3 d8e7 g2g4 c8e6 d4f5 e6f5 g4f5 b6d5 e3d4 e4e3 d2e1 c5d4 d1d4 f7f6 e5e6 e7c5 d4c4 c5d6 h1g1 d5f4 e1e3 a8d8 f1d3 f4d3 c2d3 d6d3 e3d3 d8d3');
INSERT INTO evaluations VALUES(16021,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(16021,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(16021,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(16021,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(16021,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(16021,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(16021,36,'cp 106','e7e5 g1f3 e5d4 d1d4 c7c6 c1e3 d6d5 e4e5 f8c5 d4d2 f6e4 c3e4 d5e4 f3d4 d7b6 e1c1 e8g8 h2h3 b6d5 f1c4 a7a5 c4d5 d8d5 d4b3 d5d2 c1d2 c5b4 d2e2 a5a4 b3c5 b4c5 e3c5 f8e8 a2a3 h7h5 e2e3 c8e6 h1g1 a8a5 c5b4 a5d5 d1d5 c6d5 g2g4 h5g4 h3g4 e6d7 c2c4 d7c6 c4d5 c6d5');
INSERT INTO evaluations VALUES(15924,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(15924,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(15924,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(15924,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(15924,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(15924,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(15924,36,'cp 82','g1f3 e5d4 d1d4 c7c6 c1e3 d6d5 e4e5 f6g4 e3g1 d8b6 d4b6 a7b6 h2h3 g4h6 g2g4 h6g8 g1d4 h7h5 h1g1 f8c5 e1c1 g8e7 f1d3 h5g4 h3g4 h8h3 f3g5 c5d4 g5h3 d4g1 d1g1 d7c5 a2a3 d5d4 c3e4 c5d3 c2d3 e7g6 e5e6 f7e6 f4f5');
INSERT INTO evaluations VALUES(15923,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(15923,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(15923,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(15923,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(15923,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(15923,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(15923,36,'cp 56','e5d4 d1d4 c7c6 c1d2 d8b6 e1c1 f8e7 f1c4 e8g8 h2h3 d6d5 e4d5 b6d4 f3d4 d7b6 c4d3 f6d5 c3d5 b6d5 c2c4 d5b4 d3b1 c6c5 d4f3 c8e6 h1e1 f8e8 b2b3 a8d8 d2c3 d8d1 e1d1 e8d8 d1d8 e7d8 f3g5 d8g5 f4g5 b7b6 c1d2 b4c6');
INSERT INTO evaluations VALUES(15928,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(15928,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(15928,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(15928,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(15928,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(15928,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(15928,36,'cp 92','d1d4 c7c6 c1e3 d6d5 e4e5 f6g4 e3g1 d8b6 d4b6 a7b6 h2h3 g4h6 g2g4 h6g8 g1d4 h7h5 h1g1 f8c5 e1c1 g8e7 f1d3 h5g4 h3g4 h8h3 d4c5 d7c5 d1f1 b6b5 f4f5 b5b4 c3e2 c5d3 c2d3 a8a2 c1b1 b4b3 g1h1 h3h6 h1h6 g7h6 e2c3 a2a6 f1h1 c6c5 c3b5 a6a4 h1h6 a4g4 h6h8 g4g8 b5d6 e8d8 h8g8 e7g8 d6f7 d8c7 f7d6');
INSERT INTO evaluations VALUES(18689,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18689,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(18689,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18689,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18689,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(18689,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(18689,36,'cp 0','c6d4 b5c4 a7a6 c4b3 b7b5 f3d4 c5d4 c3e2 c8b7 d2d3 f8c5 c2c3 d4c3 b2c3 g8f6 e1g1 e8g8 h2h3 d8c7 a2a4 b7c6 c1g5 c5e7 g5f4 e7d6 f4e3 d6e7');
INSERT INTO evaluations VALUES(19125,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19125,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19125,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19125,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19125,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19125,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19125,36,'cp 0','b5c4 g8f6 e1g1 a7a6 c4b3 d7d6 f1e1 d4b3 a2b3 e6e5 d2d3 f8e7 f3d2 c8e6 d2f1 e8g8 f1e3 d8d7 e3c4 d7d8');
INSERT INTO evaluations VALUES(41712,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(41712,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41712,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(41712,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(41712,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(41712,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(41712,36,'cp 19','d1e2 d6e7 e1g1 e8g8 f1e1 f8e8 c1g5 c8e6 b1d2 b8d7 d2f1 h7h6 g5h4 c7c6 h2h3 e7d6 e2d2 d8b6 c2c3 f6h5 f1g3 h5f4 g3f5 f4d3 f5d6 d3e1 a1e1 a7a5 d2c2 d7f8 h4g3 f8g6 d6e8 a8e8 e1e3 e6d7 e3e8 d7e8');
INSERT INTO evaluations VALUES(41710,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41710,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41710,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(41710,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(41710,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(41710,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(41710,36,'cp 18','c8e6 f3g5 d8e7 g5e6 f7e6 c2c3 e8g8 b1d2 b8d7 d2f3 e6e5 d4e5 d7e5 f3e5 d6e5 e1g1 a8e8 g2g3 e7d7 c1e3 f6g4 e2c2 g7g6 a1e1 e5d6 c2d2 g4e3 e1e3 e8e3 d2e3 b7b6 g1g2 d7c6 f2f4 d5d4 e3e4 c6e4 d3e4 d4c3 b2c3 g8f7 g2f3 f7f6 e4d3 f6f7 f1e1 f8e8');
INSERT INTO evaluations VALUES(28444,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(28444,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(28444,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28444,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28444,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28444,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28444,36,'cp 23','f3g5 d8e7 g5e6 f7e6 b1d2 b8c6 d2f3 e6e5 d4e5 c6e5 f3e5 d6e5 e1g1 e8g8 g2g3 a8e8 c2c3 e7d7 c1e3 f6g4 e2c2 g7g6 a1e1 e5d6 c2d2 e8e5 h2h4 c7c6 g1g2 g4e3 e1e3 d7f7 f2f4 e5e6 e3e2 e6e2 d3e2 f7e6 e2d3 f8e8');
INSERT INTO evaluations VALUES(12595,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(12595,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(12595,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(12595,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(12595,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(12595,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(12595,36,'cp -18','f8b4 f1c4 b4c3 b2c3 e8g8 e1g1 d7d6 d1d3 h7h6 f1d1 d8a5 a1b1 f8d8 h2h3 a5c7 c1a3 f6e8 f3d2 c6a5 c4d5 c8e6 c3c4 a8c8 a3b4 a5c4 d3c4 c7c4 d2c4 e6d5 c4d6 e8d6 d1d5 d6c4 b1d1 d8e8 g1f1 c4b6 d5a5 a7a6 a5c5 c8c5 b4c5 b6a4 c5e3 e8c8');
INSERT INTO evaluations VALUES(7863,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(7863,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(7863,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(7863,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(7863,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(7863,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(7863,36,'cp -13','f1c4 b4c3 b2c3 e8g8 e1g1 d7d6 d1d3 f8e8 c1g5 h7h6 g5f6 d8f6 f1b1 b7b6 c4b5 c8b7 f3d2 e8d8 b5c6 b7c6 c3c4 f6e7 a2a4 e7c7 d2f1 c6d7 f1e3 d7e6 b1b4 a8c8 a4a5 b6a5 b4a4 d8d7 g2g3 c7c6 a4a5 e6c4 e3c4 c6c4 a5a7 d7a7 a1a7 c4d3 c2d3 c8c3 d3d4');
INSERT INTO evaluations VALUES(7867,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(7867,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(7867,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(7867,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(7867,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(7867,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(7867,36,'cp -17','b4c3 b2c3 d7d6 d1d3 e8g8 e1g1 f8e8 f1d1 c8e6 c1g5 a8c8 c4e6 e8e6 f3h4 c6e7 a1b1 b7b6 h4f5 e7f5 e4f5 e6e8 h2h3 d6d5 g5f6 d8f6 d3d5 e5e4 d5d7 e4e3 f2e3 f6c3 d1d3 c3c2 b1e1 h7h6 e3e4 c2c5 g1h1 c5c4 d3d4');
INSERT INTO evaluations VALUES(21297,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(21297,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(21297,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(21297,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(21297,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(21297,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(21297,36,'cp 98','d4e5 c8g4 e5d6 f8d6 h2h3 g4h5 c4b5 g8e7 b1c3 a7a6 b5e2 d8d7 c1e3 e8c8 f3d4 h5e2 d1e2 d6e5 d4b3 f7f5 e1g1 f5f4 e3c5 d7e6 c5e7 e6e7 c3d5 e7f7 f1d1 c8b8 c2c3 e5d6 b3d4 c6e5 d5b4 g7g5 b4d3 e5g6 d4f3 h8e8 e4e5 f7g7 g1h1');
INSERT INTO evaluations VALUES(21272,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(21272,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(21272,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(21272,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(21272,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(21272,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(21272,36,'cp 77','c8g4 e5d6 f8d6 h2h3 g4h5 c4b5 g8e7 b1c3 a7a6 b5e2 d8d7 c1e3 e8c8 f3d2 h5e2 d1e2 d7e6 e1c1 d6e5 e2c4 e6c4 d2c4 e5c3 b2c3 f7f5 d1d8 h8d8 e4f5 e7f5 h1e1 f5h4 g2g3 h4f3 e1d1 g7g5 d1d8 c8d8 c1d1 d8e7 c4d2 c6e5 e3d4 e7e6 d4e5');
INSERT INTO evaluations VALUES(26727,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(26727,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(26727,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(26727,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(26727,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(26727,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(26727,36,'cp 90','e5d6 f8d6 h2h3 g4h5 c4b5 g8e7 b1c3 a7a6 b5e2 d8d7 c1e3 e8c8 f3d2 h5e2 d1e2 d7e6 d2b3 d6b4 e1g1 f7f5 f1d1 e6g6 d1d8 h8d8 e3d2 f5e4 e2e4 g6e4 c3e4 b4d2 b3d2 c6b4 c2c3 b4d3 b2b3 d3f4 g1f1 e7d5');
INSERT INTO evaluations VALUES(26725,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(26725,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(26725,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(26725,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(26725,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(26725,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(26725,36,'cp 87','f8d6 c4b5 g8e7 h2h3 g4h5 b1c3 a7a6 b5e2 d8d7 c1e3 e8c8 f3d2 h5e2 d1e2 d7e6 e1c1 d6e5 e2c4 e6c4 d2c4 e5c3 b2c3 f7f5 d1d8 h8d8 e4f5 e7f5 e3d2 f5h4 c4e3 g7g5 h1f1 d8f8 c1d1 b7b6 f2f3 h4g6 g2g3 g6e5 d1e2 f8f7');
INSERT INTO evaluations VALUES(23823,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(23823,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(23823,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(23823,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(23823,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(23823,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(23823,36,'cp 101','h2h3 g4h5 c4b5 g8e7 b1c3 a7a6 b5e2 d8d7 c1e3 e8c8 f3d2 h5e2 d1e2 d7e6 d2b3 d6b4 e1g1 f7f5 f1d1 f5e4 b3c5 e6g6 d1d8 h8d8 c3e4 d8e8 e3d2 b4d2 e2d2 e8d8 d2e2 c6d4 e2d3 e7f5 e4g3 g6c6 g3f5 c6c5 f5d4 c5d4 a1d1 d4d3 d1d3');
INSERT INTO evaluations VALUES(44074,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(44074,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(44074,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(44074,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(44074,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(44074,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(44074,36,'cp 48','d2d4 e5d4 d1d4 f8e7 g1f3 e8g8 a2a4 d8b6 e1g1 b6d4 f3d4 f6e4 d4f5 e7d8 c3e4 d7d5 f5d6 d8c7 d6c8 f8c8 c4d3 d5e4 d3e4 b8a6 a4a5 c8e8 e4f3 a8d8 c1e3 a6b4 f1c1 a7a6 e3c5 b4d5 c1d1 f7f5 g2g3 d5f6 d1d8 e8d8 f3e2 g7g6 c5e3 c7e5 c2c3 d8d7 g1g2 f6d5 e3a7 g8f7');
INSERT INTO evaluations VALUES(44088,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(44088,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(44088,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(44088,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(44088,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(44088,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44088,36,'cp 67','e5d4 d1d4 f8e7 a2a4 e8g8 g1f3 d8b6 e1g1 b6d4 f3d4 f6e4 d4f5 e7d8 c3e4 d7d5 f5d6 d8c7 d6c8 f8c8 c4d3 d5e4 d3e4 b8a6 a4a5 c8e8 e4f3 a8d8 c1e3 a6b4 f1c1 a7a6 e3c5 b4d5 g2g3 c7d6 c5a7 d6b8 a7b8 d8b8 c1d1 b8d8 g1f1 g8f8 f3d5 d8d5');
INSERT INTO evaluations VALUES(39230,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39230,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39230,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39230,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(39230,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(39230,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(39230,36,'cp -13','e5d4 f1c4 d7d6 f3d4 d8f6 d4f3 c8e6 c4e6 f6e6 e1g1 b8c6 b1c3 e8c8 a2a3 h7h6 b2b4 c5b6 c3d5 g8f6 a3a4 f6e4 c2c4 c6e5 f3d4 b6d4 d1d4 e4f6 c1e3 e5c6 d4d1 h8e8 a4a5 e6g4 d5f6 g4d1 f1d1 g7f6 b4b5');
INSERT INTO evaluations VALUES(39302,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39302,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39302,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39302,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39302,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39302,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39302,36,'cp -4','f1c4 d7d6 f3d4 d8f6 d4f3 c8e6 c4e6 f6e6 e1g1 g8e7 b1c3 b8c6 f1e1 e8c8 a2a3 c6e5 b2b4 e5f3 d1f3 c5d4 c1b2 e6f6 f3f6 d4f6 c3d1 f6b2 d1b2 h8e8 g1f1 e7c6 a1d1 a7a5 c2c3 b7b6 b2c4 g7g6 f2f3 c8b7 c4e3 a5b4 a3b4 d8a8');
INSERT INTO evaluations VALUES(39292,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39292,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39292,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39292,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39292,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39292,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39292,36,'cp 0','d7d6 f3d4 d8f6 d4f3 c8e6 c4e6 f6e6 e1g1 g8e7 b1c3 b8c6 c3d5 e8c8 f3g5 e6g6 d1g4 c8b8 d5f4 g6f6 f4h5 f6g6');
INSERT INTO evaluations VALUES(38740,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38740,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38740,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38740,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38740,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38740,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38740,36,'cp 0','f3d4 d8f6 d4f3 c8e6 c4e6 f6e6 b1c3 b8c6 e1g1 e8c8 a2a3 g8e7 f1e1 c6e5 b2b4 e5f3 d1f3 c5d4 a1b1 e6f6 f3f6 d4f6 c1d2 a7a6 g2g3 b7b6 h2h4 g7g6 b1d1 c8b7 g1g2 h8e8 a3a4 f6g7 g2h3 g7d4 h3g2');
INSERT INTO evaluations VALUES(38738,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38738,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38738,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38738,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38738,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38738,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38738,36,'cp 0','d8f6 d4f3 c8e6 c4e6 f6e6 b1c3 b8c6 e1g1 g8e7 c3d5 e8c8 f3g5 e6g6 d1g4 c8b8 d5f4 g6f6 f4h5 f6g6');
INSERT INTO evaluations VALUES(17777,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(17777,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(17777,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(17777,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(17777,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(17777,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(17777,36,'cp 194','b5c6 b7c6 d4e5 f7f6 e5d6 f8d6 c1e3 g8e7 d1d4 c8e6 b1d2 e8g8 a2a3 e6f7 d4c3 e7g6 e1c1 c6c5 h1e1 d8e7 g2g3 a6a5 d2b3 f8b8 b3a5 g6e5 f3d2 e7e6 f2f4 e6a2 d2b3 c5c4 f4e5 d6e5 e3d4 c4b3 d4e5 f6e5 a5c6 a2a1 c1d2');
INSERT INTO evaluations VALUES(17747,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(17747,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(17747,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(17747,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(17747,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(17747,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(17747,36,'cp 199','b7c6 d4e5 g8e7 e5d6 d8d6 d1d6 c7d6 b1c3 e7g6 c1e3 f8e7 f3d4 c8d7 e1c1 h7h5 h2h3 e8c8 d4f3 d8e8 d1d4 c6c5 d4d1 d7c6 e3g5 f7f6 f3h4 g6h4 g5h4 c8d7 f2f3 g7g5 h4f2 d7e6 f2g3 e8g8 h1e1 g5g4 h3g4 h5g4 f3f4 g8d8 f4f5 e6f7 c3d5 c6d5 d1d5');
INSERT INTO evaluations VALUES(18007,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(18007,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(18007,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(18007,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(18007,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(18007,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(18007,36,'cp 223','d4e5 g8e7 e5d6 d8d6 d1d6 c7d6 b1c3 e7g6 c1e3 f8e7 f3d4 c8d7 e1c1 h7h5 h2h3 e8c8 d4f3 d8e8 d1d4 f7f5 d4a4 f5e4 c3e4 c8c7 a4a6 e8a8 a6a8 h8a8 a2a3 c6c5 e4c3 d7c6 h1d1 c6f3 g2f3 c7c6 f3f4 g6h4 d1d3 h4f5 c3d5 a8b8 d5e7 f5e7');
INSERT INTO evaluations VALUES(18003,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(18003,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(18003,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(18003,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(18003,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(18003,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(18003,36,'cp 201','g8e7 e5d6 d8d6 d1d6 c7d6 b1c3 e7g6 c1e3 f8e7 f3d4 c8d7 h2h3 e8c8 e1c1 h7h5 d4f3 d8e8 d1d4 c6c5 d4d2 f7f6 d2d1 d7c6 f3d2 c8c7 d2c4 g6e5 c4e5 f6e5 f2f3 g7g5 b2b3 c7b7 a2a4 e8f8 c1b2 g5g4 h3g4 h5g4 f3g4 h8h1 d1h1');
INSERT INTO evaluations VALUES(50208,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50208,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(50208,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(50208,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50208,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(50208,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50208,36,'cp 0','d7d5 c1f4 f8g7 e2e3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(46785,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46785,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46785,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46785,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46785,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46785,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46785,36,'cp 0','c1f4 f8g7 e2e3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(55185,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55185,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(55185,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55185,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(55185,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(55185,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(55185,36,'cp 22','c7c5 g1f3 b8c6 d2d3 a7a6 c1e3 e7e6 d3d4 c5d4 f3d4 c8d7 d1e2 d8c7 e1c1 g8f6 c1b1 f8b4 c3d5 f6d5 c4d5 e8g8 d4c6 d7c6 e2c4 e6d5 c4b4 a8c8 e3b6 c7e5 b6d4 e5c7 d1d2 f8e8 b2b3 c6d7 b4b6 c7c6 h1d1 f7f6 b6c6 c8c6');
INSERT INTO evaluations VALUES(34323,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(34323,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(34323,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34323,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(34323,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(34323,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(34323,36,'cp 87','c1d2 g8f6 g1f3 c8f5 f1c4 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 d7b6 c4b3 a7a5 a2a4 f8b4 d2g5 b4c3 b2c3 e8c8 c1b2 d8d7 e2f3 c6c5 d4c5 c7c5 d1d7 b6d7 h1d1 h8d8 d1d4 c5b6 g5e3 d7c5 d4d8 c8d8 e3c5 b6c5 f3b7 f6d7 b7f3 d8e7 f3d3 c5b6 b2a2 b6c5 f2f3');
INSERT INTO evaluations VALUES(34324,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(34324,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(34324,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(34324,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(34324,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(34324,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(34324,36,'cp 76','g8f6 h2h3 a5b6 g1f3 c8f5 f1c4 e7e6 d1e2 f5g6 e1c1 b8d7 f3h4 b6c7 c4b3 f8d6 c1b1 d6f4 d2f4 c7f4 h4g6 h7g6 g2g3 f4c7 h3h4 d7b6 c3e4 f6e4 e2e4 b6d5 c2c3 d5f6 e4f3 a8d8 a2a3 c7b6 b1a2 e8g8');
INSERT INTO evaluations VALUES(33589,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(33589,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(33589,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(33589,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(33589,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(33589,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(33589,36,'cp 68','h2h3 a5b6 g1f3 c8f5 f1c4 b8d7 d1e2 e7e6 e1c1 f5g6 f3h4 b6c7 c1b1 b7b5 c4b3 d7b6 g2g4 e8c8 d2g5 a7a5 a2a3 f8e7 h1e1 f6d5 c3d5 b6d5 g5e7 c7e7 h4g6 h7g6 a3a4 e7b4 e2f3 h8f8 a4b5 c6b5 b3d5 d8d5 h3h4 c8b7');
INSERT INTO evaluations VALUES(56459,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(56459,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(56459,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(56459,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(56459,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(56459,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(56459,36,'cp -32','g8f6 g1f3 f8d6 c2c4 c7c6 e1g1 e8g8 c4d5 c6d5 b1c3 h7h6 e2e4 d5d4 c3b5 c8g4 h2h3 g4f3 d1f3 d6b4 h3h4 b8d7 g2h3 f8e8 a2a3 b4f8 a3a4 d7c5 f3d1 a7a6 b5a3 b7b5 a4b5 a6b5 b2b4 c5a6 a3b5 f8b4 c1d2 d8e7 a1b1 b4d2 d1d2 e8d8 f1c1 a6c5 d2c2');
INSERT INTO evaluations VALUES(48872,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(48872,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48872,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48872,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(48872,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48872,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(48872,36,'cp -12','g1f3 f8d6 e1g1 e8g8 c2c4 d5d4 e2e3 c7c5 e3d4 e5d4 c1g5 b8d7 b1a3 f8e8 b2b4 d6f8 a3c2 c5b4 c2d4 h7h6 g5e3 f6g4 e3c1 d7c5 a2a3 g4e5 f3e5 d8d4 c1e3 d4e5 d3d4 e5f5 d4c5 f8c5 a3b4 c5e3 f2e3 f5e5 d1d4 e5e3 d4e3 e8e3');
INSERT INTO evaluations VALUES(48871,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(48871,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(48871,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(48871,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(48871,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48871,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(48871,36,'cp -28','f8d6 e2e4 d5e4 d3e4 e8g8 d1e2 b8a6 b1c3 c7c6 f3h4 a6c5 h4f5 c5e6 c1e3 d6c5 e3c5 e6c5 e1g1 d8b6 b2b3 c5e6 a1d1 g7g6 e2d2 f8d8 d2h6 d8d1 c3d1 b6d8 f5e3 b7b6 d1b2 c8a6 f1d1 d8e7 e3c4 e6d4');
INSERT INTO evaluations VALUES(41737,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(41737,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(41737,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(41737,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(41737,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(41737,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(41737,36,'cp -15','c2c4 c7c6 e1g1 e8g8 c4d5 c6d5 b1c3 h7h6 e2e4 d5d4 c3b5 c8g4 h2h3 g4f3 d1f3 d6e7 c1d2 b8a6 f1c1 d8b6 b5a3 e7a3 b2a3 b6d6 a3a4 f8c8 g2f1 c8c1 a1c1 a6b4 f3f5 b4c6 f2f4 f6d7 f5g4 d6e7 a4a5 d7f6 g4e2');
INSERT INTO evaluations VALUES(36577,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(36577,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(36577,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36577,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36577,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36577,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(36577,36,'cp 25','c8g4 d3e2 c7c5 d4d5 e7e6 e1g1 e6d5 e4d5 b8d7 a2a4 f6e8 a1a3 a7a6 c1g5 e8f6 h2h3 g4f3 e2f3 d8c7 g2g3 a8e8 g1g2 h7h6 g5d2 e8e7 d1c1 h6h5 c1c2 f6h7 a3a1 f8e8 a1e1 e7e1 f1e1 e8e1 d2e1 h5h4 g3h4 c7d8 h4h5 d8g5 f3g4 f7f5');
INSERT INTO evaluations VALUES(27855,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(27855,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(27855,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(27855,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(27855,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(27855,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(27855,36,'cp 32','d3e2 c7c5 d4d5 e7e6 e1g1 e6d5 e4d5 b8d7 c1d2 d7b6 d1b3 f8e8 h2h3 g4f5 a1e1 f6e4 c3e4 f5e4 f3g5 h7h6 g5e4 e8e4 b3c2 d8e7 e2d1 a8e8 b2b3 e4e1 f1e1 e7f8 e1e8 f8e8 g2g3 b6d7 d2f4 e8e7 c2e2 e7f8 g1g2 g7d4 e2d2 g6g5 f4e3 d4e3 d2e3');
INSERT INTO evaluations VALUES(49354,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(49354,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49354,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49354,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(49354,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49354,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(49354,36,'cp -19','h7h6 g5f6 d8f6 b1c3 f8b4 a1c1 c7c5 g1f3 c5d4 f3d4 b8c6 e2e3 c6d4 e3d4 b7b5 a2a3 b4c3 c1c3 b5c4 f1c4 c8b7 e1g1 a8c8 h2h3 c8c7 d4d5 e8g8 d1d2 e6d5 c4b3 c7c3 b2c3 f6e5 f1d1 f8c8 b3d5 b7d5 d2d5 e5d5 d1d5 c8c3');
INSERT INTO evaluations VALUES(49230,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(49230,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(49230,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(49230,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(49230,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49230,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49230,36,'cp -5','g5f6 d8f6 b1c3 f8b4 a1c1 c7c5 g1f3 c5d4 f3d4 b8c6 e2e3 e8g8 a2a3 b4c3 c1c3 f8d8 d1f3 f6e7 f1e2 d7d5 c4d5 c6d4 e3d4 d8d5 e1g1 e7d6 f1d1 c8d7 e2c4 d5f5 f3g3 d6g3 h2g3 d7c6 c4d3 f5g5 d3c4 a8d8 f2f4 g5g6');
INSERT INTO evaluations VALUES(49227,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49227,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49227,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49227,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49227,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49227,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49227,36,'cp 0','d8f6 b1c3 f8b4 a1c1 c7c5 g1f3 e8g8 e2e3 c5d4 f3d4 b8c6 a2a3 b4c3 c1c3 f8d8 d1f3 f6e5 f3g3 e5f6');
INSERT INTO evaluations VALUES(33957,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33957,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33957,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33957,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33957,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33957,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33957,36,'cp 0','b1c3 f8b4 a1c1 c7c5 g1f3 c5d4 f3d4 b8c6 e2e3 c6d4 e3d4 b7b5 a2a3 b4c3 c1c3 b5c4 f1c4 c8b7 e1g1 e8g8 d1d2 f8c8 h2h3 a7a5 f1c1 a5a4 d4d5 f6e5 c1e1 e5f6');
INSERT INTO evaluations VALUES(48360,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(48360,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48360,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(48360,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(48360,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(48360,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(48360,36,'cp 32','f8e7 c2c4 e8g8 e1g1 d5c4 d1c2 a7a6 a2a4 c8d7 c2c4 d7c6 c1g5 c6d5 c4c2 d5e4 c2c1 h7h6 g5f6 e7f6 b1c3 e4h7 f1d1 b8c6 d4d5 e6d5 c3d5 d8d6 f3d2 c6d4 c1c4 b7b5 d5f6 d6f6 d2e4 h7e4 c4d4 f6d4 d1d4 e4g2 g1g2');
INSERT INTO evaluations VALUES(41005,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41005,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(41005,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(41005,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(41005,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41005,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(41005,36,'cp 32','e1g1 e8g8 c2c4 d5c4 d1c2 a7a6 a2a4 c8d7 c2c4 d7c6 c1g5 c6d5 c4c2 d5e4 c2c1 b8c6 b1d2 c6d4 f3d4 e4g2 d4e6 f7e6 g1g2 d8d5 d2f3 c7c5 f1d1 d5c6 c1c2 c5c4 e2e4 a8e8 a1c1 f6e4 g5e7 e8e7 c2c4');
INSERT INTO evaluations VALUES(41004,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(41004,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(41004,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(41004,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(41004,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(41004,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41004,36,'cp 21','b7b5 a2a4 c7c6 a4b5 c6b5 b1c3 d8b6 f3e5 c8b7 e2e4 f6e4 c3e4 d5e4 d1e2 e8g8 g2e4 e7f6 c2c3 f6e5 d4e5 b6c7 f1d1 h7h6 d1d6 b8d7 e2b5 b7e4 d6d7 a7a6 d7c7 a6b5 c7a7 a8a7 a1a7 f8a8 a7a8 e4a8 c1e3 a8e4');
INSERT INTO evaluations VALUES(21901,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(21901,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(21901,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(21901,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(21901,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(21901,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21901,36,'cp 52','f7f5 d2d3 d8f6 d1b3 g8e7 c1g5 f6g6 b1d2 a7a6 e1g1 b7b5 c4d5 e7d5 e4d5 h7h6 g5h4 c6e7 a2a4 g6f7 h4e7 f8e7 a4b5 a8b8 b3c4 b8b5 c4c7 e8g8 c3c4 b5b7 c7c6 b7b2 f1b1 b2b1 a1b1 e5e4 f3e1 a6a5 b1b8 e7g5 d2b3 c8d7 c6d6 f8b8 d6b8 g8h7');
INSERT INTO evaluations VALUES(22324,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(22324,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(22324,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(22324,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(22324,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(22324,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(22324,36,'cp 50','d2d3 f5f4 b2b4 d8f6 d1b3 g8h6 a2a4 h6f7 d3d4 f8e7 h2h4 e8g8 c1b2 g8h8 b1d2 h7h5 d4d5 c6b8 c4e2 c8g4 b3c4 c7c6 f3g5 g4e2 c4e2 f6g6 a4a5 f8c8 c3c4 b8d7 a1a3 h8g8 a3h3 b7b6 a5a6 c6d5 e4d5 e7g5 h4g5');
INSERT INTO evaluations VALUES(31795,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(31795,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(31795,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(31795,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(31795,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(31795,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(31795,36,'cp 76','g1f3 g8f6 b1c3 c7c6 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 d1c2 e8g8 f1d1 h7h6 g5e4 f6d5 e4f2 d5c3 b2c3 b7b5 c4d3 a7a5 c3c4 b5c4 c2c4 d8c7 c4c3 e7d6 g1h1 c8a6 c1a3 b8d7 a1c1 a6d3 a3d6 c7d6');
INSERT INTO evaluations VALUES(31796,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(31796,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(31796,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(31796,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(31796,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(31796,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(31796,36,'cp 69','g8f6 b1c3 c7c6 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 d1c2 e8g8 f1d1 h7h6 g5e4 f6e4 c3e4 b8d7 c1d2 d7f6 e4f6 e7f6 c2e4 a7a5 a1c1 c8d7 a2a3 g7g6 e4c2 f6g7 d2e1 d8e7 e3e4 b7b5 c4b3 f8b8 e1g3 b5b4 g3b8 a8b8 c2c5 e7c5 d4c5 g7b2');
INSERT INTO evaluations VALUES(29475,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(29475,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(29475,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(29475,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(29475,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(29475,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(29475,36,'cp 71','b1c3 c7c6 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 d1c2 e8g8 f1d1 h7h6 g5e4 f6e4 c3e4 b8d7 c1d2 a7a5 a2a3 d7f6 c4a2 f6e4 c2e4 c8d7 a1c1 e7d6 e4d3 d8h4 f3f4 f7f5 g2g3 h4g4 a2b1 g4f3 d2c3 a5a4 e3e4 f3d3 b1d3 f5e4 d3e4');
INSERT INTO evaluations VALUES(29473,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(29473,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(29473,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(29473,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(29473,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(29473,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(29473,36,'cp 71','c7c6 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 e3e4 h7h6 g5h3 c6c5 d4d5 e6d5 c4b5 c8d7 e4e5 f6h7 d1d5 d7b5 c3b5 d8b6 d5c4 b8c6 b5d6 e7d6 e5d6 e8g8 c1e3 b6b2 e3c5 c6e5 c4a4 e5d3 c5d4 b2b4 a4b4 d3b4 f1d1 b4c6 d4c3 f8d8 a1b1 d8d7 h3f2 h7f6 c3f6 g7f6');
INSERT INTO evaluations VALUES(28928,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(28928,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(28928,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(28928,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(28928,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(28928,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(28928,36,'cp 80','f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 e3e4 h7h6 g5h3 c6c5 d4d5 e6d5 c4b5 c8d7 e4e5 f6h7 d1d5 d7b5 c3b5 d8b6 d5c4 b8c6 b5d6 e7d6 e5d6 e8g8 c1e3 b6b2 e3c5 c6e5 c4b3 b2b3 a2b3 b7b6 c5a3 h7f6 h3f4 a7a5 f1d1 f8c8 a3b2 e5d7 a1c1 c8c1 d1c1');
INSERT INTO evaluations VALUES(39580,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39580,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39580,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39580,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39580,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39580,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39580,36,'cp 0','g8f6 f1e2 c7c5 c2c3 e8g8 b1d2 c5d4 e3d4 d7d6 g1f3 d8c7 e1g1 b8c6 d2c4 f6d5 f4g3 b7b5 c4d2 b5b4 a1c1 b4c3 b2c3 c7a5 d2c4 a5d8 c4d2 d8a5');
INSERT INTO evaluations VALUES(48655,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(48655,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48655,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(48655,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(48655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48655,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48655,36,'cp 0','c7c6 d1c2 g7g6 c1g5 f8e7 e2e3 c8f5 f1d3 f5d3 c2d3 b8d7 g5h6 f6g4 h6f4 g4f6 e1g1 e8g8 h2h3 f8e8 d3c2 a7a5 a1c1 e7f8 f1d1 f6e4 c3e4 e8e4 a2a4 e4e6 f3e1 f8d6 f4d6 e6d6 e1d3 d6e6 d1e1 d8b6');
INSERT INTO evaluations VALUES(45767,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45767,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45767,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(45767,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45767,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45767,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45767,36,'cp 1','d1c2 g7g6 c1g5 f8e7 e2e3 c8f5 f1d3 f5d3 c2d3 b8d7 g5h6 f6g4 h6f4 e8g8 h2h3 g4f6 e1g1 a7a5 d3c2 f8e8 a1c1 e7b4 a2a3 b4f8 f1e1 f6e4 c3e4 e8e4 f3e5 d7e5 f4e5 f8d6 f2f3 e4e5 d4e5 d6e5 e1e2 d8f6 e3e4 d5d4 c1d1 a8e8 a3a4 e5d6');
INSERT INTO evaluations VALUES(45770,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(45770,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45770,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45770,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45770,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45770,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45770,36,'cp 0','g7g6 c1g5 f8e7 e2e3 c8f5 f1d3 f5d3 c2d3 b8d7 g5h6 f6g4 h6f4 g4f6');
INSERT INTO evaluations VALUES(49059,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(49059,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(49059,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(49059,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(49059,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(49059,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(49059,36,'cp 162','d2d3 d8e7 d3e4 e7e5 e4d5 f8d6 b1c3 b8d7 c1d2 d7f6 e1c1 e8g8 h2h3 e5e2 f1e2 a7a6 h1e1 b7b5 a2a3 f8e8 e2f3 c8d7 d2g5 d7f5 g5h4 h7h6 h4g3 g7g5 c3e2 f5e4 f3e4 f6e4 g3d6 e4d6 e2g3 a6a5 c1d2 f7f5 f2f3 f5f4 g3e4 g8g7 e4d6 c7d6 e1e8');
INSERT INTO evaluations VALUES(49056,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(49056,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(49056,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(49056,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(49056,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(49056,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(49056,36,'cp 161','d8e7 d3e4 e7e5 e4d5 f8d6 b1c3 b8d7 c1e3 e8g8 e1c1 d7f6 e2d2 f8e8 e3d4 e5f4 d2f4 d6f4 c1b1 c8d7 f2f3 a7a6 a2a4 a8d8 f1c4 d7f5 b2b3 f6d7 b1b2 f4d6 h2h3 d7b6 c4d3 f5d3 d1d3 f7f5 h1f1 h7h5 d3d1 d6b4 d4b6');
INSERT INTO evaluations VALUES(33846,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(33846,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(33846,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(33846,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(33846,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(33846,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(33846,36,'cp 177','d3e4 e7e5 e4d5 f8d6 b1c3 b8d7 c1e3 e8g8 e1c1 d7f6 e2d2 c8f5 e3d4 e5f4 d4f6 g7f6 g2g3 f4d2 d1d2 f8e8 f1d3 f5g4 d3e2 f6f5 e2g4 f5g4 h2h3 g4h3 h1h3 e8e1 d2d1 a8e8 h3h4 e1d1 c1d1 e8e5 d1d2 d6c5 h4f4 g8g7 d2d3');
INSERT INTO evaluations VALUES(33845,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(33845,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(33845,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(33845,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(33845,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(33845,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(33845,36,'cp 170','e7e5 e4d5 f8d6 b1c3 b8d7 c1e3 e8g8 e1c1 d7f6 e2d2 f8e8 e3d4 e5f4 d2f4 d6f4 c1b1 c8d7 g2g3 f4e5 f1g2 f6g4 d4e5 e8e5 d1d2 g8f8 h2h3 g4f6 h1d1 a8e8 b1c1 a7a5 a2a3 g7g5 b2b3 f8g7 c1b2 h7h5 d2d4');
INSERT INTO evaluations VALUES(33926,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(33926,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(33926,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(33926,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(33926,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(33926,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(33926,36,'cp 156','e4d5 f8d6 b1c3 b8d7 c1e3 e8g8 e1c1 f8e8 e2d2 d7f6 e3d4 e5f4 d2f4 d6f4 c1b1 a7a6 f1d3 c8d7 h1e1 h7h5 a2a3 a8d8 d4f6 g7f6 g2g3 e8e1 d1e1 f4e5 f2f4 e5d4 c3e4 g8g7 c2c3 d4b6 c3c4 c7c6 e4c3 c6d5 c3d5');
INSERT INTO evaluations VALUES(33923,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(33923,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(33923,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(33923,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(33923,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(33923,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(33923,36,'cp 170','f8d6 b1c3 b8d7 c1e3 e8g8 e2d2 d7f6 e1c1 f8e8 e3d4 e5f4 d2f4 d6f4 c1b1 c8d7 f1e2 a8d8 e2f3 b7b6 h2h3 g8f8 b2b3 d7f5 b1b2 f6e4 h1e1 e4c3 b2c3 e8e1 d1e1 f4d6 c3b2 f7f6 c2c4 d8e8 e1e8 f8e8 a2a3 d6e5 d4e5');
INSERT INTO evaluations VALUES(51936,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(51936,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(51936,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(51936,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(51936,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(51936,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(51936,36,'cp 33','g1f3 b8c6 d2d4 c5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 e6e5 e1g1 e8g8 c1e3 b4c3 b2c3 d7d6 c4c5 d6d5 e4d5 f6d5 e3d2 d8f6 d1c2 g7g6 d3e4 c8f5 f1e1 a8b8 c3c4 d5b4 d2b4 f5e4 c2c3 e4f5 a1d1 f8e8 d1d6');
INSERT INTO evaluations VALUES(51935,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(51935,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(51935,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(51935,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(51935,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51935,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(51935,36,'cp 0','b8c6 d2d4 c5d4 f3d4 g8f6 d4c6 b7c6 b1c3 f8b4 f1d3 e6e5 e1g1 e8g8 c1e3 b4c3 b2c3 d7d6 c4c5 d6d5 d1a4 d5e4 d3c2 d8c7 a1d1 c8f5 d1d6 a8c8 c2b3 f6e8 d6d2 e8f6');
INSERT INTO evaluations VALUES(18769,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(18769,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18769,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18769,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(18769,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18769,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18769,36,'cp 12','d2d4 c5d4 f3d4 g8f6 d4c6 b7c6 b1c3 f8b4 f1d3 e6e5 c1e3 e8g8 e1g1 b4c3 b2c3 d7d6 c4c5 d6d5 e4d5 f6d5 e3d2 d8f6 d1c2 h7h6 f1e1 f8d8 d3h7 g8h8 h7e4 a8b8 c3c4 d5b4 d2b4 b8b4 c2c3 b4b8 e4c2 c8e6 c3e5');
INSERT INTO evaluations VALUES(18789,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(18789,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(18789,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(18789,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(18789,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18789,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(18789,36,'cp 6','c5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 e6e5 e1g1 e8g8 h2h3 b4c3 b2c3 d7d6 c1e3 c8e6 d1d2 d8c7 f2f4 e5f4 e3f4 f6d7 f1d1 d7e5 d3f1 c7b6 c4c5 b6c5 d2d4 e5c4 f1c4 c5c4 f4d6 f8e8 d4c4 e6c4 d1d4 a8d8 e4e5 c4e6 c3c4 c6c5 d6c5 d8d4 c5d4 e6c4');
INSERT INTO evaluations VALUES(18901,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(18901,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(18901,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18901,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(18901,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(18901,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(18901,36,'cp 14','f3d4 g8f6 d4c6 b7c6 b1c3 f8b4 f1d3 e6e5 c1e3 b4c3 b2c3 d7d6 h2h3 e8g8 e1g1 c8e6 d1d2 d8c7 f2f4 e5f4 e3f4 f6d7 f1d1 d7e5 d3f1 c7b6 f4e3 b6a5 d2d6 e6c4 e3d4 c4f1 d1f1 f7f6 f1f5 f8e8 d4e5 e8e5 f5e5 a5e5 d6e5 f6e5 a1d1 a8b8 d1d7 b8b1 g1h2 a7a6 d7c7 b1a1 c7c6 a1a2 c6c7 a2a1');
INSERT INTO evaluations VALUES(18895,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(18895,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(18895,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(18895,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(18895,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18895,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(18895,36,'cp 17','g8f6 d4c6 b7c6 b1c3 f8b4 f1d3 e6e5 c1e3 d7d6 h2h3 b4c3 b2c3 c8e6 d1d2 e8g8 e1g1 d8c7 f2f4 e5f4 e3f4 f6d7 f1d1 d7e5 d3f1 c7b6 f4e3 b6a5 d2d6 e6c4 e3d4 c4f1 d1f1 f8e8 f1f5 f7f6 d4e5 e8e5 f5e5 a5e5 d6e5 f6e5 a1d1 a8b8 d1d7 b8b1 g1h2');
INSERT INTO evaluations VALUES(12469,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(12469,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(12469,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(12469,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(12469,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(12469,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(12469,36,'cp 24','d4c6 b7c6 b1c3 f8b4 f1d3 e6e5 e1g1 e8g8 h2h3 b4c3 b2c3 d8e7 d1f3 d7d6 c4c5 d6c5 c1a3 c8e6 f3e3 f6d7 f1d1 f8b8 d1d2 h7h6 a1d1 e7g5 e3g5 h6g5 d3f1 b8b7 d2d6 a8b8 f1e2 b7b1 a3c1 d7f8 c1g5 e6a2 d1b1 b8b1 g1h2 f8e6 g5e3');
INSERT INTO evaluations VALUES(19400,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19400,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19400,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(19400,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19400,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19400,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(19400,36,'cp 13','g7g6 e2e3 f8g7 f1g2 e7e6 g1e2 g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 e1g1 e8g8 f1e1 f8e8 h2h4 c8d7 c3d5 e6d5 c1e3 d7g4 d1d2 g4e2 e1e2 g7d4 e3d4 e8e2 d2e2 c6d4 e2d3 d8f6 g2d5 d4c6 d3f3 f6f3 d5f3 a8d8 a1b1 d8d7 f3g2 g8g7 g2c6');
INSERT INTO evaluations VALUES(19257,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(19257,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(19257,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19257,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(19257,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(19257,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(19257,36,'cp 11','f1g2 f8g7 a1b1 e7e6 a2a3 a7a5 e2e4 g8e7 g1e2 b7b6 e1g1 c8a6 d2d3 e8g8 b2b3 a8b8 c3b5 a6b5 c4b5 c6d4 e2d4 g7d4 c1h6 f8e8 b3b4 a5b4 a3b4 d7d6 h2h4 d8d7 h4h5 b8a8 d1d2 d7b5 e4e5 d4e5 b4c5 b5c5 g2a8 e8a8 f1e1');
INSERT INTO evaluations VALUES(19258,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(19258,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19258,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(19258,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19258,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19258,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(19258,36,'cp 7','f8g7 e2e3 e7e6 g1e2 g8e7 e1g1 e8g8 f1e1 d7d5 c4d5 e6d5 d2d4 c5d4 e3d4 c8f5 c1g5 d8d7 a1c1 a8e8 d1d2 f5h3 g5h6 g7h6 d2h6 h3g2 g1g2 e7f5 h6d2 e8d8 h2h4 h7h6 a2a3 g8g7 e2f4 f5d4 c3d5 d7f5 c1c5');
INSERT INTO evaluations VALUES(5327,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(5327,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5327,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(5327,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5327,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5327,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(5327,36,'cp 4','g1f3 d7d6 a1b1 g8f6 e1g1 e8g8 d2d3 a8b8 a2a3 a7a6 b2b4 c5b4 a3b4 b7b5 c4b5 a6b5 c1e3 c8d7 d1c1 d8c8 f3d4 c6d4 e3d4 d7h3 d4f6 g7f6 c3d5 h3g2 d5f6 e7f6 g1g2 c8d7 c1f4 g8g7 f1c1 f8e8 e2e4 e8c8 b1a1 b8a8 h2h3 c8c1 a1a8 c1d1');
INSERT INTO evaluations VALUES(5326,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(5326,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5326,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(5326,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5326,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5326,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(5326,36,'cp 10','d7d6 e1g1 c8f5 h2h3 e7e5 d2d3 g8e7 a2a3 e8g8 a1b1 a7a5 c1g5 f7f6 g5d2 f5e6 e2e4 a8b8 c3d5 e6d7 b2b4 a5b4 d5e7 d8e7 a3b4 c5b4 d2b4 c6b4 b1b4 d6d5 d1b3 d5e4 d3e4 d7e6 f1b1');
INSERT INTO evaluations VALUES(5489,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5489,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(5489,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5489,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(5489,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5489,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(5489,36,'cp 8','e1g1 c8f5 h2h3 e7e5 d2d3 g8e7 a2a3 e8g8 f3d2 a7a6 c3d5 f5e6 a1b1 a8b8 b2b3 b7b5 e2e3 d8d7 g1h2 f7f5 c1b2 a6a5 d1e2 h7h6 b2a1 b5b4 a3a4 e6f7 b1d1 g8h7 f2f4 b8e8 e3e4 f5e4 d3e4 e7d5 c4d5 c6d4 a1d4');
INSERT INTO evaluations VALUES(5488,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(5488,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5488,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5488,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(5488,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(5488,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(5488,36,'cp 0','c8f5 d2d3 d8d7 a1b1 f5h3 g2h3 d7h3 e2e4 h3d7 a2a3 a8d8 c1d2 f7f5 e4f5 d7f5 c3d5 g8f6 d2c3 e8g8 d5e3 f5h5 g1g2 b7b6 b2b4 f6g4 e3g4 h5g4 c3g7 g8g7 b4b5 c6d4 f3d4 g4d4 d1e2 e7e5 b1e1 f8f5 f2f4 d8f8 e2e4');
INSERT INTO evaluations VALUES(1481,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1481,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1481,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1481,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1481,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(1481,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1481,36,'cp 0','c5d4 c3d4 c8d7 d3e2 g8h6 b1c3 h6f5 c3a4 b6a5 a4c3 f8b4 e1g1 b4c3 b2c3 a5c3 c1d2 c3a3 d2c1');
INSERT INTO evaluations VALUES(1525,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1525,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1525,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1525,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1525,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1525,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1525,36,'cp 0','c3d4 c8d7 d3e2 g8h6 b1c3 h6f5 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(1506,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1506,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1506,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1506,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1506,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1506,36,'cp 0','c8d7 d3e2 g8h6 b1c3 h6f5 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(27542,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27542,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27542,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27542,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27542,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27542,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27542,36,'cp 0','d3e2 g8h6 b1c3 h6f5 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(27544,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27544,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27544,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27544,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27544,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27544,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27544,36,'cp 0','g8e7 b1c3 e7f5 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(52485,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(52485,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(52485,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52485,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(52485,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(52485,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52485,36,'cp 12','b8c6 c2c4 g7g6 b1c3 f8g7 f1g2 e7e6 d2d4 c5d4 c3b5 d7d5 c4d5 e6d5 e1g1 g8f6 f3d4 e8g8 c1f4 a7a6 b5c3 f8e8 a1c1 c6d4 d1d4 f6h5 d4d5 h5f4 g3f4 c8e6 d5d8 e8d8 g2b7 a8b8 b7a6 b8b2 a2a4 b2b4 e2e3 g7c3 c1c3 b4a4 a6b5 a4a7 g1g2 e6f5 b5c6 d8d2 f1c1');
INSERT INTO evaluations VALUES(19505,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(19505,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19505,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(19505,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(19505,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19505,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(19505,36,'cp 10','c2c4 g7g6 b1c3 f8g7 f1g2 d7d6 e1g1 c8f5 d2d3 d8d7 a1b1 f5h3 g2h3 d7h3 e2e4 h3d7 a2a3 a8d8 c1d2 f7f5 e4f5 d7f5 c3d5 g8f6 d2c3 e8g8 c3f6 g7f6 b2b4 e7e6 d5e3 f5h5 h2h4 c5b4 a3b4 c6d4 f3d4 h5d1 f1d1 f6d4 g1g2');
INSERT INTO evaluations VALUES(19408,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(19408,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(19408,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(19408,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(19408,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19408,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19408,36,'cp 16','g7g6 b1c3 f8g7 f1g2 d7d6 e1g1 a8b8 e2e3 g8f6 d2d4 e8g8 b2b3 c8e6 d1e2 e6g4 d4c5 d6c5 c1b2 d8d7 a1d1 d7f5 e3e4 f6e4 c3e4 g7b2 e2b2 f5e4 d1e1 e4f5 f3e5 c6e5 b2e5 f8e8 f2f3 f5e5 e1e5 g4f5 e5c5 b8d8 c5c7 d8d2 f1f2 d2d7 c7d7 f5d7 f3f4');
INSERT INTO evaluations VALUES(19261,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(19261,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19261,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(19261,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19261,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19261,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(19261,36,'cp 17','f1g2 f8g7 b1c3 d7d6 a2a3 g8f6 e1g1 e8g8 d2d3 a7a6 a1b1 a8b8 c1e3 f6g4 e3d2 b7b5 c4b5 a6b5 b2b4 c5b4 a3b4 c8e6 f3g5 e6d7 g5e4 g4f6 e4f6 g7f6 c3d5 f6g7 d2c3 e7e6 c3g7 g8g7 d5e3 b8c8 g2c6 c8c6 b1a1 e6e5 d1d2 c6c8 f1c1 c8c1 a1c1 d8b6');
INSERT INTO evaluations VALUES(19262,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(19262,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19262,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(19262,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(19262,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19262,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19262,36,'cp 9','f8g7 b1c3 d7d6 a2a3 g8f6 d2d3 e8g8 e1g1 a8b8 a1b1 a7a6 b2b4 c5b4 a3b4 b7b5 c4b5 a6b5 c1e3 c8d7 d1c1 d8c8 f3d4 c6d4 e3d4 d7h3 d4f6 g7f6 c3d5 h3g2 d5f6 e7f6 g1g2 c8d7 c1b2 g8g7 h2h4 f8e8 b1a1 b8a8 a1a8 e8a8 f1a1 a8e8 b2a2 d7g4 e2e4');
INSERT INTO evaluations VALUES(5331,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5331,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(5331,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(5331,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5331,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5331,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5331,36,'cp 0','b1c3 d7d6 a2a3 g8f6 e1g1 e8g8 d2d3 a8b8 c1e3 a7a6 a1b1 b7b5 c4b5 a6b5 b2b4 c5b4 a3b4 f6g4 e3d2 g4e5 f3e5 c6e5 b1c1 c8b7 h2h4 d8d7 c3d5 d7e6 d1b3 e6d7 b3d1');
INSERT INTO evaluations VALUES(42752,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42752,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(42752,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(42752,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(42752,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(42752,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(42752,36,'cp 5','g1f3 b4c3 d2c3 d7d6 d1e2 b8d7 c1g5 a7a6 a2a4 h7h6 g5h4 d7f8 f3d2 f8g6 h4f6 d8f6 g2g3 g6e7 a4a5 c8e6 e1c1 e8c8 b2b4 e6c4 e2c4 f6e6 c4e6 f7e6 c3c4 g7g5 c2c3 h6h5 f2f3 c8b8 c1c2 g5g4');
INSERT INTO evaluations VALUES(42750,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42750,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(42750,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(42750,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(42750,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42750,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42750,36,'cp 2','b4c3 d2c3 d7d6 c1g5 h7h6 g5h4 b8d7 d1e2 a7a6 e1c1 b7b5 c4b3 d8e7 f3d2 c7c6 c3c4 e8g8 d2f1 f8d8 f2f3 d7c5 g2g4 c5b3 c2b3 c8e6 f1e3 b5c4 b3c4 a6a5 h1g1 a5a4 e3f5 e6f5 g4f5 d6d5 a2a3 g8h8 c1b1 d5e4 d1d8 a8d8');
INSERT INTO evaluations VALUES(42788,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42788,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(42788,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42788,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(42788,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(42788,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(42788,36,'cp 12','d2c3 d7d6 c1e3 e8g8 c4d3 b7b6 f3d2 f6g4 d2f1 g4e3 f1e3 b8d7 d1d2 d7c5 f2f3 c7c6 e1g1 a7a5 f1d1 d8c7 b2b4 c5d3 d2d3 f8d8 a2a3 a5b4 a3b4 a8a1 d1a1 b6b5 c3c4 b5c4 d3c4 c7b6 c4c3 d6d5 e4d5 c6d5 c3e5 d5d4 e3f5 c8f5 e5f5');
INSERT INTO evaluations VALUES(42786,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42786,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(42786,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42786,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(42786,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(42786,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(42786,36,'cp 8','d7d6 c1g5 b8d7 f3d2 h7h6 g5h4 d7f8 d2f1 f8g6 h4f6 d8f6 f1e3 e8g8 g2g3 g6e7 d1e2 c8e6 h1f1 e6h3 f1g1 h3e6 e1c1 e6c4 e3c4 f6e6 c1b1 f7f5 f2f3 a8e8 e4f5 e7f5 g1e1 b7b5 c4a5 e6d7 e2e4 g8h8 a5c6 a7a6 c6b4 f5e7');
INSERT INTO evaluations VALUES(41919,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41919,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(41919,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41919,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(41919,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41919,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41919,36,'cp 12','c1g5 b8d7 g5h4 h7h6 f3d2 d7f8 d1e2 a7a6 a2a4 f8g6 h4f6 d8f6 f2f3 c8e6 a4a5 e8c8 e1c1 c8b8 g2g3 g6e7 b2b4 e6c4 e2c4 f6g6 b4b5 a6b5 c4b5 f7f5 a5a6 b7b6 d2c4 g6e8 b5b3 b8a8 c4a3 e8c6 h1e1 f5e4');
INSERT INTO evaluations VALUES(14209,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(14209,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(14209,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(14209,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(14209,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(14209,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(14209,36,'cp 57','g1f3 e7e6 a2a3 d5c4 b1d2 b7b5 a3a4 b5a4 f1c4 c8d7 c4b5 f6d5 d1a4 d5b6 a4d1 c6b4 b5d7 d8d7 e1g1 d7b5 d2e4 b5c6 e4g3 f8d6 c1d2 a7a5 a1c1 c6b7 d2b4 a5b4 d1c2 h7h6 g3e4 e8g8 e4c5 d6c5 c2c5 b6d5 f3e5 a8a2 e5d3 f8b8 e3e4 d5f6 c5c4');
INSERT INTO evaluations VALUES(38382,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(38382,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(38382,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38382,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(38382,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(38382,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(38382,36,'cp 54','c8g4 b1c3 b8c6 c1e3 e7e5 d4d5 c6d4 f1e2 g4f3 e2f3 c7c5 d5c6 b7c6 e1g1 a8b8 a1b1 a7a5 a2a3 g8f6 e3d4 e5d4 d1d4 e8g8 d4d2 f6d7 b2b4 d8f6 f1c1 a5b4 a3b4 f6e6 f3e2 d7b6 d2d3 b6d7 d3e3 d7e5 b4b5 e5c4 e2c4 e6c4 b5c6 c4c6');
INSERT INTO evaluations VALUES(28113,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(28113,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(28113,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(28113,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(28113,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(28113,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(28113,36,'cp 61','b1c3 b8c6 c1e3 e7e5 d4d5 c6d4 f1e2 d4e2 d1e2 g7h6 a2a4 h6e3 e2e3 g4f3 e3f3 g8f6 a4a5 e8g8 e1g1 a7a6 f3e3 f6g4 e3g3 g4f6 b2b4 f6h5 g3e3 f7f5 c4c5 d8h4 f2f3 f8f7 c5c6 b7c6 d5c6 h5f6 b4b5 a6b5 c3b5 f5e4 a5a6');
INSERT INTO evaluations VALUES(28110,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(28110,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(28110,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(28110,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(28110,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(28110,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(28110,36,'cp 50','b8c6 c1e3 e7e5 d4d5 c6d4 f1e2 g4f3 e2f3 h7h5 e1g1 d8e7 c3b5 d4f3 d1f3 a7a6 b5c3 g7h6 b2b4 h6e3 f3e3 h5h4 h2h3 g8f6 c4c5 f6h5 c5c6 b7b6 g2g4 h5f4 c3e2 f4e2 e3e2 e8g8 g1g2 e7g5 e2e3 g5e3 f2e3 a6a5 b4b5 g8g7 a1c1 a8e8');
INSERT INTO evaluations VALUES(23821,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(23821,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(23821,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(23821,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(23821,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(23821,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(23821,36,'cp 50','c1e3 e7e5 d4d5 c6d4 f1e2 g4f3 e2f3 c7c5 d5c6 b7c6 e1g1 a8b8 a1b1 a7a5 a2a3 g8f6 e3d4 e5d4 d1d4 e8g8 d4d2 f6d7 b2b4 d8f6 f1c1 a5b4 a3b4 f6e6 f3e2 h7h5 e2f1 d7e5 d2e2 e5d7 b4b5 f8c8 e2d2 d7c5 d2c2 e6f6 c3e2 c5e6 c2d2 c6b5 c4b5 c8c1');
INSERT INTO evaluations VALUES(23820,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(23820,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(23820,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(23820,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(23820,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(23820,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(23820,36,'cp 45','e7e5 d4d5 c6d4 f1e2 g4f3 e2f3 c7c6 d5c6 b7c6 e1g1 a7a5 a2a3 a8b8 a1b1 g8f6 e3d4 e5d4 d1d4 e8g8 d4d2 f6d7 b2b4 d8f6 f1c1 a5b4 a3b4 f6e6 f3e2 f8c8 f2f4 b8a8 b1b3 c8e8 c1d1 d7f6 d2d3 f6e4 c3e4 e6e4 e2f3 e4d3');
INSERT INTO evaluations VALUES(10259,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(10259,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(10259,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(10259,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(10259,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(10259,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(10259,36,'cp 65','f3e5 c6e5 d2d4 c5d6 f2f4 e5c6 e4e5 d6e7 e5f6 e7f6 d4d5 d8e7 e1f2 e7c5 c1e3 f6d4 d1f3 d4e3 f3e3 c5e3 f2e3 c6e7 e3d2 d7d6 a1e1 c8d7 f1d3 f7f5 c3b5 d7b5 d3b5 e8d8 b5c4 a7a6 c4d3 b7b5 c2c4 b5c4 d3c4 e7c8 e1e6 c8b6 d2d3 b6d7 b2b4 a6a5 c4b5 a5b4 a3b4 a8a2 h1e1 a2g2 e6e7 d7f6 e7f7 f6d5');
INSERT INTO evaluations VALUES(10047,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(10047,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(10047,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(10047,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(10047,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(10047,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(10047,36,'cp 61','c6e5 d2d4 c5d6 f2f4 e5c6 e4e5 d6e7 e5f6 e7f6 d4d5 d8e7 e1f2 e7c5 c1e3 f6d4 d1f3 d4e3 f3e3 c5e3 f2e3 c6e7 a1e1 d7d6 e3d2 c8d7 f1d3 f7f5 c3b5 d7b5 d3b5 e8d8 b5c4 a7a6 a3a4 h8e8 h2h4 e7g8 e1e8 d8e8 c4d3 g8e7 a4a5 h7h5 c2c4 g7g6 b2b4 e8f8 h1b1 e7g8 b1b3');
INSERT INTO evaluations VALUES(10767,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(10767,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(10767,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(10767,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(10767,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(10767,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(10767,36,'cp 68','d2d4 c5d6 f2f4 e5c6 e4e5 d6e7 e5f6 e7f6 d4d5 d8e7 e1f2 e7c5 c1e3 f6d4 d1f3 d4e3 f3e3 c5e3 f2e3 c6e7 e3d2 d7d6 a1e1 c8d7 f1d3 f7f5 c3b5 d7b5 d3b5 e8d8 b5c4 h8e8 h2h4 a8b8 h1h3 e7g8 h3g3 e8e1 d2e1 g7g6 h4h5 g8e7 g3h3 d8d7 h5g6 h7g6 h3h7 b7b5 c4d3 a7a6 g2g3 b8e8 d3e2 c7c6 c2c4 b5c4 e2c4');
INSERT INTO evaluations VALUES(10765,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(10765,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(10765,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(10765,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(10765,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(10765,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(10765,36,'cp 84','c5d6 f2f4 e5c6 e4e5 d6e7 e5f6 e7f6 d4d5 d8e7 e1f2 e7c5 c1e3 f6d4 d1e2 d4e3 e2e3 c5e3 f2e3 c6e7 e3d2 d7d6 a1e1 c8d7 f1d3 a7a6 e1e3 e8f8 h1e1 e7g8 c3e4 a8e8 c2c4 g8h6 e3e2 d7g4 e2f2 g4d7 h2h3 h6g8 f4f5 h7h5 b2b4 h8h6 e1e3 h5h4 f2f4 h6h5 f5f6 g8f6 e4f6 g7f6 f4f6');
INSERT INTO evaluations VALUES(47953,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(47953,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(47953,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(47953,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(47953,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(47953,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(47953,36,'cp 54','f3d4 f8e7 c1f4 b8c6 d1d2 e8g8 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 h1e1 a8d8 c4e6 f7e6 f4e3 d8d1 c3d1 f8e8 d1c3 c7c6 e3e5 f6d7 e5d4 e6e5 d4d2 d7f6 f2f3 h7h6 g2g3 e7c5 d2e2 a6a5 e1d1 e5e4 c3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(20357,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20357,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(20357,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20357,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(20357,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(20357,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(20357,36,'cp 13','d2d4 c8d7 f1e2 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 e2d3 f5e3 f2e3 a5c4 d3c4 d5c4 a4c3 e8g8 c3e4 d7c6 f3d2 b7b5 d1g4 g8h8 e4g3 f7f5 e5f6 e7f6 d2e4 c6e4 g3e4 e6e5 d4e5 a7e3 g1h1 f6e7 e4d6 e7d6');
INSERT INTO evaluations VALUES(36543,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(36543,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(36543,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(36543,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(36543,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(36543,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(36543,36,'cp 16','f1e2 e7e5 d4e5 d6e5 f3e5 d8d1 e2d1 f8d8 d1e2 f6e4 c3e4 g7e5 e4c3 b8a6 c1d2 a6b4 e1c1 e5c3 d2c3 b4a2 c1c2 c8f5 c2b3 a2c3 b3c3 a7a5 e2f3 d8b8 d1d4 g8f8 h1d1 c7c6 h2h3 a5a4 d1a1 a8a6 f3g4 c6c5 d4d5 f5e4 d5c5 e4g2 c5c7 a6c6 c7c6');
INSERT INTO evaluations VALUES(36542,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36542,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36542,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(36542,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36542,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36542,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(36542,36,'cp 16','e7e5 e1g1 d8e7 b2b4 c8f5 b4b5 b8d7 c1b2 e5e4 f3d2 h7h5 d1c2 f8c8 f1d1 c7c5 b5c6 b7c6 b2a3 c6c5 d2b3 d7b6 a1b1 f6g4 h2h3 g4f2 g1f2 e7f6 d4c5 d6c5 c3b5 f6h4 f2g1 h4g3 a3c5 c8c5 b3c5');
INSERT INTO evaluations VALUES(57765,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(57765,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(57765,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(57765,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(57765,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(57765,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(57765,36,'cp -14','e7e5 e2e4 g8f6 g1f3 f6e4 d2d3 e4f6 f3e5 b8c6 e5c6 d7c6 d1f3 f8d6 b1c3 e8g8 f1e2 f6d5 c3d5 c6d5 e1g1 c7c6 c1f4 d8c7 f4d6 c7d6 a2a3 f8e8 f1e1 c8d7 f3g3 d6c5 c2c3 c5b6 a1b1 d7f5 e2g4 f5g4 g3g4');
INSERT INTO evaluations VALUES(57239,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(57239,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(57239,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(57239,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(57239,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(57239,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(57239,36,'cp -24','e2e4 g8f6 g1f3 f6e4 d2d3 e4f6 f3e5 b8c6 e5c6 d7c6 d1f3 c8e6 f1e2 f8e7 b1c3 d8d7 e1g1 e8c8 c1f4 f6d5 c3d5 e6d5 f3e3 c8b8 a1e1 h8e8 e3g3 g7g6 e2g4 f7f5 g4f3 d5f3 g3f3 e7f6 c2c3 e8e1');
INSERT INTO evaluations VALUES(57160,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(57160,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(57160,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(57160,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(57160,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(57160,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(57160,36,'cp -27','g8f6 b1c3 f8b4 a2a3 b4a5 g1f3 e8g8 f1e2 d7d6 b2b4 a5b6 e1g1 a7a5 a1b1 a5b4 a3b4 b8c6 e2c4 c6e7 d2d3 e7g6 c4b3 c8e6 c1e3 b6e3 f2e3 e6b3 b1b3 h7h6 b4b5 f8e8 b3b1 d6d5 e4d5 f6d5 c3d5 d8d5 c2c4');
INSERT INTO evaluations VALUES(49666,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(49666,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(49666,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(49666,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(49666,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(49666,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(49666,36,'cp -16','b1c3 f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 b4f8 f1e2 f6e4 c3e4 e8e4 e1g1 d7d5 e2f3 e4e8 f1e1 c7c6 e1e8 d8e8 d3f4 b8d7 d2d4 a7a5 a2a4 d7f6 f4d3 c8f5 c2c3 e8d8 c1g5 h7h6 g5h4 f8d6 d3e5 d6c7 e5g4');
INSERT INTO evaluations VALUES(49618,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(49618,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(49618,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(49618,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(49618,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(49618,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(49618,36,'cp -26','f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 f1e2 b4f8 c3e4 e8e4 e1g1 d7d5 e2f3 e4e8 f1e1 c7c6 e1e8 d8e8 d3f4 a7a5 d2d3 b8d7 c1e3 d7e5 e3d4 c8f5 d1e2 e5f3 e2f3 e8d7 a1e1 a5a4 a2a3 c6c5 d4e5');
INSERT INTO evaluations VALUES(42738,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(42738,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42738,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42738,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42738,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(42738,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(42738,36,'cp -14','a2a3 b4a5 g1f3 e8g8 f1e2 d7d6 b2b4 a5b6 e1g1 a7a5 a1b1 b8c6 e2c4 c6d4 d2d3 c8e6 f3d4 b6d4 c4e6 f7e6 c3b5 d4b6 c2c4 h7h6 b5c3 d8e8 c1e3 b6e3 f2e3 a5b4 a3b4 c7c6 b1a1 e8g6 a1a8 f8a8');
INSERT INTO evaluations VALUES(49446,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49446,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49446,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49446,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49446,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49446,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49446,36,'cp 0','d7d5 e4e5 f6d7 c3e2 c7c5 d2d4 c5d4 e2d4 b8c6 c1f4 f8e7 h2h4 d8b6 c2c3 b6b2 d4b5 a8b8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(53000,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(53000,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(53000,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(53000,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(53000,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(53000,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(53000,36,'cp -44','e7e6 h5e2 g8f6 g1f3 f8e7 d2d4 c5d4 f3d4 b8c6 d4c6 b7c6 e1g1 d6d5 c4d3 e8g8 c2c4 f6e4 d3e4 d5e4 e2e4 d8b6 b1c3 f7f5 e4e2 e7f6 f1d1 e6e5 c1e3 b6b4 a1c1 f5f4 e3d2 b4b2 c4c5 g8h8 c3e4 b2a2 e4f6 g7f6');
INSERT INTO evaluations VALUES(53967,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(53967,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(53967,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(53967,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(53967,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(53967,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(53967,36,'cp -32','h5e2 f8e7 g1f3 g8f6 d2d4 c5d4 f3d4 b8c6 d4c6 b7c6 e1g1 d6d5 c4d3 e8g8 c2c4 f6e4 d3e4 d5e4 e2e4 d8b6 b1c3 e7f6 c1e3 b6b2 c3b5 b2e5 e4c6 e5b8 a1b1 c8b7 c6d6 b8d6 b5d6 b7c6 d6b5 a7a6 b5d6 f8b8 f1d1 f6e7 c4c5 b8b1 d1b1 e7d6 c5d6 f7f6');
INSERT INTO evaluations VALUES(14260,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(14260,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(14260,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(14260,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14260,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14260,36,'cp 0','g1f3 c8g4 f1e2 e7e6 h2h3 g4f5 f3h4 f5g6 c3b1 a7a6 b1d2 e6e5 d4e5 c6e5 e1g1 d8e7 c2c4 d5c4 d2c4 g6e4 c4e5 e7e5 h4f3 e5f5 f3d4 f5e5');
INSERT INTO evaluations VALUES(41816,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(41816,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41816,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(41816,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(41816,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41816,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41816,36,'cp 21','c7c5 b1d2 e8g8 d4c5 d6c5 c2c4 c5e7 a2a3 b8c6 f1d3 d5c4 d2c4 b7b5 c4d6 c8a6 d3b5 e7d6 b5c6 d6g3 h2g3 a8c8 d1d8 f8d8 c6a4 f6e4 f3d4 g7g6 d4c6 d8d7 c6b4 d7d2 b4a6 d2f2 a6b4 f2g2 a4c6 g2g3 c6e4 g3e3 e1d2 e3e4');
INSERT INTO evaluations VALUES(40575,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40575,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(40575,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40575,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40575,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40575,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40575,36,'cp 11','d4c5 d6c5 b1d2 e8g8 c2c4 d8b6 d1c2 c8d7 a1c1 c5e7 c4d5 e6d5 g3c7 b6c6 c2d1 c6a4 d1a4 d7a4 c7g3 b8d7 f1e2 f6e4 e1g1 e4g3 h2g3 g7g6 f3d4 a8c8 b2b3 a4c6 d2f3 e7a3 c1c2 g8g7 d4c6 c8c6 f3d4 c6c2');
INSERT INTO evaluations VALUES(11702,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(11702,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(11702,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(11702,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(11702,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(11702,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(11702,36,'cp 0','f8c5 c3d5 e8g8 d2d3 f6d5 b3d5 a8b8 e1g1 d8e8 c1e3 c5e3 f2e3 d7d6 f3g5 c6d8 d1h5 h7h6 f1f6 h6g5 f6h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g5 g8h7 g5h5 h7g8 h5g6 g8h8 g6h6 h8g8');
INSERT INTO evaluations VALUES(7455,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(7455,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(7455,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(7455,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(7455,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(7455,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7455,36,'cp 8','f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 d1f3 b5b4 c3d5 c8b7 e1g1 f6d5 e4d5 e8g8 a2a3 a6a5 a3b4 a5b4 a1a8 d8a8 c1d2 c7c5 c2c3 b4c3 d2c3 e5c3 b2c3 a8a3 b3c4 b7a6 f3e2 a6c4 e2c4 a3a5 f1b1 a5c7 g2g3 d7d6');
INSERT INTO evaluations VALUES(52297,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(52297,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(52297,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(52297,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52297,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(52297,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52297,36,'cp 37','f8g7 d2d4 c5d4 f3d4 b8c6 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1d1 b5c4 d4c5 f8d8 d1d4 c4d5 c5e7 h6g7 e7d8 g7d4 b3d4 a8d8 g2g3 a7a5 b2b3 d5e4 a1d1 a5a4 g3g4 a4b3 a2b3 e4b7 f2f3 g8g7 h2h4');
INSERT INTO evaluations VALUES(37816,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(37816,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(37816,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(37816,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(37816,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(37816,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(37816,36,'cp 40','d2d4 c5d4 f3d4 b8c6 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8d8 e1e7 a7a6 d4c3 d8d5 h2h3 h6f8 e7e4 a8d8 e4d4 d5d4 b3d4 b5e8 a1e1 f8g7 f2f3 h7h5');
INSERT INTO evaluations VALUES(37802,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(37802,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(37802,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(37802,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(37802,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(37802,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(37802,36,'cp 42','c5d4 f3d4 b8c6 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8e8 a1d1 b5c4 d4c5 a8d8 e1e7 e8e7 c5e7 d8d5 d1d5 c4d5 e7f6 h6f4 g2g3 f4c7 g1f1 a7a5 f1e1 c7b6 f6d4 d5b3 d4b6 b3a2 b6a5 g8f8 e1d2 f8e8 d2c3 e8d7 c3d4 d7c6 d4e5 a2e6 h2h4 h7h5');
INSERT INTO evaluations VALUES(37870,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(37870,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(37870,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(37870,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(37870,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(37870,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(37870,36,'cp 28','f3d4 b8c6 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1e1 f8e8 a1d1 b5c4 d4c5 a8d8 e1e7 e8e7 c5e7 d8d5 d1d5 c4d5 e7f6 h6f4 f6d4 a7a6 g2g3 f4c7 g1f1 f7f5 d4e3 g8f7 f1e2');
INSERT INTO evaluations VALUES(44545,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(44545,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44545,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44545,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(44545,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44545,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(44545,36,'cp 16','d4d5 b7b5 f2f3 a7a6 e2e4 d7d6 a2a4 b5b4 b1d2 e7e6 d5e6 c8e6 f4g3 f8e7 g1h3 b8c6 h3f4 d6d5 f1d3 e8g8 e4d5 e6d5 e1g1 c6a5 b2b3 f8e8 f1e1 c5c4 b3c4 d5b7 a1b1 e7c5 g1h1 e8e1 d1e1 d8b6 f4d5 f6d5 c4d5 b7d5 d2b3');
INSERT INTO evaluations VALUES(44591,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44591,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44591,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(44591,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(44591,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44591,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44591,36,'cp 6','b7b5 f2f3 d7d6 e2e4 a7a6 a2a4 b5b4 b1d2 e7e6 d5e6 c8e6 g1h3 b8c6 f4g3 f8e7 h3f4 e8g8 f1d3 d6d5 e4d5 e6d5 e1g1 e7d6 f4d5 f6d5 g3d6 d8d6 d2c4 d6h6 d3e4 d5e3 d1d2 e3c4 d2h6 g7h6 e4c6 a8d8 b2b3 c4e3 f1f2 d8d6 f2e2 d6c6 e2e3 f8d8 a1e1 c5c4 e3e4 c4b3 c2b3');
INSERT INTO evaluations VALUES(20084,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(20084,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(20084,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(20084,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(20084,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(20084,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(20084,36,'cp 13','f1b5 c8d7 d4d3 g8f6 c2c4 c6e5 f3e5 d6e5 b5d7 d8d7 d3d7 f6d7 e1e2 e7e6 h1d1 f8c5 b1c3 e8e7 c1d2 c5d4 a1c1 b7b6 d2e3 d4e3 e2e3 h8c8 b2b3 f7f6 d1d2 d7c5 c1d1 c8d8 f2f3 a7a5 d2d8 a8d8 d1d8 e7d8 e3d2 d8e7');
INSERT INTO evaluations VALUES(19908,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19908,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(19908,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(19908,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(19908,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(19908,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(19908,36,'cp 9','c8d7 d4d3 g8f6 c2c4 c6e5 b5d7 d8d7 f3e5 d6e5 d3d7 f6d7 e1e2 e7e6 h1d1 f8c5 b1c3 e8e7 c1d2 h8c8 a1c1 c5d4 c3b5 a7a5 b2b3 d7c5 f2f3 c8d8 b5c7 a8a7 h2h4 h7h5 c7b5 a7a8 c1c2 b7b6 d2c1 d8d7 c1e3 a8d8 c2d2 d4e3 d2d7 d8d7 d1d7 e7d7 e2e3 f7f6 g2g4 h5g4');
INSERT INTO evaluations VALUES(26032,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26032,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(26032,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(26032,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26032,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(26032,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(26032,36,'cp 22','b5c6 b7c6 d4d3 e7e5 c2c4 g8f6 b1c3 d7e6 e1g1 f8e7 c4c5 e8g8 c1g5 d8c7 c5d6 c7d6 d3c2 h7h6 f1d1 d6b4 g5h4 b4c5 d1c1 a8c8 h2h3 f6h5 h4e7 c5e7 c3e2 c6c5 c1d1 f7f6 f3h4 f8d8 h4f5 e7c7');
INSERT INTO evaluations VALUES(26012,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26012,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(26012,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(26012,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(26012,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26012,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26012,36,'cp 16','d7c6 c2c4 g8f6 b1c3 g7g6 e1g1 f8g7 d4d3 e8g8 f1d1 d8b6 d3e2 a8c8 h2h3 a7a6 a1b1 b6c7 a2a4 f6d7 f3d4 c7b6 c1e3 d7c5 a4a5 b6a5 d4c6 b7c6 b2b4 a5a3 b4c5 g7c3 b1b7 c3b4 e3h6 f8e8 c5d6 b4d6 b7d7');
INSERT INTO evaluations VALUES(26231,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(26231,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(26231,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(26231,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(26231,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(26231,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(26231,36,'cp 13','c2c4 g8f6 b1c3 g7g6 d4d3 f8g7 e1g1 e8g8 f1d1 d8b6 d3e2 a8c8 h2h3 a7a6 a1b1 b6c7 a2a4 b7b6 b2b3 c7b7 c3d5 e7e6 d5f6 g7f6 c1h6 f8d8 f3d2 d6d5 c4d5 e6d5 e4e5 d8e8 d2f3 d5d4 d1d4 c6f3 e2f3 b7f3 g2f3 e8e5 d4d6 e5e6 b1d1 c8e8 d6d5 b6b5 a4b5 a6b5 d5b5 f6g7 h6g7');
INSERT INTO evaluations VALUES(31848,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(31848,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31848,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31848,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(31848,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31848,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31848,36,'cp 0','f2f3 g8f6 g2g4 f5g6 g4g5 f6h5 e2e4 d5e4 f3f4 e7e5 f4e5 h7h6 d1g4 b8c6 c1e3 h6g5 e1c1 h5f4 g1e2 g6h5 g4f5 h5g6');
INSERT INTO evaluations VALUES(31845,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31845,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31845,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31845,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(31845,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(31845,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31845,36,'cp -9','g8f6 g2g4 f5g6 h2h4 h7h6 g1h3 b8c6 c1f4 c6b4 e1f2 e7e6 a2a3 b4c6 a1c1 f8d6 e2e3 d8d7 f1b5 a7a6 f4d6 c7d6 b5f1 h6h5 g4g5 f6g8 h3f4 g8e7 f1h3 d7d8 c3e2 e8g8 f4g6 e7g6 f3f4 g6h4');
INSERT INTO evaluations VALUES(29566,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29566,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(29566,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(29566,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29566,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(29566,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29566,36,'cp -8','g2g4 f5g6 g4g5 f6h5 e2e4 d5e4 f3f4 e7e5 f4e5 b8c6 c1e3 f8b4 f1e2 c6e7 e2h5 e7d5 h5g6 d5e3 d1d2 d8g5 g6e4 b4c3 d2c3 g5h4 e1d2 h4e4 g1e2 e3d5 c3d3 e4h4 a1f1 h4h6 d2e1 e8g8 f1f3 a8d8 h1g1 h6b6 d3b3 c7c5 b3b6 a7b6 c2c3 c5d4 c3d4 f7f6 e1f2 d8c8 e5f6 d5f6');
INSERT INTO evaluations VALUES(29554,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29554,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29554,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(29554,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(29554,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29554,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(29554,36,'cp 0','f5g6 g4g5 f6h5 e2e4 d5e4 f3f4 e7e5 f4e5 b8c6 c1e3 f8b4 f1e2 c6e7 e2h5 e7d5 h5g6 d5e3 d1d2 d8g5 g6e4 b4c3 d2c3 g5h4 e1d2 h4e4 g1e2 e3d5 c3d3 e4h4 a1f1 h4h6 d2e1 f7f6 d3b5 c7c6 b5b7 e8g8 h1g1 f6e5 g1g7 g8h8 f1f8 a8f8 g7f7 f8d8 c2c4 d5e3 b7e7 e3c2 e1f1 h6h3 f1g1 h3e3 g1f1');
INSERT INTO evaluations VALUES(29787,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29787,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29787,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29787,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29787,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29787,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29787,36,'cp 0','h2h4 h7h6 g1h3 b8c6 c1f4 a7a6 a2a3 e7e6 e2e3 h6h5 g4g5 f6d7 f4h2 d7b6 d1d2 f8d6 h2d6 d8d6 h3f4 e8g8 e1c1 c6a5 e3e4 b6c4 d2h2 d6b6 c3a4 b6b5 b2b3 a5b3 c2b3 b5b3 f1c4 d5c4 a4c5 b3a3 h2b2 a3f3 d1f1 f3e3 b2d2 e3a3 d2b2');
INSERT INTO evaluations VALUES(51486,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(51486,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(51486,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(51486,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(51486,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(51486,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(51486,36,'cp 38','d2d4 c8b7 b1c3 e7e6 f2f3 f8b4 e2e4 d8h4 g2g3 b4c3 b2c3 h4e7 g1h3 d7d6 f1d3 b8c6 h3f4 e8c8 d1e2 g8f6 h2h4 c6a5 g3g4 c7c5 c1e3 h7h6 h4h5 f6d7 e1c1 b7a6 c1b1 d7f6 e3d2 e7d7 f4g2 c8b8 d2c1 d7a4 g2e3 b8a8 b1a1 d8c8');
INSERT INTO evaluations VALUES(10116,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(10116,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(10116,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10116,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(10116,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10116,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10116,36,'cp 12','h7h6 c1e3 c5e3 f2e3 d7d6 d1d2 c8e6 c4b3 e8g8 b1c3 e6b3 a2b3 d6d5 e4d5 f6d5 e1g1 a7a5 f1f2 f7f5 e3e4 d5c3 d2c3 d8d6 e4f5 f8f5 a1f1 a8f8 c3c4 g8h8 f3d2 c6d4 f2f5 f8f5 f1f5 d4f5 d2e4 d6d4 c4d4 f5d4 c2c3 d4e6 g1h1');
INSERT INTO evaluations VALUES(9772,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(9772,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(9772,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(9772,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9772,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(9772,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(9772,36,'cp 23','b1c3 a7a6 c1e3 c5e3 f2e3 d7d6 d3d4 e8g8 e1g1 b7b5 c4d3 f8e8 d1d2 e5d4 e3d4 b5b4 c3d5 f6d5 e4d5 c6e7 d2b4 c8f5 d3f5 e7f5 b4d2 f5e3 f1e1 e3d5 c2c4 d5f6 e1e8 d8e8 b2b3 e8d7 g1h2 c7c6 d2f4 a8e8 a1e1 e8e1 f3e1');
INSERT INTO evaluations VALUES(9754,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9754,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(9754,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9754,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9754,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9754,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9754,36,'cp 13','a7a6 c3d5 d7d6 e1g1 e8g8 c2c3 c5a7 d5f6 d8f6 c4b3 c8e6 c1e3 a7e3 f2e3 c6e7 g1h2 c7c6 a2a4 e6b3 d1b3 a8b8 f3d2 f6g6 a4a5 d6d5 b3b4 e7c8 e4d5 c6d5 e3e4 f8d8 b4c5 g6e6 e4d5 e6d5 c5d5 d8d5 d2c4 f7f6');
INSERT INTO evaluations VALUES(7252,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(7252,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7252,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(7252,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(7252,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7252,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(7252,36,'cp 4','c3d5 d7d6 e1g1 e8g8 c2c3 c5a7 d5f6 d8f6 c4b3 c8e6 c1e3 a8d8 e3a7 c6a7 d3d4 a7c8 f1e1 c8e7 b3e6 f6e6 d1b3 e6b3 a2b3 f7f6 h3h4 e7c6 a1a4 g8f7 d4d5 c6e7 f3d2 f6f5 e1d1 f7e8 h4h5 e8d7 b3b4 f5e4 d2e4 f8f4');
INSERT INTO evaluations VALUES(38666,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(38666,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38666,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38666,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38666,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38666,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38666,36,'cp 12','e2e3 g8f6 f4d6 d8d6 c2c4 e8g8 b1c3 d5c4 f1c4 c7c5 c4e2 b8c6 c3b5 d6e7 d4c5 e7c5 e1g1 f8d8 a1c1 c5e7 d1c2 a7a6 b5c3 c8d7 h2h3 h7h6 f1d1 c6b4 c2b3 d7e8 a2a3 b4d5 f3d4 d5c3 c1c3 f6e4 c3c2 a8c8 c2c8 d8c8');
INSERT INTO evaluations VALUES(38665,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(38665,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(38665,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(38665,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(38665,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(38665,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(38665,36,'cp 12','g8f6 f4d6 d8d6 c2c4 e8g8 b1c3 d5c4 f1c4 c7c5 e1g1 b8c6 c3b5 d6e7 d4c5 e7c5 c4e2 f8d8 a1c1 c5e7 d1c2 a7a6 b5c3 c8d7 a2a3 a8c8 b2b4 c6a7 c2b2 h7h6 f1d1 d7e8 h2h3 d8d1 e2d1 a7b5 c3b5 c8c1 b2c1 e8b5 c1c8 g8h7 d1c2 g7g6');
INSERT INTO evaluations VALUES(50387,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50387,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(50387,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(50387,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(50387,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(50387,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(50387,36,'cp 12','c2c4 e7e5 f1g2 f8c5 b1c3 c7c6 e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 b8c6 e1g1 e8g8 c1e3 c8g4 d1d3 g4e2 d3e2 c6d4 e3d4 b6d4 a1d1 d8b6 d1d2 g7g6 g2d5 a8d8 e2f3 g8g7 f1d1 d4f6 h2h4 f6b2 d5b7 d8d2 d1d2 f8d8');
INSERT INTO evaluations VALUES(49549,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49549,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(49549,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(49549,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(49549,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(49549,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(49549,36,'cp -11','f8c5 g1f3 b8c6 h2h3 d7d6 f1e2 c6d4 c1e3 d4f3 e2f3 c5e3 f2e3 c7c6 b1c3 b7b5 c4b5 c6b5 e1g1 e8g8 b2b4 a7a5 d1d2 a5b4 c3b5 d8b6 d2b4 b6e3 g1h1 e3d3 a2a4 c8d7 a1d1 d3c2 b4d6 c2a4 b5c3 a4c6 d6e5');
INSERT INTO evaluations VALUES(42584,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(42584,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(42584,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(42584,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(42584,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(42584,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(42584,36,'cp -8','g1f3 d7d6 h2h3 b8c6 f1e2 c6d4 f3d4 c5d4 b1c3 c7c6 c1e3 b7b5 c4b5 d4e3 f2e3 c6b5 e1g1 e8g8 c3b5 d8b6 d3d4 f6e4 e2d3 d6d5 a1c1 a8b8 d3e4 d5e4 c1c5 a7a6 b5c3 b6b2 g1h2 c8e6 d4d5 b2b4 c3e4 e6h3 d1d3 f7f5 h2h3 f5e4 f1f8 b8f8 d3c3 b4c3');
INSERT INTO evaluations VALUES(42582,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(42582,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(42582,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(42582,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(42582,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(42582,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(42582,36,'cp -23','b8c6 f1e2 d7d6 h2h3 h7h6 a2a3 a7a5 c1e3 f6h5 e3d2 c6e7 f3h4 h5f6 b1c3 e8g8 e1g1 e7g6 h4g6 f7g6 d2e3 c5e3 f2e3 c8e6 d1d2 g6g5 f1f2 g5g4 h3g4 d8d7 g4g5 h6g5 d3d4 f6g4 f2f1 a5a4');
INSERT INTO evaluations VALUES(38254,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(38254,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(38254,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(38254,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(38254,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(38254,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(38254,36,'cp 60','b1c3 g8f6 c1f4 b8c6 d1d2 e8g8 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 h1e1 a8d8 f4e3 d8d1 c3d1 f8e8 g2g3 e7d6 f2f3 e6e5 d1f2 d6d5 c1b1 b7b6 a2a3 h7h6 g3g4 d5c4 h2h4 a6a5 g4g5 f6d5 e3d2 h6g5 h4g5');
INSERT INTO evaluations VALUES(38251,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(38251,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38251,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(38251,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(38251,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(38251,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(38251,36,'cp 60','g8f6 c1f4 b8c6 d1d2 c6d4 d2d4 e8g8 e1c1 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 h1e1 a8d8 f4e3 d8d1 c3d1 f8e8 g2g3 e7d6 f2f3 c7c5 d1c3 d6c6 h2h4 b7b5 g3g4 a6a5 g4g5 f6d5 e3e4 e8f8 e1d1 d5e7 h4h5 c6e4 c3e4 f8f3');
INSERT INTO evaluations VALUES(50332,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50332,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(50332,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50332,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50332,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50332,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50332,36,'cp 0','e7e6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(49419,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49419,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49419,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49419,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49419,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49419,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49419,36,'cp 0','c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(49401,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49401,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49401,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49401,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49401,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49401,36,'cp 0','f6h5 f4d2 h5f6 d2f4');
INSERT INTO evaluations VALUES(57009,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(57009,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(57009,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(57009,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57009,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(57009,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57009,36,'cp 13','g8f6 f1g2 d7d5 g1f3 d5c4 d1a4 c8d7 a4c4 c7c5 f3e5 d8c8 b1c3 b8c6 e5d7 c8d7 e1g1 a8c8 e2e3 h7h5 f1d1 h5h4 d2d4 c5d4 e3d4 f8e7 c1e3 c6b4 c4e2 b4d5 e3d2 e8g8 a1c1 b7b6 a2a3 f8d8 c3d5 f6d5');
INSERT INTO evaluations VALUES(49386,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(49386,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49386,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(49386,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(49386,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(49386,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(49386,36,'cp 14','f1g2 d7d5 d2d4 f8b4 c1d2 b4e7 g1f3 e8g8 d1c2 c7c6 e1g1 b7b6 c4d5 c6d5 f1c1 c8b7 f3e5 f6e4 a2a3 e7d6 d2b4 a7a5 b4d6 e4d6 c2c7 b7a6 e2e3 f7f6 e5f3 d8c7 c1c7 f8c8 c7c8 d6c8 b1c3 b8c6 c3a4');
INSERT INTO evaluations VALUES(49387,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(49387,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(49387,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(49387,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(49387,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(49387,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49387,36,'cp 24','d7d5 g1f3 d5c4 d1a4 c8d7 a4c4 c7c5 d2d4 b7b5 c4c2 b8a6 e1g1 a8c8 c2d1 c5d4 f3d4 d8b6 b1c3 f8e7 c1e3 b6b8 c3e4 f6d5 e3g5 e7g5 e4g5 h7h6 g2d5 e6d5 g5f3 a6c5 d4b3 e8g8 b3c5 c8c5 d1d2 f8c8 a1c1 c5c1 f1c1 c8c1 d2c1');
INSERT INTO evaluations VALUES(48243,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(48243,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(48243,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(48243,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48243,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(48243,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(48243,36,'cp 24','g1f3 d5c4 d1a4 c8d7 a4c4 c7c5 d2d4 b7b5 c4c2 b8a6 e1g1 a8c8 c2d1 c5d4 f3d4 d8b6 d4b3 a6c5 c1e3 f6g4 e3c5 f8c5 b3c5 c8c5 a2a4 g4f6 b2b4 c5c4 d1b3 e8g8 b1d2 c4c8 a4a5 b6d4 d2f3 d4e4 e2e3 e4c4 b3c4 c8c4');
INSERT INTO evaluations VALUES(48242,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48242,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(48242,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(48242,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(48242,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(48242,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(48242,36,'cp 23','d5d4 e1g1 c7c5 e2e3 b8c6 d2d3 h7h6 e3d4 c5d4 c1f4 f8d6 f4d6 d8d6 b1d2 a7a5 a1c1 e8g8 f1e1 a5a4 c4c5 d6e7 f3e5 c6e5 e1e5 e7c7 d2f3 c8d7 e5e1 a8a5 f3e5 a5c5 c1c5 c7c5 e5d7 f6d7 d1a4 d7e5 b2b4 c5d6 a4b3 f8c8 g2b7 c8b8');
INSERT INTO evaluations VALUES(28956,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(28956,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(28956,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(28956,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(28956,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28956,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(28956,36,'cp 33','c4d5 c6d5 d1b3 f5c8 g1f3 e7e6 f3e5 b8c6 f2f4 f8e7 c1d2 e8g8 f1d3 c8d7 e1g1 a8c8 a2a3 a7a6 f1c1 c6a5 b3a2 d7e8 b2b4 a5c6 a2b3 c6a7 a3a4 a7c6 c3e2 c6e5 d4e5 f6e4 d3e4 c8c1 a1c1 d5e4 e2d4 e8c6 h2h3 c6d5 b3c2');
INSERT INTO evaluations VALUES(28940,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(28940,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(28940,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(28940,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(28940,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28940,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28940,36,'cp 21','c6d5 d1b3 f5c8 g1f3 e7e6 c1d2 b8c6 f1d3 f8d6 e1g1 e8g8 f1c1 h7h6 h2h3 c8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d8b6 b3b6 a7b6 d4d5 e6d5 e4d5 d6e7 d2c3 f8d8 a2a3 d7f5 c1d1 b6b5 d5b3 f5d3 f3e5 c6e5 c3e5 a8c8');
INSERT INTO evaluations VALUES(29131,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(29131,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(29131,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(29131,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(29131,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(29131,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29131,36,'cp 24','d1b3 f5c8 f1d3 e7e6 g1f3 b8c6 c1d2 f8d6 e1g1 e8g8 f1d1 h7h6 e3e4 d5e4 c3e4 f6e4 d3e4 d8b6 b3b6 a7b6 d2e1 c6a5 f3e5 f7f6 b2b4 f6e5 b4a5 b6a5 d4e5 d6e5 a1b1 e5f6 b1b5 f8d8 d1c1 c8d7 b5b7 a8c8 c1c8 d7c8');
INSERT INTO evaluations VALUES(47206,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(47206,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(47206,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(47206,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(47206,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(47206,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47206,36,'cp 17','g1f3 c7c5 c2c4 d5c4 d1a4 b8c6 d4c5 e7e6 f1g2 f8c5 b1c3 d8a5 a4c4 e8g8 e1g1 c8d7 c1d2 c5e7 e2e4 a5h5 h2h3 f8d8 d2e3 d7e8 f1d1 h7h6 c4b3 b7b6 b3c4 a8c8 d1d8 c8d8 a1c1 e7c5 g3g4 c6e5 c4c5 e5f3 g2f3 h5c5 e3c5 b6c5');
INSERT INTO evaluations VALUES(31066,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(31066,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(31066,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(31066,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(31066,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(31066,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(31066,36,'cp 40','e7e6 d1b3 d8c7 c1f4 d5c4 e2e4 c4b3 f4c7 f5g6 a2b3 g8f6 f3d2 f8b4 f2f3 a7a5 e1f2 b7b5 g2g4 e8g8 h2h4 h7h5 g4g5 f6e8 c7f4 f7f5 g5f6 e8f6 f1h3 f6e4 d2e4 f8f4 f2g3 f4f8 e4g5 g6f7 c3e4 f8e8 g5f7 g8f7');
INSERT INTO evaluations VALUES(31444,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(31444,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(31444,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(31444,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(31444,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(31444,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(31444,36,'cp 54','d1b3 d8c7 g2g3 b8d7 f3h4 f5g4 h2h3 d5c4 b3c4 g4h5 e2e4 f8e7 f1e2 h5e2 c4e2 g8f6 e1g1 e8g8 c1f4 c7b6 f1d1 a7a5 h4g2 a5a4 g2e3 a4a3 b2b3 b6a5 a1c1 a5h5 d4d5 h5e2 c3e2 c6d5 c1c7 d5e4 d1d7 f6d7 c7d7 e7f6 b3b4 f8d8 d7b7');
INSERT INTO evaluations VALUES(31441,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(31441,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(31441,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(31441,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(31441,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(31441,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(31441,36,'cp 55','d8c7 c1f4 d5c4 f4c7 c4b3 e2e4 f5g6 a2b3 g8f6 f3d2 f8b4 f2f3 a7a5 e1f2 b7b5 g2g4 e8g8 h2h4 h7h5 g4g5 f6e8 c7f4 f7f5 g5f6 e8f6 h1g1 g6h7 f2e3 c6c5 f1b5 c5d4 e3d4 f6e4 c3e4 f8f4 a1c1 b8a6 d4e3 f4f8 c1c2 f8d8 g1g5 h7e4 d2e4');
INSERT INTO evaluations VALUES(52916,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52916,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52916,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(52916,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52916,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(52916,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(52916,36,'cp 26','g8f6 d1e2 f6e4 e2e4 b8d7 d2d4 d7f6 e4d3 g7g6 f1e2 c8f5 d3b3 d8b6 c2c3 f6d5 e1g1 a7a5 f1e1 f8g7 e2f1 b6b3 a2b3 e8g8 c1d2 b7b6 g2g3 f8d8 f1g2 e7e6 f3e5 g7e5 e1e5 f7f6 e5e1 f5c2 a1c1 c2d3 c3c4 d5b4');
INSERT INTO evaluations VALUES(45048,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(45048,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(45048,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(45048,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(45048,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(45048,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45048,36,'cp 27','d1e2 f6e4 e2e4 b8d7 c2c3 d7f6 e4c2 c8g4 f3e5 g4h5 d2d4 e7e6 f1e2 h5e2 c2e2 d8c7 c1f4 f8d6 e1g1 e8g8 f1e1 f6d5 f4g3 d6e5 g3e5 c7d7 h2h4 h7h6 h4h5 b7b5 e2g4 f7f6 e5f4 g8h8 f4g3 f8e8 e1e2 a7a5 a1e1 a5a4');
INSERT INTO evaluations VALUES(45049,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(45049,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(45049,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(45049,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(45049,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(45049,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45049,36,'cp 25','f6e4 e2e4 d8a5 c2c3 c8f5 e4f4 e7e6 d2d4 b8d7 a2a4 f8e7 f1e2 d7f6 f4g3 f6h5 g3e5 a5d8 e2c4 e8g8 e5e2 e7d6 e1g1 d8c7 f3e5 h5f6 c1f4 f6g4 f1e1 g4e5 f4e5 a7a5 e5g3 d6g3 h2g3 h7h6 c4b3 a8d8');
INSERT INTO evaluations VALUES(51492,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(51492,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51492,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(51492,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(51492,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(51492,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(51492,36,'cp 31','g8f6 g2g3 e7e6 f1g2 c8b7 c2c4 f8b4 c1d2 c7c5 d2b4 c5b4 a2a3 b4a3 b1a3 e8g8 e1g1 a7a5 d1d2 b8a6 a3b5 d7d6 d2f4 d6d5 b5d6 d8d7 d6b7 d7b7 f3e5 a6b4 g2f3 a8c8 f1c1 h7h6 h2h4 b7e7 g1g2 f6d7 c4d5 b4d5 f4e4 d7f6 e5c6 f6e4 c6e7 d5e7 f3e4');
INSERT INTO evaluations VALUES(43879,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(43879,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(43879,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(43879,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43879,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(43879,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(43879,36,'cp 25','g2g3 c8b7 c2c4 e7e6 f1g2 f8b4 c1d2 c7c5 e1g1 b4d2 d1d2 c5d4 d2d4 b8c6 d4h4 c6e7 b1c3 e8g8 a1c1 d8b8 f1d1 d7d5 c4d5 e7d5 h4d4 d5c3 d4c3 b7d5 f3e5 f8c8 c3e3 c8c1 d1c1 b8b7 g2d5 b7d5 e3f3 a8d8 f3d5 f6d5 e5c6');
INSERT INTO evaluations VALUES(43880,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(43880,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43880,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(43880,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(43880,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(43880,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(43880,36,'cp 36','c8b7 c2c4 e7e6 f1g2 f8e7 e1g1 e8g8 b1c3 f6e4 c1d2 d7d5 c4d5 e6d5 d1c2 b8d7 f1d1 e4c3 b2c3 c7c5 d2f4 f8e8 a2a4 e7f6 h2h4 a8c8 e2e3 h7h6 c2b3 g7g5 h4g5 h6g5 f4d6 e8e6 d6c5 d7c5');
INSERT INTO evaluations VALUES(28813,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(28813,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(28813,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(28813,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(28813,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28813,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(28813,36,'cp 37','c2c4 e7e6 f1g2 f8b4 c1d2 b4d2 d1d2 d7d5 e1g1 e8g8 f1c1 d5c4 b1a3 b8d7 a3c4 a7a5 c1d1 d8e7 a1c1 f8d8 d2c3 c7c5 d4c5 e7c5 c3d2 f6e4 d2d4 a8c8 f3e5 c5d4 d1d4 d7e5 d4d8 c8d8');
INSERT INTO evaluations VALUES(28803,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(28803,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(28803,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28803,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(28803,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(28803,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(28803,36,'cp 34','e7e6 f1g2 f8b4 c1d2 c7c5 e1g1 b4d2 d1d2 c5d4 d2d4 b8c6 d4h4 e8g8 b1c3 c6e7 f1d1 d8b8 a1c1 d7d5 c4d5 e7d5 h4d4 d5c3 d4c3 b7d5 f3e5 f8c8 c3e3 c8c1 d1c1 b8b7 g2d5 e6d5 e3a3 a8e8 a3d6 b7e7 d6e7 e8e7 c1c8 e7e8 c8e8 f6e8 g1g2 e8d6');
INSERT INTO evaluations VALUES(28720,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(28720,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(28720,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(28720,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(28720,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(28720,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(28720,36,'cp 26','f1g2 f8b4 c1d2 c7c5 e1g1 e8g8 d2b4 c5b4 a2a3 b4a3 b1a3 a7a5 a3b5 b8a6 d1d2 d7d6 d2f4 d6d5 b2b3 b7c6 f1c1 c6b5 c4b5 a6b4 f3e5 d8e7 g2f3 a8c8 h2h4 b4c2 e5c6 c8c6 b5c6 c2a1 c1a1 f8c8 a1c1 e7c7 f4e5 c7e5 d4e5 f6e8 f3g2 e8c7 e2e4 g8f8');
INSERT INTO evaluations VALUES(28721,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(28721,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(28721,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(28721,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28721,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(28721,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(28721,36,'cp 26','f8b4 c1d2 c7c5 d2b4 c5b4 a2a3 b4a3 b1a3 e8g8 e1g1 d7d6 a3b5 a7a5 b2b3 b8a6 d1c1 d6d5 c1f4 b7c6 b5d6 d8e7 c4c5 b6c5 a1a5 c5d4 f3d4 a6b4 a5a8 f8a8 f1c1 c6d7 g2f3 e6e5 d6f5 e5f4 f5e7 g8f8 d4c6 d7c6 e7c6 b4c6');
INSERT INTO evaluations VALUES(28275,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(28275,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(28275,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(28275,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(28275,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(28275,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(28275,36,'cp 23','c1d2 c7c5 d2b4 c5b4 a2a3 b4a3 b1a3 e8g8 e1g1 d7d6 d1d2 a7a5 b2b3 b8c6 a3b5 c6b4 d2f4 f6e4 h2h4 d8e7 f3g5 e6e5 f4c1 e5d4 b5d4 a8e8 c1f4 e4g5 f4g5 b7g2 g5e7 e8e7 g1g2 d6d5 a1c1 d5c4 c1c4 e7d7 e2e3 f8d8 f1c1 g7g6 c4c7 d7c7 c1c7');
INSERT INTO evaluations VALUES(28274,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(28274,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(28274,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(28274,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(28274,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(28274,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(28274,36,'cp 31','c7c5 e1g1 e8g8 d2b4 c5b4 d1d3 d7d5 c4d5 e6d5 f1c1 b8c6 b1d2 a7a5 d3b5 c6a7 b5a4 d8d6 f3e5 f8c8 g2h3 c8c7 c1c7 d6c7 e2e3 b7c8 h3f1 c8e6 a4d1 c7d6 e5d3 a8c8 d2f3 c8c7 a1c1 c7c1 d1c1 d6c6 c1c6 a7c6 f1e2 h7h6 f3e5 c6e5 d4e5');
INSERT INTO evaluations VALUES(16058,30,'cp -145',NULL);
INSERT INTO evaluations VALUES(16058,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(16058,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(16058,33,'cp -148',NULL);
INSERT INTO evaluations VALUES(16058,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(16058,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(16058,36,'cp -133','c6d4 f3d1 d7d5 e4d5 b7b5 c4b3 a7a5 c2c3 a5a4 c3d4 a4b3 c1g5 b3a2 b1c3 f8e7 d4e5 f6d5 g5e7 e8e7 g1e2 d5c3 b2c3 c8e6 d1d2 c7c5 e1g1 d8b6 c3c4 b5c4 d2g5 e7d7 d3c4 e6c4 f1e1 h8d8 e2c3 b6e6 c3e4 d7c6 g5e3 d8d4 e4c3 e6f5 e1c1');
INSERT INTO evaluations VALUES(16081,30,'cp -154',NULL);
INSERT INTO evaluations VALUES(16081,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(16081,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(16081,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(16081,34,'cp -140',NULL);
INSERT INTO evaluations VALUES(16081,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(16081,36,'cp -146','f3d1 d7d5 e4d5 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 d3d4 d8e7 d4e5 e7e5 d1e2 e5e6 g1f3 f8g7 e2e6 c8e6 b3b4 a5a4 e1g1 c7c6 b1a3 e8g8 h2h3 f8e8 f1d1 d5c3 b2c3 g7c3 c1d2 c3a1 d1a1 a8d8 d2e3 e6d5 f3d4 d5e4 a1c1 d8d6 g1h2 f7f6 c1c5 e4d5 c5c1 e8d8');
INSERT INTO evaluations VALUES(16082,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(16082,31,'cp -139',NULL);
INSERT INTO evaluations VALUES(16082,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(16082,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(16082,34,'cp -161',NULL);
INSERT INTO evaluations VALUES(16082,35,'cp -148',NULL);
INSERT INTO evaluations VALUES(16082,36,'cp -147','d7d5 e4d5 b7b5 c4b3 a7a5 c2c3 a5a4 b3a4 a8a4 c3d4 e5d4 c1g5 f8b4 b1d2 e8g8 g1f3 f8e8 e1f1 b4e7 g5f6 e7f6 d2e4 f6g7 d1b3 c8b7 b3b5 a4a7 a1e1 e8e7 d5d6 c7d6 h2h4 a7a5 b5b4 b7a6 e4d6 a5a2 b4b3 a2a5 d6c4 e7e1 f3e1 a5a1 h4h5 g7f8 h5g6 h7g6 h1h3 a1b1');
INSERT INTO evaluations VALUES(14537,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(14537,31,'cp -139',NULL);
INSERT INTO evaluations VALUES(14537,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(14537,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(14537,34,'cp -124',NULL);
INSERT INTO evaluations VALUES(14537,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(14537,36,'cp -136','e4d5 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 d3d4 d8e7 d4e5 e7e5 d1e2 e5e6 e2e6 c8e6 g1f3 f8g7 b3b4 a5a4 e1g1 e8c8 b1a3 c7c6 a3c2 c8b7 c1g5 f7f6 g5h4 e6f7 f1e1 h8e8 a1d1 g7f8 f3d4 e8e1 d1e1 d8d7 g2g4 g6g5 h4g3 d5b6 h2h3');
INSERT INTO evaluations VALUES(14536,30,'cp -145',NULL);
INSERT INTO evaluations VALUES(14536,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(14536,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(14536,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(14536,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(14536,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(14536,36,'cp -138','b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 d3d4 d8e7 d4e5 e7e5 d1e2 e5e6 e2e6 c8e6 g1f3 f8g7 b3b4 a5a4 e1g1 c7c6 b1a3 e8c8 a3c2 c8b7 c2d4 e6g4 c1g5 d8d7 h2h3 g4f3 d4f3 h7h6 g5e3 h8d8 g2g3 d5b6 e3b6 b7b6 f1e1 g7f6 g1g2 c6c5 b4c5 b6c5 e1e4 d7d1 a1d1 d8d1 e4e8 d1b1 f3d2');
INSERT INTO evaluations VALUES(19367,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(19367,31,'cp -148',NULL);
INSERT INTO evaluations VALUES(19367,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(19367,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(19367,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(19367,35,'cp -144',NULL);
INSERT INTO evaluations VALUES(19367,36,'cp -129','e7e6 c4b3 g8f6 h5e2 d7d5 d2d3 c6d4 e2e3 f6g4 e3d2 c5c4 c2c3 d4b3 a2b3 c4d3 d2d3 d5e4 d3d8 e8d8 h2h3 g4e5 c1e3 b7b5 f2f3 c8b7 f3e4 b7e4 g1f3 e5f3 g2f3 e4f3 e1g1 f3h5 b1a3 f8e7 f1f2 e7a3 a1a3 a7a5 c3c4 b5b4 e3b6 d8c8 a3a5 a8a5 b6a5');
INSERT INTO evaluations VALUES(18728,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(18728,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(18728,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(18728,33,'cp -143',NULL);
INSERT INTO evaluations VALUES(18728,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(18728,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(18728,36,'cp -136','c4b3 g8f6 h5e2 c6d4 e2e3 d7d5 d2d3 f6g4 e3d2 g4e5 c2c3 d4b3 a2b3 f8e7 f2f4 e5c6 g1f3 e8g8 e1g1 a7a5 d2c2 b7b6 f1d1 d8c7 c1e3 c8a6 b1a3 a8d8 g1h1 d8d7 e4d5 d7d5 a3c4 b6b5 c4e5 c6e5 f3e5 f7f6 c3c4 f6e5 c4d5');
INSERT INTO evaluations VALUES(53381,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53381,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53381,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(53381,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(53381,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(53381,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(53381,36,'cp 13','b8c6 c2c3 g8f6 h2h3 e7e5 d4e5 c6e5 d1e2 d8e7 c1e3 c8d7 g1f3 e5d3 e2d3 e7e4 d3e2 f8d6 b1d2 e4e6 e1g1 e8g8 e2d3 f8e8 f1e1 f6e4 a1d1 d7c6 f3d4 e6g6 d4c6 b7c6 c3c4 e4f6 d3g6 h7g6 c4d5 f6d5 g1f1 d5e3 e1e3 e8e3 f2e3 a8e8 d2c4 d6c5 d1d7 c5e3');
INSERT INTO evaluations VALUES(19859,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19859,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(19859,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19859,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(19859,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19859,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(19859,36,'cp 23','c2c3 g8f6 c1f4 c8g4 d1b3 c6a5 b3a4 a5c6 b1d2 e7e6 h2h3 g4h5 g1f3 f8d6 f4d6 d8d6 e1g1 a7a6 f3e5 e8g8 f1e1 f8c8 a4b3 b7b5 a2a3 a8b8 e5c6 d6c6 b3c2 h5g6 d3g6 h7g6 d2f3 c8c7 c2e2 c6b6 h3h4 b8c8 f3e5 a6a5 g2g3 b5b4 a3b4');
INSERT INTO evaluations VALUES(41065,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41065,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41065,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41065,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(41065,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(41065,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(41065,36,'cp 37','e2e3 h7h6 g5h4 e8g8 f1d3 c7c6 d1c2 f8e8 g1e2 a7a5 e1g1 b8a6 a2a3 a6c7 a1c1 c7e6 h2h3 c8d7 h4g3 b7b5 g3e5 a8c8 f1d1 e7f8 e2g3 b5b4 c3a4 c6c5 e5f6 d7a4 c2a4 d8f6 d3b5 e8d8 d4c5 c8c5 c1c5 e6c5');
INSERT INTO evaluations VALUES(41061,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(41061,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(41061,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(41061,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(41061,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(41061,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41061,36,'cp 28','h7h6 g5h4 c7c6 f1d3 e8g8 d1c2 f8e8 g1e2 a7a5 e1g1 b8d7 a2a3 b7b5 f1c1 c8a6 c3b1 a8c8 b2b4 a6b7 b1d2 a5b4 a3b4 e7b4 d3f5 b4e7 h4f6 e7f6 d2b3 f6e7 b3a5 d8c7 e2f4 e7d6 f4d5 d6h2 g1h1 c6d5 c2c7 c8c7 h1h2 c7c1 a1c1');
INSERT INTO evaluations VALUES(40722,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(40722,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40722,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(40722,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40722,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40722,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40722,36,'cp 29','g5h4 e8g8 f1d3 c7c6 g1e2 f8e8 e1g1 b8d7 d1c2 a7a5 a1c1 d7f8 h4g3 f6h5 g3e5 e7f6 e5f6 h5f6 a2a3 f8e6 h2h3 c8d7 e2g3 b7b5 f2f3 d8b6 f1d1 a5a4 c3e2 a8c8 c2d2 c6c5 d4c5 e6c5 d3c2 b6e6 c2f5 e6e3 d2e3 e8e3 f5d7');
INSERT INTO evaluations VALUES(40725,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40725,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(40725,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40725,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40725,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(40725,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40725,36,'cp 32','e8g8 f1d3 c7c6 g1e2 b8d7 a2a4 f6e4 h4e7 d8e7 d3e4 d5e4 a4a5 d7f6 d1a4 c8e6 e1g1 h6h5 f1c1 h5h4 h2h3 f8e8 e2f4 e6f5 a4a3 e7a3 b2a3 g7g5 f4e2 g8g7 c1b1 a8b8 b1b2 f5e6 a1b1 e8e7 b1c1 b7b6 a5b6 a7b6');
INSERT INTO evaluations VALUES(36206,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36206,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36206,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(36206,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(36206,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(36206,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(36206,36,'cp 40','f1d3 f8e8 g1e2 a7a5 h2h3 c7c6 e1g1 b8d7 a1c1 f6h5 h4e7 d8e7 d1d2 e7h4 g1h2 a5a4 e2g1 h4e7 a2a3 h5f6 d2c2 b7b5 c3e2 a8a6 h2h1 f6e4 g1f3 e7d8 h1g1 d8e7 f1e1 e4d6');
INSERT INTO evaluations VALUES(36209,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36209,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(36209,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(36209,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(36209,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36209,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(36209,36,'cp 21','c7c6 g1e2 b7b5 e1g1 b8d7 a1c1 f8e8 a2a4 a7a6 c3b1 d8b6 d1c2 c8b7 b1d2 a6a5 h4f6 d7f6 d2b3 b5a4 b3c5 e8b8 c5a4 b6b4 c1a1 b7c8 c2c6 c8d7 c6c2 d7a4 a1a4 b4b2 c2b2 b8b2 e2c3 e7d6');
INSERT INTO evaluations VALUES(35731,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(35731,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(35731,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35731,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(35731,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(35731,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(35731,36,'cp 32','g1e2 a7a5 e1g1 b8d7 a1c1 f8e8 h2h3 f6h5 h4e7 d8e7 d1d2 d7f8 c3a4 e7d8 b2b3 f8e6 f1e1 c8d7 d3f5 a8c8 f5g4 h5f6 g4e6 d7e6 a4c5 d8c7 e2g3 c8a8 c5d3 c7d6 d3f4 f6e4 g3e4 d5e4 c1c5 b7b6 c5e5');
INSERT INTO evaluations VALUES(35730,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(35730,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(35730,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(35730,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(35730,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(35730,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(35730,36,'cp 22','a7a5 e1g1 b8d7 a1c1 f8e8 h2h3 d7f8 h4g3 e7d6 a2a3 d6g3 e2g3 f8e6 c3b1 c8d7 b1d2 d8b6 b2b3 c6c5 d4c5 e6c5 d3f5 a8c8 f5d7 c5d7 g3f5 b6e6 f5d4 e6e7 c1c8 e8c8 d1a1 e7c5 d2f3 c5c3 a1c3 c8c3');
INSERT INTO evaluations VALUES(45437,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(45437,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(45437,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45437,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(45437,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(45437,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45437,36,'cp 13','b8c6 c1f4 c8f5 d1b3 c6a5 b3a4 f5d7 a4c2 a8c8 e2e3 b7b5 a2a3 a7a6 f1d3 e7e6 e1g1 f8e7 h2h3 e8g8 f1c1 h7h6 f3e5 d8b6 d3h7 g8h8 g2g4 a5c4 h7d3 h8g8 c2e2 b6d8 h3h4 c4e5 d4e5 f6h7 g4g5 h6g5');
INSERT INTO evaluations VALUES(13150,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(13150,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(13150,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(13150,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(13150,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(13150,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(13150,36,'cp 16','c1f4 c8f5 d1b3 c6a5 b3a4 f5d7 a4c2 a8c8 e2e3 e7e6 f1d3 b7b5 a2a3 a7a6 e1g1 f8e7 h2h3 e8g8 f1c1 d8b6 f3e5 h7h6 f4g3 d7c6 a3a4 b5b4 c3e2 c6b7 c2d1 e7d6 b2b3 c8c1 a1c1 f8c8 c1c8 b7c8 d1c2 b6c7 f2f3 c7c2 d3c2 c8b7');
INSERT INTO evaluations VALUES(13130,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(13130,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13130,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(13130,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(13130,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(13130,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(13130,36,'cp 15','a7a6 e2e3 c8g4 h2h3 g4f3 d1f3 e7e6 f1d3 c6b4 d3b1 a8c8 e1g1 f8d6 a2a3 b4c6 f4g5 h7h6 g5h4 e8g8 b1d3 d6e7 f1c1 c6a5 c1c2 d8d7 f3d1 b7b5 a1c1 a5b3 c1b1 b3a5 b2b3 c8c6 h4f6 e7f6 c3e2 f8c8 c2c6 c8c6');
INSERT INTO evaluations VALUES(11141,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11141,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(11141,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11141,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(11141,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(11141,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(11141,36,'cp 12','d1b3 e7e6 e2e3 f8e7 f1d3 e8g8 a1c1 b7b5 e1g1 c8d7 c3e2 d8b6 f3e5 c6e5 f4e5 f8c8 c1c8 a8c8 f1c1 c8c1 e2c1 b6a5 g1f1 g7g6 b3c2 f6e8 a2a3 a5b6 c1b3 f7f6 e5g3 e8d6');
INSERT INTO evaluations VALUES(39615,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(39615,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(39615,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(39615,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(39615,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(39615,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(39615,36,'cp 62','d2d4 d7d6 c1e3 a7a6 a2a4 g8f6 f1e2 b8c6 g1f3 f6g4 e3g5 h7h6 g5h4 g6g5 h4g3 e7e5 d4e5 g4e5 f3e5 c6e5 c3d5 c8e6 c2c3 d8d7 a4a5 h6h5 h2h4 g5g4 e1g1 e8c8 a1b1 c8b8 d1d2 e6d5 e4d5 e5g6');
INSERT INTO evaluations VALUES(49392,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(49392,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49392,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49392,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49392,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(49392,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49392,36,'cp 0','d7d5 g1f3 f8e7 b1d2 e8g8 f1e2 b7b6 b2b3 c7c5 c4d5 f6d5 c1b2 c8b7 e1g1 b8d7 a1c1 a8c8 a2a3 e7f6 d2e4 f6e7');
INSERT INTO evaluations VALUES(49480,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(49480,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(49480,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(49480,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49480,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(49480,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(49480,36,'cp -20','f8b4 c1d2 a7a5 e2e3 b8a6 f1d3 a6c5 d3c2 b4d2 b1d2 c7c6 g1e2 c6d5 c4d5 b7b5 e1g1 d7d6 e2c3 c8g4 d1e1 g4d7 c3e4 c5e4 d2e4 f6e4 c2e4 e8g8 e1d2 d8b6 a2a3 g7g6 h2h3 f8c8 f1c1 h7h6 b2b4 c8c4 e4d3 c4c1 a1c1 g8g7 f2f3 a5b4');
INSERT INTO evaluations VALUES(42699,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(42699,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42699,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42699,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(42699,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42699,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(42699,36,'cp -12','c1d2 a7a5 e2e3 b8a6 f1d3 a6c5 d3c2 b4d2 b1d2 c7c6 g1e2 c6d5 c4d5 b7b5 e2c3 c8a6 c3e4 c5e4 d2e4 b5b4 d1f3 f6e4 f3e4 d7d6 c2d3 a6d3 e4d3 e8g8 a1c1 a8c8 c1c6 c8c6 d5c6 d8c7 d3b5 f8c8 e1g1 c7c6 b5a5 c6c5 a5a4 g7g6 h2h3 g8g7 a2a3 c8b8 f1d1 b4a3 b2a3 b8b2 a4d7');
INSERT INTO evaluations VALUES(42698,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(42698,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(42698,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(42698,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42698,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42698,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(42698,36,'cp -16','a7a5 e2e3 b8a6 f1d3 a6c5 d3c2 b4d2 b1d2 c7c6 g1e2 c6d5 c4d5 b7b5 e2c3 e8g8 a2a3 d8b6 e1g1 b5b4 c3e4 c8a6 e4f6 b6f6 a3b4 a5b4 f1e1 g7g6 d2e4 c5e4 c2e4 f6d6 e4d3 g8g7 a1a6 a8a6 d3a6 d6a6 h2h4 a6b6');
INSERT INTO evaluations VALUES(39790,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(39790,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(39790,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(39790,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(39790,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(39790,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(39790,36,'cp -7','e2e3 b8a6 f1d3 a6c5 d3c2 f6e4 a2a3 b4d2 b1d2 e4d2 d1d2 d7d6 g1e2 a5a4 e1c1 c8f5 e2c3 e8g8 f2f3 f5c2 d2c2 d8g5 h1e1 f7f5 c3b5 g5e7 f3f4 h7h6 c1b1 e7d7 f4e5 d6e5 b5c3 b7b6 g2g3 g8h8 c2e2 a8e8 e3e4 c5b3');
INSERT INTO evaluations VALUES(57578,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(57578,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(57578,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(57578,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(57578,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(57578,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(57578,36,'cp 52','g1f3 g8f6 g2g3 g7g6 d2d4 f8g7 f1g2 d7d6 e1g1 e8g8 b1c3 c7c6 d4d5 e7e5 e2e4 c6d5 c4d5 b8a6 c1e3 f5f4 g3f4 c8g4 f4e5 f6d7 e3g5 g4f3 g5d8 f3d1 d8e7 d1f3 e5e6 f3g2 g1g2 d7c5 e7d6 g7c3 d6f8 g8f8 b2c3 c5e4');
INSERT INTO evaluations VALUES(57576,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(57576,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(57576,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(57576,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(57576,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(57576,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(57576,36,'cp 72','g7g6 d2d4 f8g7 b1c3 g8f6 b2b4 a7a5 b4b5 d7d6 e2e3 b8d7 f1e2 e7e5 d4e5 f6e4 c1b2 e4c3 b2c3 d8e7 d1d2 d6e5 c3a5 b7b6 a5b4 d7c5 e1g1 c8b7 f1d1 e8g8 a1c1 b7f3 g2f3 e7g5 g1h1 f8d8 d2c2 g5e7 b4c5 e7c5 f3f4 d8d1 c1d1 e5e4');
INSERT INTO evaluations VALUES(56140,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(56140,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(56140,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(56140,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(56140,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(56140,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(56140,36,'cp 147','d1e2 g8f6 b1c3 f8e7 c3e4 f6d5 e4c3 d5b4 e2d1 c8f5 f1b5 b8c6 b5a4 e8g8 e1g1 f5g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 c3e2 d8a5 e2d4 a5a4 d2d3 e7f6 d4b3 b4d5 a2a3 f8e8 b3d2 a8d8 d2e4 f6d4 c2c3 a4d1 f1d1 d4e5 g1f1 f7f5 e4g3 f5f4 g3e4 f4f3 g2g3 e5c7');
INSERT INTO evaluations VALUES(56141,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(56141,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(56141,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(56141,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(56141,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(56141,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(56141,36,'cp 110','g8f6 d2d3 d8d5 b1d2 b8c6 d3e4 d5e6 h2h3 f8c5 e4e5 f6d7 d2b3 c5b6 c1f4 a7a5 c2c3 e8g8 e2e4 a5a4 f1c4 e6g6 e4g6 h7g6 b3d2 d7c5 e1g1 c8f5 f1e1 c6a5 c4e2 f5e6 f4e3 f8d8 f3d4 e6d5 c3c4 d5e4 d2e4 c5e4');
INSERT INTO evaluations VALUES(48574,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(48574,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(48574,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(48574,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(48574,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(48574,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(48574,36,'cp 130','d2d3 d8d5 b1c3 f8b4 f3d2 b4c3 b2c3 d5c6 c1b2 e8g8 d3e4 f8d8 f2f3 c8e6 d2b3 c6b6 b2c1 b6d6 c1e3 d6a3 e2b5 a3b2 f1d3 b2c3 e3d2 c3c6 b3c5 b7b6 b5c6 b8c6 c5e6 f7e6 d3c4 g8f7 d2f4 a8c8 c4b3 c6a5');
INSERT INTO evaluations VALUES(48573,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(48573,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(48573,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(48573,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(48573,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(48573,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(48573,36,'cp 113','d8d5 b1d2 b8c6 d3e4 d5e6 e2c4 h7h6 f1d3 g7g5 h2h3 f8g7 c2c3 e6c4 d3c4 e8g8 e1g1 f8e8 f1e1 c8e6 c4b5 a7a5 d2f1 a5a4 a2a3 e8d8 f1g3 f6d7 g3f5 e6f5 e4f5 d7c5 c1e3 c5d3 e1e2 c6e5 f3e5');
INSERT INTO evaluations VALUES(33906,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(33906,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(33906,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(33906,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(33906,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(33906,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(33906,36,'cp 122','b1c3 f8b4 f3d2 b4c3 b2c3 d5c6 c1b2 c8e6 d3e4 a7a6 c3c4 b8d7 f2f3 d7b6 e2d3 e8g8 b2d4 a8d8 c4c5 b6a4 d3e3 c6d7 c2c3 d7e7 a1b1 f8e8 e1f2 b7b5 c5b6 c7c5 d4f6 e7f6 e4e5 f6g6 d2e4 e6f5 b6b7 g6e6 b1e1 e6b6 f1c4 f5e4 e3e4 a4c3 e4c2 c3b5 a2a4');
INSERT INTO evaluations VALUES(31521,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31521,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31521,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(31521,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(31521,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31521,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31521,36,'cp 0','f5g4 c1e3 g8e7 h2h3 g4f3 d1f3 b8d7 e1c1 a7a6 c1b1 d8c7 c3a4 c7a5 a4c3');
INSERT INTO evaluations VALUES(32103,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(32103,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(32103,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(32103,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(32103,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(32103,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(32103,36,'cp 113','d1f3 d8d7 c2c3 g7g6 d2d3 b8c6 e1g1 f8g7 a2a4 g8h6 a4a5 a7a6 c1g5 f7f5 g5h4 f5f4 b1d2 h6f7 f3d1 e8g8 d1g4 d7e8 g4e6 e8e6 c4e6 c6d8 e6c4 h7h5 f2f3 g8h7 h4f2 g6g5 a1b1 g7f6 b2b4 f8g8 d3d4 g5g4 h3g4 h5g4 f3g4');
INSERT INTO evaluations VALUES(32101,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(32101,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(32101,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(32101,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(32101,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(32101,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(32101,36,'cp 108','d8d7 e1g1 b8c6 c2c3 g7g6 d2d3 f8g7 a2a4 g8h6 b2b4 e8g8 b1d2 g8h8 f1e1 h6g8 f3d1 f7f5 d2f3 a8e8 a4a5 c6d8 d1a4 d7a4 a1a4 g7h6 f3g5 g8f6 b4b5 f6d7 a4a1 f5e4 d3e4 h8g7 c1e3 h6g5 e3g5 d7c5 f2f3 d8e6 g5e3');
INSERT INTO evaluations VALUES(56895,30,'cp 863',NULL);
INSERT INTO evaluations VALUES(56895,31,'cp 878',NULL);
INSERT INTO evaluations VALUES(56895,32,'cp 882',NULL);
INSERT INTO evaluations VALUES(56895,33,'cp 890',NULL);
INSERT INTO evaluations VALUES(56895,34,'cp 891',NULL);
INSERT INTO evaluations VALUES(56895,35,'cp 891',NULL);
INSERT INTO evaluations VALUES(56895,36,'cp 905','h5e5 g8e7 e5h8 d7d5 b1c3 c8e6 d2d4 a7a6 c1g5 d8d6 e4e5 d6c6 h8h7 b8d7 g1e2 d7b6 e2f4 e7f5 f4e6 c6e6 c3e2 f5g7 e1c1 e6f5 g5h4 g7e6 h7h8 b6c4 c2c3 e6f4 e2f4 f5f4 c1b1 c4d2 b1a1 d2c4 f1c4 d5c4');
INSERT INTO evaluations VALUES(56893,30,'cp 886',NULL);
INSERT INTO evaluations VALUES(56893,31,'cp 892',NULL);
INSERT INTO evaluations VALUES(56893,32,'cp 894',NULL);
INSERT INTO evaluations VALUES(56893,33,'cp 894',NULL);
INSERT INTO evaluations VALUES(56893,34,'cp 894',NULL);
INSERT INTO evaluations VALUES(56893,35,'cp 894',NULL);
INSERT INTO evaluations VALUES(56893,36,'cp 898','d8e7 e5h8 g8f6 d2d3 f6g4 g1f3 e7f6 h8h7 b8c6 b1c3 c6b4 c1g5 f6g7 h7g7 f8g7 a1c1 d7d6 a2a3 b4a6 d3d4 c7c6 f1d3 a6c7 h2h3 g4f6 g5e3 e8f8 e1g1 c7e6 f3g5 e6g5 e3g5 f6e8 g5e3 c8d7 f2f4 g7f6 d3c4 f8g7 e4e5 f6e7 d4d5 e8c7 d5c6 b7c6');
INSERT INTO evaluations VALUES(35246,30,'cp 909',NULL);
INSERT INTO evaluations VALUES(35246,31,'cp 917',NULL);
INSERT INTO evaluations VALUES(35246,32,'cp 904',NULL);
INSERT INTO evaluations VALUES(35246,33,'cp 917',NULL);
INSERT INTO evaluations VALUES(35246,34,'cp 913',NULL);
INSERT INTO evaluations VALUES(35246,35,'cp 917',NULL);
INSERT INTO evaluations VALUES(35246,36,'cp 919','e5h8 g8f6 d2d3 f6g4 g1f3 f7f6 h2h3 g4e5 f3e5 f6e5 c1h6 d7d6 h6f8 e7f8 h8h7 f8f7 h7h8 f7f8 h8f8 e8f8 b1c3 c7c6 a2a4 a7a5 h3h4 f8e7 g2g3 c8e6 f1h3 e6f7 f2f3 b8a6 e1d2 a8g8 h3g4 a6c5 d2e3 g8g7 b2b3 d6d5 c3e2 d5e4 d3e4 e7d6 c2c4 c5e6');
INSERT INTO evaluations VALUES(34109,30,'cp 909',NULL);
INSERT INTO evaluations VALUES(34109,31,'cp 898',NULL);
INSERT INTO evaluations VALUES(34109,32,'cp 888',NULL);
INSERT INTO evaluations VALUES(34109,33,'cp 888',NULL);
INSERT INTO evaluations VALUES(34109,34,'cp 894',NULL);
INSERT INTO evaluations VALUES(34109,35,'cp 890',NULL);
INSERT INTO evaluations VALUES(34109,36,'cp 890','g8f6 d2d3 f6g4 g1f3 f7f6 h2h3 g4e5 f3e5 f6e5 c1h6 d7d6 h6f8 e7f8 h8h7 f8f7 h7h8 f7f8 h8f8 e8f8 b1c3 c7c6 a2a4 a7a5 h3h4 f8e7 g2g3 b8a6 e1d2 c8e6 f1h3 e6f7 f2f3 b7b6 h3g4 a8f8 a1f1 f8e8 c3e2 a6c5 b2b3 d6d5 e2c3 e7d6 e4d5 c6d5 d3d4 c5d7 d4e5 d7e5');
INSERT INTO evaluations VALUES(33504,30,'cp 905',NULL);
INSERT INTO evaluations VALUES(33504,31,'cp 901',NULL);
INSERT INTO evaluations VALUES(33504,32,'cp 904',NULL);
INSERT INTO evaluations VALUES(33504,33,'cp 909',NULL);
INSERT INTO evaluations VALUES(33504,34,'cp 909',NULL);
INSERT INTO evaluations VALUES(33504,35,'cp 909',NULL);
INSERT INTO evaluations VALUES(33504,36,'cp 919','d2d3 f6g4 g1f3 f7f6 h2h3 g4e5 f3e5 f6e5 c1h6 d7d6 h6f8 e7f8 h8h7 f8f7 h7h8 f7f8 h8f8 e8f8 b1c3 c7c6 a2a4 a7a5 c3d1 b8a6 f2f3 f8e7 d1e3 a6c5 e3c4 c8e6 e1f2 e6f7 b2b3 a8a6 h3h4 c5e6 a1a3 e6d4 c2c3 f7c4 d3c4 d4e6 f2e3 e6c5 h4h5 g6h5 h1h5');
INSERT INTO evaluations VALUES(33503,30,'cp 905',NULL);
INSERT INTO evaluations VALUES(33503,31,'cp 904',NULL);
INSERT INTO evaluations VALUES(33503,32,'cp 905',NULL);
INSERT INTO evaluations VALUES(33503,33,'cp 913',NULL);
INSERT INTO evaluations VALUES(33503,34,'cp 909',NULL);
INSERT INTO evaluations VALUES(33503,35,'cp 911',NULL);
INSERT INTO evaluations VALUES(33503,36,'cp 905','d7d6 c1g5 b8d7 b1c3 c7c6 g1f3 h7h6 g5h6 f6g4 h6g5 f7f6 g5d2 e7g7 h8g7 f8g7 d3d4 d7f8 d2f4 e8e7 h2h3 g4h6 e1c1 h6f7 f1c4 f8e6 f4d2 g6g5 a2a4 c8d7 d4d5 c6d5 c3d5 e7f8 d2c3 e6c5 d5f6 g7f6 c3f6 c5e4 f6d4 d7c6 h1e1 e4c5 d4c5 d6c5');
INSERT INTO evaluations VALUES(57040,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(57040,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(57040,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(57040,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(57040,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(57040,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(57040,36,'cp -41','c7c5 f3e2 b8c6 g2g3 g7g6 f1g2 g8e7 d2d3 f8g7 g1f3 h7h6 e1g1 e8g8 e4e5 d7d6 e5d6 d8d6 b1d2 d6c7 c2c3 e7d5 d2c4 b7b6 f1e1 c8b7 f3e5 c6e5 c4e5 a8d8 h2h4 g8h7 c1d2 d5f6 g2b7 c7b7 a1d1 f6d5 h4h5 g7e5 h5g6 f7g6');
INSERT INTO evaluations VALUES(52000,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(52000,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(52000,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(52000,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(52000,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(52000,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(52000,36,'cp -34','f3e2 b8c6 g2g3 g7g6 d2d3 g8e7 f1g2 f8g7 g1f3 e8g8 e1g1 h7h6 e4e5 b7b6 b1d2 d7d6 e5d6 e7f5 d2c4 c8b7 c1f4 b6b5 c4d2 c6d4 e2d1 e6e5 f3d4 b7g2 d4f5 g2f1 f5h6 g8h7 f4e3 g7h6 e3h6 h7h6 d2f1 d8d6');
INSERT INTO evaluations VALUES(16392,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(16392,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(16392,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(16392,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(16392,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(16392,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(16392,36,'cp -57','c6a5 c4b5 a7a6 b5a4 b7b5 a4c2 d7d5 d2d3 f8e7 f3e2 d5e4 d3e4 c7c5 g1f3 d8c7 e1g1 e8g8 b1d2 c5c4 f1d1 a5b7 a2a4 c8e6 d2f1 h7h6 f3h4 b5b4 c3b4 e7b4 c1d2 b4d2 f1d2 a6a5 h4f5 e6f5 e4f5 b7d6 h2h3 f8e8');
INSERT INTO evaluations VALUES(17052,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(17052,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(17052,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(17052,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(17052,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(17052,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(17052,36,'cp -48','c4b5 a7a6 b5a4 c7c5 d2d3 f8e7 f3e2 b7b5 a4c2 d7d5 g1f3 d5e4 d3e4 d8c7 c1g5 h7h6 g5f6 e7f6 b1d2 c8e6 e1g1 e8g8 f1d1 f8d8 h2h3 a5c4 d2c4 e6c4 c2d3 c4e6 c3c4 g7g6 c4b5 c5c4 b5b6 c7c6 b6b7 a8b8 e2c2 b8b7 d3e2 d8d1 a1d1 g8g7 f3d2 b7d7');
INSERT INTO evaluations VALUES(31158,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31158,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(31158,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31158,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(31158,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31158,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(31158,36,'cp 12','e7e6 g1f3 f5g4 c1e3 d8b6 a1b1 b8d7 f1e2 g8e7 e1g1 e7f5 f3e1 g4e2 c3e2 c6c5 c2c3 f8e7 e2g3 f5e3 f2e3 c5d4 c3d4 e8g8 g3h5 e7g5 d1e2 f7f6 e1f3 g5h6 e5f6 d7f6 h5f6 g7f6 b1e1 a8c8 g1h1 b6a6 e2a6 b7a6 e3e4 d5e4');
INSERT INTO evaluations VALUES(31523,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(31523,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(31523,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31523,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(31523,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31523,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31523,36,'cp 0','f1d3 g8e7 g1f3 f5g4 c3e2 c6c5 c2c3 b8c6 c1e3 g4f3 g2f3 c5d4 c3d4 e7g6 e1g1 g6h4 f3f4 g7g6 a2a3 a7a6 b2b4 h4f3 g1g2 f3h4 g2h1 h4f3');
INSERT INTO evaluations VALUES(31524,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31524,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31524,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31524,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(31524,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31524,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(31524,36,'cp 0','f5d3 c2d3 g8e7 g1f3 e7f5 h2h4 h7h5 c1g5 f7f6 g5f4 c6c5 e1g1 b8c6 d4c5 f8c5 a1c1 c5b6 c3a4 e8g8 e5f6 d8f6 f4g5 f6f7 a4b6 a7b6 a2a4 f5d4 f3d4 c6d4 g5e3 d4f5 e3g5');
INSERT INTO evaluations VALUES(31488,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31488,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31488,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31488,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31488,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31488,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31488,36,'cp 12','c2d3 g8e7 g1f3 e7f5 h2h4 h7h5 c3e2 c6c5 d4c5 f8c5 d3d4 c5e7 c1g5 d8a5 e1f1 b8c6 h1h3 g7g6 f1g1 a5b6 g5e7 c6e7 f3g5 a8c8 h3b3 b6a6 g2g3 c8c4 b3d3 a6b6 d1d2 e8g8 a1c1 c4c1 e2c1 a7a5 c1e2 f8c8 d3b3 b6a6 g1g2 c8c4');
INSERT INTO evaluations VALUES(38374,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(38374,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(38374,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(38374,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(38374,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(38374,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(38374,36,'cp 50','e2e4 g8f6 f1e2 e8g8 g1f3 e7e5 e1g1 c8g4 d4d5 a7a5 a2a3 b8a6 c1e3 b7b6 d1c2 g4d7 f3d2 f6e8 a1b1 f7f5 f2f3 f5f4 e3f2 h7h5 b2b4 g6g5 b4a5 a6c5 a5b6 c7b6 c2b2 a8a6 h2h3 e8f6 c3b5 g5g4 h3g4 h5g4');
INSERT INTO evaluations VALUES(46172,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(46172,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(46172,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46172,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(46172,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46172,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(46172,36,'cp 17','c4d5 f6d5 g1f3 f8e7 f1d3 b8c6 a2a3 c5d4 e3d4 e8g8 e1g1 b7b6 d1e2 c8b7 e2e4 g7g6 c3d5 d8d5 e4d5 e6d5 c1d2 a8c8 f1e1 e7f6 f3e5 f8d8 d2c3 c6e5 d4e5 f6e7 a1c1 d5d4 c3d2 c8c1 e1c1 d8c8 b2b4 c8c1 d2c1 f7f6 e5f6 e7f6 f2f3');
INSERT INTO evaluations VALUES(46133,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46133,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46133,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(46133,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(46133,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(46133,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(46133,36,'cp 0','f6d5 f1d3 b8c6 g1f3 f8e7 a2a3 c5d4 e3d4 e8g8 e1g1 b7b6 d1c2 d5f6 c1e3 h7h6 f1e1 c8b7 a1d1 e7d6 c2d2 f8e8 e3h6 g7h6 d4d5 e6d5 e1e8 f6e8 d2h6 d8f6 d3h7 g8h8 h7g6 h8g8 g6h7');
INSERT INTO evaluations VALUES(46065,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46065,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46065,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(46065,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46065,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46065,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46065,36,'cp 7','g1f3 b8c6 e3e4 d5c3 b2c3 c5d4 c3d4 f8b4 c1d2 e8g8 d2b4 c6b4 d1d2 d8d6 f1c4 f8d8 e1g1 b7b6 a2a3 b4c6 f1d1 h7h6 d2e3 c8d7 d4d5 e6d5 a1c1 d5d4 f3d4 c6d4 d1d4 d6e7 c1d1 d7e6 c4e6 f7e6 e3b3 d8d4 d1d4 a8c8');
INSERT INTO evaluations VALUES(46066,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46066,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46066,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46066,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46066,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46066,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46066,36,'cp 0','b8c6 e3e4 d5c3 b2c3 c5d4 c3d4 f8b4 c1d2 e8g8 d2b4 c6b4 d1d2 d8d6 f1c4 f8d8 e1g1 b7b6 f1e1 c8a6 a1c1 h7h6 c4a6 b4a6 g2g3 a8c8 c1c8 d8c8 e1c1 c8c1 d2c1 a6b4 c1c8 g8h7 c8c4 b4c6 g1g2 c6e7 c4d3 h7g8 d3a6 e7c6 a6c4 c6e7');
INSERT INTO evaluations VALUES(13462,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(13462,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(13462,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(13462,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13462,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13462,36,'cp 20','f1d3 f8e7 e1g1 e8g8 a2a3 c5d4 e3d4 b7b6 d1c2 d5f6 c1e3 h7h6 f1e1 c8b7 a1d1 e7d6 c2c1 f8e8 f3e5 a8c8 d3b1 d6f8 c1c2 c6e7 e3f4 e7f5 d4d5 b7d5 c2e2 g7g6 b1f5 g6f5 g2g4 f6g4 e5g4 f5g4 c3d5 e6d5 e2g4 f8g7 e1e8 d8e8 d1d5');
INSERT INTO evaluations VALUES(9333,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9333,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(9333,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(9333,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9333,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(9333,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9333,36,'cp 7','a7a6 c3d5 f6d5 c4d5 e8g8 c2c3 a6a5 h2h3 h7h6 d5b3 c8e6 b3e6 f7e6 c1e3 c5e3 f2e3 c6e7 d1b3 d8d7 d3d4 a5a4 b3c4 e7g6 f1f2 c7c6 a1d1 b7b5 c4d3 d7e7 d4e5 d6e5 g1h2 f8d8 d3e2 d8d1 e2d1 g8h7');
INSERT INTO evaluations VALUES(7170,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(7170,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7170,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7170,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(7170,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7170,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(7170,36,'cp 14','a2a4 e8g8 c3d5 f6d5 c4d5 d8f6 c2c3 c5a7 d5b3 a6a5 h2h3 c8e6 b3e6 f6e6 f1e1 a8e8 c1e3 a7e3 e1e3 c6e7 b2b4 b7b6 d3d4 e7g6 d1c2 g6f4 h3h4 a5b4 c3b4 e5d4 f3d4 e6f6 a1d1 f4e6 d4c6 f6h4 e3g3');
INSERT INTO evaluations VALUES(39336,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(39336,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(39336,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(39336,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(39336,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(39336,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(39336,36,'cp 51','c3d5 b4c5 b2b4 c5e7 g1f3 g8f6 f3e5 e8g8 f1e2 f6d5 e4d5 e7b4 e1g1 b4e7 e5c4 d7d6 d2d4 f8e8 c4e3 b8d7 a2a4 g7g6 a4a5 e7f8 c1b2 f8g7 c2c4 d7f6 f1e1 b7b6 a5b6 c7b6 d1c2 a7a5 c4c5 d6c5 d4c5 b6c5');
INSERT INTO evaluations VALUES(39313,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(39313,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(39313,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(39313,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(39313,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39313,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(39313,36,'cp 43','b4c5 b2b4 c5e7 g1f3 g8f6 f3e5 e8g8 f1e2 f6d5 e4d5 e7b4 e1g1 b4e7 d2d4 d7d6 e5c4 f8e8 c4e3 e7f8 c2c4 b8d7 a1b1 d7f6 d1c2 g7g6 c1b2 f8g7 f1e1 h7h5 h2h3 c7c6 e2f3 c6d5 f3d5 f6d5 e3d5');
INSERT INTO evaluations VALUES(21069,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(21069,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(21069,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(21069,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(21069,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(21069,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(21069,36,'cp 53','g1f3 c8f5 e2e3 a7a6 f1e2 e7e6 e1g1 h7h6 c2c4 g8f6 d2d4 f8e7 e2d3 f5d3 d1d3 c6b4 d3e2 d5c4 e2c4 a8c8 a2a3 b4d5 b3b4 c7c6 b1d2 a6a5 e3e4 d5f4 c4b3 a5b4 a3b4 e8g8 g2g3 f4h3 g1g2 h3g5 f1e1 f6d7 h2h3 g5f3 d2f3 d8b6');
INSERT INTO evaluations VALUES(21067,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(21067,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(21067,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(21067,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(21067,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(21067,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(21067,36,'cp 38','c8f5 e2e3 a7a6 f1e2 e7e6 e1g1 g8f6 c2c4 h7h6 d2d4 f8e7 b1c3 e8g8 e2d3 f5d3 d1d3 e7b4 d3e2 f8e8 a1c1 c6e7 a2a3 b4c3 c1c3 e7g6 c3c1 c7c6 f1d1 d8e7 a3a4 a6a5 f3e1 f6d7 e2g4 g8h8 g2g3');
INSERT INTO evaluations VALUES(26584,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(26584,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(26584,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(26584,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(26584,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(26584,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(26584,36,'cp 44','e2e3 a7a6 f1e2 g8f6 e1g1 e7e6 c2c4 h7h6 d2d4 f8e7 b1d2 f6e4 a2a3 c6b8 d2e4 f5e4 f3d2 e4h7 b3b4 e8g8 d1b3 c7c6 b2c3 b8d7 a3a4 h7g6 g2g3 g6h7 c4c5 e7f6 b4b5 e6e5 b5c6 b7c6');
INSERT INTO evaluations VALUES(26583,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26583,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(26583,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(26583,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(26583,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(26583,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(26583,36,'cp 47','a7a6 f1e2 e7e6 e1g1 h7h6 c2c4 g8f6 d2d4 f8e7 f3e5 c6b8 e2d3 f5d3 e5d3 e8g8 d1e2 c7c6 b1d2 b8d7 a1c1 f6e4 d2b1 e4d6 a2a4 a6a5 f2f3 d5c4 b3c4 c6c5 d4c5 d6f5 f1d1 d8c7 e3e4 f5h4 b1c3 d7c5 c3b5 c7b6 b2d4 f8d8 d4e3');
INSERT INTO evaluations VALUES(53292,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(53292,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(53292,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53292,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(53292,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(53292,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(53292,36,'cp -11','f1b5 c8d7 b5d7 d8d7 d2d3 g8f6 b1c3 b8c6 g1f3 e7e6 e1g1 f8e7 h2h3 e8g8 e4e5 f6e8 d3d4 c5d4 c3e2 e8c7 e2d4 c7a6 c1e3 a6c5 c2c3 c6d4 f3d4 f7f6 d4f3 f6f5 e3c5 e7c5 g1h2 a8c8 f3d4 c5d4 d1d4 a7a6 f1d1 c8c6 d4f2 d7c7 a2a4 c6c4 d1d4 c4d4');
INSERT INTO evaluations VALUES(53066,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53066,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53066,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53066,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53066,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53066,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53066,36,'cp -14','g7g6 f1e2 g8f6 b1c3 b8c6 e1g1 f8g7 c3d5 e7e5 c1g5 h7h6 g5f6 g7f6 d5f6 d8f6 c2c3 e8g8 d1d2 g8g7 f3e1 f6e7 f2f4 e5f4 f1f4 f7f5 f4f1 f5e4 d3e4 f8f1 e2f1 c6e5 e1d3 e5d3 f1d3 c8e6 d3b5 a8d8 a1d1 e6a2 c3c4');
INSERT INTO evaluations VALUES(52622,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52622,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52622,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52622,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52622,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52622,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52622,36,'cp 0','f1e2 g8f6 b1c3 b8c6 c3d5 f8g7 c2c3 f6d5 e4d5 c6e5 e1g1 e8g8 f1e1 b7b6 d3d4 e5f3 e2f3 c5d4 c3d4 c8f5 c1g5 g7f6 g5h6 f6g7');
INSERT INTO evaluations VALUES(21856,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(21856,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21856,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(21856,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(21856,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21856,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(21856,36,'cp 29','g8e7 d2d4 a7a6 b5e2 c6d4 f3d4 e5d4 d1d4 e7c6 d4d1 f8e7 b1c3 e8g8 e1g1 f8e8 f2f3 a8b8 c3d5 e7h4 c1e3 c6e7 c2c4 c7c5 d5e7 d8e7 e3f4 e8d8 d1d2 h4f6 a1d1 f6d4 g1h2 b7b5 b2b3 c8e6 f4e3 d4e3 d2e3 b5c4 e2c4 e6c4 b3c4 a6a5 f3f4');
INSERT INTO evaluations VALUES(55218,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(55218,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(55218,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(55218,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(55218,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(55218,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(55218,36,'cp 38','e2e3 c7c5 f1c4 g8f6 e1g1 e7e6 d2d4 a7a6 d4c5 d8d1 f1d1 f8c5 c4e2 e8g8 b1d2 b8c6 a2a3 a6a5 d2c4 b7b6 c1d2 c8a6 e2f1 a6b7 d2e1 f8d8 d1d8 a8d8 f3d2 f6d7 d2e4 c5e7 c4d6 b7a8 d6b5 a8b7 a1c1 h7h6');
INSERT INTO evaluations VALUES(55217,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55217,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(55217,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55217,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(55217,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55217,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(55217,36,'cp 19','c8e6 b1a3 e6d5 f1c4 e7e6 e1g1 g8f6 c4d5 d8d5 b2b3 c7c5 c1b2 f8e7 d2d4 e8g8 d4c5 d5c5 f3d4 b8d7 a1c1 c5g5 d1f3 g5d5 a3c4 a8c8 f1d1 f8d8 f3d5 f6d5 g1f1 d7c5 f1e2 a7a6');
INSERT INTO evaluations VALUES(31790,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(31790,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31790,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(31790,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31790,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(31790,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(31790,36,'cp 15','b1c3 b8c6 f3g5 e6f5 f1c4 e7e6 g5f3 g8f6 c4b5 f8d6 d1e2 f5g6 e3e4 f6d7 d2d4 g6h5 b5c6 b7c6 h2h3 h7h6 e1g1 e8g8 f1d1 f7f5 e4e5 d6b4 e2c4 d8e7 c4c6 d7b6 c3e2 b6d5 a2a3 b4a5 c6c5 a5b6 c5e7 d5e7 e2c3 h5f3 g2f3');
INSERT INTO evaluations VALUES(48520,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(48520,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(48520,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(48520,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(48520,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(48520,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(48520,36,'cp 71','e4f6 g7f6 c1e3 b8c6 g1e2 c8d7 g2g3 c6b4 f1g2 d7c6 g2c6 b4c6 c2c4 d8d7 d1c2 f8b4 e2c3 b4c3 c2c3 e8c8 e1c1 c6e7 h1e1 e7f5 b2b3 f5e3 c3e3 c8b8 d4d5 e6d5 d1d5 d7h3 e3e7 d8d5 c4d5 h8c8 e7f7 h3h6 c1b1 h6h2 f7f6 h2h5 f6d4 a7a5 b1b2 h5f3 e1e7');
INSERT INTO evaluations VALUES(48152,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(48152,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(48152,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(48152,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(48152,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(48152,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(48152,36,'cp 93','g7f6 c1e3 b8c6 g1e2 c8d7 g2g3 c6b4 f1g2 d7c6 g2c6 b4c6 c2c4 d8d7 d1c2 f8b4 e2c3 e8c8 e1c1 b4c3 c2c3 c6e7 d4d5 e7f5 b2b3 d7e7 e3f4 e6e5 f4d2 c7c6 d5c6 b7b6 h1e1 f5d4 c3e3 d4c6 f2f4 e7e6 d2c3 d8d1 e1d1 h8d8 d1d8 c8d8 f4e5 f6e5 e3d3 d8e8 d3h7');
INSERT INTO evaluations VALUES(12073,30,'cp -187',NULL);
INSERT INTO evaluations VALUES(12073,31,'cp -242',NULL);
INSERT INTO evaluations VALUES(12073,32,'cp -198',NULL);
INSERT INTO evaluations VALUES(12073,33,'cp -218',NULL);
INSERT INTO evaluations VALUES(12073,34,'cp -237',NULL);
INSERT INTO evaluations VALUES(12073,35,'cp -254',NULL);
INSERT INTO evaluations VALUES(12073,36,'cp -217','h7h6 g5h3 c8g4 f2f3 g4h3 g2h3 d8d4 b1c3 d4h4 e1e2 f8b4 d1e1 h4h5 e2f2 b4c3 d2c3 e8g8 b2b4 a8e8 h1f1 e5e4 f2g1 a5c4 a4c6 e8e6 c6e4 f6e4 f3e4 h5h3 g1h1 e6e4 e1f2 g7g5 f1g1 h3h5 f2g3 f8e8 c1f4 e4e2 a1f1 g8f8 f4c1 e2c2 g1g2 c2g2 h1g2 f8g8 g3f3 e8e2 f1f2 e2f2 f3f2');
INSERT INTO evaluations VALUES(12050,30,'cp -222',NULL);
INSERT INTO evaluations VALUES(12050,31,'cp -209',NULL);
INSERT INTO evaluations VALUES(12050,32,'cp -214',NULL);
INSERT INTO evaluations VALUES(12050,33,'cp -216',NULL);
INSERT INTO evaluations VALUES(12050,34,'cp -221',NULL);
INSERT INTO evaluations VALUES(12050,35,'cp -227',NULL);
INSERT INTO evaluations VALUES(12050,36,'cp -250','g5h3 d8d4 b1c3 c8g4 f2f3 g4h3 g2h3 d4h4 e1e2 f8b4 d1e1 h4h5 e2f2 e8g8 f2g2 e5e4 h1f1 e4f3 f1f3 b4c3 d2c3 a8e8 e1f1 e8e4 b2b3 f8e8 c1f4 e4e2 g2g1 f6d5 f4g3 e8e3 f3f5 h5g6 f5f3 g6e4 f3e3 e4e3 g3f2 e3e4 a1d1 d5e3 d1d4 e4g6 d4g4 e3f1 g4g6 f7g6 g1f1 e2c2 f2a7 c2a2 a7b6 a2a1 f1f2 a5c4 b6d4');
INSERT INTO evaluations VALUES(39520,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(39520,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(39520,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(39520,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(39520,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(39520,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(39520,36,'cp 0','f1c4 g8f6 e4e5 f6h5 b1c3 d7d6 d2d4 b8d7 e1g1 e8g8 f3e1 g7g6 c4e2 d6e5 e2h5 g6h5 d4e5 d7e5 d1h5 e7g5 h2h4 g5f6 f1f4 e5g6 f4e4 f6d4 g1h1 f7f5 c1g5 f5e4 g5d8 f8f1 h1h2 d4g1 h2h1 g1f2 h1h2');
INSERT INTO evaluations VALUES(39515,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(39515,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(39515,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(39515,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(39515,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(39515,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(39515,36,'cp -52','e7h4 e1f1 h4e7 b1c3 g8f6 d1e2 e8g8 d2d4 c7c6 c1f4 d7d5 c4d3 b8a6 a2a3 a6c7 a1e1 c8e6 h2h3 d5e4 c3e4 e6d5 f1f2 d5e4 d3e4 e7d6 f4e5 d6e5 d4e5 f6e4 e2e4 c7e6 f2g3 d8c7 h1f1 a8d8 g3h2 f8e8 e4h4 h7h6 h4g3');
INSERT INTO evaluations VALUES(39431,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(39431,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(39431,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(39431,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(39431,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(39431,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(39431,36,'cp -47','e1f1 h4e7 b1c3 g8f6 d2d3 d7d6 c1f4 c7c6 h2h3 b8d7 a2a3 d7c5 f4e3 e8g8 f1f2 c8e6 e3c5 d6c5 h1e1 f6h5 f2g1 e6c4 d3c4 d8b6 a1b1 a8d8 d1e2 f7f6 e4e5 f6f5 e2f1 h5f4 g1h1 f4e6 e1d1 b6c7 d1d8 f8d8 b1e1');
INSERT INTO evaluations VALUES(39430,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(39430,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(39430,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(39430,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(39430,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(39430,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(39430,36,'cp -45','h4e7 b1c3 g8f6 d1e2 e8g8 d2d4 c7c6 e4e5 f6h5 g2g4 f4g3 f3g5 e7g5 e2h5 g5c1 a1c1 d7d5 c4d3 h7h6 h1g1 d8g5 h5g5 h6g5 g1g3 f7f6 e5f6 f8f6 f1g1 c8f5 d3f5 f6f5 c1f1 f5f1 g1f1 b8d7 g3g5 a8f8 f1g2 f8e8 g2f3 e8e1 c3e2 e1h1 g5g2 g8f7 e2f4 a7a5 g2e2 d7f6 f4d3 f6e4');
INSERT INTO evaluations VALUES(39514,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(39514,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(39514,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(39514,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(39514,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(39514,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(39514,36,'cp -26','d2d4 g8f6 e4e5 f6h5 f3g5 e7g5 d1h5 d7d5 e5d6 e8g8 c4d3 g5h6 d6c7 d8c7 b1c3 b8d7 c3d5 c7d8 c1d2 d7b6 d5b6 d8b6 d2c3 c8e6 f1f2 f8d8 h5c5 b6c5 d4c5 g7g5 a1d1 d8d5 b2b4 a8d8 d1e1 h6g7 c3g7 g8g7 h2h4 h7h6 h4g5 h6g5 h1h7 g7f6 h7h6 f6e7 h6h7 d5d4 d3f5');
INSERT INTO evaluations VALUES(39517,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(39517,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(39517,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(39517,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(39517,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(39517,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(39517,36,'cp -10','g8f6 e4e5 f6h5 f3g5 d7d5 e5d6 e7g5 d1h5 e8g8 c4d3 g5h6 b1c3 b8d7 d6c7 d8c7 c3d5 c7d8 c1d2 d7b6 d5b6 d8b6 d2c3 f8e8 f1f2 e8e3 h1f1 c8e6 f2g1 a8e8 a1e1 b6d8 e1e3 f4e3 a2a4 a7a6 h2h3 d8d6 f1f6 d6g3 f6f3 g3g5 h5g5 h6g5 d4d5 e6d7 a4a5 h7h6 f3f1 e3e2 f1e1 d7b5 d3b5');
INSERT INTO evaluations VALUES(49663,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(49663,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49663,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(49663,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49663,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(49663,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(49663,36,'cp 0','f1g2 f8c5 g1e2 d7d5 e4d5 f6d5 d2d4 e5d4 e2d4 d8e7 d4e2 d5f6 e1g1 e8g8 e2f4 f8d8 f4d3 b8c6 f1e1 c8e6 g2c6 b7c6 b1c3 c5d4 c1e3 c6c5 e3d4 c5d4 c3e4 e6d5 e4c3 d5e6');
INSERT INTO evaluations VALUES(49664,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49664,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49664,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49664,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49664,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49664,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49664,36,'cp -5','f8c5 b1c3 e8g8 g1e2 c7c6 e1g1 d7d5 e4d5 c6d5 d2d4 e5d4 e2d4 c8g4 d4e2 d8d7 c1e3 c5e3 f2e3 b8c6 d1d2 g4e2 c3e2 a8d8 e2f4 c6e7 a1d1 d7c7 c2c3 c7b6 d2d4 b6b2 d4c5 d8d7 e3e4 e7c6 e4d5 c6e5 c5d4 f8e8 d4a7');
INSERT INTO evaluations VALUES(42655,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(42655,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42655,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42655,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42655,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42655,36,'cp 0','d2d3 e8g8 g1f3 d7d6 e1g1 a7a5 c2c3 b8c6 f1e1 f8e8 a2a4 h7h6 b1a3 f6g4 e1f1 g4f6');
INSERT INTO evaluations VALUES(42646,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42646,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42646,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42646,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42646,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42646,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42646,36,'cp 0','e8g8 g1f3 d7d6 e1g1 a7a5 c2c3 b8c6 f1e1 f8e8 a2a4 h7h6 b1a3 f6g4 e1f1 g4f6');
INSERT INTO evaluations VALUES(52922,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52922,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(52922,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(52922,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(52922,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(52922,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52922,36,'cp 39','c3e4 g8f6 d1e2 f6e4 e2e4 b8d7 d2d4 d7f6 e4f4 d8d6 f1d3 d6f4 c1f4 c8g4 f3e5 g4h5 c2c3 f6d5 f4g3 f7f6 e5c4 a7a5 a2a4 h5g6 d3e2 h7h5 h2h4 e7e6 e1g1 e8f7 g3d6 f8e7 g2g3 h8d8 d6e7 f7e7 f1e1 e7f7 g1g2');
INSERT INTO evaluations VALUES(42444,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(42444,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(42444,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(42444,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42444,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42444,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(42444,36,'cp -5','c7c5 g1f3 h7h6 g5f6 d8f6 a1c1 c5d4 f3d4 b8c6 e2e3 c6d4 e3d4 b7b5 a2a3 b4c3 c1c3 b5c4 f1c4 c8b7 e1g1 e8g8 d1d2 f8c8 h2h3 a7a5 f1e1 f6g5 d2g5 h6g5 e1e5 f7f6 e5b5 b7e4 d4d5 e6d5 c4d5 e4d5 c3c8 a8c8 b5d5');
INSERT INTO evaluations VALUES(39909,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39909,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(39909,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(39909,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(39909,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39909,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39909,36,'cp 0','g1f3 h7h6 g5f6 d8f6 a1c1 e8g8 e2e3 b8c6 a2a3 b4c3 c1c3 c5d4 f3d4 f8d8 d1f3 f6e5 f3g3 e5f6');
INSERT INTO evaluations VALUES(39908,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39908,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39908,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39908,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39908,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39908,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39908,36,'cp -8','h7h6 g5f6 d8f6 a1c1 e8g8 e2e3 b7b6 f1e2 c8b7 e1g1 b4c3 c1c3 d7d6 d4d5 e6d5 c4d5 f6e7 f1e1 f7f5 e2b5 b8d7 b5c6 d7f6 f3h4 b7c6 d5c6 e7e4 h4g6 f8e8 c3d3 d6d5 f2f3 e4e6 g6f4 e6c6 f4d5 f6d5 d3d5 c6f6 b2b3 a8c8 d1e2 c8c6');
INSERT INTO evaluations VALUES(49562,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(49562,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(49562,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(49562,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49562,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49562,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49562,36,'cp 10','f8b4 g1f3 d7d6 c1d2 e8g8 a2a3 b4c3 d2c3 e5e4 f3d2 e4d3 e2e3 c8f5 d1b1 b8d7 f1d3 f5d3 b1d3 d8e7 d3c2 f8e8 e1g1 a7a5 b2b4 f6e4 d2e4 e7e4 c2e4 e8e4 c3d4 e4e8 f2f3 a5b4 a3b4');
INSERT INTO evaluations VALUES(42778,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(42778,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(42778,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42778,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(42778,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42778,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42778,36,'cp 22','g1f3 d7d6 c1d2 b4a5 g2g3 c7c6 f1g2 e8g8 e1g1 h7h6 d1c2 f8e8 c3e4 f6e4 d3e4 a5b6 f3h4 a7a5 h4f5 c8f5 e4f5 b8d7 f1d1 d7f6 e2e3 d6d5 c4d5 c6d5 c2b3 a8c8 a1c1 c8c6 c1c6 b7c6 b3c2 e5e4 c2c6');
INSERT INTO evaluations VALUES(42775,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(42775,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42775,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(42775,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42775,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(42775,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(42775,36,'cp 10','d7d6 c1d2 c7c5 e2e3 e8g8 f1e2 b8c6 a2a3 b4c3 d2c3 f8e8 a1c1 c8f5 e1g1 a7a5 b2b3 h7h6 f3d2 d8e7 f1e1 f5g6 d1c2 g6h5 e2f1 a8d8 c3a1 h5g6 c2c3 b7b6 h2h3 d6d5 c4d5 f6d5');
INSERT INTO evaluations VALUES(39307,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(39307,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(39307,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(39307,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(39307,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(39307,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(39307,36,'cp 126','d2d4 c5b6 c1f4 g8e7 a2a4 a7a6 b1c3 e8g8 a4a5 b6a7 f1e2 d7d5 c3d5 e7d5 e4d5 d8d5 e1g1 b8c6 c2c3 c8f5 f4c7 a8c8 c7g3 c6a5 f3d2 a7b6 g1h1 f5g6 e2f3 d5d7 d4d5 g6d3 f1e1 c8e8 e1e8 f8e8 d5d6 a5c4 d2c4 d3c4 b2b3 c4b5 c3c4 b5c6 f3c6');
INSERT INTO evaluations VALUES(39299,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(39299,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(39299,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(39299,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(39299,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(39299,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(39299,36,'cp 139','c5b6 c1f4 g8e7 a2a4 a7a6 a4a5 b6a7 f1c4 e8g8 e1g1 d7d6 g1h1 c8e6 b1d2 e7g6 c4e6 f7e6 f4g3 b8d7 c2c3 d8e7 f1e1 b7b5 a5b6 a7b6 e4e5 d6e5 f3e5 g6e5 g3e5 d7e5 e1e5 f8f5 e5e1 e7f7 d2f3 a6a5 d1e2 a5a4 e2e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(39091,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(39091,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(39091,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(39091,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(39091,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(39091,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(39091,36,'cp 150','c1f4 g8e7 a2a4 a7a6 a4a5 b6a7 f1c4 e8g8 e1g1 d7d5 e4d5 e7d5 f4g5 f7f6 b1c3 c7c6 g5h4 c8g4 g1h1 g8h8 f1e1 d5c3 b2c3 b8d7 c4d3 f8e8 d1b1 e8e1 h4e1 d8c8 b1b3 a7b8 f3h4 d7f8 a1b1 b8c7 b3b7 c8b7 b1b7 c7a5 h1g1 g7g6 g1f1 h8g8 h4f3 g4e6 c3c4 a5e1 f1e1 f8d7 b7c7 a6a5');
INSERT INTO evaluations VALUES(39090,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(39090,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(39090,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(39090,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(39090,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(39090,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(39090,36,'cp 103','g8e7 a2a4 a7a6 a4a5 b6a7 f1c4 e8g8 e1g1 b8c6 b1c3 d7d6 g1h1 c8g4 c3e2 d6d5 e4d5 e7d5 c2c3 c6e7 d1b3 g4e6 f3g5 d5f4 e2f4 e6c4 b3c4 d8d7 c4d3 d7f5 d3f5 e7f5 f4d5 f5d6 d5c7 a8d8 c7d5 f8e8 f1e1 h7h6 g5f3 e8e1 a1e1 d6c4 d5e7 g8f8 b2b4 c4d6 g2g4 d8d7 e7d5');
INSERT INTO evaluations VALUES(47207,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(47207,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47207,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47207,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(47207,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(47207,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(47207,36,'cp 0','c7c5 g1f3 c5d4 e1g1 h7h6 b2b3 e7e6 c1b2 f8c5 f3d4 e6e5 d4f3 e5e4 f3d4 b8c6 c2c4 c6d4 b2d4 c5d4 d1d4 d5c4 d4d8 e8d8 b1c3 c4b3 a2b3 c8d7 f1d1 d8c7 a1c1 d7c6 c3b5 c7b6 b5d6 h8d8 d6c4 b6c5 c4d6');
INSERT INTO evaluations VALUES(45388,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45388,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45388,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45388,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45388,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45388,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45388,36,'cp 0','g1f3 c5d4 e1g1 h7h6 b2b3 e7e6 c1b2 f8c5 f3d4 e6e5 d4f3 e5e4 f3d4 b8c6 d4c6 b7c6 c2c4 e8g8 b1c3 e4e3 f2f4 d5c4 d1d8 f8d8 c3a4 c5e7 g2c6 a8b8 b2e5 b8b4 e5c3 b4b8');
INSERT INTO evaluations VALUES(57707,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(57707,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(57707,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(57707,33,'cp 235',NULL);
INSERT INTO evaluations VALUES(57707,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(57707,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(57707,36,'cp 235','d2d4 h7h6 h2h4 g5h4 g1f3 d7d6 b1c3 g8f6 h1h4 c8g4 d1d3 c7c6 e4e5 d6e5 f3e5 g4e6 f2f4 f6d7 d3f3 d7e5 d4e5 d8c7 c1e3 b8d7 f3f2 e8c8 e3a7 f7f6 e1c1 f6e5 f4e5 d7e5 d1d8 c7d8 f2e3 d8c7 h4e4');
INSERT INTO evaluations VALUES(57704,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(57704,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(57704,32,'cp 239',NULL);
INSERT INTO evaluations VALUES(57704,33,'cp 242',NULL);
INSERT INTO evaluations VALUES(57704,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(57704,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(57704,36,'cp 239','e7e6 b1c3 d7d5 e4d5 e6d5 d1e2 c8e6 e2e5 g8f6 c1g5 b8d7 e5f4 f8d6 f4d2 c7c6 f2f3 d8c7 g1e2 b7b5 e2f4 h7h6 f4e6 f7e6 g5h6 e8c8 h6g5 e6e5 c3e2 a7a5 d4e5 d6e5 e1c1 d7b6 e2d4 b6c4 f1c4 d5c4 c2c3 f6d5 g5d8');
INSERT INTO evaluations VALUES(56889,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(56889,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(56889,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(56889,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(56889,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(56889,35,'cp 235',NULL);
INSERT INTO evaluations VALUES(56889,36,'cp 229','b1c3 a7a6 d1h5 g8f6 h5g5 d7d5 e4e5 f6d7 g5d8 e8d8 f2f4 c7c5 c3e2 b8c6 g1f3 c6b4 e1d1 b4c6 g2g4 b7b5 c2c3 b5b4 f4f5 e6f5 g4f5 b4c3 b2c3 c5d4 c3d4 d7b6 e2g3 b6c4 f1c4 d5c4 d4d5');
INSERT INTO evaluations VALUES(56887,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(56887,31,'cp 239',NULL);
INSERT INTO evaluations VALUES(56887,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(56887,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(56887,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(56887,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(56887,36,'cp 218','d7d6 c1e3 a7a6 h2h4 g5h4 d1h5 d8f6 g1f3 b8c6 e1c1 f8h6 f1e2 c8d7 e3h6 f6h6 h5h6 g8h6 d4d5 e6d5 c3d5 a8c8 h1h4 h6g8 d1h1 c6e7 d5f4 h7h6 e2d3 e7c6 g2g3 c6e5 f3e5 d6e5 f4d5 d7c6 f2f4 c8d8 d5c7 e8d7');
INSERT INTO evaluations VALUES(55723,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(55723,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(55723,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(55723,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(55723,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(55723,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(55723,36,'cp 49','d6d5 e4d5 e6d5 b1c3 g8f6 g1f3 f8e7 f1d3 d5c4 d3c4 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 d8d4 f3b7 b8d7 c4b3 d4e5 b7f3 e7d6 g2g3 d7c5 b3c4 c5e6 c3d5 f6d5 f3d5 e5f6 g1g2 a8d8 d5f3 f6f3 g2f3 e6d4 f3g2 d4e6 c4d5 d6c5 f1d1 d8d7 a1b1');
INSERT INTO evaluations VALUES(983,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(983,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(983,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(983,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(983,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(983,35,'cp -155',NULL);
INSERT INTO evaluations VALUES(983,36,'cp -167','e4c3 d5a5 e1g1 c8g4 h2h3 g4h5 d2d3 e8c8 a2a3 c6d4 g2g4 h5g6 f3d4 e5d4 c3e2 f7f6 c1f4 g6e8 e2g3 a5b5 d1e2 b5b2 f1e1 e8d7 e2e4 d8e8 e4d5 b2c3 e1e8 d7e8 a1b1 c3c6 d5d4 f8c5 d4c4 b7b5 c4e4 c6e4 d3e4 c5a3 g3f5 a7a5 f5g7 a3b4 g7e6 e8c6 e6d4 c6e4 d4b5 e4c2');
INSERT INTO evaluations VALUES(987,30,'cp -183',NULL);
INSERT INTO evaluations VALUES(987,31,'cp -178',NULL);
INSERT INTO evaluations VALUES(987,32,'cp -155',NULL);
INSERT INTO evaluations VALUES(987,33,'cp -177',NULL);
INSERT INTO evaluations VALUES(987,34,'cp -169',NULL);
INSERT INTO evaluations VALUES(987,35,'cp -156',NULL);
INSERT INTO evaluations VALUES(987,36,'cp -184','d5d6 d2d3 c8g4 h2h3 g4h5 c3e4 d6d5 e4g3 h5f3 d1f3 d5f3 g2f3 c6d4 e1d1 d4f3 h3h4 g7g6 h4h5 g6g5 d1e2 f3d4 e2d1 f7f6 c1e3 d4e6 g3f5 a7a5 a2a4 a8d8 d1e2 e8d7 e2f3 b7b6 d3d4 e5d4 f5d4 e6d4 e3d4 d7e6');
INSERT INTO evaluations VALUES(1011,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(1011,31,'cp -194',NULL);
INSERT INTO evaluations VALUES(1011,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(1011,33,'cp -158',NULL);
INSERT INTO evaluations VALUES(1011,34,'cp -195',NULL);
INSERT INTO evaluations VALUES(1011,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(1011,36,'cp -172','e1g1 c8g4 h2h3 g4h5 c3e4 d6d5 e4c3 d5a5 d2d3 e8c8 a2a3 c6d4 g2g4 h5g6 f3d4 e5d4 c3e2 f7f6 c1d2 a5b5 f1e1 g6f7 e2f4 f8d6 d1f3 b5d7 f4h5 f7d5 f3f5 d7f5 g4f5 d5f3 h5f4 d6e5 g1f1 g7g6 f4e6 d8d5 a1c1 e5d6 d2f4 d6f4 e6f4 d5f5 f4e2 f3e2 e1e2');
INSERT INTO evaluations VALUES(1010,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(1010,31,'cp -171',NULL);
INSERT INTO evaluations VALUES(1010,32,'cp -186',NULL);
INSERT INTO evaluations VALUES(1010,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(1010,34,'cp -193',NULL);
INSERT INTO evaluations VALUES(1010,35,'cp -183',NULL);
INSERT INTO evaluations VALUES(1010,36,'cp -183','c8g4 h2h3 g4h5 c3e4 d6d5 e4c3 d5a5 c3e4 e8c8 e4g3 h5g6 f3h4 f8e7 h4g6 h7g6 d2d3 a5d5 f1e1 f7f5 c1d2 h8h4 d2c3 e7f6 a2a3 c8b8 d1e2 d5d7 b2b4 c6d4 c3d4 h4d4 a1b1 d4h4 e2f3 h4f4 f3d1 a7a6 d1e2 g6g5 b1b3 b7b6 g3h5 f4h4 h5f6');
INSERT INTO evaluations VALUES(15432,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(15432,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(15432,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(15432,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(15432,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(15432,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(15432,36,'cp 24','d3d4 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 b1c3 e8g8 f2f4 d8d7 g1h1 d7g4 c1e3 g4d1 a1d1 a7a5 e4e5 d6e5 f4e5 f6g4 e3d4 f7f6 e5f6 e7f6 d4f6 g4f6 h1g1 f6g4 f1e1 g4f6 b2b3 a8e8 e1e8 f8e8');
INSERT INTO evaluations VALUES(21052,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(21052,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(21052,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(21052,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(21052,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(21052,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(21052,36,'cp 73','e4d5 d8d5 b1c3 d5d6 d2d4 g8f6 d4d5 c6b4 c3b5 d6d8 c2c4 e7e6 a2a3 b4a6 d1a4 c7c6 d5c6 b7c6 b5c3 a6c5 a4c2 d8c7 b2b4 c5d7 g2g3 a7a5 a1b1 a5b4 a3b4 c8a6 f1e2 f8e7 e1g1 e8g8 f1d1 f8b8 c4c5 a6e2 c2e2 b8e8 d1e1 e8d8 c1f4 c7c8 g1g2 h7h6 b1d1 g8h8');
INSERT INTO evaluations VALUES(20962,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(20962,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(20962,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(20962,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(20962,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(20962,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(20962,36,'cp 63','d8d5 b1c3 d5d6 d2d4 g8f6 c3b5 d6d8 c2c4 e7e6 a2a3 a7a6 b5c3 g7g6 f1e2 f8g7 e1g1 e8g8 h2h3 b7b6 f1e1 c8b7 c1f4 f8e8 a1c1 c6d4 d1d4 d8d4 f3d4 e6e5 e2f3 b7f3 f4e5 f3b7 e5c7 e8c8 c7b6 c8c4 d4b3 f6d5 c3d5 b7d5 c1d1 d5e6');
INSERT INTO evaluations VALUES(1615,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(1615,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(1615,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(1615,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(1615,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(1615,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(1615,36,'cp 77','b1c3 d5d6 d2d4 g8f6 c3b5 d6d8 c2c4 e7e6 b5c3 f8b4 d1c2 e8g8 a2a3 b4c3 c2c3 c6e7 f1e2 b7b6 b2b4 f6e4 c3b2 c8b7 e1g1 a7a6 f1d1 e4d6 a3a4 b6b5 c4b5 a6b5 a4a5 e7f5 c1f4 b7d5 f3e5 f7f6 e5d3 f8e8 d3c5 d6c4');
INSERT INTO evaluations VALUES(1612,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(1612,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(1612,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(1612,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(1612,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(1612,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(1612,36,'cp 69','d5d6 d2d4 g8f6 c3b5 d6d8 c2c4 e7e6 f1e2 f8b4 b5c3 e8g8 e1g1 b7b6 d1b3 a7a5 a2a3 b4c3 b2c3 c8b7 c1f4 c6e7 f3e5 f6e4 f1e1 f7f6 e5d3 e6e5 d4e5 f6e5 f4e5 e7g6 c4c5 g8h8 c5b6 g6e5 d3e5 d8f6 e5g4 f6b6 b3b6 c7b6 e2d3 e4c3 e1e7 b7d5 a1c1 a8c8');
INSERT INTO evaluations VALUES(1652,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(1652,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(1652,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(1652,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(1652,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(1652,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(1652,36,'cp 77','d2d4 g8f6 c3b5 d6d8 c2c4 e7e6 f1e2 f8b4 b5c3 e8g8 e1g1 b7b6 d1b3 a7a5 a2a3 b4c3 b2c3 c8b7 f1e1 c6e7 f3e5 f6d7 c1f4 d7e5 f4e5 e7g6 e5g3 g6h4 e2f1 h4f5 g3f4 d8d7 a1d1 b7c6 b3b2 c6a4 d1d2 f8e8 d4d5 e6d5 e1e8 a8e8 d2d5 d7c8 c4c5 c7c6 c5b6 c6d5');
INSERT INTO evaluations VALUES(1650,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(1650,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(1650,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(1650,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(1650,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(1650,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(1650,36,'cp 76','g8f6 c3b5 d6d8 c2c4 e7e6 f1e2 f8b4 b5c3 b7b6 d1b3 e8g8 e1g1 a7a5 a2a3 b4c3 b2c3 c8b7 f1e1 c6e7 f3e5 f6d7 c1f4 d7e5 f4e5 e7g6 e5g3 g6h4 e2f1 h4f5 g3f4 d8d7 b3b2 f8e8 a1d1 h7h6 f1d3 a8d8 h2h3 f7f6 c4c5 f5h4 f2f3');
INSERT INTO evaluations VALUES(903,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(903,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(903,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(903,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(903,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(903,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(903,36,'cp 91','c3b5 d6d8 c2c4 c8g4 d4d5 a7a6 b5c3 c6e5 f1e2 g4f3 e2f3 d8d7 e1g1 e8c8 f1e1 d7f5 f3e4 f6e4 e1e4 e5d7 d1e2 e7e6 g2g4 f5g6 d5e6 f7e6 e4e6 d7f6 h2h3 h7h5 g4g5 f8d6 c3e4 h8e8 e4d6 d8d6 e6e8 f6e8 c1e3 d6d8 a1d1 d8d1 e2d1 e8d6 b2b3 d6f5 d1f3 f5e3 f3e3');
INSERT INTO evaluations VALUES(22706,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(22706,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(22706,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(22706,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(22706,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(22706,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(22706,36,'cp -13','g8f6 d1b3 d8e7 b3c2 g7g6 b2b4 f8g7 d2d3 e8g8 a2a4 d7d6 g1e2 f6h5 b1d2 g8h8 e1g1 a7a6 c1a3 c8g4 f2f3 g4d7 c4d5 f7f5 b4b5 a6b5 a4b5 c6a7 d5b7 a8b8 b7a6 a7b5 a6b5 b8b5 c3c4 b5a5 a3b2 a5a1 f1a1 d7c6 a1a7 f5e4 d2e4 c6e4 f3e4');
INSERT INTO evaluations VALUES(16393,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(16393,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(16393,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(16393,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(16393,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16393,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(16393,36,'cp -14','d1b3 d8e7 b3c2 g7g6 b2b4 f8g7 a2a4 e8g8 d2d3 d7d6 b1d2 g8h8 g1e2 f6h5 e1g1 a7a6 c1b2 c8d7 c4d5 a8e8 a1e1 c6d8 f2f4 h5f4 e2f4 e5f4 f1f4 c7c6 d5c4 b7b5 c4b3 c6c5 f4f2 d8c6 d2f3 d7g4 b3d5 g4f3 f2f3 c6e5');
INSERT INTO evaluations VALUES(16382,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(16382,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(16382,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(16382,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(16382,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16382,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(16382,36,'cp -14','d8e7 b3c2 g7g6 b2b4 f8g7 a2a4 e8g8 d2d3 f8d8 c4b3 d7d5 g1f3 c8e6 b1d2 d5d4 b4b5 d4c3 b5c6 c3d2 c1d2 e6b3 c2b3 b7c6 b3c4 f6d7 d2e3 c6c5 e1g1 d8b8 f1c1 g7f8 a1b1 b8b1 c1b1 a8b8 b1b8 d7b8 c4b5 b8d7 b5b7 e7d6 b7a7 d6d3');
INSERT INTO evaluations VALUES(16800,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(16800,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(16800,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(16800,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(16800,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(16800,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(16800,36,'cp 78','c1g5 f8e7 g5e7 d8e7 b1c3 c6e5 e2e4 g4f6 f3d4 e8g8 d4f5 e7c5 d1d4 c5d4 f5d4 d7d6 f1e2 a7a6 e1g1 f8e8 f1d1 c8e6 f2f4 e5c6 h2h3 c6d4 d1d4 e6d7 b2b3 d7c6 e2f3 g8f8 g1f2 b7b5 c3d5 b5c4 d4c4 c6d5 e4d5');
INSERT INTO evaluations VALUES(16787,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(16787,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(16787,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(16787,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(16787,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(16787,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(16787,36,'cp 72','f8e7 g5e7 d8e7 b1c3 c6e5 e2e4 g4f6 f3d4 d7d6 f1e2 e8g8 e1g1 e5g6 d1d3 f8e8 f1e1 c8d7 e2f1 e7e5 d3d2 a8d8 a1d1 h7h5 c3d5 f6d5 c4d5 e5f6 d1c1 d8c8 h2h3 a7a6 c1d1 g6f4 b2b3 c7c6 g2g3 f4h3 f1h3 d7h3 d5c6');
INSERT INTO evaluations VALUES(10955,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(10955,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(10955,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(10955,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(10955,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(10955,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(10955,36,'cp 73','g5e7 d8e7 b1c3 c6e5 e2e4 d7d6 f1e2 g4f6 f3d4 e8g8 e1g1 e5g6 f2f3 c7c6 d1d2 c8d7 a1d1 a8d8 f1e1 f8e8 d4c2 d7e6 e2f1 a7a6 a2a4 h7h5 a4a5 h5h4 c2d4 e6c8 b2b4 e7e5 d4f5 c8f5 e4f5');
INSERT INTO evaluations VALUES(10821,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(10821,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(10821,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(10821,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(10821,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(10821,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(10821,36,'cp 77','d8e7 b1c3 c6e5 e2e4 g4f6 f3d4 d7d6 f1e2 e8g8 d1d2 e5c6 f2f3 c6d4 d2d4 c8d7 e1g1 e7e5 f1d1 f8e8 a1c1 a7a5 b2b3 h7h6 g2g3 a8d8 d1d2 d8c8 e2f1 e5d4 d2d4 g8f8 c1d1 g7g5 f1g2 e8e5 f3f4 g5f4 g3f4 e5c5 g1f2 c8e8');
INSERT INTO evaluations VALUES(601,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(601,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(601,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(601,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(601,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(601,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(601,36,'cp 67','b1c3 c6e5 e2e4 d7d6 f1e2 g4f6 f3d4 e8g8 e1g1 e5g6 f2f3 c8d7 d1d2 e7e5 a1d1 a8e8 g2g3 h7h5 d4b3 e5e7 h2h4 a7a6 b3d4 g6e5 d4c2 e8d8 b2b3 f8e8 c2e3 d7e6 f1e1 c7c6');
INSERT INTO evaluations VALUES(18305,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(18305,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(18305,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(18305,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(18305,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(18305,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(18305,36,'cp 194','b1c3 c8b7 c3b5 g8e7 c1g5 a8c8 b5a7 e7c6 d4d2 f7f6 a7c6 d7c6 g5f4 d8d2 e1d2 f8b4 d2c1 e8d7 g2g3 c8a8 h2h4 b4d6 f2f3 d7e7 a2a4 h8d8 f4d6 d8d6 f1e2 h7h5 b2b3 c6c5 e2d3 b7a6 d3a6 a8a6 h1d1');
INSERT INTO evaluations VALUES(18303,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(18303,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(18303,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(18303,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(18303,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(18303,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(18303,36,'cp 194','c8b7 c3b5 g8e7 c1g5 a8c8 b5a7 e7c6 d4d2 f7f6 a7c6 d7c6 g5f4 d8d2 e1d2 f8b4 c2c3 c8d8 d2c2 b4d6 f4d2 c6c5 f2f3 g7g5 a2a4 d8a8 h2h4 g5g4 h4h5 e8d7 a1e1 b7c6 b2b3 a8e8 f1d3 d7d8 f3f4 c5c4 b3c4 c6a4 c2b2 a4c6 e4e5 c6g2');
INSERT INTO evaluations VALUES(25865,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(25865,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(25865,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(25865,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(25865,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(25865,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(25865,36,'cp 207','c1f4 g8f6 f2f3 f8c5 d4e5 e8f8 e1c1 d7d6 e5f5 d8e7 f5h3 e7e6 h3h4 a7a6 c1b1 a8e8 c3a4 f6d7 a4c5 d7c5 h4f2 h7h5 h2h4 f8g8 f4c1 e6f6 c1e3 f6e7 b2b3 g8h7 e3c1 h8f8 f2d2 e7e6 c1b2 h7g8 f1c4 e6g6');
INSERT INTO evaluations VALUES(5290,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5290,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(5290,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(5290,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(5290,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(5290,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(5290,36,'cp 0','e1g1 d7d6 f1e1 a7a6 a2a4 e7e6 d2d3 b7b6 c3e2 g8f6 c2c3 e8g8 c4b3 h7h6 h2h3 e6e5 d3d4 e5d4 c3d4 c5d4 e2d4 c6d4 d1d4 c8b7 d4d3 f8e8 e4e5 d6e5 d3g6 b7d5 b3d5 d8d5 g6g3 f6e4 g3g4 e4f6');
INSERT INTO evaluations VALUES(5289,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5289,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5289,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(5289,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(5289,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(5289,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(5289,36,'cp 0','d7d6 f1e1 a7a6 a2a4 e7e6 d2d3 b7b6 c3e2 g8f6 c2c3 e8g8 h2h3 c8b7 c4b3 c6e5 e2g3 b6b5 f3e5 d6e5 c1e3 d8c7 d1e2 f8c8 e1d1 h7h6 f2f3 f6d7 h3h4 h6h5 g3f1 d7b6 a4a5 b6d7 e3g5 d7f6 g5e3');
INSERT INTO evaluations VALUES(5460,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5460,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(5460,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(5460,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(5460,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(5460,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(5460,36,'cp 21','f1e1 a7a6 a2a4 e7e6 d2d3 b7b6 c3e2 g8f6 c2c3 e8g8 c4b3 c6e5 e2g3 e5f3 d1f3 a8b8 d3d4 f6d7 d4d5 e6d5 b3d5 b6b5 a4b5 a6b5 g3e2 b5b4 h2h3 b4c3 e2c3 d7e5 f3g3 d8b6 e1d1 c8e6 c1g5 b6b2 g5e7 e6d5');
INSERT INTO evaluations VALUES(15484,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15484,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15484,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(15484,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15484,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(15484,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15484,36,'cp 0','c6a5 c4b3 c7c6 h2h3 a5b3 a2b3 f8e7 e1g1 e8g8 c3e2 a7a5 c1d2 c8d7 e2g3 f8e8 f1e1 c6c5 f3h2 b7b5 g3f5 e7f8 h2g4 f6g4 h3g4 g7g6 f5e3 f8h6 f2f3 e8f8 e3d5 h6g7 d5e3');
INSERT INTO evaluations VALUES(15777,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15777,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15777,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15777,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15777,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15777,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15777,36,'cp 0','c4b3 c7c6 e1g1 f8e7 h2h3 a5b3 a2b3 e8g8 c3e2 a7a5 c1d2 c8d7 e2g3 f8e8 f1e1 c6c5 f3h2 b7b5 g3f5 e7f8 h2g4 f6g4 h3g4 g7g6 f5e3 f8h6 f2f3 e8f8 e3d5 h6g7 d5e3');
INSERT INTO evaluations VALUES(51859,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(51859,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(51859,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(51859,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(51859,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(51859,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(51859,36,'cp 71','d2d4 d7d6 f2f4 f8e7 g1f3 e5d4 d1d4 g8f6 c1e3 d6d5 e1c1 d5e4 d4d8 e7d8 f3g5 b8d7 g5e4 e8g8 e3d4 f6e4 c3e4 d7f6 e4f6 d8f6 d4f6 g7f6 f1e2 c8e6 e2f3 a7a5 a2a3 f8d8 d1d8 a8d8 h1d1 d8d1 c1d1 f6f5 d1d2 g8f8 b2b4 a5b4 a3b4 f8e7');
INSERT INTO evaluations VALUES(51854,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(51854,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(51854,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(51854,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(51854,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(51854,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(51854,36,'cp 87','d7d6 f2f4 e5d4 d1d4 g8f6 g1f3 f8e7 c1d2 e8g8 e1c1 b7b5 f1d3 f6d7 h1e1 d7c5 e4e5 b8a6 e5d6 d8d6 d4d6 e7d6 d3e2 b5b4 c3b1 f8e8 f3d4 c8d7 e2f3 e8e1 d1e1 a8c8 d2e3 c5a4 e1d1 a4b6 b1d2 b6d5 d2c4 d5e3 c4d6 e3d1 d6c8 d7c8');
INSERT INTO evaluations VALUES(7451,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(7451,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(7451,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(7451,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(7451,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(7451,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(7451,36,'cp 32','a2a4 c8b7 d2d3 e8g8 b1c3 c6a5 f3e5 a5b3 c2b3 d7d5 c1g5 c5e7 e5g4 d5e4 g5f6 e7f6 g4f6 d8f6 d3e4 b5b4 c3d5 f6b2 a1c1 b2e5 d1f3 a8e8 d5b4 b7e4 f3c3 a6a5 b4c6 e5c3 c1c3 e4c6 c3c6 e8e2 c6c7 e2b2 c7c5 b2b3 f1c1 b3a3 c5a5 g7g6 g2g3 a3a2 c1c5 a2a1 g1g2');
INSERT INTO evaluations VALUES(6600,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(6600,31,'cp 241',NULL);
INSERT INTO evaluations VALUES(6600,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(6600,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(6600,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(6600,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(6600,36,'cp 209','c2c3 b4d6 f1d3 g8f6 e1g1 e8g8 f1e1 f6e8 a2a3 f7f6 b1d2 d6e7 b2b4 a7a6 d2c4 g8h8 c4e3 g7g6 e3d5 e8g7 a3a4 d7d6 a4a5 c8e6 c1d2 e6f7 h2h3 a8c8 d5e7 c6e7 c3c4 g6g5 b4b5 g7e6 b5a6 b7a6 c4c5 d6c5 d4d5');
INSERT INTO evaluations VALUES(6599,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(6599,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(6599,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(6599,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(6599,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(6599,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(6599,36,'cp 230','b4d6 f1d3 a7a5 e1g1 g8e7 d4d5 c6b8 b1d2 e8g8 d2c4 d6c5 c4e5 d7d6 e5c4 b8d7 a2a4 f7f6 c1e3 b7b6 f3d4 d7e5 d3e2 f6f5 c4e5 d6e5 d4e6 c8e6 d5e6 c5d6 f2f3 g8h8 e2b5 f5f4 e3f2 e7g8 d1b3 d8e7 a1d1 g8f6 g1h1 g7g5 b5c4 h7h6 d1d3');
INSERT INTO evaluations VALUES(6775,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(6775,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(6775,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(6775,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(6775,34,'cp 233',NULL);
INSERT INTO evaluations VALUES(6775,35,'cp 242',NULL);
INSERT INTO evaluations VALUES(6775,36,'cp 235','f1d3 a7a5 e1g1 g8e7 d4d5 c6b8 b1d2 e8g8 d2c4 d6c5 f3e5 d7d6 e5f3 a5a4 d3c2 c8g4 h2h3 g4h5 g2g4 h5g6 c1f4 f8e8 h3h4 h7h5 f3g5 h5g4 c2a4 b8d7 b2b4 c5a7 d1g4 d7f6 g4g2 e8f8 a4b3 b7b5 c4a5 a7b6 a5c6 e7c6 d5c6 f6h5 f4h2');
INSERT INTO evaluations VALUES(48087,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(48087,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(48087,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(48087,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(48087,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(48087,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(48087,36,'cp 43','e2e4 d5c3 b2c3 c7c5 g1f3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 d7f6 f1e1 b7b6 e4e5 f6d5 f3g5 c8b7 c4d3 h7h6 g5e4 a8c8 e4d6 c8c7 a1c1 b7a8 h2h3 d8e7 a2a4 f8d8 d3b5 a7a5 g1h2 e7h4 f2f3 h4g5 d2g5 h6g5 c1c7 d5c7');
INSERT INTO evaluations VALUES(48093,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(48093,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(48093,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(48093,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(48093,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(48093,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(48093,36,'cp 32','d5c3 b2c3 c7c5 g1f3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7c5 d2d4 e6d5 e4d5 d8d6 f3e5 c8b7 a1d1 f8d8 f1e1 b6b5 c4b5 d6d5 d4g4 d5e6 g4h4 e6f6 h4f6 g7f6 e5g4 g8g7 g4e3 b7e4 d1d8 a8d8 f2f3 e4g6 e1d1 d8d1 e3d1 h7h6 a2a4');
INSERT INTO evaluations VALUES(48144,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48144,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(48144,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(48144,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(48144,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(48144,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(48144,36,'cp 37','b2c3 c7c5 g1f3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7f6 d2f4 e6d5 a1d1 c8b7 e4d5 a8c8 d5d6 f6h5 f4d4 b7f3 g2f3 d8g5 g1h1 g5f5 d1d3 h5f4 d3e3 c8d8 f1d1 f4h3 d4h4 g7g5 h4g3 f5c2 c4e2 f8e8 e3e8 d8e8 d6d7 e8d8 g3h3 c2e2');
INSERT INTO evaluations VALUES(48143,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(48143,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(48143,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(48143,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(48143,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(48143,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(48143,36,'cp 41','c7c5 g1f3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7f6 d2f4 e6d5 a1d1 c8d7 e4d5 b6b5 c4b3 a7a5 d5d6 a5a4 b3c2 d8b6 f3e5 d7e6 d1d3 f6d7 e5d7 e6d7 d3d4 b6c5 c2b1 f8e8 h2h3 g7g6 f1d1 c5e5 f4d2');
INSERT INTO evaluations VALUES(46063,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(46063,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(46063,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(46063,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46063,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(46063,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(46063,36,'cp 26','g1f3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 f1e1 c8b7 c4b3 a8c8 a1d1 h7h6 d4d5 e6d5 e4d5 d8f6 h2h3 c8e8 f3h2 e8e1 d2e1 d7c5 h2g4 f6d6 b3c2 f8d8 g4e3 g7g6 e1c3 h6h5 a2a4 c5d7 a4a5 d7f6 a5b6 a7b6');
INSERT INTO evaluations VALUES(46062,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(46062,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(46062,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(46062,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(46062,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(46062,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(46062,36,'cp 32','c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 f1e1 c8b7 a1d1 a8c8 c4b3 d8f6 d4d5 e6d5 e4e5 f6g6 f3d4 d7c5 b3c2 g6h5 d4f5 c8c7 f2f4 b7c8 f5d6 c8g4 d1c1 h5h6 h2h3 g4e6 d2d4 g7g6 d6b5 c7d7 c2b3 c5e4 d4e3 g8g7 b5d4 f8d8');
INSERT INTO evaluations VALUES(46081,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(46081,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(46081,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(46081,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(46081,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(46081,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(46081,36,'cp 37','c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 f1e1 c8b7 a1d1 a8c8 c4b3 d8f6 h2h3 h7h6 f3h2 f6g6 d4d5 e6d5 e4d5 c8e8 h2g4 d7c5 g4e3 c5b3 a2b3 e8d8 d2c3 f8e8 d1d4 b7c8 h3h4 g6d6 e1c1 d8d7 c1d1 d6f6 c3d2 c8b7 b3b4 e8d8');
INSERT INTO evaluations VALUES(46080,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(46080,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(46080,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(46080,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(46080,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(46080,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(46080,36,'cp 44','f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7f6 d2f4 e6d5 a1d1 c8b7 e4d5 a8c8 d5d6 f6h5 f4d4 b7f3 g2f3 d8g5 g1h1 g5f5 d1d3 f8d8 f1e1 h5f4 d3e3 f4g6 a2a4 f5c5 d4c5 b6c5 e3e7 d8d6 c4f7 g8f8 e7a7');
INSERT INTO evaluations VALUES(40591,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(40591,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(40591,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(40591,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(40591,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(40591,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(40591,36,'cp 45','c1d2 b4d2 d1d2 e8g8 f1c4 b7b6 d4d5 b8a6 d5e6 f7e6 d2e3 a6c5 e1g1 c8a6 c4a6 c5a6 a1d1 d8c7 f3d4 a6c5 d4b5 c7f4 e3f4 f8f4 f2f3 f4f8 d1d6 f8d8 f1d1 d8d6 d1d6 g8f8 g1f2 f8e7 e4e5 a8d8 d6d8');
INSERT INTO evaluations VALUES(40590,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(40590,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(40590,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(40590,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40590,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(40590,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(40590,36,'cp 28','b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7c5 f1e1 e6d5 e4d5 d8d6 d2d4 c8d7 f3e5 d7a4 e1e3 c5d7 f2f4 a8d8 a2a3 d6c5 d4c5 d7c5 a1e1 d8d6 g2g4 f8e8 e5c6 e8e3 e1e3 g8f8 c6a7 d6f6 e3f3 c5e4 f4f5 h7h5 h2h3 h5g4 h3g4 g7g6 c4d3 e4g5 f3e3');
INSERT INTO evaluations VALUES(40593,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(40593,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(40593,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(40593,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(40593,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(40593,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40593,36,'cp 48','d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7c5 f1e1 e6d5 e4d5 d8d6 f3e5 c8d7 e1e3 d7a4 d2d4 c5d7 c4b3 a4b3 e5d7 d6d7 a2b3 d7d6 h2h4 f8d8 h4h5 h7h6 a1e1 d6c5 d4d3 c5d6 g2g3 a7a5 e3e5 a8c8 d3b5 c8c5 e5e8 d8e8 b5e8 g8h7 e8f7 c5d5');
INSERT INTO evaluations VALUES(40592,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(40592,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(40592,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(40592,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(40592,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(40592,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40592,36,'cp 50','e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7c5 f1e1 e6d5 e4d5 d8d6 f3e5 c8d7 e1e3 d7a4 c4e2 a8d8 e2f3 c5d7 e5c4 d6f4 a1c1 d8c8 g2g3 f4f6 f3e2 a4b5 e3e4 c8c5 c1e1 f6d8 c4e3 b5e2 d2e2 d7f6 e4d4 h7h6 e2f3 f8e8 e1d1 d8d6');
INSERT INTO evaluations VALUES(36148,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(36148,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36148,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36148,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(36148,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36148,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(36148,36,'cp 44','f1c4 b8d7 e1g1 b7b6 d4d5 d7c5 f1e1 e6d5 e4d5 d8d6 f3e5 c8d7 d2d4 d7a4 h2h4 c5d7 c4b3 a4b3 e5d7 d6d7 a2b3 d7d6 e1e3 a8d8 e3g3 g7g6 g3d3 f8e8 h4h5 d6e5 h5g6 h7g6 a1a7 d8d6 a7a4 e5h5 a4a1 e8d8 a1d1 d6d5 d4d5 d8d5 d3d5 h5e2 d5d2 e2b5 d2d3');
INSERT INTO evaluations VALUES(54189,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(54189,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54189,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(54189,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(54189,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54189,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54189,36,'cp 7','g8f6 g1f3 f8e7 b1d2 e8g8 f3e5 c5c4 a2a4 b8c6 f1e2 e7d6 e5c6 b7c6 f4d6 d8d6 b2b3 c4b3 d2b3 f6d7 e1g1 c6c5 c3c4 c8a6 d4c5 d7c5 b3c5 d6c5 c4d5 a6e2 d1e2 c5d5 e2b5 d5b5 a4b5 f8c8 g2g3 g7g6 a1a5 c8c5 f1b1 g8g7 g1g2 a8b8 a5a7 b8b5');
INSERT INTO evaluations VALUES(46195,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(46195,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46195,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46195,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(46195,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(46195,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46195,36,'cp 13','g1f3 f8d6 f1b5 b8c6 e1g1 e8g8 b5c6 b7c6 f4d6 d8d6 d1a4 c5d4 c3d4 c6c5 a4a3 f6e4 f1c1 c5c4 b2b3 d6a3 b1a3 c4b3 a2b3 a8b8 a3b1 b8b3 a1a7 b3b7 a7b7 c8b7 b1d2 f7f6 d2e4 d5e4 f3d2 f8a8 h2h4 b7d5 g2g4 a8a2 d2c4 a2a6 g4g5 g8f7 c4d2 a6a5 g1g2 f7g6 g5f6 g7f6');
INSERT INTO evaluations VALUES(15705,30,'cp 223',NULL);
INSERT INTO evaluations VALUES(15705,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(15705,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(15705,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(15705,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(15705,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(15705,36,'cp 213','d2d3 f8c5 f1e2 f6g4 e2g4 d8h4 d1f3 h4g4 f3g4 c8g4 f2f3 g4e6 b1c3 e8c8 h2h4 b7b6 c1d2 c5d4 e1c1 c6c5 c3e2 a7a5 b2b3 c8b7 a2a4 f7f5 d1e1 d8e8 c2c4 e6d7 c1c2 h8g8 h4h5 d4f6 e1f1 f5e4 d3e4 h7h6 g2g3');
INSERT INTO evaluations VALUES(15702,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(15702,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(15702,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(15702,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(15702,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(15702,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(15702,36,'cp 207','f8c5 f1e2 f6g4 e2g4 d8h4 d1f3 h4g4 f3g4 c8g4 f2f3 g4e6 b1c3 e8c8 c1d2 c5d4 e1c1 c6c5 b2b3 b7b6 c3e2 a7a5 d2c3 d4e3 c1b1 f7f6 a2a4 c8b7 c3d2 e3f2 h1f1 f2d4 d2c3 h7h5 c3d4 c5d4 f1f2 c7c5 d1g1 f6f5 e4f5');
INSERT INTO evaluations VALUES(9495,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(9495,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(9495,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(9495,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(9495,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(9495,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(9495,36,'cp 227','f1e2 f6g4 e2g4 d8h4 d1f3 h4g4 f3g4 c8g4 f2f3 g4e6 b1c3 e8c8 h2h4 b7b6 b2b3 h8e8 h4h5 f7f5 c1d2 a7a5 e1c1 c5a3 c1b1 a3b4 d1e1 c6c5 a2a4 c8b7 b1c1 f5e4 e1e4 b4a3 c1d1 a3b4 e4e2 h7h6 e2e4 e6d7 e4e8 d8e8 c3e4');
INSERT INTO evaluations VALUES(9494,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(9494,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(9494,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(9494,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(9494,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(9494,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(9494,36,'cp 212','f6g4 e2g4 d8h4 d1f3 h4g4 f3g4 c8g4 f2f3 g4e6 b1c3 e8c8 h2h4 b7b6 c1d2 c5d4 e1c1 c6c5 h4h5 a7a5 g2g4 c5c4 c3e2 d4f6 d3d4 f6d4 h5h6 d4e5 h6g7 e5g7 d2c3 g7c3 e2c3 c7c6 d1d8 c8d8 c1d2 b6b5 c3e2 d8c7 d2e3 b5b4');
INSERT INTO evaluations VALUES(22742,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(22742,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(22742,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(22742,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(22742,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(22742,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(22742,36,'cp 32','f1g2 g8f6 b1c3 f8c5 a2a3 d7d6 e2e3 c5b6 d2d3 c6e7 g1f3 c7c6 b2b4 e8g8 e1g1 b6c7 a3a4 h7h6 c1b2 e7g6 a4a5 d6d5 b4b5 a8b8 b2a3 f8e8 a5a6 c6b5 c3b5 b7a6 b5c7 d8c7 d1c2 d5c4 d3c4 a6a5 f3d2 c8e6');
INSERT INTO evaluations VALUES(22743,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(22743,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(22743,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(22743,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(22743,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(22743,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(22743,36,'cp 26','g8f6 b1c3 f8c5 g1f3 e8g8 e1g1 d7d6 e2e3 c8f5 d2d3 d8d7 d3d4 c5b6 d4e5 d6e5 d1d7 f5d7 b2b3 f8e8 h2h3 a8d8 f1d1 e5e4 f3e1 d7f5 c1b2 f5g6 g1f1 h7h6 c3e2 f6d7 e2f4 g6h7 a1b1 d7c5 f4d5');
INSERT INTO evaluations VALUES(16464,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(16464,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(16464,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(16464,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(16464,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(16464,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(16464,36,'cp 24','b1c3 f8c5 a2a3 d7d6 b2b4 c5b6 d2d3 e8g8 e2e3 h7h6 g1f3 c8e6 c1b2 a7a5 b4b5 c6e7 a3a4 c7c6 e1g1 f8e8 b2a3 e7g6 d1c2 d6d5 b5c6 b7c6 c4d5 c6d5 c3b5 d8d7 b5d6 e8b8 a1c1 e6h3 c1b1 d5d4 d6c4 b6c7 b1b8 a8b8 e3d4 h3g2 g1g2');
INSERT INTO evaluations VALUES(21186,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(21186,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(21186,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(21186,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(21186,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(21186,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(21186,36,'cp 91','d4d5 c6b8 c2c4 e7e5 f1e2 g7g6 b1c3 b8d7 g2g4 f8h6 c1h6 g8h6 d1d2 h6g8 h2h4 h7h6 e1c1 g8f6 f2f3 d7c5 g1h3 a7a5 h3f2 d8e7 d1g1 h6h5 g4g5 f6d7 b2b3 d7f8 c1b1 c8d7 b1a1 e8c8 f2d3 b7b6 d3c5');
INSERT INTO evaluations VALUES(21166,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(21166,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(21166,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(21166,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(21166,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(21166,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(21166,36,'cp 92','c6b8 c2c4 e7e5 b1c3 b8d7 c1e3 g7g6 d1d2 h7h5 h2h3 d8e7 c3b5 e8d8 b5a7 f8h6 a7c8 d8c8 f1d3 f7f5 e3h6 g8h6 g1e2 e7h4 e2c3 d7c5 d3c2 c5e4 c3e4 f5e4 e1g1 h6f5 d2e2 f5d4 e2e4 g6g5 e4h4 g5h4 c2d3 c8d7');
INSERT INTO evaluations VALUES(48499,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(48499,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(48499,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(48499,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(48499,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(48499,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(48499,36,'cp 63','f1c4 c7c5 d4d5 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e7e2 c4e2 e8g8 f3d2 d6e5 d2c4 e5d4 a2a4 b8d7 f1d1 d7e5 c4e5 d4e5 c1e3 f8e8 d5d6 e8e6 e2f3 e6d6 e3c5 d6d1 a1d1 c8g4 f3b7 g4d1 b7a8 e5c3 b2c3 d1a4 a8b7 f6d7 c5b4 a4b5 f2f4 d7b6 b4c5 b6a4 c5d4 b5c4');
INSERT INTO evaluations VALUES(48472,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(48472,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(48472,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(48472,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(48472,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(48472,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(48472,36,'cp 56','c7c5 g1f3 c5d4 d1d4 d8d4 f3d4 a7a6 e1g1 f8c5 f1d1 e8g8 c1f4 f8e8 e4e5 f6d7 d4f3 b8c6 c3e4 c5e7 h2h3 d7b6 c4b3 e8d8 d1d8 c6d8 a1d1 b6d5 b3d5 e6d5 e4c3 c8e6 c3d5 d8c6 d5c7 a8d8 d1d8 c6d8 c7e6 f7e6 g1h1 d8c6');
INSERT INTO evaluations VALUES(46119,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(46119,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(46119,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(46119,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(46119,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(46119,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(46119,36,'cp 54','g1e2 c5d4 e2d4 f8c5 c1e3 e8g8 e1g1 d8e7 a1c1 b8d7 c4e2 f8d8 c3a4 c5d6 d1c2 d7f8 d4b5 d6b8 f1d1 d8d1 c1d1 c8d7 g2g3 d7c6 a4c3 a7a6 b5d4 c6d7 c2b3 b7b5 f2f4 e6e5 c3d5 e7e8 d4f5 d7f5 e4f5 e5f4 g3f4 f8d7 e2f3');
INSERT INTO evaluations VALUES(55120,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55120,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55120,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55120,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55120,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55120,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(55120,36,'cp 0','g1f3 g8f6 d2d4 e7e6 c1f4 f8d6 e2e3 e8g8 f4g3 b7b6 f1d3 c8b7 b1d2 c7c5 d1e2 d8c7 e3e4 f6e4 d2e4 d5e4 g3d6 c7d6 d3e4 b7e4 e2e4 b8d7 e1g1 d6c7 a1d1 a8c8 d4c5 d7c5 e4e2 f8d8 d1d4 h7h6 d4d8 c7d8 f1d1 d8c7 d1d4');
INSERT INTO evaluations VALUES(55118,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(55118,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(55118,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(55118,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(55118,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55118,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(55118,36,'cp 2','g8f6 d2d4 e7e6 c1f4 f8d6 e2e3 e8g8 f1d3 c7c5 e1g1 d6f4 e3f4 d8c7 g2g3 c5d4 c3d4 c7b6 d1d2 b8c6 b1c3 c8d7 f1c1 f8c8 h2h3 c6a5 a1b1 c8c7 d2e2 a8c8 g3g4 a5c4 a2a3 d7c6 c1e1 c4d6 e2e3');
INSERT INTO evaluations VALUES(47265,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(47265,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47265,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(47265,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(47265,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(47265,36,'cp 11','d2d4 e7e6 c1f4 f6h5 f4g5 f8e7 g5e7 d8e7 c3c4 e8g8 e2e3 f8d8 c4d5 e6d5 f1e2 c8f5 e1g1 h5f6 b1c3 c7c6 f3h4 f5e6 d1c2 c6c5 d4c5 e7c5 h4f3 b8c6 a1d1 d5d4 f3d4 c6d4 d1d4 d8d4 e3d4 c5d4 f1d1 d4b4 e2f3 a8c8');
INSERT INTO evaluations VALUES(15274,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(15274,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(15274,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(15274,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(15274,34,'cp -158',NULL);
INSERT INTO evaluations VALUES(15274,35,'cp -160',NULL);
INSERT INTO evaluations VALUES(15274,36,'cp -148','d8d5 e4c3 d5d6 e1g1 c8e6 d2d3 e8c8 f3g5 h7h6 g5e6 d6e6 f2f4 e5f4 c1f4 g7g5 f4g3 f7f5 d1e2 e6f7 a1e1 c6d4 e2e5 f8g7 e5e7 d8d7 e7f7 d7f7 c3d5 d4c2 e1c1 g7d4 g1h1 c2e3 d5e3 d4e3 f1f5 f7d7 c1f1 e3d4 f1e1 d4b2');
INSERT INTO evaluations VALUES(42461,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(42461,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(42461,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(42461,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(42461,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(42461,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(42461,36,'cp 164','e4e5 f6d5 d1g4 e8f8 c1g5 h7h5 g4h4 b4e7 g1f3 d5c3 b2c3 d7d6 h4f4 b7b6 f1c4 d6d5 c4b5 c8a6 a2a4 c7c6 g5e7 d8e7 b5a6 b8a6 h2h4 f8e8 h1h3 c6c5 h3g3 g7g6 f3g5 c5d4 c3d4 a6b4 e1f1 b4c2 a1c1 a8c8 g3f3 h8f8 g5h7');
INSERT INTO evaluations VALUES(42436,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(42436,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(42436,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(42436,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(42436,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(42436,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(42436,36,'cp 122','f6d5 d1g4 e8f8 a2a3 h7h5 g4f3 b4c3 b2c3 d7d6 c3c4 d5e7 c1b2 c8d7 a1d1 d6e5 d4e5 c7c5 g1h3 b8c6 f1d3 d8a5 d1d2 c6d4 f3f4 d4f5 e1g1 d7c6 f1d1 a8d8 d3f5 d8d2 f4d2 a5d2 d1d2 e6f5 d2d8 c6e8 h3g5 e7c6 d8c8 h8h6 h2h4 h6g6 b2c3 b7b6 c8c7 c6d8 c7a7 d8e6 f2f4 e6g5 f4g5 g6e6 a7a6 f8e7');
INSERT INTO evaluations VALUES(42409,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(42409,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(42409,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(42409,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(42409,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(42409,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(42409,36,'cp 132','d1g4 e8f8 c1g5 d8e8 g1e2 b8c6 e1c1 d5c3 e2c3 b4c3 b2c3 h7h6 g5e3 c6e7 f1d3 e7d5 e3d2 b7b5 h2h4 e8e7 h1h3 h6h5 g4g5 a7a6 g2g4 g7g6 d3e2 e7g5 h4g5 c8b7 g4h5 g6h5 d1h1 d5b6 h3h5 f8g7 h5h8 a8h8 h1h8 g7h8 c1d1');
INSERT INTO evaluations VALUES(56091,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(56091,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(56091,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(56091,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(56091,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(56091,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(56091,36,'cp 61','d2d4 g8f6 f1c4 f8e7 e1g1 b8c6 a2a3 e8g8 f1e1 b7b6 d4d5 c6a5 c4a2 f6d5 c3d5 e6d5 d1d5 d8d5 a2d5 c7c6 e1e7 c6d5 c1d2 c8e6 d2a5 b6a5 f3d4 f8e8 e7e8 a8e8 a3a4 g8f8 f2f3 f8e7 g1f2 e7d6 b2b3 g7g5 a1h1 e6d7 h2h4 g5g4 h1c1');
INSERT INTO evaluations VALUES(9948,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9948,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9948,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9948,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9948,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9948,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9948,36,'cp 0','d7c6 f3e5 c5f2 e1f2 d8d4 f2f1 d4e5 d2d3 e5e7 d1f3 c8e6 c1f4 e8c8 a2a3 f6d7 f3g3 d7f6');
INSERT INTO evaluations VALUES(9476,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9476,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9476,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9476,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9476,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9476,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9476,36,'cp 0','f3e5 c5f2 e1f2 d8d4 f2f1 d4e5 d2d3 e5e7 d1f3 c8e6 c1f4 e8c8 a2a3 f6d7 f3g3 d7f6');
INSERT INTO evaluations VALUES(9468,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9468,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9468,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9468,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9468,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9468,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9468,36,'cp 0','c5f2 e1f2 d8d4 f2f1 d4e5 d2d3 e5e7 d1f3 c8e6 c1f4 e8c8 a2a3 f6d7 f3g3 d7f6');
INSERT INTO evaluations VALUES(9499,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9499,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9499,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9499,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9499,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9499,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9499,36,'cp 0','e1f2 d8d4 f2f1 d4e5 d2d3 e5e7 d1f3 c8e6 c1f4 e8c8 a2a3 f6d7 f3g3 d7f6');
INSERT INTO evaluations VALUES(9498,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9498,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9498,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9498,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9498,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9498,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9498,36,'cp 0','d8d4 f2f1 d4e5 d2d3 e5e7 d1f3 c8e6 c1f4 e8c8 a2a3 f6d7 f3g3 d7f6');
INSERT INTO evaluations VALUES(523,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(523,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(523,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(523,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(523,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(523,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(523,36,'cp 0','f2f1 d4e5 d2d3 e5e7 d1f3 c8e6 c1f4 e8c8 a2a3 f6d7 f3g3 d7f6');
INSERT INTO evaluations VALUES(57562,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(57562,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(57562,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(57562,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(57562,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(57562,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(57562,36,'cp 187','d2d4 e7e6 g1e2 c7c6 e2g3 d7d5 f1d3 g8e7 b1d2 g7g6 h2h4 f8g7 d2f3 c6c5 d4c5 b8d7 c1e3 d8a5 e3d2 a5c7 b2b4 e8g8 e1g1 a7a5 c2c3 a5b4 c3b4 f6f5 e4d5 e7d5 f1e1 d7e5 f3e5 g7e5 d3c4 e5a1 d1a1');
INSERT INTO evaluations VALUES(57559,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(57559,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(57559,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(57559,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(57559,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(57559,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(57559,36,'cp 167','e7e6 g1e2 b8c6 c1e3 f6f5 e4f5 e6f5 d4d5 c6e5 e2d4 g8f6 d4f5 d7d6 f5g3 f8e7 b1c3 c7c6 f2f4 e5g4 e3d2 e8g8 h2h3 g4e5 d1e2 e5g6 e1c1 g8h8 e2f3 c8d7 f1c4 b7b5 d5c6 b5c4 c6d7 d8d7 g3e4 d6d5 e4f6 e7f6 h1e1 f6c3 f3c3 g6f4 d2f4 f8f4');
INSERT INTO evaluations VALUES(57378,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(57378,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(57378,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(57378,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(57378,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(57378,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(57378,36,'cp 187','g1e2 f8d6 e2f4 d8e7 f1e2 b8c6 b1c3 a7a6 e1g1 b7b6 c1e3 c8b7 f4d3 g8h6 c3a4 e8g8 c2c4 b6b5 d3c5 b5a4 c5b7 a8b8 b7d6 c7d6 a1b1 f6f5 e3h6 g7h6 d4d5 c6e5 e4f5 e6f5 f2f4 e5g6 e2d3 e7f6 g1h1 f8e8 b2b4 a4b3 a2b3 e8e3 d1d2');
INSERT INTO evaluations VALUES(22575,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(22575,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(22575,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(22575,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(22575,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(22575,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(22575,36,'cp -12','d2d4 d7d6 e5d6 f8d6 g1f3 g8f6 f1b5 e8g8 e1g1 a7a6 b5c6 b7c6 c1e3 d8e7 c2c4 c6c5 d4c5 d6c5 e3c5 e7c5 d1d4 c5d4 f3d4 a8b8 b2b3 c8b7 b1c3 f8d8 d4e2 g8f8 a1d1 d8d1 f1d1 f8e7 c3a4 f6d7 e2f4 b7c6 a4c3 b8d8 f4d3 a6a5');
INSERT INTO evaluations VALUES(22572,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(22572,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(22572,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(22572,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(22572,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(22572,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(22572,36,'cp -24','d7d6 e5d6 f8d6 g1f3 g8f6 f1b5 a7a6 b5c6 b7c6 c1e3 d8e7 e1g1 c6c5 d4c5 d6c5 e3c5 e7c5 f1e1 e8g8 f3e5 c8b7 b1c3 a8d8 d1e2 f6d5 c3d5 d8d5 a1d1 f8d8 c2c4 d5d1 e1d1 d8d6 e5g4 a6a5 g4e3 h7h5 d1d6 c7d6 h2h4 c5e5 e2d3');
INSERT INTO evaluations VALUES(22061,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(22061,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(22061,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(22061,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(22061,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(22061,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22061,36,'cp 0','e5d6 f8d6 g1f3 g8f6 f1b5 a7a6 b5c6 b7c6 e1g1 c6c5 d4c5 d6c5 d1d8 e8d8 f3e5 d8e8 b1d2 a6a5 a2a4 c8b7 d2b3 c5d6 e5c4 d6b4 c4e5');
INSERT INTO evaluations VALUES(21844,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(21844,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(21844,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(21844,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(21844,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21844,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(21844,36,'cp -24','f8d6 g1f3 g8f6 f1e2 e8g8 e1g1 e6e5 d4e5 c6e5 f3e5 d6e5 d1d8 f8d8 b1d2 f6d5 e2f3 c7c6 f1e1 f7f6 d2c4 e5c7 c4e3 c8e6 g2g3 g8f7 a2a3 g7g6 e3d5 e6d5 f3d5 d8d5 c1e3 a8d8 a3a4 c7b6 e3b6 a7b6');
INSERT INTO evaluations VALUES(5927,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(5927,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(5927,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(5927,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(5927,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(5927,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(5927,36,'cp -12','g1f3 g8f6 f1e2 e8g8 b1d2 e6e5 d4e5 c6e5 f3e5 d6e5 d2c4 d8d1 e2d1 f8e8 e1g1 e5d4 c4e3 c7c6 d1f3 c8e6 f1d1 a8d8 a2a4 h7h5 h2h3 g7g5 g2g4 h5g4 h3g4 a7a6 g1g2 g8g7 a4a5 g7g6 a1a4 d4a7 a4b4 d8d1 f3d1 e8e7 d1f3 a7c5 b4a4');
INSERT INTO evaluations VALUES(54707,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(54707,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(54707,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(54707,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(54707,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(54707,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(54707,36,'cp 88','e2e3 b8c6 f1d3 c8e6 g1f3 d8d7 b1c3 e6f7 c3b5 a8c8 e1g1 a7a6 b5c3 c6b4 d3e2 e7e6 a2a3 b4c6 e3e4 f8d6 f4g3 g8e7 b2b4 d6g3 h2g3 e8g8 f1e1 f8e8 d1d2 f7g6 b4b5 a6b5 e2b5 g6e4 c3e4');
INSERT INTO evaluations VALUES(54704,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(54704,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(54704,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(54704,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(54704,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(54704,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(54704,36,'cp 92','b8c6 f1d3 c8e6 g1f3 e6g4 h2h3 g4h5 c2c4 d5c4 d3c4 e7e5 f4h2 f8b4 b1c3 e5d4 d1b3 d4c3 b2c3 b4d6 h2d6 c7d6 c4g8 d8b6 g8d5 b6b3 a2b3 h5f7 a1d1 f7d5 d1d5 e8c8 e1e2 h8e8 g2g4 c8c7 h1d1 h7h6 f3d4 c6d4 d5d4 g7g6 c3c4 a7a5 f2f4 e8e6');
INSERT INTO evaluations VALUES(20842,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(20842,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(20842,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(20842,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(20842,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(20842,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(20842,36,'cp 112','f1d3 c8e6 g1f3 e6g4 h2h3 g4h5 f4h2 e7e6 c2c4 f8d6 h2d6 d8d6 e1g1 g8e7 b1d2 e8g8 a1c1 h5g6 d3g6 e7g6 d1c2 a7a5 a2a3 f8c8 f1e1 g6e7 d2f1 a5a4 c4d5 e6d5 h3h4 c6d8 h4h5 d8e6 f1g3 c8e8 e3e4 d5e4 e1e4');
INSERT INTO evaluations VALUES(20839,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(20839,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(20839,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(20839,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(20839,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(20839,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(20839,36,'cp 99','c8e6 g1f3 e6g4 h2h3 g4h5 f4h2 e7e6 c2c4 f8d6 h2d6 d8d6 e1g1 g8e7 b1d2 e8g8 a1c1 g8h8 f1e1 e6e5 d1b3 e5d4 f3d4 h5f7 d4f3 a7a5 a2a3 a5a4 b3c2 f8d8 c4d5 d6d5 d3c4 d5h5 d2e4 f7c4 c2c4 e7d5 e4c5');
INSERT INTO evaluations VALUES(56038,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(56038,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(56038,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(56038,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(56038,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(56038,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(56038,36,'cp 74','g1f3 g8f6 f1c4 a7a6 a2a4 f8b4 e1g1 e8g8 d1e2 b7b6 c1g5 b8d7 f3e5 c8b7 f1e1 d8c8 e5d7 c8d7 g5f6 g7f6 d4d5 b4c3 b2c3 b7d5 e2g4 g8h8 g4h4 f8g8 h4f6 g8g7 c4d5 d7d5 g2g3 d5f5 f6f5 e6f5 e1e7 a8c8 a1e1 f5f4 g1g2 f4g3 f2g3 c8d8 e7c7 d8d2 g2f3 d2h2');
INSERT INTO evaluations VALUES(53019,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(53019,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(53019,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(53019,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(53019,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(53019,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(53019,36,'cp -9','e7e5 b1c3 b8c6 d2d3 g7g6 a1b1 f8g7 f1e2 g8e7 e1g1 a7a5 a2a3 e8g8 c1d2 h7h6 c3b5 b7b6 b2b4 a5b4 a3b4 a8b8 h2h3 c8e6 f3e1 g8h7 e1c2 f7f5 b5c3 c6d4 b4c5 b6c5 b1b8 d8b8 c2d4 c5d4 c3b5 f5e4 d3e4');
INSERT INTO evaluations VALUES(53884,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(53884,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(53884,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(53884,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(53884,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(53884,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53884,36,'cp -1','b1c3 b8c6 d2d3 g7g6 a1b1 f8g7 f1e2 g8e7 e1g1 b7b6 a2a3 a7a5 c1d2 e8g8 c3b5 h7h6 b2b4 a5b4 a3b4 c8e6 h2h3 f7f5 b4c5 b6c5 b5c3 g8h7 c3d5 e6d7 d5b6 a8b8 b6d7 d8d7 b1b8 f8b8 e4f5 g6f5 f3h4 g7f6');
INSERT INTO evaluations VALUES(53879,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53879,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53879,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53879,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53879,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53879,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53879,36,'cp -7','g7g6 d2d3 b8c6 a1b1 g8e7 c1d2 f8g7 a2a3 a7a5 f1e2 e8g8 c3b5 b7b6 e1g1 h7h6 h2h3 f7f5 b2b4 a5b4 a3b4 c8e6 d1b3 a8b8 b1a1 g8h8 d2c3 f5e4 d3e4 c6d4 f3d4 e5d4 c3d2 e7c6 f2f4 d8d7 b3b2 c6b4 d2b4 c5b4 b2b4');
INSERT INTO evaluations VALUES(48207,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(48207,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(48207,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(48207,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(48207,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(48207,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(48207,36,'cp 18','e2e4 d5e4 c3e4 b8d7 g1f3 h7h6 e4f6 d7f6 g5e3 f8d6 f1d3 e8g8 d1e2 b7b6 e1c1 c8b7 f3e5 f6d5 e3d2 d8h4 e5f3 h4f6 e2e4 f6f5 e4f5 e6f5 f3e5 d5e7 h1g1 b7e4 d3e4 f5e4 g1e1 f7f5 f2f3 e4f3 g2f3 f8f6 c2c4 a8d8 f3f4 e7g6');
INSERT INTO evaluations VALUES(48488,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(48488,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(48488,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(48488,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(48488,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(48488,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(48488,36,'cp 33','e2e4 f8b4 f1c4 f6e4 e1g1 e4f6 d1a4 b8c6 f3e5 a8b8 d4d5 f6d5 c3d5 e6d5 e5c6 b7c6 a4c6 d8d7 c6d5 e8g8 c1e3 b4d6 b2b3 c8b7 d5h5 d7c6 h5g4 b7c8 g4h4 c8b7 f2f3 b7a6 a1c1 a6c4 h4c4 c6b5 e3f2 b5c4 c1c4 f8e8 f1e1 e8e1 f2e1 f7f6 g1f1 b8b5 g2g3 g8f7');
INSERT INTO evaluations VALUES(48498,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(48498,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48498,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(48498,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(48498,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(48498,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(48498,36,'cp 34','f8b4 f1c4 f6e4 e1g1 e4f6 d1a4 b8c6 f3e5 a8b8 d4d5 e6d5 e5c6 b7c6 c3d5 f6d5 a4c6 d8d7 c6d5 e8g8 c1e3 d7d5 c4d5 b4a5 b2b3 a5b6 f1e1 c8f5 h2h4 b8d8 d5f3 f8e8 e3f4 e8e1 a1e1 f5e6 g1f1 h7h6 e1c1 e6f5 h4h5 c7c5 f1e2 f5e6 g2g3');
INSERT INTO evaluations VALUES(42172,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(42172,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(42172,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(42172,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(42172,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42172,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(42172,36,'cp 27','f1c4 f6e4 e1g1 e4f6 d1a4 b8c6 f3e5 a8b8 e5c6 b7c6 a4c6 c8d7 c6f3 e8g8 f1d1 h7h6 b2b3 d8e7 c1f4 d7c8 f4e5 c8b7 f3g3 f6h5 g3e3 h5f6 h2h3 b8d8 d1d3 b4d6 a1c1 f8e8 c1d1 a7a6 e3e2 f6d7 c3a4 d7e5 d4e5 d6a3');
INSERT INTO evaluations VALUES(42158,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(42158,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(42158,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(42158,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(42158,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(42158,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(42158,36,'cp 15','f6e4 e1g1 e4c3 b2c3 b4d6 f3g5 h7h6 d1h5 e8g8 g5e4 f7f5 e4d6 c7d6 c1f4 d6d5 c4d3 d8e8 h5h3 b8c6 f1e1 c8d7 d3e2 b7b6 h3g3 f8f6 h2h4 e8g6 g3h2 f6f7 h4h5 g6f6 h2g3 c6a5 e2d3 a5c4 a2a4 a8c8 d3c4 c8c4');
INSERT INTO evaluations VALUES(41959,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(41959,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41959,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(41959,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(41959,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(41959,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41959,36,'cp 25','e1g1 e4f6 d1a4 b8c6 f3e5 a8b8 d4d5 e6d5 c3d5 f6d5 e5c6 b7c6 a4c6 d8d7 c6d5 e8g8 c1e3 b4d6 b2b3 c8b7 d5h5 d7c6 h5g4 b7c8 g4g5 h7h6 g5d5 c6d5 c4d5 a7a5 d5c6 c8g4 f1e1 g4e6 e1d1 b8d8 h2h3 d6e5 a1c1 d8d1 c1d1 e5d6 g1f1 g7g5 e3a7 f8d8 f1e2');
INSERT INTO evaluations VALUES(4807,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(4807,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4807,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(4807,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4807,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(4807,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(4807,36,'cp 19','d2d4 e5d4 d1d4 f7f6 c1e3 g8e7 b1d2 c8e6 d4c3 d8d7 a1d1 e8c8 f3d4 e7g6 f2f3 d7e7 d4e6 e7e6 d2c4 d6f4 c3b3 f4e3 c4e3 e6e5 g2g3 h7h5 d1d8 h8d8 f1d1 d8d1 e3d1 e5e8 d1f2 g6e5 h2h3 b7b6 f3f4 e5f7 b3d3 c8b7 g1g2 f7d6 b2b3 a6a5 g2f3 e8f7 d3c3');
INSERT INTO evaluations VALUES(4795,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(4795,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(4795,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4795,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(4795,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4795,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4795,36,'cp 18','e5d4 d1d4 f7f6 c1e3 g8e7 b1d2 c8e6 d4c3 d8d7 a1d1 e8c8 f3d4 h8e8 f2f3 g7g5 d4e6 d7e6 d2c4 e7g6 c4d6 d8d6 d1d6 c7d6 f1d1 g5g4 e3f2 g4f3 c3f3 g6e5 f3h5 e8e7 h2h3 e6a2 d1d6 a2b2 d6f6 e5d7 f6f7 e7f7 h5f7 b2c2');
INSERT INTO evaluations VALUES(4813,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(4813,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(4813,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(4813,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4813,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(4813,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(4813,36,'cp 0','d1d4 f7f6 c1e3 g8e7 b1d2 c8e6 d4c3 d8d7 a1d1 e8c8 f3d4 h8e8 f2f3 f6f5 d4f5 e6f5 d2c4 d7e6 e4f5 e6f5 c4d6 d8d6 d1d6 c7d6 e3f2 e7d5 c3d2 e8e6 a2a4 g7g5 a4a5 d5f4 f2e3 f5e5 f1e1 e5b2 e3f4 e6e1 d2e1 g5f4 e1e8 c8c7 e8f7 c7c8');
INSERT INTO evaluations VALUES(4811,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4811,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4811,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4811,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4811,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4811,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4811,36,'cp 0','f7f6 c1e3 c8e6 b1d2 g8e7 d4c3 d8d7 b2b4 e8g8 a2a3 a8d8 a1d1 f6f5 f3g5 f5e4 d2e4 e6a2 f1e1 h7h6 e4d6 c7d6 g5e4 a2d5 e3d4 e7f5 f2f3 d7f7 d4f2 d8e8 c3d3 b7b5 e4g3 f5h4 g3e4');
INSERT INTO evaluations VALUES(4818,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4818,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4818,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4818,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(4818,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4818,36,'cp 12','c1e3 c8e6 b1d2 g8e7 d4c3 d8d7 b2b4 e8g8 a2a3 a8d8 a1d1 f6f5 f3g5 f5e4 d2e4 e6d5 c3d3 d7f5 c2c4 d5e4 d3e4 f5e4 g5e4 e7f5 d1e1 d6e5 e4c5 f5e3 e1e3 e5d4 c5b7 d4e3 b7d8 f8d8 f2e3 d8d3 g1f2 d3a3 f1d1 a3c3 d1d4 g8f7 h2h3 f7f6 d4d7 h7h5 d7c7 c3c4');
INSERT INTO evaluations VALUES(57762,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(57762,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(57762,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(57762,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(57762,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(57762,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(57762,36,'cp -87','e7e5 b1c3 b8c6 e2e3 g8f6 f1b5 d7d5 d2d3 c8d7 g1e2 a7a6 b5a4 f8d6 e1g1 c6e7 e3e4 b7b5 a4b3 d5d4 c3b1 e8g8 a2a4 h7h6 f3f4 c7c5 f4e5 d6e5 b1d2 e7g6 a4b5 a6b5 a1a8 d8a8 d2f3 e5c7 e2g3 f6g4 g3f5 d7f5 e4f5');
INSERT INTO evaluations VALUES(57230,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(57230,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(57230,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(57230,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(57230,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(57230,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(57230,36,'cp -72','b1c3 b8c6 e2e4 f8c5 f1c4 a7a6 d2d3 d7d6 c3d5 g8e7 c1g5 f7f6 g5d2 e7d5 c4d5 f6f5 h2h4 d8f6 g1e2 h7h6 h4h5 c6e7 d5b3 c8d7 c2c3 e8c8 d1c2 g7g5 e1c1 c5b6 c1b1 c8b8 d2c1 h8f8 h1f1 f6g7 a2a3 e7c6 b1a1 f5f4 g2g4 d7e8 b3a2 b6e3 c1e3 f4e3 c2b3 b8a8');
INSERT INTO evaluations VALUES(10384,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(10384,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10384,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(10384,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(10384,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10384,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10384,36,'cp 0','d7d5 e4d5 f6d5 e1g1 b4c3 b2c3 c8g4 f1e1 e8g8 c1d2 f8e8 h2h3 g4h5 c4b3 h7h6 a2a4 d5f6 g2g4 f6g4 h3g4 h5g4 e1e4 g4h5 g1g2 d8f6 a4a5 f6g6 g2h2 g6f5 h2g3 f5g6');
INSERT INTO evaluations VALUES(9123,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(9123,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(9123,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9123,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(9123,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9123,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9123,36,'cp 0','e4d5 f6d5 e1g1 b4c3 b2c3 c8g4 f1e1 e8g8 c1d2 g4h5 h2h3 f8e8 a1b1 a7a6 a2a4 d5b6 c4b3 h7h6 d1e2 b6d5 b3a2 a8b8 e2e4 d5f6 e4e3 f6d5');
INSERT INTO evaluations VALUES(9104,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9104,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9104,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9104,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9104,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9104,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9104,36,'cp 0','f6d5 e1g1 b4c3 b2c3 c8g4 f1e1 e8g8 c1d2 f8e8 h2h3 g4h5 a1b1 a7a6 c4b3 h7h6 g2g4 h5g6 c3c4 d5b6 c4c5 b6d7 b3d5 a8b8 b1b7 b8b7 d5c6 b7b8 d1e2 e8e6 c6d5 e6e8');
INSERT INTO evaluations VALUES(9183,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9183,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9183,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9183,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9183,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9183,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9183,36,'cp 0','e1g1 b4c3 b2c3 c8g4 f1e1 e8g8 c1d2 g4h5 h2h3 f8e8 c4b3 h7h6 a2a4 d5f6 g2g4 f6g4 h3g4 h5g4 e1e4 g4h5 b3a2 d8f6 g1g2 a8d8 d1e2 f6g6 g2h3 g6f5 h3g3 f5g6');
INSERT INTO evaluations VALUES(10029,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(10029,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10029,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10029,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(10029,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10029,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10029,36,'cp 5','c6d4 b5a4 d4f3 d1f3 e8g8 c1e3 c5e3 f3e3 d7d6 b1c3 c7c6 a4b3 a7a5 a1d1 f8e8 f2f4 c8e6 f4e5 d6e5 b3e6 e8e6 h2h3 d8e7 c3e2 f6d7 e2g3 g7g6 f1f2 e7c5 g3f1 a5a4 e3c5 d7c5 f1e3 a8d8 d1f1 e6e7 b2b4 a4b3 a2b3 c5e6 e3c4');
INSERT INTO evaluations VALUES(10758,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(10758,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(10758,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(10758,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10758,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(10758,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(10758,36,'cp 1','b5a4 d4f3 d1f3 e8g8 c1e3 c5e3 f3e3 d7d6 b1c3 c7c6 a4b3 c8e6 d3d4 d8b6 f1d1 f8e8 d1d2 e6b3 a2b3 d6d5 e4d5 c6d5 d4e5 b6e3 f2e3 e8e5 a1d1 a8d8 e3e4 f6e4 c3e4 e5e4 d2d5 d8e8 d1d2 g8f8 g1f2 b7b6 d5d7 e8e7 d7d8 e7e8 d8e8 e4e8');
INSERT INTO evaluations VALUES(14801,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14801,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(14801,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(14801,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(14801,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(14801,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(14801,36,'cp 22','f8d6 f4g3 e8g8 b1d2 b7b6 f1d3 c8b7 b2b4 c6e7 d1c2 c7c5 b4c5 b6c5 a1b1 b7c6 d4c5 d6c5 e1g1 e7g6 c3c4 d5c4 d3c4 c5e7 f3d4 c6d5 c2a4 d5c4 d2c4 f6e4 f1c1 a8c8 a4a7 e7c5 a7b7 e4g3 h2g3 c5d4 e3d4 d8d4');
INSERT INTO evaluations VALUES(9236,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(9236,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(9236,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(9236,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9236,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(9236,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(9236,36,'cp 26','b1d2 e8g8 f4g3 c6e7 f1d3 b7b6 g3e5 d6e5 f3e5 c7c5 h2h4 c8b7 h4h5 f6e4 h5h6 g7g6 d1e2 f7f6 e5g4 f6f5 g4e5 e4d2 e1d2 e7c6 f2f4 c6e5 f4e5 d8g5 a1g1 c5d4 c3d4 a8c8 g2g4 f5f4 g1f1 f4e3 e2e3 g5g4 f1g1 g4f3 a2a3 f3f2 e3f2 f8f2 d2e3 f2b2');
INSERT INTO evaluations VALUES(9234,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(9234,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(9234,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(9234,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(9234,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(9234,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9234,36,'cp 25','e8g8 f4g3 b7b6 f1d3 c8b7 e1g1 c6e7 b2b4 c7c5 b4c5 b6c5 a1b1 b7c6 d4c5 d6c5 f3d4 c6d7 d2f3 e7g6 c3c4 c5d4 f3d4 d5c4 d3c4 f6e4 d1e1 h7h6 b1b7 e4g3 h2g3 g6e5 c4e2 e5c6 d4b3 d8e7 e2b5 f8c8 b5a4 c6e5 e1a5');
INSERT INTO evaluations VALUES(20996,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(20996,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(20996,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20996,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20996,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20996,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20996,36,'cp -2','g8f6 f1b5 d8d6 f2f4 a7a6 b5d3 g7g6 c1d2 f8g7 g1f3 e8g8 f3e5 c6b4 a2a3 b4d3 c2d3 c7c5 d4c5 d6c5 a1c1 c5d6 d3d4 c8d7 e1g1 a8c8 h2h3 c8c7 g2g4 f8c8 d1f3 e7e6 g4g5 f6e4 d2e1 f7f6 g5f6 g7f6 c3e4 d5e4');
INSERT INTO evaluations VALUES(42043,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(42043,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(42043,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42043,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42043,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42043,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(42043,36,'cp 3','e8g8 f1d3 d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 c1g5 c8b7 d1e2 b8d7 f3e5 a8c8 a1c1 h7h6 e5d7 d8d7 g5f6 g7f6 f1d1 f6f5 d4d5 e6d5 c4d5 b4c3 d5b7 d7b7 b2c3 b7e4 e2h5 c8c6 g2g3 e4g4 h5g4 f5g4 d1d7 h6h5 d7a7 f8c8 a7a4 c6c3 c1c3 c8c3 g1g2 g8g7');
INSERT INTO evaluations VALUES(53968,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53968,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(53968,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53968,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53968,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53968,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53968,36,'cp 0','a7a6 a2a4 d8c7 c4b3 g8f6 d2d4 c5d4 f3d4 b8c6 c1e3 f8e7 e1g1 c6a5 d1e2 c8d7 b3a2 a8c8 a1d1 a5c4 e3c1 c7c5 d4b3 c5a7 b3d4');
INSERT INTO evaluations VALUES(50730,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50730,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50730,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50730,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(50730,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50730,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(50730,36,'cp 0','d2d4 b7b5 c4d3 c5d4 f3d4 g8f6 a2a4 b5b4 c3a2 e6e5 d4f3 b8c6 e1g1 f8e7 c2c3 d6d5 e4d5 d8d5 d3c2 d5d1 f1d1 b4c3 a2c3 a8b8 d1e1 e7d6 e1d1');
INSERT INTO evaluations VALUES(50743,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50743,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50743,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50743,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(50743,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(50743,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50743,36,'cp 0','c5d4 f3d4 g8f6 e1g1 b7b5 c4d3 c8b7 d1e2 b8d7 b2b4 d7e5 a2a4 b5a4 c3a4 f8e7 d4b3 e8g8 c1e3 a8b8 e3a7 b8a8');
INSERT INTO evaluations VALUES(50760,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50760,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50760,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50760,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(50760,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(50760,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(50760,36,'cp 9','f3d4 g8f6 a2a4 f8e7 c1e3 b8c6 h2h3 e8g8 e1g1 c8d7 c4d3 d8c7 d1e2 d6d5 e4d5 e6d5 d4c6 d7c6 e3d4 a8e8 e2f3 f6e4 c3d5 c6d5 d3e4 d5e4 f3e4 e7f6 e4d3 f6d4 d3d4 c7c2 f1c1 c2e4 c1c4 e4d4 c4d4 e8e2 b2b3 f8e8 d4d7');
INSERT INTO evaluations VALUES(50756,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(50756,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50756,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(50756,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(50756,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(50756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50756,36,'cp 1','g8f6 a2a4 f8e7 e1g1 e8g8 c1e3 b8c6 h2h3 c8d7 c4d3 d8c7 d1e2 d6d5 e4d5 e6d5 d4c6 d7c6 e3d4 a8e8 e2f3 f6e4 c3e4 d5e4 d3e4 c6e4 f3e4 e7c5 e4d3 c5d4 d3d4 c7c2 f1c1 c2e4 c1c4 e4d4 c4d4 e8e7 a1d1 g7g6 b2b4 f8c8 d4d7 g8f8 d1d6 c8c2 d7e7 f8e7');
INSERT INTO evaluations VALUES(51864,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(51864,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(51864,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(51864,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(51864,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(51864,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(51864,36,'cp 163','d2d4 e5d4 d1d4 g8f6 b1c3 d7d5 c1g5 f8e7 e4d5 h7h6 g5f6 e7f6 d4e3 d8e7 c3e4 c6d5 e4f6 g7f6 e3e7 e8e7 e1c1 c8e6 g2g3 b8c6 f1g2 h8d8 f3d4 a8c8 h1e1 c6d4 d1d4 c8c5 g2f1 e7f8 f1d3 f8g7 a2a3 f6f5 c1d2 b7b6 e1e3 a7a5 h2h3 d8c8 h3h4 e6d7');
INSERT INTO evaluations VALUES(51856,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(51856,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(51856,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(51856,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(51856,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(51856,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(51856,36,'cp 165','e5d4 d1d4 g8f6 b1c3 d7d5 c1g5 f8e7 e4d5 h7h6 g5f6 e7f6 d4e3 d8e7 c3e4 c6d5 e4f6 g7f6 e3e7 e8e7 e1c1 c8e6 g2g3 b8c6 f3d4 h8d8 f1g2 e7f8 d1d2 d8d7 h1e1 a8e8 a2a3 h6h5 e1e3 h5h4 g3h4 f6f5 e3c3 c6d4 d2d4 f8g7 c3c5 e8h8 g2f3');
INSERT INTO evaluations VALUES(51883,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(51883,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(51883,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(51883,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(51883,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(51883,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(51883,36,'cp 159','d1d4 g8f6 b1c3 d7d5 c1g5 f8e7 e4d5 h7h6 g5f6 e7f6 d4e3 d8e7 c3e4 c6d5 e4f6 g7f6 e3e7 e8e7 e1c1 c8e6 g2g3 h6h5 f1g2 b8c6 f3h4 a8d8 d1d2 e7f8 h1d1 e6g4 d1e1 g4e6 b2b3 f8g7 f2f4 d5d4 g2c6 b7c6');
INSERT INTO evaluations VALUES(54177,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54177,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54177,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54177,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(54177,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(54177,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54177,36,'cp 19','g1f3 b8c6 c4d5 e6d5 f1e2 g8f6 e1g1 c5d4 f3d4 f8d6 b2b3 d8e7 b1c3 c6d4 d1d4 d6e5 e2b5 e8f8 d4a4 e7c7 c1b2 e5h2 g1h1 h2e5 a1c1 c7e7 a4h4 a7a6 b5d3 c8e6 f2f4 e5d6 e3e4 f6e4 h4e7 f8e7 c3e4 d5e4 d3e4 a8c8 c1c8 h8c8 b2g7');
INSERT INTO evaluations VALUES(34259,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(34259,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(34259,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(34259,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(34259,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(34259,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(34259,36,'cp 64','b1c3 g8f6 g1f3 b8c6 f3d2 a5c5 e2e4 e7e6 d2b3 c5b6 a2a3 f8e7 c1e3 b6c7 f2f4 e8g8 d1f3 a7a6 e4e5 f6e8 f1d3 b7b5 e1c1 c8b7 c3e4 d7d6 e5d6 e8d6 e4c5 c6b4 d3h7 g8h7 f3h3 h7g8 a3b4 a8c8 c1b1 b7e4 e3d2 f8d8 d2c3 d6f5 d1d8 c7d8');
INSERT INTO evaluations VALUES(34257,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(34257,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(34257,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(34257,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(34257,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(34257,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(34257,36,'cp 62','g8f6 g1f3 b8c6 f3d2 a5c5 e2e4 e7e6 d2b3 c5b6 a2a3 f8e7 c1e3 b6c7 f2f4 e8g8 d1f3 a7a6 e4e5 f6e8 f1d3 d7d6 e5d6 e8d6 e1c1 a8b8 c1b1 b7b5 e3c5 d6b7 c5e7 c6e7 h1e1 b7d6 f3h5 f7f5 h5e2 d6c4 d3c4 b5c4 b3d2 e7d5 c3d5 e6d5 g2g3 g8h8');
INSERT INTO evaluations VALUES(42487,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(42487,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(42487,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(42487,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(42487,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(42487,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(42487,36,'cp 41','b1d2 b4e7 e2e4 d7d5 e4e5 f6d7 f1d3 c7c5 e1g1 b8c6 d2b3 e8g8 c4d5 e6d5 b3c5 e7c5 d4c5 d7c5 c1g5 d8c7 d3b1 c8e6 f1e1 c5e4 b1e4 d5e4 e1e4 f8e8 h2h3 h7h6 g5h4 e6f5 e4f4 f5g6 a1c1 c7b6 d1e2 b6a5 c1e1 a5a2');
INSERT INTO evaluations VALUES(42486,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(42486,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(42486,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(42486,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42486,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(42486,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(42486,36,'cp 33','b4e7 e2e4 d7d5 e4e5 f6d7 f1d3 c7c5 e1g1 b8c6 f1e1 e8g8 a2a3 a7a5 d3c2 c5d4 c4d5 e6d5 d2b3 d8b6 b3d4 c6d4 d1d4 b6d4 f3d4 e7c5 d4b5 d7b6 c1e3 c5e3 e1e3 b6c4 e3e2 a5a4 c2d3 c8g4 f2f3 g4e6 d3c4 d5c4 b5d6 a8d8 d6b7');
INSERT INTO evaluations VALUES(16654,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(16654,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(16654,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(16654,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(16654,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(16654,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(16654,36,'cp -4','f6e4 f1e1 d7d5 c4d5 d8d5 b1c3 d5d7 c3e4 f8e7 c1g5 e8g8 g5e7 d7e7 f3d4 c8d7 e4c3 e7b4 d4b5 a8d8 b5c7 b4b2 c3d5 b2d4 c2c3 d4d1 a1d1 d7c8 c3c4 b7b6 c7b5 d8d7 d5c7 f8d8 d1d7 c8d7 e1d1 g8f8 f2f3 a7a5 d1d5 a5a4 h2h4 f8g8 h4h5 c6a5 c4c5 b6c5 d5c5');
INSERT INTO evaluations VALUES(16861,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(16861,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(16861,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16861,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(16861,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(16861,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(16861,36,'cp -3','f1e1 d7d5 c4d5 d8d5 b1c3 d5d7 c3e4 f8e7 c1g5 e8g8 g5e7 d7e7 f3d4 c8d7 e4c3 e7b4 d4b5 a8d8 b5c7 b4b2 c3d5 d7e6 c2c4 e6d5 c7d5 f8e8 d1b3 e8e2 a1b1 b2b3 a2b3 e2d2 b1d1 d2b2 b3b4 g8f8 d1a1 a7a6 a1b1 b2b1 e1b1 c6e5 d5e3 d8d3 g1f1 d3a3 b4b5 a6b5 b1b5');
INSERT INTO evaluations VALUES(16862,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(16862,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(16862,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16862,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(16862,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16862,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16862,36,'cp -9','d7d5 c4d5 d8d5 b1c3 d5d7 c3e4 f8e7 c1g5 e8g8 g5e7 d7e7 f3d4 c8d7 e4c3 e7c5 d4b3 c5f5 b3c5 f5c5 d1d7 a8c8 h2h3 f8d8 d7g4 c5b4 c3e4 b4b2 a1b1 b2c2 b1b7 c6d4 b7a7 c2c6 e4g3 c6f6 e1e4 f6b6 a7a4 c7c5 g3f5 d4f5 g4f5 b6c7');
INSERT INTO evaluations VALUES(15366,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15366,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(15366,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15366,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(15366,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15366,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15366,36,'cp 0','c4d5 d8d5 b1c3 d5d7 c3e4 f8e7 c1g5 e8g8 g5e7 d7e7 f3d4 c8d7 e4c3 e7b4 d4b5 a8d8 b5c7 b4b2 c3d5 b2d4 c2c3 d4d1 a1d1 d7c8 c3c4 b7b6 c7b5 h7h6 f2f3 g8h7 g1f2 d8d7 d5e3 f8d8 d1d5 c6b4 d5d7 d8d7 e1d1 d7d1 e3d1 c8e6 d1b2 b4a2 b5a7 a2c3 f2e1');
INSERT INTO evaluations VALUES(15272,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15272,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15272,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15272,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15272,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15272,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15272,36,'cp -9','d8d5 b1c3 d5d7 c3e4 f8e7 c1g5 e8g8 g5e7 d7e7 f3d4 c8d7 e4c3 e7b4 d4b5 a8d8 b5c7 b4b2 c3d5 d7f5 c2c4 b2c2 d1c2 f5c2 a1c1 c2f5 h2h3 h7h5 c1d1 d8d7 c7e8 g8h8 g1h2 f5e6 e8c7 e6d5 c7d5 f8d8 e1e4 c6a5 d1b1 g7g6 h2g3 h8g7 b1b5 b7b6');
INSERT INTO evaluations VALUES(974,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(974,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(974,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(974,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(974,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(974,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(974,36,'cp 0','b1c3 d5d7 c3e4 f8e7 c1g5 e8g8 g5e7 d7e7 f3d4 c8d7 e4c3 e7b4 d4b5 a8d8 b5c7 b4b2 c3d5 d7e6 c2c4 e6d5 c7d5 f8e8 e1e8 d8e8 a1b1 b2a2 b1b7 a2c4 d5e7 c6e7 b7e7 c4b5 e7a7 b5d5 a7a1 d5d1 a1d1 e8a8 g1f1 g7g5 g2g3 a8a5 h2h3 h7h5 d1b1 g8g7 f1e2 g7f6 e2e3 a5a2 e3f3 a2a5');
INSERT INTO evaluations VALUES(973,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(973,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(973,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(973,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(973,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(973,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(973,36,'cp -5','d5d7 c3e4 f8e7 c1g5 e8g8 g5e7 d7e7 f3d4 c8d7 e4c3 e7c5 d4b3 c5f5 b3c5 f5c5 d1d7 c5a5 d7d5 a5b4 d5b5 f8d8 a1d1 a7a6 b5b4 c6b4 d1c1 g8f8 a2a3 b4c6 e1d1 d8d4 c3e2 d4d1 c1d1 a8d8 d1d8 c6d8 f2f3 f8e7 g1f2 c7c5 g2g4 e7f6 f2e3 f6e5 e2f4');
INSERT INTO evaluations VALUES(15691,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(15691,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(15691,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(15691,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(15691,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(15691,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(15691,36,'cp -85','d8d4 e5f3 d4e4 d1e2 e4e2 e1e2 f6d5 d2d3 f7f6 b1c3 e8f7 h1e1 a7a5 a2a4 f8d6 c3e4 d6e7 c1d2 h8e8 e2f1 c6c5 e4c3 d5b4 a1c1 c8f5 d2f4 a8c8 c3b5 e7d8 e1e8 f7e8 h2h3 e8d7 f3d2 g7g5 f4h2 h7h5 d2e4 b7b6');
INSERT INTO evaluations VALUES(1012,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(1012,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(1012,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(1012,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(1012,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(1012,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(1012,36,'cp -80','e5f3 d4e4 d1e2 e4e2 e1e2 f6d5 b1c3 f7f6 d2d3 e8f7 h1e1 h7h5 c3e4 h5h4 h2h3 g7g5 e2f1 g5g4 h3g4 c8g4 e4g5 f7g8 g5h3 a7a5 e1e4 g4f3 g2f3 g8f7 c1d2 a5a4 a1e1 a4a3 b2b3 f8d6 e4e6 c6c5 f3f4 b7b5 c2c4 b5c4 d3c4 d5b4 d2b4 c5b4');
INSERT INTO evaluations VALUES(1013,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(1013,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(1013,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(1013,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(1013,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(1013,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(1013,36,'cp -87','d4e4 d1e2 e4e2 e1e2 f6d5 d2d3 f7f6 b1c3 e8f7 h1e1 f8e7 c1d2 a7a5 e2f1 c8d7 e1e2 a5a4 a1e1 h8e8 a2a3 b7b5 c3d5 c6d5 f3d4 g7g5 d2b4 c7c5 e2e7 e8e7 b4c5 e7e1 f1e1 h7h5 f2f3 a8e8 e1d1 e8c8 c5b4 c8g8 d1e2 g8h8');
INSERT INTO evaluations VALUES(1014,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(1014,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(1014,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(1014,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(1014,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(1014,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(1014,36,'cp -75','d1e2 e4e2 e1e2 f6d5 h1e1 f7f6 b1c3 e8f7 d2d3 f8e7 c1d2 a7a5 e2f1 c8d7 e1e2 a5a4 a1e1 h8e8 a2a3 b7b5 c3d5 c6d5 f3d4 g7g5 d2b4 c7c5 e2e7 e8e7 e1e7 f7e7 b4c5 e7f7 f1e1 a8e8 e1d1 e8g8 d1e1 h7h5 f2f3 h5h4 e1d2 g8c8 c5d6 f7g6 d6b4 c8e8 d2d1 e8h8 d1d2 g6f7 b4c5');
INSERT INTO evaluations VALUES(1015,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(1015,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(1015,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(1015,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(1015,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(1015,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(1015,36,'cp -86','e4e2 e1e2 f6d5 b1c3 f7f6 d2d3 e8f7 h1e1 f8e7 c1d2 a7a5 e2f1 c8d7 a2a3 d5c3 d2c3 a5a4 e1e2 h8e8 a1e1 c6c5 c3e5 e7d8 e5f4 e8e2 f1e2 h7h5 e2f1 g7g5 f4e3 d8e7 h2h4 d7c6 f3d2 g5h4 e3f4 a8c8 d2e4 e7f8 f2f3');
INSERT INTO evaluations VALUES(1017,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(1017,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(1017,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(1017,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(1017,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(1017,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(1017,36,'cp -78','e1e2 f6d5 d2d3 f7f6 b1c3 e8f7 h1e1 b7b6 e2f1 d5c3 b2c3 g7g5 a2a4 a7a5 f3d2 c8d7 c1b2 f8e7 c3c4 h8e8 d2e4 g5g4 e4g3 e7b4 e1c1 b6b5 g3e4 b5a4 a1a4 c6c5 a4a1 a5a4 b2f6 a4a3');
INSERT INTO evaluations VALUES(1016,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(1016,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(1016,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(1016,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(1016,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(1016,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(1016,36,'cp -85','f6d5 b1c3 f7f6 d2d3 e8f7 h1e1 b7b6 c3d5 c6d5 c1f4 c7c5 h2h4 h7h6 e2f1 g7g5 f4g3 c8d7 b2b3 a8c8 a2a4 f8e7 f3d2 h8e8 e1e2 e7f8 e2e8 c8e8 h4g5 h6g5 f2f3 f6f5 a1e1 f5f4 g3f2 e8e1 f2e1 f8d6 f1f2 f7f6');
INSERT INTO evaluations VALUES(56017,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(56017,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(56017,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56017,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(56017,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(56017,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(56017,36,'cp 15','g8f6 f1c4 a7a6 a2a4 c7c5 g1f3 b8c6 e1g1 c5d4 e3d4 f8e7 d4d5 e6d5 c3d5 f6d5 c4d5 e8g8 d1b3 c6a5 b3a2 a5c6 f1d1 c8e6 c1f4 d8c8 a2b3 e6d5 d1d5 f8d8 d5d8 c6d8 a1e1 c8d7 g2g3 a8c8 e1d1 d7e6 b3e6 d8e6 d1d7 e7f6 d7b7 e6c5 b7a7 c5a4 b2b3 a4c5 g1g2 c5b3');
INSERT INTO evaluations VALUES(48486,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(48486,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(48486,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(48486,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48486,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48486,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(48486,36,'cp 12','f1c4 a7a6 a2a4 c7c5 g1f3 b8c6 e1g1 c5d4 e3d4 f8e7 c1e3 e8g8 d1e2 c6b4 f3e5 c8d7 f1d1 d7e8 c4b3 a8c8 h2h3 e8c6 e5c6 c8c6 a1c1 h7h6 e3f4 f6d5 f4e5 a6a5 c3e4 c6c1 d1c1 b4c6 e2g4 c6e5');
INSERT INTO evaluations VALUES(44530,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(44530,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(44530,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44530,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(44530,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(44530,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(44530,36,'cp 42','g1f3 b8c6 g2g3 d7d5 d2d4 c5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 f1g2 f6d5 d1e3 c8a6 e3d5 c6d5 g2d5 a8c8 c1d2 e7e6 d5f3 f8c5 d2c3 f7f6 e1c1 e8e7 e2e3 c8c7 g3g4 a6b7 f3b7 c7b7 g4g5 e7f7 g5f6 g7f6 h1g1 h7h5 h2h3 h8b8 d1d2 b8e8 g1g3 c5b4 c3b4 b7b4 d2d7 e8e7');
INSERT INTO evaluations VALUES(44528,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44528,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(44528,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(44528,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(44528,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(44528,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(44528,36,'cp 31','b8c6 g2g3 d7d5 d2d4 c5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 f1g2 f6d5 d1e3 c8a6 e3d5 c6d5 g2d5 a8c8 c1d2 e7e6 d5f3 f8c5 d2c3 e8g8 e1g1 f8d8 f1d1 f7f6 h2h4 g8f7 e2e3 d8d1 a1d1 a6b5 g1f1 f7e7 g3g4 h7h6 f1e1 c8b8 d1c1 c5d6 h4h5 b5e8 f3e2');
INSERT INTO evaluations VALUES(12804,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(12804,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(12804,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(12804,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(12804,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(12804,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(12804,36,'cp 23','g2g3 d7d5 d2d4 c5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 f1g2 f6d5 d1e3 c8a6 e3d5 c6d5 g2d5 a8c8 c1d2 e7e6 d5f3 f8c5 d2c3 f7f6 g3g4 h7h6 h2h4 a6b5 e2e3 b5a4 a1c1 c8b8 h1g1 e8f7 f3e2 a4b5 g4g5 h6g5 h4g5 e6e5 c1d1 h8h2 g5f6 g7f6');
INSERT INTO evaluations VALUES(12805,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(12805,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(12805,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(12805,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(12805,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(12805,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(12805,36,'cp 36','d7d5 d2d4 c5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 f1g2 f6d5 d1e3 c8a6 e3d5 c6d5 g2d5 a8c8 c1d2 e7e6 d5f3 f8c5 d2c3 e8g8 e1g1 f8d8 f1d1 f7f6 h2h4 g8f7 e2e3 d8d1 a1d1 a6b5 g1f1 a7a5 a2a3 f7e7 h4h5 h7h6 g3g4 c5d6 f3e2 d6e5 c3e5 f6e5 d1c1 b5a6');
INSERT INTO evaluations VALUES(55635,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(55635,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55635,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(55635,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(55635,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(55635,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(55635,36,'cp 12','c2c4 g8f6 b1c3 c7c6 d2d4 f8e7 d4e5 d6e5 d1d8 e7d8 g1f3 d8c7 g2g3 e8g8 f1g2 f8e8 f3d2 c8f5 d2e4 f6e4 c3e4 c7a5 e1f1 a5b4 e2e3 b8d7 f1e2 d7f6 e4f6 g7f6 h1d1 e8d8 d1d8 a8d8 a1d1 d8d1 e2d1 f5g4 f2f3 g4e6 d1c2 b4c5 e3e4 b7b5 c4b5 c6b5');
INSERT INTO evaluations VALUES(55555,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(55555,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55555,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55555,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55555,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55555,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(55555,36,'cp 14','g8f6 b1c3 f8e7 d2d4 c7c6 e2e3 e8g8 g1e2 f8e8 d1c2 c8e6 a1d1 b8d7 e2g3 d8a5 f1e2 d6d5 e1g1 a8c8 a2a3 d5c4 e2c4 e6c4 b3c4 e5d4 d1d4 d7c5 h2h3 c8d8 c3e2 c5e6 d4d1 g7g6 g1h1 e6c5 d1d8 e8d8');
INSERT INTO evaluations VALUES(47829,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(47829,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(47829,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47829,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(47829,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(47829,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(47829,36,'cp 8','b1c3 f8e7 d2d4 e8g8 e2e3 c7c6 g1e2 f8e8 h2h3 b8d7 g2g4 d6d5 c4d5 f6d5 f1g2 d5c3 b2c3 e5d4 e2d4 d7f6 e1g1 h7h5 d4f5 h5g4 d1d8 e7d8 h3g4 d8c7 c3f6 g7f6 f1d1 c8f5 g4f5 a8d8 g2f3 g8f8 g1g2 f8e7 a1c1 d8d1 c1d1 e8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(57737,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(57737,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(57737,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(57737,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(57737,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(57737,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(57737,36,'cp 148','d2d4 e7e6 f1d3 d7d5 b1c3 f8b4 g1e2 d5e4 d3e4 g8f6 c1g5 c7c5 d4d5 e6d5 g5f6 d8f6 d1d5 b8c6 a2a3 b4c3 e2c3 f6e7 e1c1 c8e6 d5d6 e7d6 d1d6 e8e7 d6d3 h8d8 d3g3 g7g6 g3g5 c5c4 h1e1 a8c8 e4d5 e7f6 d5e6 f7e6 c3e4 f6g7 e4c5 e6e5 c5b7 d8d5 h2h3');
INSERT INTO evaluations VALUES(57734,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(57734,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(57734,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(57734,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(57734,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(57734,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(57734,36,'cp 123','e7e6 g1f3 d7d5 e4d5 e6d5 f1d3 c7c5 c2c3 c5d4 c3d4 f8b4 b1c3 g8e7 e1g1 b8c6 a2a3 b4c3 b2c3 c8f5 d3f5 e7f5 f1e1 c6e7 a1b1 d8c7 d1a4 e8f8 c3c4 f7f6 c4d5 b7b6 d5d6 c7d6 b1b5 a8c8 a4a7 d6c7 a7c7 c8c7 b5b6 f8f7 b6b1 h8d8 c1f4');
INSERT INTO evaluations VALUES(56966,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(56966,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(56966,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(56966,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(56966,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(56966,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(56966,36,'cp 130','g1f3 d7d5 e4d5 e6d5 f1d3 c7c5 c2c3 c5d4 c3d4 f8e7 e1g1 e7f6 f1e1 g8e7 c1f4 b8c6 f4e5 f6e5 d4e5 e8g8 h2h3 c8f5 b1c3 d8d7 f3g5 f5d3 d1d3 d7f5 d3f5 e7f5 c3d5 f5d4 e1e4 a8d8 d5f4 f8e8 a1e1 d4c2 e1c1 c2d4 e5e6 d4e6 f4e6 f7e6 g5e6 d8d2');
INSERT INTO evaluations VALUES(56965,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(56965,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(56965,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(56965,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(56965,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(56965,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(56965,36,'cp 140','d7d5 e4d5 e6d5 f1d3 c7c5 c2c3 c5d4 c3d4 f8b4 b1c3 g8e7 e1g1 b8c6 a2a3 b4c3 b2c3 c8f5 d3f5 e7f5 f1e1 c6e7 a1b1 b7b6 f3e5 d8c8 a3a4 e8g8 c1a3 h5h4 d1f3 f7f6 e5d3 f8f7 a3b4 c8d7 a4a5 e7c6 h2h3 g7g5 e1e2 f5g7 b1e1 b6a5 b4c5 d7f5');
INSERT INTO evaluations VALUES(51967,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(51967,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51967,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51967,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(51967,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(51967,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(51967,36,'cp 55','e2e4 c5d4 f3d4 b8c6 b1c3 a7a6 c1e3 g8f6 a2a3 d8c7 f2f4 d7d5 e4e5 f6d7 d1g4 g7g6 e1c1 c6d4 e3d4 d7b8 c3a4 b8d7 g4g3 b7b5 a4c3 f8c5 c3d5 e6d5 e5e6 e8g8 e6d7 c8d7 d4e5 c7b6 d1d5 d7e6 d5d1 a8c8 f1d3');
INSERT INTO evaluations VALUES(54983,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(54983,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54983,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(54983,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(54983,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(54983,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54983,36,'cp 3','c7c5 g1f3 g7g6 f1b5 c8d7 b5d7 d8d7 c2c3 g8f6 d1e2 f8g7 d2d4 e8g8 e1g1 c5d4 c3d4 b7b5 c1d2 b8a6 a2a4 a6c7 a4b5 c7b5 b1c3 f8b8 e4e5 d6e5 d4e5 f6d5 f1d1 d5c3 b2c3 d7c6 c3c4 b5c7 d2e3 c7e6');
INSERT INTO evaluations VALUES(53574,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(53574,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(53574,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(53574,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(53574,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(53574,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(53574,36,'cp 28','f1d3 d5c4 d3c4 a7a6 g1f3 g8f6 e1g1 c6c5 d4c5 d8d1 f1d1 f8c5 b1d2 e8g8 c4e2 f8d8 f3e5 b7b5 e5d3 c5b6 d2b3 c8b7 c1d2 b7d5 d2c3 b8d7 d3b4 h7h6 c3d4 b6d4 b3d4 d7e5 b4d5 f6d5 a2a4 b5a4 a1a4');
INSERT INTO evaluations VALUES(52079,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(52079,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(52079,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52079,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(52079,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52079,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(52079,36,'cp 24','c3d4 d7d5 e4e5 g8e7 b1c3 e7f5 g1f3 b8c6 a2a3 a7a6 c1g5 d8b6 c3a4 b6a7 a1c1 c8d7 g5e3 c6a5 f1d3 f5e3 f2e3 f8e7 e1g1 a5c4 d3c4 d5c4 a4c3 e8g8 f3d2 b7b5 d2e4 a6a5 d1g4 g8h8 e4d6 e7d6 e5d6 f7f6 g4g3 a8d8 g1h1 h7h6');
INSERT INTO evaluations VALUES(52066,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52066,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52066,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(52066,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(52066,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(52066,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(52066,36,'cp 61','d7d5 e4e5 g8e7 g1f3 b8c6 b1c3 e7f5 f1b5 c8d7 e1g1 a8c8 b5a4 d8b6 c3e2 f8e7 a2a3 c6a5 a4c2 d7b5 f1e1 b5e2 e1e2 e8g8 b2b4 a5c4 a1b1 g7g6 b1b3 c8c6 b4b5 c6c7 a3a4 h7h5 c2f5 e6f5 e5e6 f7e6 c1f4 c7c8 d1d3');
INSERT INTO evaluations VALUES(54385,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(54385,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(54385,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(54385,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(54385,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(54385,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(54385,36,'cp 34','e4e5 g8e7 g1f3 b8c6 b1c3 e7f5 f1b5 d8b6 b5a4 c8d7 e1g1 f8e7 c3e2 c6b4 a4d7 e8d7 g2g4 f5h4 f3h4 e7h4 c1e3 h4e7 a2a3 b4c6 a1c1 a8c8 c1c3 c8c7 h2h3 h7h5 g4g5 h8c8 d1c2 g7g6 f1c1 d7e8 e2f4 a7a5 c2a4');
INSERT INTO evaluations VALUES(50349,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(50349,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(50349,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(50349,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(50349,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(50349,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(50349,36,'cp -113','f6e4 d2d4 d7d6 c2c4 g7g6 f1d3 e4f6 b1c3 f8g7 e1g1 e8g8 f1e1 b8d7 c1e3 b7b6 d4d5 d7c5 d3e2 e7e5 d5e6 c5e6 f3d4 c8b7 e2f3 b7f3 d1f3 f6d7 a1d1 d8h4 g2g3 h4h3 c3d5 d7e5 f3g2 h3g2 g1g2 e5c4 d4e6 c4e3 f2e3 f7e6');
INSERT INTO evaluations VALUES(50449,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(50449,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(50449,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(50449,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(50449,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(50449,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(50449,36,'cp -111','d2d4 g7g6 f1d3 e4f6 e1g1 d7d6 c2c4 f8g7 b1c3 e8g8 f1e1 b8d7 c1g5 h7h6 g5h4 e7e6 h2h3 b7b6 d3c2 a7a6 f3h2 c8b7 d1d2 d8c8 d4d5 e6d5 c4d5 b6b5 a1d1 c7c5 h2g4 f6g4 h3g4 f8e8 c3e4');
INSERT INTO evaluations VALUES(50442,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(50442,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(50442,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(50442,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(50442,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(50442,35,'cp -110',NULL);
INSERT INTO evaluations VALUES(50442,36,'cp -106','g7g6 f1d3 e4f6 e1g1 f8g7 c2c4 d7d6 f1e1 e7e6 b1c3 b8d7 d3c2 e8g8 c1g5 h7h6 g5f4 b7b6 d1d2 g8h7 h2h3 a7a6 e1e3 c8b7 f4g3 h7g8 a2a3 f6h5 g3h2 f8e8 a1d1 d8e7 a3a4 h5f6 d1e1 e7f8 b2b4');
INSERT INTO evaluations VALUES(50277,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(50277,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(50277,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(50277,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(50277,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(50277,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(50277,36,'cp -91','f1d3 e4f6 e1g1 f8g7 c2c4 d7d6 b1c3 e8g8 f1e1 b8d7 c1g5 h7h6 g5f4 e7e6 d3c2 b7b6 d1d2 g8h7 h2h3 c8a6 b2b3 a6b7 e1e3 a7a6 a2a4 f6h5 f4h2 h7g8 a1e1 f8e8 d4d5 e6e5 f3d4 d7f8 d4c6 b7c6 d5c6 f8e6 c3d5 a6a5');
INSERT INTO evaluations VALUES(22785,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(22785,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(22785,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(22785,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(22785,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(22785,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(22785,36,'cp -20','g8f6 c3d5 f6d5 e4d5 c6e7 d3d4 d7d6 f1c4 g7g6 g1f3 f8g7 a2a4 e7f5 d4e5 d6e5 e1g1 e8g8 f1e1 f5d6 f3d2 a7a5 c4f1 f8e8 d2c4 b7b6 c4d6 d8d6 b2b3 c8b7 c2c4 e5e4 a1b1 c7c6 d5c6 d6c6 d1e2 h7h5 e1d1 a8d8 c1g5 d8d1 b1d1 c6c7');
INSERT INTO evaluations VALUES(16526,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(16526,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16526,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16526,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(16526,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(16526,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(16526,36,'cp -21','g2g3 f8b4 f1g2 d7d5 e4d5 f6d5 c1d2 d5c3 b2c3 b4a5 g1f3 f7f6 a2a4 c8e6 c3c4 a5d2 f3d2 d8d7 d2b3 e6g4 d1d2 b7b6 e1g1 e8g8 a4a5 g4h3 g2d5 g8h8 f1e1 a8d8 a1b1 h3e6 d5g2 e6f7 a5a6 h7h6 h2h3');
INSERT INTO evaluations VALUES(16525,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(16525,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(16525,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(16525,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(16525,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(16525,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(16525,36,'cp -26','f8b4 g1f3 d7d5 e4d5 f6d5 c1d2 d5c3 b2c3 b4d6 f1g2 e8g8 a1b1 a8b8 f3g5 d8d7 e1g1 h7h6 g5e4 d6a3 c3c4 b7b6 e4c3 c8b7 c3d5 c6e7 d5e7 d7e7 g2b7 b8b7 d1h5 f7f6 h5g6 g8h8 g6e4 b7b8 f2f4 a3c5 d2e3 f8e8 f4e5 e7e5 e4e5 c5e3 e5e3');
INSERT INTO evaluations VALUES(55733,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(55733,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(55733,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(55733,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(55733,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(55733,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(55733,36,'cp 61','b8d7 g1f3 g8f6 c1f4 c7c6 h2h3 f8e7 f1d3 d8c7 e1g1 e6e5 f4g3 e8g8 a2a4 f8e8 f1e1 e7f8 a4a5 h7h6 d1d2 a8b8 c3a4 e5d4 f3d4 d7e5 b2b4 c8d7 d3f1 b7b5 a5b6 a7b6 c2c4 c6c5 d4b5 d7b5 c4b5 e5d7 d2c2 b8c8 b4c5 d7c5');
INSERT INTO evaluations VALUES(22385,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(22385,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(22385,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(22385,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(22385,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(22385,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(22385,36,'cp 89','g1f3 g8f6 f1e2 c7c6 e1g1 e6e5 a2a4 f8e7 a4a5 e8g8 f1e1 d8c7 c1e3 h7h6 h2h3 f8e8 d4d5 c6d5 c3d5 f6d5 e4d5 f7f5 c2c4 e7f6 e2f1 d7f8 b2b4 f8g6 c4c5 e5e4 f3d4 f5f4 d4b5 c7d8 b5d6 f6a1 d1a1 f4e3');
INSERT INTO evaluations VALUES(22384,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(22384,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(22384,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(22384,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(22384,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(22384,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(22384,36,'cp 78','g8f6 f1e2 c7c6 e1g1 e6e5 a2a4 f8e7 a4a5 e8g8 f1e1 f8e8 d4d5 c6d5 e4d5 a7a6 f3d2 b7b6 a5b6 d7b6 d2b3 c8b7 b3a5 b7d5 c3d5 b6d5 e2f3 e5e4 f3e4 f6e4 e1e4 d5f6 a5c6 d8c7 c6e7 e8e7 e4e7 c7e7 c1e3 h7h6 e3d4 f6e4 h2h3 e7c7 c2c3 c7c4 d1c2 e4f6');
INSERT INTO evaluations VALUES(15929,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(15929,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(15929,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(15929,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(15929,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(15929,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(15929,36,'cp 88','c1f4 c7c6 h2h3 f8e7 f1d3 d8c7 e4e5 d6e5 d4e5 f6d5 c3d5 c6d5 e1g1 d7c5 c2c3 c5d3 d1d3 e8g8 f3d4 c8d7 f1e1 e7c5 e1e3 f7f6 a1e1 f6f5 e3g3 f8f7 h3h4 a8c8 d3d2 c5d4 d2d4 c7c4 d4d2 d5d4 c3d4');
INSERT INTO evaluations VALUES(55784,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(55784,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(55784,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(55784,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(55784,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(55784,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(55784,36,'cp 154','d5e6 c8e6 g1f3 g8f6 b1c3 f8d6 d2d4 e8g8 f1e2 c7c6 c1g5 b8d7 c3e4 d6e7 e4g3 h7h6 g5c1 b7b5 e1g1 e7d6 b2b3 f8e8 f1e1 d8c7 c2c4 b5c4 b3c4 c6c5 d4d5 e6g4 c1b2 g4f3 e2f3 e8e1 d1e1 a8e8 e1d2 d7e5 f3e2');
INSERT INTO evaluations VALUES(55752,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(55752,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(55752,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(55752,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(55752,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(55752,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(55752,36,'cp 136','c8e6 b1c3 b8c6 g1f3 c6d4 f1e2 g8e7 f3d4 d8d4 e1g1 e7f5 d2d3 e8c8 c3e4 d4d7 c2c3 h7h6 f1e1 g7g5 g2g3 c7c5 e2f1 f8e7 d1f3 c8b8 f1g2 d7d3 f3d3 d8d3 f2f4 g5f4 c1f4 b8a8');
INSERT INTO evaluations VALUES(32257,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(32257,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(32257,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(32257,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(32257,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(32257,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(32257,36,'cp 137','g1f3 g8f6 d2d4 c7c5 f1b5 b8c6 b1c3 d8c7 c1e3 a7a6 b5c6 c7c6 d1e2 c5d4 e3d4 e8c8 e1c1 f8d6 d4e5 e6c4 e2e3 f6g4 e3d4 g4e5 f3e5 c6c7 d4c4 d6e5 c4c7 e5c7 g2g3 d8d1 h1d1 h8e8 a2a3 c7b6 f2f4 b6e3 c1b1 g7g6');
INSERT INTO evaluations VALUES(32255,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(32255,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(32255,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(32255,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(32255,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(32255,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(32255,36,'cp 132','g8f6 b1c3 f8d6 d2d4 e8g8 f1d3 b8c6 e1g1 h7h6 a2a3 e6g4 c1e3 g4h5 h2h3 d8d7 c3b5 d6e7 d3e2 a7a6 b5c3 a8d8 f3h4 h5e2 c3e2 f8e8 h4f3 b7b5 d1d3 e7d6 e3d2 c6e7 a1e1 c7c5 d4c5 d6c5 d3d7 d8d7 e2c1 e8c8 c1d3 c5d6 d2a5 e7g6 a5c3');
INSERT INTO evaluations VALUES(30043,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(30043,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(30043,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(30043,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(30043,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(30043,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(30043,36,'cp 125','d2d4 c7c5 f1b5 b8c6 b1c3 d8b6 e1g1 e8c8 b5c6 b7c6 c3a4 b6b5 c2c4 b5c4 c1e3 e6g4 a1c1 c4b5 a4c3 b5a6 d1b3 g4f3 g2f3 c5c4 b3c2 f8b4 c3e4 f6d5 a2a3 b4d6 b2b3 c4b3 c2b3 d6c7 b3c2 h8e8 c2c6 a6c6');
INSERT INTO evaluations VALUES(30037,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(30037,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(30037,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(30037,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(30037,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(30037,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(30037,36,'cp 125','c7c5 f1b5 b8c6 e1g1 d8b6 b1c3 e8c8 b5c6 b6c6 c1e3 f8d6 d1c1 h8e8 d4c5 d6b8 f1e1 e6g4 f3d4 c6c7 f2f4 f6e4 c3e4 e8e4 c2c3 d8e8 h2h3 g4d7 b2b4 g7g5 c1d2 g5f4 e3f2 e4e1 a1e1 e8e1 f2e1 c7e5 d4f3 e5e3 e1f2 e3d2 f3d2 b8c7 f2d4');
INSERT INTO evaluations VALUES(22725,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(22725,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(22725,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(22725,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(22725,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(22725,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(22725,36,'cp -40','g1f3 f8c5 d2d3 d7d6 a2a3 a7a5 c1e3 g8e7 d3d4 e5d4 f3d4 c8e6 b1c3 e8g8 f1e2 c6d4 e3d4 e7c6 d4c5 d6c5 e1g1 c6d4 c3b5 d4b5 c4b5 d8f6 d1c2 b7b6 f1d1 f8d8 d1d8 a8d8 a1d1 d8d1 c2d1 g7g6 d1d2 f6e5 f2f3 c5c4 a3a4 h7h6 d2c3 e5c5 g1f1 h6h5');
INSERT INTO evaluations VALUES(55402,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(55402,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(55402,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55402,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55402,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(55402,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(55402,36,'cp 28','g8f6 e1g1 f8e7 f1e1 c7c5 h2h3 b8c6 c2c3 d8c7 b1a3 a7a6 d3d4 b7b5 c4d3 e5d4 c3d4 c5d4 a3c2 c6e5 c2d4 e8g8 c1f4 e5d3 d1d3 f8e8 a1c1 c7d7 d4f5 c8b7 f3d2 e7f8 d3g3 g8h8 f5d6 f8d6 f4d6 b7e4 d2e4');
INSERT INTO evaluations VALUES(47619,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(47619,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47619,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(47619,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(47619,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(47619,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(47619,36,'cp 25','e1g1 f8e7 f1e1 e8g8 d3d4 e5d4 f3d4 f8e8 b1c3 c7c6 c4f1 b7b5 c1f4 e7f8 a2a3 a7a6 d1d2 c6c5 d4e2 b8c6 e2g3 c8e6 a1d1 c6d4 f4e3 e6g4 d1a1 d8b6 a1b1 a8d8 e3d4 c5d4 c3d5 f6d5 e4d5 e8e1 b1e1 g4d7 h2h3 d8e8');
INSERT INTO evaluations VALUES(47618,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47618,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47618,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(47618,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47618,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(47618,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(47618,36,'cp 28','f8e7 f1e1 e8g8 h2h3 f8e8 a2a4 e7f8 b1c3 b8a6 d3d4 a6b4 d4e5 d6e5 d1d8 e8d8 e1e2 b4c6 c1e3 a7a6 c3d5 c8e6 d5f6 g7f6 c4e6 f7e6 g2g4 g8f7 g1g2 d8d6 e2d2 a8d8 a1d1 d6d2 d1d2 f7e8 d2d8 c6d8 f3d2 d8c6 d2c4');
INSERT INTO evaluations VALUES(40689,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(40689,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40689,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40689,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40689,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40689,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(40689,36,'cp 22','f1e1 e8g8 h2h3 a7a6 a2a4 c7c5 c1d2 b8c6 b1c3 c8e6 c3d5 e6d5 e4d5 c6d4 f3d4 c5d4 a4a5 f6d7 d1g4 f7f5 g4g6 f8f6 g6h5 f6f8 f2f4 e7h4 e1e2 a8c8 d2b4 d8f6 b4d6 f6d6 h5h4 e5e4 h4g3 c8e8 a1f1 d7f6 g3f2 e4e3 f2f3 b7b5 a5b6');
INSERT INTO evaluations VALUES(57659,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(57659,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(57659,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(57659,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(57659,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(57659,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(57659,36,'cp 18','g8f6 c2c4 f8g7 g1f3 e8g8 f1e2 c7c5 d4d5 d7d6 b1c3 f8e8 e3e4 e7e6 f3d2 b7b6 e1g1 b8a6 d5e6 e8e6 f2f3 a6c7 d2b3 c8b7 c1g5 h7h6 g5h4 g6g5 h4f2 f6h5 f1e1 g7c3 b2c3 d8f6 d1c2 a8e8 e2f1 f6g6 b3d2 h5f4 a1d1');
INSERT INTO evaluations VALUES(55123,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(55123,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55123,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(55123,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55123,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(55123,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(55123,36,'cp -23','g8f6 g2g3 c7c5 c2c4 d5c4 d1a4 c8d7 a4c4 e7e6 f3e5 b8c6 e5d7 d8d7 f1g2 a8c8 c4a4 a7a6 c1d2 h7h5 d2g5 b7b5 a4d1 f8e7 e1g1 e8g8 a2a4 f8d8 a4b5 a6b5 b1c3 c8b8 g5f4 b8b6 f4g5 f6d5 g5e7 d7e7 e2e3 d5c3 b2c3 g7g6 d1c2 d8c8 d3d4 b5b4 c3b4 c5b4');
INSERT INTO evaluations VALUES(47268,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(47268,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(47268,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(47268,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(47268,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(47268,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(47268,36,'cp 0','g2g3 c7c5 c2c4 d5c4 d1a4 c8d7 a4c4 e7e6 f3e5 b8c6 e5d7 d8d7 f1g2 f8e7 c4a4 f6d5 b1c3 e8g8 e1g1 f8d8 c3d5 e6d5 c1e3 a8b8 a1c1 b7b6 e3f4 b8c8 f4d2 c6d4 a4d1 d4c6');
INSERT INTO evaluations VALUES(47270,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47270,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(47270,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(47270,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(47270,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(47270,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(47270,36,'cp -15','c7c5 c2c4 d5c4 d1a4 c8d7 a4c4 e7e6 f3e5 b8c6 e5d7 d8d7 f1g2 f8e7 c4a4 f6d5 b1c3 e8g8 e1g1 f8d8 c1d2 a7a6 a4d1 b7b5 a1c1 a8c8 c3d5 e6d5 e2e3 g7g6 b2b3 e7f8 d1e2 a6a5 h2h4 h7h5 f1d1 c6b4 d2b4 a5b4 c1c2 d7e6');
INSERT INTO evaluations VALUES(45400,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(45400,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(45400,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(45400,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(45400,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(45400,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45400,36,'cp -7','f1g2 b8c6 c1f4 e7e6 e1g1 f8d6 f4d6 d8d6 e2e4 d5e4 d3e4 d6d1 f1d1 f6e4 f3d2 e4d2 b1d2 e8e7 g2c6 b7c6 d2e4 a8b8 b2b3 e6e5 e4c5 c8f5 c2c3 h8d8 b3b4 d8d5 a2a3 b8d8 d1e1 f7f6 f2f4 a7a5 f4e5 f6e5 c3c4 d5d2 e1e5 e7f6');
INSERT INTO evaluations VALUES(45401,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45401,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45401,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45401,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(45401,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(45401,36,'cp 0','b8c6 c1f4 e7e6 e1g1 f8d6 f4d6 d8d6 e2e4 e8g8 f1e1 d6c7 b1c3 f8d8 e4d5 e6d5 d3d4 c5d4 f3d4 c8g4 d1d3 a8c8 c3e2 c7b6 d4c6 b7c6 b2b3 c6c5 h2h3 g4d7 e2f4 d7b5 d3f5 b5d7');
INSERT INTO evaluations VALUES(13118,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13118,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13118,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(13118,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(13118,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(13118,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(13118,36,'cp -10','c1f4 e7e6 e1g1 f8d6 f4d6 d8d6 e2e4 d5e4 d3e4 f6e4 f3d2 e4f6 b1c3 e8g8 d2e4 f6e4 c3e4 d6e5 d1d6 e5b2 d6c5 a8b8 f1d1 e6e5 d1b1 b2d4 c2c3 d4d8 e4d6 d8c7 g2c6 b7c6 b1b8 c7b8 c5e5 c8h3');
INSERT INTO evaluations VALUES(41992,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(41992,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(41992,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41992,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(41992,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(41992,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(41992,36,'cp 45','e2e3 e8g8 c4d5 e6d5 f1d3 h7h6 g5h4 b4e7 g1e2 f8e8 e1g1 c7c6 d1c2 b8d7 h2h3 d7f8 h4g3 e7d6 a1c1 d6g3 e2g3 h6h5 b2b4 a7a6 c3a4 d8e7 a2a3 g7g6 g3e2 f8e6 f1e1 e6g7 e2f4 c8f5 a4c5 f5d3 f4d3 a8c8');
INSERT INTO evaluations VALUES(41990,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(41990,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(41990,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41990,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41990,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41990,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(41990,36,'cp 39','h7h6 g5h4 b8d7 c4d5 e6d5 f1d3 e8g8 g1e2 c7c6 e1g1 f8e8 d1c2 b4d6 h2h3 a7a5 a1e1 d7f8 b2b3 d6b4 e1c1 c8d7 g1h1 b7b5 h4g3 a8c8 g3h2 f8e6 f2f3 c6c5 c3b5 c5d4');
INSERT INTO evaluations VALUES(41641,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(41641,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(41641,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(41641,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(41641,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41641,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(41641,36,'cp 52','g5h4 e8g8 c4d5 e6d5 f1d3 c7c6 g1e2 b8d7 e1g1 f8e8 d1c2 a7a5 h2h3 d7f8 a2a3 b4d6 c3b1 c8d7 b1d2 a8c8 d2f3 d8e7 f1e1 b7b6 a1d1 b6b5 h4g3 f8e6 g3d6 e7d6 e2c3 b5b4 a3b4 a5b4 c3a4 c6c5 a4c5 e6c5 d4c5 c8c5 c2d2');
INSERT INTO evaluations VALUES(41647,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(41647,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(41647,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(41647,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(41647,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(41647,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(41647,36,'cp 49','e8g8 c4d5 e6d5 f1d3 c7c6 g1e2 b8d7 e1g1 f8e8 d1c2 b4e7 a2a3 a7a6 h2h3 d7f8 h4g3 e7d6 b2b4 d6g3 e2g3 c8d7 f1c1 h6h5 a3a4 h5h4 g3e2 d8e7 b4b5 a6b5 a4b5 f8e6 b5c6 b7c6 a1b1 c6c5 d4c5');
INSERT INTO evaluations VALUES(54046,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(54046,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(54046,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(54046,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(54046,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(54046,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(54046,36,'cp -35','b8c6 c2c3 d8a5 b5c6 b7c6 d4c5 a5c5 g1f3 c5b6 e1g1 g8e7 b2b4 e7f5 b1a3 a7a5 a3c2 c8a6 f1e1 a5b4 c3b4 a6b5 g2g4 f5e7 c1e3 b6b7 c2d4 e7g6 a2a3 b7d7 d4b5 c6b5 g4g5 a8c8 h2h4 h7h5 g5h6 g7h6 h4h5 g6e7 g1h2 h8g8 e1g1 g8g1 h2g1 e7f5');
INSERT INTO evaluations VALUES(20197,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(20197,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(20197,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(20197,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(20197,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(20197,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(20197,36,'cp -37','g1f3 c5d4 f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 b1d2 g8e7 d2f3 e7f5 c2c3 f8c5 b2b4 c5e7 a2a4 e8g8 e1g1 c6c5 b4b5 a7a6 g2g4 f5h4 f3h4 e7h4 b5b6 d8b6 g4g5 b6a5 c1d2 d5d4 c3d4 a5d8 d4c5 h4g5 d2e3 g5e3 d1d8 f8d8 f2e3 d8d3 c5c6 a8c8 f1c1 d3e3');
INSERT INTO evaluations VALUES(10136,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10136,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10136,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10136,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10136,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10136,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(10136,36,'cp 1','d7d6 d2d3 a7a6 e1g1 c6d4 h2h3 c8e6 f3d4 c5d4 c3e2 d4b6 e2g3 e8g8 c4e6 f7e6 g3h5 h7h6 c1e3 b6e3 f2e3 d8e7 h5f6 f8f6 f1f6 e7f6 d1e2 a8f8 a1f1 f6f1 e2f1 f8f1 g1f1 g8f7 c2c4 c7c5 f1f2 a6a5 f2g1 f7f6 a3a4 f6f7 g1f1 g7g5 f1e2');
INSERT INTO evaluations VALUES(9359,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(9359,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(9359,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9359,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9359,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9359,36,'cp 0','d2d3 a7a6 e1g1 c6d4 h2h3 c8e6 f3d4 c5d4 c3e2 d4b6 e2g3 d8e7 c4e6 f7e6 a3a4 a6a5 g3h5 f6d7 d1g4 e8g8 c1h6 f8f7 h6g5 e7e8 g5e3 b6e3 f2e3 f7f1 a1f1 e8g6 g4h4 a8f8 h5g3 d7f6 b2b3 g6e8 f1f3 e8g6');
INSERT INTO evaluations VALUES(9361,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(9361,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9361,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9361,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9361,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9361,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9361,36,'cp 0','a7a6 h2h3 e8g8 c1g5 h7h6 g5h4 c6d4 c3d5 g7g5 f3g5 f6d5 d1h5 d8f6 g5e6 d4c2 e1d1 c2a1 c4d5 f7e6 h4f6 f8f6 h5e8 g8g7 d5a2 f6f2 e8e7 g7h8 e7e8 h8h7 h3h4 a1c2 e8e7 h7h8 e7e8');
INSERT INTO evaluations VALUES(7177,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7177,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7177,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7177,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7177,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7177,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7177,36,'cp 0','h2h3 e8g8 c1g5 h7h6 g5h4 c6d4 c3d5 g7g5 f3g5 f6d5 d1h5 d8f6 g5e6 d4c2 e1d1 c2a1 c4d5 f6h4 h5h4 f7e6 h4h6 f8f2 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(7176,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7176,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7176,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7176,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7176,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7176,36,'cp 0','e8g8 c1g5 h7h6 g5h4 c6d4 c3d5 g7g5 f3g5 f6d5 d1h5 d8f6 g5e6 d4c2 e1d1 c2a1 c4d5 f6h4 h5h4 f7e6 h4h6 f8f2 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(2615,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2615,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2615,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2615,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2615,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2615,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2615,36,'cp 0','c1g5 h7h6 g5h4 c6d4 c3d5 g7g5 f3g5 f6d5 d1h5 d8f6 g5e6 d4c2 e1d1 c2a1 c4d5 f6h4 h5h4 f7e6 h4h6 f8f2 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(2608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2608,36,'cp 0','h7h6 g5h4 c8e6 c3d5 g7g5 f3g5 f6d5 e4d5 h6g5 d5e6 g5h4 e6f7 f8f7 c4f7 g8f7 d1h5 f7g7 h5g4 g7h8 g4h5 h8g7');
INSERT INTO evaluations VALUES(2589,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(2589,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2589,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2589,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2589,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2589,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2589,36,'cp 0','g5h4 c6d4 c3d5 g7g5 f3g5 f6d5 d1h5 d8f6 g5e6 d4c2 e1d1 c2a1 c4d5 f7e6 h4f6 f8f6 h5e8 g8g7 d5a2 f6f8 e8e7 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(55315,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55315,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(55315,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(55315,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(55315,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(55315,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(55315,36,'cp -1','b8d7 g1f3 c7c6 a4c4 e7e5 f3e5 d7e5 d4e5 d8a5 b1c3 a5e5 g2g3 c8e6 c4a4 f8c5 f1g2 g8f6 c1f4 e5d4 e1g1 e8g8 a4c2 d4b4 f4d2 b4c4 b2b3 c4g4 c3e4 c5d4 a1e1 f6e4 g2f3 g4f5 f3e4 f5h5 d2c3 d4c3 c2c3 f8d8 e4f3 h5g5 h2h4 g5h6');
INSERT INTO evaluations VALUES(43229,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(43229,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(43229,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(43229,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(43229,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(43229,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(43229,36,'cp 34','e2e4 d7d5 e4e5 f6e4 f1d3 e4c3 b2c3 c7c5 e1g1 b8c6 d3b5 d8c7 a2a4 e8g8 c1a3 b7b6 h2h3 c8f5 f1e1 f8d8 d1e2 c6a5 a1d1 e7e6 a3c1 a7a6 b5d3 f5d3 c2d3 c5d4 c3d4 d8c8 c1g5 c7d7 d1b1 h7h6 g5e3');
INSERT INTO evaluations VALUES(57582,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(57582,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(57582,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(57582,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(57582,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(57582,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57582,36,'cp 19','g8f6 c1g5 d7d5 g5f6 e7f6 e2e3 c8e6 d1f3 g7g6 h2h3 h7h5 f1d3 e6f7 g1e2 b8c6 a2a3 d8d7 h3h4 c6e7 b2b4 e7c8 e1g1 c7c6 c3a4 c8d6 a4c5 d7c8 b4b5 d6e4 b5c6 e4c5 d4c5 b7c6 e2d4 f8c5 f3g3 c5d4 e3d4 e8g8 f1b1 g8g7 a3a4');
INSERT INTO evaluations VALUES(50099,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50099,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(50099,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(50099,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(50099,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(50099,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(50099,36,'cp 36','c1g5 d7d5 g5f6 e7f6 h2h4 c8e6 e2e3 g7g6 f1d3 b8c6 g1e2 d8d7 a2a3 e6f7 b2b4 c6e7 d1b1 e7c8 a3a4 f8e7 b4b5 e8g8 a4a5 c8d6 h4h5 g6g5 a5a6 b7b6 b1a2 d6e4 h5h6 g8h8 f2f3 e4c3 e2c3 f5f4 e3f4 g5f4');
INSERT INTO evaluations VALUES(50116,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(50116,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(50116,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(50116,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(50116,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(50116,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(50116,36,'cp 36','d7d5 e2e3 e7e6 g1f3 b8d7 f1e2 f8e7 e1g1 e8g8 g5f4 b7b6 a1c1 f6e8 h2h3 g7g5 f4h2 e7d6 c3b1 d6h2 f3h2 g8h8 c2c4 d8e7 d1c2 c7c5 c4d5 e6d5 b2b3 c8b7 b1c3 e8g7 d4c5 b6c5 h2f3');
INSERT INTO evaluations VALUES(7335,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(7335,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(7335,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7335,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(7335,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(7335,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(7335,36,'cp 20','c2c3 b7b5 a4b3 d7d6 d2d4 c5b6 a2a4 c8b7 c1g5 h7h6 g5f6 d8f6 a4b5 a6b5 a1a8 b7a8 d4d5 c6a7 b1d2 e8g8 d1e2 a8b7 f1a1 b7c8 b3c2 g7g6 c2d3 c8d7 d2f1 h6h5 g2g3 c7c6 d5c6 d7c6 f1e3 a7c8 f3d2 b6e3 e2e3 g8g7 b2b4 h5h4');
INSERT INTO evaluations VALUES(7328,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7328,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(7328,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7328,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7328,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(7328,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(7328,36,'cp 23','b7b5 a4c2 d7d5 f3e5 c6e5 d2d4 d5e4 d4c5 e8g8 d1d8 f8d8 a2a4 b5a4 c1f4 e5c4 c2a4 c8d7 b2b3 c4b2 a4d7 d8d7 c5c6 d7e7 b3b4 b2d3 f4g3 f6h5 f1d1 h5g3 h2g3 f7f6 f2f3 d3b2 d1d4 e4f3 g2f3');
INSERT INTO evaluations VALUES(7463,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(7463,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(7463,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(7463,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(7463,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(7463,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(7463,36,'cp 35','a4b3 d7d6 a2a4 c8b7 d2d3 h7h6 f3h4 e8g8 a4b5 a6b5 a1a8 b7a8 d1f3 g8h7 g2g4 g7g6 f3h3 f6g8 g4g5 a8b7 g5h6 b7c8 h3g3 d8d7 b1a3 b5b4 a3c2 b4c3 b2c3 d7h3 b3a4 h3g3 h2g3 c8d7 c1e3 c5e3 c2e3 h7h6 a4c6');
INSERT INTO evaluations VALUES(7457,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(7457,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(7457,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(7457,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7457,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(7457,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(7457,36,'cp 20','d7d6 a2a4 c8b7 d2d3 h7h6 f3h4 e8g8 a4b5 a6b5 a1a8 b7a8 d1f3 g8h7 h4f5 f6g8 b1d2 b5b4 f5g3 c6e7 f3e2 b4c3 b2c3 e7g6 d2f3 c5b6 d3d4 d8e7 b3c2 a8b7 c1e3 f8a8 f1b1 g8f6 e2d1 a8a3 g3f5 e7f8 d4e5 d6e5 e3b6 c7b6');
INSERT INTO evaluations VALUES(7418,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(7418,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(7418,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(7418,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(7418,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(7418,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(7418,36,'cp 21','a2a4 c8b7 d2d3 h7h6 f3h4 e8g8 a4b5 a6b5 a1a8 b7a8 d1f3 g8h7 h4f5 f6g8 c1e3 c5e3 f3e3 c6e7 f5g3 g8f6 b1d2 c7c5 d3d4 e7g6 f1a1 a8b7 h2h3 d8c7 d4d5 f8a8 a1a8 b7a8 b3d1 c5c4 d1e2 g6f4 e2f1 a8b7 e3a7 g7g6 d2b1 f4d3 f1d3 c4d3');
INSERT INTO evaluations VALUES(39614,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(39614,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(39614,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(39614,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(39614,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39614,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(39614,36,'cp 44','c7c5 d2d4 c5d4 f3d4 b8c6 c1e3 g8f6 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1d1 b5c4 d4e3 h6e3 f2e3 c4b3 a2b3 f8b8 d1d3 b8b5 e3e4 b5b4 e4e5 b4e4 d5d6 e4e5 d6d7 a8d8 a1a7 e5e2 a7c7 e2b2');
INSERT INTO evaluations VALUES(45123,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(45123,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(45123,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(45123,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(45123,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(45123,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(45123,36,'cp -62','e7e6 d2d3 d6d5 c4b5 c8d7 b5d7 d8d7 e4e5 f6g8 e1g1 b8c6 d1e2 g8e7 c2c3 h7h6 g5f3 g7g5 h2h3 h8g8 b1a3 e8c8 d3d4 g5g4 h3g4 g8g4 a3c2 e7f5 c2e3 f5e3 c1e3 f8e7 f3h2 g4g7 g2g3 d8h8 g1g2 h6h5 f1h1 c5d4 e3d4 c6d4 c3d4 c8b8 a1c1 h5h4 h2g4');
INSERT INTO evaluations VALUES(45976,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(45976,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(45976,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(45976,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(45976,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(45976,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(45976,36,'cp -62','d2d3 d6d5 c4b5 c8d7 b5d7 d8d7 b1c3 b8c6 c1f4 e6e5 f4g3 d5d4 c3b1 h7h5 g5f3 d7e6 b1d2 f6d7 h2h3 f8e7 h3h4 f7f6 a2a4 g7g5 d2c4 g5g4 f3d2 c6b4 e1g1 e8c8 f2f3 g4f3 f1f3 d8g8 c2c3 b4c6 g3f2 g8g4 c3d4');
INSERT INTO evaluations VALUES(45973,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(45973,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(45973,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(45973,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(45973,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(45973,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(45973,36,'cp -68','d6d5 c4b5 c8d7 b5d7 d8d7 d1e2 b8c6 g5f3 h7h6 e4e5 f6g8 c2c3 g8e7 h2h4 f7f6 b1a3 e8c8 a3c2 c6e5 f3e5 f6e5 e2e5 e7c6 e5e2 e6e5 c1d2 f8d6 e1c1 d6c7 c1b1 h8e8 h4h5 c8b8 f2f3 b8a8 c2e3 d5d4 e3g4 b7b5 h1e1 d4c3 d2c3 b5b4');
INSERT INTO evaluations VALUES(6220,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(6220,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(6220,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(6220,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6220,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6220,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(6220,36,'cp 12','g8f6 c2c3 c5b6 b1d2 d7d6 f1e1 e8g8 c4b3 a7a5 d2f1 c8e6 c1e3 b6e3 f1e3 c6e7 b3c2 e7g6 d3d4 a5a4 h2h3 b7b5 d1d2 c7c5 a2a3 a8b8 a1d1 f8e8 d4e5 d6e5 d2d6 d8b6 d6b6 b8b6 f3d2');
INSERT INTO evaluations VALUES(43201,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43201,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(43201,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(43201,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(43201,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(43201,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(43201,36,'cp 56','e2e4 c7c5 d4c5 d8a5 d1d2 a5c5 f2f3 b8c6 e1c1 d7d6 g1e2 c8e6 c1b1 a8c8 f4e3 c5a5 e2f4 f6d7 c3d5 a5d2 d1d2 h7h5 f1e2 a7a6 c2c4 g7h6 b2b3 c6e5 h1e1 e6d5 e4d5 e8g8 g2g3 b7b5 c4b5 a6b5');
INSERT INTO evaluations VALUES(43268,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(43268,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(43268,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(43268,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(43268,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(43268,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(43268,36,'cp 70','c7c5 d4c5 d8a5 d1d2 a5c5 e1c1 d7d6 e4e5 f6h5 f4e3 c5e5 f1b5 b8c6 g1f3 e5f5 c3d5 c8e6 b5c6 b7c6 d5c7 e8d7 c7a8 h8a8 h1e1 f5d5 d2b4 d5b5 b4b5 c6b5 f3d4 g7d4 d1d4 a7a5 f2f3 h5g7 b2b3 h7h5 c1b2 f7f6 e3f2 e6f7 e1d1 g7e6 d4d2 g6g5 f2b6 a5a4 b6f2 e6f4 g2g3 f4e6');
INSERT INTO evaluations VALUES(46050,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(46050,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(46050,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(46050,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(46050,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(46050,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(46050,36,'cp 82','g2g4 h7h6 c1e3 b8c6 h1g1 f8e7 d4b3 d6d5 e4d5 f6d5 c3d5 e6d5 d1d2 e8g8 e1c1 c6e5 b3d4 f8e8 h2h3 c8d7 g4g5 h6g5 e3g5 a8c8 f2f4 e5c6 g1g3 c6d4 d2d4 d8c7 f1d3 c7c5 d4c5 c8c5 g5e7 e8e7 f4f5 e7e5 d1g1');
INSERT INTO evaluations VALUES(46046,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(46046,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(46046,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(46046,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(46046,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(46046,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(46046,36,'cp 52','h7h6 c1e3 b8c6 h2h3 d6d5 f1b5 c8d7 e4d5 e6d5 a2a3 f8d6 d1d3 e8g8 e1c1 c6e5 d3d2 d7b5 d4b5 e5c4 d2d3 c4e3 d3e3 f8e8 e3f3 e8e5 h3h4 a7a6 b5d6 d8d6 f3f4 a8d8 d1d4 b7b5 h1d1 d6e7 f2f3 d8e8 d4d3 e5e1 f4d4');
INSERT INTO evaluations VALUES(47463,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(47463,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(47463,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(47463,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(47463,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(47463,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(47463,36,'cp 90','d2d4 c8f5 f1d3 f5d3 d1d3 e7e6 c2c4 d5f6 e1g1 c7c6 b1c3 f8e7 c1f4 e8g8 h2h3 f8e8 a1b1 b8d7 b2b4 d7f8 f4e5 f8g6 f1d1 d8d7 a2a4 a8d8 c4c5 f6d5 c3e4 d5f4 d3c2 f7f6 e5f4 g6f4 b1b3');
INSERT INTO evaluations VALUES(15176,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(15176,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(15176,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(15176,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(15176,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(15176,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(15176,36,'cp -50','e2e3 c8f5 d2d3 d8d7 a2a3 e8c8 b1d2 f8d6 h2h3 h7h5 g1f3 h8e8 f3g5 c8b8 b3b4 e5e4 d3e4 f6e4 g5e4 f5e4 d2e4 d5e4 d1e2 h5h4 g3g4 d7e6 a1d1 g7g6 e1g1 f7f5 c2c4 a7a6 d1d5 c6e7 d5d1 d6e5 b2e5 e6e5 d1d8 e8d8 f1d1 d8d1 e2d1 f5g4 h3g4');
INSERT INTO evaluations VALUES(15171,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(15171,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(15171,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(15171,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(15171,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(15171,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(15171,36,'cp -61','c8f5 d2d3 f8d6 b1c3 d8d7 a2a3 e8c8 d1d2 d5d4 e3d4 e5d4 c3e2 h8e8 e1c1 d6c5 g1f3 a7a5 h1e1 c8b8 c1b1 h7h6 e2c1 d7d6 d2f4 d6f4 g3f4 f5d7 c1e2 f6g4 e2d4 c6d4 b2d4 c5d4 f3d4 g4f2 d1c1 f2g4 e1e8 d8e8 g2h3 f7f5 b3b4 g4h2 h3f5 d7f5');
INSERT INTO evaluations VALUES(25254,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(25254,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(25254,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(25254,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(25254,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(25254,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(25254,36,'cp -71','d2d3 d8d7 b1d2 e8c8 a2a3 f8d6 h2h3 h8e8 b3b4 c8b8 d2b3 h7h5 d1d2 d5d4 g1e2 d4e3 f2e3 f5e6 e1c1 e6b3 c2b3 d7e6 g2c6 b7c6 e3e4 e6b3 d2c2 b3c2 c1c2 f6d7 g3g4 h5g4 h3g4 f7f6 e2g3 c6c5 g3f5 c5b4 a3b4 d6b4 h1h7 e8g8 h7g7 g8g7 f5g7');
INSERT INTO evaluations VALUES(35046,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(35046,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(35046,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(35046,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(35046,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(35046,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(35046,36,'cp 166','d7d6 f1d3 h4d8 g1f3 a7a6 d1e2 f8e7 e4e5 d6d5 c1e3 b7b6 c3d1 c7c5 c2c3 b8c6 e1g1 h7h5 f1e1 g8h6 e2d2 c8b7 h2h3 d8c7 e3f4 e8c8 d1e3 c8b8 f3g5 d8f8 d2e2 b8a8 e2h5');
INSERT INTO evaluations VALUES(50340,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50340,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(50340,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(50340,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50340,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50340,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50340,36,'cp 0','d7d5 g1f3 c7c5 c2c4 d5c4 d1a4 d8d7 a4c4 c5d4 c4d4 b8c6 d4d7 c8d7 f1g2 e7e5 e1g1 h7h6 f1d1 f8c5 b1c3 e8c8 c3a4 d7e6 c1d2 c5d6 f3e1 c6d4 a4c3 c8b8 e1f3 d4c6 f3e1');
INSERT INTO evaluations VALUES(52459,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(52459,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(52459,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(52459,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(52459,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(52459,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(52459,36,'cp -17','e2e4 e7e6 g1f3 d7d5 d2d3 f8e7 f1e2 g8h6 b1d2 b8c6 d2f1 b7b5 f1g3 e8g8 e1g1 a7a5 c2c3 d5d4 c3d4 c6d4 g1h1 f7f5 f3d4 d8d4 e4f5 h6f5 e2f3 a8b8 g3f5 e6f5 f1e1 e7d6 d1e2 c8b7 f3b7 b8b7');
INSERT INTO evaluations VALUES(21009,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21009,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(21009,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21009,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(21009,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(21009,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(21009,36,'cp 31','g1f3 g8f6 c1f4 c8f5 b1d2 h7h6 d1b3 a8b8 e2e3 e7e6 c3c4 f8d6 c4d5 e6d5 f4d6 c7d6 f1e2 e8g8 e1g1 c6a5 b3a3 b7b5 b2b3 b5b4 a3b2 a5c6 f1c1 d8b6 d2f1 f8c8 h2h3 a7a5 f1g3 f5h7 a2a3 b4a3 a1a3 g8h8');
INSERT INTO evaluations VALUES(21004,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(21004,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(21004,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(21004,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(21004,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(21004,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(21004,36,'cp 36','g8f6 c1f4 e7e6 e2e3 f8d6 b1d2 e8g8 f4g3 b7b6 f1d3 c8b7 e1g1 c6e7 b2b4 c7c5 b4c5 b6c5 a1b1 b7c6 d4c5 d6c5 f3d4 c6d7 h2h3 c5b6 d1b3 b6c7 g3c7 d8c7 b3a3 e7c6 d4c6 d7c6 f1c1 h7h6 c3c4 f8c8 c4d5 f6d5 d3f1 c7e7 a3e7 d5e7');
INSERT INTO evaluations VALUES(14279,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(14279,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(14279,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(14279,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(14279,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(14279,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(14279,36,'cp 30','c1f4 c8f5 b1d2 e7e6 f3h4 f8d6 h4f5 e6f5 f4g5 h7h6 g5f6 d8f6 g2g3 c6e7 f1g2 c7c6 e2e3 f6e6 e1g1 e8g8 a2a4 a7a5 d1b3 g7g6 f1c1 f8b8 c3c4 d5c4 b3c4 e7d5 c4b3 d5b4 b3e6 f7e6 d2c4 d6c7 b2b3 g8g7 g2f3');
INSERT INTO evaluations VALUES(14221,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14221,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(14221,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(14221,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(14221,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(14221,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(14221,36,'cp 30','c8f5 b1d2 e7e6 f3h4 f8d6 h4f5 e6f5 f4g5 h7h6 g5f6 d8f6 g2g3 e8g8 f1g2 c6e7 e1g1 c7c6 e2e3 f6e6 a1c1 b7b5 b2b3 a8b8 d2f3 f8c8 c1c2 g7g6 d1e2 a7a6 c3c4 g8g7 f3e1 e7g8 c4d5 c6d5 e1d3 c8c2 e2c2 b8c8 d3c5');
INSERT INTO evaluations VALUES(24820,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(24820,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24820,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24820,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(24820,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(24820,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24820,36,'cp 31','b1d2 e7e6 f3h4 f8d6 h4f5 e6f5 f4g5 h7h6 g5f6 d8f6 g2g3 c6e7 f1g2 e8g8 e1g1 c7c6 e2e3 f6e6 a1c1 b7b5 b2b3 a8b8 d2f3 f8c8 c1c2 g7g6 d1e2 a7a5 c3c4 g8g7 c4b5 c6b5 f1c1 b5b4 f3e1 c8c2 c1c2 b8c8 h2h4 h6h5 c2c8 e6c8');
INSERT INTO evaluations VALUES(24819,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(24819,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(24819,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24819,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(24819,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(24819,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24819,36,'cp 21','e7e6 f3h4 f8d6 h4f5 e6f5 f4g5 h7h6 g5f6 d8f6 e2e3 c6e7 g2g3 c7c6 f1g2 e8g8 e1g1 f6e6 d1a4 a7a5 c3c4 d6b4 c4d5 e7d5 d2b1 f8d8 a2a3 b4e7 a4b3 h6h5 f1c1 h5h4 b1c3 d5c3 c1c3 e6b3 c3b3 a8a7 b3c3 g7g6 a1c1 g8g7');
INSERT INTO evaluations VALUES(25157,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25157,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25157,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(25157,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(25157,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(25157,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(25157,36,'cp 28','f3h4 f8d6 h4f5 e6f5 f4g5 h7h6 g5f6 d8f6 g2g3 c6e7 f1g2 c7c6 e2e3 e8g8 e1g1 g7g6 c3c4 d5c4 d2c4 f8d8 a1b1 f6e6 d1b3 a8b8 c4d2 g6g5 b3e6 f7e6 d2c4 g8g7 b2b4 e7d5 a2a3 h6h5 f1c1 d6e7 b1b3 h5h4 g3h4 g5h4 b4b5');
INSERT INTO evaluations VALUES(52850,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52850,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52850,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52850,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(52850,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(52850,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(52850,36,'cp -11','c8f5 d2d4 e7e6 b1c3 c6c5 g1f3 h7h5 c1e3 c5c4 f1e2 f8b4 e1g1 b4c3 b2c3 d8a5 e3d2 g8e7 a2a4 b8c6 h2h3 f5g6 f3h4 g6e4 d2e1 g7g6 e2f3 e4f3 d1f3 a5d8 f3e2 c6a5 e1f2 b7b6 h4f3 d8d7 f2h4');
INSERT INTO evaluations VALUES(31174,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(31174,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(31174,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(31174,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(31174,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(31174,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31174,36,'cp -8','b1c3 e7e6 d2d4 c6c5 g1f3 b8c6 c1e3 c5c4 f1e2 f8b4 e1g1 b4c3 b2c3 d8a5 d1d2 g8e7 a2a4 h7h6 f3h4 f5e4 f1b1 a5c7 e2f3 e8c8 d2e2 h8g8 g2g3 e4f3 h4f3 c8b8 e3c1 b8a8 c1a3 b7b6 g3g4 f7f5 a3d6 c7d7 h2h3 d8c8 g1h1 c6a5');
INSERT INTO evaluations VALUES(5045,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(5045,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(5045,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(5045,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(5045,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(5045,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(5045,36,'cp 9','f1c4 d7d6 h2h3 d8a5 c1d2 g8f6 e1g1 e8g8 c3d5 a5d8 d5f6 g7f6 a1b1 a7a5 a2a4 d6d5 c4d3 c8a6 d3a6 a8a6 d1e2 a6a8 b2b3 f8e8 e4e5 f6g7 e5e6 f7f5 f1e1 d8d6 e2d3 d5d4 b1d1 g7f6 g2g4 f5g4 h3g4 d6d5 f2f4');
INSERT INTO evaluations VALUES(5032,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(5032,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(5032,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(5032,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(5032,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5032,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5032,36,'cp 0','d7d6 h2h3 g8f6 e4e5 d6e5 d1d8 e8d8 c4b3 e7e6 c1d2 d8c7 c3a4 a7a5 e1c1 h7h5 h3h4 c8b7 a4c5 c7b6 h1e1 b6c5 d2e3 c5b5 a2a4 b5b4 e3d2 b4c5 d2e3');
INSERT INTO evaluations VALUES(5062,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5062,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5062,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5062,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5062,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5062,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5062,36,'cp 0','e1g1 g8f6 e4e5 d6e5 d1d8 e8d8 f1e1 c8f5 e1e5 f6d5 e5e2 e7e6 c3a4 d8e7 c1d2 h8d8 c4b3 a8b8 d2a5 d8d7 a5d2');
INSERT INTO evaluations VALUES(5061,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5061,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5061,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5061,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5061,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5061,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5061,36,'cp 0','g8f6 e4e5 d6e5 d1d8 e8d8 f1e1 c8f5 e1e5 f6d5 e5e2 e7e6 c3a4 d8e7 c4b3 h8d8 c1d2 a8b8 d2a5 d8d7 a1e1 g6g5 g2g3 f5g4 a4c5 d7d6 c5e4 d6d7');
INSERT INTO evaluations VALUES(7715,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7715,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7715,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7715,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7715,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7715,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7715,36,'cp 0','e4e5 d6e5 d1d8 e8d8 f1e1 c8f5 e1e5 f6d5 e5e2 e7e6 c3a4 d8e7 c1d2 a8b8 a1e1 h8d8 c4b3 h7h6 h2h4 e7f8 f2f3 h6h5 d2g5 g7f6 g2g3 f6g5 h4g5 f8e7 e2e5 d5b4 e5e2');
INSERT INTO evaluations VALUES(20843,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(20843,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(20843,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(20843,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(20843,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(20843,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(20843,36,'cp 60','g7g5 f4g3 h7h5 h2h3 c8f5 a2a3 e7e6 c2c4 f8d6 g3d6 d8d6 b2b4 g8h6 b1c3 c6e7 b4b5 c7c6 a3a4 e8f7 c4c5 d6d8 a4a5 a7a6 b5b6 d8d7 f1d3 f7g7 d1c2 h6f7 e1c1 h5h4 c1b2 e6e5 d3f5 d7f5');
INSERT INTO evaluations VALUES(20902,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(20902,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(20902,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(20902,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(20902,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(20902,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(20902,36,'cp 85','f4g3 h7h5 h2h3 h5h4 g3h2 g5g4 h3g4 c8g4 d1d3 d8d7 d3g6 e8d8 f3h4 e7e5 d4e5 d7e6 h2g3 c6e5 g3e5 f6e5 g6e6 g4e6 g2g3 f8e7 b1d2 c7c6 f1h3 e6f7 e1c1 d8c7 h4f5 e7f6 f2f4 e5f4 e3f4 g8e7 f5e7 f6e7 h3f5 a8g8 h1h8 g8h8 d2f3 e7d6 d1e1');
INSERT INTO evaluations VALUES(20903,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(20903,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(20903,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(20903,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(20903,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(20903,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(20903,36,'cp 74','h7h5 h2h3 c8f5 c2c4 e7e6 a2a3 h5h4 g3h2 f8d6 b1d2 c6e7 d1b3 b7b6 h2d6 d8d6 f1e2 g8h6 b3c3 a7a5 b2b4 e8g8 e1g1 g8h8 f3h2 d6d7 f1c1 a5b4 a3b4 c7c6 h2g4 a8a1 c1a1 h6g4 h3g4 f5g6 c3a3 d5c4 d2c4 e7d5');
INSERT INTO evaluations VALUES(20905,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(20905,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(20905,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(20905,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(20905,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(20905,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(20905,36,'cp 94','h2h3 c8f5 c2c4 e7e6 b1c3 f8d6 g3d6 d8d6 a2a3 c6e7 f3d2 c7c6 d1b3 d6d7 f1e2 f5g6 h3h4 g5g4 c4d5 e6d5 g2g3 e7f5 e2f1 f5g7 c3e2 a7a5 f1g2 g8e7 b3c3 a5a4 e2f4 e8g8 b2b3 g6f7 f4d3 g7e6 e1g1 f7g6 f1c1 b7b6 c3c2 g6d3 c2d3 f6f5 b3a4');
INSERT INTO evaluations VALUES(44929,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(44929,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(44929,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44929,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44929,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44929,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(44929,36,'cp 41','g1f3 e7e6 b1d2 f8d6 f4g3 e8g8 f1d3 c6c5 c2c3 d8b6 d4c5 d6c5 d1c2 c5d6 e3e4 d6g3 h2g3 h7h6 e1g1 d5e4 d2e4 b8d7 f1e1 f6e4 d3e4 d7f6 f3e5 c8d7 e4f3 d7c6 f3c6 b7c6 a2a4 c6c5 e5c4 b6c7 a1d1 f8d8 a4a5 d8d1 e1d1 a8d8');
INSERT INTO evaluations VALUES(44931,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44931,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44931,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44931,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(44931,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44931,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(44931,36,'cp 35','e7e6 f1d3 f8d6 f4g3 c6c5 d4c5 d6c5 e1g1 e8g8 b1d2 f6h5 g3h4 c5e7 h4e7 d8e7 c2c4 h5f6 a1c1 c8d7 c4d5 f6d5 d1c2 h7h6 d3h7 g8h8 h7e4 b8c6 e4d5 e6d5 c2c5 e7e6 f3d4 c6d4 c5d4 f8c8 a2a3 e6b6 d2f3 h8g8 f1d1 c8c1 d1c1');
INSERT INTO evaluations VALUES(45672,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(45672,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(45672,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(45672,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(45672,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(45672,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(45672,36,'cp 38','f1d3 f8d6 f4g3 c6c5 d4c5 d6c5 b1d2 e8g8 e1g1 f6h5 g3h4 c5e7 h4e7 d8e7 c2c4 h5f6 c4d5 f6d5 d1c2 h7h6 d3h7 g8h8 h7e4 b8c6 a2a3 f8d8 d2b3 c8d7 e4d5 e6d5 h2h3 a8c8 c2d2 d7f5 a1c1 f5e4 f3d4 e7f6 f2f3 e4h7 g1h1 a7a6 f1d1');
INSERT INTO evaluations VALUES(45669,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(45669,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45669,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(45669,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(45669,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(45669,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(45669,36,'cp 42','f8d6 f4g3 c6c5 d4c5 d6c5 b1d2 e8g8 a1c1 c5e7 c2c4 d8b6 d1c2 b8a6 d3e2 a6b4 c2b1 b4c6 e1g1 f6h5 g3h4 e7h4 f3h4 h5f6 a2a3 a7a5 h4f3 a5a4 c1c3 b6d8 e2d3 d8e7 h2h3 h7h6 f1c1 c8d7 c4d5 f6d5 c3c5');
INSERT INTO evaluations VALUES(40554,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(40554,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(40554,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(40554,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(40554,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(40554,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(40554,36,'cp 43','f4g3 c6c5 d4c5 d6c5 e1g1 e8g8 b1d2 f6h5 g3h4 c5e7 h4e7 d8e7 c2c4 h5f6 c4d5 f6d5 d1c2 h7h6 d3h7 g8h8 h7e4 b8c6 a2a3 f8d8 a1c1 f7f5 e4d5 d8d5 e3e4 f5e4 c2e4 c8d7 d2c4 a8d8 c4e3 d5d6 h2h3 d7e8 f1d1 d6d1 c1d1 d8d1 e3d1 e6e5 b2b4');
INSERT INTO evaluations VALUES(40557,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40557,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(40557,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(40557,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40557,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(40557,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40557,36,'cp 42','c6c5 d4c5 d6c5 b1d2 c5e7 c2c4 e8g8 e1g1 b8c6 a2a3 f6h5 b2b4 h5g3 h2g3 a7a5 d1a4 d5c4 d3c4 c8d7 b4b5 c6a7 f1d1 a7c8 a4c2 c8d6 a3a4 a8c8 f3e5 d7e8 c2b2 c8c5 d2e4 c5e5 b2e5 d6c4 d1d8 c4e5 d8b8');
INSERT INTO evaluations VALUES(1843,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(1843,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(1843,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(1843,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(1843,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(1843,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(1843,36,'cp 44','a7a6 b5a4 f8c5 c2c3 g8e7 f1e1 c5a7 d2d4 h7h6 c1e3 e8g8 f3e5 c6e5 d4e5 f6e5 e3a7 a8a7 b1d2 d7d6 d2f3 e5f4 e4e5 d6d5 h2h3 c7c5 a4c2 c8f5 c2f5 f4f5 f3h4 f5g5 d1g4 g5g4 h3g4 b7b5 f2f4 a7d7 h4f3 e7c6 a1d1 f8d8 g1f2 d5d4 e5e6 d7d5 e6e7 d8e8 c3d4 c6e7 d4c5 d5c5');
INSERT INTO evaluations VALUES(1235,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(1235,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(1235,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(1235,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(1235,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(1235,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(1235,36,'cp 56','b5a4 f8c5 c2c3 g8e7 f1e1 c5a7 d2d4 h7h6 c1e3 e8g8 f3e5 c6e5 d4e5 f6e5 e3a7 a8a7 b1d2 d7d6 d2f3 e5f4 f3d4 a7a8 a4c2 c8d7 g2g3 f4g4 f2f3 g4h5 f3f4 h5d1 a1d1 f8e8 a2a4 g8f8 a4a5 e7c6 d4c6 d7c6 g1f2 g7g6 b2b4 a8d8');
INSERT INTO evaluations VALUES(1239,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(1239,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(1239,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(1239,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(1239,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(1239,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(1239,36,'cp 44','f8c5 c2c3 g8e7 f1e1 h7h6 d2d4 c5a7 c1e3 e8g8 d4e5 c6e5 f3e5 f6e5 e3a7 a8a7 b1d2 d7d6 d2f3 e5f4 e4e5 d6d5 a4b3 c8e6 h2h3 c7c5 b3c2 e6f5 c2f5 f4f5 f3h4 f5g5 g2g3 f7f5 h4f3 g5h5 g1h2 a6a5 a2a4 a7a6 b2b4 a5b4 c3b4 c5b4');
INSERT INTO evaluations VALUES(91,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(91,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(91,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(91,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(91,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(91,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(91,36,'cp 69','c2c3 g8e7 f1e1 c5a7 d2d4 h7h6 c1e3 e8g8 f3e5 c6e5 d4e5 f6e5 e3a7 a8a7 b1d2 d7d6 d2f3 e5h5 e4e5 d6d5 d1d4 b7b6 d4b4 e7g6 a4c6 c8b7 c6b7 a7b7 a1d1 f8d8 b4a4 c7c5 a4a6 b7b8 a6f1 g6f4 e1e3 f4e6 h2h3 d5d4 c3d4 e6d4 f3d4 d8d4 d1d3 h5f5 d3d4 c5d4');
INSERT INTO evaluations VALUES(90,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(90,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(90,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(90,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(90,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(90,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(90,36,'cp 66','g8e7 f1e1 c5a7 d2d4 h7h6 c1e3 e8g8 f3e5 c6e5 d4e5 f6e5 e3a7 a8a7 b1d2 d7d6 d2f3 e5f4 e4e5 d6d5 h2h3 a7a8 a4b3 c7c6 g2g3 f4f5 g3g4 f5f4 f3h4 h6h5 h4g2 f4h6 f2f4 g7g6 d1f3 g8g7 g1f2 f8h8 f3g3 h5g4 h3g4 h6h3 b3d1 a8b8 d1f3 h3g3 f2g3 c8e6');
INSERT INTO evaluations VALUES(5970,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5970,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5970,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5970,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5970,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5970,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5970,36,'cp 0','c6a5 c4b3 a5b3 a2b3 g8e7 d3d4 e5d4 f3d4 e8g8 e1g1 d6d5 c1e3 b7b6 d4f5 e7f5 e3c5 b6c5 e4f5 c8f5 c3d5 f5c2 d1c2 d8d5 a1a5 a8b8 a5a7 f8e8 a7c7 d5b3 c2b3 b8b3 c7c5 b3b2 f1c1 h7h6 c5c8 e8c8 c1c8 g8h7 g2g3 b2e2 c8f8 h7g6 h2h4');
INSERT INTO evaluations VALUES(16692,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16692,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(16692,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16692,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16692,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(16692,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16692,36,'cp 7','e4e5 f6d5 c3d4 d7d6 f1b5 d6e5 f3e5 f8b4 c1d2 e8g8 b5c6 b7c6 d2b4 d5b4 e1g1 c8f5 d1f3 f5e6 f3c3 b4a2 c3c5 d8d6 c5d6 c7d6 e5c6 f8e8 b1c3 a2c3 b2c3 a8c8 c6a7 c8c3 a7b5 c3c6 f1c1 c6b6 b5c7 e8c8 c7d5 c8c1 a1c1 b6b7 d5e3 g8f8 f2f3 f8e7 d4d5 e6d7 e3c4 b7b5 c1e1 e7f8');
INSERT INTO evaluations VALUES(16354,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(16354,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(16354,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(16354,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16354,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16354,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16354,36,'cp 0','f6d5 c3d4 d7d6 f1b5 d6e5 f3e5 f8b4 c1d2 e8g8 b5c6 b7c6 d2b4 d5b4 e1g1 c8f5 d1f3 f5e6 f3c3 b4a2 c3c5 d8d5 e5c6 d5c5 d4c5 f8e8 b1c3 a2c3 b2c3 a7a6 f1d1 e6c4 c6a5 c4e6');
INSERT INTO evaluations VALUES(16774,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16774,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16774,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16774,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16774,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16774,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(16774,36,'cp 0','c3d4 d7d6 f1b5 d6e5 e1g1 e5d4 f3d4 c8d7 b1c3 d5c3 b2c3 f8e7 b5c6 d7c6 d4c6 b7c6 d1a4 e8g8 a4c6 d8d6 c6a4 d6f6 c1e3 f6c3 a1c1 c3e5 e3f4 e5f5 f4c7 f8c8 a4f4 f5d5 f4c4 d5f5');
INSERT INTO evaluations VALUES(16773,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(16773,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(16773,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16773,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16773,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16773,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16773,36,'cp 0','d7d6 f1b5 d6e5 e1g1 e5d4 f3d4 c8d7 b1c3 d5c3 b2c3 f8e7 b5c6 b7c6 d1f3 e8g8 d4c6 d7c6 f3c6 d8d6 c6a4 d6f6 c1e3 f6c3 a1c1 c3e5 e3f4 e5f5 c1c7 e7c5 a4d7 f5d7 c7d7 f8e8 d7b7 e8e2 f4e3 c5e3 f2e3 e2a2 f1f7 a2a1 g1f2');
INSERT INTO evaluations VALUES(48726,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48726,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(48726,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(48726,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(48726,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(48726,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(48726,36,'cp 0','d2d4 f8d6 c3b5 e8g8 b5d6 d8d6 h2h3 c8f5 f1e2 b8c6 e1g1 h7h6 c2c3 f8e8 f1e1 c6e7 e2f1 a7a5 f3e5 e7g6 d1f3 f5c2 f3e2 c2f5');
INSERT INTO evaluations VALUES(49018,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(49018,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49018,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(49018,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(49018,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(49018,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(49018,36,'cp 14','d5b6 g1f3 b8c6 d2d3 f8e7 a2a3 a7a5 c1e3 e8g8 e3b6 c7b6 e1g1 c8e6 e2e3 a8c8 d3d4 e5d4 e3d4 e6c4 f1e1 e7f6 f3e5 d8d4 e5c4 d4c4 c3d5 f6b2 d5b6 c4c5 b6c8 b2a1 d1a1 f8c8 a1b2 c8d8 b2b7 c6d4 b7e7 c5e7 e1e7 d4c2 a3a4 c2b4 g2f1 g8f8 e7b7 d8d1 b7b5');
INSERT INTO evaluations VALUES(47744,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47744,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47744,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(47744,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(47744,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(47744,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(47744,36,'cp 26','g1f3 b8c6 d2d3 c8e6 a2a3 f8e7 e1g1 e8g8 b2b4 c6d4 c1b2 d4f3 g2f3 c7c6 c3e4 b6d7 d1c2 f7f5 e4d2 e7d6 f3g2 h7h6 d2c4 d6c7 a3a4 a7a6 a1c1 a8b8 c4d2 c7d6 b2c3 d8e7 c2b2 d6c7 f1e1 f5f4 a4a5 e6d5');
INSERT INTO evaluations VALUES(47743,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47743,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(47743,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(47743,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47743,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(47743,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(47743,36,'cp 18','b8c6 a2a3 f8e7 b2b4 e8g8 a1b1 f8e8 d2d3 a7a5 b4b5 c6d4 e1g1 e7f8 f3d2 a5a4 e2e3 d4e6 d2f3 e6c5 f3e5 e8e5 d3d4 e5e8 d4c5 d8d1 f1d1 f8c5 h2h3 c8f5 b1a1 a8d8 c1b2 d8d1 a1d1 b6c4 b2c1 c5a3 c1a3 c4a3 g2b7');
INSERT INTO evaluations VALUES(14688,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(14688,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(14688,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(14688,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(14688,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(14688,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(14688,36,'cp 17','a2a3 a7a5 e1g1 f8e7 d2d3 e8g8 c3a4 c8e6 c1e3 b6d5 e3c5 b7b6 c5e7 d5e7 a4c3 d8d7 d1a4 a8d8 f1d1 e6g4 b2b4 g4f3 g2f3 c6d4 b4a5 d4f3 e2f3 d7a4 c3a4 d8a8 a1c1 e7d5 d3d4 e5d4 d1d4 a8a5 a4c3 d5c3 c1c3 h7h6 f3f4 f8e8 c3c7 a5a3');
INSERT INTO evaluations VALUES(53269,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53269,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53269,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(53269,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53269,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53269,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53269,36,'cp 0','g8f6 f1e2 c5d4 e3d4 c8g4 h2h3 g4f3 e2f3 b8c6 e1g1 e7e6 c1e3 f8d6 c2c4 e8g8 c4c5 d6c7 b1c3 b7b6 d1a4 c6e7 b2b4 e7f5 a1d1 c7b8 f1e1 f5h4 f3e2 h4f5');
INSERT INTO evaluations VALUES(45384,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(45384,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45384,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45384,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(45384,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45384,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45384,36,'cp 10','f1e2 c5d4 e3d4 c8g4 c2c3 e7e6 c1f4 b8c6 b1d2 f8d6 f4d6 d8d6 e1g1 g4f5 f1e1 e8g8 e2b5 f8c8 a2a4 d6c7 d1e2 h7h6 h2h3 a8b8 e2e3 c6e7 f3e5 a7a6 b5d3 f5d3 e5d3 e7f5 e3e2 a6a5 d3c5 f5d6');
INSERT INTO evaluations VALUES(55812,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55812,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(55812,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55812,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55812,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(55812,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(55812,36,'cp -6','g8f6 c4d5 f6d5 f4g3 c7c5 d4c5 f8c5 e2e4 c5b4 b1d2 d5f6 g3b8 b4d2 d1d2 a8b8 d2d8 e8d8 f1d3 d8e7 e1e2 e6e5 a1c1 c8g4 e2e3 b8c8 g1e2 g4e6 f2f3 a7a6 c1c8 h8c8 a2a3 f6d7 g2g3 b7b5 h1c1 c8c1 e2c1 e7d6 b2b4 d7b6 d3e2');
INSERT INTO evaluations VALUES(48248,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48248,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48248,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(48248,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(48248,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(48248,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48248,36,'cp 1','c4d5 f6d5 f4g3 c7c5 d4c5 f8c5 e2e4 c5b4 b1d2 d5f6 g3b8 b4d2 d1d2 a8b8 d2d8 e8d8 f1d3 d8e7 e1e2 c8d7 f2f3 b8c8 e2e3 h8d8 g1e2 d7e8 a1c1 f6d7 c1c8 d8c8 e2c3 d7b6 a2a3 f7f6 g2g3 b6d7 f3f4 d7b6 d3e2 g7g5 f4g5 f6g5 e4e5');
INSERT INTO evaluations VALUES(19950,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(19950,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(19950,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19950,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19950,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19950,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19950,36,'cp 8','g8f6 f1e1 g7g6 c2c3 f8g7 d2d4 c5d4 c3d4 c8g4 b1c3 e8g8 c4e2 d8b6 d4d5 g4f3 g2f3 c6d4 c1e3 d4e2 e1e2 b6c7 e2c2 a8c8 a1c1 c7d7 c3e2 c8c2 c1c2 a7a6 g1g2 f8c8 c2c8 d7c8 b2b3 f6h5 e2g3 h5g3 h2g3 h7h5 e3d4 c8c7 d4g7 g8g7 d1d2 b7b5 d2d4 f7f6 d4b2 a6a5 f3f4 c7c5 b2d2 b5b4 g2g1');
INSERT INTO evaluations VALUES(13255,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(13255,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13255,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13255,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(13255,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(13255,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(13255,36,'cp 22','f1e1 g7g6 c2c3 f8g7 d2d4 c5d4 c3d4 c8g4 b1c3 e8g8 c4e2 d8b6 d4d5 g4f3 g2f3 c6d4 c1e3 d4e2 e1e2 b6d8 e2c2 a8c8 a1c1 d8d7 c3e2 c8c2 c1c2 a7a6 b2b3 f8c8 c2c8 d7c8 g1g2 f6h5 h2h3 f7f5 d1d3 g8f7 a2a4 g7f6 b3b4 c8d7 b4b5 a6b5 a4b5 f5e4');
INSERT INTO evaluations VALUES(13256,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13256,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13256,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13256,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13256,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(13256,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(13256,36,'cp 14','g7g6 c2c3 f8g7 d2d4 c5d4 c3d4 c8g4 b1c3 e8g8 c4e2 d8b6 d4d5 g4f3 g2f3 c6d4 c1e3 d4e2 e1e2 b6c7 e2c2 a8c8 a1c1 c7d7 c3e2 c8c2 c1c2 f8c8 g1g2 c8c2 d1c2 a7a6 h2h3 f6h5 b2b3 g7f6 a2a4 e7e5 d5e6 f7e6 c2d2 d6d5 e2g3 h5g3 f2g3 f6g7 e4d5 e6d5');
INSERT INTO evaluations VALUES(53001,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(53001,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(53001,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(53001,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(53001,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(53001,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(53001,36,'cp -32','f1b5 b8d7 b5c4 g7g6 h5e2 f8g7 g1f3 g8f6 a2a4 e8g8 e1g1 b7b6 d2d3 c8b7 a4a5 a7a6 a5b6 d7b6 b1d2 f6d7 f1e1 e7e6 h2h3 d8c7 c2c3 a6a5 c4b5 a5a4 d2c4 b6c4 d3c4 d7b6 c1f4 e6e5 f4h2 g7h6 e1d1 g8h8 f3d2');
INSERT INTO evaluations VALUES(52986,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(52986,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(52986,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(52986,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(52986,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(52986,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(52986,36,'cp -43','b8d7 b5c4 g7g6 h5e2 f8g7 g1f3 g8f6 e1g1 e8g8 b1c3 b7b6 d2d3 a7a6 a2a4 c8b7 f1e1 h7h6 c1f4 d8c7 h2h3 e7e6 f4h2 f6h5 a1d1 d7e5 f3e5 d6e5 f2f3 h5f4 e2f1 g6g5 g1h1 a8d8 c3e2 c7d7 a4a5 b6b5');
INSERT INTO evaluations VALUES(17520,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17520,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(17520,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(17520,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(17520,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(17520,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(17520,36,'cp 65','c2c4 e7e5 d4f5 d7d6 b1c3 g8f6 f5e3 c8e6 f1d3 g7g6 e1g1 f8g7 e3d5 e6d5 c3d5 f6d5 c4d5 c6d4 c1e3 e8g8 e3d4 e5d4 g2g3 d8b6 d1e2 b6b4 h2h4 a8e8 h4h5 f7f5 h5g6 h7g6 g1g2 e8e7 a1c1 f5e4 d3e4 b4b6');
INSERT INTO evaluations VALUES(17517,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(17517,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(17517,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(17517,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(17517,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(17517,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17517,36,'cp 70','e7e5 d4f5 d7d6 f5e3 g7g6 b1c3 g8f6 e3d5 f8g7 c1g5 h7h6 g5e3 f6d5 c3d5 e8g8 e3b6 d8h4 f1d3 c8g4 d1d2 g4e6 e1g1 a8c8 a1e1 c6d4 b6d4 e5d4 f2f4 f8e8 a2a4 e6d5 c4d5 c8c5 g2g3 h4e7 h2h4 h6h5 f1f2 c5c8 a4a5 c8c5');
INSERT INTO evaluations VALUES(17457,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(17457,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(17457,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(17457,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(17457,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(17457,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(17457,36,'cp 69','d4f5 d7d6 b1c3 g7g6 f5e3 c8e6 f1d3 g8f6 e1g1 f8g7 e3d5 e6d5 c3d5 f6d5 c4d5 c6d4 c1e3 e8g8 e3d4 e5d4 g2g3 a8c8 f2f4 d8d7 f1f2 c8c5 g1g2 f8c8 h2h4 h7h5 d1d2 g7h6 g2h2 a6a5 a1f1 c8c7 f1e1 c7c8 a2a3 a5a4 e1f1 d7d8 d2b4');
INSERT INTO evaluations VALUES(37809,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(37809,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(37809,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(37809,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(37809,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(37809,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(37809,36,'cp 105','d4d5 d7d6 f1c4 g8f6 b1c3 e8g8 e1g1 b8a6 c1f4 a6c7 a2a4 f6h5 f4e3 e7e6 d5e6 c8e6 c4e2 h5f6 f3g5 d8e7 d1d2 a8d8 g5e6 e7e6 e2f3 a7a6 a4a5 c7b5 c3b5 a6b5 f1e1 f6e4 f3e4 e6e4 e3g5 e4c6 g5d8 f8d8 d2g5 d8f8 c2c3 d6d5 g5g3 d5d4 a1d1');
INSERT INTO evaluations VALUES(37842,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(37842,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(37842,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(37842,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(37842,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(37842,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(37842,36,'cp 84','d7d6 b1c3 g8f6 f1c4 b8a6 e1g1 e8g8 c1f4 f6h5 f4e3 a6c7 a2a4 e7e5 d5e6 c8e6 c4e2 h5f6 f3g5 e6c8 e4e5 d6e5 e3c5 f8e8 d1d8 e8d8 a1d1 d8d1 f1d1 c8f5 e2c4 c7e6 g5e6 f5e6 c4e6 f7e6 a4a5 e5e4');
INSERT INTO evaluations VALUES(37988,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(37988,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(37988,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(37988,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(37988,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37988,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(37988,36,'cp 91','f1c4 b8a6 e1g1 g8f6 b1c3 e8g8 c1f4 a6c7 a2a4 f6h5 f4e3 b7b6 f1e1 e7e5 d5e6 c8e6 c4e6 c7e6 d1d2 g7c3 d2c3 d8f6 c3b3 a8d8 a1d1 f8e8 h2h4 f6e7 f3g5 h5f4 g5e6 f4e6 g2g3 e7d7 b3c4 d7c6 g1g2 h7h5 f2f3 d8b8 c4a6 c6c7 c2c3 g8h7');
INSERT INTO evaluations VALUES(57558,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(57558,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(57558,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(57558,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(57558,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(57558,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(57558,36,'cp 192','e2e4 b8c6 g1f3 e7e6 d4d5 e6d5 e4d5 c6e5 f3e5 d8e7 b1c3 f6e5 f1d3 g8f6 e1g1 d7d6 c1g5 e7f7 f2f4 c8g4 d1e1 e8c8 c3e4 f8e7 f4e5 d6e5 h2h3 g4h5 d5d6 d8d6 e4d6 e7d6 e1h4 h5g6 g5f6 g6d3 c2d3 g7f6 h4f6');
INSERT INTO evaluations VALUES(53073,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(53073,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53073,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53073,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53073,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53073,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53073,36,'cp 0','g8f6 b1c3 b8c6 d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 f1d3 f8e7 e1g1 e8g8 f1e1 c8b7 d1f3 d8c7 c1g5 h7h6 g5h4 a8d8 a1d1 b7c8 f3e2 g8h8 b2b4 c8e6 d3c4 a7a5 c4e6 f7e6 b4b5 d8b8 h4f6 e7f6 b5c6 c7c6 e2d3');
INSERT INTO evaluations VALUES(45210,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45210,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45210,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45210,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45210,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45210,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(45210,36,'cp 5','b1c3 b8c6 f1c4 e7e6 d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 e1g1 a7a6 a2a4 c8d7 c4d3 d6d5 e4d5 e6d5 f1e1 e7b4 d4c6 b7c6 d1f3 d7e6 e3g5 h7h6 g5h4 f8e8 a4a5 b4d6 h4g3 d6c7 g3c7');
INSERT INTO evaluations VALUES(45186,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(45186,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45186,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(45186,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(45186,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45186,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45186,36,'cp 7','b8c6 f1c4 e7e6 d2d4 c5d4 f3d4 f8e7 e1g1 c8d7 c1e3 e8g8 c4d3 a7a6 a2a4 d8c7 f2f4 c6d4 e3d4 d7c6 d1d2 f6d7 b2b3 a8e8 a1d1 e7f6 d3c4 e8e7 f1e1 f6h4 d4f2 h4f2 d2f2 d7f6 f2e3 f8d8 c4d3 e7e8 c3e2');
INSERT INTO evaluations VALUES(13301,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(13301,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(13301,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(13301,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(13301,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13301,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(13301,36,'cp 0','f1c4 e7e6 d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 e1g1 a7a6 a2a4 c8d7 c4d3 a8c8 f2f4 c6d4 e3d4 d7c6 d1e1 f6h5 a4a5 e7f6 d4b6 d8e7 e4e5 d6e5 f4e5 f6h4 e1e3 h4g5 e3f2 g5h4');
INSERT INTO evaluations VALUES(13233,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13233,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13233,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13233,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(13233,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13233,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13233,36,'cp 0','e7e6 d2d4 c5d4 f3d4 f8e7 e1g1 e8g8 c1e3 c8d7 a2a4 a7a6 f2f4 a8c8 c4d3 c6d4 e3d4 d7c6 d1e1 f6h5 a4a5 e7f6 d4b6 d8e7 e4e5 d6e5 f4e5 f6h4 e1e3 h4g5 e3f2 g5h4');
INSERT INTO evaluations VALUES(13925,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13925,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13925,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13925,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13925,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13925,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13925,36,'cp 13','d2d4 c5d4 f3d4 f8e7 e1g1 c8d7 c1e3 e8g8 a2a4 a7a6 c4d3 c6d4 e3d4 d7c6 a4a5 f6d7 b2b4 a8c8 b4b5 a6b5 c3b5 e7f6 d4f6 d7f6 b5a7 c8c7 a7c6 c7c6 d1d2 d6d5 e4e5 f6d7 d3b5 c6c8 c2c4 d7e5 c4d5 d8d5 d2d5 e6d5');
INSERT INTO evaluations VALUES(18922,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(18922,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(18922,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(18922,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(18922,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(18922,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(18922,36,'cp 39','c3d4 d7d5 e4e5 g8e7 b1c3 e7f5 f1b5 c8d7 e1g1 d8b6 b5a4 h7h6 c3e2 f8e7 a1b1 b6a6 e2c3 c6b4 a4d7 e8d7 a2a3 b4c6 b2b4 a6c4 c3e2 h8c8 b1b2 h6h5 h2h3 a7a5 b4b5 c6d8 a3a4 e7a3 b2b1 a3e7 e2g3 f5g3');
INSERT INTO evaluations VALUES(57127,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(57127,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(57127,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57127,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(57127,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(57127,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57127,36,'cp 0','e5e4 f3d4 b8c6 e2e3 c6d4 e3d4 d8f6 d4d5 f8c5 d1e2 f6e7 d2d3 e4d3 e2e7 g8e7 f1d3 d7d6 b1c3 c8f5 d3e2 e7g6 c1e3 c5e3 f2e3 e8g8 e1d2 f5d7 b2b3 a8e8 e3e4 g6f4 e2f3 f4g6');
INSERT INTO evaluations VALUES(57273,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(57273,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57273,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(57273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57273,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(57273,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57273,36,'cp 0','f3d4 b8c6 e2e3 c6d4 e3d4 d8f6 d4d5 f8c5 d1e2 f6e7 d2d3 e4d3 e2e7 g8e7 f1d3 d7d6 b1c3 c8f5 d3e2 e7g6 c1e3 c5e3 f2e3 g6e5 e1d2 f5d7 h1f1 h7h5 c3b5 e8d8 h2h4 d7b5 c4b5 g7g6 f1f6 d8e7 f6f4 e7d7 f4f6');
INSERT INTO evaluations VALUES(57276,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(57276,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57276,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57276,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(57276,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57276,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57276,36,'cp 0','b8c6 e2e3 c6d4 e3d4 d8f6 d4d5 f8c5 d1e2 f6e7 d2d3 e4d3 e2e7 g8e7 f1d3 d7d6 b1c3 c8f5 d3e2 f5d7 c1e3 c5e3 f2e3 e7g6 e1d2 g6e5 h1f1 h7h5 c3b5 e8d8 h2h4 a7a5 b5c3 d8e7 b2b3 a8g8 c3e4 e5g6 f1h1 g6e5');
INSERT INTO evaluations VALUES(11709,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11709,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(11709,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(11709,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(11709,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(11709,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11709,36,'cp 9','a4b3 f8c5 c2c3 d7d6 c1g5 h7h6 g5h4 c8e6 e1g1 g7g5 h4g3 e6b3 d1b3 e8g8 b1d2 f6h5 a2a4 b5b4 b3c4 c6a5 c4d5 c7c6 d5a2 d8f6 a1e1 a8b8 d3d4 b4c3 b2c3 h5g3 h2g3 e5d4 c3d4 c5d4');
INSERT INTO evaluations VALUES(50369,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(50369,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50369,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(50369,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50369,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50369,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(50369,36,'cp 19','e2e4 e7e5 g1f3 b8c6 d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 f1d3 e8g8 e1g1 d7d5 e4d5 c6d5 c1g5 c7c6 d1f3 b4d6 h2h3 h7h6 g5f6 d8f6 f3f6 g7f6 c3e2 d6e5 a1b1 a7a5 f2f4 e5c7 e2g3 f8e8 g3h5 g8f8 h5f6 e8e3 g1h1 c7d6 f6g4 c8g4 h3g4');
INSERT INTO evaluations VALUES(36297,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(36297,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36297,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36297,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(36297,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36297,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36297,36,'cp 17','e2e3 h7h6 g5h4 b7b6 f1e2 d5c4 e2c4 c8b7 e1g1 b8d7 d1e2 f6d5 c3d5 e6d5 h4e7 d8e7 c4d3 f8c8 f1c1 c7c5 b2b3 a7a5 d3b5 d7f6 c1c2 c8c7 a1c1 f6e4 h2h3 a8c8 d4c5 c7c5 b5d3 g7g6 c2c5 c8c5 c1c5 e7c5 f3d4 c5c3 d3e4 c3c1 g1h2 d5e4');
INSERT INTO evaluations VALUES(16699,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(16699,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(16699,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(16699,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(16699,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(16699,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(16699,36,'cp -95','f8b4 f1d3 d7d5 e4d5 d8d5 g1f3 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 c6e5 d3e2 e8c8 c1d2 e5c6 e1c1 c6d4 h1e1 h8e8 c3e4 b4d2 d1d2 f6h5 c2c3 d4e2 d2e2 e8f8 e1g1 h5f4 e2e3 g7g6 h3h4 h7h6 e4f6 d8d6 f6g4 h6h5 g4e5 d6f6 c1c2 f8d8 a2a4 d8d5 e5c4 b7b6 b2b4 d5f5 c4d2 f4d5 e3e4 c8b7');
INSERT INTO evaluations VALUES(10489,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(10489,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(10489,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(10489,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(10489,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(10489,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(10489,36,'cp -84','c1d2 e8g8 f1d3 f8e8 g1e2 b4c3 d2c3 f6e4 d3e4 e8e4 e1g1 d7d5 d1d3 c6e5 d3d2 f7f6 a1d1 e5c4 d2d3 c8g4 f1e1 e4e2 e1e2 g4e2 d3e2 d8d6 b2b3 c4b6 a2a4 a8d8 h2h4 b6c8 h4h5 h7h6 d1e1 d6d7 c3b4 b7b6 e2e6 d7e6 e1e6 c7c5 b4a3 c8d6 e6e7 a7a5 g1f1 g8f8 e7c7');
INSERT INTO evaluations VALUES(10490,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(10490,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(10490,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(10490,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(10490,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(10490,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(10490,36,'cp -101','e8g8 f1d3 b4c3 d2c3 f8e8 g1e2 f6e4 d3e4 e8e4 e1g1 d7d5 d1d3 c6e5 d3d2 f7f6 a1d1 e5c4 d2d3 c8g4 f1e1 g4e2 e1e2 e4e2 d3e2 d8e8 e2e8 a8e8 d1d5 c4b2 g1f1 b2a4 c3d4 a7a6 d5d7 e8c8 f1e2 b7b5 e2d3 h7h5 d7d5 h5h4 g2g3 c7c5 d4e3 h4g3 h2g3 c8c6 d5d7 g8h7 d7a7 h7g6 a7b7 c6e6');
INSERT INTO evaluations VALUES(4471,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(4471,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(4471,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(4471,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(4471,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(4471,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(4471,36,'cp -85','f1d3 f8e8 g1e2 b4c3 d2c3 f6e4 d3e4 e8e4 e1g1 d7d5 d1d3 c6e5 d3d2 e5c4 d2d3 f7f6 e2g3 e4e7 b2b3 c4e5 c3e5 e7e5 c2c4 c7c6 a1e1 c8e6 e1e5 f6e5 c4d5 d8d5 d3e2 a8d8 h2h3 d5d4 e2h5 h7h6 f1e1 a7a5 h5e5 d4e5 e1e5 d8d1 g1h2 e6d5 g3f5');
INSERT INTO evaluations VALUES(17252,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(17252,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(17252,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(17252,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(17252,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(17252,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(17252,36,'cp 43','g1f3 d7d6 b1c3 c8f5 e2e3 c6b4 a1c1 c7c6 a2a3 b4d5 c3d5 f6d5 f4g3 e7e6 f1d3 f5d3 d1d3 d5f6 c2c4 f8e7 e1g1 e8g8 h2h3 a7a5 d3c2 d6d5 c4c5 f6e4 g3h2 e7f6 f3e5 f6e5 h2e5 d8g5 g1h1');
INSERT INTO evaluations VALUES(17251,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(17251,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(17251,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(17251,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(17251,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(17251,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(17251,36,'cp 40','d7d6 b1c3 c8f5 e2e3 c6b4 a1c1 c7c6 a2a3 b4d5 c3d5 f6d5 f4g5 h7h6 g5h4 d8a5 f3d2 d5f6 h4f6 e7f6 f1e2 f8e7 c2c4 e8g8 e1g1 f5e6 d2e4 f6f5 e4c3 g7g6 g2g3 e7f6 d1d3 h6h5 h2h4 a5d8 f1d1 a7a5 g1g2 d8c7 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(14480,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(14480,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(14480,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(14480,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(14480,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14480,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(14480,36,'cp 47','b1c3 c8f5 a2a3 e7e6 e2e3 h7h6 f1d3 g7g5 d3f5 e6f5 f4g3 f8g7 d1d3 d8d7 h2h4 g5g4 f3g1 e8g8 g1e2 c6e7 h4h5 f8e8 g3h4 d6d5 h4f6 g7f6 e1g1 f6g5 e2g3 g5h4 c3e2 h4g3 e2g3 d7e6 a3a4 g8h8 a4a5 h8g7 d3b3');
INSERT INTO evaluations VALUES(14473,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(14473,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14473,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(14473,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(14473,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(14473,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14473,36,'cp 43','c8f5 e2e3 c6b4 a1c1 c7c6 a2a3 b4d5 c3d5 f6d5 f4g3 d5f6 f1d3 e7e6 d3f5 e6f5 g3h4 f8e7 h4f6 e7f6 c2c4 e8g8 g2g3 g7g6 h2h4 h7h6 e1f1 d8d7 f1g2 f8e8 d1d3 a7a5 h1d1 g8g7 a3a4 e8e7 b2b3 e7e8 h4h5 g6g5 d4d5 c6c5 d1h1 f5f4 e3f4');
INSERT INTO evaluations VALUES(22872,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(22872,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(22872,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(22872,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(22872,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(22872,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(22872,36,'cp 0','d2d4 e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 f8b4 f1e2 b4c3 b2c3 d7d6 e1g1 e8g8 f2f3 d8e8 d1e1 b7b6 f3e4 f5e4 e1h4 c8d7 c1a3 g8h8 c4c5 d6c5 h4g3 c6e7 d4c5 e7f5 g3h3 f5h6 h3g3');
INSERT INTO evaluations VALUES(22760,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(22760,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(22760,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22760,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22760,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22760,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(22760,36,'cp 0','e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 f8b4 f1e2 b4c3 b2c3 d7d6 e1g1 e8g8 c4c5 d6d5 a2a4 g8h8 d1e1 c8e6 c1a3 f8f7 f2f3 c6a5 a3b4 a5c6');
INSERT INTO evaluations VALUES(22927,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(22927,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(22927,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(22927,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22927,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(22927,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(22927,36,'cp 0','f3d2 f7f5 c2c4 g8f6 b1c3 f8b4 f1e2 b4c3 b2c3 d7d6 c4c5 d6d5 e1g1 e8g8 a2a4 g8h8 a4a5 a7a6 f2f3 d8e8 d1e1 e4f3 f1f3 c8e6 f3f1 e8d7 e1g3 h7h6 h2h3 a8e8 e2d3 f6e4 d2e4 f5e4 f1f8 e8f8 d3e2 h8h7 c1d2 c6e7 a1f1 e7f5');
INSERT INTO evaluations VALUES(22926,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22926,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22926,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22926,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22926,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22926,36,'cp 0','f7f5 c2c4 g8f6 b1c3 f8b4 d2b1 d7d5 a2a3 b4e7 f1e2 c8e6 c4d5 f6d5 c3d5 e6d5 b1c3 d5f7 d1c2 d8d7 e1g1 e8g8 c1d2 e7d6 f2f3 e4f3 e2f3 c6a5 c3d5 a5c6');
INSERT INTO evaluations VALUES(29532,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(29532,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(29532,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(29532,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29532,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29532,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29532,36,'cp 0','e2e3 e7e6 g1f3 c7c6 f1d3 f5g6 e1g1 f8e7 d1e2 f6h5 f4e5 e8g8 h2h3 h5f6 e5h2 d8b6 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(29528,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(29528,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(29528,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(29528,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29528,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29528,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29528,36,'cp 4','e7e6 f1d3 c7c6 d3f5 e6f5 c3e2 b8d7 c2c3 f8e7 d1b3 d8b6 g1f3 e8g8 b3c2 f8e8 e1g1 g7g6 a2a4 a7a5 h2h3 f6e4 b2b3 a8c8 c3c4 e7b4 f1d1 b6d8 f3e1 g6g5 f4h2 g5g4 f2f3 g4f3 e1f3 d7f6 d1f1 d8e7 c4c5');
INSERT INTO evaluations VALUES(30079,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30079,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30079,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30079,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30079,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30079,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(30079,36,'cp 0','f1d3 f5g6 h2h4 a7a6 h4h5 g6d3 d1d3 h7h6 e3e4 c7c5 g1e2 c5d4 e2d4 f8b4 e4d5 f6d5 f4e5 e8g8 e5g7 g8g7 d4e6 f7e6 d3g6 g7h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(30084,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30084,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30084,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30084,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30084,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30084,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30084,36,'cp 0','f5g6 h2h4 a7a6 h4h5 g6d3 d1d3 h7h6 e3e4 c7c5 g1e2 c5d4 e2d4 f8b4 e4d5 f6d5 f4e5 e8g8 e5g7 g8g7 d4e6 f7e6 d3g6 g7h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(40939,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40939,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(40939,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(40939,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(40939,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(40939,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(40939,36,'cp 27','e4e5 f6d7 g5e7 d8e7 f2f4 a7a6 g1f3 c7c5 h2h4 b8c6 d1d2 c5d4 f3d4 c6d4 d2d4 b7b5 e1c1 b5b4 c3a4 c8b7 h4h5 e8g8 h5h6 g7g6 a4b6 d7b6 d4b6 f8c8 g2g4 a6a5 c1b1 c8c6 b6e3 a8c8 h1h2 b7a6 f1a6 c6a6 f4f5 a5a4');
INSERT INTO evaluations VALUES(41013,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(41013,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(41013,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(41013,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(41013,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(41013,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41013,36,'cp 24','f6d7 g5e7 d8e7 f2f4 a7a6 d1g4 g7g6 g1f3 c7c5 d4c5 b8c6 g4g3 d7c5 f1d3 c8d7 a1d1 e8c8 a2a3 c8b8 h2h3 d8c8 e1g1 c8c7 g1h2 h8c8 c3e2 c6a7 f3d4 c5d3 c2d3 e7c5 g3h4 a7b5 d4f3 c5e3 f1e1 c7c2 h4h7');
INSERT INTO evaluations VALUES(42294,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(42294,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(42294,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(42294,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(42294,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(42294,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42294,36,'cp 21','g5e7 d8e7 f2f4 a7a6 d1g4 g7g6 f1d3 c7c5 d4c5 b8c6 h2h4 h7h5 g4g3 d7c5 g1f3 b7b5 e1g1 b5b4 c3e2 a6a5 d3b5 c8d7 f3g5 a8b8 b5c6 d7c6 e2d4 c5e4 g3d3 e7c5 f1f3 e8g8 g1h2 e4g5 h4g5 c6b5');
INSERT INTO evaluations VALUES(42291,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(42291,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(42291,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42291,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(42291,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42291,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42291,36,'cp 26','d8e7 f2f4 a7a6 d1d2 c7c5 c3e2 b8c6 g1f3 e8g8 c2c3 c5d4 e2d4 f7f6 e5f6 e7f6 f1e2 c6d4 f3d4 e6e5 d4f3 f6f4 d2d5 g8h8 d5d2 d7f6 e1g1 f4d2 f3d2 c8e6 f1d1 e5e4 a2a4 g7g6 d2c4 e6c4 e2c4 a8d8 a4a5 h8g7 c4e2 h7h5 d1d8 f8d8 a1d1 f6d5 d1d4');
INSERT INTO evaluations VALUES(33440,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(33440,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(33440,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(33440,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(33440,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(33440,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(33440,36,'cp 22','f2f4 a7a6 d1g4 g7g6 g1f3 c7c5 d4c5 b8c6 f1d3 d7c5 g4g3 c8d7 a2a3 b7b5 e1g1 b5b4 a3b4 c5d3 c2d3 e7b4 g3f2 e8g8 f4f5 g6f5 d3d4 c6e7 f2h4 e7g6 h4h6 f8c8 g2g4 b4f8 h6f8 g8f8 g4f5 e6f5 c3d5');
INSERT INTO evaluations VALUES(33435,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(33435,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(33435,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(33435,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(33435,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(33435,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(33435,36,'cp 42','a7a6 d1g4 g7g6 g1f3 c7c5 d4c5 b8c6 f1d3 d7c5 e1g1 c8d7 a2a3 c5d3 c2d3 e7c5 g1h1 c6d4 a1c1 c5b6 f3d4 b6d4 g4d1 e8g8 h2h3 a8c8 d1d2 c8c6 g2g4 f8c8 h1h2 d4b6 h2g3 h7h6 c1e1 d5d4 c3e2 c6c2');
INSERT INTO evaluations VALUES(33337,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(33337,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(33337,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(33337,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(33337,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(33337,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(33337,36,'cp 30','d1g4 g7g6 g4g3 c7c5 d4c5 b8c6 g1f3 d7c5 f1d3 c8d7 g3f2 f7f5 e5f6 c5d3 c2d3 e7f6 g2g3 e8g8 d3d4 g6g5 f3g5 f6d4 f2d4 c6d4 e1c1 d4c6 h1e1 a8c8 c1b1 d5d4 c3e4 e6e5 f4e5 d7f5 g3g4 f5g6 g5e6 f8e8 e6f4 e8e5 f4g6 h7g6 e4d6 c8f8 e1e5 c6e5 d1d4 e5c6');
INSERT INTO evaluations VALUES(6611,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(6611,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(6611,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(6611,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(6611,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(6611,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6611,36,'cp 18','b4c3 d2c3 d7d6 f1d3 d8f6 h2h3 h7h6 d1e2 g8e7 c1e3 c8e6 e1c1 e8c8 f3d2 c8b8 b2b4 d6d5 e4d5 e7d5 d2e4 f6g6 g2g4 f7f5 g4f5 e6f5 c1b2 d5b6 e3b6 a7b6 f2f3 g6f7 e4f2 h8f8 d3c4 f7e7 c4d5 h6h5 h1e1 e7h4 d5c6 d8d1 e1d1 b7c6 d1e1 f5h3 f2h3 h4h3');
INSERT INTO evaluations VALUES(6651,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(6651,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6651,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(6651,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(6651,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(6651,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6651,36,'cp 19','d2c3 d7d6 f1d3 g8e7 d3c4 h7h6 c1e3 d8d7 f3d2 d7g4 d1g4 c8g4 f2f3 g4d7 e1c1 a7a5 b2b3 f7f5 e3f2 f5e4 d2e4 e7g6 c1b2 g6f4 h1e1 e8c8 c4f1 h8f8 c3c4 g7g5 e4c3 b7b6 h2h4 g5g4 f3g4 d7g4 d1d2 f4e6 b3b4');
INSERT INTO evaluations VALUES(6649,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(6649,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(6649,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(6649,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(6649,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6649,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(6649,36,'cp 13','d7d6 f1d3 g8e7 d3c4 h7h6 f3h4 d8d7 d1f3 c6d8 c4d3 d7g4 f3g3 h6h5 f2f3 g4g3 h2g3 d8e6 h4f5 e7f5 e4f5 e6c5 g3g4 c5d3 c2d3 c8d7 e1f2 f7f6 c1e3 e8f7 h1h5 h8h5 g4h5 d7f5 d3d4 a7a5 d4e5 d6e5 g2g4 f5d7 a1d1 d7c6 f3f4 c6a4 d1d2 e5f4 e3f4');
INSERT INTO evaluations VALUES(6062,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(6062,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(6062,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(6062,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(6062,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(6062,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(6062,36,'cp 15','f1d3 d8e7 f3d2 c8e6 e1g1 e7d7 d2c4 g8e7 f2f4 e5f4 c1f4 e6c4 d3c4 e7g6 f4g3 e8g8 h2h3 a8e8 c4d5 a7a5 g1h1 b7b6 d1e2 c6e5 a1d1 d7e7 g3h2 e7d7 h2g1 a5a4 g1e3 g6e7 d5a2 e7g6 e3f2 d7e7 d1e1');
INSERT INTO evaluations VALUES(48014,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(48014,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(48014,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(48014,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(48014,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48014,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(48014,36,'cp 19','e4c3 d2c3 f8e7 c1f4 b8d7 d1d2 e8g8 e1c1 d7c5 f1c4 d8e8 c1b1 c8e6 c4e6 f7e6 f4e3 c5e4 d2e2 d6d5 f3d2 e4d2 e2d2 e7d6 h2h4 e8g6 d1g1 g6f7 h4h5 e6e5 h5h6 g7g6 f2f3 b7b6 e3g5 a8e8 g1d1 c7c6 a2a3 f7f5 h1e1');
INSERT INTO evaluations VALUES(48025,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(48025,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(48025,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(48025,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48025,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(48025,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48025,36,'cp 15','d2c3 f8e7 c1f4 e8g8 d1d2 b8d7 e1c1 d7c5 f1c4 d8e8 h1e1 c8e6 f4g5 a7a5 b2b3 e7g5 f3g5 e8d7 g5e6 f7e6 f2f4 d6d5 g2g3 c7c6 c4d3 c5d3 d2d3 a5a4 b3b4 a4a3 d3d4 f8e8 c1b1');
INSERT INTO evaluations VALUES(48024,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(48024,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(48024,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48024,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48024,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(48024,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(48024,36,'cp 20','f8e7 c1f4 b8d7 d1d2 e8g8 e1c1 d7c5 f1c4 d8e8 c1b1 c8e6 c4e6 f7e6 f4e3 b7b6 d2e2 e8g6 h1g1 g6f7 f3d2 c5d7 f2f4 e6e5 f4e5 d7e5 e3d4 a8e8 g1f1 f7d5 b2b3 e5c6 f1f8 e8f8 d2f3 d5f7 e2e4 c6d4 f3d4 e7f6');
INSERT INTO evaluations VALUES(40874,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(40874,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(40874,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40874,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40874,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(40874,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40874,36,'cp 23','c1f4 e8g8 d1d2 b8a6 f1c4 a6c5 e1c1 c8d7 h1e1 d7e6 c4e6 c5e6 f4e3 f8e8 d2d3 d8d7 a2a3 a7a6 h2h4 d7b5 g2g4 b5d3 c2d3 d6d5 c1c2 e7d6 a3a4 f7f6 a4a5 g8f7 b2b4 a8d8 d1a1 d6f4 g4g5 f6g5 f3g5 e6g5');
INSERT INTO evaluations VALUES(40872,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40872,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40872,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40872,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(40872,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40872,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40872,36,'cp 0','e8g8 d1d2 b8a6 f1c4 a6c5 e1c1 d8e8 h1e1 c8e6 c4e6 f7e6 f4e3 c5e4 d2e2 e8g6 g2g3 g6g4 f3d4 g4e2 e1e2 e6e5 d4e6 f8c8 f2f3 g8f7 e6g7 e4c3 b2c3 f7g7 f3f4 e5f4 e3d4 g7f7 d1f1 e7g5 h2h4 g5h6 g3g4 c7c5 g4g5 h6g5 h4g5 c5d4 f1f4 f7g6 c3d4 c8f8 e2e6 g6g5 f4f8 a8f8 e6d6 f8f7 d6d8 a7a5 c1b2');
INSERT INTO evaluations VALUES(19079,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19079,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19079,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19079,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19079,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19079,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19079,36,'cp 12','g1f3 g8f6 f1d3 f8c5 a2a3 e8g8 c2c4 d7d6 b1c3 c6d4 h2h3 c8e6 e1g1 a8c8 c1d2 d4f3 d1f3 c5d4 b2b3 d4c3 d2c3 b7b5 f1d1 f8e8 f3e2 c8c6 c3b4 b5c4 b3c4 d8c7 a1c1 a7a5 b4d2 e8c8 e2e1 a5a4 d2b4 f6d7 d3e2 d7c5');
INSERT INTO evaluations VALUES(48019,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48019,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48019,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(48019,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(48019,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(48019,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(48019,36,'cp 7','e4f6 d3d4 d6d5 f1d3 f8d6 e1g1 e8g8 c1g5 c8e6 f1e1 b8d7 b1d2 h7h6 g5h4 c7c5 d4c5 d7c5 d2f1 c5e4 d3e4 d5e4 f3d2 d6e5 c2c3 d8c7 h4f6 e5f6 d2e4 f6e7 e4g3 a8d8 d1c2 a7a5 a1d1 e7c5 b2b3 b7b5 d1d8 f8d8');
INSERT INTO evaluations VALUES(47966,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(47966,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47966,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(47966,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(47966,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47966,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(47966,36,'cp 18','d3d4 d6d5 f1d3 f8d6 d1e2 d6e7 e1g1 e8g8 f1e1 f8e8 c1g5 c8e6 b1d2 h7h6 g5h4 b8d7 c2c3 c7c6 a2a4 e7d6 f3e5 d8b6 d2f3 e6g4 e2c2 g4f3 e5f3 e8e1 a1e1 a8e8 e1e8 f6e8 c2e2 e8f6 d3f5 b6c7 f5d7 f6d7 h4g3 d7f6 g3d6 c7d6 e2e5 d6e5 f3e5 f6e4 f2f3 e4g5 e5d3 g5e6');
INSERT INTO evaluations VALUES(47944,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(47944,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(47944,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(47944,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47944,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(47944,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47944,36,'cp 19','d6d5 f1d3 f8d6 d1e2 d6e7 e1g1 e8g8 f1e1 f8e8 c1g5 c8e6 c2c4 b8c6 b1c3 c6b4 c4c5 h7h6 g5f6 e7f6 d3b5 e6d7 e2d2 e8e1 a1e1 a7a6 b5d7 d8d7 a2a3 b4c6 h2h3 c6e7 c3e2 a8e8 e2f4 f6g5 f3g5 h6g5 f4h5 g5g4 h3g4 d7g4 d2h6 g7h6');
INSERT INTO evaluations VALUES(48626,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(48626,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(48626,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(48626,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(48626,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(48626,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48626,36,'cp -11','f8d6 f1e2 c8e6 e1g1 e8g8 b1d2 f8e8 e2d3 b8d7 c2c3 c7c6 f1e1 d8c7 g5h4 e6g4 d1c2 g4h5 h4g3 d6g3 h2g3 h5g6 d3g6 h7g6 c3c4 c7b6 a1c1 d7f8 c4d5 e8e1 c1e1 c6d5 c2b3 b6b3 d2b3 f8e6');
INSERT INTO evaluations VALUES(41687,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(41687,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(41687,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(41687,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(41687,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41687,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(41687,36,'cp -3','f1e2 e8g8 e1g1 c7c6 c2c4 h7h6 g5h4 d5c4 e2c4 c8f5 d1b3 b7b5 c4d3 f5d3 b3d3 b8d7 f1e1 d8b6 b1c3 f8e8 d3f5 d6f8 h2h3 b6b7 a1d1 e8e1 d1e1 a8e8 e1e8 f6e8 c3e4 e8d6 f5f4 d6e4 f4e4 d7b6 f3e5 b6d5 h4g3');
INSERT INTO evaluations VALUES(41686,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(41686,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(41686,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(41686,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41686,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(41686,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(41686,36,'cp -10','e8g8 e1g1 c8f5 e2d3 f5d3 d1d3 b8d7 b1d2 h7h6 g5f6 d7f6 f1e1 f8e8 e1e8 d8e8 g2g3 a7a5 a1e1 e8a4 a2a3 a4c6 c2c3 a5a4 f3e5 c6b6 d3c2 a8e8 e5d3 e8e1 d3e1 c7c5 d4c5 d6c5 e1d3');
INSERT INTO evaluations VALUES(36929,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(36929,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(36929,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(36929,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(36929,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(36929,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36929,36,'cp -3','e1g1 c8f5 e2d3 f5d3 d1d3 b8d7 b1d2 c7c6 c2c4 d5c4 d2c4 d8c7 d3b3 a7a5 g5h4 a5a4 b3c2 f6d5 f1e1 d6b4 h4g3 c7d8 e1e2 f8e8 e2e8 d8e8 a2a3 b4e7 g3d6 d7f8 d6e7 e8e7 a1d1 f8e6 g2g3 h7h6 d1e1 e7c7 c4e5 c7b6');
INSERT INTO evaluations VALUES(36928,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(36928,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36928,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36928,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(36928,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(36928,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36928,36,'cp -9','c8e6 b1d2 f8e8 f1e1 b8d7 e2d3 c7c6 c2c3 d8c7 g5h4 g7g6 h4g3 d6g3 h2g3 c7b6 d1c2 c6c5 d4c5 d7c5 d2b3 c5e4 b3d4 e6d7 c2b3 b6c7 d3b5 d7b5 d4b5 c7b6 b5d4 b6b3 a2b3 a7a6 b3b4 g8g7 d4c2 f6d7');
INSERT INTO evaluations VALUES(21974,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(21974,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(21974,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(21974,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(21974,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(21974,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(21974,36,'cp 25','b1c3 g7g6 d3d4 e5d4 f3d4 f8g7 f1b5 c8d7 d4c6 b7c6 b5a4 g8f6 c1g5 e8g8 e1g1 h7h6 g5e3 f8e8 f2f3 d6d5 d1d2 d5e4 e3h6 e4f3 h6g7 g8g7 f1f3 d7f5 d2d8 e8d8 h2h3 a8b8 g2g4 f5e6 b2b3 e6d5 f3f2 f6e4 c3e4 d5e4 a1e1');
INSERT INTO evaluations VALUES(21962,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21962,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(21962,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(21962,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(21962,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(21962,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(21962,36,'cp 4','g7g6 d3d4 e5d4 f3d4 f8g7 d4c6 b7c6 f1d3 g8f6 e1g1 e8g8 c1f4 a8b8 a1b1 f8e8 f1e1 h7h6 b2b3 f6g4 c3e2 c6c5 h2h3 g4e5 d1d2 g6g5 f4e5 g7e5 b1d1 c8e6 e1f1 e6d7 g1h1 d8f6 d3c4 d7c6 f2f3 g8h8 d1b1 h8h7');
INSERT INTO evaluations VALUES(52428,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(52428,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(52428,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(52428,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(52428,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(52428,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(52428,36,'cp -9','c5d4 d1d4 g8f6 b1c3 b8c6 d4d1 e7e6 g1f3 d7d5 e2e3 f8b4 f1e2 e8g8 e1g1 d5c4 e2c4 d8e7 c1d2 e6e5 a2a3 b4c5 b2b4 e5e4 b4c5 e4f3 d1f3 c6e5 f3e2 c8f5 f1d1 f8d8 h2h3 a8c8 d2e1 e7c5 d1d8 c8d8 c4b3 f5d3');
INSERT INTO evaluations VALUES(52509,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(52509,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(52509,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(52509,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(52509,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(52509,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(52509,36,'cp -12','d1d4 b8c6 d4d3 g7g6 e2e4 f8g7 b1c3 g8f6 g1f3 d7d6 c1d2 c8g4 a1c1 f6d7 f1e2 d7c5 d3b1 e8g8 e1g1 f7f5 b2b4 g7c3 d2c3 f5e4 f3d4 g4e2 d4e2 c5d3 c1d1 d8d7 e2g3 d7e6 g3e4 e6e4 b1d3 e4d3 d1d3 a7a5 a2a3 a5b4 a3b4 f8f4 b4b5 c6a5 c3a5');
INSERT INTO evaluations VALUES(52507,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(52507,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(52507,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(52507,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(52507,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(52507,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(52507,36,'cp -20','b8c6 d4d1 g8f6 b1c3 e7e6 g1f3 d7d5 e2e3 f8b4 f1e2 e8g8 e1g1 d5c4 e2c4 a7a6 d1e2 b7b5 c4d3 b4d6 c3e4 f6e4 d3e4 c8b7 b2b3 f7f5 e4c6 b7c6 c1b2 c6d5 f1d1 d8e7 b2e5 d5f3 e5d6 f3e2 d6e7 e2d1 e7f8 g8f8 a1d1');
INSERT INTO evaluations VALUES(19532,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19532,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(19532,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(19532,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(19532,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19532,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(19532,36,'cp -6','d4d3 g8f6 b1c3 g7g6 e2e4 f8g7 f1e2 d7d6 g1f3 c8g4 c1e3 g4f3 g2f3 f6d7 f3f4 d8a5 e1g1 d7c5 d3c2 g7c3 b2c3 b7b6 a2a4 a5a6 g1h1 a6c8 e3d4 h8g8 d4c5 d6c5 e2f3 c8c7 e4e5 e8c8 f1b1');
INSERT INTO evaluations VALUES(571,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(571,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(571,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(571,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(571,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(571,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(571,36,'cp 17','f1c4 c5b6 e1g1 e8g8 d4c2 d7d6 b1a3 c8e6 e3b6 e6c4 a3c4 a7b6 c4e3 f6g6 f2f3 f7f5 e4f5 e7f5 f3f4 c6e7 e3d5 e7d5 d1d5 g6f7 d5b7 a8a2 a1a2 f7a2 b7c7 b6b5 c7c6 a2b2 c6d5 g8h8 d5d3 h7h6 f1e1 d6d5 h2h3 b2b3 c2d4 f5d4 d3d4 b3c4');
INSERT INTO evaluations VALUES(568,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(568,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(568,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(568,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(568,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(568,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(568,36,'cp 29','e8g8 e1g1 c5b6 a2a4 d7d6 d1d2 f6g6 f2f3 f8d8 d4c6 e7c6 b1a3 c6e5 c4a2 c8d7 a2b3 d7e6 b3e6 b6e3 d2e3 f7e6 e3e2 a7a6 a3c4 e5c4 e2c4 g6f7 c4e2 e6e5 f1d1 f7b3 d1d2 b3b6 g1h1 d8f8');
INSERT INTO evaluations VALUES(2226,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(2226,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(2226,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(2226,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(2226,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(2226,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(2226,36,'cp 20','e1g1 c5b6 f1e1 d7d6 c4f1 d6d5 d4c6 f6c6 e4d5 e7d5 e3b6 a7b6 d1d2 c6d6 h2h3 f8d8 b1a3 d5f6 d2e3 c8d7 a3c4 d6c5 e3g3 b6b5 b2b4 c5f8 c4e5 d7e6 f1b5 a8a2 a1a2 e6a2 b5f1 f8d6 b4b5 a2e6 c3c4 c7c6 b5c6 b7c6 g3g5');
INSERT INTO evaluations VALUES(2225,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(2225,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(2225,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(2225,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(2225,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(2225,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(2225,36,'cp 14','c5b6 b1a3 f8d8 d1h5 c6d4 c3d4 h7h6 e4e5 f6g6 h5g6 e7g6 a1c1 d7d6 e5d6 c7d6 c4b3 d6d5 a3b5 c8e6 b5c3 g6e7 c3a4 e7c6 a4b6 a7b6 a2a4 c6a5 b3d1 e6d7 e3f4 a8c8 b2b3 c8c1 f4c1 a5c6 c1b2 d7e6 d1e2');
INSERT INTO evaluations VALUES(2223,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(2223,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(2223,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(2223,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(2223,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(2223,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(2223,36,'cp 22','a2a4 d7d6 f1e1 c8d7 d4c6 e7c6 b1d2 b6e3 e1e3 c6e5 h2h3 e5c4 d2c4 a7a6 e4e5 f6f4 b2b3 d7e6 e5d6 e6c4 b3c4 f4d6 d1d6 c7d6 e3e7 f8c8 e7b7 c8c4 a1a3 c4c6 b7d7 g7g6 a3b3 g8g7 b3b7 a8f8 b7a7 h7h5 d7c7 c6c7 a7c7 f8b8 c7c6 b8b1 g1h2 d6d5 c6a6');
INSERT INTO evaluations VALUES(20866,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(20866,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(20866,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(20866,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(20866,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(20866,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(20866,36,'cp 53','f2f4 c8f5 c2c3 e7e6 g1f3 g8h6 f1e2 f8e7 b1d2 a7a5 e1g1 e8g8 a2a4 f6e5 f4e5 d8e8 f3e1 e7g5 d2f3 g5c1 d1c1 c6e7 c1d2 e7g6 e1d3 b7b6 d3f2 c7c5 e2b5 e8e7 g2g3 a8c8 a1e1 h6g4 h2h4 g4f2 f1f2 f5e4 f3g5 f8f2 g1f2 c5d4');
INSERT INTO evaluations VALUES(20864,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(20864,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(20864,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(20864,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(20864,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(20864,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(20864,36,'cp 44','c8f5 c2c3 e7e6 g1e2 g8h6 e2g3 f5g6 f1e2 f8e7 e1g1 e8g8 b1d2 f6e5 f4e5 f8f1 d2f1 e7g5 e2g4 h6g4 d1g4 g5c1 a1c1 d8d7 f1e3 c6e7 h2h4 a8f8 h4h5 g6f5 e3f5 e7f5 g3f5 f8f5 h5h6 b7b6 c1f1 f5f1 g1f1 c7c5 d4c5 b6c5');
INSERT INTO evaluations VALUES(26472,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(26472,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(26472,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(26472,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(26472,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(26472,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(26472,36,'cp 50','c2c3 e7e6 g1f3 f8e7 f1e2 g8h6 b1d2 e8g8 e5f6 e7f6 e1g1 a7a5 a2a4 h6f7 f3e5 f7e5 f4e5 f6g5 d2f3 g5c1 d1c1 h7h6 c1e3 f5h7 f1f2 c6e7 a1f1 e7f5 e3d2 f5h4 f3e1 f8f2 f1f2 d8e7 e1d3 a8f8');
INSERT INTO evaluations VALUES(45129,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45129,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45129,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(45129,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(45129,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(45129,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(45129,36,'cp 2','g1f3 a7a6 d2d4 c5d4 f3d4 e7e6 c4b3 b7b5 c1e3 f8e7 a2a3 e8g8 d1e2 c8b7 f2f3 f6d7 e1c1 d7c5 b3a2 d8c7 h2h4 b8c6 e3g5 f8b8 g5e7 c7e7 e2d2 b5b4 d4c6 b7c6 d2d6 e7d6 d1d6 b4c3 d6c6 c3b2 c1b1 a6a5');
INSERT INTO evaluations VALUES(49589,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(49589,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(49589,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(49589,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(49589,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49589,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(49589,36,'cp -18','e5d4 d1d4 b8c6 d4d3 f8c5 c1f4 d7d6 d3d2 c8e6 g1f3 e8g8 f1d3 f8e8 e1g1 e6g4 f1e1 f6h5 f4g5 f7f6 g5e3 g4f3 g2f3 c6e5 d3e2 g7g5 c3d5 g8h8 g1h1 c7c6 f3f4 h5f4 d5f4 g5f4 e3c5 d6c5 d2f4 e5g6 f4e3 d8d4 e2h5 d4e3 e1e3 a8d8 e3g3 e8e4 h5g6 h7g6');
INSERT INTO evaluations VALUES(49773,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(49773,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(49773,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(49773,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(49773,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(49773,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(49773,36,'cp -6','d1d4 b8c6 d4d3 f8c5 c1f4 d7d6 d3d2 c8e6 g1f3 e8g8 f1d3 f8e8 e1g1 e6g4 a1e1 g4f3 g2f3 c6e5 g1g2 c7c6 f4g3 c5b4 a2a3 b4c3 d2c3 f6h5 f3f4 e5d3 c2d3 d8d7 f4f5 d6d5 g2h1 f7f6 f1g1 g8h8 c3d4');
INSERT INTO evaluations VALUES(49767,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(49767,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(49767,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(49767,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(49767,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(49767,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(49767,36,'cp -26','b8c6 d4d3 f8c5 c1f4 d7d6 d3d2 c8e6 g1f3 e8g8 f1d3 f8e8 e1g1 e6g4 a1e1 g4f3 g2f3 c6e5 d3e2 c7c6 c3a4 c5b6 c2c4 b6a5 b2b4 a5c7 f4g3 d8d7 e1d1 d7h3 a4c3 h7h6 g1h1 c7b6 f1g1 f6h5 e2f1 h5g3 g1g3 h3h4 f3f4 e5g4');
INSERT INTO evaluations VALUES(16687,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(16687,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(16687,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(16687,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(16687,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(16687,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(16687,36,'cp 0','d4d3 f8c5 c1f4 d7d6 d3d2 c8e6 e1c1 c5b4 d2d3 e8g8 f2f3 f6d7 c1b1 c6e5 d3d4 a7a5 c3b5 b4c5 d4c3 c5b4');
INSERT INTO evaluations VALUES(16701,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16701,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16701,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16701,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(16701,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16701,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(16701,36,'cp -12','f8c5 c1f4 d7d6 d3d2 c8e6 g1f3 e8g8 a2a3 a7a6 f1d3 f6g4 f4g3 h7h6 e1g1 g4e5 f3e5 c6e5 d3e2 d8g5 f1d1 f8e8 d2g5 h6g5 h2h3 a8c8 g1f1 e5c4 b2b4 c5a7 a3a4 f7f6 c3d5 c7c6 e2c4 c6d5 e4d5 c8c4 d5e6 e8e6 d1d6');
INSERT INTO evaluations VALUES(31072,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(31072,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(31072,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(31072,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(31072,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(31072,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(31072,36,'cp 104','c4d5 c6d5 d1b3 b8c6 b1c3 g8f6 b3b7 f5d7 c1f4 a8c8 e2e3 f6h5 f4g3 h5g3 h2g3 c8b8 b7a6 e7e6 a6e2 d8b6 a1b1 f8b4 e2d2 b4c3 d2c3 e8g8 f1d3 h7h6 e1g1 f8c8 c3d2 c6b4 f3e5 d7e8 a2a3 b4d3 e5d3 e8b5 f1c1 b5d3 c1c8 b8c8 d2d3 b6c7 g1f1 c7a5 b2b4 a5a4 d3b3 a4a6 b4b5 a6a5');
INSERT INTO evaluations VALUES(31055,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(31055,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(31055,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(31055,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(31055,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(31055,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(31055,36,'cp 110','c6d5 d1b3 f5c8 e2e4 d5e4 f1b5 c8d7 f3g5 g8h6 b5c4 e7e6 d4d5 f8e7 g5e4 d8b6 c1h6 g7h6 b1c3 b8a6 b3b6 a7b6 e1c1 e8g8 h1e1 a8c8 d5e6 d7e6 c4e6 f7e6 f2f3 f8d8 d1d8 c8d8 e1d1 d8d1 c1d1 a6b4 a2a3 b4c6 d1c2 c6d4 c2d2 g8f7');
INSERT INTO evaluations VALUES(31320,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(31320,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(31320,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(31320,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(31320,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(31320,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(31320,36,'cp 129','d1b3 b8c6 b1c3 a8c8 c1f4 g8f6 b3b7 f5d7 h2h3 g7g5 f4h2 g5g4 h3g4 f6g4 b7b3 e7e5 e2e3 e5d4 e3d4 g4h2 h1h2 f8g7 f1b5 e8g8 b3d5 a7a6 b5a4 d8b6 d5b3 b6b3 a2b3 d7g4 a4c6 c8c6 f3e5 f8e8 h2h4 h7h5 f2f3 g7e5 d4e5 e8e5 e1f2 g4e6 b3b4 g8g7 h4d4 e5g5 a1d1 g5g6 d4e4 c6c4 e4c4 e6c4 d1d2');
INSERT INTO evaluations VALUES(31307,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(31307,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(31307,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(31307,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(31307,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(31307,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(31307,36,'cp 124','b8c6 b1c3 a8c8 c1f4 g8f6 b3b7 f5d7 h2h3 g7g5 f4h2 g5g4 h3g4 f6g4 b7b3 e7e5 e2e3 g4h2 h1h2 e5d4 e3d4 f8g7 f1b5 e8g8 b3d5 a7a6 b5a4 d8b6 d5b3 b6b3 a2b3 d7g4 a4c6 c8c6 f3e5 f8e8 h2h4 g4c8 e1c1 c6b6 e5c4 b6b3 c1c2 b3b8 c4d6 e8e6 d6c8 b8c8 d4d5 e6f6 d1d2 f6d6 h4a4 g8f8 g2g4 h7h6 f2f3');
INSERT INTO evaluations VALUES(51659,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(51659,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(51659,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(51659,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(51659,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(51659,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(51659,36,'cp 202','a4b5 e7e6 f1c4 f8e7 d1e2 g8f6 c1d2 e8g8 g1f3 c6b5 c4b5 c8b7 e1g1 a7a6 b5c4 b8c6 b1c3 a6a5 c4b5 c6b4 f1c1 d8b8 h2h3 f6e4 c3e4 b7e4 f3e5 f8c8 c1c3 b8b7 f2f3 e4c6 e5c6 b4c6 b5c6 c8c6');
INSERT INTO evaluations VALUES(51637,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(51637,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(51637,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(51637,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(51637,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(51637,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(51637,36,'cp 199','g8f6 f1c4 e7e6 g1f3 f8e7 e1g1 e8g8 c1d2 c6b5 c4b5 c8b7 d1e2 a7a6 b5c4 b8c6 f1c1 a6a5 d2e1 c6b4 e2d1 d8b8 c4e2 f8c8 c1c8 b8c8 b1c3 e7d6 f3d2 b7c6 h2h3 d6c7 e2f3 c8b7 f3c6 b4c6');
INSERT INTO evaluations VALUES(4879,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(4879,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(4879,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(4879,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(4879,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(4879,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(4879,36,'cp 57','c2c3 g8e7 d2d4 e5d4 c3d4 c5b4 b1c3 d7d5 e1g1 e8g8 e4d5 e7d5 c1g5 d8d6 c3e4 d6e6 a4c2 h7h6 g5h4 f7f5 e4c5 e6f7 c5d3 g7g5 c2b3 g5h4 f3e5 c6e5 d3b4 c8e6 b4d5 e5c6 d5f6 f7f6 b3e6 f6e6 d4d5 e6g6 d5c6 g6c6 f1e1 f8e8');
INSERT INTO evaluations VALUES(4876,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(4876,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(4876,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(4876,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(4876,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(4876,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(4876,36,'cp 50','g8e7 d2d4 e5d4 c3d4 c5b4 b1c3 d7d5 e1g1 e8g8 e4d5 e7d5 c1g5 d5c3 b2c3 b4e7 g5e7 c6e7 f1e1 b7b5 a4b3 c8b7 f3e5 b7d5 h2h4 d8d6 h4h5 h7h6 b3c2 d5e6 a2a4 a8d8 a4b5 a6b5 e5d3 e6c4 d1f3 c4d5 f3f4 d6f4 d3f4 f8e8');
INSERT INTO evaluations VALUES(30779,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(30779,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(30779,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(30779,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(30779,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(30779,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(30779,36,'cp 26','g1f3 e7e6 b1c3 g8f6 e4e5 d7d5 c4f1 f6d7 d2d4 c7c5 c3e2 b6b5 c2c3 b8c6 a2a3 a7a5 e2f4 b5b4 f1d3 b4c3 b2c3 b7a6 d3a6 a8a6 h2h4 d8b6 e1g1 f8e7 c3c4 d5c4 d4d5 c6d4 c1e3 d4f3 d1f3 e8g8 d5e6 f7e6 f3e4 c4c3 g2g3 b6b3 a1b1');
INSERT INTO evaluations VALUES(30777,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(30777,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(30777,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(30777,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(30777,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(30777,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(30777,36,'cp 18','e7e6 b1c3 g8f6 e4e5 d7d5 c4f1 f6d7 d2d4 c7c5 c3e2 b6b5 a2a3 b8c6 c2c3 a7a5 e2f4 b5b4 f1d3 b4c3 b2c3 b7a6 d3a6 a8a6 f4h5 g7g6 h5f4 f8e7 h2h4 d8a8 e1g1 e8g8 f1e1 a5a4 f4d3 h7h5 c1g5 e7g5 h4g5 c5d4 c3d4');
INSERT INTO evaluations VALUES(30626,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30626,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(30626,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(30626,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(30626,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(30626,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(30626,36,'cp 19','b1c3 g8f6 e4e5 d7d5 c4f1 f6d7 d2d4 c7c5 c3e2 b8c6 c2c3 b6b5 a2a3 a7a5 e2f4 b5b4 c1e3 f8e7 h2h4 c5d4 c3d4 a5a4 f1d3 b4a3 b2a3 b7a6 d3a6 d8a5 e1f1 a5a6 f1g1 e8g8 f3g5 f8b8 d1h5 e7g5 h5g5 h7h6 g5g4 c6e7 f4h5 g7g6 h5f4');
INSERT INTO evaluations VALUES(55059,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(55059,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(55059,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(55059,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(55059,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(55059,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(55059,36,'cp -69','e7e6 e2e3 c7c5 c2c3 f8d6 f3f4 b8d7 g1f3 g8e7 a2a4 b7b6 f1d3 e8g8 f3e5 f7f6 d1h5 e7f5 e5f3 f5h6 e1g1 c8b7 b1d2 a8c8 f1e1 f6f5 f3g5 d8e7 h5e2 c5c4 d3c2 d7f6 d2f3 f6e4 c1d2 a7a5 h2h3 h6f7 g5f7 f8f7');
INSERT INTO evaluations VALUES(55864,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(55864,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(55864,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(55864,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(55864,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(55864,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(55864,36,'cp -57','e2e3 c7c5 f3f4 g8f6 c2c3 f8d6 f1d3 e8g8 g1f3 b7b6 e1g1 c8a6 d3a6 b8a6 c1d2 a6c7 d2e1 c7e8 e1h4 d6e7 b1d2 e8d6 d1e2 f6e4 h4e7 d8e7 d2e4 d6e4 f3d2 e4d6 a2a4 e7c7 g2g4 c5d4 e3d4 c7d7 a1e1 a8c8 e2g2 d7a4 f4f5');
INSERT INTO evaluations VALUES(23428,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(23428,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(23428,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(23428,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(23428,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(23428,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(23428,36,'cp 0','e7e5 g2g3 e5e4 f3h4 g8f6 f1g2 g7g5 h4f5 d7d5 c4d5 d8d5 f5e3 d5e5 b1c3 c8e6 h2h4 h8g8 h4g5 g8g5 d2d3 g5g3 f2g3 e5g3 e1f1 g3f4 f1e1 f4g3');
INSERT INTO evaluations VALUES(55031,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(55031,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(55031,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(55031,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(55031,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(55031,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(55031,36,'cp -31','g8f6 g1f3 e7e5 f3e5 f8b4 c2c4 e8g8 e2e3 f8e8 c4d5 f6d5 f1d3 h7h6 a2a3 b4f8 d1c2 c8e6 e1g1 b8d7 e5f3 c7c5 b1c3 d5f6 a3a4 a8c8 d3e2 d7b8 d2d4 b8c6 d4c5 f8c5 c2b1 d8e7 c3b5 a7a6 b5d4 c6d4 b2d4 c5d4');
INSERT INTO evaluations VALUES(47129,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(47129,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(47129,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(47129,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(47129,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(47129,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(47129,36,'cp -24','e2e3 e7e5 b2e5 f8b4 c2c3 b4d6 e5d6 d8d6 d2d4 c8f5 g1f3 e8g8 f1d3 f5g6 d3e2 c7c5 f3h4 b8c6 h4g6 h7g6 b1d2 a7a6 a1b1 b7b5 a2a4 b5b4 d4c5 d6c5 c3b4 c6b4 b1c1 c5e7 e1g1 f8c8 c1c8 a8c8 d2f3 a6a5 f3d4 e7e5');
INSERT INTO evaluations VALUES(47127,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(47127,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(47127,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(47127,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(47127,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(47127,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(47127,36,'cp -27','e7e5 b2e5 f8b4 c2c3 b4d6 e5d6 d8d6 d2d4 c7c5 f1d3 b8c6 g1e2 c5c4 d3c2 b7b5 b1d2 a8b8 e1g1 e8g8 a2a3 a7a5 h2h3 f8e8 e2g3 b5b4 a3b4 a5b4 c3b4 c8d7 b4b5 b8b5 g3e2 c6b4 e2c3 b5b8 f1e1 h7h6 d2f3 b4c2 d1c2 f6e4 c3e4 e8e4 e1b1 e4e8 b1b8 e8b8');
INSERT INTO evaluations VALUES(57291,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(57291,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(57291,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(57291,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(57291,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(57291,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(57291,36,'cp 59','c2c4 g8f6 b1c3 f8b4 g1e2 c7c6 d4d5 c6d5 c4d5 e8g8 c1d2 f8e8 e2g3 d7d6 f1e2 b4c5 e1g1 d8e7 d1b1 g7g6 e2d1 c5b4 c3e4 b4d2 e4d2 f6d5 d2e4 d5b6 d1b3 c8e6 f1d1 e6b3 a2b3 d6d5 e4c3 e8d8 a1a5 b8c6 c3d5 b6d5 a5d5');
INSERT INTO evaluations VALUES(57278,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(57278,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(57278,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(57278,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(57278,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(57278,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(57278,36,'cp 63','g8f6 b1c3 c7c6 d4d5 f8b4 g1e2 c6d5 c4d5 e8g8 c1d2 d7d6 e2g3 f8e8 f1e2 b4c5 e1g1 d8e7 d1b1 g7g6 b2b4 c5b6 e2d1 a7a5 a2a3 a5b4 a3b4 a8a1 b1a1 b8a6 d1a4 e8d8 a1b1 a6c7 a4c2 c7d5 c3d5 f6d5 c2e4 d5f6 e4f3');
INSERT INTO evaluations VALUES(49748,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(49748,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(49748,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(49748,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(49748,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(49748,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(49748,36,'cp 77','b1c3 f8b4 g1e2 c7c6 d4d5 c6d5 c4d5 e8g8 e2g3 f8e8 c1d2 d7d6 f1e2 b4c5 e1g1 b8d7 d1b1 d8e7 b2b4 c5b6 a2a4 a7a6 a4a5 b6d8 e2d1 b7b5 a5b6 d7b6 d1c2 g7g6 f1c1 c8b7 c2e4 a8c8 e4f3 b6d5 c3d5 c8c1 b1c1 f6d5 c1b2 d5b6 d2c3');
INSERT INTO evaluations VALUES(49746,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(49746,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(49746,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(49746,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(49746,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(49746,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(49746,36,'cp 78','f8b4 g1e2 c7c6 d4d5 c6d5 c4d5 e8g8 e2g3 f8e8 f1e2 d7d6 e1g1 b8d7 c1d2 b4c5 d1b1 d8e7 b2b4 c5b6 a2a4 a7a6 a4a5 b6d8 e2d1 b7b5 d1c2 d7e5 g3e4 e5c4 d2e1 f6e4 c3e4 c8b7 e4d2 b7d5 c2h7 g8f8 d2c4 d5c4 h7d3 d6d5 d3c4 d5c4 e1c3');
INSERT INTO evaluations VALUES(42838,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(42838,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(42838,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(42838,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(42838,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(42838,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(42838,36,'cp 67','g1e2 c7c6 d4d5 c6d5 c4d5 e8g8 e2g3 f8e8 f1e2 b7b6 e1g1 c8a6 c1d2 a6e2 c3e2 b4d2 d1d2 e8e5 e2c3 d7d6 d2c2 b8a6 a1d1 d8e7 d1d4 a8e8 h2h3 h7h5 f1d1 h5h4 g3e2 a6c7 b2b4 a7a6 c2b3 g7g5 a2a4 b6b5 d1c1 g5g4 e2f4 b5a4 b3a4 g4h3 g2h3 c7b5 c3b5');
INSERT INTO evaluations VALUES(22975,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(22975,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(22975,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(22975,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(22975,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(22975,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(22975,36,'cp -42','f8c5 b1c3 d7d6 g1f3 g8e7 a4c4 e8g8 f1e2 c8e6 c4d3 c6e5 f3e5 d6e5 d3g3 e7g6 c1e3 d8d6 e1g1 c7c6 f1d1 d6e7 a2a4 a8d8 a4a5 d8d1 c3d1 f8d8 e3g5 f7f6 g5e3 e7d6 e3c5 d6c5 g3c3 c5c3 d1c3 d8d2 e2d3');
INSERT INTO evaluations VALUES(6579,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(6579,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(6579,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(6579,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(6579,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(6579,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(6579,36,'cp -44','g1f3 d7d6 b1c3 g8e7 c1g5 e8g8 e1c1 f7f6 g5e3 c5e3 f2e3 g8h8 f1e2 c8d7 a4a3 a7a6 c3d5 f8e8 h1f1 a8c8 h2h3 e7g8 a3a4 g8h6 a4b3 c8b8 f3d2 c6e5 e2h5 e8f8 d2f3 h6f7 h5f7 e5f7 d5f4 d8e8 f4d5 c7c6 d5b6 d7e6 b3c3 e8e7 f3d4 f7e5 d4e6');
INSERT INTO evaluations VALUES(44553,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(44553,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(44553,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(44553,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(44553,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44553,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(44553,36,'cp 65','d4d5 b7b5 e2e4 f6e4 f1d3 e4f6 d3b5 g7g6 b1c3 f8g7 e1g1 e8g8 f1e1 d7d6 d1e2 a7a6 b5a4 c8d7 a4b3 f8e8 h2h3 a6a5 a2a4 b8a6 c1g5 h7h6 g5h4 a6b4 a1d1 d8c7 b3c4 e7e6 d5e6 d7e6 c4e6 e8e6 e2d2 d6d5 e1e6 f7e6');
INSERT INTO evaluations VALUES(44593,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(44593,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(44593,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(44593,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(44593,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(44593,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(44593,36,'cp 53','b7b5 e2e4 f6e4 f1d3 e4f6 d3b5 g7g6 e1g1 f8g7 b1c3 e8g8 f1e1 d7d6 d1e2 a7a6 b5a4 c8d7 a4b3 f8e8 c1f4 a6a5 a2a4 b8a6 a1d1 a6c7 b3c4 a8b8 b2b3 f6h5 f4d2 d7f5 c3e4 h5f6 e4f6 e7f6 d2e3 f5e4 e2d2 e4f3 g2f3 b8a8 g1g2 e8e5 f3f4');
INSERT INTO evaluations VALUES(43916,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(43916,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(43916,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43916,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(43916,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(43916,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(43916,36,'cp 69','e2e4 f6e4 f1d3 d8a5 b1d2 e4f6 e1g1 c5c4 d3e4 d7d6 f3d4 g7g6 a2a4 c8d7 e4f3 f8g7 c2c3 e8g8 b2b4 c4b3 d2b3 a5b6 d4b5 d7b5 c1e3 b6b7 a4b5 b7b5 b3a5 f6d7 d1d2 d7c5 f1b1 b5d3 d2d3 c5d3 a5c6 b8c6 d5c6 d3e5 b1b7 f8c8 c6c7 e5f3 g2f3 g7c3 a1a7 a8a7 e3a7');
INSERT INTO evaluations VALUES(52123,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(52123,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(52123,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(52123,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(52123,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(52123,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(52123,36,'cp 101','f1c4 d7d6 d2d3 b8c6 g1e2 g8f6 f2f4 a7a6 e1g1 f8e7 a2a4 c8d7 e2g3 e5f4 c1f4 e8g8 h2h3 c6d4 c3d5 f6d5 c4d5 d4e6 f4e3 e6c7 d5a2 e7g5 d1d2 g5e3 d2e3 g7g6 f1f3 d7e6 a2e6 c7e6 a1f1 d6d5 f3f6 d5e4 g3e4');
INSERT INTO evaluations VALUES(52099,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(52099,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(52099,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(52099,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(52099,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(52099,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(52099,36,'cp 100','d7d6 g1e2 g8f6 d2d3 f8e7 e1g1 b8c6 f2f4 a7a6 a2a4 e5f4 e2f4 e8g8 h2h3 a8b8 g1h1 c8d7 b2b3 c6d4 a4a5 d4c6 c3d5 f6d5 f4d5 e7g5 c1d2 d7e6 d2c3 h7h6 d5b6 d8e7 d1e2 c6d4 c3d4 c5d4 e2f2 g5h4 f2d4');
INSERT INTO evaluations VALUES(53873,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(53873,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(53873,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(53873,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(53873,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(53873,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(53873,36,'cp 101','g1e2 b8c6 d2d3 g8f6 f2f4 a7a6 a2a4 f8e7 e1g1 c8d7 e2g3 e5f4 c1f4 e8g8 d1d2 g8h8 h2h3 c6d4 a4a5 a8c8 g1h1 d4c6 c3d5 f6d5 c4d5 d7e6 g3h5 d8d7 d2f2 g7g6 h5g3 c6b4 f4h6 b4d5 e4d5');
INSERT INTO evaluations VALUES(14301,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(14301,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14301,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(14301,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(14301,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(14301,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(14301,36,'cp 46','f1g2 c8f5 c2c4 e7e6 e1g1 f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 c6e5 d4e5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c7c6 d5c4 a8d8 c1e3 b7b5 c4d3 f5d3 e2d3 d8d3 a1d1 d3d1 f1d1 f8d8 d1c1 d8c8 g1g2 f7f6 e5f6 g7f6 c1d1 g8f7 d1d7 f7e6 d7b7 a5a4 b2b3 a4b3 a2b3 b5b4 g2f3 c6c5');
INSERT INTO evaluations VALUES(18875,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(18875,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(18875,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(18875,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(18875,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(18875,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(18875,36,'cp -32','a7a6 e1g1 b7b5 c4d3 c8b7 c1f4 a8c8 d1e2 f8c5 a1c1 c6e7 f4d2 e7g6 b2b4 c5e7 a2a4 g8f6 a4b5 a6b5 c3b5 c8c1 f1c1 e8g8 b5d4 d8b8 d4b3 g6f4 d2f4 b8f4 g2g3 f4b8 b4b5 b7a8 c1a1 e7d8 f3d2 h7h6 e4e5 f6d5 d2e4 d8e7 e4d6 e7d6 e5d6 b8d6');
INSERT INTO evaluations VALUES(17413,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(17413,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(17413,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(17413,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(17413,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(17413,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(17413,36,'cp -13','e1g1 b7b5 c4d3 g8e7 c1e3 e7g6 a2a4 b5a4 c3a4 a8b8 d1e2 f8e7 d3a6 c8a6 e2a6 c6b4 a6e2 e8g8 g2g3 h7h6 f1d1 d7d6 h2h4 b4c6 a4c5 e7f6 c5d3 d8c7 a1c1 c7b7 c1c2 f8d8 d1c1 g6e5 f3e5 c6e5 e3f4 e5d3 e2d3 b7a8 f4d6');
INSERT INTO evaluations VALUES(17412,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(17412,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(17412,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(17412,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(17412,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(17412,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(17412,36,'cp -16','g8e7 c1e3 b7b5 c4d3 e7g6 a2a4 b5a4 c3a4 a8b8 d1e2 f8e7 d3a6 c8a6 e2a6 c6b4 a6e2 e8g8 f1d1 d7d6 g2g3 b4c6 a4c5 e7f6 c5d3 h7h6 h2h4 d8c8 d1c1 c8b7 c1c2 c6b4 d3b4 b7b4 e3d2 b4b3 d2c3 f8c8 a1c1 f6c3 b2c3 b3b6 c2d2 b6c7');
INSERT INTO evaluations VALUES(11115,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(11115,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(11115,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(11115,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11115,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11115,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11115,36,'cp 0','c1e3 b7b5 c4b3 c6a5 a1c1 e7c6 c3d5 e6d5 e4d5 a5b3 a2b3 f8e7 d5c6 d7c6 f3d4 e7f6 d4c6 d8d1 f1d1 e8g8 e3d4 f6g5 d4e3');
INSERT INTO evaluations VALUES(52301,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(52301,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52301,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(52301,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(52301,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52301,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52301,36,'cp 36','d7d5 e4d5 d8d5 d2d4 f8g7 b1d2 c5d4 f1c4 d5h5 d1b3 g8h6 f3d4 e8g8 c4e2 h5c5 e1g1 b8c6 d4c6 c5c6 e2f3 c6c7 b3c4 c7b8 a2a4 h6f5 a4a5 f5d6 c4c5 d6f5 a5a6 f5h4 c5e7 h4f3 d2f3 b7a6 f1d1 c8b7 d1d7');
INSERT INTO evaluations VALUES(52648,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(52648,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52648,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(52648,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52648,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(52648,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52648,36,'cp 43','e4d5 d8d5 d2d4 f8g7 b1d2 c5d4 f1c4 d5c5 d1b3 g8h6 d2e4 c5b6 f3d4 e8g8 e1g1 b8c6 b3b6 a7b6 d4b5 c8f5 e4g3 c6e5 g3f5 h6f5 c4e2 h7h5 f1d1 f8d8 d1d8 a8d8 c1f4 g7f6 f4e5 f6e5 a1d1 d8d1 e2d1 f5d6 b5d6 e5d6 d1f3 e7e6 g2g3 h5h4 g3h4 d6e5 f3b7 e5f6 h4h5 g6h5 b7f3');
INSERT INTO evaluations VALUES(52647,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(52647,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52647,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52647,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(52647,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52647,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(52647,36,'cp 36','d8d5 d2d4 g8f6 c3c4 d5e4 f1e2 c5d4 b1d2 e4c6 e1g1 c6d7 d2b3 d4d3 d1d3 d7d3 e2d3 f8g7 b3d4 e8g8 f1e1 f8d8 d3f1 g8f8 c1f4 c8d7 a1d1 b8c6 b2b3 c6d4 f3d4 d7g4 d1d2 g4d7 h2h3 d7e8 g2g3 a8c8');
INSERT INTO evaluations VALUES(34280,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(34280,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(34280,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(34280,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(34280,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(34280,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(34280,36,'cp 39','d2d4 f8g7 b1d2 c5d4 f1c4 d5c5 d1b3 g8h6 d2e4 c5b6 f3d4 e8g8 e1g1 h6f5 d4f5 c8f5 e4g3 f5d7 f1e1 b8c6 b3b6 a7b6 c4f1 d7e6 a2a3 g7e5 c1e3 e5c7 g3e2 f8d8 h2h3 a8c8 e2d4 c6d4 e3d4 d8d7 a1d1');
INSERT INTO evaluations VALUES(34279,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(34279,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(34279,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(34279,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(34279,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(34279,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(34279,36,'cp 47','f8g7 b1d2 c5d4 f1c4 d5d8 f3d4 g8f6 e1g1 e8g8 f1e1 d8c7 d4b5 c7d8 d2b3 b8c6 c1f4 c8g4 d1d8 f8d8 h2h3 g4f5 b5d4 c6d4 b3d4 f5e4 c4f1 e7e6 f4e5 e4d5 c3c4 d5c6 d4c6 b7c6 e1d1 f6d7 e5g7');
INSERT INTO evaluations VALUES(33022,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33022,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(33022,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(33022,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(33022,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(33022,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(33022,36,'cp 28','b1d2 c5d4 f1c4 d5d8 f3d4 g8f6 e1g1 e8g8 d2f3 d8c7 c4b3 b8d7 d1e2 d7c5 f1e1 e7e6 d4b5 c7b6 c1f4 c8d7 b5d6 f6h5 f4e3 b6d6 a1d1 d6c7 e3c5 c7c5 d1d7 b7b6 h2h3 a7a5 f3e5 g7e5 e2e5 c5e5 e1e5 h5f4');
INSERT INTO evaluations VALUES(39604,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(39604,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(39604,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(39604,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(39604,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(39604,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(39604,36,'cp 26','d7d5 e4d5 d8d5 b1d2 c7c5 g1f3 c5d4 f1c4 d5d8 f3d4 g8f6 d2f3 e8g8 e1g1 d8c7 c4b3 b8d7 f1e1 d7c5 d1e2 e7e6 d4b5 c7b6 c1f4 c8d7 b5d6 f6h5 f4e3 b6d6 a1d1 d6c7 e3c5 c7c5 d1d7 b7b6 h2h3 a7a5 d7b7');
INSERT INTO evaluations VALUES(38472,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(38472,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(38472,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(38472,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(38472,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(38472,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(38472,36,'cp 44','e4d5 d8d5 b1d2 c7c5 g1f3 c5d4 f1c4 d5c5 d1b3 g8h6 d2e4 c5b6 f3d4 e8g8 e1g1 h6f5 d4f5 c8f5 e4g3 f5d7 f1e1 e7e6 h2h4 h7h6 h4h5 g6g5 c1e3 b6b3 a2b3 a7a6 b3b4 b7b5 c4b3 d7c6 e1d1 b8d7 d1d6 f8c8 a1d1 d7f6 f2f3');
INSERT INTO evaluations VALUES(56806,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(56806,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(56806,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(56806,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(56806,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(56806,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(56806,36,'cp 112','f1d3 g7g6 g1f3 c7c5 e1g1 f8g7 c2c4 c5d4 c4d5 f6e5 f1e1 d8d5 b2b4 g8e7 a2a4 e8g8 b1d2 d5d8 c1a3 b8d7 f3g5 d7b6 b4b5 b6d5 d2c4 h7h6 g5e4 b7b6 a4a5 b6a5 a3d6 d5f4 c4a5 c8d7 a5b7 d8e8 b7c5 f4d3 c5d3 d7b5 d3e5 d4d3');
INSERT INTO evaluations VALUES(5405,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(5405,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(5405,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(5405,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(5405,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(5405,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(5405,36,'cp 64','c2c4 d7d6 b1c3 g7d4 e3d4 c6d4 d1d4 g8f6 f1e2 e8g8 e1g1 c8d7 f1d1 d8a5 c3d5 f6d5 e4d5 a8c8 b2b3 f8e8 a2a4 e7e5 d5e6 e8e6 e2f1 h7h5 d4d2 a5d2 d1d2 g8f8 a1d1 f8e7 g2g3 a7a5 f1h3 f7f5 h3g2 d7c6 h2h4 c8e8 g2d5 c6d5');
INSERT INTO evaluations VALUES(9838,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9838,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(9838,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(9838,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(9838,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9838,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(9838,36,'cp 21','e1g1 b4c3 b2c3 d7d6 c4b3 e8g8 f1e1 a7a6 h2h3 b7b5 f3h2 c8e6 b3e6 f7e6 a2a4 c6e7 h2g4 a8b8 a4b5 a6b5 c1e3 e7g6 g2g3 d8d7 a1a6 d7c8 a6a3 c8d7 d1e2 f6g4 e2g4 g8h7 a3a6 g6e7 g1g2 e7g8 g4e2 g8f6 e1a1 b8b7');
INSERT INTO evaluations VALUES(9837,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(9837,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9837,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(9837,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(9837,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9837,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(9837,36,'cp 6','b4c3 b2c3 e8g8 f1e1 f8e8 c1d2 d7d6 h2h3 b7b6 c4b5 c8d7 b5a6 d7c8 a6c8 a8c8 c3c4 f6h7 f3h2 h7f8 f2f4 e5f4 d2f4 f8g6 f4d2 c6e5 a2a4 d8h4 a4a5 h4g3 h2f1');
INSERT INTO evaluations VALUES(9854,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(9854,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(9854,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(9854,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9854,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(9854,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9854,36,'cp 23','b2c3 e8g8 f1e1 d7d6 h2h3 c6a5 c4b3 b7b5 c1e3 c8d7 f3d2 f8e8 f2f4 a5b3 a2b3 e5f4 e3f4 a7a5 f4g3 f6h7 c3c4 a5a4 c4b5 d8g5 g1h2 a4b3 d2b3 g5b5 e4e5 a8a1 d1a1 b5c6 c2c4 h7g5 e5d6 e8e1 a1e1 c7d6 b3d4 c6c5 e1a1 g5e6 a1a8 g8h7');
INSERT INTO evaluations VALUES(9851,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(9851,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(9851,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(9851,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9851,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(9851,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9851,36,'cp 4','d7d6 f1e1 e8g8 h2h3 c6a5 c4b3 b7b5 c1e3 f8e8 f3d2 c7c5 d1f3 c8d7 a2a4 a5b3 c2b3 b5a4 b3a4 d8e7 d3d4 e5d4 c3d4 a8d8 d4c5 d6c5 f3g3 f6e4 d2e4 e7e4 e3c5 e4f5 c5d4 e8e1 a1e1 f5g5 g3g5 h6g5 d4a7 d8a8 a7b6 d7a4 e1a1 a4c6 a1a8 c6a8 b6d4 a8c6 d4e3 f7f6 f2f3 g8h7');
INSERT INTO evaluations VALUES(42490,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(42490,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(42490,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(42490,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(42490,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(42490,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(42490,36,'cp -25','b4c3 b2c3 d7d6 e2e4 f6e4 d1g4 e4f6 g4g7 h8g8 g7h6 d8e7 g1e2 b8c6 f2f3 c6a5 e2g3 a5b3 a1b1 b3c1 b1c1 c8d7 f1d3 e8c8 e1g1 d8f8 h6h4 e7d8 g3e4 f6e8 h4d8 c8d8 c1b1 b7b6 e4g3 f7f5 g3h5 e8g7 h5g7 g8g7 g1f2 c7c5 b1e1 f8g8');
INSERT INTO evaluations VALUES(42511,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42511,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42511,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42511,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(42511,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(42511,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(42511,36,'cp 0','b2c3 d7d6 e2e4 f6e4 d1g4 e4f6 g4g7 h8g8 g7h6 d8e7 g1e2 b8c6 e2g3 e6e5 d4d5 c6a5 a1b1 g8g6 h6e3 b7b6 f1d3 c8a6 e1g1 a6c4 d3c4 a5c4 e3d3 c4a5 f2f4 e5e4 d3b5 e7d7 b5e2 d7g4 e2a6 g4c8 a6e2 c8g4');
INSERT INTO evaluations VALUES(42510,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42510,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(42510,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42510,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42510,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42510,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(42510,36,'cp -24','b7b6 d1c2 c8b7 f2f3 b8c6 e2e4 c6a5 f1d3 b7a6 c2e2 c7c5 d4d5 d7d6 g1h3 e6d5 c4d5 a6d3 e2d3 c5c4 d3c2 a5b3 a1b1 b3c1 c2c1 e8g8 e1g1 a7a6 h3f4 b6b5 f4e2 f6d7 e2d4 d7c5 d4f5 g7g6 f5d4 c5d3 c1d2 d8f6 d4c6');
INSERT INTO evaluations VALUES(39837,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(39837,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(39837,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(39837,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(39837,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(39837,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(39837,36,'cp -28','f2f3 b8c6 e2e4 c6a5 c4c5 b6c5 d1a4 a5b7 c1g5 h7h6 g5h4 c5d4 e4e5 g7g5 h4f2 f6d5 c3d4 c7c5 h2h4 c5d4 f2d4 g5h4 g1e2 d8a5 e2c3 a5a4 c3a4 b7a5 d4c5 a5b3 a1b1 b3c5 a4c5 h4h3 h1h3 f7f6 e5f6 d5f6 e1f2 e8e7 c5e4 f6e4 f3e4 e6e5 h3h4 d7d6 f1e2');
INSERT INTO evaluations VALUES(32906,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(32906,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(32906,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(32906,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(32906,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(32906,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(32906,36,'cp 120','h2h3 g4h5 g2g4 h5g6 d2d4 e7e6 f3e5 b8c6 e5c4 a5b4 c1e3 e8c8 a2a3 b4e7 f1g2 e7e8 d1f3 g8f6 c4e5 f6d5 e5g6 h7g6 c3d5 e6d5 e1c1 c6a5 c1b1 f8e7 g4g5 c8b8 h3h4 a5c4 e3c1 c7c6 g2f1 c4d6 f1d3 b8a8 d1e1 a7a6 c1f4 e8d7 f4e5');
INSERT INTO evaluations VALUES(32905,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(32905,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(32905,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(32905,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(32905,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(32905,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(32905,36,'cp 129','g4h5 d2d4 e7e6 g2g4 h5g6 f3e5 f8b4 f1g2 a5a6 d1e2 c7c6 a2a3 a6e2 e1e2 b4c3 b2c3 g8e7 e5g6 h7g6 c3c4 b8d7 a3a4 e6e5 c2c3 f7f6 a4a5 e8c8 h3h4 c8c7 c1d2 a7a6 a1b1 b7b6 a5b6 d7b6 c4c5');
INSERT INTO evaluations VALUES(32899,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(32899,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(32899,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(32899,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(32899,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(32899,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(32899,36,'cp 144','g2g4 h5g6 d2d4 e7e6 f3e5 b8c6 e5c4 a5b4 c1e3 e8c8 a2a3 b4e7 f1g2 e7e8 b2b4 h7h5 b4b5 c6e7 d1f3 e7d5 e1g1 h5g4 h3g4 f7f5 c3d5 e6d5 e3f4 e8e4 f3g3 e4d4 c4e5 g6e8 g4g5 f8d6 f1d1 d4a4 g2d5 g8e7 d5e6 c8b8 c2c4 e8b5 c4b5 a4e4 f2f3 d6c5 g1f1 h8h1 f1g2 h1g1 d1g1 e4e2 g2h3 d8h8 g3h4 h8h4 h3h4 e7g6 e5g6 c5g1 a1g1 e2e6');
INSERT INTO evaluations VALUES(32898,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(32898,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(32898,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(32898,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(32898,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(32898,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(32898,36,'cp 125','h5g6 d2d4 e7e6 f3e5 b8c6 e5c4 a5b4 c1e3 e8c8 a2a3 b4e7 f1g2 e7e8 b2b4 h7h5 b4b5 c6e7 g4g5 e7d5 d1d2 g8e7 e1g1 e7f5 c4a5 e8d7 c3e2 f5h4 g2h1 g6f5 c2c4 f5h3 e2g3 h3f1 a1f1 h4g6 c4d5 e6d5 d2d3 h5h4 g3f5 g6e7 f5e7 f8e7 h1f3 e7d6');
INSERT INTO evaluations VALUES(32840,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(32840,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(32840,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(32840,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(32840,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(32840,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(32840,36,'cp 147','d2d4 e7e6 f3e5 b8c6 e5c4 a5b4 c1e3 e8c8 a2a3 b4e7 f1g2 e7e8 b2b4 h7h5 b4b5 c6e7 g4g5 e7d5 d1d2 g8e7 e1g1 e7f5 c4a5 f5h4 g2h1 g6f5 e3f4 h4g6 c3e2 f5h3 c2c4 h3f1 g1f1 h5h4 a1b1 e8d7 c4d5 e6d5 d2d3 d7g4 f4d2 g4e6 b1e1 h4h3');
INSERT INTO evaluations VALUES(57140,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(57140,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(57140,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(57140,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(57140,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(57140,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(57140,36,'cp -42','e5e4 f3d2 d7d5 c2c4 c7c6 b1c3 g8f6 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 b8d7 d1b3 b4c3 b2c3 e8g8 e1g1 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 f8c8 c4f7 g8f7 f3e5 d7e5 d4e5 g7g6 e5f6 c8c2 f1f2 a8c8 c1d2 f5e4 h2h3 c2b2 a2a3 c8c2 a1d1 a7a6 d2e1 c2f2 e1f2');
INSERT INTO evaluations VALUES(57293,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(57293,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(57293,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(57293,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(57293,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(57293,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(57293,36,'cp -25','f3d2 g8f6 c2c4 c7c6 b1c3 d7d5 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 e8g8 e1g1 b8d7 d1b3 b4c3 b2c3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 f8c8 c4b3 f5d3 f1e1 a7a5 a2a4 d3c4 b3d1 f6e4 f3d2 c4d5 d1b3 e4c3 e3e4 d5b3 d2b3 d7f6 b3c5 b7b6 c1d2 b6c5 d2c3 c5d4');
INSERT INTO evaluations VALUES(57294,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(57294,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(57294,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(57294,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(57294,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(57294,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(57294,36,'cp -29','d7d5 c2c4 c7c6 b1c3 g8f6 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 e8g8 e1g1 b8d7 d1b3 b4c3 b2c3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 f8c8 f3d2 f5g6 f1e1 d7b6 c4e2 b6d5 e2f3 d5b4 f3b7 b4c2 b7c8 a8c8 c1b2 c2a1 e1a1 c8e8 e3e4 g6e4 d2e4 e8e4 a1b1');
INSERT INTO evaluations VALUES(56234,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56234,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(56234,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(56234,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(56234,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56234,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(56234,36,'cp -13','c2c4 c7c6 b1c3 g8f6 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 b4c3 b2c3 e8g8 e1g1 b8d7 d1b3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 f8c8 c4b3 f5d3 f1e1 a7a5 f3e5 d7e5 d4e5 a5a4 b3a4 a8a4 e5f6 a4c4 e3e4 d3e4 f6g7 g8g7 c1d2 g7g6 a1c1 c4c2 c1c2 c8c2 e1e4 c2d2 e4e7 d2a2');
INSERT INTO evaluations VALUES(56188,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(56188,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(56188,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(56188,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(56188,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(56188,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(56188,36,'cp -24','c7c6 b1c3 g8f6 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 e8g8 e1g1 b8d7 d1b3 b4c3 b2c3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 a8c8 c4b3 f5d3 f1e1 f8e8 c1a3 d3c4 f3d2 c4b3 a2b3 a7a6 e3e4 c8c2 a1d1 c2c3 h2h3 c3d3 e4e5 f6d5 d2e4 d3b3');
INSERT INTO evaluations VALUES(53813,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(53813,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(53813,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(53813,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(53813,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(53813,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(53813,36,'cp 100','d2d4 b8d7 d4e5 d6e5 e1g1 f8c5 f3g5 g8h6 d1h5 e8g8 g5f3 d8f6 a2a4 g8h8 b1d2 a7a5 h2h3 h6g8 d2b3 c5b6 c1g5 f6g6 h5g6 h7g6 f1e1 f7f6 g5e3 b6e3 e1e3 g8h6 f3d2 h6f7 c4e2 f7d6 f2f3 f8d8 e3c3 d7f8');
INSERT INTO evaluations VALUES(53814,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(53814,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(53814,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(53814,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(53814,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(53814,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(53814,36,'cp 107','b8d7 e1g1 e5d4 d1d4 d8f6 d4e3 f8e7 b1c3 g8h6 h2h3 d7e5 f3e5 d6e5 f1d1 c8e6 c4b3 e8g8 e3e2 b7b5 c1e3 a7a6 c3b1 e6b3 a2b3 g8h8 d1d7 h6g8 e2g4 g8h6 g4f3 f6e6 d7c7 e7d8 c7a7 a8a7 e3a7 a6a5 a7c5 f8e8 b1d2');
INSERT INTO evaluations VALUES(18537,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(18537,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(18537,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(18537,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(18537,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(18537,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(18537,36,'cp 112','e1g1 e5d4 d1d4 d8f6 d4e3 f8e7 b1c3 g8h6 h2h3 e8g8 a2a4 f8e8 f1e1 g8h8 a4a5 d7e5 f3e5 d6e5 c3d1 h6g8 e3b3 e8f8 f2f3 f6g6 d1f2 g8f6 c1e3 f6h5 g2g4 h5f4 e3f4 e5f4 f2d3 a8b8 g1g2 e7h4 e1e2 h8g8 d3e5 g6f6 b3c3 f6g5 e5d3 h4g3');
INSERT INTO evaluations VALUES(20126,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(20126,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(20126,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(20126,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(20126,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(20126,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(20126,36,'cp 139','b1c3 c8d7 f2f4 d7c6 c1e3 g8f6 e1c1 f6g4 e3g1 e7e5 d4d2 e5f4 d2f4 g4e5 f4g3 g7g6 g1c5 a7a6 h2h4 h7h5 c3d5 c6d5 d1d5 d8c7 f1e2 a8c8 c2c3 b7b5 c5b4 f8e7 b4a3 c8b8 c1b1 e8g8 e2h5 e5c4 h5g6 f7g6');
INSERT INTO evaluations VALUES(20125,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(20125,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(20125,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(20125,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(20125,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(20125,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(20125,36,'cp 107','c8d7 f2f4 d7c6 c1e3 g8f6 e1c1 f6g4 f1c4 g4e3 d4e3 d8b6 e3e2 g7g6 e4e5 e8c8 e5d6 e7d6 c3d5 b6a5 h1e1 c8b8 e2d3 f8g7 e1e7 d8d7 e7d7 c6d7 d5c3 a5c7 d3d6 c7d6 d1d6 d7c6 g2g3 f7f5 d6d3 h8e8 c1d2 a7a6 c4d5 g7c3 d2c3 e8e2 d5c6 b7c6');
INSERT INTO evaluations VALUES(21060,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(21060,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21060,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21060,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(21060,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(21060,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21060,36,'cp 3','e2e3 g8f6 d2d4 c8g4 f1e2 e7e6 e1g1 f8e7 a2a3 e8g8 c2c4 g4f3 g2f3 d5c4 e2c4 f6d5 g1h1 e7f6 b1c3 c6e7 c3d5 e7d5 c4d3 g7g6 d1c2 d5e7 c1d2 f8e8 d3e4 c7c6 f1g1 a7a5 a1d1 f6g7 d2c3 e7d5 c2e2 d5c3 b2c3');
INSERT INTO evaluations VALUES(21059,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21059,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21059,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21059,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(21059,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21059,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21059,36,'cp 4','c8g4 d2d4 g8f6 f1e2 e7e6 e1g1 c6e7 f3e5 g4e2 d1e2 c7c6 c2c4 e7f5 c4c5 f6d7 e5d7 d8d7 b2b4 a7a5 b4a5 f8e7 a2a4 e8g8 g2g4 f5h6 h2h3 f7f6 e2b2 a8a5 b2b3 h6f7 c1d2 a5a7 b1c3 f8a8 a1b1 d7c8 d2e1');
INSERT INTO evaluations VALUES(26634,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26634,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(26634,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26634,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(26634,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(26634,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26634,36,'cp 10','d2d4 e7e6 f1e2 a7a6 e1g1 g8f6 b1d2 c6e7 f3e5 g4e2 d1e2 e7f5 c2c4 h7h5 c4c5 g7g6 b2b4 c7c6 a2a4 f8g7 c1b2 e8g8 a1a3 f8e8 a3b3 e8e7 f1b1 d8e8 b1e1 e8f8 b4b5 a6b5 a4b5 c6b5 b3b5');
INSERT INTO evaluations VALUES(26606,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(26606,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(26606,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(26606,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26606,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26606,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(26606,36,'cp 0','e7e6 c2c4 g8f6 f1e2 f8e7 a2a3 e8g8 e1g1 d5c4 e2c4 c6a5 c4e2 c7c5 d4c5 d8d1 e2d1 g4f3 g2f3 f6d5 b2b4 a5c4 e3e4 d5e3 f1e1 e3d1 e1d1 a7a5 d1d7 f8d8 d7d8 a8d8 g1f2 d8d1 b1c3 d1d3 c3b1 d3d1');
INSERT INTO evaluations VALUES(43187,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(43187,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43187,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43187,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(43187,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43187,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43187,36,'cp 13','g1f3 e8g8 f1e2 c7c5 d4d5 d7d6 b1c3 e7e6 f3d2 b8a6 e3e4 a6c7 e1g1 f8e8 a2a4 b7b6 a1a3 c8d7 e2d3 a7a6 h2h3 e6d5 e4d5 b6b5 a4b5 a6b5 c4b5 a8a3 b2a3 c7d5 c3d5 f6d5 d2e4 d7f5 d3c4 d5b6 c4f7 g8f7 e4d6 f7g8 d6e8 d8e8');
INSERT INTO evaluations VALUES(18400,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(18400,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(18400,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(18400,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(18400,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(18400,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(18400,36,'cp -11','g8f6 f1d3 f8b4 e1g1 e8g8 f2f4 b4d6 g1h1 e5f4 c1f4 d6f4 f1f4 d7d6 d1d2 d8a5 a1d1 a5e5 b2b3 a7a5 d1f1 c8e6 a2a4 f6d7 c3e2 e5c5 e2g3 d7e5 g3f5 e6f5 e4f5 f7f6 b3b4 c5b6 b4b5 d6d5 b5c6 b6c6');
INSERT INTO evaluations VALUES(30749,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(30749,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(30749,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(30749,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(30749,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(30749,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(30749,36,'cp 168','h2h4 h7h5 f1c4 f8g7 g1f3 g8h6 e1g1 h6g4 c1g5 c7c5 c3d5 b8c6 c2c3 d8c8 a1c1 e7e6 d5e3 g4e3 f2e3 f7f6 g5f4 e8g8 c4d3 c6e7 d1e2 c8e8 c1e1 a8c8 a2a4 e8d8 e4e5 c5d4 e3d4 b7f3 f1f3 f6e5 f4e5 g7e5 d4e5 f8f3 e2f3 e7f5 d3f5');
INSERT INTO evaluations VALUES(26604,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(26604,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(26604,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(26604,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(26604,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(26604,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(26604,36,'cp 41','c2c4 e7e6 b1c3 g8f6 e2e3 f8b4 h2h3 g4f3 d1f3 c6a5 c4d5 f6d5 a1c1 e8g8 f4e5 c7c5 d4c5 a5c6 e5d6 d8a5 f1d3 f8d8 a2a3 b4c5 d6c5 a5c5 e1g1 d5c3 c1c3 c5b6 b2b4 a8c8 f3e4 g7g6 f1c1 c6e7 h3h4 c8c3 c1c3 e7f5 c3c5');
INSERT INTO evaluations VALUES(51536,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(51536,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(51536,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(51536,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(51536,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(51536,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(51536,36,'cp 130','e2e4 a7a6 b1d2 e7e6 a2a4 b5b4 f1d3 c8b7 c2c4 d7d6 f2f4 g7g6 g1f3 f8g7 d2b3 g8e7 d1e2 b8d7 c4c5 a6a5 e1g1 e8g8 c1e3 h7h6 f1c1 f7f5 e4e5 d6e5 d4e5 b7f3 g2f3 e7d5 d3b5 g6g5 f4g5 d5e3 e2e3 h6g5 c1d1');
INSERT INTO evaluations VALUES(51537,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(51537,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(51537,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(51537,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(51537,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(51537,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(51537,36,'cp 106','c8b7 b1d2 g8f6 e4e5 f6d5 g1f3 a7a6 a2a4 b5a4 f1d3 e7e6 e1g1 d5b6 d2c4 b6c4 d3c4 b8c6 a1a4 c6b4 c2c3 b4d5 c4d3 f8e7 f3d2 d5f4 d3e4 b7e4 d2e4 f4g6 d1g4 e8g8 f1e1 a6a5');
INSERT INTO evaluations VALUES(30849,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(30849,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(30849,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(30849,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(30849,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(30849,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(30849,36,'cp 120','b1d2 a7a6 a2a4 b5b4 g1f3 e7e6 f1d3 d7d6 e1g1 g8f6 f1e1 f8e7 d2b3 a6a5 c2c3 e8g8 c3b4 a5b4 a4a5 h7h6 h2h3 b7a6 c1e3 b8d7 a1c1 a6d3 d1d3 c7c5 f3d2 d8c7 e4e5 d6e5 d4c5 f6d5 a5a6 d5e3');
INSERT INTO evaluations VALUES(52764,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52764,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52764,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52764,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52764,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52764,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(52764,36,'cp 0','g8f6 c1f4 c8f5 e2e3 e7e6 f1d3 f5g6 e1g1 f8e7 d1e2 f6h5 f4e5 b8d7 h2h3 h5f6 e5h2 d8b6 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(44942,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44942,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44942,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(44942,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44942,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44942,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44942,36,'cp 0','c1f4 b8d7 a2a3 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(44922,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44922,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44922,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44922,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44922,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44922,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44922,36,'cp 0','b8d7 a2a3 f6h5 f4c1 h5f6');
INSERT INTO evaluations VALUES(12240,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12240,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12240,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12240,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12240,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12240,36,'cp 0','a2a3 d8b6 b2b4 f6h5 f4c1 h5f6');
INSERT INTO evaluations VALUES(12765,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(12765,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(12765,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(12765,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(12765,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(12765,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(12765,36,'cp 36','b5c6 d7c6 d2d3 d8c7 a2a4 e7e5 b1a3 c8e6 b2b3 e8c8 d1e2 f8e7 a3c4 f6d7 f3g5 d7f8 g5e6 f8e6 e2g4 g7g6 c1b2 h7h5 g4g3 e7f6 b2e5 h5h4 e5c7 h4g3 c7d8 f6a1 e1d2 e6d8 h1a1');
INSERT INTO evaluations VALUES(12757,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(12757,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(12757,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(12757,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(12757,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(12757,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(12757,36,'cp 39','d7c6 d2d3 d8c7 a2a4 e7e5 b1a3 g7g6 a3c4 c8e6 b2b3 f6d7 e1g1 f7f6 f3h4 e8c8 f2f4 e6c4 b3c4 f8d6 f4f5 g6g5 h4f3 a7a5 c1d2 h7h5 d1e2 b7b6 a1b1 h8h7 g1h1 d8h8 b1b3 c7d8 b3b2 d6c7 f1b1 d8e7 e2f2 g5g4 f3h4 e7f7');
INSERT INTO evaluations VALUES(13415,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(13415,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(13415,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(13415,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(13415,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(13415,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(13415,36,'cp 45','d2d3 d8c7 a2a4 e7e5 b1a3 g7g6 a3c4 c8e6 b2b3 f6d7 e1g1 f7f6 f3h4 e8c8 f2f4 e6c4 b3c4 f8d6 f4f5 g6g5 h4f3 d8g8 g2g3 h7h5 a4a5 d7b8 g1g2 b8a6 h2h4 g5g4 f3d2 c8b8 d2b3 h8h7 c1e3 b8a8 g2h2 c7d8 c2c3 g8h8 d1d2 d8e7 d2b2 e7c7');
INSERT INTO evaluations VALUES(52487,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52487,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52487,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(52487,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52487,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(52487,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52487,36,'cp 9','f1g2 b8c6 c2c4 g7g6 e2e3 f8g7 g1e2 e7e6 d2d4 c5d4 e3d4 g8e7 b1c3 d7d5 c4d5 e7d5 h2h4 h7h6 e1g1 e8g8 c3d5 e6d5 e2f4 g7d4 d1a4 f8e8 f4d5 c8h3 a4b3 h3g2 g1g2 g8g7 c1f4 a8c8 a1d1 d8d7 f4e3 d4e3 d5e3');
INSERT INTO evaluations VALUES(52488,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(52488,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52488,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(52488,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52488,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(52488,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(52488,36,'cp 0','b8c6 c2c4 g7g6 e2e3 f8g7 g1e2 e7e6 d2d4 c5d4 e3d4 g8e7 b1c3 d7d5 c4d5 e7d5 h2h4 e8g8 e1g1 d8b6 c3a4 b6d8');
INSERT INTO evaluations VALUES(19507,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(19507,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(19507,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(19507,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19507,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19507,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(19507,36,'cp 12','c2c4 g7g6 g1f3 f8g7 b1c3 d7d6 e1g1 c8g4 h2h3 g4f3 g2f3 e7e6 a2a3 g8e7 a1b1 e8g8 f3g2 a8c8 d1a4 a7a5 d2d3 b7b6 f1d1 h7h6 e2e3 d6d5 c4d5 e6d5 c3e2 d8d7 d3d4 f8d8 a4b5 c6d4 e2d4 c5d4 b5d7 d8d7 e3d4');
INSERT INTO evaluations VALUES(19410,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19410,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(19410,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(19410,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(19410,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19410,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19410,36,'cp 17','g7g6 b1c3 f8g7 g1f3 d7d6 e1g1 c8g4 h2h3 g4f3 g2f3 e7e6 f3g2 g8e7 a2a3 e8g8 a1b1 d6d5 d2d3 b7b6 c4d5 e6d5 e2e4 d5e4 d3e4 c6d4 b2b4 c5c4 c1b2 a8c8 f2f4 d8d7 g1h2 f8d8 e4e5 d7e6');
INSERT INTO evaluations VALUES(52352,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(52352,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(52352,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(52352,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(52352,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52352,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(52352,36,'cp -11','d7d6 e5d6 e7d6 d2d4 g8f6 f1d3 d6d5 g1f3 f8d6 d1e2 d6e7 e1g1 e8g8 f1e1 e7d6 c1g5 h7h6 g5h4 c8g4 b1d2 b8d7 h2h3 g4f3 d2f3 f8e8 e2e8 d8e8 e1e8 a8e8 h4f6 d7f6 g2g3 g7g6 a1e1 e8e1 f3e1 f6e4 g1g2 f7f5 a2a4 g8f7 c2c3 d6c7 g3g4 f5g4 h3g4');
INSERT INTO evaluations VALUES(52946,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(52946,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(52946,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(52946,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(52946,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(52946,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(52946,36,'cp -9','e5d6 e7d6 d2d4 d6d5 f1d3 g8f6 g1f3 f8d6 d1e2 d6e7 e1g1 e8g8 c1f4 f8e8 e2d1 b8d7 h2h3 d7f8 b1c3 f8g6 f4h2 e7d6 h2d6 d8d6 f1e1 e8e1 d1e1 g6f4 e1d2 c8e6 a1e1 f4d3 d2d3 d6b4 e1b1 h7h6 a2a3 b4c4 d3c4 d5c4 b1e1 a8d8 c3e4 f6e4 e1e4');
INSERT INTO evaluations VALUES(52940,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52940,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52940,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(52940,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52940,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52940,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(52940,36,'cp -5','e7d6 d2d4 d6d5 f1d3 g8f6 g1f3 f8d6 d1e2 d6e7 e1g1 e8g8 c1f4 f8e8 e2d1 d8b6 d1c1 b8d7 h2h3 d7f8 f1e1 f8e6 f4e3 b6c7 c2c4 d5c4 d3c4 e6f8 e3f4 c7b6 b1d2 c8e6 c1c2 f6d5 f4g3 a8d8 c4b3 d5b4 c2c3 b4d5');
INSERT INTO evaluations VALUES(53845,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(53845,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(53845,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(53845,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(53845,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(53845,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(53845,36,'cp 0','d2d4 g8f6 g1f3 d6d5 f1d3 f8d6 d1e2 d6e7 e1g1 e8g8 c1f4 f8e8 e2d1 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(51488,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(51488,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(51488,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(51488,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(51488,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(51488,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51488,36,'cp 12','g8f6 e2e3 c8b7 c2c4 e7e6 b1c3 f8e7 d1c2 e8g8 a1d1 d7d5 g1f3 h7h6 a2a3 c7c5 c4d5 f6d5 d4c5 d8c8 c3d5 b7d5 e3e4 d5b7 f1b5 b6c5 e1g1 b7a6 c2e2 a6b5 e2b5 a7a6 b5e2 b8c6 f4d6 e7d6 d1d6 c8c7 e4e5 c6e7 d6a6 a8d8 f1d1 d8d1 e2d1');
INSERT INTO evaluations VALUES(43874,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43874,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43874,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(43874,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(43874,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(43874,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(43874,36,'cp 15','e2e3 c8b7 c2c4 e7e6 b1c3 f8e7 d1c2 e8g8 a1d1 d7d5 g1f3 h7h6 a2a3 c7c5 c4d5 f6d5 d4c5 d8c8 c3d5 b7d5 e3e4 d5b7 f1b5 b6c5 e1g1 b7a6 c2e2 a6b5 e2b5 a7a6 b5b3 f8d8 d1d8 e7d8 h2h3 b8d7 f1d1 d8e7 f3d2 c8c6 f4g3 e7f6 b3c2 f6d4 d2c4 d7f6');
INSERT INTO evaluations VALUES(43872,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(43872,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(43872,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(43872,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43872,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(43872,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(43872,36,'cp 28','c8b7 c2c4 e7e6 b1c3 f8e7 d1c2 e8g8 a1d1 d7d5 g1f3 h7h6 a2a3 e7d6 f4g3 d6g3 h2g3 c7c5 c4d5 e6d5 f3e5 b8c6 f1b5 d8c8 c2a4 c6e7 e5d7 f6d7 b5d7 c8c7 d7g4 b7c6 a4c2 c7b7 g4f3 a8d8');
INSERT INTO evaluations VALUES(28804,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(28804,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(28804,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(28804,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(28804,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28804,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(28804,36,'cp 30','c2c4 e7e6 b1c3 f8e7 d1c2 e8g8 g1f3 e7b4 f1d3 b4c3 c2c3 d7d6 e1g1 b8d7 c3c2 d8e7 f4g3 h7h6 e3e4 f6h5 a1e1 e6e5 c2c3 a8e8 e1e3 h5g3 h2g3 d7f6 b2b4 e5d4 f3d4 f6e4 d3e4 b7e4 f2f3 c7c5 e3e4');
INSERT INTO evaluations VALUES(21436,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(21436,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(21436,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(21436,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(21436,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(21436,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21436,36,'cp 70','e4e5 f7f6 f1b5 c8d7 g1f3 a7a6 b5c6 d7c6 h2h4 g8h6 c1h6 g7h6 d1d2 f8g7 e1c1 e8g8 d1e1 d8e7 d2e3 a8e8 c1b1 e7f7 c3d1 c6b5 g2g3 b7b6 e5f6 f7f6 f3e5 c7c5 f2f4 e8c8 d1f2 f6f5 e5f3 g8h8 e3e6 f5e6 e1e6 c5d4 e6b6');
INSERT INTO evaluations VALUES(21461,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(21461,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(21461,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(21461,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(21461,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(21461,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(21461,36,'cp 61','f7f6 f1b5 c8d7 g1f3 d8e7 b5c6 d7c6 c3e2 e7f7 c2c3 c6d7 a2a4 g8e7 e2f4 e7c6 f4d3 f7h5 b2b4 a7a6 e1g1 f6e5 d3e5 c6e5 f3e5 h5d1 f1d1 f8d6 c1f4 h8f8 f4g3 g7g5 d1e1 h7h5 h2h3 e8c8 b4b5 a6a5 e1e2 d8e8 a1e1 b7b6');
INSERT INTO evaluations VALUES(22234,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(22234,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(22234,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(22234,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(22234,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(22234,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(22234,36,'cp 75','f1b5 c8d7 g1f3 a7a6 b5c6 d7c6 h2h4 g8h6 c1h6 g7h6 d1d2 f8g7 c3e2 f6e5 f3e5 c6b5 a2a4 b5e2 d2e2 e8g8 e5f3 d8e7 e1g1 c7c5 f1e1 c5d4 e2e6 e7e6 e1e6 a8c8 e6e2 c8c4 f3e1 d4d3 e1d3 f8c8 a1e1 c4c2 e2c2 c8c2 e1e8 g8f7 e8b8 c2c7 a4a5 f7e7 b2b4 e7d6');
INSERT INTO evaluations VALUES(10295,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(10295,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(10295,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(10295,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(10295,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(10295,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(10295,36,'cp 12','e4e5 d7d5 e5f6 d5c4 f6g7 h8g8 c1g5 c5e7 f1e1 c8e6 g5e7 d8e7 f3d4 a8d8 c2c3 g8g7 d1f3 c6d4 c3d4 c7c6 b1c3 d8d4 a1d1 d4d1 f3d1 e8f8 d1d4 g7g6 d4h8 g6g8 h8e5 g8g6 g2g3 e7f6 e5b8 f8g7 b8b7 e6h3 e1e3 f6d6 b7e7 d6e7 e3e7 g6d6 f2f3 d6d2 e7e2 d2d3 g1f2 h3e6 c3e4 e6d5 e2e3 d5e4 f3e4 d3d1 f2f3 d1a1 e3a3 a1b1 a3a7');
INSERT INTO evaluations VALUES(10051,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(10051,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(10051,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(10051,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(10051,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(10051,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(10051,36,'cp 38','d7d5 e5f6 d5c4 f6g7 h8g8 c1g5 f7f6 f1e1 e8f7 g5h6 g8e8 b1d2 e8e1 f3e1 d8d5 d2c4 f7g6 h6f4 g6g7 b2b3 b7b5 c4b2 c5b4 b2d3 b4c3 a1c1 c8f5 f4c7 c6e7 c7g3 e7g6 d3f4 g6f4 g3f4 a8e8 e1d3 d5e6 d1f3 f5e4 f3g3 e4g6 h2h3 a7a5 a2a4 b5b4 g1h2 g7h8');
INSERT INTO evaluations VALUES(9079,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(9079,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(9079,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(9079,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(9079,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(9079,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(9079,36,'cp 26','e5f6 d5c4 f6g7 h8g8 c1g5 c5e7 g5e7 e8e7 f1e1 e7f6 c2c3 c8g4 h2h3 g4f3 d1f3 f6g7 e1e4 g8e8 b1d2 e8e4 d2e4 d4d3 f3g4 g7h8 g4f5 d8g8 e4d2 g8g6 f5d5 a7a6 d5c4 a8e8 c4f4 e8e2 a1d1 h8g7 f4c7 c6e5 c7b7 g6f6 b7a7 e2d2 d1d2 e5f3 g2f3 f6g5 g1f1 g5d2 a7d4 f7f6 d4e3 d2d1 f1g2 d3d2 e3e7 g7g6');
INSERT INTO evaluations VALUES(9062,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9062,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(9062,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(9062,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(9062,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(9062,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(9062,36,'cp 43','d5c4 f6g7 h8g8 c1g5 c5e7 g5e7 e8e7 f1e1 e7f6 c2c3 d4d3 b1d2 f6g7 d2c4 c8f5 d1a4 d8f6 a1d1 a8e8 e1e8 g8e8 c4e3 f5g6 a4h4 f6h4 f3h4 c6e5 h4g6 h7g6 f2f4 e5d7 d1d3 d7c5 d3d5 e8e3 d5c5 e3e2 b2b4 e2a2 c5c7 b7b6');
INSERT INTO evaluations VALUES(9059,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9059,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(9059,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(9059,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(9059,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(9059,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(9059,36,'cp 33','f6g7 h8g8 c1g5 c5e7 g5e7 e8e7 f1e1 e7f6 c2c3 d4d3 b1d2 f6g7 d2c4 c8f5 d1d2 d8f6 e1e3 a8d8 a1e1 h7h6 c4e5 g7h7 g2g3 c6e5 e3e5 f5g6 d2e3 b7b6 a2a3 a7a6 f3d2 a6a5 f2f4 c7c5 h2h3 a5a4 e3f3');
INSERT INTO evaluations VALUES(8767,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(8767,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(8767,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(8767,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(8767,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(8767,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(8767,36,'cp 30','h8g8 c1g5 f7f6 f1e1 e8f7 g5h6 g8e8 b1d2 e8e1 f3e1 d8d5 d2c4 f7g6 h6f4 g6g7 c4d2 c5b6 e1d3 c8d7 b2b3 c6e7 d2c4 c7c5 a2a4 e7g6 f4d6 a8c8 d1d2 b6c7 d6c7 c8c7 f2f3 d7f5 a1e1 f5d3 d2d3 c7e7 e1e7 g6e7 g1f2 b7b6 d3e4 d5e4 f3e4');
INSERT INTO evaluations VALUES(7042,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(7042,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(7042,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(7042,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(7042,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(7042,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(7042,36,'cp 49','c1g5 f7f6 f1e1 e8f7 g5h6 g8e8 b1d2 e8e1 f3e1 d8d5 d2c4 f7g6 h6f4 g6g7 b2b3 b7b5 c4b2 c5b4 b2d3 b4c3 a1b1 c8f5 f4c7 c6e7 c7g3 e7g6 h2h4 a8e8 h4h5 g6e5 h5h6 g7g8 d1h5 f5g6 h5h4 d5e6 d3f4 e6f5 f2f3 g6f7 g3f2 e5g6 f4g6 f7g6 e1d3 f5g5 b1d1 e8e2');
INSERT INTO evaluations VALUES(26567,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(26567,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(26567,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(26567,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(26567,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(26567,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(26567,36,'cp 53','f1b5 e7e6 c2c4 g8e7 c4d5 e6d5 b1c3 a7a6 b5a4 g7g6 e1g1 f8g7 a4c6 e7c6 d1b3 c6e7 b3b7 f5d3 f1d1 a8b8 b7a7 b8b6 c3d5 d8d5 d1d3 d5d8 d3c3 e7c6 c3c6 b6c6 c1d2 e8g8 d2c3 c6b6 f3d2 f8e8 a2a4 d8b8 a7b8 b6b8 d2c4 c7c5 d4c5 g7c3 b2c3 e8c8 c4b6 c8c5 b6d7 b8c8 d7c5 c8c5 a1c1 c5c4 f2f3 c4a4');
INSERT INTO evaluations VALUES(26508,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(26508,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(26508,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(26508,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(26508,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(26508,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(26508,36,'cp 62','e7e6 c2c4 g8e7 c4d5 e6d5 b1c3 e7g6 e1g1 f8e7 b5e2 a8b8 a2a3 e8g8 b2b4 a7a6 d1b3 f5e6 c3a4 e7d6 a4c5 e6g4 h2h3 g4f5 c1d2 c6e7 a3a4 d8c8 h3h4 h7h5 b4b5 a6b5 a4b5 b7b6 c5a6 b8a8 f1c1');
INSERT INTO evaluations VALUES(54304,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(54304,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(54304,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(54304,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(54304,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(54304,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(54304,36,'cp -40','b8c6 c2c3 g8e7 f1e2 h7h5 e1g1 d5d4 e2d3 b7b6 a2a4 c8b7 b1a3 e7f5 d3e4 f8e7 d2d3 g7g6 c1d2 a8b8 c3c4 h5h4 b2b3 d8d7 a1b1 e8g8 a3c2 g8g7 d2e1 b7a8 d1e2 f8h8 h2h3 d7c7 c2a3 a7a6 f3g5');
INSERT INTO evaluations VALUES(20351,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(20351,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(20351,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(20351,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(20351,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(20351,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(20351,36,'cp -32','c2c3 d5d4 f1d3 f7f5 d3c4 g8e7 d2d3 b7b6 e1g1 a7a6 a2a4 e7d5 c1d2 f8e7 c3d4 c6d4 b1c3 d5b4 g1h1 d4f3 d1f3 a8b8 a1d1 e8g8 d2e3 b4c2 f3f2 c2d4 e3d4 d8d4 f2d4 c5d4 c3e2 f8d8 d1c1 g8f7 f1g1 b6b5');
INSERT INTO evaluations VALUES(20347,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(20347,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(20347,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(20347,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(20347,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(20347,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(20347,36,'cp -39','g8e7 f1e2 d5d4 d2d3 e7f5 g2g4 f5h4 e1g1 h4g6 d1e1 f8e7 e1g3 e8g8 c3d4 f7f5 e5f6 e7f6 d4c5 b7b6 b1c3 c6d4 f3d4 f6d4 g1h1 b6c5 e2f3 a8b8 c3e2 d4f6 f3e4 c8a6 a1b1 a6b7 e4b7 b8b7');
INSERT INTO evaluations VALUES(12966,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(12966,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(12966,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(12966,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(12966,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(12966,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(12966,36,'cp -36','f1e2 d5d4 d2d3 b7b6 b1a3 e7f5 e1g1 h7h5 f3g5 g7g6 c3c4 a7a6 e2f3 c8b7 g5e4 f8e7 a3c2 b6b5 b2b3 a8b8 a1b1 e8f8 c1d2 f8g7 d1e2 c6a7 c2e1 d8b6 e4g3 f5g3 h2g3 b7f3 e2f3 a7c6 b1c1 a6a5 f3e2 a5a4 b3a4 b5c4 c1c4');
INSERT INTO evaluations VALUES(10904,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(10904,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(10904,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(10904,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(10904,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(10904,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(10904,36,'cp 46','d2d4 e5d4 f3d4 e8g8 b1c3 e7b4 d4c6 d7c6 c4d3 f8e8 h2h3 f6d7 f2f4 d7c5 f4f5 c5d7 d1h5 d8e7 c1d2 b4d6 a1e1 b7b5 g1h1 d7f6 h5h4 f6d7 d2g5 f7f6 g5e3 c8a6 a2a4 d7e5 d3e2 h7h6 e2h5 e8d8 a4b5 c6b5');
INSERT INTO evaluations VALUES(10916,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(10916,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(10916,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(10916,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(10916,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(10916,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(10916,36,'cp 37','e5d4 f3d4 e8g8 b1c3 e7b4 c1g5 h7h6 g5f6 d8f6 c3d5 f6d4 d1d4 c6d4 d5b4 d7d6 f1d1 d4c6 b4d5 c8e6 f2f4 c6b4 g1f2 a7a5 a2a3 b4d5 c4d5 a8b8 b2b4 e6d5 d1d5 a5b4 a3b4 f8e8 f2f3 e8e7 a1a7 b8e8 a7b7 e7e4 b7c7 e4b4');
INSERT INTO evaluations VALUES(22098,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(22098,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(22098,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(22098,33,'cp 229',NULL);
INSERT INTO evaluations VALUES(22098,34,'cp 233',NULL);
INSERT INTO evaluations VALUES(22098,35,'cp 251',NULL);
INSERT INTO evaluations VALUES(22098,36,'cp 267','f3d4 e5d4 d1d4 g8f6 c1f4 f8e7 b1c3 e8g8 f2f3 f6h5 f4e3 f7f5 e1c1 e7f6 d4d2 f6c3 d2c3 f5e4 f3e4 f8e8 d1e1 h5f6 e3g5 h7h6 g5f6 d8f6 c3f6 g7f6 e1e3 c8d7 f1d3 e8e5 h1f1 a8f8 c1d2 g8f7 e3f3 f7e7 g2g3 c7c6 d5c6');
INSERT INTO evaluations VALUES(22094,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(22094,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(22094,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(22094,33,'cp 261',NULL);
INSERT INTO evaluations VALUES(22094,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(22094,35,'cp 242',NULL);
INSERT INTO evaluations VALUES(22094,36,'cp 268','e5d4 d1d4 g8f6 b1c3 f8e7 b2b3 c7c5 d5c6 b7c6 c1b2 d8b6 e1c1 e8g8 f1e2 f8d8 c1b1 c8e6 f2f3 d6d5 e4d5 b6d4 d1d4 f6d5 c3d5 e6d5 e2d3 g7g6 d4a4 e7c5 h1d1 h7h5 a4a6 c5b6 c2c4 d5e6 b1c2 h5h4 d1e1 h4h3');
INSERT INTO evaluations VALUES(22087,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(22087,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(22087,32,'cp 261',NULL);
INSERT INTO evaluations VALUES(22087,33,'cp 268',NULL);
INSERT INTO evaluations VALUES(22087,34,'cp 261',NULL);
INSERT INTO evaluations VALUES(22087,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(22087,36,'cp 264','d1d4 g8f6 b1c3 f8e7 b2b3 c7c5 d5c6 b7c6 c1b2 d8b6 e1c1 e8g8 f1e2 f8e8 c1b1 h7h5 h2h3 b6d4 d1d4 c8b7 h1d1 e7f8 d4b4 b7c8 b4a4 a7a5 f2f4 d6d5 e4d5 f6d5 e2f3 d5e3 d1d2 c8b7 f3h5 f8b4 h5f3');
INSERT INTO evaluations VALUES(22086,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(22086,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(22086,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(22086,33,'cp 254',NULL);
INSERT INTO evaluations VALUES(22086,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(22086,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(22086,36,'cp 251','g8f6 b1c3 f8e7 b2b3 c7c5 d5c6 b7c6 c1b2 d8b6 e1c1 e8g8 f2f3 c8e6 c1b1 a8d8 f1e2 b6d4 d1d4 d6d5 e4d5 f6d5 c3d5 e6d5 h1d1 e7c5 d4d2 c5e3 d2d3 e3g5 c2c4 d5e6 d3d8 f8d8 d1d8 g5d8');
INSERT INTO evaluations VALUES(15730,30,'cp 273',NULL);
INSERT INTO evaluations VALUES(15730,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(15730,32,'cp 262',NULL);
INSERT INTO evaluations VALUES(15730,33,'cp 263',NULL);
INSERT INTO evaluations VALUES(15730,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(15730,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(15730,36,'cp 262','b1c3 f8e7 b2b3 c7c5 d5c6 b7c6 c1b2 e8g8 f1e2 d8b6 e1c1 f8d8 c1b1 c8e6 f2f3 b6d4 d1d4 d6d5 e4d5 f6d5 c3d5 e6d5 h1d1 h7h5 d4a4 e7d6 e2d3 d6c7 a4h4 g7g6 b2f6 d8d7 d1e1 c7b6 h4a4 d7d6 f6c3 a8d8 b1c1 b6c5 h2h4 f7f5 e1e2');
INSERT INTO evaluations VALUES(15729,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(15729,31,'cp 273',NULL);
INSERT INTO evaluations VALUES(15729,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(15729,33,'cp 268',NULL);
INSERT INTO evaluations VALUES(15729,34,'cp 272',NULL);
INSERT INTO evaluations VALUES(15729,35,'cp 264',NULL);
INSERT INTO evaluations VALUES(15729,36,'cp 253','f8e7 b2b3 c7c5 d5c6 b7c6 c1b2 e8g8 e1c1 d8b6 f1e2 f8d8 f2f3 h7h5 e2c4 f6d7 a2a4 h5h4 c3b5 d7f6 d4b6 a7b6 b5d4 c8d7 h1e1 d8e8 c1b1 e7f8 g2g4 d6d5 e4d5 e8e1 d1e1');
INSERT INTO evaluations VALUES(8675,30,'cp 261',NULL);
INSERT INTO evaluations VALUES(8675,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(8675,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(8675,33,'cp 281',NULL);
INSERT INTO evaluations VALUES(8675,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(8675,35,'cp 261',NULL);
INSERT INTO evaluations VALUES(8675,36,'cp 283','c1f4 e8g8 f2f3 c7c5 d5c6 b7c6 e1c1 f6e8 f1c4 c8e6 c4b3 a7a5 h1e1 d8c7 d4d2 a8d8 c1b1 e8f6 c3a4 h7h6 a4c5 e6b3 c5b3 a5a4 b3d4 f6h5 f4e3 a4a3 g2g4 h5f6 d4f5 f8e8 h2h4 a3b2 g4g5 f6h5 g5h6 g7g6 f5e7 c7e7');
INSERT INTO evaluations VALUES(18917,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(18917,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(18917,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(18917,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(18917,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(18917,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(18917,36,'cp 70','d4d5 e6d5 e4d5 f8b4 b1c3 c6e5 g1f3 d8e7 f1e2 b4c3 b2c3 e5f3 g2f3 e7e5 d5d6 g8f6 d1d2 b7b6 h1g1 c8a6 g1g7 e8f8 g7g1 f6d5 c3c4 d5c3 c1b2 e5e2 d2e2 c3e2 e1e2 f7f6 b2f6 a6c4 e2e3 h8g8 g1g4 f8f7 g4c4 f7f6 c4c7 f6e6 a1d1 g8g5');
INSERT INTO evaluations VALUES(18866,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(18866,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(18866,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(18866,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(18866,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(18866,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(18866,36,'cp 69','e6d5 e4d5 c6e5 b1c3 f8b4 g1f3 d8e7 f1e2 e5f3 g2f3 g8f6 h1g1 e8g8 c1h6 g7g6 h6f8 g8f8 a2a3 b4c3 b2c3 e7e5 d1d4 e5d4 c3d4 f6d5 e2c4 d5e7 e1d2 b7b6 a3a4 c8b7 f3f4 e7f5 d2d3 f5d6 g1e1 a8c8 c4b5 c8d8 a1c1 a7a6 b5c4 h7h6 d3d2 d6f5 d4d5 f5d6');
INSERT INTO evaluations VALUES(19048,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(19048,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(19048,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(19048,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(19048,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(19048,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(19048,36,'cp 71','e4d5 f8b4 b1c3 c6e5 g1f3 d8e7 f1e2 e5f3 g2f3 g8f6 h1g1 e8g8 c1h6 g7g6 h6f8 g8f8 a2a3 b4c3 b2c3 e7e5 d1d4 e5d4 c3d4 f6d5 a1c1 b7b6 e1d2 c8b7 g1e1 b7c6 e2d3 d5f4 d3e4 a8d8 e1e3 f8g7 c1c4 b6b5 c4c5 f4e6 c5c1 e6f4 e4c6 d7c6 c1c6 d8d4 d2c3 d4d8 c3c2 f4e6 c2b2 d8d2 c6c2 d2d1');
INSERT INTO evaluations VALUES(19046,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(19046,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(19046,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(19046,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(19046,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(19046,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(19046,36,'cp 83','f8b4 b1c3 d8e7 f1e2 c6e5 g1f3 e5f3 g2f3 g8f6 h1g1 e8g8 c1h6 g7g6 h6f8 g8f8 a2a3 b4c3 b2c3 e7e5 d1d4 e5d4 c3d4 b7b6 d5d6 c8b7 a1c1 a8e8 g1g5 e8e6 c1c7 b7c6 c7a7 e6d6 a7a6 c6b7 a6a4 f6d5 e2c4 d5f4 g5g4 d6d4 c4b3');
INSERT INTO evaluations VALUES(18273,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18273,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18273,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18273,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18273,36,'cp 0','g8e7 d2d4 f7f6 b1c3 e7g6 d1e2 a7a5 e2c4 d8d7 f1e1 f8e7 c1e3 c8a6 c4b3 a6c8');
INSERT INTO evaluations VALUES(20665,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(20665,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(20665,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(20665,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(20665,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(20665,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(20665,36,'cp 39','d2d4 c5d4 f3d4 f8e7 c1e3 g8f6 d1e2 e6e5 d4f3 e8g8 a1d1 c8e6 e1g1 a8c8 c4b3 a7a6 h2h3 d8c7 f3g5 e6b3 c2b3 b7b5 g5f3 c7b7 e3g5 f6e8 g5e7 c6e7 d1d2 f7f5 f1d1 h7h6 e4f5 e7f5 c3e4 c8c6 a2a4 b5a4');
INSERT INTO evaluations VALUES(20687,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(20687,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(20687,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(20687,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(20687,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(20687,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(20687,36,'cp 46','c5d4 f3d4 g8f6 c1e3 f8e7 d1e2 e6e5 d4f5 c8f5 e4f5 c6d4 c4b5 e8f8 e2d3 a7a6 b5a4 b7b5 a4b3 d4b3 a2b3 d6d5 e3g5 d5d4 c3e4 f6e4 g5e7 f8e7 d3e4 h8e8 f2f4 e7f8 f4e5 d8e7 e5e6 f7e6 f5e6 e7e6 e1g1 f8g8 e4d4 e6e5 d4b6 e5b2 a1a6 a8a6 b6a6 b2e5 a6c6 e5e3 g1h1 e3e2 c6d5 g8h8');
INSERT INTO evaluations VALUES(20716,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(20716,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(20716,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20716,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(20716,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(20716,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(20716,36,'cp 47','f3d4 f8e7 c1e3 g8f6 d1e2 e6e5 d4f3 e8g8 c4b3 h7h6 e1g1 c6a5 f1d1 c8d7 b3d5 d8c8 f3d2 d7e6 d2f1 c8d7 a2a4 f8c8 f1g3 a5c4 e3c1 a7a6 h2h3 e7d8 b2b3 c4a5 e2d3 a5c6 c1e3 e6d5 c3d5 f6d5 d3d5');
INSERT INTO evaluations VALUES(20711,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(20711,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(20711,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(20711,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(20711,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(20711,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(20711,36,'cp 36','g8f6 c1e3 f8e7 d1e2 e6e5 d4f3 e8g8 e1g1 c6a5 c4d5 a7a6 f1d1 c8d7 h2h3 a8c8 a2a4 d8c7 f3d2 d7e6 d5a2 c7c6 e2d3 e7d8 d1c1 c6d7 c3d5 f6d5 a2d5 a5c6 c2c4 d8a5 d2f3 a5b4 c1d1 e6d5 e4d5 c6e7 d3b3 b4c5 e3c5 c8c5');
INSERT INTO evaluations VALUES(13985,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(13985,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(13985,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(13985,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(13985,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(13985,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(13985,36,'cp 45','c1e3 f8e7 d1e2 e6e5 d4f3 e8g8 e1g1 c6a5 c4d5 a7a6 f1d1 d8c7 f3h4 c8g4 f2f3 g4e6 h4f5 a8c8 f5e7 c7e7 a2a4 h7h6 d1d2 e6d5 c3d5 f6d5 d2d5 a5c4 e3c1 c4b6 d5d1 b6d7 b2b3 c8c3 c1a3 c3c6 d1d5 d7f6 d5d2 e7c7 a1d1 f8d8 h2h3 f6h5 e2f2 h5f6');
INSERT INTO evaluations VALUES(13983,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(13983,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(13983,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(13983,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(13983,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(13983,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(13983,36,'cp 29','f8e7 d1e2 e6e5 d4f3 c8e6 c4b3 e8g8 e1g1 h7h6 a1d1 d8c8 f1e1 a7a6 e3b6 f8e8 h2h3 c6b8 b6e3 b7b5 f3h2 b8d7 h2g4 f6g4 h3g4 d7c5 f2f3 c5b3 a2b3 c8b7 d1d2 a8c8 c3d5 e6d5 d2d5 b5b4 e1d1 c8c6 e3f2');
INSERT INTO evaluations VALUES(8196,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(8196,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(8196,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(8196,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(8196,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(8196,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(8196,36,'cp 43','d1e2 e6e5 d4f5 c8f5 e4f5 c6d4 c4b5 e8f8 e2d3 a7a6 b5a4 b7b5 a4b3 d4b3 a2b3 d6d5 e3g5 d5d4 c3e4 f6e4 g5e7 f8e7 d3e4 h8e8 f2f4 e7f8 f4e5 d8e7 e5e6 f7e6 f5e6 e7e6 e4e6 e8e6 e1d2 a8e8 h1f1 f8g8 f1f2 h7h5 d2d3 e6d6 f2f3 e8e6 g2g3 g8h7 a1f1 g7g5 d3d2 e6e5 f3f8 d6e6 f8f2 e5e3');
INSERT INTO evaluations VALUES(12357,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(12357,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(12357,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(12357,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(12357,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(12357,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(12357,36,'cp -10','f8e7 e1g1 e8g8 f1e1 d7d6 a2a4 a7a6 h2h3 b7b6 c1f4 c8b7 f4g3 h7h6 c3e2 d6d5 e4d5 f6d5 c2c3 e7f6 d3d4 c5d4 e2d4 c6d4 f3d4 d5e7 d1g4 f6d4 a1d1 e7f5 c3d4 d8g5 g4g5 h6g5 d4d5 f5g3');
INSERT INTO evaluations VALUES(7903,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(7903,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(7903,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(7903,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(7903,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(7903,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(7903,36,'cp -14','e1g1 e8g8 f1e1 d7d6 a2a3 a7a6 c4a2 b7b5 c3e2 c8b7 c2c3 c6e5 e2g3 d8b6 c1f4 e5f3 d1f3 a6a5 a1c1 a8c8 f3e2 b5b4 a3b4 a5b4 f4e3 f8d8 a2b1 g7g6 d3d4 b4c3 b2c3 c5d4 c3d4 c8c1 e1c1 d8c8 c1c8 b7c8 e2c2 g8g7 h2h3 d6d5 e4d5 e6d5 c2c8 b6b1 c8c1');
INSERT INTO evaluations VALUES(7902,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(7902,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(7902,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(7902,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(7902,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(7902,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(7902,36,'cp -25','e8g8 f1e1 d7d5 c4b3 a7a6 a2a4 b7b6 c1f4 c6a5 b3a2 d5d4 c3e2 c8b7 h2h3 f6h5 f4c1 b6b5 c2c3 d4c3 b2c3 d8c7 d1c2 f8d8 c1e3 h5f6 e3f4 c7b6 f4e3 a5c6 d3d4 b5b4 e2g3 b4c3 d4c5 e7c5');
INSERT INTO evaluations VALUES(3027,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(3027,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(3027,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(3027,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(3027,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(3027,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(3027,36,'cp -16','c4b5 d8c7 b5c6 c7c6 f3e5 c6c7 c1f4 d7d6 e5c4 c7d8 a2a4 d6d5 c4d2 c8d7 f1e1 d7c6 h2h3 d5d4 c3e2 f6e8 f4h2 f7f6 e2f4 e8c7 d1g4 d8d7 f4h5 e7d8 h2d6 f8f7 e4e5 f6f5 g4g3 c7d5 d6c5 c6a4 b2b3 a4c6');
INSERT INTO evaluations VALUES(53322,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(53322,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53322,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(53322,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(53322,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(53322,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(53322,36,'cp 18','f1g2 g8f6 e1g1 g7g6 c2c4 f8g7 d2d4 c5d4 c4d5 f6d5 f3d4 e8g8 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 a1d1 d8f8 d1d2 g7c3 d2b2 c3b2 f4e3 b2a3 e3a7 c8e6 f1b1 f8c8 c6f3 e6a2 b1a1 c8c2 a7e3');
INSERT INTO evaluations VALUES(53323,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53323,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(53323,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(53323,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53323,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(53323,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(53323,36,'cp 17','g8f6 e1g1 g7g6 c2c4 f8g7 c4d5 f6d5 d2d4 c5d4 f3d4 e8g8 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 a1d1 d8f8 c3c4 c8a6 c6b5 a6b5 c4b5 b2b5 d1d2 h7h5 f1c1 g7e5 d2d7 e5f4 g3f4 a7a6 e2e4 e7e5 g1g2 g8g7');
INSERT INTO evaluations VALUES(45404,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(45404,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45404,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45404,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(45404,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(45404,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45404,36,'cp 18','e1g1 g7g6 d2d4 c5d4 f3d4 f8g7 c2c4 e8g8 c4d5 f6d5 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 f1d1 d8f8 a2a4 b2e2 a1c1 c8e6 f4e3 e6g4 d1d2 a7a6 c6b7 a6a5 c3c4 e2d2 e3d2 f8b8 c1b1 g4d7 d2a5 d7a4 c4c5');
INSERT INTO evaluations VALUES(45403,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(45403,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45403,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(45403,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45403,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(45403,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(45403,36,'cp 24','g7g6 c2c4 f8g7 c4d5 f6d5 d2d4 c5d4 f3d4 e8g8 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 c8f5 c6a8 g7c3 c1g5 c3a1 f1a1 d8a8 g5e7 a8e8 e7c5 a7a5 g1f1 e8b8 f1e1 b8b1 a1b1 f5b1 a2a3 f7f5 e1d2 g8f7 d2e3 b1a2 e3f4 a2c4 f4g5 c4e2 g5h6 e2c4 f2f4 c4d5');
INSERT INTO evaluations VALUES(43290,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(43290,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(43290,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43290,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(43290,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43290,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(43290,36,'cp 34','c2c4 c7c6 b2b3 f6e4 d2d4 e8g8 c1b2 d7d5 e1g1 c8f5 f3h4 f5e6 d1c2 e4d6 b1d2 a7a5 c4c5 d6e8 e2e3 b8a6 a2a3 g7f6 h4f3 e6f5 c2c3 a6c7 h2h3 f6g7 g1h2 c7b5 c3c1 a5a4 b3a4 a8a4 f1e1 e8c7 f3e5');
INSERT INTO evaluations VALUES(43171,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43171,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(43171,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(43171,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43171,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43171,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(43171,36,'cp 11','c7c6 d2d4 d7d5 c4d5 c6d5 f3e5 c8f5 b1c3 e8g8 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 h7h6 h2h4 e7e6 e1g1 f8c8 a1c1 c6e5 f4e5 f6e4 g2e4 d5e4 e5g7 g8g7 a2a3 c8c6 g1h2 a8c8 f1d1 b6b5 d4d5 c6c4 d5e6 f5e6');
INSERT INTO evaluations VALUES(39979,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(39979,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(39979,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(39979,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(39979,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(39979,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(39979,36,'cp 22','b2b3 f6e4 d2d4 d7d5 e1g1 e8g8 c1b2 c8f5 b1c3 e4c3 b2c3 f5e4 a1c1 b8d7 c4d5 c6d5 f3e5 e4g2 g1g2 a8c8 d1d2 d7e5 d4e5 d8d7 c3b2 b7b6 e2e4 c8c1 f1c1 e7e6 e4d5 e6d5 c1d1 f8d8 d2d4 h7h5 h2h3 d7e6 d1d2 d8d7 d4e3 d7d8');
INSERT INTO evaluations VALUES(55763,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(55763,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(55763,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(55763,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(55763,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(55763,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(55763,36,'cp -41','b7b6 c5b6 a7b6 e2e3 f8b4 c1d2 b4d6 g1f3 g8f6 a2a4 e8g8 f1e2 f6e4 e1g1 e4d2 d1d2 b8d7 b1c3 c8b7 f1b1 c7c5 c3b5 d6e7 b1c1 d8b8 b2b3 g7g6 b5c3 f8d8 e2b5 e7f6 b5d7 d8d7 c3b5 f6g7 b3b4 c5c4 a4a5 b6a5 b4a5 b7c6 b5c3 d7d8');
INSERT INTO evaluations VALUES(51357,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(51357,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(51357,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(51357,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(51357,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(51357,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(51357,36,'cp -55','c5b6 a7b6 e2e3 f8b4 c1d2 b4d6 g1f3 g8f6 b1c3 e8g8 a2a4 c7c5 f1e2 b8c6 e1g1 d6b8 c3b5 f6e4 d2e1 d8e7 f3d2 e4f6 f2f4 c5d4 b5d4 c6d4 e3d4 f6e4 g2g3 c8b7 a1c1 f8d8 d1b3 b8d6 e2d3 b7a6 d3a6 a8a6 d2e4 d5e4 b3b5 a6a8 c1c6 e7d7 c6b6 a8a4 b5d7');
INSERT INTO evaluations VALUES(51342,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(51342,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(51342,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(51342,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(51342,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(51342,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(51342,36,'cp -42','a7b6 e2e3 g8f6 f1e2 f8b4 c1d2 b4e7 g1f3 e8g8 b1c3 c7c5 e1g1 f6e4 d2e1 e4d6 a2a4 c8a6 f3e5 d8e8 e2a6 a8a6 d1e2 a6a8 e1d2 b8c6 e5c6 e8c6 c3b5 f8c8 b2b3 d6b5 a4b5 c6b7 a1a6 c5d4 e3d4 e7f6 d2e3 c8c3');
INSERT INTO evaluations VALUES(15580,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(15580,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(15580,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(15580,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(15580,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(15580,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(15580,36,'cp 11','d3d4 e5d4 f3d4 f6e4 c3e4 d8e7 f2f3 d6d5 f1b5 c8d7 b5c6 b7c6 e1g1 d5e4 f1e1 e8c8 e1e4 e7b4 d4b3 b4b7 c1g5 f7f6 g5f4 d7f5 e4d4 b7b6 c2c3 d8d4 d1d4 b6d4 b3d4 f5d7 a1e1 f8c5 g1f1 c5d4 c3d4 h8e8 e1c1 c8b7 d4d5 g7g5 d5c6 d7c6 f4g3 h7h5 c1d1 b7c8 g3f2 g5g4');
INSERT INTO evaluations VALUES(15551,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(15551,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(15551,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(15551,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(15551,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(15551,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(15551,36,'cp 12','e5d4 f3d4 f6e4 c3e4 d8e7 f2f3 d6d5 f1b5 c8d7 e1g1 d5e4 b5c6 b7c6 f1e1 e8c8 e1e4 e7b4 d4b3 b4b7 d1f1 f8d6 e4e1 b7b5 c1e3 b5f1 e1f1 a7a6 e3d4 f7f6 f1e1 h8e8 b3c5 a6a5 a2a3 h7h5 b2b4 e8e1 a1e1 a5b4 a3b4 d6c5 b4c5 d8e8 g1f2 e8e1 f2e1 d7e6 c2c3 c8d7 g2g3');
INSERT INTO evaluations VALUES(15649,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(15649,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(15649,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(15649,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(15649,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(15649,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(15649,36,'cp 5','f3d4 f6e4 c3e4 d8e7 f2f3 d6d5 f1b5 c8d7 e1g1 d5e4 b5c6 b7c6 f1e1 e8c8 e1e4 e7b4 d4b3 b4b7 d1e2 f8d6 c1e3 h8e8 e3f2 b7b5 a1e1 e8e4 e2e4 b5h5 e4h4 h5h4 f2h4 f7f6 h4f2 a7a6 f2d4 d8e8 e1e8 d7e8 b3c5 e8g6 c2c3 a6a5 g1f2 g6f7 a2a3 h7h5 h2h4 d6c5 d4c5 g7g5 h4g5 f6g5 f2e3');
INSERT INTO evaluations VALUES(41287,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(41287,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(41287,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(41287,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41287,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(41287,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41287,36,'cp 29','e8g8 c1e3 c7c5 g1e2 d8a5 e2c1 c5d4 c1b3 a5h5 b3d4 b8c6 d1d2 h5a5 e1c1 c6d4 e3d4 c8e6 c1b1 f8c8 b2b3 a7a6 c3a4 a5d2 d1d2 f6d7 d4g7 g8g7 b1b2 h7h5 h2h4 c8c7 a4c3 a8c8 a2a4 b7b6 f1e2 d7c5 g2g3 f7f6 f3f4 e6d7 e2f3 a6a5 c3d5');
INSERT INTO evaluations VALUES(36570,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(36570,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(36570,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(36570,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36570,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(36570,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(36570,36,'cp 42','g1e2 a7a6 c1e3 b8d7 a2a4 e7e6 e2c1 b7b6 f1e2 c7c5 d4c5 d6c5 c1b3 d8c7 e1g1 a8b8 d1c2 d7e5 a4a5 e5c6 a5b6 c7b6 b3c5 b6b2 c2b2 b8b2 c5a4 b2c2 a1c1 c2c1 f1c1 f6d7 f3f4 g7d4');
INSERT INTO evaluations VALUES(36569,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36569,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(36569,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36569,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(36569,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(36569,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36569,36,'cp 27','a7a6 c1e3 b8d7 a2a4 e7e5 d4d5 f6h5 g2g4 h5f4 d1d2 h7h5 g4h5 d8h4 e2g3 d7c5 e3c5 d6c5 e1c1 c8d7 c1c2 g7h6 h5g6 f7g6 d2f2 b7b6 f2e1 h6g5 g3e2 h4e1 d1e1 g5h4 e1d1 g8g7 a4a5 b6a5 d1a1 a5a4 c3a4 d7a4 a1a4 h4f2 h2h4 a6a5 h1h2 f2d4 e2f4 f8f4 f1e2 a8b8 b2b3');
INSERT INTO evaluations VALUES(27208,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(27208,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(27208,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(27208,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(27208,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(27208,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(27208,36,'cp 67','e1g1 a7a6 b5c4 f8e7 d2d4 g8f6 f1e1 d7g4 d4d5 c6b8 c4f1 b8d7 b2b4 e8g8 b1d2 c7c6 c3c4 g4f3 d2f3 a6a5 d5c6 b7c6 b4b5 d7c5 f1d3 a5a4 d3c2 c6b5 c4b5 d8a5 a1b1 f8b8 c1d2 a5c7 h2h3 f6d7 a2a3 h7h6');
INSERT INTO evaluations VALUES(27207,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(27207,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(27207,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(27207,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(27207,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(27207,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(27207,36,'cp 57','a7a6 b5c4 f8e7 d2d4 g8f6 f1e1 d7g4 d4d5 c6b8 c4f1 e8g8 b2b4 a6a5 h2h3 g4h5 b1d2 c7c6 d5c6 b8c6 b4b5 c6a7 c3c4 f6d7 c1a3 a7c8 d2b1 d8c7 b1c3 c8b6 d1e2 f8c8 e1d1 c7c4 c3d5 e7d8 a3d6 c4e2 f1e2 b6d5 d1d5');
INSERT INTO evaluations VALUES(25712,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(25712,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(25712,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(25712,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25712,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(25712,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(25712,36,'cp 47','b5c4 f8e7 d2d4 g8f6 f1e1 d7g4 d4d5 c6b8 c4d3 g4h5 h2h3 b8d7 a2a4 h8g8 b1d2 g7g5 d2f1 g5g4 h3g4 h5g4 a4a5 h7h5 b2b4 f6h7 f1e3 d7f6 c3c4 h7g5 e3g4 g5f3 d1f3 h5g4 f3f5 d8d7 b4b5 d7f5 e4f5 e8c8 g2g3 f6d7 b5a6');
INSERT INTO evaluations VALUES(21088,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(21088,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21088,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(21088,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(21088,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(21088,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21088,36,'cp 0','d5d4 d2d3 e7e6 g2g3 d8d5 c2c4 d4c3 b1c3 f8b4 d3d4 b7b6 c1d2 b4c3 b2c3 c8b7 f1g2 d5c4 g2f1');
INSERT INTO evaluations VALUES(55099,30,'cp -152',NULL);
INSERT INTO evaluations VALUES(55099,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(55099,32,'cp -191',NULL);
INSERT INTO evaluations VALUES(55099,33,'cp -179',NULL);
INSERT INTO evaluations VALUES(55099,34,'cp -181',NULL);
INSERT INTO evaluations VALUES(55099,35,'cp -193',NULL);
INSERT INTO evaluations VALUES(55099,36,'cp -205','c8g4 c2c4 c7c6 d1b3 e7e6 h2h3 g4h5 b3b7 b8d7 d2d4 g8e7 g1f3 h5f3 g2f3 e7f5 e2e3 f8e7 f3g4 f5d6 b7c6 a8c8 c6a6 c8c4 b1d2 c4c2 g4d1 c2c7 d2f3 d6e4 c1d2 d8b8 b2b3 e8g8 e1g1 e7b4 d2b4 b8b4 d1e2 f8c8 e2d3');
INSERT INTO evaluations VALUES(31937,30,'cp -191',NULL);
INSERT INTO evaluations VALUES(31937,31,'cp -173',NULL);
INSERT INTO evaluations VALUES(31937,32,'cp -175',NULL);
INSERT INTO evaluations VALUES(31937,33,'cp -182',NULL);
INSERT INTO evaluations VALUES(31937,34,'cp -182',NULL);
INSERT INTO evaluations VALUES(31937,35,'cp -198',NULL);
INSERT INTO evaluations VALUES(31937,36,'cp -197','c2c4 c7c6 c4d5 c6d5 d1b3 g8f6 b1c3 e7e6 b3b7 b8d7 c3b5 a8c8 b7a6 f6e4 g1f3 f8c5 e1g1 e8g8 d2d4 c5b6 c1e3 d7b8 a6a3 b8c6 a1c1 d8d7 c1c2 c6a5 c2c8 f8c8 a3a4 a5c4 b5c3 d7a4 c3a4 c4e3 a4b6 a7b6 f2e3');
INSERT INTO evaluations VALUES(31892,30,'cp -195',NULL);
INSERT INTO evaluations VALUES(31892,31,'cp -175',NULL);
INSERT INTO evaluations VALUES(31892,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(31892,33,'cp -210',NULL);
INSERT INTO evaluations VALUES(31892,34,'cp -168',NULL);
INSERT INTO evaluations VALUES(31892,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(31892,36,'cp -153','c7c6 d1b3 e7e6 b3b7 b8d7 b1c3 g8e7 c4d5 e6d5 d2d4 a8b8 b7a6 b8b6 a6d3 e7g6 h2h3 g4e6 g1f3 f8d6 h3h4 h7h6 h4h5 g6e7 g2h3 e6h3 h1h3 e8g8 h3h1 d7f6 b2b3 d8c8 h1g1 e7f5 c1b2 f6h5 e2e3 f8e8 e1c1 h5f6 c3a4 b6b5 c1b1 f6e4 d3c2 f7f6 d1c1 b5b7 a4c5 d6c5 d4c5 b7e7 b1a1');
INSERT INTO evaluations VALUES(31105,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(31105,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(31105,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(31105,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(31105,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(31105,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(31105,36,'cp -163','d1b3 e7e6 b3b7 b8d7 b1c3 g8e7 c4d5 e6d5 d2d4 a8b8 b7a6 b8b6 a6d3 e7g6 h2h3 g4e6 g1f3 h7h6 h3h4 f8d6 g2h3 e6h3 h1h3 d7f6 h4h5 g6e7 h3h1 e8g8 b2b3 d8c8 h1g1 e7f5 c1b2 f8e8 e1c1 f6e4 g1g2 e4c3 b2c3 a7a5 c1c2 d6b4 f3e5 f7f6 c3b4 b6b4');
INSERT INTO evaluations VALUES(31103,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(31103,31,'cp -179',NULL);
INSERT INTO evaluations VALUES(31103,32,'cp -186',NULL);
INSERT INTO evaluations VALUES(31103,33,'cp -175',NULL);
INSERT INTO evaluations VALUES(31103,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(31103,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(31103,36,'cp -157','e7e6 b3b7 b8d7 b1c3 g8e7 c4d5 e6d5 d2d4 a8b8 b7a6 b8b6 a6d3 e7g6 h2h3 g4e6 g1f3 h7h6 h3h4 f8d6 g2h3 e6h3 h1h3 d7f6 h3h1 h6h5 d3c2 d8d7 c3a4 b6b5 c1g5 f6e4 f3d2 d6b4 a4c3 b4c3 b2c3 e4d2 c2d2 e8g8 a2a4');
INSERT INTO evaluations VALUES(39172,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(39172,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(39172,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(39172,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(39172,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(39172,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(39172,36,'cp 54','c1d2 b4d2 d1d2 g8f6 b1c3 d7d5 c4d5 e6d5 e2e3 e8g8 g1e2 f8e8 f2f3 d8e7 e1f2 c7c5 d4c5 e7c5 a1c1 b8c6 e2d4 c8d7 f1b5 c5b4 a2a3 b4a5 b5a4 c6d4 d2d4 d7a4 c3a4 a5b5 a4c5 b5b6 b2b4 a7a5');
INSERT INTO evaluations VALUES(39170,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(39170,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(39170,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(39170,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(39170,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(39170,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(39170,36,'cp 41','b4d2 d1d2 g8f6 b1c3 d7d5 c4d5 e6d5 e2e3 e8g8 f1d3 b7b6 g1e2 c7c5 e1g1 c8b7 a1d1 b8c6 e2g3 d8c7 c3b5 c7d8 f1e1 g7g6 d3e2 d8e7 b5c3 a8d8 e2f3 c6b4 a2a3 b4a6 h2h3 f8e8 e3e4 d5e4 c3e4 f6e4 g3e4 c5d4 e4g3 e7c7 e1e8 d8e8 f3b7 c7b7 d2d4');
INSERT INTO evaluations VALUES(39177,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(39177,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(39177,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(39177,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(39177,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(39177,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(39177,36,'cp 36','d1d2 d7d5 b1c3 g8f6 c4d5 e6d5 e2e3 e8g8 f1d3 b7b6 g1e2 c7c5 a1d1 c8b7 e1g1 b8c6 e2g3 d8c7 c3b5 c7d7 d4c5 b6c5 d2c3 c6b4 d3e2 b4a6 c3c1 d7e6 b5c3 f8d8 b2b3 e6e7 g3f5 e7e6 e3e4 f6e4 c3e4 e6f5 e4c5 a6c5');
INSERT INTO evaluations VALUES(39176,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(39176,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(39176,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(39176,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(39176,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(39176,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(39176,36,'cp 39','g8f6 b1c3 e8g8 e2e4 d7d5 e4e5 f6e4 c3e4 d5e4 d2e3 c7c5 e1c1 c5d4 e3e4 f7f5 e4d4 d8a5 a2a3 b8c6 d4d2 a5e5 f2f4 e5f6 g1f3 f8d8 f1d3 c8d7 d2e3 c6a5 f3e5 a5c6 h1e1 c6e5 e3e5 a8c8 d1d2 f6e5 e1e5 g8f8 e5a5 a7a6 b2b4 d7c6');
INSERT INTO evaluations VALUES(42509,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(42509,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(42509,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(42509,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(42509,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(42509,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(42509,36,'cp 38','b1c3 d7d5 c4d5 e6d5 e2e3 e8g8 f1d3 b7b6 a1d1 f8e8 g1e2 c8a6 e1g1 a6d3 d2d3 b8c6 f1e1 d8e7 d3f5 e7d8 f5c2 a7a6 h2h3 d8d6 e2f4 c6e7 f4d3 c7c6 a2a3 a8c8 b2b4 h7h5 d3e5');
INSERT INTO evaluations VALUES(42494,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(42494,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(42494,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(42494,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(42494,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(42494,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(42494,36,'cp 49','e8g8 e2e3 b7b6 f1e2 c8b7 e2f3 d7d5 c4d5 e6d5 g1e2 c7c6 e1g1 d8e7 e2f4 b8d7 a2a3 a7a5 f3e2 f8e8 f1c1 h7h6 d2d1 e7d6 c1c2 b7a6 e2a6 a8a6 a1c1 b6b5 f4d3 a6a8 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(14193,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(14193,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(14193,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(14193,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(14193,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(14193,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(14193,36,'cp -29','f6e4 g5h4 d8d6 a2a3 e4c3 b2c3 c6a5 d1d3 d6c6 f3d2 b7b6 e2e3 e7e6 a3a4 c8b7 d3b5 c6d7 b5d7 e8d7 h4g3 a8c8 f1b5 c7c6 b5e2 c6c5 e2b5 b7c6 b5a6 c8e8 e1g1 c5c4 a6b5 f8d6 e3e4 f7f6 e4e5 f6e5 g3e5 d6e5 d4e5 d7c7 f2f4 e8d8');
INSERT INTO evaluations VALUES(14432,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(14432,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(14432,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(14432,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(14432,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(14432,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(14432,36,'cp -24','g5h4 d8d6 a2a3 e4c3 b2c3 c6a5 d1d3 d6c6 f3d2 b7b6 e2e3 c8b7 d3b5 e7e6 b5c6 b7c6 h4g3 e8d7 a3a4 a7a6 f2f3 h7h5 d2b3 a5c4 f1c4 d5c4 b3d2 h5h4 d2c4 h4g3 c4e5 d7e8 e5c6 g3h2 a1b1 f8d6 e1f2 a6a5 c3c4 f7f6 g2g4 g7g5 c2c3 h8h4');
INSERT INTO evaluations VALUES(3604,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3604,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(3604,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(3604,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(3604,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(3604,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(3604,36,'cp 4','c3a4 c5b6 a2a3 c8e6 c4e6 f7e6 a4b6 a7b6 a3a4 d8d7 c2c3 c6e7 b2b4 b6b5 a4a5 e7g6 c1e3 d7c6 a1c1 d6d5 f1e1 f8f7 d1e2 f6d7 f3h2 a8f8 h2g4 c6d6 e1f1 c7c6 g4h2 d5d4 e3d2 g6f4 d2f4');
INSERT INTO evaluations VALUES(3637,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(3637,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(3637,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(3637,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(3637,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(3637,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(3637,36,'cp 19','c5b6 a2a3 c8e6 c4e6 f7e6 c2c3 c6e7 a4b6 a7b6 f1e1 e7g6 d3d4 e5d4 c3d4 d8d7 a3a4 d7c6 d1b3 d6d5 e4d5 f6d5 f3e5 g6e5 e1e5 f8f5 e5f5 e6f5 c1f4 b6b5 a4a5 c6e6 f4e5 c7c6 a1a3 e6e8 b3c2 f5f4 a3a1');
INSERT INTO evaluations VALUES(3565,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(3565,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(3565,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(3565,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3565,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(3565,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(3565,36,'cp 20','a2a3 c6e7 a4b6 a7b6 f1e1 c8e6 c4e6 f7e6 c2c3 e7g6 d3d4 f6d7 d1e2 d8f6 c1e3 f8f7 a1d1 a8f8 f3h2 f6h4 e2g4 h4g4 h3g4 b6b5 d4d5 e6d5 d1d5 b7b6 d5b5 d7c5 e3c5 b6c5 h2f3 g6f4 g1h2');
INSERT INTO evaluations VALUES(15436,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(15436,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(15436,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(15436,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(15436,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(15436,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(15436,36,'cp 51','f8e7 d2d4 e5d4 f3d4 c8d7 d4c6 b7c6 b5d3 e8g8 h2h3 d7c8 c1e3 f6d7 b1d2 a7a5 f2f4 e7h4 e1f1 h4f6 c2c3 c8a6 d3a6 a8a6 a2a4 d8e7 d1c2 a6a8 a1e1 f8e8 e3f2 e7e6 b2b3 e6e7 f2g3 f6h4 g3h4 e7h4');
INSERT INTO evaluations VALUES(8563,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(8563,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(8563,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(8563,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(8563,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(8563,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(8563,36,'cp 68','c2c3 e8g8 d2d4 c8d7 b5a4 a7a5 h2h3 f8e8 d4d5 c6b8 c3c4 b8a6 a4d7 d8d7 b1c3 c7c6 c1g5 h7h6 g5f6 e7f6 a1b1 f6d8 a2a3 d8b6 e1e2 d7e7 e2d2 e8f8 d1c2 b6c5 g2g3 g7g6 b1d1 g8g7 c3a4 c5a7 g1g2');
INSERT INTO evaluations VALUES(8554,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(8554,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(8554,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(8554,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(8554,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(8554,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(8554,36,'cp 44','e8g8 d2d4 c8d7 b5a4 a7a5 h2h3 f8e8 d4d5 c6b8 a4d7 b8d7 c3c4 c7c6 b1c3 d7c5 a2a4 a8c8 d1c2 e8f8 a1b1 h7h6 c1d2 c6d5 c4d5 d8d7 b2b4 a5b4 b1b4 e7d8 b4c4 d8a5 c3b5 a5d2 f3d2 f6h5 g2g3 h5f6 g1g2');
INSERT INTO evaluations VALUES(3354,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(3354,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(3354,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(3354,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(3354,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(3354,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(3354,36,'cp 61','d2d4 c8g4 b5c6 b7c6 d4e5 g4f3 d1f3 d6e5 b1d2 d8d6 d2c4 d6e6 f3e2 f6d7 c4a5 e7c5 b2b3 c5b6 a5c4 a7a5 c1g5 b6c5 e1d1 d7b6 g5e3 c5e3 c4e3 g7g6 b3b4 e6e7 e2c2 f8d8 d1d8 e7d8 g2g3 h7h5 h2h4 d8e7 a2a3 e7e6');
INSERT INTO evaluations VALUES(3349,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(3349,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(3349,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(3349,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(3349,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(3349,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(3349,36,'cp 62','c8g4 b5c6 g4f3 d1f3 b7c6 d4e5 d6e5 b1d2 d8d6 f3e2 f6d7 d2c4 d6e6 c4a5 e7c5 c1e3 c5b6 a5c4 d7f6 h2h3 b6e3 c4e3 f6e4 e2c4 e4f6 c4e6 f7e6 e3c4 e5e4 e1e2 f8d8 c4d2 d8d6 d2e4 f6e4 e2e4 a8d8 a1e1');
INSERT INTO evaluations VALUES(23688,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(23688,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(23688,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(23688,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(23688,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(23688,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(23688,36,'cp 65','b5c6 b7c6 d4e5 g4f3 d1f3 d6e5 b1d2 d8d6 d2c4 d6e6 f3e2 f6d7 c4a5 e7c5 c1e3 c5b6 a5c4 a7a5 b2b3 b6e3 c4e3 d7c5 f2f3 f8d8 e2c4 e6c4 e3c4 f7f6 e1d1 a5a4 b3b4 c5e6 c4a5 c6c5 a5c6 d8d1 a1d1 g8f7 g1f2 h7h5');
INSERT INTO evaluations VALUES(53571,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(53571,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(53571,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(53571,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53571,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53571,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(53571,36,'cp 7','g8f6 g1f3 b8d7 d1c2 f8d6 f1d3 e8g8 e1g1 d5c4 d3c4 a7a6 f1d1 b7b5 c4d3 d8c7 c3e4 f6e4 d3e4 d7f6 e4d3 c6c5 d4c5 c7c5 c2c5 d6c5 c1d2 c8b7 d2a5 c5e7 a2a4 b5a4 a1a4 f8c8 f3e5 b7d5 d1a1 d5b3');
INSERT INTO evaluations VALUES(54716,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(54716,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(54716,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(54716,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(54716,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(54716,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(54716,36,'cp 141','c2c4 e7e6 a2a3 f6f5 c1f4 f8e7 b1c3 c7c6 e2e3 g8f6 d1c2 e8g8 f1d3 f6e4 h2h3 b8d7 e1g1 g7g5 f4h2 e7d6 c3e2 d8e7 c4c5 d6h2 f3h2 e4f6 b2b4 g5g4 h3g4 f5g4 e3e4 d5e4 d3e4 f6e4 c2e4 d7f6 e4e5 e7g7');
INSERT INTO evaluations VALUES(54701,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(54701,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(54701,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(54701,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(54701,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(54701,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(54701,36,'cp 148','e7e6 a2a3 f8e7 c1f4 c7c6 e2e3 f6f5 f1d3 g8f6 b1c3 e8g8 h2h3 f6e4 d1c2 b8d7 e1g1 g7g5 f4h2 f8f7 c3e2 e7d6 h2d6 e4d6 c4c5 d6e4 f3h2 e4f6 a1e1 a7a5 f2f3 d7f8 b2b4 f8g6 b4b5 c8d7 b5c6 d7c6 a3a4');
INSERT INTO evaluations VALUES(56792,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(56792,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(56792,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(56792,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(56792,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(56792,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(56792,36,'cp 151','a2a3 f6f5 c1f4 g8f6 e2e3 f8e7 f1d3 c7c6 h2h3 f6e4 d1c2 e8g8 b1c3 b8d7 e1g1 g8h8 a1d1 d8e8 c3e2 a7a5 d1c1 e7f6 f4h2 e8e7 e2f4 e7e8 b2b4 g7g5 f4e2 e8e7 b4b5 e7a3 e2g3 e4g3 h2g3 a3e7 c4d5 c6d5 g3c7 a5a4');
INSERT INTO evaluations VALUES(43180,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(43180,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43180,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43180,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(43180,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(43180,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43180,36,'cp 14','d7d5 c4d5 f6d5 g1f3 c7c5 f1c4 d5b6 c4b3 e8g8 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 c1g5 c5b3 a2b3 f8e8 f1d1 b6c8 g2g4 c8d6 d1e1 h7h6 g5h4 a7a6 e1e2 a8c8 a1e1 c8c7 h4g3 c7d7 g1g2 a6a5 h3h4 b7b5 h4h5 g6g5');
INSERT INTO evaluations VALUES(41444,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(41444,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41444,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(41444,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(41444,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(41444,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41444,36,'cp 21','g1f3 e8g8 c4d5 f6d5 f1c4 d5b6 c4b3 c7c5 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 c1g5 c5b3 a2b3 f8e8 f1d1 b6c8 g2g4 c8d6 d1d2 h7h6 g5h4 g7f6 h4g3 g8g7 f3e3 d8b6 e3b6 a7b6 a1e1 e8d8 g1g2 a8c8');
INSERT INTO evaluations VALUES(41445,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(41445,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(41445,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(41445,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41445,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(41445,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41445,36,'cp 23','e8g8 c4d5 f6d5 f1c4 d5b6 c4b3 c7c5 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 f1d1 c5b3 a2b3 b6c8 g2g4 c8d6 d1e1 f8e8 g1g2 d8d7 c1f4 a7a6 e1e2 a8d8 a1e1 h7h6 f4g3 b7b5 g3e5 a6a5 e5g7 g8g7 e2e5 g7h7 f3e3');
INSERT INTO evaluations VALUES(36730,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36730,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36730,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36730,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36730,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36730,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36730,36,'cp 8','c4d5 f6d5 f1c4 d5b6 c4b3 c7c5 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 c1e3 g7c3 e3c5 c3b2 d5d6 e7d6 a1d1 b2e5 f3b7 d8c8 b7c8 f8c8 c5d6 e5d6 d1d6 a8b8 f1e1 b8b7 e1e4 a7a5 g2g4 a5a4 b3d1 b6c4 d6d4 c4b2 d1e2 c8c2 g1g2 b7c7 g2g3 h7h6 e2f3 a4a3 e4e8 g8g7');
INSERT INTO evaluations VALUES(36710,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36710,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(36710,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(36710,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36710,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36710,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36710,36,'cp 16','f6d5 f1c4 d5b6 c4b3 c7c5 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 f1d1 c5b3 a2b3 b6c8 h3h4 h7h6 f3e2 c8d6 c1e3 h6h5 a1a4 d8d7 a4a7 a8a7 e3a7 d7c7 a7d4 d6f5 d4g7 g8g7 e2e4 c7b6 b3b4 b6f6 g2g3 f5d6 e4d4 f8c8 d1d3 c8c7 g1g2');
INSERT INTO evaluations VALUES(36812,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36812,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36812,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36812,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36812,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36812,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36812,36,'cp 20','f1c4 d5b6 c4b3 c7c5 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 f1d1 c5b3 a2b3 b6c8 h3h4 h7h6 h4h5 g6g5 c1e3 f8e8 f3e2 a7a6 e3d4 g7d4 d1d4 d8b6 d4g4 c8d6 b3b4 b6c7 a1e1 c7d7 c3e4 d6e4 e2e4 a8d8');
INSERT INTO evaluations VALUES(36811,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36811,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(36811,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36811,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36811,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(36811,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36811,36,'cp 21','d5c3 b2c3 c7c5 e1g1 d8c7 d1e2 b8c6 c1a3 c7a5 a3b2 c8g4 h2h3 g4f3 e2f3 a8c8 f1c1 c8c7 a2a4 e7e6 c4f1 f8d8 f3e2 h7h5 e2b5 a7a6 b5c4 d8c8 c4b3 c5d4 c3d4 a5b4 b3a2 c6e7 c1c7 c8c7 a1c1 c7c1');
INSERT INTO evaluations VALUES(36824,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36824,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36824,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(36824,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(36824,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36824,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(36824,36,'cp 20','b2c3 c7c5 e1g1 d8c7 d1e2 b8c6 c1a3 c7a5 a3b2 c8g4 h2h3 g4f3 e2f3 a8c8 f1c1 c8c7 a2a4 e7e6 c4f1 f8c8 f3d1 c6e7 g2g3 h7h5 h3h4 b7b6 d1b3 c5d4 c3d4 c7c1 a1c1 c8c1 b2c1 e7d5 b3d1 g7f8 g1g2 a5b4 d1c2 b4c3 f1d3 c3c2 d3c2 d5b4');
INSERT INTO evaluations VALUES(36823,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36823,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(36823,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36823,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(36823,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36823,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(36823,36,'cp 12','c7c5 e1g1 d8c7 d1e2 b8c6 c1a3 c7a5 a3b2 c8g4 h2h3 g4f3 e2f3 a8c8 f1c1 e7e6 a2a4 f8d8 c4e2 d8d7 g2g3 h7h5 b2a3 c5d4 c3d4 e6e5 d4d5 d7d5 a1b1 d5d7 e2b5 g7f8 a3f8 g8f8 f3f6 f8g8 b5c6 b7c6');
INSERT INTO evaluations VALUES(36042,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(36042,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(36042,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(36042,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36042,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36042,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36042,36,'cp 34','e1g1 d8c7 d1e2 b8c6 c1a3 c7a5 a3b2 c8g4 h2h3 g4f3 e2f3 a8c8 f1c1 e7e6 a2a4 f8d8 c4f1 d8d7 h3h4 h7h5 g2g3 c8d8 f1b5 d8c8 b2a3 g7f8 f3e4 a7a6 b5f1 c8c7 f1g2 c5d4 a3f8 g8f8 c3d4 c6e7 c1c7 d7c7 e4f4 b7b5 a1b1 b5a4 d4d5 e6d5 f4d4 f8g8 b1b8 c7c8');
INSERT INTO evaluations VALUES(36041,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(36041,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(36041,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36041,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36041,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36041,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(36041,36,'cp 30','d8c7 d1e2 b8c6 c1a3 c7a5 a3b2 c8g4 h2h3 g4f3 e2f3 a8c8 f1c1 e7e6 a2a4 f8d8 g2g3 h7h5 h3h4 d8d7 c4f1 c6e7 f3d1 b7b6 d1d3 d7c7 d3b5 a5b5 a4b5 e7f5 c1d1 c5d4 c3d4 c7c2 b2a3 c8c7 f1g2 c7d7 a3b4 c2b2');
INSERT INTO evaluations VALUES(15879,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(15879,31,'cp 301',NULL);
INSERT INTO evaluations VALUES(15879,32,'cp 294',NULL);
INSERT INTO evaluations VALUES(15879,33,'cp 302',NULL);
INSERT INTO evaluations VALUES(15879,34,'cp 294',NULL);
INSERT INTO evaluations VALUES(15879,35,'cp 283',NULL);
INSERT INTO evaluations VALUES(15879,36,'cp 312','c4d5 d8f6 d5c6 f6c6 f3c6 b7c6 d2d3 c8e6 b1c3 e6d5 f2f3 a7a5 a2a4 g7g5 h2h3 e8d7 c1d2 f8d6 c3e4 d5e4 f3e4 h7h5 e1e2 g5g4 h3h4 d6b4 d2b4 a5b4 h1f1 d7e6 a4a5 c6c5 f1f5 c5c4 d3c4 c7c5');
INSERT INTO evaluations VALUES(15878,30,'cp 308',NULL);
INSERT INTO evaluations VALUES(15878,31,'cp 290',NULL);
INSERT INTO evaluations VALUES(15878,32,'cp 311',NULL);
INSERT INTO evaluations VALUES(15878,33,'cp 307',NULL);
INSERT INTO evaluations VALUES(15878,34,'cp 302',NULL);
INSERT INTO evaluations VALUES(15878,35,'cp 300',NULL);
INSERT INTO evaluations VALUES(15878,36,'cp 296','d8f6 d5c6 b7c6 f3h5 f6f7 h5f7 e8f7 d2d3 f8c5 b1c3 a8b8 b2b3 c8f5 c1e3 c5a3 e1d2 a7a5 f2f3 h7h5 h1e1 h8e8 e3f2 g7g5 f2g3 a3d6 h2h3 h5h4 g3h2 b8d8 a2a3 d8b8 e1e3 b8d8 c3e4 f7g6 a3a4 f5e4 e3e4 c6c5');
INSERT INTO evaluations VALUES(1031,30,'cp 316',NULL);
INSERT INTO evaluations VALUES(1031,31,'cp 321',NULL);
INSERT INTO evaluations VALUES(1031,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(1031,33,'cp 312',NULL);
INSERT INTO evaluations VALUES(1031,34,'cp 301',NULL);
INSERT INTO evaluations VALUES(1031,35,'cp 319',NULL);
INSERT INTO evaluations VALUES(1031,36,'cp 288','d5c6 b7c6 f3h5 f6f7 h5f7 e8f7 d2d3 h7h5 c1d2 h5h4 f2f3 a8b8 b2b3 h4h3 g2g4 c6c5 b1c3 c8b7 e1e2 f7e6 d2e3 f8d6 a1f1 b8f8 c3e4 b7e4 f3e4 f8f1 h1f1 g7g6 f1f3 d6e7 f3g3 a7a6 e3d2 e6d7 g3f3 h8h4 g4g5 h4g4 f3h3 g4g2 e2d1');
INSERT INTO evaluations VALUES(1027,30,'cp 320',NULL);
INSERT INTO evaluations VALUES(1027,31,'cp 299',NULL);
INSERT INTO evaluations VALUES(1027,32,'cp 327',NULL);
INSERT INTO evaluations VALUES(1027,33,'cp 303',NULL);
INSERT INTO evaluations VALUES(1027,34,'cp 313',NULL);
INSERT INTO evaluations VALUES(1027,35,'cp 295',NULL);
INSERT INTO evaluations VALUES(1027,36,'cp 322','f6c6 f3c6 b7c6 d2d3 f8c5 b1c3 a7a5 c1e3 c5e3 f2e3 c8e6 e1d2 e8e7 a1f1 h8f8 b2b3 e6d5 h1g1 f8f6 e3e4 d5e6 c3a4 h7h5 d2c3 e7d6 f1f6 g7f6 g1f1 a8g8 g2g3 d6e7 f1f2 g8g4 a4c5 e6c8 d3d4 e5d4 c3d4 h5h4');
INSERT INTO evaluations VALUES(1035,30,'cp 314',NULL);
INSERT INTO evaluations VALUES(1035,31,'cp 332',NULL);
INSERT INTO evaluations VALUES(1035,32,'cp 324',NULL);
INSERT INTO evaluations VALUES(1035,33,'cp 309',NULL);
INSERT INTO evaluations VALUES(1035,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(1035,35,'cp 332',NULL);
INSERT INTO evaluations VALUES(1035,36,'cp 320','f3c6 b7c6 d2d3 f8c5 b1c3 a7a5 c1e3 c5e3 f2e3 c8e6 e1d2 e8e7 a1f1 a5a4 f1f2 h8d8 b2b3 a4b3 a2b3 d8f8 h1f1 f8f2 f1f2 a8a1 c3e4 e6d5 e4c5 a1h1 g2g3 d5e6 d2c3 h7h5 c5e4 h1e1 e4g5 e1b1 b3b4 b1a1 g5e6 e7e6 e3e4 a1a3 c3b2 a3a8 b2b3 a8a1 f2f5 g7g6 f5f2 e6e7');
INSERT INTO evaluations VALUES(1028,30,'cp 319',NULL);
INSERT INTO evaluations VALUES(1028,31,'cp 327',NULL);
INSERT INTO evaluations VALUES(1028,32,'cp 329',NULL);
INSERT INTO evaluations VALUES(1028,33,'cp 319',NULL);
INSERT INTO evaluations VALUES(1028,34,'cp 336',NULL);
INSERT INTO evaluations VALUES(1028,35,'cp 329',NULL);
INSERT INTO evaluations VALUES(1028,36,'cp 333','b7c6 d2d3 c8e6 b1c3 e6d5 f2f3 g7g5 c3e4 f8e7 c1d2 e8d7 e1g1 d5e4 f3e4 a8b8 b2b3 d7e6 g1f2 b8f8 f2e2 h7h5 f1f8 h8f8 a1f1 f8f1 e2f1 g5g4 f1f2 e7h4 f2e2 h4e7 d2e3 a7a6 e3d2 e7c5 h2h3 g4h3 g2h3 c5b6 e2f3 b6c5 f3g3 c5e7 d2a5 e6d7 a5c3 d7e6 c3d2 e6d7 d2a5');
INSERT INTO evaluations VALUES(690,30,'cp 310',NULL);
INSERT INTO evaluations VALUES(690,31,'cp 330',NULL);
INSERT INTO evaluations VALUES(690,32,'cp 326',NULL);
INSERT INTO evaluations VALUES(690,33,'cp 337',NULL);
INSERT INTO evaluations VALUES(690,34,'cp 336',NULL);
INSERT INTO evaluations VALUES(690,35,'cp 341',NULL);
INSERT INTO evaluations VALUES(690,36,'cp 337','d2d3 c8e6 b1c3 a7a5 a2a4 e8d7 h2h4 f8c5 c1d2 c5d4 f2f3 h7h5 a1b1 e6f5 c3e4 h8f8 e4g3 f5g6 b2b3 f8f6 e1f1 f6d6 f1e2 d6e6 g3e4 g6e4 d3e4 e6g6 d2g5 c6c5');
INSERT INTO evaluations VALUES(43163,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43163,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(43163,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(43163,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(43163,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(43163,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43163,36,'cp 20','f1g2 c7c5 e2e3 e7e6 g1e2 d7d5 b2b3 e8g8 d2d4 c5d4 e3d4 b8c6 e1g1 c8d7 a1b1 a8c8 c3b5 f6e8 c1f4 b7b6 c4d5 e6d5 d1d2 d7f5 b1c1 d8d7 f1d1 f5h3 g2f3 h3g4 f3g4 d7g4 g1g2 g4d7 h2h3 a7a6 b5c3 b6b5 f4h6 g7h6 d2h6 c6e7 d1e1 e7f5');
INSERT INTO evaluations VALUES(48623,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48623,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(48623,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(48623,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(48623,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48623,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(48623,36,'cp -19','f8e7 g1f3 e8g8 f1d3 f8e8 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 c7c6 f2f4 d8b6 g1g2 b6b2 d3g6 f7g6 d1d3 e7b4 c3e2 b2a3 f1b1 f6e4 d3a3 b4a3 b1b7 b8a6 g5h4 a6b4 c2c3 b4c2 a1h1 a3d6 h4f2 c6c5 d4c5 d6e5 f4e5 e8e5 e2f4 e4c3');
INSERT INTO evaluations VALUES(41067,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(41067,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41067,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(41067,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(41067,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(41067,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41067,36,'cp 0','g1f3 e8g8 f1d3 h7h6 g5f4 b8c6 h2h3 c6b4 e1g1 c7c6 f1e1 b4d3 d1d3 f8e8 c3e2 e7f8 a2a4 f6e4 e2g3 d8f6 f4e5 f6g6 d3b3 a7a5 c2c3 e8e7 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(41063,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(41063,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(41063,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(41063,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(41063,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(41063,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(41063,36,'cp -10','e8g8 f1d3 h7h6 g5f4 b8c6 h2h3 c6b4 e1g1 c7c6 c3e2 f8e8 e2g3 b4d3 d1d3 e7f8 f1e1 e8e1 a1e1 c8e6 c2c3 d8a5 a2a3 a8e8 d3c2 a5d8 e1e3 f6d7 a3a4 a7a5 g3h5 b7b5 a4b5 c6b5 c2d3 d8b6 f3e5 d7e5');
INSERT INTO evaluations VALUES(36388,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(36388,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(36388,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(36388,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(36388,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(36388,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(36388,36,'cp -9','f1d3 h7h6 g5f4 b8c6 h2h3 c6b4 e1g1 c7c6 c3e2 f8e8 e2g3 b4d3 d1d3 e7f8 f1e1 e8e1 a1e1 c8e6 c2c3 d8b6 d3c2 a7a5 f3e5 a8e8 e5d3 e6c8 e1e8 f6e8 d3c5 b6b5 a2a4 b5b6 c2d2 f8c5 d4c5 b6c5');
INSERT INTO evaluations VALUES(36392,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(36392,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36392,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(36392,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(36392,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36392,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(36392,36,'cp -8','b8c6 h2h3 h7h6 g5f4 c6b4 e1g1 c7c6 c3e2 f8e8 e2g3 b4d3 d1d3 e7f8 a2a4 f6e4 f1e1 d8f6 f4e5 f6g6 a4a5 e4g5 d3b3 g5f3 b3f3 c8e6 f3b3 f7f6 e5f4 g6f7 f4d2 f8d6 d2b4 d6g3 b3g3 e6f5 a1c1 a7a6 c2c3 f5g6 b4d6 e8e1 c1e1 a8e8');
INSERT INTO evaluations VALUES(55289,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55289,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55289,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(55289,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(55289,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55289,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55289,36,'cp -2','f3e4 g8f6 e4e5 f6d5 d2d4 c8f5 g1f3 e7e6 b1c3 f8b4 c1d2 b4e7 c3d5 d8d5 d2e3 b8c6 c2c3 e8c8 d1a4 f7f6 f1b5 d5e4 e1f2 f6e5 b5c6 e4c6 a4c6 b7c6 f3e5 h8f8 h1e1 e7d6 f2g1 d6e5 d4e5 d8d5 b2b4 a7a6 e3d4 f8d8 a1d1 h7h6');
INSERT INTO evaluations VALUES(55280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55280,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(55280,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(55280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55280,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(55280,36,'cp 0','e7e5 g1f3 g8f6 c2c3 f8d6 d2d4 b8c6 f1b5 c8g4 d4e5 d6e5 b5c6 b7c6 d1d8 a8d8 b1d2 e5f4 d2b3 f4c1 a1c1 f6e4 f3e5 g4e6 e1g1 c6c5 e5f7 e6f7 c1e1 e8g8 e1e4 f7b3 a2b3 d8d2 e4e5 d2b2 e5c5 b2b3 c5c7 a7a5 f1f8 g8f8 c7a7 b3c3 a7a5 f8f7 g1f2 c3c8');
INSERT INTO evaluations VALUES(56616,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(56616,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(56616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56616,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56616,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56616,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56616,36,'cp 0','g1f3 g8f6 c2c3 f8d6 d2d4 b8c6 f1b5 c8g4 d4e5 d6e5 d1d8 a8d8 b1d2 e5f4 d2b3 f4c1 b5c6 b7c6 a1c1 f6e4 f3e5 g4e6 e1g1 c6c5 e5f7 e6f7 c1e1 e8g8 e1e4 f7b3 a2b3 d8d2 e4e5 d2b2 e5c5 b2b3 c5c7 f8f1 g1f1 a7a5 c3c4 a5a4 c7a7 a4a3 c4c5 b3c3 f1e2 h7h5 c5c6 c3c6 a7a3 c6c7 a3a8 g8f7');
INSERT INTO evaluations VALUES(56613,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56613,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56613,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56613,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56613,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56613,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56613,36,'cp 0','g8f6 c2c3 f8d6 d2d4 b8c6 f1b5 c8g4 d4e5 d6e5 d1d8 a8d8 b1d2 e5f4 d2b3 f4c1 a1c1 f6e4 b5c6 b7c6 f3e5 g4e6 e1g1 c6c5 e5f7 e6f7 c1e1 e8g8 e1e4 f7b3 a2b3 d8d2 e4e7 d2b2 e7c7 h7h6 f1f8 g8f8 c7c5 b2b3 c5a5 b3c3 a5a7 c3c2 h2h3 c2c5 a7b7 c5c4 b7a7');
INSERT INTO evaluations VALUES(48940,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48940,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48940,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48940,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48940,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48940,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48940,36,'cp 0','c2c3 f8d6 d2d4 b8c6 f1b5 c8g4 d4e5 d6e5 d1d8 a8d8 b1d2 e5f4 d2b3 f4c1 a1c1 f6e4 b5c6 b7c6 f3e5 g4e6 e1g1 c6c5 e5f7 e6f7 c1e1 e8g8 e1e4 f7b3 a2b3 d8d2 e4e5 d2b2 f1f8 g8f8 e5c5 b2b3 c5c7 a7a5 c7a7 b3c3 a7a5 f8f7 g1f2 h7h6 a5a6 c3c2 f2f3 f7g8');
INSERT INTO evaluations VALUES(44895,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(44895,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(44895,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44895,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(44895,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(44895,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(44895,36,'cp -31','d5c4 a2a4 f6d5 f4g3 e7e6 e2e4 d5c3 b2c3 b7b5 d1b1 d8a5 b1c1 b5b4 f1c4 b4c3 g1e2 f8b4 e1g1 e8g8 c1c2 c8a6 c4a6 b8a6 a1c1 f8d8 f1d1 a5b6 e2c3 d8d4 d1d4 b6d4 c3e2 d4c5 c2b2 c5a5 e2d4 a5a4 d4c6');
INSERT INTO evaluations VALUES(45018,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(45018,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(45018,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(45018,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(45018,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45018,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(45018,36,'cp -39','a2a4 f6d5 f4g3 e7e6 e2e4 d5c3 b2c3 b7b5 g1f3 a7a6 f1e2 f8e7 e1g1 c8b7 a1b1 d8c8 d1d2 b8d7 d2b2 a8a7 f1e1 e8g8 e2d1 d7f6 d1c2 c8a8 g3e5 h7h6 h2h3 f8c8 f3d2 c6c5 a4b5 a6b5 b2b5 c5d4 e5d4 b7c6 b5c4 c6e4 c2e4 f6e4 c4d3');
INSERT INTO evaluations VALUES(16436,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(16436,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(16436,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(16436,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(16436,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(16436,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(16436,36,'cp 26','b1c3 f8b4 d2d3 b4c5 f1e2 d7d6 a2a3 c8g4 b2b4 c5b6 c1e3 e8g8 e1g1 g4f3 e2f3 c6d4 a1b1 c7c6 g2g3 d4f3 d1f3 f6d7 e3b6 d8b6 g1g2 h7h6 f1d1 a7a5 b4b5 b6c5 a3a4 a8d8 b5c6 b7c6 b1b7');
INSERT INTO evaluations VALUES(19970,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(19970,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(19970,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(19970,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(19970,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(19970,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19970,36,'cp 25','d2d4 e7e5 d4d5 c6d4 f1d3 g8f6 h2h3 f6d7 b1c3 b7b6 a1b1 g7g6 e1g1 f8e7 a2a3 e8g8 c1h6 f8e8 b2b4 e7f8 h6e3 f7f5 e4f5 g6f5 b4c5 b6c5 f3g5 d7f6 d3c2 h7h6 g5e6 d4e6 d5e6 c8e6 c3d5 f5f4 e3d2 f8g7 d5f6 d8f6 c2e4 a8c8 d1c2 g8h8 b1b7');
INSERT INTO evaluations VALUES(19986,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19986,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(19986,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(19986,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19986,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(19986,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(19986,36,'cp 18','e7e5 d4d5 c6d4 f3d4 c5d4 b1d2 d8c7 a2a4 a7a5 b2b3 g8f6 c4c5 d6c5 f1b5 c8d7 b5d3 f8d6 d2c4 e8g8 e1g1 b7b5 c4d6 c7d6 a4b5 d6b6 c1g5 d7b5 d3b5 b6b5 g5f6 g7f6 d1f3 b5b6 h2h3 g8g7 a1a3 f8b8 f1a1 a8a6 f3g4 g7h8 g4h5 a6a7 h3h4 b8g8 h5f3 g8b8 f3f5 a7a6');
INSERT INTO evaluations VALUES(39223,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(39223,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(39223,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(39223,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(39223,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39223,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(39223,36,'cp 35','b1c3 b8c6 f1g2 g8f6 g1f3 e8g8 e1g1 d7d6 e2e3 c5b6 d2d3 h7h6 b2b3 c8e6 h2h3 f8e8 c1b2 d6d5 c3a4 d5c4 d3c4 e5e4 f3d4 e6d7 a4b6 a7b6 d4b5 c6e5 a2a4 h6h5 b2d4 d7b5 a4b5 d8d6 d1c2 e5f3 g2f3 a8a1 f1a1');
INSERT INTO evaluations VALUES(39217,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(39217,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(39217,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(39217,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(39217,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(39217,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(39217,36,'cp 29','b8c6 f1g2 g8f6 g1f3 e8g8 e1g1 d7d6 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 a4b6 a7b6 f1e1 f8e8 e3e4 c8h3 b2b3 c7c6 d3d4 d6d5 c4d5 c6d5 d4e5 d5e4 d1d8 a8d8 f3h1 e8e5 c1b2 e5d5 b2f6 g7f6 h1e4 d5d2 e4b7 h3e6 b7g2 h7h6 b3b4 d2a2');
INSERT INTO evaluations VALUES(6463,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(6463,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(6463,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6463,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6463,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(6463,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(6463,36,'cp 31','g1f3 g8f6 f1g2 e8g8 e1g1 d7d6 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 a4b6 a7b6 a2a4 c7c6 f3g2 b6b5 c4b5 c6b5 c1d2 b5a4 a1a4 c8d7 a4a8 d8a8 d2b4 a8a6 d1b3 f8d8 b3a3 a6a3 b2a3 d7c6 e3e4 h7h6 f1d1 d6d5');
INSERT INTO evaluations VALUES(6462,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(6462,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(6462,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(6462,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(6462,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6462,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(6462,36,'cp 32','g8f6 f1g2 e8g8 e1g1 d7d6 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 a4b6 a7b6 f1e1 f8e8 e3e4 c8h3 a2a4 h7h6 c1e3 d8d7 b2b3 f6h7 d3d4 d7e6 e3c1 e6d7 f3h1 h7g5 d1d3 g5e6 c1b2 d7e7 h1g2 h3g2 g1g2 e6d4 b2d4 e5d4');
INSERT INTO evaluations VALUES(10178,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(10178,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(10178,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(10178,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(10178,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(10178,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(10178,36,'cp 49','f1g2 d7d6 e1g1 e8g8 e2e3 c8f5 d2d3 a7a5 h2h3 h7h6 a2a3 d8d7 g1h2 f5h7 b2b3 d6d5 d3d4 e5d4 e3d4 c5a7 f3e5 d7f5 e5c6 b7c6 c1e3 f8e8 a1c1 f5d7 f1e1 a8d8 c4c5 a7b8 d1d2 f6e4 c3e4 d5e4 d2c3 f7f5');
INSERT INTO evaluations VALUES(6614,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(6614,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(6614,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6614,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(6614,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(6614,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(6614,36,'cp 13','b4c3 d2c3 h7h6 c1e3 d7d6 f3d2 d8f6 d1e2 c8e6 e1c1 e6c4 d2c4 g8e7 f2f3 f6e6 b2b3 f7f5 c1b2 f5f4 e3f2 b7b5 c4d2 a7a6 g2g3 e8g8 h1g1 g8h7 g3f4 f8f4 f2e3 f4f7 g1g3 a8e8 d1g1 e8f8 e2g2');
INSERT INTO evaluations VALUES(6635,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(6635,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(6635,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(6635,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(6635,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(6635,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(6635,36,'cp 14','d2c3 g8f6 d1e2 d7d6 c4d3 c8e6 h2h3 d8e7 c1e3 a7a6 f3g5 e6d7 e1c1 h7h6 g5f3 d6d5 e4d5 f6d5 e3d2 e8c8 d3e4 e7c5 h1e1 f7f6 f3h4 c8b8 g2g4 d7e6 h4f5 g7g5 d2e3 d5e3 e2e3 c5e3 f5e3 e6a2');
INSERT INTO evaluations VALUES(6634,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(6634,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(6634,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(6634,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(6634,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(6634,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(6634,36,'cp 12','g8f6 d1e2 d7d6 h2h3 d8e7 c4d3 c8d7 c1d2 h7h6 e1c1 e8c8 c1b1 h8e8 d2e3 f6h5 d3b5 c8b8 f3d2 h5f4 e2f3 a7a6 b5c6 d7c6 e3f4 e5f4 f3f4 d6d5 f2f3 d5e4 f3e4 b7b6 f4g3 f7f6 h1e1 a6a5 e4e5');
INSERT INTO evaluations VALUES(10429,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(10429,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(10429,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(10429,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(10429,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(10429,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(10429,36,'cp 32','d1e2 h7h6 b2b3 d7d6 c1b2 c8e6 c4b5 f6d7 e1c1 a7a6 b5c6 b7c6 h2h3 d8b8 c3c4 b8b6 a2a4 a6a5 f3e1 e8c8 b2c3 c8b7 e1d3 f7f6 f2f4 c6c5 f4f5 e6f7 e2d2 d8a8 g2g4 d7b8 h3h4 b8c6 h1h3 c6d4 d3b2 d4c6 d2e2 c6d4 e2g2');
INSERT INTO evaluations VALUES(56553,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(56553,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(56553,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(56553,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(56553,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(56553,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(56553,36,'cp 25','f1c4 e5d4 e3d4 f8d6 g1f3 g8f6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 b1c3 b8d7 c1f4 d7b6 c4b3 c7c6 f1e1 f8e8 h2h3 c8f5 g2g4 f5g6 g4g5 f6d5 c3d5 b6d5 b3d5 c6d5 a1c1 f7f6 c1c7 f6g5 e1e7 e8e7 c7e7 g5f4 e7b7 g6f5 h3h4 h7h6 f3e5 a7a5 b7c7');
INSERT INTO evaluations VALUES(56542,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(56542,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(56542,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(56542,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(56542,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(56542,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(56542,36,'cp 17','e5d4 e3d4 f8d6 g1f3 g8f6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 b1c3 e7d6 c1g5 b8d7 c4b3 a7a5 a2a3 d7b6 f1e1 a5a4 b3a2 c7c6 g5f6 g7f6 d4d5 c6d5 a2d5 b6d5 c3d5 g8g7 a1d1 c8e6 d5f6 d6h2 g1h1 g7f6 h1h2');
INSERT INTO evaluations VALUES(56643,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(56643,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(56643,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(56643,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(56643,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(56643,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(56643,36,'cp 30','e3d4 f8d6 g1f3 g8f6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 b1c3 b8d7 c1g5 e7d6 c4b3 a7a5 f1e1 h7h6 g5h4 d7b6 h4f6 g7f6 c3e4 a5a4 b3c2 d6e7 e4c5 f8e8 f3h4 g8f8 c2f5 c8f5 h4f5 e7c5 d4c5 b6d7 c5c6 b7c6 e1e8 a8e8 a1d1 d7e5 f5h6');
INSERT INTO evaluations VALUES(56641,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(56641,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(56641,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56641,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(56641,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(56641,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(56641,36,'cp 21','f8d6 g1f3 g8f6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 b1c3 e7d6 c1g5 b8d7 c4b3 a7a5 a2a3 h7h6 g5h4 d7b6 h4f6 g7f6 c3e4 d6e7 f1e1 a5a4 e4g3 a4b3 e1e7 b6d5 e7e1 c8e6 f3d2 d5f4 e1e4 f4d3 e4e3 d3b2');
INSERT INTO evaluations VALUES(50663,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50663,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(50663,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50663,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(50663,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(50663,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(50663,36,'cp 37','d2d4 c5d4 f3d4 g8f6 c1e3 e7e5 d4b3 c8e6 f2f3 h7h5 f1e2 b8d7 e1g1 a8c8 a2a4 f8e7 a4a5 e8g8 d1d2 h5h4 f1d1 d8c7 b3c1 h4h3 g2h3 f8d8 g1h1 d7c5 c3d5 f6d5 e4d5 e6f5 b2b4 c5d7 c2c4 e5e4 f3f4 e7f6 c1b3 f6a1 d1a1 d8e8 b3d4');
INSERT INTO evaluations VALUES(50659,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50659,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(50659,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(50659,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(50659,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50659,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50659,36,'cp 24','c5d4 f3d4 g8f6 c1e3 e7e5 d4b3 c8e6 f2f3 h7h5 d1d2 f8e7 f1e2 e8g8 c3d5 e6d5 e4d5 d8c7 e1g1 a6a5 a2a4 b8d7 f1c1 d7b6 c2c4 b6d7 c1f1 b7b6 g2g3 a8e8 a1c1 d7c5 b3c5 b6c5 e2d3 c7d7 g1g2 d7a4 d3f5');
INSERT INTO evaluations VALUES(50691,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50691,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(50691,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(50691,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50691,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(50691,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(50691,36,'cp 20','f3d4 g8f6 c1e3 e7e5 d4b3 c8e6 f2f4 e5f4 e3f4 b8c6 b3d4 d8b6 d4e6 f7e6 d1d3 c6b4 d3e2 f8e7 g2g3 b4d5 c3d5 e6d5 e1c1 d5e4 e2c4 d6d5 d1d5 f6d5 c4d5 a8d8 d5e4 b6d4 e4d4 d8d4 c2c3 d4d8 f1g2 b7b5 g2b7 b5b4 f4c7 b4c3 c7d8 c3b2 c1b2 e8d8 h1d1 d8c7');
INSERT INTO evaluations VALUES(55039,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(55039,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55039,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(55039,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55039,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(55039,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55039,36,'cp 1','e7e6 g1f3 g8f6 d2d4 b7b6 b2b3 c8b7 b1d2 b8d7 c1b2 f8e7 c4d5 f6d5 a1c1 e8g8 f1e2 c7c5 e1g1 c5d4 f3d4 a8c8 c1c8 d8c8 d1a1 c8a8 h2h3 d5f6 d4f3 h7h6 a2a3 f8c8 f1c1 a8b8 b3b4 c8c1 a1c1 e7d6 g1h1');
INSERT INTO evaluations VALUES(55805,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(55805,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(55805,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(55805,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(55805,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(55805,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(55805,36,'cp 0','g1f3 g8f6 d2d4 f8e7 b1c3 e8g8 b2b3 c7c5 d4c5 d5c4 f1c4 d8d1 e1d1 e7c5 c1b2 c5e7 d1e2 a7a6 c4d3 b8d7 h1d1 b7b6 a1c1 d7c5 f3e5 c8b7 d3b1 a6a5 g2g4 f8d8 f2f4 b7a6 e2f3 a6b7');
INSERT INTO evaluations VALUES(55806,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55806,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55806,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55806,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55806,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55806,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55806,36,'cp 0','g8f6 d2d4 f8e7 b2b3 e8g8 f1d3 c7c5 e1g1 c5d4 e3d4 b7b6 c1b2 b8c6 f1e1 c8b7 b1c3 a8c8 a1c1 c6a5 f3e5 a5c6');
INSERT INTO evaluations VALUES(48240,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48240,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48240,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48240,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48240,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48240,36,'cp 3','b2b3 f8e7 c1b2 e8g8 b1c3 c7c5 c4d5 f6d5 d1c2 b7b6 c3d5 e6d5 d2d4 b8c6 f1e2 d8d6 e1g1 f8d8 f1c1 a7a5 h2h3 d6h6 e2b5 c8a6 b5a6 a8a6 c1d1 h6e6 d4c5 b6c5 a1c1 e6g6 c2e2 a6a8 f3e5 c6e5 b2e5');
INSERT INTO evaluations VALUES(48231,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48231,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(48231,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48231,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48231,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(48231,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48231,36,'cp 0','f8e7 c1b2 e8g8 b1c3 c7c5 c4d5 f6d5 d1c2 b7b6 c3d5 e6d5 d2d4 b8c6 f1e2 a7a5 e1g1 d8d6 f1d1 f8d8 a2a3 d6g6 c2g6 h7g6 d4c5 b6c5 e2b5 c8b7 a1c1 c6a7 b5e2 a7c6');
INSERT INTO evaluations VALUES(17808,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(17808,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(17808,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(17808,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(17808,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(17808,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(17808,36,'cp 37','e1g1 g8f6 c2c3 f6e4 d2d4 c8d7 f1e1 e4f6 a4c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 d2c4 d6e5 c4e5 a6a5 e5d3 a8a6 d3c5 a6b6 b2b3 d8c8 e1e5 f6g4 e5e2 h8e8 f2f3 g4f6 g5f6 g7f6 g1f2 b6c6 c5e4 f6f5 e4f6 e8d8 f6h7 c6c3 h7g5 f5f4');
INSERT INTO evaluations VALUES(20437,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(20437,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(20437,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(20437,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(20437,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(20437,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(20437,36,'cp 50','c3d4 g8e7 f1b5 c8d7 b1c3 d8b6 b5a4 a8c8 e1g1 e7f5 c3e2 f8e7 a2a3 c6a5 a4c2 d7b5 f1e1 e8g8 b2b4 b5e2 e1e2 a5c6 c2f5 e6f5 d1d3 a7a5 e2b2 a5b4 a3b4 c8a8 a1b1 f8c8 h2h4 g7g6 h4h5 c6d8 c1e3 d8e6 g2g3 b6d8 b4b5 d8d7 e3h6 e7f8 h6f8 e6f8 h5h6 f8e6 b2c2');
INSERT INTO evaluations VALUES(23437,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(23437,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(23437,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(23437,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(23437,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(23437,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(23437,36,'cp 50','e7e5 d4d5 c6e7 c2c4 e7g6 b1c3 a7a5 a2a3 f8c5 c3e4 b7b6 c1d2 g8f6 e4c5 b6c5 h2h4 g6e7 f1d3 d7d6 g1e2 e8g8 e3e4 c7c6 f2f3 c6d5 c4d5 d8b6 h4h5 h7h6 e2c3 b6b2 a3a4 b2b6 a1b1 b6d8 d1c2 c8a6 d3a6 a8a6 c3b5');
INSERT INTO evaluations VALUES(22759,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(22759,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(22759,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(22759,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(22759,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(22759,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(22759,36,'cp 46','d4d5 c6e7 g1e2 d7d6 e3e4 f7f5 b1c3 g7g6 e2g3 g8f6 e4f5 e7f5 f1b5 c8d7 d1d3 f8g7 g3e4 f6e4 c3e4 e8g8 b5d7 d8d7 e1g1 h7h6 c2c4 f8f7 c1e3 b7b6 a1e1 a8f8 b2b3 g8h8 b3b4 f5d4 f2f3 d4f5 e3f2');
INSERT INTO evaluations VALUES(22661,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(22661,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(22661,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(22661,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(22661,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(22661,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(22661,36,'cp 57','c6e7 g1e2 f7f5 b1c3 d7d6 e3e4 g7g6 e4f5 e7f5 e2g3 g8f6 f1b5 c8d7 d1d3 f8g7 b5d7 d8d7 g3e4 e8g8 e4f6 f8f6 c3e4 f6f8 e1g1 b7b6 c1e3 a7a5 b2b3 a5a4 c2c4 f5d4 f2f3 g8h8 a1e1 d7f5 e3d4 e5d4 e1e2 f8e8 f1e1 a4b3');
INSERT INTO evaluations VALUES(23377,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(23377,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(23377,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(23377,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(23377,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(23377,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(23377,36,'cp 57','g1e2 d7d6 e3e4 g7g6 h2h4 h7h5 c2c4 f8g7 e2c3 f7f5 e4f5 e7f5 c1g5 g8f6 b1d2 d8d7 f1d3 f6h7 d2e4 h7g5 h4g5 d7e7 d1d2 c8d7 c3b5 d7b5 c4b5 e8c8 e1c1 e7f7 d3c4 f5d4 c1b1 f7f5 c4d3 f5f7 e4c3 h5h4 a2a4 d8f8 c3e4 f7d5');
INSERT INTO evaluations VALUES(50187,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50187,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(50187,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(50187,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(50187,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(50187,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(50187,36,'cp 14','f8g7 g1f3 e8g8 f1e2 c7c5 d4d5 d7d6 b1c3 e7e6 f3d2 f8e8 e1g1 b8a6 e3e4 a6c7 a2a4 b7b6 e2d3 c7a6 d2f3 a6b4 h2h3 c8a6 d5e6 f7e6 d3e2 a6b7 c1g5 b4c6 d1d2 e6e5 a1d1 c6d4 f3d4 e5d4');
INSERT INTO evaluations VALUES(48792,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(48792,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(48792,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48792,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(48792,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(48792,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(48792,36,'cp 8','c4d5 f6d5 f1g2 d5b6 g1f3 b8c6 d2d3 f8c5 c1d2 e8g8 c3e4 c5e7 a1c1 c8e6 e1g1 a7a5 e4c5 e7c5 c1c5 d8d6 c5b5 e6a2 b2b3 a5a4 b3a4 b6a4 b5b7 a4c5 b7b5 a8b8 b5b8 f8b8 d2e3 a2d5 e3c5 d6c5 d1d2 d5f3 g2f3 c6d4 d2e3 h7h6 f1a1 c5c2 e3e5');
INSERT INTO evaluations VALUES(48603,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(48603,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48603,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(48603,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(48603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48603,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48603,36,'cp 6','f6d5 f1g2 d5b6 g1f3 b8c6 d2d3 f8e7 a2a3 a7a5 c1e3 e8g8 e1g1 c8e6 e3b6 c7b6 e2e3 a8c8 d3d4 e5d4 e3d4 e6g4 d1a4 g4f3 g2f3 c6d4 f3b7 c8b8 a1d1 b6b5 a4d4 d8d4 d1d4 b8b7 c3d5 e7c5 d4d2 b5b4 a3b4 a5b4 d2d3 f8d8 f1d1');
INSERT INTO evaluations VALUES(49017,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(49017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49017,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(49017,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(49017,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(49017,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49017,36,'cp 5','f1g2 d5b6 g1f3 b8c6 d2d3 f8e7 a2a3 a7a5 c1e3 e8g8 e1g1 c8e6 e3b6 c7b6 d1a4 h7h6 a1d1 c6d4 f3d4 e5d4 c3e4 e6g4 f1e1 a8c8 d1b1 c8c7 e4d2 c7c5 g2b7 b6b5 a4d1 d8b6 b7f3 g4h3 d2e4 c5c7 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(57584,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(57584,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(57584,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(57584,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(57584,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(57584,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(57584,36,'cp 42','e7e6 g1f3 g8f6 f1e2 b7b6 c2c4 c8b7 e1g1 f8e7 f3g5 e8g8 b1c3 c7c5 b2b3 f6e8 g5f3 e8f6 c1b2 f6e4 a1c1 e4c3 b2c3 d7d6 f3e1 e7f6 e1d3 d8e7 d3f4 g7g5 f4h5 f6h8 e2f3 b8c6 d4d5 c6e5 f3e2 e5g6 c3h8 g8h8 d5e6 e7e6');
INSERT INTO evaluations VALUES(57358,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(57358,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(57358,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(57358,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(57358,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(57358,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(57358,36,'cp 61','g1h3 b7b6 f1d3 b8c6 e1g1 g8f6 c2c4 g7g6 b2b3 f8g7 c1a3 d7d6 b1c3 e8g8 a1c1 d8e7 d1d2 c8b7 d4d5 c6d8 c1e1 e6e5 f2f4 e5e4 d3c2 a7a6 e1c1 b6b5 h3f2 d8f7 c3e2 c7c5 d5c6 b7c6 f1d1 c6b7 a3b2 f8e8');
INSERT INTO evaluations VALUES(22791,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22791,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22791,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22791,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(22791,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(22791,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(22791,36,'cp 0','d2d4 e5d4 c3d4 d7d5 e4d5 d8d5 g1f3 c8g4 f1e2 f8b4 b1c3 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c3a4 e8c8 c1e3 e7f5 h1d1 b4e7 a1c1 e7f6 c1c4 h8e8 e2f3 f5h4 f3g3 h4f5');
INSERT INTO evaluations VALUES(22764,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(22764,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(22764,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(22764,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22764,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22764,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22764,36,'cp 0','e5d4 c3d4 d7d5 e4d5 f8b4 b1c3 d8d5 g1f3 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 a1c1 e7f5 h1d1 h8e8 c3a4 d8d5 c1c4 b4d6 h2h3 f5e3 f2e3 d5g5 e2f2 g5f5 f2e2');
INSERT INTO evaluations VALUES(22942,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(22942,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(22942,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(22942,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(22942,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(22942,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(22942,36,'cp 8','c3d4 d7d5 e4d5 f8b4 b1c3 d8d5 g1f3 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8d7 e2d3 a8b8 a1c1 e7d5 e3d2 h8e8 c1c2 e8e6 g2g3 b4e7 a2a3 h7h5 h1c1 d5c3 d2c3 h5h4 c2e2 e6e2 d3e2 e7d6 g3h4 d6h2 h4h5 h2d6');
INSERT INTO evaluations VALUES(22920,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22920,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(22920,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22920,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22920,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22920,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22920,36,'cp 0','d7d5 e4d5 d8d5 g1f3 c8g4 f1e2 f8b4 b1c3 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 a1c1 e7f5 h1d1 h8e8 c3a4 d8d5 e2d3 e8d8 d3e4 d8e8');
INSERT INTO evaluations VALUES(21608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21608,36,'cp 0','e4d5 d8d5 g1f3 c8g4 f1e2 g4f3 e2f3 f8b4 b1c3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 a1c1 e7f5 h1d1 h8e8 c3a4 d8d5 e2d3 e8d8 d3c4 b4a5 g2g4 f5d6 c4b3 d6b5 c1c5 a5b6 c5d5 d8d5 b3c4 b5d6 c4d3 d6b5');
INSERT INTO evaluations VALUES(21561,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21561,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21561,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21561,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21561,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21561,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21561,36,'cp 0','f8b4 b1c3 d8d5 g1f3 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 a1c1 e7f5 h1d1 h8e8 c3a4 d8d5 e2d3 e8d8 d3e4 d8e8');
INSERT INTO evaluations VALUES(32411,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(32411,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(32411,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(32411,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(32411,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(32411,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(32411,36,'cp 145','d8f6 f3b3 b8d7 b3b7 a8b8 b7a6 f6a6 c4a6 d7c5 a6d3 c5d3 c2d3 g8e7 b1c3 e7c6 b2b3 b8d8 e1e2 c6d4 e2d1 c7c6 c1e3 f8a3 c3e2 d4e6 d1c2 e8g8 a1d1 c6c5 e2c3 d8d7 h1f1 e6d4 c2b1 a7a5 f2f4 e5f4 e3f4 f7f6 f4e3');
INSERT INTO evaluations VALUES(32938,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(32938,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(32938,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(32938,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(32938,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(32938,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(32938,36,'cp 144','f3b3 b8d7 b3b7 a8b8 b7a6 f6a6 c4a6 d7c5 a6d3 c5d3 c2d3 g8e7 b1d2 e7c6 d2c4 b8d8 c1e3 d8d3 e1e2 d3d7 h1d1 f7f6 a2a3 a7a6 b2b4 h7h5 d1d5 d7d5 e4d5 c6d8 a1b1 e8d7 b4b5 a6b5 b1b5 d8f7 b5b8 f8e7 b8h8 f7h8 e2d3 h8f7 a3a4 e7b4 d3e4 f7d6 c4d6 d7d6 f2f4 e5f4 e3f4 d6d7 f4g3');
INSERT INTO evaluations VALUES(32937,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(32937,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(32937,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(32937,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(32937,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(32937,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(32937,36,'cp 160','b8d7 b3b7 a8b8 b7a6 f6a6 c4a6 d7c5 a6d3 c5d3 c2d3 g8e7 b1d2 e7c6 e1e2 c6d4 e2d1 d4c6 d2c4 b8d8 c1e3 d8d3 d1e2 d3d7 h1d1 f7f6 a2a3 a7a6 b2b4 f8e7 d1d5 d7d5 e4d5 c6d8 a1c1 h7h5 e3c5 e7c5 b4c5 c7c6 d5d6 d8e6 c1b1 e8f7 b1b7 f7g6 b7b6 h8c8 b6a6 e6c5 a6a7 c8b8 a3a4');
INSERT INTO evaluations VALUES(44335,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(44335,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(44335,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(44335,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(44335,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(44335,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(44335,36,'cp 97','d7d6 e2e4 g7g6 f2f4 f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5c4 d7b6 c4e2 c8g4 c1e3 b6d7 h2h3 g4f3 f1f3 e8g8 e3f2 f8e8 d1d2 a8c8 e2c4 d7b6 c4f1 b6d7 a1e1 c5c4 e4e5 d6e5 c3e4 b4d3 f1d3 c4d3 e4d6 e5e4 d6e4 f7f5');
INSERT INTO evaluations VALUES(45930,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(45930,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(45930,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(45930,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(45930,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(45930,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(45930,36,'cp 120','e2e4 g7g6 f2f4 f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5e2 e8g8 c1e3 d7f6 f3d2 a8b8 a4a5 b7b5 a5b6 f8e8 e3f2 f6d7 h2h3 h7h6 d2f3 d7b6 f1e1 c8b7 d1d2 f7f5 e4f5 g6f5 f3h4 d8d7 a1d1 b4d5 c3d5 b7d5 b2b4 g8h7 e2h5 e8e1 d2e1 g7f6 b4c5 d6c5 h5g6 h7g7');
INSERT INTO evaluations VALUES(45926,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(45926,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(45926,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(45926,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(45926,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(45926,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(45926,36,'cp 101','g7g6 f2f4 f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 c1e3 e8g8 e1g1 a7a6 b5e2 f8e8 e3f2 f7f5 e4e5 d6e5 d5d6 e5f4 d1d2 d7e5 a1e1 c8d7 f3e5 e8e5 e2c4 g8h8 d2f4 d8f6 e1e5 f6e5 f4d2 b4c6 f1e1 e5f6 f2c5 a8c8 c4f1 h7h6 h2h3 f5f4 c3d5 f6b2 d2b2 g7b2 d5b6 b2c3');
INSERT INTO evaluations VALUES(45524,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(45524,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(45524,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(45524,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(45524,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(45524,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(45524,36,'cp 95','f2f4 f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5e2 e8g8 c1e3 f8e8 e3f2 f7f5 e4e5 d6e5 d5d6 e5f4 d1d2 d7e5 a4a5 e5f3 e2f3 g7d4 f2d4 c5d4 c3d5 b4d5 f3d5 c8e6 d5b7 a8b8 b7c6 d8d6 c6e8 b8e8 d2f4 e8d8 f4d6 d8d6 a1a3 d4d3 a3c3 d6d4 g1f2 e6c4');
INSERT INTO evaluations VALUES(45526,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(45526,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(45526,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(45526,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(45526,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(45526,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(45526,36,'cp 99','f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5c4 d7b6 c4e2 c8g4 c1e3 e8g8 d1d2 b6d7 g1h1 d8e7 a1e1 a8e8 e3f2 g4f3 e2f3 c5c4 f3e2 e8c8 f2g1 f8e8 h2h3 a6a5 c3b5 d7f6 e2f3 b4d3 e1e2 f6d7 e4e5 d6e5 d5d6 e7h4 b5c7 d3f4 c7e8 f4e2');
INSERT INTO evaluations VALUES(40453,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(40453,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(40453,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(40453,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(40453,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(40453,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(40453,36,'cp 95','f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 c1e3 e8g8 e1g1 a7a6 b5e2 f8e8 e3f2 f7f5 e4e5 d6e5 d5d6 e5f4 d1d2 d7e5 f2c5 d8a5 c3b5 a6b5 c5b4 e5f3 e2f3 a5b6 g1h1 b5a4 f1e1 c8d7 e1e7 e8e7 d6e7 d7e8 f3d5 g8h8 a1d1 a8c8 h2h3 g7f6 d2f4 h8g7');
INSERT INTO evaluations VALUES(40440,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(40440,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(40440,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(40440,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(40440,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(40440,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(40440,36,'cp 80','f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5c4 d7b6 c4e2 c8g4 c1e3 e8g8 d1d2 f8e8 h2h3 g4f3 f1f3 b6d7 e3f2 a8c8 a1d1 g7h6 g1h1 d7e5 f3g3 c5c4 d1f1 e5d3 f2e3 d8h4 h1h2 h4e7 a4a5 h6g7 h2h1 e8f8 f4f5 g7e5 g3f3 f7f6 c3d1 g6g5 d1f2 f8e8 f2d3 b4d3 e2d3');
INSERT INTO evaluations VALUES(42595,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(42595,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(42595,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(42595,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(42595,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(42595,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(42595,36,'cp 26','g1f3 b8c6 f1g2 e8g8 e1g1 d7d6 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 a4b6 a7b6 a2a4 c7c6 f3g2 c8g4 d1c2 d8d7 e3e4 b6b5 a4a5 b5c4 d3c4 h7h6 b2b4 g4h3 c1b2 d7e6 g2h3 e6h3 f2f3 h3e6 b4b5 c6b5 c4b5 f8c8 c2d3');
INSERT INTO evaluations VALUES(42594,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(42594,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(42594,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42594,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(42594,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(42594,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(42594,36,'cp 25','b8c6 f1g2 e8g8 e1g1 d7d6 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 a4b6 a7b6 f1e1 c7c6 e3e4 c8h3 d3d4 h7h6 a2a4 b6b5 c4b5 c6b5 d4e5 d6e5 a4a5 h3e6 d1d8 f8d8 c1e3 e6c4 f3g2 f6g4 e3c5 d8d2 h2h3 g4f6 b2b4 d2c2 e1c1');
INSERT INTO evaluations VALUES(48342,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(48342,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48342,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(48342,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48342,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(48342,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(48342,36,'cp 7','b2b3 b7b6 c1b2 f8e7 g1f3 c8b7 e1g1 e8g8 c2c4 c7c5 d4c5 d5c4 d3c4 e7c5 d1e2 d8e7 b1c3 b8d7 c4a6 b7f3 e2f3 c5a3 f3e2 a3b2 e2b2 f8d8 f1d1 d7c5 a6f1 d8d1 a1d1 a8d8 d1d8 e7d8 f1e2 c5e4 b3b4 e4c3 b2c3');
INSERT INTO evaluations VALUES(14366,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(14366,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(14366,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(14366,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(14366,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(14366,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(14366,36,'cp 91','g1f3 a7a6 a2a3 c8f5 e1g1 e7e6 b2b4 f8d6 b1d2 f6e4 c1b2 e8g8 f1e1 c6e7 d2e4 f5e4 f3d2 e4d5 e3e4 d5c4 d2c4 c7c6 d1g4 e7g6 d4d5 c6d5 e4d5 e6e5 a1d1 d8c7 c4e3 a6a5 b4a5 c7a5 e3f5 a5b6 b2c3 f8d8 h2h3 d6c5 g1h1 f7f6');
INSERT INTO evaluations VALUES(14367,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(14367,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(14367,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(14367,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(14367,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(14367,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(14367,36,'cp 101','a7a6 b1c3 b7b5 c4d3 c6b4 d3e2 e7e6 e3e4 c7c5 e1g1 c8b7 a2a4 c5d4 f3d4 f8d6 a4b5 e8g8 b5a6 b4a6 d4b5 d6e7 d1d8 f8d8 f2f3 e7c5 g1h1 a6b4 a1a8 b7a8 f1d1 d8d1 c3d1 b4c2 b5a3 c2a3 b2a3 a8c6 d1c3 f6d7 e2b5 c6b5 c3b5 d7b6 c1f4 h7h6 h2h3 c5e7');
INSERT INTO evaluations VALUES(11330,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(11330,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(11330,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(11330,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(11330,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(11330,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(11330,36,'cp 128','b1c3 b7b5 c4b3 e7e6 e3e4 c6a5 b3c2 c8b7 e1g1 c7c5 d4d5 e6d5 c3d5 f6d5 e4d5 f8e7 f1e1 e8g8 d1d3 g7g6 c1h6 e7f6 h6f8 d8f8 f3e5 a8d8 d3f3 f6e5 e1e5 a5c4 e5e2 b7d5 f3f6 f8d6 f6d6 d8d6 c2b3 c4b6 b3d5 d6d5 a1c1 c5c4 h2h3 g8g7 b2b3 d5d3 b3c4 b6c4');
INSERT INTO evaluations VALUES(11329,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(11329,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(11329,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(11329,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(11329,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(11329,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(11329,36,'cp 101','e7e6 a2a3 b7b5 c4d3 f8d6 d1c2 c8b7 c1d2 e8g8 c3e4 h7h6 e4f6 d8f6 d3h7 g8h8 h7e4 c6d8 e4b7 d8b7 c2c6 f8b8 e1g1 b7d8 c6e8 h8h7 d2b4 a6a5 b4d6 c7d6 e8d7 d6d5 f1c1 b8b7 d7e8 f6e7 e8e7 b7e7 c1c5');
INSERT INTO evaluations VALUES(11381,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(11381,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(11381,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(11381,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(11381,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(11381,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(11381,36,'cp 127','a2a3 b7b5 c4d3 f8d6 c1d2 h7h6 a1c1 c8b7 c3e4 e8g8 e1g1 e6e5 e4f6 d8f6 d3e4 c6a5 e4b7 a5b7 f3e5 d6e5 d4e5 f6e5 c1c6 f8d8 d1c2 b7c5 d2a5 c5e6 a5c7 e6c7 c6c7 e5d5 c7c5 d5d2 f1c1 d8d7 h2h3 a8d8 b2b4 d2c2 c5c2');
INSERT INTO evaluations VALUES(11383,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(11383,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(11383,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(11383,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(11383,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(11383,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(11383,36,'cp 89','b7b5 c4d3 f8d6 d1c2 c8b7 e3e4 e6e5 c3b5 e5d4 b5d4 c6d4 f3d4 e8g8 f2f3 c7c5 d4f5 d6e5 f5e3 b7c6 e1g1 d8c7 g2g3 a8d8 g1g2 c7d7 e3c4 e5g3 h2g3 d7d3 c2d3 d8d3 c4e5 d3d6 c1e3 f8c8 e5c6 d6c6 a1c1 h7h6 f1d1 c5c4 c1c3 f6e8 g3g4 f7f6 d1d5 e8c7 d5d7 c7e6 d7a7');
INSERT INTO evaluations VALUES(45355,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(45355,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(45355,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(45355,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(45355,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(45355,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(45355,36,'cp -14','d8b6 d1c2 b8c6 b1d2 f6h5 f4e5 c5d4 e5d4 c6d4 c3d4 c8d7 f1d3 a8c8 c2b3 b6b3 d2b3 e7e6 g1f3 h5f6 e1e2 f8d6 a1c1 e8e7 f3e5 d7a4 b3c5 d6c5 d4c5 a7a5 b2b3 a4c6 f2f3 f6d7 e5d7 c6d7 h1d1 h7h6 e2f2 h8d8 c1c2');
INSERT INTO evaluations VALUES(33570,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(33570,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(33570,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(33570,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(33570,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(33570,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(33570,36,'cp -10','d1c2 b8c6 b1d2 f6h5 f4e5 c5d4 e5d4 c6d4 c3d4 c8d7 f1d3 e7e6 g1f3 f8d6 a1c1 e8e7 c2b3 h8c8 e1e2 c8c1 h1c1 b6b3 d2b3 a8c8 c1c8 d7c8 b3c5 g7g6 c5a4 f7f6 a4c3 c8d7 g2g4 h5g7 g4g5 d7c6 d3b5 c6b5 c3b5 d6b8 b5c3 a7a6 a2a4 b8d6 a4a5');
INSERT INTO evaluations VALUES(33571,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(33571,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(33571,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(33571,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(33571,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(33571,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(33571,36,'cp -13','b8c6 b1d2 f6h5 f4e5 c5d4 e5d4 c6d4 c3d4 c8d7 f1d3 e7e6 e1e2 f8d6 g1f3 a8c8 c2b3 b6b3 a2b3 a7a6 g2g4 h5f6 g4g5 f6g8 f3e5 g8e7 h2h4 e7c6 e5d7 e8d7 h4h5 h7h6 g5g6 f7f6 d2f3 c6b4 f3h4 c8c7 f2f4 h8b8 f4f5 b4d3 e2d3');
INSERT INTO evaluations VALUES(39297,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(39297,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(39297,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(39297,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(39297,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(39297,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(39297,36,'cp -3','g8f6 b1c3 e8g8 c1g5 f8e8 d1d3 b8c6 d4c6 b7c6 e1c1 d8e7 d3f3 d7d6 d1d2 a7a5 g5f6 e7f6 f3f6 g7f6 f1d3 a8b8 h1f1 c5b4 a2a3 b4c3 b2c3 c6c5 d2d1 c8e6 c3c4 b8b6 c1d2 e8b8 d1a1 b6b2 d2c3 g8f8 g2g3 a5a4 f2f3 f8e7 f1e1');
INSERT INTO evaluations VALUES(42679,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42679,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42679,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42679,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42679,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42679,36,'cp 0','b1c3 e8g8 c1g5 f8e8 d1d3 b8c6 d4c6 b7c6 e1c1 d8e7 d3f3 d7d6 f1e2 a8b8 a2a3 e7e5 g5f4 e5e7');
INSERT INTO evaluations VALUES(42678,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42678,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42678,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42678,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42678,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42678,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42678,36,'cp 0','e8g8 c1g5 f8e8 d1d3 b8c6 d4c6 b7c6 e1c1 d8e7 d3f3 d7d6 f1e2 a8b8 a2a3 e7e5 g5f4 e5e7');
INSERT INTO evaluations VALUES(37254,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37254,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37254,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37254,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37254,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37254,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37254,36,'cp 0','c1g5 f8e8 d1d3 b8c6 d4c6 b7c6 e1c1 d8e7 d3f3 d7d6 f1e2 a8b8 a2a3 e7e5 g5f4 e5e7');
INSERT INTO evaluations VALUES(49447,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(49447,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(49447,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(49447,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(49447,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(49447,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(49447,36,'cp 41','e4e5 f6d5 c3d5 e6d5 g1f3 d7d6 d2d4 d6e5 f3e5 f8d6 f1d3 e8g8 e1g1 c7c5 d1h5 g7g6 h5h6 c5c4 c1g5 d6e7 g5e7 d8e7 a1e1 e7d6 d3e2 b8c6 e5f3 f7f6 h2h4 c6e7 h4h5 e7f5 h6d2 g8g7 f3h2 c8d7 b2b3 a8e8 b3c4 d5c4 e2c4 d6d4');
INSERT INTO evaluations VALUES(49351,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(49351,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(49351,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(49351,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(49351,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(49351,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(49351,36,'cp 33','f6d5 c3e4 f7f5 e4c3 d7d6 c3d5 e6d5 g1f3 f8e7 d2d4 e8g8 f1e2 d6e5 f3e5 e7d6 c1f4 b8c6 d1d2 c6e5 d4e5 d6e7 h2h4 c8e6 h4h5 c7c5 c2c3 d5d4 e1g1 h7h6 a2a4 d8c7 f1d1 f8d8 d2c2 d4c3 d1d8 a8d8 b2c3 g8h8 a4a5');
INSERT INTO evaluations VALUES(49327,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(49327,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(49327,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(49327,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(49327,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(49327,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49327,36,'cp 32','c3e4 f7f5 e4c3 d7d6 c3d5 e6d5 g1f3 f8e7 d2d4 e8g8 f1e2 d6e5 f3e5 b8c6 c1f4 e7d6 e5c6 b7c6 f4d6 c7d6 e1g1 d8h4 f1e1 a8b8 b2b3 c8d7 c2c4 d5c4 d1d2 c4b3 a2b3 b8a8 e2f3 a7a5 e1c1 h4h6 d2h6 g7h6 f3c6 d7c6 c1c6');
INSERT INTO evaluations VALUES(21474,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(21474,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21474,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(21474,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(21474,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(21474,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(21474,36,'cp 83','d2d4 f7f6 f1b5 c8d7 b1c3 a7a6 b5c6 d7c6 c3e2 f6e5 f3e5 d8f6 e2g3 f8d6 g3h5 f6e7 c1f4 g8f6 e1g1 e8g8 h5f6 e7f6 f4g3 c6a4 d1d2 a8c8 f1e1 c7c5 b2b3 a4b5 a2a4 b5e8 e5d7 e8d7 g3d6 f6d4 d2d4 c5d4 d6f8 g8f8 e1e2 c8c3');
INSERT INTO evaluations VALUES(21414,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(21414,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(21414,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(21414,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(21414,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(21414,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(21414,36,'cp 70','e2e4 g8f6 e4e5 f6e4 c3e2 f7f6 e2g3 f6e5 d4e5 c8d7 f1d3 e4c5 d3e2 f8e7 e1g1 e8g8 c2c4 d7e8 c1e3 f8f3 e2f3 d5c4 f3g4 d8d5 f2f4 e8g6 g1h1 a8d8 d1e1 g6d3 g3e2 d5d7 a1d1 d7e8 e1g3 h7h5 g4f3 c6b4 g3h3 g7g6 h3g3');
INSERT INTO evaluations VALUES(21434,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(21434,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(21434,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(21434,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(21434,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(21434,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(21434,36,'cp 51','g8f6 e4e5 f6e4 c3e2 f7f6 e2g3 f6e5 d4e5 f8e7 c2c3 e4c5 h2h4 e8g8 c1e3 c8d7 f1e2 a7a5 h4h5 h7h6 e1g1 d8e8 d1c2 e8f7 a1d1 a5a4 f1e1 d7e8 a2a3 c5d7 c3c4 d7e5 c4d5 e6d5 f3e5 c6e5 c2c7 e7f6 c7f7 e8f7 e3d4 a8d8 d4c3 d5d4 c3d4 d8d4 d1d4 e5f3 e2f3 f6d4');
INSERT INTO evaluations VALUES(14869,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(14869,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(14869,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(14869,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(14869,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(14869,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(14869,36,'cp 59','e4e5 f6e4 c3e2 f7f6 e2g3 f6e5 d4e5 f8e7 c2c3 e4c5 c1e3 e8g8 h2h4 c8d7 h4h5 h7h6 f1e2 a7a5 e1g1 d8e8 d1c2 b7b6 a1e1 a8d8 e1d1 g8h8 b2b3 d7c8 f1e1 c8a6 e2a6 c5a6 c2e2 a6c5 c3c4 c5e4 g3e4 d5e4 d1d8 e7d8');
INSERT INTO evaluations VALUES(14886,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(14886,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14886,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(14886,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(14886,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(14886,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(14886,36,'cp 44','f6e4 c3e2 f7f6 e2g3 f6e5 d4e5 f8e7 c2c3 e4c5 c1e3 e8g8 h2h4 c8d7 f1e2 b7b6 h4h5 h7h6 e1g1 f8f3 e2f3 c6e5 f3e2 a7a5 f2f4 e7h4 e3f2 e5f7 c3c4 c7c6 a1c1 f7d6 b2b3 d8e7 e2g4 h4g3 f2g3 d6e4 g3f2 e4f2 f1f2 c5e4 f2f1 a8f8 g4h3');
INSERT INTO evaluations VALUES(14721,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14721,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(14721,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(14721,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(14721,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(14721,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14721,36,'cp 75','c3e2 f7f6 e2g3 f6e5 d4e5 f8e7 c2c3 e4c5 h2h4 e8g8 c1e3 c8d7 f1e2 b7b6 h4h5 h7h6 e1g1 f8f3 e2f3 c6e5 f3e2 a7a5 f2f4 e7h4 e3f2 e5f7 c3c4 c7c6 e2d3 f7d6 b2b3 h4g3 f2g3 d6e4 g3f2 e4f2 f1f2 d8f6 a1c1 a8f8 d3b1 f6h4 d1e2 d7e8 e2e5 h4h5 e5c7 e8g6 c4d5 g6b1 c1b1');
INSERT INTO evaluations VALUES(20846,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(20846,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(20846,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(20846,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(20846,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(20846,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(20846,36,'cp 78','c2c4 e7e5 c4d5 e5f4 d5c6 b7c6 d1c2 f8d6 e2e4 f4e3 f2e3 g8h6 c2c6 c8d7 c6e4 e8f8 b1c3 a8b8 f1c4 g7g6 c4b3 h6g4 e4d3 f6f5 e1g1 f8g7 h2h3 g4f6 f3e5 d8e7 e5d7 e7d7 e3e4 f5e4 c3e4 f6e4 d3e4 h8e8');
INSERT INTO evaluations VALUES(42533,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(42533,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(42533,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(42533,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(42533,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(42533,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(42533,36,'cp 219','d2d4 c5b6 b1c3 b8c6 c1e3 e8g8 e5c6 b7c6 e4e5 f6d5 c3d5 c6d5 d1f3 c7c6 f1d3 d7d6 e5d6 d8d6 e1c1 a7a5 h2h4 a5a4 c2c3 b6c7 h4h5 c8a6 d3c2 f7f5 a2a3 a6c4 g2g3 h7h6 e3f4 d6d7 f4c7 d7c7 d1e1 c4b3 c2b3 a4b3 e1e5 f5f4');
INSERT INTO evaluations VALUES(42529,30,'cp 239',NULL);
INSERT INTO evaluations VALUES(42529,31,'cp 228',NULL);
INSERT INTO evaluations VALUES(42529,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(42529,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(42529,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(42529,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(42529,36,'cp 235','c5b6 b1c3 e8g8 c1g5 d7d6 e5c4 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 f7f5 e4f5 c8f5 d5e3 f5e6 f1e2 b8d7 e1g1 g8h8 e2d3 e6g8 a2a4 a7a5 f2f4 d8f6 d1f3 c7c6 a1e1 a8e8 c4b6 d7b6');
INSERT INTO evaluations VALUES(42368,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(42368,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(42368,32,'cp 215',NULL);
INSERT INTO evaluations VALUES(42368,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(42368,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(42368,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(42368,36,'cp 248','b1c3 b8c6 f1c4 e8g8 e5f7 f8f7 c4f7 g8f7 e4e5 d8e7 e1g1 f6e8 c3d5 e7f8 d5b6 a7b6 f2f4 d7d6 e5e6 f7g8 f4f5 h7h6 c2c3 e8f6 g2g4 c6e7 d1e2 f6h7 h2h4 f8f6 e2h2 g7g6 f1f2 g6f5 g4f5 e7f5 d4d5 g8h8 h2f4 c8e6 d5e6 a8f8 c1e3 f6e6 e3d4 f5d4 f4d4 h7f6');
INSERT INTO evaluations VALUES(42367,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(42367,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(42367,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(42367,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(42367,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(42367,35,'cp 230',NULL);
INSERT INTO evaluations VALUES(42367,36,'cp 204','b8c6 c1e3 e8g8 e5c6 b7c6 e4e5 f6d5 c3d5 c6d5 d1f3 c7c6 f1d3 d7d6 e5d6 d8d6 e1c1 a7a5 h2h4 a5a4 a2a3 c8a6 d3a6 a8a6 h4h5 h7h6 g2g4 b6d8 e3f4 d6f6 f3e3 f6f4 e3f4 d8g5 f4g5 h6g5 h1e1 a6b6 e1e2 f7f5 g4f5 f8f5 d1g1 b6b7 e2e8 g8h7 e8c8');
INSERT INTO evaluations VALUES(1862,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(1862,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(1862,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(1862,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(1862,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(1862,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(1862,36,'cp 115','b1c3 d7d6 c3d5 f6d8 e1g1 g8f6 f1e1 c8g4 c2c3 f8e7 h2h3 g4h5 b2b4 e8g8 a2a4 f6d5 c4d5 a7a6 a4a5 g8h8 d1e2 h7h6 c1b2 h5f3 e2f3 e7g5 b4b5 a6b5 a5a6 a8a6 a1a6 b7a6 d5c6 g5d2');
INSERT INTO evaluations VALUES(1857,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(1857,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(1857,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(1857,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(1857,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(1857,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(1857,36,'cp 140','d7d6 d2d4 c6d4 f3d4 e5d4 c3b5 f6e7 e1g1 c8e6 d1d4 a7a6 c1g5 e7d7 b5c3 h7h6 g5h4 e6c4 d4c4 f8e7 c3d5 a8c8 c4c3 e8f8 h4g3 h6h5 h2h4 h8h6 f1e1 h6e6 a1d1 g8f6 f2f3 f6d5 d1d5 e6h6 c3d3 f8g8 c2c4 e7f6 b2b4 c7c6 d5f5 d7e6 g3f4');
INSERT INTO evaluations VALUES(1743,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(1743,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(1743,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(1743,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(1743,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(1743,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(1743,36,'cp 150','d2d4 c6d4 f3d4 e5d4 c3b5 f6e7 e1g1 c8e6 d1d4 e7d7 c4e2 g8e7 f2f4 e7c6 d4e3 f7f6 c2c4 g7g6 b2b3 f8g7 c1b2 e8g8 f4f5 e6f7 b5c3 a8e8 a1d1 d7e7 c3d5 f7d5 c4d5 c6e5 e3g3 a7a6 g1h1 e5d7 f5g6 h7g6 g3g6 e7e4 g6e4 e8e4 e2d3 e4e7 d3f5 f8e8 b2d4 d7c5 f5g6 e8f8 b3b4 c5e4');
INSERT INTO evaluations VALUES(55863,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(55863,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55863,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(55863,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55863,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(55863,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(55863,36,'cp 11','c7c5 f1g2 c5d4 e1g1 g8f6 f3d4 e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 h7h6 d2e4 d8b6 e4d6 f8d6 g2d5 e6d5 d1d5 a8d8 c1e3 b6b4 a1c1 d6e7 d5c4 b4c4 c1c4 f7f6 f1c1 e8f7 g1g2 d8d5 e3c5 e7c5 c4c5 h8d8 c1c4 f6f5 a2a3 f7f6 c5d5 d8d5');
INSERT INTO evaluations VALUES(54218,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(54218,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(54218,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54218,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54218,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54218,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54218,36,'cp 16','f1g2 c5d4 e1g1 g8f6 f3d4 e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 h7h6 d2e4 d8b6 e4d6 f8d6 g2d5 e6d5 d1d5 a8d8 a2a4 a7a6 d5c4 b6b4 c4b4 d6b4 c1e3 e8e7 a1c1 d8d5 c1c4 f7f5 e3b6 e7f7 g1g2 b4e7 e2e3 h8c8 f1c1 g7g5 b3c5 e7c5 c4c5 d5c5');
INSERT INTO evaluations VALUES(54219,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(54219,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54219,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54219,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(54219,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(54219,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54219,36,'cp 14','c5d4 e1g1 g8f6 f3d4 e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 h7h6 d2e4 d8b6 e4d6 f8d6 g2d5 e6d5 d1d5 a8d8 c1e3 b6b4 a1c1 d6e7 d5c4 b4c4 c1c4 f7f6 f1c1 d8d5 g1g2 e8f7 h2h4 f6f5 b3c5 e7c5 c4c5 h8d8 c5d5 d8d5 a2a3');
INSERT INTO evaluations VALUES(54392,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(54392,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54392,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(54392,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(54392,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(54392,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54392,36,'cp 11','e1g1 g8f6 f3d4 e6e5 d4f3 b8c6 c2c4 d5d4 e2e3 f8e7 e3d4 e5d4 b2b4 e7b4 f3d4 e8g8 d4c6 b7c6 d1a4 d8e7 c1g5 c8e6 g5f6 e7f6 a4b4 a8b8 b4c3 b8b1 c3f6 b1f1 a1f1 g7f6 c4c5 e6a2 f1a1 a2e6 a1a7 f8d8 g1f1 d8d1 f1e2 d1c1 a7c7 c1c5 c7c6 c5c6');
INSERT INTO evaluations VALUES(21905,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(21905,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21905,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(21905,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(21905,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21905,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21905,36,'cp 28','c8e6 c4e6 f7e6 c2c3 g8f6 d2d3 a7a5 e1g1 f8e7 a2a4 e8g8 f1e1 d6d5 e4d5 d8d5 d1e2 f8d8 b1d2 d5d3 f3e5 d3e2 e1e2 c6e5 e2e5 g8f7 g1f1 d8d5 d2c4 f6d7 e5e4 d5d1 e4e1 d1e1 f1e1 e6e5 c1e3 f7e6 e1d2 e7c5 f2f3 c5e3 c4e3 d7c5 d2c2 e5e4 f3e4 c5e4 e3c4');
INSERT INTO evaluations VALUES(27064,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(27064,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(27064,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(27064,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(27064,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(27064,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(27064,36,'cp 31','c4e6 f7e6 c2c3 g8f6 d2d3 a7a5 d1e2 f8e7 d3d4 e8g8 e1g1 e5d4 c3d4 d6d5 e4d5 e6d5 b1c3 d8d7 c1g5 h7h6 g5h4 a8e8 h4f6 e7f6 e2b5 e8d8 a1d1 b7b6 f1e1 c6e7 b5b3 d7d6 b3c2 c7c5 b2b3 d8c8 c3b5 d6d7 a2a4 d7f5 c2f5 e7f5');
INSERT INTO evaluations VALUES(27048,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(27048,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(27048,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(27048,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(27048,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(27048,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(27048,36,'cp 33','f7e6 c2c3 g8f6 d2d3 a7a5 e1g1 f8e7 d1e2 d8d7 d3d4 e5d4 c3d4 d6d5 e4d5 e6d5 b1c3 e8g8 c1f4 e7d6 f4d6 d7d6 a1e1 a8e8 e2d2 h7h6 c3b5 d6d7 e1e8 f8e8 a2a3 f6h5 f1e1 e8e1 f3e1 h5f6 e1f3 f6h5 a3a4 h5f6 f3e5 d7e7 e5g6');
INSERT INTO evaluations VALUES(27131,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(27131,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(27131,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(27131,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27131,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(27131,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(27131,36,'cp 29','d2d3 g8f6 c2c3 a7a5 e1g1 f8e7 d1e2 d8d7 d3d4 a5a4 f1e1 e8g8 d4e5 c6e5 f3e5 d6e5 b1d2 h7h6 d2f3 e7d6 c3c4 d7c6 c1d2 f6d7 d2c3 c6b6 a1b1 b6c6 b1d1 f8f7 d1d2 f7f4 b2b4 a4b3 a2b3 b7b5');
INSERT INTO evaluations VALUES(16358,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(16358,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(16358,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(16358,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(16358,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(16358,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(16358,36,'cp 132','f6e4 d1e2 e4c5 b1c3 c5e6 h2h4 d7d6 e5d6 f8d6 c1e3 d6b4 e2c4 d8f6 f1d3 b4c3 b2c3 c6a5 c4a4 a5c6 e1g1 e8g8 a4e4 g7g6 h4h5 c8d7 a1b1 b7b6 d3b5 a8e8 b1d1 e6c5 e4f4 f6f4 e3f4 a7a6 b5a6 c5a6 d1d7 a6c5 d7c7 c5e6 c7c6 e6f4 f3d4 f4h5 c3c4 h5f4 c6b6');
INSERT INTO evaluations VALUES(16810,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(16810,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(16810,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(16810,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(16810,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(16810,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(16810,36,'cp 122','d1e2 e4c5 b1c3 c5e6 h2h4 h7h6 c1e3 f8b4 e2c4 a7a6 e1c1 b7b5 c4d3 b4c3 d3c3 c8b7 a2a3 c6e7 f3d4 e7d5 c3d2 d5e3 d2e3 e6d4 e3d4 d8e7 f1d3 e8c8 h1e1 b7g2 a3a4 g2c6 d4a7 e7b4 a7a6 c8b8 a4b5 c6f3 d3e2 f3b7 a6a2 b4h4 a2f7 h8f8');
INSERT INTO evaluations VALUES(16809,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(16809,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(16809,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(16809,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(16809,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(16809,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(16809,36,'cp 101','e4c5 b1c3 c5e6 h2h4 h7h6 c1e3 f8b4 e2c4 a7a6 e1c1 b7b5 c4d3 b4c3 d3c3 c8b7 a2a3 c6e7 f3d4 e7d5 c3d2 d5e3 d2e3 e8g8 f1d3 e6d4 e3d4 f7f6 f2f3 b7c6 d4f4 f6e5 f4e5 a8c8 h4h5 d8f6 e5f6 f8f6 d3g6 g8f8 d1e1 c8d8 h1h4 d7d5 h4g4 c6d7');
INSERT INTO evaluations VALUES(30748,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(30748,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(30748,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(30748,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(30748,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(30748,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(30748,36,'cp 162','f8g7 h2h4 h7h6 c1f4 a7a6 d1e2 d7d6 e1c1 e7e6 d4d5 e6e5 f4e3 b8d7 f3d2 h6h5 f2f3 g7h6 e3h6 g8h6 g2g4 h5g4 f3g4 f7f6 d2c4 d8e7 c4e3 e8c8 a2a4 c8b8 b2b4 f6f5 g4g5 f5f4 g5h6 f4e3 e2e3 e7f8');
INSERT INTO evaluations VALUES(22675,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(22675,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(22675,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(22675,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(22675,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(22675,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(22675,36,'cp 61','c6e5 g1f3 d7d6 b1c3 g8f6 c1f4 e5g6 f4g3 f8e7 d1d2 e8g8 e1c1 a7a6 h2h4 f6h5 g3h2 b7b5 c3d5 c8g4 f1e2 h5f6 f3d4 f6d5 e2g4 d5b6 h4h5 g6e5 h2e5 d6e5 d4c6 e7g5 c6d8 g5d2 d1d2 f8d8 h1d1 d8d2 d1d2 a8f8 f2f3 b6c4');
INSERT INTO evaluations VALUES(23082,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(23082,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(23082,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(23082,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(23082,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(23082,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(23082,36,'cp 70','b1c3 f8c5 c1f4 d7d6 d1d2 g8f6 e1c1 c8e6 g1f3 e5g6 f4e3 c5b4 d2d4 b4c5 d4d3 e8g8 e3c5 d6c5 h2h3 d8e7 d3e3 a8d8 g2g3 d8d1 c3d1 f8d8 d1c3 a7a6 a2a4 h7h6 f3d2 g6e5 f2f4 e5c6 g3g4 c6d4 h1g1 e6d7 g4g5 h6g5 f4g5 f6e8 c3d5 e7d6 g5g6 f7g6 d2f3');
INSERT INTO evaluations VALUES(23079,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(23079,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(23079,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(23079,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(23079,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(23079,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(23079,36,'cp 65','f8c5 c1f4 d7d6 d1d2 g8f6 e1c1 c8e6 g1f3 e5g6 f4e3 c5b4 d2d4 b4c5 d4d3 e8g8 e3c5 d6c5 d3e3 d8e7 h2h3 c7c6 a2a4 f8e8 g2g3 b7b6 f1g2 a8d8 f3d2 a7a6 h1e1 e7d7 d2f3 d7e7 d1d8 e8d8 f3d2 e7d7 a4a5 b6a5 c3a4 e6h3 g2h3 d7h3 a4c5');
INSERT INTO evaluations VALUES(29019,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(29019,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(29019,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(29019,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(29019,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(29019,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(29019,36,'cp 100','f3e5 g4c8 c4d5 c6d5 c1f4 b8c6 e2e3 c8d7 f1d3 e7e6 a1c1 f8e7 e1g1 a8c8 d1f3 h7h6 f1d1 e8g8 d3b1 e7b4 e5d7 f6d7 a2a3 b4e7 e3e4 d5e4 c3e4 f7f5 e4c3 e7g5 d4d5 g5f4 d5c6 f4c1 c6d7 c8c7 d1c1 c7d7 f3g3 d8b6');
INSERT INTO evaluations VALUES(29058,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(29058,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(29058,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(29058,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(29058,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(29058,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(29058,36,'cp 101','g4c8 c4d5 c6d5 c1f4 e7e6 e2e3 b8c6 f1d3 c8d7 a1c1 a8c8 e1g1 f8e7 d1f3 h7h6 f1d1 e8g8 d3b1 a7a6 a2a3 b7b5 e3e4 d5e4 c3e4 f6e4 f3e4 f7f5 e4e1 c6e5 f4e5 e7g5 c1c5 d7c6 h2h4 g5h4 d1c1 c6d7 b1a2 h4g5 c5c8 d7c8 c1c6 g5f6 a2e6 c8e6 c6e6 d8d7 e6a6');
INSERT INTO evaluations VALUES(53606,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(53606,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(53606,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(53606,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(53606,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(53606,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(53606,36,'cp 56','f8b4 e4e5 c6c5 a2a3 b4c3 b2c3 h7h6 d4c5 b8d7 d1d4 g8e7 d4g4 d8a5 c1d2 e8g8 f1d3 f7f5 e5f6 d7f6 g4h4 a5c5 e1g1 e7f5 h4b4 f6e4 d3e4 c5b4 a3b4 d5e4 f3e5 b7b6 d2e3 c8b7 c3c4 a7a5 c4c5 a5b4 c5b6 f5e3 f2e3 f8f1 a1f1');
INSERT INTO evaluations VALUES(38038,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(38038,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(38038,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(38038,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(38038,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(38038,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(38038,36,'cp 63','e4e5 b7b6 c1d2 b4c3 d2c3 c8a6 f1a6 b8a6 d1e2 a6b8 c3b4 g8e7 b4a3 h7h6 c2c3 b8d7 e1g1 c6c5 f3h4 a7a5 e2g4 g7g6 h4f3 e7f5 f1e1 d8c7 a1c1 e8c8 d4c5 d7c5 a3c5 c7c5 g4a4 c8b7 c3c4 d5d4');
INSERT INTO evaluations VALUES(38056,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(38056,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(38056,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(38056,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(38056,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(38056,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(38056,36,'cp 66','b7b6 c1d2 b4c3 d2c3 c8a6 f1a6 b8a6 d1e2 a6b8 c3b4 g8e7 b4d6 e7f5 d6a3 h7h5 c2c3 a7a5 b2b3 b8a6 a3c1 a5a4 e1g1 a6c7 c1e3 c7b5 a1c1 d8d7 h2h3 a4b3 a2b3 a8a3 c1a1 e8g8 a1a3 b5a3 f1a1 a3b5 e2d3 f8c8 c3c4 d5c4 b3c4');
INSERT INTO evaluations VALUES(16795,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(16795,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(16795,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(16795,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(16795,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(16795,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(16795,36,'cp 51','f8b4 b1c3 b4c3 b2c3 d8e7 d1d5 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 h7h6 e1g1 c6e5 f3e5 d6e5 f4e3 c7c6 c4c5 f6g4 h2h3 c8f5 d3d2 a8d8 d2c1 g4e3 c1e3 e7g5 e3e5 f8e8 e5c7 d8d7 c7a5 e8e2 a1d1 d7d1 f1d1 g5e7 a5d8 e7d8 d1d8 g8f7 d8b8 e2b2 g3g4 f5d3 a2a4 f7f6 b8f8 f6e7 f8a8 a7a6 a8b8 a6a5');
INSERT INTO evaluations VALUES(10670,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(10670,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(10670,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(10670,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(10670,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(10670,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(10670,36,'cp 56','b1d2 d8e7 e2e3 g4e5 f1e2 e5f3 e2f3 d7d6 a2a3 b4d2 d1d2 e8g8 e1g1 c6e5 f3e2 a7a5 b2b3 e5d7 a1d1 d7c5 b3b4 a5b4 a3b4 c5e4 d2d4 e4g5 d4b2 g5e6 f4g3 e6g5 e2d3 g5e4 g3f4 f8e8 d1e1 e4f6 e3e4 f6d7 f2f3 e7f6 b2f6 d7f6 f4d2 f6d7');
INSERT INTO evaluations VALUES(10669,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(10669,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(10669,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(10669,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(10669,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(10669,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(10669,36,'cp 49','d8e7 e2e3 g4e5 f3e5 c6e5 f1e2 d7d6 a2a3 b4d2 d1d2 e8g8 e1g1 a7a5 b2b3 e5d7 e2d1 b7b6 e3e4 d7c5 f1e1 a5a4 b3b4 c5b3 d1b3 a4b3 d2c3 c8e6 c3b3 b6b5 b3c3 e6c4 a3a4 h7h6 f2f3 c7c5 a4a5 a8a6 g1h1 e7e6 h2h3 c5b4 c3b4 f8a8 f4e3');
INSERT INTO evaluations VALUES(597,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(597,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(597,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(597,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(597,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(597,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(597,36,'cp 57','e2e3 g4e5 f3e5 c6e5 f1e2 e8g8 e1g1 a7a5 a2a3 b4d2 d1d2 d7d6 b2b3 e5d7 a1d1 d7c5 b3b4 c5e4 d2d4 f8e8 f2f3 e4g5 d4c3 g5e6 f4g3 a5b4 a3b4 h7h5 h2h4 e6f8 e3e4 c8e6 f1e1 f8g6 c4c5 g6h4 c5d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(596,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(596,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(596,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(596,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(596,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(596,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(596,36,'cp 51','g4e5 f3e5 c6e5 f1e2 e8g8 e1g1 a7a5 a2a3 b4d2 d1d2 d7d6 b2b3 e5d7 a1d1 d7c5 b3b4 c5e4 d2d4 f8e8 f2f3 e4g5 d4c3 g5e6 f4g3 b7b6 e2d3 c8b7 h2h4 a5b4 a3b4 a8a7 d3b1 e7f6 c3f6 g7f6 h4h5 e8a8 h5h6 a7a4 g3e1 a4a1 e3e4');
INSERT INTO evaluations VALUES(604,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(604,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(604,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(604,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(604,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(604,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(604,36,'cp 59','f3e5 c6e5 a2a3 b4d2 d1d2 d7d6 f1e2 e8g8 e1g1 a7a5 b2b3 f8e8 a1c1 e5d7 e2d1 d7c5 d1c2 c8e6 c1d1 f7f5 f2f3 e7f7 d2c3 c5d7 a3a4 b7b6 f4g3 d7c5 e3e4 f5f4 g3f2 h7h6 e4e5 d6e5 f2c5 b6c5 c3e5');
INSERT INTO evaluations VALUES(591,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(591,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(591,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(591,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(591,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(591,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(591,36,'cp 45','c6e5 f1e2 e8g8 e1g1 a7a5 a2a3 b4d2 d1d2 d7d6 b2b3 e5d7 a1d1 h7h6 f1e1 d7c5 b3b4 a5b4 a3b4 c5e4 d2d4 g7g5 e2d3 f8e8 f2f3 e4f6 f4g3 f6h5 g3f2 e7f6 d4f6 h5f6 e3e4 c8e6 e4e5 d6e5 e1e5 e8d8 e5e1 c7c6');
INSERT INTO evaluations VALUES(611,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(611,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(611,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(611,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(611,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(611,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(611,36,'cp 55','a2a3 b4d2 d1d2 d7d6 f1e2 e8g8 e1g1 a7a5 b2b3 f8e8 a1d1 e5d7 e2d3 d7c5 d3c2 f7f5 h2h3 b7b6 f1e1 e7f7 f2f3 c8b7 f4g3 b7c6 g3h2 h7h6 g1h1 c5d7 c2b1 f7f6 d2c2 d7c5 b3b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(613,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(613,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(613,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(613,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(613,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(613,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(613,36,'cp 46','b4d2 d1d2 d7d6 f1e2 e8g8 e1g1 a7a5 b2b3 c8e6 a1c1 e5d7 f1e1 d7c5 d2b2 e6f5 c1d1 f8e8 e2f1 e7f6 b2f6 g7f6 b3b4 a5b4 a3b4 c5d7 f2f3 f5e6 e3e4 a8a2 c4c5 d6c5 b4c5 e8a8 d1b1 d7c5 f4c7 c5b3 b1d1 h7h6 c7d6');
INSERT INTO evaluations VALUES(616,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(616,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(616,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(616,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(616,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(616,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(616,36,'cp 58','d1d2 d7d6 f1e2 e8g8 e1g1 a7a5 b2b3 c8e6 a1c1 e5d7 f1d1 h7h6 d2c2 e7f6 f4g3 a8e8 e3e4 f6g6 b3b4 a5b4 a3b4 b7b6 d1e1 f7f5 e4f5 e6f5 c2d2 g6g5 g3f4 g5f6 e2f1 e8e1 c1e1 f5e6 f4g3 d7e5 e1e4 f8e8 f2f3 e6f7 e4e3');
INSERT INTO evaluations VALUES(615,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(615,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(615,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(615,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(615,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(615,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(615,36,'cp 58','d7d6 f1e2 e8g8 e1g1 a7a5 b2b3 c8e6 a1c1 e5d7 e3e4 d7c5 d2e3 e6d7 f2f3 h7h6 c1d1 f8e8 f4g3 e7g5 e3c3 f7f5 b3b4 a5b4 a3b4 a8a2 f1e1 c5a4 c3d4 f5e4 f3e4 d7g4 e2g4 g5g4 c4c5 d6c5 d4d5 g4e6');
INSERT INTO evaluations VALUES(495,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(495,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(495,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(495,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(495,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(495,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(495,36,'cp 49','f1e2 e8g8 e1g1 a7a5 b2b3 c8e6 a1c1 e5d7 d2c2 e7f6 f4g3 f6g6 c2c3 h7h6 f1e1 f7f5 e2f3 d7c5 h2h3 c5e4 f3e4 f5e4 c3c2 b7b6 c1d1 e6d7 d1d5 g8h8 d5d4 a8e8 b3b4 a5b4 a3b4 d7c8 e1d1 g6f7 b4b5');
INSERT INTO evaluations VALUES(496,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(496,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(496,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(496,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(496,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(496,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(496,36,'cp 48','e8g8 e1g1 a7a5 b2b3 c8e6 a1c1 e5d7 h2h3 d7c5 d2b2 h7h6 b3b4 c5d7 e2f3 a8a7 g1h2 a5b4 a3b4 e7f6 b2f6 d7f6 c4c5 d6c5 b4c5 c7c6 f4e5 f8a8 c1b1 e6d5 e5f6 d5f3 g2f3 g7f6 h2g3 g8f8 f1d1 f8e7 b1b6 a8b8');
INSERT INTO evaluations VALUES(2207,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(2207,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(2207,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(2207,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(2207,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(2207,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(2207,36,'cp 58','e1g1 a7a5 b2b3 c8e6 a1c1 e5d7 h2h3 f8e8 f1e1 d7c5 d2c3 e6f5 b3b4 c5d7 e2f1 a5b4 a3b4 b7b6 e3e4 f5e6 f4d2 a8a2 f2f4 f7f6 c1a1 e8a8 a1a2 a8a2 c3d3 e6f7 d2c3 f7e6 d3b1 a2a8 c3b2 e6f7 b1c2 h7h6');
INSERT INTO evaluations VALUES(2208,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(2208,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(2208,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(2208,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(2208,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(2208,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(2208,36,'cp 55','f8e8 b2b3 a7a5 f1e1 c8e6 a1c1 e5d7 e2d3 d7c5 d3c2 f7f5 b3b4 a5b4 a3b4 c5e4 c2e4 f5e4 c4c5 d6c5 c1c5 c7c6 h2h3 e8d8 d2c2 e6d5 b4b5 a8a2 c2c3 e7f6 c3f6 g7f6 e1d1 d8a8 b5c6 b7c6');
INSERT INTO evaluations VALUES(52126,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52126,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52126,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(52126,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52126,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(52126,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52126,36,'cp 24','g1f3 b8c6 f1c4 g8f6 d2d3 f8e7 e1g1 e8g8 f1e1 d7d6 a2a4 h7h6 b1a3 c8e6 c4e6 f7e6 c1d2 f6d7 a3c4 a8c8 c4e3 d8e8 d1b3 c6d8 a4a5 e8h5 e3f5 h5f7 f5e7 f7e7 d2e3 d7f6 h2h3 d8c6 b3a4 f6d7 a1d1 f8f3 g2f3 e7f6 g1g2');
INSERT INTO evaluations VALUES(52125,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52125,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52125,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(52125,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52125,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52125,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52125,36,'cp 16','b8c6 f1c4 g8f6 f3g5 d7d5 e4d5 f6d5 d2d4 c5d4 g5f7 e8f7 d1f3 f7e6 a2a4 f8e7 f3e4 g7g6 e1g1 e6f7 c4d5 f7g7 d5c6 b7c6 c3d4 e5d4 e4c6 c8d7 c6b7 h8f8 b1d2 d8c8 b7e4 e7f6 d2f3 d7c6 e4d3 c8f5 d3f5 g6f5 c1f4 c6f3 g2f3 f8d8 f1d1 a8c8 a1c1 c8c1 f4c1 g7f7 b2b3 d8c8 c1f4 c8c2');
INSERT INTO evaluations VALUES(52640,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(52640,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(52640,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(52640,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(52640,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(52640,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(52640,36,'cp 85','c1e3 b8c6 b1c3 g8f6 f2f3 h7h5 f1c4 c8d7 d1d2 c6d4 e3d4 f8h6 d2f2 e8g8 e1g1 a7a6 c4b3 a8c8 d4f6 e7f6 a1d1 d7e6 g1h1 g8h7 f1e1 c8c6 c3d5 e6d5 b3d5 c6c7 c2c4 b7b5 b2b3 d8c8 f3f4 b5c4 d5c4 f6f5 f2d2');
INSERT INTO evaluations VALUES(52639,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(52639,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(52639,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(52639,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(52639,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(52639,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(52639,36,'cp 73','g8f6 f2f3 b8c6 b1c3 h7h5 f1c4 c8d7 d1d2 c6d4 e3d4 f8h6 d2f2 d8a5 c4b3 d7c6 a1d1 h5h4 e1g1 e7e5 d4e3 h6e3 f2e3 a5c5 e3c5 d6c5 c3d5 c6d5 e4d5 a8d8 f1e1 h8h5 d5d6 e8f8 a2a4 e5e4 b3c4 h5e5 a4a5 e4e3 b2b3 f8e8 c2c3');
INSERT INTO evaluations VALUES(47625,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(47625,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(47625,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(47625,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(47625,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(47625,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(47625,36,'cp 65','d2d4 b8d7 f1e1 f8e7 b1c3 e8g8 a2a4 f8e8 h2h3 c7c6 a4a5 e5d4 f3d4 e7f8 c1f4 a7a6 f4g3 b7b5 a5b6 d8b6 d4b3 d7e5 c4f1 a6a5 c3a4 b6b8 b3d4 c8a6 f1a6 a8a6 b2b3 e5g6 d1d3 b8b7 f2f3 a6a8 g3f2');
INSERT INTO evaluations VALUES(47632,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(47632,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(47632,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(47632,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(47632,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(47632,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(47632,36,'cp 49','b8d7 f1e1 f8e7 c4b3 e8g8 c2c4 f6h7 b1c3 h7g5 c1e3 a7a5 h2h3 g5f3 d1f3 e5d4 e3d4 d7e5 f3e3 e5c6 a1d1 c8e6 f2f4 e7h4 e1f1 c6d4 e3d4 h4f6 d4d3 f6c3 d3c3 d8f6 c3f6 g7f6 d1d4 a5a4 b3c2 f8d8 f1d1 g8f8 c4c5 e6a2 c5d6 c7d6 d4a4');
INSERT INTO evaluations VALUES(10366,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(10366,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(10366,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(10366,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(10366,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(10366,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(10366,36,'cp -25','e8g8 f1e2 d7d5 e4d5 f6d5 c3d5 b4d2 f3d2 d8d5 e1g1 d5b5 d2c4 c8e6 a2a4 b5c5 c2c3 a7a5 d1b3 e6c4 d3c4 c6d8 f1d1 c7c6 d1d7 b7b6 b3a3 c5a3 a1a3 d8e6 b2b4 a8d8 d7d8 f8d8 c4c5 d8d2 c5b6 d2e2 g1f1');
INSERT INTO evaluations VALUES(4430,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(4430,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(4430,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(4430,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(4430,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(4430,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(4430,36,'cp -16','f1e2 b4c3 d2c3 f8e8 b2b4 a7a6 a2a4 d7d5 b4b5 c6d4 c3d4 e5d4 e4e5 f6g4 e1g1 g4e5 f3d4 d8f6 c2c3 e5g6 d1d2 c8d7 g2g3 c7c5 b5c6 b7c6 f2f4 h7h6 d4f3 a8b8 d3d4 a6a5 e2d3 d7h3 f1e1 h3f5 e1e8 b8e8');
INSERT INTO evaluations VALUES(4431,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(4431,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(4431,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(4431,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(4431,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(4431,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(4431,36,'cp -28','b4c3 d2c3 f8e8 e1g1 d7d5 f1e1 d8d6 e4d5 f6d5 c3d2 h7h6 a2a3 c8f5 b2b4 a8d8 b4b5 c6d4 f3d4 e5d4 d1b1 c7c5 b5c6 d6c6 b1b2 c6c5 e2f3 e8e1 d2e1 b7b6 e1b4 c5c8 a1e1 f5e6 h2h3 d5b4 a3b4 c8c3 b2c3 d4c3 f3c6 d8d4');
INSERT INTO evaluations VALUES(55487,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55487,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(55487,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(55487,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(55487,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(55487,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55487,36,'cp -38','f8d6 d2d4 g8f6 g1f3 e8g8 b1c3 b8d7 c1g5 c7c5 f1e2 h7h6 g5h4 b7b6 d4c5 d6c5 c3e4 c5e7 e4f6 d7f6 e1g1 c8b7 f1e1 d8c7 h4g3 c7c6 e2d3 f8d8 d1e2 a7a6 a1d1 b6b5 c2c3 d8d5 g3e5 a8d8 d3c2 d5d1 e1d1 f6d7 e5d4');
INSERT INTO evaluations VALUES(38687,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(38687,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(38687,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(38687,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(38687,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(38687,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(38687,36,'cp -31','d2d4 g8f6 g1f3 e8g8 b1c3 b7b6 f1d3 c8b7 e1g1 b8d7 f1e1 c7c5 c3e4 f6e4 d3e4 b7e4 e1e4 d7f6 c1g5 d6e7 e4e1 c5d4 f3d4 d8d5 g5e3 a8c8 d1e2 d5c4 e2c4 c8c4 c2c3 f6d5 g2g3 d5e3 e1e3 f8d8 a1d1 e7f6 e3d3 f6d4 d3d4 c4d4 d1d4');
INSERT INTO evaluations VALUES(38683,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(38683,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(38683,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(38683,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(38683,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(38683,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(38683,36,'cp -44','g8f6 g1f3 e8g8 f1e2 c7c5 b1c3 b8c6 c1g5 h7h6 g5h4 b7b6 d1d2 c8b7 d4c5 d6c5 e1c1 c6b4 a2a3 b4d5 c3d5 d8d5 h4f6 d5d2 d1d2 g7f6 f3d4 a8d8 h1d1 b7g2 b2b4 d8d4 d2d4 c5d4 d1d4 g2c6 c2c4 g8h8 f2f4 f6f5 b4b5');
INSERT INTO evaluations VALUES(41838,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(41838,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(41838,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(41838,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(41838,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(41838,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(41838,36,'cp -26','g1f3 e8g8 f1e2 c7c5 b1c3 b8c6 d4c5 d6c5 e1g1 h7h6 d1d8 f8d8 f1d1 c8d7 c1f4 a8c8 a2a3 a7a6 h2h3 g8f8 d1d2 d7e8 d2d8 c8d8 a1d1 d8d1 c3d1 f8e7 f4c1 c5d6 d1c3 b7b6 c1e3 b6b5 a3a4 b5b4 c3d1');
INSERT INTO evaluations VALUES(41837,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(41837,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(41837,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(41837,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(41837,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(41837,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(41837,36,'cp -34','e8g8 f1e2 c7c5 b1c3 b8c6 d4c5 d6c5 e1g1 h7h6 d1d8 f8d8 f1d1 c8d7 c1f4 a8c8 a2a3 c5b6 g1f1 a7a6 d1d3 d7e8 d3d8 b6d8 a1d1 c6a5 f3e5 d8c7 a3a4 g7g5 f4g3 a5c6 e5c6 e8c6 g3c7 c8c7');
INSERT INTO evaluations VALUES(10347,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(10347,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(10347,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(10347,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(10347,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(10347,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(10347,36,'cp 38','f3e5 b4c3 b2c3 d8e7 f2f3 c6e5 d4e5 e7e5 d1d4 d7d6 c1e3 e8g8 d4e5 d6e5 c3c4 c8e6 a2a4 b7b6 a4a5 f6d7 f1d3 f8d8 e1d2 g8f8 a1a3 f8e7 h1a1 a8c8 d2c3 h7h5 a5b6 a7b6 g2g3 e7d6 a3a7 d6e7 c3b4 d7b8 e3b6');
INSERT INTO evaluations VALUES(12960,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12960,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12960,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(12960,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12960,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12960,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(12960,36,'cp 26','b1a3 c8d7 a3c2 d8a5 c1d2 a5b6 d4c5 b6c7 c2d4 c6e5 d4b5 d7b5 f1b5 e7c6 d2f4 e5d3 d1d3 c7f4 d3d4 f4d4 c3d4 f8e7 f3e5 a8c8 e1d2 e7f6 f2f4 a7a5 a2a3 e8e7 b5c6 f6e5 f4e5 b7c6 b2b4 f7f6 h1b1 f6e5 d4e5 c8c7 d2c3 h8f8 b4a5 f8a8 a3a4 a8a5 c3d4 e7f7 b1b8 c7a7');
INSERT INTO evaluations VALUES(12963,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12963,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(12963,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(12963,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(12963,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(12963,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12963,36,'cp 34','c8d7 a3c2 d8a5 c1d2 a5b6 d4c5 b6c7 c2d4 c6e5 d4b5 d7b5 f1b5 e7c6 d2f4 e5d3 d1d3 c7f4 d3d4 f4d4 c3d4 f8e7 f3e5 a8c8 e1d2 e7f6 f2f4 a7a5 h1b1 e8e7 b5c6 b7c6 b2b4 a5b4 b1b4 c8c7 b4b6 f6e5 f4e5 h8a8 a2a4 c7a7 a4a5 a7a5 a1a5 a8a5 b6c6 a5a4 d2c3 a4c4 c3d3');
INSERT INTO evaluations VALUES(32474,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(32474,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(32474,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(32474,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(32474,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(32474,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(32474,36,'cp 52','b5c4 f8e7 d2d4 e5d4 f3d4 g8f6 e1g1 e8g8 b1c3 b8c6 a2a4 f8e8 c1f4 e7f8 f1e1 h7h6 d4c6 b7c6 h2h3 d7e6 c4f1 f6d7 b2b3 d8f6 c3e2 d6d5 d1d2 d5e4 e2g3 f6g6 g3e4 e6h3 a1d1 e8e4 e1e4');
INSERT INTO evaluations VALUES(32478,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(32478,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(32478,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(32478,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(32478,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(32478,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(32478,36,'cp 44','f8e7 d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 e1g1 b8c6 a2a4 f8e8 d4c6 b7c6 h2h3 e7f8 c4d3 a8b8 f2f4 d7c8 b2b3 d6d5 e4e5 f6d7 c1b2 f7f6 d1h5 g7g6 d3g6 h7g6 h5g6 f8g7 c3d5 c6d5 e5f6 d7f6 b2f6 d8f6');
INSERT INTO evaluations VALUES(54728,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(54728,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(54728,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(54728,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(54728,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(54728,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54728,36,'cp 47','d2d4 c8f5 f1b5 c7c6 b5d3 f5d3 d1d3 e7e6 g1f3 c6c5 e1g1 b8c6 e5f6 g8f6 f1e1 d8d7 c1f4 c6d4 f3d4 c5d4 b1d2 f8c5 f4e5 e8g8 d2f3 f6e4 c2c3 c5d6 e5d6 d7d6 d3d4 f8f3 e1e4 f3f6 e4e2 d6b6 d4b6 a7b6 f2f3 h7h6');
INSERT INTO evaluations VALUES(54725,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(54725,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(54725,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(54725,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(54725,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(54725,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(54725,36,'cp 51','c8f5 f1b5 c7c6 b5d3 f5d3 d1d3 c6c5 g1f3 b8c6 e1g1 d8d7 f1e1 e7e6 e5f6 g8f6 c1f4 c6d4 f3d4 c5d4 b1d2 f8c5 e1e2 e8g8 d2f3 f6e4 f4e5 d7a4 e5d4 f8f4 d3b3 a4b3 a2b3 c5d4 f3d4 e6e5 d4f3 f4f5 c2c4 e4c5 c4d5 c5b3 a1d1');
INSERT INTO evaluations VALUES(31743,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(31743,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(31743,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(31743,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31743,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(31743,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(31743,36,'cp 54','f1b5 c7c6 b5d3 f5d3 d1d3 c6c5 g1f3 e7e6 e5f6 g8f6 e1g1 b8c6 f1e1 d8d7 c1f4 c6d4 f3d4 c5d4 b1d2 f8c5 d2f3 e8g8 e1e2 f6e4 f4e5 d7f7 e5d4 c5d6 c2c4 b7b6 a1e1 a8d8 d3b3 h7h6 h2h3 d6c5 c4d5 e6d5 b3d3 f7h5 d4e5');
INSERT INTO evaluations VALUES(18678,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(18678,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(18678,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(18678,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18678,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(18678,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(18678,36,'cp 0','b7c6 f1d3 d7d5 e1g1 g8f6 e4e5 f6d7 f1e1 a7a5 c2c4 f8b4 b1c3 d5d4 a2a3 b4e7 c3e2 c6c5 e2f4 c8b7 f4h5 e8g8 d1g4 g7g6 c1h6 f8e8 h5g7 e8f8');
INSERT INTO evaluations VALUES(18388,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18388,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18388,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18388,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18388,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18388,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18388,36,'cp 0','f1d3 d7d5 e1g1 g8f6 e4e5 f6d7 f1e1 a7a5 c2c4 f8b4 b1d2 c8a6 a2a3 b4c5 d1g4 d8b6 g4g3 e8g8 a1b1 f7f6 b2b4 a5b4 a3b4 c5b4 g3h4 h7h6 d2f3 d7e5 e1e5 f6e5 c1h6 f8f3 g2f3 g7h6 h4h6 g8f7 h6h7 f7f6 h7h4 f6f7');
INSERT INTO evaluations VALUES(18387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18387,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18387,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18387,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18387,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18387,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18387,36,'cp 0','d7d5 e1g1 g8f6 e4e5 f6d7 f1e1 a7a5 c2c4 f8b4 b1d2 c8a6 a2a3 b4c5 d1g4 d8b6 g4g3 e8g8 a1b1 f7f6 b2b4 a5b4 a3b4 c5b4 g3h4 h7h6 d2f3 d7e5 e1e5 f6e5 c1h6 f8f3 g2f3 g7h6 h4h6 g8f7 h6h7 f7f6 h7h4 f6f7');
INSERT INTO evaluations VALUES(18487,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18487,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18487,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18487,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18487,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18487,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(18487,36,'cp 2','e1g1 g8f6 b1d2 f8d6 f1e1 d8c7 h2h3 e8g8 c2c4 c8b7 d1e2 e6e5 c4d5 c6d5 e4d5 f6d5 d2e4 d6b4 c1d2 b4d2 e2d2 a8d8 d3f1 c7b6 e4g3 d5f4 d2c3 f7f6 g3f5 g8h8 a1c1 b6e6 f5e3 e6a2 e1d1 a2f7 d1d8 f8d8 c3c7 f7c7 c1c7');
INSERT INTO evaluations VALUES(16394,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(16394,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(16394,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(16394,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(16394,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(16394,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(16394,36,'cp -24','d7d5 e4d5 f6d5 g1f3 d5b6 c4b3 c8f5 d3d4 e5d4 f3d4 c6d4 c3d4 d8f6 d1e2 f8e7 e1g1 e8c8 f1d1 c8b8 b1c3 h8e8 a2a4 a7a6 a4a5 b6c8 h2h3 f5d7 c1e3 c8d6 c3d5 f6f5 d5e7 e8e7 d4d5 d8e8 e2d2 f5g6 b3c2 d7f5 c2f5 g6f5');
INSERT INTO evaluations VALUES(15088,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(15088,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(15088,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(15088,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(15088,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(15088,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(15088,36,'cp -41','e4d5 f6d5 g1f3 d5b6 c4b3 c8f5 d3d4 e5d4 f3d4 c6d4 c3d4 d8f6 d1e2 f8e7 e1g1 e8c8 f1d1 h8e8 b1c3 c8b8 c1e3 f6g6 e2f3 e7b4 d1e1 f5e6 h2h4 e6b3 a2b3 a7a6 e3f4 e8e1 a1e1 f7f5 f4e5 g6f7 e1e2 b4d6 h4h5 g7g5 h5g6 h7g6 e5d6 d8d6');
INSERT INTO evaluations VALUES(14987,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(14987,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(14987,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(14987,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(14987,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(14987,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(14987,36,'cp -39','f6d5 g1f3 d5b6 c4b3 c8f5 d3d4 e5d4 f3d4 c6d4 c3d4 d8f6 d1e2 f8e7 e1g1 e8c8 f1d1 h8e8 b1c3 c8b8 c1e3 f6g6 e2f3 e7b4 d1e1 f5e6 b3e6 g6e6 d4d5 e6d7 e3b6 a7b6 e1e8 d8e8 a1d1 e8e5 g2g3 g7g5 f3f6 b4c3 b2c3 e5d5 d1d5 d7d5');
INSERT INTO evaluations VALUES(15314,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(15314,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(15314,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(15314,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(15314,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(15314,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(15314,36,'cp -23','g1f3 d5b6 c4b3 c8f5 d3d4 e5d4 c3d4 f8d6 e1g1 e8g8 a2a4 a7a5 b1c3 f8e8 c1d2 d8d7 c3b5 f5e6 a1c1 h7h6 h2h3 a8d8 f1e1 e6d5 e1e8 d8e8 b3d5 b6d5 b5d6 d7d6 d1b3 b7b6 b3b5 e8e6 b5a6 d5e7 a6c4 d6d5 c4d5 e7d5 c1c4 c6e7');
INSERT INTO evaluations VALUES(15312,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(15312,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(15312,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(15312,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(15312,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(15312,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(15312,36,'cp -34','d5b6 c4b3 c8f5 d3d4 e5d4 c3d4 f8d6 e1g1 e8g8 b1c3 h7h6 c1e3 f8e8 h2h3 d8d7 c3b5 f5e6 a1c1 e6b3 d1b3 d7f5 b5d6 c7d6 a2a3 a8d8 f1e1 e8e6 f3h4 f5d5 b3c2 c6d4 e3d4 e6e1 c1e1 d5d4 h4f3 d4f6');
INSERT INTO evaluations VALUES(15423,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(15423,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(15423,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(15423,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(15423,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(15423,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(15423,36,'cp 74','c8d7 d2d4 f8e7 d4d5 c6b8 a2a4 e8g8 b2b3 d7g4 b5e2 b8d7 a4a5 a7a6 f1e1 g4h5 f3d2 h5e2 d1e2 g7g6 c3a4 b7b5 a5b6 c7b6 c2c4 f6h5 d2f3 h5g7 c1h6 f7f5 b3b4 a6a5 b4a5 a8a5 a4c3 e7f6 h2h3 d7c5 f3d2 a5a1 e1a1');
INSERT INTO evaluations VALUES(25503,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(25503,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(25503,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(25503,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(25503,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(25503,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(25503,36,'cp 64','d2d4 f8e7 d4d5 c6b8 a2a4 e8g8 b2b3 d7g4 b5e2 b8d7 a4a5 a7a6 c1e3 c7c6 f3d2 g4e2 d1e2 d8c7 e2d3 c6d5 c3d5 f6d5 d3d5 a8c8 d2c4 c7c6 d5c6 c8c6 a1d1 f7f5 e4f5 f8f5 f2f3 f5f8 f1e1 f8c8 g2g3 h7h6 d1d3 d7f6 e3b6');
INSERT INTO evaluations VALUES(20038,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(20038,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(20038,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(20038,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20038,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(20038,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(20038,36,'cp 17','f1b5 c8d7 g1f3 g8f6 d2d3 e7e6 e1g1 a7a6 b5c4 d8c7 d1e1 c6d4 c4b3 d4b3 a2b3 f8e7 e1g3 e8g8 g3h3 d7c6 c1d2 b7b5 g2g4 f6d7 f4f5 e6f5 g4f5 e7f6 d2f4 d7e5 f4e5 d6e5 h3h5 g8h8 g1h1 c6b7 f3g5');
INSERT INTO evaluations VALUES(19928,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19928,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19928,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(19928,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(19928,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(19928,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19928,36,'cp 0','c8d7 g1f3 g7g6 b5c6 d7c6 d2d4 c5d4 f3d4 f8g7 c1e3 d8b6 e1g1 b6b2 d1d3 b2a3 a1b1 a3a5 h2h3 a8d8 c3b5 g8f6 d4c6 b7c6 b5a7 a5a4 d3d4 a4a6 d4d3');
INSERT INTO evaluations VALUES(26059,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(26059,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(26059,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(26059,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(26059,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(26059,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(26059,36,'cp 8','g1f3 g8f6 d2d3 e7e6 e1g1 a7a6 b5c4 d8c7 d1e1 e8c8 h2h3 c8b8 e1f2 d7e8 a2a3 d6d5 c4a2 d5e4 d3e4 c6d4 c1e3 d4f3 f2f3 e8c6 a2c4 b7b5 c4d3 c5c4 d3e2 c7b7 a1d1 d8d1 f1d1 f6e4 c3e4 c6e4');
INSERT INTO evaluations VALUES(26058,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(26058,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(26058,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26058,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(26058,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(26058,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26058,36,'cp 8','g7g6 e1g1 f8g7 d2d3 g8f6 a2a4 e8g8 h2h3 a7a6 b5c4 e7e6 c1e3 c6e7 c4b3 d7c6 a4a5 b7b5 a5b6 d8b6 d1e1 a6a5 b3a4 f6d7 b2b3 c6a4 a1a4 e7c6 e1d2 c6d4 c3e2 d4f3 f1f3 f7f5 f3f1 b6a6 e2c3 d7b6 a4a2 d6d5 e4d5 b6d5 c3d5 e6d5');
INSERT INTO evaluations VALUES(26005,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(26005,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(26005,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(26005,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26005,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26005,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26005,36,'cp 13','e1g1 g8f6 a2a4 f8g7 d2d3 e8g8 c1e3 a7a6 b5c4 e7e6 h2h3 c6e7 c4b3 d7c6 a4a5 b7b5 a5b6 d8b6 d1d2 b6b7 b3a4 f6d7 a1a2 c6a4 c3a4 f8b8 c2c3 h7h6 e3f2 a6a5 f1a1 e7c6 d2e1 d7b6 e1d1 b6d7 b2b3');
INSERT INTO evaluations VALUES(26004,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26004,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26004,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26004,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26004,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26004,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26004,36,'cp 0','f8g7 d2d3 g8f6 a2a3 e8g8 h2h3 c6d4 b5d7 f6d7 f3d4 c5d4 c3e2 a8c8 c1d2 d8b6 a1b1 e7e6 d2e1 b6c6 b1c1 c6b6');
INSERT INTO evaluations VALUES(23805,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23805,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23805,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23805,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23805,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23805,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(23805,36,'cp 16','d2d3 g8f6 a2a4 e8g8 h2h3 a7a6 b5c4 c6d4 c1e3 f6e8 d1d2 b7b5 a4b5 d4f3 f1f3 a6b5 a1a8 d8a8 c3b5 a8a1 d2c1 a1b2 c1b2 g7b2 f3f1 e8f6 f1b1 d7b5 c4b5 b2d4 e3d4 c5d4 b5c4 e7e6 b1b7 d6d5 e4d5 f6d5 c4d5 e6d5 g1f2 f8c8 b7b2 g8g7 f2f3');
INSERT INTO evaluations VALUES(23806,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(23806,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(23806,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(23806,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(23806,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(23806,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23806,36,'cp 11','g8f6 a2a4 e8g8 c1e3 a7a6 b5c4 e7e6 h2h3 c6e7 c4b3 d7c6 a4a5 b7b6 a5b6 d8b6 d1e1 a6a5 b3a4 f6d7 b2b3 c6a4 c3a4 b6c7 a1b1 e7c6 e1f2 a8c8 b1d1 c6b4 c2c3 b4c6 d3d4 c5d4 f3d4 c6d4 e3d4 g8h8 d4e3 h8g8');
INSERT INTO evaluations VALUES(48284,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(48284,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(48284,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48284,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48284,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48284,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48284,36,'cp 0','c7c5 g1f3 d8b6 b2b3 f8d6 f4g3 e8g8 e1g1 d6g3 h2g3 c8d7 d1e2 b8c6 c2c3 a8c8 b1d2 c5d4 e3d4 c8c7 f1b1 f8c8 d2f1 b6a5 b3b4 a5b6 f1d2 c6e7 c3c4 d5c4 d2c4 b6c6 a2a4 c6d5 c4e3 d5d6 e3c4');
INSERT INTO evaluations VALUES(46201,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46201,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(46201,36,'cp 11','g1f3 d8b6 b2b3 f8d6 f4g3 e8g8 e1g1 d6g3 h2g3 b8c6 b1c3 c5d4 e3d4 c8d7 c3e2 f8e8 a2a4 a7a5 d1d2 e6e5 d4e5 c6e5 f3e5 e8e5 f1e1 g7g6 c2c3 h7h5 e2d4 a8e8 e1e5 e8e5 a1e1 e5e1 d2e1 b6c5 e1d2 f6g4');
INSERT INTO evaluations VALUES(46202,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(46202,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46202,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46202,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(46202,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(46202,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(46202,36,'cp 0','d8b6 b2b3 f8d6 f4g3 c8d7 e1g1 e8g8 g3e5 d6e7 d4c5 b6c5 a2a3 f8d8 b1d2 d7b5 e5d4 c5c6 c2c4 d5c4 d2c4 c6d5 a3a4 b5e8 c4e5 b8c6 e5c6 e8c6 d3c4 d5e4 c4d3');
INSERT INTO evaluations VALUES(51932,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(51932,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(51932,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(51932,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(51932,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(51932,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(51932,36,'cp -13','c5d4 d1d4 b8c6 d4e3 f8e7 g1f3 d7d5 e4d5 e6d5 c4b3 a7a5 b1c3 g8f6 e3d3 e7b4 a2a3 b4c3 b2c3 a5a4 b3a2 d8e7 c1e3 c8g4 f3d4 e8g8 e1g1 h7h6 d4c6 b7c6 f1e1 g4e6 c3c4 f8e8 c4d5 e6d5 e3d2 e7c7 e1e8 a8e8 a2d5 c6d5 h2h3 e8c8 d2e3 c7e5');
INSERT INTO evaluations VALUES(52046,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(52046,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(52046,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(52046,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52046,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52046,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(52046,36,'cp -8','d1d4 b8c6 d4e3 g8f6 g1f3 f6e4 e3e4 d7d5 c4d5 d8d5 e4d5 e6d5 c1e3 c8g4 b1d2 f7f6 d2b3 f8d6 e1c1 c6e7 h2h3 g4f3 g2f3 e8f7 c2c3 h8c8 f3f4 c8c4 d1d4 c4d4 b3d4 d6c5 h1d1 c5d4 e3d4 e7f5 a2a4 f7e6 c1c2 b7b6 d4e3 a8d8');
INSERT INTO evaluations VALUES(52044,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(52044,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(52044,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(52044,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(52044,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52044,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(52044,36,'cp -13','b8c6 d4e3 f8e7 g1f3 g8f6 e1g1 d8c7 e3e2 e8g8 c4b3 d7d6 c2c4 b7b6 b1c3 a7a6 b3c2 c8b7 c1d2 c6e5 f3e5 d6e5 a2a3 f8d8 d2e3 a6a5 a1d1 d8d1 f1d1 a8d8 h2h4 b7a6 b2b3 d8d1 e2d1 a6b7 h4h5');
INSERT INTO evaluations VALUES(18887,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(18887,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(18887,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18887,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(18887,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(18887,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(18887,36,'cp -9','d4e3 f8e7 g1f3 d7d5 e4d5 e6d5 c4b3 a7a5 b1c3 g8f6 e3d3 e7b4 a2a3 b4c3 b2c3 a5a4 b3a2 d8e7 c1e3 e7a3 e1g1 a3e7 a2b3 a4a3 c3c4 d5c4 b3c4 e8g8 e3c1 f8d8 d3b3 c6a5 b3a3 e7a3 a1a3 c8f5 c4b3 a5b3 a3a8 d8a8 c2b3 f5d3 f1e1 a8a1 c1g5 a1e1 f3e1');
INSERT INTO evaluations VALUES(18881,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(18881,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(18881,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(18881,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(18881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18881,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(18881,36,'cp -9','f8e7 g1f3 d7d5 e4d5 e6d5 c4b3 a7a5 b1c3 g8f6 e3d3 e7b4 b3a4 f6e4 e1g1 b4c3 b2c3 e8g8 c1a3 f8e8 f1e1 c8g4 a4c6 b7c6 f3d4 c6c5 f2f3 c5d4 f3e4 d4c3 e4d5 d8g5 e1e8 a8e8 d3c3 g5d5 a3b2 d5g5 a1e1 g4d7 e1e8 d7e8 c3g3 g5g3 h2g3 e8c6 g1h1 f7f6 b2c3 g8f7 c3a5');
INSERT INTO evaluations VALUES(15801,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(15801,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(15801,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(15801,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(15801,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(15801,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(15801,36,'cp 88','c2c3 d4f5 d2d3 f6d5 e1g1 f7f6 g5f3 f8e7 f1e1 c7c6 d3d4 e5d4 d1e2 e8f7 f3d4 e7d6 c1d2 g7g6 e2f3 f7g7 b1a3 f5d4 c3d4 d5b6 c4b3 c8f5 a3b1 d8c7 h2h3 h7h5 b1c3 a8e8 c3e4 e8e4 e1e4 f5e4');
INSERT INTO evaluations VALUES(15798,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(15798,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(15798,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(15798,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(15798,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(15798,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(15798,36,'cp 55','b7b5 c4f1 f6d5 c3d4 d8g5 f1b5 e8d8 e1g1 c8b7 d1f3 a8b8 d4e5 d5e3 f3h3 g5g2 h3g2 e3g2 d2d4 f7f6 f2f4 g2h4 b5e2 h4f5 b1c3 f5d4 f1d1 c7c5 c1e3 d8c7 a1c1 f6e5 f4e5 d4e2 c3e2 b7f3 e3c5 f8c5 c1c5 c7b6 c5c2 h8d8 d1d8 b8d8 e5e6 f3e2 c2e2 b6c6 e6e7 d8e8 g1f2 c6d6 b2b4 e8e7 e2e7 d6e7 f2f3 e7f6 f3e4 g7g5');
INSERT INTO evaluations VALUES(11885,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(11885,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(11885,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(11885,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(11885,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(11885,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(11885,36,'cp 86','c4f1 f6d5 c3d4 d8g5 f1b5 e8d8 e1g1 c8b7 d1f3 a8b8 d4e5 d5e3 f3h3 g5g2 h3g2 e3g2 d2d4 f7f6 b5e2 g2h4 f2f4 h4f5 b1c3 f5d4 f1d1 c7c5 c1e3 f6e5 f4e5 d8e8 e2b5 b7c6 b5c6 d4c6 c3d5 e8d7 d5b4 c6d4 b4c2 b8b2 a1c1 d7e6 c2d4 c5d4 e3d4 b2a2 c1c8 a2a4 g1h1 e6d7 c8a8 d7e6 a8d8 h8g8 d4c5 a4f4 d8c8 f8c5 c8g8');
INSERT INTO evaluations VALUES(11886,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(11886,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(11886,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(11886,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(11886,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(11886,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(11886,36,'cp 52','f6d5 c3d4 d8g5 f1b5 e8d8 e1g1 c8b7 d1f3 a8b8 d4e5 d5e3 f3h3 g5g2 h3g2 e3g2 d2d4 f7f6 b5e2 g2h4 b1d2 f8e7 d2c4 h4f3 e2f3 b7f3 c1e3 h8f8 f1c1 f6f5 e3d2 d8d7 b2b3 f3e4 c1c3 d7e6 a1c1 f5f4 f2f3 e4d5 c4a5 e7b4 c3c2 d5f3 d2b4 b8b4 a5c6');
INSERT INTO evaluations VALUES(48359,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(48359,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(48359,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(48359,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(48359,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(48359,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(48359,36,'cp 33','f1g2 f8b4 c2c3 b4e7 c3c4 c7c6 b2b3 e8g8 e1g1 b7b6 c1b2 b8d7 d1c2 c8b7 f1d1 c6c5 c4d5 f6d5 b1c3 c5d4 c3d5 b7d5 f3d4 a8c8 c2b1 d5g2 g1g2 e7f6 b1e4 h7h6 e4f3 d7e5 f3b7 c8c7 b7e4 d8a8 e4a8 f8a8 a1c1 c7d7 d4b5 a8d8 d1d7 d8d7 c1c8 g8h7 b2c3 d7d5 b5d4 h7g6 c8c7 d5d7');
INSERT INTO evaluations VALUES(50180,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(50180,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(50180,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(50180,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(50180,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(50180,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50180,36,'cp 53','b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 f8g7 d2d4 c7c5 a1b1 e8g8 f1e2 c5d4 c3d4 d8a5 c1d2 a5a2 e1g1 a2e6 d1c2 b7b6 d2e3 e6d6 f1c1 c8b7 d4d5 b7a6 e2a6 b8a6 g2g3 a6c5 e3c5 d6c5 c2c5 b6c5 c1c5 f8c8');
INSERT INTO evaluations VALUES(50173,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(50173,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(50173,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(50173,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(50173,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(50173,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(50173,36,'cp 33','d7d5 d2d4 f8g7 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 f1d3 b7b6 d2a5 c6a5 e1e2 c8g4 h2h3 g4f3 e2f3 a5c6 d4d5 c6d4 f3g3 f7f5 h1d1 f5e4 d3e4 d4f5 e4f5 g6f5 a2a4 f8d8');
INSERT INTO evaluations VALUES(46748,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(46748,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(46748,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(46748,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(46748,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(46748,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(46748,36,'cp 29','c4d5 f6d5 d2d4 f8g7 e2e4 d5c3 b2c3 c7c5 a1b1 e8g8 f1e2 c5d4 c3d4 d8a5 c1d2 a5a2 e1g1 c8g4 d2e3 b8c6 d4d5 g4f3 e2f3 c6e5 b1b7 g7f6 d1b1 a2b1 f1b1 a7a5 f3e2 a5a4 f2f4 a4a3 f4e5 f6e5 e2c4 a3a2 c4a2 a8a2 b7e7 e5f6 e7d7 a2e2 e3h6 f8c8 d5d6 e2e4 g1f1 e4d4');
INSERT INTO evaluations VALUES(42550,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42550,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42550,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42550,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42550,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42550,36,'cp 0','a7a5 g1f3 d7d6 c1e3 c5e3 f2e3 e8g8 a2a4 c7c6 d1e2 b8d7 h2h3 d7c5 e1g1 c8e6 c4e6 f7e6 a1d1 d8c7 e2e1 h7h6 b2b3 f8f7 f3h4 g8h7 e1g3 g7g5 h4f3 a8f8 f3h2 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(57053,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(57053,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(57053,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(57053,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(57053,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(57053,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(57053,36,'cp -22','g8f6 b1c3 d7d5 d2d3 a7a6 e2e4 d5e4 d3e4 d8d1 e1d1 b7b6 f1d3 c8b7 c1d2 b8d7 e4e5 f6d5 c3d5 b7d5 c2c4 d5b7 d1e2 f8e7 a1d1 e8c8 b2b4 h7h6 h2h4 f7f6 e5f6 g7f6 g2g4 h8g8 h1g1 d7c5 d3h7 g8g7 h7c2 c5e4 d2e1 d8d1 e2d1');
INSERT INTO evaluations VALUES(49459,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(49459,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(49459,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(49459,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(49459,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(49459,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(49459,36,'cp -13','d2d3 d7d5 b1c3 b7b6 e2e4 d5e4 c3e4 c8b7 c1d2 f8e7 d1e2 b8c6 e1c1 c6d4 f3d4 d8d4 d1e1 e8g8 g2g3 f6e4 d3e4 f8d8 d2c3 d4a4 e2c4 a4c4 f1c4 a7a6 c4f1 a6a5 c3e5 e7b4 c2c3 b4d6 e5d6 d8d6 f1b5');
INSERT INTO evaluations VALUES(22610,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(22610,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(22610,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(22610,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(22610,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(22610,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(22610,36,'cp 108','e2e4 d7d5 b1d2 g8f6 e4e5 f6d7 d2b3 f8e7 c1f4 e8g8 c2c3 f7f6 e5f6 g7f6 f1b5 f8f7 f4g3 e7d6 e1g1 c6e7 f1e1 d7f8 b5d3 f7g7 a2a4 a7a5 b3c1 g8h8 f3h4 b7b6 d3c2 c7c5 g3d6 d8d6 d1f3 e7g8 d4c5 b6c5 c1e2');
INSERT INTO evaluations VALUES(11596,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(11596,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(11596,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(11596,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(11596,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11596,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(11596,36,'cp 33','e7g6 d2d4 e5d4 f3d4 f8c5 d4f5 e8g8 c1e3 c5e3 f5e3 f8e8 f2f3 b7b5 a4b3 a6a5 a2a3 c8a6 f1f2 b5b4 e3d5 b4a3 b1a3 c6b4 d5c3 d8e7 a3b5 a6b5 c3b5 e7c5 b3a4 e8b8 c2c3 b4a6 d1d7 g6e5 d7d5 c7c6 d5c5 a6c5 b5d4 c5a4 a1a4 e5d3 d4c6 d3b2 c6b8');
INSERT INTO evaluations VALUES(11549,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(11549,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(11549,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11549,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11549,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(11549,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(11549,36,'cp 47','d2d4 e5d4 f3d4 f8c5 d4f5 e8g8 c1e3 c5e3 f5e3 f8e8 f2f3 b7b5 a4b3 a6a5 a2a3 c8a6 d1d2 b5b4 f1e1 c6e5 g1h1 c7c5 b3d5 a8b8 d2f2 d8f6 b1d2 a5a4 a3b4 b8b4 b2b3 a6b5 g2g3 a4b3 d5b3 b5c6 a1a5 d7d6 f3f4 b4e4 d2e4');
INSERT INTO evaluations VALUES(43588,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43588,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43588,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43588,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43588,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43588,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43588,36,'cp 0','e2e3 b7b5 a2a3 e7e6 b2b4 b8d7 f1e2 c8b7 g1f3 f8d6 e1g1 e8g8 d1e1 h7h6 f3e5 d7b6 e5d3 b7c6 d3e5');
INSERT INTO evaluations VALUES(43586,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(43586,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(43586,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(43586,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(43586,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43586,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43586,36,'cp 0','e7e6 g2g4 f8b4 g1e2 e8g8 a2a3 b4e7 g4g5 f6h5 h2h4 h5f4 e2f4 c7c5 f1d3 c5c4 d3h7 g8h7 d1h5 h7g8 g5g6 f7g6 f4g6 f8f5 h5h8 g8f7 h8h7 b8c6 h1g1 e7f6 e1c1 d8g8 g6h8 f7f8 h8g6');
INSERT INTO evaluations VALUES(43650,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43650,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43650,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43650,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43650,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43650,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43650,36,'cp 0','g2g4 f8b4 g1e2 e8g8 a2a3 b4e7 g4g5 f6h5 h2h4 h5f4 e2f4 c7c5 f1d3 c5c4 d3h7 g8h7 d1h5 h7g8 g5g6 f7g6 f4g6 f8f5 h5h8 g8f7 h8h7 b8c6 h1g1 d8g8 g6h8 f7f8 h8g6');
INSERT INTO evaluations VALUES(52741,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(52741,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(52741,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(52741,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(52741,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52741,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(52741,36,'cp 25','d2d4 g8f6 b1c3 e7e6 c1g5 b8d7 e2e3 d8a5 c4d5 f6d5 d1d2 f8b4 a1c1 h7h6 g5h4 d5c3 b2c3 b4a3 c1c2 b7b6 f1d3 c8a6 e1g1 e8g8 c3c4 a5d2 f3d2 f8e8 f2f4 a3b4 f4f5 e6e5 d2e4 e5d4 e3d4 a8c8 h4f2');
INSERT INTO evaluations VALUES(12759,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12759,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12759,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12759,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(12759,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(12759,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12759,36,'cp 0','b7c6 f1d3 e7e5 e1g1 f8e7 b1d2 d7d5 e4d5 c6d5 f1e1 d8c7 d3b5 c8d7 b5d7 f6d7 d2f3 c7d6 d1e2 f7f6 c1e3 d7b6 a1d1 e8g8 e3b6 a7b6 c2c4 a8a5 f3d4 e5d4 e2e7 d6e7 e1e7 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(12017,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12017,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12017,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12017,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12017,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12017,36,'cp 0','f1d3 e7e5 e1g1 f8e7 b1d2 d7d5 e4d5 c6d5 f1e1 d8c7 d3b5 c8d7 b5d7 f6d7 d2f3 c7d6 d1e2 f7f6 c1e3 d7b6 a1d1 e8g8 e3b6 a7b6 c2c4 a8a5 f3d4 e5d4 e2e7 d6e7 e1e7 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(12016,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12016,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12016,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12016,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12016,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12016,36,'cp 0','e7e5 e1g1 f8e7 b1d2 d7d5 e4d5 c6d5 f1e1 d8c7 d3b5 c8d7 b5d7 f6d7 d2f3 c7d6 d1e2 f7f6 c1e3 d7b6 a1d1 e8g8 e3b6 a7b6 c2c4 a8a5 f3d4 e5d4 e2e7 d6e7 e1e7 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(11996,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11996,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11996,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11996,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11996,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11996,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11996,36,'cp 0','e1g1 f8e7 b1d2 d7d5 e4d5 c6d5 f1e1 d8c7 d3b5 c8d7 b5d7 f6d7 d2f3 c7d6 d1e2 f7f6 c1e3 d7b6 a1d1 e8g8 e3b6 a7b6 c2c4 a8a5 f3d4 e5d4 e2e7 d6e7 e1e7 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(11995,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11995,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11995,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11995,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11995,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11995,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11995,36,'cp 0','f8e7 b1d2 d7d5 e4d5 c6d5 f1e1 d8c7 d3b5 c8d7 b5d7 f6d7 d2f3 c7d6 d1e2 f7f6 c1e3 d7b6 a1d1 e8g8 e3b6 a7b6 c2c4 a8a5 f3d4 e5d4 e2e7 d6e7 e1e7 a5a2 f2f3 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(19227,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(19227,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19227,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(19227,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(19227,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19227,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19227,36,'cp 5','f8g7 e1g1 d7d6 f1e1 a7a6 a2a4 e7e6 d2d3 g8e7 c3e2 d6d5 c4a2 b7b6 c2c3 e8g8 e4d5 e7d5 c1g5 d8d7 d1d2 c8b7 e2g3 f7f6 g5h6 a8d8 a1d1 d5c7 h6e3 c6a5 b2b4 b7f3 g2f3 a5b7 d3d4 c5b4 c3b4 g8h8');
INSERT INTO evaluations VALUES(19936,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(19936,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(19936,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(19936,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(19936,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(19936,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(19936,36,'cp -98','e7e6 g5f3 g8f6 d2d3 d6d5 e4d5 e6d5 c4b3 d8e7 d1e2 c8g4 e2e7 f8e7 b1d2 e8g8 h2h3 g4e6 b3a4 f8c8 e1g1 h7h6 c2c3 a7a6 a4c6 c8c6 f1e1 e7d6 a2a4 a6a5 b2b3 a8e8 d3d4 b7b6 c1a3 c6c8 c3c4 f6d7 a3b2 c5d4 f3d4 d5c4');
INSERT INTO evaluations VALUES(20663,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(20663,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(20663,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(20663,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(20663,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(20663,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(20663,36,'cp -87','g5f3 g8f6 d2d3 d6d5 e4d5 e6d5 c4b3 h7h6 c1f4 g7g5 f4g3 f8g7 e1g1 e8g8 f1e1 f6h5 b1c3 c8e6 b3a4 h5g3 h2g3 d8a5 a4c6 b7c6 a1b1 e6g4 d1d2 a8e8 a2a3 e8e1 b1e1 f8b8 d2c1 g7c3 b2c3 a5c3 c1e3 c3c2 e1c1 c2a2 e3e5 b8b1 e5e8 g8g7');
INSERT INTO evaluations VALUES(52407,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(52407,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52407,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(52407,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(52407,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(52407,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(52407,36,'cp -23','a7a6 b5e2 b8c6 g1f3 g8f6 b1c3 e7e5 e1g1 d7d6 d2d3 f8e7 c3d5 e8g8 c2c4 b7b5 b2b3 a8b8 c1d2 c8g4 d2c3 g4d7 c3d2 h7h6 h2h3 f6e8 f3h2 e8c7 f2f4 e5f4 d2f4 c7d5 c4d5 c6e5 f4e5 d6e5');
INSERT INTO evaluations VALUES(50585,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(50585,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(50585,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(50585,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(50585,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(50585,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(50585,36,'cp -11','b5e2 b8c6 g1f3 e7e5 d2d3 g8f6 b1c3 f8e7 c3d5 e8g8 e1g1 d7d6 c2c3 c8e6 d5f6 e7f6 c1e3 d8e7 f3d2 a8d8 e2g4 d6d5 e4d5 e6d5 d1e2 b7b6 d2e4 f6h4 g2g3 h7h6 g3h4 f7f5 g4f3 f5e4 d3e4 d5e6 f3g2 b6b5 a1d1 e6a2 h4h5');
INSERT INTO evaluations VALUES(38671,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(38671,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(38671,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38671,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38671,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38671,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38671,36,'cp 12','g8f6 g1f3 c7c5 d4c5 d6c5 b1d2 e8g8 c2c4 c5e7 a2a3 a7a5 f1e2 d8b6 d1c2 c8d7 a1c1 b8c6 e1g1 a5a4 f1d1 f8c8 c2b1 h7h6 f3e5 c6e5 g3e5 b6a6 h2h3 b7b5 c4d5 c8c1 d1c1 f6d5 d2f3 a8c8 e5d4 f7f6 f3d2 c8c1 b1c1 a6b7');
INSERT INTO evaluations VALUES(41812,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41812,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41812,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41812,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41812,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41812,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41812,36,'cp 6','c2c4 c7c5 d4c5 d6c5 g1f3 d5c4 d1d8 e8d8 f1c4 f6e4 g3f4 f7f6 b1d2 e4d2 f3d2 e6e5 f4g3 d8e7 d2e4 c5b6 e1e2 b8c6 h1d1 c8e6 c4e6 e7e6 d1d6 e6e7 f2f3 h8d8 d6d8 a8d8 g3e1 h7h5 h2h3 d8c8 a1c1 f6f5 e4g5 g7g6');
INSERT INTO evaluations VALUES(41777,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(41777,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(41777,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(41777,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(41777,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(41777,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41777,36,'cp 12','c7c5 d4c5 d6c5 b1d2 c5e7 g1f3 e8g8 a2a3 a7a5 f1d3 b8d7 e1g1 d7c5 d3e2 c8d7 g3e5 d7c6 c4d5 c6d5 a1c1 d8b6 e5d4 a5a4 f3e5 b6a5 e2c4 a8c8 d1e2 d5c4 c1c4 a5a6 f1c1 f6e4 d2e4 c5e4 e2c2 c8c4 c2c4');
INSERT INTO evaluations VALUES(45022,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(45022,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(45022,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(45022,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45022,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45022,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(45022,36,'cp 31','a2a4 c8f5 f3e5 b8d7 e5c4 d7b6 c4e5 a7a5 e2e3 b6d7 e5c4 f5g6 f1e2 d7b6 c4e5 b6d7 e5g6 h7g6 e1g1 e7e5 d4e5 d7e5 d1c2 f8b4 b2b3 d8e7 c1b2 e8g8 a1d1 f8d8 c3a2 b4d6 b2d4 e5d7 h2h3 d6e5 a2c3');
INSERT INTO evaluations VALUES(45053,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(45053,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(45053,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(45053,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(45053,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45053,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(45053,36,'cp 29','c8f5 f3e5 b8d7 e5c4 d7b6 c4e5 a7a5 e2e3 b6d7 e5c4 f5g6 f1e2 d7b6 c4e5 b6d7 d1b3 d8c7 h2h4 d7e5 d4e5 f6d7 e5e6 f7e6 e3e4 e6e5 c1e3 g6f7 e2c4 e7e6 c4e6 d7c5 e6f7 c7f7 b3f7 e8f7 e1c1 f7e6 h4h5 f8e7 c1c2 h8d8 f2f3 d8d7 d1d7 c5d7 h5h6 g7g6');
INSERT INTO evaluations VALUES(29073,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(29073,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(29073,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(29073,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(29073,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(29073,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(29073,36,'cp 31','f3e5 b8d7 e5c4 d7b6 c4e5 a7a5 h2h4 e7e6 f2f3 h7h6 e2e4 f5h7 c1e3 f6d7 e5d3 f8d6 e3f2 e8g8 a1c1 d8e7 f1e2 f7f6 e1g1 g8h8 d1b3 d6c7 f3f4 a8d8 g2g4 h7g6 h4h5 g6h7 c1e1 e6e5 d4e5 f6e5 f4f5');
INSERT INTO evaluations VALUES(29061,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(29061,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29061,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(29061,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(29061,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(29061,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(29061,36,'cp 18','b8d7 e5c4 d7b6 c4e5 a7a5 e2e3 b6d7 e5c4 f5g6 f1e2 d7b6 d1b3 f6d7 e3e4 b6c4 b3c4 e7e5 c1e3 f8b4 a1d1 g6e4 d4e5 e4c2 d1d2 c2f5 c4f4 f5e6 f4g3 d8e7 g3g7 e8c8 e1g1 d8g8 g7h6 d7e5 g2g3 e5g4 e2g4 g8g4');
INSERT INTO evaluations VALUES(55537,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55537,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55537,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55537,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(55537,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(55537,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55537,36,'cp 0','c2c4 f8e7 b1c3 g8f6 e2e3 e8g8 d1c2 f8e8 f1e2 e7f8 g1f3 c7c6 e1g1 c8g4 h2h3 g4h5 d2d3 b8d7 c2b3 a7a5 a2a3 h7h6 f1c1 h5g6 e3e4 g6h5 c1e1 f6h7 f3h2 h5g6 h2f3');
INSERT INTO evaluations VALUES(55539,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55539,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55539,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55539,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55539,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55539,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55539,36,'cp 0','f8e7 e2e3 g8f6 b1c3 e8g8 d1c2 f8e8 g1f3 e7f8 f1e2 c7c6 e1g1 c8g4 h2h3 g4h5 c2b3 b8d7 d2d3 a7a5 a2a3 h7h6 f1d1 h5g6 f3h4 g6h7 h4f3');
INSERT INTO evaluations VALUES(47115,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(47115,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(47115,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(47115,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(47115,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(47115,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(47115,36,'cp -89','f6e4 g1f3 e4g5 f3g5 d5c4 b1c3 c7c6 f2f4 e7e6 e2e3 h7h6 g5f3 b7b5 f3e5 f8b4 d1h5 e8g8 f1e2 c8b7 a2a3 b4c3 b2c3 c6c5 e1g1 b8d7 e5d7 d8d7 h5c5 b7d5 a3a4 a7a6 a4b5 f8c8 c5b4 a6b5 f1b1 d5c6 e2f3 c6f3 g2f3 d7d5 g1f2 a8a1');
INSERT INTO evaluations VALUES(47479,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(47479,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(47479,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(47479,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(47479,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(47479,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(47479,36,'cp -76','g1f3 e4g5 f3g5 d5c4 b1c3 e7e6 h2h4 a7a6 a2a4 h7h6 g5e4 b8c6 e2e3 c6a5 e4d2 f8b4 d2c4 c7c5 d4c5 d8d1 e1d1 a5c4 f1c4 c8d7 c4e2 e8e7 e2f3 b4c5 h4h5 a8c8 h1h4 c5d6 f3b7 c8b8 b7a6 b8b2 c3e2 h8b8');
INSERT INTO evaluations VALUES(54863,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(54863,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(54863,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(54863,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(54863,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(54863,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(54863,36,'cp 63','c2c4 c7c6 b1c3 g8f6 c4d5 c6d5 c1f4 b8c6 e2e3 f6h5 f4e5 f7f6 e5g3 e7e6 a1c1 f8e7 a2a3 h5g3 h2g3 e8g8 f1d3 c8d7 e1g1 a8c8 e3e4 d5e4 d3e4 f6f5 e4c2 e7f6 f1e1 g8g7 d4d5 e6d5 c3d5 f6b2 c1b1');
INSERT INTO evaluations VALUES(54847,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(54847,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(54847,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(54847,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(54847,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(54847,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(54847,36,'cp 54','c7c6 c4d5 c6d5 c1f4 g8f6 e2e3 b8c6 b1c3 f6h5 f4e5 f7f6 e5g3 e7e6 f1b5 c8d7 e1g1 a7a6 b5a4 f8b4 e3e4 b4c3 e4d5 e6d5 b2c3 e8g8 d1b3 c6e7 a4d7 d8d7 f1b1 b7b5 f3e1 h5g7 e1d3 g7e6 a2a4 f8c8 a4b5 a6b5 a1a8 c8a8 d3c5');
INSERT INTO evaluations VALUES(52576,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(52576,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(52576,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(52576,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(52576,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(52576,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(52576,36,'cp 62','b1c3 g8f6 c4d5 c6d5 c1f4 b8c6 e2e3 f6h5 f4e5 f7f6 e5g3 e7e6 f1d3 c8d7 e1g1 f8e7 e3e4 d5e4 c3e4 h5g3 h2g3 d8b6 d4d5 e6d5 e4c3 e8c8 c3d5 b6c5 d5e7 c5e7 d1c2 c8b8 f1e1 e7d6 c2c3 h8e8 d3e2 d7g4 a2a3 h7h5 a1d1 d6e7 d1d8 e8d8');
INSERT INTO evaluations VALUES(52572,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(52572,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(52572,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(52572,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(52572,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(52572,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(52572,36,'cp 55','g8f6 c4d5 c6d5 c1f4 b8c6 e2e3 f6h5 f4e5 f7f6 e5g3 e7e6 f1d3 c8d7 e3e4 d5e4 d3e4 f6f5 e4c6 d7c6 g3e5 f8g7 e5g7 h5g7 e1g1 e8g8 f3e5 d8g5 e5c6 b7c6 d1a4 f5f4 f2f3 a8b8 b2b3 g7f5 c3e4 g5e7 a1c1');
INSERT INTO evaluations VALUES(44710,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(44710,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(44710,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(44710,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(44710,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(44710,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(44710,36,'cp 65','c4d5 c6d5 c1f4 b8c6 h2h3 f6e4 e2e3 e4c3 b2c3 f8g7 f1d3 e8g8 e1g1 c6a5 a1b1 b7b6 d1e2 d8d7 f3e5 d7e6 f1c1 c8b7 h3h4 h7h5 f2f3 a8c8 e3e4 e6f6 f4g3 e7e6 d3a6 b7a6 e2a6 d5e4 f3e4 a5c6 e5c6 c8c6 a6a7');
INSERT INTO evaluations VALUES(44702,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(44702,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(44702,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(44702,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(44702,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(44702,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(44702,36,'cp 76','c6d5 c1f4 b8c6 e2e3 f6h5 f4e5 f7f6 e5g3 e7e6 f1d3 f8e7 e3e4 d5e4 d3e4 e8g8 e1g1 c6a5 g3h4 h5g7 e4b1 b7b6 a2a3 c8a6 f1e1 d8d7 b2b4 a5c4 d1e2 a8c8 b1a2 g8h8 a1d1');
INSERT INTO evaluations VALUES(44810,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(44810,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(44810,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(44810,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(44810,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(44810,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(44810,36,'cp 73','c1f4 b8c6 e2e3 f6h5 f4e5 f7f6 e5g3 e7e6 f1d3 c8d7 e3e4 d5e4 d3e4 f6f5 e4c6 d7c6 g3e5 f8g7 e5g7 h5g7 e1g1 e8g8 f3e5 d8g5 e5c6 b7c6 f1e1 f8d8 a1c1 a8b8 b2b3 d8d6 d1f3 d6d4 f3c6 b8d8 c3b5 d4d7 h2h3 g5e7 c6a6 e6e5 c1c8 d8c8 a6c8 d7d8 c8c4 g8h8 b5c3');
INSERT INTO evaluations VALUES(44808,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(44808,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(44808,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(44808,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(44808,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(44808,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(44808,36,'cp 50','b8c6 e2e3 f6h5 f4e5 f7f6 e5g3 e7e6 f1d3 f8e7 e3e4 c6b4 e4d5 e6d5 d3b5 c8d7 b5a4 d7a4 d1a4 b4c6 a2a3 d8d7 b2b4 a8c8 e1g1 c6d4 f3d4 c8c3 a4a7 e8g8 f1e1 f8e8 a7b6 c3c8 b6b5 d7b5 d4b5 h5g3 h2g3 f6f5 a1d1 e7f8 e1e8 c8e8 d1d5');
INSERT INTO evaluations VALUES(38806,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(38806,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38806,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38806,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(38806,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(38806,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(38806,36,'cp 11','g8f6 d1a4 b8c6 e2e3 e8g8 c1d2 d5c4 f1c4 b4d6 a4c2 e6e5 f3e5 c6e5 d4e5 d6e5 f2f4 e5c3 d2c3 f6g4 c3d4 c7c5 d4c5 f8e8 e1g1 g4e3 c5e3 e8e3 f1e1 d8b6 c2f2 e3e1 a1e1 b6f2 g1f2 g8f8 c4d5 c8f5 e1e5 f5d7 d5e4 h7h6 e5a5 b7b6 e4a8 b6a5 a2a3 f8e7 g2g3 e7d6 a8f3 d6c5 f2e3 d7e6');
INSERT INTO evaluations VALUES(30804,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(30804,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(30804,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(30804,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30804,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(30804,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(30804,36,'cp 31','c2c4 e7e6 g2g3 f8b4 c1d2 c7c5 f1g2 b4d2 d1d2 c5d4 d2d4 g8f6 e1g1 b8c6 d4e3 c6e7 b1c3 e8g8 f1d1 a8c8 e3d4 d7d5 f3e5 h7h6 c4d5 f6d5 c3d5 b7d5 e2e4 d5c6 e5c4 e7g6 a1c1 d8e7 c4d6 c8c7 d4e3 f8d8 e4e5 c6g2 g1g2 c7d7 b2b4 g8h8');
INSERT INTO evaluations VALUES(30791,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(30791,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(30791,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(30791,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30791,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(30791,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(30791,36,'cp 26','e7e6 g2g3 f8b4 c1d2 c7c5 f1g2 b4d2 d1d2 c5d4 d2d4 g8f6 b1c3 b8c6 d4h4 e8g8 e1g1 c6e7 a1c1 d8b8 h4d4 d7d5 c4d5 e7d5 f1d1 d5c3 d4c3 f8c8 c3e5 c8c1 d1c1 b7f3 e5b8 a8b8 g2f3 f6e8 c1d1 b8c8 d1d7 c8c2 d7a7 c2b2 a7a8');
INSERT INTO evaluations VALUES(30647,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(30647,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(30647,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30647,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(30647,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(30647,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(30647,36,'cp 37','g2g3 g8f6 f1g2 f8b4 c1d2 c7c5 d2b4 c5b4 a2a3 d8c7 b2b3 d7d5 a3b4 d5c4 e1g1 c4b3 d1b3 e8g8 b1c3 b8d7 f1c1 f8c8 b3b2 c7b8 c1c2 a7a5 a1c1 a8a7 b4b5 b8a8 d4d5 f6d5 c3d5 c8c2 d5e7 g8f8 b2c2');
INSERT INTO evaluations VALUES(30648,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(30648,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(30648,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(30648,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30648,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(30648,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(30648,36,'cp 27','f8b4 c1d2 c7c5 d2b4 c5b4 f1g2 g8f6 d1d3 e8g8 e1g1 d7d6 f1d1 a7a5 b1d2 b8d7 e2e4 e6e5 a2a3 d8c7 h2h3 f8e8 a3b4 a5b4 a1a8 e8a8 d3b3 f6e4 d2e4 b7e4 d4e5 d6e5 b3b4 d7f6 b4d6 c7d6 d1d6 e4f3 g2f3 e5e4 f3e2');
INSERT INTO evaluations VALUES(55705,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(55705,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(55705,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(55705,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(55705,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(55705,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55705,36,'cp -13','g1f3 d6e5 f3e5 b8d7 e5f3 b7b6 g2g3 g8f6 f1g2 c8b7 e1g1 f8e7 b2b3 e8g8 c1b2 c7c5 a2a4 f6d5 b1a3 e7f6 b2f6 d7f6 f3e5 d8c7 d1e2 a7a6 a3c4 a8d8 f2f4 d5b4 c2c3 b4d5 e2f2 f6d7 e5d7 d8d7 c4e5');
INSERT INTO evaluations VALUES(55704,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55704,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55704,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55704,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(55704,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(55704,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(55704,36,'cp -12','d6e5 f3e5 b8d7 e5f3 b7b6 g2g3 f8e7 f1g2 c8b7 e1g1 g8f6 b2b3 e8g8 c1b2 c7c5 d1e2 a7a6 d2d3 b6b5 c2c4 d8c7 b1d2 h7h6 f1d1 a8c8 a2a4 b5b4 f3e5 b7g2 g1g2 c8d8 e5f3 c7b6 h2h4 f8e8 d2e4 b6b7 d1e1');
INSERT INTO evaluations VALUES(55994,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(55994,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(55994,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(55994,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(55994,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(55994,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(55994,36,'cp -9','f3e5 b8d7 e5f3 b7b6 g2g3 c8b7 f1g2 f8e7 e1g1 g8f6 b2b3 e8g8 c1b2 c7c5 a2a4 f6d5 d1e2 e7f6 b2f6 d7f6 f1d1 d5b4 d2d4 c5d4 d1d4 d8e7 f3e5 a8d8 d4d8 f8d8 g2b7 e7b7 b1a3 h7h6 a1e1 b7c7 a3c4 b4c6 e5c6 c7c6 c4e5 c6c7 e5c4 f6d5 e2e5');
INSERT INTO evaluations VALUES(55980,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55980,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(55980,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55980,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(55980,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(55980,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55980,36,'cp -10','b8d7 e5f3 b7b6 g2g3 f8e7 f1g2 c8b7 e1g1 c7c5 a2a4 g8f6 d1e2 a7a6 b2b3 e8g8 c1b2 f6e8 f1d1 e7f6 b2f6 e8f6 d2d4 c5d4 d1d4 f8e8 f3e5 b7g2 g1g2 d8c7 e5d7 f6d7 b1d2 c7c2 a1d1 a8d8 e2f3 d7f6 d2c4 d8d4');
INSERT INTO evaluations VALUES(22395,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(22395,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(22395,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(22395,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(22395,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(22395,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(22395,36,'cp -21','e5f3 b7b6 g2g3 c8b7 f1g2 g8f6 e1g1 f8e7 b2b3 a7a6 a2a4 c7c5 b1a3 e8g8 c1b2 d8c7 f1e1 f6e8 a3c4 a8d8 f3e5 b7g2 g1g2 e7f6 d1e2 h7h6 a1d1 b6b5 e5d7 d8d7 c4e5 d7d5 g2g1 b5a4 b3a4 e8d6 e2a6 c5c4');
INSERT INTO evaluations VALUES(41476,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(41476,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(41476,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(41476,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(41476,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41476,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(41476,36,'cp 11','b1d2 e8g8 h2h3 c7c5 c2c3 b7b6 a2a4 b8c6 f1b5 c8b7 e1g1 f6d7 a4a5 a7a6 b5c6 b7c6 b2b4 c5b4 c3b4 b6a5 b4a5 c6b5 f1e1 a8c8 d1b3 d7b8 b3a3 b8c6 d2b3 f8e8 b3c5 g7f8 f3e5 c6e5 f4e5 f7f6 c5e6 d8d7 e6f8');
INSERT INTO evaluations VALUES(41477,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(41477,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(41477,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41477,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(41477,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(41477,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(41477,36,'cp 14','e8g8 h2h3 c7c5 c2c3 b7b6 f1d3 b8c6 f3e5 c6e5 f4e5 a7a5 d1f3 d8d7 a2a4 f8d8 e1g1 c8b7 f1e1 a8c8 f3g3 b7c6 d3c2 f6h5 g3g4 d7g4 h3g4 h5f6 g4g5 f6d7 e5g7 g8g7 f2f4 h7h5 g5h6 g7h6 d2f3 h6g7');
INSERT INTO evaluations VALUES(36760,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(36760,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36760,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36760,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(36760,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36760,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36760,36,'cp 15','h2h3 c7c5 c2c3 f6d7 f1e2 b8c6 e1g1 e7e5 f3e5 c6e5 d4e5 d7e5 d2f3 f8e8 f4e5 g7e5 e3e4 d5d4 f3e5 e8e5 c3d4 d8d4 d1d4 c5d4 e2d3 c8e6 f1c1 b7b6 f2f4 e5c5 c1c5 b6c5 b2b3 g8g7 a1c1 a8c8 g2g4 a7a5 d3b5 c8c7 f4f5 e6d7 b5d7 c7d7 c1c5 g6f5');
INSERT INTO evaluations VALUES(36757,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(36757,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36757,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36757,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36757,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36757,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(36757,36,'cp 10','c7c5 c2c3 f6d7 f1e2 b8c6 e1g1 f8e8 e3e4 d5e4 d2e4 c5d4 c3d4 d7b6 e2b5 b6d5 b5c6 b7c6 f4e5 f7f6 e5h2 d8b6 d1d2 c8f5 e4g3 e7e6 g3f5 e6f5 h2d6 b6b7 d2c2 b7d7 d6a3 g7h6 f1e1 g8f7 a1d1 a7a6 e1e8 d7e8 f3e1 a8d8 c2c4');
INSERT INTO evaluations VALUES(39581,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39581,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39581,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39581,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39581,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39581,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39581,36,'cp 0','g8f6 b1c3 d7d5 e2e3 e8g8 f1e2 c7c5 e1g1 c5d4 e3d4 b8c6 f3e5 f6d7 e5f3');
INSERT INTO evaluations VALUES(31141,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31141,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31141,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31141,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(31141,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31141,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31141,36,'cp 0','e7e6 b1d2 c6c5 h2h3 g4h5 f1e2 g8e7 d4c5 e7c6 c2c4 f8c5 e1g1 e8g8 d2b3 c5b6 c1g5 d8d7 c4d5 d7d5 d1d5 e6d5 a1d1 c6e5 g2g4 e5f3 e2f3 h5g6 f3d5 b8c6 f1e1 f8e8 g5e3 b6e3 e1e3 e8e3 f2e3 a8d8 d1d2 g8f8 g1g2 h7h5 e3e4 h5g4');
INSERT INTO evaluations VALUES(31486,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(31486,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31486,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(31486,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31486,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(31486,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(31486,36,'cp 12','c1e3 d8b6 b1d2 c6c5 c2c4 c5d4 d1a4 b6c6 a4c6 b8c6 f3d4 f8c5 d4c6 c5e3 f2e3 b7c6 c4d5 e6d5 f1d3 g8e7 a1c1 a7a5 h2h3 g4d7 e1g1 a5a4 d2f3 e8g8 f3d4 f8b8 f1f2 b8b7 f2c2 e7g6 d4c6 d7c6 c2c6 b7b2 c6c8 a8c8 c1c8 g6f8 d3h7 g8h7 c8f8 b2a2 f8f7 a2e2 f7a7 e2e3 a7a4 e3e5');
INSERT INTO evaluations VALUES(44211,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(44211,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(44211,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(44211,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(44211,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(44211,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(44211,36,'cp 55','d4d5 b7b5 d5e6 f7e6 c4b5 d7d5 a2a4 a7a6 b1c3 d5d4 c3b1 a6b5 e2e3 c8d7 e3d4 c5d4 f1e2 a8a4 a1a4 b5a4 d1d4 d8a5 c1d2 a5d5 d4b6 f8d6 e1g1 d5c5 b6c5 d6c5 f1c1 c5d6 d2e3 f6d5 e3c5 d6c5 c1c5');
INSERT INTO evaluations VALUES(44259,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(44259,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(44259,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(44259,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(44259,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(44259,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(44259,36,'cp 62','b7b5 d5e6 f7e6 c4b5 d7d5 b1c3 b8d7 e2e4 d5d4 c3a4 c8b7 f1d3 b7e4 d3e4 f6e4 e1g1 d7f6 b2b4 c5b4 f3d4 d8d5 c1e3 f8d6 d1e2 e8g8 a1d1 a8e8 f1e1 d6c7 d4f3 d5f5 e3d4 c7a5 f3e5 e4c3 a4c3 b4c3 d4c5 c3c2 c5f8 c2d1q e1d1');
INSERT INTO evaluations VALUES(43899,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43899,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43899,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(43899,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(43899,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(43899,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(43899,36,'cp 54','d5e6 f7e6 c4b5 d7d5 c1g5 b8d7 e2e3 c5c4 b1c3 f8b4 f1e2 d8b6 e1g1 b4c3 b2c3 f6e4 g5f4 e4c3 d1c2 c3e2 c2e2 e8g8 a2a4 d7c5 a4a5 b6d8 f1b1 c5d3 f4g3 c8d7 f3e5 d3c5 e2c2 d7e8 c2c3 c5b3 b1b3 c4b3 c3b3 a8b8 b5b6 a7b6 a5a6 b8a8 b3b2');
INSERT INTO evaluations VALUES(52699,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(52699,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(52699,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(52699,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(52699,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(52699,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(52699,36,'cp -88','e7e5 e2e3 a7a5 b1c3 b8d7 c3a4 g8f6 f1e2 f8e7 d1c2 e8g8 a2a3 d8c7 g1h3 f8e8 e1g1 g7g6 h3g5 c7b8 b2b4 a5b4 a3b4 e5d4 e3d4 b7b5 c5b6 d7b6 b4b5 b6a4 a1a4 a8a4 c2a4 e7d6 g5f3');
INSERT INTO evaluations VALUES(53682,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(53682,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(53682,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(53682,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(53682,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(53682,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(53682,36,'cp -75','e2e3 g8h6 b1c3 b8d7 g1e2 h6f5 e2g3 f5g3 h2g3 f8e7 b2b4 e7f6 c3e2 e8g8 c1b2 f8e8 a1c1 g7g6 d4e5 d7e5 e2d4 c8g4 d1d2 h7h5 b4b5 d8e7 b5c6 b7c6 f1e2 g4e2 d4e2 a7a5 b2d4 e5c4 d4f6 e7f6 d2d4 f6e7 e1g1 a5a4');
INSERT INTO evaluations VALUES(53108,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(53108,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(53108,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(53108,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(53108,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(53108,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(53108,36,'cp 81','g7g6 e2e4 e7e6 b1c3 f8g7 g1f3 e6d5 c4d5 g8f6 f3d2 e8g8 f1e2 b8d7 e1g1 f8e8 f1e1 h7h6 a2a4 a7a6 a1a3 g6g5 a4a5 d7e5 d2f1 a8b8 d1c2 g5g4 f1g3 b7b6 a5b6 d8b6 c3d1 h6h5 c1g5 f6h7 g5d2 h5h4 g3h5 g7h8 d2c3 h4h3 d1e3 h7g5');
INSERT INTO evaluations VALUES(52631,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(52631,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(52631,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52631,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(52631,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(52631,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(52631,36,'cp 76','e2e4 e7e6 g1f3 e6d5 c4d5 f8g7 b1c3 g8f6 f3d2 e8g8 f1e2 b8d7 e1g1 f8e8 a2a4 g6g5 f1e1 a7a6 a4a5 d7e5 d2f1 h7h6 f1g3 a8b8 d1c2 b7b5 a5b6 d8b6 a1a3 g5g4 c3d1 h6h5 c1g5 f6h7 g5d2 h5h4');
INSERT INTO evaluations VALUES(52629,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(52629,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(52629,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(52629,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(52629,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(52629,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(52629,36,'cp 72','f8g7 f1d3 e7e6 b1c3 e6d5 e4d5 d8e7 d1e2 e7e2 e1e2 b8d7 g1f3 d7e5 f3e5 g7e5 h1e1 c8f5 d3f5 g6f5 e2d3 e8d7 a1b1 g8f6 c1g5 h8g8 g5f6 e5f6 g2g3 f6c3 d3c3 a8e8 b2b4 e8e1 b1e1 f5f4 a2a4 f4g3 h2g3 b7b6 b4c5 b6c5 c3d3 g8g4');
INSERT INTO evaluations VALUES(37982,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(37982,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(37982,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(37982,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(37982,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(37982,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(37982,36,'cp 72','g1f3 e7e6 b1c3 e6d5 c4d5 g8f6 f3d2 e8g8 f1e2 f8e8 e1g1 b8d7 a2a4 g6g5 f1e1 a7a6 a4a5 h7h6 d1c2 d7e5 d2f1 a8b8 f1g3 b7b5 a5b6 d8b6 c3d1 b6b3 c2b3 b8b3 d1e3 g5g4 a1a3 b3a3 b2a3 c8d7 f2f3 d7b5 g1f2');
INSERT INTO evaluations VALUES(37981,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(37981,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(37981,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(37981,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(37981,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(37981,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(37981,36,'cp 68','c8g4 h2h3 g4f3 d1f3 b8d7 f3d3 g7d4 f1e2 e7e5 e1g1 g8f6 b1d2 e8g8 d2f3 b7b6 d3c2 g8g7 g2g3 d8e7 e2d3 h7h6 g1g2 a7a6 a2a4 f8h8 c2e2 a8b8 f1h1 b8b7 a1b1 h8b8 h3h4 f6g8 b2b3 d7f6 f3d4 c5d4');
INSERT INTO evaluations VALUES(23016,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(23016,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(23016,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(23016,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(23016,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(23016,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(23016,36,'cp 8','c7c6 b5a4 d7d5 d2d3 f8c5 b1d2 g8f6 d2b3 c5b6 b3d4 d5e4 d3e4 f6e4 e1g1 b6d4 c3d4 e8g8 f1e1 e4f6 a4b3 c8f5 h2h3 d8d7 c1g5 f6d5 d1f3 f8e8 a1c1 f5e6 f3g3 g8h8 g5d2 d5f6 b3e6 e8e6 e1e6');
INSERT INTO evaluations VALUES(19097,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(19097,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(19097,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(19097,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(19097,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19097,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19097,36,'cp 7','b5a4 d7d5 d2d3 f8c5 b1d2 g8f6 d2b3 c5b6 b3d4 b6d4 c3d4 d5e4 d3e4 f6e4 e1g1 e8g8 f1e1 c8f5 a4b3 d8d6 d4d5 e4c5 d5c6 d6c6 d1d5 c5b3 d5c6 b7c6 a2b3 a7a5 c1d2 f5e6 e1e3 a8b8 d2c3 e6b3 c3a5 f8e8 e3e8 b8e8 a1e1 e8e1 a5e1 b3d5 f2f3 d5e6 g1f2');
INSERT INTO evaluations VALUES(19115,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(19115,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(19115,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(19115,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19115,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19115,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19115,36,'cp 3','d7d5 d2d3 f8c5 b1d2 g8f6 d2b3 c5b6 b3d4 b6d4 c3d4 d5e4 d3e4 f6e4 e1g1 e8g8 f1e1 c8f5 a4b3 f8e8 c1f4 e4f6 e1e8 d8e8 f4g5 f6d5 b3d5 c6d5 d1b3 h7h6 g5f4 e8d7 h2h3 a8c8 a1e1 g8h7 e1c1 c8c1 f4c1 f5g6 c1d2 d7f5 b3b7 f5b1 g1h2 b1a2');
INSERT INTO evaluations VALUES(21965,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(21965,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(21965,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(21965,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(21965,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(21965,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(21965,36,'cp 67','d2d4 g8f6 f1d3 g7g6 h2h3 f8g7 e1g1 e8g8 f1e1 a7a5 d3c2 b7b6 b1a3 f8e8 d4d5 c6e7 a3b5 f6h5 a2a4 c8d7 a1a3 h7h6 a3b3 h5f4 f3h2 g6g5 c3c4 e7g6 b3c3 g8h8 c3g3');
INSERT INTO evaluations VALUES(21946,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(21946,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(21946,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(21946,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(21946,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(21946,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(21946,36,'cp 49','g8f6 f1d3 g7g6 e1g1 f8g7 h2h3 e8g8 f1e1 a7a5 d3c2 h7h6 c1e3 a5a4 b1d2 b7b5 b2b4 a4b3 c2b3 e5d4 c3d4 f8e8 a2a4 b5b4 b3c2 c6a5 a1b1 b4b3 d2b3 a5b3 c2b3 f6e4 d1c1 e4g5 f3g5 h6g5 c1c6 c8f5 c6f3 f5e6 b3e6 e8e6');
INSERT INTO evaluations VALUES(15554,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(15554,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(15554,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(15554,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(15554,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(15554,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(15554,36,'cp 59','f1b5 c8d7 b5d3 g7g6 h2h3 f8g7 e1g1 h7h6 c1e3 e8g8 b1d2 a7a5 a1c1 a5a4 f1e1 f8e8 d4d5 c6e7 c3c4 c7c5 b2b4 a4b3 a2b3 f6h7 b3b4 b7b6 c1a1 a8a1 d1a1 f7f5 b4c5 b6c5 g2g3 f5f4');
INSERT INTO evaluations VALUES(15419,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15419,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(15419,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(15419,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(15419,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(15419,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(15419,36,'cp 48','c8d7 b5d3 g7g6 e1g1 f8g7 d4d5 c6e7 c3c4 e8g8 f1e1 b7b6 b2b4 d7g4 a2a4 a7a5 b4a5 a8a5 h2h3 g4f3 d1f3 f6d7 f3d1 a5a8 b1c3 d7c5 d3c2 f7f5 c1e3 c5b7 a1b1 f5f4 e3c1 b7c5 c1a3');
INSERT INTO evaluations VALUES(25498,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(25498,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(25498,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(25498,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(25498,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(25498,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(25498,36,'cp 51','b5d3 g7g6 e1g1 f8g7 f1e1 d7g4 d4d5 c6e7 h2h3 g4f3 d1f3 e8g8 b1d2 f6d7 f3d1 a7a5 a2a4 g8h8 b2b4 f7f5 b4a5 a8a5 d2b3 a5a7 c1g5 g7f6 g5d2 b7b6 d1c2 f5e4 d3e4 e7g8 a4a5 f6g5 a5b6 a7a1 e1a1 d7b6 a1a7 g5d2');
INSERT INTO evaluations VALUES(25538,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25538,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(25538,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(25538,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(25538,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25538,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(25538,36,'cp 51','g7g6 e1g1 f8g7 h2h3 h7h6 c1e3 g6g5 d4e5 d6e5 f3h2 c6e7 d1c2 e7g6 f1d1 d8e7 b1d2 g5g4 h3g4 f6g4 h2g4 d7g4 f2f3 g4d7 c2b3 b7b6 d3a6 e8g8 d2f1 a8d8 a6c4 g6f4 d1d2 d7e6 c4e6 d8d2 f1d2 f7e6');
INSERT INTO evaluations VALUES(55051,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55051,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(55051,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(55051,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(55051,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(55051,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(55051,36,'cp -21','c7c5 d4c5 e7e6 c1b2 g8f6 e2e3 f8c5 a2a3 e8g8 g1f3 a7a6 b1d2 b7b5 f1e2 c8b7 c2c4 b5c4 b3c4 b8d7 e1g1 d8e7 d2b3 d5c4 b3c5 e7c5 a1c1 b7f3 g2f3 a8b8 b2d4 c5a3 e2c4 e6e5 d4a1 f8c8 g1g2 a6a5 d1e2 a5a4 f1d1 a3e7');
INSERT INTO evaluations VALUES(53259,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(53259,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(53259,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(53259,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(53259,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(53259,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(53259,36,'cp -15','d4c5 e7e6 e2e3 d8f6 d1d4 b8c6 d4f6 g8f6 g1f3 f8c5 a2a3 e8g8 c1b2 c5e7 b1d2 f6d7 c2c4 d7c5 c4d5 e6d5 f1e2 c8f5 e1g1 f8c8 f1d1 c5d3 b2c3 a7a5 h2h3 f5g6 f3d4 c6d4 c3d4 c8c2 e2f3 a8c8 f3d5 d3c1 d5b7 c1e2 g1h1 e2d4');
INSERT INTO evaluations VALUES(49256,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(49256,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49256,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(49256,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(49256,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(49256,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49256,36,'cp 35','g1f3 b8c6 e1g1 f8c5 c2c3 d7d6 b2b4 c5b6 a2a4 a7a6 b1d2 e8g8 f1e1 c8e6 h2h3 d8d7 c4b3 e6b3 d2b3 a8e8 f3h4 d6d5 a4a5 b6a7 c1e3 a7e3 e1e3 b7b6 h4f5 d5e4 d3e4 d7d1 a1d1 e8d8 d1b1 b6a5 b3a5 c6a5 b4a5 d8b8 f5e7 g8h8 b1b8 f8b8');
INSERT INTO evaluations VALUES(49253,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(49253,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49253,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49253,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(49253,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(49253,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(49253,36,'cp 19','d7d6 e1g1 f8e7 f1e1 e8g8 h2h3 a7a6 a2a4 c8e6 b1c3 e6c4 d3c4 a6a5 b2b3 b8a6 d1e2 f6d7 c3d5 c7c6 d5e7 d8e7 c1a3 a6c5 a1d1 f7f5 e4f5 f8f5 f3d2 f5f4 g2g3 f4f8 a3c5 d7c5 d2e4 c5e4 e2e4 g8h8');
INSERT INTO evaluations VALUES(23268,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(23268,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(23268,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(23268,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(23268,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(23268,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(23268,36,'cp 57','d2d3 f5e4 d3e4 g8f6 e1g1 d7d6 b5c4 c8g4 h2h3 g4h5 b1c3 d8d7 c1e3 f8e7 c4e2 h5g6 f3h4 g6f7 h4f5 e8g8 d1d2 a7a6 a1d1 a8e8 f5e7 c6e7 f2f3 f7e6 f1f2 e7g6 e2f1 f8f7 a2a4 f6h5 a4a5 h5g3 c3d5 g3f1 f2f1 g8h8 c2c4 e8f8 c4c5 c7c6');
INSERT INTO evaluations VALUES(23267,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(23267,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(23267,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(23267,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(23267,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(23267,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(23267,36,'cp 53','f5e4 d3e4 g8f6 e1g1 d7d6 b5c4 c8g4 h2h3 g4h5 b1c3 d8d7 c1e3 f8e7 c4e2 h5g6 f3h4 g6f7 h4f5 e8g8 d1d2 a7a6 a2a4 g8h8 f5e7 d7e7 a1e1 h7h6 e2d3 a6a5 c3e2 c6b4 f2f4 a8d8 e2c3 f7g8 f4f5 d6d5 c3d5 f6d5 e4d5 b4d3 d2d3 d8d5');
INSERT INTO evaluations VALUES(23263,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(23263,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(23263,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23263,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(23263,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(23263,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(23263,36,'cp 65','d3e4 g8f6 e1g1 d7d6 b5c4 c8g4 c1e3 d8d7 h2h3 g4h5 b1c3 f8e7 c4e2 h5g6 f3h4 g6f7 h4f5 e8g8 d1d2 a7a6 a2a4 g8h8 f2f3 f7e6 f5e7 d7e7 f1f2 f6h5 e2f1 e7f6 g1h2 a8e8 c3e2 f6f7 b2b3 d6d5 e4d5 e6d5');
INSERT INTO evaluations VALUES(23261,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(23261,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(23261,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(23261,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(23261,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(23261,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(23261,36,'cp 46','g8f6 e1g1 d7d6 b5c4 c8g4 c1e3 d8d7 h2h3 g4h5 b1c3 f8e7 c4e2 h5g6 f3h4 g6f7 h4f5 e8g8 d1d2 a7a6 a2a4 g8h8 f5e7 d7e7 a4a5 f7e6 e2d3 a8e8 c3e2 f6h5 b2b4 e7f7 f2f4 e5f4 e2f4 h5f4 e3f4 f7e7 a1b1 c6e5 f4e5 f8f1 d3f1 d6e5');
INSERT INTO evaluations VALUES(17098,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(17098,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(17098,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(17098,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(17098,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(17098,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(17098,36,'cp 48','e1g1 d7d6 b5c4 c8g4 h2h3 g4h5 b1c3 d8d7 c1e3 f8e7 c4e2 h5g6 f3h4 g6f7 h4f5 e8g8 d1d2 a7a6 a2a4 g8h8 f5e7 d7e7 a1d1 f7e6 e2d3 f6h5 c3e2 e7f6 f2f3 h7h6 b2b3 h5f4 e2f4 e5f4 e3f2 c6e5 d3e2 e6g8 a4a5 a8e8');
INSERT INTO evaluations VALUES(17097,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17097,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(17097,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(17097,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(17097,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(17097,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(17097,36,'cp 63','d7d6 b5c4 c8g4 h2h3 g4h5 b1c3 d8d7 c1e3 f8e7 c4e2 h5g6 f3h4 g6f7 h4f5 e8g8 d1d2 a7a6 a2a4 g8h8 f5e7 d7e7 a1e1 a8e8 f2f3 f6h5 f1f2 f7e6 c3d5 e6d5 e4d5 c6d4 e2a6 d4f3 f2f3 b7a6 f3f8 e8f8 d2e2 e7f7 e2a6 f7d5 a6e2 h5g3 e2g4 g3f5 e3f2');
INSERT INTO evaluations VALUES(15880,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15880,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(15880,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(15880,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(15880,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(15880,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(15880,36,'cp 82','b5c4 c8g4 h2h3 g4h5 b1c3 d8d7 c3d5 e8c8 d1d3 c8b8 b2b4 h5f3 d5f6 g7f6 d3f3 c6d4 f3f6 f8e7 f6h6 d7c6 c4d5 c6e8 b4b5 d4b5 c1e3 b5d4 a1b1 c7c6 d5c4 e8f8 h6f8 h8f8 e3d4 e5d4 f2f4 e7f6');
INSERT INTO evaluations VALUES(34833,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(34833,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(34833,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(34833,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(34833,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(34833,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(34833,36,'cp 93','d2d4 d5d8 c2c4 c7c5 b1c3 g8f6 c1e3 f6g4 e3g5 f8e7 g5e7 d8e7 f1e2 c5d4 f3d4 g4f6 e2f3 a7a6 d1e2 e6e5 d4c2 e8g8 e1g1 b8c6 a1d1 f8e8 c3d5 e7d6 f1e1 f6d5 c4d5 c6b4 c2b4 d6b4 a2a3 b4b6 d5d6 c8d7 b2b4');
INSERT INTO evaluations VALUES(34825,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(34825,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(34825,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(34825,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34825,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(34825,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(34825,36,'cp 87','d5d8 c2c4 c7c5 b1c3 g8f6 c1e3 f6g4 e3g5 f7f6 g5h4 c5d4 f3d4 g4e5 h4g3 f8c5 d4b3 d8d1 a1d1 c5b4 f1e2 b8c6 e1g1 b4c3 b2c3 c8d7 f2f4 e5f7 e2h5 g7g6 h5f3 e8c8 b3d4 f7d6 d1b1 h8e8 c4c5 d6f5 g3f2 f5d4 c3d4');
INSERT INTO evaluations VALUES(56042,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(56042,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(56042,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(56042,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(56042,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(56042,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(56042,36,'cp 78','c2c4 g8f6 b1c3 c7c5 c1e3 c5d4 f3d4 a7a6 g2g3 e6e5 d4e2 d8c7 a1c1 f8c5 c3d5 f6d5 c4d5 c7a5 d1d2 a5d2 e1d2 c5e3 d2e3 e8g8 f1g2 b8d7 d5d6 f8d8 h1d1 d7f6 e2c3 c8g4 d1d2 f6e8 h2h3 g4e6 c1d1 a8b8 c3e4 b7b6 d6d7 e8c7 b2b3 g8f8 e4g5 b6b5 g2e4 b8b6 g5e6 c7e6 e4h7');
INSERT INTO evaluations VALUES(35273,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(35273,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(35273,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(35273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35273,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(35273,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35273,36,'cp -5','f1c4 b8c6 b1c3 c6d4 h5d1 g8f6 d2d3 c7c6 a2a3 e7d8 c4a2 f8c5 c3e2 d4e2 d1e2 e8g8 f2f4 e5f4 c1f4 d7d5 e1c1 d5e4 d3e4 d8e7 g1f3 e7e4 e2e4 f6e4 h1e1 e4f6 f4d6 c5d6 d1d6');
INSERT INTO evaluations VALUES(35268,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(35268,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(35268,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(35268,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(35268,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(35268,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35268,36,'cp -12','b8c6 b1c3 c6d4 h5d1 c7c6 g1f3 g8f6 d2d3 b7b5 c4b3 a7a5 a2a3 d4b3 c2b3 d7d6 d3d4 e7c7 c1g5 f8e7 e1g1 c7b8 a3a4 b5b4 d4e5 d6e5 c3b1 c8a6 f1e1 e8g8 b1d2 f8d8 a1c1 h7h6 g5f6 e7f6 c1c6 b8b7 c6c5 f6e7');
INSERT INTO evaluations VALUES(2048,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2048,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2048,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2048,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2048,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2048,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2048,36,'cp 0','b1c3 c6d4 h5d1 g8f6 g1f3 c7c6 d2d3 b7b5 c4b3 a7a5 a2a3 d4b3 c2b3 d7d6 d3d4 e7c7 c1g5 f8e7 e1g1 e5d4 f3d4 e8g8 f1e1 f8e8 d1c2 c7b6 a1d1 c8g4 d1d2 h7h6 g5h4 b6b7 h2h3 g4d7 d4f3 d7e6 f3d4');
INSERT INTO evaluations VALUES(2044,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(2044,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2044,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2044,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2044,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(2044,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2044,36,'cp -7','c6d4 h5d1 c7c6 g1f3 d7d6 h2h3 b7b5 c4b3 a7a5 a2a3 g7g6 e1g1 f8g7 f3d4 e5d4 c3e2 e7e4 d2d3 e4h4 f1e1 g8e7 c1f4 d6d5 d1d2 h7h6 f4d6 c8e6 e2f4 e8d7 g2g3 h4g5 d2e2 g5f6 d6c5 h8e8');
INSERT INTO evaluations VALUES(47763,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(47763,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(47763,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(47763,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(47763,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(47763,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47763,36,'cp 23','c7c6 h2h3 b8d7 g2g4 d7c5 f1g2 c6d5 e4d5 f8e7 c1e3 e8g8 g1e2 f6e8 e2g3 c8d7 d1d2 a8c8 g2f1 e7h4 e1c1 a7a6 c1b1 h4g3 f2g3 f7f5 e3c5 c8c5 g4f5 d7f5 g3g4 f5g6 f1d3 g6d3 d2d3 b7b5 h1f1 e8f6 a2a3 h7h6 c3e4 f6e4 f1f8 d8f8 d3e4');
INSERT INTO evaluations VALUES(45849,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(45849,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(45849,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45849,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45849,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45849,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45849,36,'cp 27','h2h3 c6d5 e4d5 f8e7 g2g4 e8g8 g1e2 b7b5 a2a3 a7a6 f1g2 b8d7 e1g1 h7h6 e2g3 d7b6 d1d3 f8e8 f2f4 f6d7 d3f3 e5f4 c1f4 b6c4 c3e4 e7h4 f3f2 f7f5 g4f5 d7f6 e4f6 d8f6 b2b3 c4e5 g2e4 c8d7 a1d1 a8c8 g1h1 g8h8');
INSERT INTO evaluations VALUES(31844,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(31844,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(31844,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(31844,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(31844,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(31844,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(31844,36,'cp -13','e7e6 c1f4 b8d7 e2e3 f8d6 f4d6 c7d6 f1d3 f5g6 e1g1 g8f6 a2a4 e8g8 d1b1 a7a6 f1c1 f6e4 b1a2 e4c3 b2c3 a8c8 c3c4 d5c4 d3c4 d8a5 c4f1 c8c3 a2b2 a5c7 a1b1 f8b8 f1d3 d6d5 h2h3 c3c6 g1h1 f7f6 f3h4 g6e4 h4f3 e4f3 g2f3');
INSERT INTO evaluations VALUES(32309,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(32309,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(32309,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(32309,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32309,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32309,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32309,36,'cp -10','c1f4 b8d7 e2e3 f8d6 f4d6 c7d6 f1d3 f5g6 e1g1 g8f6 f3h4 g6e4 a2a4 a7a6 f2f4 h7h6 h4f3 e4f5 f3d2 d8b6 a1b1 e8g8 h2h3 f5d3 c2d3 e6e5 d2f3 a8e8 d1d2 e5e4 f3e1 e8c8 g2g4 e4d3 e1d3 f6e4');
INSERT INTO evaluations VALUES(32288,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(32288,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(32288,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(32288,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(32288,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(32288,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(32288,36,'cp -12','b8d7 e2e3 f8d6 f4d6 c7d6 a2a4 g8f6 f1d3 f5g6 e1g1 e8g8 d1d2 f6e4 d2e1 a8c8 c3e2 a7a6 a4a5 c8c7 e2f4 d7f6 h2h3 d8e7 f4g6 h7g6 e1e2 f8c8 a1e1 c8e8 c2c4 d5c4 d3c4 d6d5 c4d3 e4d6 e1c1');
INSERT INTO evaluations VALUES(49358,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49358,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49358,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49358,36,'cp 0','c7c5 d4d5 d8a5 g5d2 a5d8');
INSERT INTO evaluations VALUES(44171,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44171,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44171,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44171,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44171,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44171,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44171,36,'cp 0','d4d5 d8a5 g5d2 a5d8');
INSERT INTO evaluations VALUES(24470,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(24470,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24470,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24470,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(24470,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(24470,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24470,36,'cp 19','c1g5 e7e6 d1d3 c6d4 b5d7 d8d7 d3d4 f8e7 e1c1 a8c8 g5f6 e7f6 d4d6 d7d6 d1d6 f6c3 b2c3 c8c3 h1d1 e8g8 d6d7 b7b6 d7a7 f8c8 d1d2 g7g5 c1b2 c3c4 f2f3 g8g7 a7b7 c4b4 b2c1 g7g6 a2a3 b4b5 b7d7 b5a5 d7d3 h7h5 c1b2 a5c5 d3d6 c5b5 b2a2 c8c4 d6d4 c4c3 d4b4');
INSERT INTO evaluations VALUES(24467,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24467,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24467,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(24467,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24467,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(24467,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(24467,36,'cp 29','e7e6 d1d3 c6d4 b5d7 d8d7 d3d4 f8e7 e1c1 a8c8 g5f6 e7f6 d4d6 d7d6 d1d6 f6c3 b2c3 c8c3 h1d1 e8g8 d6d7 b7b6 d7a7 f8c8 d1d2 g7g5 a7a4 b6b5 a4b4 c3c5 c1b2 g8g7 c2c3 c5c3 b4b5 c3c4 b5g5 g7f6 g5b5 c4e4 a2a3 e4h4 h2h3 h4c4 b5b4 c4c5 b4f4 f6g6 a3a4 g6g7 f4b4 h7h5 b4b3 c5c4 a4a5 c4a4');
INSERT INTO evaluations VALUES(24439,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24439,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(24439,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24439,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(24439,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24439,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24439,36,'cp 32','d1d3 c6d4 b5d7 d8d7 d3d4 f8e7 e1c1 a8c8 g5f6 e7f6 d4d6 d7d6 d1d6 f6c3 b2c3 c8c3 h1d1 e8g8 d6d7 b7b6 d7a7 f8c8 d1d2 g7g5 a7a4 g8g7 a4b4 c3c6 b4b5 g7g6 f2f3 h7h5 c1b2 c8c7 a2a3 c7c8 b5b3 g6f6 g2g3 f6g6 b3b4 g5g4 f3g4');
INSERT INTO evaluations VALUES(38621,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(38621,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(38621,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(38621,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38621,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38621,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(38621,36,'cp 7','g8e7 e1g1 e8g8 c2c4 d5c4 d3c4 h7h6 b1c3 b8d7 c1d2 c7c6 c3e4 d7b6 c4b3 b6d5 h2h3 e7g6 b3d5 c6d5 e4d6 d8d6 d1b3 f8e8 f1e1 c8e6 f3e5 g6e5 e1e5 f7f6 e5e1 b7b6 b3f3 e6d7 e1e8 a8e8 d2c3 d7c6 f3d3 e8e4');
INSERT INTO evaluations VALUES(41145,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(41145,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41145,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(41145,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41145,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41145,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(41145,36,'cp 17','e1g1 e8g8 c1g5 f7f6 g5h4 b8c6 c2c3 c8f5 h4g3 d8d7 f1e1 a8e8 a2a4 f5d3 d1d3 c6d8 g3d6 d7d6 b1d2 d6d7 d2b3 b7b6 b3d2 e7c8 a4a5 c8d6 e1e8 f8e8 a5b6 a7b6 d2f1 d8e6 g2g3 c7c6 h2h4 d6c4 b2b3 c4d6 f1e3 d6e4 c3c4 b6b5 c4b5 c6b5');
INSERT INTO evaluations VALUES(41144,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41144,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41144,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(41144,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(41144,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(41144,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41144,36,'cp 18','e8g8 c1g5 b8c6 c2c3 f7f6 g5h4 c8f5 h4g3 d8d7 f1e1 a8e8 a2a4 f5d3 d1d3 c6d8 g3d6 d7d6 b1d2 d6d7 d2b3 b7b6 b3d2 d8b7 h2h4 c7c6 h4h5 b7d6 d3a6 e7c8 a4a5 d7g4 h5h6 g7h6 d2f1 g8h8 e1e8 f8e8 f1e3');
INSERT INTO evaluations VALUES(36431,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36431,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36431,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36431,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(36431,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36431,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(36431,36,'cp 8','c2c4 d5c4 d3c4 b8c6 h2h3 c6a5 c4d3 c8f5 b1c3 f5d3 d1d3 a5c6 c1d2 c6b4 d3c4 b4d5 c4b3 d5c3 b2c3 b7b6 a2a4 h7h6 c3c4 e7c6 d4d5 c6e5 f3d4 d8f6 d4b5 f8e8 b5d6 c7d6 a4a5 e5d7 a1e1 e8e1 f1e1');
INSERT INTO evaluations VALUES(47874,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47874,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47874,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47874,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47874,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(47874,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(47874,36,'cp 0','c7c5 g2g3 b8c6 f1g2 g7g6 e1g1 f8g7 c1g5 h7h6 g5f6 g7f6 f3d2 e8g8 a2a4 f6g7 d2c4 a8b8 c3d5 c8e6 a4a5 h6h5 d1e2 c6d4 e2d1');
INSERT INTO evaluations VALUES(45912,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45912,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45912,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(45912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45912,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(45912,36,'cp 0','g2g3 b8c6 f1g2 g7g6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 h6h5 e1g1 f8h6 d1e1 e8g8 f3h4 h6g5 h4f3');
INSERT INTO evaluations VALUES(52791,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(52791,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52791,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52791,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52791,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(52791,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52791,36,'cp 22','d5e4 d2e4 c8f5 e4g3 f5g6 h2h4 h7h6 g1f3 b8d7 f1d3 g6d3 d1d3 e7e6 c1f4 d8a5 f4d2 a5c7 g3e4 g8f6 e1c1 a8d8 h1g1 f6g4 g2g3 f8e7 d3e2 e8g8 d2f4 c7a5 c1b1 a5b5 d1d3 d7f6 e4c3 b5a6 f3e5 g4e5 f4e5 c6c5 g3g4 f6d7 g4g5 h6g5 h4g5 d7e5 e2e5 d8d4 d3d4 c5d4 c3d5 e6d5 e5e7');
INSERT INTO evaluations VALUES(52905,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52905,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52905,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(52905,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52905,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(52905,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(52905,36,'cp 26','d2e4 c8f5 e4g3 f5g6 h2h4 h7h6 g1f3 b8d7 f1d3 g6d3 d1d3 e7e6 c1f4 d8a5 c2c3 f8e7 a2a4 g8f6 b2b4 a5d8 a4a5 f6d5 f4d2 e8g8 g3h5 e7d6 e1g1 b7b5 a1e1 d8c7 d3e2 d7f6 h5f6 d5f6 d2c1 a7a6 c1d2 a8e8 g2g3 f6d5');
INSERT INTO evaluations VALUES(55401,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(55401,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(55401,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(55401,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(55401,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(55401,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(55401,36,'cp 46','g8f6 d2d4 b8d7 e1g1 f8e7 c4d3 e8g8 f1e1 f8e8 h2h3 a7a5 c1e3 d6d5 f3e5 f6e4 b1d2 e7d6 d2e4 d5e4 d3e4 d7e5 d4e5 d6e5 d1c2 d8f6 a1d1 g8h8 a2a4 c7c6 e3b6 c8e6 e1e3 e5f4 e3d3 e6c4 d3d7 f6e5 d1e1 c4d5');
INSERT INTO evaluations VALUES(47615,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(47615,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(47615,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(47615,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(47615,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(47615,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(47615,36,'cp 57','d2d4 b8c6 e1g1 f8e7 f1e1 c8g4 c4b5 f6d7 h2h3 g4h5 c1e3 e7g5 b5c6 b7c6 b1d2 g5e3 e1e3 e8g8 d1a4 e5d4 c3d4 c6c5 a4a5 f8e8 b2b3 h5f3 d2f3 d8b8 a5c3 b8b6 a1d1 a7a5 a2a4 b6b4 c3d3 c5d4 f3d4 d7c5 d3e2 a8c8 e4e5 e8e5 e3e5');
INSERT INTO evaluations VALUES(47631,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47631,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(47631,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(47631,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(47631,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(47631,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(47631,36,'cp 56','b8c6 e1g1 f8e7 f1e1 c8g4 c4b5 f6d7 h2h3 g4h5 c1e3 e7g5 b5c6 b7c6 b1d2 g5e3 e1e3 e8g8 d1a4 e5d4 c3d4 c6c5 a4a5 f8e8 b2b3 h5f3 d2f3 d7f6 d4c5 e8e4 a1e1 e4e3 e1e3 d6c5 a5c5 d8d6 c5d6 c7d6 e3d3 a8d8 f3d4 d8d7 f2f3 g7g6 d4c6 g8g7');
INSERT INTO evaluations VALUES(14647,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(14647,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(14647,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14647,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(14647,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(14647,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(14647,36,'cp 72','d1b3 d8e7 e1g1 g7g5 d4e5 d6e5 b1d2 f8g7 b3c2 f6h5 g2g3 e8g8 b2b4 e7f6 a2a4 f8d8 f1e1 a7a6 h2h4 g5g4 f3h2 h5f4 d2b3 f4h3 g1g2 h3f4 g2f1 f4h3 c1e3 b7b6 a1d1 h6h5 d1d8 c6d8 a4a5 b6a5');
INSERT INTO evaluations VALUES(16430,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(16430,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(16430,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(16430,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(16430,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(16430,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(16430,36,'cp 8','g1f3 f8c5 f3e5 c6e5 d2d4 c5b4 d4e5 f6e4 d1d4 e4c3 b2c3 b4e7 f1d3 d7d6 c1f4 e8g8 e1g1 d6e5 f4e5 e7f6 a1e1 d8d4 c3d4 f6e5 e1e5 f8d8 d4d5 g8f8 c4c5 b7b6 f1c1 c7c6 d3e4 c6d5 e4d5 a8b8 c5c6 f7f6 e5h5');
INSERT INTO evaluations VALUES(57120,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(57120,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(57120,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(57120,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(57120,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(57120,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(57120,36,'cp -67','d7d5 c4b3 g8f6 d2d4 b8d7 c2c4 d5c4 b3c4 e5d4 e3d4 d7b6 c4b3 f8b4 c1d2 c8g4 g1e2 a7a5 a2a3 b4d6 b1c3 e8g8 f2f3 g4f5 e1g1 f8e8 d2g5 a5a4 b3a2 f5e6 g5f6 d8f6 c3e4 f6h4 e4d6 c7d6 a2e6 e8e6 e2c3 a8e8 c3a4 b6c4 g1h1');
INSERT INTO evaluations VALUES(56339,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(56339,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(56339,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(56339,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(56339,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(56339,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(56339,36,'cp -45','c4b3 g8f6 d2d3 c7c6 g1f3 f8d6 b1c3 a7a5 a2a4 c8e6 e1g1 h7h6 e3e4 e8g8 d3d4 f6e4 d4e5 e4c3 b2c3 d6c7 f1e1 b8d7 c1f4 d7c5 f3d4 d8d7 d1d2 a8e8 a1d1 f7f6 e5f6 c7f4 d2f4 f8f6 f4d2 e6f7 e1e8 d7e8');
INSERT INTO evaluations VALUES(56421,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(56421,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(56421,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(56421,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(56421,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(56421,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(56421,36,'cp -78','g8f6 d2d3 h7h6 g1f3 b8d7 b1c3 e5e4 f3d4 d7c5 e1g1 c7c6 d3e4 d5e4 f2f4 e4f3 d1f3 f8e7 d4f5 c8f5 f3f5 e8g8 e3e4 d8d7 e4e5 d7f5 f1f5 c5b3 a2b3 e7c5 g1f1 f6d7 c1d2 g7g6 e5e6 f7e6 f5f8 c5f8 d2e3 a7a5 c3e4 h6h5');
INSERT INTO evaluations VALUES(48852,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(48852,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(48852,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(48852,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(48852,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(48852,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(48852,36,'cp -46','d2d3 a7a5 b1c3 c7c6 g1f3 f8d6 a2a4 h7h6 e3e4 c8e6 e1g1 e8g8 d3d4 f6e4 d4e5 e4c3 b2c3 d6c7 f1e1 b8d7 c1f4 d7c5 f3d4 d8d7 d1d2 e6f5 a1d1 a8d8 f2f3 f5g6 f4g3 b7b5 a4b5 c6b5 d2e2 a5a4 b3a2 d8b8 d1b1 c5e6');
INSERT INTO evaluations VALUES(34538,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34538,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(34538,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(34538,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(34538,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(34538,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(34538,36,'cp 71','g8f6 g1f3 c7c6 f1d3 c8g4 c1e3 g4h5 h2h3 b8d7 d1e2 e7e6 e1c1 h5g6 d3g6 h7g6 h3h4 b7b5 g2g3 f8e7 c1b1 f6d5 c3e4 d6c7 e3d2 a7a5 d1e1 a5a4 e4g5 e7g5 f3g5 d7b6 a2a3 b6c4 b1a1 c4d2 e2d2 d5f6 d2e2');
INSERT INTO evaluations VALUES(33659,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(33659,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(33659,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(33659,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(33659,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(33659,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(33659,36,'cp 79','g1f3 c7c6 f3e5 g7g6 c1f4 f6h5 f4e3 h5f6 d1d2 c8e6 f1e2 f8g7 e1g1 b8d7 e5f3 d7b6 e3f4 d6d8 a2a4 e8g8 h2h3 f6d5 f4e5 d5c3 d2c3 e6d5 e5g7 g8g7 a4a5 d5f3 e2f3 b6d5 c3d2 d8d6 c2c4 d5b4');
INSERT INTO evaluations VALUES(49008,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(49008,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(49008,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(49008,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(49008,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(49008,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(49008,36,'cp 150','f3e5 c8e6 e5f3 f8d6 d2d4 e8g8 c4b3 b8c6 e1g1 c6a5 c1g5 d8d7 c2c4 d5f4 d4d5 e6g4 c4c5 a5b3 a2b3 h7h6 g5f4 d6f4 a1a4 g4f3 d1f3 f4e5 b1c3 e5c3 b2c3 d7b5 a4b4 b5c5 b4b7 a7a5 f1a1 a8b8 b7b8 f8b8 c3c4 c5d4 a1a2 d4b6 h2h3 g7g6 a2a1 b6b3 f3b3 b8b3');
INSERT INTO evaluations VALUES(48992,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(48992,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(48992,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(48992,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(48992,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(48992,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(48992,36,'cp 151','c8e6 b1c3 b8d7 d1h5 g7g6 h5e2 d5f4 e2e4 d7e5 e4e5 d8d6 e5e4 e8c8 d2d3 f4d5 c3d5 e6d5 c4d5 d6d5 e4d5 d8d5 c1e3 f8d6 e1d2 h8e8 g2g3 h7h5 h2h4 b7b6 c2c3 c8d7 a2a3 d5f5 a3a4 c7c5 a1d1 f5e5 d2e2 e5e7 d3d4 c5c4 e2f3 f7f6 h1e1 d7c6');
INSERT INTO evaluations VALUES(30006,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(30006,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(30006,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(30006,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(30006,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(30006,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(30006,36,'cp 155','b1c3 b8d7 d1h5 g7g6 h5e2 d5f4 e2e4 d7e5 e4e5 d8d6 e5e4 e8c8 d2d3 f4d5 c4d5 e6d5 c3d5 d6d5 e4d5 d8d5 c1e3 f8d6 e1d2 h8e8 g2g3 c7c5 b2b3 h7h5 a2a4 b7b6 a4a5 b6b5 c2c4 b5c4 b3c4 d5f5 a1a4 c8d7 h1a1 a7a6 d2e2');
INSERT INTO evaluations VALUES(21422,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(21422,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(21422,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(21422,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(21422,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21422,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21422,36,'cp 54','a2a3 g8f6 b1d2 c6b8 c2c4 f8e7 f1d3 e8g8 e1g1 a7a5 b2b3 b7b6 c1b2 c8b7 c4d5 e6d5 a1c1 e7d6 f3e5 c7c5 f1e1 d8e7 d1c2 f8e8 e3e4 f6e4 d2e4 d5e4 d3e4 b7e4 e1e4 b8d7 e4e3 a8c8 c2f5 d7e5 d4e5');
INSERT INTO evaluations VALUES(21430,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(21430,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(21430,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(21430,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(21430,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(21430,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(21430,36,'cp 49','g8f6 c2c4 d5c4 b1d2 e6e5 d4e5 f6g4 d1a4 c8d7 e5e6 d7e6 f1c4 e6d7 a4b3 g4e5 c4e2 g7g6 e1g1 f8g7 f3e5 c6e5 d2f3 a8b8 c1d2 e5f3 e2f3 e8g8 d2b4 f8e8 a1d1 d8c8 b4c3 d7e6 b3c2 g7c3 c2c3');
INSERT INTO evaluations VALUES(14867,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(14867,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(14867,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(14867,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(14867,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(14867,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(14867,36,'cp 66','b2b4 c6b8 c2c4 a7a5 b4b5 c7c5 c1b2 b8d7 b1d2 f8e7 f1e2 c5d4 f3d4 d7c5 c4d5 e6d5 a3a4 e8g8 e1g1 f8e8 h2h3 e7d6 b2a3 d8e7 d2b3 f6e4 a3c5 e4c5 b3c5 d6c5 a1c1 c5b6 e2f3 c8e6 d4e6');
INSERT INTO evaluations VALUES(31155,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(31155,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(31155,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(31155,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31155,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31155,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31155,36,'cp 0','e7e6 g1f3 c6c5 c2c4 d5c4 b1c3 b8c6 d4d5 e6d5 d1d5 c6d4 d5b7 a8b8 b7a7 f8e7 f1c4 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(31515,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31515,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31515,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31515,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31515,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31515,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31515,36,'cp 0','g1f3 c6c5 c2c4 d5c4 b1c3 b8c6 d4d5 e6d5 d1d5 c6d4 d5b7 f8e7 e1f2 g8h6 c3d5 e7h4 f3h4 d8h4 g2g3 h4d8 d5c7 e8f8 b7a8 h6g4 f2g1 d8a8 c7a8 d4f3 g1g2 f3e1 g2g1');
INSERT INTO evaluations VALUES(31514,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31514,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31514,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31514,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31514,36,'cp 0','c6c5 c2c4 d5c4 b1c3 b8c6 d4d5 e6d5 d1d5 c6b4 d5b7 f8e7 e1f2 e8f8 c1e3 g8h6 h2h3 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(31660,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31660,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31660,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31660,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31660,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31660,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31660,36,'cp 0','c2c4 d5c4 b1c3 b8c6 d4d5 e6d5 d1d5 g8e7 d5c4 c6d4 c1e3 f5e6 c4a4 e6d7 a4c4');
INSERT INTO evaluations VALUES(55544,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(55544,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(55544,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(55544,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(55544,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(55544,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(55544,36,'cp -22','g8f6 g1e2 g7g6 d2d4 e5d4 e2d4 f8g7 b1c3 e8g8 c1g5 h7h6 g5h4 b8d7 e1c1 f8e8 h1e1 c7c6 c4b3 d7c5 h2h3 g6g5 h4g3 g5g4 f3f4 c5b3 a2b3 f6h5 f4e3 d6d5 c1b1 g4h3 g2h3 d8g5 g3d6 g5e3 e1e3 d5e4 c3e4 a7a5 d6c5 g8h8 e4d6 e8e3 f2e3 c8h3 d1g1 h8h7');
INSERT INTO evaluations VALUES(47813,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(47813,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(47813,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(47813,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(47813,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(47813,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(47813,36,'cp 0','g1e2 g7g6 b1c3 f8g7 h2h4 b8c6 h4h5 c8g4 f3g3 g4e2 c4e2 c6d4 h5h6 g7f8 e2d1 c7c6 d2d3 f8e7 c3e2 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(31039,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(31039,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31039,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31039,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(31039,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(31039,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31039,36,'cp 19','h7h6 g1f3 b8d7 f1d3 g6d3 d1d3 e7e6 c1f4 g8f6 e1c1 f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e2 d8d5 f3e5 d5e4 e2e4 f6e4 e5d3 f8d8 c2c3 e4f6 b1c2 f6d5 f4e3 b7b6 g2g3 d5e3 f2e3 a8c8 e3e4 c6c5 d4c5 b6c5 d3e5 c5c4 d1d8 e7d8');
INSERT INTO evaluations VALUES(30966,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(30966,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(30966,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(30966,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(30966,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(30966,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(30966,36,'cp 20','g1f3 b8d7 c1f4 e7e6 f1d3 g6d3 d1d3 d8a5 f4d2 f8b4 c2c3 b4e7 c3c4 a5c7 g3h5 e8c8 e1g1 g7g5 f1e1 g8f6 h5f6 e7f6 d3e4 h8g8 a1d1 g5g4 d2f4 c7b6 f3e5 d7e5 f4e5 f6e5 e4e5 d8d7 b2b3 b6c7 g2g3 c7e5 e1e5 b7b6 e5h5');
INSERT INTO evaluations VALUES(30965,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(30965,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(30965,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(30965,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(30965,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(30965,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30965,36,'cp 30','b8d7 f1d3 g6d3 d1d3 e7e6 c1f4 g8f6 e1c1 f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e2 d8d5 f3e5 d5e4 e2e4 f6e4 f4e3 f8d8 c2c3 e4f6 b1c2 f6d5 e5d3 h6h5 a2a4 a7a5 d1e1 a8c8 e1e2 d5e3 e2e3 c6c5 d4c5 e7c5 d3c5 c8c5 h1d1');
INSERT INTO evaluations VALUES(44208,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(44208,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(44208,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(44208,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(44208,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(44208,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(44208,36,'cp 91','d4d5 d7d6 e2e4 g7g6 h2h3 f8g7 g1f3 e6d5 e4d5 e8g8 f1d3 f8e8 c1e3 b8d7 e1g1 a7a6 e3f4 f6h5 f4g5 f7f6 g5h4 h5f4 d3c2 b7b5 h4g3 g7h6 c4b5 a6b5 g3f4 h6f4 c3b5 d7e5 a2a4 c8a6 c2d3 f6f5 f3e5 f4e5 f1e1');
INSERT INTO evaluations VALUES(16552,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(16552,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16552,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(16552,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(16552,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(16552,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(16552,36,'cp -24','d7d5 e4d5 f6d5 f1g2 c8g4 h2h3 g4e6 f3g5 e6f5 b1c3 d5c3 b2c3 d8d7 g3g4 f5g6 g5e4 h7h5 e4g3 h5h4 g3e4 f8e7 e1g1 e8c8 g1h2 d8f8 c1e3 f7f5 g4f5 g6h5 g2f3 h5f3 d1f3 f8f5 f3g2 g7g5 f1g1 c8b8 a1b1 h8f8 c3c4 c6d4 e4g5');
INSERT INTO evaluations VALUES(15158,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(15158,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(15158,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(15158,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(15158,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(15158,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15158,36,'cp 0','e4d5 f6d5 f1g2 f8c5 b1c3 d5c3 b2c3 e8g8 e1g1 c5b6 f1e1 f8e8 c1e3 b6a5 e3d2 a5b6');
INSERT INTO evaluations VALUES(15013,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15013,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15013,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15013,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15013,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15013,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(15013,36,'cp -11','f6d5 f1g2 f8c5 e1g1 e8g8 b1c3 d5c3 b2c3 f7f6 f1e1 c8g4 h2h3 g4h5 c1e3 c5e3 e1e3 d8d6 d3d4 a8d8 d1d3 e5d4 f3d4 h5g6 d3c4 g6f7 c4a4 c6d4 c3d4 d6d4 a4d4 d8d4 g2b7 f8b8 b7g2 g8f8 e3a3 b8b2 a3a7 b2c2 g2f1 f8e7');
INSERT INTO evaluations VALUES(15358,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15358,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(15358,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15358,36,'cp 0','f1g2 f8c5 e1g1 e8g8 b1c3 d5c3 b2c3 c8g4 h2h3 g4h5 a2a4 f7f5 d3d4 c5e7 d4e5 d8d1 f1d1 c6e5 f3e5 h5d1 g2b7 a8d8 e5c6 d8d7 c6e5');
INSERT INTO evaluations VALUES(15359,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(15359,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15359,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15359,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(15359,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(15359,36,'cp -10','f8c5 e1g1 e8g8 f1e1 c8g4 h2h3 g4h5 b1c3 d5c3 b2c3 f8e8 g3g4 h5g6 d3d4 e5d4 e1e8 d8e8 c3d4 a8d8 c2c3 c5d6 f3h4 g6e4 c1g5 d6e7 g5e7 e8e7 g2e4 e7e4 d1b1 e4b1 a1b1 b7b6 h4f5 f7f6 b1d1 g8f7 g1g2');
INSERT INTO evaluations VALUES(44348,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(44348,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(44348,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(44348,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(44348,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(44348,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(44348,36,'cp 91','c4d5 d7d6 e2e4 g7g6 f2f4 f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5e2 e8g8 c1e3 d7f6 f3d2 a8b8 e3f2 b7b5 f2h4 d8e8 a4b5 a6b5 h2h3 c5c4 g1h1 b4d3 e2d3 c4d3 d1f3 f6d7 a1a7 b5b4 c3a4 b8b7 a7b7 c8b7 f3d3 f7f5');
INSERT INTO evaluations VALUES(22890,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(22890,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(22890,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(22890,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(22890,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(22890,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(22890,36,'cp -56','c3b2 c1b2 f8b4 b1c3 g8f6 e1g1 b4c3 b2c3 e8g8 e4e5 f6e4 c3b2 e4c5 d1c2 b7b6 a1d1 d8e7 b2c1 c5e6 c2e4 c8b7 e4g4 a8d8 f1e1 c6a5 c4d3 f7f5 d3f5 b7f3 g4f3 g7g6 f5e6 d7e6 f3h3 a5c4 d1d8 f8d8');
INSERT INTO evaluations VALUES(22893,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(22893,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(22893,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(22893,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(22893,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(22893,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(22893,36,'cp -37','c1b2 f8b4 b1c3 g8f6 e1g1 b4c3 b2c3 e8g8 e4e5 f6e4 c3b2 d7d6 f1e1 c8f5 c4d5 e4c5 f3d4 f5g6 d4c6 b7c6 d5c6 c5d3 c6a8 d8a8 d1d2 d3e1 a1e1 d6e5 b2e5 c7c5 h2h3 f7f6 e5c7 a8c6 e1e7 f8c8 c7a5 a7a6 d2d7 c6d7 e7d7 g6e8 d7a7 c5c4 a5c3 e8b5 a2a4 b5a4 a7a6 a4b5 a6a3 c8d8 f2f4 d8d5 a3a8 g8f7');
INSERT INTO evaluations VALUES(22892,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(22892,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(22892,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(22892,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(22892,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(22892,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(22892,36,'cp -47','f8b4 b1c3 g8f6 e1g1 b4c3 b2c3 e8g8 e4e5 f6e4 c3b2 d7d6 f1e1 e4c5 e5d6 c7d6 a1c1 c8e6 f3d4 c6d4 d1d4 d8g5 f2f4 g5f6 d4f6 g7f6 e1e3 f8c8 c4e6 f7e6 b2f6 c8c7 c1d1 c7f7 f6b2 f7d7 b2a3 a8d8 e3g3 g8f7 g3h3 f7g7 a3b2 g7g8 h3h6 b7b6 g2g4 d8f8 f4f5 e6f5 g4f5 f8f5 h6d6 d7d6');
INSERT INTO evaluations VALUES(6688,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(6688,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(6688,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(6688,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(6688,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(6688,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(6688,36,'cp -48','b1c3 g8f6 e1g1 b4c3 b2c3 e8g8 e4e5 f6e4 c3b2 d7d6 f1e1 c8f5 c4d5 e4c5 f3d4 f5g6 d4c6 b7c6 d5c6 a8b8 d1d4 b8b2 d4b2 c5d3 b2d4 d3e1 a1e1 d6e5 d4a7 d8d6 e1c1 g6c2 c6b5 c7c6 b5f1 d6d2 a7a3 c2e4 a3e3 d2e3 f2e3 f8b8 a2a4 g8f8 a4a5');
INSERT INTO evaluations VALUES(6680,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(6680,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(6680,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(6680,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(6680,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(6680,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(6680,36,'cp -41','g8f6 e1g1 b4c3 b2c3 e8g8 e4e5 f6e4 c3b2 d7d6 f1e1 e4c5 e5d6 c7d6 a1c1 c8e6 f3d4 c6d4 d1d4 d8g5 f2f4 g5f6 d4f6 g7f6 e1e3 f8c8 c4e6 f7e6 b2f6 c8c7 c1d1 c7d7 f4f5 e6f5 d1f1 c5e4 f1f5 e4f6 f5f6 d6d5 e3g3 g8h8 g3d3 a8c8 a2a4 d5d4 f6f4 c8c4 a4a5 c4a4 g1f2 h7h6 f2g3 a4a5');
INSERT INTO evaluations VALUES(10495,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(10495,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(10495,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(10495,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(10495,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(10495,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(10495,36,'cp -31','e1g1 b4c3 b2c3 e8g8 e4e5 f6e4 c3b2 d7d6 f1e1 e4c5 e5d6 c7d6 a1c1 c8g4 h2h3 g4h5 d1d5 h5g6 h3h4 c5a4 b2a3 a4b6 d5d6 d8d6 a3d6 f8d8 c4a6 c6a5 d6b4 b7a6 b4a5 d8d5 a5d2 g6h5 c1c2 a8d8 e1e7 h5f3 g2f3 h7h5 d2g5 d5d7 g1g2 d7e7');
INSERT INTO evaluations VALUES(54966,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(54966,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(54966,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54966,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(54966,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54966,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(54966,36,'cp 18','e7e5 c2c4 g8f6 b1c3 b8d7 g1f3 g7g6 a2a4 a7a5 f1e2 f8g7 b2b3 e8g8 e1g1 f8e8 c1a3 e5d4 f3d4 d7c5 d1c2 c7c6 a1d1 d8e7 e2f3 c8g4 f3g4 f6g4 a3b2 e7e5 d4f3 e5e7 h2h3 g4f6 c3e2 f6e4 b2g7 g8g7 f1e1');
INSERT INTO evaluations VALUES(55579,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(55579,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(55579,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55579,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(55579,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(55579,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(55579,36,'cp 33','c2c4 g8f6 b1c3 b8d7 f1e2 g7g6 g1f3 f8g7 e1g1 e8g8 b2b4 e5d4 e3d4 f8e8 f1e1 d6d5 c1f4 d5c4 e2c4 d7b6 c4b3 e8e1 d1e1 c8f5 a1c1 b6d5 c3d5 f6d5 f4e5 c7c6 b4b5 a8c8 h2h3 f5e6 b5c6 c8c6 c1c6 b7c6 b3c4');
INSERT INTO evaluations VALUES(55567,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(55567,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(55567,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(55567,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(55567,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(55567,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(55567,36,'cp 25','g8f6 b1c3 g7g6 d4e5 d6e5 d1d8 e8d8 g1f3 b8d7 f3g5 d8e7 f1e2 h7h6 g5f3 b7b6 e1g1 c8b7 b2b3 a7a6 c1b2 e7e6 a1d1 f8d6 f3e1 h8d8 f2f3 h6h5 e1d3 d7c5 d3f2 a6a5 c3b5 f6e8 b5d6 e8d6 e3e4 f7f6 b2a3 d6f7 a3c5 d8d1 f1d1 b6c5 f2d3');
INSERT INTO evaluations VALUES(47842,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(47842,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(47842,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(47842,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(47842,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47842,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(47842,36,'cp 29','b1c3 g7g6 g1f3 d8e7 e3e4 f8g7 d4d5 a7a5 h2h3 f6d7 a2a3 e8g8 c1e3 f7f5 b2b4 d7f6 c4c5 f6e4 c3e4 f5e4 f3g5 f8f4 a1c1 e7g5 c5d6 c7c6 d5c6 b8c6 d6d7 g5d8 d7c8q a8c8 f1c4 g8h8 e1g1 c6d4 e3f4 e5f4');
INSERT INTO evaluations VALUES(47839,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(47839,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(47839,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(47839,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(47839,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(47839,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47839,36,'cp 27','g7g6 g1f3 d8e7 e3e4 f8g7 d4d5 a7a5 h2h3 f6d7 a2a3 e8g8 a1b1 f7f5 b2b4 a5b4 a3b4 d7f6 f1d3 b8d7 e1g1 d7b6 d1c2 c8d7 c1g5 b6a4 c3a4 d7a4 c2b2 a4d7 f1e1 h7h6 g5f6 e7f6 c4c5 f6e7 b2d2 g8h7 d3c4 h7h8');
INSERT INTO evaluations VALUES(50403,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(50403,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(50403,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50403,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50403,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50403,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50403,36,'cp 0','d5b6 c4b3 d7d5 e5d6 d8d6 b1c3 a7a5 a2a4 d6g6 g1e2 g6g2 h1g1 g2f3 d2d4 c8g4 d1d2 h7h5 g1g3 f3h1 g3g1');
INSERT INTO evaluations VALUES(50298,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(50298,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(50298,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(50298,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50298,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(50298,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50298,36,'cp 15','c4b3 d7d5 d2d4 c8f5 c2c3 e7e6 g1e2 c7c5 c1e3 b8c6 e1g1 a8c8 b1d2 c5d4 e2d4 f5d3 f1e1 f8e7 f2f4 e8g8 b3c2 d3g6 c2g6 h7g6 d4f3 c6a5 e3d4 a5c6 d4f2 c6a5 e1e2 d8c7 f3d4 a7a6 h2h4 a5c6 d2f3 c6d4 f2d4');
INSERT INTO evaluations VALUES(50304,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(50304,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50304,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50304,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50304,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50304,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50304,36,'cp 4','d7d5 d2d4 c8f5 g1e2 e7e6 c2c3 c7c5 c1e3 b8c6 b1d2 a8c8 e1g1 c5d4 e2d4 f5d3 f1e1 f8e7 f2f4 e8g8 b3c2 d3c2 d1c2 c6b4 c2b1 b4c6 d4f3 b6a4 d2b3 d8d7 b1c2 b7b6 b3d4 c6d4 f3d4 e7c5 c2f2 c5d4 e3d4 a4c5');
INSERT INTO evaluations VALUES(46933,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(46933,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(46933,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46933,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46933,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46933,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(46933,36,'cp 9','e5d6 d8d6 d1e2 b8c6 g1f3 e7e5 b1c3 d6e7 d2d3 c8e6 f3e5 e6b3 e5c6 e7e2 e1e2 b3c2 e2d2 b7c6 h1e1 e8d7 d2c2 a7a5 c1d2 f8d6 a2a3 a5a4 e1e3 a8a5 a1e1 a5f5 e1e2 f5h5 h2h3 h8b8 e3f3 f7f6 e2e1 c6c5 e1e4 f6f5 e4e2 b8a8');
INSERT INTO evaluations VALUES(46916,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46916,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46916,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46916,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46916,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46916,36,'cp 0','d8d6 b1c3 a7a5 a2a4 d6g6 g1e2 g6g2 h1g1 g2f3 g1g3 f3h1 g3g1');
INSERT INTO evaluations VALUES(53806,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(53806,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(53806,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(53806,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(53806,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(53806,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(53806,36,'cp 51','b8d7 a2a4 g8f6 e1g1 f8e7 d2d4 e8g8 h2h3 f8e8 f1e1 h7h6 a4a5 e5d4 f3d4 e7f8 c1f4 a7a6 d1d2 d7c5 f2f3 d6d5 e4d5 e8e1 a1e1 c6d5 c4a2 d8a5 c3d5 a5d2 d5f6 g7f6 f4d2 c8d7 b2b4 c5a4 a2b3 b7b5 c2c4 b5c4 b3c4 a4b6 c4d3 g8h8 g1h1 h8g8');
INSERT INTO evaluations VALUES(19437,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19437,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(19437,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19437,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(19437,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19437,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19437,36,'cp -2','g7g6 d2d3 d7d6 g1e2 f8g7 f1g2 e7e6 e1g1 g8e7 c1g5 a8b8 a2a4 a7a6 d1d2 h7h6 g5e3 c8d7 c3d1 d8c7 f1e1 b8d8 a4a5 c6d4 e2c1 d4c6');
INSERT INTO evaluations VALUES(19272,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19272,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19272,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19272,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19272,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19272,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19272,36,'cp 0','d2d3 f8g7 g1e2 e7e6 f1g2 d7d6 e1g1 g8e7 c1g5 a8b8 a2a4 a7a6 d1d2 h7h6 g5e3 c8d7 h2h3 c6d4 b2b4 b7b5 a4b5 a6b5 a1a6 d7c8 a6a1');
INSERT INTO evaluations VALUES(52130,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(52130,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(52130,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(52130,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(52130,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(52130,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(52130,36,'cp 231','f3e5 b8c6 e5c6 d7c6 d2d3 d8c7 f1e2 f8d6 d1d2 f7f5 f2f4 c8e6 b1c3 f5e4 d3e4 e8c8 d2e3 g8h6 e1g1 c8b8 g1h1 h8g8 e3f2 g7g5 e4e5 d6e7 c3e4 g5g4 c1e3 h6f5 e3c5 e7c5 f2c5 g4g3 e4f6 g3h2 a1e1 f5g3 h1h2 g3f1 e2f1 c7g7 f6g8');
INSERT INTO evaluations VALUES(52094,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(52094,31,'cp 244',NULL);
INSERT INTO evaluations VALUES(52094,32,'cp 247',NULL);
INSERT INTO evaluations VALUES(52094,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(52094,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(52094,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(52094,36,'cp 236','b8c6 e5c6 d7c6 d2d3 d8c7 g2g3 h7h5 b1d2 h5h4 f1g2 g8f6 d2c4 c8e6 c1f4 c7d8 d1e2 h4h3 g2f3 b7b5 c4e5 a8c8 e1g1 f8d6 e5f7 e8f7 f4d6 d8d6 e4e5 d6e7 e5f6 e7f6 a2a4 b5b4 b2b3 h8e8 f3e4 g7g6 a1b1 f6c3 e2f3 c3f6');
INSERT INTO evaluations VALUES(41271,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(41271,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(41271,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(41271,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(41271,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(41271,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(41271,36,'cp 25','e8g8 f1e2 c8f5 f3d2 a7a5 e3e4 f5c8 d2f3 e7e5 c1e3 d8e7 d4e5 d6e5 c3d5 e7d6 e3g5 f6d5 c4d5 c7c6 e1g1 c6d5 d1d5 d6b4 a2a3 b4b2 e2b5 b8c6 f1b1 b2c3 b1c1 c3b2 a1b1 b2a3 b5c6 b7c6');
INSERT INTO evaluations VALUES(45248,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(45248,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45248,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(45248,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(45248,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45248,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45248,36,'cp 21','b8c6 c1e3 f6g4 d4c6 b7c6 e3d2 g7g6 d2c3 g4f6 e4e5 f6d5 d3e4 d5c3 e4c6 c8d7 c6d7 d8d7 b1c3 f8g7 e5d6 g7c3 b2c3 e8g8 e1g1 a8c8 d1d4 f8d8 h2h3 a7a6 d6e7 d7e7 d4b4 e7b4 c3b4 c8c2 a2a3 d8d2 a1d1 d2d1 f1d1 c2a2 d1d3 a2b2 d3d7 h7h5 g1h2');
INSERT INTO evaluations VALUES(13373,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13373,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13373,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13373,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13373,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(13373,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(13373,36,'cp 16','c1e3 f6g4 d4c6 b7c6 e3d2 g7g6 d2c3 g4f6 e4e5 f6d5 d3e4 d5c3 e4c6 c8d7 c6d7 d8d7 b1c3 f8g7 e5d6 g7c3 b2c3 d7e6 d1e2 e6d6 e2b5 e8f8 a1d1 d6f6 e1g1 f8g7 d1d3 a8c8 b5b3 f6c6 h2h4 h8d8 b3b4 e7e6 f1d1 d8d3 d1d3 c6c5 b4a4 c8c6 g2g3');
INSERT INTO evaluations VALUES(13374,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(13374,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(13374,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(13374,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(13374,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(13374,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13374,36,'cp 13','f6g4 d4c6 b7c6 e3d2 g7g6 d2c3 g4f6 e4e5 f6d5 d3e4 d5c3 e4c6 c8d7 c6d7 d8d7 b1c3 f8g7 e5d6 d7e6 d1e2 g7c3 b2c3 e6d6 e2b5 e8f8 a1d1 d6f6 e1g1 f8g7 d1d3 h8c8 b5d5 e7e6 d5b7 a8b8 b7a7 b8a8 a7b6 f6e5 b6b2 e5a5 b2a1 a5a2 a1a2 a8a2');
INSERT INTO evaluations VALUES(53148,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53148,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53148,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53148,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53148,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53148,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53148,36,'cp 0','c8d7 b5e2 d7f5 f3h4 f5c8 h4f3');
INSERT INTO evaluations VALUES(30934,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30934,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30934,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30934,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30934,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30934,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30934,36,'cp 0','b5e2 d7f5 f3h4 f5c8 e1g1 b8c6 h4f3 c8f5 f3h4');
INSERT INTO evaluations VALUES(55876,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(55876,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(55876,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(55876,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(55876,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(55876,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55876,36,'cp -13','d5e4 d3e4 g8f6 e4f3 f8e7 b1c3 c7c5 g1e2 e8g8 c1e3 d8a5 d1d2 f8d8 e1g1 c5d4 e2d4 b8c6 d4c6 b7c6 d2c1 c8b7 f1e1 d8d7 e1d1 a8d8 d1d7 f6d7 c1e1 d7e5 f3e2 c6c5 c3b5 a5b6 a1d1 d8d1 e1d1 b6c6');
INSERT INTO evaluations VALUES(56051,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(56051,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(56051,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(56051,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56051,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(56051,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(56051,36,'cp -20','d3e4 g8f6 e4f3 f8e7 b1c3 c7c5 g1e2 e8g8 c1e3 d8a5 d1d2 b8c6 e1c1 f8d8 d2e1 c5d4 e2d4 c6e5 e1e2 e5f3 e2f3 c8d7 f3b7 e7a3 b7b3 a3b4 a2a3 b4c3 b3c3 a5c3 b2c3 d8c8 d4b3 c8c3 e3d4 c3c4 d4f6 d7a4 d1d4 a4b3 d4c4 b3c4 f6d4 f7f6 h1d1 c4b5 d4c5 g8f7 f2f3 e6e5 d1d6');
INSERT INTO evaluations VALUES(56044,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(56044,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56044,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(56044,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(56044,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56044,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56044,36,'cp -9','g8f6 e4f3 f8e7 g1e2 e8g8 b1c3 c7c5 e1g1 b8c6 d4c5 e7c5 c1f4 d8e7 e2c1 f8d8 c1d3 h7h6 f4g3 c6d4 f1e1 c5d6 c3e4 d4f3 d1f3 f6e4 f3e4 e7c7 g3d6 d8d6 d3e5 f7f6 e5c4 d6c6 c4e3 c8d7 a1d1 a8d8 d1d2 d7e8 d2d8');
INSERT INTO evaluations VALUES(48517,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48517,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(48517,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(48517,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(48517,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(48517,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(48517,36,'cp -17','e4f3 f8e7 g1e2 e8g8 b1c3 c7c5 c1e3 d8a5 e1g1 b8c6 d4c5 e7c5 e3c5 a5c5 c3a4 c5b5 e2c3 b5a6 d1d6 f8d8 d6c7 f6e8 c7g3 c6d4 a1d1 a6a5 b2b3 h7h6 f1e1 e8f6 h2h3 d4f3 g3f3 c8d7 f3b7 a8b8 b7f3 b8c8 c3e4 f6e4 f3e4 d7c6 e4f4 c6a4 f4a4 a5a4');
INSERT INTO evaluations VALUES(48515,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48515,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(48515,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48515,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(48515,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48515,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48515,36,'cp -12','f8e7 g1e2 e8g8 b1c3 c7c5 c1e3 d8a5 d4c5 f8d8 d1c1 b8c6 f3c6 b7c6 e1g1 e7c5 e3c5 a5c5 a2a3 c5e5 f1e1 c6c5 c1f4 e5f4 e2f4 c8a6 c3a4 a8c8 a1d1 d8d1 e1d1 g8f8 b2b3 a6b5 a4c3 a7a6 f2f3 f8e7 h2h4 c8d8 d1d8 e7d8 f4d3 b5d3 c2d3 f6d5 c3e4');
INSERT INTO evaluations VALUES(57396,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(57396,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(57396,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(57396,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(57396,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(57396,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(57396,36,'cp 170','e2e4 b8c6 f1c4 f6e5 g1f3 g8f6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8e7 f1e1 e8g8 e1e5 d4f5 g5e4 e7f6 e4f6 d8f6 e5e4 f5d6 e4f4 c8f5 b1c3 f6g6 c1e3 f5c2 d1d4 a7a6 h2h3 d6f5 d4c4 f5e3 f2e3 f8f4 c4f4 g6d6 a1f1 c2d3 f1d1 d3g6 f4d6 c7d6 g1f2');
INSERT INTO evaluations VALUES(57391,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(57391,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(57391,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(57391,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(57391,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(57391,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(57391,36,'cp 194','b8c6 f1c4 f6e5 g1f3 g8f6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8e7 f1e1 e8g8 e1e5 h7h6 d1d4 h6g5 d4c4 e7f6 e5e3 b7b5 c4b5 d8d6 b1c3 f6e5 h2h3 g5g4 c3e4 d6g6 h3g4 c8g4 b5c4 e5f4 e3d3 a8e8 d5d6 g8h8 c1f4 e8e4 c4c7 e4f4 d6d7 g4d7 c7d7 f4f2');
INSERT INTO evaluations VALUES(49415,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(49415,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(49415,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(49415,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(49415,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(49415,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(49415,36,'cp -12','d7d5 c1f4 f8b4 d1d3 c7c5 d4c5 d8a5 a2a3 b4c3 d3c3 a5c3 b2c3 b8d7 c3c4 d7c5 c4d5 f6d5 f4e5 f7f6 e5d4 b7b6 d4c5 b6c5 e2e4 d5c3 f3d2 e8e7 f2f3 a8b8 f1d3 c3a4 h2h4 c8d7 e1e2 h8d8 e2e3 d7b5 d3b5 b8b5');
INSERT INTO evaluations VALUES(1205,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(1205,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(1205,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(1205,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(1205,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(1205,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(1205,36,'cp -73','d4e4 d1e2 e4e2 e1e2 b7b6 d2d3 c8b7 a2a4 g8e7 c1e3 a6a5 b1c3 f7f6 a1e1 h7h5 h2h4 e7f5 e2d1 e8c8 e3d2 f8b4 c3e4 h8e8 e4g3 f5g3 f2g3 c8d7 h1f1 b4d6 e1e8 d8e8 d2f4 c6c5 f4d6 c7d6 f1e1 e8e6 e1e6');
INSERT INTO evaluations VALUES(38148,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(38148,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(38148,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38148,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38148,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(38148,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(38148,36,'cp 50','a2a3 b4c3 b2c3 g8e7 h2h4 d8c7 a1b1 b7b6 h4h5 h7h6 d1g4 e8f8 h1h3 c8a6 h3f3 c5d4 c3d4 f8g8 f1a6 b8a6 b1b2 a8f8 g1e2 a6b8 g4h3 b8c6 f3c3 c7d7 e1f1 c6a5 h3d3 a5c4 b2b1 d7c6 g2g4 c6c7 c1f4 e7c6 a3a4 c7d7');
INSERT INTO evaluations VALUES(38150,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(38150,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(38150,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(38150,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(38150,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(38150,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(38150,36,'cp 44','b4c3 b2c3 g8e7 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 a3a4 a7a6 e1f1 c5c4 d3e2 e8c8 d1c1 c8b8 g2g3 b8a8 f1g2 d8c8 c1a3 c7d8 d2f4 c6a5 f3h2 e7c6 h2g4 d8e7 a3e7 c6e7 g4e3 e7c6 e2g4 c8c7 g2g1');
INSERT INTO evaluations VALUES(38152,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(38152,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(38152,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38152,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(38152,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(38152,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(38152,36,'cp 49','b2c3 g8e7 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 a3a4 c6a5 d4c5 a5c4 e1f1 e8g8 h1h3 c4d2 d1d2 f7f5 h3g3 g8h8 f1g1 c7c5 a1b1 c5c7 d2f4 b7b6 d3b5 f8g8 f3d4 a8f8 b1e1 f8f7 g3f3 g8c8 e1a1 c8g8 b5d7 c7d7');
INSERT INTO evaluations VALUES(38151,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(38151,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(38151,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(38151,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38151,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38151,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(38151,36,'cp 40','g8e7 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 d1c1 c6a5 e1g1 a5c4 d2f4 c7a5 d4c5 a5c3 a1b1 c3a3 c1a3 c4a3 b1b7 d7c6 b7b3 a3c4 f3d4 a7a5 f1a1 e8g8 b3c3 a5a4 f4g3 c6d7 d3c4 d5c4 c3c4 a4a3 c4b4 a3a2');
INSERT INTO evaluations VALUES(51990,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(51990,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(51990,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(51990,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(51990,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(51990,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(51990,36,'cp 11','g2g3 d7d5 e4d5 e6d5 g1e2 h7h6 f1g2 g8f6 e1g1 b8c6 e2f4 f8e7 b1c3 c8g4 d1e1 d5d4 h2h3 g4d7 c3d5 e8g8 e1d1 e7d6 f1e1 a8c8 c1d2 b7b6 a2a3 d6f4 d5f4 f8e8 e1e8 d8e8 g3g4 c6e5 d1e2 d7c6 a1e1 c6g2 f4g2 e5g6');
INSERT INTO evaluations VALUES(51989,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(51989,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51989,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51989,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51989,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51989,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51989,36,'cp 0','d7d5 e4d5 e6d5 g1e2 h7h6 f1g2 g8f6 e1g1 b8c6 e2f4 f8e7 h2h3 e8g8 b1c3 d5d4 c3d5 f6d5 f4d5 c8e6 d5f4 e6f5 f4d5');
INSERT INTO evaluations VALUES(55105,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(55105,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(55105,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(55105,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(55105,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(55105,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55105,36,'cp -30','e7e5 c1b2 f8d6 c2c4 c7c6 b1c3 g8e7 f1d3 b8a6 c4d5 a6b4 d3b1 c6d5 g1e2 c8f5 b1f5 e7f5 d2d3 d5d4 e3d4 f5d4 e2d4 e5d4 c3e4 e8g8 e1g1 b4c6 b2c1 h7h6 c1d2 d6a3 d2c1 a3b4 c1f4 f7f5 e4g3 d8d5 a1c1 d5d7 c1c4');
INSERT INTO evaluations VALUES(17816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17816,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(17816,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(17816,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(17816,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17816,36,'cp 0','f8d6 d3d4 f7f6 c1e3 d8e7 b1d2 e5d4 f3d4 e7f7 d1e2 g8e7 d2c4 e8g8 e1c1 f6f5 f2f3 f5e4 f3e4 c8e6 c4d6 c7d6 h1f1 f7g6 f1f8 a8f8 d4e6 g6e6 e2d3 e6a2 d3d6 a2a1 c1d2 a1a5 d2e2 a5h5 e2d2');
INSERT INTO evaluations VALUES(4804,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4804,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(4804,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4804,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4804,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4804,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4804,36,'cp 14','d3d4 f7f6 c1e3 d8e7 b1d2 g8h6 d4e5 f6e5 d2c4 h6f7 f3d2 a6a5 c4d6 c7d6 e1g1 c8e6 a2a4 e8g8 d1e2 f7h8 f2f4 e5f4 e3f4 h8g6 f4g3 d6d5 f1f8 a8f8 e4d5 e7c5 g3f2 c5d5 e2e3 g6f4 d2e4 d5f5 a1e1 e6d5 f2g3 f8f7 g3f4 f5f4 e3f4 f7f4 e4c3 f4b4 c3d5');
INSERT INTO evaluations VALUES(46755,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(46755,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(46755,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46755,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(46755,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(46755,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(46755,36,'cp 36','f8g7 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 a1b1 e8g8 f1e2 c8g4 e1g1 g4f3 e2f3 c5d4 c3d4 g7d4 b1b7 b8c6 d1a4 d8c8 b7b1 f8d8 f1d1 a8b8 b1b8 c6b8 e4e5 d4e5 d1d8 c8d8 a4a7 e5d4 a7b7 g8g7 g2g3 b8d7 b7d5 e7e5 g1g2 d8c7 d5c6 c7a7 c6c2 a7c5 c2c5 d7c5 c1d2');
INSERT INTO evaluations VALUES(36564,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36564,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(36564,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(36564,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(36564,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36564,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(36564,36,'cp 50','e7e5 d4d5 f6h5 g2g3 f7f5 e4f5 g6f5 f3g5 h5f6 g3g4 d8e8 g4f5 c8f5 c1e3 b8a6 f1d3 e5e4 d3e2 a6c5 d1d2 f6d7 e1c1 d7e5 h1g1 g8h8 e3c5 d6c5 g5e4 e8e7 e4g5 e5f7 g5e6 f5e6 d5e6 e7e6 c3d5 a8e8 e2d3 f7e5 d5c7 g7h6 c7e6 h6d2 d1d2 e8e6 g1g5 e5f3 g5f5 f8f5 d3f5 e6e1 d2d1 e1e2 f5d3 e2f2');
INSERT INTO evaluations VALUES(36865,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(36865,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(36865,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(36865,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(36865,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(36865,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(36865,36,'cp 43','d4d5 f6h5 g2g3 f7f5 e4f5 g6f5 f3g5 h5f6 g3g4 d8e8 g4f5 c8f5 c1e3 b8a6 h1g1 a6c5 h3h4 e8d7 d1d2 g8h8 e1c1 b7b6 f1e2 a8e8 g1g2 c5e4 c3e4 f6e4 g5e4 f5e4 g2g3 f8g8 d1g1 g7f6 h4h5 e4f5 h5h6 g8g3 f2g3');
INSERT INTO evaluations VALUES(36838,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(36838,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(36838,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(36838,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36838,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36838,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(36838,36,'cp 19','f6h5 g2g3 f7f5 e4f5 g6f5 c1g5 d8e8 f3h4 b8a6 f1e2 h5f6 d1c2 a6b4 c2d2 a7a5 a2a3 b4a6 d2c2 a6c5 g5e3 b7b6 g3g4 f6e4 h4f5 c8f5 g4f5 e4c3 c2c3 f8f5 e1c1 g8h8 d1g1 e8f7 g1g4 e5e4 c3c2 a8f8 c1b1 g7e5 h1g1 f5f2 e3f2 f7f2');
INSERT INTO evaluations VALUES(7488,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(7488,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(7488,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(7488,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(7488,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(7488,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(7488,36,'cp 59','d2d4 d7d6 d4e5 d6e5 d1d8 e7d8 a2a4 c8b7 f1e1 e8g8 a4b5 a6b5 a1a8 b7a8 b1c3 b5b4 c3d5 f6e4 e1e4 c6a5 b3a2 b4b3 c2b3 a8d5 e4e5 c7c6 b3b4 a5c4 e5e2 c4b6 a2d5 b6d5 c1d2 d8f6 f3e5 f8e8 f2f4 f6e5 e2e5 e8a8 e5e4 f7f5 e4c4 a8a1 g1f2 a1a2 c4c2');
INSERT INTO evaluations VALUES(7485,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(7485,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(7485,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(7485,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(7485,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(7485,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(7485,36,'cp 49','d7d6 d4e5 d6e5 d1d8 e7d8 a2a4 a8b8 a4b5 a6b5 b1c3 e8g8 c1e3 d8e7 c3d5 e7d6 d5f6 g7f6 f1d1 c8e6 b3e6 f7e6 c2c3 b5b4 a1a6 c6d8 f3d2 b4c3 b2c3 d8f7 g1f1 f6f5 d2c4 b8a8 d1a1 a8a6 a1a6 f8b8 f1e2 g8g7 g2g3 f5f4 g3f4 e5f4 e3d4 e6e5');
INSERT INTO evaluations VALUES(7383,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(7383,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(7383,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(7383,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(7383,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(7383,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(7383,36,'cp 65','d4e5 d6e5 d1d8 e7d8 a2a4 a8b8 a4b5 a6b5 b1c3 e8g8 c1e3 f8e8 f3g5 e8f8 f1d1 d8e7 c3d5 e7d6 d5f6 g7f6 g5f3 c6e7 f3h4 c8g4 f2f3 g4e6 b3e6 f7e6 a1a6 b8a8 d1a1 a8a6 a1a6 e7g6 h4g6 h7g6 h2h4 g8f7 g1f1 f8h8 g2g3 f7e7 a6a5 b5b4 f1e2 e7d7');
INSERT INTO evaluations VALUES(47245,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(47245,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(47245,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(47245,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(47245,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(47245,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(47245,36,'cp -25','e2e3 c7c5 b2b3 g7g6 f1b5 c8d7 b5d7 d8d7 c1a3 d7c8 b1c3 f8g7 d1e2 e8g8 e1g1 b7b6 c3d1 c8c7 a3b2 b8c6 d2d3 a8d8 g2g3 f6e8 b2g7 e8g7 d1f2 f7f6 e3e4 d5e4 d3e4 e7e5 e2c4 g8h8 f4e5 c6e5');
INSERT INTO evaluations VALUES(47244,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(47244,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(47244,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(47244,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(47244,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(47244,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(47244,36,'cp -33','c7c5 b2b3 g7g6 c1b2 f8g7 f1e2 e8g8 e1g1 c8f5 d2d3 f6g4 b2g7 g8g7 d1d2 d5d4 e3d4 c5d4 b1a3 g4e3 f1f2 b8c6 h2h3 e3d5 f3e5 c6e5 f4e5 a8b8 e2f3 b7b5 a1e1 f5e6 f3d5 d8d5 d2g5 d5c5 a3b1 b5b4');
INSERT INTO evaluations VALUES(45393,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(45393,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(45393,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(45393,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(45393,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(45393,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(45393,36,'cp -47','b2b3 g7g6 c1b2 f8g7 f1e2 e8g8 e1g1 d8c7 f3e5 f8d8 b1a3 c8f5 e2f3 b8c6 e5c6 c7c6 d2d3 a8c8 d1e2 b7b5 c2c4 b5c4 d3c4 c6a6 a1d1 h7h5 c4d5 a6e2 f3e2 f6d5 b2g7 g8g7 g1f2 d5b4 d1d8 c8d8 f1c1 d8d2');
INSERT INTO evaluations VALUES(45392,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(45392,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(45392,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(45392,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(45392,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(45392,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(45392,36,'cp -35','g7g6 f1b5 c8d7 b5d7 d8d7 c1a3 d7c8 e1g1 f8g7 b1c3 e8g8 d1e2 b7b6 a3b2 b8c6 a1e1 e7e6 d2d3 d5d4 c3d1 f6d5 g2g3 d4e3 b2g7 g8g7 d1e3 d5e3 e2e3 c8c7 a2a4 a8d8 e3f2 d8d5 e1e4 h7h5 h2h3 f8d8 f3e5 d5d4 f1e1 d4e4 e1e4 c6e5 e4e5 c7d7 e5e4 d7d6');
INSERT INTO evaluations VALUES(20283,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(20283,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(20283,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(20283,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(20283,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(20283,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(20283,36,'cp 105','e4d5 e6d5 f1b5 g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6h5 f1d1 c8g4 b5c6 g4f3 d5e7 g8f8 e4f3 h5f3 c6f3 c5e7 c2c3 a8b8 g1f1 b7b6 f1e2 a7a5 a2a4 d8d1 a1d1 b8d8 f3d5 e7c5 h2h3 d8d6 d1d2 d6f6 d5f3 f6e6 e2d1 e6d6 d2d6 c5d6');
INSERT INTO evaluations VALUES(20228,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(20228,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(20228,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(20228,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(20228,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(20228,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(20228,36,'cp 96','e6d5 f1b5 g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 a7a6 b5c6 b7c6 d3c4 c5f2 f1f2 c6d5 c4d5 a8b8 c2c4 c8b7 d5d2 h6b6 d2c3 b7f3 c3f3 b6d4 f3d5 d4d5 c4d5 f8d8 f2d2 f7f6 b2b3 g8f7 d5d6 d8d7 a1c1 a6a5 c1c6');
INSERT INTO evaluations VALUES(20515,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(20515,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(20515,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(20515,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(20515,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(20515,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(20515,36,'cp 109','f1b5 g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6g6 e4g6 h7g6 a1d1 c8f5 c2c3 f5e4 d5e3 e4f3 g2f3 c5b6 g1g2 c6e7 f3f4 d8d1 f1d1 a8d8 d1d8 b6d8 g2f3 g8f8 b5c4 e7f5 e3d5 f5h4 f3e2 h4f5 a2a4 f5e7 d5e3 d8c7 e2f3 b7b6 c4b5');
INSERT INTO evaluations VALUES(57079,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(57079,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(57079,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(57079,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(57079,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(57079,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(57079,36,'cp 38','g1f3 d7d5 f1g2 g8f6 d2d4 f8e7 c2c4 e8g8 e1g1 d5c4 d1c2 b7b5 a2a4 c8b7 a4b5 a7a6 b5a6 b8a6 c2c4 b7d5 c4c3 c7c5 c1e3 f6e4 c3c1 a6b4 a1a8 d8a8 b1c3 e4c3 b2c3 c5d4 e3d4 b4c6 c1a1 a8b7 f1b1 f8a8 b1b7 a8a1 g2f1 e7f6 d4f6 g7f6 f3d2 a1c1 b7b1');
INSERT INTO evaluations VALUES(5536,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(5536,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(5536,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(5536,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(5536,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(5536,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(5536,36,'cp 78','f2f3 g8f6 d1d2 e8g8 g2g4 c8e6 d4e6 f7e6 e1c1 c6e5 f1e2 a8c8 h2h4 f6d7 f3f4 e5c4 e2c4 c8c4 d2d3 c4c6 c3e2 d8a5 c1b1 d7c5 e3c5 a5c5 h4h5 g6h5 h1h5 c5c4 c2c3 c4d3 d1d3 b7b6 b1c2 c6c4 d3e3 f8c8 c2d3 c4c7 g4g5 b6b5');
INSERT INTO evaluations VALUES(5538,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(5538,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(5538,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(5538,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(5538,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(5538,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(5538,36,'cp 70','d8a5 d1d2 a7a6 d4b3 g7c3 b3a5 c3d2 e3d2 c6a5 d2a5 f7f6 h2h4 g8h6 a5b6 h6f7 e1c1 c8e6 b2b3 a8c8 c1b2 f7e5 a2a4 h7h5 f1e2 e8f7 a4a5 e6d7 f3f4 e5c6 g2g3 d7e6 d1d2 c6b8 e2f3 b8d7 b6d4 d7c5 f3g2 e6d7 e4e5 d7c6');
INSERT INTO evaluations VALUES(21380,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(21380,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(21380,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(21380,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(21380,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(21380,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(21380,36,'cp 61','g1f3 g8f6 c4d5 e6d5 c1g5 f8e7 e2e3 e8g8 f1d3 f6h5 g5e7 c6e7 d1c2 g7g6 b2b4 c7c6 a2a4 h5f6 e1g1 c8f5 f1c1 a7a6 h2h3 f5d3 c2d3 d8d6 a1b1 g8g7 d3c2 a8c8 c2b2 d6e6 b4b5 c6b5 a4b5 a6b5 b2b5');
INSERT INTO evaluations VALUES(17712,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(17712,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(17712,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(17712,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(17712,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(17712,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(17712,36,'cp 66','c2c4 g8f6 b1c3 c8f5 a1c1 e7e6 g1f3 h7h6 f4g3 f8b4 f1e2 e8g8 e1g1 b4c3 c1c3 f6e4 c3c1 c6e7 g3h4 g7g5 h4g3 c7c6 f3d2 e4g3 h2g3 a6a5 a2a4 g8g7 d1b3 d8c7 e2d3 f5d3 b3d3 h6h5 e3e4 f7f6 d3e3 e7g6 f1d1 h5h4');
INSERT INTO evaluations VALUES(17705,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(17705,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(17705,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(17705,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(17705,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(17705,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(17705,36,'cp 72','g8f6 b1c3 c8f5 a1c1 e7e6 g1f3 h7h6 a2a3 f8d6 c4d5 f6d5 f4g3 d5c3 c1c3 e8g8 f1d3 f5d3 d1d3 c6a7 e1g1 d6g3 h2g3 c7c6 c3c2 a6a5 f1c1 a7c8 a3a4 d8e7 d3b3 e7b4 b3b4 a5b4 b2b3 c8e7');
INSERT INTO evaluations VALUES(57270,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(57270,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(57270,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(57270,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(57270,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(57270,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(57270,36,'cp -30','g8f6 g1f3 f8b4 c2c3 d4c3 b2c3 d7d5 e4d5 b4e7 c4b3 c8g4 e1g1 e8g8 f1e1 f8e8 c1b2 e7c5 e1e8 d8e8 b1d2 b8d7 d1f1 g4h5 c3c4 h5g6 h2h3 e8f8 b3a4 c5d6 f1d1 f8e7 a4c2 d7c5 c2g6 h7g6 d1c2 a8e8 a1e1 e7d7 g1f1 c7c6 d5c6 e8e1 f3e1 d7c6 d2b3 f6e4 b3d4');
INSERT INTO evaluations VALUES(17588,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(17588,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(17588,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(17588,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(17588,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(17588,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(17588,36,'cp 31','f1c4 e7e6 c1e3 g8f6 d1e2 c6d4 e3d4 b7b5 c4b3 f8e7 e1c1 e8g8 e4e5 d6e5 d4e5 f6d7 e2e3 d8e8 e5d4 c8b7 f2f4 e7f6 h1e1 e8e7 g2g4 f6d4 d1d4 a8d8 e1d1 d7c5 c1b1 d8d4 d1d4 e7c7 h2h4 c5b3 a2b3 f8d8 g4g5 d8d4');
INSERT INTO evaluations VALUES(17581,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(17581,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17581,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(17581,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(17581,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(17581,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(17581,36,'cp 41','e7e6 c1e3 g8f6 d1e2 c6d4 e3d4 b7b5 c4b3 f8e7 e1c1 e8g8 a2a3 d8c7 c1b1 a8b8 e4e5 f6d7 e5d6 e7d6 c3e4 d6e5 f2f3 a6a5 d4e5 d7e5 e4d6 c7c5 h1e1 e5c6 f3f4 c6d4 e2e5 c5e5 e1e5 d4b3 c2b3 c8b7 d6b7 b8b7 e5c5 g7g6 g2g4 f8e8 h2h4 g8g7 d1d6 e8h8 d6a6 h7h5');
INSERT INTO evaluations VALUES(17681,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(17681,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(17681,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(17681,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17681,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(17681,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(17681,36,'cp 33','c1e3 g8f6 d1e2 c6d4 e3d4 b7b5 c4b3 f8e7 e1c1 e8g8 a2a3 d8c7 f2f3 a8b8 e4e5 f6d7 e5d6 e7d6 c3e4 d6f4 c1b1 c8b7 h1e1 a6a5 e4f2 h7h6 f2d3 f4h2 f3f4 a5a4 b3a2 h2f4 d4g7 g8g7 d3f4 c7f4 d1d7 b7c6 d7e7 b5b4 a3b4 b8b4 e2e6 c6g2');
INSERT INTO evaluations VALUES(17679,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(17679,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(17679,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(17679,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(17679,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(17679,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(17679,36,'cp 31','g8f6 d1e2 c6d4 e3d4 f8e7 c4b3 b7b5 e1c1 e8g8 a2a3 d8c7 c1b1 a8b8 e4e5 f6d7 e5d6 e7d6 c3e4 d6e5 f2f3 a6a5 d4e5 c7e5 e2e1 a5a4 b3a2 d7f6 e1g3 e5g3 e4f6 g7f6 h2g3 g8g7 b1c1 b8b6 b2b4 a4b3 a2b3 f6f5 h1h4 e6e5 h4b4 b6g6 b4b5 g6g3 b5e5 g3g2 a3a4 g2g6');
INSERT INTO evaluations VALUES(11263,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(11263,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(11263,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(11263,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(11263,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(11263,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(11263,36,'cp 70','d1e2 c6d4 e3d4 b7b5 c4b3 f8e7 e1c1 e8g8 a2a3 d8c7 c1b1 a8b8 e4e5 f6d7 e5d6 e7d6 c3e4 d6e5 h1e1 h7h6 f2f3 c8b7 d4e5 d7e5 e4d6 e5g6 g2g3 b8d8 d6b7 d8d1 e1d1 c7b7 d1d6 a6a5 c2c3 b7c7 e2d2 g6e7 d6d7 c7c5 g3g4 e7c6');
INSERT INTO evaluations VALUES(38933,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38933,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38933,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(38933,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(38933,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(38933,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(38933,36,'cp 13','f1d3 g8f6 a2a3 b4c3 b2c3 e8g8 g1e2 c8g4 e1g1 g4h5 c1g5 b8d7 a1b1 d7b6 f2f3 h5g6 d1d2 h7h6 g5h4 f8e8 d3b5 c7c6 b5d3 d8d6 a3a4 e8e7 a4a5 b6c4 d3c4 d5c4 e2g3 a8e8 f3f4 g6e4 d2f2 c6c5 d4c5 d6a6 g3e4 e7e4 h4f6 a6f6 b1b7 f6c3 b7a7 e4e2');
INSERT INTO evaluations VALUES(38935,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(38935,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38935,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38935,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(38935,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(38935,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38935,36,'cp 19','g8f6 a2a3 b4c3 b2c3 e8g8 g1e2 c8g4 e1g1 g4h5 c1g5 d8d6 a1b1 b7b6 d1d2 h5g6 d2f4 d6f4 g5f4 f8c8 f2f3 c7c5 d4c5 c8c5 f4d6 c5c6 d6b4 a7a5 b4e7 b8d7 d3b5 c6e6 e7f6 d7f6 b1b2 e6e7 f1b1 h7h6 b5c6 a8c8 e2d4');
INSERT INTO evaluations VALUES(42243,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(42243,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(42243,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42243,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42243,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(42243,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(42243,36,'cp 20','a2a3 b4a5 g1f3 e8g8 e1g1 a5c3 b2c3 c8g4 f1e1 b8c6 c1f4 f8e8 e1e8 f6e8 d1e2 d8e7 e2e3 e8d6 f3e5 c6e5 f4e5 g4e6 a1e1 a8e8 e3f3 d6e4 c3c4 e4g5 f3e2 d5c4 d3c4 c7c6 h2h4 e6c4 e2c4 g5e6 d4d5 c6d5 c4d5');
INSERT INTO evaluations VALUES(14800,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(14800,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(14800,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14800,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(14800,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(14800,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(14800,36,'cp 25','b1d2 f8d6 g1f3 e8g8 f4g3 b7b6 f1d3 c8b7 e1g1 c6e7 b2b4 c7c5 b4c5 b6c5 d4c5 d6c5 a1b1 b7c6 f3d4 c6d7 d1e2 h7h6 d2b3 c5b6 b1c1 b6c7 g3c7 d8c7 c3c4 f8c8 c4c5 e6e5');
INSERT INTO evaluations VALUES(14799,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14799,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(14799,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(14799,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(14799,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(14799,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(14799,36,'cp 29','f8d6 g1f3 e8g8 f4g3 b7b6 f1d3 c6e7 e1g1 c8b7 b2b4 e7f5 g3e5 f6e4 d1c2 e4d2 f3d2 g7g6 c2b2 a7a5 a2a4 d8e7 b4b5 a8c8 f1e1 f8d8 a1b1 b7a8 e5f4 a8b7 b2b3 f5g7 b3c2 d8e8 f4g3 g7f5 g3e5 f7f6 e5d6 f5d6 c3c4 d5c4');
INSERT INTO evaluations VALUES(9233,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(9233,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(9233,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(9233,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(9233,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(9233,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(9233,36,'cp 25','g1f3 e8g8 f4g3 c6e7 f1d3 b7b6 e1g1 a7a5 e3e4 d5e4 d2e4 f6e4 d3e4 a8b8 f1e1 c8b7 e4b7 b8b7 d1e2 d8d7 h2h4 d6g3 f2g3 c7c5 g3g4 c5d4 a1d1 d7c8 d1d4 b7d7 d4d7 c8d7 h4h5 d7d5 e2e4 h7h6 a2a3 b6b5 e4d5 e7d5');
INSERT INTO evaluations VALUES(52524,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(52524,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52524,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52524,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52524,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52524,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52524,36,'cp 19','e3d4 d7d5 g1f3 b8c6 c2c4 g8f6 b1c3 c8g4 c4d5 f6d5 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d5c3 b2c3 d8d5 b5e2 f7f6 e1g1 e7f7 f1d1 d5c6 c1e3 f8d6 c3c4 h8d8 c4c5 d6e7 d1c1 c6d5 a2a4');
INSERT INTO evaluations VALUES(52515,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52515,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(52515,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(52515,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(52515,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(52515,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52515,36,'cp 15','d7d5 f1d3 b8c6 c2c3 g8f6 c1f4 c8g4 g1f3 e7e6 h2h3 g4h5 e1g1 f8d6 f4d6 d8d6 b1d2 e8g8 f1e1 a8c8 d1e2 c6e7 a2a4 d6b6 e1b1 e7c6 e2e3 a7a6 g2g4 h5g6 d3g6 h7g6 g1g2 b6c7 b1e1 c8e8 f3e5 c6e5 e3e5 c7b6 d2f3 b6b2');
INSERT INTO evaluations VALUES(15755,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(15755,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(15755,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(15755,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(15755,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(15755,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(15755,36,'cp 105','f8e7 e1c1 e8g8 f1e2 c7c6 e4e5 d6e5 d4e5 c8d7 c3e4 f8e8 g5f6 e7f6 e4f6 d8f6 e5f6 g7f6 d1d7 e8e2 h1f1 b7b5 c1d1 e2e8 f1e1 a7a5 e1e8 a8e8 b2b3 e8e5 d7d6 e5f5 d1e2 f5c5 e2d3 f6f5 c2c4 f5f4 h2h4 b5c4 b3c4 g8g7 f2f3 c5h5 d6c6 h5h4 c6c5 h4h2 c5a5 h2g2');
INSERT INTO evaluations VALUES(8686,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(8686,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(8686,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(8686,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(8686,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(8686,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(8686,36,'cp 99','f1e2 e8g8 e1c1 c7c6 e4e5 d6e5 d4e5 c8d7 c3e4 f8e8 g5f6 e7f6 e4f6 d8f6 e5f6 g7f6 d1d7 e8e2 h1f1 a8b8 c1d1 e2e5 f1e1 g8g7 e1e2 h7h5 h2h4 a7a5 d1d2 b7b5 e2e5 f6e5 d2e3 b8c8 d7d6 f7f6 f2f4 e5f4 e3f4 c8e8 d6c6 b5b4 f4f3 e8d8 f3e3 d8e8 e3d3 e8d8 d3c4 d8d2 c4b5 d2g2 b5a5');
INSERT INTO evaluations VALUES(23104,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(23104,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(23104,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(23104,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(23104,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(23104,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(23104,36,'cp 49','d4f3 d1f3 g8f6 d2d4 d7d6 e1g1 f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 h7h6 a2a4 b7b6 a4a5 a8e8 e6g4 e4g4 h3g4 h8g8 g1f1 f8f6 a1a4 f6g6 g2g3 e5e4 a5b6 a7b6 g4g5 h6g5 d1d4 e8e5 d4e4 g6e6 e4e5 e6e5');
INSERT INTO evaluations VALUES(23095,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(23095,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(23095,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(23095,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(23095,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(23095,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(23095,36,'cp 52','d1f3 g8f6 d2d4 d7d6 e1g1 f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 h7h6 d1d4 e4h7 d4d2 b7b6 e6g4 h6h5 g4e2 h7g6 d2d5 f8f6 a1d1 a8f8 e2d3 g6d3 d1d3 h8g8 c3c4 a7a5');
INSERT INTO evaluations VALUES(23094,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(23094,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(23094,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(23094,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(23094,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(23094,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(23094,36,'cp 58','g8f6 d2d4 d7d6 e1g1 f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 b7b6 d1d4 e4c6 e6c4 c6e8 d4d5 e8g6 a2a4 h7h6 a1d1 a8e8 a4a5 e8e6 c4e2 h8h7');
INSERT INTO evaluations VALUES(16974,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(16974,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(16974,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(16974,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(16974,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(16974,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(16974,36,'cp 57','d2d4 d7d6 e1g1 f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 b7b6 d1d4 e4c6 e6c4 c6e8 d4d5 e8g6 a1d1 h7h6 a2a4 a8e8 a4a5 e8e6 b2b4 e5e4 c4e2 h8h7 e2g4 g6g4 h3g4');
INSERT INTO evaluations VALUES(16985,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(16985,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(16985,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(16985,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(16985,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(16985,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(16985,36,'cp 53','d7d6 e1g1 f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 h7h6 d1d4 e4c6 d4d2 b7b6 a1d1 a7a5 e6d5 c6d5 d2d5 a5a4 c3c4 a8b8 g1f1 h8g8 f1e2 f8f6 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(15749,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(15749,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(15749,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(15749,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(15749,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(15749,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(15749,36,'cp 56','e1g1 f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 b7b6 a2a4 h7h6 a4a5 a8d8 d1d2 h8h7 a1d1 b6a5 e6d5 e4d5 d2d5 d6e7 e3a7 d8d6 a7c5 d6d5 d1d5 f8d8 d5d8 e7d8 g1f1 a5a4 c5b4 h7g6 f1e2 g6f5');
INSERT INTO evaluations VALUES(15750,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(15750,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(15750,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(15750,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(15750,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(15750,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(15750,36,'cp 57','f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 b7b6 d1d4 e4c2 d4d2 c2e4 c3c4 e4h4 a1d1 a8e8 e6g4 h4g4 h3g4 h7h6 d2d5 h8g8 g1f1 e8e6 a2a3 e6g6 b2b4 e5e4 g4g5 h6g5 d5g5 g6g5 e3g5 f8f5 g5e3 g8f7 f1e2 f7e6');
INSERT INTO evaluations VALUES(52104,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(52104,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(52104,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(52104,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(52104,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(52104,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(52104,36,'cp 108','d2d3 d7d6 g1e2 g8f6 e1g1 f8e7 f2f4 b8c6 b1c3 a7a6 a2a4 c8d7 e2g3 e5f4 c1f4 e8g8 h2h3 c6d4 c3d5 f6d5 c4d5 d4e6 f4e3 e6c7 d5b3 d7e6 b3e6 f7e6 d1g4 e7f6 c2c3 d8e8 g3h5 e8g6 h5f6 f8f6 e3g5 f6f1 a1f1 a8f8 f1f8 g8f8 g4f4 f8e8 f4d6 g6g5 d6c7 g5e3 g1h1');
INSERT INTO evaluations VALUES(52101,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(52101,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(52101,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(52101,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(52101,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(52101,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(52101,36,'cp 90','d7d6 g1e2 g8f6 e1g1 f8e7 b1c3 b8c6 f2f4 a7a6 a2a4 c8d7 e2g3 e5f4 c1f4 e8g8 h2h3 c6d4 c3d5 f6d5 c4d5 d7e6 g3h5 d8d7 c2c3 d4c6 d1d2 g7g6 h5g3 e7h4 d5c6 b7c6 f1f3 d7e7 f4h6 f8e8 a1f1 f7f6 d2f4');
INSERT INTO evaluations VALUES(57033,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(57033,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(57033,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(57033,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(57033,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(57033,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(57033,36,'cp -29','d7d5 c1b2 b8c6 e4d5 e6d5 g1f3 g8f6 f1e2 f8d6 e1g1 e8g8 f1e1 c8f5 d2d3 f8e8 b1d2 h7h6 a2a3 a7a5 g2g3 d6e7 a1c1 d8d7 d3d4 f6e4 d2e4 d5e4 f3e5 c6e5 d4e5 e7c5 d1d7 f5d7');
INSERT INTO evaluations VALUES(55886,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(55886,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(55886,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(55886,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(55886,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(55886,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(55886,36,'cp -19','c1b2 b8c6 e4d5 e6d5 g1f3 g8f6 f1e2 f8d6 e1g1 e8g8 f1e1 f8e8 d2d3 a7a5 a2a3 c6e5 f3e5 d6e5 b2e5 e8e5 b1d2 d8d6 h2h3 e5e7 e2f1 e7e1 d1e1 h7h6 d2f3 g8f8 d3d4 c8f5 f1d3 f6e4 c2c4 c7c6 f3e5 f5h7 d3e4');
INSERT INTO evaluations VALUES(55887,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55887,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(55887,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(55887,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(55887,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(55887,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(55887,36,'cp -30','b8c6 e4d5 e6d5 g1f3 g8f6 f1e2 f8d6 e1g1 e8g8 f1e1 c8g4 h2h3 g4h5 d2d3 d8d7 b1d2 f8e8 c2c4 d6f4 a2a3 e8e6 d1c2 d5c4 b3c4 a8e8 e2f1 e6e1 a1e1 e8e1 f3e1 c6d4 c2c3 c7c5 d2b3 d4b3 c3b3 f4d2 b2f6 g7f6 e1c2 h5g6 b3b2 d7e7 d3d4 g8g7 d4c5 e7c5');
INSERT INTO evaluations VALUES(21275,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(21275,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(21275,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(21275,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(21275,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(21275,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(21275,36,'cp 96','d2d4 f8e7 d4e5 d6e5 d1d8 e7d8 c4b5 f7f6 b5c6 b7c6 c1e3 h6h5 e1c1 g8e7 d1d2 e7g6 f3e1 d8e7 e1d3 g6f8 c3a4 f8e6 b2b3 g7g5 h2h3 e8f7 d3c5 h8d8 d2d8 e7d8 h1d1 d8e7 c1b2 a7a5 f2f3 g5g4 h3g4 h5g4 c5d3 c8a6 f3g4 a6b5 a4c5');
INSERT INTO evaluations VALUES(21293,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(21293,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(21293,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(21293,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(21293,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(21293,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(21293,36,'cp 83','f8e7 d4e5 d6e5 d1d8 e7d8 c4b5 f7f6 b5c6 b7c6 c1e3 h6h5 e1c1 g8e7 d1d2 e7g6 f3e1 d8e7 e1d3 g6f8 f2f3 f8e6 c3a4 e8f7 b2b3 h8d8 h1d1 a7a5 h2h3 e7a3 c1b1 a3e7 d3c5 d8d2 d1d2 g7g5 e3f2 e6c5 a4c5');
INSERT INTO evaluations VALUES(26532,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26532,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(26532,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26532,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(26532,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(26532,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(26532,36,'cp 9','e7e6 f1d3 f5d3 d1d3 f8d6 f4g3 g8f6 g1f3 e8g8 b1d2 c6e7 e3e4 d5e4 d2e4 f6e4 d3e4 d8d7 e1g1 b7b6 a1d1 h7h6 f1e1 d7b5 e4c2 b5f5 c2e2 e7g6 h2h3 f8e8 b2b4 a7a5 b4b5 d6g3 f2g3 f5d5 c3c4 d5d6 f3e5 g6e5 e2e5 d6b4');
INSERT INTO evaluations VALUES(26807,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26807,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(26807,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(26807,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26807,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26807,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(26807,36,'cp 21','f1d3 f5d3 d1d3 f8d6 f4g3 g8f6 g1f3 e8g8 b1d2 c6e7 e3e4 f6e4 d2e4 d5e4 d3e4 d8d7 e1g1 d7b5 e4c2 a7a5 f1e1 h7h6 a1d1 b5c6 a2a3 b7b5 g3e5 a8d8 h2h4 f8e8 e1e4 e7f5 c2e2 b5b4 a3b4 a5b4 c3c4 d6e5 f3e5');
INSERT INTO evaluations VALUES(26811,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(26811,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(26811,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(26811,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26811,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26811,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26811,36,'cp 11','f5d3 d1d3 f8d6 f4g3 g8f6 g1f3 e8g8 b1d2 c6e7 e3e4 f6e4 d2e4 d5e4 d3e4 h7h6 e1g1 a7a5 f1e1 d8d7 a1d1 f8e8 a2a3 d7c6 e4d3 b7b6 d3c2 c6b5 h2h3 b5f5 c2e2 e7g6 f3h4 g6h4 g3h4 a5a4 h4g3 d6g3 f2g3 a8d8 g3g4 f5d5 e2a6 d5a8 a6a8 d8a8');
INSERT INTO evaluations VALUES(31345,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31345,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(31345,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(31345,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(31345,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(31345,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31345,36,'cp 0','e7e6 e1g1 g8f6 c1f4 b8c6 c2c3 f8d6 f4d6 d8d6 b1d2 e8g8 f1e1 g4h5 a2a4 a7a6 e2d3 f8e8 d1b3 d6c7 h2h3 e8c8 b3d1 c8e8');
INSERT INTO evaluations VALUES(31716,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31716,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(31716,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31716,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(31716,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(31716,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(31716,36,'cp 11','e1g1 g8f6 c1f4 b8c6 b1d2 f8d6 f4d6 d8d6 c2c3 g4f5 f1e1 e8g8 d2b3 h7h6 a2a4 d6c7 h2h3 f8c8 b3c5 b7b6 c5a6 c7d8 e2b5 f6e4 d1e2 e4d6 b5c6 c8c6 f3e5 c6c8 a4a5 b6a5 a6c5 d6e4 c5d7 e4d6 f2f3 d6c4 e5c4 d5c4 d7e5 d8c7 e2f2');
INSERT INTO evaluations VALUES(31715,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31715,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(31715,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31715,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(31715,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31715,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31715,36,'cp 10','g8f6 c1f4 g4f3 e2f3 f8d6 f4d6 d8d6 b1d2 d6b6 c2c3 b8c6 a1b1 e8g8 f3e2 f8e8 d2f3 e8e7 f1e1 a8e8 e2f1 g7g6 g2g3 g8g7 d1c2 a7a6 b1d1 h7h5 a2a3 b6c7 c2d2 c6a5 g1g2 b7b5 f1d3 a5c4 d3c4 b5c4');
INSERT INTO evaluations VALUES(29398,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29398,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29398,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(29398,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29398,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29398,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(29398,36,'cp 11','c1f4 b8c6 b1d2 f8d6 f4d6 d8d6 c2c3 g4f5 a2a4 d6c7 f1e1 e8g8 d2b3 h7h6 e2b5 a7a5 f3e5 c6a7 b5d3 f5d3 e5d3 a7c8 d1e2 f8e8 d3e5 f6e4 b3d2 e4f6 d2f3 c8d6 e5d3 a8c8 f3e5 d6f5 g2g3 f6e4');
INSERT INTO evaluations VALUES(8974,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(8974,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(8974,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(8974,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(8974,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(8974,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(8974,36,'cp 64','c1d2 e8g8 a2a3 b4d6 c4c5 d6e7 f1b5 f6e4 e1g1 c8d7 b5d3 f7f5 b2b4 a7a6 d2e1 g7g5 g1h1 g5g4 f3g1 e7g5 c3e2 d8e7 a3a4 e7g7 b4b5 a6b5 a4b5 c6e7 a1a5 c7c6 d1a1 a8a5 a1a5 c6b5 f2f3 g4f3 g2f3 e4f6 e1f2 g8h8 d3b5 d7c6');
INSERT INTO evaluations VALUES(8972,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(8972,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(8972,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(8972,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(8972,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(8972,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(8972,36,'cp 52','e8g8 a2a3 b4d6 c4c5 d6e7 f1b5 f6e4 e1g1 c8d7 b5d3 f7f5 b2b4 a7a6 c3e2 g7g5 f3e5 c6e5 d4e5 g5g4 f2f3 g4f3 g2f3 e4d2 d1d2 g8h8 g1h1 d8e8 e2d4 e8h5 f3f4 f8g8 f1g1 h5h4 d2c2 h4h5 c2f2 e7h4 f2f3 g8g1 a1g1 h5f3 d4f3 h4e7');
INSERT INTO evaluations VALUES(3674,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(3674,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(3674,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(3674,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(3674,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(3674,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(3674,36,'cp 64','a2a3 b4d6 c4c5 d6e7 f1b5 f6e4 b2b4 a7a5 a1b1 e7f6 e1g1 a5b4 a3b4 e4d2 d1d2 e6e5 b5c6 e5d4 f3d4 f6d4 d2d4 b7c6 b4b5 c8f5 b1b2 c6b5 b2b5 f5d7 b5b7 d7c6 b7b1 d8g5 f1d1 f8d8 b1a1 a8a1 d1a1 g5f5 h2h3 h7h6 a1a7 d8b8 a7c7');
INSERT INTO evaluations VALUES(3679,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(3679,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(3679,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(3679,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(3679,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(3679,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(3679,36,'cp 58','b4d6 c4c5 d6e7 f1b5 f6e4 e1g1 c8d7 b5d3 f7f5 d2e1 a7a5 a1b1 g7g5 g1h1 d8e8 f3g1 a5a4 c3b5 e7d8 f2f3 e4f6 g1h3 h7h6 h3f2 g8h8 b2b4 a4b3 d1b3 c6a5 e1a5 a8a5 a3a4 a5a8 b5c3 b7b6 c5b6 c7b6 d3b5 d8c7 f2d3 c7d6 b5d7 f6d7');
INSERT INTO evaluations VALUES(31965,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(31965,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(31965,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(31965,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(31965,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(31965,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(31965,36,'cp 63','b1c3 e7e6 g1f3 f8e7 e3e4 f5g6 f1c4 g8f6 d1e2 c7c5 d4c5 e7c5 e4e5 f6d7 e1g1 e8g8 h2h3 h7h6 f1d1 b8c6 c3e4 g6e4 e2e4 d8e7 a2a3 f8d8 b2b4 c5b6 c4d3 f7f5 e5f6 d7f6 e4e2 c6d4 f3d4 b6d4 c1b2 d4b2 e2b2 a8c8 a1c1');
INSERT INTO evaluations VALUES(31962,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(31962,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(31962,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(31962,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(31962,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(31962,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(31962,36,'cp 57','e7e6 f1c4 g8f6 g1f3 f8e7 e1g1 e8g8 f3h4 f5e4 c3e4 f6e4 h4f3 c7c5 d1c2 e4f6 d4c5 d8c7 c1d2 c7c5 a1c1 b8d7 c4e2 c5c2 c1c2 f8c8 f1c1 d7b6 g1f1 c8c2 c1c2 a8c8 c2c8 b6c8 f3e5 f6e4 d2e1 c8b6 e5d3 e7d6');
INSERT INTO evaluations VALUES(32367,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(32367,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(32367,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(32367,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(32367,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(32367,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(32367,36,'cp 61','g1f3 g8f6 f1c4 f8e7 e1g1 e8g8 f3h4 f5e4 c3e4 f6e4 h4f3 c7c5 d1c2 e4f6 d4c5 d8c7 c1d2 c7c5 a1c1 b8d7 c4e2 c5c2 c1c2 f8c8 f1c1 d7b6 g1f1 c8c2 c1c2 a8c8 c2c8 b6c8 e2d3 c8b6 e3e4 f6d7 f1e2 a7a6 b2b3 g8f8 h2h3 h7h6 d3c2 f8e8 g2g3 e7f6 d2e3');
INSERT INTO evaluations VALUES(32368,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(32368,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(32368,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(32368,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(32368,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(32368,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(32368,36,'cp 63','g8f6 f1c4 f8e7 e1g1 e8g8 f3h4 f5e4 c3e4 f6e4 h4f3 c7c5 d1c2 e4f6 d4c5 d8c7 c1d2 c7c5 b2b4 c5c6 a1c1 b8d7 c2b1 f8c8 c4d3 c6d6 f1d1 d7e5 d3e2 c8c1 b1c1 a8d8 d2e1 e5f3 e2f3 d6b8 d1d8 e7d8 c1b1 f6d7 a2a4 d8f6 a4a5');
INSERT INTO evaluations VALUES(30202,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(30202,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(30202,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(30202,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(30202,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(30202,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(30202,36,'cp 69','f1c4 f8e7 e1g1 e8g8 f3h4 f5e4 c3e4 f6e4 h4f3 c7c5 d1c2 e4f6 d4c5 d8a5 c1d2 a5c5 b2b4 c5c6 a1c1 b8d7 c2b1 a8c8 c4e2 c6d6 a2a3 c8c1 f1c1 d6b8 b1c2 d7b6 c2c7 b6d5 c7b8 f8b8 d2e1 g8f8 e2d3 a7a6');
INSERT INTO evaluations VALUES(30195,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(30195,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(30195,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(30195,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(30195,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(30195,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(30195,36,'cp 58','f8e7 d1e2 c7c5 e1g1 e8g8 f1d1 f5g4 h2h3 g4f3 e2f3 b8c6 c4b5 d8c7 b5c6 b7c6 d4c5 e7c5 c1d2 f8d8 c3a4 c5f8 b2b3 d8d7 d2c3 d7d1 a1d1 f6d5 c3e1 a8d8 f3e4 h7h6 a4b2 f8e7 b2c4 d5f6 d1d8 e7d8 e4d4 f6d5 c4e5');
INSERT INTO evaluations VALUES(37398,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(37398,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(37398,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(37398,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(37398,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37398,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(37398,36,'cp 7','d7d5 c4d5 f6d5 f1c4 d5b6 c4b3 c7c5 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 c1e3 g7c3 e3c5 c3b2 d5d6 e7d6 a1d1 b2e5 f3b7 d8c8 b7c8 f8c8 c5d6 e5d6 d1d6 c8c7 f2f4 a8c8 f4f5 b6c4 b3c4 c7c4 f5g6 f7g6 f1b1 c8c7 d6a6 g8g7 g1h2 c7d7');
INSERT INTO evaluations VALUES(13209,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13209,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(13209,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13209,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(13209,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(13209,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(13209,36,'cp 23','c8d7 c1g5 e7e6 d1d3 c6d4 b5d7 d8d7 d3d4 f8e7 e1c1 a8c8 g5f6 e7f6 d4d6 d7d6 d1d6 f6c3 b2c3 c8c3 h1d1 e8g8 d6d7 b7b6 c1b2 c3c4 b2b3 c4e4 d7a7 b6b5 a7b7 e4e5 f2f4 e5f5 g2g3 f8a8 d1d7 g7g6 h2h4 g8g7 d7c7 h7h6 a2a3 g6g5 h4g5');
INSERT INTO evaluations VALUES(14615,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(14615,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(14615,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14615,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(14615,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14615,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(14615,36,'cp 30','d3d4 e5d4 f3d4 f8e7 e1g1 e8g8 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 f1e1 a8d8 h2h3 f8e8 c4d3 e7f8 d4e3 d7c6 e3f3 f6d7 f3g3 g8h8 f4d2 d7c5 b2b3 d6d5 e4d5 e6d5 e1e8 c6e8 c3d5 d8d5 g3f3 e8d7 d2c3 c5d3');
INSERT INTO evaluations VALUES(26050,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26050,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(26050,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(26050,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(26050,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(26050,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(26050,36,'cp 42','b5c6 d7c6 d2d4 c5d4 f3d4 g8f6 d1e2 d8b6 d4c6 b7c6 b2b3 g7g6 c1b2 f8h6 e1g1 b6c7 c3a4 e8g8 e4e5 f6d7 e5d6 e7d6 c2c4 f8e8 e2f3 h6g7 b2g7 g8g7 f3c3 g7g8 f1e1 e8e1 a1e1 d7f8 a4b2 a8d8 b2d3 f8e6 h2h3 d8e8 c3b2 c6c5');
INSERT INTO evaluations VALUES(26006,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(26006,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(26006,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(26006,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26006,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(26006,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(26006,36,'cp 35','d7c6 d2d4 c5d4 f3d4 g8f6 d1e2 e7e5 d4c6 b7c6 c1g5 f8e7 e1g1 e8g8 g5f6 e7f6 f1d1 a8b8 a1b1 b8b4 a2a3 b4d4 b2b4 d8d7 b4b5 f8c8 b5c6 d4d1 e2d1 d7c6 c3d5 f6d8 g2g3 g7g6 b1b3 g8h8 b3c3 c6d7 c3c8 d7c8 d5e3 c8d7 h2h4 h8g7 g1g2');
INSERT INTO evaluations VALUES(26225,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(26225,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(26225,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(26225,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(26225,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(26225,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(26225,36,'cp 30','d2d4 c5d4 f3d4 g8f6 d1e2 e7e5 d4c6 b7c6 e1g1 f8e7 c3a4 e8g8 f1d1 d8d7 c2c4 d7e6 c1g5 f6e8 g5e7 e6e7 e2d3 a8d8 a4c3 e8c7 b2b4 c7e6 a1c1 e6f4 d3f3 h7h6 b4b5 d8c8 a2a4 f4e6 h2h3 g8h8 a4a5 e6d4 f3d3 f8d8');
INSERT INTO evaluations VALUES(26224,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26224,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26224,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(26224,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(26224,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26224,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(26224,36,'cp 28','c5d4 f3d4 g8f6 d1e2 e7e5 d4c6 b7c6 e1g1 f8e7 c3a4 e8g8 f1d1 d8d7 c2c4 d7e6 c1g5 f6e8 g5e7 e6e7 b2b4 e8c7 a4c3 c7e6 d1d2 a8d8 b4b5 e7c7 b5c6 c7c6 c3d5 f8e8 a1c1 d8d7 e2g4 d7b7 h2h3 h7h6');
INSERT INTO evaluations VALUES(26233,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(26233,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(26233,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(26233,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(26233,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26233,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(26233,36,'cp 23','f3d4 g8f6 d1e2 e7e5 d4c6 b7c6 c1g5 f8e7 e1g1 e8g8 g5f6 e7f6 a1d1 d8d7 b2b3 a8d8 f1e1 f6g5 e2c4 g8h8 c4d3 f7f5 e4f5 d7f5 d3f5 f8f5 c3e4 g5h4 c2c4 d6d5 e4g3 h4g3 h2g3 h8g8 c4d5 c6d5');
INSERT INTO evaluations VALUES(26228,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26228,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(26228,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(26228,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(26228,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(26228,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26228,36,'cp 28','g8f6 d1e2 e7e5 d4c6 b7c6 e1g1 f8e7 c3a4 e8g8 f1d1 d8d7 c2c4 d7e6 c1g5 f6e8 g5e7 e6e7 c4c5 a8d8 a1c1 e8f6 d1d6 d8d6 c5d6 e7d6 a4c5 h7h5 h2h4 f8d8 g2g3 d6b8 c1c2 b8b5 b2b3 f6g4 e2b5 d8d1 g1g2 c6b5 c5a6 f7f6 a6c7 b5b4');
INSERT INTO evaluations VALUES(24595,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(24595,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(24595,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(24595,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(24595,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(24595,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24595,36,'cp 30','d1e2 e7e5 d4c6 b7c6 e1g1 f8e7 f1d1 e8g8 c3a4 d8a5 b2b3 a8d8 c1d2 a5c7 c2c4 d6d5 c4d5 c6d5 a1c1 c7d6 e4d5 f6d5 a4c3 d6e6 c3d5 d8d5 d2e3 d5d1 c1d1 a7a6 d1c1 f8c8 c1c8 e6c8 h2h3 c8c6 g2g4 g7g6 e3d2 f7f6 h3h4 c6c2 g4g5 f6g5 h4g5');
INSERT INTO evaluations VALUES(50267,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(50267,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50267,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(50267,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(50267,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(50267,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50267,36,'cp 13','f8g7 c2c4 c7c6 b2b3 f6e4 d2d4 d7d5 e1g1 c8f5 c1b2 e8g8 b1c3 e4c3 b2c3 b8d7 e2e3 f5e4 d1e2 e7e6 f1d1 f8e8 a1c1 h7h5 f3d2 e4g2 g1g2 d7f6 b3b4 d8d7 a2a4 d5c4 e2c4 f6d5 b4b5 c6b5 c4b5 d7c6 b5c6 b7c6');
INSERT INTO evaluations VALUES(50266,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(50266,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(50266,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(50266,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50266,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(50266,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50266,36,'cp 32','c2c4 f8g7 d2d4 c7c6 d1b3 d7d6 f1g2 b8d7 e1g1 e7e5 b1c3 e8g8 b3c2 a7a5 d4e5 d7e5 f3e5 d6e5 c3a4 c8e6 a4c5 d8e7 c5e6 e7e6 b2b3 f8d8 c1b2 a5a4 a1d1 a4b3 d1d8 a8d8 a2b3');
INSERT INTO evaluations VALUES(50181,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50181,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50181,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(50181,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50181,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50181,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(50181,36,'cp 18','c7c6 f1g2 f8g7 b2b3 f6e4 d2d4 e8g8 c1b2 d7d5 e1g1 c8f5 b1c3 e4c3 b2c3 b8d7 a1c1 f5e4 e2e3 e7e6 d1e2 f8e8 f1d1 a8c8 f3d2 e4g2 g1g2 d7f6 a2a4 h7h5 b3b4 f6d7 e3e4 a7a5 b4a5 e6e5 d4e5 g7e5');
INSERT INTO evaluations VALUES(51675,30,'cp 492',NULL);
INSERT INTO evaluations VALUES(51675,31,'cp 464',NULL);
INSERT INTO evaluations VALUES(51675,32,'cp 496',NULL);
INSERT INTO evaluations VALUES(51675,33,'cp 497',NULL);
INSERT INTO evaluations VALUES(51675,34,'cp 519',NULL);
INSERT INTO evaluations VALUES(51675,35,'cp 522',NULL);
INSERT INTO evaluations VALUES(51675,36,'cp 525','d1f3 b8c6 f3c6 c8d7 c6a6 e7e6 g1f3 g8f6 f3e5 f6d5 c1d2 d8c8 b2b3 d5b4 a6c8 d7c8 a1a5 f7f6 a5b5 f6e5 d2b4 c8d7 b4f8 h8f8 f1c4 d7b5 c4b5 e8e7 d4e5 f8c8 e1e2 c8c2 b1d2 a8d8 b5d3 c2a2 h1c1 a7a5 c1c7 e7f8 c7c3 f8g8 c3c4 g8f7 c4c6 f7f8 c6e6 a5a4 b3a4 a2a4 d3h7');
INSERT INTO evaluations VALUES(51676,30,'cp 498',NULL);
INSERT INTO evaluations VALUES(51676,31,'cp 497',NULL);
INSERT INTO evaluations VALUES(51676,32,'cp 498',NULL);
INSERT INTO evaluations VALUES(51676,33,'cp 509',NULL);
INSERT INTO evaluations VALUES(51676,34,'cp 490',NULL);
INSERT INTO evaluations VALUES(51676,35,'cp 496',NULL);
INSERT INTO evaluations VALUES(51676,36,'cp 505','a7a6 f3a8 d8c7 a8f3 e7e5 d4e5 c7e5 b1c3 g8f6 f3g3 e5g3 h2g3 b8d7 g1f3 d7c5 f3d4 f8d6 c3b5 a6b5 d4b5 d6e5 a1a8 e8g8 f2f4 f6e4 f4e5 e4g3 b5d6 c8a6 a8a6 c5a6 h1g1 g3f1 g1f1 a6c5 d6c4 f8c8 c1d2 c5b3 c4d6 c8c2 d2c3 b3c5 f1f7 c2g2 e1f1 g2g3 f1f2 g3g4 f2f3');
INSERT INTO evaluations VALUES(44864,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44864,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44864,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(44864,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44864,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(44864,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44864,36,'cp 11','c6d5 c1f4 b8c6 e2e3 c8f5 a1c1 a7a6 g1f3 e7e6 f1e2 a8c8 e1g1 f8e7 c3a4 e8g8 a2a3 c6a5 a4c5 a5c4 b2b3 c4a3 c5b7 c8c1 d1c1 d8b6 b7c5 a3c2 c5a4 b6b7 a4c5');
INSERT INTO evaluations VALUES(39274,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(39274,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(39274,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(39274,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(39274,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(39274,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(39274,36,'cp 24','g1f3 d7d6 c2c3 b8c6 b2b4 c5b6 a2a4 a7a6 f4e5 d6e5 f1c4 g8f6 d2d3 e8g8 b1a3 d8d6 a4a5 b6a7 d1e2 c8e6 c1e3 a7e3 e2e3 f6g4 e3d2 a8d8 a1d1 g4f6 e1g1 e6g4 a3b1 g4e6 c4e6 d6e6 d2e3 f6g4 e3e2 e6h6 g1h1');
INSERT INTO evaluations VALUES(39273,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(39273,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(39273,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(39273,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(39273,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(39273,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(39273,36,'cp 17','d7d6 c2c3 b8c6 b2b4 c5b6 a2a4 a7a6 f4e5 d6e5 f1c4 g8f6 d2d3 e8g8 b1a3 d8d6 a4a5 b6a7 d1e2 c8e6 c1e3 a7e3 e2e3 f6g4 e3d2 a8d8 c4e6 d6e6 e1g1 g4f6 d2e3 f6h5 a3c4 h5f4 c4e5 c6e5 f3e5 e6e5 d3d4 f4g2 g1g2 e5e6 e3f3 f7f6 a1e1 e6c4 g2g1 f8e8 g1h1');
INSERT INTO evaluations VALUES(38734,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(38734,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(38734,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(38734,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(38734,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(38734,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(38734,36,'cp 30','c2c3 b8c6 b2b4 c5b6 a2a4 a7a6 f4e5 d6e5 f1c4 g8f6 d1e2 e8g8 d2d3 d8d6 a4a5 b6a7 c1e3 a7e3 e2e3 f6g4 e3e2 c6e7 b1a3 b7b5 a5b6 a6a5 c4a2 a5b4 a3c4 d6c5 d3d4 e5d4 c3d4 c5b5 c4e5 b5e2 e1e2 c8a6 a2c4 g4e5 f3e5 a6c4 e5c4 c7b6');
INSERT INTO evaluations VALUES(38733,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(38733,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(38733,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(38733,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(38733,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(38733,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(38733,36,'cp 12','g8f6 f4e5 d6e5 f3e5 d8e7 d2d4 c5d6 e5f3 f6e4 f1e2 e8g8 e1g1 b8d7 b1d2 d7f6 d2c4 c7c5 c4d6 e7d6 c1e3 f6g4 d1c1 f8e8 e3f4 d6b6 f1e1 c5d4 f3d4 c8d7 c1c2 e4c5 h2h3 g4f6 e2f3 d7c6 f3c6 b7c6 a2a4 a7a5 f4e5 c5d7 e5g3');
INSERT INTO evaluations VALUES(41897,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(41897,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(41897,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(41897,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(41897,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(41897,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(41897,36,'cp 12','f4e5 d6e5 f3e5 d8e7 d2d4 c5d6 c1f4 b8d7 e5d7 e7e4 d1e2 e8d7 f4d6 e4e2 f1e2 d7d6 b1d2 c7c6 e1g1 h8e8 a1e1 c8e6 a2a4 a7a5 d2c4 d6c7 e2d3 b7b6 c4e5 e6d5 h2h3 h7h6 d3c4 d5c4 e5c4 e8e1 f1e1 b6b5 a4b5 c6b5 c4e5 a8e8 g1f2 e8e7 f2e3 f6d5 e3e4 c7d6 g2g4');
INSERT INTO evaluations VALUES(52961,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(52961,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(52961,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(52961,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(52961,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(52961,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(52961,36,'cp 77','e2e4 d6d5 e4d5 c6d5 b1c3 g8f6 g1f3 b8c6 c1g5 e7e6 c4c5 h7h6 g5f6 g7f6 f1b5 f8g7 e1g1 e8g8 f1e1 c6e7 d1d2 b7b6 b2b4 a7a6 b5d3 d8c7 d3c2 a8b8 d2d3 f6f5 c2a4 a6a5 a2a3 a5b4 c3b5 c7d8 a3b4 e7c6 b5d6');
INSERT INTO evaluations VALUES(42476,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42476,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(42476,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(42476,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(42476,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(42476,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(42476,36,'cp 3','e8g8 g1f3 d7d5 f1d3 d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 c1g5 c8b7 d1e2 b4c3 b2c3 b8d7 c4b3 d8c7 c3c4 h7h6 g5h4 c7f4 h4g3 b7f3 e2f3 f4f3 g2f3 f8d8 f1d1 a8c8 a2a4 f6h5 a4a5 h5g3 h2g3 g7g5 g1g2 h6h5 f3f4 g5f4 g3f4 g8g7 d1g1');
INSERT INTO evaluations VALUES(37174,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(37174,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(37174,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(37174,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(37174,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37174,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37174,36,'cp 12','g1f3 b7b6 f1d3 d7d5 e1g1 c7c5 d4c5 d5c4 d3c4 d8d1 f1d1 b4c5 b2b3 b8c6 c1b2 c8b7 a1c1 a8c8 c4e2 f8d8 a2a3 c5e7 d1d8 c6d8 g1f1 f6d5 b3b4 e7f6 c3d5 c8c1 b2c1 b7d5 f3d2 d8c6 d2c4 g8f8 f2f3 f6c3 e3e4 d5c4 e2c4 c6d4 c4d3');
INSERT INTO evaluations VALUES(21536,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(21536,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(21536,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(21536,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(21536,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(21536,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(21536,36,'cp 185','f3e5 a7a6 e5c6 d8g5 b5f1 b7c6 b1c3 f8b4 a2a3 b4a5 d2d3 g5g6 d1f3 g8f6 c1d2 a5b6 f1e2 c8g4 f3g3 g4e2 g3g6 h7g6 e1e2 d5e4 d3e4 e8c8 f2f3 b6d4 c3a4 d4e5 d2c3 e5h2 e2f2 h2f4 h1h8 d8h8 a4c5 f6h5 a1h1 f4g3 f2e3 g3d6 c5a6 c8b7');
INSERT INTO evaluations VALUES(21526,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(21526,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(21526,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(21526,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(21526,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(21526,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(21526,36,'cp 165','a7a6 e5c6 d8g5 b5f1 b7c6 b1c3 g5g6 d1f3 f8c5 e4d5 g8e7 f1e2 h7h5 e1g1 c8g4 f3d3 c6d5 d3g6 f7g6 h2h3 g4e2 c3e2 c5d6 d2d3 e8d7 f1e1 a6a5 c1f4 e7f5 a1d1 h8b8 d1b1 b8f8 a2a3 a8b8 f4d6 d7d6 b2b4 a5b4 a3b4 b8a8 c2c3');
INSERT INTO evaluations VALUES(32132,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(32132,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(32132,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(32132,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32132,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(32132,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(32132,36,'cp 19','c7c5 a1b1 b8c6 a2a3 g8f6 e1g1 f8e7 b2b4 c5b4 a3b4 g4f3 e2f3 c6d4 f3e2 d4e2 d1e2 e8g8 d2d3 a8c8 c1d2 h7h6 h2h3 d8d7 b1b3 a7a6 b4b5 a6b5 f1b1 d6d5 b3b5 d5e4 b5b7 e4d3 c2d3 d7d8 c3e4 c8c2 e4g3 e7d6 g3f5 c2a2');
INSERT INTO evaluations VALUES(44534,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(44534,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(44534,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(44534,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44534,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(44534,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44534,36,'cp 21','b1c3 b8c6 d2d4 c5d4 f3d4 e7e6 a2a3 f8e7 e2e4 e8g8 d4f3 b7b6 c1f4 c8b7 f1e2 a8c8 e1g1 d7d6 a1c1 c6e5 f3e5 d6e5 f4e5 f6e4 c3e4 b7e4 b2b4 e7f6 d1d8 f8d8 e5f6 g7f6 f1d1 d8d1 c1d1 e4b7 f2f3 b7a6 d1d7 a6c4 e2c4 c8c4');
INSERT INTO evaluations VALUES(42741,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(42741,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(42741,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(42741,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(42741,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(42741,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(42741,36,'cp -28','d7d5 e4d5 f6d5 c1d2 e8g8 c3d5 b4d2 d1d2 d8d5 f1e2 c7c5 c2c3 b7b6 e1g1 b8c6 a2a3 c8f5 a1d1 a8d8 f1e1 f8e8 h2h3 a7a5 f3h4 f5d7 e2f3 d5d6 d2g5 g7g6 f3c6 f7f6 c6d5 g8g7 g5g3 d6d5 f2f4 e5f4 g3f4 e8e1 d1e1 d8e8 e1e8 d7e8 f4c7 e8f7 c7b6');
INSERT INTO evaluations VALUES(42261,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(42261,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(42261,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(42261,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(42261,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(42261,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(42261,36,'cp -27','e4d5 f6d5 c1d2 e8g8 c3d5 b4d2 d1d2 d8d5 f1e2 c7c5 e1g1 b8c6 a2a3 b7b6 c2c3 f7f6 b2b4 c8e6 b4c5 b6c5 a1b1 c6e7 d2b2 a8d8 f1d1 e6f5 f3h4 f5d7 b2b7 d5b7 b1b7 g7g5 h4f3 e7d5 e2f1 d7g4 b7a7 g4f3 g2f3 d5c3 d1c1 c3b5 a7a6 b5d4');
INSERT INTO evaluations VALUES(42214,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(42214,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(42214,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(42214,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(42214,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(42214,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(42214,36,'cp -37','f6d5 c1d2 e8g8 c3d5 b4d2 d1d2 d8d5 f1e2 c7c5 e1g1 b8c6 a2a3 b7b6 c2c3 c8b7 f1e1 a8d8 a1d1 c6e7 d3d4 c5d4 e2d3 d4c3 d2c3 e7g6 d3c2 d5c6 c3c6 b7c6 c2g6 c6f3 d1d8 f8d8 g2f3 h7g6 e1e5 g8f8 e5e4 a7a5 e4c4 d8d3 g1g2 f8e7');
INSERT INTO evaluations VALUES(42277,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(42277,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(42277,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(42277,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(42277,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(42277,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(42277,36,'cp -31','c1d2 e8g8 c3d5 b4d2 d1d2 d8d5 f1e2 c7c5 e1g1 b8c6 a2a3 b7b6 c2c3 c8b7 f1e1 a8d8 a1d1 c6e7 f3h4 d5e6 e2f3 b7d5 f3d5 d8d5 d2e2 e7c6 b2b4 f8d8 h2h3 h7h6 h4f3 c5b4 a3b4 e6f5 e2e4 f5e4 e1e4 d5d3 d1d3 d8d3');
INSERT INTO evaluations VALUES(42278,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(42278,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(42278,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(42278,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(42278,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(42278,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(42278,36,'cp -28','e8g8 c3d5 b4d2 d1d2 d8d5 f1e2 c7c5 e1g1 b8c6 a2a3 f7f6 c2c3 c8e6 b2b4 b7b6 f1e1 c6e7 e2f1 f8d8 b4c5 b6c5 d2b2 a8b8 b2c2 e7g6 d3d4 c5d4 c3d4 b8c8 c2e2 e6g4 d4e5 g4f3 e2f3 d5f3 g2f3 g6e5 g1g2 c8c2 a1c1 d8d2 c1c2 d2c2 g2g3');
INSERT INTO evaluations VALUES(25934,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(25934,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25934,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(25934,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25934,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25934,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25934,36,'cp 0','b5c6 b7c6 c2c3 c5d4 c3d4 c6c5 e1g1 c5c4 b2b3 a8c8 c1d2 h7h6 f1e1 a7a5 b3c4 d5c4 d1e2 g8e7 e1c1 e7d5 c1c4 f8e7 b1c3 d5c3 d2c3 e8g8 c4c8 d8c8 c3a5 c8b7 e2d3 f8a8 d3b3 d7b5 a5d2 a8a3 b3b2 b5c6 b2b7 c6b7 a1b1');
INSERT INTO evaluations VALUES(25901,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25901,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(25901,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(25901,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25901,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25901,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25901,36,'cp -1','b7c6 c2c3 c5d4 c3d4 c6c5 e1g1 h7h6 b1c3 c5d4 c3e2 f8c5 e2d4 g8e7 c1e3 c5b6 h2h4 e8g8 a1c1 e7g6 h4h5 g6h4 d1d2 d8e7 a2a3 f8c8 c1c8 a8c8 f1c1 c8c1 d2c1 a7a5 f3e1 b6d4 e3d4 d7b5 d4c5 e7b7 c1c3 b7a6 c5d6');
INSERT INTO evaluations VALUES(25820,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25820,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25820,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(25820,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25820,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25820,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25820,36,'cp 0','c2c3 c5d4 c3d4 c6c5 e1g1 c5c4 b2b3 c4b3 a2b3 f8b4 c1d2 d8b8 b1a3 g8e7 a3c2 b4d2 d1d2 b8b3 c2b4 d7b5 f1c1 e8g8 a1b1 b3a3 b1a1');
INSERT INTO evaluations VALUES(25818,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25818,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25818,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25818,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25818,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25818,36,'cp 0','c5d4 c3d4 c6c5 e1g1 c5c4 b2b3 c4b3 a2b3 f8b4 c1d2 d8b6 b1a3 g8e7 a3c2 b4d2 d1d2 b6b3 c2b4 d7b5 f1c1 e8g8 a1b1 b3a4 b1a1');
INSERT INTO evaluations VALUES(25825,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25825,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25825,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25825,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25825,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25825,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25825,36,'cp 0','c3d4 c6c5 e1g1 c5c4 b2b3 c4b3 a2b3 f8b4 c1d2 d8b6 b1a3 b4d2 d1d2 g8e7 a3c2 b6b3 c2b4 d7b5 f1c1 e8g8 a1b1 b3a4 b1a1');
INSERT INTO evaluations VALUES(25823,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25823,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25823,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25823,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25823,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25823,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25823,36,'cp 0','c6c5 e1g1 c5c4 b2b3 c4b3 a2b3 f8b4 c1d2 d8b6 b1a3 b4d2 d1d2 g8e7 a3c2 b6b3 c2b4 d7b5 f1c1 e8g8 a1b1 b3a4 b1a1');
INSERT INTO evaluations VALUES(25829,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25829,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25829,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25829,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25829,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25829,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25829,36,'cp 0','e1g1 c5c4 b2b3 c4b3 a2b3 f8b4 c1d2 d8b6 b1a3 g8e7 a3c2 b4d2 d1d2 b6b3 c2b4 d7b5 f1c1 e8g8 a1b1 b3a4 b1a1');
INSERT INTO evaluations VALUES(20584,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(20584,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(20584,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(20584,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(20584,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(20584,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20584,36,'cp 26','g8f6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 f1e2 e8g8 e1g1 b8d7 b2b4 a7a6 b5a3 a6a5 b4a5 a8a5 a3c4 a5a8 c1e3 b7b6 a1b1 a8a4 c4b6 a4a2 e2c4 a2a3 d1d2 c8a6 d2b4 d7b6 c4a6 a3a6 e3b6 d8a8 c2c4 a6a3 b4b5 e7d8');
INSERT INTO evaluations VALUES(56393,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(56393,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(56393,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56393,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(56393,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56393,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56393,36,'cp -4','d5e4 f3e5 g8f6 f1e2 e4d3 e5d3 f8d6 b1a3 b8c6 a3c4 c6d4 e1g1 e8g8 e2f3 d6e7 f1e1 f8e8 b2b3 c7c6 c1b2 c8f5 c4e5 e7f8 a2a4 a7a5 g2g4 f5d3 d1d3 d4f3 d3f3 d8d5 f3d5 f6d5 e5d3 e8e1 a1e1 a8d8');
INSERT INTO evaluations VALUES(56620,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56620,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56620,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56620,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56620,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56620,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56620,36,'cp 0','f3e5 g8f6 f1e2 d8d4 c1f4 d4b2 b1d2 f6d5 e5c4 b2b4 f4e5 e4d3 e2f3 d5c3 d1c1 b8d7 e5c7 c3e2 c1d1 e2c3');
INSERT INTO evaluations VALUES(56516,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56516,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56516,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56516,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56516,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56516,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56516,36,'cp 0','g8f6 f1e2 d8d4 c1f4 d4b2 b1d2 f6d5 e5c4 b2b4 f4e5 e4d3 e2f3 d5c3 d1c1 c3e2 c1d1');
INSERT INTO evaluations VALUES(48901,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48901,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48901,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48901,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48901,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48901,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48901,36,'cp 0','f1e2 d8d4 c1f4 d4b2 b1d2 f6d5 e5c4 b2b4 f4e5 e4d3 e2f3 d5c3 d1c1 c3e2 c1d1');
INSERT INTO evaluations VALUES(19277,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(19277,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(19277,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(19277,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(19277,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(19277,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(19277,36,'cp 98','d2d4 d7d5 e4d5 d8d5 c1e3 f8g7 d4c5 d5d1 e1d1 g8f6 b1d2 e8g8 a2a4 f8d8 a4a5 f6d5 f1c4 c8e6 d1c2 a8c8 a5a6 b7a6 c4a6 e6f5 c2c1 c8b8 a1a4 c6e5 f3e5 d5e3 e5c6 b8b2 c1b2 d8d2 b2a3 d2f2 c6e7 g8f8 e7f5 f2f5 a6e2 f5c5');
INSERT INTO evaluations VALUES(19267,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(19267,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(19267,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(19267,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(19267,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(19267,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(19267,36,'cp 47','d7d5 e4d5 d8d5 c1e3 c5d4 c3d4 d5a5 b1c3 c8e6 f1b5 g8f6 e1g1 f6d5 a2a4 a7a6 b5c6 b7c6 c3e4 d5e3 f2e3 f8g7 e4c5 e6c8 h2h3 e8g8 a1c1 a5c7 d1d2 a6a5 d2f2 c7d6 b2b3 a8b8');
INSERT INTO evaluations VALUES(26957,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(26957,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26957,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(26957,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(26957,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(26957,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(26957,36,'cp 50','h2h3 g4f3 d1f3 g8f6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 e2g3 e8g8 a1b1 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 g7g6 b5c6 b7c6 c1e3 f8b8 b1c1 d6a3 c1c2 h7h5 e3g5 f6h7 g5c1 a3f8 c1e3 b8d8 f1d1 d8d7 f3e2 a8d8 g3f1 h7f6 e3g5 d7d4 f1e3');
INSERT INTO evaluations VALUES(26955,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(26955,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(26955,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(26955,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(26955,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(26955,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(26955,36,'cp 59','g4f3 d1f3 g8f6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 a1b1 e8g8 e2g3 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 g7g6 b5c6 b7c6 c1e3 f8b8 f3e2 h7h5 e2c2 c6c5 b1b8 a8b8 f1c1 h5h4 g3e2 d6e7 c2a4 e6d7 e2c3');
INSERT INTO evaluations VALUES(26971,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(26971,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(26971,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(26971,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(26971,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(26971,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(26971,36,'cp 55','d1f3 g8f6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 a1b1 e8g8 e2g3 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 g7g6 b5c6 b7c6 c1e3 f8b8 b1d1 h7h5 d3d4 e5d4 e3d4 d6e5 g3e2 b8d8 e2f4 e6e7 d4e3 d8d1 f1d1 g8g7');
INSERT INTO evaluations VALUES(26970,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(26970,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(26970,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(26970,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(26970,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(26970,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(26970,36,'cp 51','g8f6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 e2g3 e8g8 a1b1 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 f8e8 b5c6 b7c6 c1g5 f6d7 b1c1 d6f8 c1c2 a7a5 g3f5 c6c5 f1c1 a5a4 g5d2 e8b8 f3g4 g8h8 f5e3 e6g4 h3g4 f7f6');
INSERT INTO evaluations VALUES(57217,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(57217,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(57217,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(57217,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(57217,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(57217,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(57217,36,'cp -54','d7d5 d2d4 b8c6 d4e5 c6e5 b1d2 g8f6 g1f3 c8g4 h2h3 g4h5 f1e2 e5f3 e2f3 h5f3 d2f3 d8d7 e1f1 f6e4 b2b3 h7h5 c1b2 f7f6 f1g2 e8c8 d1e2 f8d6 a1d1 c7c6 f3d4 g7g6 c2c4 d5c4 e2c4 h5h4 g3g4 h8e8 c4c2');
INSERT INTO evaluations VALUES(56472,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(56472,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(56472,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(56472,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(56472,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(56472,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(56472,36,'cp -40','d2d4 b8c6 g1f3 e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 c8e6 a2a3 a7a6 c4d5 f6d5 d1c2 d8d7 c3d5 e6d5 f1c4 d7f7 c4d5 f7d5 d2c4 e8c8 c1d2 d8d7 h2h4 f8e7 e1c1 h8d8 c1b1 c8b8 d1c1 h7h6 h1d1 g7g5 c2b3 g5h4');
INSERT INTO evaluations VALUES(50174,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(50174,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(50174,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(50174,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50174,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(50174,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(50174,36,'cp 28','f8g7 f1g2 c7c5 e2e3 e7e6 b2b3 d7d5 g1e2 e8g8 e1g1 f8e8 d2d4 c5d4 e3d4 b8c6 f1e1 a7a6 c1e3 c8d7 c4d5 f6d5 c3d5 e6d5 e2f4 d7f5 g3g4 f5e6 a1c1 d8d6 f4e6 f7e6 h2h3 a8d8 d1d2 d8d7 e1d1 d6b4 d2b4 c6b4');
INSERT INTO evaluations VALUES(51419,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(51419,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(51419,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(51419,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(51419,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(51419,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(51419,36,'cp 79','c8b7 f1d3 g8f6 h2h4 d7d6 d1e2 a7a6 g1f3 b8d7 e1g1 b6b5 f1d1 c7c5 d4c5 d7c5 e4e5 b5b4 e5f6 b4c3 f6g7 f8g7 b2c3 e8g8 a1b1 b7f3 e2f3 d8h4 b1b4 h4f6 f3g3 c5d3 g3d3 f6c3 d3d6 a8d8 d6d8 c3b4');
INSERT INTO evaluations VALUES(48214,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(48214,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48214,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(48214,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48214,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(48214,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(48214,36,'cp 10','e2e3 c7c5 b1d2 b8d7 f1e2 f8e7 e1g1 e8g8 c2c3 d8c7 g5f4 e7d6 f4d6 c7d6 c3c4 b7b6 a1c1 c8b7 c4d5 e6d5 b2b3 f8c8 f1e1 h7h6 h2h3 d6e7 f3h4 g7g6 h4f3 c8c7 d4c5 b6c5 d1c2 a7a5 e2b5 b7c6 b5c6 c7c6 g1h1 c6c7 e3e4');
INSERT INTO evaluations VALUES(48212,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48212,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(48212,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(48212,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(48212,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48212,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(48212,36,'cp 4','c7c5 b1d2 b8d7 f1e2 f8e7 e1g1 e8g8 c2c3 a7a6 f3e5 d7e5 d4e5 f6d7 g5e7 d8e7 f2f4 f7f6 e5f6 d7f6 d2f3 c8d7 a2a4 d7e8 f3e5 f6d7 e5d7 e8d7 d1d2 a8d8 b2b4 d8c8 e2d3 d7e8 e3e4 d5e4 d3e4 e8g6 a1e1 c5b4 c3b4 g6e4');
INSERT INTO evaluations VALUES(46150,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46150,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(46150,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(46150,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46150,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(46150,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46150,36,'cp 0','b1d2 b8d7 f1e2 f8e7 e1g1 e8g8 c2c3 a7a6 f3e5 d7e5 d4e5 f6d7 g5e7 d8e7 f2f4 f7f6 e5f6 d7f6 d2f3 c8d7 a2a4 d7e8 h2h3 e8g6 f3e5 c5c4 d1d4 g6c2 d4b6 a8e8 b6d4');
INSERT INTO evaluations VALUES(50388,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(50388,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(50388,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(50388,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(50388,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(50388,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(50388,36,'cp 21','e7e5 c2c4 f8c5 d2d3 e8g8 b1c3 c7c6 g1f3 d7d6 e1g1 a7a5 h2h3 f8e8 d3d4 e5d4 f3d4 a5a4 a1b1 d8a5 e2e3 h7h6 c1d2 a4a3 c3e2 c5b4 d2b4 a5b4 b2b3 d6d5 b1c1 d5c4 b3c4 b8a6 g3g4 c8e6 d1c2 a8d8 d4e6 e8e6 e2d4');
INSERT INTO evaluations VALUES(53008,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53008,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(53008,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53008,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53008,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53008,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53008,36,'cp 0','g8f6 g1f3 e7e6 a2a4 b8c6 e1g1 f8e7 c4b3 c6a5 b3a2 c8d7 c1f4 e8g8 b1d2 a7a6 c2c3 a5c6 a2c4 c6a5');
INSERT INTO evaluations VALUES(45137,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45137,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45137,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(45137,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45137,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45137,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45137,36,'cp 0','g1f3 e7e6 a2a4 b8c6 e1g1 f8e7 c4b3 c6a5 b3a2 a7a6 c2c3 e8g8 c1f4 c8d7 b1d2 a5c6 h2h3 b7b5 f1e1 h7h6 a2b3 d8c7 f4g3 f8d8 d2f1 d7e8 b3c2 b5b4 f1e3 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(20019,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(20019,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20019,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20019,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20019,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20019,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(20019,36,'cp 3','f1e2 e7e5 e1g1 g8f6 c2c4 f8e7 b1c3 e8g8 a1b1 c8d7 a2a3 a7a5 c1d2 f6e8 c3d5 e8c7 h2h3 c7e6 b2b3 g8h8 b3b4 c5b4 a3b4 c6b4 d2b4 a5b4 b1b4 d7c6 d3d4 e5d4 f3d4 e7g5 d4f3 g5f6 e2d3 h8g8 b4b1 e6c5 d3c2');
INSERT INTO evaluations VALUES(20018,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(20018,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20018,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20018,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(20018,36,'cp 12','g8f6 e1g1 c8d7 a2a3 e7e5 c2c4 f8e7 b1c3 e8g8 a1b1 a7a5 c1d2 h7h6 h2h3 a8b8 b2b3 f6e8 c3d5 e8c7 b3b4 a5b4 a3b4 b7b5 b4c5 d6c5 d5c7 d8c7 c4b5 c6d4 f3d4 c5d4 b5b6 b8b6 d2a5 b6b1 a5c7 b1d1 e2d1 e7f6 d1g4 d7g4');
INSERT INTO evaluations VALUES(13328,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(13328,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13328,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13328,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13328,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13328,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(13328,36,'cp 0','e1g1 e7e5 c2c4 f8e7 b1c3 e8g8 a1b1 c8d7 a2a3 a7a5 c1d2 a8b8 h2h3 h7h6 b2b3 f6e8 c3d5 e8c7 b3b4 c5b4 a3b4 c7d5 c4d5 c6b4 d2b4 a5b4 b1b4 b7b5 d1b3 e7f6 f1b1 d8e7 f3e1 f6g5 e1f3');
INSERT INTO evaluations VALUES(13327,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13327,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13327,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13327,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(13327,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(13327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13327,36,'cp 0','e7e5 c2c4 f8e7 b1c3 e8g8 a2a3 c8d7 h2h3 h7h6 c1e3 a7a5 f3d2 c6d4 a3a4 d7c6 c3d5 c6d7 d5c3');
INSERT INTO evaluations VALUES(22036,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(22036,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(22036,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(22036,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(22036,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(22036,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(22036,36,'cp 43','f1b5 g8f6 b1c3 f8e7 e4e5 f6g4 e5d6 d8d6 d4d6 e7d6 c3e4 e8g8 e4d6 c7d6 e1g1 g4e5 f3e5 d6e5 c2c3 f8d8 f2f4 h7h6 f1e1 c8e6 f4e5 d8d5 b5f1 a8d8 b2b3 g7g5 c3c4 d5d7 c1e3 b7b6 c4c5 b6c5 e3c5 d7d2 e1e2');
INSERT INTO evaluations VALUES(21865,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(21865,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(21865,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(21865,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(21865,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21865,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(21865,36,'cp 40','g8f6 c1f4 f8e7 b1c3 e8g8 d4d2 c8d7 e1c1 a7a6 b5d3 f8e8 h1e1 b7b5 e4e5 d6e5 f3e5 c6e5 f4e5 h7h6 e5f6 e7f6 d3e4 d7g4 e4a8 g4d1 d2d8 e8d8 c3d5 d1c2 d5f6 g7f6 a8b7 d8b8 b7c6 b8b6 c6b5 b6b5 c1c2 b5f5 e1e2 g8g7 a2a4 f5d5 e2e8');
INSERT INTO evaluations VALUES(55079,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(55079,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(55079,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(55079,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(55079,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(55079,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(55079,36,'cp -86','d5e4 f3e4 g8f6 e4d3 c8d7 g1f3 b8c6 d3e3 e7e5 f3e5 c6e5 e3e5 f8e7 f1b5 e8g8 b5d7 d8d7 e1g1 a8e8 h2h3 e7d6 e5c3 f6e4 c3f3 e4g5 f3g4 g5e6 d2d4 f7f5 g4d1 f5f4 b1d2 e6d4 d2f3 d4f3 d1f3 d7f5 f3b7 e8e2 c2c3 f5e5 a1b1 h7h5 b7f3 f8f5 c1e3');
INSERT INTO evaluations VALUES(55290,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(55290,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(55290,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(55290,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(55290,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(55290,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(55290,36,'cp -92','f3e4 g8f6 e4e3 e7e5 f1c4 f8d6 g1f3 e8g8 d2d3 h7h6 b1d2 b8c6 e1g1 c6b4 c4b3 a7a5 a2a3 b4d5 e3e2 a5a4 b3a2 d5f4 e2d1 f8e8 d2c4 f6d5 c1f4 d5f4 g2g3 f4h3 g1g2 c8g4 c4e3 g4d7 a2d5 c7c6 d5e4 d6c5 f3d2 h3g5 d1h5 d8f6 g2g1 g5e4 d2e4');
INSERT INTO evaluations VALUES(55281,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(55281,31,'cp -102',NULL);
INSERT INTO evaluations VALUES(55281,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(55281,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(55281,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(55281,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(55281,36,'cp -84','g8f6 e4d3 c8d7 g1f3 b8c6 d3e3 e7e5 f3e5 c6e5 e3e5 f8e7 d2d3 e8g8 f1e2 f8e8 b1c3 d7c6 e5f5 e7b4 c1e3 f6d5 e1g1 d5c3 b2c3 b4c3 a1d1 g7g6 f5c5 d8d5 c5d5 c6d5 a2a4 a8d8 e2f3 d5f3 g2f3 d8d6 d1b1 b7b6 b1b5 a7a6 b5b3 d6c6 f1b1 f7f6 g1f1 g8f7 f3f4 f7e7');
INSERT INTO evaluations VALUES(47403,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(47403,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(47403,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(47403,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(47403,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(47403,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(47403,36,'cp -86','e4d3 c8d7 g1f3 b8c6 d3e3 e7e5 f3e5 c6e5 e3e5 f8e7 f1b5 e8g8 b5d7 d8d7 e1g1 f8e8 e5c3 f6e4 c3f3 e7c5 c2c3 c5f2 g1h1 f2b6 d2d4 c7c5 b1d2 e4d2 c1d2 c5d4 a1d1 a8d8 c3d4 d7d5 f3g3 e8e6 d2c3 f7f6 f1e1 d8e8 a2a4 e6e1 d1e1 e8e1 g3e1 g8f7 e1e2 b6d4');
INSERT INTO evaluations VALUES(10024,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(10024,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(10024,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(10024,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(10024,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(10024,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(10024,36,'cp 34','e8g8 f1e1 d7d6 d2d4 c5b6 h2h3 c8d7 a2a4 a7a6 b5c4 f8e8 d4d5 c6e7 a4a5 b6a7 c4f1 c7c6 c3c4 b7b6 b2b4 b6a5 b4a5 a7c5 c1d2 e7g6 b1c3 a8b8 d1c2 h7h6 c3a4 c5a7 e1b1 c6d5 c4d5 d7b5 f1d3 d8d7 d3b5 b8b5 b1b5');
INSERT INTO evaluations VALUES(4268,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(4268,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(4268,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(4268,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(4268,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(4268,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(4268,36,'cp 34','f1e1 d7d6 d2d4 c5b6 h2h3 c8d7 b5a4 f8e8 a4c2 e5d4 c3d4 c6b4 b1c3 b4c2 d1c2 h7h6 c1f4 b6a5 a1d1 b7b5 e1e3 a8c8 c3e2 a5b6 e2g3 d7e6 b2b3 d6d5 e4e5 f6e4 g3e4 d5e4 c2e4 d8d7 e4d3 e8d8 f4g3 d7d5 d3d2');
INSERT INTO evaluations VALUES(4269,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(4269,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(4269,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(4269,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(4269,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(4269,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(4269,36,'cp 36','d7d6 d2d4 c5b6 h2h3 c8d7 b5a4 f8e8 a4c2 e5d4 c3d4 c6b4 b1c3 b4c2 d1c2 h7h6 c1f4 c7c6 a1d1 d6d5 e4e5 f6h7 c3e2 h7f8 e2g3 f8g6 f4e3 b6c7 g3f5 d7e6 a2a3 e6f5 c2f5 d8e7 f5c2 g6h4 f3h4');
INSERT INTO evaluations VALUES(3781,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3781,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3781,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(3781,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(3781,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(3781,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(3781,36,'cp 28','d2d4 c5b6 h2h3 c8d7 a2a4 a7a6 b5c4 f8e8 f3g5 e8f8 c1e3 d8e7 g5f3 e5d4 e3d4 c6d4 c3d4 f6h5 a4a5 b6a7 b1c3 h5f4 c3d5 f4d5 c4d5 c7c6 d5c4 d7e6 d4d5 e6d7 d1d2 c6d5 c4d5 d7e6 e1e2 e6d5 d2d5 a8c8 a1d1 c8c5 d5d6 e7d6 d1d6 c5a5');
INSERT INTO evaluations VALUES(3776,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(3776,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(3776,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(3776,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(3776,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(3776,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3776,36,'cp 37','c5b6 h2h3 c8d7 a2a4 a7a6 b5c4 f8e8 f3g5 e8f8 c1e3 d8e7 g5f3 e5d4 e3d4 c6d4 c3d4 d7e6 b1d2 b6a5 c4e6 f7e6 e1e2 b7b5 d1c2 a5b4 c2b3 b4d2 f3d2 b5a4 a1a4 a8b8 a4b4 b8b4 b3b4 f6h5 g2g3 e7g5 g1h2 g5b5 b4c4 b5c4');
INSERT INTO evaluations VALUES(32401,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(32401,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(32401,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(32401,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(32401,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(32401,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(32401,36,'cp 47','h2h3 g4h5 d1e2 f8e7 e2b5 b8d7 b1c3 g8f6 f3e5 e8g8 e5d7 f6d7 b5d5 h5g6 f1e2 g6c2 d5b7 c7c5 b7d5 d7f6 d5d8 a8d8 d4c5 e7c5 e1g1 f6e4 c3e4 c2e4 e2b5 e4d3 b5d3 d8d3 c1f4 f8d8 f4e5 h7h6 a1e1 d3d2 e5c3');
INSERT INTO evaluations VALUES(32399,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(32399,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(32399,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(32399,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(32399,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(32399,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(32399,36,'cp 57','g4h5 d1e2 f8e7 e2b5 b8d7 b5d5 g8f6 d5b3 e8g8 f1e2 c7c5 e1g1 c5d4 f3d4 d7c5 b3c4 h5g6 e2f3 a8c8 c1e3 c5e4 c4b3 f6d5 d4f5 g6f5 f1d1 d5e3 d1d8 f8d8 f2e3 e7c5 b1c3 c5e3 g1h1 e4c3 b2c3 f5e6 c3c4 e3d4 a1d1 c8c4 b3b7 d8c8 f3g4 e6g4 h3g4 d4b6');
INSERT INTO evaluations VALUES(32396,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(32396,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(32396,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(32396,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(32396,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(32396,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(32396,36,'cp 44','d1e2 f8e7 e2b5 b8d7 b5d5 g8f6 d5b3 c7c5 f1e2 e8g8 c2c3 c5d4 c3d4 d7b8 e1g1 b8c6 c1f4 c6d4 f3d4 d8d4 f4e3 d4b4 g2g4 h5g6 b1c3 e7c5 e2f3 b4b3 a2b3 c5e3 f2e3 a8e8 a1a7 e8e3 a7b7 g6d3 f1f2 f6e4 b7e7 e4f2 e7e3 f2h3 g1h2 h3f2');
INSERT INTO evaluations VALUES(1091,30,'cp 302',NULL);
INSERT INTO evaluations VALUES(1091,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(1091,32,'cp 298',NULL);
INSERT INTO evaluations VALUES(1091,33,'cp 290',NULL);
INSERT INTO evaluations VALUES(1091,34,'cp 292',NULL);
INSERT INTO evaluations VALUES(1091,35,'cp 288',NULL);
INSERT INTO evaluations VALUES(1091,36,'cp 299','c4f7 e8d8 f7b3 d8e8 b1c3 d7d6 d2d3 c6d4 h2h3 d4b3 a2b3 h7h6 g5f3 e8f7 d3d4 f7g8 e1g1 g8h7 d1d3 h8g8 f1e1 c7c6 c3e2 f6d7 e2g3 g7g6 c1e3 a7a6 e1d1 f8g7 d4e5 d7e5 f3e5 d6e5 e3b6 g8f8 d3e3 f8f7 d1d8');
INSERT INTO evaluations VALUES(1069,30,'cp 279',NULL);
INSERT INTO evaluations VALUES(1069,31,'cp 290',NULL);
INSERT INTO evaluations VALUES(1069,32,'cp 329',NULL);
INSERT INTO evaluations VALUES(1069,33,'cp 334',NULL);
INSERT INTO evaluations VALUES(1069,34,'cp 330',NULL);
INSERT INTO evaluations VALUES(1069,35,'cp 305',NULL);
INSERT INTO evaluations VALUES(1069,36,'cp 313','e8d8 f7b3 d8e8 e1g1 h7h6 g5f3 e8d8 c2c3 d7d6 d2d4 a7a5 f1e1 c8d7 b1a3 d8c8 a3c4 g7g5 a2a4 e7e8 b3c2 g5g4 f3h4 e8h5 h4f5 c8b8 b2b4 a5b4 d4d5 c6e7 f5e7 f8e7 c3b4 g4g3 f2g3 h5d1 e1d1 h6h5 c4a5 f6g4 d1f1 h5h4 g3h4 e7h4 g2g3 h4e7 f1f7');
INSERT INTO evaluations VALUES(2300,30,'cp 309',NULL);
INSERT INTO evaluations VALUES(2300,31,'cp 300',NULL);
INSERT INTO evaluations VALUES(2300,32,'cp 299',NULL);
INSERT INTO evaluations VALUES(2300,33,'cp 319',NULL);
INSERT INTO evaluations VALUES(2300,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(2300,35,'cp 333',NULL);
INSERT INTO evaluations VALUES(2300,36,'cp 323','f7b3 d8e8 e1g1 d7d6 c2c3 a7a5 d2d4 h7h6 g5f3 g7g5 b1a3 c8d7 c1e3 e8d8 a3c4 d8c8 d4e5 d6e5 b3a4 f6g4 d1d3 c6b8 a4d7 b8d7 a1d1 g4e3 f2e3 f8g7 f3d2 a5a4 b2b4 c8b8 d3d5 d7f6 d5e5 e7e5 c4e5 a8a6 e5c4');
INSERT INTO evaluations VALUES(2299,30,'cp 331',NULL);
INSERT INTO evaluations VALUES(2299,31,'cp 325',NULL);
INSERT INTO evaluations VALUES(2299,32,'cp 322',NULL);
INSERT INTO evaluations VALUES(2299,33,'cp 321',NULL);
INSERT INTO evaluations VALUES(2299,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(2299,35,'cp 305',NULL);
INSERT INTO evaluations VALUES(2299,36,'cp 324','d8e8 e1g1 d7d6 c2c3 h7h6 g5f3 g7g5 d2d4 a7a5 b1a3 c8d7 c1e3 e8d8 a3c4 d8c8 b3a4 e5d4 e3d4 g5g4 f3h4 c6d4 c3d4 e7e4 a4d7 c8d7 f1e1 e4d5 c4e3 d5f7 e3g4 f6d5 d1d3 f8g7 g2g3 a8e8 a1c1 d7d8 h4f5 e8e1 c1e1 a5a4 f5g7 f7g7');
INSERT INTO evaluations VALUES(56873,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(56873,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(56873,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(56873,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(56873,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(56873,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(56873,36,'cp 105','f8g7 b1c3 d7d6 f1d3 a7a6 c1g5 g8e7 d1d2 h7h6 g5e3 b8d7 e1g1 c7c5 d4c5 d6c5 a1d1 b7b6 e3f4 c8b7 f1e1 e7c6 e4e5 g6g5 f4g3 g5g4 c3e4 g4f3 e4d6 e8f8 d6b7 d8g5 d3e4 g5d2 d1d2 d7e5 b7c5 b6c5 g3e5 g7e5 e4c6 a8c8 c6b7 e5f4 b7c8 f4d2 e1d1 f8e7 c8a6 d2a5 c2c3 a5c7 g2f3');
INSERT INTO evaluations VALUES(50341,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(50341,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(50341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50341,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(50341,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50341,36,'cp 0','d7d5 g1f3 c7c5 e2e3 e7e6 f1e2 b7b6 c2c4 c8b7 c4d5 e6d5 b1c3 f8d6 b2b3 e8g8 e1g1 b8d7 c3b5 d6e7 c1b2 a7a6 b5c3 f8e8 a1c1 e7d6 f3h4 a8c8 h4f5 d6f8 f5g3 f8d6');
INSERT INTO evaluations VALUES(52745,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(52745,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(52745,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(52745,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(52745,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(52745,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(52745,36,'cp -12','d5c4 a2a4 g8f6 b1c3 f6d5 f4g3 d5c3 b2c3 d8a5 d1d2 c6c5 e2e4 c5d4 c3d4 c4c3 d2c2 e7e6 g1f3 f8b4 f1c4 b8d7 e1g1 d7b6 c4b3 e8g8 f1d1 c8d7 d4d5 e6d5 e4d5 a8c8 f3e5 g7g6 e5d7 b6d7 d5d6 f8e8 a1c1 a5c5 d1d5 c5c6');
INSERT INTO evaluations VALUES(52880,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(52880,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(52880,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(52880,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(52880,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(52880,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(52880,36,'cp -32','b1c3 g8f6 a2a4 f6d5 f4g3 d5c3 b2c3 d8a5 d1c2 c6c5 e2e3 c5d4 e3d4 c8e6 f1e2 b8d7 e2f3 e6f5 c2b2 e7e5 g1e2 f8e7 d4e5 e8g8 b2b5 a5b5 a4b5 d7c5 e1g1 f5e4 f3e4 c5e4 g3f4 a7a6 b5a6 a8a6 a1a6 b7a6 f1a1 f8a8 a1a4 a6a5 a4c4 e4c5 e2d4');
INSERT INTO evaluations VALUES(52878,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(52878,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(52878,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(52878,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(52878,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(52878,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(52878,36,'cp -13','g8f6 a2a4 f6d5 f4g3 d5c3 b2c3 d8a5 d1c1 b8d7 e2e3 d7b6 g3c7 e7e5 g1f3 e5d4 e3d4 c8e6 f1e2 f8e7 f3e5 a8c8 c7b6 a5b6 e1g1 b6a6 c1e3 e8g8 e2c4 e6c4 e5c4 a6c4 e3e7 c4c3 e7b7 c3d4 a4a5 f8d8 a1e1 d4d7 a5a6 g8f8 h2h4 h7h6 e1e5 d7b7 a6b7');
INSERT INTO evaluations VALUES(37805,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(37805,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(37805,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(37805,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(37805,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(37805,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(37805,36,'cp 70','c5d4 c3d4 g8f6 f1d3 e8g8 e1g1 b8c6 b1c3 d7d6 d4d5 c6b4 d3c4 c8g4 c1e3 g4f3 g2f3 a8c8 c4e2 f6d7 a2a3 g7c3 b2c3 b4a6 e3h6 f8e8 d1d4 f7f6 h6e3 a6c5 a1b1 g8g7 g1h1 g7f7 f3f4 d8a5 f2f3 a5a3 e2b5 a3a5 c3c4');
INSERT INTO evaluations VALUES(37871,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(37871,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(37871,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(37871,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(37871,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(37871,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(37871,36,'cp 82','c3d4 d7d5 e4e5 a7a6 f1d3 f7f6 e1g1 b8c6 b1d2 f6e5 f3e5 c6e5 d4e5 g8h6 d1a4 d8d7 a4d4 h6f5 d4f4 d7c6 d2f3 e8g8 f1e1 c8e6 h2h3 d5d4 d3f5 f8f5 f3d4 f5f4 d4c6 f4f7 c6d4 e6d5 f2f3 a8f8 c1g5 h7h6');
INSERT INTO evaluations VALUES(37867,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(37867,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(37867,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(37867,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(37867,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(37867,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(37867,36,'cp 75','d7d5 e4e5 g8h6 f1b5 c8d7 b5d7 d8d7 e1g1 e7e6 b1c3 h6f5 c3a4 e8g8 g2g4 f5e7 c1g5 e7c6 a4c5 d7c7 d1d2 b8d7 c5d7 c7d7 d2f4 d7c7 f4e3 f7f6 e5f6 g7f6 e3e6 c7f7 e6f7 f8f7 g5e3 a8e8 a1d1 h7h5 g4h5 g6h5 g1h1 f7g7 f1g1 g7g1 d1g1 g8f7 g1g3 f7e6 h1g2 e6f5 h2h3');
INSERT INTO evaluations VALUES(38015,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(38015,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(38015,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(38015,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(38015,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(38015,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(38015,36,'cp 78','e4e5 a7a6 f1e2 b8c6 e1g1 c8g4 b1d2 g8h6 h2h3 g4f3 d2f3 e8g8 d1b3 d8d7 f1d1 f7f6 e5f6 e7f6 e2d3 h6f7 c1e3 g7h6 e3h6 f7h6 a1c1 g8g7 c1c5 f8d8 a2a3 h6f7 h3h4 a8c8 d1c1 c6e7 g2g3 f7d6 c1e1');
INSERT INTO evaluations VALUES(38017,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(38017,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(38017,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(38017,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(38017,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(38017,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(38017,36,'cp 58','g8h6 f1b5 c8d7 b5d7 d8d7 b1c3 e7e6 e1g1 h6f5 c3a4 e8g8 g2g4 f5e7 a4c5 d7c7 c1e3 b8d7 c5d3 a8c8 a1c1 c7a5 d1e2 c8c1 f1c1 f8c8 c1a1 a5a6 h2h3 e7c6 e2d2 a6a5 d2d1 a5a6 h3h4 f7f6 e5f6 d7f6 f3g5 c6d8 g1g2 f6g4 d1g4 a6d3 g5e6 g7f6 h4h5 c8c2');
INSERT INTO evaluations VALUES(41215,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(41215,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(41215,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(41215,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(41215,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(41215,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(41215,36,'cp 53','d1d2 h7h6 g5f4 b8c6 e1c1 a7a6 f2f3 b7b5 f4e3 e7e6 g2g4 c8b7 h2h4 f6d7 d4d5 c6e5 f1e2 b5b4 c3a4 e6d5 e4d5 d7f6 c2c4 b4c3 a4c3 c7c6 d5c6 b7c6 d2d6 d8d6 d1d6 f6d7 c3d5 c6d5 d6d5 e8g8 e3d4 f8e8');
INSERT INTO evaluations VALUES(41216,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(41216,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41216,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(41216,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41216,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(41216,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(41216,36,'cp 40','h7h6 g5f4 b8c6 e1c1 a7a6 f2f3 b7b5 f4e3 e7e6 g2g4 h6h5 g4h5 f6h5 c3e2 c8d7 g1h3 e6e5 c2c3 d8f6 h3g5 b5b4 c3b4 c6d4 e2d4 e5d4 e3d4 f6d4 d2d4 g7d4 d1d4 h5f4 h2h4');
INSERT INTO evaluations VALUES(41171,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(41171,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(41171,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(41171,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(41171,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(41171,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(41171,36,'cp 38','g5f4 b8c6 e1c1 a7a6 f2f3 b7b5 f4e3 e7e6 c1b1 b5b4 c3e2 a6a5 e2c1 h6h5 g1h3 c8a6 f1a6 a8a6 h1e1 e8g8 e3h6 f8e8 h6g7 g8g7 c1e2 e6e5 h3g5 c6d4 e2d4 e5d4 d2d4 e8e7 f3f4 d8a8 f4f5 e7e5');
INSERT INTO evaluations VALUES(49633,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(49633,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(49633,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(49633,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(49633,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(49633,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(49633,36,'cp -67','d7d5 f3e5 f8d6 d3e2 f6e4 e5f3 e8g8 e1g1 f8e8 d2d4 c8f5 c2c4 d5c4 e2c4 b8c6 b1c3 d8d7 c1e3 h7h6 a1c1 e4c3 c1c3 a8d8 a2a3 f5g4 c4a2 d7f5 a2b1 f5h5 h2h3 g4f3 d1f3 h5f3 g2f3 c6e7 c3b3 b7b6 b1e4 e7g6 f1d1');
INSERT INTO evaluations VALUES(48832,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(48832,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(48832,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(48832,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(48832,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(48832,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(48832,36,'cp -66','f3e5 f8d6 d3e2 f6e4 e5f3 e8g8 e1g1 h7h6 d2d3 e4f6 c2c4 f8e8 b1c3 b8c6 c4d5 c6b4 c3b5 c8g4 c1d2 f6d5 a2a3 b4c6 b5c3 d6c5 h2h3 g4h5 a1c1 c5b6 c3e4 h5f3 e2f3 c6d4 e4c5 b6c5 c1c5 c7c6');
INSERT INTO evaluations VALUES(48757,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(48757,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(48757,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(48757,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(48757,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(48757,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(48757,36,'cp -66','f8d6 d3e2 f6e4 e5f3 e8g8 e1g1 h7h6 d2d3 e4f6 c2c4 c7c5 c4d5 f6d5 b1c3 d5c3 b2c3 b8c6 d3d4 c5d4 f3d4 c8d7 e2f3 c6d4 c3d4 a8b8 a2a4 f8e8 c1e3 a7a5 h2h3 b7b5 a4b5 d7b5 f3e2 d6b4 e2b5 b8b5 d1d3 b5b6 d3c4');
INSERT INTO evaluations VALUES(48638,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48638,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(48638,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(48638,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48638,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48638,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(48638,36,'cp 0','b1c3 f8b4 g1f3 e8g8 f1e2 b8c6 c1e3 c8g4 e1g1 d5c4 e2c4 d8d6 a1c1 a8e8 h2h3 g4h5 c3b5 d6d7 b5c3');
INSERT INTO evaluations VALUES(48636,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48636,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48636,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48636,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(48636,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48636,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48636,36,'cp 0','f8b4 g1f3 e8g8 f1d3 b8c6 e1g1 c8g4 c1e3 d5c4 d3c4 d8d6 a1c1 a8e8 h2h3 g4h5 c3b5 d6d7 b5c3');
INSERT INTO evaluations VALUES(11707,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(11707,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(11707,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(11707,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(11707,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(11707,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(11707,36,'cp 12','a4b3 f8c5 c3d5 e8g8 d2d3 f6d5 b3d5 a8b8 e1g1 d8e8 c1e3 c5e3 f2e3 d7d6 d1d2 c6e7 d5b3 h7h6 f1f2 e7g6 d2c3 c7c5 c3a5 e8c6 b3d5 c6d7 a1f1 g6e7 c2c4 e7d5 c4d5 c5c4 d3c4 b5c4 f3d2 b8b2 d2c4 b2f2');
INSERT INTO evaluations VALUES(9655,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9655,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(9655,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9655,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9655,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9655,36,'cp 30','f6e4 f2e3 d8h4 g2g3 e4g3 h2g3 h4d4 e3f3 e8g8 h1h4 e5e4 f3g2 d4c4 f7g5 c4d4 d1e2 h7h6 h4e4 d4c5 d2d4 c6d4 c1e3 b7b6 e3d4 c5g5 d4g7 f8f5 b1d2 c8a6 g7h6 g5h6 e2a6 h6d2 a6e2 d2e2 e4e2 a8f8 a1d1 d7d6 d1d4 g8g7 d4h4 g7g6 c2c4 g6g7 h4e4 f8f7 b2b3 f5f1 e4e7');
INSERT INTO evaluations VALUES(9660,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9660,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9660,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9660,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9660,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9660,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9660,36,'cp 0','f2e3 d8h4 g2g3 e4g3 h2g3 h4d4 e3f3 e8g8 h1h4 e5e4 f3g2 d4c4 f7g5 c4d4 h4f4 f8f4 g3f4 h7h6 c2c3 d4d3 d1h5 h6g5 h5e8 g8h7 e8h5');
INSERT INTO evaluations VALUES(9659,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9659,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9659,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9659,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9659,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9659,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9659,36,'cp 0','d8h4 g2g3 e4g3 h2g3 h4d4 e3f3 e8g8 h1h4 e5e4 f3g2 d4c4 f7g5 c4d4 h4f4 f8f4 g3f4 h7h6 d1h5 h6g5 h5e8 g8h7 b1c3 d4f6 c3e4 f6f4 d2d3 f4g4 g2h2 g4h4 h2g1 h4g4');
INSERT INTO evaluations VALUES(529,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(529,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(529,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(529,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(529,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(529,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(529,36,'cp 0','g2g3 e4g3 h2g3 h4d4 e3f3 e8g8 h1h4 e5e4 f3g2 d4c4 f7g5 c4d4 h4f4 f8f4 g3f4 h7h6 d1h5 h6g5 h5e8 g8h7 b1c3 d4f6 c3e4 f6f4 e8h5 h7g8 e4g5 d7d6 h5h7 g8f8 h7h8 f8e7 h8g7 e7d8 g7g8 d8e7');
INSERT INTO evaluations VALUES(528,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(528,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(528,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(528,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(528,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(528,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(528,36,'cp 0','e4g3 h2g3 h4d4 e3f3 e8g8 h1h4 e5e4 f3g2 d4c4 f7g5 c4d4 h4f4 f8f4 g3f4 h7h6 d1h5 h6g5 h5e8 g8h7 b1c3 d4f6 c3e4 f6f4 d2d3 f4g4 g2f2 g4h4 f2g1 h4g4');
INSERT INTO evaluations VALUES(49654,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(49654,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(49654,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(49654,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(49654,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(49654,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(49654,36,'cp -82','d2d4 e5d4 d1d4 b8c6 d4f2 d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 d5e4 f2c5 c8e6 g1e2 d8d5 c5d5 f6d5 f3e4 d5c3 e2c3 e8c8 a1d1 c6e5 f1e2 c7c6 h2h4 d8d1 e1d1 h8d8 d1c1 c8c7 h1d1 d8d1 c1d1 a7a5 d1d2 b7b5 a2a3 c7d6 d2e3 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(49592,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(49592,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(49592,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(49592,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(49592,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(49592,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(49592,36,'cp -68','e5d4 d1d4 b8c6 d4f2 d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 d5e4 f2c5 c8e6 g1e2 d8d5 c5d5 f6d5 f3e4 d5c3 e2c3 e8c8 f1e2 h7h5 a1d1 h5h4 d1d8 h8d8 a2a3 c6e5 e1f2 c7c6 f2e3 c8c7 b2b3 h4h3 g2g3 b7b5 c3d1 a7a5 d1b2 d8a8 a3a4 b5b4 h1d1 a8h8 b2d3 e5d3 d1d3');
INSERT INTO evaluations VALUES(49777,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(49777,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(49777,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(49777,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(49777,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(49777,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(49777,36,'cp -74','d1d4 b8c6 d4f2 d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 d5e4 f2c5 c8e6 g1e2 d8d5 c5d5 f6d5 f3e4 d5c3 e2c3 e8c8 f1e2 c6e5 a1d1 d8e8 c3b5 a7a5 b5d4 c7c6 e1g1 e8e7 a2a3 f7f6 b2b4 c8c7 b4a5 h8a8 a5a6 b7a6 d4b3 a6a5 h2h3 e6c4 f1e1 c4b3 c2b3');
INSERT INTO evaluations VALUES(49770,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(49770,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(49770,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(49770,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(49770,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(49770,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(49770,36,'cp -67','b8c6 d4f2 d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 d5e4 f2c5 c8e6 g1e2 d8d5 c5d5 f6d5 f3e4 d5c3 e2c3 e8c8 f1e2 c6e5 a1d1 d8e8 c3b5 a7a6 b5d4 e6g4 d1d2 f7f6 e2g4 e5g4 d2e2 e8e7 h2h3 g4e5 d4f5 e7d7 e2d2 g7g6 f5e3 h8d8 d2d7 e5d7 h1f1 d8e8 f1f4 h7h5 e3d1 f6f5');
INSERT INTO evaluations VALUES(16690,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(16690,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(16690,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(16690,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(16690,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(16690,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(16690,36,'cp -62','d4f2 d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 d5e4 f2c5 c8e6 g1e2 d8d5 c5d5 f6d5 f3e4 d5c3 e2c3 e8c8 f1e2 f7f6 a1d1 d8d1 e1d1 h8d8 d1c1 c6e5 h1d1 d8e8 b2b3 c7c6 a2a3 c8c7 c1d2 a7a5 d1f1 e5d7 h2h4 b7b5 e2d3 b5b4 a3b4 a5b4 c3e2');
INSERT INTO evaluations VALUES(52108,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52108,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52108,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(52108,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52108,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(52108,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(52108,36,'cp 11','d2d3 b8c6 b1c3 d7d6 g1e2 a8b8 g2g3 a7a6 f1g2 b7b5 e1g1 g8f6 b2b3 f8e7 c3d5 e8g8 h2h3 b5c4 d5e7 d8e7 b3c4 c8d7 f2f4 h7h6 f4f5 b8b7 g1h2 f8b8 g3g4 e7d8 d1d2 d8a5 d2a5 c6a5 e2c3 a5c6 c3d5');
INSERT INTO evaluations VALUES(52106,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(52106,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(52106,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(52106,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52106,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(52106,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52106,36,'cp 2','d7d6 b1c3 b8c6 g2g3 h7h5 h2h4 g7g6 a2a3 f8h6 f1h3 h6c1 a1c1 c8h3 g1h3 d8d7 c1b1 g8e7 e1f1 e8g8 f1g2 g8g7 b2b4 b7b6 b4b5 c6d4 a3a4 a7a6 h1f1 a6b5 c3b5 e7c6 b5c3 a8a6 g2h2 c6b4 f2f4 d7g4');
INSERT INTO evaluations VALUES(53883,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(53883,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(53883,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(53883,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53883,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53883,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53883,36,'cp 12','g2g3 b8c6 f1g2 a7a6 b1c3 a8b8 g1e2 b7b5 e1g1 g8f6 b2b3 f8e7 f2f4 e8g8 f4f5 b5c4 b3c4 c6d4 e2d4 e5d4 c3e2 c8b7 e2f4 f6d7 a1b1 d7e5 h2h4 d8a5 a2a4 b7c6 b1b8 f8b8 f4d5');
INSERT INTO evaluations VALUES(46637,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(46637,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(46637,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(46637,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(46637,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(46637,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(46637,36,'cp 50','f8g7 g1f3 e8g8 d1d2 d6d5 e4d5 f6d5 g5h6 d5c3 h6g7 g8g7 d2c3 d8d6 e1c1 b8d7 h2h4 d7f6 f3e5 h7h5 f2f3 c8e6 f1c4 e6c4 c3c4 d6d5 b2b3 d5c4 b3c4 f8d8 h1e1 f6d7 a2a4 e7e6 a4a5 d7e5 e1e5 b7b6 c1b2 g7f6 e5e4 d8e8 d1e1 a8d8 b2c3');
INSERT INTO evaluations VALUES(51902,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(51902,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(51902,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(51902,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(51902,34,'cp -119',NULL);
INSERT INTO evaluations VALUES(51902,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(51902,36,'cp -89','b8c6 g1f3 d8c7 d1e2 g7g5 h2h3 f8g7 d2d4 c5d4 c1g5 c6e5 b1d2 e5g6 d2b3 a7a6 b3d4 d7d5 c4b3 h7h6 g5d2 e6e5 c2c3 g8e7 d4c2 e8g8 e1c1 f8d8 h3h4 a6a5 h4h5 g6f8 c2e3 a5a4 b3c2 a4a3 b2b3 c8e6');
INSERT INTO evaluations VALUES(18706,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(18706,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(18706,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(18706,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(18706,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(18706,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(18706,36,'cp -105','g1f3 d8c7 d1e2 g7g5 h2h3 f8g7 d2d4 c6d4 f3d4 c5d4 e1g1 c7e5 e2e5 g7e5 c1g5 b7b6 f2f4 e5d6 f4f5 c8b7 f5e6 f7e6 b1a3 d6e7 g5e7 e8e7 a3b5 g8f6 b5d4 h8g8 f1f2 f6e4 f2e2 a8f8 a1e1 f8f4 c2c3 h7h5 c4d3 e4g3 e2f2 f4f6 f2f6 e7f6 d4c2 a7a5 c2e3 d7d5 d3c2');
INSERT INTO evaluations VALUES(18705,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(18705,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(18705,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(18705,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(18705,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(18705,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(18705,36,'cp -110','d8c7 d1e2 g7g5 h2h3 f8g7 d2d4 c6d4 f3d4 c5d4 c1g5 c7e5 e2e5 g7e5 b1d2 d7d5 c4d3 f7f6 g5h4 e5c7 e1g1 e8f7 f2f4 g8e7 d2b3 e7c6 a1d1 c7d6 c2c4 d4c3 b2c3 a7a5 a2a4 c6e7 d3b5 h8d8 c3c4 b7b6 c4d5 e7d5');
INSERT INTO evaluations VALUES(2105,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(2105,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(2105,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(2105,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(2105,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(2105,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(2105,36,'cp -102','d1e2 g7g5 h2h3 f8g7 d2d4 c6d4 f3d4 c5d4 c1g5 c7e5 e2e5 g7e5 b1d2 d7d5 c4d3 e5c7 d2b3 f7f6 g5h4 e8f7 b3d4 g8e7 d4b5 c7b8 b5c3 a7a6 a1d1 c8d7 c3e2 h7h5 c2c3 d7b5 e1g1 h8g8 f2f4 b5d3 d1d3 e7f5 h4f2 h5h4 f1d1 b8d6 e2d4 f5d4 d3d4 d6c5');
INSERT INTO evaluations VALUES(11857,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(11857,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(11857,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(11857,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(11857,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(11857,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11857,36,'cp 29','d2d4 f6d7 e1g1 a7a5 f1e1 f8e7 d4e5 d6e5 c1e3 e8g8 c3a4 c8a6 f3d2 a6b5 c2c4 b5a4 d1a4 d7c5 e3c5 e7c5 d2b3 c5d4 a1d1 c6c5 b3c5 d8f6 c5d3 f8b8 e1e2 g7g6 a4c2 a5a4 c4c5 f6c6 h2h3 b8d8');
INSERT INTO evaluations VALUES(11852,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11852,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(11852,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(11852,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(11852,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(11852,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11852,36,'cp 24','f6d7 e1g1 a7a5 c1e3 f8e7 b2b3 e8g8 h2h3 d8e8 d4e5 d7e5 f3e5 d6e5 d1h5 e7b4 c3a4 e8e7 a1d1 f7f5 f2f3 h7h6 e3f2 c8e6 f2h4 e7f7 h5f7 f8f7 h4g3 f5e4 f3e4 b4d6 g3f2 a8f8 a4c5 d6c5 f2c5 f7f1 d1f1 f8f1 g1f1');
INSERT INTO evaluations VALUES(6233,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(6233,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(6233,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(6233,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(6233,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(6233,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(6233,36,'cp 118','e5d4 c3d4 c5b6 b1c3 d7d6 h2h3 g8f6 e1g1 e8g8 f1e1 c6e7 a2a4 b6a5 c1d2 c7c6 a1c1 d6d5 e4e5 f6e8 c4a2 c8f5 f3h4 f5h7 a2b1 h7b1 c1b1 e8c7 d1g4 d8c8 g4g3 e7f5 h4f5 c8f5 b2b4 a5b6 d2h6 c7e6 b1d1 f5h7 h6e3 f7f5 e5f6 b6c7 g3g4 f8f6 b4b5 a8e8');
INSERT INTO evaluations VALUES(6252,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(6252,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(6252,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(6252,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(6252,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(6252,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(6252,36,'cp 132','c3d4 c5b6 h2h3 g8f6 b1c3 d7d6 e1g1 e8g8 f1e1 c6e7 a2a4 b6a5 c1d2 c7c6 a1c1 a5c3 d2c3 d6d5 e4e5 f6e4 c4d3 c8f5 f3h4 f5h7 f2f3 e4g5 f3f4 g5e4 d1g4 f7f5 g4h5 a7a5 g1h2 a8c8 h5e2 g8h8 c1d1 d8c7 h2g1');
INSERT INTO evaluations VALUES(6248,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(6248,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(6248,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(6248,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(6248,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(6248,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(6248,36,'cp 119','c5b6 h2h3 g8f6 b1c3 d7d6 e1g1 e8g8 f1e1 c6e7 a2a4 b6a5 c1d2 c7c6 d1c2 d6d5 e4e5 c8f5 c4d3 f5d3 c2d3 f6d7 c3e2 a5d2 d3d2 f8e8 a4a5 d7f8 f3h4 d8d7 f2f4 f8g6 h4g6 f7g6 a1a3 b7b6 e1a1 b6a5 d2a5 e8b8 b2b4 e7f5 a5c5 d7b7 a3b3 b7b6 g1h2 b6c5 d4c5');
INSERT INTO evaluations VALUES(6198,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(6198,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(6198,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(6198,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(6198,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(6198,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(6198,36,'cp 139','b1c3 d7d6 h2h3 g8f6 e1g1 e8g8 f1e1 c6e7 a2a4 c7c6 b2b4 f8e8 a4a5 b6c7 c4d3 e7g6 c1e3 a7a6 a1c1 c8d7 d3b1 a8c8 d1c2 b7b5 c1d1 d8e7 c2d3 c7b8 d3d2 d7e6 d4d5 e6d7 d5c6 d7c6 c3d5 c6d5 e4d5');
INSERT INTO evaluations VALUES(6195,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(6195,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(6195,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(6195,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(6195,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(6195,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(6195,36,'cp 103','d7d6 h2h3 g8f6 e1g1 e8g8 f1e1 c6e7 b2b4 c8e6 c4d3 e7g6 a2a4 a7a5 b4b5 f6h5 d3c2 h5f4 c3e2 f4e2 d1e2 f8e8 c1e3 c7c6 a1b1 e6d7 b5c6 b7c6 e1d1 b6c7 b1b3 a8b8 b3b8 d8b8 e2d3 b8c8 d1b1 c6c5 c2b3 c5d4 f3d4');
INSERT INTO evaluations VALUES(5721,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(5721,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(5721,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(5721,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(5721,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(5721,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(5721,36,'cp 126','h2h3 g8f6 e1g1 e8g8 f1e1 c6e7 a2a4 c7c6 b2b4 d6d5 e4e5 f6e4 c3e4 d5c4 a4a5 b6c7 d1c2 c8f5 c2c4 d8d5 c4d5 e7d5 c1d2 f5e4 e1e4 c7d8 f3e1 f7f5 e4e2 d8e7 g2g3 a7a6 e1d3 a8d8 d3c5 e7c5 b4c5 d5c7 d2e3 g8f7 e2b2 c7b5 b2b4 g7g5 g1g2 f7e6 a1e1');
INSERT INTO evaluations VALUES(37589,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(37589,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(37589,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(37589,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(37589,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(37589,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(37589,36,'cp 141','b1c3 d7d6 e4e5 c8b7 f1b5 c7c6 b5d3 c6c5 e5d6 d8d6 c3b5 d6e6 c1e3 b8a6 e1g1 g8f6 d4c5 e8g8 b5d4 e6c8 c5c6 b7c6 d4c6 c8c6 d3a6 b6b5 d1e2 c6a6 a2a4 f6d5 a4b5 a6e6 a1a6 e6e4 c2c3 d5e3 e2e3 e4c2 e3d2 c2d2 f3d2 a8b8 c3c4 b8a8 d2b3');
INSERT INTO evaluations VALUES(37582,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(37582,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(37582,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(37582,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(37582,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(37582,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(37582,36,'cp 138','d7d6 e4e5 d6e5 f3e5 c8b7 f1b5 c7c6 d1f3 g8f6 e5c6 b8c6 b5c6 b7c6 f3c6 f6d7 c1f4 a8c8 c6a4 e8g8 a1d1 d7f6 e1g1 d8d7 a4b3 a7a6 f4e5 d7c6 d4d5 c6c4 d1d4 c4b3 a2b3 b6b5 g2g3 f8d8 f1d1 a6a5 d5d6 e7d6 d4d6 d8d6 d1d6');
INSERT INTO evaluations VALUES(37550,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(37550,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(37550,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(37550,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(37550,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(37550,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(37550,36,'cp 132','e4e5 d6e5 f3e5 c7c6 c1f4 g8f6 d1f3 c8e6 e5c6 b8c6 f3c6 e6d7 c6c7 e8g8 c7d8 f8d8 e1c1 a7a6 f4c7 d8c8 c7e5 b6b5 a2a3 c8d8 f2f3 h7h5 f1d3 d7c6 h2h4 e7e6 h1e1 a8a7 c3e2 g7h6 c1b1 f6d7 e5d6 a7b7 d6b4 d7b6 b4c5');
INSERT INTO evaluations VALUES(33169,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(33169,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(33169,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(33169,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(33169,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(33169,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(33169,36,'cp 197','f1d3 d7d5 g1f3 h4d8 e1g1 g8e7 d1e2 e8g8 e4e5 h7h6 c3d1 c7c5 c2c3 b4a5 d4c5 a5c7 b2b4 b7b6 c1e3 e7g6 e3d4 b8c6 d3g6 f7g6 d1e3 b6c5 d4c5 f8f7 e3c2 g6g5 h2h3 c7b6 c5b6 a7b6 f1e1 d8e8 b4b5 c6e7 f3d4 e7g6 a2a4');
INSERT INTO evaluations VALUES(33171,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(33171,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(33171,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(33171,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(33171,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(33171,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(33171,36,'cp 186','d7d6 g1f3 h4d8 e1g1 g8e7 c3a4 b4a5 a2a3 c7c6 b2b4 a5c7 e4e5 b8d7 d1e2 b7b6 f1d1 d6d5 c1f4 d7f8 f4g3 h7h6 d1e1 g7g5 a1d1 f8g6 c2c4 d5c4 d3g6 e7g6 a4c3 c8b7 c3e4 e8g8 e2c4 c6c5 e4f6 g8g7');
INSERT INTO evaluations VALUES(54076,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(54076,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(54076,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(54076,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(54076,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(54076,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(54076,36,'cp -73','b8c6 c1e3 c6e5 g1f3 e5f3 d1f3 g8e7 f1d3 e7c6 e1g1 f8e7 f1d1 e8g8 c2c3 e6e5 d3c4 d5d4 c3d4 e5d4 b1c3 e7c5 c3e4 c5e7 e3d4 c6d4 f3c3 c8g4 d1d4 d8b6 a2a3 a8c8 h2h3 g4e6 a1d1 e6c4 d4c4 c8c4 c3c4 b6b2 c4c7 b2a3 c7b7 g8h8 e4c3');
INSERT INTO evaluations VALUES(20209,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(20209,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(20209,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(20209,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(20209,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(20209,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(20209,36,'cp -46','b1d2 d8c7 f2f4 f8c5 d2b3 c5b6 f1d3 g8h6 g1f3 f7f6 e5f6 e8g8 d1e2 f8f6 c1d2 a7a5 c2c3 h6g4 b3d4 e6e5 h2h3 e5e4 h3g4 c8g4 d3e4 d5e4 e2e4 h7h6 e1g1 a8d8 d2e3 c6d4 e3d4 f6f4 e4e3 f4f3 d4b6 f3e3 b6c7 d8d5');
INSERT INTO evaluations VALUES(4915,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(4915,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(4915,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(4915,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(4915,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(4915,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(4915,36,'cp 75','c2c3 c5b6 a2a4 c8b7 a4b5 a6b5 a1a8 d8a8 b1a3 b7a6 e1g1 g8f6 d2d3 a8d8 b3d5 d8e7 c1g5 h7h6 g5h4 e7d6 d5c6 d7c6 a3c2 e8g8 h4g3 f6d7 d3d4 c6c5 f3e5 d7e5 g3e5 d6e6 d1d3 c5d4 c3d4 f8e8 f1a1 a6b7 c2e3 f7f6');
INSERT INTO evaluations VALUES(4917,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(4917,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(4917,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(4917,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(4917,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(4917,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(4917,36,'cp 80','c5b6 a2a4 b5a4 b3a4 g8e7 e1g1 e8g8 b1a3 d7d5 e4d5 d8d5 a4b3 d5d3 b3c2 d3d5 f1e1 c8e6 d2d3 a6a5 a3c4 f7f6 c4b6 c7b6 c1e3 e7f5 d3d4 a8d8 d4e5 f5e3 e1e3 d5d1 c2d1 c6e5 f3e5 f6e5 e3e5 e6f7 d1f3 h7h6 e5b5 d8d6 h2h3');
INSERT INTO evaluations VALUES(14755,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(14755,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(14755,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(14755,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(14755,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(14755,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(14755,36,'cp 56','f8e7 e2e3 e8g8 a2a3 h7h6 g5f4 e7d6 f4d6 c7d6 c4d5 f6d5 c3d5 e6d5 f1d3 c6a5 e1g1 c8e6 a1c1 a5c4 d3c4 d5c4 d4d5 e6d7 e3e4 b7b5 d1d2 a8c8 f1e1 a7a6 f3d4 d8g5 f2f4 g5g6 f4f5 g6g5 d2g5 h6g5');
INSERT INTO evaluations VALUES(8473,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(8473,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(8473,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(8473,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(8473,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(8473,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(8473,36,'cp 64','e2e3 e8g8 a1c1 h7h6 g5f4 e7d6 f4g3 d6g3 h2g3 d8e7 a2a3 f8d8 d1c2 c6b8 f1e2 b8d7 e1g1 d5c4 f1d1 a7a6 e2c4 b7b5 c4e2 c7c5 d4c5 e7c5 c3e4 c5c2 e4f6 g7f6 c1c2 c8b7 c2c7 b7f3 g2f3 d7e5 d1d8 a8d8 f3f4 d8d2 e2h5 e5g6 h5g6 f7g6 c7e7 d2b2 e7e6 g8g7 e6e7 g7g8 e7a7 b5b4 a7a6 b2b1 g1h2 b4a3 a6a3');
INSERT INTO evaluations VALUES(8472,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(8472,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(8472,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(8472,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(8472,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(8472,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(8472,36,'cp 56','e8g8 a1c1 h7h6 g5f4 e7d6 f4d6 c7d6 c4d5 f6d5 c3d5 e6d5 d1d2 c8f5 f1e2 a8c8 e1g1 d8b6 f3h4 f5d7 e2d3 c6e7 f1e1 c8c1 e1c1 f8c8 c1c8 d7c8 h2h3 g7g6 g2g4 c8d7 g1h2 g8g7 h4g2 g6g5 h2g3 d7b5 d3b1 e7g8 f2f3 b5a6 a2a3');
INSERT INTO evaluations VALUES(3320,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(3320,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(3320,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(3320,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(3320,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(3320,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(3320,36,'cp 56','a1c1 h7h6 g5f4 e7d6 f4g3 d6g3 h2g3 d8e7 a2a3 f8d8 d1c2 c6b8 f1e2 b8d7 e1g1 d5c4 f1d1 c7c5 d4c5 e7c5 c3e4 c5f8 c2c4 f6e4 c4e4 d7f6 d1d8 f8d8 e4d4 d8d4 f3d4 f6e8 f2f4 c8d7 e2f3 g8f8 d4b3 a8b8 b3c5 d7c8 e3e4 b7b6 c5d3 c8a6 d3b4 a6b7 g1f2 e8d6');
INSERT INTO evaluations VALUES(17518,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(17518,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(17518,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(17518,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(17518,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(17518,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(17518,36,'cp 26','e7e6 c1e3 d8c7 d1f3 f8d6 e1c1 d6e5 f3e2 b7b5 d4c6 c7c6 e3d4 e5d4 d1d4 g8e7 e2d2 c6c7 a2a3 a8b8 d2g5 e7g6 g5g3 c7g3 h2g3 g6e7 f1e2 c8b7 h1d1 b7c6 e4e5 a6a5 c3e4 e7f5 e4d6 f5d6 d4d6 f7f6 e5f6 g7f6 e2f3 c6f3 g2f3');
INSERT INTO evaluations VALUES(55246,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(55246,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(55246,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(55246,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(55246,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(55246,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(55246,36,'cp -120','e7e5 d4e5 d8d1 e1d1 e4f3 g1f3 b8c6 f1d3 c8g4 b1d2 e8c8 h1e1 f8c5 h2h3 g4h5 g2g4 h5g6 d2e4 c5b6 c1d2 h7h5 e4g5 g6d3 c2d3 d8d3 g5f7 h5g4 f7h8 g4f3 d1c2 d3d5 h8g6 f3f2 e1f1 c6e5 g6e5 d5e5 d2c3 e5e2 c2d3 e2e3 d3d2 e3h3 c3g7 h3g3');
INSERT INTO evaluations VALUES(56568,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(56568,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(56568,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(56568,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(56568,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(56568,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(56568,36,'cp -101','d4e5 d8d1 e1d1 e4f3 g1f3 b8c6 f1d3 c8g4 b1d2 e8c8 h1e1 f8c5 e1e4 g4f5 e4c4 f5d3 c2d3 d8d5 d2e4 c5b6 d1c2 g8e7 c1d2 h7h6 e4c3 d5d7 a1e1 e7f5 c3e2 h8d8 c4f4 g7g6 e2c1 h6h5 c2b1 c6d4 f3d4 d7d4 b1c2 d4f4 d2f4 b6d4 c1b3');
INSERT INTO evaluations VALUES(51951,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(51951,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51951,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(51951,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(51951,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51951,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51951,36,'cp 13','c5d4 f3d4 g8f6 e2e3 f8c5 f1e2 e8g8 b1c3 b8c6 e1g1 d8e7 d4b3 c5d6 c1d2 b7b6 f2f4 d6b4 a2a3 b4c3 d2c3 c8a6 d1c2 d7d5 c4d5 a6e2 d5d6 e7d6 c2e2 a8d8 a1d1 d6e7 e2f3 d8d1 f1d1');
INSERT INTO evaluations VALUES(52053,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(52053,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52053,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52053,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52053,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(52053,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52053,36,'cp 11','f3d4 g8f6 e2e3 f8c5 f1e2 e8g8 b1c3 b8c6 e1g1 d8e7 b2b3 f8d8 c1b2 d7d5 c4d5 c5d4 e3d4 f6d5 c3d5 e6d5 a1c1 c8f5 f1e1 e7d6 e2f1 d8e8 e1e8 a8e8 c1c5 a7a6 h2h3 h7h6 a2a3 e8d8 d1d2 f5e4 f1e2 f7f5 d2e3');
INSERT INTO evaluations VALUES(52047,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(52047,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52047,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52047,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52047,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(52047,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52047,36,'cp 9','g8f6 e2e3 f8c5 b1c3 b8c6 f1e2 e8g8 e1g1 h7h6 c3a4 c5e7 d4c6 d7c6 d1c2 d8c7 b2b3 b7b6 f1d1 c6c5 e2f3 c8b7 f3b7 c7b7 c1b2 a8d8 a4c3 d8d1 a1d1 f8d8 h2h3 d8d1 c3d1 f6d7 f2f3 b7c7 d1c3 e7f6 g1h1 d7e5 c3b5');
INSERT INTO evaluations VALUES(44276,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(44276,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(44276,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(44276,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44276,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(44276,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44276,36,'cp 12','e2e3 f8c5 f1e2 e8g8 b1c3 b8c6 e1g1 h7h6 c3a4 c5e7 d4c6 d7c6 d1c2 b7b6 b2b3 d8c7 c1b2 c8b7 e2f3 a8d8 a1d1 c6c5 f3b7 c7b7 h2h3 d8d1 f1d1 f8d8 a4c3 d8d1 c3d1 a7a6 f2f3 f6d7 c2d3 b6b5 d1f2 b7c6 g1f1 e7f6');
INSERT INTO evaluations VALUES(44274,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44274,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44274,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(44274,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(44274,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44274,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44274,36,'cp 11','f8c5 f1e2 d7d5 c4d5 f6d5 d4b3 c5b6 c1d2 e8g8 b1c3 b8c6 e1g1 d8h4 g2g3 h4e7 c3d5 e6d5 d2c3 f8d8 a2a4 a7a5 f1e1 c8f5 e2d3 f5e4 d3e4 e7e4 b3d2 e4f5 d2f3 a8c8 d1e2 f5g6 a1d1 h7h6 e2b5 g6e4');
INSERT INTO evaluations VALUES(57498,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(57498,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(57498,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(57498,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(57498,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(57498,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(57498,36,'cp 46','e2e3 d7d6 d2d4 g8f6 g1h3 c7c6 f1e2 f8e7 b2b3 e8g8 e1g1 b8a6 c1b2 f6e4 d4e5 d6e5 c3e4 f5e4 b2e5 c8h3 g2h3 e7f6 d1d8 a8d8 e5f6 g7f6 a1d1 g8f7 d1d8 f8d8 f1d1 d8d1 e2d1 f6f5 f2f3 f7f6 d1c2 f6e5 a2a3 b7b5 c4b5 c6b5 b3b4 a6c7 f3f4 e5d5 c2b3 d5d6 g1g2 c7d5 b3d5 d6d5 g2f1 d5c4 f1g1');
INSERT INTO evaluations VALUES(57495,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(57495,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(57495,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(57495,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(57495,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(57495,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(57495,36,'cp 34','d7d6 d2d4 g8f6 g1h3 f8e7 f1e2 c7c6 b2b3 e8g8 e1g1 f6e4 c1b2 b8a6 c3a4 e5d4 b2d4 e7f6 h3f4 d8e7 d4f6 e4f6 d1d4 g7g5 f4h5 f6h5 e2h5 a6c5 a4c5 d6c5 d4c3 a7a5 a1e1 a5a4 e3e4 f5f4 e4e5 a4b3 a2b3');
INSERT INTO evaluations VALUES(29521,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(29521,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(29521,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29521,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(29521,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(29521,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(29521,36,'cp 26','e7e6 f3h4 f5g6 h4g6 h7g6 d1b3 b8c6 e2e3 f8b4 c1d2 d8d7 e1c1 b4c3 d2c3 a7a5 c4d5 e6d5 f2f3 e8g8 f1b5 d7e7 c3d2 c6d8 h2h4 c7c6 b5d3 d8e6 c1b1 b7b5 g2g4 f8c8 d1c1 b5b4 a2a4 b4a3 b3a3 e7a3 b2a3 c6c5 g4g5 f6h5');
INSERT INTO evaluations VALUES(20055,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(20055,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(20055,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(20055,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(20055,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(20055,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(20055,36,'cp -15','a7a6 e1g1 e7e6 d1e2 f8e7 f1d1 c8d7 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 h2h3 a8c8 a1c1 d8e8 a2a3 d7e6 b2b4 h7h6 b4b5 e6d5 d1d5 c6b8 c1c8 e8c8 f3d2 a6b5 d5b5 b8d7 e2c4 e7g5 e3g5 h6g5 b5d5 d7f6 d5d6 c8c4 d2c4 f8c8 c4e5 f6e4 d6b6 c8c7 g1h2');
INSERT INTO evaluations VALUES(17580,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(17580,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(17580,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(17580,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(17580,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(17580,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17580,36,'cp 0','e1g1 e7e6 d1e2 f8e7 f1d1 c8d7 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 a1c1 a8c8 a2a3 d8e8 h2h3 g8h8 b2b4 f7f5 e4f5 d7f5 b4b5 a6b5 e2b5 c8b8 b5c4 b8c8');
INSERT INTO evaluations VALUES(17579,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(17579,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17579,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17579,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(17579,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17579,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(17579,36,'cp -16','e7e6 c1f4 f8e7 d1e2 e6e5 f4e3 g8f6 f1d1 e8g8 c3d5 f6d5 c4d5 c6b4 d5b3 c8d7 h2h3 b4c6 b3d5 d8e8 a2a3 g8h8 f3d2 a8c8 d2f1 d7e6 e2d3 e8d7 e3b6 h7h6 f1g3 g7g6 a1c1 h8h7 g3f1 e7d8 d5e6 d7e6 b6e3');
INSERT INTO evaluations VALUES(17678,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17678,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17678,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17678,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17678,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17678,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17678,36,'cp 0','d1e2 f8e7 f1d1 c8d7 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 a1c1 a8c8 h2h3 h7h6 a2a3 d7e6 d5c6 c8c6 c1c6 b7c6 f3e5 d8a8 e5c4 f8d8 e3b6 d8d7 b6e3');
INSERT INTO evaluations VALUES(17677,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(17677,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(17677,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17677,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17677,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17677,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17677,36,'cp 0','g8e7 c1g5 h7h6 g5e3 e7g6 a1d1 f8e7 c3a4 e8g8 a4b6 a8b8 c4d3 e6e5 b6c8 d8c8 d3c2 c8g4 h2h3 g4h5 f1e1 g6h4 f3h4 h5h4 a2a3 b8c8 c2b1 e7g5 e3b6 g5d8 b6e3 d8g5');
INSERT INTO evaluations VALUES(35105,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(35105,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(35105,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(35105,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(35105,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(35105,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(35105,36,'cp 147','g1f3 d7d6 d2d4 e5d4 f3d4 b8c6 d4c6 b7c6 b1c3 f8e7 h2h3 f6e5 e1g1 g8f6 c1e3 e8g8 f2f4 e5h5 d1d2 a7a5 c4d3 f8e8 b2b3 e7f8 a1e1 c8b7 d2f2 b7a6 e1d1 a6d3 c2d3 c6c5 e3d2 h5g6 d1e1 c7c6 g2g4');
INSERT INTO evaluations VALUES(30676,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(30676,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(30676,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(30676,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(30676,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(30676,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(30676,36,'cp 118','g1f3 c7c5 c2c3 g8f6 d1e2 c5d4 c3d4 b8c6 a2a3 f8e7 e1g1 e8g8 b1c3 c6a5 c1f4 a8c8 a1d1 d7d5 e4e5 f6e4 c3e4 d5e4 d3e4 b7e4 e2e4 d8d5 e4d5 e6d5 f1e1 a5c4 e1e2 a7a5 g2g3 a5a4 d1c1 c4a5 c1c8 f8c8 h2h4 b6b5 h4h5');
INSERT INTO evaluations VALUES(55839,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55839,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(55839,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(55839,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(55839,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(55839,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(55839,36,'cp -31','g8f6 g1f3 c7c5 d4c5 f8c5 f1d3 b8c6 e1g1 e8g8 e3e4 d5d4 c3a4 c5d6 c2c3 e6e5 h2h3 f8e8 f1e1 c8d7 c3d4 c6d4 a4c3 d6c5 f3d4 e5d4 c3b1 d7c6 b1d2 a7a6 d1f3 c5d6 d2b3 c6e4 d3e4 e8e4 e1e4 f6e4');
INSERT INTO evaluations VALUES(53911,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(53911,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(53911,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(53911,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(53911,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(53911,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(53911,36,'cp 60','f8e7 f1b5 c8d7 b1c3 d7b5 c3b5 g8f6 c2c4 e8g8 b5c3 f8e8 e1g1 b8c6 c1f4 f6h5 f4e3 h5f6 h2h3 f6d7 d1d2 d7e5 f3e5 d6e5 c3d5 d8d6 f1d1 c6d4 e3d4 c5d4 b2b4 b7b6 a2a4 a7a5 b4a5 a8a5 d1b1 e7d8 b1b5 g8h8 d2d3 d6c6 g1h1 c6d7 b5a5');
INSERT INTO evaluations VALUES(37905,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(37905,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(37905,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(37905,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(37905,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(37905,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(37905,36,'cp 78','b1c3 g8f6 f1b5 c8d7 d1e2 b8c6 c1f4 a7a6 b5c6 d7c6 e1c1 e8g8 f3e5 c6b5 e5c4 f8e8 f2f3 b5c4 e2c4 b7b5 c4e2 c5c4 c3d5 f6d5 d1d5 c4c3 b2b3 b5b4 h1d1 d8c7 c1b1 a6a5 f4d6 e7d6 d5d6 a5a4 d6d7 c7c6 d1d6 c6c5 e2c4 c5c4 b3c4 a8c8 d6a6');
INSERT INTO evaluations VALUES(37904,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(37904,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(37904,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(37904,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(37904,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(37904,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(37904,36,'cp 75','c8e6 f1b5 b8c6 e1g1 g8f6 e4e5 d6e5 f3e5 d8b6 d1f3 a8c8 b2b3 e8g8 e5c6 b7c6 b5d3 c5c4 b3c4 b6d4 f1d1 d4g4 f3g4 f6g4 c3e4 g4e5 e4g5 e5d3 g5e6 d3f2 d1d7 e7f6 e6f8 f2h3 g2h3 f6a1 c1e3 c8f8 d7a7 f7f5 e3c5 f8e8 a2a4 e8e2');
INSERT INTO evaluations VALUES(36506,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(36506,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(36506,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36506,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(36506,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(36506,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36506,36,'cp 14','f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 f1e1 c7c5 c2c3 c5d4 f3d4 e7e5 d4b5 d7c5 d1d8 f8d8 b5c7 a8b8 c1g5 f7f6 g5e3 b7b6 e4d5 g8h8 e1d1 c8f5 e3c5 b6c5 d5c4 g7f8 b2b3 f8h6 d1d5 d8d5 c7d5 h8g7 g2g3 b8d8 d5e7 e5e4 e7c6 d8d2 c6a7');
INSERT INTO evaluations VALUES(36706,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36706,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36706,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(36706,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36706,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36706,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(36706,36,'cp 13','c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 c2c3 c7c5 f1e1 c5d4 f3d4 d7c5 e4c2 e7e5 d4b3 c5e6 d1d8 f8d8 b3a5 d8d5 b2b4 d5d7 a5c4 d7c7 c2b3 e6f8 c4e3 c8e6 c3c4 e5e4 a1b1 g7c3 e1d1 a7a5 b4b5 a5a4 b5b6 a4b3 b6c7 b3a2');
INSERT INTO evaluations VALUES(36703,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36703,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36703,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36703,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36703,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36703,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36703,36,'cp 10','d6d5 c4d3 d5e4 d3e4 b8d7 d1e2 c7c5 c2c3 c5d4 f3d4 a7a6 e4c2 d8c7 f1e1 f8e8 d4f3 d7f6 h2h3 c8e6 f3e5 f6d7 e5d7 e6d7 a2a4 d7c6 a4a5 b7b6 a5b6 c7b6 a1a3 a6a5 c2a4 e8c8 c1e3 b6c7 a4c6 c7c6 h3h4 h7h6 h4h5');
INSERT INTO evaluations VALUES(36821,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36821,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36821,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(36821,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(36821,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36821,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(36821,36,'cp 21','c4d3 d5e4 d3e4 b8d7 c2c3 c7c5 d1e2 c5d4 f3d4 a7a6 e4c2 f8e8 d4f3 d7f8 c1e3 d8c7 a2a4 c8e6 f1d1 e6c4 e2e1 a8d8 a4a5 d8d1 e1d1 f8d7 c2b3 c4b5 f3d4 b5c4 b3c4 c7c4 d1b3 c4c7 a1d1 e8c8 b3a4 d7e5 d4f3 e5f3 g2f3');
INSERT INTO evaluations VALUES(36822,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36822,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36822,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36822,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36822,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36822,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36822,36,'cp 17','d5e4 d3e4 b8d7 c2c3 c7c5 d1e2 c5d4 f3d4 d8a5 f1d1 d7f6 e4c2 f8e8 h2h3 c8d7 a2a4 a8d8 c2b3 a7a6 d4c2 e7e6 c2b4 a5c7 c1g5 h7h6 g5h4 a6a5 b4c2 d7c6 c2d4 c6d5 b3d5 d8d5 d4b5 c7b6 h4f6 d5d1 a1d1 g7f6 b5d6 e8d8');
INSERT INTO evaluations VALUES(36833,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36833,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36833,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(36833,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36833,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36833,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36833,36,'cp 29','d3e4 b8d7 c2c3 c7c5 d1e2 c5d4 f3d4 d8c7 d4b5 c7b8 f1d1 a7a6 b5a3 b8c7 e2c4 c7e5 e4f3 e5f6 c4f4 d7e5 f4f6 g7f6 f3e4 a8b8 c1f4 c8e6 e4d5 f8d8 d5e6 f7e6 a3c2 g8f7 f4e5 f6e5 c2e3 d8d1 a1d1 e5f6');
INSERT INTO evaluations VALUES(36828,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36828,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(36828,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(36828,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36828,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36828,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36828,36,'cp 22','b8d7 c2c3 c7c5 d1e2 c5d4 f3d4 a7a6 e4c2 e7e5 d4f3 d8c7 f1e1 f8e8 c1f4 b7b5 f4g3 c8b7 a2a4 c7c5 a4b5 a6b5 c2d3 b7c6 a1a8 e8a8 d3e4 c6e4 e2e4 a8a4 e4c2 c5c6 h2h3 b5b4 e1d1 b4c3 c2c3 c6c3 b2c3 d7f8 g3e5 g7e5 f3e5 a4a3 d1c1 f8e6 e5d3');
INSERT INTO evaluations VALUES(49384,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(49384,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(49384,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(49384,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(49384,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(49384,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49384,36,'cp 12','b1c3 d7d5 d2d4 f8b4 c4d5 e6d5 c1g5 h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b2c3 b8d7 f1e2 c7c5 d4c5 d7c5 b3b4 a8c8 e1g1 g7g5 h4g3 f8e8 f1d1 c5e4 c3c4 d8b6 b4b6 a7b6 d1c1 f5e6 f3d4 d5c4 d4e6 e4g3 h2g3 e8e6 e2c4 e6c6');
INSERT INTO evaluations VALUES(38989,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38989,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(38989,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(38989,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(38989,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38989,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(38989,36,'cp 31','d7d6 f1d3 c7c5 b1a3 c5d4 e5d6 g8f6 d3b5 c8d7 d1d4 a7a6 b5d7 b8d7 e1g1 a8c8 f1d1 e8g8 a1b1 c8c6 c1g5 h7h6 g5h4 d8c8 c2c3 f8d8 d4e3 b7b5 a3c2 d7b6 h4f6 g7f6 e3h6 c6d6 d1d6 d8d6 h6f4 c8d8 c2e3 b6d5 e3d5 d6d5 h2h3');
INSERT INTO evaluations VALUES(38542,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(38542,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38542,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(38542,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(38542,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(38542,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(38542,36,'cp 56','f1d3 c7c5 e5d6 d8d6 b1c3 g8f6 c3b5 d6c6 c1f4 e8g8 b5a7 a8a7 f4b8 a7a5 e1g1 c5d4 f3d4 c6d5 d4b3 a5a8 c2c4 d5g5 b8d6 f8e8 b3c5 c8d7 c5d7 f6d7 d1e2 g5a5 d3e4 d7c5 e4f3 e8d8 f1d1 a5b6 d6c5 b6c5 d1d8 a8d8 a1b1 c5c7 b2b4 c7e7');
INSERT INTO evaluations VALUES(14900,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(14900,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(14900,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(14900,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(14900,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(14900,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(14900,36,'cp 117','a7a6 a2a3 f8d6 e3e4 e6e5 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 c4f7 h8f8 f7b3 e5c3 b2c3 f6e4 e1g1 c8f5 c1e3 f8e8 f1e1 e4d6 a1d1 d8c8 e3d4 g7g6 d4e5 f5d7 f2f3 a6a5 e5d6 c7d6 e1e8 d7e8 d1d6 c8c7 d6f6 a8d8 f6f8 c7d6 f8h8 d8c8 h8h7 c8c3 h7b7 e8c6 b7b6');
INSERT INTO evaluations VALUES(37383,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(37383,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(37383,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(37383,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(37383,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(37383,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(37383,36,'cp 37','d2d4 b8c6 g1f3 d7d5 c4d5 f6d5 e1g1 d5b6 d4d5 c6a5 e2e4 c7c6 f1e1 f8e8 c1f4 c6d5 e4d5 a5c4 d1b3 c8f5 h2h3 a8c8 a1c1 c4a5 b3b5 g7c3 c1c3 d8d5 f3d4 d5b5 d4b5 c8c3 b2c3 f5d7 b5c7 e8d8 e1e7 d7c6 f4h6 b6c4');
INSERT INTO evaluations VALUES(37403,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(37403,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37403,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(37403,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(37403,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(37403,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(37403,36,'cp 48','b8c6 g1f3 d7d5 c4d5 f6d5 e1g1 d5b6 d4d5 c6a5 e2e4 c7c6 c1f4 g7c3 b2c3 c6d5 e4d5 d8d5 f3d4 d5c5 f1e1 c8d7 e1e5 c5a3 d1e1 a5c4 e5e7 f8e8 e7e8 a8e8 e1c1 a3c1 a1c1 c4b2 f4h6 b2d3 c1a1 d3c5 h2h3 f7f6 g3g4 c5e4 a1c1 b6c4');
INSERT INTO evaluations VALUES(57674,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(57674,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(57674,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(57674,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(57674,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(57674,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(57674,36,'cp -3','d7d5 e4d5 g8f6 d2d4 d8d5 g1f3 f8g7 b1c3 d5a5 f3e5 c7c5 e5c4 a5c7 d4d5 e8g8 f1e2 b8d7 a2a4 d7b6 c4e5 a7a6 a4a5 b6d7 e5c4 b7b5 a5b6 d7b6 d5d6 c7b8 d6e7 f8e8 c4d6 e8e7 d6c8 b8c8 e1g1 f6e4 c3e4 e7e4');
INSERT INTO evaluations VALUES(55521,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(55521,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(55521,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(55521,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(55521,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(55521,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(55521,36,'cp 173','g1f3 b8c6 c1g5 f7f6 e5f6 g8f6 e2e3 h7h6 g5f6 d8f6 b1c3 f6f7 f1b5 a7a6 b5a4 c8d7 d1d5 f7d5 c3d5 e8c8 h2h4 b7b5 a4b3 c6a5 c2c3 a5b3 a2b3 c8b7 f3d4 d8e8 b3b4 h8g8 h4h5 c7c6 d5f4 f8e7 f2f3 e7g5 e1f2 g8f8 d4e2 d7c8 h1c1 b7c7 c3c4 b5c4 c1c4');
INSERT INTO evaluations VALUES(55519,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(55519,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(55519,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(55519,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(55519,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(55519,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(55519,36,'cp 177','b8c6 c1g5 f7f6 e5f6 g8f6 e2e3 h7h6 g5f6 d8f6 b1c3 f6f7 f1b5 a7a6 b5a4 c8d7 d1d5 f7d5 c3d5 e8c8 h2h4 d7g4 d5f4 f8e7 c2c3 e7f6 h4h5 g4d7 a4b3 h8e8 f4d5 c6e5 d5f6 g7f6 f3e5 e8e5 b3d1 d8e8 h1h4 f6f5 d1e2 d7c6');
INSERT INTO evaluations VALUES(21890,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(21890,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(21890,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(21890,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(21890,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(21890,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(21890,36,'cp 162','c1g5 f7f6 e5f6 g8f6 e2e3 h7h6 g5f6 d8f6 b1c3 f6f7 f1b5 a7a6 b5a4 c8d7 d1d5 f7d5 c3d5 e8c8 h2h4 d7g4 a1d1 b7b5 a4b3 c6e5 f3e5 d6e5 f2f3 g4d7 e1e2 a6a5 a2a3 a5a4 b3a2 c7c6 d5b6 c8c7 b6d7 d8d7 d1d7 c7d7 c2c3 f8c5 h1c1 d7c7 h4h5');
INSERT INTO evaluations VALUES(21879,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(21879,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(21879,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(21879,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(21879,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(21879,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(21879,36,'cp 189','f7f6 e5f6 g8f6 b1c3 h7h6 g5f6 d8f6 e2e3 e8d8 h2h4 g7g6 d1d5 c8e6 d5e4 f8g7 e1c1 f6f7 f1a6 a8b8 a6d3 g7c3 b2c3 e6a2 e4g6 f7g6 d3g6 d8e7 c1d2 c6e5 d1a1 a2d5 f3e5 d6e5 a1a7 d5g2');
INSERT INTO evaluations VALUES(51110,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(51110,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(51110,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(51110,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(51110,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(51110,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(51110,36,'cp 55','e7e6 d2d4 d7d5 b1c3 f8b4 e4e5 c8d7 f1d3 g8e7 a2a3 b4c3 b2c3 d7b5 a3a4 b5d3 c2d3 b7b6 e1g1 e8g8 d1e2 c7c5 c1a3 c5d4 f3d4 f8e8 a3d6 d8d7 f1b1 e7c8 d6a3 b8c6 d4f3 h7h6 h2h4 c6a5 f3d4 a5c6 d4c6 d7c6 e2g4');
INSERT INTO evaluations VALUES(51041,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(51041,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(51041,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(51041,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(51041,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(51041,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(51041,36,'cp 68','d2d4 d7d5 b1c3 f8b4 e4d5 e6d5 f1d3 g8e7 e1g1 c8f5 d3f5 e7f5 d1d3 f5e7 c3e2 b4d6 h2h4 b8d7 f1e1 e8g8 c1f4 f8e8 f4d6 c7d6 h4h5 h7h6 f3h4 d7f6 e2g3 d8b6 b2b3 g8f8 e1e2 a8c8 a2a4 e7c6 e2e8 c8e8 h4f5 c6e7 f5e3 b6c7');
INSERT INTO evaluations VALUES(51037,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(51037,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(51037,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(51037,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(51037,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51037,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(51037,36,'cp 48','d7d5 b1c3 g8f6 e4e5 f6d7 c3e2 c7c5 c2c3 d8a5 h2h4 c5d4 e2d4 b8c6 c1f4 a5b6 d1c2 h7h6 h4h5 d7c5 d4b3 c8d7 f1e2 c5e4 e2d3 e4g5 f3d4 c6d4 b3d4 f8c5 f4e3 d7b5 d4b5 a6b5 e3c5 b6c5 c2b3 e8g8 e1g1 c5c7 a1e1 c7b6 d3c2 f7f6 e5f6 f8f6');
INSERT INTO evaluations VALUES(50967,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(50967,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(50967,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(50967,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(50967,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(50967,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(50967,36,'cp 70','b1c3 g8f6 e4e5 f6d7 c3e2 c7c5 c2c3 d8a5 h2h4 b8c6 a2a3 c5d4 b2b4 a5c7 c3d4 b7b5 e2f4 d7b6 f1e2 a6a5 b4a5 b6c4 e1g1 c7a5 c1e3 g7g6 h4h5 c8d7 f4d3 f8e7 d1c1 g6h5 d3f4 h5h4 f4h5 h8g8 e3h6 g8g6 a1b1 e8c8 h5f6 e7f6 e5f6');
INSERT INTO evaluations VALUES(50965,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(50965,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(50965,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(50965,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(50965,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(50965,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(50965,36,'cp 74','g8f6 e4e5 f6d7 c3e2 c7c5 c2c3 d8a5 h2h4 b8c6 a2a3 c5d4 b2b4 a5c7 c3d4 b7b5 e2f4 d7b6 f1e2 a6a5 b4a5 b6c4 e1g1 c7a5 c1e3 g7g6 h4h5 f8e7 f3h2 a5c3 h2g4 a8a3 a1a3 c3a3 e2c4 b5c4 d1f3 a3b2 f4e2 g6h5 g4f6 e7f6 f3f6 h8g8 e2f4');
INSERT INTO evaluations VALUES(33249,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(33249,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(33249,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(33249,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(33249,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(33249,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(33249,36,'cp 35','d4f3 b8c6 b1c3 g8e7 c1g5 f6e6 f1d3 d7d6 d1e2 f7f6 c3b5 e6d7 g5e3 c5e3 e2e3 a7a6 b5c3 d7e6 e1g1 c8d7 a2a3 e8g8 f3d4 c6d4 e3d4 c7c5 d4e3 f6f5 e4f5 e7f5 e3e6 d7e6 f1e1 f5d4 c3e2 d4c6 f2f3 d6d5 e2g3 a8e8 a1d1 g8h8');
INSERT INTO evaluations VALUES(33255,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(33255,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(33255,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(33255,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(33255,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(33255,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(33255,36,'cp 35','b8c6 b1c3 g8e7 c1g5 f6e6 f1d3 f7f6 g5d2 d7d6 e1g1 c8d7 a2a3 c6e5 b2b4 e5d3 c2d3 c5b6 a3a4 a7a6 a4a5 b6a7 d2e3 a7e3 f2e3 e8g8 f3d4 e6f7 a1c1 a8e8 d1e1 e7c6 c3d5 c6d4 e3d4 c7c6 d5b6 d7e6 h2h3 f7h5 d4d5 c6d5');
INSERT INTO evaluations VALUES(298,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(298,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(298,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(298,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(298,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(298,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(298,36,'cp 28','b1c3 g8e7 c1g5 f6e6 f1d3 f7f6 g5f4 d7d6 c3a4 c5b6 e1g1 e7g6 f4g3 g6e5 a4b6 a7b6 f3e5 d6e5 f2f4 e5f4 g3f4 c6e5 d1h5 g7g6 h5h6 c8d7 f4g3 e6e7 d3e2 e7c5 g1h1 e8c8 c2c3 h8e8 h6h7 c5e3 a1e1 e3e4 f1f6 d7c6');
INSERT INTO evaluations VALUES(297,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(297,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(297,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(297,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(297,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(297,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(297,36,'cp 39','g8e7 c1g5 f6e6 f1d3 f7f6 g5f4 d7d6 c3a4 c5b6 e1g1 e6f7 a4b6 a7b6 a2a4 c8g4 h2h3 e7g6 f4d2 g4f3 d1f3 g6e5 f3g3 e5d3 g3d3 e8g8 b2b3 f6f5 a1e1 a8e8 d2c3 f5e4 e1e4 e8e4 d3e4 h7h6 e4d3 f7f5 d3e2 f8f7 f1e1 c6e7 e2d2 e7d5 e1e8 f7f8 e8f8 g8f8');
INSERT INTO evaluations VALUES(42640,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(42640,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(42640,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(42640,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(42640,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(42640,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(42640,36,'cp 147','f3e5 d8e7 e5g4 d7d5 g4f6 e7f6 d1f3 f6f3 g2f3 b8c6 c2c3 d5d4 b1d2 f7f5 f3f4 c8e6 e4e5 g7g5 d2b3 c5b6 c3c4 a7a5 a2a4 e8c8 f1g2 g5f4 g2c6 b7c6 c1f4 h8g8 f4d2 d8e8 f2f4 c6c5 e1f2 g8g4 h2h3 g4g6 h1g1 g6h6 d2a5 h6h3 a5b6 c7b6');
INSERT INTO evaluations VALUES(42532,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(42532,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(42532,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(42532,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(42532,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(42532,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(42532,36,'cp 143','d8e7 e5g4 b8c6 b1c3 d7d5 g4f6 e7f6 d1d2 d5e4 c3e4 f6e7 f1e2 c5b6 e1g1 c8e6 c2c3 e8c8 f1e1 f7f5 e4g5 e6d5 e2f3 e7d7 f3d5 d7d5 d3d4 h7h6 g5e6 d8d6 e6f4 d5f7 d2d1 g7g5 f4d3 h8d8 d1f3 g5g4 f3g3 d6g6 d3f4');
INSERT INTO evaluations VALUES(51834,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(51834,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(51834,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(51834,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(51834,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(51834,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51834,36,'cp 31','d7d5 c4b3 a7a5 a2a4 g8f6 b1c3 f8d6 d2d4 e5d4 d1d4 d8e7 c1e3 c6c5 d4d2 c5c4 b3a2 d5e4 f3d4 d6b4 a2c4 e8g8 h2h3 b8c6 a1d1 c8d7 e1g1 g8h8 c4b3 a8d8 d2e2 h7h6 d4b5 d7f5 d1d8 f8d8 f1d1 d8d1 c3d1 f5e6 b3e6 e7e6 b5d4 e6d5 c2c3 b4e7 d4c6 d5c6 e2b5 c6b5 a4b5 f6d7 e3d4');
INSERT INTO evaluations VALUES(53747,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(53747,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53747,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(53747,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(53747,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53747,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53747,36,'cp 29','c4b3 a7a5 a2a4 g8f6 b1c3 f8d6 d2d4 e5d4 d1d4 d8e7 c1e3 c6c5 d4d2 c5c4 b3a2 d5e4 f3d4 e8g8 h2h3 d6b4 a2c4 c8d7 e1g1 b8c6 a1d1 f8e8 c4b3 a8c8 e3g5 h7h6 g5h4 c6d4 d2d4 b4c3 h4f6 c3d4 f6e7 e8e7 d1d4 d7c6 f1d1 g8h7 d4d6 f7f5 d6d8 c8d8 d1d8');
INSERT INTO evaluations VALUES(53762,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(53762,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(53762,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(53762,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(53762,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(53762,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(53762,36,'cp 22','a7a5 a2a4 g8f6 b1c3 f8d6 d2d4 e5d4 d1d4 d8e7 e1g1 d6c5 d4e5 d5e4 e5e7 c5e7 f3g5 e8g8 g5e4 b8a6 f1e1 a6c5 e4c5 e7c5 h2h3 h7h6 c1f4 c8d7 f4e5 c5b4 e1d1 a8d8 e5f6 g7f6 c3e4 g8g7 e4d6 b4d6 d1d6 d7e6 a1d1 d8d6 d1d6 e6b3 c2b3 f8e8 g1f1 e8e7 d6d1 e7e4 d1d7 b7b6 f2f3');
INSERT INTO evaluations VALUES(41301,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(41301,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41301,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(41301,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(41301,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41301,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41301,36,'cp 0','c7c5 d4d5 e8g8 g1f3 e7e6 f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 g1h1 b7b6 c1g5 d8d7 h2h3 g4e5 f3e5 f8f1 d1f1 g7e5 c3e4 c8b7 g5f6 a8f8 f6e5 f8f1 a1f1 d7d8 e5f6 d8c8 d6d7 c8d7 f6e5 d7d8 e4f6 g8h8 f6e8 h8g8');
INSERT INTO evaluations VALUES(40349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40349,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40349,36,'cp 11','d4d5 e8g8 g1f3 e7e6 f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 g1h1 b7b6 c3e4 d8d7 f3g5 g4f6 e2c4 f6e4 f1f8 g7f8 g5e4 f8g7 c1g5 c6a5 c4d3 c5c4 d3c2 c8b7 d1g4 a8f8 g5e7 f8f5 a1e1 a5c6 e4g5');
INSERT INTO evaluations VALUES(40368,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40368,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40368,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40368,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40368,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40368,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40368,36,'cp 21','e8g8 g1f3 e7e6 f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 e8f8 e1g1 b8c6 g1h1 b7b6 c3e4 d8d7 f3g5 f8f1 d1f1 g4e5 e4f6 g7f6 f1f6 d7g7 f6g7 g8g7 b2b4 h7h6 c1b2 h6g5 e2f3 c8b7 f3c6 b7c6 b2e5 g7g8 b4c5 a8c8 c5b6 a7b6 h1g1 c6d7 a1b1 c8c5 e5f6 b6b5 b1f1 c5f5 f1f5 e6f5');
INSERT INTO evaluations VALUES(35980,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(35980,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(35980,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(35980,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(35980,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(35980,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(35980,36,'cp 20','g1f3 e7e6 f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 g1h1 b7b6 c3e4 d8d7 f3g5 f8f1 d1f1 g4e5 e4f6 g7f6 f1f6 d7g7 f6g7 g8g7 b2b4 h7h6 c1b2 h6g5 e2f3 c8b7 f3c6 b7c6 b2e5 g7g8 b4c5 c6d7 c5b6 a7b6 e5d4 a8c8 d4b6 c8c2 h1g1 d7c6 b6f2 c2d2 a1c1 c6b7 a2a4 d2d6 f2g3 d6d2');
INSERT INTO evaluations VALUES(35979,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(35979,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(35979,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(35979,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(35979,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35979,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(35979,36,'cp 26','e7e6 f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 g1h1 b7b6 c3e4 d8d7 f3g5 f8f1 d1f1 g4e5 e4f6 g7f6 f1f6 d7g7 f6g7 g8g7 b2b4 h7h6 g5f3 e5f7 b4c5 b6c5 c1e3 f7d6 e3c5 d6e4 c5g1 e4c3 e2c4 c8d7 a1e1 g7f7 e1c1 c3d5 c4b5 d5e7 c1d1 a8d8 g1a7 c6a7');
INSERT INTO evaluations VALUES(36095,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36095,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36095,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36095,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36095,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(36095,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36095,36,'cp 18','f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 e8f8 e1g1 b8c6 g1h1 b7b6 c3e4 d8d7 f3g5 f8f1 d1f1 g4e5 e4f6 g7f6 f1f6 d7g7 f6g7 g8g7 b2b4 h7h6 g5f3 e5f7 b4c5 b6c5 c1a3 f7d6 a3c5 d6e4 c5g1 c8d7 e2b5 e4f6 a1e1 a8c8 a2a4 g6g5 h2h3 g7f7 b5c6 d7c6 e1c1 c6b7 f3e5 f7e8 c1c8 b7c8 g1a7');
INSERT INTO evaluations VALUES(36094,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36094,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36094,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36094,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36094,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36094,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36094,36,'cp 25','e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 g1h1 b7b6 c3e4 d8d7 f3g5 f8f1 d1f1 g4e5 e4f6 g7f6 f1f6 d7g7 f6g7 g8g7 b2b4 h7h6 g5f3 e5f7 b4c5 b6c5 c1a3 f7d6 a3c5 d6e4 c5g1 c8d7 a1e1 e4f6 e2b5 a8c8 a2a4 g6g5 h2h3 g7f7 e1d1 f6d5 g1a7 c6a7 b5d7 c8c7 h1h2 d5f6 f3e5 f7e7');
INSERT INTO evaluations VALUES(36082,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(36082,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36082,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36082,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36082,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(36082,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36082,36,'cp 12','c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 c3e4 c8d7 c1g5 d8b6 g1h1 h7h6 h2h3 h6g5 h3g4 c6e5 f3g5 b6b2 a1b1 f8f1 d1f1 b2a2 e4c5 a8f8 f1e1 f8f2 b1b7 a2e2 e1e2 f2e2 c5d7 e5d7 b7d7 g7e5 h1g1 e2d2');
INSERT INTO evaluations VALUES(36073,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36073,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36073,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36073,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36073,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36073,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36073,36,'cp 13','f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 g1h1 b7b6 c3e4 d8d7 f3g5 f8f1 d1f1 g4e5 e4f6 g7f6 f1f6 d7g7 f6g7 g8g7 b2b4 h7h6 g5f3 e5f7 b4c5 b6c5 c1a3 f7d6 a3c5 d6e4 c5g1 c8d7 a1e1 e4f6 e2b5 a8e8 a2a4 e6e5 e1c1 e8e6 a4a5 g6g5 b5c6 d7c6 g1a7 g5g4 f3e5 e6e5 c1c6 e5a5');
INSERT INTO evaluations VALUES(44516,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(44516,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(44516,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(44516,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(44516,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(44516,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(44516,36,'cp -17','h5c5 f6e4 c5e3 d7d5 b1c3 e4c3 d2c3 b8c6 f2f4 g7g6 c1d2 f8g7 g1f3 e8g8 f1d3 c8g4 e1c1 a8b8 d1e1 e7e6 f3e5 g4f5 c1b1 d8c7 d3f5 e6f5 e3c5 f8e8 e5d3 e8d8 h2h3 c7d7 c5f2 b7b5 f2c5 d5d4 g2g4');
INSERT INTO evaluations VALUES(44508,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(44508,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(44508,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44508,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(44508,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(44508,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(44508,36,'cp -24','f6e4 c5e3 e4f6 d2d4 g7g6 f1e2 f8g7 g1f3 b8c6 e1g1 e8g8 c2c3 e7e6 f1d1 b7b6 b1d2 f6d5 e3g5 d7d6 g5d8 f8d8 d2b3 c8b7 a2a4 h7h6 a4a5 c6e7 f3e1 g6g5 e2f1 d8c8 g2g3 b7c6 e1d3 c6b5 d1e1 g5g4 b3d2');
INSERT INTO evaluations VALUES(44643,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(44643,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(44643,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(44643,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(44643,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(44643,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(44643,36,'cp -23','c5e3 d7d5 b1c3 e4f6 d2d4 c8f5 f1d3 f5d3 e3d3 e7e6 g1f3 b8d7 c1f4 f8e7 e1g1 e8g8 a2a4 f6h5 f4e5 a8c8 c3e2 a7a5 b2b3 d8b6 a1c1 e7a3 c1a1 a3b4 c2c3 b4e7 h2h3 f8d8 e5h2 h7h6 c3c4 b6a6 f1c1 h5f6');
INSERT INTO evaluations VALUES(44677,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(44677,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(44677,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(44677,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(44677,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(44677,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(44677,36,'cp -20','e4f6 d2d4 g7g6 g1f3 b8c6 f1e2 f8g7 e1g1 e8g8 f1d1 d7d5 c2c3 f8e8 h2h3 c8f5 b1d2 e7e6 d2b3 f6d7 e2d3 f5d3 e3d3 d8b6 d3c2 b6c7 c1g5 f7f6 g5h4 d7b6 b3d2 a8c8 h4g3 e6e5 d4e5 f6e5 g3h2');
INSERT INTO evaluations VALUES(44639,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(44639,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(44639,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(44639,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(44639,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(44639,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(44639,36,'cp -12','g1f3 g7g6 f1e2 b8c6 e1g1 f8g7 d2d4 e8g8 f1d1 d7d5 c2c3 f8e8 f3e5 c8f5 b1d2 c6e5 d4e5 f6d7 d2f3 e7e6 e3d4 f5e4 c1f4 h7h6 a2a4 e4f3 e2f3 d8b6 h2h4 a8c8 d4b6 a7b6 f3e2 d7e5');
INSERT INTO evaluations VALUES(56215,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(56215,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(56215,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(56215,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(56215,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(56215,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(56215,36,'cp -31','f8e7 c1g5 b8d7 g5e7 g8e7 b1c3 e8g8 e1c1 f8e8 d1e1 d7f6 e2d2 c8f5 f1d3 f5d3 d2d3 e7g6 g1f3 c7c6 g2g3 d8c7 e1e8 f6e8 h1e1 g6f8 c3d1 e8d6 c2c3 a7a5 d1e3 a8e8 d3c2 c7d7 f3e5 d7h3 f2f3 f8g6 e5d3 d6c4 c2f2 e8e3 e1e3 c4e3 f2e3');
INSERT INTO evaluations VALUES(38316,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(38316,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(38316,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(38316,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(38316,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(38316,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(38316,36,'cp -27','c1g5 b8c6 c2c3 h7h6 g5e7 c6e7 g1f3 g8f6 e2c2 e7g6 f1e2 e8g8 e1g1 f6e4 b1d2 c8f5 f1e1 g6f4 e2f1 e4d6 c2b3 c7c6 e1e3 d8c7 f3h4 f5h7 g2g3 f4e6 a1e1 a7a5 a2a4 a8e8 f1g2 e6g5 e3e8 f8e8 e1e8 d6e8 h4f3');
INSERT INTO evaluations VALUES(38307,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(38307,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(38307,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(38307,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(38307,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(38307,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(38307,36,'cp -28','b8c6 c2c3 h7h6 g5e7 c6e7 e2d2 g8f6 f1d3 c8f5 d3f5 e7f5 g1f3 d8e7 d2e2 f6e4 e1g1 e8c8 f1e1 h8e8 e2c2 f7f6 b1d2 f5d6 e1e2 e7f7 a1e1 a7a6 a2a4 g7g5 d2e4 d5e4 f3d2 f6f5 f2f3 e4f3 d2f3 d6e4');
INSERT INTO evaluations VALUES(57691,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(57691,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(57691,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(57691,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(57691,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(57691,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(57691,36,'cp 16','d2d4 g8f6 g1f3 f8g7 c2c4 e8g8 b1c3 d7d5 c4d5 f6d5 f1c4 d5b6 c4e2 c7c5 e1g1 c5d4 f3d4 c8d7 d1b3 b8c6 d4c6 d7c6 e3e4 e7e6 c1e3 d8e7 e2b5 a8c8 a1c1 f8d8 b5c6 b7c6 f1d1 d8d1 c1d1 c6c5 g2g3 c5c4 b3b5 e7c7 d1c1');
INSERT INTO evaluations VALUES(40381,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(40381,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40381,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(40381,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40381,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(40381,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40381,36,'cp 30','f2f3 d8b6 c4b5 c8d7 c1e3 b6c7 d1d2 a7a6 b5d3 b8c6 e1c1 h7h5 c1b1 b7b5 d4c6 d7c6 h2h3 f6d7 c3d5 c6d5 e4d5 d7b6 h1e1 e8f8 e3g5 g7f6 h3h4 f8g7 f3f4 b6c4 d3c4 b5c4 g5f6 e7f6 f4f5 a8b8 d2c3 b8e8 e1e8 h8e8 f5g6 f7g6 d1e1 e8e1 c3e1 c7d7 e1e2 d7b7 e2c4');
INSERT INTO evaluations VALUES(40378,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40378,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(40378,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(40378,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(40378,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(40378,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(40378,36,'cp 28','d8b6 c4b5 c8d7 c1e3 b6c7 d1d2 a7a6 b5d3 b8c6 e1c1 h7h5 h2h3 b7b5 d4c6 d7c6 c1b1 c7b7 g2g4 a6a5 e3d4 b5b4 c3e2 e7e5 d4e3 h5g4 h3g4 h8h1 d1h1 d6d5 g4g5 f6h5 h1d1 a8d8 e4d5 d8d5 d2e1 d5d7 f3f4 c6e4 f4e5 e4d3 c2d3 g7e5');
INSERT INTO evaluations VALUES(23455,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23455,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23455,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(23455,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(23455,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(23455,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(23455,36,'cp 17','g1f3 g8f6 e2e3 d7d5 f1b5 e7e6 d1e2 c8d7 d2d3 f8c5 b2b3 c6e7 b5d7 d8d7 f3e5 d7d8 b1d2 a7a5 a2a4 c7c6 d2f3 c5b4 c1d2 b4d2 e2d2 e7g6 g2g4 g6e5 f3e5 f6d7 e5f3 b7b5 e1g1 e8g8 d2f2 f8e8 g1h1 d8c7 g4g5 b5a4 b3a4');
INSERT INTO evaluations VALUES(23454,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(23454,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(23454,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23454,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(23454,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(23454,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(23454,36,'cp -2','d7d5 e2e3 c8g4 f1b5 d8d7 d2d4 e7e6 e1g1 g4f3 d1f3 a7a6 b5a4 f7f5 c2c4 d5c4 a4c6 d7c6 f3c6 b7c6 b1d2 c6c5 d2c4 c5d4 e3d4 f8d6 c1d2 g8f6 a1c1 h7h5 c1c2 e8f7 g2g3 h5h4 g1g2 h4g3 h2g3 f6g4 c4e5 d6e5 d4e5 f7g6 b2b3');
INSERT INTO evaluations VALUES(57722,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(57722,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(57722,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(57722,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(57722,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(57722,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(57722,36,'cp 75','e2e4 e7e6 f1d3 d7d5 b1c3 c7c5 e4d5 e6d5 d4c5 b8c6 c3a4 f8c5 a4c5 d8a5 d1d2 a5c5 d2e3 c5e3 c1e3 d5d4 e3f4 g8e7 e1c1 e8g8 g1f3 f8d8 f4c7 d8e8 c7d6 c8g4 h1e1 e7f5 d6c5 e8e1 d1e1 g4f3 g2f3 g7g6 d3e4 g8g7 e4c6 b7c6');
INSERT INTO evaluations VALUES(57723,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(57723,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(57723,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(57723,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(57723,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(57723,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(57723,36,'cp 80','e7e6 f1d3 d7d5 b1c3 d5e4 c3e4 g8f6 g1f3 c8d7 e4f6 d8f6 e1g1 f8d6 d1e2 b8c6 c2c3 e8g8 f1d1 f8e8 e2e4 f6f5 e4h4 d6e7 d3f5 e7h4 d4d5 c6e7 f5e4 e6d5 f3h4 g7g5 e4f3 g5h4 c1h6 h4h3 h6g5 h3g2 g1g2');
INSERT INTO evaluations VALUES(56962,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(56962,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(56962,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(56962,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(56962,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(56962,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(56962,36,'cp 82','b1c3 d7d5 f1d3 b8c6 g1f3 c6b4 d3b5 c7c6 b5e2 g8f6 e4e5 f6e4 c1e3 c6c5 d4c5 e4c5 e1g1 b4c6 c3b5 f8e7 c2c4 e8g8 d1c2 c6b4 c2c3 c5e4 c3b3 c8d7 b5d6 a7a5 a2a3 b4a6 d6b7 a5a4 b7d8 a4b3 d8b7 f8b8 b7d6 e4d6 e5d6 e7d6');
INSERT INTO evaluations VALUES(56960,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(56960,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(56960,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(56960,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(56960,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(56960,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(56960,36,'cp 86','d7d5 f1d3 d5e4 c3e4 g8f6 g1f3 c8d7 e4f6 d8f6 e1g1 f8d6 d1e2 b8c6 c2c3 c6e7 f3e5 d6e5 d4e5 f6h4 b2b4 a7a6 a2a4 e8g8 c1d2 e7f5 d2e1 h4f4 g2g3 f4g5 d3e4 d7c6 e4c6 b7c6 f2f4 g5g6 e1f2 h6h5 f2c5 f8d8 a4a5');
INSERT INTO evaluations VALUES(55471,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(55471,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(55471,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(55471,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(55471,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(55471,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(55471,36,'cp 89','f1d3 b8c6 g1f3 c6b4 d3e2 g8f6 e4e5 f6e4 e1g1 c7c5 c1e3 b4c6 d4c5 e4c5 c3b5 a7a6 b5d4 c8d7 d4c6 d7c6 f3d4 f8e7 c2c4 e8g8 b2b4 c5e4 c4c5 d8c7 f2f4 a6a5 a2a3 e4c3 d1d3 c3e2 d3e2 c6d7 f1c1 a5b4 a3b4 f8c8 e2c2 a8a1');
INSERT INTO evaluations VALUES(55540,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(55540,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(55540,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(55540,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(55540,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55540,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(55540,36,'cp 37','b8c6 g1f3 c6a5 c4b5 c7c6 b5d3 c8e6 e1g1 f8e7 c3e2 g8f6 f3g5 c6c5 g5e6 f7e6 c2c3 a5c6 d3c2 e8g8 d2d3 d6d5 e2g3 d8d7 d1e2 d5d4 a2a4 f8f7 c2b3 a8f8 b3c4 c6a5 c4a2 e7d6 c1d2 a5c6 a1e1');
INSERT INTO evaluations VALUES(21899,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(21899,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(21899,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(21899,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21899,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(21899,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21899,36,'cp 38','g1f3 c6a5 c4b5 c7c6 b5d3 c6c5 c3d5 g8f6 c2c3 f8e7 e1g1 e8g8 d5e7 d8e7 h2h3 c8e6 d3b1 c5c4 d2d4 c4d3 b1d3 h7h6 b2b3 a5c6 c3c4 f6d7 d3c2 f7f5 e4f5 e6f5 c1e3 f5c2 d1c2 d7f6 a1e1 a8d8 e1d1 e7f7 f1e1');
INSERT INTO evaluations VALUES(19388,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(19388,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(19388,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(19388,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(19388,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(19388,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(19388,36,'cp -20','e7e6 b1c3 g8f6 c1f4 a7a6 f1e2 d7d6 a2a3 f8e7 g1f3 f6d7 e1g1 e8g8 f4e3 d7e5 a1b1 e5f3 e2f3 c6d4 b2b4 c8d7 f3h5 b7b5 c3e2 e7f6 e2d4 c5d4 e3d2 a8c8 b1c1 h7h6 h5g4 d8e7 h2h3 a6a5 b4a5 b5c4 c1c4 c8c4 d3c4 d7c6');
INSERT INTO evaluations VALUES(18768,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(18768,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(18768,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(18768,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(18768,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(18768,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(18768,36,'cp -28','b1c3 g8f6 c1f4 a7a6 g2g3 d7d5 c4d5 e6d5 f1g2 c8e6 g1f3 d5d4 c3e2 h7h6 f4d2 f8d6 b2b4 c5b4 f3d4 c6d4 e2d4 e6g4 d4f3 e8g8 e1g1 a6a5 d2e3 f8e8 h2h3 g4d7 d1d2 a5a4 a1b1 d6f8 b1b4 f8b4 d2b4 d8a5 b4a5 a8a5 f1b1 b7b5 f3d4 a4a3');
INSERT INTO evaluations VALUES(18766,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(18766,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(18766,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(18766,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(18766,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(18766,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(18766,36,'cp -26','g8f6 c1f4 a7a6 f1e2 d7d6 a2a3 f8e7 g1f3 e8g8 e1g1 f6g4 a1b1 e7f6 d1d2 g4e5 f4e3 e5f3 e2f3 c6d4 f3d1 c8d7 c3e2 b7b5 b2b3 d7c6 e2d4 c5d4 e3f4 f6e7 c4b5 a6b5 b1c1 c6d7 d2b4 e6e5 f4d2 d8b6 a3a4 f8c8 c1c8 a8c8');
INSERT INTO evaluations VALUES(12396,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(12396,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(12396,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(12396,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(12396,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(12396,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(12396,36,'cp -19','c1f4 a7a6 f1e2 d7d6 g1f3 f8e7 a1b1 e8g8 a2a3 f6d7 b2b4 c6d4 e1g1 b7b6 d1d2 c8b7 f4g3 d4e2 d2e2 d8c7 h2h3 d7e5 b4c5 e5f3 e2f3 b6c5 g3f4 e7f6 f1c1 a8b8 f3g3 f8d8 b1b3 b7a8 b3b8 c7b8 c1b1 b8c7 c3e2');
INSERT INTO evaluations VALUES(6972,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(6972,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(6972,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(6972,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(6972,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(6972,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(6972,36,'cp 43','g8f6 d2d4 e5d4 f3d4 d7d6 b1c3 e8g8 c1f4 c6e5 c4b3 c8g4 d4e2 g4d7 a2a4 a7a6 f1e1 b7b5 c3d5 f6d5 b3d5 a8b8 a4b5 a6b5 e2d4 e7f6 f4g3 e5c4 b2b3 c4b6 d1d2 f6d4 d2d4');
INSERT INTO evaluations VALUES(19711,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(19711,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19711,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(19711,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19711,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(19711,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19711,36,'cp 23','b7c6 e1g1 e7e5 c2c3 g8f6 f1e1 f8e7 d2d4 e5d4 c3d4 c5d4 f3d4 c8d7 c1f4 h7h6 b1c3 e8g8 d1d3 f8e8 a1d1 d8b6 b2b3 e7f8 h2h3 b6b7 d4c2 d6d5 e4d5 e8e1 c2e1 f6d5 c3d5 c6d5 e1f3 d7e6 f4d2 a8c8 d2e3 f8e7');
INSERT INTO evaluations VALUES(18461,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(18461,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(18461,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(18461,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(18461,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(18461,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(18461,36,'cp 27','e1g1 e7e5 c2c3 f7f5 d2d4 f5e4 f3g5 c8f5 f2f3 e4e3 c1e3 f8e7 g5e4 e5d4 c3d4 c5d4 e3d4 g8f6 e4f6 e7f6 f1e1 e8f7 b1c3 d6d5 c3e2 h8e8 d4f6 d8f6 d1d2 f5d7 a1c1 f6d6 a2a3 a7a5 e2f4 e8e1 c1e1 d6c5 g1f1 f7g8 d2c3 c5c3 b2c3 g8f7 f1f2 a8c8 f4d3 f7f6 d3c5 d7f5 g2g4 f5g6 e1e6 f6f7');
INSERT INTO evaluations VALUES(18460,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(18460,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(18460,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(18460,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(18460,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(18460,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(18460,36,'cp 27','d8c7 f1e1 e7e5 c2c3 g8f6 d2d4 c5d4 c3d4 f8e7 h2h3 e8g8 b1c3 h7h6 d1c2 c8e6 c1e3 c6c5 d4d5 e6d7 c3e2 a7a5 b2b3 f6h5 g2g4 h5f6 e2g3 f6h7 f3d2 e7g5 e3g5 h7g5 g1g2 a8b8 d2c4 d7b5 h3h4 g5h7 g3f5 b5c4 b3c4');
INSERT INTO evaluations VALUES(44784,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44784,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44784,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44784,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(44784,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(44784,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44784,36,'cp 25','f8g7 f2f3 d8b6 c4b5 c8d7 c1e3 b6c7 d1d2 a7a6 b5d3 b8c6 e1c1 h7h5 h2h3 b7b5 d4c6 d7c6 c1b1 f6d7 c3e2 a8c8 e3d4 g7h6 d2c3 e7e5 d4f2 c6b7 c3c7 c8c7 g2g4 h6g7 e2c3 b5b4 c3a4 d6d5 g4h5 d5e4 d3e4 b7e4 f3e4 h8h5 c2c3 d7f6 f2b6');
INSERT INTO evaluations VALUES(56473,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(56473,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(56473,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(56473,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(56473,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(56473,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(56473,36,'cp -44','b8c6 d2d4 e5e4 g1e2 f7f5 c2c4 d5c4 b1d2 g8f6 e1g1 f8d6 f2f3 e4f3 d2f3 e8g8 b2b3 c4b3 d1b3 g8h8 d4d5 c6e5 f3e5 d6e5 e2d4 a7a5 a2a4 c8d7 c1a3 e5d4 e3d4 f8e8 d5d6 c7d6 g2b7 d7e6 b3d3 e6d5 b7d5 f6d5 d3f5');
INSERT INTO evaluations VALUES(21759,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(21759,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(21759,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(21759,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(21759,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(21759,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(21759,36,'cp -51','d2d4 g8f6 d4e5 c6e5 b1d2 f8d6 g1f3 c8g4 h2h3 g4h5 e1g1 e8g8 b2b3 f6e4 g3g4 e5f3 d2f3 h5g6 c1b2 c7c6 d1d4 f7f6 c2c4 d6c5 d4d1 d5c4 d1c1 b7b5 b3c4 b5b4 f3h4 f6f5 h4f5 g6f5 g4f5 d8g5 c1c2 a8e8 f5f6 g7f6 a1d1 f8f7 g1h1 g5h4 b2d4 c5d4 d1d4 f6f5 c4c5 a7a5');
INSERT INTO evaluations VALUES(21690,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(21690,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(21690,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(21690,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(21690,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(21690,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(21690,36,'cp -51','g8f6 d4e5 c6e5 b1d2 f8d6 g1f3 c8g4 h2h3 g4h5 e1g1 e8g8 b2b3 d8e7 c1b2 a8d8 d1c1 e5f3 g2f3 h5g6 c2c4 c7c6 c4d5 f6d5 a2a3 h7h5 h3h4 f8e8 f1e1 d6c7 e3e4 d5f6 c1c3 c7b6 a1d1 b6d4 c3c2 f6g4 b2d4 d8d4 d2c4 g6e4 f3e4 d4e4 c2e4 e7e4 e1e4');
INSERT INTO evaluations VALUES(15115,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(15115,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(15115,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(15115,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(15115,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(15115,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(15115,36,'cp -38','d4e5 c6e5 b1d2 c7c6 b2b3 c8g4 g1e2 f8b4 e1g1 d8e7 c1b2 e8c8 a2a3 b4d6 h2h3 g4h5 b2d4 c6c5 d4b2 d6c7 d1e1 f6e4 f2f3 e4d2 e1d2 f7f6 f3f4 e5c6 e2c3 d5d4 e3d4 c5d4 c3e4 c7b6 a1e1 c8b8');
INSERT INTO evaluations VALUES(16498,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(16498,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(16498,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(16498,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(16498,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(16498,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(16498,36,'cp -31','f1b5 f6e4 e1g1 f8e7 f1e1 e4d6 b5c6 d7c6 f3e5 e8g8 c1b2 f8e8 d2d3 f7f6 e5g4 d6f5 b1d2 f5d4 h2h3 c6c5 b2d4 c5d4 d1f3 a7a5 a2a3 d8d7 e1e2 d7c6 a1e1 c6f3 d2f3 g8f7 a3a4 h7h5 g4h2 c8d7 f3d4 e7b4 e2e8 d7e8 e1e2 a8d8 h2f3 c7c5');
INSERT INTO evaluations VALUES(57442,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(57442,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(57442,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(57442,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(57442,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(57442,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(57442,36,'cp 171','b1c3 b8c6 f1c4 d7d5 c3d5 c6a5 c4e2 c8e6 d5c3 a5c6 g1f3 a7a6 e1g1 f8c5 g1h1 g8e7 f4e5 c6e5 d2d4 e5f3 d4c5 f3e5 d1e1 e8g8 c1e3 d8e8 b2b3 a8d8 h2h3 e7g6 e1g3 e8e7 a2a3 c7c6 a1d1 e6f7 h3h4 d8d1 c3d1 g8h8 d1f2');
INSERT INTO evaluations VALUES(31825,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31825,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31825,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(31825,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31825,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31825,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(31825,36,'cp 19','c7c6 f1d3 f5g6 g1f3 e7e6 d1c2 b8d7 e1g1 g8f6 b1d2 g6d3 c2d3 f8e7 f1d1 e8g8 e3e4 a7a5 d3e2 a5a4 h2h3 f6e4 d2e4 d5e4 e2e4 d8c7 c1f4 e7d6 f4d6 c7d6 d1d2 d6c7 e4e3 h7h6 a1d1 f8d8 g2g3 b7b5 d1c1 b5c4 c1c4');
INSERT INTO evaluations VALUES(57476,30,'cp 526',NULL);
INSERT INTO evaluations VALUES(57476,31,'cp 502',NULL);
INSERT INTO evaluations VALUES(57476,32,'cp 509',NULL);
INSERT INTO evaluations VALUES(57476,33,'cp 501',NULL);
INSERT INTO evaluations VALUES(57476,34,'cp 494',NULL);
INSERT INTO evaluations VALUES(57476,35,'cp 517',NULL);
INSERT INTO evaluations VALUES(57476,36,'cp 506','d1h5 g7g6 h5e5 d8e7 e5h8 e7e4 e1d1 g8e7 b1c3 e4f5 f2f3 d7d5 c3b5 b8a6 b5d4 f5f7 f1a6 b7a6 h1e1 c8d7 d2d3 e8c8 c1g5 d8e8 h8f6 f7f6 g5f6 c7c5 d4e6 e7c6 e6f8 e8f8 f6e7 f8f5 e7c5 a6a5 c2c3 h7h5 d1d2 g6g5 h2h3 c8b7 b2b4 d5d4');
INSERT INTO evaluations VALUES(57467,30,'cp 523',NULL);
INSERT INTO evaluations VALUES(57467,31,'cp 521',NULL);
INSERT INTO evaluations VALUES(57467,32,'cp 509',NULL);
INSERT INTO evaluations VALUES(57467,33,'cp 515',NULL);
INSERT INTO evaluations VALUES(57467,34,'cp 513',NULL);
INSERT INTO evaluations VALUES(57467,35,'cp 517',NULL);
INSERT INTO evaluations VALUES(57467,36,'cp 519','g7g6 h5e5 d8e7 e5h8 e7e4 e1d1 g8e7 b1c3 e4f5 f2f3 d7d5 c3b5 b8a6 b5d4 f5f7 f1a6 b7a6 h1e1 c8d7 d2d3 e8c8 c1g5 d8e8 h8f6 f7f6 g5f6 c7c5 d4e6 e7c6 e6f8 e8f8 f6e7 f8f5 e7c5 a6a5 d3d4 g6g5 d1e2 h7h5 h2h4 c6b8 h4g5 f5g5');
INSERT INTO evaluations VALUES(57542,30,'cp 527',NULL);
INSERT INTO evaluations VALUES(57542,31,'cp 526',NULL);
INSERT INTO evaluations VALUES(57542,32,'cp 521',NULL);
INSERT INTO evaluations VALUES(57542,33,'cp 525',NULL);
INSERT INTO evaluations VALUES(57542,34,'cp 533',NULL);
INSERT INTO evaluations VALUES(57542,35,'cp 528',NULL);
INSERT INTO evaluations VALUES(57542,36,'cp 530','h5e5 d8e7 e5h8 e7e4 e1d1 g8e7 b1c3 e4f5 f2f3 d7d5 c3b5 e8d7 h8h7 b8c6 h2h4 a7a6 b5c3 d7e8 h4h5 c8e6 h5g6 f5g6 c3e2 g6f6 h7h8 f6h8 h1h8 e7g6 h8h7 f8d6 d2d4 c6e7 c1g5 e8d7 c2c3 a8f8 d1e1 f8f5 g5d2 e6g8 h7h3 g8e6');
INSERT INTO evaluations VALUES(57539,30,'cp 506',NULL);
INSERT INTO evaluations VALUES(57539,31,'cp 528',NULL);
INSERT INTO evaluations VALUES(57539,32,'cp 530',NULL);
INSERT INTO evaluations VALUES(57539,33,'cp 523',NULL);
INSERT INTO evaluations VALUES(57539,34,'cp 509',NULL);
INSERT INTO evaluations VALUES(57539,35,'cp 518',NULL);
INSERT INTO evaluations VALUES(57539,36,'cp 513','d8e7 e5h8 e7e4 e1d1 g8e7 b1c3 e4f5 f2f3 f5f7 c3e4 f7g7 e4f6 e8f7 h8g7 f8g7 f6h7 f7g8 h7g5 d7d5 h2h4 c7c5 c2c4 e7f5 f1d3 b8c6 c4d5 c6e5 d3b1 c5c4 h4h5 e5d3 b1d3 c4d3 g5e4 c8d7 a2a4 g6h5 h1h5 f5d4 a1a3 a8d8 d5d6 g8f7 a3c3 d7a4 d1e1 d4c2 e1f2 g7d4 f2g3 d4c3');
INSERT INTO evaluations VALUES(35243,30,'cp 508',NULL);
INSERT INTO evaluations VALUES(35243,31,'cp 517',NULL);
INSERT INTO evaluations VALUES(35243,32,'cp 532',NULL);
INSERT INTO evaluations VALUES(35243,33,'cp 533',NULL);
INSERT INTO evaluations VALUES(35243,34,'cp 528',NULL);
INSERT INTO evaluations VALUES(35243,35,'cp 537',NULL);
INSERT INTO evaluations VALUES(35243,36,'cp 550','e5h8 e7e4 e1d1 g8e7 b1c3 e4f5 f2f3 f5f7 c3e4 f7g7 e4f6 e8f7 h8g7 f8g7 f6h7 d7d5 h7g5 f7g8 c2c3 c7c5 g2g4 c8d7 h2h4 b8c6 b2b3 g7f6 c1a3 b7b6 d1c2 c6b4 c2b2 a7a5 f1e2 a5a4 f3f4 e7c6 b2b1 b4a6 h4h5 a6c7 b1b2 a8e8 e2d3 f6g5 f4g5 c6e5 d3c2 c7b5');
INSERT INTO evaluations VALUES(34107,30,'cp 548',NULL);
INSERT INTO evaluations VALUES(34107,31,'cp 540',NULL);
INSERT INTO evaluations VALUES(34107,32,'cp 540',NULL);
INSERT INTO evaluations VALUES(34107,33,'cp 537',NULL);
INSERT INTO evaluations VALUES(34107,34,'cp 532',NULL);
INSERT INTO evaluations VALUES(34107,35,'cp 533',NULL);
INSERT INTO evaluations VALUES(34107,36,'cp 538','e7e4 e1d1 g8e7 b1c3 e4f5 f2f3 d7d5 c3b5 e8d7 h8h7 b8c6 d2d4 a7a6 b5c3 d7d6 h2h4 c8e6 f1d3 f5f6 c3e2 f8g7 c1g5 f6f8 e2f4 e6g8 f4g6 g8h7 g6f8 a8f8 d3h7 c6d4 c2c3 d4e6 d1c2 f8h8 g5e7 d6e7 h7f5 e6f4 a1e1 e7f6 g2g4 f4g2 e1e6 f6f7 h4h5 g2f4');
INSERT INTO evaluations VALUES(34102,30,'cp 544',NULL);
INSERT INTO evaluations VALUES(34102,31,'cp 543',NULL);
INSERT INTO evaluations VALUES(34102,32,'cp 547',NULL);
INSERT INTO evaluations VALUES(34102,33,'cp 551',NULL);
INSERT INTO evaluations VALUES(34102,34,'cp 549',NULL);
INSERT INTO evaluations VALUES(34102,35,'cp 540',NULL);
INSERT INTO evaluations VALUES(34102,36,'cp 555','e1d1 g8e7 b1c3 e4f5 f2f3 d7d5 c3b5 b8a6 b5d4 f5f7 f1a6 b7a6 h1e1 c8d7 d2d3 e8c8 c1g5 d8e8 h8f6 f7f6 g5f6 c7c5 d4e6 e7c6 e6f8 e8f8 f6e7 f8f5 d1d2 g6g5 e7c5 a6a5 c2c3 c8b7 a1b1 h7h5 b2b4 a5b4 c3b4 a7a6 b4b5 a6b5 g2g4 f5f3 g4h5 f3h3');
INSERT INTO evaluations VALUES(34103,30,'cp 542',NULL);
INSERT INTO evaluations VALUES(34103,31,'cp 543',NULL);
INSERT INTO evaluations VALUES(34103,32,'cp 546',NULL);
INSERT INTO evaluations VALUES(34103,33,'cp 521',NULL);
INSERT INTO evaluations VALUES(34103,34,'cp 532',NULL);
INSERT INTO evaluations VALUES(34103,35,'cp 544',NULL);
INSERT INTO evaluations VALUES(34103,36,'cp 515','g8e7 b1c3 e4f5 f2f3 d7d5 c3b5 e8d7 h8h7 a7a6 b5c3 d7c6 d2d4 f5f6 f1d3 c8e6 h1e1 e6g8 h7h3 b7b6 c3e2 c6b7 c2c3 g8e6 g2g4 c7c5 f3f4 b8c6 c1d2 a8e8 d4c5 b6c5 f4f5 g6f5 g4g5 f6f7 e2f4 e6d7 h3h5 f7h5 f4h5 f5f4 h5f4 b7a8');
INSERT INTO evaluations VALUES(16450,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(16450,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(16450,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(16450,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16450,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(16450,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(16450,36,'cp 9','d7d5 c4d5 f6d5 f1g2 f8c5 e1g1 e8g8 d2d3 f8e8 c1d2 a7a5 a2a3 d5c3 d2c3 c5b6 f3d2 c8e6 a1c1 e6d5 g2e4 d8d7 d1a4 a8d8 e4d5 d7d5 d2c4 h7h6 c4b6 c7b6 a4c4 d5c4 d3c4 c6d4 c3d4 e5d4 f1e1 d4d3 e2d3 e8e1 c1e1 d8d3 e1e8 g8h7');
INSERT INTO evaluations VALUES(15097,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(15097,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(15097,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(15097,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15097,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(15097,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(15097,36,'cp 11','c4d5 f6d5 f1g2 f8c5 e1g1 e8g8 d2d3 a7a5 d1a4 d5e7 c1d2 e7f5 a4b5 c5b4 e2e3 f5d6 b5a4 c8f5 e3e4 f5d7 d2e3 c6d4 a4d1 d4f3 g2f3 d6b5 a1c1 b5c3 b2c3 b4a3 c1b1 b7b6 d3d4 a3d6 d1c2 d8e7 f3g2 a8d8 f1d1 h7h6');
INSERT INTO evaluations VALUES(15001,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(15001,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(15001,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15001,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(15001,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(15001,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(15001,36,'cp 14','f6d5 f1g2 f8c5 e1g1 e8g8 d2d3 h7h6 c3d5 d8d5 b2b3 d5d6 c1b2 c6d4 e2e3 d4f3 d1f3 d6d3 a1d1 d3g6 f3d5 c5e7 b2e5 c8g4 d1d4 c7c6 d5a5 g6e6 d4f4 f8e8 e5c3 g4h3 g2h3 e6h3 f1d1 h3e6 a5c7 a8c8');
INSERT INTO evaluations VALUES(15344,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15344,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(15344,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15344,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(15344,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(15344,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(15344,36,'cp 1','f1g2 f8c5 d2d3 e8g8 e1g1 h7h6 c3d5 d8d5 c1d2 d5d6 a1c1 c8e6 a2a3 a7a5 d2c3 f8d8 f3d2 c5a7 g2c6 b7c6 d2f3 f7f6 c3a5 a7f2 f1f2 a8a5 f3d2 a5a7 d1c2 e6d5 d2c4 d6e6 c4e3 e6f7 b2b4 f7e6 e3d5 c6d5');
INSERT INTO evaluations VALUES(15345,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15345,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15345,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15345,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15345,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15345,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15345,36,'cp 13','f8c5 e1g1 e8g8 d2d3 h7h6 c3d5 d8d5 a2a3 a7a5 c1d2 d5d6 d2c3 f8d8 a1c1 c8e6 f3d2 d6d7 b2b4 a5b4 a3b4 c6b4 g2b7 a8a2 c3b4 c5b4 d2c4 b4d6 b7f3 d7a4 d1a4 a2a4 c1c2 e6c4 d3c4 g8f8 f1b1 d6e7 b1b5 e7d6 f3d5');
INSERT INTO evaluations VALUES(10154,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(10154,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10154,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(10154,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(10154,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(10154,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(10154,36,'cp -26','c5d4 f3d4 c6d4 c1g5 d8e7 b1c3 c7c6 f2f4 d7d6 d1d2 c8e6 c4d3 h7h6 f4e5 d6e5 g5f6 g7f6 d2f2 h8g8 f2f6 e6h3 f6e7 e8e7 f1f2 h3g4 c3d1 g4d1 a1d1 f7f6 d3c4 g8d8 d1f1 b7b5 c4b3 d8d6 c2c4 b5c4 b3c4 a7a5 h2h4 a5a4 f1c1 a8b8 g1h2 b8b6 c4f1 b6b4 c1c3 d4e6');
INSERT INTO evaluations VALUES(10444,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(10444,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(10444,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(10444,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(10444,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(10444,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(10444,36,'cp -31','f3d4 c6d4 c1g5 d8e7 b1c3 c7c6 f2f4 d7d6 d1d2 c8e6 c4d3 h7h6 f4e5 d6e5 g5f6 g7f6 d2f2 h8g8 f2f6 e6h3 f6e7 e8e7 f1f2 h3g4 a1f1 g8g7 c3d1 g4d1 f1d1 d4e6 f2f5 g7g5 c2c3 a8g8 g2g3 g5f5 e4f5 e6c5 d3c2 h6h5 g1f2 h5h4 f2f3 h4g3 h2g3 a7a5 b2b4 a5b4 c3b4');
INSERT INTO evaluations VALUES(10441,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(10441,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(10441,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(10441,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(10441,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(10441,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(10441,36,'cp -33','c6d4 c1g5 d8e7 b1c3 c7c6 f2f4 d7d6 d1d2 c8e6 c4d3 h7h6 f4e5 d6e5 g5f6 g7f6 d2f2 h8g8 f2f6 e6h3 f6e7 e8e7 f1f2 h3g4 a1f1 a8f8 d3c4 d4e6 g1h1 e6f4 h2h3 g4h5 h1h2 h5g6 g2g3 f4e6 g3g4 e6f4 c3e2 f4e2 f2e2 f8d8 c2c3 g8f8 a2a4 f7f6 a4a5 h6h5 f1a1 h5g4 h3g4 d8d6 a1f1');
INSERT INTO evaluations VALUES(10789,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(10789,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(10789,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(10789,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(10789,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(10789,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(10789,36,'cp -20','c1g5 d8e7 b1c3 c7c6 f2f4 d7d6 d1d2 c8e6 c4d3 h7h6 f4e5 d6e5 g5f6 g7f6 d2f2 h8g8 f2f6 e6h3 f6e7 e8e7 f1f2 h3g4 a1f1 a8f8 d3c4 d4e6 g1h1 e6f4 c3d1 g4d1 f1d1 f8d8 d1f1 d8d4 c4f7 e7f7 g2g3 d4b4 g3f4 b4b2 f4e5 f7e6 f2f6 e6e5');
INSERT INTO evaluations VALUES(10785,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(10785,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(10785,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(10785,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(10785,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(10785,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(10785,36,'cp -43','h7h6 g5h4 d7d6 f2f4 c8g4 h4f6 g4d1 f6d8 a8d8 f1d1 d4c2 b1c3 c2a1 d1a1 e5f4 c3d5 e8d7 d5f4 h8e8 c4d3 e8e5 a1f1 c7c6 h2h4 g7g6 f4e2 d7e7 e2d4 d6d5 d4f3 e5e6 e4e5 c6c5 b2b3 e6b6 f1c1 d8c8 h4h5 g6h5 d3f5 c8c7 g1f2 b6a6 c1c2 f7f6 e5e6 a6b6 f2e3 b6b4');
INSERT INTO evaluations VALUES(9905,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(9905,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(9905,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(9905,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(9905,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(9905,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(9905,36,'cp -12','g5h4 d7d6 f2f4 c8g4 h4f6 g4d1 f6d8 a8d8 f1d1 d4c2 b1c3 c2a1 d1a1 e5f4 c3d5 e8d7 d5f4 h8e8 c4d3 e8e5 a1f1 c7c6 h2h4 g7g6 f4e2 d7e7 e2d4 d6d5 d4f3 e5e6 e4d5 e6e3 d3c2 e3e2 c2d1 e2b2 d5c6 b2b1 d1a4 b1f1 g1f1 b7c6 a4c6 d8d1 f1e2 d1b1 c6d5 f7f5 d5b3 e7f6 e2e3 g6g5 f3d2 b1h1 h4g5 h6g5 g2g3 h1c1 e3f2 g5g4 d2c4');
INSERT INTO evaluations VALUES(9906,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(9906,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9906,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(9906,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(9906,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(9906,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9906,36,'cp -24','d7d6 f2f4 c8g4 h4f6 g4d1 f6d8 a8d8 f1d1 d4c2 b1c3 c2a1 d1a1 e5f4 c3d5 e8d7 d5f4 h8e8 c4d3 e8e5 a1f1 c7c6 h2h4 g7g6 f4e2 d7e7 e2d4 d6d5 d4f3 e5e6 e4d5 e6e3 d3c2 e3e2 c2d1 e2b2 d5c6 b2b1 d1a4 b1f1 g1f1 b7c6 a4c6 d8b8 c6d5 f7f5 d5b3 e7f6 f1e2 b8b4 g2g3 b4e4 e2f2 e4g4 f3d2 f5f4 f2f3 g4g3 f3f4 g3h3 d2f3 h3h1');
INSERT INTO evaluations VALUES(55129,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(55129,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(55129,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(55129,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(55129,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(55129,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(55129,36,'cp -25','e7e5 d3d4 b8c6 g1f3 e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 c8e6 f1e2 d5c4 e2c4 d8d7 d1b3 e6c4 d2c4 e8c8 e1g1 c8b8 f1d1 f8d6 c1d2 g7g5 d4d5 c6e5 c3b5 e5c4 b3c4 d6e5 d2c3 e5c3 c4c3 f6d5 d1d5 d7d5');
INSERT INTO evaluations VALUES(39203,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(39203,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(39203,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(39203,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(39203,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(39203,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(39203,36,'cp 0','g8f6 g1e2 d7d6 d2d3 c7c6 b1c3 b7b5 c4b3 c8e6 b3e6 f7e6 f3h3 d8e7 f2f4 a7a5 f4e5 d6e5 a2a3 e8g8 e2g1 c5a7 g1f3 b8d7 f3g5 a8e8 g5f3');
INSERT INTO evaluations VALUES(39595,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(39595,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(39595,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(39595,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(39595,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(39595,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(39595,36,'cp 17','e3e4 d7d6 g1f3 g8f6 f1d3 e7e5 d4e5 d6e5 e1g1 d8e7 c2c3 e8g8 a2a4 f8d8 d1c2 c7c5 b1d2 b8c6 a4a5 h7h6 f1e1 f6h5 d3f1 h5f4 d2c4 c8g4 f3d2 g4e6 c4e3 a8b8 c2a4 e7d7 d2c4 b7b5 a5b6 a7b6');
INSERT INTO evaluations VALUES(5963,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(5963,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(5963,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(5963,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5963,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(5963,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5963,36,'cp 14','g8h6 g5f3 h6g4 e1g1 a7a5 h2h3 g4f6 c1e3 c5e3 f2e3 e8g8 a2a4 c8e6 c4e6 f7e6 b1d2 b7b6 c2c3 d8d7 d3d4 a8e8 d1e2 d6d5 f3e5 c6e5 d4e5 f6e4 d2e4 d5e4 e2c4 d7d2 f1f8 e8f8 c4e6 g8h8 a1f1 f8f1 g1f1 d2c1 f1f2 c1d2 f2g3 d2e3 g3g4 e3e2 g4f4 e2f2 f4e4 f2g2 e4d3 g2f1 d3d4');
INSERT INTO evaluations VALUES(9461,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(9461,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(9461,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(9461,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(9461,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(9461,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9461,36,'cp 15','g5f3 h6g4 e1g1 a7a5 h2h3 g4f6 c1e3 c5e3 f2e3 e8g8 a2a4 c8e6 c4e6 f7e6 b1d2 c6e7 f3h2 e7g6 d1e2 d8e7 f1f2 a8e8 a1f1 f6d7 h2g4 f8f2 f1f2 e8f8 f2f8 g6f8 d3d4 e5d4 e3d4 c7c5 e2c4 c5d4 c4d4 d7c5 e4e5');
INSERT INTO evaluations VALUES(9463,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(9463,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(9463,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9463,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(9463,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(9463,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(9463,36,'cp 0','h6g4 e1g1 g4f6 a2a4 a7a5 c1e3 c5e3 f2e3 e8g8 h2h3 c8e6 c4e6 f7e6 b1d2 d8d7 d1e2 h7h6 f1f2 f8f7 a1f1 a8f8 b2b3 b7b6 d2c4 d7e8 c4d2');
INSERT INTO evaluations VALUES(54295,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(54295,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(54295,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(54295,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(54295,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(54295,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(54295,36,'cp -60','c5d4 e3d4 b8c6 g1f3 g8e7 b1d2 c6d4 f3d4 e7c6 d2f3 d8c7 d4c6 b7c6 f1d3 a8b8 a1b1 f7f5 e1g1 c6c5 c2c3 a7a5 a2a4 c5c4 d3e2 f8c5 b2b3 c4b3 b1b3 c8d7 b3b8 c7b8 d1d3 e8g8 d3a6 b8b6 a6b6 c5b6 f1b1 f8b8 f3d4 d7a4');
INSERT INTO evaluations VALUES(54435,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(54435,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(54435,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(54435,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(54435,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(54435,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(54435,36,'cp -61','e3d4 b8c6 g1f3 g8e7 b1d2 c6d4 f3d4 e7c6 d2f3 d8c7 d4c6 b7c6 f1d3 a8b8 a1b1 c6c5 c2c3 c5c4 d3c2 f8c5 e1g1 e8g8 b2b4 c4b3 a2b3 a7a5 d1d2 c8a6 f1c1 c5a3 c1e1 a3e7 e1e3 e7c5 f3d4 g7g6 h2h4 f8c8 b1d1');
INSERT INTO evaluations VALUES(54417,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(54417,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(54417,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(54417,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(54417,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(54417,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(54417,36,'cp -52','b8c6 g1f3 g8e7 b1d2 e7f5 d2b3 f5d4 b3d4 d8c7 d4c6 b7c6 f1d3 a8b8 a1b1 c6c5 c2c3 c5c4 d3c2 f8c5 e1g1 c8d7 f1e1 e8g8 b2b3 c4b3 a2b3 f8c8 h2h4 d7b5 d1d2 h7h6 b3b4 c5b6 b1c1 a7a5 f3d4 b5d7 h4h5 a5b4 c3b4');
INSERT INTO evaluations VALUES(20413,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(20413,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(20413,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(20413,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(20413,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(20413,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(20413,36,'cp -60','g1f3 g8e7 b1d2 d8c7 f1b5 e7f5 e1g1 f5d4 f3d4 c8d7 d2f3 a7a6 d4c6 a6b5 c6d4 b5b4 f1e1 c7b6 e1e3 f8c5 c2c3 a8a4 b2b3 a4a8 c3b4 c5b4 h2h4 e8g8 a2a4 f8c8 h4h5 h7h6 g2g4 c8c3 e3c3 b4c3 a1c1 c3b2 c1c2 b2a3');
INSERT INTO evaluations VALUES(20412,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(20412,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(20412,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(20412,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(20412,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(20412,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(20412,36,'cp -50','g8e7 b1d2 c6d4 f3d4 e7c6 d2f3 d8c7 d4c6 b7c6 f1d3 a8b8 a1b1 c6c5 c2c3 f8e7 e1g1 e8g8 f1e1 c8d7 d3c2 c5c4 f3d4 f7f5 d1d2 a7a5 d2f4 g8h8 f4d2 a5a4 g2g3 e7c5 b2b3 c4b3 a2b3 a4a3 b3b4 c5b6 b1a1 b8a8 e1e3 f8c8');
INSERT INTO evaluations VALUES(47232,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(47232,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(47232,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(47232,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(47232,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47232,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(47232,36,'cp 42','e4e5 f6g8 d3d4 c7c5 g1f3 c5d4 d1d4 b8c6 d4f4 e7e6 b1c3 c6b4 f1d3 c8d7 e1g1 b4d3 c2d3 g8e7 h2h4 e7f5 c3e2 f8e7 h4h5 h7h6 e2g3 f5h4 f3d4 g7g5 h5g6 h4g6 f4e3 h6h5 c1d2 h5h4 g3e2 h4h3 g2g3 f7f6 f2f4 f6e5 f4e5');
INSERT INTO evaluations VALUES(47295,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(47295,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(47295,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(47295,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(47295,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(47295,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(47295,36,'cp 32','f6g8 d3d4 c7c5 g1f3 c5d4 d1d4 e7e6 b1c3 b8c6 d4g4 d5d4 c3e2 d4d3 e2c3 d3c2 g4e4 g8e7 a2a3 e7d5 e4c2 c8d7 c1d2 a8c8 f1d3 d5c3 d2c3 c6e7 d3e4 e7d5 e4d5 e6d5 e1g1 f8e7 a1d1 d7e6');
INSERT INTO evaluations VALUES(43264,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(43264,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(43264,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(43264,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(43264,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(43264,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(43264,36,'cp 138','e4e5 f6e4 f1d3 d7d5 c2c4 c7c6 e1g1 e8g8 f1e1 b8a6 a2a3 c8f5 c4d5 c6d5 f3h4 e7e6 h4f5 e6f5 c1e3 a8c8 d1b3 a6c7 b1c3 e4c3 b2c3 c7e6 a1c1 f5f4 e3d2 c8c7 d3e2 d8d7 a3a4 h7h5 c1b1 b7b6 b3b5 f8c8 b5d7 c7d7 g2g3 g7h6');
INSERT INTO evaluations VALUES(43124,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(43124,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(43124,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(43124,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(43124,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(43124,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(43124,36,'cp 126','f6e4 f1d3 d7d5 c2c4 c7c6 e1g1 e8g8 f1e1 b8a6 a2a3 c8g4 h2h3 g4f5 c1e3 a6c7 a3a4 f5e6 a4a5 f7f5 c4d5 e6d5 b1c3 e4c3 b2c3 b7b5 d1e2 c7e6 h3h4 d8d7 h4h5 f5f4 e3c1 g6g5 c1a3 f8f7 a1c1');
INSERT INTO evaluations VALUES(20974,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(20974,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(20974,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(20974,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(20974,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(20974,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(20974,36,'cp 69','g8f6 g1f3 e7e6 a2a3 f8e7 e2e3 e8g8 b2b4 c6b8 c4c5 c7c6 f1d3 b7b6 e1g1 c8a6 d3a6 b8a6 d1a4 a6b8 c1d2 d8c7 c5b6 c7b6 a4c2 b8d7 c3a4 b6b7 f1c1 f8c8 a1b1 b7a6 h2h3 f6e4 d2e1 e7f6 f3d2 e4d2 e1d2 e6e5 d4e5 d7e5');
INSERT INTO evaluations VALUES(46772,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(46772,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46772,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(46772,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46772,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46772,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46772,36,'cp 25','b1d2 f8g7 c2c3 e8g8 f1e2 b8d7 h2h3 b7b6 g1f3 c7c5 e1g1 c8b7 a2a4 a7a6 d1b3 a8c8 f1d1 e7e6 c3c4 h7h5 a1c1 c5d4 e3d4 d5c4 d2c4 f6d5 f4d6 f8e8 d6h2 g7h6 c1c2 h6f8 c4d6 f8d6 h2d6');
INSERT INTO evaluations VALUES(46773,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(46773,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46773,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46773,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46773,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(46773,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(46773,36,'cp 11','f8g7 c2c3 e8g8 f1e2 c7c5 g1f3 b8d7 h2h3 b7b6 e1g1 c8b7 a2a4 a7a6 e2d3 c5c4 d3c2 b6b5 f3e5 d7e5 f4e5 b7c6 f1e1 d8b6 e3e4 a8d8 d1e2 b6b7 a4b5 a6b5 e5h2 d5e4 d2e4 f6d7 b2b4 c4b3 c2b3');
INSERT INTO evaluations VALUES(41473,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(41473,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41473,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(41473,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41473,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(41473,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(41473,36,'cp 27','c2c3 e8g8 f1e2 c7c5 g1f3 b8d7 h2h3 b7b6 e1g1 f8e8 a2a4 a7a6 f3e5 c8b7 b2b4 c5c4 e3e4 d5e4 e2c4 e7e6 e5d7 f6d7 d1e2 e6e5 f4e3 e5d4 c3d4 d7f6 a1c1 f6d5 b4b5 a6a5 f1d1 d5e3 f2e3 a8c8 c4b3 c8c1 d1c1');
INSERT INTO evaluations VALUES(41464,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41464,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41464,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41464,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41464,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41464,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41464,36,'cp 19','e8g8 f1e2 b8d7 h2h3 b7b6 g1f3 c8b7 e1g1 c7c5 a2a4 a7a6 e2d3 c5c4 d3c2 b6b5 d1e2 f8e8 f3e5 d7e5 f4e5 e7e6 b2b3 c4b3 d2b3 b7c6 f1b1 b5a4 e5f6 g7f6 b3c5 c6b5 e2d1 d8c7 c2a4 b5a4 c5a4 f6e7 a4b6');
INSERT INTO evaluations VALUES(36744,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36744,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(36744,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36744,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(36744,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36744,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36744,36,'cp 20','f1e2 c8f5 a2a4 c7c5 g1f3 d8b6 d1b3 b6b3 d2b3 c5c4 b3d2 b8c6 h2h3 f6d7 e1g1 e7e5 d4e5 c6e5 f3e5 g7e5 f4e5 d7e5 d2f3 e5f3 e2f3 f5d3 f1d1 f8d8 d1d2 d8d7 a1d1 g8g7 b2b4 g7f6 a4a5 a8d8');
INSERT INTO evaluations VALUES(32093,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(32093,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(32093,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(32093,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(32093,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(32093,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(32093,36,'cp 21','g8f6 d2d4 b8d7 d4e5 d6e5 a2a4 a7a5 d1e2 g4h5 b1d2 f8d6 e2e1 d7b6 f3h4 h5g4 h2h3 g4d7 h4f5 d7f5 e4f5 e8g8 c4b3 e5e4 d2e4 f8e8 e4f6 d8f6 e1c3 f6f5 f1d1 d6e5 c3d3 f5d3 d1d3 a8d8 d3d8 e8d8');
INSERT INTO evaluations VALUES(29908,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(29908,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(29908,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(29908,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(29908,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(29908,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(29908,36,'cp 33','d2d4 b8d7 d4e5 d6e5 b1d2 h7h6 f1e1 f8c5 h2h3 g4h5 d2f1 h5g6 f1g3 d8e7 c4d3 f6h5 g3f5 g6f5 e4f5 e8g8 g2g4 h5f6 g4g5 h6g5 f3g5 c5d6 g5e4 f6e4 e1e4 d7c5 e4g4 c5d3 c2d3 f7f6');
INSERT INTO evaluations VALUES(43172,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(43172,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(43172,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(43172,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(43172,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(43172,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(43172,36,'cp 29','d2d4 c7c5 d4d5 e8g8 g1f3 e7e6 d5e6 d7e6 d1d8 f8d8 f3e5 b8d7 e5d3 f6g4 c1g5 d8e8 b1a3 a7a6 e1g1 e6e5 e2e4 b7b6 f2f3 g4h6 a3c2 f7f6 g5e3 h6f7 b2b4 f6f5 b4c5 b6c5 e3c5 d7c5');
INSERT INTO evaluations VALUES(43193,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(43193,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(43193,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(43193,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43193,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(43193,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(43193,36,'cp 25','c7c5 d4d5 e8g8 g1f3 e7e6 d5e6 d7e6 d1d8 f8d8 f3e5 b8d7 e5d3 e6e5 b1c3 e5e4 c3e4 f6e4 g2e4 d7b6 d3c5 b6c4 c5b7 c8b7 e4b7 a8b8 c1g5 g7b2 g5d8 b2a1 e1g1 a1b2 b7a6 b8d8 a6c4 d8c8 c4d5 c8c1 f1c1 b2c1 g1g2 g8f8 e2e4 f8g7 g2f1 c1a3 d5b3 f7f6 f2f4 a3c5 f1g2 c5b6');
INSERT INTO evaluations VALUES(56683,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(56683,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(56683,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(56683,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(56683,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(56683,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(56683,36,'cp 35','d6e5 f1d3 f8c5 d1e2 g8e7 b1c3 c8e6 e1g1 e8g8 d3c4 d8c8 f1d1 b8c6 h2h3 g8h8 c4e6 c8e6 c3d5 c5d6 c1e3 b7b6 a2a4 a8e8 c2c3 h7h6 d5e7 c6e7 f3d2 f6f5 e4f5 e7f5 d2e4 a7a5 d1d2 e6g6 e2g4 g6e6 e4d6 f5d6 g4e6 e8e6');
INSERT INTO evaluations VALUES(56837,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(56837,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(56837,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(56837,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(56837,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(56837,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(56837,36,'cp 45','f1d3 f8c5 e1g1 g8e7 d1e2 c8e6 b1c3 b8c6 f1d1 e8g8 d3c4 d8c8 h2h3 g8h8 c4e6 c8e6 c3d5 c5d6 c1e3 b7b6 a2a4 f6f5 e3g5 f5e4 e2e4 e6f5 d5c3 c6d4 g5e7 f5e4 c3e4 d4f3 g2f3 d6e7 d1d7 e7d6 e4d6 c7d6 a1d1 f8f3 d1d6 h8g8 d7e7 f3h3 d6d7');
INSERT INTO evaluations VALUES(10235,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(10235,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10235,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(10235,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(10235,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(10235,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(10235,36,'cp 13','b2b4 c5b6 f1e2 d7d5 b1d2 e8g8 a2a4 a7a6 e1g1 h7h6 h2h3 c8e6 f1e1 d5e4 d3e4 f6h5 d1c2 d8f6 d2c4 h5f4 c1f4 f6f4 c2c1 f8d8 c1f4 e5f4 a1b1 g7g5 c4b6 c7b6 b4b5 a6b5 e2b5 d8c8 e1d1 c8c7 d1d6 a8d8 b1d1 d8d6 d1d6 c7d7 d6d7 e6d7 e4e5');
INSERT INTO evaluations VALUES(10072,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10072,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10072,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(10072,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10072,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(10072,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(10072,36,'cp 1','c5b6 f1e2 d7d5 e4d5 f6d5 e1g1 a7a6 a2a3 d8f6 c3c4 d5f4 c1f4 f6f4 c4c5 b6a7 b1c3 c8e6 b4b5 c6d4 b5a6 b7a6 d1a4 e6d7 a4a5 e8g8 a1e1 d7e6 c3e4 d4b3 a5c7 f8c8 c7e5 f4e5 f3e5 a7c5 e2f3 c5a3');
INSERT INTO evaluations VALUES(9917,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9917,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9917,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9917,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9917,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(9917,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9917,36,'cp 0','a2a4 a7a6 f1e2 e8g8 b1d2 f8e8 d1c2 h7h6 e1g1 d7d5 f1e1 f6g4 e1f1');
INSERT INTO evaluations VALUES(19942,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(19942,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19942,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19942,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19942,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19942,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19942,36,'cp 0','g8f6 d2d3 e7e6 e1g1 f8e7 b1d2 e8g8 f1e1 c6a5 c4b5 c8d7 b5d7 d8d7 d2c4 a5c6 a2a4 b7b6 c1f4 a8d8 d1e2 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(13242,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13242,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13242,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13242,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13242,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13242,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13242,36,'cp 0','d2d3 e7e6 e1g1 f8e7 f1e1 e8g8 b1d2 c6a5 c4b5 c8d7 b5d7 d8d7 d2c4 a5c6 a2a4 d7c7 c1f4 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(13245,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13245,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13245,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13245,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13245,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13245,36,'cp 0','e7e6 b1d2 f8e7 e1g1 e8g8 f1e1 c6a5 c4b5 c8d7 b5d7 d8d7 a2a4 a5c6 d2c4 d7c7 c1f4 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(13935,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13935,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13935,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13935,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13935,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13935,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13935,36,'cp 0','e1g1 f8e7 f1e1 e8g8 b1d2 c6a5 c4b5 c8d7 b5d7 d8d7 a2a4 a5c6 d2c4 d7c7 c1f4 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(13934,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13934,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13934,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13934,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13934,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(13934,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13934,36,'cp 0','f8e7 f1e1 e8g8 b1d2 c6a5 c4b5 c8d7 b5d7 d8d7 d2c4 a5c6 a2a4 a8d8 c1f4 f6h5 f4d2 d7c7 c4a3 a7a6 d1e2 c6a5 e2d1');
INSERT INTO evaluations VALUES(8165,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8165,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8165,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8165,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8165,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8165,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8165,36,'cp 0','b1d2 e8g8 f1e1 c6a5 c4b5 c8d7 b5d7 d8d7 a2a4 a5c6 d2c4 a8d8 c1f4 f6h5 f4d2 d7c7 c4a3 a7a6 d1e2 c6a5 e2d1');
INSERT INTO evaluations VALUES(41213,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(41213,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(41213,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(41213,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(41213,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(41213,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(41213,36,'cp 50','e8g8 d1d2 d6d5 e4d5 f6d5 g5h6 d5c3 h6g7 g8g7 d2c3 b8d7 e1c1 d7f6 f3e5 c8e6 f1c4 e6c4 c3c4 c7c6 f2f3 d8d5 h1e1 e7e6 a2a4 d5c4 e5c4 f6d5 c4d6 a8b8 c2c4 f8d8 d6e4 d5c7 g2g3 h7h6 e4f2 b7b6');
INSERT INTO evaluations VALUES(36483,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(36483,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(36483,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(36483,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(36483,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(36483,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(36483,36,'cp 47','d1d2 d6d5 e4d5 f6d5 g5h6 d5c3 h6g7 g8g7 d2c3 b8d7 e1c1 d7f6 f3e5 d8d6 f1c4 f6d5 c3f3 c8e6 h2h4 h7h5 h1e1 c7c6 c4d3 d5b4 a2a3 b4d3 f3d3 f8d8 d3e3 d6d5 f2f3 d5a2 e3c3 a2a1 c1d2 a1a2 e5d3 a2d5 e1e4 e6f5 e4e7 d5d4 c3d4 d8d4 d2c3');
INSERT INTO evaluations VALUES(36484,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(36484,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(36484,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(36484,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(36484,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(36484,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(36484,36,'cp 39','d6d5 e4d5 f6d5 g5h6 d5c3 h6g7 g8g7 d2c3 b8d7 e1c1 d7f6 f3e5 d8d6 h2h4 h7h5 f2f3 c8e6 f1c4 a8d8 h1e1 e6c4 c3c4 f6d7 a2a3 d7e5 d4e5 d6e6 c4e6 f7e6 d1d8 f8d8 c2c3 d8d5 e1e4 b7b6 c1c2 c7c6 g2g3 a7a5 c3c4');
INSERT INTO evaluations VALUES(50653,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(50653,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(50653,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(50653,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(50653,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(50653,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(50653,36,'cp 27','e1g1 g8f6 f1e1 e7e6 c4f1 f8e7 c2c3 e8g8 d2d4 b7b5 a2a4 c5d4 c3d4 c8b7 f1d3 b8c6 b1c3 c6b4 d3b1 b5a4 a1a4 a6a5 h2h3 b7c6 a4a3 f8e8 f3h2 d8b6 e4e5 f6d7 d1g4 d6e5 d4e5 a5a4 c1h6 e7f8 c3e4 c6e4 g4e4 g7g6');
INSERT INTO evaluations VALUES(50652,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(50652,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(50652,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(50652,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50652,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(50652,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(50652,36,'cp 20','g8f6 f1e1 e7e6 c4f1 f8e7 c2c3 e8g8 d2d4 b7b5 a2a4 c5d4 c3d4 c8b7 f1d3 b8c6 b1c3 c6b4 d3b1 b5a4 a1a4 a6a5 h2h3 d8b6 a4a3 d6d5 e4e5 f6e4 f3h2 f8c8 h2g4 b6d8 c3e4 d5e4 a3g3 g8h8 b1e4 b7e4 e1e4 d8d5 e4e1');
INSERT INTO evaluations VALUES(43379,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(43379,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(43379,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43379,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(43379,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43379,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(43379,36,'cp 20','f1e1 e7e6 c4f1 f8e7 d2d4 c5d4 f3d4 b7b5 a2a4 b5a4 a1a4 e8g8 d4b3 c8d7 a4a1 a6a5 c1f4 d7c6 b1c3 e6e5 f4g5 b8d7 g5f6 d7f6 f1b5 c6b5 c3b5 a5a4 b3d2 d6d5 e4d5 f6d5 d2c4 a8b8 b5a3 d5f6 d1d8 f8d8 e1e5 e7b4 e5e2 b4a3 b2a3');
INSERT INTO evaluations VALUES(14836,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(14836,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14836,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(14836,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(14836,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(14836,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(14836,36,'cp 25','f1d3 c6b4 e1g1 c7c5 e3e4 b4d3 c2d3 c5d4 d1a4 c8d7 a4d4 d5e4 d3e4 d7c6 d4c4 d8b6 f3e5 b6c5 c4c5 f8c5 e5c6 b7c6 c1f4 c5d4 a1c1 e8g8 f1e1 f8d8 g1f1 h7h6 b2b3 g7g5 f4c7 d8d7 c7g3 a8c8 c3a4 c6c5 f2f3 f6h5 g3f2');
INSERT INTO evaluations VALUES(14844,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(14844,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14844,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(14844,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(14844,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14844,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14844,36,'cp 1','c6b4 e1g1 c7c5 e3e4 c5d4 f3d4 d5e4 d3e4 f8e7 e4f3 e6e5 d4b5 d8d1 f3d1 e8g8 f1e1 b4c6 d1f3 f8d8 b5c7 a8b8 c7a6 b7a6 f3c6 c8d7 c6f3 e7d6 h2h3 d7e6 g2g4 a6a5 a2a4 b8c8 e1d1 d6b4 d1d8 c8d8 c3b5 b4c5 c1e3 c5e3 f2e3 h7h6');
INSERT INTO evaluations VALUES(51168,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(51168,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(51168,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(51168,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(51168,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(51168,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(51168,36,'cp 108','g1f3 e7e6 b1d2 c8b7 a2a4 b5b4 f1d3 c7c5 d4d5 e6d5 e1g1 f8e7 f1e1 d7d6 e4d5 g8f6 d2c4 e8g8 c1g5 b8d7 d3f5 f6d5 c4d6 b7c6 f5d7 d8d7 f3e5 d7c7 d6f7 f8f7 e5f7 g8f7 d1h5 f7g8 g5e7 d5e7 h5c5 e7d5 a1d1 a8d8 c2c4 b4c3 b2c3 c7d6 c5d6 d8d6');
INSERT INTO evaluations VALUES(51167,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(51167,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(51167,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(51167,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(51167,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(51167,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(51167,36,'cp 89','c8b7 b1d2 e7e6 a2a4 b5b4 f1d3 c7c5 d4d5 e6d5 e1g1 g8f6 f1e1 d7d6 e4d5 f8e7 c2c4 b4c3 d2e4 f6e4 d3e4 c3b2 c1b2 e8g8 a1b1 b7c8 f3e5 e7f6 e5c4 b8d7 d1c2 f6b2 b1b2 d7f6 b2b6 f8e8 b6d6 c8d7 c4b6 f6e4 e1e4 d8c7 d6d7 c7b6 d5d6 b6c6 e4e8 a8e8');
INSERT INTO evaluations VALUES(30360,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(30360,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(30360,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(30360,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(30360,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(30360,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(30360,36,'cp 122','f1d3 e7e6 e1g1 c7c5 c2c3 g8f6 e4e5 f6d5 d4c5 f8c5 a2a4 b8c6 a4b5 a6b5 a1a8 b7a8 d3b5 d8b8 d1e2 c6e7 b2b4 c5b6 c1d2 e7g6 c3c4 d5e7 d2c3 e8g8 g2g3 a8f3 e2f3 b6c7 f1d1 e7d5 f3e2 d5c3 b1c3 c7e5 c3e4 f8d8 e4c5 d7d6 c5d7 d8d7 b5d7 b8b4');
INSERT INTO evaluations VALUES(30359,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(30359,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(30359,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(30359,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(30359,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(30359,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(30359,36,'cp 103','e7e6 c2c3 c7c5 e1g1 g8f6 e4e5 f6d5 d4c5 f8c5 a2a4 d8c7 a4b5 a6b5 a1a8 b7a8 d3b5 b8c6 d1e2 e8g8 b5d3 h7h6 b2b4 c5b6 c1d2 f7f6 e2e4 f6f5 e4e2 d5e7 b1a3 e7g6 a3c4 c6e7 f1a1 a8f3 e2f3 g6e5 c4e5 c7e5 a1e1 e5b8 f3e2 e7c6 d3b5 b8c7 g2g3 c7b8 b5d3 b6c7');
INSERT INTO evaluations VALUES(30351,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(30351,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(30351,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(30351,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(30351,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(30351,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(30351,36,'cp 122','c2c3 c7c5 e1g1 g8f6 e4e5 f6d5 d4c5 f8c5 a2a4 d8c7 a4b5 a6b5 a1a8 b7a8 d3b5 b8c6 d1e2 e8g8 b5d3 h7h6 c1d2 f7f6 b2b4 c5b6 e2e4 f6f5 e4e2 d5e7 b1a3 e7g6 a3c4 b6a7 f1a1 c6e7 a1a2 e7d5 c4d6 d5c3 d2c3 c7c3 a2a7 g6f4 e2d2 c3d3 d2d3 f4d3 b4b5 a8d5 b5b6 d5f3 g2f3 d3e5');
INSERT INTO evaluations VALUES(6759,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(6759,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(6759,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(6759,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(6759,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(6759,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(6759,36,'cp 94','c6e7 a2a3 c7c6 c2c4 g8f6 b1c3 c6d5 e4d5 e8g8 f1e2 b7b6 e1g1 e7g6 f1e1 d6b8 e2d3 d7d6 b2b4 a7a5 a1b1 a5b4 a3b4 f6h5 f3d2 h5f6 d1c2 f6h5 d3f5 d8g5 f5c8 f8c8 d2e4 g5d8 c4c5 d6c5 b4c5 b6c5 g2g3 f7f5 e4g5 d8d7');
INSERT INTO evaluations VALUES(7012,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(7012,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(7012,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(7012,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(7012,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(7012,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(7012,36,'cp 113','a2a3 c7c6 c2c4 c6d5 e4d5 g8f6 b1c3 e8g8 f1e2 b7b6 e1g1 e7g6 f1e1 d6b8 h2h3 d7d6 c1e3 h7h6 e2d3 g6e7 a3a4 c8f5 d3f5 e7f5 e3d2 d8c8 d1e2 f8e8 g2g4 f5d4 f3d4 e5d4');
INSERT INTO evaluations VALUES(47137,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(47137,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(47137,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(47137,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(47137,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(47137,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(47137,36,'cp 72','c4d5 c7c6 d5c6 b8c6 e2e3 e7e5 f1c4 c8f5 e1g1 a7a6 d2d4 f8d6 f3e5 c6e5 d4e5 d6e5 d1d8 a8d8 b1d2 b7b5 c4b3 f6e4 d2f3 e5f6 f3d4 f5g6 f1d1 e4c5 c1d2 f6d4 e3d4 d8d4 d1e1 e8d7 d2e3 c5b3 a2b3 d4d6 e3f4 d6e6 e1e6 d7e6 a1a6 e6f5 f4g3 h8c8 f2f3');
INSERT INTO evaluations VALUES(47107,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(47107,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(47107,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(47107,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(47107,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(47107,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(47107,36,'cp 69','c7c6 d5c6 b8c6 d2d4 e7e5 f3e5 f8b4 b1c3 d8d4 e5c6 b4c3 b2c3 d4c3 d1d2 c3c6 c1a3 c8e6 e2e3 c6b6 f1d3 a8d8 a1b1 b6b1 d3b1 d8d2 e1d2 e8d7 h1c1 h8c8 c1c8 d7c8 a3b2 f6e8 b1h7 e6a2 h7d3 c8d7 d3b5 d7e7 b5e8 e7e8 b2g7 a7a5 f2f3 a5a4 h2h4 a4a3 h4h5 a2b1 e3e4 a3a2');
INSERT INTO evaluations VALUES(8824,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(8824,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(8824,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8824,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8824,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8824,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8824,36,'cp 0','g7g5 f3h2 c6a5 d1f3 a5c4 d3c4 c8e6 c3d5 f6d7 h2g4 e6d5 c4d5 h6h5 g4e3 g5g4 h3g4 d8h4 g4g5 h4g5 e3f5 g5g4 c1e3 c5e3 f3e3 h8g8 g2g3 g4h3 e3f3 d7f6 a1e1 g8g5 f5d6 c7d6 f3f6 g5g3 f2g3 h3g3 g1h1 g3h3 h1g1');
INSERT INTO evaluations VALUES(8761,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8761,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8761,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8761,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8761,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8761,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8761,36,'cp 0','f3h2 c6a5 d1f3 a5c4 d3c4 c8e6 c3d5 f6d7 h2g4 e6d5 c4d5 h6h5 g4e3 g5g4 h3g4 d8h4 g4g5 h4g5 e3f5 g5g4 c1e3 c5e3 f3e3 h8g8 g2g3 g4h3 e3h6 e8c8 f5e7 c8b8 e7g8 d8g8 f1e1 g8g3 f2g3 h3g3 g1f1 g3f3 f1g1');
INSERT INTO evaluations VALUES(20090,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(20090,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(20090,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(20090,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20090,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(20090,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(20090,36,'cp -18','f1c4 e7e6 e1g1 f8e7 d1e2 c8d7 f1d1 a7a6 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 a1c1 a8c8 a2a3 d8e8 b2b4 h7h6 h2h3 d7e6 b4b5 e6d5 d1d5 c6b8 c1c8 e8c8 f3d2 a6b5 d5b5 b8d7 d2c4 c8c7 c4b6 d7b6 b5b6 c7d7 e2g4 d7g4 h3g4');
INSERT INTO evaluations VALUES(51987,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(51987,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(51987,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51987,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51987,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51987,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51987,36,'cp 0','g8f6 e4e5 f6d5 c2c4 d5b4 b1c3 d7d6 e5d6 f8d6 a2a3 b4c6 g2g3 e8g8 f1g2 e6e5 e1g1 b8a6 f3g5 d6e7 g5h7 g8h7 g2e4 h7g8 d1h5 g7g6 e4g6 f7g6 h5g6 g8h8 g6h5 h8g8');
INSERT INTO evaluations VALUES(44249,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44249,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44249,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(44249,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44249,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44249,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44249,36,'cp 0','e4e5 f6d5 c2c4 d5e7 b2b4 b7b6 b4b5 c8b7 c1b2 d7d5 e5d6 e7f5 g2g3 f8d6 b1d2 b8d7 d2e4 d6e7 f1g2 d7f6 e4d2 f6d7');
INSERT INTO evaluations VALUES(16650,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(16650,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(16650,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(16650,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(16650,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(16650,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(16650,36,'cp -78','f6e4 c4f7 e8f7 d1h5 g7g6 h5d5 f7g7 d4c6 d8e8 d5e5 e8e5 c6e5 f8b4 c2c3 b4d6 e5c4 h8e8 e1g1 d6f8 f1e1 g7f7 b1d2 e4c5 e1e8 f7e8 c4e3 c7c6 d2f3 d7d6 h2h3 f8e7 b2b3 a7a5 c1a3 a5a4 a3c5 d6c5 a1e1 c8e6 c3c4 e8d7 e3g4 d7d6 f3e5 a4b3 a2b3 a8a2');
INSERT INTO evaluations VALUES(16857,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(16857,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(16857,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(16857,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(16857,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(16857,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(16857,36,'cp -73','c4f7 e8f7 d1h5 g7g6 h5d5 f7g7 d4c6 d8e8 d5e5 e8e5 c6e5 f8b4 c2c3 b4d6 e5f3 b7b6 e1g1 c8a6 f1e1 e4c5 c1e3 h8e8 b2b3 a6b7 b1d2 g7g8 e1d1 d6e7 a2a4 a7a5 f3d4 e7f6 d4b5 c5e6 d2f3 b7f3 g2f3 a8d8 d1d2 d7d5 a1d1 c7c6 b5d4 e6d4 e3d4 f6d4 d2d4');
INSERT INTO evaluations VALUES(48288,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48288,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48288,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48288,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48288,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48288,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48288,36,'cp 2','f8d6 f4d6 c7d6 c2c4 d5c4 f1c4 d8b6 c4d3 b8c6 d2c4 b6b4 d1d2 b4d2 c4d2 c6b4 d3b1 c8d7 a2a3 b4d5 e3e4 d5b6 b1d3 b6a4 a1b1 e8e7 f2f3 h8c8 g1e2 a7a6 e1f2 c8c7 h2h4 d7b5 d3b5 a6b5 h4h5 c7c2 f2e3');
INSERT INTO evaluations VALUES(41792,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(41792,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41792,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(41792,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41792,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(41792,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41792,36,'cp 1','f4d6 c7d6 c2c4 e8g8 c4d5 f6d5 g1f3 c8d7 f1d3 b8c6 e1g1 h7h6 d2e4 d7e8 e4c3 c6e7 d1d2 d5c3 b2c3 b7b5 a1b1 a8b8 e3e4 d8c7 f1c1 e8c6 d2e3 f8c8 h2h4 a7a6 e3f4 c6e8 f4g3 f7f6 e4e5 d6e5');
INSERT INTO evaluations VALUES(57608,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(57608,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(57608,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(57608,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(57608,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(57608,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(57608,36,'cp 64','g2g3 g8f6 c2c4 e7e6 f1g2 f8e7 b2b3 e8g8 c1b2 b7b6 e1g1 c8b7 e2e3 a7a5 a2a3 d7d6 b1c3 d8e8 d2d3 b8a6 d3d4 e8h5 a1c1 f6g4 h2h3 g4f6 f1e1 a8d8 f3d2 h5d1 c1d1 b7g2 g1g2 c7c6 c3a4 d8b8 d2f3 a6c7');
INSERT INTO evaluations VALUES(57609,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(57609,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(57609,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(57609,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(57609,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(57609,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(57609,36,'cp 63','e7e6 f1g2 g8f6 e1g1 f8e7 c2c4 e8g8 b2b3 b7b6 e2e3 c8b7 c1b2 a7a5 a2a3 d8e8 b1c3 d7d6 d2d4 f6e4 c3b5 e7d8 f3d2 b8d7 a1c1 h7h5 d2e4 f5e4 g2h3 d6d5 c4d5 b7d5 b5c7 d8c7 c1c7 d7f6 a3a4');
INSERT INTO evaluations VALUES(55905,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(55905,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55905,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(55905,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(55905,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(55905,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55905,36,'cp -12','f8e7 g2g3 a7a5 f1g2 g8f6 g1e2 a5a4 e1g1 e8g8 f1e1 c7c5 c2c3 b8c6 d3d4 c5d4 e2d4 d8c7 d1e2 e7c5 e4d5 c6d4 c3d4 c5d4 d2f3 d4c5 d5e6 c8e6 c1e3 c7b6 f3g5 e6g4 e3c5 b6c5 g5e4 f6e4 e2e4 g4e6');
INSERT INTO evaluations VALUES(31828,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31828,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31828,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31828,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31828,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31828,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(31828,36,'cp 8','e7e6 e2e3 g8f6 g1f3 c7c6 f1d3 f5g6 e1g1 f8e7 d1e2 d8b6 f1b1 e8g8 d3g6 h7g6 e3e4 c6c5 d4c5 e7c5 e4d5 f6d5 c3d5 e6d5 b2b4 c5d6 f4d6 b6d6 c2c4 d5c4 e2c4 b8c6 b4b5 c6e5 f3e5 d6e5 b1d1 f8e8 c4b3 a7a6 b5a6');
INSERT INTO evaluations VALUES(32289,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32289,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32289,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32289,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32289,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32289,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32289,36,'cp 10','e2e3 g8f6 g1f3 c7c6 f1d3 f5g6 e1g1 f8e7 f1e1 f6h5 f4b8 a8b8 g2g3 d8b6 a1b1 h5f6 d3g6 h7g6 e3e4 d5e4 c3e4 e8g8 d1e2 f6e4 e2e4 f8d8 c2c3 c6c5 d4c5 e7c5 e4e2 b6c7 b1d1 c5e7 h2h3 b7b6 d1d8 b8d8 e1d1 d8d1 e2d1 c7c4');
INSERT INTO evaluations VALUES(32287,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(32287,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(32287,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32287,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(32287,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(32287,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(32287,36,'cp 0','f8d6 f4d6 c7d6 g2g4 f5e4 c3e4 d5e4 h2h4 d8b6 c2c4 b6b2 g1e2 b8c6 a1b1 b2a2 e2c3 a2a5 d1d2 d6d5 c4d5 e6d5 b1b5 a5a1 b5b1');
INSERT INTO evaluations VALUES(56018,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(56018,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(56018,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(56018,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(56018,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(56018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56018,36,'cp 0','a7a6 c1g5 f7f6 g5d2 b8c6 e2e4 c6d4 f1c4 g8e7 f3d4 d8d4 d1e2 d4d7 e1g1 e7c6 d2e3 f8d6 f2f4 d7e7 e4e5 f6e5 c3e4 e5f4 e4d6 c7d6 f1f4 h8f8 f4f8 e8f8 c4d3 f8g8 e2h5 g7g6 d3g6 h7g6 h5g6 e7g7 g6e8 g7f8 e8g6');
INSERT INTO evaluations VALUES(50982,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50982,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50982,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(50982,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(50982,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50982,36,'cp -3','e2e3 b7b5 a2a4 b5b4 c3b1 g8f6 f1c4 c7c5 b1d2 c8b7 d1e2 b8c6 e1g1 c5d4 d2b3 c6a5 b3a5 d8a5 e3d4 f8d6 c1g5 a5f5 g5f6 b7f3 e2f3 f5f3 g2f3 g7f6 d4d5 d6e5 d5e6 f7e6 c4e6 e5b2 a1a2 b2c3 a2e2 a8d8 f1c1');
INSERT INTO evaluations VALUES(34201,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(34201,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(34201,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(34201,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(34201,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(34201,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(34201,36,'cp 155','f8b4 d2d4 e5d4 c1g5 f6d6 d1d4 d6d4 f3d4 g8f6 e1c1 b4c3 b2c3 f6e4 h1e1 e8g8 c4f7 f8f7 e1e4 d7d5 e4e8 f7f8 e8f8 g8f8 d1e1 c8d7 g5e7 f8g8 e7d6 c6c5 d4b3 c5c4 b3c5 b8c6 c5b7 a8e8 e1e8 d7e8 d6c5 e8d7 c1d2 h7h6 c5d4 d7e6 h2h3 g8f7 g2g4 h6h5 f2f3 a7a6 d2e3 h5g4 h3g4');
INSERT INTO evaluations VALUES(32997,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(32997,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(32997,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(32997,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(32997,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(32997,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(32997,36,'cp 160','d2d4 e5d4 c1g5 f6d6 d1d4 d6d4 f3d4 g8f6 f2f3 e8g8 c4b3 b8a6 e1c1 a6c5 c3e2 f6e8 c2c3 c5b3 a2b3 f7f6 g5f6 b4c3 f6g7 c3b2 c1b2 e8g7 h2h4 d7d5 e4d5 c6d5 h4h5 g7e6 h5h6 f8e8 e2c3 e6d4 d1d4 c8e6 h1h5 a8d8 h5g5 g8f8 c3d5');
INSERT INTO evaluations VALUES(9331,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9331,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9331,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9331,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9331,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9331,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9331,36,'cp 0','c6a5 c4b3 e8g8 e1g1 a5b3 a2b3 a7a6 c3e2 f6h5 e2c3');
INSERT INTO evaluations VALUES(9533,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(9533,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9533,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(9533,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(9533,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9533,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9533,36,'cp 9','c4b3 a7a6 c3e2 h7h6 e2g3 a5b3 a2b3 e8g8 e1g1 f8e8 f1e1 c5b6 c1d2 f6h7 b3b4 h7f8 c2c4 f8g6 d2c3 c7c5 b4c5 d6c5 g3f5 b6c7 f5e3 c8e6 b2b3 d8d6 e3d5 a8d8 g1h1 b7b5 c3a5 c7a5 a1a5');
INSERT INTO evaluations VALUES(14364,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(14364,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(14364,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(14364,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(14364,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(14364,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(14364,36,'cp 10','e7e5 g1f3 e5d4 e3d4 f8d6 c1g5 e8g8 e1g1 h7h6 g5h4 c8g4 f1e1 g8h8 e1e3 d6e7 c4b3 g4f3 d1f3 c6d4 f3b7 a8b8 b7a7 b8a8 a7b7 a8b8 b7a6 d4b3 a2b3 b8b3 a6e2 b3b4 h4g3 e7d6 g3d6 d8d6 a1a6 b4b6 a6b6 c7b6');
INSERT INTO evaluations VALUES(15211,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(15211,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(15211,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15211,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(15211,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(15211,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(15211,36,'cp 12','g1f3 e5d4 e3d4 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 a2a4 d6e7 h4g3 g4f3 d1f3 d8d4 c4b5 e7d6 b5c6 b7c6 f3c6 d6g3 h2g3 d4b6 c6b6 c7b6 e1e7 h8g8 a1e1 f8d8 g1f1 d8d2 e1e2 d2d3 e7c7 d3d7 c7d7 f6d7 e2d2 d7c5 f2f3 g8f8');
INSERT INTO evaluations VALUES(15212,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(15212,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(15212,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(15212,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15212,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(15212,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(15212,36,'cp 4','e5d4 e3d4 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 a1c1 a8b8 c3e4 d6e7 h4f6 e7f6 e4f6 d8f6 c4d5 f8d8 d5c6 b7c6 b2b3 b8b5 e1e4 f6g6 e4e5 g4f3 d1f3 b5e5 d4e5 d8d2 h2h4 d2a2 f3c6 g6c6 c1c6 a2b2 c6a6 b2b3');
INSERT INTO evaluations VALUES(15234,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(15234,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(15234,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(15234,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(15234,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15234,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(15234,36,'cp 18','e3d4 f8d6 c1g5 e8g8 e1g1 h7h6 g5h4 c8g4 f1e1 g8h8 a2a4 g4f3 d1f3 c6d4 f3d3 d4c6 a1d1 c6e5 d3f1 e5c4 f1c4 f8e8 c4f7 e8e1 d1e1 d8g8 f7g8 h8g8 h4f6 g7f6 g2g3 g8f7 c3b5 d6e5 f2f4 e5b2 e1b1 a7a6 b5c7 a8c8 b1b2 c8c7 g1g2 b7b5 a4b5 a6b5 b2b5 c7c2 g2h3 f7g6 b5b6');
INSERT INTO evaluations VALUES(15233,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(15233,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(15233,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(15233,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(15233,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(15233,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15233,36,'cp 0','f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 c4d5 g4f3 d5f3 c6d4 f3b7 a8b8 c3d5 c7c5 h4f6 g7f6 d1h5 h8g7 h5g4 g7h8');
INSERT INTO evaluations VALUES(45947,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(45947,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(45947,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(45947,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(45947,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(45947,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(45947,36,'cp 65','f1b5 b8d7 d4f5 a7a6 b5a4 b7b5 a4b3 d7c5 c1g5 c8f5 e4f5 f8e7 e1g1 e8g8 g5f6 e7f6 c3d5 e5e4 c2c3 f8e8 b3c2 e8e5 f1e1 a8c8 g2g3 h7h5 h2h4 e5f5 c2e4 c5e4 e1e4 c8c5 d5e3 f5e5 e4e5 f6e5 d1f3 g7g6 f3b7 d8c8 b7c8 c5c8 f2f4 e5g7 a1d1');
INSERT INTO evaluations VALUES(45891,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(45891,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(45891,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(45891,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(45891,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(45891,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(45891,36,'cp 65','b8d7 d4f5 a7a6 b5a4 b7b5 a4b3 d7c5 c1g5 c8f5 e4f5 f8e7 e1g1 e8g8 g5f6 e7f6 c3d5 e5e4 c2c3 f8e8 b3c2 e8e5 f1e1 a8c8 d1d2 g7g6 a1d1 f6g5 d2d4 e5f5 c2e4 c5e4 d4e4 f5e5 e4d4 c8c4 d4a7 e5e1 d1e1 c4c5 a7b7 d8c8 b7c8 c5c8 g2g3 c8c5 e1d1 h7h5 g1g2 g8g7 h2h4 g5d8');
INSERT INTO evaluations VALUES(1588,30,'cp 246',NULL);
INSERT INTO evaluations VALUES(1588,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(1588,32,'cp 280',NULL);
INSERT INTO evaluations VALUES(1588,33,'cp 274',NULL);
INSERT INTO evaluations VALUES(1588,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(1588,35,'cp 305',NULL);
INSERT INTO evaluations VALUES(1588,36,'cp 257','d2d4 e6d6 d4d5 c6e5 c1f4 e5d3 d1d3 d6f4 d3b5 e8d8 g1f3 g7g6 a1d1 f8g7 e1g1 a7a6 b5a5 b7b6 a5a3 c8b7 g2g3 f4d6 a3a4 b6b5 a4b3 g7c3 b3c3 g8f6 f3g5 h8f8 a2a4 b5b4 c3b3 h7h6 g5f3 b7d5 c2c4 b4c3 b3c3 a8b8 e2c4 c7c6 f1e1 d8e8');
INSERT INTO evaluations VALUES(1579,30,'cp 278',NULL);
INSERT INTO evaluations VALUES(1579,31,'cp 290',NULL);
INSERT INTO evaluations VALUES(1579,32,'cp 285',NULL);
INSERT INTO evaluations VALUES(1579,33,'cp 289',NULL);
INSERT INTO evaluations VALUES(1579,34,'cp 280',NULL);
INSERT INTO evaluations VALUES(1579,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(1579,36,'cp 267','e6d6 d4d5 c6e5 c1f4 e5d3 d1d3 d6f4 d3b5 e8d8 g1f3 g7g6 a1d1 f8g7 e1g1 a7a6 b5a5 b7b6 a5a3 c8b7 g2g3 f4d6 a3a4 d6d7 a4b3 d8e8 f1e1 e8f8 c3e4 d7f5 e2f1 h7h6 c2c4 g8f6 f3d4 f5d7 e4c5 b6c5 b3b7 a8e8 d4c6 h6h5 h2h3 h5h4 g3g4 d7d6 b7a6');
INSERT INTO evaluations VALUES(19205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19205,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(19205,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19205,36,'cp 14','b7c6 f1c4 d7d6 e1g1 f8g7 f1e1 d8b6 c4b3 g8f6 h2h3 e8g8 b1c3 c8b7 c1g5 h7h6 g5e3 b6c7 d1d2 g8h7 a1d1 a8d8 f2f3 c6c5 c3d5 f6d5 b3d5 b7d5 e4d5 g7b2 e3h6 f8h8 h6g5 b2f6 c2c4 d8b8 g5f6 e7f6 d1b1 b8b1');
INSERT INTO evaluations VALUES(18420,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18420,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(18420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18420,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(18420,36,'cp 0','f1c4 f8g7 b1c3 d7d6 h2h3 g8f6 e4e5 d6e5 d1d8 e8d8 c4b3 e7e6 c1d2 d8e7 e1c1 h8d8 c3a4 d8d2 c1d2 e5e4 a4c5 e7d6 c5a4');
INSERT INTO evaluations VALUES(18413,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18413,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18413,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18413,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18413,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18413,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(18413,36,'cp 0','d7d6 e1g1 f8g7 b1c3 g8f6 c1f4 e8g8 d1d2 f6e4 c3e4 d6d5 d2b4 d5c4 a1d1 c6c5 b4c5 d8b6 c5e7 c8f5 e4d6 f5c2 d1d2 c2d3 f1c1 a8d8 h2h3 g7b2 d6c4 d3c4 d2d8 b6d8 e7d8 f8d8 c1c4 b2d4 g2g4 g8f8 c4a4 d4b6 g1g2 f8e7 f4e3 d8c8 a4e4 e7d7 e3b6 a7b6 e4d4 d7e6 d4e4');
INSERT INTO evaluations VALUES(18763,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(18763,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18763,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(18763,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(18763,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(18763,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(18763,36,'cp -2','c6d4 d2d3 g8e7 g2g3 e7c6 f1g2 d7d6 a1b1 f8e7 a2a3 d4f3 d1f3 e7f6 c1d2 e8g8 b2b4 a7a6 e1g1 a8b8 c3e2 c8d7 f1c1 b7b6 f3e3 c6d4 e2d4 f6d4 e3e2 e6e5 d2e3 d4e3 e2e3 d8c7 h2h4 f8c8 f2f4 b6b5');
INSERT INTO evaluations VALUES(21029,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(21029,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(21029,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(21029,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(21029,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(21029,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(21029,36,'cp 36','f1g2 g8f6 g1f3 c8f5 e1g1 e7e6 c2c4 f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 c6e5 d4e5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c7c6 d5c4 b7b5 c4d3 f5d3 e2d3 f8d8 c1e3 d8d3 f1d1 d3d1 a1d1 a8d8 d1c1 d8c8 g1g2 f7f6 e5f6 e7f6 b2b3 f6e7 g2f3 c6c5 f3e4 c5c4 b3c4 b5c4 e4e5');
INSERT INTO evaluations VALUES(21030,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(21030,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(21030,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(21030,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(21030,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(21030,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(21030,36,'cp 35','g8f6 g1f3 c8f5 c2c4 e7e6 e1g1 f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 c6e5 d4e5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c7c6 d5g2 f8d8 e2e4 f5e6 c1e3 f7f6 e5f6 e7f6 f2f4 f6b2 a1b1 b2d4 e3d4 d8d4 b1b7 e6a2 f1a1 d4d2 e4e5 a2d5 g2d5 d2d5 a1a4 d5b5 b7c7 b5b4 a4a3 b4b2 c7c6 a8d8 g1f1 b2h2 a3a5');
INSERT INTO evaluations VALUES(14303,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14303,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(14303,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(14303,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(14303,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(14303,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(14303,36,'cp 53','g1f3 c8f5 c2c4 e7e6 e1g1 f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 c6e5 d4e5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c7c6 d5g2 f8d8 e2e4 f5e6 c1e3 f7f6 e5f6 e7f6 f2f4 f6b2 a1b1 b2d4 e3d4 d8d4 b1b7 e6a2 f1a1 a2f7 e4e5 f7d5 g2d5 d4d5 b7c7 a5a4 a1a3 d5b5 c7c6 b5b4 c6c7 b4b2 c7e7');
INSERT INTO evaluations VALUES(25554,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(25554,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(25554,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(25554,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(25554,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(25554,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25554,36,'cp 47','d1e2 f8e7 g5f3 f6d5 f3e5 c7c6 b5d3 d7e6 e1g1 d5f4 e2e4 f4d3 e5d3 e8g8 d3f4 e6d7 d2d4 a5c4 e4d3 b7b5 f1e1 f8e8 b1d2 e7d6 e1e8 d8e8 f4e2 e8e6 d2c4 b5c4 d3d1 e6g4 h2h3 g4e4 e2g3 d6g3 f2g3 a8e8 c1f4');
INSERT INTO evaluations VALUES(25555,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(25555,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(25555,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25555,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(25555,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(25555,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(25555,36,'cp 29','f8e7 g5f3 f6d5 f3e5 c7c6 b5d3 d7e6 e1g1 d5f4 e2e4 f4d3 e5d3 e8g8 d3f4 e6d7 d2d4 a5c4 e4d3 c4d6 c1e3 d7f5 d3b3 g7g5 f4d3 f5e6 b3a4 a7a5 c2c3 e6f5 a4d1 a5a4 h2h3 d6c4 d1e2 d8d5 b1d2 f5d3 e2d3 c4b2 d3c2 b2c4 d2c4 d5c4');
INSERT INTO evaluations VALUES(24224,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24224,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(24224,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(24224,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(24224,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(24224,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24224,36,'cp 30','g5f3 f6d5 f3e5 c7c6 b5d3 d7e6 e1g1 d5f4 e2e4 f4d3 e5d3 e8g8 d3f4 e6d7 d2d4 a5c4 e4e2 c4d6 c1e3 f8e8 b1d2 d6f5 e2d3 f5h4 a1e1 g7g5 f4e2 d7f5 d3c4 d8d5 c4d5 c6d5 e1c1 f5g6 e2c3 a8d8 f2f3 f7f6 e3f2 h4f5 f1e1 e7b4 a2a3 e8e1 f2e1');
INSERT INTO evaluations VALUES(21739,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(21739,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(21739,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(21739,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(21739,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(21739,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(21739,36,'cp -87','h7h5 h2h4 g8f6 g1f3 f6g4 d2d3 f8b4 c2c3 b4e7 b1a3 c8e6 a3c2 f7f6 d1d2 d8d7 b2a3 a8d8 a3e7 c6e7 a2a4 c7c6 e1g1 e8g8 f1b1 e7f5 a4a5 f5d6 b1e1 d6b5 c3c4 d5c4 b3c4 e6c4 d3c4 d7d2 f3d2');
INSERT INTO evaluations VALUES(22417,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(22417,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(22417,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(22417,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(22417,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(22417,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(22417,36,'cp 80','g1f3 e7e5 c2c4 g7g6 f1e2 g8f6 b1c3 f8g7 c1e3 d8e7 d1c2 e8g8 e1g1 e5d4 f3d4 c7c6 a1d1 f8e8 e3g5 d7c5 f2f3 a7a5 c2c1 a5a4 d4c2 c8e6 g5f4 a8d8 c1d2 e7c7 g1h1 c7b6 f4g5 d8d7 g5e3 b6b2 d1b1');
INSERT INTO evaluations VALUES(22416,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(22416,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(22416,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(22416,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(22416,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(22416,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(22416,36,'cp 71','e7e5 c2c4 g8f6 b1c3 g7g6 d1c2 f8g7 f1e2 e8g8 e1g1 e5d4 f3d4 f8e8 c1g5 d7c5 f2f3 c7c6 a1d1 a7a5 g1h1 a5a4 c2c1 d8c7 d4c2 c8e6 g5e3 f6d7 e3d4 g7d4 c2d4 d7f6 f1e1 f6d7 d4c2 g8h8 c2e3 a8d8 f3f4');
INSERT INTO evaluations VALUES(34585,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(34585,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(34585,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(34585,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(34585,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(34585,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(34585,36,'cp -78','d5e4 g1e2 e4c4 d2d4 g7g6 b1c3 f8g7 a2a3 c4c6 h2h4 h7h5 e2f4 a7a6 c1e3 g8f6 d1b3 c6d6 a1d1 b8c6 f1c4 e8g8 f4g6 c6a5 b3a2 a5c4 a2c4 c8e6 d4d5 b7b5 c4b3 f7g6 d5e6 d6e5 c3e2 a8d8 d1d8 f8d8 b3c2 e5e6 c2g6 e6e4 g6e4 f6e4 e3c1 e4c5 e2f4 c5b3 e1g1 b3c1');
INSERT INTO evaluations VALUES(34636,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(34636,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(34636,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(34636,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(34636,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(34636,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(34636,36,'cp -69','g1e2 e4c4 d2d4 g7g6 b1c3 f8g7 a2a3 c4c6 e2g3 c6e6 f1e2 g8f6 e1g1 e8g8 e2f3 e6b6 f1e1 b8c6 d4d5 c6d4 e1e7 d4f3 d1f3 c8g4 f3e3 b6d6 e3e5 d6e5 e7e5 g4d7 c1f4 c7c6 a1d1 h7h6 d5c6 d7c6 e5e1 f8d8 f2f3 d8d1 e1d1');
INSERT INTO evaluations VALUES(18398,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18398,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(18398,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(18398,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18398,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(18398,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18398,36,'cp -8','g8f6 e1g1 f8c5 b1c3 e8g8 h2h3 d7d6 d1f3 c8e6 c4b3 c5d4 f1d1 h7h6 c3a4 a7a5 c2c3 d4a7 c1e3 e6b3 a2b3 d6d5 e3a7 a8a7 f3e3 a7e7 e4d5 c6d5 a4c5 f8e8 c3c4 d5d4 e3d2 d8b6 d2a5 b6d6 b3b4 e5e4 a5a3 g8h7 a3g3 d6g3 f2g3');
INSERT INTO evaluations VALUES(11988,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(11988,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(11988,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(11988,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(11988,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(11988,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(11988,36,'cp -6','b1c3 f8b4 e1g1 e8g8 c1g5 h7h6 g5h4 d8e7 d1e2 g7g5 h4g3 d7d6 a1d1 b4c3 b2c3 f8d8 d1b1 g8g7 f1d1 c8e6 c4e6 e7e6 c3c4 a8b8 b1b3 a7a5 f2f3 f6h5 b3b8 d8b8 a2a3 c6c5 e2d2 h5f4 d2d6 e6d6 d1d6 b8b2 g3f4 e5f4 d6c6 b2c2');
INSERT INTO evaluations VALUES(11985,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(11985,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(11985,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(11985,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(11985,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11985,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(11985,36,'cp -12','f8b4 e1g1 e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 d8e7 d1e2 d7d6 a1d1 b4c3 b2c3 a8b8 d1d2 f8d8 f1d1 a7a5 a2a4 g8g7 c4b3 b8b6 f2f3 f6h5 g3f2 c8a6 e2e3 c6c5 g2g3 h5f6 c3c4 a6b7 d2d3 g5g4 f2e1 g4f3 e1a5 f3f2 e3f2 f6e4');
INSERT INTO evaluations VALUES(7664,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7664,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7664,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(7664,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(7664,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7664,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7664,36,'cp 0','e1g1 e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 d8e7 d1e2 d7d6 a2a3 b4c3 b2c3 g8h7 f1d1 a8b8 d1d2 a7a5 a1d1 f8d8 a3a4 h7g7 h2h3 d8d7 d2d3 d7d8 d3d2');
INSERT INTO evaluations VALUES(7663,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7663,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7663,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7663,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7663,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7663,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7663,36,'cp -4','e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 d8e7 d1e2 d7d6 a2a3 b4c3 b2c3 a8b8 a1b1 b8b1 f1b1 g8g7 f2f3 f6h5 g3f2 h5f4 e2e3 c8e6 c4f1 c6c5 c3c4 e7c7 f2e1 f8b8 b1b8 c7b8 e3b3 b8c7 e1d2 f4h5 b3b5 h5f6');
INSERT INTO evaluations VALUES(32119,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(32119,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(32119,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(32119,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(32119,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(32119,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(32119,36,'cp 55','b8d7 h2h3 g4f3 e2f3 f8e7 e1g1 e7g5 c1g5 d8g5 b1c3 g5d8 a2a4 a7a5 d1d2 g8f6 a1d1 e8g8 f3g4 e5d4 d2d4 d7e5 g4e2 d8e7 f1e1 e5c6 d4c4 c6b4 c3b5 e7e4 c4e4 f6e4 b5c7 a8c8 c2c3 c8c7 c3b4 a5b4 d1d4 e4c5 d4b4 b7b6');
INSERT INTO evaluations VALUES(50590,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(50590,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50590,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50590,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(50590,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(50590,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(50590,36,'cp -2','e7e6 d2d4 b7b5 c4d3 c5c4 d3e2 g8f6 e4e5 f6d5 c2c3 b8c6 b2b3 c6a5 f3d2 c8b7 e1g1 d8c7 d2e4 f8e7 e2f3 e8g8 f1e1 b7c6 c1a3 e7a3 b1a3 f7f5 e5f6 d5f6 b3b4 c6e4 f3e4 f6e4 e1e4 a5b7 d1e2 a8e8 e4e5 b7d6 a1e1 f8f5 a3c2 f5e5 d4e5 d6f5');
INSERT INTO evaluations VALUES(21283,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(21283,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(21283,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(21283,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(21283,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21283,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21283,36,'cp 38','e1g1 g7g5 b1c3 f8g7 a2a4 g8e7 a4a5 a7a6 f3e1 e7g6 d1h5 c6e7 g2g3 d8d7 c3d5 e8g8 d5e3 d7h3 h5h3 c8h3 e1g2 h3e6 c1d2 a8e8 b2b3 e7c6 c2c3 c6e7 f2f3 e6c4 b3c4 f7f5 e3d5 e7d5 c4d5 f5f4 g2e1 f8f7');
INSERT INTO evaluations VALUES(21282,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(21282,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(21282,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21282,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(21282,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(21282,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21282,36,'cp 25','c6a5 c4d5 c7c6 d5f7 e8f7 b2b4 g8f6 b4a5 d8a5 b1d2 f8e7 d2c4 a5c7 d3d4 c8e6 d1d3 b7b5 c4e3 h8d8 c2c4 f7g8 c4b5 c6b5 d4d5 e6d7 f3h4 f6h5 h4f5 e7f8 d3d1 h5f6 d1f3 c7c3 a1b1 c3d3 f5h6 g8h7 e3g4 d3f3 g2f3 f6d5 e4d5 g7h6 g4f6 h7g6');
INSERT INTO evaluations VALUES(39199,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39199,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39199,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39199,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(39199,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(39199,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(39199,36,'cp -4','g8f6 g1f3 d7d6 d2d4 e5d4 f3d4 e8g8 e1g1 c8g4 f2f3 b8c6 c1e3 g4e6 c4b5 c6d4 e3d4 c5d4 d1d4 c7c6 b5d3 d6d5 e4d5 f6d5 c3d5 e6d5 f1e1 g7g6 h2h4 d5e6 d4d8 f8d8 g1f2 d8d4 g2g4 a8d8 e1e5 e6c4 e5e7 c4d3 c2d3 c6c5');
INSERT INTO evaluations VALUES(20939,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20939,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(20939,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(20939,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(20939,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(20939,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(20939,36,'cp 28','c8d7 g1f3 e7e6 e1g1 c6e5 b5e2 e5f3 e2f3 g8f6 c2c4 f8d6 b1d2 e8g8 b2b3 c7c5 d4c5 d6e5 a1b1 d8c7 h2h3 c7c5 c4d5 e6d5 c1b2 c5e7 b2e5 e7e5 f3e2 a8c8 d2f3 e5c3 f3d4 h7h6 b1c1 c3a5 c1c8 f8c8 d1a1 f6e4');
INSERT INTO evaluations VALUES(27281,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(27281,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(27281,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(27281,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(27281,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(27281,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(27281,36,'cp 32','g1f3 e7e6 e1g1 c6e5 b5e2 e5f3 e2f3 f8d6 b1d2 g8f6 c2c4 e8g8 b2b3 c7c5 d4c5 d6e5 a1b1 d8c7 h2h3 c7c5 c4d5 e6d5 c1b2 e5b2 b1b2 a8c8 d1a1 d7b5 f1d1 f8d8 b2b1 c5c3 d2f1 c3a1 b1a1 c8c2 f1g3 g7g6 a2a4 b5a6 a1c1 c2c1 d1c1');
INSERT INTO evaluations VALUES(27282,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(27282,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(27282,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(27282,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(27282,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(27282,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(27282,36,'cp 26','e7e6 e1g1 c6e5 f3e5 d7b5 c2c4 b5c4 e5c4 d5c4 d1a4 c7c6 a4c4 g8f6 b1c3 f8e7 f1d1 e8g8 e3e4 a7a5 c1f4 a5a4 a1c1 d8d7 c4e2 b7b5 d4d5 c6d5 e4d5 f6d5 c3d5 e6d5 c1c7 d7f5 f4g3 e7f6 e2b5 f5e4 d1d5 a4a3 b2b4');
INSERT INTO evaluations VALUES(27172,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(27172,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(27172,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(27172,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(27172,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(27172,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(27172,36,'cp 26','b5e2 g8f6 b2b3 g7g5 c2c4 g5g4 f3d2 h7h5 c1b2 h5h4 a2a3 f8h6 b1c3 d8e7 d1c2 e8c8 b3b4 c8b8 e1c1 d7c8 c1b1 a7a6 b4b5 a6b5 c4b5 c6a5 d2b3 a5b3 c2b3 b7b6 a3a4 c8b7 e2d3 h4h3 g2h3 g4h3');
INSERT INTO evaluations VALUES(19714,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19714,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19714,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(19714,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(19714,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19714,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19714,36,'cp 10','d8a5 b1c3 c8g4 h2h3 g4f3 d1f3 e7e6 e1g1 g8f6 a2a3 a8c8 f3g3 a5d8 c3e2 d8c7 g3c7 c8c7 c1f4 c7c8 b5d3 f8e7 f1b1 e8g8 b2b3 f8d8 g1f1 f6e4 f2f3 e4d6 c2c3 g7g5 f4d2 h7h5 g2g3 f7f5 b1e1 g8f7 a3a4 e7f6 b3b4');
INSERT INTO evaluations VALUES(1410,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(1410,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(1410,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(1410,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(1410,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(1410,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(1410,36,'cp 23','b1c3 c8g4 c1d2 a8c8 a2a4 a5d8 h2h3 g4f3 d1f3 e7e6 c3e2 a7a6 b5d3 c6b4 e1g1 g8f6 f1c1 f8d6 b2b3 e8g8 c2c4 d5c4 b3c4 b4d3 f3d3 e6e5 d4d5 e5e4 d3e3 d6c5 e3g3 f6e8 d2g5 d8d7 g3b3 e8d6 g5e3 h7h5 e2g3 c5e3 b3e3 d6c4 e3d4 c4d6 g3e4 c8c1 a1c1 d6e4');
INSERT INTO evaluations VALUES(1409,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(1409,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(1409,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(1409,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(1409,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(1409,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(1409,36,'cp 10','c8g4 c1d2 a8c8 a2a4 a5d8 h2h3 g4f3 d1f3 e7e6 c3e2 a7a6 b5d3 c6b4 e1g1 g8f6 c2c3 b4d3 f3d3 f8e7 e2c1 f6e4 d2f4 e8g8 d3e2 a6a5 c1d3 e7d6 f2f3 e4f6 f4d6 d8d6 e2e3 h7h6 f1e1 b7b6 e3e5 d6d8 e1e2 f6e8 g2g4 c8c4 g1h1 e8d6 h1g2');
INSERT INTO evaluations VALUES(42235,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42235,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42235,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42235,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42235,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42235,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(42235,36,'cp -3','f1d3 e8g8 e1g1 c8g4 h2h3 g4h5 c1g5 c7c6 g2g4 h5g6 f3e5 b8d7 e5d7 d8d7 g5f6 g7f6 d1f3 b4c3 b2c3 g8g7 a2a4 f8e8 a4a5 b7b5 a5b6 a7b6 a1a8 e8a8 f1e1 b6b5 g1g2 a8e8 e1e3 e8e3 f3e3 h7h6 e3f3 g6d3 f3d3');
INSERT INTO evaluations VALUES(42242,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(42242,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(42242,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(42242,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(42242,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(42242,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(42242,36,'cp 0','e8g8 e1g1 c8g4 c1g5 c7c6 h2h3 g4h5 g2g4 h5g6 f3e5 b8d7 e5d7 d8d7 g5f6 g7f6 d1f3 b4c3 b2c3 g8g7 a2a4 h7h6 a4a5 a8e8 f3f4 g6d3 c2d3 e8e2 a1e1 e2c2 e1c1 c2a2 c1a1');
INSERT INTO evaluations VALUES(37113,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(37113,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37113,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37113,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37113,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37113,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37113,36,'cp 0','e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 c1g5 c7c6 f3e5 b8d7 e5d7 d8d7 g5f6 g7f6 d1f3 b4c3 b2c3 g8g7 a2a4 f8e8 g1h2 h7h6 a4a5 b7b5 a5b6 a7b6 a1a8 e8a8 d3g6 f7g6 f1e1 a8e8 e1e8 d7e8 f3g3 e8d7 g3b8 b6b5 h2g2 h6h5 g4h5 g6h5 b8g3 g7h6 g3e3 h6g6 e3g3');
INSERT INTO evaluations VALUES(37112,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37112,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37112,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37112,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37112,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37112,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37112,36,'cp 0','c8g4 c1g5 c7c6 h2h3 g4h5 g2g4 h5g6 f3e5 b8d7 e5d7 d8d7 g5f6 g7f6 d1f3 b4c3 b2c3 g8g7 a2a4 f8e8 g1h2 e8e6 f3f4 g6d3 c2d3 h7h6 h3h4 a7a5 h2g2 b7b5 f1b1 b5a4 a1a4 d7e7 b1b6 e6e1 a4a2 e7e6 f4f5 e6f5 g4f5 a5a4 g2f3');
INSERT INTO evaluations VALUES(43252,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43252,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43252,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43252,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(43252,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43252,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(43252,36,'cp 11','d7d5 c2c4 c7c6 c4d5 c6d5 b1c3 e8g8 e1g1 f6e4 f3e5 c8f5 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 f1c1 d8d7 h2h4 h7h6 e2e3 f7f6 c3c4 g6g5 c4d5 e6d5 h4g5 h6g5 g3g4 g5f4 g4f5 f4e3 f2e3 g7h6 a4c6 d7c6 c1c6 h6e3 g1h1 e3d4 g2d5 g8h8');
INSERT INTO evaluations VALUES(41508,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(41508,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(41508,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41508,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41508,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(41508,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(41508,36,'cp 12','c2c4 c7c6 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 e1g1 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 f1c1 c3c4 c1c4 d5c4 a1c1 g7d4 c1c4 e6e5 e2e3 f5e6 f4h6 e6c4 h6f8 g8f8 a4c4 d4b6 c4c3 d8e7 g2d5 f8g7 g1g2 h7h6 c3e1 h6h5 d5b3 e7c7');
INSERT INTO evaluations VALUES(41454,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(41454,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(41454,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41454,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(41454,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(41454,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(41454,36,'cp 16','c7c6 c4d5 c6d5 f3e5 e8g8 b1c3 c8f5 c1f4 b8c6 e5c6 b7c6 e1g1 f6h5 f4e3 h5f6 a1c1 f6g4 e3f4 e7e5 d4e5 g4e5 c3a4 f8e8 b2b3 d8e7 d1d2 a8d8 a4c5 h7h5 h2h3 g7f6 f1d1 g6g5 f4e3 f5g6 c5a4 d8c8 e3d4 h5h4');
INSERT INTO evaluations VALUES(40280,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40280,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40280,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(40280,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40280,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40280,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(40280,36,'cp 11','c4d5 c6d5 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 a8c8 e1g1 e8g8 f1c1 d8e7 h2h4 f6h5 e5g7 h5g7 c1c2 g7f5 a1c1 c8c6 e2e3 f5d6 e3e4 d6e4 c3e4 d5e4 g2e4 c6c2 c1c2 f8c8 e4b7 c8c2 d2c2 e7b4 b7c8 b4d4 c8d7 d4d7');
INSERT INTO evaluations VALUES(22777,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(22777,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(22777,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(22777,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(22777,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(22777,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(22777,36,'cp -63','g8f6 g1f3 f8c5 b1c3 e8g8 f1e2 f8e8 e1g1 a7a5 f1e1 d7d6 d2d3 h7h6 d1d2 c8d7 a2a3 c5b6 c3d1 d6d5 e4d5 f6d5 d1e3 d5f4 e3c4 d8f6 e2f1 f4h3 g2h3 f6f3 c4b6 c7b6 f1g2 f3h5 e1e3 h5g5 g1h1 a8d8 a1d1 d7e6');
INSERT INTO evaluations VALUES(16501,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(16501,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(16501,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(16501,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(16501,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(16501,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(16501,36,'cp -59','b1c3 f8c5 g1f3 e8g8 f1d3 a7a6 c3a4 c5d4 b2d4 e5d4 e1g1 d7d5 e4e5 f6e4 f1e1 f8e8 c2c4 d4c3 d2c3 c8g4 d3e4 d5e4 d1d8 a8d8 f3d4 c6d4 c3d4 d8d4 a1c1 g4d7 a4c5 d7c6 c5a6 e8a8 a6c7 a8a2 c7b5 d4d2 b5c3 a2c2 c1c2 d2c2 c3e2 g7g6 h2h3');
INSERT INTO evaluations VALUES(16504,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(16504,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16504,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(16504,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(16504,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(16504,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(16504,36,'cp -54','f8c5 g1f3 e8g8 f1e2 a7a5 e1g1 f8e8 f1e1 d7d6 d2d3 h7h6 d1d2 c8d7 a2a3 c5b6 c3d1 d6d5 e4d5 f6d5 e2f1 d8f6 d1e3 d5f4 e3c4 f4h3 g2h3 f6f3 c4b6 c7b6 f1g2 f3h5 d2e2 h5e2 e1e2 a8c8 a1e1 f7f6 h3h4 d7g4 e2e4 g4f5 e4e2 b6b5 f2f4 f5g4 e2d2 e5f4');
INSERT INTO evaluations VALUES(10205,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(10205,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(10205,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(10205,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(10205,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(10205,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(10205,36,'cp -41','g1f3 d7d6 f1e2 a7a6 d2d3 h7h6 e1g1 c5a7 a2a3 e8g8 d1d2 b7b5 b3b4 c8e6 a3a4 c6d4 f3d4 a7d4 c3d1 d4b2 d1b2 c7c5 f1b1 c5b4 d2b4 d8c7 a4b5 a6b5 c2c3 a8a1 b1a1 d6d5 b2d1 f8c8 b4b5 d5e4 d3e4');
INSERT INTO evaluations VALUES(10204,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(10204,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(10204,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(10204,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(10204,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(10204,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(10204,36,'cp -63','d7d6 f1e2 a7a6 d2d3 c8e6 e1g1 h7h6 a2a3 c5a7 b3b4 e8g8 d1d2 f8e8 c3d1 d6d5 f3e5 c6d4 d1e3 d5e4 a1d1 c7c6 d3e4 f6e4 d2d3 d4e2 d3e2 d8h4 e2f3 e4g5 f3g3 h4g3 h2g3 a8d8 d1d8 e8d8 f1d1 d8d1 e3d1');
INSERT INTO evaluations VALUES(9391,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(9391,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(9391,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(9391,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(9391,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(9391,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(9391,36,'cp -56','f1b5 e8g8 b5c6 b7c6 e1g1 c8g4 h2h3 g4h5 d1e2 d6d5 e4d5 h5f3 e2f3 c6d5 c3e2 f8e8 e2g3 c5d6 c2c4 c7c6 f1c1 e8e6 d2d3 d8d7 c4d5 c6d5 c1c2 a8c8 a1c1 c8c2 c1c2 d6f8 g3f1 d5d4 f1d2 f6d5 d2e4 f8b4 b2c1 d5c3 e4c3');
INSERT INTO evaluations VALUES(23073,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(23073,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(23073,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(23073,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(23073,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(23073,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(23073,36,'cp 195','b1c3 f8b4 f2f4 e5c6 d4c6 d7c6 d1d8 e8d8 c1d2 h7h5 e1c1 c8g4 d1e1 g4d7 h2h3 g8h6 f1c4 h5h4 h1f1 b4e7 a2a4 a7a6 a4a5 b7b5 c4d3 d8c8 f4f5 f7f6 e4e5 f6e5 e1e5 e7d6 e5e1 c8b7 c3e4 h6f5 e4d6 f5d6');
INSERT INTO evaluations VALUES(23072,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(23072,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(23072,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(23072,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(23072,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(23072,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(23072,36,'cp 154','f8b4 f2f4 e5c6 d4c6 b7c6 d1d4 d8h4 g2g3 h4e7 c1d2 g8f6 e1c1 b4c5 d4c4 e8g8 e4e5 f6d5 c3a4 c5e3 f1d3 e3d2 d1d2 d5b6 c4e4 g7g6 h1e1 f8d8 d2d1 e7f8 a4c3 d7d5 e5d6 d8d6 e4e2 d6e6 e2f2 e6e1 d1e1 c8e6 a2a3 b6d7 f4f5 e6d5 c1b1 d7c5');
INSERT INTO evaluations VALUES(6885,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(6885,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(6885,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(6885,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(6885,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(6885,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(6885,36,'cp 150','f2f4 e5c6 d4c6 b7c6 d1d4 d8h4 g2g3 h4e7 c1d2 g8f6 e1c1 b4c5 d4c4 e8g8 e4e5 f6d5 c3a4 c5e3 f1d3 e3d2 d1d2 d5b6 c4e4 g7g6 h1e1 f8d8 e1e2 b6a4 e4a4 a7a5 a2a3 c8a6 e2e3 e7e6 b2b3 e6e7 d3a6 a8a6 c1b2 a6a8 a4d4 e7e6 d2d1 d7d5 e5d6 e6d6 d4d6 d8d6');
INSERT INTO evaluations VALUES(6210,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(6210,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(6210,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(6210,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6210,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6210,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(6210,36,'cp 0','g8f6 d2d3 e8g8 h2h3 a7a6 c1e3 d7d6 e3c5 d6c5 a2a4 d8d6 f3d2 b7b6 c3d5 a8b8 d5e3 f8d8 d2f3 b6b5 c4a2 d6e7 f3d2 e7d6');
INSERT INTO evaluations VALUES(43126,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43126,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43126,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43126,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43126,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43126,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(43126,36,'cp 21','h7h6 g5f4 d7d5 e2e3 c7c5 d4c5 d8a5 a1c1 d5c4 f1c4 e8g8 g1f3 a5c5 c4b3 b8c6 e1g1 c5a5 h2h3 f8d8 d1e2 g6g5 f4h2 g5g4 h3g4 c8g4 f1d1 a7a6 h2g3 e7e6 b3c2 f6d5 c3d5 d8d5 e2c4 a5b4 c4b4 c6b4 d1d5 b4d5');
INSERT INTO evaluations VALUES(43110,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(43110,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(43110,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(43110,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(43110,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(43110,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(43110,36,'cp 17','g5d2 c7c5 d4d5 e7e6 g1f3 d7d6 e2e4 e8g8 d5e6 c8e6 f1d3 b8c6 e1g1 f6d7 h2h3 d7e5 f3e5 d6e5 c3d5 d8h4 d2c3 a8d8 d1e2 h4g5 e2e3 g5e3 f2e3 b7b6 d3e2 d8d7 a2a4 c6a5 c3a5 b6a5 b2b3');
INSERT INTO evaluations VALUES(56288,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(56288,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(56288,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(56288,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(56288,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(56288,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(56288,36,'cp -35','f8e7 d2d4 g8f6 e2d1 c7c5 d4c5 e7c5 f1d3 d8e7 d1e2 e7e2 d3e2 e8g8 e1g1 b8c6 b1d2 f8e8 e2d3 c5b6 h2h3 c6b4 d2b3 b4d3 c2d3 c8f5 f1d1 f5g6 a2a4 a7a5 c1d2 f6d7 a1c1 g6d3 d2a5 d3e4 a5b6 d7b6 a4a5');
INSERT INTO evaluations VALUES(38323,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(38323,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(38323,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(38323,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(38323,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(38323,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(38323,36,'cp -38','d2d4 g8f6 e2d1 c7c5 d4c5 e7c5 f1d3 d8e7 d1e2 e7e2 d3e2 e8g8 e1g1 b8c6 b1d2 c8g4 c2c3 f8e8 e2d3 c5b6 h2h3 g4h5 f1e1 h7h6 g1f1 g7g5 e1e8 a8e8 g2g4 h5g6 d3g6 f7g6 d2b3 f6e4 b3d4 c6d4 f3d4 b6d4 c3d4 e8c8 c1e3 c8c2 b2b3 e4d2 f1g2');
INSERT INTO evaluations VALUES(38314,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(38314,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(38314,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(38314,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(38314,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(38314,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(38314,36,'cp -35','g8f6 e2d1 c7c5 d4c5 e7c5 f1d3 d8e7 d1e2 e7e2 d3e2 e8g8 e1g1 b8c6 b1d2 c5b6 e2d3 c8g4 h2h3 g4h5 c2c3 f8e8 f1e1 e8e1 f3e1 a8e8 g1f1 h7h6 d2f3 f6e4 c1e3 b6e3 f2e3 g7g5 a1d1 g8g7 d3c2 e4f6 f1e2 h5g6 c2g6 g7g6 e1c2');
INSERT INTO evaluations VALUES(41104,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(41104,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(41104,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(41104,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(41104,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(41104,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(41104,36,'cp -33','e2d1 c8g4 f1d3 e8g8 e1g1 f8e8 h2h3 g4h5 f1e1 b8d7 a2a4 d7f8 c2c3 c7c6 a4a5 d8c7 c1e3 h5g6 a5a6 b7b5 f3e5 e7d6 e5g6 h7g6 b1d2 e8e7 d2f3 a8e8 e3g5 e7e1 f3e1 f8e6 g5e3 d6h2 g1f1 h2f4 e3f4 e6f4 e1f3 f6e4 f1g1 c7b6 d1c2 c6c5 a1e1 f4d3 c2d3 c5c4 d3c2');
INSERT INTO evaluations VALUES(11513,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11513,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(11513,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(11513,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(11513,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(11513,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(11513,36,'cp 76','d2d4 e5d4 f3d4 f8b4 d4c6 d7c6 d1d8 e8d8 c1d2 h8e8 e1c1 c8e6 a2a3 e6c4 a3b4 a6a5 b4a5 a8a5 b2b3 c4e6 c1b2 d8c8 f2f3 f6d7 c3e2 a5a6 d2f4 f7f6 f4g3 c6c5 e2c3 d7e5 d1d2 e8d8 d2d8 c8d8 h1e1 e6f7 c3b5 d8d7 f3f4 d7c6 b5c3 e5d7 e1d1 b7b6 f4f5 d7e5');
INSERT INTO evaluations VALUES(11530,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(11530,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(11530,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(11530,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(11530,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(11530,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(11530,36,'cp 63','e5d4 f3d4 f8b4 d4c6 d7c6 d1d8 e8d8 f2f3 c8e6 c4e6 b4c3 b2c3 f7e6 c3c4 a6a5 a2a4 e6e5 e1e2 c6c5 c1b2 h8e8 h1d1 d8c8 d1d3 b7b6 g2g4 e8e7 a1d1 c8b7 d3d8 a8d8 d1d8 g7g6 h2h4 b7c6 b2c3 f6d7 d8g8 d7f6 g8h8 c6b7 g4g5 f6h5');
INSERT INTO evaluations VALUES(11546,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(11546,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(11546,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(11546,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(11546,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(11546,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(11546,36,'cp 75','f3d4 f8b4 d4c6 d7c6 d1d8 e8d8 c1g5 d8e8 g5d2 h7h5 c4d3 h5h4 h2h3 c8e6 f2f4 f6h5 e1c1 b4d6 e4e5 d6b4 f4f5 e6d5 c3d5 b4d2 d1d2 c6d5 c2c4 h5g3 h1e1 d5d4 d3e4 g3e4 e1e4 h8h5 d2f2 c7c5 c1d2 e8d7 d2d3 a8e8 f2f4');
INSERT INTO evaluations VALUES(11545,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(11545,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(11545,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(11545,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(11545,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(11545,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(11545,36,'cp 60','f8b4 d4c6 d7c6 d1d8 e8d8 c1g5 d8e8 g5d2 h7h5 f2f3 h5h4 e1c1 c8e6 c4e6 f7e6 c3a4 b4d6 b2b3 b7b5 a4b2 f6d7 h2h3 e6e5 d2e3 d7c5 b2d3 c5d3 d1d3 a6a5 a2a4 a8b8 c1d2 b5a4 b3a4 b8b4 h1a1 e8d7 d3c3 h8g8 d2e2 g7g5 e3d2 b4b2');
INSERT INTO evaluations VALUES(10129,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(10129,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10129,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(10129,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10129,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10129,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(10129,36,'cp 0','e8g8 d2d3 d7d6 c2c3 a7a6 h2h3 h7h6 b2b4 c5a7 a2a4 c8e6 b1d2 f6h5 c4a2 d8f6 d2f1 f6g6 f3h4 g6f6');
INSERT INTO evaluations VALUES(4347,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4347,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4347,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4347,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(4347,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(4347,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(4347,36,'cp 18','d2d3 d7d6 c2c3 a7a6 b2b4 c5a7 a2a4 h7h6 b1d2 c8e6 c4b3 e6b3 d1b3 d8e8 d2c4 e8e6 b3c2 f6g4 c1e3 b7b5 c4d2 g4e3 f2e3 c6e7 d3d4 c7c6 d2b3 e7g6 a4b5 a6b5 h2h3 f8c8 g1h2 c8f8 a1a6');
INSERT INTO evaluations VALUES(4343,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(4343,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4343,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4343,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(4343,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(4343,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(4343,36,'cp 6','d7d6 c2c3 a7a6 b1d2 c8e6 b2b4 c5a7 c4b3 h7h6 a2a4 e6b3 d2b3 f8e8 d1c2 d8d7 h2h3 f6h5 f3h4 a8d8 c1e3 a7e3 e1e3 h5f4 a4a5 d7e6 d3d4 e5d4 b3d4 c6d4 c3d4 e6f6 h4f5 g7g6 f5h6 g8g7 h6g4 f6d4');
INSERT INTO evaluations VALUES(3835,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(3835,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(3835,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(3835,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(3835,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(3835,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(3835,36,'cp 15','c2c3 a7a6 c4b3 h7h6 b1d2 c8e6 d2f1 e6b3 d1b3 f8e8 c1e3 c5e3 f1e3 d8d7 a2a4 c6e7 a4a5 e7g6 a1d1 g6f4 b3c2 d6d5 g2g3 d7h3 f3h4 a8d8 f2f3 g7g6 c2f2 f4e6 h4g2 d5e4 d3e4 d8d1 e1d1 e6c5');
INSERT INTO evaluations VALUES(20262,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(20262,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(20262,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(20262,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(20262,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(20262,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(20262,36,'cp 10','g8f6 b1d2 f8d6 f4g3 e8g8 f1d3 g7g6 e1g1 b7b6 g3h4 c8b7 f1e1 d6e7 d4c5 b6c5 e3e4 a8c8 e4d5 e6d5 d3f1 d8c7 d1c2 c8e8 b2b4 c5b4 c3b4 f6e4 h4e7 c7e7 b4b5 c6e5 f3e5 e7e5 d2e4 d5e4');
INSERT INTO evaluations VALUES(57061,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(57061,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(57061,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(57061,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57061,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(57061,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(57061,36,'cp 5','g1f3 g8f6 e2e3 d7d5 c1b2 f8d6 c2c4 e8g8 b1c3 c7c6 d2d4 b8d7 a1c1 d8e7 f1e2 b7b6 e1g1 c8b7 e2d3 a8c8 f1e1 c6c5 c4d5 e6d5 d3f1 f8d8 g2g3 h7h6 f1g2 e7e6 d1e2 d8e8 e1d1 c5c4 c3b5');
INSERT INTO evaluations VALUES(57059,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(57059,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(57059,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(57059,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57059,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(57059,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(57059,36,'cp 2','g8f6 c1b2 f8e7 e2e3 c7c5 c2c4 e8g8 f1e2 d7d5 e1g1 b8c6 c4d5 d8d5 d2d4 f8d8 d4c5 d5c5 b1d2 c5f5 d1b1 f5b1 f1b1 c8d7 a2a3 a7a5 b1d1 d7e8 a1c1 a8c8 h2h3 f6d5 e2b5 d5c7 b5c6 e8c6 d2c4 a5a4 d1d8 c8d8 b3a4 c6a4');
INSERT INTO evaluations VALUES(49462,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(49462,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(49462,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(49462,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(49462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49462,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49462,36,'cp 0','e2e3 c7c5 d2d4 c5d4 e3d4 b7b6 c1b2 f8e7 f1d3 e8g8 e1g1 c8b7 b1d2 d7d6 f1e1 b8d7 d2f1 a8c8 f1g3 f6d5 d1d2 d5f6');
INSERT INTO evaluations VALUES(13366,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(13366,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13366,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13366,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13366,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13366,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(13366,36,'cp 2','f6g4 d4c6 b7c6 e3c1 d8b6 d1d2 g7g6 c3a4 f8h6 a4b6 h6d2 c1d2 a7b6 f2f3 g4e5 a2a4 c8a6 d2c3 f7f6 c3e5 f6e5 e1d2 a6f1 h1f1 e8d7 b2b4 a8a6 a1a3 h8a8 f1a1 g6g5 h2h3 h7h5 c2c4 e7e6 d2d3 c6c5 b4b5 a6a5 d3c2 d7c8 c2b3 c8d7 b3c3');
INSERT INTO evaluations VALUES(13413,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(13413,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13413,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(13413,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(13413,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(13413,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(13413,36,'cp 0','d4c6 b7c6 e3c1 g7g6 h2h3 g4e5 b2b3 f8g7 c1b2 e5d7 d1d2 d7c5 f2f3 e8g8 h3h4 a7a5 h4h5 a5a4 c3a4 c5a4 b2g7 g8g7 h5g6 h7g6 d2h6 g7f6 b3a4 d8a5 e1d1 a5a4 f1d3 a4d4 d1e2 c8e6 h6h4 f6g7 h4h7 g7f6');
INSERT INTO evaluations VALUES(44178,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(44178,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44178,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44178,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44178,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44178,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44178,36,'cp 0','g1f3 b8c6 d2d4 f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 d2b3 e6e5 c1e3 f8e7 e1c1 c8d7 b3a5 a8c8 g2g3 e7c5 d1d5 c5e3 f2e3 e8e7 h1d1 c8c7 d5e5 e7f6 e5d5 f6e7');
INSERT INTO evaluations VALUES(1584,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(1584,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(1584,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(1584,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(1584,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(1584,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(1584,36,'cp 167','e6d7 d2d4 e7e6 e1g1 g8f6 f1e1 a7a6 e2d3 f8e7 c3e4 b7b6 c2c3 c8b7 d1e2 f6e4 d3e4 e8g8 f3e5 c6e5 e4b7 a8a7 b7e4 e5g6 c1e3 e7d6 g2g3 g6e7 e4c2 e7d5 e3g5 d6e7 g5d2 c7c5 d4c5 e7c5 a1d1 d7c6');
INSERT INTO evaluations VALUES(47141,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(47141,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(47141,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(47141,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(47141,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(47141,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(47141,36,'cp -33','d5c4 b1c3 c7c6 a2a4 f6d5 f4g3 b7b5 e2e4 d5c3 b2c3 a7a6 g1f3 e7e6 f1e2 f8d6 g3d6 d8d6 e4e5 d6c7 f3g5 b8d7 e1g1 h7h6 g5e4 e8g8 f2f4 d7b6 a4a5 b6d5 d1c1 c7e7 h2h3 a8b8 f1f3 f7f5 e4c5 e7c7 f3g3 g8h8 c1d2 c7e7 a1d1 b5b4 e2c4 b4c3');
INSERT INTO evaluations VALUES(47348,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(47348,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(47348,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(47348,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(47348,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(47348,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(47348,36,'cp -24','b1c3 c7c6 a2a4 f6d5 f4g3 d5c3 b2c3 d8a5 d1c2 c6c5 e2e3 c5d4 e3d4 c8e6 f1e2 b8d7 e2f3 e6f5 c2b2 e7e5 g1e2 f8e7 d4e5 e8g8 b2b5 a5b5 a4b5 d7c5 e1g1 f8d8 f1e1 c5b3 a1d1 d8d1 e1d1 b3c5 h2h3 f5d3 e2f4 a7a5 b5a6 a8a6');
INSERT INTO evaluations VALUES(47345,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(47345,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(47345,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(47345,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(47345,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(47345,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(47345,36,'cp -27','f6d5 e2e3 d5f4 e3f4 g7g6 f1c4 f8g7 g1f3 b8c6 d4d5 c6a5 d1a4 c7c6 d5c6 e8g8 e1g1 a5c6 f1d1 d8a5 a4a5 c6a5 c4d5 e7e6 d5e4 a5c4 a1b1 c4b2 b1b2 g7c3 b2b3 c3a5 e4b7 a8b8 b7c8 b8b3 a2b3 f8c8 g2g3 a5b6 g1g2 c8c2 d1d2 c2d2 f3d2 g8f8 d2f3 f7f6 f3d2 f8e7 b3b4 a7a6');
INSERT INTO evaluations VALUES(55953,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(55953,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(55953,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55953,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55953,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(55953,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55953,36,'cp 24','g1f3 c7c5 c2c4 d5d4 e2e3 b8c6 d2d3 g8f6 e1g1 h7h6 e3d4 c5d4 c1f4 f8d6 f4d6 d8d6 b1d2 e8g8 a1c1 d6b4 d1c2 a7a5 f1e1 a5a4 f3e5 c8d7 d2f3 f8d8 c2e2 d7e8 h2h4 b4d6 e5c6 e8c6 f3e5 f6d7 e5c6 b7c6 c4c5 d7c5 e2c2 c5d7 c2c6 d6c6 g2c6');
INSERT INTO evaluations VALUES(22102,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(22102,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(22102,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(22102,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(22102,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(22102,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(22102,36,'cp 30','d6e5 d1d8 e8d8 f1c4 d8e8 b1c3 f8b4 c1d2 g8f6 f2f3 e8e7 a2a3 b4c5 c3a4 c5d6 d2e3 h7h5 e1e2 f6d7 c4a2 d7f8 h1d1 f8e6 h2h4 e6f4 e2f2 c8e6 a2e6 f4e6 a4c3 c7c6 a3a4 f7f6 a4a5 h8d8 c3e2 d6c7 d1d8 c7d8');
INSERT INTO evaluations VALUES(22194,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(22194,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(22194,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(22194,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(22194,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(22194,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(22194,36,'cp 31','d1d8 e8d8 a2a4 g8f6 f2f3 a7a5 f1c4 d8e8 b1a3 f8b4 e1e2 e8e7 c1e3 f6e8 a3b5 c7c6 b5c3 f7f6 c3d1 h7h5 h2h4 b4d6 b2b3 g7g6 d1b2 e8c7 e3b6 c7e8 h1d1 d6c7 b6c5 e8d6 c5a3 c8e6 c4e6 e7e6 c2c3 b7b6');
INSERT INTO evaluations VALUES(2266,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(2266,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(2266,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(2266,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(2266,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(2266,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(2266,36,'cp 24','e8d8 f1c4 d8e8 a2a4 g8f6 f2f3 a7a5 e1e2 f8c5 b1c3 e8e7 g2g4 c7c6 c3d1 f6e8 d1f2 e8d6 f2d3 c5b6 c4b3 f7f6 h2h4 h7h5 g4g5 c8e6 b3e6 e7e6 c1e3 b6e3 e2e3 b7b6 b2b3 c6c5 d3b2 f6g5 h4g5 h5h4 a1d1 d6f7 b2c4 f7g5 c4b6 a8b8');
INSERT INTO evaluations VALUES(2325,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(2325,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(2325,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2325,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(2325,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(2325,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(2325,36,'cp 40','f1c4 d8e8 b1c3 f8b4 f2f3 e8e7 c1d2 g8f6 a2a3 b4c5 c3a4 c5d6 d2e3 h7h5 e1c1 a7a6 a4c5 b7b6 c5d3 a6a5 h2h4 c8e6 c4e6 e7e6 a3a4 f6d7 c2c3 h8d8 c1c2 d6c5 e3d2 c5d6 d2g5 f7f6 g5e3 d7c5 b2b3 c5b7 d3b2');
INSERT INTO evaluations VALUES(2323,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(2323,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(2323,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(2323,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(2323,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(2323,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(2323,36,'cp 30','d8e8 b1c3 f8b4 c1d2 g8f6 a2a3 b4c5 f2f3 e8e7 c3a4 c5d6 b2b3 f6d7 h2h4 d7b6 a4b2 h7h5 d2e3 b6c4 b2c4 c8e6 c4d6 c7d6 a3a4 h8c8 c2c4 a7a5 e3d2 f7f5 e4f5 e6f5 d2c3 c8c5 e1c1 b7b6 c1b2');
INSERT INTO evaluations VALUES(14809,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(14809,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14809,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(14809,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(14809,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(14809,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14809,36,'cp 22','f8d6 f4g3 e8g8 b1d2 d8e7 c2c4 e6e5 c4d5 c6b4 d3b1 e5d4 g3d6 e7d6 e3e4 d6a6 d1b3 c7c5 b3c4 f8e8 c4a6 b4a6 e1g1 c8g4 f1c1 a8c8 a2a3 g4f3 g2f3 f6h5 b1d3 h5f4 d3f1 h7h6 c1e1 c8d8 f1a6 b7a6 d2c4 f4g6 a1d1 g6f4');
INSERT INTO evaluations VALUES(9241,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(9241,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(9241,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(9241,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(9241,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(9241,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(9241,36,'cp 36','f4g3 d6g3 h2g3 d8e7 c2c4 d5c4 d3c4 e8g8 b1c3 e6e5 d4d5 c6a5 c4e2 b7b6 e3e4 f8d8 d1b1 a7a6 e1g1 a5b7 f1e1 b7d6 b1c1 c8d7 e2d3 c7c5 d5c6 d7c6 c1e3 f6d7 e1d1 b6b5 d3e2 d7f6 e3c5 c6e4 c5e5 e7e5 f3e5 e4b7 a1c1 a8c8');
INSERT INTO evaluations VALUES(9258,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9258,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9258,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(9258,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(9258,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(9258,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(9258,36,'cp 25','d8e7 c2c4 d6g3 h2g3 d5c4 d3c4 e8g8 b1c3 e6e5 d1b1 f8d8 d4d5 c6a5 c4e2 b7b6 e3e4 a7a6 e1g1 a5b7 f1d1 c8d7 a2a4 a6a5 e2b5 b7c5 f3d2 f6e8 b5d7 d8d7 b2b4 a5b4 b1b4 c5b7 b4e7 d7e7 d1c1 b7c5 c3b5');
INSERT INTO evaluations VALUES(32764,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(32764,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(32764,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(32764,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(32764,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(32764,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(32764,36,'cp 25','c2c4 b8c6 b1c3 g8f6 d2d4 c5d4 f3d4 g7g6 f2f3 f8g7 c1e3 e8g8 e1g1 a7a6 c3a4 a8b8 d4c6 d7c6 a1c1 b7b5 c4b5 c6b5 b2b3 f8c8 d1d2 f6d7 f1d1 b5b7 h2h3 g7f6 c1c8 b8c8 d1c1 c8c1 d2c1 e7e6 a4c3 b7b4 c3e2 b4a5 c1d2 a5b5 g1f2 f6e7 e2c3 b5b7 c3a4 b7c6');
INSERT INTO evaluations VALUES(32759,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(32759,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(32759,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(32759,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(32759,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(32759,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(32759,36,'cp 26','g8f6 b1c3 b8c6 d2d4 c5d4 f3d4 g7g6 f2f3 f8g7 c1e3 e8g8 e1g1 a7a6 d4b3 d7c7 a1c1 c6e5 c3d5 f6d5 c4d5 c7d7 b3d4 a8c8 b2b3 c8c5 h2h3 f8c8 d1d2 c5c1 f1c1 c8c1 d2c1 d7d8 c1d2 d8c8 d4e2 g8f8 g1f1 c8d8 a2a4 e5d7 a4a5 d8c7');
INSERT INTO evaluations VALUES(36469,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36469,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(36469,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(36469,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(36469,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(36469,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(36469,36,'cp 41','e2e3 h7h6 g5h4 g6g5 h4g3 f6h5 d1c2 e7e6 f1e2 f7f5 e1g1 d8e7 b2b4 c8d7 a1c1 d7e8 c4c5 a7a6 c5d6 c7d6 b4b5 a6b5 e2b5 b8c6 a2a4 a8c8 c2d2 c6a5 b5e8 f8e8 c3b5 a5c4 d2e2 h5f6');
INSERT INTO evaluations VALUES(36466,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(36466,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(36466,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(36466,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36466,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(36466,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36466,36,'cp 28','h7h6 g5h4 g6g5 h4g3 f6h5 d1c2 e7e6 f1e2 d8e7 f3d2 h5g3 h2g3 c7c5 d4d5 b8a6 e1g1 a6c7 d5e6 c8e6 e2f3 a8b8 f1d1 f8d8 a2a4 c7a6 c3d5 e6d5 c4d5 a6b4 c2b3 d8e8 f3e2 e7f6 d2c4 b7b6 a1a3 e8e7 g3g4 b8d8');
INSERT INTO evaluations VALUES(36442,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(36442,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(36442,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(36442,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(36442,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(36442,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36442,36,'cp 31','g5h4 g6g5 h4g3 f6h5 d1c2 e7e6 f1e2 d8e7 e1g1 b8d7 f3d2 h5g3 h2g3 f7f5 b2b4 c7c6 a1d1 g8h8 f1e1 a7a6 e2d3 e7f7 g3g4 d7f6 g4f5 e6f5 d2f1 f5f4 c3e4 f6e4 d3e4 c8g4 f2f3 g4e6 d4d5 e6d7 e3f4 f7f4 f1e3');
INSERT INTO evaluations VALUES(36445,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(36445,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(36445,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36445,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36445,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36445,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36445,36,'cp 27','c7c5 f1e2 g6g5 h4g3 f6h5 e1g1 e7e6 d1d2 h5g3 h2g3 b8c6 d4d5 e6d5 c3d5 c6e7 d5c3 c8e6 f1d1 d8b6 a1c1 a8d8 d2c2 d6d5 c4d5 e7d5 c3d5 d8d5 d1d5 e6d5 b2b3 f8c8 e2d3 d5e6 d3c4 e6c4 c2c4 b6e6 c4e6 f7e6');
INSERT INTO evaluations VALUES(18581,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(18581,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(18581,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(18581,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(18581,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(18581,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(18581,36,'cp 43','c2c3 g8f6 e4f6 d7f6 g1f3 c8g4 h2h3 g4f3 d1f3 d8d5 f1e2 e7e6 f3g3 e8c8 a2a4 h7h5 g3d3 g7g6 e1g1 d5f5 d3c4 f5d5 h3h4 d5c4 e2c4 f8h6 c1h6 h8h6 a4a5 h6h8 c4b3 f6d5 f1e1 a7a6 a1d1 d5e7 e1e5 c8c7 g2g3');
INSERT INTO evaluations VALUES(11505,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(11505,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(11505,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(11505,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(11505,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(11505,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(11505,36,'cp 163','b5c6 d7c6 f3e5 f8c5 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 f2f4 a8d8 d2b3 c5e3 e4e3 e6g4 b3c5 b7b6 c5e4 e7f7 e4f2 g4e6 a2a3 h7h6 a1d1 d8d5 h2h3 a6a5 b2b4 f7g6 g3g4 a5b4 a3b4');
INSERT INTO evaluations VALUES(11496,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(11496,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(11496,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(11496,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(11496,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(11496,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(11496,36,'cp 175','d7c6 f3e5 f8c5 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 f2f4 a8d8 d2b3 c5e3 e4e3 e6g4 b3c5 b7b6 c5e4 e7f7 a2a3 h7h6 e4f2 g4e6 a1d1 d8d5 h2h3 a6a5 d1d2 d5d2 e3d2 c6c5 d2e3 f7f5 e1e2 g7g5 f2g4 g8g7 g4f6 g5f4 g3f4 f5h3');
INSERT INTO evaluations VALUES(11455,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(11455,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(11455,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(11455,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(11455,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(11455,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(11455,36,'cp 156','f3e5 f8c5 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 f2f4 a8d8 d2f3 e6d5 e3c5 e7c5 e4e3 c5e3 e1e3 h7h6 h2h4 c6c5 g1f2 d5e6 e3e2 d8d3 e2d2 d3d2 f3d2 f8d8 d2f1 h6h5 f2f3 a6a5 f1e3 g7g6 f3e2 g8f7 a2a3 b7b5 a1d1 d8d1 e2d1 a5a4 d1e2 c7c6 e2f3 e6f5');
INSERT INTO evaluations VALUES(11440,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(11440,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(11440,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(11440,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(11440,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(11440,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(11440,36,'cp 150','f8c5 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 f2f4 a8d8 d2b3 c5e3 e4e3 e6g4 b3c5 b7b6 c5e4 e7f7 e4f2 g4e6 a2a3 d8d5 a1d1 a6a5 h2h3 h7h6 d1d4 d5d4 e3d4 c6c5 d4e3 f7f5 e1e2');
INSERT INTO evaluations VALUES(7198,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(7198,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(7198,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(7198,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(7198,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(7198,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(7198,36,'cp 182','c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 e3c5 e7c5 e4e3 c5e3 e1e3 a6a5 f2f4 b7b6 d2e4 c6c5 a2a3 a8d8 g1f2 h7h6 f2f3 h6h5 a1e1 e6f5 f3g2 g8f7 h2h3 f7e7 e1e2 a5a4 e4f2 f5d7');
INSERT INTO evaluations VALUES(42386,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(42386,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(42386,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(42386,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(42386,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(42386,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(42386,36,'cp 103','d2d4 b8c6 f3e5 d6e5 d4e5 c6e5 c4e2 d7d5 b1c3 c8e6 f2f4 e5c6 f4f5 d5d4 f5e6 d4c3 e6f7 e8f7 e2c4 f7e8 d1d8 a8d8 b2c3 e8d7 e1g1 h8f8 c1g5 d8e8 c4d3 c6e5 f1f5 d7c8 a1f1 f6d7 f5f8 d7f8 g5f4 f8d7 g1f2 e5g4 f2f3 g4e5 f3e3 e5g4 e3d2 g4f6 f1e1 d7e5 f4g3 a7a6 c3c4 f6d7 d2e3 g7g6 e1f1 c7c5 h2h3 h7h6');
INSERT INTO evaluations VALUES(42387,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(42387,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(42387,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(42387,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(42387,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(42387,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(42387,36,'cp 93','b8c6 f3e5 d6e5 d4e5 c6e5 c4e2 d7d5 b1c3 c8e6 f2f4 e5c6 f4f5 d5d4 f5e6 d4c3 e6f7 e8f7 e2c4 f7e8 d1d8 a8d8 b2c3 e8d7 c1g5 d8e8 e1g1 h8f8 c4d3 c6e5 f1f5 d7c8 a1f1 f6d7 f5f8 d7f8 g5f4 f8d7 f4g3 a7a6 g1f2 e5g4 f2e2 g4f6 e2f3 d7e5 f3e3 e5g4 e3d2 b7b5 h2h3 g4e5 g3e5 e8e5 d2e3 c7c5 c3c4');
INSERT INTO evaluations VALUES(10733,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(10733,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(10733,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(10733,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(10733,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(10733,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(10733,36,'cp 104','f3e5 d6e5 d4e5 c6e5 c4e2 d7d5 b1c3 c8e6 f2f4 e5c6 f4f5 d5d4 f5e6 d4c3 e6f7 e8f7 e2c4 f7e8 d1d8 a8d8 b2c3 e8d7 e1g1 d8f8 c4d3 d7c8 c1e3 c6e5 a2a4 a7a6 g2g3 f8d8 a1d1 h8e8 a4a5 d8d6 e3c5 d6e6 c5d4 f6d7 g1g2 c7c5 d4g1 e5d3 c2d3');
INSERT INTO evaluations VALUES(53837,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(53837,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(53837,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(53837,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(53837,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(53837,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(53837,36,'cp -6','g8f6 d2d3 b8d7 c2c3 g7g6 e1g1 f8g7 f1e1 e8g8 h2h3 b7b5 a4c2 c8b7 a2a4 a7a6 d3d4 f8e8 a4b5 a6b5 a1a8 d8a8 d4e5 d6e5 c2d3 b7a6 c1e3 c6c5 b1d2 c5c4 d3f1 e8c8 e3g5 a6b7 g5f6 d7f6 f3e5 f6e4 d2e4 b7e4 e5g4 e4d5 g4e3');
INSERT INTO evaluations VALUES(45880,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45880,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45880,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45880,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45880,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(45880,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(45880,36,'cp -10','d2d3 b8d7 c2c3 g7g6 e1g1 f8g7 f1e1 e8g8 a4c2 d6d5 a2a4 f8e8 h2h3 d8c7 a4a5 a8b8 b2b4 b7b5 a5b6 a7b6 c2b3 c8b7 d1c2 c6c5 b1a3 c5b4 c3b4 c7c2 a3c2 d5e4 f3g5 b7d5 b3d5 f6d5 d3e4 d5f4 e1d1 d7f6 c1f4 e5f4');
INSERT INTO evaluations VALUES(45879,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45879,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45879,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(45879,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(45879,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(45879,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(45879,36,'cp -15','f8e7 c2c3 c8g4 b1d2 b8d7 d2f1 e8g8 h2h3 g4e6 f1g3 d6d5 e1g1 d5e4 d3e4 e6c4 f1e1 d8c7 a4b3 c4b3 a2b3 f8d8 b3b4 d7c5 d1c2 c5d3 e1d1 d3c1 d1d8 a8d8 c2c1 g7g6 c1e3 c6c5 b4b5 a7a5 f3d2');
INSERT INTO evaluations VALUES(15454,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(15454,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(15454,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(15454,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(15454,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(15454,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(15454,36,'cp 53','c6e7 a2a4 g7g6 f1b5 c8d7 c1g5 f8g7 b5d7 d8d7 a4a5 c7c6 e1g1 c6d5 g5f6 g7f6 c3d5 e7d5 d1d5 a8c8 d5d3 e8g8 f1d1 f6e7 c2c4 c8c5 b2b4 c5c7 d3e3 d7c6 f3d2 b7b5 a5b6 c6b6 e3b6 a7b6 a1a6 e7g5 d2f1 g5h6 h2h4 h6f4');
INSERT INTO evaluations VALUES(14091,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(14091,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(14091,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14091,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(14091,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(14091,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(14091,36,'cp 58','a2a4 g7g6 f1b5 c8d7 c1g5 f8g7 d1d3 a7a6 b5d7 d8d7 f3d2 e8g8 a4a5 f6h5 g2g3 f7f6 g5e3 f6f5 f2f3 h5f6 e1c1 a8b8 d2c4 e7c8 c1b1 f5e4 f3e4 f6g4 d1f1 f8f1 d3f1 g4e3 c4e3 b7b6 f1d3 b6a5 d3a6');
INSERT INTO evaluations VALUES(23452,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(23452,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(23452,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(23452,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(23452,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(23452,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(23452,36,'cp -21','d7d5 e4d5 d8d5 b1c3 d5e6 d1e2 c6d4 e2e6 c8e6 f1b5 d4b5 c3b5 e8d7 g1f3 f7f6 d2d3 e6d5 e1f2 g8h6 c2c4 d5f7 h2h3 f7g6 c1e3 h6f5 e3c5 f5d6 c5d6 e7d6 h1d1 h7h5 d3d4 h5h4 d4d5 a7a5 d1e1 c7c6 b5d4 g6f7 d4f5 c6d5');
INSERT INTO evaluations VALUES(21054,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(21054,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(21054,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(21054,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(21054,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(21054,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(21054,36,'cp -30','e4d5 d8d5 b1c3 d5e6 d1e2 c6d4 e2e6 c8e6 f1b5 d4b5 c3b5 e8d7 g1f3 f7f6 d2d3 e6d5 e1f2 b7b6 c1d2 e7e6 c2c4 d5b7 a2a4 a7a5 h1e1 g8h6 d2c3 a8e8 h2h3 f8c5 d3d4 c5f8 c4c5 h6f5 g2g4 f5e7 c5b6 c7b6 a1c1 b7d5 g4g5');
INSERT INTO evaluations VALUES(20958,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(20958,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(20958,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(20958,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(20958,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(20958,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(20958,36,'cp -25','d8d5 b1c3 d5e6 d1e2 c6d4 e2e6 c8e6 f1b5 d4b5 c3b5 e8d7 g1f3 f7f6 d2d3 a7a6 b5c3 h7h5 f3h4 e6f7 c1d2 e7e6 c3e4 g8e7 g2g3 e7f5 h4f3 a6a5 a2a4 b7b6 e1c1 a8d8 h1e1 d7c8 e4c3 f5d4 f3d4 d8d4 b2b3 d4d8 c3e4');
INSERT INTO evaluations VALUES(1610,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(1610,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(1610,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(1610,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(1610,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(1610,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(1610,36,'cp -25','b1c3 d5e6 d1e2 c6d4 e2e6 c8e6 f1b5 d4b5 c3b5 e8d7 g1f3 f7f6 e1f2 e6f7 d2d3 e7e6 h1e1 g8e7 c1d2 e7c6 a2a3 a7a6 b5c3 f7h5 a1c1 h5f3 f2f3 e6e5 f4e5 c6e5 f3f2 h7h5 h2h3 h5h4 c3e4 h8h5 b2b4 b7b6');
INSERT INTO evaluations VALUES(35173,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(35173,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(35173,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(35173,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(35173,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(35173,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(35173,36,'cp 75','b1c3 f8c5 d1e2 d7d6 g1f3 h4d8 c3a4 c5b6 a4b6 a7b6 d2d4 b8c6 c1g5 g8f6 d4d5 c6e7 e1c1 e7g6 h2h4 e8g8 h4h5 g6f4 g5f4 e5f4 f3d4 h7h6 f2f3 c8d7 c1b1 d8e8 e2d2 e8e5 c4b5 d7b5 d4b5 f8c8 d2d4 c7c6 d4e5 d6e5 d5c6 b7c6 b5d6 c8d8 d6c4');
INSERT INTO evaluations VALUES(35168,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(35168,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(35168,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(35168,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(35168,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(35168,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(35168,36,'cp 59','f8c5 d1e2 d7d6 g1f3 h4d8 e1g1 h7h6 c3a4 c5b6 a4b6 a7b6 d2d4 d8f6 e2e3 c8g4 c4b5 g4d7 b5e2 b8c6 d4d5 c6e7 f3e1 e7g6 e1d3 g8e7 c2c4 c7c5 d5c6 e7c6 f2f4 e5f4 d3f4 g6f4 e3f4 f6f4 c1f4 c6e5 f4e5 d6e5');
INSERT INTO evaluations VALUES(33287,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(33287,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(33287,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(33287,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(33287,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(33287,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(33287,36,'cp 65','d1e2 d7d6 g1f3 h4d8 e1g1 h7h6 c3a4 c5b6 a4b6 a7b6 d2d4 d8f6 c1e3 g8e7 e2d2 e7g6 a2a4 e8g8 b2b3 b8c6 d4d5 c6e7 g1h1 g6f4 e3f4 f6f4 d2f4 e5f4 f3d4 e7g6 c4d3 c8d7 d3b5 g6e5 b5d7 e5d7 d4b5 f8c8 a1d1 d7e5 f2f3 g8h8');
INSERT INTO evaluations VALUES(42040,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42040,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42040,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42040,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42040,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(42040,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(42040,36,'cp 0','e8g8 g1f3 b7b6 c4d5 e6d5 a1c1 c8b7 f1d3 b4e7 e1g1 c7c5 f3e5 b8c6 a2a3 e7d6 e5g4 d6e7');
INSERT INTO evaluations VALUES(37038,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37038,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37038,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37038,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37038,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37038,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37038,36,'cp 1','a2a3 b4e7 g1f3 b7b6 c4d5 e6d5 b2b4 c7c5 b4c5 b6c5 f1b5 b8d7 d4c5 d7c5 e1g1 a8b8 f3d4 c8d7 a3a4 a7a6 b5c6 c5e6 c3e2 e6d4 e2d4 e7c5 a1c1 c5d4 e3d4 b8c8 c6b7 c8c1 d1c1 d7a4 b7a6 a4d7 a6d3');
INSERT INTO evaluations VALUES(37050,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37050,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37050,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37050,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37050,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37050,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37050,36,'cp 12','b4e7 g1f3 b7b6 c4d5 e6d5 b2b4 c7c5 b4c5 b6c5 d4c5 e7c5 f1b5 a7a6 b5e2 b8c6 c3a4 c5a7 a1c1 c8b7 a4c5 a7c5 c1c5 f6e4 c5c2 d5d4 e3d4 c6d4 f3d4 d8d4 e1g1 e4d2 c2d2 d4h4 d2b2 b7e4 e2f3 a8d8 d1a4 f8e8 f3e4 e8e4 a4a6');
INSERT INTO evaluations VALUES(43259,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(43259,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(43259,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(43259,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(43259,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(43259,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(43259,36,'cp 186','e4e5 f6g8 c1f4 d7d5 h2h3 e7e6 g1f3 g8e7 d1d2 b8d7 h3h4 c7c5 h4h5 a7a6 h5h6 g7f8 d4c5 e7c6 c3a4 d7c5 a4c5 f8c5 c2c3 e8g8 a1d1 f7f6 e5f6 d8f6 f4g3 f8f7 f1e2 c8d7 e1g1 c5e7 g3h4 f6f4 h4e7 f4d2 d1d2 f7e7 f3g5 d7e8 f2f4');
INSERT INTO evaluations VALUES(43123,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(43123,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(43123,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(43123,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(43123,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(43123,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(43123,36,'cp 174','f6g8 c1f4 d7d5 d1d2 a7a6 h2h3 e7e6 g1f3 b8d7 h3h4 g8e7 h4h5 c7c5 h5h6 g7f8 d4c5 e7c6 c3a4 d7c5 a4c5 f8c5 c2c3 c8d7 f1e2 c5e7 a1d1 c6a7 e2d3 d8c7 e1g1 d7b5 d3b5 a6b5 f3d4 c7b6 f4e3 b5b4 d4e2 b4c3 e2c3 b6c7 a2a4 a7c6');
INSERT INTO evaluations VALUES(39538,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(39538,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(39538,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(39538,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(39538,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(39538,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(39538,36,'cp 188','c1f4 d7d5 d1d2 a7a6 h2h3 e7e6 g1f3 b8d7 h3h4 g8e7 h4h5 c7c5 h5h6 g7f8 d4c5 e7c6 c3a4 d7c5 a4c5 f8c5 f1e2 e8g8 c2c3 f7f6 e5f6 d8f6 f4g3 c8d7 e1c1 c5e7 e2d3 b7b5 c1b1 e6e5 d3c2 d7e6 g3h4 f6f7 h4e7 f7e7 f3g5 a8d8 g5e6 e7e6 c2b3 c6e7 f2f3 e6f5 b1a1');
INSERT INTO evaluations VALUES(52030,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(52030,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(52030,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(52030,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(52030,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(52030,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(52030,36,'cp 75','b1c3 g8f6 e2e4 d7d6 f1c4 f8e7 g1e2 e6d5 c3d5 e8g8 e1g1 f6d5 c4d5 b8a6 e2f4 a6c7 c2c3 f8e8 f2f3 c7d5 f4d5 c8e6 c1f4 e7f8 a2a4 f7f5 e4f5 e6f5 f1f2 d8d7 f4g3 d7f7 h2h3 a8d8 c3c4 f5e6 f2d2 e6d5 d2d5');
INSERT INTO evaluations VALUES(52028,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(52028,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(52028,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52028,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(52028,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(52028,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(52028,36,'cp 58','d7d6 e2e4 g8f6 f1c4 f8e7 g1e2 e6d5 c4d5 e8g8 e1g1 b8c6 e2f4 f8e8 f2f3 a7a6 a2a4 a8b8 d5c4 c8d7 c1e3 c6e5 c4e2 b7b5 a4b5 a6b5 f4d5 f6d5 c3d5 d7c6 e3f2 c6d5 d1d5 d8b6 c2c3 b5b4');
INSERT INTO evaluations VALUES(53164,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(53164,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53164,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(53164,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53164,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53164,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(53164,36,'cp 10','g8f6 c4d5 d8d5 d1d5 f6d5 b1a3 e7e5 e2e4 d5b4 c1e3 c8d7 g1f3 f7f6 f1b5 b8a6 e1e2 f8c5 b5d7 e8d7 a3b5 c5e3 f2e3 h8c8 h1d1 d7e6 a2a3 c8c5 b5c3 b4c6 d1d3 c5a5 a1d1 a6c5 d3d6 e6e7 f3h4 g7g6 b2b4 a5a3 b4c5 a3c3 d6d7 e7e8 d7h7 c3c2 e2f1');
INSERT INTO evaluations VALUES(54815,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(54815,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(54815,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(54815,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(54815,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54815,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(54815,36,'cp 71','e2e4 g8f6 b1c3 f8g7 f1e2 e8g8 g1f3 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 b8d7 a1b1 h7h6 a2a3 b7b6 d1c2 g8h7 c1e3 h6h5 g2g4 f6g8 g4g5 g8e7 f3g2 f7f5 g5f6 g7f6 f1e1 g6g5 c2d1 e7g6 d1h5 h7g7 b2b4');
INSERT INTO evaluations VALUES(54816,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(54816,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(54816,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(54816,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(54816,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(54816,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(54816,36,'cp 71','f8g7 b1c3 g8f6 f1e2 e8g8 g1f3 e7e5 e1g1 e5d4 f3d4 f8e8 f2f3 c7c6 g1h1 a7a6 d4c2 c8e6 c1g5 h7h6 g5e3 b7b5 e3f4 g7f8 d1d2 e6c4 e2c4 b5c4 a1d1 b8d7 f4d6 f8d6 d2d6 d8b6 c3a4 b6b5 d6a3 b5g5 c2e3 d7e5 a4b6');
INSERT INTO evaluations VALUES(15085,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(15085,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(15085,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(15085,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(15085,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(15085,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(15085,36,'cp 160','e4d5 c6d4 h2h3 f8d6 e1g1 e8g8 a2a4 c7c6 d5c6 d4c6 d2d3 h7h6 c3e4 f8e8 c2c3 c8e6 c4e6 e8e6 b2b4 d8c7 c1e3 a8d8 d1b3 d6e7 a1d1 f6d5 e3d2 d5f6 b3b1 b7b6 f1e1 c7d7 e4f6 e7f6 b4b5');
INSERT INTO evaluations VALUES(14984,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(14984,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(14984,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(14984,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(14984,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(14984,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(14984,36,'cp 159','c6d4 h2h3 f8d6 e1g1 e8g8 d2d3 a7a6 a2a4 c8d7 a4a5 d4f3 d1f3 h7h6 c1d2 a8c8 c3e4 f6e4 d3e4 f7f5 c4d3 f5e4 f3e4 d7f5 e4c4 f5d3 c4d3 d8h4 a1e1 h4d4 d3d4 e5d4 e1e4 d4d3 c2c4 c8e8 e4e8 f8e8 d2e3 e8e4 b2b3 d3d2 e3d2 e4d4 d2e3 d4d3 g2g4 d3c3 f1e1 c3b3 c4c5 d6e5');
INSERT INTO evaluations VALUES(15812,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(15812,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(15812,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(15812,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(15812,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(15812,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(15812,36,'cp 159','h2h3 f8d6 d2d3 a7a6 a2a4 e8g8 e1g1 h7h6 f1e1 d4f3 d1f3 c8d7 c1d2 a8c8 e1e2 c7c6 c3e4 f6e4 d3e4 d8e7 c2c3 g8h7 e2e1 h7h8 g2g4 c6d5 c4d5 d7e6 c3c4 e6d5 c4d5 d6b4 d2b4 e7b4 f3a3 c8c4 b2b3 b4a3 a1a3');
INSERT INTO evaluations VALUES(44097,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44097,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(44097,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44097,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44097,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44097,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44097,36,'cp 0','d7d5 d2d4 e5e4 a2a3 f8e7 f2f3 e4f3 g1f3 h7h6 e1g1 e8g8 f3h4 d5c4 e2e4 c8g4 g2f3 g4e6 f3g2');
INSERT INTO evaluations VALUES(45814,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45814,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45814,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45814,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45814,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45814,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45814,36,'cp 0','d2d4 e5e4 a2a3 f8e7 f2f3 e4f3 g1f3 h7h6 e1g1 e8g8 f3h4 d5c4 e2e4 c8g4 g2f3 g4e6 f3g2');
INSERT INTO evaluations VALUES(17812,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(17812,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(17812,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(17812,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(17812,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(17812,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(17812,36,'cp -17','e5d4 d1d4 d8d4 f3d4 c8d7 f2f3 e8c8 b1c3 g8e7 h2h4 h7h5 e1f2 e7g6 c1e3 g6e5 a1d1 d8e8 b2b3 g7g6 d4e2 f8d6 e2g3 b7b5 c3e2 h8f8 h1e1 c8b7 e2f4 a6a5 f4d3 e5d3 d1d3 a5a4 b3a4 d7e6 a4b5 c6b5 e1a1 e8a8');
INSERT INTO evaluations VALUES(17822,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(17822,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(17822,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(17822,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(17822,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(17822,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(17822,36,'cp -12','d1d4 d8d4 f3d4 c8d7 b1c3 e8c8 f2f3 d8e8 e1f2 g8e7 h2h4 c6c5 d4e2 f7f5 c1f4 f5e4 c3e4 d7c6 e2g3 e7d5 f4d2 c5c4 a1e1 d5b4 a2a3 b4c2 e1c1 c6a4 h4h5 e8d8 g3f5 b7b5 g2g4 b5b4 a3b4 f8b4 d2b4 c2b4 f5g7 b4d3 f2g3 d3c1');
INSERT INTO evaluations VALUES(17821,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(17821,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(17821,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(17821,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(17821,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(17821,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(17821,36,'cp -17','d8d4 f3d4 c8d7 f2f3 e8c8 b1c3 d8e8 e1f2 g8e7 c1e3 b7b6 a1d1 f7f5 h2h4 f5e4 c3e4 c6c5 d4e2 e7f5 e3f4 d7c6 e2g3 f5g3 e4g3 g7g6 h4h5 f8g7 c2c3 a6a5 d1e1 h8f8 e1e8 f8e8 h5g6 h7g6 h1e1 e8e1 f2e1 c6d5 a2a3 c8d7 f4e3 d7c6 g3e2 d5b3 e3g5 b3c2 e2f4 g7e5');
INSERT INTO evaluations VALUES(1204,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(1204,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(1204,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(1204,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(1204,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(1204,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(1204,36,'cp -11','f3d4 c8d7 b1c3 e8c8 f2f3 d8e8 e1f2 g8e7 c1e3 b7b6 h2h4 c6c5 d4e2 f7f5 a1d1 f5e4 c3e4 e7f5 e3f4 d7c6 e4g3 f5g3 e2g3 g7g6 h4h5 f8g7 c2c3 a6a5 d1e1 h8f8 e1e8 f8e8 a2a3 a5a4 h1e1 e8e1 f2e1 c8d7 h5g6');
INSERT INTO evaluations VALUES(1201,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(1201,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(1201,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(1201,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(1201,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(1201,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(1201,36,'cp -16','c8d7 f2f3 e8c8 b1c3 d8e8 e1f2 g8e7 c1e3 f7f5 h2h4 f5e4 c3e4 e7d5 e3d2 c6c5 d4e2 d5b4 h1c1 d7c6 e2c3 h7h6 a2a3 b4d5 c3d5 c6d5 c1e1 d5e4 e1e4 e8e4 f3e4 f8e7 a1h1 e7f6 d2c3 f6c3 b2c3 c8d7 h1d1 d7e7 f2e3 b7b6 c3c4 h8a8 a3a4 a6a5 e3f4 a8f8 f4e5 f8f2');
INSERT INTO evaluations VALUES(27499,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(27499,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(27499,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(27499,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(27499,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(27499,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(27499,36,'cp -14','f2f3 e8c8 b1c3 d8e8 e1f2 g8e7 d4e2 f7f5 h2h4 c6c5 c1f4 d7c6 e2g3 f5e4 c3e4 e7d5 f4d2 d5b4 h1c1 c5c4 a2a3 b4d5 c1e1 e8e5 g3e2 h7h6 e2c3 f8e7 c3d5 e5d5 d2c3 d5f5 e4g3 e7c5 f2f1 f5f4 h4h5 h8f8 c3g7 f4f3 g2f3');
INSERT INTO evaluations VALUES(47094,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(47094,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(47094,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(47094,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(47094,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(47094,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47094,36,'cp 23','f6d5 g1f3 c8g4 e1g1 e7e6 d2d4 f8e7 h2h3 g4h5 c4e2 e8g8 f3e5 h5e2 d1e2 c7c5 d4c5 e7c5 c2c4 d5e7 f1d1 d8c8 b1c3 b8c6 e5d7 c6d4 e2h5 c8d7 h5c5 e7c6 b2b4 f8d8 c1e3 e6e5 b4b5 d7e7 c3a4 b7b6 c5e7 c6e7 e3d4 e5d4 c4c5 b6c5 a4c5 a8c8 c5b3 d4d3 a1c1 c8c1 d1c1 d3d2 c1d1');
INSERT INTO evaluations VALUES(47437,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(47437,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(47437,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(47437,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(47437,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(47437,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(47437,36,'cp 34','d2d4 e7e6 g1f3 f8e7 e1g1 e8g8 d1e2 a7a6 c4b3 b7b5 a2a4 c7c6 f1e1 c8b7 b3d5 d8d5 b1c3 d5d8 c1e3 d8c7 c3e4 b8d7 e4c5 a6a5 e3f4 c7f4 c5d7 f8d8 d7c5 f4c7 a4b5 c6b5 e2b5 b7c6 b5e2 a5a4 c2c3 a4a3 b2a3');
INSERT INTO evaluations VALUES(47438,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(47438,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(47438,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47438,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47438,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(47438,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(47438,36,'cp 19','e7e6 g1f3 f8e7 e1g1 e8g8 f1e1 c7c5 b1c3 b8c6 c4d5 e6d5 d4c5 d5d4 c3e4 c8f5 c1f4 f5e4 e1e4 e7c5 d1e2 d8b6 a1b1 a7a5 f4d2 b6a6 e2a6 a8a6 f3e1 f8d8 e1d3 c5d6 b1e1 f7f6 a2a4 g8f7 g1f1 a6b6 b2b3 d8c8');
INSERT INTO evaluations VALUES(16145,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(16145,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(16145,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(16145,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(16145,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(16145,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(16145,36,'cp 25','f8c5 c2c3 c5b6 d2d4 d7d6 a2a4 a7a5 c1e3 f6g4 c4a2 e8g8 b1d2 b6a7 h2h3 g4e3 f2e3 g8h8 e1f1 f7f5 d1e1 c6e7 f3h4 f5f4 e3f4 e5f4 e4e5 d6d5 a2b1 c7c5 d2f3 c5d4 f3d4 a7d4 c3d4 g7g5 h4f3 d8b6 e1e2 c8f5 b1d3 b6b4 d3f5 e7f5 a1e1');
INSERT INTO evaluations VALUES(9780,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(9780,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(9780,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(9780,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(9780,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(9780,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(9780,36,'cp 30','c2c3 d7d6 d2d4 c5b6 a2a4 a7a5 c1e3 c8g4 d1b3 e8g8 b1d2 e5d4 f3d4 g4d7 h2h3 c6e5 c4f1 f8e8 a1d1 a8b8 b3c2 d6d5 d4b5 b6e3 e1e3 d7b5 a4b5 c7c6 e4d5 c6d5 d2b3 d8b6 b3d4');
INSERT INTO evaluations VALUES(9762,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(9762,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(9762,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(9762,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9762,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(9762,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(9762,36,'cp 24','c5b6 d2d4 d7d6 a2a4 a7a5 c1e3 f6g4 h2h3 g4e3 f2e3 e8g8 b1d2 c6e7 c4a2 e5d4 e3d4 c7c6 d1c2 d6d5 e1e2 c8e6 a1e1 d5e4 a2b1 e6f5 d2e4 e7d5 e4f6 d8f6 c2f5 f6f5 b1f5 g7g6 f5c2 b6c7 g1f2 a8d8 g2g3 g8g7 c2b3 d8d7 b3d5 d7d5 f3d2 f8d8');
INSERT INTO evaluations VALUES(14225,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(14225,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(14225,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(14225,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(14225,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(14225,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(14225,36,'cp 16','c6b4 d3b5 c7c6 b5a4 c8f5 a2a3 b4a6 g1f3 e7e6 b1d2 a6b8 f3e5 f8d6 c2c4 e8g8 e1g1 d6c7 a4c2 b8d7 c2f5 e6f5 c4d5 f6d5 f4g3 f8e8 d1b3 a8b8 e5d3 g7g6 g3c7 d8c7 f1c1 d7f6 d2f3 g8g7 h2h3');
INSERT INTO evaluations VALUES(14499,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(14499,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14499,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(14499,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(14499,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(14499,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(14499,36,'cp 21','d3b5 c7c6 b5a4 c8f5 a2a3 b4a6 b1d2 a6b8 g1f3 e7e6 f3e5 h7h5 c2c4 b8d7 a4c2 d7e5 f4e5 d5c4 c2f5 e6f5 d1c2 g7g6 d2c4 f8e7 a1c1 e8g8 c2b3 b7b5 c4d2 d8d7 e1g1 f8c8 h2h3 a7a5 e5f6 e7f6 c1c3 a5a4 b3c2');
INSERT INTO evaluations VALUES(44840,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44840,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(44840,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44840,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44840,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44840,36,'cp 0','b8c6 d2d3 e7e5 f1e2 f8b4 c1d2 c8f5 d2b4 c6b4 d1a4 b4c6 a4a3 c6d4 e2d1 d4c6');
INSERT INTO evaluations VALUES(13188,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13188,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(13188,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13188,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13188,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13188,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(13188,36,'cp 12','d2d3 e7e5 f1e2 f8b4 c1d2 c8f5 g1f3 d8e7 f3h4 f5e6 e1g1 e8g8 h4f3 a8d8 d2b4 c6b4 d1b3 e6f5 f1d1 e7c5 h2h3 f6h5 e2f1 h5f4 f3e5 c5e5 b3b4 f5d3 b4b5 e5b5 c4b5 d3f1 d1f1 g7g6 b1a3 d8d2 b2b3 f8d8 a3c4 d2c2');
INSERT INTO evaluations VALUES(13186,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(13186,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(13186,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(13186,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13186,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(13186,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13186,36,'cp 8','c8f5 g1f3 e7e5 f1e2 f8b4 c1d2 d8e7 f3h4 f5e6 e1g1 e8g8 h4f3 a8d8 d2b4 c6b4 d1b3 e6f5 f1d1 f8e8 a2a3 b4d3 e2d3 e5e4 d3e4 d8d1 b3d1 f6e4 d1e1 e7d8 b1c3 e4g5 e1d1 g5f3 d1f3 d8d2 a1f1 g7g6 h2h3 d2b2 c3b5 f5e4 f3g3 e4c6 b5a7 b2b6 a7c6 b6c6 g1h1 c6c4');
INSERT INTO evaluations VALUES(40478,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(40478,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(40478,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(40478,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(40478,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(40478,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(40478,36,'cp 66','f1d3 e8g8 e1g1 c6c5 c4d5 e6d5 d4c5 b8d7 c3e2 d7c5 b2b3 d8e7 c1b2 c5e4 h2h3 b4d6 a1c1 f8e8 d3c2 c8d7 d1d4 d7b5 c2d3 a7a6 e2f4 b5d3 d4d3 a8d8 c1c2 h7h6 f1c1');
INSERT INTO evaluations VALUES(40483,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(40483,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(40483,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(40483,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(40483,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(40483,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(40483,36,'cp 50','e8g8 e1g1 b7b6 d1c2 c8b7 a2a3 b4c3 c2c3 d5c4 d3c4 f6e4 c3c2 c6c5 d4c5 e4c5 b2b4 b7e4 c2e2 c5a4 f3d4 b8c6 d4c6 e4c6 f2f3 c6b7 e3e4 a8c8 c4d3 h7h6 c1f4 d8e7 f4e3 a4c3 e2d2 f8d8 f1c1');
INSERT INTO evaluations VALUES(36104,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(36104,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36104,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(36104,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(36104,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(36104,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(36104,36,'cp 52','e1g1 b7b6 d1c2 b8d7 a2a3 b4c3 c2c3 d5c4 d3c4 c6c5 d4c5 d7c5 b2b4 c5e4 c3e5 a7a5 b4b5 c8b7 c1b2 a8c8 a1c1 b7d5 c4e2 c8c1 f1c1 e4f2 f3g5 f2e4 g5e4 d5e4 b2d4 h7h6 h2h3 d8b8 c1c7 f6e8 c7d7 b8e5 d4e5');
INSERT INTO evaluations VALUES(36103,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(36103,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(36103,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(36103,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36103,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(36103,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(36103,36,'cp 37','b8d7 d1c2 b7b6 c4d5 c6d5 c1d2 c8b7 a2a3 b4e7 f1c1 d8b8 c2a4 f8c8 c3e2 c8c1 a1c1 b8d8 d2b4 e7b4 a3b4 a7a5 b4b5 a8c8 c1c8 d8c8 b2b4 c8a8 h2h3 a5b4 a4b4 a8f8 b4b2 h7h6 e2c3 f8d8 f3e5 d7e5 d4e5');
INSERT INTO evaluations VALUES(2924,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(2924,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(2924,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(2924,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(2924,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(2924,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(2924,36,'cp 52','d1c2 b7b6 c4d5 c6d5 c1d2 c8b7 c2b3 b4e7 f1c1 a8c8 a2a3 a7a6 c3e2 c8c1 a1c1 d8b8 d2b4 e7b4 a3b4 b8d6 h2h3 h7h6 c1a1 f8a8 e2f4 a8c8 d3e2 f6e4 f3e1 e4d2 b3d1 d2e4 e2a6 b7a6 a1a6 d6b4 f4d3 b4b5 a6a7 b5c6');
INSERT INTO evaluations VALUES(57246,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(57246,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57246,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57246,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57246,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57246,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(57246,36,'cp 0','c2c4 g8f6 d2d3 d7d5 c4d5 f6d5 g1f3 b8c6 e2e3 a7a5 f1e2 f8d6 e3e4 d5f6 b1c3 e8g8 c1e3 f6g4 e3d2 g4f6 d2e3');
INSERT INTO evaluations VALUES(57132,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57132,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57132,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57132,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57132,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(57132,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(57132,36,'cp 0','g8f6 d2d3 d7d5 c4d5 f6d5 g1f3 b8c6 e2e3 a7a5 f1e2 f8e7 e1g1 c8e6 e3e4 d5f6 b1c3 e8g8 c1e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(49577,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49577,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49577,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49577,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49577,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49577,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49577,36,'cp 0','d2d3 b8c6 g1f3 d7d5 c4d5 f6d5 e2e3 a7a5 f1e2 f8e7 e1g1 c8e6 e3e4 d5f6 b1c3 e8g8 c1e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(49575,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49575,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49575,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49575,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49575,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49575,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49575,36,'cp 0','d7d5 c4d5 f6d5 g1f3 b8c6 e2e3 a7a5 f1e2 c8e6 e1g1 f8e7 e3e4 d5f6 b1c3 e8g8 c1e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(52978,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(52978,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(52978,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(52978,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(52978,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(52978,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(52978,36,'cp 87','c2c4 g7g6 b1c3 f8g7 d2d4 g8f6 f1e2 e8g8 e1g1 e7e5 d4d5 b8a6 c1e3 f6g4 e3g5 f7f6 g5h4 c6c5 f3e1 g4h6 g2g4 b7b6 g1h1 a6c7 e1g2 c8d7 h2h3 h6f7 h4g3 a8b8 a2a4 a7a6 g3h2 g7h6 f2f4 e5f4');
INSERT INTO evaluations VALUES(3819,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3819,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3819,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3819,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3819,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3819,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3819,36,'cp 0','c6a5 c3a4 a5c4 a4c5 d6c5 d3c4 f6e4 f3e5 f8e8 d1d8 e8d8 g2g4 e4d6 b2b3 b7b6 f2f3 f7f6 e5d3 h7h5 a2a4 a7a5 d3f4 h5g4 h3g4 c8b7 g1f2 d8d7 f1h1 a8e8 c1e3 d7e7 a1e1 g8f7 h1h7 f7g8');
INSERT INTO evaluations VALUES(3926,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3926,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3926,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3926,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3926,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3926,36,'cp 0','c3a4 a5c4 a4c5 d6c5 d3c4 f6e4 f3e5 b7b6 b2b3 f7f6 e5d3 c8f5 d1f3 d8d7 f1d1 e4c3 d1d2 c3e4');
INSERT INTO evaluations VALUES(39196,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(39196,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(39196,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(39196,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(39196,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(39196,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(39196,36,'cp -17','d8e7 h5e2 g8f6 d2d3 e8g8 g1f3 c7c6 e1g1 h7h6 c4b3 d7d6 a2a4 f8e8 a4a5 b8d7 b1c3 c5b4 a1a4 b4c3 b2c3 d7c5 a4a1 c5b3 c2b3 c8e6 c3c4 f6d7 h2h3 c6c5 c1e3 f7f5 e4f5 e6f5 f3d2 d7f6 f1e1');
INSERT INTO evaluations VALUES(33323,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(33323,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(33323,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(33323,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(33323,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(33323,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(33323,36,'cp -12','h5e2 g8f6 d2d3 c7c6 g1f3 e8g8 e1g1 f8e8 c4b3 h7h6 c2c3 d7d6 c1e3 c5e3 e2e3 c8e6 b3c2 b8d7 d3d4 a7a5 a2a4 d6d5 f3e5 f6e4 b1d2 e6f5 e3f4 e7e6 d2e4 f5e4 e5d3 e6g6 f4g3 g6g3 h2g3 e4d3 c2d3');
INSERT INTO evaluations VALUES(55011,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(55011,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(55011,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(55011,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(55011,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(55011,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(55011,36,'cp -100','e7e5 d2d3 c8e6 c2c3 a7a6 b5a3 c7c5 g1f3 b8c6 f3g5 e6c8 f2f4 e5f4 c1f4 h7h6 g5f3 g8e7 f1e2 e7g6 f4g3 f8d6 g3d6 d8d6 c3d4 c6d4 f3d4 d6d4 d1d2 c8e6 d2c3 a8d8 c3d4 d8d4 e1d2 e8g8 d2e3 f7f5 h1c1 f5e4');
INSERT INTO evaluations VALUES(56115,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(56115,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(56115,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(56115,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(56115,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(56115,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(56115,36,'cp -74','d2d3 c8e6 f1e2 a7a6 b5a3 f8b4 c1d2 b4d2 d1d2 g8e7 g1f3 f7f6 c2c3 c7c5 e1g1 b8c6 e2d1 e7g6 c3d4 c6d4 a1c1 a8c8 f3d4 d8d4 g2g3 e8g8 d1b3 e6b3 a2b3 b7b5 d2a5 c8c6 a5c3 d4d7 b3b4 f8c8 c3b3 d7f7 b3f7 g8f7 b4c5 c6c5 c1c5');
INSERT INTO evaluations VALUES(56113,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(56113,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(56113,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(56113,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(56113,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(56113,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(56113,36,'cp -105','c8e6 c2c3 a7a6 b5a3 c7c5 g1f3 b8c6 f3g5 e6c8 f2f4 e5f4 c1f4 h7h6 g5f3 g8e7 f1e2 e7g6 f4g3 f8d6 g3d6 d8d6 c3d4 c6d4 f3d4 d6d4 d1d2 e8g8 d2c3 d4c3 b2c3 g6f4 a3c4 f4g2 e1d2 c8e6 c4d6 a8b8 d3d4 c5d4 c3d4 g2f4 d4d5 e6d7 a2a4 f4e2 d2e2');
INSERT INTO evaluations VALUES(5256,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(5256,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(5256,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(5256,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(5256,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(5256,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(5256,36,'cp 0','b7c6 f1c4 d7d6 e1g1 g8f6 e4e5 d6e5 d1d8 e8d8 f1e1 c8f5 e1e5 f6d5 e5e2 e7e6 c3a4 d8e7 c1d2 h8b8 c4b3 g7f6 a1d1 g6g5 d2a5 b8b5 a5d2');
INSERT INTO evaluations VALUES(52887,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(52887,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(52887,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(52887,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52887,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52887,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52887,36,'cp 30','b7b5 a2a4 b5b4 c3b1 c8a6 d1c2 d8d4 g1f3 d4b6 c1e3 b6c7 b1d2 e7e6 f1c4 a6c4 d2c4 g8f6 e3g5 b4b3 c2e2 f8b4 f3d2 b4d2 g5d2 e8g8 d2a5 c7b7 a5c3 c6c5 c3f6 g7f6 e1g1 b8c6 e4e5 c6e5 c4e5 f6e5');
INSERT INTO evaluations VALUES(51653,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51653,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(51653,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51653,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(51653,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(51653,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(51653,36,'cp 13','a2a4 b5b4 c3b1 c8a6 d1c2 d8d4 g1f3 d4b6 c1e3 b6c7 f1c4 a6c4 c2c4 e7e6 e1g1 g8f6 b1d2 f6d7 a1c1 f8e7 d2b3 d7b6 c4d4 e8g8 b3c5 b8d7 d4b4 e7c5 e3c5 d7c5 c1c5 b6d7 c5c1 c6c5 b4c4 f8d8 b2b4 a8c8 f1d1 d7b6');
INSERT INTO evaluations VALUES(51663,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(51663,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(51663,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51663,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(51663,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51663,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(51663,36,'cp 11','b5b4 c3b1 c8a6 d1c2 d8d4 g1f3 d4b6 c1e3 b6c7 f1c4 a6c4 c2c4 e7e6 b1d2 g8f6 e1g1 f8d6 e4e5 d6e5 f3e5 c7e5 c4b4 b8d7 f1c1 f6d5 b4b7 e5b8 b7c6 e8g8 c6d7 d5e3 f2e3 f8d8 d7c6 d8d2 b2b4 b8d8 b4b5 h7h6 h2h3 a8b8 c6c3 b8b7 c1c2 d2d3 c3c8 d8c8 c2c8 g8h7');
INSERT INTO evaluations VALUES(51668,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(51668,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(51668,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(51668,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(51668,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(51668,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(51668,36,'cp 14','c3b1 c8a6 d1c2 d8d4 g1f3 d4b6 c1e3 b6c7 b1d2 e7e6 f1c4 a6c4 d2c4 g8f6 e3g5 b4b3 c2e2 f8b4 f3d2 b4d2 c4d2 b8d7 e1g1 e8g8 d2c4 h7h6 g5d2 d7b6 d2a5 c6c5 c4d2 c7c6 a5c3 a8d8 a4a5 b6a4 c3f6 g7f6 d2b3 c5c4 e2g4 g8h7');
INSERT INTO evaluations VALUES(38218,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(38218,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(38218,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(38218,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(38218,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(38218,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(38218,36,'cp 25','c7c5 d2d3 b8c6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 d8e7 d1e2 e8g8 e1g1 c8e6 d5e6 f7e6 a4a5 a8c8 f1b1 e7d7 c2c4 f6d8 b2b4 c5b4 b1b4 b7b6 a5a6 d8f6 h2h3 d6d5 b4b2 f8f7 c4d5 e6d5 e4d5 d7d5');
INSERT INTO evaluations VALUES(37900,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(37900,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37900,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(37900,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(37900,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(37900,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(37900,36,'cp 32','d2d3 b8c6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 e8g8 c2c3 c8e6 d1b3 e6g4 f3d2 f6g5 d2c4 a8b8 e1g1 d8e7 a4a5 g4e6 d5e6 f7e6 b3d1 f8f7 g2g3 b8f8 d1g4 f7f3 h2h4 g5h6 g1g2 f3f6 a1e1 f6f7 b2b3 e7d7 e1b1 f7f3 g4f3 f8f3');
INSERT INTO evaluations VALUES(37901,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(37901,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(37901,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(37901,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(37901,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(37901,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(37901,36,'cp 29','b8c6 a2a4 g8f6 c1g5 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 e8g8 a4a5 a8b8 e1g1 c8e6 d5e6 f7e6 c2c3 b8c8 h2h3 f8f7 d1d2 f6e7 d2e2 d8d7 f1d1 d7b5 d1b1 b5d7 a1a2 e7d8 e2d1 h7h6 b2b3 d8e7 c3c4 c8f8 b3b4 c5b4 d1a4');
INSERT INTO evaluations VALUES(48277,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(48277,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(48277,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(48277,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(48277,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(48277,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48277,36,'cp -24','f8b4 f1d3 c7c5 d4c5 b8d7 g1f3 d7c5 e1g1 b4c3 b2c3 c8d7 c3c4 a8c8 f4e5 e8g8 e5d4 d5c4 d3c4 d8e7 d1b1 b7b6 a2a4 d7c6 c4e2 c6f3 e2f3 e6e5 d4b2 e7c7 b1a2 c5e4 f1d1 f8e8 a1c1 e4c3 b2c3 c7c3 h2h3 h7h6 f3e2 e8d8');
INSERT INTO evaluations VALUES(42004,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(42004,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(42004,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(42004,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(42004,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(42004,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(42004,36,'cp -12','f1d3 c7c5 d4c5 b8d7 g1f3 d7c5 e1g1 b4c3 b2c3 c8d7 f4e5 a8c8 c3c4 e8g8 e5d4 d7a4 d1b1 c5d3 c2d3 d5c4 d3c4 d8e7 f1c1 b7b6 b1b2 f8d8 c4c5 b6c5 b2a3 f6e4 a3a4 c5d4 f3d4 e7f6 f2f3 e4c3 a4a3 c3d5 c1c8 d8c8 a1c1 c8c1 a3c1 h7h5 h2h3 f6g6 g1f2 d5b4 c1c7 g6b1 c7d7 e6e5 d4f5 b4d3 f2g3 b1e1 g3h2');
INSERT INTO evaluations VALUES(42008,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(42008,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(42008,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(42008,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(42008,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(42008,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(42008,36,'cp -25','c7c5 d4c5 b8d7 g1f3 e8g8 e1g1 b4c3 b2c3 d7c5 f4g5 c5e4 d3e4 d5e4 f3d2 e6e5 g5f6 d8f6 d2e4 f6g6 e4g3 g6c6 a2a4 c6c3 d1d6 f8e8 f1c1 c8g4 d6d3 c3c6 c1b1 b7b6 a4a5 g4e6 d3b5 c6b5 b1b5 e6d7 b5d5 d7c6');
INSERT INTO evaluations VALUES(11504,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(11504,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(11504,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(11504,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(11504,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(11504,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(11504,36,'cp 160','b5c6 b7c6 f3e5 f8b4 c2c3 b4e7 e1g1 e8g8 f1e1 a6a5 d3d4 c6c5 d4d5 f8e8 d1c2 e7f8 e5f3 d7d6 c3c4 c8g4 f3d2 g7g5 d2f1 f6h5 c1d2 f8g7 d2c3 g4d7 f1e3 g7d4 c2d1 d4e3 d1h5 e3d4 h5f3 c7c6 c3d4 c5d4 b1d2 c6c5');
INSERT INTO evaluations VALUES(11495,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(11495,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(11495,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(11495,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(11495,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(11495,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(11495,36,'cp 189','b7c6 f3e5 f8b4 c2c3 b4e7 e1g1 e8g8 f1e1 a6a5 d3d4 c6c5 d4d5 f8e8 d1f3 e7f8 c1f4 d7d6 e5c6 d8d7 b1d2 c8b7 f4g5 d7g4 g5f6 g4f3 g2f3 g7f6 a2a4 b7c6 d5c6 a8a6 d2c4 f8h6 g1g2 e8a8 a1d1 a6c6 g2h3 f6f5 e4f5 g8g7 h3g4 a8g8 c4a5');
INSERT INTO evaluations VALUES(11734,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(11734,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(11734,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(11734,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(11734,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(11734,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(11734,36,'cp 195','f3e5 f8b4 c2c3 b4e7 e1g1 e8g8 f1e1 a6a5 d3d4 c6c5 d4d5 f8e8 d1f3 e7f8 c1f4 d7d6 e5c6 d8d7 b1d2 c8b7 f4g5 d7g4 g5f6 g4f3 g2f3 g7f6 a2a4 b7c6 d5c6 f8h6 d2c4 a8a6 g1g2 f6f5 e4f5 e8e1 a1e1 g8g7 g2g3 a6c6 e1d1 c6a6 f3f4 a6a8 b2b3 g7f6 g3g4 f6e7 h2h3 h6g7');
INSERT INTO evaluations VALUES(11727,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(11727,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(11727,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(11727,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(11727,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(11727,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(11727,36,'cp 179','f8b4 c2c3 b4e7 e1g1 a6a5 f1e1 e8g8 d3d4 c6c5 d4d5 f8e8 d1f3 e7f8 c1f4 d7d6 e5c6 d8d7 b1d2 d7g4 f3g4 c8g4 f4g5 f6h5 d2c4 a5a4 e4e5 f7f6 e5d6 f8d6 c4d6 c7d6 g5e3 g4d7 e3d2 g7g5 c3c4 h7h6 f2f3 h5g7 d2c3');
INSERT INTO evaluations VALUES(48674,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(48674,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48674,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48674,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(48674,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(48674,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(48674,36,'cp 26','g1f3 c8g4 e1g1 f8e7 c1f4 e8g8 h2h3 g4h5 b1d2 c7c5 d4c5 e7c5 c2c3 c5d6 f4d6 d8d6 f1e1 b8c6 d1b3 d6c7 e1e3 f8e8 a1e1 e8e3 e1e3 h5g6 d3g6 h7g6 g2g3 a8d8 b3a4 c7b6 d2b3 f6e4 g1g2 a7a6 a4a3 a6a5 b3d4 c6d4 f3d4');
INSERT INTO evaluations VALUES(50182,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(50182,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(50182,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(50182,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(50182,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50182,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50182,36,'cp 46','f1g2 c7c5 g1f3 f8g7 d2d4 d7d5 b1c3 e8g8 d4c5 d8a5 e1g1 d5c4 d1d4 b8c6 d4c4 c8e6 c4b5 a5b5 c3b5 a8d8 c1e3 f6e4 b5d4 c6d4 f3d4 e4c5 f1d1 e6c8 a1b1 c5a4 d1d3 d8d7 b2b3 a4c5 d3d2 f8d8 b1d1 g7f6 a2a3 f6d4 d2d4 d7d4 d1d4');
INSERT INTO evaluations VALUES(50183,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50183,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(50183,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(50183,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50183,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(50183,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50183,36,'cp 28','c7c5 b2b3 f8g7 c1b2 e7e6 g1f3 d7d5 d2d4 c5d4 e1g1 e8g8 c4d5 e6d5 f3d4 b8c6 b1d2 f8e8 e2e3 h7h5 d2f3 f6e4 a1c1 d8b6 d1e2 c6d4 b2d4 g7d4 f3d4 c8d7 h2h4 a7a5 e2d3 a5a4 c1c2 e4c5 d3d2 c5e4 g2e4 d5e4 b3a4 a8a4 c2b2 b6a6 f1b1');
INSERT INTO evaluations VALUES(44416,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44416,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44416,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44416,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(44416,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(44416,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(44416,36,'cp 24','b1c3 f8g7 e2e3 e8g8 d2d4 c5d4 e3d4 e7e6 g1e2 d7d5 b2b3 b8c6 e1g1 c8d7 a1b1 a8c8 c3b5 f6e8 c1f4 b7b6 c4d5 e6d5 b1c1 d7f5 d1d2 d8d7 h2h4 a7a6 b5c3 e8f6 h4h5 f8e8 f4g5 f6e4 c3e4 f5e4 f2f3 e4f5 g3g4 f5e6 h5h6 g7f8 g5f6 f8e7 f6e7 d7e7 e2c3');
INSERT INTO evaluations VALUES(44413,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44413,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44413,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(44413,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(44413,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44413,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(44413,36,'cp 22','f8g7 e2e3 e7e6 g1e2 e8g8 d2d4 c5d4 e3d4 d7d5 b2b3 b8c6 e1g1 c8d7 a1b1 a7a6 c4c5 f8e8 b3b4 e6e5 d4e5 e8e5 c1f4 e5e8 f4g5 d7f5 b1b3 d5d4 c3d5 f5g4 f1e1 e8e5 g5f6 g7f6 d5b6 a8b8 b6c4 e5e8 c4d6 e8e5 d1d2 g4e6 d6b7');
INSERT INTO evaluations VALUES(55719,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(55719,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(55719,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(55719,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(55719,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(55719,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(55719,36,'cp 50','d6d5 c1f4 f8d6 f4d6 d8d6 e2e3 g8f6 a1c1 d6e7 c4d5 e6d5 f1d3 e8g8 g1e2 f8e8 d1d2 b7b6 e1g1 c8a6 e2f4 c7c6 c1c2 a6d3 d2d3 e7b7 f1c1 b8d7 h2h3 c6c5 d4c5 d7c5 d3b5 f6d7');
INSERT INTO evaluations VALUES(51800,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(51800,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(51800,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(51800,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(51800,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(51800,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(51800,36,'cp 61','d7d5 a2a3 d5e4 c3e4 b8d7 c1f4 g8f6 e4f6 d7f6 g1f3 c6c5 f4e5 c8d7 f1e2 d8a5 c2c3 d7a4 d1d2 c5d4 d2d4 f8e7 e1g1 e8g8 f1e1 a8d8 d4h4 f6d5 h4g4 d5f6 g4g3 a5b6 b2b4 e7d6 c3c4 d6e5 g3e5 f6d7 e5g3 d7f6 c4c5');
INSERT INTO evaluations VALUES(52466,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(52466,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(52466,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(52466,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(52466,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(52466,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(52466,36,'cp 4','g1f3 d7d5 e2e3 e7e6 c1b2 b8c6 f1e2 g8f6 e1g1 f8e7 d2d4 c5d4 f3d4 e8g8 c2c4 c6d4 b2d4 d5c4 e2c4 c8d7 b1d2 d7c6 d2f3 e7d6 d1e2 d8e7 f3e5 c6e4 f2f3 e4d5 c4b5 a7a6 b5d3 f8d8 e3e4 d6c5 d4c5 e7c5 e2f2');
INSERT INTO evaluations VALUES(52464,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52464,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52464,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(52464,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(52464,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52464,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(52464,36,'cp 5','g8f6 c1b2 b8c6 e2e3 g7g6 d2d4 c5d4 f3d4 f8g7 d4c6 b7c6 f1e2 d8a5 b1d2 e8g8 c2c4 f8e8 a2a3 c6c5 e1g1 a8b8 b3b4 c5b4 a3b4 a5b4 b2d4 c8b7 d1c2 b7c6 a1a7 b8a8 f1b1 b4d6 c4c5 d6e6 b1a1 f6d5 a7a8 e8a8 a1a8 c6a8 d4g7 g8g7 e2f3');
INSERT INTO evaluations VALUES(44581,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(44581,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(44581,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44581,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44581,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44581,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44581,36,'cp 0','c1b2 b8c6 e2e3 g7g6 d2d4 c5d4 f3d4 f8g7 d4c6 b7c6 f1e2 e8g8 c2c4 d8a5 b1d2 f8e8 a2a3 c6c5 e1g1 a8b8 d1c2 a5c7 b2c3 a7a5 e2f3 d7d6 f1e1 c8b7 f3b7 b8b7 h2h3 e8b8 a1b1 c7b6 g1h1 b6c7');
INSERT INTO evaluations VALUES(44582,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44582,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44582,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44582,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44582,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44582,36,'cp 0','e7e6 e2e3 f8e7 g2g4 f6g4 h1g1 e7f6 b1c3 g4h6 d2d4 d7d5 d4c5 b8d7 f1b5 e8g8 b5d7 c8d7 d1d2 h6f5 e1c1 d8e7 e3e4 d5e4 d2d7 e4f3 c3e4 f6b2 c1b2 e7h4 d7b7 a8b8 b7c6 b8c8 c6b7');
INSERT INTO evaluations VALUES(44255,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44255,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(44255,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44255,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44255,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(44255,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44255,36,'cp 8','e2e3 b8c6 c2c4 f8e7 b1c3 e8g8 d2d4 d7d5 c4d5 f6d5 d4c5 e7c5 f1c4 d8a5 a1c1 d5c3 b2c3 c5b4 d1c2 c8d7 e1g1 b4c3 c2c3 a5c3 c1c3 f8c8 f1d1 d7e8 a2a3 c6a5 d1c1 a5c4 c3c4 c8c4 c1c4 a8d8 f3d4 h7h6 c4c7');
INSERT INTO evaluations VALUES(44254,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44254,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44254,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44254,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(44254,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44254,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44254,36,'cp 3','b8c6 d2d4 c5d4 e3d4 b7b6 f1d3 c8b7 a2a3 d7d5 b1d2 f8d6 f3e5 e8g8 d1e2 a8c8 f2f4 f6e4 d2e4 d5e4 e2e4 d8h4 g2g3 c6e5 g3h4 b7e4 d3e4 e5g6 f4f5 e6f5 e4f5 c8e8 e1f2 g6h4 f5d3 h7h5 a1e1 e8e6 e1e6 f7e6 f2e2 h4g2 e2d1 f8f2');
INSERT INTO evaluations VALUES(30797,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(30797,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(30797,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30797,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(30797,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(30797,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(30797,36,'cp 35','c2c4 g8f6 b1c3 e7e6 e2e3 d7d5 g1f3 f8d6 f4d6 d8d6 d1a4 b7c6 a4c2 e8g8 c4d5 f6d5 f3g5 d5f6 f1d3 h7h6 d3h7 g8h8 g5e4 f6e4 h7e4 c6e4 c3e4 d6e7 a1d1 f8c8 e1g1 b8d7 e4c3 c7c6 c2e2 h8g8 e3e4 d7f6 e4e5');
INSERT INTO evaluations VALUES(30788,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(30788,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(30788,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(30788,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(30788,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(30788,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(30788,36,'cp 33','g8f6 e2e3 e7e6 b1c3 d7d5 c4d5 f6d5 c3d5 d8d5 a2a3 c7c5 a1c1 f8e7 g1f3 e8g8 f1c4 d5f5 d4c5 b6c5 e1g1 f8d8 d1e2 b8d7 f4c7 d8c8 c4d3 b7e4 d3e4 f5e4 c7a5 a8b8 a5c3 h7h6 f1d1 e4b7 f3d2 b7b5 e2b5 b8b5 f2f4');
INSERT INTO evaluations VALUES(47612,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(47612,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(47612,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(47612,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(47612,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(47612,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(47612,36,'cp 29','f8e7 h2h3 c8e6 c4b3 e6b3 a2b3 b8c6 d3d4 e8g8 c1e3 a7a5 e1g1 e5d4 f3d4 c6d4 e3d4 c7c6 d1f3 f8e8 f1e1 f6h7 a1d1 e7f8 d4e3 b7b5 e3f4 d8d7 f3d3 e8d8 c3e2 h7f6 e2g3 a5a4 c2c4 b5c4 b3c4 d7b7');
INSERT INTO evaluations VALUES(40681,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40681,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40681,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(40681,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40681,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(40681,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(40681,36,'cp 14','e1g1 e8g8 a2a4 c8e6 h2h3 d8d7 a4a5 c7c5 c1d2 b8c6 c3d5 e7d8 c4b3 e6d5 e4d5 c6b4 d2b4 c5b4 f1e1 b7b6 d3d4 e5e4 f3d2 d7e7 a5a6 b6b5 d2e4 f6e4 f2f3 e4c3 b2c3 e7c7 c3b4 c7c3 d1d3 c3d3 c2d3 d8f6 e1e4 a8c8 f3f4');
INSERT INTO evaluations VALUES(40680,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40680,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(40680,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(40680,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(40680,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(40680,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40680,36,'cp 18','e8g8 a2a4 c8e6 h2h3 d8d7 f1e1 e7d8 c1d2 c7c5 f3h2 b8c6 c3d5 a7a6 d5f6 d8f6 h2g4 f6h4 g4e3 h4g5 e3d5 g5d2 d1d2 a8e8 d5e3 c6d4 a4a5 d7d8 c2c3 d4c6 b2b4 c5b4 c3b4 c6d4 e1f1 f7f5 e4f5 d4f5 e3f5 f8f5');
INSERT INTO evaluations VALUES(36180,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36180,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36180,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36180,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(36180,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36180,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36180,36,'cp 35','h2h3 c8e6 f1e1 a7a6 a2a4 e6c4 d3c4 a6a5 b2b3 b8a6 c1a3 f6d7 d1d2 d7c5 a1d1 c7c6 c3e2 c5e4 d2e3 f7f5 f3e5 e7g5 f2f4 g5h4 g2g3 h4f6 g3g4 f6h4 e1f1 a6c5 g4f5 f8f5 a3c5 e4c5 e5g6 f5f7 g6h4 d8h4 d1d6');
INSERT INTO evaluations VALUES(36178,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36178,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36178,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(36178,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(36178,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36178,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(36178,36,'cp 13','a7a6 c1e3 c8e6 a2a3 b7b5 c4e6 f7e6 c3e2 c7c5 b2b4 b8c6 b4c5 d6c5 f3d2 f6h5 d2b3 c5c4 d3c4 b5c4 b3d2 h5f4 a1b1 c4c3 e2c3 e7a3 d2c4 d8g5 d1g4 g5g4 h3g4 a3b4 c3a4 f4e2 g1h2 f8b8 a4b6 a8a7 g4g5 h6g5');
INSERT INTO evaluations VALUES(35207,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(35207,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(35207,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(35207,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(35207,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(35207,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(35207,36,'cp 77','g2g3 h4e7 b1c3 e5f4 f1g2 f4g3 h2g3 c7c6 d2d4 d7d5 d1e2 h7h5 c1f4 g8f6 e1c1 c8g4 g2f3 b8d7 f3g4 f6g4 e2f3 g7g5 f4d2 d5e4 c3e4 e8c8 d2g5 f7f6 g5f4 e7e6 c1b1 e6d5 g1h3 f6f5 e4g5 d5f3 g5f3 f8h6 f4h6 h8h6 f3h4');
INSERT INTO evaluations VALUES(35206,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(35206,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(35206,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(35206,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(35206,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(35206,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(35206,36,'cp 48','h4e7 b1c3 e5f4 f1g2 f4g3 h2g3 c7c6 d2d4 d7d5 d1e2 h7h6 c1d2 g8f6 e1c1 c8g4 g2f3 g4e6 e4d5 f6d5 f3d5 c6d5 h1h5 b8c6 c3d5 e6d5 h5d5 e7e2 g1e2 c6e7 d5b5 e8c8 d1f1 a7a6 b5e5 f7f6 e5h5 d8e8 c2c3 e7g6 e2f4 g6f4 d2f4');
INSERT INTO evaluations VALUES(35288,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(35288,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(35288,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(35288,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(35288,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(35288,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(35288,36,'cp 97','b1c3 e5f4 f1g2 f4g3 h2g3 c7c6 d2d4 d7d5 d1e2 c8e6 g1h3 d5e4 c1g5 f7f6 g5f4 b8a6 c3e4 e8c8 e1c1 a6c7 e4c5 e6d5 e2g4 d5e6 c5e6 e7e6 g4e6 c7e6 d4d5 c6d5 g2d5 e6c5 b2b4 g8e7 d5c4 c5d7 c4e6 e7c6 b4b5 c6e5 f4e5 f6e5 h3g5 c8c7 g5f7');
INSERT INTO evaluations VALUES(57226,30,'cp -116',NULL);
INSERT INTO evaluations VALUES(57226,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(57226,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(57226,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(57226,34,'cp -110',NULL);
INSERT INTO evaluations VALUES(57226,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(57226,36,'cp -98','d7d5 f3g3 d8e7 f2f4 e5e4 b1c3 g8f6 b2b3 b8c6 f1b5 c8d7 c1b2 h7h5 g1e2 e8c8 h2h3 e7e8 b5c6 d7c6 e1g1 h5h4 g3f2 c8b8 a2a3 f8e7 e2d4 c6d7 b3b4 h8h5 f2e2 d7c8 d2d3 e4d3 c2d3 g7g5 e3e4 g5g4 h3g4 c8g4');
INSERT INTO evaluations VALUES(56482,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(56482,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(56482,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(56482,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(56482,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(56482,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(56482,36,'cp -67','f3g3 d8e7 f2f4 e5e4 b1c3 g8f6 b2b3 h7h5 c1b2 h5h4 g3f2 c8d7 h2h3 e7e6 e1c1 b8c6 g1e2 e8c8 g2g4 h4g3 e2g3 a7a6 f1e2 c8b8 h3h4 d7c8 c1b1 f8b4 a2a3 b4c5 h4h5 e6e7 b3b4 c5b6 f2g2 h8h6');
INSERT INTO evaluations VALUES(44933,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(44933,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44933,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(44933,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(44933,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44933,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(44933,36,'cp 36','e2e3 e7e6 c2c4 f8d6 f4d6 d8d6 b1c3 e8g8 a1c1 b8d7 c4d5 e6d5 f1d3 d6e7 d1c2 f8e8 e1g1 f6e4 h2h3 d7f6 f3e5 c8f5 c3e2 g7g6 g2g4 f5e6 f2f3 e4d6 b2b3 f6d7 f3f4 f7f6 e5f3 a8c8 f4f5');
INSERT INTO evaluations VALUES(23398,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(23398,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(23398,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(23398,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(23398,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(23398,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(23398,36,'cp 151','e4f5 d7d5 f1b5 c8f5 f3e5 f5d7 b5c6 d7c6 e1g1 g8f6 f1e1 c6d7 d2d4 e7e6 c1g5 f8e7 d1e2 e8g8 e5d7 d8d7 e2e6 d7e6 e1e6 a8e8 b1d2 h7h6 g5h4 g8f7 e6e1 e7d6 h4g3 d6g3 h2g3 g7g5 d2f3 e8e4 c2c3 f7g7 f3d2 e4e1 a1e1 f8e8 g1f1 a7a5 f2f3 e8e1 f1e1 a5a4 g3g4');
INSERT INTO evaluations VALUES(23389,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(23389,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(23389,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(23389,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(23389,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(23389,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(23389,36,'cp 145','d7d5 f1b5 c8f5 f3e5 f5d7 b5c6 d7c6 e1g1 g8f6 d2d4 c6d7 c1g5 e7e6 f1e1 f8d6 c2c4 d5c4 b1c3 e8g8 e5c4 h7h6 g5h4 d6e7 d1b3 a7a5 b3b7 d8b8 b7b3 b8b3 a2b3 e7b4 c4e5 d7e8 h4f6 f8f6 e5d3 b4d6 f2f3 e8g6 c3e4 g6e4 e1e4');
INSERT INTO evaluations VALUES(20879,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(20879,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(20879,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(20879,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(20879,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(20879,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(20879,36,'cp 122','f1b5 c8f5 f3e5 f5d7 b5c6 d7c6 d1e2 g8f6 d2d4 e7e6 b1c3 f8e7 c1g5 d8c8 h2h4 e8g8 h4h5 h7h6 e5g6 f8f7 g6e7 f7e7 g5f6 g7f6 e1c1 e6e5 e2d2 c8f8 h1h3 a8d8 h3f3 e7f7 d2e3 d8e8 d4e5 e8e5 f3g3 f7g7 g3g7 f8g7 e3a7 e5h5 f2f4 g7d7 b2b3 g8f7 a7f2 h5f5');
INSERT INTO evaluations VALUES(20872,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(20872,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(20872,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(20872,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(20872,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(20872,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(20872,36,'cp 134','c8f5 f3e5 f5d7 b5c6 d7c6 d1e2 g8f6 d2d4 e7e6 b1c3 f8b4 e5c6 b7c6 e2e6 d8e7 e6e7 e8e7 c1d2 e7d7 e1c1 b4c3 d2c3 f6e4 c3e1 a8e8 f2f3 e4d6 e1g3 e8e2 d1d2 e2e7 b2b3 h8e8 g3e5 c6c5 h1d1 c7c6 c2c4 d6f5 c4d5 c6d5 d2c2 c5c4 b3c4');
INSERT INTO evaluations VALUES(26476,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(26476,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(26476,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(26476,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(26476,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(26476,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(26476,36,'cp 138','f3e5 f5d7 b5c6 d7c6 d2d4 g8f6 d1e2 g7g6 e5c6 b7c6 c1f4 f8g7 f4e5 d8b8 b2b3 e8g8 e1g1 c6c5 d4c5 b8b4 b1d2 b4c5 c2c4 c7c6 a1c1 g7h6 c1c2 f6d7 d2f3 d7e5 f3e5 c5d6 c4d5 c6d5 c2c6 d6b4 f1e1 h6g7 e5d7 f8f7 c6c7 g8h8');
INSERT INTO evaluations VALUES(55878,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(55878,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(55878,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(55878,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(55878,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(55878,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(55878,36,'cp -99','d5e4 f3e5 g8e7 d1h5 g7g6 h5h4 f8g7 h4e4 b8d7 f1e2 e7f5 c2c3 c7c5 e1g1 c5d4 c3d4 e8g8 f1d1 d7f6 e4d3 b7b6 b1c3 c8b7 g2g4 f5e7 c1f4 f6d5 f4g3 f7f6 c3d5 b7d5 e5c4 f6f5 e2f3 f5f4 g3h4 d5f3 d3f3 d8d5 f3d5 e7d5');
INSERT INTO evaluations VALUES(56054,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(56054,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(56054,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(56054,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(56054,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(56054,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(56054,36,'cp -101','f3e5 g8f6 b1c3 f8b4 f1c4 c7c5 c1e3 d8a5 e1g1 e8g8 d4c5 b4c5 e5g4 b8d7 g4f6 d7f6 e3c5 a5c5 d1e2 e4e3 f2e3 c8d7 a1d1 d7c6 d1d4 a7a6 c4b3 h7h6 h2h3 b7b5 a2a4 b5b4 c3d1 a6a5 d1f2 c5g5 f2d3 a8c8 e3e4 f8d8 d4d8 c8d8 e2f2');
INSERT INTO evaluations VALUES(55976,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(55976,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(55976,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(55976,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(55976,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(55976,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(55976,36,'cp -125','g8e7 b1c3 e7f5 d4d5 f8d6 f1b5 c7c6 d5c6 e8g8 c6b7 c8b7 e5c4 d6c7 d1d8 f8d8 b5a4 b8a6 c3b5 c7b6 a4b3 a6c5 c1g5 d8d7 b5c3 e4e3 g5e3 b7g2 h1g1 f5e3 f2e3 g2c6 c4b6 a7b6 b3c4 a8d8 a1d1 d7d1 c3d1 c6e4 c4b3 c5b3 c2b3 e4b7 d1c3 f7f5 a2a4 g8f7');
INSERT INTO evaluations VALUES(1913,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(1913,31,'cp 285',NULL);
INSERT INTO evaluations VALUES(1913,32,'cp 267',NULL);
INSERT INTO evaluations VALUES(1913,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(1913,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(1913,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(1913,36,'cp 294','b1c3 f8b4 a1b1 b2a3 c3d5 b4d2 d1d2 e8d8 f3g5 c6e5 b1b3 a3f8 d5c7 d8c7 d2a5 c7b8 a5e5 d7d6 e5f4 g8f6 g2g3 f8e7 f1g2 h8d8 e1g1 b7b6 g5f3 c8b7 f4g5 d8g8 f1d1 h7h6 g5f4 f6e4 f3d4 g7g5 f4g4 g8c8 d4f5 e7e5 f5h6 e4f6 g4h3 b7g2 h3g2 e5e7');
INSERT INTO evaluations VALUES(1912,30,'cp 267',NULL);
INSERT INTO evaluations VALUES(1912,31,'cp 282',NULL);
INSERT INTO evaluations VALUES(1912,32,'cp 289',NULL);
INSERT INTO evaluations VALUES(1912,33,'cp 268',NULL);
INSERT INTO evaluations VALUES(1912,34,'cp 292',NULL);
INSERT INTO evaluations VALUES(1912,35,'cp 316',NULL);
INSERT INTO evaluations VALUES(1912,36,'cp 292','f8b4 a1b1 b2a3 c3d5 b4d2 d1d2 e8d8 f3g5 g8h6 f2f4 a3a2 d5c3 a2a3 e2e3 h8e8 f1c4 a3e7 c4b3 c6e5 c3d5 e7c5 f4e5 e8e5 g5f3 e5e8 f3d4 c7c6 d5c3 c5g5 c3d1 g5f6 d4f3 b7b6 e1g1 c8a6 f1e1 d8c7 e3e4 f6d6 d2g5 d6c5 g5c5 b6c5 b1a1 c5c4 a1a6 c4b3 c2b3 f7f6 f3d4 h6f7 a6a5 f7g5 d1c3 g5e6 d4e6 e8e6');
INSERT INTO evaluations VALUES(330,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(330,31,'cp 300',NULL);
INSERT INTO evaluations VALUES(330,32,'cp 297',NULL);
INSERT INTO evaluations VALUES(330,33,'cp 290',NULL);
INSERT INTO evaluations VALUES(330,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(330,35,'cp 312',NULL);
INSERT INTO evaluations VALUES(330,36,'cp 310','a1b1 b2a3 c3d5 b4d2 d1d2 e8d8 f3g5 g8h6 f2f4 h8e8 e2e3 a3a2 d5c3 a2a3 f1c4 a3e7 g5e4 c6e5 f4e5 e7e5 b1b5 e5e7 b5h5 c7c6 c4e2 f7f5 e4d6 e8f8 e3e4 d8c7 d6f5 h6f5 e4f5 g7g6 f5g6 h7g6 h5g5 d7d6 h2h4 c8f5 e1g1 a8e8 h4h5 e7e6 e2f3 e6e3 d2e3 e8e3');
INSERT INTO evaluations VALUES(329,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(329,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(329,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(329,33,'cp 309',NULL);
INSERT INTO evaluations VALUES(329,34,'cp 265',NULL);
INSERT INTO evaluations VALUES(329,35,'cp 310',NULL);
INSERT INTO evaluations VALUES(329,36,'cp 302','b2a3 c3d5 b4d2 d1d2 e8d8 f3g5 g8h6 f2f4 h8e8 e2e3 a3a2 d5c3 a2a3 g5e4 f7f6 e5f6 a3f8 f6g7 f8g7 h2h3 a7a5 g2g4 a5a4 f1g2 a4a3 e1g1 h6f7 c3d5 a3a2 b1a1 g7b2 e4f6 e8e6 g4g5 h7h6 h3h4 c6e7 e3e4 h6g5 h4g5 e7d5 f6d5 e6e8');
INSERT INTO evaluations VALUES(331,30,'cp 314',NULL);
INSERT INTO evaluations VALUES(331,31,'cp 310',NULL);
INSERT INTO evaluations VALUES(331,32,'cp 311',NULL);
INSERT INTO evaluations VALUES(331,33,'cp 295',NULL);
INSERT INTO evaluations VALUES(331,34,'cp 309',NULL);
INSERT INTO evaluations VALUES(331,35,'cp 312',NULL);
INSERT INTO evaluations VALUES(331,36,'cp 269','c3d5 b4d2 d1d2 e8d8 f3g5 g8h6 f2f4 a3a2 d5c3 a2a3 e2e3 h8e8 g5e4 f7f6 e5f6 a3f8 f6g7 f8g7 h2h3 a7a5 g2g4 a5a4 f1g2 a4a3 e1g1 a8a5 c3d5 b7b6 c2c4 c8a6 d2d3 a3a2 b1a1 h6f7 d5f6 e8e7 g4g5 h7h6 h3h4 h6g5 h4g5 f7e5 f4e5 c6e5 d3d4 a6c4 f1f2 e7e6 d4c3 e6f6 f2f6');
INSERT INTO evaluations VALUES(323,30,'cp 328',NULL);
INSERT INTO evaluations VALUES(323,31,'cp 298',NULL);
INSERT INTO evaluations VALUES(323,32,'cp 324',NULL);
INSERT INTO evaluations VALUES(323,33,'cp 304',NULL);
INSERT INTO evaluations VALUES(323,34,'cp 298',NULL);
INSERT INTO evaluations VALUES(323,35,'cp 306',NULL);
INSERT INTO evaluations VALUES(323,36,'cp 309','b4d2 d1d2 e8d8 f3g5 g8h6 f2f4 h8e8 e2e3 a3a2 d5c3 a2a3 g5e4 f7f6 e5f6 a3f8 f6g7 f8g7 h2h3 a7a5 g2g4 a5a4 f1g2 a8a5 e1g1 b7b6 c3d5 a4a3 f1d1 h6f7 b1a1 e8e6 f4f5 e6e4 g2e4 g7e5 e4g2 e5g3 d2f2 g3f2 g1f2 c6e5 d5f6 b6b5 g2d5 b5b4 f6h7 a5c5 d5b3 f7d6 d1d4');
INSERT INTO evaluations VALUES(34917,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(34917,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(34917,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(34917,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(34917,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(34917,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(34917,36,'cp 0','d2d4 f8b4 b1c3 g8f6 f1d3 d5e6 d1e2 e6e2 e1e2 f6h5 e2f2 e8g8 a2a3 b4d6 c3b5 b8c6 h1e1 c8g4 b2b4 a7a6 b5d6 c7d6 c2c4 g4f3 g2f3 c6d4 c1b2 d4e6 d3e4 a6a5 b2c3 a8b8 b4a5 f8c8 e4d5 e6c7 d5e4');
INSERT INTO evaluations VALUES(34897,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34897,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34897,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34897,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34897,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34897,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34897,36,'cp 0','f8b4 c2c3 b4d6 c3c4 d5e6 f1e2 c7c6 e1g1 g8f6 b1c3 e8g8 e2d3 e6g4 c1d2 g4h5 d1b3 b8d7 a1e1 f8d8 d3e2 d6c7 c4c5 h5g6 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(19940,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(19940,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(19940,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(19940,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(19940,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(19940,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(19940,36,'cp 39','g1f3 e7e5 d2d3 h7h6 e1g1 f8e7 c3d5 g8f6 c2c3 e8g8 c1e3 f6d5 c4d5 c8e6 d5e6 f7e6 d3d4 e5d4 c3d4 c5d4 f3d4 c6d4 e3d4 e7f6 d4f6 d8f6 d1d6 a8d8 d6b4 b7b6 f2f3 d8d4 b4c3 f8d8 a1c1 d8d7 a2a3 d4d2 c3f6 g7f6 b2b4 d7d3');
INSERT INTO evaluations VALUES(6983,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(6983,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(6983,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(6983,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(6983,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(6983,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(6983,36,'cp 46','g8f6 e1g1 e8g8 b1c3 e7b4 c1g5 h7h6 g5f6 d8f6 c3d5 f6d4 d1d4 c6d4 d5b4 b7b5 c4d3 a7a5 b4d5 c7c6 d5e3 f8d8 a2a4 b5b4 f1d1 g8f8 d3f1 d4e6 e3c4 e6c5 c4d6 b4b3 c2c3 c8a6 f1a6 a8a6 f2f3 a6a8 g2g3 a8b8 e4e5 f7f6 e5f6 g7f6 d6f5 d7d5 f5h6');
INSERT INTO evaluations VALUES(10939,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(10939,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(10939,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(10939,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(10939,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(10939,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(10939,36,'cp 37','e1g1 e8g8 b1c3 e7b4 c3d5 f6d5 c4d5 b4c5 c1e3 c6d4 e3d4 d8e7 d1d3 c5d4 d3d4 d7d6 f1e1 c8e6 a1d1 f8e8 b2b3 h7h6 a2a4 a7a6 d4b4 a8b8 d1d3 e6d5 d3d5 e7f6 d5f5 f6e6 e1e3 e8e7 f2f3');
INSERT INTO evaluations VALUES(16015,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(16015,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(16015,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(16015,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(16015,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(16015,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(16015,36,'cp 61','e7e5 f1c4 f8e7 e1g1 c7c6 a2a4 e8g8 f1e1 h7h6 h2h3 f8e8 a4a5 d8c7 c4f1 e5d4 f3d4 e7f8 c1f4 a8b8 b2b4 a7a6 d1d2 c6c5 b4c5 d7c5 f2f3 b7b5 a5b6 c7b6 f4e3 b6b7 f1c4 b7c7 c4a2 c8e6 d4e6');
INSERT INTO evaluations VALUES(49258,30,'cp 277',NULL);
INSERT INTO evaluations VALUES(49258,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(49258,32,'cp 256',NULL);
INSERT INTO evaluations VALUES(49258,33,'cp 243',NULL);
INSERT INTO evaluations VALUES(49258,34,'cp 245',NULL);
INSERT INTO evaluations VALUES(49258,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(49258,36,'cp 263','f3e5 d7d5 e4d5 f8d6 d2d4 e8g8 b1c3 f6d7 c1f4 f8e8 e1g1 d7e5 f4e5 d6e5 d4e5 e8e5 d1d4 b8c6 d4f4 e5f5 f4d2 c6e5 c4b3 a7a5 f2f4 e5d7 a1e1 d7c5 b3c4 c8d7 e1e5 d8f8 d2d4 f5f6 b2b3 d7f5 f1e1 c5d7 c3b5 f5c2 e5e7 a8d8 b5c7 d7b6 c4b5 c2g6');
INSERT INTO evaluations VALUES(49246,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(49246,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(49246,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(49246,33,'cp 274',NULL);
INSERT INTO evaluations VALUES(49246,34,'cp 272',NULL);
INSERT INTO evaluations VALUES(49246,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(49246,36,'cp 259','d7d5 e4d5 f8d6 d2d4 e8g8 b1c3 f6d7 c1f4 f8e8 e1g1 d7e5 f4e5 d6e5 d4e5 e8e5 d1d4 b8c6 d4f4 e5f5 f4d2 c6e5 c4b3 a7a5 f2f4 e5d7 a1e1 d7f6 a2a4 c8d7 h2h3 f5h5 f1f3 c7c6 d5d6 d8b6 g1h1 b6b4 e1e7 a8f8 c3e4 b4d2 e4d2 b7b5 d2e4 f6e4 e7d7 h5f5 a4b5 f5b5 h1g1 e4f6 d7e7');
INSERT INTO evaluations VALUES(47722,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(47722,31,'cp 277',NULL);
INSERT INTO evaluations VALUES(47722,32,'cp 267',NULL);
INSERT INTO evaluations VALUES(47722,33,'cp 260',NULL);
INSERT INTO evaluations VALUES(47722,34,'cp 274',NULL);
INSERT INTO evaluations VALUES(47722,35,'cp 276',NULL);
INSERT INTO evaluations VALUES(47722,36,'cp 289','e4d5 f8d6 d2d4 e8g8 b1c3 f6d7 f2f4 d6e5 f4e5 d7e5 c4e2 e5g6 e1g1 c7c6 d5c6 b8c6 c1e3 c6e7 d1d2 e7f5 e3f2 f5d6 b2b3 c8d7 e2d3 a8c8 c3e4 d6e4 d3e4 d7c6 e4c6 b7c6 a1e1 d8g5 d2g5 h6g5 c2c4 c8e8 e1e8 f8e8 f1e1 e8e1 f2e1 g6f4 g1f1 f7f6 e1f2');
INSERT INTO evaluations VALUES(47704,30,'cp 283',NULL);
INSERT INTO evaluations VALUES(47704,31,'cp 299',NULL);
INSERT INTO evaluations VALUES(47704,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(47704,33,'cp 274',NULL);
INSERT INTO evaluations VALUES(47704,34,'cp 268',NULL);
INSERT INTO evaluations VALUES(47704,35,'cp 281',NULL);
INSERT INTO evaluations VALUES(47704,36,'cp 261','f8d6 d2d4 e8g8 b1c3 f6d7 f2f4 d6e5 f4e5 d7e5 c4e2 e5g6 e1g1 c7c6 d5c6 b8c6 c1e3 c6e7 d1d2 c8d7 e3f2 f8e8 a1e1 e7f5 e2d3 d8a5 b2b3 a8d8 e1e8 d8e8 a2a4 a7a6 f2e1 f5e3 f1f3 a5g5 d3e4 e3f5 e4f5 d7f5 d2g5 h6g5');
INSERT INTO evaluations VALUES(55266,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(55266,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55266,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(55266,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(55266,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(55266,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(55266,36,'cp 24','g1f3 c7c5 f3e5 b8d7 e5d7 c8d7 b1d2 e7e5 d2c4 f7f6 a2a4 d7e6 b2b3 g8e7 f1e2 e7c6 e1g1 d8d7 f2f4 f8e7 f4e5 f6e5 h2h3 e8c8 e2g4 e6g4 d1g4 d7g4 h3g4 d8f8 f1f5 g7g6 f5f3 c6b4 f3f8 e7f8 c1d2 b4c2 a1f1 c2e3 d2e3 d4e3 f1f7');
INSERT INTO evaluations VALUES(55264,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(55264,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(55264,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(55264,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(55264,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(55264,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(55264,36,'cp 22','c7c5 f3e5 b8d7 e5d7 c8d7 b1d2 e7e5 d2c4 f7f6 a2a4 d7e6 b2b3 f8d6 f1e2 d6c7 e1g1 g8e7 e2h5 e7g6 g2g3 e8g8 h5g4 d8d7 g4e6 d7e6 f2f4 e5f4 c1f4 c7f4 g3f4 f6f5 d1h5 b7b6 h5h3 a8e8 a1e1 g8h8 e4e5 h7h6 e1e2 h8h7 h3g2 e8e7');
INSERT INTO evaluations VALUES(53216,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(53216,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53216,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53216,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(53216,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53216,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(53216,36,'cp 15','f3e5 b8d7 e5d7 c8d7 b1d2 e7e5 d2c4 f7f6 a2a4 d7e6 b2b3 g8e7 f1e2 e7c6 e1g1 d8d7 f2f4 f8e7 a4a5 e5f4 c1f4 e8g8 d1e1 c6b4 e2d1 e6c4 b3c4 e7d6 f4d2 b4c6 e1f2 a8e8 f2f5 d7c7 f5h3 c6e5 a1b1 b7b6 a5b6 a7b6 g1h1 g8h8 d2f4 g7g6 f4g3');
INSERT INTO evaluations VALUES(16600,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(16600,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16600,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16600,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16600,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16600,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16600,36,'cp 0','g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a2a4 d8f6 a4a5 b6c7 a1e1 c7d6 b2c3 d6f4 d2d1 f4d6');
INSERT INTO evaluations VALUES(16601,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16601,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16601,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16601,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16601,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16601,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16601,36,'cp 0','e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a2a4 d8f6 a1e1 a7a5 g2g3 f6f3 d3e2 f3f6');
INSERT INTO evaluations VALUES(29525,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(29525,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(29525,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(29525,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(29525,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(29525,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(29525,36,'cp 35','d1b3 b8c6 b3b7 f5d7 b7b3 a8b8 b3d1 e7e5 c4d5 f6d5 d4e5 f8b4 b1d2 c6e5 g1f3 e5f3 d1f3 d5f6 a2a3 b4d2 c1d2 b8b2 d2c3 b2c2 c3f6 d8f6 f3f6 g7f6 f1d3 c2b2 e1g1 e8g8 f1c1 c7c6 h2h3 f8b8 d3e4 b8b6 c1c5 b2b5 a1c1 b5c5 c1c5 b6a6 c5c3 c6c5 e4d3 a6d6 g2g4 h7h6 g1g2 g8h8 d3c4 d7c6 f2f3 d6d2 g2g3');
INSERT INTO evaluations VALUES(39418,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(39418,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(39418,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(39418,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(39418,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(39418,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(39418,36,'cp -42','b1c3 g8f6 d1b3 b4c3 b3c3 e8g8 g1f3 d7d6 e1g1 b8d7 c4b3 d7c5 b3c2 b7b6 f1e1 c8g4 a1d1 d8e7 e4e5 d6e5 e1e5 c5e6 d1e1 g4f3 c3f3 a8e8 c2b3 e7b4 h2h3 c7c5 b2c3 b4h4 b3e6 f7e6 e5e6 e8e6 e1e6 f6h5 f3e3 h4a4 f2f3 a4a2 e6e7 a2d5 e7a7 h7h6 a7e7 f8f7 e7e8 g8h7 c3e5');
INSERT INTO evaluations VALUES(39412,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(39412,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(39412,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(39412,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(39412,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(39412,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(39412,36,'cp -62','g8f6 d1b3 b4c3 b3c3 e8g8 g1f3 d7d6 e1g1 b8d7 e4e5 d7e5 f3e5 d6e5 a1d1 d8e7 f1e1 c8e6 e1e5 f8e8 f2f4 e7f8 c4e6 e8e6 e5e6 f7e6 c3c7 f8f7 c7e5 h7h6 f4f5 a8e8 h2h3 a7a6 f5e6 e8e6 d1d8 e6e8 d8e8 f6e8 e5b8 f7d7 b8g3 d7b5 g3b8 g8f7 b2e5 b5c6');
INSERT INTO evaluations VALUES(42824,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(42824,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(42824,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(42824,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(42824,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(42824,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(42824,36,'cp -47','d1b3 b4c3 b3c3 e8g8 g1f3 b8c6 e1g1 d7d6 a1d1 c8e6 f1e1 e6c4 c3c4 f6d7 e4e5 d6e5 c4g4 d7f6 g4g3 d8c8 f3e5 c6e5 e1e5 f8d8 d1e1 h7h6 h2h3 c7c5 e5e7 c8f5 e7b7 d8d2 b2f6 f5f6 e1e4 a8f8 b7a7 f6b6 a7e7 d2a2 e7e8 f8e8 e4e8 g8h7');
INSERT INTO evaluations VALUES(41227,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(41227,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41227,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(41227,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(41227,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41227,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(41227,36,'cp 32','d2d4 e8g8 c4b3 c8g4 h2h3 g4f3 d1f3 b8c6 f3d1 a7a6 c1e3 e7e5 d4e5 d6e5 d1d8 a8d8 f2f3 c6d4 e1c1 d8d7 d1d2 f8d8 h1d1 d4b3 a2b3 d7d2 e3d2 g7f8 c3e2 f6h5 c2c3 h5g7 b3b4 f7f6 c1c2 g8f7 d1a1 g7e6 e2c1 d8d7 c1b3 c7c5 b4c5 f8c5 b3c5 e6c5 d2e3');
INSERT INTO evaluations VALUES(44517,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44517,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(44517,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44517,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44517,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44517,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44517,36,'cp 0','a7a6 d2d3 b8c6 c3d5 e7e6 d5f6 d8f6 a2a4 f6g6 g1f3 g6g2 h1g1 g2h3 g1g3 h3h5 g3g5 h5h3');
INSERT INTO evaluations VALUES(23135,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(23135,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(23135,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(23135,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(23135,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(23135,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(23135,36,'cp 139','c7c6 e1g1 g8f6 f1e1 d7d6 b1d2 c8g4 f2f3 g4c8 c4b3 f8e7 f3f4 e8g8 f4f5 b7b5 a2a4 c8b7 d2f3 b5a4 b3a4 c6c5 h2h3 a7a5 c1f4 b7a6 f3d2 h7h6 a4c6 a8b8 b2b3 a6b5 c6b5 b8b5 d2c4');
INSERT INTO evaluations VALUES(19107,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(19107,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(19107,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(19107,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(19107,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(19107,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(19107,36,'cp 141','d1h5 d8e7 e1g1 g7g6 h5a5 b7b6 a5e1 a7a5 a2a4 f8g7 b1d2 g8f6 d2f3 e8g8 f3d4 f8e8 e1d1 c8b7 c1g5 h7h6 g5h4 e7b4 d1c1 b4c5 d4b3 c5h5 h4g3 d7d5 e4d5 f6d5 d3d4 c6c5 c2c3 c5d4 b3d4 a8c8 c4b5 e8d8 c1d2 h5g5 d2g5 h6g5');
INSERT INTO evaluations VALUES(56805,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(56805,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(56805,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(56805,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(56805,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(56805,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(56805,36,'cp 68','f6e5 f3e5 g8f6 b1d2 f8d6 c2c3 c7c5 f1b5 c8d7 e5d7 b8d7 d1e2 d8e7 d4c5 d6c5 d2b3 e8g8 c1g5 c5b6 e1c1 a7a6 b5d3 e7d6 f2f3 h7h6 g5d2 e6e5 c1b1 a8e8 d3g6 e8e7 d2c1 d6c6 g6c2 e5e4 h1f1 f8e8 f3e4 f6e4 e2d3 d7f6 d3f3 a6a5 a2a3 a5a4 b3d4 b6d4 d1d4 b7b5');
INSERT INTO evaluations VALUES(56782,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(56782,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(56782,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(56782,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(56782,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(56782,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(56782,36,'cp 82','f3e5 g8f6 c1e3 f8d6 f2f4 e8g8 b1d2 b8d7 f1e2 c7c5 e1g1 c5d4 e3d4 d6c5 d2b3 c5b6 e2d3 d7e5 f4e5 b6d4 b3d4 f6e4 f1f8 d8f8 d3e4 d5e4 d1g4 f8b4 g4e4 b4b2 a1f1 b2b4 e4f4 b4e7 f4f3 a8b8 c2c3 c8d7 f3h3');
INSERT INTO evaluations VALUES(56774,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(56774,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(56774,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(56774,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(56774,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(56774,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(56774,36,'cp 69','g8f6 b1d2 f8d6 d2f3 c7c5 d4c5 d8a5 c1d2 a5c5 f1d3 b8d7 d2c3 f6e4 c3d4 c5c7 e5d7 c8d7 f3e5 e8g8 d3e4 d7e8 d1g4 d5e4 g4e4 f8f5 f2f4 e8a4 c2c3 d6e5 d4e5 c7b6 b2b3 a4c6 e4d4 b6d4 c3d4 a8c8 e1g1 f5f8 f1e1 c6d5 a1c1 c8c1');
INSERT INTO evaluations VALUES(49133,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(49133,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(49133,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(49133,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(49133,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(49133,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(49133,36,'cp 74','b1d2 f8d6 c2c3 c7c5 f1b5 c8d7 e5d7 b8d7 d1e2 d8e7 d4c5 d6c5 d2b3 e8g8 c1g5 c5b6 e1c1 e7d6 h1e1 h7h6 g5e3 b6e3 e2e3 e6e5 f2f3 f8e8 b3d2 a7a6 b5a4 b7b5 a4b3 g8h8 d2e4 d6c6 e4f6 d7f6 e3d3 a8d8 c1b1 e8e7 b3c2 c6b6 g2g4');
INSERT INTO evaluations VALUES(31311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31311,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(31311,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(31311,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(31311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31311,36,'cp -3','b8c6 f1b5 a7a6 b5c6 b7c6 c1g5 d8b8 e1g1 e7e6 c3a4 f8d6 c2c4 d5c4 a1c1 f5e4 c1c4 e4d5 c4c3 f7f6 g5h4 g8e7 h4g3 e7f5 g3d6 f5d6 d1e2 e8g8 a4c5 f8e8 f1e1 a6a5 h2h3 d6b5 c3e3 b5d6');
INSERT INTO evaluations VALUES(26284,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26284,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26284,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(26284,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(26284,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26284,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26284,36,'cp 0','f1b5 a7a6 b5c6 b7c6 f3e5 h7h5 e5c6 d8b6 c6e5 e7e6 c3e2 f7f6 e5d3 h5h4 h2h3 f8d6 c1f4 d6f4 d3f4 b6b2 e1g1 g8e7 f1e1 e8f7 f4d3 b2b5 a1b1 b5c6 e2c1 c6c3 c1e2');
INSERT INTO evaluations VALUES(26236,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26236,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26236,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26236,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26236,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26236,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26236,36,'cp 0','a7a6 b5c6 b7c6 f3e5 h7h5 e5c6 d8c7 c6b4 e7e6 b4d3 f8d6 c3e2 a8c8 c2c3 f7f6 c1e3 e8f7 d3c5 g8e7 h2h4 d6c5 d4c5 c7b7 d1b3 b7d7 b3d1');
INSERT INTO evaluations VALUES(25630,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25630,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25630,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25630,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25630,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25630,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25630,36,'cp 0','b5c6 b7c6 f3e5 h7h5 e5c6 d8c7 c6b4 e7e6 b4d3 f8d6 c3e2 a8b8 h2h3 g8f6 e1g1 e8g8 c2c3 f8c8 b2b3 f6e4 c1e3 d6h2 g1h1 h2d6');
INSERT INTO evaluations VALUES(35124,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(35124,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(35124,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(35124,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(35124,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(35124,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(35124,36,'cp 44','b8c6 c2c3 f8e7 f1e2 d7d6 b1a3 f6g6 a3c4 h7h5 c4e3 h5h4 e1g1 g8f6 g1h1 c8e6 c3c4 h4h3 g2g3 a7a5 f3h4 g6h7 e3d5 e7d8 c1e3 e6d7 d1b3 e8g8 b3b7 a8b8 d5f6 d8f6 b7c7 f8d8 e3b6 f6h4 g3h4 h7h4 f1g1 c6d4 b6d4 e5d4');
INSERT INTO evaluations VALUES(1871,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(1871,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(1871,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(1871,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(1871,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(1871,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(1871,36,'cp 42','a2a4 f6d8 f1e2 g8f6 e1g1 f8e7 c2c3 d7d5 e4d5 f6d5 f1e1 e8g8 e2f1 c8g4 b2b4 e7f6 b4b5 c6a5 b1d2 d5c3 d1c2 c3d5 f3e5 g4f5 c1a3 f8e8 a3b2 c7c6 e5f3 e8e1 a1e1 f6b2 c2b2 f5e6 d2e4 c6b5');
INSERT INTO evaluations VALUES(50840,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(50840,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(50840,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(50840,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(50840,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50840,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50840,36,'cp 26','c2c4 e7e6 g1f3 g8f6 b2b3 c7c5 c4d5 f6d5 c1b2 c5d4 f3d4 f8e7 b1d2 e8g8 f1d3 c8d7 e1g1 b8c6 d1b1 h7h6 d3h7 g8h8 h7e4 c6d4 b2d4 e7f6 b1b2 d7c6 a1c1 f6d4 b2d4 d5e7 d4b2 a8c8 e4f3 c6f3 d2f3 c8c1 f1c1');
INSERT INTO evaluations VALUES(50831,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(50831,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(50831,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(50831,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50831,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50831,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50831,36,'cp 13','e7e6 g1f3 g8f6 b2b3 c7c5 c4d5 f6d5 c1b2 c5d4 f3d4 f8e7 b1d2 e8g8 f1d3 b7b5 e1g1 c8b7 a2a4 b5a4 a1a4 b8d7 d1b1 d5f6 d3e2 a6a5 f1d1 e7b4 a4a2 b7d5 b1a1 d8e7 e2c4 f8c8 c4d5 f6d5 d2c4');
INSERT INTO evaluations VALUES(50948,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50948,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50948,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(50948,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50948,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50948,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(50948,36,'cp 12','g1f3 g8f6 b2b3 c7c5 f1e2 c5d4 e3d4 f8b4 c1d2 b4e7 e1g1 b8c6 d2e3 e8g8 b1c3 b7b6 c4d5 e6d5 a2a3 h7h6 f3e5 c8b7 e5g4 b7c8 b3b4 f6g4 e2g4 e7g5 f1e1 f8e8 g4f3 c8e6 a1c1 a8c8 e3g5 d8g5 b4b5 a6b5 c3b5 e6d7 e1e8 c8e8');
INSERT INTO evaluations VALUES(50949,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(50949,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(50949,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50949,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(50949,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50949,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50949,36,'cp 1','g8f6 b2b3 c7c5 f1e2 c5d4 e3d4 f8b4 c1d2 b4e7 d2e3 b7b6 e1g1 c8b7 b1c3 e8g8 f3e5 b8d7 e5d7 f6d7 a1c1 a8c8 c4d5 e6d5 e2g4 e7a3 c1c2 c8c6 c3e2 d7f6 c2c6 b7c6 e2g3 c6b5 g4e2 d8d7 e2b5 d7b5 g3f5 f8d8');
INSERT INTO evaluations VALUES(43647,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43647,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(43647,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(43647,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43647,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43647,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(43647,36,'cp 15','b1c3 b8d7 d1c2 d5c4 a2a4 c7c5 f1c4 d8c7 c2e2 f8d6 h2h3 e8g8 e1g1 c5d4 e3d4 b7b6 d4d5 e6d5 c3d5 f6d5 c4d5 c8b7 d5b7 c7b7 f1e1 d7f6 c1g5 f8e8 e2d3 d6e7 a1c1 h7h6 g5e3 a8c8 d3b3 e7c5 e3c5 e8e1 f3e1');
INSERT INTO evaluations VALUES(43645,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(43645,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43645,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(43645,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43645,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43645,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43645,36,'cp 0','b8d7 b2b3 b7b6 f1d3 c8b7 e1g1 f8d6 a2a4 d8e7 c1b2 e8g8 h2h3 f8d8 a4a5 d6b4 a5b6 c7b6 c3a2 b4d6 a2c3');
INSERT INTO evaluations VALUES(53878,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(53878,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(53878,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(53878,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(53878,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(53878,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(53878,36,'cp 82','c2c3 b8c6 e1g1 c8g4 h2h3 g4h5 d2d3 g8f6 b1d2 f8e7 f1e1 e8g8 a2a4 g8h8 d2f1 a7a6 f1g3 h5g6 f3h4 f6e8 h4f5 g6f5 g3f5 e7g5 c1g5 d8g5 f5e3 g7g6 d1b3 a8b8 c4f7 g5f6 f7e8 f6f2 g1h1 f8e8 e3g4 f2f4 b3d5 h7h5 g4h2 b8d8 h2f1 f4f2 f1e3');
INSERT INTO evaluations VALUES(53874,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(53874,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(53874,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(53874,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(53874,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(53874,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(53874,36,'cp 69','b8c6 e1g1 c8g4 h2h3 g4h5 d2d3 g8f6 b1d2 f8e7 f1e1 e8g8 a2a4 a8b8 d2f1 g8h8 f1g3 h5g6 c1d2 a7a6 f3h4 f6e8 h4f5 e7g5 d2g5 d8g5 f5e3 g5d8 c4b3 c6e7 d1f3 h7h6 a4a5 e8f6 e3c4 f6h7 h3h4 f7f5 g3f5 e7f5 e4f5 g6f5 f3g3');
INSERT INTO evaluations VALUES(20596,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(20596,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(20596,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(20596,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(20596,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(20596,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(20596,36,'cp 75','e1g1 c8g4 h2h3 g4h5 a2a4 g8f6 f1e1 f8e7 d2d3 e8g8 b1d2 g8h8 d2f1 a8b8 f1g3 h5g6 c1d2 a7a6 b2b4 c5b4 c3b4 b7b5 c4d5 d8d7 a4b5 a6b5 d5c6 d7c6 d1e2 c6d7 d3d4 e5d4 f3d4 d6d5 e4e5 f6e4 a1a6 e4g3 f2g3 f8c8 e1a1 h8g8');
INSERT INTO evaluations VALUES(20595,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(20595,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(20595,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(20595,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(20595,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(20595,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(20595,36,'cp 76','c8g4 h2h3 g4h5 d2d3 f8e7 b1d2 g8f6 f1e1 e8g8 a2a4 a8b8 d2f1 g8h8 f1g3 h5g6 c1d2 a7a6 b2b4 c5b4 c3b4 b7b5 c4d5 d8d7 a4b5 a6b5 d1b3 h7h6 a1a6 f6d5 e4d5 c6d4 f3d4 e5d4 e1c1 b8c8 a6c6 f8e8 d2f4 h8h7 g3e4');
INSERT INTO evaluations VALUES(50256,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(50256,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(50256,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(50256,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50256,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(50256,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50256,36,'cp 9','g1f3 f8g7 g2g3 b7b6 f1g2 e8g8 c2c4 c8b7 d2d4 c7c5 e1g1 d7d5 d4c5 d5c4 d1c1 c4b3 a2b3 d8c8 c5b6 a7b6 a1a8 b7a8 f1d1 c8c1 d1c1 f8d8 b1d2 b8d7 b2d4 a8b7 h2h4 f6d5 d4g7 g8g7 d2c4 d8c8 f3d4');
INSERT INTO evaluations VALUES(50252,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50252,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(50252,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(50252,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(50252,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(50252,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(50252,36,'cp 12','f8g7 g2g3 e8g8 f1g2 c7c6 e1g1 f8e8 c2c4 d7d5 d2d4 c8f5 b1c3 b8d7 e2e3 f6e4 a1c1 e4c3 b2c3 f5e4 d1e2 d5c4 b3c4 e7e5 e2b2 d8b6 f3e5 e4g2 g1g2 d7e5 d4e5 b6b2 c3b2 g7e5 b2e5 e8e5 f1d1 a8e8 d1d7 e8e7 c1d1 g8g7');
INSERT INTO evaluations VALUES(43277,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43277,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43277,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43277,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(43277,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43277,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43277,36,'cp 21','g2g3 e8g8 f1g2 b7b6 c2c4 c8b7 d2d4 e7e6 e1g1 d8e7 b1c3 d7d5 c4d5 e6d5 a1c1 b8d7 e2e3 f6e4 f3e1 a8c8 e1d3 f8d8 f1e1 c7c6 h2h4 d7f6 c3e2 c6c5 b2a3 h7h6 d4c5 b6c5 e2f4 g6g5 h4g5 h6g5 f4h5 f6h5 d1h5');
INSERT INTO evaluations VALUES(43279,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(43279,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(43279,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(43279,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43279,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43279,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(43279,36,'cp 12','b7b6 c2c4 e8g8 d2d4 e7e6 f1g2 c8b7 e1g1 d8e7 b1c3 d7d5 c4d5 e6d5 e2e3 f6e4 a1c1 b8a6 a2a3 c7c5 d4c5 a6c5 b3b4 g7c3 b2c3 e4c3 c1c3 c5e4 c3c1 a8c8 c1c8 f8c8 d1a1 e7f6 a1f6 e4f6 f3d4 f6e4 h2h4');
INSERT INTO evaluations VALUES(56517,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(56517,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(56517,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(56517,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(56517,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(56517,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(56517,36,'cp 117','f1c4 g8h6 d2d4 b8d7 e1g1 d7e5 d4e5 d8d1 f1d1 h6g4 b1c3 g4e5 c3b5 f8d6 c4d5 c8g4 b5d6 c7d6 d1d4 e8c8 d5e4 h8e8 c1e3 h7h6 d4b4 d6d5 e4d3 d5d4 b4d4 d8d4 e3d4 e5d3 c2d3 g7g5 d4a7 e8e2 b2b4 e2d2 d3d4 b7b5 a7c5 g4e6 a2a3 c8b7 h2h3 e6c4 a1e1');
INSERT INTO evaluations VALUES(56511,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(56511,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(56511,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(56511,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(56511,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(56511,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(56511,36,'cp 102','g8h6 d2d4 b8d7 e1g1 d7e5 d4e5 d8d1 f1d1 f8c5 b1c3 e4e3 c1e3 c5e3 f2e3 c8d7 c3b5 e8c8 b5a7 c8b8 a7b5 h6g4 c4e2 g4e3 d1d3 d7b5 d3e3 b5e2 e3e2 h8e8 a1f1 e8e7 g1f2 b7b5 f2g3 b8b7 h2h4 h7h6 a2a3 d8e8 f1e1 b7c6 e2e3 e8d8 g3f4 d8d2 e1e2 d2d1 e5e6 f7e6 e3e6 e7e6 e2e6 c6d7');
INSERT INTO evaluations VALUES(48957,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(48957,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(48957,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(48957,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(48957,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(48957,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(48957,36,'cp 97','d2d4 b8d7 e1g1 d7e5 d4e5 d8d1 f1d1 h6g4 b1c3 g4e5 c3b5 f8d6 c4d5 c8g4 b5d6 c7d6 d1d4 e8c8 d5e4 g4e6 c1f4 d8d7 b2b3 h8d8 a1e1 g7g6 f2f3 d6d5 f4e5 d5e4 d4e4 d7d2 e1e2 b7b5 g1f2 a7a6 e5f4 d2d4 e4d4 d8d4 f4e3 d4d7 h2h3 d7d8 f2e1 d8d7 e3f4');
INSERT INTO evaluations VALUES(6441,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6441,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6441,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6441,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6441,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6441,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6441,36,'cp 0','g8f6 d2d3 d7d6 b1c3 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 e1g1 e8g8 c1g5 c5b6 g5h4 d4e6 d1f3 e6d4');
INSERT INTO evaluations VALUES(10144,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10144,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(10144,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10144,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(10144,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10144,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10144,36,'cp 0','b1c3 d7d6 d2d3 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 e1g1 c5b6 g1h1 e8g8 c1g5 h7h6 g5h4 d4e6 c3e2 g7g5 h4g3 e6f4 c4a2 d6d5 g3f4 g5f4 e4d5 c6d5 d3d4 d8d6 d4e5 d6e5 e2c3 d5d4 c3e2 f4f3 e2g3 f3g2 h1g2 g8h8 d1f3');
INSERT INTO evaluations VALUES(48680,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48680,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48680,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48680,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48680,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(48680,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48680,36,'cp 0','c1f4 c7c6 e2e3 c8f5 f1e2 d8b6 f3h4 f5e4 d1c1 h7h6 h4f3 e4h7 b1c3 b8d7 a2a3 f6h5 f4e5 h5f6 e5g3 f6h5');
INSERT INTO evaluations VALUES(48646,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48646,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48646,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48646,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48646,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48646,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48646,36,'cp 0','c7c6 e2e3 c8f5 f1e2 d8b6 f3h4 f5e4 d1c1 h7h6 h4f3 e4h7 a2a3 f6h5 f4e5 b8d7 b1c3 h5f6 e5g3 f6h5');
INSERT INTO evaluations VALUES(19918,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(19918,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(19918,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(19918,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(19918,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(19918,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19918,36,'cp 14','c5d4 f3d4 c8d7 c2c4 g8f6 b1c3 g7g6 h2h3 c6d4 b5d7 d8d7 d1d4 f8g7 c1e3 e8g8 d4d3 d7c6 e3d4 a7a6 c3d5 f6d5 e4d5 c6d7 d4g7 g8g7 e1g1 b7b5 f1e1 b5c4 d3d4 g7g8 a1c1 c4c3 c1c3 a8c8 c3c6 c8c6 d5c6 d7c6 e1e7 f8e8 e7e8 c6e8 d4d6 e8e1 g1h2');
INSERT INTO evaluations VALUES(19910,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(19910,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19910,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19910,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19910,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19910,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(19910,36,'cp 11','f3d4 c8d7 e1g1 g8f6 d4c6 b7c6 b5d3 g7g6 b1c3 f8g7 d1e1 e8g8 b2b3 e7e6 g1h1 d8c7 f2f3 c6c5 c1f4 f6h5 f4e3 d7c6 a1d1 h5f6 d3c4 f8d8 e3g5 h7h6 g5f4 g6g5 f4d2 d6d5 c4d3 d5e4 c3e4 f6e4 d3e4 c6e4');
INSERT INTO evaluations VALUES(6412,30,'cp -605',NULL);
INSERT INTO evaluations VALUES(6412,31,'cp -622',NULL);
INSERT INTO evaluations VALUES(6412,32,'cp -619',NULL);
INSERT INTO evaluations VALUES(6412,33,'cp -617',NULL);
INSERT INTO evaluations VALUES(6412,34,'cp -609',NULL);
INSERT INTO evaluations VALUES(6412,35,'cp -613',NULL);
INSERT INTO evaluations VALUES(6412,36,'cp -611','d8g5 e1g1 g5g6 b1c3 g8f6 c3d5 e8g8 c2c3 d7d6 d5f6 g7f6 b2b4 c5b6 a2a4 a7a5 b4b5 c6e7 d2d4 g8h8 f1e1 f8g8 g2g3 c8h3 d1d2 h3e6 c4d3 g6g7 g1h1 a8e8 d2h6 g7h6 c1h6 e5d4 h6d2 e7g6 f2f4 e6g4 h2h3 g4h3 h1h2 h3d7 c3d4 b6d4 a1c1 d4b6 d2e3 b6e3 e1e3 e8c8 d3e2 h8g7');
INSERT INTO evaluations VALUES(311,30,'cp -609',NULL);
INSERT INTO evaluations VALUES(311,31,'cp -607',NULL);
INSERT INTO evaluations VALUES(311,32,'cp -598',NULL);
INSERT INTO evaluations VALUES(311,33,'cp -602',NULL);
INSERT INTO evaluations VALUES(311,34,'cp -617',NULL);
INSERT INTO evaluations VALUES(311,35,'cp -609',NULL);
INSERT INTO evaluations VALUES(311,36,'cp -609','e1g1 g5g6 b1c3 g8f6 c3d5 e8g8 d5f6 g7f6 c4e2 d7d5 e2h5 g6e4 d2d3 e4h4 g2g3 h4h3 h5f3 c8e6 a2a3 h3f5 b2b4 c5d4 a1b1 d4c3 f3g2 c6d4 c1e3 c7c6 e3d4 c3d4 a3a4 f5g6 d1d2 d4b6 a4a5 b6c7 b1b3 f6f5 d3d4 e5e4');
INSERT INTO evaluations VALUES(310,30,'cp -621',NULL);
INSERT INTO evaluations VALUES(310,31,'cp -617',NULL);
INSERT INTO evaluations VALUES(310,32,'cp -621',NULL);
INSERT INTO evaluations VALUES(310,33,'cp -634',NULL);
INSERT INTO evaluations VALUES(310,34,'cp -625',NULL);
INSERT INTO evaluations VALUES(310,35,'cp -621',NULL);
INSERT INTO evaluations VALUES(310,36,'cp -605','g5g6 d2d3 g8f6 g1h1 e8g8 b1c3 d7d6 f2f3 c8d7 c3a4 c5b6 a4b6 a7b6 a2a4 g8h8 c2c3 h7h6 b2b4 f6h7 c1e3 f7f5 e3f2 d7e6 c4e6 g6e6 a4a5 c6e7 f2h4 e7g6 h4f2 f5e4 f3e4 h7f6 a5a6 a8a6 d1e2 a6a1 f1a1');
INSERT INTO evaluations VALUES(39224,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(39224,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(39224,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(39224,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(39224,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39224,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39224,36,'cp 0','c7c6 b1c3 g8f6 e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 e8g8 e1g1 c8g4 c1e3 b8c6 d1d3 g4e2 d3e2 c6d4 e3d4 b6d4 f1d1 d8f6 d1d2 a8d8 a1d1 d4b6 e2f3 f6f3 g2f3 d5d4 f3b7 f8e8 g1f1 g7g6 b7g2 b6c5 g2c6');
INSERT INTO evaluations VALUES(37660,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37660,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37660,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37660,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(37660,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(37660,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(37660,36,'cp 12','b1c3 g8f6 e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 c1e3 e8g8 g1e2 b8c6 e1g1 c8g4 d1d3 g4e2 d3e2 c6d4 e3d4 b6d4 a1d1 d8b6 b2b3 a8d8 e2g4 d4c5 d1d5 b6a5 g4c4 d8d5 c4d5 b7b6 a2a4 a5c3 f1d1 h7h6 d5f3 c3c2 d1d7 c2b1 g2f1');
INSERT INTO evaluations VALUES(37659,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(37659,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(37659,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(37659,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(37659,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(37659,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(37659,36,'cp 11','g8f6 e2e3 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 g1e2 e8g8 c1e3 b6a5 e3d2 b8c6 e1g1 c8g4 f1e1 f8e8 h2h3 a5d2 d1d2 g4e2 e1e2 e8e2 d2e2 c6d4 e2e5 d4c6 e5d5 d8b6 d5b3 a8d8 g2c6 b7c6 a1d1 d8d1 b3d1 g7g6 d1d2 c6c5 d2c3');
INSERT INTO evaluations VALUES(51941,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51941,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(51941,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(51941,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(51941,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(51941,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(51941,36,'cp 31','g1f3 g8f6 g2g3 f8e7 f1g2 d7d5 c4d5 e6d5 d2d4 c5c4 e1g1 e8g8 c1g5 c8e6 f3e5 b8c6 d1d2 f6e8 g5e7 c6e7 e2e4 d5e4 f1d1 d8c7 c3e4 f7f6 e5f3 a8d8 e4c3 e7d5 f3h4 d5c3 b2c3 e8d6 d1e1 e6f7 a2a4 b7b6 a1b1 g8h8');
INSERT INTO evaluations VALUES(51939,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(51939,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51939,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(51939,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(51939,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(51939,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(51939,36,'cp 25','b8c6 g2g3 d7d5 c4d5 e6d5 d2d4 g8f6 f1g2 c5d4 f3d4 f8c5 d4b3 c5b6 c3d5 c8e6 e2e4 e6g4 d1d3 f6d5 e4d5 c6e5 d3c3 d8e7 e1g1 a8c8 c3e1 e8g8 c1f4 e5f3 g2f3 e7e1 f1e1 g4f3 d5d6 f8d8 a1c1 c8c1 e1c1 f7f6 b3d2 f3c6 b2b3 g8f7 d2c4 d8d7 c4b6 a7b6 h2h3 h7h6');
INSERT INTO evaluations VALUES(18776,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18776,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(18776,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(18776,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(18776,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(18776,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(18776,36,'cp 30','g2g3 g8f6 f1g2 d7d5 c4d5 e6d5 d2d4 c5d4 f3d4 f8c5 d4b3 c5b6 c3d5 c8e6 e2e4 e6g4 d1d3 f6d5 e4d5 c6e5 d3c3 e8g8 e1g1 f8e8 c1e3 a8c8 e3b6 a7b6 c3d2 e5f3 g2f3 g4f3 d5d6 d8d7 b3d4 f3e4 f2f3 e4g6 d2b4 e8d8 a1d1 d7d6 b4d6 d8d6 d4b5 d6d1 f1d1');
INSERT INTO evaluations VALUES(18777,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(18777,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18777,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(18777,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(18777,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(18777,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(18777,36,'cp 30','g8f6 f1g2 d7d5 c4d5 e6d5 d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 f6e4 c5d4 a7a5 f2f3 e4d6 d1d3 d7f5 d3d2 d8d7 f1e1 a5a4 d4f2 d6c4 d2c3 f5h3 b2b4 a8c8 g2h3 d7h3 e2e4 h3d7');
INSERT INTO evaluations VALUES(12400,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(12400,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12400,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(12400,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(12400,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12400,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(12400,36,'cp 23','f1g2 d7d5 c4d5 e6d5 d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 f6e4 c5d4 a7a5 f2f3 e4d6 d1d2 a5a4 f1d1 d6c4 d2c3 c4d6 d4c5 d6b7 c5e3 b7d6 d1e1 a4a3 b2b3 d6b5 c3d2 d8f6 e3f2 a8d8 e1d1');
INSERT INTO evaluations VALUES(12401,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(12401,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(12401,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(12401,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(12401,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(12401,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(12401,36,'cp 31','d7d5 c4d5 e6d5 d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 f6e4 c5d4 a7a5 c1c2 h7h5 f2f3 e4d6 d4f2 a8b8 d1c1 a5a4 f1d1 d8f6 h2h3 h5h4 g3g4 d6c4 f2d4 f6d6 e2e4 c4e5 e4d5');
INSERT INTO evaluations VALUES(13541,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(13541,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(13541,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(13541,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(13541,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(13541,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(13541,36,'cp 31','c4d5 e6d5 d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 h7h5 f2f3 h5h4 g3g4 d8a5 a2a4 a8b8 d1c2 f6h7 f1d1 h7f8 c5f2 a5b4 f2a7 b8b7 a7d4 f8e6 d4f2 b4b2 f2h4 e6f4 c2b2');
INSERT INTO evaluations VALUES(13519,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(13519,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13519,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(13519,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(13519,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(13519,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(13519,36,'cp 31','e6d5 d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 f6e4 c5d4 a7a5 f1e1 h7h6 f2f3 e4g5 d1d2 a5a4 c1c2 a8b8 d4f2 g5e6 f3f4 e6c7 e2e4 d5e4 e1e4 e8e4 g2e4 c7d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(13767,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(13767,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(13767,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(13767,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(13767,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(13767,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(13767,36,'cp 24','d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5b6 a4b6 a7b6 d1c2 c8b7 b2b4 f8e8 c1e3 h7h5 a2a4 b7a6 f1d1 a8c8 a4a5 b6a5 g2f3 h5h4 c2f5 d8d6 a1a5 a6c4 d1b1 h4g3 h2g3 c8d8 g1g2 d6d7 f5d7 f6d7 a5a7 d7e5 e3d4 e5f3 e2f3 c4d3 b1g1');
INSERT INTO evaluations VALUES(13749,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(13749,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(13749,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13749,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(13749,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(13749,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(13749,36,'cp 15','c5d4 f3d4 f8c5 d4c6 b7c6 c3a4 c5b4 c1d2 d8a5 a4c3 a5a6 a2a3 b4e7 e1g1 e8g8 d2e3 c8g4 d1c2 h7h5 h2h3 g4e6 e3d4 f8d8 a1d1 a8b8 f1e1 a6c8 e2e4 d5e4 d4f6 d8d1 c2d1 e7f6 c3e4 f6e7 b2b4 h5h4');
INSERT INTO evaluations VALUES(13795,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(13795,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(13795,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(13795,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(13795,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13795,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13795,36,'cp 31','f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5b6 a4b6 a7b6 d1c2 c8b7 b2b4 f8e8 c1e3 f6e4 a2a4 c6c5 b4c5 b6c5 e3c5 a8c8 g2e4 e8e4 c2b2 b7a6 b2b6 a6e2 b6d8 c8d8 f1b1 d5d4 a4a5 e2a6 a1a4 h7h5 c5b6 d8d5 b1d1 f7f6 a4d4 e4d4 d1d4 d5d4 b6d4 g8f7 g1g2 f7e6 g2f3 e6d5 d4b6 d5c6 f3e3 a6c4');
INSERT INTO evaluations VALUES(13791,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(13791,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13791,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(13791,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(13791,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(13791,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(13791,36,'cp 35','f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5b6 a4b6 a7b6 d1c2 c8b7 b2b4 f8e8 c1e3 f6e4 f1d1 b6b5 g2e4 e8e4 c2e4 d5e4 d1d8 a8d8 e3b6 d8d7 g1f1 f7f6 f1e1 g8f7 a2a4 b5a4 a1a4 h7h5 a4a7 g7g5 b6c5 f7g6 h2h4 g6f5 c5e3 g5h4 g3h4 b7c8 a7a5 f5g6');
INSERT INTO evaluations VALUES(34437,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(34437,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(34437,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(34437,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(34437,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(34437,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(34437,36,'cp 21','c2c3 b8c6 a2a3 g8h6 f1d3 c5d4 c3d4 h6f5 d3f5 e6f5 b1c3 c8e6 b2b4 h7h6 h2h4 a8c8 e1g1 f8e7 d1d3 a7a6 h4h5 e8g8 c1e3 c6a7 f1c1 c8c6 c3a4 c6c1 a1c1 b6b5 d3d1 f8c8 a4c5 a7c6 c1b1 b7b6 c5d3 c6a7 a3a4 b5d7 d3f4 c8c4');
INSERT INTO evaluations VALUES(44540,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(44540,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(44540,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44540,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44540,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(44540,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(44540,36,'cp 4','c5d4 f3d4 e7e6 b1c3 f8b4 g2g3 e8g8 f1g2 d7d5 c4d5 f6d5 d1b3 d8b6 d4c2 b4c3 b2c3 b6b3 a2b3 d5c3 c2b4 a7a5 c1e3 f8d8 e3d2 c3b1 a1b1 a5b4 d2b4 b8c6 b4c3 e6e5 e1g1 c8e6 b1b2 f7f6 f2f4 e6d5 f4e5 d5g2 g1g2 f6e5 f1c1 d8c8 c1b1 b7b5 c3d2 c6d4');
INSERT INTO evaluations VALUES(44611,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44611,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(44611,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(44611,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(44611,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(44611,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(44611,36,'cp 16','f3d4 e7e6 b1c3 f8b4 g2g3 e8g8 f1g2 d7d5 c4d5 f6d5 d1b3 d8b6 d4c2 b4c3 b2c3 b6b3 a2b3 d5c3 c2b4 a7a5 c1e3 f8d8 e3d2 c3b5 b4d3 b5d4 a1a5 a8a5 d2a5 d8d7 b3b4 b8c6 g2c6 d4c6 f2f3 f7f6 e1f2 d7d5 h1c1 e6e5 d3c5 d5d2 c5e4');
INSERT INTO evaluations VALUES(44606,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(44606,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44606,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44606,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(44606,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44606,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44606,36,'cp 13','e7e6 e2e3 f8c5 f1e2 e8g8 e1g1 b8c6 a2a3 d7d5 d4c6 b7c6 b2b4 c5d6 c1b2 d8e7 c4d5 c6d5 b1c3 a7a6 c3a4 c8b7 a1c1 f8c8 c1c8 a8c8 e2d3 h7h6 d1d2 f6g4 g2g3 g4e5 b2e5 d6e5 a4c5 d5d4 c5b7 e7b7 e3d4 e5d4');
INSERT INTO evaluations VALUES(50328,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(50328,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(50328,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(50328,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(50328,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(50328,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(50328,36,'cp -79','d7d5 g1f3 c8f5 e2e3 e7e6 f1d3 f8e7 d3f5 e6f5 d1d3 d8c8 e1g1 e8g8 b2b3 f8e8 c2c4 f6e4 c1a3 e7a3 b1a3 b8d7 a1c1 d7f6 f3e5 c7c6 a3b1 a7a5 a2a4 c8e6 h2h3 e4g3 f1e1 f6e4 b1c3 f7f6 e5f3 e4c3 d3c3 g3e4');
INSERT INTO evaluations VALUES(21183,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(21183,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(21183,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(21183,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(21183,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(21183,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(21183,36,'cp 98','c2c4 e7e5 d4d5 c6b8 e2e4 g7g6 b1c3 b7b6 h2h3 f8h6 f1e2 h6c1 d1c1 b8d7 b2b4 a7a5 a2a3 g8e7 e1g1 e8g8 c1h6 f7f6 g1h2 g8h8 h6d2 e7g8 c3b5 f6f5 e4f5 g6f5 f3g5 d7f6 f2f4 d8e7 d2e3 a5b4 a3b4 a8a1 f1a1 h7h6');
INSERT INTO evaluations VALUES(21173,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(21173,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21173,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(21173,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(21173,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(21173,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(21173,36,'cp 91','g7g6 d4d5 c6b8 e2e4 e7e5 c4c5 d6c5 b1c3 b8d7 f1e2 g8f6 c1e3 f8e7 a2a4 e8g8 h2h3 a7a6 e1g1 f6e8 e3h6 e8d6 h6f8 e7f8 f1e1 f8h6 d1c2 b7b6 f3d2 d7f6 d2c4 d6c4 e2c4 f6e8');
INSERT INTO evaluations VALUES(38142,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(38142,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(38142,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(38142,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(38142,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(38142,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(38142,36,'cp -28','f1d3 b8c6 d1e2 d8c7 g1f3 f7f6 c1f4 f6e5 f4e5 c6e5 f3e5 g8f6 d3b5 e8d8 e1g1 a7a6 b5a4 c5d6 f1e1 f6e4 e5f3 d6c5 e1f1 c7f4 a4b3 h8f8 b1d2 e4d2 f3d2 a6a5 c2c4 a5a4 b3c2 b7b6 a1c1 a4a3 b2b3 d8c7 c2h7 c8b7 c4d5 e6d5 e2e7 c7b8 c1c5 b6c5 h7d3');
INSERT INTO evaluations VALUES(34823,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(34823,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(34823,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(34823,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(34823,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(34823,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(34823,36,'cp 64','d5d8 g1f3 g8f6 c1g5 f8e7 f1d3 e8g8 d1e2 b8c6 e1c1 c6b4 h1e1 b4d3 e2d3 b7b6 d4d5 c8b7 d3e3 e7c5 e3f4 d8d6 f4h4 e6d5 c3b5 d6d8 f3e5 h7h6 g5f6 d8f6 h4f6 g7f6 e5d3 a8c8 d3c5 b6c5 e1e7 f8e8 e7e8 c8e8 b5c7 e8e2 c7d5 e2f2 d5e3 b7g2');
INSERT INTO evaluations VALUES(34320,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(34320,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(34320,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34320,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(34320,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(34320,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(34320,36,'cp 60','g8f6 f1c4 c8f5 c1d2 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 b7b5 c4b3 d7b6 e2f3 a8d8 c3e2 f8e7 g2g4 b6c4 d2f4 c7d7 e2c3 f6d5 h4h5 g6h5 g4h5 d5f4 f3f4 c4d6 c1b1 d6f5 c3e4 a7a5 a2a4 e7d6 f4g4 d7c7 g4e2 b5a4 b3a4');
INSERT INTO evaluations VALUES(33587,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(33587,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(33587,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(33587,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(33587,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(33587,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(33587,36,'cp 68','f1d3 c8g4 h2h3 g4h5 c1d2 b8d7 g2g4 h5g6 d1e2 e7e6 d3g6 h7g6 e1c1 f8b4 c1b1 b4c3 d2c3 a5b5 e2b5 c6b5 d4d5 f6d5 c3g7 h8g8 g7d4 a7a6 h3h4 a8c8 d1e1 e8e7 h4h5 g6h5 g4h5 f7f6 h5h6 g8h8 h6h7 e7f7');
INSERT INTO evaluations VALUES(33590,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(33590,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(33590,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(33590,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(33590,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(33590,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(33590,36,'cp 61','c8g4 h2h3 g4h5 c1d2 e7e6 d1e2 f8b4 a2a3 h5g6 e1g1 g6d3 e2d3 b4c3 d2c3 a5c7 c3d2 b8d7 c2c4 e8g8 f1e1 f8d8 d3e2 h7h6 d2c3 d8e8 a3a4 a7a5 a1d1 e8d8 f3e5 d7e5 d4e5 d8d1 e1d1 f6d7');
INSERT INTO evaluations VALUES(32581,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(32581,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(32581,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32581,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(32581,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(32581,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(32581,36,'cp 61','h2h3 g4h5 c1d2 e7e6 d1e2 f8b4 a2a3 h5g6 e1g1 g6d3 e2d3 b4c3 d2c3 a5a6 d3d1 e8g8 c3b4 f8e8 a1c1 b8d7 c2c4 a8d8 d1c2 a6b6 b4c3 b6c7 f1e1 a7a5 c1d1 h7h6 b2b3 b7b5 a3a4 b5a4 b3a4 d8c8 c2d3 c6c5 d4d5 e6d5');
INSERT INTO evaluations VALUES(10392,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(10392,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(10392,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(10392,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(10392,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(10392,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(10392,36,'cp 40','c2c3 b4c5 b2b4 c5b6 e1g1 d7d6 c4b3 h7h6 b1d2 a7a6 a2a4 e8g8 d2c4 b6a7 c1e3 a7e3 c4e3 c6e7 a4a5 c8e6 f1e1 e7g6 d1c2 e6b3 c2b3 d8d7 a1d1 f8e8 h2h3 a8b8 b3c2 d6d5 e4d5 f6d5 d3d4 e5e4 f3e5 g6e5 d4e5 e8e5 d1d5 e5d5');
INSERT INTO evaluations VALUES(10390,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(10390,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(10390,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(10390,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(10390,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(10390,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(10390,36,'cp 34','b4c5 b2b4 c5b6 e1g1 d7d6 c4b3 e8g8 a2a4 a7a6 c1g5 c8e6 b1d2 h7h6 g5h4 e6b3 d1b3 g7g5 h4g3 d8e8 f1e1 e8e6 b3b2 a6a5 b4b5 c6e7 d2c4 e7g6 c4b6 c7b6 a1d1 a8c8 h2h3 f6h5 f3h2 f7f5 e4f5 f8f5');
INSERT INTO evaluations VALUES(10112,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(10112,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(10112,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(10112,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(10112,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(10112,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(10112,36,'cp 36','e1g1 d7d6 b2b4 c5b6 a2a4 a7a6 f1e1 e8g8 h2h3 c8e6 b1d2 h7h6 c4e6 f7e6 d1b3 d8e8 d2c4 b6a7 d3d4 e5d4 b4b5 a6b5 a4b5 d4c3 b5c6 a7f2 g1f2 a8a1 c6b7 e8b8 e4e5 f6d5 e5d6 c7d6 e1e6 a1c1 c4d6 c1h1 b3d5 h1c1 d6f7 f8f7');
INSERT INTO evaluations VALUES(10110,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(10110,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(10110,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10110,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(10110,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(10110,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(10110,36,'cp 52','d7d6 b2b4 c5b6 a2a4 a7a6 c4b3 c6e7 a4a5 b6a7 c1e3 a7e3 f2e3 e8g8 h2h3 a8b8 b1d2 b7b6 d3d4 e7g6 a5b6 c7b6 c3c4 d8e7 b3c2 b8a8 a1a3 c8b7 d4d5 a6a5 b4a5 a8a5 a3a5 b6a5 d1a1 e7c7 a1a3 h7h6 f1a1 f8a8 c4c5 d6c5 d2c4');
INSERT INTO evaluations VALUES(23418,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(23418,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(23418,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(23418,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23418,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(23418,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23418,36,'cp 0','d7d5 d2d4 d5e4 b1c3 a7a6 b5c6 b7c6 c3e4 g8f6 e4f6 e7f6 g1e2 f8e7 e1g1 e8g8 c2c4 c8e6 b2b3 a6a5 h2h3 a5a4 c1b2 d8d7 e2g3 a4b3 a2b3 a8a1 b2a1 f8b8 f1e1 e7d6 g3e4 d6f8 e4g3');
INSERT INTO evaluations VALUES(20941,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(20941,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(20941,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(20941,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(20941,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(20941,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(20941,36,'cp 0','d2d4 d5e4 b1c3 c8d7 d4d5 c6b4 b5d7 d8d7 a2a3 b4a6 d1d4 e7e6 d4e4 a6c5 d5e6 d7e6 e4e6 c5e6 g1f3 g8f6 c1e3 f6g4 e3d2 f8c5 h1f1 e8c8 h2h3 g4h6 e1c1 h6f5 c3e4 h8e8 f3g5 e6g5 e4c5 g5e6 c5e6 e8e6 f1e1 d8e8 d2b4');
INSERT INTO evaluations VALUES(20940,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20940,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20940,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(20940,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20940,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(20940,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(20940,36,'cp 0','d5e4 b1c3 a7a6 b5c6 b7c6 c3e4 g8f6 e4f6 e7f6 g1e2 f8d6 e1g1 e8g8 f1e1 f8e8 c1e3 c8e6 b2b3 d8d7 c2c4 d6a3 h2h3 a6a5 e2g3 a3b2 a1b1 b2c3 e1f1 a8d8 g3e2 c3b4 e2g3');
INSERT INTO evaluations VALUES(39160,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(39160,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(39160,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(39160,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(39160,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(39160,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(39160,36,'cp 125','c2c3 b4e7 f1d3 c7c5 g1f3 d7d5 e4e5 a7a6 e1g1 b8c6 c1e3 d8b6 d1d2 c8d7 d4c5 e7c5 e3c5 b6c5 d2g5 c5f8 b1d2 f7f6 g5e3 g8h6 h2h3 h6f7 e5f6 g7f6 c3c4 e8c8 c4d5 e6d5 f1c1 c8b8 d2b3');
INSERT INTO evaluations VALUES(39158,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(39158,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(39158,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(39158,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(39158,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(39158,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(39158,36,'cp 125','b4e7 f1d3 d7d6 g1f3 b8d7 e1g1 e6e5 a2a4 g8f6 a4a5 e8g8 d1c2 b7b6 a5b6 c7b6 h2h3 a7a5 c3c4 f8e8 b1c3 e7f8 f1e1 h7h6 c1e3 c8b7 e1d1 d8e7 d4d5 d7c5 e3c5 b6c5 a1a3 e7d8 d1a1 b7c8 c2d2 g7g6 c3b1 f6h5 a3a5 a8a5 a1a5 h5f4');
INSERT INTO evaluations VALUES(39498,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(39498,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(39498,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(39498,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(39498,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(39498,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(39498,36,'cp 112','g1f3 d7d5 e4e5 a7a6 f1d3 c7c5 e1g1 c8d7 a2a4 c5c4 d3c2 b8c6 b1d2 b7b5 b2b3 g8h6 h2h3 d8a5 c1b2 a5d8 f1e1 h6f5 b2c1 h7h6 b3c4 b5c4 d2f1 c6a5 g2g4 f5h4 f3h4 e7h4 c1a3 a5b3 c2b3 c4b3 d1b3 a8b8');
INSERT INTO evaluations VALUES(39497,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(39497,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(39497,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(39497,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(39497,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(39497,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(39497,36,'cp 125','c7c5 f1d3 d7d5 e4e5 b8c6 e1g1 c8d7 d4c5 e7c5 b2b4 c5b6 a2a4 g8e7 b1d2 h7h6 f1e1 e8g8 d2b3 e7g6 c1a3 f8e8 d3c2 c6e7 a4a5 b6c7 a3c1 e7f5 g2g3 b7b6 h2h4 a8c8 c2d3 b6a5 b3a5 c7b6 h4h5 g6f8 a5b7');
INSERT INTO evaluations VALUES(31911,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31911,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(31911,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31911,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31911,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31911,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(31911,36,'cp 13','e7e6 c2c4 c7c6 d1b3 d8b6 c1d2 b8d7 d2c3 f8e7 b1d2 g8f6 f1d3 b6c7 a1c1 e8g8 h2h3 g4h5 c3b4 e7b4 b3b4 h5g6 d3g6 h7g6 e1g1 a7a5 b4c3 c6c5 d4c5 c7c5 c3b3 a5a4 b3d1 c5b4 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(32344,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(32344,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(32344,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(32344,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32344,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(32344,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(32344,36,'cp 27','c2c4 c7c6 d1b3 d8b6 c1d2 b8d7 d2c3 g4f5 b1d2 h7h6 f1e2 f8e7 e1g1 g8f6 f1c1 e8g8 a2a3 f6e4 c3b4 c6c5 b4c5 e4c5 d4c5 d7c5 b3b6 a7b6 c4d5 e6d5 b2b4 c5e6 d2b3 f5h7 b3d4 f8c8 c1c8 a8c8 e2b5');
INSERT INTO evaluations VALUES(32337,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(32337,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32337,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(32337,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(32337,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(32337,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(32337,36,'cp 24','c7c6 h2h3 g4h5 g2g4 h5g6 d1b3 d8b6 f3e5 g6e4 f2f3 e4b1 a1b1 f8d6 f3f4 h7h5 g4g5 d6e5 d4e5 g8e7 c4d5 b6b3 a2b3 c6d5 c1d2 e8g8 f1d3 b8c6 e1e2 f8c8 h1c1 e7f5 b1a1 g7g6 b3b4 a7a6 c1c5 f5g3 e2f3 g3f5 d2c3 g8g7');
INSERT INTO evaluations VALUES(31464,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31464,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(31464,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31464,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(31464,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31464,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31464,36,'cp 28','h2h3 g4h5 g2g4 h5g6 d1b3 d8b6 f3e5 g6e4 f2f3 e4b1 a1b1 f8d6 f3f4 b6b3 a2b3 b8d7 c1d2 d7f6 f1d3 f6e4 h1g1 g8e7 e5f3 e4d2 e1d2 c6c5 d4c5 d6c5 d2e2 e8g8 h3h4 b7b6 h4h5 a8d8 g4g5 a7a5 g5g6 f7g6 h5g6');
INSERT INTO evaluations VALUES(21079,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(21079,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(21079,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(21079,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(21079,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(21079,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(21079,36,'cp 45','d2d4 c8f5 c2c4 e7e6 a2a3 g8f6 b2b4 f8e7 c1b2 f6e4 b1c3 e8g8 c4c5 e7f6 f1e2 c6b8 e1g1 c7c6 c3e4 f5e4 f3d2 e4g6 a1c1 b8d7 f2f4 h7h6 b2c3 a7a6 d2b3 f6e7 c3e1 d7f6 b3a5 d8c7 e1g3 f6e4 e2d3 f7f5 g1h1 g8h8');
INSERT INTO evaluations VALUES(11930,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11930,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11930,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11930,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(11930,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11930,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11930,36,'cp 0','d8e7 d1e2 f6d5 g2g3 g7g6 c2c4 c8a6 b2b3 f8g7 c1b2 d7d6 f1g2 e8c8 e2d2 d5b4 a2a3 b4d3 d2d3 g7e5 e1g1 e5b2 a1a2 b2g7 g2c6 d6d5 d3f3 d5c4 c6b5 a6b7 b5c6');
INSERT INTO evaluations VALUES(699,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(699,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(699,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(699,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(699,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(699,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(699,36,'cp 0','d1e2 f6d5 c2c4 d5b6 b1c3 c8a6 e2e4 e7e6 c4c5 a6f1 c5b6 f1c4 b6c7 d7d5 e5d6 f8d6 e4e6 c4e6 c1e3 f7f6 e1g1 e8f7 a1c1 h8c8 f1d1 d6c7 b2b3 h7h5 h2h3 g7g5 c3e4 e6d5 e4c3 d5e6');
INSERT INTO evaluations VALUES(698,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(698,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(698,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(698,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(698,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(698,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(698,36,'cp 0','f6d5 c2c4 d5b6 b1c3 e7e6 c1d2 d7d6 e5d6 b6c4 e2e6 c8e6 f1c4 e6c4 d6c7 f8d6 b2b3 c4e6 a1c1 d6c7 d2e3 e8g8 e1g1 f7f6 c3e2 f8c8 f1d1 a7a5 d1d4 e6d5 e2c3 d5e6 c3a4 e6d5 h2h3 g8f7 d4h4 f7g8 h4d4');
INSERT INTO evaluations VALUES(701,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(701,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(701,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(701,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(701,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(701,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(701,36,'cp 0','c2c4 d5b6 b1c3 c8a6 e2e4 e7e6 b2b3 e8c8 c1b2 f7f6 f2f4 f6e5 e1c1 a6b7 e4e5 f8d6 e5e6 d6f4 c1c2 d7e6 d1d8 h8d8 f1d3 c6c5 h1e1 b7g2 e1e6 g7g6 e6e7 d8d7 e7e8 d7d8');
INSERT INTO evaluations VALUES(700,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(700,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(700,36,'cp 0','d5b6 b1c3 c8a6 e2e4 e7e6 b2b3 e8c8 c1b2 f7f6 f2f4 f6e5 e1c1 a6b7 e4e5 f8d6 e5e6 d6f4 c1c2 d7e6 f1d3 c6c5 h1e1 b7g2 e1e6 h8f8 d3h7 d8d1 c3d1 f8h8 h7f5 c8b7 h2h4 h8h4 b2g7 h4h5 f5e4 g2e4 e6e4 f4d6 e4e2 d6f4');
INSERT INTO evaluations VALUES(694,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(694,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(694,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(694,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(694,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(694,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(694,36,'cp 0','b1c3 c8a6 e2e4 e7e6 b2b3 e8c8 c1b2 f7f6 f2f4 f6e5 e1c1 a6b7 e4e5 f8d6 e5e6 d6f4 c1b1 d7e6 d1e1 c6c5 c3e4 b7e4 e1e4 d8d1 b1c2 d1d2 c2b1');
INSERT INTO evaluations VALUES(20037,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20037,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(20037,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20037,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20037,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(20037,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20037,36,'cp 10','g7g6 f1b5 c8d7 d2d3 f8g7 e1g1 g8f6 b5c4 e8g8 h2h3 c6d4 f3d4 c5d4 c3d5 f6d5 c4d5 e7e6 d5b3 d8c7 c1d2 b7b5 a2a4 b5a4 b3a4 d7a4 a1a4 f8c8 a4c4 c7d7 f4f5 e6f5 e4f5 g6f5 b2b3 d6d5 c4c8 a8c8 g1h1 g8h8 d2g5 h7h6 g5h4 d7e6 f1e1');
INSERT INTO evaluations VALUES(19678,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(19678,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19678,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19678,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19678,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(19678,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(19678,36,'cp 15','f1b5 c8d7 e1g1 f8g7 d2d3 g8f6 b5c4 e8g8 a2a4 a7a6 c1e3 c6d4 h2h3 f6e8 d1d2 b7b5 a4b5 d4f3 f1f3 a6b5 a1a8 d8a8 c3b5 d7b5 c4b5 a8a1 d2c1 a1b2 c1b2 g7b2 b5c4 e8c7 f4f5 b2d4 e3f2 d4f2 g1f2 g6f5 f3f5 e7e6 f5g5 g8h8 g5h5');
INSERT INTO evaluations VALUES(19641,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19641,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19641,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(19641,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19641,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(19641,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(19641,36,'cp 5','c8d7 d2d3 f8g7 e1g1 g8f6 b5c4 e8g8 a2a4 a7a6 c1e3 e7e6 h2h3 c6e7 c4b3 d7c6 a4a5 b7b6 a5b6 d8b6 d1e1 b6c7 e1f2 a6a5 b3a4 f6d7 a4c6 e7c6 c3b5 c7b8 c2c4 c6b4 d3d4 c5d4 e3d4 g7d4 f3d4 d6d5 e4e5 d5c4 b5d6');
INSERT INTO evaluations VALUES(10520,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10520,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(10520,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10520,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(10520,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10520,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10520,36,'cp 0','b1d2 d7d5 e4d5 f6d5 e1g1 e8g8 a1c1 f8e8 f1e1 e8e1 d1e1 c6e7 d2e4 h7h6 f3e5 b7b6 a2a3 c8e6 e4g3 d8d6 g3e4');
INSERT INTO evaluations VALUES(10518,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10518,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10518,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(10518,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10518,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10518,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10518,36,'cp 5','d7d5 e4d5 f6d5 e1g1 e8g8 a1c1 d5f6 d2b3 d8d6 c4b5 c6e7 b3c5 b7b6 c5d3 c8e6 a2a3 c7c6 b5a4 a8c8 a4b3 e6b3 d1b3 c8c7 d3e5 d6d5 b3c2 f8c8 f1e1 c6c5 c2d3 c5d4 c1c7 c8c7 d3d4 e7g6 g2g3 d5d4 f3d4 g6e5 e1e5 g8f8');
INSERT INTO evaluations VALUES(9117,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9117,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(9117,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9117,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(9117,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9117,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(9117,36,'cp 8','e4d5 f6d5 e1g1 e8g8 a1c1 d5f6 f1e1 c6d4 d2e4 f6e4 e1e4 d4f3 d1f3 c8e6 c4e6 f7e6 f3e2 d8d7 e4e6 f8e8 e6e4 e8e4 e2e4 c7c6 c1e1 a8f8 b2b3 h7h6 h2h4 b7b6 e4e6 d7e6 e1e6 c6c5 e6d6 f8e8 g2g3 e8e2');
INSERT INTO evaluations VALUES(9101,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9101,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9101,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(9101,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(9101,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9101,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9101,36,'cp 0','f6d5 e1g1 e8g8 a1c1 f8e8 f1e1 e8e1 d1e1 c6e7 d2e4 h7h6 f3e5 b7b6 a2a3 c8e6 e4g3 d8d6 g3e4');
INSERT INTO evaluations VALUES(9176,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9176,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9176,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9176,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9176,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(9176,36,'cp 0','e1g1 e8g8 a1c1 f8e8 f1e1 e8e1 d1e1 c6e7 d2e4 h7h6 f3e5 b7b6 h2h3 c8e6 e4c3 c7c6 c3d5 e6d5 e1c3 a8c8 c3b3 e7f5 c4d5 c6d5 c1c8 d8c8 b3d5 c8c1 g1h2 c1f4 h2g1');
INSERT INTO evaluations VALUES(9175,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9175,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9175,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(9175,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(9175,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9175,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9175,36,'cp 0','e8g8 a1c1 f8e8 f1e1 e8e1 d1e1 c6e7 d2e4 h7h6 f3e5 b7b6 e4c3 c8e6 h2h3 c7c6 c3d5 e6d5 e1b4 a8c8 c4d5 c6d5 c1c8 d8c8 b4d6 c8c1 g1h2 c1f4 h2g1');
INSERT INTO evaluations VALUES(3751,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3751,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(3751,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(3751,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(3751,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3751,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3751,36,'cp 0','a1c1 f8e8 f1e1 e8e1 d1e1 c6e7 e1e5 c8e6 c4d5 e6d5 c1c7 e7g6 e5g3 g6f8 h2h3 f8e6 c7c3 d8b6 d2b3 h7h6 f3h4 e6d4 b3d4 b6d4 c3d3 d4e4 d3e3 e4b1 g1h2 b1b2 h4f5 d5e6 f5h6 g8h7 h6g4 h7g8 g3h4 e6g4 h3g4 b2d4 e3h3 d4f4 h2h1 f4c1 h1h2');
INSERT INTO evaluations VALUES(55850,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(55850,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55850,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55850,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55850,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55850,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55850,36,'cp 0','g8f6 g1f3 b7b6 e1g1 c8b7 b2b3 f8d6 c2c4 e8g8 c1b2 b8d7 d1e2 c7c5 c4d5 f6d5 b1d2 c5d4 f3d4 d7c5 d3c2 d8h4 h2h3 d5b4 c2b1 b4d5');
INSERT INTO evaluations VALUES(18829,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(18829,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(18829,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(18829,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(18829,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18829,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(18829,36,'cp 11','g2g3 f8e7 f1g2 d7d5 e1g1 g8f6 d1e2 e8g8 a2a4 a7a6 c1f4 b7b6 b1a3 c8b7 h2h4 c6a5 f1d1 f6h5 f4d2 h5f6 e4e5 f6d7 d2a5 b6a5 c2c4 a8b8 c4d5 b7d5 a3c4 f7f5 f3d2 f5f4 e2g4 f4g3 f2g3 g8h8 d1e1');
INSERT INTO evaluations VALUES(18831,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18831,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(18831,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(18831,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(18831,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18831,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18831,36,'cp 2','d7d5 d1e2 g8f6 f1g2 f8e7 e1g1 e8g8 a2a4 h7h6 c1f4 c8d7 f1e1 f8e8 c2c3 a7a6 b1d2 b7b5 a4b5 a6b5 f3e5 c6e5 f4e5 d7c6 h2h4 f6d7 e5f4 d5e4 g2e4 d8b6 e4c6 b6c6 e2e4 c6e4 d3e4 e8d8 e1d1 a8a1');
INSERT INTO evaluations VALUES(20298,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20298,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(20298,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(20298,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(20298,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(20298,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(20298,36,'cp 0','d1e2 g8f6 f1g2 f8e7 e1g1 e8g8 a2a4 h7h6 c1f4 c8d7 f1e1 f8e8 c2c3 a8c8 e4e5 f6h7 h2h4 h7f8 h4h5 d8b6 f4e3 b6c7 e3f4');
INSERT INTO evaluations VALUES(19065,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(19065,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19065,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(19065,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19065,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(19065,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19065,36,'cp 12','f8b4 c2c3 b4e7 f1c4 g8f6 c1g5 f6e4 c4f7 e8f7 d1d5 f7f8 g5e7 f8e7 d5e4 d7d5 e4d3 e7f7 e1g1 e5e4 d3e3 h8e8 f2f3 e4f3 e3f3 f7g8 f3f7 g8h8 b1d2 c6e5 f7f4 d8b6 g1h1 c8d7 a1e1 e5g6 f4d4 b6d4 e1e8 a8e8 b3d4 g6e5 d2f3 e5d3 b2b3 h8g8 h1g1 d7g4');
INSERT INTO evaluations VALUES(5211,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(5211,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(5211,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(5211,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(5211,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5211,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5211,36,'cp 4','c2c3 b4e7 f1c4 g8f6 c1g5 f6e4 c4f7 e8f7 d1d5 f7e8 g5e7 e8e7 d5e4 d7d5 e4d3 e5e4 d3e3 e7f7 e1g1 h8e8 f2f3 f7g8 f3e4 d5e4 b1d2 d8d5 e3f4 e4e3 a1e1 c8h3 g2h3 e3d2 f4d2 d5h5 d2d7 e8e1 f1e1 h7h6 b3d4 c6d4 c3d4 a8f8 d7b7 h5a5 b7b3 g8h8 e1f1 a5g5 b3g3 f8f1 g1f1 g5c1 f1g2 c1b2 g3f2 b2a3 f2f7 h8h7 g2h1 a3d3');
INSERT INTO evaluations VALUES(17485,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(17485,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(17485,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(17485,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(17485,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(17485,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(17485,36,'cp 46','d2d4 c5d4 f3d4 e7e6 e1g1 d8c7 d4c6 b7c6 d1e2 g8e7 b1c3 e7g6 f2f4 f8e7 e4e5 c8b7 c1d2 c6c5 c3e4 e8g8 b2b3 g6h4 d2c3 h4f5 a1d1 c7c6 c4d3 c5c4 b3c4 c6b6 g1h1 b6e3 d1e1 b7e4 e2e3 f5e3 e1e3 e4c6 e3e2 a6a5 h2h3 e7c5 h1h2 f8d8 g2g4 d7d5 e5d6 c5d6 c3e5');
INSERT INTO evaluations VALUES(17491,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(17491,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17491,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(17491,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17491,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(17491,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(17491,36,'cp 35','c5d4 f3d4 e7e6 e1g1 d8c7 d4c6 b7c6 d1e2 g8e7 b1c3 e7g6 f2f4 f8e7 e4e5 c8b7 c3e4 c6c5 c4d3 e8g8 b2b3 c7c6 c1d2 g6h4 d2c3 h4f5 h2h3 a6a5 a2a4 f7f6 e2f2 f6e5 c3e5 d7d6 e5b2 g8h8 e4g5 e7f6 b2f6 f8f6 d3e4');
INSERT INTO evaluations VALUES(17516,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(17516,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(17516,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(17516,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(17516,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(17516,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(17516,36,'cp 41','f3d4 e7e6 e1g1 d8c7 d4c6 b7c6 d1e2 g8e7 b1c3 e7g6 f2f4 f8e7 e4e5 c8b7 c3e4 c6c5 c4d3 e8g8 b2b3 c7c6 c1d2 g6h4 d2c3 h4f5 a2a4 a6a5 g1h1 g7g6 e4d2 f5d4 c3d4 c5d4 d3e4 c6a6 d2c4 b7e4 e2e4 e7c5 a1d1');
INSERT INTO evaluations VALUES(17515,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(17515,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(17515,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(17515,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(17515,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(17515,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(17515,36,'cp 22','e7e6 e1g1 d8c7 d4c6 b7c6 d1e2 g8e7 b1c3 e7g6 f2f4 f8e7 e4e5 c8b7 c3e4 c6c5 c1d2 g6h4 b2b3 e8g8 d2c3 h4f5 a1d1 a6a5 a2a4 c7c6 d1e1 e7h4 g2g3 d7d5 e5d6 c6e4 e2e4 b7e4 e1e4 f5d6 e4e3 d6c4 b3c4 h4f6 c3f6 g7f6 f4f5 e6e5 f1d1');
INSERT INTO evaluations VALUES(17420,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(17420,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(17420,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(17420,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(17420,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17420,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(17420,36,'cp 40','b1c3 g8f6 d4c6 d8c7 d1e2 d7c6 a2a4 f8b4 e1g1 e6e5 f2f3 f6h5 g2g3 e8g8 c1e3 h5f6 f1c1 a8b8 c3d1 b7b5 a4b5 a6b5 c4d3 f6d7 c2c4 b5c4 d3c4 b4c5 e2f2 c5e3 f2e3 b8b7 b2b3 f8d8 e3c3 c7b6 d1e3 d7f6');
INSERT INTO evaluations VALUES(54175,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54175,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54175,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54175,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54175,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54175,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54175,36,'cp 0','c5d4 f3d4 e6e5 d4b5 a7a6 d1a4 c8d7 c4d5 g8f6 e2e4 f8c5 c1e3 c5e3 b5d6 e8f8 a4a3 e3h6 d6c8 f8e8 c8d6');
INSERT INTO evaluations VALUES(54341,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54341,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54341,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54341,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54341,36,'cp 0','f3d4 e6e5 d4b5 a7a6 d1a4 c8d7 c4d5 g8f6 e2e4 f8c5 c1e3 c5e3 b5d6 e8f8 a4a3 e3h6 d6f5 f8e8 f5d6');
INSERT INTO evaluations VALUES(54334,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54334,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54334,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54334,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54334,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54334,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54334,36,'cp 0','e6e5 d4b5 a7a6 d1a4 c8d7 c4d5 g8f6 e2e4 f8c5 c1e3 c5e3 b5d6 e8f8 a4a3 e3h6 d6f5 f8e8 f5d6');
INSERT INTO evaluations VALUES(53402,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(53402,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(53402,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53402,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(53402,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(53402,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(53402,36,'cp 19','e4d5 g8f6 d2d4 f6d5 b1c3 b8c6 g1f3 c8f5 f1b5 e7e6 e1g1 f8e7 f3e5 a8c8 d1a4 e8g8 b5c6 d5b6 a4a7 b7c6 a7b7 c6c5 d4c5 e7c5 b7f3 d8d4 f3g3 b6c4 e5c4 d4c4 h2h3 c4b4 g3f4 f8d8 f4b4 c5b4 c1e3 b4c3 b2c3 c8c3');
INSERT INTO evaluations VALUES(55771,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(55771,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(55771,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(55771,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(55771,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(55771,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(55771,36,'cp -93','e6d5 g1f3 g8h6 d2d4 f8e7 b1c3 e8g8 f1e2 e7h4 g2g3 h4e7 c1e3 f8e8 e1g1 c8h3 f1e1 c7c6 f3e5 b8d7 c3a4 d7f6 e2f3 d8c8 a4c5 h6g4 c5d3 h7h5 a2a3 c8f5 d3f2 g4f2 e3f2 e7d6 d1d3 g7g6 c2c3 g8g7 d3f5 h3f5 c3c4 a8d8 c4c5 d6c7');
INSERT INTO evaluations VALUES(56262,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(56262,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(56262,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(56262,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(56262,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(56262,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(56262,36,'cp -78','d2d4 f8e7 g1f3 g8f6 f3e5 c7c5 f1b5 c8d7 b5d7 b8d7 c2c3 e8g8 e1g1 f8e8 c1e3 c5d4 e3d4 e7c5 b1d2 d8b6 d2b3 c5d4 d1d4 d7e5 f4e5 f6g4 d4b6 a7b6 b3d4 e8e5 a2a4 a8e8 f1d1 g7g6 h2h3 g4e3 d1d3 e5e4 d4f3 f7f5 b2b3 f5f4');
INSERT INTO evaluations VALUES(56200,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(56200,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(56200,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(56200,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(56200,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(56200,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(56200,36,'cp -85','f8e7 g1f3 g8f6 f3e5 e8g8 c1e3 c7c5 d4c5 b8d7 b1c3 d7e5 f4e5 f6g4 e3d4 g4e5 f1e2 e5c6 e1g1 c8e6 d4f2 d5d4 c3e4 b7b6 e2f3 a8c8 c5b6 d8b6 b2b3 b6a5 g1h1 c6e5 f2d4 e5f3 d1f3 c8c2 d4c3 a5c7 a1d1');
INSERT INTO evaluations VALUES(13091,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13091,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13091,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13091,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13091,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13091,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13091,36,'cp 0','b1d2 c8f5 g1f3 d8b6 d4c5 b6b2 f3d4 e7e5 f4e5 f5g4 a1b1 b2a3 f2f3 c6e5 f3g4 f8c5 d4b5 a3a2 b1a1 a2b2 a1c1 b2a2');
INSERT INTO evaluations VALUES(13090,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13090,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13090,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13090,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13090,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13090,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13090,36,'cp 0','c8f5 g1f3 d8b6 d4c5 b6b2 f3d4 e7e5 f4e5 f5g4 a1b1 b2a3 f2f3 c6e5 f3g4 f8c5 d4b5 a3a2 b1a1 a2b2 a1c1 b2a2');
INSERT INTO evaluations VALUES(24609,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24609,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24609,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24609,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24609,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24609,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24609,36,'cp 0','g1f3 d8b6 d4c5 b6b2 f3d4 e7e5 f4e5 f5g4 a1b1 b2a3 f2f3 c6e5 f3g4 f8c5 d4b5 a3a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(24610,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24610,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24610,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24610,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24610,36,'cp 0','d8b6 d4c5 b6b2 f3d4 e7e5 f4e5 f5g4 a1b1 b2a3 f2f3 c6e5 f3g4 f8c5 d4b5 a3a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(56428,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(56428,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(56428,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(56428,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(56428,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(56428,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(56428,36,'cp -66','h7h5 h2h4 b8c6 a2a3 g8f6 d2d3 c8f5 b1c3 d5d4 c3e4 f6e4 d3e4 f5e6 g1f3 f7f6 e2e3 f8c5 e3d4 c6d4 c1b2 d4c6 e1g1 d8e7 b3b4 c5b6 c2c3 g7g5 a3a4 a7a5 b4b5 c6d8 d1c2 g5h4 f3h4 e7g7 g1h1 d8f7 h4f5 g7h7');
INSERT INTO evaluations VALUES(39605,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(39605,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(39605,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(39605,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(39605,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39605,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(39605,36,'cp 17','d7d5 e4d5 d8d5 g1f3 g8h6 b1c3 d5a5 d1d2 h6f5 f1c4 e8g8 e1g1 b8d7 f1e1 c7c5 c3d5 a5d2 e3d2 d7b6 d5b6 a7b6 d4c5 b6c5 c2c3 f5d6 c4b3 c5c4 b3c2 g7f6 a2a4 f8d8 d2f4 c8g4 f3d2 g4e6 d2e4 d6e4 c2e4 e6d5 a4a5 e7e6 a1a4 d5e4');
INSERT INTO evaluations VALUES(47799,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(47799,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(47799,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(47799,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(47799,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47799,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(47799,36,'cp 27','g1f3 f8e7 d2d4 e5d4 f3d4 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 e7d6 f1e1 b8d7 d1c2 h7h6 c1d2 d7f6 e4d3 c8d7 h2h3 e8e1 a1e1 d8f8 d4f5 d7f5 d3f5 a8e8 g1f1 e8e1 d2e1 f8e8 c2e2 e8a4 b2b3 a4a3 f5d3 d6f8 g2g3 c7c6 b3b4');
INSERT INTO evaluations VALUES(55075,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(55075,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(55075,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(55075,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55075,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(55075,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(55075,36,'cp -34','d5e4 d3e4 g8f6 e4f3 e7e5 b1c3 d8d6 g1e2 b8c6 c3b5 d6d7 b5c3 f8e7 e1g1 e8g8 d2d3 f8e8 f1e1 h7h6 h2h3 a7a5 a2a3 e7f8 c1e3 a5a4 c3e4 f6e4 d3e4 d7d8 d1d8 e8d8 e1d1 c8e6 e2c3 c6d4 e3d4 e5d4 c3b5 c7c5 b5c7');
INSERT INTO evaluations VALUES(55286,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(55286,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(55286,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(55286,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(55286,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(55286,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(55286,36,'cp -27','d3e4 g8f6 e4f3 e7e5 b1c3 b8c6 g1e2 d8d6 c3b5 d6d7 b5c3 f8d6 e1g1 h7h6 f1e1 e8g8 d2d3 f8e8 h2h3 a7a5 a2a3 d6f8 c1e3 b7b6 e2g3 c8b7 a1b1 a5a4 d1d2 e8d8 b1d1 a8a5 d2e2 f6e8 f3g4 c6d4 e3d4 d7d4 c3e4');
INSERT INTO evaluations VALUES(55274,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(55274,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(55274,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(55274,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(55274,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(55274,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55274,36,'cp -44','g8f6 e4f3 e7e5 b1c3 d8d6 g1e2 b8c6 c3b5 d6d7 b5c3 f8e7 d2d3 e8g8 e1g1 f8e8 f1e1 h7h6 a2a3 e7f8 c1e3 a7a5 c3e4 f6h7 d3d4 e5d4 e2d4 c6e5 a3a4 e5f3 d1f3 d7d5 e4c3 d5f3 d4f3 c8g4 h2h3 g4f3 g2f3');
INSERT INTO evaluations VALUES(47390,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(47390,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(47390,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(47390,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(47390,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(47390,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(47390,36,'cp -29','e4f3 e7e5 b1c3 d8d6 g1e2 b8c6 c3b5 d6d7 b5c3 f8e7 e1g1 e8g8 d2d3 h7h6 a2a3 f8e8 f1e1 e7f8 h2h3 a7a5 c1e3 c6d4 e3d4 e5d4 c3e4 f6h7 c2c3 d4c3 b2c3 c7c6 d3d4 d7c7 a3a4 c8e6 d1d2 e6d5 d2f4 c7f4 e2f4 a8d8 a1d1 d5e4 e1e4');
INSERT INTO evaluations VALUES(49378,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49378,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(49378,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(49378,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(49378,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(49378,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(49378,36,'cp -26','d7d5 c4d5 e6d5 d2d4 f8b4 f1g2 e8g8 g1f3 f8e8 e1g1 b8d7 d1c2 c7c6 c1f4 d7b6 f3e5 a7a5 a2a3 b4d6 a1d1 a5a4 h2h3 f6h5 f4c1 f7f5 g2f3 h5f6 c1f4 c8e6 e5d3 d6f4 d3f4 g7g6 f1e1 e6f7 f4d3 f6e4 f3g2');
INSERT INTO evaluations VALUES(48238,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48238,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(48238,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48238,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(48238,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48238,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(48238,36,'cp 0','d2d4 c7c5 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 b2b4 c5c7 f2f4 b7b6 f1g2 c8b7 g2b7 c7b7 g1f3 e8g8 e1g1 f8d8 b4b5 b7e4 f3g5 e4g6 g5f3 g6e4');
INSERT INTO evaluations VALUES(48257,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(48257,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48257,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(48257,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(48257,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48257,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(48257,36,'cp -33','c7c5 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 b2b4 c5c7 f2f4 e8g8 f1g2 b8c6 a1c1 e6e5 b4b5 c6e7 g1f3 e5e4 f3d4 f8d8 e2e3 e7f5 d4f5 c8f5 e1g1 a7a6 d1e2 a6b5 c4b5 f5g4 e2f2 c7b6 d2c3 b6b5 h2h3 g4e2 f1e1 e2d3 g3g4 h7h6 c3d4 a8c8 g1h1');
INSERT INTO evaluations VALUES(44563,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(44563,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(44563,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(44563,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(44563,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(44563,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(44563,36,'cp 95','e4e5 f6g8 d2d4 c5d4 d1d4 b8c6 d4f4 d7d5 g1f3 e7e6 f1d3 c6b4 e1g1 b4d3 c2d3 g8e7 c3b5 e7g6 f4g3 c8d7 c1g5 d8b8 b5d4 f8c5 g5e3 c5d4 e3d4 e8g8 d4c5 f8e8 h2h4 g6e7 h4h5 h7h6 a1c1 b8d8 c1c3 e7f5 g3f4 a8c8 f1c1 a7a6 g2g4 f5h4 f3d4 f7f5 c5d6 c8c3 c1c3 f5g4 c3c7');
INSERT INTO evaluations VALUES(44512,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(44512,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(44512,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(44512,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(44512,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(44512,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(44512,36,'cp 99','f6g8 d2d4 c5d4 d1d4 e7e6 d4f4 b8c6 g1f3 d7d5 f1d3 c6b4 e1g1 b4d3 c2d3 g8e7 c3b5 e7g6 f4g3 c8d7 b5d6 f8d6 e5d6 h7h5 h2h4 e8g8 d3d4 d8f6 f3e5 g6e5 d4e5 f6f5 c1g5 a8c8 f1c1 a7a6 c1c7 c8c7 d6c7 f8c8 a1c1');
INSERT INTO evaluations VALUES(52410,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(52410,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(52410,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(52410,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(52410,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(52410,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(52410,36,'cp 109','d2d4 c5d4 d1d4 b8c6 d4f4 e7e6 g1f3 d7d5 c1d2 a7a6 e1c1 f7f6 g2g4 f6e5 f3e5 g8h6 f4g3 f8d6 f2f4 d8c7 d1e1 e8g8 f1d3 c6b4 c1b1 h6f7 g4g5 d6e5 f4e5 b4d3 c2d3 f7d8 c3e2 d8c6 g5g6 h7h6 e2f4 c8d7 f4h5');
INSERT INTO evaluations VALUES(13251,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(13251,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13251,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(13251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13251,36,'cp 0','e1g1 e7e6 a2a4 f8e7 c4b3 c6a5 b3a2 a7a6 c2c3 c8d7 f1e1 b7b5 c1f4 e8g8 e4e5 d6e5 f3e5 a5c6 d1f3 c6e5 f4e5 d8b6 b1d2 d7c6 f3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(13250,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13250,36,'cp 0','e7e6 f1e1 f8e7 d1e2 a7a6 a2a4 e8g8 c2c3 c8d7 c1f4 c6a5 c4a2 b7b5 e4e5 d6e5 f3e5 a5c6 e2f3 c6e5 f4e5 d8b6 b1d2 d7c6 f3g3 f6h5 g3e3 h5f6');
INSERT INTO evaluations VALUES(13937,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13937,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13937,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13937,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13937,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13937,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13937,36,'cp 0','f1e1 f8e7 a2a4 e8g8 b1d2 d8c7 c2c3 c8d7 h2h3 a7a6 d2f1 h7h6 c1f4 c6e5 f4e5 d6e5 a4a5 d7b5 f1e3 f6d7 d1b3 b5c6 b3d1 a8d8 c4b3 c6b5 b3c4');
INSERT INTO evaluations VALUES(37821,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(37821,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(37821,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(37821,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(37821,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37821,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(37821,36,'cp 78','d2d4 c5d4 c3d4 d7d5 e4e5 g8h6 f1b5 c8d7 b5d7 d8d7 b1c3 b8c6 e1g1 e7e6 c3a4 a8c8 a4c5 d7c7 c1g5 h6f5 a1c1 h7h6 g5f6 g7f6 e5f6 e8g8 d1d2 c7d8 f3e5 c8c7 e5c6 b7c6 d2f4 g6g5 f4e5 d8d6 c1c3 d6e5 d4e5 f8d8 h2h3');
INSERT INTO evaluations VALUES(38377,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(38377,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38377,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38377,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(38377,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(38377,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(38377,36,'cp 50','b1c3 g8f6 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 a2a3 b8a6 c1e3 b7b6 f3d2 g4e2 d1e2 d8d7 f1b1 a5a4 h2h3 f6e8 b2b4 a4b3 d2b3 f7f5 f2f3 f5e4 c3e4 a6c5 b3c5 d6c5 e3d2 e8d6 e4d6 d7d6 e2e4 a8a4 b1b3 d6d7 g2g4');
INSERT INTO evaluations VALUES(38373,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38373,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(38373,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(38373,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(38373,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(38373,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(38373,36,'cp 63','g8f6 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 a2a3 b8a6 c1e3 g4d7 f3d2 h7h5 h2h3 b7b6 a1b1 f6h7 b2b4 a5b4 a3b4 f7f5 f2f3 f5f4 e3f2 g7f6 d1b3 f6h4 c4c5 b6c5 e2a6 c5b4 a6b7 b4c3 b7a8 c3d2 a8c6 h4g3 c6d7 d8d7 f2g3 f4g3 b3d3');
INSERT INTO evaluations VALUES(52112,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(52112,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(52112,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(52112,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(52112,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(52112,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(52112,36,'cp 41','g2g3 b8c6 a2a3 g7g6 f1g2 a7a5 d2d3 d7d6 c3d5 h7h6 c1d2 g8f6 e2e3 f8g7 g1e2 e8g8 d5f6 g7f6 e1g1 f6g7 d1b3 h6h5 e2c3 h5h4 h2h3 g8h7 c3d5 a8b8 g1h2 f7f5 b3b6 c8e6 b6d8 h4g3 f2g3 b8d8 d5c7');
INSERT INTO evaluations VALUES(52111,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(52111,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52111,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(52111,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(52111,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(52111,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(52111,36,'cp 31','g7g6 a2a3 a7a5 d2d3 d7d6 c3d5 h7h6 e2e3 b8c6 f1g2 f8g7 g1e2 g8e7 d5e7 d8e7 e2c3 e8g8 e1g1 c8e6 c1d2 f8b8 a1b1 e7d7 h2h4 e6f5 g2e4 f5h3 f1e1 h3e6 e4d5 e6d5 c4d5 c6e7 b2b4 a5b4 a3b4 b7b5 e3e4 g8h8');
INSERT INTO evaluations VALUES(53989,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53989,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(53989,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(53989,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53989,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53989,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53989,36,'cp 0','c5d4 e1g1 b8c6 f3d4 g8f6 f1e1 f8e7 c4f1 e8g8 c2c4 d6d5 e4d5 e6d5 b1c3 c8g4 d1d2 c6d4 d2d4 d5c4 d4d8 e7d8 f1c4 a8c8 c4b3 c8c6 b3a4 c6c8');
INSERT INTO evaluations VALUES(54018,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(54018,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(54018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54018,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54018,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(54018,36,'cp 5','f3d4 g8f6 c4d3 b8c6 d4c6 b7c6 e1g1 f8e7 c2c4 d6d5 e4d5 c6d5 c4d5 f6d5 b1c3 e8g8 d1c2 h7h6 c3d5 e6d5 c1f4 d8b6 a1c1 c8a6 f4e3 a6d3 c2d3 b6b2 d3d5 b2a3 c1c7 f8d8 d5h5 a8c8');
INSERT INTO evaluations VALUES(54016,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(54016,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(54016,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(54016,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(54016,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(54016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54016,36,'cp 3','g8f6 b1c3 a7a6 e1g1 f8e7 a2a4 e8g8 c1e3 b8c6 h2h3 c8d7 c4d3 c6d4 e3d4 d7c6 d1f3 f6d7 f1d1 d8a5 f3g3 a5g5 g3g5 e7g5 a4a5 f8c8 c3a4 g5d8 a4b6 d7b6 d4b6 d8b6 a5b6 g7g6 f2f4 g8g7 f4f5');
INSERT INTO evaluations VALUES(46041,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(46041,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46041,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46041,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(46041,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(46041,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46041,36,'cp 10','b1c3 a7a6 a2a4 f8e7 e1g1 b8c6 c1e3 e8g8 h2h3 c8d7 c4d3 d8c7 d1e2 d6d5 e4d5 e6d5 d4c6 d7c6 e3d4 a8e8 e2f3 f6e4 c3d5 c6d5 d3e4 d5e4 f3e4 e7f6 e4d3 f6d4 d3d4 c7c2 f1c1 c2e4 c1c4 e4d4 c4d4 e8e2 b2b4 f8e8 d4d7 e2e7 a1d1 g7g6 b4b5 e7d7 d1d7 a6b5 a4b5 b7b6');
INSERT INTO evaluations VALUES(32486,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(32486,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(32486,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(32486,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(32486,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(32486,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(32486,36,'cp 104','c1e3 g8f6 b1c3 f8e7 d1d2 f6g4 e3f4 b8c6 d4c6 d7c6 f1e2 g4f6 f2f3 f6d7 e1c1 e7f6 c3b5 a7a6 b5d4 f6d4 d2d4 d8f6 d4f6 d7f6 h2h4 f6d7 h4h5 e8c8 g2g4 f7f6 b2b3 h8e8 a2a4 h7h6 f4d2 d7e5 f3f4 e5f7');
INSERT INTO evaluations VALUES(55726,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(55726,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(55726,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(55726,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(55726,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(55726,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(55726,36,'cp 50','c2c4 d6d5 e2e3 g8f6 b1c3 f8d6 f4e5 c7c6 g1f3 d6e7 f1d3 d5c4 d3c4 b8d7 e5g3 e8g8 d1c2 b7b5 c4d3 h7h6 e1g1 c8b7 a1d1 a7a6 f3e5 d7e5 g3e5 a8c8 e5f6 e7f6 c3e4 f6e7 e4c5');
INSERT INTO evaluations VALUES(14384,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(14384,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(14384,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(14384,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(14384,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(14384,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(14384,36,'cp 87','g1f3 c8g4 d4d5 c6e5 c1e3 c7c6 d5c6 d8d1 a1d1 e5c6 f1c4 a8c8 e1g1 e7e6 d1c1 g4f3 g2f3 f8e7 c4b5 e8g8 b5c6 c8c6 e3a7 f8a8 a7d4 f6h5 f1d1 h5f4 g1f1 f4g6 d4e3 h7h6 d1d7 b7b6 d7b7 e7f6 f3f4 f6c3 c1c3 c6c3 b2c3 a8a2 b7b6 a2c2 b6b8 g8h7 b8c8 g6h4 c8c7 h7g6 c3c4');
INSERT INTO evaluations VALUES(48522,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(48522,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(48522,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(48522,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(48522,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(48522,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(48522,36,'cp 132','f1d3 c7c5 d4d5 a7a6 c1f4 g7g6 g1f3 f8g7 d5d6 b8c6 e1g1 h7h6 f1e1 g6g5 e4e5 g5f4 e5f6 g7f6 c3d5 e8g8 d5f4 e6e5 f4h5 c8g4 h5g3 f6g7 h2h3 g4f3 d1f3 d8d6 f3e4 d6g6 e4g6 f7g6 d3g6 c6d4 g6e4 a8d8 e4b7');
INSERT INTO evaluations VALUES(48524,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(48524,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(48524,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(48524,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(48524,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(48524,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(48524,36,'cp 128','c7c5 d4d5 e6d5 e4d5 f8d6 d1e2 d8e7 c3b5 e8f8 g1f3 f6d5 d3c4 d5b6 b5d6 e7d6 c1e3 b8c6 e1g1 b6c4 a1d1 d6e7 e2c4 f8g8 e3c5 c8e6 c4b5 e7c7 f3g5 e6g4 f2f3 g4h5 g5e4 h7h6 a2a3 h5g6 c5d6 c7b6 b5b6 a7b6 e4c3 g6c2 d1d5 a8d8 d6c7 d8d5 c3d5 g8h7 d5b6');
INSERT INTO evaluations VALUES(19554,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(19554,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(19554,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(19554,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(19554,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(19554,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(19554,36,'cp -15','d4c3 b1c3 e7e6 f1e2 g8f6 e1g1 d7d6 d1b3 f8e7 f1d1 e6e5 c1e3 e8g8 a1c1 a7a6 e3g5 a8b8 g5f6 e7f6 b3a3 f6e7 c3d5 c8g4 d5e7 d8e7 a3d6 e7d6 d1d6 b8d8 d6d5 g4e6 d5c5 c6d4 e2c4 e6c4 c1c4 d4c6 c5d5 f7f6 h2h3 d8d5 e4d5');
INSERT INTO evaluations VALUES(19564,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(19564,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(19564,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(19564,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(19564,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(19564,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(19564,36,'cp -23','b1c3 e7e6 f1e2 f8e7 e1g1 d7d6 d1b3 g8f6 f1d1 e6e5 c1g5 e8g8 g5f6 e7f6 a1c1 a8b8 b3a3 f6e7 c3d5 e7g5 f3g5 d8g5 a3d6 c8e6 d6a3 f8d8 a3e3 h7h6 e2b5 g8h7 e3g5 h6g5 b5c6 b7c6 d5e3 d8d1 e3d1 b8d8 b2b3 d8d2 d1c3');
INSERT INTO evaluations VALUES(19431,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(19431,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(19431,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19431,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(19431,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(19431,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(19431,36,'cp 0','g1f3 d7d6 f1b5 e7e5 e1g1 f8e7 d2d3 g8f6 b1c3 c8e6 h2h3 h7h6 c3e2 e8g8 c1d2 d8c7 b5c6 c7c6 e2g3 b7b5 c2c4 c6d7 f3h2 a8c8 a2a4 a7a6 f2f4 e5f4 d2f4 d7b7 d1f3 b5c4 b3c4 c8b8 a1e1 g8h7 g3f5 e6f5 e4f5');
INSERT INTO evaluations VALUES(19430,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19430,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(19430,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19430,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(19430,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(19430,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19430,36,'cp -6','d7d6 f1b5 e7e5 e1g1 f8e7 b1c3 g8f6 d2d3 e8g8 c3e2 f6e8 f3d2 c6d4 b5c4 e8c7 a2a4 c8g4 f2f3 g4e6 c4e6 c7e6 d2c4 f7f5 c2c3 d4e2 d1e2 f5f4 c1a3 b7b6 b3b4 d8d7 b4b5 g8h8 f1b1 a8e8 g1h1');
INSERT INTO evaluations VALUES(19992,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(19992,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(19992,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(19992,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(19992,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19992,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(19992,36,'cp -13','f1b5 e7e5 e1g1 f8e7 b1c3 g8f6 h2h3 c8d7 b5c4 e8g8 d2d3 d7e6 f3g5 c6d4 a2a4 h7h6 g5f3 a7a6 c1d2 d8d7 a4a5 a8e8 c3d5 e6d5 e4d5 d4b5 f3h2 f6h7 c4b5 a6b5 c2c4 f7f5 b3b4 h7f6 b4c5 d6c5 d2c3');
INSERT INTO evaluations VALUES(22849,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(22849,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(22849,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(22849,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(22849,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(22849,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(22849,36,'cp -64','d7d5 d2d3 c8e6 b1d2 f7f6 a2a3 d8d7 f1g2 e8c8 b3b4 c8b8 c2c4 g7g5 b4b5 c6e7 d1c2 h7h5 h2h3 g8h6 d2b3 e7g6 e2e3 d5c4 d3c4 e6f5 c2c3 e5e4 b3d4 g6e5 e1c1 d7e7 d4f5 d8d1 c1d1 h6f5 g2e4 e5g4');
INSERT INTO evaluations VALUES(21738,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(21738,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(21738,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(21738,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(21738,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(21738,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(21738,36,'cp -53','d2d3 g8f6 f1g2 f8c5 g1f3 d8e7 e1g1 e5e4 d3e4 d5e4 f3d4 c8d7 d4c6 d7c6 e2e3 h7h5 b1d2 e8c8 d1e2 c5a3 b2a3 e7a3 d2c4 a3e7 a1d1 c6d5 h2h4 g7g5 h4g5 f6h7 f2f4 e4f3 g2f3 d5c4 e2c4 h7g5 g1g2 e7e3 c4e2 e3e2 f3e2 d8d1 e2d1');
INSERT INTO evaluations VALUES(52706,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(52706,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52706,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(52706,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(52706,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52706,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(52706,36,'cp 13','c6d5 e4d5 g8f6 b1c3 f6d5 d2d4 b8c6 g1f3 c8f5 f1b5 e7e6 e1g1 f8e7 f3e5 a8c8 d1a4 a7a6 b5c6 b7c6 c1e3 d5c3 b2c3 f7f6 e5c4 e8g8 a4a6 d8c7 a2a4 c8a8 a6b6 e7d6 b6c7 d6c7 f1c1 f8c8 a4a5 a8a6 c4b6 c7b6 a5b6 a6b6 h2h3 b6b7 f2f3 f5g6 c1e1 c8b8 c3c4 g8f7');
INSERT INTO evaluations VALUES(14904,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(14904,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(14904,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(14904,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(14904,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(14904,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(14904,36,'cp 82','b1c3 f8b4 a2a3 b4c3 b2c3 c6a5 c4d3 b7b6 a3a4 e8g8 e3e4 h7h6 e1g1 c8b7 f1e1 c7c5 h2h3 c5d4 c3d4 a8c8 c1f4 f8e8 a1c1 a5c6 d3b5 e8e7 d1d2 a7a6 b5f1 e7d7 c1d1 c6a5 f1d3 b7c6 d3c2 c6b7');
INSERT INTO evaluations VALUES(17247,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(17247,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(17247,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(17247,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(17247,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(17247,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(17247,36,'cp 52','g1f3 e7e6 e2e3 d7d5 b1c3 f8e7 a2a3 e8g8 b2b4 c6b8 c4c5 c7c6 f1d3 a7a5 c1b2 b7b6 c3a4 b6c5 b4c5 c8a6 d3a6 a8a6 d1d3 a6a7 e1g1 b8d7 f1b1 f6e4 b2c3 f7f5 a4b6 g7g5 b6d7 d8d7');
INSERT INTO evaluations VALUES(17246,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(17246,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17246,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(17246,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(17246,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(17246,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(17246,36,'cp 61','e7e6 a2a3 d7d5 b1c3 f8e7 c1f4 f6e4 e2e3 e4c3 b2c3 e8g8 c4d5 e6d5 f1d3 e7d6 f4d6 d8d6 e1g1 h7h6 f1e1 c8e6 d1e2 c6e7 a3a4 b7b6 h2h3 f8d8 e3e4 d5e4 d3e4 e7d5 e4c2 c7c5 c3c4 d5b4 d4d5 e6d7 f3e5 a7a5 e5d7 b4c2 e2c2');
INSERT INTO evaluations VALUES(12592,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(12592,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(12592,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(12592,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(12592,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(12592,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(12592,36,'cp -25','f8b4 c1d2 b4c3 d2c3 e8g8 d1f3 a7a5 a2a4 c6b4 c3b4 a5b4 f1b5 d8c7 a1c1 b7b6 e1g1 c8a6 c2c4 b4c3 c1c3 c7d8 b3d2 a6b5 a4b5 a8a5 f3d3 d7d5 c3c6 h7h6 b2b4 a5a8 e4d5 f6d5 f1d1 d5b4 d3d8 f8d8 c6b6');
INSERT INTO evaluations VALUES(7860,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(7860,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(7860,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(7860,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(7860,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(7860,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(7860,36,'cp -16','c1d2 b4c3 d2c3 e8g8 f1d3 d7d5 e4d5 f6d5 d1d2 d8f6 e1g1 f8d8 d3e4 c8e6 b3c5 d5c3 d2c3 e6d5 e4d5 d8d5 c5b7 c6d4 c3c4 d5b5 c2c3 b5b7 c3d4 e5d4 b2b3 b7d7 a1d1 g7g6 d1d3 a8d8 f1d1 a7a6 h2h3 d7d6 c4c2 h7h5 c2d2 d6d5 a2a3 a6a5 b3b4 a5b4 a3b4 d8d7');
INSERT INTO evaluations VALUES(7861,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(7861,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(7861,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(7861,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(7861,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(7861,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(7861,36,'cp -31','b4c3 d2c3 e8g8 d1f3 d7d6 h2h3 a7a5 a2a4 c6b4 f1d3 b4d3 f3d3 c8e6 e1g1 d8c7 f1d1 f8c8 d3e3 h7h6 e3e1 b7b6 b3c1 c8e8 c1d3 a8d8 f2f3 f6d7 d1d2 f7f5 a1d1 f5e4 f3e4 d7f6 d3f2 e6f7 g1h2 f7g6 e1e2 c7c6 e2b5 c6b5 a4b5 f6e4 f2e4 g6e4');
INSERT INTO evaluations VALUES(7871,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(7871,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(7871,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(7871,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(7871,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(7871,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(7871,36,'cp -15','d2c3 e8g8 d1f3 d7d6 h2h3 c8e6 f1d3 a7a5 a2a4 c6b4 c3b4 a5b4 e1g1 e6d7 d3b5 d8c7 f3d3 d7b5 a4b5 g7g6 c2c4 b4c3 b2c3 f8c8 a1a8 c8a8 c3c4 c7b6 f1c1 a8a2 c4c5 d6c5 b3c5 g8g7');
INSERT INTO evaluations VALUES(7869,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(7869,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(7869,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(7869,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(7869,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(7869,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(7869,36,'cp -23','e8g8 d1f3 d7d6 h2h3 a7a5 a2a4 c6b4 f1d3 d8b6 b3d2 c8e6 f3e2 a8c8 c3b4 b6b4 b2b3 d6d5 e1g1 d5e4 d2e4 f6e4 e2e4 b4e4 d3e4 f7f5 e4b7 c8c2 a1c1 c2c1 f1c1 f8b8 b7a6 b8b4 c1c5 e6b3 c5a5 b4a4 a5b5 a4a6 b5b3 g8f7 g2g3 g7g5 b3b7 f7g6 b7e7 e5e4 g3g4 g6f6 e7h7 a6a1 g1h2 f5f4');
INSERT INTO evaluations VALUES(20956,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(20956,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(20956,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(20956,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(20956,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(20956,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(20956,36,'cp 15','d8d5 g1f3 c8g4 f1e2 e8c8 c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 a1d1 d8d1 e1d1 g8f6 e2c4 f8d6 d1c1 e5e6 c4e6 f7e6 f2f3 c8d7 e3d2 a7a6 h2h3 h7h5 h1e1 e6e5 a2a4 h8e8 c3e4 b7b5 b2b3 b5a4 b3a4 f6e4 e1e4 g7g6 c1d1 e8h8 d1e2 d7e6 d2e3');
INSERT INTO evaluations VALUES(17497,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(17497,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17497,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17497,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(17497,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(17497,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(17497,36,'cp 38','d2d4 c5d4 f3d4 e7e6 c1f4 d7d6 d4c6 b7c6 f1e2 f8e7 e4e5 d6d5 e1g1 g8h6 f4h6 g7h6 e2d3 e8g8 d1h5 f7f5 h5h6 g8h8 c3e2 e7g5 h6h5 c6c5 b2b3 f8g8 f2f4 g5e7 g1h1 c8b7 e2g3 g8g7 h5h6 c5c4 h6e6 g7g3 h2g3 c4d3');
INSERT INTO evaluations VALUES(17494,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(17494,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(17494,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(17494,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(17494,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17494,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(17494,36,'cp 48','c5d4 f3d4 e7e6 c1e3 d8c7 d1f3 b7b5 f3g3 c7g3 h2g3 c8b7 d4c6 b7c6 f2f3 f8b4 e3d4 f7f6 a2a3 b4d6 c3e2 g8e7 d4f2 f6f5 e1c1 d6e5 e4f5 e7f5 d1e1 e5f6 g3g4 f5d6 e2d4 c6d5 d4f5 d6f5 g4f5 e8f7 f1d3 g7g6 f5g6 h7g6');
INSERT INTO evaluations VALUES(17524,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(17524,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(17524,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(17524,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(17524,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(17524,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(17524,36,'cp 37','f3d4 e7e6 d4c6 b7c6 f1d3 d8c7 e1g1 c8b7 d1e2 d7d5 c1g5 f8d6 f2f4 h7h6 g5h4 g7g5 e4e5 g5h4 e5d6 c7d6 a1e1 g8f6 f4f5 e6e5 e2e5 d6e5 e1e5 e8d7 f1e1 d7d6 b2b4 a6a5 b4a5 h4h3 e5e7 h8e8 e7e8 f6e8 g1f2 h3g2 f2g2');
INSERT INTO evaluations VALUES(33231,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(33231,31,'cp 298',NULL);
INSERT INTO evaluations VALUES(33231,32,'cp 317',NULL);
INSERT INTO evaluations VALUES(33231,33,'cp 305',NULL);
INSERT INTO evaluations VALUES(33231,34,'cp 318',NULL);
INSERT INTO evaluations VALUES(33231,35,'cp 295',NULL);
INSERT INTO evaluations VALUES(33231,36,'cp 305','c3d5 f6d8 f3e5 d7d6 e5c4 c7c6 d5f4 a7a6 d2d4 c5a7 c2c3 g8e7 a2a4 e8g8 f1d3 d6d5 e4d5 e7d5 e1g1 b8d7 f1e1 h7h6 f4d5 c6d5 c4d6 d7f6 d6c8 a8c8 g2g3 f8e8 e1e8 d8e8 c1e3 a7b8 d1f3 e8c6 g1g2 c8e8 h2h4 b8c7 a4a5 c7b8');
INSERT INTO evaluations VALUES(33257,30,'cp 303',NULL);
INSERT INTO evaluations VALUES(33257,31,'cp 301',NULL);
INSERT INTO evaluations VALUES(33257,32,'cp 319',NULL);
INSERT INTO evaluations VALUES(33257,33,'cp 319',NULL);
INSERT INTO evaluations VALUES(33257,34,'cp 308',NULL);
INSERT INTO evaluations VALUES(33257,35,'cp 321',NULL);
INSERT INTO evaluations VALUES(33257,36,'cp 319','f6d8 f3e5 g8f6 d2d4 c5e7 d5f6 e7f6 e5g4 e8g8 f1c4 b8c6 c2c3 f8e8 c4d5 d7d6 e1g1 c6a5 g4f6 d8f6 d1a4 c7c6 a4a5 b7b6 a5a4 b6b5 a4b3 c6d5 b3d5 a8b8 f2f3 b5b4 d5a5 b4c3 b2c3 c8d7 c1e3 f6d8 a5a7 d6d5 e3f4 b8b2 e4d5 d7b5 f1e1 d8d5 e1e8 b5e8 a1e1');
INSERT INTO evaluations VALUES(39309,30,'cp 323',NULL);
INSERT INTO evaluations VALUES(39309,31,'cp 310',NULL);
INSERT INTO evaluations VALUES(39309,32,'cp 297',NULL);
INSERT INTO evaluations VALUES(39309,33,'cp 305',NULL);
INSERT INTO evaluations VALUES(39309,34,'cp 306',NULL);
INSERT INTO evaluations VALUES(39309,35,'cp 301',NULL);
INSERT INTO evaluations VALUES(39309,36,'cp 317','f3e5 d7d6 e5c4 c7c6 d5f4 g8e7 d2d4 c5b6 c4b6 a7b6 f1d3 e8g8 e1g1 c6c5 d3c4 b8c6 c2c3 e7g6 f1e1 c5d4 c3d4 d8h4 f4g6 h7g6 f2f3 c6d4 c1e3 d4c6 d1d6 f8e8 e3f2 h4h6 f2b6 c8e6 b6e3 h6h4 c4e6 e8e6 d6d5 a8d8 d5g5 h4g5');
INSERT INTO evaluations VALUES(9305,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9305,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9305,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9305,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9305,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9305,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9305,36,'cp 0','h7h6 g5f6 d8f6 c3d5 f6d8 e1g1 a7a6 c2c3 c5a7 d5e3 d8f6 a2a4 c6e7 b2b4 e8g8 d3d4 e7g6 b4b5 f8e8 e3d5 f6d8 d5e3');
INSERT INTO evaluations VALUES(8809,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(8809,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(8809,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8809,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8809,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8809,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8809,36,'cp 5','g5f6 d8f6 c3d5 f6d8 c2c3 c6e7 d5e3 c5e3 f2e3 e8g8 e1g1 c7c6 c4b3 e7g6 d1e1 a7a5 a2a4 g8h7 b3c2 c8e6 d3d4 d8b6 b2b3 h7h8 c3c4 b6c7 h2h3 b7b5 c2d3 b5b4 e1g3 c6c5 f1f2 c7b6');
INSERT INTO evaluations VALUES(8807,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(8807,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(8807,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(8807,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8807,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8807,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8807,36,'cp 0','d8f6 c3d5 f6d8 e1g1 a7a5 c2c3 c5a7 d5e3 d8f6 f1e1 e8g8 e3d5 f6d8 d5e3');
INSERT INTO evaluations VALUES(499,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(499,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(499,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(499,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(499,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(499,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(499,36,'cp 11','c3d5 f6d8 c2c3 c6e7 d5e3 c5e3 f2e3 e8g8 e1g1 e7g6 d1b3 a8b8 f1f2 d8e7 a1f1 b7b5 c4d5 c8e6 h2h4 a7a5 d5e6 f7e6 g2g3 g6h8 b3d1 a5a4 f3h2 f8f2 f1f2 b8f8 f2f8 e7f8 d1g4 f8f7 g1g2');
INSERT INTO evaluations VALUES(501,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(501,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(501,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(501,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(501,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(501,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(501,36,'cp 0','f6d8 c2c3 a7a6 d3d4 e5d4 c3d4 c5a7 d5e3 d8f6 e3d5');
INSERT INTO evaluations VALUES(8897,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8897,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8897,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8897,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8897,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8897,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8897,36,'cp 0','c2c3 a7a6 e1g1 c5a7 d5e3 e8g8 c4b3 d8f6 e3d5 f6d8');
INSERT INTO evaluations VALUES(32130,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(32130,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32130,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(32130,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(32130,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(32130,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(32130,36,'cp 45','g4f3 d1f3 g8f6 f1c4 b8c6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 a1b1 e8g8 e2g3 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 f8d8 c1g5 d7e6 b4b5 c6b5 b1b5 b7b6 f1c1 d6c5 b5b3 g8h8 b3c3 d8d7 g5f6 e6f6 f3f6 g7f6 g1f1');
INSERT INTO evaluations VALUES(32150,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(32150,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(32150,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(32150,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(32150,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(32150,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(32150,36,'cp 59','d1f3 g8f6 f1c4 b8c6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 e2g3 e8g8 a1b1 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 f8e8 b5c6 b7c6 c1g5 f6d7 b1c1 d6f8 c1c2 h7h6 g5d2 e8d8 g3f5 c6c5 f1d1 d7b6 d2e3 c5c4 d3c4 d8d1 f3d1 b6c4');
INSERT INTO evaluations VALUES(32149,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(32149,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(32149,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(32149,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(32149,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(32149,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(32149,36,'cp 62','g8f6 f1c4 b8c6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 e2g3 e8g8 a1b1 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 f8e8 b5c6 b7c6 f3e2 c6c5 c1e3 a8b8 b1b8 e8b8 f1c1 b8c8 e2c2 d6f8 g3f1 c5c4');
INSERT INTO evaluations VALUES(29954,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(29954,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(29954,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(29954,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(29954,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(29954,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(29954,36,'cp 53','f1c4 b8c6 c3e2 c6a5 c4b5 c7c6 b5d3 d6d5 e1g1 f8d6 a1b1 e8g8 e2g3 d8d7 b2b4 a5c4 d3c4 d5c4 d2d3 c4d3 c2d3 d7e6 b4b5 f8e8 b5c6 b7c6 c1g5 f6d7 b1c1 d6f8 c1c2 a8b8 f1c1 b8b6 g3f5 c6c5 f3g4 g7g6 f5h6 g8g7 g4e6 e8e6 h6g4 e6d6');
INSERT INTO evaluations VALUES(19586,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(19586,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(19586,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(19586,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(19586,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(19586,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(19586,36,'cp 71','f3d4 c5d4 c2c3 a7a6 b5a4 b7b5 a4c2 d4c3 b1c3 c8b7 d2d4 a8c8 c1e3 e7e6 e1g1 g8f6 e4e5 b5b4 e5f6 b4c3 f6g7 f8g7 b2c3 c8c3 d1d2 c3c8 c2d3 e8g8 a2a4 d8h4 f1b1 c8b8 e3f4 g7d4 d3h7 h4h7 d2d4 h7g6');
INSERT INTO evaluations VALUES(19583,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(19583,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(19583,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(19583,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(19583,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(19583,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(19583,36,'cp 68','c5d4 c2c3 a7a6 b5a4 b7b5 a4c2 g7g6 c3d4 f8g7 d4d5 e7e6 b1c3 c8b7 d2d4 e6d5 e1g1 g8e7 f1e1 e8g8 e4e5 d7d6 c1g5 d6e5 d4e5 d8c7 d1d2 a8e8 e5e6 f7e6 e1e6 d5d4 c2b3 d4c3 e6e7 g8h8 e7c7 c3d2 g5d2 e8e2 d2e1 f8e8 e1c3 g7c3 b2c3');
INSERT INTO evaluations VALUES(19580,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(19580,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(19580,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(19580,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(19580,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(19580,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(19580,36,'cp 71','c2c3 a7a6 b5a4 b7b5 a4c2 g7g6 c3d4 f8g7 d4d5 e7e6 b1c3 c8b7 e1g1 g8e7 d2d4 e6d5 f1e1 e8g8 e4e5 d7d6 c1g5 d6e5 d4e5 a8c8 d1d2 b5b4 c3a4 h7h6 g5h4 c8c4 h4f6 g7f6 e5f6 e7c6 a1d1 c4d4 d2h6 d8f6 a4c5 b7c8 c2b3 d4d1 e1d1 c6d4 b3d5 f8d8 d1d4');
INSERT INTO evaluations VALUES(19577,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(19577,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(19577,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(19577,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(19577,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(19577,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(19577,36,'cp 59','a7a6 b5a4 b7b5 a4c2 g7g6 c3d4 f8g7 e1g1 g7d4 b1c3 c8b7 c3e2 d4g7 d2d4 g8f6 f2f3 e8g8 c1e3 f6e8 d1d2 e8d6 b2b3 e7e6 f1c1 f8e8 e2f4 a8c8 f4d3 f7f5 e4e5 d6f7 a2a4 d7d6 a4b5 d6e5 b5a6');
INSERT INTO evaluations VALUES(17531,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(17531,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(17531,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17531,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(17531,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(17531,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(17531,36,'cp 65','b5a4 b7b5 a4c2 g7g6 c3d4 f8g7 d4d5 e7e6 e1g1 g8e7 d2d4 e6d5 e4d5 d7d6 f1e1 h7h6 a2a4 b5b4 c1f4 e8g8 b1d2 c8b7 d2e4 e7f5 c2b3 a6a5 b3c4 g7d4 d1d2 a8c8 a1c1 d8b6 e4g3 f5g3 h2g3 h6h5 e1e7 b7a6 c4a6 c8c1 d2c1 b6a6');
INSERT INTO evaluations VALUES(39609,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(39609,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(39609,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(39609,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(39609,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(39609,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(39609,36,'cp 30','c7c5 d4d5 d7d6 a2a4 a7a6 g1f3 g8f6 b1c3 e8g8 f1c4 e7e6 e1g1 e6d5 c3d5 b8c6 d5f6 d8f6 c2c3 c8e6 f4f5 g6f5 c4e6 f7e6 e4f5 f6f5 d1d6 c6e5 c1e3 e5c4 d6c5 f5c5 e3c5 f8e8 f1e1 c4b2 c5d4 g7d4 c3d4 b2c4 e1e4 b7b5 a4b5 a6b5 a1a8 e8a8 e4e6 a8a1 g1f2');
INSERT INTO evaluations VALUES(37812,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(37812,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(37812,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(37812,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37812,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(37812,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(37812,36,'cp 50','d4d5 d7d6 a2a4 a7a6 g1f3 g8f6 b1c3 e8g8 f1c4 e7e6 e1g1 e6d5 c3d5 b8c6 d5f6 d8f6 c2c3 c8e6 f4f5 g6f5 c1g5 f6g6 e4f5 g6f5 c4d3 f5g4 d1d2 f7f6 g5f4 c6e5 h2h3 g4h5 d3e4 e5c4 d2e2 f8e8 e2f2 g8h8 a1e1 d6d5 e4d3 e6f7 d3c4 e8e1 f1e1 d5c4 f4d6 a8e8 e1e8 f7e8');
INSERT INTO evaluations VALUES(47157,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(47157,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(47157,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(47157,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(47157,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(47157,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(47157,36,'cp -19','c7c5 e2e3 b8c6 g1f3 d8b6 b1c3 c5c4 a1b1 b6a5 f3d2 c8f5 g2g4 f5g6 g4g5 f6e4 f1g2 e4d2 d1d2 e7e6 a2a3 f8e7 h3h4 g6f5 h4h5 h7h6 g5h6 g7h6 g2h3 f5h3 h1h3 e8c8 c3e2 a5a6 d2c3 d8g8 b2b4 c4b3 b1b3 a6a5 f4e5 f7f6 e5h2 a5c3 b3c3 c8d7 e2f4 g8g5 f4g6 h8g8 e1f1');
INSERT INTO evaluations VALUES(45363,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(45363,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45363,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45363,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45363,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(45363,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(45363,36,'cp 0','e2e3 d8b6 b1c3 c8d7 a1b1 e7e6 g1f3 d7c6 f1e2 b8d7 e1g1 f8e7 a2a3 c5c4 b2b4 c4b3 b1b3 b6a5 e2d3 a7a6 a3a4 e7b4 c3a2 b4e7');
INSERT INTO evaluations VALUES(45358,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(45358,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(45358,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(45358,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(45358,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(45358,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(45358,36,'cp -18','d8b6 b1c3 c8d7 a1b1 e7e6 g1f3 d7c6 f1e2 b8d7 e1g1 f8e7 a2a3 c5c4 b2b4 c4b3 b1b3 b6a5 e2b5 a8c8 b5d3 e8g8 d1e2 a7a6 a3a4 f6e4 d3e4 d5e4 f3e5 d7e5 f4e5 f7f6 e2g4 c6d7 b3b7 f6e5 b7d7 a5c3 g4e6 f8f7');
INSERT INTO evaluations VALUES(17891,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(17891,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(17891,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(17891,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(17891,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(17891,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(17891,36,'cp 30','f8c5 c3d5 d7d6 c2c3 c5a7 d2d4 e5d4 c3d4 g8e7 e1g1 c8g4 c1e3 e7d5 c4d5 e8g8 a1c1 c6b4 d5b3 b4c6 b3a4 a7b6 a4c6 b7c6 h2h3 g4h5 c1c6 d8e8 d4d5 b6e3 f2e3 e8e4 c6c7 e4e3 f1f2 f8e8 d1d2 h7h6 c7c6 e3d2 f3d2 e8e1 g1h2');
INSERT INTO evaluations VALUES(4857,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(4857,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(4857,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(4857,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(4857,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(4857,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(4857,36,'cp 39','c3d5 g8f6 c2c3 h7h6 d2d4 e5d4 c3d4 c5a7 d5f6 d8f6 e4e5 f6f5 c4d3 f5e6 c1e3 d7d6 e1g1 d6e5 f1e1 e6d5 d1c2 e8g8 d3c4 d5d8 d4e5 a7e3 e1e3 c8e6 c4e6 f7e6 c2b3 d8d5 b3b7 a8b8 b7a6 b8b2 a6a3 b2c2 e3d3 d5c5 a3c5 c2c5 d3d7 c6e5');
INSERT INTO evaluations VALUES(52842,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(52842,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(52842,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(52842,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(52842,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52842,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(52842,36,'cp 21','g1f3 c8g4 e1g1 g8f6 h2h3 g4h5 d2d4 e7e6 c2c4 b8d7 d1b3 d8b6 c4c5 b6a6 b1c3 f8e7 c1f4 b7b6 c5b6 a6b6 c3a4 b6a6 f1c1 e8g8 g3g4 h5g6 e2e3 f8c8 g2f1 a6a5 b3d1 h7h5 a2a3 h5g4 b2b4 a5d8 h3g4 f6g4 f1a6 g6f5 d1e2 g8h8 a6c8 a8c8');
INSERT INTO evaluations VALUES(52841,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52841,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(52841,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52841,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52841,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52841,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(52841,36,'cp 26','c8g4 e1g1 g8f6 h2h3 g4f3 e2f3 e7e6 f3f4 f8e7 d2d3 e8g8 b1d2 a7a5 d2f3 b7b5 f1e1 b8d7 d1e2 f8e8 c2c3 a5a4 f3e5 d8b6 c1d2 e7f8 h3h4 c6c5 e5d7 f6d7 f4f5 c5c4 d3c4 b5c4 f5e6 e8e6 d2e3');
INSERT INTO evaluations VALUES(31136,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31136,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(31136,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(31136,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31136,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(31136,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(31136,36,'cp 17','e1g1 b8d7 d2d4 e7e6 c2c4 g8f6 h2h3 g4h5 c4d5 e6d5 f3h4 h5g6 h4g6 h7g6 h3h4 f8e7 b1c3 e8g8 d1b3 d8b6 e2e3 e7d6 c1d2 f8e8 f1c1 f6e4 c3e4 d5e4 b3b6 d7b6 c1c2 f7f5 a1c1 b6d5 g2f1 a7a5');
INSERT INTO evaluations VALUES(31135,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31135,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(31135,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(31135,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31135,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(31135,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31135,36,'cp 18','b8d7 d2d4 g8f6 c2c4 e7e6 d1b3 d8b6 b1c3 f8e7 h2h3 g4f3 g2f3 e8g8 f1d1 b6b4 c4d5 c6d5 c1f4 b4b3 a2b3 f8c8 d1c1 h7h6 h3h4 c8c6 e2e4 d5e4 c3e4 c6c1 f4c1 f6d5 c1d2 a7a6 e4c3 d7b6 c3a4 b6a4 b3a4 a8c8 f3d5 e6d5');
INSERT INTO evaluations VALUES(4184,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(4184,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(4184,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(4184,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(4184,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(4184,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(4184,36,'cp 32','d3d4 c5b6 d4e5 f6g4 c1f4 d8e8 d1d5 c6e7 d5d2 e7g6 c4b3 g4e5 f3e5 g6e5 a2a4 a7a5 f4e3 b6e3 d2e3 d7d6 f2f4 e5g4 e3d4 b7b6 c3c4 e8c6 b1c3 g4f6 e4e5 d6e5 f4e5 f6d7 c3d5 c6c5 d4c5 d7c5 b3c2 c8e6 d5c7 a8d8 c7e6');
INSERT INTO evaluations VALUES(4202,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(4202,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(4202,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(4202,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(4202,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(4202,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(4202,36,'cp 22','c5b6 d4e5 f6g4 c1f4 d8e8 e5e6 f7e6 f4g3 g8h8 h2h3 g4f6 e4e5 f6e4 g1h2 d7d5 e5d6 c7d6 b1d2 e4g3 f2g3 d6d5 c4b3 b6c7 f3h4 c6e7 f1f8 e8f8 d1h5 c8d7 a1f1 f8e8 h5e2 e7f5 h4f5 e6f5 e2f3 f5f4 g3f4 d7b5');
INSERT INTO evaluations VALUES(14345,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(14345,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(14345,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(14345,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(14345,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(14345,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14345,36,'cp 34','d2d4 c8f5 e1g1 e7e6 c2c4 f8e7 f3e5 e8g8 b1c3 h7h6 c1f4 d5c4 e5c4 d8d4 d1d4 c6d4 a1d1 a8d8 e2e3 b7b5 c4a5 b5b4 e3d4 b4c3 b2c3 f6d5 f4d2 e7a3 c3c4 d5b6 d2f4 d8c8 g2b7 f5h3 f1e1 a3b4 a5c6 b4e1');
INSERT INTO evaluations VALUES(24817,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24817,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24817,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24817,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(24817,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24817,36,'cp 13','e7e6 b1d2 f8d6 f4g3 e8g8 f1e2 f8e8 e1g1 d6g3 h2g3 e6e5 d4e5 c6e5 d1b3 e5f3 e2f3 a8b8 f1d1 d8e7 c3c4 d5c4 d2c4 h7h5 a1c1 a7a6 b3c3 f5e6 a2a4 e6c4 c3c4 c7c6 b2b4 g7g6 b4b5 c6b5 a4b5 a6b5 c4b5 e8d8 b5g5 d8d1 c1d1');
INSERT INTO evaluations VALUES(25154,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(25154,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(25154,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(25154,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25154,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(25154,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(25154,36,'cp 7','f1e2 f8d6 f4g3 e8g8 b1d2 f8e8 f3h4 c6e7 e1g1 e7g6 h4f5 e6f5 g3d6 d8d6 g2g3 c7c6 c3c4 g6e7 a2a4 a7a5 d1b3 g7g6 a1c1 d6c7 d2f3 g8g7 f1d1 a8d8 f3e5 d5c4 e2c4 e7d5 c4d5 d8d5 e5d3 e8e7 c1c5 d5c5 d3c5 g7g8 d1c1 g8g7 c5d3 c7d6');
INSERT INTO evaluations VALUES(56355,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(56355,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(56355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56355,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56355,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56355,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(56355,36,'cp -22','e5d4 e3d4 f8b4 b1c3 g8f6 f1e2 b8c6 g1f3 d5c4 e1g1 e8g8 e2c4 c8g4 a2a3 b4a5 c1e3 d8d6 a1c1 g4h5 h2h3 a8e8 b2b4 a5b6 c3a4 c6d4 a4b6 d4f3 g2f3 d6d1 f1d1 a7b6 g1g2 e8e7 a3a4 f8e8 a4a5 b6a5 b4a5 e8a8 e3g5 e7e8 g5f6 g7f6');
INSERT INTO evaluations VALUES(56193,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(56193,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(56193,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(56193,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(56193,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56193,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(56193,36,'cp -9','e3d4 f8b4 b1c3 g8f6 f1d3 e8g8 g1e2 d5c4 d3c4 b8d7 c1g5 d7b6 c4d3 f8e8 e1g1 h7h6 g5h4 c7c6 f1e1 b4e7 d3c2 f6d5 h4g3 c8e6 a2a3 d5c3 e2c3 e7f6 d1d3 g7g6 c3e4 e6f5 e4f6 d8f6 e1e8 a8e8 d3f3 f6g5 c2f5 g5f5 f3f5');
INSERT INTO evaluations VALUES(27203,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(27203,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(27203,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(27203,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(27203,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(27203,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(27203,36,'cp 42','f8e7 f1e1 g8f6 d2d4 c6d4 b5d7 d8d7 f3d4 e5d4 d1d4 e8g8 c1e3 c7c6 a1d1 f8d8 a2a4 d7c7 h2h3 h7h6 d4c4 a8c8 c4b3 b7b6 b3c4 c7b7 b2b3 d6d5 e4d5 f6d5 c3d5 d8d5 c4g4 e7f6 d1d5');
INSERT INTO evaluations VALUES(43018,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(43018,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(43018,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(43018,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(43018,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(43018,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(43018,36,'cp -5','b8c6 e5f4 e8g8 f1e2 d7d5 g1f3 d5e4 d3e4 f8e8 b1c3 e7d6 f4h4 h7h6 e1g1 c6e7 e4e5 e7g6 h4d4 g6e5 c1f4 e5f3 e2f3 d6f4 d4f4 c7c6 f1e1 c8e6 c3e4 f6e4 f3e4 d8g5 f4g5 h6g5 a2a3 a8d8 f2f3 g8f8 a1d1 e6d5 e4d5 e8e1 d1e1 d8d5 g1f2 d5d2 e1e2');
INSERT INTO evaluations VALUES(10870,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(10870,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(10870,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(10870,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(10870,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(10870,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(10870,36,'cp -9','e5f4 d7d5 g1f3 e8g8 f1e2 d5e4 d3e4 f8e8 b1c3 e7d6 f4h4 h7h6 e1g1 c6e7 e4e5 e7g6 h4d4 g6e5 c1f4 c7c5 d4d2 d8b6 b2b3 c8f5 f3e5 d6e5 f4e5 e8e5 a1d1 e5e8 e2c4 a8d8 d2c1 a7a6 c3d5 f6d5 c4d5 f5e6 d5e6 b6e6 d1d8 e8d8 f1e1 e6c6');
INSERT INTO evaluations VALUES(10950,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(10950,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(10950,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(10950,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(10950,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(10950,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(10950,36,'cp -9','e8g8 g1f3 d7d5 f1e2 d5e4 d3e4 f8e8 b1c3 e7d6 f4h4 h7h6 e1g1 c6e7 e4e5 e7g6 h4d4 g6e5 c1f4 c7c5 d4d2 d8b6 b2b3 c8f5 f4e5 d6e5 f3e5 e8e5 a1d1 e5e8 e2c4 a8d8 d2c1 f5e6 c4e6 b6e6 d1d8 e8d8 f1d1 d8d1 c1d1 c5c4 h2h3 e6e5 d1d2 c4b3 c2b3');
INSERT INTO evaluations VALUES(44071,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(44071,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(44071,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(44071,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(44071,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(44071,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(44071,36,'cp 8','d7d5 c4b3 f8b4 e1g1 e8g8 f3e5 f8e8 d2d4 b4c3 b2c3 f6e4 f1e1 c8e6 c3c4 b8d7 e5d7 d8d7 c4c5 e6f5 f2f3 e4c3 d1d2 c3b5 e1e8 a8e8 a2a4 b5c7 d2f2 c7e6 a4a5 e6c7 c1f4 a7a6 f4g3 c7b5 f2d2 f7f6 b3a4 f5g6 c2c3 h7h6 a1e1 e8e1 d2e1 d7f7 a4b5 a6b5');
INSERT INTO evaluations VALUES(45800,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(45800,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45800,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(45800,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45800,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45800,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(45800,36,'cp 19','c4b3 f8b4 e1g1 e8g8 d2d3 b4c3 b2c3 d8c7 e4d5 c6d5 c1g5 b8d7 d1d2 h7h6 g5h4 b7b6 h4g3 a7a5 a2a4 f8e8 h2h3 c8b7 f1e1 a8c8 a1b1 e8e7 f3d4 c7c3 d2c3 c8c3 d4b5 c3c6 d3d4 e5e4 b5d6 b7a6 d6f5 e7e6 f5e3 e6e8 e3d5');
INSERT INTO evaluations VALUES(45818,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(45818,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(45818,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45818,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45818,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(45818,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(45818,36,'cp 25','f8b4 e1g1 e8g8 d2d3 a7a5 a2a4 d5e4 c3e4 f6e4 d3e4 b8d7 b3c4 h7h6 c1e3 d8e7 d1e2 b4d6 b2b3 d7f6 f3d2 f8d8 e3b6 d8e8 a1d1 f6d7 b6e3 d7f8 g1h1 c8e6 c4e6 f8e6 d2c4 d6c7 g2g3 b7b5 c4b2 b5a4');
INSERT INTO evaluations VALUES(53334,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(53334,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(53334,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(53334,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(53334,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(53334,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(53334,36,'cp -65','b8c6 g1f3 g8h6 f1b5 e7e6 c2c3 c8d7 b1a3 f8e7 e1g1 d8b6 b5e2 e8g8 d1b3 b6c7 d2d4 f7f6 g1h1 f6e5 f4e5 h6f5 b3d1 c5d4 a3b5 c7b6 b5d4 f5d4 c3d4 d7e8 c1e3 e8g6 e2d3 g6d3 d1d3 b6b2 f1b1 b2a3 d3a3');
INSERT INTO evaluations VALUES(19824,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(19824,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(19824,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(19824,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(19824,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(19824,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(19824,36,'cp -51','g1f3 g8h6 f1b5 c8d7 b1a3 e7e6 c2c3 d8b6 b5e2 f7f6 d1b3 b6d8 e1g1 h6f5 b3d1 f6e5 f4e5 f8e7 a3c2 e8g8 d2d4 c5d4 c2d4 d8b6 g1h1 f5d4 c3d4 d7e8 c1e3 b6b2 a1b1 b2a2 b1a1 a2b2 a1b1 b2a3 b1b3 a3a5 e2d3 e8g6 d3g6 h7g6 b3b7 a5a6 b7c7');
INSERT INTO evaluations VALUES(44736,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(44736,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(44736,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(44736,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(44736,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(44736,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(44736,36,'cp 96','e4e5 d6e5 f3e5 e7e6 c4b5 b8d7 d1f3 f8g7 b5d7 c8d7 f3b7 e8g8 e1g1 d8b6 b7f3 b6c7 e5d7 c7d7 d2d3 f6d5 c3e4 f7f5 e4d2 e6e5 d2c4 d5b6 c4b6 a7b6 f1e1 h7h6 c1e3 g8h7 a2a3 d7a4 a1c1 a4d7 h2h3 f8e8 c1d1 d7a4 c2c3 a8d8 f3e2 a4b3');
INSERT INTO evaluations VALUES(23373,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(23373,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(23373,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(23373,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(23373,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(23373,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(23373,36,'cp 84','e7g6 h2h4 h7h5 c1g5 g8f6 b1c3 f8b4 d1d3 b4c3 d3c3 f6e4 g5d8 e4c3 d8c7 c3d5 c7d6 f7f6 e1c1 d5e7 f1d3 a7a5 h1e1 g6f4 d3b5 f4e6 f3d2 e7f5 d2c4 e6d4 a2a4 e8f7 f2f4 d4b5 a4b5 e5f4 c4b6 f5d6 d1d6 a8b8 b6d7 c8d7 d6d7 f7g6 e1e7 h8e8 e7g7 g6f5 d7b7 b8b7 g7b7');
INSERT INTO evaluations VALUES(22991,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(22991,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22991,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(22991,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(22991,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(22991,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(22991,36,'cp 91','h2h4 h7h5 c1g5 g8f6 b1c3 f8b4 d1d3 b4c3 d3c3 f6e4 g5d8 e4c3 d8c7 c3d5 c7e5 g6e5 f3e5 d7d6 e1c1 c8e6 e5f3 e8c8 f3g5 d5f6 f1d3 f6g4 d1d2 h8e8 h1d1 c8c7 b2b3 a7a6 a2a4 g4f6 f2f3 d6d5 c1b2 d8d7 d1e1 d7e7 d2e2 f6d7');
INSERT INTO evaluations VALUES(22987,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(22987,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(22987,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(22987,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(22987,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(22987,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22987,36,'cp 87','h7h5 b1c3 f8b4 d1d3 g8f6 c1g5 b4c3 d3c3 f6e4 g5d8 e4c3 d8c7 c3d5 c7e5 g6e5 f3e5 d7d6 e1c1 c8e6 e5f3 e8c8 f3g5 d5b4 a2a3 b4c6 f1e2 c8c7 h1e1 d8d7 d1d2 e6f5 e2d1 a7a6 d1f3 c6e5 f3e4 f5e4 e1e4 e5c6 g2g3 d6d5 e4e1 g7g6 e1e3');
INSERT INTO evaluations VALUES(22547,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(22547,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22547,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(22547,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(22547,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(22547,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(22547,36,'cp 87','b1c3 g8f6 c1g5 f8b4 d1d3 b4c3 d3c3 f6e4 g5d8 e4c3 d8c7 c3d5 c7e5 g6e5 f3e5 d7d6 e1c1 c8e6 e5f3 e8c8 f3g5 h8f8 d1d4 d5f6 f1d3 c8c7 h1d1 f8e8 f2f3 e8e7 a2a4 a7a6 b2b4 f6d7 g5e4 d7e5 b4b5 d6d5 e4c3 c7b6 d3e2 e7d7 a4a5 b6a5 b5a6');
INSERT INTO evaluations VALUES(21153,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(21153,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(21153,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(21153,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(21153,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(21153,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(21153,36,'cp 109','d2d4 e7e6 f1b5 f8d6 e1g1 g8e7 f1e1 a7a6 b5f1 e8g8 c3e4 e7g6 e4d6 c7d6 b2b3 h7h6 c2c4 d6d5 c1b2 c8d7 f1d3 d8f6 d1d2 f8d8 c4c5 g6h4 f3h4 f6h4 a2a3 c6e7 d2e2 h4f6 b2c3 a6a5 b3b4 a5b4 a3b4 a8a1 e1a1');
INSERT INTO evaluations VALUES(21128,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(21128,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(21128,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(21128,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(21128,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(21128,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(21128,36,'cp 123','e7e6 f1b5 f8d6 e1g1 g8e7 f1e1 e8g8 c3e4 h7h6 c2c4 e7g6 c1d2 c8d7 b5a4 b7b6 a1c1 d6b4 d2b4 c6b4 a4b3 a8b8 a2a3 b4c6 d1d2 c6e7 d4d5 e7f5 b3c2 e6d5 c4d5 f5d6 e4g3 d8f6');
INSERT INTO evaluations VALUES(21507,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(21507,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(21507,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(21507,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(21507,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(21507,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(21507,36,'cp 98','f1b5 f8d6 e1g1 g8e7 f1e1 e8g8 c3e4 h7h6 c2c4 e7g6 c1d2 c8d7 b5a4 b7b6 d2c3 c6a5 c4c5 d6e7 a4c2 d7c6 e4g3 c6d5 c2g6 f7g6 f3e5 g6g5 d1d3 a7a6 c3a5 b6a5 g3e4 a8b8 b2b3 a5a4 b3a4 b8b2');
INSERT INTO evaluations VALUES(21350,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(21350,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(21350,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(21350,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(21350,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(21350,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(21350,36,'cp 118','g8f6 e1g1 f8e7 a2a3 e8g8 f1e1 a7a5 h2h3 h7h6 b2b3 e7d6 b5d3 c6e7 c1b2 c8d7 f3e5 e7f5 c3e2 d7c6 e2f4 f5h4 e5c6 b7c6 f4h5 a8b8 d1e2 f6h5 e2h5 d8f6 h5g4 h4f5 g4e4 f8e8 a1b1 e8d8 e4c6 f5d4 c6e4 d6c5');
INSERT INTO evaluations VALUES(57174,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(57174,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(57174,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(57174,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(57174,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(57174,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(57174,36,'cp 76','f4e5 d7d6 e5d6 f8d6 g1f3 g8f6 g2g3 e8g8 f1g2 f8e8 e1g1 c7c5 b1a3 d6c7 b2b3 b8c6 c1b2 a8b8 b2f6 d8f6 a3c4 f6d8 a2a4 h7h6 e2e3 c8g4 d1e1 c6b4 a1c1 a7a6 f3h4 b7b5 a4b5 a6b5 e1f2 b5c4 f2f7 g8h8');
INSERT INTO evaluations VALUES(57108,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(57108,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(57108,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(57108,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(57108,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(57108,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(57108,36,'cp 63','d7d6 e5d6 f8d6 g1f3 g8f6 g2g3 b8c6 f1g2 h7h5 d2d3 h5h4 g3h4 f6g4 b1c3 c8f5 e2e4 f5e6 c1g5 f7f6 g5d2 d6c5 d1e2 c5f2 e1f1 f2b6 c3d5 g4e5 d2c3 e6g4 h2h3 g4f3 g2f3 h8h4 f1e1 d8d6 e1d1 e8c8 d5b6 a7b6 e2f2');
INSERT INTO evaluations VALUES(55522,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(55522,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(55522,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(55522,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(55522,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(55522,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(55522,36,'cp 64','g1f3 b8c6 e5d6 f8d6 b1c3 g8f6 g2g3 c8g4 f1g2 d8d7 e1g1 e8c8 d2d4 h8e8 a2a3 h7h5 d1d3 g4f5 e2e4 f6e4 f3h4 e4c3 h4f5 c3e2 g1h1 e2c1 a1c1 d6f8 c2c3 g7g6 f5h4 f8h6 c1d1 a7a6 f1f6 e8e6 f6e6 d7e6 h4f3 e6d7 d1e1 c8b8');
INSERT INTO evaluations VALUES(55520,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(55520,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(55520,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(55520,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(55520,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(55520,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(55520,36,'cp 65','d6e5 f3e5 g8f6 e5f3 b8c6 d2d4 f8d6 c1g5 e8g8 c2c3 c8f5 g2g3 d6e7 f1g2 f6g4 g5c1 d8d7 e1g1 a8e8 h2h3 g4f6 g3g4 f5g6 f3h4 g6e4 b1d2 e4g2 h4g2 e7d6 f1f6 g7f6 e2e4 c6e7 d2c4 e7g6 d1f3 e8e6 d4d5 g6e5 c4e5 e6e5 c1h6 d7e7 h6f8 g8f8 a1e1 h7h5 g2h4 h5g4');
INSERT INTO evaluations VALUES(56635,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(56635,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(56635,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(56635,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(56635,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(56635,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(56635,36,'cp 83','f3e5 f8d6 e5f3 g8f6 g2g3 e8g8 f1g2 c7c5 d2d3 b8c6 b1c3 c8g4 e1g1 h7h6 c1e3 d8e7 d1d2 f8e8 e3f2 a8d8 a1e1 c6e5 g1h1 e7d7 f2g1 e5f3 e2f3 e8e1 f1e1 g4e6 b2b3 d6f8 d2e2 b7b6 g1f2 d8e8 e2d2 e8d8 h1g1 b6b5 d2e2');
INSERT INTO evaluations VALUES(55485,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(55485,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(55485,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(55485,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(55485,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(55485,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(55485,36,'cp 155','f8d6 b1c3 g8f6 c3b5 e8g8 b5d6 c7d6 e2e3 c8f5 g1f3 b8c6 f1e2 f6e4 e1g1 d6d5 f3d4 f5g6 d4c6 b7c6 c2c4 d5c4 e2c4 d8b6 f2f3 f8d8 d1e1 e4c5 e3e4 c5d3 e1e3 b6e3 c1e3 d3b2 c4a6 b2d3 f1b1 f7f6 b1b3 d3e5 a6e2 h7h5 h2h4');
INSERT INTO evaluations VALUES(38650,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(38650,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(38650,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(38650,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(38650,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(38650,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(38650,36,'cp 159','b1c3 g8f6 c3b5 e8g8 b5d6 c7d6 e2e3 c8f5 g1f3 b8c6 f1e2 d8a5 c2c3 d6d5 f3d4 f5g6 e1g1 a8c8 f2f3 f8e8 d1d2 c6e5 f1d1 h7h6 b2b3 a5c3 c1b2 c3c5 a1c1 c5b6 g1f1 a7a6 h2h3 e5c6 d4c6 c8c6 c1c6 b7c6');
INSERT INTO evaluations VALUES(38647,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(38647,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(38647,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(38647,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(38647,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(38647,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(38647,36,'cp 146','b8c6 c3b5 g8f6 b5d6 c7d6 e2e3 c8f5 g1f3 d6d5 f1e2 e8g8 e1g1 f6e4 f3d4 f5g6 d4c6 b7c6 c2c4 d8e7 c4d5 f8d8 f2f3 e4c5 f1f2 d8d5 d1f1 a8d8 e3e4 d5d7 c1f4 c5d3 e2d3 d7d3 f2c2 e7d7 a1e1 d3d1 e1d1 d7d1 f1d1 d8d1 g1f2');
INSERT INTO evaluations VALUES(41711,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41711,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(41711,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(41711,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41711,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(41711,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(41711,36,'cp 22','e8g8 c1g5 c7c6 b1d2 c8e6 f1e1 h7h6 g5h4 b8d7 c2c3 d8c7 h4g3 d6g3 h2g3 f8e8 d1c2 c6c5 d3b5 b7b6 d2f1 a7a6 b5d7 c7d7 f1e3 a8d8 e1e2 d7b7 a1e1 f6e4 f3e5 f7f6 e5d3 e6f7 e3f5 c5c4');
INSERT INTO evaluations VALUES(36944,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36944,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36944,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36944,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36944,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36944,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36944,36,'cp 8','c1g5 c7c6 b1d2 c8e6 g5h4 b8d7 f3g5 d8c7 g5e6 f7e6 d2f3 e6e5 f3e5 d7e5 d4e5 d6e5 h4g3 e5g3 h2g3 c7b6 c2c4 d5c4 d3c4 g8h8 d1c2 b6c5 a1c1 a8e8 c4e2 c5c2 c1c2 g7g6 f2f3 h8g7 f1d1 e8d8 d1d8');
INSERT INTO evaluations VALUES(36925,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36925,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36925,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(36925,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(36925,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(36925,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(36925,36,'cp 9','c7c6 b1d2 c8e6 f1e1 h7h6 g5h4 b8d7 c2c3 e6g4 h4g3 g4f3 d1f3 d8c7 d2f1 a8e8 g3d6 c7d6 f3g3 d6g3 f1g3 d7b6 g3f5 e8e1 a1e1 f8e8 e1e8 f6e8 b2b3 g8f8 g1f1 b6c8 f1e2 c8e7 f5e3 g7g6 g2g3 f7f5 e2d2 f8f7 e3g2');
INSERT INTO evaluations VALUES(36117,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(36117,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36117,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36117,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(36117,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(36117,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(36117,36,'cp 13','b1d2 c8e6 c2c4 h7h6 g5h4 b8d7 d1b3 b7b6 f1e1 f8e8 b3c2 d8c7 h4g3 c6c5 c4d5 e6d5 e1e8 f6e8 d3h7 g8h8 h7e4 d5e4 d2e4 d6g3 h2g3 a8c8 a1c1 c5d4 c2c7 c8c7 c1c7 e8c7 f3d4 d7e5 e4d6 h8g8 g1f1 g8f8 d4f5 e5d3 f1e2 d3b2');
INSERT INTO evaluations VALUES(43406,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43406,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(43406,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(43406,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(43406,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(43406,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43406,36,'cp 1','e7e5 d4b3 c8e6 c1e3 h7h5 c3d5 e6d5 e4d5 g7g6 d1d2 b8d7 e1c1 d7b6 c1b1 b6d5 e3g5 f8e7 c2c4 d5b6 b3a5 d8d7 g5h6 a8d8 h6g7 h8g8 g7f6 e7f6 f1d3 d7c7 d3e4 b6c4 a5c4 c7c4 e4b7 c4e6 b7d5 e6f5 d5e4 f5g5 d2e2 e8f8 h2h4 g5g3 e2a6 g3g2 a6a7');
INSERT INTO evaluations VALUES(43450,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(43450,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(43450,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(43450,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(43450,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(43450,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43450,36,'cp 27','d4b3 c8e6 c1e3 h7h5 c3d5 e6d5 e4d5 g7g6 d1d2 f8g7 e1c1 b8d7 c1b1 b7b5 c2c4 b5c4 f1c4 d7b6 c4e2 b6d5 e3g5 e8g8 d2d5 f6d5 g5d8 d5f4 d8b6 f4e2 d1d6 f8e8 b3c5 e8c8 c5e4 g7f8 d6d2 a8b8 d2e2 b8b6 h1c1 c8b8 c1c7 a6a5');
INSERT INTO evaluations VALUES(43460,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(43460,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(43460,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43460,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(43460,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43460,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(43460,36,'cp 19','c8e6 c1e3 f8e7 d1d2 h7h5 c3d5 f6d5 e4d5 e6f5 f1e2 a6a5 a2a4 e8g8 e1g1 b8d7 f3f4 e5f4 e3f4 d7f6 b3d4 f5g6 e2f3 f8e8 d4b5 a8c8 c2c3 e7f8 d2d4 d8d7 a1e1 e8e1 f1e1 c8e8 e1e8 f6e8 b5a3 e8f6 h2h3 b7b6 g1h2 d7f5 a3c4 f5b1');
INSERT INTO evaluations VALUES(28546,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(28546,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(28546,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(28546,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(28546,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(28546,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(28546,36,'cp 36','c1e3 h7h5 d1d2 b8d7 c3d5 f6d5 e4d5 e6f5 b3a5 f8e7 f1e2 e7h4 g2g3 h4f6 e1g1 f5h3 f1e1 e8g8 a5b7 d8c7 b7a5 e5e4 a5c6 e4f3 e2f3 f6b2 a1b1 b2f6 e3f4 f8e8 f3h5 d7e5 f4e5 f6e5 c2c4 e5f6 h5f3 e8e1 b1e1 c7d7 f3g2 h3g2 g1g2 a8e8 e1e8 d7e8 d2d3 e8e1');
INSERT INTO evaluations VALUES(47320,30,'cp -197',NULL);
INSERT INTO evaluations VALUES(47320,31,'cp -198',NULL);
INSERT INTO evaluations VALUES(47320,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(47320,33,'cp -194',NULL);
INSERT INTO evaluations VALUES(47320,34,'cp -212',NULL);
INSERT INTO evaluations VALUES(47320,35,'cp -201',NULL);
INSERT INTO evaluations VALUES(47320,36,'cp -195','c8g4 d1d2 e4d3 f1d3 b8c6 f2f3 g4h5 d2e2 c6d4 e2f2 e7e5 b1c3 h5g6 g5e4 f6e4 f3e4 f8c5 c1e3 d8d7 e1c1 e8c8 c1b1 f7f6 d3c4 d7c6 c4d5 c6e8 a2a3 c8b8 d1d2 c5b6 h1d1 d8d6 g2g4 h7h5 d1g1 h5g4 g1g4 f6f5 e4f5 g6f5 g4g7 f5c2 b1a2');
INSERT INTO evaluations VALUES(29734,30,'cp -194',NULL);
INSERT INTO evaluations VALUES(29734,31,'cp -174',NULL);
INSERT INTO evaluations VALUES(29734,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(29734,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(29734,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(29734,35,'cp -195',NULL);
INSERT INTO evaluations VALUES(29734,36,'cp -190','d1d2 e4d3 f1d3 b8c6 f2f3 g4h5 d2e2 c6d4 e2f2 h5g6 g5e4 e7e5 b1c3 f6e4 f3e4 f8c5 c1e3 d8d7 e1c1 e8c8 d3c4 f7f6 a2a3 c8b8 c1b1 a7a6 d1d2 d7e7 h1d1 d8d6 c3a4 c5a7 a4c3 h8d8 g2g4 h7h5 f2g2 h5g4 g2g4 e7e8 h2h4 g6h5 g4g7');
INSERT INTO evaluations VALUES(52370,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(52370,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52370,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(52370,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52370,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(52370,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(52370,36,'cp 16','d7d5 g1f3 c8f5 f1d3 f5g6 c2c4 g8f6 e1g1 e7e6 d1c2 b8d7 b1d2 c6c5 d3g6 h7g6 c4d5 f6d5 d4c5 f8c5 d2b3 a8c8 c2e2 e8g8 f1d1 d8c7 c1d2 c5e7 e3e4 d5f6 a1c1 c7b8 g2g3 b8a8 e4e5 f6d5 h2h4 c8c1 d2c1 d7b6 c1g5');
INSERT INTO evaluations VALUES(21855,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(21855,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(21855,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(21855,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(21855,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(21855,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21855,36,'cp 20','a7a6 b5c6 b7c6 d3d4 e5d4 d1d4 g8f6 b1c3 f8e7 c1f4 e8g8 e1c1 c6c5 d4d2 f6h5 f4g5 e7g5 f3g5 h7h6 g5f3 c8b7 h1e1 f8e8 e4e5 b7f3 g2f3 d8h4 c3d5 d6e5 d5c7 e8d8 c7d5 h5f6 e1e5 a8b8 d2e3 f6d5 e5d5 d8d5 d1d5 h4h2 d5c5 h2h1 c1d2');
INSERT INTO evaluations VALUES(17780,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(17780,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(17780,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17780,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(17780,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(17780,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(17780,36,'cp 16','b5c6 b7c6 d3d4 e5d4 d1d4 g8f6 b1c3 f8e7 c1f4 e8g8 e1c1 f6d7 h2h4 d7c5 d4e3 c8g4 d1d2 g4f3 e3f3 c5e6 f4e3 d8c8 f3g4 c8b7 c3d1 a8b8 e3g5 f8e8 g5e7 e8e7 h1h3 e6c5 f2f3 g8h8 h3g3');
INSERT INTO evaluations VALUES(17750,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(17750,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(17750,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(17750,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(17750,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(17750,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(17750,36,'cp 11','b7c6 d3d4 e5d4 d1d4 g8f6 b1c3 f8e7 e4e5 d6e5 d4d8 e7d8 f3e5 f6d5 c1d2 d8f6 e1c1 f6e5 d1e1 f7f6 f2f4 d5c3 d2c3 c8e6 f4e5 f6f5 h1f1 c6c5 b2b3 e8c8 f1f3 c5c4 b3c4 d8d7 a2a3 h8d8 f3h3 g7g6 h3h4 c8b7 g2g3 d8e8 h4h6 e6c4');
INSERT INTO evaluations VALUES(18014,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(18014,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18014,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(18014,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(18014,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(18014,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(18014,36,'cp 12','d3d4 e5d4 d1d4 g8f6 b1c3 f8e7 c1f4 e8g8 e1c1 f6d7 h2h4 d7c5 d4d2 c8g4 d1e1 c5e6 f4e3 f7f5 e4f5 g4f5 c3e2 f5e4 f3d4 e6d4 e2d4 d8d7 f2f3 e4d5 e3g5 a8e8 d4e2 d5f7 g5e7 e8e7 e2c3 f8e8 b2b3 c6c5 e1e7 d7e7');
INSERT INTO evaluations VALUES(40962,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(40962,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40962,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40962,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40962,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(40962,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(40962,36,'cp 5','b1c3 e8g8 b2b3 c7c5 c1b2 c5d4 f3d4 d5c4 f1c4 a7a6 e1g1 b7b5 c4e2 a8a7 a1c1 a7d7 d1c2 c8b7 e2f3 d8b6 a2a4 b7f3 d4f3 b8c6 c2e2 b5a4 c3a4 b6b3 c1c6 b3a4 b2f6 a4c6 f3e5 c6c7 e5d7 c7d7 f6e7 d7e7 e2a6 e7f6');
INSERT INTO evaluations VALUES(22541,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(22541,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(22541,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(22541,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(22541,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(22541,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(22541,36,'cp 55','g8f6 e4e5 d7d5 c4b5 f6d7 b5c6 b7c6 f3d4 d7b6 d4c6 d8d7 c6d4 c7c5 e5e6 d7b7 e6f7 e8f7 d4e2 f8e7 a2a4 a7a5 e2f4 h8f8 b1c3 d5d4 c3b5 f7g8 c2c3 d4c3 b5c3 c8f5 c1e3 e7g5 e3c5 g5f4 d1b3 f8f7 b3b6 b7b6 c5b6 f7b7 c3d5 f4e5 f1e1 e5b2 a1a2 b2c3 d5c3 b7b6 a2d2 a8b8 c3b5');
INSERT INTO evaluations VALUES(17656,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(17656,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17656,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(17656,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(17656,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17656,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(17656,36,'cp 36','g8f6 a3c4 b7b5 c4e3 f8e7 f1d3 c6b4 d3e2 e8g8 e1g1 c8e6 a2a3 b4c6 e3d5 a8b8 d5f6 e7f6 c3d5 f6g5 c1g5 d8g5 c2c3 f8c8 g2g3 g5d8 a3a4 h7h6 f1e1 b8b7 d5e3 c6a5 a4b5 a6b5 e3f5 e6f5 e4f5 d6d5 f5f6 a5c4 f6g7 c4b2');
INSERT INTO evaluations VALUES(11240,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(11240,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(11240,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11240,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(11240,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(11240,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(11240,36,'cp 34','a3c4 b7b5 c4e3 f8e7 f1d3 c6b4 d3e2 c8e6 e1g1 e8g8 a2a3 b4c6 e3d5 a8b8 c1e3 e6d5 c3d5 f6e4 e2f3 e4c5 f1e1 b8c8 a3a4 c5a4 b2b3 a4c5 e3c5 d6c5 a1a6 e7g5 g2g3 f7f5 d5b6 e5e4 b6c8 c6d4 f3e4 f5e4 c8d6 f8f6 h2h4 f6d6 a6d6 d8d6 h4g5 d4f3 g1g2 f3e1 d1e1 d6d5');
INSERT INTO evaluations VALUES(23463,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23463,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(23463,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23463,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23463,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23463,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23463,36,'cp 0','d7d5 d2d4 c8f5 c1f4 e7e6 e2e3 f8d6 f4g3 h7h5 a2a3 h5h4 g3d6 c7d6 h2h3 g8f6 f1d3 c6e7 e1g1 a8c8 a1c1 d8b6 c3a4 b6c6 a4c3');
INSERT INTO evaluations VALUES(21071,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21071,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21071,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21071,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21071,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21071,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21071,36,'cp 4','d2d4 c8f5 c1f4 e7e6 e2e3 f8d6 f4g3 g8f6 f1d3 c6e7 d1e2 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 f5e4 e2e4 d8d7 e1g1 b7b5 f1e1 h7h6 f3e5 d6e5 e4e5 e7d5 c2c3 f8e8 e5e2 d7c6 a2a3 c6c4');
INSERT INTO evaluations VALUES(51786,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(51786,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(51786,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(51786,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(51786,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(51786,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(51786,36,'cp 36','d7d5 e2e4 d5e4 c3e4 c6c5 e4c5 b8c6 c1e3 e6e5 d4e5 d8a5 d1d2 c6b4 c5e4 c8f5 a1d1 b4c2 e1e2 a5e5 d2c2 f5e4 c2c3 e5c3 b2c3 g8e7 g1f3 e7f5 f3d4 f5e3 e2e3 e4c6 d4c6 b7c6 e3f3 f8c5 f1d3 e8c8 h1e1 f7f6 h2h4 c8c7 h4h5 h7h6 d3f5 d8d1 e1d1 h8d8 d1e1 d8d2 e1e2 d2e2 f3e2 c5b6 f2f3');
INSERT INTO evaluations VALUES(32165,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(32165,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(32165,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(32165,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(32165,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(32165,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(32165,36,'cp 62','h2h3 g4f3 d1f3 g8f6 d2d4 e5d4 e4e5 f6d7 f3b7 d7b6 f1d3 d6e5 f4e5 f8b4 b1d2 b4d2 c1d2 d8h4 e1d1 b8d7 e5e6 f7e6 h1f1 h4h5 d1c1 h8f8 f1f8 e8f8 b7c7 f8g8 c7d6 h5e5 d6e5 d7e5 d3e2 a8c8 b2b3 d4d3 e2d1 c8d8 c1b2 d3c2 b2c2 d8d2 c2d2');
INSERT INTO evaluations VALUES(32164,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32164,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(32164,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32164,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(32164,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(32164,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(32164,36,'cp 52','g4f3 d1f3 g8f6 d2d4 e5d4 e4e5 f6d7 f3b7 d7b6 b1d2 d6e5 f4e5 c7c6 f1d3 d8d7 b7d7 b8d7 d2f3 f8b4 e1f2 e8g8 a2a3 b4e7 h1e1 f7f6 e5f6 e7f6 c1f4 d7c5 a1c1 f8e8 f4d6 c5d3 c2d3 b6a4 c1c2 g8f7 d6c5 a4c5 c2c5 e8e3 e1e3 d4e3 f2e2');
INSERT INTO evaluations VALUES(32162,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(32162,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(32162,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(32162,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(32162,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(32162,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(32162,36,'cp 72','d1f3 g8f6 d2d4 e5d4 e4e5 d6e5 f4e5 f6d7 f3b7 d7b6 b1d2 c7c6 f1d3 d8d7 b7d7 b8d7 d2f3 f8b4 e1f2 e8g8 a2a3 b4e7 h1e1 f7f6 e5f6 e7f6 c1f4 d7c5 a1c1 f8e8 e1e8 a8e8 b2b4 c5d3 c2d3 b6a4 c1c6 a4b2 f2g3 b2d3 c6c7 d3f4 g3f4 d4d3 c7d7 e8e2 d7d3 f6b2 b4b5 b2c1 f4f5 g7g6 f5f6 c1b2 f3d4 e2g2');
INSERT INTO evaluations VALUES(32161,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(32161,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(32161,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32161,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(32161,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(32161,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(32161,36,'cp 68','g8f6 d2d4 e5d4 e4e5 f6d7 f3b7 d7b6 b1d2 b8d7 b7e4 f8e7 d2f3 e8g8 f1d3 g7g6 e1g1 d7c5 e4d4 d6e5 d4d8 a8d8 f3e5 c5d3 e5d3 f8e8 g1f2 e7f6 a1b1 c7c5 f2f3 c5c4 d3f2 c4c3 f1d1 c3b2 c1b2 f6b2 b1b2 d8c8 f2e4 e8e6 c2c3 f7f5');
INSERT INTO evaluations VALUES(15480,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(15480,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(15480,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(15480,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(15480,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(15480,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(15480,36,'cp 20','c8g4 h2h3 g4h5 f1e1 f8e7 a2a4 d8d7 c4e2 e8g8 d2d3 h5g6 f3h4 a8d8 h4g6 h7g6 c3d5 f6d5 e4d5 c6d4 e2f1 c7c6 c1e3 d7c7 e3d4 e5d4 g2g3 c6c5 h3h4 d8e8 f1h3 e7f6 h4h5 g6g5 a4a5 e8e5 e1e5 f6e5');
INSERT INTO evaluations VALUES(25302,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(25302,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(25302,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(25302,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(25302,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(25302,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(25302,36,'cp 25','h2h3 g4h5 f1e1 f8e7 a2a4 d8d7 c4e2 e8g8 d2d3 h5g6 f3h4 c6d4 h4g6 d4e2 d1e2 h7g6 e1d1 f8e8 d3d4 e5d4 d1d4 d7e6 c1f4 f6h5 f4h2 e7f6 d4d3 f6e5 e2d2 h5f6 h2e5 e6e5 a1e1 e5c5 d3d4 a8d8 e1d1 e8e7 f2f3');
INSERT INTO evaluations VALUES(25301,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(25301,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(25301,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(25301,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(25301,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(25301,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(25301,36,'cp 18','g4h5 f1e1 f8e7 a2a4 d8d7 c4e2 e8g8 d2d3 h5g6 f3h4 c6d4 h4g6 d4e2 e1e2 h7g6 d3d4 c7c6 c1e3 e5d4 d1d4 a7a6 d4b6 a8b8 a1d1 d7e6 b6b3 e6b3 c2b3 f6d7 e3f4 d7e5 f4g3 b8d8 f2f3 f8e8 g1f1 g6g5');
INSERT INTO evaluations VALUES(25297,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(25297,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(25297,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(25297,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25297,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(25297,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(25297,36,'cp 20','f1e1 f8e7 a2a4 d8d7 c4e2 h7h6 a4a5 a7a6 d2d3 h5g6 c3d5 f6d5 e4d5 c6d8 c2c4 f7f5 c1d2 e8g8 d1a4 c7c6 d2c3 e7f6 f3d2 d7e7 e2f3 a8c8 g2g3 g8h8 a4d1 e7d7 f3g2 g6e8 f2f4 e5f4 g3f4 f6c3 b2c3 c6d5');
INSERT INTO evaluations VALUES(22811,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(22811,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(22811,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(22811,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(22811,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(22811,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(22811,36,'cp -62','b1c3 f8c5 f1c4 d7d6 d2d3 a7a6 c3d5 g8e7 c1g5 f7f6 g5d2 e7d5 c4d5 f6f5 f3f4 d8h4 g2g3 h4e7 g1f3 f5e4 d3e4 e5f4 d2f4 c8g4 h2h3 g4h5 h1f1 c6b4 d5b7 a8b8 b7d5 c7c6 d5b3 e7e4 d1e2 e4e2 e1e2 b4d5 e2d2 e8d7 a1e1 h7h6 f3h4 d5f4 f1f4 b8e8 g3g4 e8e1 d2e1 h8e8 e1d1 g7g5 f4f5');
INSERT INTO evaluations VALUES(22780,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(22780,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(22780,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(22780,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(22780,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(22780,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(22780,36,'cp -87','f8c5 f1c4 a7a5 d2d3 c6d4 a2a4 c7c6 c4a2 g8f6 g1e2 e8g8 e2d4 c5d4 d1e2 d7d5 c3d1 b7b5 a4b5 c6b5 a2d5 f6d5 e4d5 c8b7 c2c3 d4c5 c1e3 d8d5 e3c5 d5c5 e2f2 c5d6 d1e3 a5a4 e1g1 a4a3 a1a3 a8a3 b2a3 d6a3');
INSERT INTO evaluations VALUES(31964,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(31964,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(31964,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(31964,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(31964,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(31964,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(31964,36,'cp 207','e2e4 f5g6 f1c4 e7e6 d1b3 g8f6 b3b7 b8d7 f2f3 e6e5 c1e3 f8d6 g1e2 e8g8 b7a6 e5d4 e3d4 d7b6 c4b3 d8e7 e1g1 c7c6 a1d1 f6d7 a6b7 c6c5 d4f2 c5c4 b3c2 f8d8 c3b5 d6e5 f3f4 e5b2');
INSERT INTO evaluations VALUES(31968,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(31968,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(31968,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(31968,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(31968,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(31968,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(31968,36,'cp 190','f5g6 f1c4 e7e6 d1b3 g8f6 c1g5 f8e7 e4e5 f6g8 g5e3 b8d7 a1d1 d8c8 h2h4 c7c6 h4h5 g6f5 g1e2 d7b6 c4d3 f5d3 d1d3 c8d7 e1g1 e8c8 f1d1 c8b8 e3d2 g8h6 d2h6 g7h6 d3f3 h8f8 e2f4 d7c7 a2a4 b8a8 a4a5 b6d5 f4d5');
INSERT INTO evaluations VALUES(31809,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(31809,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(31809,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(31809,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(31809,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(31809,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(31809,36,'cp 211','f1c4 e7e6 d1b3 b8d7 h2h4 g8f6 e4e5 f6g8 h4h5 g6f5 b3b7 a8b8 b7a7 f8b4 a7a4 h7h6 g1e2 g8e7 c4b5 b4c3 b2c3 e8g8 c1a3 b8a8 a3e7 d8e7 a4d1 c7c5 e1g1 c5d4 c3d4 f5g4 a2a4 g4h5 d1d2 f8d8 e2g3 h5g6 f2f4 d7b6 f4f5 g6f5');
INSERT INTO evaluations VALUES(31808,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(31808,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(31808,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(31808,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(31808,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(31808,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(31808,36,'cp 210','e7e6 d1b3 b8d7 h2h4 g8f6 e4e5 f6g8 h4h5 g6f5 b3b7 a8b8 b7a7 f8b4 g1e2 h7h6 a7a4 g8e7 c4b5 b4c3 e2c3 e8g8 e1g1 d7b6 a4d1 b6d5 a2a4 d5b4 a4a5 c7c6 b5e2 f5c2 d1d2 e7f5 d4d5 e6d5 c3a2 b4a2 a1a2 c2b3 a2a3 c6c5 e2d3 f5d4 d2c3 b3c4 a5a6 c4d3 c3d3 b8a8 c1e3');
INSERT INTO evaluations VALUES(32263,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(32263,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(32263,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(32263,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(32263,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(32263,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(32263,36,'cp 198','d1b3 g8f6 c1g5 h7h6 g5f6 g7f6 a1d1 b8d7 g1e2 f8d6 e1g1 d8e7 b3b7 e8g8 b7a6 f8b8 a6c6 e7d8 b2b3 b8b6 c6a4 c7c6 f2f4 d8c7 c4d3 f6f5 e4e5 d6e7 d3c2 h6h5 a4c4 a8b8 c4d3 g8g7 d3d2 h5h4 h2h3 b6b4 d1c1 c7b6');
INSERT INTO evaluations VALUES(11565,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(11565,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(11565,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(11565,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(11565,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11565,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(11565,36,'cp 29','d2d4 b7b5 a4b3 d7d5 d4e5 c8e6 c2c3 f8e7 b1d2 e4c5 b3c2 d5d4 d2b3 d4d3 c2b1 c5b3 a2b3 e6f5 c1e3 e8g8 e3d4 d8d5 f1e1 d3d2 d1d2 f5b1 a1b1 d5b3 e1e4 b3e6 h2h4 f8d8 d2e2 b5b4 h4h5 a6a5 d4e3 a5a4 c3b4');
INSERT INTO evaluations VALUES(11563,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(11563,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(11563,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(11563,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(11563,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(11563,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(11563,36,'cp 26','b7b5 a4b3 d7d5 d4e5 c8e6 c2c3 f8e7 b1d2 e4c5 b3c2 d5d4 d2b3 d4d3 b3c5 d3c2 d1d8 a8d8 c5e6 f7e6 c1e3 d8d5 f1c1 c6e5 f3e5 d5e5 e3d4 e5g5 c1c2 e8f7 a2a4 g5d5 a4b5 a6b5 g2g3 h8d8 a1a7 d8d7 d4e3 e7c5 e3c5 d5c5 a7b7 d7d5 c2e2 h7h6 b7b8 g7g5 b8c8 c5c6 g1g2');
INSERT INTO evaluations VALUES(11722,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(11722,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(11722,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(11722,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(11722,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(11722,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(11722,36,'cp 25','a4b3 d7d5 d4e5 c8e6 c1e3 f8e7 c2c3 d8d7 b1d2 a8d8 a2a4 e4d2 e3d2 e8g8 a4b5 a6b5 d1e2 e6f5 f1e1 d8b8 h2h3 h7h6 d2f4 b5b4 b3a4 b4c3 b2c3 d7e6 a4c6 e6c6 f3d4 c6g6');
INSERT INTO evaluations VALUES(11719,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(11719,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(11719,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(11719,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(11719,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11719,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(11719,36,'cp 16','d7d5 d4e5 c8e6 c2c3 f8e7 b1d2 e4c5 b3c2 d5d4 d2b3 d4d3 b3c5 d3c2 d1d8 a8d8 c5e6 f7e6 c1e3 d8d5 f1c1 c6e5 f3e5 d5e5 c1c2 e8f7 a2a4 h8d8 a4b5 a6b5 g2g3 e7c5 e3c5 e5c5 a1a7 b5b4 a7b7 b4c3 c2c3 c5c3 b2c3 d8c8 g1f1 f7e7 f1e2 e7d7 b7b5 c8a8 f2f4');
INSERT INTO evaluations VALUES(11623,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(11623,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(11623,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11623,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(11623,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11623,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(11623,36,'cp 24','d4e5 c8e6 c2c3 f8e7 c1e3 e8g8 b1d2 e6g4 d2e4 d5e4 d1d5 d8d5 b3d5 e4f3 d5c6 f3g2 f1c1 a8b8 a2a4 b5b4 c6g2 b4b3 a4a5 f8d8 g2f1 b8a8 f2f4 g4e6 c1d1 d8d1 a1d1 f7f6 f4f5 e6f5 f1c4 g8f8 e5f6 e7f6 g1f2 a8e8 e3c5 f6e7 c5e7 f8e7');
INSERT INTO evaluations VALUES(11622,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(11622,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(11622,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11622,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(11622,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(11622,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(11622,36,'cp 25','c8e6 c2c3 f8e7 c1e3 d8d7 b1d2 e4d2 d1d2 c6a5 f3d4 c7c5 d4e6 f7e6 b3d1 a5c4 d2e2 c4e5 e2h5 e5f7 a2a4 a8c8 a4b5 a6b5 a1a6 c8c6 a6a8 c6c8 a8c8 d7c8 f1e1 e7f6 d1c2 g7g6 h5e2 e8g8 e2b5 c5c4 e1a1 f7e5 e3d4 c8c6 b5c6 e5c6');
INSERT INTO evaluations VALUES(24339,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24339,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(24339,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24339,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24339,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24339,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24339,36,'cp 21','c2c3 f8e7 c1e3 e8g8 b1d2 e6g4 d2e4 d5e4 d1d5 d8d5 b3d5 e4f3 d5c6 f3g2 f1c1 a8b8 a2a4 b5b4 c6g2 b4b3 a4a5 f8d8 g2f1 b8a8 f2f4 g7g6 g1f2 e7h4 f2g2 h4e7 g2g3 h7h5 f1c4 g4e6 c4e6 f7e6 g3f2 d8d5 c1d1 d5d1 a1d1 a8d8 d1d8 e7d8 f2f3 g8f7 f3e4 d8e7');
INSERT INTO evaluations VALUES(24341,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(24341,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24341,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(24341,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(24341,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(24341,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24341,36,'cp 13','f8e7 b1d2 e4c5 b3c2 d5d4 d2b3 d4d3 b3c5 d3c2 d1d8 a8d8 c5e6 f7e6 c1e3 d8d5 f1c1 c6e5 f3e5 d5e5 c1c2 e8f7 g1f1 h8d8 a2a4 e7g5 e3g5 e5g5 a4b5 a6b5 g2g3 g5d5 a1a7 d8d7 a7b7 f7e7 b2b4 c7c6 b7b6 d7d6 f1g2 h7h5 h2h4 e7f6 b6b8 d6d8 b8d8 d5d8');
INSERT INTO evaluations VALUES(19409,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(19409,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19409,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(19409,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(19409,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19409,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19409,36,'cp 6','b1c3 g7g6 f1g2 f8g7 e2e3 e7e6 g1e2 g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 e1g1 e8g8 f1e1 f8e8 h2h4 h7h6 c3d5 e6d5 c1e3 c8g4 d1d2 d8f6 b2b4 g4e2 e1e2 c6d4 e3d4 f6d4 a1d1 d4d2 e2d2 g7c3 d2d5 c3b4 d5d7 b4c5 g2b7 a8b8 b7d5 e8e7 d7e7 c5e7');
INSERT INTO evaluations VALUES(25690,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(25690,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(25690,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(25690,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(25690,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(25690,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(25690,36,'cp 6','h2h3 h7h5 d2d3 d8f6 c1e3 g4f3 d1f3 f6f3 g2f3 f7f5 e4f5 g8e7 f1e1 e7f5 b1d2 f8d6 g1f1 e8g8 c2c3 d6e7 f1e2 a8d8 e1g1 d8d5 d2e4 f8d8 a1d1 g8f7 a2a3 b7b6 e4g3 g7g6 g3e4 c6c5 a3a4 d8e8 e3g5 d5d8 g5e7');
INSERT INTO evaluations VALUES(25689,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(25689,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(25689,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(25689,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(25689,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(25689,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(25689,36,'cp 9','h7h5 d2d3 d8f6 c1e3 g4f3 d1f3 f6f3 g2f3 f8d6 b1d2 g8e7 f1d1 c6c5 c2c3 e7c6 g1f1 f7f6 f1e2 g7g5 d2c4 b7b5 c4d6 c7d6 d3d4 c5d4 c3d4 e8e7 d4e5 c6e5 f3f4 g5f4 e3f4 a8c8 f4e3 c8c2 d1d2 h8c8 f2f4 c2d2 e3d2 e5c4 b2b3 c4d2 e2d2 c8g8 d2e3 g8g2');
INSERT INTO evaluations VALUES(25652,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(25652,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(25652,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(25652,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(25652,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(25652,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(25652,36,'cp 13','d2d3 d8f6 c1e3 g4f3 d1f3 f6f3 g2f3 f8d6 b1d2 g8e7 f1d1 e8g8 g1f1 c6c5 c2c3 f7f6 a2a3 a6a5 a3a4 a8d8 a1c1 b7b6 d2c4 g7g5 d3d4 e5d4 c3d4 c5d4 c4d6 d8d6 d1d4 d6d4 e3d4 c7c5 d4e3 g8f7 f3f4 g5f4 e3f4 e7c6 f4c7 f8e8 f2f3');
INSERT INTO evaluations VALUES(25651,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(25651,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(25651,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(25651,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25651,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25651,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25651,36,'cp 0','d8f6 c1e3 g4f3 d1f3 f6f3 g2f3 f8d6 b1d2 g8e7 f1d1 f7f5 e4f5 e7f5 g1f1 a8d8 c2c3 e8g8 a2a4 d8d7 f1e2 d6e7 b2b4 f5h4 a4a5 g8h8 a1b1 h8g8');
INSERT INTO evaluations VALUES(27485,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27485,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27485,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27485,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27485,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(27485,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27485,36,'cp 1','c1e3 g4f3 d1f3 f6f3 g2f3 f8d6 b1d2 g8e7 f1d1 e8g8 c2c3 f7f5 e4f5 a8d8 g1f1 e7f5 f1e2 d6e7 b2b4 d8d7 a2a3 f5h4 a1b1 f8d8 d2c4 d7d5 c4a5 d8b8 d1g1 g8f7 a3a4 h4g6 c3c4 d5d7 b4b5 c6b5 a4b5 b8a8 b5b6 c7c6 a5b3 a8d8 b3a5');
INSERT INTO evaluations VALUES(19519,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(19519,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(19519,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(19519,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(19519,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(19519,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(19519,36,'cp -50','e7e6 g1f3 f8b4 e1g1 g8e7 c1g5 e8g8 a1c1 h7h6 g5h4 b4c3 c1c3 d7d5 e4d5 e6d5 c4b3 c8e6 f3d4 c6d4 h4e7 d8e7 d1d4 f8d8 c3g3 e7f8 f1e1 g8h8 h2h4 b7b6 d4d3 a8c8 b3c2 f8g8 c2b3 d8e8 b3a4 e8e7 a4b3 c8d8 d3d4 g8f8 e1e5 d8c8');
INSERT INTO evaluations VALUES(18876,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(18876,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(18876,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(18876,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(18876,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(18876,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(18876,36,'cp -36','g1f3 f8b4 e1g1 g8e7 c3b5 d7d5 e4d5 e6d5 c4d5 e7d5 d1d5 e8g8 d5d8 f8d8 c1e3 c8g4 b5d4 c6d4 f3d4 a7a5 h2h3 g4d7 a1c1 h7h6 c1c7 a8b8 d4f3 d7c6 f3d4 c6d5 a2a4 b4d6 c7c2 d5e4 c2d2 d6e5 f1d1 d8d5 d4e2 b8d8 d2d5 d8d5 d1d5');
INSERT INTO evaluations VALUES(40988,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40988,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(40988,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(40988,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40988,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(40988,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(40988,36,'cp 23','c2c4 e8g8 b1c3 b7b6 f1d3 c8a6 c4d5 f6d5 e1g1 d5f4 e3f4 a6d3 d1d3 c7c5 d4d5 b8c6 f1d1 c6b4 d3e4 e6d5 c3d5 b4d5 d1d5 d8c7 a1d1 a8d8 b2b3 d8d5 d1d5 e7f6 g2g3 c7c6 h2h4 f8e8 e4c4 c6c8 a2a4 e8d8 g1g2 g7g6 h4h5 d8d5 c4d5 g6h5 d5h5');
INSERT INTO evaluations VALUES(40949,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40949,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(40949,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(40949,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(40949,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40949,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(40949,36,'cp 25','e8g8 b1c3 b7b6 f1d3 c8a6 c4d5 f6d5 c3d5 d8d5 e1g1 d5b7 f3e5 a6d3 d1d3 f8d8 d3c2 c7c5 d4c5 e7c5 f1d1 b8d7 b2b4 d7e5 b4c5 e5c6 h2h3 b6c5 c2c5 a8c8 d1d8 c6d8 c5a3 a7a6 g1h2 b7c6 f2f3 h7h6 a1d1');
INSERT INTO evaluations VALUES(36315,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(36315,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(36315,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36315,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(36315,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(36315,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(36315,36,'cp 35','b1c3 b7b6 f1d3 d5c4 d3c4 c8a6 d1e2 a6c4 e2c4 c7c5 d4c5 b6c5 e1g1 b8c6 f1d1 d8a5 f4d6 a8c8 h2h3 h7h6 c3e4 f6e4 c4e4 f8d8 d6e7 c6e7 d1d8 c8d8 e4b7 e7g6 a1c1 a5b6 b7b6 a7b6 g1f1 g8f8 f1e2 d8d7 a2a3 f8e7 f3d2');
INSERT INTO evaluations VALUES(34567,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(34567,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(34567,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(34567,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(34567,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(34567,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(34567,36,'cp 213','g1f3 g6d6 d2d4 g8f6 f3e5 a7a6 e2f3 e7e6 e1g1 d6d8 d1e2 f8e7 f1d1 e8g8 c1f4 f6d5 f4g3 d5c3 b2c3 b8d7 e5d7 d8d7 c3c4 a8b8 a1b1 h7h6 c4c5 b7b5 c5c6 d7d8 c2c4 e7d6 c4b5 d6g3 h2g3 a6b5 b1b5 b8a8 e2c2 d8f6 b5c5 f8d8 d4d5 e6d5 c5d5 d8d5');
INSERT INTO evaluations VALUES(34564,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(34564,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(34564,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(34564,33,'cp 204',NULL);
INSERT INTO evaluations VALUES(34564,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(34564,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(34564,36,'cp 201','g6b6 e1g1 e7e6 d2d4 g8f6 c1f4 a7a6 c3a4 b6a7 c2c4 f8d6 f4e3 b7b6 c4c5 d6e7 f3e5 e8g8 e2f3 f6d5 a1c1 b6b5 a4c3 c7c6 f3e4 a7c7 d1c2 g7g6 e3h6 f8e8 c1e1 b8d7 f2f4 f7f5 e4f3 d7e5 f4e5 e7f8 h6f8 e8f8 f3d5 e6d5');
INSERT INTO evaluations VALUES(52789,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(52789,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52789,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(52789,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52789,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(52789,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(52789,36,'cp 3','e7e6 b1c3 f8b4 c1d2 g8e7 f1d3 e8g8 a2a3 b4c3 d2c3 b7b6 e4e5 c6c5 f3f4 c8a6 g1f3 h7h6 g2g4 a6d3 d1d3 b8c6 e1g1 c5d4 c3d4 c6d4 f3d4 a8c8 c2c3 e7c6 d4b5 f7f6 e5f6 f8f6 a1e1 d8d7 b5d4');
INSERT INTO evaluations VALUES(53614,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(53614,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(53614,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(53614,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53614,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53614,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(53614,36,'cp 38','b1c3 f8b4 c1d2 b8d7 a2a3 b4a5 f1d3 g8e7 f3f4 a5c3 d2c3 d5e4 d3e4 d8c7 d1f3 e8g8 g1e2 a7a5 e4d3 b7b6 f3g3 e7d5 c3d2 c8b7 c2c4 d5f6 a1d1 b6b5 c4c5 b5b4 e1g1 b4a3 b2a3 b7a6 d3a6 a8a6 f4f5 c7g3 e2g3 f8e8 f5e6 f7e6 f1e1 a6a8 d2f4 f6d5 f4d6 d7f6 g3e4 f6e4 e1e4 a8d8');
INSERT INTO evaluations VALUES(53607,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53607,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53607,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53607,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(53607,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(53607,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(53607,36,'cp 23','f8b4 c1d2 g8e7 a2a3 b4a5 f1d3 b8d7 f3f4 a5c3 d2c3 d5e4 d3e4 d8c7 d1f3 e8g8 g1e2 a7a5 e4d3 b7b6 f3g3 e7d5 c3d2 d7f6 e1g1 d5e7 d2c3 c8a6 d3a6 a8a6 d4d5 e7d5 c3e5 c7a7 f4f5 b6b5 e2d4 f6h5 g3h4 f7f6 f5e6 h5f4 e5f4 a7d4 g1h1 d5f4 f1f4 d4b2 a1e1 b2c2 e6e7 f8e8');
INSERT INTO evaluations VALUES(23289,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(23289,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(23289,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(23289,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(23289,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(23289,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(23289,36,'cp 154','d2d4 e5d4 f3g5 c6e5 d1d4 d8e7 c4g8 h7h6 b1c3 h6g5 c1g5 e7g5 d4e5 f8e7 c3d5 f5e4 e5e4 c7c6 d5e7 g5e7 e4e7 e8e7 g8c4 d7d5 c4d3 g7g5 e1g1 h8f8 a1e1 e7d6 e1e3 c8f5 d3f5 f8f5 f1e1 a7a5 e3e7 b7b5 e1e6 d6c5 e7c7 a8a6 c7g7 a5a4 g2g4 f5f4 g7g5 a4a3 b2b3 c5b4 h2h4 b4c3 e6e2 b5b4');
INSERT INTO evaluations VALUES(23290,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(23290,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(23290,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(23290,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(23290,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(23290,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(23290,36,'cp 170','e5d4 e4e5 d7d5 e5d6 c7d6 e1g1 d6d5 c4b3 g8f6 f3d4 f8c5 c1e3 c6a5 d4c6 b7c6 e3c5 f6e4 f1e1 e8f7 c5e3 h8e8 f2f3 e4g5 b1d2 g5e6 e3f2 a5b3 d2b3 c8d7 b3d4 d8f6 d1d2 f7g8 d4e6 e8e6 e1e6');
INSERT INTO evaluations VALUES(19726,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(19726,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(19726,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(19726,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(19726,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19726,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(19726,36,'cp -31','g1f3 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 d8b6 c2c4 e7e6 d2d4 f8d6 e5g6 h7g6 b1c3 d6b4 d1a4 b4c3 b2c3 d5c4 a4c4 g8e7 c1a3 a7a6 b5a4 a8c8 c4c5 b6c7 c5d6 c7d6 a3d6 b7b5 a4b3 c6a5 d6b4 a5c4 a2a4 e7d5 a4b5 a6b5 a1a7 d5b4 c3b4 c4d6 e1e2');
INSERT INTO evaluations VALUES(19722,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(19722,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(19722,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(19722,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(19722,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(19722,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(19722,36,'cp -20','c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 d8b6 c2c4 e7e6 d2d4 f8d6 e5g6 h7g6 b1c3 d6b4 d1a4 b4c3 b2c3 d5c4 c1a3 g8f6 a4c4 e8c8 b5c6 b6c6 c4c6 b7c6 f2f3 f6d5 e1d2 h8h4 c3c4 d5f6 a3c5 d8h8 c5a7 h4h3 h1h3 h8h3 d2e2 f6d7 a2a4 h3h2 e2d3 c8b7 a7c5 d7c5 d4c5 e6e5 d3e4 h2e2 e4d3');
INSERT INTO evaluations VALUES(26249,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(26249,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(26249,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(26249,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(26249,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(26249,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(26249,36,'cp -35','h2h3 g4h5 g2g4 h5g6 f3e5 d8c7 d2d4 e7e6 c1f4 f8d6 b1c3 g8f6 d1d2 a7a6 b5c6 b7c6 e1c1 c6c5 e5g6 h7g6 f4d6 c7d6 d4c5 d6c5 f2f4 e8g8 f4f5 e6f5 g4f5 g6f5 h1g1 g7g6 c3d5 c5d6 d5f4 d6d2 d1d2 f8e8 b2b3 a6a5 c2c4 e8e4 f4d5 g8g7 d5f6 g7f6');
INSERT INTO evaluations VALUES(31065,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(31065,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(31065,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(31065,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(31065,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(31065,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31065,36,'cp 10','e7e6 g1f3 f8e7 f1d3 f5g6 e1g1 g8f6 b2b3 b8d7 d1c2 e8g8 d3g6 h7g6 e3e4 d5e4 c3e4 d8a5 e4g5 a8e8 c1b2 a5f5 a1c1 f6h5 c2f5 g6f5 g2g3 h5f6 f1d1 f6e4 g5e4 f5e4 f3e5 e8d8 h2h4 d7e5');
INSERT INTO evaluations VALUES(31442,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(31442,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31442,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31442,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(31442,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(31442,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(31442,36,'cp 13','g1f3 f8e7 f1d3 f5g6 d1e2 g8f6 e3e4 d5e4 c3e4 g6e4 d3e4 d8a5 c1d2 e7b4 a2a3 b4d2 f3d2 e8g8 e1g1 b8d7 d2f3 f6e4 e2e4 h7h6 h2h3 a5c7 a1d1 a8d8 e4c2 a7a5 d1d3 c6c5 f1d1 c5d4 d3d4 d7c5 d4d8 f8d8 d1d8 c7d8 b2b4');
INSERT INTO evaluations VALUES(31443,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31443,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31443,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31443,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(31443,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(31443,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(31443,36,'cp 5','f8e7 f1d3 f5g6 d3g6 h7g6 e1g1 g8f6 d1d3 e8g8 f1d1 b8d7 e3e4 d5c4 d3c4 b7b5 c4e2 b5b4 c3a4 d8a5 b2b3 c6c5 f3e5 f8c8 e5c4 a5a6 c1b2 c5d4 b2d4 c8c7 f2f3 f6h5 g2g3 a8c8 d4e3 h5f6');
INSERT INTO evaluations VALUES(22620,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(22620,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(22620,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(22620,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(22620,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(22620,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(22620,36,'cp 49','g8f6 d2d4 f8b4 f1d3 d7d5 e1g1 b4c3 e4d5 c6d4 b2c3 d4f3 d1f3 d8d5 f3g3 b7b6 a2a4 c8b7 d3b5 c7c6 b5e2 e8c8 a4a5 f6e4 g3e3 c6c5 f2f3 e4f6 a1b1 d5d6 c3c4 e6e5 e3a3 h8e8 c1g5 c8c7 g5h4 d8d7 h4g3 f6h5 f1d1 h5g3 d1d6 g3e2 g1f1 d7d6 f1e2 e8d8');
INSERT INTO evaluations VALUES(16297,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(16297,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(16297,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(16297,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(16297,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(16297,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(16297,36,'cp 71','d2d4 f8b4 f1d3 d7d5 e1g1 b4c3 e4d5 e6d5 d1e2 d8e7 e2e7 c6e7 b2c3 e8g8 c1g5 f8e8 f1e1 c8d7 g5f6 g7f6 f3h4 e7g6 h4f5 d7f5 d3f5 e8e1 a1e1 g8f8 g1f1 h7h5 g2g3 b7b6 h2h4 g6e7 f5d3 a8c8 d3a6 c8d8 e1b1 d8d6 a6d3 d6c6 f1e1 e7c8');
INSERT INTO evaluations VALUES(16293,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(16293,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(16293,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(16293,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(16293,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(16293,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(16293,36,'cp 58','f8b4 f1d3 d7d5 e1g1 b4c3 e4d5 e6d5 d1e2 d8e7 e2e7 c6e7 b2c3 e8g8 f1e1 e7g6 a2a4 a7a5 c3c4 d5c4 d3c4 f8e8 e1e8 f6e8 c4f1 b7b6 c2c4 h7h5 g2g3 e8d6 h2h4 c8d7 c1b2 f7f6 f3d2 a8d8 f1e2 d7g4 f2f3 g4e6 d4d5 e6d7 b2c3 g6e5 g1f2');
INSERT INTO evaluations VALUES(9969,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(9969,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(9969,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(9969,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(9969,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(9969,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(9969,36,'cp 69','f1d3 d7d5 e1g1 b4c3 e4d5 c6d4 b2c3 d4f3 d1f3 d8d5 f3g3 c8d7 c3c4 d5c6 a2a4 e8c8 f1e1 e6e5 e1e5 h8e8 c1b2 g7g6 g3e3 e8e5 b2e5 b7b6 h2h3 d7f5 a4a5 d8e8 e3f4 f6d7 e5g7 f5d3 c2d3 f7f6 f4h6 c6d6 a5b6 a7b6 h6h7 d6d3 a1a8 d7b8 g7f6');
INSERT INTO evaluations VALUES(9971,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(9971,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(9971,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(9971,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(9971,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(9971,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(9971,36,'cp 52','d7d5 e1g1 b4c3 e4d5 c6d4 b2c3 d4f3 d1f3 d8d5 f3g3 c8d7 c3c4 d5c6 f1e1 e8c8 a2a4 e6e5 e1e5 h8e8 c1b2 a7a6 a4a5 g7g6 h2h3 e8e5 b2e5 d7f5 g3h4 f6e4 a1e1 g6g5 h4h5 d8e8 e5d4 f5g6 h5h6 e4d2 e1e5 d2c4 e5g5 g6d3 h6c6 b7c6');
INSERT INTO evaluations VALUES(8963,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(8963,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(8963,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(8963,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(8963,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(8963,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(8963,36,'cp 63','e1g1 b4c3 e4d5 e6d5 d1e2 d8e7 e2e7 c6e7 b2c3 e8g8 f1e1 e7g6 g2g3 b7b6 h2h4 h7h6 c1a3 f8d8 f3d2 c8b7 f2f3 a7a5 g1f2 b7a6 d2f1 c7c6 f1e3 a6d3 c2d3 h6h5 e1b1 b6b5 e3f5 d8e8 a3d6 a5a4 d6b4 e8e6 b1e1 a8e8 e1e6');
INSERT INTO evaluations VALUES(34616,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(34616,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(34616,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(34616,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(34616,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(34616,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(34616,36,'cp 71','e5a5 d2d4 g8f6 e2f4 c7c6 f1c4 c8f5 c1d2 a5c7 h2h3 e7e5 d1e2 f8d6 c4d3 e8g8 d3f5 e5f4 e1c1 b8d7 e2f3 a8e8 h1e1 b7b5 a2a3 a7a5 f5d7 c7d7 d2f4 b5b4 f4d6 d7d6 a3b4 a5b4 c3e4 f6e4 e1e4 e8e4 f3e4 d6f6 e4f3 f6f3 g2f3 f8d8 h3h4 g8f8');
INSERT INTO evaluations VALUES(41393,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(41393,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(41393,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(41393,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(41393,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(41393,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(41393,36,'cp 75','d1d2 c7c6 f4h6 g7h6 d2h6 e7e5 g1f3 d8e7 d4e5 d6e5 f1c4 b8d7 f3g5 h8f8 g5h7 f6h7 h6h7 b7b5 c4e2 e7g5 h7h3 d7c5 h3f3 f8h8 c3d1 c5e6 c2c3 g5e7 f3e3 b5b4 g2g3 e8f8 e1g1 a7a5 f2f3 a5a4 e2c4 b4c3 b2c3 f8g7 h2h4 e7c5 e3c5');
INSERT INTO evaluations VALUES(41394,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(41394,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(41394,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(41394,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(41394,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(41394,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(41394,36,'cp 74','c7c6 f4h6 g7h6 d2h6 e7e5 e1c1 d8e7 f1e2 b7b5 f2f4 b5b4 c3a4 f6e4 f4e5 d6d5 d1f1 b8d7 e2d3 e7g5 h6g5 e4g5 h2h4 g5e6 h4h5 e8g8 g1e2 d7b6 h5g6 h7g6 b2b3 b6a4 b3a4 g8g7 f1f6 c8d7 h1h4 c6c5 d4c5 e6c5');
INSERT INTO evaluations VALUES(40316,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(40316,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(40316,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(40316,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(40316,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(40316,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(40316,36,'cp 42','f4h6 g7h6 d2h6 e7e5 d4e5 d6e5 f1c4 b7b5 c4b3 d8e7 g1f3 b8d7 e1g1 a7a5 a2a3 d7c5 h2h3 f6h5 b3a2 f7f6 f1e1 a5a4 a1d1 c8d7 h6e3 h5f4 d1d2 e8c8 e1d1 d7e6 a2e6 c5e6 c3d5 c6d5 e4d5 e7c5 d5e6 c5e3 f2e3 f4e6 d2d6 c8c7');
INSERT INTO evaluations VALUES(52093,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(52093,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(52093,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(52093,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(52093,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(52093,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(52093,36,'cp 74','c1e3 b8c6 b1c3 g8f6 g1f3 h7h6 f3d2 d7d5 c5d6 d8d6 c3b5 d6e7 c2c4 c8e6 g2g3 e6c4 d2c4 e7b4 b5c3 b4c4 f1g2 f8c5 e3c5 c4c5 e2e3 e8g8 e1g1 a8d8 d1e2 d8d7 a1c1 c5b6 f1d1 d7d1 c1d1 f8d8 d1d8 c6d8 a2a3 a7a5 e2c2 d8e6 c3d5 f6d5 g2d5');
INSERT INTO evaluations VALUES(52091,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(52091,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(52091,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(52091,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(52091,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(52091,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(52091,36,'cp 65','g8f6 b1c3 b8c6 g1f3 h7h6 f3d2 d7d5 c5d6 d8d6 c3b5 d6e7 c2c4 c8e6 g2g3 e6c4 d2c4 e7b4 b5c3 b4c4 f1g2 f8c5 e3c5 c4c5 e2e3 e8g8 e1g1 f8d8 d1c2 a8c8 a2a3 c6d4 c2a4 b7b5 e3d4 c5c3 b2c3 b5a4 d4e5 f6d5 g2h3 c8c3 f1d1 g7g6');
INSERT INTO evaluations VALUES(57696,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(57696,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(57696,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57696,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57696,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(57696,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(57696,36,'cp 22','g1f3 c7c5 c2c3 f8g7 d2d4 c5d4 c3d4 g8f6 f1g2 d7d5 f3e5 e8g8 b1c3 c8f5 c1f4 b8c6 e5c6 b7c6 e1g1 f6e4 c3a4 h7h6 b2b3 g6g5 f4e3 f8e8 a1c1 a8c8 f2f3 e4d6 e3f2 e7e5 d4e5 d6b5 e2e4 d5e4 f3e4 f5g6 e5e6 e8e6 a4c5 e6e5 d1g4');
INSERT INTO evaluations VALUES(57695,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(57695,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(57695,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(57695,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(57695,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(57695,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(57695,36,'cp 13','c7c5 c2c3 b7b6 e2e4 c8b7 d1e2 f8g7 f1g2 e7e5 b1a3 g8e7 e1g1 e8g8 d2d4 e5d4 c3d4 d7d5 e4d5 b7d5 d4c5 b6c5 c1e3 d8b6 a1d1 b8c6 a3c4 b6b5 b2b3 a8d8 e3g5 f8e8 f1e1 g7c3 c4e3 b5e2 e1e2 d5e6 d1d8 e8d8 g5e7 c6e7 f3g5 c3g7');
INSERT INTO evaluations VALUES(52317,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52317,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(52317,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52317,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(52317,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52317,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52317,36,'cp 24','c2c3 g8f6 f1g2 f8g7 d2d4 c5d4 c3d4 d7d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 h7h6 h2h4 e7e6 e1g1 f6h5 e5d7 f8d8 d7b6 h5f4 g3f4 a8a6 b6a4 d8a8 a4c5 a6b6 b2b3 c6d4 e3d4 g7d4 c3a4 b6c6 c5b7 d4a1 f1a1 c6c2 b7c5');
INSERT INTO evaluations VALUES(50422,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(50422,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(50422,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(50422,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(50422,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(50422,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(50422,36,'cp 195','d2d4 e7e6 g1h3 h7h6 f1d3 d7d5 e1g1 c7c5 c2c3 b8c6 c1e3 b7b6 h3f4 c8b7 a2a3 f8e7 d1g4 g7g5 f4h5 d8d7 b1d2 e4d2 e3d2 e8c8 b2b4 c5d4 c3d4 c8b8 d2e3 b8a8 f1c1 a8b8 d3b5');
INSERT INTO evaluations VALUES(50415,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(50415,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(50415,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(50415,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(50415,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(50415,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(50415,36,'cp 182','e7e6 g1h3 h7h6 f1d3 d7d5 c2c3 c7c5 e1g1 c8d7 h3f4 b8c6 c1e3 f8e7 d3e4 d5e4 f4h5 g7g6 h5g3 c5d4 c3d4 c6b4 b1c3 d7c6 d1d2 d8d7 g3e4 e8c8 f2f3 c8b8 f1d1 g6g5 a2a3 b4d5 a1c1 b8a8 c3d5 e6d5');
INSERT INTO evaluations VALUES(49335,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(49335,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(49335,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(49335,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(49335,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(49335,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(49335,36,'cp 203','g1h3 h7h6 c2c3 d7d5 f1d3 c7c5 e1g1 c8d7 h3f4 f8e7 c1e3 e8g8 b1d2 c5d4 c3d4 e7g5 d2b3 f7f5 f2f3 g5f4 e3f4 e4g5 d1d2 b8c6 a1c1 g5f7 c1c3 a7a5 f4e3 c6b4 d3b1 a8c8 c3c8 d7c8 b3c1 d8b6 h2h4 b4c6 c1e2 c8d7');
INSERT INTO evaluations VALUES(39002,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(39002,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(39002,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(39002,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(39002,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(39002,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(39002,36,'cp 79','b1c3 a7a6 d2d4 b7b5 c4d3 c8b7 e1g1 d7d6 f1e1 b8d7 c1g5 g8e7 a2a4 b5b4 c3e2 h7h6 g5f4 e8g8 e2g3 e6e5 d4e5 d7e5 f3e5 d6e5 f4e3 d8d6 d3c4 g8h7 d1e2 e7c6 c2c3 c6a5 a1d1 d6e7 c4d5 b7d5 e4d5 b4c3 b2c3 f7f5 e3d4 a8e8 e2a6');
INSERT INTO evaluations VALUES(38993,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(38993,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(38993,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(38993,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(38993,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(38993,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(38993,36,'cp 63','g8e7 d2d4 d7d5 c4b3 e8g8 e1g1 d5e4 c3e4 b7b6 f1e1 h7h6 c2c3 c8b7 c1f4 b8d7 d1e2 e7f5 a1d1 a8c8 b3c4 d7f6 c4d3 f6d5 f4e5 d8e7 d3a6 b7a6 e2a6 f7f6 e5g3 e7d7 a2a4 g6g5 a6e2 a7a5 h2h3 c8e8 g3h2 f8f7 g2g4 f5d6');
INSERT INTO evaluations VALUES(43063,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(43063,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(43063,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(43063,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(43063,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(43063,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(43063,36,'cp 69','d2d4 e8g8 e1g1 d7d5 c4b3 d5e4 c3e4 b7b6 f1e1 c8b7 d1e2 e7f5 c2c3 h7h6 c1f4 b8d7 b3c2 a8c8 h2h3 d8e7 c2d3 f5d6 a2a4 f8d8 e4d6 c7d6 a4a5 d7f6 f3d2 b7c6 a5a6 f6d5 f4g3 e7f6 d2e4 f6e7 a1d1');
INSERT INTO evaluations VALUES(43066,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(43066,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(43066,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43066,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(43066,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(43066,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(43066,36,'cp 62','e8g8 e1g1 d7d5 c4b3 d5e4 c3e4 b7b6 f1e1 h7h6 c2c3 c8b7 c1f4 b8d7 d1e2 e7d5 f4g3 d5f6 e4f6 d7f6 h2h3 d8e7 g3h4 a8d8 a1d1 g6g5 h4g3 c7c5 d4c5 e7c5 b3c2 c5c6 g3e5 d8d1 e1d1 g5g4 h3g4 f6g4 e5d4 f8d8 d4g7 d8d1 c2d1 g8g7');
INSERT INTO evaluations VALUES(37334,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(37334,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(37334,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(37334,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(37334,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(37334,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(37334,36,'cp 65','e1g1 d7d5 c4b3 d5e4 c3e4 b7b6 c2c3 b8d7 f1e1 c8b7 d1e2 h7h6 c1f4 e7d5 f4g3 d5f6 e4f6 d7f6 h2h3 d8e7 a1d1 a8d8 e2e3 f6d5 e3c1 g8h7 f3e5 e7g5 c1g5 h6g5 e5d3 d8c8 c3c4 d5f6 d3e5 f6h5 g3h2 h5f4 f2f3 c8d8 b3c2 g7e5 e1e5');
INSERT INTO evaluations VALUES(37333,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(37333,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(37333,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(37333,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(37333,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(37333,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(37333,36,'cp 66','d7d5 c4b3 d5e4 c3e4 b7b6 f1e1 h7h6 c2c3 c8b7 c1f4 b8d7 d1e2 e7d5 f4g3 d5f6 e4f6 d7f6 g3h4 a7a5 a2a4 b7d5 b3d5 d8d5 h4g3 a8c8 h2h3 f8e8 f3e5 c8d8 a1d1 f6d7 e5c4 d5c6 g3h2 d7f6 b2b3 f6d5 e2f3 e8f8 h2e5');
INSERT INTO evaluations VALUES(19380,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(19380,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(19380,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(19380,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(19380,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(19380,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(19380,36,'cp -43','g8f6 d2d3 e7e6 b1c3 d7d5 e4d5 e6d5 c4a2 c8e6 g1e2 h7h6 e1g1 f8d6 e2f4 d6f4 c1f4 e8g8 f1e1 d5d4 c3e4 e6a2 a1a2 f6e4 e1e4 d8d5 a2a1 a8e8 f2f3 d5d7 f4g3 b7b6 d1d2 c6e7 d2f4 f7f5 e4e1 e7d5 f4d6 d7f7 e1e8 f8e8 a1e1');
INSERT INTO evaluations VALUES(12792,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(12792,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(12792,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(12792,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(12792,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(12792,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(12792,36,'cp -38','b1c3 e7e6 d2d3 d7d5 e4d5 e6d5 c4a2 f8e7 g1e2 e8g8 e1g1 c8e6 f1e1 e7d6 e2f4 f8e8 h2h3 d6f4 c1f4 d5d4 e1e6 e8e6 c3e4 f6e4 a2e6 f7e6 d3e4 e6e5 f4h2 d8e7 d1g4 h7h6 h2g3 c5c4 a1e1 a8f8 f2f4 e5f4 g3f4');
INSERT INTO evaluations VALUES(12790,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(12790,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(12790,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(12790,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(12790,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(12790,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(12790,36,'cp -39','e7e6 d2d3 d7d5 e4d5 e6d5 c4a2 f8e7 g1e2 e8g8 e1g1 d5d4 c3e4 f6e4 d3e4 c8e6 a2e6 f7e6 e2f4 d8d7 b2b3 e6e5 f4d5 e7d6 d1d3 c6e7 d5e7 d7e7 f2f3 f8f7 c1d2 a8f8 a3a4 d6c7 d2e1 a7a6 e1g3 h7h5 g1h1 g8h7 a1e1');
INSERT INTO evaluations VALUES(12379,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(12379,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(12379,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(12379,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(12379,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(12379,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(12379,36,'cp -28','c4a2 d7d5 e4d5 e6d5 d2d3 f8e7 g1e2 c8e6 e1g1 e8g8 f1e1 e7d6 e2f4 f8e8 h2h3 d6f4 c1f4 d5d4 e1e6 e8e6 c3e4 f6e4 a2e6 f7e6 d3e4 e6e5 f4g3 d8e7 d1d3 h7h6 d3c4 g8h7 a1b1 a8c8 b2b4 c5b4 a3b4 a7a6 b4b5 c6a7 c4d5 a7b5 g3e5 c8c2 e5d4 b5d4');
INSERT INTO evaluations VALUES(20989,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(20989,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20989,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20989,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20989,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20989,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(20989,36,'cp 1','g1f3 c8f5 a2a3 e7e6 e2e3 g8f6 c2c4 c6e7 f3e5 g7g6 f1e2 f8g7 b1c3 f6e4 e1g1 c7c6 c3e4 f5e4 e2d3 f7f6 e5f3 e7f5 c4c5 e8g8 d1c2 e4d3 c2d3 e6e5 f4e5 f6e5 f3e5 d8e7 e5f3 g7h6 f3e5');
INSERT INTO evaluations VALUES(20988,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20988,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20988,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(20988,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20988,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(20988,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20988,36,'cp 0','c8f5 a2a3 g8f6 e2e3 e7e6 c2c4 g7g6 b1c3 f8g7 f1e2 c6e7 c1d2 c7c6 e1g1 e8g8 f3e5 f6e4 g2g4 e4c3 d2c3 f5e4 e2f3 f7f6 e5d3 e4f3 d1f3 d5c4 d3c5 d8c8 c5e4 c8c7 e4c5');
INSERT INTO evaluations VALUES(26551,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(26551,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26551,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(26551,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(26551,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(26551,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(26551,36,'cp 0','e2e3 e7e6 a2a3 f8e7 c2c4 g8f6 b1c3 f6e4 c1d2 e4d2 f3d2 e7f6 c4d5 e6d5 f1b5 e8g8 e1g1 c6e7 g2g4 c7c6 b5e2 f5d7 g1h1 e7c8 g4g5 f6e7 e2d3 c8d6 f1g1 g7g6 d2f3 d6f5 f3e5 g8g7 d1g4');
INSERT INTO evaluations VALUES(26550,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(26550,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(26550,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26550,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26550,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(26550,36,'cp 0','c6b4 f1b5 c7c6 b5a4 a7a5 a2a3 b7b5 b1c3 b5a4 a3b4 a5b4 c3a4 e7e6 c1d2 g8f6 a4c5 f8c5 d4c5 a8a1 d1a1 e8g8 e1g1 f5c2 d2b4 c2e4 b4c3 f6g4 a1e1 f7f6 f3d4 d8d7 h2h3 e6e5 d4f3 e4f3 f1f3 e5e4 f3f1 g4h6 g2g4 f8a8 c3d4 a8a2 d4c3');
INSERT INTO evaluations VALUES(51908,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(51908,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(51908,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(51908,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(51908,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(51908,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(51908,36,'cp -37','b8c6 c2c3 d7d6 e5d6 f8d6 d2d4 c5d4 f3d4 c8d7 d4f5 d6e5 f5d6 e5d6 d1d6 g8e7 f1d3 e6e5 b1d2 e8g8 d3c2 d8c8 d2e4 d7e6 c1e3 f8d8 d6a3 e7d5 e1g1 d5e3 f2e3 h7h6 c2b3 e6b3 a3b3 c8c7 a1d1 c6a5 b3c2 a5c4 c2e2 c7c6 e4g3 d8d1 f1d1');
INSERT INTO evaluations VALUES(18721,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(18721,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(18721,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(18721,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(18721,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(18721,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(18721,36,'cp -22','c2c3 d7d6 e5d6 f8d6 d2d4 c5d4 f3d4 c8d7 d4b5 d6b8 b5d6 b8d6 d1d6 g8e7 f1d3 e6e5 b1a3 d7e6 d6d8 a8d8 d3e2 f7f6 f2f4 e7g6 f4e5 g6e5 e1g1 e8g8 c1f4 a7a6 b2b3 e6g4 e2g4 e5g4 a1d1 g4e5 f4e5 c6e5 a3c2 f8e8 c2b4 a6a5 b4c2');
INSERT INTO evaluations VALUES(18718,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(18718,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(18718,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(18718,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(18718,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(18718,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(18718,36,'cp -37','d7d6 e5d6 f8d6 d2d4 c5d4 f3d4 c8d7 d4b5 d6b8 b5d6 b8d6 d1d6 g8e7 f1d3 e6e5 b1a3 d7f5 d6d8 a8d8 d3e2 f7f6 a3c4 e8f7 a2a4 f5d3 c1e3 d3e2 e1e2 d8d7 h1d1 e7d5 g2g3 g7g5 f2f3 h7h5 e3f2 h5h4 c4e3 c6e7 e3d5 e7d5 g3h4');
INSERT INTO evaluations VALUES(20660,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(20660,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(20660,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(20660,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(20660,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(20660,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(20660,36,'cp -28','e5d6 f8d6 d2d4 c5d4 f3d4 c8d7 d4f5 d6f8 f5d6 f8d6 d1d6 g8e7 f1d3 e6e5 b1a3 d7e6 d6d8 a8d8 d3e2 h7h6 a3c4 e8g8 e1g1 g7g5 a2a4 d8d7 f1e1 f7f6 a4a5 f8d8 c1e3 e7d5 g2g3 d5e3 c4e3 d7d2 b2b4 g8f7 b4b5 c6e7 a1d1');
INSERT INTO evaluations VALUES(54852,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(54852,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54852,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54852,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(54852,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54852,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54852,36,'cp 37','c2c4 f8g7 b1c3 g8f6 e2e3 c7c5 d4c5 d8a5 a1c1 d5c4 f1c4 e8g8 g1e2 a5c5 d1b3 b8c6 c3b5 c5b4 b3b4 c6b4 b5c7 b7b5 c4b5 a8b8 e1g1 b4a2 c1a1 a7a6 a1a2 a6b5 e2d4 c8d7 a2a7 b5b4 f1a1 f8c8 f4e5 f6e4 c7d5 g7e5 d5e7 g8f8 a7d7');
INSERT INTO evaluations VALUES(47610,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(47610,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(47610,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(47610,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(47610,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(47610,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(47610,36,'cp 45','b8c6 c4b5 f8e7 d2d4 c8d7 d4d5 c6b4 b5d7 d8d7 d1e2 c7c6 d5c6 d7c6 f3h4 g7g6 h4f3 b4a6 a2a4 a6c5 f1e1 a8c8 f3d2 c5e6 d2b3 h6h5 e2f3 c6d7 c1e3 a7a6 a4a5 e8g8 a1d1 c8c4 b3d2 c4c8 d2f1 c8c4 e3b6 c4b4 b2b3 e6d4 b6d4 b4d4 f3e2 d4d1');
INSERT INTO evaluations VALUES(14611,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14611,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14611,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(14611,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(14611,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(14611,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(14611,36,'cp 46','c4b5 f8e7 d2d4 c8d7 d4d5 c6b4 b5d7 d8d7 d1e2 c7c6 d5c6 d7c6 f3h4 g7g6 h4f3 b4a6 a2a4 a6c5 f1e1 a7a6 f3d2 a8c8 d2b3 c5e6 e2f3 c6d7 a4a5 h6h5 c1e3 e8g8 a1d1 g8g7 d1d3 c8c4 b3d2 c4b4 b2b3 e7d8 e1d1 d8a5 d2c4 b4c4 b3c4 a5c3 d3c3');
INSERT INTO evaluations VALUES(57345,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(57345,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(57345,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(57345,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(57345,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(57345,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(57345,36,'cp 54','g8f6 d1c2 b7b6 c1f4 f8b4 e2e3 b4c3 c2c3 c8b7 f2f3 f6h5 g1h3 d8h4 g2g3 h4f6 f1g2 h5f4 h3f4 e8g8 c3d2 c7c5 e1g1 b8c6 a1d1 a8d8 d4d5 c6e5 b2b3 f8e8 h2h4 e6d5 f4d5 b7d5 d2d5 g8h8 d5d2');
INSERT INTO evaluations VALUES(45227,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(45227,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(45227,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(45227,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(45227,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(45227,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(45227,36,'cp 74','b1c3 e7e6 e2e4 g7g6 f1e2 e6d5 c4d5 f8g7 g1f3 c8g4 c1f4 e8g8 e1g1 a7a6 a2a4 g4f3 e2f3 d8e7 e4e5 d6e5 d5d6 e7e6 f4g5 e5e4 f1e1 e6f5 g5f6 e4f3 f6e7 b8c6 d1f3 f5f3 g2f3 g7c3 b2c3 c6e7 d6e7 f8e8 a1b1 b7b5 a4b5 a6b5 b1b5 c5c4 g1g2');
INSERT INTO evaluations VALUES(45225,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(45225,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(45225,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(45225,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(45225,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(45225,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(45225,36,'cp 87','g7g6 g1f3 e7e6 e2e4 f8g7 h2h3 e6d5 e4d5 e8g8 f1d3 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1c2 e4f6 a1e1 b8a6 a2a3 a6c7 e3d2 e8e1 f1e1 d8d7 b2b3 a8e8 e1f1 b7b5 f1d1 e8a8 d2a5 b5c4 b3c4 a8b8 a5d2 d7d8 d2g5');
INSERT INTO evaluations VALUES(3789,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(3789,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(3789,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(3789,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(3789,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3789,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(3789,36,'cp 7','h7h6 g5h4 c8g4 h2h3 g4e6 c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 c2c3 d8e8 f1e1 e8g6 e1e3 f6h5 g1h2 h5f4 h4g3 g6h5 d3d4 f8f7 h2g1 e5d4 c3d4 d6d5 g1h2 a8f8 e4d5 e6d5 g3f4 f7f4 a1c1 h5g6 h2g1 g6f6 d1b3 c6d4 b3d5 f6f7 d5b7 d4f3 e3f3');
INSERT INTO evaluations VALUES(3583,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(3583,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(3583,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(3583,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(3583,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(3583,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(3583,36,'cp 15','g5h4 c8g4 h2h3 g4e6 c4e6 f7e6 c3a4 c5b6 c2c3 d8e8 a4b6 a7b6 f1e1 e8g6 e1e3 f6h5 g1h2 h5f4 h4g3 f8f7 g3f4 f7f4 h2g1 g6h5 d1b3 h5f7 a1e1 a8f8 d3d4 e5d4 c3d4 e6e5 d4e5 f7b3 e3b3 d6e5 a2a3 f8d8 b3c3 g7g5');
INSERT INTO evaluations VALUES(3593,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(3593,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(3593,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3593,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3593,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(3593,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(3593,36,'cp 17','c8g4 h2h3 g4e6 c4e6 f7e6 c3a4 c5b6 c2c3 d8e8 a4b6 a7b6 f1e1 e8g6 e1e3 f6h5 g1h2 h5f4 h4g3 f8f7 g3f4 f7f4 d1e2 g6h5 d3d4 e5d4 c3d4 a8f8 a1d1 e6e5 d4d5 c6e7 d1c1 c7c6 d5c6 b7c6 c1d1 f4f6 e3d3 d6d5 a2a3 f6f4 e4d5');
INSERT INTO evaluations VALUES(23713,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(23713,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(23713,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(23713,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(23713,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(23713,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(23713,36,'cp 22','h2h3 g4e6 c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 f1e1 d8e8 c2c3 f6h5 h4g3 e8g6 g1h2 h5f4 e1e3 g6h5 d3d4 f8f7 g3f4 f7f4 d1e2 e5d4 c3d4 f4f7 h2g1 e6e5 d4e5 c6e5 f3e5 h5e5 a2a3 c7c6 a1e1 a8f8 f2f3 f7e7 a3a4 f8f7 e1d1');
INSERT INTO evaluations VALUES(23712,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(23712,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(23712,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(23712,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(23712,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(23712,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(23712,36,'cp 26','g4e6 c4e6 f7e6 c3a4 c5b6 c2c3 d8e8 f1e1 f6h5 a4b6 a7b6 h4g3 e8g6 g1h2 h5f4 e1e3 g6h5 d3d4 f8f7 g3f4 e5d4 c3d4 f7f4 d1e2 a8f8 h2g1 h5f7 a1f1 f7g6 d4d5 e6d5 e4d5 c6d4 f3d4 f4d4 e3e7 f8f7 f1c1 f7e7 e2e7');
INSERT INTO evaluations VALUES(38356,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(38356,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(38356,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38356,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(38356,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(38356,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(38356,36,'cp 33','d2d4 g8f6 d1e2 e8g8 e4e5 d6e5 d4e5 f6d5 h2h3 c8e6 e1g1 b8d7 f1e1 c7c6 a2a4 d8c7 c1d2 f8d8 b1c3 d5c3 d2c3 e6c4 e2c4 d7f8 h3h4 f8e6 h4h5 d8d5 c4h4 a8d8 a4a5 b7b5 a5b6 a7b6 e1e3 d5d1 a1d1 d8d1 g1h2 b6b5 f3g5 e6g5');
INSERT INTO evaluations VALUES(46686,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46686,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46686,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46686,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46686,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46686,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46686,36,'cp 0','f8g7 f1e2 c7c5 h2h3 b8c6 e1g1 d8b6 b1d2 b6b2 a1b1 b2a2 d2c4 c8f5 e2d3 f5d3 c2d3 a2a6 b1a1 a6b5 a1b1');
INSERT INTO evaluations VALUES(43161,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(43161,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(43161,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(43161,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(43161,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(43161,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(43161,36,'cp 58','e2e4 d7d6 f1e2 e7e5 d2d4 e8g8 e1g1 b8a6 a2a3 e5d4 f3d4 f8e8 f2f3 c7c6 c1g5 a6c5 g1h1 a7a5 d1c2 a5a4 a1d1 d8c7 c2d2 f6d7 g5h6 g7h6 d2h6 d7f6 d4c2 c8e6 h6f4 f6h5 f4e3 c7b6 g2g4 h5f6 c2b4 a8d8 e3g5 f6d7 d1d6');
INSERT INTO evaluations VALUES(43151,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(43151,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(43151,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43151,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(43151,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(43151,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(43151,36,'cp 52','d7d6 f1e2 e7e5 d2d4 e8g8 e1g1 b8a6 f1e1 c8g4 c1e3 g4f3 e2f3 e5d4 e3d4 f8e8 d1d2 a6c5 a1d1 c5e6 d4e3 h7h5 d2c2 f6g4 f3g4 h5g4 c4c5 g7e5 c5d6 c7d6 c2e2 d8h4 g2g3 h4h3 c3d5 a8c8 d1d2 g8g7 f2f4 g4f3 e2f3 h3h5 f3h5 g6h5 e1f1 e6c7 f1f5 c7d5 f5g5 g7f6 e4d5');
INSERT INTO evaluations VALUES(41252,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(41252,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(41252,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(41252,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(41252,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(41252,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(41252,36,'cp 60','d2d4 e8g8 f1e2 e7e5 e1g1 b8a6 c1e3 d8e7 d1c2 e5d4 f3d4 c7c6 a1d1 f8e8 f2f3 d6d5 c4d5 f6d5 c3d5 c6d5 e2a6 b7a6 d4c6 e7b7 d1d5 c8e6 d5c5 b7b2 c2b2 g7b2 c5c2 b2a3 e3c5 a3c5 c2c5 e6a2 f1a1 a2e6 a1a6 e6d7 c6e5 d7c8 a6a5 c8e6 g1f2 e8c8 c5c8 e6c8 f2e3 f7f5');
INSERT INTO evaluations VALUES(51106,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(51106,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(51106,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(51106,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(51106,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(51106,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(51106,36,'cp 61','e2e4 e7e6 f1d3 d7d5 b1c3 c7c5 d4c5 f8c5 e4d5 e6d5 c3d5 c5f2 e1f2 d8d5 d1f3 g8f6 f3d5 f6d5 g1f3 e8g8 a2a3 h7h6 h1d1 b8c6 c2c4 d5f6 b2b4 f8e8 c1b2 f6e4 f2g1 c8f5 d1e1 e4g3 a1d1 e8e1 f3e1 f5d3 d1d3 g3f5 g1f2 f7f6 g2g4 f5e7 d3d7 a8d8 d7d8 c6d8 f2f3 g8f7 e1d3 d8c6 f3e4');
INSERT INTO evaluations VALUES(8825,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(8825,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(8825,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(8825,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(8825,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(8825,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(8825,36,'cp 13','c3a4 c5b6 a4b6 a7b6 a2a3 e8g8 e1g1 c8e6 c4e6 f7e6 a3a4 d8e8 c1d2 c6e7 a4a5 b6a5 a1a5 a8a5 d2a5 b7b6 a5c3 e8f7 f3h2 e7g6 g2g3 g6e7 d1e2 c7c5 g1g2 e7c6 h3h4 c6d4 c3d4 c5d4 c2c3 d4c3 b2c3 b6b5');
INSERT INTO evaluations VALUES(22722,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(22722,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(22722,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(22722,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(22722,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(22722,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(22722,36,'cp -77','f8c5 d2d3 d7d6 g1e2 g8e7 g2g3 c8g4 f1g2 h7h5 h2h3 g4e2 c3e2 h5h4 g3g4 e7g6 e1g1 c6d4 c1e3 d4e2 d1e2 a7a5 e2d2 e8g8 b2b3 c7c6 a1b1 d8b6 a2a4 a8b8 d2e2 c5e3 f2e3 b6c5 g1h2 b7b5 g4g5 b5a4 b3a4 b8b1 f1b1');
INSERT INTO evaluations VALUES(6449,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(6449,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(6449,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(6449,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(6449,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(6449,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(6449,36,'cp -64','d2d3 d7d6 g1e2 g8e7 g2g3 c8g4 f1g2 h7h5 h2h3 g4e2 c3e2 h5h4 g3g4 e7g6 c1e3 c6d4 e2d4 c5d4 e1g1 c7c5 a1b1 e8g8 b2b4 d8e7 d1d2 f8b8 a2a3 e7f6 e3g5 f6e6 g5e3 b7b5 f1c1 a7a6 g2f3 e6f6 f3d1 g6f4 e3f4 f6f4');
INSERT INTO evaluations VALUES(6450,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(6450,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(6450,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(6450,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(6450,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(6450,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(6450,36,'cp -78','d7d6 a2a3 a7a5 c1e3 g8e7 g1f3 f7f5 e4f5 e7f5 e3c5 d6c5 g2g3 e8g8 f1g2 f5d4 f3d4 c6d4 e1g1 c7c6 c3e4 b7b6 d1h5 c8f5 a1e1 d8d7 h2h3 a8e8 g1h2 f8f7 e1e3 e8f8 h2g1 d7c7 h5d1 h7h6 d1h5 g8h8 e4g5 f7f6 g5e4 f6g6 g1h1');
INSERT INTO evaluations VALUES(43289,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(43289,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43289,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(43289,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(43289,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43289,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43289,36,'cp 20','d7d5 c2c4 c7c6 b2b3 f6e4 d2d4 c8f5 f3h4 f5e6 c1b2 e8g8 d1c2 e4d6 b1d2 a7a5 e2e4 d5e4 d2e4 d6e4 g2e4 a5a4 h4f3 d8b6 f1e1 e6g4 d4d5 c6d5 e4d5 a4b3 c2b3 b6b3 a2b3 a8a1 b2a1 b8c6 d5c6 b7c6');
INSERT INTO evaluations VALUES(41516,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(41516,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(41516,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(41516,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(41516,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(41516,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41516,36,'cp 20','c2c4 c7c6 d2d4 e8g8 c4d5 c6d5 b1c3 b8c6 f3e5 c8f5 e5c6 b7c6 c1f4 f8e8 a1c1 f6h5 f4e3 h5f6 b2b3 h7h5 e3f4 a8c8 c3a4 f6d7 d1d2 e7e5 d4e5 g7e5 f4e5 e8e5 h2h4 d8e7 e2e3 f5e4 f1d1 d7f6');
INSERT INTO evaluations VALUES(22749,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(22749,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(22749,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(22749,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(22749,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(22749,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(22749,36,'cp 76','d4d5 f8b4 b1d2 c6b8 g1f3 b4d2 c1d2 d7d6 e2e4 g8e7 f1d3 a7a5 a2a3 e8g8 e1g1 h7h6 d1c2 a5a4 c4c5 b8a6 c5d6 c7d6 c2a4 f7f5 d2e3 f5e4 d3e4 c8f5 e4f5 e7f5 a4e4 f5e3 e4e3 d8d7 f3d2 d7f7 d2e4 f7d5 e4c3 d5f7');
INSERT INTO evaluations VALUES(22657,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(22657,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(22657,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(22657,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(22657,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(22657,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(22657,36,'cp 72','c6b8 g1f3 f8b4 b1d2 b4d2 f3d2 d7d6 e2e4 a7a5 f1e2 b8d7 d2b1 g8e7 b1c3 e8g8 f2f3 f7f5 c1e3 d7f6 d1d2 f5e4 f3e4 c8g4 e1g1 g4e2 d2e2 e7g6 f1f5 d8e7 a1f1 g6f4 e3f4 e5f4 c4c5 f6d7 c5d6 c7d6 f5f4 f8f4 f1f4 a8f8 f4f8 g8f8 e2f3 f8g8 f3f4 h7h6 g1f1 g7g5 f4f5');
INSERT INTO evaluations VALUES(19159,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(19159,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(19159,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(19159,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(19159,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(19159,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(19159,36,'cp 186','d8e7 e5e7 g8e7 b1a3 e7c6 c1f4 c6d4 c2c3 d4e6 f4g3 b7b6 f2f3 c8b7 a3c4 f8e7 a1d1 b7c6 c4d6 e7d6 d1d6 h7h5 h2h4 b6b5 e1f2 a7a5 f1e2 e8g8 d6d2 f7f5 e4f5 f8f5 e2d3 f5d5 h1d1 a5a4 d3g6 d5d2 d1d2 a4a3 b2b3 d7d5 g6h5 d5d4 c3d4 e6d4');
INSERT INTO evaluations VALUES(1379,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(1379,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(1379,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(1379,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(1379,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(1379,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(1379,36,'cp 178','e5e7 g8e7 b1a3 e7c6 c1f4 c6d4 f2f3 b7b6 c2c3 d4e6 f4g3 c8b7 a3c4 f8e7 a1d1 b7c6 c4d6 e7d6 g3d6 e6d8 e1f2 d8b7 d6f4 a7a5 d1d2 a5a4 h2h4 h7h5 c3c4 b7d8 f4d6 d8e6 f1d3 e8c8 b2b4 a4b3 a2b3 c8b7 h1b1 d8a8 b3b4 c5b4 b1b4 h8c8 b4b1 f7f6 g2g3');
INSERT INTO evaluations VALUES(1342,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(1342,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(1342,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(1342,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(1342,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(1342,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(1342,36,'cp 193','g8e7 b1a3 e7g6 a3b5 e8d8 f1c4 a7a6 b5c3 d8e8 e1g1 d7d6 c3d5 a8b8 f1d1 c8e6 c2c3 f8e7 c4e2 e7d8 c1e3 g6e5 a2a4 h7h5 a4a5 e5g4 e3f4 g4e5 d5e3 g7g5 f4g3 h5h4 g3e5 d6e5 e2c4 d8e7 c4e6 f7e6 e3g4 b8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(50037,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(50037,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(50037,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(50037,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(50037,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(50037,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(50037,36,'cp 251','d2d4 e5d4 e1g1 e7g6 f3d4 b8c6 d4f5 c6a5 c4d3 g6e5 b1c3 a5c6 d3e2 g7g6 f2f4 e5f7 f5g3 f8g7 c1e3 e8g8 d1d2 d7d6 e2c4 g8h8 h2h3 d8d7 a1d1 b7b6 f1f2 c8b7 f4f5 f7e5 c4e6 d7d8 a2a4 c6a5 e6a2 a5c4');
INSERT INTO evaluations VALUES(50039,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(50039,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(50039,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(50039,33,'cp 243',NULL);
INSERT INTO evaluations VALUES(50039,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(50039,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(50039,36,'cp 243','e5d4 e1g1 e7g6 f3d4 b8c6 d4f5 c6e5 c4b3 d7d5 f2f4 e5f7 b1c3 c7c6 d1h5 c8f5 h5f5 f8c5 g1h1 d8d7 f5h5 e8c8 c3a4 c5b6 e4d5 c6c5 c1e3 d7c7 d5d6 f7d6 e3c5 c8b8 a4b6 a7b6 c5g1 d6e4 f4f5 g6e5 h5e2 e4c5 f1e1 h8e8 b3c4 e5c4 e2c4 c7d6 e1e8 d8e8');
INSERT INTO evaluations VALUES(2389,30,'cp 252',NULL);
INSERT INTO evaluations VALUES(2389,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(2389,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(2389,33,'cp 249',NULL);
INSERT INTO evaluations VALUES(2389,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(2389,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(2389,36,'cp 230','f3e5 f7f6 e5d3 c8e6 c1d2 d8c6 b1c3 e8c8 e1c1 c6b4 d2e3 b4d3 d1d3 a7a6 d3d8 c8d8 c3a4 f8d6 a4c5 e6c8 f1e2 g8h6 f2f3 h8e8 h1d1 h6f7 e3f4 f7e5 f4g3 g7g5 c5b3 d8e7 c2c3 b7b6 b3d4 e5g6 g3d6 e7d6');
INSERT INTO evaluations VALUES(2385,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(2385,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(2385,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(2385,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(2385,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(2385,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(2385,36,'cp 248','c8e6 b1c3 f7f6 e5d3 d8c6 c1e3 e8c8 e1c1 f6f5 e4e5 g8e7 h2h4 e6c4 f2f4 h7h5 a2a3 g7g6 g2g3 f8h6 d1e1 c6d4 e3d4 d8d4 b2b3 c4a6 d3c5 a6f1 h1f1 e7c6 c1b2 d4d2 c3e2 d2d5 b3b4 b7b6 e2c3 d5d2 c5b3 d2g2 f1g1 g2h2 g1h1 h2h1 e1h1 h6g7 h1e1');
INSERT INTO evaluations VALUES(19959,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(19959,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(19959,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19959,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19959,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(19959,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(19959,36,'cp -6','e7e6 b1c3 a7a6 a2a3 g8f6 c4a2 f8e7 d2d3 e8g8 e1g1 b7b5 g1h1 d6d5 e4e5 f6d7 d3d4 c5d4 c3e2 a6a5 e2d4 c6d4 f3d4 e7c5 f4f5 d7e5 c1f4 c5d4 d1d4 e5c4 h2h3 c8b7 a2c4 d5c4 f4d6');
INSERT INTO evaluations VALUES(20686,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20686,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20686,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(20686,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20686,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20686,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(20686,36,'cp 0','b1c3 g8f6 d2d3 a7a6 a2a4 d6d5 c4b3 c6d4 b3a2 d5e4 d3e4 c8d7 f3e5 d8c7 e1g1 a8d8 e5d7 d8d7 e4e5 d4b3 d1e2 b3d4');
INSERT INTO evaluations VALUES(20683,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20683,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20683,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20683,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20683,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20683,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20683,36,'cp 6','g8f6 d2d3 a7a6 a2a4 d6d5 c4b3 d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 c6a5 b3a2 b7b5 f4f5 c5c4 c1d2 a5b7 a2b1 b5b4 b1a2 b4b3 c2b3 b7c5 b3c4 c5e4 f5e6 c8e6 d2e3 f8b4 d1c2 e8c8 f3d4 e6d7 c4c5 e4c5 a2f7 d7a4 b2b3 a4e8 f7c4 c8b7 c2b2 e8g6');
INSERT INTO evaluations VALUES(13950,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13950,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13950,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13950,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13950,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13950,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13950,36,'cp 0','d2d3 a7a6 a2a4 d6d5 c4b3 c6d4 b3a2 d5e4 d3e4 c8d7 e1g1 d8c7 f3e5 a8d8 e5d7 d8d7 e4e5 d4b3 d1e2 b3d4');
INSERT INTO evaluations VALUES(13951,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13951,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(13951,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(13951,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(13951,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(13951,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(13951,36,'cp -5','f8e7 e1g1 e8g8 c4b3 c6a5 d1e1 a5b3 a2b3 a7a5 c1d2 b7b6 e1g3 f6h5 g3g4 g7g6 g4h3 f7f5 f3g5 d8d7 e4f5 e6f5 c3d5 e7d8 f1e1 c8b7 h3f3 h5f6 c2c4 b6b5 g5e6 f8e8 h2h3 f6d5 c4d5 d7f7 d2c3 b5b4 e6d8 a8d8 e1e8 d8e8');
INSERT INTO evaluations VALUES(8175,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(8175,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(8175,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8175,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(8175,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8175,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8175,36,'cp -3','e1g1 e8g8 a2a3 a8b8 c4a2 b7b5 d1e1 a7a5 e4e5 f6d7 e5d6 e7d6 c3e4 d6e7 c1e3 d8c7 e1f2 c5c4 a1d1 d7f6 e4f6 e7f6 d3c4 b5c4 c2c3 c6e7 e3d4 f6d4 d1d4 c8a6 f1d1 f8d8 f3e5 e7f5 d4d7 d8d7 e5d7');
INSERT INTO evaluations VALUES(8174,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8174,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8174,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(8174,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8174,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8174,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8174,36,'cp 0','e8g8 a2a3 a8b8 c4a2 b7b5 d1e1 d8c7 e1g3 b5b4 a3b4 c6b4 a2b1 b4c6 b1a2');
INSERT INTO evaluations VALUES(3126,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3126,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3126,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3126,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3126,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3126,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3126,36,'cp 11','a2a3 a8b8 c4a2 b7b5 d1e1 d8c7 e4e5 f6d7 e1g3 b5b4 a3b4 c5b4 e5d6 c7d6 c3e4 d6c7 f4f5 c7g3 e4g3 d7f6 f5e6 c8e6 a2e6 e7c5 g1h1 f7e6 c1g5 c5b6 c2c3 e6e5 c3b4 f6g4 g3e4 g4e3 f1c1 c6b4 f3e5 e3c2 g5e7 c2a1 e7f8 b8f8 c1a1');
INSERT INTO evaluations VALUES(33817,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(33817,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(33817,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(33817,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(33817,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(33817,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(33817,36,'cp 100','f1e2 e7d8 e1g1 f8e7 e2d3 e4f6 c2c4 d6d5 b1c3 d5c4 d3c4 e8g8 a2a3 c8g4 h2h3 g4f3 d1f3 d8d4 f3b7 b8d7 c4a2 d4e5 b7f3 e7d6 g2g3 a8b8 f1d1 a7a5 g1g2 h7h6 a1b1 f8e8 c1e3 d7c5 c3d5 c5e4 d5f6 e5f6 f3f6');
INSERT INTO evaluations VALUES(33815,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(33815,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(33815,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(33815,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(33815,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(33815,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(33815,36,'cp 83','e7d8 e1g1 f8e7 e2d3 e4f6 c2c4 d6d5 b1c3 d5c4 d3c4 e8g8 f1e1 c8g4 h2h3 g4f3 d1f3 b8c6 e1d1 c6d4 f3b7 c7c5 c1f4 e7d6 f4d6 d8d6 c3b5 d6f4 b5d4 c5d4 c4f7 g8h8 g2g3 f4e5 d1e1 e5f5 f7e6 f5g6 b7f3 a8d8 a1d1 d4d3');
INSERT INTO evaluations VALUES(5836,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(5836,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(5836,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(5836,33,'cp 204',NULL);
INSERT INTO evaluations VALUES(5836,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(5836,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(5836,36,'cp 201','c2c3 b4f8 b2b4 f7f6 b4b5 c6a5 f1d3 c7c6 a2a4 c6c5 c1e3 c5c4 d3c2 b7b6 e1g1 f6f5 b1d2 g7g6 f3g5 d8e7 d2f3 g8h6 h2h3 h6f7 g5f7 e7f7 f3g5 f7e7 d1d2 f8h6 g2g4');
INSERT INTO evaluations VALUES(5835,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(5835,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(5835,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(5835,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(5835,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(5835,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(5835,36,'cp 189','b4f8 b2b4 a7a5 b4b5 c6a7 f1d3 c8d7 a2a4 h7h6 b1a3 g8e7 e1g1 c7c6 b5c6 d7c6 a3b5 e7c8 c1e3 c8b6 f3d2 a7b5 a4b5 c6d7 d1g4 g7g6 c3c4 d5c4 d2c4 b6d5 f1c1 h6h5 g4g3 h5h4 g3g4 f8e7 c4d6 e7d6 e5d6');
INSERT INTO evaluations VALUES(51923,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(51923,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(51923,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(51923,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(51923,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(51923,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(51923,36,'cp -87','b8c6 d2d3 d7d5 c4b3 d5e4 f3e4 g8f6 e4h4 f8e7 g1f3 a7a6 h4g3 b7b5 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 b1a3 d8b6 e1g1 e8g8 c1h6 f6e8 f1a1 a8b7 c2c3 e7f6 h6g5 c6e7 g5f6 e8f6');
INSERT INTO evaluations VALUES(32008,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(32008,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(32008,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(32008,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(32008,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(32008,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(32008,36,'cp 99','d2d4 c7c6 h2h3 g4f3 d1f3 g8f6 c1e3 e7e6 a2a3 f6d5 c3d5 c6d5 f1d3 f8e7 h3h4 h7h5 c2c3 g7g6 g2g3 b8c6 e1g1 e8g8 a1e1 e7f6 g1g2 e6e5 d4e5 c6e5 f3d1 e5d3 d1d3 d8d7 e1d1 f8d8 e3d4 f6d4 d3d4 d7b5 f1e1 b5b2 d1b1 b2a3 b1b7');
INSERT INTO evaluations VALUES(32001,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(32001,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(32001,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(32001,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(32001,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(32001,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(32001,36,'cp 84','c7c6 h2h3 g4f3 d1f3 g8f6 c1e3 e7e6 a2a3 f8e7 f1d3 b8d7 e1c1 e8g8 c1b1 b7b5 f3c6 b5b4 a3b4 e7b4 c3a2 a8b8 c6a4 b4d6 c2c4 e6e5 a2c3 e5d4 e3d4 d6c5 d4c5 d7c5 d3h7 f6h7 d1d8 c5a4 d8f8 h7f8 c3a4 b8b4 a4c3 b4c4 c3b5 c4e4');
INSERT INTO evaluations VALUES(31208,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(31208,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(31208,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(31208,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(31208,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(31208,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(31208,36,'cp 98','h2h3 g4f3 d1f3 e7e6 c1e3 g8f6 a2a3 f6d5 f1d3 f8e7 c3d5 c6d5 c2c3 b8c6 e1g1 e8g8 a1e1 e7f6 f3g3 a8c8 e1e2 c6a5 e3h6 g8h8 h6c1 a5c4 a3a4 h8g8 c1h6 d8d7 h3h4 a7a5 h6f4 d7a4');
INSERT INTO evaluations VALUES(46688,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(46688,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(46688,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46688,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(46688,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46688,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(46688,36,'cp 30','b1c3 f8g7 e2e4 e8g8 d1d2 b8c6 e1c1 c8g4 f1e2 g4f3 e2f3 e7e5 d4e5 d6e5 f4e3 d8c8 c1b1 f8e8 f3e2 c6d4 f2f3 c7c6 c3a4 b7b6 c2c3 d4e2 d2e2 f6h5 g2g3 g7f8 b2b3 h5g7 a4b2 c8c7 d1d2 e8d8 h1d1 d8d2 d1d2 g7e6 h2h4 h7h5 b2d3');
INSERT INTO evaluations VALUES(46678,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(46678,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(46678,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(46678,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(46678,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(46678,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(46678,36,'cp 42','f8g7 d1d2 e8g8 e2e4 c8g4 e1c1 b8c6 f4e3 g4f3 g2f3 e7e5 d4e5 d6e5 h2h4 d8d2 d1d2 f8d8 d2d8 a8d8 c3a4 b7b6 c2c3 a7a5 c1c2 g7f8 f1b5 c6a7 b5a6 f6h5 b2b3 h5f4 a4b2 f7f6 e3f4 e5f4 a6b7 f8c5 b2d3 c5d6 a2a4 g8g7 h4h5 g6g5 h1d1 g7f8');
INSERT INTO evaluations VALUES(41305,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(41305,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(41305,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(41305,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41305,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(41305,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(41305,36,'cp 26','e2e4 e8g8 d1d2 c8g4 e1c1 b8c6 f4e3 e7e5 d4e5 d6e5 d2d8 a8d8 d1d8 f8d8 f3d2 a7a6 f2f3 g4e6 f1c4 e6c4 d2c4 b7b5 c4d2 c6d4 d2b3 g7f8 b3a5 d8d7 c3d1 f6e8 d1f2 f7f6 c2c3 d4e6 h1d1 d7d1 c1d1 g8f7 f2d3 e8d6 b2b3 g6g5');
INSERT INTO evaluations VALUES(41392,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41392,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(41392,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(41392,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(41392,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(41392,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(41392,36,'cp 24','e8g8 d1d2 c8g4 e1c1 b8c6 f4e3 e7e5 d4e5 d6e5 d2d8 f8d8 d1d8 a8d8 f3d2 g4e6 f2f3 f6h5 a2a4 a7a6 f1c4 e6c4 d2c4 f7f5 h1d1 d8d1 c1d1 h5f6 c3d5 f5e4 d5f6 g7f6 f3e4 f6e7 d1e2 b7b5 a4b5 a6b5 c4d2 g8f7 e2d3 c6b4 d3c3 f7e6 d2f3 b4c6 e3f2 c6b4');
INSERT INTO evaluations VALUES(36680,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36680,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36680,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(36680,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(36680,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(36680,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(36680,36,'cp 36','d1d2 c8g4 e1c1 b8c6 f1e2 g4f3 e2f3 e7e5 d4e5 d6e5 f4e3 d8c8 c1b1 c6d4 e3d4 e5d4 d2d4 f6g4 d4d2 g7c3 b2c3 f8e8 b1a1 g4f6 d2d4 c8e6 e4e5 e6e5 f3b7 a8b8 d4e5 e8e5 b7f3 b8b6 h1e1 e5c5 e1e3 h7h5 d1d4 c5a5 d4d8 g8g7 c3c4 b6b4 f3e2 b4a4');
INSERT INTO evaluations VALUES(36681,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(36681,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(36681,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(36681,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(36681,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(36681,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36681,36,'cp 37','b8c6 e1c1 c8g4 f1e2 g4f3 e2f3 e7e5 d4e5 d6e5 f4e3 d8c8 c1b1 f8e8 c3d5 f6d5 e4d5 c6d4 f3e4 d4f5 f2f3 f5d6 h1e1 a7a5 c2c3 c8d7 d2e2 f7f5 e4d3 e5e4 f3e4 f5e4 d3c2 b7b5 g2g4 e8f8 h2h4 b5b4 c3b4 a5b4 h4h5 f8f3 h5h6 g7h8 e3c5 a8e8 c5b4 d7g4 a2a3 g4h3 e2d2');
INSERT INTO evaluations VALUES(17646,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(17646,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(17646,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(17646,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17646,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(17646,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(17646,36,'cp 43','b5a3 g8f6 a3c4 b7b5 c4e3 f8e7 f1d3 c6b4 d3e2 c8e6 e1g1 e8g8 a2a3 b4c6 e3d5 c6d4 e2d3 f6d5 c3d5 e7g5 c2c3 g5c1 a1c1 d4c6 d5e3 d8b6 h2h3 a8c8 b2b4 c8c7 d1e2 c6e7 c3c4 f8c8 f1d1 b6d4 e3d5 e7d5');
INSERT INTO evaluations VALUES(19317,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(19317,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(19317,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(19317,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(19317,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(19317,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19317,36,'cp 12','g8f6 b1c3 f8g7 f1c4 e8g8 c4b3 d7d6 f2f3 c8d7 d1d2 c6d4 e3d4 b7b5 a2a4 b5a4 b3a4 a7a5 e1g1 a8b8 b2b3 b8b4 d4f2 d7e6 f1e1 f6d7 a4d7 e6d7 c3d5 b4b5 a1a3 d7c6 c2c4 b5b7 f2d4 g7d4 d2d4 c6d5 e4d5 d8c7 d4c3 f8b8 c3a5');
INSERT INTO evaluations VALUES(12740,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12740,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(12740,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(12740,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(12740,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(12740,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(12740,36,'cp 25','b1c3 f8g7 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 f1d1 d3c4 b5a7 c4b3 a2b3 e7e6 g2g3 h6g7 d4g7 g8g7 a7c6 a8a1 d1a1 f8c8 c6e5 c8c2 a1a7 g7f6 e5f7 c2b2 f7d6 b2b3 a7h7');
INSERT INTO evaluations VALUES(12736,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(12736,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(12736,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(12736,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(12736,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(12736,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(12736,36,'cp 28','f8g7 f1c4 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1d1 b5c4 d4c5 f8d8 d1d4 c4d5 c5e7 h6g7 e7d8 g7d4 b3d4 a8d8 f2f3 d5a2 d4c6 d8d6 c6e7 g8f8 a1a2 f8e7 a2a7 e7f6 a7b7 h7h5 g1f1 d6d1 f1e2 d1h1 h2h3 g6g5 b7b3 h5h4 e2e3 h1d1 b3b4 d1b1 e3d4');
INSERT INTO evaluations VALUES(22090,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(22090,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(22090,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(22090,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(22090,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(22090,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(22090,36,'cp 113','c1e3 g8f6 h2h3 c7c6 c2c3 b4a6 f1a6 b7a6 d1a4 d8c7 d5c6 f8e7 b1d2 a8b8 e1g1 e8g8 c3c4 a6a5 b2b3 b8b4 a4a3 b4b8 a1c1 c8e6 f1e1 f8d8 e3g5 h7h6 g5h4 d8c8 c1d1 d6d5 a3c1 d5c4 d2c4 c7c6 h4g3 e7b4 g3e5 b4e1 d1e1 e6c4 e5b8 c4b5 c1c6 b5c6 b8a7 f6e4');
INSERT INTO evaluations VALUES(21003,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(21003,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(21003,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(21003,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(21003,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(21003,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(21003,36,'cp -49','e7e5 g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 f1e2 c8e6 b1c3 d8d7 a2a3 f8d6 d1c2 d7f7 e1g1 d5c4 c3b5 a7a6 b5d6 c7d6 d2c4 f7c7 c1d2 a8c8 f1c1 c7d8 c2d1 e8g8 d1e1 f6d5 c4a5 c6a5 d2a5 d8g5 c1c8 f8c8 a1c1 c8c1 e1c1 f5f4 e3f4 d5f4');
INSERT INTO evaluations VALUES(48273,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48273,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(48273,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(48273,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(48273,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(48273,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(48273,36,'cp 0','f8e7 g1f3 e8g8 b1c3 c7c5 d4c5 d5c4 d1d8 f8d8 f1c4 e7c5 c1d2 b8c6 e1g1 c8d7 f1d1 c5e7 c4e2 d7e8 d2e1 d8d1 a1d1 a8c8 c3b5 f6d5 b5c3 d5f6');
INSERT INTO evaluations VALUES(40972,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40972,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40972,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(40972,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40972,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40972,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40972,36,'cp 0','g1f3 e8g8 b1c3 b8d7 c4d5 e6d5 f1d3 f8e8 d1c2 c7c6 e1g1 e7d6 f1e1 h7h6 h2h3 d7b6 a3a4 a7a5 c1d2 c8e6 f3e5 a8c8 e5f3');
INSERT INTO evaluations VALUES(40973,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40973,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40973,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40973,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40973,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40973,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40973,36,'cp 0','e8g8 b1c3 b8d7 c4d5 e6d5 f1d3 f8e8 e1g1 e7d6 f1e1 c7c6 d1c2 h7h6 h2h3 d7b6 a3a4 a7a5 c1d2 c8e6 f3e5 a8c8 e5f3');
INSERT INTO evaluations VALUES(36340,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36340,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36340,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36340,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36340,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36340,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36340,36,'cp 0','b1c3 b8d7 c4d5 e6d5 f1d3 c7c6 e1g1 e7d6 d1c2 h7h6 f1e1 f8e8 h2h3 d7b6 a3a4 a7a5 c1d2 c8e6 f3e5 a8c8 e5f3');
INSERT INTO evaluations VALUES(36338,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36338,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36338,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36338,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36338,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36338,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36338,36,'cp 0','b8d7 c4d5 e6d5 f1d3 c7c6 e1g1 e7d6 d1c2 f8e8 f1e1 d7b6 h2h3 h7h6 c1d2 c8e6 a3a4 d8c8 f3e5 e6h3 g2h3 c8h3 d3f5 h3h4 f2f4 d6e5 f4e5 h4g3 g1h1 g3f3 h1g1');
INSERT INTO evaluations VALUES(3994,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3994,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3994,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3994,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3994,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3994,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3994,36,'cp 0','c4d5 e6d5 f1d3 f8e8 e1g1 e7d6 f1e1 h7h6 c3b5 d6f8 b5c3');
INSERT INTO evaluations VALUES(47324,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(47324,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(47324,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(47324,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(47324,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(47324,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(47324,36,'cp 40','a7a6 d2d4 b7b5 c4d3 c8b7 d1e2 e7e6 c1g5 b8d7 d4d5 d7c5 d5e6 f7e6 e1g1 f8d6 g5f6 d8f6 c3e4 c5e4 d3e4 b7e4 e2e4 e8g8 f1e1 a8e8 c2c3 h7h6 e1e2 f6f5 a1e1 f5e4 e2e4 f8f6 h2h3 g8f7 a2a3 f6f4 e4f4 d6f4 g2g3 f4d6 g1g2 e8c8 h3h4 c7c5');
INSERT INTO evaluations VALUES(43600,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(43600,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(43600,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(43600,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(43600,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(43600,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(43600,36,'cp 45','f3e5 e7e6 d2d4 b8d7 c1f4 d7e5 f4e5 f6d7 e5g3 d7f6 c4e2 f8d6 d1d3 e8g8 g3h4 a8b8 e1c1 b7b5 e2f3 d6e7 h1e1 b5b4 c3a4 c8b7 f3b7 b8b7 d3a6 d8d5 f2f3 b7b8 b2b3 d5h5 h4g3 f6d5 a6e2 b8a8 g3e5 h5h6 c1b2');
INSERT INTO evaluations VALUES(38615,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(38615,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(38615,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(38615,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38615,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(38615,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(38615,36,'cp -12','g8f6 c3b5 d6b4 c2c3 b4a5 f1d3 e8g8 e1g1 c7c6 b5a3 c8g4 a3c2 a5c7 c2e3 g4h5 e3f5 f8e8 c1g5 b8d7 f5g3 c7g3 h2g3 h5g6 g5f4 g6d3 d1d3 h7h6 a1e1 d8a5 a2a3 e8e4 f3d2 e4e1 f1e1 a8e8 e1e8 f6e8 d2f3 e8f6 f3e5 d7e5 f4e5 f6d7 e5d6 a5b6 d3c2');
INSERT INTO evaluations VALUES(51540,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(51540,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(51540,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(51540,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(51540,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(51540,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(51540,36,'cp 200','f1b5 c8b7 d2d3 e7e6 g1f3 c7c5 e1g1 f8e7 c2c4 d8c7 b1c3 a7a6 b5a4 b8c6 f1e1 g8f6 c1d2 e8g8 c3d5 e6d5 e4d5 f8e8 d5c6 b7c6 a4c6 c7c6 f3e5 c6c7 h2h3 d7d6 e5f3 h7h6 a1c1 c7b7 d1a4 f6d7 c1d1 e7f6 b2b3 d6d5 e1e8 a8e8');
INSERT INTO evaluations VALUES(51526,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(51526,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(51526,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(51526,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(51526,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(51526,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(51526,36,'cp 176','c8b7 d2d3 c7c5 g1f3 e7e6 e1g1 f8e7 c2c4 d8c7 f1e1 b8c6 b1c3 g8f6 c1d2 a7a6 b5a4 e8g8 c3d5 e6d5 e4d5 f8e8 d5c6 b7c6 a4c6 c7c6 b2b3 d7d5 a1c1 e7f8 d1c2 h7h6 h2h3 d5d4 e1e8 a8e8 c1e1 e8e1 f3e1 f6d7 c2d1 d7e5 d2f4 e5g6 f4g3');
INSERT INTO evaluations VALUES(30843,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(30843,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(30843,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(30843,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(30843,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(30843,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(30843,36,'cp 221','d2d3 g8f6 g1f3 c7c5 e1g1 e7e6 c2c4 b8c6 b1c3 f8e7 e4e5 f6g4 c1f4 d8b6 h2h3 g4h6 f4h6 g7h6 c3e4 h8g8 b5c6 b7c6 d1d2 g8g6 g2g3 b6b4 d2e2 e8f8 a1d1 f8g8 f1e1 b4b7 d3d4 a7a5 d4c5 a5a4 f3d4 c6e4 e2e4 b7e4 e1e4 e7c5 g1g2');
INSERT INTO evaluations VALUES(11759,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(11759,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(11759,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(11759,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(11759,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(11759,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(11759,36,'cp 96','e7g6 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 d2d3 f8e7 g2g3 c6a5 b3a2 b5b4 h2h4 h7h5 c2c3 b4b3 a2b3 a5b3 d1b3 a8c6 c1g5 f7f6 g5e3 d7d5 b1d2 e8g8 f1a1 g8h8 b3c2 d8d7 b2b4 f8d8 d2b3 d5e4 d3e4 d7g4 f3h2');
INSERT INTO evaluations VALUES(11716,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(11716,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(11716,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(11716,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(11716,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(11716,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(11716,36,'cp 109','a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 b1c3 b5b4 c3d5 f8d6 c2c3 e8g8 d5e3 c6a5 b3c2 f8e8 f1e1 c7c5 g2g3 h7h6 d2d3 b4c3 b2c3 d6f8 h2h4 d7d6 c1d2 a8c6 d1e2 c6d7 e1a1 a5c6 c2b3 g6e7 h4h5 d8c7 f3h4 e8b8 b3c4 c6a5 c4a2');
INSERT INTO evaluations VALUES(12776,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12776,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12776,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12776,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12776,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12776,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(12776,36,'cp 0','g1f3 e7e6 d2d4 f6e4 c3e4 d7d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 f2f3 e4e3 c4b5 e8f7 d2c4 a7a6 b5a4 f8c5 c1e3 c5e3 c4e3 b7b5 a4b3 f7f6 e1c1 d8f7 h1e1 h8d8 d1d8 f7d8 e1d1 f6e7 h2h4 g7g6 d1e1 e7f6 e1d1');
INSERT INTO evaluations VALUES(29527,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(29527,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(29527,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(29527,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(29527,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(29527,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(29527,36,'cp 114','c4d5 f5b1 d1a4 d8d7 a4d7 b8d7 a1b1 f6d5 e2e4 d5b6 c1d2 e7e6 b2b4 f8d6 f1d3 e8g8 e1e2 c7c6 h1c1 f8e8 d2e3 e8c8 c1c2 a7a5 b4a5 a8a5 f3d2 d6c7 d2c4 a5a8 f2f4 b6c4 d3c4 c7b6 c4b3 d7f6 f4f5 f6e4 f5e6');
INSERT INTO evaluations VALUES(43535,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(43535,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(43535,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(43535,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(43535,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(43535,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(43535,36,'cp 40','f2f3 d6d5 e4d5 f6d5 c3d5 d8d5 f1d3 d5a5 c2c3 e6e5 d4b3 a5c7 d1c2 c8e6 e1c1 b8d7 h2h4 f8e7 g2g4 g7g6 h4h5 e8c8 c2f2 f7f5 g4f5 g6f5 f3f4 e5e4 b3d4 d7c5 d3c2 e6d7 h1g1 h8g8 g1g8 d8g8 d1g1 g8g1 f2g1 c5d3 c2d3 e4d3 g1g8 e7d8');
INSERT INTO evaluations VALUES(43537,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(43537,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43537,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(43537,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(43537,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(43537,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(43537,36,'cp 37','h7h5 f1c4 d8c7 c4b3 b7b5 a2a4 b5b4 c3a2 e6e5 d4e2 c7b7 c2c3 b8c6 c3b4 c6b4 e1g1 f8e7 a2b4 b7b4 e2c3 a8b8 b3a2 c8e6 f1f2 e8g8 f2d2 b4b7 a4a5 f8c8 h2h3 e6a2 a1a2 b7b3 e3b6 b3d1 d2d1 b8b7 a2a1 h5h4 d1c1 e7d8 c3d5 c8c1 a1c1 f6d5 b6d8');
INSERT INTO evaluations VALUES(16545,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(16545,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(16545,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(16545,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(16545,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(16545,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(16545,36,'cp -25','b1c3 f8c5 e1g1 e8g8 h2h3 f8e8 f1e1 c6d4 f3d4 c5d4 d3c4 c7c6 d1f3 d7d6 d2d3 b7b5 c4b3 a7a5 a2a4 c8e6 b3e6 f7e6 c3e2 d4b6 e1f1 e8f8 e2g3 b5b4 c1e3 f6d7 f3g4 d8e7 e3b6 d7b6 g3h5 b6d7 f2f4 g8h8 f4e5 d7e5 g4g3 e5g6');
INSERT INTO evaluations VALUES(16520,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(16520,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(16520,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(16520,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(16520,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(16520,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(16520,36,'cp -12','f8c5 e1g1 e8g8 h2h3 a7a5 d3c4 d7d6 d2d3 a5a4 a2a3 c6d4 f3d4 c5d4 c3e2 d4b6 e2g3 c8e6 c4e6 f7e6 g3h5 d8e8 h5f6 f8f6 c1g5 f6f7 d1g4 e8b5 g5e3 f7f6 a1b1 b5c6 e3b6 c6b6 b2b3 a4b3 b1b3 b6a7 b3b4 b7b6');
INSERT INTO evaluations VALUES(10216,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(10216,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(10216,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(10216,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(10216,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(10216,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(10216,36,'cp -16','e1g1 d7d6 h2h3 e8g8 d3c4 h7h6 f1e1 c6d4 f3d4 c5d4 d2d3 c7c6 d1f3 c8e6 c3e2 d4b6 e2g3 d6d5 c4b3 a7a5 a2a4 f8e8 e4d5 e6d5 b3d5 d8d5 f3d5 f6d5 c1d2 a8d8 e1e4 f7f6 g3f5 g8f7 a1e1 d5e7 f5e7 e8e7 d2e3');
INSERT INTO evaluations VALUES(10215,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(10215,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(10215,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(10215,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(10215,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(10215,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(10215,36,'cp -24','d7d6 a2a3 a7a5 d3c4 e8g8 h2h3 c6d4 d2d3 c7c6 f3d4 c5d4 c3e2 d4c5 e2g3 d6d5 c4a2 c8e6 d1e2 c5d6 e2f3 h7h6 g3f5 f8e8 e4d5 f6d5 f5d6 d8d6 f1e1 d5c7 b2b3 a5a4 b3a4 e6a2 a1a2');
INSERT INTO evaluations VALUES(9395,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(9395,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(9395,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(9395,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(9395,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(9395,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(9395,36,'cp -23','a2a3 e8g8 d3c4 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 d2d3 d6d5 e4d5 c6d5 c4a2 d4e6 c3a4 c5d6 c2c4 d5c4 d3c4 d8e7 a4c3 a8d8 c3d5 f6d5 c4d5 e6f4 f1e1 e7f6 c1e3 e5e4 d1d4 f6f5 e3f4 f5f4 d4e4 f4h2 g1f1');
INSERT INTO evaluations VALUES(34871,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(34871,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(34871,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(34871,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(34871,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(34871,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(34871,36,'cp 157','f1b5 c7c6 b5a4 e6g6 d1e2 f8e7 d2d3 g8f6 a4b3 a7a5 c1d2 b8d7 f3e5 d7e5 e2e5 b7b5 e1g1 e8g8 a2a4 f8e8 c3e4 f6e4 e5e4 g6e4 d3e4 e8d8 f1d1 c8a6 a4b5 a6b5 c2c4 a5a4 c4b5 a4b3 a1a8 d8a8 b5c6 a8c8 d2c3 c8c6 g1f1 c6c4 f2f3 e7f6');
INSERT INTO evaluations VALUES(34864,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(34864,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(34864,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(34864,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(34864,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(34864,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(34864,36,'cp 156','c7c6 b5a4 e6g6 d1e2 f8e7 d2d3 g8f6 a4b3 a7a5 c1d2 b8d7 f3e5 d7e5 e2e5 b7b5 e1g1 e8g8 a2a4 f8e8 c3e4 f6e4 e5e4 c8f5 e4f3 b5b4 a1e1 f5d7 b3c4 g6f6 f3f6 e7f6 b2b3 f6d4 e1e8 a8e8 d2e3 c6c5 h2h3 d7c6 f1e1 h7h6 e1e2 d4e3');
INSERT INTO evaluations VALUES(34447,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(34447,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(34447,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(34447,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(34447,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(34447,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(34447,36,'cp 176','b5a4 e6g6 d1e2 c8e6 e1g1 b8d7 d2d4 e5d4 f3d4 f8c5 d4e6 g6e6 e2e6 f7e6 c3e4 e6e5 a4b3 g8f6 e4g5 c5b6 g5e6 h8g8 c1d2 a7a5 a1d1 e8e7 f1e1 h7h6 e6f4 f6g4 f4g6 e7f6 d2e3 b6e3 f2e3 g8b8 d1d6 f6g5 h2h4 g5h5 d6d7 h5g6 h4h5 g6h7 e1f1 b8f8 f1f7 f8f7 b3f7');
INSERT INTO evaluations VALUES(34448,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(34448,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(34448,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(34448,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(34448,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(34448,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(34448,36,'cp 148','e6g6 d1e2 c8e6 e1g1 b8d7 d2d4 e6g4 a4b3 e8c8 f3e5 d7e5 e2e5 g8f6 e5g5 g4h5 g5g6 h5g6 c3e2 f6d5 f2f4 f7f6 f4f5 g6f7 c1d2 f8d6 a1e1 h8e8 e2c3 d6c7 c3e4 e8e7 e4g3 d8e8 e1e7 e8e7 g1f2 e7e8 f1h1 b7b5 h2h4 a7a5 h4h5');
INSERT INTO evaluations VALUES(53017,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(53017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53017,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(53017,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53017,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53017,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53017,36,'cp 0','b8c6 g2g3 g7g6 f1g2 f8g7 g1e2 a7a6 e1g1 a8b8 a2a4 c8g4 a1b1 g4e2 c3e2 d8b6 d2d3 g8f6 f2f4 e8g8 g1h1 b6c7 e2c3 e7e6 c1d2 f6d7 f4f5 c7d8 d2e3 g7d4 e3h6 d4g7');
INSERT INTO evaluations VALUES(19965,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19965,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19965,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19965,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19965,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19965,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19965,36,'cp 0','g2g3 g7g6 f1g2 a7a6 g1e2 f8g7 e1g1 a8b8 a2a4 c8g4 a1b1 g4e2 c3e2 d8b6 f2f4 g8f6 d2d3 b6c7 c1d2 e8g8 e2c3 c7d8 g1h1 e7e6 f4f5 f6d7 d2e3 g7d4 e3h6 d4g7');
INSERT INTO evaluations VALUES(19964,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19964,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19964,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19964,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19964,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19964,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19964,36,'cp 0','g7g6 f1g2 a8b8 g1e2 f8g7 d2d3 a7a6 c1e3 g8f6 a2a4 f6g4 e3c1 e8g8 e1g1 c6b4 c3d5 b4c6');
INSERT INTO evaluations VALUES(51858,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(51858,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(51858,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(51858,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(51858,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(51858,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(51858,36,'cp 60','d7d6 f1c4 f8e7 d2d4 b8d7 a2a4 g8f6 e1g1 e8g8 h2h3 h7h6 f1e1 f8e8 a4a5 e5d4 f3d4 e7f8 c1f4 d7e5 c4a2 c8d7 f4g3 b7b5 a5b6 d8b6 a2b3 a7a5 f2f4 e5g6 g1h2 a8d8 h2h1 b6b4 d1d2 d6d5 e4e5 f6h5 g3h2 f8c5 a1d1');
INSERT INTO evaluations VALUES(4284,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(4284,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(4284,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4284,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(4284,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4284,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4284,36,'cp 15','d7d6 b1c3 h7h6 g5h4 c8g4 h2h3 g4e6 c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 c2c3 d8e8 h4g3 e8g6 f3h4 g6h5 d1b3 h5e8 f1e1 f6h5 e1e3 c6a5 b3c2 b6b5 a2a3 c7c5 b2b4 a5c6 b4c5 d6c5 c2d1 c5c4 d3c4 b5c4 d1e2 e8f7 h4f3 h5g3 f2g3');
INSERT INTO evaluations VALUES(3796,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(3796,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(3796,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(3796,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(3796,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3796,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(3796,36,'cp 15','b1c3 h7h6 g5h4 c8g4 h2h3 g4e6 c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 c2c3 d8e8 f1e1 e8g6 e1e3 f8f7 h4f6 g6f6 a2a3 a8f8 d1e2 c6e7 d3d4 e7g6 d4e5 g6e5 a1f1 f6f4 f3e5 f4e5 a3a4 g8h8 g2g3 e5g5 h3h4 g5g6 g1g2');
INSERT INTO evaluations VALUES(34371,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(34371,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(34371,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(34371,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(34371,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(34371,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(34371,36,'cp 162','b1c3 d5d8 d2d4 g8f6 d4c5 d8d1 e1d1 c8g4 c1e3 b8d7 d1c1 e7e6 f1b5 a7a6 c5c6 a6b5 c6d7 f6d7 c3b5 a8c8 b2b3 g4f5 c2c3 d7c5 h1d1 f8e7 f3e1 e8g8 c1b2 f8d8 b3b4 c5d3 e1d3 f5d3 a2a4 e6e5 f2f4 f7f6 e3b6 d8d7 f4e5 f6e5');
INSERT INTO evaluations VALUES(34365,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(34365,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(34365,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(34365,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(34365,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(34365,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(34365,36,'cp 161','d5d8 d2d4 g8f6 d4c5 d8d1 e1d1 c8g4 c1e3 b8c6 a1b1 e8c8 d1c1 c6b4 a2a3 b4d5 c3d5 f6d5 e3d2 f7f6 b2b4 e7e5 f1c4 b7b6 c5b6 d5b6 c4b3 e5e4 f3g1 f8d6 h2h3 g4h5 g2g4 h5e8 g1e2 h7h5 g4h5 e8h5 e2d4 h5g4 d2e3 h8h3');
INSERT INTO evaluations VALUES(53223,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(53223,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(53223,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(53223,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(53223,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(53223,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(53223,36,'cp 151','d2d4 g8f6 d4c5 d8d1 e1d1 c8g4 c1e3 b8c6 d1c1 e8c8 a1b1 c6b4 a2a3 b4d5 c3d5 f6d5 e3d2 f7f6 b2b4 e7e5 f1b5 h7h5 b5c4 g7g5 h2h3 g4f5 h1d1 d5c7 c1b2 f8e7 d2e3 f5d7 d1d2 g5g4 f3h4 g4h3 h4g6 h8e8 g2h3 d7h3 b1h1 h3g4 g6e7 e8e7');
INSERT INTO evaluations VALUES(53205,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(53205,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(53205,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(53205,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(53205,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(53205,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(53205,36,'cp 149','g8f6 d4c5 d8d1 e1d1 c8g4 c1e3 b8c6 d1c1 e8c8 a1b1 c6b4 a2a3 b4d5 c3d5 f6d5 e3d2 f7f6 b2b4 e7e5 f1c4 b7b6 c5b6 d5b6 c4b3 a7a5 h2h3 g4h5 h1d1 c8b7 c1b2 a5a4 b3a2 e5e4 g2g4 h5e8 f3h4 h7h5 h4f5 h5g4 h3g4 h8h3 f5g3 e8c6 d2e3');
INSERT INTO evaluations VALUES(36473,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(36473,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(36473,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36473,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(36473,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36473,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36473,36,'cp 14','h7h6 g5c1 e7e6 f1d3 c7c5 e1g1 c5d4 f3d4 b8c6 c1e3 d6d5 e4d5 e6d5 h2h3 f8e8 a1c1 c6d4 e3d4 d5c4 d3c4 c8e6 c4e6 e8e6 d4e3 d8e7 d1f3 a8d8 f1d1 d8d1 c1d1 a7a6 b2b3 e6e5 c3d5 e5d5 d1d5 f6d5 f3d5 e7a3 g1h2 a3a2 d5b7 a2a1 b7b6 a1c3 g2g3');
INSERT INTO evaluations VALUES(36443,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36443,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36443,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(36443,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(36443,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36443,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(36443,36,'cp 14','g5c1 c7c5 f1e2 e7e6 e1g1 b8c6 d4d5 c6d4 d5e6 c8e6 f3d4 c5d4 c3b5 f6e4 b5d4 f8e8 c1e3 e6d7 e2f3 d8a5 d1d3 e4c5 d3c2 c5e6 f3d5 e6d4 e3d4 d7f5 c2d1 g7d4 d1d4 f5e6 d5e6 e8e6 b2b3 d6d5 c4d5 e6d6 d4e3 h6h5 a1d1 d6d5 d1d5 a5d5');
INSERT INTO evaluations VALUES(35459,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(35459,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(35459,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35459,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35459,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(35459,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(35459,36,'cp 19','f7f6 c1e3 b8d7 f3d2 a7a5 b1c3 g8h6 f2f3 f8c5 e3c5 d7c5 e1c1 c7c6 d2b3 c5d7 c3a4 d8c7 c2c4 b7b6 g2g3 h6f7 h2h4 f7d6 h4h5 h8d8 d1d2 d8f8 c4c5 b6c5 h5h6 g7g6 b3c5 d7c5 a4c5 a5a4 f1h3 a8a5 c5e6 c8e6 h3e6');
INSERT INTO evaluations VALUES(35473,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(35473,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(35473,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(35473,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(35473,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(35473,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(35473,36,'cp 8','b1c3 c8e6 a2a4 f8b4 a4a5 b4c3 b2c3 b8d7 f3d2 g8h6 f2f3 h6f7 c1a3 f7d6 f1e2 c7c5 d2f1 b7b5 f1e3 d8c7 e1f2 a7a6 h1d1 h7h5 a3c1 h5h4 e3d5 e6d5 d1d5 d6b7 c1e3 g7g5 g2g3 a8d8 f2g2 c7c8 f3f4 h4g3 h2g3');
INSERT INTO evaluations VALUES(35471,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(35471,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(35471,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(35471,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(35471,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(35471,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(35471,36,'cp 0','c7c6 c1e3 b8d7 a2a4 f8c5 e3c5 d7c5 a4a5 d8e7 f3d2 g8h6 c3d1 h8d8 f2f3 h6f7 d1e3 f7d6 b2b4 c5e6 e1c1 e6c7 c2c3 h7h5 h2h4 g7g6 g2g4 c8e6 g4h5 g6h5 h1g1 d8g8 f1c4 d6c4 d2c4 a8d8 c1c2 d8d1 g1d1 e6c4');
INSERT INTO evaluations VALUES(35446,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35446,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35446,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(35446,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35446,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(35446,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35446,36,'cp 3','a2a4 a7a5 c1e3 d8c7 h2h3 c8e6 f3d2 b8d7 f1c4 e6c4 d2c4 f8c5 e3d2 g8e7 c3e2 b7b6 e2c1 e7c8 c1d3 h8d8 e1e2 c8d6 c4d6 c5d6 d2e3 d6c5 d3c5 d7c5 f2f3 c5e6 h3h4 c6c5 a1d1 c7c6 d1d8 a8d8 h4h5 b6b5 a4b5 c6b5');
INSERT INTO evaluations VALUES(7599,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(7599,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(7599,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(7599,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(7599,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(7599,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(7599,36,'cp 25','f1d3 e8g8 e1g1 f8e8 h2h3 c5d4 c3e2 d4b6 e2g3 d7d5 e4d5 c6d5 d1f3 b6d4 c2c3 d4e5 c1f4 c7c5 f1e1 e5f4 f3f4 c8e6 e1e2 d8b8 f4f3 h7h6 a1e1 b8d6 g3f5 e6f5 e2e8 a8e8 e1e8 f6e8 f3f5 e8f6 f5c8 d6f8 c8c6 f8e8 c6e8 f6e8');
INSERT INTO evaluations VALUES(7600,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(7600,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(7600,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(7600,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7600,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(7600,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(7600,36,'cp 31','e8g8 c3a4 c5b6 e1g1 d7d5 e4e5 f6e4 d1e1 f7f5 a4b6 a7b6 b2b4 f5f4 d3e4 d5e4 e1e4 f4f3 f1e1 f3g2 a2a4 c8f5 e4c4 d8d5 c4d5 c6d5 c1e3 f5e4 a4a5 b6a5 a1a5 a8a5 b4a5 f8a8 e3d2 g8f7 e5e6 f7e8 c2c4 a8a6 e6e7 c7c6 d2b4 c6c5 b4d2 a6e6 c4d5 e4d5');
INSERT INTO evaluations VALUES(2872,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(2872,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(2872,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(2872,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(2872,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(2872,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(2872,36,'cp 37','c3a4 c5b6 e1g1 d7d5 e4e5 f6e4 d1e1 f8e8 d3e4 d5e4 a4b6 a7b6 e1e4 a8a5 c1f4 a5d5 e4f3 c8f5 c2c3 d8d7 f1d1 f5e4 d1d5 c6d5 f3e3 e4f5 e3d2 f5e6 a1d1 c7c5 h2h3 h7h6 a2a3 d7a4 g1h2 e8d8 f4g3 d5d4 c3d4 a4d4');
INSERT INTO evaluations VALUES(42693,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(42693,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(42693,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(42693,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(42693,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(42693,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42693,36,'cp -13','e2e3 d7d6 b1c3 a7a5 f1d3 e8g8 a2a3 c7c6 g1e2 c6d5 c4d5 b8a6 e1g1 a6c7 d3c2 b7b5 e3e4 f6h5 c1d2 c8a6 c3a2 b5b4 a3b4 a5b4 a2b4 c5b4 d2b4 h5f4 e2f4 a6f1 a1a8 c7a8 d1f1 e5f4 c2d3 a8c7 h2h3');
INSERT INTO evaluations VALUES(42691,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(42691,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(42691,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(42691,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(42691,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(42691,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(42691,36,'cp -17','e8g8 b1c3 d7d6 a2a3 a7a5 f1d3 c7c6 g1e2 b8a6 e1g1 a6c7 d1c2 c6d5 c4d5 c7d5 c3d5 f6d5 d3h7 g8h8 h7f5 d5e7 f5d3 c8e6 c1d2 d8d7 e2g3 f7f5 d2c3 c5b6 a1d1 e7d5 d3c4 d5c3 c2c3 a8c8 c4e6 d7e6 c3d3 c8d8 g3e2 d6d5');
INSERT INTO evaluations VALUES(37257,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(37257,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(37257,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(37257,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(37257,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(37257,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(37257,36,'cp -9','f1d3 d7d6 a2a3 c7c6 b1c3 a7a5 g1e2 c6d5 c4d5 b8a6 e1g1 a6c7 d1c2 c7d5 c3d5 f6d5 d3h7 g8h8 h7f5 d5e7 f5d3 c8d7 c1d2 a8c8 c2d1 f7f5 b2b4 a5b4 d2b4 d8e8 a1c1 b7b6 d3a6 c8a8 a6d3 e8g6 e2g3 d7e6 d1e2 e7d5 f1d1 c5b4 a3b4 d5b4 d3b1 d6d5');
INSERT INTO evaluations VALUES(14206,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(14206,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(14206,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(14206,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(14206,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14206,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(14206,36,'cp 26','c8f5 a2a3 e7e6 g1f3 h7h6 b2b4 c6e7 c4c5 g7g5 f1d3 f8g7 e1g1 a7a6 a3a4 f5d3 d1d3 e7g6 c1b2 g5g4 f3e1 c7c6 d3c2 d8e7 e1d3 g6h4 a1e1 h6h5 e3e4 f6e4 c3e4 d5e4 e1e4 a8d8');
INSERT INTO evaluations VALUES(24801,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(24801,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(24801,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(24801,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24801,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(24801,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(24801,36,'cp 32','a2a3 e7e6 g1f3 f8e7 f1d3 e8g8 e1g1 f5g6 b2b4 c6b8 c4c5 b8d7 c3e2 g6d3 d1d3 c7c6 d3c2 d8c7 c1d2 f8e8 d2c3 f6e4 a1b1 a7a6 f1c1 e4c3 c2c3 e7f6 a3a4 e6e5 b4b5 g7g6 c3b3 a6b5 a4b5 e5d4 e2d4 f6d4 f3d4');
INSERT INTO evaluations VALUES(24805,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(24805,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24805,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24805,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(24805,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(24805,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(24805,36,'cp 31','e7e6 g1f3 f8e7 f1d3 e8g8 e1g1 f5g6 b2b4 c6b8 c4c5 b8d7 c3e2 g6d3 d1d3 c7c6 c1b2 d8c7 f3d2 b7b6 d3c2 a7a5 b2c3 a5b4 a3b4 c7b7 e2f4 a8a1 f1a1 f8a8 a1a8 b7a8 f4d3 e7d8 d2b3 d8c7 c5b6 d7b6 b3c5 h7h6 d3e5 b6c4 e5c4');
INSERT INTO evaluations VALUES(25144,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(25144,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(25144,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(25144,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(25144,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25144,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(25144,36,'cp 38','g1f3 f8e7 f1d3 f5g6 e1g1 e8g8 b2b4 c6b8 c4c5 b8d7 c3e2 g6d3 d1d3 c7c6 c1b2 d8c7 f1c1 b7b6 b2c3 f6e4 a1b1 b6c5 b4c5 a7a5 e2g3 e4c3 d3c3 g7g6 g3e2 e7f6 g2g3 e6e5 d4e5 d7e5 f3e5 f6e5 e2d4 f8b8');
INSERT INTO evaluations VALUES(25145,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(25145,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(25145,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(25145,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(25145,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(25145,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(25145,36,'cp 33','f8e7 f1d3 e8g8 e1g1 f5d3 d1d3 c6a5 c4c5 b7b6 d3c2 b6c5 d4c5 d8b8 b2b4 a5c4 f3d4 a7a5 d4c6 b8e8 b4b5 e7c5 e3e4 e8d7 c1g5 f6g4 h2h3 g4e5 e4d5 e6d5 f1d1 d5d4 c3e4 c5b6 a3a4 e5c6 c2c4');
INSERT INTO evaluations VALUES(17173,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(17173,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(17173,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(17173,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(17173,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(17173,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(17173,36,'cp 1','d7d5 f3e5 d5e4 d1e2 c8e6 e5c6 b7c6 b5a6 d8d7 e1g1 e7g6 e2e4 f8e7 d2d4 e8g8 b1a3 f8e8 a3c4 e7f6 e4d3 c6c5 c4e5 g6e5 d4e5 d7d3 a6d3 f6e5 c1e3 e5d6 d3b5 e8d8 b5c6 a8b8 b2b3 f7f6 g2g3 g8f7 f1d1 d6e5 d1d8 b8d8 a1c1 e5d6 c3c4');
INSERT INTO evaluations VALUES(29661,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(29661,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(29661,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(29661,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(29661,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(29661,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(29661,36,'cp 48','f3e5 g4e6 c1f4 g7g6 d1d2 f8g7 e2e3 a7a6 f1e2 b8d7 e1g1 e8g8 a2a4 d7e5 f4e5 a6a5 f2f3 c7c6 b2b3 f6e8 e5g7 g8g7 e3e4 f7f6 e4e5 f6e5 d4e5 g7g8 e2d3 d8b6 g1h1 e8g7 c3e2');
INSERT INTO evaluations VALUES(29699,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(29699,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(29699,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(29699,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(29699,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(29699,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(29699,36,'cp 29','g4e6 c1f4 g7g6 d1d2 f8g7 e1c1 e8g8 e2e3 c7c5 h2h4 b8c6 f2f3 f6h5 f4h2 f7f5 f1e2 c5d4 e3d4 f5f4 h1e1 d8a5 a2a3 a7a6 e5c6 b7c6 e2c4 d5c4 e1e6 a8d8 h2g1 c6c5 d4d5 h5g3 e6e7 g3f5');
INSERT INTO evaluations VALUES(35187,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(35187,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(35187,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(35187,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(35187,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(35187,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(35187,36,'cp 163','g1f3 h4e4 f1e2 b8c6 d4e5 d7d5 e1g1 c8f5 e2d3 e4g4 h2h3 g4h5 g2g4 f5g4 h3g4 h5g4 g1h1 e8c8 f3g5 g4d1 f1d1 c6e5 c1f4 e5d3 d1d3 g8h6 b1c3 c7c6 a1g1 f8d6 f4d2 h8e8 d3f3 d5d4 c3b1 f7f5 g5h3 h6g4 f3f5 g7g6 f5f7 h7h5 d2g5 d8d7 f7d7 c8d7 b1d2 e8e2 g1f1 b7b6 h1g2 g4e3 g5e3');
INSERT INTO evaluations VALUES(35185,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(35185,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(35185,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(35185,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(35185,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(35185,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(35185,36,'cp 158','h4e4 f1e2 b8c6 d4e5 d7d5 e1g1 c8f5 e2d3 e4g4 h2h3 g4h5 g2g4 f5g4 h3g4 h5g4 g1h1 e8c8 f3g5 g4d1 f1d1 c6e5 c1f4 e5d3 d1d3 g8h6 b1c3 c7c6 c3b5 c8d7 b5d4 d8e8 c2c4 f7f6 g5e6 d5c4 e6f8 h8f8 d3d2 d7c8 f4h6 g7h6 d4f5 e8d8 f5d6 c8c7 a1g1 b7b5 g1g7 c7b6 g7h7 a7a5 h7h6 a5a4 h1g2 f8g8 g2f1');
INSERT INTO evaluations VALUES(35191,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(35191,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(35191,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(35191,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(35191,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(35191,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(35191,36,'cp 164','f1e2 b8c6 d4e5 d7d5 e1g1 c8f5 e2d3 e4g4 h2h3 g4h5 g2g4 f5g4 h3g4 h5g4 g1h1 e8c8 f3g5 g4d1 f1d1 c6e5 c1f4 e5d3 d1d3 g8h6 b1c3 c7c6 c3b5 c8d7 b5d4 d8e8 c2c4 f7f6 g5e6 d5c4 e6f8 h8f8 d3d2 d7c8 f4h6 g7h6 d4f5 c8c7 f5d6 e8d8 a1g1 f8f7 d6c4 d8d2 c4d2 f7d7 d2f1 f6f5');
INSERT INTO evaluations VALUES(35189,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(35189,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(35189,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(35189,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(35189,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(35189,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(35189,36,'cp 157','b8c6 d4e5 d7d5 e1g1 c8f5 e2d3 e4g4 h2h3 g4h5 g2g4 f5g4 h3g4 h5g4 g1h1 e8c8 f3g5 g4d1 f1d1 c6e5 c1f4 e5d3 d1d3 g8h6 b1c3 c7c6 c3b5 c8d7 b5d4 f8c5 a1d1 h8e8 h1g2 c5d4 d3d4 d7c8 d4a4 a7a6 f4e3 d8d6 g5f3 d6g6 g2f1 h6f5 e3b6 g6h6 a4f4 f5d6 f4g4 h6h1 f3g1');
INSERT INTO evaluations VALUES(50481,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(50481,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(50481,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(50481,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(50481,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(50481,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(50481,36,'cp -61','b7b5 c4e2 c8b7 d2d3 d8c7 e1g1 g8f6 c1g5 b8c6 a2a4 b5b4 c3b1 f8e7 b1d2 d7d5 f1e1 e8g8 e2f1 a6a5 g5h4 a8d8 e4e5 f6d7 h4e7 c6e7 d3d4 c5c4 g2g3 b7a6 c2c3 d8b8 h2h4 e7c6 d1e2 b4c3 b2c3 b8b2 h4h5');
INSERT INTO evaluations VALUES(51195,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(51195,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(51195,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(51195,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(51195,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(51195,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(51195,36,'cp -51','c4e2 d8c7 a2a3 c8b7 e1g1 f8e7 d2d4 c5d4 f3d4 g8f6 e2d3 e8g8 c1e3 e7c5 d1e2 d7d6 d4b3 c5e3 e2e3 h7h6 a1d1 b8d7 f1e1 a8c8 f2f3 d7e5 g1h1 f8d8 d1d2 e5c4 d3c4 b5c4 b3c1 c8b8 h2h3');
INSERT INTO evaluations VALUES(51215,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(51215,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(51215,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(51215,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(51215,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(51215,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(51215,36,'cp -62','c8b7 d2d3 d8c7 e1g1 b8c6 f1e1 f8d6 g2g3 g8e7 e2f1 e8g8 f1g2 h7h6 c1e3 a8c8 a2a3 f7f5 e4f5 e7f5 c3e4 c6d4 c2c3 d4f3 g2f3 a6a5 d1e2 f5e3 e2e3 d6e7 f3g2 b5b4 e4d2 e7g5 f2f4 g5f6 a1c1 b4c3 b2c3 c8b8 d2c4 b7g2 g1g2');
INSERT INTO evaluations VALUES(30333,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(30333,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(30333,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(30333,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(30333,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(30333,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(30333,36,'cp -39','d2d3 d8c7 e1g1 b8c6 f1e1 f8d6 c3b1 c6d4 b1d2 g8e7 e2f1 e8g8 a2a4 d4f3 d2f3 e7g6 a4b5 a6b5 a1a8 f8a8 c2c3 a8a1 g2g3 h7h6 d1e2 a1a2 c1e3 b7c6 e2c2 c7a5 h2h4 a5a4 c2d2 b5b4 c3b4 a4b4 d2b4 c5b4 e1b1 g6e5 f3d2');
INSERT INTO evaluations VALUES(30334,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(30334,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(30334,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(30334,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(30334,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(30334,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(30334,36,'cp -46','d8c7 e1g1 b8c6 f1e1 f8d6 c3b1 c6d4 b1d2 d4f3 d2f3 g8e7 a2a4 e8g8 e2f1 e7g6 a4b5 a6b5 a1a8 f8a8 c2c3 h7h6 g2g3 a8a1 d1b3 b7c6 c1e3 a1a4 b3c2 c7b7 h2h4 b7a8 h4h5 g6f8 f3d2 d6e7 d3d4 c5c4 d4d5 e6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(13832,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(13832,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(13832,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(13832,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(13832,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13832,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(13832,36,'cp 23','a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 a8b8 a3c2 f6g5 a2a4 b5a4 c2b4 c6b4 c3b4 c8d7 f1a6 e8g8 e1g1 d7c6 a6c4 d8d7 b2b3 a4b3 c4b3 c6d5 b3d5 b8b4 a1a6 g7g6 a6d6 d7e7 d6c6 f8b8 g2g3 b4b6 c6b6 b8b6 h2h4 g5f6 d1c2 g8g7 g1g2 h7h5 c2c3 b6b8 f1a1 b8b6');
INSERT INTO evaluations VALUES(11213,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(11213,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(11213,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11213,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(11213,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(11213,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(11213,36,'cp 22','b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 a8b8 a3c2 f6g5 a2a4 b5a4 c2b4 c6b4 c3b4 e8g8 a1a4 f7f5 e4f5 f8f5 f1c4 g8h8 e1g1 e5e4 d1d4 d8e8 d4a7 b8b7 a7a8 e4e3 f2e3 f5f1 c4f1 g5e3 d5e3 e8e3 g1h1 b7c7 a4a6 e3f4 a6a1 f4b4 a8d5');
INSERT INTO evaluations VALUES(11221,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(11221,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(11221,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11221,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(11221,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(11221,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11221,36,'cp 27','b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 f6g5 a3c2 e8g8 a2a4 b5a4 a1a4 a6a5 f1c4 a8b8 b2b4 c8d7 a4a3 g8h8 e1g1 a5b4 c2b4 d8c8 c4a6 c8e8 d1h5 h7h6 f1a1 c6b4 c3b4 d7e6 a6c4 e6d5 c4d5 b8b4 g2g3 b4b8');
INSERT INTO evaluations VALUES(11800,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(11800,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(11800,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(11800,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(11800,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(11800,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11800,36,'cp 31','c3d5 f8e7 g5f6 e7f6 c2c3 e8g8 a3c2 f6g5 a2a4 b5a4 a1a4 a6a5 f1c4 a8b8 b2b4 c8d7 a4a1 a5b4 c2b4 g8h8 e1g1 d8c8 c4a6 c8e8 d1e2 c6e7 d5e7 g5e7 f1d1 d7e6 h2h3 h7h6 b4d5 e7g5 a6c4 e8c6 a1a7 b8c8 c4a6 c6c5 a6c8');
INSERT INTO evaluations VALUES(11795,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(11795,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(11795,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11795,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11795,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(11795,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(11795,36,'cp 25','f8e7 g5f6 e7f6 c2c3 f6g5 a3c2 e8g8 a2a4 b5a4 a1a4 a6a5 f1c4 g8h8 c2e3 a8b8 b2b3 g7g6 e1g1 f7f5 d1d3 c8e6 f1a1 d8d7 e3f1 g5h4 d3e2 h4d8 f1d2 c6e7 d5e7 d7e7 b3b4 f5e4 c4e6 e7e6 b4a5 d6d5 a4b4');
INSERT INTO evaluations VALUES(7507,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(7507,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(7507,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(7507,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(7507,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(7507,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(7507,36,'cp 40','g5f6 e7f6 c2c3 e8g8 a3c2 f6g5 a2a4 b5a4 a1a4 g8h8 f1c4 f7f5 e4f5 c8f5 e1g1 a6a5 c2e3 f5e6 b2b3 a8b8 a4a1 e6f7 e3f5 f7g6 f5g3 c6e7 d5e7 g5e7 c4d5 d8c7 c3c4 c7b6 g3e4 e7d8 g2g3 b6b3 d1b3 b8b3 e4d6');
INSERT INTO evaluations VALUES(7505,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(7505,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(7505,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(7505,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(7505,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(7505,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(7505,36,'cp 28','e7f6 c2c3 f6g5 a3c2 e8g8 a2a4 b5a4 a1a4 g8h8 f1c4 f7f5 e4f5 c8f5 e1g1 a6a5 b2b4 a8b8 b4a5 c6a5 c4d3 f5d3 d1d3 a5b3 a4g4 b3c5 d3e2 g5h6 g4b4 d8e8 b4b8 e8b8 c2b4 e5e4 b4a6 c5a6 e2a6 b8e8 a6e2 e8e5 c3c4');
INSERT INTO evaluations VALUES(7517,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(7517,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(7517,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7517,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(7517,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(7517,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(7517,36,'cp 20','c2c3 e8g8 a3c2 f6g5 a2a4 b5a4 a1a4 g8h8 f1c4 f7f5 e4f5 c8f5 e1g1 a6a5 c2e3 f5e6 h2h3 a8b8 b2b3 h7h6 d1e2 d8d7 f1d1 g5h4 g2g3 h4d8 h3h4 c6e7 d5e7 d7e7 c4e6 e7e6 e2c4 e6f6 a4a2 d8b6 c4e4 b6e3');
INSERT INTO evaluations VALUES(7516,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(7516,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(7516,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(7516,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(7516,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(7516,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(7516,36,'cp 29','e8g8 a3c2 f6g5 a2a4 b5a4 a1a4 g8h8 f1c4 f7f5 e4f5 c8f5 e1g1 a6a5 b2b4 a8b8 b4a5 c6a5 c4d3 f5d3 d1d3 a5b3 a4b4 b3c5 d3e2 b8c8 b4g4 g5h6 g4g3 c5d7 h2h3 d7f6 d5f6 f8f6 f1e1 f6f4 g3f3 d8b6 e1d1');
INSERT INTO evaluations VALUES(2835,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(2835,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(2835,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(2835,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(2835,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(2835,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2835,36,'cp 21','a3c2 f6g5 a2a4 b5a4 a1a4 g8h8 f1c4 f7f5 e4f5 c8f5 c2e3 f5e6 e1g1 a6a5 b2b3 h7h6 d1e2 g5e3 e2e3 c6e7 d5e7 d8e7 c4e6 e7e6 c3c4 d6d5 c4d5 e6d5 f1b1 d5f7 a4c4 f7a7 e3e2 a8b8 g2g3');
INSERT INTO evaluations VALUES(2834,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(2834,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(2834,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(2834,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(2834,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(2834,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2834,36,'cp 25','f6g5 a2a4 b5a4 a1a4 g8h8 f1c4 f7f5 e4f5 c8f5 e1g1 a6a5 c2e3 f5d7 a4a1 a8b8 d1c2 d7e6 b2b3 d8d7 f1d1 g5h4 d1d2 h4g5 c2d3 e6g8 d2a2 h7h6 c4b5 d7b7 b5c6 b7c6 c3c4 g8h7 d3c3 c6b7 a2a5 b7b3');
INSERT INTO evaluations VALUES(2836,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(2836,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(2836,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(2836,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(2836,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(2836,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(2836,36,'cp 32','a2a4 b5a4 a1a4 a6a5 f1c4 g8h8 h2h4 g5h6 c2e3 a8b8 b2b3 h6e3 d5e3 c6e7 e1g1 f7f5 e4f5 e7f5 e3f5 c8f5 h4h5 h7h6 c4d5 f5h7 a4a3 h7g8 d5g8 f8g8 c3c4 g8f8 d1d2 b8b4 f1a1 d8f6 a3a5 b4b3 a5d5 f6f4 d2f4 f8f4 f2f3');
INSERT INTO evaluations VALUES(10379,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(10379,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10379,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(10379,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10379,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10379,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(10379,36,'cp 7','b4c3 b2c3 f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 e8g8 f1e1 f6d7 d2d3 f8e8 c1f4 d7e5 f4e5 c8e6 e5f4 d8d7 d1f3 f7f6 f3g3 e6f7 h2h3 e8e1 a1e1 a8c8 g3g4 d7g4 h3g4 c8e8 e1e8 f7e8 f4c7 c6c5 c7b8 a7a5 b8c7 e8a4 c7a5 a4c2 d3d4 c2b1 a2a3 b1e4 f2f3 e4c6 d4d5 c6d5');
INSERT INTO evaluations VALUES(10431,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(10431,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(10431,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10431,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(10431,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(10431,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(10431,36,'cp 15','b2c3 f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 e8g8 f1e1 f8e8 d2d3 f6d7 c1f4 d7e5 e1e5 e8e5 f4e5 f7f6 e5f4 c8e6 d1f3 d8d7 a2a4 a7a5 c3c4 a8c8 f4e3 b7b5 f3h5 b5c4 d3c4 e6f5 h2h3 c8e8 a1e1 d7e6 h5e2 c6c5');
INSERT INTO evaluations VALUES(10426,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(10426,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(10426,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10426,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(10426,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(10426,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(10426,36,'cp 6','e8g8 f1e1 f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 f8e8 d2d3 f6d7 c1f4 d7e5 e1e5 e8e5 f4e5 c8d7 c3c4 c6c5 d1f3 d7c6 f3g3 f7f6 e5c7 d8d4 a1e1 a8e8 e1e8 c6e8 h2h3 h7h6 c7f4 d4a1 g1h2 g8h7 f4e3 a1a2 e3c5 e8g6 g3f4 a2c2 c5a7 c2d3');
INSERT INTO evaluations VALUES(41367,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(41367,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41367,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(41367,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(41367,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(41367,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41367,36,'cp 31','b8c6 d4d5 c6b4 d3c4 e8g8 e1g1 c8g4 c4e2 e7e6 a2a3 b4a6 d5e6 g4e6 e2a6 b7a6 f1e1 d8b8 b2b3 b8b7 c1g5 h7h6 g5f6 g7f6 d1d2 f6g7 a1d1 a8d8 h2h3 e6c8 e1e3 f8e8 f3d4 c8d7 a3a4 a6a5 d4f3 g7c3 d2c3 e8e4 e3e4');
INSERT INTO evaluations VALUES(45255,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(45255,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(45255,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(45255,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(45255,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(45255,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(45255,36,'cp 116','b1c3 e7e5 g1f3 e5d4 f3d4 b8c6 g2g3 c8g4 f1e2 g4e2 d4e2 f8e7 c1e3 e8g8 e1g1 a8c8 e2f4 d8d7 f2f3 a7a6 g1g2 f8e8 f1f2 c6e5 a1c1 c8c6 f2c2 e8c8 h2h4 e7d8 h4h5 h7h6 e3d4 b7b5 b2b3 d7b7 c3d5 c6c2 c1c2 c8c2 d1c2 f6d5');
INSERT INTO evaluations VALUES(45252,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(45252,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(45252,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(45252,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(45252,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(45252,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(45252,36,'cp 113','e7e5 g1f3 e5d4 f3d4 b8c6 g2g3 c8g4 f1e2 g4e2 d4e2 f8e7 e1g1 e8g8 c1e3 f6g4 e3f4 f8e8 d1b3 d8d7 a1d1 a8c8 b3a4 g4e5 f2f3 e7d8 g1h1 a7a6 g3g4 d8a5 c3d5 b7b5 a4b3 a5d8 e2g3 c6a5 b3b4 a5c6');
INSERT INTO evaluations VALUES(54133,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(54133,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(54133,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(54133,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(54133,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(54133,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(54133,36,'cp 88','e6d5 f1b5 b8c6 e1g1 g8f6 f1e1 c8e6 c1g5 f8e7 d4c5 e7c5 g5f6 g7f6 d1d2 e8g8 b5d3 g8g7 b1c3 f8g8 a1d1 a8c8 c3e2 c6e5 f3d4 g7h8 d2h6 f6f5 h2h3 d8g5 h6g5 g8g5 g1h1 e6d7 c2c3 e5d3 d1d3 c8e8 d3d1 e8e4 g2g3 c5d6 d4f3 g5g6 d1d5 d7c6 d5f5');
INSERT INTO evaluations VALUES(56382,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(56382,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(56382,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(56382,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(56382,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(56382,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(56382,36,'cp -50','d5d4 c3b1 f8d6 f1c4 c7c5 d2d3 h7h6 c4b5 b8d7 b1d2 d6c7 a2a4 a7a6 b5c4 g8e7 d2f1 e7c6 c4d5 c6b4 d5b3 d7f8 f1g3 b4c6 e1g1 f8g6 b3d5 c6e7 d5c4 e8g8 c1d2 g6f4 c2c3 e7g6 g3f5 g8h7');
INSERT INTO evaluations VALUES(56574,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(56574,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(56574,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(56574,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(56574,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(56574,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(56574,36,'cp -48','c3b1 f8d6 f1c4 c7c5 a2a4 b8c6 d2d3 h7h6 b1a3 a7a6 c1d2 d8f6 f3g1 g8e7 d1f3 f6f3 g1f3 f7f6 c2c3 g7g5 c4b3 c8d7 a3c4 d6c7 c3d4 e5d4 a4a5 e7g6 h2h3 e8e7 e1e2 h6h5 c4b6 c7b6 a5b6 h8b8 h1c1 e7d6');
INSERT INTO evaluations VALUES(56592,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(56592,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(56592,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(56592,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(56592,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(56592,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(56592,36,'cp -54','f8d6 f1c4 c7c5 a2a4 h7h6 d2d3 b8c6 b1a3 a7a6 c1d2 d8f6 f3g1 f6g6 d1f3 f7f5 g1e2 g8f6 e4f5 c8f5 f3g3 g6h7 f2f3 e8c8 g3f2 f5d7 e2g3 h7g6 b2b3 d6c7 e1c1 h8f8 f2e2 c8b8 c1b2 g6e8 g3e4 e8e7 h1f1 c7a5 e4f6 f8f6 d2a5 c6a5 c4d5 d7c6 d5c6 f6c6');
INSERT INTO evaluations VALUES(52211,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(52211,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(52211,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(52211,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(52211,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(52211,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(52211,36,'cp 237','f1c4 d7d5 e4d5 b7b5 c4b3 b5b4 c3e2 c6d5 d2d4 e5e4 e2f4 g8e7 f3d2 f6f5 e1g1 g7g6 c2c4 b4c3 b2c3 f8h6 f4d5 e7d5 d2e4 h6c1 b3d5 f5e4 d5a8 c1f4 d1b3 e8f8 g2g3 f4d6 a8e4 f8g7 a1e1 h8f8 c3c4 d8f6 c4c5 d6c7 e4g2 c8a6 e1e6 f6d4 e6e7 g7h6 e7e4');
INSERT INTO evaluations VALUES(52200,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(52200,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(52200,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(52200,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(52200,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(52200,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(52200,36,'cp 226','d7d6 d2d4 g7g6 e1g1 a7a5 c1e3 b7b5 c4e2 f8h6 d1d2 h6e3 d2e3 d8b6 f1d1 e8f7 a2a4 b5b4 c3a2 f7g7 c2c3 c8a6 e2a6 a8a6 f3d2 e5d4 c3d4 g8h6 a2c1 h6f7 d2c4 b6c7 c1b3 a6a7 d4d5 c6d5');
INSERT INTO evaluations VALUES(54651,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(54651,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(54651,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(54651,33,'cp 242',NULL);
INSERT INTO evaluations VALUES(54651,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(54651,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(54651,36,'cp 249','a2a4 g7g6 a4a5 f8g7 d2d4 g8h6 h2h4 d8e7 h4h5 g6g5 d4d5 c6c5 c4b5 b8d7 f3d2 f6f5 e4f5 e5e4 f5f6 g7f6 c3e4 e8g8 e1g1 f6g7 e4g3 h6f5 g3f5 f8f5 b5d3 f5f8 h5h6 g7f6 d2c4 d7e5 c4e5 f6e5 d1h5 c8f5 c1g5 e7f7 h5e2 e5b2 a1b1');
INSERT INTO evaluations VALUES(54513,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(54513,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(54513,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(54513,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54513,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(54513,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(54513,36,'cp 42','c5c4 g2g3 b8c6 c1g5 f8e7 g5e7 g8e7 f1g2 d8b6 c3a4 b6b5 c2c3 b7b6 e1g1 e8g8 f1e1 h7h6 b2b3 c8e6 a1b1 b5a6 b3b4 a6c8 b1b2 e7g6 h2h4 e6g4 b2e2 c8d7 a4b2 f8e8 e2e8 a8e8 e1e8 d7e8 d1d2 c6e7 b2d1 g6f8 f3h2 g4d1 d2d1 f8h7 h2f1 h7f6 f1e3');
INSERT INTO evaluations VALUES(54577,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(54577,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54577,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54577,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(54577,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(54577,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(54577,36,'cp 28','g2g3 b8c6 c1g5 f8e7 g5e7 g8e7 f1g2 d8b6 a1b1 h7h6 e1g1 e8g8 f1e1 b6a5 d1d2 c8e6 h2h4 a8d8 a2a3 f8e8 e1e2 a5b6 c3a4 b6b5 a4c5 e6f5 a3a4 b5b6 b2b4 a7a5 b4b5 c6b4 b1c1 b6c7 f3h2 c7b6 c2c3');
INSERT INTO evaluations VALUES(39216,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(39216,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(39216,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(39216,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(39216,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39216,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(39216,36,'cp 42','d7d6 d2d4 e5d4 f3d4 c8d7 e2e3 g8f6 f1e2 b8c6 a2a3 a7a5 d4b5 d7f5 b2b3 h7h6 e1g1 e8g8 c1b2 c6e5 c3d5 f6d5 c4d5 c7c6 d5c6 b7c6 b5d4 f5d7 d1c2 f8e8 a1d1 c5b6 d4f3 e5f3 e2f3 d6d5 c2c3 f7f6 d1c1 a8b8');
INSERT INTO evaluations VALUES(38713,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(38713,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(38713,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(38713,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38713,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(38713,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(38713,36,'cp 60','d2d4 e5d4 f3d4 c8d7 e2e3 b8c6 f1e2 g8f6 a2a3 a7a5 b2b3 e8g8 d4b5 f8e8 c1b2 d7f5 e1g1 c6e5 b5d4 f5d7 d1c2 c5d4 e3d4 e5g6 f1e1 c7c6 e2f1 d6d5 e1e8 d8e8 c3a4 e8e7 a4c5 b7b6 c5a4 a8b8 c4c5 b6c5 d4c5 h7h6 a1d1 g6f4');
INSERT INTO evaluations VALUES(38720,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(38720,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(38720,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(38720,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(38720,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(38720,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(38720,36,'cp 44','e5d4 f3d4 c8d7 e2e3 g8f6 a2a3 a7a5 b2b3 b8c6 f1e2 e8g8 d4b5 f8e8 e1g1 c6a7 b5a7 c5a7 c1b2 d7f5 e2f3 c7c6 c3e2 d8e7 d1d2 f5g6 e2g3 a5a4 b3b4 f6e4 d2e2 e4f2 f1f2 a7e3 g1h1 e3f2 e2f2 f7f6');
INSERT INTO evaluations VALUES(38744,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(38744,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(38744,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38744,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38744,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(38744,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(38744,36,'cp 38','f3d4 c8d7 e2e3 g8f6 f1e2 b8c6 a2a3 a7a5 b2b3 h7h6 d4c6 b7c6 e1g1 e8g8 c1b2 d8e7 e2f3 f8e8 h2h3 a8b8 f1e1 f6h7 c3e4 h7g5 e4g5 e7g5 b2c3 d7h3 c3a5 e8e6 b3b4 c5b6 a5b6 b8b6 d1d4 c6c5 d4d5 h3f5 e3e4 f5g6 d5g5 h6g5');
INSERT INTO evaluations VALUES(38741,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(38741,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(38741,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(38741,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(38741,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38741,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(38741,36,'cp 44','c8d7 e2e3 b8c6 a2a3 c6d4 e3d4 c5b6 c1e3 c7c6 f1d3 g8f6 e1g1 e8g8 d1d2 f8e8 a1e1 d7e6 d4d5 c6d5 c4d5 b6e3 e1e3 e6d7 e3e8 f6e8 f1e1 e8f6 d2e3 d8b6 e3b6 a7b6 h2h3 h7h6 e1e7 g8f8');
INSERT INTO evaluations VALUES(45101,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(45101,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(45101,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(45101,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(45101,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(45101,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(45101,36,'cp 109','f2f4 d8a5 f1d3 e7e5 g1e2 e5d4 e2d4 a5b6 d4b3 a7a5 a2a4 c8g4 d3e2 g4e6 b3d4 b8a6 e1g1 a6c5 e2f3 e8c8 c1e3 e6c4 f1e1 b6b2 d1d2 b2b6 e3f2 b6c7 a1b1 d6d5 e4d5 f6d5 c3d5 c4d5 f3d5 d8d5 e1e8 d5d8 e8d8 c7d8 d2c3 c5e4 c3h3 d8d7 h3d7 c8d7 b1b7 d7c8');
INSERT INTO evaluations VALUES(45106,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(45106,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(45106,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(45106,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(45106,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(45106,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(45106,36,'cp 95','d8a5 f1d3 e7e5 g1f3 e5d4 f3d4 a5b6 d4b3 a7a5 a2a4 f8e7 d1f3 b6c7 e1g1 e8g8 h2h3 b8a6 c1e3 a6b4 b3d4 f8e8 a1e1 e7f8 e3f2 d6d5 e4e5 f6d7 g1h1 d7c5 f3g3 b4d3 c2d3 g8h8 f4f5 f7f6 e5f6 c7g3 f2g3');
INSERT INTO evaluations VALUES(33600,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(33600,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(33600,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(33600,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(33600,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(33600,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(33600,36,'cp 86','f1d3 e7e5 g1e2 e5d4 e2d4 a5b6 d4b3 a7a5 a2a4 b8a6 f4f5 a6b4 d1f3 d6d5 e4d5 c6d5 c1g5 b4d3 f3d3 f8b4 d3e2 e8f8 e1c1 b4c3 b2c3 c8f5 g5f6 g7f6 d1d5 f5g6 e2d2 f8g7 d5d6 b6c7 d2d4 h8e8 d4f6 g7g8 h1d1 c7c4 f6d4 e8e4 d4c4 e4c4 d6d8 a8d8 d1d8 g8g7 b3a5');
INSERT INTO evaluations VALUES(46711,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(46711,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(46711,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(46711,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(46711,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(46711,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(46711,36,'cp 58','g1f3 f8g7 e1g1 e8g8 f1e1 b8c6 h2h3 e7e5 c2c3 a7a5 d3c2 h7h6 a2a4 f8e8 d4d5 c6e7 c1d2 f6d7 c3c4 d7c5 b2b4 a5b4 d2b4 c8d7 a4a5 c5a6 b4d2 f7f5 b1c3 e8f8 c3a4 f5e4 c2e4 e7f5 a1b1 d7a4 d1a4 a6c5 a4c2 g8h7');
INSERT INTO evaluations VALUES(46708,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(46708,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(46708,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(46708,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(46708,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46708,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(46708,36,'cp 61','f8g7 e1g1 b8c6 f1e1 e7e5 c2c3 e8g8 h2h3 a7a5 a2a4 f8e8 d4d5 c6e7 d3c2 h7h6 b1a3 e8f8 b2b4 a5b4 c3b4 f6h5 b4b5 b7b6 a3c4 f7f5 a1a3 f5e4 c2e4 h5f6 e4c2 c8f5 c2b3 g8h7 a4a5 e5e4 f3d4 f6d5 a5b6 d5b6 c4b6 c7b6');
INSERT INTO evaluations VALUES(41380,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(41380,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(41380,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(41380,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(41380,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(41380,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(41380,36,'cp 74','e1g1 b8c6 f1e1 e8g8 h2h3 e7e5 c2c3 a7a5 a2a4 f8e8 d4d5 c6e7 d3c2 h7h6 b1a3 e8f8 a3b5 f6e8 c3c4 f7f5 e4f5 g6f5 a1a3 e7g6 f3h2 g6e7 f2f4 e5e4 h2f1 c7c6 b5c3 e8f6 f1e3 c6d5 c4d5 g8h8');
INSERT INTO evaluations VALUES(41379,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(41379,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(41379,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(41379,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(41379,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(41379,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(41379,36,'cp 62','b8c6 f1e1 e7e5 c2c3 e8g8 h2h3 a7a5 a2a4 f8e8 d3c2 h7h6 d4d5 c6e7 b1a3 e8f8 a3b5 f6e8 c3c4 f7f5 e4f5 c8f5 c2b3 e8f6 c4c5 g8h7 c5d6 c7d6 f3h4 a8c8 c1e3 f5e4 b5c3 e4f5 a1c1 f5d7 g2g4 d8e8 b3c2 e8f7 d1d2 c8c3 b2c3 f6d5');
INSERT INTO evaluations VALUES(14815,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(14815,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(14815,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(14815,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14815,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(14815,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(14815,36,'cp 48','f8d6 f4g3 b7b6 c2c4 e8g8 a1c1 c8b7 f1e2 c6e7 c4d5 e6d5 e1g1 f6e4 g3d6 d8d6 d1a4 a7a5 f1d1 f8c8 a4c2 e4d2 f3d2 c7c5 d4c5 c8c5 c2a4 h7h6 d2b1 c5c1 d1c1 b7c6 a4d4 d6b4 d4d2 b4d2 b1d2');
INSERT INTO evaluations VALUES(9243,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(9243,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(9243,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(9243,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(9243,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(9243,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(9243,36,'cp 43','f4g3 e8g8 c2c4 b7b6 a1c1 c8b7 f1e2 c6e7 c4d5 e6d5 e1g1 f6e4 g3d6 d8d6 d1c2 a8c8 c2a4 a7a5 f1d1 h7h6 a2a3 b7c6 a4c2 c6b7 d2e4 d5e4 f3d2 f7f5 d2c4 d6e6 b2b4 a5b4 a3b4 b7d5 b4b5 e7g6 e2h5 f5f4 h5g6 e6g6 e3f4 d5c4 c2c4 g6f7 d4d5 f7f4');
INSERT INTO evaluations VALUES(9260,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(9260,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(9260,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(9260,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(9260,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(9260,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(9260,36,'cp 41','b7b6 c2c4 e8g8 a1c1 c8b7 f1d3 c6b4 d3b1 c7c5 d4c5 b6c5 a2a3 b4a6 e1g1 d6g3 h2g3 h7h6 f1e1 d8b6 b2b3 a8c8 c4d5 e6d5 e3e4 d5e4 d2e4 b7e4 b1e4 f6e4 e1e4 a6b8 b3b4 f8d8 d1e2 b8d7 e2c2 c5b4 c2c8 d8c8 c1c8 g8h7 e4b4');
INSERT INTO evaluations VALUES(41205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41205,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41205,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(41205,36,'cp 21','h7h6 g5h4 g6g5 h4g3 f6h5 e2e3 h5g3 h2g3 c7c5 f1e2 g5g4 f3h4 h6h5 a1c1 b8c6 d4d5 c6e5 d1c2 d8a5 a2a3 c8d7 e1g1 a7a6 c2b1 b7b5 c4b5 a6b5 b2b4 c5b4 a3b4 a5b6 h4f5 e8f8 b1e4 e5g6 e2d3 d7f5 e4f5 g7c3 c1c3 g6e5 f1c1');
INSERT INTO evaluations VALUES(41168,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41168,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41168,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41168,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(41168,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41168,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41168,36,'cp 38','g5h4 g6g5 h4g3 f6h5 e2e3 c7c5 f1e2 h5g3 h2g3 g5g4 f3h4 h6h5 a1c1 b8c6 d4d5 c6e5 e1g1 c8d7 c1b1 a8b8 d1c2 a7a6 b2b4 c5b4 b1b4 b7b5 c4b5 b8c8 b4b3 a6b5 h4f5 e8f8 c2b1 e5c4 f5g7 f8g7 c3b5 c4e5 b1b2 h5h4 b5d4 h4g3 f2g3 c8b8 f1c1 b8b3 a2b3');
INSERT INTO evaluations VALUES(41172,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(41172,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41172,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(41172,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41172,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41172,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41172,36,'cp 9','g6g5 h4g3 f6h5 e2e3 h5g3 h2g3 c7c5 f1e2 g5g4 f3h4 h6h5 a1c1 b8c6 d4d5 c6e5 d1c2 c8d7 a2a3 a7a6 h4f5 d7f5 c2f5 d8b6 f5c2 b6a5 f2f4 g4f3 g2f3 e8c8 e1f2 c8b8 c1b1 d8c8 f2g2 g7f6 c3e4 a5c7 e4f6 e7f6 b1d1 h5h4 h1h4 c8g8 h4h8 g8h8');
INSERT INTO evaluations VALUES(31192,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(31192,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(31192,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(31192,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(31192,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(31192,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(31192,36,'cp -28','d8d4 g1e2 d4d8 e2g3 f5e4 g3e4 b8d7 d1e2 e7e6 e1g1 g8f6 c1f4 d8a5 f1d1 f6e4 d3e4 f8e7 a2a3 d7f6 b2b4 a5h5 e4f3 h5f5 g2g3 e8g8 c2c4 a7a5 b4b5 a8d8 b5c6 b7c6 a3a4 e7c5 g1g2 c5d4 f4c7 d8d7 f3c6 d4a1');
INSERT INTO evaluations VALUES(32778,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(32778,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(32778,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(32778,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(32778,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(32778,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(32778,36,'cp -17','g1f3 d4d8 d1e2 f5e4 d3e4 d8c7 e1g1 e7e6 g2g3 f8d6 f3d2 g8f6 d2c4 e8g8 c4d6 c7d6 f1d1 d6c5 b2b3 b8d7 c1b2 f8e8 a2a4 a8d8 e4g2 d7f8 d1d2 d8d2 e2d2 h7h6 b2f6 g7f6 a4a5 g8g7 a5a6 b7b6 a1d1 e6e5');
INSERT INTO evaluations VALUES(32777,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(32777,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(32777,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(32777,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(32777,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(32777,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(32777,36,'cp -31','d4d8 d1e2 e7e6 e1g1 f5e4 d3e4 d8c7 g2g3 f8d6 f3d2 g8f6 d2c4 e8g8 c4d6 c7d6 f1d1 d6e7 e4g2 f8d8 d1d8 e7d8 a2a4 b8d7 b2b4 h7h6 a4a5 a7a6 c2c4 e6e5 b4b5 a6b5 c4b5 a8a5 a1a5 d8a5 b5c6 b7c6 g2c6 d7c5 e2c4 a5b6');
INSERT INTO evaluations VALUES(31197,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(31197,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(31197,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(31197,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(31197,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(31197,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(31197,36,'cp -19','d1e2 f5e4 d3e4 d8c7 g2g3 e7e6 e1g1 f8d6 f3d2 g8f6 d2c4 e8g8 c4d6 c7d6 f1d1 d6e7 e4g2 b8d7 a2a4 f8d8 a4a5 a7a6 b2b3 d7f8 c1d2 d8d7 c2c4 a8d8 b3b4 e6e5 d2c3 h7h6 d1e1 e5e4 g2e4 e7e4 e2e4 f6e4 e1e4 f8e6 e4e1 e6d4');
INSERT INTO evaluations VALUES(53383,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53383,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53383,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53383,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53383,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(53383,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(53383,36,'cp 9','g8f6 b1c3 b8c6 c1f4 c8f5 e2e3 e7e6 d1b3 f8b4 f1b5 e8g8 b5c6 b4c3 b3c3 a8c8 f3e5 f6g4 e5g4 f5g4 c3b4 c8c6 b4b7 d8c8 b7c8 f8c8 e1g1 f7f6 f2f3 g4f5 h2h4 a7a5 b2b3 h7h5 f1f2 g8f7 a1e1 c6c2 f2c2 c8c2 e1a1 f7e8 a2a3 c2b2 b3b4 a5b4');
INSERT INTO evaluations VALUES(45456,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(45456,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45456,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(45456,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45456,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(45456,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(45456,36,'cp 22','b1c3 b8c6 c1f4 c8f5 e2e3 e7e6 d1b3 f8b4 f1b5 e8g8 b5c6 b4c3 b3c3 a8c8 f3e5 f6g4 e5g4 f5g4 c3b4 c8c6 b4b7 d8c8 b7c8 f8c8 e1g1 f7f6 h2h4 a7a5 f2f3 g4f5 h4h5 g7g5 h5g6 h7g6 e3e4 g6g5 e4f5 g5f4 f5e6 c6e6 a1d1 c8c2 f1f2');
INSERT INTO evaluations VALUES(57657,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(57657,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(57657,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(57657,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(57657,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(57657,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(57657,36,'cp 39','d7d5 e2e4 d5e4 c3e4 f8g7 g1f3 g8f6 e4f6 e7f6 f1e2 e8g8 e1g1 f8e8 f1e1 a7a5 c1f4 c8f5 a2a3 b8d7 d1d2 f5e4 a1d1 d7b6 c2c4 e4f3 g2f3 f6f5 g1g2 a5a4 f4g5 g7f6 g5e3 h7h5 d4d5 b6d7 d5d6 c7c6 c4c5 b7b6 c5b6 d7b6 e3b6 d8b6 e2c4 e8e1 d1e1');
INSERT INTO evaluations VALUES(54857,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(54857,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(54857,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(54857,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(54857,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(54857,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54857,36,'cp 48','e2e4 d5e4 c3e4 f8g7 g1f3 g8f6 e4f6 e7f6 f1e2 e8g8 e1g1 f8e8 f1e1 c8f5 c1e3 b8c6 c2c3 c6e7 d1b3 e7d5 e3d2 d5b6 c3c4 f5e4 a1d1 a7a5 d2f4 c7c6 e2f1 a5a4 b3b4 g6g5 f4e3 e4c2 d1d2 c2e4 e1d1 e4f3 g2f3');
INSERT INTO evaluations VALUES(54868,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(54868,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(54868,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(54868,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(54868,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(54868,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(54868,36,'cp 40','d5e4 c3e4 f8g7 g1f3 g8f6 e4f6 e7f6 f1e2 e8g8 e1g1 a7a5 f1e1 f8e8 c1f4 c7c6 d1d2 b8a6 c2c4 g7f8 a2a3 a6c7 h2h3 c7e6 f4e3 e6g7 e2f1 c8d7 a1d1 a5a4 e3f4 e8e1 f3e1 g6g5 f4e3 g7f5 e1c2 d7e6 d4d5 c6d5');
INSERT INTO evaluations VALUES(54895,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(54895,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(54895,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(54895,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(54895,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(54895,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(54895,36,'cp 53','c3e4 g8f6 e4f6 e7f6 f1e2 f8g7 g1f3 e8g8 e1g1 f8e8 f1e1 a7a5 c1f4 b8a6 d1d2 c7c6 c2c4 a6c7 a1d1 c7e6 f4e3 g7f8 a2a3 c8d7 d2c3 a5a4 e2f1 d8c7 h2h3 e6g7 e3h6 e8e1 f3e1 c7a5 c3a5 a8a5 e1d3 d7f5 h6d2 a5a8');
INSERT INTO evaluations VALUES(54893,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(54893,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(54893,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(54893,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(54893,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(54893,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(54893,36,'cp 41','g8f6 e4f6 e7f6 g1f3 f8g7 f1e2 e8g8 e1g1 f8e8 f1e1 a7a5 c1f4 b8a6 c2c4 c7c6 d1d2 g7f8 a2a3 a6c7 h2h3 c8f5 a1d1 f5e4 f3h2 a5a4 h2g4 c7e6 f4e3 f6f5 g4h6 f8h6 e3h6 d8h4 d4d5 c6d5 c4d5 e4d5 d2d5 h4h6 d5b7 h6f4 e2f3');
INSERT INTO evaluations VALUES(43012,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(43012,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(43012,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(43012,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(43012,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(43012,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(43012,36,'cp -63','b8c6 e5f4 d7d5 e4d5 c6b4 c4b3 b4d5 f4d4 e8g8 b1c3 c7c5 d4d3 c8e6 a2a3 b7b5 b3d5 f6d5 g1f3 c5c4 d3e4 d5f6 e4d4 e6f5 d2d3 c4d3 c2d3 f5d3 c1e3 f8e8 e1c1 d3g6 d4d8 e7d8 h1e1 f6e4 d1d5 a7a6 d5e5 d8c7 e5e8 a8e8');
INSERT INTO evaluations VALUES(10867,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(10867,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(10867,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(10867,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(10867,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(10867,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(10867,36,'cp -71','e5f4 d7d5 e4d5 c6b4 c4b3 b4d5 f4d4 c7c5 d4d3 e8g8 b1c3 c8e6 g1f3 d5b4 d3d8 a8d8 e1g1 c5c4 b3a4 a7a6 a2a3 b4d5 c3d5 e6d5 c2c3 e7d6 a4c2 f8e8 d2d4 c4d3 c2d3 d5f3 g2f3 d6h2 g1h2 d8d3 c1e3 e8d8 c3c4 f6d7 f1c1 d7e5 h2g2 h7h6 f3f4 e5g4 c1e1 g4e3 e1e3');
INSERT INTO evaluations VALUES(10936,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(10936,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(10936,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(10936,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(10936,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(10936,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(10936,36,'cp -83','d7d5 e4d5 c6b4 c4b3 b4d5 f4d4 c7c5 d4d3 e8g8 b1c3 c8e6 g1f3 d5b4 d3d8 a8d8 e1g1 c5c4 b3a4 a7a6 a2a3 b4d5 c3d5 e6d5 f1e1 e7d6 c2c3 f6g4 g1f1 d6c5 e1e2 b7b5 a4c2 f8e8 d2d4 c4d3 c2d3 g4f2 e2e8 d8e8 d3e2 d5c4 e2c4 b5c4 a1b1 f2d3 c1d2 e8b8');
INSERT INTO evaluations VALUES(43514,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43514,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43514,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43514,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43514,36,'cp 0','d8b6 d4b3 b8d7 d1e2 b6c7 g2g4 f8e7 f4f5 b7b5 a2a3 h7h6 g5d2 d7e5 e1c1 a8b8 h1g1 b5b4 a3b4 b8b4 g1g3 c7b7 e2g2 f6h7 c3a2 b4a4 a2c3');
INSERT INTO evaluations VALUES(33506,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33506,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33506,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33506,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33506,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33506,36,'cp 1','d4b3 b8d7 d1d2 b6c7 e1c1 b7b5 a2a3 c8b7 f1d3 f8e7 g5f6 e7f6 d3b5 f6c3 b5d7 e8d7 b2c3 h8c8 h1e1 a6a5 a3a4 c7c3 d2d6 d7e8 d6d7 e8f8 d7d3 c3b4 d3d6 f8e8 d6b4 a5b4 d1d6 a8a4 c1b2 c8c4 d6d4 b7a6 e1e3 c4d4 b3d4 a6c4 d4b3 a4a3 g2g3 c4b3 c2b3');
INSERT INTO evaluations VALUES(12368,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12368,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12368,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12368,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(12368,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12368,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12368,36,'cp 0','e1g1 d7d5 c4b5 c8d7 f1e1 a7a6 b5c6 d7c6 e4e5 f6d7 d3d4 c5d4 f3d4 d8h4 c2c3 f8c5 e1e3 h4g5 e3e1');
INSERT INTO evaluations VALUES(12367,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12367,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12367,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12367,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12367,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12367,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12367,36,'cp 0','d7d5 c4b5 c8d7 f1e1 a7a6 b5c6 d7c6 e4e5 f6d7 d3d4 f8e7 c2c3 a6a5 b1d2 a5a4 d2f1 d8a5 c1d2 a5b6 d2c1');
INSERT INTO evaluations VALUES(13531,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13531,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13531,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13531,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13531,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13531,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13531,36,'cp 0','c4b5 c8d7 f1e1 a7a6 b5c6 d7c6 e4e5 f6d7 d3d4 c5d4 f3d4 d8h4 c2c3 f8c5 e1e3 h4g5 e3e1');
INSERT INTO evaluations VALUES(53331,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(53331,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(53331,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(53331,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(53331,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(53331,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(53331,36,'cp -42','c5d4 d1d4 b8c6 f1b5 a7a6 b5c6 b7c6 g1f3 e7e6 e1g1 c6c5 d4d3 g8e7 c2c4 c8b7 b1c3 d5d4 c3e4 e7g6 b2b4 c5b4 a2a3 a8c8 f1d1 f8e7 h2h4 d8c7 e4d6 e7d6 e5d6 c7d6 a3b4 e8g8 c4c5 d6d5 d3d4 d5h5 g2g4 g6h4 g4h5 h4f3 g1f1 f3d4 d1d4 h7h6 c1e3');
INSERT INTO evaluations VALUES(53448,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(53448,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(53448,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(53448,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(53448,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(53448,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(53448,36,'cp -46','d1d4 b8c6 f1b5 e7e6 g1f3 g8e7 e1g1 c8d7 b5c6 b7c6 c2c4 c6c5 d4g4 h7h5 g4h3 e7f5 c1g5 f8e7 g5e7 d8e7 b1c3 d5c4 f3d2 a8d8 d2c4 d7c6 c4e3 f5d4 e3c4 e7g5 f2f4 g5g4 h3g4 h5g4 a1c1 h8h5 f1e1 h5f5 c3e2 d4e2 e1e2 f5f4 c4d6 e8e7');
INSERT INTO evaluations VALUES(53443,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(53443,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(53443,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(53443,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(53443,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(53443,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(53443,36,'cp -50','b8c6 f1b5 e7e6 g1f3 c8d7 b5c6 b7c6 c2c4 c6c5 d4d3 g8e7 b1c3 d5c4 d3c4 e7f5 e1g1 a8b8 f1d1 d8c7 c3e2 f8e7 e2g3 e8g8 g3f5 e6f5 c1g5 d7c6 g5e7 c7e7 c4e2 c6f3 e2f3 e7e5 b2b3 g7g6 f3g3 f5f4 g3f3 f8d8 d1e1');
INSERT INTO evaluations VALUES(19888,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(19888,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(19888,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(19888,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(19888,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(19888,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(19888,36,'cp -30','f1b5 e7e6 g1f3 g8e7 e1g1 c8d7 b5c6 b7c6 c2c4 c6c5 d4d3 a8b8 b1c3 d5c4 d3c4 e7f5 f1d1 d8c7 a2a4 b8b4 c4a6 c7b7 a6b7 b4b7 f3d2 a7a6 a1b1 d7c6 b2b3 f8e7 c1a3 f5d4 d2e4 d4e2 c3e2 c6e4 b1c1 b7b3 a3c5 e7c5 c1c5 e8g8');
INSERT INTO evaluations VALUES(19729,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(19729,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(19729,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(19729,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(19729,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(19729,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(19729,36,'cp -39','e7e6 g1f3 g8e7 e1g1 c8d7 b5c6 e7c6 d4f4 c6e7 c1e3 e7g6 f4g3 f8e7 c2c3 e8g8 b1d2 f7f6 e3d4 f6f5 d2b3 a8c8 b3c1 a7a5 f1d1 a5a4 b2b4 d7b5 c1d3 b5d3 d1d3 g8h8 a2a3 f8f7 a1d1 b7b5 f3d2 f5f4 g3f3 d8d7');
INSERT INTO evaluations VALUES(49395,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(49395,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49395,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(49395,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(49395,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(49395,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(49395,36,'cp 9','d7d5 f1g2 f8b4 c1d2 b4e7 d1c2 e8g8 g1f3 c7c6 e1g1 b7b6 f3e5 c8b7 b1c3 f6d7 e5g4 b8a6 a2a3 a6c7 c4d5 c6d5 c2a4 a7a5 f1c1 b6b5 a4d1 e7d6 e2e4 b5b4 a3b4 a5b4 a1a8 d8a8 c3a4 d5e4 a4c5 b7c6 d2b4');
INSERT INTO evaluations VALUES(48270,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48270,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(48270,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(48270,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(48270,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(48270,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48270,36,'cp 17','f1g2 f8b4 c1d2 b4e7 g1f3 c7c6 d1c2 b7b6 e1g1 e8g8 f3e5 c8b7 c4d5 c6d5 f1c1 f6e4 a2a3 e4d6 c2a4 d8e8 a4e8 f8e8 d2b4 b8a6 b4d6 e7d6 e5d3 a8c8 b1c3 a6c7 e2e3 e8d8 g2f1 b7a6 d3e5 d6e5 d4e5 a6f1');
INSERT INTO evaluations VALUES(48271,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(48271,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(48271,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(48271,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(48271,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48271,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48271,36,'cp 0','f8b4 c1d2 b4e7 g1f3 e8g8 d1c2 c7c6 e1g1 b7b6 f3e5 c8b7 c4d5 c6d5 f1c1 f6e4 c2c7 b7a6 c7c2');
INSERT INTO evaluations VALUES(42061,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(42061,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42061,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(42061,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(42061,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(42061,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(42061,36,'cp 20','c1d2 b4e7 g1f3 c7c6 d1c2 e8g8 e1g1 b7b6 f1d1 b8d7 d2f4 c8b7 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 d8c7 e2e3 a8d8 a1c1 c6c5 d4d5 e6d5 g2d5 d7f6 d5b7 c7b7 c2e2 a7a6 e3e4 d8d4 c4e3 f8d8 e3f5 d4d1 c1d1 d8d1 e2d1 g7g6 f5e7 b7e7 e4e5 f6h5 d1f3 h5g7 f3a8 e7e8 a8a6');
INSERT INTO evaluations VALUES(42060,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(42060,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(42060,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42060,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(42060,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(42060,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(42060,36,'cp 8','b4e7 d1c2 c7c6 g1f3 b8d7 d2f4 e8g8 e1g1 b7b6 f1d1 c8b7 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 a8c8 e2e3 d8c7 a2a3 a7a5 a1c1 b7a6 c4d2 d7f6 d2f3 c6c5 d4c5 e7c5 f3e5 f8d8 d1d8 c7d8 e5c6 d8c7 b2b4 c5d6 b4b5');
INSERT INTO evaluations VALUES(56727,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(56727,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(56727,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(56727,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(56727,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(56727,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(56727,36,'cp 139','f3g5 g8f6 f1c4 d8e7 c4f7 e8d8 f7b3 d8e8 g5f7 h8g8 b1c3 c8e6 b3e6 e7e6 f7g5 e6d7 f2f4 b8c6 e1g1 h7h6 g5f3 f8e7 f4f5 a8d8 f3h4 d6d5 e4d5 c6b4 d1e2 b4d5 c1d2 d7c6 h4g6 e7d6 g6e5 d6e5 e2e5 e8f7 e5g3 g8e8 a1e1 e8e1 f1e1 c6c5 g1h1 d5e7 d2e3 c5c4 g3f3 e7c6 e3g1 c6b4 f3e2 c4e2 e1e2 b4c6');
INSERT INTO evaluations VALUES(35147,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(35147,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(35147,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(35147,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(35147,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35147,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(35147,36,'cp 8','b8c6 d4b5 f8c5 d1f3 c5b6 f3g3 d7d6 b1c3 g8e7 c1g5 f6g6 f1d3 b6a5 a2a3 f7f6 g5e3 a7a6 b2b4 a6b5 b4a5 c8e6 d3b5 e8g8 g3g6 h7g6 a5a6 b7a6 b5e2 f8b8 e1c1 c6e5 d1d4 g6g5 d4a4 e6d7 a4a5 e7c6 a5a4');
INSERT INTO evaluations VALUES(42559,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(42559,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42559,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(42559,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42559,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42559,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42559,36,'cp 4','c7c6 g1f3 e8g8 e1g1 d7d6 c2c3 b8d7 c4b3 c5b6 f1e1 f8e8 c1e3 b6e3 e1e3 a7a5 a2a4 d8c7 b3c2 d7f8 d3d4 c8e6 b1d2 f8g6 f3g5 e6c8 d2f1 h7h6 g5f3 b7b5 d1d2 a8b8 a4b5 c6b5 f1g3 c8e6 a1e1');
INSERT INTO evaluations VALUES(39871,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39871,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39871,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39871,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39871,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39871,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39871,36,'cp 0','g1f3 d7d5 e4d5 c6d5 c4b3 d8a5 c1d2 a5c7 e1g1 b8c6 d2g5 c8e6 b1c3 e8c8 c3b5 c7b6 b5c3');
INSERT INTO evaluations VALUES(50430,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(50430,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(50430,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(50430,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(50430,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(50430,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(50430,36,'cp 27','g5f4 d7d5 e2e3 c7c5 b1d2 e4d2 d1d2 b8c6 g1f3 c5d4 e3d4 c8f5 f1d3 f5d3 d2d3 e7e6 e1g1 f8d6 f4d6 d8d6 c2c3 h7h6 f1e1 e8g8 e1e2 f8c8 g2g3 c6a5 g1g2 a5c4 f3e5 c8c7 e5c4 c7c4 e2e3 a7a5 a2a3 a5a4 a1e1 d6c6 d3e2');
INSERT INTO evaluations VALUES(50445,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(50445,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(50445,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(50445,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(50445,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50445,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(50445,36,'cp 25','d7d5 e2e3 c7c5 b1d2 e4d2 d1d2 b8c6 g1f3 c5d4 e3d4 c8f5 f1d3 f5d3 d2d3 e7e6 e1g1 f8d6 f4d6 d8d6 g2g3 h7h6 c2c3 e8g8 f1e1 a8c8 e1e2 c6a5 f3e1 d6d8 e1g2 c8c6 d3f3 d8f6 f3g4 a5c4 a1e1 a7a5 h2h4');
INSERT INTO evaluations VALUES(47502,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(47502,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(47502,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47502,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(47502,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47502,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(47502,36,'cp 17','e2e3 b8d7 f1d3 g7g5 d3e4 g5f4 e4d5 f4e3 f2e3 e7e6 d5f3 d8g5 d1d3 f8h6 g1e2 g5e3 d3e3 h6e3 b1d2 d7b6 a2a4 a7a5 b2b3 e3h6 h2h4 b6d5 d2c4 e8g8 e1f2 b7b6 c2c3 c8b7 g2g4 h6g7 a1e1 f7f5 g4g5');
INSERT INTO evaluations VALUES(47497,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(47497,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(47497,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(47497,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47497,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(47497,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(47497,36,'cp 21','b8c6 f1b5 e7e6 b1d2 f8d6 d2e4 d5e4 g1e2 e8g8 f4g3 c8d7 e1g1 c6e7 b5d7 d8d7 c2c4 f7f5 a1c1 h7h6 a2a3 g7g5 g3d6 c7d6 d4d5 e6d5 c4d5 e7g6 e2d4 g6e5 d1b3 e5d3 c1c3 a8c8 c3d3 e4d3 b3d3');
INSERT INTO evaluations VALUES(10247,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(10247,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10247,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10247,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(10247,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10247,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(10247,36,'cp 0','e1g1 e8g8 b1c3 d7d5 f3e5 c6e5 d3d4 c5d4 d1d4 e5c6 d4d3 d5e4 d3d8 f8d8 c1f4 c8f5 f4c7 d8d7 c7f4 c6d4 e2d1 a8d8 a1c1 f5g6 c3e2 d4c6 e2c3');
INSERT INTO evaluations VALUES(10246,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10246,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10246,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10246,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10246,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10246,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10246,36,'cp 0','e8g8 b1c3 d7d5 f3e5 c6e5 d3d4 c5d4 d1d4 e5c6 d4d3 d5e4 c3e4 f6e4 d3e4 f8e8 e4c4 c8e6 c4c3 c6d4 e2d3 d8d5 c1f4 e6f5 a1e1 f5d3 c3d3 c7c5 f4e3 e8d8 e3d4 d5d4 d3d4 d8d4 e1e7 b7b6 f1e1 g7g6 g1f1 d4d2 e1e2 d2d1 e2e1');
INSERT INTO evaluations VALUES(4390,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(4390,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4390,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(4390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4390,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4390,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4390,36,'cp 0','b1c3 d7d5 f3e5 c6e5 d3d4 c5d4 d1d4 e5c6 d4d3 d5e4 c3e4 f6e4 d3e4 f8e8 e4c4 c6d4 e2d3 c8e6 c4c3 d8f6 c3c7 a8c8 c7b7 d4c2 d3c2 c8c2 c1e3 c2b2 b7a7 e6d5 a7c5 f6e6 f1d1 d5g2 g1g2 e6e4 g2g1 e4g4 g1f1 g4h3 f1g1');
INSERT INTO evaluations VALUES(4382,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4382,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4382,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4382,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4382,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4382,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4382,36,'cp 0','d7d5 c1g5 d5e4 d3e4 c8e6 c3d5 e6d5 g5f6 d8f6 d1d5 c5b6 c2c3 a8d8 d5b3 c6b8 g2g3 g7g6 a2a4 a7a5 a1d1 b8d7 g1g2 g8g7 h2h4 h7h5 e2c4 d7c5 b3c2 c5d7');
INSERT INTO evaluations VALUES(27195,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(27195,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(27195,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(27195,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(27195,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(27195,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27195,36,'cp 77','c6b8 d1e2 c7c6 b5c4 g8f6 b1c3 f8e7 e1g1 c6d5 c3d5 f6d5 c4d5 d7c6 f1d1 b8d7 c2c4 e8g8 c1e3 d8c7 a1c1 a8c8 h2h3 d7f6 d5c6 b7c6 e3g5 f8d8 g5f6 e7f6 b2b4 g7g6 g2g3 c7b7 a2a3 a7a5 b4b5 a5a4 g1g2 f6g7');
INSERT INTO evaluations VALUES(32471,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(32471,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(32471,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(32471,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(32471,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(32471,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(32471,36,'cp 75','d1e2 c7c6 b5c4 g8f6 b1c3 f8e7 e1g1 e8g8 f1d1 d8c7 a2a4 c6d5 c4d5 d7c6 d5b3 b8d7 c1g5 h7h6 g5f6 d7f6 f3d2 g7g6 d2f1 g8g7 e2d3 a7a5 f1e3 h6h5 b3d5 a8d8 c3b5 c7d7 b2b3 f6d5 e3d5 f7f5 d5e7 d7e7 b5d6');
INSERT INTO evaluations VALUES(35744,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(35744,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(35744,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(35744,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(35744,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(35744,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35744,36,'cp 30','c4d5 e6d5 f1d3 c8e6 h2h3 b8d7 e1g1 f6e8 g5e7 d8e7 d1c2 g7g6 a1e1 e8g7 b2b4 a7a6 c2b2 e6f5 d3f5 g7f5 e3e4 d5e4 c3e4 d7b6 e4g3 e7f6 g3f5 f6f5 e1e7 b6c4 b2c3 c4d6 f1e1 a8e8 a2a4 e8e7 e1e7');
INSERT INTO evaluations VALUES(35741,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(35741,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(35741,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(35741,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(35741,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(35741,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(35741,36,'cp 28','f6d5 g5e7 d8e7 f1d3 d5c3 b2c3 b7b6 e1g1 c8b7 a2a4 c6c5 a4a5 h7h6 a5a6 b7f3 d1f3 b8d7 d3b5 d7f6 f1d1 a8c8 c3c4 f8d8 g2g3 c5d4 e3d4 d8d6 d1d2 c8c7 f3f4 e7d8 a1e1 f6h5 f4e3 h5f6 e1d1 h6h5 e3f4 g7g6 d2d3');
INSERT INTO evaluations VALUES(56056,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(56056,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(56056,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(56056,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(56056,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(56056,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(56056,36,'cp 47','d2e4 b8d7 g1f3 g8f6 e4f6 d7f6 c2c3 a7a6 c1f4 c7c5 f1e2 c5d4 d1d4 d8d4 f3d4 c8d7 e1c1 f8c5 e2f3 e8c8 d1d2 c5b6 h1e1 b6c7 f4e5 h7h5 h2h4 c7e5 e1e5 g7g6 g2g3 h8e8 c3c4 c8c7 b2b3 f6g4 e5e2 g4f6 d4c2 d7c6 f3c6 b7c6');
INSERT INTO evaluations VALUES(20418,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(20418,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(20418,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(20418,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(20418,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(20418,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20418,36,'cp 0','d4c6 b7c6 b1d2 g8e7 f1d3 c6c5 e1g1 e7c6 d2f3 c5c4 d3e2 f8e7 b2b3 c8a6 c2c3 e8g8 a1b1 d8c7 c1f4 a8b8 h2h4 h7h6 f1e1 c7b6 f4e3 b6c7');
INSERT INTO evaluations VALUES(20152,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20152,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20152,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20152,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20152,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20152,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20152,36,'cp 0','b7c6 b1d2 g8e7 f1d3 c6c5 e1g1 e7c6 d2f3 c5c4 d3e2 f8e7 b2b3 c8a6 c2c3 e8g8 f1e1 d8c7 c1f4 f8d8 a2a3 c7b6 f4e3 b6c7');
INSERT INTO evaluations VALUES(18493,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(18493,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18493,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18493,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(18493,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18493,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18493,36,'cp 0','b1d2 g8e7 f1d3 c6c5 e1g1 e7c6 d2f3 c5c4 d3e2 f8e7 b2b3 c8a6 c2c3 e8g8 c1f4 a8b8 a1b1 d8c7 f1e1 f8d8 h2h4 h7h6 h4h5 c7b7 f4e3 b7c7');
INSERT INTO evaluations VALUES(34724,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(34724,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(34724,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(34724,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(34724,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(34724,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(34724,36,'cp 138','b1c3 g8f6 f1c4 b8c6 c1g5 e5d4 f3d4 c6d4 d1d4 h7h6 g5f4 g7g5 f4e3 f8g7 h2h4 e8f8 d4d3 f6g4 c3d5 e7d8 e3d4 g4e5 d3e2 c7c6 d5e3 b7b5 c4b3 c6c5 d4e5 g7e5 e2b5 g5h4 c2c3 c8d7 b5b7 a7a5 b7d5 d7e6 d5d3 d8g5 b3e6 f7e6 e1c1');
INSERT INTO evaluations VALUES(34723,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(34723,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(34723,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(34723,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(34723,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(34723,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(34723,36,'cp 112','g8f6 f1c4 c8g4 c1e3 b8d7 h2h3 g4f3 d1f3 g7g6 e3g5 c7c6 d4d5 f8g7 a2a4 e8g8 a4a5 h7h6 g5e3 b7b5 d5c6 b5c4 c6d7 f8b8 a1b1 e7e6 e1g1 b8b7 f1d1 b7d7 c3d5 f6d5 d1d5 a8c8 b2b4 c4b3 c2b3 c8c6 b1c1 c6c1 e3c1 d7c7 c1e3');
INSERT INTO evaluations VALUES(33783,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(33783,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(33783,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(33783,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(33783,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(33783,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(33783,36,'cp 141','f1c4 c8g4 c1e3 b8c6 c4b5 g4d7 d4d5 c6b8 b5e2 e7d8 f3d2 f8e7 a2a4 e8g8 d2c4 b8a6 e1g1 c7c6 f2f4 e5f4 f1f4 f8e8 d1d2 c6d5 e4d5 a6c5 e3c5 d6c5 d5d6 e7f8 a1d1 a8c8 h2h3 b7b6 f4f1 g8h8 b2b3 h7h6 e2f3 d7e6');
INSERT INTO evaluations VALUES(46639,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(46639,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(46639,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46639,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(46639,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(46639,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(46639,36,'cp 28','f8g7 d2d4 e8g8 c4b3 c8g4 h2h3 g4f3 d1f3 b8c6 f3d1 e7e5 d4e5 d6e5 d1d8 f8d8 c1g5 c6d4 e1c1 c7c6 a2a4 d8d7 h1e1 d4b3 c2b3 a8d8 f2f3 a7a6 c1c2 h7h6 g5h4 d7d1 c3d1 d8d7 d1e3 f6h5 g2g3 f7f6 e1d1 d7d1 e3d1 g8f7');
INSERT INTO evaluations VALUES(25153,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25153,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25153,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25153,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(25153,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(25153,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(25153,36,'cp 17','f8d6 f4g3 e8g8 f1e2 d8e7 f3h4 h7h6 h4f5 e6f5 e1g1 d6g3 h2g3 g7g6 c3c4 d5c4 e2f3 c6d8 d2c4 c7c6 a2a4 d8e6 a4a5 a7a6 d1b3 h6h5 a1a4 a8c8 b3a3 e7a3 b2a3 c8c7 f1b1 f8d8 a4b4 c6c5 b4b7 c5d4 b7c7 e6c7');
INSERT INTO evaluations VALUES(24105,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24105,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24105,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(24105,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24105,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(24105,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(24105,36,'cp 14','f4g3 e8g8 f1e2 f8e8 f3h4 c6e7 e1g1 e7g6 h4f5 e6f5 g3d6 d8d6 g2g3 a7a5 c3c4 g6e7 d1c2 c7c6 c4d5 f6d5 e2f3 g7g6 d2c4 d6c7 a2a3 h7h5 f1c1 d5f6 c2a4 e7d5 c4e5 g8g7 c1c5 f6e4 f3e4 f5e4 a4b3');
INSERT INTO evaluations VALUES(24110,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24110,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(24110,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(24110,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(24110,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(24110,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(24110,36,'cp 5','e8g8 f1e2 f8e8 f3h4 h7h6 h4f5 e6f5 g3h4 a8b8 h4f6 d8f6 g2g3 c6e7 e1g1 g7g6 a1c1 g8g7 c3c4 c7c6 d2f3 d5c4 e2c4 e7c8 c4b3 f6e7 h2h4 c8b6 h4h5 b6d7 d1c2 d7f6 h5g6 f7g6 g1g2 b8d8 f1h1 a7a5 a2a3');
INSERT INTO evaluations VALUES(23724,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(23724,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(23724,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(23724,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(23724,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(23724,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(23724,36,'cp 8','a2a3 a7a6 f1e2 f8e8 f3h4 c6e7 c3c4 e7g6 h4f3 d6g3 h2g3 c7c6 d1b3 d8c7 e1g1 a8d8 f1d1 h7h5 a1c1 a6a5 b3c3 a5a4 c3b4 d8a8 d2b1 c7e7 b4e7 g6e7 b1c3 f5g4 f3e5 g4e2 c3e2 g7g6 f2f3 g8g7 c4c5 e8b8 e2c3');
INSERT INTO evaluations VALUES(17018,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(17018,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(17018,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(17018,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(17018,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(17018,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(17018,36,'cp 51','h7h6 g5h4 d7d6 b1c3 g7g5 h4g3 f8g7 f1b5 c8d7 e1c1 d7b5 c3b5 a7a6 b5c3 e8g8 d4b4 a6a5 b4b5 f6d7 h2h4 f8e8 g3d6 c7d6 d1d6 g7c3 b2c3 g5h4 d6d7 d8g5 b5g5 h6g5 f2f3 a8c8 c1d2 e8e5 h1b1 b7b5 d7d5 e5d5 e4d5 c8c5 d5d6 c5d5 d2e3 g8f8 e3e4 d5d6 b1b5');
INSERT INTO evaluations VALUES(16214,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(16214,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(16214,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(16214,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(16214,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(16214,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(16214,36,'cp 43','g5h4 g7g5 h4g3 f8g7 b1c3 d7d6 f1b5 c8d7 e1c1 d7b5 c3b5 a7a6 b5c3 e8g8 f2f3 f6h5 d4b4 f8e8 c3d5 h5g3 h2g3 a6a5 b4b7 a8b8 b7c7 b8b2 c7d8 e8d8 a2a4 b2a2 c1d2 a2a4 d1b1 d8e8 b1b7 a4d4 d2c1 d4a4 h1d1 g7e5 d5e7 g8f8 e7f5 a4a1 c1d2 e5c3 d2e2 a1d1 e2d1 a5a4 b7a7');
INSERT INTO evaluations VALUES(53263,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(53263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53263,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53263,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53263,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53263,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53263,36,'cp 0','e2e4 d5e4 g1e2 c5d4 d1d4 c8d7 c1e3 b8c6 d4c4 g8f6 e1c1 c6a5 c4d4 a5c6');
INSERT INTO evaluations VALUES(53277,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53277,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53277,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53277,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53277,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53277,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53277,36,'cp 5','d5e4 d4d5 g8f6 g1e2 e7e6 e2f4 e6d5 f4d5 f6d5 d1d5 d8d5 c3d5 f8d6 c1e3 c8f5 e1c1 b8d7 d5f4 d6e5 f1e2 d7f6 f4d5 b7b6 f2f4 e4f3 e2f3 a8d8 e3g5 d8d5 f3d5 f6d5 d1d5 f7f6 h1e1 e8f7 e1e5 f6e5 d5e5 f5e6 a2a4 h7h6 g5h4 g7g5 h4f2 f7e7 a4a5');
INSERT INTO evaluations VALUES(53207,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(53207,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53207,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53207,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53207,36,'cp 0','d4d5 g8f6 g1e2 e7e6 e2f4 f6d5 f4d5 e6d5 d1d5 d8d5 c3d5 f8d6 c1e3 c8f5 e1c1 b8d7 f1e2 e8c8 d5c3 d6e5 c3b5 b7b6 b5a7 c8b7 a7b5 d7f6 e2c4 f5e6 c4e6 f7e6 c2c4 f6g4 h2h3 g4e3 f2e3 b7c6 d1d8 h8d8 h1f1 d8d3 f1f7 d3e3 b5a7 c6d6 a7c8 d6c6');
INSERT INTO evaluations VALUES(19204,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(19204,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(19204,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(19204,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(19204,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(19204,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(19204,36,'cp 55','b7c6 e1g1 f8g7 f1e1 g8h6 c2c3 e8g8 d2d4 c5d4 c3d4 d7d5 e4e5 f7f6 e5f6 e7f6 b1d2 f8e8 d2b3 h6f5 c1d2 g7f8 e1e8 d8e8 h2h3 f5d6 d2b4 e8d8 a1c1 a7a5 b4d6 d8d6 d1d2 a5a4 b3c5 c8f5 d2c3 g6g5 h3h4 h7h6 b2b4 a8e8 h4g5');
INSERT INTO evaluations VALUES(18412,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(18412,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(18412,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(18412,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(18412,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(18412,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(18412,36,'cp 37','e1g1 f8g7 f1e1 g8h6 d2d3 e8g8 h2h3 f7f6 c2c3 h6f7 c1e3 d7d6 d3d4 e7e5 b1d2 c5d4 c3d4 e5d4 e3d4 c6c5 d4c3 c8b7 d2c4 a8b8 f3d2 f7g5 c4e3 g5e4 d2e4 b7e4 e3d5 f8e8 b2b3 f6f5 c3g7 g8g7');
INSERT INTO evaluations VALUES(18411,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(18411,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(18411,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(18411,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(18411,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(18411,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(18411,36,'cp 36','f8g7 f1e1 g8h6 c2c3 e8g8 d2d4 c5d4 c3d4 d7d5 e4e5 f7f6 e5f6 e7f6 b1d2 h6f5 d2b3 f8e8 c1d2 e8e1 d1e1 f5d6 a1c1 d6c4 d2a5 d8f8 e1c3 f8f7 h2h3 g7h6 c1e1 c8f5 f3d2 c4d2 b3d2 a8e8 d2f3 e8e1 f3e1 f7d7 a5b4 f5e4 e1d3');
INSERT INTO evaluations VALUES(5027,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(5027,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(5027,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(5027,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(5027,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5027,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(5027,36,'cp 33','f1e1 g8h6 c2c3 e8g8 d2d4 c5d4 c3d4 d7d5 e4e5 f7f6 e5f6 e7f6 b1d2 h6f5 d2b3 f8e8 h2h3 f5d6 c1f4 d6c4 d1c2 c8f5 c2c3 g6g5 f4g3 g7f8 e1e8 d8e8 a1e1 e8f7 b3c5 a7a5 f3d2 c4d2 c3d2 a5a4 e1c1 h7h6 a2a3');
INSERT INTO evaluations VALUES(5028,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(5028,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(5028,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(5028,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(5028,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(5028,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(5028,36,'cp 49','g8h6 c2c3 e8g8 d2d4 c5d4 c3d4 d7d5 e4e5 f7f6 e5f6 e7f6 b1d2 f8e8 d2b3 e8e1 d1e1 h6f5 c1d2 g7f8 d2b4 f5d6 e1c3 c8d7 a2a3 d6b5 c3e3 b5d6 a1e1 d8c7 e3c3 a8e8 c3c1 g6g5 b3c5 d7f5 e1e8 d6e8');
INSERT INTO evaluations VALUES(53340,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(53340,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(53340,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(53340,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(53340,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(53340,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(53340,36,'cp -48','c8g4 c2c3 b8c6 f1b5 d8b6 b5c6 b6c6 h2h3 g4h5 e1g1 e7e6 d2d4 h7h6 b1d2 a7a5 d4c5 f8c5 d2b3 h5f3 d1f3 g8e7 c1e3 c5b6 b3d4 b6d4 e3d4 e8g8 f1d1 b7b5 a2a3 e7f5 d1d2 f8b8 a1d1 b5b4');
INSERT INTO evaluations VALUES(31285,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(31285,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(31285,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(31285,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(31285,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(31285,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(31285,36,'cp -12','h2h3 g4f3 f1b5 b8d7 d1f3 e7e6 c2c4 g8e7 b1c3 a7a6 b5d7 d8d7 d2d3 d7c6 c4d5 e7d5 c3d5 c6d5 f3d5 e6d5 c1e3 f7f6 e5f6 g7f6 g2g4 f8d6 e1e2 e8f7 e2f3 h7h5 a1e1 a8d8 e3f4 d8d7 f4d6 d7d6 f3g2 b7b6 b2b3 d6e6');
INSERT INTO evaluations VALUES(31284,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(31284,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(31284,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(31284,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(31284,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(31284,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(31284,36,'cp -25','g4f3 f1b5 b8d7 d1f3 e7e6 c2c4 d8c7 e1g1 d5c4 d2d3 c4d3 f1d1 g8e7 d1d3 e7c6 b1a3 d7e5 f3g3 a8d8 d3d8 c7d8 c1g5 f7f6 g3e5 f6e5 b5c6 b7c6 g5d8 e8d8 a3c4 f8e7 c4e5 d8c7 b2b3 h8d8 e5c4 e7f6 a1e1 c7d7 g1f1 d7e7 f1e2 d8d4 f2f3 e7d7 e1c1 d4d5 c4e3 d5e5 c1d1 d7c8 f3f4 e5e4 e2f3');
INSERT INTO evaluations VALUES(1473,30,'cp -260',NULL);
INSERT INTO evaluations VALUES(1473,31,'cp -273',NULL);
INSERT INTO evaluations VALUES(1473,32,'cp -262',NULL);
INSERT INTO evaluations VALUES(1473,33,'cp -247',NULL);
INSERT INTO evaluations VALUES(1473,34,'cp -309',NULL);
INSERT INTO evaluations VALUES(1473,35,'cp -300',NULL);
INSERT INTO evaluations VALUES(1473,36,'cp -291','b6b2 b1d2 c5c4 a1c1 b2a2 g2g3 a2a5 f1g2 a5c7 f3g5 h7h6 g5h3 b7b5 h3f4 a8b8 e1g1 a7a5 f4h5 g7g6 h5f6 g8f6 e5f6 f8d6 f1e1 c8d7 d1g4 b5b4 c3b4 c6b4 g2f1 c4c3 d2b3 e6e5 g4h4 e5e4 b3c5 c3c2 e3h6 d7f5 c5a6 b4a6 f1a6 e8d7 a6e2');
INSERT INTO evaluations VALUES(1538,30,'cp -305',NULL);
INSERT INTO evaluations VALUES(1538,31,'cp -290',NULL);
INSERT INTO evaluations VALUES(1538,32,'cp -287',NULL);
INSERT INTO evaluations VALUES(1538,33,'cp -300',NULL);
INSERT INTO evaluations VALUES(1538,34,'cp -290',NULL);
INSERT INTO evaluations VALUES(1538,35,'cp -284',NULL);
INSERT INTO evaluations VALUES(1538,36,'cp -306','b1d2 c5c4 a1c1 b2a2 f3g5 b7b5 d1h5 g7g6 h5f3 c6d8 f3d1 h7h6 g5f3 g8e7 h2h4 e7f5 h4h5 g6g5 f3h2 d8c6 f1e2 a2a5 e1g1 f8e7 h2g4 a5c7 d2f3 a7a5 c1a1 a8b8 e3d2 b5b4 f1e1 e8g8 g4f6 e7f6 e5f6 c7d8 g2g4 f5d6 f3e5 c6e5 d4e5 d6e4 c3b4 a5b4 a1a7 b4b3 d2a5');
INSERT INTO evaluations VALUES(1537,30,'cp -309',NULL);
INSERT INTO evaluations VALUES(1537,31,'cp -316',NULL);
INSERT INTO evaluations VALUES(1537,32,'cp -291',NULL);
INSERT INTO evaluations VALUES(1537,33,'cp -320',NULL);
INSERT INTO evaluations VALUES(1537,34,'cp -325',NULL);
INSERT INTO evaluations VALUES(1537,35,'cp -324',NULL);
INSERT INTO evaluations VALUES(1537,36,'cp -317','c5c4 a1c1 b2a2 g2g3 a2a5 f1g2 a5c7 f3g5 b7b5 e1g1 h7h6 g5h3 a7a5 h3f4 c8d7 h2h4 a5a4 c1b1 a8b8 b1a1 c6a5 a1a2 a5b3 g2h3 b5b4 c3b4 f8b4 d1g4 h8h7 d2b1 g7g6 g1h2 g8e7 g4d1 c4c3 f4e2 c7c4 e2c3 b4c3 b1c3 c4c3');
INSERT INTO evaluations VALUES(1543,30,'cp -293',NULL);
INSERT INTO evaluations VALUES(1543,31,'cp -302',NULL);
INSERT INTO evaluations VALUES(1543,32,'cp -294',NULL);
INSERT INTO evaluations VALUES(1543,33,'cp -307',NULL);
INSERT INTO evaluations VALUES(1543,34,'cp -287',NULL);
INSERT INTO evaluations VALUES(1543,35,'cp -283',NULL);
INSERT INTO evaluations VALUES(1543,36,'cp -317','a1c1 b2a2 g2g3 a2a5 f1g2 b7b5 e1g1 a5c7 f3g5 h7h6 g5h3 a7a5 h3f4 c8d7 h2h4 a5a4 c1b1 a8b8 d1e2 g8e7 g2h3 g7g6 d2f3 b5b4 c3b4 c6b4 b1a1 f8g7 e3d2 e8g8 d2c3 b4c6 e2e3 b8b3 f3d2');
INSERT INTO evaluations VALUES(36295,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(36295,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(36295,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36295,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(36295,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(36295,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(36295,36,'cp 0','g1f3 h7h6 g5f6 e7f6 a1c1 c7c6 f1e2 d5c4 e2c4 b8d7 e1g1 c6c5 c3e4 c5d4 e4f6 d7f6 d1d4 b7b6 c4b5 c8b7 b5c6 b7c6 c1c6 a8c8 f1c1 c8c6 c1c6 f6d5 d4c4 d5e7 c6c7 e7d5');
INSERT INTO evaluations VALUES(57041,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57041,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57041,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57041,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(57041,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57041,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(57041,36,'cp -12','d7d5 e4d5 e6d5 f1g2 g8f6 g1f3 f8e7 e1g1 e8g8 f1e1 f8e8 b1c3 d5d4 c3e4 f6e4 e1e4 d4d3 c2d3 b8c6 d3d4 e7f6 e4e8 d8e8 d4d5 c6b4 d1b3 e8e2 g2f1 e2e4 d2d3 e4f3 b3b4 f3d5 c1f4 a7a5 b4d2 d5d7 a1e1 a5a4 f1g2');
INSERT INTO evaluations VALUES(55911,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55911,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(55911,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(55911,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(55911,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(55911,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55911,36,'cp -8','e4d5 e6d5 f1g2 b8c6 g1e2 g8f6 d2d4 f8e7 e1g1 e8g8 c1g5 h7h6 g5f6 e7f6 c2c3 c6e7 e2f4 c7c6 b1d2 d8b6 b2b3 g7g6 f1e1 c8f5 h2h4 f8e8 d2f1 a8d8 f1e3 h6h5 a2a4 a7a5 e3f5 e7f5 d1c2');
INSERT INTO evaluations VALUES(55782,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(55782,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55782,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(55782,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(55782,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(55782,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55782,36,'cp -11','e6d5 f1g2 g8f6 g1f3 h7h6 e1g1 f8e7 f1e1 e8g8 b1c3 f8e8 d2d3 c7c5 d3d4 c8e6 a2a3 b8c6 d4c5 e7c5 b2b4 c5f2 g1f2 d8b6 f2f1 f6g4 d1e2 e6d7 c3d5 e8e2 d5b6 e2e1 f3e1 a7b6 f1g1 c6b4 g2b7 a8a7');
INSERT INTO evaluations VALUES(56294,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(56294,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(56294,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(56294,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(56294,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(56294,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56294,36,'cp 0','g1f3 f8e7 f1g2 g8f6 e1g1 e8g8 f1e1 f8e8 b1c3 c8f5 d2d3 h7h6 a2a3 c7c5 d3d4 c5d4 f3d4 f5g4 d4e2 e7c5 c3a4 c5d6 h2h3 g4h5 a4c3 d6c5 c3a4 c5d6');
INSERT INTO evaluations VALUES(56290,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(56290,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(56290,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(56290,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(56290,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(56290,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(56290,36,'cp -10','f8e7 f1g2 g8f6 e1g1 e8g8 f1e1 h7h6 b1c3 f8e8 d2d3 c8f5 h2h3 c7c6 a2a3 b8d7 c3e2 a7a5 b2b3 e7d6 e2d4 f5h7 e1e8 f6e8 c1b2 d8b6 d4e2 e8f6 b2c3 c6c5 a3a4 d5d4 c3d2 a8e8 d1f1 h7f5 a1e1 d7e5 f3e5');
INSERT INTO evaluations VALUES(22047,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(22047,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(22047,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(22047,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(22047,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(22047,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(22047,36,'cp 46','b1c3 g8f6 c1d2 g7g6 e1c1 f8g7 h2h4 h7h5 c1b1 e8g8 f2f3 c6e5 e3e1 c7c6 d2g5 b7b5 f3f4 e5d7 e4e5 d6e5 f4e5 f8e8 c3e4 e8e5 e4f6 d8f6 g5f6 e5e1 d1e1 d7f6 g1f3 a7a6 f1e2 f6d5 e2d3 c8d7 g2g3 a8f8 f3e5 d7e8 h1f1 a6a5 a2a3 g7f6 c2c3 g8g7 d3c2 a5a4 b1a1');
INSERT INTO evaluations VALUES(22041,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(22041,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(22041,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(22041,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(22041,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(22041,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(22041,36,'cp 42','g8f6 c1d2 g7g6 e1c1 f8g7 h2h4 h7h5 c1b1 e8g8 f2f3 c6e5 g2g4 h5g4 f3f4 e5d7 g1e2 f8e8 h4h5 f6h5 f4f5 c7c6 e2g3 d7f6 g3h5 g6h5 e3g5 f6e4 c3e4 d8g5 e4g5 c8f5 h1h5 e8e7 f1d3 f5d3 c2d3 f7f5 b1c1 a8e8 d1h1 e7e2 c1c2');
INSERT INTO evaluations VALUES(15650,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(15650,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(15650,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(15650,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(15650,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(15650,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(15650,36,'cp 48','c1d2 g7g6 e1c1 f8g7 h2h4 h7h5 e3g3 e8g8 f2f3 d8e8 d2g5 a7a6 g1h3 c8h3 g3h3 e8e6 h3e6 f7e6 c3e2 e6e5 e2c3 g8h7 g5f6 f8f6 c3d5 f6f7 c2c3 g7h6 c1b1 h6f4 g2g4 h5g4 f3g4 a8h8 g4g5 h7g7 f1c4 b7b5 c4e2');
INSERT INTO evaluations VALUES(15651,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(15651,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(15651,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(15651,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(15651,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(15651,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(15651,36,'cp 50','g7g6 e1c1 f8g7 h2h4 h7h5 e3g3 e8g8 f2f3 d8e8 g1e2 a7a6 c1b1 b7b5 e2f4 c8e6 d2c1 c6e5 f1e2 e8d7 g3e1 e6c4 b2b3 c4e2 e1e2 f8b8 c3d5 f6d5 f4d5 c7c6 d5e3 e5c4 g2g4 c4e3 e2e3 h5g4 f3g4 b8e8 h4h5');
INSERT INTO evaluations VALUES(53307,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53307,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(53307,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53307,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(53307,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(53307,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53307,36,'cp 12','e2e3 g8f6 g1f3 e7e6 d2d4 c5d4 e3d4 f8d6 f1d3 b8c6 a2a3 e8g8 b1d2 b7b6 f3e5 c8b7 e1g1 a8c8 d1e2 c6e7 c2c3 e7f5 a3a4 d6e7 f1e1 a7a5 e5g4 f6g4 e2g4 c8c7 d2f3 f5d6 a1b1 d6e4');
INSERT INTO evaluations VALUES(53301,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53301,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53301,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53301,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53301,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53301,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53301,36,'cp 1','e7e6 g1f3 b8c6 d2d4 c5d4 e3d4 g8f6 f1d3 f8d6 e1g1 e8g8 c2c4 b7b6 b1c3 c6e7 c3b5 d6b8 d1e2 a7a6 b5c3 c8b7 g2g3 b8d6 f1d1 h7h6 f3e5 e7c6 c3a4 d5c4 b3c4 d6e7 e5c6 b7c6 a4c3 c6b7 d4d5');
INSERT INTO evaluations VALUES(54264,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54264,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54264,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54264,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54264,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54264,36,'cp 0','g1f3 b8c6 f1b5 g8e7 e1g1 a7a6 b5c6 e7c6 d2d4 f8e7 d4c5 e8g8 d1d2 f7f6 b2a3 e6e5 b1c3 c8e6 f1d1 d8a5 c3a4 a5b5 a4c3');
INSERT INTO evaluations VALUES(54265,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54265,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54265,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54265,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54265,36,'cp 5','b8c6 f1e2 g8f6 e1g1 f8e7 d2d4 c5d4 f3d4 e8g8 d4c6 b7c6 c2c4 c8b7 b1d2 e7d6 a1c1 d8e7 d2f3 a7a5 c4d5 c6d5 f3d4 a8c8 d4b5 d6c5 c1c2 c5b4 c2c8 f8c8 d1a1 b7a6 f1c1 c8c1 a1c1 e7d7 b5d4 a6e2 d4e2 g8h8');
INSERT INTO evaluations VALUES(20308,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20308,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(20308,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(20308,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(20308,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(20308,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20308,36,'cp 1','f1e2 g8f6 e1g1 f8e7 d2d4 c5d4 f3d4 e8g8 c2c4 c6d4 b2d4 b7b6 b1c3 c8b7 c4d5 f6d5 c3d5 b7d5 e2a6 e7f6 a1c1 f6d4 d1d4 d8e7 d4a4 d5b7 a6b7 e7b7 h2h3 h7h6 c1c6 a8c8 f1c1 c8c6 a4c6 b7a6 c6c2 f8d8 a2a4 a6d3 c2d3 d8d3 c1c8 g8h7 c8c7 h7g6 c7a7 d3b3 a4a5');
INSERT INTO evaluations VALUES(19340,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(19340,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(19340,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(19340,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(19340,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(19340,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(19340,36,'cp 65','e1g1 e7e6 f1e1 g8f6 c4b5 d7d6 e4e5 f6d7 b5c6 b7c6 d2d3 d6d5 b1d2 f8e7 d2f1 h6h5 h2h3 g7g5 f1g3 c8b7 f3h2 d8c7 g3h5 e8c8 f2f4 d8g8 c2c3 g5f4 c1f4 e7h4 d1f3 c5c4 d3d4 c6c5 f4g3 h4g3 h5g3 c5d4 c3d4');
INSERT INTO evaluations VALUES(19339,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(19339,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(19339,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(19339,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(19339,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(19339,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(19339,36,'cp 37','e7e6 d2d4 c5d4 f3d4 d7d6 f1e1 g8f6 c4f1 c6d4 d1d4 d6d5 e4d5 d8d5 d4h4 c8d7 b1c3 d5f5 f1d3 f5g4 h4g4 f6g4 h2h3 g4f6 c3b5 e8d8 c1d2 d7b5 d3b5 f8d6 g2g3 d8e7 a2a4 a8c8 d2c3 h8d8 a1d1 c8c7 a4a5 d6c5 d1d8 e7d8');
INSERT INTO evaluations VALUES(18654,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(18654,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(18654,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(18654,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(18654,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(18654,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(18654,36,'cp 35','f1e1 g8e7 c2c3 d7d5 e4d5 e7d5 d2d4 c5d4 f3d4 c6d4 d1d4 d5b6 d4d8 e8d8 e1d1 c8d7 c4e2 d8c8 c1e3 f8e7 a2a4 h8d8 a4a5 b6d5 e3d4 e7f6 d4f6 d5f6 b2b4 c8c7 c3c4 d7e8 d1c1 a8c8 b1c3 c7b8 f2f3 d8d2 a1a2 c8d8 a2d2 d8d2 g1f2');
INSERT INTO evaluations VALUES(48419,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48419,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(48419,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(48419,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(48419,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(48419,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48419,36,'cp 28','e1g1 c7c5 c2c4 d5d4 d2d3 b8c6 e2e3 f8d6 e3d4 c5d4 c1g5 h7h6 g5f6 d8f6 b1d2 f6e7 d2e4 e8g8 e4d6 e7d6 a1c1 d6c7 f1e1 c8d7 d1e2 a7a5 f3e5 d7e8 b2b3 a8d8 a2a3 b7b6 c4c5 c7e5 e2e5 c6e5');
INSERT INTO evaluations VALUES(48418,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(48418,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(48418,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(48418,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(48418,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(48418,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(48418,36,'cp 14','c7c5 c2c4 d5d4 e2e3 b8c6 d2d3 h7h6 e3d4 c5d4 c1f4 f8d6 f4d6 d8d6 b1d2 e8g8 f1e1 a8b8 d1e2 f6d7 d2e4 d6e7 a2a3 a7a5 f3h4 g7g5 h4f3 f7f5 e4d2 e6e5 c4c5 a5a4 b2b4 a4b3 d2b3 e7f7 f3d2 d7f6 g2c6 b7c6 e2e5 b8b3 d2b3 f7b3 e5d4');
INSERT INTO evaluations VALUES(46281,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46281,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(46281,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(46281,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(46281,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(46281,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(46281,36,'cp 33','c2c4 d5d4 e2e3 b8c6 d2d3 h7h6 e3d4 c5d4 c1f4 f8d6 f4d6 d8d6 b1d2 e8g8 f1e1 a8b8 d1e2 f6d7 d2e4 d6e7 f3h4 a7a5 a2a3 a5a4 f2f4 d7c5 e1f1 c8d7 f4f5 c5e4 g2e4 e7g5 f5e6 d7e6 h4f5 b8e8 a1e1 e6f5 f1f5 g5g6 c4c5 c6e5');
INSERT INTO evaluations VALUES(46163,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(46163,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46163,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(46163,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(46163,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46163,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(46163,36,'cp 26','d5d4 e2e3 b8c6 e3d4 c5d4 d2d3 h7h6 c1f4 f8d6 f4d6 d8d6 f1e1 e8g8 b1d2 a8b8 d1e2 f6d7 d2e4 d6e7 a1c1 b7b6 f3h4 a7a5 f2f4 d7c5 e4c5 e7c5 b2b3 f8e8 g3g4 c8b7 g2e4 c6b4 a2a3 b4a6 e4b7 b8b7 e2g2 b7c7 g4g5 h6g5 f4g5 c5a3 g2g3');
INSERT INTO evaluations VALUES(12074,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12074,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12074,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12074,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12074,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12074,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12074,36,'cp 0','f8e7 b5d3 e8g8 e1g1 h7h6 g5e4 f6d5 e4g3 d5b4 b2b3 f7f5 g3f5 e7c5 f3h5 b4d3 f5h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(7703,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7703,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7703,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7703,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7703,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7703,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7703,36,'cp 0','b5d3 e8g8 e1g1 h7h6 g5e4 f6d5 e4g3 d5b4 b2b3 f7f5 g3f5 e7c5 f3h5 b4d3 f5h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(41398,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(41398,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(41398,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(41398,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(41398,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(41398,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(41398,36,'cp 59','f1d3 e7e5 d4e5 d6e5 g1f3 e5f4 c1f4 e8g8 d1d2 b8c6 e1c1 f6g4 h1e1 g4e5 d2e3 e5f3 g2f3 d8d4 d3f1 d4b6 e3b6 a7b6 f4c7 g7c3 b2c3 a8a2 c7b6 c8e6 f3f4 c6a5 c1d2 f8c8 b6a5 a2a5 d1a1 a5c5 f1d3 e6c4 e1b1 c4d3 c2d3 c5c3');
INSERT INTO evaluations VALUES(41399,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(41399,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(41399,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(41399,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(41399,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(41399,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(41399,36,'cp 55','b8c6 e4e5 d6e5 f4e5 f6d5 g1f3 c8g4 c3d5 d8d5 c2c3 f7f6 e5f6 e7f6 d1e2 d5e6 h2h3 e6e2 e1e2 g4d7 e2f2 e8c8 c1d2 h8e8 h1e1 g7f8 g2g4 d7e6 d2f4 f8d6 f4d6 d8d6 d3c4 c6a5 c4b3 a5b3 a2b3 c8b8 f3d2 f6f5 g4f5 g6f5');
INSERT INTO evaluations VALUES(17268,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(17268,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(17268,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(17268,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(17268,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(17268,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(17268,36,'cp 208','d4d5 c6e5 f2f4 e5g6 e4e5 f6g8 f1d3 e7e6 d1e2 f8b4 c2c3 b4c5 g1f3 d8e7 b2b4 c5b6 a2a4 a7a5 d5e6 f7f6 e5f6 g8f6 f4f5 g6f8 b4a5 b6c5 b1d2 d7e6 d2b3 c5d6 e1g1 e6f5 e2e7 e8e7 f3d4 f8g6 d4f5 c8f5 d3f5 g6e5 g1h1 e7e8 c1f4');
INSERT INTO evaluations VALUES(17222,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(17222,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(17222,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(17222,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(17222,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(17222,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(17222,36,'cp 199','c6e5 f2f4 e5g6 e4e5 f6g8 f1d3 g8h6 d1e2 e7e6 c1d2 d7d6 d5e6 c8e6 e5d6 d8f6 d6c7 a7a6 d3g6 f6g6 b1c3 f8d6 e1c1 e8g8 g1f3 d6c7 h2h3 f8e8 h1e1 e6d7 f3e5 c7e5 f4e5 h6f5 e2f2 f5e7 d2f4 d7e6 c3e4 e8c8 e4d6 c8c6 a2a3 h7h6 d1d2 a8d8 g2g4 g8h7');
INSERT INTO evaluations VALUES(17314,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(17314,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(17314,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(17314,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(17314,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(17314,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(17314,36,'cp 181','f2f4 e5g6 e4e5 f6g8 f1d3 g8h6 d1e2 e7e6 c1d2 f8e7 b1c3 d7d6 e1c1 e6d5 c3d5 e8g8 d2b4 f8e8 d5e7 d8e7 e5d6 c7d6 e2e7 g6e7 d3b5 e8d8 g1f3 c8g4 d1d6 d8d6 b4d6 e7d5 f3e5 a8d8 d6c5 a7a6 b5c4 d5f4 g2g3 f4g6 e5g4');
INSERT INTO evaluations VALUES(55404,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(55404,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(55404,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(55404,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(55404,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(55404,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(55404,36,'cp 40','g8f6 d2d4 e5d4 e1g1 f8e7 f1e1 e8g8 f3d4 c7c5 d4f3 c8e6 c4e6 f7e6 e4e5 d6e5 f3e5 d8d1 e1d1 f8c8 a2a4 b8c6 e5c4 f6d5 c2c3 d5b6 b1d2 c8d8 d1e1 g8f7 g1f1 e7f6 f1e2 b6c4 d2c4 d8d5 c1f4 a8d8 a1d1 f7e7 c4e3 d5d1 e1d1 d8d1 e3d1 c5c4');
INSERT INTO evaluations VALUES(47624,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(47624,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(47624,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(47624,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(47624,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(47624,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(47624,36,'cp 48','d2d4 e5d4 e1g1 f8e7 f1e1 e8g8 f3d4 c7c5 d4f3 c8e6 c4e6 f7e6 e4e5 d6e5 f3e5 d8d1 e1d1 f8c8 a2a4 b8c6 e5c4 c8d8 d1e1 f6d5 c2c3 g8f7 b1d2 d5b6 g1f1 e7f6 a4a5 b6c4 d2c4 d8d5 c1f4 a8d8 e1e4 e6e5 f4e3 f7e6 f1e2 c6e7 h3h4 e7f5 h4h5 f5e3');
INSERT INTO evaluations VALUES(34692,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(34692,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(34692,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(34692,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(34692,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(34692,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(34692,36,'cp 12','g7g6 g2g3 g8f6 f1g2 f8g7 c1d2 c7c5 g1f3 b8c6 e1g1 e8g8 f1e1 c8g4 c3d5 a5d8 d5e3 g4e6 d2c3 a8b8 f3g5 e6d7 e3c4 c6d4 a2a4 d7c6 g2c6 d4c6 d1e2 b8c8 a4a5 d8c7 c3e5 c6e5 e2e5 c7e5 c4e5');
INSERT INTO evaluations VALUES(18676,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(18676,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18676,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(18676,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(18676,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(18676,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18676,36,'cp 36','b7c6 d2d3 g8e7 h2h4 d7d6 h4h5 e6e5 c2c3 c8g4 b1a3 d8d7 h5h6 f7f6 h6g7 f8g7 c1h6 g7h6 h1h6 d6d5 a3c2 d5d4 d1d2 e7g6 e1c1 e8g8 c3d4 g4f3 g2f3 c5d4 d1g1 f8f7 f3f4 f7g7 f4e5 f6e5 d2g5');
INSERT INTO evaluations VALUES(18380,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(18380,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(18380,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(18380,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(18380,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(18380,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(18380,36,'cp 28','d2d3 g8e7 h2h4 h7h5 e4e5 d7d6 b1c3 f7f6 e5d6 e7f5 c3e4 f8d6 e1g1 e6e5 c1e3 f5e3 e4d6 d8d6 f2e3 c8g4 d1e1 e8g8 f3d2 f6f5 e1g3 a8e8 a1e1 e8e6 d2f3 g4f3 g3f3 e5e4 d3e4 e6e4 f3h5 d6g3 h5g5 g3h4 g5h4');
INSERT INTO evaluations VALUES(18378,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(18378,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(18378,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18378,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18378,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(18378,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(18378,36,'cp 38','g8e7 h2h4 h7h5 e4e5 d7d6 b1c3 e7f5 c3e4 f7f6 e5d6 f8d6 e1g1 e6e5 f1e1 d6e7 c1e3 d8b6 c2c3 e8g8 f3d2 f5e3 e1e3 c8g4 d1a4 f6f5 e4g5 b6b2 d2b3 e5e4 a4c4 g8h8 e3e1 e7f6 b3c5 b2c3 a1c1 c3c4 c1c4 f6g5 h4g5 h8h7 d3e4 f5e4');
INSERT INTO evaluations VALUES(57062,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57062,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(57062,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(57062,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57062,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(57062,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57062,36,'cp 7','g8f6 e2e3 d7d5 g1f3 f8d6 d2d4 b7b6 f1b5 c7c6 b5d3 c6c5 c2c4 d5c4 d3c4 e8g8 e1g1 c8b7 d4c5 d6c5 d1e2 d8e7 a2a3 a7a5 f1d1 f8d8 b1d2 b8d7 c4b5 h7h6 h2h3 a8c8 b5a6 b7a6 e2a6 e6e5 d2c4 e5e4 f3d4 d7e5 c4e5');
INSERT INTO evaluations VALUES(49465,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(49465,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(49465,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49465,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(49465,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(49465,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49465,36,'cp 7','e2e3 c7c5 g1f3 f8e7 c2c4 e8g8 b1c3 d7d5 c4d5 f6d5 f1e2 b8c6 e1g1 d5c3 b2c3 b7b6 d2d4 c5d4 f3d4 c6d4 c3d4 c8b7 e2f3 b7f3 d1f3 e7f6 a1d1 f6d4 d1d4 d8e7 f1d1 a8d8 d4d8 f8d8 d1d8 e7d8 h2h3 h7h6 f3c6 a7a5 g2g3 g7g6 c6b7 h6h5 b7c6 g8g7 g1g2 a5a4');
INSERT INTO evaluations VALUES(49461,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(49461,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49461,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49461,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49461,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(49461,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(49461,36,'cp 0','d7d5 c2c4 f8d6 b1c3 e8g8 g1f3 f8e8 d2d4 b8d7 d1c2 b7b6 f1d3 c8b7 e1g1 d5c4 b3c4 b7f3 g2f3 c7c5 d4d5 f6d5 c3d5 d8g5 g1h1 g5h5 d5f4 d6f4 e3f4 h5f3 h1g1 f3g4 g1h1');
INSERT INTO evaluations VALUES(48406,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(48406,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(48406,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48406,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(48406,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(48406,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(48406,36,'cp 11','c2c4 f8d6 g1f3 e8g8 b1c3 f8e8 d2d4 b8d7 f1e2 b7b6 a2a4 f6e4 e1g1 e4c3 b2c3 a7a5 c3b2 c8b7 c4d5 e6d5 b2a3 c7c5 e2b5 e8e6 a3b2 a8c8 a1c1 e6g6 d1e2 d8e7 f1d1 e7e6 g2g3 d7f6 f3e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(48230,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(48230,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48230,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(48230,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(48230,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(48230,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48230,36,'cp 6','f8d6 g1f3 e8g8 b1c3 f8e8 d2d4 b8d7 f1e2 b7b6 a2a4 c8b7 e1g1 a7a6 a4a5 d5c4 b3c4 c7c5 a5b6 d7b6 d4c5 d6c5 c3a4 b6a4 a1a4 f6e4 d1a1 c5f8 f1d1 d8c7 b2e5 c7e7 e5d4 e4c5 d4c5 e7c5 a4a5 c5b6 d1c1 e8c8 f3d4 f8b4 a5a2');
INSERT INTO evaluations VALUES(54174,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(54174,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(54174,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(54174,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(54174,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54174,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(54174,36,'cp 13','b8c6 g1f3 g8f6 c4d5 f6d5 a2a3 c5d4 e3d4 f8e7 f1d3 e8g8 e1g1 b7b6 d1e2 c8b7 e2e4 g7g6 c3d5 d8d5 e4d5 e6d5 h2h4 e7f6 c1e3 c6a5 e3g5 g8g7 f1e1 a8c8 f3e5 f6g5 h4g5 a5b3 a1d1 b3d4 d3g6 d4c2 g6c2 c8c2 e5d3 b7a6 e1e7 a6d3 d1d3');
INSERT INTO evaluations VALUES(20252,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(20252,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(20252,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(20252,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(20252,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(20252,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20252,36,'cp 20','g1f3 g8f6 c4d5 f6d5 a2a3 c5d4 e3d4 f8e7 f1d3 e8g8 e1g1 b7b6 d1e2 c8b7 e2e4 g7g6 c3d5 d8d5 e4d5 e6d5 h2h4 f8e8 c1g5 g8g7 a1c1 a8c8 g5e3 c6a5 e3f4 a5c4 f1e1 e7f6 e1e8 c8e8 b2b3 c4b2 d3b5 e8c8 f4e3 c8c1 e3c1 b2d1 c1d2');
INSERT INTO evaluations VALUES(20253,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(20253,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(20253,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(20253,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(20253,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(20253,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(20253,36,'cp 19','g8f6 c4d5 e6d5 f1b5 c5d4 e3d4 f8d6 c1g5 e8g8 e1g1 c8g4 b5e2 d6e7 h2h3 g4h5 f1e1 h7h6 g5h4 f8e8 f3e5 h5e2 e1e2 f6e4 e5c6 b7c6 h4e7 d8e7 d1c2 e7b7 c3e4 d5e4 a1e1 f7f5 f2f3 b7d7 f3e4 d7d4 g1h2 e8e4 e2e4 f5e4 e1e4 d4f6');
INSERT INTO evaluations VALUES(54233,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(54233,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(54233,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(54233,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(54233,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(54233,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(54233,36,'cp -12','c5d4 c3d4 d5e4 d2e4 g8f6 e4f6 d8f6 a2a3 f8d6 g1f3 h7h6 c1d2 b8c6 f1d3 e8g8 e1g1 e6e5 d4d5 c6e7 d1b3 c8f5 f1e1 f8d8 d3f5 f6f5 d2b4 e7g6 a1d1 d6b4 b3b4 d8d7 f3h4 g6h4 b4h4 f7f6 h4a4 d7d6 d1d2 f5d7 a4e4 a8d8');
INSERT INTO evaluations VALUES(54395,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(54395,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(54395,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(54395,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(54395,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54395,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(54395,36,'cp -15','c3d4 d5e4 d2e4 g8f6 f1d3 f6e4 d3e4 f8b4 c1d2 b4d2 d1d2 b8d7 g1e2 d7f6 e4f3 e8g8 e1g1 d8b6 a1c1 f8d8 f1d1 c8d7 h2h3 d7e8 c1c5 a8c8 c5c8 d8c8 d1c1 c8c1 e2c1 g8f8 d2c3 h7h6 c1d3 b6d6 d3e5 b7b6 g1h1');
INSERT INTO evaluations VALUES(54384,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(54384,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(54384,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(54384,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(54384,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(54384,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(54384,36,'cp -13','d5e4 d2e4 c8d7 g1f3 d7c6 e4c3 b8d7 f1c4 d7b6 c4d3 f8d6 e1g1 g8e7 d1e2 e8g8 f1d1 e7g6 f3e5 d6e5 d4e5 b6d5 d3g6 f7g6 c3d5 c6d5 f2f3 h7h6 e2e3 d8e8 a2a3 e8b5 c1d2 b5b3 d2c3 g6g5 d1d4 a7a5');
INSERT INTO evaluations VALUES(54458,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(54458,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(54458,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(54458,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(54458,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(54458,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(54458,36,'cp -8','d2e4 g8f6 f1d3 f6e4 d3e4 f8b4 c1d2 b4d2 d1d2 b8d7 g1e2 e8g8 e1g1 h7h6 a1c1 d7f6 e4f3 a8b8 f1d1 b7b6 e2c3 c8b7 d4d5 e6d5 c3d5 b7d5 f3d5 d8d5 d2d5 f6d5 d1d5 f8c8 c1c8 b8c8 g1f1 g8f8 d5d7 a7a5 f1e2 c8c2 d7d2 c2c1 d2d1 c1d1 e2d1 f8e7 d1c2 e7d6');
INSERT INTO evaluations VALUES(54457,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(54457,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54457,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(54457,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(54457,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(54457,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(54457,36,'cp -14','g8f6 f1d3 f6e4 d3e4 f8b4 c1d2 b4d2 d1d2 b8d7 g1e2 e8g8 e1g1 d8b6 e4f3 f8d8 a1c1 d7f6 f1d1 c8d7 c1c5 a8c8 d1c1 d7c6 f3c6 b7c6 d2c2 f6d5 b2b3 d5e7 c2d2 h7h6 c1c4 d8d7 h2h3 b6b8 d2e3 c8d8 e2c3 b8b6 c3a4');
INSERT INTO evaluations VALUES(46381,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46381,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(46381,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(46381,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(46381,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(46381,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(46381,36,'cp -12','f1d3 f6e4 d3e4 f8b4 c1d2 b4d2 d1d2 b8d7 g1e2 d7f6 e4f3 e8g8 e1g1 h7h6 a1c1 a8b8 f1d1 b7b6 e2c3 c8b7 d4d5 b7d5 c3d5 e6d5 f3d5 f6d5 d2d5 d8f6 b2b3 f8e8 d5f3 f6f3 g2f3 e8e2 d1d7 e2a2 c1c7 b8f8 c7a7 a2b2 g1g2 b2b3');
INSERT INTO evaluations VALUES(11126,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(11126,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(11126,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(11126,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(11126,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(11126,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(11126,36,'cp 77','d4c6 d7c6 d1d8 e8d8 c1f4 d8e8 a2a4 f6d7 a4a5 e7e5 f4d2 a8b8 f2f3 f7f6 h2h4 h7h5 d2e3 f8c5 e1f2 e8e7 f1e2 c5e3 f2e3 d7c5 b2b4 c5e6 c3a4 e6d4 a1c1 c8e6 a4c5 e6f7 g2g3 d4e6 c5d3 e6d4 h1d1 h8d8 d3c5 g7g5 h4g5');
INSERT INTO evaluations VALUES(11118,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(11118,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(11118,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(11118,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(11118,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(11118,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(11118,36,'cp 62','d7c6 d1d8 e8d8 c1f4 d8e8 a2a4 f6d7 a4a5 e7e5 f4d2 a8b8 f2f3 f7f6 d2e3 f8b4 e1f2 h7h5 h2h4 d7f8 f1e2 c8e6 c3b1 f8d7 h1d1 b4c5 e3c5 d7c5 b1c3 e8e7 b2b4 c5d7 c3a4 g7g5 d1h1 g5g4 f3g4 h5g4 a1d1 f6f5 e4f5 e6f5 e2d3');
INSERT INTO evaluations VALUES(25931,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(25931,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(25931,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(25931,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25931,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(25931,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25931,36,'cp -13','b5c6 b7c6 g1f3 c5d4 c3d4 c6c5 e1g1 c5c4 b2b3 a8c8 c1d2 h7h6 h2h3 g8e7 d2b4 e7g6 b4f8 g6f8 d1d2 f8g6 b1c3 e8g8 f1c1 c4b3 a2b3 d8b6 c3e2 a7a6 c1b1 d7b5 e2c3 c8c7 c3a4 b5a4 b3a4 b6a7');
INSERT INTO evaluations VALUES(25900,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(25900,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(25900,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(25900,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(25900,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25900,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(25900,36,'cp -11','b7c6 g1f3 c5d4 c3d4 c6c5 e1g1 c5d4 b1d2 g8e7 d2b3 h7h6 g2g3 e7c6 b3d4 f8c5 d4c6 d7c6 f3d4 c5d4 d1d4 e8g8 c1e3 d8d7 f1c1 f8c8 b2b3 a7a5 c1c3 c6b5 a1c1 c8c3 d4c3 a5a4 b3b4 b5e2 a2a3 d7b5 c3c6 b5c6 c1c6');
INSERT INTO evaluations VALUES(25819,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(25819,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25819,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(25819,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25819,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25819,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25819,36,'cp 0','g1f3 c5d4 c3d4 c6c5 e1g1 c5c4 b2b3 a8c8 c1d2 h7h6 d1e2 g8e7 f1c1 e7g6 b3c4 d5c4 c1c4 d7c6 b1c3 f8e7 d2e3 c6f3 c4c8 d8c8 e2f3 c8c3 a1c1 c3a5 c1c8 e7d8 f3c6 e8f8 c6d6 f8e8 d6c6');
INSERT INTO evaluations VALUES(44923,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(44923,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44923,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(44923,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(44923,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(44923,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44923,36,'cp 7','e2e3 c8f5 f1d3 e7e6 d3f5 e6f5 h2h3 b8d7 g1e2 g7g6 e1g1 f8e7 e2c1 h7h5 c1d3 h5h4 c3e2 e8g8 b2b3 f6h5 c2c4 h5f4 e2f4 d7f6 c4d5 f6d5 f4d5 d8d5 a1c1 f8d8 d1e2 a7a5 d3c5 b7b6 c5d3 e7d6');
INSERT INTO evaluations VALUES(44921,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(44921,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(44921,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44921,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44921,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44921,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44921,36,'cp 0','c8f5 f1d3 e7e6 d3f5 e6f5 c3e2 b8d7 g1f3 f6h5 f4e5 h5f6 e5f4');
INSERT INTO evaluations VALUES(28975,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(28975,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28975,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(28975,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(28975,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(28975,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28975,36,'cp 6','f1d3 e7e6 g1f3 f5g6 e1g1 b8d7 f1e1 d8a5 h2h4 f8e7 d1e2 f6e4 d3e4 d5e4 f3d2 f7f5 d2c4 a5d8 c4d6 e7d6 f4d6 d7b6 d6g3 e8g8 b2b3 d8e7 e2d2 g6h5 c3e2 h5e2 d2e2 b6d5 c2c4 d5b4 e1d1 c6c5 f2f3 b4d3 f3e4 f5e4 d4c5 e7c5');
INSERT INTO evaluations VALUES(28979,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(28979,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(28979,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(28979,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(28979,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(28979,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(28979,36,'cp 9','e7e6 g1f3 b8d7 e1g1 f5g6 f1e1 f8b4 a2a3 b4a5 b2b4 a5c7 d3g6 h7g6 f4g5 d8b8 g2g3 a7a5 e3e4 f6e4 c3e4 d5e4 e1e4 e8g8 e4e1 a5b4 a3b4 a8a1 d1a1 c7d6 c2c3 b8c7 a1b2 e6e5 h2h4 e5d4 f3d4 d7f6 g5f6 g7f6');
INSERT INTO evaluations VALUES(56207,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(56207,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(56207,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(56207,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(56207,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(56207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56207,36,'cp 0','c7c6 c1g5 f8e7 g5e7 d8e7 e2e3 c8f5 b1c3 a7a5 f1d3 f5d3 d1d3 g8f6 e1g1 e8g8 d3c2 b8d7 a1c1 f8d8 f1e1 d8e8 a2a3 e7d8 c2b3 a8b8 b3c2');
INSERT INTO evaluations VALUES(53725,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53725,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53725,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53725,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53725,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53725,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53725,36,'cp 0','b1c3 c8f5 c1f4 b8d7 e2e3 g8f6 h2h3 d8b6 f1d3 b6b2 d3f5 b2c3 e1f1 c3a3 a1b1 a3a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(53716,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53716,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53716,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53716,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53716,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53716,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53716,36,'cp 0','c8f5 c1f4 b8d7 e2e3 g8f6 h2h3 d8b6 f1d3 b6b2 d3f5 b2c3 e1f1 c3a3 a1b1 a3a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(31577,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31577,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31577,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31577,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31577,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31577,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31577,36,'cp 0','c1f4 b8d7 e2e3 g8f6 h2h3 d8b6 f1d3 b6b2 d3f5 b2c3 e1f1 c3a3 a1b1 a3a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(14215,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14215,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(14215,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(14215,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(14215,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(14215,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(14215,36,'cp 15','c8f5 f1b5 e7e6 g1f3 f8d6 f3e5 e8g8 b5c6 b7c6 c3a4 d6e5 f4e5 f6d7 e1g1 d7e5 d4e5 f7f6 d1d4 f6e5 d4e5 d8d6 e5c3 e6e5 a4c5 e5e4 f2f3 e4f3 f1f3 d6g6 f3g3 g6f6 c3f6 f8f6 a1f1 a7a5 f1f2 a8e8 b2b3');
INSERT INTO evaluations VALUES(24808,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24808,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(24808,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(24808,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24808,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(24808,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(24808,36,'cp 17','g1f3 e7e6 f1b5 f8d6 f3e5 e8g8 b5c6 b7c6 e5c6 d8e8 c6e5 a8b8 a1b1 d6b4 e1g1 b4c3 b2c3 e8a4 b1b4 b8b4 c3b4 a4b4 g2g4 f5g6 f2f3 b4a4 d1b1 f6d7 b1b3 g6c2 b3a4 c2a4 e5d7 a4d7 f1b1 f7f6 b1b7 e6e5 f4g3 e5d4 e3d4 a7a6 b7c7 d7b5 h2h4 h7h6 h4h5');
INSERT INTO evaluations VALUES(24809,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24809,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(24809,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24809,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(24809,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(24809,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(24809,36,'cp 14','e7e6 f1b5 f6d7 b5d3 f5d3 d1d3 f8d6 f4g3 d8e7 e3e4 d5e4 d3e4 d7f6 e4e3 d6g3 h2g3 e8c8 e1c1 f6d5 c3d5 d8d5 c2c3 h7h6 h1h4 h8d8 d1e1 a7a6 e3e2 c8b8 h4e4 e7d7 c1b1 b8a8 e1d1 a8b8 a2a3 d5d6 g3g4 d6d5 f3d2 c6a5');
INSERT INTO evaluations VALUES(25148,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(25148,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(25148,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(25148,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(25148,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(25148,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(25148,36,'cp 18','f1b5 f6d7 b5d3 f5d3 d1d3 f8d6 f4g3 d8e7 e3e4 d5e4 d3e4 d7f6 e4e3 e8c8 e1c1 f6d5 c3d5 e6d5 d1e1 h7h6 e3e7 d6e7 f3e5 c6e5 g3e5 e7f8 h2h4 h6h5 e5f4 c8d7 e1e5 c7c6 h1e1 f8d6 e5f5 d6f4 f5f4 f7f6 c1d2 d8e8 e1e8');
INSERT INTO evaluations VALUES(25126,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(25126,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(25126,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(25126,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(25126,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(25126,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(25126,36,'cp 12','f6d7 b5d3 f5d3 d1d3 f8d6 f4g3 d8e7 e3e4 d5e4 c3e4 d6g3 h2g3 f7f6 e1c1 e8c8 a2a3 c8b8 h1e1 h8e8 c1b1 h7h5 d3c3 a7a6 d1d3 d7b6 e4c5 e7d6 e1h1 d6d5 b2b3 g7g6 b1a1 d5d6 a1b2');
INSERT INTO evaluations VALUES(21117,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(21117,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(21117,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(21117,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(21117,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(21117,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(21117,36,'cp 40','e7e5 d4d5 c6e7 f1c4 a7a6 g1f3 e7g6 e3e4 h7h6 e1g1 f8d6 a2a4 f7f5 f1e1 f5f4 c3e2 d8f6 c1d2 g8e7 e2c1 c8g4 c4e2 e8g8 c1b3 g6h4 f3h4 g4e2 d1e2 f6h4 f2f3 h4g5 g1h1 c7c6 d5c6 e7c6 d2c3 g5h5');
INSERT INTO evaluations VALUES(21790,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(21790,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(21790,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(21790,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(21790,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(21790,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(21790,36,'cp 62','d4d5 c6e7 f1c4 a7a6 g1f3 e7g6 a2a4 g8f6 a4a5 f8d6 f3d2 d8e7 d2e4 f6e4 c3e4 e8g8 e1g1 c8f5 e4d6 c7d6 f2f3 f8c8 b2b3 f5d7 d1d2 c8c7 c1a3 d7b5 c4b5 a6b5 a3b4 e7d7 a1c1 a8c8 h2h3 g6e7 c1c7 d7c7');
INSERT INTO evaluations VALUES(21569,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(21569,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(21569,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(21569,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(21569,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(21569,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21569,36,'cp 32','c6e7 f1c4 a7a6 a2a4 e7g6 g1f3 f8d6 e3e4 h7h6 e1g1 f7f5 f1e1 f5f4 b2b4 g8f6 b4b5 c8g4 c3b1 g6h4 b1d2 e8g8 b5a6 b7a6 c4e2 g4f3 e2f3 d6b4 f3g4 f6g4 d1g4 c7c6 e1d1 c6d5 e4d5 h4f5 c1b2 d8e7 d2e4 f5d6 g4g6 d6e4');
INSERT INTO evaluations VALUES(26965,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(26965,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(26965,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(26965,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(26965,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(26965,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(26965,36,'cp 74','c2c3 d8d7 h2h3 g4h5 a2a4 f8e7 a4a5 a7a6 b1d2 g8f6 d2f1 d6d5 e4d5 f6d5 f3e5 c6e5 d1h5 e5c4 d3c4 d5f6 h5f3 e8c8 f1e3 h8e8 e1g1 e7f8 e3c2 f6e4 c1e3 d7e6 c2d4 e6c4 f3f5 c8b8 f5h7 g7g6 d4f3 f8c5 e3c5 e4c5 f1e1 c5d3 e1e8 d8e8 b2b4 c4d5 h7g7 b8a8');
INSERT INTO evaluations VALUES(26962,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(26962,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(26962,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(26962,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(26962,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(26962,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(26962,36,'cp 63','d8d7 h2h3 g4h5 b1d2 g8f6 a2a4 f8e7 a4a5 a7a6 d2f1 d6d5 e4d5 f6d5 f3e5 c6e5 d1h5 e5c4 d3c4 d5f6 h5f3 e8c8 f1e3 h8e8 e1g1 d7d3 f3f5 d3d7 f5d7 d8d7 f1e1 e7c5 g1f1 c5a7 e3f5 e8e1 f1e1 f6e4 c1e3 a7e3 f5e3 d7d2');
INSERT INTO evaluations VALUES(29537,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(29537,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(29537,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(29537,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29537,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29537,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(29537,36,'cp 4','f5d3 c2d3 e7e6 g1f3 f8d6 d1b3 b7b6 f4d6 d8d6 b1c3 e8g8 e1g1 c7c5 a1c1 f8c8 e3e4 d6d8 e4d5 f6d5 c3d5 d8d5 b3d5 e6d5 d4c5 b6c5 c1c3 a7a5 f1c1 b8a6 b2b3 c8b8 d3d4 a6b4 c1a1 c5d4 f3d4 b4a2');
INSERT INTO evaluations VALUES(29640,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(29640,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(29640,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(29640,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29640,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(29640,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29640,36,'cp 9','c2d3 e7e6 b1c3 f8d6 d1b3 b7b6 f4d6 d8d6 b3a4 c7c6 g1f3 e8g8 e1g1 f8c8 f1c1 a7a5 a2a3 h7h6 h2h3 d6d8 a4d1 c6c5 c1c2 c5d4 f3d4 b8c6 d4c6 c8c6 a1c1 a8c8 c3e2 d8d7 c2c6 c8c6 c1c6 d7c6 e2d4 c6b7 b2b3');
INSERT INTO evaluations VALUES(29638,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29638,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29638,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(29638,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(29638,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29638,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(29638,36,'cp 19','e7e6 b1c3 f8d6 f4d6 c7d6 b2b4 d8d7 g1e2 b8c6 b4b5 c6e7 e1g1 e8g8 d1b3 f8c8 f1b1 d7d8 b1c1 a7a6 h2h3 a6a5 b3a3 h7h6 c3a4 a8b8 c1c8 e7c8 a1b1 f6d7 e2c3 d7b6 a4b6 d8b6');
INSERT INTO evaluations VALUES(32091,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(32091,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(32091,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(32091,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(32091,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(32091,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(32091,36,'cp 50','g8f6 h2h3 g4e6 c4e6 f7e6 c2c3 f8e7 e1g1 e8g8 f1e1 b8c6 b2b4 a7a6 a2a4 h7h6 b1d2 d8e8 d2f1 e7d8 d1e2 c6e7 d3d4 e5d4 c3d4 e7g6 f1g3 e8f7 a1a3 c7c6 e2d3 d8c7 c1d2 a8c8 d2e3 f6d7 b4b5 c7a5 b5c6 b7c6');
INSERT INTO evaluations VALUES(29904,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(29904,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(29904,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(29904,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(29904,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(29904,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(29904,36,'cp 51','h2h3 g4e6 c4e6 f7e6 c2c3 f8e7 e1g1 e8g8 c1e3 c7c5 a2a3 a7a5 b1d2 b8c6 d3d4 e5d4 c3d4 c5d4 f3d4 c6d4 e3d4 a8c8 d1e2 d8d7 d4c3 a5a4 c3b4 d6d5 e4e5 f6e8 e2d1 f8f4 b4e7 d7e7 d2f3 f4e4 a1c1 c8c1 d1c1 e7d7 c1c3');
INSERT INTO evaluations VALUES(29902,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(29902,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(29902,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(29902,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(29902,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(29902,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(29902,36,'cp 45','g4e6 c4e6 f7e6 c2c3 f8e7 e1g1 e8g8 f1e1 b8c6 b2b4 a7a6 a2a4 h7h6 b1d2 d8e8 d2f1 e8g6 c1e3 f6d7 f1h2 c6d8 d1b3 c7c6 d3d4 d8f7 d4e5 d7e5 f3e5 f7e5 f2f4 e5d3 f4f5 g6f7 e1d1 d3e5 f5e6 f7g6');
INSERT INTO evaluations VALUES(57137,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(57137,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(57137,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(57137,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(57137,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(57137,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(57137,36,'cp 12','e5d4 d1d4 b8c6 d4e3 g8e7 c3d5 d7d6 g1f3 c6e5 d5c3 e7g6 h2h4 h7h5 f3e5 d6e5 g2g3 c7c6 c1d2 f8e7 e1c1 d8c7 e3e4 e8g8 f1h3 c8h3 h1h3 c7c8 h3h1 b7b5 d2g5 f7f5 e4f3 c8e6 f3h5 b5b4 g5e7 g6e7');
INSERT INTO evaluations VALUES(57303,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(57303,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(57303,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(57303,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(57303,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(57303,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(57303,36,'cp 39','d1d4 b8c6 d4d3 g8f6 c1g5 h7h6 g5f6 d8f6 e1c1 f6f2 e2e3 f2h4 g1f3 h4g4 c1b1 f8e7 c3d5 e7d6 f1e2 c6e7 f3d4 g4g5 d4b5 e7d5 b5d6 c7d6 d3d5 g5d5 d1d5 b7b6 d5d6 c8b7 e2f3 b7f3 g2f3 e8c8 b1c1 d8e8 h1d1 e8e7 e3e4 h8e8 b2b3');
INSERT INTO evaluations VALUES(57297,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(57297,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(57297,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(57297,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(57297,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(57297,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(57297,36,'cp 25','b8c6 d4d3 f8c5 c1e3 c6b4 d3d2 c5e3 d2e3 d8e7 e3e7 g8e7 e1c1 a7a6 e2e4 d7d6 a2a3 b4c6 f2f4 c8d7 g1f3 h7h6 d1e1 e8c8 f1d3 d7g4 h1f1 g7g5 e4e5 e7g6 d3g6 f7g6 e5d6 g5f4 h2h3 g4f3 d6c7 c8c7 g2f3 g6g5 e1e6 h8g8 e6h6');
INSERT INTO evaluations VALUES(22929,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(22929,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(22929,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(22929,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(22929,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(22929,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(22929,36,'cp 25','d4d3 f8c5 c1e3 c6b4 d3d2 c5e3 d2e3 d8e7 e3e7 g8e7 e1c1 a7a6 a2a3 b4c6 e2e4 d7d6 f2f4 h7h6 f1d3 c8d7 g1f3 e8c8 d1e1 d7g4 e1e3 f7f6 g2g3 g7g5 c3d5 h8f8 h1e1 e7g6 b2b4 h6h5 f4g5 g4f3 e3f3 f6g5 f3f5 f8f5 e4f5');
INSERT INTO evaluations VALUES(19691,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(19691,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(19691,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(19691,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(19691,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(19691,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(19691,36,'cp 81','c1e3 g8f6 f2f3 f8g7 d1d2 e8g8 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4g5 f6d7 d4g7 g8g7 c3d5 a5d2 e1d2 d7b6 d5e7 c8e6 a2a4 a8e8 a4a5 b6c4 f1c4 e6c4 e7d5 c4d5 e4d5 e8e5 c2c4 f7f6 g5f6 f8f6 h1e1 e5e1 a1e1 f6f3 e1e7 f3f7 e7e8 f7f4 e8b8 b7b6 a5a6 f4h4 b8b7 g7f6 b7a7 h4c4 a7a8 f6g7');
INSERT INTO evaluations VALUES(19690,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(19690,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(19690,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(19690,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(19690,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(19690,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(19690,36,'cp 64','g8f6 f2f3 h7h5 f1c4 c8d7 d1d2 f8g7 e1c1 a7a6 h1e1 e8g8 e3h6 c6d4 h6g7 g8g7 d2d4 a8c8 e4e5 d6e5 e1e5 f8e8 c4b3 d8c7 c3d5 f6d5 e5e7 g7g8 e7e8 c8e8 d4d5 d7e6 d5d6 c7d6 d1d6 e6b3 c2b3 e8e2 d6d2 e2e1 c1c2 e1h1 h2h3 g8g7 c2c3 g7f6');
INSERT INTO evaluations VALUES(42391,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(42391,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(42391,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(42391,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(42391,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(42391,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(42391,36,'cp 91','f3e5 d6e5 d2d4 e5d6 e4e5 d6b4 e5f6 d8f6 f1d3 d7d5 e1g1 c7c6 f1e1 b4e7 e1e5 h7h6 d1e1 b8a6 a2a3 a6c7 f2f4 g7g6 c3e2 c8d7 e2g3 e8d8 f4f5 e7d6 c2c3 d6e5 d4e5 h8e8 c1h6 f6e5 e1e5 e8e5 h6f4 d7f5 g3f5 e5f5 d3f5 g6f5 a1f1 c7e6 f4g3');
INSERT INTO evaluations VALUES(56975,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(56975,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(56975,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(56975,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(56975,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(56975,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(56975,36,'cp 62','g1f3 b8c6 d2d4 e5d4 e1g1 g8f6 e4e5 d7d5 c4b5 f6d7 b5c6 b7c6 f3d4 d7b6 d4c6 d8d7 c6d4 c7c5 e5e6 d7b7 e6f7 e8f7 d4e2 f8e7 a2a4 h8d8 a4a5 b6d7 b1d2 d7f6 d2f3 f7g8 f1e1 e7d6 c1f4 c8g4 d1d3 g4f3 d3f3 b7b2 f3d3 c5c4 d3f5 b2b7 e2d4 d6f4 f5f4 f6e4 f2f3 e4c5 f4g3 b7f7 d4c6');
INSERT INTO evaluations VALUES(56973,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(56973,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(56973,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(56973,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(56973,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(56973,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(56973,36,'cp 55','b8c6 d2d4 e5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 e4e5 d7d5 e5f6 d8f6 c4d3 d5d4 d1e2 e8f8 e1g1 d4c3 b2c3 b4d6 a1b1 c8f5 f1e1 f5d3 e2d3 f8g8 e1e4 f6g6 c1e3 g8h7 e4c4 h8b8 d3g6 h7g6 b1b3 b8b5 c4c6 b5a5 a2a3 f7f6 g2g3 a5a3 b3a3 d6a3');
INSERT INTO evaluations VALUES(4303,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(4303,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(4303,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(4303,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(4303,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(4303,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(4303,36,'cp -63','d7d5 e4d5 c8g4 g5f3 f6d5 b1c3 d5c3 b2c3 g8h8 a1b1 f7f5 d3d4 c5d6 c4e2 g4f3 e2f3 e5e4 f3e2 b7b6 d1e1 d8e7 f2f3 e7e6 e2b5 c6e7 c1d2 e6g6 f3e4 f5e4 e1h4 c7c6 f1f8 a8f8 b5e2 e7d5 b1f1 f8e8 h4g4');
INSERT INTO evaluations VALUES(3718,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(3718,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(3718,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(3718,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(3718,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(3718,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(3718,36,'cp -63','e4d5 c8g4 g5f3 f6d5 b1c3 d5c3 b2c3 f8e8 f1e1 g4h5 h2h3 h7h6 c1e3 c5a3 e3c1 a3f8 c4b3 d8f6 g2g4 h5g6 g4g5 h6g5 f3g5 f8c5 d1f3 f6f3 g5f3 g8f8 b3d5 f7f6 c1e3 c5e3 f2e3 e5e4 f3h4 g6h7 e1b1 a8d8 c3c4 e4d3 c2d3 e8e3 b1b7 e3h3 d5c6 h3h4');
INSERT INTO evaluations VALUES(3710,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(3710,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(3710,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(3710,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(3710,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(3710,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(3710,36,'cp -76','c8g4 g5f3 f6d5 b1c3 d5c3 b2c3 f8e8 f1e1 h7h6 c1e3 c5a3 e3c1 a3f8 h2h3 g4h5 c4b3 d8f6 b3d5 a8d8 d5e4 h5g6 e4g6 f6g6 f3h4 g6f6 d1g4 g7g6 a1b1 b7b6 h4f3 f8g7 a2a4 f6e6 g4e6 e8e6 g1f1 f7f5 g2g4 e5e4 g4f5');
INSERT INTO evaluations VALUES(19428,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(19428,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(19428,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(19428,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(19428,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(19428,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(19428,36,'cp 208','d4d5 c6b8 b1c3 d7d6 f2f4 g7g6 a2a4 f8g7 f1b5 c8d7 b5c4 a7a6 g1f3 d7g4 e1g1 b8d7 h2h3 g4f3 d1f3 g8f6 c1d2 e8g8 g1h1 d7b6 c4d3 c5c4 d3e2 a8c8 f3f2 b6d7 a4a5 d7c5 f2e3 b7b5 a5b6 d8b6 e2c4');
INSERT INTO evaluations VALUES(19511,30,'cp 220',NULL);
INSERT INTO evaluations VALUES(19511,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(19511,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(19511,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(19511,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(19511,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(19511,36,'cp 177','c6a5 g1f3 a7a6 a2a3 e7e6 d5d6 a5c6 b1c3 b7b5 c1g5 f7f6 g5e3 c8b7 e3c5 g7g6 d1d2 g8h6 h2h4 h6f7 e1c1 f8h6 c5e3 h6f8 h4h5 g6g5 e3c5 a8c8 c1b1 g5g4 f3d4 c6e5 d4b3 c8c6 c5b4 f8d6 b3a5 d6b4 a5c6 b4c3 c6d8 c3d2 d8b7 d2f4 g2g3 f4g5 f1e2 e8g8');
INSERT INTO evaluations VALUES(27532,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(27532,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(27532,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(27532,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(27532,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(27532,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(27532,36,'cp 57','d4c5 f8c5 e1g1 a7a5 d1e2 f7f6 c3c4 d5c4 d3c4 f6e5 b1d2 g8f6 b2b3 e8g8 c1b2 a8d8 f3e5 c6e5 e2e5 b6d6 e5d6 c5d6 d2f3 b7b5 c4e2 f6d5 g2g3 a5a4 b3a4 b5a4 f3e5 a4a3 b2d4 d5e7 f1c1 d7e8 a1b1 e7f5 d4c3 d8c8 e2g4 g8h8 b1b7');
INSERT INTO evaluations VALUES(27517,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(27517,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(27517,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(27517,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(27517,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(27517,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(27517,36,'cp 52','f8c5 e1g1 a7a5 d1e2 f7f6 c3c4 d5c4 d3c4 f6e5 b1d2 g8f6 b2b3 e8g8 c1b2 a8d8 a1d1 g8h8 f3e5 c6e5 b2e5 d7c6 h2h3 c6d5 c4b5 c5d6 e5d6 b6d6 d2c4 d6f4 e2e5 f4e5 c4e5 f6e4 d1c1 d8c8 b5c4 h8g8 f1e1 d5c4 c1c4 c8c4 e5c4 e4f2 c4a5');
INSERT INTO evaluations VALUES(27307,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(27307,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(27307,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(27307,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(27307,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(27307,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(27307,36,'cp 59','e1g1 a7a5 d1e2 f7f6 c3c4 d5c4 d3c4 f6e5 b1d2 g8f6 b2b3 a8d8 c1b2 e8g8 f3e5 c6e5 b2e5 d7c6 c4e6 g8h8 d2c4 b6a6 e6f5 b7b5 c4e3 d8e8 e2b2 a6a7 a1c1 c6a8 b2c3 c5b6 e5d6 a7b7 c3c2 f8g8 f1e1 f6d5 f5e4 d5e3 e4b7 e3c2 e1e8 g8e8 b7a8 c2d4 a8f3 h8g8 d6c5 d4f3 g2f3');
INSERT INTO evaluations VALUES(27306,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(27306,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(27306,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(27306,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(27306,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(27306,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(27306,36,'cp 45','a7a5 d1e2 f7f6 c3c4 d5c4 d3c4 f6e5 b1d2 g8f6 b2b3 a8d8 c1b2 e8g8 f3e5 c6e5 e2e5 b6d6 e5d6 c5d6 d2f3 b7b5 c4e2 f6d5 f3e5 d7e8 f1d1 d5f4 e2f3 f4g6 e5d3 e8f7 a1c1 a5a4 d3c5 d6c5 c1c5 a4b3 a2b3 d8d1 f3d1 f8d8 d1g4 g6f4');
INSERT INTO evaluations VALUES(55880,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(55880,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(55880,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(55880,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(55880,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(55880,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(55880,36,'cp -82','d5e4 c1e3 g8h6 d1d2 c7c5 e3h6 d8h4 g2g3 h4h6 d2h6 g7h6 d4c5 f8g7 c2c3 e4f3 g1f3 b8d7 c5c6 b7c6 b1d2 e8e7 e1c1 c8b7 f1g2 c6c5 f3h4 b7g2 h4g2 d7e5 c1c2 h6h5 g2e3 h5h4 g3h4 h8d8 e3c4 f7f5 c4a5 a8c8 d2c4 e5c4 a5c4 d8d1');
INSERT INTO evaluations VALUES(56055,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(56055,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(56055,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(56055,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(56055,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(56055,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(56055,36,'cp -88','c1e3 g8h6 d1d2 c7c5 e3h6 d8h4 g2g3 h4h6 d2h6 g7h6 d4c5 e4f3 g1f3 f8c5 b1d2 c8d7 e1c1 d7c6 f1d3 b8d7 d3e4 a8c8 c2c3 e8e7 e4c6 c8c6 a2a4 h8c8 c1c2 c6d6 d2c4 d6d1 h1d1 f7f6 c4a5 b7b6 a5b3 h6h5 f3d4 d7e5');
INSERT INTO evaluations VALUES(56053,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(56053,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(56053,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(56053,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(56053,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(56053,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(56053,36,'cp -83','g8h6 d1d2 c7c5 e3h6 d8h4 g2g3 h4h6 d2h6 g7h6 d4c5 e4f3 g1f3 c8d7 b1d2 f8c5 e1c1 d7c6 f1d3 b8d7 d3e4 a8c8 c2c3 h6h5 e4c6 c8c6 d2e4 c6c7 c1c2 c5e7 h1e1 h8g8 a2a4 g8g4 d1d4 d7f6 f3d2 h7h6 a4a5 f6e4 d4e4 g4g5 b2b4 g5f5 e1e2 e7g5 d2c4 e8e7 c4e5');
INSERT INTO evaluations VALUES(23425,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(23425,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(23425,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(23425,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(23425,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(23425,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(23425,36,'cp 21','e7e5 g2g3 f8c5 g1f3 g8f6 f1g2 d7d6 e1g1 e8g8 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 a4b6 a7b6 f1e1 h7h6 h2h4 c7c6 e3e4 c8e6 a2a3 d8d7 f3g2 b6b5 c4b5 c6b5 d3d4 e6c4 d4d5 b5b4 c1d2 b4a3 a1a3 a8a3 b2a3');
INSERT INTO evaluations VALUES(49932,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(49932,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(49932,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49932,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(49932,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(49932,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(49932,36,'cp 37','f8b4 c1d2 b7b6 e2e3 c8b7 f1e2 b4c3 d2c3 f6e4 a1c1 e8g8 e1g1 c7c5 c3e1 b8c6 d4d5 c6e7 d5e6 d7e6 f3e5 d8c7 e5d3 e6e5 b2b4 e7g6 d1b3 c7e7 b4c5 e4c5 d3c5 b6c5 f2f3 a8c8 e1c3 f8d8 c1d1 b7c6 d1d8 c8d8 b3b2');
INSERT INTO evaluations VALUES(42913,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(42913,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(42913,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(42913,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(42913,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42913,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(42913,36,'cp 51','c1d2 b7b6 e2e3 e8g8 f1d3 c7c5 e1g1 c8b7 a2a3 b4c3 d2c3 f6e4 c3e1 d7d6 f3d2 e4d2 d1d2 b8c6 d4c5 d6c5 f2f3 c6e5 d3e2 d8e7 b2b4 a8d8 d2c3 e5g6 e1f2 h7h6 b4c5 b6c5 a3a4 e6e5 a4a5 d8d6 f1d1');
INSERT INTO evaluations VALUES(42912,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(42912,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(42912,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(42912,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(42912,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(42912,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(42912,36,'cp 40','b7b6 e2e3 b4c3 d2c3 f6e4 f1d3 e4c3 b2c3 c8b7 e1g1 e8g8 f1e1 d8f6 e3e4 f5e4 d3e4 b7e4 e1e4 b8c6 d1e2 c6e7 f3e5 a8d8 a1d1 e7g6 e5d3 f6f5 g2g3 f5a5 h2h4 a5c3 d4d5 e6d5 c4d5 c3f6 g1g2 c7c6 d5c6 d7c6');
INSERT INTO evaluations VALUES(18765,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(18765,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(18765,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(18765,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(18765,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18765,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18765,36,'cp 2','g2g3 g8e7 d2d3 g7g6 g1e2 f8g7 f1g2 e8g8 a1b1 b7b6 e1g1 d7d6 a2a3 c8b7 b2b4 c6d4 b4b5 d4e2 c3e2 a7a6 a3a4 a6b5 a4b5 a8a2 h2h4 d8a8 h4h5 g6h5 e2f4 h5h4 f4h5 g7d4 c1g5 f7f6 g5e3 d4e3 f2e3 f6f5 g3h4 g8h8');
INSERT INTO evaluations VALUES(18764,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(18764,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(18764,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18764,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18764,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18764,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18764,36,'cp 0','g8e7 d2d3 c6d4 f1g2 e7c6 g1e2 f8e7 e1g1 d7d6 f2f4 e8g8 e2d4 c5d4 c3b1 c8d7 b1a3 c6b4 d1e2 a7a5 c1d2 d8b6 h2h4 b4a6 a1b1 a6b4');
INSERT INTO evaluations VALUES(13152,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13152,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(13152,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(13152,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(13152,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13152,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(13152,36,'cp 1','h2h3 e7e6 f1d3 c6b4 e1g1 b4d3 d1d3 f8e7 f1e1 e8g8 c1g5 b7b6 c3e2 f6e4 g5e7 d8e7 d3e3 c8b7 e2f4 a8e8 c2c3 f7f6 f3d2 e7f7 d2e4 d5e4 c3c4 e6e5 d4e5 f6e5 f4d5 b7d5 c4d5 e8d8 e1e2 f7d5 a1c1 d5a2 e3e4 a2a5 b2b3 d8d2');
INSERT INTO evaluations VALUES(13149,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(13149,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(13149,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(13149,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(13149,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(13149,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(13149,36,'cp 0','e7e6 f1d3 c6b4 e1g1 b4d3 d1d3 f8e7 c1g5 e8g8 c3e2 b7b6 f1e1 f6e4 g5e7 d8e7 c2c3 c8b7 f3d2 e7g5 d2f3');
INSERT INTO evaluations VALUES(16750,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(16750,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(16750,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16750,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(16750,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(16750,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(16750,36,'cp 9','e4d6 b5c6 d7c6 d4e5 d6f5 d1d8 e8d8 f1d1 d8e8 b1c3 c8e6 h2h3 f8e7 c1g5 h7h5 d1d3 a8c8 a1d1 f7f6 e5f6 g7f6 g5f4 e8f7 b2b3 h8e8 g1f1 b7b6 c3e2 c6c5 c2c4 e8d8 e2g3 f5g3 f2g3 d8d3 d1d3 e6f5 d3d2 c7c6 f1f2 c8e8');
INSERT INTO evaluations VALUES(16875,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(16875,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(16875,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(16875,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(16875,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(16875,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(16875,36,'cp 14','b5c6 d7c6 d4e5 d6f5 d1d8 e8d8 h2h3 f8e7 b1c3 h7h5 c1g5 c8e6 a1d1 d8e8 b2b3 h5h4 f1e1 a8d8 d1d8 e8d8 e1d1 d8e8 a2a4 a7a6 g5c1 b7b6 c1b2 a6a5 g1f1 e6c8 c3e2 h8h6 c2c4 h6e6 e2d4 e6g6 b2c3 f5d4 f3d4');
INSERT INTO evaluations VALUES(16231,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(16231,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(16231,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(16231,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(16231,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(16231,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(16231,36,'cp 20','d7c6 d4e5 d6f5 d1d8 e8d8 f1d1 d8e8 b1c3 f8e7 c1g5 e7g5 f3g5 h7h6 g5e4 e8e7 f2f4 f5e3 d1d2 h8d8 d2d8 e7d8 a1c1 d8e7 g1f2 e3d5 f2f3 h6h5 h2h3 d5c3 e4c3 h5h4 c1d1 c8f5 d1d2 a8d8 d2d8 e7d8 c3e4 b7b6 a2a3 c6c5 c2c4 a7a5 e4c3 c7c6 c3e4 d8d7 e4g5 d7e7');
INSERT INTO evaluations VALUES(15720,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(15720,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(15720,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(15720,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(15720,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(15720,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(15720,36,'cp 25','d4e5 d6f5 d1d8 e8d8 f1d1 d8e8 b1c3 a7a5 a2a4 c8e6 f3g5 e6c8 c3e2 f8e7 g5f3 b7b6 h2h3 c8b7 e2d4 f5d4 f3d4 h8g8 c1f4 a8d8 g2g4 d8d7 d4f5 e7d8 g1g2 c6c5 f2f3 g7g6 d1d7 e8d7 a1d1 d7c8 f5h6 g8f8 g4g5 c5c4 g2g3 d8e7 h6g4 e7b4');
INSERT INTO evaluations VALUES(15718,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(15718,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(15718,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(15718,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(15718,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(15718,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(15718,36,'cp 13','d6f5 d1d8 e8d8 f1d1 d8e8 b1c3 a7a5 a2a4 c8e6 f3g5 e6c8 c3e2 b7b6 c1f4 f8e7 h2h3 c6c5 g5e4 c8b7 e4c3 g7g5 f4h2 h7h5 c3d5 b7d5 d1d5 a8d8 d5d8 e8d8 g2g4 f5h4 a1d1 d8c8 h2g3 h4g6 e2c3 h5h4 g3h2 h8e8 b2b3 c5c4 g1g2 c4b3 c2b3 f7f6 e5f6 e7f6');
INSERT INTO evaluations VALUES(15713,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(15713,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15713,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(15713,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(15713,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(15713,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(15713,36,'cp 20','d1d8 e8d8 f1d1 d8e8 b1c3 c8e6 f3g5 e6c8 g5e4 a7a5 d1d3 c8e6 h2h3 a8d8 d3d8 e8d8 c1g5 d8c8 a1d1 f8b4 g2g4 f5e7 f2f4 b4c3 e4c3 e7d5 c3d5 e6d5 g1h2 h7h5 h2g3 h5g4 h3g4 h8h1 d1h1 d5h1 f4f5 h1e4 c2c4 c6c5 g5e7 e4d3 e7c5 d3c4 a2a3 b7b6 c5e7 c4b3 g3f4 c7c5 e7f8 g7g6 f4g5 g6f5');
INSERT INTO evaluations VALUES(2253,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2253,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(2253,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2253,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(2253,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(2253,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(2253,36,'cp 8','e8d8 f1d1 d8e8 b1c3 f8e7 c1g5 e7g5 f3g5 h7h6 g5e4 e8e7 f2f4 h6h5 g1f2 b7b6 e4g5 c8b7 d1d2 a8d8 a1d1 d8d2 d1d2 h5h4 g5f3 a7a5 a2a3 c6c5 c3d5 b7d5 d2d5 a5a4 g2g4 h4g3 h2g3 h8h1 g3g4 f5d4 f3d4 c5d4 d5d4 h1h2 f2g3 h2c2 d4a4 c2b2 f4f5 b2c2');
INSERT INTO evaluations VALUES(2287,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2287,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(2287,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2287,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(2287,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(2287,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(2287,36,'cp 26','f1d1 d8e8 b1c3 f8e7 g2g4 f5h4 f3h4 e7h4 h2h3 h7h5 g4g5 f7f6 e5f6 g7f6 c3e4 f6g5 c1g5 h4g5 e4g5 e8e7 d1d3 c8f5 a1e1 e7f6 g5e4 f6g6 d3g3 g6h6 g3e3 h8f8 g1g2 a8d8 g2g3 f8g8 g3h2 g8f8 e1e2 b7b6 h2g3 d8d1 e2e1 d1e1 e3e1 f8g8 g3h4 g8g2 e4g3 f5c2 e1e6 h6h7 e6e7 h7h6');
INSERT INTO evaluations VALUES(21064,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(21064,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(21064,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(21064,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(21064,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21064,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21064,36,'cp 0','e7e5 f1b5 g8e7 g1f3 e5e4 f3e5 a7a6 d1h5 c8e6 e5c6 e7c6 b5c6 b7c6 e1g1 h7h6 f2f3 e4f3 h5f3 d8d7 f3h5 e6g4 h5e5 g4e6');
INSERT INTO evaluations VALUES(19045,30,'cp 297',NULL);
INSERT INTO evaluations VALUES(19045,31,'cp 297',NULL);
INSERT INTO evaluations VALUES(19045,32,'cp 276',NULL);
INSERT INTO evaluations VALUES(19045,33,'cp 287',NULL);
INSERT INTO evaluations VALUES(19045,34,'cp 287',NULL);
INSERT INTO evaluations VALUES(19045,35,'cp 277',NULL);
INSERT INTO evaluations VALUES(19045,36,'cp 279','e4d5 c6d4 f3d4 c5d4 d1d4 g8f6 b1c3 f8d6 c1f4 d6f4 d4f4 e8g8 f1e2 d7d6 e1c1 f8e8 h1e1 h7h6 a2a3 e8e5 e2d3 f6d7 f2f3 a7a6 c1b1 e5e1 d1e1 d7e5 c3e4 d8e7 g2g4 e5d3 c2d3 c8d7 f4d6 e7h4 e1c1 d7b5 c1c3 h4d8 d6d8 a8d8 d5d6 b5c6');
INSERT INTO evaluations VALUES(19040,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(19040,31,'cp 292',NULL);
INSERT INTO evaluations VALUES(19040,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(19040,33,'cp 296',NULL);
INSERT INTO evaluations VALUES(19040,34,'cp 294',NULL);
INSERT INTO evaluations VALUES(19040,35,'cp 296',NULL);
INSERT INTO evaluations VALUES(19040,36,'cp 295','c6d4 f3d4 c5d4 d1d4 g8f6 b1c3 f8d6 c1f4 d6f4 d4f4 e8g8 e1c1 d7d6 f1e2 f8e8 h1e1 h7h6 a2a3 a7a6 g2g4 e8e2 c3e2 c8g4 f2f3 g4h5 e2g3 h5g6 g3f5 a8c8 f5e7 d8e7 e1e7 c8c2 c1b1 c2d2 b1a2 d2d1 e7b7 f6d5 f4c4 g6d3 c4c6 g8h7 a2b3 d1d2 b3a4 d3g6 a4a5 d5f4 c6c7 f4d3');
INSERT INTO evaluations VALUES(19174,30,'cp 277',NULL);
INSERT INTO evaluations VALUES(19174,31,'cp 283',NULL);
INSERT INTO evaluations VALUES(19174,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(19174,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(19174,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(19174,35,'cp 292',NULL);
INSERT INTO evaluations VALUES(19174,36,'cp 324','f3d4 c5d4 d1d4 g8f6 b1c3 f8d6 c1g5 e8g8 e1c1 f8e8 g5f6 d8f6 d4f6 g7f6 g2g3 d6c5 f1d3 g8g7 c3e4 c5b4 g3g4 d7d6 h2h3 e8e5 c2c3 b4c5 e4c5 d6c5 c3c4 c8d7 h1e1 a8e8 e1e5 f6e5 c1d2 h7h5 f2f3 h5g4 h3g4 e8h8 d3e4 g7f6');
INSERT INTO evaluations VALUES(19173,30,'cp 278',NULL);
INSERT INTO evaluations VALUES(19173,31,'cp 321',NULL);
INSERT INTO evaluations VALUES(19173,32,'cp 321',NULL);
INSERT INTO evaluations VALUES(19173,33,'cp 322',NULL);
INSERT INTO evaluations VALUES(19173,34,'cp 304',NULL);
INSERT INTO evaluations VALUES(19173,35,'cp 320',NULL);
INSERT INTO evaluations VALUES(19173,36,'cp 326','c5d4 d1d4 g8f6 b1c3 f8d6 c1g5 d8e7 f1e2 d6e5 d4e3 h7h6 g5h4 e5c3 e3c3 e8g8 h4f6 g7f6 a1d1 d7d6 d1d3 e7e5 d3e3 e5c3 e3c3 c8f5 e1d2 f8e8 h1e1 g8g7 e2f3 e8e1 d2e1 a8c8 e1d2 c8c3 d2c3 b7b6 g2g3 g7f8 c3b3 f5h3 c2c4 h3d7 a2a3 f6f5');
INSERT INTO evaluations VALUES(19180,30,'cp 337',NULL);
INSERT INTO evaluations VALUES(19180,31,'cp 322',NULL);
INSERT INTO evaluations VALUES(19180,32,'cp 332',NULL);
INSERT INTO evaluations VALUES(19180,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(19180,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(19180,35,'cp 315',NULL);
INSERT INTO evaluations VALUES(19180,36,'cp 337','d1d4 g8f6 b1c3 d7d6 f1d3 f8e7 e1g1 e8g8 f1e1 f8e8 d4h4 c8d7 h2h3 h7h6 h4g3 f6h5 g3f3 h5f6 c1e3 e7f8 e3d4 a7a6 a2a3 a8c8 a3a4 f6h7 f3g3 h7f6 a4a5 e8e1 a1e1 f6h5 g3f3 d8g5 d4e3 g5h4 e3d2 h5f6 g2g4 c8e8 e1e8 f6e8 b2b3 f8e7');
INSERT INTO evaluations VALUES(19179,30,'cp 337',NULL);
INSERT INTO evaluations VALUES(19179,31,'cp 309',NULL);
INSERT INTO evaluations VALUES(19179,32,'cp 335',NULL);
INSERT INTO evaluations VALUES(19179,33,'cp 342',NULL);
INSERT INTO evaluations VALUES(19179,34,'cp 329',NULL);
INSERT INTO evaluations VALUES(19179,35,'cp 332',NULL);
INSERT INTO evaluations VALUES(19179,36,'cp 327','g8f6 b1c3 f8d6 c1g5 e8g8 e1c1 f8e8 g5f6 d8f6 d4f6 g7f6 g2g3 d6c5 f1d3 d7d6 c3e4 g8g7 e4c5 d6c5 d1e1 c8d7 e1e8 a8e8 c2c4 a7a5 c1d2 b7b6 a2a3 e8a8 d2e3 a8e8 e3f4 h7h5 h1d1 e8e5 f2f3 g7f8 d1d2 a5a4 h2h4 f8e7 d2e2 e5e2');
INSERT INTO evaluations VALUES(52754,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52754,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52754,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52754,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(52754,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(52754,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52754,36,'cp 1','g8f6 e2e3 c8f5 f2f3 e7e6 g2g4 f5g6 h2h4 h7h5 g4g5 f6d7 g1e2 d7b6 f4h2 f8d6 e2f4 e8g8 d1d2 c6c5 f4g6 f7g6 h2d6 d8d6 c3b5 d6c6 d4c5 c6c5 e1c1 b8d7 f3f4 a7a6 b5d4 c5d6 d4f3 a8c8 c1b1 g8h7 c2c3 d7c5');
INSERT INTO evaluations VALUES(55928,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(55928,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(55928,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(55928,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(55928,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(55928,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(55928,36,'cp -17','b1c3 g8f6 d2d3 a7a6 e2e4 d5e4 d3e4 d8d1 e1d1 b7b6 e4e5 f6d7 c3e4 c8b7 f1d3 b8c6 c2c3 e8c8 d1c2 h7h6 h1e1 a6a5 a2a4 c6e7 c1d2 e7d5 d3b5 d7c5 e4c5 f8c5 a1d1 d5e7 b5d3 h8f8 e1e2 b7d5 d3e4 d5c4 e2e1 e7f5');
INSERT INTO evaluations VALUES(10098,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10098,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10098,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(10098,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(10098,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(10098,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10098,36,'cp 17','g1f3 h7h6 c1e3 c5b6 c3d5 d7d6 a2a4 f6d5 c4d5 e8g8 e1g1 d8f6 d1d2 c6e7 a4a5 e7d5 e4d5 b6e3 f2e3 f6e7 d2b4 c7c6 e3e4 c6d5 e4d5 e7c7 c2c4 c8d7 b4c3 f7f6 b2b4 f6f5 c4c5 f5f4');
INSERT INTO evaluations VALUES(12739,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(12739,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(12739,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(12739,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(12739,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(12739,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(12739,36,'cp 69','d4c6 d7c6 d1d8 e8d8 f1c4 d8e8 e4e5 f6d7 e5e6 f7e6 c4e6 d7c5 e6c8 a8c8 c1e3 c5e6 e1c1 f8g7 c3e4 b7b6 d1e1 c8d8 c2c3 d8d5 g2g3 e8f7 c1c2 g7f6 e1e2 h8d8 h1e1 d5e5 e4d2 e5d5 a2a4 h7h5 f2f3 d8d7 d2e4 d5e5 b2b4 e5f5 e4f6 f5f6 f3f4');
INSERT INTO evaluations VALUES(12709,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(12709,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(12709,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(12709,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(12709,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(12709,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(12709,36,'cp 64','d7c6 d1d8 e8d8 f1c4 d8e8 e4e5 f6d7 e5e6 f7e6 c4e6 f8g7 c1e3 d7f8 e6c8 a8c8 c3e4 b7b6 e1c1 f8e6 d1e1 e8f7 c2c3 h8d8 e1e2 d8d5 g2g3 c8d8 c1c2 g7f6 h1e1 d5e5 e4d2 e5d5 a2a4 h7h5 d2e4 d5e5 f2f3 d8d7 h2h3 h5h4 g3g4 g6g5 e4d2 b6b5 a4a5');
INSERT INTO evaluations VALUES(23204,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(23204,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(23204,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(23204,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(23204,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(23204,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(23204,36,'cp 147','f6e5 f1c4 g8f6 f3g5 d7d5 e4d5 c6a5 c4d3 d8d5 e1g1 c8g4 f2f3 g4h5 b1c3 d5d7 c1e3 e8c8 e3a7 f8b4 d1e2 b4c3 b2c3 h5g6 d3b5 d7f5 a7e3 f6d5 g5e4 f5e6 e4c5 e6f6 a2a4 d5e3 e2e3 f6f4 e3e1 d8d6 g2g3 f4g5 c5e4 g6e4 e1e4');
INSERT INTO evaluations VALUES(23278,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(23278,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(23278,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(23278,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(23278,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(23278,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(23278,36,'cp 184','f1c4 g8f6 f3g5 d7d5 e4d5 c6a5 c4d3 d8d5 e1g1 c8g4 f2f3 g4h5 b1c3 d5d7 c1e3 f8b4 d1e2 b4c3 b2c3 e8g8 a1d1 d7e7 e2f2 b7b6 f2h4 e7e8 d3f5 h5g6 f5e6 g8h8 d1e1 h7h6 f3f4 f6h7 g5h7 h8h7 f4f5 g6f7 e3h6 g7h6 e1e5 a5c6 e5e3 a8d8 f1e1 e8e7 f5f6 e7c5 g1h1');
INSERT INTO evaluations VALUES(23275,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(23275,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(23275,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(23275,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(23275,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(23275,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(23275,36,'cp 173','g8f6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8e7 f1e1 e8g8 e1e5 h7h6 d1d4 h6g5 e5e2 e7f6 d4d3 c8g4 e2e3 c7c5 c2c3 d8d7 b1d2 g4f5 d2e4 a8d8 d5d6 f5e4 e3e4 d7d6 d3c4 g8h8 c1e3 b7b6 h2h3 d6d7 a1e1 f8e8 a2a4 e8e4 c4e4 d8e8 e4c2');
INSERT INTO evaluations VALUES(17102,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(17102,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(17102,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(17102,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(17102,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(17102,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(17102,36,'cp 173','f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8e7 f1e1 e8g8 e1e5 h7h6 d1d4 h6g5 d4c4 e7f6 e5e3 b7b5 c4b5 f6d4 b1c3 d4e3 c1e3 c8f5 b5c4 d8f6 d5d6 f8f7 d6c7 a8c8 c3b5 a7a6 b5d4 c8c7 c4d5 c7d7 d5f5 f6f5 d4f5 f7f5 h2h3 f5d5 a1e1 d7c7 e1c1 c7d7 g1f1 d5d1 f1e2');
INSERT INTO evaluations VALUES(47181,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(47181,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(47181,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(47181,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(47181,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(47181,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(47181,36,'cp -22','c8f5 f1d3 e7e6 d3f5 e6f5 g1e2 c7c6 e1g1 f8d6 b1d2 b8d7 d1b3 d8c8 c3c4 e8g8 b3c2 g7g6 b2b3 c8d8 a2a4 f8e8 d2f3 d6c7 e2f4 g8g7 c4d5 f6d5 f4d5 c6d5 c1a3 d7f6 f1c1 a8c8 g2g3 f6e4');
INSERT INTO evaluations VALUES(29571,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(29571,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(29571,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(29571,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(29571,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(29571,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(29571,36,'cp -12','f1d3 e7e6 d3f5 e6f5 b2b3 h7h5 g1h3 c7c6 c1a3 b8d7 a3f8 d7f8 b1d2 f8e6 d1c2 g7g6 e1g1 d8d6 d2f3 e8g8 c3c4 f8e8 a1c1 a8c8 c4c5 d6e7 a2a3 f6e4 f3e5 e4g5 h3g5 e7g5 h2h3 g5e7');
INSERT INTO evaluations VALUES(49844,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(49844,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(49844,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(49844,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(49844,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(49844,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(49844,36,'cp 66','f8b4 b1c3 b4c3 b2c3 d8e7 d1d5 b8c6 g1f3 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 b7b6 f1g2 c8b7 e1g1 c6a5 f3d2 b7g2 g1g2 h7h6 d2b3 a5b7 f2f3 g7g5 f4e3 e8c8 e3d4 h8e8 d3f5 f6d7 e2e4 c7c5 d4e3 c8c7 f5d5 d7e5');
INSERT INTO evaluations VALUES(42898,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(42898,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(42898,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(42898,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(42898,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(42898,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(42898,36,'cp 88','b1c3 b4c3 b2c3 d8e7 d1d5 b8c6 g1f3 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 c8g4 e1g1 a8e8 a1e1 b7b6 h2h3 g4d7 c4c5 b6c5 e2e4 f6h5 f4c1 e7f7 e4e5 c6e5 f3e5 d6e5 c1a3 e8d8 e1d1 h5f6 a3c5 f8e8 d3e3 a7a5 d1d2 d7e6 d2d8 e8d8 f1e1');
INSERT INTO evaluations VALUES(42896,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(42896,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(42896,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(42896,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(42896,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(42896,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(42896,36,'cp 66','b4c3 b2c3 b8c6 g1f3 d8e7 d1d5 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 h7h6 e1g1 c6e5 f4e5 d6e5 f3d2 f8d8 d3c2 c7c6 f1d1 c8e6 a2a4 a7a5 c2b2 d8d7 d2f1 a8d8 d1d7 d8d7 b2b6 e7d8 b6c5 d8c7 h2h3 e6f7 a1b1 f7g6');
INSERT INTO evaluations VALUES(41293,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41293,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41293,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(41293,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(41293,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41293,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(41293,36,'cp 33','e8g8 g1f3 c8g4 d3e2 f6d7 c1e3 g4f3 e2f3 b8c6 d4d5 c6a5 f3e2 g7c3 b2c3 e7e5 d5e6 f7e6 e1g1 f8f7 f2f4 d8e7 e2d3 b7b6 d1d2 d7c5 e3c5 b6c5 a1e1 a8f8 g2g3 a5c6 h2h4 g8h8 g1g2 e6e5 f4f5 f8g8 f1h1 f7f6 e1f1');
INSERT INTO evaluations VALUES(36581,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(36581,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(36581,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(36581,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(36581,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(36581,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(36581,36,'cp 35','g1e2 b8c6 e1g1 e7e5 d4d5 c6d4 c1g5 c7c5 d1d2 a7a6 a1b1 d8c7 b2b3 f6h5 f2f3 f7f6 g5h4 f6f5 g1h1 g7f6 h4f6 h5f6 b3b4 b7b6 e4f5 c8f5 b4c5 b6c5 d3f5 g6f5 d2g5 c7g7 g5g7 g8g7 b1b6 f8d8 g2g4 f5g4 f3g4 d4e2 c3e2 f6g4');
INSERT INTO evaluations VALUES(36580,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(36580,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(36580,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(36580,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36580,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(36580,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36580,36,'cp 21','b8c6 e1g1 e7e5 d4d5 c6d4 c1g5 c7c5 d1d2 a7a6 a1b1 d8c7 b2b3 f6h5 g1h1 f7f6 g5e3 f6f5 e4f5 g6f5 f2f3 g8h8 b3b4 b7b6 e3h6 f8g8 b4c5 b6c5 d2g5 c7f7 h6g7 h5g7 g5h6 g7e8 b1b6 f7e7 f1b1 g8g6 h6d2 e7h4 e2d4 c5d4');
INSERT INTO evaluations VALUES(3501,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(3501,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(3501,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(3501,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(3501,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(3501,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(3501,36,'cp 33','c1g5 h7h6 g5d2 e7e5 d4d5 c6d4 e1g1 f6d7 f2f3 a7a5 d2e3 f7f5 e4f5 d4f5 d1d2 f5e3 d2e3 g8h7 c3e4 d7f6 e2g3 f6e4 g3e4 b7b6 a1e1 f8f7 e1e2 c8d7 e3d2 d8e7 a2a3 d7f5 f1e1 a8f8 b2b4 a5b4 a3b4 f8a8 c4c5 f5e4 e2e4 b6c5 b4c5');
INSERT INTO evaluations VALUES(20814,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(20814,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(20814,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(20814,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(20814,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(20814,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(20814,36,'cp 43','g7g6 f1b5 c8d7 h2h3 d7b5 c3b5 a7a6 b5c3 f8g7 d1d3 g8f6 c1g5 d8d7 f3d2 h7h6 g5f6 g7f6 e1c1 b7b5 c1b1 b5b4 c3e2 a6a5 d2c4 e7c8 g2g3 c8b6 c4b6 c7b6 h3h4 h6h5 f2f4 a5a4 h1f1 e5f4 g3f4 e8g8 e4e5 d6e5 f4f5 f6g7 f5g6 f7g6 d3g6 d7g4');
INSERT INTO evaluations VALUES(20801,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(20801,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(20801,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(20801,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(20801,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(20801,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(20801,36,'cp 61','f1b5 c8d7 h2h3 d7b5 c3b5 a7a6 b5c3 d8d7 c1e3 f8g7 d1d3 g8f6 a2a4 e8g8 a4a5 h7h6 e1g1 f6h5 a1a4 f7f5 a4b4 d7c8 f3h2 f8f7 f2f3 h5f4 d3c4 c7c6 c4b3 a8b8 e3a7 b7b5 a5b6 b8b7 b4c4 c8d7 d5c6 e7c6');
INSERT INTO evaluations VALUES(33381,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(33381,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(33381,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(33381,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(33381,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33381,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33381,36,'cp 11','g2g3 c7c6 f1g2 e8g8 c1d2 b8a6 b1c3 a6c7 e1g1 d6d5 a2a3 f8e8 f1e1 e7f8 e1e8 f6e8 f3d4 g7g6 a3a4 a7a5 c3e2 c8d7 d2e3 c7a6 c2c3 a6c5 e2c1 e8c7 d4b3 c7e6 b3c5 f8c5 d3d4 c5d6 c1d3 a8e8 a1e1 g8f8 h2h3');
INSERT INTO evaluations VALUES(54049,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(54049,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(54049,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(54049,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(54049,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(54049,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(54049,36,'cp -34','c8d7 b5d7 d8d7 c2c3 g8e7 d2d4 e7f5 e1g1 b8c6 a2a4 a8c8 b1a3 c5d4 c3d4 f8b4 a3c2 e8g8 h2h4 b4a5 c2e3 f5e3 c1e3 h7h6 h4h5 a7a6 a1c1 c6e7 d1b3 c8c1 f1c1 f8c8 g2g3 c8c1 e3c1 d7c6 c1f4 c6c8 f4e3 e7c6 b3d1 c8d7 g3g4');
INSERT INTO evaluations VALUES(30891,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(30891,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(30891,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(30891,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(30891,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(30891,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(30891,36,'cp -37','b5d7 d8d7 c2c3 g8e7 d2d4 e7f5 a2a4 b8c6 e1g1 c5d4 c3d4 f8b4 b1c3 e8g8 d1d3 a8c8 c3e2 h7h6 h2h3 d7e7 e2f4 b4a5 g2g4 f5h4 f3h4 e7h4 c1e3 f7f5 e5f6 h4f6 a1c1 a5b6 f4h5 f6f3 d3g6 f3f7 g6d3 b6d8');
INSERT INTO evaluations VALUES(30858,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(30858,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(30858,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(30858,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(30858,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(30858,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(30858,36,'cp -49','b8d7 c2c3 g8e7 d2d4 e7c6 d1e2 f8e7 e1g1 e8g8 c1e3 d8b6 f1c1 a7a6 c1c2 f8c8 g2g3 a6a5 a2a4 b6b3 e2d1 c5d4 b1d2 b3b6 c3d4 b6a6 h2h4 b7b5 h4h5 h7h6 d1e2 b5a4 e2a6 a8a6 a1a4 d7b6');
INSERT INTO evaluations VALUES(26151,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(26151,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(26151,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(26151,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(26151,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(26151,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(26151,36,'cp -36','c2c3 g8e7 e1g1 e7c6 d2d4 f8e7 d1e2 e8g8 c1e3 d8b6 f1c1 a8c8 c1c2 h7h6 b1d2 c5d4 c3d4 c6b4 c2c8 f8c8 f3e1 b6a6 e2a6 b4a6 a2a3 d7b6 g1f1 a6b8 e1d3 a7a5 f1e2 a5a4 g2g3 g7g5 f2f4 c8c2 f4g5 e7g5');
INSERT INTO evaluations VALUES(48833,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(48833,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48833,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(48833,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48833,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(48833,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(48833,36,'cp 12','e4d5 d8d5 b1c3 f8b4 f1e2 b4c3 b2c3 e8g8 e1g1 b8c6 c3c4 d5d6 a2a4 f8e8 c1a3 d6d8 f1e1 c8g4 h2h3 g4h5 a3b2 d8d6 a4a5 h7h6 f3h4 h5e2 d1e2 c6d4 b2d4 e5d4 e2d2 c7c5 h4f5 d6c7 e1e8 a8e8 a1b1 b7b6 a5b6 a7b6 f5g3');
INSERT INTO evaluations VALUES(48610,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(48610,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(48610,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48610,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(48610,36,'cp 6','d8d5 b1c3 f8b4 f1e2 b8c6 e1g1 d5d8 c3e4 f6e4 d3e4 d8d1 f1d1 b4c5 d1d5 c5d6 c1e3 f7f6 e2b5 c8d7 c2c3 a7a6 b5e2 d7e6 d5d1 a6a5 a2a4 e8f7 f3d2 d6e7 e2b5 h7h5 h2h4 h8d8 g1f1 g7g6 f2f3 c6a7 b5c4');
INSERT INTO evaluations VALUES(33910,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33910,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(33910,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33910,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(33910,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33910,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(33910,36,'cp 14','b1c3 f8b4 f1e2 b8c6 e1g1 d5d8 c3e4 f6e4 d3e4 d8d1 f1d1 b4c5 d1d5 c5d6 c1e3 f7f6 d5d1 c6b4 f3e1 a7a5 e2c4 b4a6 e1d3 a6b4 d1d2 e8e7 a2a4 b7b6 a1d1 c8e6 c4e6 e7e6 h2h4 h8d8 f2f3 d6e7 g1f2 h7h5 f2e2 b4d3 d2d3');
INSERT INTO evaluations VALUES(33909,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(33909,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(33909,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(33909,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(33909,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(33909,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(33909,36,'cp 12','f8b4 f1e2 b8c6 e1g1 d5d8 c3e4 f6e4 d3e4 d8d1 f1d1 b4c5 d1d5 c5d6 c1e3 f7f6 d5d1 c6b4 f3e1 a7a5 a2a3 b4c6 e2b5 c8d7 a3a4 e8e7 h2h4 h7h6 g1f1 d7e6 b5c6 b7c6 e1d3 g7g5 d3c5 d6c5 e3c5 e7e8 b2b3 g5h4');
INSERT INTO evaluations VALUES(46774,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(46774,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(46774,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(46774,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(46774,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(46774,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(46774,36,'cp 15','f8g7 c2c3 e8g8 h2h3 c7c5 b1d2 b7b6 a2a4 c8b7 a4a5 c5c4 f1e2 b6b5 f3e5 f6d7 h3h4 d7e5 f4e5 b8d7 e5g7 g8g7 h4h5 d7f6 a5a6 b7c6 g2g4 d8c7 g4g5 f6e4 f2f4 e4g3 h1h3 g3f5 d2f1 g7h8 h5g6 f7g6 e2g4 e7e6');
INSERT INTO evaluations VALUES(48194,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(48194,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(48194,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(48194,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(48194,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(48194,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(48194,36,'cp 137','d5e6 c8e6 g1f3 f8d6 d2d4 e8g8 f1e2 c7c6 c1g5 b8d7 c3e4 d6e7 e4g3 d8b6 b2b3 c6c5 e1g1 f8e8 c2c3 a8d8 d1c2 h7h6 g5f4 f6d5 f4d2 e7d6 e2c4 b6c6 a2a4 a7a6 a4a5 d5f4 c4e6 f4e6 f1e1 d7f6 h2h3 c6d7 f3e5 d6e5 e1e5');
INSERT INTO evaluations VALUES(48066,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(48066,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(48066,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(48066,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(48066,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(48066,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(48066,36,'cp 127','c8e6 g1f3 f8d6 d2d4 e8g8 f1e2 c7c6 c1g5 b8d7 c3e4 d6e7 e4g3 d8b6 g5c1 c6c5 e1g1 f8e8 b2b3 c5d4 c1b2 a8d8 b2d4 e7c5 c2c3 c5d4 f3d4 d7c5 d1c2 e6d5 a1d1 c5e4 g3f5 e4f2 f1f2 f6e4 c2c1 e4f2');
INSERT INTO evaluations VALUES(30041,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(30041,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(30041,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(30041,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(30041,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(30041,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(30041,36,'cp 132','g1f3 f8d6 d2d4 e8g8 f1e2 c7c6 c1g5 b8d7 c3e4 d6e7 e4g3 h7h6 g5c1 c6c5 e1g1 d8b6 c2c4 c5d4 f3d4 d7e5 d4e6 b6e6 d1b3 a8c8 c1e3 b7b6 a1c1 f8d8 b3c3 f6g4 e3d4 e7f6 g1h1 e6e7 h2h3 e5c6 d4f6 g4f6');
INSERT INTO evaluations VALUES(30040,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(30040,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(30040,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(30040,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(30040,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(30040,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(30040,36,'cp 124','f8d6 d2d4 e8g8 f1d3 b8c6 e1g1 h7h6 c3b5 c6b4 c2c4 b4d3 d1d3 f8e8 b5d6 d8d6 h2h3 b7b5 c4b5 e6d5 a2a4 c7c6 b5c6 d6c6 f3e5 c6b7 f2f3 f6h5 c1d2 a8c8 f1c1 b7b2 a4a5 c8c1 a1c1 e8d8 c1c2 b2b8 e5c6 d5c6 c2c6');
INSERT INTO evaluations VALUES(28447,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(28447,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(28447,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(28447,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(28447,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(28447,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(28447,36,'cp 134','d2d4 e8g8 f1d3 b8c6 e1g1 h7h6 c3b5 c6b4 c2c4 b4d3 d1d3 f8e8 b5d6 d8d6 h2h3 b7b5 c4b5 e6d5 a2a4 c7c6 b5c6 d5c6 c1d2 d6d7 d3a3 d7b7 b2b4 e8e2 a3d3 a8e8 b4b5 c6f3 g2f3 b7d7');
INSERT INTO evaluations VALUES(28445,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(28445,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(28445,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(28445,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(28445,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(28445,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(28445,36,'cp 125','e8g8 f1d3 h7h6 e1g1 b8c6 c3b5 c6b4 c2c4 b4d3 d1d3 f8e8 b5d6 d8d6 h2h3 b7b5 c4b5 e6d5 a2a4 c7c6 b5c6 d5c6 d3a3 d6a3 a1a3 a8c8 f3e5 c6a8 f1d1 f6d5 a3g3 g7g5 e5g4 g8g7 g4e3 g7g6 c1d2 d5e3 f2e3 a8c6 a4a5 f7f6');
INSERT INTO evaluations VALUES(57080,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(57080,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(57080,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(57080,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(57080,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(57080,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(57080,36,'cp 31','d7d5 g1f3 c7c5 c2c4 d5d4 e2e3 b8c6 e1g1 g8f6 e3d4 c5d4 d2d3 a7a5 b1a3 f8d6 c1g5 h7h6 g5f6 d8f6 f3d2 e8g8 d2e4 f6e7 e4d6 e7d6 a3b5 d6c5 b2b3 c8d7 d1d2 c6e5 f2f4 e5g4 g2b7 g4e3 b7a8 f8a8 f1e1 d7b5 c4b5 c5b5');
INSERT INTO evaluations VALUES(15425,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(15425,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15425,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(15425,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(15425,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(15425,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(15425,36,'cp 24','g7g6 d1e2 c8d7 e1g1 f8g7 d3d4 e5d4 e4e5 d6e5 b5c6 b7c6 e2e5 d8e7 e5e7 e8e7 f3d4 h8e8 c3a4 e7f8 a4c5 d7g4 h2h3 g4e2 f1e1 e2b5 c1f4 f6d5 f4e5 g7e5 e1e5 e8e5 c5d7 f8e8 d7e5 f7f6 a1d1 a8d8');
INSERT INTO evaluations VALUES(34720,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(34720,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(34720,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(34720,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(34720,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(34720,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(34720,36,'cp 105','d2d4 c8e6 c4e6 e7e6 d1d3 c7c6 c2c4 g8f6 e1g1 f8e7 b1c3 b8d7 a2a4 e8g8 h2h3 e7d8 a4a5 e6e7 f1e1 a7a6 c1g5 h7h6 g5h4 e7e6 d4d5 c6d5 e4d5 e6e8 b2b4 a8c8 h4f6 d7f6 f3d2 f6h5 c3e4 h5f4 d3e3 e8e7 b4b5 a6b5');
INSERT INTO evaluations VALUES(34722,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(34722,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(34722,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(34722,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(34722,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(34722,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(34722,36,'cp 112','c8e6 d4d5 e6c8 h2h3 g7g6 a2a4 a7a6 b1c3 b8d7 g2g4 h7h5 h1g1 f8h6 g4g5 h6g7 c4e2 e7d8 c1e3 g8e7 f3h4 a8b8 g1g3 d7c5 b2b3 b7b6 b3b4 c5d7 d1d2 e8g8 e1f1 a6a5 b4b5 b8a8 a1d1 f8e8 f1g2 c8b7 g2h1');
INSERT INTO evaluations VALUES(6542,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(6542,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(6542,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(6542,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(6542,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(6542,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(6542,36,'cp 106','c4f7 e8f7 d1d5 f7e8 d5h5 e8f8 h5c5 d8e7 c5e3 d7d5 e4d5 e7e3 f2e3 c6b4 e1g1 b4d5 b1c3 d5c3 b2c3 b7b6 e3e4 c8a6 f1e1 g8e7 c1f4 f8f7 f3g5 f7f6 h2h4 e7g6 e4e5 f6f5 f4g3 h7h6 g5f3 f5g4 h4h5 a6b7 g3h2 g6f4 e5e6 b7f3 g2f3 g4f5 h2f4 f5f4 e6e7 h8e8 a1d1');
INSERT INTO evaluations VALUES(6154,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(6154,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(6154,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(6154,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(6154,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(6154,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(6154,36,'cp 108','e8f7 d1d5 f7f8 d5c5 d8e7 c5e3 d7d5 e4d5 e7e3 f2e3 c6b4 e1g1 b4d5 b2c3 g8f6 c3c4 d5b6 b1d2 f8f7 f3e5 f7e8 c1a3 c8e6 c4c5 b6d7 e5f3 e8d8 f3d4 h8e8 c5c6 b7c6 d4c6 d8c8 e3e4 e6g8 f1b1 a7a5 b1b2 d7e5 c6e7 c8d7 b2b5 e5d3 e7g8 e8g8 a1d1 g8b8 b5b8 a8b8 d2f3');
INSERT INTO evaluations VALUES(4663,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(4663,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(4663,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(4663,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(4663,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(4663,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(4663,36,'cp 92','d1d5 f7f8 d5c5 d8e7 c5e3 d7d5 e4d5 e7e3 f2e3 c6b4 e1g1 b4d5 b2c3 g8f6 c3c4 d5b6 e3e4 f6e4 f3e5 e4f6 b1c3 c8e6 c3e4 b6c4 e4f6 g7f6 f1f6 f8e7 c1g5 c4e5 f6h6 e7d6 a1d1 d6c5 h6e6 e5c4 e6h6 a8e8 h2h4 c7c6 g5f6 h8g8 h6h7 g8g2 g1g2 c4e3 g2g3 e3d1 h7b7 e8e1 h4h5 d1e3 b7a7 c5b6 a7h7 c6c5 h5h6 e3f5 g3f3 e1f1 f3e2');
INSERT INTO evaluations VALUES(4666,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(4666,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(4666,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(4666,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(4666,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(4666,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(4666,36,'cp 100','f7f8 d5c5 d8e7 c5e3 d7d5 e4d5 c6b4 e1g1 e7e3 f2e3 b4d5 b2c3 g8f6 c3c4 d5b6 e3e4 f6e4 f3e5 e4f6 b1c3 c8e6 c3e4 b6c4 e4f6 g7f6 f1f6 f8e7 f6h6 c4e5 c1g5 e7d6 a1d1 d6c5 h6e6 e5c4 h2h4 h7h5 d1d7 a8c8 g5f6 c4b6 d7f7 h8g8 f7h7 c5b5 e6e5 c7c5 h7b7 b5c6 b7a7 b6d5 e5h5 g8g2 g1g2 d5f4 g2f3 f4h5 f6e5 c6d5 a7e7 c5c4 f3g4');
INSERT INTO evaluations VALUES(2521,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(2521,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(2521,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(2521,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(2521,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(2521,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(2521,36,'cp 94','d5c5 d8e7 c5e3 d7d5 e4d5 c6b4 e1g1 e7e3 f2e3 b4d5 b2c3 g8f6 c3c4 d5b6 e3e4 f6e4 f3e5 e4f6 b1c3 f8g8 c3b5 c8d7 b5c7 a8c8 c7d5 b6d5 c4d5 d7b5 f1d1 c8c5 a1b1 a7a6 c1b2 c5d5 d1c1 b5c6 e5c6 b7c6 c1c6 g8f7 b1f1 h8e8 b2f6 g7f6 c6f6 f7g8 f6a6 e8e2 f1b1 d5d7 a6a5 d7d2 a5g5 g8f7 b1b7 f7f6 h2h4 d2a2 b7h7');
INSERT INTO evaluations VALUES(2520,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(2520,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(2520,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(2520,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(2520,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(2520,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(2520,36,'cp 88','d8e7 c5e3 d7d5 e4d5 e7e3 f2e3 c6b4 e1g1 b4d5 b2c3 g8f6 c3c4 d5b6 e3e4 f6e4 f3e5 e4f6 b1c3 c8e6 c3e4 b6c4 e4f6 g7f6 f1f6 f8e7 c1g5 c4e5 f6h6 e7d6 a1d1 d6c5 h6e6 e5c4 e6h6 a8e8 h2h4 c4d6 g5f6 h8f8 f6d4 c5b5 h6h7 f8f4 h4h5 c7c5 d4a1');
INSERT INTO evaluations VALUES(9760,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9760,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9760,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9760,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9760,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(9760,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(9760,36,'cp 0','c5b6 e1g1 d7d6 a2a4 e8g8 c3d5 f6d5 c4d5 d8f6 d1d2 c6e7 a4a5 e7d5 e4d5 b6e3 f2e3 f6e7 e3e4 f7f5 d2c3 f5e4 d3e4 c8g4 f3d2 e7g5 f1f8 a8f8 a1f1 f8f1 d2f1 g5c1 c3c7 g4e2 c7c8 g8h7 c8f5 h7g8');
INSERT INTO evaluations VALUES(9713,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(9713,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(9713,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9713,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(9713,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9713,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(9713,36,'cp 0','a2a4 d7d6 e1g1 e8g8 c3d5 f6d5 c4d5 d8f6 h2h3 c8e6 d5e6 f7e6 c2c3 f8f7 f3h2 a8f8 e3b6 a7b6 b2b4 c6b8 d1e2 f6e7 e2g4 b8c6 g4d1 e7e8 d1g4');
INSERT INTO evaluations VALUES(22877,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(22877,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(22877,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(22877,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(22877,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(22877,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(22877,36,'cp 33','c2c4 f8c5 b1c3 g8f6 g1f3 e8g8 e1g1 d7d6 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 f3g2 c8g4 d1c2 d8c8 b2b3 g4h3 c1b2 h3g2 g1g2 c8e6 d3d4 c7c6 a4b6 a7b6 a2a4 e5e4 d4d5 c6d5 b2f6 e6f6 c4d5');
INSERT INTO evaluations VALUES(38763,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(38763,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(38763,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(38763,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38763,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38763,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(38763,36,'cp 11','e6d5 f1d3 g8f6 a2a3 b4c3 b2c3 e8g8 g1e2 c8g4 e1g1 g4h5 f1e1 f8e8 a1b1 d8c8 c1f4 b8c6 d1d2 h5g6 e2g3 e8e1 b1e1 c8d7 f4e5 f6e8 h2h3 g6d3 c2d3 b7b6 g3h5 f7f6 e5f4 c6e7 d2e2 e7g6');
INSERT INTO evaluations VALUES(50915,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(50915,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(50915,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(50915,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(50915,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(50915,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(50915,36,'cp 54','d2d4 e5d4 e1g1 f8e7 f3d4 g8f6 b1c3 e8g8 c1f4 f8e8 a2a4 b8c6 h2h3 e7f8 d4c6 b7c6 d1d3 f6d7 d3d2 d7c5 f1e1 c8e6 c4f1 a6a5 a1d1 d8b8 b2b3 b8b4 f4e3 a8d8 f2f3 h7h6 e3d4 b4b7 f3f4 d6d5 f4f5 e6c8 e4d5');
INSERT INTO evaluations VALUES(50916,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(50916,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(50916,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(50916,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(50916,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(50916,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(50916,36,'cp 54','e5d4 f3d4 g8f6 e1g1 f8e7 b1c3 e8g8 c1f4 f8e8 a2a4 b8c6 d4c6 b7c6 f1e1 c8g4 d1d2 g4e6 c4e6 f7e6 a1d1 e6e5 f4g5 f6g4 g5e7 d8e7 f2f3 g4f6 c3e2 e7f7 e2g3 f6d7 b2b3 d7c5 d2c3 a6a5 g3f5 f7f6 f5e3');
INSERT INTO evaluations VALUES(15937,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(15937,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(15937,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(15937,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(15937,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(15937,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(15937,36,'cp 40','c4d5 e6d5 e2e3 f8e7 f1d3 c7c6 g1f3 e8g8 h2h3 f6e4 g5f4 d7f6 f3e5 e7d6 e1g1 d8e7 f1e1 f8e8 f2f3 e4c3 b2c3 f6d7 e5g4 d7b6 f4d6 e7d6 d1c2 c8g4 h3g4 h7h6 a2a4 b6d7 c2f2 a7a5 a1b1 b7b6 e3e4 d5e4 f3e4');
INSERT INTO evaluations VALUES(15933,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(15933,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(15933,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15933,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(15933,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(15933,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(15933,36,'cp 32','e6d5 e2e3 f8e7 f1d3 e8g8 g1f3 c7c6 e1g1 f6e4 g5f4 d7f6 f3e5 g7g5 f4g3 c8f5 a2a4 e7b4 d3e4 f6e4 c3e4 f5e4 h2h4 f7f6 f2f3 e4f5 e5d3 b4d6 g3d6 d8d6 h4g5 f5d3 d1d3 f6g5 a4a5 a8e8 e3e4 d6g3 e4d5 g5g4 d5c6 b7c6 a1c1 f8f3 f1f3 g4f3 d3f3 g3f3 g2f3');
INSERT INTO evaluations VALUES(15979,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15979,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(15979,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15979,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(15979,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(15979,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(15979,36,'cp 36','e2e3 c7c6 f1d3 f8e7 g1f3 d7f8 d1c2 f8e6 g5h4 g7g6 e1g1 e8g8 a1b1 a7a5 a2a3 f6e8 h4e7 d8e7 b2b4 a5b4 a3b4 e8d6 b4b5 e6g5 f3d2 c8g4 b5c6 b7c6 c3a4 g8g7 b1b6 d6f5 a4c5 f5h4 b6b7 e7d6 f2f3 g4f5 d3f5 h4f5');
INSERT INTO evaluations VALUES(15977,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(15977,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(15977,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(15977,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(15977,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(15977,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(15977,36,'cp 21','c7c6 f1d3 f8e7 g1f3 f6e4 g5f4 d7f6 e1g1 e7d6 f4d6 e4d6 b2b4 e8g8 d1c2 d8e7 a2a4 a7a5 b4b5 c6b5 c3b5 d6b5 d3b5 c8e6 c2c5 e7c5 d4c5 f8c8 f1c1 f6e4 c5c6 b7c6 b5c6 a8b8 f3d4 g8f8 c6b5 c8c1 a1c1 e4d6');
INSERT INTO evaluations VALUES(12188,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(12188,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(12188,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(12188,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12188,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(12188,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(12188,36,'cp 27','f1d3 f8e7 g1f3 f6e4 g5f4 d7f6 e1g1 e7d6 d1c2 e8g8 c3e4 d5e4 d3e4 f6e4 c2e4 f8e8 e4d3 d6f4 e3f4 d8d5 f1e1 c8f5 d3c3 f5e4 f3h4 d5d6 e1e3 e4d5 e3e5 d6d7 h2h3 f7f6 e5e1 e8e1 a1e1 a8e8 a2a3 e8e1 c3e1 g8f7 e1d2 d7e6 f2f3 g7g6 d2c2 e6e7 g1f2');
INSERT INTO evaluations VALUES(12191,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(12191,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(12191,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12191,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(12191,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(12191,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12191,36,'cp 28','f8e7 g1f3 f6e4 g5f4 d7f6 h2h3 e8g8 e1g1 c8f5 f3e5 d8c8 a2a4 e4c3 b2c3 f5d3 d1d3 b7b6 f1c1 f8e8 c3c4 c6c5 d4c5 c8c5 e5f3 h7h6 f4e5 a8d8 e5d4 d5c4 c1c4 c5d5 c4c7 d8d7 a1c1 d7c7 c1c7');
INSERT INTO evaluations VALUES(7794,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(7794,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(7794,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(7794,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(7794,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(7794,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(7794,36,'cp 23','g1f3 f6e4 g5f4 d7f6 h2h3 e8g8 e1g1 c8f5 f3e5 d8c8 d1c2 e4d6 d3f5 d6f5 f1b1 f8e8 b2b4 f5d6 a2a4 f6e4 c3e4 d6e4 b4b5 g7g5 f4h2 c6c5 d4c5 e7c5 e5d3 c5d6 c2c8 d6h2 g1h2 a8c8 a4a5');
INSERT INTO evaluations VALUES(7796,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(7796,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(7796,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(7796,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(7796,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(7796,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(7796,36,'cp 23','f6e4 g5f4 d7f6 e1g1 e7d6 d3e4 f6e4 c3e4 d5e4 f4d6 d8d6 f3d2 d6e7 a1c1 e8g8 c1c5 f7f5 d2b1 c8e6 b1c3 a8d8 d1c2 e7f7 b2b3 e6d5 c3d5 d8d5 g2g3 f8d8 b3b4 a7a6 a2a4 h7h6 f1c1 g8h8 c5d5 d8d5 c2b1 f7h5 b1b3');
INSERT INTO evaluations VALUES(47091,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(47091,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(47091,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(47091,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(47091,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(47091,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(47091,36,'cp -57','e7e5 e2e3 e5e4 b2b4 b7b6 d1a4 c8d7 a4c2 f8e7 b1c3 e8g8 a1b1 c7c6 g1e2 b8a6 f2f3 a8b8 a2a3 b6c5 d4c5 e4f3 g2f3 a6c7 f1g2 a7a5 e1g1 f6h5 c3a4 a5b4 a3b4 e7f6 a4b6 d7e6 f3f4 g7g6 c1d2 c7b5 e3e4 b5d4');
INSERT INTO evaluations VALUES(22695,30,'cp -146',NULL);
INSERT INTO evaluations VALUES(22695,31,'cp -140',NULL);
INSERT INTO evaluations VALUES(22695,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(22695,33,'cp -165',NULL);
INSERT INTO evaluations VALUES(22695,34,'cp -152',NULL);
INSERT INTO evaluations VALUES(22695,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(22695,36,'cp -163','g8f6 h5h4 h7h6 d2d3 f8e7 c1g5 d7d6 f1e2 e8g8 g5f6 e7f6 h4g3 c6e7 h2h4 e7g6 g3h2 c7c6 b1d2 c8e6 g2g3 d8b6 b2b3 d6d5 h2g2 f8e8 e1g1 b6a5 f1c1 a5c7 a2a4 a7a5 c2c3 c7d7 d3d4 e6h3');
INSERT INTO evaluations VALUES(16378,30,'cp -151',NULL);
INSERT INTO evaluations VALUES(16378,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(16378,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(16378,33,'cp -159',NULL);
INSERT INTO evaluations VALUES(16378,34,'cp -156',NULL);
INSERT INTO evaluations VALUES(16378,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(16378,36,'cp -144','h5h4 h7h6 d2d3 f8e7 c1g5 d7d6 f1e2 e8g8 g5f6 e7f6 h4g3 c6e7 h2h4 e7g6 g3h2 g6f4 e2f1 a7a5 g2g3 f4e6 f1h3 d6d5 e1g1 d5e4 d3e4 c7c6 b1c3 d8b6 c3d1 b6b4 c2c3 b4e4 h2g2 e4g6 d1e3 e6c5 h3c8 a8c8 e3c4');
INSERT INTO evaluations VALUES(16493,30,'cp -152',NULL);
INSERT INTO evaluations VALUES(16493,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(16493,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(16493,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(16493,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(16493,35,'cp -146',NULL);
INSERT INTO evaluations VALUES(16493,36,'cp -145','f8e7 d2d3 h7h6 c1g5 d7d6 f1e2 e8g8 g5f6 e7f6 h4g3 c6e7 h2h4 e7g6 g3h2 a7a5 g2g3 a5a4 a2a3 c7c6 b1d2 f8e8 e1g1 b7b5 c2c3 g6f8 h2g2 h6h5 f3e1 g7g6 e1c2 f6g7 c2b4 d8b6 f1b1 f8e6 b1e1 e6c5 d2f1 g7h6');
INSERT INTO evaluations VALUES(33631,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(33631,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(33631,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(33631,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(33631,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(33631,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(33631,36,'cp 9','f8e7 b1c3 c8g4 f1e2 b8c6 c1e3 e8c8 h2h3 g4f3 e2f3 d6d5 e3d2 d5d4 c3e4 c6e5 f3e2 f6e4 d3e4 d4d3 e2d3 e5d3 c2d3 d8d3 e1c1 h8d8 d2c3 d3d1 h1d1 g7g6 f2f3 d8d1 c1d1 c8d7 d1c2 d7e6 c3d2 e7d6 c2d3 f7f5 e4f5 g6f5');
INSERT INTO evaluations VALUES(46718,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(46718,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(46718,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(46718,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(46718,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(46718,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(46718,36,'cp 40','f8g7 g1f3 c7c5 f1b5 c8d7 e4e5 f6g4 b5d7 d8d7 h2h3 c5d4 d1d4 g4h6 d4f2 b8c6 c1d2 d6e5 f4e5 e8g8 e1c1 a8d8 h1e1 d7e6 c1b1 h6f5 f3g5 e6c4 e1e4 f5d4 d2e3 f7f6 e5f6 f8f6 g5f3 f6f8 a2a3 e7e5 e3d4 e5d4 f2e2 c4e2 c3e2 g7f6 e2d4 c6d4 f3d4 f8e8 e4e8 d8e8 d4b3 e8e2');
INSERT INTO evaluations VALUES(31958,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(31958,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(31958,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(31958,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(31958,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(31958,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(31958,36,'cp 56','g8f6 g1f3 e7e6 b1c3 c7c6 e1g1 b8d7 c4d3 f5d3 d1d3 c6c5 d3e2 c5d4 e3d4 f8b4 d4d5 d8e7 d5e6 e7e6 f1e1 e8g8 e2e6 f7e6 e1e6 d7c5 e6e2 c5d3 a1b1 a8c8 c1e3 b4c3 b2c3 b7b6 e2c2 f6d5 e3d4 c8c4 g2g3');
INSERT INTO evaluations VALUES(29706,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(29706,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(29706,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(29706,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(29706,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(29706,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(29706,36,'cp 64','g1f3 e7e6 e1g1 b8d7 b1c3 c7c6 d1e2 b7b5 e3e4 f6e4 c4e6 f5e6 e2e4 a8c8 f3g5 d7f6 e4e2 d8d7 f1e1 f8b4 g5e6 f7e6 c1f4 e8g8 e2e6 d7e6 e1e6 b4c3 b2c3 f6d5 f4d2 f8e8 e6e5 e8e5 d4e5 c6c5 e5e6 b5b4 a1c1 b4c3 d2c3 d5c3 c1c3 g8f8');
INSERT INTO evaluations VALUES(29707,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(29707,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(29707,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(29707,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(29707,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(29707,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(29707,36,'cp 60','e7e6 b1c3 c7c6 e1g1 f8e7 f3h4 f5g6 h4g6 h7g6 d1e2 b7b5 c4d3 b8d7 b2b3 e8g8 c1b2 a7a6 c3e4 a8c8 f1d1 d8b6 g2g3 f6e4 d3e4 c6c5 d4c5 d7c5 e4g2 f8d8 h2h4 c5d7 a1c1 c8c1 d1c1 e7f6 b3b4 f6b2 e2b2');
INSERT INTO evaluations VALUES(30199,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(30199,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(30199,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(30199,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(30199,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(30199,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(30199,36,'cp 77','b1c3 c7c6 e1g1 f8e7 f3h4 f5g6 h4g6 h7g6 a2a3 d8c7 g2g3 b8d7 d1e2 e6e5 c4a2 a8d8 f1d1 d7b6 d1e1 e7d6 c1d2 e8g8 e2d3 b6d5 c3d5 c6d5 a1c1 c7d7 d4e5 d6e5 d2b4 f8e8 c1c2 d8c8 e1d1 d7h3 c2c8 e8c8');
INSERT INTO evaluations VALUES(37758,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37758,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37758,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37758,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37758,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37758,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(37758,36,'cp 0','b8c6 e1g1 d7d6 f1e1 a7a6 a2a4 e7e6 d2d3 g8e7 c3e2 d6d5 c4a2 b7b6 c2c3 e8g8 e4d5 e7d5 c1g5 d8d7 d1d2 c8b7 e2g3 f7f6 g5h6 a8d8 h2h3 d5c7 h6f4 c7d5');
INSERT INTO evaluations VALUES(34555,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(34555,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(34555,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(34555,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(34555,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34555,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(34555,36,'cp 66','g8f6 d2d4 e6b6 a2a4 c8d7 g2g3 e7e6 f1g2 d7c6 d4d5 f6d5 c3d5 c6d5 g2d5 e6d5 e1g1 b8c6 c1e3 b6a6 d1d5 a6e2 f1e1 e2a6 e3c5 f8e7 b2b4 e8g8 c5e7 f8e8 d5d7 a6c4 d7c7 h7h6 c7b7 a8b8 b7d7 c6e7 e1e7 e8e7 d7e7 b8b4 e7d8 g8h7');
INSERT INTO evaluations VALUES(33676,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(33676,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(33676,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(33676,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(33676,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(33676,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(33676,36,'cp 87','d2d4 e6b6 a2a4 e7e6 a4a5 b6c6 e2f4 f8b4 c1d2 e8g8 f4d3 b4c3 b2c3 b8d7 d1f3 c6f3 g2f3 b7b6 a5a6 f6d5 h1g1 f8d8 f1e2 d7f6 c3c4 d5e7 d2f4 f6e8 c2c3 c8d7 h2h4 g8f8 h4h5 h7h6 d3e5 a8c8 e1c1 d7a4 d1d3 e8f6 e5g4');
INSERT INTO evaluations VALUES(33670,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(33670,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(33670,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33670,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(33670,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(33670,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(33670,36,'cp 65','e6b6 a2a4 c8d7 a4a5 b6d6 c1f4 d6b4 g2g4 d7c6 g4g5 f6d5 f4d2 d5c3 b2c3 b4d6 h1g1 e7e6 g1g3 f8e7 d2f4 d6d8 h2h4 e8g8 h4h5 c6e4 d1d2 b8c6 d2e3 e4c2 f1g2 d8d7 a5a6 b7b6 e1f1 a8d8 f1g1 e7d6 h5h6 g7g6 f4d6 d7d6 g2c6 d6c6');
INSERT INTO evaluations VALUES(19946,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19946,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19946,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19946,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19946,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19946,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19946,36,'cp 0','e7e6 e1g1 f8e7 c2c3 g8f6 b1d2 e8g8 f1e1 c6a5 c4b5 c8d7 b5d7 d8d7 a2a4 a5c6 d2c4 a8d8 c1f4 f6h5 f4e3 d7c7 f3d2 h5f6 e3f4 c7d7');
INSERT INTO evaluations VALUES(20673,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20673,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20673,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20673,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20673,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20673,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20673,36,'cp 0','e1g1 f8e7 c2c3 g8f6 f1e1 e8g8 a2a4 d6d5 c4b3 d5e4 d3e4 d8d1 b3d1 f8d8 d1c2 b7b6 b1a3 h7h6 h2h3 c8a6 a3b5 d8d7 f3d2 c6a5 d2f3');
INSERT INTO evaluations VALUES(20672,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20672,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20672,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20672,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20672,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20672,36,'cp 0','g8f6 a2a4 f8e7 c4b3 c6a5 b3a2 a7a6 c2c3 e8g8 c1f4 b7b5 b1d2 c8d7 e4e5 d6e5 f3e5 a5c6 d1f3 c6e5 f4e5 d8c8 f3g3 d7c6 d2b3 c6d5 b3d2');
INSERT INTO evaluations VALUES(28817,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(28817,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(28817,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(28817,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(28817,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(28817,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(28817,36,'cp 111','e4e5 f6d5 c3d5 b7d5 c1g5 d8c8 g1h3 e7e6 h3f4 d5b7 f1d3 c7c5 c2c3 b8c6 e1g1 f8e7 d1g4 g7g6 g5e7 c6e7 f1e1 c5d4 c3d4 c8d8 h2h4 e7d5 a1d1 d5f4 g4f4 e8g8 d3c4 f7f6 f4h6 d8e7 d4d5 f6e5 e1e5 e7f6');
INSERT INTO evaluations VALUES(28785,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(28785,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(28785,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(28785,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(28785,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(28785,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(28785,36,'cp 106','f6d5 c3d5 b7d5 c1g5 d8c8 g1h3 e7e6 h3f4 c8b7 g5h4 c7c5 d1g4 c5d4 f1d3 b8c6 f4d5 c6e5 g4d4 e5d3 d4d3 b7d5 d3d5 e6d5 e1c1 f7f6 h1e1 e8f7 d1d5 d7d6 f2f4 h7h5 c1d2 a8c8 h4f2 g7g6 d2d3 f8g7 d5d6 h8d8 d6d8 c8d8 d3e2');
INSERT INTO evaluations VALUES(28833,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(28833,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(28833,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(28833,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(28833,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(28833,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(28833,36,'cp 122','c3d5 b7d5 c1g5 d7d6 d1d2 h7h6 g5h4 d8d7 e5d6 d7d6 h4g3 d6d7 g1f3 b8c6 f1b5 d5f3 g2f3 a7a6 b5c6 d7c6 d4d5 c6c4 b2b3 c4c5 e1c1 g7g6 g3e5 h8g8 h1e1 e8c8 d2e2 c5a3 c1b1 f8g7 f3f4 g7e5 f4e5 e7e6 c2c4 e6d5');
INSERT INTO evaluations VALUES(28831,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(28831,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(28831,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(28831,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(28831,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(28831,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(28831,36,'cp 98','b7d5 c1g5 d7d6 g1e2 h7h6 g5h4 d6e5 d4e5 g7g5 e2c3 d5c6 h4g3 d8d1 a1d1 e7e6 h2h4 h8g8 c3b5 c6b5 f1b5 c7c6 b5e2 b8d7 h4g5 h6g5 e2f3 a8c8 h1h7 f8g7 c2c3 e8e7 e1e2 f7f5 e5f6 d7f6 g3d6 e7f7 f3h5 f6h5 h7h5 f7g6 h5h3 g8d8 h3g3 g7f6');
INSERT INTO evaluations VALUES(28605,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(28605,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(28605,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(28605,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(28605,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(28605,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(28605,36,'cp 99','c1g5 d7d6 d1d2 d8d7 e5d6 d7d6 g1e2 e7e6 e2c3 b8c6 e1c1 d6d7 c3d5 d7d5 c2c3 f8e7 g5e3 e8c8 c1b1 e6e5 d4e5 d5e5 d2c2 c8b8 f1a6 e7g5 e3g5 e5g5 g2g3 g5a5 a6e2 a5c5 f2f4 h8e8 e2f3 e8e3 f3g2 d8d1 h1d1 g7g6 a2a3 c6e7');
INSERT INTO evaluations VALUES(42762,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(42762,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(42762,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42762,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42762,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(42762,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(42762,36,'cp 10','g1e2 c7c6 g2g3 b4c3 e2c3 d7d5 e4d5 c6d5 d2d4 b8c6 d4e5 c6e5 c4d5 e8g8 f1g2 c8g4 d1d4 e5f3 g2f3 g4f3 e1g1 f6d5 d4d3 f3g2 f1d1 d5c3 d3d8 a8d8 d1d8 f8d8 g1g2 c3d5 c1d2 f7f6 a1c1 d8d7 g2f3 g8f7 c1c4 d5b6 c4c2 b6d5 h2h4 a7a6 h4h5 f7e6 c2c5 d5e7 f3e2 e7c6 d2c3 e6f7 c5c4');
INSERT INTO evaluations VALUES(42763,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(42763,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(42763,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(42763,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(42763,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(42763,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42763,36,'cp 5','c7c6 g2g3 b4c3 e2c3 d7d5 e4d5 c6d5 d2d4 b8c6 d4e5 c6e5 c4d5 e8g8 f1g2 c8g4 d1d2 e5f3 g2f3 g4f3 e1g1 f6d5 d2d3 f3g2 f1d1 d5c3 d3d8 f8d8 d1d8 a8d8 g1g2 c3d5 c1d2 f7f6 a1c1 d8d7 g2f3 g8f7 c1c5 g7g6 c5c4 d5b6 c4c2 b6d5 d2e3 d5e3 f3e3 g6g5 a2a4 h7h5 a4a5 f7g6 b2b4');
INSERT INTO evaluations VALUES(53486,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(53486,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(53486,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(53486,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(53486,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(53486,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(53486,36,'cp 76','d2d4 d8c7 a2a4 f8e7 h2h3 g8f6 e1g1 e8g8 f1e1 f8e8 c1e3 b8d7 d4d5 d7b6 c4b3 c6d5 e4d5 a7a5 d1d3 c8d7 f3d2 e7f8 c3b5 c7d8 a1d1 a8c8 c2c4 d7b5 a4b5 b6d7 b3a4 d7c5 e3c5 c8c5');
INSERT INTO evaluations VALUES(53494,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(53494,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(53494,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(53494,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(53494,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(53494,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(53494,36,'cp 71','d8c7 e1g1 g8f6 a2a4 f8e7 f3h4 a7a5 d4e5 d6e5 f2f4 e7d6 d1e1 e8g8 h4g6 f8e8 g6e5 d6e5 f4e5 c7e5 e1f2 b8d7 c1e3 f6g4 f2f7 g8h8 f7f4 g4e3 f4e3 d7f6 c4f7 f6g4 e3f4 e5d4 g1h1 e8f8 f4g3 g4e3 f1f3 c8g4 f3e3 f8f7 h2h3 g4e6 a1d1 d4b4 g3g6 f7e7 e3g3');
INSERT INTO evaluations VALUES(23432,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(23432,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(23432,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(23432,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(23432,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(23432,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23432,36,'cp 42','g8f6 g1f3 d7d6 b1c3 c8f5 e2e3 c6b4 a1c1 c7c6 a2a3 b4d5 c3d5 f6d5 f4g3 d5f6 f1d3 e7e6 e1g1 f8e7 f1e1 f5e4 c2c4 e8g8 d1c2 e4d3 c2d3 a7a5 d3b3 d8c8 c4c5 d6d5 f3e5 f6e4 e5d3 e7f6 g3e5 f6e5 d3e5 f8d8 b3c2 c8c7');
INSERT INTO evaluations VALUES(7390,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(7390,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(7390,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(7390,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(7390,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(7390,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(7390,36,'cp 25','c2c3 e8g8 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 a2a3 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 c2d3 f6h5 d3f1 h5f4 d1c2 f4e6 b3b4 a8c8 a1d1 e5d4 c3d4 c6c5 d4d5 e6g5 f3g5 h6g5 c2b1 c5c4 b1c2 d7e5 b2c3 d8d7');
INSERT INTO evaluations VALUES(7394,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(7394,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(7394,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(7394,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7394,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7394,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(7394,36,'cp 19','e8g8 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 a2a3 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 c2d3 f6h5 d1c2 h5f4 d3f1 a6a5 a1d1 f4e6 c3c4 b5b4 c4c5 e6d4 b2d4 e5d4 c5d6 f8d6 d2c4 d7e5 f3d4 e5c4 f1c4 d8c7 d4f5 d6f8 a3b4 a5b4 e1e3');
INSERT INTO evaluations VALUES(2724,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(2724,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(2724,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(2724,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(2724,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(2724,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(2724,36,'cp 24','a2a4 c8d7 d2d4 h7h6 h2h3 f8e8 c1d2 e7f8 b3c2 a8b8 d4d5 c6e7 b2b3 e7g6 c3c4 b5c4 b3c4 d8c8 e1e3 g6f4 f3e1 f6h5 a4a5 f8e7 e1d3 e7g5 e3f3 e8f8 b1c3 f4d3 c2d3 g5d2 d1d2 h5f4');
INSERT INTO evaluations VALUES(53265,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53265,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53265,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53265,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53265,36,'cp 0','e7e6 c1f4 f8d6 f4d6 d8d6 d4c5 d6c5 d1d4 c5d4 c3d4 b8c6 b1c3 c8d7 e2e3 e8e7 a1c1 g8f6 f1e2 a8c8 e1g1 f6e8 f3e5 e8d6 e5d3 c6a5 d3e5');
INSERT INTO evaluations VALUES(54208,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54208,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54208,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(54208,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54208,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54208,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54208,36,'cp 12','c1f4 f8d6 f4d6 d8d6 d4c5 d6c5 d1d4 c5d4 c3d4 e8e7 e2e3 b8c6 b1c3 c8d7 a1c1 g8f6 f1b5 c6b4 b5d7 f6d7 e1e2 h7h5 a2a3 b4c6 c1c2 d7b6 h1c1 h8c8 f3e5 a7a6 e5d3 c6a5 c3b1 c8c2 c1c2 a8c8 c2c8 b6c8 d3c5 c8d6 e2d3');
INSERT INTO evaluations VALUES(54190,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54190,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(54190,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(54190,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(54190,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(54190,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54190,36,'cp 9','f8d6 f4d6 d8d6 d4c5 d6c5 d1d4 c5d4 c3d4 b8c6 b1c3 g8f6 e2e3 c8d7 f1b5 c6b4 e1e2 d7b5 c3b5 e8e7 h1c1 h8c8 f3e5 f6e8 a2a3 b4c6 e5c6 b7c6 b5c3 c6c5 d4c5 c8c5 b2b4 c5c4 c3b5 a7a5 c1c4 d5c4 a1c1 a5b4 a3b4 a8c8 b5c3 e8d6 f2f3 c8a8 b4b5 e7d7 e2d2');
INSERT INTO evaluations VALUES(4316,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4316,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(4316,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(4316,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(4316,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4316,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4316,36,'cp 0','d2d3 c6a5 f3e5 d8e7 e5g4 a5c4 c1g5 c4b6 e4e5 d7d6 g5f6 g7f6 d3d4 c8g4 d1g4 g8h8 d4c5 e7e5 c3e2 f6f5 g4h4 d6c5 e1c1 e5e2 h1e1 e2c4 h4f6 h8g8 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(4320,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4320,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4320,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4320,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4320,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4320,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4320,36,'cp 0','c6a5 c4b3 d7d6 e1g1 a7a6 c3e2 f6h5 e2c3');
INSERT INTO evaluations VALUES(14775,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(14775,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(14775,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(14775,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(14775,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14775,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(14775,36,'cp 58','g1f3 f8d6 c4c5 d6e7 a1b1 a7a5 a2a3 e8g8 b2b4 a5b4 a3b4 f6e4 d1c2 e4c3 c2c3 e6e5 d4e5 f7f6 e5e6 c8e6 f1e2 e6f5 b1b2 f5e4 e1g1 d8d7 b2d2 d7g4 c3b3 c6e5 f3e1 g4e6 f2f3 e4f5 g2g4 f5g6 b3d5 e6d5 d2d5 a8a2 d5d2 f8a8 d2a2');
INSERT INTO evaluations VALUES(57567,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(57567,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(57567,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(57567,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(57567,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(57567,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(57567,36,'cp 152','e2e4 g8h6 d2d4 h6f7 c1e3 c7c6 f1e2 d7d5 e4d5 c6d5 c2c4 e7e6 b1c3 d5c4 e1g1 f8d6 e2c4 e8g8 a1c1 b8a6 c4b3 a6c7 d4d5 e6e5 f3d2 c7e8 d2c4 c8f5 c4d6 e8d6 f2f4 e5f4 e3f4 f7e5');
INSERT INTO evaluations VALUES(57561,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(57561,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(57561,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(57561,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(57561,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(57561,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(57561,36,'cp 153','g8h6 d2d4 h6f7 c1e3 c7c6 f1d3 e7e5 b1c3 d7d6 d4d5 f8e7 f3d2 e8g8 a2a4 d8e8 d1e2 e7d8 a4a5 c6d5 d3b5 c8d7 b5d7 e8d7 c3d5 b8c6 a5a6 c6e7 e2d3 b7a6 a1a6 f6f5 d5e7 d8e7 e4f5 d6d5 e1g1 d5d4 c2c3 f7d6 c3d4 d7f5 d3f5 d6f5 d4d5');
INSERT INTO evaluations VALUES(49781,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(49781,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(49781,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(49781,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(49781,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(49781,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(49781,36,'cp 75','e4e5 f6g8 g1f3 d7d6 d1e2 c8e6 d2d4 d6e5 d4d5 e6d5 e2e5 g8e7 c1f4 b8c6 e5e3 d5e6 a1d1 d8c8 e3e4 e7g6 f1b5 f8d6 b5c6 b7c6 e4c6 e6d7 c6e4 d7e6 f4d6 c7d6 e1d2 e8g8 d2c1 c8c5 h1e1 a8c8 a2a3 f8d8 d1d2 h7h6 e4d4 c5a5 d4b4 a5b4 a3b4 e6g4 c3b5 c8b8 c2c4 g4f3 g2f3 g6h4 b5d6');
INSERT INTO evaluations VALUES(49507,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(49507,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(49507,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(49507,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(49507,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(49507,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(49507,36,'cp 71','f6g8 g1f3 d7d6 d1e2 d6e5 d2d4 c8e6 d4d5 e6d5 e2e5 g8e7 c1f4 b8c6 e5e3 d5e6 a1d1 d8c8 e3e4 e7g6 f1b5 f8d6 b5c6 b7c6 e4c6 e6d7 c6e4 d7e6 f4d6 c7d6 e4a4 e6d7 a4d4 e8g8 e1d2 c8b7 d2c1 f8c8 d4d6 d7e6 d6d4 c8c4 d4d2 a8c8 h1e1 e6g4 d2d5 g4f3 d5f3');
INSERT INTO evaluations VALUES(57106,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(57106,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(57106,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(57106,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(57106,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(57106,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(57106,36,'cp 63','g1f3 d7d6 d2d4 d6e5 d1e2 f8e7 e2e5 g8f6 c1f4 c7c6 f1c4 b8d7 e5e2 d7b6 c4b3 a7a5 a2a4 b6d5 c3d5 f6d5 f4g3 e8g8 e1g1 c8g4 h2h3 g4f3 e2f3 e7d6 g3e5 d6e5 d4e5 d8b6 g1h1 a8e8 a1e1 e8e7 f3g3 b6b4 f1f5 f8e8 e1f1 b4e4 f1f3 e7e5 f5f7');
INSERT INTO evaluations VALUES(57105,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(57105,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(57105,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(57105,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(57105,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(57105,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(57105,36,'cp 78','d7d6 d2d4 d6e5 f1b5 c7c6 b5c4 e5d4 c1f4 d8e7 f3e5 f7f6 e1g1 f6e5 f4e5 d4c3 f1e1 b8d7 e5c3 g8f6 d1f3 e8d8 e1e7 f8e7 a1d1 h8e8 f3c6 e7b4 c6f3 b4c3 f3c3 e8e7 c4b5 a7a6 c3a5 d8e8 d1e1 e7e1 a5e1 e8d8 b5d3 g7g6 e1h4 a8a7 d3e4 b7b6 e4c6 d8e7 h4g3 e7e6 g3c3 c8b7 c3b3 e6e7');
INSERT INTO evaluations VALUES(55516,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(55516,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(55516,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(55516,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(55516,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(55516,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(55516,36,'cp 101','d2d4 d6e5 f1b5 c7c6 b5c4 e5d4 c1f4 d8e7 f3e5 f7f6 e1g1 f6e5 f4e5 d4c3 c4f7 e7f7 f1f7 e8f7 d1h5 f7e7 a1e1 e7d8 h5g5 g8e7 e5d6 c8d7 d6e7 f8e7 g5e7 d8c8 e7g7 h8d8 g7c3 a7a5 e1e7 h7h5 c3e5 a5a4 h2h4 a4a3 b2b4 b8a6 c2c3 a6c7 e5h5 c7d5 e7g7 d5c3 h5f7 a8a6 h4h5');
INSERT INTO evaluations VALUES(55511,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(55511,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(55511,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(55511,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(55511,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(55511,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(55511,36,'cp 93','d6e5 f1b5 c7c6 b5c4 e5d4 c1f4 d8e7 f3e5 f7f6 e1g1 d4c3 c4f7 e7f7 e5f7 e8f7 d1d8 c8e6 f4b8 f8e7 d8c7 g8h6 a1e1 h8e8 e1e6 f7e6 f1e1 e6f7 c7b7 c3b2 b8g3 h6f5 g1f1 f5g3 h2g3 f7f8 b7b2 e7c5 b2b7 e8e1 f1e1 a8e8 e1d1 e8d8 d1e2 d8d6 g3g4 c5b6 e2e1 d6e6 e1f1 e6e5 b7c6 h7h5 g4h5 e5h5 f1e2');
INSERT INTO evaluations VALUES(14228,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(14228,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(14228,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(14228,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(14228,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(14228,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14228,36,'cp 27','c8f5 g1f3 e7e6 c2c4 f8d6 f4g5 e8g8 a2a3 h7h6 g5h4 d6e7 f1e2 f6e4 h4e7 c6e7 d2e4 f5e4 e1g1 a7a5 d1b3 a5a4 b3b4 d5c4 b4c4 e7f5 c4c3 e4d5 a1c1 c7c6 f3d2 f5d6 f2f3 f7f5 e2d3 d6b5 c3c5 b7b6 c5c2 b5d6 c2c3 d8g5 c1e1 d6f7');
INSERT INTO evaluations VALUES(24825,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(24825,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(24825,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24825,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24825,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24825,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(24825,36,'cp 36','c2c4 e7e6 a2a3 f8d6 f4d6 c7d6 c4d5 e6d5 g1e2 e8g8 e2c3 c6e7 f1e2 a8c8 a1c1 d8d7 e1g1 a7a6 f1e1 f6e4 c3a2 c8c1 a2c1 e4f6 d2f1 f8c8 c1d3 h7h5 f2f3 d7c7 d1b3 c7d7 h2h3 g7g6 e1c1 c8c1 d3c1 f5b1');
INSERT INTO evaluations VALUES(46307,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46307,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46307,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46307,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(46307,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(46307,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(46307,36,'cp 13','f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 h7h6 g5h4 d8d6 h4g3 d6d8 b5c6 b7c6 f3e5 d8b6 a2a4 f8e8 a4a5 b6a6 f1e1 c8f5 h2h3 a8c8 f2f3 f6d7 g3f2 f7f6 e5g4 e8e1 d1e1 d7f8 f2g3 f8g6 e1d2 h6h5 g4e3 f5e6 g1h2 h5h4');
INSERT INTO evaluations VALUES(40637,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40637,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40637,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40637,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40637,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40637,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40637,36,'cp 25','c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 h7h6 g5h4 d8d6 h4g3 d6d8 f1e1 f6e4 b5c6 b7c6 f3e5 c8f5 e5c6 d8e8 c6e5 a8c8 d1f3 f5h7 a1c1 f7f6 e5d3 e8d7 h2h3 h7g6 g1h2 f8e8 d3f4 g6f7 e1e3 a7a5 f3g4');
INSERT INTO evaluations VALUES(40629,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40629,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40629,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40629,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40629,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40629,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(40629,36,'cp 22','e6d5 f1d3 h7h6 e1g1 e8g8 c1f4 c8g4 d1d2 f6h5 f4b8 g4f3 b8e5 f7f6 e5g3 f3e4 f1e1 h5g3 h2g3 e4d3 d2d3 d8d7 a2a4 b4c3 b2c3 f8e8 a4a5 e8e1 a1e1 b7b6 g1h2 b6a5 d3a6 a8b8');
INSERT INTO evaluations VALUES(27222,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(27222,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(27222,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(27222,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(27222,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(27222,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(27222,36,'cp 50','g7g6 b1c3 f8g7 d4c6 b7c6 b5d3 a8b8 d1e1 g8f6 a1b1 e8g8 f2f3 d8e7 c1g5 e7e5 g5d2 f6h5 f3f4 e5e7 e1f2 f7f5 b1e1 b8b2 e4f5 e7f7 f5g6 h7g6 c3e4 d7f5 e4g5 f7a2 d3f5 g6f5 f2f3 h5f6');
INSERT INTO evaluations VALUES(50358,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(50358,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(50358,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(50358,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(50358,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(50358,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(50358,36,'cp -26','d7d5 g2g3 g7g6 f1g2 f8g7 e1g1 c7c5 d2d3 e8g8 f3e5 f6g4 e5g4 c8g4 h2h3 g4e6 e2e4 d8d7 g1h2 b8c6 b1c3 g7c3 b2c3 d5e4 d3e4 d7d1 f1d1 a8c8 g3g4 f8d8 d1e1 e6c4 e4e5 c4e6 a2a4 b7b6 c1e3 e6d5 e1d1 d5g2 h2g2 d8d1');
INSERT INTO evaluations VALUES(21572,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(21572,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(21572,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(21572,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(21572,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(21572,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21572,36,'cp 32','d8d5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1d4 d5d4 c3d4 c8e6 f1e2 e8c8 e1c1 a7a6 h1e1 e7f5 d4c3 e4e3 f2f3 d8d1 e1d1 h7h5 d1e1 h8d8 e2d3 f5e7 d3e4 e7d5 e4d5 d8d5 e1e3 g7g6 e3e4 d5c5 a2a3 c5c4 e4c4 e6c4 c1d2 c4e6 b2b3');
INSERT INTO evaluations VALUES(1718,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(1718,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(1718,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(1718,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(1718,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(1718,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(1718,36,'cp 42','b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1d4 d5d4 c3d4 c8g4 h2h3 g4e6 e1c1 e8c8 f1e2 a7a6 h1e1 e7f5 d4c3 f5h4 e2f1 d8d1 e1d1 h4f5 g2g4 f5h4 d1d4 h7h5 d4e4 h5g4 h3g4 h4f3 c1d1 h8h1 d1e2 f3h2 f1g2 h1g1 g2f3 h2f3 e2f3 b7b5 e4f4');
INSERT INTO evaluations VALUES(1712,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(1712,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(1712,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(1712,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(1712,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(1712,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(1712,36,'cp 35','f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1d4 d5d4 c3d4 c8e6 f1e2 e7f5 e1c1 f5h4 g2g3 h4f3 b2b3 h7h5 e2f3 e4f3 d4e3 h5h4 g3g4 h4h3 d1d4 a8d8 d4e4 e8g8 h1g1 e6d5 e4f4 f8e8 c2c4 d5c6 f4f5');
INSERT INTO evaluations VALUES(180,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(180,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(180,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(180,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(180,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(180,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(180,36,'cp 41','c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1d4 d5d4 c3d4 c8e6 e1c1 e8c8 f1e2 a7a6 d4c3 g7g6 h1e1 h7h5 b2b3 d8d1 e2d1 e6d5 g2g3 h8d8 c3d2 b7b5 g3g4 h5g4 d1g4 e7f5 h2h4 d8e8 g4f5 g6f5 h4h5 c8d7 e1h1 e8h8 h5h6 d7e6');
INSERT INTO evaluations VALUES(181,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(181,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(181,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(181,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(181,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(181,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(181,36,'cp 36','b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1d4 d5d4 c3d4 c8e6 e1c1 e8c8 f1e2 d8d5 d4c3 h8e8 b2b3 d5d1 e2d1 c7c5 h1e1 e6d5 d1e2 c8c7 e2b5 e7c6 g2g3 h7h5 a2a4 e8d8 e5e6 f7e6 c3g7 d8g8 g7h6 c6d4 b5f1 c7d7');
INSERT INTO evaluations VALUES(184,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(184,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(184,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(184,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(184,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(184,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(184,36,'cp 33','d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1d4 d5d4 c3d4 c8e6 e1c1 e8c8 f1e2 d8d5 d4c3 h8e8 b2b3 d5d1 h1d1 e7d5 c3d4 e8d8 d4c5 d5c3 d1d8 c8d8 e2f1 b7b6 c5f8 g7g6 a2a4 d8d7 c1d2 c3b1 d2e3 b1c3 e3f4 c7c5 f1b5 d7c7 f4g5 c3d1 f2f3');
INSERT INTO evaluations VALUES(182,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(182,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(182,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(182,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(182,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(182,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(182,36,'cp 42','e5e4 f3e5 c6e5 d4e5 g8e7 d1d4 d5d4 c3d4 c8e6 e1c1 e8c8 f1e2 d8d5 d4c3 h8e8 b2b3 d5d1 e2d1 c7c5 h1e1 e6d5 d1e2 e7f5 e2g4 g7g6 e1e3 h7h5 g4f5 g6f5 e3h3 e8g8 g2g3 g8h8 e5e6 h8h7 e6f7 d5f7 c1d2 c8d7 d2e3 d7e6 e3f4 b7b5 f4g5');
INSERT INTO evaluations VALUES(18735,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(18735,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(18735,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(18735,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(18735,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(18735,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(18735,36,'cp -35','a7a6 a2a4 g8f6 g1f3 d7d5 c4b3 d5e4 c3e4 f8e7 e4f6 e7f6 c2c3 e8g8 e1g1 b7b6 c1f4 a8a7 d1e2 a7d7 f1e1 c8b7 a1d1 f8e8 f4g3 g7g6 b3c2 h7h6 h2h3 e8f8 f3d2 f6g7 d2c4 b6b5 a4b5 a6b5 c4e5 c6e5 g3e5');
INSERT INTO evaluations VALUES(17353,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(17353,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(17353,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(17353,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(17353,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(17353,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(17353,36,'cp -15','a2a4 g8f6 g1f3 d7d5 c4b3 d5e4 c3e4 b7b6 c1f4 f8e7 c2c3 f6h5 f4e3 d8c7 e1g1 e8g8 d3d4 h5f4 f1e1 c8b7 d4c5 b6c5 b3c2 f8d8 e3f4 c7f4 e4d2 c6a5 d1e2 a8b8 c2e4 f4c7 e4b7 b8b7 d2e4 c5c4 a1d1 h7h6 d1d8 c7d8');
INSERT INTO evaluations VALUES(17400,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(17400,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(17400,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(17400,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(17400,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(17400,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(17400,36,'cp -24','g8f6 g1f3 d7d5 c4b3 d5e4 c3e4 d8c7 d1d2 f8e7 e1g1 e8g8 d2f4 c7f4 c1f4 c6a5 f3d2 f6e4 d3e4 b7b5 c2c3 c8b7 b3c2 f8c8 a4b5 a6b5 h2h3 h7h6 e4e5 b7d5 f1e1 a5c6 d2e4 a8a1 e1a1');
INSERT INTO evaluations VALUES(11081,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(11081,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(11081,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(11081,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(11081,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(11081,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(11081,36,'cp -26','g1f3 d7d5 c4b3 d5e4 c3e4 f8e7 c2c3 e8g8 e1g1 b7b6 d1e2 c8b7 e4f6 e7f6 c1f4 b6b5 f3d2 c6e7 d2e4 e7g6 e4f6 d8f6 f4g3 g6f4 g3f4 f6f4 e2e3 f4d6 f1d1 f8d8 a4b5 a6b5 a1a8 d8a8 e3g3 d6b6 d3d4');
INSERT INTO evaluations VALUES(11080,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(11080,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(11080,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(11080,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(11080,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(11080,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(11080,36,'cp -23','d7d5 c4b3 d5e4 c3e4 f8e7 e1g1 e8g8 c2c3 b7b6 d1e2 c8b7 e4f6 e7f6 c1f4 b6b5 f1d1 b5b4 d3d4 b4c3 b2c3 c5d4 a1b1 d8e7 c3d4 f8d8 d4d5 e6d5 e2d2 a6a5 d1e1 e7b4 b3d5 b4d2 f4d2 d8d5 b1b7 a8d8 b7b5 h7h5 b5d5 d8d5 d2e3 g7g6 g2g3');
INSERT INTO evaluations VALUES(11203,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(11203,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(11203,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(11203,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(11203,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(11203,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(11203,36,'cp -16','c4b3 d5e4 c3e4 f8e7 e1g1 e8g8 d1e2 b7b6 c2c3 c8b7 e4f6 e7f6 c1f4 b6b5 a4b5 a6b5 a1a8 b7a8 f3d2 c6e7 d2e4 e7g6 e4f6 d8f6 f4g3 b5b4 f1c1 g6f4 g3f4 f6f4 e2e3 f4c7 d3d4 c5c4 b3d1 c7b7 d1f3 b7b5 c3b4 b5b4 e3c3 b4c3 c1c3 a8f3 c3f3');
INSERT INTO evaluations VALUES(42149,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(42149,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(42149,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(42149,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(42149,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(42149,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(42149,36,'cp 87','a2a3 b4c3 b2c3 e8g8 g1e2 c7c5 e1g1 b8c6 e2g3 d8c7 c1b2 f8d8 d1e2 b7b6 c4d3 c6e7 f2f3 c8b7 e2f2 e7g6 f1d1 c5c4 d3g6 h7g6 e3e4 f6h5 g3h5 g6h5 a3a4 f7f5 b2a3 d8d7 e4f5 a8d8 f2h4 e6f5 h4h5');
INSERT INTO evaluations VALUES(42157,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(42157,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(42157,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(42157,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(42157,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(42157,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(42157,36,'cp 81','b4c3 b2c3 c7c5 g1e2 e8g8 e1g1 b8c6 e2g3 b7b6 c1b2 c8b7 c4d3 a8c8 f1e1 h7h6 a1c1 c6e7 c3c4 e7g6 a3a4 f8e8 h2h3 c8c7 a4a5 b6a5 d4d5 e6d5 b2f6 d8f6 c4d5 g6e5 d3e2 c7c8 e3e4 b7a6 e2a6 f6a6 g3f5 a6f6 d1e2');
INSERT INTO evaluations VALUES(42183,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(42183,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(42183,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(42183,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(42183,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(42183,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(42183,36,'cp 70','b2c3 c7c5 g1e2 b8c6 e1g1 e8g8 e2g3 d8c7 c1b2 f8d8 d1e2 c6e7 c4d3 c8d7 a3a4 a8c8 a4a5 c5c4 d3c2 f6d5 b2a3 e7g6 e2d2 b7b5 e3e4 d5f4 e4e5 d7c6 f2f3 d8d7 a5a6 c6d5 d2e3 c8a8 g3e4 d5e4 f3e4');
INSERT INTO evaluations VALUES(42182,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(42182,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(42182,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(42182,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(42182,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(42182,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(42182,36,'cp 67','e8g8 g1e2 c7c5 e1g1 b7b6 e2g3 d8c7 c4d3 c8a6 d3a6 b8a6 e3e4 a6b8 c1g5 b8d7 e4e5 f6d5 d1g4 c5d4 c3d4 g8h8 g3e4 c7c2 a1e1 c2d3 e4d6 h7h6 g5h4 f7f5 e5f6 d7f6 g4e6 d3d4 h4g3 d4g4 f2f3 g4e6 e1e6 h8g8 h2h3 f8d8 f1d1 d5c3 d1d2');
INSERT INTO evaluations VALUES(37090,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(37090,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(37090,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(37090,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(37090,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(37090,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(37090,36,'cp 89','g1e2 c7c5 e1g1 b8c6 e2g3 b7b6 c4d3 c8b7 c1b2 d8e7 c3c4 c6a5 d1e2 f8d8 a1d1 a8c8 f1e1 g7g6 h2h3 f6e8 b2c3 b7c6 d1b1 e7g5 c3a5 b6a5 d3e4 c5d4 e3d4 c6a4 e2d3 g5f6 d4d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(32851,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(32851,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(32851,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(32851,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(32851,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(32851,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(32851,36,'cp 0','b8c6 e1g1 e8c8 h2h3 g4h5 b1c3 d5d7 d2d3 g8f6 f1e1 h8g8 c1e3 h7h6 c3b5 g7g5 d3d4 d7f5 e2d3 f5d7 d3e2');
INSERT INTO evaluations VALUES(27623,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(27623,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(27623,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(27623,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(27623,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(27623,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(27623,36,'cp 13','d2d4 e8c8 c1e3 e7e5 b1c3 d5a5 f3e5 g4e2 d1e2 c6e5 d4e5 a5e5 a1d1 d8d1 e1d1 g8f6 e2c4 f8d6 d1c1 a7a6 h2h3 h8f8 a2a4 e5e6 c4e6 f7e6 f2f3 f6d5 e3d2 d5c3 d2c3 c8d7 h1e1 f8f4 b2b3 g7g6 c3d2 f4h4 c2c4 d6c5 e1d1 c5d6 a4a5 e6e5 d1e1 d7e6 c1c2 h7h5');
INSERT INTO evaluations VALUES(5968,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5968,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(5968,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(5968,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5968,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5968,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5968,36,'cp 3','g8f6 c3a4 c5b6 a4b6 a7b6 d2d3 e8g8 a2a3 h7h6 f1e1 c8e6 c4e6 f7e6 c2c3 d8e8 c1e3 e8h5 a3a4 f8f7 f3d2 h5h4 h2h3 a8f8 b2b4 f6h5 d1g4 h4g4 h3g4 h5f4 e3f4 f7f4 f2f3 f4f7 d2c4 f8a8 c4e3 g8h7 b4b5 c6e7 d3d4');
INSERT INTO evaluations VALUES(16448,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(16448,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(16448,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(16448,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(16448,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16448,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16448,36,'cp 0','f8b4 d1b3 d7d6 f1e2 a7a5 e1g1 e8g8 d2d3 f8e8 c1d2 c8f5 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(10410,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10410,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10410,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10410,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10410,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10410,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10410,36,'cp 0','d1c2 b4c3 c2c3 d8e7 d2d4 e5e4 f3g1 d7d5 b2b3 h7h5 a2a3 h5h4 h2h3 g7g5 g1e2 c8e6 c1b2 h8h6 c3c2 c6a5 b2c3 a5c6');
INSERT INTO evaluations VALUES(10411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10411,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10411,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10411,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10411,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10411,36,'cp 0','b4c3 c2c3 d8e7 d2d4 e5e4 f3g1 d7d5 b2b3 h7h5 g1e2 h5h4 h2h3 g7g5 c1b2 c8e6 a2a3 h8h6 c3c2 c6a5 b2c3 a5c6');
INSERT INTO evaluations VALUES(24848,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24848,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(24848,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24848,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(24848,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(24848,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24848,36,'cp 36','f1b5 d8d6 f3e5 f6d7 b5d3 f5d3 e5d3 a7a6 e1g1 d7f6 b2b3 e7e5 d4e5 c6e5 c1b2 e5d3 c2d3 f8e7 c3e2 e8g8 e2g3 d6e6 d1f3 c7c6 g3f5 f8e8 f5e7 e8e7 b2f6 e6f6 f3f6 g7f6 d3d4 g8g7 a1c1 e7e4 g2g3 a6a5 c1c5 a5a4 g1g2 a4b3 a2b3');
INSERT INTO evaluations VALUES(24778,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(24778,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(24778,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24778,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24778,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24778,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24778,36,'cp 24','d8d6 f3e5 f6d7 e5c6 b7c6 b5d3 e7e6 e1g1 f5d3 c2d3 c6c5 d4c5 d6c5 c1d2 f8d6 a1c1 c5b6 d1a4 c7c6 e3e4 e8g8 e4d5 d7c5 a4c6 c5d3 c1b1 b6c6 d5c6 a8c8 d2e3 c8c6 f1d1 d3b2 b1b2 d6e5 b2b7 e5c3 b7a7 c3b4 a2a4 b4c5 e3c5 c6c5 a4a5 g7g5 a5a6 c5a5');
INSERT INTO evaluations VALUES(48637,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48637,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48637,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48637,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(48637,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(48637,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(48637,36,'cp -21','f8b4 b1c3 e8g8 f1e2 b8c6 c1e3 d5c4 e2c4 c8g4 e1g1 d8d6 c4e2 b4a5 a1c1 a5b6 h2h3 g4h5 d4d5 b6e3 f2e3 h5f3 e2f3 c6e5 c3e4 f6e4 f3e4 g7g6 d1d4 a7a6 c1c3 f8e8 e4b1 e8e7 e3e4 a8e8 d4a7 b7b5 a7d4 e5d7 d4f2');
INSERT INTO evaluations VALUES(42253,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(42253,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(42253,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(42253,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(42253,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(42253,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(42253,36,'cp -11','b1c3 e8g8 f1e2 d5c4 e2c4 c8g4 e1g1 b8c6 c1g5 h7h6 g5h4 b4a5 a1c1 g4f3 d1f3 d8d4 h4f6 d4f6 f3f6 g7f6 c3e4 g8g7 e4g3 a8d8 c1d1 a5b6 g3f5 g7g6 f5h4 g6h7 g2g3 c6d4 g1g2 c7c6 d1d3 b6c5 a2a3 b7b5 c4a2 c5b6 f1d1 f6f5');
INSERT INTO evaluations VALUES(5941,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(5941,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(5941,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(5941,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(5941,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(5941,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(5941,36,'cp 21','c3a4 c8d7 a4c5 d6c5 b5c6 d7c6 f3e5 c6e4 d1e2 g8f6 f2f3 d8e7 e5f7 e8f7 f3e4 e7e4 e2e4 f6e4 d2d3 e4f6 c1f4 c5c4 e1c1 c4d3 d1d3 c7c6 h1f1 h8e8 f4e3 b7b6 f1f3 a8d8 d3d8 e8d8 b2b3 f7e6 a2a4 d8d5 e3f4 b6b5 a4b5 c6b5');
INSERT INTO evaluations VALUES(22602,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(22602,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(22602,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(22602,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(22602,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(22602,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(22602,36,'cp 49','e2e3 g8f6 g1f3 b7b6 a2a3 c8b7 c2c4 f6h5 f4g5 f8e7 g5e7 c6e7 f1e2 e8g8 e1g1 d7d6 b1c3 h5f6 d1d3 c7c5 f1d1 e7g6 c3b5 a7a6 b5c3 a8b8 b2b4 d8e7 d4c5 d6c5 d3d6 b7f3 g2f3 e7d6 d1d6 f8c8 a1d1 c5b4');
INSERT INTO evaluations VALUES(22600,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(22600,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(22600,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(22600,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(22600,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(22600,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(22600,36,'cp 29','g8f6 g1f3 h7h6 c2c3 d7d6 h2h3 g7g6 b1d2 f8g7 f4h2 e8g8 d1c2 f6d7 a2a4 e6e5 f1e2 a7a5 e1g1 d8e7 e2b5 d7b6 d2c4 b6c4 b5c4 g8h8 g1h1 c8d7 a1e1 f7f5 f3d2 a8d8 f2f3 g6g5 d4d5 c6b8 c4d3');
INSERT INTO evaluations VALUES(16284,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(16284,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(16284,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(16284,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(16284,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(16284,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(16284,36,'cp 50','g1f3 h7h6 c2c3 g7g6 h2h3 f8g7 b1d2 e8g8 f4h2 d8e7 f1c4 d7d6 d1c2 f6d7 a2a4 f7f5 e1g1 g8h7 a4a5 a7a6 b2b4 e6e5 d4e5 d7e5 f3e5 c6e5 c4d5 c7c6 d5c4 h7h8 a1d1 c8d7 f1e1 a8c8 c4e2');
INSERT INTO evaluations VALUES(16285,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(16285,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(16285,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(16285,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(16285,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(16285,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(16285,36,'cp 32','h7h6 c2c3 g7g6 h2h3 d7d6 f4h2 f8g7 f1c4 d8e7 a2a4 e6e5 a4a5 a7a6 b1d2 e5e4 f3g1 d6d5 c4b3 e8g8 g1e2 b7b6 a5b6 c7b6 e1g1 f8d8 e2g3 h6h5 f2f3 c6a5 f3e4 a5b3 d1b3 d5e4 g3e2 b6b5 b3a3 e7a3 a1a3');
INSERT INTO evaluations VALUES(30888,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(30888,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(30888,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(30888,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(30888,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(30888,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(30888,36,'cp -40','b5d7 d8d7 c2c3 b8c6 g1e2 f7f6 f2f4 c5d4 c3d4 g8h6 e1g1 f8e7 b1d2 f6e5 f4e5 h6f5 d2f3 e8g8 c1d2 a7a5 a1c1 f8f7 d1c2 a8f8 c2d3 f5h4 f3h4 e7h4 g2g3 h4d8 d2e3 a5a4 a2a3 d8b6 f1f4 f7f4 e2f4');
INSERT INTO evaluations VALUES(30857,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(30857,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(30857,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(30857,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(30857,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(30857,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(30857,36,'cp -35','d8d7 c2c3 b8c6 g1f3 g8e7 b1d2 c5d4 c3d4 f7f6 e1g1 f6e5 f3e5 c6e5 d4e5 e7c6 d2f3 f8e7 c1e3 e8g8 a1c1 f8f7 a2a3 a8d8 d1b3 f7f3 g2f3 c6e5 e3f4 e5c4 f1e1 e7f6 a3a4 h7h5 c1c4 d5c4 b3c4 d7d5 c4e4 g8f7 b2b4 h5h4');
INSERT INTO evaluations VALUES(32753,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(32753,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(32753,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(32753,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(32753,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(32753,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(32753,36,'cp -31','c2c3 b8c6 g1f3 g8e7 b1d2 c5d4 c3d4 f7f6 e1g1 f6e5 f3e5 c6e5 d4e5 e7c6 d2f3 f8e7 c1e3 e8g8 a1c1 f8f5 d1a4 a8d8 a4b3 h7h6 f1d1 e7g5 e3g5 h6g5 h2h3 d8f8 b3e3 f5f3 g2f3 f8f5 g1g2 c6e5 e3a7 e5g6');
INSERT INTO evaluations VALUES(32751,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(32751,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(32751,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(32751,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(32751,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(32751,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(32751,36,'cp -31','b8c6 g1f3 g8e7 b1d2 c5d4 c3d4 f7f6 e1g1 f6e5 f3e5 c6e5 d4e5 e7f5 d2b3 f8e7 b3d4 e8g8 c1e3 b7b6 a1c1 f5d4 d1d4 e7c5 d4d3 a8c8 h2h3 a7a5 b2b3 h7h6 e3c5 b6c5 f2f4 a5a4 g1h2 f8f5 f1f3 a4b3 a2b3 d7f7 d3e3 f7c7 e3d2');
INSERT INTO evaluations VALUES(27566,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(27566,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(27566,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(27566,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(27566,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(27566,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(27566,36,'cp -27','g1f3 g8e7 b1d2 c5d4 c3d4 f7f6 d2b3 f6e5 f3e5 c6e5 d4e5 e7c6 f2f4 a7a5 e1g1 a5a4 b3d4 f8c5 c1e3 e8g8 a1c1 b7b6 h2h3 c6d4 e3d4 a8c8 b2b3 a4b3 a2b3 d7f7 g1h2 f7g6 d4c5 b6c5 d1g4 g6f7 c1c2 h7h6 c2c3 f7f5 g4f5 f8f5 g2g4');
INSERT INTO evaluations VALUES(27565,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(27565,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(27565,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(27565,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(27565,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(27565,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(27565,36,'cp -52','g8e7 b1d2 c5d4 c3d4 f7f6 e1g1 f6e5 f3e5 c6e5 d4e5 e7c6 d2f3 f8e7 c1e3 e8g8 a1c1 f8f5 d1a4 a8d8 b2b3 c6e5 f3d4 d7a4 b3a4 f5f6 e3g5 f6f7 d4e6 e7g5 e6g5 f7f6 c1e1 e5c6 g5e6 d8d6 e6c5 b7b6 c5d3 g8f7 e1c1 d6e6 f1e1 d5d4 e1e6 f6e6 f2f4');
INSERT INTO evaluations VALUES(38412,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(38412,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(38412,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(38412,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(38412,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(38412,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(38412,36,'cp 49','g8f6 f1d3 e8g8 e1g1 b8d7 f1e1 e7e5 h2h3 a7a5 d3c2 f8e8 d4d5 d7c5 c3c4 c7c6 b1c3 c8d7 d5c6 d7c6 b2b3 b7b6 a1b1 d8d7 c1g5 h7h6 g5c1 d7b7 f3d2 b7c7 c1a3 a8d8 c3d5 f6d5 c4d5 c6d7');
INSERT INTO evaluations VALUES(41373,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(41373,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(41373,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(41373,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(41373,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(41373,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(41373,36,'cp 61','f1d3 e8g8 e1g1 b8d7 c1e3 b7b6 b1d2 e7e5 f1e1 h7h6 h2h3 c8b7 d1c2 f8e8 a1d1 a7a5 d4d5 c7c6 c3c4 d7c5 d3f1 a8c8 c2b1 b7a6 b2b3 c6d5 e4d5 b6b5 c4b5 a6b7 d1c1 f6d5 e3c5 c8c5 c1c5');
INSERT INTO evaluations VALUES(41374,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(41374,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(41374,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(41374,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(41374,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(41374,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(41374,36,'cp 51','b8d7 e1g1 e7e5 c1e3 b7b6 f1e1 e8g8 e3g5 h7h6 g5h4 d8e8 b1d2 a7a6 d1c2 f6h5 a2a4 g8h8 a1d1 c8b7 d4e5 d6e5 b2b4 h5f4 d3c4 d7f6 h4f6 g7f6 g2g3 f4e6 c4d5 b7c6 d1a1 a8d8 d2c4 c6d5 e4d5 d8d5 c4e5 h8g7 e5g4 e8c6');
INSERT INTO evaluations VALUES(11767,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(11767,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(11767,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(11767,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(11767,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(11767,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(11767,36,'cp 110','a4b3 e7g6 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 b1c3 b5b4 c3d5 f8d6 d2d3 c6a5 b3a2 b4b3 c2b3 a8d5 e4d5 e8g8 d3d4 e5e4 f3g5 h7h6 b3b4 a5b3 a2b3 h6g5 g2g3 d8f6 b4b5 f6f5 f2f3 d6b4 f3e4 f5e4 c1g5 b4e7 g5e7');
INSERT INTO evaluations VALUES(52170,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(52170,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(52170,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(52170,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(52170,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(52170,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(52170,36,'cp 94','e4d5 d7d6 g1f3 b8d7 b1c3 g8f6 f1d3 f8e7 e1g1 e8g8 f1e1 h7h6 c1f4 f6h5 f4e3 f8e8 h2h3 a7a6 a2a4 h5f6 e3f4 d7b6 d3f1 c8f5 g2g4 f5g6 f3h4 g6h7 h4g2 e7f8 e1e8 d8e8 a4a5 b6d7 a1a3 f6e4 a3b3 e4c3 b2c3 a8b8 c3c4');
INSERT INTO evaluations VALUES(52160,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(52160,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(52160,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(52160,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(52160,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(52160,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(52160,36,'cp 95','d7d6 g1f3 b8d7 f1d3 g8e7 d1e2 g7g6 e1g1 f8g7 f1e1 d7e5 f3e5 d6e5 b1c3 e8g8 c1g5 f7f6 g5e3 b7b6 a1d1 e7f5 d3f5 c8f5 a2a4 f8f7 f2f3 g6g5 a4a5 d8d7 h2h3 g7f8 e3f2 f5g6 c3e4 a8d8 e2c4 h7h5 c4a6 g6e4 e1e4 b6b5 b2b3 f8d6 c2c4 d8b8');
INSERT INTO evaluations VALUES(53921,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(53921,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(53921,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(53921,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(53921,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(53921,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(53921,36,'cp 98','g1f3 b8d7 b1c3 g8f6 f1d3 f8e7 e1g1 a7a6 a2a4 d7b6 d3e2 e8g8 h2h3 f8e8 f1e1 e7f8 c1g5 h7h6 g5h4 c8f5 a4a5 b6d7 f3d2 b7b5 a5b6 d7b6 d2f1 a6a5 f1e3 f5d7 e2d3 a5a4 d1f3 d8e7 d3f5 d7f5 f3f5 e7d7 f5d7 f6d7');
INSERT INTO evaluations VALUES(53919,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(53919,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(53919,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(53919,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(53919,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(53919,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(53919,36,'cp 91','b8d7 f1d3 g8e7 d1e2 g7g6 e1g1 f8g7 f1e1 d7e5 f3e5 d6e5 b1c3 e8g8 c1g5 f7f6 g5e3 b7b6 a1d1 e7f5 d3f5 g6f5 f2f4 f8f7 e2h5 d8d7 f4e5 f6e5 e3h6 g7f6 h6g5 f6g7 d5d6 c8b7 c3d5 b7d5 d1d5 d7e6 e1d1 f7d7 c2c4 e6g6 h5h4 a8e8 g5e7 g8f7 d1e1');
INSERT INTO evaluations VALUES(12782,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(12782,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(12782,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(12782,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(12782,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(12782,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(12782,36,'cp -5','e7e6 e1g1 d7d5 c4b5 c8d7 f1e1 f8e7 c2c4 d5c4 d3c4 c6b4 b5d7 d8d7 b1c3 d7d1 e1d1 a7a6 e4e5 f6d7 c3e4 b4c6 c1f4 e8c8 b2b3 h7h6 h2h3 d7b6 f4e3 b6d7');
INSERT INTO evaluations VALUES(22793,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(22793,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22793,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(22793,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(22793,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(22793,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(22793,36,'cp -20','g8f6 e1g1 f8c5 b1c3 e8g8 d3c4 d7d6 d2d3 c8g4 h2h3 g4h5 c1g5 h7h6 g5e3 c5e3 f2e3 c6a5 c4b3 a5b3 a2b3 c7c6 d1e1 h5f3 f1f3 f6h7 f3f5 a7a5 e1g3 b7b5 a1f1 d8e7 c3e2 h7g5 g1h2 a8d8 g3f2 a5a4 b3a4 b5a4 f2g3 g7g6');
INSERT INTO evaluations VALUES(34670,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(34670,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(34670,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(34670,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(34670,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(34670,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(34670,36,'cp 44','g8f6 g1f3 b8c6 h2h3 c6e5 b2b4 e5f3 d1f3 a5f5 f3e2 e7e6 a2a3 c8d7 d2d4 f5h5 e2h5 f6h5 e1g1 h5f6 f1e1 c7c6 c4d3 f8e7 c1d2 e8g8 c3e4 f6e4 d3e4 f8d8 c2c3 h7h6 a1b1 g8f8 a3a4 d7e8 a4a5 e7d6 e4d3 a8c8');
INSERT INTO evaluations VALUES(33727,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(33727,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(33727,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(33727,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(33727,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(33727,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(33727,36,'cp 67','d2d4 c8g4 f2f3 g4d7 c1d2 a5b6 g1e2 e7e6 e1g1 f8e7 a2a4 a7a5 g1h1 e8g8 c3e4 d7c6 c4d3 f8d8 b2b3 f6d5 c2c3 b8d7 d1c2 d7f8 a1d1 d8d7 e4g5 d5f6 d2c1 f8g6 c3c4 a8e8 g5e4 e8d8 c1g5 g6f8 e4f6 e7f6 g5f6 g7f6');
INSERT INTO evaluations VALUES(33729,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(33729,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(33729,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(33729,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(33729,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(33729,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(33729,36,'cp 62','c8g4 f2f3 g4d7 c1d2 a5b6 g1e2 e7e6 a2a4 a7a5 e1g1 f8e7 g1h1 d7c6 d2e3 e8g8 e3g1 f8d8 b2b3 b8a6 e2g3 c6d5 c3d5 e6d5 c4d3 a6b4 d3e2 b4c6 c2c3 e7f8 e2d3 c6e7 d1c2 c7c6 g3e2 d8e8 a1e1 e7g6 g2g3 b6c7');
INSERT INTO evaluations VALUES(32646,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32646,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(32646,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(32646,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(32646,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(32646,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(32646,36,'cp 47','f2f3 g4d7 c1d2 a5b6 g1e2 e7e6 d2e3 a7a6 c4b3 b8c6 d1d2 c6a5 d4d5 a5b3 a2b3 b6d6 e3f4 d6b6 d5e6 f7e6 e1c1 e8c8 c1b1 f8c5 d2e1 c5f2 e1f1 d7b5 c3b5 d8d1 f1d1 a6b5 e2c1 f6d5 f4d2 h8d8 c1d3 f2d4 b3b4 b6d6 d1e2 e6e5 h1e1 d8e8');
INSERT INTO evaluations VALUES(44102,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(44102,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44102,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44102,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(44102,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(44102,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(44102,36,'cp 20','d2d4 e5e4 d4d5 f8b4 c1d2 d8e7 b1c3 e8g8 g1h3 c6d5 c3d5 f6d5 c4d5 b4d2 d1d2 d7d6 h3g5 f7f5 h2h4 a7a5 a1c1 e7f6 g5h3 b8a6 h3f4 c8d7 e1g1 a6c5 b2b3 b7b5 c1c2 b5b4 f1c1 f8b8 f4e6 c5e6 d5e6 d7e6 d2d6 b8d8');
INSERT INTO evaluations VALUES(44106,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(44106,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44106,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(44106,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(44106,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(44106,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44106,36,'cp 7','e5e4 d4d5 f8b4 c1d2 d8e7 b1c3 e8g8 g1h3 c6d5 c3d5 f6d5 c4d5 b4d2 d1d2 d7d6 h3g5 f7f5 h2h4 h7h6 g5h3 b8d7 a1c1 d7e5 e1g1 b7b6 f2f3 e4f3 e2f3 c8d7 f1e1 e7f6 b2b3 a8c8 h3f4 c8c5 f4e6 c5c1 e1c1 f8c8 c1c8 d7c8 e6d4 f6g6 f3f4 e5g4');
INSERT INTO evaluations VALUES(44121,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44121,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44121,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(44121,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44121,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44121,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(44121,36,'cp 14','d4d5 f8b4 c1d2 d8e7 b1c3 e8g8 g1h3 c6d5 c3d5 f6d5 c4d5 b4d2 d1d2 d7d6 h3g5 f7f5 h2h4 h7h6 g5h3 b8d7 a1c1 d7e5 h3f4 b7b6 h4h5 c8d7 g2h3 a8c8 e1g1 g8f7 g1g2 e7g5 d2b4 g5e7 b4b3 f8e8 b3a3 a7a5 a3b3 c8c5 b3b6 e5c4 b6a7 e8c8 a7a6 e7g5 f1d1');
INSERT INTO evaluations VALUES(53632,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53632,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53632,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53632,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53632,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53632,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53632,36,'cp 3','c6c5 g1f3 b8c6 a2a3 c8d7 c3e2 d8b6 c2c3 a7a5 a1b1 a5a4 c1e3 g8h6 e3h6 g7h6 g2g3 c5d4 e2d4 f8g7 d4c6 b7c6 f1d3 e8g8 d1c2 f7f5 e5f6 g7f6 e1g1 f6g7 c2e2 c6c5 b1e1 c5c4 d3c2 b6b2');
INSERT INTO evaluations VALUES(54289,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54289,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54289,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54289,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(54289,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54289,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54289,36,'cp 3','g1f3 b8c6 f1d3 c5d4 c3b5 f8c5 a2a3 a7a6 b2b4 a6b5 b4c5 g8e7 e1g1 b5b4 c1b2 b4a3 b2a3 e8g8 a3b2 a8a1 d1a1 d8c7 f3d4 c6d4 b2d4 e7c6 f1b1 c6d4 a1d4 c8d7 b1b6 h7h6 h2h3 f8a8 g1h2 a8c8 f2f4 c7c5 d4c5 c8c5 b6b7 d7a4 f4f5');
INSERT INTO evaluations VALUES(54286,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54286,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54286,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54286,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54286,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54286,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54286,36,'cp 16','b8c6 d4c5 f8c5 f1d3 a7a6 c1f4 h7h6 e1g1 g8e7 a2a3 c8d7 f4g3 e8g8 f1e1 d8c7 c3e2 e7g6 e2f4 g6f4 g3f4 f7f6 f4g3 c6e5 f3e5 f6e5 d1h5 d7b5 d3b5 a6b5 e1e5 c5d4 e5e6 c7c2 e6e2 c2f5 h5f5 f8f5 a1e1 a8a4 g3d6 d4f6 d6e7 d5d4 e7f6');
INSERT INTO evaluations VALUES(20329,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20329,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(20329,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20329,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20329,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20329,36,'cp 12','d4c5 f8c5 f1d3 a7a6 c1f4 h7h6 a2a3 g8e7 e1g1 c8d7 f1e1 e8g8 c3e2 d8b6 e1f1 b6c7 a1c1 b7b5 h2h4 e7g6 d3g6 f7g6 c2c3 f8f7 f4g3 a8f8 e2d4 c6e7 d1d2 c5a7 f1e1 c7b6 b2b4 b6c7 e1e2 e7f5 d4f5 g6f5 g3f4');
INSERT INTO evaluations VALUES(37179,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(37179,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(37179,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(37179,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(37179,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37179,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(37179,36,'cp 21','e2e3 b7b6 g1f3 c7c5 f1d3 d7d5 c4d5 e6d5 h2h3 f8e8 a2a3 b4a5 d4c5 b6c5 e1g1 a5b6 c3a4 c8a6 d3a6 b8a6 b2b4 c5c4 d1c2 a6c7 f1d1 c7e6 a4b6 a7b6 d2e1 h7h6 a3a4 f6e4 f3d4 d8d7 d4e2');
INSERT INTO evaluations VALUES(37172,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37172,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37172,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37172,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(37172,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(37172,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37172,36,'cp -5','b7b6 g1f3 d7d5 c4d5 e6d5 a1c1 c7c5 d4c5 b6c5 f1b5 a7a6 b5a4 d5d4 e3d4 c5d4 c3e4 d8e7 d2b4 e7b4 e4d2 c8d7 a4d7 b8d7 e1g1 b4b2 c1c2 b2b4 c2c4 b4a5 c4a4 a5b6 a4d4 d7e5 d1a4 e5f3 d2f3 f8c8 f1d1 h7h6 d4d6 b6b5 a4b3 b5b3 a2b3 f6e4 d6d7');
INSERT INTO evaluations VALUES(17234,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17234,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(17234,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17234,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(17234,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17234,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17234,36,'cp 0','d2d3 e7e6 b1d2 d7d5 c4b5 a7a6 b5c6 b7c6 g1f3 a6a5 a2a4 d5e4 d2e4 f6e4 d3e4 d8d1 e1d1 c8b7 c1f4 f8d6 f4d6 c7d6 f3d2 b7a6 a1a3 f7f6 d1c1 c6c5 d2b3 a6c8 b3d2');
INSERT INTO evaluations VALUES(17231,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17231,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17231,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17231,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17231,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17231,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17231,36,'cp 0','e7e6 b1d2 d7d5 c4b5 c8d7 g1f3 d5e4 d3e4 a7a6 b5e2 g7g5 d2c4 f6e4 e2d3 e4c5 c1g5 f7f6 f3e5 c5d3 c2d3 f8b4 e1f1 e8g8 g5h6 f6e5 d1g4 g8f7 g4h5 f7g8');
INSERT INTO evaluations VALUES(16259,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(16259,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16259,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(16259,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(16259,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(16259,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(16259,36,'cp -12','b1d2 d7d5 c4b3 a7a5 b3a4 c8d7 g1f3 d5e4 d3e4 c6d4 a4d7 d8d7 e4e5 d4f3 d2f3 f6d5 c2c4 d5b6 d1d7 b6d7 e1e2 f8c5 c1f4 e8c8 f4d2 b7b6 a1d1 h8f8 b2b3 f7f6 e5f6 g7f6 g2g3 f8g8 d2c3 c5b4 c3d2 e6e5 d2e3 h7h5 d1d5 d8e8');
INSERT INTO evaluations VALUES(22910,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(22910,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(22910,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(22910,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(22910,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(22910,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22910,36,'cp -34','g8f6 b1c3 f8b4 c1g5 d8e7 g5f6 e7f6 e1c1 b4c3 c4c3 f6c3 b2c3 d7d6 f2f4 e8g8 g1f3 f8e8 f1d3 c8d7 f3d4 g8f8 d1e1 c6d8 g2g3 d8e6 e4e5 d6e5 e1e5 g7g6 d4e6 e8e6 e5e6 d7e6 c3c4 a8e8 h1e1 b7b6 c1d2 e6d7 e1b1 h7h5 b1c1 f8e7 c1e1 d7e6');
INSERT INTO evaluations VALUES(16671,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(16671,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(16671,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16671,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(16671,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(16671,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(16671,36,'cp -22','b1c3 f8b4 g1e2 d7d6 c1g5 c8e6 c4d3 h7h6 g5h4 b4c5 e1c1 c6e5 d3d2 f6e4 c3e4 d8h4 e4c5 d6c5 e2f4 e8g8 f4e6 f7e6 d2e3 h4f4 d1e1 f4e3 e1e3 e5g4 e3e6 g4f2 h1g1 a8e8 f1c4 e8e6 c4e6 g8h8 g1e1 f8e8 e1e2 b7b5 c1d2 e8d8 d2e1 f2d1 c2c3 c5c4 e6f5 h8g8 e2e7 d1b2 e7c7');
INSERT INTO evaluations VALUES(16668,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(16668,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(16668,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(16668,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(16668,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(16668,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(16668,36,'cp -20','f8b4 g1e2 d7d6 c1g5 h7h6 g5h4 c8e6 c4d3 b4c5 e1c1 c6e5 d3d2 f6e4 c3e4 d8h4 e4c5 d6c5 e2f4 a8d8 f1b5 c7c6 d2e3 h4f6 d1d8 e8d8 h1d1 e5d7 f4e6 f6e6 e3f4 d8c8 b5c4 e6f6 f4f6 d7f6 c4f7 h8f8 f7e6 c8c7 b2b3 f8e8 e6c4 b7b5 c4d3 f6d5 c1d2 e8f8 f2f3 d5f4 d1e1 f4g2 e1e7 c7b6 e7g7');
INSERT INTO evaluations VALUES(10513,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(10513,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10513,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(10513,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(10513,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(10513,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(10513,36,'cp -8','g1e2 d7d5 e4d5 f6d5 c1d2 d5c3 e2c3 d8e7 c4e4 e7e4 c3e4 c8f5 f1d3 e8c8 e1c1 f5g6 a2a3 b4d2 d1d2 h8e8 e4c5 g6d3 c5d3 c6e5 d3e5 e8e5 h2h4 d8e8 h1d1 e5e2 c2c3 e2d2 d1d2 e8e1 c1c2 e1h1 g2g3 h1h2 c2b3');
INSERT INTO evaluations VALUES(36563,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(36563,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(36563,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(36563,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(36563,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(36563,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(36563,36,'cp 29','c7c5 g1e2 d8a5 e2c1 a5b6 c1b3 c5d4 e3d4 b6d8 d4e3 b8c6 f1e2 c8e6 a1c1 f6d7 e1g1 d7b6 b3d2 f7f5 e4f5 g6f5 g1h1 g8h8 b2b3 d6d5 c4d5 b6d5 c3d5 e6d5 d1e1 a8c8 f3f4 c6b4 c1c8 d8c8 e3a7 c8a8 d2f3 a8a7 e1b4 f8c8 b4d2 a7c5');
INSERT INTO evaluations VALUES(35949,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(35949,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(35949,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(35949,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(35949,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(35949,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(35949,36,'cp 34','g1e2 d8a5 e2c1 c5d4 c1b3 a5b6 e3d4 b6d8 f1e2 b8c6 d4e3 c8e6 b3d4 c6d4 e3d4 d8a5 e1g1 a8c8 b2b3 a7a6 a2a4 f6d7 f3f4 g7d4 d1d4 a5c5 d4c5 c8c5 a4a5 d7f6 h2h3 h7h5 f1d1 f8c8 d1d4 h5h4 g1f2 f6d7 f2e3 c5c7 e2g4 c7c5 c3d5 g8f8 d5b6 d7b6 g4e6');
INSERT INTO evaluations VALUES(35948,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(35948,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(35948,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(35948,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(35948,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(35948,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(35948,36,'cp 28','d8a5 e2c1 a5b6 c1b3 c5d4 e3d4 b6d8 d4e3 b8c6 f1e2 c8e6 b3d4 c6d4 e3d4 f6d7 b2b3 d8a5 e1g1 g7d4 d1d4 a5c5 d4c5 d7c5 f3f4 a7a5 a1d1 g8g7 g1f2 h7h5 d1d2 e6d7 f1e1 a8b8 f2e3 f7f6 e1b1 h5h4 b1d1 g6g5 c3d5 g5f4 d5f4');
INSERT INTO evaluations VALUES(52638,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(52638,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(52638,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(52638,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(52638,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(52638,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(52638,36,'cp 62','f8g7 c1e3 g8f6 f2f3 e8g8 d1d2 b8c6 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8b6 c3a4 b6a5 b2b3 c8f5 g2g4 f5e6 d4e5 a5b4 c1b2 a8b8 d1d4 b4d6 e5d6 e7d6 a4c3 b8b4 d4b4 d5b4 f1e2 f8d8 h1d1 g8f8 f3f4 b4d5 c3d5 e6d5 c2c4 d8e8');
INSERT INTO evaluations VALUES(37999,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(37999,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(37999,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(37999,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(37999,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(37999,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(37999,36,'cp 91','f2f3 a7a6 c1e3 g8f6 d1d2 h7h5 f1c4 b8d7 e3g5 d8c7 c4b3 b7b5 e1c1 c8b7 h1e1 e8g8 g5h6 d7c5 c1b1 e7e6 a2a3 a8c8 h6g7 g8g7 b3a2 f8d8 d2g5 f6h7 g5g3 h7f8 c3e2 c5d7 e2f4 g7h7 f4d3 h7h8 c2c3 d7e5');
INSERT INTO evaluations VALUES(6423,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(6423,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(6423,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(6423,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(6423,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(6423,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(6423,36,'cp 29','d1h5 d8f6 g1f3 g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 f3d2 g6e7 a2a3 e6c4 d2c4 f7f6 f1f2 a7a6 a3a4 e8f7 h2h4 h6h5 a4a5 f7e6 g1h2 a8c8 h2h3 c6b4 d3d4 c8d8 d4d5 e6d7');
INSERT INTO evaluations VALUES(34826,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(34826,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(34826,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(34826,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(34826,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(34826,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(34826,36,'cp 76','g1f3 g8f6 b1c3 f8b4 g2g3 c8d7 f1g2 d7c6 e1g1 d5a5 c1d2 a5b6 c3a4 b6a5 d2b4 a5b4 b2b3 e8g8 a1c1 b8d7 d1c2 c6f3 g2f3 c7c6 a4c5 d7c5 c2c5 b4c5 c1c5 f8d8 e2e3 f6d7 c5c2 d7f6 f3g2');
INSERT INTO evaluations VALUES(34824,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(34824,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(34824,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(34824,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(34824,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(34824,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(34824,36,'cp 58','g8f6 a2a3 c7c5 b1c3 d5d7 c1f4 c5d4 f3d4 f6d5 c3d5 d7d5 e2e3 a7a6 a1c1 b8d7 d1b3 d7f6 b3d5 f6d5 f4g3 f8e7 f1c4 e8g8 c4d5 e6d5 e1e2 f8e8 h1d1 h7h5 h2h3 h5h4 g3h2 e7f6 c1c5 c8e6 c5c7 e8e7 d1c1 e7c7 c1c7 f6d4 e3d4 a8e8');
INSERT INTO evaluations VALUES(33886,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(33886,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(33886,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(33886,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(33886,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(33886,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(33886,36,'cp 103','a2a3 c7c5 b1c3 d5d8 c1f4 c8d7 e2e4 c5d4 d1d4 f8e7 a1d1 b8c6 d4e3 d8a5 f1e2 e6e5 f4g5 d7e6 e1g1 e8g8 c3d5 e6d5 g5f6 e7f6 d1d5 a5c7 f1c1 c7e7 e2c4 c6d4 e3d3 d4f3 d3f3 g7g6 c1d1 a8d8 g2g3 d8d5');
INSERT INTO evaluations VALUES(56689,30,'cp 221',NULL);
INSERT INTO evaluations VALUES(56689,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(56689,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(56689,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(56689,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(56689,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(56689,36,'cp 219','e1g1 g8e7 d2d4 e7g6 c2c3 c7c6 d1b3 d8b6 b3c2 b6d8 c1e3 b8a6 d4e5 d6e5 f1d1 d8e7 b1d2 c8e6 c2b3 e6f7 h2h4 e7c7 h4h5 g6f4 c4f7 c7f7 b3f7 e8f7 e3f4 e5f4 d2b3 f8e7 d1d7 b7b6 f3d4 h8d8 d7b7 d8b8 b7e7 f7e7 d4c6 e7f7 c6b8 a8b8 a1d1');
INSERT INTO evaluations VALUES(56688,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(56688,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(56688,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(56688,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(56688,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(56688,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(56688,36,'cp 215','b8c6 d2d4 h7h5 b2b4 g7g5 b4b5 c6a5 c4e2 c8e6 c2c4 g8e7 d4d5 e6f7 f3e1 d8d7 e1c2 f8g7 c1d2 b7b6 b1c3 a5b7 c2b4 f7g6 a2a4 b7c5 a4a5 g6e4 c3e4 c5e4 a5b6 c7b6 e2h5 e8f8 d2e3 f6f5 b4c6 e4c3 d1f3 e5e4 f3h3 c3e2 g1h1 g7a1');
INSERT INTO evaluations VALUES(22304,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(22304,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(22304,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(22304,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(22304,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(22304,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(22304,36,'cp 224','d2d4 g7g6 h2h3 c6a5 c4e2 g8h6 d4e5 d6e5 d1d8 e8d8 f1d1 d8e8 b1c3 c8e6 c3b5 a8c8 b5a7 c8d8 d1d8 e8d8 c1e3 h6f7 a1d1 f7d6 f3d2 d8e8 a2a4 b7b6 a7b5 d6b5 e2b5 e8f7 c2c3 c7c6 b5e2 c6c5 e2b5 f8d6 d1b1 h8c8 b2b4 a5b7 b5a6 c8b8 a6b7 b8b7 b4c5');
INSERT INTO evaluations VALUES(18996,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(18996,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(18996,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(18996,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(18996,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18996,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18996,36,'cp 10','e5d4 c3d4 c5d4 f3d4 f8b4 b1c3 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4d6 e2d4 c8d7 f1d1 a7a5 d4f3 h7h6 d3c4 f8d8 b2b3 a5a4 b3a4 d7e8 a4a5 a8a5 c4b3 d8a8');
INSERT INTO evaluations VALUES(19036,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(19036,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19036,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(19036,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(19036,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(19036,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19036,36,'cp 10','c3d4 c5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4d6 e2d4 c8d7 f1d1 a7a5 d3f5 d7f5 d4f5 d6e5 f5d4 d5e7 c1e3 c6c5 d4b5 f8c8 b5d6 e5d6 d1d6 e7f5 d6d7 f5e3 f2e3 a5a4');
INSERT INTO evaluations VALUES(19035,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19035,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(19035,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(19035,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19035,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19035,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19035,36,'cp 12','c5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 f1d1 b4e7 e2f4 f8d8 f4d5 d8d5 c1e3 c8e6 b2b3 a8d8 d3e2 d5d1 a1d1 d8d1 e2d1 c6c5 d1e2 f7f5 f2f3 e7d6 g1f2 g8f7 g2g3 e6d5 e2a6 f7e7 f3f4 h7h6 f2e1 d5e6');
INSERT INTO evaluations VALUES(14861,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(14861,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(14861,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(14861,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(14861,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(14861,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(14861,36,'cp 48','e1g1 f8e7 c2c4 e8g8 b2b3 b7b6 b1d2 c6a5 c1b2 c8b7 c4d5 e6d5 a1c1 e7d6 d2b1 f8e8 b1c3 a5c6 e2e3 c6e7 f3e5 a7a6 d1c2 e7g6 c3e2 a6a5 e2f4 d8e7 f4d3 a8d8 f1d1 e7e6 d3f4 g6f4 e3f4 b7a6');
INSERT INTO evaluations VALUES(14860,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(14860,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(14860,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(14860,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(14860,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(14860,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(14860,36,'cp 39','f8e7 c2c4 e8g8 b2b3 b7b6 b1d2 c8b7 c1b2 c6a5 c4d5 e6d5 f1e1 f8e8 a1c1 e7d6 f3e5 a8c8 e2e3 d8e7 d2f3 f6e4 f3h4 e7e6 e5d3 e6d7 d1e2 a5c6 a2a3 g7g5 h4f3 f7f6 f3d2 c6e7 d2e4 d5e4');
INSERT INTO evaluations VALUES(8495,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(8495,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(8495,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(8495,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(8495,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(8495,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(8495,36,'cp 53','c2c4 e8g8 b2b3 b7b6 c1b2 c8b7 b1d2 c6a5 c4d5 e6d5 a1c1 e7d6 a2a3 f8e8 f3e5 c7c5 e2e3 c5d4 e5g4 f6g4 d1g4 d6e5 b2d4 a5c6 d4e5 c6e5 g4d4 d8f6 c1c3 a8d8 c3c7 e8e7 f1c1 h7h6 b3b4 e7c7 c1c7 d8d7 c7d7');
INSERT INTO evaluations VALUES(8480,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(8480,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(8480,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(8480,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(8480,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(8480,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(8480,36,'cp 36','e8g8 b2b3 b7b6 c1b2 c8b7 b1d2 c6a5 c4d5 e6d5 a1c1 e7d6 e2e3 d8e7 d2b1 f6e4 f3h4 g7g6 b1c3 a8c8 d1e2 f8d8 h4f3 d6b4 g2h3 c8a8 c3b5 a5c6 f3e5 c6e5 d4e5 c7c5 h3g2 a7a6');
INSERT INTO evaluations VALUES(3326,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(3326,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(3326,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(3326,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(3326,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(3326,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(3326,36,'cp 42','b2b3 b7b6 b1d2 c6a5 c4d5 e6d5 c1b2 c8e6 f3g5 e6d7 a1c1 a8c8 g5f3 d7e6 f3e5 f8e8 e2e3 h7h6 e3e4 c7c6 f2f4 d5e4 d2e4 e6d5 e4f6 e7f6 d1h5 c8c7 g2f3 d8d6 h5g4 d5f3 g4f3 c7e7 f1d1 e8d8 f3g4 d6e6');
INSERT INTO evaluations VALUES(2054,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(2054,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(2054,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(2054,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(2054,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(2054,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(2054,36,'cp 109','b1c3 g8f6 e1g1 d7d6 h2h3 c8e6 c4b3 e6b3 a2b3 e7d7 d2d4 f8e7 d4e5 d6e5 d1e2 e8g8 f1d1 e7d6 c1g5 d7e6 c3d5 f6e8 e2b5 a8b8 b5c4 a7a6 c2c3 h7h6 g5h4 a6a5 f3d2 g8h8 c4d3 e6c8 h4g3 f7f6 d2c4 c8e6');
INSERT INTO evaluations VALUES(2050,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(2050,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(2050,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(2050,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(2050,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(2050,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(2050,36,'cp 109','g8f6 e1g1 d7d6 h2h3 c8e6 c4b5 a7a6 b5c6 b7c6 d2d4 f6d7 f1e1 f7f6 d4d5 c6d5 e4d5 e6f7 f3d4 d7b6 a2a4 e7d7 d1f3 b6c8 b2b4 c8e7 d4e6 c7c6 e6f8 e8f8 d5c6 d7c6 c3e4 f7d5 c1a3 d5e4 e1e4 f8f7 b4b5 a6b5 a4b5 c6c2 a1c1 a8a3 c1c2 a3f3');
INSERT INTO evaluations VALUES(1100,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(1100,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(1100,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(1100,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(1100,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(1100,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(1100,36,'cp 102','e1g1 d7d6 h2h3 c8e6 c4b3 e6b3 a2b3 e7d7 d2d4 f8e7 d4e5 d6e5 d1e2 e8g8 f1d1 e7d6 c1g5 d7e6 c3d5 f6e8 c2c3 f7f6 g5e3 c6e7 b3b4 e7d5 d1d5 a7a6 f3d2 d6e7 d2c4 e8d6 e3c5 d6c4 c5e7 e6e7');
INSERT INTO evaluations VALUES(1099,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(1099,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(1099,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(1099,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(1099,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(1099,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(1099,36,'cp 102','d7d6 c3d5 e7d8 f1e1 f8e7 h2h3 e8g8 d5e7 d8e7 d2d4 e5d4 f3d4 c8d7 c1g5 c6e5 c4f1 e5g6 d1f3 f8e8 b2b3 e7e5 g5f6 e5f6 f3f6 g7f6 g2g3 g8f8 a1d1 a7a6 a2a4 g6e7 g1g2 e7c6 g3g4 h7h5 g2g3 h5g4 h3g4 c6d4 d1d4');
INSERT INTO evaluations VALUES(914,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(914,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(914,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(914,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(914,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(914,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(914,36,'cp 108','c3d5 e7d8 f1e1 f8e7 d5e7 d8e7 h2h3 e8g8 d2d4 e5d4 f3d4 c8d7 c1f4 a7a6 f4h2 f8e8 c4f1 h7h6 b2b3 c6d4 d1d4 d7b5 c2c4 b5c6 e4e5 d6e5 e1e5 e7d8 e5e8 f6e8 d4c3 d8d7 c3a5 a8d8 a1e1 f7f6 h2f4 e8d6 f1d3 d8e8');
INSERT INTO evaluations VALUES(19359,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(19359,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(19359,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(19359,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(19359,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(19359,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(19359,36,'cp 64','e2e4 g8f6 f1d3 d8a5 c1d2 a5c5 g1f3 d7d5 e4d5 c5d5 b1c3 d5d6 c3e4 f6e4 d3e4 g7g6 d2c3 d6d1 a1d1 h8g8 e1g1 c8f5 e4d5 e7e6 d5c6 b7c6 d1d2 f8e7 f3e5 a8c8 f1d1 g6g5 e5c4 f5e4 c4d6 e7d6 d2d6 e4d5 c3f6 g8g6 f6d4 a7a5 a2a3 g5g4');
INSERT INTO evaluations VALUES(19353,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(19353,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(19353,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(19353,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(19353,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(19353,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(19353,36,'cp 65','g8f6 f1d3 d8a5 c1d2 a5c5 g1f3 d7d5 e4d5 f6d5 b1c3 c8g4 c3d5 c5d5 d1e2 g4f3 g2f3 e8c8 d3e4 d5e5 d2e3 e5c7 e4c6 c7c6 e3a7 e7e6 a7e3 f8c5 h1g1 c5e3 f2e3 d8d5 a1d1 d5d1 e2d1 c6c7 d1e2 c7a5 e2d2 a5a2 d2c3 c8b8 g1g7');
INSERT INTO evaluations VALUES(34829,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(34829,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(34829,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34829,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(34829,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(34829,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(34829,36,'cp 90','d5d8 d2d4 g8f6 c1g5 f8e7 f1d3 e8g8 d1e2 b8c6 e1c1 c6b4 h1e1 b4d3 e2d3 b7b6 d4d5 c8b7 d5e6 d8d3 d1d3 f7e6 e1e6 e7d6 g5f6 f8f6 e6f6 g7f6 c1d1 a8e8 c3b5 d6f4 g2g3 a7a6 b5d4 f4d6 f3h4 b7e4 d3d2 g8f7 c2c3 d6c5 f2f3 e4g6 h4g6 h7g6 h2h3');
INSERT INTO evaluations VALUES(39596,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(39596,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(39596,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39596,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(39596,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(39596,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(39596,36,'cp 28','d7d6 c2c4 g8f6 f1e2 e8g8 b1c3 e7e5 d4e5 d6e5 f3e5 d8d1 c3d1 f8d8 d1c3 f6e4 c3e4 g7e5 e4c5 b8c6 c5d3 e5h8 f2f3 a7a5 e1f2 c6b4 h1d1 c8e6 d1d2 e6c4 d3b4 d8d2 c1d2 c4e2 b4d5 e2c4 d5c7 a8c8 d2a5 h8b2 a1b1 b2a3 c7b5 a3c5');
INSERT INTO evaluations VALUES(38401,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(38401,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38401,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(38401,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(38401,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(38401,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(38401,36,'cp 34','c2c4 g8f6 b1c3 b8d7 f1e2 e7e5 b2b3 e8g8 e1g1 c7c6 d4e5 d6e5 c1a3 f8e8 d1c2 g7f8 a3f8 d7f8 b3b4 d8e7 c4c5 c8g4 f3d2 g4e2 c3e2 b7b6 c5b6 a7b6 a2a3 c6c5 b4c5 e7c5 c2c5 b6c5 a1c1 f8d7 e2c3 e8d8');
INSERT INTO evaluations VALUES(38376,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(38376,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38376,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(38376,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(38376,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(38376,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(38376,36,'cp 26','g8f6 b1c3 e8g8 f1e2 b8a6 e1g1 c8f5 a2a3 e7e5 b2b4 c7c6 d4e5 d6e5 d1b3 d8e7 f1d1 a6c7 c1b2 a8d8 d1d8 f8d8 a1d1 d8d1 b3d1 h7h5 h2h3 a7a5 b4a5 c7a6 c3a4 f6e4 f3d2 a6c5 a4c5 e4c5 d2b3 c5a4');
INSERT INTO evaluations VALUES(41281,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(41281,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41281,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41281,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(41281,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41281,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41281,36,'cp 35','b1c3 e8g8 f1e2 e7e5 d4e5 d6e5 f3e5 d8d1 c3d1 f8d8 d1c3 f6e4 c3e4 g7e5 e4c3 b8a6 c1d2 a6b4 e1c1 e5c3 d2c3 b4a2 c1c2 c8f5 c2b3 a2c3 b3c3 a7a5 d1d4 d8b8 h1a1 g8f8 g2g4 f5e6 g4g5 c7c6 d4d6 b8c8 b2b3 f8e7 c4c5 c8d8 d6d8 e7d8 f2f4 d8e7');
INSERT INTO evaluations VALUES(5387,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5387,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5387,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(5387,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(5387,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(5387,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5387,36,'cp 0','d4c6 b7c6 b1c3 d7d6 h2h3 g8f6 e4e5 d6e5 d1d8 e8d8 c1d2 d8e8 e1c1 h7h5 c4b3 c8f5 d2e3 g7h6 e3h6 h8h6 g2g4 h5g4 h3g4 f6g4 f2f3 g4h2 d1d2 h2g4');
INSERT INTO evaluations VALUES(5253,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5253,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5253,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5253,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5253,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5253,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5253,36,'cp 0','b7c6 b1c3 d7d6 h2h3 g8f6 e4e5 d6e5 d1d8 e8d8 c1d2 d8e8 e1c1 h7h5 c4b3 c8f5 d2e3 g7h6 e3h6 h8h6 g2g4 h5g4 h3g4 f6g4 f2f3 g4h2 d1d2 h2g4');
INSERT INTO evaluations VALUES(5037,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5037,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5037,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5037,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5037,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5037,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5037,36,'cp 0','b1c3 d7d6 h2h3 g8f6 e4e5 d6e5 d1d8 e8d8 c1d2 d8e8 c4b3 h7h5 e1c1 c8f5 d2e3 g7h6 e3h6 h8h6 g2g4 h5g4 h3g4 f6g4 f2f3 g4h2 d1d2 h2g4');
INSERT INTO evaluations VALUES(33203,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(33203,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(33203,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(33203,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(33203,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(33203,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(33203,36,'cp 119','g8e7 e1g1 h7h6 d2d4 c5d4 f3d4 e5d4 e4e5 f6b6 c3e4 d7d5 e5d6 c7d6 c1f4 e8g8 e4d6 e7f5 d6c8 f8c8 c4d3 f5e7 d1g4 b6e6 g4f3 b8c6 f4g3 c8e8 f1e1 e6d7 h2h3 c6b4 d3e4 d4d3 e4d3 b4d3 e1d1 d7c6 f3d3 a8d8 d3b3 e7f5 d1d8 e8d8');
INSERT INTO evaluations VALUES(33478,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(33478,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(33478,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(33478,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(33478,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(33478,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(33478,36,'cp 156','e1g1 e8g8 c3a4 f6c6 a4c5 c6c5 c4b3 d7d6 d2d4 e5d4 f3d4 b8c6 c2c3 c5b6 c1e3 c6d4 c3d4 d6d5 e4d5 a7a5 e3f4 a5a4 b3c4 c8d7 d1d2 e7f5 a1c1 a8c8 f1e1 b6d4 c4d3 c8e8 e1e8 f8e8 d3f5 d4d2 f4d2 d7f5 c1c7 h7h6 d5d6 e8e2');
INSERT INTO evaluations VALUES(40952,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(40952,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(40952,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40952,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(40952,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40952,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40952,36,'cp 6','g1f3 e8g8 b2b3 c7c5 d4c5 e7c5 c1b2 b8c6 a2a3 d5c4 b3b4 d8d1 a1d1 c5e7 f1c4 a7a6 c4d3 b7b5 c3e4 f6d5 e1g1 c8b7 d1c1 d5b6 e4d2 f8d8 d3e2 a8c8 d2b3 c6b8 f3d2 b6a4 b2a1 b8d7 c1c8 d8c8 f1c1 c8c1 b3c1 e6e5 c1b3');
INSERT INTO evaluations VALUES(46945,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(46945,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(46945,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(46945,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(46945,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(46945,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(46945,36,'cp 64','e7e5 d4e5 d8d1 c3d1 c8d7 f2f4 d7c6 d1c3 f8b4 g1e2 b8a6 a2a3 b4c3 e2c3 a6c5 c1e3 c5e4 c3e4 c6e4 a1d1 g7g6 e1f2 h7h5 d1d4 e4f5 f1e2 c7c6 d4d6 b6d5 h1d1 f5e6 h2h3 d5e3 f2e3 e8e7 g2g4 h5g4 e2g4 e6g4 h3g4 a8d8 d6d8 h8d8 d1d8 e7d8 e3d4 d8e7 f4f5 b7b6 d4e4 a7a6 e4f4');
INSERT INTO evaluations VALUES(47742,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(47742,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(47742,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(47742,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(47742,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(47742,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(47742,36,'cp 62','d4e5 d8d1 c3d1 c8d7 f2f4 d7c6 d1c3 f8b4 g1e2 b8d7 a2a3 b4c3 e2c3 d7c5 f1e2 c5e4 c3e4 c6e4 e1g1 g7g6 f1d1 e4d5 a3a4 a7a5 g1f2 e8c8 b2b4 d5c4 b4a5 c4e2 d1d8 h8d8 a5b6 e2g4 a1a2 c7b6 c1e3 c8d7 e3b6 d8c8 f2g3 g4f5 a2d2 d7e7 g3h4 c8c2');
INSERT INTO evaluations VALUES(6070,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(6070,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(6070,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(6070,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(6070,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(6070,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(6070,36,'cp 133','b1c3 g8f6 e1g1 e8g8 c3d5 h7h6 d1a4 b4a5 b2b4 a5b6 d5b6 c7b6 d4d5 c6e7 f1e1 e7g6 a4b3 c8d7 h2h3 b6b5 c4d3 f8e8 c1b2 a8c8 a1d1 g6e5 f3e5 d6e5 b2c1 a7a6 c1e3 f6h5 d1c1 h5f4 d3f1 d8g5 g1h2 f7f5 e4f5 g5f5 d5d6 d7e6 b3d1 c8c1');
INSERT INTO evaluations VALUES(6067,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(6067,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(6067,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(6067,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(6067,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(6067,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(6067,36,'cp 129','g8f6 e1g1 e8g8 c3d5 h7h6 d1a4 b4a5 b2b4 a5b6 d5b6 c7b6 d4d5 c6e7 f1e1 c8d7 a4b3 e7g6 h2h3 f8e8 c4d3 b6b5 c1b2 a8c8 a2a3 a7a6 a1d1 g6e5 f3e5 d6e5 d3f1 d8e7 b3d3 f6h5 d3d2 h5f4 g1h2 e7d6 g2g3 f4g6 f2f3 f7f5 e4f5 d7f5 f1g2 f5d7');
INSERT INTO evaluations VALUES(9452,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(9452,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(9452,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(9452,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(9452,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(9452,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(9452,36,'cp 166','e1g1 e8g8 c3d5 h7h6 d1a4 b4a5 c4b5 f6h5 h2h3 a8b8 f1d1 f8e8 b2b4 a7a6 b5f1 a5b6 d5b6 c7b6 d1e1 c6e7 a4d1 h5f6 d4d5 e7g6 f3d2 c8d7 a2a4 b6b5 a4a5 b8c8 c1b2 f6h7 a1a3 h7g5 a3e3 f7f5 e4f5 e8e3 e1e3 g6f4 d2e4 d7f5');
INSERT INTO evaluations VALUES(9451,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(9451,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(9451,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(9451,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(9451,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(9451,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(9451,36,'cp 164','e8g8 c3d5 h7h6 d1a4 b4a5 c4b5 f6h5 h2h3 a8b8 f1d1 f8e8 b2b4 a7a6 b5f1 a5b6 d5b6 c7b6 d1e1 c6e7 a4d1 e7g6 d4d5 b6b5 c1b2 c8d7 a1c1 b8c8 c1c8 d7c8 d1b3 h5f6 b3d3 f6h5 g2g3 c8d7 b2c3 h5f6 d3d4 g6e5');
INSERT INTO evaluations VALUES(3887,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(3887,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(3887,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(3887,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(3887,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(3887,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(3887,36,'cp 175','c3d5 h7h6 d1a4 b4a5 c4b5 f6h5 h2h3 c8d7 b5c6 b7c6 a4a5 c6d5 e4e5 d7e6 f1e1 d6e5 d4e5 d8d7 c1e3 f8e8 a5b4 f7f5 b4c3 a7a5 a1c1 f5f4 e3d4 g7g5 c3c7 d7c7 c1c7 h5g7 a2a4 e6f5 e1c1 g7e6 c7c6 h6h5 d4b6 g5g4 f3d4 e6d4 b6d4 g4h3 g2h3 f5h3');
INSERT INTO evaluations VALUES(47441,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(47441,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(47441,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(47441,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(47441,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47441,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(47441,36,'cp 19','d5c3 b2c3 e7e5 g1f3 e5d4 c3d4 c7c5 e2e3 c5d4 d1d4 d8d4 f3d4 b8c6 d4c6 b7c6 f1c4 a8b8 e1e2 f8d6 c1d2 e8e7 a1c1 c8d7 h1d1 f7f6 e3e4 c6c5 d2c3 d7a4 d1d3 d6f4 c1h1 h8d8 d3d8 e7d8 g2g3 f4e5 c3e5 f6e5 e2e3 b8b2');
INSERT INTO evaluations VALUES(47534,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(47534,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(47534,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(47534,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(47534,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(47534,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(47534,36,'cp 24','b2c3 e7e5 g1f3 e5d4 c3d4 c7c5 d4d5 f8d6 g2g3 e8g8 f1g2 f8e8 e1g1 b7b5 d1d3 h7h6 f1e1 b8a6 d3c3 c5c4 c1b2 f7f6 c3c2 c8d7 f3d4 a8c8 a2a3 a6c5 d4c6 d8c7 b2c3 d7c6 d5c6 d6e5 a1d1 e5c3 c2c3 g8h8 g2f3 e8e5 c3b4');
INSERT INTO evaluations VALUES(47533,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(47533,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(47533,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(47533,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(47533,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(47533,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47533,36,'cp 16','e7e5 g1f3 e5d4 c3d4 c7c5 e2e3 c5d4 f1b5 c8d7 b5d7 d8d7 d1d4 b8c6 d4d7 e8d7 g2g4 f8e7 c1b2 f7f6 e1e2 a8c8 h2h4 d7e6 h4h5 h8d8 a1c1 d8d5 c1c4 b7b5 c4c2 d5c5 c2c5 e7c5 h1c1 c5e7 e3e4 c8c7 c1c6 c7c6 f3d4 e6d7 d4c6');
INSERT INTO evaluations VALUES(14796,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(14796,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(14796,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(14796,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(14796,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(14796,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(14796,36,'cp 20','a2a3 f8d6 f4g5 d6e7 e2e3 h7h6 g5h4 a7a6 f1d3 b7b6 c3e2 c8b7 e1g1 e8g8 h2h3 a6a5 c2c3 f6e4 h4e7 c6e7 a1c1 c7c5 d3b1 e7f5 e2f4 d8e7 f3e5 f8c8 b1e4 d5e4 f4h5 e7g5 e5d7 f5h4 d1g4 g5g4 h3g4 c5d4');
INSERT INTO evaluations VALUES(14819,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(14819,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(14819,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(14819,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(14819,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(14819,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14819,36,'cp 11','f8d6 f4g5 d6e7 e2e3 h7h6 g5h4 c6b8 d1e2 b8d7 e3e4 d5e4 c3e4 e8g8 e4f6 d7f6 e1c1 b7b6 c1b1 c8b7 e2e1 c7c5 h1g1 a8c8 d4c5 d8c7 f3e5 f8d8 d1d8 c7d8 c5c6 b7c6 e5c6 c8c6 f1b5 c6c7 e1e3 f6d5');
INSERT INTO evaluations VALUES(9244,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(9244,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(9244,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(9244,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(9244,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(9244,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(9244,36,'cp 22','f4g5 d6e7 e2e3 h7h6 g5h4 c6b8 h4f6 e7f6 d1d2 g7g6 h2h4 f6g7 f3e5 b8d7 f2f4 d7e5 f4e5 c8d7 e3e4 c7c5 e4d5 d8b6 d5e6 f7e6 e1c1 c5d4 d2d4 b6d4 d1d4 g7e5 d4g4 e8e7 f1b5 e5c3 b5d7 e7d7 b2c3 d7e7 g4g6 a8g8 g6g8 h8g8');
INSERT INTO evaluations VALUES(47226,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(47226,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(47226,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(47226,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(47226,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(47226,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(47226,36,'cp -27','d5d4 c3e2 c7c5 f2f4 e7e5 g2g3 b8c6 g1f3 f8d6 f1g2 e8g8 e1g1 c8g4 b2b3 f6d7 a2a4 a7a6 f3e5 c6e5 f4e5 d7e5 d1d2 b7b5 e2f4 g4d7 a4b5 a6b5 a1a8 d8a8 f4d5 a8d8 d2f2 d7e6 c1f4 e6d5 f4e5 d6e5 e4d5');
INSERT INTO evaluations VALUES(47387,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(47387,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(47387,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(47387,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(47387,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(47387,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(47387,36,'cp -32','c3e2 c7c5 f2f4 b8c6 g2g3 e7e6 g1f3 b7b5 f1h3 a7a5 a2a4 b5a4 a1a4 f8e7 e1g1 f6d7 g3g4 e8g8 g4g5 d7b6 a4a1 c5c4 h3g2 c4d3 c2d3 a5a4 g1h1 a8b8 b2b3 a4b3 d1b3 b6d7 b3d1 d7c5 e2g3 c5b3');
INSERT INTO evaluations VALUES(47388,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(47388,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(47388,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(47388,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(47388,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(47388,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(47388,36,'cp -39','c7c5 f2f4 e7e5 g2g3 b8c6 g1f3 f8d6 f1g2 c8g4 e1g1 f6d7 h2h3 g4f3 g2f3 e8g8 h3h4 b7b5 g1g2 a7a5 f4f5 d6e7 a2a4 b5a4 c2c4 d4c3 e2c3 d7b6 c3a4 b6a4 d1a4 d8b6 a4c4 f8b8 f3e2 c6d4 e2d1 b8d8 c4c3 a8b8');
INSERT INTO evaluations VALUES(47555,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(47555,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(47555,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(47555,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(47555,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(47555,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(47555,36,'cp -111','e8f7 b1c3 c8e6 d2d4 f7g8 f1d3 c7c5 d4d5 e6f7 e1g1 b8d7 b2b3 a7a6 a2a4 d8a5 c1b2 a8e8 f2f4 h7h5 d1f3 c5c4 d3c4 h8h6 a1e1 h5h4 f3h3 f7h5 h3e3 a5d8 h2h3 h6g6 e3f2 g6g3 c4d3 d7c5 e4e5 c5d3');
INSERT INTO evaluations VALUES(35488,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(35488,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(35488,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(35488,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(35488,34,'cp -110',NULL);
INSERT INTO evaluations VALUES(35488,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(35488,36,'cp -104','b1c3 c8e6 d2d4 f7g8 f1d3 c7c5 d4c5 b8c6 c5d6 f8d6 h2h3 h7h6 c1e3 g7g5 d1d2 d8a5 d3c4 e6c4 d2d6 g8f7 e1c1 a8d8 d6c5 a5c5 e3c5 d8d1 h1d1 h8d8 d1h1 b7b5 a2a3 a7a5 b2b3 b5b4 c3d1 c4e6 d1e3 f7g8 f2f3 f6d7 a3b4 a5b4');
INSERT INTO evaluations VALUES(35487,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(35487,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(35487,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(35487,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(35487,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(35487,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(35487,36,'cp -136','c8e6 f2f4 f7g8 d2d4 b8d7 c1e3 c7c5 d1f3 a8c8 h2h3 e6f7 e1c1 d8a5 e4e5 c5d4 e3d4 d6e5 f4e5 d7e5 f3b7 c8c7 b7b5 e5c6 b5a5 c6a5 c3b5 c7b7 b5a7 h7h5 b2b3 h8h6 f1d3 f6d5 d3e4 b7d7 d4b2 d7a7 e4d5 f7d5 d1d5 a5c6 b2c3 a7a2');
INSERT INTO evaluations VALUES(56879,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(56879,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(56879,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(56879,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(56879,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(56879,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(56879,36,'cp 102','d2d4 f8g7 b1c3 d7d6 f1d3 a7a6 c1g5 g8e7 d1d2 h7h6 g5e3 b7b6 e1g1 b8d7 c3e2 c8b7 c2c4 d7f6 d4d5 e6d5 c4d5 c7c6 d5c6 e7c6 a1d1 f6g4 e3f4 g4e5 f3e5 d6e5 f4e3 b6b5 g1h1 c6d4 f2f3 d4e2 d3e2 e8g8 e3h6 d8d2 h6d2 f8d8 d2a5');
INSERT INTO evaluations VALUES(1867,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(1867,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(1867,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(1867,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(1867,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(1867,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(1867,36,'cp 166','d4d5 c6b8 c1e3 d7d6 c2c4 f6d8 b1c3 b7b6 h2h3 b8d7 b2b4 g8f6 f1d3 f8e7 d3c2 f6h5 c2a4 e8g8 a4c6 a8b8 c3b5 a7a5 b4a5 b6a5 e1g1 h5f4 d1d2 f7f5 e3f4 f5e4 f3g5 d7b6 g5e6 c8e6 d5e6 e5f4 d2c2 e4e3 f2e3 f4e3 f1f8 d8f8 a1f1');
INSERT INTO evaluations VALUES(1847,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(1847,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(1847,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(1847,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(1847,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(1847,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(1847,36,'cp 156','c6b8 c1e3 d7d6 c2c4 f6d8 b1c3 g8f6 d1a4 c8d7 a4b3 d7c8 c4c5 f8e7 f1b5 f6d7 e1g1 e8g8 c5c6 b7c6 d5c6 d7f6 h2h3 c8e6 b3a4 a7a6 f1d1 d8e8 a1c1 e7d8 a2a3 h7h6 b2b4 g8h7 d1e1 d8e7 f3d2 a6b5 a4a8');
INSERT INTO evaluations VALUES(42162,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(42162,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(42162,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(42162,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(42162,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(42162,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(42162,36,'cp 87','e4e5 f6d5 g1e2 c7c5 e1g1 e8g8 d4c5 b8c6 c3d5 e6d5 c4d5 b4c5 d5c6 b7c6 d1c2 d8b6 c1f4 c8a6 a1c1 a8d8 c2c5 a6e2 c5b6 a7b6 f1e1 e2b5 c1d1 c6c5 f2f3 d8d3 d1d3 b5d3 a2a3 f8e8 e1d1 c5c4 g1f2 f7f6');
INSERT INTO evaluations VALUES(42125,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(42125,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(42125,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(42125,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(42125,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(42125,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(42125,36,'cp 67','f6d5 c1d2 e8g8 d1g4 f7f5 e5f6 d5f6 g4h4 f6d5 h4d8 f8d8 g1f3 c7c5 d4c5 b4c5 e1g1 h7h6 a2a3 b8c6 c4a2 c6d4 f3d4 c5d4 a1d1 d4f6 c3d5 e6d5 d2a5 d8e8 a2d5 c8e6 d5b7 a8b8 b7c6 e8c8 d1d6 e6c4 c6d7 c4f1 d7c8 b8c8 g1f1');
INSERT INTO evaluations VALUES(41943,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(41943,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(41943,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(41943,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(41943,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(41943,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(41943,36,'cp 88','g1e2 c7c5 e1g1 e8g8 d4c5 b8c6 c3d5 e6d5 c4d5 b4c5 d5c6 b7c6 c1f4 d8d1 f1d1 c8g4 d1d2 f7f6 e5f6 f8f6 f4g3 a8e8 e2c3 f6e6 g1f1 e6e7 h2h3 g4e6 a1c1 c5b6 d2d6 e6c4 f1g1 e7e6 c1d1 e6d6 d1d6 e8e6 d6d7 h7h5 b2b3 e6e1 g1h2 c4f7 c3a4 f7d5 a4b6');
INSERT INTO evaluations VALUES(47193,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(47193,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(47193,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(47193,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(47193,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(47193,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(47193,36,'cp -4','e7e6 g1f3 b7b6 c2c4 c8b7 f1d3 d5c4 d3c4 f8d6 e1g1 b8d7 b1d2 e8g8 b2b3 e6e5 d4e5 d7e5 c4e2 d8e7 c1b2 f6e4 d2e4 b7e4 b2e5 d6e5 f3e5 e7e5 d1d4 e5d4 e3d4 f8e8 e2c4 a7a6 f2f3 e4b7 f1e1 b6b5 c4f1 b7d5 e1e8 a8e8 g1f2 c7c6 a1c1');
INSERT INTO evaluations VALUES(48347,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48347,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(48347,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(48347,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(48347,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48347,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(48347,36,'cp 0','c2c4 b7b6 g1f3 c8b7 b2b3 b8d7 f1e2 f8d6 e1g1 e8g8 c1a3 d6a3 b1a3 d8e7 d1c1 h7h6 c1b2 c7c5 a1c1 a8c8 f1d1 a7a6 d4c5 d7c5 f3e5 c5e4 e2d3 e4c5');
INSERT INTO evaluations VALUES(52102,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(52102,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(52102,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52102,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(52102,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52102,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(52102,36,'cp 18','b8c6 d2d3 f8e7 b1c3 g8f6 f3d2 d7d6 d2f1 c8g4 f2f3 g4e6 f1e3 a8b8 a2a4 a7a6 c3d5 b7b5 c4a2 c6b4 d5b4 c5b4 a4b5 a6b5 a2e6 f7e6 e1g1 e8g8 g1h1 b8c8 a1a7 c8c7 a7c7 d8c7 c1d2 d6d5 e4d5 f6d5 e3g4');
INSERT INTO evaluations VALUES(19534,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(19534,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(19534,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(19534,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(19534,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(19534,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(19534,36,'cp -35','g8f6 g1f3 b7b5 c4e2 e7e6 a2a3 a7a6 g2g3 c8b7 f1g2 c6a5 b1d2 a8c8 c2c3 d8c7 e1g1 f8e7 f3d4 e8g8 f1e1 f8e8 d4b3 a5c4 a3a4 e6e5 a4b5 a6b5 d2c4 b5c4 b3d2 h7h6 d2f1 c7c6 f1d2 d7d6 d2f1 e8d8 c1e3 f6e4');
INSERT INTO evaluations VALUES(12877,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(12877,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(12877,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(12877,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(12877,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(12877,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(12877,36,'cp -25','f1d3 e7e6 g1f3 a7a6 e1g1 b7b5 c4c3 c8b7 a2a4 b5b4 c3e1 d8c7 b1d2 d7d5 e1e2 f8e7 f1e1 d5e4 d2e4 e8g8 c1e3 f6d5 e4c5 e7f6 a1b1 d5e3 e2e3 f8d8 d3e4 a8c8 e1d1 g7g6 c5b7 c7b7 d1d8 c8d8');
INSERT INTO evaluations VALUES(54913,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54913,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54913,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54913,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54913,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54913,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54913,36,'cp 0','g5h4 c7c5 g1f3 b8c6 e2e3 c5d4 e3d4 d8b6 b1c3 e7e6 f1d3 c8d7 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(54924,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(54924,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54924,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54924,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54924,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54924,36,'cp 0','c7c5 g1f3 b8c6 e2e3 c5d4 e3d4 d8b6 b1c3 e7e6 f1d3 c8d7 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(21462,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(21462,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(21462,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(21462,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(21462,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(21462,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(21462,36,'cp 76','f7f6 g1f3 f6e5 f3e5 c6e5 d4e5 c8d7 f1d3 d8h4 b1d2 f8c5 d1e2 g8e7 b2b4 c5b6 h2h3 e8g8 g2g3 h4h6 d2f1 g7g5 f1e3 b6e3 c1e3 h6g7 h3h4 g5g4 h4h5 f8f3 h1h4 e7f5 h5h6 g7g6 e1c1 d7a4 d3f5 g6f5 d1d4 a4e8 e2c2 f5c2 c1c2 e8g6 c2b2 a8f8 c3c4 c7c6 h4g4 f8f5 b4b5 f5e5 b5c6');
INSERT INTO evaluations VALUES(22236,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(22236,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(22236,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(22236,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(22236,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(22236,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(22236,36,'cp 97','g1f3 f6e5 f3e5 c6e5 d4e5 c7c5 f1d3 g8e7 b1d2 e7c6 d1h5 e8d7 d2f3 d7c7 e1g1 c8d7 h5g4 a8c8 c3c4 h7h5 g4h3 c7b8 c4d5 e6d5 e5e6 d7e8 f1d1 f8d6 d3f5 d5d4 c1g5 d8c7 e6e7 c6e7 g5e7 d6e7 f5c8 c7c8');
INSERT INTO evaluations VALUES(53170,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53170,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53170,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(53170,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(53170,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(53170,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53170,36,'cp 14','e7e6 c1e3 g8f6 c2c4 b8c6 b1c3 d5c4 d1d8 c6d8 f3e5 f6d5 e3d4 d5c3 d4c3 f7f6 e5c4 f8c5 a1d1 e6e5 c4d6 c5d6 d1d6 c8e6 a2a3 d8f7 d6d1 a8d8 d1c1 f7d6 e2e3 d8c8 f2f3 h7h5 e1f2 e5e4 h2h4 e8f7 c1d1 h8d8 f1e2 b7b6');
INSERT INTO evaluations VALUES(54073,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(54073,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(54073,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54073,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(54073,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54073,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54073,36,'cp 25','c2c4 f8c5 c4d5 g8f6 d1c2 d8b6 e2e3 e6d5 f1e2 b8c6 a2a3 e8g8 b2b4 c5e7 c1b2 c8g4 e1g1 g4f3 e2f3 a7a5 b1c3 a5b4 c3a4 b6a6 a3b4 c6b4 c2b3 f8d8 f1d1 a8c8 a4c3 a6c4 b3c4 d5c4 f3b7 c8b8 d1d8 e7d8 a1a8 b8a8 b7a8 b4d3');
INSERT INTO evaluations VALUES(54059,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54059,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(54059,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54059,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54059,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54059,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(54059,36,'cp 9','f8c5 c4d5 e6d5 d1c2 d8b6 e2e3 g8f6 b1c3 b8c6 c1d2 c5d6 f1e2 e8g8 a1c1 c8e6 e1g1 a8c8 f1d1 a7a6 c2b1 b6c7 c3a4 c7e7 h2h3 h7h6 d2c3 f6d7 c3e1 b7b5 a4c3 d7b6 e2f1 e7f6 c3e2');
INSERT INTO evaluations VALUES(38119,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(38119,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(38119,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(38119,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38119,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(38119,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(38119,36,'cp 16','c4d5 g8f6 d1c2 d8b6 e2e3 e6d5 b1c3 b8c6 c1d2 c5d6 f1e2 e8g8 e1g1 a7a6 a1c1 c8e6 c3a4 b6d8 f3g5 a8c8 d2c3 d5d4 g5e6 f7e6 e3d4 c6d4 c2d3 d4e2 d3e2 e6e5 c3d2 e5e4 h2h3 d8e7 c1c8 f8c8 d2e3 f6d5 a4b6 d5b6 e3b6');
INSERT INTO evaluations VALUES(38112,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(38112,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(38112,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(38112,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(38112,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38112,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(38112,36,'cp 10','e6d5 d1c2 d8b6 e2e3 g8f6 f1e2 b8c6 a2a3 c5d6 b2b4 a7a5 b4b5 c6e5 c1b2 e5f3 e2f3 b6b5 b1c3 b5c6 a1c1 c8e6 c2e2 c6a6 c3b5 e8g8 b5d6 a6d6 e1g1 f8c8 e2b5 d6a6 b5a6 b7a6 f3e2 f6e4 c1c8 a8c8 e2a6');
INSERT INTO evaluations VALUES(38197,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(38197,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(38197,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38197,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(38197,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38197,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(38197,36,'cp 17','d1c2 d8b6 e2e3 g8f6 f1e2 b8c6 e1g1 d5d4 e3d4 c6d4 f3d4 c5d4 b1d2 e8g8 d2f3 c8g4 f3d4 b6d4 e2g4 d4g4 h2h3 g4e6 c1d2 f8c8 c2a4 h7h6 f1e1 e6c6 a4f4 c6c2 d2c3 f6d5 f4f3 d5c3 e1e7 c2g6 b2c3 b7b6 a1d1 c8d8 d1d8 a8d8 e7a7');
INSERT INTO evaluations VALUES(16155,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(16155,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(16155,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(16155,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(16155,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(16155,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(16155,36,'cp 13','e5d4 e4e5 d7d5 c4b5 f6e4 f3d4 f8c5 b1d2 c5d4 c3d4 e8g8 b5c6 b7c6 e1g1 c6c5 d2e4 d5e4 d4c5 a7a5 f1e1 c8f5 c1f4 d8e7 e5e6 f5e6 e1e4 e7c5 e4e3 f8d8 d1e2 d8d5 f4e5 a8d8 a1e1 e6f5 h2h3 d8e8 e5f4 e8e3 e2e3 c5b5 b2b3');
INSERT INTO evaluations VALUES(16187,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(16187,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(16187,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(16187,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16187,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(16187,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(16187,36,'cp 14','e4e5 d7d5 c4b5 f6e4 f3d4 c8d7 b5c6 b7c6 e1g1 f8e7 c1e3 a7a5 b1d2 e4d2 d1d2 e8g8 f2f4 f8e8 f4f5 c6c5 f5f6 c5d4 f6e7 d8e7 c3d4 e7b4 a1c1 b4d2 e3d2 e8c8 f1f3 d7e6 h2h3 a8b8 a2a3 b8b2 d2a5 b2e2 c1c7 e2a2 c7c8 e6c8');
INSERT INTO evaluations VALUES(16109,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(16109,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(16109,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(16109,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16109,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16109,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16109,36,'cp 0','d7d5 c4b3 f6e4 e1g1 c8g4 c3d4 e4g5 c1g5 h6g5 b1c3 g4f3 d1f3 c6d4 f3d5 d8d5 b3d5 c7c6 f1d1 d4e6 d5e6 f7e6 c3e4 f8e7 g2g3 a8d8 d1d8 e8d8 g1g2 c6c5 a1c1 b7b6 f2f4 g5f4 g3f4 d8c7 g2g3 c7c6 c1d1 h8d8 d1g1 d8h8 g1d1');
INSERT INTO evaluations VALUES(49400,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(49400,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(49400,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(49400,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(49400,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49400,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49400,36,'cp 0','d7d5 c3b5 f8b4 c2c3 b4a5 g1f3 a7a6 b5a3 e8g8 e2e3 d8e7 a3b1 f6h5 f4g5 f7f6 g5h4 g7g5 f3d2 h5g7 h4g3 b8c6 h2h4 g7f5 h4g5 f6g5 g3h2 e6e5 d4e5 c8e6 d1h5 e6f7 h5g4 f7e6 g4h5');
INSERT INTO evaluations VALUES(49632,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(49632,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(49632,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(49632,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49632,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(49632,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(49632,36,'cp -27','d7d5 g1f3 d5e4 f3e5 b8d7 e5d7 c8d7 b1d2 e4d3 f1d3 d8e7 d1e2 e7e2 d3e2 e8c8 e1g1 f8d6 d2f3 f6e4 e2d3 e4c5 d3c2 h8e8 c1e3 h7h6 f1e1 a7a5 h2h4 c5a4 a1b1 a4b6 e3b6 c7b6 f3d4 d6c5 d4f5 e8e1 b1e1 d7f5 c2f5 c8c7 e1b1');
INSERT INTO evaluations VALUES(48831,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(48831,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(48831,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(48831,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(48831,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(48831,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(48831,36,'cp -30','g1f3 d5e4 f3e5 b8d7 e5d7 d8d7 c1g5 d7e6 d3e4 e6e4 d1e2 f8e7 e2e4 f6e4 g5e7 e8e7 b1d2 e4d2 e1d2 c8e6 f1d3 a8d8 h1e1 e7f6 e1e3 g7g6 d2c2 d8d6 a1e1 h8d8 b2b3 h7h5 h2h4 b7b6 g2g3 e6g4 b3b4 c7c5 a2a3 g4e6 c3c4 d6d4 c2c3 d8d6 b4c5 b6c5 e3f3 f6g7 e1e5');
INSERT INTO evaluations VALUES(9327,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(9327,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9327,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9327,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9327,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9327,36,'cp 0','a7a6 d2d3 c6a5 c4b3 e8g8 c3e2 f6h5 e2c3');
INSERT INTO evaluations VALUES(7167,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7167,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(7167,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7167,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(7167,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7167,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7167,36,'cp 12','a2a4 e8g8 d2d3 c8e6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 c6e7 d5e7 d8e7 b2b4 c5b6 a4a5 b6a7 d1e2 a8e8 c4e6 f7e6 d3d4 e5d4 c3d4 c7c6 a1d1 d6d5 e4d5 e6d5 e2e7 e8e7 f1e1 e7e1 d1e1 f8f4 e1e7 a7d4 e7b7 d4c3');
INSERT INTO evaluations VALUES(43521,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43521,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43521,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43521,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43521,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43521,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43521,36,'cp 0','b7b5 c4d3 c8b7 d1e2 b8d7 b2b4 d7e5 a2a4 b5a4 a1a4 e5d3 c2d3 f8e7 c1d2 e8g8 a4a1 d8b8 a1b1 f8c8 h2h3 b7c6 d4c6 c8c6 d3d4 c6b6 f1c1 b6b4 b1b4 b8b4 c3d5 b4b7 c1c7 b7b1 c7c1');
INSERT INTO evaluations VALUES(43749,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43749,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43749,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43749,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(43749,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43749,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43749,36,'cp 5','c4d3 c8b7 d1e2 b8d7 b2b4 d7e5 a2a4 b5a4 a1a4 e5d3 c2d3 f8e7 c1d2 e8g8 a4a1 d8b8 a1b1 f8c8 f1c1 b7c6 d4c6 c8c6 d3d4 b8b7 b4b5 a6b5 b1b5 b7c7 b5b1 a8c8 h2h3 c7d7 e4e5 f6e8 d4d5 e6d5 c3d5 c6c1 b1c1 c8c1 d2c1');
INSERT INTO evaluations VALUES(43756,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43756,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43756,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43756,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43756,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43756,36,'cp 0','c8b7 d1e2 b8d7 b2b4 d7e5 a2a4 b5a4 a1a4 e5d3 c2d3 f8e7 c1d2 e8g8 a4a1 d8d7 c3a4 a8b8 f1b1 b7c6 a4c3 c6b7');
INSERT INTO evaluations VALUES(28520,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(28520,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(28520,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28520,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28520,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28520,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28520,36,'cp 0','f1e1 f8e7 a2a4 b5b4 c3a2 e8g8 a2b4 d6d5 e4e5 e7b4 c2c3 b4e7 e5f6 e7f6 c1d2 b8d7 b2b4 a8c8 d1e2 d7e5 d3a6 b7a6 e2a6 e5c4 d2e3 e6e5 d4b3 c8a8 a6b7 a8b8 b7c6 b8c8 c6b5 c8b8 b5c5 b8c8 c5a7 c8a8');
INSERT INTO evaluations VALUES(51598,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(51598,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(51598,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(51598,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(51598,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(51598,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(51598,36,'cp 88','d2d4 c5d4 f3d4 a7a6 b1c3 e7e6 c1e3 d8c7 d1d2 f8b4 f2f3 g8e7 e1c1 c8b7 d4e2 e8g8 a2a3 b4c5 e3f4 e6e5 f4g5 f7f6 g5e3 c5e3 d2e3 b6b5 h2h4 b8c6 e3d2 a8d8 d2d6 c7d6 d1d6 e7c8 d6d2 c8e7 h4h5 d7d6');
INSERT INTO evaluations VALUES(51593,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(51593,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(51593,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(51593,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(51593,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(51593,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(51593,36,'cp 90','c5d4 f3d4 a7a6 b1c3 e7e6 c1e3 d8c7 d1d2 f8b4 d4e2 b8c6 e3f4 e6e5 f4g5 h7h6 g5h4 c8b7 a2a3 b4e7 c3d5 c7d6 h4g3 g8f6 d5f6 d6f6 e1c1 d7d6 e2c3 c6d4 c1b1 e8g8 f2f4 a8c8 f1d3 e5f4 g3f4 d4e6 f4g3 e6c5 h1e1 c5d3 d2d3');
INSERT INTO evaluations VALUES(51604,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(51604,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(51604,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(51604,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(51604,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(51604,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(51604,36,'cp 95','f3d4 a7a6 c1e3 e7e6 b1c3 d8c7 d1d2 g8f6 f2f3 h7h5 e1c1 c8b7 c1b1 f8e7 f1e2 b8c6 d4c6 d7c6 e3g5 e8g8 d2e3 e7c5 e3c1 c5e7 f3f4 a8d8 e4e5 f6g4 g5e7 c7e7 h2h3 g4h6 e2h5 c6c5 h1g1 h6f5 g2g4 d8d1 g1d1 f5d4 c1e3');
INSERT INTO evaluations VALUES(51603,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(51603,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(51603,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(51603,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(51603,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(51603,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(51603,36,'cp 95','a7a6 c1g5 d8c7 b1c3 e7e6 d1d2 f8b4 e1c1 g8f6 f2f3 b4c3 d2c3 c7c3 b2c3 d7d6 f1e2 h7h6 g5h4 e8e7 d4b3 b8d7 d1d2 g7g5 h4f2 c8b7 h1d1 f6e8 a2a4 b7c6 a4a5 b6b5 f2g3 d7e5 d2d6 e8d6 g3e5 h8d8');
INSERT INTO evaluations VALUES(55455,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(55455,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(55455,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(55455,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(55455,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(55455,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(55455,36,'cp 117','f1d3 a7a6 e1g1 b8c6 b1d2 g7g6 c2c3 f8g7 a2a4 e6e5 d4e5 c6e5 f3e5 d6e5 d2c4 g8e7 d1e2 e8g8 f1d1 c8e6 c1e3 b7b6 f2f3 d8c8 e3f2 f8d8 c4e3 e6b3 d3c2 d8d1 a1d1 b3e6 a4a5 b6b5 e3d5 e6d5 e4d5 c8d7 c3c4');
INSERT INTO evaluations VALUES(55454,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(55454,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(55454,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(55454,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(55454,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(55454,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(55454,36,'cp 99','a7a6 e1g1 b8c6 b1d2 g7g6 a2a4 f8g7 c2c3 g8f6 f1e1 e8g8 a4a5 e6e5 h2h3 f6h5 d4d5 c6e7 c3c4 f7f5 a1a3 h5f4 d3c2 c8d7 b2b4 a8c8 d2f1 f5e4 c2e4 e7f5 e4c2 f5h4 f3h4 d8h4 f1g3');
INSERT INTO evaluations VALUES(50898,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(50898,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(50898,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(50898,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(50898,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(50898,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(50898,36,'cp 117','e1g1 b8c6 c2c3 g7g6 a2a4 f8g7 a4a5 g8f6 f1e1 e8g8 h2h3 b7b5 a5b6 c7b6 c1e3 c8b7 b1d2 b6b5 d1e2 d8c7 f3h2 b5b4 d3a6 b4c3 b2c3 c6e7 a6b7 c7b7 e2d3 d6d5 h2g4 f6g4 h3g4 d5e4 d2e4 e7d5 a1a8 f8a8 c3c4 d5e3');
INSERT INTO evaluations VALUES(50897,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(50897,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(50897,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(50897,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(50897,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(50897,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(50897,36,'cp 115','b8c6 c2c3 g7g6 a2a4 f8g7 b1d2 g8f6 f1e1 e6e5 h2h3 f6d7 d3c4 e8g8 d2f1 a6a5 f1g3 g8h7 c1e3 e5d4 c3d4 c6b4 a1c1 d6d5 c4b3 d7b6 e4d5 b6d5 f3e5 d8h4 d1f3 c7c6 e5f7 b4d3 b3d5 d3e1');
INSERT INTO evaluations VALUES(29561,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(29561,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(29561,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(29561,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(29561,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(29561,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(29561,36,'cp -9','e7e6 f1d3 a7a6 d3f5 e6f5 e1g1 f8d6 c3e2 c7c6 b2b3 e8g8 a2a4 d8e7 c2c4 b8d7 d1c2 g7g6 e2f4 f8e8 c2a2 e7d8 f4d3 d7f8 c1a3 d6c7 a2c2 f8e6 a1d1 f6e4 g2g3 e6g7 d1c1 g6g5 f3e1 d8d7 f2f3 e4d6 a3d6 d7d6 c4d5');
INSERT INTO evaluations VALUES(30114,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(30114,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(30114,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(30114,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(30114,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(30114,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(30114,36,'cp -3','f1d3 a7a6 d3f5 e6f5 e1g1 f8d6 c3e2 c7c6 b2b3 e8g8 a2a4 d8e7 c2c4 g7g6 d1c2 f8e8 c4c5 d6c7 b3b4 b8d7 c1d2 f6e4 b4b5 g6g5 b5b6 c7b8 d2c1 e7f6 f3d2 e4d2 c1d2 f5f4 a1e1 d7f8 e3f4 b8f4 d2f4');
INSERT INTO evaluations VALUES(30121,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(30121,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(30121,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(30121,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(30121,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(30121,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(30121,36,'cp -22','a7a6 d3f5 e6f5 c3e2 f8d6 e1g1 c7c6 b2b3 d8e7 c2c4 e8g8 d1c2 g7g6 e2f4 b8d7 f4d3 f8e8 a2a4 f6e4 c2a2 e7e6 a2c2 g6g5 c1a3 d6b8 c4d5 c6d5 a1c1 h7h5 f3e5 b8e5 d3e5 d7e5 d4e5 e6e5');
INSERT INTO evaluations VALUES(6509,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(6509,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(6509,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(6509,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(6509,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(6509,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6509,36,'cp 0','d7d6 c3a4 a7a5 a4c5 d6c5 f1b5 g8e7 f3e5 d8d4 e5c6 d4e4 d1e2 e4e2 e1e2 c8d7 c6a7 a8a7 b5c4 e7f5 d2d3 d7e6 h1e1 e6c4 d3c4 e8d7 e2d3 f5d4 d3c3 d4e6 e1e4 a7a8 b2b3 b7b6 c1f4 h8e8 a1d1 d7c6 d1e1 c6d7');
INSERT INTO evaluations VALUES(6027,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(6027,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(6027,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(6027,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(6027,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(6027,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6027,36,'cp 10','c3a4 a7a5 a4c5 d6c5 f1c4 d8f6 d2d3 g8e7 h2h3 b7b6 c1e3 e8g8 c2c3 e7g6 c4d5 h7h6 d3d4 c5d4 c3d4 a8b8 a1c1 c6d4 f3d4 e5d4 d1d4 f6d4 e3d4 g6f4 c1c7 f4d5 e4d5 f8d8 d4c3 d8d5 c7a7 c8e6 e1g1 b8d8 f1e1 d8d7 a7d7 d5d7');
INSERT INTO evaluations VALUES(29520,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29520,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(29520,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(29520,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29520,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(29520,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(29520,36,'cp 21','e7e6 g1f3 c7c6 f3h4 f5g6 f1e2 b8d7 g2g3 f8d6 e1g1 e8g8 h4g6 h7g6 b2b3 f8e8 c1b2 d8e7 d1c2 a8d8 f1d1 d6a3 b2a3 e7a3 a1c1 a3e7 h2h4 d7b6 e2d3 d5c4 b3c4 e6e5 d3f1 e5d4 e3d4 e7d7 d4d5 c6d5 c4d5 b6d5 c3d5');
INSERT INTO evaluations VALUES(30067,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(30067,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(30067,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(30067,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(30067,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(30067,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(30067,36,'cp 22','g1f3 c7c6 f3h4 f5g6 f1e2 b8d7 e1g1 f8d6 g2g3 a7a6 a2a4 e8g8 h4g6 h7g6 a4a5 a8b8 c1d2 d5c4 d1a4 e6e5 a4c4 d6c7 b2b4 d8e7 f1b1 g6g5 e2f1 c7d6 f1g2 e5e4 c3a4 f6d5 a4c5 d7f6 c4c2 d6c5 b4c5 g5g4');
INSERT INTO evaluations VALUES(30068,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(30068,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30068,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30068,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(30068,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(30068,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(30068,36,'cp 24','c7c6 f3h4 f5g6 f1e2 b8d7 e1g1 a7a6 g2g3 f8e7 h4g6 h7g6 b2b3 e7b4 d1c2 e6e5 c1b2 e5d4 e3d4 e8g8 e2f3 a8c8 a2a3 b4c3 b2c3 f8e8 f1e1 e8e1 a1e1 d7f8 a3a4 f8e6 f3g2 g6g5 a4a5 g5g4 c2e2 e6g5 e2e3');
INSERT INTO evaluations VALUES(55054,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(55054,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(55054,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(55054,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(55054,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(55054,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(55054,36,'cp -42','c7c5 d4c5 g8f6 a2a3 b8c6 e2e3 e7e6 c2c4 f8c5 g1f3 e8g8 b1c3 d5c4 d3d8 f8d8 f1c4 c8d7 e1g1 a8c8 c1d2 d7e8 f1d1 c6a5 c4a2 c5e7 d2e1 a5c4 d1d8 e7d8 a2c4 c8c4 f3e5 c4c8 h2h3 f6d7 e5d7 e8d7 g1f1 d8e7 f2f3 d7e8 a1c1');
INSERT INTO evaluations VALUES(2149,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(2149,31,'cp 287',NULL);
INSERT INTO evaluations VALUES(2149,32,'cp 289',NULL);
INSERT INTO evaluations VALUES(2149,33,'cp 290',NULL);
INSERT INTO evaluations VALUES(2149,34,'cp 282',NULL);
INSERT INTO evaluations VALUES(2149,35,'cp 290',NULL);
INSERT INTO evaluations VALUES(2149,36,'cp 297','c4d5 f7e7 d5c6 b7c6 f3e3 e7f7 d2d3 f8d6 b1d2 h8f8 d2e4 f6g6 e1g1 f7g8 c1d2 c8e6 b2b3 h7h6 a1e1 a8e8 f2f3 a7a6 e3f2 f8f7 g1h1 e6d5 e1e2 g8h8 f1e1 e8f8 c2c4 d5e4 e2e4 c6c5 f2e2 h8g8');
INSERT INTO evaluations VALUES(2147,30,'cp 303',NULL);
INSERT INTO evaluations VALUES(2147,31,'cp 297',NULL);
INSERT INTO evaluations VALUES(2147,32,'cp 308',NULL);
INSERT INTO evaluations VALUES(2147,33,'cp 294',NULL);
INSERT INTO evaluations VALUES(2147,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(2147,35,'cp 297',NULL);
INSERT INTO evaluations VALUES(2147,36,'cp 297','f7e7 d5c6 b7c6 f3e3 e7f7 d2d3 f8d6 e1g1 h8f8 b1c3 f7g8 c1d2 c8e6 a1e1 f6g6 c3e4 a8e8 b2b3 h7h6 f2f3 a7a6 e1e2 e6d5 g1h1 f8f7 e3f2 f7f8 f1e1 c6c5 e1f1 g6h5 c2c4 d5c6 d2e3 c6e4');
INSERT INTO evaluations VALUES(2145,30,'cp 291',NULL);
INSERT INTO evaluations VALUES(2145,31,'cp 303',NULL);
INSERT INTO evaluations VALUES(2145,32,'cp 291',NULL);
INSERT INTO evaluations VALUES(2145,33,'cp 294',NULL);
INSERT INTO evaluations VALUES(2145,34,'cp 300',NULL);
INSERT INTO evaluations VALUES(2145,35,'cp 297',NULL);
INSERT INTO evaluations VALUES(2145,36,'cp 291','d5c6 b7c6 f3e3 e7f7 d2d3 f8d6 e1g1 c8e6 b1c3 h8f8 c3e4 f6g6 c1d2 f7g8 a1e1 a8e8 e3g3 g6h5 g3g5 h5f7 b2b3 h7h6 g5h4 f7g6 f2f3 f8f7 h4f2 a7a6 e1e2 e8f8 g1h1 g8h7 f2g3 g6h5 g3e1 h7g8 e1f2 g7g5 f1e1');
INSERT INTO evaluations VALUES(39107,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(39107,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(39107,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(39107,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(39107,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(39107,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(39107,36,'cp 69','c7c5 f1c4 b8c6 d2d3 h7h6 f3g1 a7a6 a2a4 g8f6 g1e2 c6b4 e1g1 e8g8 f2f4 e5f4 c1f4 d6e7 e2g3 d7d6 d1d2 d6d5 e4d5 f6d5 c3d5 b4d5 a1e1 c8e6 f4e5 d8d7 d2f2 d5f6 e5c3 b7b5 c4e6 f7e6 f2e2 f6d5 c3e5 f8f1 e1f1 e7f6 g3h5 f6e5 e2e5 a8e8 a4b5');
INSERT INTO evaluations VALUES(13238,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(13238,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(13238,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(13238,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(13238,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(13238,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(13238,36,'cp -16','e7e6 d3d4 c5d4 f3d4 f8e7 h2h3 e8g8 e1g1 c6d4 d1d4 b7b6 f1d1 c8b7 e4e5 d6e5 d4e5 d8b8 e5b8 f8b8 c1f4 b8d8 a2a4 a7a6 c4f1 a8c8 f4e3 f6d7 d1d2 g8f8 a1d1 f8e8 f1e2 d7c5 d2d4 d8d4 d1d4 e7f6');
INSERT INTO evaluations VALUES(13930,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(13930,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(13930,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13930,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(13930,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(13930,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(13930,36,'cp 0','e1g1 f8e7 d3d4 c5d4 f3d4 e8g8 h2h3 c6d4 d1d4 c8d7 b2b3 d7c6 c1b2 a7a6 a2a4 d8a5 f1e1 f8e8 a1d1 a8d8 c4d3 a5e5 d4e5 d6e5 f2f3 f6d7 c3b1 f7f6 d3e2 e7b4 b2c3 b4c5 g1f1 d7f8 d1d8 e8d8 c3a5 d8d7 e1d1 d7d1 e2d1 g8f7 b1c3');
INSERT INTO evaluations VALUES(13929,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(13929,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(13929,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(13929,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(13929,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13929,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(13929,36,'cp -16','f8e7 d3d4 c5d4 f3d4 e8g8 h2h3 c8d7 c1e3 d8c7 c4d3 a7a6 f2f4 c6d4 e3d4 d7c6 d1f3 f6d7 f3g3 e7f6 d4f6 d7f6 g1h2 c7b6 a1b1 a8d8 a2a3 d6d5 e4e5 f6e4 g3e1 e4c3 e1c3 g7g6 h3h4 d5d4 c3e1 c6b5 e1e2 b5d3 c2d3 d8d5');
INSERT INTO evaluations VALUES(8160,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8160,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(8160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8160,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(8160,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(8160,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(8160,36,'cp -7','d3d4 c5d4 f3d4 e8g8 h2h3 c8d7 c1e3 a7a6 c4d3 c6d4 e3d4 d7c6 a2a3 e6e5 d4e3 d6d5 e4d5 f6d5 c3d5 d8d5 f2f3 a8d8 d1e2 d5e6 a1e1 e7d6 e3f2 h7h6 f2b6 d8d7 e1d1 e6e7 b6f2 f8e8 d3e4');
INSERT INTO evaluations VALUES(15756,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(15756,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(15756,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(15756,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(15756,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(15756,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(15756,36,'cp 79','f8e7 c1g5 c7c6 e1c1 d8b6 g5f6 g7f6 d4d2 h7h5 f2f4 c8g4 d1e1 b6c5 c4e2 g4e6 c3a4 c5f2 g2g3 e6h3 a4c3 f2b6 e2c4 h3e6 c4e6 f7e6 d2e2 e8c8 e2c4 d6d5 e4d5 e6d5 c3d5 d8d5 e1e7 h5h4 g3g4 h8d8 a2a3 d5d4 c4f7 d4f4 f7e6 c8b8 c1b1');
INSERT INTO evaluations VALUES(8688,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(8688,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(8688,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(8688,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(8688,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(8688,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(8688,36,'cp 73','c1f4 e8g8 e1c1 c8e6 c1b1 f6d7 c4e2 e7f6 d4e3 d8e7 h2h4 f6e5 g2g3 d7b6 b2b3 a8d8 f4e5 d6e5 f2f4 d8d1 h1d1 e5f4 g3f4 e7h4 a2a4 h4e7 e3d4 f7f6 a4a5 b6c8 c3d5 e7d8 e2c4 g8h8 d4c5 e6g8 c5c7 d8c7 d5c7 g8c4 b3c4');
INSERT INTO evaluations VALUES(57287,30,'cp -170',NULL);
INSERT INTO evaluations VALUES(57287,31,'cp -146',NULL);
INSERT INTO evaluations VALUES(57287,32,'cp -157',NULL);
INSERT INTO evaluations VALUES(57287,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(57287,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(57287,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(57287,36,'cp -162','d8h4 e1e2 h4e7 e2f2 d7d5 e4d5 e7h4 f2e2 c8g4 g1f3 g4f3 g2f3 b8d7 d1e1 h4f6 e2d1 g8e7 e1e4 g7g5 b1d2 e8c8 f1d3 d7b6 h2h4 b6d5 h4g5 f6g5 d2c4 g5g2 h1e1 h7h5 c4e5 g2g7 c1d2 h5h4 c2c3 h4h3 d1c2 h3h2 e1h1 f7f6 e5g4 f6f5 e4e6 c8b8 g4h2 d5e3 d2e3 f4e3 e6e3 e7d5');
INSERT INTO evaluations VALUES(35221,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(35221,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(35221,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(35221,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(35221,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(35221,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(35221,36,'cp -133','e1e2 h4e7 e2f2 d7d5 e4d5 e7h4 f2e2 c8g4 g1f3 b8d7 c1f4 g4f3 e2f3 f8d6 d1e2 g8e7 f4d6 c7d6 e2e4 h4h5 f3f2 d7f6 e4d3 e8g8 f1e2 h5g5 b1c3 a8e8 h2h4 g5f4 f2g1 e7f5 a1f1 f4d4 d3d4 f5d4 e2d3 f6g4 h1h3 f7f5 c3b5 d4b5 d3b5 e8e7');
INSERT INTO evaluations VALUES(35220,30,'cp -139',NULL);
INSERT INTO evaluations VALUES(35220,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(35220,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(35220,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(35220,34,'cp -143',NULL);
INSERT INTO evaluations VALUES(35220,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(35220,36,'cp -151','h4e7 e2f2 e7h4 f2e2 d7d5 e4d5 c8g4 g1f3 g4f3 g2f3 b8d7 d1e1 h4f6 e1d2 g7g5 d2d3 e8c8 h2h4 h7h6 b1c3 f8g7 h4g5 h6g5 c3e4 f6g6 h1g1 g6b6 g1g5 g7d4 c1f4 f7f6 g5g2 f6f5 e4c3 g8f6 a1e1 c8b8 e2d1 d4c3 d5d6 c7d6 b2c3 d7e5 d3b5 e5f3 b5b6 a7b6');
INSERT INTO evaluations VALUES(27199,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(27199,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(27199,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(27199,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(27199,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(27199,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(27199,36,'cp 42','g8f6 b5d3 g7g6 e1g1 f8g7 f1e1 d7g4 d4d5 c6e7 b1d2 a7a5 h2h3 g4d7 c3c4 b7b6 d2b1 e8g8 b1c3 f6h5 c1d2 h7h6 a2a3 f7f5 e4f5 d7f5 c3e4 h5f6 e4f6 f8f6 d3e4 f5e4 e1e4 g6g5 d1e2 d8d7 a1e1 e7g6 e2d3 g8h8');
INSERT INTO evaluations VALUES(6967,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(6967,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(6967,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6967,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(6967,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(6967,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(6967,36,'cp 15','g8f6 d2d3 d7d6 a2a4 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 c1e3 d6d5 c4a2 d5e4 c3e4 f6d5 e3d4 e5d4 d1h5 g7g6 h5e5 e7f6 e5g3 f6h4 g3f3 g8g7 a1e1 h4e7 e1e2 d8d7 f1e1 a8e8 a2d5 c6d5 e4g3 e7h4 f3g4 d7g4 h3g4 h4g3 f2g3 e8e2 e1e2');
INSERT INTO evaluations VALUES(30912,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(30912,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(30912,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(30912,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30912,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(30912,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(30912,36,'cp 41','d1e2 f8e7 e1g1 g8f6 f1e1 a7a6 b5d7 d8d7 d2d4 b8c6 d4c5 e8c8 c1e3 h8e8 e2d3 f6g4 b1d2 g4e3 e1e3 e7c5 e3e8 d8e8 d3h7 e8e2 a1f1 g7g6 h7h8 d7d8 h8g7 d8e7 a2a3 c5a7 g7c3 e7c5 c3h8 c8d7 h8h7 c5e7 h2h4 d7c7 h4h5 g6h5 h7h5 e7e6 c2c3 c7b8 b2b4 a7b6 d2b3 e6f6 b3d4 c6d4 c3d4');
INSERT INTO evaluations VALUES(30913,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(30913,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30913,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(30913,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(30913,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(30913,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(30913,36,'cp 28','f8e7 e1g1 g8f6 f1e1 a7a6 b5d7 d8d7 d2d4 b8c6 d4c5 e8c8 c1e3 h8e8 e2d3 d5d4 e3g5 f6d5 b1d2 f7f6 g5h4 d5f4 d3c4 c6a5 c4f1 e7c5 h4g3 e8e1 f3e1 c5d6 d2e4 d6c7 g3f4 c7f4 e1d3 d7f5 a1e1 a5c6 g2g3 f4d6 e4d6 d8d6 f1e2 d6d8 a2a3 c8b8 h2h3');
INSERT INTO evaluations VALUES(28062,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(28062,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(28062,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(28062,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(28062,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(28062,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(28062,36,'cp 26','e1g1 g8f6 f1e1 a7a6 b5d7 d8d7 d2d4 b8c6 d4c5 e8c8 c1e3 h8e8 e2d3 f6g4 b1d2 g4e3 e1e3 e7c5 e3e8 d8e8 d3h7 e8e2 a1f1 g7g6 h7h8 d7d8 h8g7 d8e7 a2a3 c5a7 h2h3 c6e5 a3a4 e5f3 d2f3 e2c2 f3e5 a7c5 e5d3 c5b6 b2b4 c2c4 g7g8 b6d8 g2g3 c8b8 f1e1 c4e4 e1e4 d5e4');
INSERT INTO evaluations VALUES(28063,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28063,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(28063,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(28063,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(28063,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(28063,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(28063,36,'cp 13','g8f6 f1e1 a7a6 b5d7 d8d7 d2d4 b8c6 d4c5 e8c8 c1e3 h8e8 e2d3 f6g4 b1d2 g4e3 e1e3 e7c5 e3e8 d8e8 a1d1 g7g6 g2g3 c5a7 g1g2 f7f6 d2b3 c6e5 f3e5 e8e5 d3f3 e5f5 f3c3 d7c6 c3c6 b7c6 f2f4 f5h5 h2h4 h5f5 g2f3 h7h5 c2c3 a7b6 b3d4 b6d4 d1d4 c8c7 b2b4 c7b6 d4d1 c6c5 b4c5 b6c5');
INSERT INTO evaluations VALUES(54929,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(54929,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(54929,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(54929,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54929,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(54929,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(54929,36,'cp 43','c2c4 e7e6 b1c3 g8f6 g2g3 f8e7 f1g2 e8g8 e1g1 d5c4 f3e5 b8c6 g2c6 b7c6 e5c6 d8e8 c6e7 e8e7 d1a4 a7a5 a4c4 c8a6 c4c6 f6d5 c3d5 e6d5 f1e1 e7b4 c1e3 b4b2 c6d5 a6e2 a1b1 b2c2 b1c1 c2d3 c1c7 a8e8 d5b3 d3b3 a2b3 e2f3 e1a1 e8b8 c7c3 f3d5 a1a5 b8b3 c3b3 d5b3');
INSERT INTO evaluations VALUES(54916,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(54916,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54916,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(54916,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(54916,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(54916,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(54916,36,'cp 30','e7e6 b1c3 g8f6 g2g3 f8e7 f1g2 e8g8 b2b3 b7b6 e1g1 b8d7 c1b2 c8b7 d1d3 c7c5 c4d5 f6d5 c3d5 b7d5 e2e4 d5b7 a1d1 e7f6 f1e1 a8c8 d3e2 d8c7 d4d5 f6b2 e2b2 e6d5 e4d5 c8e8 e1e8 f8e8 f3h4 d7f6 h4f5');
INSERT INTO evaluations VALUES(55461,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(55461,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(55461,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(55461,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(55461,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(55461,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(55461,36,'cp 26','b1c3 g8f6 c1f4 f8d6 f4d6 d8d6 c4d5 e6d5 e2e3 c8f5 f3h4 f5h7 f1d3 e8g8 d3h7 g8h7 h4f5 d6a6 d1c2 g7g6 f5g3 b8c6 c2e2 a6e2 g3e2 c6e7 c3b5 c7c6 b5d6 a8b8 f2f3 e7c8 d6c8 f8c8 e1d2 h7g7 h1c1 c8e8');
INSERT INTO evaluations VALUES(55458,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(55458,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(55458,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(55458,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(55458,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(55458,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(55458,36,'cp 25','g8f6 c1f4 f8d6 f4d6 d8d6 e2e3 e8g8 a1c1 b7b6 f1e2 c8b7 c4d5 f6d5 c3d5 e6d5 d1c2 c7c6 f3e5 b8d7 e5d7 d6d7 e1g1 f8c8 c2a4 c8c7 h2h3 d7d6 c1c3 b7c8 e2f3 a7a5 f1c1 c8d7 a2a3 a8b8 a4c2');
INSERT INTO evaluations VALUES(47674,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(47674,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(47674,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(47674,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(47674,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47674,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(47674,36,'cp 25','c4d5 e6d5 c1f4 f8d6 f4d6 d8d6 d1c2 e8g8 e2e3 a7a6 h2h3 c7c5 d4c5 d6c5 f1e2 b8c6 e1g1 c8e6 a1d1 f8d8 f3d4 c6d4 d1d4 f6e4 f1c1 e4c3 b2c3 b7b5 a2a4 b5a4 d4a4 a6a5 c1a1 c5b6 e2f3');
INSERT INTO evaluations VALUES(47662,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(47662,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(47662,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(47662,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(47662,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(47662,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(47662,36,'cp 12','e6d5 c1f4 c7c5 e2e4 f6e4 f1b5 b8c6 c3e4 d5e4 e1g1 d8b6 d1a4 c8e6 f3e5 a7a6 b5c6 b7c6 e5c6 b6b5 a4b5 a6b5 d4c5 f8c5 a1c1 c5e7 c6e7 e8e7 f4e5 h8c8 c1c8 e6c8 e5g7 a8a2 f1e1 c8b7 g7d4 b7d5 f2f3');
INSERT INTO evaluations VALUES(47714,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(47714,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47714,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47714,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47714,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(47714,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47714,36,'cp 15','c1f4 c7c6 e2e3 c8f5 f3e5 b8d7 f1d3 d7e5 d3f5 e5c4 e1g1 f8d6 b2b3 c4a3 f4d6 d8d6 d1d2 e8g8 c3a4 a3b5 a4c5 d6e7 f2f3 a8e8 f1e1 b7b6 c5a4 b5d6 f5d3 e8c8 a4c3 f8e8 a1c1 c6c5 c3b5 d6b5');
INSERT INTO evaluations VALUES(47709,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(47709,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47709,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(47709,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(47709,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47709,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47709,36,'cp 25','c7c6 e2e3 c8f5 f3e5 f8e7 f1d3 f5d3 e5d3 e8g8 d1b3 d8b6 f2f3 f6h5 b3b6 a7b6 f4e5 b8d7 g2g4 h5f6 e1f2 d7e5 d3e5 e7d6 e5d3 h6h5 g4g5 f6e8 a2a4 g7g6 b2b4 e8g7 f3f4 g7f5 h1b1 f7f6 g5f6 a8e8');
INSERT INTO evaluations VALUES(10396,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10396,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(10396,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(10396,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10396,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10396,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(10396,36,'cp 11','b4c3 d2c3 f6e4 c4f7 e8f7 d1d5 f7e8 d5e4 d7d6 e1g1 d8f6 f3d4 f6g6 e4g6 h7g6 d4c6 b7c6 f1e1 e8f7 f2f4 e5f4 c1f4 c8f5 c3c4 h8b8 b2b4 b8e8 c2c3 a7a5 e1e8 f7e8 c4c5 a5b4 c3b4 d6c5 b4c5 a8a5 f4c7 a5c5 a1e1 e8d7 c7e5 c5a5 e5g7 a5a3');
INSERT INTO evaluations VALUES(10440,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(10440,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10440,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(10440,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10440,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(10440,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(10440,36,'cp 9','d2c3 f6e4 c4f7 e8f7 d1d5 f7e8 d5e4 d7d6 e1g1 d8f6 f3d4 f6g6 e4g6 h7g6 d4c6 b7c6 f1e1 e8d7 c1g5 c6c5 f2f4 e5f4 b2b4 d7c6 b4c5 c6c5 e1e7 c5c6 a1f1 c8f5 f1f4 a8c8 e7g7 f5c2 h2h4 a7a5 g2g4 h8e8 g1f2 c6b6 h4h5 g6h5 g4h5');
INSERT INTO evaluations VALUES(10438,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(10438,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10438,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10438,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10438,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(10438,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10438,36,'cp 12','f6e4 c4f7 e8f7 d1d5 f7e8 d5e4 d7d6 e1g1 d8f6 f3d4 f6g6 e4g6 h7g6 d4c6 b7c6 f1e1 c8f5 f2f4 e8d7 f4e5 h8e8 c1f4 f5c2 c3c4 a7a5 a1c1 c2b3 e5d6 c7d6 c4c5 e8e1 c1e1 d6c5 f4d2 d7d6 g1f2 a8a7 d2c3 a5a4 e1e2 a7e7 e2d2 d6c7 f2g3 b3e6 d2e2 c7d7 h2h3 e6d5 e2e7 d7e7');
INSERT INTO evaluations VALUES(10685,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(10685,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(10685,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(10685,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(10685,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(10685,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(10685,36,'cp 16','c4f7 e8f7 d1d5 f7e8 d5e4 d8f6 e1g1 d7d6 f3d4 f6g6 e4g6 h7g6 d4c6 b7c6 f1e1 e8d7 e1e4 d7e6 c3c4 a7a5 e4e2 c6c5 f2f4 e5e4 e2e4 e6f7 c1d2 c8f5 e4e2 a5a4 a1e1 h8e8 e2e8 a8e8 e1e8 f7e8 d2a5 f5c2 a5c7 e8d7 c7a5 d7e6 g1f2 c2b3 f2g3 e6d7 g3g4 b3c4');
INSERT INTO evaluations VALUES(50239,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(50239,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(50239,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(50239,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(50239,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(50239,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(50239,36,'cp 130','e4e5 f6h5 d2d4 d7d6 g1f3 d6e5 f3e5 f8g7 d1f3 e8g8 c1e3 h5f6 e1c1 c7c6 f1c4 b8d7 c4b3 d8a5 e5c4 a5f5 f3f5 g6f5 h2h3 d7b6 c4a5 b6d5 c3d5 f6d5 e3g5 h7h6 g5d2 g7d4 d2h6 d4g7 h6d2 g7d4 f2f3 d5e3 d1e1 f5f4 h3h4 e7e5 g2g4 c8e6 b3e6 f7e6');
INSERT INTO evaluations VALUES(50149,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(50149,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(50149,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(50149,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(50149,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(50149,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(50149,36,'cp 96','f6h5 d2d4 d7d6 c1d2 h5g7 g1f3 d6e5 d4e5 g7e6 d1e2 c7c5 e1c1 f8g7 e2e4 b8c6 d2e3 d8a5 f1b5 e8g8 b5c6 b7c6 e4c6 a8b8 c6a4 a5c7 c3d5 c7b7 b2b3 f8d8 a4c4 g7f8 d5f4 c8d7 d1d7 b7d7 f4e6 d7e6 c4e6 f7e6 e3c5 d8c8 c5e3 a7a5 f3d2 g8h8 h2h4 a5a4 b3a4 b8b4');
INSERT INTO evaluations VALUES(50275,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(50275,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(50275,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(50275,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(50275,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(50275,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(50275,36,'cp 125','d2d4 d7d6 f2f4 d6d5 g1f3 h5g7 c1e3 h7h5 g2g3 b7b6 d1d2 e7e6 e1c1 c8a6 f1a6 b8a6 h2h3 d8d7 g3g4 f8e7 e3f2 e8c8 f2h4 e7h4 f3h4 c8b7 h4f3 a6b8 a2a3 b8c6 c1b1 d8c8 b2b3 c6a5 b1a2 h5g4 h3g4 h8h1 d1h1 a5c6');
INSERT INTO evaluations VALUES(22792,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(22792,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(22792,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(22792,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(22792,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(22792,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(22792,36,'cp -25','g8f6 f1e2 d7d5 b1d2 f8d6 g1f3 a7a5 e1g1 e8g8 f1e1 h7h6 e4d5 f6d5 h2h3 c8f5 e2f1 d5b6 d2e4 f8e8 e4d6 d8d6 c1e3 b6d5 a2a3 a8d8 d1b3 e5e4 d3e4 f5e4 f3h2 d6g6 b3b7 c6e5 b7a6 e5d3 a6g6');
INSERT INTO evaluations VALUES(16543,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(16543,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16543,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16543,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(16543,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(16543,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(16543,36,'cp -17','b1d2 d7d5 f1e2 f8d6 g1f3 h7h6 e1g1 e8g8 f1e1 a7a5 e4d5 f6d5 h2h3 c8f5 e2f1 f8e8 d2c4 a5a4 c1d2 d5b6 c4b6 c7b6 d3d4 e5d4 e1e8 d8e8 c3d4 f5e6 a2a3 e6b3 d1e1 e8e1 a1e1 c6a5 f1b5 a8d8 d2c3 a5c4 f3d2 c4d2');
INSERT INTO evaluations VALUES(17896,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(17896,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17896,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17896,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17896,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17896,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17896,36,'cp 0','g8f6 d2d3 d7d5 e4d5 f6d5 e1g1 d5b6 c4b3 c8g4 h2h3 g4h5 f1e1 d8e7 d3d4 e8c8 c2c3 f7f6 b1d2 c8b8 a2a4 a6a5 d1c2 h5g6 c2d1');
INSERT INTO evaluations VALUES(11524,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11524,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11524,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11524,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11524,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11524,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11524,36,'cp 0','d2d3 f8c5 e1g1 e8g8 c1e3 c5e3 f2e3 b7b5 c4b3 d7d6 a2a4 c8d7 b1c3 c6a5 b3a2 a5c6');
INSERT INTO evaluations VALUES(11521,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11521,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11521,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11521,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11521,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11521,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11521,36,'cp 0','f8c5 e1g1 e8g8 c1e3 c5e3 f2e3 b7b5 c4b3 d7d6 a2a4 a8b8 d1e1 h7h6 h2h3 c6e7 e1g3 f6h5 g3h4 h5f6');
INSERT INTO evaluations VALUES(7309,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7309,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7309,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7309,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7309,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7309,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7309,36,'cp 0','e1g1 e8g8 c1e3 c5e3 f2e3 b7b5 c4b3 d7d6 a2a4 a8b8 a4b5 a6b5 d1e1 c6e7 b1c3 c7c6 h2h3 d8b6 f3g5 h7h6 e1g3 h6g5 g3g5 g8h7 g5h4 h7g8');
INSERT INTO evaluations VALUES(21510,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(21510,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(21510,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(21510,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(21510,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(21510,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(21510,36,'cp 86','h7h6 f1d3 c6b4 d3c4 g8f6 d1e2 f6e4 e2e4 c8d7 f3e5 b4c6 e5d7 d8d7 c2c3 e8c8 e1g1 f8d6 a2a4 c8b8 b2b4 e6e5 d4d5 c6e7 e4e2 f7f5 a4a5 f5f4 a5a6 b7b6 f1e1 h8f8 c4b5 d7f5 b5d3 f5f7 d3e4 e7d5 e2b5 d5e7 e4f3 f7e8 b5d3 e8g6 d3g6 e7g6');
INSERT INTO evaluations VALUES(33735,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(33735,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(33735,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33735,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(33735,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(33735,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(33735,36,'cp 43','c8g4 g1f3 a5f5 f1d3 g4f3 d3f5 f3d1 a1d1 e7e6 f5h3 b8a6 e1g1 a8d8 c3e2 c7c5 c2c3 f8e7 d2c1 e8g8 g2g3 c5d4 e2d4 a6c5 f1e1 h7h6 h3f1 a7a6 h2h4 f6d5 d4b3 c5b3 a2b3 d8d6 d1d3 f8d8 e1d1 d5f6 b3b4 d6d3 f1d3');
INSERT INTO evaluations VALUES(32652,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(32652,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(32652,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(32652,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(32652,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(32652,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(32652,36,'cp 49','g1f3 a5f5 f1d3 g4f3 d3f5 f3d1 a1d1 e7e6 f5h3 b8d7 e1g1 c7c6 c3e2 f8e7 g2g3 a7a5 a2a4 h7h5 h3g2 h5h4 c2c3 h4h3 g2h1 e7d6 e2c1 e8g8 c1d3 e6e5 d4e5 d6e5 d3e5 d7e5 d2f4 f8e8 f1e1 e5d7 f4e3 d7e5 e3d4 e5c4');
INSERT INTO evaluations VALUES(32650,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(32650,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(32650,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(32650,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(32650,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(32650,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(32650,36,'cp 39','a5f5 f1d3 g4f3 d3f5 f3d1 a1d1 e7e6 f5h3 b8d7 e1g1 c7c6 c3e2 d7b6 b2b3 a7a5 g2g3 a5a4 c2c4 f8e7 d1a1 e8g8 f1b1 f8d8 d2e3 a4b3 a2b3 a8a1 b1a1 d8a8 a1a8 b6a8 e3f4 f6e8 e2c1 a8c7 h3g2 c7a6 c1d3 g7g6 g3g4 a6b4 d3b4 e7b4');
INSERT INTO evaluations VALUES(9839,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9839,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(9839,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9839,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9839,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9839,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9839,36,'cp 0','d7d6 a2a3 b4c3 d2c3 e8g8 e1g1 c8e6 c4e6 f7e6 h2h3 a7a5 c3d2 d8e8 b2b4 b7b6 d1e2 e8h5 b4b5 c6e7 a1e1 e7g6 f3h2 h5h4 h2f3');
INSERT INTO evaluations VALUES(8882,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8882,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8882,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8882,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8882,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8882,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8882,36,'cp 0','a2a3 b4c3 d2c3 e8g8 e1g1 c8e6 c4e6 f7e6 h2h3 a7a5 c3d2 d8e8 b2b4 b7b6 d1e2 e8h5 b4b5 c6e7 a1e1 e7g6 f3h2 h5h4 h2f3');
INSERT INTO evaluations VALUES(8883,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8883,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8883,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8883,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(8883,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8883,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8883,36,'cp 0','b4c3 d2c3 e8g8 e1g1 c8e6 c4e6 f7e6 h2h3 d8e8 b2b4 f6h5 g2g3 b7b5 d1e2 a7a5 f3h4 a5b4 a3b4 e8f7 g1g2 g8h7 f1b1 h5f6 b1f1');
INSERT INTO evaluations VALUES(50155,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(50155,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50155,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(50155,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(50155,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50155,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50155,36,'cp 0','f8g7 b1d2 c7c5 g5f6 g7f6 d2e4 d8b6 e4f6 e7f6 d4c5 b6b2 d1d4 b2d4 f3d4 b8a6 d4b3 b7b6 c5b6 a7b6 a2a4 e8e7 e1d2 d7d6 e2e3 a6c5 b3c5 d6c5 f1b5 a8a5 h1b1 c8e6 d2c3 e7d6 b1d1 d6e7');
INSERT INTO evaluations VALUES(43136,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(43136,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(43136,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(43136,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(43136,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43136,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(43136,36,'cp 8','b1d2 c7c5 g5f6 g7f6 d2e4 d8b6 e4f6 e7f6 d4c5 b6b2 d1d4 b2d4 f3d4 b8a6 d4b3 b7b6 c5b6 a7b6 a2a4 e8e7 e1d2 d7d6 e2e3 a6c5 b3c5 d6c5 f1b5 h8d8 d2c3 c8e6 h1d1 f6f5 d1d8 e7d8 a1d1 d8c7 h2h4 a8d8 d1d8 c7d8 f2f3 d8e7 g2g3 h7h6 f3f4 f7f6');
INSERT INTO evaluations VALUES(43137,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43137,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43137,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(43137,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(43137,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(43137,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43137,36,'cp 0','c7c5 g5f6 g7f6 d2e4 d8b6 e4f6 e7f6 d4c5 b6b2 d1d4 b2d4 f3d4 b8a6 d4b3 e8e7 e2e3 b7b6 c5b6 a7b6 e1d2 a6c5 f2f3 c8a6 b3d4 h8c8 f1a6 a8a6 h1b1 c5b7 b1b4 b7a5 a2a4 a5c4 d2d3 c4e5 d3d2');
INSERT INTO evaluations VALUES(48246,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48246,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(48246,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48246,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(48246,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(48246,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(48246,36,'cp -10','c7c5 b1c3 c5d4 e3d4 f8b4 c4c5 b4c3 b2c3 e8g8 g1f3 b7b6 f4d6 f8e8 f1b5 c8d7 b5d3 f6e4 d3e4 d5e4 f3d2 b6c5 d4c5 d7b5 c3c4 b5c6 e1g1 b8d7 d1h5 e6e5 a1d1 d8c8 f1e1 d7f6 h5h4 h7h6 h4g3 c6b7 d2b3 b7c6');
INSERT INTO evaluations VALUES(46168,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(46168,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(46168,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(46168,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(46168,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(46168,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(46168,36,'cp 0','d4c5 f8c5 g1f3 e8g8 f1e2 b7b6 e1g1 c8b7 c4d5 f6d5 f4g3 b8d7 b1d2 a8c8 a2a3 a7a6 h2h3 d5f6 f3e5 d7e5 g3e5 c5e7 e5d4 e7c5');
INSERT INTO evaluations VALUES(50345,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(50345,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(50345,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(50345,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(50345,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(50345,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(50345,36,'cp -88','f6e4 g1f3 d7d5 c2c4 b8c6 f1e2 e7e6 b1c3 f8b4 d1c2 e4d6 c4d5 b4c3 b2c3 e6d5 e1g1 c6e7 a2a4 e8g8 f1e1 c8g4 c2b3 f8e8 h2h3 g4f3 e2f3 c7c6 c1f4 b7b6 b3b4 d6c4 f4g5 f7f6 g5f4 e7g6 f4g3 h7h6');
INSERT INTO evaluations VALUES(50444,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(50444,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(50444,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(50444,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(50444,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(50444,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(50444,36,'cp -104','g1f3 g7g6 f1d3 e4f6 c2c4 d7d6 b1c3 f8g7 e1g1 e8g8 f1e1 b8d7 c1f4 b7b6 d3e2 c8b7 d1d2 e7e6 f4g5 a7a6 a1d1 d8e7 h2h4 h7h5 g5h6 a8d8 h6g7 g8g7 f3g5 c7c5 d4d5 e6d5 c4d5');
INSERT INTO evaluations VALUES(45179,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(45179,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(45179,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(45179,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(45179,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(45179,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45179,36,'cp 18','b1c3 c5d4 f3d4 a7a6 h2h3 e7e5 d4b3 c8e6 f1e2 f8e7 e1g1 h7h6 d1d3 b8d7 a2a4 a8c8 c1e3 e8g8 a4a5 d8c7 f1c1 d7c5 b3c5 d6c5 e2f3 f8d8 d3e2 d8d7 c1d1 c8d8 d1d7 d8d7 g2g3 c7c6');
INSERT INTO evaluations VALUES(47806,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(47806,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(47806,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(47806,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47806,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47806,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(47806,36,'cp 21','g1f3 f8e7 e1g1 e8g8 b1c3 h7h6 h2h3 c8e6 f1e1 d8d7 a2a4 e7d8 c4e6 f7e6 d3d4 e5d4 d1d4 e6e5 d4c4 d7f7 c4f7 f8f7 e1d1 b8a6 f3e1 a6c5 f2f3 a7a5 c1e3 c5e6 e3f2 d8e7 e1d3 a8f8 d1d2 c7c6 a1d1 f6d7');
INSERT INTO evaluations VALUES(12072,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(12072,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(12072,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(12072,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12072,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12072,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(12072,36,'cp 0','h7h6 g5f3 e5e4 f3e5 f8d6 d2d4 e4d3 e5d3 e8g8 e1g1 d8c7 h2h3 c6c5 b1a3 c8a6 c1f4 a8d8 d1c1 c5c4 f4d6 c7d6 d3f4 f6e4 c1e3 e4d2 f1e1 f8e8 e3g3 d2e4 g3e3');
INSERT INTO evaluations VALUES(12049,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(12049,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12049,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12049,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(12049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12049,36,'cp 0','g5f3 e5e4 f3e5 f8c5 e1g1 d8d4 e5g4 c8g4 e2g4 a5c4 b1c3 e4e3 g4f3 e3f2 g1h1 c4e5 c3e2 d4h4 g2g3 h4c4 b2b3 c4e6 e2f4 e6f5 c1b2 e5f3 d1f3 e8g8 h1g2 f6e4 a1d1 e4g5 f3d3 g5e4 d3f3');
INSERT INTO evaluations VALUES(12052,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12052,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12052,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12052,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12052,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12052,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12052,36,'cp 0','e5e4 f3e5 f8c5 e1g1 e8g8 g1h1 f8e8 f2f4 e4f3 e5f3 c5b6 b1c3 c6c5 e2b5 e8e6 d2d3 c8d7 b5d7 d8d7 c1f4 a8d8 f3e5 d7e8 e5g4 f6g4 d1g4 a5c6 g4g3 c6d4 g3f2 d4f5 f4d2 f5d4');
INSERT INTO evaluations VALUES(12055,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(12055,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(12055,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(12055,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12055,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12055,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12055,36,'cp 0','f3e5 f8c5 e1g1 e8g8 g1h1 f8e8 f2f4 e4f3 e5f3 c5b6 b1c3 c6c5 e2b5 e8e6 d2d3 c8d7 a2a4 a7a6 b5d7 d8d7 c1f4 a8d8 f3d2 c5c4 d2c4 a5c4 d3c4 d7c6 c3d5 f6d5 c4d5 d8d5 d1f3 e6f6 b2b3 c6d7 f3e2 f6e6 e2f3');
INSERT INTO evaluations VALUES(12046,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(12046,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(12046,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12046,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(12046,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(12046,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12046,36,'cp 0','f8d6 d2d4 e4d3 e5d3 e8g8 e1g1 d8c7 g2g3 c8f5 b1d2 c6c5 b2b3 c5c4 b3c4 a8b8 f1e1 f8e8 e2f1 f5g4 f1e2 g4f5');
INSERT INTO evaluations VALUES(7723,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7723,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7723,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7723,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7723,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7723,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7723,36,'cp 0','d2d4 e4d3 e5d3 e8g8 e1g1 d8c7 g2g3 c8f5 b1d2 c6c5 b2b3 c5c4 b3c4 a8d8 c1b2 f6e4 d2e4 f5e4 e2f3 e4f5 f3g4 f5g6 c4c5 d6c5 b2e5 c7c6 g4f3 c6c8 f3g4');
INSERT INTO evaluations VALUES(7722,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(7722,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(7722,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(7722,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7722,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(7722,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7722,36,'cp 2','e4d3 e5d3 e8g8 e1g1 d8c7 e2f3 d6h2 g1h1 h2d6 b2b3 f8e8 c1b2 f6h7 b1d2 c8f5 f3g4 f5d3 c2d3 c6c5 d2f3 a5c6 f1e1 c6e5 f3e5 d6e5 b2e5 e8e5 e1e5 c7e5 a1c1 a8e8 h1g1 e5b2 c1c5 b2a2 d3d4 a2a3 g4f3 a3b2 d4d5 h7g5 d5d6 g5f3 g2f3 e8e2');
INSERT INTO evaluations VALUES(7724,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7724,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7724,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7724,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7724,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7724,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7724,36,'cp 0','e5d3 e8g8 g2g3 c8f5 e1g1 d8c7 b1d2 c6c5 b2b3 c5c4 b3c4 a8d8 c1b2 f6e4 d2e4 f5e4 e2f3 e4f5 f3g4 f5g6 c4c5 d6c5 b2e5 c7c6 g4f3 c6c8 f3g4');
INSERT INTO evaluations VALUES(7726,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7726,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7726,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7726,36,'cp 0','e8g8 e1g1 d8c7 g2g3 c8f5 b1d2 c6c5 b2b3 c5c4 b3c4 a8d8 c1b2 f6e4 d2e4 f5e4 e2f3 e4f5 f3g4 f5g6 c4c5 d6c5 b2e5 c7c6 g4f3 c6c8 f3g4');
INSERT INTO evaluations VALUES(2916,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2916,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2916,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2916,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2916,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2916,36,'cp 0','e1g1 d8c7 g2g3 c8f5 b1d2 c6c5 b2b3 c5c4 b3c4 a8d8 c1b2 f6e4 d2e4 f5e4 e2f3 e4f5 f3g4 f5g6 c4c5 d6c5 b2e5 c7c6 g4f3 c6c8 f3g4');
INSERT INTO evaluations VALUES(38225,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(38225,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38225,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(38225,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(38225,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(38225,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(38225,36,'cp 24','g8f6 f1e1 c7c5 a2a4 e8g8 d2d3 b8c6 c2c3 c8e6 b1a3 d8c7 d1e2 a7a6 c1d2 e6d7 c4a2 h7h6 h2h3 f8b8 a3c2 b7b5 a4b5 a6b5 c2e3 d7e6 a2e6 f7e6 e3c2 g8h7 b2b4 a8a4 d2e3 b8a8 a1a4 b5a4');
INSERT INTO evaluations VALUES(40778,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40778,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40778,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(40778,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(40778,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(40778,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40778,36,'cp 37','b1c3 c7c5 d2d3 b8c6 c1g5 e8g8 g5f6 e7f6 c3d5 g7g6 a2a4 f6g7 c2c3 g8h8 d1d2 c8e6 f1b1 f7f5 d2g5 d8d7 a4a5 e6d5 c4d5 c6e7 d5a2 b7b5 a5b6 a7b6 g5e3 b6b5 f3g5 g7h6 g5f7 f8f7 e3h6 f7f8 a2b3 e7g8 b3g8 a8a1 b1a1 h8g8');
INSERT INTO evaluations VALUES(40762,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(40762,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40762,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40762,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(40762,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(40762,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(40762,36,'cp 35','b8c6 d2d4 c8g4 d4e5 d6e5 c4b5 d8d6 b5c6 b7c6 d1e2 d6e6 e2e3 e8g8 b2b3 a7a5 f3e1 f6d7 e1d3 e6g6 f2f3 g4e6 c1b2 e7d6 f1e1 a8e8 c3a4 f7f5 a4c5 d7c5 d3c5 f5e4 c5e4 a5a4 a1d1 e6d5 e3d3 a4b3 a2b3');
INSERT INTO evaluations VALUES(8582,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8582,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(8582,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(8582,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(8582,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(8582,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(8582,36,'cp 42','d2d4 c8g4 d4e5 d6e5 c4b5 d8d6 b5c6 b7c6 d1e2 d6e6 e2e3 e8g8 b2b3 f6d7 h2h3 g4h5 c3e2 h5g6 e2g3 c6c5 f1d1 f8d8 c1b2 f7f6 c2c4 a7a5 f3h4 a5a4 h4f5 e7f8 e3f3 e6c6 b2c3 g6f5 g3f5 a4b3 a2b3 a8a1 c3a1');
INSERT INTO evaluations VALUES(8617,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(8617,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8617,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(8617,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(8617,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(8617,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(8617,36,'cp 38','c8g4 d4e5 d6e5 c4b5 d8d6 b5c6 b7c6 d1e2 d6e6 e2e3 e8g8 b2b3 f6d7 h2h3 g4h5 c3e2 h5g6 e2g3 c6c5 f1d1 f8d8 c1b2 f7f6 f3h4 c5c4 g3f5 e7c5 e3f3 d7f8 d1d5 e6b6 b2c1 f8e6 c1e3 c5e3 f5e3 e6g5 f3d1 c7c6 h4g6 c6d5 g6e7 g8f8 e3d5 g5e4 d1e2 e4g3');
INSERT INTO evaluations VALUES(52523,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52523,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52523,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(52523,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(52523,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52523,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(52523,36,'cp 11','c3d4 g8f6 g1f3 d7d5 b1c3 b8c6 c1f4 c8f5 e2e3 e7e6 f1b5 f6d7 a1c1 f8e7 e1g1 d8b6 d1b3 a7a6 b5c6 b7c6 b3b6 d7b6 f4c7 b6d7 f1d1 c6c5 c7a5 a8c8 d4c5 c8c5 a5b4 c5c4 b4e7 e8e7 b2b3 c4c8 f3d4 f5g6 f2f3');
INSERT INTO evaluations VALUES(52514,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(52514,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52514,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52514,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(52514,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52514,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52514,36,'cp 11','g8f6 b1c3 d7d5 c1f4 b8c6 e2e3 c8f5 a1c1 a7a6 g1f3 e7e6 f1e2 a8c8 e1g1 f8d6 f4d6 d8d6 c3a4 e8g8 a4c5 c8c7 a2a3 f5e4 h2h3 f8c8 f3d2 h7h6 b2b4 a6a5 d2e4 f6e4 c5e4 d5e4 b4a5 c6a5 c1c7 c8c7');
INSERT INTO evaluations VALUES(44622,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44622,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(44622,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(44622,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(44622,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44622,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(44622,36,'cp 17','b1c3 d7d5 c1f4 b8c6 g1f3 c8f5 e2e3 e7e6 f1b5 f6d7 a1c1 a8c8 e1g1 f8e7 a2a3 g7g5 f4e5 e8g8 e5g3 a7a6 b5d3 f5g6 h2h3 d7b6 g3h2 g6d3 d1d3 f7f5 c3e2 b6c4 c1c2 b7b5 f1c1 d8e8 f3e1 b5b4 a3a4 g5g4 h3g4');
INSERT INTO evaluations VALUES(31156,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(31156,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(31156,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(31156,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(31156,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(31156,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31156,36,'cp 0','f5e4 f2f3 e4g6 e5e6 f7e6 h2h4 h7h5 f1d3 g6d3 d1d3 d8d6 g1e2 b8d7 d3g6 e8d8 c1g5 g8f6 g4h5 e6e5 b1c3 e5d4 e2d4 e7e5 d4f5 d6e6 g5f6 e6f6 e1c1 d8c7 c3e2 f8c5 f3f4 e5f4 e2f4 c5e3 f5e3 f6f4 g6g5 f4f3 h1f1 f3h5 g5g7 a8f8 g7c3 h5h4 f1f8 h8f8 d1d5 c7c8 d5d1');
INSERT INTO evaluations VALUES(31147,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(31147,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(31147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31147,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(31147,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(31147,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(31147,36,'cp -18','g1f3 e7e6 c1e3 b8d7 b1c3 d8a5 f1d3 e4d3 d1d3 c6c5 h2h4 g8e7 d4c5 e7c6 e1c1 f8c5 e3c5 d7c5 d3e3 c5a4 c3a4 a5a4 a2a3 a4e4 h1e1 e4e3 f2e3 h7h5 g4g5 e8e7 c2c3 g7g6 e3e4 d5e4 e1e4 h8d8 d1d2 a7a6 d2d8 a8d8');
INSERT INTO evaluations VALUES(57650,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(57650,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(57650,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(57650,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(57650,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(57650,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(57650,36,'cp 8','c7c5 b1c3 b8c6 e2e3 f8g7 g1e2 e7e6 f1g2 g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 h2h4 e8g8 e1g1 d8b6 f1e1 c6d4 c3d5 d4e2 e1e2 e6d5 d1d5 f8d8 c1e3 d8d5 e3b6 c8g4 g2d5 g4e2 b6e3 a8d8 d5g2 g7b2 a1e1 e2g4 g2b7 g4e6 e3a7 e6a2 a7b6');
INSERT INTO evaluations VALUES(55910,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(55910,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(55910,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55910,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(55910,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(55910,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(55910,36,'cp -12','e4d5 e6d5 d2d4 b8c6 c2c3 c8e6 g1e2 f8d6 e2f4 g8f6 h2h3 e8g8 c1e3 d8d7 b1d2 a8e8 f1d3 e6f5 d3f5 d7f5 e1c1 d6f4 e3f4 e8e7 h1e1 f8e8 e1e7 e8e7 g2g4 f5d7 d2b3 b7b6 b3d2 a7a5 g4g5 f6e4 d2e4 e7e4 f3g3 e4e2 g3g4 d7g4 h3g4 e2f2 f4c7');
INSERT INTO evaluations VALUES(55780,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(55780,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55780,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55780,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55780,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(55780,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(55780,36,'cp -13','e6d5 d2d4 b8c6 c2c3 c8e6 g1e2 f8d6 e2f4 g8f6 h2h3 e8g8 c1e3 c6e7 f1d3 e6d7 b1d2 c7c6 e1g1 e7g6 f4h5 f6e8 f3e2 g6h4 e3f4 g7g6 f4d6 e8d6 e2e5 d6e8 e5g3 d7f5 h5f4 e8g7 d2f3 h4f3 g3f3 d8f6 d3f5 f6f5 f3g3 a8e8 a1e1 e8e4 e1e4 d5e4 g3e3 f8e8');
INSERT INTO evaluations VALUES(56291,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(56291,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(56291,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(56291,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(56291,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(56291,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(56291,36,'cp -11','d2d4 b8c6 f1b5 g8f6 h2h3 c8d7 c2c3 h7h6 g1e2 f8d6 e1g1 e8g8 b5d3 f6e4 c1f4 d8f6 d3e4 d5e4 f3e3 a8e8 b1d2 g7g5 f4d6 f6d6 f2f4 e4f3 e3f3 f7f5 e2g3 c6e7 d2c4 d6e6 c4e5 f5f4 f3g4 e6g4 h3g4 f4g3 e5d7 f8f1 a1f1');
INSERT INTO evaluations VALUES(56209,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(56209,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(56209,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(56209,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(56209,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(56209,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(56209,36,'cp -14','g8f6 h2h3 b8c6 f1b5 c8d7 c2c3 h7h6 g1e2 c6e7 b5d7 d8d7 c1f4 e7g6 b1d2 g6f4 e2f4 f8e7 e1c1 e8g8 d1e1 c7c6 f4d3 e7d6 e1e2 a8e8 d3e5 d7c7 h1e1 d6e5 e2e5 e8e5 e1e5 f8e8 e5e2 e8e2 f3e2 b7b6 a2a4');
INSERT INTO evaluations VALUES(56350,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(56350,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(56350,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56350,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(56350,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(56350,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(56350,36,'cp 26','c4d5 d8d5 g1f3 b8c6 b1c3 d5d6 f1g2 g8f6 d2d3 f8e7 e1g1 e8g8 c1g5 c6d4 f3d2 c7c6 e2e3 d4e6 g5f6 e7f6 c3e4 d6e7 d1c2 f8d8 e4f6 e7f6 h2h4 e6c7 d2b3 a7a5 f1d1 h7h6 d3d4 e5d4 d1d4 c8e6 d4d8 f6d8');
INSERT INTO evaluations VALUES(56170,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(56170,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(56170,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(56170,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(56170,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(56170,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(56170,36,'cp 14','g8f6 b1c3 f6d5 f1g2 d5b6 g1f3 b8c6 e1g1 f8e7 b2b3 e8g8 c1b2 c8f5 d2d4 e5d4 c3b5 f5e4 f3d4 c6d4 d1d4 d8d4 b5d4 e4g2 g1g2 e7f6 a1b1 f8e8 e2e3 c7c6 d4f3 b6d5 b3b4 a8d8 f1c1 a7a6 a2a3 g7g6 g2f1 g8g7 f1e2 d8d6 f3d2 f6b2 b1b2 d5f6 c1c4 e8d8');
INSERT INTO evaluations VALUES(51405,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(51405,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(51405,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(51405,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(51405,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(51405,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(51405,36,'cp 67','f8b4 e2e4 c8b7 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 f1d3 d7d6 g1h3 g8f6 e1g1 b8c6 d1e2 e6e5 d4d5 c6b8 c4c5 b6c5 c3c4 b8d7 h3f2 e8g8 c1d2 b7a6 f2d1 a6c8 d1e3 h7h6 d2c3 f8e8 d3c2 a7a5 e3f5 e7d8 f3f4 f6h7');
INSERT INTO evaluations VALUES(37562,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(37562,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(37562,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(37562,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(37562,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(37562,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(37562,36,'cp 39','e2e4 c8b7 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 g1e2 g8f6 h2h4 d7d6 d1a4 b7c6 a4c2 c6b7 e1f2 b8c6 c2a4 f6d7 c1g5 f7f6 g5e3 e8g8 e2f4 a8e8 h4h5 e7f7 d4d5 c6d8 f1h3 d7e5 h1e1 a7a5 a2a3 f6f5 e4f5');
INSERT INTO evaluations VALUES(56681,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(56681,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(56681,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(56681,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(56681,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(56681,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(56681,36,'cp 82','g7g6 a2a4 f8h6 c1h6 g8h6 b1c3 b8d7 h2h4 a7a6 f3d2 d8e7 g2g4 f6f5 g4f5 d7f6 f5g6 h7g6 f1e2 h6f7 d2c4 c8d7 d1d2 f6h5 e2h5 h8h5 d2e3 h5h4 e1c1 h4h1 d1h1 b7b6 b2b4 e7f6 c1b2 f6f4 e3f4 e5f4 f2f3');
INSERT INTO evaluations VALUES(22714,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(22714,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(22714,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(22714,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(22714,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(22714,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(22714,36,'cp -32','g8f6 d2d3 d7d5 e4d5 f6d5 g1f3 c8g4 e1g1 d5b6 c4b5 f8d6 b1d2 e8g8 b5c6 b7c6 h2h3 g4h5 d2e4 f7f5 e4g3 h5g6 f1e1 f8e8 c1g5 d8d7 g5d2 a7a6 f3h4 c6c5 d1e2 a6a5 h4g6 h7g6 g3f1 a5a4 a1d1 a8d8');
INSERT INTO evaluations VALUES(16415,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16415,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(16415,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(16415,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(16415,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(16415,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(16415,36,'cp -12','d2d3 d7d5 e4d5 f6d5 g1f3 d5b6 c4b3 d8f6 b1c3 c8e6 c3b5 f8d6 b5d6 c7d6 b3e6 f6e6 e1g1 e8g8 f1e1 d6d5 b2b4 f7f6 a3a4 c6e7 c2c3 f8c8 c1d2 b6d7 h2h3 d7f8 b4b5 h7h6 a1b1 b7b6 d1b3 e6f7 a4a5 f8d7');
INSERT INTO evaluations VALUES(16413,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16413,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(16413,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(16413,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(16413,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(16413,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(16413,36,'cp -27','d7d5 e4d5 f6d5 g1f3 c8g4 e1g1 d5b6 h2h3 g4h5 c4a2 d8e7 b1c3 e8c8 f1e1 c8b8 c1e3 e7e8 c3e4 f7f5 e4g3 h5f3 d1f3 g7g6 a1d1 c6d4 e3d4 d8d4 g3e2 d4d8 d3d4 e5e4 f3e3 e8a4 a2b3 a4c6 f2f3 e4f3 e3f3 c6f3 g2f3 f8d6 g1g2 c7c6 f3f4');
INSERT INTO evaluations VALUES(15093,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(15093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15093,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(15093,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(15093,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(15093,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(15093,36,'cp -8','e4d5 f6d5 g1f3 c8g4 h2h3 g4h5 e1g1 d5b6 c4a2 d8f6 c1e3 h5f3 d1f3 f6f3 g2f3 f8e7 b1d2 h8f8 f3f4 e5f4 e3f4 e8c8 d2f3 e7f6 a1b1 h7h6 f1e1 b6d5 f4g3 d5e7 g1g2 e7f5 h3h4 a7a6 e1e4 c6d4 f3d4 d8d4 e4d4 f5d4 c2c3 d4f5 h4h5 f5g3');
INSERT INTO evaluations VALUES(14991,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(14991,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(14991,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(14991,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(14991,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(14991,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(14991,36,'cp -7','f6d5 g1f3 d5b6 c4a2 c8f5 b1c3 d8d7 f3g5 f5g6 f2f4 e5f4 d1e2 f8e7 c1f4 e8c8 e1c1 e7f6 e2d2 f6c3 d2c3 b6d5 c3c4 d5f4 c4f4 h8e8 d1e1 c6d4 f4f2 a7a6 g5f3 d4f3 f2f3 c8b8 a2c4 f7f6 g2g3 e8e7 e1e7 d7e7 h1f1 e7e5 f3f4 e5e2 f4f2 e2h5 c4a2 h5g5 c1b1 d8e8');
INSERT INTO evaluations VALUES(15327,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(15327,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(15327,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(15327,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(15327,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(15327,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(15327,36,'cp -6','g1f3 c8g4 h2h3 g4h5 e1g1 d5b6 c4a2 d8f6 c1e3 h5f3 d1f3 f6f3 g2f3 a8d8 f1e1 f8e7 g1g2 b6d5 b1c3 d5e3 f2e3 e8g8 g2f1 e7h4 e1d1 g7g6 f1e2 g8g7 c3e4 b7b6 c2c3 f7f5 e4d2 c6e7 a3a4 h4f6 a2c4 c7c6 d3d4 e5d4 c3d4 f8e8 f3f4 e7d5 c4d5 d8d5 d2f3');
INSERT INTO evaluations VALUES(15326,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(15326,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(15326,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(15326,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(15326,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(15326,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(15326,36,'cp -12','c8g4 h2h3 g4h5 e1g1 d5b6 c4a2 d8f6 c1e3 h5f3 d1f3 f6f3 g2f3 f8e7 f1e1 a8d8 g1g2 b6d5 b1c3 d5e3 f2e3 e8g8 e1g1 g7g6 g2f1 g8g7 f1e2 d8d7 a1b1 f7f5 a2e6 d7d6 e6b3 e7h4 g1g2 g7h6 c3d5 d6d7 f3f4 e5f4 d5f4');
INSERT INTO evaluations VALUES(28017,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(28017,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(28017,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(28017,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(28017,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(28017,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(28017,36,'cp 108','d7d6 c1f4 e7e6 d1e2 a7a6 c4b3 h7h6 e1c1 b8d7 c1b1 b6b5 h2h4 g8e7 a2a3 d7b6 f4e3 d8d7 f3e1 e7c6 d4d5 c6a5 e3b6 a5b3 c2b3 c7b6 e1c2 e8c8 e2d2 g7c3 d2c3 c8b8 f2f3 d7c7 c2b4 e6d5');
INSERT INTO evaluations VALUES(27896,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(27896,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(27896,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(27896,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(27896,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(27896,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(27896,36,'cp 129','h2h4 h7h6 c1f4 e7e6 d1e2 a7a6 c4b3 b8d7 e1c1 b6b5 a2a3 d7b6 h4h5 g6g5 f4e3 g8f6 d4d5 e6e5 f3d2 b7c8 f2f3 e8g8 c1b1 c8d7 c3a2 a6a5 c2c4 b5b4 c4c5 d6c5 e3c5 b4a3 c5a3 f8e8 a2c3 a5a4 b3c4 d8b8 g2g4');
INSERT INTO evaluations VALUES(55553,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55553,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(55553,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55553,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(55553,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(55553,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55553,36,'cp -6','c7c5 g2g3 b8c6 f1g2 g7g6 d2d3 f8g7 b1c3 g8e7 e1g1 e8g8 a1b1 a7a5 b2b3 a8b8 a2a3 c8e6 c1d2 b7b6 c3b5 f7f6 h2h3 d8d7 g1h2 g8h8 f3h4 g6g5 h4f3 e6f7 d2c3 e7g6 f3e1 c6d4');
INSERT INTO evaluations VALUES(46754,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(46754,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46754,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(46754,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46754,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(46754,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46754,36,'cp 12','f8g7 c4d5 f6d5 g1f3 e8g8 f1c4 d5c3 b2c3 c7c5 e1g1 d8c7 d1e2 b8c6 c1a3 c7a5 a3b2 c8g4 h2h3 g4f3 e2f3 a8c8 f1c1 c5d4 c3d4 e7e5 d4e5 g7e5 f3e2 c8c7 a1b1 e5b2 e2b2 f8d8 c1d1 d8d1 b1d1 a5e5 b2d2 g8g7 f2f4 e5c5 c4b3 c6a5 b3d5 a7a6 a2a4 a5c4');
INSERT INTO evaluations VALUES(52858,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52858,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(52858,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52858,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(52858,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(52858,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(52858,36,'cp 44','d2d4 c6c5 g1f3 c5d4 d1d4 b8c6 d4f4 g7g6 b1c3 g8h6 f1b5 a7a6 b5c6 b7c6 e1g1 h6f5 c3a4 d8a5 c1d2 a5b5 b2b3 c6c5 c2c4 d5c4 f4c4 b5c4 b3c4 f5d4 a4c5 d4f3 g2f3 f8g7 f3f4 c8f5 d2e3 g6g5 f1d1 a8c8 f4g5 g7e5');
INSERT INTO evaluations VALUES(22833,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(22833,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(22833,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(22833,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(22833,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(22833,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(22833,36,'cp -32','e5f4 c1f4 d7d5 e4d5 d8d5 b1c3 d5a5 f1e2 g8f6 e2f3 c6d4 d1d2 c8e6 f3b7 a8b8 g1e2 d4c2 d2c2 b8b7 e1g1 f8e7 g1h1 e8g8 e2d4 e6d7 a1e1 f8e8 d4b3 a5a6 f4e3 h7h6 b3c5 e7c5 e3c5 e8e1 f1e1 d7b5 c3b5 a6b5 e1e7 b5b2 c2b2 b7b2 e7c7 b2a2 c5a7');
INSERT INTO evaluations VALUES(22965,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(22965,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(22965,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(22965,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(22965,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(22965,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(22965,36,'cp -30','c1f4 d7d5 e4d5 d8d5 b1c3 d5a5 f1e2 f8a3 d1c1 g8f6 f4d2 a3e7 g1f3 e8g8 a2a3 c8g4 e1g1 a8e8 g1h1 e7d6 d2g5 e8e6 g5f6 e6f6 c3e4 f6e6 e2d1 g4f3 d1f3 a5e5 e4d6 e5d6 f3c6 d6c6 c1d2 f8e8 d2f2 f7f6 f1e1 e6e1 a1e1 e8e1 f2e1 g8f7 c2c3 g7g6 e1e2');
INSERT INTO evaluations VALUES(22958,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(22958,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(22958,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(22958,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(22958,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(22958,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(22958,36,'cp -34','d7d5 e4d5 d8d5 b1c3 d5a5 f1e2 f8a3 d1c1 g8f6 f4d2 a3e7 g1f3 e8g8 e1g1 c8g4 a2a3 a8e8 g1h1 e7d6 d2g5 e8e6 g5f6 e6f6 c3e4 f6e6 e2d1 g4f3 d1f3 a5e5 e4d6 e5d6 f3c6 d6c6 c1d2 f8e8 a1e1 h7h6 h2h3 b7b6 e1e6 e8e6 h1g1 c6e8 d2f4 c7c5 c2c3 a7a5 f1f2 a5a4 d3d4 e6e3 d4d5 e3d3');
INSERT INTO evaluations VALUES(37355,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(37355,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(37355,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(37355,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(37355,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(37355,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(37355,36,'cp 15','h7h6 g5h4 c7c5 d4d5 d7d6 e2e3 c8f5 f1d3 f5d3 d1d3 b8d7 e1g1 g6g5 h4g3 f6h5 a1b1 f7f5 h2h3 d8e8 g3h2 e8f7 c3b5 a8c8 b5a7 c8a8 a7b5 a8a2 b5c7 f7f6 f3d2 f8c8 c7e6 d7f8 e6g7 h5g7 d3b3 c8a8 b3b7 f8g6 g1h1');
INSERT INTO evaluations VALUES(37348,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(37348,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(37348,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(37348,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37348,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(37348,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(37348,36,'cp 16','g5h4 c7c5 d4d5 d7d6 e2e3 g6g5 h4g3 c8f5 f1d3 f5d3 d1d3 f6h5 e3e4 b7b5 c3b5 g7b2 a1b1 b2g7 e1g1 b8d7 b1b3 a7a6 b5c3 d8a5 c3d1 h5g3 h2g3 g5g4 b3b7 a5a4 f3h4 d7e5 d3d2 a4c4 h4f5 f8b8 b7e7 c4e4 d1e3 b8b1 f5d6 b1f1 g1f1 e4b1 d2d1');
INSERT INTO evaluations VALUES(37349,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(37349,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(37349,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(37349,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(37349,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(37349,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37349,36,'cp 16','c7c5 d4d5 d7d6 e2e3 c8f5 f1d3 f5d3 d1d3 g6g5 h4g3 b8d7 e1g1 f6h5 e3e4 a7a6 c3d1 b7b5 d3c2 g7d4 f1e1 f8e8 d1e3 d4e3 e1e3 e7e5 d5e6 e8e6 a1d1 d8e7 b2b3 a8e8 c4b5 a6b5 a2a4 h5g3 h2g3 e6e4 e3e4 e7e4 c2e4');
INSERT INTO evaluations VALUES(34067,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(34067,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(34067,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(34067,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(34067,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(34067,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(34067,36,'cp 187','f1d3 e4f6 e1g1 d7d6 e5f3 e7d8 d1e2 f8e7 f1e1 c8e6 c2c4 d6d5 b1c3 d5c4 d3c4 e6c4 e2c4 e8g8 c4b3 b8c6 d4d5 c6a5 b3a4 b7b6 c1g5 h7h6 g5h4 f8e8 e1e2 a5b7 a1e1 b7d6 f3e5 e7f8 a4d1 d8c8 h4f6 g7f6 e5c6 c8f5 h2h3 e8e2 e1e2 f5h5 b2b4 g8h8 a2a3 f8g7');
INSERT INTO evaluations VALUES(34064,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(34064,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(34064,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(34064,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(34064,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(34064,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(34064,36,'cp 165','e4f6 e1g1 d7d6 e5f3 e7d8 f1e1 f8e7 d1e2 c8e6 c2c4 d6d5 b1c3 b8c6 c4c5 e8g8 h2h3 d8d7 a2a3 f8e8 d3b5 d7c8 e2d1 a7a6 b5d3 e6f5 c1f4 f5g6 b2b4 c6d8 f4e5 d8c6 e5f6 e7f6 d3g6 e8e1 d1e1 h7g6 c3d5');
INSERT INTO evaluations VALUES(34012,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(34012,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(34012,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(34012,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(34012,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(34012,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(34012,36,'cp 144','e1g1 d7d6 e5f3 e7d8 d1e2 f8e7 f1e1 c8e6 c2c4 d6d5 b1c3 b8c6 c1g5 e8g8 c4d5 e6d5 a1d1 d5f3 e2f3 h7h6 g5h4 c6d4 f3e3 d4e6 d3f5 d8c8 e3h3 g8h8 h4f6 e7f6 c3d5 f6g5 h3g4 c7c6 h2h4 c6d5 h4g5 c8c4 g5h6 c4g4 f5g4 g7h6');
INSERT INTO evaluations VALUES(34011,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(34011,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(34011,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(34011,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(34011,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(34011,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(34011,36,'cp 188','d7d6 e5f3 e7d8 f1e1 f8e7 d1e2 c8e6 c2c4 d6d5 b1c3 d5c4 d3c4 e6c4 e2c4 e8g8 c4b3 d8d6 b3b7 b8c6 c1f4 d6f4 b7c6 f4d6 c6a4 d6b6 a4c2 e7b4 e1e5 b4d6 c3a4 b6a6 e5e1 a8e8 a2a3 h7h6 h2h3 f6d5 a4c5 a6c6 e1e8 f8e8');
INSERT INTO evaluations VALUES(21348,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(21348,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(21348,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21348,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21348,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(21348,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21348,36,'cp 28','c8d7 b1c3 c6a5 b5d3 c7c5 d4c5 a5c6 g1f3 f8c5 e1g1 a7a6 c3e2 d8c7 c1f4 g8e7 a1c1 c7b6 c1b1 e7g6 d3g6 f7g6 b2b4 c5e7 a2a3 a8d8 c2c3 e8g8 d1d2 b6b5 f1e1 h7h6 h2h4 b5c4 e2d4 c6d4 c3d4 d8c8 e1c1 c4b5 d2e3 c8c1 b1c1');
INSERT INTO evaluations VALUES(27176,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(27176,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(27176,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(27176,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(27176,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(27176,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(27176,36,'cp 22','b1c3 c6a5 b5d3 c7c5 d4c5 a5c6 g1f3 f8c5 e1g1 a7a6 c3e2 d8c7 c1f4 g8e7 a1c1 c7b6 c1b1 c6b4 a2a3 b4d3 c2d3 e7g6 f4g3 b6d8 h2h4 g6e7 h4h5 e7f5 g3f4 h7h6 b1c1 c5b6 e2g3 f5g3 f4g3 e8g8 g3h4 d8e8 d3d4 d7b5 f1e1 a8c8 d1d2 e8d7 g2g4 c8c1 e1c1');
INSERT INTO evaluations VALUES(27174,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(27174,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(27174,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(27174,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(27174,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(27174,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(27174,36,'cp 18','c6a5 b5d3 c7c5 d4c5 a5c6 g1f3 f8c5 e1g1 a7a6 c3e2 d8c7 c1f4 g8e7 a1c1 c7b6 c1b1 c6b4 a2a3 b4d3 c2d3 e7g6 f4g3 b6d8 h2h4 g6e7 h4h5 e7f5 g3f4 h7h6 d3d4 c5e7 d1d2 e8g8 f1c1 d7b5 e2c3 b5c6 d2d3 d8d7 g2g4 f5h4 f3h4 e7h4');
INSERT INTO evaluations VALUES(5581,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(5581,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(5581,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(5581,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(5581,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(5581,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(5581,36,'cp 86','b1c3 f7f6 f1d3 f6e5 f3e5 g8f6 e1g1 e8g8 c1e3 c8d7 a1c1 a8c8 d3b1 b4d6 f2f4 c6e7 d1e2 d6b4 e3f2 d8e8 f1e1 e7g6 e5d7 e8d7 g2g3 c8c6 b1d3 a7a6 a2a3 b4c3 b2c3 f8c8 c3c4 d5c4 c1c4 g6e7 c4c6 c8c6');
INSERT INTO evaluations VALUES(5578,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(5578,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(5578,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(5578,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(5578,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(5578,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(5578,36,'cp 88','f7f6 f1d3 f6e5 f3e5 g8f6 c1e3 e8g8 e1g1 c8d7 a2a3 b4a5 c3a4 a5c7 e5d7 f6d7 d1g4 d8e7 a1c1 c7d6 f2f4 g7g6 g1h1 a8e8 b2b4 e7g7 d3b5 d7f6 g4e2 g7c7 a4c5 d6c5 d4c5 a7a6 b5c6 c7c6 e3d4 f6e4');
INSERT INTO evaluations VALUES(38885,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(38885,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(38885,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(38885,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(38885,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(38885,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(38885,36,'cp 86','f1c4 g8f6 e4e5 f6d5 c1d2 e8g8 d1g4 f7f5 e5f6 d5f6 g4h4 f6d5 h4d8 f8d8 g1f3 d5b6 c4b3 b8c6 a2a3 b4e7 c3b5 b6d5 e1g1 a7a6 b5c3 e7f6 d2e3 a6a5 f1e1 c6e7 c3e4 a5a4 b3a2 a8a5 e3d2 a5b5 d2c3 g8f8 h2h4 e7f5 h4h5 d5c3 b2c3 f5d6');
INSERT INTO evaluations VALUES(38870,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(38870,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(38870,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(38870,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(38870,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(38870,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(38870,36,'cp 73','g8f6 e4e5 f6d5 c1d2 e8g8 d1g4 f7f5 e5f6 d5f6 g4h4 f6d5 h4d8 f8d8 g1f3 c7c5 d4c5 b4c5 e1g1 h7h6 f1d1 b8c6 d2e1 d5f6 d1d8 c6d8 f3e5 c5d6 e5g6 c8d7 a1d1 d6c7 f2f4 d7e8 g6e5 a8c8 c4b3 g8f8 e1f2 b7b6 d1e1 c7e5');
INSERT INTO evaluations VALUES(51594,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(51594,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(51594,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(51594,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(51594,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(51594,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(51594,36,'cp 162','d4d5 a7a6 c2c4 d7d6 b1c3 e7e5 h2h4 h7h5 g2g3 g7g6 c1g5 d8c7 f1h3 c8h3 g1h3 b8d7 g5d2 f8h6 h3g5 g8f6 e1g1 c7d8 d1e2 d8e7 g1g2 a8b8 f2f3 b8b7 b2b3 e8g8 a1e1 h6g7 f3f4 e5f4 g3f4 f6g4');
INSERT INTO evaluations VALUES(51601,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(51601,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(51601,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(51601,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(51601,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(51601,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(51601,36,'cp 157','a7a6 c2c4 d7d6 b1c3 e7e5 h2h4 h7h5 g2g3 g7g6 c1g5 d8c7 g1h3 f8h6 f2f3 b8d7 h3f2 d7f8 f1e2 a8b8 g5d2 c7e7 d1c2 c8d7 a2a4 h6d2 c2d2 g8h6 e1c1 f7f6 d1g1 d7c8 g1f1 f6f5 e2d1 f8d7 e4f5');
INSERT INTO evaluations VALUES(22831,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(22831,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(22831,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(22831,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(22831,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(22831,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(22831,36,'cp -50','e5f4 g1f3 g7g5 g2g3 f8g7 d2d4 d7d6 d4d5 c6e5 g3f4 g5f4 c1f4 c8g4 f1e2 e5f3 e2f3 d8f6 e1g1 g4h3 f3g2 h3g2 g1g2 f6g6 g2h3 g7c3 b2c3 e8c8 d1g4 d8d7 f4g5 f7f6 g5h4 g6e8 h4f6 g8f6 f1f6 h8g8 g4f5 c8b8 a1f1 d7e7 f1f4 a7a6 f6e6 g8g6 f5f8 e7e6 d5e6 g6e6 f8e8 e6e8');
INSERT INTO evaluations VALUES(22962,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(22962,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(22962,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(22962,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(22962,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(22962,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(22962,36,'cp -40','g1f3 g7g5 g2g3 d7d6 d2d4 f8g7 d4d5 c6e5 g3f4 g5f4 c1f4 c8g4 f1e2 g4f3 e2f3 d8f6 f4e5 f6e5 d1d3 g8f6 e1c1 e8c8 d3e3 a7a6 h1g1 h8g8 c1b1 c8b8 g1g5 e5e7 g5f5 f6d7 c3e2 g7f6 e2d4 d8e8 d1f1 d7c5 a2a3 g8g5 b1a2 c5e4 e3e4 e7e4 f3e4 f6d4 e4d3');
INSERT INTO evaluations VALUES(22961,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(22961,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(22961,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(22961,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(22961,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(22961,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(22961,36,'cp -42','g7g5 g2g3 f8g7 d2d4 d7d6 d4d5 c6e5 g3f4 g5f4 c1f4 c8g4 f1e2 g4f3 e2f3 d8f6 f4e5 f6e5 d1d3 g8f6 e1c1 e8c8 f3e2 c8b8 d3g3 f6d7 d1f1 d8g8 g3e5 d7e5 c3b5 g7h6 c1d1 h6e3 c2c3 g8g2 b5d4 h7h5 h2h3 h5h4 d4f3 e5g6 f3e1 g2g3 f1f7 g6e5 f7f5 a7a5 d1c2 e3h6 e1f3');
INSERT INTO evaluations VALUES(22493,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(22493,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(22493,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(22493,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(22493,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(22493,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22493,36,'cp -40','g2g3 f8g7 d2d4 d7d6 d4d5 c6e5 g3f4 g5f4 c1f4 c8g4 f1e2 g4f3 e2f3 d8f6 f4e5 f6e5 d1d3 g8f6 e1c1 e8c8 f3g2 c8b8 g2h3 d8e8 h3f5 e5f4 d3d2 g7h6 d2f4 h6f4 c1b1 a7a6 d1f1 f4d2 f1f2 d2c3 b2c3 h8g8 f2f4 e8e5 b1b2 b8a7 c3c4 a7b6 f4h4 h7h5 h1f1 f6g4 f5g4 h5g4 f1f7 e5e4');
INSERT INTO evaluations VALUES(33209,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(33209,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(33209,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(33209,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(33209,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(33209,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(33209,36,'cp 42','b8c6 c2c3 d7d6 b2b4 c5b6 c4b3 a7a6 e1g1 g8e7 a2a4 c8e6 b1d2 e6b3 d1b3 e8g8 d2c4 f6e6 f3g5 e6d7 c4b6 c7b6 f2f4 h7h6 g5h3 d7e6 b3e6 f7e6 c1e3 b6b5 e3b6 b5a4 a1a4 e7c8 b6e3 b7b5 a4a2 a6a5 f4e5');
INSERT INTO evaluations VALUES(283,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(283,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(283,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(283,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(283,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(283,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(283,36,'cp 53','c2c3 d7d6 b2b4 c5b6 c4b3 g8e7 a2a4 a7a5 b4b5 c6d8 b1d2 f6g6 e1g1 c8h3 f3h4 g6f6 g2h3 f6h4 d1g4 h4g4 h3g4 e7g6 d2c4 d8e6 c4b6 c7b6 b3e6 f7e6 c1e3 g6f4 f1d1 g7g5 d3d4 e5d4 e3f4 g5f4 c3d4 h8g8 h2h3 h7h5 g1g2 h5g4');
INSERT INTO evaluations VALUES(281,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(281,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(281,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(281,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(281,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(281,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(281,36,'cp 39','d7d6 b2b4 c5b6 c4b3 a7a6 b1a3 g8e7 a3c4 b6a7 a2a4 h7h6 c1e3 e8g8 e3a7 c6a7 c4e3 c8e6 e1g1 a7c6 f1e1 e6b3 d1b3 f6e6 e3d5 e6d7 c3c4 e7g6 a1d1 a8e8 d5e3 a6a5 b4b5 c6b4 b3c3 c7c5 b5c6 b7c6 d3d4 e5d4 f3d4 e8e4 d4f5');
INSERT INTO evaluations VALUES(22891,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(22891,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(22891,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(22891,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(22891,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(22891,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(22891,36,'cp -12','b1c3 g8f6 g1f3 f8d6 e1g1 e8g8 f1e1 c6e5 f3e5 d6e5 f2f4 e5c3 b2c3 f8e8 c1a3 d7d5 e4d5 c8f5 d1b3 d8d7 b3b7 f5e4 b7c6 d7f5 a3e7 f6d5 c4d5 e4d5 c6c5 a8b8 c3c4 d5e4 c5f5 e4f5 e7c5 a7a5 e1e8 b8e8 c5f2 h7h5 a1e1 e8b8 f2a7 b8b4 e1e5 f5d3 e5a5 b4c4 a5h5');
INSERT INTO evaluations VALUES(22889,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(22889,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(22889,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(22889,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(22889,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(22889,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(22889,36,'cp -35','d7d6 g1f3 g8f6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5f6 g7f6 e5c4 a8d8 c4a5 b7a8 a5c4 h8g8 g2g3 d8d1 a1d1 g8g5 d1e1 g5h5 b2b3 a8b7 c3e2 b7c8 e2f4 h5c5 h2h3');
INSERT INTO evaluations VALUES(22008,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(22008,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(22008,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(22008,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(22008,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(22008,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(22008,36,'cp -10','g1f3 g8f6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5d7 f8d8 d7f6 e7f6 g3c7 d8d1 a1d1 c6c5 c7d6 f6d4 c3b5 d4b2 d6c5 a7a5 h2h4 g5h4 b5d6 b7c6 d6f5 c6e4 f5h6 g8g7 h6g4 f7f5 c5d4 b2d4 d1d4 f5g4 d4e4');
INSERT INTO evaluations VALUES(22007,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(22007,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(22007,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(22007,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(22007,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(22007,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(22007,36,'cp -11','g8f6 d1b3 d8d7 e1g1 c6a5 b3b4 a5c4 b4c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5d7 f8d8 d7f6 e7f6 g3c7 d8d1 a1d1 c6c5 c7d6 f6d4 c3b5 d4b2 d6c5 a7a5 h2h4 g5h4 b5d6 b7a6 d1d5 b2g7 c5b6 a8b8 b6a5 b8b2 d6f5 g7f8 a5d8 h4h3 g2h3 b2a2 f5e7 f8e7');
INSERT INTO evaluations VALUES(15620,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(15620,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(15620,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(15620,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(15620,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(15620,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(15620,36,'cp -19','d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5d7 f8d8 d7f6 e7f6 g3c7 d8d1 a1d1 a8e8 c7d6 f6d4 h2h4 c6c5 c3b5 d4b2 d6c5 e8e5 d1b1 e5c5 b1b2 a7a5 b5d4 b7d5 h4g5 h6g5 a2a3 g8g7 f2f3 c5c4 b2d2 g7f6');
INSERT INTO evaluations VALUES(46964,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(46964,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(46964,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(46964,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46964,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(46964,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(46964,36,'cp 4','c2c4 g7g6 b1c3 f8g7 e2e4 e7e6 g1f3 e8g8 f1d3 c7c6 e1g1 c6d5 c4d5 e6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 a2a4 a8c8 c1f4 f8e8 e1e8 f6e8 a1e1 d8b6 d3d2 h7h5 e1e7 b6b4 c3d1 b4d2');
INSERT INTO evaluations VALUES(46960,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(46960,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46960,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(46960,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(46960,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(46960,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46960,36,'cp 0','g7g6 b1c3 f8g7 e2e4 e7e6 f1d3 e6d5 c4d5 c7c6 g1f3 c6d5 e4d5 e8g8 e1g1 c8g4 c1g5 g4f3 d1f3 b8d7 h2h3 h7h6 g5f4 d8b6 f3e2 f8e8 e2d2 d7c5 f4e3 b6b4 d3c2 a7a6 a1b1 b4h4 e3f4 c5e4 c2e4 f6e4 c3e4 e8e4 f4d6 e4d4');
INSERT INTO evaluations VALUES(46625,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(46625,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(46625,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(46625,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46625,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(46625,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46625,36,'cp 0','b1c3 f8g7 g1f3 e8g8 e2e4 e7e6 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 a2a4 a8c8 c1g5 d8b6 d3d2 f8e8 e1e8 f6e8 a4a5 b6b4 a1a4 b4b3 a4a3 b3b4 a3a4');
INSERT INTO evaluations VALUES(46624,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(46624,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46624,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46624,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46624,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(46624,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46624,36,'cp 0','f8g7 g1f3 e8g8 e2e4 e7e6 f1d3 c7c6 e1g1 c6d5 e4d5 e6d5 c4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 a2a4 a8c8 c1f4 d8b6 c3e4 f6e4 e1e4 b6b2 a1b1 c8c3 b1b2 c3d3 b2b7 d3d5 b7a7 g6g5 f4g3 g7e5 a7a6 e5g3 f2g3 f8b8');
INSERT INTO evaluations VALUES(41189,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41189,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41189,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41189,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41189,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41189,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41189,36,'cp 0','g1f3 e8g8 e2e4 e7e6 f1d3 c7c6 e1g1 c6d5 e4d5 e6d5 c4d5 c8g4 f1e1 b8d7 h2h3 g4f3 d1f3 d7e5 f3d1 e5d3 d1d3 a7a6 a2a4 a8c8 c1g5 d8b6 d3d2 b6b4 a4a5 f8e8 e1e8 c8e8 g5f4 f6d7 a1a4 b4b3 a4a3 b3b4');
INSERT INTO evaluations VALUES(41188,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41188,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41188,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41188,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41188,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41188,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41188,36,'cp 0','e8g8 e2e4 e7e6 f1d3 c7c6 e1g1 e6d5 c4d5 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 a7a6 d3f1 f8e8 c1g5 d8b6 e1e8 a8e8 a1b1 f6h5 g2g4 h5f6 g5f4 d7e5 f3d1 b6b4 f4g5 f6d7 g5e3 e5c4 f1c4 b4c4 b1c1');
INSERT INTO evaluations VALUES(47176,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(47176,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(47176,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(47176,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(47176,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(47176,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(47176,36,'cp -54','c7c5 e2e3 c8f5 f1d3 f5g6 g1f3 b8c6 e1g1 e7e6 d3g6 h7g6 d1e2 a7a6 f1d1 d8c7 c3a4 c5d4 e3d4 b7b5 c1f4 c7b7 a4c3 f8e7 c3b1 e8g8 a2a3 a6a5 c2c3 b5b4 h2h4 f8c8 g1g2 f6e4 a3b4 a5b4 a1a8');
INSERT INTO evaluations VALUES(45373,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(45373,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(45373,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(45373,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(45373,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(45373,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(45373,36,'cp -44','e2e3 b8c6 g1f3 c5d4 e3d4 c8g4 f1g2 e7e6 c3e2 f8d6 c2c3 e8g8 e1g1 b7b5 c1f4 a7a5 h2h3 g4f5 f1e1 a8c8 f3e5 c6e7 e5d3 f5d3 d1d3 d6f4 e2f4 b5b4 c3b4 a5b4 a2a4 b4a3 b2a3 d8a5 h3h4 e7f5 e1c1 c8c1 a1c1 f5d6');
INSERT INTO evaluations VALUES(49579,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(49579,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(49579,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(49579,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(49579,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(49579,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(49579,36,'cp -15','e5d4 e3d4 d7d5 b1c3 f8b4 g1f3 e8g8 f1e2 c8e6 f3g5 d8d7 c4c5 e6g4 e1g1 b4c3 e2g4 f6g4 b2c3 b8c6 c1f4 h7h6 g5f3 f8e8 f1e1 e8e1 f3e1 b7b6 e1d3 g4f6 d1c2 a8e8 f2f3 f6h5 f4g3 h5g3 h2g3');
INSERT INTO evaluations VALUES(49753,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(49753,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(49753,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(49753,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(49753,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(49753,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(49753,36,'cp -12','e3d4 d7d5 b1c3 f8b4 g1f3 e8g8 f1d3 b8c6 e1g1 d5c4 d3c4 c8g4 c1e3 d8d6 a1c1 f8e8 h2h3 g4h5 c3a4 d6d7 a2a3 b4f8 a4c3 a8d8 f1e1 a7a6 c4f1 b7b5 g2g4 h5g6 f1g2 h7h5 f3h4 c6d4 h4g6 f7g6 e3d4 e8e1 d1e1 d7d4');
INSERT INTO evaluations VALUES(49745,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(49745,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(49745,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(49745,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(49745,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49745,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(49745,36,'cp -17','f8b4 b1c3 e8g8 g1f3 d7d5 f1d3 d5c4 d3c4 b8c6 c1e3 c8g4 e1g1 d8d6 a1c1 a8d8 h2h3 g4h5 c3a4 d6d7 a2a3 b4e7 c4e2 f6d5 e2b5 f8e8 a4c5 e7c5 c1c5 h7h6 f1e1 a7a6 b5a4 d5e3 e1e3 h5f3 d1f3 d7d4 a4c6 b7c6 e3e8 d8e8 c5c6 d4b2 c6c7');
INSERT INTO evaluations VALUES(42837,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(42837,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(42837,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(42837,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(42837,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(42837,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(42837,36,'cp 0','b1c3 e8g8 g1f3 d7d5 f1e2 d5c4 e2c4 c8g4 e1g1 g4f3 d1f3 d8d4 f3b7 d4c4 b7a8 b4d6 a8a7 d6h2 g1h2 c4f1 f2f3 f1e1 a7c7 f6g4 f3g4 e1h4 h2g1 h4e1');
INSERT INTO evaluations VALUES(42834,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(42834,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(42834,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(42834,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(42834,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(42834,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(42834,36,'cp -1','e8g8 g1f3 d7d5 f1e2 c8e6 c4c5 f6e4 d1b3 b4c3 b2c3 b7b6 e1g1 b6c5 d4c5 d8f6 f3d4 e4c5 b3c2 f8e8 c1e3 b8c6 d4c6 e6f5 c2d1 f6c6 c3c4 e8d8 c4d5 d8d5 e2f3 f5e4 f3e4 c5e4 a1c1 c6e6 d1b3 a8c8 f1d1 e4f6 d1d5 e6d5 e3a7 d5b3');
INSERT INTO evaluations VALUES(37280,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(37280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37280,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(37280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37280,36,'cp 0','f1e2 d7d5 g1f3 c8e6 c4c5 f6e4 d1b3 b4c3 b2c3 b7b6 e1g1 h7h6 f1e1 b6c5 d4c5 f8e8 b3c2 e4c5 c1e3 c5e4 c3c4 b8c6 a1c1 c6b4 c2b2 b4c6');
INSERT INTO evaluations VALUES(50223,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50223,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(50223,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(50223,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50223,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50223,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50223,36,'cp 20','d7d5 c2c4 c7c6 c4d5 c6d5 b1c3 f8g7 f1g2 e8g8 f3e5 c8f5 e1g1 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 d8b6 e2e4 d5e4 a1b1 b6d8 b1b7 f8e8 a4c6 e7e5 f4e3 e5d4 c3d4 f5e6 g2e4 a8c8 c6a4 c8c4 a4a6 g7d4 f1d1');
INSERT INTO evaluations VALUES(46801,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(46801,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(46801,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46801,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(46801,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(46801,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(46801,36,'cp 15','c2c4 c7c6 c4d5 c6d5 f1g2 f8g7 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 e7e6 e1g1 f8c8 f1c1 h7h6 h2h4 c6e5 f4e5 g6g5 a2a3 c8c6 h4g5 h6g5 g2f1 f6g4 e5g7 g8g7 f1b5 c6c7 c1e1 a8c8 f2f3 g4f6 b5f1 f5h7 a1c1 g5g4');
INSERT INTO evaluations VALUES(46760,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46760,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46760,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(46760,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(46760,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(46760,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46760,36,'cp 13','c7c6 c4d5 c6d5 b1c3 f8g7 f1g2 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 h7h6 h2h4 e7e6 e1g1 f8d8 g2f3 f6h5 e5c6 b7c6 f4c7 d8d7 c7b6 d7b7 b6c5 h5f6 b2b4 f6d7 c5d6 e6e5 d4e5 d7e5 d6e5 g7e5');
INSERT INTO evaluations VALUES(44712,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44712,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44712,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44712,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44712,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44712,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44712,36,'cp 12','c4d5 c6d5 f1g2 f8g7 f3e5 e8g8 b1c3 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 h7h6 h2h4 e7e6 e1g1 f8c8 f1c1 c6e5 f4e5 g6g5 a2a3 f6g4 e5g7 g8g7 c3b5 f5d3 c1c8 a8c8 b5c3 g4f6 a1c1 g7g6 c3a2 c8c1 a2c1 d3a6');
INSERT INTO evaluations VALUES(55112,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(55112,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(55112,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(55112,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(55112,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(55112,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(55112,36,'cp -27','e7e6 d2d4 c7c5 f1d3 g8f6 d4c5 b8c6 g1f3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 c1e3 c5e3 f2e3 a8c8 a1c1 d8c7 d1e2 f8d8 g1h2 c7e7 f1e1 e6b3 d3c2 b3c2 c1c2');
INSERT INTO evaluations VALUES(55939,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55939,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(55939,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(55939,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(55939,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(55939,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(55939,36,'cp -27','d2d4 c7c5 g1f3 g8f6 d4c5 f8c5 f1d3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 h2h3 h7h6 a2a3 a7a5 c2c3 d4c3 e2c3 c8e6 c1e3 c5e3 f2e3 d8c7 a1c1 f8d8 d1e2 a8c8 c3a4 c7e7 a4c5 e6a2 d3b5 c6a7 b2b4 c8c7 b5d3 b7b6 e2a2');
INSERT INTO evaluations VALUES(13573,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(13573,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13573,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(13573,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(13573,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(13573,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(13573,36,'cp -16','f6h5 f4e5 d8b6 d1c2 c5d4 e5d4 b6c7 d3b5 f8d6 c3c4 d5c4 e1g1 e8g8 d4c3 c6e5 c3e5 d6e5 f3e5 c7e5 a2a4 a7a6 b5c4 c8d7 b1d2 d7c6 f1c1 a8c8 c4f1 h5f6 d2c4 e5g5 c2c3 f8d8 c3e5 g5e5 c4e5 c6e8 a4a5 h7h6 h2h3 f6e4 c1c8 d8c8 a1d1');
INSERT INTO evaluations VALUES(13468,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(13468,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(13468,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(13468,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(13468,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(13468,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(13468,36,'cp -14','f4e5 d8b6 d1c2 c5d4 e3d4 f7f6 e5g3 c8d7 a2a4 a8c8 c2e2 g7g6 e1g1 f8g7 c3c4 e8g8 b1c3 c6d4 f3d4 b6d4 f1d1 d4c5 c4d5 e6d5 d3b5 h5g3 h2g3 d7c6 e2e6 g8h8 c3d5 c8e8 b2b4 e8e6 b4c5 f6f5 d5f4 e6e4 a1c1 g7f6 b5c6 b7c6');
INSERT INTO evaluations VALUES(44518,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(44518,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(44518,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(44518,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(44518,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(44518,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(44518,36,'cp -23','e7e6 g1f3 d7d5 c4b5 b8d7 e1g1 a7a6 b5d7 c8d7 b1d2 d5e4 d3e4 f8e7 b2b3 e8g8 c1b2 b7b5 f3e5 d8c7 e5d7 f6d7 c2c4 e7d6 d2f3 b5c4 b3c4 d6e5 f3e5 d7e5 b2e5 c7e5 d1e2 a6a5 g2g3 f8d8 a2a4 e5c7 a1b1');
INSERT INTO evaluations VALUES(44183,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(44183,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(44183,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(44183,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(44183,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(44183,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(44183,36,'cp -6','g1f3 d7d5 c4b5 c8d7 b5d7 d8d7 b1c3 d5e4 d3e4 d7d1 e1d1 b8c6 d1e2 c6d4 f3d4 c5d4 c3b5 a8c8 c2c3 f6e4 b5d4 a7a6 h1d1 f8e7 d4b3 e4d6 a2a4 f7f6 d1d4 d6f5 d4d3 h7h5 c1e3 f5e3 e2e3 c8c7 e3e2 g7g5 a1d1');
INSERT INTO evaluations VALUES(44181,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(44181,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(44181,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(44181,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(44181,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(44181,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(44181,36,'cp -7','d7d5 c4b5 c8d7 b5d7 d8d7 b1c3 b8c6 e1g1 h7h6 c1f4 f8e7 f3e5 c6e5 f4e5 e8g8 e4d5 f6d5 c3d5 e6d5 f1e1 f8e8 h2h3 e7f8 d1g4 d7c6 d3d4 c5c4 e1e3 e8e6 c2c3 a8e8 a1e1 b7b5 a2a3 f7f6 e5f4 e6e3 e1e3 e8e3 f4e3 g8f7 g4h5 g7g6 h5e2 a7a6 e3d2 f8e7');
INSERT INTO evaluations VALUES(46156,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(46156,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(46156,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(46156,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(46156,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(46156,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(46156,36,'cp -7','c4b5 c8d7 b5d7 d8d7 b1c3 b8c6 e1g1 h7h6 c1f4 f8e7 f3e5 c6e5 f4e5 e8g8 e4d5 f6d5 c3d5 e6d5 h2h3 a8e8 f1e1 e7g5 d1g4 d7g4 h3g4 e8e6 g1f1 f8e8 e5c3 d5d4 e1e6 f7e6 c3e1 g8f7 f2f3 e6e5 c2c4');
INSERT INTO evaluations VALUES(19712,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19712,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19712,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19712,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19712,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19712,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19712,36,'cp 0','b7c6 c2c4 g7g6 f1e2 f8g7 b1c3 g8f6 e1g1 e8g8 d1c2 f6d7 a1b1 a7a5 c1e3 d8c7 b1d1 c7b7 f2f4 d7c5 f4f5 a8b8 b2b3 b7b4 e3d2 b4a3 d2c1 a3b4');
INSERT INTO evaluations VALUES(18468,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(18468,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18468,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(18468,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(18468,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18468,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18468,36,'cp 0','c2c4 g7g6 f1e2 g8f6 b1c3 f8g7 e1g1 e8g8 d1c2 f6d7 a1b1 a7a5 c1e3 d7c5 b1d1 d8c7 f2f4 c7b7 e2f3 a8b8 b2b3 b7b4 e3d2 b4b6 d2e3');
INSERT INTO evaluations VALUES(16396,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(16396,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(16396,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(16396,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(16396,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(16396,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(16396,36,'cp -135','d7d5 e4d5 c6d4 f3d1 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 g1f3 f8d6 e1g1 e8g8 b1d2 f7f5 b3b4 a5a4 d2b3 f5f4 f1e1 f8e8 b3d2 c8g4 h2h3 g4h5 d2e4 d5f6 b2b3 f6e4 e1e4 a4b3 a1a8 d8a8 d1b3 h5f7 b3d1 a8d5 e4e1 e8a8 f3g5 f7g6 g5e4 d6e7 d1g4 d5d3 g4e6 g8f8 e6e5');
INSERT INTO evaluations VALUES(9958,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(9958,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9958,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(9958,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(9958,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(9958,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(9958,36,'cp 23','b7c6 f1d3 e8g8 c3a4 c5b6 e1g1 f8e8 a4b6 a7b6 f1e1 d7d5 e4d5 e8e1 d1e1 d8d5 c1f4 c8f5 d3f1 d5d4 f4e5 d4c5 e5f6 g7f6 c2c3 c5e5 e1d2 g8g7 a1e1 e5d6 d2d6 c7d6 a2a3 f5e6 f2f4 d6d5 g1f2 f6f5 e1e5 c6c5 f2g3 g7f6 h2h3 h7h6');
INSERT INTO evaluations VALUES(19467,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(19467,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19467,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19467,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19467,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19467,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19467,36,'cp 6','f1g2 g8f6 b1c3 e7e6 g1e2 d7d5 e4d5 e6d5 d2d4 c5d4 e2d4 c8g4 d4e2 f8b4 e1g1 b4c3 b2c3 e8g8 h2h3 g4f5 c1g5 h7h6 g5f6 d8f6 d1d5 f5c2 d5b5 a8e8 b5b7 e8e2 b7c6 f6c6 g2c6 f8c8 c6b5 e2e5 c3c4 c2f5 f1e1 e5e1 a1e1 f5h3 e1e7 h3e6 e7a7 e6c4 b5c4 c8c4 g1g2 c4c3 a7b7 c3a3 b7b2 g8h7 b2b7');
INSERT INTO evaluations VALUES(19468,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19468,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19468,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(19468,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(19468,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19468,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(19468,36,'cp 0','g8f6 b1c3 e7e6 g1f3 d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 c8g4 d4e2 f8b4 e1g1 b4c3 b2c3 e8g8 h2h3 g4f5 c1g5 h7h6 g5f6 d8f6 d1d5 f5c2 d5b5 a8e8 b5b7 e8e2 b7c6 f6c6 g2c6 c2f5 c6f3 e2e5 f1e1 f8e8 e1e5 e8e5 c3c4 f5h3 f3d5 h3e6');
INSERT INTO evaluations VALUES(12844,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(12844,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(12844,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(12844,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(12844,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(12844,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(12844,36,'cp 2','b1c3 e7e6 g1e2 d7d5 e4d5 e6d5 d2d4 c5d4 e2d4 c8g4 d4e2 f8b4 e1g1 b4c3 b2c3 e8g8 h2h3 g4f5 c1g5 h7h6 g5f6 d8f6 d1d5 f5c2 d5b5 a8e8 b5b7 e8e2 b7c6 f6c6 g2c6 f8c8 c6b5 e2e5 c3c4 c2d3 f1d1 d3c4 b5c4 c8c4 d1d8 g8h7 d8d7 e5f5 d7a7 c4c2 a1f1 h6h5 a7a4 c2b2 g3g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(12825,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(12825,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(12825,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(12825,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(12825,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(12825,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12825,36,'cp 0','e7e6 g1e2 d7d5 e4d5 e6d5 d2d4 c5d4 e2d4 c8g4 d4e2 f8b4 h2h3 b4c3 b2c3 g4f5 c1g5 h7h6 g5f6 d8f6 d1d5 e8g8 e1g1 f5c2 d5b5 a8e8 b5b7 e8e2 b7c6 f6c6 g2c6 c2f5 c6f3 e2e5 f3g4 f5g4 h3g4 f8c8 f1d1 c8c3 d1d7 a7a5 a1b1 c3c4 b1b8 g8h7 d7b7 c4g4 b7f7 e5e2 f7f8 e2a2 b8a8 a5a4');
INSERT INTO evaluations VALUES(12428,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(12428,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(12428,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(12428,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(12428,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12428,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(12428,36,'cp 2','g1f3 d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 c8g4 d4e2 f8b4 h2h3 b4c3 b2c3 g4f5 c1g5 h7h6 g5f6 d8f6 d1d5 f5c2 e1g1 e8g8 d5b5 a8e8 b5b7 e8e2 b7c6 f6c6 g2c6 c2f5 c6f3 e2e5 f1e1 e5c5 e1e7 c5c3 f3d5 c3c5 d5b3 f5h3 e7a7 h3e6 g1h1 e6d5 h1h2 f8e8');
INSERT INTO evaluations VALUES(12427,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12427,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12427,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(12427,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(12427,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(12427,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12427,36,'cp 3','d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 c8g4 d4e2 f8b4 h2h3 b4c3 b2c3 g4f5 c1g5 h7h6 g5f6 d8f6 d1d5 e8g8 e1g1 f5c2 d5d2 c2a4 e2f4 a8d8 f4d5 f6d6 f1e1 a4b5 d2e3 b5c4 d5e7 c6e7 e3e7 d8d7 e7d6 d7d6 e1e7 b7b6 e7a7 f8d8 a2a4 d6d1 a1d1 d8d1 g1h2');
INSERT INTO evaluations VALUES(24811,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24811,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(24811,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24811,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(24811,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24811,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(24811,36,'cp 21','e2e3 e7e6 f1b5 f6d7 e1g1 f8d6 f4g3 e8g8 f1e1 c6e7 b5d3 f5d3 c2d3 d6g3 h2g3 c7c5 d4c5 d7c5 e1e2 c5d7 a1c1 d8b6 e2c2 f8c8 a2a3 h7h6 b2b4 b6a6 a3a4 a6b6 c1b1 a7a6 a4a5 b6c7 e3e4 d5e4');
INSERT INTO evaluations VALUES(39630,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(39630,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(39630,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(39630,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(39630,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(39630,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(39630,36,'cp 40','d2d4 c7c5 d4d5 d7d6 a2a4 a7a6 g1f3 g8f6 b1c3 e8g8 f1d3 e7e6 d5e6 f7e6 d3c4 b8c6 e1g1 g8h8 e4e5 d6e5 d1d8 f8d8 f3e5 c6e5 f4e5 f6g4 c4e2 g4e5 c3e4 b7b6 a4a5 d8d4 e4g5 h7h6 g5f7 e5f7 f1f7 h8g8 f7e7 b6a5 a1a5 a8b8 a5c5');
INSERT INTO evaluations VALUES(51933,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51933,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(51933,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(51933,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(51933,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51933,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51933,36,'cp 0','b8c6 g2g3 c6d4 f1g2 g8e7 g1e2 e7c6 d2d3 f8e7 e1g1 d7d6 f2f4 e8g8 e2d4 c5d4 c3b1 c8d7 b2b3 a7a6 b1a3 a8b8 h2h4 b7b5 g1h2 d8c7 a3c2 e7f6 a1b1 f6e7');
INSERT INTO evaluations VALUES(31313,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31313,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31313,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31313,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31313,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31313,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31313,36,'cp 0','g2g4 f5d7 c1e3 e7e6 d1d2 b8c6 e1c1 a8c8 c1b1 d8b6 c3a4 b6b4 a4c3');
INSERT INTO evaluations VALUES(12650,30,'cp 277',NULL);
INSERT INTO evaluations VALUES(12650,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(12650,32,'cp 276',NULL);
INSERT INTO evaluations VALUES(12650,33,'cp 266',NULL);
INSERT INTO evaluations VALUES(12650,34,'cp 263',NULL);
INSERT INTO evaluations VALUES(12650,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(12650,36,'cp 234','d4d3 d7d6 c1f4 f8e7 e1c1 e8g8 f4d6 c8e6 d6e7 d8e7 d3d6 e7d6 d1d6 a8d8 d6d8 f8d8 f2f3 g8f8 f1e2 f6e8 h2h4 h7h6 b2b3 e8c7 h1d1 d8d1 c1d1 f8e7 d1d2 f7f6 f3f4 b7b6 d2e3 e7d7 g2g3 e6f7 e2f3 b6b5 h4h5 d7d6');
INSERT INTO evaluations VALUES(12653,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(12653,31,'cp 245',NULL);
INSERT INTO evaluations VALUES(12653,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(12653,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(12653,34,'cp 232',NULL);
INSERT INTO evaluations VALUES(12653,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(12653,36,'cp 231','d7d6 c1f4 f8e7 e1c1 e8g8 f4d6 c8e6 d6e7 d8e7 d3d6 e7d6 d1d6 f8d8 d6d8 a8d8 f1e2 g8f8 f2f3 f6e8 h1e1 f8e7 e2d3 e7f8 e1d1 f8e7 e4e5 h7h6 f3f4 e6g4 d3e2 g4e6 d1d2 d8d2 c1d2 f7f6 e5f6 e8f6 e2f3 b7b6 h2h3 a7a6 g2g4');
INSERT INTO evaluations VALUES(14013,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(14013,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(14013,32,'cp 245',NULL);
INSERT INTO evaluations VALUES(14013,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(14013,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(14013,35,'cp 256',NULL);
INSERT INTO evaluations VALUES(14013,36,'cp 242','c1f4 f8e7 e1c1 e8g8 f4d6 c8e6 d6e5 d8a5 d3b5 a5b5 f1b5 a8d8 e5c7 d8d4 f2f3 a7a6 b5e2 b7b5 c7e5 d4d7 e5f4 d7d8 g2g3 h7h6 f4c7 d8d4 c7b6 d4d7 d1d7 f6d7 b6a5 d7e5 c3d5 e7d8 a5d8 f8d8 f3f4 e6d5 e4d5');
INSERT INTO evaluations VALUES(57381,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(57381,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(57381,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(57381,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(57381,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(57381,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(57381,36,'cp 148','g8f6 c1g5 d7d5 e2e3 c8e6 g1f3 f8e7 f1e2 c7c5 c2c4 b8c6 e1g1 e8g8 c4d5 f6d5 g5e7 d8e7 d1c2 c6b4 c2c1 g8h8 a2a3 b4c6 b1c3 a7a6 c3d5 e6d5 c1c3 a8d8 a1d1 b7b5 b2b3 h7h6 d1d2 b5b4 a3b4 c5b4 c3b2 a6a5 f1d1 d5f3 e2f3 c6e5');
INSERT INTO evaluations VALUES(50008,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(50008,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(50008,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(50008,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(50008,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(50008,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(50008,36,'cp 158','c1g5 d7d5 e2e3 f8e7 g1f3 c8e6 f1e2 e8g8 e1g1 c7c5 c2c4 b8c6 b1d2 g8h8 a1c1 d5c4 e2c4 e6c4 d2c4 b7b5 c4e5 c6e5 f3e5 c5c4 d1c2 d8c7 e5f3 c7b7 h2h3 a8e8 g5h4 a7a6 c1d1 b7e4 c2e2 e8d8 a2a3 d8d1');
INSERT INTO evaluations VALUES(50005,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(50005,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(50005,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(50005,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(50005,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(50005,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(50005,36,'cp 151','d7d5 e2e3 f8e7 f1e2 e8g8 g1f3 c8e6 e1g1 c7c5 c2c4 b8c6 b1d2 g8h8 a1c1 d5c4 e2c4 e6f5 h2h3 a7a6 f3h4 f5d7 d2f3 b7b5 c4d3 d7e6 b2b3 c6b4 d3b1 d8d1 f1d1 f8d8 b1f5 e6f5 h4f5 b4a2 f5e7 a2c1 d1c1 d8e8');
INSERT INTO evaluations VALUES(22580,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(22580,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(22580,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(22580,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(22580,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(22580,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(22580,36,'cp -30','d7d5 e4d5 e6d5 c4b5 f8d6 e1g1 g8e7 f1e1 e8g8 d2d4 e7g6 b1d2 c8g4 h2h3 g4f5 d2f1 c6e7 b5d3 c7c6 a2a4 d8c7 h3h4 f5g4 f1e3 g4f3 d1f3 g6h4 d3h7 g8h7 f3h5 h7g8 h5h4 c7d7 c1d2 f7f5 h4h5 a8e8 a1d1 d6c7 d2b4 f8f7');
INSERT INTO evaluations VALUES(21372,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(21372,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(21372,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(21372,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(21372,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(21372,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(21372,36,'cp -18','e4d5 e6d5 c4b5 f8d6 e1g1 g8e7 d2d4 a7a6 b5d3 e8g8 c2c3 h7h6 b1d2 c8f5 d3f5 e7f5 f1e1 d8d7 d2f1 a8e8 f1g3 f5e7 c1d2 f7f5 g3f1 c6d8 h2h4 d8f7 a2a4 c7c6 h4h5 f5f4 f1h2 f7g5 f3g5');
INSERT INTO evaluations VALUES(21360,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(21360,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(21360,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(21360,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(21360,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(21360,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(21360,36,'cp -26','e6d5 c4b5 f8d6 e1g1 g8e7 d2d4 e8g8 f1e1 e7g6 b1d2 c8g4 c2c3 c6e7 b5d3 d8d7 h2h3 g4f5 d3f1 c7c6 b2b4 a8e8 a2a4 e7c8 b4b5 e8e1 f3e1 d6c7 e1f3 c8d6 f3e5 d7e6 b5c6 g6e5 d4e5 e6e5 c6b7 d6b7');
INSERT INTO evaluations VALUES(21581,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(21581,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(21581,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(21581,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(21581,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(21581,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(21581,36,'cp -26','c4b5 f8d6 e1g1 g8e7 f1e1 e8g8 d2d4 e7g6 b1d2 c8g4 h2h3 g4f5 d2f1 c6e7 b5d3 c7c6 f1g3 f5d3 d1d3 d8d7 c1d2 f8e8 a1d1 d6g3 f2g3 e7f5 f3e5 g6e5 d4e5 f5e7 c2c4 d7e6 c4d5 e6d5 d2c3 d5d3 d1d3 e7d5 d3d4 a7a5 a2a4 h7h6');
INSERT INTO evaluations VALUES(21524,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(21524,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(21524,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(21524,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(21524,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(21524,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(21524,36,'cp -9','f8d6 e1g1 g8e7 f1e1 e8g8 d2d4 e7g6 c1g5 c6e7 b5d3 c7c6 c2c4 d5c4 d3c4 h7h6 g5e7 d6e7 h2h3 c8f5 b1c3 d8c7 c4d3 f5d3 d1d3 f8d8 a1d1 d8d7 d4d5 a8d8 d3e4 c6d5 c3d5 c7a5 d5e7 d7e7 d1d8 a5d8 e4c2 e7e1 f3e1 g6e7 e1d3');
INSERT INTO evaluations VALUES(44004,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(44004,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(44004,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(44004,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(44004,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44004,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(44004,36,'cp 64','c5d4 c3d4 b8c6 g1f3 d8b6 e1g1 a7a5 f1e1 f8e7 h2h3 a5a4 a1b1 c6b4 d2f1 e8g8 c1g5 f7f6 e5f6 d7f6 f3e5 b4d3 d1d3 e7b4 e1e2 f6e4 g5e3 b4d6 b1c1 e4f6 d3d1 c8d7 e2c2 a4a3 e5d7 f6d7 b2a3 d7f6 c1b1 b6a6 e3g5 d6a3 g5f6 f8f6 f1e3 f6f8');
INSERT INTO evaluations VALUES(44018,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(44018,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(44018,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(44018,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(44018,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(44018,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44018,36,'cp 70','c3d4 b8c6 g1f3 f7f6 e5f6 d7f6 e1g1 f8d6 f1e1 e8g8 h2h3 c8d7 d2b3 a7a5 b3c5 d6c5 d4c5 d8c7 c1g5 c6b4 d3b1 c7c5 a2a3 b4c6 g5f4 c5e7 f4g3 a8c8 g3h2 h7h6 d1d2 d7e8 b1f5 f6e4 f5e4 d5e4 e1e4 e8g6 h2d6 e7d7');
INSERT INTO evaluations VALUES(44015,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(44015,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(44015,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(44015,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44015,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(44015,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(44015,36,'cp 60','b8c6 g1f3 f7f6 e5f6 d7f6 e1g1 f8d6 f1e1 e8g8 h2h3 c8d7 d2b3 a7a5 b3c5 d6c5 d4c5 d8c7 c1g5 c6b4 d3b1 c7c5 a2a3 b4c6 d1d2 a5a4 b1c2 c5a5 d2d1 f6e4 g5e3 h7h6 a1c1 a5b5 c2e4 d5e4 f3g5 h6g5 d1d7 f8f6 c1c5 b5d3 d7d3 e4d3 c5c3 a8f8');
INSERT INTO evaluations VALUES(12136,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(12136,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(12136,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(12136,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(12136,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(12136,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(12136,36,'cp 73','g1f3 f7f6 e5f6 d7f6 e1g1 f8d6 h2h3 e8g8 f1e1 d8c7 a2a3 c8d7 b2b4 e6e5 d1b3 g8h8 d4e5 c6e5 f3e5 d6e5 c1b2 e5b2 b3b2 c7f4 d3f1 a8e8 e1e8 f8e8 g2g3 f4f5 a1c1 a7a6 b2d4 d7b5 f1g2 f5d3 d4d3 b5d3 d2f3 d3c4 f3d4');
INSERT INTO evaluations VALUES(12135,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(12135,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(12135,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(12135,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(12135,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(12135,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(12135,36,'cp 52','d8b6 e1g1 a7a5 f1e1 f8e7 h2h3 a5a4 a2a3 c6d4 f3d4 b6d4 d2f3 d4b6 d3c2 d7c5 d1d4 c8d7 c1e3 a8c8 a1c1 b6b5 d4g4 g7g6 e3h6 b5b2 g4f4 h8f8 f3g5 f7f5 h6f8 e8f8 g5h7 f8g7 h7f6 e7f6 e5f6 b2f6 f4b4 f6e7 c1d1 d7c6 b4d4 g7f7 e1e3 c5e4 c2a4 c6a4 d4a4');
INSERT INTO evaluations VALUES(735,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(735,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(735,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(735,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(735,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(735,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(735,36,'cp 27','e1g1 a7a5 h2h3 f8e7 f1e1 a5a4 a2a3 c6d4 f3d4 b6d4 d2f3 d4b6 c1e3 d7c5 a1c1 c8d7 d3b1 b6b5 h3h4 a8c8 f3g5 b5b2 g5h7 b2a3 d1g4 g7g6 h7g5 c5d3 c1c8 d7c8 b1d3 a3d3 g4a4 b7b5 a4a8 e8g8 e1c1 d3a3 a8b8 d5d4 e3d2 a3a6 b8c7 a6b7 g5e4 b7c7');
INSERT INTO evaluations VALUES(734,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(734,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(734,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(734,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(734,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(734,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(734,36,'cp 32','a7a5 h2h3 f8e7 f1e1 a5a4 a1b1 c6b4 d2f1 d7f8 c1g5 e7g5 f3g5 b4d3 d1d3 f8g6 d3a3 c8d7 g5f3 b6d8 f1e3 d8e7 a3d3 e8g8 a2a3 a8c8 b1d1 d7c6 d3d2 c6b5 g2g3 f7f6 e3g2 e7f7 e5f6 f7f6 f3g5 h7h6 g5e6');
INSERT INTO evaluations VALUES(5730,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(5730,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(5730,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(5730,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5730,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(5730,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(5730,36,'cp 21','e1g1 g8f6 c3a4 c5b6 a4b6 a7b6 a2a4 e8g8 f1e1 d8e7 c2c3 c8e6 c4e6 e7e6 d3d4 e5d4 f3d4 c6d4 c3d4 f8e8 f2f3 b6b5 a4a5 f6d7 c1e3 c7c6 d1d3 e6c4 d3c4 b5c4 e3f4 d6d5 e4d5 c6d5 e1e8 a8e8 g1f2 d7f8');
INSERT INTO evaluations VALUES(5729,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(5729,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5729,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(5729,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5729,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(5729,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(5729,36,'cp 0','c6a5 c3a4 c5b6 a4b6 a7b6 c4d5 g8e7 c2c3 e8g8 a2a4 g8h8 d5a2 e7g6 a2d5');
INSERT INTO evaluations VALUES(57475,30,'cp 836',NULL);
INSERT INTO evaluations VALUES(57475,31,'cp 840',NULL);
INSERT INTO evaluations VALUES(57475,32,'cp 851',NULL);
INSERT INTO evaluations VALUES(57475,33,'cp 850',NULL);
INSERT INTO evaluations VALUES(57475,34,'cp 858',NULL);
INSERT INTO evaluations VALUES(57475,35,'cp 859',NULL);
INSERT INTO evaluations VALUES(57475,36,'cp 850','d1h5 g7g6 h5e5 d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 h7h6 f1c4 g8e7 g1e2 d7d6 c3b5 e8d8 b5c3 b7b6 a2a4 c8b7 h2h4 b8a6 c1e3 a6c5 e3c5 b6c5 f2f4 d8d7 c4b5 c7c6 b5d3 a7a5 e1c1 h6h5 c1d2 a8e8 b2b3 d7c7 g2g3 e8f8');
INSERT INTO evaluations VALUES(57466,30,'cp 858',NULL);
INSERT INTO evaluations VALUES(57466,31,'cp 853',NULL);
INSERT INTO evaluations VALUES(57466,32,'cp 859',NULL);
INSERT INTO evaluations VALUES(57466,33,'cp 846',NULL);
INSERT INTO evaluations VALUES(57466,34,'cp 855',NULL);
INSERT INTO evaluations VALUES(57466,35,'cp 851',NULL);
INSERT INTO evaluations VALUES(57466,36,'cp 855','g7g6 h5e5 d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 g7c3 b2c3 d7d6 h2h4 b8d7 g1e2 g8f6 f2f3 d7e5 c1g5 e8f7 e2f4 c8e6 f4e6 f7e6 e1d2 c7c6 g5f6 e6f6 f3f4 e5g4 f1d3 f6e6 c3c4 g4f6 h1b1 b7b6 a2a4 f6h5 f4f5 g6f5 e4f5 e6f6 a4a5');
INSERT INTO evaluations VALUES(57541,30,'cp 855',NULL);
INSERT INTO evaluations VALUES(57541,31,'cp 850',NULL);
INSERT INTO evaluations VALUES(57541,32,'cp 855',NULL);
INSERT INTO evaluations VALUES(57541,33,'cp 855',NULL);
INSERT INTO evaluations VALUES(57541,34,'cp 863',NULL);
INSERT INTO evaluations VALUES(57541,35,'cp 855',NULL);
INSERT INTO evaluations VALUES(57541,36,'cp 867','h5e5 d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 g7c3 b2c3 d7d6 h2h4 b8d7 g1e2 d7f6 f2f3 c8e6 c1e3 e6f7 e2f4 g8e7 e1f2 e7c6 c3c4 f6d7 f4d5 e8c8 g2g3 f7e6 h4h5 b7b6 f1e2 c6e5 f3f4 d8f8 f2g2 e6g4 e2g4 e5g4 e3d4 c7c5 d4c3 f8e8 h5g6 h7g6');
INSERT INTO evaluations VALUES(57538,30,'cp 859',NULL);
INSERT INTO evaluations VALUES(57538,31,'cp 857',NULL);
INSERT INTO evaluations VALUES(57538,32,'cp 862',NULL);
INSERT INTO evaluations VALUES(57538,33,'cp 850',NULL);
INSERT INTO evaluations VALUES(57538,34,'cp 859',NULL);
INSERT INTO evaluations VALUES(57538,35,'cp 865',NULL);
INSERT INTO evaluations VALUES(57538,36,'cp 867','d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 g7c3 b2c3 b7b6 g1f3 b8c6 f3d4 c8b7 f2f3 c6e7 h2h4 e8c8 f1d3 d8f8 a2a4 a7a5 c1g5 h7h6 g5e3 c7c5 d4b5 b7a6 e1f2 e7c6 b5d6 c8c7 d3a6 c7d6');
INSERT INTO evaluations VALUES(35242,30,'cp 858',NULL);
INSERT INTO evaluations VALUES(35242,31,'cp 862',NULL);
INSERT INTO evaluations VALUES(35242,32,'cp 867',NULL);
INSERT INTO evaluations VALUES(35242,33,'cp 869',NULL);
INSERT INTO evaluations VALUES(35242,34,'cp 867',NULL);
INSERT INTO evaluations VALUES(35242,35,'cp 859',NULL);
INSERT INTO evaluations VALUES(35242,36,'cp 867','e5h8 e7g7 h8g7 f8g7 b1c3 g7c3 b2c3 d7d6 h2h4 c8e6 h4h5 e6f7 h5g6 f7g6 f2f3 b8c6 g1h3 c6e5 h3f4 e8c8 c1e3 a7a6 c3c4 g8e7 e1f2 e7c6 a2a3 e5d7 f4g6 h7g6 f1e2 a6a5 a1d1 a5a4 c4c5 d7c5 e3c5 d6c5 d1d8 c8d8');
INSERT INTO evaluations VALUES(34106,30,'cp 858',NULL);
INSERT INTO evaluations VALUES(34106,31,'cp 869',NULL);
INSERT INTO evaluations VALUES(34106,32,'cp 865',NULL);
INSERT INTO evaluations VALUES(34106,33,'cp 855',NULL);
INSERT INTO evaluations VALUES(34106,34,'cp 854',NULL);
INSERT INTO evaluations VALUES(34106,35,'cp 855',NULL);
INSERT INTO evaluations VALUES(34106,36,'cp 861','e7g7 h8g7 f8g7 b1c3 g7c3 b2c3 d7d6 h2h4 b8d7 h4h5 g6h5 h1h5 d7f6 h5h4 h7h5 g1e2 c8e6 f2f3 e6f7 c1g5 f6h7 g5e3 g8f6 e2f4 h7f8 a1b1 e8c8 e3d4 f6d7 d4a7 f7a2 b1a1 a2f7 f4h5 f7h5 h4h5 f8e6 c3c4 d7c5 e1f2');
INSERT INTO evaluations VALUES(40923,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40923,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40923,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40923,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40923,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(40923,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(40923,36,'cp 8','h7h6 g5f6 e7f6 b1c3 e8g8 a1c1 c7c6 f1e2 b8d7 e1g1 d5c4 e2c4 c6c5 c3e4 c5d4 e4f6 d7f6 d1d4 d8d4 f3d4 f8d8 f2f3 c8d7 g1f2 a8c8 e3e4 g8f8 e4e5 f6g8 c4d3 g8e7 d3e4 e7c6 d4c6 d7c6 e4c6 c8c6 c1c6 b7c6');
INSERT INTO evaluations VALUES(40710,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(40710,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(40710,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(40710,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40710,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40710,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40710,36,'cp 12','g5h4 e8g8 b1c3 b7b6 c4d5 f6d5 h4e7 d8e7 c3d5 e6d5 a1c1 c8e6 f1d3 c7c5 b2b3 b8c6 e1g1 a8c8 f1e1 c8c7 d3b1 c5d4 f3d4 c6d4 c1c7 e7c7 d1d4 f8c8 e1d1 c7c3 d4a4 a7a5 h2h3 c8d8 a4f4');
INSERT INTO evaluations VALUES(40716,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40716,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(40716,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(40716,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40716,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40716,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(40716,36,'cp 12','e8g8 b1c3 b7b6 f1e2 d5c4 e2c4 b8d7 d1e2 c8b7 e1g1 f6d5 c3d5 e6d5 h4e7 d8e7 c4d3 c7c5 b2b3 a7a5 f1c1 f8c8 h2h3 a5a4 d3b5 b7c6 b3a4 c5c4 e2b2 c6b5 a4b5 a8a3 c1c3 a3c3 b2c3 e7d6 a1c1 d7f6');
INSERT INTO evaluations VALUES(36202,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36202,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36202,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(36202,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36202,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36202,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36202,36,'cp 16','b1c3 b7b6 f1e2 d5c4 e2c4 c8b7 e1g1 b8d7 d1e2 f6d5 c3d5 e6d5 h4e7 d8e7 c4d3 f8c8 f1c1 a7a5 b2b3 c7c5 a1b1 c8c7 h2h3 d7f6 f3e5 a8c8 e2b2 f6e4 d4c5 e4c5 d3e2 c5e4 c1c7 c8c7 e5f3 c7c3 b1e1 b7c6 e1d1 e7f6 d1c1');
INSERT INTO evaluations VALUES(54371,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(54371,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(54371,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(54371,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(54371,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(54371,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(54371,36,'cp 39','c1f4 f8d6 f4g3 d8b6 a1b1 h7h6 e2e3 g8f6 g1f3 d6g3 h2g3 b8c6 a2a3 e8g8 f1d3 b6c7 e1g1 f8d8 b1c1 c8d7 d1e2 c7d6 c1c2 d7e8 f3h2 e6e5 c3b5 d6e7 d4e5 c6e5 b5d4 e5d3 e2d3 a8c8 d4f5 e7f8 f1c1 c8c2 c1c2 f6e4 h2f3 e8d7');
INSERT INTO evaluations VALUES(54359,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54359,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(54359,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(54359,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(54359,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(54359,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54359,36,'cp 25','f8d6 f4g3 d8b6 a1b1 g8f6 e2e3 e8g8 g1f3 h7h6 f1d3 d6g3 h2g3 b8c6 a2a3 b6c7 e1g1 f8d8 b1c1 c8d7 c1c2 c7d6 d1e2 a7a6 f1c1 d7e8 e2e1 d6e7 c3e2 a8c8 b2b4 a6a5 f3e5 f6e4 d3e4 d5e4 e5c6 c8c6 c2c6');
INSERT INTO evaluations VALUES(17250,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(17250,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(17250,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17250,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(17250,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(17250,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(17250,36,'cp 26','d7d6 g1f3 f6h5 f4g3 h5g3 h2g3 e7e6 f1b5 c8d7 d4d5 c6e5 b5d7 d8d7 f3e5 d6e5 d5e6 d7e6 d1d3 e6d7 d3d7 e8d7 g3g4 d7e6 b1d2 f8e7 e1e2 h7h6 a2a4 a7a5 d2c4 b7b6 c2c3 h8d8 a1d1 d8d1 h1d1 a8d8 d1d8 e7d8 e2d3 d8e7 c4d2 f7f6 f2f3 e6d6 d3e4');
INSERT INTO evaluations VALUES(14478,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(14478,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14478,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(14478,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(14478,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(14478,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14478,36,'cp 25','g1f3 f6h5 f4g3 h5g3 h2g3 e7e6 f1b5 c8d7 c2c4 a7a6 b5c6 d7c6 b1c3 c6d7 e3e4 b7b5 b2b3 b5c4 b3c4 g7g6 e1g1 f8g7 a1c1 e8g8 d1e2 d7c6 f1d1 d8e7 d4d5 e6d5 c3d5 c6d5 c4d5 f8e8 d1e1 f7f5 c1c4 f5e4 c4e4 e7e4 e2e4 e8e4');
INSERT INTO evaluations VALUES(14479,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(14479,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(14479,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(14479,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(14479,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(14479,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(14479,36,'cp 21','f6h5 f4g3 h5g3 h2g3 e7e6 f1b5 c8d7 c2c4 a7a6 b5a4 g7g6 b1c3 f8g7 e1g1 e8g8 a1c1 f8e8 d1c2 b7b5 c4b5 c6b4 c2b3 a6b5 a4b5 b4a2 c3a2 d7b5 b3b5 a8a2 b2b4 a2b2 b5c4 d6d5 c4c7 b2b4 c7f4 b4c4 c1c4 d5c4 f1c1');
INSERT INTO evaluations VALUES(54205,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(54205,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(54205,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(54205,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(54205,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(54205,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(54205,36,'cp 73','e2e4 c5d4 c1g5 f8e7 g5e7 g8e7 d1d4 e8g8 e1c1 b8c6 d4c5 b7b6 c5e3 d8c7 e4d5 e7d5 c3d5 e6d5 f1b5 c8g4 h1e1 a8c8 c2c3 h7h6 h2h3 g4d7 b5c6 d7c6 e3e5 c7e5 e1e5 f8e8 e5e8 c8e8 b2b3 g8f8 a2a4 f8e7 f3d4 c6d7 f2f3 e7d6 c1d2');
INSERT INTO evaluations VALUES(48796,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(48796,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48796,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(48796,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(48796,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(48796,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(48796,36,'cp 21','c4d5 f6d5 g1f3 b8c6 e1g1 d5b6 b1c3 f8e7 b2b3 e8g8 c1b2 c8f5 a1c1 f8e8 d2d3 f7f6 d1c2 a7a5 f1d1 d8d7 c3e4 a5a4 d3d4 c6b4 c2b1 a4b3 a2b3 e5d4 b2d4 b6d5 f3h4 f5h3 d4c5 h3g2 h4g2 f6f5 c5b4 e7b4');
INSERT INTO evaluations VALUES(48616,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(48616,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48616,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(48616,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(48616,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(48616,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(48616,36,'cp 23','f6d5 g1f3 b8c6 e1g1 d5b6 d2d3 f8e7 b1c3 e8g8 c1e3 c8e6 a1c1 f8e8 f1e1 f7f5 a2a3 g8h8 d1c2 e7f6 c3a4 b6a4 c2a4 c6d4 f3d4 e5d4 e3f4 c7c6 h2h4 d8d7 c1c5 h7h6 a4c2 d7f7 b2b4 e6b3 c2d2');
INSERT INTO evaluations VALUES(49026,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(49026,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(49026,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(49026,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(49026,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(49026,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(49026,36,'cp 33','b1c3 d5b6 g1f3 b8c6 a2a3 a7a5 e1g1 f8e7 d2d3 c8e6 c1e3 e8g8 c3a4 e5e4 f3e1 b6d5 g2e4 d5e3 f2e3 e7g5 e1g2 e6h3 e4c6 b7c6 d1c1 f8e8 f1f3 a8b8 a4c5 g5h6 g2f4 h3g4 f3f2 d8e7 e3e4 g7g5 f4g2 h6g7 e2e3 b8b2 f2b2 g7b2 c1b2 e7c5 a1c1 c5d6');
INSERT INTO evaluations VALUES(27168,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(27168,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(27168,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(27168,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(27168,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(27168,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(27168,36,'cp 62','b7c6 b5c4 f8e7 b1c3 g8f6 e1g1 e8g8 c4d3 f8e8 h2h3 c6c5 d3c4 d7c6 d1f3 e8f8 c1f4 f6d7 b2b3 d7b6 c4b5 c6b7 a1e1 f7f5 f3e3 f5e4 c3e4 b6d5 b5c4 g8h8 c4d5 b7d5 e4c3 d5g8 e3e7 f8f4 e1e3 d8e7');
INSERT INTO evaluations VALUES(25805,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(25805,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(25805,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(25805,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(25805,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(25805,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(25805,36,'cp 53','b5c4 g8f6 e1g1 f8e7 b1c3 e8g8 c4d3 c6c5 c1f4 d7c6 b2b3 f8e8 a1b1 f6d7 d1d2 e7f6 f2f3 f6d4 f4e3 d4e3 d2e3 d7e5 b1d1 f7f5 d3b5 f5e4 b5c6 e5c6 c3e4 g8h8 e3f2 d8e7 f1e1 e7f7 e4g5 f7f5 f2d2 c6d4 c2c3');
INSERT INTO evaluations VALUES(25808,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25808,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(25808,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(25808,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(25808,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(25808,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(25808,36,'cp 60','d8h4 e1g1 g8f6 b1d2 f8e7 c4d3 e8g8 h2h3 h4h5 d1h5 f6h5 d2c4 g7g5 f1e1 h5f4 d3f1 a8e8 a1b1 e7f6 c1d2 f6g7 d2a5 d6d5 c4d2 f4e6 d2b3 e8d8 b1d1 g7b2 e4d5 c6d5 d1d5 d7c6 d5f5 h7h6 f1c4 b2g7 c4e6 f7e6 f5f8 d8f8');
INSERT INTO evaluations VALUES(42485,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(42485,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(42485,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42485,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(42485,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(42485,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(42485,36,'cp 16','b4e7 d2f4 e8g8 b1c3 d7d5 e2e3 b7b6 c4d5 f6d5 c3d5 d8d5 a2a3 c7c5 d4c5 d5c5 a1c1 c5a5 d1d2 a5d2 f3d2 b8a6 f1e2 c8b7 e1g1 a8c8 b2b4 f8d8 d2f3 a6b8 f4c7 d8d7 c7g3 b8c6 h2h4 d7d8 f1d1 h7h6 e2d3 e7f6 b4b5');
INSERT INTO evaluations VALUES(42987,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(42987,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(42987,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(42987,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(42987,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(42987,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(42987,36,'cp 24','d2f4 e8g8 b1c3 d7d5 e2e3 b7b6 a1c1 c8b7 h2h3 c7c5 d4c5 b6c5 f1e2 d8b6 e1g1 f8d8 c3a4 b6a5 d1b3 b7c6 a4c3 b8d7 f1d1 a5b6 c4d5 f6d5 c3d5 c6d5 b3b5 b6b5 e2b5 a7a6 b5e2 a6a5 f4c7 d8c8 e3e4 c8c7 e4d5 e6d5 d1d5 d7b6');
INSERT INTO evaluations VALUES(53249,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(53249,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(53249,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(53249,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53249,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(53249,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(53249,36,'cp -17','g8f6 g1f3 b8c6 e2e3 d8b6 b2b3 c5d4 e3d4 c8f5 f1d3 f5d3 d1d3 e7e6 e1g1 f8e7 c3c4 d5c4 b3c4 a8d8 f1d1 d8d7 b1c3 e8g8 a1b1 b6a6 h2h3 f8d8 f4e3 b7b6 c3b5 a6c8 d3c2 h7h6 a2a4 a7a6 b5c3 c6a5 f3e5 a5c4 e5d7 d8d7');
INSERT INTO evaluations VALUES(45357,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(45357,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(45357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45357,36,'cp 0','g1f3 b8c6 a2a3 c5c4 b1d2 c8f5 f3h4 f5d7 h4f3');
INSERT INTO evaluations VALUES(45356,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(45356,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(45356,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(45356,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(45356,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(45356,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(45356,36,'cp -19','b8c6 a2a3 e7e6 e2e3 f8d6 d4c5 d6c5 b1d2 e8g8 b2b4 c5d6 f4d6 d8d6 c3c4 f8d8 c4d5 e6d5 b4b5 c6e5 f1e2 c8g4 e1g1 e5f3 e2f3 g4f3 d2f3 a8c8 a3a4 c8c4 f3d2 c4c3 a1c1 d8c8 c1c3 c8c3 d2f3 f6e4 d1d4 c3c5 f1e1 g7g6 g2g3');
INSERT INTO evaluations VALUES(13093,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13093,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(13093,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(13093,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(13093,36,'cp 0','a2a3 f6h5 f4e5 f7f6 e5g3 g7g5 d4c5 e7e5 b2b4 g5g4 e2e4 g4f3 e4d5 h5f4 g3f4 e5f4 d5c6 d8e7 e1d2 c8g4 c6b7 a8d8 d2c2 f3g2 f1b5 e8f7 b5c4 f7e8 c4b5');
INSERT INTO evaluations VALUES(4878,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(4878,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(4878,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(4878,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(4878,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(4878,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(4878,36,'cp 16','g8f6 c2c3 b7b5 a4b3 d7d6 d2d4 c5b6 a2a4 c8b7 c1e3 e5d4 c3d4 f6e4 d4d5 c6e7 e3b6 c7b6 a4b5 a6b5 a1a8 b7a8 f3d4 e4c5 b1c3 b5b4 b3a4 e8f8 c3a2 c5a4 d1a4 e7d5 a2b4 g7g6 b4c6 a8c6 d4c6 d8f6 f1d1 f6f4 a4f4 d5f4 g2g3');
INSERT INTO evaluations VALUES(14962,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(14962,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(14962,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(14962,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(14962,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(14962,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14962,36,'cp 64','a7a6 b5c3 f6h5 c3d5 c8e6 e2e4 h5f4 d5f4 d8e7 f4e6 e7e6 d1d5 e6d5 e4d5 c6e5 a1d1 e8c8 f1e2 e5d7 g1h3 f8c5 e1g1 h8e8 e2h5 g7g6 h5f3 d7e5 f3e2 e5d7 f1e1 d7f6 h3g5 d8d7 e2f3 e8e1 d1e1 c8d8 g2g3 f6d5 f3d5 d7d5 g5f7 d8c8 f7e5 d5d2');
INSERT INTO evaluations VALUES(11385,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(11385,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(11385,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(11385,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(11385,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(11385,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(11385,36,'cp 89','b5c3 f6h5 c3d5 c8e6 e2e4 h5f4 d5f4 d8e7 c2c3 c6e5 f1e2 a8d8 d1c2 g7g6 f4e6 e7e6 g1h3 f8h6 f2f4 e5c4 e2c4 e6c4 c2e2 c4e6 e1g1 e8g8 a2a3 d8d6 e2e1 e6c4 e4e5 d6b6 f1f2 f7f6 e1e2 c4c5 e5f6 f8f6 g2g3 f6e6 e2g4 e6e3 a1d1 b6e6 h3g5 h6g5 f4g5 e6e8 g4d4 c5d4');
INSERT INTO evaluations VALUES(26602,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(26602,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(26602,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(26602,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(26602,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(26602,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(26602,36,'cp 44','e7e6 c2c4 g8f6 b1c3 f8b4 a1c1 e8g8 f1e2 d5c4 e2c4 c6a5 c4d3 c7c5 d4c5 b4c5 e1g1 g4f3 g2f3 c5e7 f4g3 a8c8 d1e2 a5c6 f1d1 d8a5 a2a3 h7h6 c3e4 f8d8 g1g2 f6d5 h2h4 g8h8 e4c3 e7d6 c3d5 d6g3 f2g3 a5d5');
INSERT INTO evaluations VALUES(26837,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(26837,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(26837,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(26837,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(26837,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(26837,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(26837,36,'cp 35','c2c4 g8f6 a2a3 f8d6 f4g5 h7h6 g5h4 d6e7 b1d2 f6e4 h4e7 c6e7 h2h3 g4h5 f1e2 e4d2 f3d2 h5e2 d1e2 e8g8 e1g1 f8e8 d2f3 c7c6 b2b4 d5c4 e2c4 a7a5 f1b1 b7b6 b4a5 a8a5 b1b3 c6c5 a1b1 c5d4 f3d4 d8d5');
INSERT INTO evaluations VALUES(56069,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(56069,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(56069,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(56069,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(56069,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(56069,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(56069,36,'cp 44','b8d7 d2d4 g8f6 e4f6 d7f6 c2c3 a7a6 c1f4 c7c5 f1e2 c5d4 d1d4 d8d4 f3d4 c8d7 e2f3 e8c8 e1c1 f8c5 f4e5 h7h5 h1e1 c5b6 h2h3 f6e8 d4b3 e8f6 d1d2 d7b5 e5d4 b6d4 b3d4 b5c4 b2b3 c4d5 c3c4 d5f3 d4f3 d8d2 c1d2 f6d7 d2c3 a6a5');
INSERT INTO evaluations VALUES(22621,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(22621,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(22621,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(22621,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(22621,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(22621,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22621,36,'cp 91','d2d4 d7d5 e4e5 a7a6 g1f3 f7f6 c3e2 c8d7 c2c3 d8e7 g2g3 e7f7 f1g2 f6e5 d4e5 h7h6 e2f4 g7g5 f4d3 g5g4 f3d4 h6h5 h2h4 c6d4 c3d4 g8e7 c1g5 d7b5 g5f6 f8g7 f6e7 f7e7 e1g1 c7c5 d4c5 e8g8 b2b4 e7f7 f2f4 g4f3 f1f3');
INSERT INTO evaluations VALUES(22613,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(22613,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(22613,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(22613,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(22613,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(22613,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(22613,36,'cp 75','d7d5 e4e5 f7f6 f1b5 c8d7 g1f3 a7a6 b5c6 d7c6 h2h4 g8h6 c1h6 g7h6 d1e2 f8g7 e1c1 e8g8 h1e1 c6e8 e5f6 f8f6 f3e5 h6h5 g2g3 d8d6 f2f4 c7c5 d4c5 d6c5 e2e3 c5e3 e1e3 a8c8 c3e2 e8g6 e5g6 f6g6 e2d4 g7d4 d1d4 b7b5 a2a4 g8h8 a4b5 a6b5');
INSERT INTO evaluations VALUES(49508,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(49508,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(49508,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(49508,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(49508,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(49508,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(49508,36,'cp -83','f6h5 d1e2 f8e7 d2d4 e8g8 b1c3 d7d5 g2g3 f7f6 e5e6 f6f5 h1g1 b8c6 c1f4 e7b4 e2e3 f8e8 e1c1 c8e6 f1b5 e6d7 e3d3 c6e7 b5d7 d8d7 a2a3 b4d6 f4e5 e8f8 c3e2 e7g6 e2f4 g6f4 g3f4 c7c6 d3b3 g8h8 c1b1 h5f6 e5f6');
INSERT INTO evaluations VALUES(49836,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(49836,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(49836,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(49836,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(49836,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(49836,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(49836,36,'cp -53','d2d4 d7d5 c2c4 g7g5 g2g4 h5g7 b1c3 d5c4 f1c4 c8g4 h2h3 g4h5 h3h4 b8c6 c3d5 f8e7 h4g5 h5g4 c1f4 d8d7 d5e3 e8c8 d4d5 e7c5 d5c6 d7c6 c4d5 d8d5 e3d5 h8d8 f3d4 c6d5 d1g4 g7e6 e1c1 c5d4 f4d2 d5e5 d2c3 d4c3 b2c3 e5c3 c1b1 d8d2 d1d2 c3d2 g4f5 d2g5 f5f7 g5b5 b1c1 b5c4 c1b2 c4d4 b2c2');
INSERT INTO evaluations VALUES(49835,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(49835,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(49835,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(49835,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(49835,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(49835,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(49835,36,'cp -51','d7d5 c2c4 g7g5 g2g4 h5g7 b1c3 d5c4 f1c4 c8g4 h2h3 g4h5 h3h4 b8c6 c3d5 f8e7 h4g5 h5g4 c1f4 d8d7 d5e3 e8c8 d4d5 e7c5 d5c6 d7c6 c4d5 d8d5 e3d5 h8d8 f3d4 c6d5 d1g4 g7e6 e1c1 c5d4 f4d2 d5e5 d2c3 d4c3 b2c3 e5c3 c1b1 d8d2 d1d2 c3d2 g4f5 d2g5 f5f7 g5b5 b1c1 b5c4 c1b2 c4d4 b2c2 d4e4 c2b2 e4h1 f7e6 c8d8 e6f6 d8d7 f6f7 d7d6 f7f6 d6c5 f6c3 c5d5 c3d3 d5e6 d3e3 e6d6');
INSERT INTO evaluations VALUES(6970,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6970,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(6970,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(6970,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(6970,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6970,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(6970,36,'cp 32','g8f6 d2d4 f6e4 d4e5 d7d6 d1e2 e4c5 e5d6 c7d6 b1a3 d6d5 c1e3 e8g8 a1d1 c8e6 e1g1 a7a6 f1e1 c6a5 e2c2 a8c8 c4e2 a5c6 a3c4 h7h6 c4d2 f8e8 d2f1 e7d6 f3d4 c6d4 e3d4 e6d7 f1e3 c5e6 e3d5 e6d4 d1d4');
INSERT INTO evaluations VALUES(10893,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(10893,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(10893,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(10893,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(10893,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(10893,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(10893,36,'cp 16','d2d4 f6e4 d4e5 d7d6 d1e2 d6d5 e1g1 e4f2 c4b3 f2e4 f1d1 e7c5 c1e3 c5e3 e2e3 c6e7 b3d5 e7d5 e3e4 c7c6 b1a3 e8g8 a3c4 f7f5 e4h4 d8h4 f3h4 c8e6 c4d6 d5e3 d1d2 a8d8 a1e1 f5f4 b2b3 d8d7 c3c4 d7e7 h2h3 g7g5 h4f3 g5g4 h3g4 e6g4 g1f2 e7g7 e1g1 g4h5 c4c5 g7e7 g1c1 h5f3 g2f3 e7e5 d6b7');
INSERT INTO evaluations VALUES(10915,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(10915,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10915,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(10915,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(10915,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(10915,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(10915,36,'cp 12','f6e4 d4e5 d7d6 d1e2 d6d5 e1g1 e4f2 c4b3 f2e4 f1d1 e7c5 c1e3 c5e3 e2e3 c6e7 b3d5 e7d5 e3e4 c7c6 g1h1 d8e7 c3c4 d5c7 b1c3 e8g8 h2h3 f7f6 d1e1 f6e5 e4e5 e7e5 f3e5 f8e8 a1d1 a7a5 b2b3 c7a6 g2g4 a6c5 h1g2 c8e6 e5d3 c5d3 d1d3 e6d5 g2f1 d5e6 c3e4 e6f7');
INSERT INTO evaluations VALUES(48182,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48182,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48182,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48182,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48182,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48182,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(48182,36,'cp 0','e6d5 c1f4 c7c6 b1c3 c8f5 e2e3 b8d7 a1c1 f6h5 f4g5 h5f6');
INSERT INTO evaluations VALUES(39142,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(39142,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(39142,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(39142,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(39142,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(39142,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(39142,36,'cp 112','d2d4 c5b6 f1d3 d7d5 b1d2 c7c5 d4c5 b6c5 e1g1 g8e7 a2a3 a7a5 c2c4 e8g8 e4d5 e6d5 d1c2 h7h6 f1d1 d5c4 d2c4 d8c7 c4e5 c5d6 d3h7 g8h8 c1f4 d6e5 f3e5 c7c2 h7c2 c8e6 c2e4 b8c6 e5d7 f8d8 d7b6 d8d1 a1d1 a8e8 h2h3 h8g8 d1e1 e7g6 f4g3');
INSERT INTO evaluations VALUES(39139,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(39139,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(39139,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(39139,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(39139,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(39139,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(39139,36,'cp 112','c5b6 b1d2 d7d5 f1d3 c7c5 d4c5 b6c5 e1g1 g8e7 c2c3 e8g8 f1e1 c5b6 b2b4 e7g6 c1a3 b8d7 e4d5 e6d5 c3c4 d8f6 c4d5 d7e5 d3g6 e5f3 d1f3 f7g6 f3f6 f8f6 d2e4 f6f5 a1d1 c8d7 e4c5 a8e8 e1e8 d7e8 h2h3 h7h6 a3b2 e8f7');
INSERT INTO evaluations VALUES(39084,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(39084,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(39084,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(39084,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(39084,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(39084,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(39084,36,'cp 121','a2a4 a7a6 b1d2 d7d5 b2b4 e6e5 e4d5 e5d4 d2c4 d8d5 c4b6 c7b6 c1a3 g8e7 d1d4 d5e6 f1e2 e8g8 c2c4 b8c6 d4b2 e7g6 e1g1 f8e8 f1e1 e6f6 b2f6 g7f6 b4b5 a6b5 c4b5 c6e5 f3e5 f6e5 a3b2 c8f5 h2h3 f5c2 e2f3 a8a4 a1a4');
INSERT INTO evaluations VALUES(50921,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(50921,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(50921,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(50921,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(50921,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(50921,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(50921,36,'cp 113','d2d4 e5d4 f3d4 f8e7 c1f4 g8f6 d1f3 c8g4 f3g3 c7c5 d4e2 g4e6 e1c1 b8c6 f4d6 f6h5 g3f3 e7d6 f3h5 c6d4 c1b1 g7g6 h5h6 d8f6 h6e3 d4e2 f1e2 d6e5 h1f1 e5d4 e3g3 e8c8 f2f4 d4c3 d1d8 h8d8 b2c3 f6e7 g3e3 e7c7 f4f5 g6f5 e4f5 e6d7 e2c4 c7b6 c4b3 d7c6');
INSERT INTO evaluations VALUES(50917,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(50917,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(50917,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(50917,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(50917,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(50917,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(50917,36,'cp 90','e5d4 f3d4 g8f6 c1f4 f8e7 d1f3 c8g4 f3g3 d8d7 g3e3 b8c6 f2f3 g4e6 e1c1 f6h5 d4c6 d7c6 f4g5 c6b6 c3d5 b6e3 g5e3 e6d5 d1d5 h5f6 d5d1 h7h5 a2a4 h5h4 f1c4 f6d7 c2c3 d7e5 c4e2 g7g5 e3d4 e8c8 a4a5');
INSERT INTO evaluations VALUES(50930,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(50930,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(50930,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(50930,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(50930,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(50930,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(50930,36,'cp 102','f3d4 f8e7 c1f4 g8f6 d1f3 c8g4 f3g3 d8d7 g3e3 b8c6 f2f3 g4e6 e1c1 e8g8 g2g4 c6d4 e3d4 a8d8 d4f2 d6d5 g4g5 f6h5 f4e3 d5d4 d1d4 d7c6 d4d3 f7f5 f2h4 g7g6 f1e2 e6c4 d3d8 f8d8 h1d1 c4e2 d1d8 e7d8 c3e2 c6b5 e2f4 h5f4 h4f4 f5e4 f3e4');
INSERT INTO evaluations VALUES(50927,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(50927,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(50927,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(50927,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(50927,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(50927,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(50927,36,'cp 86','f8e7 c1f4 g8f6 d1f3 c8g4 f3g3 d8d7 g3e3 b8c6 f2f3 g4e6 e1c1 c6d4 e3d4 b7b5 c3d5 c7c5 d5f6 g7f6 d4e3 e8c8 f1e2 h8e8 d1d2 d6d5 e4d5 e6d5 e3c3 d7c6 a2a3 d5e6 h1d1 d8d2 d1d2 e8d8 d2d8 e7d8 h2h4 f6f5 c3h8 c6d5 h4h5 c8b7 h8g7 d5d4 g7d4 c5d4 h5h6 d8b6 b2b3 b7a8');
INSERT INTO evaluations VALUES(52467,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52467,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52467,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(52467,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(52467,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52467,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(52467,36,'cp 0','g8f6 g1f3 e7e6 c2c4 f8e7 e2e3 e8g8 d2d4 c5d4 e3d4 b7b6 f1d3 d7d5 e1g1 b8c6 f1e1 c8b7 b1c3 c6a5 f3e5 a5c6');
INSERT INTO evaluations VALUES(44583,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44583,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44583,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44583,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44583,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44583,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44583,36,'cp 9','g1f3 e7e6 e2e3 f8e7 f1e2 e8g8 e1g1 b8c6 d2d4 c5d4 f3d4 d7d5 d4f3 f6e4 d1c1 e7f6 c2c4 d5c4 e2c4 f6b2 c1b2 d8f6 b2f6 e4f6 f1d1 f8d8 d1d8 c6d8 f3e5 c8d7 e5d7 f6d7 b1c3 a8c8 a1c1 g8f8');
INSERT INTO evaluations VALUES(19563,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(19563,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(19563,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(19563,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(19563,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(19563,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(19563,36,'cp -79','g8f6 c3d3 e7e6 g1f3 d8b6 b1c3 f8c5 d3d2 f6g4 c3d1 e8g8 f1d3 f7f5 e4f5 d7d5 h2h3 g4e5 f3e5 c6e5 e1g1 e5d3 d2d3 f8f5 a2a4 f5f7 d1e3 c8d7 e3g4 b6c7 c1e3 c5d6 f2f4 a8f8 g4e5 d6e5 f4e5 f7f1 a1f1 f8f1 d3f1 d7a4 e3a7 a4c2 a7d4 h7h6 f1f4 c2g6');
INSERT INTO evaluations VALUES(12894,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(12894,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(12894,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(12894,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(12894,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(12894,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(12894,36,'cp -75','c3d3 e7e6 g1f3 d8b6 b1c3 f8c5 d3d2 f6g4 c3d1 d7d5 e4d5 e6d5 f1d3 e8g8 e1g1 d5d4 b2b3 g4e5 f3e5 c6e5 d1b2 e5d3 b2d3 c5d6 c2c4 d4c3 d2c3 f8e8 c1b2 d6f8 d3f4 b6h6 f4d5 c8f5 c3g3 e8e2 g3f3 e2b2 f3f5 a8d8 f1d1 h6g6 f5g6 h7g6 h2h3 b7b6');
INSERT INTO evaluations VALUES(12897,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(12897,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(12897,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(12897,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(12897,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(12897,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(12897,36,'cp -62','e7e6 g1f3 d8b6 b1c3 f8c5 d3d2 f6g4 c3d1 d7d5 e4d5 e6d5 f1d3 e8g8 e1g1 d5d4 b2b3 c6e5 f3e5 g4e5 d1b2 e5d3 b2d3 c5d6 c1b2 f8d8 f1e1 c8f5 d2g5 f5g6 e1e2 a8c8 a1d1 d6c7 g5h4 b6b5 a2a4 b5h5 h4h5 g6h5 f2f3 f7f6 b2a3 c7b8 g1f2 b8h2 g2g3 h5g6 a3e7 d8e8');
INSERT INTO evaluations VALUES(56245,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(56245,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(56245,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(56245,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(56245,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56245,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56245,36,'cp -20','e4d5 g8f6 f1c4 f6d5 b1c3 d5b6 c4b3 b8c6 g1e2 f8d6 d2d4 e8g8 e1g1 c8f5 c3b5 f8e8 e2c3 a7a6 b5d6 d8d6 d4d5 c6a5 b3a4 e8d8 f1e1 b6d5 d1f3 d5c3 f3c3 a5c6 a4c6 b7c6 c3f3 d8e8 e1e8 a8e8 c1e3 f5e4 f3g3 e4d5');
INSERT INTO evaluations VALUES(56150,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(56150,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(56150,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(56150,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(56150,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(56150,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(56150,36,'cp -27','g8f6 f1b5 b8d7 g1f3 a7a6 b5d7 c8d7 e1g1 f6d5 f1e1 f8e7 b1c3 d7e6 f3d4 d5f4 d4e6 f4e6 d2d4 d8d4 d1d4 e6d4 c3d5 d4e6 e1e6 f7e6 d5c7 e8f7 c7a8 h8a8 c1e3 a8c8 a1c1 c8c1 e3c1 e6e5 g1f1 f7e6 f1e2 e6d5 b2b3 g7g6 h2h3 e5e4 c1e3 b7b5 e3a7 e7c5 a7c5 d5c5 e2e3 c5d5 a2a4 b5a4 b3a4');
INSERT INTO evaluations VALUES(48583,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(48583,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(48583,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(48583,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48583,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(48583,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48583,36,'cp -13','f1b5 b8d7 g1f3 a7a6 b5d3 f6d5 e1g1 f8e7 f1e1 d7f6 b1c3 e8g8 d3c4 c8e6 f3g5 d5c3 g5e6 c3d1 e6d8 e7d8 e1d1 d8e7 d2d4 f8d8 a2a4 c7c6 d1d3 d8d7 c1e3 a6a5 e3d2 g8f8 g2g3 f6d5 d2c3 e7b4 a1e1 a8d8 d3f3 d7e7 e1e7 f8e7 g1g2');
INSERT INTO evaluations VALUES(10058,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10058,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10058,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(10058,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(10058,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(10058,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10058,36,'cp 0','h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 a7a6 d3d4 e5d4 c3d4 c5a7 f1c4 d7d6 h2h3 e8g8 e1g1 b7b5 c4b3 c8b7 a2a4 c6e7 a4b5 a6b5 d1d3 e7d5 b3d5 b7d5 e4d5 d8d7 a1a5 a7b6 d3b5 d7b5 a5b5 f8e8 f1c1 a8a4 h3h4 e8e7 g2g3 g7g6 g1g2 h6h5 c1c6 b6d4 f3d4 a4d4 b5b8 g8g7');
INSERT INTO evaluations VALUES(9736,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9736,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9736,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9736,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9736,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9736,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9736,36,'cp 0','g5f6 d8f6 c3d5 f6d8 c2c3 a7a6 d3d4 e5d4 c3d4 c5a7 f1c4 d7d6 h2h3 b7b5 c4b3 c8b7 a2a4 c6e7 a4b5 a6b5 d1d3 e8g8 d3b5 a8b8 b5d3 e7d5 b3d5 b7d5 e4d5 f8e8 e1d1 b8b2 a1a7 d8b8 a7a1 b2b3 d3c2 b3b2');
INSERT INTO evaluations VALUES(9724,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9724,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9724,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9724,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9724,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9724,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9724,36,'cp 0','d8f6 c3d5 f6d8 c2c3 a7a6 d3d4 e5d4 c3d4 c5a7 f1c4 d7d6 h2h3 b7b5 c4b3 c8b7 a2a4 c6e7 a4b5 a6b5 d1d3 e8g8 d3b5 a8b8 b5d3 e7d5 b3d5 b7d5 e4d5 f8e8 e1d1 b8b2 a1a7 d8b8 a7a1 b2b3 d3c2 b3b2');
INSERT INTO evaluations VALUES(549,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(549,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(549,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(549,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(549,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(549,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(549,36,'cp 3','c3d5 f6d8 c2c3 d7d6 d3d4 c5b6 f1b5 e8g8 e1g1 e5d4 d5b6 a7b6 f3d4 c6d4 c3d4 c8d7 b5d7 d8d7 a2a3 a8a4 a1c1 b6b5 f1e1 f8a8 d1d2 d6d5 e4e5 a4c4 c1c3 d7e6 c3g3 a8a4 b2b4 e6c6 d2d3 c4c1 d3d2');
INSERT INTO evaluations VALUES(551,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(551,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(551,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(551,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(551,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(551,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(551,36,'cp 0','f6d8 c2c3 d7d6 d3d4 c5b6 d4e5 c6e5 a2a4 e5f3 d1f3 e8g8 f1d3 c7c6 d5b6 d8b6 e1c1 c8e6 d3c2 f8d8 d1d4 d6d5 e4d5 d8d5 d4d5 e6d5 f3d3 g7g6 d3g3 a8e8 c2g6 f7g6 g3g6 g8f8 g6h6 f8g8 h6g6 g8f8');
INSERT INTO evaluations VALUES(9889,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9889,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9889,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9889,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(9889,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9889,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9889,36,'cp 0','c2c3 a7a6 d3d4 e5d4 c3d4 c5a7 f1c4 d7d6 h2h3 c6e7 e1g1 e7g6 d1d2 e8g8 d5f4 d8f6 f4d5');
INSERT INTO evaluations VALUES(15458,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(15458,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(15458,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(15458,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(15458,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(15458,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(15458,36,'cp 51','c6e5 c1f4 e5g6 f4g3 f8e7 d1d2 e8g8 e1c1 h7h6 f1c4 a7a6 c4b3 b7b5 e4e5 d6e5 d2e2 c8d7 f3e5 g6e5 e2e5 c7c6 c3e4 a6a5 a2a3 a5a4 b3a2 f6e4 e5e4 e7g5 c1b1 d8c8 d1d6 c6c5 a2d5 a8b8 d6d7 c8d7 g3b8 f8b8 h1d1 g5f6');
INSERT INTO evaluations VALUES(15746,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(15746,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(15746,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(15746,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(15746,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(15746,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(15746,36,'cp 47','f3d4 c7c6 f2f4 e5g6 g2g3 f8e7 d1d3 e8g8 f1g2 f8e8 e1g1 g6f8 h2h3 f8e6 d4e6 c8e6 a2a4 d6d5 e4e5 f6d7 f4f5 d7e5 f5e6 e5d3 e6f7 g8h8 f7e8q d8e8 c2d3 e8g6 g1h2 g6d3 c1f4 e7c5 a1d1 d3g6 c3e2 g6c2 d1c1 c2e2 c1c5 e2b2 h2h1');
INSERT INTO evaluations VALUES(51929,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(51929,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(51929,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(51929,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(51929,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(51929,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(51929,36,'cp -48','g8f6 d2d3 d7d5 e4d5 e6d5 c4a2 h7h6 b1c3 b8c6 g1e2 c8e6 e2f4 d8d7 e1g1 e8c8 f4e6 f7e6 c3e2 f8d6 b2b4 c8b8 b4c5 d6c5 a1b1 a7a6 c2c4 b8a8 h2h3 g7g5 c1e3 d5d4 e3d2 c5a3 e2g3 g5g4 h3h4 a3d6 d1a4 d7c7 f1e1 h8f8 e1e6 d6g3 f2g3');
INSERT INTO evaluations VALUES(44189,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(44189,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(44189,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(44189,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(44189,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(44189,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(44189,36,'cp -35','d2d3 b8c6 b1d2 d7d5 c4b5 d8b6 b5a4 b6c7 g1f3 d5e4 d3e4 a7a6 a4c6 c7c6 e1g1 b7b5 f1e1 c8b7 b2b3 f8e7 c1b2 e8g8 d1e2 f6d7 a3a4 h7h6 h2h4 c5c4 a4b5 a6b5 b3c4 b5b4 f3e5 d7e5');
INSERT INTO evaluations VALUES(44188,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(44188,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(44188,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(44188,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(44188,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(44188,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(44188,36,'cp -45','d7d5 e4d5 f6d5 g1f3 f8e7 e1g1 e8g8 d3d4 c5d4 f3d4 e7f6 d4f3 b7b6 c4d5 e6d5 b1c3 f6c3 b2c3 b8c6 f1e1 h7h6 c1f4 c8f5 f3e5 a8c8 e5c6 c8c6 d1d2 d8d7 e1e3 c6g6 a1e1 d7a4 f4e5 f5c2 e5d4 f8d8');
INSERT INTO evaluations VALUES(1918,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(1918,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(1918,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(1918,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(1918,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(1918,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(1918,36,'cp 82','b1c3 f8b4 f1e2 h4e4 d4b5 b4c3 b2c3 e8d8 e1g1 g8f6 c1g5 d7d6 b5d4 e4e5 g5h4 c8d7 e2f3 d8c8 f1e1 e5c5 f3c6 b7c6 h4f6 g7f6 d1d3 a8b8 e1e7 b8b6 a1e1 c5g5 c3c4 h8g8 g2g3 g5h5 d3c3 a7a6 c4c5 h5c5 c3c5 d6c5 d4b3 d7e6 b3c5 g8d8');
INSERT INTO evaluations VALUES(1916,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(1916,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(1916,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(1916,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(1916,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(1916,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(1916,36,'cp 40','f8b4 f1e2 h4e4 d4b5 b4c3 b2c3 e8d8 e1g1 g8f6 c1g5 e4f5 g5h4 d7d6 b5d4 f5f4 g2g3 f4h6 f2f4 h8e8 e2h5 e8e7 h5f3 c8d7 a1b1 c6d4 c3d4 a8b8 f3b7 d8e8 h4g5 h6g6 d1f3 d6d5 g5f6 g7f6 f4f5 g6g5 b7d5 b8b1 f1b1 e8f8 d5e4 f8g7 e4d3 e7e3 f3f4 g5f4 g3f4 d7c6 b1b8 c6d5 b8c8');
INSERT INTO evaluations VALUES(336,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(336,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(336,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(336,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(336,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(336,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(336,36,'cp 64','f1e2 h4e4 d4b5 b4c3 b2c3 e8d8 e1g1 g8f6 e2f3 e4c4 a1b1 c6e5 c1g5 a7a6 b5d4 d7d6 f3b7 c8b7 b1b7 d8c8 b7b4 c4c3 b4b3 c3c5 d4f5 a8b8 f5g7 f6d5 g5d2 b8b5 g7f5 h8g8 f1e1 g8g6 h2h3 e5c4 f5g3 c5c6 d1e2');
INSERT INTO evaluations VALUES(335,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(335,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(335,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(335,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(335,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(335,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(335,36,'cp 63','h4e4 d4b5 b4c3 b2c3 e8d8 e1g1 g8f6 c1g5 d7d6 b5d4 e4e5 g5h4 e5f4 g2g3 f4h6 f2f4 h8e8 e2h5 c6d4 c3d4 e8e7 h4g5 h6h5 g5f6 h5d5 d1f3 c8e6 f3d5 e6d5 f6e7 d8e7 a2a4 b7b6 g1f2 e7d7 f2e3 d5c4 f1e1 g7g6 e3d2 c4e6 d2d3 h7h5 c2c4 e6f5 d3d2 a7a5');
INSERT INTO evaluations VALUES(52122,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(52122,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(52122,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(52122,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(52122,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(52122,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(52122,36,'cp 37','d7d6 f1c4 f8e7 d2d3 b8c6 a2a4 c6b4 h2h3 g8f6 e1g1 e8g8 f1e1 b4c6 c1d2 h7h6 f3h2 c8e6 h2f1 d6d5 e4d5 f6d5 c3d5 e6d5 d2c3 d5c4 d3c4 f7f6 d1f3 c6d4 f3b7 d4c2 a1d1 c2d4 b7e4 g8h8 f1e3 d8d7 a4a5 a8d8 a5a6');
INSERT INTO evaluations VALUES(49890,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(49890,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(49890,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(49890,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(49890,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(49890,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(49890,36,'cp -99','d7d5 f3e5 f8c5 d2d3 c5f2 e1f1 f2b6 c4b3 d8e7 d3e4 e7e5 e4d5 e5e2 f1e2 c8g4 e2d3 g4f5 d3e2 b8d7 c1e3 d7c5 b1c3 e8c8 h1e1 h8e8 e2f3 f7f6 e3f2 f5g6 e1e8 d8e8 a1e1 e8e1 f2e1 c5b3 c2b3 b6c5 h2h3 c8d7 g2g4 a7a6 c3e4 c5b6 e4c3 b6d4 e1g3 d4c5 c3a4 c5b4 f3e3 h7h5 a4c3 b4c5 e3f3 h5g4 h3g4 g6c2');
INSERT INTO evaluations VALUES(49072,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(49072,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(49072,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(49072,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(49072,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(49072,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(49072,36,'cp -98','f3e5 f8c5 d2d3 c5f2 e1f1 f2b6 c4b3 d8e7 d3e4 e7e5 e4d5 e5e2 f1e2 c8g4 e2d3 g4f5 d3e2 b8a6 c1e3 a6c5 h1e1 e8c8 b1c3 h8e8 e2f3 f7f6 h2h3 h7h5 a1c1 g7g5 b3c4 f5g6 c4b5 e8e5 g2g4 f6f5 e3c5 e5e1 c1e1 b6c5 e1e6 f5g4 h3g4 h5g4 f3g4 g6c2 e6e2 c2h7 b5c4 a7a6 g4g5 b7b5');
INSERT INTO evaluations VALUES(49046,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(49046,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(49046,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(49046,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(49046,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(49046,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(49046,36,'cp -89','f8c5 d2d3 c5f2 e1f1 f2b6 c4b3 d8e7 d3e4 e7e5 e4d5 e5e2 f1e2 c8g4 e2d3 g4f5 d3e2 b8d7 c1e3 d7c5 h1e1 e8c8 b1c3 h8e8 e2f3 h7h5 h2h3 f7f6 a1c1 a7a6 b3c4 f5g6 h3h4 g6f5 g2g3 f5g4 f3f2 g7g5 c4e2 g5h4 e2g4 h5g4 g3h4 f6f5 b2b4 e8e3 f2e3 c5b3 e3f4 b3c1 e1c1 d8h8 c1f1 b6d4');
INSERT INTO evaluations VALUES(42275,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(42275,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(42275,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(42275,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(42275,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(42275,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(42275,36,'cp -94','d2d3 c5f2 e1f1 f2b6 c4b3 d8e7 d3e4 e7e5 e4d5 e5e2 f1e2 c8g4 e2d3 g4f5 d3e2 b8d7 c1e3 d7c5 h1e1 e8c8 b1c3 h8e8 e2f3 h7h5 h2h3 f7f6 e3f2 g7g5 e1e8 d8e8 a1e1 e8e1 f2e1 h5h4 e1f2 f5g6 g2g3 h4g3 f3g3 a7a6 h3h4 c5b3 f2b6 b3d2 b6f2 g5h4 g3h4 g6c2 h4g4 c8d7 f2e3');
INSERT INTO evaluations VALUES(1878,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(1878,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(1878,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1878,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(1878,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1878,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(1878,36,'cp 22','f8c5 c2c3 d7d6 f1e2 f6g6 e1g1 g8f6 f2f3 d6d5 d1d2 d5e4 d4c6 c5e3 d2e3 b7c6 b1d2 e8g8 f3e4 f8e8 a1e1 c8e6 e4e5 f6d5 e3f3 g6g5 d2e4 g5e3 f3e3 d5e3 f1f3 e3c4 f3g3 e8e7 e2c4 e6c4 b2b3 c4e6 e4c5 a8d8 a2a4 e6d5 a4a5');
INSERT INTO evaluations VALUES(30653,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(30653,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(30653,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(30653,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(30653,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(30653,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(30653,36,'cp 78','f1d3 f8b4 e1f1 b4e7 g1f3 g8f6 b1c3 d7d6 g2g3 e8g8 f1g2 h7h6 h1e1 b8d7 a1b1 a7a6 b2b4 c7c6 c1f4 a6a5 a2a3 a5b4 a3b4 a8a3 d1c2 d8a8 f4c1 a3a7 c1b2 d6d5 e4d5 c6d5 c4c5 b6c5 b4c5 f8c8 f3e5 d7e5 e1e5');
INSERT INTO evaluations VALUES(30651,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(30651,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(30651,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(30651,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(30651,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(30651,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(30651,36,'cp 64','f8b4 e1f1 b4e7 b1c3 d7d6 g1f3 b8d7 f1g1 g8f6 h2h3 e8g8 d1e2 a7a6 a2a4 e6e5 d4d5 c7c6 g1h2 d7c5 d3c2 c6d5 c4d5 a6a5 c1e3 b7a6 e2d2 a8c8 h1g1 h7h6 g1b1 c5d7 h2g1 f8e8 b2b4 a5b4 b1b4 d8c7 a1a3 e7f8');
INSERT INTO evaluations VALUES(27980,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(27980,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(27980,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(27980,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(27980,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(27980,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(27980,36,'cp 69','e1f1 b4e7 b1c3 d7d6 g1f3 b8d7 f1g1 g8f6 h2h3 e8g8 d1e2 a7a6 a2a4 e6e5 d4d5 c7c6 d5c6 b7c6 d3c2 d8c7 g1h2 a6a5 b2b3 d7c5 h1e1 f8e8 c1d2 g7g6 a1d1 c7b7 d2h6 a8d8 c3d5 f6d5 c4d5 c6d7 e2d2 e7f6 h6e3');
INSERT INTO evaluations VALUES(50102,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(50102,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(50102,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50102,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(50102,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(50102,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(50102,36,'cp 56','c2c4 e7e6 f1d3 b7b6 g1f3 c8b7 e1g1 f8e7 a2a3 e8g8 b2b4 f6e4 f3d2 c7c5 b4c5 b6c5 d2e4 f5e4 d3e2 c5d4 e3d4 b8c6 b1c3 c6a5 c1e3 d7d5 c4c5 b7c6 a1b1 e7f6 c3b5 f6e7 f2f3 e4f3 f1f3 f8f3');
INSERT INTO evaluations VALUES(50084,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(50084,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(50084,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(50084,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(50084,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(50084,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(50084,36,'cp 45','e7e6 g1f3 b7b6 f1d3 c8b7 e1g1 f8e7 a2a3 e8g8 d1c2 b7f3 g2f3 g8h8 b1c3 d7d5 c4d5 f6d5 c3e2 e7g5 f3f4 g5e7 g1h1 c7c5 d4c5 b6c5 e3e4 f5e4 d3e4 e7d6 e4h7 d8h4 h7d3 b8c6 f1g1 d6f4 c1f4 d5f4 e2f4 f8f4 c2c5');
INSERT INTO evaluations VALUES(49937,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(49937,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(49937,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(49937,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(49937,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(49937,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(49937,36,'cp 53','f1d3 b7b6 g1h3 b8c6 e1g1 g7g6 b2b3 f8g7 c1a3 d7d6 b1c3 e8g8 d3e2 d8d7 a1c1 c8b7 e2f3 c6d8 f3b7 d8b7 d1d2 b7d8 f2f3 d8f7 h3f2 a8e8 f1d1 d7e7 d1e1 g7h6 c1d1 c7c5 a3b2 c5d4 d2d4');
INSERT INTO evaluations VALUES(12351,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(12351,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(12351,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(12351,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(12351,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(12351,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(12351,36,'cp -27','f6e4 c3e4 d7d5 c4b5 d5e4 f3e5 d8c7 e5c6 a7a6 c6a7 a6b5 a7b5 c7c6 a2a4 f8e7 d2d4 e4d3 d1d3 e8g8 d3e2 f8d8 c1f4 f7f6 f4g3 e7f8 f1d1 d8d1 e2d1 e6e5 c2c4 c8f5 f2f3 c6d7 d1d7 f5d7 a4a5 g8f7 g3e1 d7c6 b5c7 a8b8 a5a6 b7a6');
INSERT INTO evaluations VALUES(19464,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(19464,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(19464,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19464,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(19464,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19464,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(19464,36,'cp -12','g8f6 b1c3 d7d5 e4d5 f6d5 f1g2 d5c3 b2c3 g7g6 e1g1 f8g7 d2d4 c5d4 f3d4 c6d4 c3d4 d8d4 d1d4 g7d4 a1b1 e8g8 c1a3 d4f6 f1e1 c8e6 b1b7 a8c8 g2e4 f8e8 b7a7 e6a2 g1g2 a2c4 a7b7 c4e6 b7a7 c8c4 e4d3 c4c3 g2f1 c3c8 a7a6 e6d5');
INSERT INTO evaluations VALUES(12842,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(12842,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(12842,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(12842,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(12842,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(12842,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12842,36,'cp -17','b1c3 d7d5 e4d5 f6d5 f1g2 d5c3 b2c3 g7g6 e1g1 f8g7 d2d4 e8g8 d4d5 g7c3 c1h6 c6d4 h6f8 d8f8 a1b1 a8b8 f3d4 c5d4 d1d3 c8d7 f1e1 a7a5 e1e5 b7b5 d5d6 e7d6 e5b5 b8b5 b1b5 d7b5 d3b5 f8c8 b5c6 c8c6 g2c6 g8g7 c6b5 g6g5 a2a4 h7h5 b5c6 h5h4 g3h4');
INSERT INTO evaluations VALUES(5411,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(5411,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(5411,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(5411,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(5411,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(5411,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(5411,36,'cp 109','b1c3 d7d6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 h7h6 b5d3 g8h7 c1e3 f7f5 d4d5 f5e4 d3e4 e6d5 c3d5 c8f5 e4f5 e7f5 e3d2 f5h4 f3h2 f8f5 d5e3 f5f7 d2c3 d8g5 c3g7 f7g7 h2f3 g5f6 f3h4 f6h4 d1d6 a8d8 d6c5 d8d2 e3d5 d2b2 e1e8');
INSERT INTO evaluations VALUES(5408,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(5408,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(5408,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(5408,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(5408,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(5408,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(5408,36,'cp 105','d7d6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 a7a6 b5c6 e7c6 c1e3 b7b5 d1d2 c8b7 a1d1 a8c8 e3h6 c6e7 a2a3 g8h8 h6g7 h8g7 f3h2 e7g8 h2g4 d8h4 f2f3 g8f6 g4f6 h4f6 d2f2 g7g8 c3e2 f8e8 e2g3 f6f4');
INSERT INTO evaluations VALUES(5542,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(5542,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(5542,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(5542,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(5542,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(5542,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(5542,36,'cp 132','h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 a7a6 b5c6 e7c6 c1g5 c6e7 d4d5 h7h6 g5e3 e6d5 e4d5 e7f5 e3f4 b7b5 c3e4 c8b7 d1d2 a8c8 g2g4 f5h4 f3h4 d8h4 e4d6 h4h3 d6b7 c8c4 e1e3 h3g4 f4g3 g7d4 e3d3');
INSERT INTO evaluations VALUES(5541,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(5541,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(5541,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(5541,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(5541,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(5541,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(5541,36,'cp 102','e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 c8d7 c1f4 c6d4 f3d4 e6e5 f4e3 e5d4 e3d4 g7d4 d1d4 d7b5 c3b5 d6d5 a1d1 d5e4 d4e4 d8b6 b5d6 e7c6 d1d2 a8d8 a2a3 d8d7 e4h4 b6d8 h4f4 d8c7 e1d1 d7e7 b2b4 e7e6 b4b5 c6a5 d2d4 b7b6 f4f3 h7h5 f3g3 a5b7 d6b7 c7b7 d4d7 b7e4 d7a7');
INSERT INTO evaluations VALUES(11732,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(11732,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(11732,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(11732,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(11732,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(11732,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(11732,36,'cp 162','f3e5 f8b4 e1g1 e8g8 f2f3 d8e7 e5d3 b4a5 d3f4 d7d5 d2d3 d5e4 d3e4 a5b6 g1h1 a6a5 b2b3 f8d8 d1e1 c8a6 c3e2 a6e2 e1e2 b6d4 a1b1 a5a4 b3b4 c6c5 b4b5 d8d7 f4d3 a8d8 c1g5 c5c4 d3b4 d4b6 e2c4');
INSERT INTO evaluations VALUES(11726,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(11726,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(11726,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(11726,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(11726,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(11726,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(11726,36,'cp 169','f8b4 e1g1 e8g8 f2f3 d8e7 e5d3 b4a5 d3f2 d7d5 d2d3 d5e4 f3e4 a5b6 g1h1 f6e8 d1e2 f7f5 c1e3 e8d6 a1e1 f5e4 f2e4 c8f5 e3b6 c7b6 e4g3 e7e2 g3e2 a8e8 e2d4 f5d7 h1g1 c6c5 f1f8 g8f8 e1f1 f8g8 d4f3 d7c6 g1f2');
INSERT INTO evaluations VALUES(6533,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6533,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(6533,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6533,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(6533,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(6533,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(6533,36,'cp -8','g8h6 g5f7 h6f7 c4f7 e8f7 d1h5 g7g6 h5c5 d7d5 c2c3 d5e4 e1g1 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 f2f3 b8b6 g5f4 g4e6 f4e5 a8d8 d6e4 b6c6 e5f6 d8f8 f6c3 e6d5 e4f2 c6e6 a1d1 d5c6 f2g4');
INSERT INTO evaluations VALUES(10562,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(10562,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10562,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(10562,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(10562,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(10562,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(10562,36,'cp -5','g5f7 h6f7 c4f7 e8f7 d1h5 g7g6 h5c5 d7d5 c2c3 d5e4 e1g1 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 g5f6 b8b6 f6e5 b6c6 f2f3 g4e6 d6e4 a8d8 e5f6 d8d3 f6c3 h7h6 a1e1 g8f7 e4f2 d3d5 a2a3');
INSERT INTO evaluations VALUES(9658,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9658,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9658,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(9658,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9658,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9658,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9658,36,'cp -4','h6f7 c4f7 e8f7 d1h5 g7g6 h5c5 d7d5 c2c3 d5e4 e1g1 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 g5f6 b8b6 f6e5 g4e2 d6e4 b6c6 f2f3 a8d8 e5c3 d8d1 a1d1 e2d1 e4f6 g8g7 f6g4 g7f8 h2h4 c6e6 g4f2 d1e2');
INSERT INTO evaluations VALUES(9664,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(9664,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9664,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9664,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9664,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(9664,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(9664,36,'cp 0','c4f7 e8f7 d1h5 g7g6 h5c5 d7d5 c2c3 d5e4 e1g1 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 f2f3 b8b6 a1d1 g4e6 b2b3 a7a5 h2h4 a5a4 b3a4 a8a4 d6c8 e6c8 d1d8 g8g7 d8c8 a4a2 c8c7 g7f8 c7c8 f8g7');
INSERT INTO evaluations VALUES(9651,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(9651,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(9651,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(9651,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9651,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9651,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(9651,36,'cp -5','e8f7 d1h5 g7g6 h5c5 d7d5 e1g1 d5e4 c2c3 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 f2f3 b8b6 a1d1 g4e6 b2b3 a7a5 h2h4 a5a4 b3a4 a8a4 d6e4 a4a8 d1d2 e6f5 d2e2 f5e4 e2e4 b6b1 g1h2 b1b2 a2a4 b2c2');
INSERT INTO evaluations VALUES(4692,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4692,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4692,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(4692,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(4692,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(4692,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(4692,36,'cp -9','d1h5 g7g6 h5c5 d7d5 c2c3 d5e4 e1g1 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 f2f3 b8b6 a1d1 g4e6 b2b3 a7a5 h2h4 a5a4 b3a4 a8a4 d6e4 a4a8 d1c1 a8c8 c1e1 c8b8 g1h2 b8a8 e4f6 g8g7 f6e4 b6c6 h2g3 e6a2');
INSERT INTO evaluations VALUES(4691,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(4691,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(4691,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(4691,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(4691,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(4691,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(4691,36,'cp -1','g7g6 h5c5 d7d5 c2c3 d5e4 e1g1 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 f2f3 b8b6 a1d1 g4e6 b2b3 a7a5 h2h4 a5a4 b3a4 a8a4 d6e8 e6f5 d1d2 a4a8 e8f6 g8g7 f6d5 b6d6 g2g4 a8a5 d5c7 d6d2');
INSERT INTO evaluations VALUES(4683,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(4683,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(4683,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(4683,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(4683,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(4683,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(4683,36,'cp -9','h5c5 d7d5 e1g1 d5e4 c2c3 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 f2f3 b8b6 a1d1 g4e6 b2b3 a7a5 h2h4 a5a4 b3a4 a8a4 d6e4 a4a8 d1d2 e6f5 e4g3 f5b1 a2a4 b6b7 g1h2 a8a4 d2d6 a4a8 g3f1 b1f5 f1e3');
INSERT INTO evaluations VALUES(4682,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(4682,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(4682,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(4682,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(4682,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(4682,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(4682,36,'cp -5','d7d5 c2c3 d5e4 e1g1 d8d6 c5d6 c7d6 c3d4 c6d4 b1c3 c8g4 c3e4 d4c2 e4d6 f7e7 d6b7 c2a1 c1g5 e7f7 f1a1 h8b8 b7d6 f7g8 f2f3 b8b6 a1d1 g4e6 b2b3 a7a5 h2h4 a5a4 b3a4 a8a4 d6e4 e6f5 d1d8 g8f7 e4d6 f7e6 d6f5 g6f5 d8h8 a4a2 h8h7 b6b2 g1h2 b2g2 h2h3 g2g1 h7e7 e6d6 e7f7 d6d5 f7f5 d5e6 f5f6 e6e5');
INSERT INTO evaluations VALUES(39240,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(39240,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(39240,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(39240,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(39240,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(39240,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(39240,36,'cp 69','g8f6 f3e5 f6e4 d2d4 c5e7 f1d3 e4f6 e1g1 e8g8 f1e1 c7c5 d4c5 e7c5 b1d2 d7d5 d2b3 c5d6 b3d4 a7a6 c1g5 f8e8 e5f3 e8e1 d1e1 b8d7 a1d1 d6f8 g5h4 d7c5 d3c2 d8b6 h4f6 b6f6 e1e8 g7g6 d1e1 a8b8 f3e5 g8g7 h2h3');
INSERT INTO evaluations VALUES(42628,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(42628,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(42628,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(42628,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(42628,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(42628,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(42628,36,'cp 89','f3e5 f6e4 d2d4 c5e7 f1d3 e4f6 e1g1 d7d6 e5f3 e8g8 f1e1 d6d5 c1g5 b8c6 b1d2 h7h6 g5f4 f6h5 f4e3 e7d6 f3e5 h5f6 e3f4 c6e7 d1c2 f6h5 f4g3 h5g3 h2g3 g7g6 e1e2 g8g7 d2f3 c7c6 a1e1');
INSERT INTO evaluations VALUES(46759,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(46759,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(46759,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(46759,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(46759,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(46759,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(46759,36,'cp 84','c4d5 f8g7 d1a4 c7c6 d5c6 b8c6 b1c3 c8f5 e2e3 a7a6 f3e5 e8g8 e5c6 b7c6 f1e2 f6d5 e2f3 a8c8 e1g1 e7e5 f1d1 e5d4 e3d4 a6a5 c3e4 f5e4 f3e4 f8e8 e4f3 d8b6 f3d5 c6d5 c1e3 e8e4 a4b3 b6e6 b3d3 c8a8 a1c1 g7f8 e3d2 f8g7 d2c3 h7h5 h2h3 e4e2');
INSERT INTO evaluations VALUES(46739,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(46739,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(46739,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(46739,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(46739,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(46739,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(46739,36,'cp 60','f8g7 d1a4 c7c6 d5c6 b8c6 e2e3 e8g8 b1c3 c8f5 f1e2 a7a6 e1g1 b7b5 a4d1 a8c8 a2a3 f8e8 f1e1 e7e5 d4e5 d8d1 e1d1 f6e4 c1d2 e4c5 a3a4 c5b3 e3e4 f5e6 a4b5 a6b5 e2b5 b3a1 d1a1 e8d8 d2g5 c6e5 g5d8 e5f3 g2f3 c8d8 b5e2 h7h6 g1g2');
INSERT INTO evaluations VALUES(19813,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(19813,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19813,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(19813,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19813,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(19813,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(19813,36,'cp 39','d2d4 g8f6 e1g1 c5d4 f3d4 e7e6 c2c4 f8c5 d4b3 c5b6 c4d5 f6d5 b1c3 e8g8 c3d5 e6d5 d1d5 d8d5 g2d5 c8g4 a2a4 f8d8 d5c6 b7c6 a4a5 b6c7 f2f3 a8b8 b3c5 g4h3 f1e1 c7e5 c1e3 b8b2 g1f2 b2d2 e3d2 e5d4 e2e3 d4a1 d2b4 a1f6');
INSERT INTO evaluations VALUES(19791,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(19791,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(19791,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(19791,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(19791,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(19791,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(19791,36,'cp 42','g8f6 e1g1 e7e6 c2c4 d5c4 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 c3c4 c8d7 c1b2 f8d8 f3d4 a8c8 d4b5 b7b6 b5d6 e7d6 d1d6 d7e8 d6d8 c6d8 a1c1 e8c6 c4c5 c6g2 g1g2 f6d7 c5b6 c8c1 b2c1 a7b6 c1b2 f7f6 f2f4 g8f7 g2f3 e6e5');
INSERT INTO evaluations VALUES(13109,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(13109,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(13109,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(13109,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(13109,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(13109,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(13109,36,'cp 22','e1g1 e7e6 c2c4 d5c4 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 f3d4 c6d4 c3d4 f6d5 c1d2 f7f5 b3a5 e7f6 a1c1 b7b6 a5c6 c8b7 a2a4 f8c8 c6e5 c8c1 d1c1 a8c8 c1c8 b7c8 e2e4 f5e4 g2e4 c8b7 f2f3 g8f8 e5d7 f8g8 d7f6 g7f6 a4a5 b7c6');
INSERT INTO evaluations VALUES(13108,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(13108,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(13108,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13108,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(13108,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(13108,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(13108,36,'cp 19','e7e6 c2c4 d5c4 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 c1b2 c8d7 c3c4 f8c8 a2a4 d7e8 a4a5 f6d7 f3d2 a8b8 b2a3 e7a3 a1a3 c6b4 d1c1 b7b5 a3a1 d7f6 b3d4 b5c4 d2c4 c8d8 e2e3 b4d3 c1d1 d3c5 d1b1 b8b1 a1b1 h7h6 c4e5');
INSERT INTO evaluations VALUES(13597,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(13597,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(13597,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(13597,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(13597,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13597,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(13597,36,'cp 26','c2c4 d5c4 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 f3d4 c6d4 c3d4 f6d5 c1d2 b7b6 e2e4 d5f6 a2a4 c8a6 d2e3 f8d8 d1c1 a8c8 a4a5 c8c1 b3c1 a6c4 a5b6 a7b6 f2f3 b6b5 g2f1 e7b4 c1a2 b4f8 a2c3 c4f1 g1f1 b5b4');
INSERT INTO evaluations VALUES(13556,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(13556,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(13556,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(13556,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(13556,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(13556,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(13556,36,'cp 15','d5c4 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 c1b2 c8d7 c3c4 f8c8 a2a4 d7e8 a4a5 f6d7 f3d2 a8b8 b2a3 c6b4 a1b1 d7e5 c4c5 b4a6 a3b2 e5g6 d2e4 e8a4 b2d4 a4b3 b1b3 a6c5 e4c5 e7c5 b3b7 c5d4 b7b8');
INSERT INTO evaluations VALUES(13502,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(13502,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(13502,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(13502,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(13502,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(13502,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13502,36,'cp 31','d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 c1b2 c8d7 c3c4 f8c8 c4c5 d7e8 f3d2 b7b5 c5b6 a7b6 d2c4 a8b8 c4d6 e7d6 d1d6 c6a5 b3a5 b6a5 b2e5 h7h6 d6a6 b8b5 e5d4 f6d5 a6a7 b5b4 e2e3 a5a4 h2h3 d5c3 g2f1 b4b2 f1a6 c3e2 g1f1');
INSERT INTO evaluations VALUES(20340,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(20340,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(20340,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(20340,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(20340,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(20340,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(20340,36,'cp -74','c5d4 e3d4 g8e7 b1d2 e7f5 d2b3 f5d4 b3d4 f8c5 d4c6 b7c6 c2c3 d8b6 d1c2 a7a5 a1d1 e8g8 f1d3 h7h6 e1g1 c8a6 c3c4 f8b8 b2b3 a5a4 d1b1 b6a7 d3h7 g8h8 h7d3 b8b4 f1d1 h8g8 c4d5 a6d3 d1d3 a4b3 d3b3 b4b3 a2b3 c6d5 h2h4 c5e7');
INSERT INTO evaluations VALUES(20440,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(20440,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(20440,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(20440,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(20440,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(20440,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(20440,36,'cp -64','e3d4 g8e7 b1d2 c6d4 f3d4 e7c6 d2f3 f8c5 d4c6 b7c6 c2c3 d8b6 d1c2 a7a5 f1d3 c8a6 a1d1 h7h6 e1g1 e8g8 c3c4 f8b8 d1b1 b6a7 b2b3 a5a4 g2g3 b8b4 f1d1 c5e7 h2h4 a8b8 c4d5 c6d5 d3a6 a7a6 d1c1 a4a3 g1g2 e7f8 h4h5 a6b7 c2d3 b8c8 c1c8 b7c8');
INSERT INTO evaluations VALUES(10887,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10887,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(10887,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(10887,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(10887,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(10887,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10887,36,'cp 6','e8g8 d2d3 d7d6 a2a3 c8g4 c1e3 c6d4 e3d4 e5d4 c3d5 f6d7 h2h3 g4f3 d1f3 e7h4 f3e2 a7a6 g2g3 h4g5 f2f4 g5h6 f4f5 c7c6 d5f4 g8h8 g1g2 a6a5 h3h4 b7b5 c4a2 h6f4 f1f4 d7f6 a1f1 c6c5 c2c3 d4c3 b2c3 d6d5');
INSERT INTO evaluations VALUES(4574,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(4574,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(4574,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(4574,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(4574,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(4574,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(4574,36,'cp 39','d2d3 h7h6 h2h3 a7a5 a2a3 d7d6 c1e3 c8e6 c4e6 f7e6 d3d4 c6d4 f3d4 e5d4 d1d4 e6e5 d4c4 g8h8 a1d1 c7c6 a3a4 d8d7 d1d3 f6h5 f1d1 f8f6 c3e2 a8f8 e2g3 b7b5 a4b5 h5g3 f2g3 c6b5 c4b3');
INSERT INTO evaluations VALUES(4577,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(4577,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4577,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(4577,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(4577,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(4577,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(4577,36,'cp 30','d7d6 a2a4 a7a5 h2h3 c8e6 d1e2 d8d7 f1d1 a8e8 c1e3 d6d5 e4d5 f6d5 c3d5 e6d5 d3d4 d5f3 e2f3 e5d4 c4b5 e7f6 c2c3 e8e6 b5c6 d7c6 f3c6 e6c6 e3d4 h7h6 a1b1 f6d4 d1d4 c6b6 d4c4 b6c6 c4c6 b7c6');
INSERT INTO evaluations VALUES(3375,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(3375,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3375,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(3375,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(3375,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(3375,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(3375,36,'cp 32','a2a4 h7h6 f1e1 c6a5 c4a2 c7c5 c1d2 a5c6 h2h3 c8e6 c3d5 e6d5 a2d5 f6d5 e4d5 c6b4 d2b4 c5b4 a4a5 d8c7 e1e4 c7c5 e4c4 c5b5 f3h4 e7h4 c4h4 b5d5 h4b4 d5c6 c2c4 d6d5 d1b3 d5c4');
INSERT INTO evaluations VALUES(32319,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(32319,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32319,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(32319,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(32319,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(32319,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(32319,36,'cp 14','c2c4 c7c6 f1d3 f8b4 b1c3 f5d3 d1d3 g8f6 c1d2 e8g8 e1g1 b4a5 a1b1 b8d7 c4d5 c6d5 f1c1 a8c8 d3b5 a5c3 d2c3 d8b6 b5b6 d7b6 f3e5 f6e4 c3b4 c8c1 b1c1 f8c8 e5d3 c8c1 d3c1 b6c8 c1d3 b7b6 f2f3 e4d6 b2b3 f7f6 b4a3');
INSERT INTO evaluations VALUES(32283,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(32283,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(32283,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(32283,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(32283,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(32283,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(32283,36,'cp 12','c7c6 f1d3 f5g6 d1c2 g8f6 e1g1 b8d7 d3g6 h7g6 b1d2 c6c5 c4d5 f6d5 d4c5 f8c5 d2b3 c5b6 a2a4 e8g8 f1d1 a7a5 c2e2 d8c7 c1d2 d7e5 a1c1 e5f3 e2f3 c7e5 d2c3 e5f5 h2h3 f5f3 g2f3 f8d8 c3d4 b6d4 d1d4 g8f8');
INSERT INTO evaluations VALUES(31446,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(31446,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(31446,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31446,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31446,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31446,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31446,36,'cp 23','b1c3 f8e7 f1d3 f5d3 d1d3 g8f6 e1g1 e8g8 f1d1 d5c4 d3c4 b7b5 c4d3 b8d7 a2a3 a7a5 e3e4 b5b4 c3a4 d7b6 a4b6 d8b6 f3e5 f8d8 a3b4 a5b4 a1a8 d8a8 c1g5 b6c7 e5c4 h7h6 g5h4 c7a7 d3f3 a8d8 h4f6 e7f6 e4e5');
INSERT INTO evaluations VALUES(47069,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(47069,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(47069,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(47069,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(47069,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(47069,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(47069,36,'cp 59','d6e5 f3e5 c7c6 f1e2 c8f5 e1g1 b8d7 e5f3 e7e6 c2c4 d5f6 c1f4 d8b6 d1c1 f8e7 b1c3 e8g8 h2h3 h7h6 c4c5 b6d8 b2b4 a7a6 a2a4 f6d5 c3d5 e6d5 b4b5 a6b5 a4b5 a8a1 c1a1 f8e8 f1e1 e7f6 f4d6 d7f8 a1a4 d8c8 b5c6 b7c6 f3e5');
INSERT INTO evaluations VALUES(47478,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(47478,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(47478,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(47478,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(47478,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(47478,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(47478,36,'cp 61','f3e5 c7c6 f1e2 c8f5 e1g1 b8d7 e5f3 e7e6 c2c4 d5f6 c1f4 d8b6 b2b3 f8e7 a2a3 e8g8 b1c3 b6d8 h2h3 f6e4 c3e4 f5e4 e2d3 e4d3 d1d3 a7a5 f1d1 f8e8 g2g3 h7h6 g1g2 d8c8 d1e1 e8d8 h3h4 d7f6 f4e5');
INSERT INTO evaluations VALUES(47469,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(47469,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(47469,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(47469,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(47469,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(47469,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(47469,36,'cp 48','c7c6 f1e2 c8f5 e1g1 b8d7 e5f3 e7e6 c2c4 d5f6 c1f4 f8e7 b1c3 e8g8 h2h3 d8b6 c3a4 b6a5 a2a3 b7b5 c4b5 c6b5 a4c5 a5b6 g2g4 f5e4 c5e4 f6e4 d1d3 e4d6 a3a4 a7a6 f3e5 f8d8 a1c1 a8c8 e5d7 d8d7 c1c8 d6c8 f1c1 c8d6 a4b5');
INSERT INTO evaluations VALUES(19100,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(19100,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(19100,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(19100,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(19100,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(19100,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(19100,36,'cp 53','b5c4 d7d5 e4d5 c6d5 c4b5 c8d7 f1e1 g8e7 c2c4 a7a6 b5d7 d8d7 b1a3 f7f6 b2b4 e8f7 d1b3 e7f5 c1b2 a8d8 c4c5 a6a5 b4a5 f8c5 a1c1 h8e8 e1e8 d8e8 a3b5 h7h5 a2a4 e8c8 g1f1 c8c6 h2h3 h5h4 b3f3 g7g6 f3b3 d7e6 c1e1 e6d7 f1g1 c6a6 e1c1 a6c6 c1b1 d7e6 b2a3 e6e4 d2d3 e4e5');
INSERT INTO evaluations VALUES(19109,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(19109,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(19109,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(19109,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(19109,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(19109,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(19109,36,'cp 69','d7d5 e4d5 c6d5 c4b5 c8d7 f1e1 g8e7 b1a3 a7a6 b5d7 d8d7 c2c4 f7f6 b2b4 e8f7 d1b3 e7f5 c1b2 a8d8 c4c5 a6a5 b4a5 f8c5 a1c1 h8e8 e1e8 d8e8 a3b5 e8c8 h2h3 c8c6 a2a4 h7h5 g1f1 h5h4 b3d3 c6c8 d3f3 g7g6 f3d3 c5b4 c1c8 d7c8 b5d4 f5d4 d3d4 c8c4 d4c4 d5c4 b2c3 b4c3 d2c3 f7e6 g2g3');
INSERT INTO evaluations VALUES(20739,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(20739,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(20739,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(20739,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(20739,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(20739,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(20739,36,'cp 60','e4d5 c6d5 c4b5 c8d7 f1e1 g8e7 c2c4 a7a6 b5d7 d8d7 b1a3 f7f6 b2b4 e8f7 d1b3 e7f5 c1b2 a8d8 c4c5 a6a5 b4a5 f8c5 a1c1 h8e8 e1e8 d8e8 a3b5 e8c8 h2h3 c8c6 a2a4 h7h5 b3f3 h5h4 b2a3 c5a3 c1c6 d7c6 f3f5 a3d6');
INSERT INTO evaluations VALUES(20737,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(20737,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(20737,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(20737,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(20737,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(20737,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(20737,36,'cp 62','c6d5 c4b5 c8d7 f1e1 g8e7 c2c4 a7a6 b5d7 d8d7 b1a3 f7f6 c4d5 e8f7 a3c4 e7d5 d1f3 a8e8 e1e8 f7e8 d2d3 b7b5 c4d2 e8f7 d2b3 d5b4 c1d2 d7d5 f3d5 b4d5 a2a3 f8d6 a1c1 d5e7 f2f4 h8c8 c1c8 e7c8 g1f2 c8b6 f2f3 f6f5 b3d4 g7g6 h2h3 d6e7 d4c6 b6d5 c6b8 e7f6 b2b3 f6b2 b8a6 b2a3');
INSERT INTO evaluations VALUES(20784,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(20784,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(20784,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(20784,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(20784,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(20784,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(20784,36,'cp 59','c4b5 c8d7 f1e1 g8e7 b1a3 a7a6 b5d7 d8d7 c2c4 f7f6 b2b4 e8f7 d1b3 e7f5 c1b2 a8d8 c4c5 a6a5 b4a5 f8c5 a1c1 h8e8 e1e8 d8e8 a3b5 h7h5 g1f1 e8c8 h2h3 h5h4 a2a4 c8c6 b3d3 g7g6 c1b1 c5e7 b2d4 f5d4 d3d4 c6c4 d4a7 e7c5 a7b8 f7g7 b5c3');
INSERT INTO evaluations VALUES(20782,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(20782,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(20782,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(20782,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(20782,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(20782,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(20782,36,'cp 78','c8d7 f1e1 g8e7 b1a3 a7a6 b5d7 d8d7 c2c4 f7f6 b2b4 e8f7 d1b3 e7f5 c1b2 a8d8 c4c5 a6a5 b4a5 f8c5 a1c1 d8c8 a3b5 h7h5 a2a3 g7g6 h2h3 c8c6 g1f1 h8c8 b3d3 c6e6 e1e6 d7e6 g2g4 h5g4 h3g4 f5d6 b2d4 d6c4 d4c5');
INSERT INTO evaluations VALUES(25994,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(25994,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(25994,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(25994,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(25994,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(25994,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(25994,36,'cp 58','f1e1 g8e7 c2c4 d7b5 c4b5 d8d7 b1a3 f7f6 b2b4 e8f7 d1b3 e7g6 a3c2 d7b5 c2d4 b5d7 d2d3 f8e7 a2a4 h8e8 h2h3 a7a6 c1d2 a8c8 a1b1 f7g8 b1c1 c8c1 e1c1 e7d8 b4b5 d8c7 b5b6 c7e5 d4f3 g6f4 g1f1 h7h5 d2f4');
INSERT INTO evaluations VALUES(48881,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(48881,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(48881,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(48881,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(48881,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(48881,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(48881,36,'cp -44','d2d4 b8d7 d4e5 d7e5 b1d2 c7c6 g1f3 e5f3 d2f3 f8b4 c1d2 b4e7 e1g1 e8g8 d1e2 d8b6 d2c3 f6e4 c3e5 f7f6 e5c3 a7a5 a2a4 c8g4 e2d3 f8e8 f3d4 e7f8 g2f3 g4d7 d4e2 a8d8 d3d4 b6d4 e2d4 e4c3 b2c3');
INSERT INTO evaluations VALUES(48803,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(48803,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(48803,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(48803,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(48803,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(48803,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(48803,36,'cp -47','b8c6 d4e5 c6e5 b1d2 a7a5 g1f3 e5f3 d2f3 f8b4 c2c3 b4e7 c3c4 e8g8 c4d5 d8d5 e1g1 f8d8 d1d5 f6d5 f1d1 c7c6 a2a4 c8f5 f3d4 f5g4 g2f3 h7h5 c1d2 e7f6 f3g4 h5g4 d2c3 g7g6 a1c1 f6g7 b2b3 d8e8 g1f1 d5c3 c1c3');
INSERT INTO evaluations VALUES(10342,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(10342,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(10342,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(10342,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(10342,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(10342,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(10342,36,'cp -52','e8g8 a2a3 b4a5 e1g1 a5c3 d2c3 h7h6 g5f3 f6e4 f1e1 e4d6 c4d5 f8e8 h2h3 d8f6 f3h2 f6g6 h2g4 a7a5 a3a4 g8h7 d5c6 d7c6 g4e5 g6f6 e5d3 e8e1 d1e1 c8f5 c1f4 a8e8 e1d1 f5g6 f4g3 b7b6 f2f3 f6g5 g3f4 g5d5 d1f1 g6d3 f1d3 d5d3 c2d3 e8e2 f4d6 c7d6');
INSERT INTO evaluations VALUES(4423,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(4423,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(4423,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(4423,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(4423,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(4423,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(4423,36,'cp -38','a2a3 b4a5 e1g1 h7h6 g5f3 a5c3 d2c3 f6e4 f1e1 e4d6 c4d5 f8e8 h2h3 d8f6 f3h2 f6g6 h2g4 a7a5 a3a4 g8h7 d5c6 d7c6 g4e5 g6f6 e5d3 e8e1 d1e1 c8f5 c1f4 a8e8 e1f1 f6g6 f4g3 f7f6 g3d6 c7d6 a1e1 e8e1 f1e1 f5d3 c2d3 g6d3');
INSERT INTO evaluations VALUES(4425,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(4425,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(4425,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(4425,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(4425,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(4425,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(4425,36,'cp -47','b4a5 e1g1 a5c3 d2c3 h7h6 g5f3 f6e4 c4d5 e4d6 h2h3 d8f6 f1e1 f8e8 f3h2 f6g6 h2g4 g8h7 d5c6 d7c6 g4e5 g6f6 e5d3 e8e1 d1e1 c8f5 c1f4 a8e8 e1f1 f6g6 f4g3 f5e4 g1h2 e4d3 c2d3 g6e6 c3c4 e6e2 b2b4 e2f1 a1f1 e8e2 a3a4 e2a2');
INSERT INTO evaluations VALUES(34781,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(34781,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(34781,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(34781,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(34781,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(34781,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(34781,36,'cp 213','b1c3 c8g4 d1d3 b8d7 c1g5 f6g6 g5e3 c7c6 f1e2 f8e7 h2h3 g4f3 g2f3 e5d4 d3d4 g6f6 d4d2 f6e6 h1g1 g7g6 f3f4 g8f6 d2d4 d6d5 e4d5 c6d5 f4f5 g6f5 c3b5 e8d8 d4f4 e6e5 f4e5 d7e5 e1c1 e5d7 e3a7 e7c5 a7c5 d7c5');
INSERT INTO evaluations VALUES(34779,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(34779,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(34779,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(34779,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(34779,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(34779,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(34779,36,'cp 215','c8g4 d1d3 b8d7 c1g5 f6g6 g5e3 c7c6 e1c1 f8e7 f1e2 a8d8 h1g1 f7f5 d4e5 d6e5 h2h3 f5e4 c3e4 g4f5 e4d6 e7d6 d3d6 g6d6 d1d6 e8e7 d6d1 g8f6 g1e1 f5e4 e2c4 h7h6 e3a7 b7b6 d1d7 e7d7 a7b6 d8e8 a2a4 e4d5 f3d2 d7c8 g2g3 f6d7 b6e3 c8c7 a4a5 e8a8 b2b4 d5c4 d2c4');
INSERT INTO evaluations VALUES(56079,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(56079,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(56079,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(56079,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(56079,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(56079,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(56079,36,'cp 61','c3e4 b8d7 c2c3 d7f6 e4f6 g8f6 d2d4 a7a6 c1f4 c7c5 f1e2 c5d4 d1d4 d8d4 f3d4 c8d7 e1c1 e8c8 e2f3 f8c5 f4e5 c5b6 h1e1 h7h5 h2h3 h5h4 c1c2 f6e8 d4b3 d7b5 d1d8 c8d8 b3d4 d8c8 d4b5 a6b5 c2b3 b6f2 e1d1 f2c5');
INSERT INTO evaluations VALUES(55770,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(55770,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(55770,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(55770,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(55770,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(55770,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(55770,36,'cp -11','e6d5 e4d5 g8f6 f1b5 b8d7 g1f3 f8d6 e1g1 e8g8 d2d4 d7b6 c1g5 b6d5 b1c3 c8e6 b5d3 h7h6 g5h4 d5f4 f1e1 c7c6 f3e5 f4d3 d1d3 f8e8 h2h3 d6e7 e5g6 f6d5 g6e7 d5e7 a1d1 d8a5 e1e5 a5b4 h4e7 e8e7 d4d5 a8d8 d3f3 c6d5 c3d5 d8d5 d1d5 e6d5 f3d5');
INSERT INTO evaluations VALUES(42537,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42537,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42537,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42537,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42537,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42537,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42537,36,'cp 0','c7c6 g1f3 d7d5 e4d5 c6d5 c4b3 b8c6 b1c3 d5d4 c3d5 c8e6 d5f6 g7f6 g5h4 c5e7 e1g1 h8g8 f3d2 e6b3 a2b3 d8d5 g2g3 d5e6 g1h1 e6d5');
INSERT INTO evaluations VALUES(57387,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(57387,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(57387,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(57387,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(57387,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(57387,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(57387,36,'cp 74','f8c5 d1h5 g7g6 h5f3 d7d6 b1c3 a7a6 h2h4 h7h5 c1d2 b8d7 e1c1 d7f8 f3g3 d8e7 f2f3 c8d7 g1h3 e8c8 f1c4 d7h3 g3h3 f8d7 c3e2 c8b8 b2b4 c5a7 c4d3 e7e8 h3e6 g8e7 c1b1 h8f8 d2h6 f8f7 a2a4');
INSERT INTO evaluations VALUES(39466,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(39466,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(39466,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(39466,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(39466,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(39466,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(39466,36,'cp 77','b1c3 d7d6 d1h5 g7g6 h5f3 b8d7 c1d2 a7a6 h2h4 h7h5 e1c1 d7f8 f3g3 c8d7 c1b1 d8e7 d2e3 c5e3 g3e3 f8h7 g1h3 f6f5 e4f5 g6f5 g2g3 g8f6 f2f3 e8g8 a2a3 e7g7 h3g5 b7b5 f1d3 g7e7 d1e1 g8h8');
INSERT INTO evaluations VALUES(41290,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(41290,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(41290,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(41290,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(41290,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41290,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(41290,36,'cp 63','e8g8 c1e3 e7e5 d4d5 b8a6 h2h4 a6c5 d1c2 b7b6 h4h5 f6h5 e2h5 g6h5 c2e2 h5h4 e2h5 d8f6 h1h4 f6g6 h5g6 h7g6 e1c1 f7f5 e3c5 b6c5 g1f3 g7f6 h4h6 f5e4 c3e4 f6g7 h6h4 f8f4 d1h1 c8f5 f3g5 a8f8 f2f3 f5e4 g5e4 f8b8 h4h7');
INSERT INTO evaluations VALUES(55894,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(55894,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(55894,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(55894,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(55894,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(55894,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(55894,36,'cp -39','d5d4 c3e2 c7c5 g2g3 b8c6 f1g2 g8f6 f2f4 b7b5 g1f3 f8e7 e1g1 c8b7 a2a4 b5a4 a1a4 a7a5 b2b3 e8g8 a4a1 f6g4 h2h3 g4e3 c1e3 d4e3 g1h2 d8b6 e2c3 c6d4 f1e1 b6b4 c3a4 e3e2 e1e2 d4e2 d1e2 b7a6 e4e5 a8b8');
INSERT INTO evaluations VALUES(56058,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(56058,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(56058,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(56058,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(56058,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(56058,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(56058,36,'cp -21','c3e2 c7c5 f2f4 b8c6 g1f3 f8e7 g2g3 f7f5 a2a3 f5e4 d3e4 g8h6 c2c3 e8g8 c3d4 d8b6 f1g2 c5d4 d1d3 e6e5 e1g1 h6f7 b2b4 c8g4 g1h1 a7a5 b4b5 c6d8 f4e5 g4f3 f1f3 d8e6 e2f4 f7e5 f4d5 e5d3 d5b6 f8f3 g2f3');
INSERT INTO evaluations VALUES(56060,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(56060,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(56060,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(56060,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(56060,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(56060,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(56060,36,'cp -20','c7c5 f2f4 b8c6 g1f3 f8e7 a2a4 d8c7 g2g3 b7b6 f1g2 g8f6 e1g1 c8b7 c1d2 e8g8 b2b3 a7a6 h2h3 b6b5 g3g4 f6d7 g4g5 c5c4 b3c4 b5c4 f3d4 e7c5 d2e3 a8d8 d1d2 c4d3 c2d3 c6d4 e2d4 e6e5 d4f5 e5f4');
INSERT INTO evaluations VALUES(57065,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(57065,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(57065,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(57065,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(57065,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(57065,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(57065,36,'cp 29','d2d4 d7d5 e2e4 g8f6 e4e5 f6d7 f2f4 c7c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 e8g8 d2f2 c6d4 e3d4 d8c7 f1d3 c7b6 d4c5 d7c5 e1g1 b6c7 f1e1 c8d7 a1d1 c5d3 d1d3 d7b5 c3b5 a6b5');
INSERT INTO evaluations VALUES(14052,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(14052,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(14052,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(14052,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(14052,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(14052,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(14052,36,'cp 138','c2c3 e7f5 f1d3 f7f6 d1e2 f6e5 d4e5 a7a5 h2h4 d8d7 g2g4 f5h6 f3g5 b7b6 g5h7 h8h7 d3h7 c8a6 h7d3 a6d3 e2d3 c6e5 d3e2 e5g4 c1h6 g4h6 b1d2 e8c8 e1c1 c8b7 h1e1 d7a4 c1b1 a4h4 e2e6 h4g4 e6e2 g4e2 e1e2');
INSERT INTO evaluations VALUES(14049,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(14049,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(14049,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(14049,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(14049,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(14049,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(14049,36,'cp 127','e7f5 f1d3 f7f6 b1d2 f6e5 d4e5 a7a5 h2h4 a5a4 d1e2 f8c5 g2g4 f5e7 h4h5 h7h6 g4g5 e7f5 b2b4 a4b3 d2b3 b7b6 g5h6 g7h6 h1g1 d8d7 g1g6 c6e7 g6f6 e8d8 d3b5 e7c6 b3c5 b6c5 c1f4 c8b7 e1c1 d8c8 d1g1');
INSERT INTO evaluations VALUES(23163,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(23163,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(23163,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(23163,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(23163,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(23163,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(23163,36,'cp 92','f3e5 a5c4 e5c4 g8f6 d2d3 d7d5 e4d5 f6d5 c1d2 b7b5 c4a5 d8e7 d1e2 c8d7 a2a3 e7e2 e1e2 e8c8 h1e1 f8d6 b1c3 d5c3 d2c3 h8e8 e2f1 g7g6 a5b3 e8e1 a1e1 d8e8 e1e8 d7e8 g2g3 c7c5 f1e2 e8c6 b3d2 f7f5 e2e3 c8d7 d2f3 c6d5 f3e5 d6e5 c3e5 d7e6 e5c3 c5c4 d3d4 d5b7 e3f4');
INSERT INTO evaluations VALUES(23160,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(23160,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(23160,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(23160,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(23160,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(23160,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(23160,36,'cp 87','a5c4 e5c4 g8f6 d2d3 d7d5 e4d5 f6d5 e1g1 f8e7 c1d2 c8e6 b1c3 d5c3 d2c3 b7b5 c4e3 e8g8 f1e1 b5b4 c3e5 f7f6 e5f4 f8e8 d3d4 c7c6 c2c4 b4c3 b2c3 d8a5 d1d3 a5a4 c3c4 e6f7 d3d1 a4d1 e1d1 e7b4 d4d5 e8e4 f4c7 c6d5');
INSERT INTO evaluations VALUES(23063,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(23063,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(23063,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(23063,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(23063,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(23063,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(23063,36,'cp 109','e5c4 g8f6 d2d3 d7d5 e4d5 f8e7 b1c3 f6d5 c1d2 c8e6 e1g1 e8g8 d1f3 f8e8 c4e3 d5c3 d2c3 b7b5 b2b3 d8d7 a2a4 b5b4 c3b2 a8d8 a1e1 f7f6 h2h3 e7f8 e1e2 e6f7 f1e1 e8e6 e3c4 d8e8 g1f1 h7h6 e2e6 e8e6 e1e6 d7e6');
INSERT INTO evaluations VALUES(23140,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(23140,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(23140,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(23140,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(23140,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(23140,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(23140,36,'cp 87','g8f6 d2d3 d7d5 e4d5 f8e7 e1g1 c8g4 d1d2 f6d5 b2b3 e8g8 c1b2 g4e6 b1c3 d5c3 b2c3 b7b5 c4e3 d8d7 a2a4 a7a6 f1e1 f8e8 d3d4 c7c6 h2h3 e7g5 d2d3 g5e3 e1e3 e6f5 e3e8 a8e8 d3d2 f7f6 a4b5 a6b5 c3b4 f5g6 b4c5 d7d5');
INSERT INTO evaluations VALUES(8052,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8052,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8052,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8052,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8052,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8052,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8052,36,'cp 0','e8g8 d4b3 d7d5 c3d5 f6d5 e4d5 c6b4 d1d2 b4d5 e1c1 c8e6 f1c4 d8c7 c4d5 f8d8 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(3078,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3078,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3078,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3078,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3078,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3078,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3078,36,'cp 0','d4b3 d7d5 c3d5 f6d5 e4d5 c6b4 d1d2 b4d5 e1c1 c8e6 f1c4 d8c7 c4d5 f8d8 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(52302,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(52302,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(52302,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(52302,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(52302,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(52302,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(52302,36,'cp 41','d2d4 d7d5 d4c5 d5e4 d1d8 e8d8 b1d2 g8f6 f2f3 b8d7 d2e4 f6e4 f3e4 d7c5 e4e5 c8d7 g1f3 d8e8 f1e2 c5e6 e1g1 h7h5 h2h4 a8c8 a2a4 a7a6 a4a5 f8h6 c1h6 h8h6 f3d4 e6d4 c3d4 d7g4 e2d3 c8d8 d3e4 d8d4 e4b7 g4e2 f1c1 e2b5 c1c7 e8f8');
INSERT INTO evaluations VALUES(52292,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(52292,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(52292,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52292,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52292,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(52292,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52292,36,'cp 48','d7d5 d4c5 d5e4 d1d8 e8d8 b1d2 g8f6 f2f3 b8d7 d2e4 f6e4 f3e4 d7c5 e4e5 c8d7 g1f3 d8e8 f1c4 d7f5 e1g1 f5d3 c4d3 c5d3 c1e3 f8g7 e3d4 d3f4 f1d1 f4e6 h2h4 h8f8 d4e3 f7f6 e5f6 g7f6 d1e1 e8f7 g1f2 f8d8 g2g4 d8d7 a1d1 d7d1');
INSERT INTO evaluations VALUES(29017,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(29017,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(29017,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(29017,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(29017,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(29017,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(29017,36,'cp 21','e7e6 h2h3 g4f3 d1f3 f8b4 f1d3 e8g8 c4d5 c6d5 e1g1 b8c6 c1d2 b4d6 a1c1 a8c8 f3e2 f8e8 f1d1 g7g6 d2e1 h7h5 f2f3 d6b8 e1h4 g8g7 c3a4 b7b6 a4c3 c6e7 h4g5 e7f5 d3f5 e6f5 g5f6 g7f6');
INSERT INTO evaluations VALUES(29276,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(29276,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(29276,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(29276,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(29276,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(29276,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(29276,36,'cp 28','h2h3 g4f3 d1f3 f8b4 f1d3 e8g8 e1g1 b8d7 f3e2 f8e8 c1d2 d8e7 f1d1 d7b6 a2a3 b4c3 d2c3 d5c4 d3c4 b6c4 e2c4 a7a5 a1c1 e8d8 c4e2 f6d5 c3e1 h7h6 e2f3 e6e5 d4e5 e7e5 d1d4 a5a4 f3g4 d5f6 d4d8 a8d8 g4a4');
INSERT INTO evaluations VALUES(29274,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(29274,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(29274,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(29274,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(29274,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(29274,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(29274,36,'cp 19','g4f3 d1f3 f8b4 f1d3 e8g8 e1g1 b8d7 f3e2 f8e8 c1d2 d8e7 f1d1 b4c3 d2c3 e6e5 d4e5 d7e5 c3d4 e5d3 e2d3 d5c4 d3c4 f6e4 f2f3 e4g5 c4c5 e7c5 d4c5 g5e6 c5b4 a8d8 g1f2 f7f6 b4c3 g8f7 a2a4 d8d1 a1d1 e8d8 d1d8 e6d8');
INSERT INTO evaluations VALUES(46776,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(46776,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(46776,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(46776,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46776,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46776,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(46776,36,'cp 28','b1d2 f8g7 e2e3 e8g8 h2h3 b7b6 c2c3 c7c5 f1e2 a7a6 e1g1 c8b7 a2a4 b8d7 d1b3 a8c8 f1d1 e7e6 c3c4 a6a5 f4d6 f8e8 a1c1 g7f8 d6f8 e8f8 f3e5 b7a6 e2d3 c5d4 e3d4 d5c4 d2c4 f6d5 c4d6 c8c1 d1c1 a6d3');
INSERT INTO evaluations VALUES(49016,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(49016,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(49016,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(49016,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(49016,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(49016,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(49016,36,'cp 159','f3e5 d5c3 d2c3 d8d1 e1d1 b8d7 c1f4 d7e5 f4e5 c8g4 d1c1 e8c8 f1d3 a7a5 h2h4 h7h5 a2a4 g4e6 c1d2 f8c5 a1e1 c5f2 e1e2 f2c5 e5g7 h8g8 g7f6 d8d5 f6g5 c5d6 d2c1 d5g5 h4g5 g8g5 b2b3 c8d7 h1h4 d6g3 h4d4 d7e7 c1b1 b7b6');
INSERT INTO evaluations VALUES(48993,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(48993,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(48993,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(48993,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(48993,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(48993,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(48993,36,'cp 147','d5c3 d2c3 d8d1 e1d1 b8d7 e5d7 c8d7 f1d3 e8c8 c1e3 d7c6 h1g1 f8e7 g2g3 h8e8 d1d2 g7g6 h2h4 f7f5 g1e1 c6e4 a2a4 a7a5 e3g5 e7g5 h4g5 c7c5 b2b3 e8h8 e1e3 h7h5 g5h6 h8h6 d2e2 e4c6 a1e1 h6h2 d3b5 c6b5 a4b5 b7b6 e3e5 c8c7');
INSERT INTO evaluations VALUES(49084,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(49084,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(49084,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(49084,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(49084,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(49084,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(49084,36,'cp 152','d2c3 d8d1 e1d1 b8d7 f1b5 c7c6 h1e1 d7e5 e1e5 c8e6 b5d3 f8d6 e5e1 e8c8 d1e2 a7a5 a2a4 h8e8 c1e3 c8c7 e2f1 h7h6 h2h3 e8e7 a1b1 e6a2 b1d1 a2e6 b2b3 b7b6 h3h4 e6d7 e3d4 e7e1 d1e1 f7f6 d4e3 d6e5 c3c4 c6c5 d3g6 c7d6 e3d2 d7g4');
INSERT INTO evaluations VALUES(49083,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(49083,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(49083,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(49083,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(49083,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(49083,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(49083,36,'cp 139','d8d1 e1d1 b8d7 f1b5 c7c6 h1e1 d7e5 e1e5 c8e6 b5d3 f8d6 e5e1 e8c8 d1e2 c8c7 e2f1 h7h6 h2h3 a7a5 e1e2 h8e8 c1e3 f7f6 a2a3 e6f7 h3h4 d8d7 a1e1 d7e7 e3d2 b7b6 e2e7 d6e7 d2e3 e7d6 f1e2 d6c5 e2d2 c5e3 e1e3 c7d6 g2g3 e8e3 d2e3 d6e5');
INSERT INTO evaluations VALUES(21328,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(21328,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(21328,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(21328,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(21328,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(21328,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(21328,36,'cp 157','b1c3 g8f6 e4e5 f6d7 e5d6 f8d6 d4g7 d8e7 c4e2 d6e5 g7g4 d7f6 g4c4 c8f5 c1f4 e8c8 f4e5 e7e5 e1g1 f5e6 c4b5 e5b5 e2b5 c7c6 b5d3 a7a5 f1e1 f6d7 c3e4 c8c7 f2f3 b7b5 b2b3 c7b6 g1f2 d7e5 d3f1 e5g6 g2g3 e6f5 e1e2 f5e4 e2e4 d8d2 e4e2 h8d8 a1e1 a5a4 e2d2 d8d2 e1e2 d2d1 b3a4 b5a4');
INSERT INTO evaluations VALUES(21327,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(21327,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(21327,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(21327,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(21327,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(21327,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(21327,36,'cp 143','g8f6 e4e5 f6d7 e5d6 f8d6 d4g7 d8e7 c4e2 d6e5 g7g4 d7f6 g4c4 a7a6 c1e3 c8e6 c4a4 e6d7 a4b3 c7c6 e1c1 b7b5 e3d4 e5d4 d1d4 d7e6 b3a3 e7a3 b2a3 a8d8 d4h4 e8f8 e2f3 f8g7 a3a4 f6d5 c3e2 d5e7 e2d4 e6a2 d4c6 e7c6 f3c6 d8b8 a4b5 a6b5 h4g4 g7f6 g4f4 f6g7 h1d1 a2e6 c6d7 h8d8 d7e6 d8d1 c1d1');
INSERT INTO evaluations VALUES(14672,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(14672,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(14672,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(14672,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(14672,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(14672,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(14672,36,'cp 159','c1f4 f8e7 e1c1 e8g8 h2h4 f6g4 f2f3 e7f6 d4d2 f6c3 b2c3 g4f6 e4e5 c8e6 c4e6 d6e5 d2d8 a8d8 e6f7 g8f7 f4e5 c7c6 c3c4 d8d1 h1d1 f8e8 e5g3 a7a6 d1d4 g7g5 c1d2 g5g4 d2d3 g4f3 g2f3 e8g8 g3c7 g8c8 c7b6 f7e6 a2a4 f6d7');
INSERT INTO evaluations VALUES(14676,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(14676,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(14676,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(14676,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(14676,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(14676,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(14676,36,'cp 137','f8e7 h2h4 e8g8 e1c1 c8e6 c4e6 f7e6 e4e5 d6e5 d4e5 d8c8 e5e2 f8f7 f2f3 e7b4 c3e4 f6d5 f4h2 c8f8 c1b1 a8d8 c2c4 d5f4 h2f4 f7f4 d1d8 f8d8 h1d1 d8e8 e2e3 f4h4 e3a7 e8c6 a7d4 b4e7 d4d7 g8f7 d7c6 b7c6 a2a4 f7e8 b1c2 h4h5 c2c3 h5a5 d1a1 c6c5 g2g4 a5a6 a4a5');
INSERT INTO evaluations VALUES(42044,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42044,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42044,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42044,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42044,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42044,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(42044,36,'cp 5','e8g8 e2e3 b7b6 c4d5 e6d5 a1c1 b4e7 f1e2 c7c5 e1g1 c8b7 a2a3 a7a6 d4c5 b6c5 d1b3 d8c7 f1d1 b8d7 b3c2 f8e8 b2b4 c5c4 f3d4 e7d6 h2h3 d6h2 g1h1 h2d6 d4f5 d6f8 h1g1 h7h6 d2e1 a8d8 b4b5 d7e5 a3a4 f8c5 f5d4');
INSERT INTO evaluations VALUES(37045,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37045,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37045,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(37045,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37045,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37045,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37045,36,'cp 0','e2e3 b7b6 c4d5 e6d5 f1d3 f8e8 f3e5 c8a6 d3a6 b8a6 e1g1 c7c5 d1a4 a6c7 d4c5 b6c5 e5f3 a7a5 f1d1 b4c3 d2c3 d8d7 a4c2 f6e4 c3e5 c7e6 f3d2 e4d6 d2f3');
INSERT INTO evaluations VALUES(37041,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(37041,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37041,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37041,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37041,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(37041,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(37041,36,'cp 0','b7b6 a1c1 c8b7 c4d5 e6d5 f1d3 f8e8 e1g1 b4f8 f3e5 c7c5 f2f4 b8c6 c3d5 d8d5 d3c4 d5d8 e5f7 d8c7 f7g5 g8h8 g5f7');
INSERT INTO evaluations VALUES(35652,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35652,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35652,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35652,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35652,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35652,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(35652,36,'cp 4','a1c1 c8a6 d1b3 c7c5 f1e2 d5c4 e2c4 a6b7 a2a3 b7f3 g2f3 b4c3 d2c3 c5d4 c3d4 b8c6 d4f6 d8f6 b3c3 c6e5 f3f4 e5c4 c3f6 g7f6 h1g1 g8h8 c1c4 a8c8 c4c8 f8c8 e1d2 c8d8 d2c2 d8c8 c2b3 c8c5 g1d1 c5b5 b3c3 b5c5 c3d2 h8g7 d2e2 c5c2 d1d2 c2d2 e2d2 g7h6 d2e2 h6h5 b2b4');
INSERT INTO evaluations VALUES(52401,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(52401,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52401,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(52401,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(52401,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(52401,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(52401,36,'cp 20','g1f3 d7d5 c2c4 g8f6 d2d4 c8f5 f3h4 f5c8 b1d2 e7e6 h4f3 c6c5 c4d5 e6d5 b2b3 b8c6 c1b2 c5d4 f3d4 c6d4 b2d4 c8d7 f1d3 f8a3 d1f3 e8g8 e1g1 f8e8 d4f6 d8f6 f3f6 g7f6 a1b1 a8c8 f1d1');
INSERT INTO evaluations VALUES(52402,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52402,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52402,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52402,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(52402,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(52402,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(52402,36,'cp 20','g8f6 c2c4 d7d5 d2d4 c8f5 f1d3 f5g6 e1g1 e7e6 d1c2 d5c4 c2c4 b8d7 d3g6 h7g6 h2h3 f8d6 b1d2 e8g8 c4b3 b7b6 e3e4 d6e7 f1d1 a8c8 e4e5 f6d5 d2e4 c6c5 e4c3 c5c4 b3a4 d8c7 c1g5 f7f6 g5h4 c7c6 a4c2 f6e5 h4e7 d5e7 f3e5 d7e5 d4e5');
INSERT INTO evaluations VALUES(44096,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44096,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44096,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44096,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44096,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(44096,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44096,36,'cp 21','d2d4 e5d4 d1d4 d7d5 g1f3 b8a6 c1g5 f8e7 c4d5 a6b4 e1c1 b4d5 e2e4 d5c3 d4c3 d8b6 g5e3 b6b4 c3b4 e7b4 h2h3 b4a5 f1d3 a5b6 e3b6 a7b6 a2a3 e8e7 d3c2 h7h6 f3d4 c8e6 f2f4 f6h5 d4e6 f7e6 h1g1 h5f6 e4e5 f6d5 c1b1');
INSERT INTO evaluations VALUES(44105,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(44105,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44105,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44105,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(44105,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44105,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(44105,36,'cp 4','e5d4 d1d4 d7d5 g1f3 b8a6 c1g5 f8e7 c4d5 a6b4 e1c1 b4d5 e2e4 d5c3 d4c3 d8b6 g5e3 b6b4 c3b4 e7b4 f3e5 f6g4 e5g4 c8g4 d1d3 b4a5 d3a3 a5c7 f1g2 a7a5 f2f3 g4d7 f3f4 h7h5 b2b4 h5h4 e4e5 h4g3 h2g3 h8h1 g2h1 d7g4');
INSERT INTO evaluations VALUES(22731,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22731,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22731,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(22731,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(22731,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(22731,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(22731,36,'cp 0','g8f6 d1b3 g7g6 g1f3 a7a5 d2d4 d7d6 f1e2 f8g7 e1g1 e8g8 d4e5 d6e5 f1d1 d8e7 e3e4 c8g4 c1e3 c6d4 f3d4 e5d4 e3d4 g4e2 c3e2 f6e4 b3b7 a8b8 b7a7 b8a8 a7b7');
INSERT INTO evaluations VALUES(16446,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16446,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(16446,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(16446,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16446,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(16446,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(16446,36,'cp 8','d1b3 a7a5 a2a3 g7g6 d2d3 f8g7 g1f3 d7d6 b3c2 h7h6 b2b3 e8g8 c1b2 d6d5 c4d5 f6d5 f1e2 d5c3 c2c3 d8e7 c3c2 f8d8 e1g1 c8e6 f3d2 f7f5 a1c1 d8d7 e2f3 a8d8 f3e2');
INSERT INTO evaluations VALUES(36546,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(36546,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(36546,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36546,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(36546,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36546,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36546,36,'cp 26','c7c5 d4c5 d6c5 d1d8 f8d8 f3e5 f6e8 e5d3 e8d6 c1f4 d6c4 a1c1 b8c6 g2c6 b7c6 c3a4 c4b6 a4c5 b6d5 f4d2 a7a5 a2a3 c8h3 h1g1 h7h5 b2b3 d5b6 f2f3 g7d4 e2e3 d4g7 e1e2 a5a4 c5a4 b6a4 b3a4 a8a4 d2b4 g7h6 b4e7');
INSERT INTO evaluations VALUES(35944,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35944,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(35944,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(35944,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(35944,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(35944,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(35944,36,'cp 52','d4d5 e7e6 e1g1 e6d5 c4d5 f8e8 f1e1 b8d7 c1f4 f6g4 f4g5 f7f6 g5d2 g4e5 a2a4 a7a6 a1b1 f6f5 f3g5 d7f8 d1c1 c8d7 e1d1 b7b5 g5e6 f8e6 d5e6 d7e6 a4b5 a6b5 c3b5 d8b6 b5c3 a8c8 d2f4 e6b3 d1d2 e5f7 h2h4 b6a6 c3d5');
INSERT INTO evaluations VALUES(35981,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(35981,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(35981,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35981,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(35981,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(35981,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(35981,36,'cp 36','e7e6 e1g1 e6d5 c4d5 f8e8 f1e1 b8d7 c1f4 f6g4 f4g5 d8a5 a2a4 a5b6 d1c2 c5c4 c3d1 d7c5 c2c4 a7a5 h2h3 h7h6 g5f4 g4f6 f3d2 c8d7 f4e3 a8c8 e3d4 b6d8 d1e3 f6e4 g2e4 c5e4 c4d3 g7d4 d3d4 e4d2 d4d2 d7h3 e1c1 d8f6 c1c8 e8c8 a1d1 h3d7');
INSERT INTO evaluations VALUES(30671,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(30671,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(30671,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(30671,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(30671,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(30671,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(30671,36,'cp 88','f8b4 g1e2 d7d5 e4e5 d8d7 e1g1 c7c5 e2f4 g7g6 a2a3 b4c3 b2c3 c5c4 d3e2 b8c6 d1d2 e8c8 a3a4 c8b8 c1a3 b8a8 f1b1 g8e7 g2g3 e7f5 f4g2 h7h6 e2f3 d8b8 a3b4 b7a6 g3g4 f5e7 g2e3 h6h5 g4g5 e7f5 e3f5 g6f5 b4d6');
INSERT INTO evaluations VALUES(27952,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(27952,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(27952,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(27952,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(27952,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(27952,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(27952,36,'cp 100','g1e2 d7d5 e4e5 d8d7 e1g1 b4e7 c3b1 b7a6 d3a6 b8a6 c2c3 c7c5 e2g3 a8c8 h2h3 a6b8 b1d2 c5d4 c3d4 d7c6 d1f3 e7d8 f3g4 g7g6 d2f3 c6c2 f3e1 h7h5 g4f3 c2a4 c1e3 g8e7 b2b3 a4a6 e3g5 b8c6 f3d1 h5h4 g5f6 h4g3 f6h8');
INSERT INTO evaluations VALUES(27951,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(27951,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(27951,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(27951,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(27951,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(27951,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(27951,36,'cp 99','d7d5 e4e5 d8d7 e1g1 c7c5 e2f4 g7g6 a2a3 b4c3 b2c3 c5c4 d3e2 b8c6 d1d2 e8c8 a3a4 c8b8 c1a3 g8e7 f1b1 b8a8 g2g3 e7f5 a3b4 a7a5 g3g4 a5b4 g4f5 g6f5 c3b4 h8g8 g1f1 f7f6 e5f6 e6e5 d4e5 c6e5 d2d4 d7d6 b4b5 d6f6 a4a5');
INSERT INTO evaluations VALUES(29923,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(29923,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(29923,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(29923,33,'cp 246',NULL);
INSERT INTO evaluations VALUES(29923,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(29923,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(29923,36,'cp 244','f3b3 d6d5 b3b7 b8d7 c4d5 a8b8 b7a7 f8c5 a7a6 f6d5 e4d5 b8b6 a6e2 b6g6 e1g1 d8g5 e2f3 g5h4 g1h1 f7f5 d2d3 e8g8 b1d2 e5e4 d3e4 d7e5 f3e2 f5f4 d2f3 e5f3 e2f3 c5f2 h1h2 f2e1 f3e2 e1a5 f1f3 g6g3 e4e5 g3f3 e2f3 h4e1 b2b4 e1e5 a1b1 a5b6 c1d2 e5d4 d2c3 d4c4 a2a4');
INSERT INTO evaluations VALUES(43251,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(43251,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(43251,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(43251,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(43251,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(43251,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(43251,36,'cp 22','f1g2 c7c6 c2c4 d7d5 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 e7e6 e1g1 h7h6 h2h4 f8d8 f1c1 d8c8 g2f1 c6e5 f4e5 f6e4 c3e4 g7e5 c1c8 a8c8 e4c3 e5d6 a1c1 e6e5 f1e2 e5d4 e3d4 f5e6');
INSERT INTO evaluations VALUES(50313,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(50313,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(50313,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(50313,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(50313,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(50313,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(50313,36,'cp -18','a2a3 d7d5 g1f3 b8d7 c1b2 e7e6 e2e3 d7b6 f1e2 f8d6 e1g1 a7a5 b4b5 e8g8 a3a4 c7c6 b5c6 b7c6 d2d3 d8e7 b1d2 c8b7 c2c4 d5c4 d3c4 d6b4 d1c2 f8d8 f1d1 c6c5 f3e5 b6d7 e5d7 d8d7 b2c3 b4c3 c2c3');
INSERT INTO evaluations VALUES(52334,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52334,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52334,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(52334,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(52334,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52334,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52334,36,'cp 39','g8f6 b1c3 b8c6 d4a4 d7d6 e4e5 d6e5 f3e5 d8d4 e5c6 d4a4 c3a4 b7c6 b2b3 f6d5 c1a3 c8f5 e1c1 h7h5 f1e2 e7e5 a3f8 h8f8 h1e1 f7f6 g2g3 e8c8 a2a3 h5h4 b3b4 c8c7 c2c4 d5b6 a4c5 d8d1 e1d1 b6d7 c5d7 f5d7 c4c5 f8h8 g3g4');
INSERT INTO evaluations VALUES(44447,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(44447,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(44447,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(44447,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44447,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44447,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(44447,36,'cp 24','d4a4 b8c6 b1c3 d7d6 e4e5 d6e5 f3e5 d8d4 e5c6 d4a4 c3a4 b7c6 b2b3 f6d5 c1a3 c8f5 e1c1 e8c8 f1d3 f5d3 d1d3 e7e5 a3f8 h8f8 h1e1 f8e8 a4c5 f7f5 f2f3 d5f6 d3d8 c8d8 c2c4 f6d7 e1d1 e8e7 b3b4 d8c7 c5d7 e7d7 d1d7 c7d7 c1d2 c6c5 a2a3');
INSERT INTO evaluations VALUES(50519,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(50519,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(50519,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(50519,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(50519,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(50519,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(50519,36,'cp 100','d4d5 g8f6 b1c3 d7d6 a2a4 e6d5 e4d5 f8e7 f1e2 e8g8 e1g1 f8e8 c1f4 f6h5 f4e3 h5f6 f3d2 b8d7 e3f4 e7f8 f1e1 d7b6 e2f3 e8e1 d1e1 c8f5 e1d1 h7h6 h2h3 f5g6 g2g4 b6d7 d2c4 d8c7 f3e2 d7b6 c4e3 g6h7 a4a5 b6d7 e3c4');
INSERT INTO evaluations VALUES(50547,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(50547,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(50547,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(50547,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(50547,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(50547,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(50547,36,'cp 85','g8f6 b1c3 d7d6 a2a4 f8e7 f1e2 e8g8 e1g1 f8e8 c1f4 e6d5 e4d5 f6h5 f4e3 h5f6 f3d2 b8d7 e3f4 d7e5 f1e1 b7b6 f4e5 d6e5 e2f3 e7d6 d2c4 a8b8 h2h3 g7g6 g2g4 h7h6 d1d3 d8c7 b2b3 g8g7 a1d1 c8b7 c3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(56153,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(56153,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(56153,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(56153,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(56153,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(56153,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(56153,36,'cp -101','g8f6 b1d2 b7b5 a2a4 c7c6 f1e2 f8c5 d5c6 d8d4 g1h3 c8h3 e1g1 b5b4 g2h3 b8c6 d2c4 a8d8 d1d4 c6d4 e2d1 f6e4 c4e5 e8g8 e5d3 a7a5 c1e3 f8e8 f1e1 c5f8 f2f3 e4c5 d3c5 f8c5 g1f1 d4f3 d1f3 c5e3 a1d1 d8d1 e1d1 e3f4 d1d7 f4h2 f1g2');
INSERT INTO evaluations VALUES(48596,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(48596,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(48596,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(48596,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(48596,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(48596,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(48596,36,'cp -106','f1c4 f6e4 g1f3 f8b4 b1d2 e8g8 e1g1 e4d6 a2a3 b4d2 f3d2 f8e8 f1e1 b8d7 b2b3 d7b6 c4f1 d6f5 d2c4 d8d5 d1d5 b6d5 e1e5 c8e6 c1b2 c7c6 c4a5 e8e7 e5e1 e7d7 g2g3 a8d8 a1d1 d5b6 d1d7 d8d7 f1h3 f5d4');
INSERT INTO evaluations VALUES(15120,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(15120,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(15120,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(15120,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(15120,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(15120,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(15120,36,'cp 140','f3e5 f6e4 d1h5 c6e5 h5e5 c8e6 c3e4 d5e4 e5e4 f8e7 c1e3 e6d5 e4g4 e7f6 c2c4 d5e6 g4f4 c7c6 f1e2 g7g5 f4e4 d8a5 e3d2 a5f5 e4f5 e6f5 d2e3 e8c8 e1c1 h8e8 d1d2 g5g4 h1d1 h7h5 g2g3 h5h4 d4d5 c6d5 d2d5 f5e6 d5c5 c8b8 c5h5 d8d1 e2d1 h4g3');
INSERT INTO evaluations VALUES(22846,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(22846,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(22846,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(22846,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(22846,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(22846,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(22846,36,'cp -19','e5e4 f3d4 c6d4 b2d4 d7d5 e2e3 c8f5 d2d3 g8f6 b1d2 f8e7 f1e2 d8d7 h2h3 e8c8 a2a4 c8b8 d3e4 f6e4 d2e4 f5e4 e2f3 d7e6 d1e2 h8e8 e1c1 e7a3 c1b1 g7g6 f3e4 e6e4 e2g4 b7b6 g4e4 e8e4 b1a1 b8c8 d4b2 a3b2 a1b2 c7c6');
INSERT INTO evaluations VALUES(22970,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(22970,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(22970,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(22970,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(22970,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(22970,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(22970,36,'cp -30','f3d4 c6d4 b2d4 d7d5 e2e3 h7h5 d2d3 c8f5 d3e4 d5e4 d1d2 a7a6 d2c3 g8f6 b1d2 f8e7 c3b2 d8d7 f1e2 c7c5 d4e5 a8d8 e1c1 d7c6 f2f3 e8g8 c1b1 b7b5 d2e4 f5e4 f3e4 f8e8 e2f3 c6e6 d1d8 e7d8 h1d1 h5h4');
INSERT INTO evaluations VALUES(22914,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(22914,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(22914,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(22914,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(22914,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(22914,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(22914,36,'cp -37','c6d4 b2d4 d7d5 e2e3 c8f5 d2d3 h7h5 d3e4 d5e4 d1d2 a7a6 d2c3 g8f6 b1d2 f8e7 c3b2 d8d7 h2h3 c7c5 d4e5 a8d8 a1d1 e8g8 f1e2 d7e6 a2a4 d8d5 e5g3 d5d7 e1g1 f6d5 d2c4 h5h4 g3e5 b7b5 a4b5 a6b5 c4a3 e6g6 g1h1 b5b4 a3b5 f8c8 e2c4');
INSERT INTO evaluations VALUES(23149,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(23149,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(23149,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(23149,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(23149,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(23149,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(23149,36,'cp -19','b2d4 d7d5 e2e3 c8f5 d2d3 h7h5 d3e4 d5e4 d1d2 c7c6 d2c3 g8f6 b1d2 f8e7 f1e2 e8g8 h2h3 h5h4 e1g1 f8e8 c3b2 d8d6 a1d1 f6d5 d2c4 d6e6 a2a4 a8d8 d1d2 a7a6 a4a5 e6h6 g1h1 f5e6 f1d1 h6g6');
INSERT INTO evaluations VALUES(23144,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(23144,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(23144,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(23144,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(23144,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(23144,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(23144,36,'cp -22','d7d5 e2e3 c8f5 d2d3 h7h5 d3e4 d5e4 d1d2 c7c6 d2c3 g8f6 b1d2 f8d6 d2c4 d6e7 c3b2 e8g8 a2a4 d8c8 f1e2 f8d8 h2h3 h5h4 e1g1 c8e6 a1d1 f6d5 b2c1 f7f6 d4b2 e7b4 d1d4 b7b5 a4b5 c6b5 c2c3 b4e7');
INSERT INTO evaluations VALUES(22175,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(22175,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(22175,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(22175,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(22175,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(22175,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(22175,36,'cp -15','e2e3 h7h5 d2d3 c8f5 d3e4 d5e4 d1d2 c7c6 d2c3 g8f6 b1d2 f8e7 f1e2 e8g8 c3b2 d8c7 h2h3 a8d8 d4e5 c7a5 e5c3 e7b4 c3b4 a5b4 e1c1 h5h4 d2b1 b4c5 b2a3 c5a3 b1a3 f6d5 a3b1 b7b5 c2c4 d5b4 b1c3 b4d3 e2d3 e4d3');
INSERT INTO evaluations VALUES(22174,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(22174,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(22174,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(22174,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(22174,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(22174,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(22174,36,'cp -41','h7h5 d2d3 c8f5 d3e4 d5e4 d1d2 a7a6 d2c3 g8f6 b1d2 f8e7 f1e2 f6d5 c3b2 f7f6 c2c4 d5b4 e1g1 h5h4 c4c5 h4h3 g2g3 b4d3 b2c3 d8d5 b3b4 a6a5 a2a3 e8g8 f2f3 a5b4 a3b4 a8a1 f1a1 f8d8 f3e4 f5e4 d2e4 d5e4');
INSERT INTO evaluations VALUES(5571,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(5571,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(5571,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(5571,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(5571,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(5571,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(5571,36,'cp -26','b1c3 a7a6 f1d3 f7f6 d1e2 d8c7 c1f4 g8e7 g2g3 f6f5 c3b1 e7g6 b1d2 c6d4 f3d4 c5d4 e1c1 d4e5 f4e5 c7e5 e2f1 e8g8 f2f4 e5c7 d2f3 e6e5 f3e5 g6e5 f4e5 c7e5 f1f3 c8e6 h1e1 e5f6 f3f4 a8e8 h2h4 g7g6 e1e5 e6f7 d1e1 e8e5');
INSERT INTO evaluations VALUES(5566,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(5566,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(5566,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(5566,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(5566,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(5566,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(5566,36,'cp -41','g8e7 c3a4 b7b6 c1f4 h7h6 f1d3 e8g8 e1g1 c6b4 d1d2 c8a6 f1c1 a8c8 a2a3 b4d3 c2d3 e7g6 a4c5 b6c5 d3d4 g6f4 d2f4 c5c4 h2h4 c8b8 a1b1 b8b3 f3d2 b3d3 c1c3 d8b6 c3d3 c4d3 d2f3 b6b3 h4h5');
INSERT INTO evaluations VALUES(8220,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(8220,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(8220,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(8220,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(8220,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(8220,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(8220,36,'cp -36','c3a4 c5b4 c2c3 b4a5 f1d3 a5c7 d1e2 b7b6 c1f4 f7f6 e5f6 g7f6 f4c7 d8c7 f3d4 c7e5 e1g1 e5e2 d3e2 c6d4 c3d4 e7f5 a1d1 c8d7 a4c3 e8e7 e2g4 f5h4 g2g3 h7h5 c3d5 e6d5 g4d7 e7d7 g3h4 h8g8 g1h1 a8c8 d1c1 c8c4 c1c4 d5c4');
INSERT INTO evaluations VALUES(31853,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31853,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(31853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31853,36,'cp 0','e7e6 g1f3 g8f6 d3f5 e6f5 d1d3 d8c8 b2b3 b8a6 e1g1 c7c6 c2c4 c8e6 d3e2 a6c7 f3g5 e6c8 g5f3');
INSERT INTO evaluations VALUES(32315,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32315,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32315,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32315,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32315,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32315,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32315,36,'cp 0','d3f5 e6f5 g1f3 g8f6 d1d3 d8c8 b2b3 b8a6 e1g1 c7c6 c2c4 c8e6 d3e2 a6c7 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(32272,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(32272,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(32272,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(32272,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(32272,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(32272,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32272,36,'cp 8','e6f5 g1f3 c7c6 c2c4 d5c4 d1c2 d8c8 e1g1 g8f6 c1d2 c8e6 f3g5 e6d7 a2a4 f8d6 c2c4 e8g8 b1a3 b8a6 c4b3 h7h6 g5f3 f8b8 a3c4 d7e6 b3c2 d6c7 f1c1 b8e8 a4a5 f6e4 d2e1 a8d8 c4e5 e6d5');
INSERT INTO evaluations VALUES(56416,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(56416,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(56416,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(56416,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(56416,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(56416,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(56416,36,'cp -95','d2d4 e5d4 d1d4 c8e6 g1f3 g8f6 b1c3 f6g4 d4a4 c7c6 e2e4 b7b5 a4d4 f8e7 e1g1 e8g8 e4d5 c6d5 d4f4 b8c6 c3b5 a8c8 c2c3 d8d7 f4g3 g4f6 f3d4 f6e4 g2e4 d5e4 c1h6 e7f6 d4c6 c8c6 a1d1 d7b7 a2a4 g8h8 h6f4 a7a6 b5d6 b7b2 d6e4 f6c3');
INSERT INTO evaluations VALUES(56360,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(56360,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(56360,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(56360,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(56360,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(56360,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(56360,36,'cp -100','e5d4 d1d4 c8e6 g1f3 g8f6 f3e5 f6d7 e5d7 d8d7 b1c3 b8c6 d4a4 c6b4 a4d7 e8d7 e1d1 a8d8 a2a3 b4c6 c1f4 c6e7 g2f3 e7g6 f4g3 f7f6 h2h3 g6e5 f3g2 h7h5 g4h5 h8h5 e2e3 c7c6 c3e2 g7g5 h3h4 f8d6 h4g5 h5h1 g2h1 d8h8 h1g2 f6g5 b2b3 e6g4 d1d2 g4f3 g2f3 e5f3 d2c3 d6g3 f2g3');
INSERT INTO evaluations VALUES(14287,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14287,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14287,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14287,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14287,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(14287,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(14287,36,'cp -4','c6b4 b1c3 c7c6 e1g1 b4d3 c2d3 c8g4 h2h3 g4h5 d1b3 d8b6 f3e5 e7e6 c3a4 b6b3 a2b3 e8c8 a4c5 f8c5 d4c5 a7a6 c1d2 f6d7 e5d7 d8d7 d3d4 f7f6 f2f3 e6e5 d4e5 f6e5 b3b4 h8e8 a1e1 h5g6 e1d1 g6c2 d1c1 c2b3 d2c3 h7h5 f1e1 b3c4 g1f2 g7g5 c1d1');
INSERT INTO evaluations VALUES(11939,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(11939,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(11939,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(11939,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(11939,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(11939,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11939,36,'cp 7','f8b4 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 f8e8 c1g5 b4d6 f1e1 c8e6 a1d1 h7h6 g5h4 a7a5 c3e2 d6e5 e2f4 d8d6 h4g3 e6d7 c2c3 d6b8 e1e2 a5a4 d1e1 e5d6 f4h5 f6h5 e2e8 d7e8 g3d6 b8d6 f3h5 a4a3 b2b4 d6f6');
INSERT INTO evaluations VALUES(39201,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(39201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39201,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(39201,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(39201,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(39201,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(39201,36,'cp 1','g8f6 g1f3 d7d6 c2c3 e8g8 e1g1 h7h6 h2h3 c5b6 f1e1 c7c6 b1d2 b8d7 c4b3 b6c7 d3d4 f8e8 b3c2 d7f8 d2f1 d6d5 f3e5 f6e4 c1f4 c8f5 f1e3 f5h7 f2f3 e4d6 f4h2 h7c2 d1c2 a7a5 a2a4 f8e6 f3f4 f7f6');
INSERT INTO evaluations VALUES(57725,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(57725,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(57725,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(57725,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(57725,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(57725,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(57725,36,'cp 87','d2d4 e7e6 f1d3 d7d5 b1c3 b8c6 g1f3 c6b4 d3b5 c7c6 b5e2 g8f6 e4e5 f6e4 a2a3 e4c3 b2c3 b4a6 c3c4 a6c7 c4d5 e6d5 e2d3 f8e7 c2c3 c8g4 h2h3 g4h5 d1c2 e8g8 f3h2 c6c5 e1g1 c5d4 c3d4 c7e6 c1e3 a8c8 c2b3 d8b6 b3b6 a7b6 a3a4 h5g6 d3g6 f7g6');
INSERT INTO evaluations VALUES(38795,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(38795,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(38795,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(38795,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38795,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38795,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(38795,36,'cp -14','d5e4 d1g4 g8f6 g4g7 h8g8 g7h6 d8d4 e1c1 b8d7 f2f3 a7a6 d2e1 d4e5 c3e4 b4e1 d1e1 e5a5 e4f6 d7f6 e1d1 g8g6 h6d2 a5d2 d1d2 b7b6 g2g3 c8b7 f1g2 a8d8 d2e2 h7h5 f3f4 b7g2 e2g2 d8d4 g2e2 d4e4 c1d2 g6g8 g1f3 e4e2 d2e2 e8e7');
INSERT INTO evaluations VALUES(38863,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(38863,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(38863,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(38863,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(38863,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(38863,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(38863,36,'cp 0','d1g4 g8f6 g4g7 h8g8 g7h6 d8d4 e1c1 b8d7 f2f3 a7a6 d2e1 d4e5 c3e4 b4e1 d1e1 e5a5 e4f6 d7f6 e1d1 g8g6 h6d2 a5d2 c1d2 e8e7 d2e2 c8d7 e2f2 g6g5 f1d3 g5a5 a2a3 d7b5 g1e2 b5d3 d1d3 a5b5 h1b1 b5h5 b1h1');
INSERT INTO evaluations VALUES(57639,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(57639,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(57639,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(57639,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(57639,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(57639,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(57639,36,'cp -39','c7c5 g1f3 f8g7 c2c3 d7d6 f1b5 c8d7 b5d7 d8d7 e5d6 d7d6 e1g1 g8f6 b1a3 e8g8 f1e1 e7e6 d1e2 b8c6 d2d4 c5d4 a3b5 d6c5 b5d4 c6d4 f3d4 f6d5 d4f3 a8c8 a2a4 f8e8 g2g3 a7a6 c1e3 d5e3 e2e3 c5c4 e3a7 c4c7 a4a5 c7c6 f3e5');
INSERT INTO evaluations VALUES(52264,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(52264,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52264,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(52264,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(52264,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(52264,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(52264,36,'cp -33','c2c3 f8g7 g1f3 d7d6 f1b5 c8d7 b5d7 d8d7 e5d6 d7d6 e1g1 g8f6 b1a3 e8g8 f1e1 e7e6 d2d4 c5d4 a3c4 d6c7 d1d4 f8d8 d4f4 c7f4 c1f4 b8d7 a1d1 f6d5 f4g5 f7f6 g5c1 d7c5 f3d4 g7f8 d4b3 b7b6 b3c5 f8c5 c4e3 d5e3 c1e3 c5e3');
INSERT INTO evaluations VALUES(29650,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(29650,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(29650,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(29650,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(29650,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(29650,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(29650,36,'cp 59','f3e5 g4e6 e2e3 b8d7 b1d2 g7g6 c2c3 f8g7 d1b3 d7e5 f4e5 d8c8 a2a4 e8g8 b3a3 f8e8 f1e2 c7c6 h2h3 a7a5 b2b3 f6e4 e5g7 g8g7 d2e4 d5e4 c3c4 f7f6 a3b2 e6f7 b2c2 c8e6 e1g1 e8c8 f1d1 b7b6 c4c5 b6c5 c2c5');
INSERT INTO evaluations VALUES(29698,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(29698,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29698,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(29698,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(29698,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(29698,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(29698,36,'cp 42','g4f5 e2e3 e7e6 b1d2 h7h5 c2c4 b8d7 h2h3 f8b4 a2a3 d7e5 f4e5 b4d2 d1d2 e8g8 f1d3 f5d3 d2d3 d5c4 d3c4 d8d5 c4d5 f6d5 e3e4 f7f6 e5h2 d5e7 h2c7 f8c8 c7a5 f6f5 e4f5 e7f5 a1d1 c8c4 e1g1 c4d4 d1d4 f5d4 f1d1 d4c6 a5c3 a8d8');
INSERT INTO evaluations VALUES(38990,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(38990,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(38990,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(38990,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(38990,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(38990,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(38990,36,'cp 65','f2f4 g8e7 b1d2 d7d6 c2c3 d6e5 f4e5 c7c5 d2e4 c5d4 e4d6 e8f8 c3d4 b8c6 g1f3 e7f5 d6f5 g6f5 c1e3 h7h6 f1c4 c6e7 e1g1 f8g8 d1e1 a7a6 e1g3 b7b5 c4d3 c8b7 g3h3 b7d5 f1f2 d5c4 d3c4 b5c4 h3h5');
INSERT INTO evaluations VALUES(38986,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(38986,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38986,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(38986,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(38986,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(38986,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(38986,36,'cp 48','g8e7 b1c3 b7b6 g1f3 c8b7 d1e2 b8c6 c1d2 e7f5 e2f2 c6e7 f1e2 c7c5 d4c5 b6c5 f2c5 h7h5 c5f2 a8b8 h2h3 d7d5 e2d3 d5d4 c3b5 b7f3 f2f3 a7a6 g2g4 a6b5 g4f5 g6f5 d2b4 e7d5 b4d6 b8b6 e1c1 b6d6 e5d6 d8d6');
INSERT INTO evaluations VALUES(43056,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(43056,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(43056,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(43056,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(43056,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(43056,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(43056,36,'cp 48','g1f3 d7d6 b1c3 d6e5 f4e5 b8d7 c3e4 e8g8 h2h4 c7c5 c1g5 c5d4 d1d4 e7f5 d4c3 f7f6 e5f6 d7f6 e4f6 g7f6 g5f6 d8f6 c3f6 f8f6 e1c1 b7b6 h1h3 c8b7 f3e5 f5e7 g2g4 a8c8 h4h5 g6h5 g4h5 f6f5 e5c4 f5f2 f1d3 e7f5 d1e1 b7d5 c4e5 d5a2 h5h6 g8f8 e5g4 f2g2 g4f6 f5d4 f6h7 f8e7');
INSERT INTO evaluations VALUES(9690,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(9690,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(9690,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(9690,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(9690,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(9690,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(9690,36,'cp -87','d2d3 c6a5 c4b5 e8g8 c1g5 c7c6 b5a4 b7b5 a4c2 d7d5 b1d2 h7h6 g5f6 g7f6 f3g3 f6h4 g3e3 d5d4 e3e2 c6c5 g1f3 h4f6 c3d4 e5d4 e1g1 c8e6 e4e5 f6g7 d2e4 d8e7 f1e1 a8d8 e2d2 a5c6 e4f6 g7f6 e5f6 e7f6 d2h6 e6d5 f3g5');
INSERT INTO evaluations VALUES(9692,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(9692,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(9692,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(9692,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(9692,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(9692,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(9692,36,'cp -89','c6a5 c4b5 c7c6 b5a4 d7d5 c1g5 e8g8 a4c2 b7b6 b1d2 h7h6 g5f6 g7f6 f3g3 f6g5 g1f3 g5f4 g3h4 d8h4 f3h4 f4d2 e1d2 d5e4 b2b4 e4d3 c2d3 f8d8 d2c2 a5b7 h4f3 c6c5 f3e5 c5b4 c3b4 a7a5 a2a3 c8e6 h1c1 b7d6 e5c6 d8d7 c2b2 a5b4 c6b4 g8g7');
INSERT INTO evaluations VALUES(14167,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(14167,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14167,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(14167,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14167,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(14167,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(14167,36,'cp 44','e7e6 e1g1 c8d7 b2b3 f8d6 c1b2 e8g8 b5e2 c6e7 c2c4 c7c5 d4c5 d6c5 b1d2 e7g6 a1c1 a8c8 a2a3 a7a5 c4d5 e6d5 d2b1 d7e6 b1c3 d8e7 c3b5 f8d8 a3a4 e6d7 b5d4 f6e4 c1c2 c5b4 c2c8 d8c8 e2b5 e4c3 b2c3 b4c3');
INSERT INTO evaluations VALUES(14731,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(14731,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(14731,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(14731,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(14731,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(14731,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(14731,36,'cp 55','e1g1 f8e7 b2b3 e8g8 c1b2 a7a6 b5c6 b7c6 f3e5 c8b7 e5d3 a6a5 b1d2 a5a4 a1c1 a4b3 a2b3 f6d7 c2c4 e7d6 d2f3 f8e8 c4c5 d6e7 f1e1 b7a6 d3f4 a6b5 d1c2 a8a2 c1a1 a2a1');
INSERT INTO evaluations VALUES(14950,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(14950,31,'cp 263',NULL);
INSERT INTO evaluations VALUES(14950,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(14950,33,'cp 259',NULL);
INSERT INTO evaluations VALUES(14950,34,'cp 268',NULL);
INSERT INTO evaluations VALUES(14950,35,'cp 260',NULL);
INSERT INTO evaluations VALUES(14950,36,'cp 260','f3e5 c8d7 e5d7 d8d7 e4e5 f6e4 c3e2 a7a6 b5c6 d7c6 e1g1 e4c5 d2d4 c5e6 c2c3 h7h5 d1d3 g7g6 a2a4 f8e7 h2h3 h5h4 d3f3 g6g5 c1e3 h8f8 f3h5 e8c8 a1e1 f7f6 f2f4 f6e5 f4g5 e5e4 f1f8 d8f8 h5h4 c6a4');
INSERT INTO evaluations VALUES(14943,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(14943,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(14943,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(14943,33,'cp 245',NULL);
INSERT INTO evaluations VALUES(14943,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(14943,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(14943,36,'cp 256','c8d7 e5d7 d8d7 e4e5 f6e4 c3e2 a7a6 b5c6 d7c6 e1g1 e4c5 d2d4 c5e6 c2c3 h7h5 e2f4 h5h4 h2h3 f8e7 f4d3 e8c8 f1e1 h8g8 f2f4 g7g6 c1e3 c8b8 a2a4 b8a7 a4a5 d8f8 a1c1 f7f5 e5f6 e7f6 d1g4');
INSERT INTO evaluations VALUES(22459,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(22459,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(22459,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(22459,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(22459,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(22459,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(22459,36,'cp 81','d2d4 e5d4 f3d4 f8g7 d4c6 b7c6 b1c3 d7d6 e1g1 c8e6 c4b3 a7a5 c1e3 g8f6 e4e5 d6e5 b3e6 d8d1 e6f7 e8f7 a1d1 a5a4 f1e1 h8b8 e3c1 b8b4 e1e5 b4c4 d1d3 f6d5 e5e4 d5c3 e4c4 c3e2 g1f1 e2c1 d3d7 f7g8 a2a3 g7b2 c4c6 b2a3 c6c7 a3b4 d7g7 g8h8 g7h7 h8g8');
INSERT INTO evaluations VALUES(22464,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(22464,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(22464,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(22464,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(22464,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(22464,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(22464,36,'cp 84','e5d4 f3d4 d7d6 b1c3 f8g7 d4c6 b7c6 e1g1 c8e6 c4b3 a7a5 c1e3 g8f6 e4e5 d6e5 b3e6 d8d1 e6f7 e8f7 f1d1 a5a4 a1c1 g7f8 e3g5 f8b4 g5f6 f7f6 c3e4 f6f5 f2f3 h8d8 d1d8 a8d8 g1f2 h7h5 c1a1 d8b8 b2b3 b4a3 a1d1 a4b3 a2b3 a3e7 d1d7');
INSERT INTO evaluations VALUES(22482,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22482,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(22482,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(22482,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(22482,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(22482,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(22482,36,'cp 71','f3d4 f8g7 d4c6 b7c6 e1g1 d7d6 b1c3 c8e6 c4b3 a7a5 h2h3 g8e7 c1e3 e8g8 e3d4 e7c8 d4g7 g8g7 f2f4 c8b6 d1d3 a5a4 b3e6 f7e6 f1f2 d6d5 e4d5 e6d5 f4f5 d8d6 a1f1 g7g8 c3e2 a8e8 b2b3 a4b3 a2b3 b6d7 d3d2 f8f5 f2f5 g6f5 f1f5');
INSERT INTO evaluations VALUES(22481,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(22481,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(22481,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(22481,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(22481,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(22481,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(22481,36,'cp 76','f8g7 d4c6 b7c6 e1g1 d7d6 b1c3 c8e6 c4b3 a7a5 c1e3 g8f6 e4e5 d6e5 b3e6 d8d1 e6f7 e8f7 f1d1 a5a4 a1c1 g7f8 e3g5 f8e7 d1d3 h8b8 b2b3 a4b3 a2b3 b8d8 d3h3 h7h5 c1e1 d8d4 h3e3 f6d5 c3d5 d4d5 g5e7 f7e7 e3e5 d5e5 e1e5 e7f6 e5e1');
INSERT INTO evaluations VALUES(14378,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(14378,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(14378,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14378,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(14378,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14378,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(14378,36,'cp 50','g1f3 e7e5 d4d5 e5e4 d5c6 d8d1 e1d1 e4f3 g2f3 b7c6 f1c4 f6d7 d1c2 d7b6 c4d3 c8e6 h2h4 b6d5 a2a3 d5c3 b2c3 h7h5 c1b2 g7g6 c3c4 h8g8 a3a4 f8e7 b2e5 e8c8 a1b1 e6f5 e3e4 f5e6 c4c5 e7c5 d3a6 c8d7 b1b7 c5d6');
INSERT INTO evaluations VALUES(12475,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(12475,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12475,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(12475,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(12475,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12475,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12475,36,'cp 0','f8b4 f1d3 d7d5 e4d5 f6d5 d4c6 d5e3 f2e3 b7c6 d1g4 b4c3 b2c3 e8g8 e1g1 a8b8 g4e4 g7g6 e4c6 d8a5 a1b1 b8b1 f1b1 a5e5 b1e1 e5a5');
INSERT INTO evaluations VALUES(7831,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7831,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7831,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7831,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7831,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7831,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7831,36,'cp 0','f1d3 d7d5 e4d5 f6d5 d4c6 d5e3 f2e3 b7c6 d1g4 b4c3 b2c3 e8g8 e1g1 a8b8 g4e4 g7g6 e4c6 d8g5 c6f3 c8b7 f3f2 b7c6 h2h4 g5g4 d3e2 g4e4 e2d3');
INSERT INTO evaluations VALUES(7837,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7837,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7837,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7837,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7837,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7837,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7837,36,'cp 0','e6e5 d4c6 b7c6 e1g1 e8g8 a2a3 b4c3 b2c3 d7d5 f2f3 f8e8 d1d2 a7a5 a3a4 c8a6 d2e2 d8c8 e2f2 h7h6 a1b1 c8c7 f2e2 a6c8 e2d2 c8a6');
INSERT INTO evaluations VALUES(17255,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(17255,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(17255,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(17255,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(17255,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(17255,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(17255,36,'cp 12','d7d5 c1f4 c8f5 e2e3 e7e6 f1b5 f8d6 f3e5 e8g8 b5c6 b7c6 e5c6 d8e8 c6e5 a8b8 a1b1 d6b4 e1g1 b4c3 b2c3 b8b1 d1b1 f6e4 b1b4 f7f6 e5d3 e8c6 f1e1 e4c3 d3c5 c3e4 c5e4 f5e4 f2f3 e4g6 e1c1 f8a8 c2c4 g6d3 b4b5 c6b5 c4b5 d3b5 f4c7 b5c4 c7d6 c4a2 c1c7');
INSERT INTO evaluations VALUES(55938,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(55938,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55938,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(55938,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55938,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(55938,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55938,36,'cp 11','g1f3 g8f6 e2e3 f8d6 d2d4 e8g8 f1d3 b7b6 c2c4 c8b7 e1g1 b8d7 b1d2 c7c5 c4d5 e6d5 f3e5 d8e7 e5d7 f6d7 a1c1 a7a5 d3b5 f8d8 d1e2 d7f8 a2a4 f8e6 b5d3 e6c7 f2f4 b7a6 d3a6 c7a6 d2f3');
INSERT INTO evaluations VALUES(55935,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(55935,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(55935,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(55935,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55935,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55935,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(55935,36,'cp 4','g8f6 e2e3 f8e7 d2d4 e8g8 f1d3 c7c5 e1g1 b8d7 b1d2 b7b6 d1e2 c8b7 a1d1 a8c8 f3e5 c8c7 f2f4 c5d4 e3d4 f6e8 g2g4 e8d6 b2a3 d6c8 a3b2');
INSERT INTO evaluations VALUES(7772,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(7772,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(7772,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(7772,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(7772,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(7772,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(7772,36,'cp 46','c4d5 e6d5 f1d3 f6e4 g5f4 d7f6 e1g1 e8g8 h2h3 c8f5 f3e5 e4c3 b2c3 f5d3 d1d3 d8c8 c3c4 c8e6 a1c1 e7d6 c4c5 d6e5 f4e5 b7b5 f2f3 f8e8 f1e1 a7a5 d3c2 a8c8 e3e4 c8d8 c2f2 d5e4 f3e4 b5b4');
INSERT INTO evaluations VALUES(7769,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(7769,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(7769,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(7769,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(7769,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(7769,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(7769,36,'cp 34','e6d5 f1d3 f6e4 g5f4 d7f6 e1g1 e8g8 h2h3 c8f5 f3e5 e4c3 b2c3 f5d3 d1d3 f6d7 a2a4 d7e5 f4e5 d8d7 a1b1 b7b6 c3c4 d5c4 d3c4 f8d8 f1d1 a8c8 e5h2 h7h6 h2e5 e7f8 e3e4 d7e6 c4e6 f7e6 a4a5 b6a5');
INSERT INTO evaluations VALUES(7793,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(7793,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(7793,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(7793,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(7793,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(7793,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(7793,36,'cp 30','d1c2 h7h6 g5h4 e8g8 f1d3 f8e8 e1g1 f6e4 d3e4 e7h4 e4h7 g8h8 f3h4 d8h4 h7d3 h4e7 f1b1 d7b6 b1e1 c8d7 a1c1 a8d8 c3e2 b6c8 e2g3 h8g8 f2f3 d7e6 c2f2 c8d6 a2a4 e7f6 a4a5 a7a6 e3e4 d5e4 f3e4 f6f2 g1f2');
INSERT INTO evaluations VALUES(12458,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12458,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12458,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12458,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12458,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12458,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(12458,36,'cp 0','d8c7 e1g1 f8e7 f1e1 d7d6 c3b5 c7b8 c4d3 e8g8 c2c4 a7a6 d4c6 b7c6 b5c3 f6d7 e1e3 f8d8 e3g3 d7f8 d1h5 g7g6 h5h6 e6e5 a1b1 b8c7 c1e3 f8e6 g3h3 e6f8 h3f3 f8e6');
INSERT INTO evaluations VALUES(44672,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(44672,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(44672,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(44672,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44672,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44672,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(44672,36,'cp 0','g1f3 b8c6 c3d4 d7d6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5c6 d7c6 f1e1 f8e7 d4d5 b6d5 c3d5 c6d5 d1d5 e6d5 c1f4 a8d8 a1c1 d8d7 c1c8 d7d8');
INSERT INTO evaluations VALUES(44671,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44671,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(44671,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44671,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44671,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44671,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44671,36,'cp 0','b8c6 c3d4 d7d6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5c6 d7c6 c1e3 b6d5 a1c1 f8e7 d1f3 d5c3 f3c6 b7c6 c1c3 e8d7 b2b3 e7a3 c3c4 h8b8 c4a4 a3d6 e3d2 b8b5 f1c1 a6a5 g1f1 b5d5 a4c4 a8c8 c4a4');
INSERT INTO evaluations VALUES(12922,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12922,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(12922,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(12922,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(12922,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(12922,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(12922,36,'cp 1','f1c4 d5b6 c4b3 d7d5 e5d6 d8d6 e1g1 c8e6 f3d4 c6d4 c3d4 e6b3 a2b3 g7g6 b1c3 f8g7 d1f3 e8g8 f3b7 f8b8 b7e4 d6d4 e4e7 b6d5 c3d5 d4d5 a1a7 a8a7 e7a7 d5b3 a7c7 g7b2 c1b2 b3b2 c7a7 b2b4 a7e3 b8b7 f1d1 g8g7 g2g3 b4e7 e3e7 b7e7');
INSERT INTO evaluations VALUES(12918,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(12918,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(12918,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12918,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12918,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(12918,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(12918,36,'cp 9','d5b6 c4b3 d7d5 e5d6 d8d6 f3d4 c6d4 c3d4 c8e6 e1g1 e6b3 a2b3 e7e6 d1f3 d6d7 b1c3 f8e7 f1d1 e8g8 d4d5 b6d5 c3d5 e6d5 d1d5 d7c6 c1e3 e7f6 a1c1 c6e6 g2g3 b7b6 c1d1 f8e8 e3d4 a8d8 d4f6 d8d5 d1d5 e6f6 f3f6');
INSERT INTO evaluations VALUES(12755,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(12755,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12755,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(12755,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12755,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(12755,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(12755,36,'cp 0','c4b3 d7d6 e5d6 d8d6 e1g1 c8e6 f3d4 c6d4 c3d4 e6b3 a2b3 g7g6 b1c3 f8g7 d1f3 e8g8 f3b7 f8b8 b7e4 d6d4 e4d4 g7d4 c3b5 d4e5 a1a7 b6d5 a7a4 d5b6');
INSERT INTO evaluations VALUES(46426,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(46426,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(46426,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(46426,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(46426,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(46426,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(46426,36,'cp 101','f1b5 b8c6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6g6 e4g6 h7g6 a1d1 c8g4 b5c6 b7c6 d5e3 g4f3 g2f3 f7f6 g1g2 a8b8 b2b3 g6g5 c2c3 g8f7 f1h1 c5e3 f2e3 a7a5 g2f2 f7e6 f2g3 d8d1 h1d1 c6c5 c3c4 a5a4 b3a4 b8a8');
INSERT INTO evaluations VALUES(46388,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(46388,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(46388,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(46388,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(46388,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(46388,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(46388,36,'cp 87','b8c6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6g6 e4g6 h7g6 f1d1 c8f5 c2c3 f5e4 d5e3 e4f3 g2f3 c6e5 b5e2 g6g5 a2a4 c5e7 e3d5 g8f8 f3f4 g5f4 d5f4 d8d1 a1d1 a8d8 f4d5 e5g6 a4a5 e7d6');
INSERT INTO evaluations VALUES(55239,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(55239,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(55239,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(55239,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(55239,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(55239,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(55239,36,'cp -79','g1f3 g8f6 b1c3 c7c6 f1d3 c8g4 c1e3 e7e6 d1e2 f8b4 e1g1 g4h5 e2e1 h5g6 a1d1 b8d7 e1h4 g6d3 d1d3 b4e7 e3g5 h7h6 f3e5 d7e5 d4e5 f6d5 g5e7 d8e7 h4g3 e8g8 c3e4 f7f5 e5f6 d5f6 g3d6 a8e8 d3f3 e7d6 e4d6 e8e7 d6e4 f6d7 f3f8 d7f8 c2c3 f8g6 e4c5 b7b6');
INSERT INTO evaluations VALUES(55237,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(55237,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(55237,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(55237,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(55237,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(55237,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(55237,36,'cp -72','g8f6 b1c3 c7c6 f1d3 c8g4 c1e3 e7e6 d1e2 f8b4 h2h3 g4h5 a2a3 b4c3 b2c3 e8g8 e1g1 b8d7 e2e1 h5g6 c3c4 a8c8 a1d1 d8c7 d3g6 h7g6 f3g5 c6c5 d4d5 e6d5 c4d5 c7e5 e3f4 e5e1 d1e1 f8e8 d5d6 e8e1 f1e1 f6d5 f4g3 f7f6');
INSERT INTO evaluations VALUES(47370,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(47370,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(47370,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(47370,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(47370,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(47370,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(47370,36,'cp -94','b1c3 c7c6 f1d3 c8g4 c1e3 e7e6 d1e2 f8b4 h2h3 g4h5 a2a3 b4c3 b2c3 e8g8 e1g1 b8d7 e2e1 c6c5 d4c5 a8c8 e1h4 d7c5 e3g5 c5d7 a1b1 h7h6 h4h5 f6h5 g5d8 f8d8 b1b7 d7c5 b7a7 c5d3 c2d3 h5g3 f1e1 c8c3 a3a4 c3d3 a7b7 d3a3 f3e5 f7f6 e5c6');
INSERT INTO evaluations VALUES(14696,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(14696,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(14696,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(14696,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14696,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(14696,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(14696,36,'cp 18','b7c6 d2d4 f6d7 e1g1 a7a5 b2b3 f8e7 d4e5 d6e5 c3a4 e8g8 d1e2 e7d6 c1g5 d8e8 f3h4 h7h6 g5d2 d7b6 a4b6 c7b6 h4f5 e8d8 a1d1 d8f6 f5d6 f6d6 f2f4 c8a6 c2c4 d6d4 e2f2 d4f2 g1f2 e5f4 d2f4 a5a4 d1d6 a4b3 a2b3 f8e8 f4e3 e8e4 e3b6');
INSERT INTO evaluations VALUES(41358,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41358,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(41358,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(41358,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(41358,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(41358,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(41358,36,'cp 46','e8g8 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 f1e2 c7c5 e1g1 c5d4 f3d4 b8c6 e2f3 d5c4 d4c6 b7c6 b2b3 c4a6 a2a4 c8f5 a1b1 f8d8 d1c1 h7h5 f1d1 e7e5 d1d2 e5e4 f3e2 a6a5 e3g5 d8d2 c1d2 a5d2 g5d2 f5e6 b1d1 a8d8 d2e3 d8d1 e2d1 a7a5 d1e2');
INSERT INTO evaluations VALUES(36659,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(36659,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(36659,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(36659,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36659,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(36659,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(36659,36,'cp 52','c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 f1e2 c7c5 c2c4 d5d6 e1g1 c5d4 f3d4 b8c6 d4c6 d6c6 d1b3 c8f5 e2f3 f5e4 f3e4 c6e4 f1e1 e4c6 a1d1 f8d8 b3c2 a7a5 b2b3 g7f6 a2a4 h7h5 d1d2 e7e6 e1d1 d8d2 d1d2 c6c7 d2d3 a8d8 d3d8 c7d8 c2e4 b7b6');
INSERT INTO evaluations VALUES(36652,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36652,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(36652,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(36652,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(36652,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(36652,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(36652,36,'cp 45','d6d5 e4d5 f6d5 c3d5 d8d5 f1e2 c7c5 c2c4 d5d6 e1g1 c5d4 f3d4 b8c6 d4c6 d6c6 d1b3 c8f5 e2f3 f5e4 f3e4 c6e4 f1e1 e4c6 a1d1 h7h5 b3c2 a7a5 b2b3 e7e6 a2a4 g7f6 e3d4 f6d4 d1d4 f8d8 e1d1 c6b6 d4d7 a8c8 d1d2 d8d7 d2d7 b6b4');
INSERT INTO evaluations VALUES(31109,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(31109,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(31109,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(31109,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(31109,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(31109,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(31109,36,'cp 106','f3e5 g4c8 c4d5 c6d5 b1c3 g8f6 c1f4 e7e6 e2e3 f8e7 a1c1 e8g8 f1d3 f6d7 e5f3 d7f6 e1g1 f6h5 f4e5 b8c6 c3b5 c6e5 d4e5 g7g6 d1b3 h5g7 e3e4 g7e8 f1d1 d8b6 e4d5 e6d5 b3d5 c8e6 d5d4 b6d4 f3d4 e6a2 f2f4 a2d5 b5c3 e7c5 c3d5');
INSERT INTO evaluations VALUES(31138,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(31138,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(31138,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(31138,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(31138,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(31138,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(31138,36,'cp 118','g4c8 c4d5 c6d5 b1c3 g8f6 c1f4 e7e6 e2e3 f8e7 f1d3 e8g8 e1g1 f6d7 e5f3 d7f6 c3b5 b8c6 f4c7 d8d7 f3e5 c6e5 c7e5 a7a6 b5c7 a8a7 a1c1 d7d8 d1f3 c8d7 c1c2 d7b5 d3b5 a6b5 a2a3 a7a5 e5f6 e7f6 f3e2 d8d7 f1c1 f6e7 e2g4');
INSERT INTO evaluations VALUES(52886,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52886,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(52886,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(52886,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(52886,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(52886,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(52886,36,'cp 41','a2a4 e7e6 e2e3 b7b5 a4b5 c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8e7 f3e5 d8c7 c1a3 e8g8 e1g1 e7a3 b1a3 b8d7 a3b5 c7b6 c4d3 d7e5 d4e5 f6d7 d1b1 c8b7 b5d6 b7c6 b1b6 a7b6 f2f4 d7c5 d3c4 c5b7 a1a8 f8a8 f1b1 b7d6 e5d6');
INSERT INTO evaluations VALUES(52938,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(52938,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(52938,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(52938,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52938,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52938,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(52938,36,'cp 31','g8f6 e2e3 e7e6 f1c4 c6c5 e1g1 c5d4 e3d4 b8c6 b1c3 f8e7 d1e2 e8g8 f1d1 c6b4 a4a5 c8d7 f3e5 d7e8 c1f4 a7a6 e2f3 f6d5 f4d2 a8c8 f3g3 g8h8 c3d5 b4d5 c4d3 h7h6 g3f3 h8g8 f3e4 f7f5 e4f3 e7f6 a1c1 c8c1 d1c1');
INSERT INTO evaluations VALUES(17170,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(17170,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(17170,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17170,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(17170,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17170,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(17170,36,'cp 32','e5d4 f3d4 c6d4 d1d4 e7c6 d4d3 f8e7 b1c3 a7a6 b5a4 e8g8 c3d5 c6b4 d5b4 e7b4 c2c3 b4c5 a4c2 d7d6 c1e3 c5e3 d3e3 c8e6 e1g1 f8e8 f2f4 g8h8 a2a4 f7f6 h2h3 d8e7 a4a5 a8d8 f1e1 e7f7 a1d1 e6b3 e3f2');
INSERT INTO evaluations VALUES(17178,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(17178,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(17178,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17178,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(17178,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17178,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(17178,36,'cp 32','f3d4 c6d4 d1d4 e7c6 d4e3 f8e7 b1c3 e8g8 e1g1 f8e8 c3d5 e7d6 c1d2 c6e7 d5e7 d8e7 b5d3 d6e5 c2c3 d7d5 f1e1 g7g6 a1d1 c8d7 h2h3 d5e4 d3e4 c7c6 e3g5 e5f6 g5a5 e7e5 a5a4 e5c7 e4f3 d7e6 d2e3 f6g7 a4h4 a7a5 e3g5');
INSERT INTO evaluations VALUES(17177,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(17177,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(17177,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(17177,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(17177,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(17177,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(17177,36,'cp 41','c6d4 d1d4 e7c6 d4e3 f8e7 e1g1 e8g8 b1c3 f8e8 c3d5 e7d6 c1d2 c6e7 d5e7 d8e7 d2c3 c7c6 b5d3 d6e5 e3d2 a7a5 a2a3 e7c5 c3e5 c5e5 c2c3 d7d6 f2f4 e5c5 g1h1 f7f6 b2b4 c5h5 a1e1 c8e6 d3e2 h5f7 d2d6 e6c4 d6d4 c4e2 e1e2 a8d8');
INSERT INTO evaluations VALUES(17201,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(17201,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17201,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17201,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(17201,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(17201,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(17201,36,'cp 29','d1d4 e7c6 d4e3 f8e7 e1g1 e8g8 b1c3 f8e8 f1d1 c6e5 h2h3 c7c6 b5f1 d7d6 c3e2 e7f6 a1b1 e5g6 c2c4 c8e6 e2g3 f6e5 b2b3 d8h4 e3f3 e5g3 f2g3 h4e7 c1f4 a8d8 g1h1 h7h6 f3f2 b7b6 d1d2');
INSERT INTO evaluations VALUES(17200,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(17200,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(17200,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(17200,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(17200,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(17200,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(17200,36,'cp 29','e7c6 d4e3 f8e7 e1g1 e8g8 b1c3 f8e8 f1d1 c6e5 h2h3 b7b6 b5a4 c8b7 a4b3 a7a5 a2a4 d7d6 c3d5 e5d7 d5e7 d8e7 f2f3 d7c5 b3c4 b7a6 c4d5 a6b7 d5b7 c5b7 e3f2 b7c5 b2b3 f7f5 e4f5 e7f6 c1a3 f6f5 d1e1 c5e6 a1d1 f5f7 a3b2 e6f4');
INSERT INTO evaluations VALUES(31327,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(31327,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(31327,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(31327,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(31327,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(31327,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(31327,36,'cp -41','d2d4 e7e6 g1f3 f8d6 d1e2 b8c6 c2c3 d8a5 e1g1 h7h6 f3e5 g8e7 c1f4 a5c7 f1e1 e8g8 b1d2 c6e5 f4e5 d6e5 e2e5 c7e5 e1e5 a7a5 a2a4 f8c8 b3d1 f5h7 e5e1 c8c7 h2h4 e7c8 g2g4 c8d6 f2f3 c7c6 d1b3 a8c8 h4h5');
INSERT INTO evaluations VALUES(31306,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(31306,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(31306,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(31306,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(31306,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(31306,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(31306,36,'cp -42','e7e6 g1f3 f8d6 d1e2 b8c6 c2c3 d8a5 e1g1 h7h6 f3e5 g8f6 c1f4 a5c7 b1a3 a7a6 a3c2 e8g8 f4g3 f5h7 f1e1 f6e4 e5f3 e4g3 h2g3 c6a5 c2e3 a8c8 b3c2 h7c2 e2c2 b7b5 a2a3 c8b8 e1e2 a5c4 a3a4 b5b4 c3b4');
INSERT INTO evaluations VALUES(31690,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(31690,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(31690,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(31690,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(31690,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(31690,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(31690,36,'cp -35','g1f3 f8d6 d1e2 b8c6 c2c3 d8a5 e1g1 h7h6 f3e5 g8e7 c1f4 a5c7 b1d2 d6e5 f4e5 c6e5 e2e5 c7e5 d4e5 e7c6 f2f4 e8e7 h2h3 g7g5 a1e1 h8g8 g1h2 f5g6 f1f2 g5f4 f2f4 b7b5 f4h4 h6h5 b3d1 a8b8 d1f3 a7a5 d2b3 b5b4 f3h5');
INSERT INTO evaluations VALUES(31689,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(31689,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(31689,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(31689,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(31689,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(31689,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(31689,36,'cp -44','f8d6 d1e2 b8c6 c2c3 d8a5 e1g1 h7h6 f3e5 g8e7 c1f4 a5c7 b1d2 d6e5 f4e5 c6e5 e2e5 c7e5 d4e5 e7c6 f2f4 e8e7 f1f2 a7a5 b3a4 h8b8 a1e1 b7b5 a4d1 a5a4 a2a3 b8d8 d1e2 a8b8 g2g4 f5c2 d2f3 c2b3 f4f5 b3c4 f5f6 g7f6 e5f6 e7d6 e1f1 c4e2 f2e2');
INSERT INTO evaluations VALUES(56999,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56999,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(56999,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(56999,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(56999,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(56999,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(56999,36,'cp -10','a2a3 d7d5 g1f3 g8f6 e2e3 c7c5 b4c5 f8c5 d2d4 c5d6 c2c4 e8g8 c1b2 b8c6 f1d3 d5c4 d3c4 b7b5 c4d3 b5b4 b1d2 c8b7 e1g1 a7a5 d1b1 b4a3 b2a3 c6b4 d2c4 d6e7 f1c1 b4d3 a3e7 d8e7 b1d3 b7f3 g2f3 f8c8 d3a3 e7a3 a1a3');
INSERT INTO evaluations VALUES(57001,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(57001,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(57001,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(57001,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(57001,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(57001,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(57001,36,'cp -23','d7d5 g1f3 g8f6 e2e3 c7c5 b4c5 f8c5 d2d4 c5d6 f1d3 e8g8 c2c4 d5c4 d3c4 b8d7 b1c3 e6e5 c4e2 a7a6 a3a4 e5d4 f3d4 d8c7 c1b2 d7c5 h2h3 b7b6 e1g1 c8b7 e2f3 f8d8 f3b7 c7b7 d1f3 a8c8 f3b7 c5b7 f1d1 b7a5 a1c1 d6b4 c3e2');
INSERT INTO evaluations VALUES(11917,30,'cp 456',NULL);
INSERT INTO evaluations VALUES(11917,31,'cp 458',NULL);
INSERT INTO evaluations VALUES(11917,32,'cp 453',NULL);
INSERT INTO evaluations VALUES(11917,33,'cp 438',NULL);
INSERT INTO evaluations VALUES(11917,34,'cp 455',NULL);
INSERT INTO evaluations VALUES(11917,35,'cp 456',NULL);
INSERT INTO evaluations VALUES(11917,36,'cp 482','e4e5 f6g8 f1c4 c8b7 e1g1 f8c5 d4g4 e8f8 c1g5 d8e8 a1d1 b7c6 f1e1 h7h6 g5d2 h6h5 g4f4 a7a5 c3e4 c6e4 e1e4 c7c6 f4f3 b6b5 c4f7 e8f7 e4f4 g8f6 e5f6 g7g6 h2h4 f8g8 d1e1 f7d5 f6f7 g8h7 f3d5 c6d5 d2c3 h8c8 e1d1 a5a4 d1d5');
INSERT INTO evaluations VALUES(11913,30,'cp 492',NULL);
INSERT INTO evaluations VALUES(11913,31,'cp 467',NULL);
INSERT INTO evaluations VALUES(11913,32,'cp 478',NULL);
INSERT INTO evaluations VALUES(11913,33,'cp 477',NULL);
INSERT INTO evaluations VALUES(11913,34,'cp 469',NULL);
INSERT INTO evaluations VALUES(11913,35,'cp 494',NULL);
INSERT INTO evaluations VALUES(11913,36,'cp 466','f6g8 f1c4 c8b7 e1g1 b7c6 f1e1 f8c5 d4g4 e8f8 c1g5 d8e8 a1d1 h7h6 g5d2 h6h5 g4f4 h5h4 h2h3 h8h6 c3e4 c5e7 f4g4 b6b5 c4b3 a7a5 a2a3 a5a4 b3a2 h6e6 a2e6 f7e6 d2g5 e8g6 g4f4 f8e8 g5e7 g8e7 f4h4 g6f5 f2f3 e7g6 h4g3 e8f7 g1h2 f7g8 e4c5 a8f8 c5d7');
INSERT INTO evaluations VALUES(43823,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(43823,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(43823,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43823,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(43823,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(43823,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(43823,36,'cp 17','f8b4 e2e3 e8g8 f1d3 d7d5 e1g1 c7c5 d4c5 d5c4 d3c4 d8d1 f1d1 b4c5 b2b3 c8b7 c1b2 b8c6 c4e2 a8c8 a2a3 f8d8 b3b4 c5f8 d1d8 c8d8 a1c1 h7h6 g1f1 f8e7 b2a1 d8c8 c1d1 c8d8 d1d8 c6d8 a1b2 f6d7 f3d2 f7f5 d2c4');
INSERT INTO evaluations VALUES(39824,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(39824,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(39824,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(39824,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(39824,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(39824,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(39824,36,'cp 17','e2e3 c7c5 f1d3 d7d5 e1g1 e8g8 d4c5 d5c4 d3c4 d8d1 f1d1 b4c5 a2a3 c8b7 c4e2 b8d7 c1d2 f8d8 d2e1 c5f8 c3b5 f6d5 a1c1 a7a5 b5d4 a8c8 e2b5 c8c1 d1c1 d7c5 f3e5 a5a4 d4c6 d8c8 c6b4 h7h6 e5c4 d5b4 e1b4');
INSERT INTO evaluations VALUES(39822,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(39822,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(39822,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(39822,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(39822,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(39822,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(39822,36,'cp 12','c7c5 f1d3 d7d5 c4d5 e6d5 e1g1 e8g8 d4c5 b6c5 c3e2 c5c4 d3c2 b8c6 h2h3 a8b8 c1d2 c8d7 b2b3 d8a5 d2b4 c6b4 d1d2 f8c8 a2a3 c4c3 d2d1 b4c2 d1c2 d7b5 f3d4 b5e2 c2e2 a5b6 a1c1 a7a5 e2d3 f6e4 c1c2 g7g6 f1b1 b6d6 f2f3');
INSERT INTO evaluations VALUES(56104,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56104,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(56104,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(56104,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(56104,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(56104,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(56104,36,'cp -28','c7c6 b5e2 g8f6 d2d4 f8d6 c2c4 e8g8 e1g1 d5c4 e2c4 c8f5 c1g5 b8d7 g5h4 d8c7 b1c3 f8e8 a1c1 h7h6 c4d3 f5d3 d1d3 a7a6 a2a3 e8e7 f1e1 e7e1 c1e1 a8e8 e1e8 f6e8 d3c2 g7g6 b2b4 d7f8 c3e4 c7d7 e4c5 d6c5');
INSERT INTO evaluations VALUES(53677,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(53677,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(53677,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(53677,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(53677,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(53677,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(53677,36,'cp -27','b5e2 g8f6 e1g1 f8d6 d2d4 e8g8 c2c4 d5c4 e2c4 c8f5 c1g5 b8d7 g5h4 d8c7 b1c3 f8e8 a1c1 a7a6 h4g3 d7b6 c4b3 a8d8 f1e1 e8e1 d1e1 d8e8 e1d1 f5e6 b3e6 e8e6 g3d6 c7d6 g2g3 e6e7 d1b3 b6d5 f3e5 h7h6 c1e1 d5c3');
INSERT INTO evaluations VALUES(53735,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(53735,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(53735,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(53735,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(53735,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(53735,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(53735,36,'cp -33','f8d6 c2c4 g8f6 d2d4 d5c4 e2c4 e8g8 e1g1 c8f5 c1g5 b8d7 c4d3 f5d3 d1d3 f8e8 g5h4 d8a5 b1c3 h7h6 a2a3 e8e6 b2b4 a5h5 f1e1 a8e8 e1e6 e8e6 h4g3 d7f8 a1e1 d6g3 h2g3 e6e1 f3e1 h5g5 d3c2 g5g4 e1f3 g4g6 c2b3 g6f5 f3e5 f8e6');
INSERT INTO evaluations VALUES(41289,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(41289,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(41289,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(41289,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(41289,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(41289,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(41289,36,'cp 50','e8g8 c1e3 e7e5 d4d5 b8a6 d1d2 c7c6 g2g4 c8d7 f2f3 c6d5 c4d5 a6c5 h3h4 h7h5 g4g5 f6e8 b2b4 c5a4 c3a4 d7a4 g1e2 a8c8 e2c3 a4d7 f1e2 f7f6 a2a4 f6g5 h4g5 f8f4 h1g1 a7a6 a1c1 b7b5 a4b5 a6b5 e3f4 e5f4');
INSERT INTO evaluations VALUES(36574,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(36574,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(36574,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(36574,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(36574,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(36574,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(36574,36,'cp 47','c1e3 e7e5 d4d5 b8a6 g1e2 a6c5 f2f3 c7c6 d1d2 c6d5 c4d5 c8d7 g2g4 a8c8 b2b4 c5a4 c3a4 d7a4 a1c1 c8c1 d2c1 a4b5 c1d2 a7a6 e2g3 h7h5 f1b5 a6b5 a2a3 d8e7 e1f2 f8c8 f2g2 c8c4 g4h5 f6h5 g3h5 g6h5 h1g1');
INSERT INTO evaluations VALUES(36558,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(36558,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(36558,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(36558,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(36558,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(36558,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(36558,36,'cp 57','e7e5 d4d5 b8a6 g2g4 c7c6 f2f3 c8d7 d1d2 c6d5 c4d5 a6c5 h3h4 h7h5 g4g5 f6e8 b2b4 c5a4 c3a4 d7a4 g1e2 f7f5 e2c3 a4d7 f1e2 f5e4 c3e4 d7f5 a2a4 a7a6 e1g1 d8d7 e2d3 g8h7 b4b5 e8c7 b5a6 c7a6 a1b1 a6c5 e3c5 d6c5');
INSERT INTO evaluations VALUES(36859,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(36859,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(36859,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(36859,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(36859,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(36859,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(36859,36,'cp 76','d4d5 b8a6 g2g4 c7c6 f2f3 c8d7 d1d2 a6c5 h3h4 c6d5 c4d5 h7h5 g4g5 f6e8 b2b4 c5a4 c3a4 d7a4 g1e2 a4d7 e2c3 f7f6 f1e2 a7a6 a2a4 e8c7 e1f2 g8h7 h1c1 f6g5 h4g5 f8f4 c1g1 d8e7 e3f4 e5f4 a1c1 b7b5 c3d1 a8c8 a4b5 c7b5 e2b5 c8c1');
INSERT INTO evaluations VALUES(36834,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(36834,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(36834,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(36834,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(36834,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(36834,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(36834,36,'cp 68','b8a6 g2g4 c7c6 f2f3 c8d7 d1d2 c6d5 c4d5 a8b8 e3a7 b8a8 a7e3 a6c5 a1c1 f6e8 h3h4 f7f5 g4f5 g6f5 e4f5 g7f6 e3f2 d7f5 b2b4 c5d7 g1e2 g8h8 f1g2 f8g8 e2g3 f5g6 g2h3 f6h4 h3e6');
INSERT INTO evaluations VALUES(20331,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(20331,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(20331,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20331,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20331,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(20331,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(20331,36,'cp -12','c8d7 g1f3 g8h6 f1d3 d8b6 d1e2 c5d4 e3d4 c6d4 c3d4 h6f5 d3f5 f8b4 b1c3 e6f5 a2a4 b6a5 e1g1 a8c8 f1c1 e8g8 c3a2 h7h6 h2h3 b4e7 e2e3 d7e6 a2c3 c8c6 c3e2 f8c8 c1c6 b7c6 b2b3 c6c5 d4c5 e7c5');
INSERT INTO evaluations VALUES(25944,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25944,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(25944,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25944,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25944,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25944,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25944,36,'cp 0','g1f3 d8b6 d1d2 c5d4 c3d4 c6a5 f1d3 a8c8 b1c3 a5c4 d3c4 c8c4 e1g1 f8b4 f1c1 g8e7 a2a3 b4c3 c1c3 c4c3 d2c3 e8g8 e3g5 e7c6 c3d2 h7h6 g5f6 f8c8 h2h3 g8h7 d2c2 h7g8');
INSERT INTO evaluations VALUES(25945,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25945,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25945,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25945,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25945,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25945,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25945,36,'cp 0','d8b6 d1d2 a8c8 f1e2 c5d4 c3d4 f8b4 b1c3 c6a5 e1g1 a5c4 e2c4 c8c4 f1c1 g8e7 a2a3 b4c3 c1c3 c4c3 d2c3 e8g8 e3g5 e7c6 c3d2 h7h6 g5f6 f8c8 h2h3 g8h7 d2d3 h7g8');
INSERT INTO evaluations VALUES(27525,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27525,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27525,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27525,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27525,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27525,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27525,36,'cp 0','d1d2 a8c8 f1e2 c5d4 c3d4 f8b4 b1c3 g8e7 a1c1 c6a5 e1g1 h7h6 h2h4 a5c4 e2c4 c8c4 a2a3 b4c3 c1c3 c4c3 d2c3 e8g8 c3d2 g8h7 f1c1 d7a4 d2d3 h7g8 d3d2');
INSERT INTO evaluations VALUES(27524,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27524,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27524,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27524,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27524,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27524,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27524,36,'cp 0','a8c8 f1e2 c5d4 c3d4 f8b4 b1c3 g8e7 e1g1 c6a5 f1c1 a5c4 e2c4 c8c4 a2a3 b4c3 c1c3 c4c3 d2c3 h7h6 a1c1 e8g8 c3d2 f8c8 c1c8 e7c8 h2h4 b6b3 e3h6 g7h6 d2h6 b3c2 h6f6 c8b6 f6d8 g8g7 d8f6 g7f8 f6h8 f8e7 h8f6');
INSERT INTO evaluations VALUES(22805,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(22805,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22805,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22805,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22805,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22805,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22805,36,'cp 0','g8f6 b1c3 f8b4 f1b5 e8g8 e1g1 b4c3 d2c3 d7d6 f1e1 c8e6 a2a4 a7a5 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(16565,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16565,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16565,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16565,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16565,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16565,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16565,36,'cp 0','b1c3 f8b4 f1b5 e8g8 e1g1 b4c3 d2c3 d7d6 f1e1 c8e6 a2a4 a7a5 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(16505,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16505,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(16505,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16505,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16505,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16505,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16505,36,'cp 0','f8b4 f1b5 e8g8 e1g1 b4c3 d2c3 d7d6 f1e1 c8e6 b5d3 a7a5 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(10356,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10356,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(10356,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10356,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10356,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10356,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10356,36,'cp 0','a2a3 b4c3 d2c3 f6e4 f3e5 d8e7 e5c6 d7c6 c1e3 e4g3 f2g3 e7e3 d1e2 e3e2 f1e2 c8e6 e1c1 a8d8 d1e1 e8g8 e2f3 f8e8 b2b3 g8f8 c3c4 e6c8 b3b4 c8e6 e1e4 e6f5 e4e8 d8e8 h1d1 e8e3 d1d8 e3e8');
INSERT INTO evaluations VALUES(10377,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10377,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10377,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10377,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10377,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10377,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10377,36,'cp 0','b4c3 d2c3 f6e4 f3e5 d8e7 e5c6 d7c6 c1e3 e4g3 f2g3 e7e3 d1e2 e3e2 f1e2 c8e6 e1c1 a8d8 d1e1 e8e7 e2f3 h8e8 c3c4 e7f8 b2b3 h7h6 e1e3 e6f5 e3e8 d8e8 h1d1 e8e5 d1d8 e5e8');
INSERT INTO evaluations VALUES(47016,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(47016,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47016,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(47016,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(47016,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(47016,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(47016,36,'cp 26','g1f3 g7g6 f1e2 f8g7 c2c4 e8g8 b1c3 b8d7 e1g1 e7e5 b2b4 f8e8 c1b2 a7a5 b4b5 e5d4 f3d4 d7b6 a1c1 f6d7 b2a3 d7c5 c3d5 b6d5 c4d5 d8e7 h2h3 c8d7 e2f3 h7h5 d1d2 g7d4 e3d4 c5e4 f3e4 e7e4 f1e1 e4d5');
INSERT INTO evaluations VALUES(47017,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(47017,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(47017,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(47017,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(47017,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(47017,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(47017,36,'cp 26','g7g6 f1e2 f8g7 c2c4 e8g8 b1c3 b8d7 e3e4 e7e5 e1g1 f8e8 d4d5 d7c5 d1c2 a7a5 c1g5 b7b6 f3d2 h7h6 g5e3 f6h7 g1h1 c8d7 f1g1 h6h5 a1f1 a5a4 g2g3 f7f5 f2f3 h7f6 e4f5 g6f5 a2a3 d8e7 f1e1');
INSERT INTO evaluations VALUES(46695,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(46695,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(46695,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(46695,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(46695,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(46695,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(46695,36,'cp 42','c2c4 f8g7 b1c3 e8g8 f1e2 b8d7 e1g1 e7e5 e3e4 e5d4 f3d4 c7c6 d1c2 f8e8 f1d1 d8e7 f2f3 a7a6 c1g5 h7h6 g5f4 d7e5 c2d2 g6g5 f4e3 d6d5 c4d5 c6d5 e4d5 e5f3 e2f3 e7e3 d2e3 e8e3 d5d6 c8d7 f3b7 a8b8 b7c6 b8b2 c6d7 f6d7 c3d5');
INSERT INTO evaluations VALUES(46668,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46668,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(46668,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(46668,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(46668,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46668,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(46668,36,'cp 16','f8g7 b1c3 e8g8 f1e2 a7a5 e1g1 b8a6 e3e4 e7e5 c1e3 d8e7 e3g5 f8e8 d4e5 d6e5 c3d5 e7d8 f3e5 e8e5 d5f6 g7f6 d1d8 f6d8 g5d8 e5e4 e2f3 e4c4 f1d1 c7c6 d8a5 c8e6 a5c3 a6c5 c3f6 c5d7 f6d4 f7f6 b2b3 c4c2 a2a4 d7e5 d4e5');
INSERT INTO evaluations VALUES(17802,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17802,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17802,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17802,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17802,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17802,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17802,36,'cp 0','b7b5 a4b3 c6d4 f3d4 e5d4 c2c3 d4c3 d1d5 c8e6 d5c6 e6d7 c6d5');
INSERT INTO evaluations VALUES(17982,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17982,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17982,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17982,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17982,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17982,36,'cp 0','a4b3 c6d4 f3d4 e5d4 c2c3 d4c3 d1d5 c8e6 d5c6 e6d7 c6d5');
INSERT INTO evaluations VALUES(17974,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17974,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17974,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17974,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17974,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17974,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17974,36,'cp 0','c6d4 f3d4 e5d4 c2c3 d4c3 d1d5 c8e6 d5c6 e6d7 c6d5');
INSERT INTO evaluations VALUES(18023,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18023,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18023,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18023,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18023,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18023,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18023,36,'cp 0','f3d4 e5d4 c2c3 d4c3 d1d5 c8e6 d5c6 e6d7 c6d5');
INSERT INTO evaluations VALUES(18022,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18022,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18022,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18022,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18022,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18022,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18022,36,'cp 0','e5d4 c2c3 d4c3 d1d5 c8e6 d5c6 e6d7 c6d5');
INSERT INTO evaluations VALUES(18026,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18026,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18026,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18026,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18026,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18026,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18026,36,'cp 0','c2c3 d4c3 d1d5 c8e6 d5c6 e6d7 c6d5');
INSERT INTO evaluations VALUES(25929,30,'cp -206',NULL);
INSERT INTO evaluations VALUES(25929,31,'cp -209',NULL);
INSERT INTO evaluations VALUES(25929,32,'cp -188',NULL);
INSERT INTO evaluations VALUES(25929,33,'cp -199',NULL);
INSERT INTO evaluations VALUES(25929,34,'cp -176',NULL);
INSERT INTO evaluations VALUES(25929,35,'cp -179',NULL);
INSERT INTO evaluations VALUES(25929,36,'cp -181','c6e5 f3e5 d7b5 a2a4 b5a6 d4c5 f8c5 b2b4 c5d6 e5f7 e8f7 b4b5 g8f6 e1g1 a8c8 b5a6 b7a6 g2g3 h8e8 c1g5 h7h6 g5f6 d8f6 a1a2 f6f5 a2e2 f7g8 e2e3 c8c6 b1d2 d6e5 c3c4 d5c4 d2f3 e5f6 f3d4 f6d4 d1d4 e8c8 e3c3 c6c5 f1e1 e6e5 d4e3 c8c6 e3e4 f5e4 e1e4 a6a5 g1g2');
INSERT INTO evaluations VALUES(25991,30,'cp -196',NULL);
INSERT INTO evaluations VALUES(25991,31,'cp -180',NULL);
INSERT INTO evaluations VALUES(25991,32,'cp -195',NULL);
INSERT INTO evaluations VALUES(25991,33,'cp -181',NULL);
INSERT INTO evaluations VALUES(25991,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(25991,35,'cp -197',NULL);
INSERT INTO evaluations VALUES(25991,36,'cp -163','f3e5 d7b5 b1a3 b5c6 e1g1 g8f6 c1g5 c5d4 g5f6 g7f6 e5c6 b7c6 c3d4 f6f5 a3c2 f8g7 b2b4 d8f6 d1d2 e8g8 f1b1 f8c8 a2a3 h7h5 d2d1 h5h4 h2h3 g7f8 c2e1 f8d6 g1h1 a7a5 e1f3 a5b4');
INSERT INTO evaluations VALUES(25990,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(25990,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(25990,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(25990,33,'cp -188',NULL);
INSERT INTO evaluations VALUES(25990,34,'cp -200',NULL);
INSERT INTO evaluations VALUES(25990,35,'cp -185',NULL);
INSERT INTO evaluations VALUES(25990,36,'cp -184','d7b5 b1a3 b5c6 e1g1 g8f6 c1g5 c5d4 g5f6 g7f6 e5c6 b7c6 c3d4 f8d6 a1c1 d8b6 a3c2 h7h5 f1e1 a8b8 d1f3 h8h6 b2b3 e8e7 h2h3 b8g8 g1h1 e7d7 e1d1 h5h4 c2e3 f6f5 c1c2 g8c8 d1c1 b6d4 e3f5 e6f5 f3f5 h6e6');
INSERT INTO evaluations VALUES(26450,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(26450,31,'cp -177',NULL);
INSERT INTO evaluations VALUES(26450,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(26450,33,'cp -191',NULL);
INSERT INTO evaluations VALUES(26450,34,'cp -176',NULL);
INSERT INTO evaluations VALUES(26450,35,'cp -184',NULL);
INSERT INTO evaluations VALUES(26450,36,'cp -176','c1f4 c5d4 a2a4 b5d7 c3d4 g8e7 f4g5 h7h5 g5e7 f8e7 e1g1 e7d6 b1c3 d6e5 d4e5 a8c8 d1b3 d8c7 f2f4 e8g8 f1f2 a7a5 h2h3 c7c5 b3d1 g7g6 c3e2 c5b4 e2d4 c8c4');
INSERT INTO evaluations VALUES(14286,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14286,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14286,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(14286,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(14286,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(14286,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14286,36,'cp 0','g1f3 c6b4 c1d2 b4d3 c2d3 e7e6 e1g1 b7b6 b1c3 f8e7 b2b4 e8g8 b4b5 c8b7 a2a4 f6d7 d1b3 a7a5 f1c1 e7b4 h2h3 a8c8 c1c2 d8e7 a1c1 b4a3 c1a1');
INSERT INTO evaluations VALUES(11929,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(11929,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(11929,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(11929,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(11929,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(11929,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(11929,36,'cp -48','d8e7 d2d4 d7d6 e5f3 e7e4 c1e3 e4g6 e1g1 c8h3 f3e1 f8e7 b1c3 h7h6 d1f3 h3d7 h2h3 g6f5 f3e2 e8g8 e1d3 a7a5 b2b3 f8e8 a1e1 d7c8 g2g4 f5d7 f2f4 a5a4 e2f3 a4b3 a2b3 f6h7 e3f2 c8a6 g1h2 h7f6 e1e3 e7f8 f4f5');
INSERT INTO evaluations VALUES(696,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(696,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(696,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(696,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(696,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(696,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(696,36,'cp -48','d2d4 d7d6 e5f3 e7e4 c1e3 e4g6 e1g1 c8h3 f3e1 f8e7 b1c3 h7h5 g1h1 h3d7 d1d2 e8g8 h2h3 f8e8 e1f3 f6d5 c3d5 c6d5 f1e1 a7a5 e3f4 a5a4 b2b4 d7f5 f4g5 f7f6 f3h4 g6g5 d2g5 f6g5 h4f5 e7f6 h1g1 g5g4 e1e8 a8e8 g1f1 g4h3 g2h3');
INSERT INTO evaluations VALUES(695,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(695,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(695,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(695,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(695,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(695,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(695,36,'cp -56','d7d6 e5f3 e7e4 c1e3 e4g6 e1g1 c8h3 f3e1 f8e7 b1c3 h7h5 g1h1 h3d7 d1d2 e8g8 e1f3 f6d5 c3d5 c6d5 a1c1 f8e8 h2h3 a7a5 f1e1 e7d8 b2b3 a5a4 e3f4 e8e4 c2c4 g6f5 f4g5 e4e1 c1e1 f7f6 g5e3 g7g5 c4c5 a4b3 a2b3');
INSERT INTO evaluations VALUES(48361,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(48361,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(48361,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(48361,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(48361,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(48361,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(48361,36,'cp 25','g1f3 f8e7 e1g1 c7c6 b2b3 e8g8 c2c4 b7b6 d1c2 c8b7 b1c3 b8d7 f1d1 b7a6 c1f4 f6h5 f4d2 h5f6 e2e4 d5c4 e4e5 f6d5 c3e4 c4b3 a2b3 a6e2 e4g5 g7g6 d1e1 e2f3 g5f3 c6c5 d4c5 d7c5 b3b4 c5d7 f3d4 d5b4 c2a4 b4d5 d4c6');
INSERT INTO evaluations VALUES(13092,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13092,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(13092,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(13092,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(13092,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(13092,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13092,36,'cp 0','d8b6 d1c2 g7g6 b1d2 f8g7 h2h3 e8g8 f1e2 c5d4 e3d4 f6h5 f4h2 g7h6 e1g1 h5f4 d2b3 c8f5 c2d1 a7a5 f3h4 f5c8 h4f3');
INSERT INTO evaluations VALUES(798,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(798,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(798,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(798,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(798,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(798,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(798,36,'cp 0','d1c2 g7g6 b1d2 f8g7 h2h3 e8g8 f1e2 c5d4 e3d4 f6h5 f4h2 g7h6 e1g1 h5f4 d2b3 c8f5 c2d1 a7a5 f3h4 f5c8 h4f3');
INSERT INTO evaluations VALUES(800,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(800,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(800,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(800,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(800,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(800,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(800,36,'cp 0','g7g6 b1d2 f8g7 h2h3 e8g8 f1e2 c5d4 e3d4 f6h5 f4h2 g7h6 d2b3 c8f5 c2d1 h5f4 e1g1 a7a5 f3h4 f5c8 h4f3');
INSERT INTO evaluations VALUES(1779,30,'cp -858',NULL);
INSERT INTO evaluations VALUES(1779,31,'cp -857',NULL);
INSERT INTO evaluations VALUES(1779,32,'cp -863',NULL);
INSERT INTO evaluations VALUES(1779,33,'cp -858',NULL);
INSERT INTO evaluations VALUES(1779,34,'cp -864',NULL);
INSERT INTO evaluations VALUES(1779,35,'cp -873',NULL);
INSERT INTO evaluations VALUES(1779,36,'cp -867','g5g2 f7g5 f8e7 c4f1 g2h1 c2c3 e7g5 c3d4 h1e4 d1e2 e4e2 f1e2 g8e7 b1c3 g5h4 d2d3 e8g8 c3e4 d7d5 e4g3 e7f5 e2d1 c8d7 d1g4 a8e8 e1d1 e8e7 g4f5 d7f5 c1e3 f5d3 d1d2 d3a6 d2c3 g7g6 a2a4 g8f7 a1c1 c7c6 b2b4 a6c4 c1e1 a7a6 g3e2 c4e2 e1e2');
INSERT INTO evaluations VALUES(1781,30,'cp -871',NULL);
INSERT INTO evaluations VALUES(1781,31,'cp -860',NULL);
INSERT INTO evaluations VALUES(1781,32,'cp -863',NULL);
INSERT INTO evaluations VALUES(1781,33,'cp -855',NULL);
INSERT INTO evaluations VALUES(1781,34,'cp -863',NULL);
INSERT INTO evaluations VALUES(1781,35,'cp -863',NULL);
INSERT INTO evaluations VALUES(1781,36,'cp -878','f7g5 g2h1 c4f1 f8e7 c2c3 e7g5 c3d4 h1e4 d1e2 e4e2 f1e2 g8e7 b2b3 g5h4 b1c3 e8g8 c1a3 d7d6 e1c1 c8e6 c3e4 h4f2 a3b2 f2h4 e2c4 e6c4 b3c4 f8f4 d2d3 d6d5 e4c5 f4f2 c5e6 f2h2 d1f1 h2f2 f1f2 h4f2 b2a3 g8f7 e6c7 a8c8 a3e7 f2e3 c1b1 c8c7');
INSERT INTO evaluations VALUES(10067,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(10067,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(10067,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(10067,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(10067,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(10067,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(10067,36,'cp -65','e8g8 b1c3 c6d4 g5f3 c7c6 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d2c3 d7d5 c4d3 f8e8 d1d2 c8g4 a2a4 a7a5 h2h3 g4h5 d2g5 h5e2 g5f6 g7f6 d3e2 e8e2 c1e3 c5e3 f2e3 g8g7 f1f2 e2e3 a1f1 e3e6 f2f4 h7h5 f1f3 a8e8 b2b4 e6e3 f3e3 e8e3 b4a5 e3e4 g2g3 g7g6 g1g2 e4f4 g3f4 g6f5 g2f3 c6c5');
INSERT INTO evaluations VALUES(4307,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(4307,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(4307,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(4307,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(4307,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(4307,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(4307,36,'cp -53','b1c3 c6d4 g5f3 c7c6 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d2c3 d7d5 c4d3 f8e8 d1d2 c8g4 a2a4 a7a5 h2h3 g4h5 d2g5 h5e2 g5f6 g7f6 d3e2 e8e2 c1e3 c5e3 f2e3 g8g7 f1f2 e2e3 a1f1 e3e6 f2f4 h7h5 f4f5 g7g6 f5f3 a8e8 b2b4 e8a8 f3g3 g6h6 f1f4 a5b4 f4b4 a8a7 b4f4 b7b6 g3g8 a7a5');
INSERT INTO evaluations VALUES(4295,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(4295,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(4295,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(4295,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(4295,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(4295,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(4295,36,'cp -72','c6d4 g5f3 c7c6 f3d4 e5d4 e4e5 d4c3 e5f6 d8f6 d2c3 d7d5 c4d3 f8e8 d1d2 c8g4 a2a4 a7a5 h2h3 g4h5 d2g5 h5e2 g5f6 g7f6 d3e2 e8e2 c1e3 c5e3 f2e3 g8g7 f1f2 e2e3 a1f1 e3e6 f2f5 a8d8 f5f4 h7h5 b2b4 d8a8 c3c4 d5c4 b4a5 a8a5 f4c4 b7b6 c4b4 a5a8 b4b3 a8a4 b3b6 a4c4');
INSERT INTO evaluations VALUES(53719,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(53719,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(53719,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(53719,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(53719,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(53719,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(53719,36,'cp 1','f1d3 f8d6 g1f3 g8e7 c3e2 c8f5 d3f5 e7f5 e2g3 f5e7 e1g1 e8g8 f1e1 f8e8 c1g5 h7h6 g5h4 d6g3 h4g3 b8d7 c2c3 d7f6 d1d3 f6e4 e1e2 e7g6 a1e1 d8d7 h2h3 a7a5 f3e5 g6e5 g3e5 f7f6 e5f4 a5a4 f2f3 e4d6');
INSERT INTO evaluations VALUES(53721,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(53721,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53721,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(53721,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53721,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53721,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(53721,36,'cp 10','f8d6 g1f3 g8e7 c3e2 e8g8 e2g3 e7g6 e1g1 g6f4 f1e1 f4d3 d1d3 b8a6 c1d2 a6c7 h2h3 f7f6 f3h4 c8e6 a2a4 f8e8 h4f5 d6g3 f5g3 d8d7 h3h4 e6f7 d3f3 c7e6 c2c3 h7h6 h4h5 e6f8 g3f5 g8h8 a4a5 e8e1 a1e1 a8e8 e1e8 f7e8');
INSERT INTO evaluations VALUES(38086,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(38086,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38086,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(38086,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38086,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(38086,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(38086,36,'cp 16','g1f3 g8e7 c3e2 e8g8 e2g3 e7g6 e1g1 g6f4 f1e1 f4d3 d1d3 b8a6 c1d2 a6b4 d3c3 b4a6 h2h3 f7f6 c3d3 a6c7 f3h4 c8e6 h4f5 d6g3 f5g3 d8d7 d2f4 a8e8 f4c7 d7c7 e1e3 e6f7 a1e1 a7a5 e3e8 f8e8 e1e8 f7e8 d3e2 c7d7 h3h4 h7h6 h4h5');
INSERT INTO evaluations VALUES(38085,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38085,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(38085,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(38085,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(38085,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(38085,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38085,36,'cp 9','g8e7 c3e2 e8g8 e2g3 e7g6 e1g1 g6f4 f1e1 f4d3 d1d3 b8a6 c1d2 a6c7 h2h3 f7f6 f3h4 c8e6 h4f5 f8e8 h3h4 d6g3 f5g3 d8d7 d2f4 e6f7 h4h5 e8e1 a1e1 a8e8 e1e8 c7e8 f4b8 a7a6 a2a4 h7h6 f2f3 f7e6 b2b3');
INSERT INTO evaluations VALUES(40253,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(40253,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(40253,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(40253,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(40253,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40253,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(40253,36,'cp 12','c3e2 c8f5 d3f5 e7f5 e2g3 f5e7 e1g1 e8g8 f1e1 f8e8 c1g5 f7f6 g5d2 d8d7 f3h4 d6g3 h2g3 b8a6 d1h5 a6c7 d2b4 g7g6 h5h6 c7e6 h6d2 e7c8 h4f3 c8d6 b4d6 d7d6 a2a4 e8e7 a4a5 a8e8 e1e2 e6c7 e2e7 e8e7 g3g4 g6g5 f3h2 e7e4 c2c3 g8g7 h2f1 e4g4');
INSERT INTO evaluations VALUES(16680,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(16680,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(16680,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(16680,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(16680,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(16680,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(16680,36,'cp -78','d4d5 c6e5 f2f4 e5g6 e4e5 f8c5 e5f6 e8g8 f1e2 d8f6 c3e4 f6e7 e4c5 e7c5 d1d3 f8e8 c1d2 d7d6 e1c1 c8f5 d3f5 e8e2 g2g3 c5c4 b2b3 c4d4 c2c3 d4f2 f5d3 a8e8 h2h4 h7h5 d1f1 f2g2 h1g1 g2h3 g1h1 h3g4 f1e1 g8f8 d3c4 e8e7 e1e2 g4e2 h1e1 e2c4 b3c4 e7e1 d2e1 g6e7 c1c2 a7a6 e1f2 f8e8 f2d4 e7f5 d4f2');
INSERT INTO evaluations VALUES(16333,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(16333,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(16333,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(16333,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(16333,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(16333,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(16333,36,'cp -62','c6e5 f2f4 e5g6 e4e5 f8c5 e5f6 e8g8 f1e2 d8f6 c3e4 f6e7 d1d3 f8e8 e4c5 e7c5 c1d2 d7d6 e1c1 c8f5 d3f5 e8e2 g2g3 c5c4 b2b3 c4d4 c2c3 d4f2 f5d3 a8e8 h2h4 h7h5 h1f1 f2g2 f1g1 g2h3 g1h1 h3g4 h1e1 a7a6 c3c4 g6f8 e1e2 e8e2 d1e1 e2g2 e1e3 f7f6 a2a4 f8d7 e3e8 g8f7 d3e4 g2g1 d2e1 g4g3 e4e7 f7g6');
INSERT INTO evaluations VALUES(16924,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(16924,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(16924,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(16924,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(16924,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(16924,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(16924,36,'cp -71','f2f4 e5g6 e4e5 f8b4 e5f6 d8f6 d1e2 e8d8 e2c4 a7a5 f1e2 b7b5 c4d3 g6f4 c1f4 f6f4 h1f1 f4h4 g2g3 h4h6 f1f7 h8e8 h2h4 a8a6 a2a3 b4c3 b2c3 a6f6 f7f6 h6f6 d5d6 c7d6 e1c1 f6e5 e2g4 d8c7 c1b2 c8b7 g4f3 b7f3 d3f3 b5b4 a3b4 a5b4 d1d4 b4c3 f3c3 c7d8 d4c4 e8e6 c4c7 e5c3 c7c3 g7g6 c3c4 d6d5 c4c5 e6e5');
INSERT INTO evaluations VALUES(16925,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(16925,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(16925,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(16925,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(16925,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(16925,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(16925,36,'cp -69','e5g6 e4e5 f8b4 e5f6 d8f6 d1e2 e8d8 e2c4 h8e8 f1e2 b4c3 c4c3 f6e7 c3f3 g6h4 f3f2 b7b6 e1d1 c8b7 c2c4 c7c6 g2g3 h4f5 e2d3 f5d6 h1e1 e7e1 f2e1 e8e1 d1e1 c6d5 c4d5 b7d5 b2b3 d8c7 c1b2 a8e8 e1f2 f7f6 b2a3 g7g6 a1c1 d5c6 a3b2 e8e6 c1e1 c6e4 d3f1 a7a5 b2d4 d6f5 e1c1 e4c6 d4b2');
INSERT INTO evaluations VALUES(17058,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(17058,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(17058,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(17058,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(17058,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(17058,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(17058,36,'cp -69','e4e5 f8b4 e5f6 d8f6 d1e2 e8d8 e2c4 h8e8 f1e2 b4c3 c4c3 f6e7 c3f3 g6h4 f3f2 b7b6 e1d1 c8b7 c2c4 c7c6 g2g3 h4f5 e2d3 f5d6 h1e1 e7e1 f2e1 e8e1 d1e1 c6d5 c4d5 b7d5 b2b3 g7g6 c1b2 d6e4 a1c1 d8e7 b2d4 e4d6 d4b2 d5e4 d3a6 e4c6 e1d2 d6e4 d2e3 e4c5 a6f1 a8e8 b2a3 d7d6');
INSERT INTO evaluations VALUES(17059,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(17059,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(17059,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(17059,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(17059,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(17059,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(17059,36,'cp -70','f8b4 e5f6 d8f6 d1e2 e8d8 e2c4 h8e8 f1e2 b4c3 c4c3 f6e7 c3f3 g6h4 f3f2 b7b6 e1d1 c8b7 c2c4 c7c6 g2g3 h4f5 e2d3 f5d6 h1e1 e7e1 f2e1 e8e1 d1e1 c6d5 c4d5 b7d5 b2b3 d8c7 c1b2 a8e8 e1f2 d5e4 a1c1 c7b7 d3f1 f7f6 c1d1 b7c7 b2a3 e8e6 d1e1 e4f5 e1e6 d7e6 f1c4 f5g4 a3b2 d6c4 b3c4 c7c6 b2a3');
INSERT INTO evaluations VALUES(44956,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(44956,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44956,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(44956,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(44956,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(44956,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(44956,36,'cp 29','f1d3 c8g4 c2c4 e7e6 e1g1 b8d7 b1d2 f8e7 b2b3 g4f5 d1c2 f5d3 c2d3 e8g8 e3e4 d5e4 d2e4 h7h6 c1f4 f6e4 d3e4 f8e8 f1d1 d8a5 h2h3 a8d8 a2a3 a5f5 e4f5 e6f5 b3b4 d7f8 f4e3 e7f6 g2g3 f8e6 a1b1 a7a6 g1g2');
INSERT INTO evaluations VALUES(44952,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44952,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44952,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44952,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44952,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44952,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44952,36,'cp 21','c8g4 c2c4 e7e6 b1d2 f8e7 b2b3 g4f5 d1c2 f5d3 c2d3 c6c5 e1g1 c5d4 f3d4 e8g8 c1b2 b8c6 f1d1 c6d4 b2d4 d8a5 d3e2 a8c8 c4d5 f6d5 d2c4 a5a6 a2a4 f8d8 a1c1 d5f6 e2f3 b7b5 a4b5 a6b5 e3e4 d8d7');
INSERT INTO evaluations VALUES(29048,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(29048,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(29048,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29048,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(29048,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(29048,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(29048,36,'cp 17','c2c4 e7e6 b1d2 g4f5 d1c2 f5d3 c2d3 f8e7 e1g1 a7a5 b2b3 e8g8 c1b2 b8d7 f1d1 h7h6 a2a3 f8e8 h2h3 d8c7 a1c1 a8c8 d3c2 b7b6 c4d5 e6d5 f3e5 c6c5 c2f5 c7b7 e5d7 b7d7 f5d7 f6d7 d4c5 d7c5 b2d4 c5d7 a3a4 e7f6');
INSERT INTO evaluations VALUES(9036,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(9036,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(9036,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(9036,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(9036,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(9036,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(9036,36,'cp 121','a2a3 b4c3 b2c3 c6a5 d1a4 c7c6 c4d3 c8d7 e1g1 e8g8 a4d1 c6c5 f1e1 h7h6 e3e4 d7c6 a1b1 a8c8 c1f4 f6h5 f4d2 h5f6 h2h3 b7b5 d2f4 c5d4 c3d4 a5c4 a3a4 a7a6 d1e2 f6d7 a4b5 a6b5 b1c1 d7b6 d4d5 e6d5 e4e5 f8e8 f3d4');
INSERT INTO evaluations VALUES(9038,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(9038,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(9038,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(9038,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(9038,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(9038,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(9038,36,'cp 96','b4c3 b2c3 c6a5 c4d3 b7b6 a3a4 c7c5 e1g1 e8g8 e3e4 h7h6 f1e1 c8b7 c1f4 a8c8 h2h3 c5d4 c3d4 f8e8 f4g3 a5c6 d3b1 f6h5 g3h2 h5f6 a1a2 c6a5 h2g3 e8e7 a2d2 e7d7 f3e5 d7c7 d4d5 e6d5 e4d5 c7c1 d1c1 c8c1 e1c1');
INSERT INTO evaluations VALUES(9049,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(9049,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(9049,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(9049,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(9049,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(9049,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(9049,36,'cp 86','b2c3 c6a5 d1a4 c7c6 c4d3 c8d7 e1g1 c6c5 a4d1 d7c6 f1e1 c6f3 d1f3 e8g8 a1b1 a8c8 h2h3 d8d7 d3b5 d7d8 b5f1 f8e8 a3a4 c5d4 c3d4 e8e7 c1d2 h7h6 e1d1 e7d7 f1b5 a5c6 b5d3 c6a5 d2e1 a7a6 d3f1 a5c6 f3e2 c6e7 e2b2 e7d5 d1c1 c8c1 b1c1');
INSERT INTO evaluations VALUES(9048,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(9048,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(9048,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(9048,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(9048,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(9048,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(9048,36,'cp 102','c6a5 c4b5 c8d7 b5d3 c7c5 e1g1 d7c6 f1e1 c6f3 d1f3 e8g8 a1b1 a8c8 h2h3 d8d7 d3f1 c5d4 c3d4 f8d8 c1d2 a5c6 d2c3 h7h6 b1d1 c6e7 e3e4 d7c6 d1c1 c6d7 c3b2 c8c1 e1c1 d7a4 f1d3 d8d7 d3b1 a4a5 d4d5 e6d5 b2f6 g7f6');
INSERT INTO evaluations VALUES(19795,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19795,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(19795,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19795,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(19795,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(19795,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19795,36,'cp 30','e4d5 d8d5 d2d4 e7e6 b1a3 g8f6 c1e3 a7a6 d4c5 d5d1 e1d1 f6d5 e3d4 c6d4 c3d4 b7b6 c5c6 f8a3 b2a3 a8a7 f1d3 a7c7 a1c1 d5e7 d1d2 f7f6 d3e4 e8g8 d2e3 f8d8 f3d2 f6f5 e4f3 f5f4 e3f4 d8d4 f4e3');
INSERT INTO evaluations VALUES(19769,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19769,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19769,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19769,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(19769,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(19769,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(19769,36,'cp 25','d8d5 d2d4 e7e6 b1a3 g8f6 c1e3 c5d4 a3b5 d5d7 b5d4 f8e7 d4c6 d7c6 f1d3 e8g8 d1e2 b7b6 e3d4 c8b7 e1c1 a8c8 c1b1 f8d8 h1e1 f6d7 d3c2 c6c4 e2c4 c8c4 c2e4 b7e4 e1e4 c4c8 b1c2 h7h6 h2h4 d7f8 a2a4 f7f6 e4e1 e6e5 d4e3 g8f7 h4h5 d8d1');
INSERT INTO evaluations VALUES(1401,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(1401,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(1401,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(1401,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(1401,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(1401,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(1401,36,'cp 21','d2d4 e7e6 c1e3 c5d4 c3d4 f8b4 b1c3 g8f6 f1d3 e8g8 e1g1 d5h5 d3e2 h5a5 a1c1 c8d7 d1b3 b4d6 f3d2 a5b4 b3b4 d6b4 a2a3 b4a5 d2c4 a5c3 b2c3 b7b6 e2f3 a8d8 e3d2 h7h6 f1e1 c6e7 c4e3 d7c6 f3c6 e7c6');
INSERT INTO evaluations VALUES(1395,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(1395,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(1395,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(1395,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(1395,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(1395,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(1395,36,'cp 22','e7e6 b1a3 g8f6 c1e3 c5d4 a3b5 d5d7 b5d4 f8e7 d4c6 d7c6 f3e5 c6d5 d1a4 f6d7 e5f3 e8g8 a1d1 d5f5 f1e2 d7c5 a4c4 b7b6 f3d4 f5g6 d4c6 e7f6 e3c5 b6c5 e1g1 c8b7 e2f3 g6g5 c4e2 b7c6 f3c6 a8d8 g2g3 d8d1');
INSERT INTO evaluations VALUES(1541,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(1541,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(1541,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(1541,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(1541,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(1541,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(1541,36,'cp 21','f1d3 g8f6 e1g1 c5d4 c3d4 d5d6 b1c3 f8e7 c3b5 d6d8 c1f4 f6d5 f4g3 c6b4 a1c1 e8g8 d3b1 c8d7 b5c3 d7c6 h2h4 d5f6 f3e5 a8c8 f1e1 d8b6 e5c6 c8c6 g3e5 b4d5 c3d5 f6d5 c1c6 b7c6 d1c2 g7g6 h4h5 e7f6 a2a3 f6e5 d4e5 d5f4 h5g6');
INSERT INTO evaluations VALUES(4349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4349,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(4349,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(4349,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(4349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4349,36,'cp -8','d7d6 d2d3 c6d4 h2h3 c7c6 c4a2 a7a5 f3d4 c5d4 c3e2 d4b6 e2g3 c8e6 d1e2 f6d7 g1h1 d8h4 h1h2 h4e7 f2f4 e5f4 c1f4 e6a2 a1a2 f7f5 e2d2 f5e4 d3e4 d7f6 a2a1 f6e4 g3e4 e7e4 f4d6 f8f1 a1f1 e4d5 d2d5 c6d5');
INSERT INTO evaluations VALUES(3839,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(3839,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(3839,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(3839,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3839,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(3839,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3839,36,'cp 0','d2d3 c6d4 h2h3 c7c6 c4a2 c5b6 f3d4 b6d4 d1f3 h7h6 c1e3 d4b6 a1d1 c8e6 d3d4 e5d4 e3d4 b6d4 d1d4 d8b6 d4b4 b6c7 b4d4');
INSERT INTO evaluations VALUES(3841,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3841,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3841,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3841,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3841,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3841,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3841,36,'cp 0','c6d4 h2h3 c7c6 c4a2 c5b6 f3d4 b6d4 c3e2 d4b6 e2g3 c8e6 a2e6 f7e6 g3h5 h7h6 c1e3 f6h5 d1h5 b6e3 f2e3 d8b6 h5g4 b6e3 g1h2 a8e8 f1f8 g8f8 a1f1 f8g8 f1f6 e3c1 b2b3 c1c2 f6h6 c2f2 h6e6 e8e6 g4e6 g8h7 e6f5 f2f5 e4f5 h7g8 h2g3 c6c5 g3f3 d6d5 b3b4');
INSERT INTO evaluations VALUES(18754,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(18754,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(18754,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(18754,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(18754,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(18754,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(18754,36,'cp -62','g8f6 d2d3 d7d5 e4d5 e6d5 c4a2 d8e7 d1e2 c8g4 e2e7 f8e7 b1d2 e8c8 h2h3 g4h5 c2c3 h8e8 e1g1 e7d6 d3d4 c5c4 a2b1 h5g6 b1g6 f7g6 b2b3 c6a5 b3b4 a5c6 a3a4 e8e2 f1e1 e2e1 f3e1 d8e8 e1f3 c6d8 d2f1 d8e6 c1d2 e6f4 d2f4 d6f4 g2g3');
INSERT INTO evaluations VALUES(12385,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(12385,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(12385,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(12385,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(12385,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(12385,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(12385,36,'cp -77','d2d3 d7d5 e4d5 e6d5 c4a2 d8e7 d1e2 c8g4 e2e7 f8e7 b1d2 e8c8 h2h3 g4h5 c2c3 h8e8 e1g1 e7d6 d3d4 c5c4 a2b1 h5g6 b1g6 f7g6 b2b3 c4b3 d2b3 e8e2 c1d2 f6e4 a3a4 d8e8 f1e1 e2e1 a1e1 c8d7 g1f1 h7h6 e1c1 g6g5 c3c4 d5c4 c1c4 e4d2 f3d2 b7b6 g2g3 c6e7 f1e2 e7d5 e2d3 d5b4 d3c3 e8e2 c4b4 d6b4 c3b4');
INSERT INTO evaluations VALUES(12382,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(12382,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(12382,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(12382,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(12382,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(12382,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(12382,36,'cp -68','d7d5 e4d5 e6d5 c4a2 d8e7 d1e2 c8g4 e2e7 f8e7 b1d2 e8c8 h2h3 g4h5 c2c3 e7d6 e1g1 h8e8 d3d4 c5c4 a2b1 h5g6 b1g6 f7g6 f1e1 e8e1 f3e1 c6a5 d2f3 a5b3 a1b1 d8e8 c1e3 h7h6 e1c2 g6g5 b1d1 a7a6 d1e1 b7b5 a3a4 c8d7 a4b5 a6b5 e1d1 f6e4 d1e1 d7c6');
INSERT INTO evaluations VALUES(13534,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(13534,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(13534,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(13534,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(13534,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(13534,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(13534,36,'cp -54','e4d5 e6d5 c4a2 d8e7 d1e2 c8g4 e2e7 f8e7 b1d2 e8c8 h2h3 g4h5 c2c3 h8e8 e1g1 e7d6 d3d4 c5c4 a2b1 h5g6 b1g6 h7g6 g2g3 c6a5 a1b1 e8e7 f1e1 d8e8 e1e7 e8e7 d2f1 e7e2 c1d2 a5b3 b1e1 e2e1 f3e1 c8d7 h3h4 d7c6 e1f3 f6e4 d2e3 f7f6 a3a4');
INSERT INTO evaluations VALUES(13514,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(13514,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(13514,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(13514,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(13514,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(13514,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(13514,36,'cp -86','e6d5 c4a2 h7h6 e1g1 f8d6 f1e1 c8e6 d3d4 e8g8 d4c5 d6c5 b1c3 f8e8 b2b4 f6e4 c3e4 d5e4 b4c5 e4f3 a2e6 e8e6 c1e3 f3g2 d1d8 a8d8 g1g2 d8d7 e1d1 d7e7 e3f4 e6g6 g2f1 g6g4 f4d6 e7d7 d6g3 d7d1 a1d1 g4d4 d1b1 d4d7 a3a4 f7f6 h2h4 g8f7 g3d6 f7g6');
INSERT INTO evaluations VALUES(13742,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(13742,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(13742,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(13742,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(13742,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(13742,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(13742,36,'cp -70','c4a2 h7h6 e1g1 f8d6 f1e1 c8e6 d3d4 e8g8 d4c5 d6c5 b1c3 f8e8 b2b4 f6e4 c3e4 d5e4 a2e6 e8e6 b4c5 e4f3 d1d8 a8d8 c1e3 f3g2 g1g2 f7f6 e1d1 g8f7 d1d8 c6d8 c2c3 e6e7 a1d1 d8c6 d1d5 f7e6 d5d6 e6f5 e3d4 f5g4 d6d5');
INSERT INTO evaluations VALUES(13776,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(13776,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(13776,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(13776,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(13776,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(13776,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(13776,36,'cp -98','h7h6 e1g1 f8d6 b1c3 c8e6 d3d4 c5c4 b2b3 e8g8 b3c4 d5c4 f1e1 f6d5 c3b5 d6b8 a2c4 a7a6 c4d5 e6d5 b5c3 d5f3 d1f3 c6d4 f3d3 d8d6 g1f1 f8d8 c3e2 b8a7 c1e3 d6f6 e2d4 a7d4 e3d4 d8d4 d3c3 a8d8 e1e3 f6h4 e3h3 h4f4 h3g3 g7g6 a1e1 b7b5');
INSERT INTO evaluations VALUES(31781,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(31781,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(31781,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(31781,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(31781,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(31781,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(31781,36,'cp 135','c4d5 e6d5 b1c3 g8f6 f2f3 b7b6 e2e4 d5b7 c1e3 e7e6 d1b3 f8e7 e1c1 e8g8 c1b1 d8c8 g2g4 c7c5 d4d5 e6d5 e4d5 f8d8 g1h3 b8d7 g4g5 f6h5 h3f2 d7e5 b3c2 e5g6 f2e4 c8b8 a2a3 e7f8 f1c4 h5f4 b1a1 g8h8');
INSERT INTO evaluations VALUES(31775,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(31775,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(31775,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(31775,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(31775,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(31775,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(31775,36,'cp 128','e6d5 b1c3 g8f6 f2f3 b7b6 c3d5 d8d5 e2e4 d5d7 g2g4 e7e6 g4g5 f6h5 f3f4 g7g6 c1e3 h7h6 g5h6 h8h6 f4f5 e6f5 e4f5 b8c6 f5g6 h6g6 d1h5 g6e6 h5f3 f8h6 f1h3 e6e3 f3e3 h6e3 h3d7 e8d7 g1f3 a8e8 a1d1 c6d4 h2h4 c7c5 e1f1 d7c6 f1g2 e8e4 f3d4 e3d4 h4h5 d4b2');
INSERT INTO evaluations VALUES(31819,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(31819,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(31819,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(31819,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(31819,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(31819,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(31819,36,'cp 137','b1c3 g8f6 f2f3 d5c4 d1a4 b7b5 c3b5 c4b5 a4b5 c7c6 b5a4 b8d7 c1d2 d8b6 e1c1 a8b8 b2b3 e7e5 e2e4 e5d4 d2a5 b6c5 c1b1 d7b6 a5b6 c5b6 a4d4 b6c7 e4e5 f6d5 f1c4 b8d8 g1e2 f8a3 d1d3 e8g8 h1d1 h7h6 d4e4 d5b6 f3f4 b6c4 e4c4 d8d3 d1d3 f8d8');
INSERT INTO evaluations VALUES(31818,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(31818,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(31818,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(31818,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(31818,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(31818,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(31818,36,'cp 123','g8f6 f2f3 b7b6 c3d5 d8d5 e2e4 d5d7 g1e2 e7e6 c1e3 f8b4 e2c3 e8g8 a2a3 b4c3 b2c3 f8d8 d1b3 c7c5 f1b5 d7c7 e1g1 a7a6 b5e2 b8c6 b3b2 c5d4 c3d4 c7d7 f1d1 b6b5 e3f2 h7h6 e2f1 d8c8 h2h3 c6a5 b2b4 a5c4 a3a4 a8b8 a1b1 d7e8 f2g3 b8b6 a4a5 b6b7 b1b3 e8e7 b4e7 b7e7');
INSERT INTO evaluations VALUES(29513,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(29513,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(29513,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(29513,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(29513,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(29513,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(29513,36,'cp 134','f2f3 b7b6 c3d5 d8d5 e2e4 d5d7 g1e2 e7e6 e2c3 f8b4 c1e3 e8g8 a2a3 b4c3 b2c3 f8d8 d1b3 c7c5 f1b5 d7c7 e1g1 a7a6 b5e2 b8c6 b3b2 c5d4 c3d4 b6b5 a3a4 b5b4 f1c1 c7b7 a4a5 h7h6 e3f2 a8b8 e2f1 c6e7 b2b3 d8c8 f2g3 b8a8 g3d6 e7c6 d4d5 e6d5 e4d5 c6a5 c1c8 a8c8 a1a5');
INSERT INTO evaluations VALUES(56272,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(56272,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(56272,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(56272,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56272,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56272,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56272,36,'cp 7','d2d4 c7c6 c1f4 g8f6 e2e3 c8f5 f1d3 f5g6 g1f3 b8d7 e1g1 f8e7 h2h3 g6d3 d1d3 e8g8 d3c2 f8e8 f3e1 d7b6 e1d3 f6e4 a2a4 e7d6 b2b3 a7a5 a1c1 d6f4 d3f4 e4c3 c2c3 b6d7 c3b2 e8e7 f4d3 h7h5');
INSERT INTO evaluations VALUES(39942,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(39942,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(39942,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(39942,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(39942,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(39942,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(39942,36,'cp 109','e4e5 f6d5 d1g4 e8f8 f1d3 d7d6 e1g1 b4c3 b2c3 d6e5 d4b5 b8c6 c3c4 d5f4 f1d1 h7h5 g4f3 f4d3 d1d3 d8f6 f3f6 g7f6 b5c7 a8b8 d3g3 c8d7 c1a3 c6e7 g3d3 d7a4 a1d1 a4c2 d3d8 f8g7 a3e7 c2d1 d8d1 b8c8 d1d7 f6f5 c4c5 g7g6 h2h4 f5f4 a2a4 a7a6 a4a5 g6f5');
INSERT INTO evaluations VALUES(39939,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(39939,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(39939,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(39939,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(39939,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(39939,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(39939,36,'cp 90','f6d5 d1g4 e8f8 f1d3 d7d6 e1g1 b4c3 b2c3 d6e5 d4b5 b8c6 c3c4 d5f4 f1d1 h7h5 g4f3 f4d3 d1d3 d8f6 f3f6 g7f6 b5c7 a8b8 d3g3 c8d7 c1a3 c6e7 a1d1 d7a4 g3d3 a4c2 d3d8 f8g7 d8h8 b8h8 d1c1 h8c8 c7b5 c2a4 b5c3 c8c4 a3e7 a4c6 f2f3 f6f5 g1f2 f7f6 c3e2 c4a4 a2a3 g7g6 e7b4 a7a5 b4d6 g6f7 c1b1');
INSERT INTO evaluations VALUES(39889,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(39889,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(39889,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(39889,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(39889,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(39889,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(39889,36,'cp 92','d1g4 e8f8 f1d3 d7d6 e1g1 b4c3 b2c3 d6e5 d4b5 b8c6 c3c4 d5f4 c1f4 h7h5 g4f3 e5f4 d3e4 h8h6 f3f4 d8e7 f1e1 e6e5 f4e3 f7f5 e4d5 c8d7 a1d1 a7a6 b5c7 a8c8 d5c6 d7c6 c7d5 c6d5 d1d5 h6e6 e1d1 f8g8 d5d7 e7f6 d7b7 c8c4 e3f3 e6c6 f3h5 c6c7 h5e8 g8h7 b7c7');
INSERT INTO evaluations VALUES(19963,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(19963,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(19963,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(19963,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(19963,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(19963,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(19963,36,'cp -31','c8g4 f1e2 g4f3 e2f3 g7g6 d2d3 e7e6 e1g1 f8g7 a1b1 g8e7 c1g5 c6d4 f3g4 e8g8 d1d2 d8c7 g5e3 e7c6 f2f4 f7f5 e4f5 d4f5 e3f2 c7d7 g4f3 f5d4 f3e4 a8b8 g2g3 a7a6 b1e1 b7b5 g1h1 c6b4 a2a3 b4c2 e1b1');
INSERT INTO evaluations VALUES(26337,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(26337,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(26337,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(26337,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(26337,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(26337,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(26337,36,'cp -17','f1e2 g7g6 d2d3 g4f3 e2f3 e7e6 e1g1 g8e7 c1g5 f8g7 f3g4 c6d4 f2f4 e8g8 g1h1 d8d7 f4f5 g6f5 g5e7 d7e7 e4f5 g7e5 g2g3 e7g5 c3e4 g5h6 a1b1 a8b8 d1d2 h6d2 e4d2 b7b5 c4b5 b8b5 f5e6 f7e6 d2c4 e5f6 b2b3 b5b4 h1g2');
INSERT INTO evaluations VALUES(26338,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(26338,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(26338,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(26338,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(26338,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(26338,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(26338,36,'cp -23','g7g6 d2d3 f8g7 a1b1 g4f3 e2f3 e7e6 e1g1 g8e7 f3g4 e8g8 c1g5 a8b8 g1h1 d8d7 d1d2 a7a6 c3e2 b7b5 g4h3 c6d4 e2d4 g7d4 b2b3 g8h8 g5h6 f8c8 a2a3 e7g8 h6e3 d4e3 d2e3 d7e7 f2f4 b5c4 d3c4 g8f6');
INSERT INTO evaluations VALUES(41354,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(41354,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(41354,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(41354,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(41354,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(41354,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(41354,36,'cp 77','d1d2 a7a6 e3h6 g7h6 d2h6 b8d7 g1f3 e7e5 e1c1 d8e7 f1c4 d7b6 c4b3 c8e6 d4e5 d6e5 b3e6 e7e6 d1d3 e6e7 h1d1 f6g4 h6g7 h8f8 a2a4 c7c6 d1d2 b6c4 d2e2 f7f6 g7e7 e8e7 b2b3 c4d6 a4a5 d6b5 c3a4 a8d8 a4c5 d8d3 c2d3 f8b8');
INSERT INTO evaluations VALUES(41355,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(41355,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(41355,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(41355,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(41355,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(41355,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(41355,36,'cp 88','c7c6 g1f3 e8g8 e3h6 b7b5 f1d3 g7h6 d2h6 e7e5 d4e5 f6g4 h6f4 d6e5 f4g3 f7f6 a2a4 b5b4 c3b1 d8e7 d3c4 g8h8 b1d2 g4h6 d2b3 h6f7 f3d2 f7d6 h2h4 d6c4 d2c4 c8e6 b3a5 e7c5 b2b3 c5d4 e1g1 d4e4 a1e1 e4f5 c4d6 f5g4 a5b7 g4g3 f2g3');
INSERT INTO evaluations VALUES(53998,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(53998,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(53998,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(53998,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(53998,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(53998,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(53998,36,'cp 75','d2d4 c5d4 f3d4 a7a6 g2g4 b7b5 a2a3 g8e7 h2h4 e7c6 f2f4 c6d4 d1d4 b8c6 d4f2 f8e7 g4g5 a8b8 f1g2 e8g8 c1e3 b5b4 a3b4 b8b4 b2b3 d8c7 f2d2 a6a5 e1g1 f7f5 e4f5 f8f5 g2e4 f5f7 h4h5 d6d5 e4h7 g8h7');
INSERT INTO evaluations VALUES(53993,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(53993,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53993,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(53993,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(53993,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(53993,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(53993,36,'cp 45','c5d4 f3d4 a7a6 g2g4 g8e7 h2h4 b7b5 a2a3 e7c6 c1e3 c6d4 d1d4 b8c6 d4d2 h7h6 f2f4 f8e7 e1c1 d8a5 c1b1 a8b8 c3a2 b5b4 a3b4 c6b4 a2b4 a5b4 d2b4 b8b4 d1d4 a6a5 g4g5 h6g5 f4g5 b4d4 e3d4 e8g8 f1d3 e6e5 d4e3 c8e6');
INSERT INTO evaluations VALUES(54027,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(54027,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(54027,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(54027,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(54027,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(54027,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(54027,36,'cp 44','f3d4 a7a6 g2g4 b7b5 a2a3 g8e7 h2h4 e7c6 c1e3 c6d4 d1d4 b8c6 d4d2 a8b8 g4g5 d8a5 a1b1 b5b4 a3b4 a5b4 b2b3 b4a5 c3d1 a5c7 h4h5 c8b7 b1a1 h7h6 f1a6 h6g5 a6b7 c7b7 e3g5 f8e7 g5e7 c6e7');
INSERT INTO evaluations VALUES(54020,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(54020,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(54020,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(54020,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(54020,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(54020,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(54020,36,'cp 47','a7a6 g2g4 b7b5 a2a3 g8e7 h2h4 e7c6 c1e3 c6d4 d1d4 b8c6 d4d2 f8e7 g4g5 d8a5 a1d1 a8b8 f2f4 h7h6 f1g2 h6g5 h4g5 h8h1 g2h1 b5b4 a3b4 a5b4 b2b3 e7d8 e1f2 c8d7 h1f3 d8a5 c3b1 b4d2 b1d2 a5b4 d2c4 e8e7 c4b6 a6a5 b6d7 e7d7 f4f5 c6e5 f3e2');
INSERT INTO evaluations VALUES(50765,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(50765,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(50765,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50765,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(50765,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(50765,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(50765,36,'cp 64','g2g4 g8e7 h2h4 b7b5 a2a3 e7c6 c1e3 c6d4 d1d4 b8c6 d4d2 a8b8 f2f4 f8e7 g4g5 d8a5 a1d1 b5b4 a3b4 a5b4 b2b3 h7h6 f1g2 h6g5 h4g5 h8h1 g2h1 e7d8 e1f2 d8a5 c3b1 a5b6 h1f3 e8e7 e3b6 b8b6 d2e3 b4c5 b1d2 c5e3 f2e3');
INSERT INTO evaluations VALUES(16361,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(16361,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16361,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(16361,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(16361,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(16361,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16361,36,'cp -36','h7h6 g5f6 d8f6 g1e2 c6a5 b1d2 a5c4 d2c4 f8c5 e1g1 d7d6 c4e3 c7c6 d1d2 e8g8 c2c3 c5b6 f2f4 c8g4 a1e1 g4e2 f4e5 f6e6 d2e2 d6e5 a2a3 f8d8 f1f3 d8d7 g1h1 b6e3 e2e3 a7a5 e1f1 a5a4 h2h3 a8d8');
INSERT INTO evaluations VALUES(16111,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(16111,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(16111,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(16111,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(16111,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(16111,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(16111,36,'cp -36','g5f6 d8f6 g1e2 f8c5 e1g1 c6a5 b1d2 a5c4 d2c4 d7d6 c4e3 c7c6 c2c3 c5e3 f2e3 f6g5 d1d2 c8e6 e2g3 g7g6 c3c4 a8d8 f1f3 h6h5 h2h3 d6d5 c4d5 c6d5 d2b4 g5e7 b4c3 f7f6 d3d4 h5h4 d4e5 h4g3 e5f6 e7h7 a1f1 d5e4');
INSERT INTO evaluations VALUES(16096,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(16096,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(16096,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(16096,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(16096,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(16096,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(16096,36,'cp -29','d8f6 g1e2 c6a5 b1d2 a5c4 d2c4 f8c5 e1g1 d7d6 c2c3 c8e6 d1a4 c7c6 b2b4 e6c4 d3c4 c5b6 a4b3 e8g8 a2a4 a7a6 a1d1 a8d8 a4a5 b6a7 g1h1 f6e6 f2f3 g7g6 f1e1 a7f2 e1f1 f2h4 e2g1 f7f5 b3a2 f5f4 a2e2 h6h5 b4b5 g8g7 b5a6 b7a6 e2d3 h4e7 g1e2');
INSERT INTO evaluations VALUES(1044,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(1044,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(1044,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(1044,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(1044,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(1044,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(1044,36,'cp -15','g1e2 c6a5 b1d2 a5c4 d2c4 f8c5 e1g1 d7d6 c2c3 c8e6 d1a4 c7c6 b2b4 e6c4 d3c4 c5b6 a4b3 e8g8 a2a4 a7a6 a1d1 a6a5 d1d3 h6h5 e2g3 g7g6 d3f3 f6e7 f1b1 a8a7 g3f1 g8g7 h2h3 f8a8 b4b5 a8c8 f1d2 b6c5 b3c2 a7a8 d2b3 c5b6 c2e2 c6b5 c4b5');
INSERT INTO evaluations VALUES(26206,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(26206,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(26206,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(26206,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(26206,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26206,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26206,36,'cp 21','g8f6 d1e2 a7a6 f1d1 e7e6 d2d4 c5d4 f3d4 d8a5 c2c4 f8e7 b2b3 e8g8 c1b2 a5h5 e2h5 f6h5 b1c3 h5f6 d1d2 d7c5 f2f3 f8c8 a1d1 c8c7 d4c2 h7h5 g2g3 a8b8 c3e2 c5d7 e2d4 b8c8 g1g2 b7b5 c4b5 a6b5 c2b4');
INSERT INTO evaluations VALUES(24574,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(24574,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(24574,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(24574,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24574,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24574,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(24574,36,'cp 33','d1e2 a7a6 f1d1 e7e6 b2b3 d8c7 c1b2 f8e7 c2c4 d7e5 d2d4 e5f3 e2f3 e8g8 a2a4 c5d4 b2d4 f6d7 b1c3 e7f6 d4f6 d7f6 f3e3 a8c8 h2h3 f8d8 a1c1 f6d7 e3g3 d7c5 c1b1 c7c6 g3e3 c5d7 d1d2 c6c7 g1h1 h7h6 f2f4 d7c5');
INSERT INTO evaluations VALUES(24576,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24576,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(24576,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(24576,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(24576,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(24576,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(24576,36,'cp 9','a7a6 f1d1 e7e6 b2b3 f8e7 d2d4 c5d4 f3d4 a8c8 c2c4 d8a5 g2g3 a5e5 f2f3 e5a5 g1h1 h7h5 c1d2 a5e5 d2e3 e5a5 b1d2 e8g8 e3g1 f8e8 a2a4 g7g6 a1c1 e7f8 e2f1 f8g7 c1b1 a5c7 d4c2');
INSERT INTO evaluations VALUES(21341,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(21341,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(21341,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(21341,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(21341,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(21341,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(21341,36,'cp 75','b7c6 b1c3 g8f6 f1d3 a8b8 e1g1 g7g6 b2b3 f8g7 c1g5 d8e7 h2h3 h7h6 g5d2 f6h5 d1e2 e8g8 a1d1 f8e8 f1e1 c8b7 e2e3 a7a6 e3e2 a6a5 d3c4 d6d5 e4d5 e7e2 e1e2 e8e2 c3e2');
INSERT INTO evaluations VALUES(18281,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(18281,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(18281,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(18281,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(18281,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(18281,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(18281,36,'cp 82','b1c3 g8f6 f1d3 a8b8 e1g1 g7g6 b2b3 f8g7 c1g5 h7h6 g5d2 f6g4 d1e2 a7a5 d3a6 e8g8 a1e1 g4f6 f2f4 f8e8 a6c8 d8c8 e2f3 c8g4 f3d3 e8e7 h2h3 g4d7 e1e2 b8e8 f1e1 d7d8 d3f3 e7e6 g1h1 d8a8 f3d3 c6c5 h1g1');
INSERT INTO evaluations VALUES(18279,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(18279,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(18279,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(18279,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(18279,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(18279,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(18279,36,'cp 62','g8f6 f1d3 a8b8 e1g1 g7g6 b2b3 f8g7 c1g5 d8e7 h2h3 e7e5 g5d2 f6h5 d1f3 e8g8 c3e2 c8b7 a1e1 c6c5 d2c3 e5g5 f3g4 g5g4 h3g4 h5f4 c3g7 f4d3 g7f8 d3e1 f1e1 g8f8 f2f3 a7a5 e2c3 b7c6 g4g5 a5a4 c3a4 c6a4 b3a4');
INSERT INTO evaluations VALUES(36555,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(36555,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(36555,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(36555,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36555,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36555,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(36555,36,'cp 37','c7c6 b1c3 c8f5 f1e1 f6e4 d1b3 e4c3 b2c3 d8c8 e2e4 f5g4 c1g5 f8e8 f3d2 b7b6 g5f4 e7e5 f4e3 g4h3 g2h3 c8h3 b3a3 e5d4 e3d4 g7d4 c3d4 c6c5 a3e3 b8c6 d4c5');
INSERT INTO evaluations VALUES(35905,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(35905,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(35905,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(35905,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(35905,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(35905,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(35905,36,'cp 32','b1c3 c8f5 d1b3 d8b6 f1e1 b8a6 c1g5 f8e8 e2e4 f5g4 g5e3 b6b3 a2b3 f6d7 h2h3 g4f3 g2f3 e8d8 e1e2 c6c5 d4d5 g7d4 f3g4 d7f6 e3f4 a6b4 a1d1 f6d7 f4g5 h7h5 g4f3 f7f6 g5f4 d7e5 f3g2 e5d3 e2d2 d3f4');
INSERT INTO evaluations VALUES(35903,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35903,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(35903,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(35903,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(35903,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(35903,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(35903,36,'cp 25','c8f5 f1e1 f6e4 d1b3 e4c3 b2c3 f5c8 e2e4 b8d7 c1e3 d8c7 h2h4 e7e5 a2a4 c6c5 a4a5 h7h6 b3a3 a8b8 a3c1 h6h5 c1a3 b7b6 d4d5 d7f6 f3g5 c7e7 e3d2 g7h6 e1b1 f6h7 a3c1 b8b7 d2e3 h6g5 h4g5 f7f6 a5b6 a7b6');
INSERT INTO evaluations VALUES(29560,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(29560,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(29560,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(29560,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29560,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(29560,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(29560,36,'cp -7','f1d3 e7e6 g1f3 a7a6 e1g1 f5g6 h2h3 c7c5 d4c5 b8d7 b2b4 a6a5 b4a5 a8a5 c3b5 d7c5 a2a4 f8e7 d3g6 h7g6 c2c4 e8g8 c1d2 a5a8 d2b4 d5c4 d1d8 f8d8 b4c5 e7c5 f1c1 a8c8 c1c4 c5f8 a1c1 c8c4 c1c4 d8d1 g1h2 d1a1 c4c8 a1a4 b5d6 f6d5 d6b7');
INSERT INTO evaluations VALUES(29570,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(29570,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(29570,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(29570,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(29570,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(29570,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(29570,36,'cp -13','e7e6 g1f3 a7a6 d3f5 e6f5 c3e2 c7c6 e1g1 f8d6 b2b3 e8g8 a2a4 d8e7 c1b2 b8d7 c2c4 f8e8 d1c1 g7g5 b2a3 d6a3 c1a3 e7a3 a1a3 f5f4 f3g5 f4e3 e2g3 e3f2 g1f2 c6c5 d4c5 d5c4 b3c4 h7h6 g5h3 f6g4 f2g1 g4e3 f1f3 e3c2 a3c3 c2d4 f3f4 d4e6 f4f5 e6c5 h3f4 e8e1 g1f2');
INSERT INTO evaluations VALUES(30120,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(30120,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(30120,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(30120,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(30120,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(30120,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(30120,36,'cp -21','g1f3 a7a6 e1g1 f5g6 d1e1 g6h5 g1h1 f8e7 f3e5 b8d7 e3e4 d7e5 d4e5 f6e4 c3e4 d5e4 d3e4 c7c6 c1d2 h5g6 d2c3 e8g8 h2h3 f8e8 f2f4 g6e4 e1e4 d8d5 e4d5 c6d5 c3d4 e8c8 c2c3');
INSERT INTO evaluations VALUES(38673,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(38673,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38673,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38673,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(38673,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(38673,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38673,36,'cp 10','g8f6 e2e3 c7c5 d4c5 d6c5 b1d2 e8g8 c2c4 d8b6 d1c2 c5e7 f1e2 c8d7 e1g1 f8c8 a1c1 a7a5 f3e5 d7e8 c2b1 b6a6 b2b3 b8d7 e5d7 e8d7 e2d3 d5c4 d2c4 b7b5 c4e5 d7e8 h2h3 c8c1 f1c1 a6b7');
INSERT INTO evaluations VALUES(41817,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41817,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(41817,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(41817,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(41817,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(41817,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41817,36,'cp 20','e2e3 c7c5 d4c5 d6c5 b1d2 e8g8 c2c4 d8b6 d1c2 c8d7 a1c1 c5e7 c4d5 e6d5 g3c7 b6b4 f1d3 f8c8 a2a3 b4a4 c2b1 d7b5 e1g1 b5d3 b1d3 a4a6 d3a6 b8a6 c7a5 a6c5 a5b4 a7a5 b4c5 c8c5 c1c5 e7c5 d2b3 c5d6 b3d4');
INSERT INTO evaluations VALUES(12373,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(12373,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12373,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(12373,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12373,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12373,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(12373,36,'cp 12','f1e1 f6e4 e1e4 d7d5 c4d5 d8d5 b1c3 d5f5 d2d4 c5d4 f3d4 c6d4 e4d4 f8e7 c1e3 e8g8 d4f4 f5e5 d1f3 e5c7 f4d4 e7f6 d4d3 f6e5 f3h5 b7b6 a1d1 c8b7 d3d7 c7c6 d7b7 e5c3 d1d7 c3b2 d7f7 c6c2 h2h3 c2b1 g1h2 b1e4 f7d7 f8f5 h5g4 e4g4 h3g4 f5a5 g4g5 a5a2 f2f4 a2a5 d7e7 a5d5 e7e6 a7a5 e6b6 b2d4 b7b8 a8b8 b6b8 g8f7 e3d4 d5d4');
INSERT INTO evaluations VALUES(12374,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(12374,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(12374,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12374,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(12374,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12374,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(12374,36,'cp 12','f6e4 e1e4 d7d5 c4d5 d8d5 b1c3 d5f5 d2d4 c5d4 f3d4 c6d4 e4d4 f8e7 c1e3 e8g8 d4f4 f5e5 d1f3 e5c7 f4d4 e7f6 d4d3 c8d7 e3f4 c7c8 f3h5 e6e5 f4g5 d7g4 h5h4 f6g5 h4g5 f7f6 g5d2 f8d8 c3d5 g8h8 h2h3 g4h5 b2b3 c8c5 c2c4 h5g6 d3c3 d8d7 a1d1 a8d8 b3b4');
INSERT INTO evaluations VALUES(43506,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(43506,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(43506,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(43506,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43506,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(43506,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43506,36,'cp -6','f8e7 c4b3 b7b5 a2a3 e8g8 f2f4 c8b7 f4f5 f6e4 g5e7 d8e7 e1g1 e4c3 b2c3 b8d7 f5e6 f7e6 f1e1 e7f6 d1d2 d7c5 d4e6 c5b3 c2b3 f8f7 e6g5 f7c7 e1e6 f6c3 a1e1 a8f8 d2d6 c3c5 d6c5 c7c5 e6b6 c5g5 b6b7 h7h6 e1e7 f8f4 b7d7 g8h7 e7f7 f4f7 d7f7 g5g6 g1f2 g6c6 f7a7 c6d6 f2e2 d6g6 e2f3 g6c6');
INSERT INTO evaluations VALUES(39697,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(39697,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(39697,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(39697,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39697,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(39697,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(39697,36,'cp -7','c4b3 e8g8 f2f4 b7b5 a2a3 c8b7 f4f5 f6e4 g5e7 d8e7 e1g1 e4c3 b2c3 b8d7 f5e6 f7e6 f1e1 e7f6 d1d2 d7c5 d4e6 c5b3 c2b3 f8f7 e6g5 f7c7 e1e6 f6c3 a1e1 a8f8 d2d6 c3c5 d6c5 c7c5 e6b6 c5g5 b6b7 h7h6 b7e7 g8h7 e7e3 f8c8 e1e2 c8c6 g1f2 g5c5 h2h3 a6a5 b3b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(39700,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(39700,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(39700,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(39700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39700,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(39700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39700,36,'cp 0','b7b5 a2a3 e8g8 f2f4 c8b7 f4f5 f6e4 g5e7 d8e7 e1g1 e4c3 b2c3 b8d7 f5e6 f7e6 f1e1 e7f6 e1e2 d7c5 d4e6 c5b3 c2b3 f8f7 d1d6 a8e8 a1e1 f6c3 h2h3 c3f6 d6d2 f7e7 e6c5 e7e2 e1e2 e8e2 d2e2 f6d4 e2f2 d4d1 f2f1 d1d6 c5b7 d6b6 f1f2 b6b7 f2c5 b7e4 c5c8 g8f7 c8a6 e4e3 g1h2 e3e5 h2h1 e5e1 h1h2');
INSERT INTO evaluations VALUES(39800,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39800,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39800,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39800,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39800,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39800,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39800,36,'cp 0','a2a3 e8g8 f2f4 c8b7 f4f5 f6e4 g5e7 d8e7 e1g1 e4c3 b2c3 b8d7 f5e6 f7e6 f1e1 e7f6 e1e2 d7c5 d4e6 c5b3 c2b3 f8f7 d1d6 a8e8 a1e1 f6c3 h2h3 c3f6 d6d2 f7e7 e6c5 e7e2 e1e2 e8e2 d2e2 f6d4 e2f2 d4a1 f2f1 a1d4');
INSERT INTO evaluations VALUES(26987,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(26987,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(26987,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(26987,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(26987,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(26987,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(26987,36,'cp 122','h2h3 g4h5 d4d5 c6b8 c1e3 h5f3 d1f3 b8d7 a2a4 f8e7 a4a5 a7a6 f3g3 g7g6 c4e2 e7h4 g3f3 h4g5 b1d2 g5e3 f3e3 d8e7 e1c1 f7f5 g2g4 g8f6 f2f3 e8g8 h3h4 f5f4 e3c3 f8c8 c1b1 g8g7 h4h5 g6g5 h5h6 g7h8 d2c4 c8b8 c3a3 f6e8 b1a2 b7b5 a5b6 c7b6 b2b3 b6b5 c4a5');
INSERT INTO evaluations VALUES(26986,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(26986,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(26986,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(26986,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(26986,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(26986,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(26986,36,'cp 119','g4h5 d4d5 c6b8 c1e3 h5f3 d1f3 b8d7 b1d2 g7g6 e1c1 a7a6 c4e2 f8e7 a2a4 e7g5 a4a5 h7h6 g2g3 g8f6 h3h4 g5e3 f3e3 d8b8 g3g4 b8a7 e3a7 a8a7 g4g5 f6h5 e2h5 g6h5 d1g1 h6g5 h4g5 e8e7 g5g6 d7f6 f2f3 f7g6 g1g6');
INSERT INTO evaluations VALUES(26953,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(26953,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(26953,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(26953,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(26953,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(26953,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(26953,36,'cp 117','d4d5 c6b8 c1e3 h5f3 d1f3 b8d7 a2a4 g7g6 b1d2 a7a6 a4a5 f8h6 e3h6 g8h6 f3e3 h6g8 c4e2 d8e7 e1c1 f7f5 g2g4 g8f6 f2f3 e8g8 h3h4 f5f4 e3c3 f8c8 c1b1 g8g7 d2b3 h7h6 c3d2 c8b8 d1g1 c7c5 d5c6 b7c6 g4g5 h6g5 g1g5 f6h7 g5g4 d7f6 g4g1');
INSERT INTO evaluations VALUES(41397,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41397,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(41397,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(41397,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(41397,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(41397,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(41397,36,'cp 48','c7c5 d4c5 d8a5 f1d3 a5c5 d1e2 e8g8 c1e3 c5a5 e1g1 c8g4 e2f2 g4f3 f2f3 b8c6 g1h1 f6d7 e3d2 d7c5 d3c4 c6d4 f3d1 a5c7 c3d5 c7d8 c2c3 d4c6 d1f3 e7e6 d5e3 a7a6 c4e2 b7b5 a1d1 c5a4 d2c1 a4c5 h2h3 b5b4 c3b4');
INSERT INTO evaluations VALUES(40350,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(40350,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40350,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(40350,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(40350,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(40350,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(40350,36,'cp 43','d4c5 d8a5 f1d3 a5c5 d1e2 b8c6 c1e3 c5a5 e1g1 c8g4 g1h1 e8g8 e2f2 g4f3 f2f3 f6d7 e3d2 d7c5 d3c4 a5c7 c3d5 c7d8 d2c3 e7e6 c3g7 g8g7 d5c3 a7a6 a2a4 d8e7 a1d1 g7g8 h2h3 a8e8 b2b3 c6b4 f3f2 b4c6');
INSERT INTO evaluations VALUES(51920,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(51920,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(51920,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(51920,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(51920,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(51920,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(51920,36,'cp -25','d7d5 e4d5 e6d5 c4b5 b8c6 d2d4 c5d4 c3d4 f8d6 b1c3 g8e7 g1e2 e8g8 c1f4 c8g4 f2f3 g4f5 e1g1 c6a5 f4g5 f5e6 e2f4 d6f4 g5f4 d8b6 f1e1 e7f5 g1h1 b6d4 d1d4 f5d4 a1d1 d4b5 c3b5 h7h6 b5c3 f8e8 c3d5 e6d5 e1e8 a8e8');
INSERT INTO evaluations VALUES(54153,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(54153,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(54153,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(54153,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54153,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(54153,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(54153,36,'cp -29','e4d5 e6d5 c4b5 b8c6 d2d4 c5d4 c3d4 f8d6 b1c3 g8e7 g1e2 e8g8 c1f4 c8g4 f4d6 d8d6 d1d2 g4e2 c3e2 e7f5 e1g1 g7g6 b5c6 b7c6 e2g3 f5g7 f1c1 f8b8 c1c3 a7a5 g3e2 b8b6 a1c1 a8e8 b2b3 d6b4 e2g3 g7e6 g3e2 e8e7 h2h3 b6b5');
INSERT INTO evaluations VALUES(54118,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(54118,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(54118,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(54118,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(54118,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(54118,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(54118,36,'cp -21','e6d5 c4b5 b8c6 d2d4 c5d4 c3d4 f8d6 b1c3 g8e7 g1e2 e8g8 c1f4 c8g4 f4d6 d8d6 e1g1 g4e2 c3e2 g7g6 a1c1 a8c8 b5c6 c8c6 c1c6 e7c6 d1d2 g8g7 h2h3 f8e8 e2c3 a7a6 f1d1 h7h6 d1c1 e8e6 c1d1 c6e7 d1e1 e6e1 d2e1 e7c6 e1d2');
INSERT INTO evaluations VALUES(54497,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(54497,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(54497,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(54497,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(54497,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(54497,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(54497,36,'cp -18','c4b5 b8c6 d2d4 c5d4 c3d4 f8d6 b1c3 g8e7 g1e2 e8g8 c1f4 c8g4 f4d6 d8d6 e1g1 g4e2 c3e2 a8d8 b5c6 e7c6 e2c3 f8e8 d1d2 g7g6 f1e1 g8g7 e1e8 d8e8 a1d1 h7h5 h2h3 e8d8 d1e1 d6b4 e1d1 d8d6 a2a3 b4c4 d2d3 h5h4 d3e3 d6d7');
INSERT INTO evaluations VALUES(54470,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(54470,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(54470,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(54470,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(54470,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(54470,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(54470,36,'cp -35','b8c6 g1f3 f8d6 d2d4 c5d4 f3d4 g8e7 b1d2 a7a6 b5a4 b7b5 d4c6 e7c6 a4c2 d8e7 d1e2 c6e5 h2h3 e8g8 e1g1 f8e8 d2b3 c8d7 b3d4 e7f6 e2d2 e5c4 d2d3 g7g6 f1d1 b5b4 d4f3 d7f5 d3d4 f6d4 f3d4 b4c3 b2c3');
INSERT INTO evaluations VALUES(20477,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(20477,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(20477,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(20477,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(20477,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(20477,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(20477,36,'cp -6','g1f3 f8d6 d2d4 g8f6 d4c5 d6c5 e1g1 h7h6 d1e2 c8e6 c1e3 c5e3 e2e3 e8g8 b1d2 f8e8 b5c6 b7c6 f1e1 d8c7 b2b4 a7a5 e3e5 c7e7 f3d4 a5b4 d4c6 e7c5 c6b4 a8c8 a2a4 e6d7 e5d4 e8e1 a1e1 d7a4 d4c5 c8c5 e1e3');
INSERT INTO evaluations VALUES(21024,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(21024,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(21024,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(21024,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21024,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21024,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21024,36,'cp 23','c8f5 a2a3 g8f6 e2e3 e7e6 c2c4 f6e4 b2b4 f8e7 b1d2 e8g8 c1b2 e4d2 f3d2 c6b8 f1e2 b8d7 c4c5 c7c6 e1g1 h7h6 d2b3 a7a6 e2d3 f5d3 d1d3 f8e8 b2c3 e7f6 f2f4 f6h4 c3e1 h4e1 a1e1 d7f6 g1h2');
INSERT INTO evaluations VALUES(26568,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(26568,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(26568,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(26568,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(26568,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26568,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26568,36,'cp 21','a2a3 g8f6 e2e3 e7e6 c2c4 f6e4 b2b4 f8e7 f3d2 e8g8 d2e4 f5e4 b1c3 e4g6 c4d5 e6d5 f1e2 c6b8 b4b5 c7c6 e1g1 b8d7 a3a4 a7a6 b5c6 b7c6 a4a5 e7b4 c1b2 b4a5 d1a4 a5c3 b2c3 d8g5 e2f3 g6d3 f1e1 f8c8 c3b4 g8h8 e1c1 g5g6 g1h1');
INSERT INTO evaluations VALUES(26577,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26577,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(26577,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(26577,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(26577,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(26577,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(26577,36,'cp 18','g8f6 e2e3 e7e6 c2c4 f6e4 f3d2 e4d2 b1d2 a7a5 d2f3 f8e7 c1d2 e8g8 f1e2 e7f6 d2c3 d5c4 e2c4 c6e7 d1e2 e7d5 e3e4 d5c3 b2c3 f5g6 e1g1 c7c5 f1d1 c5d4 c3d4 a8c8 a1c1 d8b6 a3a4 f8d8 d4d5 e6d5 c4d5 c8c1 d1c1');
INSERT INTO evaluations VALUES(24872,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24872,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(24872,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24872,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24872,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24872,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24872,36,'cp 26','e2e3 e7e6 c2c4 f6e4 d1b3 a8b8 b1c3 d5c4 b3c4 f8e7 c1d2 e8g8 c3e4 f5e4 f1d3 d8d5 c4d5 e6d5 e1e2 f7f6 h1c1 e4d3 e2d3 g8f7 b2b4 a7a6 a1b1 e7d6 b4b5 a6b5 b1b5 c6e7 d2b4 f7e6 c1b1 b7b6 b4d6 e6d6 a3a4');
INSERT INTO evaluations VALUES(24870,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24870,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(24870,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(24870,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(24870,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(24870,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24870,36,'cp 27','f6e4 b1d2 e7e6 d2e4 f5e4 c2c4 d5c4 f1c4 d8f6 c4d3 e4d3 d1d3 f6g6 d3g6 h7g6 e1e2 a7a5 a3a4 f7f6 c1d2 g6g5 h3h4 g5h4 h1h4 h8h4 f3h4 f8b4 d2b4 c6b4 h4f3 e8e7 a1h1 b4d5 f3d2 b7b6 g2g4 d5b4 h1c1 b4d5 d2c4 e7d7');
INSERT INTO evaluations VALUES(43871,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(43871,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(43871,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(43871,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(43871,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(43871,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(43871,36,'cp 47','f2f3 b8c6 b1c3 d7d5 d1a4 d8d7 c3d5 f6d5 c4d5 d7d5 e2e4 d5d6 f1b5 c8d7 d4d5 d6b4 e1f2 b4a4 b5a4 c6d4 a4d7 e8d7 c1f4 c7c5 d5c6 d4c6 a1d1 d7e8 g1e2 e7e6 d1c1 f8c5 f4e3 c5e3 f2e3 a8c8 c1c3 e8e7 h1c1 e7d7 f3f4 c6b4 a2a3 c8c3 e2c3 b4c6 e4e5 f7f6 e5f6 g7f6 g2g4 h8d8 c1d1 d7e7 d1d8');
INSERT INTO evaluations VALUES(20955,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(20955,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(20955,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(20955,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(20955,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(20955,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(20955,36,'cp 58','d8d5 e2e3 e7e5 b1c3 f8b4 a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 c7c6 f3e5 e8g8 c1b2 c6d5 c4d5 d8a5 d1d2 a5d2 e1d2 f6d5 f1d3 f8d8 h1c1 c8e6 h2h3 a7a6 d2e2 f7f6 e5f3 e6f7 g2g4 f7g6 d3c4 g6f7 c4b3 a8c8 c1c8');
INSERT INTO evaluations VALUES(1607,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(1607,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(1607,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(1607,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(1607,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(1607,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(1607,36,'cp 46','e2e3 e7e5 b1c3 f8b4 a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 e8g8 f3e5 c7c6 c1b2 c6d5 c4d5 d8a5 d1d2 a5d2 e1d2 f6d5 f1d3 f8d8 h1c1 c8e6 d2e2 a8c8 c1c8 d8c8 a1c1 c8c1 b2c1 f7f6 e5f3 e7c6 c1b2 e6f7 f3d2 d5b6 d3c2 c6e5 b2d4 f7e6 d4b6 a7b6 c2b3 e6b3');
INSERT INTO evaluations VALUES(1604,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(1604,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(1604,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(1604,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(1604,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(1604,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(1604,36,'cp 49','e7e5 b1c3 f8b4 a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 e5e4 f3d4 c7c6 d5c6 e7c6 f1e2 c6e5 a3a4 c8g4 e2g4 f6g4 e1g1 e8g8 c1a3 f8e8 h2h3 g4f6 d1b3 b7b6 d4b5 d8d3 b3d3 e4d3 b5d6 e8e6 f1d1 f6e8 c4c5 e8d6 c5d6');
INSERT INTO evaluations VALUES(1716,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(1716,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(1716,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(1716,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(1716,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(1716,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(1716,36,'cp 43','b1c3 f8b4 a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 e8g8 f3e5 c7c6 f1e2 f6e4 d1b3 c6d5 e1g1 d8c7 e5f3 d5c4 e2c4 c8g4 f3d2 e4c5 b3c2 g4e6 c1b2 e6c4 d2c4 a8d8 a1d1 d8d1 f1d1');
INSERT INTO evaluations VALUES(1710,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(1710,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(1710,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(1710,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(1710,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(1710,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(1710,36,'cp 58','f8b4 a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 c7c6 f3e5 c6d5 c4d5 e8g8 f1c4 d8a5 d1d2 a5d2 c1d2 e7d5 e1g1 c8e6 f2f3 f8d8 f1c1 d5b6 c4e6 f7e6 d2b4 a8c8 b4a5 f6d7 e5d7 d8d7 a5b6 c8c1 a1c1 a7b6 c1b1 d7d6 g1f2 h7h6 b1b4 d6d2 f2g3 d2d3 e3e4 d3a3 b4b6');
INSERT INTO evaluations VALUES(178,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(178,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(178,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(178,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(178,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(178,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(178,36,'cp 62','a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 c7c6 f3e5 c6d5 c4d5 e8g8 f1c4 d8a5 d1d2 a5d2 c1d2 e7d5 e1g1 c8e6 f2f3 f8d8 f1c1 d5b6 c4e6 f7e6 d2b4 f6d7 e5d7 d8d7 g1f2 a8c8 c1c8 b6c8 a1c1 c8e7 b4e7 d7e7 c1c8 g8f7 h2h4 h7h6 f2g3 f7f6 e3e4 e7d7 c8b8 b7b6 b8a8 f6e7 a3a4');
INSERT INTO evaluations VALUES(47365,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(47365,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(47365,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(47365,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(47365,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(47365,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(47365,36,'cp 131','g1f3 e7e6 e2e4 f8e7 c1f4 e8g8 d1c2 c7c5 d4c5 d8a5 f1e2 b8c6 e1g1 a5c5 a2a3 e6e5 b2b4 c5b6 f3e5 c6d4 c2b2 d4e2 b2e2 c8e6 a1d1 f8e8 d1c1 a7a5 f4e3 b6d6 e5f3 e6g4 e2b5 g4f3 g2f3 d6e6 f1d1 a5b4 a3b4');
INSERT INTO evaluations VALUES(47359,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(47359,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(47359,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(47359,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(47359,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(47359,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(47359,36,'cp 133','c7c5 d4d5 e7e6 e2e4 e6d5 e4d5 a7a6 c1f4 f8d6 d1e2 e8f8 e2e3 h7h6 f1e2 b7b5 f3d2 d6f4 e3f4 b8d7 a2a4 b5b4 c3d1 d7b6 d5d6 f6d5 f4g3 c8b7 d1e3 d8g5 e1g1 g5g3 h2g3 d5e3 f2e3 b6d7 a4a5 a8e8 d2c4 g7g6 e2f3');
INSERT INTO evaluations VALUES(20004,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(20004,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(20004,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(20004,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(20004,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(20004,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(20004,36,'cp -14','g7g6 d3d4 c5d4 f3d4 g8f6 f1c4 c6d4 d1d4 f8g7 e1g1 e8g8 a2a4 c8d7 f1d1 d7c6 d4d3 a8c8 c4b3 f6d7 c1e3 d8a5 e3d4 g7d4 d3d4 a5c5 a4a5 c8c7 f2f3 f8c8 d4c5 d7c5 b3c4 g8f8 d1d4 c5d7 c4f1 h7h5 a5a6 b7a6 a1a6 f8g7');
INSERT INTO evaluations VALUES(19671,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(19671,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(19671,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19671,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(19671,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(19671,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(19671,36,'cp -11','d3d4 c5d4 f3d4 g8f6 f1c4 f8g7 e1g1 e8g8 h2h3 c6d4 d1d4 a7a6 a2a4 c8d7 d4d3 d7c6 f1d1 a8c8 c1e3 f6d7 d3d2 b7b5 a4b5 a6b5 c3b5 g7b2 a1a7 c6b5 c4b5 d7c5 b5d3 b2f6 d2a5 d8a5 a7a5 c8c7 d1b1 g8g7 a5b5 c5d3 c2d3 f8c8');
INSERT INTO evaluations VALUES(34916,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(34916,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(34916,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(34916,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(34916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34916,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34916,36,'cp 0','d5e6 d1e2 b8c6 d2d4 e6e2 f1e2 f8d6 c3b5 c6b4 e1d1 c8f5 c1d2 b4d5 b5d6 c7d6 h1e1 e8c8 c2c4 d5e3 d2e3 f4e3 e2f1 g8f6 e1e3 h8e8 e3e8 d8e8 d1d2 f6e4 d2c2 e4f2 c2d2');
INSERT INTO evaluations VALUES(34877,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(34877,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(34877,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(34877,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34877,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(34877,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(34877,36,'cp 31','f1e2 f8d6 e1g1 g8e7 f3g5 e6h6 c3e4 e8g8 d2d4 e7g6 e2c4 b8c6 c2c3 c6d8 e4d6 h6g5 d6c8 a8c8 d1f3 d8e6 f3b7 c7c6 c1d2 c6c5 b7a6 g5f5 a1e1 c5d4 c4e6 f7e6 a6e6 f5e6 e1e6 c8b8 b2b3 d4c3 d2c3 b8c8 c3b4 f8d8 f1e1 c8c2');
INSERT INTO evaluations VALUES(34878,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34878,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(34878,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(34878,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(34878,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(34878,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(34878,36,'cp 25','f8d6 e1g1 g8e7 d2d4 e8g8 f3g5 e6h6 c3e4 b8c6 e2c4 e7g6 c2c3 c6d8 d1f3 d8e6 g5h3 d6e7 c1f4 e6f4 h3f4 g6f4 f3f4 h6f4 f1f4 c8e6 c4e6 f7e6 f4f2 f8f2 g1f2 a8d8 a1d1 g8f7 c3c4 h7h5 f2e3 c7c6 h2h3 b7b6 g2g4 h5g4 h3g4 d8h8 d1f1 f7g6');
INSERT INTO evaluations VALUES(33074,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(33074,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(33074,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(33074,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(33074,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(33074,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(33074,36,'cp 35','e1g1 g8e7 d2d4 e6h6 c3e4 e8g8 f3g5 e7g6 e2c4 b8c6 c2c3 c6d8 d1f3 d8e6 g5h3 d6e7 c1f4 e6f4 h3f4 g6f4 f3f4 h6f4 f1f4 c8e6 c4e6 f7e6 f4f2 f8f2 g1f2 a8d8 a1d1 d8d5 f2e2 g8f7 c3c4 d5f5 g2g4 f5f4 h2h3 f7e8 d1f1 f4f1 e2f1 b7b6 b2b3 h7h6 f1e2 e8d7 e2f3');
INSERT INTO evaluations VALUES(33073,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(33073,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(33073,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(33073,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(33073,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(33073,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(33073,36,'cp 15','g8e7 d2d4 e8g8 f3g5 e6h6 c3e4 e7g6 e2c4 b8c6 c2c3 c6d8 d1f3 d8e6 g5h3 d6e7 c1f4 g6f4 h3f4 e6f4 f3f4 h6f4 f1f4 c8e6 c4e6 f7e6 f4f2 f8f2 g1f2 a8d8 a1d1 g8f7 f2e3 c7c6 d1f1 f7g6 a2a4 d8d5 c3c4 d5h5 h2h3 h5f5 f1d1 f5h5 c4c5 a7a5 e3d3 h5h4 d1c1 e7f6 c1f1 h7h5 f1d1');
INSERT INTO evaluations VALUES(55558,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(55558,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(55558,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(55558,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(55558,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(55558,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(55558,36,'cp 29','f7f5 f1g2 f8e7 g1f3 g8f6 b2b4 e8g8 d2d3 d8e8 a2a4 c7c6 b4b5 c8e6 e1g1 h7h6 c1a3 b8d7 a4a5 a7a6 b5c6 b7c6 d1a4 a8c8 e2e3 g7g5 a1b1 g8h8 a4b3 f5f4 f3d2 f4e3 f2e3 f6g4 f1f8 e8f8 b3b7 g4e3 b7a6 e3g2 g1g2');
INSERT INTO evaluations VALUES(56741,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(56741,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(56741,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(56741,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(56741,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(56741,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(56741,36,'cp 34','d2d4 f8e7 f1h3 b8a6 d4e5 d6e5 d1d8 e7d8 g1f3 d8f6 e2e4 g8e7 e1g1 a6b4 f1d1 e8g8 f3e1 b4c6 h3g2 f5e4 c3e4 e7f5 e4f6 g7f6 b2b3 f5d4 c1b2 a7a5 e1f3 d4f3 g2f3 c8e6 f3e4 a5a4 f2f3 f8d8 f3f4 f6f5 e4d5 e6d5 c4d5 c6b4 f4e5');
INSERT INTO evaluations VALUES(19321,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(19321,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(19321,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(19321,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(19321,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(19321,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(19321,36,'cp 104','f8g7 b1c3 d7d6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 c8d7 c1f4 a7a6 b5c6 d7c6 d1d2 a8c8 f4h6 c6d7 f3h2 b7b5 a2a3 a6a5 h2f3 f8e8 c3e2 d8b6 e2g3 b5b4 a1c1 c8c1 e1c1 e8c8 c1c8 e7c8 e4e5 d7c6 a3b4 b6b4');
INSERT INTO evaluations VALUES(22953,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(22953,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22953,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(22953,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(22953,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(22953,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(22953,36,'cp -22','d7d5 d3d5 d8d5 e4d5 c6b4 f1b5 c8d7 b5d7 e8d7 e1d1 b4d5 g1f3 a8d8 h1e1 f8c5 f3e5 d7c8 c1d2 c5f2 e5f7 f2e1 d1e1 g8f6 f7h8 d8h8 b1c3 h8e8 e1f1 d5e3 d2e3 e8e3 a1d1 c7c6 d1d4 a7a5 f1f2 e3e5 f2f3 b7b5 g2g4 c8c7 a2a3 a5a4 h2h3 h7h5 d4f4 h5g4 h3g4');
INSERT INTO evaluations VALUES(39106,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(39106,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(39106,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(39106,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(39106,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(39106,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(39106,36,'cp 168','d4e5 d6e5 g1f3 d7d6 f3e5 d6e5 d1d8 e8d8 b1c3 g8f6 f1e2 c8e6 f2f4 a7a5 c1e3 f6g4 e2g4 e6g4 f4f5 f7f6 h2h3 g4h5 g2g4 h5f7 e1c1 d8c8 h3h4 h8d8 g4g5 d8d1 h1d1 b8d7 d1g1 a8a6 g5g6 h7g6 f5g6 f7g8 g1h1 d7f8 h4h5');
INSERT INTO evaluations VALUES(39096,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(39096,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(39096,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(39096,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(39096,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(39096,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(39096,36,'cp 159','d6e5 g1f3 d7d6 f3e5 d6e5 d1d8 e8d8 b1c3 g8f6 f1e2 h8e8 e1g1 c7c6 c1e3 c8e6 f2f4 e6g4 a1d1 b8d7 f4e5 g4e2 c3e2 e8e5 e3d4 e5e7 d4f6 g7f6 f1f6 d8c7 f6f5 d7e5 b2b3 e5g4 d1d3 e7e4 f5f7 c7b6 e2g3 e4e1 f7f1 e1f1 g3f1 a8e8 h2h3 g4f6 g1f2 f6e4 f2f3 e4g5 f3g4 g5e4 d3e3 e4f2 g4f3');
INSERT INTO evaluations VALUES(39323,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(39323,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(39323,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(39323,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(39323,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(39323,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(39323,36,'cp 142','g1f3 d7d6 f3e5 d6e5 d1d8 e8d8 b1c3 g8f6 f1e2 c8d7 c1e3 b8c6 e1c1 d8c8 h2h3 a7a6 a2a3 h8e8 e2f3 d7e6 g2g3 e8d8 d1d8 c8d8 h1d1 f6d7 f3g2 h7h6 f2f4 f7f6 b2b3 b7b6 c1b2 a6a5 c3d5 c6e7 g3g4 e6f7');
INSERT INTO evaluations VALUES(39322,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(39322,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(39322,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(39322,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(39322,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(39322,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(39322,36,'cp 145','d7d6 f3e5 d6e5 d1d8 e8d8 b1c3 g8f6 f1e2 c8d7 c1g5 h7h6 g5e3 d7c6 f2f3 b8d7 e1c1 d8e7 d1d2 f6e8 h1d1 e8d6 a2a4 a7a5 b2b3 f7f6 c1b2 h8d8 e2b5 d7f8 e3c5 f8e6 c5a3 e6d4 b5c6 b7c6 f3f4 d8d7 h2h3 e7e6 f4e5 f6e5 a3c5');
INSERT INTO evaluations VALUES(1855,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(1855,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(1855,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(1855,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(1855,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(1855,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(1855,36,'cp 15','b1c3 f8c5 g1f3 g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 f3d2 g6e7 a2a3 e6c4 d2c4 f7f6 f1f2 e8g8 g1f1 f6f5 f1e2 f5f4 a1f1 f4e3 e2e3 c6d4 e3d2 f8f2 f1f2 a8f8 f2f8 g8f8 b2b4 a7a6 a3a4 b7b5 a4b5 a6b5 c4e3 f8e8');
INSERT INTO evaluations VALUES(1851,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(1851,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(1851,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(1851,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(1851,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(1851,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(1851,36,'cp 11','f8c5 g1f3 g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 f3d2 g6e7 a2a3 e6c4 d2c4 f7f6 f1f2 h6h5 h2h3 a7a6 a1f1 b7b5 c4d2 e8d7 g1h2 h8f8 d2f3 c6d8 c3d5 e7d5 e4d5 c7c6 d5c6 d8c6 f3h4 d6d5');
INSERT INTO evaluations VALUES(269,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(269,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(269,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(269,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(269,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(269,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(269,36,'cp 17','g1f3 g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 c3d5 e8d7 a2a4 c6e7 b2b4 c7c6 d5c3 e6c4 d3c4 d7e6 a4a5 h8d8 c3a4 g6f8 c4c5 d6d5 a4c3 d5e4 c3e4 e7d5 f3g5 h6g5 e4g5 e6d7 f1f7 d7e8 f7b7 d5e3 b7g7 e3c2');
INSERT INTO evaluations VALUES(268,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(268,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(268,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(268,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(268,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(268,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(268,36,'cp 9','g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 f3d2 g6e7 a2a3 e6c4 d2c4 f7f6 g1f2 e8d7 f2e2 h6h5 g2g3 a7a6 a3a4 h8h6 a1b1 h6g6 b2b4 h5h4 g3h4 a8h8 b4b5 a6b5 a4b5 g6g2 f1f2');
INSERT INTO evaluations VALUES(18818,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(18818,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(18818,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(18818,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(18818,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(18818,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(18818,36,'cp -13','g8f6 f1e2 d7d5 e4d5 e6d5 d3d4 c5d4 f3d4 f8d6 e1g1 e8g8 c1e3 f8e8 f1e1 h7h6 d1d2 d6c7 e3f4 c7a5 e2f3 e8e1 a1e1 c8g4 f3g4 f6g4 d4b3 d5d4 b3a5 d8a5 c3e2 a5b6 b2b3 g4f6 e2g3 a8c8 c2c4 d4c3 d2c3 f6d5 c3f3 d5f4 f3f4');
INSERT INTO evaluations VALUES(12431,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(12431,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(12431,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(12431,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(12431,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(12431,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(12431,36,'cp -15','f1e2 d7d5 e4d5 e6d5 d3d4 c5d4 f3d4 f8d6 e1g1 e8g8 f1e1 f8e8 c1e3 a7a6 e2f3 h7h6 f3d5 f6d5 c3d5 d6h2 g1h2 d8d5 h2g1 c8f5 c2c3 a8d8 d1b3 f5e4 b3d5 e4d5 b2b3 f7f6 a1d1 c6d4 e3d4 e8e1 d1e1 g8f7 d4b6 d8d7 e1d1 d5c6 d1d7 c6d7 f2f3 d7c6');
INSERT INTO evaluations VALUES(12430,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(12430,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(12430,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(12430,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(12430,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(12430,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(12430,36,'cp -17','d7d5 e4d5 e6d5 d3d4 c5d4 f3d4 f8d6 e1g1 e8g8 c1g5 d6e5 g5e3 f8e8 f1e1 h7h6 h2h3 e5c7 e2f1 c8d7 d1d3 a7a6 d4c6 d7c6 a1d1 d8d7 g2g3 a8d8 e3d4 f6e4 c3e4 d5e4 d3c3 f7f5 d4e3 d7d1 e1d1 d8d1');
INSERT INTO evaluations VALUES(36591,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(36591,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36591,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36591,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36591,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(36591,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(36591,36,'cp 18','f1e2 c7c5 g1f3 c5d4 f3d4 b8c6 c1e3 c8g4 d4c6 g4e2 c6d8 e2d1 a1d1 f8d8 b2b3 f6g4 e3d4 e7e5 f4e5 d6e5 d4c5 g7f8 d1d8 a8d8 c5f8 g8f8 c3d5 f8g7 e1g1 g4h6 g2g3 f7f5 f1e1 h6f7 h2h4 d8e8 e4f5 g6f5 e1d1 e8d8 d1d2 g7g6 g1g2 d8e8 g2f3');
INSERT INTO evaluations VALUES(36590,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(36590,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(36590,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36590,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(36590,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36590,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(36590,36,'cp 36','e7e5 f4e5 d6e5 d4d5 b8a6 g1f3 a6c5 c1g5 h7h6 g5f6 d8f6 b2b4 c5d7 c4c5 a7a5 a2a3 f6f4 g2g3 f4g4 e1g1 a5b4 a3b4 a8a1 d1a1 d7f6 a1b1 g4h3 f1e1 c8d7 e2f1 h3h5 f1g2 f8a8 b1c2 c7c6 f3d2 f6g4 h2h3 g4f6 g3g4 h5g5 d2f3 g5f4 c2c1 f4c1 e1c1');
INSERT INTO evaluations VALUES(32667,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(32667,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(32667,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(32667,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(32667,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32667,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32667,36,'cp 0','f1e2 b8c6 h2h3 g4h5 d2d4 e8c8 c2c4 d5e4 c1e3 e7e6 a1c1 f8e7 e2d3 h5f3 d3e4 e7b4 e1f1 f3d1 e4c6 d1h5 c6a4 b4e7 f2f3 h5g6 f1f2 h7h5 g2g3 e7f6 c1d1 c7c5 d4d5 b7b6 b2b3 e6d5 c4d5 g6f5 a4b5 f6e5 h1e1 e5c3 e1h1');
INSERT INTO evaluations VALUES(32668,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(32668,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(32668,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32668,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32668,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32668,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32668,36,'cp 0','b8c6 h2h3 g4h5 d2d4 e8c8 c2c4 d5e4 c1e3 e7e6 a1c1 f8e7 e2d3 h5f3 d3e4 e7b4 e1f1 f3d1 e4c6 d1h5 c6a4 b4e7 f2f3 h5g6 f1f2 h7h5 g2g3 e7f6 c1d1 c7c5 d4d5 b7b6 d1d2 g6f5 h3h4 e6d5 c4d5 f5d7 a4b3 f6e5 h1e1 h8e8 d5d6 d7e6 b3a4 e6d7');
INSERT INTO evaluations VALUES(27429,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27429,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27429,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27429,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27429,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27429,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27429,36,'cp 0','h2h3 g4h5 d2d4 e8c8 c2c4 d5e4 c1e3 e7e6 a1c1 f8e7 e2d3 h5f3 d3e4 e7b4 e1f1 f3d1 e4c6 d1h5 c6a4 b4e7 f2f3 h5g6 f1f2 h7h5 g2g3 e7f6 c1d1 c7c5 d4d5 b7b6 d1d2 g6f5 h3h4 e6d5 c4d5 f5d7 a4b3 f6e5 h1e1 h8e8 d5d6 d7e6 b3a4 e6d7');
INSERT INTO evaluations VALUES(27427,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27427,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27427,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27427,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27427,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27427,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27427,36,'cp 0','g4h5 d2d4 e8c8 c2c4 d5e4 c1e3 e7e6 a1c1 f8e7 e2d3 h5f3 d3e4 e7b4 e1f1 f3d1 e4c6 d1h5 c6a4 b4e7 f2f3 h5g6 f1f2 h7h5 g2g3 e7f6 c1d1 c7c5 d4d5 b7b6 d1d2 e6d5 c4d5 g6f5 h3h4 f5d7 a4b3 f6e5 h1e1 h8e8 d5d6 d7e6 b3a4 e6d7');
INSERT INTO evaluations VALUES(27418,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27418,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27418,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27418,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27418,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27418,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27418,36,'cp 0','d2d4 e8c8 c2c4 d5e4 c1e3 e7e6 a1c1 f8e7 e2d3 h5f3 d3e4 e7b4 e1f1 f3d1 e4c6 d1h5 c6a4 b4e7 f2f3 h5g6 f1f2 h7h5 g2g3 e7f6 c1d1 c7c5 d4d5 b7b6 d1d2 e6d5 c4d5 g6f5 h3h4 f5d7 a4b3 f6e5 h1e1 h8e8 d5d6 d7e6 b3a4 e6d7');
INSERT INTO evaluations VALUES(39015,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(39015,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(39015,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(39015,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(39015,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(39015,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(39015,36,'cp 91','g1f3 d7d5 c1f4 g8f6 a1c1 e8g8 e2e3 a7a6 d1b3 b8c6 f1d3 b7b6 c4d5 e6d5 e1g1 c8b7 f1d1 c6a5 b3c2 a8c8 h2h3 f8e8 d3e2 g7f8 b2b3 f8a3 c1b1 a3b4 c2b2 f6e4 c3e4 d5e4 f3d2 d8e7 f4g3');
INSERT INTO evaluations VALUES(39014,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(39014,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(39014,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(39014,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(39014,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(39014,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(39014,36,'cp 93','d7d5 c1f4 g8f6 a1c1 a7a6 e2e3 e8g8 d1b3 b8c6 f1d3 c6a5 b3b4 d5c4 d3c4 a5c6 b4a4 c8d7 a4c2 f6h5 f4e5 c6e5 d4e5 d7c6 c3e4 d8e7 e1g1 a8d8 b2b3 c6d5 f1d1 c7c6 a2a4 d5e4 c2e4 d8d1 c1d1 f8d8 d1d8 e7d8 g2g3 d8d1 g1g2 d1d7 e4h4 g7f8');
INSERT INTO evaluations VALUES(38568,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(38568,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(38568,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(38568,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(38568,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(38568,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(38568,36,'cp 106','c1f4 g8f6 a1c1 e8g8 e2e3 a7a6 d1b3 b8c6 f1e2 c6a5 b3b4 a5c6 b4a3 f6h5 f4g5 f7f6 g5h4 g6g5 h4g3 d5c4 e2c4 g8h8 e1g1 h5g3 h2g3 d8e7 a3e7 c6e7 g3g4 c8d7 b2b4 f8d8 f1d1 d8c8 a2a4 e7c6 c1b1 b7b6 d1c1 c6e7 c3e4');
INSERT INTO evaluations VALUES(6083,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(6083,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(6083,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(6083,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(6083,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(6083,35,'cp -110',NULL);
INSERT INTO evaluations VALUES(6083,36,'cp -111','d8d4 e5d3 d4e4 d1e2 e4e2 e1e2 c5d4 f2f3 c8f5 c2c3 d4b6 g2g4 f5d7 a2a4 a7a5 b1a3 e8c8 a3c4 d8e8 e2d1 h7h5 g4g5 g8e7 c4b6 c7b6 d3f2 e7g6 d2d3 d7f5 h1e1 e8d8 d1e2 h8e8 f2e4 c6c5 c1e3 c5c4 e2f2 d8d3');
INSERT INTO evaluations VALUES(207,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(207,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(207,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(207,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(207,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(207,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(207,36,'cp -107','e5d3 d4e4 d1e2 e4e2 e1e2 c5d4 c2c3 d4f6 f2f3 b7b6 d3f4 c8a6 e2f2 e8c8 d2d4 d8e8 h1e1 e8e1 f2e1 g8e7 e1f2 h8e8 b1d2 e7d5 d2e4 f6h4 g2g3 h4e7 g3g4 e7h4 f2g2 d5f6 g2h3 f6e4 f3e4 h4d8 e4e5 c6c5 c1d2 c5d4 c3d4 c7c5 d4c5 e8e5 c5b6 a7b6 a1e1 d8f6 b2b3 a6b7 e1e5 f6e5');
INSERT INTO evaluations VALUES(208,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(208,31,'cp -130',NULL);
INSERT INTO evaluations VALUES(208,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(208,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(208,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(208,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(208,36,'cp -117','d4e4 d1e2 e4e2 e1e2 c5d4 h1e1 c8f5 c2c3 d4f6 e2f1 f6e7 d3c5 b7b6 c5e4 c6c5 c3c4 e8c8 b1c3 e7h4 g2g3 h4f6 e4f6 g8f6 f1g2 d8d3 b2b3 h8d8 f2f3 h7h6 g3g4 f5g6 h2h4 h6h5 g4g5 f6d7 c3d5 d3d4 d5e7 c8b7 e7g6 f7g6');
INSERT INTO evaluations VALUES(210,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(210,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(210,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(210,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(210,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(210,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(210,36,'cp -103','d1e2 e4e2 e1e2 c5d4 f2f3 b7b6 h1e1 g8e7 e2d1 c8f5 d3f4 h7h5 d2d3 f7f6 f4e2 d4e5 h2h3 e8f7 a2a4 g7g5 b1d2 f5e6 d2e4 e7g6 e1f1 h5h4 f3f4 g5f4 e2f4 g6f4 c1f4 e5f4 f1f4 a8g8 d1d2 g8g2 d2c3 g2g6 a1h1 c6c5 a4a5 f6f5 a5b6 a7b6 b2b3 h8a8 h1f1 g6g2');
INSERT INTO evaluations VALUES(211,30,'cp -116',NULL);
INSERT INTO evaluations VALUES(211,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(211,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(211,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(211,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(211,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(211,36,'cp -125','e4e2 e1e2 c5d6 b1c3 g8f6 d3e1 e8g8 d2d3 f8e8 c1e3 a7a5 e1f3 a5a4 a2a3 f6d5 e2d2 d5e3 f2e3 c8d7 c3e4 d6f8 e4c3 b7b5 e3e4 f7f6 e4e5 f6e5 a1e1 f8d6 h1f1 c6c5 d2c1 h7h6 f3d2 d7c6 c3e4 d6e7 d2f3 e7f8 g2g4 c5c4 d3c4 b5c4');
INSERT INTO evaluations VALUES(213,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(213,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(213,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(213,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(213,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(213,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(213,36,'cp -97','e1e2 c5d4 f2f3 b7b6 h1e1 g8e7 e2d1 c8f5 d3f4 h7h5 d2d3 f7f6 b1d2 h5h4 f4e2 d4e5 h2h3 f5e6 d2e4 a7a5 a2a4 e8d7 e4f2 e7d5 f2g4 e5d6 e2d4 a8e8 d4e6 e8e6 g4e3 e6e5 c1d2 d5e3 e1e3 e5e3 d2e3');
INSERT INTO evaluations VALUES(212,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(212,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(212,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(212,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(212,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(212,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(212,36,'cp -121','c5d4 f2f3 b7b6 h1e1 g8e7 e2d1 c8f5 d3f4 h7h5 d2d3 f7f6 b1d2 e8f7 f4e2 d4e5 h2h3 f5e6 d2e4 e7f5 a2a4 h5h4 a4a5 e6d7 c2c3 c6c5 c1f4 e5f4 e2f4 h8d8 d1c2 d7b5 c3c4 b5c6 c2c3 c6b7 b2b3 f5d4 c3b2 a8b8 e4f2');
INSERT INTO evaluations VALUES(56552,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(56552,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(56552,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(56552,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(56552,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(56552,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(56552,36,'cp 16','g1f3 e5d4 d1d4 d8d4 f3d4 g8f6 c1f4 f8b4 f4c7 e8g8 e2e3 c8d7 a2a3 b4c3 b2c3 f8c8 c7b8 a8b8 f2f3 f6d5 e1d2 b7b5 e3e4 d5b6 f1e2 c8e8 e2d1 b8d8 d1c2 h7h5 h1b1 h5h4 b1d1 a7a5 a1b1 h4h3 g2h3');
INSERT INTO evaluations VALUES(56551,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56551,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56551,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56551,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56551,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56551,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56551,36,'cp 0','e5d4 d1d4 d8d4 f3d4 g8f6 c1f4 f8b4 f4c7 e8g8 e2e3 c8d7 c7e5 f6e4 a2a3 b4c3 b2c3 f8c8 a1c1 b7b5 f2f3 e4c5 f1e2 b8c6 d4c6 c8c6 e1f2 f7f6 e5f4 d7e6 h1d1 a8c8 e3e4 c6a6 c1a1 c5a4 d1c1 a4b2 c1b1 b2a4');
INSERT INTO evaluations VALUES(56655,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(56655,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(56655,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56655,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(56655,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(56655,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(56655,36,'cp 24','d1d4 d8d4 f3d4 f8c5 e2e3 g8e7 f1c4 b8c6 d4f3 e7g6 e1g1 g6e5 c4e2 c8g4 f3e5 g4e2 c3e2 c6e5 f1d1 e5c6 b2b3 e8g8 c1b2 f8d8 g1f1 f7f6 e2c3 g8f7 f1e2 d8d1 a1d1 a8d8 d1d8 c6d8 g2g4 c5d6 h2h3 f7e6 f2f4 g7g6 c3b5');
INSERT INTO evaluations VALUES(4050,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(4050,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(4050,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(4050,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(4050,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(4050,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(4050,36,'cp 86','e1g1 e7e5 c1e3 e5d4 f3d4 a7a6 d1c2 c7c6 h2h3 f8e8 a1d1 d8c7 d4b3 b7b5 c4b5 c6b5 c2d2 g7f8 c3d5 c7b8 d1c1 f6e4 d2d1 c8b7 d5c7 e4g3 f1e1 e8e3 f2e3 a8a7 e2f3 b7f3 d1f3 a7c7 f3g3 c7c1 b3c1 f8g7 e1d1 g7b2');
INSERT INTO evaluations VALUES(4049,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(4049,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(4049,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(4049,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(4049,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(4049,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(4049,36,'cp 75','e7e5 c1e3 e5d4 f3d4 c7c6 d1c2 a7a6 h2h3 d8e7 a1d1 d7c5 e2f3 f8e8 f1e1 a6a5 c2d2 a5a4 e3f4 f6d7 f3e2 e7f8 f2f3 g7e5 f4e3 d7f6 e2d3 f6h5 d4e2 c5d3 d2d3 h5g3 d3d2 g3e2 c3e2');
INSERT INTO evaluations VALUES(3957,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(3957,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(3957,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(3957,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(3957,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(3957,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(3957,36,'cp 62','c1e3 e5d4 f3d4 f8e8 d1c2 c7c6 a1d1 a7a6 h2h3 d8c7 d4b3 b7b5 e3f4 d7e5 c4c5 d6c5 b3c5 f6d7 c5d7 c8d7 f4e3 d7e6 c3d5 e6d5 e4d5 a8d8 f1e1 d8d6 a2a4 e8d8 a4b5 a6b5 d5c6 c7c6 d1d6 d8d6 c2c6 e5c6 e2b5 c6d4 e3f4');
INSERT INTO evaluations VALUES(3951,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(3951,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(3951,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(3951,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(3951,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(3951,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(3951,36,'cp 88','e5d4 f3d4 f8e8 d1c2 a7a6 a1d1 c7c6 h2h3 d8c7 d4b3 b7b5 e3f4 d7e5 c4c5 d6c5 b3c5 c7e7 f4e3 e5d7 b2b4 a6a5 a2a3 d7c5 e3c5 e7c7 c5d6 c7a7 b4a5 c8e6 d6b4 a8c8 c2d2 e6b3 d1b1 c8d8 b4d6 b3e6 e4e5 f6d7 a3a4 d7e5 a4b5 d8d6');
INSERT INTO evaluations VALUES(53138,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(53138,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(53138,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(53138,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(53138,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(53138,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(53138,36,'cp -13','e7e6 f1c4 a7a6 e1g1 f8e7 d1e2 b8c6 f1d1 d8c7 c1f4 g8f6 c4b3 e8g8 a1c1 c7b8 c3a4 b7b5 c1c6 b5a4 b3a4 c8d7 c6c4 d7b5 a4b5 a6b5 c4c6 a8a2 f4d6 e7d6 c6d6 h7h6 d6d4 f8c8 d4b4 b8b7 e2b5 b7b5 b4b5 f6e4 b5b7 c8a8 b7b4 e4f6 h2h3 a2a1 d1a1 a8a1 g1h2');
INSERT INTO evaluations VALUES(54029,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(54029,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(54029,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(54029,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(54029,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54029,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54029,36,'cp -8','f1c4 a7a6 e1g1 b8c6 d1e2 f8e7 f1d1 c8d7 c1f4 e6e5 f4e3 g8f6 a1c1 a8c8 a2a3 e8g8 c3d5 b7b5 c4a2 f6d5 a2d5 d8e8 e2d2 d7e6 b2b4 e8d7 d5c6 c8c6 f3e5 d6e5 d2d7 e6d7 d1d7 c6c1 e3c1 f8d8 d7d8 e7d8 c1e3 f7f6 f2f3 g8f7 h2h3 d8e7 e3b6 f7g6 g1f2 f6f5 f2e2 f5e4');
INSERT INTO evaluations VALUES(54014,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54014,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54014,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(54014,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54014,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54014,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54014,36,'cp 0','f8e7 e1g1 b8c6 d1e2 c8d7 f1d1 g8f6 c3b5 d8b8 c1f4 c6e5 d1d6 e7d6 b5d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 d6f8 d7f8 e2d3 a8c8 b2b3 g5e5 h2h4 h7h6 g2g3 g7g5 d3d6 e5d6 d1d6 g8g7 g1g2 g5h4 g3h4 f8g6 d6d7');
INSERT INTO evaluations VALUES(37909,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37909,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37909,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37909,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37909,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37909,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37909,36,'cp 0','e1g1 b8c6 d1e2 a7a6 f1d1 c8d7 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 a1c1 a8c8 a2a3 d8e8 e2d2 d7e6 h2h3 e8d8 b2b4 h7h6 d2d3 e6d5 d3d5 g8h7 a3a4 c6b4 d5b7 c8b8 b7a7 b8a8 a7b7');
INSERT INTO evaluations VALUES(37908,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37908,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37908,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37908,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(37908,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(37908,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37908,36,'cp 0','g8f6 d1e2 e8g8 e4e5 d6e5 f1d1 d8a5 f3e5 b8c6 c1f4 c6e5 f4e5 a7a6 c4b3 b7b6 d1d4 f8d8 d4a4 a5c5 a4c4 c5a5');
INSERT INTO evaluations VALUES(40198,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(40198,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(40198,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(40198,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(40198,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(40198,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40198,36,'cp 0','d1e2 b8c6 f1d1 c8d7 c3b5 d8b8 c1f4 c6e5 b5d6 e7d6 d1d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 d6f8 d7f8 g2g3 a8d8 d1d8 g5d8 e2e3 b7b6 h2h4 d8d6 g1g2 d6e5 b2b3 g7g6 f2f4 e5b2 e3e2 b2b1 e4e5 h7h6 a2a4 b1a1 e2d3 a1b2 d3e2');
INSERT INTO evaluations VALUES(40197,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40197,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40197,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40197,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40197,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40197,36,'cp 0','b8c6 f1d1 c8d7 c3b5 d8b8 c1f4 c6e5 d1d6 e7d6 b5d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 d6f8 d7f8 g2g3 a8d8 d1d8 g5d8 e2e3 b7b6 h2h4 d8d6 b2b3 d6e5 g1g2 h7h6 f2f4 e5b2 e3e2 b2c1 e2d3 g7g6 a2a4 c1b2 d3e2 b2b1 e4e5 b1c1 e2f1 c1d2 f1e2');
INSERT INTO evaluations VALUES(8191,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8191,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8191,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8191,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8191,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8191,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8191,36,'cp 0','f1d1 c8d7 c3b5 d8b8 c1f4 c6e5 b5d6 e7d6 d1d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 d6f8 d7f8 g2g3 a8d8 d1d8 g5d8 e2e3 b7b6 h2h4 d8d6 b2b3 d6e5 g1g2 g7g6 f2f4 e5b2 e3e2 b2b1 e4e5 h7h5 a2a4 b1c1 e2d3 c1b2 d3e2');
INSERT INTO evaluations VALUES(8192,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8192,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8192,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8192,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8192,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8192,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8192,36,'cp 0','c8d7 c3b5 d8b8 c1f4 c6e5 b5d6 e7d6 d1d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 d6f8 d7f8 g2g3 a8d8 d1d8 g5d8 e2e3 b7b6 h2h4 d8d6 b2b3 d6e5 g1g2 g7g6 f2f4 e5b2 e3e2 b2c3 e2d3 c3c1 a2a4 h7h6 d3f1 c1c2 f1e2 c2c3 e2d3 c3b2 d3e2');
INSERT INTO evaluations VALUES(9761,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(9761,31,'cp 251',NULL);
INSERT INTO evaluations VALUES(9761,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(9761,33,'cp 250',NULL);
INSERT INTO evaluations VALUES(9761,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(9761,35,'cp 254',NULL);
INSERT INTO evaluations VALUES(9761,36,'cp 252','d2d4 c5b6 f3e5 e8g8 e5c6 b7c6 e4e5 f6d5 b1d2 d7d6 f1e1 c8e6 c4b3 d8d7 d2f3 d5e7 c1f4 a7a5 h2h3 e6b3 a2b3 e7g6 f4g3 d6e5 g3e5 f8e8 e5g3 e8e1 f3e1 a8e8 e1d3 d7d5 g1h2 d5f5 d1d2 f5d7 a1e1 e8e1 d2e1');
INSERT INTO evaluations VALUES(9794,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(9794,31,'cp 239',NULL);
INSERT INTO evaluations VALUES(9794,32,'cp 247',NULL);
INSERT INTO evaluations VALUES(9794,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(9794,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(9794,35,'cp 248',NULL);
INSERT INTO evaluations VALUES(9794,36,'cp 255','c5b6 f3e5 e8g8 e5c6 b7c6 e4e5 f6d5 b1d2 d7d6 f1e1 d6e5 e1e5 f8e8 e5e8 d8e8 c4f1 d5f6 d2c4 c8e6 a2a4 a7a5 c1e3 f6d5 e3d2 d5f6 c4b6 c7b6 c3c4 e6f5 d2f4 e8e6 f2f3 a8e8 d1b3 f6d7 f4g3 e6e3 b3e3 e8e3 d4d5 c6d5 c4d5 e3b3');
INSERT INTO evaluations VALUES(15521,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(15521,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(15521,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(15521,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(15521,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(15521,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(15521,36,'cp 31','c8g4 c4b5 f6d7 d4d5 c6d4 b5e2 d4e2 d1e2 f8e7 h2h3 g4f3 e2f3 e7g5 c1g5 d8g5 h3h4 g5g6 h1h3 a7a6 h4h5 g6g5 h3g3 g5h6 e1e2 g7g6 a1h1 h8f8 g3h3 h6g5 h5g6 f7g6 f3g3 g5g3 h3g3 d7f6 g3f3 e8e7 a2a4 f8g8');
INSERT INTO evaluations VALUES(25333,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(25333,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(25333,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(25333,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(25333,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(25333,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(25333,36,'cp 37','c4b5 f6d7 d4e5 d6e5 c1e3 f8d6 d1d2 e8g8 e1c1 d7b6 b5e2 d8e7 c3d5 b6d5 e4d5 c6d8 h2h3 g4f3 e2f3 f7f5 h1e1 h7h6 g2g3 e5e4 f3e2 d8f7 c2c3 d6c5 f2f4 c5e3 d2e3 f7d6 g3g4 e7h4 g4f5');
INSERT INTO evaluations VALUES(55773,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55773,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(55773,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55773,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(55773,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(55773,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55773,36,'cp 0','e6d5 d2d4 c7c6 c1f4 g8f6 d1c2 g7g6 e2e3 c8f5 f1d3 f5d3 c2d3 f8g7 h2h3 e8g8 g1f3 b8d7 e1g1 f8e8 d3c2 a7a5 a2a4 g7f8 f1c1 f8b4 c3a2 b4e7 a2c3');
INSERT INTO evaluations VALUES(26108,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26108,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(26108,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(26108,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(26108,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(26108,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(26108,36,'cp 42','e1g1 g8e7 f1d1 c6c5 c2c4 c5d4 b1a3 e7c6 c4d5 d7c5 d3b1 e6d5 a3c2 d4d3 c2e1 f8e7 e1d3 c5e6 d3f4 e6f4 c1f4 d8d7 b1d3 a8d8 a2a3 e8g8 a1c1 a7a6 d3b3 d5d4 h2h3 d7c7 b3c4 c7b6 e5e6 b6b5 c4b5 a6b5 f3g5 e7g5 f4g5 d8c8 e6e7 f8e8 f2f4');
INSERT INTO evaluations VALUES(26109,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26109,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(26109,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(26109,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(26109,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(26109,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(26109,36,'cp 31','g8e7 c2c4 d5c4 d3c4 d7b6 c4e2 d8d5 b1c3 d5c4 e2d2 e7d5 c3e4 h7h6 b2b3 c4a6 a2a4 a8d8 c1a3 f8a3 a1a3 b6c8 h2h4 a6b6 a4a5 b6b4 d2b4 d5b4 a3a4 b4d5 g2g3 e8g8 h4h5 b7b5 a5b6 a7b6 f1a1 c8e7');
INSERT INTO evaluations VALUES(24537,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(24537,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24537,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24537,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(24537,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(24537,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(24537,36,'cp 35','c2c4 d5c4 d3c4 e7d5 b1c3 f8e7 c4b3 d7b6 c3e4 e8g8 c1d2 a7a5 a2a3 h7h6 a1e1 a5a4 b3c2 b6d7 c2c1 g8h7 g2g4 f7f5 e5f6 d7f6 c1c2 h7g8 f3e5 f6e4 c2e4 e7f6 e5f3 d5c7 d2f4 f8f7 f4c7 d8c7 e1e3 a8f8 f1e1 c7b6 e4e6 b6b2 h2h4');
INSERT INTO evaluations VALUES(55612,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55612,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(55612,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(55612,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55612,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(55612,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(55612,36,'cp 7','g8f6 b1c3 c7c5 e2c4 h7h6 d2d3 b8c6 e1g1 f8e7 c1d2 e8g8 a2a3 g8h8 a3a4 c8d7 h2h3 a7a6 f3h2 d7e6 f1e1 a8b8 h2f1 f6h7 f1e3 e7g5 e1f1 h7f6 c3d5 b7b5 a4b5 a6b5 c4a2 b8a8 d5f6 a8a2 a1a2 e6a2 f6h5 a2e6 e3f5 e6f5 e4f5 g5d2 d1d2');
INSERT INTO evaluations VALUES(47895,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(47895,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(47895,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(47895,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(47895,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(47895,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47895,36,'cp 3','b1c3 c7c5 e2c4 h7h6 d2d3 b8c6 e1g1 f8e7 a2a4 e8g8 f1e1 c8e6 c3d5 d8d7 c2c3 e7d8 f3d2 a7a6 h2h3 a8b8 d5f6 d8f6 d2f1 g8h7 c1e3 f6e7 c4d5 e6d5 e4d5 c6d8 a4a5 f7f5 f2f4 e7h4 g2g3 h4f6 d1f3 d8f7 f1d2 e5f4 e3f4');
INSERT INTO evaluations VALUES(47867,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47867,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(47867,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47867,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47867,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(47867,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47867,36,'cp 0','c7c5 e2c4 h7h6 d2d3 b8c6 h2h3 c8e6 c3d5 c6a5 c4b5 a5c6');
INSERT INTO evaluations VALUES(6692,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(6692,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(6692,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(6692,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(6692,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(6692,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(6692,36,'cp 29','b1d2 g8f6 d4d5 c6e7 d1b3 b4d2 f3d2 e8g8 e1g1 d7d6 c1a3 a8b8 b3d3 c8d7 a1e1 b7b5 c4b3 b5b4 a3b2 e7g6 d2c4 d7b5 f2f4 a7a5 d3d4 a5a4 b3c2 f8e8 f1f2 h7h6 g2g3 b4b3 a2b3 a4b3 c2b3 b5a6 c4a5');
INSERT INTO evaluations VALUES(6691,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(6691,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(6691,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6691,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(6691,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(6691,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(6691,36,'cp 35','g8f6 d4d5 c6e7 d1b3 b4d2 f3d2 d7d6 e1g1 e8g8 f1e1 c7c6 c1b2 e7g6 h2h3 b7b5 c4f1 c6d5 e4d5 c8d7 a1d1 a7a6 b2f6 d8f6 d2e4 f6f4 b3a3 a8e8 a3d6 f4d6 e4d6 e8e1 d1e1 f8d8 d6b7 d8c8 g2g3 g8f8 f1g2 c8c2 d5d6 c2d2');
INSERT INTO evaluations VALUES(10502,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(10502,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(10502,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(10502,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(10502,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(10502,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(10502,36,'cp 37','d4d5 c6e7 d1b3 b4d2 f3d2 d7d6 e1g1 e8g8 h2h3 e7g6 f1e1 c7c6 c1b2 b7b5 c4f1 c6d5 e4d5 c8d7 b3g3 f8e8 g3d6 e8e1 a1e1 d7f5 d6a6 d8d5 d2f3 f5e6 a2a3 g6f4 e1e3 f6e4 a6b5 d5b5 f1b5 e6d5 b5f1 f4e6 e3e1 e4c5 f3d4 h7h6 d4e6 c5e6');
INSERT INTO evaluations VALUES(17020,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(17020,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(17020,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(17020,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(17020,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(17020,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(17020,36,'cp 23','f8d6 c1e3 d8e7 b1c3 d6e5 d4d2 e8g8 e3d4 d7d5 c4d5 c7c6 d4e5 e7e5 d5c4 f6e4 d2e3 c8f5 e1c1 a8e8 c3e4 e5e4 e3e4 f5e4 f2f3 e4f5 g2g4 f5c8 a2a4 e8e3 d1d3 f8e8 d3e3 e8e3 h1d1 g8f8 d1d8 e3e8 d8d4 a7a5 c1d2 f7f6 b2b3 g7g6 h2h4 f8e7');
INSERT INTO evaluations VALUES(21640,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(21640,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(21640,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(21640,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21640,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21640,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(21640,36,'cp 16','d5e4 d1d8 c6d8 f3g5 c8f5 b1c3 f8b4 c1d2 b4c3 d2c3 h7h6 g5h3 d8e6 e1c1 g8e7 h3g1 h6h5 h2h4 f5g4 d1e1 e8c8 f1c4 e6f4 h1h2 e7d5 c3d2 h8e8 e1e4 f4g6 e5e6 d5f6 e4e1 g4e6 c4e6 e8e6 e1e6 f7e6 f2f3 e6e5 g2g3 d8f8 h2e2 f6d5 e2e4 f8f7 a2a4 d5e7 a4a5 e7c6 e4e1 a7a6');
INSERT INTO evaluations VALUES(21784,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(21784,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(21784,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(21784,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(21784,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21784,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21784,36,'cp 28','d1d8 c6d8 f3g5 c8f5 b1c3 f8b4 c1d2 b4c3 d2c3 h7h6 g5h3 d8e6 e1c1 g8e7 a2a3 f5g4 d1e1 e8c8 f1c4 g4f5 h3g1 e6f4 g1e2 f4e2 c4e2 e7d5 c3d2 h8e8 e2h5 d8d7 f2f4 e4f3 h5f3 e8d8 h1f1 f5e6 f1f2 d5e7 f3h5 e7c6 d2c3 g7g6 h5d1 b7b6 b2b3 a7a6 d1f3 c6d4');
INSERT INTO evaluations VALUES(2262,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(2262,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(2262,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2262,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(2262,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(2262,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(2262,36,'cp 20','c6d8 f3g5 c8f5 b1c3 f8b4 c1d2 b4c3 d2c3 h7h6 g5h3 d8e6 e1c1 g8e7 f1c4 a7a6 a2a3 a8d8 d1d8 e8d8 h1d1 d8c8 h3g1 b7b5 c4a2 e7g6 g1e2 g6h4 e2d4 f5g4 d1e1 h6h5 g2g3 h4f3 d4f3 e4f3 c1d2 h8d8 d2e3 c7c5 a2e6 f7e6 e3f4 d8d5');
INSERT INTO evaluations VALUES(2388,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(2388,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(2388,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(2388,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(2388,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(2388,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(2388,36,'cp 19','f3g5 c8f5 b1c3 f8b4 c1d2 b4c3 d2c3 h7h6 g5h3 d8e6 e1c1 g8e7 f1c4 a7a6 a2a3 b7b5 c4a2 a8d8 d1d8 e8d8 h3g1 h6h5 g1e2 h5h4 h1e1 d8d7 h2h3 h8h6 c3d2 h6g6 a2e6 g6e6 e2f4 e6c6 e1d1 d7c8 d2b4 e7g6 f4e2 g6e5');
INSERT INTO evaluations VALUES(2386,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(2386,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(2386,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(2386,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(2386,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(2386,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(2386,36,'cp 20','c8f5 b1c3 f8b4 c1d2 b4c3 d2c3 g8e7 e1c1 h7h6 g5h3 d8e6 f1c4 a7a6 h1e1 b7b5 c4e2 h6h5 h3g1 h5h4 f2f3 e7g6 c1b1 e4f3 e2f3 a8d8 g1e2 c7c5 a2a3 e8e7 g2g4 h4g3 e2g3 d8d1 e1d1 g6h4 f3b7 f5g4 d1d3 h4f5 b7a6 f5g3 h2g3 h8h1 b1a2');
INSERT INTO evaluations VALUES(23570,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(23570,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(23570,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(23570,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(23570,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(23570,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23570,36,'cp 28','b1c3 f8b4 c1d2 b4c3 d2c3 h7h6 g5h3 d8e6 e1c1 g8e7 h3g1 h6h5 h2h4 f5g4 g1e2 e7f5 b2b3 a8d8 d1e1 c7c6 a2a4 g4e2 f1e2 e4e3 f2e3 f5e3 g2g3 e3f5 h1h3 f5d4 e2d1 d4f5 d1f3 g7g6 g3g4 h5g4 f3g4 f5h4 e1f1 b7b6 a4a5 d8d5 a5a6 c6c5 h3h2 d5d8 h2f2');
INSERT INTO evaluations VALUES(43591,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43591,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43591,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43591,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43591,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43591,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43591,36,'cp 29','g1f3 e7e6 b1d2 c7c5 c2c3 f8d6 f4g3 d8c7 d4c5 c7c5 f1d3 c5b6 d1c2 e8g8 c3c4 d6g3 h2g3 h7h6 e1g1 c8d7 c4c5 b6a7 b2b4 b7b6 d2b3 f8c8 f1c1 b6c5 b3c5 d7b5 d3b5 a6b5 f3d4 b8c6 d4b5 c6b4 b5a7 b4c2 a7c8 c2a1');
INSERT INTO evaluations VALUES(43593,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(43593,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(43593,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(43593,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(43593,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(43593,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(43593,36,'cp 23','e7e6 c2c3 f8d6 b1d2 e8g8 f3e5 c7c5 f1d3 b8c6 d1f3 d8c7 e5c6 c7c6 e1g1 c6c7 f4d6 c7d6 d4c5 d6c5 e3e4 c5b6 a1b1 d5e4 d2e4 f6e4 d3e4 f7f5 e4c2 c8d7 f1e1 a8d8 f3e3 b6e3 e1e3 a6a5');
INSERT INTO evaluations VALUES(43660,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(43660,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43660,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43660,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(43660,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(43660,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43660,36,'cp 34','c2c3 b8d7 b1d2 c7c5 f3e5 d8b6 a1b1 d7e5 d4e5 f6e4 d2e4 d5e4 d1a4 c8d7 a4e4 d7c6 e4c2 f8e7 h2h4 a8d8 f2f3 d8d7 f1e2 e8g8 b1d1 d7d1 e2d1 b6c7 e1f2 b7b5 f4g5 h7h6 g5e7 c7e7 d1e2 c5c4 a2a3');
INSERT INTO evaluations VALUES(43656,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(43656,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(43656,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43656,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(43656,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43656,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(43656,36,'cp 25','c7c5 b1d2 b8d7 f1d3 d8b6 d1c2 f8d6 e1g1 e8g8 a2a4 d6f4 e3f4 b6c7 g2g3 a8b8 b2b3 b7b6 f3e5 c8b7 b3b4 c5c4 d3e2 b8a8 f1e1 b6b5 e5d7 f6d7 a4a5 b7c6 d2f3 d7f6 f3e5 f8c8 e2f3 c6e8 f3g2 h7h6');
INSERT INTO evaluations VALUES(43565,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43565,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(43565,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(43565,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(43565,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(43565,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(43565,36,'cp 21','b1d2 f8d6 f4g3 d8b6 d1c2 e8g8 f1e2 h7h6 e1g1 d6g3 h2g3 c8d7 c2b3 b6a7 d4c5 f8c8 c3c4 c8c5 c4d5 e6d5 b3a3 c5c7 b2b4 b8c6 a3b2 a8e8 f1e1 f6g4 b4b5 a6b5 e2b5 a7b6 a2a4 c6e5 e1c1 d7b5 b2b5 b6b5 a4b5 e8c8 c1c7');
INSERT INTO evaluations VALUES(43564,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43564,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(43564,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43564,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(43564,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43564,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(43564,36,'cp 27','f8d6 f1d3 e8g8 f4g3 d8b6 e3e4 d6g3 h2g3 f6e4 d2e4 d5e4 d3e4 f7f5 e4c2 b8c6 c2b3 c5d4 f3d4 c6d4 c3d4 c8d7 e1g1 a8d8 d1d2 d7e8 f1e1 b6d4 b3e6 g8h8 d2d4 d8d4 e6f5 d4d2 a1c1 e8g6 f5g6 h7g6 f2f3 d2b2 c1c7 b2a2 e1e7 a6a5 e7g7');
INSERT INTO evaluations VALUES(57153,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(57153,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(57153,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(57153,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(57153,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(57153,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(57153,36,'cp -21','g8f6 b1c3 f8c5 g1f3 e8g8 e1g1 b8c6 h2h3 a7a6 d3c4 d7d6 c4b3 h7h6 d2d3 c6a5 f1e1 a5b3 a2b3 c8d7 c1e3 c5e3 e1e3 c7c5 f3d2 b7b5 d2f1 g8h7 d1d2 a6a5 e3f3 b5b4 c3e2 d7e6 f1e3 d6d5 e4d5 f6d5');
INSERT INTO evaluations VALUES(49636,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49636,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(49636,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49636,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(49636,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(49636,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(49636,36,'cp -14','b1c3 f8c5 g1f3 e8g8 e1g1 f8e8 d3c4 d7d6 a2a3 c7c6 d2d3 b7b5 c4a2 c5b6 f3g5 e8f8 c1e3 h7h6 g5f3 f8e8 h2h3 c8e6 a2e6 e8e6 e3b6 d8b6 d3d4 b8d7 d4d5 c6d5 c3d5 f6d5 d1d5 e6e8 a3a4 b5b4 d5b3');
INSERT INTO evaluations VALUES(49623,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(49623,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(49623,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(49623,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49623,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49623,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(49623,36,'cp -17','f8c5 g1f3 e8g8 e1g1 f8e8 d3c4 d7d6 a2a3 c7c6 d2d3 b7b5 c4a2 c5b6 h2h3 b8d7 c1e3 d7f8 e3b6 d8b6 d3d4 h7h6 f1e1 f8g6 d1d3 a7a5 a1d1 c8d7 d3d2 a8d8 d2e3 b6c7 b2b4 a5b4 a3b4 d7e6 a2e6 e8e6 d4e5');
INSERT INTO evaluations VALUES(14224,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(14224,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(14224,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(14224,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(14224,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(14224,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(14224,36,'cp 24','g1f3 c8f5 b1d2 e7e6 f3h4 d8d7 e2e3 f8d6 f4d6 d7d6 f1e2 e8g8 e1g1 f6e4 h4f5 e6f5 d2e4 f5e4 c3c4 c6e7 d1b3 a8b8 c4d5 e7d5 a1c1 c7c6 c1c5 f8d8 b3c2 d6e6 h2h3 d5b4 c2c4 b4d5 g1h1 g7g6 c4c2 d5b4 e2c4 b4c2 c4e6');
INSERT INTO evaluations VALUES(12148,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(12148,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(12148,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(12148,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(12148,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(12148,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(12148,36,'cp 26','e2e3 h7h6 g5h4 f8e7 f1e2 d5c4 e2c4 b7b5 c4d3 a7a6 h4f6 d7f6 e1g1 c6c5 d4c5 e7c5 d3e4 a8a7 d1d8 e8d8 f3e5 a7c7 f1d1 d8e8 e4c6 c8d7 e5d7 f6d7 d1d7 c7c6 a1d1 c6c8 h2h4 b5b4 c3e4 c5e7');
INSERT INTO evaluations VALUES(12147,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(12147,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(12147,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(12147,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(12147,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12147,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12147,36,'cp 26','h7h6 g5f4 f8e7 c4d5 f6d5 f4g3 c6c5 c3d5 e6d5 d4c5 d8a5 d1d2 a5d2 e1d2 d7c5 f1b5 e8f8 h1d1 a7a6 b5e2 c5e4 d2e1 c8e6 g3e5 e7d6 e5d6 e4d6 f3d4 f8e7 f2f3 d6c4 e1f2 h8c8 h2h4 a6a5 a1b1 c8c7 g2g4 a8c8 e2d3 a5a4 h4h5 c4d6 b2b3 a4b3 b1b3');
INSERT INTO evaluations VALUES(12099,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(12099,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(12099,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(12099,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(12099,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(12099,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(12099,36,'cp 26','g5h4 f8e7 c4d5 f6d5 h4g3 c6c5 c3d5 e6d5 d4c5 d8a5 d1d2 a5d2 e1d2 d7c5 f1b5 e8f8 h1d1 a7a6 b5e2 c5e4 d2e1 e4g3 h2g3 c8e6 d1d3 e7c5 f3d4 f8e7 a1d1 a8c8 d4e6 f7e6 e3e4 d5e4 d3d7 e7f6 d7b7 a6a5');
INSERT INTO evaluations VALUES(12100,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(12100,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(12100,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(12100,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(12100,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12100,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(12100,36,'cp 16','f8e7 d1c2 e8g8 a2a3 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3e1 e7f8 h2h3 c8c7 a1c1 c7c1 e1c1 f8c8 c1c8 b7c8 f3e5 d7e5 d4e5 f7f6 e5f6 g7f6 f2f4 g8f7 g1f2 f7e7 g2g4 e7d6 b2b4 e6e5');
INSERT INTO evaluations VALUES(7757,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(7757,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(7757,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(7757,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(7757,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(7757,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(7757,36,'cp 23','a2a3 e8g8 d1c2 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3e1 c8c7 h2h3 d7f6 a1c1 f6e4 f3e5 e7d8 b2b4 d8c8 c1c7 c8c7 h3h4 f7f6 d3e4 d5e4 e5g6 b7d5 e1d1 c7c3 g6e7 g8h8 e7d5 e6d5');
INSERT INTO evaluations VALUES(51950,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51950,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(51950,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(51950,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(51950,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(51950,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(51950,36,'cp 10','d7d5 g1f3 c5d4 e3d4 f8b4 b1c3 g8e7 a2a3 b4c3 b2c3 d8c7 c4d5 e6d5 c1d2 b8c6 f1e2 h7h6 e1g1 c8f5 f3h4 f5h7 f1e1 e8g8 e2f1 c6a5 h2h3 c7d7 h4f3 e7c6 f3e5 c6e5 e1e5 f7f6 e5e1 f8e8 d1g4 d7g4 h3g4');
INSERT INTO evaluations VALUES(27275,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(27275,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(27275,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(27275,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(27275,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(27275,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(27275,36,'cp 102','b5d7 d8d7 c2c4 e7g6 b1c3 f8e7 a2a4 g8f6 a4a5 e8g8 e1g1 a7a6 h2h3 c7c6 d1a4 a8c8 f1d1 f6h5 d5c6 b7c6 c4c5 d7e6 c5d6 e7d6 a4a2 e6a2 c3a2 d6e7 c1e3 h5f4 a2c3 f8d8 g2g3 f4e6 c3a4 f7f6 e3b6 d8d1 a1d1 g8f8 f3d2');
INSERT INTO evaluations VALUES(27148,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(27148,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(27148,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(27148,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(27148,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(27148,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(27148,36,'cp 90','d8d7 c2c4 e7g6 b1c3 g8f6 g2g3 h7h5 d1e2 f8e7 c1g5 a7a6 h2h4 d7g4 f3d2 f6d7 g5e3 e7d8 a2a4 f7f5 e4f5 g6e7 e2g4 h5g4 a4a5 e7f5 d2e4 f5d4 e1c1 d4f5 c1c2 e8g8 b2b4 f5e3 f2e3 f8f3 c2d3 d7f6 d1a1 d8e7 h4h5 f6e4 c3e4 a8f8');
INSERT INTO evaluations VALUES(27677,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(27677,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(27677,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(27677,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(27677,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(27677,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(27677,36,'cp 104','c2c4 g8f6 b1c3 h7h6 h2h3 e7g6 a2a4 f8e7 a4a5 a7a6 c1e3 e8g8 e1g1 f6h7 d1a4 d7c8 a4c2 h7g5 f3h2 g5h7 b2b4 e7g5 h2f3 g5e3 f2e3 c8d8 c4c5 d8e7 c5c6 b7c6 d5c6 h7f6 f3d2 a8d8 c2d3 d6d5 c3d5 f6d5 e4d5 e7b4');
INSERT INTO evaluations VALUES(27674,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(27674,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(27674,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(27674,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(27674,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(27674,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(27674,36,'cp 100','g8f6 b1c3 h7h6 h2h3 e7g6 a2a4 f8e7 a4a5 e8g8 e1g1 a7a6 c1e3 f6h7 d1a4 d7c8 a4c2 h7g5 f3h2 f7f5 e4f5 f8f5 e3g5 e7g5 g2g3 c8d7 c3e4 a8f8 a1a3 g5d8 c2d1 g6e7 a3b3 b7b6 a5b6 c7b6 b3a3 a6a5 b2b3');
INSERT INTO evaluations VALUES(10414,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(10414,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(10414,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(10414,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(10414,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(10414,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(10414,36,'cp 56','c3d5 b4c5 e2e3 f6d5 c4d5 c6e7 g1f3 e5e4 f3g5 e7d5 g5e4 c5e7 e4c3 d5c3 d2c3 a7a5 a2a4 d7d6 e1g1 e7f6 e3e4 e8g8 c1e3 c8d7 d1c2 f8e8 f1e1 a8b8 f2f4 d8e7 a1d1 b7b5 e3a7 b8d8 a4b5 d7b5 a7d4 f6d4 c3d4');
INSERT INTO evaluations VALUES(10325,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(10325,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(10325,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(10325,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(10325,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(10325,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(10325,36,'cp 47','b4c5 e2e3 f6d5 c4d5 c6e7 g1e2 c7c6 e2c3 e8g8 e1g1 e7d5 c3d5 c6d5 d2d4 e5d4 e3d4 c5b6 g2d5 d7d6 a2a4 c8h3 f1e1 b6a5 e1e4 d8d7 c1e3 a8e8 d1h5 g7g6 h5f3 e8e4 f3e4 h3f5 e4f3 f5g4 f3f4 g4e6 d5e6 f7e6 f4e4 d6d5 e4c2 a5d8 a4a5 a7a6 e3f4');
INSERT INTO evaluations VALUES(19565,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(19565,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(19565,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(19565,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19565,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(19565,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19565,36,'cp 0','g8f6 b1c3 a7a6 c3d5 e7e6 d5f6 d8f6 c2c3 b7b5 g1f3 c8b7 c1g5 c6e5 d3e3 e5f3 g2f3 f6e5 g5f4 e5h5 h1g1 f8c5 e3e2 e8g8 g1g5 h5h4 g5g4 h4h5');
INSERT INTO evaluations VALUES(12898,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12898,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12898,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12898,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12898,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12898,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12898,36,'cp 0','b1c3 a7a6 c3d5 e7e6 d5f6 d8f6 c2c3 b7b5 c1e3 c6e5 d3d2 f6h4 d2d4 e5c6 d4d3 c6e5');
INSERT INTO evaluations VALUES(12893,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12893,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12893,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12893,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12893,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12893,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12893,36,'cp 0','a7a6 c3d5 e7e6 d5f6 d8f6 c2c3 b7b5 c1e3 c6e5 d3d2 f6h4 d2d4 e5c6 d4d3 c6e5');
INSERT INTO evaluations VALUES(49939,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(49939,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(49939,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(49939,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(49939,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(49939,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(49939,36,'cp 55','g2g3 f8e7 f1g2 e8g8 b1c3 d7d5 e1g1 c7c6 d1c2 f6e4 c1f4 b8d7 a1d1 a7a5 f3e5 d7e5 f4e5 e4c3 c2c3 e7f6 d1c1 d8e7 e5f6 e7f6 e2e3 f6e7 g2f3 c8d7 a2a3 g7g5 f3g2 d7e8 b2b4 a5b4 a3b4 h7h6 c1a1 e8h5 b4b5 c6b5');
INSERT INTO evaluations VALUES(49940,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(49940,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(49940,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(49940,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(49940,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(49940,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(49940,36,'cp 45','f8e7 f1g2 d7d5 d1c2 c7c6 b1d2 f6e4 f3e5 b8d7 e5d3 e7f6 e2e3 e8g8 e1g1 f8e8 a2a4 a7a5 b2b3 b7b6 c1b2 c8b7 f1d1 a8c8 d3f4 d8e7 f4h5 c6c5 d2e4 f5e4 h5f6 g7f6 c2e2 c5d4 e3d4 f6f5 b2a3 e7f7 a3d6');
INSERT INTO evaluations VALUES(42938,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(42938,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(42938,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(42938,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(42938,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42938,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(42938,36,'cp 62','b1c3 e8g8 f1g2 d7d5 e1g1 b8d7 d1c2 c7c6 f3g5 d7b8 a2a3 f6e4 g5h3 e7f6 f1d1 a7a5 b2b3 e4c3 c2c3 a5a4 b3b4 f8e8 h3f4 e6e5 f4h5 e5e4 h5f6 d8f6 c4d5 c6d5 f2f3 b8c6 c1f4');
INSERT INTO evaluations VALUES(42934,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(42934,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(42934,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(42934,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(42934,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42934,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(42934,36,'cp 33','e8g8 f1g2 d7d5 e1g1 c7c6 d1c2 b8d7 f3g5 d7b8 f1d1 a7a5 a2a3 h7h6 g5f3 b8d7 c4d5 c6d5 c1f4 d7b6 f3e5 c8d7 e5g6 f8f7 d1c1 d7e8 a3a4 e7b4 c3b5 a8c8 c2d1 c8c1 d1c1 f7d7 g6e5 d7e7 c1c2 e8b5 a4b5');
INSERT INTO evaluations VALUES(55787,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55787,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55787,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55787,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(55787,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(55787,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(55787,36,'cp 0','f8e7 g5e7 d8e7 c2c4 g8f6 e2e3 e8g8 c4d5 f6d5 b1d2 b7b6 e3e4 d5f6 a1c1 c8b7 f1d3 f8d8 d1e2 b8c6 e4e5 f6d5 e2e4 g7g6 d3b1 f7f5 e5f6 d5f6 e4e3 f6d5 e3e4');
INSERT INTO evaluations VALUES(38282,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38282,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38282,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38282,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38282,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38282,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38282,36,'cp 3','g5e7 d8e7 c2c4 g8f6 e2e3 e8g8 c4d5 e6d5 f1d3 c8g4 e1g1 c7c6 h2h3 g4f3 d1f3 a7a5 b1c3 f8e8 a1c1 b8d7 f1e1 g7g6 a2a3 a8d8 f3e2 d7f8 f2f3 f8e6 e2f2 g8g7 f3f4 f6h5 g2g4 h5f6 g4g5');
INSERT INTO evaluations VALUES(38205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38205,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38205,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38205,36,'cp 0','d8e7 c2c4 g8f6 e2e3 e8g8 b1c3 f8d8 c4d5 e6d5 f1e2 c7c6 e1g1 c8f5 a2a3 b8d7 d1b3 d8b8 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(33069,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33069,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33069,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33069,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33069,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33069,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(33069,36,'cp 0','c2c4 g8f6 e2e3 e8g8 b1c3 f8d8 d1a4 c8d7 a4c2 d5c4 f1c4 c7c5 d4c5 e7c5 c4e2 b8c6 e1g1 c6e5 f3d4 e5c6');
INSERT INTO evaluations VALUES(46976,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46976,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(46976,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46976,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46976,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46976,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(46976,36,'cp 0','e7e6 g1f3 d6d5 e4d5 e6d5 c4b3 f8e7 c3e2 b8c6 d2d4 c8g4 d1d3 c6b4 d3c3 e8g8 e1g1 b4c6 c3d3 c6b4');
INSERT INTO evaluations VALUES(48039,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48039,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48039,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48039,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48039,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48039,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48039,36,'cp 0','g1f3 d6d5 e4d5 e6d5 c4b3 f8e7 c3e2 b8c6 d2d4 c8g4 d1d3 c6b4 d3c3 e8g8 e1g1 b4c6 c3d3 c6b4');
INSERT INTO evaluations VALUES(48038,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48038,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48038,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48038,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48038,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48038,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48038,36,'cp 0','d6d5 e4d5 e6d5 c4b3 f8e7 c3e2 e8g8 c2c3 c7c5 e1g1 b8c6 h2h3 c8f5 d2d4 c5c4 b3c2 f5c2 d1c2 h7h6 c1f4 e7d6 f4d6 d8d6 e2g3 f8e8 f3h4 d6f4 h4f5 c6e7 a1e1 e7f5 g3f5 g7g6 f5e3 h6h5 a2a4 g8g7 g2g3 f4h6 c2d1');
INSERT INTO evaluations VALUES(2474,30,'cp -183',NULL);
INSERT INTO evaluations VALUES(2474,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(2474,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(2474,33,'cp -175',NULL);
INSERT INTO evaluations VALUES(2474,34,'cp -174',NULL);
INSERT INTO evaluations VALUES(2474,35,'cp -181',NULL);
INSERT INTO evaluations VALUES(2474,36,'cp -212','c3e4 d7d5 e4g3 e5e4 f3g1 d8f6 d2d3 e4d3 d1d3 c6b4 d3d2 f8c5 e1d1 h8e8 a2a3 b4a6 f2f3 c5e3 d2d3 e3c1 a1c1 f7g8 b2b4 c7c5 d3d2 c8d7 g1e2 f6b2 b4c5 b2a3 d2c3 a3c5 c3c5 a6c5 h1e1 a7a5 c1b1 a5a4 e2c3 e8d8 c3d5');
INSERT INTO evaluations VALUES(2472,30,'cp -217',NULL);
INSERT INTO evaluations VALUES(2472,31,'cp -186',NULL);
INSERT INTO evaluations VALUES(2472,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(2472,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(2472,34,'cp -198',NULL);
INSERT INTO evaluations VALUES(2472,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(2472,36,'cp -189','d7d5 e4g3 e5e4 f3g1 d8f6 d2d3 e4d3 d1d3 c6b4 d3d2 f8c5 g1e2 f6f2 e1d1 f2g2 d2f4 f7g8 c1e3 c5e3 f4e3 c8g4 d1d2 g2f3 e3f3 g4f3 h1f1 a8f8 e2d4 f3g4 f1f8 g8f8 a1f1 f8e8 f1f4 g4c8 d4f5 g7g6 f5e3 d5d4 f4d4 b4c6 d4f4 e8e7 e3d5 e7d6 d5c3 h8d8 d2c1 d6e7 c3b5 d8d7 f4f1 h7h6 a2a3 b7b6 b2b4 e7d8 b5c3 c6e5 c3b5');
INSERT INTO evaluations VALUES(2435,30,'cp -186',NULL);
INSERT INTO evaluations VALUES(2435,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(2435,32,'cp -192',NULL);
INSERT INTO evaluations VALUES(2435,33,'cp -176',NULL);
INSERT INTO evaluations VALUES(2435,34,'cp -178',NULL);
INSERT INTO evaluations VALUES(2435,35,'cp -207',NULL);
INSERT INTO evaluations VALUES(2435,36,'cp -215','e4g3 e5e4 f3g1 d8f6 d2d3 e4d3 d1d3 c6b4 d3d2 f8c5 g1e2 h7h5 h1f1 c8e6 a2a3 b4c6 f2f3 c5b6 g3e4 f6h4 g2g3 h4e7 e4g5 f7g8 g5e6 e7e6 e1d1 a8e8 e2f4 e6f7 f1e1 e8e1 d2e1 g7g5 f4e6 f7f3 e1e2 g5g4 e2f3 g4f3 c1f4 h8h7 e6g5 f3f2 d1d2 h7e7 a1f1 c6e5 f4e5 e7e5 g5h3');
INSERT INTO evaluations VALUES(2438,30,'cp -215',NULL);
INSERT INTO evaluations VALUES(2438,31,'cp -219',NULL);
INSERT INTO evaluations VALUES(2438,32,'cp -218',NULL);
INSERT INTO evaluations VALUES(2438,33,'cp -229',NULL);
INSERT INTO evaluations VALUES(2438,34,'cp -176',NULL);
INSERT INTO evaluations VALUES(2438,35,'cp -190',NULL);
INSERT INTO evaluations VALUES(2438,36,'cp -177','e5e4 f3g1 d8f6 d2d3 e4d3 d1d3 c6b4 d3d2 f8c5 g1e2 f6f2 e1d1 f2g2 d2f4 f7g8 c1e3 c5e3 f4e3 c8g4 d1d2 g2f3 e3f3 g4f3 h1f1 a8f8 e2d4 f3g4 f1f8 g8f8 a1f1 f8e8 f1f4 g4c8 d4f5 g7g6 f5e3 b4a6 e3d5 c7c6 d5f6 e8e7 f6e4 a6c7 e4g5 c8e6 f4e4 e7f6 h2h4 h7h6 g5e6 c7e6 d2e3 h8d8 e4a4 a7a6 a4b4 b7b5');
INSERT INTO evaluations VALUES(2432,30,'cp -187',NULL);
INSERT INTO evaluations VALUES(2432,31,'cp -190',NULL);
INSERT INTO evaluations VALUES(2432,32,'cp -189',NULL);
INSERT INTO evaluations VALUES(2432,33,'cp -186',NULL);
INSERT INTO evaluations VALUES(2432,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(2432,35,'cp -194',NULL);
INSERT INTO evaluations VALUES(2432,36,'cp -185','f3g1 d8f6 d2d3 e4d3 d1d3 c6b4 d3d2 f8c5 g1e2 h7h5 h1f1 c8e6 a2a3 b4c6 f2f3 c5b6 g3e4 f6h4 g2g3 h4e7 e4g5 f7g8 g5e6 e7e6 e1d1 a8e8 e2f4 e6f7 a3a4 b6a5 d2g2 h8h6 c1d2 h5h4 a1a3 h4h3 g2h1 a5d2 d1d2 g7g5 f4d3 h6e6 d2c1 e6e2 c1b1 c6d4 f3f4 d4c2 a3c3');
INSERT INTO evaluations VALUES(2433,30,'cp -221',NULL);
INSERT INTO evaluations VALUES(2433,31,'cp -228',NULL);
INSERT INTO evaluations VALUES(2433,32,'cp -206',NULL);
INSERT INTO evaluations VALUES(2433,33,'cp -211',NULL);
INSERT INTO evaluations VALUES(2433,34,'cp -195',NULL);
INSERT INTO evaluations VALUES(2433,35,'cp -184',NULL);
INSERT INTO evaluations VALUES(2433,36,'cp -194','d8f6 d2d3 e4d3 d1d3 c6b4 d3d2 f8c5 g1e2 h7h5 h1f1 c8e6 a2a3 b4c6 f2f3 c5b6 g3e4 f6h4 g2g3 h4e7 e4g5 f7g8 g5e6 e7e6 e1d1 a8e8 e2f4 e6f7 a3a4 b6a5 d2f2 c6e5 f4d3 e5c4 b2b3 h5h4 g3g4 a5b6 f2g2 f7f6 a1a2 c4e3 c1e3 b6e3 f3f4 f6d4 a4a5 h4h3 g2f3 e8e4 f1e1 c7c5 a2a4 c5c4 b3c4 d5c4 g4g5');
INSERT INTO evaluations VALUES(23198,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(23198,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(23198,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(23198,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(23198,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(23198,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(23198,36,'cp 122','a7a6 b5c4 f8c5 c2c3 c5a7 d2d4 d7d6 c1e3 g8e7 b1d2 c6a5 d1a4 a5c6 a4c2 c6a5 d4e5 f6e5 c2a4 c7c6 c4e2 b7b5 a4a3 e8g8 b2b4 a5b7 c3c4 a7e3 f2e3 c8g4 h2h3 g4h5 a3c3 h7h6 a2a4 b5a4 c4c5 a6a5 c5d6 b7d6 a1a4 a5b4 a4b4 a8a2 c3b3 h5f7 e2c4');
INSERT INTO evaluations VALUES(17950,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(17950,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(17950,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(17950,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(17950,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(17950,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(17950,36,'cp 117','b5c4 f8c5 c2c3 c5a7 d2d4 d7d6 c1e3 g8e7 f1e1 b7b5 c4f1 c8g4 d4d5 a7e3 e1e3 c6b8 a2a4 b5a4 d1a4 b8d7 b1d2 e8g8 h2h3 g4f3 e3f3 a6a5 a4c2 f6f5 e4f5 d7b6 g2g4 e7d5 d2c4 g8h8 c4b6 d5b6 c3c4');
INSERT INTO evaluations VALUES(22864,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(22864,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22864,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(22864,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(22864,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(22864,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(22864,36,'cp -11','d7d5 f1g2 g8f6 e1g1 f8e7 c2c4 e8g8 c4d5 f6d5 b1c3 c8e6 c3d5 e6d5 c1e3 d8d7 a1c1 f8d8 d1a4 h7h6 f1d1 a8b8 a2a3 d7e6 f3d2 a7a5 g2d5 d8d5 d2e4 f7f5 e4c3 d5d7 a4c4 e6f7 f2f3 e7g5 c4f7 g8f7 e3g5');
INSERT INTO evaluations VALUES(21750,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(21750,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(21750,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(21750,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(21750,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(21750,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(21750,36,'cp -21','f1g2 g8f6 e1g1 h7h6 c2c4 d5c4 d3c4 f8d6 b1c3 e8g8 b2b3 e5e4 f3d4 c6d4 d1d4 f8e8 c1b2 d6e5 d4d8 e8d8 a1d1 c8f5 f2f3 c7c6 f3e4 f5g6 g2f3 g6e4 c3e4 e5b2 e4d6 a8b8 h2h3 b2a3 d6f5 a7a5 g3g4 d8d7 d1d7 f6d7');
INSERT INTO evaluations VALUES(17978,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(17978,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(17978,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17978,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(17978,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(17978,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17978,36,'cp 61','a2a4 b5b4 d2d4 g8f6 d4e5 d6e5 d1d8 c6d8 f3e5 f8d6 e5d3 f6e4 b1d2 e4c5 d3c5 d6c5 d2e4 c5a7 c1e3 a7e3 f2e3 c8e6 h1f1 h7h5 b3e6 d8e6 f1f5 h8h6 a1d1 h5h4 f5a5 e8e7 e4c5 e6c5 a5c5 c7c6 c5e5 e7f6 e5a5 f6e6 d1d4 h4h3 g2g3');
INSERT INTO evaluations VALUES(31201,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(31201,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(31201,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(31201,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(31201,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(31201,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(31201,36,'cp 72','e4g3 f5g4 c2c3 g8f6 d2d4 g4f3 d1f3 d8d5 f1e2 b8d7 f3d5 c6d5 h2h4 e7e6 h4h5 h7h6 c1f4 f8e7 a2a4 e8g8 a4a5 a7a6 g3f1 f6e4 f1d2 d7f6 f2f3 e7d6 f4d6 e4d6 e1f2 f8c8 g2g4 a8b8 e2d3');
INSERT INTO evaluations VALUES(31204,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(31204,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(31204,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(31204,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(31204,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(31204,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(31204,36,'cp 52','f5g4 c2c3 g8f6 d2d4 g4f3 d1f3 d8d5 f3d5 c6d5 f1d3 e7e6 h2h4 f8d6 g3e2 b8c6 f2f3 h7h5 c1g5 a7a6 e1f2 f6h7 g5e3 c6a5 b2b4 a5c4 d3c4 d5c4 e3f4 d6e7 a2a4 h7f6 f4e5 e8g8 e2f4');
INSERT INTO evaluations VALUES(57035,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57035,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(57035,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57035,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57035,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(57035,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57035,36,'cp -15','d7d5 e4e5 d5d4 g1f3 b8c6 d1e2 d8d7 b2b4 a7a6 a2a4 b7b6 b1a3 c8b7 b4b5 a6b5 a4b5 c6a5 f3d4 f8c5 a3c2 c5d4 c2d4 g8e7 e2g4 e8g8 f1e2 c7c5 b5c6 a5c6 a1a8 f8a8 d4f3 d7c7 e1g1 c6e5 f3e5 c7e5 d2d4 e5f5 g4f5');
INSERT INTO evaluations VALUES(55896,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55896,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(55896,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(55896,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(55896,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(55896,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(55896,36,'cp -1','e4e5 d5d4 g1f3 b8c6 f1b5 d4d3 d1a4 d8d5 a4c4 a7a6 c4d5 e6d5 b5d3 c8g4 e1g1 e8c8 f1e1 d8e8 d3e2 g4f3 e2f3 c6e5 g1f1 g8f6 f3e2 h7h5 d2d3 f8d6 h2h3 e5g6 b1d2 g6f4 d2f3 e8e2 e1e2 f4e2 f1e2 h8e8 c1e3 f6d7 b2b4 d7f8 e2d2 f8e6 g2g3 c8d7 a2a4 f7f6 h3h4 a6a5 b4b5 c7c5 b5c6 b7c6 a1b1');
INSERT INTO evaluations VALUES(55967,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(55967,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55967,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(55967,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(55967,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55967,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55967,36,'cp -12','d5d4 g1f3 b8c6 f1b5 d4d3 d1a4 d8d5 a4c4 a7a6 c4d5 e6d5 b5d3 c8g4 e1g1 e8c8 f1e1 d8e8 d3e2 g4f3 e2f3 c6e5 g1f1 g8f6 f3e2 f8d6 d2d3 f6g4 c1f4 h7h5 h2h3 g7g5 f4g3 g4f6 b1d2 h5h4 g3e5 d6e5 d2f3 e5f4 f3d4 c8d7 d4b3 c7c6 a2a4 e8e6 a4a5 h8e8');
INSERT INTO evaluations VALUES(15479,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(15479,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15479,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15479,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15479,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(15479,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15479,36,'cp 0','c6a5 c4b5 c7c6 b5e2 c6c5 e1g1 f8e7 a2a3 a5c6 e2c4 f6e4 c3e4 d6d5 c4b5 d5e4 f3e5 d8d6 e5c6 b7c6 b5e2 e8g8 d2d3 f8d8 d3e4 d6g6 d1e1 e7f6 c2c3 a8b8 e2f3 c8a6 f3e2');
INSERT INTO evaluations VALUES(31791,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(31791,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(31791,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(31791,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(31791,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(31791,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(31791,36,'cp 51','c7c6 g1f3 b7b5 f1e2 g8f6 e1g1 a7a5 a2a4 b5b4 c3b1 g7g6 b1d2 f8g7 e2c4 e6c4 d2c4 c6c5 b2b3 b8d7 c1b2 e8g8 a1c1 d8c7 f3e5 f6d5 d1f3 c5d4 b2d4 d7e5 f3d5 e5c4 c1c4 c7d6 d5d6 e7d6 d4g7 g8g7 c4c6 d6d5 f1c1 f8d8 c6c5 d5d4');
INSERT INTO evaluations VALUES(31019,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(31019,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(31019,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(31019,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(31019,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(31019,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(31019,36,'cp 76','g1f3 g8f6 f3g5 e6g4 f2f3 g4h5 f1c4 b8d7 e1g1 d8c7 d1e2 a7a5 g1h1 a5a4 e3e4 h7h6 g5h3 e7e5 c1e3 h5g6 f1d1 f8d6 d4d5 e8g8 a1c1 d6c5 e2d2 c5e3 d2e3 b7b5 d5c6 b5c4 c6d7 f6d7');
INSERT INTO evaluations VALUES(31020,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(31020,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(31020,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(31020,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(31020,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(31020,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(31020,36,'cp 81','g8f6 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 d1c2 e8g8 f1d1 h7h6 g5e4 f6e4 c3e4 b8d7 c1d2 d7f6 d2e1 f6e4 c2e4 d8b6 a2a3 a7a5 e1c3 c8d7 c4d3 f7f5 e4e5 e7f6 e5d6 b6d8 d3c4 d8e7 d6c5 e7c5 d4c5 f6c3 d1d7 c3b2');
INSERT INTO evaluations VALUES(49164,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(49164,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(49164,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(49164,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(49164,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(49164,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(49164,36,'cp 31','c7c5 c2c3 b8c6 d2f3 c5d4 c3d4 f7f6 e5f6 d8f6 g1e2 e6e5 d4e5 f8b4 c1d2 d7e5 f3e5 f6e5 d2b4 c6b4 d3b5 b4c6 e1g1 e8g8 d1d2 c8d7 f1e1 e5f6 e2c3 d7e6 c3d5 e6d5 d2d5 g8h8 d5d2 a8d8 d2e3 f6b2 a1b1 b2d4 e3d4 c6d4 b5c4 d4c2');
INSERT INTO evaluations VALUES(44006,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(44006,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44006,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(44006,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(44006,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(44006,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(44006,36,'cp 44','c2c3 c5d4 c3d4 b8c6 g1e2 f7f6 e5f6 d7f6 d2f3 f8d6 e1g1 e8g8 h2h3 c8d7 f1e1 f6h5 c1g5 d8b6 e2c3 h7h6 g5e3 h5f4 d3c2 c6b4 c2b1 d7e8 e3f4 f8f4 e1e6 e8h5 a2a3 h5f3 g2f3 a8f8 a3b4 f4f3 e6e3 f3f2 d1d3 f8f5 d3f5 f2f5 b1f5 b6d4');
INSERT INTO evaluations VALUES(19050,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(19050,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(19050,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(19050,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(19050,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(19050,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(19050,36,'cp 43','g8f6 f3g5 d7d5 e4d5 c6d4 e1g1 f8d6 g5e4 f6e4 c3e4 e8g8 c1e3 d4f5 e3g5 f7f6 g5d2 g8h8 a1c1 d6e7 f1e1 a7a5 d2c3 b7b6 a2a3 c8b7 e4g3 f5g3 h2g3 b6b5 c4b5 d8d5 d1d5 b7d5 e1d1 d5e6 b5c6 a8a7 b2b4 e6b3 d1d7 a7d7 c6d7 a5b4 a3b4 f8d8 d7c6 d8c8');
INSERT INTO evaluations VALUES(12578,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(12578,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(12578,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(12578,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(12578,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(12578,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(12578,36,'cp 57','f3g5 d7d5 e4d5 c6d4 e1g1 f8d6 c1e3 d4f5 g5e4 f6e4 c3e4 e8g8 e3g5 f7f6 g5d2 g8h8 f1e1 d6e7 d2c3 b7b6 e4g3 f5g3 h2g3 c8f5 a2a3 a8c8 d1e2 e7d6 c3b4 d8e7 b4d6 e7d6 g3g4 f5d7 a1c1 g7g6 g4g5 f6g5 e2e5 d6e5 e1e5 h8g7 e5e7 f8f7 d5d6');
INSERT INTO evaluations VALUES(12576,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(12576,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(12576,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(12576,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(12576,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(12576,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12576,36,'cp 39','d7d5 e4d5 c6d4 e1g1 f8d6 c1e3 d4f5 g5e4 f6e4 c3e4 e8g8 e3g5 f7f6 g5d2 g8h8 e4d6 d8d6 d2c3 c8d7 d1f3 a8c8 c4d3 f5e7 f1d1 d6d5 d3e4 d5e6 e4b7 c8b8 a1c1 d7c6 b7c6 e7c6 f3e4 f8d8 a2a3 d8d7 h2h3 h7h6 f2f4 b8c8 d1d2');
INSERT INTO evaluations VALUES(13797,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(13797,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(13797,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(13797,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(13797,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(13797,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(13797,36,'cp 40','e4d5 c6d4 e1g1 f8d6 g5e4 f6e4 c3e4 e8g8 c1e3 d4f5 e3g5 f7f6 g5d2 g8h8 a1c1 d6e7 e4g3 e7d6 f1e1 f5g3 h2g3 c8d7 a2a3 a8c8 d1e2 d8e7 g3g4 f6f5 g4f5 d7f5 c4d3 c8c1 d2c1 e7f7 d3f5 f7f5 c1e3 a7a6 e1d1 h7h6');
INSERT INTO evaluations VALUES(13787,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(13787,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(13787,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(13787,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(13787,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(13787,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(13787,36,'cp 40','c6d4 e1g1 f8d6 g5e4 f6e4 c3e4 e8g8 c1e3 d4f5 e3g5 f7f6 g5d2 g8h8 f1e1 b7b6 d2c3 d6e7 a2a3 f5d6 e4d6 e7d6 d1h5 g7g6 h5e2 c8d7 h2h4 d8e7 c3d2 a8c8 a1c1 h7h5 c4b5 c8c1 e1c1 f8d8 b5c6 d7g4 e2e3 g4f5 c1c4 h8g7 e3h6 g7g8 d2b4');
INSERT INTO evaluations VALUES(10108,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(10108,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(10108,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(10108,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(10108,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(10108,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(10108,36,'cp -27','d2d3 e8g8 c1g5 c5e7 b1d2 h7h6 g5f6 e7f6 g1e2 c6a5 a2a4 a5c4 d2c4 d7d6 e1g1 c8e6 a4a5 a7a6 c4e3 g7g6 h2h3 d8e7 b2b4 f6g5 d3d4 a8e8 a1d1 e6b3 d1b1 e5d4 e2d4 b3a4 b1a1 a4c6 e3d5 c6d5 e4d5 e7e4');
INSERT INTO evaluations VALUES(10113,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(10113,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(10113,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(10113,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(10113,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(10113,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(10113,36,'cp -42','e8g8 c1g5 c5e7 b1d2 h7h6 g5f6 e7f6 g1e2 c6a5 a2a4 a5c4 d2c4 d7d6 a4a5 c8e6 e1g1 a7a6 c4e3 g7g6 d3d4 h6h5 a1d1 d8e7 h2h3 a8e8 d1d3 f6g5 b2b4 g5e3 d3e3 e6c4 f1e1 g8h7 e2c1 f7f5 e4f5 f8f5 f3e4 e7f7 f2f3 e8f8');
INSERT INTO evaluations VALUES(4337,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(4337,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(4337,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(4337,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(4337,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(4337,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(4337,36,'cp -28','c1g5 c5e7 b1d2 h7h6 g5f6 e7f6 g1e2 c6a5 a2a4 a5c4 d2c4 d7d6 e1g1 c8e6 c4e3 g7g6 d3d4 a7a5 h2h3 h6h5 d4e5 d6e5 f1d1 d8e7 e3d5 e6d5 d1d5 c7c6 d5d3 f8d8 a1d1 d8d3 d1d3 a8d8 d3d8 e7d8 e2g3 f6g5 g3f1 b7b5 a4b5 c6b5 f1e3');
INSERT INTO evaluations VALUES(4282,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(4282,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(4282,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(4282,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(4282,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(4282,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(4282,36,'cp -36','c5e7 b1d2 h7h6 g5f6 e7f6 g1e2 c6a5 a2a4 a5c4 d2c4 d7d6 a4a5 c8e6 c4e3 g7g6 e1g1 a7a6 d3d4 d8e7 h2h3 a8e8 a1d1 h6h5 b2b4 f6g5 f1e1 f7f6 d1b1 e7f7 b4b5 f6f5 e4f5 g6f5 f3g3 f7g7 h3h4 g5h6 b5a6');
INSERT INTO evaluations VALUES(29647,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(29647,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29647,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(29647,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29647,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29647,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29647,36,'cp 14','e7e6 b1d2 f8d6 f4d6 d8d6 c2c4 e8g8 f1e2 b8d7 e1g1 a7a5 h2h3 g4h5 d1c2 c7c6 a1c1 h5g6 e2d3 g6d3 c2d3 h7h6 f1e1 f8e8 d3c2 b7b6 e3e4 d5e4 d2e4 f6e4 c2e4 d7f6 e4e5 e8d8 e5d6 d8d6 c4c5 b6c5 d4c5');
INSERT INTO evaluations VALUES(30167,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(30167,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(30167,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(30167,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30167,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(30167,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(30167,36,'cp 15','b1d2 f8d6 f4d6 d8d6 f1e2 b8d7 c2c4 e8g8 e1g1 a7a5 h2h3 g4h5 a2a4 f8c8 c4d5 e6d5 d1b3 b7b6 d2b1 h7h6 b1c3 c7c6 f1c1 d6e6 f3h4 h5e2 c3e2 e6e4 h4f3 e4e6 b3d1 f6e8 b2b3 d7f6 e2f4 e6e7');
INSERT INTO evaluations VALUES(30168,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30168,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(30168,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30168,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30168,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30168,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(30168,36,'cp 19','f8d6 h2h3 g4h5 f4d6 c7d6 g2g4 h5g6 f3h4 b8c6 h4g6 h7g6 d1e2 d8c7 f1g2 c6a5 c2c3 a8b8 e1g1 a5c4 d2c4 c7c4 e2c2 b8c8 f1c1 c8c7 b2b3 c4a6 c3c4 e8g8 g4g5 f6d7 g2f1 a6c6 h3h4 f8c8 c2d2 d5c4 c1c4 c6b6 c4c7 c8c7');
INSERT INTO evaluations VALUES(49887,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(49887,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(49887,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(49887,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(49887,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(49887,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(49887,36,'cp -79','e4c5 f3e5 d7d5 c4b3 f8d6 d3d4 d6e5 d4e5 c5b3 a2b3 d5d4 c2c3 b8c6 b3b4 e8g8 e1g1 c6e5 d1d4 d8d4 c3d4 e5c6 b1c3 c6b4 c1f4 c8e6 f4c7 f8c8 d4d5 c8c7 d5e6 f7e6 f1d1 a7a6 h2h3 g8f7 c3b5 c7e7 a1a3 b4c6 b5d6 f7g8 d6e4 a8d8 d1d8 c6d8 e4c5 g8f7 f2f3');
INSERT INTO evaluations VALUES(49803,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(49803,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(49803,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(49803,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(49803,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(49803,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(49803,36,'cp -68','f3e5 d7d5 c4b3 f8d6 d3d4 c5b3 a2b3 d6e5 d4e5 d5d4 d1d3 e8g8 e1g1 b8c6 f2f4 f7f6 d3c4 g8h8 b1d2 f6e5 f4e5 c8f5 d2f3 f5e4 c1g5 d8d5 c4d5 e4d5 g5h4 a8e8 h4g3 h8g8 a1e1 d5f3 g2f3 g8f7 f3f4 f7e6 g3h4 f8f7 g1g2 h7h6 h4g3 e8d8 e1d1');
INSERT INTO evaluations VALUES(33785,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33785,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(33785,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(33785,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33785,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33785,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33785,36,'cp 0','e7e2 f1e2 f8e7 b1c3 h7h6 g5e3 b8c6 e1c1 a7a6 h2h3 c8e6 a2a3 e8c8 d3d4 d6d5 f3e5 c6e5 d4e5 f6e4 c3e4 d5e4 c2c3 d8d1 h1d1 h8d8 d1d8 c8d8 b2b4 d8d7 c1d2 c7c5 b4b5 a6b5 e2b5 d7c7 g2g4 g7g5 a3a4 b7b6 d2c2');
INSERT INTO evaluations VALUES(33866,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33866,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(33866,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(33866,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(33866,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(33866,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33866,36,'cp 6','f1e2 f8e7 b1c3 h7h6 g5e3 b8c6 f3d4 c8d7 d4c6 d7c6 e1g1 a7a5 f1e1 e8g8 a2a4 f6d5 c3d5 c6d5 e3d4 f8e8 f2f4 g7g6 g1f2 f7f5 e2f3 d5f3 f2f3 g8f7 h2h3 h6h5 g2g3 e7f6 d4f6 f7f6 d3d4 c7c6 c2c4 f6f7 f3f2 d6d5 c4d5');
INSERT INTO evaluations VALUES(33865,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(33865,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(33865,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33865,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(33865,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(33865,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(33865,36,'cp 7','f8e7 b1c3 h7h6 g5e3 b8c6 f3d4 c8d7 e2f3 e8c8 d4c6 d7c6 f3c6 b7c6 e1c1 d8e8 d1e1 c8d7 e3d2 d6d5 h2h3 e7c5 c3d1 h6h5 e1e8 h8e8 h1e1 e8e1 d2e1 c5d6 d1c3 d7e6 f2f3 d5d4 c3e2 c6c5 e1f2 f6d5 c2c3 d4c3 b2c3 e6f5 c3c4 d5b4');
INSERT INTO evaluations VALUES(33373,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(33373,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(33373,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(33373,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(33373,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(33373,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33373,36,'cp 0','b1c3 h7h6 g5e3 b8c6 f3d4 c8d7 d4c6 d7c6 d3d4 f6d5 c3d5 c6d5 e1d2 e7f6 h1g1 a7a5 g2g3 c7c6 a2a4 e8d7 e2d3 b7b6 c2c3 d5b3 a1a3 b3d5');
INSERT INTO evaluations VALUES(33372,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(33372,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(33372,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33372,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33372,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(33372,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(33372,36,'cp 3','h7h6 g5e3 b8c6 f3d4 c8d7 d4c6 d7c6 d3d4 c6g2 h1g1 g2h3 g1g7 h3e6 e2f3 d6d5 c3e2 e8d7 e2f4 a8g8 f4e6 f7e6 g7g8 h8g8 e1e2 h6h5 a2a4 a7a5 c2c3 c7c6 h2h3 e7d6 e3d2 b7b6 d2e3 d6e7 a1h1 h5h4 h1a1 e7d6 e3d2 c6c5 d2e3 c5c4 a1f1');
INSERT INTO evaluations VALUES(54923,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(54923,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(54923,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(54923,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(54923,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(54923,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(54923,36,'cp 50','c2c4 e7e6 c4d5 e6d5 e2e3 f8d6 f4d6 d8d6 f1d3 g8e7 b1c3 e8g8 g1e2 b7b6 e1g1 c8a6 a1c1 a6d3 d1d3 c7c6 h2h3 a7a5 e2g3 f8e8 b2b3 b8d7 c3a4 e7g6 c1c2 g6e7 f1c1 a8b8 g3e2 h6h5 e2f4 h5h4 f4e2 d6h6');
INSERT INTO evaluations VALUES(56997,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(56997,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(56997,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(56997,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(56997,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(56997,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(56997,36,'cp -53','d7d5 d2d3 g8f6 h5e2 e6e5 g1f3 f8d6 e4d5 f6d5 g2g3 b8c6 b1c3 d6b4 c1d2 e8g8 f1g2 c8g4 h2h3 c6d4 e2d1 g4f3 g2f3 d4f3 d1f3 b4c3 d2c3 f7f6 f3f5 f8f7 c3d2 d8d7 f5d7 f7d7 e1e2 a8d8 a1e1 b7b5');
INSERT INTO evaluations VALUES(55788,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(55788,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(55788,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(55788,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(55788,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(55788,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(55788,36,'cp -46','d2d3 b8c6 h5e2 e6e5 g1f3 g8e7 g2g3 d5d4 f1g2 e7g6 h2h4 f8e7 b1d2 c8g4 d2f1 g6f8 f1h2 g4d7 e1g1 a7a5 h4h5 f8e6 a2a3 e6c5 f3d2 g7g5 g2f3 d7e6 f3g4 d8d7 a1b1 h7h6 g4e6 d7e6 e2g4 e6g4 h2g4');
INSERT INTO evaluations VALUES(21716,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21716,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21716,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21716,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21716,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21716,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21716,36,'cp 0','e4d5 d8d5 b1c3 d5d6 f1e2 c8e6 e1g1 e8c8 a1b1 f7f5 b2b4 g8f6 c1e3 c8b8 f1e1 e6c8 e2f1 e5e4 c3b5 d6e7 e3f4 f6d5 f4g5 d5f6');
INSERT INTO evaluations VALUES(21577,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21577,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21577,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21577,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21577,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21577,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21577,36,'cp 0','d8d5 b1c3 d5d6 f1e2 g8f6 e1g1 c8f5 c1e3 a7a6 f1e1 h7h6 f3d2 d6e6 d2c4 b7b5 c4d2 f8e7 a2a4 b5b4 c3e4 e8g8 e4f6 e6f6 e2g4 f5g6 g4f3 f6e6 d2b3 f8e8 e3c5 e7c5 b3c5 e6d6 c5b7 d6d7 b7c5');
INSERT INTO evaluations VALUES(1727,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(1727,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(1727,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(1727,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(1727,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(1727,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(1727,36,'cp 3','b1c3 d5d8 c1e3 f8b4 d3d4 e5e4 f3d2 g8f6 c3e2 e8g8 c2c3 b4e7 e2f4 e7d6 f4h5 f6h5 d1h5 f7f5 e3g5 d6e7 f1c4 g8h8 h2h4 c8d7 c4f7 d7e8 f7e8 d8e8 h5e8 a8e8 g5e7 c6e7 d2c4 e7d5 c4e5 h8g8 e1g1 d5f4 g2g3 f4g6 f2f4 g6e5 f4e5');
INSERT INTO evaluations VALUES(1723,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1723,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(1723,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1723,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(1723,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(1723,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(1723,36,'cp 0','d5d8 c1e3 f8b4 d3d4 e5e4 f3e5 c6e5 d4e5 c8e6 d1d4 d8d4 e3d4 b4c3 d4c3 e8c8 f1e2 g8e7 e1g1 e7d5 c3d2 d5e7');
INSERT INTO evaluations VALUES(21830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21830,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21830,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21830,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21830,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21830,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21830,36,'cp 0','f1e2 g8e7 e1g1 e7g6 c1g5 f8e7 g5e7 g6e7 d3d4 e5d4 c3b5 e7f5 e2d3 e8g8 d3e4 a7a6 e4c6 a6b5 c6b5 c7c5 b5d3 f5d6 f1e1 b7b6 d1d2 c8f5 a2a4 f5d3 d2d3 h7h6 c2c3 d4c3 d3c3 f8e8 b2b3 d6e4 c3c2 d8f6 h2h3 e4c3 e1e8 a8e8 a1e1 e8e1 f3e1 c3d5 c2d2 f6a1');
INSERT INTO evaluations VALUES(52382,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(52382,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(52382,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(52382,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(52382,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52382,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52382,36,'cp 0','e7e5 g1e2 g8f6 d2d4 d7d6 b1c3 d8b6 d4e5 d6e5 h2h3 b8d7 e2g3 d7c5 f1e2 c5e6 e2d3 b6c7 e1g1 h7h5 a2a4 h5h4 g3e2 a7a5 c1e3 f8d6 f1d1 e8g8 d3c4 c7e7 e2c1 e6d4 e3d4 e5d4 d1d4 d6c5 d4d2 f8e8 g1h1 c8d7');
INSERT INTO evaluations VALUES(51865,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(51865,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(51865,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(51865,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(51865,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(51865,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51865,36,'cp 0','g1e2 g8f6 d2d4 d7d5 e4d5 e5e4 f3g3 c6d5 b1c3 b8c6 c1e3 f8d6 g3g7 h8g8 g7h6 g8g6 h6h4 g6g4 h4h6 d6f8 h6h3 g4g6 h3h4 g6g4');
INSERT INTO evaluations VALUES(33629,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(33629,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(33629,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(33629,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(33629,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(33629,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(33629,36,'cp 72','e7e4 e1d1 b8c6 b1c3 e4e6 c7g3 f8e7 g1e2 e6f5 e2f4 g7g5 f4h3 d7d5 g3g5 d5d4 g5f5 c8f5 c3b5 f5h3 g2h3 f6e4 d1e2 e7h4 d2d3 e4f2 h1g1 e8c8 c1f4 h8e8 e2d2 c6e5 b5a7 c8c7 a1e1 f7f6 f4g3 h4g5 d2e2 f2d3 c2d3 c7b6 f1g2 b6a7 g2e4 e5c4 d3c4 e8e4 e2f3 e4e3 e1e3 g5e3 g1g2 d8e8 a2a3 a7b6');
INSERT INTO evaluations VALUES(33625,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(33625,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(33625,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(33625,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(33625,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(33625,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(33625,36,'cp 62','e1d1 b8c6 b1c3 e4g6 f2f3 f8c5 c7g3 d7d5 f1b5 d5d4 c3e2 f6d5 g3g6 h7g6 b2b3 d5c7 a2a4 c7b5 a4b5 c6d8 b3b4 c5b6 e2f4 c8d7 d2d3 d8e6 f4d5 b6d8 b5b6 a7a6 a1a5 a8c8 g1e2 d7c6 e2f4 e8g8 f4e6 f7e6 d5c7 d8c7 b6c7 c8c7');
INSERT INTO evaluations VALUES(33626,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(33626,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(33626,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(33626,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(33626,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(33626,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(33626,36,'cp 64','b8c6 b1c3 e4g6 f2f3 f8c5 c7g3 d7d5 f1b5 d5d4 c3e2 f6d5 g3g6 h7g6 b2b3 d5c7 a2a4 c7b5 a4b5 c6d8 c1a3 c5b6 e2f4 c8d7 a3b2 d8e6 g1e2 e6f4 e2f4 e8c8 f4d5 b6c5 b3b4 d7b5 b4c5 d8d5 a1a7 c8b8 a7a3 h8e8 h2h4 b5e2 d1c1 d4d3 b2g7 d5c5 g7c3 e8d8 c2d3 d8d3');
INSERT INTO evaluations VALUES(48054,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(48054,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(48054,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(48054,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(48054,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(48054,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(48054,36,'cp 73','b8d7 h2h3 f8e7 c1f4 e8g8 f1d3 c7c5 d4d5 e6d5 e4d5 d7b6 e1g1 h7h6 d3e2 c8f5 a2a4 f6e4 a4a5 e4c3 b2c3 b6d7 e2d3 f5d3 d1d3 f8e8 a5a6 b7a6 a1a6 d7b6 c3c4 e7f6 f1b1 d8e7 g2g4 a8c8 g1g2 f6d4 f4g3 e7b7 a6a3 b7e7 a3b3 e7e2 g2g1');
INSERT INTO evaluations VALUES(3816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3816,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3816,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3816,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3816,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3816,36,'cp 0','d2d3 c6a5 c4b3 a7a6 c3e2 f6h5 e2c3');
INSERT INTO evaluations VALUES(23470,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(23470,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(23470,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(23470,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(23470,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(23470,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(23470,36,'cp 47','d2d4 e7e5 d4d5 c6e7 g1e2 d7d6 e3e4 g7g6 e2g3 h7h5 h2h4 c8d7 f1d3 a7a6 c2c4 f8h6 b1c3 f7f5 e4f5 h6c1 d1c1 e7f5 c1g5 d8g5 h4g5 e8c8 g3e4 h5h4 b2b4 g8e7 e4f6 f5d4 e1c1 b7b5 f2f3 d7f5 c3e4 b5c4 d3c4');
INSERT INTO evaluations VALUES(48856,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(48856,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(48856,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(48856,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(48856,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(48856,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(48856,36,'cp -46','d2d3 c7c5 g1f3 b8c6 e1g1 c8e6 f3g5 e6g4 h2h3 g4h5 g5f3 f8e7 g3g4 h5g6 f3h4 d8d7 b1d2 e8g8 c2c4 d5d4 d2f3 a8d8 f3g5 f6e8 a1b1 e8c7 f2f4 e7g5 f4g5 d8e8 d1e1 e5e4 e1g3 c6b4 h4g6 f7g6 d3e4 f8f1 g2f1 e8e4');
INSERT INTO evaluations VALUES(48853,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(48853,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(48853,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(48853,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(48853,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(48853,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(48853,36,'cp -67','f8d6 g1f3 e8g8 e1g1 f8e8 b1d2 c7c6 e2e4 a7a5 a2a3 c8g4 h2h3 g4h5 d1e1 d8c7 c1b2 b8d7 f3h4 a8d8 g1h1 a5a4 h4f5 d6f8 b3a4 h5g6 f5h4 c7a5 h4g6 h7g6 e4d5 c6d5 d2f3 a5a4 b2e5 d7e5 f3e5 a4c2 a1c1 c2b2');
INSERT INTO evaluations VALUES(52152,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(52152,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(52152,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(52152,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52152,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(52152,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(52152,36,'cp 35','f1c4 b8c6 e1g1 f8e7 c2c3 g8f6 c3d4 c5d4 e4e5 f6g4 f1e1 d7d6 e5d6 d8d6 h2h3 g4f6 b1c3 e8g8 c3b5 d6c5 b2b3 a7a6 b5d4 c6d4 f3d4 b7b5 c4f1 f8d8 c1b2 e7f8 d1f3 c5d5 f3g3 f8d6 g3h4 c8b7 d4c2 d5f5 c2e3 f5f4 h4f4 d6f4');
INSERT INTO evaluations VALUES(52145,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(52145,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(52145,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(52145,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52145,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52145,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(52145,36,'cp 25','b8c6 e1g1 f8e7 c2c3 g8f6 c3d4 c5d4 f1e1 e8g8 e4e5 f6e8 f3d4 d7d5 c4b3 e7c5 c1e3 c5b6 b1c3 e8c7 c3a4 b6d4 e3d4 c7e6 d4c5 e6c5 a4c5 c8e6 a1c1 d8b6 d1d2 f8d8 f2f4 c6a5 b3d1 a8c8 d2f2 a5c4 c5e6 f7e6 f2b6 c4b6 d1f3');
INSERT INTO evaluations VALUES(19034,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(19034,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(19034,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(19034,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(19034,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(19034,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(19034,36,'cp 30','e1g1 f8e7 c2c3 g8f6 c3d4 c5d4 e4e5 f6g4 f1e1 d7d6 e5d6 d8d6 c1g5 g4f6 b1c3 e8g8 c3b5 d6d8 b5d4 c6d4 d1d4 d8d4 f3d4 e7c5 d4b3 c5b4 e1d1 b4e7 a1c1 f8d8 b3a5 d8d1 c1d1 c8e6 g2g3 e6c4 a5c4 a8d8 d1d8 e7d8 c4d6 b7b6 g1g2');
INSERT INTO evaluations VALUES(20102,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(20102,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20102,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(20102,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(20102,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(20102,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20102,36,'cp 23','c2c4 g7g6 f1d3 f8g7 g1e2 c8g4 e1g1 a8c8 h2h3 g4e2 d3e2 g8f6 b1c3 f6d7 c1d2 e8g8 f1e1 d7c5 e2f1 e7e6 a1b1 g7c3 d2c3 e6e5 b2b4 c5e6 c3d2 b7b6 d2h6 f8e8 h6e3 h7h5 d1d3 c6d4 g1h1 d8h4 b1d1 a7a5 a2a3 a5b4 a3b4 c8a8 d1d2');
INSERT INTO evaluations VALUES(20065,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20065,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20065,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20065,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20065,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20065,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20065,36,'cp 0','g7g6 f1d3 f8g7 g1e2 c8g4 e1g1 a8c8 f2f3 d8b6 g1h1 g7b2 c1b2 b6b2 b1c3 g4e6 a1b1 b2a3 b1b7 g8f6 f3f4 e8g8 b7b3 a3c5 b3b5 c5a3');
INSERT INTO evaluations VALUES(22037,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(22037,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(22037,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(22037,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(22037,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(22037,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(22037,36,'cp 83','f3d4 g8f6 f1b5 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 d8d7 h2h3 f8b8 e4e5 f6d5 c3d5 c6d5 f3d5 c7c6 d5d4 b8b4 d4d3 b4b2 f4f5 e7f8 a2a4 b2b7 a1d1 d6d5 e3d4 a8e8 d4c3 f8b4 g1h1 b4c3 d3c3 b7b4 d1e1 b4e4 e5e6 f7e6 e1e4 d5e4 f5f6 d7f7 c3a5');
INSERT INTO evaluations VALUES(39551,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(39551,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(39551,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(39551,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(39551,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39551,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(39551,36,'cp -12','e7e6 g1e2 g8e7 b1c3 c7c5 d2d3 b8c6 c4b3 d7d6 f3g3 e8g8 e1g1 c6a5 c3b5 e6e5 g3h4 a5b3 a2b3 f7f6 c2c4 a7a6 b5c3 c8e6 c1d2 d8d7 c3d5 e6d5 c4d5 f6f5 f2f4 g7f6 h4e1 e5f4 e2f4 f5e4 d3e4 f6b2 a1a2 b2e5 f4e6');
INSERT INTO evaluations VALUES(39004,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(39004,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(39004,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39004,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(39004,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(39004,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39004,36,'cp 0','c4b3 c7c5 d2d3 b8c6 b1c3 g8e7 h2h4 d7d5 g1e2 d5d4 c3b1 h7h5 f3g3 e8g8 e1g1 b7b6 b1d2 c6e5 d2c4 e5c4 b3c4 e7c6 c1f4 c8b7 a2a3 d8d7 f1b1 c6a5 c4a2 a5c6');
INSERT INTO evaluations VALUES(10258,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(10258,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(10258,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(10258,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(10258,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(10258,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(10258,36,'cp -33','d7d5 f1d3 d5e4 c3e4 c5b6 e1g1 c8f5 d1e2 f5g6 e4f6 d8f6 d3g6 h7g6 d2d3 f6f5 c1e3 e8c8 e3b6 a7b6 f1e1 f7f6 e2e4 c6d4 f3d4 d8d4 e4f5 g6f5 a3a4 g7g5 a4a5 b6a5 a1a5 c8d7 e1a1 d4b4 b2b3 b4b6 a5d5 b6d6 d5c5 d6c6 c5c6 d7c6 a1a7 h8d8');
INSERT INTO evaluations VALUES(9094,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(9094,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(9094,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(9094,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(9094,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(9094,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(9094,36,'cp -35','f1d3 d5e4 c3e4 c5b6 e1g1 c8f5 d1e2 f5g6 e4f6 d8f6 d3g6 h7g6 d2d3 f6f5 c1e3 e8c8 e3b6 a7b6 f1e1 f7f6 e2e4 c6d4 f3d4 d8d4 e4f5 g6f5 a3a4 d4d6 g1f1 d6c6 c2c3 g7g5 d3d4 e5d4 c3d4 f5f4 a4a5 b6a5 a1a5 c8d7');
INSERT INTO evaluations VALUES(12295,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(12295,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(12295,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(12295,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(12295,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(12295,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12295,36,'cp 0','b7c6 d1f3 f8c5 f1c4 e8g8 e1g1 c5d4 c3e2 d4b6 c1g5 d7d6 a1d1 c8g4 f3a3 b6c5 b2b4 c5b6 h2h3 g4h5 a3d3 h7h6 g5f6 d8f6 d3d6 f6g5 g1h1 a8d8 d6c6 d8d1 f1d1 b6f2 c4d3 f8b8');
INSERT INTO evaluations VALUES(12788,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(12788,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(12788,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(12788,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(12788,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(12788,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(12788,36,'cp -79','d2d3 b7b5 c4b3 d7d5 e4d5 c6d4 f3d1 a7a5 c2c3 d4b3 a2b3 f6d5 g1f3 g7g6 e1g1 f8g7 d3d4 b5b4 d4c5 e8g8 c5c6 b4c3 b1c3 d5c3 b2c3 c8a6 f1e1 d8d1 e1d1 g7c3 a1a2 f8c8 c1d2 c3d2 f3d2 a6b5 f2f3 c8c6 d1a1 c6a6 d2e4 e7e6 g1f2 b5d3 e4d2 e6e5');
INSERT INTO evaluations VALUES(12783,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(12783,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(12783,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(12783,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(12783,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(12783,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(12783,36,'cp -78','c6d4 f3d1 d7d5 e4d5 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 g1e2 b5b4 b1d2 e7e6 e1g1 f8e7 d2c4 e8g8 c1d2 d8c7 e2g3 f8d8 f1e1 e7f8 d1g4 f7f6 g4f3 a8a6 h2h4 g8h8 c3b4 c5b4 e1c1 e6e5 d3d4 e5d4 g3e2 c8b7');
INSERT INTO evaluations VALUES(30757,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(30757,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(30757,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(30757,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(30757,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(30757,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(30757,36,'cp 161','d2d4 h7h6 c1f4 f8g7 f1c4 a7a6 e1g1 b6b5 c4d5 b8c6 f3e5 e7e6 d5c6 b7c6 e5c6 d7c6 d1f3 g8e7 a1d1 e8g8 e4e5 e7d5 c3e4 d8h4 f4c1 h4h5 f3h5 g6h5 f2f4 d5e7 a2a3 g8h7 e4c5 a6a5 g1f2 f8d8 c5b7 d8d5 c2c3');
INSERT INTO evaluations VALUES(54705,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(54705,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(54705,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(54705,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(54705,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(54705,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(54705,36,'cp 78','c7c5 e2e3 c5d4 e3d4 e7e6 c2c4 b8c6 b1c3 f8d6 f4d6 d8d6 f1d3 g8e7 e1g1 e8g8 f1e1 g8h8 a2a3 c8d7 b2b4 a7a6 c4c5 d6f4 c3a4 e6e5 a4b6 d7g4 d1c1 f4c1 a1c1 g4f3 b6a8 f8a8 g2f3 c6d4');
INSERT INTO evaluations VALUES(52970,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(52970,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(52970,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(52970,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(52970,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(52970,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(52970,36,'cp 80','e7e5 f2f4 e5d4 d1d4 g8f6 c1d2 b8d7 e1c1 d8b6 g1f3 f8e7 f1c4 e8g8 h2h3 b6d4 f3d4 d6d5 e4d5 d7b6 c4d3 f8d8 g2g4 f6d5 c3d5 b6d5 c2c4 d5b6 d2e3 e7c5 b2b3 c5d4 e3d4 d8d4 d3h7 g8h7 d1d4 g7g6 h1d1 c8e6 d4d8');
INSERT INTO evaluations VALUES(38473,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(38473,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(38473,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(38473,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(38473,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(38473,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(38473,36,'cp 82','e4d5 d8d5 f1e2 d5d8 e1g1 e7e6 c2c3 g8e7 a2a4 e8g8 c1f4 b7b6 a4a5 a7a6 h2h3 b8d7 f4h2 b6b5 b2b4 c8b7 b1d2 e7f5 f1e1 f5d6 d2b3 b7d5 f3d2 f8e8 e2f3 a8c8 d1c2 d8g5 f3d5 e6d5 a1d1 c8d8 h2g3 g5h6 b3c5');
INSERT INTO evaluations VALUES(38446,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(38446,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(38446,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(38446,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(38446,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(38446,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(38446,36,'cp 95','d8d5 f1e2 d5d8 e1g1 e7e6 a2a4 g8e7 c2c3 e8g8 a4a5 a7a6 b1d2 b7b5 a5b6 c7b6 d2e4 c8b7 e2d3 b8d7 h2h4 d7f6 e4f6 g7f6 d1e2 b6b5 h4h5 d8d5 h5h6 f8d8 f1e1 d5h5 d3e4 b7e4 e2e4 h5f5 e4b7 f5d5 b7c7 d5d6 c7d6 d8d6 c1f4 d6d7');
INSERT INTO evaluations VALUES(33065,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(33065,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(33065,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(33065,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(33065,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(33065,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(33065,36,'cp 88','f1e2 d5d8 e1g1 e7e6 a2a4 g8e7 c2c3 e8g8 c1f4 b7b6 a4a5 a7a6 b1d2 c8b7 h2h3 b8d7 f4h2 e7f5 f1e1 b6b5 b2b4 b7d5 e2d3 f5d6 d1e2 d8e7 d3c2 a8d8 h2f4 f8e8 f4g3 c7c6 g3h2 d8c8 h2f4 c8d8 e2e3 f7f6 f4h2 e6e5 d4e5 f6e5');
INSERT INTO evaluations VALUES(39573,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(39573,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(39573,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(39573,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(39573,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(39573,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(39573,36,'cp 9','g8f6 b1c3 d7d5 c4d5 f6d5 f1c4 d5b6 c4b3 c7c5 g1f3 e8g8 e1g1 c5d4 e3d4 c8g4 d4d5 b8d7 h2h3 g4f3 d1f3 d7c5 f1d1 c5b3 a2b3 b6c8 h3h4 d8d7 h4h5 h7h6 c1e3 d7f5 f3f5 g6f5 c3b5 a7a6 e3f4 f8d8 b5c7 a8a7 a1a2 c8d6 f2f3 a6a5');
INSERT INTO evaluations VALUES(23445,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(23445,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(23445,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(23445,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(23445,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(23445,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(23445,36,'cp -14','e7e5 b1c3 f8b4 g1e2 g8e7 a2a3 b4c3 e2c3 d7d5 e4d5 e7d5 f1e2 e8g8 e1g1 c6d4 f1e1 f8e8 e2f3 d5c3 b2c3 d4f3 d1f3 d8d6 d3d4 c8d7 d4e5 d6g6 c1f4 b7b6 h2h4 h7h6 f3g3 g6g3 f4g3 d7f5 f2f3 f5c2 e5e6 e8e6 e1e6 f7e6 g3c7');
INSERT INTO evaluations VALUES(21521,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(21521,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21521,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(21521,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(21521,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(21521,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21521,36,'cp 7','f8d6 c2c4 d5c4 e1g1 g8e7 b5c4 e8g8 h2h3 e7f5 c1e3 d8f6 b1c3 f5e3 f2e3 f6g6 g1h1 g6h6 d1e2 c8g4 e2f2 a8e8 a1e1 g4e6 c4e6 f7e6 c3e4 h6g6 e4d6 c7d6 e3e4 h7h6 f2d2 e6e5 d4d5 c6d4 f3d4 f8f1 e1f1 e5d4 d2d4 g6e4 d4a7 e4d5');
INSERT INTO evaluations VALUES(5893,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5893,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(5893,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5893,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5893,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(5893,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(5893,36,'cp 11','c2c4 d5c4 e1g1 g8e7 b5c4 e8g8 h2h3 e7f5 c1e3 f5e3 f2e3 c6a5 c4d3 c7c5 b1c3 c5d4 e3d4 a5c6 g1h1 c6e7 f3g5 e7f5 d3f5 d8g5 c3e4 g5h6 e4d6 h6d6 f5e4 c8e6 d4d5 e6d7 d1d4 b7b6 a2a3 a8c8 a1d1 c8e8 d4d3 g7g6 e4f3 e8e7 d1e1 f8e8 e1e7 d6e7 d3d2 e7e5 d5d6 g8g7');
INSERT INTO evaluations VALUES(18908,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(18908,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(18908,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(18908,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(18908,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(18908,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(18908,36,'cp -9','g8f6 b1d2 d7d5 e4d5 c6d4 e3d4 d8d5 d2f3 c8d7 f1e2 f8c5 a2a3 c5d4 d1d4 d7c6 e1c1 a8c8 h1g1 d5d4 f3d4 e8e7 d4c6 c8c6 d1d4 b7b6 c2c3 c6c5 g1d1 g7g6 h2h4 h7h6 a3a4 h8c8 f2f3 c8c7 e2d3 h6h5 c1d2 g6g5 h4g5');
INSERT INTO evaluations VALUES(12483,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(12483,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(12483,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(12483,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(12483,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(12483,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(12483,36,'cp 0','b1d2 d7d5 e4d5 c6d4 e3d4 d8d5 d2f3 c8d7 f1e2 f8c5 a2a3 c5d4 d1d4 d7c6 e1c1 d5d4 d1d4 e8e7 g2g3 b7b6 h1e1 c6f3 e2f3 a8c8 f3e2 g7g6 f2f4 c8c7 c2c3 h7h5 h2h4 h8c8 e1d1 c7c5 c1d2 c8c7 d2c1');
INSERT INTO evaluations VALUES(56737,30,'cp 279',NULL);
INSERT INTO evaluations VALUES(56737,31,'cp 294',NULL);
INSERT INTO evaluations VALUES(56737,32,'cp 291',NULL);
INSERT INTO evaluations VALUES(56737,33,'cp 304',NULL);
INSERT INTO evaluations VALUES(56737,34,'cp 304',NULL);
INSERT INTO evaluations VALUES(56737,35,'cp 288',NULL);
INSERT INTO evaluations VALUES(56737,36,'cp 301','d2d4 b8c6 f3g5 g8h6 d4d5 c6e7 b1c3 g7g6 h2h4 f8g7 h4h5 e7g8 h5g6 h7g6 c1d2 f5f4 g5e6 c8e6 d5e6 c7c6 d1e2 d8b6 e1c1 g8f6 f2f3 e8c8 c1b1 c8b8 a2a4 b6c7 d2e1 h6g8 h1h8 g7h8 a4a5 h8g7 a5a6 b7b5 c4a2');
INSERT INTO evaluations VALUES(56738,30,'cp 297',NULL);
INSERT INTO evaluations VALUES(56738,31,'cp 300',NULL);
INSERT INTO evaluations VALUES(56738,32,'cp 267',NULL);
INSERT INTO evaluations VALUES(56738,33,'cp 278',NULL);
INSERT INTO evaluations VALUES(56738,34,'cp 292',NULL);
INSERT INTO evaluations VALUES(56738,35,'cp 272',NULL);
INSERT INTO evaluations VALUES(56738,36,'cp 300','b8c6 f3g5 g8h6 d4d5 c6e7 b1c3 g7g6 h2h4 f8g7 h4h5 e7g8 h5g6 h7g6 c1d2 d8e7 d1e2 f5f4 g5e6 c8e6 d5e6 g8f6 e1c1 c7c6 g2g3 g6g5 a2a4 e8c8 a4a5 c8b8 a5a6 h6g4 f2f3 h8h1 d1h1 g4e3 d2e3 f4e3 e2e3 d6d5 e4d5 f6d5 c3d5');
INSERT INTO evaluations VALUES(12817,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(12817,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(12817,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(12817,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(12817,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(12817,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(12817,36,'cp 18','c5d4 f3d4 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 f1e2 e8g8 e1g1 f7f5 f2f4 a7a6 b5a3 b8d7 a3c4 e5f4 c1f4 d7b6 c4a5 e7f6 c2c3 g7g5 f4e3 f5f4 e3d4 b6d5 e2c4 d8a5 d4f6 a5c5 d1d4 f8f6 c4d5 g8f8 d4c5 d6c5 a4a5 c8f5 a1d1 a8e8 d5b7');
INSERT INTO evaluations VALUES(12887,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(12887,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(12887,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(12887,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(12887,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(12887,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(12887,36,'cp 40','f3d4 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 c2c4 a7a6 b5c3 f8e7 f1d3 f7f5 f2f3 e8g8 e1g1 b8d7 g1h1 a6a5 c1e3 b7b6 a2a3 d7c5 d3c2 e7g5 e3g1 c8a6 d1e2 a8c8 b2b4 c5d7 c2d3 g7g6 a1b1 g5f4 b4b5 a6b7 c3a4 d8h4 e2f2 h4f2 g1f2 d7c5 a4c5 d6c5 f1d1');
INSERT INTO evaluations VALUES(53139,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(53139,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(53139,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(53139,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(53139,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(53139,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53139,36,'cp -13','g1f3 b8c6 f1c4 a7a6 e1g1 e7e6 c1f4 f8e7 d1e2 e6e5 f4e3 g8f6 f1d1 c8d7 c3d5 f6d5 c4d5 e8g8 a2a3 a8c8 a1c1 d8e8 h2h3 d7e6 e2d3 h7h6 f3d2 b7b5 d2f1 c6d4 d5e6 f7e6 e3d4 e5d4 d3d4 e7f6 d4d2 c8c1 d1c1');
INSERT INTO evaluations VALUES(34822,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(34822,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(34822,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(34822,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34822,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(34822,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(34822,36,'cp 81','f8b4 e2e3 g8f6 c1d2 b4c3 d2c3 e8g8 g1e2 b8c6 d1c2 f8d8 f2f3 d5d7 e2f4 d7e7 a2a3 e6e5 d4e5 c6e5 e3e4 g7g5 f4d3 e5d3 f1d3 f6d5 c3d2 d5f4 d3f1 f4e6 d2c3 e6d4 c2f2 d4b3 a1b1 g5g4 f2g3 e7g5 f1e2 c8e6 g3f2 c7c5 e1g1');
INSERT INTO evaluations VALUES(33109,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(33109,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(33109,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(33109,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(33109,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(33109,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(33109,36,'cp 69','e2e3 g8f6 c1d2 b4c3 d2c3 e8g8 g1e2 b8c6 d1c2 f8d8 f2f3 d5d7 e2f4 a7a5 a2a3 b7b6 a1d1 d7d6 f1d3 c8a6 d3a6 a8a6 e1g1 a6a8 f4e2 f6d5 c3d2 d5e7 g1h1 e6e5 d4e5 d6e5 e3e4 e5c5 c2c5 b6c5');
INSERT INTO evaluations VALUES(38433,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(38433,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(38433,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(38433,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(38433,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(38433,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(38433,36,'cp 72','d2d4 g8f6 c1e3 e8g8 d1d2 c7c6 e3h6 b8d7 h2h4 e7e5 h4h5 f6h5 e1c1 d8e7 g2g4 h5f4 h6g5 e7e8 g5f4 e5f4 g4g5 e8e7 d2f4 f8e8 f1d3 d7f8 d1e1 c8e6 e4e5 d6d5 h1h4 b7b5 e1h1 a7a5 c3e2 a5a4 f4h2 a4a3 b2b3 c6c5 e2f4 c5c4');
INSERT INTO evaluations VALUES(12591,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(12591,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12591,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(12591,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12591,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12591,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12591,36,'cp 35','c1g5 f8b4 c2c3 b4e7 g5f6 e7f6 b1a3 d7d6 f1c4 a7a6 e1g1 e8g8 a3c2 c6e7 a2a4 f6g5 c2b4 g8h8 b3d2 f7f5 f1e1 g5d2 d1d2 f5e4 e1e4 e7g6 a1d1 c8f5 e4e3 g6f4 d2d6 d8g5 e3g3 f5g4 d1e1 a8d8');
INSERT INTO evaluations VALUES(12573,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(12573,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12573,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(12573,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(12573,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12573,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(12573,36,'cp 21','f8b4 c2c3 b4e7 g5f6 e7f6 b1a3 d7d6 f1c4 a8b8 d1e2 a7a6 b3d2 e8g8 e1g1 c6e7 c4b3 f6g5 f1d1 b7b5 c3c4 e7c6 d2f3 c8g4 c4b5 c6d4 d1d4 e5d4 b5a6 d8a5 a3c2 g5f6 a1d1 b8b6 c2d4 a5a6 b3c4 a6a5 b2b3 f6d4 d1d4 g4f3 g2f3 a5g5 g1h1');
INSERT INTO evaluations VALUES(10050,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10050,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10050,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10050,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10050,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10050,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10050,36,'cp -6','d7d5 c4e2 f6e4 c3d4 c5b4 c1d2 e4d2 b1d2 e8g8 e1g1 f7f6 d2b3 f8e8 e2b5 c8g4 b5c6 b7c6 a1c1 f6e5 d4e5 e8e5 b3d4 e5e4 d4c6 d8d6 c6b4 e4b4 c1c7 a8f8 c7c3 b4b2 h2h3 g4h5 c3d3 b2a2 d3d5 h5f3 d1f3 f8f3 d5d6 f3f8 d6d7 h7h6 d7d1 f8b8 d1c1 b8e8');
INSERT INTO evaluations VALUES(9078,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(9078,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9078,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(9078,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(9078,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9078,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9078,36,'cp 5','c4b5 f6e4 c3d4 c5b6 h2h3 f7f6 b1c3 e8g8 e1g1 f6e5 b5c6 b7c6 c3e4 d5e4 d1b3 d8d5 f3e5 c8a6 f1d1 a6e2 d1e1 b6d4 e1e2 d4e5 e2e4 a8e8 e4e2 e5d4 c1e3 d5b3 a2b3 d4b6 e2c2 b6e3 f2e3 e8e3 a1a3 f8c8 c2c6 e3e2 a3a7 e2b2 a7b7 h7h6 b7c7 c8c7 c6c7');
INSERT INTO evaluations VALUES(9065,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9065,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(9065,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9065,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9065,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9065,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9065,36,'cp 7','f6e4 c3d4 c5b6 b1c3 e8g8 c1e3 f7f5 e5f6 e4f6 b5c6 b7c6 f3e5 d8d6 e1g1 f6d7 e5f3 c8b7 c3a4 d6g6 f1e1 a8e8 a4b6 c7b6 b2b4 b7a6 a2a4 a6c4 b4b5 h7h6 b5c6 g6c6 d1c2 c6e6 a1c1 e6e4 c2e4 e8e4 a4a5');
INSERT INTO evaluations VALUES(9137,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9137,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9137,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9137,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9137,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9137,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(9137,36,'cp 14','c3d4 c5b6 c1e3 e8g8 b1c3 f7f5 e5f6 e4f6 b5c6 b7c6 f3e5 d8d6 e1g1 f6d7 e5f3 a8b8 a2a3 d6g6 c3a4 c8a6 f1e1 b8e8 a4b6 c7b6 a1c1 e8e4 h2h3 h7h6 e3d2 g8h7 e1e4 d5e4 f3e1 g6f6 d2e3 f6g6 b2b4 h7h8 f2f4 e4f3');
INSERT INTO evaluations VALUES(9136,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(9136,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(9136,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(9136,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9136,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(9136,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9136,36,'cp 7','c5b6 b1c3 e8g8 c1e3 c8g4 h2h3 g4h5 d1c2 h5g6 c2b3 c6e7 e1g1 c7c6 b5d3 e4c3 b3c3 f7f6 e5f6 f8f6 e3g5 f6f3 g2f3 d8d7 g5e7 d7e7 d3g6 e7g5 g1h1 g5g6 f1e1 a8f8 f3f4 g6f6 a1d1 f6f4 c3e3 h7h6 e3f4 f8f4 e1e8 g8h7');
INSERT INTO evaluations VALUES(8904,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(8904,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(8904,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(8904,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(8904,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(8904,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(8904,36,'cp 6','h2h3 e8g8 e1g1 f7f6 b1c3 f6e5 b5c6 b7c6 c3e4 d5e4 d1b3 d8d5 f3e5 c8a6 f1d1 a6e2 d1e1 b6d4 e1e2 d4e5 e2e4 a8e8 e4e2 e5d4 c1e3 d5b3 a2b3 f8f7 a1d1 c6c5 d1d3 f7e7 g2g3 g8f7 g1g2 d4e3 f2e3 f7g6 d3c3 e7e5 e2e1 e8d8 g2f2 d8d2 e1e2');
INSERT INTO evaluations VALUES(6971,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(6971,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(6971,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6971,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(6971,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(6971,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(6971,36,'cp 21','g8f6 e1g1 d7d6 a2a4 e8g8 a4a5 a7a6 f1e1 c8e6 b1d2 h7h6 c2c3 d8d7 d1b3 a8b8 h2h3 e6c4 d2c4 d7e6 b3c2 f8e8 c1d2 e7f8 c4e3 d6d5 c3c4 d5e4 d3e4 c6d4 f3d4 e5d4 e3d5 f6d5 c4d5 e6g6 d2f4 f8d6 f4d6 c7d6');
INSERT INTO evaluations VALUES(52458,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(52458,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(52458,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(52458,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(52458,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(52458,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(52458,36,'cp -30','g8f6 g2g3 g7g6 d2d3 d7d5 f1g2 f8g7 e1g1 e8g8 f3e5 f6g4 e5g4 c8g4 b1c3 d5d4 c3e4 d8c7 d1e1 g4f5 h2h3 b8d7 e1f2 a8e8 f2f3 d7f6 e4f6 e7f6 g3g4 f5c8 f4f5 g6g5 c2c3 h7h6 c3d4 c5d4 c1d2 c8d7 f3b7 c7b7 g2b7');
INSERT INTO evaluations VALUES(44578,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(44578,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(44578,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(44578,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(44578,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(44578,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(44578,36,'cp -24','g2g3 g7g6 d2d3 d7d5 f1g2 f8g7 e1g1 e8g8 f3e5 f6g4 e5g4 c8g4 b1c3 d5d4 c3e4 d8b6 a2a3 g4d7 d1e1 f7f5 e4d2 d7c6 d2c4 b6c7 e1f2 b7b5 c4e5 g7e5 f4e5 c7e5 c1h6 f8f6 a1e1 c6g2 f2g2 b8c6 e2e4 d4e3 h6e3 e5b2 e3c5');
INSERT INTO evaluations VALUES(55630,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55630,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(55630,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(55630,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55630,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55630,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(55630,36,'cp 10','g8f6 g1e2 c7c6 c2c4 c8f5 e2g3 f5g6 d2d4 b8d7 f1e2 h7h5 h2h4 e5d4 d1d4 d6d5 c4d5 f8c5 d4d2 f6d5 b1c3 d5c3 b2c3 d8e7 b3b4 c5d6 a1d1 d6g3 f2g3 d7f6 c3f6 g7f6 e1g1 e8g8 g3g4 a8d8 d2c3 h5g4 d1d8 f8d8 f1f6 g4g3 f6f4 b7b5 f4g4 d8b8');
INSERT INTO evaluations VALUES(47908,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(47908,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(47908,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(47908,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47908,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(47908,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47908,36,'cp 19','g1e2 c7c6 c2c4 c8f5 e2g3 f5g6 d2d4 h7h5 h2h4 b8d7 f1e2 e5d4 d1d4 d6d5 c4d5 f8c5 d4d2 f6d5 b1c3 d5c3 b2c3 d8e7 b3b4 c5d6 a1d1 d6g3 f2g3 d7f6 c3f6 g7f6 e1g1 e8g8 d2c3 f8e8 f1f6 e7e3 c3e3 e8e3 e2c4 g8g7 f6f4 e3e7 g3g4 a7a5 g4h5 g6h5 g2g4 b7b5 c4b3 a5a4 g4h5 a4b3 a2b3 a8a2 d1d6 a2b2 f4g4 g7h7');
INSERT INTO evaluations VALUES(15431,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(15431,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(15431,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(15431,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(15431,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(15431,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(15431,36,'cp 33','f8e7 c2c3 e8g8 f1e1 a7a6 b5a4 f8e8 h2h3 e7f8 c1g5 b7b5 a4c2 h7h6 g5h4 c6b8 b1d2 b8d7 d3d4 c7c5 d4d5 a8b8 c3c4 f8e7 h4g3 g7g6 a2a3 f6h5 g3h2 h5f4 h2f4 e5f4 e4e5 d7e5 f3e5 d6e5 e1e5');
INSERT INTO evaluations VALUES(8560,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8560,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8560,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(8560,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(8560,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(8560,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(8560,36,'cp 40','c2c3 e8g8 f1e1 a7a6 b5a4 f8e8 h2h3 b7b5 a4c2 e7f8 c1g5 h7h6 g5h4 c8d7 b1d2 g7g5 h4g3 c6e7 d3d4 e7g6 d4e5 d6e5 d2f1 c7c5 f1e3 c5c4 b2b3 d7c6 b3c4 c6e4 d1d8 e8d8 c4b5 a6b5 f3e5 e4c2 e3c2 g6e5 g3e5');
INSERT INTO evaluations VALUES(8556,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(8556,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(8556,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(8556,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(8556,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(8556,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(8556,36,'cp 28','e8g8 b5a4 f8e8 f1e1 a7a6 b1d2 e7f8 d3d4 b7b5 a4c2 c8b7 h2h3 h7h6 d4d5 c6b8 b2b3 c7c6 c3c4 b8d7 a2a4 b5c4 b3c4 a6a5 d2b3 d7b6 c2d3 b7a6 b3a5 c6d5 c4d5 a6d3 a5c6 d8c7 d1d3 b6a4 d3c4 a4c5');
INSERT INTO evaluations VALUES(3356,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(3356,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(3356,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(3356,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(3356,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(3356,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(3356,36,'cp 36','b1d2 f8e8 f1e1 a7a6 b5a4 e7f8 d3d4 b7b5 a4c2 c8b7 h2h3 h7h6 a2a3 c6b8 b2b3 b8d7 c1b2 a8c8 a3a4 c7c6 d2f1 d8c7 f1g3 g7g6 c2d3 f8g7 d1d2 c8d8 d3f1 d6d5 e4d5 e5d4 d5c6 d4c3 d2c3 c7c6 a4b5 a6b5');
INSERT INTO evaluations VALUES(50085,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(50085,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(50085,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50085,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(50085,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(50085,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(50085,36,'cp 41','e7e6 g2g3 f6e4 f1g2 d7d5 d1c2 f8e7 b1c3 c7c6 c1f4 b8d7 a1d1 e8g8 e1g1 a7a5 f3e5 d7e5 f4e5 e7f6 e5f6 d8f6 c3e4 f5e4 f2f3 e4f3 f1f3 f6e7 f3f8 e7f8 e2e4 d5c4 c2c4 f8b4 b2b3 c8d7 c4b4 a5b4');
INSERT INTO evaluations VALUES(28714,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(28714,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(28714,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(28714,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28714,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(28714,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(28714,36,'cp 18','a2a3 d7d5 c4d5 f6d5 e2e3 f8e7 f1d3 e8g8 e1g1 c7c5 e3e4 d5c3 b2c3 b8c6 c1e3 c6a5 d1e2 a8c8 f1e1 d8c7 h2h4 f8d8 a1c1 a5c6 a3a4 c5d4 c3d4 e7a3 c1a1 a3b4 e1d1 h7h6 d3b5 c7e7 e3f4 c6a5 d1c1');
INSERT INTO evaluations VALUES(28723,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(28723,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(28723,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(28723,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(28723,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(28723,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(28723,36,'cp 12','d7d5 c1f4 d5c4 e2e3 f8d6 f4d6 d8d6 f1c4 e8g8 e1g1 b8d7 d1e2 c7c5 c4a6 b7a6 e2a6 e6e5 d4e5 d7e5 f3e5 d6e5 h2h3 f6e4 c3e4 e5e4 a1d1 e4e7 d1d2 f8d8 f1d1 d8d2 d1d2 g7g6 a3a4 a8f8 b2b3 f8b8 g2g3 h7h6 a6b5 b8d8');
INSERT INTO evaluations VALUES(28658,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(28658,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(28658,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(28658,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(28658,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28658,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28658,36,'cp 20','c4d5 f6d5 e2e3 f8e7 f1d3 c7c5 e3e4 d5c3 b2c3 e8g8 e1g1 d8c7 f1e1 b8d7 d1e2 f8d8 a3a4 h7h6 g2g3 c5d4 c3d4 d8c8 c1d2 a7a5 a1b1 c7d8 g1g2 c8c7 h2h4 a8c8 d3a6 b7a6 e2a6 c7c6 h4h5 d7f6 a6b7');
INSERT INTO evaluations VALUES(55865,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55865,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(55865,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(55865,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(55865,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(55865,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(55865,36,'cp 9','g1f3 c7c5 c2c4 d5c4 f1g2 b8c6 d1a4 c8d7 d4c5 c6a5 a4a3 a8c8 c1e3 d8c7 a3c3 g8f6 b2b4 c4b3 a2b3 a5c6 f3d4 c6d4 e3d4 f8c5 c3c5 c7c5 d4c5 c8c5 a1a7 d7c8 b1a3 c5c3 e1g1 c3b3 a3c4 e8g8 f1d1 b3c3 c4b6 c3c7');
INSERT INTO evaluations VALUES(44245,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(44245,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(44245,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(44245,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(44245,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(44245,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(44245,36,'cp 62','e4e5 f6d5 c3d5 e6d5 d2d4 d7d6 f1b5 b8c6 d1e2 f8e7 e5d6 d8d6 d4c5 d6c5 c1e3 c5d6 e1g1 e8g8 a1d1 c8f5 b5d3 f5d3 d1d3 a8d8 c2c3 f8e8 h2h3 h7h6 f1d1 d6e6 d1d2 e7f6 e2d1 c6a5 b2b3 a5c6 d3d5 d8d5 d2d5 f6c3');
INSERT INTO evaluations VALUES(44162,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(44162,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(44162,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(44162,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(44162,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(44162,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(44162,36,'cp 57','f6d5 c3d5 e6d5 d2d4 d7d6 f1b5 b8c6 d1e2 f8e7 e5d6 d8d6 d4c5 d6c5 c1e3 c5d6 e1g1 e8g8 a1d1 c8f5 b5d3 f5d3 d1d3 a8d8 f1d1 d6e6 c2c3 h7h6 d1e1 f8e8 h2h3 e6f5 e1d1 f5e6 e2c2 e7f6 e3d4 c6d4 f3d4 e6b6 g2g3 a7a6 c2d2');
INSERT INTO evaluations VALUES(27947,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(27947,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(27947,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(27947,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(27947,34,'cp -108',NULL);
INSERT INTO evaluations VALUES(27947,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(27947,36,'cp -102','b4c3 d2c3 b7e4 d4d5 g8f6 d5e6 f7e6 f1d3 e4b7 d1e2 d8e7 e1c1 b8c6 f3e5 e8c8 e5c6 b7c6 f2f4 f6d5 c3d2 d5b4 d2b4 e7b4 h1f1 g7g6 g2g3 b4c5 h2h4 h8e8 f1e1 a7a5 d3e4 e8f8 a2a3 c8b7 e2d3 f8f6 e1e3 c6e4 e3e4 d7d5 e4e5');
INSERT INTO evaluations VALUES(27997,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(27997,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(27997,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(27997,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(27997,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(27997,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(27997,36,'cp -99','d2c3 b7e4 d4d5 g8f6 d5e6 f7e6 f1d3 e4b7 d1e2 b8c6 e1c1 d8e7 f3e5 e8c8 e5c6 b7c6 f2f4 f6d5 c3d2 d5b4 d2b4 e7b4 h1f1 b4c5 g2g3 h7h6 h2h4 h8f8 a2a3 d8e8 d3e4 e8e7 f1f3 c6e4 e2e4 c8b8 f3c3 c5h5 e4c4 d7d6 d1e1 h5f5 c4c6');
INSERT INTO evaluations VALUES(27995,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(27995,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(27995,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(27995,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(27995,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(27995,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(27995,36,'cp -95','b7e4 d4d5 g8f6 d5e6 f7e6 f1d3 e4b7 d1e2 d8e7 e1c1 b8c6 f3e5 e8c8 h1e1 f6d5 e5c6 b7c6 c3d2 h8f8 g2g3 d5b4 d3h7 c6f3 d2b4 e7b4 e2a6 c8b8 d1d2 f3b7 a6e2 f8h8 e2d3 g7g5 h7e4 d7d5 e4f3 b4e7 f3g4 e6e5 f2f3 a7a5 a2a3 e7f6 c1b1');
INSERT INTO evaluations VALUES(27891,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(27891,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(27891,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(27891,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(27891,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(27891,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(27891,36,'cp -78','d4d5 g8f6 d5e6 f7e6 f1d3 e4b7 e1g1 e8g8 d1e2 d8e7 a1e1 b8c6 a2a3 e7c5 f3d2 c5h5 d2e4 h5e2 e4f6 g7f6 e1e2 c6e7 e2e3 e7d5 e3g3 g8f7 d3h7 d5c3 h7g6 f7e7 g3c3 f8h8 f1e1 c7c5 c3g3 a8g8 f2f4 b7c6 g3g4 h8h6 f4f5 e6e5 h2h4 d7d5 h4h5 d5d4 c2c3 g8d8');
INSERT INTO evaluations VALUES(27886,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(27886,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(27886,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(27886,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(27886,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(27886,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(27886,36,'cp -107','g8f6 d5e6 f7e6 f1d3 e4b7 d1e2 b8c6 e1c1 d8e7 f3e5 e8c8 e5c6 b7c6 f2f4 f6d5 c3d2 d5b4 d2b4 e7b4 h1f1 h8f8 g2g3 g7g6 e2e5 b4e7 h2h4 e7f6 e5e2 c8b8 a2a3 d8e8 d1e1 a7a5 e2e5 b8c8 e5f6 f8f6 e1e5 d7d6 e5g5 c8d7');
INSERT INTO evaluations VALUES(37743,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(37743,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(37743,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(37743,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(37743,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(37743,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(37743,36,'cp 91','f1d3 d7d6 e1g1 b8d7 c1e3 g8f6 b1d2 e7e5 c2c3 e8g8 f1e1 a7a5 h2h3 b7b6 d1e2 f8e8 d4e5 d6e5 a2a4 g7f8 e1d1 d7c5 d3c4 d8c7 f3g5 c5e6 e2f3 e6g5 e3g5 f6h5 g5e3 a8b8 a1c1 e8d8 b2b3 h5f4');
INSERT INTO evaluations VALUES(37740,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(37740,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(37740,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(37740,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(37740,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(37740,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(37740,36,'cp 83','d7d6 e1g1 g8f6 c1e3 b8d7 b1d2 a7a5 a2a4 d8c7 c2c3 e8g8 f1e1 b7b6 h2h3 e7e5 d4e5 d6e5 d1e2 c8b7 a1d1 f8d8 d3c4 h7h6 f3h2 d7f8 h2g4 f6g4 e2g4 b7c8 g4h4 c8e6 c4e6 f8e6 d2c4 b6b5 d1d8 c7d8 h4d8 e6d8 a4b5 c6b5 c4b6');
INSERT INTO evaluations VALUES(36528,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(36528,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(36528,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(36528,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(36528,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36528,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36528,36,'cp 14','e2e4 c8g4 f1e2 b8c6 d4d5 e7e5 f4g5 g4f3 e2f3 c6d4 e1g1 c7c5 d5c6 b7c6 b2b4 a8b8 a1b1 d8c7 d1d2 b8b7 f1c1 f8b8 a2a3 a7a6 f3e2 h7h6 g5e3 a6a5 e2f3 a5b4 a3b4 f6d7 b4b5 d7c5 d2d1 h6h5 f3e2 g8h7 h2h3 c5e6 e2f1 c7d8 b5c6 d4c6 c3d5');
INSERT INTO evaluations VALUES(36586,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36586,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36586,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(36586,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(36586,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36586,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36586,36,'cp 16','c8g4 f1e2 b8c6 d4d5 e7e5 f4g5 g4f3 e2f3 c6d4 e1g1 c7c6 d5c6 b7c6 b2b4 d8c7 f3e2 f8b8 a1b1 b8b7 e2d3 a8b8 a2a3 d4e6 g5d2 e6f4 d1c2 f4e6 g1h1 e6d4 c2d1 f6d7 d2g5 d7f6 c3e2 d4e6 g5d2 a7a5 b4b5 e6c5 d1c2 d6d5 e4d5');
INSERT INTO evaluations VALUES(27860,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(27860,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(27860,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(27860,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(27860,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(27860,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(27860,36,'cp 12','f1e2 b8c6 d4d5 e7e5 f4g5 g4f3 e2f3 c6d4 e1g1 c7c5 d5c6 b7c6 b2b4 d8c7 f3e2 f8b8 a1b1 b8b7 e2d3 a8b8 a2a3 d4e6 g5d2 e6f4 d1c2 f4e6 g1h1 e6d4 c2d1 f6d7 d2g5 d7f8 c3e2 a7a5 e2d4 e5d4 d1d2 a5b4 a3b4 f8d7 f2f4 b8a8 b4b5 a8a3 f4f5');
INSERT INTO evaluations VALUES(39590,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(39590,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(39590,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(39590,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(39590,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(39590,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(39590,36,'cp 67','e2e4 d7d6 c1e3 a7a6 d1d2 b7b5 a2a4 b5b4 c3d1 b8c6 g1e2 g8f6 f2f3 a8b8 d1f2 e7e5 d4d5 c6a5 b2b3 c7c6 d5c6 d6d5 e4d5 d8d5 e1c1 d5d2 d1d2 a5c6 g2g4 e8g8 g4g5 f6h5 e2g3 c6d4 g3h5 g6h5 e3d4');
INSERT INTO evaluations VALUES(54425,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(54425,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54425,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(54425,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(54425,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(54425,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(54425,36,'cp -6','b8c6 d4f4 d8c7 b1c3 a7a6 f4g3 f7f6 e5f6 c7g3 f6f7 e8f7 h2g3 g8f6 c1e3 h7h6 e1c1 b7b5 f3g5 f7g8 g5h3 c8b7 f2f3 a8d8 e3b6 d8c8 g3g4 f8d6 b6e3 c6e5 c1b1 e5c4 e3c1 c4a3 b2a3 c8c3 h3f4 e6e5 f4e6');
INSERT INTO evaluations VALUES(20429,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(20429,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(20429,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20429,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20429,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(20429,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20429,36,'cp 0','d4f4 g8h6 b1c3 f8e7 f1d3 f7f6 e5f6 e7f6 f4g3 h6f7 h2h4 e6e5 f3g5 e5e4 g5f7 e8f7 d3e2 c6d4 e1g1 f6h4 g3e3 d4c2 e3f4 d8f6 f4c7 f6e7 c7f4');
INSERT INTO evaluations VALUES(19067,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(19067,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(19067,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(19067,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(19067,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(19067,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(19067,36,'cp 51','f1c4 g8f6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8c5 d1h5 g7g6 h5d1 c5b6 g5e4 f7f6 a2a4 e8g8 c1e3 c8f5 d5d6 g8g7 g1h1 a8c8 d1d2 g6g5 a1d1 d8d7 f2f3 d7e6 e3d4 e5d4 c3b5 f5e4 f3e4 e6e4 f1e1 e4d5 b5d4 d5d4 d2d4 b6d4 d1d4 c8c2 d6d7 f8d8');
INSERT INTO evaluations VALUES(5186,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(5186,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(5186,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(5186,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(5186,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(5186,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(5186,36,'cp 75','d4c6 b7c6 d1d4 b4f8 f1e2 d7d5 e4d5 c6d5 c1f4 d8b6 d4b6 a7b6 c3b5 e8d7 e1g1 g8f6 c2c4 f8c5 f1d1 d7e7 b5c7 a8a4 b2b3 a4a7 c4d5 f6d5 c7d5 e6d5 d1d5 c8e6 d5d3 h8d8 d3d8 e7d8 f4e5 f7f6 e5c3 a7c7 g1f1 d8e7 a1e1 c5d6 c3d4 c7c2 e2d3 c2a2 d4b6');
INSERT INTO evaluations VALUES(5165,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(5165,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(5165,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(5165,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(5165,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(5165,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(5165,36,'cp 70','b7c6 d1d4 b4f8 f1e2 d7d5 e4d5 c6d5 c1f4 d8b6 d4b6 a7b6 c3b5 e8d7 e1g1 g8f6 c2c4 f8c5 f1d1 c8b7 a2a3 f6e4 c4d5 b7d5 g1f1 e4f2 d1d5 e6d5 b2b4 c5b4 f1f2 b4c5 f2f3 f7f6 a3a4 g7g5 f4d2 h8e8 b5c3 e8e5 e2d3 h7h5 h2h3 d7e6 g2g3 a8c8 a1b1 h5h4');
INSERT INTO evaluations VALUES(5014,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(5014,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(5014,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(5014,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(5014,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(5014,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(5014,36,'cp 65','d1d4 b4f8 f1e2 d7d5 e4d5 c6d5 c1f4 d8b6 d4b6 a7b6 c3b5 e8d7 e1g1 g8f6 c2c4 c8b7 f1d1 f8c5 e2f3 h8c8 c4d5 f6d5 a2a3 a8a4 f4e5 f7f6 b5c3 a4a8 e5g3 d7e7 c3d5 b7d5 f3d5 e6d5 d1d5 c8d8 d5d1 h7h5 g1f1 g7g5 f2f3 b6b5 d1c1 a8c8 c1c2 c5b6 a1c1 c8c2 c1c2 d8d1 f1e2 d1h1');
INSERT INTO evaluations VALUES(5013,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(5013,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(5013,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(5013,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(5013,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(5013,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(5013,36,'cp 74','b4f8 f1e2 d7d5 e4d5 c6d5 c1f4 d8b6 d4b6 a7b6 c3b5 e8d7 e1g1 g8f6 c2c4 f8c5 f1d1 c8b7 e2f3 h8c8 c4d5 f6d5 a2a3 a8a4 f4e5 f7f6 b5c3 a4a8 e5g3 d7e7 c3d5 b7d5 f3d5 e6d5 d1d5 c8d8 d5d1 b6b5 g1f1 h7h5 f1e2 d8d1 a1d1 b5b4 a3b4 c5b4 e2d3');
INSERT INTO evaluations VALUES(42056,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(42056,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(42056,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(42056,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(42056,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(42056,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(42056,36,'cp 15','b1d2 e8g8 a2a3 b4e7 b2b3 b7b6 c1b2 c8b7 c4d5 d8d5 f1d3 c7c5 d3c4 d5h5 d4c5 e7c5 e1g1 b8d7 d1e2 c5e7 a1c1 f8d8 f3d4 h5e2 c4e2 f6d5 d4c6 b7c6 c1c6 e7f6 b2f6 d7f6 c6c2 a8c8 f1c1 c8c2 c1c2 d5e7 g1f1 d8c8 c2c8 e7c8 d2f3 c8d6 f3d4 f6d7 b3b4');
INSERT INTO evaluations VALUES(42055,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(42055,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(42055,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42055,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42055,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42055,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42055,36,'cp 16','e8g8 a2a3 b4e7 b2b3 b7b6 c1b2 c7c5 d4c5 b6c5 f1d3 b8d7 e1g1 c8a6 f1e1 d8c7 d1c2 h7h6 h2h3 f8d8 a1c1 c7b6 b2c3 a8b8 c1d1 b8a8 d1b1 a8c8 e1d1 e7f8 d3f1 f6e8 c4d5 e6d5 b3b4 a6f1 d2f1 d7f6 f1g3 c5b4 a3b4 e8d6 f3d4 b6b7');
INSERT INTO evaluations VALUES(1819,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(1819,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(1819,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(1819,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(1819,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(1819,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(1819,36,'cp 12','f8c5 d1f3 b7c6 b1c3 f6f3 g2f3 d7d6 h2h4 a7a5 f1e2 c8e6 c1e3 c5e3 f2e3 g8f6 b2b3 e8e7 e1d2 g7g6 h1g1 h8b8 g1g5 h7h6 g5g3 f6d7 a1f1 d7f6 a2a3 b8d8 f3f4 d8g8 e2f3 h6h5 d2c1 f6g4 f3g4 e6g4 g3g1 a8b8');
INSERT INTO evaluations VALUES(21031,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(21031,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(21031,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21031,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21031,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(21031,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(21031,36,'cp 24','g1f3 c8f5 a2a3 g8f6 e2e3 e7e6 c2c4 f8e7 b2b4 e8g8 c1b2 c6b8 c4c5 b7b6 f1e2 a7a5 e1g1 c7c6 b1c3 h7h6 c3a4 b8d7 c5b6 d7b6 a4c5 b6d7 a1c1 a5b4 a3b4 f6e4 e2d3 e4c5 d4c5 f5d3 d1d3 d8c7 b2c3 a8a3 f3d4 e7f6');
INSERT INTO evaluations VALUES(15427,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15427,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(15427,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(15427,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(15427,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(15427,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(15427,36,'cp 44','d2d4 c8d7 b5d3 f8e7 d4d5 c6b8 c3c4 c7c6 b1c3 c6d5 e4d5 e8g8 a2a3 f6e8 a1b1 a7a5 b2b4 d7g4 h2h3 g4h5 c1e3 b8d7 d1c2 a5b4 d3h7 g8h8 a3b4 h5f3 g2f3 e7g5 h7e4 g5e3 f2e3 d8g5 e1e2 f7f5 b1g1');
INSERT INTO evaluations VALUES(38393,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(38393,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38393,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38393,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(38393,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38393,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38393,36,'cp 22','g1f3 g8f6 f1e2 c7c5 c2c3 d8b6 b1a3 e8g8 e1g1 b8d7 h2h3 a7a6 a1c1 b6c7 c3c4 b7b6 a3b1 c8b7 f4h2 a8c8 b1c3 c5d4 e3d4 f6e4 c3e4 b7e4 f1e1 e7e5 d4e5 d6e5 b2b4 e4b7 e2f1 f8d8 d1b3 b7f3 b3f3');
INSERT INTO evaluations VALUES(38394,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38394,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38394,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38394,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38394,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(38394,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38394,36,'cp 0','b8d7 a2a4 e7e5 d4e5 d6e5 f4g3 g8e7 e3e4 e8g8 b1d2 e7c6 f1b5 d7c5 e1g1 d8e7 d1e2 a7a5 c2c3 g8h8 f1d1 f7f6 f3e1 c8e6 d2c4 c5b3 a1a3 b3c5 a3a1');
INSERT INTO evaluations VALUES(6136,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(6136,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6136,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6136,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(6136,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(6136,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(6136,36,'cp 0','a2a4 e7e5 d4e5 d6e5 f4g3 g8e7 f1d3 e8g8 b1c3 e7c6 e1g1 d7c5 d3c4 d8d1 f1d1 c8f5 d1d5 b7b6 d5d2 a8d8 a1d1 d8d2 d1d2 a7a6 c4d5 c6b4 d5c4');
INSERT INTO evaluations VALUES(55173,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55173,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(55173,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55173,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55173,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55173,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55173,36,'cp 0','b8c6 g1f3 g7g6 f1e2 f8g7 c1f4 h7h6 h2h4 b7b6 c2c4 d8d7 b1d2 c8b7 a2a4 e7e6 d2e4 c6b4 f3d2 b4c6');
INSERT INTO evaluations VALUES(21096,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21096,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21096,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21096,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21096,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21096,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21096,36,'cp 0','g1f3 g7g6 f1e2 f8g7 c1f4 h7h6 h2h4 b7b6 c2c4 d8d7 b1d2 c8b7 a2a4 e7e6 d2e4 c6b4 f3d2 b4c6');
INSERT INTO evaluations VALUES(21095,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21095,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21095,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21095,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21095,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21095,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21095,36,'cp 0','g7g6 f1e2 f8g7 c1f4 h7h6 h2h4 b7b6 c2c4 d8d7 b1d2 c8b7 a2a4 e7e6 d2e4 c6b4 f3d2 b4c6');
INSERT INTO evaluations VALUES(53060,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53060,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53060,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53060,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53060,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53060,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53060,36,'cp 0','b8c6 g2g3 e7e6 f1g2 a8b8 g1f3 g8f6 e1g1 f8e7 f1e1 e8g8 d3d4 c5d4 f3d4 h7h6 d4c6 b7c6 b2b3 f6d7 c1b2 c8a6 d1d4 c6c5 d4d2 d8a5 a1d1 a5c7 c3a4 a6b5 a4c3');
INSERT INTO evaluations VALUES(20006,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20006,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20006,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20006,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20006,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20006,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20006,36,'cp 0','g2g3 a8b8 f1g2 e7e6 g1e2 g8f6 e1g1 f8e7 d3d4 c5d4 e2d4 e8g8 f1e1 h7h6 d4c6 b7c6 b2b3 c8a6 c3a4 a6b5 a4c3');
INSERT INTO evaluations VALUES(20005,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20005,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20005,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20005,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20005,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20005,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20005,36,'cp 0','e7e6 f1g2 a8b8 g1e2 f8e7 d3d4 c5d4 e2d4 g8f6 e1g1 e8g8 f1e1 h7h6 d4c6 b7c6 b2b3 c8a6 c1f4 e6e5 f4c1 d8c7 c1b2 f8e8 c3a4 a6b5 a4c3');
INSERT INTO evaluations VALUES(6576,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(6576,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(6576,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(6576,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(6576,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(6576,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(6576,36,'cp -11','g1f3 g8f6 f1d3 d7d5 e4d5 f6d5 e1g1 e8g8 d3e4 d5f6 e4c6 b7c6 b1c3 c8a6 d1d8 f8d8 f1e1 f6d5 c3d5 c6d5 c2c3 c5f8 b2b4 c7c5 b4c5 f8c5 c1e3 c5a3 e1b1 d8d7 b1b3 a3d6 a2a4 a8c8 f3d4 a6c4 b3b2 c4d3 d4b5 d6b8 b5d4 b8e5');
INSERT INTO evaluations VALUES(6575,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(6575,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(6575,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(6575,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(6575,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(6575,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(6575,36,'cp -21','g8f6 f1d3 d7d5 e4d5 f6d5 e1g1 e8g8 d3e4 d5f6 e4c6 b7c6 b1c3 c8a6 d1d8 f8d8 f1e1 f6d5 c3d5 c6d5 c2c3 c5f8 b2b4 c7c5 b4c5 f8c5 c1e3 c5d6 f3d4 d8d7 d4f5 d6f8 e3d4 a6d3 f5e3 f7f6 e1d1 d3g6 a2a4 a7a5 d1d2 g8f7 h2h3 a8c8');
INSERT INTO evaluations VALUES(10312,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(10312,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(10312,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(10312,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(10312,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(10312,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(10312,36,'cp -23','f1d3 d7d5 e4d5 f6d5 e1g1 e8g8 d3e4 d5f6 e4c6 b7c6 b1c3 c8a6 d1d8 f8d8 f1e1 f6d5 c1d2 c5f8 a1d1 a6b7 c3d5 c6d5 d2a5 d8d7 b2b4 a8c8 a2a3 f7f6 f3d4 g8f7 d4b5 a7a6 b5d4 c8e8 e1e8 f7e8 d4b3 f8d6 d1e1 e8f7');
INSERT INTO evaluations VALUES(10310,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(10310,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(10310,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(10310,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(10310,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(10310,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(10310,36,'cp -19','d7d5 e4d5 f6d5 e1g1 e8g8 d3e4 d5f6 e4c6 b7c6 b1c3 c8a6 d1d8 f8d8 f1e1 f6d5 c3d5 c6d5 c2c3 c5f8 b2b4 c7c5 b4c5 f8c5 a2a4 d8d7 c1a3 c5a3 a1a3 g7g6 h2h4 g8g7 f3d4 a8c8 h4h5 d7b7 d4b5 a6b5 a4b5 b7b5 a3a7');
INSERT INTO evaluations VALUES(35267,30,'cp 241',NULL);
INSERT INTO evaluations VALUES(35267,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(35267,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(35267,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(35267,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(35267,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(35267,36,'cp 219','b1c3 b8c6 c3d5 e7d8 g1f3 d7d6 c1g5 g8e7 d5e7 f8e7 g5e7 d8e7 e5d6 c7d6 e2e3 e8g8 f3d4 c8d7 f1e2 c6d4 d1d4 d7c6 e1c1 f8d8 d4g4 e7e6 g4e6 f7e6 f2f3 d6d5 d1d2 g8f7 f3f4 a8c8 d2d4 h7h6 h1f1 d8g8 g2g3 f7e7 c1d2 g7g5 c2c4 a7a5 c4d5 c6d5');
INSERT INTO evaluations VALUES(21394,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21394,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(21394,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(21394,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(21394,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21394,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(21394,36,'cp 24','g8f6 b1d2 f8d6 g1f3 d8e7 f4g3 e6e5 f1b5 e5d4 c3d4 e8g8 a1c1 d6g3 h2g3 c6b8 b5d3 c7c6 e1g1 b8d7 d1c2 f8e8 f1e1 f6e4 b2b4 d7f6 b4b5 c6b5 d3b5 c8d7 b5d7 f6d7 a2a4 e4d2 c2d2 d7f6 f3e5');
INSERT INTO evaluations VALUES(34249,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(34249,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(34249,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(34249,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(34249,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(34249,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(34249,36,'cp 100','g1f3 g8f6 e2e4 b8c6 a2a3 d7d6 c3d5 c5a5 b2b4 a5d8 f1d3 g7g6 d5f6 e7f6 c1e3 f8g7 c2c4 e8g8 a1c1 c6e5 f3e5 f6e5 e1g1 c8e6 d1d2 f8e8 a3a4 a7a5 b4b5 g7f8 h2h3 a8c8 g1h1 b7b6 f1d1 h7h5 f2f3 d8c7');
INSERT INTO evaluations VALUES(34248,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(34248,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(34248,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(34248,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(34248,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(34248,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(34248,36,'cp 115','g8f6 e2e4 b8c6 c1e3 c5a5 f3d2 a5d8 d2b3 e7e6 a2a3 d7d5 e4d5 f6d5 c3d5 e6d5 d1f3 f8d6 e1c1 c8e6 b3c5 d6c5 e3c5 d8g5 c5e3 g5h4 f3g3 h4g3 h2g3 e8g8 f2f3 a8c8 e3f2 h7h6 f1d3 f7f6 h1e1 f8e8 b2b3 a7a6 c1b1');
INSERT INTO evaluations VALUES(33562,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(33562,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(33562,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(33562,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(33562,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(33562,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(33562,36,'cp 125','e2e4 b8c6 c1e3 c5a5 f3d2 a5d8 d2b3 e7e6 d1d2 f8e7 e1c1 e8g8 a2a3 f6g4 e3c5 d7d6 d2e2 g4f6 f2f4 d6d5 c5e3 b7b5 e2b5 d8c7 e4d5 e6d5 b5a4 c8d7 f1b5 f8c8 c1b1 a7a6 b5c6 d7c6 a4a5 c7d7 e3b6 d7b7 b6c5 e7c5 b3c5');
INSERT INTO evaluations VALUES(33561,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(33561,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(33561,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(33561,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(33561,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(33561,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(33561,36,'cp 105','b8c6 c1e3 c5a5 f3d2 a5d8 d2b3 e7e6 d1d2 f8e7 e1c1 e8g8 a2a3 f6g4 e3c5 d7d6 d2e2 g4f6 f2f4 d6d5 c5e3 c6a5 e4e5 f6d7 b3a5 d8a5 e2e1 e7c5 e3d2 a5d8 f4f5 a7a6 e1g3 d8c7 d2f4 b7b5 f5f6 g7g6 g3g5 g8h8 f1d3 b5b4 a3b4 c5b4');
INSERT INTO evaluations VALUES(29765,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(29765,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(29765,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(29765,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(29765,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(29765,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(29765,36,'cp 94','h2h3 g4f3 d1f3 c7c6 c1e3 e7e6 a2a3 f8e7 e1c1 b8d7 c1b1 e8g8 f1d3 f6d5 c3d5 c6d5 f3e2 a8b8 e3f4 b8c8 f4c1 d7b6 h3h4 b6a4 d3b5 a4b6 b2b3 b6d7 g2g4 a7a6 b5d3 d8a5 e2e1 a5e1 d1e1 d7f6 g4g5 f6h5 c2c4 e7d6 c4c5 d6c7');
INSERT INTO evaluations VALUES(29762,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(29762,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(29762,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(29762,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(29762,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(29762,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(29762,36,'cp 76','g4f3 d1f3 c7c6 c1e3 e7e6 f1d3 d8a5 e1g1 b8d7 c3e2 f8d6 c2c4 e8g8 a2a3 e6e5 b2b4 a5a6 b4b5 c6b5 c4b5 a6a5 f3b7 e5e4 d3e4 a8b8 b7c6 f6e4 c6e4 d7f6 e4c6 b8b6 c6f3 b6b5 a3a4 b5b3 e2g3 a5d5 g3f5 d5f3 g2f3 b3b6 g1g2 f6d5 f5d6 b6d6');
INSERT INTO evaluations VALUES(29770,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(29770,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(29770,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(29770,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(29770,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(29770,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(29770,36,'cp 98','d1f3 c7c6 c1e3 e7e6 f1d3 d8a5 e1g1 b8d7 c3e2 f8d6 c2c4 e8g8 a2a3 f8e8 b2b4 a5d8 a1d1 d8e7 e3g5 d6c7 f1e1 a7a5 b4b5 c6b5 c4c5 h7h6 g5h4 b5b4 d3e4 b4a3 e4b7 a8b8 b7c6 a3a2 d1a1 g7g5 h4g3 c7g3 f2g3 f6d5 a1a2 d5b4 a2a5 b4c6');
INSERT INTO evaluations VALUES(29768,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(29768,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(29768,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(29768,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(29768,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(29768,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(29768,36,'cp 91','c7c6 c1e3 e7e6 f1d3 b8d7 e1c1 f8b4 c3e2 f6d5 c1b1 d5e3 f2e3 b4d6 e3e4 d8f6 f3e3 e6e5 c2c3 g7g6 g2g4 f6e7 d4d5 d7b6 d3c2 e8c8 e3f3 c6d5 e4d5 e5e4 c2e4 h8e8 e4d3 e7e3 f3g2 e3c5 h1f1 c5d5 g2d5 b6d5 f1f7 d5e3');
INSERT INTO evaluations VALUES(29079,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(29079,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(29079,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(29079,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(29079,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(29079,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(29079,36,'cp 85','c1e3 e7e6 f1d3 b8d7 e1c1 f8b4 c3e2 f6d5 c1b1 d5e3 f2e3 b4d6 e3e4 d8f6 f3e3 e6e5 h1f1 f6e7 d3c4 e8g8 d4d5 d7b6 c4b3 c6d5 b3d5 a8d8 e2c3 d6b4 d5b3 d8d1 f1d1 e7c5 e3c5 b4c5 a2a4 c5b4 c3d5 b4c5');
INSERT INTO evaluations VALUES(29078,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(29078,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(29078,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(29078,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(29078,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(29078,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(29078,36,'cp 86','e7e6 f1d3 b8d7 e1c1 f8b4 c3e2 f6d5 c1b1 d5e3 f2e3 b4d6 h3h4 d8e7 d1f1 f7f5 g2g4 f5g4 f3g4 e8c8 h4h5 d7f6 g4g5 h8f8 f1g1 d8d7 c2c3 c8b8 e3e4 e6e5 e2g3 f6d5 g3f5 e7g5 g1g5 d5f4 d3c4 b7b5 c4b3 d6c7 g5g7 d7g7 f5g7 e5d4 c3d4');
INSERT INTO evaluations VALUES(29312,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(29312,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(29312,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(29312,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(29312,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(29312,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(29312,36,'cp 85','f1d3 b8d7 e1c1 f8b4 c3e2 f6d5 c1b1 d5e3 f2e3 b4d6 h3h4 d8e7 h1f1 h8f8 h4h5 e8c8 d3h7 g7g6 e3e4 g6h5 e4e5 d6c7 e2f4 h5h4 h7e4 f7f5 e4d3 f8g8 d3c4 g8g3 f3h5 d7f8 h5h6 d8e8 f1h1 c7d8 f4h5 e7g5 h6g5 g3g5 h5f6');
INSERT INTO evaluations VALUES(19432,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(19432,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(19432,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(19432,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(19432,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(19432,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(19432,36,'cp -30','d7d6 f1b5 c8d7 d2d3 g8f6 g1f3 g7g6 b1c3 f8h6 e1g1 e8g8 c3d5 e7e5 d5f6 d8f6 b2c1 h6c1 d1c1 f6e7 c1h6 f7f5 b5c4 g8h8 e4f5 f8f5 c4d5 a8f8 c2c3 f5h5 h6d2 e7f6 f1e1 d7g4');
INSERT INTO evaluations VALUES(19994,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19994,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(19994,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(19994,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19994,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(19994,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(19994,36,'cp -10','f1b5 c8d7 d2d3 g8f6 b1d2 g7g6 a2a4 f8g7 g1e2 e8g8 e1g1 e7e6 f2f4 a7a6 b5c6 d7c6 d2f3 b7b5 a4b5 a6b5 a1a8 d8a8 d1a1 a8a1 f1a1 f8a8 a1a8 c6a8 h2h3 h7h5 b2f6 g7f6 g1f2 f6b2 f2e3 g8f8 g2g4 f8e7 d3d4 h5g4 h3g4 f7f5 e4f5 g6f5');
INSERT INTO evaluations VALUES(19919,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(19919,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(19919,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(19919,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(19919,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(19919,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(19919,36,'cp -13','c8d7 d2d3 g8f6 b1d2 g7g6 a2a4 f8g7 g1e2 e8g8 d2c4 c6b4 e1g1 d7e6 c4e3 a7a6 b5c4 d6d5 e4d5 e6d5 c4d5 b4d5 e3d5 f6d5 b2g7 g8g7 d1d2 a8c8 a1e1 e7e6 e2f4 d5f4 d2f4 d8c7 f4c1 c8d8 e1e4 d8d4 f1e1 f8d8 e4d4 d8d4 e1e4');
INSERT INTO evaluations VALUES(57520,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(57520,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(57520,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(57520,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(57520,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(57520,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(57520,36,'cp 181','f3d4 d8f6 d4b5 f6e5 d1e2 e8d8 b5c3 b8c6 e2e4 e5e4 c3e4 d7d5 e4g5 c6d4 e1d1 c8f5 d2d3 d8e8 b1c3 c7c6 c3e2 d4e2 f1e2 f8d6 f2f4 h7h6 g5f3 g8e7 c1e3 h6h5 f3e5 c6c5 h1e1 d6e5 f4e5 d5d4 e3g1 f5g4 c2c3 e7c6 e2g4 h5g4 e1e4 g4g3');
INSERT INTO evaluations VALUES(57517,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(57517,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(57517,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(57517,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(57517,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(57517,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(57517,36,'cp 174','d8f6 d4b5 f6e5 d1e2 e8d8 b5c3 b8c6 e2e4 e5e4 c3e4 d7d5 e4g5 c6d4 e1d1 c8f5 d2d3 d8e8 b1c3 c7c6 c3e2 d4e2 f1e2 f8d6 h1e1 g8e7 f2f4 h7h6 g5f3 c6c5 h2h3 f5d7 d3d4 c5d4 f3d4 h8f8 e2h5 e8d8 h5g4 f8f7 g2g3 d7g4 h3g4 d8d7 d4f3 a8f8');
INSERT INTO evaluations VALUES(35231,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(35231,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(35231,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(35231,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(35231,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(35231,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(35231,36,'cp 160','d4b5 f6e5 d1e2 e8d8 b5c3 b8c6 e2e4 e5e4 c3e4 d7d5 e4g5 c6d4 e1d1 c8f5 d2d3 d8e8 b1c3 c7c6 c3e2 d4e2 f1e2 f8d6 f2f4 g8e7 c1e3 h8f8 g2g3 c6c5 c2c3 e8d8 d1d2 d8c7 g5f3 e7c6 a1e1 a8e8 e2d1 d5d4 e3f2');
INSERT INTO evaluations VALUES(54134,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54134,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54134,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(54134,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(54134,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54134,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(54134,36,'cp 12','e6d5 g1f3 g8f6 f1b5 c8d7 b5d7 b8d7 e1g1 f8e7 d4c5 d7c5 d2b3 c5e4 c2c3 e8g8 c1e3 f8e8 f1e1 b7b6 h2h3 e7c5 b3d4 e4d6 e3g5 e8e1 d1e1 d6e4 g5f4 h7h6 e1e2 d8d7 f4e5 d7b7 a1e1 a8e8 e5f6 e4f6 e2b5 e8e1 f3e1 c5d4');
INSERT INTO evaluations VALUES(54531,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54531,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(54531,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(54531,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(54531,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(54531,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54531,36,'cp 12','f1b5 c8d7 b5d7 d8d7 g1e2 g8f6 e1g1 f8e7 d4c5 e7c5 d2b3 c5b6 d1d3 e8g8 c1g5 f6e4 g5e3 b6e3 d3e3 b8c6 a1d1 f8e8 f1e1 e8e5 e2d4 e4f6 e3d2 a8e8 d4f3 e5e1 d1e1 e8e1 d2e1 h7h6 c2c3 a7a6 g2g3 f6e4 e1e2 e4d6 e2d3 d6c4');
INSERT INTO evaluations VALUES(54465,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(54465,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(54465,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(54465,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54465,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54465,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54465,36,'cp 8','c8d7 b5d7 d8d7 d4c5 f8c5 d2b3 c5e7 g1f3 g8f6 e1g1 e8g8 f1e1 b8c6 c1g5 h7h6 g5f6 e7f6 c2c3 f8e8 d1d3 e8e1 a1e1 a8d8 g2g3 a7a6 h2h4 g7g6 g1g2 h6h5 d3e3 g8g7 e1e2 d7c7 e3d3 d8d7 e2e8 d7d8 e8d8 c7d8 b3c5');
INSERT INTO evaluations VALUES(30907,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(30907,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(30907,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(30907,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30907,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(30907,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(30907,36,'cp 17','b5d7 d8d7 d4c5 f8c5 d2b3 c5b6 g1f3 d7e6 d1e2 b8c6 a2a4 a7a6 e2e6 f7e6 e1g1 g8f6 f1e1 e8c8 a4a5 b6a7 c1e3 d5d4 e3f4 d4d3 a1c1 h8f8 c2d3 d8d3 f4e3 a7e3 e1e3 d3e3 f2e3 c8b8 f3d4 c6d4 b3d4 f8f7 h2h3 f6d5 c1e1 f7e7');
INSERT INTO evaluations VALUES(10023,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(10023,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(10023,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(10023,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10023,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(10023,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10023,36,'cp 0','c6d4 b5c4 e8g8 f3e5 d7d6 e5f7 f8f7 c4f7 g8f7 c3a4 c8g4 f2f3 f6h5 f3g4 d8h4 e1d2 h4f2 d2c3 d4e2 c3b3 e2d4');
INSERT INTO evaluations VALUES(10757,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10757,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(10757,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(10757,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10757,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10757,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10757,36,'cp 0','f3e5 e8g8 b5c4 d7d6 e5f7 f8f7 c4f7 g8f7 c3a4 c8g4 f2f3 f6e4 d3e4 d8h4 g2g3 d4f3 d1f3 g4f3 g3h4 f3h1 a4c5 d6c5 c1f4 a8d8 f4c7 d8d7 c7b8 a7a6 a1d1 d7d1 e1d1 h1e4 b8a7 c5c4 d1d2 f7e6 d2c3 e6f5 c3c4 e4c2 a7d4 g7g6 c4c5 c2b1 a2a4 b1c2 a4a5 f5f4 d4f2 c2e4 f2g3 f4f5 c5c4');
INSERT INTO evaluations VALUES(57598,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(57598,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(57598,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(57598,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(57598,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(57598,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(57598,36,'cp 17','c2c4 g8f6 g1f3 e7e6 e2e3 b7b6 f1e2 g7g6 b1c3 c8b7 a1b1 a7a5 e1g1 d8e7 a2a3 f8g7 b2b4 a5b4 a3b4 e8g8 d1b3 b8c6 b4b5 c6d8 c1b2 d8f7 b1d1 f8e8 d1a1 d7d6 f3d4 f6d7 b3c2 d7c5 d4c6 b7c6 b5c6 a8a1 b2a1');
INSERT INTO evaluations VALUES(57697,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(57697,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(57697,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(57697,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(57697,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(57697,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(57697,36,'cp 16','c7c5 c2c3 d7d5 d2d4 c5d4 c3d4 g8f6 g1f3 f8g7 b1c3 e8g8 f3e5 c8f5 c1e3 f6e4 d1b3 e4c3 b2c3 b8c6 g2d5 g7e5 d5c6 b7c6 d4e5 d8a5 b3b4 a5e5 b4d4 e5d5 f2f3 e7e5 d4c5 e5e4 c5d5 c6d5 e1f2 f5e6');
INSERT INTO evaluations VALUES(52319,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52319,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52319,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52319,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(52319,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(52319,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(52319,36,'cp 16','c2c3 f8g7 d2d4 c5d4 c3d4 d7d5 g1f3 b8c6 b1c3 g8f6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 e8g8 e1g1 a8c8 a1c1 d8e7 c1c2 c8c6 f1c1 f8c8 a2a3 e7d8 h2h4 h7h6 e5f4 h6h5 g2f3 d8a5 f4g5 a5b6 c3a4');
INSERT INTO evaluations VALUES(37397,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(37397,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(37397,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(37397,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37397,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(37397,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(37397,36,'cp 23','e3e4 d7d6 f1e2 e7e5 d4d5 b8a6 c1g5 c7c6 g1f3 h7h6 g5h4 g6g5 h4g3 f6h5 e1g1 h5f4 a1c1 d8b6 c1b1 b6d8 f3d2 a6b4 f2f3 f7f5 e4f5 f4e2 d1e2 b4c2 d2e4 c2d4 e2d1 c8f5 g3f2 a8c8 b1c1 c6c5 c3b5 d4b5 c4b5 b7b6 a2a3');
INSERT INTO evaluations VALUES(19324,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(19324,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(19324,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(19324,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(19324,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(19324,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(19324,36,'cp 107','c3d4 f8g7 b1c3 d7d6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 c8d7 c1f4 a7a6 b5c6 d7c6 d1d2 a8c8 f4h6 d6d5 h6g7 g8g7 e4e5 h7h6 a2a4 d8a5 d2f4 a5d8 c3e2 e7f5 f4d2 f5h4 f3h4 d8h4 b2b3 h4e7 e1f1');
INSERT INTO evaluations VALUES(47440,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(47440,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(47440,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(47440,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(47440,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(47440,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(47440,36,'cp 78','d5f6 g1f3 c7c5 b1c3 c5d4 d1d4 d8d4 f3d4 a7a6 c3a4 b8d7 f1e2 e7e5 d4c2 a8b8 a4c3 d7c5 c1e3 f6g4 b2b4 g4e3 f2e3 c5d7 e1c1 f8e7 c4c5 d7f6 d1d2 c8e6 h1d1 a6a5 a2a3 e8g8 c1b2 f8d8 d2d8 b8d8 d1d8 e7d8 e2f3 a5b4 a3b4');
INSERT INTO evaluations VALUES(47340,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(47340,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(47340,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(47340,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(47340,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(47340,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(47340,36,'cp 61','g1f3 c7c5 b1c3 c5d4 d1d4 d8d4 f3d4 a7a6 c3a4 b8d7 f1e2 e7e5 d4c2 a8b8 a4c3 d7c5 c1e3 c8e6 e1c1 b8c8 b2b3 f8e7 f2f3 e8g8 e3f2 f8d8 d1d8 c8d8 h1d1 d8d1 c1d1 g8f8 c2e3 c5d7 e2d3 e7c5 d3b1 c5d4 c3e2 d4b6 b1c2 b6c7 f2h4 h7h5');
INSERT INTO evaluations VALUES(47339,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(47339,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(47339,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(47339,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(47339,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(47339,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(47339,36,'cp 68','c7c5 b1c3 c5d4 d1d4 d8d4 f3d4 a7a6 c3a4 b8d7 f1e2 e7e5 d4c2 a8b8 a4c3 d7c5 c1e3 c8e6 e1c1 f8e7 f2f3 e8g8 g2g4 a6a5 e3f2 h7h5 g4g5 f6d7 h2h4 d7b6 c2e3 b8c8 b2b3 f8d8 d1d8 c8d8 c1c2 d8d4 c3b5 d4d8 h1e1');
INSERT INTO evaluations VALUES(20082,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(20082,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(20082,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(20082,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(20082,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(20082,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(20082,36,'cp 152','d4d5 c6b8 b1c3 g7g6 g1e2 g8f6 e2d4 f8g7 f1e2 e8g8 c1e3 e7e5 d5e6 c8e6 d4e6 f7e6 d1b3 d8e7 a1d1 b8c6 e1g1 a7a6 a2a4 a8b8 h2h3 f8d8 e2c4 d8e8 b3a3 g7f8 a4a5 b8c8 c4b3 d6d5 a3e7 f8e7 e4d5 c6a5 b3a2 f6d5 c3d5');
INSERT INTO evaluations VALUES(20051,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(20051,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(20051,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(20051,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(20051,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(20051,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(20051,36,'cp 147','c6b8 b1c3 g8f6 g1f3 g7g6 f3d4 f8g7 f1e2 e8g8 c1e3 e7e5 d5e6 c8e6 d4e6 f7e6 d1b3 d8e7 a1d1 b8c6 e1g1 a8b8 h2h3 a7a6 a2a4 f8d8 e2c4 d8e8 b3a3 g7f8 a4a5 d6d5 a3e7 c6e7 c4b3 f8g7 e3b6 f6d7 b3a4 e7c6 b6e3 g7c3 e4d5 e6d5 b2c3 d7f6 e3g5 f6e4 d1d5 e4g5 d5g5');
INSERT INTO evaluations VALUES(51327,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(51327,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(51327,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(51327,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(51327,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(51327,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(51327,36,'cp -29','a7a6 g1f3 e7e6 a2a4 b5b4 c3e2 b8c6 e2g3 c6a5 c1g5 f8e7 g5e7 g8e7 a1c1 c7c5 f1c4 e8g8 e1g1 c5d4 d1d4 d8d4 f3d4 c8d7 c4e2 d7a4 c1c7 e7g6 f1a1 a5b3 d4b3 a4b3 a1a6 g6f4 a6a8 f8a8 h2h4 a8a1 g1h2 g7g6 c7b7 a1a4 h4h5 g8g7');
INSERT INTO evaluations VALUES(51122,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(51122,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(51122,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(51122,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(51122,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(51122,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(51122,36,'cp -5','g1f3 e7e6 a2a4 b5b4 c3e2 b8c6 a4a5 b4b3 c1g5 g8f6 e2c3 c6b4 a1c1 b4c2 c1c2 b3c2 d1c2 c7c5 d4d5 e6d5 g5f6 g7f6 e4d5 a8b8 f1e2 f8g7 c2e4 e8f8 e1g1 f6f5 e4c2 h8g8 e2c4 g7f6 f1e1 b8b4 c3a4 g8g6 a4b6 b4b2');
INSERT INTO evaluations VALUES(51121,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(51121,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(51121,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(51121,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(51121,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(51121,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(51121,36,'cp -27','e7e6 a2a4 b5b4 c3e2 b8c6 e2g3 c6a5 c1g5 f8e7 g5e7 g8e7 a1c1 e8g8 f1c4 a5c4 c1c4 a6a5 c4c5 c8a6 d1c2 d8d6 g3e2 b4b3 c2c3 h7h6 f3d2 f7f5 d2c4 a6c4 e4e5 d6d7 c3c4 e7d5 c5b5 a8b8 b5a5 f5f4 a5b5 b8b5 a4b5 f8b8');
INSERT INTO evaluations VALUES(51142,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(51142,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(51142,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(51142,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(51142,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(51142,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(51142,36,'cp -12','a2a4 b5b4 c3e2 b8c6 c1g5 f8e7 g5e7 g8e7 e2g3 c6a5 a1c1 e8g8 f1c4 a5c4 c1c4 a6a5 c4c5 c8a6 g3e2 d8d6 d1c2 b4b3 c2c3 f8b8 f3e5 b8b4 c5a5 d6b6 c3c5 b6c5 a5c5 b4a4 e1d2 a6e2 d2e2 a4d4 e2e3 a8d8 h1c1 e7g6 c1c3 g6h4 g2g3 h4g2 e3f3 g2e1 f3e2 d4e4 e2f1 d8d1 c3b3');
INSERT INTO evaluations VALUES(47219,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(47219,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(47219,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(47219,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(47219,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(47219,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(47219,36,'cp 96','e4e5 f6d7 f1d3 c7c5 c2c3 b8c6 g1e2 d8b6 e1g1 e7e6 d3c2 f8e7 a2a3 e8g8 b2b4 c5d4 d1d3 g7g6 c3d4 f7f6 b4b5 c6d8 e5f6 d7f6 b1c3 c8d7 a3a4 d8f7 a4a5 b6c7 c2b3 e7b4 c1f4 b4d6 f4d6 f7d6 a5a6 c7b6 a6b7 b6b7');
INSERT INTO evaluations VALUES(47292,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(47292,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(47292,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(47292,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(47292,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(47292,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(47292,36,'cp 94','f6d7 c2c3 c7c5 f1d3 b8c6 g1e2 e7e6 e1g1 d8b6 d3c2 f8e7 a2a3 e8g8 d1d3 g7g6 b1d2 c5d4 c3d4 f7f6 e5f6 e7f6 d2f3 f6g7 c1e3 d7f6 a1c1 f6g4 c2a4 g4e3 d3e3 c6e7 e2g3 b6a5 b2b3 b7b5 c1c5 c8d7 f3e5');
INSERT INTO evaluations VALUES(49178,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(49178,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(49178,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(49178,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(49178,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(49178,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(49178,36,'cp 94','c2c3 c7c5 f1d3 b8c6 g1e2 e7e6 e1g1 d8b6 d3c2 f8e7 a2a3 e8g8 b2b4 c5d4 c3d4 f7f6 d1d3 g7g6 b4b5 c6d8 e5f6 d7f6 b1c3 c8d7 a3a4 d8f7 a4a5 b6c7 c2b3 a7a6 b5b6 c7d8 e2f4 a8c8 f1e1 e7b4 c1d2 f8e8');
INSERT INTO evaluations VALUES(48247,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(48247,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(48247,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48247,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(48247,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(48247,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(48247,36,'cp -15','d5c4 d1a4 b8d7 e2e3 f6d5 f1c4 f8b4 b1c3 d5f4 e3f4 b4c3 b2c3 e8g8 c4e2 b7b6 e1g1 c8b7 f1d1 d8e7 f3e5 d7e5 d4e5 h7h6 a4b4 e7b4 c3b4 a7a5 b4b5 f8d8 f2f3 c7c5 g1f2 d8d1 a1d1 b7d5 a2a4 a8c8 f4f5 c5c4 f2e3 c4c3 d1c1 c3c2');
INSERT INTO evaluations VALUES(48485,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(48485,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(48485,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48485,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48485,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48485,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48485,36,'cp -13','d1a4 b8d7 e2e3 f6d5 f1c4 f8b4 b1c3 b4c3 b2c3 d5f4 e3f4 e8g8 c4e2 b7b6 e1g1 c8b7 f1d1 d8e7 f3e5 d7e5 d4e5 h7h6 a4b4 e7b4 c3b4 a7a5 a2a3 b7d5 b4b5 f8c8 a1c1 c7c5 b5c6 d5c6 f2f3 c6a4 c1c8 a8c8 d1d6 c8c3 d6b6 c3a3 b6b8 g8h7 b8b7 h7g6 h2h4 a4c6 h4h5 g6h5');
INSERT INTO evaluations VALUES(17398,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(17398,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(17398,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(17398,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(17398,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(17398,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(17398,36,'cp -32','g8f6 d2d3 d7d5 c4a2 d5e4 c3e4 f8e7 c2c3 e8g8 e1g1 b7b6 f3g5 f6e4 g5e4 c6e5 d3d4 c5d4 c3d4 c8b7 d4e5 b7e4 c1e3 e4d3 e3b6 d8b6 d1d3 b6b2 a2c4 b2e5 c4a6 e7f6 a1d1 f8d8 d3e2 d8d1 f1d1 e5e2 a6e2 a8a4');
INSERT INTO evaluations VALUES(11079,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(11079,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(11079,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(11079,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(11079,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(11079,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(11079,36,'cp -30','d2d3 d7d5 c4b3 d5e4 c3e4 f8e7 c2c3 e8g8 e1g1 b7b6 e4f6 e7f6 c1f4 a8a7 f1e1 a7d7 d1e2 c8b7 a1d1 f8e8 f4g3 g7g6 b3c4 f6g7 g3f4 h7h6 h2h4 h6h5 f4g3 c6e7 f3e5 g7e5 e2e5 e7f5 e5f4 f5g3 f2g3');
INSERT INTO evaluations VALUES(52606,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(52606,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(52606,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(52606,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(52606,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(52606,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(52606,36,'cp 49','d2d4 c5d4 f3d4 f8g7 b1c3 g8f6 f2f3 d8b6 c4b5 b8c6 c1e3 b6c7 d1d2 c8d7 b5c6 b7c6 e3h6 e8g8 e1c1 c6c5 d4e2 c7b7 h6g7 g8g7 g2g4 f8b8 b2b3 a7a5 h2h4 c5c4 h4h5 c4b3 a2b3 a5a4 h5g6 a4b3 d2h6 g7g8 c3d5 a8a1 c1d2 a1d1 d2d1 b3c2 d1e1 b7b1 e1f2 f7g6 d5f6 e7f6');
INSERT INTO evaluations VALUES(52609,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(52609,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52609,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(52609,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52609,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(52609,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52609,36,'cp 37','c5d4 f3d4 f8g7 c1e3 g8f6 f2f3 d6d5 c4b5 c8d7 e4e5 f6g8 f3f4 d7b5 d4b5 a7a6 b5d4 e7e6 b1d2 g8e7 d1f3 b8c6 d2b3 c6d4 b3d4 d8a5 c2c3 a8c8 e1g1 e7c6 f3f2 c6d4 e3d4 e8g8 h2h3 f7f6 e5f6 g7f6 a1e1 f6d4 f2d4 a5a2 e1e6');
INSERT INTO evaluations VALUES(52636,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(52636,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52636,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52636,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(52636,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(52636,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(52636,36,'cp 38','f3d4 f8g7 c4b3 g8f6 f2f3 b8c6 c1e3 d6d5 b1c3 d5e4 d4c6 d8d1 a1d1 b7c6 f3e4 f6g4 e3c5 g4e5 h2h3 h7h5 c5e3 g7f6 c3a4 g6g5 a4c5 h5h4 e1d2 h8g8 d2c1 g5g4 b3a4 a8b8 c1b1 g8g6 e3f4 b8b6 b2b3 f6g7 d1d5 e5d7 c5d7 c8d7 d5c5 g4h3');
INSERT INTO evaluations VALUES(52635,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52635,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52635,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(52635,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(52635,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(52635,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(52635,36,'cp 17','f8g7 c1e3 g8f6 f2f3 d6d5 c4b5 c8d7 e4e5 f6g8 b5d7 d8d7 f3f4 b8c6 b1c3 e7e6 d1d3 c6d4 e3d4 g8e7 e1c1 h7h5 g2g3 a7a6 c1b1 d7c7 c3e2 e7f5 h2h3 g7f8 g3g4 f5d4 e2d4 f8c5 d4b3 e8c8 f4f5 g6f5 g4f5 c5b6 f5e6 f7e6 d3g6 c7e7 h1f1 h5h4');
INSERT INTO evaluations VALUES(37992,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(37992,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(37992,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(37992,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(37992,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(37992,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(37992,36,'cp 26','c1e3 g8f6 f2f3 d6d5 c4b5 c8d7 e4e5 f6g8 f3f4 d7b5 d4b5 a7a6 b5d4 b8c6 d4c6 b7c6 b1d2 g8h6 e1g1 h6f5 e3f2 e7e6 g2g4 f5e7 d2f3 d8a5 d1d4 h7h5 h2h3 a8c8 b2b4 c6c5 b4a5 c5d4 f1b1 h5g4 h3g4 c8c2 f3d4 c2c8 a2a3 g7f8 g1g2 g6g5 f4g5');
INSERT INTO evaluations VALUES(17848,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(17848,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(17848,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(17848,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(17848,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(17848,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(17848,36,'cp 169','d2d4 f8d6 d4e5 c6e5 c4b3 g8f6 c1f4 e5f3 d1f3 d6f4 f3f4 d7d6 e1c1 d8e7 h1e1 e7e5 f4e3 c8e6 f2f4 e5c5 e3c5 d6c5 e4e5 f6d7 c3d5 a8c8 c2c3 b7b5 b3c2 g7g6 b2b3 c5c4 b3c4 c7c6 d5e3 d7b6 c4b5 a6b5 g2g4 e8g8 f4f5 e6a2 d1d6 b6c4 e3c4');
INSERT INTO evaluations VALUES(17860,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(17860,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(17860,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(17860,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(17860,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(17860,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(17860,36,'cp 161','f8d6 d4e5 c6e5 c4b3 g8f6 c1f4 e5f3 d1f3 d6f4 f3f4 d7d6 e1c1 d8e7 h1e1 e7e5 f4e3 c8e6 f2f4 e5c5 e3c5 d6c5 e4e5 f6d7 c3d5 a8c8 a2a4 b7b5 c2c4 e6d5 d1d5 d7b6 d5c5 b6d7 c5d5 c7c6 d5d6 d7c5 b3d1 b5c4 d1c2 e8e7 d6d4 c5b3 c2b3 c4b3 e1d1 c8c7 c1d2 f7f6 e5f6 e7f6');
INSERT INTO evaluations VALUES(18802,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(18802,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(18802,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(18802,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(18802,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(18802,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(18802,36,'cp 89','d4d5 e6d5 e4d5 c6e5 g1f3 d8e7 f3e5 e7e5 f1e2 g8f6 b1a3 f8e7 e1g1 e5d5 a3b5 d5d1 f1d1 e8g8 e2f3 a7a6 b5d6 e7d6 d1d6 f8e8 c1e3 f6e4 d6b6 a6a5 a1e1 d7d6 h2h3 h7h6 e3h6 g7h6 e1e4 e8e4 f3e4 a8a6 b6a6 b7a6 e4d5 c8e6 d5e6 f7e6 g2g4 d6d5 g1g2 a5a4 f2f4 g8g7 g2f3 g7g6');
INSERT INTO evaluations VALUES(18864,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(18864,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(18864,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(18864,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(18864,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(18864,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(18864,36,'cp 75','e6d5 e4d5 c6e5 b1a3 d7d6 g1f3 e5f3 d1f3 g8f6 h2h3 a7a6 f1d3 f8e7 e1g1 e8g8 f1e1 c8d7 c3c4 f8e8 c1f4 h7h6 b2b3 e7f8 a3c2 b7b5 c2e3 f6h7 e3g4 d7g4 f3g4 d8f6 f4e3 b5c4 b3c4 g7g6 a1b1 a8b8 b1b8 e8b8 g2g3');
INSERT INTO evaluations VALUES(19044,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(19044,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(19044,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(19044,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(19044,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(19044,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(19044,36,'cp 87','e4d5 c6e5 g1f3 d8e7 f3e5 e7e5 f1e2 g8f6 c3c4 f8d6 b1c3 e8g8 g2g3 f6e4 c3e4 e5e4 e1g1 e4e7 e2d3 d6e5 f1e1 d7d6 c1f4 f7f5 f4e5 d6e5 d1d2 c8d7 f2f3 e7d6 e1e2 a8e8 a1e1 e8e7 a2a3 a7a6 d2c3 f8e8 c3c2 e8f8 a3a4 h7h5 c2c3 f8e8 c3d2');
INSERT INTO evaluations VALUES(19039,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(19039,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(19039,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(19039,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(19039,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(19039,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(19039,36,'cp 89','c6e5 g1f3 d8e7 f3e5 e7e5 f1e2 g8f6 e1g1 e5d5 b1a3 f8e7 a3b5 d5d1 f1d1 e8g8 e2f3 f6e8 d1e1 e7d8 c1f4 a7a6 b5d6 d8c7 d6c8 a8c8 f4g5 e8f6 e1e7 c8d8 a1e1 b7b5 g1f1 b5b4 c3b4 c7d6 e7e2 c5b4 e1d1 d6c5 e2e5 d7d5 g5f6 g7f6 e5d5 c5b6 d5d6 b6c7 d6a6 d8d1 f3d1 f8d8 d1c2');
INSERT INTO evaluations VALUES(46798,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(46798,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(46798,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(46798,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(46798,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(46798,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(46798,36,'cp 57','c2c4 e7e6 f1e2 b7b6 b1c3 c8b7 e1g1 f8g7 b2b4 e8g8 a2a4 d5c4 e2c4 c7c5 b4c5 b6c5 c1a3 d8a5 c3b5 f6e4 c4d3 b8d7 a1c1 a7a6 d4c5 a6b5 c5c6 b7c6 c1c6 e4c3 c6c3 b5b4 c3b3 b4a3 b3a3 d7e5 f3e5 g7e5');
INSERT INTO evaluations VALUES(46758,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(46758,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(46758,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(46758,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(46758,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(46758,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(46758,36,'cp 49','c7c6 f1e2 f8g7 e1g1 e8g8 b1c3 f8e8 a2a4 a7a5 c4d5 c6d5 d1b3 e7e6 c1d2 f6e4 d2e1 e4d6 c3b5 b8c6 b5d6 d8d6 e1d2 c8d7 f1c1 e8b8 h2h3 d7e8 c1c3 d6d8 e2b5 b8c8 b5c6 c8c6 b3b7');
INSERT INTO evaluations VALUES(44711,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(44711,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(44711,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(44711,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(44711,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(44711,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(44711,36,'cp 44','f1e2 f8g7 e1g1 e8g8 b1c3 f8e8 a2a4 a7a5 c4d5 c6d5 c1d2 f6e4 d2e1 e4c3 e1c3 b8c6 h2h3 e7e6 e2b5 c8d7 d1e2 d8b6 f1c1 e8d8 g2g4 d8c8 c1c2 c8c7 h3h4 a8c8 a1c1 c6e7 b5d7 c7d7 h4h5 d7c7');
INSERT INTO evaluations VALUES(48202,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(48202,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(48202,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(48202,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(48202,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(48202,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(48202,36,'cp 33','c4d5 e6d5 b1c3 f8e7 e2e3 e8g8 f1d3 h7h6 g5h4 c7c6 d1c2 f8e8 g1e2 f6h5 h4e7 d8e7 e1g1 b8d7 a1e1 d7f6 e2c1 f6g4 h2h3 g4f6 c2d1 g7g6 d1d2 c8d7 c1e2 b7b6 e2f4 e7d6 c3e2 h5f4 e2f4 a8d8 f2f3 c6c5');
INSERT INTO evaluations VALUES(48171,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(48171,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48171,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(48171,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(48171,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(48171,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(48171,36,'cp 15','e6d5 b1c3 f8e7 e2e3 h7h6 g5h4 c7c6 f1d3 e8g8 g1e2 a7a5 e1g1 b8d7 h2h3 b7b5 d1c2 c8a6 a2a3 f8e8 f1c1 e7f8 d3h7 g8h8 h7f5 d7b6 b2b3 a5a4 b3b4 a6c8 f5d3 d8e7 c1e1 c8d7 h4g3 b6c4 d3c4 d5c4 a1d1');
INSERT INTO evaluations VALUES(48627,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(48627,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(48627,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(48627,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48627,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(48627,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(48627,36,'cp 28','b1c3 c7c6 e2e3 h7h6 g5h4 f8e7 f1d3 e8g8 g1e2 a7a5 d1c2 b7b5 e1g1 b8d7 a2a3 f8e8 f1c1 c8a6 c3b1 a8c8 d3f5 c6c5 d4c5 c8c5 b1c3 d7e5 c1d1 a6b7 e2d4 b5b4 a3b4 a5b4 c2b3 b4c3 b3b7 c3b2 b7b2 f6e4 d4b3 e5c4 h4e7 d8e7 b2d4 e7h4');
INSERT INTO evaluations VALUES(57028,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(57028,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57028,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57028,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57028,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57028,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57028,36,'cp 0','d7d5 c1f4 c7c5 e2e3 g8f6 g1f3 b8c6 h2h3 f8d6 d4c5 d6f4 e3f4 d8a5 d1d2 a5c5 b1c3 e8g8 f1d3 f8d8 e1g1 c8d7 f1e1 c6e7 f3d4 e7c6 d4f3');
INSERT INTO evaluations VALUES(55873,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(55873,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(55873,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55873,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55873,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55873,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(55873,36,'cp 0','g1f3 g8f6 c2c4 d5c4 e2e3 a7a6 f1c4 c7c5 d4c5 d8d1 e1d1 f8c5 d1e2 c5e7 b1d2 e8g8 b2b4 c8d7 c1b2 d7a4 c4b3 a4b5 b3c4');
INSERT INTO evaluations VALUES(55871,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55871,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55871,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55871,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55871,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55871,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55871,36,'cp 0','c7c5 c2c4 d5c4 e2e4 b8c6 d4d5 c6a5 f1c4 a5c4 d1a4 d8d7 a4c4 e6d5 e4d5 f8d6 e1g1 g8e7 b1c3 e8g8 f1d1 b7b6 c3e4 c8b7 e4d6 d7d6 c1f4 d6f6 f4g5 f6d6');
INSERT INTO evaluations VALUES(54225,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54225,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54225,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54225,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54225,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54225,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54225,36,'cp 0','c2c4 d5c4 e2e4 b8c6 d4d5 c6a5 f1c4 a5c4 d1a4 d8d7 a4c4 e6d5 e4d5 f8d6 e1g1 g8e7 b1c3 e8g8 f1d1 b7b6 c3e4 c8b7 e4d6 d7d6 c1f4 d6f6 f4g5 f6d6');
INSERT INTO evaluations VALUES(54182,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54182,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54182,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54182,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54182,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54182,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54182,36,'cp 0','d5c4 e2e4 b8c6 d4d5 c6a5 f1c4 a5c4 d1a4 d8d7 a4c4 e6d5 e4d5 f8d6 e1g1 g8e7 b1c3 e8g8 f1d1 b7b6 c3e4 c8b7 e4d6 d7d6 c1f4 d6f6 f4g5 f6d6');
INSERT INTO evaluations VALUES(4285,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4285,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(4285,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4285,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(4285,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(4285,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4285,36,'cp 29','b1c3 h7h6 g5h4 c5e7 e1g1 d7d6 a2a4 g7g5 h4g3 c8g4 c3b1 g4d7 a4a5 a7a6 c2c3 g5g4 f3h4 g8h7 h4f5 d7f5 e4f5 d8d7 b1d2 d6d5 c4a2 d7f5 f2f3 f5g6 d3d4 e5d4 a2b1 d4d3 f3g4 f6g4 d1f3 c6e5 f3d5');
INSERT INTO evaluations VALUES(18948,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(18948,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(18948,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18948,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18948,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(18948,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(18948,36,'cp 0','d7d6 a2a4 g8f6 d2d3 f8e7 b5c4 e8g8 e1g1 c8e6 f1e1 d8d7 c1g5 c6b4 f3d2 e6g4 f2f3 g4e6 f3f4 e5f4 g5f4 a8d8 d2f3 e6g4 a1c1 g4e6');
INSERT INTO evaluations VALUES(7448,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(7448,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7448,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(7448,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(7448,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(7448,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7448,36,'cp 2','d7d6 a2a4 a8b8 a4b5 a6b5 c2c3 e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 c8g4 h2h3 g4h5 b3c2 b8a8 b1d2 c5b6 b2b4 h5g6 f1e1 a8a1 d1a1 f6h5 d2f1 d8a8 a1b2 a8a7 e1a1 a7b7 g3h2 f8a8 d3d4 a8a1 b2a1 e5d4 c3d4');
INSERT INTO evaluations VALUES(7412,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(7412,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7412,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7412,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(7412,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(7412,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(7412,36,'cp 14','c2c3 h7h6 a2a4 c8d7 d3d4 c5b6 c1e3 f6g4 h2h3 g4e3 f2e3 e8g8 g1h1 c6e7 a4a5 b6a7 d4e5 a7e3 e5d6 c7d6 b1a3 d7e6 b3e6 f7e6 a3c2 e3a7 c2b4 a7c5 d1b3 d6d5 f3d4 f8f1 a1f1 d8d6 e4e5 d6e5 b4d3 e5d6 d3c5 d6c5 d4e6 c5e3 e6d4 a8f8 f1f8 g8f8');
INSERT INTO evaluations VALUES(7420,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(7420,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(7420,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(7420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7420,36,'cp 0','h7h6 a2a4 c8d7 d3d4 c5b6 c1e3 f6g4 h2h3 g4e3 f2e3 e8g8 g1h1 c6e7 a4a5 b6a7 d4e5 a7e3 e5d6 c7d6 d1d6 d7c6 d6d8 a8d8 f1e1 e3f4 b1a3 f8e8 a3c2 f4g3 e1e3 g3f4');
INSERT INTO evaluations VALUES(7408,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7408,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7408,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7408,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7408,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7408,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7408,36,'cp 7','a2a4 c8d7 d3d4 c5b6 c1e3 f6g4 h2h3 g4e3 f2e3 e8g8 g1h1 c6e7 a4a5 b6a7 d4e5 a7e3 e5d6 c7d6 b1a3 d7e6 b3e6 f7e6 a3c2 e3c5 c2d4 d8d7 d1b3 d6d5 e4d5 e6d5 a1e1 e7f5 d4f5 f8f5 f3e5 f5f1 e1f1 d7e6 e5f3 a8f8 f1e1');
INSERT INTO evaluations VALUES(13234,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13234,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13234,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13234,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13234,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(13234,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13234,36,'cp 12','e7e6 d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 c4b3 c8d7 d1e2 a7a6 f2f4 b7b5 a2a3 d8c7 d4c6 d7c6 f4f5 d6d5 f5e6 f7e6 e4d5 f6d5 c3d5 c6d5 b3d5 e6d5 e3d4 e7c5 f1f8 a8f8 a1d1 c7c6 c2c3 f8e8 e2h5 e8e7 d4c5 c6c5 d1d4 e7e1 g1f2 e1b1 h5e8 c5f8 e8f8 g8f8 d4b4 b1c1 b4f4 f8e7 f2e3');
INSERT INTO evaluations VALUES(13926,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(13926,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(13926,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(13926,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13926,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13926,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(13926,36,'cp 17','d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 c4b3 c6a5 f2f4 e6e5 f4e5 f6g4 d1e2 g4e3 e2e3 d6e5 d4f5 c8f5 f1f5 a5b3 a2b3 d8d4 e3d4 e5d4 c3d5 e7d6 e4e5 d6c5 b3b4 f8d8 d5c7 a8c8 b4c5 c8c7 a1a7 c7c5 a7b7 d8f8 f5f2 c5e5 g1f1 e5h5 b7d7 h5h2 d7d4 f8e8 f2e2');
INSERT INTO evaluations VALUES(13954,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(13954,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(13954,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(13954,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13954,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13954,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(13954,36,'cp 16','c5d4 f3d4 f8e7 c1e3 a7a6 c4b3 c6d4 d1d4 e8g8 a2a4 c8d7 h2h3 b7b5 f2f4 b5a4 c3a4 d8c7 a4b6 a8b8 b6d7 f6d7 d4c3 d7c5 e3c5 c7c5 c3c5 d6c5 a1a6 b8b4 a6a4 b4a4 b3a4 c5c4 a4b5 f8b8 b5c4 b8b2 c4b3 e7c5 g1h1 g8f8 g2g4 g7g6 e4e5 c5e3 h1g2');
INSERT INTO evaluations VALUES(13990,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(13990,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(13990,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(13990,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(13990,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(13990,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(13990,36,'cp 16','f3d4 f8e7 c1e3 a7a6 c4b3 c6d4 d1d4 e8g8 a2a4 f6g4 e3c1 c8d7 h2h3 g4f6 c1e3 b7b5 f2f3 d8c7 a4b5 d7b5 c3b5 a6b5 f1b1 f6d7 c2c3 d7e5 d4b6 c7b6 e3b6 e5c4 b3c4 b5c4 b6e3 a8a1 b1a1 f8b8');
INSERT INTO evaluations VALUES(13984,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(13984,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(13984,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(13984,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(13984,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(13984,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13984,36,'cp 12','f8e7 c1e3 a7a6 c4b3 c6d4 d1d4 e8g8 a2a4 c8d7 h2h3 b7b5 f2f4 b5a4 c3a4 d8c7 a4b6 a8b8 b6d7 f6d7 d4c3 d7c5 e3c5 c7c5 c3c5 d6c5 a1a6 b8b4 a6a4 f8b8 f1a1 g7g5 f4f5 e6f5 e4e5 f5f4 a4a7 c5c4 a7e7 c4b3 c2c3');
INSERT INTO evaluations VALUES(8197,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(8197,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8197,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(8197,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(8197,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(8197,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(8197,36,'cp 23','c1e3 a7a6 c4b3 c6d4 d1d4 e8g8 a2a4 f6g4 e3d2 g4f6 h2h3 b7b6 d2e3 c8b7 a1d1 b7c6 f2f4 b6b5 f4f5 b5a4 b3a4 c6a4 c3a4 d8c7 c2c4 c7b7 f5e6 f7e6 a4b6 a8d8 e4e5 f6e4 f1f8 d8f8 e5d6 e7d6 b6a4 d6g3 a4c3 e4c3 d4c3');
INSERT INTO evaluations VALUES(8195,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(8195,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8195,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(8195,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(8195,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8195,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(8195,36,'cp 8','a7a6 c4b3 c6d4 d1d4 e8g8 a2a4 c8d7 h2h3 b7b5 f2f4 b5a4 c3a4 d7c6 a4c3 h7h6 f4f5 e6e5 d4d3 d8d7 e3f2 f8b8 f2h4 a6a5 h4f6 e7f6 a1b1 a5a4 b3d5 b8c8 g1h1 d7b7 f1f3 b7c7 b1d1 f6g5 d5c6 c7c6');
INSERT INTO evaluations VALUES(7083,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(7083,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(7083,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7083,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(7083,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(7083,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(7083,36,'cp 17','c4b3 c6d4 d1d4 e8g8 a2a4 c8d7 h2h3 b7b5 f2f4 d7c6 f4f5 e6e5 d4d3 b5b4 c3d5 c6d5 b3d5 f6d5 d3d5 d8d7 d5d3 f8c8 f1c1 c8c6 a4a5 a8b8 g1h2 b8b5 c2c4 b4c3 c1c3 c6c3 b2c3 e7d8 e3b6 d8b6 a5b6 b5b6 a1a6');
INSERT INTO evaluations VALUES(10896,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10896,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10896,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10896,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10896,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10896,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10896,36,'cp 0','e8g8 e1g1 d7d5 e4d5 f6d5 a2a4 a7a5 h2h3 d5b6 c4b5 f7f6 d3d4 e5d4 f1e1 e7c5 c3d4 c6d4 f3d4 d8d4 d1c2 g8h8 c1e3 d4b4 e3d2 b4d4');
INSERT INTO evaluations VALUES(4583,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4583,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4583,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4583,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4583,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4583,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4583,36,'cp 0','b1d2 d7d5 c4b3 c8e6 e1g1 a7a5 f1e1 d5e4 d3e4 d8d6 a2a4 f6d7 b3c2 f8d8 d1e2 d6c5 d2f1 e6c4 e2d1 d7f6 f3d2 c4e6 d1e2 d8d7 f1e3 a8d8 d2c4 h7h6 b2b3 c5a7 c1b2 e7f8 a1d1 d7d1 e1d1 d8d1 e2d1 f6d7 e3d5 b7b6 h2h3 d7c5');
INSERT INTO evaluations VALUES(51090,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(51090,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(51090,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(51090,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(51090,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(51090,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(51090,36,'cp 98','e2e4 e7e6 f1d3 d7d5 b1c3 d5e4 c3e4 g8f6 e4f6 d8f6 g1f3 h7h6 e1g1 f8d6 f1e1 b8d7 d1e2 f6e7 f3e5 e8g8 e2h5 d7e5 d4e5 d6c5 e1e4 f7f5 e5f6 f8f6 e4f4 b7b6 h5f3 a8b8 f4f6 e7f6 f3f6 g7f6 c1h6 g8f7 h6e3 c5e3 f2e3 f7e7');
INSERT INTO evaluations VALUES(55731,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(55731,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(55731,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(55731,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(55731,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(55731,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(55731,36,'cp 82','e2e4 d6d5 b1d2 g8f6 e4e5 f6d7 c2c3 c7c5 f1d3 b8c6 e1g1 a7a5 d3c2 c5d4 c3d4 b7b5 f1e1 h7h6 d2f1 a5a4 f1g3 d7b6 g3h5 c8d7 c1f4 c6b4 f4d2 b4c2 d1c2 b6c4 a2a3 d8b6 e1c1 b5b4 a3b4 c4d2 c2d2 b6b4 h5g7 e8d8 g7h5 b4d2 f3d2 d7b5 h2h3 a8c8');
INSERT INTO evaluations VALUES(23134,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(23134,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(23134,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(23134,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(23134,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(23134,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(23134,36,'cp 55','g8f6 e1g1 d7d5 e4d5 f6d5 c3d4 c7c6 d1b3 f8d6 b1c3 c8e6 g2g3 d8b6 c4d5 c6d5 b3b6 a7b6 d2d3 e8d7 c1f4 d6b4 c3b5 h8c8 a2a3 b4e7 f1c1 g7g5 f4d2 e6g4 f2f4 g5f4 d2f4 g4f5 c1c8 a8c8 b5c3 f5d3 c3d5 c8c2 a1c1 c2c1 f4c1 b6b5 d5f4 d3c4 g1f2 e7f6');
INSERT INTO evaluations VALUES(36525,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36525,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36525,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36525,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36525,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36525,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36525,36,'cp 1','c7c5 h2h3 d6d5 f1e2 c5d4 e3d4 b8c6 e1g1 d5c4 e2c4 c6a5 b2b3 b7b6 c4e2 c8b7 f3e5 a8c8 a1c1 e7e6 d1d2 a5c6 e2f3 b7a8 f1d1 c6e7 f3a8 c8a8 f4g5 a8c8 h3h4 c8c7 a2a4 d8a8 b3b4 e7f5 b4b5 f8c8 e5c6 a7a6 g2g3 a6b5 a4b5 c7d7 g5f4');
INSERT INTO evaluations VALUES(35940,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(35940,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(35940,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35940,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35940,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(35940,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(35940,36,'cp 19','h2h3 c5d4 e3d4 d6d5 f1e2 b8c6 e1g1 d5c4 e2c4 c6a5 b2b3 b7b6 c4e2 c8b7 f3e5 a8c8 a1c1 a5c6 e2f3 e7e6 d1d2 b7a8 f1d1 c6e7 f3a8 c8a8 f4g5 a8c8 h3h4 f6h5 e5g4 d8d7 g5h6 g7h6 d2h6 e7d5 c3d5 d7d5 h6h5 d5h5 g4f6 g8g7');
INSERT INTO evaluations VALUES(9481,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9481,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(9481,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(9481,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(9481,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(9481,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(9481,36,'cp 10','c1e3 c5d6 h2h3 c8e6 b1d2 d8e7 f3g5 e6d7 d1e2 h7h6 g5f3 d7e6 a2a4 f6d7 d3d4 e8c8 e1c1 g7g5 d4e5 d6e5 f3e5 d7e5 f2f4 g5f4 e3f4 e5g6 f4h2 e7b4 b2b3 c6c5 c1b1 h8e8 h1f1 b7b6 d2f3 d8d1 f1d1');
INSERT INTO evaluations VALUES(19479,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19479,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19479,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(19479,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19479,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19479,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19479,36,'cp 0','g1f3 g8f6 b1c3 e7e5 f1b5 f8e7 d2d3 d7d6 b5c4 e8g8 e1g1 a7a6 c3d5 f6d5 c4d5 c8e6 c2c4 a8c8 h2h3 b7b5 b2b3 d8d7 c1d2 f7f5 e4f5 f8f5 d5e4 f5f8 d2c3 e6f5 e4d5 f5e6');
INSERT INTO evaluations VALUES(19477,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(19477,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(19477,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(19477,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(19477,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(19477,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19477,36,'cp 0','g8f6 b1c3 e7e5 f1b5 d7d6 e1g1 f8e7 d2d3 a7a6 b5c4 e8g8 c3d5 f6d5 c4d5 c8e6 c2c4 d8d7 c1d2 f7f5 f1e1 f5e4 d5e4 e6f5 e4d5 f5e6');
INSERT INTO evaluations VALUES(12847,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12847,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12847,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(12847,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(12847,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(12847,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(12847,36,'cp -6','b1c3 e7e5 f1b5 d7d6 d2d3 a7a6 b5c4 b7b5 c4a2 c8e6 c1g5 f8e7 a3a4 b5b4 c3d5 e8g8 c2c3 a8b8 g5d2 b4c3 b2c3 c6a5 c3c4 a5c6 e1g1 h7h6 d2c3 d8d7 h2h3 g8h7 a1b1 b8b1 a2b1');
INSERT INTO evaluations VALUES(12845,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12845,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(12845,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(12845,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12845,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12845,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12845,36,'cp 0','e7e5 f1b5 d7d6 e1g1 f8e7 d2d3 e8g8 b5c4 a7a6 c3d5 f6d5 c4d5 c8e6 c2c4 d8d7 c1d2 b7b5 b2b4 a8c8 f1e1 h7h6 a1c1 c6d4 b4c5 d4f3 d1f3 c8c5 f3g3 g8h7 g3f3');
INSERT INTO evaluations VALUES(12570,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12570,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12570,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12570,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12570,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12570,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12570,36,'cp 0','f1b5 d7d6 e1g1 f8e7 d2d3 e8g8 b5c4 a7a6 c3d5 f6d5 c4d5 c8e6 c2c4 d8d7 c1d2 b7b5 b2b4 a8c8 f1e1 h7h6 a1c1 c6d4 b4c5 d4f3 d1f3 c8c5 f3g3 g8h7 g3f3');
INSERT INTO evaluations VALUES(38413,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38413,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(38413,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(38413,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(38413,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(38413,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(38413,36,'cp 50','f1d3 g8f6 g1f3 b8d7 c1e3 e8g8 e1g1 e7e5 b1d2 b7b6 f1e1 c8b7 e3g5 d8c8 a2a4 a7a5 g5h4 b7a6 d3b5 f6h5 h4g5 h7h6 g5e3 h5f6 d1c2 f8e8 a1c1 a6b5 a4b5 f6g4 d2f1 c8b7 e3d2 d6d5 e4d5 b7d5');
INSERT INTO evaluations VALUES(38414,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(38414,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(38414,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(38414,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(38414,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38414,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(38414,36,'cp 49','e7e5 g1f3 b8d7 c1e3 g8f6 e1g1 b7b6 b1d2 e8g8 f1e1 c8b7 h2h3 f8e8 d4d5 c7c6 c3c4 a7a5 a2a3 a5a4 d3c2 b6b5 d5c6 b7c6 c4b5 c6b5 d2b1 d7b6 b1c3 b5c6 e3b6 d8b6 c2a4 c6a4 c3a4 b6b5 b2b3 d6d5 e4d5 f6d5');
INSERT INTO evaluations VALUES(20265,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(20265,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(20265,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(20265,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(20265,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(20265,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(20265,36,'cp 8','b1d2 g8f6 c2c3 f8d6 f4g3 d8e7 f1d3 e6e5 d3b5 e5e4 g3d6 e7d6 f3e5 c8d7 b5c6 d7c6 e5c6 b7c6 d2b3 c5d4 c3d4 e8g8 e1g1 f8b8 d1e2 a7a5 f1c1 f6d7 b3c5 d7c5 c1c5 a5a4');
INSERT INTO evaluations VALUES(20266,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(20266,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20266,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(20266,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(20266,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(20266,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(20266,36,'cp 19','f8d6 f4d6 d8d6 d4c5 d6c5 c2c4 g8f6 c4d5 c5d5 f1e2 e8g8 e1g1 f8d8 d1c2 d5f5 f1c1 f5c2 c1c2 c6b4 c2c4 b4d5 a1c1 c8d7 g1f1 d5b6 c4c2 a8c8 c2c8 d8c8 c1c8 d7c8 f3e5 g8f8 d2c4 b6c4 e5c4 h7h6 a2a3 c8d7');
INSERT INTO evaluations VALUES(33675,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(33675,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(33675,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(33675,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(33675,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(33675,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(33675,36,'cp 125','g1f3 e6b6 d2d4 c8g4 e1g1 e7e6 c1e3 b6a5 h2h3 g4f3 e2f3 c7c6 c3e2 f8e7 c2c4 e8g8 a2a3 a5c7 b2b4 b8d7 d1b3 a7a5 a1d1 h7h6 f1e1 a5b4 a3b4 a8a6 e3f4 e7d6 f4c1 f8a8 g2g3 d6e7 e2c3 c7b6 c4c5 b6c7 c1f4 c7d8 b4b5 a6a3');
INSERT INTO evaluations VALUES(33672,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(33672,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(33672,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(33672,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(33672,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(33672,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(33672,36,'cp 104','e6b6 d2d4 c8g4 e1g1 e7e6 c1e3 b6a5 h2h3 g4f3 e2f3 c7c6 c3e2 f8e7 c2c4 e8g8 a2a3 b8d7 e2c3 a5c7 f1e1 f8e8 b2b4 e7f8 g2g3 a7a5 d1b3 e6e5 c4c5 e5d4 e3d4 a5b4 a3b4 d7e5 f3g2 e5d3 e1d1 a8a1 d1a1 c7d7 d4e3');
INSERT INTO evaluations VALUES(40079,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(40079,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(40079,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(40079,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(40079,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(40079,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(40079,36,'cp 37','f1d3 d5c4 d3c4 e8g8 e1g1 c6c5 e3e4 c5d4 f3d4 e6e5 d4f3 b8c6 c1e3 c8g4 d1e2 c6d4 e3d4 e5d4 f1d1 g4f3 e2f3 a8c8 c4b3 e7c5 c3d5 f6d5 b3d5 g7g6 g2g3 b7b6 a1c1 g8g7 f3b3 d8e7 f2f4 h7h5 g1h1 f8d8 h1g2 h5h4');
INSERT INTO evaluations VALUES(40083,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(40083,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(40083,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(40083,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(40083,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(40083,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(40083,36,'cp 41','d5c4 d3c4 c6c5 e1g1 b8c6 d4c5 d8d1 f1d1 e7c5 f3d2 e8g8 d2e4 c5e7 e4f6 e7f6 c3e4 f6e5 a1b1 b7b6 b2b3 c8b7 c1a3 f8d8 e4d6 e5d6 d1d6 d8d6 a3d6 c6a5 c4b5 a8c8 d6f4 c8d8 f2f3 d8d2 e3e4 d2a2');
INSERT INTO evaluations VALUES(40101,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(40101,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40101,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(40101,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(40101,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(40101,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(40101,36,'cp 35','d3c4 e8g8 e1g1 c6c5 d4c5 d8d1 f1d1 e7c5 c4e2 b7b6 f3e5 c8b7 b2b3 b8c6 e5d7 f6d7 d1d7 c6a5 c1b2 f8d8 a1d1 d8d7 d1d7 c5f8 g1f1 b7c6 d7d2 f8b4 e2f3 c6f3 g2f3 a5c6 d2d7 a8d8 d7c7 d8d2 c7c8 c6d8 c8b8');
INSERT INTO evaluations VALUES(40098,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40098,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(40098,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40098,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(40098,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40098,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(40098,36,'cp 28','e8g8 e1g1 c6c5 d4c5 d8d1 f1d1 e7c5 a2a3 a7a6 c4e2 b7b5 f3d2 f8d8 d2b3 d8d1 e2d1 c5d6 d1f3 a8a7 c1d2 b8d7 a1d1 d6f8 f3e2 d7c5 b3c5 f8c5 c3a2 a7d7 e2f3 c5e7 g1f1 e6e5 a2b4 e5e4 f3e2 c8b7 d2c3 d7d1 e2d1');
INSERT INTO evaluations VALUES(49516,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(49516,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(49516,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(49516,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(49516,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(49516,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(49516,36,'cp -56','h7h6 g5f6 d8f6 c2c3 f8c5 g1f3 d7d6 g2g3 a7a5 f1g2 b8c6 b1d2 g7g5 d1e2 c8e6 d2c4 e8c8 f3d2 c8b8 h2h4 a5a4 b2b4 c5a7 c4e3 c6e7 d2c4 f6g7 h4g5 h6g5 e1c1 h8h1 d1h1 d8h8 h1h8 g7h8 c1b1 a7e3');
INSERT INTO evaluations VALUES(49248,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(49248,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(49248,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(49248,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(49248,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(49248,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(49248,36,'cp -64','g5f6 d8f6 g1f3 f8c5 b1c3 c7c6 d1d2 a7a5 e1c1 d7d6 d3d4 e5d4 f3d4 e8g8 f2f3 a5a4 a2a3 b7b5 c3e2 f8d8 c2c3 d6d5 e2g3 c5b6 c1b1 d5e4 g3e4 f6e7 d2c1 a8a7 f1e2 a7d7 h1e1 e7e5 g2g4 b6d4 c3d4 d7d4 d1d4 d8d4 e4c5 e5f6 c1e3');
INSERT INTO evaluations VALUES(49245,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(49245,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(49245,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(49245,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(49245,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(49245,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(49245,36,'cp -51','d8f6 b1c3 c7c6 d1f3 f6e7 g2g3 g7g6 h2h4 h6h5 f1h3 f8h6 g1e2 b8a6 a2a3 b7b5 e1g1 e8g8 d3d4 d7d6 h3c8 f8c8 a1d1 a6c7 g1g2 a7a5 g3g4 b5b4 c3a4 e7h4 g4h5 h4g5 g2h1 b4a3 b2a3 c8b8 d4e5 d6e5');
INSERT INTO evaluations VALUES(33960,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(33960,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(33960,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(33960,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(33960,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(33960,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(33960,36,'cp -50','b1c3 c7c6 d1f3 f6e7 f3e3 d7d6 g2g3 b8d7 f2f4 g7g6 g1f3 b7b5 f1h3 h6h5 e1g1 f8h6 e3f2 b5b4 c3d1 e8g8 f4f5 d7f6 a2a3 b4b3 c2b3 a7a5 g1h1 c8a6 f2e2 a8b8 d1e3 d6d5 e4d5');
INSERT INTO evaluations VALUES(10140,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(10140,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10140,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10140,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10140,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10140,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(10140,36,'cp 0','d7d6 e1g1 e8g8 h2h3 h7h6 b1c3 a7a6 c1e3 c5e3 f2e3 c8e6 f3d2 c6e7 a3a4 e7g6 a4a5 d8e7 d1f3 c7c6 c4e6 f7e6 f3g3 g8h7 c3a4 a8d8 d2f3 f8f7 b2b4 d8f8 f1f2 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(9366,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9366,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9366,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9366,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9366,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9366,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9366,36,'cp 0','e1g1 e8g8 h2h3 a7a5 c1e3 c5e3 f2e3 c6e7 b1c3 c7c6 a3a4 e7g6 d3d4 c8e6 c4d3 e6d7 d1d2 d8e7 d3c4 h7h6 a1e1 d7e6 c4d3 e6d7');
INSERT INTO evaluations VALUES(9365,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9365,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9365,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9365,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9365,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9365,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(9365,36,'cp 0','e8g8 h2h3 a7a5 c1e3 c5e3 f2e3 c6e7 a3a4 e7g6 b1c3 c7c6 d3d4 d8b6 d1d2 c8e6 c4e6 f7e6 d4e5 d6e5 a1a3 a8d8 d2e2 f6d7 a3b3 b6a7 b3a3');
INSERT INTO evaluations VALUES(3846,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3846,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3846,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3846,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3846,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3846,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3846,36,'cp 0','h2h3 a7a5 c1e3 c5e3 f2e3 c6e7 b1c3 e7g6 d3d4 c7c6 a3a4 c8e6 d4d5 c6d5 e4d5 e6d7 f3d2 a8c8 d1e2 d8b6 c4b3 f6e8 d2c4 b6b4 c3a2 b4c5 a2c3');
INSERT INTO evaluations VALUES(3844,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3844,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3844,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3844,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3844,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3844,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3844,36,'cp 0','a7a5 c1e3 c5e3 f2e3 c6e7 b1c3 c7c6 a3a4 e7g6 d3d4 c8e6 d4d5 c6d5 e4d5 e6d7 f3d2 a8c8 d1e2 d8b6 c4b3 f6e8 d2c4 b6b4 c3a2 b4c5 a2c3');
INSERT INTO evaluations VALUES(52470,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52470,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52470,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(52470,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52470,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52470,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(52470,36,'cp 29','e2e4 a7a6 g1f3 e7e6 d2d4 c5d4 f3d4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 f8e7 a2a3 d8c7 f1e1 c7e5 a3a4 b5b4 c3b1 g8f6 b1d2 e5c7 d2b3 c6b7 a4a5 e8g8 c1d2 f8c8 h2h3 c7e5 c2c3 d7d5 f2f4 e5c7 e4e5 b4c3 d2c3');
INSERT INTO evaluations VALUES(53974,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(53974,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53974,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53974,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53974,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(53974,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53974,36,'cp 0','g8f6 a2a4 b8c6 e1g1 f8e7 c4b3 c6a5 b3a2 a7a6 c2c3 e8g8 c1f4 b7b5 b1d2 a5c6 h2h3 c8d7 f1e1 d8c7 f4g3 f8d8 a2b1 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(19075,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(19075,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(19075,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(19075,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(19075,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(19075,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(19075,36,'cp 40','b1c3 g8f6 f1c4 c6d4 e3d3 f8b4 c1g5 b4c3 d3c3 e8g8 g1f3 f6e4 g5d8 e4c3 f3d4 d7d5 c4d3 c3e4 d8h4 e5d4 e1c1 e4c5 d3b5 a7a6 b5f1 c5e6 f2f3 c8d7 h4f2 a8e8 f2d4 e6d4 d1d4 e8e1 d4d1 e1d1 c1d1 g7g6 d1d2 g8g7 f1e2 g7f6 h1d1 d7f5 a2a3 f8c8');
INSERT INTO evaluations VALUES(19066,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(19066,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19066,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19066,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(19066,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(19066,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(19066,36,'cp 18','f8b4 f1c4 b4c3 e3c3 g8f6 c3g3 e8g8 c1h6 f6h5 g3g5 d8g5 h6g5 h7h6 g5d2 h5f6 f2f3 d7d5 e4d5 c6d4 a1c1 f8d8 g1e2 d4e2 e1e2 f6d5 h1d1 c8e6 e2f2 f7f6 c4f1 d5e7 c2c4 g8f7 d2e1 e7c6 e1c3 a8c8 g2g3 c6d4 c3d4 d8d4 d1d4 e5d4 f1d3');
INSERT INTO evaluations VALUES(37862,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(37862,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(37862,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(37862,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(37862,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(37862,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(37862,36,'cp 14','g8f6 b1c3 e8g8 e4e5 f6e8 h2h4 d7d6 e5d6 d8d6 d1d2 h7h5 e1c1 b8c6 f1c4 e8f6 f2f3 c6a5 c4e2 f8d8 d4b5 d6d2 d1d2 d8d2 e3d2 a5c6 b5c7 a8b8 d2g5 c8f5 c3d5 b8d8 h1d1 f6d5 d1d5 d8d7 d5d7 f5d7 c7d5 g8f8 c2c3');
INSERT INTO evaluations VALUES(40029,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(40029,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40029,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40029,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40029,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(40029,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40029,36,'cp 19','b1c3 e8g8 e4e5 f6e8 h2h4 d7d6 e5d6 e8d6 d1d2 b8c6 e1c1 c8d7 h4h5 d6f5 d4f5 d7f5 h5g6 d8d2 d1d2 h7g6 c3d5 e7e6 d5c3 a8d8 f1e2 d8d2 e3d2 e6e5 e2c4 c6d4 c3d5 f8e8 d2g5 g8f8 c2c3 e8c8 d5e3 d4e6 e3f5 g6f5 c4e6 f7e6');
INSERT INTO evaluations VALUES(40027,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(40027,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(40027,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40027,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40027,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(40027,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(40027,36,'cp 21','e8g8 e4e5 f6e8 h2h4 d7d6 e5d6 b8c6 d1d2 e8d6 e1c1 c8d7 h4h5 d6f5 d4f5 d7f5 h5g6 f5g6 f1d3 d8a5 e3h6 g7c3 d2c3 a5c3 b2c3 f8d8 d3e2 f7f6 h6d2 g8g7 d2e3 a8c8 d1d8 c8d8 g2g4 a7a6 h1h3 d8c8 e2c4 c6e5 e3h6 g7h8 c4d5 e5g4 d5b7');
INSERT INTO evaluations VALUES(35718,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(35718,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(35718,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(35718,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(35718,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(35718,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(35718,36,'cp 15','e4e5 f6e8 h2h4 d7d6 e5d6 b8c6 d1d2 e8d6 e1c1 c8d7 h4h5 d6f5 d4f5 d7f5 h5g6 f5g6 f1d3 d8a5 e3h6 g7c3 d2c3 a5c3 b2c3 f8d8 d3e2 d8d1 h1d1 a8d8 d1d8 c6d8 e2f3 f7f6 h6e3 a7a6 c3c4 g8f7 e3b6 d8c6 b6c5 f7e6 c1d2 e6e5 g2g3 h7h5');
INSERT INTO evaluations VALUES(38431,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(38431,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(38431,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(38431,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(38431,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(38431,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(38431,36,'cp 39','b1c3 g8f6 g1f3 e8g8 f1d3 b8c6 e1g1 e7e5 d4d5 c6e7 d1e1 e5f4 c1f4 h7h6 a1d1 g6g5 f4e3 e7g6 c3e2 c7c6 d5c6 b7c6 h2h3 f6h5 e1d2 c6c5 b2b4 d8e7 g2g4 h5f6 e2g3 f8e8 b4c5 d6c5 e4e5 g6e5 f3e5');
INSERT INTO evaluations VALUES(38428,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(38428,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38428,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(38428,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(38428,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(38428,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(38428,36,'cp 57','g8f6 g1f3 e8g8 f1d3 b8c6 e1g1 e7e5 d4d5 c6e7 d1e1 e5f4 c1f4 h7h6 e1d2 g6g5 f4e3 e7g6 a1e1 f8e8 h2h3 c7c6 a2a4 g5g4 h3g4 f6g4 e3d4 g6e5 c3e2 c6d5 e4d5 e5f3 f1f3 d8g5 d2c3 g7d4 c3d4 g5e5 d4e5 g4e5 f3g3 g8f8');
INSERT INTO evaluations VALUES(45860,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(45860,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(45860,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(45860,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(45860,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(45860,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(45860,36,'cp 75','d2d4 d8c7 a2a4 f8e7 e1g1 e8g8 h2h3 b8d7 c4a2 h7h6 f1e1 f8e8 c1e3 d7f8 d4d5 c8d7 f3d2 f8g6 a2b3 e7d8 d1f3 c6d5 c3d5 f6d5 b3d5 d7e6 c2c4 d8g5 d2f1 g6f4 e1c1 c7e7 f1g3 a8c8 b2b4 a7a6 b4b5 a6b5 d5e6 f4e6 c4b5 c8c1 a1c1 g5e3');
INSERT INTO evaluations VALUES(45865,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(45865,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(45865,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(45865,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(45865,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(45865,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(45865,36,'cp 56','d8c7 a2a4 f8e7 e1g1 e8g8 h2h3 b8d7 c4a2 h7h6 f1e1 f8e8 c1e3 d7f8 d4d5 c8d7 f3d2 f8g6 a4a5 b7b5 a5b6 a7b6 d1e2 b6b5 d5c6 c7c6 c3d5 d7e6 d5f6 e7f6 a2e6 f7e6 e2g4 g8h7 a1a8 c6a8 c2c4 b5c4 d2c4 d6d5');
INSERT INTO evaluations VALUES(13528,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(13528,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13528,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13528,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(13528,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(13528,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(13528,36,'cp -9','c4b5 c8d7 e4d5 e6d5 e1g1 f8e7 c1g5 e8g8 f1e1 f8e8 d3d4 c6d4 f3d4 c5d4 b5d7 d8d7 d1d4 d7c6 g5e3 e7d6 a1d1 d6e5 d4d3 e5c3 d3c3 c6c3 b2c3 e8e6 e3d4 a8c8 g1f1 f6e4 d4a7 e4c3 d1d3 e6e1 f1e1 c3a2 d3d5 a2b4 d5b5 b4c2 e1d2 c2a3 b5b7');
INSERT INTO evaluations VALUES(13488,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13488,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(13488,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(13488,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(13488,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(13488,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(13488,36,'cp -12','c8d7 e4d5 e6d5 c1g5 f8e7 e1g1 e8g8 f1e1 d7g4 b5c6 b7c6 h2h3 g4h5 e1e5 f8e8 g5f6 h5f3 d1f3 e7f6 e5e8 d8e8 c3a4 a8b8 a1b1 e8e7 f3e3 e7e3 f2e3 f6e7 e3e4 g8f8 e4d5 c6d5 g1f1 g7g6 b2b3 b8d8 g2g4 e7d6 c2c4 d5d4 b1e1 h7h6 e1e2 d8c8 f1g2 c8e8');
INSERT INTO evaluations VALUES(43240,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(43240,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(43240,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(43240,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(43240,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(43240,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(43240,36,'cp -12','d7d5 b2b4 e8g8 e1g1 b8d7 a2a4 f8e8 d3e2 e7e5 d4e5 d7e5 f3e5 e8e5 c1b2 e5e8 c2c4 c8e6 c4d5 f6d5 b2g7 g8g7 d1d4 d8f6 d4f6 g7f6 b4b5 e8c8 b1d2 f6e7 f1c1 c7c6 e2c4 c6b5 a4b5 c8c7 h2h3 d5b6 c4e6 c7c1 a1c1 e7e6 c1c7');
INSERT INTO evaluations VALUES(41500,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(41500,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(41500,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(41500,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(41500,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(41500,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41500,36,'cp 0','b2b4 e8g8 e1g1 b8d7 a2a4 f8e8 d3e2 e7e5 d4e5 d7e5 f3e5 e8e5 c1b2 e5e8 c2c4 c8e6 b1d2 d5c4 d2c4 a7a6 b2d4 d8e7 d4c5 e7d7 c5d4');
INSERT INTO evaluations VALUES(55832,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(55832,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(55832,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(55832,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(55832,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(55832,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(55832,36,'cp -27','e2e3 g8f6 c2c3 c7c5 f1d3 f8e7 g1f3 e8g8 e1g1 b7b6 d1e2 c8b7 b2b3 b8d7 f3e5 f6e4 c1b2 d7f6 b1d2 e4d2 e2d2 a8c8 d2e2 c5d4 c3d4 f6e4 a1c1 f7f6 e5g4 d8d7 g4f2 c8c1 f1c1 f8c8 f2e4 c8c1 b2c1 d5e4 d3b5 b7c6 b5c6 d7c6 e2c4 c6d5 c1b2');
INSERT INTO evaluations VALUES(55830,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(55830,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(55830,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(55830,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(55830,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(55830,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(55830,36,'cp -39','c7c5 c2c3 f8d6 g1f3 g8f6 f1d3 e8g8 e1g1 b7b6 c1d2 c8a6 d3a6 b8a6 d2e1 a8c8 d1e2 a6c7 e1g3 c7e8 b1d2 d6e7 e3e4 f6e4 d2e4 d5e4 e2e4 c5d4 f3d4 e7c5 a1d1 e8d6 e4e2 h7h6 g3f2 d8f6 g2g3 f8d8 g1h1 d6e8');
INSERT INTO evaluations VALUES(54199,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(54199,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(54199,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(54199,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(54199,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(54199,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(54199,36,'cp -32','c2c3 g8f6 g1f3 f8e7 f1d3 b7b6 d1e2 e8g8 e1g1 c8b7 f3e5 f6e4 b2b3 c5d4 c3d4 a7a5 c1b2 f7f6 e5f3 b7a6 f1c1 d8d7 b1c3 a6d3 e2d3 d7b7 c1c2 b8a6 a2a3 e4d6 c3b5 f8c8 c2c8 a8c8 b5d6 e7d6 a1c1 c8c1 b2c1');
INSERT INTO evaluations VALUES(54197,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(54197,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(54197,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(54197,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(54197,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(54197,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(54197,36,'cp -34','g8f6 g1f3 f8e7 f1d3 b7b6 e1g1 e8g8 d1e2 f6e4 b2b3 c8b7 c1b2 b8d7 b1d2 a8c8 f3e5 e4d2 e2d2 e7d6 d2e2 d7f6 d4c5 d6c5 a1d1 d8e7 h2h3 f8d8 g1h2 f6e4 d3e4 d5e4 c3c4 f7f6 e5g4 c5b4 d1d8 c8d8 f1d1 d8d1 e2d1');
INSERT INTO evaluations VALUES(46217,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(46217,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(46217,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(46217,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(46217,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(46217,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(46217,36,'cp -31','f1d3 b7b6 g1f3 f8e7 e1g1 e8g8 d1e2 f6e4 b2b3 b8d7 c1b2 c8b7 b1d2 a8c8 f3e5 e4d2 e2d2 e7d6 e5d7 d8d7 a2a4 c8c7 d2e2 f8c8 h2h3 g7g6 c3c4 c5d4 b2d4 d7e7 a1d1 d6c5 d4e5 c7d7 c4d5 b7d5 d3a6 c8a8 e5a1 d7d8 a6c4 d5c4 b3c4');
INSERT INTO evaluations VALUES(46219,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(46219,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(46219,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(46219,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(46219,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(46219,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(46219,36,'cp -34','f8e7 g1f3 e8g8 e1g1 f6e4 c3c4 c5d4 e3d4 e4f6 b1c3 b8c6 c4c5 b7b6 c5b6 d8b6 c3a4 b6b7 c1e3 c8d7 f3e5 f6e4 d3e4 d5e4 a4c5 e7c5 d4c5 c6e5 f4e5 d7c6 d1b3 b7b3 a2b3 h7h6 a1a3 a8b8 f1c1 a7a5 e3d2 f8d8 d2a5 d8d3 a5e1 d3b3');
INSERT INTO evaluations VALUES(56747,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(56747,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(56747,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(56747,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(56747,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(56747,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(56747,36,'cp 145','d2d4 f5e4 c3e4 g8f6 c1g5 f8e7 g5f6 g7f6 e4g3 e8g8 f1d3 d6d5 d3f5 b8c6 d1d2 c8f5 g3f5 d8d7 g2g4 e7b4 c2c3 e5e4 f3h4 b4d6 e1c1 a8e8 c1b1 c6e7 d2h6 e7f5 h4f5 g8h8 h1f1 f8f7 h6e3 a7a5 f2f3 b7b5 h2h3 b5b4 c3b4 a5b4 e3b3 d7e6 f3e4 e6e4 b3c2 e4c2 b1c2');
INSERT INTO evaluations VALUES(56744,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(56744,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(56744,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(56744,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(56744,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(56744,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(56744,36,'cp 150','f5e4 c3e4 g8f6 c1g5 f8e7 g5f6 g7f6 e4g3 e8g8 f1d3 d6d5 d4e5 f6e5 f3e5 e7d6 f2f4 d8h4 e1g1 b8c6 c2c3 g8h8 d3c2 d6c5 g1h1 c8e6 d1e2 a8e8 a1e1 c5d6 e5d3 h4h6 c2b3 c6a5 e2f3 c7c6 f3f2 c6c5 b3d1 e6d7 e1e8 d7e8 d1f3 a5c4 b2b3');
INSERT INTO evaluations VALUES(56725,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(56725,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(56725,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(56725,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(56725,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(56725,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(56725,36,'cp 133','c3e4 g8f6 c1g5 f8e7 g5f6 g7f6 e4g3 e8g8 f1d3 d6d5 d4e5 f6e5 f3e5 e7d6 f2f4 d8h4 e1g1 b8c6 d1d2 c8e6 a1e1 c6e5 f4e5 f8f1 e1f1 d6e5 d2e2 e5d4 g1h1 e6g4 d3h7 g8h8 e2d3 d4e5 h7g6 e5g3 d3c3 h8g8 g6f7 g8h7 c3g3 h4g3 h2g3 c7c6 f1f4 g4d1 c2c4 d5c4 f7c4 b7b5 c4d3 h7g7 f4d4 a8c8 h1g1 g7f6');
INSERT INTO evaluations VALUES(56724,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(56724,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(56724,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(56724,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(56724,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(56724,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(56724,36,'cp 150','g8f6 c1g5 f8e7 g5f6 g7f6 e4g3 e8g8 f1d3 f6f5 d4e5 b8c6 e5d6 d8d6 e1g1 f5f4 d3c4 g8h8 d1d6 e7d6 g3e4 c6e5 f3e5 d6e5 c2c3 c8f5 f1e1 b7b6 f2f3 h8g7 a2a4 a7a5 e4f2 e5d6 c4d3 f5d7 g1f1 a8d8 d3b5 d7f5 a1d1 f5g6 b5d3 g6f7 e1e4 f7g6');
INSERT INTO evaluations VALUES(10176,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(10176,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(10176,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10176,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(10176,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(10176,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(10176,36,'cp 13','e8g8 a2a3 c5b6 g1e2 c6e7 d2d4 d7d6 c3a4 c7c6 e3e4 b6d4 e2d4 e5d4 d1d4 d6d5 c1g5 d5e4 e1c1 d8a5 g5f6 g7f6 a4c3 f6f5 g3g4 e7g6 h1e1 c8e6 g4f5 a8d8 d4e3 d8d1 e1d1 a5f5 g2e4 f5f4 e3f4 g6f4');
INSERT INTO evaluations VALUES(4367,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4367,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(4367,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(4367,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(4367,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(4367,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(4367,36,'cp 39','a2a3 d7d6 g1e2 f8e8 e1g1 c5b6 b2b3 c8f5 d2d4 d8d7 d4d5 c6e7 e3e4 f5h3 d1d3 h7h6 c1d2 a7a6 g1h1 c7c6 f2f3 c6d5 c4d5 b6c5 c3d1 a8c8 d1f2 c5f2 f1f2 e7g6 g2h3 d7h3 e2g1');
INSERT INTO evaluations VALUES(25510,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(25510,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(25510,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(25510,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(25510,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(25510,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(25510,36,'cp 55','f8e7 d2d4 e8g8 b5a4 a7a5 h2h3 f8e8 d4d5 c6b8 a4d7 b8d7 c3c4 c7c6 b1c3 d7c5 c1g5 a5a4 a1b1 h7h6 g5f6 e7f6 b2b3 a4b3 a2b3 c5a6 d1c2 f6e7 e1d1 d8a5 g2g3 a6b4 c2e2 b4a6 d1c1 a5b4 c3a2 b4a5 g1g2 a6c5 a2c3');
INSERT INTO evaluations VALUES(24201,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(24201,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(24201,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(24201,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(24201,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(24201,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(24201,36,'cp 67','d2d4 e8g8 b5a4 a7a5 h2h3 f8e8 d4d5 c6b8 a4d7 b8d7 c3c4 c7c6 b1c3 d7c5 a2a4 e7f8 a1b1 a8c8 b2b3 g7g6 g2g3 d8d7 c1g5 f8g7 g1g2 c6d5 c4d5 f6h5 c3b5 f7f5 e4f5 g6f5 f3d2 h5f6 g5f6 g7f6 d2c4 g8h8');
INSERT INTO evaluations VALUES(24193,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(24193,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(24193,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(24193,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(24193,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(24193,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(24193,36,'cp 50','e8g8 b5a4 a7a5 h2h3 f8e8 d4d5 c6b8 a4d7 b8d7 c3c4 c7c6 b1c3 d7c5 a2a4 e8f8 b2b3 g8h8 a1b1 f6d7 d1c2 g7g6 c1h6 f8g8 c3e2 c6d5 c4d5 f7f5 e4f5 g6f5 c2f5 g8g6 h6e3 d8g8 e2g3 a8f8 f5c2 g8d5 e1d1 d5e6 b3b4 a5b4 b1b4 d6d5');
INSERT INTO evaluations VALUES(23751,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(23751,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(23751,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(23751,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(23751,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(23751,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(23751,36,'cp 75','b5a4 a7a5 h2h3 f8e8 d4d5 c6b8 a4d7 b8d7 c3c4 d7c5 b1c3 c7c6 a2a4 g8h8 a1b1 e8f8 b2b3 f6d7 d1c2 f7f6 g2g4 g7g6 g1h2 f6f5 e4f5 g6f5 g4f5 f8g8 e1g1 g8g1 f3g1 d7f6 c1e3 c6d5 c3d5 f6d5 c4d5 a8c8 g1f3 d8d7 e3c5');
INSERT INTO evaluations VALUES(14614,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(14614,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(14614,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(14614,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(14614,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14614,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14614,36,'cp 1','c6a5 c4b3 c7c6 c3e2 f8e7 e2g3 e8g8 c1d2 a5b3 a2b3 c6c5 c2c4 a7a5 d2c3 c8e6 f3d2 b7b6 h2h3 f6d7 g3f5 f8e8 d1f3 d7f6 f3e2 e6f5 e4f5 d8d7 e2f3 f6h7 f1e1 a8d8 a1d1 e7g5 d2f1 h7f6 f1g3 g5f4 g3h5 f6h5 f3h5');
INSERT INTO evaluations VALUES(14680,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(14680,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14680,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(14680,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(14680,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(14680,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(14680,36,'cp 9','c4d5 c7c6 d5f7 e8f7 b2b4 f8e7 c1d2 h8f8 b4a5 f7g8 h2h3 d8a5 c3d5 a5d8 d5e7 d8e7 a2a4 f6h5 g1h2 h5f4 f3g1 d6d5 f2f3 c8e6 d1e1 f4g6 g1e2 f8f7 e1g3 g8h7 a4a5 a8d8 d2e3 c6c5 e4d5 e6d5 e2c3 g6f4 c3e4 b7b6 a5b6 a7b6 f1b1 d5e4');
INSERT INTO evaluations VALUES(54975,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(54975,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(54975,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(54975,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(54975,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54975,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(54975,36,'cp 0','c7c5 g2g3 g7g6 g1e2 g8f6 c2c3 f8g7 f1g2 e8g8 e1g1 e7e5 c3c4 b8c6 b1c3 a8b8 c1e3 a7a6 a2a4 c8d7 a1b1 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(38341,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(38341,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(38341,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(38341,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(38341,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(38341,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(38341,36,'cp -19','d6e5 f3e5 g7e5 d4e5 d8d1 e1d1 b8c6 c1f4 c8f5 b1a3 e8c8 d1c1 e7e6 f2f3 h7h5 f4g5 d8d4 a3c4 d4d5 f3f4 d5d7 c4e3 f5e4 f1b5 g8e7 g5e7 d7e7 h1d1 e7d7 d1d7 c8d7 c1d2 d7c8 a1d1 c6e7 b5e2 h8d8 d2c1');
INSERT INTO evaluations VALUES(38489,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(38489,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(38489,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(38489,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(38489,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(38489,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(38489,36,'cp -16','f3e5 b8d7 e5d7 d8d7 c2c3 e7e5 d1e2 g8e7 d4e5 e8g8 b1d2 e7d5 d2f3 f8e8 e2e4 d5f6 e4h4 f6g4 f1e2 g7e5 e1g1 e5f6 c1g5 e8e2 g5f6 g4f6 h4f6 d7e7 f6e7 e2e7 f1e1 e7e1 a1e1 g8f8 b2b3 c8d7 f3e5 d7e6 f2f3 f8e7 g1f2 f7f6 e5c4 a8d8');
INSERT INTO evaluations VALUES(38484,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(38484,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(38484,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(38484,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(38484,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(38484,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(38484,36,'cp -24','b8d7 e5d7 c8d7 f1c4 g8h6 e1g1 e8g8 f1e1 b7b5 c4b3 a7a5 c2c3 a5a4 b3c2 d7g4 f2f3 g4e6 c2e4 e6d5 c1g5 d5e4 e1e4 f8e8 b1a3 b5b4 g5h6 b4a3 h6g7 a3b2 a1b1 g8g7 b1b2 d8d6 d1e2 e7e6 e2c4 a4a3');
INSERT INTO evaluations VALUES(39569,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(39569,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(39569,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(39569,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(39569,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(39569,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(39569,36,'cp 13','d2d4 c7c6 e2e4 d7d5 e4e5 d5c4 f1c4 c6c5 d1b3 d8d4 c4f7 e8f8 b1c3 d4e5 g1e2 b8c6 c1f4 e5f5 f7d5 g8f6 e1c1 f6d5 d1d5 f5e6 f4e3 c6d4 e2d4 c5d4 e3d4 g7d4 d5d4 e6b3 a2b3 c8h3 h1e1 h7h5 c3d5 a8e8 d4b4 b7b6 b4f4 h3f5 d5e3 e8c8 c1d2');
INSERT INTO evaluations VALUES(39575,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(39575,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(39575,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(39575,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(39575,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(39575,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(39575,36,'cp 12','c7c6 f1g2 d7d5 c4d5 c6d5 g1f3 g8f6 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 a8c8 e1g1 e8g8 f1d1 d8e7 a1c1 f8d8 c1c2 c8c6 d1c1 d8c8 a2a3 e7d8 h2h3 h7h6 c3a2 c6c2 c1c2 c8c2 d2c2 a7a5 e2e3 d7c6');
INSERT INTO evaluations VALUES(4340,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(4340,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4340,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4340,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(4340,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(4340,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4340,36,'cp 16','e1g1 d7d6 c2c3 h7h6 f1e1 a7a5 c4b3 f8e8 b1d2 c8e6 b3a4 d8b8 d2f1 b7b5 a4c2 b8b6 c1e3 c5e3 f1e3 a8d8 d3d4 d6d5 e3d5 e6d5 e4d5 d8d5 d1e2 a5a4 a1d1 g8f8 f3e5 c6e5 d4e5 d5e5 e2e5 e8e5 e1e5 g7g6 c2e4 b5b4 c3b4 b6b4 e4f3 b4b2');
INSERT INTO evaluations VALUES(4339,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4339,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(4339,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(4339,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(4339,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4339,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4339,36,'cp 0','d7d6 c2c3 h7h6 f1e1 a7a6 b1d2 c8e6 b2b4 c5a7 c4b3 f8e8 a2a4 e6b3 d2b3 c6e7 c1e3 a7e3 e1e3 e7g6 d1c2 d6d5 e3e1 g6f4 e1e3');
INSERT INTO evaluations VALUES(3833,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(3833,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(3833,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(3833,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(3833,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3833,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3833,36,'cp 13','c2c3 h7h6 b2b4 c5b6 a2a4 a7a5 b4b5 c6e7 b1d2 e7g6 f1e1 f8e8 d1b3 c8e6 c4e6 e8e6 d3d4 e6e8 b3c2 a8c8 c1b2 c7c6 b5c6 c8c6 c3c4 f6d7 a1d1 d8c8 c2b3 e5d4 f3d4 b6d4 b2d4 d7c5 b3c2 g6e5 d4e3');
INSERT INTO evaluations VALUES(3826,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(3826,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(3826,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(3826,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(3826,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(3826,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(3826,36,'cp 6','a7a5 f1e1 c5b6 b1d2 c8e6 a2a4 h7h6 c4b3 f8e8 d2c4 b6c5 c1e3 c5e3 c4e3 d8d7 d1c2 e6b3 c2b3 c6e7 f3h4 d7c6 e3g4 f6g4 h3g4 a8d8 d3d4 c6d7 h4f5 e5d4 c3d4 d6d5 b3b7 d5e4 b7e4 e7f5 e4f5 e8e1 a1e1 d7f5 g4f5 d8d4');
INSERT INTO evaluations VALUES(4982,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(4982,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(4982,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4982,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4982,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4982,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4982,36,'cp 0','f3e5 d8g5 e5g4 g5g6 e1g1 d7d6 g4e3 g6e4 b1c3 e4h4 g2g3 h4d8 d2d4 c5b6 a2a4 a7a5 e3c4 g8e7 c4b6 c7b6 d4d5 e8g8 d5c6 e7c6 c1f4 c8b7 f1e1 d8d7 c3d5 a8e8 e1e8 f8e8 d5b6 d7h3 d1f3 e8b8 b6d5 c6d4 f3e4 b7d5');
INSERT INTO evaluations VALUES(4979,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4979,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4979,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4979,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4979,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4979,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4979,36,'cp 0','d8g5 e5g4 g5g6 e1g1 d7d6 g4e3 g6e4 b1c3 e4h4 g2g3 h4d8 d2d4 c5b6 a2a4 a7a5 e3c4 g8e7 c4b6 c7b6 d4d5 e8g8 d5c6 e7c6 c1f4 c8b7 d1d6 c6d4 f1d1 d4f3 g1g2 f3d2 g2g1');
INSERT INTO evaluations VALUES(12524,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(12524,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(12524,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(12524,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(12524,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(12524,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(12524,36,'cp 182','f3g5 d7d5 e4d5 c6a5 d2d3 a5c4 d3c4 h7h6 g5e4 f6e4 c3e4 d8h4 d1e2 f7f5 e4c3 f8d6 g2g3 h4f6 e2h5 f6f7 h5f7 e8f7 h2h4 c8d7 e1e2 d6e7 h4h5 a8g8 b2b3 g7g5 h5g6 g8g6 h1h2 a7a6 c1b2 e7f6 a2a4 h6h5 a1h1 h5h4 g3h4 h8h5 f2f3 e5e4 f3e4 f6e5 h2f2 e5d4');
INSERT INTO evaluations VALUES(12520,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(12520,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(12520,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(12520,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(12520,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(12520,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(12520,36,'cp 179','d7d5 e4d5 c6a5 d2d3 a5c4 d3c4 h7h6 g5e4 f6e4 c3e4 d8h4 d1e2 f7f5 e4c3 f8d6 g2g3 h4f6 e2h5 f6f7 h5f7 e8f7 e1e2 c8d7 h2h4 b7b6 a2a4 a7a5 c1d2 g7g6 a1e1 a8g8 b2b3 d6e7 f2f3 g6g5 e2f2 e7d6 g3g4 g5h4 g4f5 d7f5 c3b5 d6b8 h1h4 f5c2 d2h6 c2b3');
INSERT INTO evaluations VALUES(13782,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(13782,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(13782,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(13782,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(13782,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(13782,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(13782,36,'cp 191','e4d5 c6a5 d2d3 a5c4 d3c4 h7h6 g5e4 f6e4 c3e4 d8h4 d1e2 f7f5 e4c3 f8d6 g2g3 h4e7 e2h5 e7f7 h5f7 e8f7 h2h4 c8d7 e1e2 a8g8 c1d2 g7g6 a2a4 a7a5 h1h2 b7b6 a1h1 g8e8 f2f3 d6f8 b2b3 f8g7 e2d3 g7f6 d2e3 f7g7 d3d2 g7f7 c3d1 f6e7 d1f2 e7d6 d2e2');
INSERT INTO evaluations VALUES(13728,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(13728,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(13728,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(13728,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(13728,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(13728,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(13728,36,'cp 197','c6a5 d2d3 a5c4 d3c4 h7h6 g5e4 f6e4 c3e4 d8h4 d1e2 f7f5 e4c3 f8d6 g2g3 h4f6 e2h5 f6f7 h5f7 e8f7 h2h4 c8d7 e1e2 a8g8 b2b3 g7g6 a2a4 a7a5 c1d2 b7b6 h1h2 g8e8 a1h1 d6f8 e2d1 f8d6 d1c1 d6c7 f2f3 e5e4 f3e4 c7g3');
INSERT INTO evaluations VALUES(55613,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55613,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55613,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55613,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55613,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55613,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55613,36,'cp 2','g8f6 d2d3 c7c5 f1g2 b8c6 e1g1 g7g6 a2a3 f8g7 c2c4 e8g8 b1c3 a7a6 c3d5 b7b5 c1g5 b5c4 d1a4 c8d7 d3c4 h7h6 d5f6 g7f6 g5h6 f6g7 h6g5 d8c7 a4d1 a6a5 b2b3 a5a4 b3b4 c5b4 a3b4 c6b4 g5e7 c7c4 e7d6');
INSERT INTO evaluations VALUES(47897,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(47897,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(47897,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(47897,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47897,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47897,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47897,36,'cp 19','d2d3 g7g6 f1g2 f8g7 c2c4 c7c5 b1c3 e8g8 a2a3 b8c6 a1b1 a7a5 e1g1 c8g4 h2h3 g4d7 g1h2 a8b8 c3b5 f6e8 c1g5 g7f6 g5d2 f6g7 b2b3 b7b6 d1e2 h7h6 b3b4 a5b4 a3b4 b8a8 b4c5 b6c5 b5c3 e8c7');
INSERT INTO evaluations VALUES(47888,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47888,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(47888,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47888,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(47888,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47888,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47888,36,'cp 5','c7c5 f1g2 b8c6 e1g1 g7g6 c2c4 f8g7 b1c3 e8g8 a2a3 a7a5 a1b1 a8b8 c1g5 h7h6 g5d2 c8g4 h2h3 g4d7 c3b5 b7b6 f3e1 c6d4 a3a4 f6h7 e1c2 d4c2 d1c2 h7g5 h3h4 g5e6 c2d1 e6c7 h4h5 g8h8 b5c3 h8h7 b2b3 c7e6 g2h3 e6d4 h3d7');
INSERT INTO evaluations VALUES(41245,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41245,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41245,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(41245,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(41245,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(41245,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41245,36,'cp 36','c4b3 e8g8 g1f3 c8g4 h2h3 g4f3 d1f3 b8c6 f3d1 c6a5 e1g1 a5b3 a2b3 c7c6 f2f4 a7a5 c1e3 e7e6 d1f3 f6d7 e4e5 d6d5 e3f2 f7f6 f3g3 f6e5 f4e5 c6c5 a1e1 c5d4 c3b5 d8b6 b5d4 a5a4 b3a4');
INSERT INTO evaluations VALUES(43824,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(43824,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(43824,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(43824,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(43824,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(43824,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(43824,36,'cp 137','e2e4 d7d6 f1d3 e6e5 g1f3 b8d7 e1g1 f8e7 d1e2 c8b7 h2h3 e8g8 c1e3 e5d4 f3d4 f8e8 a1d1 e7f8 f2f3 g7g6 e3g5 c7c6 e2f2 a7a6 f3f4 f8g7 f4f5 b6b5 d4f3 d8c7 g5f4 b5b4 c3e2 d7e5 f3e5 d6e5');
INSERT INTO evaluations VALUES(43832,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(43832,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(43832,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(43832,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(43832,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(43832,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(43832,36,'cp 126','d7d6 f1d3 e6e5 g1f3 b8d7 e1g1 f8e7 d1e2 e8g8 f1d1 c7c6 h2h3 c8b7 d3c2 d8c7 c1e3 e5d4 f3d4 g7g6 e3h6 f8e8 e2d2 d7e5 b2b3 a7a6 a1c1 b6b5 d4f5 g6f5 d2g5 e5g6 e4f5 f6d7 g5e3 d7c5 f5g6 h7g6 e3f3 c5e6 c3e4 c6c5 c4b5 a6b5');
INSERT INTO evaluations VALUES(43783,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(43783,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(43783,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(43783,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(43783,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(43783,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(43783,36,'cp 137','f1d3 e6e5 g1f3 b8d7 e1g1 f8e7 c1e3 e8g8 h2h3 f8e8 d1c2 a7a6 f1d1 e5d4 f3d4 c8b7 d4f5 e7f8 a1c1 a6a5 f2f3 b7c8 d3f1 d7c5 f5d4 c8b7 c2f2 g7g6 d1d2 f6d7 d4b5 d7e5 c3d5 c7c6 f3f4 c6d5 f4e5');
INSERT INTO evaluations VALUES(46565,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(46565,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(46565,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(46565,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(46565,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(46565,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46565,36,'cp 50','g1f3 c7c5 c2c3 b8c6 b1a3 c8d7 a3c2 d8a5 c1d2 a5b6 d4c5 b6c7 c2d4 c6e5 d4b5 d7b5 f1b5 e7c6 d2f4 e5d3 d1d3 c7f4 d3d4 f4d4 c3d4 f7f6 h2h4 e8f7 a2a3 e6e5 d4e5 f8c5 a1d1 a8d8 e5f6 g7f6 e1f1 h7h5 h1h3 c5b6 b5e2 a7a5 f3e1 c6d4 e2d3 h8e8 e1c2 d4c2 d3c2 d8c8 c2d3 a5a4');
INSERT INTO evaluations VALUES(46564,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(46564,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(46564,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(46564,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(46564,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(46564,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(46564,36,'cp 33','c7c5 c2c3 b8c6 b1a3 c8d7 a3c2 d8a5 c1d2 a5b6 d4c5 b6c7 c2d4 c6e5 d4b5 d7b5 f1b5 e7c6 d2f4 e5d3 d1d3 c7f4 d3d4 f4d4 c3d4 f7f6 h2h4 e8f7 a2a3 f8e7 b2b4 h8b8 e1d2 h7h5 g2g3 a7a5 d2c3 e7d8 h1e1 d8c7 c3b3 b8d8 a1c1 f7e7 b5c6 b7c6 a3a4 e7f7');
INSERT INTO evaluations VALUES(44687,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44687,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(44687,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44687,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(44687,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44687,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(44687,36,'cp 38','c2c3 b8c6 b1a3 c8d7 a3c2 d8a5 c1d2 a5b6 d4c5 b6c7 c2d4 c6e5 d4b5 d7b5 f1b5 e7c6 d2f4 e5d3 d1d3 c7f4 g2g3 f4c7 d3d4 a7a6 b5a4 e8c8 a4c6 c7c6 b2b4 f7f6 d4e3 e6e5 e1g1 g7g5 a2a4 h7h5 h2h4 g5g4 f3e1 f8h6 e3e2 d5d4 c3d4 h8e8 b4b5 e5d4 b5c6 e8e2 c6b7 c8b7');
INSERT INTO evaluations VALUES(44685,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44685,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(44685,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(44685,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(44685,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44685,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(44685,36,'cp 26','b8c6 b1a3 c8d7 a3c2 d8a5 c1d2 a5b6 d4c5 b6c7 c2d4 c6e5 d4b5 d7b5 f1b5 e7c6 d2f4 e5d3 d1d3 c7f4 b2b4 f8e7 d3d4 f4f5 f3e5 e8g8 e5c6 b7c6 b5c6 e7f6 d4e3 a8c8 c6b7 c8c7 b7a6 f5c2 e1g1 c2c3 a1b1 f8b8 a2a3 g8f8 f1c1 c3e3 f2e3 f8e7 a3a4 h7h5 a6e2 g7g6 b4b5 b8c8');
INSERT INTO evaluations VALUES(56145,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(56145,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(56145,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(56145,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(56145,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(56145,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(56145,36,'cp 69','d2d3 d8d5 b1c3 f8b4 c1d2 b4c3 d2c3 g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 g1f3 b8c6 e5c7 c8e6 e2d2 a4e4 e1f2 e4f5 c7f4 e6d5 f4g3 d5f3 g2f3 f5c5 d2e3 c5c2 f1e2 a8d8 h1c1 c2d2 g3f4 d2d4 c1d1 d4f6 d1d8 e8d8 a1d1 d8c8 f4d6 g8d8 e3f4 f6f4 d6f4');
INSERT INTO evaluations VALUES(56144,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(56144,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(56144,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(56144,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(56144,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(56144,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(56144,36,'cp 38','d8d5 b1c3 f8b4 c1d2 b4c3 d2c3 g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 e5c7 b8c6 g1f3 c8g4 a1d1 a4a2 e2b5 a8c8 f1c4 a7a6 c4a2 a6b5 c7g3 e7f5 e1f2 f5g3 h2g3 h7h5 h1e1 e8f8 e1e4 c6a5 c2c3 g8g6 d1d5 a5c4 a2c4 b5c4 d5b5 c8c7 b5b4 g4e6 e4f4 f8g7');
INSERT INTO evaluations VALUES(34915,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(34915,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(34915,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(34915,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(34915,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(34915,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(34915,36,'cp 50','b1c3 f8b4 c1d2 b4c3 d2c3 g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 e5c7 b8c6 g1f3 c8g4 a1d1 a4a2 e2b5 a8c8 f1c4 a7a6 c7g3 a6b5 c4a2 c6b4 a2b1 b4c2 e1f2 g4f5 f3h4 g8g4 h4f5 e7f5 d1d2 g4c4 h1c1 f5g3 f2g3 c2d4 c1c4 c8c4 b1h7 d4c6 h7d3 c4b4 d2e2 e8f8 d3e4');
INSERT INTO evaluations VALUES(34914,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(34914,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(34914,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(34914,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(34914,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(34914,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(34914,36,'cp 47','f8b4 c1d2 b4c3 d2c3 g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 e5c7 b8c6 g1f3 c8e6 e2d2 a4e4 e1f2 e4f5 c2c4 e6c4 a1c1 c4a2 f1e2 f5d7 d2d7 e8d7 c7f4 a8d8 h1e1 d7e8 f3g5 g8g7 h2h4 e8f8 h4h5 e7d5 h5h6 d5f4 h6g7 f8g7 g5f3 f4e2');
INSERT INTO evaluations VALUES(33130,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(33130,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(33130,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(33130,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(33130,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(33130,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(33130,36,'cp 47','c1d2 b4c3 d2c3 g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 g1f3 b8c6 e5c7 c8g4 a1d1 a8c8 c7g3 a4a2 c2c3 g8g6 e2b5 a7a6 f1c4 a6b5 c4a2 e7f5 g3f4 g4f3 g2f3 g6g2 d1d2 f5h4 d2e2 c6e7 e1f1 g2e2 f1e2 b5b4 c3c4 e7g6 f4g3 h4f5');
INSERT INTO evaluations VALUES(33131,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(33131,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(33131,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(33131,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(33131,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(33131,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(33131,36,'cp 46','b4c3 d2c3 g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 e5c7 b8c6 g1f3 c8g4 a1d1 a4a2 e2b5 a8c8 c7g3 a2e6 e1f2 a7a6 b5c4 e6c4 f1c4 c6b4 d1d4 b4c2 d4f4 g4e6 c4e6 f7e6 f4e4 e8f7 h2h4 g8d8 g3f4 e7d5 f4g5 d8f8 g5h6 f8e8 f3e5 f7e7 h1d1 c2b4 h6g5 e7d6 e4b4 c8c2 f2g3 d6e5 b4b7');
INSERT INTO evaluations VALUES(33133,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(33133,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(33133,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(33133,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(33133,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(33133,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(33133,36,'cp 49','d2c3 g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 e5c7 b8c6 g1f3 c8g4 a1d1 a4a2 e2b5 a8c8 c7g3 a2e6 e1f2 a7a6 b5c4 e6c4 f1c4 c6b4 d1d4 b4c2 d4f4 g4e6 c4e6 f7e6 f4e4 e8f7 h2h4 g8d8 h1c1 e7d5 f3g5 f7g8 g5e6 d8e8 e4g4 g8f7 e6g5 f7g6 g5f3 g6f7 f3e5 f7f6 g4g5 c2e3 c1e1 e8g8 e5d7 f6f7 e1e3 g8g5 h4g5 d5e3');
INSERT INTO evaluations VALUES(33132,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(33132,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(33132,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(33132,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(33132,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(33132,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(33132,36,'cp 61','g8e7 d3e4 d5e4 d1e2 e4f4 c3g7 h8g8 g7e5 f4a4 g1f3 b8c6 e5c7 c8e6 a2a3 e7d5 b2b3 a4g4 c7e5 e8c8 e1c1 d5f4 d1d8 g8d8 e2e4 c6e5 e4e5 f4g2 e5c3 c8b8 c1b2 g2h4 f3h4 g4h4 f1d3 b8a8 c3a5 h4d4 b2a2 d8c8 h1e1 a8b8 d3h7 d4f4 a5e5 f4e5 e1e5');
INSERT INTO evaluations VALUES(41077,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(41077,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(41077,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(41077,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41077,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(41077,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41077,36,'cp 0','f1d3 h7h6 g5h4 f6h5 f3e5 h5f4 h4e7 d8e7 e1g1 e8g8 b1c3 b8d7 f1e1 e7g5 g2g3 f4h3 g1g2 h3f4');
INSERT INTO evaluations VALUES(41070,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41070,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41070,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41070,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41070,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41070,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41070,36,'cp 0','h7h6 g5h4 f6h5 h4e7 d8e7 f3e5 h5f4 e1g1 e8g8 b1c3 b8d7 f1e1 e7g5 g2g3 f4h3 g1g2 h3f4');
INSERT INTO evaluations VALUES(45626,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(45626,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(45626,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(45626,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(45626,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(45626,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(45626,36,'cp 42','e2e3 f8d6 f4d6 d8d6 a1c1 e8g8 g1f3 f8d8 d1b3 b8d7 c4d5 e6d5 f1d3 d8e8 e1g1 f6e4 a2a4 d7f6 h2h3 d6e7 a4a5 c8f5 f1e1 e4c3 d3f5 c3e4 a5a6 a8b8 a6b7 e7b7 b3b7 b8b7 c1c6 b7b2 f5e4 f6e4');
INSERT INTO evaluations VALUES(45624,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45624,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45624,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(45624,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(45624,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(45624,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(45624,36,'cp 19','f8d6 f4d6 d8d6 d1c2 e8g8 g1f3 d6e7 a1c1 f8d8 a2a3 c8d7 f1d3 d5c4 d3c4 d7e8 c3e4 f6e4 c2e4 b8d7 e1g1 b7b6 f1d1 d7f6 e4c2 c6c5 c4a6 a8b8 d4c5 d8d1 c2d1 b6c5 d1c2 e7d6 a6e2 d6b6 e2d3 b6b2 c2c5 b2b7');
INSERT INTO evaluations VALUES(40538,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(40538,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40538,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(40538,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(40538,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(40538,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40538,36,'cp 44','f4g3 h7h6 g1f3 e8g8 f3e5 c6c5 f1e2 c5d4 e3d4 d5c4 e1g1 b8c6 e2c4 d6e7 e5c6 b7c6 f1e1 a7a5 c3a4 c8a6 c4b3 a6b5 a4c5 e7c5 d4c5 d8d1 a1d1 f6d5 g3d6 f8d8 b3c2 d5b4');
INSERT INTO evaluations VALUES(40542,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(40542,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40542,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(40542,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(40542,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(40542,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(40542,36,'cp 27','h7h6 g1f3 e8g8 f1d3 d5c4 d3c4 d6g3 h2g3 c6c5 d4c5 d8a5 e1g1 a5c5 c4e2 b8c6 a1c1 c5e7 d1c2 c8d7 a2a3 f8d8 f1d1 a8c8 b2b4 a7a6 c2b1 d7e8 d1d8 e7d8 c3a4 f6d7 g3g4 c6e7 g4g5 c8c1 b1c1 h6g5 f3g5 d7f6 a4c5');
INSERT INTO evaluations VALUES(52502,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(52502,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(52502,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(52502,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(52502,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(52502,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(52502,36,'cp -132','c3b2 c1b2 e7e6 g1f3 b8a6 b1c3 a6c5 e1g1 a7a6 a2a4 d7d6 d1e2 g8e7 b2a3 e7g6 a1b1 f8e7 e2e3 c5d7 f1d1 e8g8 b1c1 d7e5 f3e5 g6e5 c4e2 h7h6 f2f4 e5c6 c1b1 a8b8 e3f2 d8c7 b1c1 f8d8 f2e3');
INSERT INTO evaluations VALUES(52504,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(52504,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(52504,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(52504,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(52504,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(52504,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(52504,36,'cp -119','c1b2 e7e6 g1f3 b8a6 b1c3 a6c5 e1g1 b7b6 c3d5 c8b7 f1e1 a8c8 a1c1 g8f6 d5f6 g7f6 c4d5 b7a6 d5c4 f8e7 f3d2 b6b5 c4f1 a6b7 d1g4 a7a6 g4g7 h8f8 g7h6 e6e5 h2h3 c5e6 c1c8 b7c8 e1c1 c8b7 h6h7 d8b6');
INSERT INTO evaluations VALUES(52503,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(52503,31,'cp -123',NULL);
INSERT INTO evaluations VALUES(52503,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(52503,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(52503,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(52503,35,'cp -143',NULL);
INSERT INTO evaluations VALUES(52503,36,'cp -144','e7e6 g1f3 b8a6 b1c3 a6c5 e1g1 b7b6 c3d5 c8b7 f1e1 a8c8 a1c1 g8f6 d5f6 g7f6 c4d5 f8e7 d5b7 c5b7 e4e5 f6f5 f3d4 c8c1 d1c1 h8g8 c1h6 e7f8 h6h7 d8g5 g2g3 b7c5 e1c1 g5g6 h7h3 g6h6 h3h6 f8h6 c1d1 c5e4 d4b5 h6f8 a2a4 f8c5');
INSERT INTO evaluations VALUES(52043,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(52043,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(52043,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(52043,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(52043,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(52043,35,'cp -163',NULL);
INSERT INTO evaluations VALUES(52043,36,'cp -157','b1c3 b8a6 g1f3 a6c5 e1g1 b7b6 c3d5 c8b7 f1e1 a8c8 a1c1 g8f6 d5f6 g7f6 c4d5 f8e7 d5b7 c5b7 c1b1 h8g8 g2g3 d8c7 d1e2 b7c5 f3d4 c7f4 d4f5 f4g4 f2f3 g4g5 b2c1 g5h5 f5e7 e8e7 e1d1 f6f5 e4e5 e7f8 c1a3 f8g7 d1d4 h5g6 e2d2 h7h5');
INSERT INTO evaluations VALUES(29634,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(29634,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(29634,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(29634,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(29634,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(29634,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(29634,36,'cp 148','c4d5 c7c6 f3e5 c6d5 e5g4 f6g4 e2e4 g4f6 e4e5 f6d7 c1e3 b8c6 f1d3 e7e6 e1g1 c6b4 d3e2 f8e7 a2a3 b4c6 e2d3 e8g8 b1d2 f7f6 d1h5 f6f5 b2b4 a7a6 d2b3 d8e8 h5e8 f8e8 g2g3 g7g5 a1c1 e8c8 c1c3 c6a7 b3c5 d7c5 b4c5 a7b5');
INSERT INTO evaluations VALUES(16296,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(16296,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(16296,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(16296,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(16296,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(16296,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(16296,36,'cp 137','e4e5 f6g8 a2a3 g8e7 b1c3 d7d6 f1d3 e7f5 d3f5 e6f5 d1e2 f8e7 e5d6 c7d6 d4d5 c6e5 f3e5 d6e5 e2e5 e8g8 e1g1 e7f6 e5g3 c8d7 c1f4 f8e8 f4c7 d8e7 d5d6 e7e5 g3d3 e5c5 c3d5 f6b2 a1d1 e8e4 d5e7 g8h8 d3b3 c5a3 b3b7 a8e8 b7d5 b2f6 f1e1 f6e7 e1e4');
INSERT INTO evaluations VALUES(16248,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(16248,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(16248,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(16248,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(16248,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(16248,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(16248,36,'cp 141','f6g8 f1d3 c6b4 d3e2 d7d6 e1g1 b4c6 e2d3 g8e7 d1e2 e7g6 f1d1 f8e7 c2c4 e8g8 b1c3 c6b4 d3b1 f7f5 a2a3 b4c6 e5d6 c7d6 c4c5 g8h8 b1a2 d6d5 b2b4 a7a6 c3a4 c8d7 a4b6 a8b8 d1e1 e7f6 c1b2 g6f4 e2e3 g7g5');
INSERT INTO evaluations VALUES(42609,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(42609,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(42609,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42609,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(42609,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(42609,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(42609,36,'cp -14','e5d4 f3d4 e8g8 c1g5 f8e8 d1d3 h7h6 g5h4 b8c6 d4c6 b7c6 e1c1 a8b8 d3f3 d7d6 h2h3 c5b4 h4f6 d8f6 f3f6 g7f6 f1d3 b4c3 b2c3 e8e5 c1d2 e5a5 d1a1 a5a4 c3c4 h6h5 h1b1 b8b6 h3h4 c6c5 d2c3 b6b4 b1b4 c5b4 c3b3 c8d7');
INSERT INTO evaluations VALUES(42683,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42683,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42683,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(42683,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(42683,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(42683,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(42683,36,'cp -13','f3d4 e8g8 c1g5 f8e8 d1d3 h7h6 g5h4 b8c6 d4c6 b7c6 e1c1 a8b8 d3f3 d7d6 h2h3 c5b4 h4f6 d8f6 f3f6 g7f6 f1d3 b4c3 b2c3 e8e5 c1d2 e5a5 d1a1 a5a4 c3c4 c8e6 h1b1 b8b6 d2c3 h6h5 h3h4 g8f8 g2g3 f8e7 b1b3 c6c5 f2f3 e7d7 a2a3');
INSERT INTO evaluations VALUES(48857,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(48857,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(48857,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(48857,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(48857,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(48857,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(48857,36,'cp -74','f8d6 e2e3 e8g8 g1e2 c7c5 b1c3 d5d4 c3b5 d6e7 e1g1 b8c6 a2a4 c8e6 f2f4 e5e4 f4f5 e6d7 e3d4 c5d4 b5d4 c6b4 g1h1 e7c5 d2d3 e4d3 c2d3 a8c8 g2b7 f8e8 e2f4 h7h6 d1d2 c5d4 b2d4');
INSERT INTO evaluations VALUES(41735,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(41735,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(41735,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(41735,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(41735,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(41735,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(41735,36,'cp -77','d2d3 c7c5 e2e4 d5d4 b2c1 b8c6 b1d2 c8e6 g1f3 f6d7 d2c4 d6e7 a2a4 a7a6 a4a5 e6c4 b3c4 c6a5 f3d2 a5c6 h2h4 e8g8 g2h3 d8c7 e1g1 b7b6 d1e2 a6a5 f2f4 a5a4 f4f5 g8h8 g3g4 c6b4 d2f3 b6b5');
INSERT INTO evaluations VALUES(55776,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(55776,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(55776,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55776,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(55776,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(55776,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(55776,36,'cp -30','e6d5 g2g3 f8c5 g1f3 g8f6 f1g2 e8g8 e1g1 f8e8 b1c3 h7h6 a2a3 a7a5 c1f4 c7c6 f1e1 e8e1 d1e1 c8f5 f4e3 c5f8 e1d2 b8d7 a1e1 d8c7 c3e2 a5a4 e2c3 c7a5 c3e2 c6c5 d2a5 a8a5 g1h1 a5a6 h2h3');
INSERT INTO evaluations VALUES(56282,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(56282,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56282,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(56282,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(56282,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(56282,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(56282,36,'cp -10','g2g3 g8f6 f1g2 f8c5 g1f3 e8g8 e1g1 a7a5 f1e1 a5a4 b1a3 c7c6 c2c3 c8f5 b2b4 a4b3 a2b3 f6g4 d3d4 c5d6 f3h4 f5d7 h4f3 d7e6 a3c2 a8a1 c2a1 b8d7 a1c2 g4f6 f3g5 e6g4 g2f3 g4f5 c3c4 f5g6 c2e3 f6e4 g5e4 d5e4');
INSERT INTO evaluations VALUES(40374,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(40374,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(40374,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(40374,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(40374,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(40374,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40374,36,'cp 21','f2f3 b8c6 d4b3 h7h6 g5e3 c8e6 d1d2 d6d5 e4d5 f6d5 c3d5 d8d5 d2d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5e2 h6h5 b3c5 c6e5 h1e1 d5c6 d1d8 h8d8 c2c3 g7f6 c1c2 f6h4 e1d1 e7e6 e3f4 d8d1 c2d1 h4f6 c5e4 f6g7 d1c2 c6d5 a2a4 c8d7');
INSERT INTO evaluations VALUES(57304,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(57304,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(57304,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(57304,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(57304,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(57304,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(57304,36,'cp 14','c3d4 d7d5 b1c3 c7c6 c1f4 g8f6 e2e3 c8f5 f1d3 f5d3 d1d3 b8d7 h2h3 d7b6 g1e2 f6h5 f4e5 f8d6 e1g1 e8g8 b2b3 f8e8 e5d6 d8d6 f2f3 h5g3 e2g3 d6g3 d3d2 g3g5 a1e1 a8d8 d2f2 g5f5 e1d1 f5f6 e3e4 d5e4 f3e4 f6f2 g1f2');
INSERT INTO evaluations VALUES(57285,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(57285,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(57285,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57285,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(57285,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(57285,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(57285,36,'cp 6','d7d5 b1c3 c7c6 c1f4 g8f6 e2e3 c8f5 f1d3 f5d3 d1d3 f6h5 g1f3 b8d7 e1g1 f8e7 h2h3 e8g8 f4h2 h5f6 a1c1 f8e8 d3c2 d7b6 f3e5 e7d6 f1e1 h7h6 b2b3 f6e4 e5d3 d6h2 g1h2 b6c8 h2h1 e4c3 c2c3 c8d6 f2f3');
INSERT INTO evaluations VALUES(17778,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(17778,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(17778,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(17778,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(17778,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(17778,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(17778,36,'cp 101','b5c6 b7c6 d2d4 e5d4 d1d4 c6c5 d4d3 f8e7 c1f4 g8f6 e1c1 e8g8 e4e5 d6e5 f4e5 d8d3 d1d3 f8e8 e5c7 c5c4 d3d1 e7c5 c7g3 c8b7 h1e1 e8e1 d1e1 c5b4 e1e3 a8c8 h2h3 b4c3 e3c3 f6d5 c3a3 f7f5 f3d4 f5f4 g3h2 c4c3 b2c3 d5c3 h2f4 b7g2');
INSERT INTO evaluations VALUES(17748,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(17748,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(17748,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(17748,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(17748,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(17748,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(17748,36,'cp 89','b7c6 d2d4 e5d4 d1d4 c6c5 d4d3 f8e7 c1f4 g8f6 e1c1 e8g8 e4e5 d6e5 f4e5 c8b7 d3c4 d8c8 h1e1 h7h6 e5f6 e7f6 c3d5 c8f5 c4d3 f5g6 f3e5 f6g5 f2f4 g6d3 d1d3 c5c4 d3d1 f8d8 f4g5 d8d5 g5h6 d5d1 c1d1 a8d8 d1c1 c4c3 b2c3 g7h6 g2g3 d8d5 a2a3 g8f8 c3c4 d5d4');
INSERT INTO evaluations VALUES(18009,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(18009,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(18009,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(18009,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(18009,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(18009,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(18009,36,'cp 98','d2d4 e5d4 d1d4 c6c5 d4d3 f8e7 c1f4 g8f6 e1c1 c8e6 e4e5 d6e5 d3e3 d8b8 f3e5 b8b6 b2b3 c5c4 e5c4 e6c4 e3b6 c7b6 b3c4 e7a3 c1b1 e8g8 c3d5 a8c8 d5f6 g7f6 f4h6 f8e8 h1e1 e8e6 e1e3 e6e3 f2e3 a3f8 h6f8 g8f8 d1d6 c8c4 d6b6 a6a5 b6f6 f8g7');
INSERT INTO evaluations VALUES(18004,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(18004,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(18004,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(18004,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(18004,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(18004,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(18004,36,'cp 97','e5d4 d1d4 c6c5 d4d3 f8e7 c1f4 g8f6 e1c1 e8g8 e4e5 d6e5 f4e5 d8d3 d1d3 f8e8 e5c7 c5c4 d3d1 c8b7 c7e5 f6e4 c3e4 b7e4 h1e1 e4f3 g2f3 e7g5 c1b1 f7f6 e5d4 a8d8 c2c3 g5f4 h2h3 g8f7 b1c2 h7h5 b2b3 e8e1');
INSERT INTO evaluations VALUES(18019,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(18019,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(18019,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(18019,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(18019,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(18019,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(18019,36,'cp 83','d1d4 c6c5 d4d3 f8e7 c1f4 c8e6 e1c1 g8f6 e4e5 d6e5 d3d8 a8d8 f4e5 d8d7 f3g5 f6g4 g5e6 f7e6 e5g3 h7h5 h2h3 g4f6 h3h4 h8g8 d1e1 e8f7 g3e5 g8b8 g2g3 f6d5 e1e2 d5b4 f2f4 b8e8 h1e1 e8d8 b2b3 b4c6 c1b2 c6e5 e2e5');
INSERT INTO evaluations VALUES(32985,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(32985,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(32985,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(32985,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(32985,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(32985,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(32985,36,'cp 242','e1g1 d7d6 d2d4 c5d4 f3d4 e5d4 e4e5 f6g6 c4d3 c8f5 d3f5 g6f5 d1d4 d6d5 b2b3 b8a6 c1a3 g8h6 f1e1 f5e6 c3e2 c6c5 a3c5 h6f5 d4a4 e6c6 a4f4 a6c5 f4f5 e8g8 a1d1 f8e8 e2d4 c6b6 e1e3 c5e6 c2c3 a8c8 f5g4 b6a5 g4f3 e6g5 f3f5 g5e4 e3f3 a5c7 d4b5 c7e7 f5f4 e7e5 f4f7 g8h8 d1d5');
INSERT INTO evaluations VALUES(32984,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(32984,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(32984,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(32984,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(32984,34,'cp 234',NULL);
INSERT INTO evaluations VALUES(32984,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(32984,36,'cp 240','d7d6 d2d4 c5d4 f3d4 e5d4 e4e5 f6g6 c4d3 c8f5 d3f5 g6f5 d1d4 d6d5 b2b3 b8a6 c1a3 g8h6 f1e1 a6c7 a1d1 c7e6 d4b4 e8c8 b4a5 c8b8 a3d6 d8d6 e5d6 f5f4 d6d7 f4c7 a5b4 c7d7 c3e2 h8d8 e2f4 h6f5 h2h3 d7d6 b4d6 d8d6 f4e6 f7e6 g1f1 b7b6 c2c3');
INSERT INTO evaluations VALUES(33032,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(33032,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(33032,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(33032,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(33032,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(33032,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(33032,36,'cp 230','d2d4 c5d4 f3d4 e5d4 e4e5 f6g6 c4d3 c8f5 d3f5 g6f5 d1d4 d6d5 b2b3 b8a6 c1a3 g8h6 f1e1 f5e6 c3e2 a6c7 d4b4 e8c8 e2d4 e6g6 b4a5 h6f5 d4f5 g6f5 a5a7 c7b5 a7a8 c8c7 a3d6 b5d6 e5d6 c7b6 a8a4 d8d6 e1e7 d6d7 e7d7 f5d7 a4d4 b6a6 d4g7 h8e8 g7h7 e8e4 g2g3 d7e6');
INSERT INTO evaluations VALUES(16502,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(16502,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(16502,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(16502,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(16502,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(16502,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(16502,36,'cp -46','f8c5 f1e2 e8g8 e1g1 f8e8 f1e1 a7a6 d2d3 h7h6 c3a4 c5a7 c2c4 f6h7 c1e3 d7d6 d1d2 c8g4 a4c3 g4f3 e2f3 c6d4 f3d1 c7c6 b3b4 d4e6 a1b1 a7d4 e3d4 e6d4 c3e2 d4e2 d1e2 h7g5 e2g4 g5e6 g2g3 e6d4 g1g2');
INSERT INTO evaluations VALUES(10203,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(10203,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(10203,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(10203,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(10203,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(10203,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(10203,36,'cp -37','f1e2 e8g8 e1g1 f8e8 f1e1 a7a6 d2d3 h7h6 c1b2 c6d4 c3d5 d4e2 e1e2 d7d6 d3d4 e5d4 d5f6 d8f6 b2d4 c5d4 d1d4 f6d4 f3d4 c7c5 d4f3 d6d5 a1e1 d5d4 f3d2 b7b5 h2h3 a6a5 f2f4 a5a4 e4e5 c8a6 e5e6 f7e6 e2e6 e8e6 e1e6 a6b7');
INSERT INTO evaluations VALUES(20987,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(20987,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(20987,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(20987,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(20987,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(20987,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(20987,36,'cp -17','c8f5 a2a3 g8f6 c2c4 e7e6 b1c3 f8e7 g1f3 e8g8 f1e2 d5c4 e2c4 c6b8 d1e2 c7c5 d4c5 e7c5 e1g1 a7a6 g1h1 b7b5 e3e4 f5g6 c4b3 b8c6 c1e3 d8b6 e3c5 b6c5 a1c1 c5b6 f4f5 e6f5 e4f5 g6h5 c3d5 f6d5 b3d5 c6d4 e2d1 d4f5 d5a8');
INSERT INTO evaluations VALUES(26549,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26549,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(26549,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(26549,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26549,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(26549,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(26549,36,'cp 0','a2a3 g7g6 c2c4 e7e6 g1f3 f8g7 f1e2 g8f6 b1c3 h7h5 e1g1 c6e7 c1d2 c7c6 f3e5 f6e4 d2e1 e4c3 e1c3 f5e4 e2d3 e4d3 d1d3 a7a5 h2h3 e7f5 c4c5 g7e5 f4e5 d8g5 f1f4 g5g3 f4f3 g3g5');
INSERT INTO evaluations VALUES(55866,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(55866,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(55866,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55866,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(55866,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55866,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(55866,36,'cp 8','c7c5 c2c4 d5c4 g1f3 b8c6 d1a4 c5d4 f3d4 d8d4 g2c6 c8d7 c1e3 d7c6 a4c6 d4d7 c6c4 a8c8 c4b3 f8c5 e1g1 c5e3 b3e3 a7a6 b1c3 g8f6 a1d1 d7e7 d1d4 e8g8 f1d1 h7h6 a2a3 f8d8 g1g2 d8d4 e3d4 c8c6 e2e4 c6c5 e4e5 f6d5 c3e4 c5c7 e4d6 c7c2');
INSERT INTO evaluations VALUES(54220,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54220,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(54220,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(54220,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(54220,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54220,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54220,36,'cp 17','g1f3 c5d4 e1g1 f8e7 f3d4 e6e5 d4b5 a7a6 b5c3 d5d4 c3d5 b8c6 e2e3 c8e6 c2c4 g8f6 e3d4 e5d4 d5e7 d8e7 f1e1 e8g8 g2c6 b7c6 d1d4 a8d8 d4f4 e7b4 c1d2 b4c4 f4c4 e6c4 d2g5 f8e8 b1c3 e8e1 a1e1 d8e8 e1e8 f6e8 g5e3 e8c7 b2b3 c4e6 c3a4 f7f6 e3c5 a6a5');
INSERT INTO evaluations VALUES(38277,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(38277,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(38277,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(38277,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(38277,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(38277,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(38277,36,'cp 103','f1d3 b8d7 e1g1 c7c5 d4d5 e6d5 e4d5 g8f6 b1c3 e8g8 f1e1 f8e8 h2h3 h7h6 c1f4 d7b6 d3f1 e7f8 g2g4 c8d7 d1d3 e8e1 a1e1 a7a6 f3d2 b6a4 c3a4 d7a4 b2b3 a4d7 d2c4 d8c7 d3f3 b7b5 c4e3 a8e8 c2c4 b5c4 b3c4');
INSERT INTO evaluations VALUES(38276,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(38276,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(38276,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(38276,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(38276,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(38276,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(38276,36,'cp 98','c7c5 d4d5 e6d5 e4d5 g8f6 e1g1 e8g8 c2c4 c8g4 h2h3 g4f3 d1f3 b8d7 f3e2 g7g6 b1c3 f6h5 g2g4 h5g7 f2f4 e7f6 c1d2 f6d4 g1h2 a7a6 a1e1 f8e8 e2f3 e8e1 f1e1 d8a5 c3e4 a5a2 e4d6 a2b2 d6e4 a8e8');
INSERT INTO evaluations VALUES(45953,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(45953,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45953,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45953,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(45953,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(45953,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(45953,36,'cp 22','f8e7 f1c4 c8e6 d1d3 b8c6 e1g1 e8g8 a2a3 a8c8 c4a2 h7h6 c3d5 e6d5 e4d5 e5e4 d3e3 e4f3 d5c6 c8c6 e3f3 d6d5 a2b3 d8b6 c1e3 e7c5 e3c5 c6c5 f1e1 c5c6 a1d1 c6e6 h2h3 f8e8 e1e6 f7e6 c2c4 d5d4 d1d3');
INSERT INTO evaluations VALUES(40162,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(40162,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40162,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40162,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40162,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40162,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40162,36,'cp 17','f1c4 c8e6 d1d3 b8c6 e1g1 e8g8 a2a3 a8c8 c4a2 h7h6 c3d5 e6d5 e4d5 e5e4 d3e3 e4f3 d5c6 c8c6 e3f3 c6c2 f3b7 d6d5 b7b3 c2c8 c1d2 e7c5 a1d1 a7a5 b3g3 g8h8 g3d3 f8e8 g2g3 f6e4 a2b1 d8b6');
INSERT INTO evaluations VALUES(40160,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(40160,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40160,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(40160,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(40160,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(40160,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(40160,36,'cp 22','c8e6 c4b3 b8d7 d1e2 a7a6 e1g1 d8c7 f1d1 b7b5 c1g5 a8c8 a2a3 e8g8 h2h3 d7b6 f3e1 b6c4 b3c4 e6c4 e2e3 f6h5 g5e7 c7e7 c3d5 e7b7 a3a4 c4d5 d1d5 b5a4 a1a4 c8c6 e1d3 c6c2 d5d6 h5f4 d3f4 e5f4 e3f4 b7b2');
INSERT INTO evaluations VALUES(30821,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(30821,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(30821,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30821,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30821,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30821,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(30821,36,'cp 0','d7d6 g2g3 g7g6 f1g2 c7c5 e1g1 g8f6 f1e1 f8g7 a2a4 e8g8 b1a3 b8d7 c2c3 d8c7 a1b1 f8d8 a3b5 c7b8 b5a3');
INSERT INTO evaluations VALUES(9784,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(9784,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9784,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9784,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9784,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(9784,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9784,36,'cp 11','d7d6 c3a4 c5b6 e1g1 e8g8 b2b4 c6e7 a4b6 a7b6 h2h3 d8e8 f1e1 c8e6 c4e6 f7e6 c2c4 f6h5 c4c5 e8g6 c5d6 c7d6 c1e3 e7c6 f3h4 g6f7 d3d4 e5d4 e3d4 h5f4 d4e3 c6e5 d1d6');
INSERT INTO evaluations VALUES(8857,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(8857,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(8857,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(8857,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(8857,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(8857,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(8857,36,'cp 18','c3a4 c5b6 e1g1 c8e6 c4e6 f7e6 c2c3 e8g8 h2h3 d8e8 b2b4 c6e7 a4b6 a7b6 a3a4 e7g6 c3c4 e8f7 c1d2 f6d7 d2e3 d6d5 f3h2 d5d4 e3c1 g6f4 c1f4 e5f4 h2f3 e6e5 d1b3 f7e6 g1h2 g7g5 f1b1 g8g7 a4a5 b6a5');
INSERT INTO evaluations VALUES(21610,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(21610,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(21610,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(21610,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(21610,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(21610,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(21610,36,'cp 63','e4d5 d8d5 b1c3 f8b4 f1d3 d5e6 e1f2 g8e7 c1f4 e6d7 a2a3 b4c3 b2c3 b7b6 c3c4 c6d4 d3e4 d4c6 h1e1 e8g8 f3e5 d7d1 a1d1 c8b7 c4c5 a8c8 e4f3 c6a5 f3g4 f7f5 g4e2 e7g6 e5g6 h7g6 h2h4 f8e8 e2d3 e8e1 d1e1 b7e4 d3e4 f5e4 e1e4 a5b7 e4e7 b7c5');
INSERT INTO evaluations VALUES(21560,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(21560,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(21560,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(21560,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(21560,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(21560,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(21560,36,'cp 56','d8d5 b1c3 f8b4 f1d3 d5e6 e1f2 g8e7 c1f4 e6d7 a2a3 b4c3 b2c3 b7b6 h1e1 c8b7 a3a4 e8g8 f2g1 a8e8 d1d2 e7g6 f4g3 c6e7 d2f2 e7d5 c3c4 d5b4 d4d5 b4d3 c2d3 c7c6 d5d6 c6c5 h2h4 e8e1 a1e1 f8e8 e1e8 d7e8 d3d4 c5d4 f2d4 g6f8 f3e5');
INSERT INTO evaluations VALUES(1701,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(1701,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(1701,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(1701,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(1701,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(1701,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(1701,36,'cp 79','b1c3 f8b4 f1d3 d5e6 e1f2 g8e7 c1f4 e6d7 a2a3 b4c3 b2c3 b7b6 h1e1 c8b7 a3a4 e8g8 f2g1 a8e8 d1d2 e7g6 f4g3 c6e7 d3b5 b7c6 d2d3 a7a5 c3c4 c6b5 c4b5 e7d5 h2h3 d5f4 d3d2 f4h5 g3h2 h5f6 c2c4 e8e1 a1e1 f8e8 e1e8 f6e8 h2g3 g6e7 d2c2 e7f5 g3f2 f7f6 d4d5');
INSERT INTO evaluations VALUES(1700,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(1700,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(1700,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(1700,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(1700,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(1700,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(1700,36,'cp 48','f8b4 f1d3 d5e6 e1f2 g8e7 c1f4 e6d7 a2a3 b4c3 b2c3 b7b6 h1f1 c8b7 f2g1 f7f6 d1e2 e8c8 a1e1 c8b8 f4g3 h7h5 a3a4 g7g5 h2h4 e7d5 e2d2 d7g4 g3h2 d5f4 h2f4 g5f4 e1e4 c6e5 e4f4 e5f3 f1f3 g4d7 a4a5 b7f3 f4f3 h8e8 d2f2 d7d6 a5b6 a7b6');
INSERT INTO evaluations VALUES(44783,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44783,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44783,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44783,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(44783,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44783,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(44783,36,'cp 11','f8g7 f2f3 b8c6 d4b3 h7h6 g5e3 c8e6 d1d2 d6d5 e4d5 f6d5 c3d5 d8d5 d2d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5e2 h6h5 b3c5 c6e5 h1e1 g7f6 b2b3 e7e6 h2h3 d5c6 a2a4 f6h4 d1d8 h8d8 e1d1 h4e7 d1d8 c8d8 c5e4 d8c7 e3f4 f7f6 f4g3 g6g5');
INSERT INTO evaluations VALUES(57764,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(57764,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(57764,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(57764,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(57764,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(57764,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(57764,36,'cp -81','d7d5 g2g3 c7c5 d2d3 b8c6 f1g2 h7h5 h3g5 h5h4 e2e4 d5d4 e4e5 c6e5 e1g1 e7e6 f2f4 e5g6 g3g4 f8e7 b1a3 e7g5 f4g5 g8e7 d1e2 e8g8 g2e4 d8d7 a3c4 d7c7 e2g2 c8d7 e4b7 a8b8 b7e4 e7d5');
INSERT INTO evaluations VALUES(55149,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(55149,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(55149,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(55149,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(55149,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(55149,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(55149,36,'cp -52','g2g3 e7e5 d2d4 b8c6 d4e5 c6e5 f1g2 c7c6 b1d2 g8f6 e1g1 f8d6 c2c4 e5c4 d2c4 d5c4 d1d4 e8g8 d4c4 f8e8 f1d1 d8e7 c1f4 d6c5 a1c1 c5b6 h3g5 c8f5 e2e3 a8d8 d1d8 e8d8 g5f3 f5d3 c4b3 h7h6 f4e5 d3e4 c1d1');
INSERT INTO evaluations VALUES(55147,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(55147,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(55147,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(55147,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(55147,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(55147,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(55147,36,'cp -77','c7c5 d2d3 h7h5 f1g2 b8c6 b1d2 g8f6 h3g5 h5h4 e2e4 e7e5 c2c3 d5d4 d1b3 d8c7 d2f3 h4g3 f2g3 f6h7 g5h7 h8h7 c3d4 c5d4 c1d2 c8e6 b3a4 f7f6 a1c1 c7d7 a2a3 a7a6 e1g1 e8c8 b2b4 c8b8 f3h4 d8c8 f1f2');
INSERT INTO evaluations VALUES(5299,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(5299,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(5299,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(5299,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(5299,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5299,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(5299,36,'cp 30','c2c3 e7e6 c4b5 e6e5 d2d4 c5d4 c3d4 e5d4 c1f4 a7a6 b5a4 g8e7 f4d6 b7b5 a4b3 c8b7 f1e1 e7c8 e4e5 c8d6 e5d6 e8f8 b1d2 f7f5 a1c1 d8f6 c1c6 b7c6 e1e7 a8e8 e7f7 f6f7 b3f7 f8f7 f3g5 f7f6 d2f3 c6f3 g5f3 e8e4 g2g3 h8e8 d1b3 h7h6 b3d5 d4d3 d5b7 e8d8 b7a6 e4e2');
INSERT INTO evaluations VALUES(5293,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(5293,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(5293,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(5293,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(5293,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(5293,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(5293,36,'cp 10','e7e6 c4b5 e6e5 d2d4 e5d4 c3d4 c5d4 e4e5 a7a6 b5a4 b7b5 a4b3 g8e7 c1f4 e8g8 b1d2 c8b7 f1e1 c6a5 f3d4 a5b3 d4b3 a8c8 d2e4 b7e4 e1e4 e7f5 d1d3 f8e8 a1e1 g7f8 e1d1 d8b6 g2g4 f5g7 f4g5 b6b7 e4d4 c8c4 g5f6 c4d4 b3d4 g7e6 d4e6 d7e6');
INSERT INTO evaluations VALUES(5105,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(5105,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(5105,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(5105,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(5105,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(5105,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(5105,36,'cp 25','c4b5 e6e5 d2d4 e5d4 c3d4 c5d4 c1f4 a7a6 b5a4 g8e7 b1d2 d7d5 e4d5 d8d5 a4b3 d5f5 f4d6 e8g8 f1e1 c8e6 b3e6 f7e6 d1b3 f8e8 b3b7 f5d5 d6e7 c6e7 b7d5 e7d5 d2c4 e8c8 b2b3 a6a5 f3g5 e6e5 g5e6 a5a4 e6g7 g8g7 c4e5 c8c2 e5f3 d4d3 a1d1 a4b3 a2b3');
INSERT INTO evaluations VALUES(52405,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52405,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(52405,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(52405,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(52405,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(52405,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52405,36,'cp 26','f1g2 d7d5 g1f3 c8g4 e1g1 e7e6 h2h3 g4h5 d2d4 b8d7 c2c4 g8f6 c4d5 c6d5 d1b3 d8b6 g3g4 h5g6 c1f4 f6e4 f1c1 f8d6 b3b6 a7b6 f4d6 e4d6 b1c3 e8e7 f3d2 h8c8 a2a4 d6c4 d2c4 c8c4');
INSERT INTO evaluations VALUES(52406,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(52406,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(52406,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52406,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(52406,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52406,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52406,36,'cp 25','d7d5 g1f3 c8g4 e1g1 g8f6 h2h3 g4f3 e2f3 b8d7 d2d3 g7g6 f3f4 f8g7 b1d2 e7e6 d2f3 e8g8 c2c4 f6e8 a1b1 e8d6 b2b3 d6f5 h3h4 h7h5 g2h3 f8e8 c1b2 g7b2 b1b2 d7f6 d1a1 c6c5 b2d2 d8e7 h3f5');
INSERT INTO evaluations VALUES(23105,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(23105,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(23105,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(23105,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(23105,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(23105,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(23105,36,'cp 47','d4f3 d1f3 g8f6 f3g3 d7d6 h2h4 g7g6 b1c3 c7c6 c1e3 f6h5 g3f3 d8f6 f3f6 h5f6 f2f3 c8e6 c4b3 h7h5 d3d4 e5d4 e3d4 f8g7 e1c1 e6b3 a2b3 e8e7 b3b4 a7a6 c1b1 f6e8 d4b6 a8c8 c3e2 e8c7 c2c3 c7e6 h1f1 h8f8 b6e3 f7f5 e4f5 f8f5');
INSERT INTO evaluations VALUES(23097,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(23097,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(23097,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(23097,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(23097,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(23097,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(23097,36,'cp 62','d1f3 g8f6 b1c3 d7d6 h1g1 c8g4 f3g3 g4e6 c4e6 f7e6 c1e3 g7g6 d3d4 e5d4 e3d4 f8g7 e1c1 f6h5 g3d3 g7d4 d3d4 e8g8 f2f3 a7a6 g1f1 d8f6 d4f6 h5f6 e4e5 d6e5 f1e1 f8d8 e1e5 d8d1 c3d1 g8f7 e5c5');
INSERT INTO evaluations VALUES(23096,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(23096,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(23096,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(23096,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(23096,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(23096,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(23096,36,'cp 56','g8f6 b1c3 d7d6 h1g1 c8g4 f3g3 g4e6 c4e6 f7e6 c1e3 d8e7 e1c1 g7g6 d3d4 f6h5 g3f3 e5d4 e3d4 f8h6 c1b1 h8f8 f3d3 e6e5 d4e3 h6e3 d3e3 h5f4 g2g3 f4e6 c3d5 e7g7 f2f4 c7c6 d5b4 e5f4 g3f4 g7e7 b4d3 e8c8 e3a7 e6f4 a2a3 f4d3 d1d3 c8c7 g1d1');
INSERT INTO evaluations VALUES(16978,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(16978,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(16978,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(16978,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(16978,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(16978,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(16978,36,'cp 60','b1c3 d7d6 h1g1 c8g4 f3g3 g4e6 c4e6 f7e6 c1e3 d8d7 e1c1 g7g6 f2f3 f8g7 d3d4 f6h5 g3f2 e8g8 h2h4 b7b5 d4e5 g7e5 c3e2 b5b4 c1b1 d7f7 e3h6 e5g7 h6g7 h5g7 e4e5 d6e5 f2c5 a7a5 c5e5 f7f5 e5e3 a5a4 e2d4');
INSERT INTO evaluations VALUES(16973,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(16973,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(16973,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(16973,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(16973,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(16973,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(16973,36,'cp 43','d7d6 h1g1 c8g4 f3g3 g4e6 c4e6 f7e6 c1e3 d8e7 e1c1 f6h5 g3f3 e7f7 f3h3 g7g6 c3e2 f8g7 g2g3 e8c8 c1b1 a7a6 g1f1 f7d7 h3g2 h8f8 f2f3 c8b8 e2c1 f8f7 c1b3 d8f8 c2c3 h5f6 g3g4 f6g8 h2h3 g7h6 e3h6 g8h6 d3d4');
INSERT INTO evaluations VALUES(14231,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14231,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14231,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14231,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14231,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14231,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(14231,36,'cp 5','c8f5 a2a3 e7e6 e2e3 f8d6 f4g5 c6e7 c2c4 c7c6 g5f6 g7f6 b1c3 d8d7 c4d5 e7d5 f1d3 f5g6 c3e4 e8c8 d1c2 d6c7 e1c1 d7e7 e4c5 c7b6 d3g6 h7g6 c5d3 b6c7 c1b1 c8b8 b1a1 e6e5 e3e4 d5b6 d4e5 c7e5 d3e5 d8d1 h1d1');
INSERT INTO evaluations VALUES(24830,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24830,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(24830,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(24830,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(24830,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(24830,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24830,36,'cp 14','e2e3 e7e6 a2a3 h7h6 f1d3 f5d3 d1d3 f8d6 b1d2 d8e7 e1c1 e8c8 h1e1 c8b8 f4d6 c7d6 e3e4 d5e4 d2e4 f6d5 e4c3 d5c3 d3c3 d8c8 c3d3 d6d5 c1b1 c8c7 c2c3 h8c8 b1a2 e7d6 f3e5 a7a5 d3e2 c6e5 e2e5 d6e5');
INSERT INTO evaluations VALUES(24826,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(24826,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24826,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(24826,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(24826,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24826,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24826,36,'cp 16','e7e6 a2a3 f8d6 f4g5 c6e7 c2c4 c7c6 g5f6 g7f6 b1c3 d8d7 a1c1 h8g8 f3h4 f5g6 d1f3 f6f5 c4d5 e7d5 c3d5 c6d5 f1d3 e8f8 e1g1 f8g7 g2g3 a8c8 c1c8 g8c8 h4g2 d7c7 h3h4 h7h5 g2f4 d6f4 e3f4 g7h8 f3e3');
INSERT INTO evaluations VALUES(25161,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(25161,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(25161,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(25161,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25161,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25161,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(25161,36,'cp 0','a2a3 f8d6 f4g5 c6e7 c2c4 c7c6 g5f6 g7f6 b1c3 d8d7 g2g3 e6e5 a1c1 e8g8 f3h4 f5e6 c4d5 c6d5 f1b5 d7d8 d1h5 a8c8 e1g1 d8b6 c1b1 g8h8 h5h6 e7g8 h6h5');
INSERT INTO evaluations VALUES(25168,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25168,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25168,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25168,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25168,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25168,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25168,36,'cp 0','f8d6 f4g5 c6e7 c2c4 c7c6 g5f6 g7f6 b1c3 d8d7 c4d5 e7d5 f1d3 f5g6 c3e4 e8c8 d1c2 d6c7 e4c5 d7e7 e1c1 c7d6 c5e4 d6c7');
INSERT INTO evaluations VALUES(24109,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(24109,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24109,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24109,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(24109,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24109,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(24109,36,'cp 12','f4g5 c6e7 c2c4 c7c6 g5f6 g7f6 b1c3 d8d7 f1d3 d5c4 d3c4 e7d5 c4d3 f5g6 c3e4 e8c8 e4d6 d7d6 d1c2 c8b8 e1c1 d6e7 c1b1 d5b6 h3h4 g6h5 d1g1 h5g6 d3g6 h7g6 b1a1 e6e5 g2g4 d8d5 h4h5 e5d4 h5g6 h8h1');
INSERT INTO evaluations VALUES(24926,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(24926,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(24926,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(24926,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(24926,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(24926,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(24926,36,'cp 36','e7e6 c2c4 f6e4 a2a3 e4d2 d1d2 f8d6 f4g3 h7h5 c4d5 e6d5 a1c1 c6e7 f1d3 g4f3 g2f3 d8d7 b2b4 a7a6 h1g1 d6g3 h2g3 e8c8 f3f4 c8b8 g1h1 e7c8 d2e2 g7g6 b4b5 a6b5 d3b5 d7d6 e1f1 d6a3');
INSERT INTO evaluations VALUES(25216,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(25216,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(25216,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(25216,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25216,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25216,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25216,36,'cp 46','h2h3 g4f3 d1f3 f8d6 c2c3 d6f4 f3f4 d8e7 f4h2 a7a5 g2g4 e8c8 f1b5 c6a7 b5d3 c8b8 e1c1 a5a4 a2a3 f6e8 f2f4 e8d6 c1b1 f7f6 h3h4 f6f5 g4g5 h7h6 c3c4 h6g5 f4g5');
INSERT INTO evaluations VALUES(25214,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(25214,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25214,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(25214,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(25214,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(25214,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(25214,36,'cp 23','g4f3 d1f3 f8d6 c2c3 d6f4 f3f4 d8e7 f4g3 e8g8 f1d3 e6e5 e1g1 a7a6 a2a4 f6h5 g3f3 h5f6 f3d1 a6a5 d1b3 b7b6 d3b5 c6d8 d4e5 e7e5 f1e1 e5d6 e3e4 d5e4 d2e4 f6e4 e1e4 d8e6 e4e3 a8d8 a1e1 e6c5 b3c2 c7c6');
INSERT INTO evaluations VALUES(48681,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(48681,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(48681,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48681,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(48681,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48681,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(48681,36,'cp -12','f8d6 c1g5 c8e6 e1g1 e8g8 b1d2 b8d7 f1e1 h7h6 g5h4 c7c6 h4g3 d6g3 h2g3 d8b6 b2b3 a7a5 c2c4 e6f5 a1c1 f8e8 d2f1 f5e4 f1e3 d7f8 c4d5 e4d5 e3d5 c6d5 d1d2 a5a4 b3a4 a8a4 e2d3 e8e1 c1e1 b6a5 d2a5 a4a5');
INSERT INTO evaluations VALUES(41721,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(41721,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(41721,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(41721,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(41721,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(41721,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(41721,36,'cp 0','c1g5 c8e6 b1d2 h7h6 g5h4 b8d7 e1g1 e8g8 f1e1 c7c6 c2c4 e6f5 c4d5 c6d5 d1b3 a8c8 h4g3 d6g3 h2g3 d8b6 a1c1 f8e8 e2f1 e8d8 f1e2');
INSERT INTO evaluations VALUES(6598,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(6598,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(6598,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(6598,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6598,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6598,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(6598,36,'cp 0','e5d4 f3d4 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 c1g5 c7c6 d1f3 b4d6 h2h3 h7h6 g5f6 d8f6 f3f6 g7f6 c3e2 d6e5 a1b1 a7a5 f2f4 e5d6 g1h2 f8e8 b2b3 e8e3 e2d4 c6c5 d4f5 e3d3 f5d6 d3d2 d6c8 a8c8 b1d1 d2c2 d1d5');
INSERT INTO evaluations VALUES(6666,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(6666,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6666,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(6666,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6666,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6666,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6666,36,'cp 6','f3d4 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 c1g5 c7c6 f1e1 h7h6 g5h4 b4d6 d1d2 a7a5 h2h3 a8b8 b2b3 b8b4 h4g3 d6g3 f2g3 c6c5 a1d1 c8d7 g1h2 d8c7 d2f2 f8e8 e1e8 d7e8 d1e1 c7d6 a2a3 b4b8 f2d2 d5d4 c3a4');
INSERT INTO evaluations VALUES(6653,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6653,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6653,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(6653,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(6653,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(6653,36,'cp 4','g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 f8e8 c1g5 c7c6 d1f3 h7h6 g5h4 b4d6 f1e1 c8e6 a1d1 a8b8 b2b3 b8b4 h4f6 d8f6 f3f6 g7f6 c3e2 g8f8 e2g3 d6g3 f2g3 c6c5 g3g4 a7a5 g1f2 b4b8 e1e2 f8e7 d1e1 e7d6 c2c3');
INSERT INTO evaluations VALUES(57149,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(57149,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(57149,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(57149,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(57149,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(57149,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(57149,36,'cp -87','g8f6 b1c3 f8c5 d2d3 d7d5 g1f3 b8c6 c1g5 d5e4 c3e4 c5e7 c2c3 c8e6 d1a4 d8d5 f1e2 h7h6 e4f6 g7f6 g5e3 e8c8 d3d4 h8g8 h1g1 e7f8 e1c1 d5a2 a4a2 e6a2 d4e5 f6e5 d1d8 c8d8 e2b5 f7f6 h4h5 a2e6 b5c6');
INSERT INTO evaluations VALUES(49605,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(49605,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(49605,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(49605,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(49605,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(49605,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(49605,36,'cp -85','g1f3 f6e4 f3e5 d8e7 d1e2 e7e5 d2d3 d7d5 d3e4 d5e4 b1c3 b8c6 e2e4 e5e4 c3e4 c6b4 f1b5 c8d7 b5d7 e8d7 e1d1 a8e8 e4g5 f7f6 g5h3 e8e4 g2g3 f8d6 c1d2 c7c5 h1e1 e4e1 d2e1 h8e8 e1d2 b7b6 b2b3 b4d5 c2c3 e8e4 d1c2 c5c4 a1e1 e4g4 b3c4 g4c4');
INSERT INTO evaluations VALUES(19306,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(19306,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19306,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19306,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(19306,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(19306,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19306,36,'cp 15','f8g7 d4c6 b7c6 e1g1 d7d6 b1c3 g8f6 e4e5 d6e5 d1d8 e8d8 f1e1 d8e8 f2f3 c8f5 e1e5 f6d5 c4d5 c6d5 e5d5 g7c3 b2c3 f7f6 c1a3 e8f7 a1b1 a8b8 b1b8 h8b8 g2g4 f5e6 d5a5 b8b7 a3b4 b7c7 g1f2 e6c4 h2h3 e7e5 a2a4 a7a6 f2e1 f7e6');
INSERT INTO evaluations VALUES(35023,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(35023,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(35023,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(35023,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(35023,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(35023,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(35023,36,'cp 177','d2d4 d7d6 c1e3 c8g4 b1c3 b8c6 d4d5 c6b8 c3b5 f6d8 b5a7 g8f6 h2h3 g4h5 d1d3 f8e7 f3d2 e8g8 a2a4 h5g6 e1g1 g8h8 f1e1 b8d7 a7b5 f6h5 a4a5 h5f4 d3b3 e7g5 c4f1 f7f5 e4f5 g6f5 b3c4 a8c8 a5a6 b7a6 b5a7 f4h3 g2h3 g5e3 e1e3');
INSERT INTO evaluations VALUES(35024,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(35024,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(35024,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(35024,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(35024,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(35024,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(35024,36,'cp 171','d7d6 c1e3 b8c6 b1c3 c8g4 d4d5 c6b8 c3b5 f6d8 b5a7 f8e7 h2h3 g4h5 d1d3 g8f6 f3d2 e8g8 e1g1 b8d7 a2a4 h5g6 a4a5 d7c5 e3c5 d6c5 a7b5 c7c6 d5d6 c6b5 d6e7 d8e7 c4b5 f8d8 d3e3 e7c7 d2c4 g6e4 c4b6 a8b8 f1e1 e4c2');
INSERT INTO evaluations VALUES(34758,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(34758,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(34758,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(34758,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(34758,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(34758,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(34758,36,'cp 198','b1c3 b8c6 c1e3 c8g4 d1d3 c6b4 c4f7 f6f7 d3b5 c7c6 b5b4 g8f6 e1c1 a7a5 b4b6 a8a6 b6b3 f7b3 a2b3 e5d4 d1d4 g4f3 g2f3 f8e7 d4d3 e8f7 f3f4 d6d5 e4d5 f6d5 c3d5 c6d5 d3d5 a5a4 b3a4 a6a4');
INSERT INTO evaluations VALUES(39562,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(39562,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(39562,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(39562,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(39562,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(39562,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(39562,36,'cp 8','c7c5 e2e3 b8c6 f1g2 e7e6 g1e2 g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 c3d5 e6d5 c1e3 e8g8 e1g1 c8g4 d1d2 g4e2 d2e2 g7d4 e3d4 c6d4 e2d3 d8f6 a1d1 d4c6 d3d5 f6b2 d1b1 b2f6 b1b7 c6d4 d5a5 a8d8');
INSERT INTO evaluations VALUES(37784,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(37784,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(37784,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(37784,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(37784,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(37784,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(37784,36,'cp 16','f1g2 b8c6 e2e3 e7e6 g1e2 g8e7 e1g1 e8g8 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 h2h4 h7h6 f1e1 f8e8 c3d5 e6d5 c1e3 c8g4 a1c1 d8a5 a2a3 a5a6 g2d5 a8d8 c1c5 c6e7 d5c4 a6f6 d1b3 e7f5 c5e5 f5e3 e5e8 d8e8 f2e3 f6e7 e2f4');
INSERT INTO evaluations VALUES(37785,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(37785,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(37785,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(37785,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(37785,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(37785,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37785,36,'cp 0','b8c6 g1f3 d7d6 e1g1 c8f5 h2h3 e7e5 d2d3 g8e7 a1b1 e8g8 a2a3 a7a5 f3d2 f5e6 c3d5 a8b8 b2b3 b7b5 c1b2 d8d7 g1h2 h7h6 e2e3 f7f5 f2f4 b5c4 d3c4 e5e4 b2g7 g8g7 d5c3 b8d8 d1c2 g6g5 c3a4 d7c7 a4c3');
INSERT INTO evaluations VALUES(9939,30,'cp 282',NULL);
INSERT INTO evaluations VALUES(9939,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(9939,32,'cp 269',NULL);
INSERT INTO evaluations VALUES(9939,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(9939,34,'cp 268',NULL);
INSERT INTO evaluations VALUES(9939,35,'cp 270',NULL);
INSERT INTO evaluations VALUES(9939,36,'cp 249','d4d5 c6e7 e4e5 f6e4 e1g1 d7d6 d1e2 e4c5 e5d6 c7d6 f1e1 c8g4 c1f4 e8f8 b2b4 c5a4 b1d2 e7g6 f4g3 b6d4 a1c1 h7h5 h2h4 d4e5 c4d3 e5g3 f2g3 g6e5 e2e3 g4f3 d2f3 e5f3 e3f3 a4b6 e1f1 f7f6 f3e2 d8e7 e2e7 f8e7 c1c7 b6d7 c7b7 h8e8 f1e1 e7f8');
INSERT INTO evaluations VALUES(9912,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(9912,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(9912,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(9912,33,'cp 281',NULL);
INSERT INTO evaluations VALUES(9912,34,'cp 276',NULL);
INSERT INTO evaluations VALUES(9912,35,'cp 248',NULL);
INSERT INTO evaluations VALUES(9912,36,'cp 304','c6e7 e4e5 f6e4 e1g1 d7d6 d1e2 f7f5 b1c3 e8g8 e5e6 e4f6 b2b4 a7a6 c1g5 h7h6 g5f6 f8f6 a1d1 c7c5 a2a3 c5b4 a3b4 f6f8 e2d2 b6a7 c4b3 b7b5 f3d4 a7d4 d2d4 e7g6 f1e1 d8f6 d4f6 f8f6 c3e2 c8b7 e2d4 g6e5 f2f4');
INSERT INTO evaluations VALUES(18970,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(18970,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18970,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(18970,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(18970,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(18970,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(18970,36,'cp 28','f8e7 b1c3 g8f6 e1g1 d7d6 c1g5 e8g8 g5f6 e7f6 c3d5 a8b8 a2a4 g7g6 c2c3 a7a6 d1d2 c8g4 d2e3 f6g7 f1b1 b7b5 a4b5 a6b5 a1a6 c6d4 c3d4 c5d4 e3d2 b5c4 d3c4 g4e6 b2b4 f7f5 a6a7 f5e4');
INSERT INTO evaluations VALUES(5246,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(5246,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(5246,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(5246,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(5246,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(5246,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(5246,36,'cp 29','b1c3 g8f6 e1g1 e8g8 c1g5 d7d6 g5f6 e7f6 c3d5 g7g6 c2c3 c8e6 d5e3 f6g7 a2a4 a8b8 a1b1 a7a6 c4e6 f7e6 e3c2 a6a5 b2b4 c5b4 c3b4 b7b5 a4b5 b8b5 b4a5 b5b1 d1b1 d8a5 c2e3 f8b8 b1c2 c6d4 c2c4 g8h8 f3d4 e5d4 e3g4');
INSERT INTO evaluations VALUES(29533,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(29533,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(29533,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(29533,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(29533,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(29533,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(29533,36,'cp -10','e7e6 g1f3 f8d6 f1d3 f5d3 d1d3 d6f4 e3f4 d8d6 g2g3 c7c5 b1d2 b8d7 e1g1 e8g8 d3e2 d6b6 f1b1 a7a5 d2b3 c5c4 b3d2 b6a6 a2a4 f8b8 b2b3 c4b3 e2a6 a8a6 b1b3 d7b6 g1f1 b8c8 f3e5 h7h6 f1g2 f6e4 d2e4 d5e4');
INSERT INTO evaluations VALUES(30086,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30086,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(30086,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(30086,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(30086,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(30086,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(30086,36,'cp -4','g1f3 f8d6 f1d3 f5d3 d1d3 e8g8 e1g1 h7h6 h2h3 c7c5 d4c5 d6c5 c3c4 b8d7 c4d5 f6d5 b1c3 d5f4 e3f4 a8c8 g2g3 a7a6 f1c1 d7f6 d3e2 c5b4 f3e5 d8b6 c1c2 b4c3 c2c3 c8c3 b2c3 f8d8 a1c1 b6c7 c3c4 b7b6 c1d1 d8d1 e2d1 b6b5 d1b3');
INSERT INTO evaluations VALUES(30087,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(30087,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(30087,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(30087,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(30087,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30087,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(30087,36,'cp -13','f8d6 f1d3 f5d3 d1d3 e8g8 e1g1 h7h6 f1d1 d6f4 e3f4 d8d6 f4f5 e6f5 d3f5 f8e8 b1d2 b8d7 d1e1 g7g6 f5c2 g8g7 a2a4 a7a6 h2h3 c7c6 d2f1 a6a5 a1d1 e8e1 d1e1 a8e8 e1e8 f6e8 c2e2 d6e6 e2e6 f7e6 b2b4 a5b4');
INSERT INTO evaluations VALUES(28451,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(28451,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(28451,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28451,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28451,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(28451,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28451,36,'cp 0','f4d6 c7d6 d1b3 d8e7 c3c4 d5c4 f1c4 e8g8 c4e2 b8c6 b1c3 a8c8 e1g1 d6d5 f1c1 h7h6 h2h3 c8c7 c3b5 c7c8');
INSERT INTO evaluations VALUES(46648,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46648,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(46648,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(46648,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(46648,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(46648,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(46648,36,'cp 25','f8g7 c4b3 e8g8 g1f3 c8g4 h2h3 g4f3 d1f3 b8c6 f3d1 e7e5 d4e5 d6e5 d1d8 f8d8 c1g5 c6d4 f2f3 h7h6 g5e3 f6h5 e1f2 c7c6 h1d1 h5f4 d1d2 a7a5 h3h4 b7b5 g2g3 f4e6 b3e6 d4e6 a1d1 d8d2 d1d2 a5a4 c3d1 a8d8 d2d8 e6d8 h4h5 g8h7 g3g4 d8e6');
INSERT INTO evaluations VALUES(38420,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(38420,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(38420,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(38420,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(38420,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(38420,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(38420,36,'cp 53','g8f6 b1c3 e8g8 h2h3 a7a6 f1e2 e7e5 d4e5 d6e5 f3e5 f6e4 c3e4 g7e5 c2c3 d8d1 a1d1 c8d7 e1g1 d7a4 d1d2 b8d7 e2f3 a8b8 b2b3 a4c6 f1d1 f7f5 e4g5 c6f3 g5f3 e5d6 c3c4 b8e8 e3h6 f8f7 b3b4 d7e5 f3e5 e8e5 h6f4');
INSERT INTO evaluations VALUES(41386,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(41386,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(41386,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(41386,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(41386,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(41386,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(41386,36,'cp 42','b1c3 e8g8 h2h3 d6d5 e4e5 f6e4 c3e4 d5e4 f3g5 c7c5 c2c3 b8c6 d4c5 c6e5 g5e4 d8a5 d1c2 c8f5 f1e2 e5d7 e2g4 f5e4 c2e4 d7c5 e4c2 c5a4 e1g1 b7b5 f1d1 e7e6 g4f3 a8c8 c2e2 b5b4 c3b4 a5b4 a1b1 b4a5 e2d2 a5d2 d1d2 f8d8 d2d8 c8d8 e3a7 g7b2 b1d1 d8d1 f3d1');
INSERT INTO evaluations VALUES(53076,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53076,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53076,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53076,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53076,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53076,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53076,36,'cp 0','g8f6 b1c3 g7g6 e4e5 d6e5 f3e5 f6d7 e5c4 d7f6');
INSERT INTO evaluations VALUES(45212,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45212,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45212,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45212,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45212,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45212,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45212,36,'cp 0','b1c3 g7g6 d2d4 c5d4 f3d4 b8c6 d4c6 b7c6 f1g2 d8c7 b2b3 f8g7 c1b2 e8g8 e1g1 h7h5 h2h3 f8d8 f1e1 e7e5 d1d2 c8e6 c3a4 f6d7 c2c4 d7b6 a4b6 c7b6 b2c1 b6d4 d2d4 e5d4 c1g5 d8c8 g5f4 c8d8 f4g5');
INSERT INTO evaluations VALUES(45189,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45189,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45189,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45189,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45189,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45189,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45189,36,'cp 0','g7g6 e4e5 d6e5 f3e5 f6d7 e5c4 d7f6');
INSERT INTO evaluations VALUES(12054,30,'cp -265',NULL);
INSERT INTO evaluations VALUES(12054,31,'cp -269',NULL);
INSERT INTO evaluations VALUES(12054,32,'cp -259',NULL);
INSERT INTO evaluations VALUES(12054,33,'cp -271',NULL);
INSERT INTO evaluations VALUES(12054,34,'cp -278',NULL);
INSERT INTO evaluations VALUES(12054,35,'cp -238',NULL);
INSERT INTO evaluations VALUES(12054,36,'cp -279','e5e4 d1e2 c8e6 f3g1 f8c5 b2b4 e6c4 d2d3 c5b4 c1d2 b4d2 e2d2 e4d3 c2d3 d8d3 d2d3 c4d3 b1c3 e8g8 e1c1 d3c4 a4c2 a8b8 g1e2 f6d5 h1e1 f8c8 h2h3 d5b4 c2b1 c4d5 e2f4 g8f8 d1d4 a5c4 e1d1 a7a5 b1f5 c8d8');
INSERT INTO evaluations VALUES(12057,30,'cp -248',NULL);
INSERT INTO evaluations VALUES(12057,31,'cp -265',NULL);
INSERT INTO evaluations VALUES(12057,32,'cp -278',NULL);
INSERT INTO evaluations VALUES(12057,33,'cp -278',NULL);
INSERT INTO evaluations VALUES(12057,34,'cp -278',NULL);
INSERT INTO evaluations VALUES(12057,35,'cp -253',NULL);
INSERT INTO evaluations VALUES(12057,36,'cp -266','f3g1 f8c5 d2d4 e4d3 c2d3 d8b6 g1h3 c8h3 g2h3 c5f2 e1f1 e8g8 f1g2 f2d4 b1c3 c6c5 h1f1 a8d8 d1f3 f6d7 c3d5 b6d6 c1f4 d7e5 f4e5 d4e5 d5c3 d6g6 f3g4 g6d3 a1d1 d3a6 g4e4 e5c3 b2c3 a6c4 e4c4 a5c4 d1e1 d8d3 f1f3 d3d2 f3f2 f8d8 f2d2 d8d2 g2g1 a7a5 e1e7 c4d6 a2a3 d2d3 e7a7 d3c3 a7a5');
INSERT INTO evaluations VALUES(12059,30,'cp -274',NULL);
INSERT INTO evaluations VALUES(12059,31,'cp -267',NULL);
INSERT INTO evaluations VALUES(12059,32,'cp -246',NULL);
INSERT INTO evaluations VALUES(12059,33,'cp -256',NULL);
INSERT INTO evaluations VALUES(12059,34,'cp -251',NULL);
INSERT INTO evaluations VALUES(12059,35,'cp -272',NULL);
INSERT INTO evaluations VALUES(12059,36,'cp -278','f8c5 d2d4 e4d3 c2d3 d8b6 g1h3 c8h3 g2h3 c5f2 e1f1 e8g8 f1g2 f2d4 b1c3 c6c5 h1f1 a5c6 d1b3 c6b4 f1f3 b6a6 b3c4 a6b7 c1d2 a8c8 a1e1 b4c6 c4b3 b7b3 a4b3 c6e5 f3f5 e5d3 e1e2 c5c4 b3c2 d3b2 c3b5 d4c5 d2h6 b2d3 g2f1 f8e8 e2e8 c8e8 c2d3 c4d3 f5c5 g7h6');
INSERT INTO evaluations VALUES(7721,30,'cp -285',NULL);
INSERT INTO evaluations VALUES(7721,31,'cp -296',NULL);
INSERT INTO evaluations VALUES(7721,32,'cp -287',NULL);
INSERT INTO evaluations VALUES(7721,33,'cp -284',NULL);
INSERT INTO evaluations VALUES(7721,34,'cp -276',NULL);
INSERT INTO evaluations VALUES(7721,35,'cp -294',NULL);
INSERT INTO evaluations VALUES(7721,36,'cp -267','d2d4 e4d3 c2d3 d8b6 g1h3 c8h3 g2h3 c5f2 e1f1 e8g8 f1g2 f2d4 b1c3 c6c5 h1f1 a5c6 d1b3 c6b4 f1f3 b6a6 b3c4 a6b7 c1d2 a8c8 a1e1 b4c6 c4b3 b7b3 a2b3 c6e5 f3f1 e5d3 e1b1 c8b8 f1f3 d3e5 f3f5 b8b6 d2e1 e5d3 e1g3 d3b2 b1b2 d4c3 b2e2 c3d4');
INSERT INTO evaluations VALUES(16039,30,'cp -182',NULL);
INSERT INTO evaluations VALUES(16039,31,'cp -146',NULL);
INSERT INTO evaluations VALUES(16039,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(16039,33,'cp -195',NULL);
INSERT INTO evaluations VALUES(16039,34,'cp -180',NULL);
INSERT INTO evaluations VALUES(16039,35,'cp -217',NULL);
INSERT INTO evaluations VALUES(16039,36,'cp -227','e8f7 c3e4 d7d5 e4g3 e5e4 f3g1 h7h5 d2d4 e4d3 d1d3 d8e7 d3e3 h5h4 g3e2 c8f5 c1d2 f5c2 g1f3 e7e3 f2e3 f8e7 e1g1 f7g8 e2f4 c2e4 d2c3 h8h6 f3d2 g7g5 f4h3 e4d3 f1e1 a8d8 h3f2 d3f5 h2h3 h6e6 d2f3 a7a5 f3d4 c6d4 e3d4 e7f6 e1e6 f5e6 f2d3 e6f5 d3e5 a5a4');
INSERT INTO evaluations VALUES(32907,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(32907,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(32907,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(32907,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(32907,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(32907,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(32907,36,'cp 34','e7e6 h2h3 g4h5 d2d4 b8d7 d4d5 e8c8 e1g1 e6d5 c1f4 g8f6 a2a3 h5f3 e2f3 d7b6 d1d3 f8d6 f4g5 h8e8 g5f6 g7f6 b2b4 a5a6 d3a6 b7a6 c3d5 b6d5 f3d5 d6h2 g1h2 d8d5 a1e1 e8e1 f1e1 c8d7 h2g3 d5d2 c2c3 d2a2 e1d1 d7e6 d1d4 a2a3');
INSERT INTO evaluations VALUES(32935,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(32935,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(32935,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(32935,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(32935,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(32935,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(32935,36,'cp 30','d2d4 g8f6 h2h3 g4h5 e1g1 b8d7 a2a3 c7c6 c1f4 f8e7 b2b4 a5d8 c3a4 e8g8 c2c4 d7b6 a4b2 a7a5 d1d2 h7h6 f4e3 d8c7 f1c1 c7d8 c1d1 d8c7 e3f4 c7d8 b4a5 b6d7 a5a6 a8a6 c4c5 a6a8 a3a4 f6d5');
INSERT INTO evaluations VALUES(36667,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(36667,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(36667,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(36667,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(36667,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(36667,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(36667,36,'cp 47','h2h3 b8c6 c1g5 e7e5 d4e5 d6e5 d3c4 d8d1 a1d1 h7h6 g5f6 g7f6 c3d5 f6d8 d1d2 a8b8 e1g1 b7b5 c4b3 f8e8 d5c3 b5b4 c3d5 c6d4 f3d4 e5d4 d2d4 c7c5 d4d2 e8e4 d5b4 d8g5 b4d5 c8e6 f2f4 g5h4 f1d1 g8g7 c2c3 c5c4 b3c2 e4e1 d1e1 h4e1');
INSERT INTO evaluations VALUES(36665,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36665,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36665,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(36665,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36665,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(36665,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(36665,36,'cp 35','b8c6 c1g5 e7e5 d4e5 d6e5 d3c4 d8d1 a1d1 h7h6 g5f6 g7f6 c3d5 f6d8 e1g1 g8g7 d1d2 f7f5 d5c3 d8f6 f1e1 h6h5 h3h4 a8b8 c4b3 f8e8 c3d5 f6d8 e4f5 c8f5 f3g5 e5e4 a2a4 b7b6 d5f4 d8g5 h4g5 c6e5 b3d5');
INSERT INTO evaluations VALUES(21666,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(21666,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(21666,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(21666,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(21666,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(21666,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(21666,36,'cp 163','e4d5 c6d4 f3d4 e5d4 e1g1 f8d6 f1e1 e8f8 d2d3 h7h5 d1e2 c8g4 e2e4 d8f6 b1d2 g8e7 d2b3 g4d7 b3d4 a8e8 c1e3 e7f5 e4f3 f5e3 f3f6 d6h2 g1h1 g7f6 h1h2 e3g4 h2g1 g4e5 f2f3 h5h4 g1f2 h4h3 e1h1 e5g6 g2g3 h3h2 b2b4 h8h3 d4b5 d7b5 c4b5 e8d8 b5c4 f8g7');
INSERT INTO evaluations VALUES(21567,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(21567,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(21567,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(21567,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(21567,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(21567,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(21567,36,'cp 169','c6d4 f3d4 e5d4 e1g1 f8d6 f1e1 e8f8 d2d3 h7h5 d1e2 c8g4 e2e4 c7c5 c1f4 g8f6 f4d6 d8d6 e4e5 d6e5 e1e5 h5h4 b1d2 h8h5 e5h5 g4h5 d2e4 f6e4 d3e4 g7g5 d5d6 f7f6 e4e5 f6e5 a1e1 h5f7 c4f7');
INSERT INTO evaluations VALUES(22169,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(22169,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(22169,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(22169,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(22169,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(22169,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(22169,36,'cp 194','f3d4 e5d4 e1g1 f8d6 f1e1 e8f8 d2d3 h7h5 b1d2 c8g4 f2f3 d8h4 d2f1 g4f5 f1g3 f5d7 c1d2 g8h6 d1c1 d6g3 h2g3 h4g3 d2f4 g3g6 f4c7 h6f5 c1d2 f8g8 e1e4 h5h4 c7f4 b7b5 c4b3 f5g3 g1h2 g6f6 c2c4 h8h5 c4b5 g3e4 f3e4');
INSERT INTO evaluations VALUES(22168,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(22168,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(22168,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(22168,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(22168,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(22168,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(22168,36,'cp 172','e5d4 e1g1 f8d6 f1e1 e8f8 d2d3 h7h5 b1d2 c8g4 f2f3 d8h4 d2f1 g4f5 f1g3 g8h6 d1d2 f5d7 d2g5 h4g5 c1g5 f7f6 g5d2 h6f5 g3e4 a7a5 a2a4 b7b6 e4d6 f5d6 b2b4 d6c4 d3c4 f8f7 b4a5 b6a5 e1b1 h8c8 c4c5 d7f5 b1b7 f5c2 d5d6 f7e6 b7c7 c8c7 d6c7');
INSERT INTO evaluations VALUES(22159,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(22159,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(22159,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(22159,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(22159,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(22159,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(22159,36,'cp 165','e1g1 f8d6 f1e1 g8e7 d2d3 e8g8 b1d2 h7h6 h2h3 c8d7 d2e4 e7g6 c1d2 f8e8 a2a4 d8h4 c4b5 h4d8 e4d6 c7d6 b5d7 d8d7 e1e4 e8e4 d3e4 a8e8 d1f3 g6e5 f3b3 d7c7 b3b4 c7c2 b4d4 e5d3 d2c3 f7f6 b2b3 e8e5');
INSERT INTO evaluations VALUES(22701,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(22701,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(22701,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(22701,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(22701,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(22701,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(22701,36,'cp 148','b4b5 c6a5 b2e5 a5c4 e5d4 a7a6 a2a4 d7d5 g1f3 g8e7 d2d3 c4b6 b1d2 e7f5 d4b2 a6b5 a4b5 b6a4 d1c1 d5d4 g2g3 c8e6 g3g4 f5d6 f3d4 e6g4 h1g1 g4d7 e2e3 a4b2 c1b2 a8a1 b2a1 f8e7 g1g7 e7f6 g7g1 d6b5 d2e4 b5d4 e4f6 d8f6 a1d4 f6d4 e3d4 e8d8 g1g3 b7b5 c2c3 c7c6 f1e2 d8c7');
INSERT INTO evaluations VALUES(22653,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(22653,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(22653,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(22653,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(22653,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(22653,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(22653,36,'cp 168','c6a5 b2e5 a5c4 e5d4 d7d5 a2a4 a7a6 g1f3 g8e7 d2d3 c4b6 b1d2 e7f5 d4b2 a6b5 a4b5 b6a4 d1c1 c8d7 e2e4 f5h4 f3h4 d8h4 g2g3 h4h6 f1g2 f8b4 c2c3 b4e7 e4d5 e8g8 e1g1 e7g5 c1c2 a4b2 d2e4 b2d3 c2d3 g5e7 a1a8 f8a8 d5d6 e7d6 c3c4 a8a2');
INSERT INTO evaluations VALUES(55002,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(55002,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(55002,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(55002,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55002,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(55002,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55002,36,'cp 11','e7e5 f1g2 c7c5 c2c4 g7g6 b1c3 b8c6 a2a3 f8g7 e1g1 g8e7 b2b4 e8g8 a1b1 f7f5 b4b5 c6d4 f3d4 c5d4 c3d5 c8e6 a3a4 e7d5 c4d5 e6f7 d2d3 h7h5 e2e3 d4e3 f2e3 a8c8 c1d2 g7h6 e3e4 h6d2 d1d2');
INSERT INTO evaluations VALUES(55648,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(55648,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(55648,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(55648,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(55648,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(55648,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(55648,36,'cp 40','c2c4 c7c5 f1g2 g7g6 b1c3 f8g7 a2a3 a7a5 e1g1 b8c6 d2d3 g8e7 f3e1 c8e6 c1g5 f7f6 g5d2 e8g8 c3d5 f6f5 e1c2 e6f7 d2g5 h7h6 d5e7 c6e7 g5e7 d8e7 c2e3 g8h7 e3d5 f7d5 g2d5 h6h5 a1b1 a8a7 e2e3 b7b6 g1g2 h5h4 b2b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(16215,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(16215,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(16215,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(16215,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(16215,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(16215,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(16215,36,'cp 153','d7d6 e4e5 f6d7 e5d6 f8d6 d4g7 d8e7 c4e2 d6e5 g7g4 d7f6 g4c4 c8f5 c1f4 e8c8 f4e5 e7e5 e1g1 f5e6 c4b5 e5b5 e2b5 c7c6 b5d3 a7a5 f2f3 c8c7 f1e1 b7b5 a2a3 f6d7 b2b4 d8a8 g2g4 h8d8 g1f2 h6h5 h2h3 h5g4 f3g4 d7b6 a1d1 a5b4 a3b4 b6a4');
INSERT INTO evaluations VALUES(21511,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(21511,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(21511,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(21511,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(21511,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(21511,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(21511,36,'cp 86','g1f3 g8f6 f1d3 c6b4 e4f6 g7f6 d3e2 b7b6 c2c3 b4d5 d1a4 c8d7 a4c2 f8g7 e1g1 d7c6 c3c4 d5e7 b2b4 e8g8 c1b2 f6f5 d4d5 g7b2 c2b2 e6d5 a1c1 f8e8 f1d1 d8d6 b4b5 c6b7 f3d4 a8d8 b2c3 c7c6 b5c6 b7c6 d4c6 e7c6 d1d5');
INSERT INTO evaluations VALUES(52685,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(52685,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(52685,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(52685,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(52685,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(52685,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(52685,36,'cp 75','d2d4 c5d4 f3d4 e7e6 b1c3 g8f6 c1e3 a7a6 a2a3 b7b5 c4a2 c8b7 f2f3 d6d5 e4d5 f6d5 a2d5 b7d5 d4f5 d5c6 d1d8 e8d8 e1c1 d8c8 f5d6 f8d6 d1d6 h8d8 h1d1 d8d6 d1d6 c8b7 e3f2 b8d7 b2b3 a8d8 f2g3 e6e5 c1b2 b7c7');
INSERT INTO evaluations VALUES(52686,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(52686,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(52686,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(52686,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(52686,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(52686,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(52686,36,'cp 65','c5d4 f3d4 g8f6 b1c3 a7a6 c1e3 e7e6 c4b3 b7b5 a2a3 d8c7 f2f3 b8d7 d1d2 a8b8 e1c1 d7c5 g2g4 a6a5 c1b1 c8d7 h2h4 b5b4 a3b4 a5b4 c3a2 c5b3 d4b3 c7b7 g4g5 f6h5 g5h6 g7h6 e4e5 h5g3 h1g1 g3f5 e5d6 f8d6');
INSERT INTO evaluations VALUES(52688,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(52688,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(52688,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(52688,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(52688,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(52688,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(52688,36,'cp 73','f3d4 g8f6 b1c3 a7a6 e1g1 e7e6 f2f4 b8d7 f1e1 d8a5 d1e2 f8e7 g1h1 a5h5 e2h5 f6h5 f4f5 e6e5 d4b3 b7b5 c4d5 a8b8 b3a5 h5f6 a5c6 b8b6 c6b4 d7c5 c1e3 c8d7 a2a4 e8g8 a4b5 d7b5 c3b5');
INSERT INTO evaluations VALUES(52687,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(52687,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(52687,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(52687,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(52687,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(52687,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(52687,36,'cp 50','g8f6 b1c3 e7e6 c1e3 a7a6 a2a3 d8c7 c4b3 b7b5 h2h4 b8d7 f2f3 d7c5 b3a2 a8b8 d1d2 e6e5 d4e2 c8e6 g2g4 f8e7 e2g3 c5a4 g4g5 h6g5 h4g5 h8h1 g3h1 f6h7 g5g6 h7f6 g6f7 e8f7 a2e6 f7e6 c3a4 b5a4 e1c1 c7c4 h1g3 d6d5 g3f5 d5d4 f5e7 e6e7');
INSERT INTO evaluations VALUES(40275,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(40275,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(40275,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(40275,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(40275,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(40275,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(40275,36,'cp 37','f1e2 e8g8 e1g1 f8e8 a2a4 a7a5 c4d5 c6d5 c1d2 f6e4 d2e1 e4c3 e1c3 b8c6 h2h3 c8f5 d1d2 d8d6 e2b5 e8c8 f1c1 b7b6 d2e2 e7e6 b5a6 c8d8 g2g4 f5e4 f3d2 c6b4 d2e4 d5e4 c3b4 a5b4');
INSERT INTO evaluations VALUES(40274,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(40274,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(40274,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(40274,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(40274,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(40274,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(40274,36,'cp 39','e8g8 e1g1 f8e8 a2a4 a7a5 c4d5 c6d5 c1d2 f6e4 d2e1 e4c3 e1c3 b8c6 d1b3 e7e6 h2h3 g7f8 e2b5 c8d7 f1c1 f8d6 f3e5 e8e7 e5g4 h7h5 g4f6 g8g7 e3e4 g7f6 e4e5 d6e5 d4e5 f6g7 b5c6 d7c6 c3d4 a8c8 b3g3 e7c7 d4b6 d5d4');
INSERT INTO evaluations VALUES(35873,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(35873,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(35873,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(35873,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(35873,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(35873,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(35873,36,'cp 34','e1g1 f8e8 a2a4 a7a5 f3e5 e7e6 b2b3 b8a6 c1a3 a6b4 a1c1 f6d7 f2f4 d7e5 f4e5 f7f6 e5f6 g7f6 d1e1 f6h4 e1d2 h4g5 a3b2 c8d7 f1f2 d8e7 c1f1 e8f8 c3d1 b7b6 f2f8 a8f8 d1f2 h7h5 b2c3 d7e8');
INSERT INTO evaluations VALUES(35872,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(35872,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(35872,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(35872,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(35872,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(35872,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35872,36,'cp 34','f8e8 a2a4 a7a5 c4d5 c6d5 c1d2 f6e4 d2e1 e4c3 e1c3 b8c6 d1b3 e7e6 f1c1 g7f8 h2h3 e8e7 c3d2 h7h5 e2d3 e7c7 c1c3 c6b4 c3c7 d8c7 a1c1 c7d8 d3b5 c8d7 b5d7 d8d7 d2b4 f8b4 f3e5 d7d8 e5d3 d8e7 d3b4 e7b4 b3b4 a5b4');
INSERT INTO evaluations VALUES(40765,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(40765,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40765,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(40765,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(40765,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40765,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40765,36,'cp 22','e8g8 e1g1 h7h6 h2h3 c7c5 a2a4 b8c6 c1d2 a7a6 f1e1 a8b8 c3d5 b7b5 a4b5 a6b5 c4a2 f6d5 e4d5 c6d4 b2b4 d8c7 b4c5 c7c5 f3d4 c5d4 a2b3 c8f5 d2e3 d4h4 a1a7 e7g5 a7a6 h4b4 e3g5 h6g5 a6c6 f8c8');
INSERT INTO evaluations VALUES(36221,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36221,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36221,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36221,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36221,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(36221,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(36221,36,'cp 18','e1g1 c8e6 a2a4 h7h6 h2h3 d8d7 f1e1 c7c5 c3d5 e7d8 c2c3 b8c6 d3d4 e5d4 c3d4 e6d5 e4d5 c6d4 f3d4 c5d4 d1d4 d8b6 d4d3 f8e8 c1d2 e8e1 d2e1 a7a6 b2b4 d7e7 a4a5 b6a7 e1c3 f6e4 c3d4 a7d4 d3d4 a8e8 a1c1 e7e5 d4e5 e8e5 b4b5 a6b5');
INSERT INTO evaluations VALUES(36220,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36220,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36220,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36220,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36220,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(36220,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36220,36,'cp 13','c8e6 h2h3 c7c5 c1g5 b8c6 g5f6 e7f6 c3d5 g7g6 c2c3 f6g7 a2a4 g8h8 d1d2 f7f5 d2g5 d8d7 a4a5 e6d5 c4d5 g7f6 g5d2 c6e7 d5b3 d6d5 f1e1 d7d6 b3a4 a8d8 d2e2 f5e4 d3e4 d5d4 c3d4 e5d4');
INSERT INTO evaluations VALUES(27196,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(27196,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(27196,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(27196,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(27196,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(27196,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(27196,36,'cp 30','d6e5 e1g1 g8f6 f1e1 f8d6 b1d2 d8e7 c2c3 h7h6 d2c4 e8g8 b5a4 f8d8 d1e2 a7a6 a4c2 f6h5 c1e3 d7e6 c4d6 c7d6 h2h3 h5f4 e3f4 e5f4 c2b3 e6b3 a2b3 d8e8 b3b4 a8d8 b4b5 a6b5 e2b5 c6e5 b5e2 e7e6');
INSERT INTO evaluations VALUES(27192,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(27192,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(27192,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(27192,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(27192,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(27192,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(27192,36,'cp 22','e1g1 g8f6 f1e1 f8c5 c1g5 h7h6 g5h4 c6e7 h4f6 g7f6 d1d7 d8d7 b5d7 e8d7 g2g3 a7a5 a2a4 d7e6 b1d2 h6h5 d2b3 c5d6 g1f1 e7g6 b3d2 h5h4 d2c4 d6c5 c2c3 b7b6 a1b1 h8d8 e1d1 d8d1 b1d1 a8h8 f1e2 h4g3 h2g3 g6e7 f3h4');
INSERT INTO evaluations VALUES(27191,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(27191,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(27191,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(27191,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(27191,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(27191,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(27191,36,'cp 33','g8f6 f1e1 f8d6 c1g5 h7h6 g5h4 a7a6 b5c6 d7c6 c2c4 d6b4 d1d8 a8d8 b1c3 b4c3 b2c3 g7g5 h4g3 e8g8 g3e5 f6e4 f3d4 f7f6 e5c7 d8c8 c7b6 f8e8 f2f3 e4d2 d4c6 e8e1 a1e1 c8c6 c4c5 d2c4 b6d8 g8f8 e1e4 c4e5 d8f6 e5f3 g2f3 c6f6 g1f2');
INSERT INTO evaluations VALUES(49679,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(49679,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(49679,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(49679,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(49679,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(49679,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(49679,36,'cp 23','f4e5 f6e4 g1f3 f8e7 d2d3 e4g5 d3d4 e8g8 f1d3 g5f3 d1f3 d7d6 b1c3 b8c6 c1f4 c6b4 e1c1 b4d3 d1d3 c8f5 e5d6 c7d6 d3e3 d8d7 f3d5 e7f6 c3e4 f6e7 e4c5 d6c5 d5d7 f5d7 e3e7 d7c6 d4c5 f8e8 h1e1 e8e7 e1e7 c6g2 f4d6 g2c6 e7c7 a8e8 b2b4 a7a6 a2a4');
INSERT INTO evaluations VALUES(49506,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(49506,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(49506,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49506,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49506,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49506,36,'cp 0','f6e4 g1f3 f8e7 d2d3 e4g5 d3d4 e8g8 f1d3 g5f3 d1f3 d7d6 b1c3 b8c6 c1f4 c6b4 e1c1 b4d3 d1d3 c8f5 d3d2 d6e5 d4e5 d8c8 c3d5 e7h4 g2g3 f5g4 f3e4 g4f5');
INSERT INTO evaluations VALUES(49864,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(49864,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49864,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(49864,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(49864,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49864,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(49864,36,'cp 16','g1f3 f8e7 d2d3 e4g5 d3d4 e8g8 f1d3 g5f3 d1f3 d7d6 c1f4 b8c6 b1c3 c6b4 e1c1 b4d3 d1d3 c8f5 d3d2 d6e5 d4e5 d8c8 c3d5 e7h4 g2g3 f5g4 f3e4 h4d8 d5e3 g4d7 h2h4 d7c6 e3d5 c8e6 h1d1 d8e7 d5e7 e6e7 e4d4 h7h6 b2b3 f8d8 d4c4 d8d2 d1d2 a7a6 a2a4 a8d8 d2d8 e7d8 e5e6 f7e6 c4e6 g8h8 f4e5');
INSERT INTO evaluations VALUES(49859,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49859,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49859,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(49859,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49859,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49859,36,'cp 0','f8e7 d2d3 e4g5 d3d4 e8g8 f1d3 g5f3 d1f3 d7d6 c1f4 b8c6 b1c3 c6b4 e1c1 b4d3 d1d3 c8f5 d3d2 d6e5 d4e5 d8c8 c3d5 e7h4 g2g3 f5g4 f3e4 g4f5');
INSERT INTO evaluations VALUES(23444,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(23444,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(23444,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(23444,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(23444,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(23444,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(23444,36,'cp -24','d7d5 d2d3 g8f6 b1d2 e7e5 g1f3 f8d6 f1e2 e8g8 e1g1 f8e8 f1e1 a7a5 e2f1 c8g4 e4d5 f6d5 d2e4 h7h6 h2h3 g4h5 e4g3 h5g6 d1b3 d5b6 a2a4 g8h8 g3e4 f7f5 e4d6 d8d6 f3h4 g6h7 c1e3 b6d5 b3b5 a8d8 e3c5 d6f6 h4f3');
INSERT INTO evaluations VALUES(21049,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(21049,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(21049,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(21049,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(21049,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(21049,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(21049,36,'cp -24','d2d3 g8f6 b1d2 e7e5 g1f3 a7a5 e4d5 d8d5 f1e2 f8c5 e1g1 e8g8 d2c4 f8d8 c1g5 e5e4 d3e4 d5e4 c4d2 e4g6 g5f6 g7f6 d1a4 c6e5 a4h4 e5f3 d2f3 c8f5 f1d1 f5c2 d1e1 c2e4 h4h3 f6f5 a1d1 d8d1 e2d1 a8e8 f3d2 g8g7 d1f3 g6g5');
INSERT INTO evaluations VALUES(51482,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(51482,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(51482,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(51482,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(51482,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(51482,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(51482,36,'cp 57','c8b7 d2d4 e7e6 e2e4 f8b4 f2f3 b4c3 b2c3 d8h4 g2g3 h4e7 f1d3 d7d6 g1e2 g8f6 e1g1 b8c6 c1e3 e6e5 e2c1 f6d7 c1b3 c6d8 f1f2 e8g8 d3f1 a7a5 b3d2 a5a4 f1h3 g8h8 f3f4 f8e8 h3f5 e5f4 e3f4 d7f6');
INSERT INTO evaluations VALUES(30783,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(30783,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(30783,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(30783,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(30783,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(30783,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(30783,36,'cp 37','d2d4 e7e6 f2f3 f8b4 e2e4 b4c3 b2c3 d8h4 g2g3 h4e7 g1h3 d7d6 f1d3 g8f6 e1g1 b8c6 f3f4 e8c8 d1e2 f6d7 c1a3 e7e8 a3b4 h7h5 a2a4 a7a5 b4a3 f7f6 a1b1 c8b8 b1b2 e6e5 f4e5 c6e5 d4e5 d7e5 d3c2 e8c6 h3f4 e5c4');
INSERT INTO evaluations VALUES(46350,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(46350,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46350,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(46350,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(46350,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(46350,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46350,36,'cp 32','c1f4 f8d6 f4g3 f6e4 e2e3 e4g3 h2g3 c8d7 f1d3 d7c6 g3g4 b8d7 a1c1 a8c8 e1f1 g7g6 e3e4 d5e4 d3e4 c6e4 c3e4 d6e7 g2g3 c8c1 d1c1 d7f6 e4f6 e7f6 c1c5 d8b6 c5b6 a7b6 f1e2 h7h5 g4g5 f6g7 e2d3 f7f6 h1d1 e8d7 d1c1 f6g5 f3g5');
INSERT INTO evaluations VALUES(46333,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(46333,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(46333,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46333,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46333,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(46333,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(46333,36,'cp 15','f8d6 f4d6 d8d6 e2e3 e8g8 f1d3 b8c6 e1g1 c8d7 a1c1 f8c8 h2h3 h7h6 a2a3 d7e8 f3d2 c6e7 d1e2 c8c7 c1c2 d6d8 f1c1 a7a6 d2b3 a6a5 a3a4 b7b6 b3d2 e7c6 c3b5 c7c8 d2b1 c6b4 c2c8 a8c8 b1c3 e8b5 d3b5 f6e4');
INSERT INTO evaluations VALUES(40577,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(40577,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(40577,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40577,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40577,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40577,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(40577,36,'cp 18','f4d6 d8d6 e2e3 c8d7 f1d3 b8c6 e1g1 e8g8 a1c1 f8c8 f3d2 c6a5 d1e2 d6b6 d2f3 a5c4 c1c2 c4d6 f3e5 d7e8 f1c1 c8c7 g2g4 a8c8 g4g5 f6d7 e5d7 e8d7 h2h4 c7c6 e2g4 b6d8 g4f4 d7e8 g1g2 f7f6 g5f6 d8f6 f4f6 g7f6');
INSERT INTO evaluations VALUES(22677,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(22677,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(22677,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(22677,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(22677,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(22677,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(22677,36,'cp 126','d7d6 g1f3 c8g4 b1d2 d6e5 f4e5 g8f6 c2c3 c6e5 f3e5 g4e6 e2e3 f8d6 d2c4 d6e5 d1d8 a8d8 c4e5 e8e7 f1c4 f6d7 c4e6 d7e5 e6f5 g7g6 f5c2 e5c4 b2b3 c4a3 a1c1 a3c2 c1c2 d8d3 e1e2 h8d8 g2g4 b7b6 h2h4 e7f6 b3b4 f6e5 h4h5 g6g5 c3c4 f7f5 g4f5 e5f5 c4c5 b6b5 h5h6 d3a3');
INSERT INTO evaluations VALUES(21886,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(21886,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(21886,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(21886,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(21886,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(21886,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(21886,36,'cp 140','g1f3 c8g4 b1d2 d6e5 f4e5 g8f6 c2c3 c6e5 f3e5 g4e6 e2e3 f8d6 e5c4 e8g8 f1e2 d6e7 d2f3 g7g6 d1d8 f8d8 f3d4 e6d7 a2a4 g8g7 e1g1 c7c5 d4b3 d7e6 f1d1 d8d1 a1d1 b7b6 f2f3 a8d8 b3d2 h7h5 e3e4 h5h4 g1f2 f6h5 g2g3 f7f5 f3f4 h4g3 h2g3');
INSERT INTO evaluations VALUES(21885,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(21885,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(21885,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(21885,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(21885,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(21885,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(21885,36,'cp 131','c8g4 b1d2 d6e5 f4e5 g8f6 c2c3 c6e5 f3e5 g4e6 e2e3 f8d6 e5c4 d6c5 d2b3 d8e7 f1e2 a8d8 d1c2 g7g6 c4a5 e6f5 c2c1 f5e4 e1g1 c5b6 a5c4 c7c6 c4b6 a7b6 a2a4 e8g8 b3d2 e4d3 e2d3 d8d3 c1c2');
INSERT INTO evaluations VALUES(28739,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(28739,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(28739,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(28739,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(28739,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(28739,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(28739,36,'cp 24','f1d3 c7c5 e1g1 f8e7 h2h3 d7d5 c2c3 c5c4 d3c2 b6b5 f3e5 e8g8 b1d2 a7a5 a2a3 b5b4 a3b4 a5b4 a1a8 b7a8 c3b4 e7b4 d2f3 a8c6 g2g4 c6b5 d1b1 b8d7 g4g5 d7e5 f4e5 f6d7 c2h7 g8h8 e5f4 c4c3 b2c3 b5f1 c3b4 f1h3 b4b5 h3g4');
INSERT INTO evaluations VALUES(28735,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(28735,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(28735,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(28735,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(28735,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(28735,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(28735,36,'cp 16','f8e7 h2h3 d7d5 c2c3 e8g8 b1d2 c7c5 e1g1 e7d6 d1e2 d8c7 f4d6 c7d6 e3e4 d5e4 d2e4 b7e4 d3e4 f6e4 e2e4 b8d7 a1d1 d6c7 e4e3 h7h6 c3c4 a8d8 b2b3 f8e8 f1e1 c5d4 f3d4 a7a6 e3f3 d7f6 d4c6 d8d6 d1d6 c7d6');
INSERT INTO evaluations VALUES(28313,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(28313,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(28313,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28313,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(28313,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28313,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28313,36,'cp 18','h2h3 e8g8 b1d2 c7c5 e1g1 d7d5 c2c3 e7d6 d1e2 d8c7 f4d6 c7d6 e3e4 d5e4 d2e4 f6e4 d3e4 b7e4 e2e4 b8d7 a1d1 d6c7 c3c4 h7h6 b2b3 f8d8 e4e2 a7a5 d1d2 a5a4 f1d1 a4b3 a2b3 d7f6 e2e5 a8c8 d4c5 d8d2 d1d2');
INSERT INTO evaluations VALUES(57299,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(57299,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(57299,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(57299,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(57299,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(57299,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(57299,36,'cp -33','g8f6 g1f3 d7d5 d4d1 b8c6 f1e2 c8f5 e1g1 f8d6 b1a3 e8g8 a3b5 d6e7 b5d4 c6d4 e3d4 c7c6 e2d3 f5g4 f1e1 f8e8 h2h3 g4h5 c1f4 e7d6 e1e8 f6e8 f4e5 d8c7 e5d6 c7d6 d3e2 d6f6 d1d2 h5f3 e2f3 g7g6 a1e1 g8g7');
INSERT INTO evaluations VALUES(49766,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(49766,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(49766,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(49766,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(49766,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(49766,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(49766,36,'cp -36','g1f3 d7d5 d4d1 b8c6 f1e2 c8f5 e1g1 f8d6 b1a3 e8g8 a3b5 d6e7 b5d4 c6d4 e3d4 c7c6 e2d3 f5d3 d1d3 f6e4 c2c4 f8e8 c1f4 a7a5 f3e5 a5a4 a1e1 f7f6 e5g4 e7b4 e1e2 d8d7 g4e3 b7b5 c4b5 c6b5 h2h3');
INSERT INTO evaluations VALUES(7038,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(7038,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(7038,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(7038,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(7038,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(7038,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(7038,36,'cp 98','b1c3 d7d6 c1e3 g8f6 d1d2 f6g4 e3g5 h7h6 g5h4 a7a6 e1c1 g4f6 h2h3 b7b5 f1d3 f6d7 h1e1 e8g8 c3d5 d7b6 g2g4 a8b8 d5b6 b8b6 d4d5 c6a7 d2e3 b5b4 e4e5 d6e5 f3e5 d8d5 h4e7 f8e8 d3e4');
INSERT INTO evaluations VALUES(7034,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(7034,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(7034,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(7034,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(7034,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(7034,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(7034,36,'cp 95','a7a6 c1e3 d7d6 d1d2 g8f6 d4d5 c6b8 e3h6 g7h6 d2h6 e7e5 h2h3 b8d7 e1c1 d8e7 f1e2 b7b5 a2a3 d7b6 f3d2 f6d7 g2g4 c8b7 d2b3 e8c8 b3a5 e7f6 h6e3 f6f4 h3h4 d8f8 d1f1 f7f6 h4h5 d7c5 h1h3 c8b8 b2b4 f4e3 f2e3 c5d7 a5c6 b7c6 d5c6');
INSERT INTO evaluations VALUES(33195,30,'cp 311',NULL);
INSERT INTO evaluations VALUES(33195,31,'cp 324',NULL);
INSERT INTO evaluations VALUES(33195,32,'cp 326',NULL);
INSERT INTO evaluations VALUES(33195,33,'cp 331',NULL);
INSERT INTO evaluations VALUES(33195,34,'cp 323',NULL);
INSERT INTO evaluations VALUES(33195,35,'cp 322',NULL);
INSERT INTO evaluations VALUES(33195,36,'cp 341','d2d4 d7d6 e5f3 c5b6 a2a4 a7a6 c1e3 g8e7 f1d3 e8g8 e1g1 h7h6 b1d2 e7g6 b2b4 b8c6 c2c3 f6d8 d1b3 c6e7 c3c4 b6a7 f1e1 c7c6 g2g3 c8e6 d4d5 a7e3 e1e3 e6g4 a1e1 g6e5 f3e5 d6e5');
INSERT INTO evaluations VALUES(33194,30,'cp 335',NULL);
INSERT INTO evaluations VALUES(33194,31,'cp 327',NULL);
INSERT INTO evaluations VALUES(33194,32,'cp 333',NULL);
INSERT INTO evaluations VALUES(33194,33,'cp 321',NULL);
INSERT INTO evaluations VALUES(33194,34,'cp 322',NULL);
INSERT INTO evaluations VALUES(33194,35,'cp 325',NULL);
INSERT INTO evaluations VALUES(33194,36,'cp 319','d7d6 e5f3 c5b6 a2a4 a7a6 c1e3 g8e7 b1d2 h7h6 c2c3 b8d7 a4a5 b6a7 f1d3 e7g6 e1g1 e8g8 f1e1 f8e8 h2h3 d7f8 b2b4 c8d7 d2f1 f8e6 g2g3 c7c6 h3h4 e6c7 h4h5 g6f8 f1h2 c7b5 d1d2 a8c8 e4e5 d6e5 d4e5 a7e3 e1e3 f6e7 a1e1 d7e6');
INSERT INTO evaluations VALUES(43203,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43203,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43203,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43203,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43203,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43203,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43203,36,'cp 0','e8g8 g1f3 d7d6 h2h3 f6d7 f1e2 b8c6 e1g1 e7e5 f4h2 f7f5 a2a4 g8h8 d4e5 d7e5 b1d2 d8e7 f3e5 c6e5 d2f3 a7a5 d1c2 b7b6 f3d4 c8b7 a1d1 e7g5 h2g3 g5e7');
INSERT INTO evaluations VALUES(37424,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37424,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37424,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37424,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37424,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37424,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37424,36,'cp 0','g1f3 d7d6 h2h3 b8c6 f4h2 f6d7 a2a4 e7e5 f1e2 f7f5 d4e5 d7e5 b1d2 d8e7 e1g1 g8h8 f3e5 c6e5 d2f3 a7a5 f3d4 b7b6 d1c2 c8b7 a1d1 e7g5 h2g3 g5e7');
INSERT INTO evaluations VALUES(37425,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37425,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37425,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37425,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37425,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37425,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37425,36,'cp 0','d7d6 h2h3 b8c6 f1e2 f6d7 f4h2 e7e5 a2a4 f7f5 e1g1 g8h8 d4e5 d7e5 b1d2 d8e7 f3e5 c6e5 d2f3 a7a5 d1c2 b7b6 f3d4 c8b7 a1d1 e7g5 h2g3 g5e7');
INSERT INTO evaluations VALUES(36606,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36606,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36606,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36606,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36606,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36606,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36606,36,'cp 0','h2h3 b8c6 f4h2 f6d7 f1e2 e7e5 e1g1 f7f5 a2a4 g8h8 d4e5 d7e5 b1d2 d8e7 f3e5 c6e5 d2f3 c8e6 f3d4 e6d7 d4b5 d7e6');
INSERT INTO evaluations VALUES(36603,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36603,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36603,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36603,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36603,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36603,36,'cp 0','f6d7 f4h2 e7e5 f1e2 b8c6 a2a4 g8h8 d4e5 d7e5 b1d2 d8e7 e1g1 f7f5 f3e5 c6e5 d2f3 c8e6 f3d4 e6d7 d4b5 d7e6');
INSERT INTO evaluations VALUES(48167,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(48167,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(48167,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(48167,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(48167,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(48167,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(48167,36,'cp -46','b7b6 c5b6 a7b6 g1f3 f8b4 c1d2 b4d6 a2a4 e8g8 f1e2 c7c5 b1c3 b8c6 c3b5 d6b8 e1g1 f6e4 d2e1 d8e7 a1c1 c8b7 f3d2 c5d4 b5d4 c6d4 e3d4 b8f4 c1a1 f8c8 d2f3 b7c6 f3e5 c6e8 e2d3 e7f6 f2f3 e4d6 e1g3');
INSERT INTO evaluations VALUES(43789,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(43789,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(43789,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(43789,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(43789,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(43789,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(43789,36,'cp -37','c5b6 a7b6 g1f3 f8b4 c1d2 b4d6 f1d3 c7c5 e1g1 b8c6 b1c3 e8g8 c3b5 d6b8 d2c3 f6e4 d3e4 d5e4 f3e5 b8e5 d4e5 c8a6 a2a4 a6b5 a4b5 a8a1 d1a1 c6e7 a1a6 e7d5 a6a7 d8g5 f1a1 h7h6 h2h3 g5h5 a7a4 d5c3 b2c3 h5e5 a4a7 f8b8 c3c4');
INSERT INTO evaluations VALUES(54279,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(54279,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(54279,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(54279,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(54279,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(54279,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(54279,36,'cp -76','c5d4 g1f3 b8c6 c1f4 h7h6 f3d4 g8e7 d4c6 e7c6 f1e2 g7g5 f4g3 f8g7 c4d5 e6d5 h2h4 c6e5 h4g5 h6g5 h1h8 g7h8 b1c3 c8e6 c3b5 a7a6 b5d4 e6d7 a1c1 a8c8 c1c8 d8c8 d4b3 d7e6 b3d4 e5g6 d4e6 c8e6 d1a4 e8f8');
INSERT INTO evaluations VALUES(54406,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(54406,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(54406,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(54406,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(54406,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(54406,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(54406,36,'cp -95','g1f3 b8c6 f3d4 g8e7 d4c6 e7c6 c4d5 e6d5 f1b5 d8b6 b5c6 b7c6 e1g1 c8a6 f1e1 f8c5 d1f3 e8g8 b1c3 a8e8 c1d2 c5d4 c3a4 b6c7 d2c3 d4e5 a4c5 a6c4 c3e5 e8e5 f3c3 f8e8 h2h3 e5e1 a1e1 e8e1 c3e1 c7d6 e1e8 d6f8 e8c6 c4a2');
INSERT INTO evaluations VALUES(57048,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(57048,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(57048,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(57048,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(57048,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(57048,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(57048,36,'cp -15','d7d5 e4d5 e6d5 d2d4 g8f6 g1f3 f8d6 c2c4 c7c6 c4d5 f6d5 f1d3 d8e7 d3e2 b8d7 e1g1 d7f6 b1d2 e8g8 f1e1 c8g4 h2h3 g4h5 d2c4 a8e8 c1d2 f6e4 f3e5 e4d2 e2h5 d2c4 e5c4 e7f6 e1e8 f8e8 g2g3 e8e7 d1f3 d6c7 a1d1 h7h6 f3f6');
INSERT INTO evaluations VALUES(55919,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55919,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(55919,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(55919,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(55919,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(55919,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(55919,36,'cp -17','e4d5 e6d5 g1f3 g8f6 d2d4 f8d6 c2c4 c7c6 c4d5 f6d5 f1d3 d8e7 d3e2 b8d7 e1g1 e8g8 f1e1 d7f6 h2h3 c8e6 b1d2 f8e8 d2c4 d5f4 e2f1 e7c7 c4d6 c7d6 f3e5 f4g6 c1g5 e6d5 e5c4 d5c4 e1e8 a8e8 f1c4 h7h6 g5e3 f6d5 d1b3 d6c7');
INSERT INTO evaluations VALUES(55786,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(55786,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(55786,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55786,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(55786,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(55786,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(55786,36,'cp -16','e6d5 g1f3 g8f6 d2d4 f8d6 c2c4 c7c6 c4d5 f6d5 f1c4 d8e7 c4e2 e8g8 e1g1 e7c7 b1c3 c8e6 c3e4 b8d7 e4d6 c7d6 f1e1 f8e8 c1d2 d7f8 d1b3 d5b6 b3c2 e6d5 d2b4 d6c7 f3e5 f8e6 b4c3 b6d7 e2f1 a8d8 a1d1 d7f6 c2f5');
INSERT INTO evaluations VALUES(56297,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(56297,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(56297,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(56297,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(56297,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56297,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(56297,36,'cp -21','d2d4 g8f6 b1c3 c7c6 f1d3 f8d6 d1e2 d6e7 g1f3 e8g8 e1g1 c8g4 e2e3 b8d7 c1d2 f8e8 f3e5 g4e6 h2h3 d7e5 d4e5 d5d4 e3f4 d4c3 e5f6 c3d2 f6e7 d8e7 a1d1 g7g6 d1d2 g8g7 f1e1 a8d8 f4b4 e7b4 a3b4 e6d5 e1e8 d8e8 f2f3 e8e1 g1f2');
INSERT INTO evaluations VALUES(56219,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(56219,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(56219,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(56219,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(56219,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56219,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(56219,36,'cp -4','g8f6 f1d3 f8d6 g1e2 e8g8 e1g1 f8e8 b1c3 c7c6 c1f4 c8g4 d1d2 g4h5 f4d6 d8d6 a1e1 h5g6 e2f4 b8d7 e1e8 f6e8 c3d1 g6d3 f4d3 d6c7 f1e1 e8d6 c2c3 d7f6 h2h3 a7a5 a3a4 g7g6 d1e3 a8e8 f2f3 h7h5 d2f2 d6c4 e3c4');
INSERT INTO evaluations VALUES(48699,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48699,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(48699,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(48699,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(48699,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48699,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48699,36,'cp -2','f1d3 c7c5 d4c5 f8c5 g1f3 d8e7 d1e2 e7e2 d3e2 e8g8 b1c3 b8c6 c3a4 c5e7 c1e3 f8d8 f3d4 f6g4 e2g4 c8g4 h2h3 g4d7 a4c3 c6d4 e3d4 e7g5 h3h4 d8e8 e1d1 g5f4 c3d5 f4h6 h1e1 e8d8 c2c4 d7c6 d1c2 c6d5 c4d5 d8d5 d4c3 a8c8');
INSERT INTO evaluations VALUES(36550,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(36550,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(36550,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36550,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(36550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36550,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36550,36,'cp 0','f8e8 d3e2 e7e6 e1g1 c7c5 a2a3 b8c6 d4d5 c6a5 c1d2 e6d5 c4d5 c5c4 f3d4 c8d7 a3a4 a8c8 a1a3 d8b6 e2f3 a7a6 d2c1 h7h5 h2h3 b6b4 f1e1 b4b6');
INSERT INTO evaluations VALUES(57800,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(57800,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(57800,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(57800,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(57800,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(57800,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(57800,36,'cp 12','d3e2 e7e6 e1g1 c7c5 b2b3 b7b6 d1d2 b8c6 c1b2 c5d4 f3d4 c6d4 d2d4 d6d5 f1d1 c8b7 d4f4 d8e7 c4d5 f6d5 c3d5 b7d5 b2g7 g8g7 f4d4 e7f6 d4f6 g7f6 f2f3 e8d8 a1c1 a8c8 g1f2 f6e7 e3e4 d5b7 e4e5 d8d1 c1c8 b7c8 e2d1 f7f6 f3f4 f6e5 f4e5 c8b7 d1c2 a7a5 h2h4');
INSERT INTO evaluations VALUES(21276,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(21276,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(21276,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(21276,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(21276,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(21276,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(21276,36,'cp 1','c6a5 d1e2 g8f6 d3d4 e5d4 f3d4 a5c4 e2c4 c7c5 d4e2 c8e6 c4d3 d6d5 e4d5 e6d5 e1g1 d5c6 d3g3 f8d6 c1f4 d6f4 e2f4 e8g8 a1d1 d8c8 f4d5 f6d5 c3d5 c6d5 d1d5 f8d8 g3f3 c8c7 f1d1 d8d5 f3d5 a8e8 g2g3 g7g6 b2b3 b7b5 c2c4 b5c4 b3c4 h6h5 g1g2 e8e2 d1d2 e2d2 d5d2');
INSERT INTO evaluations VALUES(44219,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44219,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44219,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44219,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(44219,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44219,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44219,36,'cp 0','f8e7 b1d2 e8g8 f1d3 d7d5 a2a4 b8c6 g1f3 f6h5 d4c5 e7c5 e1g1 h5f4 e3f4 c5d6 g2g3 f7f6 d1c2 g7g6 d3g6 h7g6 c2g6 g8h8 g6h6 h8g8 h6g6');
INSERT INTO evaluations VALUES(46141,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(46141,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(46141,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(46141,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(46141,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(46141,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(46141,36,'cp -20','c5d4 d1d4 b8c6 g5f6 g7f6 d4h4 d5c4 a1d1 c8d7 g2g3 f8g7 f1g2 e8g8 e1g1 d8e7 f3d2 f6f5 h4e7 c6e7 d2c4 d7c6 d1d2 g7c3 b2c3 c6g2 g1g2 e7d5 c4a5 b7b6 c3c4 d5f6 a5c6 g8g7 f1d1 a8c8 c6a7 c8c4 a7b5 c4a4 d1c1 f8a8 a2a3 f6d5 b5c3 a8c8 c3a4 c8c1');
INSERT INTO evaluations VALUES(46294,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(46294,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(46294,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(46294,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(46294,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(46294,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(46294,36,'cp -3','d1d4 b8c6 g5f6 g7f6 d4h4 d5c4 a1d1 c8d7 g2g3 f6f5 h4c4 a8c8 f1g2 d8b6 e1g1 f8g7 c4b5 g7c3 b5d3 c3f6 d3d7 e8f8 d7c8 f8g7 c8d7 h8d8 b2b3 d8d7 d1d7 b6b5 e2e3 c6b4 d7c7 b4d5 f3d4 f6d4 g2d5 b5d5 e3d4 d5e4 f1c1 g7g6 c7d7 h7h5 d4d5');
INSERT INTO evaluations VALUES(55665,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55665,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(55665,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55665,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(55665,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55665,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(55665,36,'cp 18','g8f6 d1d4 f8e7 c1f4 a7a6 a2a3 b7b5 c4d5 c7c6 d5a2 c6c5 d4d2 e8g8 b1c3 c8e6 e1g1 e6a2 a1a2 f8e8 f1d1 b8c6 a2a1 b5b4 a3b4 c5b4 c3d5 f6e4 d2d3 e7f8 a1a6 e4c5 a6a8 d8a8 d3f5 c6e7 f5h5 e7d5 h5d5 a8d5 d1d5 b4b3');
INSERT INTO evaluations VALUES(47933,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47933,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(47933,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47933,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(47933,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(47933,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47933,36,'cp 16','d1d4 f8e7 c1f4 b8c6 d4d3 c8e6 b1d2 e8g8 e1g1 e6c4 d3c4 f8e8 f1e1 e7f8 a1d1 d8d7 h2h3 a7a6 c4b3 h7h6 e4e5 d6e5 d2c4 d7e6 f3e5 b7b5 e5c6 e6c6 c4e5 c6e6 g1f1 e6b3 c2b3 f8d6 e5d3 e8e1 f1e1');
INSERT INTO evaluations VALUES(47931,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(47931,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(47931,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(47931,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(47931,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47931,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(47931,36,'cp 10','f8e7 c1f4 b8c6 d4d3 c8e6 b1d2 e6c4 d3c4 d8d7 e1g1 d7g4 f4g3 d6d5 c4b3 e8c8 e4d5 d8d5 d2c4 e7c5 a1e1 g4f5 c4e3 c5e3 e1e3 h8d8 h2h3 f6h5 g3h4 d5b5 b3a4 b5a5 a4g4 f7f6 a2a3 a5d5 g4f5 d5f5 c2c3');
INSERT INTO evaluations VALUES(40829,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40829,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40829,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(40829,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40829,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(40829,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(40829,36,'cp 20','c1f4 a7a6 a2a3 b7b5 c4d5 c7c5 d4d3 f6d5 d3d5 a8a7 b1c3 e8g8 e1g1 f8e8 f1d1 a7d7 d5d3 c8b7 c3d5 e7f8 c2c3 c5c4 d3c2 b8c6 a3a4 c6e7 a4b5 a6b5 f4g5 b7d5 e4d5 d8c8 g5e3 d7b7 h2h3 e7f5 e3f4 h7h6 c2d2');
INSERT INTO evaluations VALUES(18063,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(18063,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(18063,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18063,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18063,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(18063,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(18063,36,'cp 25','d7d6 d2d4 f7f6 b1c3 d8d7 f3e1 c8b7 c1e3 a6a5 d1e2 g8e7 d4e5 f6e5 f2f4 e5f4 e3f4 e7g6 e1d3 f8e7 f4g3 b7a6 f1f5 e7f6 f5a5 e8g8 a5a6 a8a6 e4e5 f8e8 d3c5 d6c5 e2a6 g6e5 a1e1 e5f7 e1e8 d7e8 a6e2 e8e2 c3e2 f7d6 e2c3');
INSERT INTO evaluations VALUES(18017,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(18017,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(18017,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(18017,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(18017,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(18017,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(18017,36,'cp 51','d2d4 f7f6 b1c3 a6a5 d4e5 d6e5 c1e3 g8e7 d1d8 e8d8 c3a4 d8e8 f1d1 g7g5 f3e1 e7g6 e3c5 h7h5 c5f8 g6f8 e1d3 f8e6 f2f3 g5g4 g1f2 e8e7 a4c5 h5h4 c5e6 c8e6 d3c5 a8g8 a1c1 h4h3 g2g3 g4f3 c5e6 e7e6 f2f3');
INSERT INTO evaluations VALUES(18006,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(18006,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(18006,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(18006,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(18006,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(18006,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(18006,36,'cp 49','f7f6 b1c3 g8e7 f3h4 d8d7 d4e5 f6e5 h2h3 e7g6 h4g6 h7g6 f2f4 d6d5 d1d3 f8c5 c1e3 d5d4 c3a4 c5e7 e3d2 c6c5 b2b3 d7e6 d3g3 e8g8 f4e5 c8b7 f1e1 c5c4 d2g5 f8e8 g5e7 e6e7 a1d1 a8d8 a4b2 e7e5 g3e5 e8e5 b2c4 e5e4 c4a5 b7a8 b3b4 d8b8 e1e4 a8e4 d1d4 e4c2 g1f2 c2a4 f2g3 a4b5 h3h4 b8e8');
INSERT INTO evaluations VALUES(32649,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(32649,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(32649,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(32649,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(32649,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(32649,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(32649,36,'cp 160','h2h3 g4f3 d1f3 c7c6 c1d2 b8d7 e1c1 e7e6 a2a3 a5b6 g2g4 b6d4 d2e3 d4e5 e3f4 e5c5 g4g5 f6d5 c3e4 c5a5 e4d6 f8d6 f4d6 d5e7 h3h4 e7f5 d6b4 a5e5 b4c3 e5c5 f1h3 e8c8 h1e1 d7b6 d1d8 h8d8 h3f5 c5f5 f3f5 e6f5 c3g7 c8d7 e1e5 f5f4 e5f5 d7e6 f5f4');
INSERT INTO evaluations VALUES(32647,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(32647,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(32647,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(32647,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(32647,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(32647,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(32647,36,'cp 163','g4h5 g2g4 h5g6 f3d2 f6d5 d2c4 d5c3 c4a5 c3d1 f1g2 d1f2 e1f2 b8d7 a5b7 e7e6 c1f4 f8e7 c2c3 e7h4 f4g3 h4g3 f2g3 e8g8 b2b4 d7f6 h1e1 f6d5 b7a5 a8e8 a1c1 a7a6 a5c6 d5e7 e1e5 e7c6 g2c6 e8d8 c1e1 d8d6');
INSERT INTO evaluations VALUES(32638,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(32638,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(32638,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(32638,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(32638,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(32638,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(32638,36,'cp 156','g2g4 h5g6 f3d2 b8d7 d2c4 a5a6 c1f4 a6e6 c4e5 e6b6 d1f3 e7e6 e5c4 b6c6 f3c6 b7c6 f4c7 d7b6 e1c1 b6c4 f1c4 h7h5 g4g5 f6d5 c3d5 e6d5 c4d3 f8e7 h1g1 e8d7 c7e5 h8g8 d3g6 f7g6 d1d3 a8f8 f2f4 e7d6 d3b3 f8b8 g1g3 a7a5 b3b8 g8b8 g3a3 d6a3 e5b8 a3f8 a2a4 d7e6 c2c3');
INSERT INTO evaluations VALUES(32637,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(32637,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(32637,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(32637,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(32637,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(32637,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(32637,36,'cp 149','h5g6 f3d2 b8d7 d2c4 a5a6 c1f4 a6e6 c4e5 e6b6 d1f3 e7e6 e5c4 b6c6 f3c6 b7c6 f4c7 d7b6 e1c1 b6c4 f1c4 h7h5 g4g5 f6d5 c3d5 e6d5 c4d3 e8d7 c7e5 f8e7 h1g1 h8g8 d3g6 f7g6 d1d3 a8f8 f2f4 e7d6 g1e1 a7a5 d3f3 d6e5 e1e5 f8f7 b2b3 a5a4');
INSERT INTO evaluations VALUES(32662,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(32662,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(32662,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(32662,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(32662,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(32662,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(32662,36,'cp 162','f3d2 b8d7 d2c4 a5a6 c1f4 a6e6 c4e5 e6b6 d1f3 e7e6 e5c4 b6c6 f3c6 b7c6 f4c7 d7b6 e1c1 b6c4 f1c4 h7h5 g4g5 f6d5 c3d5 e6d5 c4d3 e8d7 c7e5 f8e7 h1g1 h8g8 d3g6 f7g6 d1d3 a8f8 f2f4 e7d6 g1e1 a7a5 e5d6 d7d6 d3e3 f8f4 e3e6 d6c7 e6g6 f4d4 e1e6 d4c4 b2b3 c4c3');
INSERT INTO evaluations VALUES(2129,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(2129,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(2129,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(2129,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(2129,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(2129,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(2129,36,'cp 79','e1g1 a7a6 b5c6 d7c6 d2d3 e7e5 a2a4 g8f6 b1a3 c8g4 h2h3 g4h5 g2g4 h5g6 a3c4 f6d7 f3h4 d7b6 c4b6 c7b6 b2b3 e8c8 g4g5 c8b8 d1g4 f8d6 g1g2 b8a8 c1e3 b6c7 f2f4 e5f4 e3f4 f7f6 f4d6 c7d6 a1e1 d6d4 g4g3 f6g5 h4g6 h7g6 e1e2');
INSERT INTO evaluations VALUES(2128,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(2128,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(2128,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(2128,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(2128,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(2128,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(2128,36,'cp 75','a7a6 b5c6 d7c6 d2d3 e7e5 a2a4 a6a5 c1e3 f8d6 b1d2 g8e7 f3h4 h7h6 d1f3 b7b6 d2c4 c8e6 h4f5 e6c4 d3c4 e7f5 e4f5 e8c8 f3e4 c8b7 f1d1 c7c8 g2g3 d6c7 d1d8 h8d8 a1e1 f7f6 h2h4 c8d7 h4h5 d7f7 e4g4 b7a8 g1g2 a8b7 b2b3 b7b8 g4f3');
INSERT INTO evaluations VALUES(1274,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(1274,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(1274,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(1274,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(1274,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(1274,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(1274,36,'cp 71','b5c6 d7c6 a2a4 e7e5 a4a5 g8f6 d2d3 c8g4 h2h3 g4h5 c1e3 f6d7 b1d2 f7f6 d1b1 h5f7 d2c4 f7c4 d3c4 f8e7 c2c3 e8g8 b2b4 c5b4 c3b4 c6c5 b4b5 f8d8 f1d1 d7f8 d1d8 a8d8 b5a6 b7a6 b1b6 c7c8 a1b1 f8e6 b6b7 e7f8 b7c8 d8c8 b1b6');
INSERT INTO evaluations VALUES(1272,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(1272,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(1272,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(1272,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(1272,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(1272,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(1272,36,'cp 57','d7c6 a2a4 e7e5 a4a5 c8g4 d2d3 g8f6 h2h3 g4h5 c1e3 f6d7 b1d2 f7f6 c2c3 f8e7 d1c2 h5f7 d2c4 e8g8 f3d2 a8d8 g1h1 g8h8 a1d1 g7g5 d2b3 f7e6 f2f3 d8e8 f1e1 f8g8 e3f2 h7h5 c4e3 h8h7 e3f5');
INSERT INTO evaluations VALUES(22690,30,'cp -154',NULL);
INSERT INTO evaluations VALUES(22690,31,'cp -156',NULL);
INSERT INTO evaluations VALUES(22690,32,'cp -156',NULL);
INSERT INTO evaluations VALUES(22690,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(22690,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(22690,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(22690,36,'cp -133','d7d5 c4b3 g8f6 h5d1 c8f5 d2d3 d8e7 b3a4 f5d7 g1f3 e8c8 d3d4 e5e4 a4c6 d7c6 f3e5 c6e8 c2c4 d5c4 e5c4 e8b5 b2b3 e7e6 c1a3 f8a3 b1a3 b5a6 d1d2 h7h5 e1c1 h8e8 c1b1 h5h4 b1a1 e6e7 a3b1 f6d5 b1c3');
INSERT INTO evaluations VALUES(21660,30,'cp -146',NULL);
INSERT INTO evaluations VALUES(21660,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(21660,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(21660,33,'cp -156',NULL);
INSERT INTO evaluations VALUES(21660,34,'cp -158',NULL);
INSERT INTO evaluations VALUES(21660,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(21660,36,'cp -146','c4b3 g8f6 h5d1 c8f5 d2d3 d8e7 b3a4 f5d7 c1d2 e8c8 d1e2 g7g5 h2h3 h7h5 b1c3 d5d4 a4c6 d7c6 e3d4 e5d4 e2e7 f8e7 c3e2 c6g2 h1h2 g2c6 d2g5 h8g8 g5h4 e7d6 h4g3 d6b4 e1f1 g8g6 a2a3 b4c5 b2b4 c5b6 f1e1 d8h8 g3h4 h8g8 e1d2 f6d5 a1e1');
INSERT INTO evaluations VALUES(5294,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(5294,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(5294,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(5294,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(5294,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(5294,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(5294,36,'cp 93','d2d4 c5d4 c3d4 d7d6 e1g1 g8f6 d4d5 c6b8 d1c2 b8d7 h2h3 e8g8 a2a4 d7c5 b1c3 c8d7 a4a5 a8b8 c2e2 b7b5 a5b6 a7a5 e4e5 d6e5 f3e5 d8b6 c1e3 b6d6 e3d4 c5b3 c4b3 b8b3 f1c1');
INSERT INTO evaluations VALUES(5314,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(5314,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(5314,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(5314,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(5314,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(5314,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(5314,36,'cp 81','c5d4 c3d4 d7d6 b1c3 g8f6 d4d5 c6a5 c4d3 a7a6 e1g1 c8g4 c1e3 a8c8 a1c1 f6d7 c3e2 c8c1 e3c1 b7b5 d3b1 e8g8 b2b3 d7f6 f3d4 g4e2 d4e2 a5b7 c1e3 f6g4 e3g5 b7c5 e2d4 d8d7 d4c6 g7f6 g5f6 e7f6 b3b4 c5a4 d1d4 f8c8 f1c1 g4e5 h2h3 e5c6');
INSERT INTO evaluations VALUES(5389,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(5389,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(5389,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(5389,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(5389,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(5389,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(5389,36,'cp 80','c3d4 g8f6 d4d5 b7b5 c4e2 c6a5 e4e5 f6g4 c1f4 f7f6 e5e6 e8g8 e6d7 d8d7 e1g1 c8b7 b1a3 b5b4 a3c4 a5c4 e2c4 a8d8 d1b3 b7d5 f1d1 e7e6 b3b4 d7c6 c4b3 d5f3 d1d8 f8d8 a1c1 c6d7 g2f3 g4e5 f4e5 f6e5 c1d1 d7c7 b3e6 g8h8 e6d5 c7b6 b4b6 a7b6');
INSERT INTO evaluations VALUES(5388,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(5388,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(5388,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(5388,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(5388,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(5388,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(5388,36,'cp 91','g8f6 d4d5 c6a5 c4d3 b7b5 e1g1 e8g8 b1a3 a7a6 b2b4 a5c4 e4e5 f6g4 d3c4 b5c4 a3c4 d7d6 c1b2 c8b7 c4a5 d8b6 e5d6 g7b2 d6e7 f8e8 a1b1 b6b4 a5b7 e8e7 b7a5 b4a5 b1b2 a8d8 d1d4 g4h6 d4d2 a5d2 b2d2 h6f5 f1d1 f5d6 g2g4 d8e8 g1f1 e7b7 f1g2 e8c8 f3e5');
INSERT INTO evaluations VALUES(55117,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(55117,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(55117,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(55117,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(55117,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(55117,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(55117,36,'cp -34','e7e5 d2d4 b8c6 g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 a2a3 d8d7 b2b4 a7a6 c4d5 f6d5 c3d5 e6d5 f1c4 d5c4 d2c4 b7b5 c4a5 c6a5 b4a5 c7c5 d4c5 d7d1 e1d1 f8c5 c1d2 c5e7 d1e2 e8f7 h1c1 h8c8 d2b4 c8c4 c1c4');
INSERT INTO evaluations VALUES(21121,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(21121,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(21121,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(21121,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(21121,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(21121,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(21121,36,'cp 72','e7e5 f1c4 f8b4 b1c3 e5d4 e3d4 g8f6 e1g1 e8g8 h2h3 h7h6 c1f4 b4d6 f4e5 c8f5 f1e1 f6d7 d1c1 d6e5 d4e5 f5e6 c1f4 e6c4 f4c4 d7b6 c4g4 d8e7 e1e4 a8d8 a1e1 b6d5 a2a3 f8e8 g4g3 d5c3 b2c3 b7b6 e4g4 f7f6 e5f6 e7f6 e1e8 d8e8 g3c7');
INSERT INTO evaluations VALUES(21793,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(21793,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(21793,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(21793,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(21793,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(21793,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(21793,36,'cp 51','f1c4 f8b4 b1c3 e5d4 e3d4 g8f6 e1g1 e8g8 h2h3 h7h6 c1f4 b4d6 f4e5 c8f5 f1e1 f6d7 e5d6 c7d6 c4f1 a7a6 a1c1 a8b8 d1d2 d7f6 d4d5 c6e7 a2a4 f8e8 d2d4 e7g6 e1e8 f6e8 f3d2 f5d7 c1e1 b7b5 d4b4 e8f6 a4b5 d7b5 f1b5');
INSERT INTO evaluations VALUES(22663,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(22663,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(22663,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(22663,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(22663,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(22663,35,'cp -149',NULL);
INSERT INTO evaluations VALUES(22663,36,'cp -138','g8f6 d5c4 d7d5 e4d5 f6d5 g1f3 f8e7 c1d2 c8e6 c4e4 d8d7 f1b5 e6f5 e4e2 e8g8 e1g1 f8e8 b1c3 e7f6 e2c4 d5b6 c4f4 f5c2 a1c1 c2d3 b5d3 d7d3 f1d1 a8c8 d2e3 d3g6 g2g3 f6c3 c1c3 e8e4 f3h4 e4f4 h4g6 f4c4 e3b6 c4c3 b2c3 c7b6');
INSERT INTO evaluations VALUES(16339,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(16339,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(16339,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(16339,33,'cp -144',NULL);
INSERT INTO evaluations VALUES(16339,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(16339,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(16339,36,'cp -144','d5c4 d7d5 e4d5 f6d5 g1f3 f8e7 c1d2 c8e6 c4e4 d8d7 f1b5 e6f5 e4e2 e8g8 b1c3 d5b4 e1g1 e7d6 a2a3 b4c2 a1c1 c2d4 f3d4 c6d4 b5d7 d4e2 c3e2 f5d7 d2f4 d6f4 e2f4 d7c6 f2f3 f8d8 f1d1 g8f8 h2h4 d8d1 c1d1 g7g6 g1f2 f8e7 d1e1 e7f6 e1d1 a7a5 g2g4 h7h6 h4h5 a5a4 h5g6 f7g6 f4d5 c6d5 d1d5 f6e6 d5c5 c7c6');
INSERT INTO evaluations VALUES(16670,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(16670,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(16670,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(16670,33,'cp -151',NULL);
INSERT INTO evaluations VALUES(16670,34,'cp -139',NULL);
INSERT INTO evaluations VALUES(16670,35,'cp -151',NULL);
INSERT INTO evaluations VALUES(16670,36,'cp -131','d7d5 e4d5 f6d5 g1f3 f8e7 c1d2 c8e6 c4e4 d8d7 f1b5 e6f5 e4e2 e8g8 b1c3 d5b4 e1g1 e7d6 a2a3 b4c2 a1c1 c2d4 f3d4 c6d4 b5d7 d4e2 c3e2 f5d7 d2f4 d6f4 e2f4 c7c6 f1d1 f8d8 f2f3 d7e8 g1f2 b7b6 g2g4 g8f8 h2h4 c6c5 g4g5 a8c8 f2g3 e8b5 g3f2 h7h6 d1d5');
INSERT INTO evaluations VALUES(27214,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(27214,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(27214,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(27214,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(27214,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(27214,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(27214,36,'cp 37','d2d4 e5d4 f3d4 g7g6 d4c6 b7c6 b5d3 f8g7 b1c3 a8b8 e1g1 g8e7 d1d2 e8g8 b2b3 f7f5 c1b2 f5e4 d3e4 e7f5 a1e1 f5h4 f2f4 h4f5 e4d3 d8h4 d3c4 g8h8 c3d1 g7b2 d1b2 c6c5 b2d1 b8e8 d1c3 e8e1 d2e1 h4g3 c3d5 d7c6 e1g3 f5g3');
INSERT INTO evaluations VALUES(37475,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37475,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(37475,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(37475,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(37475,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(37475,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(37475,36,'cp 15','c2c4 c7c6 b2b3 d7d6 b1c3 e7e5 d2d3 b8a6 e2e4 a6c7 b3b4 d6d5 e4d5 c6d5 f3e5 f6d7 d3d4 d7e5 d4e5 d5c4 f2f4 c8g4 d1e1 f7f6 e1e4 a8c8 c1b2 d8d3 a1e1 f6e5 f4e5 g4e6 a2a3 f8f1 e1f1 d3e4 c3e4 c7d5');
INSERT INTO evaluations VALUES(37387,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(37387,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(37387,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(37387,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(37387,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37387,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(37387,36,'cp 9','c7c6 d2d4 d7d5 c4d5 c6d5 f3e5 f6e4 c1f4 c8f5 b1c3 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 f1c1 d8c8 a4a7 c3c1 a1c1 c8a8 a7a8 f8a8 e2e3 g7f8 a2a4 a8a4 g2f1 a4a2 c1c7 a2d2 c7a7 d2c2 a7a8 f5e4');
INSERT INTO evaluations VALUES(35689,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(35689,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(35689,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35689,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(35689,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(35689,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(35689,36,'cp 14','b1c3 d7d5 c4d5 c6d5 d2d4 f6e4 f3e5 c8f5 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 f1c1 d8d7 h2h4 h7h6 c1e1 d7e7 a1c1 f8c8 e2e4 d5e4 g2e4 f5e4 e1e4 e7b7 e4e1 b7b5 a4b3 c8d8 c1b1 b5a5 b3c4 a5d5 c4d5 c6d5');
INSERT INTO evaluations VALUES(35686,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(35686,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(35686,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(35686,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(35686,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35686,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35686,36,'cp 0','d7d5 c4d5 c6d5 d2d4 f6e4 f3d2 e4f6');
INSERT INTO evaluations VALUES(35864,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35864,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35864,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35864,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(35864,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(35864,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(35864,36,'cp 0','c4d5 c6d5 d2d4 f6e4 f3e5 c8f5 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 e5c6 b7c6 e4c6 a8c8 c6f3 d8d4 d1d4 g7d4 f1d1 f8d8 c1f4 d4b2 d1d8 c8d8 a1b1 b2c3 a2a3 f7f6 b1b7 d8d7 b7b8 g8f7 g1g2 g6g5 f4g3 d7d2 b8b7 d2d7');
INSERT INTO evaluations VALUES(35860,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(35860,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(35860,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(35860,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35860,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35860,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35860,36,'cp 5','c6d5 d2d4 f6e4 f3e5 c8f5 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 d4d5 e6d5 d1d5 g7e5 d5d8 a8d8 c1h6 e5g7 h6g7 g8g7 e4c6 b7c6 f1d1 a7a5 b2b3 d8d5 d1d5 c6d5 a1c1 f8b8 c1c7 a5a4 c7e7 a4b3 a2b3 b8b3 f2f3 d5d4 e7d7 b3b2 d7d4 b2e2');
INSERT INTO evaluations VALUES(36040,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(36040,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36040,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(36040,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(36040,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(36040,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36040,36,'cp 11','d2d4 f6e4 f3e5 c8f5 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 d4d5 e6d5 d1d5 g7e5 d5d8 f8d8 g1g2 c6d4 e4d3 a8c8 f2f4 e5f6 a1b1 d4e6 f1d1 h7h6 c1e3 f6d4 e3d2 e6c5 d2b4 e7e5 b4c5 c8c5 f4f5 g8g7 b2b3 c5c3 f5g6 f7g6');
INSERT INTO evaluations VALUES(36037,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(36037,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36037,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(36037,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(36037,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36037,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36037,36,'cp 12','f6e4 f3e5 c8f5 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 d4d5 e6d5 d1d5 g7e5 d5d8 f8d8 g1g2 a8c8 a1b1 b7b6 c1e3 c6b4 f1d1 e7e6 d1d8 c8d8 f2f4 e5g7 g4g5 h7h6 h2h4 b4d5 e4d5 d8d5 b2b3 g7d4 e3d2 h6g5 f4g5 d4e5 d2e3 e5d4 e3f4 d4e5 f4e5');
INSERT INTO evaluations VALUES(49795,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(49795,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(49795,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(49795,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(49795,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(49795,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(49795,36,'cp 125','g1f3 d7d6 b1c3 b8c6 e1g1 f8e7 h2h3 h6g8 c3d5 g8f6 f1e1 e8g8 d5e7 d8e7 c2c3 c8e6 c4b5 a7a5 b5a4 h7h6 a4c2 e6d7 a2a4 g8h8 d2d4 f6h7 a1b1 a8e8 d4d5 c6d8 c1e3 f7f5 e4f5 d7f5 c2f5 f8f5 b2b3');
INSERT INTO evaluations VALUES(49793,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(49793,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(49793,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(49793,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(49793,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(49793,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(49793,36,'cp 128','b8c6 e1g1 d7d6 d2d4 e5d4 f3d4 h6g4 b1c3 f8e7 c4b5 c8d7 b5c6 b7c6 d4c6 d7c6 d1g4 e8g8 c3d5 f7f5 d5e7 d8e7 e4f5 e7f7 f2f3 c6d7 g4d4 d7f5 f1f2 f8e8 c1f4 e8e7 f4g3 a8e8 h2h3 f7f6 d4f6 g7f6 c2c4 h7h5 b2b3 g8f7');
INSERT INTO evaluations VALUES(37674,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(37674,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(37674,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(37674,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(37674,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(37674,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(37674,36,'cp 99','c1e3 d8b6 b1a3 b6a5 e3d2 a5d8 d4b3 c5a3 b2a3 b7b6 d1g4 d8f6 e1c1 b8c6 g4g3 f6g6 f2f3 g6g3 h2g3 g8e7 a3a4 c8b7 f1b5 e8c8 a4a5 d7d5 e4d5 e7d5 a5b6 d5b6 b5e2 h7h6 d2c3 f7f6 b3c5 d8d1 h1d1 h8d8 d1d8 c6d8 c5b7');
INSERT INTO evaluations VALUES(37673,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(37673,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(37673,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(37673,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(37673,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(37673,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(37673,36,'cp 104','d8b6 b1a3 b6a5 e3d2 a5d8 d4b3 c5a3 b2a3 b8c6 d1g4 d8f6 f2f4 b7b6 f1d3 g8h6 g4h3 f6e7 f4f5 e8g8 e1g1 f7f6 d2h6 g7h6 h3h6 c6e5 f1f4 c8b7 a1f1 g8h8 f5e6 e5g6 f4g4 g6e5 g4g3 d7e6 a3a4 f8f7 h6e3 a8g8 g3g8 h8g8 a4a5');
INSERT INTO evaluations VALUES(32980,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(32980,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(32980,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(32980,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(32980,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(32980,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(32980,36,'cp 113','b1a3 b6a5 e3d2 a5d8 d4b3 c5a3 b2a3 d7d5 d2a5 b7b6 a5c3 g8f6 e4d5 d8d5 d1d5 f6d5 c3g7 h8g8 g7b2 c8a6 c2c4 d5f4 g2g3 a6b7 h1g1 b8d7 f2f3 f4g6 e1f2 g6e5 f1e2 a8c8 b3d2 f7f6 g1d1 e8e7 a1c1 d7c5 d2b3 c5a4 b2d4 b7a6 c4c5 a6e2 f2e2 e5c6 d4f2');
INSERT INTO evaluations VALUES(6561,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(6561,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(6561,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(6561,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(6561,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(6561,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(6561,36,'cp -80','g8f6 d4c6 b7c6 b1d2 d7d5 e4d5 d8d5 d1e2 c8e6 d2b3 c5b6 c3c4 d5d6 c1e3 d6b4 e2d2 b4d2 e3d2 f6g4 c4c5 e6b3 c5b6 a7b6 f1e2 g4f6 e1g1 b3d5 f2f3 f6d7 f1e1 d7c5 b2b4 c5a4 g1f2 e8d7 h2h4 h8e8 a2a3 c6c5 e2b5 c7c6 e1e8 d7e8 a1e1 e8d7 b5a4 a8a4 b4c5');
INSERT INTO evaluations VALUES(10303,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(10303,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(10303,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(10303,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(10303,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(10303,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(10303,36,'cp -69','d4c6 b7c6 b1d2 d7d5 e4d5 e8g8 f1e2 c6d5 e1g1 f8e8 d2b3 c5d6 c1g5 h7h6 g5h4 a8b8 e2f3 a7a5 b3a5 b8b2 a5c6 d6h2 g1h1 d8d6 h4f6 d6f6 h1h2 f6c6 f3d5 c6c3 a1c1 c3f6 d1f3 f6f3 d5f3 b2a2 c1c7 c8e6 c7c1 e8d8 h2g3 d8d2 c1b1 g7g6 b1d1 e6b3 d1d2');
INSERT INTO evaluations VALUES(12807,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(12807,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(12807,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(12807,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(12807,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(12807,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(12807,36,'cp 27','e7e6 g1f3 d8b6 e1g1 f8e7 b2b3 d7d5 c4d5 f6d5 c3d5 e6d5 d2d4 c5d4 c1b2 c8g4 b2d4 c6d4 d1d4 b6d4 f3d4 e8c8 a1d1 e7f6 h2h3 g4e6 d1d3 c8b8 f1d1 d8c8 f2f4 h8d8 g1f2 g7g6 g3g4 a7a6 g2f3 h7h6 f4f5 g6f5 d4f5');
INSERT INTO evaluations VALUES(12402,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(12402,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(12402,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(12402,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12402,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(12402,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12402,36,'cp 33','g1f3 d8b6 e1g1 f8e7 b2b3 d7d5 c4d5 f6d5 c3d5 e6d5 d2d4 c5d4 c1b2 c8g4 b2d4 c6d4 d1d4 b6d4 f3d4 e8c8 a1d1 c8b8 d1d3 e7f6 e2e3 h7h5 h2h3 g4e6 h3h4 d8d6 f1d1 h8d8 g2f3 g7g6 g1g2 f6e5 d4e2 a7a6 e2f4 e5f4 g3f4 b8c7 d3d2 c7b8');
INSERT INTO evaluations VALUES(20261,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(20261,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(20261,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(20261,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(20261,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(20261,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(20261,36,'cp 15','b1d2 g8f6 g1f3 f8d6 f4g3 e8g8 f1d3 g7g6 e1g1 b7b6 f1e1 c8b7 g3h4 d6e7 h2h3 a8c8 a2a3 f6h5 h4e7 d8e7 b2b4 c5c4 d3c2 f7f5 b4b5 c6d8 f3e5 h5f6 f2f4 d8f7 d2f3 f6e4 e5f7 f8f7 c2e4 d5e4');
INSERT INTO evaluations VALUES(20260,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(20260,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(20260,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(20260,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(20260,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(20260,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(20260,36,'cp 23','g8f6 g1f3 f8e7 f3e5 d8b6 a1b1 c6e5 d4e5 f6e4 d2e4 d5e4 h2h4 h7h5 d1c2 c8d7 c2e4 e8c8 e4c2 d7b5 f1e2 b5e2 c2e2 b6c6 e3e4 c5c4 g2g3 d8d3 e1g1 g7g6 g1h1 h8d8 b1a1 c6b6');
INSERT INTO evaluations VALUES(13564,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(13564,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(13564,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(13564,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(13564,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(13564,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(13564,36,'cp 20','g1f3 f8d6 f4g3 e8g8 f1d3 g7g6 e1g1 b7b6 g3h4 d6e7 d1e2 c8b7 a2a4 f6d7 h4g3 e7f6 h2h4 e6e5 d4e5 d7e5 f3e5 c6e5 d3a6 b7a6 e2a6 f8e8 g3e5 f6e5 d2f3 e5f6 a1d1 f6h4 a6b7 a8b8 b7d5 d8d5 d1d5');
INSERT INTO evaluations VALUES(22816,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(22816,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(22816,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(22816,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(22816,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22816,36,'cp 0','b1c3 a7a5 f1g2 f8c5 g1e2 g8f6 d2d3 e8g8 e1g1 f8e8 c3b5 c6a7 b5c3');
INSERT INTO evaluations VALUES(22782,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22782,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22782,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(22782,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22782,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(22782,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(22782,36,'cp 0','a7a5 g1e2 f8c5 f1g2 g8e7 d2d3 e8g8 e1g1 d7d6 h2h3 c8e6 g1h2 d6d5 f2f4 d5d4 c3a4 c5d6 f4f5 e6d7 b2b3 b7b5 a4b2 d6a3 g3g4 f7f6 h3h4 e7c8 e2g3 c8d6 g4g5 f6g5 c1g5 d8e8 a1b1 g8h8 g3h5 e8f7');
INSERT INTO evaluations VALUES(50407,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(50407,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(50407,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(50407,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(50407,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(50407,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(50407,36,'cp 66','d7d6 d2d4 d6e5 f3e5 c7c6 f1e2 c8f5 e1g1 b8d7 e5f3 d8c7 c2c4 d5f6 b1c3 e7e6 f3h4 f5g6 g2g3 a7a5 h4g6 h7g6 f1e1 f8b4 d1c2 e8g8 a2a3 b4e7 c1f4 e7d6 f4g5 e6e5 d4e5 d6e5 a1d1 d7c5 g5e3 f6d7 h2h4 c5e6 h4h5 g6h5 e2h5');
INSERT INTO evaluations VALUES(47071,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(47071,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(47071,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(47071,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(47071,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(47071,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(47071,36,'cp 60','c2c4 d5b6 e5d6 e7d6 d2d4 f8e7 f1e2 e8g8 e1g1 f8e8 h2h3 e7f6 b1c3 c8f5 c1f4 b8c6 b2b3 a7a5 d1d2 h7h6 a1e1 a5a4 d4d5 c6b4 c3b5 a4b3 a2b3 a8a2 d2b4 e8e2 e1e2 a2e2 f4e3 f5g6 e3b6 c7b6 b5d6 g6h5 d6f5 h5f3 g2f3 f6e5 b4e7 d8e7 f5e7 g8f8');
INSERT INTO evaluations VALUES(47066,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(47066,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(47066,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(47066,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(47066,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(47066,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(47066,36,'cp 60','d5b6 e5d6 e7d6 d2d4 f8e7 f1e2 e8g8 e1g1 e7f6 b1c3 b8c6 h2h3 c8f5 c1f4 a7a5 b2b3 a5a4 d1d2 f5g6 a1e1 a4b3 a2b3 c6b4 c3b5 b4c2 e1d1 c2a3 b5c3 a3c2 f4g3 f8e8 e2d3 c2b4 d3g6 h7g6 c3b5 b4c6 d4d5 c6e5 f3e5 f6e5 g3e5 e8e5');
INSERT INTO evaluations VALUES(46921,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(46921,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(46921,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(46921,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(46921,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(46921,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46921,36,'cp 59','e5d6 e7d6 d2d4 f8e7 f1e2 e8g8 e1g1 c7c6 h2h3 c8f5 b1c3 d6d5 c4c5 b6d7 b2b4 a7a6 a2a4 f8e8 e2d3 f5d3 d1d3 d7f8 c1e3 e7f6 b4b5 f8e6 b5c6 b8c6 a1b1 d8d7 c3e2 c6a5 e2f4 e6f4 e3f4 d7a4 f4e5 f6e5 f3e5 g7g6');
INSERT INTO evaluations VALUES(46914,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(46914,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(46914,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(46914,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(46914,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(46914,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(46914,36,'cp 72','e7d6 d2d4 f8e7 h2h3 c8f5 f1d3 f5d3 d1d3 e8g8 e1g1 b8c6 b2b3 f8e8 b1c3 c6b4 d3d2 c7c6 d4d5 c6d5 c1a3 a7a5 a3b4 a5b4 c3d5 b6d5 d2d5 d8b6 f1e1 g7g6 d5d2 e7f6 e1e8 a8e8 a1d1 f6c3 d2d6 b6d6 d1d6');
INSERT INTO evaluations VALUES(52958,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(52958,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(52958,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52958,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(52958,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(52958,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(52958,36,'cp 65','d6d5 e2e3 g8f6 g1f3 g7g6 f1d3 f8g7 e1g1 c8g4 c4d5 c6d5 d1b3 g4f3 b3b7 f3g2 g1g2 b8d7 c3d5 e8g8 d5f6 d7f6 b7f3 e7e5 b2b3 e5d4 c1a3 f8e8 e3d4 d8d4 a1d1 d4e5 d3c4 e5g5 g2h1 a8d8 a3c1 g5h4 c1f4 h4g4 f3g4 f6g4 d1d8');
INSERT INTO evaluations VALUES(31159,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(31159,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31159,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31159,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(31159,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31159,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31159,36,'cp 21','e7e6 f1e2 c6c5 g1f3 b8c6 e1g1 f5g6 a2a3 c5d4 c3d4 g8e7 h2h3 a7a6 b1c3 h7h6 c1f4 g6h5 g2g4 h5g6 a1c1 e7c8 c3a4 c8b6 b2b4 a8c8 a4c5 f8c5 d4c5 b6d7 d1d2 e8g8 f1d1 g6e4 d2e3 e4f3 e2f3');
INSERT INTO evaluations VALUES(31527,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31527,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(31527,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(31527,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31527,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(31527,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31527,36,'cp 26','f1e2 c6c5 g1f3 b8c6 e1g1 c5d4 f3d4 g8e7 d4f5 e7f5 e2d3 d5d4 f1e1 d8d7 d3f5 e6f5 c3d4 d7d4 b1c3 f8c5 c1e3 d4d1 a1d1 c5e3 e1e3 f5f4 e3e4 a8d8 c3d5 e8g8 e4f4 f8e8 f4e4 c6e5 f2f4 e5c6 e4e8 d8e8 d5c3 g8f8 g1f2 a7a6 d1d5');
INSERT INTO evaluations VALUES(9350,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(9350,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9350,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(9350,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9350,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9350,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(9350,36,'cp 5','h7h6 c2c3 a7a5 b1d2 c8e6 c4b5 e8g8 f1e1 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 a2a4 e7g6 b5c4 e6c4 e3c4 a7c5 c4e3 c5c6 d1c2 f8e8 c3c4 g6f4 a1d1 c6b6 b2b3 g8h8 f3h4 g7g6 h4f3 f4e6 e3d5 f6d5');
INSERT INTO evaluations VALUES(8847,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8847,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8847,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(8847,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8847,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8847,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(8847,36,'cp 0','c2c3 a7a5 b1d2 c8e6 f1e1 e8g8 c4b5 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 b5a4 a7b6 d1d2 e7g6 a4c2 a5a4 a2a3 e6d7 e3c4 b6a6 c4e3');
INSERT INTO evaluations VALUES(8840,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8840,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(8840,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8840,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8840,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8840,36,'cp 0','a7a5 b1d2 c8e6 f1e1 e8g8 c4b5 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 b5a4 e7g6 d3d4 c7c6 a4c2 b7b5 b2b3 a7c7 d1d2 f8b8 e1d1 a5a4 b3a4 b5a4 c3c4 a4a3 d2c3 c7a5 c3a5 a8a5 a1b1 b8b1 d1b1');
INSERT INTO evaluations VALUES(22571,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(22571,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(22571,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(22571,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(22571,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(22571,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(22571,36,'cp 96','c6e5 f2f4 e5g6 d5e6 d7e6 d1d8 e8d8 c1d2 c8d7 b1c3 f8d6 g1h3 d8e8 e1c1 a7a6 f1e2 d7c6 h1e1 e6e5 f4f5 g6e7 g2g4 g8f6 h3f2 d6c5 f2d3 c5d6 e2f3 h7h5 g4g5 f6d7 e1f1 b7b5 a2a3 a6a5 c1b1 h5h4 d2e3 b5b4 a3b4 a5b4 c3a2');
INSERT INTO evaluations VALUES(22979,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(22979,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(22979,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(22979,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(22979,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(22979,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22979,36,'cp 82','b1c3 g8f6 f2f4 e5g6 d5e6 d7e6 d1d8 e8d8 c1d2 f8b4 e1c1 d8e7 g1e2 h8d8 g2g3 b7b6 a2a3 b4c3 e2c3 e7f8 f1d3 c8b7 h1e1 d8d7 d3b5 d7d8 h2h3 h7h6 b5d3 d8d7 a3a4 c7c5 e4e5 f6e8 d3b5 d7d8 b2b3 b7f3 b5e2 f3e2 e1e2 e8c7 e2f2');
INSERT INTO evaluations VALUES(34326,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(34326,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(34326,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(34326,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(34326,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(34326,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(34326,36,'cp 69','d2d4 g8f6 c1f4 c8f5 f3e5 b8d7 e5c4 a5d8 c4e3 f5e6 a2a4 g7g6 a4a5 f6h5 f4g5 h7h6 g5h4 f8g7 f1e2 d7f6 e1g1 h5f4 e2f3 a8c8 f1e1 e8g8 h2h3 b7b5 a5b6 a7b6 d1d2 f6h5 a1d1 d8c7 c3e2 c8d8 c2c3 e6b3 d1b1 f4e2 f3e2 h5f4 e2f3 f8e8 b1a1 b3e6');
INSERT INTO evaluations VALUES(43190,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(43190,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43190,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43190,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(43190,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43190,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43190,36,'cp 10','c7c6 f1g2 d7d5 c4d5 c6d5 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 e8g8 e1g1 a8c8 h2h4 d8e7 f1c1 f8d8 e2e3 c8c6 c3e2 d8c8 c1c6 d7c6 a1c1 g7f8 e2f4 f6d7 f4d3 d7e5 d3e5 e7b4 d2b4 f8b4 c1c2 b4d6 e5c6 c8c6 c2c6');
INSERT INTO evaluations VALUES(39984,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(39984,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(39984,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(39984,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(39984,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(39984,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(39984,36,'cp 18','b1c3 d7d5 d1b3 d8b6 f1g2 e8g8 e1g1 c8f5 h2h3 f8d8 f1e1 f5e4 b3b6 a7b6 c4d5 e4f3 g2f3 f6d5 c3d5 c6d5 e2e3 b8c6 e1d1 b6b5 a2a3 e7e6 c1d2 g7f8 a1c1 d8c8 f3e2 b5b4 a3b4 c6b4 c1c8 a8c8 d1c1 c8c1 d2c1 b4c2 e2b5 f8e7');
INSERT INTO evaluations VALUES(53056,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(53056,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53056,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53056,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53056,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53056,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53056,36,'cp 0','g7g6 g1e2 f8g7 d2d3 b8c6 f1g2 e7e6 c1e3 c6d4 e3d2 d4c6');
INSERT INTO evaluations VALUES(52617,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52617,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52617,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52617,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52617,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52617,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52617,36,'cp 0','g1e2 f8g7 f1g2 b8c6 d2d3 e7e6 c1e3 g8e7 h2h4 h7h5 e1g1 e6e5 d1d2 c6d4 c3d1 c8e6 c2c3 d4e2 d2e2 e8g8 c3c4 e7c6 d1c3 d8d7 e2d2 a8b8 g1h2 a7a6 a2a4 c6d4 a1b1 d4b3 d2c2 b3d4 c2d2');
INSERT INTO evaluations VALUES(23107,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(23107,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(23107,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(23107,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(23107,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(23107,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23107,36,'cp 30','d4f3 d1f3 d8f6 f3g3 f8c5 b1c3 g8e7 d2d3 d7d6 c1g5 f6g6 g5e7 e8e7 c3d5 e7d8 b2b4 c5b6 g3g6 f7g6 a2a4 c7c6 d5b6 a7b6 e1d2 g6g5 d2e3 g5g4 h2h4 g4h3 g2h3 h8f8 f2f3 d8e7 b4b5 c6b5 a4b5 a8a1 h1a1 c8h3 a1a7');
INSERT INTO evaluations VALUES(23101,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(23101,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(23101,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(23101,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(23101,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(23101,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(23101,36,'cp 43','d1f3 d8f6 f3g3 f8c5 b1c3 g8e7 d2d3 d7d6 c1g5 f6g6 g5e7 e8e7 c3d5 e7d8 g3g6 f7g6 b2b4 c5b6 a2a4 c7c6 d5b6 a7b6 e1d2 g6g5 f2f3 c8d7 c2c3 d8e7 h2h3 d7e6 c4e6 e7e6 d3d4 a8a7 a1a3 h8a8 h1a1 h7h5 a4a5 g5g4 h3h4 d6d5 f3g4 h5g4 d2d3 b6a5');
INSERT INTO evaluations VALUES(23098,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(23098,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(23098,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(23098,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(23098,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(23098,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(23098,36,'cp 29','d8f6 f3g3 f8c5 b1c3 g8e7 d2d3 d7d6 c1g5 f6g6 g5e7 e8e7 c3d5 e7d8 b2b4 c5b6 g3g6 f7g6 a2a4 c7c6 d5b6 a7b6 e1d2 g6g5 h2h3 c8d7 c2c3 d8e7 c4b3 a8a7 f2f3 d7e6 b3e6 e7e6 a1a2 h8a8 h1a1 a7a6 d3d4 b6b5 a4a5 h7h6 d2c2 b7b6 c2d3 e5d4 c3d4 b6a5');
INSERT INTO evaluations VALUES(26024,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26024,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26024,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(26024,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(26024,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(26024,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(26024,36,'cp 17','g7g6 c1e3 g8f6 f2f3 a7a6 b5c6 b7c6 d1d2 a6a5 e1c1 a5a4 h1e1 e7e5 d4e2 h7h6 h2h3 d7e6 f3f4 f8e7 e2g3 d8a5 a2a3 e6c4 f4f5 g6g5 c1b1 a5c7 d2f2 e8g8 e3c1 f8b8 h3h4 g5g4 f2e3 d6d5 e3h6 d5d4 g3h5 f6h5 h6h5 d4c3');
INSERT INTO evaluations VALUES(26000,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(26000,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(26000,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(26000,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26000,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26000,36,'cp 25','c1e3 g8f6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 e1g1 d8a5 a1b1 f8g7 b2b4 a5e5 c3a4 f6d7 c2c4 e8g8 d1d2 a8d8 b1c1 c6c5 b4c5 d7c5 a4c5 d6c5 d2a5 d8c8 a5a6 c8a8 a6b6 a8a2 b6c5 e5c5');
INSERT INTO evaluations VALUES(25998,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25998,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(25998,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(25998,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(25998,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25998,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(25998,36,'cp 17','g8f6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 e1g1 f8g7 c3a4 e8g8 d1d2 d8b8 a1e1 d6d5 e4e5 b8e5 e3c5 e5b8 e1e7 b8b5 b2b3 f8e8 e7e8 a8e8 f1e1 e8e1 d2e1 f6d7 c5f2 b5b8 e1e7 b8e5 e7e5 g7e5 h2h3 f7f5 a4c5 d7c5 f2c5 g8f7 g1f2 f7e6 f2e2 e5d6 c5e3 h7h5');
INSERT INTO evaluations VALUES(24457,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(24457,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(24457,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24457,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24457,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(24457,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24457,36,'cp 20','f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 e1g1 f8g7 c3a4 e8g8 d1d2 d8b8 c2c4 f6d7 a1d1 c6c5 b2b3 b8b7 e3h6 g7d4 g1h1 f8e8 h2h4 d7f8 h6e3 f8e6 h4h5 a6a5 e3d4 c5d4 a4b2 g6g5 b2d3 h7h6 h1g1 b7c7 f3f4 g8h8 f4g5 e6g5');
INSERT INTO evaluations VALUES(52616,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(52616,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(52616,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(52616,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(52616,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(52616,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(52616,36,'cp 68','d2d4 c5d4 f3d4 g8f6 c1e3 b8c6 f2f3 h7h5 f1c4 c8d7 d4c6 b7c6 h2h3 f8g7 e1g1 e8g8 a1b1 d8c7 d1e1 d7c8 b1d1 c8b7 e3f4 f6d7 g1h1 a7a5 f4g5 d7e5 c4e2 a5a4 a2a3');
INSERT INTO evaluations VALUES(52613,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(52613,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(52613,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(52613,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(52613,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(52613,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(52613,36,'cp 66','c5d4 f3d4 g8f6 c1e3 b8c6 f2f3 f8g7 d1d2 e8g8 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4g5 f6d7 d4g7 g8g7 c3d5 a5d2 e1d2 d7e5 f3f4 e5g4 f1d3 e7e6 d5e3 d6d5 e4d5 e6d5 a1e1 f8d8 e3g4 c8g4 e1e7 d8d7 h1e1 a8c8');
INSERT INTO evaluations VALUES(52645,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(52645,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(52645,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(52645,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(52645,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(52645,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(52645,36,'cp 71','f3d4 g8f6 c1e3 f8g7 f2f3 b8c6 d1d2 e8g8 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 d5c3 d2c3 g7h6 d4e3 h6e3 c3e3 d8b6 d1e1 c8e6 h2h4 f8d8 h4h5 b6e3 e1e3 a7a5 f1d3 g8g7 h1h4 d8d6 b2b3 c6c5 h5g6 h7g6 h4e4 d6d4 e4e6 f7e6 e3e6 g7f7 e6c6 a8d8 c6c5');
INSERT INTO evaluations VALUES(10882,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(10882,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(10882,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(10882,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10882,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10882,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(10882,36,'cp 0','c2c4 e7c5 b1c3 h7h6 g5e3 d7d6 f1e2 c8g4 f3d2 g4e2 d1e2 c6d4 e3d4 c5d4 d2f3 e8g8 e1g1 d4b6 a1d1 f8e8 f1e1 c7c6 a2a3 a7a5 g2g3 f6g4 e1f1 g4f6');
INSERT INTO evaluations VALUES(41297,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(41297,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(41297,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(41297,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(41297,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(41297,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(41297,36,'cp 69','b1c3 e8g8 f1e2 e7e5 e1g1 b8a6 f1e1 e5d4 f3d4 a6c5 e2f1 f6g4 h2h3 d8f6 h3g4 f6d4 d1f3 c8d7 c1e3 d4f6 f3f6 g7f6 c3d5 f6b2 a1b1 b2e5 d5c7 a8c8 c7d5 g8h8 f2f4 e5g7 e3c5 c8c5 b1b7 d7g4 b7a7 g4e6 e1d1');
INSERT INTO evaluations VALUES(52129,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(52129,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52129,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52129,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52129,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(52129,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(52129,36,'cp 29','f2f4 e5f4 c1f4 b8c6 b1c3 f8e7 c3d5 d7d6 g1e2 c8e6 g2g3 g8f6 f1g2 e6d5 e4d5 c6e5 e1g1 e5g6 f4d2 b7b5 d3d4 e8g8 d4c5 d6c5 e2c3 b5b4 c3b5 f6d5 c2c4 b4c3 b5c3 d5c3 d2c3 d8d1 a1d1 a8d8 d1d8 f8d8');
INSERT INTO evaluations VALUES(51965,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(51965,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(51965,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(51965,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(51965,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(51965,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(51965,36,'cp 27','d4d5 g8f6 b1c3 a7a6 e3e4 b7b5 g1f3 c8b7 c1g5 h7h6 g5f6 d8f6 g2g3 d7d6 f1g2 f8e7 e1g1 e8g8 f1e1 a8a7 d5e6 f7e6 e4e5 d6e5 c3e4 f6f5 f3d2 f5f7 a2a4 b5b4 d2c4 b7e4 e1e4 f7f2 g1h1 b8c6 e4e1');
INSERT INTO evaluations VALUES(3626,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3626,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3626,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3626,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3626,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3626,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3626,36,'cp 0','a7a6 c3e2 c5a7 c1e3 c6e7 e2g3 e7g6 c4a2 b7b5 d1d2 a7e3 f2e3 c8e6 a2e6 f7e6 d2e2 d8e7 f1f2 c7c5 a1f1 a8c8 f3d2 f6d7 f2f8 c8f8 b2b3 f8f1 d2f1 d7f6 g3h5 f6d7');
INSERT INTO evaluations VALUES(2593,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2593,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2593,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2593,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2593,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2593,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2593,36,'cp 0','c3e2 c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 c6e7 e2g3 e7g6 d1e2 d8d7 f3h2 d7c6 g3h5 f6h5 e2h5 g6h8 h5e2 h8g6');
INSERT INTO evaluations VALUES(49371,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(49371,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(49371,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(49371,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(49371,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(49371,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49371,36,'cp 8','d7d5 c4d5 e6d5 e4e5 f6e4 g1f3 c8f5 d2d4 f8b4 c1d2 b4c3 d2c3 e8g8 f1d3 c7c5 d4c5 b8d7 d3e4 d5e4 f3d4 f5g6 d4b3 f8e8 e1g1 d8e7 e5e6 e7e6 d1d6 d7e5 d6e6 e8e6 c3e5 e6e5 a1c1 e4e3 f2e3 e5e3 b3a5');
INSERT INTO evaluations VALUES(48229,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(48229,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(48229,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(48229,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(48229,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(48229,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(48229,36,'cp 20','c4d5 e6d5 e4e5 f6e4 g1f3 c8f5 d2d4 f8b4 c1d2 b8c6 a2a3 b4a5 f1e2 e8g8 e1g1 f5g4 h2h3 g4f3 e2f3 c6d4 f3e4 d5e4 d1g4 a5c3 d2c3 d8d5 a1e1 a8d8 e1e4 d4b5 f1e1 d5e6 c3b4 f8e8 f2f3 d8d5 g4e6 f7e6');
INSERT INTO evaluations VALUES(48189,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48189,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(48189,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(48189,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(48189,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(48189,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(48189,36,'cp 22','e6d5 e4e5 f6e4 g1f3 c8f5 d2d4 f8b4 c1d2 b4c3 d2c3 e8g8 f1e2 c7c5 d4c5 e4c3 b2c3 f5e4 e1g1 d8e7 d1d2 b8c6 f3d4 e7c5 e5e6 c6d4 c3d4 c5d6 e6f7 f8f7 a1c1 a8f8 f2f3 e4g6 e2b5 h7h6 f1e1 f7c7 c1c7 d6c7 e1e5 c7d6 b5f1 f8c8 d2e3 g6f7');
INSERT INTO evaluations VALUES(48705,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(48705,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(48705,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(48705,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(48705,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(48705,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(48705,36,'cp 30','e4e5 f6e4 g1f3 c8f5 d2d3 e4c3 b2c3 c7c5 d3d4 c5c4 a2a4 f8e7 f1e2 b8c6 e1g1 e8g8 f3e1 f7f6 e5f6 e7f6 e1c2 f5e4 f1e1 d8a5 c1d2 a8e8 e2g4 g7g6 f2f3 e4d3 g2g3 f6g7 f3f4 g8h8 c2e3 e8d8 g4f3 f8e8');
INSERT INTO evaluations VALUES(19816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19816,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19816,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19816,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19816,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(19816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19816,36,'cp 0','c5d4 c3d4 c8g4 c1e3 e7e6 f1d3 f8b4 b1c3 g8e7 a1c1 b4c3 c1c3 d8b6 d3b1 b6a6 c3a3 a6b5 a3b3 b5a5 b3c3');
INSERT INTO evaluations VALUES(19889,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19889,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(19889,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(19889,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(19889,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19889,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(19889,36,'cp 0','c3d4 c8g4 c1e3 e7e6 a2a3 g8e7 b1d2 e7f5 d1b3 a7a6 f1d3 c6a5 b3a2 b7b5 h2h3 g4h5 d3f5 e6f5 e1g1 f8e7 a2b1 d8d7 e3g5 h5f3 d2f3 e7g5 f3g5 a5b3 e5e6 f7e6 f1e1 b3d4 g5e6 d4e6 b1f5 e8e7 f5g5 e7f8 g5f5');
INSERT INTO evaluations VALUES(19885,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19885,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19885,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19885,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19885,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19885,36,'cp 0','c8g4 c1e3 e7e6 a2a3 g8e7 b1d2 e7f5 d1b3 a7a6 f1d3 f8e7 h2h3 g4h5 g2g4 f5e3 f2e3 h5g6 d3g6 h7g6 e1e2 d8d7 b3d3 e8c8 b2b4 f7f6 b4b5 a6b5 a1b1 f6e5 f3e5 c6e5 d4e5 d5d4 e3d4 d7d5 d2f3 d5a2 e2e3 a2g2 b1g1 g2a2');
INSERT INTO evaluations VALUES(26309,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26309,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26309,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26309,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26309,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26309,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26309,36,'cp 0','c1e3 e7e6 f1d3 f8b4 b1c3 f7f6 h2h3 g4f3 d1f3 f6e5 f3h5 e8f8 d4e5 d5d4 h5f3 f8e8 f3h5');
INSERT INTO evaluations VALUES(26306,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26306,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26306,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26306,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26306,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26306,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26306,36,'cp 0','e7e6 f1d3 f8b4 b1c3 f7f6 h2h3 g4f3 d1f3 f6e5 f3h5 e8f8 d4e5 d5d4 h5f3 f8e8 f3h5');
INSERT INTO evaluations VALUES(26429,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26429,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26429,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26429,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26429,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26429,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26429,36,'cp 0','f1d3 f8b4 b1c3 f7f6 h2h3 g4f3 d1f3 f6e5 f3h5 e8f8 d4e5 d5d4 h5f3 f8e8 f3h5');
INSERT INTO evaluations VALUES(16732,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(16732,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(16732,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(16732,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(16732,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(16732,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(16732,36,'cp 25','b1c3 d7d6 h5d1 g7g6 a2a4 f8g7 d2d3 c6d4 g1e2 c8e6 h2h3 c7c6 e2d4 e5d4 c3e2 d6d5 e4d5 e6d5 e1g1 e8g8 e2f4 d5c4 d3c4 h6f5 f1e1 d8d7 f4d3 a8e8 e1e8 f8e8 c1f4 h7h5 b2b3 b7b6 d1d2 c6c5 a1e1 e8e1 d3e1 d7c6 d2d3 h5h4 e1f3');
INSERT INTO evaluations VALUES(45206,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45206,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45206,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45206,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45206,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45206,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45206,36,'cp 0','f1e2 g7g6 b1c3 b8c6 c3d5 f8g7 c2c3 e8g8 e1g1 f6d5 e4d5 c6e5 f1e1 b7b6 d3d4 e5f3 e2f3 c5d4 c3d4 c8d7 c1g5 f8e8 d1d2 a8c8 h2h3 b6b5 b2b3 a7a5 e1e2 d8b6 a1d1 g7f6 h3h4 h7h5 d2f4 g8g7 g5h6 g7g8');
INSERT INTO evaluations VALUES(45205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45205,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(45205,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(45205,36,'cp 0','g7g6 b1c3 b8c6 c3d5 f8g7 c2c3 e8g8 e1g1 f6d5 e4d5 c6e5 f1e1 b7b6 d3d4 e5f3 e2f3 c5d4 c3d4 c8d7 c1g5 f8e8 d1d2 a8c8 h2h3 b6b5 b2b3 a7a5 e1e2 g7f6 h3h4 h7h5 d2f4 g8g7 g5h6 g7h7 h6g5');
INSERT INTO evaluations VALUES(44773,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44773,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44773,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(44773,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44773,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44773,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44773,36,'cp 0','b1c3 b8c6 c3d5 f8g7 c2c3 e7e5 c1g5 h7h6 g5f6 g7f6 h2h4 h6h5 f3g5 f6g7 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(44763,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44763,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44763,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44763,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44763,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44763,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44763,36,'cp 0','b8c6 c3d5 f8g7 c2c3 e7e5 c1g5 h7h6 g5f6 g7f6 h2h4 h6h5 f3g5 f6g7 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(13017,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13017,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13017,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13017,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13017,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13017,36,'cp 0','c3d5 f8g7 c2c3 e7e5 c1g5 h7h6 g5f6 g7f6 h2h4 h6h5 f3g5 f6g7 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(48055,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(48055,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(48055,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(48055,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(48055,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(48055,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(48055,36,'cp 83','g1f3 b8d7 c1f4 c7c6 h2h3 d8a5 f1d3 e6e5 f4h2 f8e7 d1d2 e8g8 e1g1 a5c7 a2a4 f8e8 a4a5 e7f8 f1e1 e5d4 f3d4 d7c5 b2b4 c5d3 c2d3 b7b5 f2f4 a7a6 e4e5 d6e5 f4e5 c7d7 d4c2 f6d5 c3e4 f8e7 e4d6 e7d6 e5d6');
INSERT INTO evaluations VALUES(56287,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(56287,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(56287,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(56287,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56287,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(56287,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(56287,36,'cp -10','g8f6 d2d4 f8d6 c1g5 e8g8 e1g1 c8e6 b1d2 h7h6 g5h4 b8d7 a2a4 c7c6 h4g3 d6c7 f1e1 e6f5 c2c3 c7g3 h2g3 a7a5 d2f1 f8e8 e2d3 f5e4 d3e4 f6e4 f1d2 d8b6 d2e4 e8e4 e1e4 d5e4 f3d2 d7f6');
INSERT INTO evaluations VALUES(48745,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48745,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(48745,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48745,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48745,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(48745,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(48745,36,'cp -9','d2d4 f8d6 e1g1 e8g8 c1g5 c8e6 b1d2 b8d7 f1e1 c7c6 c2c4 h7h6 g5h4 f8e8 e2d3 d5c4 d3c4 e6d5 e1e8 d8e8 d1c2 e8c8 h4g3 d6g3 h2g3 d7b6 c4d5 b6d5 a2a4 a7a5 f3e5 c8c7 c2b3 c7b6 a1c1 b6b3');
INSERT INTO evaluations VALUES(47084,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(47084,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(47084,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(47084,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(47084,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(47084,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(47084,36,'cp 34','b8d7 g1f3 f6d5 d2d4 a7a6 b5d7 c8d7 c2c4 d5b6 d1e2 d7f5 b1c3 e7e6 e1g1 f8b4 h2h3 e8g8 c4c5 b4c3 b2c3 b6a4 c3c4 b7b6 c5b6 c7b6 c1f4 f8e8 f1e1 a8c8 a1c1 f7f6 e2d2 f5g6 d2b4 d8d7 b4a3 h7h6');
INSERT INTO evaluations VALUES(16024,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(16024,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(16024,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(16024,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(16024,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(16024,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(16024,36,'cp 62','g1f3 f6d5 d2d4 a7a6 b5e2 e7e6 c2c4 f8b4 c1d2 b4d2 d1d2 d5e7 e1g1 b7b6 b1c3 c8b7 a1d1 d7f6 f3e5 d8d6 f1e1 c7c5 d2f4 e8g8 d4c5 d6c5 e5d3 c5c8 e2f1 f8d8 d3e5 e7g6 f4e3 c8c7');
INSERT INTO evaluations VALUES(31695,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(31695,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31695,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(31695,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31695,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31695,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(31695,36,'cp 26','f1b5 b8c6 f3e5 a8c8 g2g4 f5g6 d1e2 f8d6 b5c6 c8c6 e5c6 b7c6 c1e3 h7h5 f2f3 h5g4 f3g4 h8h3 e1c1 g8f6 e3d2 f6e4 c3e4 g6e4 h1f1 d8b6 g4g5 b6b5 e2b5 c6b5 f1f2 d6h2 d1f1 h2d6 d2f4 d6e7 b2b3 b5b4 f1g1 e4f5 f2h2 h3f3 f4e5 f7f6 g5f6 g7f6 e5b8 e8f7 c1b1 a7a5 b1b2');
INSERT INTO evaluations VALUES(31626,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(31626,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(31626,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(31626,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(31626,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31626,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(31626,36,'cp 21','b8c6 f3e5 a8c8 g2g4 f5g6 d1e2 f8d6 b5c6 c8c6 e5c6 b7c6 c1e3 h7h5 f2f3 h5g4 f3g4 g8f6 e1c1 f6e4 c3e4 g6e4 h1f1 h8h2 f1f2 h2h3 e3f4 d6f4 f2f4 d8c7 d1f1 f7f6 f4e4 d5e4 e2e4 c7d6 g4g5 f6f5 f1f5 h3h1 e4h1 e6f5 h1f3 d6d4 f3c6 e8e7 c2c3 d4e3 c1c2 e3g5 c6b7 e7f6 b7a6 f6e5');
INSERT INTO evaluations VALUES(26374,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(26374,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(26374,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(26374,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(26374,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(26374,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(26374,36,'cp 12','f3e5 a8c8 g2g4 f5g6 b5c6 b7c6 d1e2 g8f6 h2h4 d8b6 h4h5 g6e4 c3e4 f6e4 c2c3 h7h6 c1f4 c6c5 f2f3 e4g5 e1c1 c5d4 d1d4 f8e7 c1b1 c8c3 d4a4 c3c8 f4e3 e7c5 h1c1 e8g8 c1c5 c8c5 e5d7 b6b5 e2b5 c5b5 d7f8 g8f8 a4a7');
INSERT INTO evaluations VALUES(26381,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(26381,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26381,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(26381,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26381,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(26381,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(26381,36,'cp 16','a8c8 g2g4 f5g6 b5c6 b7c6 d1e2 g8f6 h2h4 d8b6 h4h5 g6e4 c3e4 f6e4 c2c3 h7h6 c1f4 c6c5 f2f3 e4g5 e1c1 c5d4 d1d4 f8e7 d4a4 e8g8 c1b1 a7a5 f4g5 e7g5 f3f4 g5f6 e5d7 b6c6 d7f6 g7f6 a4d4 c8b8 b1a1 b8b5 h1d1 f8b8 d1d2 b8b7 a2a3 g8g7 e2e3');
INSERT INTO evaluations VALUES(23218,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(23218,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(23218,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(23218,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(23218,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(23218,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(23218,36,'cp 220','g8e7 d2d4 e7g6 c2c3 c6a5 c4d3 a5c6 d1b3 d7d6 a2a4 d8d7 a4a5 a7a6 d3c4 f8e7 c1e3 c6d8 b1d2 d8e6 h2h4 e8g8 h4h5 g6f4 e3f4 e5f4 f1e1 g8h8 c4d5 e6d8 e4e5 f6e5 d4e5 d7g4 h5h6 g7h6 e5d6 e7d6 d2c4 g4g7 c4d6 c7d6 a1d1');
INSERT INTO evaluations VALUES(17140,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(17140,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(17140,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(17140,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(17140,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(17140,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(17140,36,'cp 205','d2d4 c6d4 f3d4 e5d4 c2c3 d4d3 a2a4 c7c6 d1d3 d7d5 e4d5 e7d5 f1e1 f8e7 d3e4 e8f7 b1d2 d8d7 d2f3 h8e8 e4h7 d7f5 h7f5 c8f5 f3d4 f5d7 d4b5 a8c8 c1f4 g7g5 f4g3 f6f5 g3e5 d7e6 b5d4 e6d7 a4a5 e7f6 e5f6 e8e1 a1e1 f7f6 c4d5 c6d5 f2f3 f5f4 g1f2 a7a6 e1a1 c8h8 h2h3 h8e8 b2b4 e8e3');
INSERT INTO evaluations VALUES(17144,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(17144,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(17144,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(17144,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(17144,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(17144,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(17144,36,'cp 227','c6d4 f3d4 e5d4 c2c3 d4d3 a2a4 c7c6 d1d3 d7d5 e4d5 e7d5 f1e1 f8e7 d3e4 e8f7 b1d2 d8d7 d2f3 h8e8 e4h7 d7f5 h7f5 c8f5 c1e3 a7a6 f3d4 f5d7 f2f3 e7d6 a4a5 e8h8 h2h3 h8e8 d4b3 d7e6 e3c5 d6c7 b3d4 e6d7 g1f2 e8e1 a1e1 a8e8 e1e8 f7e8 d4b3 d7f5 c5d4');
INSERT INTO evaluations VALUES(43173,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(43173,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(43173,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(43173,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(43173,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43173,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43173,36,'cp 0','e8g8 e2e4 d7d6 g1f3 c8g4 f1e2 b8c6 d4d5 e7e5 f4g5 g4f3 e2f3 c6d4 e1g1 c7c5 d5c6 b7c6 b2b4 d8c7 d1d2 f6d7 f3e2 a8b8 a1b1 b8b7 b4b5 f8b8 a2a4 d7c5 g5e3 a7a6 g1h1 c7a5 e2d1 c5e6 d1g4 e6c5');
INSERT INTO evaluations VALUES(37391,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(37391,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37391,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(37391,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(37391,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(37391,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(37391,36,'cp 16','e2e4 d7d6 g1f3 e7e5 d4e5 d6e5 d1d8 f8d8 f4e5 b8a6 f1e2 d8e8 a1d1 f6e4 c3e4 g7e5 f3e5 e8e5 d1d8 g8g7 f2f3 b7b6 e1f2 c8b7 d8a8 b7a8 e4c3 a6c5 h1d1 a7a5 a2a3 e5e8 c3b5 e8e7 b5d4 a8b7 b2b4 a5b4 a3b4 c5a4 d4b5 b7c6 b5a7 c6d7 g2g4');
INSERT INTO evaluations VALUES(37416,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(37416,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(37416,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(37416,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(37416,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(37416,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(37416,36,'cp 12','d7d6 g1f3 e7e5 d4e5 d6e5 d1d8 f8d8 f4e5 b8a6 f1e2 d8e8 a1d1 f6e4 c3e4 g7e5 f3e5 e8e5 d1d8 g8g7 e4c3 b7b6 f2f3 a6c5 e1f2 c8b7 d8a8 b7a8 b2b4 c5e6 a2a3 a7a5 h1d1 e5h5 h2h3 h5h4 f2e3 h4h5 e2d3 h5e5 e3f2 g6g5 d3e4 a8e4 c3e4 a5b4 a3b4 g7g6 d1d7 h7h5 e4c3 e6f4 d7c7 f4d3 f2f1 d3b4');
INSERT INTO evaluations VALUES(36587,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36587,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36587,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36587,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(36587,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36587,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36587,36,'cp 12','g1f3 e7e5 d4e5 d6e5 d1d8 f8d8 f4e5 b8a6 f1e2 d8e8 a1d1 f6e4 c3e4 g7e5 f3e5 e8e5 d1d8 g8g7 f2f3 b7b6 e1f2 c8b7 d8a8 b7a8 e4c3 a6c5 a2a3 c5e6 b2b4 a8c6 h1d1 e5h5 h2h3 h5h4 e2f1 h4d4 d1d4 e6d4 f1d3 g7f6 f2e3 d4f5 d3f5 f6f5 e3d4 f5f4 c3e2 f4f5 h3h4 f7f6 c4c5 f5e6 e2f4 e6f5');
INSERT INTO evaluations VALUES(54986,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(54986,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54986,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54986,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(54986,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54986,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54986,36,'cp 0','e2e4 c7c5 g1f3 g7g6 b1c3 f8g7 d2d4 c5d4 f3d4 b8c6 c1e3 g8f6 f1e2 e8g8 a2a4 c6d4 e3d4 d8a5 e1g1 e7e5 d4e3 e5f4 f1f4 c8e6 e3d4 a5g5 f4f3 f8d8 a4a5 d6d5 d4e3 g5e5 e3d4');
INSERT INTO evaluations VALUES(12042,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(12042,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12042,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12042,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12042,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(12042,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12042,36,'cp 0','h2h3 g7g6 f1c4 f8g7 e4e5 d6e5 d1d8 e8d8 c1g5 d8e8 c4b3 h7h5 e1c1 c8f5 g5e3 g7h6 e3h6 h8h6 g2g4 h5g4 h3g4 f6g4 f2f3 g4h2 d1d2 h2g4');
INSERT INTO evaluations VALUES(30252,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(30252,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(30252,32,'cp 292',NULL);
INSERT INTO evaluations VALUES(30252,33,'cp 312',NULL);
INSERT INTO evaluations VALUES(30252,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(30252,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(30252,36,'cp 296','f3b3 d8e7 b3b7 e7b4 b7b4 f8b4 c1d2 b4d2 b1d2 e8e7 e1c1 a7a5 h1e1 h8d8 f2f3 g7g6 c4f1 f6d7 d2c4 f7f6 d1d5 a5a4 e1d1 b8a6 c4a5 a6b8 a5b7 d8c8 g2g3 c7c6 d5a5 a8a5 b7a5 d7c5 d1d2 c8c7 c2c3 b8d7 a5c4 e7e6 c1c2 e6e7');
INSERT INTO evaluations VALUES(30250,30,'cp 298',NULL);
INSERT INTO evaluations VALUES(30250,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(30250,32,'cp 283',NULL);
INSERT INTO evaluations VALUES(30250,33,'cp 298',NULL);
INSERT INTO evaluations VALUES(30250,34,'cp 272',NULL);
INSERT INTO evaluations VALUES(30250,35,'cp 290',NULL);
INSERT INTO evaluations VALUES(30250,36,'cp 306','d8e7 b3b7 e7b4 b7b4 f8b4 c1d2 b4d2 b1d2 e8e7 e1c1 a7a5 h1e1 h8d8 f2f3 a5a4 e1e3 d8d6 e3c3 f6d7 c4d3 d6c6 d2c4 d7b6 c4e5 c6c3 b2c3 f7f6 e5g4 a8a5 c3c4 b8d7 c1d2 e7d6 d2c3 a4a3 e4e5 d7e5 d3h7 d6c6 d1d4 e5g4 d4g4 g7g5 h7e4 c6c5');
INSERT INTO evaluations VALUES(32683,30,'cp 296',NULL);
INSERT INTO evaluations VALUES(32683,31,'cp 304',NULL);
INSERT INTO evaluations VALUES(32683,32,'cp 300',NULL);
INSERT INTO evaluations VALUES(32683,33,'cp 289',NULL);
INSERT INTO evaluations VALUES(32683,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(32683,35,'cp 285',NULL);
INSERT INTO evaluations VALUES(32683,36,'cp 279','b3b7 e7b4 b7b4 f8b4 c1d2 b4d2 b1d2 e8e7 e1c1 a7a5 h1e1 h8d8 f2f3 a5a4 d2f1 d8d1 e1d1 f6e8 f1e3 e8d6 e3d5 e7d8 d1d3 f7f6 d5e3 b8d7 c4d5 a8a6 d3d2 d8e7 b2b4 a4b3 c2b3 c7c6 d5c4 a6b6 c4g8 h7h5 e3c4 d6c4 g8c4 h5h4 c1c2 d7b8 c2b2 b8d7 b2c3');
INSERT INTO evaluations VALUES(41866,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(41866,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(41866,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(41866,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41866,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41866,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41866,36,'cp 18','e1g1 c5b6 c2c3 h7h6 f1e1 e8g8 h2h3 f8e8 b1d2 c7c6 d3d4 b8d7 c4b3 d7f8 b3c2 f8g6 d2f1 d6d5 f3e5 g6e5 d4e5 e8e5 c1f4 e5e4 c2e4 f6e4 f4e3 b6c7 c3c4 d5c4 e3h6 d8d1 a1d1 g7h6 e1e4 c8e6 f1g3 b7b5 g3h5 g8f8 g2g4');
INSERT INTO evaluations VALUES(41865,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(41865,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(41865,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(41865,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41865,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(41865,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41865,36,'cp 4','b8c6 c2c3 a7a5 h2h3 h7h6 b1d2 e8g8 f1e1 c8e6 c4b5 c5a7 d2f1 c6e7 d3d4 e5d4 f3d4 e6d7 b5d3 e7g6 f1g3 a5a4 d4f3 d7c6 c1e3 a7e3 e1e3 f6d7 f3d4 d7e5 d1d2 d8g5 a1d1 e5d3 d2d3 c6d7 d3c2 f8e8');
INSERT INTO evaluations VALUES(47133,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(47133,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(47133,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(47133,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(47133,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(47133,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(47133,36,'cp 47','c4d5 f6d5 g1f3 c7c5 d2d4 d5c3 b2c3 g7g6 e2e3 f8g7 f1b5 c8d7 b5e2 e8g8 e1g1 d7c6 c1a3 c5d4 c3d4 f8e8 d1b3 e7e6 f1c1 b8d7 a3d6 d8a5 a2a4 e8c8 d6b4 a5d5 b3d5 c6d5 a4a5 c8c1 a1c1 g7f8 b4f8 g8f8 f3d2 d7f6');
INSERT INTO evaluations VALUES(47103,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(47103,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(47103,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(47103,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(47103,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(47103,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(47103,36,'cp 39','f6d5 g1f3 c7c5 d2d4 d5c3 b2c3 g7g6 e2e4 f8g7 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 f1d3 c8g4 d2a5 c6a5 e1e2 a8d8 d4d5 e7e6 b1b5 b7b6 h2h3 g4f3 e2f3 e6d5 e4d5 a5b7 h1c1 b7d6 b5b4 d6e8 e3f4 d8d5 d3e4 d5c5 c1c5');
INSERT INTO evaluations VALUES(47456,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(47456,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(47456,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(47456,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(47456,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(47456,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(47456,36,'cp 26','g1f3 c7c5 d2d4 d5c3 b2c3 g7g6 e2e3 f8g7 f1b5 c8d7 b5e2 e8g8 e1g1 d7c6 c1a3 c5d4 c3d4 f8e8 d1b3 e7e6 f1c1 b8d7 a3d6 d8a5 a2a4 e8c8 d6b4 a5d8 a4a5 c6d5 b3a3 c8c1 a3c1 a8c8 c1a3 b7b6 a5a6 c8c2');
INSERT INTO evaluations VALUES(47452,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(47452,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47452,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47452,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47452,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(47452,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(47452,36,'cp 35','c7c5 d2d4 d5c3 b2c3 g7g6 e2e3 f8g7 f1b5 c8d7 b5d3 e8g8 e1g1 b8c6 a1b1 d8c7 c1a3 b7b6 h2h3 f8e8 d4c5 c6a5 d3b5 d7b5 b1b5 a5c4 d1b3 c4a3 b3a3 e8c8 c3c4 b6c5 a3a6 a8b8 f1b1 b8b5 b1b5 c8b8 g2g3 b8b6 a6a5 b6c6 a5c7 c6c7 b5b8 g7f8 b8a8 g8g7');
INSERT INTO evaluations VALUES(45503,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(45503,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(45503,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45503,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(45503,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(45503,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(45503,36,'cp 37','d2d4 d5c3 b2c3 g7g6 e2e3 f8g7 f1b5 c8d7 b5e2 e8g8 e1g1 d7c6 a2a4 d8d7 d1b3 f8c8 a4a5 e7e6 f1d1 d7c7 c1b2 c5d4 f3d4 a7a6 c3c4 c6e8 h2h4 b8d7 h4h5 a8b8 d4f3 d7c5 b3b4 g7b2 b4b2 g6h5 b2c3 c5e4');
INSERT INTO evaluations VALUES(16835,30,'cp -182',NULL);
INSERT INTO evaluations VALUES(16835,31,'cp -172',NULL);
INSERT INTO evaluations VALUES(16835,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(16835,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(16835,34,'cp -171',NULL);
INSERT INTO evaluations VALUES(16835,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(16835,36,'cp -173','d7d5 c4b5 f8c5 d2d4 e5d4 b1d2 e8g8 d2e4 d5e4 e1e4 d8d5 e4e1 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 c6b4 b5a4 b4d5 c1g5 c7c6 a4b3 h7h6 b3d5 c6d5 g5e7 f8c8 a2a4 a7a6 e7c5 c8c5 a1d1 c5c2 d1d4 a8d8 e1e5 g8f8 d4d5 d8d5 e5d5 c2b2 d5c5 b7b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(15380,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(15380,31,'cp -178',NULL);
INSERT INTO evaluations VALUES(15380,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(15380,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(15380,34,'cp -174',NULL);
INSERT INTO evaluations VALUES(15380,35,'cp -171',NULL);
INSERT INTO evaluations VALUES(15380,36,'cp -171','c4b5 f8c5 d2d4 e5d4 b1d2 e8g8 d2e4 d5e4 e1e4 d8d5 e4e1 c8g4 b5d3 a8e8 c1f4 c5b4 e1e2 g4f3 g2f3 b4d6 e2e8 f8e8 d3e4 d5b5 f4g3 b5b2 a1b1 b2a2 b1b7 c6d8 b7b1 g7g6 b1a1 a2c4 a1a7 d8e6 d1a1 e6g5 a7a8 e8a8');
INSERT INTO evaluations VALUES(15258,30,'cp -171',NULL);
INSERT INTO evaluations VALUES(15258,31,'cp -171',NULL);
INSERT INTO evaluations VALUES(15258,32,'cp -158',NULL);
INSERT INTO evaluations VALUES(15258,33,'cp -176',NULL);
INSERT INTO evaluations VALUES(15258,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(15258,35,'cp -180',NULL);
INSERT INTO evaluations VALUES(15258,36,'cp -159','f8c5 d2d4 e5d4 b1d2 e8g8 d2e4 d5e4 e1e4 d8d5 e4e1 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 c6b4 b5a4 b4d5 c1g5 c7c6 a4b3 h7h6 b3d5 c6d5 g5e7 f8c8 a2a4 a7a6 e7c5 c8c5 a1d1 c5c2 d1d4 a8d8 e1e7 c2b2 d4f4 b2b1 g1h2 f7f6 f4g4 g7g5');
INSERT INTO evaluations VALUES(9139,30,'cp -153',NULL);
INSERT INTO evaluations VALUES(9139,31,'cp -159',NULL);
INSERT INTO evaluations VALUES(9139,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(9139,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(9139,34,'cp -179',NULL);
INSERT INTO evaluations VALUES(9139,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(9139,36,'cp -172','d2d4 e5d4 b1d2 e8g8 d2e4 d5e4 e1e4 d8d5 e4e1 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 c6b4 b5a4 b4d5 c1g5 c7c6 a4b3 h7h6 b3d5 c6d5 g5e7 f8c8 a2a4 b7b6 e7c5 b6c5 e1e7 a7a6 e7d7 c8d8 d7c7 a8c8 c7a7 c8b8 b2b3 b8a8 a7e7 a8c8 e7a7 c8c6');
INSERT INTO evaluations VALUES(9138,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(9138,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(9138,32,'cp -156',NULL);
INSERT INTO evaluations VALUES(9138,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(9138,34,'cp -164',NULL);
INSERT INTO evaluations VALUES(9138,35,'cp -160',NULL);
INSERT INTO evaluations VALUES(9138,36,'cp -167','e5d4 b1d2 e8g8 d2e4 d5e4 e1e4 d8d5 e4e1 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 c6b4 b5a4 b4d5 a2a3 a7a5 c1g5 c7c6 a4b3 f7f6 g5d2 a5a4 b3a2 f8d8 h3h4 d8e8 e1e4 f6f5 e4e1 g8f8 h4h5 d5f6 e1e8 a8e8 h5h6 g7g6 a1e1');
INSERT INTO evaluations VALUES(9132,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(9132,31,'cp -177',NULL);
INSERT INTO evaluations VALUES(9132,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(9132,33,'cp -164',NULL);
INSERT INTO evaluations VALUES(9132,34,'cp -178',NULL);
INSERT INTO evaluations VALUES(9132,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(9132,36,'cp -178','b1d2 e8g8 d2e4 d5e4 e1e4 d8d5 e4e1 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 c6b4 b5a4 b4d5 c1g5 c7c6 a4b3 h7h6 g5h4 f8e8 e1e8 a8e8 h4g3 e8d8 a1d1 g8f8 a2a4 h6h5 b3c4 c5b6 g3e5 a7a5 h3h4 d5b4 c2c3 d4c3 d1d8 b6d8 e5c3 d8h4 c3b4 a5b4 f3f4 h4f6 b2b3 g7g6');
INSERT INTO evaluations VALUES(4578,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(4578,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(4578,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(4578,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(4578,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(4578,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(4578,36,'cp 25','a2a4 d7d6 e1g1 a7a5 h2h3 c8e6 f1e1 e6c4 d3c4 c6b4 c1e3 c7c6 b2b3 d8c7 d1d2 a8d8 a1d1 f8e8 d2e2 b4a6 e1f1 d6d5 c4d5 e7b4 e2c4 b4c3 d5d6 d8d6 d1d6 c7d6 c4c3 d6b4 c3b2 f6e4 f3e5 b4c3 e5d3 c3b2 d3b2 e4c3');
INSERT INTO evaluations VALUES(52068,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52068,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(52068,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(52068,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(52068,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52068,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52068,36,'cp 0','g8f6 f1d3 b8c6 d4e3 d7d5 e4d5 f6d5 e3e2 f8e7 e1g1 e8g8 f1d1 d8c7 c2c4 d5b4 b1c3 b4d3 d1d3 e7f6 c3e4 f6e7');
INSERT INTO evaluations VALUES(14899,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(14899,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(14899,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(14899,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(14899,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(14899,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(14899,36,'cp 133','g1f3 a7a6 a2a3 f8d6 e3e4 e6e5 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 c4f7 e5c3 b2c3 f6e4 e1g1 c8d7 f1e1 e4d6 f7d5 h8e8 c1g5 d8c8 e1e8 d6e8 a1b1 c7c6 d5b3 a6a5 b1e1 a5a4 b3c2 a8a5 g5e7 a5d5 c2h7 e8f6 e7f6 g7f6 h7c2 d5d2 c2a4');
INSERT INTO evaluations VALUES(39194,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(39194,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(39194,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(39194,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(39194,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(39194,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(39194,36,'cp 221','b1c3 b8c6 c1f4 g8e7 d1d2 e8g8 g1f3 d7d6 e5d6 c7d6 a2a3 d8a5 a1b1 d6d5 e2e3 c5b6 f1d3 d5d4 e3d4 c6d4 f3d4 b6d4 e1g1 d4c3 b2c3 e7g6 f4g3 b7b6 d3e4 c8a6 e4a8 a6f1 a8d5 f1b5 g3d6 f8e8 d2d4 e8d8 d6b4 a5a6');
INSERT INTO evaluations VALUES(49755,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(49755,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(49755,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(49755,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(49755,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(49755,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(49755,36,'cp -22','f6e4 d1e2 d8e7 d4f5 e7e6 f5d4 e6d5 d4b5 e8d8 c1f4 f8b4 c2c3 b4a5 b1d2 e4f6 e1c1 h8e8 e2d3 d5d3 f1d3 d7d5 f4g5 h7h6 g5f6 g7f6 c3c4 a7a6 b5a3 b8d7 c4d5 b7b5 d2b3 a5b6 a3c2 d7c5 b3d4 c8d7 d3f5 d7f5 d4f5 c5e4 d1e1 e4f2 e1e8 d8e8 h1e1 e8f8 e1e2 a8e8 e2e8 f8e8 f5h6');
INSERT INTO evaluations VALUES(49918,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(49918,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(49918,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(49918,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(49918,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(49918,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(49918,36,'cp -22','d1e2 d8e7 d4f5 e7e6 f5d4 e6d5 d4b5 e8d8 c1f4 f8b4 c2c3 b4a5 b1d2 h8e8 e1c1 e4f6 e2d3 d5d3 f1d3 d7d6 f4g5 h7h6 g5f6 g7f6 d2e4 f6f5 e4f6 e8e6 f6h5 b8c6 b5a3 a5b6 d3f5 e6e2 f5c8 a8c8 a3c4 e2f2 h1f1 f2f1 d1f1 d8e7 g2g3 c8d8 h5g7 d8g8 g7f5 e7e6 f5h6 g8h8 h6g4 f7f5 f1e1 e6f7 e1f1 f7g6 c4b6 f5g4');
INSERT INTO evaluations VALUES(49917,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(49917,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(49917,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(49917,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(49917,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(49917,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(49917,36,'cp -27','d8e7 d4f5 e7e6 f5d4 e6d5 d4b5 e8d8 c1f4 f8b4 c2c3 b4a5 b1d2 h8e8 e1c1 e4f6 e2d3 d5d3 f1d3 d7d6 f4g5 h7h6 g5f6 g7f6 d2e4 f6f5 e4f6 e8e6 f6h5 b8c6 b5a3 a5b6 d3f5 e6e2 f5c8 a8c8 a3c4 e2f2 h1f1 f2f1 d1f1 d8e7 g2g3 b6c5 b2b4 c5b6 f1e1 e7f8 h5f6 c8d8 e1d1 f8g7 f6d5 d8e8 a2a4 c6e5 c4e5 e8e5');
INSERT INTO evaluations VALUES(34076,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(34076,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(34076,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(34076,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(34076,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(34076,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(34076,36,'cp -54','d4f5 e7e6 f5d4 e6d5 d4b5 e8d8 c1f4 f8b4 c2c3 b4a5 b1d2 e4f6 e1c1 h8e8 e2d3 d5d3 f1d3 d7d6 f4g5 h7h6 g5f6 g7f6 d2e4 f6f5 e4f6 e8e6 f6h5 b8c6 b5a3 a5b6 d3f5 e6e2 f5c8 a8c8 a3c4 e2f2 h1f1 f2f1 d1f1 d8e7 f1e1 e7f8 h5f6 b6f2 e1e4 b7b5 f6d7 f8g7 e4g4 g7h8');
INSERT INTO evaluations VALUES(44543,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(44543,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(44543,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(44543,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44543,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44543,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44543,36,'cp 8','e7e6 g1f3 d7d5 b1d2 d8b6 a1b1 c5d4 e3d4 c8d7 c2c3 d7b5 f1b5 b6b5 a2a4 b5a6 d1b3 a6b6 b3b5 b8d7 b1a1 f8e7 f3e5 a7a5 b5b6 d7b6 e1d1 f6d7 e5d3 h7h5 h2h4 e8g8 f4e3 f8c8 d1c2 a8a6 f2f3 e7d6 e3f4 d6f4 d3f4 d7f6');
INSERT INTO evaluations VALUES(46253,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46253,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46253,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(46253,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(46253,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(46253,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(46253,36,'cp 44','c2c4 d5c4 e1g1 b8c6 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 c3c4 e6e5 c1b2 e5e4 f3d2 e4e3 f2e3 f6g4 d2f1 e7g5 b2d4 f8e8 b3c5 c6d4 e3d4 g5e3 f1e3 g4e3 d1e1 e3c2 c5b7 c8b7 g2b7 a8b8 b7f3 e8d8 d4d5 c2e1 a1e1 b8b2 c4c5 b2a2');
INSERT INTO evaluations VALUES(46182,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(46182,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(46182,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(46182,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(46182,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(46182,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(46182,36,'cp 25','d5c4 e1g1 b8c6 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 c1b2 c8d7 c3c4 f8c8 c4c5 d7e8 f3d2 b7b5 c5b6 a7b6 d2c4 a8b8 c4d6 e7d6 d1d6 c6b4 a2a3 b4d5 b2e5 e8a4 b3c1 h7h6 c1a2 c8d8 d6d8 b8d8 a1b1 f6d7 e5d4 d8c8 g2d5 e6d5 a2c3 c8c4 c3a4');
INSERT INTO evaluations VALUES(46121,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46121,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(46121,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(46121,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(46121,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(46121,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(46121,36,'cp 27','e1g1 b8c6 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 c1b2 c5e7 c3c4 c8d7 d2b3 f8c8 c4c5 d7e8 f3d2 b7b5 c5b6 a7b6 d2c4 a8b8 c4d6 e7d6 d1d6 c6e7 a1c1 c8c1 b3c1 e8b5 d6d1 b8c8 g2b7 c8b8 b7f3 b8c8 d1d6 b5c6 f3c6 e7c6 c1d3 g8f8 a2a4 f6d5');
INSERT INTO evaluations VALUES(52337,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(52337,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(52337,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(52337,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(52337,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(52337,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(52337,36,'cp 63','c3d4 d7d5 e4e5 f8g7 g1f3 a7a6 f1d3 b8c6 e1g1 f7f6 b1d2 g8h6 e5f6 g7f6 d2b3 h6f7 d3e2 e8g8 h2h3 c8f5 b3c5 b7b6 c5a4 f5e4 c1e3 e7e6 a4c3 e4f3 e2f3 c6e7 f3g4 e7f5 a1c1 f5e3 f2e3');
INSERT INTO evaluations VALUES(52332,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(52332,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(52332,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(52332,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(52332,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(52332,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(52332,36,'cp 67','d7d5 e4e5 a7a6 b1c3 b8c6 h2h3 f8g7 c1e3 g7e5 d4e5 d5d4 e3d4 c6d4 d1d2 d4c6 d2f4 d8d4 f4d4 c6d4 a1d1 d4e6 f1c4 c8d7 c4d5 e8c8 g1e2 h7h5 f2f3 h5h4 d5e6 d7e6 d1d8 c8d8 e1f2 g8h6 e2d4 e6d7 h1d1 e7e6 b2b4 d8e7 c3e4');
INSERT INTO evaluations VALUES(52654,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(52654,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(52654,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(52654,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(52654,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(52654,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(52654,36,'cp 66','e4e5 a7a6 f1d3 b8c6 g1f3 g8h6 h2h3 c8f5 e1g1 f5d3 d1d3 h6f5 b1c3 e7e6 c3e2 f5h4 f3h4 d8h4 d3b3 a8b8 c1e3 f8g7 a1c1 e8g8 c1c3 g7h6 b3d1 h6e3 f2e3 b8c8 d1d2 h4e7 f1f6 c8c7 a2a3 c6a5 c3c7 e7c7');
INSERT INTO evaluations VALUES(52655,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(52655,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(52655,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(52655,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(52655,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(52655,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(52655,36,'cp 59','a7a6 g1f3 b8c6 f1d3 g8h6 h2h3 c8f5 e1g1 f5d3 d1d3 h6f5 b1c3 e7e6 c3e2 f5h4 f3h4 d8h4 c1e3 f8g7 a1c1 h4e7 e2f4 e8g8 d3d2 f7f6 f4d3 a6a5 f1e1 a5a4 c1c2 e7f7 f2f4 c6a5 e3f2 f6e5 d4e5 g7h6 g2g3');
INSERT INTO evaluations VALUES(52204,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(52204,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(52204,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(52204,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(52204,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(52204,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(52204,36,'cp 248','b1c3 d7d5 e4d5 b7b5 c4b3 b5b4 c3e2 c6d5 d2d4 e5e4 e2f4 g8e7 f3d2 b8c6 c2c4 b4c3 b2c3 c6a5 b3d5 e7d5 d1h5 g7g6 h5d5 d8d5 f4d5 e8f7 d2e4 c8b7 d5f6 f8g7 a1b1 g7f6 e4d6 f7f8 b1b5 b7g2 h1g1 a5b7 d6b7 g2c6 c1a3 f8g8 b5b2 a8b8 b7d6 b8b2 a3b2 f6e7 b2a3 e7d6 a3d6 g8f7 e1d2 h8d8 d6f4');
INSERT INTO evaluations VALUES(42557,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(42557,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(42557,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(42557,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(42557,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(42557,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(42557,36,'cp -13','c7c6 b1c3 d7d6 a2a4 c5b4 c1g5 h7h6 g5d2 e8g8 f3g3 g8h7 g1f3 f6h5 g3h4 d8h4 f3h4 a7a5 h2h3 c8e6 c4e6 f7e6 h4f3 b8a6 e1g1 b4c5 c3e2 a6b4 d2b4 a5b4 a4a5 h5f6 f3d2 d6d5 d2b3 c5d6 g1h1 d5e4');
INSERT INTO evaluations VALUES(38199,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(38199,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(38199,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(38199,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38199,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(38199,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38199,36,'cp 8','b1c3 g8f6 c1g5 e8g8 g5f6 d8f6 d1d5 c5b4 d5d4 f6e6 d4e3 b8c6 f1d3 g7g6 e3e6 c8e6 e1g1 b4c3 b2c3 a8c8 f1e1 e6d5 f3g5 c6a5 h2h4 f8d8 a1d1 h7h6 g5e4 g8f8 d3b5 h6h5 f2f3 a7a6 b5d3 d5e4 d3e4 d8d1 e1d1 c8c3 d1d7 f8e8');
INSERT INTO evaluations VALUES(38188,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38188,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(38188,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(38188,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(38188,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(38188,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(38188,36,'cp 18','g8f6 c1g5 e8g8 g5f6 d8f6 d1d5 c5b4 d5d4 f6e6 d4e3 b8c6 f1d3 g7g6 e3e6 c8e6 e1g1 b4c3 b2c3 a8c8 f1e1 c6a5 f3g5 e6d5 h2h4 f8d8 a1d1 a5c4 d3c4 d5c4 d1d8 c8d8 a2a3 d8d2 e1e7 d2c2 e7b7 c2c3 b7c7 c3c1 g1h2');
INSERT INTO evaluations VALUES(40650,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40650,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(40650,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(40650,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(40650,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(40650,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(40650,36,'cp 21','c1g5 e8g8 g5f6 d8f6 d1d5 c5b4 d5d4 f6e6 d4e3 e6e3 f2e3 b8d7 e1c1 b4c3 b2c3 d7b6 d1d4 c8d7 f1d3 d7c6 c1d2 f8e8 f3g5 g7g6 h1f1 e8e7 g2g4 a8e8 d4f4 b6d5 f4e4 h7h6 e4e7 e8e7 g5e4 g8g7 c3c4 d5c7 e4d6 e7d7 c4c5 c7e6 d2c3 e6c5');
INSERT INTO evaluations VALUES(40646,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(40646,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(40646,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40646,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(40646,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(40646,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40646,36,'cp 17','e8g8 g5f6 d8f6 d1d5 c5b4 d5d4 f6e6 d4e3 b8c6 f1d3 g7g6 e3e6 c8e6 e1g1 b4c3 b2c3 a8c8 f1e1 e6d5 f3g5 c6a5 a1d1 f8d8 h2h4 h7h6 g5e4 g8f8 f2f3 d5e4 d3e4 d8d1 e1d1 c8c3 d1d5 a5c4 d5d7 c4b6 d7b7 c3a3 g1f2 a3a2 e4d3');
INSERT INTO evaluations VALUES(21279,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(21279,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(21279,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(21279,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(21279,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(21279,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(21279,36,'cp 58','d2d4 g8f6 d1b3 d8e7 e1g1 g7g5 d4e5 d6e5 b1d2 f8g7 a2a4 e8g8 b3c2 f8d8 c4e2 f6h5 g2g3 a7a5 f3e1 h5f6 e1g2 c8h3 d2c4 d8d7 f2f3 g7f8 g3g4 e7e6 c1e3 b7b6 e3f2 h6h5 c4e3 h5g4 f3g4 f8c5 g1h1');
INSERT INTO evaluations VALUES(21294,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(21294,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(21294,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(21294,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(21294,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(21294,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(21294,36,'cp 44','g8f6 d1b3 d8d7 c4d3 g7g6 d4d5 c6b8 b3d1 f8g7 c3c4 e8g8 b1c3 a7a5 d3c2 b8a6 c1e3 f6h5 d1d2 h5f4 e3f4 e5f4 d2f4 d7e7 e1g1 a6b4 f4d2 b4c2 d2c2 c8g4 f3e1 a8e8 f2f3 g4d7 e1d3 f7f5');
INSERT INTO evaluations VALUES(55103,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(55103,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(55103,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(55103,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(55103,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(55103,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(55103,36,'cp -53','d2d4 c7c5 e2e3 g8f6 c2c4 c5d4 e3d4 c8g4 f1e2 g4e2 d1e2 b8c6 g1f3 d5c4 e2c4 e7e6 e1g1 f8b4 b1c3 e8g8 c1g5 h7h6 g5f6 d8f6 a1d1 b4c3 b2c3 a8c8 c4b5 c6e7 c3c4 b7b6 f1e1 e7g6 h4h5 g6h4 f3h4 f6h4 g2g3 h4g4 d4d5 c8c4');
INSERT INTO evaluations VALUES(55050,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(55050,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(55050,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(55050,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55050,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(55050,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(55050,36,'cp -54','c7c5 e2e3 b8c6 c2c4 c5d4 e3d4 g8f6 b1c3 d5c4 g1f3 c8e6 c1e3 f6d5 a1c1 d8a5 e3d2 d5c3 d2c3 a5f5 c3d2 a8d8 f1c4 e6c4 c1c4 e7e6 e1g1 f8d6 d1b1 f5b1 f1b1 f7f6 g1f1 d8d7 b1c1 e8f7 f1e2 h7h5 d2e3 h8c8 c4c3 d6b4 c3c4 c8d8 a2a3 b4d6');
INSERT INTO evaluations VALUES(6655,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(6655,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(6655,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(6655,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(6655,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(6655,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(6655,36,'cp 65','c2c3 b4c5 c1e3 c5b6 d4f5 d7d5 f5g7 e8f8 g7h5 d8h4 h5g3 d5e4 b1d2 g8f6 d2c4 f6g4 c4b6 a7b6 d1c1 h8g8 f1e2 c6e5 e3f4 g8g6 g3e4 c8f5 f4g3 h4d8 e1g1 f5e4 f2f3 e4d3 c1d2 d3c4 d2d8 a8d8 e2c4 e5c4 f3g4 g6g4 b2b3 c4e3 f1f3 d8e8');
INSERT INTO evaluations VALUES(6654,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(6654,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(6654,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(6654,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(6654,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(6654,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(6654,36,'cp 39','b4c5 c1e3 c5b6 d4f5 d7d5 f5g7 e8f8 g7h5 d8h4 h5g3 d5e4 b1d2 g8f6 d2c4 c8e6 d1a4 f8g7 a4c2 h8g8 c4b6 a7b6 a2a3 a8d8 f1e2 g7f8 e1g1 g8g6 a1d1 d8e8 c2c1 f6d5 d1d5 e6d5 f1d1 d5b3 d1d7 e8e7 d7e7 c6e7 c3c4 h4f6 g3e4 f6e5 e4g3 e7f5 e3f4 f5g3 h2g3');
INSERT INTO evaluations VALUES(6562,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(6562,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(6562,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6562,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(6562,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(6562,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(6562,36,'cp 32','c1e3 c5b6 d4f5 d7d5 f5g7 e8f8 g7h5 d8h4 h5g3 d5e4 b1d2 g8f6 d2c4 c8e6 d1a4 f8g7 e1c1 h8e8 c1b1 g7h8 f1e2 b6e3 c4e3 a7a6 a4c2 h4f4 e2c4 e6c4 e3c4 b7b5 c4e3 a8d8 g3f5 c6e5 d1d8 e8d8 c3c4 f6g4 c4b5 g4e3 f5e3 a6b5 c2c3');
INSERT INTO evaluations VALUES(6560,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(6560,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(6560,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(6560,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(6560,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(6560,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(6560,36,'cp 35','c5b6 d4f5 d7d5 f5g7 e8f8 g7h5 d8h4 h5g3 d5e4 b1d2 g8f6 d2c4 c8e6 d1a4 f8g7 c4b6 a7b6 a4c2 h8g8 a2a3 a8d8 a1d1 d8d1 c2d1 g8d8 d1c1 h7h6 f1e2 c6e5 e1g1 e5g4 h2h3 g4e3 c1e3 h4g5 f2f4 g5c5 e3c5 b6c5 f1d1 d8e8 e2b5 c7c6 b5e2 g7f8 c3c4 e6d7 g1f2 e4e3 f2e1 d7e6');
INSERT INTO evaluations VALUES(6342,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(6342,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(6342,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(6342,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(6342,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(6342,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(6342,36,'cp 62','d4f5 d7d5 f5g7 e8f8 g7h5 d8h4 h5g3 d5e4 d1d2 h4e7 e3h6 g8h6 d2h6 f8e8 f1e2 e4e3 f2f4 e7d6 h6d6 c7d6 e2f3 d6d5 f3d5 c6e7 d5e4 f7f5 e4d3 h8g8 b1a3 e7d5 d3c4 c8e6 h1f1 e8f7 a3c2 a8c8 c4b3 g8g4 g3f5 e6f5 b3d5 f7g7 d5f3 e3e2 e1e2 c8e8 e2d2 g4f4 f3b7 e8d8 c2d4 b6d4 c3d4 f4d4 d2c3');
INSERT INTO evaluations VALUES(50235,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(50235,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(50235,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(50235,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(50235,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(50235,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(50235,36,'cp 153','e4e5 f6h5 g1f3 c7c6 c2c3 h5g7 b1d2 d7d6 f1c4 d6d5 c4e2 h7h5 a2a4 a7a5 f3h4 g6g5 h4f3 g5g4 f3g1 c6c5 d4c5 b8c6 h2h3 c6e5 d2b3 h8g8 h3g4 c8g4 c1f4 e5c6 e2g4 h5g4 d1g4 d8c8 g4f3 c8f5 b3d4 c6d4 c3d4 g7e6 g1e2 f5e4 f3e4 d5e4 e1c1 g8g2');
INSERT INTO evaluations VALUES(50148,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(50148,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(50148,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(50148,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(50148,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(50148,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(50148,36,'cp 132','f6h5 g1f3 c7c6 b1c3 d7d6 f1c4 d6d5 c4d3 h5g7 h2h3 c8f5 e1g1 e7e6 g2g4 f5d3 d1d3 h7h5 c3e2 f8e7 c2c3 a7a5 g1g2 b8d7 b2b3 d8c7 c1g5 c7d8 d3e3 e7g5 f3g5 d8e7 e3f4 h5g4 h3g4 h8h4 f2f3');
INSERT INTO evaluations VALUES(50274,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(50274,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(50274,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(50274,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(50274,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(50274,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(50274,36,'cp 150','g1f3 c7c6 b1c3 d7d6 f1c4 d6d5 c4d3 h5g7 h2h3 c8f5 c3e2 b8d7 d3f5 g7f5 h3h4 f5g7 e2f4 e7e6 h4h5 g7h5 f4h5 g6h5 h1h5 c6c5 e1f1 c5d4 f3d4 d8c7 d1e2 a8c8 c2c3 c7c4 g2g3 c4e2 f1e2 h7h6 c1e3 h8g8 e3h6 f8h6 h5h6 d7e5 d4e6');
INSERT INTO evaluations VALUES(54415,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(54415,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(54415,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(54415,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(54415,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(54415,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(54415,36,'cp 89','b2b3 b7b5 a2a4 d8a5 b3b4 a5a6 a4a5 b8c6 b1a3 a6b7 a3c2 f7f6 f2f4 a7a6 f1e2 g7g6 c1e3 f6f5 g1f3 f8e7 h2h3 h7h5 e3f2 b7d7 f3h4 e8f7 e1g1 g8h6 d1e1 c8b7 h4f3 a8g8 f3g5 f7e8 f2h4 e8d8 e1g3 d8c8 e2f3 e7g5 f4g5');
INSERT INTO evaluations VALUES(54410,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(54410,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(54410,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(54410,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(54410,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(54410,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(54410,36,'cp 106','b7b5 a2a4 d8a5 b3b4 a5a6 a4a5 a6b7 b1a3 b8c6 a3c2 a7a6 g1f3 f7f6 h2h4 h7h5 d1e2 g8e7 g2g3 f6e5 f3e5 c6e5 e2e5 e7c6 e5f4 f8e7 f1g2 e7f6 f4f3 b7f7 c1f4 e8g8 c2e3 c8d7 f4d6 f6e7 f3f7 f8f7 d6c7 a8c8 c7b6 g8h8 e1g1 f7f8');
INSERT INTO evaluations VALUES(51713,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(51713,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(51713,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(51713,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(51713,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(51713,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(51713,36,'cp 116','a2a4 d8a5 b3b4 a5a6 a4a5 b8c6 b1a3 a6b7 a3c2 f7f6 g1f3 a7a6 d1e2 g8e7 h2h4 h7h5 g2g3 f6e5 f3e5 c6e5 e2e5 e7c6 e5f4 f8e7 f1g2 e7f6 f4f3 b7f7 c1g5 f6g5 h4g5 f7f3 g2f3 g7g6 c2e3 e8g8 f3g2 c8d7 f2f4 g8g7 e1g1 a8d8 g2h3 c6e7 f1e1 h5h4 g1g2 e7f5 h3f5');
INSERT INTO evaluations VALUES(54380,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(54380,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(54380,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(54380,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(54380,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(54380,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(54380,36,'cp 58','f1d3 f8d6 e1g1 g8e7 f1e1 b8c6 a2a3 e8g8 b1d2 e7g6 b2b4 a7a6 c1b2 c6e7 f3e5 b7b6 g2g3 e7f5 d1h5 g6e7 h5e2 c8b7 d2f3 b7c8 g3g4 f5h6 f3g5 e7g6 h2h4 f7f5 e5g6 h7g6 c2c4 h6g4 f2f3 g4f6 c4c5 d6f4 g5e6 c8e6 e2e6 g8h7');
INSERT INTO evaluations VALUES(54374,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(54374,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(54374,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(54374,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(54374,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(54374,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(54374,36,'cp 40','f8d6 e1g1 g8e7 f1e1 b8c6 a2a3 e8g8 b1d2 h7h6 b2b4 b7b6 d1e2 c8b7 b4b5 c6a5 f3e5 d6e5 e2e5 e7f5 d2f3 a5c4 e5f4 a7a6 b5a6 b7a6 d3f5 e6f5 f4f5 a6c8 f5h5 a8a7 a3a4 f8e8 e1e8 d8e8 h5d5 a7a4 a1a4 e8a4');
INSERT INTO evaluations VALUES(30614,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(30614,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(30614,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(30614,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(30614,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(30614,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(30614,36,'cp 77','b1c3 g8f6 a2a3 d8e7 g1e2 c7c6 c1g5 h7h6 g5h4 g7g5 h4g3 c6d5 e4d5 f6d5 h2h4 h8g8 c3d5 b7d5 h4g5 h6g5 e2c3 f8g7 d1d3 b8c6 e1c1 g7c3 d3c3 e7c5 c3f6 c5e7 f6h8 g8h8 h1h8 e7f8 h8f8 e8f8 c2c4 d5e4 d1d7 f8e8 d7c7 g5g4 g3d6 a8d8 c4c5 b6c5');
INSERT INTO evaluations VALUES(30612,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(30612,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(30612,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(30612,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(30612,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(30612,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(30612,36,'cp 51','g8f6 a2a3 f8e7 g2g4 e6d5 e4d5 e7c5 f1g2 d8e7 g1e2 f6g4 c3e4 f7f5 e4c5 e7c5 e1g1 e8g8 b2b4 c5c4 f1e1 d7d6 d1d4 c4d4 e2d4 b8d7 c1b2 g4f6 c2c4 f8e8 d4f5 d7e5 b2e5 e8e5 e1e5 d6e5 a1e1 a8e8 f5e3 g8f7 c4c5 e8d8 d5d6 b7g2 d6c7 d8c8 e3c4 c8c7 c4e5 f7e6');
INSERT INTO evaluations VALUES(28685,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(28685,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(28685,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(28685,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(28685,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(28685,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(28685,36,'cp 70','a2a3 f8e7 g2g4 e6d5 e4d5 e7c5 f1g2 d8e7 g1e2 f6g4 c3e4 f7f5 e4c5 e7c5 e1g1 e8g8 c1f4 a7a5 e2d4 b8a6 c2c4 a8e8 a1c1 g7g5 f4g3 f5f4 d4b3 c5e7 d1g4 f4g3 h2g3 d7d6 c1e1 e7f6 f2f4 b7c8 g4g5 f6g5 f4g5 f8f1 g1f1 e8e1 f1e1 a6c5 b3c5 b6c5 e1f2 c8f5 f2e3 g8g7 g2e4 f5d7');
INSERT INTO evaluations VALUES(37459,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37459,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37459,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37459,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(37459,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(37459,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(37459,36,'cp 1','d7d5 c2c4 c7c6 c4d5 c6d5 b1c3 f6e4 f3e5 c8f5 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 e5c6 b7c6 e4c6 a8c8 c6f3 d8d4 d1d4 g7d4 f1d1 d4g7 a1b1 e6a2 b1a1 a2b3 d1d3 b3c4 d3d7 f8d8 d7a7 d8d1 g1g2 c4e2 f3e2 d1c1 a1c1 c8c1 b2b3 g7f6 a7d7 c1c2 d7d8 g8g7 e2c4 c2c1 d8c8 c1c2');
INSERT INTO evaluations VALUES(36780,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(36780,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(36780,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36780,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36780,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36780,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36780,36,'cp 9','c2c4 c7c6 c4d5 c6d5 b1c3 f6e4 f3e5 c8f5 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 a1c1 d8c8 a4a7 f5c2 a2a4 c3c4 e2e3 c4a4 a7a4 c2a4 c1c8 f8c8 g3g4 a4c2 f1a1 h7h5 h2h3 c2d3 g2f1 d3e4 g4g5 g7f8 h3h4');
INSERT INTO evaluations VALUES(36737,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(36737,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36737,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36737,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36737,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36737,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36737,36,'cp 8','c7c6 c4d5 c6d5 f3e5 c8f5 b1c3 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 d8d7 f1e1 a8c8 a1c1 f8e8 e2e3 f5d3 h2h4 h7h6 e3e4 d3e4 g2e4 d5e4 e1e4 e8d8 e4e1 e7e6 c1d1 c6c5 a4d7 d8d7 d4c5 d7d1 e1d1 g7c3 f4h6 c8c5 d1d8 g8h7 h6f8 c5c7 d8e8 c3g7');
INSERT INTO evaluations VALUES(35879,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(35879,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(35879,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(35879,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(35879,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(35879,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(35879,36,'cp 12','c4d5 c6d5 f3e5 f6d7 g2d5 d7e5 d4e5 b8c6 b1c3 c8h3 f1e1 d8b6 c1f4 c6e5 d1b3 b6b3 a2b3 e5c6 f4e3 f8d8 d5c6 b7c6 e3a7 g7d4 a7d4 a8a1 e1a1 d8d4 a1a8 g8g7 a8a6 d4d2 f2f3 d2b2 a6c6 b2b3 c3d5 b3b7 g1f2 h3f5 c6c7 b7c7 d5c7 e7e6 c7e8 g7f8');
INSERT INTO evaluations VALUES(35859,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(35859,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(35859,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(35859,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35859,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(35859,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(35859,36,'cp 23','c6d5 b1c3 f6e4 f3e5 c8f5 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 f1c1 c3c1 a1c1 d8b6 e2e3 b6b2 a4d1 b2a3 g2f1 a7a5 d1d2 h7h5 c1c3 a3b4 h2h4 a5a4 f1d3 f5d3 d2d3 f8a8 d3c2 a4a3 c2b3 g7f8');
INSERT INTO evaluations VALUES(36039,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36039,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36039,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(36039,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(36039,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36039,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(36039,36,'cp 19','b1c3 f6e4 f3e5 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 d8b6 e2e4 b6a6 a4b4 d5e4 b4e7 c8f5 e7a3 a6a3 c1a3 f8e8 a1e1 g7f8 a3f8 g8f8 e1e2 a8b8 f1e1 a7a5 g2e4 f5e4 e2e4 e8e4 e1e4 b8b2 a2a4 b2b3 e4e5 b3c3 e5a5 c3c4 a5a8 f8g7 a4a5 c4d4');
INSERT INTO evaluations VALUES(46194,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46194,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(46194,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(46194,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46194,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(46194,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46194,36,'cp -1','f8d6 f1b5 b8c6 b5c6 b7c6 f4d6 d8d6 g1f3 e8g8 e1g1 c5d4 c3d4 c8a6 f1e1 c6c5 d4c5 d6c5 a2a3 c5b6 b2b4 a6c4 h2h3 f8c8 d1b1 a7a5 e1c1 f6e8 b4a5 b6a5 b1b4 a5b4 a3b4 a8a1 c1a1 f7f6 a1a7 c8c7 a7c7 e8c7 f3d4 g8f7 d2c4 d5c4 b4b5 f7e7');
INSERT INTO evaluations VALUES(40570,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40570,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(40570,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40570,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(40570,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40570,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40570,36,'cp 8','f1b5 b8c6 b5c6 b7c6 f4d6 d8d6 g1f3 e8g8 e1g1 c8a6 f1e1 c5d4 c3d4 c6c5 d4c5 d6c5 a2a3 c5b6 b2b4 a6c4 f3e5 f8c8 a1c1 c4b5 h2h3 h7h6 d1b3 f6d7 e5d7 b5d7 d2f3 c8b8 e3e4 a7a5 f3e5 d7e8 e4d5 e6d5 b3d5 a5b4 a3b4 b6b4 e5c6 e8c6 d5c6 b4a3 e1d1 a3e7');
INSERT INTO evaluations VALUES(57668,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(57668,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(57668,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(57668,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(57668,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57668,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(57668,36,'cp 0','c7c5 f2f4 f8g7 g1f3 g8f6 a2a4 e8g8 f1e2 b8c6 e1g1 d7d5 e4e5 f6g4 h2h3 g4h6 c2c3 d5d4 b1d2 b7b6 d2e4 g8h8 a4a5 a8b8 d1a4 c8d7 a4d1');
INSERT INTO evaluations VALUES(52304,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52304,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52304,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52304,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52304,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52304,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52304,36,'cp 0','f2f4 b8c6 g1f3 f8g7 f1e2 g8f6 c2c3 e8g8 e1g1 b7b5 c1e3 d7d6 a2a3 f6g4 e3c1 g4f6');
INSERT INTO evaluations VALUES(4326,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(4326,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(4326,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(4326,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(4326,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(4326,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(4326,36,'cp 41','d2d4 c5b6 d4e5 f6g4 c1f4 d8e8 b1d2 g4e5 a2a4 d7d6 f3e5 c6e5 a4a5 e5c4 d2c4 b6c5 d1d3 a7a6 b2b4 c5a7 f4e3 c8e6 e3a7 e8b5 c4b2 a8a7 f1e1 a7a8 f2f3 h7h6 d3b5 a6b5 g1f2 f8d8 b2d3 e6c4 d3f4 c7c6 a1d1 c4b3 d1d4 g7g6 g2g3 d8d7 f2e2 b3c4 f4d3');
INSERT INTO evaluations VALUES(4360,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(4360,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(4360,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(4360,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(4360,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(4360,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(4360,36,'cp 40','c5b6 d4e5 f6g4 c1f4 d8e8 a2a4 g4e5 f3e5 c6e5 b1d2 d7d6 a4a5 e5c4 d2c4 b6c5 d1d3 a7a6 b2b4 c5a7 f4e3 c8e6 e3a7 e8b5 c4b2 a8a7 f1e1 a7a8 f2f3 f8d8 d3b5 a6b5 g1f2 d6d5 a1d1 d5e4 e1e4 d8d1 b2d1 a8d8 e4d4 d8d4 c3d4 c7c6 d1c3 e6c4');
INSERT INTO evaluations VALUES(4315,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4315,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4315,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4315,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4315,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4315,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(4315,36,'cp -16','d7d6 d2d3 c6a5 c4b3 c7c6 c3a4 a5b3 a2b3 c5b4 c1d2 b4d2 f3d2 b7b5 a4c3 a7a5 f2f4 e5f4 f1f4 d8b6 g1h1 c8e6 d2f3 b6e3 c3e2 f6h5 f4h4 e3e2 d1e2 h5g3 h1g1 g3e2 g1f2 f7f6 f2e2 d6d5 e4d5 e6d5 e2d2 f8e8');
INSERT INTO evaluations VALUES(17602,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(17602,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(17602,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(17602,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(17602,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(17602,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(17602,36,'cp 90','d2d3 d8c7 d1d2 g8f6 a2a4 a6a5 b1a3 e7e5 e1g1 f6d7 a3c4 f8e7 b2b3 b7b6 c1b2 f7f6 f3h4 c8a6 g2g3 g7g6 f2f4 e5f4 g3f4 e8c8 a1e1 a6c4 b3c4 h8f8 d2e2 e7d6 e2g4 g6g5 f4g5 d6h2 g1g2 f6g5 h4f5');
INSERT INTO evaluations VALUES(17599,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(17599,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(17599,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(17599,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(17599,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(17599,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(17599,36,'cp 70','d8c7 d1d2 a6a5 d2c3 b7b6 a2a4 g8f6 c1d2 g7g6 c3e5 c7d7 d2c3 f8g7 h2h3 f6h5 e5h2 c8a6 g2g4 g7c3 b1c3 h5g7 h2g3 g7e6 e1c1 e8c8 f3e5 d7e8 g4g5 f7f6 g5f6 e7f6 e5g4 e8e7 c1b1 c8b7 f2f4');
INSERT INTO evaluations VALUES(47663,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(47663,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(47663,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(47663,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(47663,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(47663,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(47663,36,'cp 142','g5f6 g7f6 c4d5 e6d5 e2e3 c7c6 f1d3 c8e6 d1b3 d8c7 f2f4 h8g8 b3c2 b8a6 d3a6 b7a6 a1c1 e8c8 g1f3 c8b8 e1g1 c7c8 c3e2 e6d7 e2g3 b8a8 c2d3 f8d6 f3h4 a6a5 h4f5 d6f8 b2b3 f8b4 g1h1');
INSERT INTO evaluations VALUES(47658,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(47658,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(47658,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(47658,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(47658,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(47658,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(47658,36,'cp 126','g7f6 c4d5 e6d5 e2e3 h8g8 g1e2 c7c6 d1c2 c8e6 e2f4 b8d7 f1d3 f8d6 e1g1 f6f5 d3f5 e6f5 c2f5 d8g5 f5g5 h6g5 f4d3 f7f5 b2b4 d7f6 f2f3 a7a6 g2g3 e8c8 a2a4 g8h8 a1e1 d8e8 b4b5 a6b5 a4b5 g5g4 e1e2 c8c7 b5c6 b7c6 f3g4 f6g4');
INSERT INTO evaluations VALUES(44874,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(44874,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(44874,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(44874,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(44874,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(44874,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(44874,36,'cp 88','d5c6 b8c6 g1f3 e7e5 f1b5 f8d6 d2d3 h7h6 h2h3 e8g8 e1g1 a7a6 b5c6 b7c6 f1e1 f8e8 c1d2 a6a5 a2a3 a5a4 c3e4 f6e4 e1e4 c8f5 e4e2 f5g6 d2c3 f7f6 f3d2 g6h5 f2f3 h5f7 g1h1 c6c5 d1e1 d8b6 d2c4 b6c7 c4d6 c7d6 a1d1 d6b6');
INSERT INTO evaluations VALUES(44843,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(44843,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(44843,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(44843,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(44843,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(44843,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(44843,36,'cp 86','b8c6 g1f3 e7e5 f1b5 f8d6 d2d3 h7h6 e1g1 e8g8 c3e4 f6e4 d3e4 d8c7 f3h4 f8d8 c2c3 c6e7 d1h5 c8e6 h4f5 e6f5 e4f5 e7d5 b5a4 e5e4 a4c2 b7b5 c2e4 d5f6 h5f3 f6e4 f3e4 b5b4 c3b4 d6h2 g1h1 h2e5 c1e3 e5b2 a1b1 b2e5');
INSERT INTO evaluations VALUES(18269,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(18269,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(18269,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(18269,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(18269,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(18269,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(18269,36,'cp -30','c6c5 e1g1 g8e7 d2d3 e7g6 c3d5 a7a5 c2c3 f8e7 d3d4 e8g8 c1e3 c8e6 d4e5 d6e5 c3c4 h7h6 f3d2 a5a4 d1c2 e7d6 a1d1 d8c8 f1e1 f8d8 d2f1 c8a6 b2b3 a4b3 a2b3 c7c6 d5c3 a6a5 h2h3 g6f4 e3f4 e5f4 e4e5');
INSERT INTO evaluations VALUES(57018,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57018,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(57018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57018,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57018,36,'cp 0','g8f6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(5981,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5981,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(5981,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(5981,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5981,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5981,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(5981,36,'cp 0','g8f6 d2d3 c8e6 c4b3 a7a5 c1e3 e6b3 a2b3 c5e3 f2e3 d6d5 e4d5 f6d5 d1e2 e8g8 b1c3 d8d7 d3d4 e5d4 e3d4 a8e8 e2d2 b7b6 a1e1 d5f6 c3b5 h7h6 c2c4 e8e1 f1e1 f8e8 e1e8 f6e8 d2e3 c6b4 f3e5 d7f5 e5f3');
INSERT INTO evaluations VALUES(9354,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9354,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(9354,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9354,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9354,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(9354,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9354,36,'cp 17','d2d3 c8e6 b1d2 a7a5 c2c3 e8g8 c4b5 h7h6 f1e1 c5a7 d2f1 c6e7 b5a4 c7c5 f1g3 e7g6 d3d4 e5d4 c3d4 c5d4 f3d4 d6d5 c1e3 e6d7 a4d7 d8d7 e4d5 d7d5 d4f5 a7e3 f5e3 d5d1 e1d1 f8d8 g3e2 g6f8 e2d4 f8e6 d4f5 f6e4 f2f3');
INSERT INTO evaluations VALUES(12452,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12452,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12452,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12452,36,'cp 0','f8e7 d1d3 e8g8 e1c1 h7h6 g5h4 d7d5 e4d5 f6d5 h4e7 d8e7 c3d5 e6d5 d3g3 e7f6 g3f3 f6g5 f3e3 g5f6');
INSERT INTO evaluations VALUES(7924,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7924,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7924,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7924,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7924,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7924,36,'cp 0','d1d3 h7h6 g5h4 d7d5 e1c1 e8g8 e4d5 f6d5 h4e7 d8e7 c3d5 e6d5 d3e3 e7f6 f1b5 c8d7 e3f3 f6g5 f3e3');
INSERT INTO evaluations VALUES(50058,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(50058,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(50058,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(50058,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(50058,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(50058,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(50058,36,'cp 113','g1f3 a7a6 f1d3 d7d6 c1e3 b8d7 c2c3 g7g6 e1g1 f8g7 f1e1 e8g8 b1d2 b7b6 d2f1 c8b7 d1d2 g8h8 f1g3 e6e5 f3g5 d7f6 d3c4 d8e8 a1d1 e5d4 e3d4 e7c6 c4d5 h7h6 g5h3 h8h7 f2f4 c6d4 d5b7 a8b8 f4f5 b8b7 c3d4');
INSERT INTO evaluations VALUES(50057,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(50057,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(50057,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(50057,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(50057,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(50057,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(50057,36,'cp 95','a7a6 f1d3 d7d6 c1e3 b8d7 c2c3 g7g6 b1d2 f8g7 d1e2 h7h6 h2h4 b7b5 e1g1 f7f5 a2a4 e8g8 e3f4 b5a4 f1e1 d7b6 e4e5 e7d5 f4g3 a8b8 g3h2 g8h8 d2c4 b6c4 d3c4');
INSERT INTO evaluations VALUES(57574,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(57574,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(57574,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(57574,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(57574,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(57574,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(57574,36,'cp 36','e7e5 e2e3 g8f6 d2d4 d7d6 f1e2 f8e7 g1h3 c7c6 b2b3 b8a6 c1b2 e8g8 e1g1 a6c7 g1h1 c7e6 f2f3 c8d7 d1d2 e5d4 e3d4 d6d5 f1e1 f8e8 e2d3 g7g6 a1d1 e7d6 d3f1 a8c8 c3e2 f6h5 g2g3 a7a6 b2c3 g6g5 h3f2 f5f4 g3f4');
INSERT INTO evaluations VALUES(47754,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47754,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47754,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47754,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47754,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(47754,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(47754,36,'cp 9','c7c6 b5e2 c6c5 e2c4 f8e7 d2d3 h7h6 h2h3 c8e6 f3h4 e8g8 h4g6 f7g6 c4e6 g8h7 c3d5 f6d5 e6d5 b8c6 c2c3 e7g5 c1g5 d8g5 d1g4 g5f6 e1g1 h6h5 g4e6 a8b8 a2a4 c6e7 d5c4 f6e6 c4e6 f8f6 e6c4 g6g5 g2g3 h5h4 g1g2 e7c6');
INSERT INTO evaluations VALUES(45840,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45840,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(45840,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45840,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(45840,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45840,36,'cp 4','b5e2 b8d7 e1g1 b7b5 d2d4 f8e7 e2d3 e5d4 f3d4 d7e5 c3e2 e8g8 e2g3 f8e8 a2a4 b5b4 f1e1 c8g4 f2f3 g4d7 d3f1 e7f8 f3f4 e5g6 c2c3 h7h6 c3b4 d6d5 b4b5 d5e4 d4c6 d8c7 d1c2 a8c8 c1e3 a7a6 g1h1 a6b5 a4b5');
INSERT INTO evaluations VALUES(55542,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(55542,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(55542,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(55542,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(55542,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55542,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(55542,36,'cp 27','g8f6 g1f3 f8e7 e1g1 e8g8 b1c3 c8e6 h2h3 c7c5 f1e1 b8c6 c1g5 a7a6 a2a4 a8b8 g5f6 e7f6 c3d5 f6g5 c2c3 g5h6 b2b4 c5b4 d5b4 e6c4 b4c6 b7c6 d3c4 d8c7 d1d3 a6a5 a1d1 b8d8 c4c5 d6d5 e4d5 d8d5');
INSERT INTO evaluations VALUES(56915,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(56915,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(56915,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(56915,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(56915,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(56915,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(56915,36,'cp 212','d2d4 e5d4 d1d4 g8f6 c1g5 f8e7 b1c3 h7h6 g5f4 b8c6 d4e3 f6g4 e3d2 d7d6 f1b5 c8d7 e1c1 e7f6 h1e1 a7a6 b5c6 d7c6 c3d5 e8g8 h2h3 g4e5 f4e5 d6e5 d2h6 f6g7 h6e3 c6d5 d1d5 d8e7 e1d1 a8d8 e3d2 d8d5 d2d5 e7b4 f3e5 c7c6 d5c4 b4e7 e5f3 f8e8 d1e1 e7d6 c2c3 b7b5 c4e2 d6f4 e2d2');
INSERT INTO evaluations VALUES(56908,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(56908,31,'cp 228',NULL);
INSERT INTO evaluations VALUES(56908,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(56908,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(56908,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(56908,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(56908,36,'cp 225','e5d4 d1d4 g8f6 c1g5 f8e7 b1c3 h7h6 g5f4 b8c6 d4e3 g6g5 f4g3 f6h5 c3d5 d7d6 e1c1 c8e6 f1b5 e8g8 d5e7 d8e7 e4e5 c6e5 f3e5 d6e5 g3e5 f8d8 b5d3 h5f6 e3g3 f6h5 g3f3 g5g4 f3e3 e7g5 e3g5 h6g5 e5c7 d8d7 c7a5 b7b6 a5c3 h5f4 d3a6 a8d8 d1d7 d8d7');
INSERT INTO evaluations VALUES(45988,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(45988,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(45988,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(45988,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(45988,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(45988,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(45988,36,'cp -21','f8e7 e4e5 d6e5 f3e5 e8g8 e1g1 c8d7 c1f4 b8c6 a2a4 d7e8 e5c6 e8c6 c4b5 d8d7 f1e1 a7a6 b5c6 d7c6 d1e2 b7b5 f4g3 h7h6 h2h3 f8d8 c3e4 f6e4 e2e4 c6e4 e1e4 b5b4 g3e5 d8d7 a1e1 a8d8 e5f4 e7f6 f4e5 f6e5');
INSERT INTO evaluations VALUES(40177,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(40177,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(40177,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(40177,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(40177,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(40177,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(40177,36,'cp -11','e4e5 d6e5 f3e5 e8g8 e1g1 c8d7 c1f4 b8c6 a2a4 d7e8 f1e1 e7d6 e5c6 e8c6 f4d6 d8d6 d1e2 b7b6 e2e5 d6d7 c4b5 f8d8 b5c6 d7c6 e1e3 f6d7 e5e4 c6e4 e3e4 d7f6 e4c4 f6d5 c3d5 d8d5 g1f1 a8d8 f1e2 a7a6 e2f1 g8f8 f1e2 d5e5 e2d1 f8e7 b2b4 c5b4 c4b4');
INSERT INTO evaluations VALUES(23179,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(23179,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(23179,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(23179,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(23179,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(23179,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(23179,36,'cp -16','d7c6 d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 f2f3 e8c8 h2h4 g8e7 d4e2 e7g6 h4h5 g6e5 c1f4 d7e6 b1c3 f8b4 a2a3 b4e7 f4e5 f6e5 a1d1 d8g8 e2c1 g7g6 h5g6 g8g6 e1f1 h7h5 c1d3 e7d6 f1f2 h5h4 c3e2');
INSERT INTO evaluations VALUES(22333,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(22333,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(22333,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(22333,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22333,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(22333,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(22333,36,'cp -22','d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 h2h4 e8c8 f2f3 c6c5 d4e2 d8e8 e1f2 f6f5 b1c3 g8f6 e2g3 f5e4 c3e4 f6e4 g3e4 d7c6 h1e1 h7h6 c1d2 f8e7 h4h5 h8f8 e4g3 e7f6 d2c3 f6c3 b2c3 b7b6 e1e8 c6e8 a1e1 e8f7 g3f5 f7a2 f5g7 a2d5');
INSERT INTO evaluations VALUES(22330,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(22330,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(22330,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(22330,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(22330,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(22330,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(22330,36,'cp -21','e5d4 d1d4 d8d4 f3d4 c8d7 f2f3 g8e7 d4e2 h7h5 h2h4 e7g6 e2f4 g6f4 c1f4 e8c8 b1c3 d7e6 a1d1 f8c5 c3a4 d8d1 e1d1 h8d8 d1c1 c5e7 b2b3 a7a5 f4e3 b7b6 a4b2 c8b7 a2a4 e7a3 c1b1 e6f7 e3f4 a3e7 h1d1 d8d1 b2d1 b6b5 d1e3 b5a4 b3a4');
INSERT INTO evaluations VALUES(29668,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(29668,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(29668,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29668,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(29668,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29668,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(29668,36,'cp 0','e7e6 c2c4 c7c6 d1b3 d8b6 e1g1 b8d7 b1d2 g4f5 f3h4 f5e4 h4f3');
INSERT INTO evaluations VALUES(30178,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(30178,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(30178,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(30178,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(30178,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(30178,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(30178,36,'cp 17','c2c4 d5c4 h2h3 g4h5 e1g1 c7c5 b1d2 b8c6 d2c4 c5d4 e3d4 f8d6 d1b3 a8b8 c1g5 d6c7 a1d1 e8g8 d4d5 e6d5 c4e3 d5d4 e3f5 d8d5 b3d5 f6d5 f3d4 h5e2 d4e2 f7f6 d1d5 f6g5 f1d1 c7b6 f5e3 b8d8 e2c3 d8d5 d1d5 f8d8 d5g5');
INSERT INTO evaluations VALUES(30156,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(30156,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30156,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30156,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30156,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(30156,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(30156,36,'cp 7','d5c4 e1g1 c7c5 d4c5 b8c6 b1a3 f8c5 a3c4 e8g8 a2a3 a7a5 h2h3 g4h5 b2b3 d8e7 c1b2 h5g6 f3e5 c6e5 c4e5 g6f5 e2f3 f8d8 d1e2 h7h6 f1d1 d8d1 e2d1 a8d8 d1e1 a5a4 b3a4 f6e4 a1d1 d8d1 e1d1 c5a3 b2a3 e7a3 d1d4 a3c3 d4c3 e4c3 f3b7');
INSERT INTO evaluations VALUES(43408,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(43408,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43408,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(43408,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43408,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43408,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(43408,36,'cp 38','e7e5 d4b3 c8e6 f2f4 b8c6 f4f5 e6b3 a2b3 f8e7 f1c4 a8c8 e1g1 e8g8 g1h1 b7b5 c4e2 c6b4 e2f3 f8e8 c1d2 h7h6 d2e1 d6d5 e4d5 e5e4 f3e4 f6e4 c3e4 b4c2 d5d6 c2a1 d6d7 e7f8 d7e8r d8e8 e4c3 b5b4 d1a1 b4c3 e1c3 e8b5 f5f6 g7g6 f1e1 c8b8 a1a4 b5a4 b3a4 f8b4 e1e4 b4c3 b2c3');
INSERT INTO evaluations VALUES(43452,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43452,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43452,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43452,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(43452,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(43452,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(43452,36,'cp 16','d4b3 c8e6 f2f4 b7b5 c1e3 b5b4 c3d5 b8d7 f1c4 f8e7 f4e5 f6d5 e4d5 e7h4 e1e2 d7e5 d5e6 e5c4 d1d5 c4e3 e6f7 e8f8 e2e3 d8g5 d5g5 h4g5 e3d3 f8f7 b3d4 g5f6 a2a3 a8b8 h1f1 h8c8 f1f2 b4a3 b2b3 a6a5 a1a3 c8c5 f2f5 d6d5 a3a4 b8c8 d3d2 f7g6 f5f3');
INSERT INTO evaluations VALUES(43461,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(43461,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(43461,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(43461,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43461,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43461,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(43461,36,'cp 8','c8e6 c1e3 b8d7 g2g4 h7h6 d1d2 b7b5 f1g2 e6c4 c3d5 a6a5 a2a4 f6d5 e4d5 c4b3 c2b3 b5a4 a1a4 f8e7 e1g1 e8g8 f1a1 a8b8 a4a5 e7g5 e3g5 d8g5 d2g5 h6g5 a5a6 b8b3 a6d6 d7f6 d6a6 b3b2 d5d6 b2d2 a6a5 e5e4 a5g5 d2d6 g5e5 d6d5 e5e7 f8d8 g2e4 f6e4 e7e4 d5d4 e4d4 d8d4');
INSERT INTO evaluations VALUES(41681,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(41681,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(41681,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(41681,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(41681,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(41681,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(41681,36,'cp -20','c7c6 f1d3 e8g8 h2h3 f8e8 c3e2 b8d7 e1g1 d8c7 d1c1 f6e4 g5f4 d7f8 f4d6 c7d6 f1e1 c8f5 c1f4 d6f4 e2f4 e4d6 d3f5 d6f5 e1e8 a8e8 c2c3 f5d6 a1e1 e8e1 f3e1 g7g5 f4d3 g8g7 f2f3 a7a5 g1f2 f8g6 a2a4 f7f5 d3c5 f5f4');
INSERT INTO evaluations VALUES(40511,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(40511,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(40511,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(40511,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(40511,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(40511,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(40511,36,'cp -20','f1d3 e8g8 h2h3 f8e8 c3e2 h7h6 g5h4 d8a5 c2c3 f6e4 e1g1 b8d7 f3d2 d7f6 d2e4 d5e4 d3c2 f6d5 a2a3 e4e3 h4g3 e3f2 f1f2 d5e3 d1d3 e3c2 d3c2 d6g3 e2g3 c8e6 g3f5 a8d8 a1e1 a5b6 c2d3 e6f5 e1e8 d8e8 d3f5 e8e1 g1h2 b6c7 f5f4 c7f4 f2f4 e1e8 h2g1');
INSERT INTO evaluations VALUES(40512,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(40512,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(40512,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(40512,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(40512,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(40512,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(40512,36,'cp -10','e8g8 c3e2 f8e8 h2h3 h7h6 g5h4 d8a5 d1d2 a5d2 f3d2 f6e4 d3e4 d5e4 h4g3 d6f8 d2c4 b8d7 c4e3 d7f6 c2c4 c8e6 b2b3 f8b4 e1d1 b7b5 d1c2 a7a5 a2a3 b4e7 e2f4 e6d7 a3a4 b5a4 b3a4 e7b4 f4e2 f6h5 g3h2 f7f5 c4c5 g7g5 h2e5 e8e5 d4e5 b4c5');
INSERT INTO evaluations VALUES(36115,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(36115,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(36115,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(36115,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(36115,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(36115,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(36115,36,'cp -4','h2h3 f8e8 c3e2 b8d7 e1g1 d8c7 d1c1 f6e4 g5f4 d7f8 f4d6 c7d6 f1e1 c8f5 f3e5 f8g6 e5g6 d6g6 e2f4 g6g5 c2c3 a7a5 d3e4 f5e4 e1e2 e4f5 e2e3 h7h6 c1d2 e8e3 d2e3 f5d7 f4d3 g5e3 f2e3 a8e8 d3c5 d7c8 a1e1');
INSERT INTO evaluations VALUES(9581,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(9581,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9581,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9581,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(9581,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(9581,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(9581,36,'cp 21','e8g8 d1c2 h7h6 g5h4 c7c6 a2a3 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3e1 d7f6 f3e5 f6d7 e5d7 e7d7 a3a4 d7c7 a4a5 b6a5 d3f1 c7b6 e1a5 b6b2 a5a7');
INSERT INTO evaluations VALUES(3987,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3987,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(3987,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(3987,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(3987,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(3987,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3987,36,'cp 21','d1c2 h7h6 g5h4 c7c6 a2a3 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3e1 d7f6 f3e5 b7c6 a1c1 e7b7 h2h3 c6e8 f2f3 f6d7 e5g4 c8c1 e1c1 b6b5 g4f2 d7b6');
INSERT INTO evaluations VALUES(3988,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(3988,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(3988,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(3988,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(3988,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(3988,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3988,36,'cp 18','h7h6 g5h4 c7c5 d4c5 d7c5 c4d5 f6d5 h4e7 d5e7 f1e2 d8b6 a2a3 a7a5 e1g1 f8d8 f3e5 c8d7 e5d7 d8d7 f1d1 d7d1 a1d1 a8d8 d1d8 b6d8 h2h3 b7b6 g2g3 e7f5 e2f1 d8c7 f1g2 c7d6 c3e4 d6d3 c2d3 c5d3 g3g4');
INSERT INTO evaluations VALUES(14266,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14266,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(14266,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(14266,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(14266,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(14266,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(14266,36,'cp 0','c1g5 h7h6 g5h4 e7e6 d1d2 b7b5 c3b5 a8b8 e2e3 a7a6 b5c3 b8b2 a2a3 f8e7 f1d3 e8g8 g1e2 b2b8 e1g1 e6e5 c3a4 d8e8 a4c3');
INSERT INTO evaluations VALUES(39570,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(39570,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(39570,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(39570,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(39570,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(39570,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(39570,36,'cp 13','c7c5 b1c3 b8c6 g1f3 d7d6 a2a3 g8f6 e1g1 e8g8 d2d3 d8d7 a1b1 b7b6 b2b4 c8b7 e2e3 a8d8 d1c2 e7e6 f1d1 h7h6 h2h3 d7c7 c2b3 c6e7 c1b2 g6g5 d3d4 c5d4 f3d4 b7g2 g1g2 d6d5 c4d5 f6d5 c3d5');
INSERT INTO evaluations VALUES(37788,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(37788,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37788,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(37788,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(37788,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(37788,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(37788,36,'cp 12','g1f3 b8c6 b1c3 d7d6 e1g1 c8g4 h2h3 g4f3 g2f3 e7e6 d2d3 g8e7 f3g2 e8g8 a1b1 d8d7 a2a3 a8c8 d1a4 f8d8 f1d1 d6d5 c1g5 b7b6 c4d5 e6d5 e2e4 h7h6 g5e7 c6e7 a4d7 d8d7 c3d5 e7d5 e4d5 c8d8 b2b4 c5b4 b1b4');
INSERT INTO evaluations VALUES(37787,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(37787,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37787,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(37787,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37787,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37787,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(37787,36,'cp 15','b8c6 b1c3 e7e6 d2d4 c5d4 c3b5 d7d5 c4d5 e6d5 e1g1 g8f6 f3d4 e8g8 c1f4 a7a6 b5c3 f8e8 f4g5 h7h6 g5e3 h6h5 e3g5 c6d4 d1d4 c8e6 a1d1 f6e4 g5d8 g7d4 c3e4 a8d8 d1d4 d5e4 d4d8');
INSERT INTO evaluations VALUES(44046,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(44046,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(44046,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(44046,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(44046,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(44046,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(44046,36,'cp 118','e2e4 d7d5 c4d5 c6d5 e4e5 f6e4 f1d3 e4c3 b2c3 d8c7 c1d2 h7h6 g1h3 b8c6 e1g1 c6a5 h3f4 g7g6 d1f3 f8e7 f1b1 c8d7 f4e2 d7c6 a2a4 a5c4 d2c1 c7d7 e2g3 a8c8 f3d1 h6h5 h2h3 c8c7 g3f1 b7b6 a1a2 c4a5 d1c2 a5c4 f1e3');
INSERT INTO evaluations VALUES(44052,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(44052,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(44052,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(44052,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(44052,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(44052,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(44052,36,'cp 118','d7d5 c4d5 c6d5 e4e5 f6e4 f1d3 e4c3 b2c3 d8c7 c1d2 b8c6 g1h3 h7h6 e1g1 c6a5 h3f4 g7g6 d1f3 f8e7 f1b1 c8d7 f4e2 d7c6 a2a4 c7d7 d2c1 a5c4 e2g3 h6h5 g3f1 a8c8 f3d1 b7b6 h2h3 c4a5 f1d2 d7c7 d2b3 a5c4 a1a2 c6d7');
INSERT INTO evaluations VALUES(45659,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(45659,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(45659,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(45659,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(45659,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(45659,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(45659,36,'cp 124','c4d5 c6d5 e4e5 f6e4 f1d3 e4c3 b2c3 d8c7 c1d2 b8c6 g1h3 c6a5 e1g1 h7h6 h3f4 g7g6 d1f3 f8e7 f1b1 c8d7 f4e2 d7c6 a2a4 a5c4 d2c1 c7a5 b1b3 a5c7 e2g3 h6h5 g3f1 a8c8 b3b1 b7b6 h2h3 c7d7 f1e3 c6b7 f3g3 c4a5 c1d2 a5c4');
INSERT INTO evaluations VALUES(9771,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9771,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9771,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9771,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9771,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9771,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9771,36,'cp 0','d7d6 c2c3 a7a6 a2a4 e8g8 c4b3 f8e8 b1d2 c5a7 f1e1 c8e6 a4a5 e6b3 d1b3 d8d7 d2f1 d6d5 c1d2 a8b8 b3c2 b8d8 c2b3');
INSERT INTO evaluations VALUES(10717,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(10717,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(10717,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(10717,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(10717,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(10717,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(10717,36,'cp 228','d2d4 d8e7 f1e1 a7a6 b5f1 e8g8 c1g5 b7b5 a2a4 h7h6 g5h4 c8b7 c2c3 f8e8 h4g3 f6h7 a4b5 a6b5 a1a8 e8a8 f1b5 a8a1 d4d5 c6d8 d1d3 c7c6 b5c4 h7f6 b1d2 a1e1 f3e1 d6c5 b2b4 c6d5 c4d5 b7d5 e4d5 e5e4');
INSERT INTO evaluations VALUES(10708,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(10708,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(10708,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(10708,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(10708,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(10708,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(10708,36,'cp 226','d8e7 f1e1 e8g8 c2c3 f6e8 b1d2 f7f6 d2f1 g8h8 f1e3 b7b6 b5c4 g7g6 c4d5 c8b7 e3g4 h7h5 c1h6 h5g4 h6f8 e7f8 f3h4 f8h6 d1g4 f6f5 h4f5 h6h5 g4h5 g6h5 a1d1 e8f6 f5d6 c7d6 d5b3 a8f8 d4e5 c6e5 d1d6');
INSERT INTO evaluations VALUES(23106,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(23106,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(23106,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(23106,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(23106,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(23106,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(23106,36,'cp 93','d4f3 d1f3 d8f6 f3c3 c7c6 d2d3 g8h6 c4b3 a7a5 a2a3 f8d6 f2f4 e8g8 h2h3 a5a4 b3a2 d6b8 f4f5 b8a7 g1h1 f6h4 c3e1 h4e1 f1e1 g8h8 c1d2 a7c5 b1c3 d7d6 e1e2 h6g8 g2g4 c8d7 e2g2 b7b5 a1f1 g8f6 h3h4 h7h5 g4g5 f6g4 g2g3');
INSERT INTO evaluations VALUES(23100,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(23100,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(23100,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(23100,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(23100,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(23100,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(23100,36,'cp 104','d1f3 d8f6 f3c3 c7c6 d2d3 g8h6 c4b3 a7a5 a2a3 f8d6 f2f4 e8g8 h2h3 a5a4 b3a2 d6c7 f4f5 f6h4 c3e1 c7b6 g1h1 h4e1 f1e1 g8h8 g2g4 f7f6 b1c3 d7d6 c1d2 c8d7 h1g2 h6g8 a1b1 b6c5 c3e2 d7e8 c2c3 e8f7 a2f7 f8f7');
INSERT INTO evaluations VALUES(23099,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(23099,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(23099,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(23099,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(23099,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(23099,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(23099,36,'cp 86','d8f6 f3c3 c7c6 d2d3 g8h6 c4b3 a7a5 a2a3 f8d6 f2f4 e8g8 h2h3 a5a4 b3a2 d6b8 f4f5 b8a7 g1h1 f6h4 c3e1 h4e1 f1e1 g8h8 c1d2 a7c5 g2g4 b7b5 b1c3 d7d6 e1f1 h6g8 c3e2 g8f6 g4g5 f6h5 h1h2 c8d7 c2c3 f7f6 h3h4 g7g6 d3d4 c5b6');
INSERT INTO evaluations VALUES(1957,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(1957,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(1957,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(1957,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(1957,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(1957,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(1957,36,'cp 100','f3c3 c7c6 d2d3 g8h6 c4b3 a7a5 a2a3 f8d6 f2f4 e8g8 h2h3 d6b8 f4f5 b8a7 g1h1 a5a4 b3a2 f6h4 c3e1 h4e1 f1e1 g8h8 c1d2 a7c5 b1c3 d7d6 g2g4 b7b5 e1e2 c8d7 e2g2 f7f6 h3h4 d7e8 a1f1 h6g8 a2e6 e8f7 e6d7 f7e8 d7e8 f8e8 h1h2 e8e7 d2c1 e7b7 g4g5 g8e7 c3e2 d6d5');
INSERT INTO evaluations VALUES(33725,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(33725,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(33725,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(33725,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(33725,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(33725,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(33725,36,'cp 35','b8c6 h2h3 c6e5 c4b5 c7c6 f3e5 c6b5 d2d4 g7g6 e1g1 f8g7 c1d2 e8g8 f1e1 b5b4 c3e2 f6e4 d2f4 c8e6 a2a3 f8d8 a3b4 a5b4 e5d3 b4b6 f4e3 b6c7 e2f4 e6c4 d1f3 e4d6 d3e5 c7c8 g2g4');
INSERT INTO evaluations VALUES(884,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(884,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(884,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(884,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(884,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(884,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(884,36,'cp 54','e1g1 c8g4 h2h3 g4f3 d1f3 e7e6 c4b5 a5b6 d2d3 e8c8 b5c6 b6c6 f3c6 b7c6 f1e1 c6c5 c3e4 f8e7 e4g5 h8f8 g1f1 d8d6 a2a3 h7h6 g5f3 f6d7 a1b1 e7f6 c1d2 d6a6 e1e4 g7g5 a3a4 f6e7 b2b3 d7f6 e4e1 f6d5 f3e5 e7d6 e5c4 f8d8 e1e4');
INSERT INTO evaluations VALUES(45625,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(45625,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(45625,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(45625,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(45625,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(45625,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(45625,36,'cp -32','d5c4 a2a4 f8b4 g2g3 e8g8 f1g2 b8d7 e1g1 a7a5 d1c2 f6d5 f3g5 d7f6 f4e5 f8e8 f1d1 c8d7 c3a2 h7h6 e5f6 d5f6 g5e4 f6e4 g2e4 b4f8 c2c4 b7b5 c4c2 a8c8 e4g2 d8c7 e2e3 b5a4 a2c3 a4a3 b2a3 e6e5 c3e2 e8d8 d4e5 c7e5');
INSERT INTO evaluations VALUES(45709,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(45709,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(45709,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(45709,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(45709,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(45709,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(45709,36,'cp -16','a2a4 f8b4 g2g3 b7b5 f1g2 c8b7 e1g1 d8b6 d1c2 e8g8 c3e4 f6e4 c2e4 c6c5 d4c5 b4c5 e4c2 b8d7 a4b5 b6b5 f3g5 d7f6 g2b7 b5b7 f4e5 b7b3 a1c1 b3c2 c1c2 f8c8 e5f6 g7f6 g5f3 c5f8 f1c1 c8b8 c1a1 b8b4 f3d4 a8b8 a1a7 f8c5 a7d7 b4b2 c2c4 c5d4 d7d4 b2e2 c4b4 b8b4 d4b4');
INSERT INTO evaluations VALUES(53717,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(53717,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53717,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53717,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(53717,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53717,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(53717,36,'cp 17','f8d6 f1d3 g8e7 c3e2 c8f5 h2h4 b8d7 d3f5 e7f5 d1d3 f5e7 c1f4 d7f6 e1g1 e8g8 a1e1 f6e4 f4d6 e4d6 e2f4 e7g6 f4g6 h7g6 c2c3 f8e8 e1e8 d8e8 f1e1 e8d7 f3e5 d7e6 d3f3 a8e8 e5d3 e6d7 e1e8 d7e8 g2g3 e8d7 g1g2 d6e4 f3e3 e4d6');
INSERT INTO evaluations VALUES(38084,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(38084,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38084,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38084,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38084,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(38084,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38084,36,'cp 14','f1d3 g8e7 c3e2 c8f5 e1g1 e8g8 f1e1 b8d7 c1f4 f5d3 d1d3 d6c7 f4c7 d8c7 e2f4 a8e8 g2g3 e7g6 f4g6 h7g6 c2c3 d7f6 a2a4 e8e1 a1e1 f8e8 e1e8 f6e8 d3e2 c7d7 f3e5 d7e6 e2e3 e8d6 e3f4 e6f5 f4f5 g6f5 f2f3 f7f6 e5d3 g8f7');
INSERT INTO evaluations VALUES(6461,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(6461,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(6461,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(6461,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(6461,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(6461,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(6461,36,'cp 53','a2a3 d7d6 e2e3 a7a6 b2b4 c5a7 c1b2 g8e7 f1e2 e8g8 e1g1 e7g6 f1e1 f8e8 d2d3 c8e6 d1c2 f7f5 a1d1 e6f7 g1h1 c6b8 e3e4 f5f4 d3d4 e5d4 f3d4 a7d4 d1d4 b8c6 d4d1 c6e5 c4c5 f4f3 e2f3 e5f3 g2f3');
INSERT INTO evaluations VALUES(21977,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(21977,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(21977,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(21977,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(21977,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(21977,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21977,36,'cp 68','d2d4 e5d4 f3d4 g8f6 f1b5 c8d7 b1c3 f8e7 d4c6 d7c6 b5c6 b7c6 e1g1 f6d7 f2f4 e7f6 c1d2 e8g8 d1f3 d8e7 a1e1 a7a5 b2b3 f6d4 g1h2 d7c5 c3e2 e7e4 f3e4 c5e4 e2d4 e4d2 f1f2 c6c5 f2d2 c5d4 d2d4 f8e8 e1e8 a8e8 d4c4');
INSERT INTO evaluations VALUES(21949,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(21949,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(21949,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(21949,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(21949,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(21949,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(21949,36,'cp 53','e5d4 f3d4 g8f6 f1b5 c8d7 b1c3 f8e7 d4c6 d7c6 b5c6 b7c6 c1e3 d8b8 a1b1 e8g8 e1g1 b8b4 d1d3 f6d7 a2a3 b4b7 b2b3 e7f6 f2f3 f8e8 f1d1 d7f8 c3e2 a7a5 e3f2 f8e6 a3a4 c6c5 e2c3 f6d4 c3b5 d4f2 g1f2 b7b6 f2f1 e6f4');
INSERT INTO evaluations VALUES(22039,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(22039,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(22039,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(22039,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(22039,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(22039,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(22039,36,'cp 52','f3d4 g7g6 b1c3 f8g7 c1e3 c6d4 e3d4 g8f6 g2g4 e8g8 f1g2 f8e8 e1g1 c8d7 g4g5 f6h5 d4g7 h5g7 f2f4 g7h5 d1d2 h7h6 h3h4 c7c6 g2f3 d6d5 e4d5 h6g5 f3h5 g5g4 d5c6 b7c6 a1d1 d7f5 d2d8 a8d8 h5g4 d8d1 c3d1 f5g4');
INSERT INTO evaluations VALUES(33281,30,'cp 349',NULL);
INSERT INTO evaluations VALUES(33281,31,'cp 339',NULL);
INSERT INTO evaluations VALUES(33281,32,'cp 343',NULL);
INSERT INTO evaluations VALUES(33281,33,'cp 343',NULL);
INSERT INTO evaluations VALUES(33281,34,'cp 370',NULL);
INSERT INTO evaluations VALUES(33281,35,'cp 382',NULL);
INSERT INTO evaluations VALUES(33281,36,'cp 388','d2d4 d7d6 e5c4 d6d5 e4d5 c5d4 d1e2 e8f8 g2g3 h4e7 c1e3 d4e3 c4e3 b8d7 b1c3 h7h5 e1c1 g8f6 e2d2 d7b6 d5d6 e7d6 d2d6 c7d6 d1d6 f6e8 d6d4 h5h4 g3g4 h8h6 f1g2 h4h3 g2e4 h6d6 h1d1 d6d4 d1d4 g7g6 g4g5 b6d7 b2b3 d7e5');
INSERT INTO evaluations VALUES(33276,30,'cp 353',NULL);
INSERT INTO evaluations VALUES(33276,31,'cp 366',NULL);
INSERT INTO evaluations VALUES(33276,32,'cp 376',NULL);
INSERT INTO evaluations VALUES(33276,33,'cp 343',NULL);
INSERT INTO evaluations VALUES(33276,34,'cp 368',NULL);
INSERT INTO evaluations VALUES(33276,35,'cp 364',NULL);
INSERT INTO evaluations VALUES(33276,36,'cp 383','d7d6 e5c4 d6d5 e4d5 c5d4 d1e2 e8f8 g2g3 h4e7 c1e3 d4e3 c4e3 b8d7 b1c3 g8f6 e1c1 h7h5 e2d2 d7b6 d5d6 e7d6 d2d6 c7d6 d1d6 h5h4 f1g2 f6e8 d6d4 h4g3 h2g3 h8h1 g2h1 a8b8 c3b5 a7a6 b5d6 e8d6 d4d6 b6d7 h1g2 f8e7 e3f5 e7f8 d6d4 b7b5 d4h4 d7f6 h4h8 f6g8 f5e7 f8e7 h8g8');
INSERT INTO evaluations VALUES(53251,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(53251,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53251,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(53251,36,'cp 0','c5d4 c2c3 d4c3 b1c3 e7e6 e2e4 g8f6 f1b5 b8c6 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 a1c1 a7a6 b5c6 b7c6 c1c6 f8b4 e1d1 c8d7 c6b6 b4a5 b6d6 a5b4');
INSERT INTO evaluations VALUES(53394,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53394,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53394,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53394,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53394,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53394,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53394,36,'cp 0','c2c3 d4c3 b1c3 g8f6 e2e4 d5e4 d1d8 e8d8 e1c1 c8d7 f3e5 d8e8 c3b5 b8a6 c1b1 e7e6 e5d7 f6d7 b5d6 f8d6 f4d6 a6b8 d1c1 b8c6 f1a6 c6a5 a6e2 a5c6');
INSERT INTO evaluations VALUES(39343,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(39343,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(39343,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(39343,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(39343,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(39343,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(39343,36,'cp 165','c2c3 b4e7 f3e5 g8f6 f1c4 e8g8 d2d3 d7d5 e4d5 b7b5 c4b3 e7d6 f2f4 b8d7 e1g1 d7e5 f4e5 d6e5 d3d4 e5d6 c1g5 d8e8 f1e1 f6e4 g5h4 f7f5 b1d2 a7a5 b3c2 e8g6 d2e4 f5e4 c2e4 c8f5 d1d3 g6h5 e4f5 f8f5 e1e4 f5f4 e4f4 d6f4 h4g3 f4g3 d3g3 h5d5 g3c7 b5b4 c7f4 a8e8 f4g3 b4c3');
INSERT INTO evaluations VALUES(39339,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(39339,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(39339,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(39339,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(39339,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(39339,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(39339,36,'cp 157','b4e7 f3e5 g8f6 f1c4 e8g8 d2d3 d7d5 e4d5 b7b5 c4b3 e7d6 f2f4 b8d7 e1g1 d7e5 f4e5 d6e5 d3d4 e5d6 c1g5 d8e8 f1e1 f6e4 c3c4 f7f6 b1d2 b5c4 d2e4 e8f7 b3c4 f6g5 e1f1 c8f5 e4g5 f7g6 d1d2 g6h6 g2g3 f5h3 f1f8 a8f8 a1e1 g8h8 c4f1 h3f5 f1g2 a7a5 g5f3 h6d2 f3d2 f5g4');
INSERT INTO evaluations VALUES(46978,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(46978,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(46978,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(46978,33,'cp -120',NULL);
INSERT INTO evaluations VALUES(46978,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(46978,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(46978,36,'cp -129','f6e4 d2d4 g7g6 e1g1 e7e6 f1e1 e4f6 c4e2 f8g7 c2c4 b7b6 c1f4 e8g8 b1c3 h7h6 h2h3 c8b7 d1d2 g8h7 a1d1 b8d7 d2e3 f6g8 g2g4 g8e7 e2f1 a7a6 f1g2 a8c8 f4g3 h7g8 g3f4 g6g5 f4g3 e7g6 d4d5');
INSERT INTO evaluations VALUES(47075,30,'cp -135',NULL);
INSERT INTO evaluations VALUES(47075,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(47075,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(47075,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(47075,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(47075,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(47075,36,'cp -125','d2d4 g7g6 e1g1 f8g7 f1e1 e4f6 c4e2 e8g8 c2c4 e7e6 b1c3 b7b6 c1f4 b8c6 d1d2 c6e7 f4g5 c8b7 a1d1 d6d5 f3e5 f6e4 c3e4 d5e4 e5g4 f7f6 g5h6 e7f5 h6g7 g8g7 g4e3 d8e7 e3f5 g6f5 f2f3 a8d8 d2e3 e7d6 f3e4 b7e4 b2b4 g7h8');
INSERT INTO evaluations VALUES(6246,30,'cp 301',NULL);
INSERT INTO evaluations VALUES(6246,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(6246,32,'cp 299',NULL);
INSERT INTO evaluations VALUES(6246,33,'cp 295',NULL);
INSERT INTO evaluations VALUES(6246,34,'cp 306',NULL);
INSERT INTO evaluations VALUES(6246,35,'cp 296',NULL);
INSERT INTO evaluations VALUES(6246,36,'cp 317','c4f7 e8e7 d4c6 d7c6 d1d8 e7d8 f2f3 g8f6 b1c3 a7a5 a2a4 h8f8 f7b3 d8e7 c3e2 f6h5 c1d2 b7b6 b3c4 g7g5 e1c1 c8e6 c4e6 e7e6 g2g4 h5f4 d2f4 g5f4 e2d4 c5d4 d1d4 h6h5 g4g5 c6c5 d4d1 c7c6 h2h4 a8d8 d1d8 f8d8 h1d1 d8b8 c2c3 b8f8 c1c2 c5c4 d1d4 b6b5 a4b5 c6b5 d4d5');
INSERT INTO evaluations VALUES(6148,30,'cp 301',NULL);
INSERT INTO evaluations VALUES(6148,31,'cp 301',NULL);
INSERT INTO evaluations VALUES(6148,32,'cp 300',NULL);
INSERT INTO evaluations VALUES(6148,33,'cp 292',NULL);
INSERT INTO evaluations VALUES(6148,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(6148,35,'cp 309',NULL);
INSERT INTO evaluations VALUES(6148,36,'cp 307','e8e7 d4c6 d7c6 d1d8 e7d8 f2f3 g8f6 b1c3 a7a5 f7b3 f6d7 c1d2 d8e7 e1c1 h8d8 a2a3 b7b5 c3e2 d7f8 e2f4 f8e6 b3e6 c8e6 f4d3 c5d4 d2f4 d4b6 d3e5 d8d1 h1d1 c6c5 f4e3 e7e8 f3f4 a8d8 d1d8 e8d8 e5d3 e6c4 c1d2 a5a4 d3c5 d8e7 d2c3 e7f7 e3d4 b6c5 d4c5 f7g8 f4f5');
INSERT INTO evaluations VALUES(24469,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24469,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(24469,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24469,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(24469,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24469,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(24469,36,'cp 0','c6d4 b5d7 d8d7 d1d4 e7e6 f1d1 f8e7 a2a4 d7c6 c1g5 e8g8 g5f6 e7f6 d4d6 c6c4 d6g3 c4c5 d1d7 f6e5 g3e3 c5b4 c3d1 a8d8 d7d8 f8d8 e3e2 d8d2 e2b5 b4d6 d1e3 e5f4 b5b7 f4e3 f2e3 d2d1 a1d1 d6d1 g1f2 d1d2 f2f3 d2d1');
INSERT INTO evaluations VALUES(51809,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(51809,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(51809,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(51809,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(51809,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(51809,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(51809,36,'cp 66','d2d4 d7d5 a2a3 g8f6 e4e5 f6d7 c3e2 c6c5 c2c3 b8c6 g1f3 f8e7 c1e3 c5d4 c3d4 b7b5 h2h4 b5b4 a3a4 d8b6 e2f4 b4b3 f1d3 c8a6 e1g1 a6d3 d1d3 c6a5 e3d2 a5c4 a4a5 b6b5 d2c3 e8g8 h4h5 f8c8 h5h6');
INSERT INTO evaluations VALUES(44539,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44539,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44539,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44539,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44539,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44539,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44539,36,'cp 0','c5d4 e3d4 d7d5 c4d5 f6d5 g1f3 b8c6 b1c3 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 c3d5 d7d5 b5e2 f7f6 e1g1 e7f7 f1d1 d5f5 c1e3 f8e7 d1d7 h8d8 a1d1 a7a5 g1g2 d8d7 d1d7 f5g6 g2f1 g6b1 f1g2');
INSERT INTO evaluations VALUES(44610,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44610,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44610,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44610,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44610,36,'cp 0','e3d4 d7d5 c4d5 f6d5 g1f3 b8c6 b1c3 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 c3d5 d7d5 b5e2 f7f6 e1g1 e7f7 f1d1 d5f5 c1e3 f8e7 d1d7 h8c8 g1g2 a7a5 e2d1 c8d8 a1c1 d8d7 d1d7 a8d8 d7a7 f5g6 g2f1 g6f5');
INSERT INTO evaluations VALUES(44607,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(44607,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(44607,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44607,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44607,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44607,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44607,36,'cp 0','d7d5 c4d5 f6d5 b1c3 b8c6 g1f3 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 c3d5 d7d5 b5e2 f7f6 e1g1 e7f7 f1d1 d5f5 d1d7 f8e7 c1e3 h8c8 e3a7 f5g5 g1h1 c8c1 a1c1 g5c1 h1g2 c1g5 g2f1 g5c1');
INSERT INTO evaluations VALUES(982,30,'cp -201',NULL);
INSERT INTO evaluations VALUES(982,31,'cp -194',NULL);
INSERT INTO evaluations VALUES(982,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(982,33,'cp -196',NULL);
INSERT INTO evaluations VALUES(982,34,'cp -200',NULL);
INSERT INTO evaluations VALUES(982,35,'cp -189',NULL);
INSERT INTO evaluations VALUES(982,36,'cp -182','c8f5 e4c3 d5d7 c1e3 e8c8 a2a3 f8e7 d1e2 h8e8 e1c1 f5g4 e2f1 f7f5 d1e1 g4h5 h2h3 h7h6 g2g4 f5g4 f3d2 g4h3 f1h3 d7h3 h1h3 h5f7 c3e4 b7b6 e1h1 e7f8 h3h2 e8e6 f2f3 f8e7 e3f2 d8f8 c2c3 a7a5 a3a4 f8d8 c1c2 e6g6 f2h4');
INSERT INTO evaluations VALUES(27448,30,'cp -180',NULL);
INSERT INTO evaluations VALUES(27448,31,'cp -200',NULL);
INSERT INTO evaluations VALUES(27448,32,'cp -186',NULL);
INSERT INTO evaluations VALUES(27448,33,'cp -185',NULL);
INSERT INTO evaluations VALUES(27448,34,'cp -187',NULL);
INSERT INTO evaluations VALUES(27448,35,'cp -181',NULL);
INSERT INTO evaluations VALUES(27448,36,'cp -180','e4c3 d5d7 a2a3 e8c8 c1e3 f8e7 d1d2 h8e8 e1c1 h7h6 f3g1 f5g6 f2f3 c6d4 g1e2 d7c6 c1b1 f7f5 e2c1 g6f7 h1e1 e7f6 c3e2 d4b5 e2c3 b5c3 d2c3 c6c3 b2c3 e5e4 e3d4 f6d4 c3d4 e4d3 c1d3 d8d4 e1e8 f7e8');
INSERT INTO evaluations VALUES(27449,30,'cp -186',NULL);
INSERT INTO evaluations VALUES(27449,31,'cp -187',NULL);
INSERT INTO evaluations VALUES(27449,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(27449,33,'cp -187',NULL);
INSERT INTO evaluations VALUES(27449,34,'cp -194',NULL);
INSERT INTO evaluations VALUES(27449,35,'cp -166',NULL);
INSERT INTO evaluations VALUES(27449,36,'cp -172','d5d7 c1e3 e8c8 a2a3 f8e7 d1d2 h8e8 e1c1 h7h6 f3g1 c6d4 g1f3 d7e6 h1e1 d4f3 g2f3 e6c6 c3e4 b7b6 d2c3 c6c3 e4c3 f5h3 d3d4 e5d4 d1d4 d8d4 e3d4 e7g5 d4e3 h3g2 c3b5 a7a6 b5d4 g2h3 e1g1 g5f6 d4e2 g7g5 e2g3');
INSERT INTO evaluations VALUES(47177,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(47177,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(47177,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(47177,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(47177,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(47177,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(47177,36,'cp -36','c7c5 e2e3 a7a6 g1f3 e7e6 d4c5 b8d7 b2b4 a6a5 c3b5 a5b4 b5d6 f8d6 c5d6 d7b8 c1b2 d8d6 d1c1 e8g8 a2a3 c8d7 f1d3 f8c8 e1g1 b4a3 b2a3 d6c7 a3b2 b8c6 b2f6 g7f6 a1a8 c8a8 c1b2 g8g7 f1b1 a8b8 f3d4 c6e7 f2f4 d7c6 d4c6 c7c6 b2b4 e7g6 g1h1');
INSERT INTO evaluations VALUES(45374,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(45374,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(45374,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(45374,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(45374,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(45374,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(45374,36,'cp -34','e2e3 b8c6 g1f3 e7e6 f1e2 a7a6 e1g1 d8c7 b2b3 c5d4 e3d4 b7b5 c3b1 f8e7 c2c4 d5c4 b3c4 b5c4 e2c4 h7h6 c1e3 c8b7 b1d2 e8g8 a1c1 e7a3 c1b1 a3d6 d1e2 f8c8 f1c1 c7e7 c4b3 c6b4 c1c4 b4d5 f3e5 d5e3 e2e3');
INSERT INTO evaluations VALUES(4914,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(4914,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4914,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(4914,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(4914,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(4914,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(4914,36,'cp 25','d7d6 a2a4 c8g4 h2h3 g4f3 d1f3 g8f6 c2c3 e8g8 d2d3 c6a5 b3a2 b5a4 b1d2 f8e8 f3d1 d8d7 a2b1 a5b3 d2b3 a4b3 d1b3 h7h6 b3d1 c7c6 d1f3 d7e6 b1c2 c5b6 f1e1 d6d5 e4d5 c6d5 c1e3 e8d8 f3g3 f6h5 g3g4 e6g4 h3g4 b6e3');
INSERT INTO evaluations VALUES(4901,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(4901,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(4901,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(4901,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(4901,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(4901,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(4901,36,'cp 34','a2a4 c8g4 h2h3 g4f3 d1f3 g8f6 c2c3 c6a5 b3a2 b5b4 c3b4 c5b4 b1c3 e8g8 d2d3 a8b8 c1g5 c7c6 c3e2 h7h6 g5e3 b4c5 d3d4 e5d4 e2d4 d8e7 d4f5 e7e5 b2b4 c5e3 f2e3 b8b4 f5h6 g7h6 f3f6 e5f6 f1f6 a5c4 a2c4 b4c4 f6h6 c4e4 h6d6 e4e3 a4a5');
INSERT INTO evaluations VALUES(30795,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(30795,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(30795,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(30795,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(30795,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30795,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(30795,36,'cp 20','e7e6 g1f3 g8f6 h2h4 d7d6 f1d3 c7c5 c2c3 f8e7 e1g1 e8g8 b1d2 b8d7 f1e1 f6h5 f4h2 h5f6 d2c4 d6d5 c4e5 c5c4 d3c2 b6b5 h4h5 b5b4 c3b4 e7b4 h5h6 g7g6 e1e2 a7a5 a1c1 d8e7 h2f4 b4d6 e5d7 e7d7 f4d6');
INSERT INTO evaluations VALUES(30656,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30656,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(30656,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(30656,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(30656,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(30656,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(30656,36,'cp 19','g1f3 g8f6 h2h4 c7c5 c2c3 f8e7 b1d2 c5d4 c3d4 e8g8 f1d3 d7d6 e1g1 b8c6 f4g3 e6e5 a2a3 c6d4 e3d4 e5e4 d3e2 e4f3 e2f3 b7f3 d1f3 a8c8 a1c1 d8d7 d2c4 h7h5 g3f4 f6g4 g2g3 g7g6 c4e3 e7f6 e3g4 d7g4 f3g4 h5g4 f4d6 f8e8 c1c8 e8c8 d6e5 f6e5 d4e5');
INSERT INTO evaluations VALUES(30657,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(30657,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30657,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30657,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30657,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(30657,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(30657,36,'cp 13','g8f6 h2h4 f8e7 c2c3 c7c5 f1d3 e8g8 b1d2 d7d6 e1g1 b8d7 f1e1 f8e8 a2a4 a7a6 f4h2 f6g4 f3g5 g4f6 e3e4 c5d4 c3d4 a8c8 g5f3 e6e5 d4d5 d7c5 d3f1 a6a5 f1b5 e8f8 a1c1 f6g4 h2g3 f7f5 e4f5 g4h6 d2e4 c5e4 c1c8 d8c8 e1e4 c8f5');
INSERT INTO evaluations VALUES(52316,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52316,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52316,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52316,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52316,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52316,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(52316,36,'cp 0','g1f3 f8g7 f1e2 d7d6 d2d3 g8f6 e1g1 e8g8 a2a4 b8c6 c2c3 a7a6 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(52315,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52315,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52315,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52315,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52315,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52315,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52315,36,'cp 0','f8g7 d2d3 b8c6 f1e2 g8f6 a2a4 e8g8 e1g1 d7d6 c2c3 a7a6 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(37834,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37834,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37834,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37834,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37834,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37834,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37834,36,'cp 0','c2c3 b8c6 d2d3 g8f6 f1e2 d7d6 e1g1 e8g8 a2a4 a7a6 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(37833,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37833,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37833,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37833,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37833,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37833,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37833,36,'cp 0','b8c6 d2d3 g8f6 f1e2 d7d6 a2a4 e8g8 e1g1 a7a6 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(5369,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5369,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5369,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5369,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5369,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5369,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5369,36,'cp 0','f1c4 e7e6 c4e2 g8f6 e4e5 f6d5 d2d4 c5d4 c3d4 d7d6 b1c3 e8g8 e1g1 d8b6 c3a4 b6d8');
INSERT INTO evaluations VALUES(5311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5311,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5311,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5311,36,'cp 0','e7e6 c4e2 g8e7 d2d4 c5d4 c3d4 d8b6 b1a3 c6d4 a3c4 b6c5 c1e3 c5b4 e3d2 b4c5');
INSERT INTO evaluations VALUES(5120,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5120,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5120,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5120,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5120,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5120,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5120,36,'cp 0','c4e2 g8e7 d2d4 c5d4 c3d4 d8b6 b1a3 c6d4 a3c4 b6c5 c1e3 c5b4 e3d2 b4c5');
INSERT INTO evaluations VALUES(52813,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(52813,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52813,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(52813,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(52813,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(52813,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52813,36,'cp 8','b1c3 d5e4 c3e4 g8f6 c2c3 e7e5 e4f6 d8f6 f3f6 g7f6 g1f3 c8g4 f1e2 b8d7 d2d4 h7h5 g2g3 e8c8 h2h3 g4e6 c1e3 f8h6 e3h6 h8h6 h3h4 e6g4 d4e5 d7e5 f3e5 f6e5 e2g4 h5g4 a1d1 d8h8 e1g1 f7f6 a2a4 c8c7 a4a5');
INSERT INTO evaluations VALUES(52797,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52797,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(52797,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52797,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(52797,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52797,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(52797,36,'cp 13','d5e4 c3e4 g8f6 c2c3 e7e5 e4f6 d8f6 f3f6 g7f6 g1f3 c8g4 f1e2 b8d7 d2d4 h7h5 c1e3 f8h6 e3h6 h8h6 e1c1 e5e4 f3g1 g4e6 f2f3 f6f5 c3c4 d7f6 g1h3 h5h4 b2b3 e8c8 h1e1 a7a5 c1b2 d8g8 e2f1 b7b5 c4b5 c6b5 h3f4');
INSERT INTO evaluations VALUES(52923,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(52923,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52923,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52923,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(52923,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(52923,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52923,36,'cp 6','c3e4 g8f6 c2c3 e7e5 e4f6 d8f6 f3f6 g7f6 g1f3 c8e6 d2d4 b8d7 f1e2 h7h5 g2g3 a7a5 e1g1 a5a4 c1e3 f8e7 f1d1 h8g8 c3c4 e6g4 g1g2 h5h4 a1c1 h4g3 h2g3 e8c8 d1h1 g8h8 d4d5 h8h1 c1h1 c6d5');
INSERT INTO evaluations VALUES(18333,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(18333,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(18333,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(18333,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(18333,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(18333,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(18333,36,'cp 62','b1c3 c8b7 c1d2 d8h4 g1f3 h4h5 f1c4 e8c8 e1c1 f8c5 e3e2 f7f5 e4e5 d8e8 h1e1 g8f6 a2a3 f6e4 b2b4 c5f2 e1f1 h5g4 c4d5 c6e5 d5b7 c8b7 f1f2 e4f2 e2f2 d7d6 h2h3 g4h5 h3h4 e5f3 g2f3 b7b8 d2g5 h8f8 f3f4 h7h6');
INSERT INTO evaluations VALUES(18332,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(18332,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(18332,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(18332,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(18332,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(18332,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(18332,36,'cp 75','c8b7 c1d2 d8h4 g1f3 h4h5 g2g4 h5g6 f1c4 e8c8 e1c1 f7f6 h1g1 c8b8 e3e2 f8c5 h2h4 g6e8 h4h5 g8e7 a2a3 a7a5 d2f4 d7d6 c1b1 c6e5 c4a2 e7c6 f4e3 g7g6 c3d5 e5f3 e2f3 c6e5 f3e2 c5e3 e2e3 h8f8');
INSERT INTO evaluations VALUES(9953,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(9953,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9953,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9953,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(9953,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9953,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9953,36,'cp -15','d7c6 d2d3 d8e7 d1e2 f6d7 e1g1 d7f8 c3d1 f7f6 d1e3 c8e6 c2c3 b4d6 f1e1 f8g6 d3d4 e8g8 e3f5 e7d7 d4e5 f6e5 f5d6 c7d6 f3g5 h7h6 g5e6 d7e6 c1e3 g6f4 e3f4 f8f4 a1d1 f4f7 e2d2 a8d8 d2c2 d8e8');
INSERT INTO evaluations VALUES(9502,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9502,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9502,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9502,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9502,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9502,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9502,36,'cp 0','d1e2 d8e7 d2d3 a7a5 e1g1 b4d6 a2a3 c8e6 h2h3 f6d7 c1e3 d7b6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(12359,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(12359,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(12359,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(12359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12359,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(12359,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(12359,36,'cp -12','g1f3 f8e7 e1g1 e8g8 c1f4 a7a6 a2a4 d7d6 h2h3 c8d7 f1e1 d8c7 f4g3 f8d8 d1e2 d7e8 c3d1 c6a5 e4e5 d6e5 g3e5 e7d6 d1c3 d6e5 e2e5 a8c8 e5e3 f6d7 f3d2 h7h6 d2b3 d7f8 b3a5 c7a5 f2f4 a5c7');
INSERT INTO evaluations VALUES(39019,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(39019,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(39019,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(39019,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(39019,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(39019,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(39019,36,'cp 79','g8e7 g1f3 e8g8 f1d3 d7d6 h2h4 e6e5 d4e5 d6e5 h4h5 c8g4 h5g6 e7g6 d3e2 d8d1 e2d1 b8a6 f3h4 g4d1 e1d1 f8d8 d1e2 g6f4 e2f3 f4e6 h4f5 e6d4 f3g4 d4f5 g4f5 c7c6 c1g5 d8d4 b2b3 a6c5 g5e3 c5e6 e3d4 e5d4');
INSERT INTO evaluations VALUES(43077,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43077,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(43077,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(43077,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(43077,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(43077,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(43077,36,'cp 78','g1f3 e8g8 f1d3 d7d6 h2h4 e6e5 d4e5 d6e5 h4h5 c8g4 h5g6 e7g6 d3e2 d8d1 e2d1 b8c6 c1e3 a8d8 c3d5 c6d4 f3d4 e5d4 e3g5 d8e8 f2f3 f7f6 d5c7 e8c8 g5d2 c8c7 f3g4 f8e8 b2b3 d4d3 d2c3 b7b5 e1d2 b5c4 b3c4 e8e4 d1f3');
INSERT INTO evaluations VALUES(43076,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(43076,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(43076,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(43076,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(43076,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(43076,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(43076,36,'cp 61','e8g8 f1d3 d7d6 h2h4 e6e5 d4e5 d6e5 h4h5 c8g4 h5g6 e7g6 d3e2 d8d1 c3d1 b8c6 d1e3 g4d7 c1d2 a8d8 g2g3 c6d4 a1d1 d7c8 e1f1 b7b6 b2b3 f8e8 f3h4 g6h4 h1h4 c7c6 d2c3 d4e2 f1e2 d8d1 e2d1 g7f8 d1e2 f7f6 c3d2 c8e6 f2f3 f8c5 h4h1 e8d8 e3f5');
INSERT INTO evaluations VALUES(37338,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(37338,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(37338,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(37338,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(37338,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(37338,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(37338,36,'cp 72','f1d3 d7d6 h2h4 e6e5 d4e5 d6e5 h4h5 b8a6 h5h6 g7f6 d3e2 d8d1 e2d1 c8e6 b2b3 e7c6 c3d5 f6d8 c1e3 f7f6 e1g1 c6b8 b3b4 b8d7 d1a4 c7c6 b4b5 c6b5 a4b5 a6c5 b5d7 c5d7 f1d1');
INSERT INTO evaluations VALUES(22520,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22520,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22520,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(22520,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(22520,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22520,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(22520,36,'cp 8','g8f6 d2d3 f8c5 b1c3 a7a6 c1e3 d7d6 e1g1 e8g8 e3c5 d6c5 c3d5 d8d6 d5e3 b7b5 c4a2 c6d4 a3a4 c7c6 a4b5 c6b5 f3d4 c5d4 e3d5 f6d5 a2d5 a8b8 h2h3 c8e6 d5e6 f7e6 d1e2 b8b6 g2g3 g8h7 g1g2 d6c5');
INSERT INTO evaluations VALUES(16152,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(16152,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(16152,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16152,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(16152,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(16152,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(16152,36,'cp 6','d2d3 f8c5 e1g1 a7a6 b2b4 c5b6 c1e3 e8g8 b1d2 d7d6 h2h3 b6e3 f2e3 c6e7 d1e1 e7g6 f3h4 g6h4 e1h4 c8e6 f1f2 b7b5 c4b3 a6a5 a1f1 f6d7 h4d8 f8d8 b3d5 e6d5 e4d5 a5b4 a3b4 d7b6 e3e4 d8d7');
INSERT INTO evaluations VALUES(16149,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16149,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(16149,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(16149,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16149,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(16149,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(16149,36,'cp 8','f8c5 e1g1 d7d6 b1c3 c6d4 c3a4 b7b5 f3d4 b5c4 d4f5 c8f5 e4f5 c4d3 a4c5 d6c5 c2d3 e8g8 c1e3 d8d7 e3c5 f8d8 d1f3 d7d3 f3d3 d8d3 f1e1 a7a5 h2h3 e5e4 c5e7 d3d4 e7f6 g7f6 a1c1 a8b8 c1c7 b8b2 c7e7 b2a2 e7e8 g8g7 e1e4 d4e4 e8e4 a2a3');
INSERT INTO evaluations VALUES(9790,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(9790,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(9790,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(9790,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(9790,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(9790,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(9790,36,'cp 9','b1c3 a7a6 c1e3 d7d6 e1g1 e8g8 c3d5 f6d5 c4d5 c5e3 f2e3 c6e7 d5b3 e7g6 f3d2 d8g5 d1f3 c8e6 b3e6 f7e6 f3e2 a6a5 f1f8 a8f8 d2f3 g5e7 h2h3 g6h4 f3d2 e7g5 a1f1 f8f1 d2f1 d6d5 g1h2 g5e7');
INSERT INTO evaluations VALUES(33199,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33199,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(33199,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(33199,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33199,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(33199,36,'cp -5','g1f3 d7d6 b1c3 g7g6 h5g5 h7h6 g5g3 g8e7 d2d3 c8e6 c1e3 c5e3 f2e3 e6c4 d3c4 a7a6 c4c5 b8d7 c5d6 f6d6 f3d2 d6e6 e1c1 e8c8 h1f1 c8b8 g3f3 f7f6 d2b3 e7c8 c3d5 h8f8 d1d3 f6f5 f1d1 e6f7 e4f5 g6f5 e3e4 f5e4 f3e4 c8d6 e4b4');
INSERT INTO evaluations VALUES(33197,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(33197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33197,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33197,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33197,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33197,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33197,36,'cp -1','d7d6 b1c3 g7g6 h5g5 h7h6 g5g3 g8e7 d2d3 c8e6 c1e3 c5e3 f2e3 e6c4 d3c4 a7a6 c4c5 b8d7 c5d6 f6d6 f3d2 d6e6 h1f1 e8c8 e1c1 c8b8 b2b3 d7b6 a2a4 e7c6 g3f3 e6e7 d2b1 d8d1 f1d1 h6h5 c1b2 h8f8 h2h3 c6b4 a4a5 b6c8 c3a4 c8d6 b1c3 f8h8 c3d5');
INSERT INTO evaluations VALUES(33078,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(33078,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(33078,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(33078,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(33078,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33078,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33078,36,'cp 18','b1c3 g7g6 h5g5 h7h6 g5g3 g8e7 d2d3 c8e6 c1e3 b8c6 e3c5 d6c5 e1c1 e8c8 f3d2 g6g5 h2h4 c6d4 d1f1 e7g6 h4g5 h6g5 c1b1 h8h1 f1h1 d8h8 h1h8 f6h8 b2b3 h8h4 c4e6 f7e6 d2f3 h4g3 f2g3 d4f3 g2f3 g6e7 a2a3 e7c6 c3e2 c8d8');
INSERT INTO evaluations VALUES(33076,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(33076,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(33076,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(33076,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(33076,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(33076,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(33076,36,'cp 7','g7g6 h5g5 h7h6 g5g3 g8e7 d2d3 c8e6 c1e3 b8c6 e3c5 d6c5 e1c1 e8c8 c1b1 c8b8 c4e6 f6e6 f3d2 b7b6 a2a3 f7f5 e4f5 e7f5 g3g4 c6d4 h2h3 h8f8 h1f1 g6g5 d1e1 d8e8 c3e2 d4c6 e2g3 f5d4 f2f3 f8f4 g4e6');
INSERT INTO evaluations VALUES(52597,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(52597,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(52597,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(52597,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(52597,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(52597,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(52597,36,'cp -21','h5h4 b8c6 g1e2 a7a6 d2d3 b7b5 c4b3 f8g7 e1g1 h7h6 a2a4 c8d7 f2f4 g8f6 b1d2 e8g8 c2c3 c5c4 d3c4 c6a5 b3d1 b5c4 f4f5 g6g5 h4h3 e7e6 e2g3 a8b8 g1h1 d8c7 d1g4 f6g4 h3g4 e6f5 g3f5');
INSERT INTO evaluations VALUES(10710,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(10710,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10710,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(10710,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10710,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10710,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(10710,36,'cp 17','d2d3 e8g8 e1g1 h7h6 b5a4 a7a6 c1e3 f8e8 a2a3 d6f8 a4b3 d7d6 h2h3 c8e6 c3d5 e6d5 e4d5 c6e7 c2c4 g7g6 b3c2 f8g7 f1e1 c7c6 d5c6 e7c6 c2a4 d6d5 d1d2 g8h7 c4d5 d8d5 a1c1 a8c8 c1c5 d5d6 e1c1 e8d8 a4c6 c8c6 c5e5');
INSERT INTO evaluations VALUES(10711,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(10711,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10711,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(10711,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(10711,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(10711,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(10711,36,'cp 13','e8g8 e1g1 a7a6 b5c4 h7h6 a2a4 d6c5 c3d5 d7d6 c1e3 f6d5 c4d5 c6e7 d5b3 c5e3 f2e3 c8e6 b3e6 f7e6 h2h3 d8d7 b2b4 f8f7 d1d2 a8f8 a4a5 e7g6 f3h2 f7f1 a1f1 f8f1 h2f1 d6d5 d2c3 d5d4 c3c4 d7d6');
INSERT INTO evaluations VALUES(4511,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4511,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(4511,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(4511,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(4511,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(4511,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4511,36,'cp 13','e1g1 h7h6 c1e3 a7a6 b5a4 f8e8 a2a3 d6f8 a4b3 d7d6 h2h3 b7b5 f3d2 c8e6 c3d5 f6d7 f2f4 e5f4 d5f4 e6b3 c2b3 c6e7 d1f3 e7g6 f4g6 f7g6 d3d4 c7c5 a1c1 d8e7 b3b4 c5d4 e3d4 a8c8 f3g3 e7e6 g3f2 c8c1 f1c1 d7f6 d4f6');
INSERT INTO evaluations VALUES(4510,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(4510,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(4510,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(4510,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(4510,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(4510,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4510,36,'cp 20','a7a6 b5c4 h7h6 a2a4 d6c5 c1e3 c5e3 f2e3 d7d6 c3d5 f6d5 c4d5 a6a5 d1d2 c6e7 d5b3 e7g6 f1f2 c8e6 b3e6 f7e6 h2h3 g6h4 f3h2 d8g5 a1f1 b7b6 b2b3 g5g3 d2e2 f8f2 f1f2 g3g5 h2f3 h4f3 e2f3 g5e7 f2f1 a8f8 f3g4 f8f1 g1f1 g7g5');
INSERT INTO evaluations VALUES(36620,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36620,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36620,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36620,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(36620,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36620,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36620,36,'cp 10','f1e2 c7c5 e1g1 d8b6 d1c1 b8c6 c2c3 c8e6 b1d2 a8c8 d4c5 b6c5 e3e4 c5a5 a2a4 f8d8 f1d1 a5c7 c1b1 e6d7 d2c4 d7e6 f3d2 b7b6 d2f3 h7h6 b1c2 c7b7 c4e3 c6a5 f3d4 b7e4 c2e4 f6e4 d4e6 f7e6');
INSERT INTO evaluations VALUES(36619,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36619,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(36619,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(36619,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(36619,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36619,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36619,36,'cp 18','c7c5 e1g1 d8b6 d1c1 b8c6 c2c3 c8e6 b1d2 a8c8 d4c5 b6c5 e3e4 c5a5 a2a4 a5c7 f1e1 a7a6 c1d1 h7h6 f3d4 e6d7 d4b3 e7e5 f4e3 d6d5 a4a5 d5d4 c3d4 e5d4 b3d4 c6a5 b2b3 f8d8 e2f1 c7c3 a1c1 c3c1 d1c1');
INSERT INTO evaluations VALUES(35962,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35962,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(35962,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(35962,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(35962,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(35962,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(35962,36,'cp 9','e1g1 d8b6 d1c1 b8c6 c2c3 c8e6 b1d2 c5d4 e3d4 a8c8 f1e1 e6d5 f4e3 e7e6 c3c4 c6e7 e2d3 e7f5 d3f5 e6f5 b2b3 b6a6 a2a4 f6h5 c1a3 f5f4 c4d5 f4e3 f2e3 f8e8 a1c1 b7b5 g1f2 g7h6 e3e4 c8c1 e1c1 h6f4 c1c6');
INSERT INTO evaluations VALUES(33248,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(33248,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(33248,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(33248,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(33248,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(33248,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(33248,36,'cp 15','b8c6 c2c3 d7d6 g2g3 f6g6 f1g2 g8f6 e1g1 f6g4 e3f4 e8g8 h2h3 g4f6 d4c6 b7c6 b2b4 c5b6 f1e1 f6h5 f4e3 h5g3 f2g3 g6g3 d1d2 c8h3 e3b6 a7b6 e1e3 g3g2 d2g2 h3g2 g1g2 a8a4 a2a3 f8a8 e4e5 d6e5 e3e5 a8d8 e5e7 c6c5 e7c7 c5b4 c3b4');
INSERT INTO evaluations VALUES(38783,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(38783,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(38783,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(38783,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(38783,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38783,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(38783,36,'cp 8','a2a3 b4c3 b2c3 g8f6 e2e3 c7c5 c3c4 b8c6 d4c5 f6e4 c4d5 e6d5 f1d3 d8a5 c1d2 a5c5 e1g1 c8g4 d1e1 e4d2 f3d2 e8g8 e1b1 b7b6 b1b2 f8c8 h2h3 g4h5 a3a4 h5g6 f1d1 c8c7 a1c1 a8d8 d2b3 c5e7 b3d4 c6d4 b2d4');
INSERT INTO evaluations VALUES(38789,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(38789,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(38789,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(38789,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(38789,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(38789,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(38789,36,'cp 0','b4c3 b2c3 c7c5 d4c5 g8f6 c3c4 b8c6 e2e3 f6e4 c4d5 e6d5 f1d3 d8a5 c1d2 a5c5 e1g1 c8g4 d1e1 g4f3 d2b4 c6b4 a3b4 c5b6 g2f3 e4g5 e1e2 e8g8 h2h4 b6h6 h4g5 h6g5 g1h1 g5h6 h1g2 h6g5 g2h1');
INSERT INTO evaluations VALUES(38839,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38839,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38839,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38839,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38839,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38839,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38839,36,'cp 13','b2c3 g8f6 e2e3 c7c5 c3c4 b8c6 d4c5 f6e4 c4d5 e6d5 f1d3 d8a5 c1d2 a5c5 e1g1 c8g4 d1e1 e4d2 f3d2 e8g8 h2h3 g4d7 e1b1 b7b6 d2f3 d7e6 c2c3 h7h6 d3a6 e6c8 a6b5 c8d7 a3a4 c5c3 f1d1 c3f6 d1d5 a8d8 f3d4 c6d4');
INSERT INTO evaluations VALUES(38838,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(38838,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(38838,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(38838,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38838,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(38838,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38838,36,'cp 0','c7c5 c3c4 b8c6 d4c5 g8f6 e2e3 f6e4 c4d5 e6d5 f1d3 d8a5 c1d2 a5c5 e1g1 c8g4 d1e1 g4f3 d2b4 c6b4 a3b4 c5c8 g2f3 e8g8 f3e4 c8g4 g1h1 g4f3 h1g1');
INSERT INTO evaluations VALUES(34805,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(34805,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(34805,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(34805,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(34805,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(34805,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(34805,36,'cp 158','f1b5 b8c6 e1g1 g8e7 d2d4 a7a6 d4e5 d6d1 f1d1 c8g4 b5a4 e7g6 a4c6 b7c6 d1d4 g4f3 g2f3 f8e7 d4e4 e8c8 f3f4 f7f6 c1e3 g6h4 g1f1 h8f8 a1d1 d8d1 c3d1 h4f3 h2h3 f8d8 d1c3 f3d2 e3d2 d8d2 e4e2 d2d4 c3e4 c8d7 e5f6 e7f6 c2c3 d4d5 f1g2 d5f5');
INSERT INTO evaluations VALUES(52482,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52482,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(52482,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(52482,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(52482,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(52482,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52482,36,'cp 0','d7d5 d2d4 g8f6 f1e2 b8c6 e1g1 c8g4 h2h3 g4f3 e2f3 c5d4 e3d4 e7e6 c1e3 a8c8 c2c3 f8d6 b1d2 e8g8 f3e2 h7h6 e2d3 a7a6 f1e1 f8e8 a2a4 d8c7 d3c2 c6b4 c2b1 b4c6');
INSERT INTO evaluations VALUES(53321,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(53321,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(53321,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53321,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(53321,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53321,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53321,36,'cp 13','d2d4 g8f6 f1e2 b8c6 e1g1 c8g4 h2h3 g4f3 e2f3 c5d4 e3d4 e7e6 c1e3 f8e7 c2c3 d8b6 d1b3 b6c7 b1d2 e8g8 f3e2 h7h6 b3d1 a7a6 e2d3 e7d6 a2a4 c6e7 d1f3 f6h7 a1c1 f7f5 c3c4 h7g5 f3e2 f8f6 c4c5');
INSERT INTO evaluations VALUES(52988,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52988,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(52988,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(52988,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(52988,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(52988,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52988,36,'cp 2','b8d7 b5e2 g8f6 g1f3 a7a6 a2a4 b7b6 d2d4 c5d4 d1d4 c8b7 c1e3 e7e6 e1c1 d8c7 f3d2 d6d5 e4d5 f8c5 d5d6 c5d6 d2c4 d6c5 e3f4 e6e5 f4e5 d7e5 d4e5 c7e5 c4e5 c5f2 e2f3 b7f3 e5f3 a8d8 c1b1 e8g8 d1d8 f8d8 h1d1 d8e8 c3d5 f6d5 d1d5 g7g6');
INSERT INTO evaluations VALUES(18600,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(18600,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(18600,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(18600,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(18600,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(18600,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(18600,36,'cp 18','g1f3 a7a6 b5d7 c8d7 d2d4 c5d4 d1d4 e7e5 d4d3 a8c8 e1g1 d7e6 f1d1 f8e7 c3d5 e6d5 d3d5 d8c7 f3g5 c7c4 d5b7 h7h6 g5f3 c8c7 b7a8 c7c8 a8a7 c8c7 a7e3 g8f6 b2b3 c4e4 c2c4 e8g8 e3b6 f8c8 b6a6 d6d5 d1e1 e4g4 f3e5 g4f5 c1a3 d5c4 b3c4');
INSERT INTO evaluations VALUES(18599,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18599,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18599,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(18599,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(18599,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(18599,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18599,36,'cp 6','a7a6 b5d7 c8d7 d2d4 c5d4 d1d4 e7e5 d4d3 a8c8 e1g1 d7e6 f1d1 h7h6 f3d2 f8e7 d2f1 c8c4 b2b3 c4d4 d3e2 g8f6 c1b2 d8a5 d1d4 e5d4 c3d1 a5e5 e2d3 e5e4 d3e4 f6e4 b2d4 e4f6 a2a4 h6h5 f1e3 e8d7 d1c3 d7c6 c3e2 g7g6 d4b2 h8c8');
INSERT INTO evaluations VALUES(17319,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(17319,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(17319,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(17319,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17319,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(17319,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(17319,36,'cp 23','b5d7 c8d7 d2d4 c5d4 d1d4 e7e5 d4d3 a8c8 e1g1 d7e6 f1d1 h7h6 f3d2 f8e7 d2f1 c8c4 b2b3 c4d4 d3f3 d4d1 f3d1 g8f6 f1e3 d8b6 c1b2 b6d4 d1f1 b7b5 a1d1 d4c5 h2h3 b5b4 c3d5 f6d5 e4d5 e6c8 f1c4 c5c4 e3c4 f7f6 a2a4 e8f7 f2f3');
INSERT INTO evaluations VALUES(22876,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(22876,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(22876,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(22876,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(22876,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(22876,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(22876,36,'cp -59','e5e4 f3g1 d7d5 d2d3 c8f5 d3e4 d5e4 b1c3 d8e7 c3d5 e7d7 d5e3 f8d6 e3f5 d7f5 c2c3 g8f6 d1c2 e8g8 g1f3 f8e8 c1e3 d6c5 f3d4 c6d4 e3d4 c5d4 c3d4 c7c6 a2a3 a8d8 e2e3 d8d5 b2b4 g7g6 a1c1 g8g7 h2h3 h7h5 c2b3 e8d8 c1c5 f5e6 c5d5');
INSERT INTO evaluations VALUES(22971,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(22971,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(22971,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(22971,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(22971,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(22971,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(22971,36,'cp -65','f3g1 d7d5 d2d3 c8f5 d3e4 d5e4 b1c3 d8e7 c3d5 e7d7 d5e3 f8d6 c1d2 d7e6 c2c3 g8f6 d1b3 d6c5 b3e6 f7e6 g2h3 c5e3 d2e3 f6d5 e3d2 e8c8 e1c1 c6e5 h3f5 e6f5 g1h3 h7h6 h3f4 e5g4 f4d5 d8d5 d2e3 g4e3 f2e3 h8d8 d1d4 c7c5 d4d5 d8d5 c3c4 d5d6 h1d1 d6d1 c1d1 c8c7 a2a4');
INSERT INTO evaluations VALUES(36556,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(36556,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(36556,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(36556,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(36556,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(36556,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(36556,36,'cp 37','b1c3 c7c5 d4d5 e7e6 e1g1 f8e8 c1g5 h7h6 g5f6 d8f6 f3e1 e6e5 e2e4 h6h5 h2h4 e8f8 d1d2 f6e7 e1f3 b8d7 g2h3 a7a6 a2a4 d7f6 h3c8 a8c8 g1g2 c8e8 d2e2 g7h6 f3h2 e8b8 f1e1 b7b6 e1h1 g8g7 b2b3');
INSERT INTO evaluations VALUES(24815,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(24815,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24815,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24815,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(24815,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(24815,36,'cp 7','e7e6 g1f3 f8d6 f4g3 e8g8 f1e2 f8e8 e1g1 d6g3 h2g3 e6e5 d4e5 c6e5 d1b3 e5f3 e2f3 a8b8 f1d1 d8e7 c3c4 d5c4 d2c4 f5e6 b3c3 c7c6 c4e5 e6f5 e5d3 b8d8 d3f4 g7g6 a2a3 a7a6 a1c1 f5e4 b2b4 h7h6 c3b2 d8d1 c1d1');
INSERT INTO evaluations VALUES(25151,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25151,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(25151,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(25151,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(25151,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(25151,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(25151,36,'cp 19','g1f3 f8d6 f4g3 e8g8 f1e2 f8e8 f3h4 c6e7 e1g1 c7c5 d4c5 d6c5 a1c1 h7h6 a2a3 c5d6 d1b3 f6e4 d2e4 f5e4 h4f3 e7f5 f3d2 f5g3 h2g3 d8b6 b3b6 a7b6 c3c4 d6c5 f1d1 e8d8 b2b4 c5e7 d2e4 d5e4');
INSERT INTO evaluations VALUES(41201,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(41201,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(41201,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(41201,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(41201,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41201,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(41201,36,'cp -24','d6e5 f3e5 e8g8 f1e2 c7c5 d4c5 d8a5 e1g1 a5c5 e5d3 c5a5 e2f3 b8c6 f1e1 f8e8 f3c6 b7c6 d1f3 c8f5 f3c6 a8c8 e1e5 a5d8 c6f3 f5d3 f3d3 d8d3 c2d3 f6g4 e5e2 g7c3 b2c3 c8c3 c1f4 c3d3 h2h3 g4f6');
INSERT INTO evaluations VALUES(41530,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(41530,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(41530,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(41530,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(41530,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(41530,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(41530,36,'cp -12','f3e5 e8g8 f1c4 f6d7 e5d7 d8d7 c1e3 b8c6 d4d5 g7c3 b2c3 c6a5 c4e2 f8d8 c3c4 c7c6 d1d2 b7b6 e1c1 c8b7 d5d6 c6c5 d2c3 a5c6 e2f3 c6d4 e3d4 c5d4 d1d4 b7f3 c3f3 a8c8 h1d1 e7d6 c1b2 d8e8 d4d6 d7a4 f3b3 c8c4 d6d8 e8d8 b3a4 c4a4 d1d8 g8g7 f2f3 a4a5 c2c4 g7f6 d8d7');
INSERT INTO evaluations VALUES(41519,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(41519,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(41519,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(41519,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(41519,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(41519,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(41519,36,'cp -16','e8g8 f1c4 f6d7 e5d7 d8d7 c1e3 b8c6 d4d5 c6a5 c4e2 g7c3 b2c3 f8d8 c3c4 c7c6 d1d2 b7b6 e1c1 c8b7 d5d6 c6c5 d2c3 a5c6 e2f3 c6d4 e3d4 c5d4 d1d4 b7f3 c3f3 a8c8 h1d1 e7d6 c1b2 d7a4 f3b3 a4e8 b3d3 c8c5 d4d6 d8d6 d3d6 c5c4 d6d8 e8d8 d1d8 g8g7 d8d7 c4b4 b2c3 a7a5');
INSERT INTO evaluations VALUES(36787,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(36787,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(36787,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(36787,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(36787,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(36787,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(36787,36,'cp -11','f1c4 f6d7 e5d7 d8d7 c1e3 b8c6 d4d5 c6a5 c4e2 g7c3 b2c3 f8d8 c3c4 c7c6 d1d2 b7b6 e1c1 c8b7 d5d6 c6c5 d2c3 a5c6 e2f3 c6d4 e3d4 c5d4 d1d4 b7f3 c3f3 a8c8 h1d1 e7d6 c1b2 d7a4 f3b3 a4e8 b3d3 c8c5 d4d6 d8d6 d3d6 c5c4 d6d8 e8d8 d1d8 g8g7 b2b3 c4e4 d8d2 e4e1 c2c4 e1g1 g2g3 g7f6 d2d7 g1b1 b3a3');
INSERT INTO evaluations VALUES(36785,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(36785,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(36785,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(36785,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(36785,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(36785,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(36785,36,'cp -11','f6d7 e5d7 d8d7 c1e3 b8c6 d4d5 c6a5 c4e2 g7c3 b2c3 f8d8 c3c4 c7c6 d1d2 b7b6 e1c1 c8b7 d5d6 c6c5 d2c3 a5c6 e2f3 c6d4 e3d4 c5d4 d1d4 b7f3 c3f3 a8c8 h1d1 e7d6 g2g3 h7h5 c1b2 d8e8 d4d6 d7a4 f3b3 a4c4 b3c4 c8c4 d6d8 c4e4 d8e8 e4e8 d1d2 g8g7 c2c4 g7f6 b2c3 e8e1 d2d7 e1e2 d7a7');
INSERT INTO evaluations VALUES(45194,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45194,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45194,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45194,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45194,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45194,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45194,36,'cp -12','b8c6 c3d5 e7e5 c2c4 f8e7 f1e2 e8g8 e1g1 a7a6 b2b3 c8g4 c1e3 g4d7 d5e7 d8e7 f3e1 b7b5 e1c2 c6d4 c2d4 c5d4 e3c1 a6a5 f2f4 a5a4 f4e5 d6e5 c1g5 a8a6 c4b5 d7b5 g5f6 g7f6 b3a4 b5a4 d1d2 a4c6 e2d1 f8b8');
INSERT INTO evaluations VALUES(13310,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(13310,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13310,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13310,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(13310,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13310,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(13310,36,'cp 0','c3d5 e7e5 c2c4 f8e7 f1e2 a7a6 e1g1 b7b5 b2b3 e8g8 h2h3 a8b8 c1d2 c8d7 a1b1 b5b4 b1a1 a6a5 a2a3 b8a8 a3b4 f6d5 c4d5 c6b4 d2c3 f7f5 f3d2 g7g6 d2c4 d7b5 c4a3 b5d7');
INSERT INTO evaluations VALUES(55005,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(55005,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(55005,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(55005,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(55005,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(55005,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(55005,36,'cp -98','c7c6 b5d3 e7e5 e4d5 c6d5 d3b5 b8c6 d2d4 d8b6 a2a4 b6d4 d1d4 e5d4 g1e2 a7a6 b5d3 g8e7 b1d2 g7g6 e1g1 f8g7 f1e1 e8g8 e2f4 c8f5 d2b3 f8e8 d3f5 e7f5 c1d2 f5d6 g1f1 d6c4 f4d5 c4b2 d5b6 a8d8');
INSERT INTO evaluations VALUES(52693,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(52693,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(52693,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(52693,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(52693,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(52693,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(52693,36,'cp -75','b5d3 e7e5 e4d5 c6d5 d3b5 b8c6 d2d4 e5d4 d1d4 g8f6 g1f3 f8e7 d4a4 e8g8 e1g1 c8d7 b1c3 f8e8 c1g5 a7a6 b5c6 b7c6 f3e5 h7h6 g5f4 f6h5 e5d7 d8d7 a1d1 h5f4 a4f4 e7f6 f4d2 d7d6 b2b3 e8e5 c3e2 a8e8 e2g3');
INSERT INTO evaluations VALUES(52803,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(52803,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(52803,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(52803,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(52803,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(52803,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(52803,36,'cp -88','e7e5 e4d5 c6d5 d3b5 b8c6 d2d4 d8b6 a2a4 b6d4 d1d4 e5d4 b1d2 a7a6 b5d3 f8c5 d2b3 c5a7 c1f4 g8f6 g1e2 e8g8 f2f3 f8e8 e1d2 c8e6 f4g3 f6d7 g3f2 d7e5 b3d4 e6d7 h1f1 c6b4 e2f4 a8c8 f1e1 a7c5 e1e5 e8e5 c2c3 c5d4 f2d4 b4d3 d2d3 d7f5 d3d2 e5e8 a4a5 f5d7 f4d5 d7c6');
INSERT INTO evaluations VALUES(53767,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(53767,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(53767,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(53767,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(53767,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(53767,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(53767,36,'cp -65','e4d5 c6d5 d3b5 b8c6 d2d4 d8b6 a2a4 e5d4 g1f3 f8b4 b1d2 a7a6 b5e2 g8f6 e1g1 b4d2 f3d2 e8g8 d2b3 f8e8 a4a5 b6c7 f1e1 c8d7 b3d4 c6a5 c1e3 a5c4 e2c4 d5c4 d1f3 d7g4 f3g3 c7g3 h2g3 b7b5');
INSERT INTO evaluations VALUES(6620,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(6620,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(6620,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(6620,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(6620,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(6620,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(6620,36,'cp 187','c2c3 b4a5 d2d4 e5d4 e1g1 a5b6 c3d4 d7d6 h2h3 g8f6 b1c3 h7h6 c1f4 e8g8 f1e1 c6e7 a2a4 b6a5 f4h2 f6h7 h2g3 e7g6 a1b1 c7c6 b2b4 a5c7 a4a5 h7g5 d4d5 g5f3 d1f3 d8g5 c4f1 c8d7 f3e3 f7f5 f2f4 g5h5 d5c6');
INSERT INTO evaluations VALUES(6617,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(6617,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(6617,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(6617,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(6617,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(6617,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(6617,36,'cp 186','b4a5 d2d4 d7d6 d4e5 c8g4 e5d6 g8f6 c1f4 e8g8 b1d2 c7d6 e1g1 a8c8 f1e1 h7h6 d2f1 f8e8 h2h3 g4h5 f1g3 c6e5 f4e5 d6e5 d1d8 c8d8 c4b5 e8e6 g3h5 f6h5 b2b4 a5c7 a2a4 c7b6 b5c4');
INSERT INTO evaluations VALUES(6729,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(6729,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(6729,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(6729,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(6729,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(6729,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(6729,36,'cp 183','d2d4 e5d4 e1g1 a5b6 c3d4 d7d6 b1c3 g8f6 h2h3 h7h6 c1f4 e8g8 f1e1 c6e7 a2a4 b6a5 f4h2 a5b4 d1b3 a7a5 e4e5 f6h7 e1e2 h7g5 f3g5 h6g5 a1d1 c8d7 c4f7 f8f7 e5e6 d7e6 e2e6 b4c3 b2c3 b7b6 b3c4 d8d7 d4d5 e7f5 d1e1 g8h8 h2g3');
INSERT INTO evaluations VALUES(6732,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(6732,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(6732,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(6732,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(6732,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(6732,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(6732,36,'cp 176','e5d4 e1g1 d7d6 d1b3 d8e7 c3d4 a5b6 c4b5 c8d7 b1c3 c6a5 b3b4 d7b5 b4b5 e7d7 b5d3 a5c6 d4d5 c6e7 a2a4 a7a6 f1e1 e7g6 a4a5 b6a7 h2h4 d7e7 c1g5 f7f6 g5e3 g6e5 d3c2 e5f3 g2f3 g8h6 e3a7 a8a7 c2a4 e7d7 a1c1 a7a8 g1h2 d7a4 c3a4 e8d8 h4h5 g7g6 e1g1 g6h5');
INSERT INTO evaluations VALUES(20635,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(20635,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(20635,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(20635,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(20635,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(20635,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(20635,36,'cp 57','d4f5 c8f5 e4f5 d8d7 f1c4 g8e7 e1g1 e7f5 f2f4 f8e7 g1h1 e8c8 c4d3 h8e8 f4e5 g7g6 e5d6 e7d6 d3f5 g6f5 c1g5 d6e7 d1h5 e7g5 h5g5 e8e5 a1d1 d7e6 d1d8 c6d8 g5h4 d8c6 h1g1 h7h6 h2h3 c8b8 h4f2 a7a6 f1d1 e6c4 g2g3');
INSERT INTO evaluations VALUES(20642,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(20642,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(20642,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(20642,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(20642,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(20642,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(20642,36,'cp 55','c8f5 e4f5 d8d7 f1c4 g8e7 e1g1 e7f5 f2f4 e8c8 f4e5 c8b8 c1g5 f8e7 g5e7 c6e7 c4f7 d7c7 e5d6 f5d6 f7b3 c7c5 g1h1 h8e8 d1f3 e7f5 f3f2 c5e5 f2f4 e5c5 c3d5 e8e2 a1e1 d8e8 e1e2 e8e2 d5c3 e2e8 f1d1 a7a6 b3f7 c5e5 f4e5');
INSERT INTO evaluations VALUES(26444,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(26444,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26444,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(26444,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(26444,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(26444,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(26444,36,'cp 53','e4f5 d8d7 f1c4 g8e7 e1g1 e7f5 f2f4 e8c8 f4e5 c8b8 c4d3 g7g6 d3f5 g6f5 e5d6 f8d6 c1g5 c6e7 d1d6 d7d6 g5f4 d6f4 f1f4 d8d2 a1e1 e7g6 f4f2 d2f2 g1f2 h8d8 e1e2 b8c7 f2e3 d8e8 e3d2 e8e2 c3e2 c7d6 d2e3 d6e5 e2d4 f5f4 e3d3 e5d5 c2c4 d5c5 d3c3 a7a5 h2h3 c5d6 d4b3 a5a4 b3d2');
INSERT INTO evaluations VALUES(26443,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(26443,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(26443,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(26443,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(26443,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(26443,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(26443,36,'cp 38','d8a5 a2a3 g8f6 c1g5 f8e7 f1b5 e8g8 e1g1 c6d4 b5d3 d6d5 f1e1 e5e4 d3e4 d5e4 d1d4 h7h6 g5f6 e7f6 d4e4 f6c3 b2c3 a5c3 e4b7 f8e8 b7b4 c3c2 f5f6 g7f6 h2h3 c2f5 b4b7 e8b8 b7c7 b8b2 c7g3 f5g6 g3f4 b2b3 e1e3 b3e3 f4e3 g8g7 a1d1 a7a5 e3f3');
INSERT INTO evaluations VALUES(49325,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(49325,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(49325,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(49325,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(49325,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(49325,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(49325,36,'cp 81','d7d6 c2c4 d5e7 e5d6 c7d6 b1c3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 d4e5 d6e5 d2d8 f8d8 f3e5 b8c6 e5c6 b7c6 g4f5 d8e8 e1d2 c6b5 f1g2 c8f5 g2a8 e8a8 c4b5 g7b2 a1d1 a8c8 f4h6 c8c2 d2e1 b2c3 e1f1 c2a2 h4h5 c3f6 d1d6 f6e7 d6a6 f5d3 f1g1 a2a6 b5a6');
INSERT INTO evaluations VALUES(48029,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(48029,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(48029,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(48029,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(48029,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(48029,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(48029,36,'cp 107','c2c4 d5e7 e5d6 c7d6 b1c3 g7g6 h2h4 h7h6 c1f4 f8g7 c3b5 e8g8 d1d2 e7f5 g2g4 e6e5 d4e5 d6e5 d2d8 f8d8 f3e5 b8c6 e5c6 b7c6 g4f5 d8e8 e1d2 c6b5 f1g2 c8f5 g2a8 e8a8 c4b5 g7b2 a1d1 a8c8 d2e2 f5e6 d1d2 b2c3 d2d3 e6c4 a2a4 c3a5 h1c1 c4d3 e2d3 c8d8 d3e4 a5b6 f4e3 d8e8 e4d3');
INSERT INTO evaluations VALUES(1464,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(1464,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(1464,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(1464,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(1464,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(1464,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(1464,36,'cp -49','c5d4 f1d3 d4c3 b1c3 g8e7 e1g1 e7g6 c1b2 f8e7 c3a4 b6d8 a1c1 c8d7 a4c5 e8g8 d3b1 a8c8 c5b7 d8b6 b7d6 c8d8 h2h4 c6e5 f3e5 g6e5 d6f5 e6f5 b2e5 e7h4 e5c7 b6b7 c7d8 h4d8 d1d2 g7g6 f1e1 d8b6 b3b4 f8d8 d2f4 d7e6 f4d2 g8g7 b1d3 b7b8');
INSERT INTO evaluations VALUES(1518,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(1518,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(1518,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(1518,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(1518,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(1518,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(1518,36,'cp -42','f1d3 d4c3 b1c3 g8e7 c1b2 e7g6 e1g1 c8d7 c3a4 b6a5 a2a3 b7b5 a4c3 a7a6 c3e2 b5b4 a3a4 f8e7 g2g3 e8g8 h2h4 d5d4 e2d4 c6e5 d3e4 e5f3 d4f3 a8d8 d1e2 a5h5 f1e1 e7f6 b2f6 g7f6 e2a6 f6f5 e4b7 f5f4 a6e2 f4g3 f2g3 d7c8 b7c8 d8c8 a4a5 c8c3');
INSERT INTO evaluations VALUES(49004,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(49004,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(49004,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(49004,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(49004,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(49004,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(49004,36,'cp 6','c8e6 b1c3 f8b4 g1e2 d5c3 b2c3 e6c4 f3b7 b4c5 d3c4 b8d7 e2g3 e8g8 e1g1 a8b8 b7d5 f8e8 g3e4 c5f8 c1g5 c7c6 d5d3 d8c8 a1d1 d7c5 e4c5 f8c5 g5e3 c5e3 d3e3 b8b2 e3a7 b2c2 d1d7 c2a2 a7c7 c8c7 d7c7 a2c2 c7c6 c2c3');
INSERT INTO evaluations VALUES(30010,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(30010,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(30010,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(30010,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(30010,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(30010,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(30010,36,'cp 3','b1c3 f8b4 g1e2 e8g8 e1g1 c7c6 c3e4 f7f6 d3d4 b8d7 c4b3 g8h8 f1d1 b7b5 a2a4 a7a6 e4c3 d7b6 a4b5 a6b5 a1a8 d8a8 d4e5 f6e5 f3h5 b4c3 b2c3 a8e8 h5e8 f8e8 e2g3 d5c3 d1d6 e6b3 c2b3 e8c8');
INSERT INTO evaluations VALUES(30007,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(30007,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(30007,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(30007,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30007,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30007,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30007,36,'cp 6','f8b4 g1e2 e8g8 e1g1 c7c6 c3e4 f7f6 d3d4 b8d7 c4b3 g8h8 f1d1 b7b5 a2a4 a7a6 c1d2 b4d2 a4b5 a6b5 e4d2 a8a1 d1a1 d8b6 h2h3 e5d4 f3e4 d7c5 e4d4 f8d8 b3d5 e6d5 d4f4 c5e6 f4g3 e6d4 g3e3 d4e2 e3e2 b6c5');
INSERT INTO evaluations VALUES(28466,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28466,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(28466,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(28466,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(28466,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(28466,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(28466,36,'cp 19','g1e2 e8g8 e1g1 c7c6 c3e4 f7f6 d3d4 b8d7 c4b3 g8h8 f1d1 b7b5 a2a4 a7a6 c1d2 b4d2 a4b5 a6b5 e4d2 a8a1 d1a1 d8b6 h2h3 e5d4 f3e4 e6g8 b3d5 g8d5 e4d4 f8e8 d4b6 d7b6 e2c3 h7h6 f2f3 b6c4 c3d5 c6d5 d2c4');
INSERT INTO evaluations VALUES(49897,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(49897,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(49897,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(49897,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(49897,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(49897,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(49897,36,'cp 106','d4e5 c7c6 d1e2 e4c5 a2a3 f8e7 g1f3 d7d5 e5d6 d8d6 b2b4 c5e6 b1c3 e8g8 c3e4 d6c7 e1g1 e6f4 e2d2 c8e6 d2f4 c7f4 c1f4 e6c4 f1e1 c4e6 f3d4 b8d7 a1d1 d7f6 d4e6 f7e6 g2g3 f6d5 f4d6 e7d6 e4d6 a7a5 b4a5 a8a5 d6b7 a5a3 b7c5 a3c3 c5e6 f8e8 e6d4 e8e1 d1e1');
INSERT INTO evaluations VALUES(49838,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(49838,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(49838,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(49838,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(49838,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(49838,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(49838,36,'cp 123','c7c6 d1e2 e4c5 a2a3 f8e7 g1f3 e8g8 e1g1 b7b5 c4a2 d7d5 e5d6 e7d6 f1d1 d8e7 b1c3 c8g4 c1g5 e7e2 c3e2 d6c7 g5e7 f8e8 e7c5 e8e2 d1e1 g4f3 g2f3 e2e1 a1e1 b8d7 e1e7 c7h2 g1h2 d7c5 e7f7 g8h8 b2b4 c5a4 f7c7 a4c3 a2b3 g7g6 c7c6 c3b1 c6a6 b1d2');
INSERT INTO evaluations VALUES(30098,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(30098,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30098,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30098,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30098,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30098,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30098,36,'cp 1','f8d6 f4d6 c7d6 c2c4 d5c4 f1c4 b8c6 e1g1 e8g8 d1b3 c6a5 b3b4 a5c4 d2c4 f5d3 f1c1 d3c4 c1c4 d8d7 h2h3 a7a5 b4d2 b7b5 c4c2 b5b4 a1c1 f8c8 f3e1 h7h6 e1d3 d7b7 f2f3 f6d5 e3e4 d5b6 c2c8 a8c8 c1c8 b7c8');
INSERT INTO evaluations VALUES(28454,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(28454,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28454,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28454,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(28454,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(28454,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28454,36,'cp 9','f4d6 c7d6 c2c4 d5c4 f1c4 b8c6 e1g1 e8g8 d1b3 d8d7 f1c1 f8c8 c4b5 c8c7 f3h4 a8c8 h4f5 e6f5 c1c3 a7a6 b5f1 d6d5 a1c1 f5f4 e3f4 c6d4 b3b4 d4c6 b4a3 g7g6 g2g3 c6d4 a3b4 d4c6 b4a4 h7h5 d2f3 f6e4');
INSERT INTO evaluations VALUES(16412,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(16412,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(16412,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(16412,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(16412,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(16412,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(16412,36,'cp -48','f6e4 c3e4 d7d5 c4d3 d5e4 d3e4 f8d6 d2d3 h7h6 g1f3 f7f5 e4c6 b7c6 e1g1 e8g8 d1e1 f8e8 e1c3 e5e4 d3e4 f5e4 f3d4 c8g4 f2f4 e4f3 d4f3 d8e7 c1d2 e7e4 a1e1 e4d5 h2h3 g4f5 d2e3 f5e4 e3d4 e8e7 c3b3 d5b3 c2b3 c6c5 d4f2 a8b8 f3d2 e4d3 e1e7 d3f1 e7e1 f1d3');
INSERT INTO evaluations VALUES(16836,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(16836,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(16836,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(16836,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(16836,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(16836,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(16836,36,'cp -28','c3e4 d7d5 c4d3 d5e4 d3e4 c8e6 g1f3 d8d7 e1g1 f7f6 c2c3 e8c8 d1a4 g7g5 d2d4 g5g4 f3h4 e5d4 c3d4 c8b8 e4c6 d7c6 a4c6 b7c6 f1e1 e6b3 c1e3 d8d5 g2g3 f8d6 h4g2 h8e8 g2f4 d6f4 e3f4 e8e1 a1e1');
INSERT INTO evaluations VALUES(17291,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(17291,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17291,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(17291,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(17291,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(17291,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(17291,36,'cp 21','d2d4 d7d5 c1f4 f6h5 f4d2 h5f6 e2e3 e7e6 f1b5 a7a6 b5c6 b7c6 c3a4 a6a5 e1g1 f6e4 c2c4 c8a6 b2b3 d5c4 d1c2 d8d5 f1c1 c4b3 a2b3 e4d2 f3d2 f8b4 d2f3 e8g8 a4c3 d5f5 c2f5 e6f5 f3e5 c6c5 d4c5 f8e8 e5c6 b4c5 a1a5');
INSERT INTO evaluations VALUES(19297,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(19297,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(19297,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(19297,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(19297,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(19297,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19297,36,'cp 38','f1b5 c6d4 f3d4 c5d4 e1g1 a7a6 b5e2 e7e6 d2d3 f8g7 b1d2 g8e7 d1e1 d7d6 d2f3 c8d7 e1f2 e7c6 c1d2 e8g8 a1e1 d8b6 b2b4 b6c7 e2d1 a8e8 a2a4 b7b5 a4b5 a6b5 h2h4 h7h5 f2g3 e8c8 g3h3 c6e7 g2g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(19218,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(19218,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(19218,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(19218,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(19218,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(19218,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(19218,36,'cp 26','c6d4 f3d4 c5d4 e1g1 a7a6 b5e2 f8g7 d2d3 e7e6 b1d2 g8e7 d2f3 d7d6 c1d2 c8d7 a1c1 d8b6 d1e1 a8c8 d2b4 e8g8 b4a3 f8e8 e1f2 c8c6 c2c4 d4c3 b2c3 b6f2 g1f2 e6e5 f4e5 d6e5 c3c4 b7b5 c4c5 h7h5');
INSERT INTO evaluations VALUES(57191,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(57191,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(57191,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(57191,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(57191,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(57191,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(57191,36,'cp -58','d7d5 d2d3 f8d6 f1g2 g8f6 b1d2 e8g8 c1b2 c7c6 g1f3 f8e8 a2a3 a7a5 e1g1 c8g4 e2e4 b8d7 h2h3 g4h5 d1e1 d8c7 f3h4 a8d8 g1h1 d5e4 d2e4 f6e4 d3e4 d6f8 h4f5 f7f6 b2c3 a5a4 b3b4 b7b5 f2f4 h5f7 a1d1 c6c5');
INSERT INTO evaluations VALUES(56427,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(56427,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(56427,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(56427,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(56427,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(56427,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(56427,36,'cp -66','d2d3 f8d6 f1g2 g8f6 b1d2 e8g8 c1b2 f8e8 e2e4 c7c6 g1f3 a7a5 a2a3 d5e4 d3e4 b7b5 e1g1 h7h6 b3b4 a8a6 b4a5 d8a5 d2c4 b5c4 d1d6 b8d7 d6b4 f6e4 f1e1 c4c3 e1e4 c3b2 b4b2');
INSERT INTO evaluations VALUES(11174,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(11174,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(11174,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(11174,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(11174,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(11174,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(11174,36,'cp 62','f2f3 e7e5 d4b3 c8e6 d1e2 a8c8 e1c1 c6a5 e2d2 h7h5 e3g5 f8e7 g5f6 a5b3 a2b3 e7f6 d2d6 d8d6 d1d6 e8e7 d6d1 h5h4 c1b1 h8d8 f1c4 e6c4 b3c4 d8d1 h1d1 c8c4 c3d5 e7e6 c2c3 c4c6 b2b3 b7b5 b1c2 f6g5 d1a1 g5f4');
INSERT INTO evaluations VALUES(11176,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(11176,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(11176,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(11176,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(11176,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11176,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(11176,36,'cp 55','e7e5 d4b3 c8e6 d1e2 a8c8 e1c1 c6a5 e2d2 h7h5 b3a5 d8a5 c1b1 b7b5 a2a3 f8e7 f1e2 c8b8 c3a2 a5d2 e3d2 d6d5 a2b4 e7b4 d2b4 b8d8 b4c3 d5d4 c3b4 d8d7 a3a4 e6c4 e2d3 h8h6 a4b5 a6b5 c2c3 h5h4 d3c4 b5c4 c3d4 d7d4 d1d4 e5d4 h1d1 h4h3 g2g4 d4d3');
INSERT INTO evaluations VALUES(11228,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(11228,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(11228,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(11228,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(11228,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(11228,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(11228,36,'cp 61','d4b3 c8e6 d1e2 a8c8 e1c1 c6a5 e2d2 h7h5 c1b1 f8e7 f1e2 b7b5 b3a5 d8a5 a2a3 c8b8 c3a2 a5d2 e3d2 d6d5 a2b4 e7b4 d2b4 b8d8 b4c3 d5d4 c3a5 d8d7 c2c3 e6b3 d1d2 d4c3 a5c3 e8g8 d2d7 f6d7 h1c1 f8c8 e2d1 b3e6 c1c2 f7f6 c3b4 d7b6 c2c8 b6c8 b1c2 g7g5');
INSERT INTO evaluations VALUES(11232,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11232,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(11232,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(11232,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(11232,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(11232,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(11232,36,'cp 53','c8e6 d1e2 a8c8 e1c1 c6a5 e2d2 h7h5 c1b1 f8e7 f1e2 b7b5 b3a5 d8a5 a2a3 c8b8 c3a2 a5d2 e3d2 d6d5 a2b4 e7b4 d2b4 b8d8 b4c3 d5d4 c3a5 d8d7 c2c3 e6b3 d1d2 d4c3 a5c3 e8g8 d2d7 f6d7 h1c1 f8c8 c3d2 d7b6 c1c8 b6c8 f3f4 c8d6 f4e5 d6e4 d2b4 g7g6 e2f3 f7f5 e5f6 b3d5 b1c2 e4f6 c2d3');
INSERT INTO evaluations VALUES(24255,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(24255,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(24255,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24255,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(24255,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(24255,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(24255,36,'cp 64','d1e2 a8c8 e1c1 c6a5 e2d2 b7b5 c1b1 a5c4 f1c4 b5c4 b3c1 d8c7 g2g4 c8b8 g4g5 f6d7 f3f4 e5f4 e3f4 d7e5 c1e2 c7b7 d2c1 f8e7 h2h3 e8g8 b1a1 f7f5 e2d4 e6c8 d4f5 c8f5 e4f5 f8f5 c3d5 e7d8 h1f1');
INSERT INTO evaluations VALUES(41324,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41324,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41324,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(41324,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41324,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(41324,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41324,36,'cp 0','c7c5 f1e2 d8b6 d1c1 c8e6 c2c3 e8g8 e1g1 b8c6 f1e1 c5d4 e3d4 e6d5 b1d2 a8c8 f4e3 b6c7 c3c4 d5f3 d2f3 e7e5 d4e5 d6e5 c1c2 c6e7 e2d3 e7c6');
INSERT INTO evaluations VALUES(48654,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(48654,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(48654,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48654,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48654,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(48654,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(48654,36,'cp -28','f8d6 c3b5 d6e7 f1d3 e8g8 g1e2 f8e8 e1g1 b7b6 b5c3 e7d6 e2g3 c8a6 d3a6 b8a6 c1d2 c7c6 d1a4 d8c8 c3e2 c8b7 a4c2 g7g6 a1c1 a8c8 f1d1 h7h5 h2h3 c6c5 d4c5 b6c5 a2a3 h5h4 d2c3');
INSERT INTO evaluations VALUES(41698,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(41698,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(41698,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(41698,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(41698,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(41698,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(41698,36,'cp -11','f1d3 e8g8 g1e2 f8e8 e1g1 b7b6 e2g3 c8b7 b2b3 b8d7 c1b2 a7a6 a1c1 f6e4 g3f5 d6f8 d1c2 a8c8 f5g3 d7f6 c3e2 g7g6 f1e1 c7c5 d3e4 f6e4 d4c5 b6c5 g3e4 e8e4 e2g3 e4b4 c2d2 d5d4 e3d4 c5d4 c1c8 b7c8');
INSERT INTO evaluations VALUES(49565,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(49565,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49565,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(49565,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(49565,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(49565,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(49565,36,'cp 19','g1f3 b8c6 b1c3 f8e7 g2g3 d7d5 c3d5 f6d5 c4d5 d8d5 f1g2 e8g8 e1g1 d5d6 a2a3 a7a5 d1c2 c8e6 c1e3 f8d8 a1c1 d6d7 f3d2 a8a6 d2e4 a5a4 e4c5 e7c5 c2c5 f7f6 c5c3 h7h6 f2f4 e5f4 g3f4 a6a8 g2c6 b7c6 c3c6');
INSERT INTO evaluations VALUES(49563,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(49563,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(49563,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(49563,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49563,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49563,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49563,36,'cp 12','b8c6 b1c3 d7d5 c4d5 f6d5 g2g3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 c1e3 e8g8 d1a4 f8d8 a1c1 f7f6 f1e1 c6b4 a4d7 d8d7 a2a3 b4d5 e3c5 e7c5 c1c5 c7c6 e1c1 g8f7 f3d2 a7a5 g2e4 a5a4 e4h7 g7g6 d2c4 a8a6');
INSERT INTO evaluations VALUES(50156,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50156,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(50156,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(50156,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50156,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50156,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(50156,36,'cp 20','g5f6 e7f6 c2c4 d7d5 c4d5 d8d5 b1c3 d5d8 h2h4 h7h5 g2g3 b8d7 e2e3 f8d6 f1g2 e8g8 a2a3 c7c6 c3e4 d7b6 g1e2 c8g4 e1g1 f8e8 e4d6 d8d6 d1c2 g4e2 c2e2 g8g7 b2b4 a7a6 e2b2 f6f5 a1c1 b6d5 c1c5');
INSERT INTO evaluations VALUES(50141,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50141,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50141,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(50141,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50141,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(50141,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50141,36,'cp 15','e7f6 c2c4 d7d5 c4d5 d8d5 b1c3 d5d8 h2h4 h7h5 g2g3 c7c6 f1h3 b8d7 e2e3 f6f5 h3g2 d7f6 g1h3 f8h6 d1b3 d8b6 e1g1 e8g8 b3b6 a7b6 a2a4 f8e8 b2b4 f6e4 c3e4 f5e4 h3g5 c8f5 b4b5 h6g5 h4g5 c6b5 a4b5');
INSERT INTO evaluations VALUES(49217,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(49217,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(49217,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(49217,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(49217,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49217,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(49217,36,'cp 17','c2c4 d7d5 c4d5 d8d5 b1c3 d5d8 h2h4 h7h5 g2g3 f8d6 f1h3 b8d7 g1f3 f6f5 e1g1 e8g8 e2e3 c7c6 d1b3 d7b6 a2a4 d6e7 f3e5 c8e6 b3c2 f8e8 h3g2 e7d6 e5d3 a7a5 a1c1 d8e7 f1d1 a8d8 b2b3');
INSERT INTO evaluations VALUES(43443,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(43443,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(43443,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(43443,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(43443,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(43443,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(43443,36,'cp 156','d4f5 b8c6 h2h4 c8e6 d1d2 h7h6 g5f6 d8f6 c3d5 f6d8 f5e3 h6h5 g2g3 g7g6 e1c1 f8h6 f1h3 e6d5 d2d5 e8g8 c1b1 h6e3 f2e3 d8a5 d5b3 a8d8 b3b7 a5c5 b7c7 c6a5 c7c5 d6c5 b2b3 a5b7 h3d7 c5c4 d7c6');
INSERT INTO evaluations VALUES(43481,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(43481,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(43481,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(43481,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(43481,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(43481,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(43481,36,'cp 123','h7h6 g5f6 d8f6 d1d2 g7g6 f5e3 h6h5 e1c1 f6d8 f1c4 b8c6 g2g3 f8h6 f2f4 e8g8 d2f2 b7b5 c4d5 c8d7 c1b1 a8c8 d5c6 d7c6 c3d5 e5f4 g3f4 f8e8 f4f5 h6e3 f2e3 c6d5 d1d5 d8e7 f5g6 f7g6 e3h6 e7g7 h6g5 e8e4');
INSERT INTO evaluations VALUES(50398,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(50398,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(50398,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(50398,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(50398,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50398,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(50398,36,'cp 15','f1c4 e7e6 d2d4 d7d6 c5d6 c7d6 g1f3 f8e7 d1e2 e8g8 e1g1 b8d7 c4d5 e6d5 b1c3 d7b6 c1f4 f8e8 e2d3 c8e6 b2b3 a7a6 h2h3 a8c8 a1c1 d8d7 c3e2 h7h6 e2g3 d6e5 c1c8 e8c8 d4e5 d7b5 d3b5 a6b5 f3d4 e6d7 f4e3 b5b4');
INSERT INTO evaluations VALUES(50395,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50395,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(50395,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(50395,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50395,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50395,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(50395,36,'cp 18','e7e6 d2d4 d7d6 c5d6 c7d6 g1f3 f8e7 e1g1 e8g8 d1e2 b8d7 c4d5 e6d5 b1c3 d7b6 c1f4 a7a6 f1e1 c8e6 e5d6 e7d6 f4d6 d8d6 f3e5 a8e8 e5d3 b6c4 a1c1 h7h6 e2f3 d6b6 b2b3 c4d6 f3f4 d6f5 d3c5 b6b4 c3e2 e8c8 c5d3 b4a3 c1c7 c8c7 f4c7');
INSERT INTO evaluations VALUES(49310,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(49310,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49310,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(49310,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(49310,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49310,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(49310,36,'cp 12','d2d4 d7d6 c5d6 c7d6 g1f3 f8e7 d1e2 e8g8 e1g1 d6e5 d4e5 c8d7 c4d5 e6d5 b1c3 d7e6 c1e3 d8d7 f3d4 b8c6 f2f4 c6d4 e3d4 f7f6 e5f6 e7f6 e2f2 f6d4 f2d4 a8c8 b2b3 d7d6 a1e1 d6c5 d4c5 c8c5 e1e6 c5c3 e6d6 c3c2 d6d5 c2a2 d5d7 a7a6');
INSERT INTO evaluations VALUES(49311,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(49311,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(49311,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(49311,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(49311,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(49311,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(49311,36,'cp 26','b7b6 c4d5 e6d5 b1c3 b6c5 c3d5 c7c6 d5c3 d7d5 e5d6 f8d6 d4c5 d8e7 g1e2 d6c5 c1f4 e8g8 e1g1 b8a6 e2g3 a6b4 d1f3 f8d8 f1d1 b4d5 c3d5 c6d5 a1c1 c5b6 d1d5 c8e6 d5d8 a8d8 b2b3 e7f6 g3e4 f6g6 e4d6 h7h6 f3g3 g6g3 h2g3 e6d5 f4e5 d5e6');
INSERT INTO evaluations VALUES(42236,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(42236,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(42236,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(42236,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(42236,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(42236,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(42236,36,'cp -25','e8g8 f1e2 c7c6 e1g1 f8e8 f1e1 b4d6 d2g5 c8e6 e2d3 b8d7 h2h3 d8c7 c3e2 h7h6 g5e3 d7f8 b2b3 f6e4 c2c4 c7b8 d1c1 d6b4 e1f1 e6f5 e3f4 b8d8 a2a3 b4d6 c4c5 d6f4 e2f4 f8g6 f1e1 g6h4 f3h4 d8h4 d3e4 e8e4 e1e4');
INSERT INTO evaluations VALUES(37110,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(37110,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(37110,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(37110,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(37110,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(37110,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(37110,36,'cp -21','f1e2 c8f5 e1g1 c7c6 f3h4 f5e6 f1e1 f8e8 e2d3 b8d7 h2h3 a7a5 h4f3 d8c7 d1c1 b4d6 c3e2 f6e4 d2f4 e6f5 f4d6 e4d6 d3f5 d6f5 c2c4 d7f6 c4d5 f6d5 c1c2 g7g6 e2c3 e8e1 a1e1 a8d8 c3e4 c7b6 c2d2 g8g7 g2g4');
INSERT INTO evaluations VALUES(37111,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(37111,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(37111,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(37111,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(37111,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(37111,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(37111,36,'cp -21','c8f5 e1g1 f8e8 f1e1 c7c6 f3h4 f5e6 e2d3 b8d7 h2h3 a7a5 h4f3 b4d6 d1c1 d8c7 c3e2 f6e4 d2f4 e6f5 f4d6 c7d6 c2c4 d5c4 c1c4 f5g6 a2a3 a5a4 c4c2 d7f6 a1d1 f6d5 f3e5 e4f6 c2d2 f6d7 d3g6 h7g6 e2f4 d7e5 f4d5 d6d5 d4e5');
INSERT INTO evaluations VALUES(38430,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(38430,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(38430,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(38430,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38430,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38430,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38430,36,'cp 35','c7c5 d4d5 g8f6 b1c3 e8g8 f1d3 b7b5 e4e5 d6e5 f4e5 f6g4 d5d6 c8b7 c1g5 f7f6 e5f6 e7f6 g5f4 c5c4 d3e2 b8d7 c3b5 d8a5 b5c3 f6f5 e1g1 a5c5 g1h1 h7h6 f3h4 g4e5 d1d2 g8h7 e2f3 e5f3 h4f3 g6g5 f4e3 c5c6 a1e1 a8e8 e3d4 e8e1 f1e1 c6d6 d4g7');
INSERT INTO evaluations VALUES(37974,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(37974,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(37974,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(37974,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(37974,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(37974,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(37974,36,'cp 26','d4d5 g8f6 b1c3 e8g8 f1d3 e7e6 d5e6 c8e6 e1g1 b8c6 f4f5 c5c4 f5e6 d8b6 g1h1 c4d3 e6f7 f8f7 d1d3 f6g4 c3d5 g4f2 f1f2 b6f2 c1e3 f7f3 e3f2 f3d3 c2d3 a8f8 f2e3 g7b2 a1b1 b2d4 e3d4 c6d4 h2h4 b7b6 h1h2 f8f2 b1c1 f2c2 c1c2 d4c2 h2g3 c2e1 d3d4 e1c2');
INSERT INTO evaluations VALUES(53026,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(53026,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(53026,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(53026,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(53026,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(53026,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(53026,36,'cp 50','g1f3 g8f6 d2d4 c5d4 f3d4 b8c6 e2e4 c6d4 d1d4 g7g6 f1e2 f8g7 e1g1 e8g8 d4e3 a7a5 f1d1 c8e6 c4c5 a8c8 c5d6 e7d6 h2h3 f8e8 e2b5 e6d7 d1d6 d8e7 b5d7 f6d7 d6d1 d7c5 f2f3 g7c3 e3c3 c5e4 c3e1 e4d6 e1e7 e8e7 c1g5 e7d7 g5f6 a5a4');
INSERT INTO evaluations VALUES(53024,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(53024,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(53024,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(53024,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(53024,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(53024,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(53024,36,'cp 58','g7g6 d2d4 c5d4 f3d4 b8c6 e2e4 g8f6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 c8d7 e1g1 a7a6 a2a4 a6a5 a1d1 d7c6 d3c2 f6d7 b2b3 b7b6 e3f4 f8e8 f2f3 d8c7 f4e3 d7c5 d1d2 e8d8 f1d1 d8d7 e2f1 a8c8 g2g3 g7e5');
INSERT INTO evaluations VALUES(53264,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(53264,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(53264,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(53264,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(53264,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(53264,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(53264,36,'cp -11','e7e6 g1f3 g8f6 b1d2 b8d7 f1d3 b7b6 d1e2 c8b7 e1g1 a7a6 b2b3 f8d6 c1b2 e8g8 c3c4 c5d4 e3d4 d8e7 f3e5 a6a5 a1c1 f8d8 f2f4 a5a4 e2e3 a4b3 a2b3 a8a2 c1c2 d6b4 b2c3 a2c2 c3b4 e7b4 d3c2');
INSERT INTO evaluations VALUES(54206,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(54206,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54206,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(54206,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54206,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54206,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(54206,36,'cp -5','g1f3 g8f6 f1e2 b8d7 e1g1 f8e7 b2b3 b7b6 c3c4 c5d4 f3d4 c8b7 b1d2 a8c8 c4d5 f6d5 c1b2 e8g8 a1c1 c8c1 d1c1 d8a8 c1a1 d5f6 d4f3 f8c8 f1c1 h7h6 b2d4 e7f8 a2a3 f6e4 d2e4 c8c1 a1c1 b7e4 h2h3 e4d5 c1c2 a8b7 b3b4');
INSERT INTO evaluations VALUES(54207,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(54207,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(54207,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(54207,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(54207,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(54207,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(54207,36,'cp -8','g8f6 f1e2 b8d7 e1g1 f8e7 b2b3 b7b6 c1b2 c8b7 b1d2 a8c8 d4c5 c8c5 a1c1 e8g8 c3c4 d5c4 e2c4 c5c7 c4e2 c7c1 d1c1 d8a8 c1b1 f8c8 f1c1 h7h6 h2h3 a7a6 b2d4 c8c1 b1c1 b7d5 a2a3 b6b5 c1b2 e7d6 b3b4 a8c6 e2d3');
INSERT INTO evaluations VALUES(46234,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(46234,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(46234,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(46234,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(46234,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(46234,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(46234,36,'cp -7','f1e2 b8d7 e1g1 f8e7 b1d2 e8g8 b2b3 b7b6 c1b2 c8b7 c3c4 a8c8 c4d5 f6d5 a1c1 c5d4 f3d4 c8c1 d1c1 d8a8 c1a1 d5f6 d4f3 f8c8 f1c1 b7d5 a2a3 c8c1 a1c1 a8b7 h2h3 e7f8 b2d4 h7h6 b3b4 a7a5 c1b1 a5b4 a3b4');
INSERT INTO evaluations VALUES(46233,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46233,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(46233,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(46233,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(46233,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(46233,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(46233,36,'cp -5','b7b6 e1g1 f8e7 c3c4 e8g8 c4d5 f6d5 d4c5 e7c5 a2a3 c8b7 b2b4 c5e7 c1b2 e7f6 d1b3 a7a5 b1d2 f6b2 b3b2 d8f6 b2f6 g7f6 b4a5 a8a5 f1c1 b7a6 f3d4 a6e2 d4e2 b8d7 d2c4 a5a8 e3e4 d5e7 f2f4 e7g6 c4d6 d7c5');
INSERT INTO evaluations VALUES(16147,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(16147,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16147,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(16147,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(16147,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(16147,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(16147,36,'cp -29','f6e4 c3e4 d7d5 c4d3 d5e4 d3e4 f8d6 d2d4 e5d4 e4c6 b7c6 d1d4 e8g8 c1e3 c6c5 d4c3 d8d7 e1c1 d7f5 f3d2 f8e8 g2g4 f5e6 d2c4 d6e7 e3c5 e7c5 h1e1 e6c6 c4a5 c6b5 e1e8 b5e8 c3c5 c8g4 d1g1 e8e4 c5e3 e4e3 f2e3 h6h5 b2b4 f7f6 a5c6 g8f7 a3a4 a8e8 c6a7');
INSERT INTO evaluations VALUES(18356,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(18356,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(18356,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(18356,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(18356,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(18356,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(18356,36,'cp 188','d2d4 e7e6 d4d5 c6e7 b1c3 c8b7 d1e2 g8f6 c1d2 g7g6 d5e6 d7e6 e1c1 f8g7 h2h4 f6d7 h4h5 h7h6 d2f4 g6g5 f3e5 g5f4 e5d7 e7d5 e4d5 d8d7 d5e6 d7e6 e2b5 c7c6 b5b4 e6e7 b4f4 e8g8 f1c4 a8e8 d1d6 b6b5 c4d3 e7g5 f4g5 h6g5 h5h6 g7e5 d6d7 b7c8 d7a7 b5b4');
INSERT INTO evaluations VALUES(18352,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(18352,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(18352,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(18352,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(18352,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(18352,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(18352,36,'cp 175','e7e6 d4d5 c6e7 b1c3 a7a6 g2g3 b6b5 f1g2 b5b4 c3a4 g8f6 d5e6 d7e6 d1d8 e8d8 f3e5 d8e8 c1e3 h7h5 e1c1 f6g4 e3d4 c8b7 a4c5 b7c6 c5d3 g4e5 d4e5 e7g6 e5c7 h5h4 f2f4 f7f6 c7b6');
INSERT INTO evaluations VALUES(18311,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(18311,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(18311,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(18311,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(18311,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(18311,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(18311,36,'cp 182','d4d5 c6e7 b1c3 a7a6 g2g3 b6b5 f1g2 g8f6 c1g5 a8b8 f3e5 e6d5 e4d5 h7h6 g5f6 g7f6 e5d3 d7d6 c3e2 f8g7 e1g1 f6f5 e2f4 e8g8 h2h4 e7g6 f4h5 g7h8 d1d2 g8h7 d3b4 g6e5 a1e1 a6a5 e1e5 a5b4 e5e3 b8a8 f1e1 a8a2 d2b4');
INSERT INTO evaluations VALUES(41270,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41270,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41270,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41270,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(41270,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(41270,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(41270,36,'cp 47','g1f3 e8g8 f1e2 b8a6 e1g1 c8f5 a2a3 f6e4 c3e4 f5e4 c1d2 c7c5 d2c3 a6c7 a1c1 c5d4 c3d4 e7e5 d4c3 f8e8 f3d2 e4c6 b2b4 b7b6 d1c2 d8e7 f1d1 a7a5 b4b5 c6b7 e2f3 b7f3 d2f3 e8c8 f3d2');
INSERT INTO evaluations VALUES(50151,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(50151,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50151,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50151,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(50151,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50151,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50151,36,'cp 0','f6e4 g5f4 b8c6 g1f3 f8g7 b1d2 e4f6 d4d5 c6b8 d2b1 f6e4 b1d2');
INSERT INTO evaluations VALUES(50276,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(50276,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(50276,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(50276,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(50276,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(50276,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50276,36,'cp 0','g5f4 b8c6 g1f3 f8g7 b1d2 e4f6 d4d5 c6b8 d2b1 f6e4 b1d2');
INSERT INTO evaluations VALUES(52469,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(52469,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52469,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(52469,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52469,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(52469,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(52469,36,'cp -27','d7d5 e2e4 d5d4 c3b1 g8f6 f1b5 c8d7 b5d7 f6d7 d2d3 b8c6 a2a4 e7e5 e1g1 a7a6 c1d2 f8e7 b1a3 e8g8 d1e2 a8b8 g2g3 h7h6 c2c3 f8e8 h2h4 d8c7 f1c1 b7b5 a4b5 a6b5 c3c4 b5b4 a3b5 c7b7 b2b3 b8a8 f3h2 a8a1 c1a1 e8a8 a1a8 b7a8 h4h5 a8a1 e2f1');
INSERT INTO evaluations VALUES(53310,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(53310,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(53310,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(53310,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(53310,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(53310,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(53310,36,'cp -12','e2e4 d5d4 f1b5 c8d7 b5d7 d8d7 c3e2 b8c6 e1g1 e7e6 d2d3 f8d6 c1f4 e6e5 f4d2 g8e7 c2c3 d4c3 b2c3 e8g8 d2e3 f8d8 d1b1 b7b6 f1d1 h7h6 d3d4 e5d4 c3d4 c6d4 e2d4 c5d4 d1d4 e7c6 d4d5 a8c8 d5d1');
INSERT INTO evaluations VALUES(53281,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(53281,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(53281,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(53281,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(53281,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(53281,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(53281,36,'cp -23','d5d4 f1b5 c8d7 b5d7 d8d7 c3e2 b8c6 e1g1 e7e6 d2d3 f8d6 c1f4 e6e5 f4d2 b7b6 a2a3 g8e7 e2g3 f7f6 d1e2 a7a5 f3h4 e8c8 h4f5 e7f5 g3f5 d6c7 f2f4 c8b7 f4e5 c7e5 d2f4 a5a4 e2f3 g7g6 f4e5 c6e5');
INSERT INTO evaluations VALUES(53212,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(53212,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(53212,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(53212,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(53212,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(53212,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(53212,36,'cp -15','f1b5 c8d7 b5d7 d8d7 c3e2 b8c6 d2d3 e7e6 e1g1 f8d6 a1b1 g8e7 c2c3 d4c3 b2c3 e8g8 d3d4 f8d8 c1e3 b7b6 e4e5 d6c7 d1c2 c5d4 c3d4 h7h6 f1c1 d8c8 c2e4 c7d8 e3d2 e7f5 h2h4 d8e7 h4h5 c6a5 c1c8 a8c8 d2a5 b6a5 e4d3 d7a4');
INSERT INTO evaluations VALUES(53812,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(53812,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(53812,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(53812,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53812,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53812,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(53812,36,'cp 0','g8f6 d2d3 f8e7 e1g1 c8g4 h2h3 g4h5 f1e1 b8d7 b1d2 e8g8 d2f1 d6d5 e4d5 c6d5 c4b3 a7a5 a2a4 e5e4 g2g4 h5g6 d3e4 d5e4 f3h4 d7c5 d1d8 a8d8 h4g6 h7g6 b3c4 f6d7 c4d5 d7f6');
INSERT INTO evaluations VALUES(45864,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45864,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45864,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45864,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45864,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45864,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45864,36,'cp 0','d2d3 f8e7 e1g1 c8g4 f1e1 e8g8 b1d2 b8d7 h2h3 g4h5 d2f1 d6d5 e4d5 c6d5 c4b3 a7a5 a2a4 e5e4 g2g4 h5g6 d3e4 d5e4 f3h4 d7c5 d1d8 a8d8 h4g6 h7g6 b3c4 f6d7 c4d5 d7f6');
INSERT INTO evaluations VALUES(9834,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(9834,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(9834,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(9834,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(9834,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(9834,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(9834,36,'cp 75','c3d5 b4e7 f1e1 d7d6 c2c3 e8g8 a2a4 f6d5 e4d5 c6b8 d2d4 e5d4 f3d4 b8d7 a4a5 d7e5 c4f1 f8e8 b2b4 e7f6 h2h3 a7a6 c1e3 e5g6 c3c4 c8d7 d1d2 f6g5 a1a3 d8f6 e3g5 f6g5 d2g5 h6g5 a3e3 e8e3 e1e3 g6e5 c4c5');
INSERT INTO evaluations VALUES(9827,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(9827,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(9827,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(9827,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(9827,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(9827,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(9827,36,'cp 88','e8g8 d5b4 c6b4 f3e5 f6e4 c2c3 b4c6 d2d4 a7a5 f2f3 e4f6 e5g4 d7d6 f1e1 f6h5 d1c2 g8h8 g4f2 c6e7 c4d3 c8d7 c1d2 b7b5 b2b3 a5a4 c3c4 a4b3 a2b3 a8a1 e1a1 b5c4 b3c4');
INSERT INTO evaluations VALUES(15224,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(15224,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(15224,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(15224,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(15224,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(15224,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(15224,36,'cp 82','d1d8 e8d8 c1g5 d8e8 g5f6 g7f6 e1c1 f8d6 f1b5 a7a6 b5c6 b7c6 f3h4 d6c5 d1d2 c8e6 c3a4 c5f8 d2d3 c6c5 d3f3 h7h5 h4f5 c5c4 a2a3 h5h4 h1d1 h8g8 g2g3 h4g3 f2g3 c7c6 h2h3 g8h8 g3g4 a8b8 a4c3 b8d8 d1d8 e8d8 c1d2 d8d7 d2e2');
INSERT INTO evaluations VALUES(2251,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(2251,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(2251,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(2251,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(2251,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(2251,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(2251,36,'cp 82','e8d8 c1g5 d8e8 g5f6 g7f6 e1c1 f8d6 f1b5 a7a6 b5c6 b7c6 f3h4 d6c5 f2f3 c8e6 g2g3 h8g8 c3a4 c5e3 c1b1 a8d8 d1d8 e8d8 h1e1 e3f2 e1e2 f2g1 c2c3 a6a5 b2b3 g8g5 b1c1 d8c8 c1d2 h7h5 a4b2 c8b7 d2c2 b7a6 b2c4 g1c5 c4e3');
INSERT INTO evaluations VALUES(2281,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(2281,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(2281,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(2281,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(2281,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(2281,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(2281,36,'cp 89','c1g5 d8e8 g5f6 g7f6 e1c1 f8d6 f1b5 a7a6 b5c6 b7c6 f3h4 d6c5 f2f3 c8e6 g2g3 h7h5 c3a4 c5e3 c1b1 a8d8 d1d8 e8d8 h1e1 e3f2 e1e2 f2g1 b2b3 h8g8 c2c3 d8c8 e2g2 g1a7 b1c2 e6h3 g2e2 g8g5 c2d2 g5g8 e2e1 c8b7 h4f5 h3f5 e4f5 a6a5');
INSERT INTO evaluations VALUES(2278,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(2278,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(2278,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(2278,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(2278,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(2278,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(2278,36,'cp 91','d8e8 e1c1 f8d6 g5f6 g7f6 f1b5 a7a6 b5c6 b7c6 f3h4 d6c5 f2f3 c8e6 g2g3 h7h5 c3a4 c5e3 c1b1 a8d8 d1d8 e8d8 h1e1 e3h6 b2b3 d8c8 a4b2 a6a5 e1d1 h6e3 c2c3 h8g8 b1c2 e3c5 d1e1 c5f2 e1f1 f2c5 h4f5 e6f5 e4f5 g8g5 b2c4');
INSERT INTO evaluations VALUES(42383,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(42383,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(42383,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(42383,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(42383,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(42383,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(42383,36,'cp 37','e8g8 e1g1 b8c6 f1e1 h7h6 c3d5 f6d5 e4d5 c6d4 d2d3 d8f6 f3d4 e5d4 c1d2 b7b6 d1h5 c8b7 e1e4 c7c6 a1e1 c6d5 c4d5 b7d5 h5d5 a8c8 c2c3 d4c3 d2c3 f6g6 d3d4 g8h7 e4e2 c8c7 d5f3 f8c8 g2g3 f7f5 h2h4 f5f4 h4h5 g6f5 e2e8 c8e8');
INSERT INTO evaluations VALUES(37153,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(37153,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(37153,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(37153,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(37153,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(37153,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(37153,36,'cp 47','e1g1 b8c6 f1e1 h7h6 c3d5 f6d5 e4d5 c6d4 d2d3 d8f6 f3d4 e5d4 c1d2 b7b6 d1h5 c7c6 e1e4 c8b7 a1e1 c6d5 c4d5 b7d5 h5d5 a8c8 c2c3 d4c3 d2c3 f6g6 e4c4 c8e8 e1e8 f8e8 d5f3 f7f5 g2g3 f5f4 c4e4 e8e4 d3e4 f4g3 h2g3');
INSERT INTO evaluations VALUES(37152,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(37152,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(37152,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(37152,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(37152,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(37152,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(37152,36,'cp 32','b8c6 d2d3 h7h6 c1e3 f8e8 a2a3 d6f8 h2h3 d7d6 f3h2 c8e6 c4e6 e8e6 h2g4 f6g4 d1g4 e6g6 g4f3 f8e7 c3d5 e7f6 c2c4 d8d7 a1d1 c6e7 d5c3 f6g5 d3d4 g5e3 f2e3 g6f6 f3h5 f6f1 d1f1 e7g6 h5f5 d7f5 e4f5 e5d4 e3d4');
INSERT INTO evaluations VALUES(4518,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(4518,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(4518,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(4518,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(4518,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(4518,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(4518,36,'cp 30','d2d3 h7h6 c1e3 f8e8 a2a3 d6f8 h2h3 d7d6 f3h2 c8e6 c4e6 f7e6 h2g4 f8e7 b2b4 a7a6 c3b1 f6h7 b1d2 e7g5 c2c3 e8f8 d1e2 g5e3 f2e3 d8e7 d2f3 h7g5 g4h2 g5f3 f1f3 f8f3 h2f3 a8f8 a1f1');
INSERT INTO evaluations VALUES(4519,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4519,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(4519,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4519,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(4519,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(4519,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(4519,36,'cp 29','h7h6 c1e3 f8e8 a2a3 d6f8 h2h3 d7d6 f3h2 c8e6 c4e6 f7e6 h2g4 f8e7 b2b4 a7a6 c3b1 f6h7 b1d2 e7g5 c2c3 e8f8 d1b3 d8d7 a1e1 f8f7 g4h2 g8h8 h2f3 g5e3 f2e3 a8f8 c3c4 h7f6 d3d4 e5d4 e3d4 f6h5 d4d5');
INSERT INTO evaluations VALUES(4234,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(4234,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(4234,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(4234,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(4234,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(4234,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(4234,36,'cp 19','h2h3 f8e8 a2a3 d6f8 f3h2 d7d6 c1e3 c8e6 c4e6 f7e6 h2g4 f8e7 b2b4 a7a6 c3b1 f6h7 b1d2 e7g5 d2f3 g5e3 f2e3 e8f8 c2c3 h7g5 g4h2 g5f3 f1f3 d8e7 a3a4 f8f3 h2f3 a8f8 d1b3 c6d8 b4b5 d8f7 c3c4');
INSERT INTO evaluations VALUES(14203,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(14203,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(14203,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14203,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(14203,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(14203,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(14203,36,'cp 5','d5c4 g1f3 e7e5 f3e5 f8b4 b1d2 f6d5 a2a3 b4d2 d1d2 d5f4 e3f4 c6d4 f1c4 c8e6 e1c1 c7c5 c4e6 f7e6 c1b1 b7b6 d2c3 d8d5 g2g3 e8g8 h2h4 a8d8 h4h5 d5e4 c3d3 e4d3 d1d3 d8d5 b2b4 d4e2 d3e3 d5d2 b4c5 b6c5');
INSERT INTO evaluations VALUES(14375,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(14375,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(14375,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(14375,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(14375,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(14375,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(14375,36,'cp 13','g1f3 e7e5 f4e5 f8b4 f3d2 b7b5 a2a3 b4d2 b1d2 c6e5 d4e5 f6g4 a3a4 g4e5 a4b5 e5d3 f1d3 c4d3 e1g1 e8g8 d2b3 c8b7 f2f3 f8e8 b3c5 d8d5 c5b7 d5b7 d1d3 a8d8 d3c3 b7b5 a1a7 b5b6 f1a1 b6e3 c3e3 e8e3 a7c7 e3b3 c7c2 d8b8 a1a7 b3b2 c2b2 b8b2');
INSERT INTO evaluations VALUES(57244,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(57244,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(57244,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(57244,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(57244,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(57244,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(57244,36,'cp -19','d7d5 d2d4 e5d4 e3d4 g8f6 g1f3 f8d6 c2c4 c7c6 c4d5 f6d5 b1c3 e8g8 c3d5 c6d5 f1d3 c8g4 e1g1 b8c6 h2h3 g4h5 c1e3 h7h6 f1e1 d8f6 d3e2 f8d8 d1b3 a8b8 a1c1 h5g6 e2d3 c6a5 b3b5 g6d3 b5d3 a5c4');
INSERT INTO evaluations VALUES(56498,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56498,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(56498,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(56498,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(56498,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(56498,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(56498,36,'cp -9','d2d4 e5d4 e3d4 g8f6 f1d3 f8d6 g1e2 e8g8 e1g1 f8e8 h2h3 b8d7 c1f4 d7f8 d1d2 d6f4 e2f4 d8d6 b1c3 c8d7 a1e1 a7a6 h3h4 a8d8 c3e2 f8g6 f4g6 h7g6 d2b4 d6b4 a3b4 f6e4 e1c1 e8e7 e2f4 c7c6 f1e1 d8e8 c2c3 e4f6 e1e7');
INSERT INTO evaluations VALUES(56371,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(56371,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(56371,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(56371,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(56371,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(56371,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(56371,36,'cp -14','e5d4 e3d4 g8f6 f1d3 f8d6 g1e2 e8g8 e1g1 f8e8 b1c3 c7c6 c1f4 c8g4 d1d2 g4h5 f4d6 d8d6 e2f4 h5g6 f1e1 b8d7 e1e8 a8e8 f4g6 h7g6 a1e1 e8a8 g2g3 b7b5 c3d1 d7b6 d2b4 d6b4 a3b4 a7a5 b4a5 a8a5 d1c3 a5a8 g1g2 b6c4');
INSERT INTO evaluations VALUES(4860,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(4860,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(4860,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(4860,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(4860,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(4860,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(4860,36,'cp 137','d2d4 e5d4 c3d4 c5a7 e1g1 d7d6 h2h3 b7b5 c4b3 c6a5 b3c2 g8e7 b1c3 e8g8 c3d5 e7g6 b2b3 c7c5 a2a4 b5a4 a1a4 c8d7 a4a1 d7b5 f1e1 a5c6 c1g5 f7f6 g5e3 g8h8 d1d2 a6a5 d5c3 b5a6 d4d5 c6e5');
INSERT INTO evaluations VALUES(4870,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(4870,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(4870,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(4870,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(4870,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(4870,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(4870,36,'cp 157','e5d4 c3d4 c5a7 b1c3 d7d6 h2h3 h7h6 e1g1 g8e7 c1e3 e8g8 a1c1 b7b5 c4d3 f7f5 d4d5 c6e5 f3e5 d6e5 e3a7 a8a7 f1e1 f5f4 d3f1 c7c6 d5c6 d8d1 e1d1 e7c6 c3d5 c6d4 d5f4 f8f4 c1c8 f4f8 d1c1 a7f7 c8f8 g8f8 f2f3 g7g5 g1f2 f8e7 f1d3 f7f4');
INSERT INTO evaluations VALUES(4881,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(4881,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(4881,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(4881,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(4881,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(4881,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(4881,36,'cp 147','c3d4 c5a7 b1c3 d7d6 h2h3 b7b5 c4b3 c6a5 b3c2 g8e7 e1g1 e8g8 b2b3 a5c6 f1e1 f7f5 b3b4 h7h6 a2a3 f5e4 c3e4 c8f5 c2b3 g8h8 c1b2 d6d5 e4c5 d8d6 e1e3 f8f6 f3e5 a8f8 d1e2 f5c8 a1e1 f6f2 e2f2 f8f2 g1f2 d6f6 f2g1 c8f5 b2c3 c6e5 e3e5 e7c6 e5d5 f6g6 c5d7 f5h3');
INSERT INTO evaluations VALUES(4880,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(4880,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(4880,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(4880,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(4880,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(4880,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(4880,36,'cp 147','c5a7 b1c3 d7d6 h2h3 h7h6 c1e3 g8e7 a2a3 b7b5 c4a2 c6a5 e1g1 e8g8 f1e1 a5c4 a2c4 b5c4 d4d5 a7e3 e1e3 a8b8 d1e2 e7g6 a1a2 f7f5 e2c4 f5e4 c4e4 d8f6 b2b4 c8d7 g1h2 f6f7 f3d4 b8e8 d4e6 g6f4 e4f4 f7f4 e6f4 f8f4 e3e8 d7e8 a2e2 e8g6 e2d2 f4c4 c3e2 g8f7 f2f3 g6c2');
INSERT INTO evaluations VALUES(4941,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(4941,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(4941,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(4941,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(4941,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(4941,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(4941,36,'cp 161','e1g1 d7d6 h2h3 b7b5 c4b3 c6a5 b3c2 g8e7 b2b3 e8g8 b1c3 a5c6 f1e1 h7h6 c1b2 c6b4 c2b1 b4c6 c3e2 f7f5 e2g3 d6d5 e4e5 f5f4 g3e2 d8e8 d1d2 e8f7 b1d3 g7g5 a1c1 b5b4 a2a3 g5g4 h3g4 c8g4 e5e6 g4e6 a3b4 e6g4 f3h2');
INSERT INTO evaluations VALUES(4940,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(4940,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(4940,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(4940,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(4940,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(4940,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(4940,36,'cp 138','d7d6 h2h3 h7h6 b1c3 g8e7 c1e3 e8g8 a1c1 b7b5 c4d3 f7f5 d4d5 c6e5 f3e5 d6e5 d1b3 g8h8 e3a7 a8a7 f1d1 a7a8 b3a3 e7g6 d3f1 a8b8 a3c5 d8g5 c5c7 b8b7 c7d6 f5f4 g1h2 c8d7 b2b3 f8d8 b3b4 d8f8 c3e2 f4f3');
INSERT INTO evaluations VALUES(4937,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(4937,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(4937,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(4937,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(4937,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(4937,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(4937,36,'cp 125','h2h3 h7h6 b1c3 g8e7 c1e3 e8g8 a1c1 b7b5 c4d3 f7f5 d4d5 c6e5 f3e5 d6e5 e3a7 a8a7 f1e1 d8d6 e4f5 c8f5 d3f5 f8f5 d1g4 d6g6 c1d1 a7b7 c3e4 b7b6 g4g6 e7g6 g2g3 g6f8 d1c1 f5f7 g1g2 f7d7 c1d1 d7e7 b2b3 f8h7 h3h4 h7f6 f2f4 e7e8 e4f6 g7f6 f4e5 e8e5 e1e5 f6e5');
INSERT INTO evaluations VALUES(14381,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(14381,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(14381,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(14381,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(14381,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(14381,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(14381,36,'cp 102','f1c4 a7a6 b1c3 b7b5 c4b3 e7e6 e3e4 c6a5 b3c2 c7c5 d4d5 e6d5 e4e5 f6g8 d1d5 c8b7 d5d2 d8d2 c1d2 a5c4 e1c1 g8e7 d2g5 h7h6 g5f4 g7g5 c3e4 b7e4 c2e4 a8c8 b2b3 c4a5 f4d2 a5c6 h2h3 f8g7 d2c3 e8g8 d1d6');
INSERT INTO evaluations VALUES(51470,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(51470,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(51470,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(51470,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(51470,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(51470,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(51470,36,'cp 139','b1c3 f8g7 g1f3 a7a6 c1f4 d7d6 d1e2 h7h6 e4e5 d6d5 c3d1 e7e6 c2c3 c8d7 e2d2 d7b5 f1b5 a6b5 e1g1 c7c5 d1e3 b8c6 a2a3 c6a5 d2e2 a5c4 e3g4 g6g5 f4e3 d8d7 b2b3 c4a5 f1b1 g8e7 a3a4 b5a4 a1a4 e7f5');
INSERT INTO evaluations VALUES(51468,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(51468,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(51468,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(51468,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(51468,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(51468,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(51468,36,'cp 138','f8g7 g1f3 c8b7 h2h4 h7h6 c1f4 a7a6 d1e2 b6b5 e4e5 e7e6 c3e4 g8e7 g2g3 e7d5 f1g2 d7d6 h4h5 g6g5 f3g5 b8d7 g5f3 d8e7 e1c1 e8c8 h1g1 d7b6 f3e1 d5b4 e1d3 b4d3 d1d3 d6e5 f4e5 g7e5 d4e5 d8d3 e2d3 h8d8 d3e3 e7b4');
INSERT INTO evaluations VALUES(37583,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(37583,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(37583,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(37583,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(37583,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(37583,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(37583,36,'cp 139','g1f3 a7a6 h2h4 h7h6 c1f4 c8b7 d1e2 b6b5 e4e5 e7e6 c3e4 g8e7 g2g3 e7d5 f1g2 b8c6 f4d2 d7d6 c2c3 d8e7 e1g1 e8g8 f1e1 a8d8 b2b3 d6e5 d4e5 f8e8 b3b4 b7a8 a1c1 d5b6 e4f6 g7f6 e5f6 e7f8');
INSERT INTO evaluations VALUES(47078,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(47078,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47078,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47078,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47078,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(47078,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47078,36,'cp 0','e7f6 e2e3 b8d7 c2c4 d5c4 f1c4 f8d6 d1c2 g7g6 g1f3 e8g8 b1c3 f6f5 e1g1 c7c6 d4d5 d7b6 c2d3 c6c5 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(48954,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(48954,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48954,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48954,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48954,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48954,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48954,36,'cp 0','e2e3 b8d7 c2c4 d5c4 f1c4 f8d6 d1c2 e8g8 g1f3 g7g6 e1g1 f6f5 b1c3 d7f6 c3b5 a7a6 b5d6 c7d6 a2a4 c8d7 c2b3 b7b5 a4b5 a6b5 c4e2 d8e7 e2d3 a8a4 h2h3 f8a8 a1a4 b5a4 b3b7 a8c8 d3b5 e7e6 f3g5 e6e7');
INSERT INTO evaluations VALUES(48953,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48953,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48953,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48953,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48953,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48953,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(48953,36,'cp -9','b8d7 c2c4 d5c4 f1c4 f8d6 d1c2 e8g8 g1f3 g7g6 e1g1 f6f5 b1c3 c7c6 d4d5 d7b6 c2b3 a8b8 a1d1 b6c4 b3c4 b7b5 c4d3 b5b4 d5c6 b4c3 d3d6 d8d6 d1d6 c3b2 f1b1 b8b6 d6d2 b6c6 d2b2 g8g7 h2h3 c6c5 f3d4 c8a6 a2a4 f8e8 d4b5 a6b5 b2b5');
INSERT INTO evaluations VALUES(56028,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(56028,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(56028,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(56028,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(56028,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(56028,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(56028,36,'cp 3','b7b5 a2a4 c7c6 a4b5 c6b5 b2b3 g8f6 b3c4 f6e4 b1d2 f8b4 f1d3 e4d2 c1d2 b4d2 d1d2 c8b7 c4b5 e8g8 e1g1 b7f3 g2f3 b8d7 d3e4 a8b8 e4c6 d7b6 d2e3 b6d5 e3e5 d8e7 f1c1 e7h4 c1e1 h4h3 e5g3 h3f5 a1a7 d5c3 g3e5 c3b5 e5f5 e6f5 c6b5');
INSERT INTO evaluations VALUES(51376,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(51376,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(51376,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51376,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(51376,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(51376,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(51376,36,'cp 0','a2a4 c7c6 a4b5 c6b5 b2b3 g8f6 b3c4 f6e4 c4c5 f8c5 f1b5 c8d7 b5d7 d8d7 d4c5 d7d1 e1d1 e4f2 d1e2 f2h1 c1e3 f7f6 b1d2 b8c6 a1h1 e6e5 d2c4 e8e7 f3d2 a8c8 d2e4 c8c7 h1a1 h8b8 e4c3 c7d7 a1a6 b8b4 c4d6 b4b2 e3d2 c6b4 a6a4 b4c6 a4a6');
INSERT INTO evaluations VALUES(51378,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51378,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(51378,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(51378,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(51378,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51378,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(51378,36,'cp 16','c7c6 a4b5 c6b5 b2b3 g8f6 b3c4 f6e4 c4c5 f8c5 f1b5 c8d7 b5d7 d8d7 d4c5 d7d1 e1d1 e4f2 d1e2 f2h1 c1e3 f7f6 b1d2 b8c6 a1h1 e6e5 d2c4 e8e7 f3d2 a8b8 h1a1 h8c8 a1a6 c8c7 g2g4 b8b4 e3f2 c6b8 a6a3 b8c6 e2d3 a7a5 d3c3 a5a4 c4b6 e7e6 d2c4');
INSERT INTO evaluations VALUES(51699,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51699,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51699,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51699,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51699,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51699,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51699,36,'cp 8','a4b5 c6b5 b2b3 g8f6 b3c4 f6e4 c4b5 f8b4 b1d2 e4d2 c1d2 b4d2 d1d2 c8b7 f1d3 e8g8 e1g1 b7f3 g2f3 b8d7 d3e4 a8b8 e4c6 d7e5 f1d1 e5c6 b5c6 b8c8 d1c1 d8d6 a1a6 f8d8 d2e3 c8c7 e3e5 d6e5 d4e5 g7g5 a6a5 d8c8 a5c5 a7a6 c1c4 h7h6 g1g2 g8g7');
INSERT INTO evaluations VALUES(51691,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51691,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(51691,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51691,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(51691,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(51691,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51691,36,'cp 11','c6b5 b2b3 g8f6 b3c4 f6e4 c4c5 f8c5 f1b5 c8d7 b5d7 d8d7 d4c5 d7d1 e1d1 e4f2 d1e2 f2h1 c1e3 b8c6 b1d2 e8e7 a1h1 e6e5 h1a1 f7f6 d2c4 a8c8 g2g4 c8c7 a1a6 h8b8 f3d2 b8b4 h2h4 c6b8 a6a3 b8c6 e2d3 e7e6 a3a6 e6d5 d3c3 b4b8 c4d6 c6d4 d6e4');
INSERT INTO evaluations VALUES(57064,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(57064,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(57064,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57064,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(57064,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(57064,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(57064,36,'cp -5','d7d5 e2e4 d5d4 c3e2 c7c5 c2c3 g8f6 c3d4 c5d4 e2d4 f6e4 b2b3 c8d7 c1b2 b8c6 f1d3 e4f6 d1e2 c6d4 f3d4 f8e7 d4f3 e8g8 f3e5 f6d5 d3e4 f7f6 e5d7 d8d7 e2d3 f6f5 e4d5 d7d5 d3d5 e6d5 a1c1 e7f6 b2f6 f8f6 c1c7');
INSERT INTO evaluations VALUES(55941,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(55941,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(55941,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(55941,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55941,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55941,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55941,36,'cp -4','e2e4 d5d4 c3e2 c7c5 c2c3 g8f6 c3d4 c5d4 e2d4 f6e4 b2b3 f8c5 f1b5 c8d7 d1c2 c5d4 f3d4 e4d6 b5d3 d8b6 c1b2 b8a6 e1g1 a6b4 c2b1 b4d3 b1d3 e8g8 a2a4 f8e8 a4a5 b6d8 a5a6 b7a6 d3a6 d6f5 a6a5 f5d4 b2d4 d7c6 a5d8 e8d8 d4a7 d8d2 a7e3 d2a2 a1a2 a8a2 f1c1');
INSERT INTO evaluations VALUES(55764,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(55764,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(55764,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(55764,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(55764,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(55764,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(55764,36,'cp -75','d5f4 e6e5 f4e2 g8f6 e2g3 h7h5 h2h4 c8g4 g1f3 b8c6 f1b5 d8d6 c2c3 d4d3 d1b3 g4f3 g3f5 d6d7 g2f3 g7g6 f5e3 f8h6 b3c4 a7a6 b5c6 b7c6 b2b3 e8g8 h1g1 g8h7 c1a3 f8e8 e1c1 a6a5 g1h1');
INSERT INTO evaluations VALUES(56063,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(56063,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(56063,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(56063,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(56063,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(56063,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(56063,36,'cp -90','e6e5 f4e2 g8f6 e2g3 h7h5 h2h4 b8c6 g1f3 c8g4 f1b5 d8d6 a2a3 f6d7 d2d3 g7g6 b5c4 d7b6 c4a2 f7f6 c1d2 d6d7 b2b4 e8c8 d1b1 b6a4 f3h2 g4e6 a2e6 d7e6 h2f3 a7a6 b1a2 e6a2 a1a2 c6e7 e1g1 f8h6 d2h6');
INSERT INTO evaluations VALUES(56600,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(56600,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(56600,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(56600,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(56600,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(56600,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(56600,36,'cp -96','f4e2 g8f6 e2g3 h7h5 h2h4 b8c6 g1f3 c8g4 f1b5 d8d6 b5c6 d6c6 f3e5 g4d1 e5c6 d1c2 c6d4 c2e4 g3e4 f6e4 d2d3 e8c8 c1e3 e4f6 e1c1 f6g4 d4f3 f8e7 c1c2 h8e8 h1e1 e7f6 f3g5 d8d6 d3d4 g4e3 e1e3 e8e3 f2e3 f6g5 h4g5 d6g6 d1f1 g6g5 f1f7 g5g2 c2d3');
INSERT INTO evaluations VALUES(22761,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(22761,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(22761,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(22761,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(22761,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(22761,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(22761,36,'cp 210','d4d5 c6b8 f3e5 g8f6 g2g3 c7c6 f1g2 c6d5 e1g1 b8c6 e5d3 d7d6 c2c3 f8e7 d3f4 e8g8 b1a3 d5d4 a3b5 d4c3 b5c3 f8e8 c1e3 c8d7 c3d5 f6d5 f4d5 e7f6 a1c1 a8c8 d1b3 c6d4 b3d3 d4c6 b2b3 f6b2 c1c2 b2a3 c2c4 h7h6 f1d1');
INSERT INTO evaluations VALUES(22662,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(22662,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(22662,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(22662,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(22662,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(22662,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(22662,36,'cp 190','c6b8 f3e5 g8f6 c1f4 f8c5 c2c4 d7d6 e5d3 c5b6 e2e3 c8f5 f1e2 b6a5 b1d2 e8g8 e1g1 c7c6 d2b3 a5b6 f4g3 c6d5 c4d5 b8a6 e2f3 a8c8 a2a4 f5d3 d1d3 a6b4 d3d2 a7a5 a1c1 f8e8 c1c8 d8c8');
INSERT INTO evaluations VALUES(48677,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(48677,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(48677,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(48677,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(48677,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(48677,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(48677,36,'cp -25','f1d3 f8d6 g1e2 e8g8 e1g1 f8e8 b1c3 b7b6 e2g3 c8b7 b2b3 b8d7 c1b2 g7g6 c3b5 d6f8 b5c3 a7a6 c3e2 c7c6 a1c1 b6b5 a2a4 f8d6 f1e1 d6b4 e1f1 b4f8 e2f4 d8b6 d1c2 a6a5 d3e2 a8b8 a4b5 c6b5 f4d3 b7a6');
INSERT INTO evaluations VALUES(22076,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(22076,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(22076,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(22076,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(22076,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(22076,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(22076,36,'cp 103','c2c4 g8f6 b1c3 g7g6 h2h3 f8g7 f1e2 e8g8 c1g5 h7h6 g5e3 f8e8 d1c2 d8e7 c2d2 e6e5 d4e5 c6e5 f3d4 e5c6 f2f3 c6d4 e3d4 b7b6 e1c1 c8b7 d4e3 e7f8 g2g4 f6d7 h3h4 a8b8 c3d5 d7c5 d5c7 e8c8 c7d5 b7d5 e4d5 c8e8 e2d3 c5d3 d2d3');
INSERT INTO evaluations VALUES(22069,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(22069,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(22069,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(22069,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(22069,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(22069,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(22069,36,'cp 100','g8f6 b1c3 g7g6 h2h3 f8g7 c1g5 e8g8 f1e2 h7h6 g5e3 f8e8 d1c2 d8e7 e1g1 e6e5 d4e5 c6e5 f3e5 e7e5 a1e1 f6e4 e2d3 e4f6 c3d5 f6d5 e3d2 d5f4 e1e5 d6e5 f1e1 c8e6 d3e4 c7c6 d2e3 a8d8 b2b4 f7f5 e3f4 e5f4 e4f3 a7a6 c4c5 d8d7 a2a4 g8h8');
INSERT INTO evaluations VALUES(55564,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(55564,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(55564,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(55564,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(55564,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(55564,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(55564,36,'cp 39','b1c3 f7f5 f1g2 f8e7 g1f3 g8f6 e1g1 e8g8 d2d4 e5e4 f3g5 c7c6 d4d5 b8a6 g5h3 a6c7 f2f3 c6d5 c4d5 b7b5 a2a3 e4f3 e2f3 a7a5 f1e1 c8d7 h3f4 f8e8 d1d4 g7g6 c1e3 e7f8 d4d2 f8h6 c3e2 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(6379,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(6379,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(6379,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(6379,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(6379,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(6379,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(6379,36,'cp 12','g8f6 e1g1 e8g8 b5c6 d7c6 f3e5 f8e8 e5d3 b4c3 d2c3 f6e4 c1f4 c8f5 f1e1 d8d7 f2f3 e4f6 g2g4 f5g6 d3e5 d7d1 a1d1 g6c2 d1d2 c2g6 c3c4 h7h5 h2h3 a8d8 d2d8 e8d8 g1g2 c6c5 g2g3 g6c2 f4g5 d8e8 b2b3 h5g4 h3g4');
INSERT INTO evaluations VALUES(11323,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(11323,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(11323,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(11323,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(11323,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(11323,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(11323,36,'cp 67','c2c4 c8f5 b1c3 e7e6 a1c1 h7h6 a2a3 f8d6 c4d5 e6d5 f4e5 f5e6 e5d6 c7d6 f1d3 e8g8 h2h3 a8c8 e1g1 c6a5 d1e2 f8e8 f3d2 c8c6 c3a2 d8d7 c1c6 a5c6 a2c3 c6e7 f1c1 e6f5 d3f5 d7f5');
INSERT INTO evaluations VALUES(11307,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(11307,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(11307,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(11307,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(11307,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(11307,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(11307,36,'cp 64','c8f5 b1c3 e7e6 a1c1 h7h6 a2a3 f8d6 c4d5 f6d5 f4g3 d5c3 c1c3 e8g8 f1d3 d8d7 e1g1 f8d8 d3e2 f5h7 d1b3 c6e7 f3e5 d6e5 g3e5 e7d5 c3c5 c7c6 f1c1 a6a5 e2f3 d5f6 e5f6 g7f6 b3c4 a5a4 b2b4');
INSERT INTO evaluations VALUES(36615,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36615,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36615,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36615,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36615,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(36615,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36615,36,'cp 0','b1c3 c7c5 d4c5 d6c5 h2h3 d8b6 e1g1 b6b2 d1d2 b8c6 a1b1 b2a3 c3b5 a3a5 d2a5 c6a5 b5c7 a8b8 c7b5');
INSERT INTO evaluations VALUES(36598,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(36598,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36598,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36598,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36598,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(36598,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36598,36,'cp 0','c7c5 e1g1 c5d4 e3d4 c8g4 h2h3 g4f3 d1f3 b8c6 c3e2 e7e5 d4e5 d6e5 f4e3 d8e7 e2c3 a8d8 f1e1 c6d4 e3d4 d8d4 f3e3 e7d7 a1d1 f8d8 c3e2 d4d6 e2c3');
INSERT INTO evaluations VALUES(53169,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(53169,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(53169,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(53169,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(53169,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(53169,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53169,36,'cp -6','e7e6 g1f3 f8c5 c2c4 g8f6 a2a3 e8g8 c4d5 e6d5 b2b4 c5d6 c1b2 a7a5 b4a5 d8a5 b2c3 a5d8 f1d3 b8c6 e1g1 d8e7 h2h3 c8d7 b1d2 h7h6 a3a4 c6b4 d3b5 d7f5 c3f6 e7f6 f3d4 f5g6 d2f3 f8c8 a1c1 c8c1 d1c1');
INSERT INTO evaluations VALUES(54071,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(54071,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(54071,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(54071,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(54071,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54071,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(54071,36,'cp 0','a2a3 f8c5 g1f3 g8f6 f1e2 e8g8 e1g1 a7a5 c2c4 d5c4 e2c4 b7b6 b1d2 d8e7 b2b3 f8d8 d1e2 c8b7 c1b2 h7h6 c4d3 b8d7 f1d1 a8c8 d2c4 c8a8');
INSERT INTO evaluations VALUES(54252,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54252,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54252,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54252,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54252,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(54252,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54252,36,'cp 0','d1e2 g8f6 g2g3 f8e7 a2a4 d8a5 c1d2 a5b6 d2c1');
INSERT INTO evaluations VALUES(45474,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45474,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45474,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45474,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45474,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45474,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45474,36,'cp 0','d2d4 c8g4 h2h3 g4f3 d1f3 b8c6 c1e3 e7e6 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(55966,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(55966,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(55966,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(55966,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(55966,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(55966,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(55966,36,'cp -65','d5d4 c3e4 d8d5 d2d3 d5e5 g1f3 e5d5 g2g3 f8e7 f1g2 g8f6 e1g1 d5d8 h2h4 f6e4 d3e4 c7c5 e4e5 c8d7 b2b4 d7c6 b4c5 e7c5 c1b2 e8g8 f3d4 c6g2 g1g2 d8d5 d1f3 d5f3 d4f3 f8c8 a1c1 b8c6 f1d1 b7b6 h4h5 h7h6 d1d7');
INSERT INTO evaluations VALUES(55985,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(55985,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(55985,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(55985,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(55985,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(55985,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(55985,36,'cp -68','c3e4 d8d5 d2d3 d5e5 g1f3 e5d5 g2g3 f8e7 f1g2 g8f6 e1g1 d5d8 f3e5 f6e4 g2e4 b8d7 e5d7 d8d7 d1g4 e8g8 c1h6 e7f6 f2f4 a8b8 a1e1 b7b6 h2h4 c7c5 a2a3 d7c7 b2b3 f8e8 h6g5 f6e7 g4h5 g7g6 h5e2 f7f6 g5h6 e7f8 h6f8 e8f8 h4h5 c7g7 h5g6 h7g6 e2h2 f6f5');
INSERT INTO evaluations VALUES(34686,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(34686,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(34686,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(34686,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(34686,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(34686,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(34686,36,'cp 100','g1f3 c7c5 e2e4 b8c6 d2d4 c5d4 f3d4 c6d4 d1d4 e7e5 d4d3 c8e6 c1d2 a7a6 d3g3 g8f6 f1e2 a8d8 a1d1 a5c7 e1g1 f8e7 f2f4 h7h5 g1h1 e8g8 f4f5 h5h4 g3h4 e6c4 h4e1 d8d4 d2e3 c4e2 e1e2 d4d1 f1d1 f8d8 h2h3 d8d1 e2d1');
INSERT INTO evaluations VALUES(34680,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(34680,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(34680,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(34680,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(34680,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(34680,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34680,36,'cp 72','c7c5 e2e4 b8c6 d2d4 c5d4 f3d4 c6d4 d1d4 e7e5 d4d3 c8e6 c1d2 a7a6 f1e2 g8f6 d3g3 f8e7 e1g1 e8g8 a1d1 a5b6 d2h6 f6e8 h6c1 b6c7 c1e3 a8d8 d1d8 e7d8 h2h4 e8f6 e3h6 f6e8 f1c1 c7a5 h6e3 d8b6');
INSERT INTO evaluations VALUES(44108,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(44108,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(44108,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(44108,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(44108,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(44108,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(44108,36,'cp 163','d2d4 e5d4 d1d4 d7d5 c1g5 f8e7 e4d5 h7h6 g5f6 e7f6 d4e3 d8e7 c3e4 c6d5 e4f6 g7f6 e3e7 e8e7 e1c1 c8e6 g2g3 b8c6 f1g2 h8e8 f3d4 e7f8 d1d2 a8d8 h1e1 e6c8 e1d1 c8g4 g2f3 f6f5 f3g4 f5g4 d4f5 h6h5 d2d5 d8d5 d1d5 e8e1 c1d2 e1a1 f5e3 a1a2');
INSERT INTO evaluations VALUES(44107,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(44107,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(44107,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(44107,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(44107,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(44107,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(44107,36,'cp 174','e5d4 d1d4 d7d5 c1g5 f8e7 e4d5 h7h6 g5f6 e7f6 d4e3 d8e7 c3e4 c6d5 e4f6 g7f6 e3e7 e8e7 e1c1 c8e6 g2g3 b8c6 f3d4 c6d4 d1d4 a8c8 f1d3 f6f5 d4a4 a7a6 h1e1 e7f6 a4h4 h6h5 c1d2 b7b5 e1e3 c8c6 e3f3');
INSERT INTO evaluations VALUES(44125,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(44125,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(44125,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(44125,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(44125,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(44125,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(44125,36,'cp 173','d1d4 d7d5 c1g5 f8e7 e4d5 h7h6 g5f6 e7f6 d4e3 d8e7 c3e4 c6d5 e4f6 g7f6 e3e7 e8e7 e1c1 c8e6 g2g3 b8c6 f1g2 h8e8 d1d2 a8d8 h1d1 d8d7 f3d4 e7f8 c2c3 d7e7 d1e1 e8d8 d4c2 e7d7 g2f3 c6e5 f3e2 e6f5 e1d1 f5e4 h2h4');
INSERT INTO evaluations VALUES(43543,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(43543,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(43543,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(43543,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(43543,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(43543,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(43543,36,'cp 45','c1e3 f8e7 a2a3 e8g8 d1d2 e6e5 d4b3 b7b5 e1c1 b8c6 f2f3 d8c7 c1b1 b5b4 c3d5 f6d5 e4d5 a6a5 a3a4 c6d8 e2b5 c8d7 d2e2 a8b8 f3f4 d7b5 a4b5 e5f4 e3d4 c7d7 h1e1 a5a4 e2e7 d7e7 e1e7 a4b3 e7d7 b3c2 b1c2 b8b5 d7d6 f7f6 d6d7');
INSERT INTO evaluations VALUES(43534,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43534,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(43534,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(43534,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(43534,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43534,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(43534,36,'cp 15','f8e7 g2g4 d6d5 e4d5 f6d5 c3d5 d8d5 e2f3 d5c4 d1e2 c4e2 f3e2 c8d7 e1c1 b8c6 e2f3 e8c8 d4c6 d7c6 d1d8 c8d8 f3c6 b7c6 h1d1 d8c7 d1d4 h7h5 g4g5 e7d6 h2h3 f7f5 g5f6 g7f6 d4a4 h8a8 a4h4 a8h8 c1d2 f6f5 h4a4');
INSERT INTO evaluations VALUES(39735,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(39735,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(39735,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(39735,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(39735,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(39735,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(39735,36,'cp 38','f2f4 e6e5 f4e5 d6e5 d4f5 d8d1 a1d1 c8f5 e4f5 e7b4 e1g1 b4c3 b2c3 b8c6 e2f3 e5e4 f3e2 h7h5 h2h3 a8c8 c3c4 c6e5 d1d4 e5c6 d4d6 e8g8 e3g5 c6e5 d6d4 c8c5 f1b1 c5a5 g5f6 g7f6 b1b2 a5a4 g1f2 b7b5 f2e3 e5c4 e2c4 b5c4 d4e4 f8c8');
INSERT INTO evaluations VALUES(46923,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46923,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(46923,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(46923,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(46923,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(46923,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(46923,36,'cp 55','d6e5 f4e5 b8c6 c1e3 c8f5 b1c3 e7e6 g1f3 f8e7 f1e2 f7f6 e1g1 e8g8 e5f6 e7f6 d1d2 d8e7 a1d1 a8d8 e3f2 g7g5 d4d5 c6e5 f3e5 f6e5 d2e3 e5f4 e3c5 f4d6 c5a5 d6b4 a5a7 e6d5 e2f3 d5c4 d1d8 f8d8 a7b7 b4c3 b2c3 g5g4 f2b6 g4f3 b7f3');
INSERT INTO evaluations VALUES(46936,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(46936,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(46936,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(46936,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(46936,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(46936,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(46936,36,'cp 62','f4e5 c8f5 b1c3 b8c6 g1f3 e7e6 c1e3 f8e7 f1e2 e8g8 e1g1 f7f6 e5f6 e7f6 d1d2 d8e7 c4c5 b6d5 e3f2 a8d8 a1e1 f5g4 c3e4 g4f3 e4f6 e7f6 e2f3 d5f4 f3c6 b7c6 d2e3 h7h5 f2g3 d8d4 h2h3 f6g5 f1f4 d4f4 g3f4 f8f4 e3e6 g8h7 e6e3 f4f5 e3e4 h7h8 e4c6');
INSERT INTO evaluations VALUES(46935,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(46935,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(46935,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(46935,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(46935,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46935,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(46935,36,'cp 60','c8f5 b1c3 e7e6 g1f3 b8c6 c1e3 f8e7 f1e2 e8g8 e1g1 f7f6 e5f6 e7f6 d1d2 d8e7 a1d1 a8d8 e3f2 d8d7 c4c5 b6d5 d1e1 e7d8 c3d5 d7d5 e2c4 d5d7 e1e6 g8h8 e6f6 f8f6 f2h4 c6d4 d2e3 d4f3 e3f3 f5g6 h4f6 g7f6 f3f6 d8f6 f1f6 h8g7 f6f4 a7a5 a2a3 d7d1 f4f1 d1d2 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(29430,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(29430,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(29430,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(29430,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(29430,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(29430,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(29430,36,'cp 67','b1c3 e7e6 g1f3 b8c6 f1e2 f8e7 c1e3 e8g8 e1g1 f7f6 e5f6 e7f6 d1d2 d8e7 g1h1 a8d8 a1d1 d8d7 d2e1 h7h6 e3g1 a7a6 h2h3 f5h7 e1g3 g8h8 b2b3 e7b4 g3e1 h7c2 d1c1 c2h7 e1d2 f8d8 c1d1 b4e7');
INSERT INTO evaluations VALUES(56367,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(56367,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(56367,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(56367,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(56367,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(56367,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(56367,36,'cp 145','d4e5 b8c6 b1c3 f8b4 c1d2 g8e7 a2a3 b4a5 d1h5 c8e6 e1c1 d8d7 g1f3 h7h6 f3g5 g7g6 h5h4 e8c8 g5e6 d7e6 f2f4 g6g5 h4h3 e6h3 g2h3 g5f4 e3f4 e7f5 f1d3 c6e7 b2b4 a5b6 c3e2 c8b8 e2g3 f5e3 d1e1 e3c4 d3c4 d5c4 a3a4');
INSERT INTO evaluations VALUES(56314,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(56314,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(56314,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(56314,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(56314,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(56314,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(56314,36,'cp 120','b8c6 b1c3 f8b4 c1d2 g8e7 a2a3 b4a5 d1h5 c8e6 g1f3 d8d7 h2h3 h7h6 f1d3 a5b6 b2b4 d5d4 f3d4 c6d4 e3d4 b6d4 e1g1 e7c6 f1e1 g7g6 h5h4 g6g5 h4g3 e8c8 a1d1 e6f5 b4b5 f5d3 c2d3 c6a5 c3e2 d4b6 d2e3 d7b5 d3d4 c8b8 g3f3');
INSERT INTO evaluations VALUES(21651,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(21651,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(21651,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(21651,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(21651,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(21651,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(21651,36,'cp 132','b1c3 f8b4 c1d2 g8e7 a2a3 b4a5 d1h5 c8e6 g1f3 d8d7 e1c1 h7h6 f3g5 c6e5 g5e4 e7c6 e4c5 d7e7 c5e6 e7e6 f2f4 e5c4 f1c4 d5c4 f4f5 e6e5 c1b1 e8c8 h5f7 h8f8 f7c4 f8f5 c4g4 e5e6 e3e4 f5f6 g4e6 f6e6 d2c1 a5b6 d1d8 c6d8');
INSERT INTO evaluations VALUES(21647,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(21647,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(21647,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(21647,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(21647,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(21647,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(21647,36,'cp 112','f8b4 f2f4 f7f6 f1d3 f6e5 d1h5 e8f8 f4e5 g8h6 g1f3 c8g4 h5g5 g4f3 g5f4 h6f7 g2f3 g7g5 f4g3 c6e5 c1d2 e5d3 c2d3 d5d4 e3d4 d8d4 e1c1 b4d6 g3h3 a8e8 c3e4 d6e5 d2c3 d4e3 c1b1 e5c3 e4c3 c7c6 h1e1 e3e1 d1e1 e8e1 b1c2 e1e7 c3e4 h7h5 h3c8 f8g7 c8f5 h8h6 e4g5');
INSERT INTO evaluations VALUES(9253,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9253,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(9253,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(9253,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9253,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9253,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(9253,36,'cp 24','e8g8 b1d2 b7b6 f1d3 c8b7 e1g1 c6e7 b2b4 e7f5 g3e5 g7g6 a2a4 a7a5 b4a5 a8a5 d2b3 a5a7 a4a5 d8e7 d3f5 e6f5 a5b6 a7a1 d1a1 f8a8 a1b2 c7b6 b3d2 f6d7 e5d6 e7d6 h2h3 d6a3 b2b5 a3d6 f1b1 b7c6 b5b2');
INSERT INTO evaluations VALUES(3762,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(3762,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(3762,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(3762,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(3762,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(3762,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(3762,36,'cp 37','b1d2 b7b6 f1d3 c8b7 b2b4 c6e7 e1g1 e7f5 g3e5 f6d7 d1c2 g7g6 f1b1 f5h4 e5g3 h4f3 d2f3 a7a5 a2a4 d6g3 h2g3 d8e7 e3e4 d5e4 d3e4 b7e4 c2e4 a5b4 c3b4 a8a7 e4c2 f8a8 b4b5 e7d6 b1d1 d6b4 d1c1 a7a4');
INSERT INTO evaluations VALUES(3761,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(3761,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(3761,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(3761,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(3761,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(3761,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(3761,36,'cp 27','b7b6 f1d3 c8b7 b2b4 c6b8 e1g1 b8d7 d1b1 d8e7 g3h4 h7h6 f1e1 g7g5 h4g3 a7a5 b4b5 c7c5 b5c6 b7c6 c3c4 d6g3 h2g3 d5c4 d2c4 b6b5 c4e5 c6f3 g2f3 d7e5 d4e5 f6d7 d3b5 d7e5 b5e2 a8b8 b1e4 e7f6 e1b1 e5g6');
INSERT INTO evaluations VALUES(25496,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(25496,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(25496,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(25496,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(25496,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(25496,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(25496,36,'cp 92','b5c6 d7c6 d1e2 e5d4 f3d4 f8e7 c1e3 f6d7 e1c1 e7f6 f2f3 a7a6 e2d2 f6d4 e3d4 f7f6 h1e1 e8g8 g2g4 b7b5 e1g1 d7e5 d2e3 d8e8 b2b3 e5g6 c3e2 a8d8 d4b2 d6d5 e3c3 c6a8 b2a3 d5e4 a3f8 d8d1 g1d1 e8f8 c3c7 e4f3');
INSERT INTO evaluations VALUES(25453,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(25453,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(25453,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(25453,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(25453,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(25453,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(25453,36,'cp 122','d7c6 d1e2 e5d4 f3d4 f8e7 c1e3 f6d7 e1c1 e7f6 d4c6 b7c6 e2c4 d7e5 c4a4 e8g8 h1e1 e5g4 e4e5 g4e5 e3d4 f8e8 f2f4 e5g4 d4f6 g4f6 a4c6 a8c8 g2g3 e8e1 d1e1 h7h5 c3b5 a7a6 b5c3 a6a5 b2b3 g8f8 a2a4 h5h4 c1b2 h4h3 c3d5 d8d7 c6d7 f6d7');
INSERT INTO evaluations VALUES(25088,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(25088,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(25088,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(25088,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(25088,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(25088,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(25088,36,'cp 111','d1e2 e5d4 f3d4 f8e7 c1e3 f6d7 e1c1 e7f6 c1b1 f6d4 e3d4 d8e7 h1e1 f7f6 c3b5 d7c5 e4e5 e8g8 e2c4 e7e6 c4e6 c5e6 e5d6 c6b5 e1e6 c7d6 e6d6 f8d8 d4e3 d8d6 d1d6 a7a6 g2g3 a8e8 b2b3 b5c6 c2c4 g8f7 b1b2 h7h5 b2c3 g7g5 a2a4 e8e6 d6d8 e6e8 d8d2 e8e6 c3b4');
INSERT INTO evaluations VALUES(56964,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(56964,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(56964,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(56964,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(56964,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(56964,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(56964,36,'cp 75','c2c3 d7d5 e4e5 c7c5 d2d4 b8c6 f1d3 d8c7 e1g1 c5d4 c3d4 c6b4 d3b5 c8d7 b5d7 c7d7 c1d2 b4c6 b1c3 g8e7 c3e2 e7g6 h2h4 f8e7 g2g3 e7d8 d1b3 g6e7 g1g2 d8a5 b3d3 e7f5 f1c1 a5d2 d3d2 e8g8 h4h5 f8c8');
INSERT INTO evaluations VALUES(6015,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(6015,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6015,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6015,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(6015,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(6015,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(6015,36,'cp 12','c3a4 c5b6 a4b6 a7b6 f1e2 g8e7 d3d4 e8g8 e1g1 h7h6 a2a4 e7g6 c2c3 d8e7 d4d5 c6b8 c1e3 f7f5 e4f5 c8f5 b2b4 b8d7 c3c4 d7f6 a1a3 g8h8 h2h3 a8e8 f1e1 e7f7 e2f1 f5d7');
INSERT INTO evaluations VALUES(12934,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(12934,31,'cp 245',NULL);
INSERT INTO evaluations VALUES(12934,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(12934,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(12934,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(12934,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(12934,36,'cp 248','e4e5 f6g8 c1f4 e7e6 b1c3 g8e7 e1c1 h7h5 f1d3 a7a6 d3e4 e7c6 d4e3 d8a5 h2h4 c6e5 c3d5 e6d5 d1d5 e5c4 d5a5 c4e3 f4e3 f8e7 e3d4 e8f8 e4f3 d7d6 a5h5 h8h5 f3h5 c8d7 h5f3 d7c6 f3c6 b7c6 g2g4 a8b8 c2c3 f7f5 g4f5 b8b5 h1g1 b5f5 d4g7 f8g8 g7d4 g8f8 g1g6 e7h4');
INSERT INTO evaluations VALUES(12930,30,'cp 238',NULL);
INSERT INTO evaluations VALUES(12930,31,'cp 258',NULL);
INSERT INTO evaluations VALUES(12930,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(12930,33,'cp 235',NULL);
INSERT INTO evaluations VALUES(12930,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(12930,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(12930,36,'cp 220','f6g8 c1f4 e7e6 b1c3 d8a5 e1c1 g8e7 c1b1 a7a6 g2g4 e7c6 d4e3 b7b5 f1g2 c8b7 d1d3 b5b4 c3e4 c6e5 e4d6 f8d6 g2b7 e5g4 e3d4 d6f4 b7a8 g4e5 d4f4 e5d3 c2d3 e8g8 a8f3 a5b5 f4c4 b5b6 d3d4 d7d5 c4d3 f8c8 h2h4 c8c4 h1d1 a6a5 d3e3 c4c8 h4h5 a5a4 d1c1 c8c1 b1c1 h7h6 f3d1 a4a3 b2a3 b4a3');
INSERT INTO evaluations VALUES(19607,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(19607,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(19607,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(19607,33,'cp 252',NULL);
INSERT INTO evaluations VALUES(19607,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(19607,35,'cp 273',NULL);
INSERT INTO evaluations VALUES(19607,36,'cp 246','b1c3 e7e6 c1f4 g8e7 e1c1 d8a5 c1b1 a7a6 g2g4 e7c6 d4e3 b7b5 f1g2 c8b7 d1d3 b5b4 c3e4 c6e5 e4d6 f8d6 g2b7 e5g4 e3f3 d6f4 f3f4 g4e5 b7a8 e5d3 c2d3 e8g8 a8f3 a5b5 f4c4 b5b6 d3d4 g7g6 h2h4 h7h5 f3e2 a6a5 c4c5 b6b7 d4d5 a5a4 h1c1 a4a3 d5e6');
INSERT INTO evaluations VALUES(49429,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(49429,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(49429,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(49429,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49429,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49429,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(49429,36,'cp 4','c7c5 f1g2 c5d4 f3d4 d7d5 e1g1 e6e5 d4b3 c8e6 e2e4 b8c6 e4d5 e6d5 g2d5 d8d5 b1c3 d5e6 c1g5 a8d8 d1f3 f8e7 f1d1 e8g8 d1d8 f8d8 a1d1 d8d1 f3d1 b7b6 g5f6 e7f6 c3d5 h7h6 c2c3 c6e7 c3c4 f6g5 h2h4 g5f6 g1g2 e5e4 d5f6 g7f6 d1d4 e7f5 d4d5 g8g7 d5e6');
INSERT INTO evaluations VALUES(44237,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44237,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44237,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(44237,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(44237,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44237,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44237,36,'cp 3','f1g2 c5d4 f3d4 d7d5 e1g1 e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 b7b6 d2e4 f8e7 e4c3 d5b4 d1d8 a8d8 a2a3 b4d5 c3d5 e6d5 c1e3 d5g2 g1g2 e8d7 a1c1 d8c8 f1d1 d7e6 c1c4 e7d6 d1c1 c6e7 c4a4 c8c1 e3c1 d6b8 a4g4 e7g6 c1e3 h8d8');
INSERT INTO evaluations VALUES(44238,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(44238,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44238,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44238,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44238,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(44238,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(44238,36,'cp 12','c5d4 f3d4 d7d5 e1g1 e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 f8e7 d2e4 b7b6 e4c3 d5b4 d1d8 a8d8 a2a3 b4d5 c3d5 e6d5 c1e3 d5g2 g1g2 f7f5 f1c1 d8c8 c1c2 e8d7 a1d1 d7e6 d1c1 e6d7 f2f4 h8d8 f4e5 c6e5 b3d4 c8c2 c1c2');
INSERT INTO evaluations VALUES(56961,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(56961,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(56961,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(56961,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(56961,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(56961,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(56961,36,'cp 80','d7d5 e4e5 c7c5 c2c3 b8c6 f1d3 c5c4 d3c2 b7b5 e1g1 b5b4 b2b3 g8e7 a2a3 c4b3 c2b3 b4c3 b1c3 c8d7 b3c2 c6a5 c3e2 a8c8 a1b1 c8b8 b1b8 d8b8 h2h3 e7c6 f1e1 b8b6 e2f4 a5c4 f4h5');
INSERT INTO evaluations VALUES(55472,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(55472,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(55472,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(55472,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(55472,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(55472,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(55472,36,'cp 80','e4e5 c7c5 c2c3 b8c6 f1d3 c5c4 d3c2 b7b5 e1g1 g8e7 a2a3 a7a5 b1d2 c8d7 b2b3 c4b3 d2b3 e7g6 c1d2 a5a4 b3c1 c6a5 c1e2 f8e7 g2g3 e8g8 h2h4 f8e8 g1g2 g6f8 f1h1 a5b3 a1a2 b3d2 d1d2 d8a5 a2a1 e8c8 c2d3 a5c7');
INSERT INTO evaluations VALUES(55473,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(55473,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(55473,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(55473,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(55473,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(55473,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(55473,36,'cp 80','c7c5 c2c3 b8c6 f1d3 c5c4 d3c2 b7b5 e1g1 b5b4 b2b3 c4b3 a2b3 b4c3 b1c3 a7a5 c1d2 f8b4 c3a4 c8a6 f1e1 g8e7 d2b4 c6b4 a4c5 e7c6 c2b1 d8b6 d1d2 a6b5 e1c1 b4a6 h2h4 a8c8 h4h5 c6b4 d2e3 c8c7 g2g4 c7c8 g1g2 a6c5');
INSERT INTO evaluations VALUES(53520,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(53520,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(53520,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(53520,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(53520,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(53520,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(53520,36,'cp 64','c2c3 b8c6 f1d3 d8c7 a2a3 c5c4 d3c2 c8d7 b1d2 e8c8 b2b4 c4b3 d2b3 f7f6 e1g1 f6e5 f3e5 c6e5 d4e5 c7e5 c1e3 c8b8 f1e1 e5c7 d1d4 b7b6 e3f4 f8d6 f4d6 c7d6 d4g7 g8e7 b3d4 e7c6 d4c6 d6c6 e1e3 c6d6 g7d4 h8f8 g2g3 d6e7');
INSERT INTO evaluations VALUES(53519,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(53519,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(53519,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(53519,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(53519,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(53519,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(53519,36,'cp 66','b8c6 f1d3 d8c7 a2a3 c5c4 d3c2 c8d7 b1d2 e8c8 b2b4 f7f6 d1e2 c8b8 a3a4 d7e8 e1g1 f6f5 d2b1 g7g5 b1a3 c7g7 a3b5 g8e7 b5d6 e7c8 d6e8 d8e8 c1d2 g5g4 f3e1 f8e7 g2g3 h6h5 b4b5 c6d8');
INSERT INTO evaluations VALUES(20136,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(20136,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(20136,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(20136,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(20136,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(20136,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(20136,36,'cp 76','f1d3 d8c7 e1g1 c5d4 c3d4 c6b4 d3b5 c8d7 b5d7 c7d7 b1c3 g8e7 c1d2 b4c6 c3e2 e7g6 h2h4 f8e7 g2g3 e7d8 g1g2 e8g8 a1c1 g6e7 h4h5 a8c8 d2f4 d8b6 f1h1 c6a5 g3g4 c8c1 f4c1 a5c6 c1f4 f7f6 e5f6 f8f6 f3e5 c6e5 f4e5 f6f7 e2f4 e7c6');
INSERT INTO evaluations VALUES(5814,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(5814,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(5814,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(5814,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(5814,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(5814,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(5814,36,'cp 85','c1g5 g8e7 e2e3 f7f6 g5f4 b4c3 b2c3 c6a5 c4d5 e6d5 f3d2 e8g8 f1e2 c8e6 e1g1 b7b6 f1e1 h7h6 d1a4 e6f7 e2d3 c7c6 f4g3 d8d7 a4c2 a8d8 a1b1 f8e8 f2f3 e7c8 e3e4 c6c5 d3b5');
INSERT INTO evaluations VALUES(5794,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(5794,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(5794,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(5794,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(5794,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(5794,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(5794,36,'cp 70','g8e7 e2e3 b4c3 b2c3 f7f6 g5f4 c6a5 c4d5 e6d5 f3d2 e8g8 f1e2 c8e6 e1g1 b7b6 f1e1 d8d7 a2a4 e6f7 h2h3 a8d8 e2d3 f7g6 d3a6 f8e8 f2f3 c7c5 d2b3 a5b3 d1b3 e7c6 a6b5 c6a5 b5d7 a5b3 d7e8');
INSERT INTO evaluations VALUES(52558,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(52558,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(52558,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(52558,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(52558,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(52558,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(52558,36,'cp 130','f1d3 g7g6 g1f3 f6e5 f3e5 f8g7 e1g1 g8f6 f1e1 e8g8 b1d2 c6c5 d4c5 d8c7 h2h3 c7c5 e5f3 b8c6 c2c4 c6b4 d3f1 c8f5 d2b3 c5c7 b3d4 d5c4 a2a3 b4d3 f1d3 c4d3 d4f5 g6f5 d1d3 f6e4 c1e3 c7d6 d3b3 g8h8 a1d1');
INSERT INTO evaluations VALUES(34981,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(34981,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(34981,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(34981,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(34981,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(34981,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(34981,36,'cp 168','f1c4 b8c6 e1g1 f8e7 d2d4 c8g4 c4f7 e8f7 d4d5 g8f6 d5c6 b7c6 d1d3 h7h6 d3c4 g4e6 c4a4 a5a4 f3e5 f7g8 c3a4 g8h7 e5c6 e7d6 c6d4 e6d7 a4c3 a7a5 f1e1 a5a4 d4b5 d6b4 e1e2 h8c8 c1f4 a4a3 b2a3 b4c3 b5c3 a8a3 e2e3');
INSERT INTO evaluations VALUES(34978,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(34978,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(34978,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(34978,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(34978,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(34978,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(34978,36,'cp 206','c8e6 c4e6 f7e6 e1g1 b8c6 f3g5 g8f6 g5e6 c6d4 e6d4 e5d4 f1e1 e8d7 c3e4 f8d6 e4d6 d7d6 d2d3 a8e8 c1f4 d6d7 f4g3 h8f8 a2a3 e8e1 d1e1 a5e1 a1e1 c7c5 e1e5 b7b6 g3h4 h7h6 h4f6 f8f6 e5e4 g7g5 g1f1 f6e6 f1e2 e6e4 d3e4 d7e6 f2f3 c5c4 b2b3 b6b5 b3c4 b5c4 a3a4 e6d6 g2g3');
INSERT INTO evaluations VALUES(42052,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(42052,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(42052,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(42052,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(42052,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42052,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(42052,36,'cp 25','c4d5 e6d5 c1g5 e8g8 e2e3 c7c5 d4c5 h7h6 g5h4 c8e6 g1f3 b8d7 f1d3 a8c8 e1g1 d7c5 c3e2 c5e4 c2b3 b4c5 h4f6 d8f6 e2f4 f8d8 a1c1 b7b6 f4e6 f6e6 d3a6 c8c7 f1d1 e4g5 a6e2 g7g6 a2a3 e6f6 f3g5');
INSERT INTO evaluations VALUES(41985,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41985,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(41985,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(41985,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(41985,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(41985,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41985,36,'cp 9','e6d5 c1g5 e8g8 e2e3 c7c5 d4c5 h7h6 g5h4 c8e6 g1f3 b8d7 f3d4 d7c5 a2a3 b4c3 c2c3 a8c8 f2f3 e6d7 c3d2 c5e6 d4e6 d7e6 f1d3 d5d4 e1g1 d4e3 d2e3 f8e8 a1d1 e6f5 d3e4 d8e7 e3a7 f5e4 f3e4 e7e4 h4f6 c8c2 f1f2 g7f6 h2h3');
INSERT INTO evaluations VALUES(42241,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(42241,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42241,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42241,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(42241,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(42241,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(42241,36,'cp 0','c1g5 h7h6 g5h4 c7c5 d4c5 e8g8 e2e3 c8e6 g1e2 b8d7 a2a3 b4c5 a1d1 a8c8 e2d4 a7a6 f1d3 d8b6 e1g1 c5d4 e3d4 g8h8 f2f3 b6d4 h4f2 d4f4 f1e1 d7e5 f2g3 f4d4 g3f2');
INSERT INTO evaluations VALUES(42224,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42224,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42224,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42224,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42224,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(42224,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(42224,36,'cp 11','e8g8 e2e3 h7h6 g5h4 c7c5 d4c5 c8e6 g1f3 b8d7 f3d4 d7c5 a2a3 b4c3 c2c3 a8c8 f2f3 e6d7 f1e2 c5e6 d4e6 f7e6 c3d2 f6e4 h4d8 e4d2 e1d2 f8d8 f3f4 g8f7 a1c1 f7e7 h2h4 e7d6 c1c8 d8c8 g2g4 c8e8 h1f1 e6e5 f4e5 e8e5 f1f7 e5e7 f7e7 d6e7 d2c3 b7b6 c3d4 e7d6 e2f3');
INSERT INTO evaluations VALUES(9743,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(9743,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(9743,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(9743,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(9743,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(9743,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9743,36,'cp 15','a7a5 e1g1 d7d6 b1d2 g7g5 h4g3 e8g8 c2c3 c5a7 f1e1 f6h7 h2h3 h6h5 d3d4 h5h4 g3h2 g5g4 h3g4 c8g4 c4e2 d8f6 g1h1 a8e8 f3g5 g4e2 g5h7 g8h7 d1e2 e5d4 f2f4 h7g7 d2f3 f8g8 a1d1 d4c3 b2c3 g7f8 e4e5 d6e5 f3e5 c6e5 f4e5 f6g6 e2f3');
INSERT INTO evaluations VALUES(12129,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(12129,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(12129,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(12129,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(12129,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(12129,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(12129,36,'cp 48','g1e2 c5d4 c3d4 f7f6 e5f6 d7f6 d2f3 f8d6 e1g1 e8g8 c1f4 d6f4 e2f4 f6e4 f4e2 c8d7 d3e4 d5e4 f3d2 d8b6 d2e4 a8d8 d4d5 c6e5 e2c3 d7b5 f1e1 e6d5 c3d5 b6h6 d1d4 e5d3 e1e3 d3f4 e4f6 g8h8 d4f4 h6f4 d5f4 f8f6 f4h5 f6h6 h5g3 d8d2 b2b3');
INSERT INTO evaluations VALUES(12130,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(12130,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(12130,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(12130,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(12130,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(12130,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(12130,36,'cp 23','c5d4 c3d4 f7f6 e5f6 d7f6 d2f3 f8d6 e1g1 e8g8 c1f4 d6f4 e2f4 f6e4 d1c1 e4g5 f3g5 d8g5 f4e2 g5c1 a1c1 c8d7 d3b1 a8c8 f1e1 g8f7 f2f4 g7g6 g1f2 f7e7 h2h3 e7d6 g2g4 a7a5 f2g3 b7b5 h3h4 b5b4 b1d3');
INSERT INTO evaluations VALUES(12140,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12140,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(12140,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(12140,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(12140,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(12140,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(12140,36,'cp 34','c3d4 f7f6 e5f6 d7f6 d2f3 f8d6 e1g1 e8g8 c1f4 d6f4 e2f4 f6e4 f4h5 c8d7 h5g3 e4g5 f3e5 c6e5 d4e5 g5f7 d1e2 d8h4 a1e1 a7a6 a2a3 f7h6 h2h3 d7c6 e2d2 h6f7 g1h2 a8d8 f2f4 f7h6 f4f5 h6g4 h2g1 h4g3 h3g4 g3g4 f5f6 f8f7 f1f4 g4g5 d3c2 g7f6 e5f6');
INSERT INTO evaluations VALUES(12137,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12137,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(12137,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(12137,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(12137,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(12137,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12137,36,'cp 26','f7f6 e5f6 d7f6 d2f3 f8d6 e1g1 e8g8 c1f4 d6f4 e2f4 f6e4 f4e2 c8d7 a1c1 h7h6 d3b1 d8b6 e2g3 a8d8 g3e4 d5e4 b1e4 d7e8 d1c2 c6d4 f3d4 d8d4 f1d1 d4d1 c1d1 e8c6 e4c6 f8c8 h2h3 b6c6 c2e2 c8f8 b2b3 a7a6 d1d4 c6c1 g1h2 c1c6 e2d2 c6c5 b3b4 c5e5 f2f4 e5f6 d4d7 f6f4 d2f4 f8f4 d7b7 f4f2');
INSERT INTO evaluations VALUES(12217,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(12217,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(12217,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(12217,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(12217,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(12217,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(12217,36,'cp 27','e5f6 d7f6 e1g1 f8d6 d2f3 e8g8 c1f4 d6f4 e2f4 f6e4 f4e2 c8d7 a1c1 h7h6 d3b1 d8b6 e2g3 a8d8 g3e4 d5e4 b1e4 d7e8 d1c2 c6d4 f3d4 d8d4 f1d1 d4d1 c1d1 e8c6 e4c6 f8c8 h2h3 b6c6 c2e2 c8f8 b2b3 f8f5 d1d8 g8f7 e2d2 f7g6 d8d4 f5g5 g2g3 g5d5 d4d5 c6d5');
INSERT INTO evaluations VALUES(12216,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(12216,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12216,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(12216,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(12216,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12216,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(12216,36,'cp 32','d7f6 e1g1 f8d6 d2f3 e8g8 c1f4 d6f4 e2f4 f6e4 f4e2 c8d7 a1c1 h7h6 d3b1 d8b6 e2g3 a8d8 g3e4 d5e4 b1e4 d7e8 d1c2 c6d4 f3d4 d8d4 f1d1 d4d1 c1d1 e8c6 e4c6 f8c8 h2h3 b6c6 c2e2 c8f8 b2b3 a7a6 a2a4 f8f5 d1d8 f5f8 d8d4 f8f5 d4c4 c6d7 e2c2 d7d5 c4c8 f5f8 c8c7');
INSERT INTO evaluations VALUES(14031,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(14031,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(14031,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(14031,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(14031,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(14031,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(14031,36,'cp 40','d2f3 f8d6 e1g1 e8g8 c1g5 d8b6 e2c3 h7h6 c3a4 b6a5 g5h4 e6e5 d4e5 c6e5 f3e5 d6e5 a4c3 e5c3 b2c3 f6e4 d1c2 c8f5 f2f3 a5c5 g1h1 e4d6 d3f5 d6f5 h4f2 c5c4 f1d1 a8e8 c2b3 f5e3 d1d2 a7a5 a1e1 a5a4 b3c4 e3c4 d2d5 e8e1 f2e1 f8e8 e1f2 e8e2');
INSERT INTO evaluations VALUES(14029,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(14029,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(14029,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14029,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14029,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14029,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14029,36,'cp 25','f8d6 e1g1 e8g8 c1g5 d8b6 e2c3 h7h6 c3a4 b6a5 g5h4 e6e5 d4e5 c6e5 f3e5 d6e5 a1c1 a5b4 h4f6 f8f6 d1h5 e5d4 a4c3 d4c3 c1c3 b4d6 h5e8 f6f8 e8b5 a7a6 b5b3 b7b5 d3c2 c8e6 c3e3 e6f5 c2f5 f8f5 f1d1 a8f8');
INSERT INTO evaluations VALUES(8421,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(8421,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(8421,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(8421,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(8421,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(8421,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(8421,36,'cp 33','c1g5 e8g8 e1g1 d8b6 e2c3 h7h6 c3a4 b6a5 g5h4 e6e5 d4e5 c6e5 f3e5 d6e5 a1c1 a5b4 h4g3 e5g3 h2g3 c8g4 d1b3 b4b3 a2b3 f8c8 f2f3 c8c1 f1c1 a8c8 c1a1 g4d7 b3b4 d7a4 a1a4 c8a8 g1f2 g8f7 f2e3 f7e6 e3d4 f6h5 g3g4 h5f4 g2g3');
INSERT INTO evaluations VALUES(52501,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(52501,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(52501,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(52501,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(52501,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(52501,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(52501,36,'cp -70','d7d5 e4d5 d8d5 g1f3 e7e6 d2d4 g8f6 f1e2 d5d8 e1g1 b8c6 a3b4 f8b4 c2c3 b4d6 b1d2 e8g8 d2c4 d6e7 f3e5 c6e5 c4e5 d8c7 c3c4 b7b6 c1f4 e7d6 e2f3 a8b8 f4g5 d6e5 d4e5 f8d8 e5f6 d8d1 f1d1 a7a5 f6g7');
INSERT INTO evaluations VALUES(53351,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(53351,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(53351,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(53351,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(53351,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(53351,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(53351,36,'cp -61','e4d5 d8d5 g1f3 e7e6 a3b4 f8b4 b1c3 d5d8 c3e4 g8f6 e4f6 d8f6 a1b1 f6e7 f1b5 b8d7 f3e5 a7a5 c2c3 b4d6 d2d4 e8g8 e5c4 f8d8 e1g1 b7b6 d1f3 c8a6 f3c6 a6b5 b1b5 d7f6 c4b6 a8a6 c1e3 a5a4 c3c4 e7c7 c6c7 d6c7 c4c5 a4a3 f1a1 e6e5 b5b4 a3a2 f2f3 f6d5 b6d5 d8d5 g2g3 e5d4 e3d4');
INSERT INTO evaluations VALUES(22678,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(22678,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(22678,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(22678,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(22678,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(22678,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(22678,36,'cp 21','d7d6 e2e4 d6e5 d1d8 e8d8 f4e5 c6e5 g1f3 f8d6 f3e5 d6e5 b1a3 g8f6 a3c4 h8e8 c4e5 e8e5 f1d3 f6e4 e1g1 f7f5 c1f4 e5e8 a1e1 a7a5 g2g4 a5a4 g4f5 c8f5 f4c7 d8c7 f1f5 e4f6 e1e8 a8e8 f5a5 e8e1 g1g2 e1b1 b2b3 a4b3 a2b3 g7g6 a5a8 f6d5 g2f3 d5b4 d3e4 b4c6');
INSERT INTO evaluations VALUES(21887,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(21887,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(21887,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(21887,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(21887,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(21887,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(21887,36,'cp 20','e2e4 d6e5 d1d8 e8d8 g1f3 f8d6 f4e5 c6e5 f3e5 d6e5 b1d2 g8f6 d2c4 h8e8 c4e5 e8e5 f1d3 f6e4 e1g1 d8e8 c1f4 e5e7 a1e1 e4d6 d3h7 g7g6 f4g5 e7e6 g2g4 a7a5 h2h4 a5a4 g5d2 d6c4 h7g8 c4d6 h4h5 e8f8 g8f7 d6f7 e1e6 c8e6 h5g6 f8g8 g6f7 e6f7 a2a3 a8e8 f1f4 e8e2');
INSERT INTO evaluations VALUES(53970,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(53970,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(53970,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53970,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(53970,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(53970,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53970,36,'cp 9','g1e2 a7a6 a2a4 g8f6 d2d4 c5d4 e2d4 f8e7 e1g1 b8c6 c1e3 e8g8 h2h3 c8d7 f2f4 d8c7 d1e2 a8d8 c4d3 c6d4 e3d4 d7c6 e2f2 f6d7 b2b3 e7f6 c3e2 e6e5 f4e5 f6e5 a1d1 g8h8 f2h4 h7h6 d4e5 d7e5 e2g3');
INSERT INTO evaluations VALUES(15695,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(15695,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(15695,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(15695,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(15695,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(15695,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(15695,36,'cp -16','f8d6 h2h3 h7h6 d2d4 d6b4 f3e5 f6e4 e1g1 b4c3 b2c3 e8g8 d1f3 e4d6 a2a4 f8e8 a4a5 a7a6 f1e1 f7f6 e5d3 e8e1 d3e1 c8e6 e1d3 e6d5 f3g4 d8c8 g4c8 a8c8 c1a3 d6c4 a3b4 c8e8 a1e1 e8e1 d3e1 b7b5 e1d3 g7g5 f2f3');
INSERT INTO evaluations VALUES(9518,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(9518,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(9518,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(9518,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(9518,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(9518,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(9518,36,'cp -15','h2h3 h7h6 d2d4 d6b4 f3e5 b4c3 b2c3 f6e4 e1g1 e8g8 d1d3 e4d6 c3c4 f8e8 c1f4 f7f6 e5g4 c8f5 d3b3 f5e6 b3d3 d8d7 f4d6 d7d6 f1e1 d6d7 g4e3 b7b6 a1d1 a8d8 d4d5 e6f7 e3f5 g8h8 f5e3 c6d5 c4d5 e8e5 c2c4 d8e8');
INSERT INTO evaluations VALUES(18579,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(18579,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(18579,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(18579,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(18579,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(18579,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(18579,36,'cp 52','g8f6 e4f6 e7f6 f1d3 f8d6 d1e2 d8e7 e2e7 d6e7 c1e3 d7b6 c2c4 c8e6 b2b3 e8c8 e1g1 h8e8 a1c1 e7a3 c1b1 c8b8 f3e1 a3f8 e1c2 b6c8 a2a4 h7h5 f1d1 h5h4 h2h3 c8d6 c4c5 d6f5 b3b4 f5e3 f2e3 g7g6 b4b5');
INSERT INTO evaluations VALUES(12253,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(12253,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(12253,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12253,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(12253,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(12253,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12253,36,'cp 43','e4f6 e7f6 f1d3 f8e7 c2c3 e8g8 d1c2 h7h6 e1g1 f8e8 c1e3 d7f8 f3d2 f8e6 f1e1 e7d6 d3h7 g8h8 h7e4 e6f4 c3c4 g7g6 e3f4 d6f4 g2g3 f4d2 c2d2 h8g7 d4d5 c6d5 e4d5 c8e6 a1d1 e6d5 e1e8 d8e8 c4d5 a8d8 h2h4 e8d7 d5d6');
INSERT INTO evaluations VALUES(12223,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(12223,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(12223,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(12223,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(12223,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(12223,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(12223,36,'cp 35','e7f6 f1d3 f8d6 d1e2 d8e7 e2e7 d6e7 c1e3 e8g8 a2a4 f8d8 h2h3 d7f8 e1c1 h7h5 b2b3 e7a3 c1d2 f8g6 c2c4 c8e6 d2c3 a3d6 d3e4 g6e7 f3e1 a7a5 e4f3 h5h4 c3b2 g7g5 e1c2 d8d7 h1e1 a8d8 b2c3');
INSERT INTO evaluations VALUES(35282,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(35282,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(35282,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(35282,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(35282,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(35282,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(35282,36,'cp 103','g8f6 f1c4 c7c6 c4b3 d7d6 d2d4 c8e6 e1g1 b8d7 h2h3 e6b3 a2b3 e7d8 d1e2 d8c7 f1d1 a7a6 c1g5 f8e7 d4d5 c6d5 g5f6 d7f6 c3d5 f6d5 d1d5 e8g8 c2c3 g7g6 f3d2 a8d8 b3b4 h7h5 d2f1 e7g5 a1d1 b7b5 f1e3 c7c6');
INSERT INTO evaluations VALUES(34047,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(34047,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(34047,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(34047,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(34047,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(34047,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(34047,36,'cp 110','f1c4 c7c6 c4b3 d7d6 d2d4 c8e6 e1g1 e6b3 a2b3 e7c7 f3h4 b8d7 g1h1 g7g6 f2f4 e5d4 d1d4 f8e7 f4f5 e8g8 c1h6 f8e8 h6g5 c7d8 h4f3 f6g4 g5f4 a7a5 a1d1 e7f8 h2h3 g4e5 d4f2 b7b5 f3e5 d7e5 f4e5 e8e5 f5g6 f7g6');
INSERT INTO evaluations VALUES(34039,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(34039,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(34039,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(34039,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(34039,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(34039,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(34039,36,'cp 90','c7c6 c4b3 d7d6 d2d4 c8e6 c1e3 g7g6 d1d3 b8a6 b3e6 e7e6 e3g5 a6c7 d4d5 c6d5 g5f6 e6f6 c3d5 f6d8 d5c3 f8e7 h2h4 h7h5 g2g3 a7a6 a2a4 b7b5 e1g1 b5b4 c3d1 a6a5 d1e3 e8g8 f3d2 c7e6 e3d5 g8h7 d2c4 a8c8 c2c3 b4c3 b2c3 e6c5');
INSERT INTO evaluations VALUES(33567,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(33567,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(33567,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(33567,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(33567,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(33567,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(33567,36,'cp 107','c4b3 d7d6 d2d4 c8e6 c1e3 g7g6 d1d3 b8a6 b3e6 e7e6 e3g5 a6c7 d4d5 c6d5 g5f6 e6f6 c3d5 f6d8 d5c3 d8d7 e1g1 d7c6 f3d2 f8h6 d2c4 e8g8 a2a4 c7e6 b2b3 e6d4 a1d1 a8d8 c3d5 g8h8 c4a5 c6c2 d3c2 d4c2 a5b7 d8b8 b7a5');
INSERT INTO evaluations VALUES(38222,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38222,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(38222,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(38222,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(38222,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(38222,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(38222,36,'cp 11','g8f6 e1g1 e8g8 f1e1 c7c5 c1g5 b8c6 b1c3 f6d7 g5e7 c6e7 a2a4 d7b6 a4a5 b6c4 d3c4 e7c6 h2h3 c6a5 c3b5 a5c6 d1d6 d8d6 b5d6 f7f6 e1d1 f8d8 d6c8 a8c8 d1d5 b7b6 g1f1 c6b4 d5d8 c8d8 a1a7 b4c2 f1e2 c2d4 f3d4 d8d4 a7a8 g8f7');
INSERT INTO evaluations VALUES(38807,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38807,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(38807,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(38807,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(38807,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(38807,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(38807,36,'cp 34','d5c4 g1f3 a7a6 e2e3 b7b5 a2a4 c8b7 a4b5 b4c3 d2c3 a6b5 a1a8 b7a8 b2b3 g8f6 b3c4 b5c4 f1c4 e8g8 e1g1 b8d7 c4e2 a8d5 d1c2 d8a8 f1a1 a8c6 f3e1 f8b8 f2f3 d5b3 c2d2 b3c4 e1d3 c4d3 e2d3 c6b6 a1b1 b6a7 b1b8 a7b8 d3e2 c7c5 d4c5 d7c5');
INSERT INTO evaluations VALUES(38880,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(38880,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(38880,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(38880,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(38880,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38880,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38880,36,'cp 25','e2e3 c7c5 g1f3 c5d4 f3d4 b4e7 f1c4 g8f6 a1c1 e8g8 e1g1 c8d7 d4f3 b8c6 c4d3 d8c7 c3e4 f8d8 e4f6 e7f6 d2c3 f6c3 c1c3 d7e8 d1b3 c7b6 b3b6 a7b6 a2a3 c6e7 f1d1 h7h6 c3c1 g8f8 f3d4 e7c6 d4b5 f8e7 b5c7');
INSERT INTO evaluations VALUES(38875,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(38875,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38875,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(38875,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(38875,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38875,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(38875,36,'cp 23','c7c5 g1f3 c5d4 f3d4 b4e7 f1c4 g8f6 a1c1 c8d7 d4f3 e8g8 e1g1 d7c6 c4e2 b8d7 f3d4 a8c8 d4c6 c8c6 c3b5 d8b6 d2c3 f6d5 c3d4 e7c5 a2a3 c5d4 b5d4 c6c1 d1c1 d5e7 c1b1 f8c8 b2b4 d7f6 f1c1 c8c1 b1c1');
INSERT INTO evaluations VALUES(50297,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(50297,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(50297,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(50297,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50297,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50297,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(50297,36,'cp 12','b6d5 f1c4 e7e6 d2d4 b7b6 c4d5 e6d5 b1c3 b6c5 c3d5 c7c6 d5c3 d7d5 e5d6 f8d6 d4c5 d8e7 g1e2 d6c5 e1g1 e8g8 c1f4 b8d7 f1e1 f8e8 c3a4 c8a6 a4c5 d7c5 f4d6 e7g5 d6c5 g5c5 e2c3 e8e1 d1e1 a8d8 a2a3 c5b6 a1d1 d8d1 c3d1 a6b5 h2h3');
INSERT INTO evaluations VALUES(2118,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(2118,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(2118,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(2118,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(2118,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(2118,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(2118,36,'cp 45','c1e3 a7a6 d1f3 b7b5 f3g3 c7g3 h2g3 c6d4 e3d4 c8b7 f2f3 g8e7 g3g4 e7c6 d4b6 c6e5 e1c1 b7c6 b6d4 e5g6 c1b1 h7h6 c3e2 a8c8 d4c3 f7f6 e2d4 c6b7 d4b3 f8e7 b3a5 b7a8 a2a4 b5a4 f1a6 c8b8 a5c4 a8c6 c4d6 e7d6 d1d6 b8a8 a6c4 g6f4 c4f1 e8e7 d6d1');
INSERT INTO evaluations VALUES(2115,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(2115,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(2115,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(2115,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(2115,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(2115,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(2115,36,'cp 47','a7a6 d1f3 c6d4 e3d4 b7b5 e1c1 g8e7 f3g3 c7g3 h2g3 c8b7 f2f3 e7c6 d4e3 a8c8 a2a4 b5b4 c3e2 c6a5 e2f4 a5b3 c1b1 f8c5 c2b3 c5e3 f4d3 e8e7 d3b4 a6a5 b4c2 e3c5 f1b5 c8c7 g3g4 h7h6 d1d3 d7d5 e4d5 b7d5 h1d1 h6h5 g4h5 h8h5 c2e3 c5e3');
INSERT INTO evaluations VALUES(1260,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(1260,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(1260,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(1260,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(1260,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(1260,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(1260,36,'cp 42','d1f3 c6d4 e3d4 b7b5 f3g3 c7g3 h2g3 c8b7 f2f3 g8e7 e1c1 e7c6 d4e3 f8b4 g3g4 f7f6 g4g5 f6g5 e3g5 h7h5 g5e3 a8c8 c3e2 h5h4 e2f4 g7g5 f4d3 b4e7 d3c5 e7c5 e3c5 c6e5 c5e3 e5f7 e3d4 h8h7 d4f2 e8e7 g2g3 c8h8 g3h4 g5h4 f1h3 d7d6 h1g1 f7e5');
INSERT INTO evaluations VALUES(52362,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(52362,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(52362,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(52362,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(52362,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52362,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52362,36,'cp 18','g8f6 d2d4 d7d5 e2e3 c8f5 f3h4 f5e4 b1c3 e7e6 f2f3 e4g6 d1b3 d8c7 c1d2 d5c4 f1c4 f8e7 h4g6 h7g6 e3e4 b7b5 c4d3 b8d7 d2e3 h8h2 e1f2 f6h5 c3e2 e7h4 f2g1 h2h1 g1h1 e8c8 a1c1 h5g3 h1g1 d7f6');
INSERT INTO evaluations VALUES(9081,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(9081,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(9081,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(9081,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(9081,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(9081,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(9081,36,'cp 57','e4d5 b7b5 c4b5 d8d5 e1g1 c5f2 f1f2 d5b5 h2h3 b5b6 b1c3 c6d4 c3a4 b6c6 c2c3 d4e6 c3c4 e8g8 g5e6 c8e6 b2b3 f8e8 c1g5 f6d7 d3d4 e5d4 d1d4 f7f6 g5f4 e6f7 a1d1 e8e7 a4c3 d7b6 f2e2 e7d7 d4f2 d7d1 c3d1 a8e8 e2e8 f7e8 f4e3 e8g6 f2d2 b6d7 e3a7');
INSERT INTO evaluations VALUES(9068,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(9068,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(9068,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(9068,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(9068,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(9068,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(9068,36,'cp 44','b7b5 c4b5 d8d5 e1g1 c5f2 f1f2 d5b5 h2h3 b5b6 b1c3 c6d4 c3a4 b6c6 c2c3 d4e6 c3c4 e6d4 b2b4 e8g8 c1b2 f8e8 b2d4 e5d4 b4b5 c6d6 d1f3 c8d7 f3f4 d6f4 f2f4 a7a6 a4c5 a6b5 c5d7 f6d7 c4b5 d7b6 f4e4 e8e4 g5e4 a8a5 a1c1 a5b5 c1c7');
INSERT INTO evaluations VALUES(12597,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12597,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12597,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12597,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12597,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12597,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12597,36,'cp 0','g1e2 f8c5 e2g3 e8g8 f1c4 d7d6 e1g1 f6g4 h2h3 d8h4 d1f3 c6d4 f3d3 c8e6 h3g4 e6c4 d3c4 d4f3 g2f3 h4g3 g1h1 g3h3 h1g1');
INSERT INTO evaluations VALUES(15502,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(15502,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(15502,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(15502,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(15502,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(15502,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(15502,36,'cp 145','f3g5 d6d5 e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5d3 h7h6 g5e4 f6e4 d3e4 f8d6 e4f3 e8g8 d2d3 f7f5 b1d2 d8c7 h2h3 c8e6 b2b3 a5b7 f1e1 e6d5 c1b2 a8d8 d1e2 f8e8 f3h5 g7g6 h5f3 b7c5 e2e3 g8h7 a1d1 c5d7 a2a3 d6c5 e3e2 a7a5 d2c4 d5c4 b3c4 d8b8');
INSERT INTO evaluations VALUES(15476,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(15476,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(15476,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(15476,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(15476,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(15476,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(15476,36,'cp 130','d6d5 e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5d3 f8c5 b1c3 e8g8 f1e1 a5b7 h2h3 h7h6 g5f3 f8e8 b2b3 c5d6 c1b2 d8c7 c3e4 f6e4 d3e4 b7c5 d2d4 c5e4 d4e5 d6e5 b2e5 e8e5 d1d4 e5d5 d4e4 c8f5 e4e2 a8d8 a1d1 d5d1 e1d1 d8d1 e2d1 c7a5');
INSERT INTO evaluations VALUES(48008,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48008,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(48008,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(48008,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48008,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(48008,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48008,36,'cp -13','d6d5 c4b3 e4c5 e1g1 c5b3 a2b3 f8d6 b1c3 c8e6 c3b5 b8c6 h2h3 e8g8 d2d4 a7a6 c1g5 f7f6 b5d6 d8d6 g5h4 d6d7 h4g3 f8e8 f1e1 e6f7 e1e8 a8e8 f3e1 f7g6 c2c3 g6f5 b3b4 c6d8 f2f3 c7c6 g3f2 e8e7 e1d3 d8e6 d3c5');
INSERT INTO evaluations VALUES(49031,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(49031,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(49031,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(49031,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(49031,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(49031,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(49031,36,'cp -12','c4b3 e4c5 e1g1 c5b3 a2b3 f8e7 c2c3 e8g8 d2d4 c8e6 c1f4 a7a5 b3b4 b8c6 b4b5 c6a7 a1a5 b7b6 a5a4 a7b5 f1e1 b5d6 a4a8 d8a8 b1a3 a8b7 f4d6 e7d6 f3g5 e6d7 d1d3 g7g6 a3b5 d6f4 g5h3 d7f5 d3f3 f4h6 h3f4');
INSERT INTO evaluations VALUES(49038,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(49038,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(49038,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(49038,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(49038,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49038,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(49038,36,'cp -7','e4c5 e1g1 c5b3 a2b3 f8e7 d2d4 e8g8 c2c3 c8e6 c1f4 a7a5 b3b4 b8c6 b4b5 c6a7 a1a5 b7b6 a5a4 a7b5 f3e5 a8a4 d1a4 d8e8 f1e1 c7c6 b1d2 e8c8 d2f1 f7f6 e5d3 f8e8 a4d1 e6g4 f2f3 g4h5 f1e3 c8d7 h2h4 h7h6 f4g3 h5f7 d3f4');
INSERT INTO evaluations VALUES(48181,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(48181,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(48181,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(48181,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(48181,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(48181,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(48181,36,'cp -26','e6d5 f1d3 f8d6 g1e2 e8g8 e1g1 f8e8 b1c3 b7b6 e2g3 c8b7 b2b3 b8d7 g3f5 d6f8 c1b2 a7a6 a1c1 g7g6 f5g3 f8d6 c3e2 d7f8 e2f4 f8e6 d1d2 f6g4 h2h3 g4f6 a2a3 h7h5 f4e6 e8e6 b3b4 c7c6 g3e2 f6e4 d2e1');
INSERT INTO evaluations VALUES(37460,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(37460,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(37460,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(37460,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(37460,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(37460,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(37460,36,'cp 31','c2c4 c7c6 b1c3 d7d5 d1b3 d8b6 e1g1 c8f5 h2h3 b8a6 b3b6 a7b6 c4d5 f6d5 c3d5 c6d5 c1d2 a6b8 f1c1 b8c6 g3g4 f5e4 e2e3 f8c8 f3e1 e4g2 g1g2 e7e5 d4e5 g7e5 e1d3 e5g7 a2a4 c6a5 d2a5 b6a5 c1c5 c8c5 d3c5 b7b6 c5d7 g7b2 a1a2 b2c3 d7b6 a8d8');
INSERT INTO evaluations VALUES(37409,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(37409,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(37409,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(37409,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(37409,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37409,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(37409,36,'cp 19','c7c6 d1b3 d8b6 e1g1 d7d5 b1c3 c8f5 h2h3 b8a6 b3b6 a7b6 c4d5 f6d5 c3d5 c6d5 c1d2 a6b8 g3g4 f5e4 e2e3 b8c6 f1c1 e7e6 f3e1 g6g5 c1c3 c6a5 g2e4 d5e4 a1c1 f8d8 c3c7 g7e5 c7c2 e5d6 f2f3 f7f5 d2a5 b6a5 a2a4 e4f3 e1f3 h7h6 g1f2 g8f7 g4f5 e6f5');
INSERT INTO evaluations VALUES(35694,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(35694,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(35694,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35694,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(35694,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(35694,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(35694,36,'cp 18','b1c3 d7d5 d1b3 d8b6 e1g1 c8f5 h2h3 b8a6 b3b6 a7b6 c4d5 f6d5 c3d5 c6d5 c1d2 a6b8 f1c1 b8c6 g3g4 f5e4 e2e3 f8c8 f3e1 e4g2 g1g2 e7e5 d4e5 g7e5 c1c2 e5g7 a2a4 d5d4 a1a3 d4e3 d2e3 c6e7 a3a1 c8c2 e1c2 g7b2 a1b1 b2c3');
INSERT INTO evaluations VALUES(35691,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35691,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(35691,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35691,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(35691,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(35691,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(35691,36,'cp 18','d7d5 d1b3 d8b6 e1g1 f8d8 h2h3 h7h6 c1d2 d5c4 b3c4 c8f5 g3g4 f5e6 c4a4 e6d5 c3d5 c6d5 f1c1 a7a5 e2e3 b8c6 c1c2 h6h5 g4h5 f6h5 a1c1 g7f6 a2a3 h5g7 g2f1 e7e5 d4e5 c6e5 f3e5 f6e5');
INSERT INTO evaluations VALUES(35875,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(35875,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(35875,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(35875,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(35875,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(35875,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(35875,36,'cp 25','c4d5 c6d5 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 a8a5 a1c1 f8c8 e1g1 h7h6 h2h4 f6e4 c3d5 a5d5 e5c6 b7c6 g3g4 f5g4 g2e4 d5a5 a2a3 g4d7 f1d1 g6g5 f4h2 g5h4 e4f3 c6c5 d4c5 g7b2 d1d7 b2c1 c5b6');
INSERT INTO evaluations VALUES(35858,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(35858,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(35858,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35858,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(35858,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(35858,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(35858,36,'cp 6','c6d5 f3e5 c8f5 e1g1 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 a1c1 c3c1 f1c1 d8b6 e2e3 b6b2 a4c6 a7a5 c6c3 b2c3 c1c3 a5a4 a2a3 f8a8 g2f1 f7f6 f4c7 g7f8 f1b5 g6g5 f2f3 g5g4 f3g4 f5g4 g1f2');
INSERT INTO evaluations VALUES(36038,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36038,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36038,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36038,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36038,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36038,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36038,36,'cp 19','f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 e7e6 e1g1 f8c8 f1c1 c6e5 f4e5 f6e4 g2e4 f5e4 c3e4 d5e4 e5f4 b6b5 g3g4 f7f6 a2a4 c8c1 a1c1 b5a4 c1c7 b7b5 c7b7 h7h5 g4h5 g6h5 b7b5 f6f5 b5b7 g7f6 g1g2 a8c8 b7a7 c8c2 a7a4');
INSERT INTO evaluations VALUES(33945,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33945,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33945,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33945,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(33945,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(33945,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(33945,36,'cp 5','d2d4 b8c6 c1e3 e5d4 f3d4 c8d7 d4c6 d5c6 d1d4 f8e7 e1c1 e7f6 d4c4 e8g8 h1g1 f8e8 g2g3 h7h6 c4c6 d7c6 f1c4 e8e5 e3d4 e5e4 d4f6 e4c4 f6d4 c6f3 d1d3 a8d8 d4c3 d8d3 c2d3 c4a4 a2a3 b7b6 g1e1 f7f6 e1e7');
INSERT INTO evaluations VALUES(51808,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(51808,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(51808,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(51808,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(51808,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(51808,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(51808,36,'cp 51','d7d5 d2d4 f8b4 e4e5 g8e7 a2a3 b4c3 b2c3 h7h6 a3a4 c6c5 h2h4 b7b6 f1b5 c8d7 b5d3 b8c6 h4h5 c5d4 c3d4 c6b4 c1a3 b4d3 d1d3 a8c8 e1g1 e8g8 a4a5 b6a5 a3e7 d8e7 a1a5 c8c7 f1a1 f8a8 g2g4 d7e8 g1g2 c7c8 c2c3 e7c7 g4g5 c7c3 d3c3 c8c3 a5a7 a8a7 a1a7 h6g5');
INSERT INTO evaluations VALUES(53619,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(53619,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(53619,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(53619,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(53619,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(53619,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(53619,36,'cp 63','d2d4 f8b4 e4e5 g8e7 a2a3 b4c3 b2c3 c6c5 a3a4 h7h6 c1a3 b8d7 f1e2 b7b6 a4a5 e8g8 e1g1 b6a5 e2d3 f8e8 f1e1 c8b7 d4c5 d8c8 c5c6 e7c6 d1e2 a5a4 a3d6 a7a5 a1a4 b7a6 e2d2 a6d3 c2d3 c6e7 f3d4');
INSERT INTO evaluations VALUES(54370,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(54370,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(54370,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54370,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54370,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(54370,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(54370,36,'cp 18','b8c6 c1f4 f8d6 f4d6 d8d6 f1d3 g8e7 e1g1 e8g8 f1e1 c8d7 d1d2 a7a6 c3e2 f7f6 c2c4 d5c4 d3c4 e7d5 e2c3 d5c3 b2c3 a8c8 c4d3 g8h8 d3c2 d7e8 c2b3 e8f7 a1d1 b7b5 d2e3 f8e8 f3d2 d6c7 d2e4 c6e7 e4c5');
INSERT INTO evaluations VALUES(20390,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(20390,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20390,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(20390,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(20390,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20390,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(20390,36,'cp 21','c1f4 f8d6 f4d6 d8d6 f1d3 g8e7 e1g1 a7a6 f1e1 e8g8 d1d2 f7f6 a1d1 c8d7 c3a4 e6e5 d4e5 c6e5 f3e5 f6e5 a4c3 d7c6 d2e2 d6c5 e2e5 c5f2 g1h1 e7g6 e5g5 f2f4 g5f4 g6f4 d3f1 a8d8 h1g1 d5d4 c3e4 h7h6 e4c5 a6a5 c5d3 f4d3');
INSERT INTO evaluations VALUES(21860,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(21860,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(21860,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(21860,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(21860,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(21860,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(21860,36,'cp 0','e5f4 d2d3 d8f6 h2h4 g8e7 d1d2 a7a6 b5a4 g7g6 d2f4 f6f4 c1f4 f8g7 c2c3 f7f5 b1d2 f5e4 d2e4 h8f8 f4d2 h7h6 e1g1 c8e6 d3d4 e8c8 e4g3 e6c4 f1e1 c4f7 e1f1 f7c4');
INSERT INTO evaluations VALUES(21869,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(21869,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(21869,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(21869,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(21869,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(21869,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(21869,36,'cp 0','d2d3 d8f6 h2h4 g8e7 d1d2 a7a6 b5a4 c8e6 d2f4 e8c8 b1c3 h7h6 f4h2 d6d5 c1f4 d5e4 d3e4 e7g6 f4c7 f8a3 a4c6 a3b2 c3a4 b2c3 e1e2 e6c4 e2e3 c3d4 f3d4 f6d4 e3f3 d4f6 f3e3');
INSERT INTO evaluations VALUES(51519,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51519,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(51519,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(51519,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(51519,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(51519,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(51519,36,'cp 2','c2c4 g8f6 g1f3 c8b7 d2d4 e7e6 f1d3 d7d5 e1g1 d5c4 d3c4 a7a6 b1d2 b8d7 b2b3 c7c5 c1b2 f8e7 a2a4 e8g8 d1e2 c5d4 b2d4 e7d6 a1d1 d6c5 d4b2 d8e7 e3e4 f8d8 c4d3 c5a3 b2a3 e7a3 e4e5 f6d5 d2e4 a3b3 e4d6');
INSERT INTO evaluations VALUES(51483,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51483,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(51483,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(51483,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(51483,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51483,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(51483,36,'cp 2','g8f6 b2b3 c8b7 c1b2 g7g6 g1f3 f8g7 g2g3 c7c5 f1g2 e8g8 e1g1 d7d5 d2d4 d5c4 b3c4 f6e8 d1e2 e8d6 f1c1 c5d4 e3d4 e7e6 d4d5 e6d5 b2g7 g8g7 e2b2 d8f6 b2f6 g7f6 c4d5 f8d8 f3d4 b8d7 b1d2 d7c5 d4c6 d8e8 d2c4 d6c4');
INSERT INTO evaluations VALUES(12459,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(12459,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(12459,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(12459,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(12459,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12459,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(12459,36,'cp 19','b1c3 d7d5 e4d5 e6d5 c4b3 f8c5 d1e2 e8f8 d4c6 b7c6 c1d2 d8e7 e2e7 f8e7 c3a4 c5d6 e1c1 h8e8 d2e3 e7f8 h2h3 h7h5 c2c4 c8f5 h1e1 d6b4 a4c3 f5e6 b3a4 e8c8 a2a3 b4c3 b2c3 d5c4 e3c5 f8g8 d1d6 e6d5 a4c2 d5g2');
INSERT INTO evaluations VALUES(19899,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(19899,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(19899,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(19899,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(19899,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(19899,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(19899,36,'cp 86','e7e5 b1c3 g8f6 f1b5 f8d6 d2d3 h7h6 e1g1 e8g8 h2h3 a7a6 b5c6 b7c6 f1e1 f8e8 c1d2 a6a5 c3e4 f6e4 d3e4 d6c5 a2a3 f7f6 b2b4 a5b4 a3b4 a8a1 d1a1 c5f8 e1b1 d8b6 d2e3 b6b5 a1c3 c8e6 e3c5 f8c5 c3c5 b5c4 f3d2 c4c5 b4c5 e8a8 b1b6 a8a1 g1h2 a1a2 b6c6 a2c2');
INSERT INTO evaluations VALUES(20562,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(20562,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(20562,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(20562,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(20562,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(20562,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(20562,36,'cp 99','b1c3 g8f6 f1b5 f8d6 d2d3 h7h6 e1g1 e8g8 h2h3 a7a6 b5c6 b7c6 f1e1 f8e8 c3e4 f6e4 d3e4 a6a5 c1d2 a5a4 d1e2 d8e7 d2c3 c8a6 e2d2 a6b5 e1d1 d6c7 d2e3 c6c5 a2a3 e8d8 d1d8 a8d8 f3d2 d8d7 b2b3 e7d6 d2c4 b5c4 b3c4 d7d8 e3f3');
INSERT INTO evaluations VALUES(20560,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(20560,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(20560,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(20560,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(20560,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(20560,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(20560,36,'cp 92','g8f6 f1b5 f8d6 d2d3 h7h6 e1g1 e8g8 h2h3 a7a6 b5c6 b7c6 f1e1 f8e8 c3e4 f6e4 d3e4 a6a5 c1d2 a5a4 d2c3 f7f6 f3h4 c8e6 h4f5 d6f8 d1h5 g8h7 f5h4 d8c7 h5g6 h7h8 a2a3 f8c5 a1d1 a8d8 h4f5 c7a7 d1d2 e8g8 g6g3 d8d2 c3d2 e6f5 e4f5');
INSERT INTO evaluations VALUES(22900,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(22900,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(22900,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(22900,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(22900,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(22900,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(22900,36,'cp -6','g1f3 f8b4 c2c3 d4c3 e1g1 g8f6 e4e5 d7d5 e5f6 d5c4 f6g7 h8g8 d1d8 c6d8 b1c3 b4c3 b2c3 f7f6 f3d4 g8g7 f1e1 e8f7 c1f4 d8e6 d4e6 c8e6 f4c7 a8g8 g2g3 g7g5 a1d1 g5f5 d1d2 b7b6 e1e3 f5c5 c7f4 c5d5 d2d5 e6d5 f2f3 h7h5 g1f2 g8d8 g3g4 h5g4 f3g4 d5e6 g4g5 f6g5 f4g5');
INSERT INTO evaluations VALUES(42234,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(42234,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(42234,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42234,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(42234,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(42234,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42234,36,'cp 11','c1g5 e8g8 e2e3 h7h6 g5h4 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 f6e4 h4d8 e4c3 b2c3 f8d8 e1d2 d7f6 h1c1 c7c5 d4c5 f6e4 d2e1 e4c5 f3d4 f5d7 c3c4 d5c4 c1c4 b7b6 a1c1 g7g6 c4c3 a8c8 f2f3 c8c7 h2h4 g8g7 a2a3 d7a4 e1f2 d8c8');
INSERT INTO evaluations VALUES(42220,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42220,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(42220,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42220,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(42220,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(42220,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(42220,36,'cp 0','h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 f6e4 h4d8 e4c3 d8e7 f8e8 b2c3 e8e7 e1g1 c7c5 f1c1 a8c8 h2h3 d7b6 a2a4 b6c4 e2c4 d5c4 c1d1 c5d4 d1d4 b7b6 a1d1 f5c2 d1a1');
INSERT INTO evaluations VALUES(41663,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41663,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(41663,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41663,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(41663,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(41663,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(41663,36,'cp 4','g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 f6e4 h4d8 e4c3 b2c3 f8d8 e1d2 d7f6 h1c1 c7c5 d4c5 f6e4 d2e1 e4c5 f3d4 f5d7 f2f3 b7b6 e1d2 a8c8 a1b1 c8c7 e2b5 d7b5 b1b5 c5a4 b5b4 a4c5 h2h4 d8c8');
INSERT INTO evaluations VALUES(44659,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(44659,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(44659,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(44659,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(44659,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(44659,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44659,36,'cp 4','c5d4 g1f3 b8c6 f1c4 d5b6 c4b3 d7d6 e5d6 d8d6 e1g1 c8e6 f3d4 c6d4 c3d4 e6b3 a2b3 g7g6 b1c3 f8g7 d1f3 e8g8 f3b7 f8b8 b7e4 d6d4 e4e7 b6d5 c3d5 d4d5 a1a7 a8a7 e7a7 d5b3 a7e7 b3b7 e7c5 g7b2 c1b2 b7b2 c5e7 b2b3 e7e5 b3b2 e5e4 b2b4 e4b4 b8b4');
INSERT INTO evaluations VALUES(17174,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(17174,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(17174,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(17174,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(17174,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(17174,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(17174,36,'cp 20','e7g6 b5a4 f8b4 c2c3 b4e7 d3d4 d7d5 f3e5 g6e5 d4e5 d5e4 d1d8 e7d8 c1f4 g7g5 f4e3 e8g8 a4c6 b7c6 b1d2 f8e8 d2e4 e8e5 e1c1 c8e6 e4d2 h7h6 h1e1 d8e7 a2a3 e5d5 e3d4 a7a5 d2e4 a8d8 f2f3 g5g4 d4f6 d5d1 e1d1 d8d1 c1d1 e7d6 e4d6 c7d6 f3g4 e6g4 d1c2');
INSERT INTO evaluations VALUES(22715,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(22715,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(22715,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(22715,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(22715,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(22715,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(22715,36,'cp -73','e5f4 d2d4 d8h4 e1f1 d7d6 g1f3 c8g4 f3h4 g4d1 b1c3 d1c2 c1f4 g8f6 a1c1 c6d4 h4f3 d4e6 f4e3 c2e4 c3e4 f6e4 c4e6 f7e6 c1c7 f8e7 c7b7 e7f6 f1e2 e8g8 h1c1 d6d5 e3d4 f6d4 f3d4 f8f2 e2e3 f2g2 d4e6 e4d6 b7d7 d6c4 e3f3 g2b2 d7d5 a8c8 f3g3 b2e2 d5d4 e2e3 g3f2 e3e6 d4c4 c8f8 f2g1 e6e2 c4c2 e2e4 c2c7 e4a4');
INSERT INTO evaluations VALUES(52368,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(52368,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(52368,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(52368,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(52368,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(52368,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(52368,36,'cp 31','d7d5 e2e4 d5e4 c3e4 g8f6 e4f6 e7f6 c2c3 f8d6 f1d3 e8g8 d1c2 h7h6 h2h3 f8e8 g1e2 b8d7 e1g1 d7f8 c1e3 c8e6 e2f4 d6f4 e3f4 a7a5 a2a4 f8g6 d3g6 f7g6 c2g6 e6f7 g6c2 d8d7 f1e1 e8e1 a1e1 a8e8 e1e8 d7e8 c3c4 e8d7 c2c3 b7b5 c4b5 c6b5 a4b5');
INSERT INTO evaluations VALUES(26918,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(26918,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(26918,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(26918,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(26918,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(26918,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(26918,36,'cp 90','h2h3 g4f3 d1f3 a7a6 b5c6 b7c6 d2d4 e5d4 c2c3 d4c3 f3c3 d8d7 f1e1 c6c5 b1a3 g8e7 c1e3 e7g6 a1d1 f8e7 e3c5 d7c6 c5d4 c6c3 d4c3 f7f6 a3c2 e8f7 c2e3 g6f4 d1d4 f4e6 d4c4 h8b8 e3d5 c7c5 e1d1 a8a7 c4a4 e7d8 b2b3 a7a8 f2f4 a6a5');
INSERT INTO evaluations VALUES(26917,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(26917,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(26917,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(26917,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(26917,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(26917,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(26917,36,'cp 71','g4f3 d1f3 a7a6 b5c6 b7c6 f3c3 c6c5 d2d3 g8f6 b2b4 c5b4 c3c6 f6d7 a2a3 b4a3 b1c3 f8e7 c3d5 a8c8 c1a3 e8g8 c6a6 f8e8 a3b4 d7f6 d5f6 e7f6 a6c6 e8e7 a1a6 d8d7 f1a1 c8b8 g1f1 h7h5 b4d2 g8h7 c6c4 d7e6 c4e6 f7e6 d2e3 e7f7 g2g3 b8b2 a6c6 h5h4');
INSERT INTO evaluations VALUES(26922,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(26922,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(26922,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(26922,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(26922,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(26922,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(26922,36,'cp 89','d1f3 a7a6 b5c6 b7c6 d2d4 e5d4 c2c3 d4c3 f3c3 d8d7 f1e1 g8f6 e4e5 d6e5 c3e5 f8e7 c1g5 e8f8 b1c3 h7h6 g5f4 e7d6 e5d4 f8g8 a1d1 a8b8 d4a7 g8h7 f4d6 c7d6 a7d7 f6d7 d1d6 d7c5 b2b3 h8c8 e1e7 h7g8 e7e5 c5e6 c3e4 b8b4 e4c5 e6c5 e5c5 b4e4');
INSERT INTO evaluations VALUES(26921,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(26921,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(26921,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(26921,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(26921,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(26921,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(26921,36,'cp 91','a7a6 b5c6 b7c6 d2d4 e5d4 c2c3 d4c3 f3c3 d8d7 f1e1 g8f6 e4e5 d6e5 c3e5 f8e7 c1g5 e8f8 b1c3 h7h6 g5f4 e7d6 e5d4 f8g8 a1d1 a8b8 d4a7 g8h7 f4d6 c7d6 a7d7 f6d7 d1d6 d7c5 b2b3 h8c8 e1e7 h7g8 e7e5 c5e6 e5a5 b8a8 g2g3 a8a7');
INSERT INTO evaluations VALUES(53981,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(53981,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(53981,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(53981,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(53981,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(53981,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(53981,36,'cp -41','d2d3 b8c6 g1e2 g8f6 a2a4 f8e7 c4b3 c6a5 b3a2 a7a6 b1c3 e8g8 e1g1 a5c6 f1d1 c8d7 f3g3 g8h8 c1g5 c6b4 a2b3 d7c6 a4a5 d6d5 g5d2 b7b5 a5b6 d8b6 e4d5 e6d5 d1e1 a8e8 c3a2 a6a5 a2b4 a5b4');
INSERT INTO evaluations VALUES(53975,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(53975,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(53975,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(53975,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(53975,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(53975,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(53975,36,'cp -44','g8f6 g1e2 f8e7 e1g1 b8c6 c4b3 e8g8 a2a4 g8h8 b1c3 a7a6 f1d1 c8d7 d3d4 c5d4 e2d4 b7b5 f3d3 b5b4 c3e2 d8c7 d4f3 h8g8 e2g3 a6a5 d3e2 d7c8 b3c4 c8b7 h2h3 a8c8 c1g5 h7h6 g5e3 f8d8 a1c1 d6d5 e4d5 f6d5 c4d5 d8d5 d1d5');
INSERT INTO evaluations VALUES(45997,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(45997,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(45997,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(45997,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(45997,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(45997,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(45997,36,'cp -36','g1e2 f8e7 e1g1 e8g8 a2a4 b8c6 c4b3 g8h8 b1c3 a7a6 f1d1 c8d7 d3d4 c5d4 e2d4 b7b5 f3d3 b5b4 c3e2 d8c7 d4f3 h8g8 e2g3 f8d8 d3e2 d7c8 c1f4 a6a5 b3c4 c8b7 b2b3 c6e5 f3e5 d6e5 f4d2');
INSERT INTO evaluations VALUES(55554,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(55554,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(55554,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(55554,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(55554,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55554,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55554,36,'cp 14','g1e2 c7c5 b1c3 b8c6 d2d3 a8b8 g2g3 a7a6 f1g2 b7b5 e1g1 f8e7 b2b3 g8f6 c3d5 e8g8 c1d2 b5c4 b3c4 h7h6 d5e3 c6d4 e2c3 c8d7 f2f4 b8b2 d2c1 b2b7 f4f5 d7c6 h2h4 d8a5 d1d2 f8b8 g3g4 f6h7 c3d5 a5d2');
INSERT INTO evaluations VALUES(26995,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(26995,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(26995,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(26995,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(26995,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(26995,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(26995,36,'cp 121','f1b5 e5d4 d1d4 g8e7 d4e3 a7a6 b5a4 g4d7 a4b3 c6a5 c1d2 a5b3 a2b3 e7c6 c3d5 f8e7 d2c3 e8g8 e3f4 f7f6 f4g3 b7b5 e1g1 a6a5 h2h4 b5b4 c3d2 f6f5 f1e1 c6e5 d5e7 d8e7 e4f5 e5f3 g3f3 e7f6 g2g4 f6h4 g1g2 c7c5 e1h1 h4f6 g2g3 f6f7');
INSERT INTO evaluations VALUES(26904,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(26904,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(26904,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(26904,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(26904,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(26904,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(26904,36,'cp 106','e5d4 d1d4 g8e7 d4e3 a7a6 b5a4 g4d7 h2h4 c6a5 b2b4 a5c4 e3e2 d7a4 e2c4 d8d7 c1b2 a4b5 c3b5 d7b5 c4b3 e7c6 a2a3 b5h5 e1c1 f8e7 b3d5 h5g4 f3g5 e8g8 f2f3 g4g2 d1g1 e7g5 c1b1 g2d2 g1g5 d2d5 g5g7 g8h8 g7g1 f7f6 e4d5 c6e5 f3f4 e5c4');
INSERT INTO evaluations VALUES(28426,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(28426,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(28426,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(28426,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(28426,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(28426,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(28426,36,'cp 67','f1e2 g4f3 g2f3 f6d7 f3f4 b8c6 e4e5 e7e6 h2h4 h7h5 c1e3 d7b6 d1c2 c6e7 e1c1 d6d5 c4c5 b6d7 c1b1 c7c6 e2d3 e7f5 c2b3 b7b6 d3f5 g6f5 c3e2 g7f8 d1c1 d8c7 b3a4 a8b8 b1a1 b6c5 d4c5 c7b7 b2b3 a7a6 h1g1 b7b5 a4b5 a6b5 c1c2 b8a8 b3b4 a8a3');
INSERT INTO evaluations VALUES(28425,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(28425,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(28425,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(28425,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(28425,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(28425,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(28425,36,'cp 62','g4f3 g2f3 f6d7 f3f4 b8c6 e4e5 e7e6 h2h4 d7b6 c1e3 h7h5 b2b3 c6e7 e2d3 a7a5 e1f1 e8g8 d1e2 d8d7 a2a4 a8d8 a1d1 d6d5 c4c5 b6c8 c3b5 e7c6 h1h3 c8e7 e3d2 f7f5 d2c3 g7h6 e2d2 g8h7 d3e2 c6b4 c3b4 a5b4 a4a5');
INSERT INTO evaluations VALUES(28430,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(28430,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(28430,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(28430,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(28430,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(28430,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(28430,36,'cp 60','g2f3 f6d7 f3f4 b8c6 e4e5 e7e6 h2h4 h7h5 c1e3 d7b6 d1c2 c6e7 e2d3 d8d7 c3e2 d6d5 c4c5 b6c4 b2b3 c4a5 e2g3 a5c6 a2a3 a7a5 e1f1 e8g8 c2e2 e7f5 g3f5 e6f5 d3b5 d7e6 h1g1 c6e7 e2d2 g8h7 a3a4 c7c6 b5e2 f7f6 b3b4 f8h8 e5f6');
INSERT INTO evaluations VALUES(47249,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(47249,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47249,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(47249,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(47249,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(47249,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47249,36,'cp 0','c8f5 g1f3 e7e6 c2c4 c7c6 f3h4 f5g4 f1e2 g4e2 d1e2 b8d7 e1g1 f8d6 b1c3 e8g8 a1d1 d6b4 h4f3 d8e7 d2d4 a7a5 f3d2 d5c4 b3c4 e6e5 d2f3 e5d4 f3d4 a5a4 d4f5 e7e6 f5d4');
INSERT INTO evaluations VALUES(29601,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29601,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29601,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(29601,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29601,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(29601,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29601,36,'cp 0','g1f3 e7e6 c2c4 h7h6 b1c3 c7c6 d2d4 f8b4 a2a3 b4d6 f1d3 e8g8 e1g1 b8d7 a3a4 d8e7 c3e2 f8e8 a1c1 f5d3 d1d3 a8c8 e2g3 a7a6 f1d1 c8d8 d1f1');
INSERT INTO evaluations VALUES(29602,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29602,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29602,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29602,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29602,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29602,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29602,36,'cp 0','e7e6 c2c4 h7h6 d2d4 c7c6 b1c3 f8b4 a2a3 b4d6 f1d3 e8g8 e1g1 b8d7 c3e2 f8e8 c4c5 f5d3 d1d3 d6c7 b3b4 d8e7 f3e5 a7a5 f2f3 d7f8 d3c2 f8d7');
INSERT INTO evaluations VALUES(30138,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(30138,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30138,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30138,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30138,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30138,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30138,36,'cp 0','c2c4 h7h6 d2d4 f8b4 b1c3 c7c6 a2a3 b4d6 f1d3 e8g8 e1g1 b8d7 d3f5 e6f5 d1c2 g7g6 a3a4 a7a5 c4d5 f6d5 c3d5 c6d5 b2a3 d6a3 a1a3 a8c8 c2b2 d8b6 a3a1 c8c6 f1c1 f8c8 h2h3 d7f6 f3e5 c6c1 a1c1 c8c1 b2c1 g8g7 c1c2 f6e4 e5d3 h6h5');
INSERT INTO evaluations VALUES(19169,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(19169,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(19169,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(19169,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(19169,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(19169,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(19169,36,'cp 143','g8f6 b1c3 d7d6 f1b5 c8d7 c1f4 f8e7 d1d3 a7a6 b5d7 d8d7 e1c1 a8d8 f4g5 e8g8 d3g3 d7c8 f2f3 c8e6 f3f4 g8h8 h1e1 h7h6 h2h4 f8e8 g5f6 e7f6 c3d5 b7b5 c1b1 d8c8 g3f3 f6h4 g2g3 h4f6 g3g4 g7g6 f3g3 h8g7 e1h1 f6d4 d1e1 b5b4 g3h4 e8h8 h1h3 c8e8 g4g5');
INSERT INTO evaluations VALUES(12657,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(12657,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(12657,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(12657,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(12657,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(12657,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(12657,36,'cp 159','b1c3 d7d6 f1b5 c8d7 c1f4 f8e7 d1d3 a7a6 b5d7 d8d7 e1c1 a8d8 f4g5 e8g8 d3g3 f8e8 h1e1 d7g4 f2f3 g4g3 h2g3 b7b5 g5f6 e7f6 c3d5 e8e6 g3g4 f6g5 c1b1 d8b8 c2c3 a6a5 a2a4 b5a4 b1c2 e6h6 d1a1 h6h2 e1e2 g7g6 c2d3 g5d8 d3c4 h7h5 g4h5 g6h5 a1a4 h2h1 a4a2 h1d1 d5f4 d1h1 e4e5 d6e5 e2e5');
INSERT INTO evaluations VALUES(12652,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(12652,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(12652,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(12652,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(12652,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(12652,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(12652,36,'cp 160','d7d6 c1f4 f8e7 f1b5 c8d7 d1d3 e8g8 b5d7 d8d7 e1c1 d7e6 h1g1 a8d8 g2g4 f6d7 g4g5 d7e5 f4e5 c5c4 d3e3 d6e5 c3d5 b7b6 c1b1 e7c5 e3e2 b6b5 h2h4 b5b4 h4h5 c4c3 b2b3 g8h8 h5h6 g7g6 g1f1 a7a5 f2f4 e5f4 f1f4');
INSERT INTO evaluations VALUES(14012,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(14012,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(14012,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(14012,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(14012,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(14012,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(14012,36,'cp 172','c1f4 f8e7 f1b5 c8d7 d1d3 e8g8 b5d7 d8d7 e1c1 f8d8 h2h4 a8b8 h4h5 b7b5 h5h6 g7g6 d3e3 d7e6 f2f3 a7a5 c1b1 b5b4 c3d5 f6d5 d1d5 a5a4 e3d3 a4a3 b2b3 d8e8 g2g4 e6f6 f4c1 e7f8 g4g5 f6e6 d3c4 e6e7 h1d1 e7e6 c1f4 b8c8 c4d3 f7f6 f4d6 f8d6 d5d6 e6e5 d3d5 e5d5 d1d5 f6g5 d5g5 g8h8 d6d7');
INSERT INTO evaluations VALUES(29995,30,'cp 297',NULL);
INSERT INTO evaluations VALUES(29995,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(29995,32,'cp 291',NULL);
INSERT INTO evaluations VALUES(29995,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(29995,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(29995,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(29995,36,'cp 260','g5e6 f7e6 c4e6 b8c6 d2d3 d8e7 e6h3 g7g6 b1c3 f8g7 c1e3 e8g8 e1g1 c6d4 g2g3 c7c6 e3d4 e5d4 c3e2 f6d7 f2f4 d7c5 d1d2 b7b5 a1e1 c5a4 b2b3 a4c3 e2c3 d4c3 d2e3 b5b4 d3d4 a7a5 h3g4 g8h8 g1h1 c6c5 d4c5');
INSERT INTO evaluations VALUES(29994,30,'cp 259',NULL);
INSERT INTO evaluations VALUES(29994,31,'cp 277',NULL);
INSERT INTO evaluations VALUES(29994,32,'cp 256',NULL);
INSERT INTO evaluations VALUES(29994,33,'cp 274',NULL);
INSERT INTO evaluations VALUES(29994,34,'cp 274',NULL);
INSERT INTO evaluations VALUES(29994,35,'cp 256',NULL);
INSERT INTO evaluations VALUES(29994,36,'cp 251','f7e6 c4e6 b8c6 d2d3 d8e7 e6h3 g7g6 a2a4 c6d8 a4a5 c7c6 c1e3 d8f7 b1d2 f8h6 e3h6 f7h6 d1e2 h6f7 e2e3 e8g8 g2g3 g8g7 e1g1 h7h5 h3g2 f6g4 e3e2 a8e8 h2h3 g4f6 d2f3 g6g5 h3h4 g5g4 f3d2 g7h8 c2c3 e7d7');
INSERT INTO evaluations VALUES(30273,30,'cp 244',NULL);
INSERT INTO evaluations VALUES(30273,31,'cp 244',NULL);
INSERT INTO evaluations VALUES(30273,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(30273,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(30273,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(30273,35,'cp 259',NULL);
INSERT INTO evaluations VALUES(30273,36,'cp 255','c4e6 b8c6 d2d3 d8e7 e6h3 g7g6 b1a3 f8g7 c2c3 c6d8 a3c4 e8g8 g2g3 d8f7 e1g1 c7c6 h3g2 e7d7 a2a4 h7h5 f2f4 e5f4 g3f4 a8e8 c4e3 f7h6 d1f3 d6d5 e4e5 f6g4 d3d4 g4e3 f3e3 g8h8 g1h1 h6f5 e3h3');
INSERT INTO evaluations VALUES(30268,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(30268,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(30268,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(30268,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(30268,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(30268,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(30268,36,'cp 251','b8c6 d2d3 d8e7 e6h3 g7g6 b1c3 f8g7 c1e3 e8g8 e1g1 a8e8 a2a4 b7b6 g2g3 g8h8 c3b5 a7a6 b5a3 c6d8 c2c3 d8f7 a4a5 b6b5 b2b4 g7h6 e3h6 f7h6 c3c4 b5c4 a3c4 h6f7 d1d2 f7d8 g1h1 d8e6');
INSERT INTO evaluations VALUES(6551,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(6551,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(6551,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6551,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6551,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6551,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6551,36,'cp 10','d7d6 c3d4 c5b6 b1c3 c6a5 c4d3 g8e7 h2h3 e8g8 a1b1 a5c6 d3c2 e7g6 a2a4 c6e7 c1d2 c7c6 c3e2 c6c5 d2c3 b6a5 c3a5 d8a5 f1e1 h7h6 d1d3 e7c6 b1b5 a5a6 d4c5 d6c5 b5c5 a6d3 c2d3 f8d8 c5c3');
INSERT INTO evaluations VALUES(6040,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(6040,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(6040,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(6040,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(6040,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(6040,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(6040,36,'cp 24','c3d4 c5b6 h2h3 c6a5 c4d3 g8e7 b1c3 e8g8 a1b1 h7h6 f1e1 e7g6 c3d5 a5c6 d3c2 c6e7 d5e7 g6e7 d1d3 f8e8 c1f4 e7g6 f4g3 c7c6 d3a3 b6c7 b1d1 b7b5 c2b3 d8f6 d4d5 c8h3 d5c6 h3g4 d1d6 c7d6 a3d6 f6d6');
INSERT INTO evaluations VALUES(6038,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(6038,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(6038,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(6038,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(6038,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(6038,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(6038,36,'cp 14','c5b6 h2h3 c6a5 c4d3 g8e7 b1c3 e8g8 a1b1 c8d7 f1e1 a5c6 d3c2 e7g6 c3d5 g6h4 e4e5 d7f5 f3h4 f5c2 d1c2 d8h4 e5d6 c7d6 d5b6 a7b6 b1b6 c6d4 c2b1 h4h5 b6b7 d4e2 g1f1 e2c1 e1c1 d6d5 b7b5 f8e8 f1g1');
INSERT INTO evaluations VALUES(5779,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5779,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5779,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(5779,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(5779,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(5779,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(5779,36,'cp 16','b1c3 c6a5 c4d3 g8e7 h2h3 e8g8 a1b1 h7h6 c3e2 a5c6 f1e1 e7g6 d3c2 c8d7 d1d3 g6h4 e1d1 h4f3 d3f3 d8h4 c1e3 c6e7 e2g3 g8h8 g1h2 a8e8 g3h5 e7g6 a2a4 e8a8 h5g3 a8e8 g3f5 h4f6 f3h5 d7f5 e4f5 g6e7 a4a5 b6a5 b1b7 a5b6');
INSERT INTO evaluations VALUES(15570,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(15570,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(15570,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(15570,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(15570,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(15570,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(15570,36,'cp 56','f1b5 f8e7 d2d4 e5d4 f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 e1g1 f6d7 f2f4 e8g8 c1d2 e7f6 d1f3 f8e8 g1h1 d7c5 a1e1 d8d7 b2b4 c5e6 e1d1 e6d4 f3d3 d7e7 d1e1 d4e6 c3d5 c6d5 e4d5 e7d7 d5e6 e8e6 e1e6 d7e6 c2c4 a8e8 f1e1 e6d7');
INSERT INTO evaluations VALUES(15422,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15422,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(15422,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(15422,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(15422,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(15422,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(15422,36,'cp 40','c8d7 d2d4 e5d4 f3d4 g7g6 b5c6 b7c6 d1e2 f8g7 e4e5 d6e5 e2e5 d8e7 e5e7 e8e7 c1g5 h7h6 g5f6 g7f6 e1c1 e7f8 h1e1 f8g7 c3e4 f6e7 d4f3 h8d8 f3e5 d7e8 e5d3 e7f8 e1e3 d8d5 e4c3 d5d4 d1e1 e8d7 d3e5 a8d8 e5d7 d4d7 e3e4 a7a5 e1e2 f8d6');
INSERT INTO evaluations VALUES(25502,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(25502,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(25502,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(25502,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(25502,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(25502,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(25502,36,'cp 57','d2d4 e5d4 f3d4 g7g6 d4c6 d7c6 c1g5 f8g7 e4e5 d6e5 b5c6 b7c6 d1f3 d8d6 a1d1 d6e6 c3b5 f6d5 d1d5 c6b5 d5b5 e8g8 f3b3 e6b3 b5b3 f7f5 e1e2 f5f4 h1d1 h7h6 g5e7 f8f7 e7c5 e5e4 d1d5 a8e8 a2a4 a7a5 e2d2 g6g5 b3b7 g7e5 c5a7 f4f3 g2g4 c7c6');
INSERT INTO evaluations VALUES(39648,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(39648,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(39648,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(39648,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(39648,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(39648,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(39648,36,'cp 16','d2d4 d7d5 g1f3 c7c6 c2c4 g8f6 c4d5 c6d5 b1c3 e8g8 e1g1 f6e4 f3e5 c8f5 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 e5c6 b7c6 e2e3 a8c8 e4f3 d8b6 d1e2 f7f5 b2b3 c6c5 c1b2 f5g4 f3g4 c5d4 b2d4 g7d4 e3d4 e6g4 e2g4 c8d8 a1d1 d8d5');
INSERT INTO evaluations VALUES(39600,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(39600,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(39600,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(39600,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(39600,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(39600,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(39600,36,'cp 15','d7d5 g1f3 c7c6 c2c4 g8f6 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 h7h6 h2h4 e7e6 e1g1 f8c8 f1c1 c6e5 f4e5 g6g5 h4g5 h6g5 a2a4 c8c6 f2f3 g5g4 e3e4 g4f3 e4f5 f3g2 f5e6 f7e6 g1g2');
INSERT INTO evaluations VALUES(38470,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38470,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38470,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(38470,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(38470,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38470,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38470,36,'cp 22','g1f3 c7c6 c2c4 g8f6 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1e3 d8b6 e1g1 b6b2 c3d5 b8c6 d5f6 g7f6 d1c1 b2c1 a1c1 c6e5 d4e5 f6e5 g2b7 a8b8 b7d5 a7a6 d5b3 e5d6 f2f3 f8c8 c1c8 f5c8 g1f2 c8d7 f1d1 a6a5 e3f4 d6f4');
INSERT INTO evaluations VALUES(41359,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(41359,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(41359,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(41359,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(41359,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(41359,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(41359,36,'cp 48','e8g8 e1g1 e7e5 d4e5 d6e5 f3e5 f6e4 c3e4 d8d1 e2d1 g7e5 f1e1 a7a5 c2c3 b8d7 a2a4 e5g7 c1f4 d7e5 d1e2 b7b6 e2b5 c8a6 b5a6 a8a6 e1e2 f8e8 a1e1 e8f8 e1d1 a6a8 e4g3 f8e8 d1d5 e5c4 b2b3 c4d6 c3c4 e8e2 g3e2 a8d8 g1f1 g7f6 f4e3 g8g7');
INSERT INTO evaluations VALUES(36661,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(36661,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(36661,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(36661,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(36661,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(36661,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(36661,36,'cp 44','c1e3 e7e5 d4e5 d6e5 d1d3 b8c6 e1c1 d8d3 e2d3 c8g4 d3e2 f8d8 d1d8 a8d8 h2h3 g4f3 e2f3 c6d4 h1d1 a7a6 c3a4 f6d7 f3g4 d7f6 c2c3 f6g4 h3g4 d4e6 d1d8 e6d8 a4c5 g7f8 c5d7 f8d6 c1c2 d8e6 a2a4 g8g7 g4g5 d6e7');
INSERT INTO evaluations VALUES(36653,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(36653,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(36653,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(36653,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(36653,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(36653,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(36653,36,'cp 28','b8d7 h2h3 e7e5 d4e5 d6e5 d1d3 d8e7 d3c4 c7c6 a2a4 a7a5 c4b3 d7c5 b3a3 b7b6 e1g1 f8e8 e3c5 b6c5 e2c4 a8b8 a1d1 f6d7 c3b1 d7b6 b1d2 b6c4 d2c4 c8a6 f3d2 g7h6 f1e1 h6d2 c4d2 c5c4 a3e7 e8e7 d2f3 b8b2 d1d8 g8g7 d8a8 b2b6 e1d1 h7h6 d1d6');
INSERT INTO evaluations VALUES(47323,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(47323,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(47323,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(47323,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(47323,34,'cp -141',NULL);
INSERT INTO evaluations VALUES(47323,35,'cp -144',NULL);
INSERT INTO evaluations VALUES(47323,36,'cp -138','f1d3 b8c6 e1g1 c8g4 f2f3 g4d7 b1c3 e7e6 c1e3 c6b4 d3c4 d7c6 d1e2 f8d6 e3f2 e8g8 g5e4 c6e4 c3e4 f6e4 e2e4 c7c6 a1d1 d8c7 g2g3 b4d5 g1h1 a8d8 e4e2 d8d7 c4d3 d6e7 c2c4 d5f6 d3b1 f8d8 d1d7 d8d7 f1d1 d7d1 e2d1 g7g6 h1g2 b7b6 f2d4 b6b5 d4f6 e7f6');
INSERT INTO evaluations VALUES(47322,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(47322,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(47322,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(47322,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(47322,34,'cp -152',NULL);
INSERT INTO evaluations VALUES(47322,35,'cp -147',NULL);
INSERT INTO evaluations VALUES(47322,36,'cp -159','b8c6 e1g1 c8g4 f2f3 g4d7 b1c3 e7e6 c1e3 c6b4 d3c4 d7c6 d1e2 f8d6 e3f2 e8g8 g5e4 c6e4 c3e4 f6e4 e2e4 c7c6 a1d1 d8c7 g2g3 b4d5 c4d3 g7g6 e4e2 f8d8 g1h1 d6f8 c2c4 d5b4 d3b1 c6c5 g3g4 f8g7 f2h4 d8d1');
INSERT INTO evaluations VALUES(14358,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(14358,31,'cp -139',NULL);
INSERT INTO evaluations VALUES(14358,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(14358,33,'cp -158',NULL);
INSERT INTO evaluations VALUES(14358,34,'cp -156',NULL);
INSERT INTO evaluations VALUES(14358,35,'cp -151',NULL);
INSERT INTO evaluations VALUES(14358,36,'cp -141','e1g1 c8g4 f2f3 g4d7 b1c3 e7e6 c1e3 c6b4 d3c4 d7c6 d1e2 f8d6 e3f2 e8g8 g5e4 c6e4 c3e4 f6e4 e2e4 c7c6 g1h1 d8c7 a1d1 b4d5 g2g3 f8d8 c4d3 g7g6 e4e2 d6f8 c2c3 a7a5 a2a4 d8d7 d3e4 a8d8 d1d2');
INSERT INTO evaluations VALUES(14357,30,'cp -145',NULL);
INSERT INTO evaluations VALUES(14357,31,'cp -155',NULL);
INSERT INTO evaluations VALUES(14357,32,'cp -157',NULL);
INSERT INTO evaluations VALUES(14357,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(14357,34,'cp -159',NULL);
INSERT INTO evaluations VALUES(14357,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(14357,36,'cp -157','c8g4 f2f3 g4d7 b1c3 e7e6 c1e3 c6b4 d3c4 d7c6 d1e2 f8d6 e3f2 e8g8 g5e4 c6e4 c3e4 f6e4 e2e4 c7c6 a1d1 d8c7 g1h1 b4d5 g2g3 f8d8 f1e1 d6e7 d1d2 e7f6 c2c3 d8d7 e1d1 g7g6 e4e2 a8d8 c4d5 e6d5 f2a7 b7b6');
INSERT INTO evaluations VALUES(20248,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(20248,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(20248,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(20248,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20248,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(20248,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20248,36,'cp 17','e4d5 e6d5 c4b5 f8d6 b1c3 g8e7 d4c5 d6c5 c1e3 c5e3 f2e3 e8g8 d1d2 a7a6 b5d3 c8e6 e1c1 a8c8 c1b1 h7h6 h2h3 c6b4 d3e2 b7b5 g2g4 b4c6 e2d3 e7g6 h1f1 d8a5 f3d4 g6e5 d4c6 e5c6 c3e2 a5d2 d1d2 c6e5 e2f4');
INSERT INTO evaluations VALUES(20225,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(20225,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(20225,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(20225,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20225,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20225,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20225,36,'cp 0','e6d5 c4b5 f8d6 d4c5 d6c5 b1c3 g8e7 c1e3 c5e3 f2e3 e8g8 d1d2 a7a6 b5d3 c8e6 e1c1 d8b6 a2a3 a8d8 h1f1 g8h8 f3g5 h7h6 g5e6 f7e6 c3a4 b6a7 a4c3');
INSERT INTO evaluations VALUES(20509,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20509,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20509,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(20509,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(20509,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20509,36,'cp 0','c4b5 f8d6 d4c5 d6c5 e1g1 g8e7 b1d2 c8g4 h2h3 g4h5 d2b3 c5d6 c2c3 e8g8 f1e1 f8e8 b3d4 a7a6 b5e2 d6c7 c1e3 d8d6 g2g3 d6d7 e2f1 h5g6 d1d2 g6e4 f3g5 e4g6');
INSERT INTO evaluations VALUES(46308,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(46308,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46308,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(46308,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(46308,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(46308,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(46308,36,'cp 27','g1f3 f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 d8d6 b5c6 d6c6 f3e5 c6e6 g5f6 e6f6 f1e1 f6d6 d1f3 c8e6 e5d3 f8e8 a2a4 f7f6 d3c5 d6c6 e1b1 b7b6 c5e6 e8e6 a4a5 e6e4 f3d3 a8c8 a5b6 a7b6');
INSERT INTO evaluations VALUES(46425,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(46425,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(46425,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(46425,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(46425,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(46425,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(46425,36,'cp 56','c1g5 c5d4 f3d4 f8e7 e2e3 e8g8 f1e2 b8c6 e1g1 c8e6 a1c1 h7h6 g5f4 a8c8 h2h3 e7d6 f4d6 d8d6 a2a3 a7a6 d4c6 c8c6 d1d4 d6c5 c1d1 f8c8 e2f3 b7b5 d4f4 c5d6 f4d6 c6d6 d1d2 g8f8 f1d1 f8e7 f3d5');
INSERT INTO evaluations VALUES(46405,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(46405,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46405,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(46405,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(46405,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(46405,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(46405,36,'cp 47','c8e6 g5f6 d8f6 e2e4 d5e4 c3e4 f6f4 f1b5 b8c6 d1e2 c5d4 f3d4 f8b4 e1f1 e8g8 d4e6 f7e6 b5c6 b7c6 g2g3 f4e5 a1c1 a8d8 a2a3 b4a5 c1c2 a5b6 f1g2 d8d4 h1e1 e5d5 f2f3 d4d3 c2d2 d3d2 e4d2 e6e5 e1c1 a7a5 d2e4 b6d4 b2b4 a5b4 a3b4 g8h8 e2c4 f8d8 c4d5 c6d5');
INSERT INTO evaluations VALUES(29359,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(29359,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(29359,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(29359,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(29359,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(29359,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(29359,36,'cp 40','g5f6 d8f6 e2e4 b8d7 e4d5 e6g4 d4c5 f8c5 f1e2 e8g8 e1g1 a8d8 d1a4 g4h5 a1d1 a7a6 d1d2 f8e8 c3e4 f6b6 e4c5 d7c5 a4d4 e8e4 d4c3 e4e8 f1d1 c5e4 c3d4 b6d6 d2d3 e4f6 d3e3 h7h6 e2c4 h5f3 g2f3 b7b5 c4e2 e8e3 d4e3 g8h8 e3c3');
INSERT INTO evaluations VALUES(19449,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(19449,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(19449,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(19449,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(19449,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(19449,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(19449,36,'cp -6','e7e5 f1e2 f8e7 c3c4 d7d6 b1c3 a7a6 g1f3 a8b8 a2a4 f7f5 e4f5 c8f5 e1g1 g8f6 f3d2 e8g8 d2e4 g8h8 e2f3 c6d4 e4f6 d4f3 d1f3 f5d7 c3d5 e7f6 a4a5 f6h4 f3e2 d7c6 c1e3');
INSERT INTO evaluations VALUES(19010,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(19010,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(19010,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19010,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19010,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(19010,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(19010,36,'cp 0','c3c4 d7d6 b1c3 g8e7 g2g3 g7g6 f1g2 f8g7 a2a3 a7a6 g1e2 e8g8 e1g1 a8b8 a1b1 b7b5 c4b5 a6b5 b2b4 c5b4 a3b4 h7h6 c1e3 c6d4 h2h3 g8h7 b1a1 c8d7 e3d2 d4c6 c3d5 e7d5 e4d5 c6e7 e2c3 d7f5 d2e3 f5d7');
INSERT INTO evaluations VALUES(44083,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(44083,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(44083,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(44083,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(44083,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(44083,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(44083,36,'cp 135','f3e5 d7d5 e4d5 f8d6 d2d4 c6d5 c4b5 b8d7 e1g1 e8g8 c1f4 d8b6 a2a4 f6e4 b1d2 d7e5 d4e5 d6c5 d1e2 b6g6 d2e4 d5e4 f1e1 c8f5 b5c4 a8d8 c2c3 a7a6 h2h3 f5h3 e2e4 g6e4 e1e4 h3e6 c4f1 c5a7 b2b4 f8e8 a1e1 e6f5');
INSERT INTO evaluations VALUES(44064,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(44064,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(44064,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(44064,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(44064,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(44064,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(44064,36,'cp 139','d7d5 e4d5 f8d6 d2d4 c6d5 c4b5 b8d7 e1g1 e8g8 c1f4 f6e4 b5d3 d8f6 e5d7 c8d7 f4d6 f6d6 b1d2 d6f4 d3e4 d5e4 f1e1 a8e8 d2c4 f4c7 c4e5 d7a4 b2b3 f7f6 b3a4 f6e5 e1e4 e5d4 e4d4 c7c3 h2h3 h7h6 d4d3 c3c4 a4a5 e8c8 c2c3 c4c5 d1d2');
INSERT INTO evaluations VALUES(45794,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(45794,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(45794,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(45794,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(45794,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(45794,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(45794,36,'cp 149','e4d5 f8d6 d2d4 c6d5 c4b5 b8d7 e1g1 e8g8 c1f4 d8b6 b1c3 d6e5 d4e5 d5d4 b5d7 f6d7 c3e2 b6b2 e2d4 d7c5 f4e3 b7b6 f1e1 c8e6 a2a4 b2c3 e3g5 c3c4 a4a5 h7h6 g5e3 a8d8 a5b6 a7b6 f2f4 c4c3 a1a7 c5e4 d1a1 f8e8 a7a3');
INSERT INTO evaluations VALUES(45727,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(45727,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(45727,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(45727,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(45727,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(45727,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(45727,36,'cp 137','f8d6 d2d4 c6d5 c4b5 b8d7 e1g1 e8g8 c1f4 d8b6 b1c3 d6e5 d4e5 d5d4 e5f6 d7f6 a2a4 d4c3 b2c3 c8g4 d1d4 a8c8 d4b6 a7b6 f4e5 g4f5 e5d4 f6e4 a1a3 e4c5 f1c1 f8d8 f2f3 f5d7 b5d7 d8d7 c1b1 c8c6 b1e1 c6c8 a4a5');
INSERT INTO evaluations VALUES(55932,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(55932,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(55932,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55932,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55932,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55932,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55932,36,'cp 6','g8f6 g1f3 f8e7 c2c4 e8g8 b1c3 c7c5 c4d5 f6d5 f1e2 b8c6 e1g1 d5c3 b2c3 b7b6 d2d4 c5d4 f3d4 c6d4 c3d4 c8b7 e2f3 b7f3 d1f3 e7f6 f1d1 f6d4 d1d4 d8e7 g2g3 a8d8 a1d1 d8d4 d1d4 h7h6 f3e2 f8d8 e2b2');
INSERT INTO evaluations VALUES(17190,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(17190,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(17190,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(17190,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(17190,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(17190,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(17190,36,'cp 136','b1c3 c6d4 d1d4 e7c6 d4e3 f8b4 c1d2 d7d6 f2f4 e8g8 e1c1 c8g4 f1e2 b4c5 e3d3 c6b4 d3g3 g4e2 c3e2 d8d7 a2a3 b4c6 c1b1 a8e8 g3d3 b7b5 e2g3 a7a5 h1f1 e8b8 g3f5 b5b4 a3a4 c6e7 f5e7 d7e7 b2b3 f8e8 d1e1 e7e6 d2c1 c7c6 c1b2');
INSERT INTO evaluations VALUES(17189,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(17189,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(17189,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(17189,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(17189,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(17189,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(17189,36,'cp 127','c6d4 d1d4 e7c6 d4e3 f8b4 c1d2 d7d6 a2a3 b4c5 e3g3 e8g8 f2f4 g8h8 e1c1 a8b8 h2h4 a7a6 c3d5 c8e6 d2c3 f7f6 f1e2 c6e7 d5b4 e7c6 b4c6 b7c6 h4h5 h7h6 e2a6 d6d5 a6c4 e6g8 c4b3 d5d4 b3g8 f8g8 c1b1 c5a3');
INSERT INTO evaluations VALUES(17206,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(17206,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(17206,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(17206,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(17206,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(17206,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(17206,36,'cp 129','d1d4 e7c6 d4e3 f8b4 c1d2 d7d6 a2a3 b4a5 e1c1 c8e6 f2f4 f7f6 c3d5 a5d2 d1d2 d8d7 f1e2 e8g8 e3b3 c6d8 b3d3 g8h8 h1d1 d8c6 e2h5 a8b8 h2h3 f6f5 e4e5 d6e5 f4e5 d7d5 d3d5 e6d5 d2d5 g7g6 h5f3 b8e8');
INSERT INTO evaluations VALUES(17204,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(17204,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(17204,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(17204,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(17204,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(17204,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(17204,36,'cp 133','e7c6 d4e3 f8b4 c1d2 d7d6 a2a3 b4a5 e1c1 c8e6 f2f4 f7f6 c3d5 a5d2 d1d2 e8g8 f1e2 d8d7 e3b3 c6d8 b3c3 e6d5 d2d5 f8e8 c3c4 d8f7 h1d1 g8f8 e2f3 c7c6 d5a5 f8g8 c4c3 d7e7 c1b1 a7a6 a5h5 a8c8 g2g3 c8d8 h5h4 g7g6 b1a1 g8g7 f3g2 h7h5');
INSERT INTO evaluations VALUES(16868,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(16868,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(16868,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(16868,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(16868,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(16868,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(16868,36,'cp 128','d4e3 f8b4 c1d2 d7d6 a2a3 b4a5 e1c1 c8e6 f2f4 f7f6 c3d5 a5d2 e3d2 d8d7 f1b5 e8g8 h1e1 g8h8 h2h3 a7a6 b5e2 c6e7 d5e3 e6g8 e2g4 d7c6 d2d4 b7b5 e3d5 a8e8 g4h5 g8f7 h5f7 f8f7 d4a7 e7c8 a7f2 c8b6 f2d4 f7d7 g2g4 h8g8 c1b1 c6a8 d5b6 c7b6 d4b6 e8e4 e1e4');
INSERT INTO evaluations VALUES(49578,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49578,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49578,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49578,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(49578,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(49578,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(49578,36,'cp -1','e5d4 d1d4 b8c6 d4d1 f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 b7b6 f1e2 d7d6 e1g1 e4c3 c2c3 c8b7 a1d1 c6e5 f3e5 e8e5 d1d4 a7a5 f1d1 a5a4 e2f1 g7g6 d4f4 d8e7 h2h3 a8a5 b2b3 a4b3 a2b3 a5a3');
INSERT INTO evaluations VALUES(49752,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(49752,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(49752,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(49752,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(49752,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(49752,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(49752,36,'cp 0','d1d4 b8c6 d4d1 f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 b4c3 d2c3 f6e4 d1c2 b7b6 f1e2 d7d6 e1g1 c8b7 f1d1 e4c3 c2c3 c6e5 f3e5 e8e5 d1d4 d8f6 e2f1 a7a5 a1d1 e5e4 a2a3 e4d4 d1d4 g7g6 f2f3 h7h5 c3d2 a8e8 e3e4 h5h4 h2h3');
INSERT INTO evaluations VALUES(49750,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(49750,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(49750,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(49750,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(49750,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(49750,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49750,36,'cp 0','b8c6 d4d1 f8b4 c1d2 e8g8 e2e3 f8e8 g1f3 d7d6 d1c2 b4c3 d2c3 f6e4 f1e2 b7b6 e1g1 c8b7 f1d1 e4c3 c2c3 c6e5 f3e5 e8e5 d1d4 d8f6 d4f4 f6g6 g2g3 a8e8 e2f1 g6e6 f4d4 g7g6 f1g2 b7g2 g1g2 e5e4 a1d1 a7a5 c3c2 e4d4');
INSERT INTO evaluations VALUES(16665,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16665,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16665,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16665,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16665,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(16665,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16665,36,'cp 0','d4d1 f8b4 c1d2 d7d6 e2e3 e8g8 g1f3 f8e8 f1d3 c8g4 d3e2 b4c3 d2c3 f6e4 d1c2 g4f5 e2d3 f5g6 e1g1 a7a5 a2a3 a5a4 a1d1 d8e7 h2h3 g6h5 d3e2 h5g6');
INSERT INTO evaluations VALUES(1347,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(1347,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(1347,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(1347,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(1347,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(1347,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(1347,36,'cp 85','b5a4 g8f6 b1c3 e7e6 e1g1 d7d5 e4d5 e6d5 d2d4 c8e6 a2a3 f8e7 d4c5 b6c5 c1e3 c5d6 c3b5 d6b8 b5d4 e6d7 c2c4 e8g8 c4c5 f6g4 d4c6 d7c6 e3d4 e7f6 a4c2 f8e8 d4f6 g4f6 b2b4 b8f4 d1d4 f6e4 a1e1 f4f6 e1e3 a7a5 f1e1 a5b4 a3b4 c6b5');
INSERT INTO evaluations VALUES(2723,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(2723,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(2723,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(2723,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(2723,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(2723,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(2723,36,'cp 28','c8b7 d2d4 f8e8 b1d2 e7f8 a2a4 h7h6 b3c2 e5d4 c3d4 c6b4 c2b1 c7c5 d4d5 f6d7 a1a3 c5c4 e1e3 d8f6 a4b5 a6b5 b2b3 c4c3 d2f1 g7g6 f3d4 a8c8 a3a7 f8g7 d4e2 c3c2 b1c2 b4c2 e3f3 f6d8 a7b7 e8e4 f1g3 e4e8 b7b5 h6h5 f3f4 g7h6 f4c4 c8c4 b3c4');
INSERT INTO evaluations VALUES(23582,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(23582,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(23582,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(23582,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(23582,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(23582,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(23582,36,'cp 30','d2d4 f8e8 b1d2 e7f8 b3c2 c6b8 b2b3 b8d7 a2a4 h7h6 c1a3 c7c6 c2d3 f6h5 a3c1 h5f6 c1b2 d8c7 d1c2 f6h5 d3f1 h5f4 c3c4 b5a4 a1a4 f4e6 b2c3 e5d4 f3d4 d7c5 a4a1 e6d4 c3d4 c5e6 d4c3 c6c5 d2f3');
INSERT INTO evaluations VALUES(23579,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(23579,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(23579,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(23579,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(23579,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(23579,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(23579,36,'cp 27','f8e8 b1d2 e7f8 b3c2 g7g6 a2a3 c6b8 a3a4 b8d7 c2d3 d8b8 a4b5 a6b5 a1a8 b7a8 d1b3 c7c6 d3f1 b8b6 b3c2 a8b7 b2b3 e5d4 c3d4 d6d5 e4e5 f6h5 f3g5 f7f6 g5f3 h5g7 e5f6 d7f6 e1e8 g7e8 f1d3 e8c7 d2f1 c7e6 c1e3 f8d6 f1g3');
INSERT INTO evaluations VALUES(45449,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(45449,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(45449,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45449,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(45449,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45449,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(45449,36,'cp 19','c8g4 h2h3 g4h5 c1f4 e7e6 c2c3 f8d6 f4d6 d8d6 e1g1 d6b6 b1d2 b8c6 d1b3 b6c7 b3a3 c7e7 a3e7 e8e7 a2a4 h5g6 d3b5 c6a5 f1e1 h8c8 f3e5 c8c7 f2f3 a7a6 b5e2 g6c2 g1f2 f6d7 e5d7 e7d7 b2b4 a5b3 d2b3');
INSERT INTO evaluations VALUES(29175,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(29175,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(29175,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29175,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29175,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(29175,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(29175,36,'cp 21','h2h3 g4h5 c1f4 e7e6 b1d2 f8d6 f4d6 d8d6 e1g1 d6b6 c2c3 b8c6 d1b3 b6c7 b3a3 c7e7 a3a4 e8g8 a1e1 f8c8 e1e3 e7c7 f1e1 c7b6 g2g4 h5g6 d3g6 h7g6 a4c2 b6c7 a2a3 b7b5 d2b3 f6e4 f3d2 e4d6 b3c5 a7a5 f2f4 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(29173,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29173,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(29173,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29173,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(29173,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(29173,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(29173,36,'cp 18','g4h5 c1f4 e7e6 e1g1 f8d6 f4d6 d8d6 b1d2 b8c6 c2c3 e8g8 f1e1 c6e7 a2a4 a7a5 d3b5 h7h6 d1e2 d6c7 g2g4 h5g6 f3e5 g6h7 f2f3 f8c8 h3h4 f6e8 f3f4 e8d6 b5d3 h7d3 e5d3 d6c4 d2f3 c4d6 e2e3 d6c4 e3f2 c4d6 d3c5 b7b6 c5d3');
INSERT INTO evaluations VALUES(29156,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(29156,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(29156,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(29156,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29156,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(29156,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(29156,36,'cp 24','c1f4 b8c6 b1d2 e7e6 e1g1 f8d6 f4d6 d8d6 c2c3 e8g8 d1b1 f8b8 a2a4 h5g6 d3g6 h7g6 b1d3 b8e8 f1e1 f6d7 e1e3 e8e7 a1e1 a8e8 d3b5 e6e5 d4e5 d7e5 f3e5 c6e5 d2b3 e5c6 e3e7 e8e7 g1f1 c6e5 e1d1 e5c4 b5d5 d6d5 d1d5 c4b2');
INSERT INTO evaluations VALUES(57209,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57209,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57209,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(57209,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(57209,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(57209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57209,36,'cp 13','b8c6 c2c4 g8f6 b1c3 d7d5 c4d5 f6d5 g2g3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 a2a3 f7f6 d1a4 e8c8 b2b4 c8b8 b4b5 c6d4 f3d4 d7d4 a4d4 d8d4 a3a4 e6g4 f2f3 g4d7 g1f2 h8e8 f3f4 e7c5 f4e5 e8e5 c1e3');
INSERT INTO evaluations VALUES(22863,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(22863,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(22863,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(22863,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(22863,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(22863,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(22863,36,'cp 0','c2c4 g8f6 g2g3 d7d5 c4d5 f6d5 b1c3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 a2a3 a7a5 d1a4 e8g8 c1d2 f7f6 a1c1 f8d8 f1e1 a8b8 a4b5 d7d5 b5d5 d8d5 d2c3 d5b5 g1f1 g8f7 f3d2 a5a4 c1c2 b5c5 d2e4 c5b5');
INSERT INTO evaluations VALUES(54998,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(54998,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(54998,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54998,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(54998,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54998,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(54998,36,'cp 44','d2d4 e7e5 c2c4 g7g6 b1c3 g8f6 g1f3 d8e7 f1e2 f8g7 b2b4 e8g8 e1g1 c8f5 c1b2 f8e8 a2a4 a7a5 b4b5 b8d7 d4e5 d6e5 b2a3 e7d8 c4c5 c7c6 f3d2 g7f8 g2g4 f5e6 g4g5 f6d5 c3e4 h7h6 a1c1 h6g5');
INSERT INTO evaluations VALUES(57354,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(57354,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(57354,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(57354,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(57354,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(57354,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(57354,36,'cp 61','e2e3 g8f6 g1f3 f8e7 f1d3 e8g8 e1g1 d7d6 c2c4 f6h5 b1c3 c7c5 d3e2 c5d4 f3d4 h5f4 e3f4 g8h8 d1d2 e6e5 d4f3 b8c6 a1d1 c8e6 c3d5 e5e4 f3e1 e7f6 e1c2 e6d5 c4d5 c6e7 g2g3 a8c8 c2e3 d8d7 a2a3 g7g6 d1c1 c8c1 f1c1 f8c8');
INSERT INTO evaluations VALUES(57352,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(57352,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(57352,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(57352,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(57352,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(57352,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(57352,36,'cp 43','b8c6 g1f3 g8f6 f1d3 g7g6 e1g1 d7d6 c2c4 f8g7 b1c3 e8g8 a1c1 f6h5 f4g5 h5f6 d4d5 c6b4 d3e2 a7a5 d5e6 c8e6 a2a3 b4a6 e3e4 f5e4 c3e4 e6f5 e4c3 a6c5 f3d4 d8d7 d4f5 d7f5 g5e3 a5a4 c3b5 f8f7 f1e1 f6e4 f2f3');
INSERT INTO evaluations VALUES(57117,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(57117,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(57117,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(57117,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(57117,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(57117,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(57117,36,'cp -43','d7d5 e2e3 g8f6 c2c4 c7c6 g1f3 f8d6 b1c3 a7a6 d2d3 e8g8 f1e2 f8e8 e1g1 d8e7 c1b2 h7h6 c3a4 b8d7 d1c2 e7d8 f1c1 d7f8 e2f1 f8g6 h2h3 d6b8 g2g3 c8f5 c4d5 c6d5 f1g2 a6a5 b4b5 b8d6 c2b3 f5e6 a4c5');
INSERT INTO evaluations VALUES(56337,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(56337,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(56337,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(56337,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(56337,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(56337,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(56337,36,'cp -55','e2e3 g8f6 c1b2 f8d6 c2c4 c7c6 g1f3 b8d7 d2d3 e8g8 b1d2 f8e8 f1e2 d7f8 e1g1 h7h6 f1e1 f8g6 d1c2 c8d7 e2f1 d8e7 g2g3 b7b6 d2b3 d7g4 f1g2 e7d7 e3e4 g4h3 c4c5 d6c7 a3a4 h3g2 g1g2 g8h8 c5b6');
INSERT INTO evaluations VALUES(56335,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(56335,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(56335,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(56335,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(56335,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(56335,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(56335,36,'cp -50','f8d6 c1b2 g8f6 c2c4 c7c6 d2d3 e8g8 g1f3 b8d7 f1e2 f8e8 b1d2 d7f8 e1g1 f8g6 f1e1 c8g4 e3e4 a7a5 d1b3 d5d4 c4c5 d6c7 g2g3 h7h6 d2c4 g4h3 e1c1 d8e7 b3c2 f6h7 e2f1 h3g4 f1g2 h7g5 f3g5 e7g5');
INSERT INTO evaluations VALUES(14807,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14807,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(14807,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14807,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14807,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(14807,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(14807,36,'cp 38','g1f3 f8d6 f4g3 d8e7 c2c4 d5c4 d3c4 d6g3 h2g3 e8g8 b1c3 f8d8 e1g1 h7h6 a2a3 e6e5 f1e1 c8e6 c4e6 e7e6 d1a4 e5d4 e3d4 e6d6 a1d1 a7a6 a4b3 c6a5 b3c2 d8e8 f3e5 a8b8 c3e4 f6e4 e1e4 a5c6 e5c6 b7c6 d1e1 e8e4 e1e4 b8d8 b2b4');
INSERT INTO evaluations VALUES(39647,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(39647,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(39647,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(39647,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(39647,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(39647,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(39647,36,'cp 16','c7c5 c2c4 b8c6 b1c3 d7d6 e1g1 c8g4 h2h3 g4f3 g2f3 e7e6 a2a3 g8e7 f3g2 e8g8 d2d3 d8d7 a1b1 a8c8 d1a4 f8d8 f1d1 b7b6 c1g5 d6d5 c4d5 e6d5 e2e4 d5e4 d3e4 c6d4 a4d7 d8d7 d1d2 h7h6 g5e3 c8d8 b1d1 e7c6 c3d5');
INSERT INTO evaluations VALUES(37836,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(37836,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(37836,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(37836,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(37836,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(37836,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(37836,36,'cp 16','c2c4 b8c6 b1c3 d7d6 e1g1 c8f5 h2h3 e7e5 d2d3 g8e7 a2a3 a7a5 a1b1 e8g8 b2b3 a8b8 g1h2 f5e6 f3g5 e6c8 c1b2 h7h6 g5e4 f7f5 e4d2 c8e6 c3d5 b7b5 e2e3 d8d7 f2f4 b5c4 d5e7 c6e7 d2c4 e6c4 d3c4 e5e4 d1d2');
INSERT INTO evaluations VALUES(51775,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(51775,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(51775,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(51775,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(51775,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51775,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(51775,36,'cp 0','c4b3 d7d5 d1e2 a7a5 a2a4 f8e7 g1f3 d5e4 e2e4 g8f6 e4e2 b7b6 b1a3 e8g8 d2d4 c8a6 b3c4 a6c4 a3c4 f6d5 e1g1 d8c7 g2g3 b8d7 c4e3 f8e8 f1d1 c6c5 c2c3 d7f6 e3d5 e6d5 c1f4 c7b7 f4e5');
INSERT INTO evaluations VALUES(56178,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(56178,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(56178,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(56178,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(56178,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(56178,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(56178,36,'cp -81','f8e7 c4b3 a7a5 c2c3 g8f6 g1f3 e8g8 e1g1 f8e8 d2d4 e7d6 e2d1 h7h6 b1d2 c8f5 f1e1 b8d7 d2f1 c7c6 e1e8 d8e8 a2a4 e8e6 b3c2 f6e4 f3e1 a8e8 d1e2 d6c7 f2f3 e4d6 e2e6 f7e6 c1e3 f5c2 e1c2 e6e5 d4e5 d7e5 e3d4');
INSERT INTO evaluations VALUES(38308,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(38308,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(38308,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(38308,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(38308,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(38308,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(38308,36,'cp -76','c4b3 a7a5 c2c3 g8f6 g1f3 e8g8 e1g1 e7d6 d2d4 a5a4 b3c2 f8e8 c1e3 c8g4 b1d2 a4a3 f1b1 b8d7 b2a3 a8a3 b1b3 d8a8 b3a3 a8a3 c3c4 d6f4 d2f1 d5c4 e2c4 d7b6 c4b3 a3d6 e3f4 d6f4 f3e5 g4e6 b3d3 g7g6 f1e3');
INSERT INTO evaluations VALUES(38318,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(38318,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(38318,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(38318,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(38318,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(38318,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(38318,36,'cp -75','g8f6 g1f3 e8g8 d2d4 a7a5 c2c3 f8e8 e1g1 e7d6 e2d1 h7h6 c3c4 d5c4 b3c4 b8d7 b1c3 d7b6 c4b3 c7c6 f3e5 c8e6 f1e1 d8c7 b3c2 b6d5 h2h3 b7b5 c3d5 e6d5 c1f4 c7b6 d1d3 a5a4 f4g3 b6b7 a2a3');
INSERT INTO evaluations VALUES(41108,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(41108,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(41108,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(41108,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(41108,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(41108,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(41108,36,'cp -85','g1f3 e8g8 d2d4 f8e8 e1g1 a7a5 c2c3 e7d6 e2d1 h7h6 c3c4 d5c4 b3c4 b8d7 h2h3 d7b6 c4d3 c8d7 b1c3 b6d5 c3b5 d6f4 f1e1 e8e1 f3e1 d7c6 e1f3 a5a4 a2a3 d8e7 c1f4 d5f4 d3f1 f6d5 f3e5 a8e8');
INSERT INTO evaluations VALUES(41107,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(41107,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(41107,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(41107,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(41107,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(41107,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(41107,36,'cp -65','e8g8 d2d4 a7a5 c2c3 f8e8 e1g1 e7d6 e2d1 c8g4 c1g5 c7c6 b1d2 b8d7 f1e1 d8c7 h2h3 e8e1 d1e1 g4f5 a2a4 a8e8 e1d1 h7h6 g5e3 d7f8 b3c2 c7d7 c2f5 d7f5 f3h4 f5e6 d1f3 f8g6 h4f5 d6b8 a1e1');
INSERT INTO evaluations VALUES(36417,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(36417,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(36417,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(36417,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(36417,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(36417,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(36417,36,'cp -59','d2d4 a7a5 c2c3 e7d6 e1g1 f8e8 e2d1 h7h6 h2h3 c8f5 c3c4 d5c4 b3c4 b8d7 c4d3 f5d3 d1d3 d7c5 d3c2 c5e4 f1e1 d8d7 a2a3 d7b5 c1d2 c7c6 b1c3 e4c3 d2c3 f6d5 c3d2 a5a4 f3e5 d5f6 a1c1 d6e5 d4e5 e8e5 e1e5 b5e5 d2c3 e5e4');
INSERT INTO evaluations VALUES(1842,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(1842,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(1842,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(1842,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(1842,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(1842,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(1842,36,'cp 35','f8c5 c2c3 g8e7 e1g1 h7h6 c1e3 c5b6 b1d2 d7d6 a2a4 e8g8 d2c4 b6e3 f2e3 f6e6 c4d2 e7g6 d3d4 c6b8 b5d3 b8d7 a4a5 e6e7 b2b4 d7f6 d1e1 c8d7 h2h3 c7c6 c3c4 f8e8 f3h4 g6h4 e1h4 b7b5 a5b6 a7b6');
INSERT INTO evaluations VALUES(261,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(261,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(261,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(261,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(261,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(261,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(261,36,'cp 34','c2c3 g8e7 e1g1 h7h6 c1e3 d7d6 d3d4 e5d4 c3d4 c5b6 d4d5 b6e3 f2e3 a7a6 b5a4 b7b5 d5c6 b5a4 b1c3 a4a3 d1d4 f6g6 c3d5 e7d5 e4d5 a3b2 a1b1 e8g8 b1b2 f8e8 b2f2 g6e4 d4e4 e8e4 f3d4 f7f5 d4f5 c8f5 f2f5 e4e3');
INSERT INTO evaluations VALUES(9040,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(9040,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(9040,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(9040,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(9040,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(9040,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(9040,36,'cp 202','c4d3 f6d7 a2a3 b4c3 b2c3 b7b6 e1g1 c6a5 f1e1 c8b7 c1f4 e8g8 d1e2 f7f6 f3d2 f8e8 d2c4 a5c4 d3c4 g8h8 f2f3 d7f8 f4g3 f8g6 a3a4 a7a5 h2h4 e6e5 c4b5 c7c6 b5c4 b7c8 h4h5 g6f4 g3f4 e5f4 h5h6 g7h6');
INSERT INTO evaluations VALUES(23044,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(23044,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(23044,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(23044,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(23044,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(23044,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(23044,36,'cp 159','c1e3 a7a5 d1d2 f8e7 b1c3 d7d6 f2f3 h7h5 a2a3 b4a6 e1c1 g8f6 g1h3 c8h3 g2h3 f6d7 h1g1 g7g6 f1b5 e8g8 c1b1 a6c5 g1g2 g8h7 b1a2 e7h4 a3a4 d8e7 d2e2 f8g8 b5d7 c5d7 c3b5 b7b6 b5c7 a8c8 c7b5 d7c5 e3c5 c8c5');
INSERT INTO evaluations VALUES(10115,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(10115,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(10115,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(10115,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(10115,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10115,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(10115,36,'cp 9','g1f3 a7a5 e1g1 d7d6 h2h3 h7h6 b1d2 e8g8 f1e1 c8e6 c4b5 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 d1c2 c7c6 b5c4 d6d5 e4d5 c6d5 c4b5 a7b6 a2a4 e7g6 d3d4 e5e4 f3e5 a8d8 e5g6 f7g6 a1d1 g8h7 b5f1 b6d6 c3c4 d5c4 f1c4 e6c4 e3c4 d6d5');
INSERT INTO evaluations VALUES(21487,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(21487,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(21487,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(21487,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(21487,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(21487,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(21487,36,'cp 113','g1f3 g8f6 b1c3 a7a6 a2a3 f8d6 e3e4 e6e5 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 c4f7 e5c3 b2c3 h8f8 f7b3 f6e4 c1e3 c8d7 f2f3 e4d6 e1c1 a6a5 h1e1 a8a6 b3d5 f8e8 c3c4 c7c6 d5g8 e8g8 d1d6 g8e8 e3g5 d8c7 e1e8 d7e8 c4c5 e8g6 g5e7 c7c8');
INSERT INTO evaluations VALUES(21488,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(21488,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(21488,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(21488,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(21488,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(21488,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(21488,36,'cp 90','g8f6 b1c3 a7a6 a2a3 b7b5 c4d3 f8d6 d1c2 c8b7 e1g1 e8g8 c3e4 e6e5 f1d1 e5d4 e3d4 h7h6 e4c5 d8c8 d1e1 c6e7 f3e5 b7d5 c1f4 f8e8 h2h3 e7c6 c2c3 d6c5 d4c5 c8b7 f2f3 f6h5 f4h2 c6e5 e1e5 h5f6 d3f5 b7c6 a1e1 a8d8 h2f4');
INSERT INTO evaluations VALUES(53567,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(53567,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(53567,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(53567,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(53567,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(53567,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53567,36,'cp 0','d5c4 a2a4 d8b6 d1c2 b6d4 e2e4 d4c5 g1f3 g8f6 f1e2 b8d7 a4a5 f8d6 f4e3 c5h5 f3e5 h5e5 f2f4 f6g4 f4e5 g4e3 c2d2 d6c5 c3a4 c5b4 a4c3');
INSERT INTO evaluations VALUES(53654,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(53654,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53654,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53654,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(53654,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(53654,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(53654,36,'cp -13','a2a4 g8f6 e2e3 b7b5 a4b5 c6b5 c3b5 f6d5 b5c3 f8b4 d1c2 b4c3 b2c3 d5f4 e3f4 c8b7 g1f3 b7f3 g2f3 b8c6 f1c4 e8g8 c4d3 d8h4 e1g1 c6e7 f4f5 e6f5 f1e1 a7a5 d3f5 h4g5 f5g4 f7f5 e1e5 a5a4 h2h4 g5h6 a1a4 a8a4 c2a4 f5g4 e5e7 g4f3 a4c4 g8h8');
INSERT INTO evaluations VALUES(44207,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44207,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44207,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44207,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(44207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44207,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(44207,36,'cp 0','c5d4 f3d4 d7d5 e2e3 b8c6 c4d5 e6d5 f1b5 c8d7 e1g1 f8d6 e3e4 d5e4 d4c6 d7c6 d1a4 e8g8 b5c6 b7c6 c3e4 f6e4 a4e4 d8f6 c1e3 f8e8 e4c2 d6e5 a1b1 e5c7 b1c1 c7b6 c2c3 f6c3 c1c3 b6e3 f2e3 a8d8 c3c6 d8d2 f1f2 d2d1 f2f1');
INSERT INTO evaluations VALUES(44283,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(44283,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44283,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(44283,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(44283,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(44283,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44283,36,'cp 19','f3d4 d7d5 e2e3 b8c6 c4d5 e6d5 f1b5 c8d7 e1g1 f8d6 d4f3 d7g4 h2h3 g4h5 b5e2 e8g8 f3h4 h5e2 c3e2 f6e4 h4f3 f8e8 c1d2 c6e5 f3e5 d6e5 d2c3 e5c3 e2c3 e4c3 b2c3 b7b6 d1b3 a8c8 a1d1 c8c5 d1d4 d8c7');
INSERT INTO evaluations VALUES(44273,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44273,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(44273,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(44273,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(44273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44273,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44273,36,'cp 0','d7d5 e2e3 b8c6 c4d5 e6d5 f1b5 c8d7 e1g1 f8d6 d4f3 d7g4 b5e2 e8g8 f3d4 g4c8 b2b3 d6e5 d4c6 b7c6 c1b2 c8f5 a1c1 a8c8 e2f3 f8e8 c3a4 e5b2 a4b2 c6c5 b2a4 d8d6 h2h3 d5d4 a4b2 c8d8 b2c4 d6a6 e3d4 c5d4');
INSERT INTO evaluations VALUES(46303,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46303,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(46303,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(46303,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(46303,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(46303,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(46303,36,'cp 11','e2e3 b8c6 c4d5 e6d5 f1e2 f8d6 c3b5 d6b8 d1c2 c6d4 b5d4 e8g8 c1d2 f8e8 e1g1 f6e4 d2c3 d8d6 g2g3 c8h3 f1c1 d6f6 e2f3 b8e5 f3e4 d5e4 c2e4 b7b6 e4h4 f6h4 g3h4 a8c8 d4e2 e5c3 c1c3 c8c3');
INSERT INTO evaluations VALUES(46302,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(46302,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(46302,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(46302,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(46302,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(46302,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46302,36,'cp 6','b8c6 c4d5 e6d5 f1b5 c8d7 e1g1 f8d6 e3e4 d5e4 d4c6 d7c6 b5c6 b7c6 c1g5 e8g8 c3e4 d6h2 g1h2 d8d1 e4f6 g7f6 a1d1 f6g5 f2f4 a8b8 b2b3 g5f4 f1f4 f8d8 d1d8 b8d8 f4a4 d8d7 h2g3 g8f8 a4a5 d7c7 g3f2 f8e7');
INSERT INTO evaluations VALUES(13627,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13627,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13627,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13627,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(13627,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(13627,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(13627,36,'cp 6','c4d5 e6d5 f1e2 f8b4 e1g1 e8g8 d1b3 d8e7 e2f3 f8d8 c1d2 b4d6 c3b5 d6c5 d4c6 b7c6 b5d4 c8d7 b3c2 a8c8 d4f5 d7f5 c2f5 f6e4 f3e4 g7g6 f5f3 d5e4 f3e2 c5b4 d2b4 e7b4 f1d1 d8d3 a1c1 c8d8 b2b3 d3d2 d1d2 d8d2');
INSERT INTO evaluations VALUES(13620,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13620,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(13620,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(13620,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(13620,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(13620,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(13620,36,'cp 0','e6d5 f1e2 f8b4 d1b3 b4c5 e1g1 e8g8 f1d1 d8e7 h2h3 f8d8 e2f3 h7h6 c3e2 c5d6 e2f4 f6e4 f3e4 d5e4 d4c6 b7c6 c1d2 d6f4 e3f4 c8e6 b3c2 d8d3 d2e3 e7d7 d1c1 e6h3 g2h3 d7h3 c2c6 h3g4 g1f1 g4h3');
INSERT INTO evaluations VALUES(50377,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50377,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50377,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50377,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50377,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(50377,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50377,36,'cp -2','e2e4 e7e5 f2f4 e5f4 c1f4 b8c6 f1e2 h7h6 a2a3 f8c5 b1c3 e8g8 e4e5 f6e8 d1d2 d7d6 e5d6 e8d6 e1c1 c6d4 g1f3 d4e2 c3e2 f8e8 h1e1 c8g4 e2g3 g4f3 g2f3 e8e1 d1e1 d8d7 g3e4 d6e4 f3e4 a8d8 d2e2 c5e7');
INSERT INTO evaluations VALUES(22776,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(22776,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(22776,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(22776,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(22776,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(22776,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22776,36,'cp -50','b1c3 g8f6 g1f3 f8c5 f1b5 c6d4 f3e5 e8g8 e1g1 d4b5 c3b5 f6e4 d2d4 c5e7 f2f3 e4f6 c2c4 d7d6 e5d3 c7c6 b5c3 d6d5 c1e3 f8e8 f1e1 d5c4 b3c4 c8e6 d3e5 f6d7 e5d7 d8d7 d4d5 c6d5 c4d5 e6f5 d1d2 e7b4 e3f2 a7a6 h2h3 a6a5 e1e8 a8e8 a1d1');
INSERT INTO evaluations VALUES(22778,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(22778,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(22778,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(22778,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(22778,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(22778,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(22778,36,'cp -59','g8f6 g1f3 f8c5 f1e2 e8g8 e1g1 f8e8 f1e1 d7d6 d2d3 a7a5 c3a4 c5a7 c2c4 c8g4 a4c3 g4f3 e2f3 c6d4 g2g3 c7c6 f3g2 b7b5 c3e2 d4e6 e1f1 h7h5 c4b5 c6b5 c1b2 f6g4 h2h3 a7f2 f1f2 g4f2 g1f2');
INSERT INTO evaluations VALUES(16503,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(16503,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(16503,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(16503,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(16503,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(16503,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(16503,36,'cp -45','g1f3 f8c5 f1e2 e8g8 e1g1 d7d6 d2d3 a7a6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 f7f5 d3d4 c5a7 e4f5 c8f5 d5e3 f5e4 d4e5 c6e5 f3e5 d6e5 d1d8 a8d8 a1d1 g8f7 e3c4 f7f6 f1e1 e4f5 a2a4 e5e4 e2f1 d8d1 e1d1 f5e6 a4a5');
INSERT INTO evaluations VALUES(55809,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(55809,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(55809,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(55809,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55809,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(55809,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55809,36,'cp 18','f1g2 g8f6 g1f3 d5c4 d1a4 c8d7 a4c4 c7c5 b1c3 b8c6 e1g1 f8e7 d2d4 b7b5 c4d3 c5c4 d3c2 b5b4 c3e4 a8c8 f1d1 c4c3 e4f6 e7f6 c1f4 e8g8 f4d6 f8e8 a1c1 f6e7 d6e7 c6e7 b2c3 c8c3 c2d2');
INSERT INTO evaluations VALUES(55810,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55810,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(55810,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(55810,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(55810,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(55810,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(55810,36,'cp 12','d5c4 d1a4 c7c6 a4c4 c6c5 g1f3 b8c6 d2d3 g8f6 c1e3 b7b6 e1g1 f8e7 d3d4 f6d5 d4c5 e8g8 e3d4 c8b7 b1c3 c6d4 f3d4 b6c5 d4f5 d5b6 f5e7 d8e7 c4f4 b7g2 g1g2 a8b8 a1b1 b6d5 f4f3 d5c3 f3c3 e7b7 g2g1 b7a6 c3c5 a6a2 c5e5 a2b3 e5f4 b3b6');
INSERT INTO evaluations VALUES(56013,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(56013,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(56013,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(56013,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(56013,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(56013,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(56013,36,'cp 22','d1a4 c7c6 a4c4 c6c5 d2d3 g8f6 b1c3 f8e7 c4b3 d8b6 b3b5 b6b5 c3b5 b8a6 f2f4 f6d5 e1f2 c8d7 b5c3 a6c7 c1d2 f7f6 g1f3 e8f7 h1c1 h8d8 g3g4 d5c3 b2c3 c5c4 d3c4 d7c6 d2e3 c7a6 c1d1 d8d1 a1d1 a6c5 f3d4');
INSERT INTO evaluations VALUES(16619,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(16619,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(16619,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(16619,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(16619,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(16619,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(16619,36,'cp 25','c2c4 d7d5 c4d5 f6d5 e1g1 d5b6 d2d3 f8e7 b1c3 e8g8 a2a3 c8e6 c1e3 a7a5 c3a4 b6d5 e3c5 e7d6 a1c1 a8c8 f3d2 f7f5 e2e4 d5e7 d2f3 b7b6 f3g5 e6f7 c5d6 d8d6 g5f7 f8f7 e4f5 c6d4 a4c3 e7f5 c3e4 d6g6 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(16463,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16463,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(16463,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(16463,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(16463,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(16463,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(16463,36,'cp 7','d7d5 c4d5 f6d5 e1g1 d5b6 b1c3 f8e7 b2b3 e8g8 c1b2 c8g4 a1c1 f7f5 d2d3 g8h8 h2h3 g4h5 c3a4 b6a4 b3a4 e7d6 f3d2 a8b8 d2c4 c6d4 b2d4 e5d4 c4d6 d8d6 d1c2 c7c6 a4a5 c6c5 c2c5 d6c5 c1c5 h5e2 f1e1 e2d3 e1e7 f5f4 g3g4');
INSERT INTO evaluations VALUES(20966,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20966,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(20966,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(20966,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20966,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(20966,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20966,36,'cp 0','e2e3 f7f6 g5h4 e7e5 d4e5 c6e5 f1e2 c7c6 g1f3 e5f3 e2f3 g8e7 h4g3 e7f5 b1c3 c8e6 d1d3 f8b4 e1c1 d8a5 e3e4 f5g3 h2g3 b4c3 e4d5 c3b2 c1b2 e8c8 h1h7 h8h7 d3h7 a5b4 b2c1 b4a3 c1b1 a3b4');
INSERT INTO evaluations VALUES(20964,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20964,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(20964,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(20964,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(20964,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(20964,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(20964,36,'cp -13','f7f6 g5h4 e7e5 d4e5 c6e5 f1e2 c7c6 b1d2 g8e7 h4g3 e7f5 g1f3 e5f3 e2f3 c8e6 d2b3 g7g6 d1e2 f5g3 h2g3 f6f5 g3g4 d8e7 g4f5 e6f5 b3d4 f5d7 e1g1 f8g7 c2c4 g7d4 e3d4 e7e2 f3e2 d7e6 c4c5 e6f5 f1e1 e8f7 a1c1 a8e8 f2f3');
INSERT INTO evaluations VALUES(20828,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(20828,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(20828,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(20828,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(20828,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(20828,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(20828,36,'cp -5','g5h4 e7e5 d4e5 c6e5 f1e2 g8h6 g1f3 e5f3 e2f3 h6f5 h4g3 c7c6 b1d2 c8e6 d2b3 e6f7 e3e4 d5e4 f3e4 f5g3 h2g3 d8d1 a1d1 a7a5 e1g1 f7e6 a2a3 e6b3 c2b3 f8e7 f1e1 g7g6 e4d3 a8d8 g1f1 e8f8 f1g1 d8d7');
INSERT INTO evaluations VALUES(20848,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(20848,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(20848,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(20848,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(20848,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(20848,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(20848,36,'cp -5','e7e5 d4e5 c6e5 f1e2 g8h6 h4g3 h6f5 g1f3 e5f3 e2f3 c8e6 b1d2 c7c6 d2b3 g7g6 g3f4 f8d6 e1g1 d6f4 e3f4 d8d6 f1e1 e8c8 b3d4 h8e8 d1d2 f5d4 d2d4 e6f5 c2c3 e8e1 a1e1 c6c5 d4a4 d6b6 e1d1 f5d7 a4c2 d7f5');
INSERT INTO evaluations VALUES(24895,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(24895,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(24895,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(24895,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(24895,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(24895,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(24895,36,'cp 83','d1b3 e7e6 b3b7 c6b4 b7b5 c7c6 b5a4 g4f5 b1a3 a7a5 c4c5 f6d7 f1e2 g7g5 e1g1 f8g7 a4d1 e8g8 c1d2 a8b8 d2b4 b8b4 b2b3 f5g6 a3c2 b4b7 a1b1 d8e7 b1c1 f8b8 e2d3 e6e5 d3g6 h7g6 d4e5 g5g4 f3e1 d7e5 c2d4 b8a8 e1d3 e5d3 d1d3 g7d4 e3d4');
INSERT INTO evaluations VALUES(16578,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16578,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16578,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(16578,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(16578,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16578,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(16578,36,'cp 0','b1c3 f8c5 g1e2 e8g8 d2d3 a7a5 e1g1 f8e8 h2h3 c6d4 e2d4 c5d4 c3b5 d4b6 c2c4 d7d6 b5c3 c7c6 g1h2 b6d4 a2a4 c8d7 f2f4 h7h6 c3e2 d4c5 e2c3');
INSERT INTO evaluations VALUES(16516,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16516,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16516,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16516,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16516,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16516,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16516,36,'cp 0','f8c5 d2d3 a7a5 g1e2 e8g8 e1g1 f8e8 h2h3 c6d4 e2d4 c5d4 c3b5 d4b6 c2c4 d7d6 b5c3 c7c6 g1h2 b6d4 a2a4 c8d7 f2f4 h7h6 c3e2 d4c5 e2c3');
INSERT INTO evaluations VALUES(10213,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10213,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10213,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10213,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10213,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10213,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10213,36,'cp 0','c3a4 c5e7 a4c3 e7c5');
INSERT INTO evaluations VALUES(29516,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(29516,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(29516,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(29516,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29516,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(29516,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(29516,36,'cp 9','e7e6 g1f3 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 b8d7 e1g1 e8g8 d1b3 d7f6 a1c1 b6d5 f1e1 a8b8 b5c4 h7h6 c4f1 a7a6 b3a3 a6a5 a3b3 d5b4 a2a3 b4c6 h2h3 d8e7 c1c3 f6d5 c3c4 d5f6');
INSERT INTO evaluations VALUES(30064,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30064,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(30064,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(30064,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(30064,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(30064,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(30064,36,'cp 21','g1f3 f8d6 f4d6 c7d6 d1b3 d5c4 f1c4 d8e7 c4b5 b8d7 e1g1 e8g8 b5e2 f8c8 b1c3 d7b6 h2h3 f5e4 c3e4 f6e4 f1c1 d6d5 a2a4 c8c1 a1c1 a8c8 c1c8 b6c8 b3c2 c8d6 e2d3 g7g6 g2g3 g8g7 a4a5 a7a6 g1g2 f7f6');
INSERT INTO evaluations VALUES(45111,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(45111,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(45111,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45111,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45111,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45111,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45111,36,'cp 0','c8d7 b5d7 d8d7 d2d4 c5d4 d1d4 g7g6 c1g5 f8g7 a1d1 h7h6 g5f6 g7f6 d4d2 f6g7 e1g1 b8c6 c3d5 e8g8 c2c3 a8d8 d2f4 b7b5 f4h4 f8e8 h2h3 d7b7 b2b3 d8c8 f1e1 c8d8');
INSERT INTO evaluations VALUES(28868,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(28868,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(28868,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(28868,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(28868,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28868,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28868,36,'cp 0','d1e2 g7g6 b5d7 b8d7 e1g1 f8g7 a2a4 e8g8 h2h3 d7b8 d2d4 c5d4 f3d4 b8c6 d4b3 a8c8 f1d1 c6e5 a4a5 a7a6 f2f4 e5c4 e4e5 f6d7 e2f3 d8c7 c3d5 c7d8');
INSERT INTO evaluations VALUES(48120,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(48120,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(48120,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(48120,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(48120,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(48120,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(48120,36,'cp 130','c3e4 d5e4 f1c4 d8d7 g1e2 c7c5 c2c3 b7b6 c1e3 b8c6 e1g1 f8e7 c4b5 e8g8 e2g3 c5d4 e3d4 c8b7 f1e1 d7c7 b5c6 c7c6 g3e4 c6c7 d1g4 a8d8 a1d1 b7e4 e1e4 e7c5 g4g3 c5d4 d1d4 d8d4 e4d4 f8d8 g3f4 d8d4 f4d4 b6b5');
INSERT INTO evaluations VALUES(48118,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(48118,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(48118,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(48118,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(48118,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(48118,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(48118,36,'cp 92','d5e4 f1c4 a7a6 a2a4 f8e7 g1e2 e8g8 e1g1 c7c5 c2c3 c5d4 c3d4 b8c6 c1e3 c8d7 e2c3 c6b4 c4b3 d7c6 d1g4 g8h8 c3e4 a8c8 f1d1 c6d5 b3d5 b4d5 g4e2 h7h6 h2h3 d8d7 e2d2 h8h7 d2d3 h7g8 a4a5 d5b4 d3b3 d7d5 b3d5 b4d5 a1c1 e7b4 e3d2 c8c1 d1c1 b4d2 e4d2 f8d8 d2c4');
INSERT INTO evaluations VALUES(48133,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(48133,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(48133,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(48133,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(48133,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(48133,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(48133,36,'cp 111','f1c4 a7a6 a2a4 f8e7 g1e2 e8g8 c2c3 c7c5 e1g1 b8c6 c1e3 c8d7 c4b3 c5d4 c3d4 c6b4 e2c3 a8c8 c3e4 d7c6 e4c3 b4d5 d1e2 g8h8 c3d5 c6d5 e2d3 d8d7 b3c2 g7g6 c2b3 h8g8 b3d5 d7d5 f1c1 f8d8 h2h3 h7h5 c1c3 e7b4 c3c2');
INSERT INTO evaluations VALUES(56157,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(56157,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(56157,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56157,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(56157,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(56157,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(56157,36,'cp 15','d8d5 g1f3 b8c6 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3d2 g8f6 e2e3 e8g8 d1b3 c6e7 b3d5 e7d5 f1c4 f8e8 e1g1 a7a5 a2a3 a5a4 h2h3 c8e6 f1c1 c7c6 c4e2 e8c8 d2c4 d5c3 c1c3 f6d5 c3c1 f7f5 c4d6');
INSERT INTO evaluations VALUES(34931,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(34931,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(34931,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(34931,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(34931,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(34931,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(34931,36,'cp 19','g1f3 b8c6 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3d2 e4e3 f2e3 g8f6 d1b3 c8e6 b3d5 e6d5 h1g1 h7h5 h2h3 c6e7 g2g4 h5g4 h3g4 g7g5 a1c1 e8c8 d2f3 d8g8 a2a4 f6e4 f3e5 f7f6 e5d3 h8h2 a4a5 b7b6 c3b4 e7c6 f1g2 c6a5 b4a5 b6a5 c1a1');
INSERT INTO evaluations VALUES(34929,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34929,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(34929,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(34929,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(34929,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(34929,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34929,36,'cp 0','b8c6 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3d2 e4e3 f2e3 g8f6 d1b3 c8e6 b3d5 e6d5 h1g1 h7h5 h2h3 c6e7 g2g4 h5g4 h3g4 g7g5 a1c1 e8c8 d2f3 d8g8 a2a4 f6e4 f1g2 e4f6');
INSERT INTO evaluations VALUES(1717,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(1717,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(1717,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(1717,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(1717,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(1717,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(1717,36,'cp 29','b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3d2 e4e3 f2e3 g8f6 d1b3 c8e6 b3b7 c6d4 b7d5 d4c2 e1f2 f6d5 c3g7 h8g8 a1c1 d5e3 g7c3 e8c8 h2h3 g8e8 h1g1 f7f5 g2g3 e6d5 g3g4 f5g4 h3g4 d8d7 f1h3 d7f7 d2f3 c8b7 g1g3 d5a2 f2g1 c7c5 f3e5 f7e7 e5d3 b7c6');
INSERT INTO evaluations VALUES(1711,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(1711,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(1711,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(1711,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(1711,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(1711,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(1711,36,'cp 4','f8b4 c1d2 b4c3 d2c3 e5e4 f3d2 e4e3 f2e3 g8f6 d1b3 c8e6 b3d5 e6d5 h1g1 h7h5 h2h3 c6e7 g2g4 h5g4 h3g4 g7g5 a1c1 e8c8 d2f3 d8g8 a2a4 f6e4 f3e5 g8e8 f1g2 f7f6 e5c4 b7b6 c3b4 e7c6 b4a3 h8h2 g2f3 c8b7 g1h1');
INSERT INTO evaluations VALUES(179,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(179,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(179,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(179,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(179,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(179,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(179,36,'cp 19','c1d2 b4c3 d2c3 e5e4 f3d2 e4e3 f2e3 g8f6 d1b3 c8e6 b3b7 c6d4 b7d5 d4c2 e1f2 f6d5 c3g7 h8g8 a1c1 d5e3 g7c3 e8c8 g2g3 g8e8 h2h3 e6a2 f1g2 e3g2 c1c2 g2e3 c2c1 e3c4 d2c4 a2c4 e2e3 d8d3 c3d4 d3d2 f2g1 c4e6 h1h2 d2h2 g1h2 e8d8 c1a1 c7c5 a1c1 d8d5 d4c5 c8b7 c5d4 d5b5 e3e4');
INSERT INTO evaluations VALUES(13273,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(13273,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(13273,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(13273,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(13273,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(13273,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13273,36,'cp 38','d2d4 c8g4 d4d5 c6d4 f1e2 g4f3 e2f3 g7g6 c1e3 f8g7 e3d4 c5d4 d1d4 e8g8 d4d2 f6d7 f3e2 d8a5 e1g1 d7c5 f1c1 c5a4 a1b1 g7c3 b2c3 a4c5 f2f3 a5c7 c1e1 b7b6 g2g3 a8e8 e2f1 e7e5 d5e6 f7e6 d2e3');
INSERT INTO evaluations VALUES(13292,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(13292,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(13292,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(13292,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(13292,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13292,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(13292,36,'cp 39','c8g4 d4d5 c6d4 f1e2 g4f3 e2f3 g7g6 c1e3 f8g7 e3d4 c5d4 d1d4 e8g8 f3e2 f6d7 d4d2 d7c5 e1g1 d8a5 f1c1 c5a4 a1b1 g7c3 b2c3 a4c5 f2f3 a5c7 c1e1 b7b6 b1d1 a8e8 g2g3 c5a4 g1g2 c7c5 f3f4 f7f6 h2h4 c5a3 e2g4 a3c3 g4e6 g8g7 d2d3 c3d3 d1d3');
INSERT INTO evaluations VALUES(26536,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(26536,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(26536,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26536,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(26536,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(26536,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(26536,36,'cp 34','f5d3 d1d3 e7e6 g1f3 f8d6 f4g3 g8f6 e1g1 e8g8 c2c4 c6e7 d3b3 b7b6 b1c3 d6g3 h2g3 a8c8 f1d1 c7c6 a1c1 c8c7 c3e2 c6c5 b3a3 d8b8 c4d5 e7d5 d4c5 c7c5 c1c5 b6c5 e2c3 d5b4 a3b3 b4d5 b3c2 f8c8 a2a3');
INSERT INTO evaluations VALUES(26596,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(26596,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(26596,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(26596,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(26596,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(26596,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(26596,36,'cp 34','d1d3 e7e6 g1f3 f8d6 f4g3 g8f6 e1g1 e8g8 c2c4 c6e7 a2a3 c7c5 d4c5 d6c5 b2b4 c5d6 g3d6 d8d6 c4c5 d6d7 f1c1 b7b6 b1d2 b6c5 c1c5 f8c8 a1c1 c8c5 c1c5 d7b7 d3c2 a8c8 c5c8 b7c8 c2c8 e7c8');
INSERT INTO evaluations VALUES(26595,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(26595,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(26595,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(26595,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(26595,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(26595,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(26595,36,'cp 35','e7e6 g1f3 f8d6 f4g3 g8f6 e1g1 e8g8 c2c4 c6e7 d3b3 b7b6 b1c3 d6g3 h2g3 a8c8 f1d1 h7h5 a1c1 c7c6 c3e2 f6e4 e2f4 g7g6 f3e5 g8g7 b3a3 a7a5 c4c5 b6c5 d4c5 d8c7 f4d3 c8a8');
INSERT INTO evaluations VALUES(26833,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26833,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(26833,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(26833,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(26833,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(26833,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(26833,36,'cp 23','g1f3 f8d6 f4g3 h7h5 a2a3 h5h4 g3d6 c7d6 h2h3 a8c8 e1g1 g8e7 d3e2 c6a5 e2b5 c8c6 b1d2 a7a6 b5d3 e7g6 c2c3 e8g8 e3e4 d5e4 d3e4 d6d5 e4g4 d8f6 g2g3 c6b6 a1b1 h4g3 f2g3 f6f5 b2b4 a5c6 g4f5 e6f5 g1g2');
INSERT INTO evaluations VALUES(26834,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(26834,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26834,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(26834,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(26834,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(26834,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(26834,36,'cp 33','f8d6 f4g3 h7h5 a2a3 h5h4 g3d6 c7d6 h2h3 a8c8 e1g1 g8e7 f1d1 d8c7 d3e2 c6a5 d1c1 c7c4 e2d1 a5b3 b1d2 b3d2 f3d2 c4a6 c2c3 c8c6 a3a4 a6d3 c3c4 e8g8 c1c3 d3f5 c4d5 c6c3 b2c3 e7d5 c3c4 d5f6 d1f3 f5f3');
INSERT INTO evaluations VALUES(14299,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(14299,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14299,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14299,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14299,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(14299,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(14299,36,'cp 40','c2c4 e7e6 e2e3 f8d6 c4c5 d6e7 a2a3 a7a5 b1c3 e8g8 f1d3 b7b6 c5b6 c7b6 d1e2 c8b7 c1d2 e7d6 e1g1 a8c8 f1c1 e6e5 d4e5 c6e5 c3b5 c8c1 a1c1 e5d3 e2d3 b7a6 a3a4 f6e4 d2c3 d6c5 c3d4 f8e8 c1d1 c5b4 d3c2 d8d7 f3e5 e8c8 e5d7 c8c2 d7b6 c2f2');
INSERT INTO evaluations VALUES(19506,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(19506,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19506,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(19506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19506,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19506,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(19506,36,'cp 8','g7g6 c2c4 f8g7 b1c3 d7d6 e1g1 c8f5 h2h3 e7e5 d2d3 g8e7 a1b1 e8g8 a2a3 a7a5 c1g5 f7f6 g5d2 f5e6 e2e4 a8b8 c3d5 b7b5 b2b4 a5b4 a3b4 b5c4 b4b5 e6d5 e4d5 c6a5 d1e2 c4d3 e2d3 c5c4 d3e2');
INSERT INTO evaluations VALUES(19299,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19299,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19299,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(19299,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(19299,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(19299,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19299,36,'cp 5','c2c4 f8g7 b1c3 d7d6 e1g1 c8f5 a2a3 d8d7 b2b4 g8f6 d2d3 c5b4 a3b4 c6b4 f3d4 f5h3 g2h3 d7h3 d1a4 f6d7 a4b4 g7d4 c1e3 d4c3 b4c3 e8g8 a1a7 a8a7 e3a7 b7b6 c3d4 h3e6 f1e1 f8a8 a7b6 a8a2 d4e3 d7b6 e3b6 a2e2 e1e2 e6e2 b6d4 h7h5 d4e4 e2d1 g1g2 d1b3 e4a8 g8g7 a8f3 b3c2 f3e4 c2b3 e4e2 g7f8 h2h3 f8g7');
INSERT INTO evaluations VALUES(38805,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(38805,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(38805,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(38805,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(38805,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(38805,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(38805,36,'cp 16','c7c5 c4d5 e6d5 d4c5 g8f6 g1f3 e8g8 f1e2 b8c6 e1g1 b4c5 a2a3 a7a6 b2b4 d5d4 e3d4 c5d4 c1b2 c8e6 d1c2 d8e7 a1d1 f8d8 f1e1 a8c8 e2f1 d4a7 f3g5 c6d4 c2d3 h7h6 g5e6 d4e6 d3e2 a7d4 d1d3 d4a7 g2g3 d8d3 e2d3 c8d8 d3e2 a7d4 f1g2 e7c7');
INSERT INTO evaluations VALUES(22016,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(22016,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(22016,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(22016,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(22016,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(22016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22016,36,'cp 28','d2d4 g7g5 h2h4 c6a5 c4d3 g5g4 f3g1 a5c6 d4d5 c6e5 c1f4 g8e7 h4h5 f8g7 g1e2 e8g8 b1c3 f7f5 e1g1 f5e4 c3e4 e7f5 d1d2 e5d3 c2d3 h7h6 e4g3 f5g3 e2g3 c8d7 f4h6 g7d4 h6e3 d8g5 a1e1 f8f1 g1f1 a8f8 f1g1 f8f7 e1e2 g5d5 e3d4 d5d4 g1h2 d4f6 d2e3 a7a6 b2b3 d7c6 d3d4 f6f4 e3f4 f7f4');
INSERT INTO evaluations VALUES(22018,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22018,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22018,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(22018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22018,36,'cp 17','g7g5 h2h4 c6a5 c4d3 g5g4 f3g1 a5c6 d4d5 c6e5 c1f4 g8e7 h4h5 f8g7 b1c3 e8g8 g1e2 f7f5 d1d2 f5e4 c3e4 e5d3 d2d3 d8e8 e1c1 c8f5 f4d2 e7g6 e2g3 g6f4 d2f4 f5e4 g3e4 f8f4 d1e1 e8e5 c2c3 e5f5 h5h6 g7e5 e1e2 f5g6 c1c2 a8e8 a2a3 e5g7 e4g3 g6d3 c2d3 g7e5 h1e1');
INSERT INTO evaluations VALUES(21251,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(21251,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(21251,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(21251,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(21251,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(21251,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(21251,36,'cp 4','h2h4 c6a5 c4d3 g5g4 f3g1 a5c6 d4d5 c6e5 c1f4 g8e7 h4h5 f8g7 d1d2 e8g8 g1e2 f7f5 b1c3 f5e4 c3e4 e5d3 d2d3 c8f5 e1g1 f5e4 d3e4 e7f5 f4d2 d8e7 e4g4 f5e3 d2e3 e7e3 g1h2 g8h8 e2f4 g7e5 g2g3 e5f4 f1f4 f8f4 g4f4 e3e2 h2g1 e2h5 f4f6 h8g8 f6e6 h5f7 a1e1 a8f8 e1e4 f7e6 d5e6');
INSERT INTO evaluations VALUES(47435,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(47435,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(47435,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(47435,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47435,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(47435,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47435,36,'cp 28','d5b6 c4e2 c7c5 d2d4 b8c6 d4c5 d8d1 e2d1 b6a4 e1g1 a4c5 b1c3 c8f5 f1e1 e7e6 f3h4 f5g6 h4g6 h7g6 g2g3 a8d8 a1b1 f8e7 h2h4 e8g8 c1f4 e7f6 c3e4 c5e4 e1e4 c6e7 f4e3 b7b6 c2c3 e7d5 e3d2 b6b5 d1b3 a7a6 b1d1');
INSERT INTO evaluations VALUES(46941,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(46941,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(46941,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46941,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(46941,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46941,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(46941,36,'cp 43','c4e2 c7c5 d2d4 b8c6 d4c5 d8d1 e2d1 b6a4 e1g1 g7g6 b1a3 f8g7 c2c3 a4c5 c1e3 c5e6 d1b3 e8g8 f1d1 e6c7 d1d2 h7h6 a1d1 c8e6 b3e6 c7e6 g1f1 f8d8 f1e2 d8d2 d1d2 a8d8 g2g3 d8d2 f3d2 f7f5 e2d3 g8f7 d3c4');
INSERT INTO evaluations VALUES(57165,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(57165,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(57165,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57165,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(57165,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57165,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(57165,36,'cp -17','f8c5 b1c3 g8f6 d2d4 e5d4 e2d4 e8g8 c1g5 h7h6 g5h4 f8e8 d1d3 b8c6 d4c6 b7c6 e1c1 a8b8 d3f3 d7d6 h2h3 c5b4 h4f6 d8f6 f3f6 g7f6 f1d3 b4c3 b2c3 e8e5 a2a4 h6h5 c1d2 e5a5 d1b1 b8b6 g2g3 c6c5 b1b6 a7b6 d3b5 g8f8 h3h4 f6f5 h1e1 f5e4 e1e4 c8e6 c3c4');
INSERT INTO evaluations VALUES(18751,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(18751,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(18751,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(18751,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(18751,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(18751,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(18751,36,'cp -41','g8f6 d2d3 d7d5 e4d5 e6d5 c4a2 c8e6 g1e2 f8e7 e1g1 e8g8 h2h3 b7b6 f1e1 h7h6 e2f4 d8d7 c3e2 e7d6 f4e6 f7e6 c1f4 g8h8 c2c3 a8e8 f4d6 d7d6 e2g3 e6e5 a2b3 b6b5 b3c2 a7a5 d1d2 a5a4 d2e2 g7g6 a1d1');
INSERT INTO evaluations VALUES(6744,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6744,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6744,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6744,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6744,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6744,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6744,36,'cp 0','d1b3 d8e7 e1g1 a5b6 c3d4 c6d4 f3d4 b6d4 b1c3 g8f6 c1g5 e8g8 c3d5 e7c5 d5f6 d4f6 g5f6 g7f6 a1d1 c7c6 b3g3 g8h8 d1d6 c5c4 g3h4 c4c3 d6d3 c3d3 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(6742,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6742,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6742,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6742,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6742,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6742,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6742,36,'cp 0','d8e7 e1g1 a5b6 c3d4 c6d4 f3d4 b6d4 b1c3 g8f6 c1g5 e8g8 c3d5 e7c5 d5f6 d4f6 g5f6 g7f6 a1d1 c7c6 b3g3 g8h8 d1d6 c5c4 g3h4 c4c3 d6d3 c3d3 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(401,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(401,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(401,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(401,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(401,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(401,36,'cp 0','e1g1 a5b6 c3d4 c6d4 f3d4 b6d4 b1c3 g8f6 c1g5 e8g8 c3d5 e7c5 d5f6 d4f6 g5f6 g7f6 a1d1 c7c6 b3g3 g8h8 d1d6 c5c4 g3h4 c4c3 d6d3 c3d3 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(402,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(402,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(402,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(402,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(402,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(402,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(402,36,'cp 0','a5b6 c3d4 c6d4 f3d4 b6d4 b1c3 g8f6 c1g5 e8g8 c3d5 e7c5 d5f6 d4f6 g5f6 g7f6 a1d1 c7c6 b3g3 g8h8 d1d6 c5c4 g3h4 c4c3 d6d3 c3d3 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(9158,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(9158,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(9158,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(9158,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(9158,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(9158,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(9158,36,'cp 99','f1e1 e8g8 f3e5 c6e5 e1e5 c7c6 d1f3 c8e6 b1d2 b7b5 c4b3 c5b6 e5e1 a7a5 a2a3 a5a4 b3a2 b6a5 e1f1 d8d7 d2e4 e6g4 e4c5 d7c8 f3g3 f8e8 h2h3 g4h5 c5e4 a5c7 g3h4 h5g6 c1d2 c8e6 e4g3 e6d7 f1e1 c7e5 a1b1 h7h6 h4g4 d7g4 h3g4 d5f6 g4g5 h6g5 d2g5 f6g4 g3e2 e5d6 g5d2 d6c5');
INSERT INTO evaluations VALUES(9159,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(9159,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(9159,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(9159,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(9159,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(9159,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(9159,36,'cp 69','e8g8 f3e5 c6e5 e1e5 c7c6 d1f3 c8e6 b1d2 b7b5 c4b3 c5b6 e5e1 d5b4 f3d1 d8d4 e1f1 e6g4 d2f3 d4f6 a2a4 a8e8 c1d2 a7a5 d3d4 b5a4 a1a4 b6c7 f1e1 e8e1 d2e1 f8e8 e1d2 g4f3 d1f3 f6f3 g2f3 e8e2 d2b4 a5b4 a4a6');
INSERT INTO evaluations VALUES(3746,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(3746,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(3746,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(3746,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(3746,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(3746,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(3746,36,'cp 74','f3e5 c6e5 e1e5 c7c6 d1f3 c8e6 b1d2 b7b5 c4b3 c5b6 a2a4 d5b4 f3e2 e6b3 d2b3 f8e8 e5e8 d8e8 c1e3 b4d5 a4b5 c6b5 d3d4 b5b4 e2d3 a7a5 b3d2 e8c6 d2c4 b6c7 e3d2 h7h5 h2h4 a5a4 g2g3 a8c8 c4e3 d5e3 d2e3 c7d6 d4d5 c6d7 c2c4 b4c3 b2c3 d7g4');
INSERT INTO evaluations VALUES(3734,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(3734,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(3734,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(3734,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(3734,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(3734,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(3734,36,'cp 73','c6e5 e1e5 c7c6 d1f3 c8e6 b1d2 b7b5 c4b3 c5b6 e5e1 d5b4 f3d1 d8d4 e1f1 e6g4 d2f3 d4d6 h2h3 g4h5 a2a4 a8e8 c1d2 a7a5 f1e1 e8e1 d2e1 f8e8 e1d2 h7h6 d3d4 b5a4 a1a4 d6d4 f3d4 h5d1 d2b4 b6d4 a4a5 e8e2 b3c4');
INSERT INTO evaluations VALUES(44898,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(44898,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(44898,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(44898,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44898,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(44898,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(44898,36,'cp 32','e2e3 e7e6 g1f3 f8d6 f4d6 d8d6 b1c3 e8g8 a1c1 f8d8 d1c2 d6e7 a2a3 h7h6 f1d3 d5c4 d3c4 c6c5 d4c5 e7c5 c4e2 c8d7 e1g1 d7e8 f1d1 d8d1 c2d1 b8c6 b2b4 c5f8 h2h3 a8d8 d1c2 c6e7 c2b2 d8c8 f3e5 e7d5 c3d5 c8c1 b2c1 f6d5');
INSERT INTO evaluations VALUES(44896,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(44896,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(44896,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44896,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(44896,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(44896,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(44896,36,'cp 31','e7e6 g1f3 f8d6 f4d6 d8d6 b1c3 e8g8 d1c2 f8d8 a1c1 d6e7 a2a3 h7h6 h2h3 b8d7 c4d5 e6d5 f1d3 b7b6 e1g1 c6c5 d3e2 c8b7 d4c5 b6c5 b2b4 c5c4 f3d4 a7a5 f1d1 a5b4 d4f5 e7f8 a3b4 d8b8 c1b1 g7g6');
INSERT INTO evaluations VALUES(45627,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(45627,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45627,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(45627,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(45627,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(45627,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(45627,36,'cp 36','g1f3 f8d6 f4d6 d8d6 d1b3 b7b6 c4d5 e6d5 f1e2 e8g8 e1g1 c8a6 e2a6 b8a6 b1c3 a6c7 a1c1 h7h6 c1c2 f8c8 f1d1 c7e6 h2h3 c8d8 b3a4 a7a5 a4b3 b6b5 c2c1 a8b8 b3c2 d8c8 c3e2 f6e4 b2b3 c8c7 f3e5 b8c8');
INSERT INTO evaluations VALUES(45628,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(45628,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(45628,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45628,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(45628,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(45628,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45628,36,'cp 20','f8d6 f4d6 d8d6 b1c3 e8g8 a1c1 b8d7 h2h3 a7a6 c4d5 c6d5 f1d3 b7b5 e1g1 c8b7 f3d2 b7c6 c3e2 b5b4 d2b3 c6a4 d1d2 a4b3 a2b3 a6a5 c1c2 d7b6 f1c1 f8c8 c2c5 g7g6 d2c2 g8g7 e2f4 f6d7 c5c7 d6c7 c2c7 c8c7 c1c7');
INSERT INTO evaluations VALUES(40540,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40540,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40540,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40540,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(40540,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40540,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40540,36,'cp 25','f4d6 d8d6 b1c3 e8g8 a1c1 d6e7 d1c2 b8d7 h2h3 h7h6 c4d5 e6d5 f1d3 f8e8 e1g1 f6e4 c3e2 e4g5 f3d2 d7f6 e2f4 c8d7 f1e1 a7a5 d2f1 g5e6 f4e6 d7e6 f2f3 f6h5 a2a3 g7g6 c2f2 e7d8 f1g3 h5g3 f2g3 h6h5');
INSERT INTO evaluations VALUES(22738,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(22738,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(22738,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22738,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(22738,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(22738,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(22738,36,'cp 3','a2a3 g8e7 b2b4 d7d5 c4d5 e7d5 c1b2 f8d6 d1c2 e8g8 g1f3 d8e7 b1c3 d5c3 c2c3 f7f6 b4b5 c6d8 f1c4 c8e6 d2d4 e5d4 f3d4 e6c4 c3c4 e7f7 c4e2 f8e8 e1g1 a7a6 f1d1 a6b5 d4b5 d6f8 a3a4 d8e6 b2c3 c7c6 b5d6 f8d6');
INSERT INTO evaluations VALUES(22745,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22745,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22745,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22745,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22745,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22745,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22745,36,'cp 0','g8f6 b2b4 d7d5 c4d5 f6d5 d1c2 f8d6 c1b2 a7a6 g1f3 d8e7 f1c4 d5b6 c4d3 h7h6 b1c3 e8g8 e1g1 f7f5 e3e4 e7f7 a1e1 b6c4 b2c1 c4b6');
INSERT INTO evaluations VALUES(16468,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(16468,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(16468,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(16468,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16468,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(16468,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(16468,36,'cp 9','b2b4 d7d5 c4d5 f6d5 c1b2 f8d6 d1c2 e8g8 g1f3 d8e7 b1c3 d5c3 c2c3 a7a6 f1c4 c8e6 e1g1 e6c4 c3c4 e5e4 f3e1 f8e8 d2d3 c6d8 a1c1 a6a5 d3e4 e7e4 b4b5 e4c4 c1c4 c7c6 b5c6 a8c8 e1f3 b7b5 c4c2 c8c6 f1c1 c6c2 c1c2 d8e6 g2g3 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(45100,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(45100,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(45100,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(45100,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(45100,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(45100,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(45100,36,'cp 72','g7g6 c1e3 f8g7 h2h3 e8g8 a2a4 b8d7 f1e2 e7e5 d4e5 d7e5 e1g1 e5f3 e2f3 d8c7 d1d2 c8e6 f1d1 f8d8 e3g5 d8d7 a4a5 d6d5 g5f4 c7d8 e4e5 f6e4 f3e4 d5e4 d2e3 d7d1 c3d1 d8c7 e3d4 c6c5 d4e4 e6f5 e4e2 a8e8 d1e3 f5e6 a5a6 b7a6 e2a6');
INSERT INTO evaluations VALUES(44726,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(44726,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(44726,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(44726,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(44726,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(44726,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(44726,36,'cp 75','c1e3 f8g7 h2h3 e8g8 a2a4 b7b6 f1d3 d8c7 a4a5 b6a5 e1g1 b8d7 f1e1 f8e8 d1d2 c8b7 e3f4 e7e5 d4e5 d6e5 f4e3 a8d8 d2c1 b7c8 c3a4 g7f8 e3g5 f8e7 d3c4 f6h5 c1e3 d7b6 c4b3 h5f4 a4c5 e7g5 f3g5');
INSERT INTO evaluations VALUES(44529,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44529,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(44529,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44529,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(44529,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44529,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(44529,36,'cp 26','d7d5 c4d5 f6d5 f1g2 d5c7 a2a3 b8c6 b2b4 e7e5 g1f3 f7f6 a1b1 c5b4 a3b4 a7a6 e1g1 f8e7 b4b5 c7b5 c3b5 a6b5 b1b5 e5e4 f3h4 c6d4 b5b1 c8g4 f2f3 g4h5 e2e3 d4f3 h4f3 d8d5 d1b3 d5b3 b1b3 e4f3 g2f3 h5f3 f1f3 e8f7 c1b2 h8d8 b2c3 d8d7 g1f2 a8a2 f2e2 e7a3 b3b5 a3d6 e3e4');
INSERT INTO evaluations VALUES(45341,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(45341,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45341,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45341,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(45341,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(45341,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(45341,36,'cp 13','c4d5 f6d5 f1g2 d5c7 a2a3 g7g6 h2h4 h7h5 d2d3 f8g7 c1e3 c7e6 a1b1 e8g8 g1f3 b8c6 e1g1 e6d4 c3e4 c8g4 e4c5 g4f3 e2f3 b7b6 c5e4 a8c8 d1a4 d4f5 b1c1 d8d7 c1c2 c8c7 f1c1 f5d4 e3d4 c6d4 a4d7 c7d7 c2c7 d4e2 g1f1 e2c1 c7d7 g7b2 d7a7 c1d3 a7e7 b2a3');
INSERT INTO evaluations VALUES(45331,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45331,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(45331,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(45331,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45331,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45331,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(45331,36,'cp 6','f6d5 f1g2 d5c7 a2a3 g7g6 d2d3 f8g7 h2h4 h7h5 c1e3 c7e6 g1f3 b8c6 e1g1 e8g8 a1b1 a8b8 c3e4 b7b6 b2b4 c5b4 a3b4 c6d4 e3d4 e6d4 f3d4 d8d4 e2e3 d4d7 d1b3 c8b7 b4b5 b8c8 d3d4 e7e6 f1c1 f8d8 c1c8 d8c8 e4g5 d7e7 g1h1 b7g2 h1g2 e7d7');
INSERT INTO evaluations VALUES(45505,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(45505,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(45505,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45505,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(45505,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(45505,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45505,36,'cp 16','f1g2 d5c7 a2a3 g7g6 h2h4 h7h5 d2d3 f8g7 c1e3 c7e6 g1f3 b8c6 e1g1 e8g8 a1b1 a8b8 c3e4 c6d4 b2b4 c5b4 a3b4 d4b5 d1d2 b7b6 e3h6 c8b7 h6g7 e6g7 d2b2 g7f5 f1c1 d8d7 c1c4 b5d6 e4d6 f5d6 c4d4 b7f3 g2f3 b8c8 g1g2 f8d8 b1c1 c8c1 b2c1 e7e6 c1e3');
INSERT INTO evaluations VALUES(45506,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(45506,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45506,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(45506,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45506,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45506,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(45506,36,'cp 16','d5c7 a2a3 g7g6 d2d3 f8g7 h2h4 h7h5 c1e3 c7e6 g1f3 b8c6 e1g1 e8g8 a1b1 a8b8 c3e4 c6d4 b2b4 c5b4 a3b4 d4b5 d1d2 b7b6 e3h6 c8b7 h6g7 g8g7 d2b2 f7f6 f1c1 b8c8 c1c8 b7c8 b1a1 c8b7 e4g5 e6g5 f3g5 b7d5 g5e4 d8d7 e4c3 d5g2 c3b5 d7b5 g1g2');
INSERT INTO evaluations VALUES(45452,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(45452,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45452,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(45452,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(45452,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45452,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(45452,36,'cp 26','c2c3 b8c6 c1f4 c8g4 d1b3 d8d7 b1d2 e7e6 g1f3 f8d6 f4d6 d7d6 e1g1 e8g8 f1e1 d6e7 f3e5 c6e5 e1e5 g4h5 a1e1 e7c7 h2h3 h5g6 d3g6 h7g6 a2a4 f8c8 d2f3 f6e4 h3h4 a7a6 f3g5 e4d6 b3c2 d6f5 c2e2');
INSERT INTO evaluations VALUES(45448,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(45448,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(45448,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(45448,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45448,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(45448,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(45448,36,'cp 22','b8c6 c1f4 c8g4 d1b3 c6a5 b3a4 a5c6 b1d2 e7e6 d3b5 d8b6 g1f3 g4f3 d2f3 a7a6 b5c6 b6c6 a4d1 b7b5 f3e5 c6b7 e5d3 f8e7 e1g1 e8g8 a2a4 f6e4 d1b3 h7h6 g2g3 f8c8 f1e1 e4d6 f4d6 e7d6 g1h1 g8h8 a4b5');
INSERT INTO evaluations VALUES(51452,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(51452,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(51452,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(51452,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(51452,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(51452,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(51452,36,'cp 240','f3e5 d7d6 f1b5 b8d7 e5d7 c8d7 b5e2 d6d5 e4d5 f8d6 e1g1 g8e7 b1c3 e8g8 d2d4 f8e8 e2d3 e7f5 c3b5 d8f6 c2c3 a7a6 b5d6 f5d6 c3c4 d7f5 d3f5 f6f5 b2b3 e8e4 f2f3 e4e7 a2a4 f5f6 c1a3 a8e8 a3d6 c7d6');
INSERT INTO evaluations VALUES(51438,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(51438,31,'cp 241',NULL);
INSERT INTO evaluations VALUES(51438,32,'cp 248',NULL);
INSERT INTO evaluations VALUES(51438,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(51438,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(51438,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(51438,36,'cp 237','d7d6 f1b5 b8d7 e5d7 c8d7 b5e2 d6d5 e4d5 f8d6 d2d3 g8e7 b1c3 e8g8 e1g1 f8e8 f1e1 e7g6 c3e4 d6e5 e2f3 f7f5 c1g5 d8c8 e4c3 e5d6 f3h5 e8e1 d1e1 g6f4 g5f4 d6f4 g2g3 f4d6 g1g2 g7g6 h5f3 c8f8 e1d2 g8g7 a1e1 a8e8 c3e2 f8f6');
INSERT INTO evaluations VALUES(55656,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(55656,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(55656,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(55656,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(55656,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(55656,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(55656,36,'cp -9','d1b3 d8d7 b1c3 b8c6 b3c2 g8f6 h2h3 f8e7 g1f3 c6e5 f3e5 d6e5 e1g1 e8g8 f1d1 d7c6 c4b5 c6e6 b5a4 e6a6 a4b5 a6a5 c1d2 c8e6 b5c4 a5b6 d2e3 c7c5 c4e6 b6e6 e3g5 a8c8 g5f6 e7f6 c2b3 a7a6 b3e6 f7e6 d1d6 c8d8 d6e6 d8d2 e6b6 f6g5 a1d1 d2f2 b6b7');
INSERT INTO evaluations VALUES(55653,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55653,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(55653,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(55653,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55653,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(55653,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(55653,36,'cp -34','d8d7 b1c3 b8c6 b3c2 g8f6 h2h3 f8e7 g1f3 c6e5 f3e5 d6e5 c1e3 e8g8 e1g1 c7c6 f1d1 d7c7 c4b3 c7b8 a2a4 a7a5 b3c4 c8d7 c2b3 e7b4 e3g5 b8e8 g5f6 g7f6 d1d3 e8e7 a1d1 d7e6 c3e2 e6c4 b3c4 a8d8');
INSERT INTO evaluations VALUES(38813,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38813,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38813,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(38813,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(38813,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38813,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38813,36,'cp 20','c1d2 b4e7 b1c3 g8f6 d2f4 e8g8 e2e3 b7b6 f1d3 c8a6 c4d5 f6d5 c3d5 d8d5 e1g1 d5b7 a1c1 c7c5 d4c5 b6c5 d3a6 b7a6 f3e5 b8c6 d1d7 c6e5 f4e5 f8e8 f1d1 f7f6 e5c7 e7f8 a2a4 a8c8 b2b3 a6b7 c7g3 b7b3 d7a7');
INSERT INTO evaluations VALUES(38812,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(38812,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38812,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38812,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(38812,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38812,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(38812,36,'cp 9','b4e7 b1c3 g8f6 d2f4 e8g8 e2e3 b8d7 c4c5 c7c6 f1d3 b7b6 b2b4 a7a5 a2a3 c8a6 d3a6 a8a6 e1g1 d8a8 h2h3 a8b7 d1e2 a5b4 a3b4 f8a8 a1a6 a8a6 f1b1 b7a8 e2b2 h7h6 f4c7 f6e8 b4b5 c6b5 b2b5 d7c5 c7b6 c5a4 c3a4 a6a4 b5d7 e7d6 f3e5 d6e5 d4e5 a4a1');
INSERT INTO evaluations VALUES(47162,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(47162,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(47162,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(47162,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(47162,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(47162,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(47162,36,'cp -80','c8f5 e2e3 e7e6 f1d3 f8e7 e1g1 e8g8 b1c3 f5g6 f3e5 c7c5 b2b3 a7a6 c1b2 c5d4 e3d4 b8c6 e5c6 b7c6 c3a4 d8c7 a1c1 a6a5 d1e1 f8e8 e1f2 c7a7 h2h3 e7b4 g1h2 f6e4 d3e4 g6e4 a4c5 f7f6 c5e4 d5e4 c2c4 a5a4');
INSERT INTO evaluations VALUES(29553,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(29553,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(29553,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(29553,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(29553,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(29553,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(29553,36,'cp -63','e2e3 e7e6 f1d3 f8e7 e1g1 e8g8 b1c3 f5g6 f3e5 c7c5 b2b3 b8c6 e5c6 b7c6 c3a4 c5d4 e3d4 f6e4 c1e3 e7a3 d3e4 g6e4 d1d2 d8f6 a4c5 a3c5 d4c5 f6g6 f1f2 f7f6 e3d4 a8e8 a1f1 h7h5 a2a3 h5h4 h2h3 f8f7 f1e1 e6e5 f4e5 f6e5 f2f7');
INSERT INTO evaluations VALUES(31153,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(31153,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(31153,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(31153,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31153,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(31153,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(31153,36,'cp 13','e7e6 b1c3 f8b4 c4d5 c6d5 f1b5 b8d7 g1f3 a7a6 b5d3 b4c3 b2c3 g8e7 e1g1 f5d3 d1d3 d7b6 a1b1 a8c8 b1b3 h7h6 c1d2 e8g8 f1b1 c8c6 h2h4 d8c7 g2g4 f8b8 h4h5 b6c4 g4g5 e7f5 g5h6 c4d2 h6h7 g8h7 d3d2 h7g8 d2g5');
INSERT INTO evaluations VALUES(31508,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(31508,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31508,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31508,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31508,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31508,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31508,36,'cp 0','b1c3 f8b4 c4d5 c6d5 a2a3 b4c3 b2c3 d8c7 f1b5 b8d7 g1e2 a7a6 b5d3 g8e7 c1g5 d7b6 e1g1 f5d3 d1d3 h7h6 g5e7 c7e7 a3a4 e8g8 e2g3 b6c4 f2f4 f7f6 e5f6 f8f6 g3h5 f6f5 h5g3');
INSERT INTO evaluations VALUES(31506,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31506,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(31506,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31506,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31506,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31506,36,'cp 2','f8b4 d1b3 d8a5 c1d2 g8e7 g1f3 e8g8 a2a3 b4c3 d2c3 a5b6 b3d1 c6c5 d4c5 b6c7 d1d2 d5c4 d2d6 f8c8 f1c4 c7c5 c4a2 f5e4 f3g5 e4d5 a2b1 h7h6 g5e4 d5e4 b1e4 c5c4 d6b4 e7d5 b4c4 c8c4 e4d5 e6d5 e1g1');
INSERT INTO evaluations VALUES(28089,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28089,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28089,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28089,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(28089,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28089,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28089,36,'cp 12','d1b3 d8a5 g1f3 g8e7 f1e2 d5c4 e2c4 c6c5 e1g1 c5d4 c3b5 e8g8 b5d4 b8c6 d4f5 e7f5 a2a3 b4e7 c1d2 a5c7 a1c1 e7c5 c4d3 c6e5 d3f5 e6f5 b3d5 e5f3 d5f3 c7b6 f3f5 c5d4 b2b4 h7h6 g2g3 a8d8 d2c3 b6e6 f5e6 f7e6 c3d4 d8d4 c1c7 f8f7 c7f7 g8f7 f1c1 f7g6 c1c7 d4d3 c7b7 d3a3 g1g2');
INSERT INTO evaluations VALUES(31977,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(31977,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(31977,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(31977,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(31977,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(31977,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31977,36,'cp -28','g2g4 f5g6 g1e2 g8f6 g4g5 f6d5 f1g2 d5c3 b2c3 f7f5 g5f6 e7f6 e1g1 f8d6 c3c4 c7c6 c4c5 d6c7 c2c3 b7b6 e2f4 c7f4 c1f4 e8g8 d1a4 b6c5 d4c5 d8c8 g2e4 g6e4 a4e4 f8e8 e4c4 g8h8 f4d6 b8d7 f1e1 d7e5 d6e5 f6e5 c4e4 c8h3 e4c6 a8c8 c6e4');
INSERT INTO evaluations VALUES(1782,30,'cp -1225',NULL);
INSERT INTO evaluations VALUES(1782,31,'cp -1225',NULL);
INSERT INTO evaluations VALUES(1782,32,'cp -1225',NULL);
INSERT INTO evaluations VALUES(1782,33,'cp -1238',NULL);
INSERT INTO evaluations VALUES(1782,34,'cp -1244',NULL);
INSERT INTO evaluations VALUES(1782,35,'cp -1240',NULL);
INSERT INTO evaluations VALUES(1782,36,'cp -1240','g2e4 d1e2 d4e2 d2d3 e4g2 f7h8 d7d5 c4d5 g2d5 c1e3 e2d4 e3d4 d5d4 b1c3 g7g5 e1c1 d4h8 f2f4 g5g4 d1e1 e8f7 e1e5 h7h6 c3e4 g8e7 e4c3 h8f6 c3d5 e7d5 e5d5 f6e6 d5a5 e6e2 f1d1 e2h2 a5a4 b7b5 a4e4 c7c6 c1b1 a7a5 e4e5 h2f4');
INSERT INTO evaluations VALUES(1783,30,'cp -1225',NULL);
INSERT INTO evaluations VALUES(1783,31,'cp -1217',NULL);
INSERT INTO evaluations VALUES(1783,32,'cp -1217',NULL);
INSERT INTO evaluations VALUES(1783,33,'cp -1221',NULL);
INSERT INTO evaluations VALUES(1783,34,'cp -1217',NULL);
INSERT INTO evaluations VALUES(1783,35,'cp -1221',NULL);
INSERT INTO evaluations VALUES(1783,36,'cp -1194','d1e2 d4e2 d2d3 e4g2 c1e3 d7d5 c4d5 g2d5 f7h8 e2d4 e3d4 d5d4 b1c3 g7g5 f1g1 f8h6 e1c1 d4h8 c1b1 g8e7 g1g3 e7f5 g3g1 e8f7 c3e4 h8d8 f2f3 c8e6 d1f1 a7a5 f1f2 f5d4 g1g3 e6c8 g3g4 c7c6 g4g2 a5a4 g2g3 d4e6 g3g4 a4a3 b2a3 a8a3');
INSERT INTO evaluations VALUES(1784,30,'cp -1243',NULL);
INSERT INTO evaluations VALUES(1784,31,'cp -1240',NULL);
INSERT INTO evaluations VALUES(1784,32,'cp -1240',NULL);
INSERT INTO evaluations VALUES(1784,33,'cp -1238',NULL);
INSERT INTO evaluations VALUES(1784,34,'cp -1247',NULL);
INSERT INTO evaluations VALUES(1784,35,'cp -1246',NULL);
INSERT INTO evaluations VALUES(1784,36,'cp -1246','d4e2 d2d3 e4g2 c1e3 d7d5 c4d5 g2d5 f7h8 e2d4 e3d4 d5d4 b1c3 g7g5 e1c1 d4h8 f1g1 f8h6 c1b1 g8e7 a2a3 h8f6 g1g2 e8f7 d1g1 f6f4 c3e2 f4h4 e2c3 g5g4 f2f3 g4f3 g2g3 h4f4 g1f1 c8g4 f1g1 f3f2 g1f1 e7f5 c3e4 f7g6 g3g4 f4g4 f1f2 g4g1 b1a2 b7b5 a2b3 a7a5 b3c3');
INSERT INTO evaluations VALUES(1780,30,'cp -1221',NULL);
INSERT INTO evaluations VALUES(1780,31,'cp -1209',NULL);
INSERT INTO evaluations VALUES(1780,32,'cp -1217',NULL);
INSERT INTO evaluations VALUES(1780,33,'cp -1231',NULL);
INSERT INTO evaluations VALUES(1780,34,'cp -1228',NULL);
INSERT INTO evaluations VALUES(1780,35,'cp -1215',NULL);
INSERT INTO evaluations VALUES(1780,36,'cp -1224','d2d3 e4g2 c1e3 d7d5 c4d5 g2d5 f7h8 e2d4 e3d4 d5d4 b1c3 g7g5 e1c1 d4h8 f1g1 f8h6 c3d5 e8d8 c1b1 h8d4 d5e3 d4f4 g1g2 h6g7 a2a3 h7h6 b1a1 g8e7 c2c3 h6h5 d1g1 g7h6 d3d4 f4f6 g1b1 b7b6 e3c4 h5h4 b1g1');
INSERT INTO evaluations VALUES(43082,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(43082,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(43082,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(43082,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(43082,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(43082,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(43082,36,'cp 192','h2h4 h7h5 c1g5 f7f6 g5e3 d7d5 f1d3 e8g8 e1g1 b8c6 f1e1 c6b4 d3f1 c7c5 a2a3 c5d4 f3d4 b4c6 e4d5 e6d5 c3e2 c8g4 f2f3 c6d4 e2d4 g4d7 c2c3 e7c6 d4b5 d7e8 e3c5 e8f7 d1d2 f8e8 e1e8 f7e8 b5d6 e8f7 d6f7 g8f7 a1d1 c6e7 c3c4 d8c7 c5e7 f7e7');
INSERT INTO evaluations VALUES(53859,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(53859,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(53859,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(53859,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(53859,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(53859,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(53859,36,'cp 125','c2c4 d6d5 e4e5 f8b4 c1d2 b4d2 b1d2 b7b6 f1d3 g8e7 a1c1 c6c5 c4d5 e7d5 e1g1 e8g8 d4c5 b6c5 d1c2 h7h6 c2c5 c8a6 d3a6 b8a6 c5a3 a6b4 d2e4 a7a5 c1c4 d5f4 e4f6 g7f6 c4f4 f6f5 f4c4 d8d3 b2b3 a8d8');
INSERT INTO evaluations VALUES(53854,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(53854,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(53854,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(53854,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(53854,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(53854,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(53854,36,'cp 124','f8e7 b1c3 g8f6 c1f4 d8a5 f1d3 b8d7 e1g1 e6e5 f4e3 e8g8 h2h3 f8e8 d1d2 a5c7 d3c2 a7a6 a2a4 b7b6 a1c1 e7f8 b2b3 c8b7 d4d5 a8c8 d5c6 c7c6 c3d5 g7g6 e3h6 f6d5 h6f8 e8f8 e4d5 c6c7 f3g5');
INSERT INTO evaluations VALUES(37898,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(37898,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(37898,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(37898,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(37898,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(37898,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(37898,36,'cp 136','b1c3 g8f6 c1f4 d8a5 f1d3 b8d7 e1g1 e6e5 f4e3 e8g8 h2h3 f8e8 d4d5 d7c5 d3c2 a5b4 d1d2 b4c4 b2b3 c4a6 b3b4 c6d5 e4d5 e5e4 f3g5 c5d3 c3e4 f6e4 g5e4 d3e5 e3d4 c8f5 f1e1 f5g6 a2a4 a6c4 d4e5 d6e5 d5d6 e7f8 a1d1 a8d8 c2b1');
INSERT INTO evaluations VALUES(37896,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(37896,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(37896,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(37896,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(37896,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(37896,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(37896,36,'cp 119','g8f6 f1e2 e8g8 e1g1 b8d7 h2h3 d8c7 c1e3 f8e8 b2b4 b7b6 a1b1 a7a5 b4b5 c6c5 d4d5 d7f8 d1c2 e6e5 g1h2 f8g6 g2g3 c7d7 f3g1 e8f8 a2a4 h7h5 f2f4 h5h4 f4f5 h4g3 h2g3 g6f4 e3f4 e5f4 f1f4 f6h7 g3g2 e7f6 f4g4 d7e7 b1d1 a8b8');
INSERT INTO evaluations VALUES(40146,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(40146,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(40146,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(40146,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(40146,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(40146,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(40146,36,'cp 129','f1d3 b8d7 h2h3 e6e5 c1e3 e8g8 e1g1 a7a5 d3c2 f8e8 d1e2 d8c7 a1c1 d7f8 d4d5 f8g6 g1h2 e8f8 g2g3 c8d7 f3g1 a8e8 c2d3 h7h5 c1d1 c7d8 a2a3 c6d5 c3d5 f6d5 c4d5 h5h4 g1f3');
INSERT INTO evaluations VALUES(40147,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(40147,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(40147,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(40147,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(40147,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(40147,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(40147,36,'cp 129','b8d7 h2h3 e6e5 c1e3 e8g8 e1g1 a7a5 f1e1 f8e8 d1e2 d8c7 a1d1 d7f8 a2a3 f8e6 d4d5 e6c5 d3c2 c8d7 b2b3 g7g6 d1b1 e8c8 f3d2 e7f8 e2f3 c7d8 e3g5 f8g7 e1d1 d8e7 g5e3 h7h6 b3b4 a5b4 a3b4 c5a6');
INSERT INTO evaluations VALUES(22474,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(22474,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(22474,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(22474,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(22474,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(22474,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(22474,36,'cp 100','d2d4 e5d4 f3d4 f8g7 c1e3 d7d6 d1d2 c6d4 e3d4 g8f6 f2f3 a7a6 g2g4 c7c5 d4e3 c8e6 a1d1 d8b6 g4g5 f6d7 d2d6 b6d6 d1d6 h7h6 g5h6 g7h6 f3f4 h6g7 e4e5 g6g5 f4g5 g7e5 d6d2 e5g3 e3f2 h8h2 h1h2 g3h2');
INSERT INTO evaluations VALUES(22468,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(22468,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(22468,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(22468,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(22468,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(22468,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(22468,36,'cp 85','e5d4 f3d4 f8g7 c1e3 d7d6 f1b5 c8d7 d1d2 g8f6 b5c6 b7c6 e3h6 g7h6 d2h6 f6g4 h6d2 e8g8 e1c1 c6c5 d4f3 d7c6 h2h3 g4f6 h1e1 f8e8 e4e5 c6f3 g2f3 f6h5 e5d6 e8e1 d2e1 c7d6 c3e4 d8h4 d1d6 h4h3 e1c3 h3h2 d6d1 h2f4 c1b1 a8e8 b2b3 f4e5 c3c5 e5c5');
INSERT INTO evaluations VALUES(22485,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(22485,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(22485,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(22485,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(22485,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(22485,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(22485,36,'cp 92','f3d4 f8g7 c1e3 d7d6 f2f3 g8f6 d1d2 e8g8 e1c1 d6d5 d4c6 b7c6 f1c4 c8e6 c4b3 d5e4 b3e6 f7e6 d2d8 a8d8 c3e4 f6e4 f3e4 a7a5 a2a4 g7e5 g2g3 h7h5 d1d8 f8d8 c2c3 g8g7 h1d1 e5d6 c1c2 d8b8 b2b3 c6c5 d1d3');
INSERT INTO evaluations VALUES(22483,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(22483,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(22483,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(22483,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(22483,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(22483,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(22483,36,'cp 79','f8g7 c1e3 d7d6 d1d2 g8f6 f2f3 e8g8 e1c1 d6d5 d4c6 b7c6 e3h6 c8e6 e4d5 e6d5 h6g7 g8g7 d2f4 d8e7 f1c4 f8d8 h1e1 e7c5 c4d3 d8d7 a2a3 c5d6 f4e5 a8d8 e1e3 g7f8 d3f1 f8g7 d1d3 h7h5 c3d5 c6d5 e5d6 d7d6');
INSERT INTO evaluations VALUES(6189,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(6189,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(6189,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(6189,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(6189,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(6189,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(6189,36,'cp 90','c1e3 d7d6 d1d2 g8f6 e1c1 e8g8 f2f3 d6d5 e4d5 f6d5 e3g5 d5c3 g5d8 c3a2 c1b1 f8d8 d2e1 c6d4 f1c4 c8f5 b1a2 f5c2 e1e7 d8f8 d1d4 g7d4 h1c1 c2a4 e7b4 a4d7 c4f7 f8f7 b4d4 d7e6 a2a3 c7c6 d4e3 a8e8 c1d1 f7e7 d1e1 a7a6 h2h4 e7d7 e3c5 e6f7 e1e8 f7e8');
INSERT INTO evaluations VALUES(52279,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(52279,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(52279,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(52279,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(52279,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(52279,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(52279,36,'cp 62','g1f3 b8c6 d2d4 c5d4 f3d4 f8g7 c1e3 d8a5 b1c3 d7d6 f1e2 g7d4 e3d4 c6d4 d1d4 g8f6 e1g1 e8g8 c3d5 f6d5 e4d5 e7e5 d5e6 c8e6 f1d1 f8d8 e2f3 a8b8 a2a4 a5c5 b2b3 c5d4 d1d4 a7a6 h2h3 h7h5 a1e1 g8f8 d4d3 b7b6 d3d2 b8c8 e1e4');
INSERT INTO evaluations VALUES(52278,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(52278,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(52278,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(52278,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(52278,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(52278,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(52278,36,'cp 64','f8g7 d2d4 c5d4 f3d4 b8c6 c1e3 d7d6 f1e2 c6d4 e3d4 d8a5 b1c3 g7d4 d1d4 g8f6 e1g1 e8g8 b2b3 c8e6 f2f4 a8c8 c3b5 c8d8 a2a3 e6d7 b3b4 a5b6 d4b6 a7b6 b5c3 d7e6 a1d1 d8a8 a3a4 f8c8 d1d4 g8f8 f1a1 c8c6 a1d1 a8c8 e4e5');
INSERT INTO evaluations VALUES(37779,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(37779,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(37779,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(37779,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(37779,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(37779,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(37779,36,'cp 66','d2d4 c5d4 f3d4 b8c6 c1e3 d8b6 d4b3 b6d8 d1d2 g8f6 b1c3 f6g4 e3f4 e8g8 f2f3 g4f6 f4e3 d7d6 f1e2 c8e6 e1g1 a7a5 a1d1 a5a4 b3d4 c6d4 e3d4 d8a5 f3f4 f8c8 b2b3 a4b3 a2b3 a5b4 f4f5 e6d7 d4f6 g7f6 c3d5 b4d2 d5f6 e7f6 d1d2 d7c6 f5g6 h7g6 f1f6');
INSERT INTO evaluations VALUES(37795,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(37795,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(37795,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(37795,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(37795,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(37795,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(37795,36,'cp 62','c5d4 f3d4 b8c6 c1e3 d7d6 b1c3 c6d4 e3d4 g7d4 d1d4 g8f6 f1e2 d8a5 e1g1 e8g8 b2b3 c8d7 c3d5 f6d5 e4d5 f8e8 f1d1 b7b5 c4b5 d7b5 b3b4 a5a6 e2b5 a6b5 a2a4 b5b6 a1c1 a8c8 d4e4 c8c1 d1c1 e7e5 e4c4 e5e4 a4a5 b6d8 h2h3 e4e3 f2e3 e8e3');
INSERT INTO evaluations VALUES(37859,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(37859,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(37859,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(37859,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(37859,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37859,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(37859,36,'cp 63','f3d4 b8c6 c1e3 d7d6 f1e2 c6d4 e3d4 g7d4 d1d4 g8f6 b1c3 c8e6 e1g1 e8g8 f2f4 d8a5 b2b3 a8c8 c3b5 c8d8 a2a3 e6d7 b5c3 d7c6 b3b4 a5b6 d4b6 a7b6 e2f3 d8c8 f1d1 f8d8 a3a4 e7e5 g2g3 c6d7 f3e2 d7c6 f4e5 d6e5');
INSERT INTO evaluations VALUES(44551,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(44551,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(44551,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(44551,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(44551,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(44551,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(44551,36,'cp 34','d4d5 e7e6 b1c3 a7a6 g2g3 b7b5 f1g2 a8a7 e3e4 d7d6 a2a3 f8e7 g1h3 e8g8 e1g1 e6d5 c3d5 f6d5 d1d5 c8e6 d5d1 e6c4 f1e1 e7f6 h3f4 b8c6 g2f1 c6e5 f1c4 e5c4 c2c3 c4b6 f4d5 f8e8 c1f4 f6e5 d5b6 d8b6 d1f3 a7e7 f4g5 e7e6 a1d1 b6b7 d1d5 e5f6 g5f6');
INSERT INTO evaluations VALUES(55869,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(55869,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(55869,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(55869,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55869,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(55869,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(55869,36,'cp -28','g8f6 c1f4 c7c5 e2e3 a7a6 g1f3 b7b5 g2g4 b8c6 h1g1 c8b7 h2h4 f6e4 c3e4 d5e4 f3e5 c6e5 f4e5 f7f6 e5g3 a8c8 d4c5 d8d1 e1d1 c8c5 c2c3 e6e5 d1c2 e8f7 f1e2 f8e7 a3a4 b5b4 c3c4 f7e6 g1d1 c5c8 d1d2 c8d8 d2d8 h8d8 a1d1 d8c8 h4h5 a6a5 b2b3 e7c5 c2c1 g7g6 g3h4');
INSERT INTO evaluations VALUES(48365,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(48365,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48365,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(48365,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(48365,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(48365,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(48365,36,'cp -31','c1f4 c7c5 e2e3 a7a6 g1f3 b7b5 g2g4 b8c6 h1g1 c8b7 h2h4 c5d4 e3d4 f8e7 g4g5 f6h5 f4h2 e8g8 c3e2 b5b4 a3b4 e7b4 c2c3 b4d6 h2d6 d8d6 d1d2 a8e8 e2g3 h5g3 g1g3 e6e5 d4e5 c6e5 e1c1 e5f3 g3f3 d5d4');
INSERT INTO evaluations VALUES(48296,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(48296,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48296,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(48296,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(48296,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(48296,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48296,36,'cp -38','a7a6 e2e3 c7c5 g1f3 b7b5 g2g4 b8c6 h1g1 c8b7 h2h4 h7h6 f1d3 d8b6 g4g5 h6g5 h4g5 f6d7 d1d2 g7g6 d3e2 b6a5 e1f1 b5b4 c3a2 b4a3 d2a5 c6a5 b2a3 c5d4 e3d4 a8c8 a2b4 a5c4 a1b1 c4a3 b1b3 a3c4 b4d3');
INSERT INTO evaluations VALUES(43663,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(43663,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(43663,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(43663,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(43663,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(43663,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(43663,36,'cp -25','g1f3 c7c5 e2e3 b7b5 d4c5 f8c5 f1d3 c8b7 d1e2 b8c6 e1g1 e8g8 e3e4 c6d4 f3d4 c5d4 e4d5 d4c3 b2c3 d8d5 f2f3 d5c5 e2e3 c5e3 f4e3 f6d5 e3d4 a8c8 a3a4 d5c3 a4b5 a6b5 d4c3 c8c3 f1b1 b7c6 d3b5 c3c2 b5c6 c2c6 b1b7 c6d6 b7a7 h7h6 h2h4 d6d4 h4h5');
INSERT INTO evaluations VALUES(39500,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(39500,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(39500,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(39500,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(39500,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(39500,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(39500,36,'cp 81','e4e5 b8c6 g1f3 d7d6 c1d2 d6e5 d4e5 g8h6 b1c3 h6f5 f1d3 f5d4 d2f4 g7g5 f4g3 h7h5 h2h4 g5g4 f3g5 d4f5 d3f5 e6f5 d1d8 e7d8 e1c1 c8e6 c3e2 d8e7 e2f4 h8h6 b2b3 a7a5 a2a4 b7b6 c1b2 a8d8 d1d8 c6d8 h1d1 d8b7 d1d2 b7c5 f2f3 g4f3 g2f3 e6d7');
INSERT INTO evaluations VALUES(39486,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(39486,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(39486,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(39486,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(39486,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(39486,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(39486,36,'cp 68','b8c6 g1f3 d7d6 c1d2 b7b6 f1d3 c8b7 d1e2 d8d7 e1g1 e8c8 d3a6 d6e5 d4e5 b7a6 e2a6 c8b8 a2a4 h7h5 h2h3 g8h6 d2h6 h8h6 b1c3 c6b4 a6e2 g7g5 f1d1 d7e8 f3d4 a7a5 c3b5 g5g4 c2c3 b4d5 b2b4 c7c6 b5d6 e7d6 e5d6 e6e5 d4f5 h6g6');
INSERT INTO evaluations VALUES(10034,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10034,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(10034,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10034,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10034,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10034,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10034,36,'cp 0','f6g4 e1f1 e8g8 b5c6 b7c6 h2h3 g4f6 f3e5 f6e4 b1c3 d7d5 d2d4 e4c3 b2c3 c5d6 f1e1 d8f6 e5c6 c8e6 c6e5 f8e8 a2a4 e6f5 e1e3 d6e5 d1e2 f6g6 d4e5 f5c2 e3g3 g6e4 g3e3');
INSERT INTO evaluations VALUES(10577,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10577,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(10577,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10577,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(10577,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10577,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10577,36,'cp 0','e1f1 a7a6 b5a4 e8g8 h2h3 g4f6 a4c6 d7c6 d2d3 f8e8 c1g5 h7h6 g5h4 d8d6 b1d2 b7b5 h4g3 f6h5 g3h2 d6f6 d3d4 e5d4 e4e5 f6g6 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(31838,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(31838,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(31838,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(31838,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(31838,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(31838,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(31838,36,'cp -72','e2e3 g8f6 c2c4 e7e6 g1f3 f8e7 c4d5 e6d5 f1d3 f5d3 d1d3 b8d7 b1c3 e8g8 c1d2 e7b4 e1g1 f8e8 c3a4 a7a5 a2a3 b4d2 f3d2 c7c6 a4c3 d8b6 d2f3 f6e4 f3e5 d7f6 c3a4 b6d8 a1c1 e4d6 a4c5 f6e4 c5d7 d8h4 d3e2');
INSERT INTO evaluations VALUES(31836,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(31836,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(31836,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(31836,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(31836,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(31836,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(31836,36,'cp -70','g8f6 f1d3 e7e6 g1f3 f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 e8g8 c2c4 c7c6 e1g1 f8e8 c1d2 f6e4 b1c3 e7a3 a1b1 b8a6 d3e2 e4d2 e2d2 c8e6 f1e1 e8e7 h2h3 a3b4 f3g5 e6d7 g5f3 f7f6 d2d3 g7g6 b1d1 a6c7 g1h2 a8e8 a2a4');
INSERT INTO evaluations VALUES(35115,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(35115,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(35115,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(35115,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(35115,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(35115,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(35115,36,'cp 155','b1c3 e5d4 c3d5 f6d8 c1f4 d7d6 d1d4 b8c6 d4e3 g8f6 e1c1 c8e6 g1f3 f6g4 e3e1 f8e7 h2h3 g4f6 g2g4 e8g8 c1b1 f8e8 g4g5 f6d7 f4e3 e7f8 h3h4 e6g4 f1e2 e8e4 h4h5 g4e6 g5g6 h7h6 d5c3 e4g4 f3d4 g4d4 e3d4 c6d4 d1d4');
INSERT INTO evaluations VALUES(51077,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(51077,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(51077,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(51077,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(51077,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(51077,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(51077,36,'cp 142','c2c4 d7d6 b1c3 e7e5 g1f3 e5d4 f3d4 g8f6 f1e2 f8e7 c1e3 c7c6 e1g1 e8g8 d1c2 f8e8 a1d1 d8c7 e3f4 c6c5 d4b3 b8c6 h2h3 c6e5 f4h2 c8e6 c3d5 e6d5 c4d5 b7b5 f2f4 e5d7 g1h1 e7f8 e2f3 c5c4 b3d4 d7c5 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(51073,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(51073,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(51073,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(51073,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(51073,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(51073,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(51073,36,'cp 130','d7d6 b1c3 e7e5 g1f3 e5d4 f3d4 g8f6 c1e3 f8e7 f1e2 c7c5 d4c2 c8d7 e1g1 e8g8 a2a4 b8c6 f2f4 f8e8 e3f2 a8b8 e2f3 d8b6 c2e3 c6d4 e3d5 d4f3 d1f3 f6d5 c3d5 b6c6 a4a5 e7d8 f1d1 f7f5 e4f5 d7f5');
INSERT INTO evaluations VALUES(6320,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(6320,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(6320,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(6320,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(6320,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(6320,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(6320,36,'cp 36','a2a4 a7a6 e1g1 d7d6 c2c3 g8f6 d2d3 e8g8 h2h3 c6e7 a4a5 b6a7 c4b3 e7g6 c1e3 h7h6 e3a7 a8a7 f1e1 d8e7 b1d2 b7b5 a5b6 c7b6 d1c2 c8e6 b3e6 f7e6 d3d4 a6a5 d4e5 d6e5 c2b3 g6f4 b4a5 b6a5 d2c4 a7c7');
INSERT INTO evaluations VALUES(6330,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(6330,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(6330,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(6330,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(6330,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6330,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(6330,36,'cp 18','a7a6 e1g1 g8f6 d2d3 e8g8 c2c3 h7h6 f1e1 d7d6 h2h3 c8e6 b1d2 b6a7 c4b3 f8e8 b3e6 f7e6 d2b3 d8d7 c1d2 d7f7 a1c1 e8f8 d3d4 f6h5 g1h2 e5d4 c3d4 e6e5 d2e3 e5d4 b3d4 a7d4 f3d4 c6b4 d4f5');
INSERT INTO evaluations VALUES(4896,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4896,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4896,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(4896,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(4896,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(4896,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(4896,36,'cp 24','e1g1 d7d6 c2c3 g8f6 d2d3 h7h6 h2h3 e8g8 b1d2 c8e6 c4e6 f7e6 d2c4 b6a7 c1e3 a7e3 f2e3 c6e7 c4d2 d8d7 d1b3 f8f7 c3c4 a8f8 c4c5 e7g6 b4b5 a6b5 a4b5 d6c5 d2c4 f6h5 c4e5 g6e5 f3e5 f7f1 a1f1 f8f1 g1f1 d7e7 b3c4 h5f6 f1g1 f6d7 e5f3');
INSERT INTO evaluations VALUES(19935,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(19935,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(19935,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19935,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(19935,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(19935,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(19935,36,'cp -12','f1b5 c8d7 e5d6 e7e6 d2d4 f8d6 d4c5 d6c5 e1g1 g8f6 a2a3 a7a6 b5d3 e8g8 c1g5 h7h6 g5h4 d8c7 b1c3 f6h5 f1e1 a8e8 d1d2 c5e7 h4e7 c6e7 f3e5 d7c6 e5c6 e7c6 a1d1 e8d8 d2e3 h5f6 e3g3 c7g3 h2g3 d8d7 c3e4 f6e4 d3e4 f8d8 d1d7 d8d7');
INSERT INTO evaluations VALUES(19917,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(19917,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(19917,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19917,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(19917,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(19917,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(19917,36,'cp -32','c8d7 e5d6 e7e6 d2d4 f8d6 d4c5 d6c5 e1g1 d8c7 b1d2 g8f6 a2a3 a8d8 b5d3 a7a5 d1e2 e8g8 d2c4 e6e5 f3e5 c6d4 e2d1 b7b5 e5d7 f6d7 c4d2 d7e5 d2e4 c5b6 c1g5 f7f6 g5e3 a5a4 d1h5 e5d3 c2d3 c7d7 e4c3 g7g6 h5h3 d7h3 g2h3');
INSERT INTO evaluations VALUES(26044,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(26044,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(26044,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(26044,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(26044,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(26044,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(26044,36,'cp -12','e5d6 e7e6 d2d4 f8d6 d4c5 d6c5 e1g1 d8c7 b1d2 g8f6 a2a3 a8d8 b5d3 a7a5 c2c3 e8g8 b2b4 c5a7 d1c2 h7h6 c1b2 c6e5 f3e5 c7e5 c3c4 d7a4 b2e5 a4c2 d3c2 d8d2 c2d1 a7d4 e5d4 d2d4 b4a5 d4c4 d1f3 c4c7 a1c1 f8c8 c1c7 c8c7 f1e1');
INSERT INTO evaluations VALUES(13747,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(13747,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13747,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(13747,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13747,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13747,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(13747,36,'cp 10','f1b5 c5d4 f3d4 c8d7 d4f3 a7a6 b5c6 b7c6 e1g1 f8b4 f3e5 b4d6 e5d7 d8d7 c3e2 d7c7 h2h3 d6h2 g1h1 h2e5 d1c2 e8g8 c1d2 a8b8 d2c3 e5c3 e2c3 c6c5 a1c1 f8c8 f1d1 c7e5 d1e1 h7h6 h1g1 b8b4 a2a3 b4b6 c2d2 c5c4 c1d1 g8h8 f2f3');
INSERT INTO evaluations VALUES(13700,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(13700,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13700,36,'cp 0','c5d4 f3d4 c8d7 e1g1 f8d6 e3e4 d5e4 c1g5 e8g8 d4c6 d7c6 b5c6 b7c6 c3e4 d6h2 g1h2 d8d1 a1d1 f6e4 g5e7 f8e8 d1d7 a7a5 f2f3 e4f6 e7f6 g7f6 f1f2 e8d8 d7b7 a8b8 b7c7 b8c8 c7e7 d8e8 e7d7 c8d8 d7b7 d8b8 b7c7 b8c8 c7b7');
INSERT INTO evaluations VALUES(18920,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18920,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(18920,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(18920,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(18920,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(18920,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(18920,36,'cp 19','d4d1 d7d5 e4d5 e6d5 f1d3 f8c5 e1g1 g8f6 f1e1 c8e6 b1d2 e8g8 d2b3 c5b6 h2h3 f6e4 c1e3 e6f5 e3b6 d8b6 d1e2 f8e8 e2e3 b6c7 f3d4 f5g6 d4b5 c7b8 a1d1 a7a6 b5d4 b8c7 d3e4 d5e4 d4c6 c7c6 e3c5 c6c5 b3c5');
INSERT INTO evaluations VALUES(18936,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(18936,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18936,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(18936,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(18936,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(18936,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18936,36,'cp 10','g8f6 f1d3 g7g6 e1g1 f8g7 f1e1 d7d5 b1c3 e8g8 e4d5 f6d5 c3d5 e6d5 h2h3 d8b6 a2a3 c8d7 a1b1 a8e8 c1f4 e8e1 d1e1 f8e8 e1d2 a7a6 b2b3 c6d4 f3d4 b6d4 a3a4 d4f6 f4g5 f6b6 b1e1 e8e1 d2e1 b6c5 e1e3 c5a3 e3d2 d7c6 d2e3');
INSERT INTO evaluations VALUES(12506,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12506,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12506,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12506,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(12506,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(12506,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(12506,36,'cp 17','f1d3 d7d5 e4d5 e6d5 e1g1 f8c5 f1e1 c8e6 b1d2 e8g8 h2h3 c5b6 d2b3 f6e4 c1e3 e6f5 e3b6 d8b6 d1e2 f5g6 e2e3 b6c7 f3d4 a8e8 d4c6 b7c6 b3c5 e4c5 e3c5 g6d3 c2d3 c7b6 b2b4 b6c5 b4c5 g7g6 a1b1 d5d4 e1e8 f8e8 g1f1 e8e5 b1b8 g8g7');
INSERT INTO evaluations VALUES(12503,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12503,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(12503,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12503,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(12503,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12503,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(12503,36,'cp 17','d7d5 e4d5 e6d5 e1g1 c8e6 f1e1 f8c5 b1d2 e8g8 h2h3 c5b6 d2b3 f6e4 c1e3 e6f5 e3b6 d8b6 d1e2 a8e8 e2e3 b6c7 f3d4 e4f6 e3d2 c6d4 b3d4 f5d3 c2d3 c7b6 d4f5 g7g6 f5e7 g8g7 d2g5 b6b2 a1b1 b2c3 e7f5 g7h8 e1c1');
INSERT INTO evaluations VALUES(37382,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(37382,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(37382,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(37382,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(37382,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37382,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37382,36,'cp 13','d7d5 c4d5 f6d5 e1g1 c7c5 d2d4 b8c6 c3d5 d8d5 c1e3 c5d4 f3d4 d5c4 g2c6 b7c6 d1c1 c8a6 c1c4 a6c4 a1c1 g7d4 e3d4 c4e2 f1e1 e2f3 e1e7 f8e8 e7e3 f3d5 a2a4 a7a5 h2h4 e8e3 f2e3 d5b3 c1c6 b3a4 c6c7 a4b3 g1f2 a5a4 d4c3 b3d5 f2e2 a4a3 b2a3');
INSERT INTO evaluations VALUES(36718,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36718,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(36718,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36718,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36718,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36718,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36718,36,'cp 12','c4d5 f6d5 e1g1 c7c5 d2d4 b8c6 c3d5 d8d5 c1e3 c5d4 f3d4 d5c4 d4c6 b7c6 d1d2 c8f5 b2b3 c4a6 a1c1 f8d8 d2b4 a6a2 g2c6 a8c8 b4e7 a2b3 e7a7 f5h3 c6g2 c8c1 e3c1 h3g2 g1g2 b3d5 g2g1 h7h5 a7a4 d5d7 a4d7 d8d7 f1e1 d7e7');
INSERT INTO evaluations VALUES(36712,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36712,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36712,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(36712,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36712,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36712,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36712,36,'cp 21','f6d5 e1g1 c7c5 d2d4 d5c3 b2c3 c5d4 f3d4 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 c8f5 c6a8 g7c3 c1a3 c3a1 a3e7 d8a8 f1a1 a8e8 e7c5 a7a6 a1e1 f5g4 e2e3 e8c8 e1c1 c8d8 g1f1 d8d1 c1d1 g4d1 f1e1 d1f3 e1d2 f3e4 c5d4 f7f5 d2c3');
INSERT INTO evaluations VALUES(36818,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(36818,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36818,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36818,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36818,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36818,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36818,36,'cp 10','e1g1 c7c5 d2d4 d5c3 b2c3 c5d4 f3d4 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 f1d1 d8f8 f4e3 b2e2 c6f3 e2c2 d1c1 c2c3 c1c3 g7c3 a1c1 c3e5 e3a7 c8d7 g1g2 e5d6 h2h4 f8d8 c1a1 d7a4 a1b1 g8f8 a7e3 d8b8 b1b8 d6b8 g2f1 b8d6 f1e2 a4b5 e2d2 d6b4 d2c2 b5a4 c2c1');
INSERT INTO evaluations VALUES(36817,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36817,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36817,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36817,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(36817,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36817,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36817,36,'cp 12','c7c5 d2d4 d5c3 b2c3 c5d4 f3d4 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 f1d1 d8f8 a2a4 g7c3 c6b5 b2b5 a4b5 c3a1 d1a1 c8d7 a1a5 f7f6 f2f3 g6g5 f4d2 f8b8 a5a7 d7b5 a7e7 b8e8 e7e8 b5e8 e2e4 e8d7 d2b4 g5g4 f3g4 d7c6 b4c3 g8f7 e4e5 f6e5 c3e5');
INSERT INTO evaluations VALUES(19886,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(19886,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(19886,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19886,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19886,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19886,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(19886,36,'cp 0','c8f5 f1d3 f5d3 d1d3 e7e6 g1f3 f8b4 b1d2 g8e7 a2a3 b4d2 c1d2 d8b6 e1g1 h7h6 f1c1 e7f5 d2c3 e8g8 c1e1 b6d8 h2h3 f5h4 e1e3 h4f5');
INSERT INTO evaluations VALUES(26313,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26313,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26313,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(26313,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26313,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26313,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26313,36,'cp 6','f1d3 f5d3 d1d3 e7e6 g1f3 g8e7 b1c3 e7f5 a2a3 f8e7 c1d2 e8g8 e1g1 f7f6 e5f6 e7f6 c3e2 d8d7 a1e1 a8e8 d2c3 d7f7 e2g3 f5h4 f3h4 f6h4 f2f4 f7g6 d3b5 g6f7 e1e3 f7d7 b5d3 h4d8 g1h1 d8c7 g3h5 f8f7 g2g4 e8f8 e3h3 c7d6');
INSERT INTO evaluations VALUES(39715,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(39715,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(39715,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(39715,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(39715,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(39715,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(39715,36,'cp 43','d1f3 d8c7 e1c1 b8d7 g2g4 b7b5 g5f6 d7f6 g4g5 f6d7 f4f5 e8g8 h1g1 b5b4 c3e2 e6e5 f5f6 e5d4 f6e7 f8e8 e2d4 d7e5 f3f4 e8e7 d4f5 e7e8 g1g2 d6d5 g2f2 d5e4 f5d6 c8g4 d6e8 a8e8 d1e1 g4f3 c1b1 c7c5 f2d2 a6a5 f4e3 c5e3 e1e3 g8f8 f1e2 h7h6 g5h6 g7h6 e2f3 e5f3');
INSERT INTO evaluations VALUES(39714,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(39714,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(39714,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(39714,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(39714,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(39714,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(39714,36,'cp 43','h7h6 g5h4 e6e5 d4f5 c8f5 e4f5 b8c6 e1c1 a8c8 g2g4 e5f4 f3f4 e8g8 h1g1 f6d5 h4e7 d5f4 e7d8 f8d8 g1g3 d6d5 g3f3 d5d4 f3f4 d4c3 b2c3 d8d1 c1d1 c6e5 d1d2 g8f8 h2h3 f7f6 f4b4 b7b5 a2a4 b5a4 b4a4 e5f3 d2c1 c8c3 c1b2 c3e3 a4a6');
INSERT INTO evaluations VALUES(39659,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(39659,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(39659,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(39659,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(39659,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(39659,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(39659,36,'cp 59','g5h4 e6e5 d4f5 c8f5 e4f5 b8c6 e1c1 e8g8 g2g4 e5f4 h1g1 c6e5 f3f4 f6d5 c3d5 e7h4 c1b1 h4g5 f4f2 g5h4 f2g2 h4g5 f1d3 a8c8 g1f1 e5d7 c2c3 f8e8 d3c2 b7b5 a2a3 a6a5 c2b3 d7c5 b3a2 g5f6 g2f2 c5e4 f2f3 f6g5 f5f6 e4f6 b1a1 e8e5 d5f6 g5f6');
INSERT INTO evaluations VALUES(39660,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(39660,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(39660,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(39660,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(39660,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(39660,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(39660,36,'cp 47','e6e5 d4f5 c8f5 e4f5 b8c6 e1c1 e8g8 g2g4 e5f4 h1g1 c6e5 f3f4 f6d5 h4e7 d5f4 e7d8 a8d8 c3d5 f4d5 d1d5 f8e8 h2h3 g8f8 f1g2 d8d7 d5d4 e8c8 a2a4 f8e7 g1d1 c8c5 c2c3 b7b5 a4b5 a6b5 g2d5 d7d8 c1b1 e7d7 d4e4 d7c7 d1d4 d8d7 e4e2 e5c6 d4d2');
INSERT INTO evaluations VALUES(50190,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50190,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(50190,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(50190,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(50190,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(50190,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(50190,36,'cp 15','c7c6 g1f3 d7d5 f1g2 f8g7 e1g1 e8g8 c4d5 c6d5 b1c3 f6e4 f3d2 e4c3 b2c3 c8e6 e2e4 b8c6 e4e5 b7b6 d2b3 d8d7 f1e1 e6h3 g2h3 d7h3 a2a4 f8c8 a4a5 h7h5 e1e3 h3d7 e3f3 e7e6 h2h4 c6a5 b3a5 b6a5 a1a5');
INSERT INTO evaluations VALUES(26653,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(26653,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26653,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26653,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(26653,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(26653,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(26653,36,'cp 8','e7e6 c2c3 f7f6 b1d2 f6e5 f3e5 c6e5 d4e5 c7c6 d2b3 d8c7 c1f4 g8e7 g2g4 f5e4 f2f3 e4g6 d1d4 h7h5 h2h4 h5g4 f3g4 g6e4 h1h2 e7g6 f4g3 e8c8 b3d2 c6c5 d4a4 g6e5 d2e4 d5e4 a4e4 f8d6 e1c1 e5d3 d1d3 d6g3 h2h1 h8h4 e4e6 c8b8 d3d8 c7d8 h1h4 g3h4 e6e5 d8c7 e5e8 c7d8');
INSERT INTO evaluations VALUES(26852,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26852,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(26852,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(26852,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(26852,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(26852,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(26852,36,'cp 12','f1b5 a7a6 b5c6 b7c6 c1e3 f5g6 e1g1 a6a5 h2h3 g8e7 f3h4 e7f5 h4f5 g6f5 b1c3 a5a4 b2b3 a4b3 a2b3 f8b4 c3e2 e8g8 c2c3 b4e7 d1d2 d8b8 b3b4 b8b5 e2g3 f5g6 e3g5 e7g5 d2g5 b5d3 g5c1 f8b8 a1a8 b8a8 c1b2 d3a6 h3h4 g6d3 f1e1');
INSERT INTO evaluations VALUES(26793,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(26793,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26793,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(26793,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(26793,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26793,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(26793,36,'cp 0','a7a6 b5c6 b7c6 c1e3 a6a5 e1g1 f5g6 d1d2 g8e7 b1c3 e7f5 c3a4 f5e3 d2e3 g6c2 b2b3 c2g6 f1c1 f8a3 c1c6 d8d7 e3c3 a3b4 c3c1 b4a3 c1c3');
INSERT INTO evaluations VALUES(51769,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51769,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51769,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51769,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(51769,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(51769,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51769,36,'cp 33','d7d6 g1f3 d6e5 f3e5 b8d7 b1c3 d7e5 d4e5 d8d1 c3d1 c8d7 h2h4 g8e7 h4h5 e7d5 h5h6 e8c8 c1g5 f7f6 h6g7 f8g7 e5f6 g7f6 g5f6 d5f6 h1h6 f6g4 h6h4 h7h5 f1d3 d7e8 a2a4 c6c5 g2g3 g4e5 d3e2 e8c6 h4h5 h8h5 e2h5');
INSERT INTO evaluations VALUES(53847,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(53847,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(53847,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(53847,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(53847,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(53847,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(53847,36,'cp 28','g1f3 d6e5 f3e5 b8d7 b1c3 d7e5 d4e5 d8d1 c3d1 g8e7 h2h4 h7h5 f1d3 b7b6 a2a3 e7f5 d1c3 f8e7 c1g5 c8b7 d3f5 e6f5 e1c1 a8d8 g5e7 e8e7 c3e2 e7e6 e2f4 e6e5 d1d8 h8d8 f4h5 d8g8 h1e1 e5d6 h5f4 c6c5 e1d1 d6c7 f2f3 g8d8 d1d8 c7d8 c1d2 d8e7 d2e3 e7d6');
INSERT INTO evaluations VALUES(53846,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(53846,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(53846,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(53846,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(53846,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53846,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(53846,36,'cp 42','d6e5 f3e5 b8d7 b1c3 d7e5 d4e5 d8d1 c3d1 g8e7 h2h4 e7g6 f2f4 h7h5 g2g3 g6e7 d1e3 b7b6 f1e2 a7a5 e1f2 c8a6 e2a6 a8a6 c1d2 g7g6 a2a4 a6a8 h1d1 a8d8 e3c4 e7f5 d2e3 f5e3 d1d8 e8d8 f2e3 f8c5 e3f3 d8e7 a1d1 h8d8 f3e2 d8d7 d1d7 e7d7 e2f3 c5d4');
INSERT INTO evaluations VALUES(15535,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(15535,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(15535,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(15535,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(15535,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(15535,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(15535,36,'cp 65','g1f3 f8e7 d2d4 a7a5 e1g1 e8g8 b2b3 f8e8 e2e4 c8d7 a1b1 c6d4 f3d4 e5d4 d1d4 a5a4 h2h3 a4b3 a2b3 e7f8 f1e1 c7c6 c1e3 d7e6 e1d1 h7h6 d4d3 f6d7 d3c2 d8a5 e3d4 a5h5 c3e2 h5a5 d4b2 a5g5 e2f4 g5e7 b2c3 a8a3 f4e6');
INSERT INTO evaluations VALUES(15534,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(15534,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(15534,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(15534,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(15534,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(15534,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(15534,36,'cp 59','f8e7 d2d4 a7a5 e1g1 e8g8 b2b3 c8f5 d4d5 c6b4 f3h4 f5d7 e2e4 d8c8 a2a3 b4a6 d1c2 a6c5 c1b2 f8e8 b3b4 c5a6 c2b3 e7f8 f1d1 h7h6 d1e1 c8d8 h4f3 c7c6 f3d2 c6d5 e4d5 d8b6 d2e4 f6e4 e1e4 a5b4 a3b4 b6b4');
INSERT INTO evaluations VALUES(8624,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(8624,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(8624,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(8624,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(8624,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(8624,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(8624,36,'cp 61','d2d4 e8g8 e1g1 a7a5 b2b3 c8g4 d4d5 c6b8 f3e1 b8a6 a2a3 f6d7 f2f3 g4h5 e2e4 h5g6 g2h3 e7g5 e1c2 f7f5 e4f5 g6f5 h3f5 f8f5 b3b4 f5f8 d1d3 a5b4 c1g5 d8g5 a3b4 d7f6 c3e4 f6e4 d3e4 g5h5 g1g2 h7h6 f1e1 h5f7 e4d3 a8e8');
INSERT INTO evaluations VALUES(22540,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(22540,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(22540,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(22540,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(22540,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(22540,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(22540,36,'cp 20','g8f6 e4e5 d7d5 c4b5 f6e4 f3d4 f8c5 b1d2 c5d4 c3d4 e8g8 b5c6 b7c6 e1g1 c6c5 d2e4 d5e4 d4c5 a7a5 f1e1 c8f5 c1f4 d8e7 e5e6 f5e6 d1d4 f8d8 d4e4 e7c5 f4e5 c5c4 e4c4 e6c4 e5c7 d8d2 c7e5 a8e8 h2h3 d2c2 a2a3 f7f6 e5c3 e8e1 a1e1 g8f7 e1e4 c4b3 e4d4 a5a4 d4d8 c2e2');
INSERT INTO evaluations VALUES(16537,30,'cp -189',NULL);
INSERT INTO evaluations VALUES(16537,31,'cp -187',NULL);
INSERT INTO evaluations VALUES(16537,32,'cp -207',NULL);
INSERT INTO evaluations VALUES(16537,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(16537,34,'cp -169',NULL);
INSERT INTO evaluations VALUES(16537,35,'cp -196',NULL);
INSERT INTO evaluations VALUES(16537,36,'cp -199','d7d5 d1e2 f8d6 e1g1 f6e4 d3e4 d5e4 e2e4 e8g8 d2d3 f7f5 e4e2 d8f6 b1a3 h7h6 a3c4 c8e6 d3d4 e5e4 f3d2 d6f4 g2g3 f4d2 c4d2 f8e8 f2f4 e4f3 d2f3 f6f7 f1f2 e6a2 c1e3 a2c4 e2d2 a8d8 f3h4 d8d5 h4f3 g7g5 a1e1 e8e4');
INSERT INTO evaluations VALUES(15145,30,'cp -176',NULL);
INSERT INTO evaluations VALUES(15145,31,'cp -188',NULL);
INSERT INTO evaluations VALUES(15145,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(15145,33,'cp -183',NULL);
INSERT INTO evaluations VALUES(15145,34,'cp -179',NULL);
INSERT INTO evaluations VALUES(15145,35,'cp -200',NULL);
INSERT INTO evaluations VALUES(15145,36,'cp -172','d1e2 f8d6 e1g1 d5e4 d3e4 f6e4 e2e4 e8g8 d2d3 f7f5 e4e2 d8f6 b2b4 a7a6 b1d2 c8e6 a2a4 e6d5 c1a3 f8e8 b4b5 e5e4 d3e4 f5e4 b5c6 e4f3 e2d3 d5c6 d2f3 d6a3 a1a3 a8d8 d3c4 g8h8 f3d4 f6f4 a3a1 h7h6 a1e1 e8e1 f1e1 d8e8 e1e8 c6e8');
INSERT INTO evaluations VALUES(21498,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(21498,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(21498,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(21498,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(21498,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(21498,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(21498,36,'cp 188','e2e4 b7b5 c3b5 f8b4 b5c3 g8f6 f1c4 f6e4 e1g1 e4d6 d1a4 b4c3 c4e6 c8d7 e6d7 d8d7 b2c3 e8g8 c1a3 f8e8 c3c4 c6e5 a4d7 e5d7 f1c1 a8c8 g1f1 c7c5 c1c2 d6b7 d4c5 d7c5 f3d4 h7h6 d4f5 c8d8 a1e1 e8e1 f1e1 c5d3 e1f1 d3f4 g2g3 d8d3 a3b2 b7d6');
INSERT INTO evaluations VALUES(21503,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(21503,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(21503,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(21503,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(21503,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(21503,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(21503,36,'cp 153','b7b5 c3b5 f8b4 b5c3 g8f6 f1c4 f6e4 e1g1 e4d6 d1a4 b4c3 c4e6 c8d7 e6d7 d8d7 b2c3 e8g8 c1f4 c6e5 a4d7 e5d7 f1c1 d7b6 a1b1 a8c8 f3d2 f8e8 g2g3 h7h6 g1f1 g7g5 f4d6 c7d6 c3c4 c8c7 b1b5 g8g7 b5a5 e8c8 c4c5 f7f5');
INSERT INTO evaluations VALUES(16818,30,'cp 317',NULL);
INSERT INTO evaluations VALUES(16818,31,'cp 286',NULL);
INSERT INTO evaluations VALUES(16818,32,'cp 319',NULL);
INSERT INTO evaluations VALUES(16818,33,'cp 308',NULL);
INSERT INTO evaluations VALUES(16818,34,'cp 305',NULL);
INSERT INTO evaluations VALUES(16818,35,'cp 328',NULL);
INSERT INTO evaluations VALUES(16818,36,'cp 303','c4f7 e8e7 d2d4 h7h6 g5e4 e7f7 d4d5 c6e7 d5d6 c7d6 e4d6 f7g8 b1c3 e7c6 d1d5 g8h7 d6f7 d8h4 f7h8 c6b4 d5b3 h7h8 a2a3 b4c6 c1e3 h4h5 e1g1 b7b6 b3d5 c8a6 d5d7 h5g6 f1d1 a8d8 d7h3 c6d4 d1d2 a6b7 h3g3 g6f6 a1d1 f8c5 e3d4 e5d4 c3e2 d4d3 e2f4 d8d4 f4d3 c5d6 f2f4 b7a6 g3f3');
INSERT INTO evaluations VALUES(16038,30,'cp 304',NULL);
INSERT INTO evaluations VALUES(16038,31,'cp 318',NULL);
INSERT INTO evaluations VALUES(16038,32,'cp 295',NULL);
INSERT INTO evaluations VALUES(16038,33,'cp 319',NULL);
INSERT INTO evaluations VALUES(16038,34,'cp 302',NULL);
INSERT INTO evaluations VALUES(16038,35,'cp 305',NULL);
INSERT INTO evaluations VALUES(16038,36,'cp 325','e8e7 d2d4 h7h6 g5e4 e7f7 d4d5 c6e7 d5d6 c7d6 e4d6 f7g8 b1c3 e7c6 d1d5 g8h7 d6f7 d8h4 f7h8 h7h8 c1e3 d7d6 d5f7 h4e7 f7e7 f8e7 e1c1 c8e6 f2f4 e6g4 d1d2 a8f8 h1f1 g4e6 d2d1 e6g4 d1e1 f8c8 h2h3 g4h5 g2g4 h5g6 c1b1 c6b4 f4f5 g6f7 e3a7 f7c4 f1f3 e7h4 e1d1');
INSERT INTO evaluations VALUES(2476,30,'cp 308',NULL);
INSERT INTO evaluations VALUES(2476,31,'cp 325',NULL);
INSERT INTO evaluations VALUES(2476,32,'cp 333',NULL);
INSERT INTO evaluations VALUES(2476,33,'cp 319',NULL);
INSERT INTO evaluations VALUES(2476,34,'cp 321',NULL);
INSERT INTO evaluations VALUES(2476,35,'cp 329',NULL);
INSERT INTO evaluations VALUES(2476,36,'cp 305','d2d4 h7h6 g5e4 e7f7 d4d5 c6e7 d5d6 c7d6 e4d6 f7g8 b1c3 e7c6 d1d5 g8h7 d6f7 d8h4 f7h8 c6b4 d5b3 h7h8 a2a3 b4c6 c1e3 d7d6 e1g1 b7b6 f2f4 c8a6 f1f2 a8d8 a1d1 c6a5 b3a4 f8e7 f4e5 h4a4 c3a4 a5c4 e3f4 d6d5 f4g3 c4e3 d1e1 e7g5 e5e6 d5d4 f2f7 e3c2 e6e7');
INSERT INTO evaluations VALUES(31766,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(31766,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(31766,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(31766,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(31766,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(31766,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(31766,36,'cp 90','d2d4 e7e6 b1c3 c7c6 h2h3 g4f3 d1f3 d6d5 f1d3 d5e4 c3e4 g8f6 c2c3 b8d7 g2g3 f8e7 d3c2 f6e4 f3e4 d7f6 e4d3 h7h6 d3e2 e8g8 a2a4 d8d5 e1g1 a7a5 f1e1 d5d8 h3h4 e7d6 e2f3 d6c7 c1e3 d8d5 f3e2 a8d8');
INSERT INTO evaluations VALUES(31764,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(31764,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(31764,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(31764,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(31764,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(31764,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(31764,36,'cp 89','e7e6 h2h3 g4h5 b1c3 b8d7 g2g4 h5g6 d1e2 c7c6 c1d2 h7h6 h3h4 h6h5 g4g5 g8e7 f1h3 d6d5 e1g1 d5e4 c3e4 e7f5 d2f4 f8e7 f1e1 e8g8 c2c3 d7b6 a1d1 b6d5 f4g3 f5g3 f2g3 e7d6 e4d6 d8d6 f3e5 a8d8 e2f3 a7a5 c3c4 d5b4 a2a3 b4c2 e1e2');
INSERT INTO evaluations VALUES(38395,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(38395,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(38395,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38395,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(38395,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(38395,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(38395,36,'cp 50','b1c3 g8f6 d1d2 e8g8 f4h6 c7c5 h6g7 g8g7 e1c1 c5d4 f3d4 b8c6 h2h3 c8d7 g2g4 c6d4 d2d4 d8b6 g4g5 b6d4 d1d4 f6h5 f1g2 a8b8 g2f3 d7c6 h1e1 h7h6 f3h5 g6h5 g5h6 g7h6 e1d1 f7f6 f2f3 b8d8 a2a4 d8c8 c1d2 a7a6 d4f4');
INSERT INTO evaluations VALUES(38387,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(38387,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38387,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(38387,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(38387,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38387,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38387,36,'cp 21','g8f6 e2e4 e8g8 h2h3 c7c5 d4c5 d8a5 f1d3 a5c5 e1g1 b7b6 f4e3 c5c7 c3b5 c7d8 c2c4 c8b7 b5c3 b8d7 d1c2 d7c5 f3d4 e7e6 f1d1 a8c8 a1c1 a7a6 d4b3 d8e7 b3c5 d6c5 d1d2 e7c7 b2b3 c8d8 c1d1 d8d7 d3e2 f8d8 d2d7 d8d7');
INSERT INTO evaluations VALUES(45916,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(45916,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(45916,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(45916,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(45916,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(45916,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(45916,36,'cp 183','f1b5 b8d7 d4f5 a7a6 b5d7 d8d7 d1e2 d7e6 c1g5 g7g6 f5e3 f8g7 e3c4 e8g8 e1c1 e6g4 e2g4 f6g4 d1d2 c8e6 c4d6 g7c3 b2c3 f7f6 g5f4 g4e5 a2a3 e5c4 d6c4 e6c4 h1d1 f8e8 f2f3 f6f5 e4f5 g6f5 d2d7 b7b5 g2g3 a8c8 d1d6 e8e1 c1b2 e1e6 d6e6 c4e6 d7d6');
INSERT INTO evaluations VALUES(45892,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(45892,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(45892,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(45892,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(45892,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(45892,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(45892,36,'cp 186','b8d7 d4f5 a7a6 b5d7 d8d7 d1e2 d7e6 c1g5 g7g6 f5e3 f8g7 e3c4 e8g8 e1c1 e6g4 e2g4 f6g4 d1d2 c8e6 c4d6 g7c3 b2c3 f7f6 g5f4 g4e5 a2a3 e5c4 d6c4 e6c4 h1d1 f8e8 f2f3 c4b5 c3c4 b5c6 h2h4 g8f7 c1b2 f6f5 e4e5 e8e6 b2c3 a8e8 d1b1 e8e7 d2d6 e6d6 e5d6 e7e8 d6d7 c6d7 b1b7');
INSERT INTO evaluations VALUES(19774,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(19774,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19774,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19774,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19774,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19774,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19774,36,'cp 0','g8f6 b1d2 c8f5 g1f3 d8b6 d4c5 b6b2 f3d4 e7e5 f4e5 f5g4 a1b1 b2a3 f2f3 c6e5 f3g4 f8c5 d4b5 a3a2 b1a1 a2b2 a1b1 b2a2');
INSERT INTO evaluations VALUES(44236,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44236,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44236,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(44236,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(44236,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44236,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44236,36,'cp 22','c2c4 d7d5 c4d5 e6d5 f1b5 b8d7 d4c5 a7a6 b5d7 c8d7 b1c3 f8c5 e1g1 a8c8 b2b3 e8g8 c1b2 d7c6 d1d3 d5d4 f3d4 c5d4 d3d4 d8d4 e3d4 f6d5 c3d5 c6d5 a1c1 f8d8 f2f3 f7f6 g1f2 h7h5 f1d1 g7g5 b2c3 d5c6');
INSERT INTO evaluations VALUES(44210,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(44210,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44210,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(44210,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(44210,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44210,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(44210,36,'cp 27','d7d5 c4d5 e6d5 f1b5 b8d7 d4c5 a7a6 b5e2 d7c5 b1c3 c8e6 e1g1 f8d6 b2b4 c5e4 c1b2 e8g8 b4b5 e4c3 b2c3 a6a5 a2a4 a8c8 d1e1 b7b6 c3d4 f6d7 e1d2 d7c5 d2b2 f7f6 d4c5 c8c5 f3d4 e6f7');
INSERT INTO evaluations VALUES(50516,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(50516,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(50516,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(50516,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50516,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50516,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50516,36,'cp 45','d7d5 e4e5 c8d7 f1d3 b8c6 a2a3 g8h6 e1g1 h6f5 d4c5 f8c5 b1d2 e8g8 d2b3 c5a7 c1f4 h7h6 f1e1 f5h4 f3h4 d8h4 f4g3 h4d8 d3c2 c6e7 b3d4 a7d4 d1d4 e7f5 d4d2 a8c8 g3f4 d8b6 a1b1 f8e8 a3a4 e8d8 g2g4 f5h4 d2d3 h4g6');
INSERT INTO evaluations VALUES(50784,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(50784,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(50784,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(50784,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(50784,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(50784,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(50784,36,'cp 41','e4e5 c8d7 f1d3 b8c6 a2a3 d8b6 d3c2 a8c8 d4c5 f8c5 e1g1 b6c7 c1f4 g8e7 f4g3 c5a7 b1d2 c6a5 f1e1 h7h6 a1b1 d7b5 b1c1 e8g8 c2b1 a5c4 d2c4 d5c4 f3d4 a7d4 d1d4 c8d8 d4g4 e7g6 b1g6 f7g6 g4g6 c7f7 g6f7 g8f7');
INSERT INTO evaluations VALUES(50788,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(50788,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(50788,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(50788,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(50788,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(50788,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(50788,36,'cp 36','c8d7 f1d3 b8c6 a2a3 d8b6 d3c2 a8c8 d4c5 f8c5 e1g1 b6c7 c1f4 g8e7 f4g3 c5a7 b1d2 c6a5 f1e1 h7h6 a1b1 d7b5 g3f4 e8g8 a3a4 e7g6 f4g3 b5e8 f3h4 a5c4 d2f3 b7b5 h4g6 f7g6 a4b5 a6b5 f3d4 a7d4 d1d4');
INSERT INTO evaluations VALUES(49639,30,'cp -127',NULL);
INSERT INTO evaluations VALUES(49639,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(49639,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(49639,33,'cp -134',NULL);
INSERT INTO evaluations VALUES(49639,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(49639,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(49639,36,'cp -94','d7d5 c1g5 b8c6 e4d5 c6d4 f3d1 d8d5 c2c3 d4c6 b1d2 c8e6 g1f3 e8c8 f1e2 h7h6 g5f6 g7f6 e1g1 f6f5 d1c2 h6h5 b2b4 h5h4 a2a4 c6e7 c3c4 d5d6 c4c5 d6d5 d3d4 e7c6 e2c4 d5d7 b4b5 c6d4 f3d4 d7d4 c4e6 f7e6 d2b3 d4d3');
INSERT INTO evaluations VALUES(48836,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(48836,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(48836,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(48836,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(48836,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(48836,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(48836,36,'cp -103','c1g5 b8c6 b1c3 c6d4 f3d1 c7c6 e4d5 d8b6 a1b1 f6d5 c3d5 c6d5 g5e3 b6a5 c2c3 d4c6 g1f3 f8d6 f1e2 a5c7 e1g1 e8g8 h2h3 c8e6 b2b4 h7h6 e3c5 a8d8 c5d6 c7d6 f1e1 e5e4 f3d4 c6d4 c3d4 e6f5 b4b5 f8e8 d1d2 d6f6 d3e4 f5e4 b1d1');
INSERT INTO evaluations VALUES(54480,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54480,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54480,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54480,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54480,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(54480,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(54480,36,'cp 17','f8c5 b1c3 g8f6 c1g5 e8g8 g5f6 d8f6 d1d5 c5b4 d5d4 f6e6 d4e3 e6e3 f2e3 b4c3 b2c3 b8d7 e1c1 d7b6 d1d4 c8d7 c3c4 a8c8 f3g5 d7c6 f1d3 g7g6 g5e4 g8g7 e4d6 c8c7 d3e4 c7e7 e4c6 b7c6 h1e1 e7e5 d4f4 h7h5 c1d2 c6c5 e3e4');
INSERT INTO evaluations VALUES(16310,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(16310,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(16310,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(16310,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(16310,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(16310,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(16310,36,'cp -31','f6e4 e1g1 f8e7 f1e1 e4d6 b5c6 d7c6 f3e5 a7a5 d2d4 e8g8 a2a4 f8e8 c1f4 e7g5 f4g3 c8e6 b1d2 f7f6 e5d3 e6f7 d1c2 g5h6 d2f1 d8d7 d3c5 d7c8 e1e8 c8e8 f2f3 b7b6 c5e4 e8d7 e4d6 c7d6 a1e1 b6b5 f1e3 g7g6 e3g4 h6g7 h2h4 h7h5 g4e3 a8e8 g3f2');
INSERT INTO evaluations VALUES(16755,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(16755,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(16755,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(16755,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(16755,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(16755,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(16755,36,'cp -23','e1g1 f8e7 f1e1 e4d6 b5c6 d7c6 f3e5 c8f5 d2d3 e8g8 c1f4 f7f6 e5f3 d8d7 b1d2 f5g6 d2e4 f8e8 e4c5 d7c8 c5e6 b7b6 f3d4 c6c5 d4c6 e7f8 e6f8 g8f8 d1f3 c8d7 c6e7 d6f5 e7g6 h7g6 g2g3 a8d8 e1e8 d8e8 h2h4 f8g8');
INSERT INTO evaluations VALUES(44091,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(44091,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(44091,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(44091,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(44091,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(44091,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(44091,36,'cp 76','g1f3 d8a5 d2d3 d7d6 a2a3 f8e7 a1b1 e8g8 f1e2 a5d8 e1g1 c6c5 b2b4 b8c6 f3e1 b7b6 f2f4 e5f4 c1f4 f6d7 c3d5 d7e5 d5e7 d8e7 e1c2 c8e6 b4b5 c6d4 c2d4 c5d4 f4g3 f7f5 e4f5 e6f5 g3f2 e7g5 f2d4 g5g6 d1d2 f5d3');
INSERT INTO evaluations VALUES(44089,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(44089,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(44089,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(44089,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(44089,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(44089,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(44089,36,'cp 61','d8a5 d2d3 d7d6 f1e2 c6c5 e1g1 b8c6 f3e1 a7a6 f2f4 f8e7 a2a4 a5d8 g1h1 c8d7 e1f3 e8g8 d1e1 c6d4 f4e5 d6e5 e2d1 d4f3 f1f3 d7g4 f3g3 d8d7 d1g4 f6g4 h2h3 g4f6 c1d2 f6h5 g3f3 f7f5 e4f5 f8f5');
INSERT INTO evaluations VALUES(5945,30,'cp 266',NULL);
INSERT INTO evaluations VALUES(5945,31,'cp 257',NULL);
INSERT INTO evaluations VALUES(5945,32,'cp 257',NULL);
INSERT INTO evaluations VALUES(5945,33,'cp 283',NULL);
INSERT INTO evaluations VALUES(5945,34,'cp 283',NULL);
INSERT INTO evaluations VALUES(5945,35,'cp 281',NULL);
INSERT INTO evaluations VALUES(5945,36,'cp 262','d2d4 e5d4 c3d4 c5b6 d4d5 a7a6 b5a4 b6a5 b1d2 b7b5 d5c6 b5a4 d1a4 a5b6 d2c4 b6a7 e1g1 g8e7 c1e3 c8g4 e3a7 g4f3 a7d4 f3e2 f1e1 e2c4 a4c4 e8g8 b2b4 e7g6 c4c3 g6f4 d4e3 f4g6 a1d1 d8e7 a2a4 a8b8 e3a7 b8a8 a7d4 f7f6 b4b5 a6b5 a4b5 e7e6 b5b6 c7b6 c6c7 f8c8 d4b6 g6e5 c3g3');
INSERT INTO evaluations VALUES(5936,30,'cp 255',NULL);
INSERT INTO evaluations VALUES(5936,31,'cp 246',NULL);
INSERT INTO evaluations VALUES(5936,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(5936,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(5936,34,'cp 286',NULL);
INSERT INTO evaluations VALUES(5936,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(5936,36,'cp 265','e5d4 c3d4 c5b6 d4d5 a7a6 b5a4 b6a5 b1d2 b7b5 d5c6 b5a4 d1a4 a5b6 d2c4 b6a7 e1g1 g8e7 c1e3 e8g8 e3a7 a8a7 c4e3 a7a8 a1d1 a8b8 e4e5 d6d5 f3d4 f7f6 f2f4 f6e5 f4e5 e7g6 f1f8 g6f8 a4a5 c8e6 b2b3 d8g5 e3d5 b8d8 d4f3 d8d5 d1d5 g5c1 g1f2');
INSERT INTO evaluations VALUES(5957,30,'cp 285',NULL);
INSERT INTO evaluations VALUES(5957,31,'cp 273',NULL);
INSERT INTO evaluations VALUES(5957,32,'cp 294',NULL);
INSERT INTO evaluations VALUES(5957,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(5957,34,'cp 291',NULL);
INSERT INTO evaluations VALUES(5957,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(5957,36,'cp 279','c3d4 c5b4 e1f1 a7a6 b5c6 b7c6 d1a4 a6a5 a2a3 c8b7 a4b3 c6c5 a3b4 c5b4 h2h4 h7h6 b1d2 g8f6 e4e5 f6d5 d2c4 d8d7 h1h3 a5a4 b3c2 e8g8 h3g3 g8h8 f1g1 f7f5 c1d2 f5f4 g3g6 d5e7 e5d6 e7g6 c2g6 c7d6');
INSERT INTO evaluations VALUES(5955,30,'cp 272',NULL);
INSERT INTO evaluations VALUES(5955,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(5955,32,'cp 278',NULL);
INSERT INTO evaluations VALUES(5955,33,'cp 281',NULL);
INSERT INTO evaluations VALUES(5955,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(5955,35,'cp 285',NULL);
INSERT INTO evaluations VALUES(5955,36,'cp 297','c5b6 d4d5 a7a6 b5a4 b6a5 b1d2 b7b5 d5c6 b5a4 d1a4 a5b6 d2c4 b6c5 e1g1 a8b8 f1d1 b8b4 a4c2 c8e6 c4a5 g8f6 a5b7 d8c8 b7c5 b4c4 c2d3 c4c5 f3d4 e8g8 f2f3 f8e8 c1f4 h7h6 b2b3 f6h5 f4e3 c5e5 a1c1 e5c5');
INSERT INTO evaluations VALUES(5766,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(5766,31,'cp 290',NULL);
INSERT INTO evaluations VALUES(5766,32,'cp 297',NULL);
INSERT INTO evaluations VALUES(5766,33,'cp 289',NULL);
INSERT INTO evaluations VALUES(5766,34,'cp 313',NULL);
INSERT INTO evaluations VALUES(5766,35,'cp 306',NULL);
INSERT INTO evaluations VALUES(5766,36,'cp 310','d4d5 a7a6 b5a4 b6a5 b1d2 b7b5 d5c6 b5a4 d1a4 a5b6 d2c4 b6c5 e1g1 c8e6 c4a5 g8f6 a5b7 d8e7 b7c5 d6c5 f1e1 h7h6 c1e3 f6g4 e3d2 e8g8 h2h3 g4f6 d2e3 a8b8 a4c2 c5c4 f3d4 f8e8 a1d1 e7b4 d1d2 b4a5 e3f4 e6d5 f2f3 a5c5 f4e3 d5c6 d4c6 c5c6 e1c1 b8b4');
INSERT INTO evaluations VALUES(5764,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(5764,31,'cp 307',NULL);
INSERT INTO evaluations VALUES(5764,32,'cp 300',NULL);
INSERT INTO evaluations VALUES(5764,33,'cp 307',NULL);
INSERT INTO evaluations VALUES(5764,34,'cp 308',NULL);
INSERT INTO evaluations VALUES(5764,35,'cp 307',NULL);
INSERT INTO evaluations VALUES(5764,36,'cp 315','a7a6 b5a4 b6a5 b1d2 b7b5 d5c6 b5a4 d1a4 a5b6 d2c4 b6a7 e1g1 g8e7 c1e3 e8g8 e3a7 a8a7 c4e3 d8e8 f3d4 e7g6 a4c2 e8e5 a1d1 f8e8 f2f3 c8e6 g2g3 e5h5 f1f2 g6e5 d4e6 f7e6 f3f4 e5g6 c2c3 a7a8 a2a3');
INSERT INTO evaluations VALUES(27226,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(27226,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(27226,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(27226,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(27226,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(27226,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(27226,36,'cp 40','d4c4 f8e7 b1c3 g8f6 c1f4 e8g8 e1c1 d8c8 h1e1 f8e8 b5c6 d7c6 f3g5 e8f8 h2h3 a7a6 g2g4 f6d7 g5f3 d7c5 f3d4 f8e8 d4f5 e7f8 c3d5 c6d5 e4d5 c8d7 b2b3 e8e1 d1e1 a8e8 e1e3 e8e3 f4e3 d7e8 c4d4 e8e4 d4e4 c5e4 c2c4 g7g6 f5d4 e4c3 e3d2 c3e4 f2f3 e4d2');
INSERT INTO evaluations VALUES(57360,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(57360,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(57360,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(57360,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(57360,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(57360,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(57360,36,'cp 58','c2c4 f8e7 g2g3 g8f6 f1g2 d7d5 d1c2 c7c6 b1c3 e8g8 e1g1 b8d7 f3g5 d7b8 f1d1 a7a5 a2a3 h7h6 g5f3 b8d7 c4d5 c6d5 c1f4 g7g5 f4e5 f6e4 a1c1 g5g4 f3e1 d7e5 d4e5 e4c3 c2c3 c8d7 e1d3 a8c8 c3b3 c8c1');
INSERT INTO evaluations VALUES(57347,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(57347,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(57347,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(57347,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(57347,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(57347,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(57347,36,'cp 58','g8f6 g2g3 f8e7 f1g2 d7d5 d1c2 c7c6 e1g1 e8g8 b2b3 b7b6 b1c3 c8b7 e2e3 e7d6 c3e2 b8d7 e2f4 f8e8 c1b2 a8c8 f1d1 d8e7 a1c1 a7a5 f3e5 b7a6 f4d3 f6e4 e5d7 e7d7 f2f3 e4f6 d3e5 d7e7 c4c5 b6c5 d4c5');
INSERT INTO evaluations VALUES(26968,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(26968,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(26968,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(26968,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(26968,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(26968,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(26968,36,'cp 55','c2c3 g8f6 d2d3 f8e7 h2h3 g4h5 a2a4 d8d7 a4a5 a7a6 b2b4 e8g8 f1e1 a8d8 b1d2 d6d5 e4d5 f6d5 d2e4 f7f5 e4g3 h5f7 d1b3 g8h8 f3e5 c6e5 e1e5 f5f4 g3e4 e7f6 e4f6 g7f6 e5d5 f7d5 c1f4 d5c4 b3c4 c7c6 d3d4 f8e8 f4g3 d7d5');
INSERT INTO evaluations VALUES(26960,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(26960,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(26960,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(26960,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(26960,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(26960,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(26960,36,'cp 54','f8e7 d2d3 g8f6 f1e1 e8g8 h2h3 g4e6 c4b3 e6b3 d1b3 b7b6 c1g5 d8d7 b1d2 a7a5 g5f6 e7f6 d2f1 a5a4 b3c2 g7g6 a1d1 f6g7 d3d4 a4a3 d4d5 c6d8 b2b3 f7f5 e4f5 d7f5 f1e3 f5c2 e3c2 d8f7 c3c4 g7h6 g2g3 f7g5 f3g5 h6g5 h3h4');
INSERT INTO evaluations VALUES(39169,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(39169,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39169,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(39169,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39169,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(39169,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(39169,36,'cp 9','g8f6 g1f3 e8g8 e2e3 d7d5 c1d2 b8d7 a1c1 b4e7 d1c2 h7h6 c4d5 e6d5 f1d3 c7c6 c3e2 f8e8 e1g1 e7d6 e2g3 a7a6 a2a3 d6c7 f1e1 a6a5 b2b4 c7g3 h2g3 f6e4 b4b5 c6c5 d4c5 d7c5 d2c3 c8g4 d3e2 a8c8 f3d4 e4c3 c2c3');
INSERT INTO evaluations VALUES(49967,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(49967,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(49967,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(49967,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49967,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(49967,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(49967,36,'cp 29','f8b4 a1c1 b4c3 c1c3 b7b6 e2e3 c8b7 c4c5 b8c6 f1e2 e8g8 g1f3 c6b4 a2a3 b4d5 c3c1 d7d6 e1g1 d8e8 g5h4 h7h6 h4f6 d5f6 f3d2 g8h8 c5d6 c7d6 d2b1 e8e7 b1c3 a8c8 d1d2 c8c7 c3b5 c7c1');
INSERT INTO evaluations VALUES(42921,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(42921,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(42921,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42921,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(42921,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(42921,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(42921,36,'cp 27','a1c1 b7b6 e2e3 b4c3 c1c3 c8b7 f2f3 h7h6 g5h4 g7g5 h4g3 f6h5 g3f2 h5f6 f1d3 d8e7 g1e2 b8c6 e1g1 e8c8 e3e4 f5e4 d3e4 f6e4 f3e4 h8f8 e2g3 c8b8 a2a3 f8f4 b2b4 d8f8 b4b5 c6a5 d4d5 f4f2 f1f2 f8f2 g1f2 e6d5 d1d4 d5e4 d4h8 b7c8');
INSERT INTO evaluations VALUES(17817,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(17817,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(17817,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(17817,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(17817,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(17817,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(17817,36,'cp -27','f7f6 d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 b1c3 e8c8 c1e3 d8e8 e1c1 g8e7 h3h4 h7h5 d4e2 e7g6 e2f4 g6f4 e3f4 f8e7 f2f3 g7g5 f4e3 g5h4 h1h4 f6f5 h4h2 e7d6 h2h4 d6g3 h4h1 e8g8 d1g1 h5h4 e3d4 h8h7 g1d1 d7e6 b2b3 f5e4 c3e4');
INSERT INTO evaluations VALUES(17840,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(17840,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(17840,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(17840,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(17840,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(17840,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(17840,36,'cp -22','d2d4 e5d4 d1d4 d8d4 f3d4 c6c5 d4e2 c8e6 c1e3 e8c8 b1c3 b7b6 a1d1 d8d1 e1d1 g8e7 h3h4 h7h5 d1c1 e7c6 f2f3 f8e7 b2b3 h8e8 h1d1 e7d8 e2g3 g7g6 c3e2 e6f7 e3d2 f6f5 d1h1 f5e4 g3e4 c5c4 b3b4 d8e7');
INSERT INTO evaluations VALUES(31981,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(31981,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(31981,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31981,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31981,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(31981,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(31981,36,'cp 6','e4g3 f5g6 d2d4 e7e6 c1e3 g8f6 h2h4 h7h6 f1d3 g6d3 d1d3 d8d5 g1f3 b8d7 e1g1 f8d6 c2c4 d5c6 e3d2 e8g8 a1e1 a8d8 b2b4 a7a6 g3e4 b7b5 c4c5 f6e4 e1e4 d6e7 f1e1 d7f6 e4e2 d8d5 d2f4 f8e8');
INSERT INTO evaluations VALUES(31983,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(31983,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31983,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31983,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31983,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31983,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31983,36,'cp 12','f5g6 d2d4 e7e6 h2h4 h7h6 g1h3 b8c6 c2c3 g6h7 h3f4 g8f6 c1e3 d8d7 f1d3 h7d3 f4d3 e8c8 d1f3 d7d5 f3d5 f6d5 e1c1 f8d6 g3e4 e6e5 e4d6 c7d6 d4e5 d5e3 f2e3 c6e5 h1f1 e5d3 d1d3 f7f6 c1d2 h8e8 f1f5 c8c7 d3d4 e8e5 g2g4 a7a6 c3c4 c7d7 f5e5 f6e5');
INSERT INTO evaluations VALUES(37998,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(37998,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(37998,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(37998,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(37998,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(37998,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(37998,36,'cp 57','b8c6 f2f3 g8f6 d1d2 e8g8 g2g4 h7h5 g4h5 f6h5 e1c1 d8a5 h2h4 g7d4 e3d4 c6d4 d2d4 c8e6 f1h3 e6h3 h1h3 a8c8 c1b1 e7e6 a2a3 c8c6 d4e3 a5e5 f3f4 e5f4 e3a7 f8c8 a7b7 h5f6 h4h5 f6h5 d1h1 f4e5 b7d7 c6c7 d7b5 e5b5');
INSERT INTO evaluations VALUES(16434,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16434,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16434,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(16434,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(16434,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(16434,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(16434,36,'cp -6','c1e3 f8b4 b1c3 d7d6 f1e2 e8g8 a2a3 b4c3 b2c3 c6e7 g1f3 f6g4 e3g5 h7h6 g5h4 f7f5 e4f5 c8f5 e1g1 g4f6 f1e1 e7g6 h4g3 b7b6 d3d4 f6e4 d4e5 d6e5 d1d8 a8d8 c4c5 e4g3 h2g3 f5d3 c5b6 a7b6 a3a4 d3e2 e1e2 d8d5 a1e1 f8d8 f3e5 g6e5 e2e5 d5e5');
INSERT INTO evaluations VALUES(16435,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(16435,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(16435,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(16435,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16435,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16435,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(16435,36,'cp -1','f8b4 b1c3 e8g8 a2a3 b4c3 b2c3 c6e7 g1f3 d7d6 f1e2 h7h6 f3d2 f6h7 c4c5 d6d5 f2f3 h7g5 d3d4 e5d4 c3d4 g5e6 d2b3 c8d7 a3a4 f7f5 e1g1 d8e8 b3d2 b7b6 e4d5 e7d5 e3f2 e6f4 e2c4 d7c6');
INSERT INTO evaluations VALUES(54784,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(54784,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(54784,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54784,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54784,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54784,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54784,36,'cp 0','c1g5 b8c6 f1b5 g8f6 d4d5 a7a6 b5a4 b7b5 d5c6 d8d1 a1d1 b5a4 g5f4 a4a3 b2b3 a8a7 g1e2 e7e6 f4e3 a7a8 e3f4');
INSERT INTO evaluations VALUES(39006,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(39006,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(39006,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(39006,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(39006,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(39006,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(39006,36,'cp 22','d7d5 e4d5 e6d5 c4d3 g8f6 e1g1 e8g8 f1e1 f8e8 e1e8 f6e8 h2h3 c7c6 c2c3 b8d7 a2a4 a7a5 c1f4 d7f8 b1d2 f8e6 f4g3 e8f6 d1c2 c6c5 d4c5 e6c5 d2b3 c5d3 c2d3 f6e4 g3h4 d8d7 f3d2 h7h6 d2e4 d5e4 d3e4');
INSERT INTO evaluations VALUES(38566,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(38566,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(38566,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(38566,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(38566,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(38566,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(38566,36,'cp 27','e4d5 e6d5 c4d3 g8f6 e1g1 e8g8 c2c3 f8e8 c1f4 b8d7 a2a4 c7c5 b1d2 c5d4 f3d4 d7c5 d2b3 c5d3 d1d3 c8d7 a4a5 a7a6 f1e1 f6e4 d3d1 d8h4 f4e3 a8c8 h2h3 e4d6 d4f3 h4h5 e3d4 e8e1 f3e1 h5d1 a1d1 c8e8 d4g7 g8g7');
INSERT INTO evaluations VALUES(38565,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38565,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(38565,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(38565,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(38565,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(38565,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38565,36,'cp 21','e6d5 c4d3 g8f6 d1e2 c8e6 f3g5 d8d6 g5e6 d6e6 b1d2 e6e2 e1e2 e8g8 h1d1 a7a5 a2a4 g7h6 e2f1 f8e8 g2g3 c7c6 f1g2 e8e7 b2b3 b8a6 c2c3 e7e6 c1b2 a6c7 d2f3 h6g7 f3e5 a8e8 d1e1 f6d7 f2f4 f7f6');
INSERT INTO evaluations VALUES(38580,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38580,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(38580,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38580,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38580,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38580,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38580,36,'cp 30','c4d3 g8f6 d1e2 c8e6 f3g5 d8d6 g5e6 f7e6 c1e3 b8d7 f2f4 c7c5 c2c3 c5d4 c3d4 f6e4 e1g1 e8g8 g1h1 d7f6 b1d2 e4d2 e3d2 f6e4 d3e4 d5e4 e2e4 g7d4 a1d1 d6b6 d2a5 b6a5 d1d4 a5a2 f1d1 a8e8 e4b7 e8b8 b7e4 b8b2');
INSERT INTO evaluations VALUES(45984,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(45984,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(45984,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45984,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45984,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45984,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45984,36,'cp 0','f8e7 d2d4 c5d4 f3d4 e8g8 c1e3 b8c6 c4b3 c8d7 d1e2 a7a6 f2f4 d8c7 f4f5 c6d4 e3d4 e6f5 e2e3 a8e8 e4f5 d7c6 e3g3 f6h5 g3g4 h5f6 g4g3');
INSERT INTO evaluations VALUES(40174,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40174,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(40174,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40174,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40174,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40174,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40174,36,'cp 0','d2d4 c5d4 f3d4 e8g8 a2a4 b7b6 d1d3 c8b7 d3g3 d8c8 c4d5 b8d7 c1h6 f6h5 g3h3 h5f6 h3g3');
INSERT INTO evaluations VALUES(40186,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40186,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40186,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40186,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40186,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40186,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40186,36,'cp 0','c5d4 f3d4 e8g8 c1e3 b7b6 d1f3 d8c7 c4b3 b8c6 a1d1 c6d4 d1d4 a7a6 f3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(40201,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40201,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40201,36,'cp 0','f3d4 e8g8 c1e3 b7b6 d1f3 d8c7 c4b3 b8c6 f3g3 f6h5 g3h3 h5f6');
INSERT INTO evaluations VALUES(40199,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40199,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40199,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40199,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40199,36,'cp 0','e8g8 d1e2 a7a6 a2a4 c8d7 c1e3 b8c6 d4c6 d7c6 a4a5 f6d7 f1d1 a8c8 h2h3 d8c7 e3d4 e7g5 d4e3');
INSERT INTO evaluations VALUES(35818,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35818,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35818,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35818,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35818,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35818,36,'cp 0','c1e3 b7b6 a2a4 c8b7 f2f3 f8e8 c4b5 e8f8');
INSERT INTO evaluations VALUES(45202,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(45202,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(45202,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(45202,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(45202,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45202,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(45202,36,'cp -7','e7e5 g2g3 f8e7 f1g2 b8c6 e1g1 e8g8 c3c4 a7a6 b1c3 a8b8 f3e1 b7b5 e1c2 c8g4 f2f3 g4d7 b2b3 b5c4 d3c4 a6a5 c1d2 c6d4 f1f2 d7c6 c2e3 d8d7 g1h1 f8e8 g2f1 b8b7 f1d3 e7d8 e3d5 f6d5 c4d5');
INSERT INTO evaluations VALUES(45914,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45914,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45914,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(45914,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45914,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45914,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(45914,36,'cp 0','g2g3 f8e7 f1g2 e8g8 e1g1 b8c6 h2h3 a8b8 f1e1 c8e6 f3g5 e6c8');
INSERT INTO evaluations VALUES(47835,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(47835,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(47835,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(47835,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(47835,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47835,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(47835,36,'cp 38','c7c6 e2e4 c8e6 d2d3 f8e7 g1f3 e8g8 e1g1 a7a6 a2a4 b8d7 a4a5 b7b6 a5b6 d8b6 c3a4 b6b7 h2h3 d7b6 a4c3 b6d7 b2b3 d7c5 a1b1 a6a5 c1a3 b7c7 f1e1 c5a6 d3d4 a6b4 g1h2 a8b8 b1a1 f8e8 a3b2 e6d7 c3a4 c6c5 d4c5 d7a4 a1a4 c7c5');
INSERT INTO evaluations VALUES(45871,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(45871,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(45871,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(45871,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(45871,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(45871,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(45871,36,'cp 45','e2e4 c6c5 g1e2 a7a6 d2d3 f8e7 e1g1 b8c6 f2f4 c8g4 h2h3 g4e2 c3e2 b7b5 h3h4 a8b8 f4f5 b5b4 g2f3 a6a5 b2b3 d8c7 g1g2 f6d7 a2a3 e7f6 f3h5 e8e7 a3b4 c5b4 c1e3 c7b7 f1h1 h8g8');
INSERT INTO evaluations VALUES(38624,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(38624,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(38624,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38624,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(38624,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38624,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(38624,36,'cp 7','g1f3 g8e7 e1g1 e8g8 c1g5 b8c6 c2c3 f7f6 g5h4 c8f5 f1e1 d8d7 h4g3 a8e8 g3d6 d7d6 b1a3 f5d3 d1d3 c6d8 e1e3 d6d7 a1e1 e7g6 a3c2 c7c6 e3e8 f8e8 g2g3 d8f7 h2h4 f7d6 c2e3 g6f8 a2a4 g7g6 d3d1 g8f7 f3d2 f8e6 e3g2');
INSERT INTO evaluations VALUES(52800,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(52800,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(52800,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(52800,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(52800,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52800,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(52800,36,'cp -25','d5d4 c3e2 e7e5 f2f4 e5f4 g1f3 c6c5 e2f4 b8c6 f1e2 f8d6 f4h5 g7g6 h5f4 g8f6 e1g1 d8c7 g2g3 h7h6 c2c4 d4c3 b2c3 g6g5 f4g2 c8e6 c1b2 h8g8 d3d4 c5d4 f3d4 g8g6 d4e6 f7e6 d1b3 c7b6 g1h1 b6b3 a2b3 e8c8 e2c4 f6g4 h2h3 g4e5 c4e2 d6c5');
INSERT INTO evaluations VALUES(52907,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(52907,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(52907,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(52907,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52907,36,'cp 0','c3e2 e7e5 f2f4 e5f4 g1f3 c6c5 e2f4 b8c6 g2g3 g8f6 f1g2 f8d6 e1g1 c8g4 c2c4 e8g8 d1c2 f6d7 f4d5 d7e5 f3h4 g4e6 c2e2 e6g4');
INSERT INTO evaluations VALUES(52908,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(52908,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(52908,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(52908,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(52908,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(52908,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(52908,36,'cp -19','e7e5 f2f4 e5f4 e2f4 c6c5 g2g3 b8c6 g1f3 g8f6 f1g2 f8d6 e1g1 e8g8 f4d5 f6g4 d1e2 g4e5 c1f4 f7f6 c2c4 c8g4 e2d2 a8b8 a2a4 d8a5 d2c2 g8h8 f3e5 c6e5 b2b3 b8e8 h2h3 g4d7 c2d2 a5d8 g1h1');
INSERT INTO evaluations VALUES(53787,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(53787,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(53787,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(53787,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(53787,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(53787,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(53787,36,'cp -13','f2f4 e5f4 e2f4 c6c5 g1f3 b8c6 g2g3 g8f6 f1g2 f8d6 f4d5 e8g8 e1g1 f6g4 d1e2 g4e5 c1f4 c8g4 e2d2 e5f3 g2f3 g4e6 c2c4 d4c3 b2c3 b7b5 a1b1 b5b4 c3b4 d6f4 g3f4 e6d5 e4d5 c6b4 d5d6 d8d6 f3a8 f8a8');
INSERT INTO evaluations VALUES(56404,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(56404,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(56404,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(56404,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(56404,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(56404,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(56404,36,'cp -76','d5e4 b1c3 f7f5 d2d3 g8f6 d3e4 d8d1 e1d1 b8c6 c1e3 f8d6 g1e2 c8d7 d1c1 e8c8 a2a3 c6e7 e4f5 h8f8 h1d1 a7a6 b2b4 h7h6 c3e4 f6e4 g2e4 e7f5 e4f5 d7f5 d1d2 c8d7 c2c4 d7e6 e2c3 c7c6 e3b6 d8d7 c1b2 d6e7 d2d7 e6d7');
INSERT INTO evaluations VALUES(56622,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(56622,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(56622,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(56622,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(56622,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(56622,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(56622,36,'cp -75','b1c3 g8f6 c3e4 f6e4 g2e4 f8c5 d2d3 e8g8 e4g2 b8c6 c1e3 c5d4 d1c1 d8d6 g1f3 d4e3 f2e3 e5e4 d3e4 d6b4 c2c3 b4e4 e1g1 c6e5 f3d4 e4d3 e3e4 c8g4 f1f2 a8e8 c1f4 d3a6 a2a4 g4d7 d4f5 f7f6 a4a5 b7b6 h2h3 d7e6');
INSERT INTO evaluations VALUES(10251,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10251,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(10251,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(10251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10251,36,'cp -9','d7d5 e4d5 f6d5 e1g1 e8g8 f1e1 c8g4 h2h3 g4h5 b1c3 d5c3 b2c3 f8e8 c1e3 c5b6 a2a4 d8f6 g3g4 h5g6 f3d2 a8d8 d2e4 f6e7 e4g3 b6a5 e3d2 e7d7 d1f3 a5b6 d2g5 f7f6 g5d2 g8h8 d2e3 e8f8 f3e2 b6e3 e2e3 h7h6 a4a5 b7b6');
INSERT INTO evaluations VALUES(9093,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(9093,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(9093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9093,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9093,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9093,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9093,36,'cp 0','e4d5 f6d5 b1c3 e8g8 e1g1 d5c3 b2c3 c8f5 f1e1 f8e8 f3h4 f5e6 g2c6 b7c6 e1e5 c5e7 h4g2 e7f6 e5c5 d8d7 c1g5 e6g4 d1f1 f6e5 g5f4 e5f6');
INSERT INTO evaluations VALUES(17363,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(17363,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(17363,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(17363,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(17363,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(17363,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(17363,36,'cp -40','b7b5 c4a2 c8b7 d2d3 f8d6 c3e2 g8f6 c2c3 d8c7 a2b1 c6e7 e1g1 c5c4 a3a4 e8g8 a4b5 a6b5 a1a8 b7a8 e2d4 f8b8 f1e1 e6e5 d4c2 c4d3 d1d3 h7h6 c1d2 d6c5 c2b4 c7b6 d2e3 d7d6 d3e2 a8e4 e3c5 e4f3 c5b6 f3e2');
INSERT INTO evaluations VALUES(18114,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(18114,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(18114,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(18114,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(18114,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(18114,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(18114,36,'cp -36','c4a2 g8f6 e1g1 c8b7 d2d3 d8c7 a3a4 b5b4 c3e2 f8d6 c2c3 h7h6 e2g3 e8g8 a2b1 b4c3 b2c3 c6e7 d1e1 e7g6 d3d4 c5d4 c3d4 d6f4 b1d3 f8b8 g3e2 f4c1 a1c1 c7d8 e2g3 b7c6 d3c2 d8c7 e1c3 b8c8 f3e5 g6e5 d4e5');
INSERT INTO evaluations VALUES(18131,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(18131,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(18131,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(18131,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(18131,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(18131,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(18131,36,'cp -35','g8f6 e1g1 c8b7 d2d3 f8d6 a3a4 d8b6 h2h3 e8g8 f1e1 b5b4 c3b1 b6c7 c1e3 d6e7 b1d2 d7d5 c2c3 f8d8 e4e5 b4c3 b2c3 c6e5 f3e5 c7e5 e3c5 e5c7 c5e7 c7e7 d2f3 h7h6 d1d2 a6a5 f3d4 e7c7 a1c1 b7a6 d2e3 c7b6 c1b1 b6d6 b1c1 d8d7 d4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(28027,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(28027,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(28027,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(28027,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(28027,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(28027,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(28027,36,'cp 102','d7d6 f1e2 g8f6 d1c2 e8g8 c1g5 b8d7 a1d1 h7h6 g5e3 e7e5 e1g1 d8e7 d4e5 d7e5 f3e5 e7e5 f2f3 f6d7 c2d2 g8h7 c3b5 f8c8 e3d4 e5e7 d4g7 h7g7 b2b3 g7h7 e2d3 a7a5 b5c3 c8f8 f3f4 d7c5 d1e1 a8e8 d3c2 e7f6 c3d5 b7d5 c4d5');
INSERT INTO evaluations VALUES(27898,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(27898,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(27898,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(27898,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(27898,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(27898,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(27898,36,'cp 98','d1c2 g8f6 f1e2 e8g8 c1e3 e7e5 d4d5 b7c8 h2h3 f6h5 g2g3 d8e7 f3g5 f7f6 g5e6 c8e6 d5e6 e7e6 c3d5 e6d7 e2h5 g6h5 e1c1 b8c6 c1b1 g8h8 g3g4 c6d4 e3d4 e5d4 f2f3 h5h4 d5f4 c7c5 f4g2 d7b7 g2h4 a8d8 h1g1 b6b5');
INSERT INTO evaluations VALUES(38695,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(38695,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(38695,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38695,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(38695,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38695,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38695,36,'cp 1','g8f6 e1g1 e8g8 d2d4 e5d4 f3d4 c8g4 f2f3 b8c6 c1e3 g4e6 c4e6 f7e6 d1d2 d8d7 d4e2 c5e3 d2e3 a7a6 a1d1 a8e8 b2b3 e6e5 c3d5 f6d5 e4d5 c6e7 e2g3 a6a5 g3e4 e7f5 e3d3 b7b6 c2c4 h7h6 a2a3 f5d4 b3b4 a5b4 a3b4 d7f7');
INSERT INTO evaluations VALUES(53078,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(53078,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(53078,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(53078,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(53078,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(53078,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(53078,36,'cp -26','f1b5 b8d7 f3g3 g7g6 b5e2 f8g7 a2a4 d8a5 b1c3 a7a6 a1a3 d7e5 d2d3 e5c6 f2f4 c6d4 e2d1 e7e6 g1f3 g8e7 e1g1 a5c7 f3d4 g7d4 c1e3 d4g7 g3f2 e8g8 a4a5 c8d7 d3d4 c5d4 e3d4 g7d4 f2d4 e7c6 d4b6 c7b6 a5b6 c6b4 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(48998,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(48998,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(48998,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(48998,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(48998,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(48998,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(48998,36,'cp 221','d1h5 g7g6 h5f3 c8e6 b1c3 c7c6 g5e6 f7e6 c3e4 b8d7 d2d3 d8e7 e1g1 e7f7 f3e2 f8e7 c1d2 f7f5 g2g3 e8c8 a2a4 h8f8 d2g5 f5f7 a1b1 c8b8 g5d2 f7f3 e2f3 f8f3 a4a5 d7c5 g1g2 f3f8 d2h6 f8f7 b2b4 c5e4 d3e4 d5c7 h6e3 a7a6 c2c3 b8a8 c4e2 d8f8');
INSERT INTO evaluations VALUES(10250,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10250,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(10250,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(10250,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(10250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10250,36,'cp -3','f1g2 d7d5 e4d5 f6d5 e1g1 e8g8 b1c3 d5c3 b2c3 f7f6 f1e1 c8g4 h2h3 g4h5 a2a4 c5b6 c1e3 a8b8 e3b6 c7b6 d3d4 d8c7 g3g4 h5f7 d4e5 b8d8 d1c1 c6e5 f3e5 f6e5 c1e3 f7d5 e3e5 c7e5 e1e5 d5g2 g1g2 d8d2 a1f1 h7h6 a4a5 d2c2 a5b6 a7b6 e5e7 f8f7 e7e4');
INSERT INTO evaluations VALUES(17361,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(17361,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(17361,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(17361,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(17361,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(17361,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(17361,36,'cp 12','d2d4 d7d5 e4d5 e6d5 f1e1 f8e7 c4f1 c5d4 f3d4 g8f6 d4c6 b7c6 c2c4 e8g8 c4d5 f6d5 d1c2 c8e6 a2a3 a6a5 b1d2 d8d7 d2c4 e7f6 c1d2 a5a4 h2h3 d7a7 c2e4 f6d4 e4h4 h7h6 a1d1 d4f6 h4g3 a7b8 g3f3 f6d4 f1d3 d4b2 d3e4');
INSERT INTO evaluations VALUES(17376,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(17376,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(17376,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(17376,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(17376,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(17376,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(17376,36,'cp 4','d7d5 e4d5 e6d5 f1e1 f8e7 c4f1 c5d4 h2h3 g8f6 f3d4 e8g8 c1e3 h7h6 b1c3 f8e8 a2a3 e7d6 d1d2 c8e6 e3f4 d6c5 d4c6 b7c6 b2b4 c5f8 a1d1 c6c5 b4c5 f8c5 f4e5 c5a3 e5f6 d8f6 c3d5 e6d5 e1e8 a8e8 d2d5 a3e7 d5b3 e8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(56993,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(56993,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(56993,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(56993,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(56993,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(56993,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(56993,36,'cp -106','e6d5 d1d5 g8f6 d5d1 d7d5 e2e3 f8e7 f1d3 e8g8 g1e2 c7c5 b2b3 e7d6 c1b2 b8c6 h2h3 a7a6 e1g1 b7b5 e2g3 c5c4 d3e2 d8e7 e2f3 f8d8 d1e2 d6e5 b2e5 e7e5 c2c3 c8b7 e2c2 d8e8 b1d2 a8d8 a1d1 g7g6 g3e2 e5c7 e2f4');
INSERT INTO evaluations VALUES(57099,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(57099,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(57099,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(57099,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(57099,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(57099,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(57099,36,'cp -87','d1d5 g8f6 d5d1 d7d5 e2e3 c7c5 f1b5 c8d7 c2c4 f8e7 b5d7 d8d7 c4d5 f6d5 g1f3 e8g8 e1g1 d7e6 d1b3 b8c6 c1d2 b7b5 e3e4 d5c7 b3e6 f7e6 b1c3 a7a6 d2g5 e7g5 f3g5 c6d4 a1c1');
INSERT INTO evaluations VALUES(57098,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(57098,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(57098,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(57098,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(57098,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(57098,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(57098,36,'cp -107','g8f6 d5d1 f8e7 e2e3 d7d5 f1d3 e8g8 g1e2 c7c5 b2b3 e7d6 c1b2 b8c6 h2h3 a7a6 e1g1 b7b5 e2g3 c5c4 d3e2 d8e7 b3c4 d6g3 f2g3 d5c4 e2f3 e7e3 g1h1 f8d8 d1e1 e3e1 f1e1 c8b7 b2f6 g7f6 b1c3 g8g7 a1d1 d8d1 c3d1 c6a5 d1e3 b7f3');
INSERT INTO evaluations VALUES(49485,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(49485,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(49485,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(49485,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(49485,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(49485,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(49485,36,'cp -102','d5d1 f8e7 e2e3 d7d5 f1d3 c7c5 b2b3 b8c6 c1b2 c8e6 g1f3 d8a5 c2c3 c5c4 d3c2 e8g8 e1g1 a5c7 a2a4 a7a6 b2a3 c4b3 c2b3 e7a3 a1a3 c7e7 h2h3 a8c8 f3d4 c6d4 d1d4 f8d8 a3a1 f6e4 f1d1 e7g5');
INSERT INTO evaluations VALUES(23435,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(23435,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(23435,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(23435,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(23435,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(23435,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(23435,36,'cp 28','g8f6 g1f3 d7d5 c1f4 c8f5 b1d2 e7e6 f3h4 f8d6 h4f5 e6f5 f4g5 h7h6 g5f6 d8f6 e2e3 c6e7 g2g3 c7c6 f1g2 a7a5 c3c4 f6e6 c4d5 e7d5 e1g1 d5f6 d1c2 e8g8 a1c1 a8b8 a2a3 g7g6 d2b1');
INSERT INTO evaluations VALUES(17258,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(17258,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(17258,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17258,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(17258,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17258,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(17258,36,'cp 23','g1f3 d7d5 c1f4 e7e6 b1d2 f8d6 f4g5 e6e5 d4e5 d6e5 f3e5 c6e5 d1a4 c8d7 a4f4 d8e7 e2e3 e8c8 d2b3 h7h6 g5f6 e7f6 f4f6 g7f6 f1e2 b7b6 b3d4 c7c5 d4c2 c8c7 h2h4 d7f5 e1c1 h6h5 f2f3 f5c2 c1c2 f6f5 a2a4 a7a5');
INSERT INTO evaluations VALUES(17257,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(17257,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(17257,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(17257,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(17257,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(17257,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(17257,36,'cp 26','d7d5 c1f4 c8f5 b1d2 e7e6 f3h4 f8d6 h4f5 e6f5 f4g5 h7h6 g5f6 d8f6 e2e3 c6e7 g2g3 e8g8 f1d3 f6e6 d1h5 c7c6 e1g1 g7g5 c3c4 d6b4 c4d5 c6d5 h5e2 b4d2 e2d2 f8c8 f1c1 g8g7 d2a5 c8c1 a1c1 a8c8 c1c8 e6c8');
INSERT INTO evaluations VALUES(31428,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(31428,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(31428,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(31428,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(31428,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(31428,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(31428,36,'cp 29','h2h4 h7h6 f3e5 g6h7 f1d3 h7d3 d1d3 b8d7 d3e2 g8f6 e1g1 f8e7 c2c4 d8b6 f1d1 e8g8 e5d7 f6d7 c1f4 a8d8 d1d3 d7f6 f4e5 d8d7 b2b3 b6d8 a1d1 e7d6 e2f3 f8e8 a2a4 d8e7 e5f6 e7f6 f3f6 g7f6 g3h5 d6e7 g2g4 e8d8 g1g2 f6f5 g4f5');
INSERT INTO evaluations VALUES(31423,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(31423,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(31423,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(31423,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(31423,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31423,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31423,36,'cp 27','h7h6 f3e5 g6h7 f1d3 h7d3 d1d3 b8d7 f2f4 f8e7 c1d2 d7e5 f4e5 e7h4 e1c1 h4g3 d3g3 g8e7 g3g7 h8g8 g7h7 d8d5 d1f1 e8c8 f1f7 g8g2 f7e7 g2g7 h7h6 g7e7 d2g5 e7d7 g5d8 d7d8 c1b1 c6c5 d4c5 d5e5 a2a3 e5c5 h6e6 c8b8 e6f6 c5d4 f6e7 d8d7 e7e6 d7d8 e6e2 a7a6 b1a2 d8c8');
INSERT INTO evaluations VALUES(31375,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(31375,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31375,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(31375,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(31375,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(31375,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(31375,36,'cp 29','c1f4 b8d7 f1d3 g6d3 d1d3 g8f6 e1c1 f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e1 d8d5 f3e5 d5e4 e1e4 f6e4 f4e3 h6h5 c2c3 e4d6 g2g3 d6f5 h1e1 a8d8 e3f4 f7f6 e5c4 g8f7 c4a5 d8d7 b1c2 g7g5 h4g5 f6g5 f4e5 g5g4');
INSERT INTO evaluations VALUES(16697,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(16697,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(16697,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(16697,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(16697,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(16697,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(16697,36,'cp -46','f8b4 a2a3 b4c3 e3c3 e8g8 f2f3 d7d5 c1g5 d5e4 g5f6 d8f6 c3f6 g7f6 f3e4 f8e8 e1c1 c8g4 f1e2 e8e4 e2g4 e4g4 g2g3 a8e8 g1h3 g4d4 h3f4 g8g7 f4d5 e8e2 d5c7 d4d1 c1d1 e2f2 c7d5 c6d4 d5e3 g7g6 h1f1 f2h2 f1f4 h2h1 d1d2 d4e6 f4a4 a7a6 a4b4 h1h3 g3g4 e6c5 d2e2 g6g5');
INSERT INTO evaluations VALUES(10486,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(10486,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(10486,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(10486,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(10486,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(10486,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(10486,36,'cp -50','a2a3 b4c3 e3c3 e8g8 f2f3 d7d5 c1g5 d5e4 g5f6 d8f6 c3f6 g7f6 f3e4 f8e8 e1c1 c8g4 f1e2 e8e4 e2g4 e4g4 g2g3 a8e8 g1h3 g4d4 h3f4 g8g7 f4d5 e8e2 d5c7 d4d1 c1d1 e2f2 c7d5 c6d4 d5e3 g7g6 h1f1 f2h2 f1f4 h2h1 d1d2 d4e6 f4a4 a7a6 a4b4 h1h3 g3g4 e6c5 b4d4 h7h5 g4h5 h3h5');
INSERT INTO evaluations VALUES(41091,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41091,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41091,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(41091,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(41091,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41091,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(41091,36,'cp 32','f1d3 e8g8 h2h3 c7c5 d4c5 e7c5 e1g1 a7a6 f1e1 h7h6 c1f4 b8c6 a2a3 c8e6 b2b4 c5e7 f3e5 c6e5 f4e5 f6d7 e5d4 a8c8 c3e2 e7f6 c2c3 f8e8 d3c2 f6d4 d1d4 d8f6 a1d1 f6d4 d1d4 d7e5 c2b3 e5c4 b3c4 c8c4');
INSERT INTO evaluations VALUES(41094,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(41094,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(41094,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41094,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41094,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(41094,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(41094,36,'cp 20','e8g8 h2h3 c7c5 d4c5 b8d7 e1g1 d7c5 f1e1 h7h6 c1f4 c8d7 d1d2 f8e8 a1d1 d7c6 f3d4 c5d3 d2d3 e7c5 e1e8 d8e8 f4e3 c5f8 b2b3 e8d7 c3e2 a8c8 a2a4 a7a6 a4a5 d7c7 d4c6 c7c6 c2c3 f8c5 e3d4 c5d4 d3d4');
INSERT INTO evaluations VALUES(36409,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36409,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36409,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36409,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36409,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(36409,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36409,36,'cp 24','h2h3 c7c5 d4c5 b8d7 e1g1 d7c5 f1e1 h7h6 c1f4 c8d7 d1d2 f8e8 a1d1 d7c6 f4e5 e7d6 e5d6 d8d6 d3f5 d6c7 f3d4 c5e4 f5e4 f6e4 d2e3 e4f6 e3g3 c7g3 f2g3 c6d7 g1f2 f6e4 c3e4 d5e4 g3g4 a7a6');
INSERT INTO evaluations VALUES(36407,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36407,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(36407,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(36407,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36407,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(36407,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(36407,36,'cp 16','f8e8 e1g1 c7c6 f1e1 b8d7 c1f4 d7f8 d1d2 a7a5 a2a3 f8e6 f4e5 g7g6 c3d1 f6h5 e5h2 h5g7 a3a4 e6g5 f3g5 e7g5 e1e8 g7e8 d1e3 e8g7 f2f4 g5f6 c2c3 c8f5 d3f5 g7f5 e3f5 g6f5 d2d3 d8d7 a1e1 b7b5 d3g3 g8h8 a4b5');
INSERT INTO evaluations VALUES(29227,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(29227,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29227,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(29227,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(29227,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(29227,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(29227,36,'cp 12','f8d6 f3e5 d5c4 g2g4 f5g6 h2h4 f6d5 f1c4 d5f4 e3f4 d6e5 d4e5 d8d1 a1d1 h7h5 c4e2 h5g4 e2g4 b8d7 h4h5 a8d8 h1h3 g6f5 g4f5 e6f5 c3e2 d7c5 e2d4 g7g6 h5h6 e8g8 b2b4 c5e4 h3d3 c6c5 d4f5 d8d3 f5e7 g8h7 d1d3 c5b4');
INSERT INTO evaluations VALUES(28365,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(28365,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(28365,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(28365,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(28365,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(28365,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(28365,36,'cp 20','f4d6 d8d6 f1d3 f5d3 d1d3 e8g8 e1g1 b8d7 a1c1 d6e7 b2b3 a8c8 h2h3 f8d8 d3b1 a7a6 f1d1 h7h6 b1b2 c8c7 a2a4 a6a5 c3e2 d8c8 e2f4 b7b6 c1c2 c8d8 d1c1 c6c5 c4d5 f6d5 f4d5 e6d5 d4c5 b6c5 b2c3 c5c4 b3c4');
INSERT INTO evaluations VALUES(28362,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(28362,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(28362,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(28362,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(28362,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(28362,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(28362,36,'cp 27','d8d6 d1b3 d6e7 f3e5 a7a5 f2f3 h7h5 f1e2 e8g8 e1g1 f8e8 a1c1 a5a4 b3d1 a4a3 b2b3 b8d7 f3f4 g7g6 e2d3 f6g4 d1d2 f5d3 d2d3 d7f6 d3d2 g4h6 h2h3 h6f5 c4c5 a8d8 d2e1 f6e4');
INSERT INTO evaluations VALUES(32548,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32548,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(32548,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(32548,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(32548,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(32548,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(32548,36,'cp 20','d1b3 d6e7 f3e5 a7a5 f1e2 d5c4 b3c4 h7h6 e1g1 e8g8 f1d1 b8d7 e5d7 f6d7 e3e4 f5g6 a1c1 f8d8 c4d3 c6c5 d3b5 c5d4 d1d4 d7f6 c1d1 d8d4 d1d4 a8d8 d4d8 e7d8 f2f3 b7b6 b5d3 d8c7 d3e3 f6d7 e2b5 d7e5');
INSERT INTO evaluations VALUES(45219,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45219,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45219,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(45219,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45219,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45219,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(45219,36,'cp 0','b8c6 f1b5 c8d7 e4e5 f6g4 h2h3 g4h6 g2g4 e7e6 d2d3 h6g8 f4f5 h7h5 f5e6 d7e6 f3g5 h5g4 e1g1 h8h3 f1f7 e6f7 d1g4 h3h6 g5f7 e8f7 g4f3 f7e6 f3g4');
INSERT INTO evaluations VALUES(13338,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(13338,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13338,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(13338,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(13338,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(13338,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13338,36,'cp 20','f1b5 c8d7 d2d3 a7a6 b5c4 e7e6 e1g1 d8c7 c4b3 f8e7 d1e1 e8g8 a2a3 b7b5 c1e3 c6a5 b3a2 b5b4 c3e2 b4a3 b2a3 a8b8 e3d2 a5c6 h2h3 a6a5 e1f2 d6d5 e4e5 f6e8 a1e1 a5a4 d2c3 d5d4 c3a1 c7d8 a2c4 e8c7');
INSERT INTO evaluations VALUES(13222,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(13222,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13222,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(13222,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(13222,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(13222,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13222,36,'cp 0','c8d7 d2d3 e7e6 e1g1 a7a6 b5c4 d8c7 c4b3 f8e7 d1e2 e8g8 a2a4 c6a5 b3a2 a5c6 a2b3');
INSERT INTO evaluations VALUES(24482,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24482,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24482,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24482,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24482,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24482,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24482,36,'cp 0','d2d3 e7e6 e1g1 a7a6 b5c4 d8c7 d1e1 c6d4 c4b3 a6a5 f4f5 a5a4 b3c4 d4c2 e1e2 c2a1 f3g5 a1c2 e2c2 c7c8 c3b5 c8c6 b5a3 c6c8 a3b5');
INSERT INTO evaluations VALUES(51850,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(51850,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(51850,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(51850,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(51850,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(51850,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(51850,36,'cp 19','d2d4 e5e4 d4d5 f8b4 c1d2 d8e7 d2b4 e7b4 d1d2 b4c4 b1c3 g8f6 e2e3 c4c5 a1c1 c5a5 f1g2 e8g8 g2e4 f6e4 c3e4 a5a2 e4d6 b8a6 g1e2 a2d5 d2d5 c6d5 e1g1 b7b6 f1d1 a6c5 d1d5 c8a6 e2d4 c5e6 b2b4 e6d4 d5d4');
INSERT INTO evaluations VALUES(51853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51853,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(51853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51853,36,'cp 0','e5e4 d4d5 f8b4 c1d2 d8e7 d2b4 e7b4 b1c3 g8f6 d1d2 b4c4 e2e3 c4b4 a2a3 b4b3 d5d6 b7b6 g1e2 c6c5 e2f4 c8b7 f1e2 b8c6 e2d1 b3c4 d1e2');
INSERT INTO evaluations VALUES(5291,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(5291,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(5291,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(5291,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(5291,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(5291,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5291,36,'cp 0','e7e6 c1f4 d7d6 c3b5 e6e5 f4g5 g8e7 b5c3 h7h6 g5e7 c6e7 c3d5 c8d7 a2a4 e7d5 c4d5 d7c6 d5c6 b7c6 e1g1 e8g8 a1b1 a8b8 g2g3 d8d7 b2b3 h6h5 d1d2 g8h7 f3g5 h7g8');
INSERT INTO evaluations VALUES(5102,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5102,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5102,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5102,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5102,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5102,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5102,36,'cp 0','c3b5 d7d6 c1f4 e6e5 f4g5 g8e7 b5c3 h7h6 g5e7 c6e7 c3d5 c8d7 d5e7 d8e7 a2a4 h6h5 e1g1 e8g8 d1d2 g8h7 f3g5 h7h8 g5f3');
INSERT INTO evaluations VALUES(3805,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(3805,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(3805,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(3805,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(3805,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(3805,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(3805,36,'cp -34','h2h3 c6a5 c4b3 a5b3 a2b3 a7a5 e1g1 h7h6 g5f3 c8d7 c1e3 c5e3 f2e3 c7c6 d1e1 b7b5 e1g3 d7e6 f3h4 f6h5 g3e1 g8h7 e1f2 h5f6 f2g3 g7g6 g3f2 h7g7 c3e2 d8e7 e2c3 e6d7 h4f3 f8e8 g1h1 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(3803,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(3803,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(3803,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(3803,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(3803,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(3803,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(3803,36,'cp -36','h7h6 g5f3 c6a5 c4b3 a5b3 a2b3 a7a5 e1g1 c8d7 c1e3 c5e3 f2e3 c7c6 d1e1 b7b5 e1f2 f8e8 f3h4 d7e6 f2g3 f6h5 g3f2 g8h8 f2e1 h8h7 h4f3 h5f6 e1h4 h7g8 h4f2 d8b6 f3h4 g8h8 g1h1 d6d5 f2g3 d5e4 f1f6 g7f6');
INSERT INTO evaluations VALUES(3587,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(3587,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(3587,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(3587,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(3587,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(3587,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(3587,36,'cp -32','g5f3 c6a5 c4b3 a5b3 a2b3 a7a5 e1g1 f8e8 c1e3 c5e3 f2e3 c7c6 d1e1 c8e6 e1h4 b7b5 h4g3 f6h5 g3f2 g8h8 f3h4 d8b6 g1h1 h5f6 f2e1 b6d8 h4f5 b5b4 c3a4 f6g8 h1g1 g7g6 f5h4 g8f6 e1f2 f6h7 h4f3 g6g5 d3d4 g5g4 h3g4');
INSERT INTO evaluations VALUES(3605,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(3605,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(3605,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(3605,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(3605,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(3605,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(3605,36,'cp -32','c6a5 c4b3 a5b3 a2b3 a7a5 e1g1 f8e8 c1e3 c5e3 f2e3 c7c6 d1d2 c8d7 d2f2 b7b5 f2g3 d7e6 f3h4 f6h5 g3e1 g8h7 h4f3 h7h8 e1f2 h5f6 f2h4 d8e7 c3e2 h8g8 h4f2 e7d8 e2g3 g8h7 f3h4 e8g8 h4f5 d6d5 d3d4');
INSERT INTO evaluations VALUES(14794,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(14794,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(14794,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(14794,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(14794,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(14794,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(14794,36,'cp 8','f8d6 f3e5 e8g8 f1d3 f6d7 e5d7 c8d7 f4g3 a7a6 e1g1 d8e7 a2a3 a8d8 d1h5 g7g6 h5e2 d8c8 e2d2 f7f5 a1e1 d6g3 h2g3 c6d8 c3e2 c7c5 c2c4 d5c4 d3c4 b7b5 c4b3 c5c4 b3c2 d7c6 e2f4 d8f7 f2f3');
INSERT INTO evaluations VALUES(9232,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9232,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(9232,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(9232,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9232,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9232,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(9232,36,'cp 20','f4g3 e8g8 a2a3 c6e7 f1d3 c7c5 d4c5 d6c5 e3e4 e7g6 e4d5 e6d5 e1g1 c8g4 h2h3 g4h5 g3h2 d8c8 d3g6 f7g6 c3d5 c8f5 d5f4 a8d8 d1e2 h5f3 e2e6 f5e6 f4e6 f3c6 e6c5 d8d2 b2b4 d2c2 f1c1 c2c1 a1c1 f8e8 f2f3 h7h6 h2g3 e8e2 c5b3');
INSERT INTO evaluations VALUES(52183,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(52183,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(52183,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(52183,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(52183,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(52183,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(52183,36,'cp 57','g1f3 g8f6 f3e5 b8c6 e5c6 d7c6 d1d8 e8d8 c1e3 f8d6 f1e2 d8e7 e3d4 d6c7 d4c5 e7e8 f2f3 c8e6 e1f2 f6d7 c5e3 e8e7 f3f4 f7f6 a1c1 c7b6 h2h4 b6e3 f2e3 h8e8 b2b3 d7b6 e2f3 a8d8 h1d1 d8d7 d1d7 b6d7 c3a4 e6f7 a4c5 d7c5 c1c5');
INSERT INTO evaluations VALUES(52182,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52182,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(52182,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(52182,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(52182,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(52182,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(52182,36,'cp 36','b8c6 f1c4 g8f6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8c5 d1h5 g7g6 h5d1 c8f5 c1e3 f7f6 g5e4 c5b6 d5d6 e8g8 a2a4 g8g7 f1e1 a8c8 a4a5 b6a5 e3d4 e5d4 d1d4 d8b6 d4d5 a5c3 b2c3 b6c6 d5d4 f5e4 e1e4 c6c3 d4a7 f8f7 a1d1 f7d7 h2h3 c3c5 a7a2 d7d6 e4e7 g7h6 d1d6 c5d6');
INSERT INTO evaluations VALUES(55069,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(55069,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(55069,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55069,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(55069,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55069,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55069,36,'cp 0','g8f6 g1f3 e7e6 e2e3 b8d7 c2c4 f8e7 b1d2 b7b6 f1e2 c8b7 e1g1 d5c4 d2c4 e8g8 b2b3 c7c5 c1b2 a8c8 a1c1 h7h6 a3a4 f6e4 d4c5 e7c5 f3d2 e4d2 c4d2 d7f6 e2f3 b7f3 d2f3 d8d1 f1d1');
INSERT INTO evaluations VALUES(47214,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47214,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47214,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47214,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47214,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(47214,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(47214,36,'cp 0','g1f3 e7e6 e2e3 b8d7 c2c4 f8e7 b1d2 b7b6 f1e2 c8b7 e1g1 e8g8 b2b3 a8c8 c4d5 f6d5 c1b2 c7c5 a1c1 e7f6 d2e4 f6e7 e4d2');
INSERT INTO evaluations VALUES(47212,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(47212,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(47212,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(47212,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(47212,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(47212,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(47212,36,'cp 0','e7e6 e2e3 f8e7 b2b3 b7b6 c1b2 b8d7 b1d2 c8b7 f1e2 c7c5 c2c4 d5c4 d2c4 a8c8 a1c1 e8g8 e1g1 h7h6 d4c5 c8c5 b3b4 c5c7 c4e5 c7c1 d1c1 d8a8 e5d7 f6d7 b2d4 e7f6 d4f6 d7f6 c1b2 f8c8 f1c1 b7d5 c1c8 a8c8 f3d4 f6e4 e2f3 c8c3');
INSERT INTO evaluations VALUES(48370,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48370,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(48370,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48370,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(48370,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48370,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48370,36,'cp 0','c1f4 c7c5 e2e3 c5d4 e3d4 d8b6 b1c3 c8d7 a1b1 b8c6 h2h3 c6e7 f1d3 e7g6 f4e5 d7c6 e5h2 f8e7 e1g1 e8g8 f1e1 a7a6 d1d2 f8e8 h2g3 e8f8 d2e3 f8e8');
INSERT INTO evaluations VALUES(48298,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48298,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(48298,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48298,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48298,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48298,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(48298,36,'cp 10','c7c5 e2e3 b8c6 h2h3 b7b6 f1b5 c8b7 c2c3 a8c8 d1a4 f6d7 b1d2 a7a6 b5e2 f8e7 e1g1 e8g8 b2b4 b6b5 a4d1 c5c4 a3a4 c8a8 a4b5 a6b5 d1c2 d7f6 e3e4 d5e4 d2e4 f6d5 f4g3 a8a1 f1a1 d8b6 e4c5 f8a8 a1a8 b7a8 f3d2 h7h6 c5d7');
INSERT INTO evaluations VALUES(46216,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(46216,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46216,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46216,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(46216,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46216,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(46216,36,'cp 10','e2e3 b8c6 h2h3 b7b6 c2c3 c5c4 b1d2 f8d6 f4d6 d8d6 d1c2 c8b7 f1e2 a8b8 e1g1 e8g8 e3e4 f6e4 d2e4 d5e4 c2e4 c6e7 e4e5 e7f5 e5d6 f5d6 f1e1 f8c8 f3d2 g7g6 e2f3 h7h5 f3b7 b8b7 e1e5 g8g7 d2f3 b7c7');
INSERT INTO evaluations VALUES(46215,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46215,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46215,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46215,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46215,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46215,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(46215,36,'cp 0','b8c6 h2h3 b7b6 c2c3 c5c4 b2b3 c4b3 d1b3 c8b7 f3e5 f8d6 e5c6 b7c6 f4d6 d8d6 a3a4 e8g8 b1d2 f8c8 f1a6 c8c7 e1g1 a8b8 f1c1 d6d8 b3b2 h7h6 a6d3 b8c8 d3a6');
INSERT INTO evaluations VALUES(13577,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13577,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13577,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13577,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13577,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13577,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13577,36,'cp 0','h2h3 a7a6 f1e2 b7b5 d4c5 f8c5 c2c3 e8g8 e1g1 f8e8 b2b4 c5f8 a3a4 b5a4 d1a4 c8d7 a4b3 d8b6 f4h2 c6e7 f3d4 e7c6');
INSERT INTO evaluations VALUES(16726,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(16726,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(16726,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(16726,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(16726,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(16726,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(16726,36,'cp -7','g1f3 f8b4 c2c3 b4c5 c1g5 e8g8 b1d2 d7d6 g5h4 d8e7 a4c2 c6e5 f3e5 d6e5 f1e2 a7a5 e1g1 h7h6 d2b3 c5d6 b3d2 g7g5 h4g3 h6h5 h2h4 f6g4 h4g5 e7g5 d2f3 g5h6 a1e1 a5a4 f3h4 c8d7 h4f5 d7f5 e4f5 h5h4');
INSERT INTO evaluations VALUES(16725,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(16725,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16725,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(16725,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(16725,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(16725,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(16725,36,'cp -30','f8b4 c2c3 b4c5 f1d3 d7d6 e1g1 a7a5 b1d2 e8g8 h2h3 c6e5 f3e5 d6e5 a4c2 d8e7 d2b3 c5b6 b3d2 a5a4 d2c4 b6c5 c1e3 c8e6 e3c5 e7c5 c4e3 f8d8 b2b4 a4b3 a2b3 g7g6 b3b4 c5e7 d3c4 e6c4 e3c4 g8g7 a1a8 d8a8 c4e3 a8d8 f1d1 d8d1 e3d1');
INSERT INTO evaluations VALUES(56535,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(56535,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(56535,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(56535,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(56535,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(56535,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(56535,36,'cp 251','d1d8 e8d8 b1c3 b8c6 c1f4 d8e8 e1c1 c8g4 f1e2 h7h5 c3e4 g8e7 h2h3 g4f5 e4g3 f5g6 e2c4 h5h4 g3e2 c6a5 c4b5 c7c6 b5d3 g6d3 d1d3 e7d5 g1f3 f8c5 b2b3 c5f2 f4g5 h8h5 c1b2 d5b4 d3d6 f2c5 g2g4 h4g3 e2g3 h5h8');
INSERT INTO evaluations VALUES(35405,30,'cp 232',NULL);
INSERT INTO evaluations VALUES(35405,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(35405,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(35405,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(35405,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(35405,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(35405,36,'cp 248','e8d8 b1c3 b8c6 c1f4 d8e8 e1c1 c8g4 f1e2 g4f5 g1h3 f8b4 c3b5 a8c8 g2g4 f5e6 h3g5 g8e7 g5e4 a7a6 b5c3 c8d8 d1d8 e8d8 f4g3 d8c8 h1d1 g7g6 g3h4 h8d8 d1d8 c8d8 f2f4 d8c8 a2a3 b4c3 e4c3 e7d5 c3d5 e6d5 c2c3 c6a5 f4f5 g6f5 g4f5 a5b3 c1d1');
INSERT INTO evaluations VALUES(35452,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(35452,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(35452,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(35452,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(35452,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(35452,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(35452,36,'cp 249','b1c3 b8c6 c1f4 d8e8 e1c1 c8f5 g1e2 f8b4 a2a3 b4a5 b2b4 a5b6 e2g3 f5g4 f1e2 h7h5 g3e4 g8e7 f2f3 g4f5 e2b5 e7g6 f4g5 g6e5 h1e1 e8f8 c3a4 f5e4 e1e4 f7f6 f3f4 e5f7 g5h4 a8d8 d1d8 f7d8 e4e2 f8f7 a4b6 a7b6');
INSERT INTO evaluations VALUES(35451,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(35451,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(35451,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(35451,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(35451,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(35451,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(35451,36,'cp 212','b8c6 c1f4 d8e8 e1c1 c8f5 g1e2 a8d8 d1d8 e8d8 e2g3 f5e6 g3e4 f8b4 c3a4 d8c8 a2a3 b4e7 a4c5 e6f5 f1c4 e7c5 e4c5 b7b6 c5d3 f5e6 c4a6 c8b8 f4d2 g8e7 h1e1 e7d5 b2b4 h7h6 h2h4 h8d8 c2c3 c6e7 d3b2 c7c6 a6d3 b6b5 h4h5 d5b6 d2e3 b8c7 g2g3 e7f5 e3c5 b6a4 d3f5 a4c5 f5e6 c5e6 e1d1 d8d1 b2d1');
INSERT INTO evaluations VALUES(2313,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(2313,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(2313,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(2313,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(2313,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(2313,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(2313,36,'cp 227','c1f4 d8e8 e1c1 c8f5 g1e2 a8d8 d1d8 e8d8 e2g3 f5e6 g3e4 f8b4 c3a4 d8c8 a2a3 b4e7 a4c5 e6f5 f1c4 e7c5 e4c5 b7b6 c5d3 f5e6 c4e6 f7e6 h2h4 g8e7 h4h5 h7h6 c2c3 h8f8 c1c2 f8f5 g2g4 f5f8 h1h3 c6a5 f4e3 e7d5 f2f4 c7c5 e3d2 a5c4 d2c1');
INSERT INTO evaluations VALUES(40544,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40544,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40544,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40544,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(40544,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(40544,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(40544,36,'cp 30','f1d3 d5c4 d3c4 b7b5 c4d3 b8d7 e1g1 e8g8 d1c2 c8b7 e3e4 e6e5 h2h3 e5d4 f3d4 d7c5 f1d1 d8c7 c1g5 d6h2 g1h1 h2e5 d4f3 a8d8 g5f6 g7f6 f3e5 c7e5 b2b4 c5e6 c3e2 c6c5 b4c5 e5c5 e2g3 b5b4 g3f5 c5c2 d3c2 d8d1 a1d1 f8d8 d1d8 e6d8 f2f3 a7a5 h1g1 b7a6');
INSERT INTO evaluations VALUES(40547,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(40547,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40547,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40547,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(40547,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(40547,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40547,36,'cp 42','d5c4 d3c4 b7b5 c4b3 c8a6 e1g1 b5b4 c3e2 e8g8 f1e1 c6c5 e2g3 b8c6 e3e4 d6g3 h2g3 c6d4 e4e5 d4f3 d1f3 f6d5 b3c2 f7f5 e5f6 d8f6 f3f6 d5f6 e1e6 a6c4 e6e5 f8e8 e5e8 a8e8 c1e3 a7a5 b2b3 c4e6 c2d3 e8c8 a1c1 a5a4 c1c5 c8c5 e3c5 a4b3 a2b3 e6b3 c5b4 g8f7 f2f3');
INSERT INTO evaluations VALUES(40561,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(40561,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(40561,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(40561,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(40561,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(40561,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(40561,36,'cp 29','d3c4 b7b5 c4b3 c8a6 e1g1 b5b4 c3e2 e8g8 f1e1 c6c5 e2g3 d6g3 h2g3 a6b7 e3e4 f6e4 d1c2 c5c4 c2c4 a7a5 c1f4 e4f6 f3e5 b7d5 c4d3 d5b3 a2b3 f6d5 e1c1 h7h6 c1c4 d5b6 c4c5 b6d7 c5c1 d7f6 d3f3 f6d5 c1c5 b8d7 e5d7 d8d7 a1a5');
INSERT INTO evaluations VALUES(40560,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(40560,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(40560,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(40560,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(40560,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(40560,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40560,36,'cp 17','b7b5 c4d3 b8d7 e1g1 e8g8 d1c2 c8b7 e3e4 e6e5 h2h3 e5d4 f3d4 d7c5 f1d1 d8c7 c1g5 d6h2 g1h1 h2e5 d4f3 a8d8 f3e5 c7e5 f2f4 e5e6 e4e5 d8d3 d1d3 c5d3 c2d3 f6d5 a1e1 b7c8 c3d5 e6d5 d3d5 c6d5 g5e7 f8e8 e7c5 a7a5 e1c1 c8e6 c5b6 a5a4 a2a3 e8c8 c1c8 e6c8 h1g1 h7h5 b6e3 f7f6 e5f6 g7f6');
INSERT INTO evaluations VALUES(56484,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(56484,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(56484,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(56484,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(56484,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(56484,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(56484,36,'cp -42','e5e4 f3d4 c7c5 d4b3 c5c4 b3d4 f8c5 c2c3 g8f6 d2d3 c4d3 e2d3 b8c6 d3e4 d5e4 c1e3 c5d4 e3d4 c8g4 d1d2 e8g8 e1g1 f8e8 f1e1 h7h6 h2h3 g4f5 b1a3 d8e7 a3c2 a8d8 a1d1 f5g6 c2e3 c6d4 c3d4 f6d5');
INSERT INTO evaluations VALUES(56260,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(56260,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(56260,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(56260,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(56260,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(56260,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(56260,36,'cp -50','f3d4 c7c5 d4b3 c5c4 b3d4 f8c5 c2c3 g8f6 d2d3 c4d3 e2d3 c8g4 d1a4 g4d7 a4d1 b8c6 d3e4 d5e4 c1e3 d7g4 d1a4 c5d4 c3d4 e8g8 b1c3 f8e8 e1g1 a7a6 d4d5 f6d5 c3e4 d5e3 f2e3 c6e5 a4d4 d8e7 d4d6 e5c4 d6e7 e8e7 a1e1 c4b2 e4d6 b2d3 e1b1 e7d7 d6b7 a8c8 g1h1');
INSERT INTO evaluations VALUES(56197,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(56197,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(56197,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(56197,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(56197,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(56197,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(56197,36,'cp -37','c7c5 d4b3 c5c4 b3d4 f8c5 c2c3 g8f6 d2d3 c4d3 e2d3 c8g4 d1a4 g4d7 a4d1 c5d4 c3d4 e8g8 d3e4 d5e4 b1c3 d7g4 d1a4 b8c6 c1e3 f8e8 e1g1 a7a6 d4d5 f6d5 c3e4 d5e3 f2e3 c6e5 a4d4 d8e7 d4d6 h7h6 d6e7 e8e7 h2h3 g4e6 e4c5 e5c4 g2b7 a8b8 b7g2 b8b2 c5e6 e7e6 g2d5');
INSERT INTO evaluations VALUES(54509,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(54509,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(54509,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(54509,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(54509,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(54509,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(54509,36,'cp -39','d4b3 c5c4 b3d4 f8c5 c2c3 g8f6 d2d3 c4d3 e2d3 c8g4 d1a4 g4d7 a4b3 b8c6 d3e4 e8g8 e1g1 d5e4 c1e3 c5b6 b1d2 c6d4 e3d4 b6d4 c3d4 d7c6 a1d1 d8e8 b3a3 a8d8 a3a7 e4e3 g2c6 e3f2 f1f2 e8c6 a7c5 c6a6 c5a3 a6a3 b2a3 d8d4 d1e1 h7h6 f2e2');
INSERT INTO evaluations VALUES(54533,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(54533,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(54533,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(54533,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(54533,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(54533,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(54533,36,'cp -50','c5c4 b3d4 f8c5 c2c3 g8f6 d2d3 c4d3 e2d3 c8g4 d1a4 g4d7 a4d1 b8c6 d3e4 d5e4 c1e3 c5d4 c3d4 d7g4 d1a4 e8g8 b1c3 f8e8 e1g1 a7a6 d4d5 f6d5 c3e4 d5e3 f2e3 c6e5 a4d4 d8e7 d4d6 e5c4 d6f4 c4e5 e4c3 a8c8 c3d5 e7d6 f4d4 e8d8 a1c1 c8c1 f1c1 g4f3');
INSERT INTO evaluations VALUES(17453,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(17453,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(17453,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(17453,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(17453,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(17453,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(17453,36,'cp 73','b5d6 f8d6 d1d6 d8e7 d6d1 g8f6 b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 f1e2 b8d7 e1g1 e8g8 c2c4 h7h6 g5e3 a6a5 b2b3 f6e4 d1c2 f7f5 f2f3 e4c5 a2a3 f5f4 e3c5 d7c5 b3b4 c5d7 e2d3 b7b6 f1e1 a5b4 a3b4 a8a1 e1a1');
INSERT INTO evaluations VALUES(17464,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(17464,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(17464,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(17464,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(17464,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(17464,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(17464,36,'cp 70','f8d6 d1d6 d8e7 d6d1 g8f6 b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 f1e2 e8g8 e1g1 h7h6 g5e3 e5e4 c2c4 b8d7 d1d2 d7e5 b2b4 f8c8 f1c1 a6a5 b4a5 f6d7 a1b1 d7c5 e3d4 f7f5 f2f4 e4f3 g2f3 e7h4 d4e5 d6e5 g1h1 c8e8 d5d6');
INSERT INTO evaluations VALUES(4733,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(4733,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(4733,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(4733,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(4733,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(4733,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(4733,36,'cp 74','d1d6 d8e7 d6d1 g8f6 b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 f1e2 e8g8 c2c4 a6a5 e1g1 b8a6 d1c2 f8c8 c2f5 a6c5 b2b3 h7h6 g5e3 e7d7 f5c2 f6g4 e2g4 d7g4 f2f4 e5e4 e3c5 c8c5 c2e4 b7b5 c4b5 c5b5 a1c1 g4d7');
INSERT INTO evaluations VALUES(4730,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(4730,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(4730,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(4730,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(4730,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(4730,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(4730,36,'cp 62','d8e7 d6d1 g8f6 b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 c2c4 b8d7 f1e2 e8g8 e1g1 f8b8 a2a4 a6a5 e2d3 e5e4 f1e1 e7e5 d1d2 d7c5 a1a3 b8e8 d3b1 f6d7 f2f3 h7h6 g5f4 e5f6 f3e4 d7e5 b2b3 e8e7 f4g3 f6g6 b1c2 b7b6 d2d4 a8e8 a3a1');
INSERT INTO evaluations VALUES(82,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(82,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(82,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(82,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(82,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(82,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(82,36,'cp 65','d6d1 g8f6 b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 f1e2 e8g8 c2c4 b8d7 e1g1 h7h6 g5e3 f6h7 f2f3 f7f5 d1c2 a8c8 a1c1 h7f6 c2d2 a6a5 c1e1 b7b6 e2d3 e5e4 e3d4 d7e5 f3e4 f5e4 d3e4 e5c4 d2d3 f6e4 f1f8 c8f8 e1e4 c4e5 d4e5');
INSERT INTO evaluations VALUES(81,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(81,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(81,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(81,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(81,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(81,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(81,36,'cp 54','g8f6 b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 c2c4 b8d7 f1e2 h7h6 g5e3 e8g8 e1g1 f6h7 f2f3 f7f5 d1c2 a8c8 a1c1 h7f6 c2f5 f6d5 f5d3 d5f4 e3f4 f8f4 d3d5 g8h8 c1d1 d7b6 d5d6 e7d6 d1d6 b6c4 d6e6 c4b2 f1b1 b2a4 e6e5 a4c3 b1b7 c3e2 e5e2');
INSERT INTO evaluations VALUES(428,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(428,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(428,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(428,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(428,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(428,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(428,36,'cp 67','b1c3 d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 f1e2 e8g8 c2c4 h7h6 g5e3 f6h7 e1g1 b8d7 f2f3 f7f5 a1c1 b7b6 b2b4 h7f6 d1d2 g8h8 e2d3 e5e4 d3b1 a8c8 e3d4 d7e5 f3e4 f6e4 d2e2 b6b5 c4b5 c8c1 f1c1 e7g5 c1c7 a6b5');
INSERT INTO evaluations VALUES(427,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(427,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(427,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(427,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(427,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(427,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(427,36,'cp 59','d7d6 c1g5 c8e6 c3d5 e6d5 e4d5 c6b8 c2c4 b8d7 f1e2 e8g8 e1g1 h7h6 g5e3 f6h7 d1c2 f7f5 a1e1 f5f4 e3d2 a8c8 e2g4 b7b5 b2b3 h7f6 g4e6 g8h8 f2f3 e7d8 e1c1 b5c4 b3c4 c8b8 c1b1 d8c7 g1h1 d7c5 e6f5 f8e8 f5g6 e8c8 d2c3 a6a5 f1c1 b8b1 c1b1');
INSERT INTO evaluations VALUES(29001,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(29001,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29001,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(29001,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29001,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29001,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29001,36,'cp 10','f6e4 g5f4 d5c4 f3d2 e4d2 d1d2 f5g6 h2h4 h7h5 e2e3 b7b5 a2a4 d8a5 f1e2 b5b4 c3a2 b8d7 e2c4 e7e5 f4g3 f8e7 e1g1 a5a4 b2b3 a4a5 d4e5 d7b6 e5e6 e8g8 e6f7 g6f7 c4f7 f8f7 g3d6 a5d5 d6b4 e7h4 a2c1 d5d2 b4d2 h4f6 a1a6 c6c5 g1h1');
INSERT INTO evaluations VALUES(38234,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(38234,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(38234,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(38234,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(38234,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38234,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(38234,36,'cp 54','d2d4 e5d4 f3d4 g8f6 c1f4 b8c6 d1d2 c6d4 d2d4 c8e6 e1c1 e8g8 c1b1 a7a6 f1c4 f6d7 c3d5 e7f6 d5f6 d8f6 d4f6 d7f6 c4d3 f6g4 f4g3 a8e8 h1e1 g4e5 d3f1 g7g5 h2h4 g5g4 e1e3 g8g7 b1c1 f7f6 b2b3 h7h5 f2f4 e5c6 f1e2 f6f5 g3f2 f5e4 e3e4');
INSERT INTO evaluations VALUES(38229,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38229,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(38229,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(38229,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(38229,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(38229,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(38229,36,'cp 54','b8d7 h2h3 g8f6 f1c4 e8g8 e1g1 e5d4 d1d4 d7c5 f1d1 c8e6 c4f1 a7a5 c1f4 f8e8 d4e3 e6d7 e4e5 e7f8 e3c1 f6e4 c3e4 c5e4 c2c3 d7c6 f3d4 d8d7 c1c2 d6d5 f1d3 h7h6 d4c6 d7c6');
INSERT INTO evaluations VALUES(56243,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56243,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56243,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(56243,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(56243,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56243,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56243,36,'cp 0','e4d5 d8d5 c3d4 b8c6 g1f3 c8g4 f1e2 f8b4 b1c3 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8d7 e3d2 a8d8 a1c1 h8e8 h1d1 d7c8 d2e3 e7d5 c3a4 e8e6 g2g3 h7h5 a2a3 b4e7 a4c5 e7c5 c1c5 c8d7 c5a5 d8b8 d1d2 h5h4 g3h4 b8e8 a5a7');
INSERT INTO evaluations VALUES(56137,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56137,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56137,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56137,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56137,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56137,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56137,36,'cp 0','d8d5 c3d4 b8c6 g1f3 c8g4 f1e2 g4f3 e2f3 f8b4 b1c3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e7d5 e2d3 a7a5 c3a4 e8d7 g2g3 h8b8 h1c1 b4d6 b2b3 d6a3 c1e1 d5b4 d3d2 b4d5');
INSERT INTO evaluations VALUES(34904,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(34904,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(34904,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34904,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34904,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(34904,36,'cp 2','c3d4 f8b4 b1c3 b8c6 g1f3 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8d7 h1c1 a8b8 e2d3 h8e8 a1b1 e7d5 e3d2 h7h5 h2h4 b4a5 c3d5 a5d2 d3d2 c6d5 c1c5 c7c6 c5a5 b8b7 g2g3 f7f6 d2d3 e8e6 b2b3 e6e4 a5a6');
INSERT INTO evaluations VALUES(34898,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34898,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34898,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34898,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34898,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34898,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34898,36,'cp 0','b8c6 g1f3 f8b4 b1c3 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 c3a4 e7f5 h1d1 b4e7 a1c1 h8e8 c1c6 e7f6 e2f3 f5d4 e3d4 d8d4 d1d4 f6d4 c6c4 e8d8 g2g3 d8d6 f3e2 d6e6 e2f1 e6d6');
INSERT INTO evaluations VALUES(1703,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1703,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1703,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1703,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1703,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1703,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1703,36,'cp 0','g1f3 c8g4 f1e2 g4f3 e2f3 f8b4 b1c3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 c3a4 e7f5 h1d1 h8e8 a1c1 b4e7 c1c6 e7f6 e2f3 f5d4 e3d4 d8d4 d1d4 f6d4 c6c4 e8d8 f3e2 d8e8');
INSERT INTO evaluations VALUES(1702,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1702,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1702,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1702,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1702,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1702,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1702,36,'cp 0','c8g4 f1e2 g4f3 e2f3 f8b4 b1c3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 e8c8 c1e3 g8e7 c3a4 e7f5 h1d1 h8e8 a1c1 b4e7 c1c6 e7f6 e2f3 f5h4 f3g3 f6d4 g3h4 d4e3 d1d8 c8d8 f2e3 e8e4 h4g3 e4a4 a2a3 d8d7 c6c5 g7g6 c5d5 d7e6 d5c5');
INSERT INTO evaluations VALUES(27644,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27644,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27644,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27644,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27644,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27644,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27644,36,'cp 0','f1e2 f8b4 b1c3 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 c3a4 e7f5 h1d1 h8e8 a1c1 b4e7 c1c6 e7f6 e2f3 f5h4 f3g3 f6d4 g3h4 d4e3 d1d8 c8d8 f2e3 e8e4 h4g3 e4a4 a2a3 a4a5 c6c4 d8d7 g3f3 h7h5 c4h4 a5d5 h4c4 d5a5');
INSERT INTO evaluations VALUES(27643,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27643,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27643,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27643,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27643,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27643,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27643,36,'cp 0','g4f3 e2f3 f8b4 b1c3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 a1c1 e7f5 h1d1 h8e8 c3a4 b4e7 c1c6 e7f6 e2f3 f5h4 f3g3 f6d4 g3h4 d4e3 d1d8 c8d8 f2e3 e8e4 h4g3 e4a4 a2a3 a4a5 c6c4 d8d7 g3f3 h7h5 c4d4 d7e7 d4c4');
INSERT INTO evaluations VALUES(23474,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(23474,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(23474,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(23474,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(23474,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(23474,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(23474,36,'cp 31','d2d4 d7d5 g1f3 c8f5 c2c4 e7e6 f1g2 f8e7 e1g1 g8f6 b1c3 e8g8 c4d5 e6d5 f3e5 e7d6 e5c6 b7c6 c1g5 h7h6 g5f6 d8f6 d1a4 d6e7 e2e4 d5e4 g2e4 f5e4 c3e4 f6g6 a1c1 g6e4 f1e1 e4g4 e1e7 a8d8 a4b3 a7a5 e7c7 d8d4');
INSERT INTO evaluations VALUES(23439,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(23439,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(23439,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(23439,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(23439,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(23439,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(23439,36,'cp 25','d7d5 g1f3 c8f5 c2c4 e7e6 f1g2 g8f6 e1g1 f8e7 f3e5 e8g8 c4d5 e6d5 b1c3 e7d6 e5c6 b7c6 c1g5 h7h6 g5f6 d8f6 d1a4 d6e7 e2e4 d5e4 c3e4 f5e4 g2e4 a8d8 a1d1 c6c5 a4a7 d8d4 d1d4 f6d4 e4f3 d4b2 a7c7 e7f6 f3d5');
INSERT INTO evaluations VALUES(46314,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(46314,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(46314,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(46314,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46314,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46314,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(46314,36,'cp 18','e3d4 f8e7 c4d5 f6d5 f1d3 e8g8 e1g1 b8c6 f1e1 e7f6 d3e4 c6e7 f3e5 c8d7 d1d3 h7h6 c3d5 e6d5 e5d7 d8d7 e4f3 d7f5 e1d1 f5d3 d1d3 a8c8 g1f1 f8d8 c1d2 c8c4 d2a5 b7b6 a5c3 g7g6 a2a3 g8g7 g2g3 e7f5 a1d1 h6h5 h2h3');
INSERT INTO evaluations VALUES(18034,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(18034,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(18034,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(18034,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(18034,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(18034,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(18034,36,'cp 90','d2d4 e5d4 e1g1 f8c5 c2c3 d4d3 d1d3 g8e7 a2a4 b5a4 b3a4 e8g8 b2b4 c5a7 b1d2 d7d6 d2c4 c8b7 c1e3 a7e3 c4e3 e7g6 d3c2 d8f6 a4c6 b7c6 f3d4 c6d7 d4e2 f8b8 c2d2 g6f4 e2f4 f6f4 f2f3 c7c5 g1h1 c5b4 c3b4 d7e6 f1d1');
INSERT INTO evaluations VALUES(18032,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(18032,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(18032,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(18032,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(18032,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(18032,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(18032,36,'cp 82','e5d4 e1g1 f8c5 c2c3 d4d3 d1d3 g8e7 a2a4 b5a4 b3a4 e8g8 b2b4 c5a7 b1a3 d7d6 a3c4 c8b7 c1e3 a7e3 c4e3 e7g6 d3c2 c6e5 f3e5 d6e5 a4b3 d8h4 b3d5 b7d5 e3d5 g6f4 g2g3 h4h3 d5e3 f4e6 f2f3 e6g5 c2e2 f8e8 g1h1 c7c6 e3f5');
INSERT INTO evaluations VALUES(18037,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(18037,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(18037,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(18037,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(18037,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(18037,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(18037,36,'cp 102','e1g1 f8c5 c2c3 d4d3 d1d3 g8e7 a2a4 b5a4 b3a4 e8g8 b2b4 c5a7 c1f4 d7d6 f4g3 a7b6 f1d1 c8b7 b1d2 e7g6 d3d5 c6e7 d5h5 a6a5 b4b5 f7f5 e4f5 f8f5 h5g4 f5f6 a4c2 d8c8 g4c8 a8c8 d2c4 b7f3 g2f3 d6d5 c4a5 f6f3');
INSERT INTO evaluations VALUES(20729,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(20729,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(20729,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(20729,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(20729,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(20729,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(20729,36,'cp 124','b1c3 g8f6 f1d3 f8e7 h2h3 e8g8 e1g1 d8b6 d3c2 c8d7 a2a3 a8c8 a1b1 a7a6 f1e1 f8d8 c1f4 c6a5 c2d3 h7h6 d1e2 a5b3 f4e3 d6d5 e4e5 f6e4 d3c2 d7c6 e2d3 b3a5 f3d2 b6c7 b1c1 a5c4 c3e4 c4e3 d3e3 d5e4 c2e4');
INSERT INTO evaluations VALUES(20727,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(20727,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(20727,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(20727,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(20727,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(20727,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(20727,36,'cp 121','g8f6 f1d3 f8e7 e1g1 e8g8 a2a3 d8b6 d3b1 c8d7 b2b4 e6e5 c1e3 e5d4 f3d4 b6d8 h2h3 c6d4 d1d4 a7a5 b1d3 h7h6 f1d1 a5b4 a3b4 a8a1 d1a1 d8c7 c3e2 f8c8 e2f4 c7b8 b4b5 e7d8 a1b1 b8c7 b1c1 c7b8 c1c8 d7c8 b5b6');
INSERT INTO evaluations VALUES(14001,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(14001,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(14001,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(14001,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(14001,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(14001,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(14001,36,'cp 113','f1d3 f8e7 e1g1 e8g8 a2a3 c8d7 c1f4 a8c8 a1c1 c6d4 f3d4 e6e5 f4e3 e5d4 e3d4 a7a6 h2h3 f8e8 d3b1 d7e6 f2f4 f6d7 g1h2 d7c5 b1c2 e6c4 f1f2 e7f6 d4f6 d8f6 d1f3 c5e6 c1d1 e6d4 f3g3 b7b5 f2d2 d4c2 d2c2');
INSERT INTO evaluations VALUES(14002,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(14002,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(14002,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(14002,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(14002,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(14002,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(14002,36,'cp 135','f8e7 e1g1 a7a6 h2h3 e8g8 d1e2 c8d7 c1f4 a8c8 a1d1 d6d5 e4e5 f6h5 f4c1 f7f5 e5f6 h5f6 d3b1 e7d6 f1e1 d6b8 f3e5 d7e8 c1g5 d8c7 f2f4 c6e5 d4e5 f6h5 e2g4 c7b6 g1h1 e8g6 f4f5 e6f5 c3d5 f5g4 d5b6');
INSERT INTO evaluations VALUES(8210,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(8210,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(8210,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(8210,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(8210,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(8210,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(8210,36,'cp 132','e1g1 c8d7 f1e1 e8g8 h2h3 d8b6 e4e5 d6e5 d4e5 f6d5 c3d5 e6d5 c1e3 b6c7 a1c1 d7e6 a2a3 c7d7 d1d2 a8c8 c1d1 a7a6 d3b1 e6f5 b1f5 d7f5 e3b6 d5d4 b2b4 h7h6 f3d4 c6d4 d2d4 f5e6 d4d7 e6b6 d7e7 f8e8');
INSERT INTO evaluations VALUES(8209,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(8209,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(8209,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(8209,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(8209,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(8209,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(8209,36,'cp 125','e8g8 a2a3 a7a6 d1c2 h7h6 h2h3 c8d7 e4e5 f6d5 c2e2 d6e5 d4e5 e7c5 c3d5 e6d5 d3b1 f7f5 b1a2 d7e6 c1e3 c5e3 e2e3 d8d7 a1d1 a8e8 e3b6 e6f7 f1e1 d5d4 e5e6 f7e6 e1e6 e8e6 a2e6 d7e6 b6b7 e6c4 b2b3 c4c2 d1d2 c2c1 g1h2 g8h7 b7a6');
INSERT INTO evaluations VALUES(3136,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(3136,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(3136,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(3136,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(3136,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(3136,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(3136,36,'cp 138','a2a3 c8d7 d1c2 h7h6 e4e5 f6d5 c2e2 d8b6 f1d1 d6e5 d4e5 a8d8 a1b1 g8h8 h2h3 e7c5 c3a4 c6d4 e2e4 f7f5 a4b6 f5e4 b6d7 d4f3 g2f3 d8d7 d3e4 d5f6 d1d7 f6d7 b2b4 c5d4 e4b7 d7e5 b1b3 e5c4 b3d3 d4e5');
INSERT INTO evaluations VALUES(16796,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(16796,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(16796,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(16796,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(16796,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(16796,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(16796,36,'cp 71','g1f3 f8b4 b1c3 b4c3 b2c3 d8e7 d1d5 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 c8e6 g2b7 a8b8 a1b1 b8f8 b1c1 h7h6 c4c5 f6d5 b7d5 e6d5 c5d6 c7d6 f4d6 f8f3 d6e5 d5a2 c1c7 f3f7 c7c8 f7f8 c8f8 g8f8 f2f3 a7a5');
INSERT INTO evaluations VALUES(26533,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(26533,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(26533,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(26533,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(26533,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(26533,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(26533,36,'cp 25','e7e6 b1d2 f8d6 f4g3 g8f6 e2e3 e8g8 f1e2 f8e8 e1g1 d6g3 h2g3 e6e5 d4e5 c6e5 d1b3 e5f3 e2f3 b7b6 c3c4 d5c4 d2c4 f6e4 a1d1 d8f6 d1d4 a8d8 f1d1 e4g5 b3a4 g5f3 g2f3 f5e6 g1g2 c7c5 d4d8 e8d8');
INSERT INTO evaluations VALUES(26810,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(26810,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26810,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26810,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(26810,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(26810,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26810,36,'cp 11','e2e3 f8d6 f4g3 g8f6 f1e2 e8g8 b1d2 f8e8 f3h4 c6e7 e1g1 c7c5 d4c5 d6c5 a1c1 a8c8 c3c4 h7h6 a2a3 d5d4 d2b3 d4e3 b3c5 c8c5 f2e3 c5c8 d1d8 e8d8 f1d1 d8d1 c1d1 f5c2 d1c1 c2a4 g3d6 e7c6 h4f3 f6e4 d6b4 c6b4 a3b4');
INSERT INTO evaluations VALUES(26808,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(26808,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(26808,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26808,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(26808,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(26808,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(26808,36,'cp 7','f8d6 f4g3 g8f6 d1b3 c6e7 f1e2 e8g8 e1g1 a7a5 g3d6 d8d6 f1c1 a5a4 b3b4 f6e4 b4d6 e4d6 c3c4 d6c4 e2c4 d5c4 c1c4 c7c6 a2a3 f5d3 c4c3 d3e4 b1d2 e4f3 d2f3 e7d5 c3c1 f8d8 h2h4 f7f6 f3d2');
INSERT INTO evaluations VALUES(23850,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(23850,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(23850,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(23850,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(23850,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(23850,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(23850,36,'cp 18','f4g3 g8f6 b1d2 e8g8 f1e2 f8e8 e1g1 d6g3 h2g3 e6e5 d4e5 c6e5 d1b3 e5f3 e2f3 a8b8 f1d1 d8e7 c3c4 d5c4 d2c4 h7h5 a1c1 f5e6 b3c3 e6c4 c3c4 c7c6 b2b4 a7a6 a2a4 g7g6 b4b5 a6b5 a4b5 c6b5 c4b5 e8d8 d1d8 b8d8 b5b7 e7b7 f3b7 d8d2');
INSERT INTO evaluations VALUES(23852,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(23852,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(23852,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(23852,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23852,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(23852,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(23852,36,'cp 4','g8f6 f1e2 e8g8 b1d2 f8e8 f3h4 c6e7 e1g1 e7g6 h4f5 e6f5 g3d6 d8d6 g2g3 g6e7 c3c4 d6e6 d1b3 c7c6 d2f3 a8b8 a1c1 g7g6 f3e5 a7a6 c1c2 g8g7 f1c1 d5c4 e2c4 e7d5 e5d3 e6d6 a2a4 a6a5 c1d1 g6g5 d1c1');
INSERT INTO evaluations VALUES(24111,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24111,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(24111,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(24111,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(24111,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(24111,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(24111,36,'cp 18','a2a3 c6e7 c3c4 e8g8 b1d2 b7b6 f1e2 d6g3 h2g3 c7c5 a1c1 d8d6 e1g1 h7h5 d1a4 a7a5 f1d1 f8c8 d2b1 f5b1 c1b1 g7g6 c4d5 e6d5 b2b3 g8g7 b1c1 c8b8 d1e1 b8c8 d4c5 b6c5');
INSERT INTO evaluations VALUES(35319,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(35319,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(35319,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(35319,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(35319,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(35319,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(35319,36,'cp 109','g1f3 e7e6 d2d4 g8f6 c1g5 f8e7 d1d2 e8g8 e1c1 f6d5 g5e7 d7e7 h2h4 e7b4 a2a3 b4a5 c3d5 a5d2 c1d2 e6d5 d1e1 c8d7 f1d3 f8e8 e1e8 d7e8 h1e1 b8c6 b2b4 a7a6 c2c3 e8d7 f3e5 c6e5 e1e5 d7e6 f2f4 c7c6 f4f5 e6d7 h4h5');
INSERT INTO evaluations VALUES(35317,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(35317,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(35317,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(35317,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(35317,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(35317,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(35317,36,'cp 118','g8f6 d2d4 e7e6 f3e5 d7d8 c1f4 f8b4 a2a3 b4c3 b2c3 b8d7 f1d3 f6d5 f4d2 d7e5 d4e5 d5e7 d1g4 e7g6 h2h4 h7h5 g4g3 e8g8 d3e4 d8e8 a3a4 c8d7 e4b7 a8b8 b7e4 d7c6 e4d3 e8d7 d2e3 a7a5 e3d4 d7e7 f2f3 f8d8 d3g6 f7g6');
INSERT INTO evaluations VALUES(34082,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(34082,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(34082,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(34082,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(34082,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(34082,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(34082,36,'cp 122','d2d4 e7e6 f3e5 d7d8 c1f4 f8b4 f1d3 e8g8 e1g1 b8c6 e5c6 b7c6 f4e5 f6d7 d1g4 g7g6 g4g3 d7e5 d4e5 b4c3 b2c3 c6c5 a1e1 c8b7 h2h4 d8e7 h4h5 g8h8 h5h6 h8g8 g3g4 b7d5 e1e3 a8d8 e3g3 c5c4 d3c4');
INSERT INTO evaluations VALUES(38849,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(38849,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(38849,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(38849,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(38849,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(38849,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(38849,36,'cp 130','c2c3 b4e7 f1d3 c7c5 g1f3 a7a6 e1g1 b8c6 f1e1 c8d7 d4c5 e7c5 b1d2 g8e7 d2b3 c5a7 c1e3 a7e3 e1e3 e7g6 d1d2 d8c7 a1e1 h7h6 h2h4 g6f4 d3f1 g7g5 g2g3 h8g8 h4h5 f4h5 f3d4 e8c8 f1h3');
INSERT INTO evaluations VALUES(38848,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(38848,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(38848,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(38848,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(38848,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(38848,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(38848,36,'cp 115','b4e7 f1d3 c7c5 g1f3 a7a6 e1g1 b8c6 f1e1 c8d7 h2h3 d8c7 a2a3 c5c4 d3c2 c6a5 b1d2 h7h6 a1b1 e7f8 d2f1 e8c8 c1f4 c8b8 f1e3 g8e7 f4g3 b8a8 h3h4 g7g6 g3h2 c7b6 e3g4 a5b3 h2f4 d7a4 g4f6 e7c6 h4h5 g6h5');
INSERT INTO evaluations VALUES(38297,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(38297,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(38297,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(38297,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(38297,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(38297,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(38297,36,'cp 134','d1g4 e7f8 f1d3 b7b6 h2h4 c8a6 d3a6 b8a6 h4h5 c7c5 g1e2 a6c7 b1d2 g8h6 g4h3 d8d7 a2a4 c5d4 c3d4 f8b4 e1f1 b4e7 h3d3 b6b5 a4a5 c7a6 d2f3 a6b4 d3b3 a8c8 c1h6 g7h6 h1h3 c8c4');
INSERT INTO evaluations VALUES(41210,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(41210,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41210,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(41210,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(41210,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(41210,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41210,36,'cp 37','h7h6 g5f4 e8g8 d1d2 e7e5 d4e5 d6e5 f4h6 f6e4 d2d8 f8d8 h6g7 e4c3 g7f6 d8d6 f6e5 d6e6 f2f4 f7f6 b2c3 f6e5 e1c1 b8c6 f4e5 e6e7 g1f3 c8g4 f1d3 g8g7 h1f1 c6e5 f3e5 e7e5 d1e1 a8e8 c1d2 b7b6 g2g3 g4f5 d3f5 g6f5 h2h4 e8d8 d2c2 e5e1 f1e1');
INSERT INTO evaluations VALUES(41169,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(41169,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(41169,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41169,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(41169,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41169,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(41169,36,'cp 38','g5f4 e8g8 d1d2 e7e5 d4e5 d6e5 f4h6 f6e4 d2d8 f8d8 h6g7 e4c3 g7f6 d8d6 f6e5 d6e6 f2f4 f7f6 b2c3 f6e5 e1c1 b8c6 f4e5 c6e5 g1f3 e5f3 g2f3 g8f7 d1d8 e6d6 d8d6 c7d6 h2h4 c8f5 h4h5 g6h5 h1h5 f7f6 c1d2 a8g8 d2e3 g8e8 e3d4 e8e1 f1h3 f5h3 h5h3 e1a1 h3h6 f6e7 h6h7 e7d8');
INSERT INTO evaluations VALUES(14627,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(14627,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(14627,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14627,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14627,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(14627,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(14627,36,'cp 31','a2a4 f8e7 f1e1 e8g8 c2c3 a7a6 b1d2 c6a5 c4a2 c7c5 d3d4 d8c7 b2b3 e5d4 c3d4 c5d4 c1b2 c8g4 a1c1 a5c6 h2h3 g4h5 b3b4 c7b6 g2g4 h5g6 f3d4 c6b4 d2c4 b6d8 a2b1 d6d5 e4d5 g6b1 d5d6 e7d6 c1b1 f8e8 e1e8 f6e8 b2c3 b4c6 b1b7');
INSERT INTO evaluations VALUES(24956,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(24956,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(24956,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(24956,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(24956,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(24956,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(24956,36,'cp 119','f3e5 g4h5 c2c4 e7e6 b1c3 f8e7 d1a4 e8g8 e5c6 b7c6 a4c6 a8b8 c4c5 d8d7 c6a6 f6e4 e1g1 f7f5 f2f3 e4c3 b2c3 e6e5 g2h3 e7f6 a6d3 h5g6 c1e3 e5d4 e3d4 d7e7 d3d2 c7c6 e2e3 f8e8 f1e1 h7h5 d2c2 f6d4 c3d4 h5h4 c2c3 h4g3 h2g3 e7g5 g1f2 g5h6 e1h1 e8e7');
INSERT INTO evaluations VALUES(24974,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(24974,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(24974,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(24974,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(24974,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(24974,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(24974,36,'cp 100','g4h5 c2c4 e7e6 b1c3 f8e7 d1a4 e8g8 e5c6 b7c6 a4c6 a8b8 c4d5 f6d5 c3d5 e6d5 c6d5 d8d5 g2d5 e7f6 f2f3 f6d4 g3g4 h5g6 e2e4 h7h5 h2h3 d4b2 c1b2 b8b2 e1g1 c7c6 d5b3 h5g4 h3g4 f8d8 a1d1 d8d2 d1d2 b2d2 f1c1 a7a5 c1c6 a5a4 c6c8 g8h7 b3c2 a4a3 c2b3');
INSERT INTO evaluations VALUES(44519,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(44519,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(44519,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(44519,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(44519,34,'cp -171',NULL);
INSERT INTO evaluations VALUES(44519,35,'cp -151',NULL);
INSERT INTO evaluations VALUES(44519,36,'cp -164','f6e4 e1g1 d7d5 c4b5 c8d7 b5d7 b8d7 c2c4 d5c4 b1a3 e7e6 a3c4 f8e7 d1e2 d7f6 b2b3 e8g8 c1b2 a8c8 f1d1 e4d6 a1c1 d6c4 c1c4 b7b5 c4c2 d8b6 d1c1 c8d8 d2d4 c5d4 e2d3 f6d5 b2d4 b6b7 d3e4 a7a6 d4b2');
INSERT INTO evaluations VALUES(44653,30,'cp -165',NULL);
INSERT INTO evaluations VALUES(44653,31,'cp -168',NULL);
INSERT INTO evaluations VALUES(44653,32,'cp -172',NULL);
INSERT INTO evaluations VALUES(44653,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(44653,34,'cp -157',NULL);
INSERT INTO evaluations VALUES(44653,35,'cp -164',NULL);
INSERT INTO evaluations VALUES(44653,36,'cp -172','e1g1 e7e6 f1e1 d7d5 c4b3 e4f6 c2c3 f8e7 d2d4 e8g8 c1g5 d8b6 b1d2 c5d4 f3d4 b8c6 d2f3 c8d7 d4c6 d7c6 f3e5 h7h6 g5e3 e7c5 b3c2 a8d8 e3d4 c5d4 c3d4 b6b2 e1e3 b2b6 e3g3 f6e4 c2e4 d5e4 d1g4');
INSERT INTO evaluations VALUES(49090,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(49090,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(49090,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(49090,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(49090,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(49090,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49090,36,'cp 30','f8c5 g1f3 b8c6 c1b2 e8g8 d2d4 e5d4 c3d4 c5b4 e1f1 c8e6 h2h4 h7h6 f1g1 e6d5 a2a3 b4e7 e2e3 a7a5 d1a4 d5e4 f3e1 e4g2 g1g2 d8d5 e1f3 e7d6 h1c1 c6e7 a4c2 f7f5 a3a4 f8f7 c2c4 d5c4 c1c4 e7d5 c4c2');
INSERT INTO evaluations VALUES(42269,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(42269,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(42269,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42269,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(42269,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42269,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(42269,36,'cp 34','g1f3 b8c6 c1b2 e8g8 d2d4 e5d4 c3d4 c5b4 e1f1 c8e6 h2h4 h7h6 f1g1 e6d5 a2a3 b4e7 d1d3 f7f5 a1c1 a8c8 e2e3 c6a5 f3e5 d5g2 g1g2 d8d5 f2f3 e7d6 h1d1 b7b5 d3c2 c7c6 g2f2 d6e5 d4e5 d5e6');
INSERT INTO evaluations VALUES(53404,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(53404,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(53404,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(53404,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(53404,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(53404,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(53404,36,'cp -76','g1f3 b8c6 f3e5 g8h6 d2d4 f7f6 e5f3 c8g4 f1e2 e7e6 e1g1 f8d6 h2h3 g4f3 f1f3 e8g8 c2c3 g8h8 a2a4 d8c7 b1a3 a7a6 a3c2 a8e8 e2d3 c6a5 a1b1 h6g8 d1e2 f6f5 b2b4 a5c6 f3e3 c7d7 g2g3 g8f6 c1d2 c6e7 b4b5 a6b5 d3b5 e7c6');
INSERT INTO evaluations VALUES(53403,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(53403,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(53403,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(53403,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(53403,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(53403,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(53403,36,'cp -60','b8c6 f3e5 g8h6 d2d4 f7f6 e5c6 b7c6 f1d3 c8g4 d1d2 e7e6 d2e3 d8d7 b1d2 f8d6 d2b3 e8g8 e1g1 a7a5 c1d2 a5a4 b3c5 d6c5 d4c5 g4f5 b2b4 a4b3 a2b3 f5d3 e3d3 h6f5 c2c4 h7h5 h2h3 a8a1 f1a1 f8d8 d2e1 d5c4 d3d7 d8d7 b3c4 d7d4 a1a6 d4f4 a6c6 f4c4 c6e6 c4c5 e6a6');
INSERT INTO evaluations VALUES(19871,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(19871,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(19871,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(19871,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(19871,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(19871,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(19871,36,'cp -61','f3e5 g8h6 d2d4 f7f6 e5c6 b7c6 f1d3 c8g4 d1d2 e7e6 d2e3 f8d6 b1c3 e8g8 h2h3 g4f5 d3f5 e6f5 e1g1 f8e8 e3d3 g7g6 c1d2 h6f7 c3a4 d6c7 f1e1 f7d6 e1e8 d8e8 b2b4 e8b8 a4c5 d6e4 c5a6 b8b6 a6c7 b6c7 c2c4 d5c4 d3c4 g8h8 d2e3 c7b6 c4e6 b6b4 e6c6 a8b8');
INSERT INTO evaluations VALUES(22223,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22223,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(22223,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(22223,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(22223,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(22223,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(22223,36,'cp 68','c8d7 b1c3 d8e7 c1f4 a7a6 b5c6 d7c6 e1g1 e7d7 c3e2 g8e7 c2c3 c6b5 e5f6 g7f6 f4g3 e8c8 f1e1 e7g6 a2a4 b5e2 d1e2 d8e8 b2b4 h8g8 b4b5 a6a5 e2d3 f8d6 a1d1 b7b6 c3c4 d5c4 d3c4 c8b8 g3d6 d7d6 g2g3');
INSERT INTO evaluations VALUES(27273,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(27273,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(27273,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(27273,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(27273,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(27273,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(27273,36,'cp 82','b1c3 a7a6 b5c6 d7c6 h2h4 g8h6 c1h6 g7h6 d1e2 d8e7 e1c1 e8c8 e2e3 f8g7 e5f6 e7f6 f3e5 c6e8 g2g3 h6h5 h1e1 g7h6 f2f4 h8g8 c1b1 h6f8 d1d3 f8d6 a2a3 c8b8 c3e2 f6f5 d3b3 e8a4 b3c3 a4b5 e2c1 g8g7 e3d2 d8g8 e1e3');
INSERT INTO evaluations VALUES(27271,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(27271,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(27271,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(27271,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(27271,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(27271,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27271,36,'cp 75','a7a6 b5c6 d7c6 h2h4 g8h6 c1h6 g7h6 d1d2 f8g7 e1c1 e8g8 d1e1 d8e7 d2e3 c6d7 h1f1 c7c5 d4c5 a8c8 e5f6 e7f6 f3e5 f6e5 e3e5 g7e5 e1e5 c8c5 f2f4 b7b5 c3e2 b5b4 b2b3 d7b5 f1e1 b5e2 e1e2 f8f4 g2g3 f4f1 c1b2 c5c6 e5e6 c6e6 e2e6');
INSERT INTO evaluations VALUES(29472,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(29472,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(29472,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(29472,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(29472,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(29472,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(29472,36,'cp 87','g1f3 c7c6 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 d1c2 e8g8 f1d1 b8d7 a2a3 h7h6 g5e4 f6e4 c3e4 d7f6 c1d2 f6e4 c2e4 e7d6 f3f4 c8d7 e4f3 g8h8 e3e4 f7f5 a1c1 d6c7 g1h1 a7a5 c4d3 d7e8 d2c3 d8e7 d1e1 a8d8');
INSERT INTO evaluations VALUES(29646,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29646,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29646,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(29646,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29646,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29646,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(29646,36,'cp 21','g1f3 e7e6 b1d2 f8d6 f4d6 d8d6 f1e2 b8d7 e1g1 e8g8 c2c4 a7a5 h2h3 g4f5 d1b3 b7b6 a1c1 f8c8 b3c3 c7c6 f1d1 d6e7 e2d3 f5d3 c3d3 c6c5 e3e4 d5e4 d2e4 c8d8 e4c3 c5d4 f3d4 a8c8 d3e3 c8c4 d4f5');
INSERT INTO evaluations VALUES(52707,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(52707,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(52707,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(52707,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(52707,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(52707,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(52707,36,'cp -59','c6d5 g1f3 c8f5 f1b5 b8c6 f3e5 a8c8 c2c3 e7e6 d2d4 f8d6 e1g1 g8e7 d1e2 f7f6 e5c6 b7c6 b5a6 c8b8 b1d2 c6c5 d4c5 d6c5 g1h1 e8g8 d2b3 c5b6 a2a4 e7c6 a6b5 d8c7 b5c6 c7c6 a4a5 b6c7 b3d4 c6c4 e2c4 d5c4 d4f5 e6f5');
INSERT INTO evaluations VALUES(25509,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(25509,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(25509,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(25509,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(25509,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(25509,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(25509,36,'cp 69','f1e1 a7a6 b5c4 d7g4 a2a4 d8d7 h2h3 g4h5 d2d3 f8e7 c1g5 c6d8 b1d2 d8e6 g5h4 e8g8 h4f6 h5f3 d2f3 e7f6 g2g3 a8d8 h3h4 c7c6 c4a2 g7g6 g1g2 g8g7 a1c1 a6a5 d3d4 e5d4 c3d4 d6d5 e4d5 c6d5 h4h5 e6g5 f3g5 f6g5');
INSERT INTO evaluations VALUES(30643,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(30643,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30643,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(30643,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(30643,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(30643,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(30643,36,'cp 10','f8b4 d1b3 a7a5 e2e3 g8f6 f1e2 e8g8 e1g1 d7d5 a2a3 d5c4 b3c4 b4e7 f3e5 d8c8 c1d2 c7c5 d4c5 b6c5 c4f4 b8d7 e5c4 d7b6 c4b6 c8c6 e2f3 c6b6 c3a4 b6b5 f3b7 b5b7 f4c4 b7c6 f1c1 f6e4 d2e1 e4d6 c4b3 c5c4 a4b6');
INSERT INTO evaluations VALUES(27974,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27974,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(27974,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(27974,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(27974,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(27974,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(27974,36,'cp 3','c1d2 g8f6 e2e3 e8g8 f1d3 d7d5 c4d5 e6d5 a1c1 b4e7 e1g1 c7c5 f3e5 b8c6 d2e1 e7d6 e5c6 b7c6 d4c5 b6c5 c3b5 c6b5 d3b5 d6h2 g1h2 d8b8 h2g1 b8b5 e1c3 f6e4 d1d5 e4c3 c1c3 b5b2 c3c5 a8d8 d5c4 b2f6 c4a4 f6e7 f1c1 d8d2 a2a3 f8d8 a4f4');
INSERT INTO evaluations VALUES(27973,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27973,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(27973,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27973,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27973,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(27973,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(27973,36,'cp 11','g8f6 g2g3 e8g8 f1g2 d7d6 e1g1 b8d7 b2b3 d8e7 a2a3 b4c3 d2c3 c7c5 e2e3 f8c8 c3b2 c8c7 d1e2 d6d5 f3d2 a8c8 a3a4 c5d4 e3d4 d7b8 a4a5 b8c6 a5a6 b7a8 a1c1 c6a5 f1d1 h7h6 c4c5 b6c5');
INSERT INTO evaluations VALUES(28268,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(28268,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(28268,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(28268,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(28268,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(28268,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(28268,36,'cp 16','g2g3 c7c5 d4c5 b4c5 f1g2 e8g8 e1g1 h7h6 d2f4 d8e7 c3a4 d7d5 c4d5 b7d5 a4c5 b6c5 a1c1 f8c8 f3e5 d5g2 g1g2 f6d5 d1d2 b8c6 e5c6 c8c6 f4e5 f7f6 e5c3 a8d8 d2c2 e6e5 f1d1 e5e4 g2g1 e7e6');
INSERT INTO evaluations VALUES(48378,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(48378,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48378,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(48378,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(48378,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48378,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(48378,36,'cp -9','c7c5 g1f3 c5d4 f3d4 e6e5 d4f3 f8b4 c1d2 d5d4 c3e2 b4d2 d1d2 b8c6 h2h3 e8g8 e1g1 c8e6 f1d1 d8d6 c2c3 d4c3 d2c3 a8c8 f3g5 c6d8 c3d2 h7h6 g5f3 d6b8 d2e3 d8c6 a2a3 c6e7 e2c3 f8d8 a1c1 e7g6 d3f1 g6f4 d1d2 d8d2 e3d2');
INSERT INTO evaluations VALUES(46264,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(46264,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(46264,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46264,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(46264,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(46264,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(46264,36,'cp -3','g1f3 c5d4 f3d4 e6e5 d4f3 f8b4 c1d2 d5d4 c3e2 b4d2 d1d2 b8c6 h2h3 c8e6 e1g1 e8g8 f1d1 d8d6 c2c3 d4c3 e2c3 f8d8 d2e1 d6b4 d1d2 c6d4 d3e2 d4f3 e2f3 b4b6 a1d1 d8d2 d1d2 a8d8 d2d8 b6d8 e1e3 d8d4 e3d4 e5d4 c3e2');
INSERT INTO evaluations VALUES(46263,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(46263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46263,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(46263,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(46263,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(46263,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(46263,36,'cp -12','c5c4 d3e2 f8b4 e4e5 f6e4 c1d2 e4d2 d1d2 b8c6 e1g1 e8g8 a2a3 b4a5 b2b4 a5c7 g2g3 c8d7 c3d1 f7f6 e5f6 g7f6 d1e3 g8h8 c2c3 d8e7 a3a4 e7g7 f3h4 c6e7 f2f4 f6f5 e3g2 e7g8 h4f3 g8f6 d2e3 f6e4 f3e5 c7e5 f4e5');
INSERT INTO evaluations VALUES(55845,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(55845,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(55845,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(55845,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(55845,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(55845,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(55845,36,'cp -12','g8f6 g1f3 f8d6 c3c4 b7b6 f1e2 e8g8 e1g1 c8b7 b2b3 d5c4 e2c4 b8d7 c1b2 d8e7 d1e2 e6e5 d4e5 d7e5 f3e5 d6e5 b2e5 e7e5 b1d2 a8d8 a1c1 h7h6 f1d1 f8e8 c4a6 b7a6 e2a6 c7c5 d2c4 e5e7 d1d8 e8d8');
INSERT INTO evaluations VALUES(48330,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(48330,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(48330,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(48330,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(48330,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48330,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(48330,36,'cp -10','g1f3 f8d6 c3c4 b7b6 b2b3 e8g8 f1e2 c8b7 e1g1 b8d7 c1a3 d6a3 b1a3 d8e7 d1c1 c7c5 c1b2 a8c8 a1c1 h7h6 f1d1 f8d8 c4d5 b7d5 a3b5 a7a6 b5c3 d5b7 d4c5 d7c5 b3b4 c5e4 d1d8 e7d8 c3e4 f6e4 f3e5 c8c1 b2c1 e4d6');
INSERT INTO evaluations VALUES(48331,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48331,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(48331,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(48331,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(48331,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(48331,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(48331,36,'cp -14','f8d6 c3c4 b7b6 f1e2 e8g8 b2b3 c8b7 b1c3 a7a6 e1g1 d8e7 c1b2 f8d8 c4d5 e6d5 a1c1 b8d7 e2d3 d8f8 c3e2 f6e4 e2f4 c7c6 d3b1 f8e8 f4d3 a8c8 f3e5 c6c5 e5d7 e7d7 f2f3 e4f6 d4c5 b6c5 b2f6 g7f6 d1e2 d7e7');
INSERT INTO evaluations VALUES(41807,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(41807,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(41807,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(41807,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(41807,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(41807,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41807,36,'cp -1','c3c4 b7b6 f1e2 c8b7 e1g1 e8g8 b1c3 a7a6 b2b3 d8e7 c1b2 f8d8 f1e1 b8d7 a1c1 d8e8 e2f1 f6e4 g2g3 a8d8 f1g2 h7h6 d1c2 e4c3 b2c3 d7f6 e1d1 f6e4 c3b2 d5c4 b3c4 c7c5 f3d2 e4d2 g2b7 e7b7 d1d2 c5d4');
INSERT INTO evaluations VALUES(32135,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(32135,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(32135,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(32135,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32135,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(32135,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(32135,36,'cp 69','d2d4 b8d7 c1e3 g8f6 b1d2 d6d5 d4e5 f6e4 d2e4 d5e4 d1a4 e4f3 a4g4 d7e5 g4e4 f8d6 g2f3 e8g8 e1c1 d8f6 f1e2 e5c6 h2h4 f8e8 e4c4 h7h6 h1g1 a8d8 h4h5 c6e5 c4d4 c7c5 d4a4 e5c6 d1d2 d6c7 d2d8 e8d8');
INSERT INTO evaluations VALUES(32116,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(32116,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(32116,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(32116,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(32116,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(32116,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(32116,36,'cp 76','b8d7 c1e3 g8f6 b1d2 f8e7 f1d3 g4h5 h2h3 e8g8 d1b3 b7b6 b3c2 h5g6 f3h4 c7c6 h4g6 h7g6 e1c1 e5d4 e3d4 d8c7 h3h4 d6d5 e4d5 c6d5 d2f3 e7c5 c1b1 c5d4 f3d4 d7c5 g2g4 c5d3 d1d3 f6g4 h4h5 g4f6 h5g6 f7g6');
INSERT INTO evaluations VALUES(49888,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(49888,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(49888,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(49888,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(49888,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(49888,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(49888,36,'cp -99','d7d5 c4b3 b8c6 d1e1 c8e6 d2d3 e4c5 f3e5 c6d4 e1d1 f8d6 e5f3 c5b3 a2b3 e6g4 f1e1 d4e6 h2h3 g4h5 c2c4 d5c4 b3c4 e8g8 b1c3 c7c6 g2g4 h5g6 d3d4 h7h5 g1g2 a7a5 c1d2 d6c7 d4d5 e6c5 f3e5 d8f6 e5g6 f6g6 d1f3 f7f5 g4g5 f5f4 c3e4 c5e4 f3e4 g6g5 g2f1 c6d5 e4d5 g5d5 c4d5');
INSERT INTO evaluations VALUES(49070,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(49070,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(49070,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(49070,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(49070,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(49070,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(49070,36,'cp -106','c4b3 b8c6 d1e1 c8e6 d2d3 e4c5 f3e5 c6d4 e1d1 f8d6 e5f3 c5b3 a2b3 e6g4 d1e1 d4e6 f3g5 e8g8 f2f3 e6g5 f3g4 d6c5 g1h1 f8e8 e1d1 c7c6 b1d2 e8e6 d2f3 g5f3 d1f3 d8e7 c1d2 a8e8 h2h3 c5d6 d2f4 e6f6 f3g3 d6f4 f1f4 e7e5 f4f3 e5g3 f3g3');
INSERT INTO evaluations VALUES(49078,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(49078,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(49078,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(49078,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(49078,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(49078,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(49078,36,'cp -101','b8c6 d1e1 c8e6 d2d3 e4c5 f3e5 c6d4 e1d1 c5b3 a2b3 d8f6 f2f4 f8d6 b1c3 e8g8 c1e3 d4f5 e3f2 d6e5 f4e5 f6e5 d1e2 e5d6 c3b5 d6c6 b5d4 f5d4 f2d4 f7f6 e2f2 e6f7 b3b4 f8e8 h2h4 f7g6 a1e1 a7a6 e1e8 a8e8 f1e1 h7h6 e1e8 g6e8 f2e3 e8f7 d4c3 g8h8 e3d4 c6d6 d4f2 d5d4 c3d4');
INSERT INTO evaluations VALUES(15387,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(15387,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(15387,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(15387,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(15387,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(15387,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(15387,36,'cp -81','d1e1 c8e6 d2d3 e4c5 f3e5 c6d4 e1d1 c5b3 a2b3 d8f6 f2f4 f8d6 b1c3 e8g8 c1e3 d4f5 e3f2 d6e5 f4e5 f6e5 f1e1 e5d6 c3b5 d6c6 b5d4 f5d4 f2d4 f7f6 b3b4 f8e8 d1f3 a7a6 h2h3 h7h6 c2c3 c6d7 e1e3 e6f5 a1e1 e8e3 e1e3 a8f8 d4c5 f8f7 f3g3 g8h7 e3e1');
INSERT INTO evaluations VALUES(56989,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(56989,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(56989,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(56989,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(56989,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(56989,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(56989,36,'cp -15','d2d4 g7g5 b1c3 d7d6 g2g3 f4g3 h2g3 f8g7 c1e3 c8g4 f1e2 g8f6 d1d2 b8d7 e1c1 d7b6 h1f1 d8e7 d1e1 e8c8 e2d3 g4h3 f1h1 h3e6 d4d5 e6g4 f3h2 g4d7 e3d4 f6e8 d4g7 e8g7 a2a4 b6a4 c3a4 d7a4 h2g4');
INSERT INTO evaluations VALUES(56986,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56986,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(56986,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56986,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(56986,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(56986,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(56986,36,'cp -41','g7g5 g2g3 f4g3 h2g3 f8g7 b1c3 d7d6 c1e3 c7c6 d1d2 g8f6 e1c1 f6g4 e4e5 d6d5 c3a4 b8d7 d1e1 b7b5 a4c5 d7c5 d4c5 g4e3 e1e3 e8g8 f3d4 d8c7 f1d3 f8e8 h1e1 c8g4 d3f5 g4f5 d4f5 e8e6 c1b1 a7a5 a2a3 b5b4 f5d4 a8e8 d4e6 e8e6 a3b4 a5b4 d2e2');
INSERT INTO evaluations VALUES(56946,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(56946,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(56946,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(56946,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(56946,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(56946,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(56946,36,'cp -32','g2g3 f4g3 h2g3 f8g7 b1c3 d7d6 c1e3 c7c6 d1d2 g8f6 e1c1 f6g4 e4e5 d6d5 c3a4 b8d7 d1e1 b7b5 a4c5 d7c5 d4c5 g4e3 e1e3 e8g8 f1d3 f8e8 f3d4 d8c7 h1e1 c8g4 d3f5 g4f5 d4f5 e8e6 c1b1 a8e8 f5d6 e8f8 g3g4 a7a5 d2h2 a5a4 d6f5');
INSERT INTO evaluations VALUES(57123,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(57123,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(57123,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(57123,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(57123,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(57123,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(57123,36,'cp -25','b8c6 c1b2 g8e7 a2a3 d7d5 c4d5 e7d5 e2e3 c8e6 d1c2 d8d7 g1f3 f7f6 b3b4 e8c8 d2d4 e5e4 f3d2 f6f5 b1c3 f8d6 f1c4 d5c3 b2c3 c6e7 e1g1 h7h5 b4b5 h5h4 c2b3 e6c4 d2c4 h4h3 g2g3 e7d5 f1c1 d7e6 a3a4 g7g5');
INSERT INTO evaluations VALUES(22728,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22728,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22728,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22728,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22728,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(22728,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(22728,36,'cp -18','c1b2 g8e7 g1f3 d7d5 c4d5 e7d5 f3e5 c6e5 b2e5 d8e7 e5c3 c8f5 e2e3 e8c8 f1c4 f5e4 d1e2 e4g2 e2g4 e7d7 g4d7 d8d7 h1g1 g2e4 a2a3 d5c3 b1c3 e4f5 d2d4 g7g6 b3b4 f8g7 a1c1 h8e8 c4b5 c7c6 b5e2 g7f8 c3a4 d7d8 a4c5 c8b8 h2h4 f8e7 h4h5');
INSERT INTO evaluations VALUES(22729,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(22729,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(22729,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(22729,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(22729,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(22729,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(22729,36,'cp -15','g8e7 e2e3 d7d5 c4d5 e7d5 a2a3 c8e6 g1f3 f7f6 b3b4 d8d7 d1c2 g7g5 h2h3 e8c8 b4b5 c6e7 e3e4 d5f4 d2d4 h7h5 g2g3 f4h3 b1c3 g5g4 a1d1 d7e8 d4d5 e6f7 b5b6 a7b6 c3b5 e7d5 d1d5 f8c5 d5d8 e8d8 f1h3 g4h3 e1g1 f7e6');
INSERT INTO evaluations VALUES(52377,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52377,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52377,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52377,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52377,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(52377,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52377,36,'cp 33','d7d5 e4e5 c6c5 d2d4 b8c6 f1e2 c5d4 c3d4 c8f5 g1f3 e7e6 e1g1 g8e7 b1c3 h7h6 h2h3 a7a6 c1e3 f5h7 a1c1 e7f5 e2d3 a8c8 e3f4 f8e7 c3e2 d8b6 c1c2 e8g8 g2g4 f5h4 d3h7 g8h7 f3h4 e7h4 d1d2 h7g8 f1c1');
INSERT INTO evaluations VALUES(52801,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(52801,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(52801,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(52801,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52801,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(52801,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(52801,36,'cp 29','e4e5 c8f5 g1f3 e7e6 d2d4 c6c5 f1e2 b8c6 e1g1 f5g6 a2a3 c5d4 c3d4 g8e7 b1c3 g6h5 f3g5 h5e2 c3e2 e7f5 g5f3 d8b6 d1d3 a7a6 g2g4 f5e7 e2f4 e7g6 f4h5 c6a5 c1e3 b6b3 d3b3 a5b3 a1b1 a8c8 h2h4 h7h6');
INSERT INTO evaluations VALUES(52853,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52853,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(52853,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(52853,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52853,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(52853,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52853,36,'cp 20','c8f5 g1f3 e7e6 f1e2 c6c5 d2d4 b8c6 e1g1 f5g6 c1e3 c5d4 c3d4 g8e7 f3h4 e7f5 h4f5 g6f5 f2f4 f5g6 b1c3 f8e7 a2a3 e8g8 b2b4 f7f6 e2g4 d8d7 d1d2 a8d8 a1c1 f6e5 f4e5 f8f1 g1f1 d8f8 f1g1 g6f5 g4f5 f8f5 c3e2 a7a6 h2h3 f5f7 d2d3 g8h8 e2f4');
INSERT INTO evaluations VALUES(31177,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31177,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31177,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(31177,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(31177,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(31177,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(31177,36,'cp 31','d2d4 e7e6 f1e2 c6c5 g1f3 b8c6 e1g1 f5g6 b1d2 c5d4 c3d4 d8b6 d2b3 g8e7 a2a4 c6b4 a4a5 b6c7 c1f4 e7c6 d1d2 f8e7 f1c1 e8g8 f4g5 e7g5 f3g5 c7e7 c1c3 f7f6 g5f3 g6e4 a1c1 e4f3 e5f6 f8f6 e2f3 g8h8 f3g4');
INSERT INTO evaluations VALUES(16737,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(16737,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(16737,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(16737,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(16737,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(16737,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(16737,36,'cp 87','d7d6 b1c3 c6a5 c4b3 a5b3 a2b3 g7g6 d3d4 f7f6 h2h4 c8e6 d1e2 a7a6 h4h5 g6g5 c1e3 h6g4 f3d2 g4e3 e2e3 e5d4 e3d4 h7h6 c3d5 f8g7 d4c4 c7c6 d5c7 d8c7 c4e6 c7e7 e6e7 e8e7 e1c1 e7e6 d2c4 a8d8 c4e3 g7f8 e3f5 d6d5 f5d4 e6e7 e4d5 d8d5');
INSERT INTO evaluations VALUES(15671,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(15671,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(15671,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(15671,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(15671,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(15671,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(15671,36,'cp 92','a2a4 c8e6 b1c3 e6c4 d3c4 h6g8 d1e2 f8e7 h2h3 d8d7 a4a5 g8f6 e1g1 e8g8 f1d1 b7b6 c1g5 b6a5 c3d5 f6d5 e4d5 c6d8 g5e7 d7e7 c4c5 d6c5 e2e5 e7e5 f3e5 f8e8 e5c4 a5a4 a1a4 d8b7 a4a6 e8e2 d1c1 b7d6 c4d6 c7d6');
INSERT INTO evaluations VALUES(32071,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(32071,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(32071,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(32071,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(32071,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(32071,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(32071,36,'cp 25','b8d7 f1e2 f8e7 e1g1 a7a5 b1c3 g4f3 e2f3 e7g5 c1g5 d8g5 a2a3 g5d8 g2g3 g8f6 g1g2 e8g8 d1d2 d7c5 b2b4 c5d7 f3e2 d7b6 f1b1 h7h5 b1d1 g7g6 h2h4 g8g7 f2f3 d8e7 c3b5 f8c8 c2c4 a5b4 a3b4');
INSERT INTO evaluations VALUES(42112,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(42112,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(42112,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42112,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42112,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(42112,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(42112,36,'cp 30','h7h6 c2c4 e8g8 b1c3 d8e7 f3e5 b8c6 a1c1 f8d8 c4d5 c6e5 d4e5 f6d5 d1g4 g8f8 g4h4 e7h4 h1h4 c7c5 c3e4 b7b6 g3g4 c8d7 g4g5 h6g5 e4g5 f7f6 g5f3 d7e8 e5f6 g7f6 a2a3 e8g6 g2g4 f8f7 g4g5 d8h8 g5f6 h8h4 f3h4');
INSERT INTO evaluations VALUES(41656,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(41656,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41656,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(41656,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41656,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(41656,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41656,36,'cp 39','f3e5 b8d7 b1d2 c7c5 c2c3 d8b6 d1c2 c5d4 c3d4 d7e5 d4e5 f6d7 c2c3 e8g8 c3d4 f7f6 e5f6 d7f6 d2f3 c8d7 f1d3 f8c8 e1g1 c8c7 a1c1 c7c1 f1c1 b6d4 e3d4 a8c8 c1c8 d7c8 f3e5 g8f8 g3g4 c8d7 f2f3 d7e8 g1f2 f8e7 f2e3 e7d6');
INSERT INTO evaluations VALUES(54812,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(54812,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(54812,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(54812,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(54812,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(54812,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(54812,36,'cp 58','f8g7 e2e4 g8f6 g1f3 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 a2a3 b8a6 c1e3 g4d7 f3d2 f6e8 a1b1 f7f5 f2f3 d8h4 b2b4 a5b4 a3b4 g7h6 e3h6 h4h6 d1c2 e8f6 c4c5 d6c5 e2a6 b7a6 f1e1 c5b4 b1b4 a6a5 b4c4');
INSERT INTO evaluations VALUES(20328,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(20328,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(20328,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(20328,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(20328,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(20328,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(20328,36,'cp -98','c5d4 c3b5 f8c5 b5d4 c5d4 f3d4 c6e5 f2f4 e5g6 f1d3 g8f6 e1g1 e8g8 b2b3 f8e8 c1b2 f6e4 d3e4 d5e4 d1e1 e6e5 f4e5 e8e5 e1e3 d8g5 e3g5 e5g5 b2c1 g5d5 c1e3 f7f6 h2h3 c8d7 f1d1 g8f7 c2c4 d5d6 d4c2 d6d1 a1d1 d7e8 c2b4 b7b6 d1d6 a8c8 b4d5 g6e7 d5e7 f7e7 e3f4');
INSERT INTO evaluations VALUES(20435,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(20435,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(20435,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(20435,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(20435,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(20435,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(20435,36,'cp -86','c3b5 f8c5 b5d4 c5d4 f3d4 c6e5 f2f4 e5g6 f1d3 g8f6 e1g1 e8g8 b2b3 f6e4 d1e1 e4c5 d3b5 d8b6 c1e3 a7a6 b3b4 a6b5 b4c5 b6a6 e3d2 g6e7 d2b4 e7c6 c2c3 c8d7 a2a3 f8c8 h2h3 b7b6 e1g3 f7f6 f1e1 c6d4 c3d4 b6c5 b4c5 g8h8 g3f2');
INSERT INTO evaluations VALUES(50971,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(50971,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(50971,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(50971,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50971,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(50971,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(50971,36,'cp 49','f1d3 c7c5 c2c3 b8c6 g1f3 c8d7 a2a3 g8h6 e1g1 h6f5 f1e1 h7h6 d4c5 f8c5 b1d2 e8g8 d2b3 c5a7 d3c2 a8c8 e1e2 d8c7 a3a4 c8d8 b3d4 a7d4 c3d4 d8c8 g2g4 f5e7 e2e1 f7f5 e5f6 f8f6 c1d2 e7g6 c2g6');
INSERT INTO evaluations VALUES(13056,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(13056,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(13056,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(13056,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(13056,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(13056,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(13056,36,'cp 0','c8g4 h2h3 g4f3 d1f3 e7e6 e1g1 d8b6 f1d1 f8e7 a2a3 a7a6 b5a4 e8g8 c3e2 b6a5 a4b3 a5b6');
INSERT INTO evaluations VALUES(24604,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24604,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24604,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24604,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24604,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24604,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24604,36,'cp 0','h2h3 g4f3 d1f3 e7e6 e1g1 d8b6 f1d1 f8e7 a2a3 a7a6 b5a4 e8g8 c3e2 b6a5 a4b3 a5b6');
INSERT INTO evaluations VALUES(24602,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(24602,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24602,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24602,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24602,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24602,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24602,36,'cp 0','g4f3 d1f3 e7e6 e1g1 d8b6 f1d1 a8c8 b5a4 f8d6 c1g5 d6e7 c3e2 e8g8 a1b1 b6a6 a4b3 b7b5 c2c3 b5b4 b3c2 c6a5 c2d3 a6b6 g5f4 e7d6 f4g5');
INSERT INTO evaluations VALUES(24607,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24607,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24607,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24607,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24607,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24607,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24607,36,'cp 0','d1f3 e7e6 e1g1 d8b6 f1d1 f8e7 a2a3 a7a6 b5a4 e8g8 c3e2 b6a5 a4b3 a5b6');
INSERT INTO evaluations VALUES(24606,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24606,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24606,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24606,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24606,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24606,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24606,36,'cp 0','e7e6 e1g1 d8b6 f1d1 f8e7 a2a3 a7a6 b5a4 e8g8 c3e2 b6a5 a4b3 a5b6 b3a4');
INSERT INTO evaluations VALUES(24697,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24697,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24697,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24697,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24697,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24697,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24697,36,'cp 0','e1g1 d8b6 f1d1 a8c8 b5a4 f8e7 c3e2 b6a6 a4b3 b7b5 c2c3 b5b4 b3c2 b4c3 b2c3 c6a5 c2a4 a5c6');
INSERT INTO evaluations VALUES(24698,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24698,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(24698,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24698,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(24698,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(24698,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(24698,36,'cp -5','f8d6 c1g5 h7h6 g5h4 d6e7 b5c6 b7c6 c3a4 f6e4 h4e7 d8e7 b2b3 e8g8 f1e1 f8e8 a1d1 e7d6 f3e3 a8c8 c2c3 d6e7 e3d3 e7c7 d3c2 c8d8 f2f3 e4d6 c2f2 d8c8 d1c1 d6b7 g1h1 e8d8 h1g1 a7a5 f3f4 c6c5 f4f5 e6f5 f2f5 c7c6');
INSERT INTO evaluations VALUES(40714,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(40714,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(40714,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(40714,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40714,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40714,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40714,36,'cp 0','e8g8 d1c2 d5c4 e2e3 f6d5 h4e7 d8e7 f1c4 d5c3 c2c3 b7b6 e1g1 c8b7 f3e5 b8d7 e5d7 e7d7 c4e2 f8c8 f1d1 d7e7 b2b4 a7a5 a2a3 b7d5 d1c1 a5b4 a3b4 a8a1 c1a1 c8a8 a1a8 d5a8 f2f3 a8b7 e2c4 e7d7 c4f1 b7d5 f1c4');
INSERT INTO evaluations VALUES(36200,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36200,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36200,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36200,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(36200,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(36200,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36200,36,'cp 6','e2e3 b7b6 f1d3 c8b7 h4f6 e7f6 c4d5 e6d5 e1g1 c7c5 a1c1 c5d4 f3d4 b8c6 c3e2 c6d4 e2d4 g7g6 d1g4 a8c8 g4f4 f6g7 c1c8 d8c8 f1d1 c8b8 f4b8 f8b8 f2f4 b8c8 g2g4 g7f6 g1f2 g8f8 f2g2 f8g7 g2f3 c8c7 d4b5 c7e7 b5d6');
INSERT INTO evaluations VALUES(21825,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(21825,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(21825,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(21825,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(21825,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(21825,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(21825,36,'cp 27','f1b5 f8c5 b5c6 b7c6 d2d3 g8e7 d1e2 e7g6 c1e3 g6f4 e3f4 e5f4 b1d2 e8g8 e1c1 c8e6 c1b1 a7a5 d3d4 c5d6 a2a4 a8b8 b2b3 b8b4 e2a6 h7h6 h1e1 d8a8 a6a8 f8a8 c2c3 b4b3 d2b3 e6b3 d1d2 b3a4 e4e5 d6e7 d2b2 a4b5 e5e6 f7e6 e1e6 g8f7 e6e1');
INSERT INTO evaluations VALUES(21547,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(21547,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(21547,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21547,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(21547,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(21547,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(21547,36,'cp 6','f8c5 b5c6 b7c6 d2d3 g8e7 d1e2 e7g6 c1g5 f7f6 g5e3 d8e7 b1d2 a7a5 a2a4 c8e6 e1g1 c5e3 e2e3 e7b4 b2b3 b4b6 e3f2 b6f2 f1f2 c6c5 d2f1 h7h5 h2h4 e6g4 g2g3 g6e7 f1e3 g4e6 f3d2 e8g8 a1e1 e7c6 d2c4 f8d8 c4b2');
INSERT INTO evaluations VALUES(22528,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(22528,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22528,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(22528,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(22528,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(22528,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(22528,36,'cp 18','a2a3 g8f6 c2c4 g7g6 b2b4 d7d6 b4b5 c6e7 g2g3 a7a6 a3a4 a6b5 c4b5 c7c6 c1a3 f8g7 f1g2 c8e6 b1d2 e8g8 e1g1 c6b5 a4b5 f8e8 d1c2 d8d7 c2b2 e8c8 f1c1 c8c1 a1c1 a8c8 g2f1 c8c1 b2c1 d7b5');
INSERT INTO evaluations VALUES(44075,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(44075,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(44075,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44075,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(44075,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(44075,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(44075,36,'cp -35','d7d5 c4b3 d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 b8d7 g1f3 f8d6 c1e3 d7f6 f3d2 f6g4 d2c4 g4e3 c4e3 a7a5 a2a4 g7g6 d1e2 h7h5 a1d1 d6c7 h2h4 e8e7 g2g3 f7f5 e4f5 g6f5 h1g1 h8d8 d1d8 c7d8 f2f4 d8c7 c2c3 c8e6 b3e6 e7e6 f4e5 c7e5');
INSERT INTO evaluations VALUES(45802,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(45802,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(45802,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(45802,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(45802,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(45802,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(45802,36,'cp -29','c4b3 d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 b8d7 g1f3 f8d6 f3d2 a7a5 a2a4 d7c5 b3c4 f7f5 f2f3 c8d7 d1e2 e8c8 b2b3 c8c7 h2h3 h8f8 c1b2 g7g6 a1d1 c5e6 c4e6 d7e6 b2c3 b7b6 d2c4 e6c4 b3c4 d8d7 d1d3 f8d8');
INSERT INTO evaluations VALUES(23370,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(23370,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(23370,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(23370,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(23370,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(23370,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(23370,36,'cp 51','e7g6 c1e3 f8b4 b1d2 b7b6 g2g3 g8f6 f1d3 e8g8 a2a3 b4c5 e3c5 b6c5 b2b4 d7d6 h2h3 h7h6 d1c2 a7a6 g1f3 c5b4 a3b4 c8d7 h3h4 c7c6 e1g1 c6d5 c4d5 a6a5 b4b5 f6g4 h4h5 g6e7 d3e2 f7f5');
INSERT INTO evaluations VALUES(22986,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(22986,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(22986,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(22986,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(22986,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(22986,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(22986,36,'cp 53','c1e3 f8b4 b1d2 b7b6 g2g3 b4c5 e3c5 b6c5 d1a4 g8f6 h2h4 h7h5 f2f3 e8g8 e1c1 d7d6 c1b1 c8d7 a4c2 c7c6 g1e2 a8b8 b1a1 d8b6 e2c3 b8b7 f1d3 f8b8 d1b1 a7a6 h1g1 b6c7 c2d1 g6e7 d1e2 c6d5 c4d5 d7b5 d2c4 g8h8 g1h1 h8g8 b1g1');
INSERT INTO evaluations VALUES(6391,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(6391,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(6391,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6391,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(6391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6391,36,'cp 0','g8e7 e1g1 e8g8 c2c3 f7f5 d3d4 e5d4 e4e5 d4c3 b1c3 h7h6 c3d5 g8h7 c1h6 g7h6 d5f6 h7g7 f6h5 g7h7');
INSERT INTO evaluations VALUES(11018,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11018,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(11018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11018,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11018,36,'cp 0','a2a4 h7h6 e1g1 d7d6 c2c3 a7a6 b5c4 e7g6 b1a3 e8g8 a3c2 g8h8 c2e3 c5a7 b2b4 d8e8 e3d5 e8d8');
INSERT INTO evaluations VALUES(21393,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(21393,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(21393,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(21393,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(21393,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(21393,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(21393,36,'cp 35','e2e3 g8f6 g1f3 f8b4 f1d3 e8g8 a2a3 b4c3 b2c3 f6e4 e1g1 d8e7 c3c4 c6a5 c4d5 e6d5 f3d2 c8f5 d2e4 f5e4 f2f3 e4d3 c2d3 b7b6 f1e1 e7d7 a1c1 a8c8 e3e4 c7c5 d4c5 c8c5 c1c5 b6c5 d1c2 f8c8 e1c1 c5c4 d3c4 a5c4');
INSERT INTO evaluations VALUES(21391,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(21391,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(21391,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(21391,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(21391,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21391,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21391,36,'cp 25','f8d6 g1f3 a7a6 f1d3 g8f6 f4g3 c6b4 e1g1 b7b6 g3d6 c7d6 a2a3 b4d3 c2d3 e8g8 h2h3 h7h6 a1c1 a6a5 c3b5 c8d7 d1b3 d8e7 c1c7 f8c8 c7b7 a8b8 b7b8 c8b8 f1c1 f6e8 c1c2 b8c8 c2c8 d7c8 e3e4 c8d7 e4d5 d7b5 b3b5');
INSERT INTO evaluations VALUES(49091,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(49091,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(49091,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(49091,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(49091,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(49091,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(49091,36,'cp 29','b2c3 c7c5 d1a4 c8d7 a4b3 b8c6 b3b7 a8b8 b7a6 d8c7 a6d3 f8e7 a1b1 e8g8 b1b8 f8b8 g1f3 d7e6 d3c2 c7a5 e1g1 c5c4 f3g5 e7g5 g2c6 g5e7 c6e4 g7g6 d2d4 e5d4 c3d4 c4c3 d4d5 e6d5 e4d5 a5d5 c2c3 d5a2 c1h6');
INSERT INTO evaluations VALUES(53135,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(53135,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(53135,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(53135,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(53135,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(53135,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(53135,36,'cp -25','d4c3 b1c3 e7e6 c1f4 e6e5 f4e3 b8c6 f1c4 f8e7 e1g1 g8f6 f3g5 e8g8 d1b3 d6d5 e4d5 c6a5 b3a4 a5c4 a4c4 c8f5 a1d1 h7h6 g5e4 f6e4 c3e4 f5e4 c4e4 e7d6 e4f5 d8c8 f5d3 c8d7 d1c1 f8c8 c1c8 a8c8 e3a7 c8a8 a7e3 a8a2 d3c3 a2a4 f1c1 a4b4 c3c8 d7c8 c1c8 g8h7');
INSERT INTO evaluations VALUES(53143,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(53143,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(53143,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(53143,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(53143,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(53143,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(53143,36,'cp -12','b1c3 b8c6 f1c4 e7e6 d1e2 c8d7 e1g1 f8e7 f1d1 a7a6 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 h2h3 d8e8 e2c4 g8h8 c4b3 a8b8 e3b6 b8c8 a1c1 f7f5 f3d2 e8g6 c1c3 e7g5 b3a3 g5d2 d1d2 f5e4 d5c6 c8c6 c3c6 d7c6 a3d6 g6d6 d2d6');
INSERT INTO evaluations VALUES(23244,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(23244,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(23244,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(23244,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(23244,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(23244,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(23244,36,'cp 111','b1c3 f8c5 f1c4 d7d6 c3a4 c8g4 c2c3 c5b6 h2h3 g4h5 a4b6 a7b6 d2d4 f6f5 f4e5 f5e4 c1g5 g8e7 g2g4 d6d5 g4h5 d5c4 f3h4 h8f8 h5h6 g7g6 d1g4 d8d7 g4e4 d7d5 e4d5 e7d5 e1e2 c6e7 h1f1 f8f1 e2f1 e8d7 f1e2 e7f5 a2a4 d7e6 e2f3 f5h4 g5h4 b6b5');
INSERT INTO evaluations VALUES(23242,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(23242,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(23242,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(23242,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(23242,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(23242,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(23242,36,'cp 100','f8c5 f1c4 d7d6 c3a4 c5b6 a4b6 a7b6 d2d4 c8g4 d4d5 c6d4 c1e3 b6b5 c4d3 g8e7 e3d4 e5d4 e1g1 c7c6 h2h3 g4f3 d1f3 e8g8 a2a3 d8d7 b2b3 f6f5 a3a4 b5a4 a1a4 a8a4 b3a4 c6c5 f1e1 g8h8 e4f5 e7f5 e1e6 g7g6 d3f5 f8f5 f3e4 f5f8 a4a5 d7f7 e6d6 f7f4 e4f4 f8f4');
INSERT INTO evaluations VALUES(38339,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(38339,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(38339,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(38339,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(38339,34,'cp -108',NULL);
INSERT INTO evaluations VALUES(38339,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(38339,36,'cp -99','d6e5 f4e5 c7c5 f1b5 b8c6 c2c3 c5d4 b5c6 b7c6 d1d4 d8d4 c3d4 g8h6 g1f3 c8g4 c1h6 g7h6 b1c3 e8g8 e1f2 a8b8 b2b3 f7f6 e5f6 f8f6 f2g3 g4h5 c3e4 f6f5 a1d1 b8b4 d1d3 h6f4 g3f2 h5f3 g2f3 f5a5 e4c3 g8f7 d3d1 h7h5 f2e2');
INSERT INTO evaluations VALUES(38488,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(38488,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(38488,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(38488,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(38488,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(38488,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(38488,36,'cp -70','f4e5 c7c5 f1b5 b8c6 d4d5 a7a6 b5a4 b7b5 d5c6 d8d1 e1d1 c8g4 g1f3 e8c8 b1d2 b5a4 h1e1 c8c7 e1e4 g4f3 g2f3 d8d4 f3f4 g8h6 e4d4 c5d4 d2f3 h6f5 d1e2 c7c6 e2d3 f7f6 d3e4 c6d7 c1d2 h8b8 b2b3 d7e6 a1c1 h7h6 c1e1 f6e5 f4e5 b8c8 e1c1 a4b3 c2b3 c8c1 d2c1');
INSERT INTO evaluations VALUES(38485,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(38485,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(38485,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(38485,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(38485,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(38485,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(38485,36,'cp -60','c7c5 f1b5 c8d7 b5d7 d8d7 d4d5 e7e6 b1c3 g7e5 d1e2 e5c3 b2c3 d7d5 g1f3 b8d7 e1g1 g8f6 c1h6 f6e4 a1e1 f7f5 f3g5 e4g5 h6g5 e6e5 e1d1 d5c6 d1d2 e8g8 f1d1 f8f7 d2d6 c6a4 h2h4 a8e8 d1d5 e5e4 h4h5 d7f8 e2d2 g6h5 g5h6');
INSERT INTO evaluations VALUES(11984,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(11984,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(11984,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(11984,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(11984,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(11984,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(11984,36,'cp -35','a8b8 a1b1 f8c5 f1c4 h7h6 g5f6 d8f6 e1g1 e8g8 c3a4 c5d6 d1e2 d6c7 b2b4 f6e7 a2a3 f8d8 f1d1 a7a5 b4b5 c6b5 b1b5 b8b5 c4b5 c8b7 e2c4 c7d6 a4c3 d6a3 c4c7 a3b4 c7b7 b4c3 d1d7 d8d7');
INSERT INTO evaluations VALUES(30092,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30092,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30092,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(30092,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(30092,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(30092,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(30092,36,'cp 5','f5d3 c2d3 c7c5 d4c5 f8c5 d1b3 d8b6 b3b6 c5b6 e1e2 b8c6 b1c3 e8e7 a1c1 d5d4 c3e4 f6e4 d3e4 d4e3 f4e3 b6e3 e2e3 h8d8 h1d1 d8d1 c1d1 f7f6 d1c1 a8d8 h2h4 h7h5 g2g3 d8d7 a2a3 e6e5 b2b4 e7e6 e3e2 e6d6 c1d1 d6e7');
INSERT INTO evaluations VALUES(30163,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(30163,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(30163,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(30163,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(30163,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(30163,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30163,36,'cp 7','c2d3 c7c5 d4c5 f8c5 d1b3 d8e7 e1g1 b8c6 f4g5 c5b6 b1c3 e7b4 b3b4 c6b4 g5f6 g7f6 d3d4 a8c8 f1d1 e8e7 a1c1 h7h5 g2g3 b4c6 c3e2 c6b4 a2a3 b4c6 g1f1 c6a5 f3d2 a7a6 e2f4 h5h4 c1c8 h8c8 g3h4 c8c2');
INSERT INTO evaluations VALUES(30160,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(30160,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(30160,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(30160,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(30160,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(30160,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30160,36,'cp 7','c7c5 d4c5 f8c5 d1b3 d8e7 e1g1 b8c6 f1c1 e8g8 d3d4 c5b6 h2h3 e7b4 b3c2 f8c8 b1c3 b4e7 c3a4 b6c7 f4c7 c8c7 a4c5 c6a5 c2a4 b7b6 c5d3 a8c8 c1c7 c8c7 f3e5 e7e8 a4d1 a5c4 a1c1 c7c8 b2b3 c4e5 d3e5');
INSERT INTO evaluations VALUES(15701,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(15701,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(15701,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(15701,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(15701,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(15701,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(15701,36,'cp 219','f8c5 h2h3 d8e7 d2d3 c8e6 g2g3 e8c8 f1g2 f6d7 d1e2 c5d4 c1e3 c6c5 c3b5 d4e3 e2e3 c8b8 b5c3 d7e5 f2f4 e5c6 e1g1 h8e8 g1h2 f7f6 e3f2 e7f8 a2a3 c5c4 a1d1 c4d3 c2d3 c6d4 c3e2 c7c5 e2d4 d8d4');
INSERT INTO evaluations VALUES(9490,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(9490,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(9490,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(9490,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(9490,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(9490,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(9490,36,'cp 225','h2h3 c8e6 d2d3 d8e7 g2g3 e8c8 d1e2 c5d4 f1g2 f6d7 c1e3 c6c5 c3b5 d4e3 e2e3 c8b8 f2f4 g7g5 e1c1 h8g8 h1g1 h7h6 b5c3 f7f6 d1e1 a7a6 e3f2 e7d6 a2a3 b8a7 g1f1 g8e8 e4e5 f6e5 c3e4 d6e7 f4e5 e6d5 e4f6 d5g2 f2g2 d7f6 f1f6');
INSERT INTO evaluations VALUES(9488,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(9488,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(9488,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(9488,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(9488,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(9488,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(9488,36,'cp 208','c8e6 d2d3 d8e7 d1e2 e8c8 g2g3 c5d4 f1g2 c8b8 c1e3 c6c5 c3b5 d4e3 e2e3 f6d7 f2f4 g7g5 e1c1 h8g8 h1g1 g8g7 f4f5 e6f5 e3e1 a7a6 b5c7 f5e6 c7d5 e6d5 e4d5 e7d6 g1f1 f7f6 e1f2 h7h5 d1e1 h5h4 g3g4 g7e7 e1e7 d6e7');
INSERT INTO evaluations VALUES(52363,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52363,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52363,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(52363,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(52363,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(52363,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52363,36,'cp 16','g8f6 d2d4 d7d5 f1g2 g7g6 c4d5 c6d5 g1f3 f8g7 f3e5 e8g8 b1c3 c8f5 c1f4 d8b6 e1g1 b6b2 c3d5 f6d5 g2d5 b8c6 d5c6 g7e5 f4e5 b7c6 f1e1 f7f6 e5c7 f8c8 c7a5 a8b8 e2e4 f5h3 d1d3 b2b5 d3d2 b5b2 d2b2 b8b2 a5c3 b2c2 e1c1 c2c1 a1c1 h3e6 c1b1 c8d8 b1b7 d8d7 b7d7 e6d7');
INSERT INTO evaluations VALUES(44470,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44470,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(44470,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44470,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(44470,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(44470,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44470,36,'cp 23','g1f3 d7d5 d2d4 g7g6 c4d5 c6d5 f1g2 f8g7 b1c3 e8g8 e1g1 f6e4 f3d2 e4c3 b2c3 b8c6 e2e4 c8e6 e4e5 d8d7 d2b3 b7b6 f1e1 e6h3 g2h3 d7h3 a2a4 h3f5 a4a5 f8c8 h2h4 h7h5 c1f4 c6a5 b3a5 b6a5 d1d2 c8c4 a1a5 a8c8 f4g5 c4c3 g5e7 c3c7');
INSERT INTO evaluations VALUES(44469,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(44469,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(44469,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(44469,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44469,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(44469,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44469,36,'cp 15','d7d5 d2d4 g7g6 c4d5 c6d5 f1g2 f8g7 f3e5 e8g8 e1g1 c8f5 b1c3 f6e4 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 e5c6 b7c6 e2e3 d8d7 e4f3 f7f5 g4g5 a8d8 c1d2 f5f4 d1e2 f4e3 d2e3 e6d5 f3d5 c6d5 f2f4 d8c8 a1c1 c8c1 f1c1');
INSERT INTO evaluations VALUES(44892,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(44892,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(44892,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44892,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(44892,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(44892,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44892,36,'cp 15','d2d4 g7g6 f1g2 f8g7 e1g1 e8g8 c4d5 c6d5 b1c3 f6e4 f3d2 e4d2 c1d2 e7e6 d2e3 b8c6 d1d2 b7b6 f1d1 c6e7 e3g5 c8b7 h2h4 a8c8 a1c1 d8d7 g2h3 h7h5 e2e3 e7f5 h3f5 e6f5 c3e2 c8c4 b2b3 c4c1 d1c1 f8c8');
INSERT INTO evaluations VALUES(44915,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(44915,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44915,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44915,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(44915,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(44915,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44915,36,'cp 8','g7g6 c4d5 c6d5 f1g2 f8g7 f3e5 b8c6 e1g1 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 a8c8 b1c3 e8g8 f1c1 d8e7 h2h4 f8d8 c1c2 c8c6 a1c1 d8c8 g2f3 e7d8 a2a3 a7a6 e2e3 h7h6 c3a2 c6c2 c1c2 c8c2 d2c2 d7c6 f3e2');
INSERT INTO evaluations VALUES(56261,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(56261,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(56261,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(56261,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(56261,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(56261,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(56261,36,'cp -100','f8e7 d2d4 g8f6 f3e5 e8g8 c1e3 c7c5 d4c5 b8d7 b1c3 d7e5 f4e5 f6g4 e3d4 g4e5 d4e5 f8e8 d1d4 e7f6 e1c1 f6e5 d4d5 d8g5 c1b1 c8e6 d5d2 g5d2 d1d2 e5c3 b2c3 e8c8 f1b5 c8c5 b5d7 c5d5 d2d5 e6d5 h1g1 a8d8 d7b5 d5e6 b5d3 d8c8 b1b2 b7b6 a2a4 c8c5 g1e1 g8f8 h2h3 h7h5');
INSERT INTO evaluations VALUES(21066,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(21066,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(21066,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(21066,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(21066,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(21066,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(21066,36,'cp 42','c1b2 g8f6 e2e3 c8g4 d2d4 e7e6 f1e2 f6e4 a2a3 c6b8 b1d2 e4d2 f3d2 g4f5 c2c4 c7c6 c4c5 f8e7 e1g1 b8d7 b3b4 e8g8 d2b3 d8c7 a3a4 a7a6 b2c3 h7h6 f2f3 e7g5 d1d2 d7f6 b4b5 a6b5 a4b5 c6b5 e2b5 a8a1 f1a1 f8b8 b5d3 f5d3 d2d3 f6d7');
INSERT INTO evaluations VALUES(45130,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45130,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45130,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45130,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45130,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45130,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45130,36,'cp 0','e7e6 c4b3 b8c6 f2f4 d6d5 g1f3 b7b5 e4d5 e6d5 c3b5 d8a5 b5c3 d5d4 d1e2 f8e7 f3e5 d4c3 e5c6 c3b2 e1f2 a5c7 h1e1 b2a1n e2e7 c7e7 e1e7 e8f8 e7f7 f8e8 f7e7 e8f8');
INSERT INTO evaluations VALUES(45990,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45990,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45990,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45990,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45990,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45990,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45990,36,'cp -8','c4b3 f8e7 f2f4 d6d5 e4e5 f6d7 a2a4 b8c6 g1f3 e7h4 g2g3 h4e7 e1g1 e8g8 d1e2 a7a6 c3d1 a8b8 c2c3 b7b5 a4b5 a6b5 d3d4 b5b4 b3c2 b4c3 b2c3 c5d4 c3d4 d8b6 c1e3 c6b4 c2d3 b4d3 e2d3 f7f6 d1f2 f6e5 d4e5 d7c5');
INSERT INTO evaluations VALUES(53075,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(53075,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53075,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(53075,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(53075,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(53075,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(53075,36,'cp 4','g8f6 b1c3 e7e5 e2c4 h7h6 d2d3 b8c6 h2h3 c8e6 c3d5 f8e7 a2a4 d8d7 c2c3 e7d8 f3h2 e8g8 d5f6 d8f6 c4e6 f7e6 d1h5 g8h7 h2g4 d7f7 h5f7 f8f7 e1e2 a8f8 h3h4 f6e7 h4h5 e7g5 c1g5 h6g5');
INSERT INTO evaluations VALUES(45211,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45211,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(45211,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45211,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45211,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45211,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(45211,36,'cp 15','b1c3 e7e5 e2c4 h7h6 d2d3 b8c6 h2h3 c8e6 c3d5 g7g6 c2c3 f8g7 d5f6 d8f6 d1b3 e6c8 a2a4 e8g8 a4a5 a8b8 c1d2 c8d7 b3a4 f6d8 a4a2 b7b6 a5b6 a7b6 e1g1 b6b5 c4d5 c6e7 d5b3 d8c8 d2e3 d7c6 f1d1 c8c7 d1b1 c7b7');
INSERT INTO evaluations VALUES(45188,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45188,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(45188,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(45188,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45188,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45188,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(45188,36,'cp 8','e7e5 e2c4 h7h6 d2d3 b8c6 h2h3 c8e6 c3d5 g7g6 c2c3 f8g7 d5f6 d8f6 d1b3 c6d8 c4e6 f6e6 b3b5 e6d7 b5d7 e8d7 a2a4 d8e6 h3h4 h6h5 f3g5 f7f5 e1e2 e6c7 g2g3 b7b6 c1e3 h8f8 g5f3 g7f6 h1b1 d6d5 e4d5 c7d5 e3g5 d7d6 g5f6 d5f6');
INSERT INTO evaluations VALUES(6522,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(6522,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(6522,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(6522,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(6522,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(6522,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(6522,36,'cp 101','f3e5 g8f6 d2d3 e8g8 e5c6 b7c6 f1e2 f8e8 e4e5 d7d6 e5f6 d8f6 b1c3 f6d4 h1f1 c8g4 h2h3 g4e2 c3e2 e8e6 c2c3 d4d5 f1f2 c5f2 e1f2 a8e8 e2d4 e6e1 d1d2 d5c5 d2e1 e8e1 f2e1 c5d5 e1f2 c6c5 d4e2 d5d3 b2b4 a7a5 b4a5 d3a6 a2a4 a6a5 c1d2 h7h5 a1a2 a5b6 a4a5 b6b3');
INSERT INTO evaluations VALUES(37357,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(37357,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(37357,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(37357,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37357,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(37357,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(37357,36,'cp 55','e2e4 d7d6 d1d2 c7c5 d4d5 a7a6 a2a4 d8a5 a1a3 a5b4 g1f3 b8d7 f1d3 h7h6 g5f4 g6g5 f4g3 f6h5 d2c2 b4b6 e1g1 h5g3 h2g3 b6d8 c2d2 e7e6 c3d1 a8b8 a3b3 d7e5 f3e5 g7e5 a4a5 e6d5 e4d5 f7f5 d3c2 e5d4 d1e3');
INSERT INTO evaluations VALUES(37361,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(37361,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(37361,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(37361,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(37361,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(37361,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(37361,36,'cp 14','d7d6 d1d2 c7c5 d4d5 b7b5 c4b5 a7a6 a2a4 d8a5 a1a3 a5b4 f2f3 b8d7 c3a2 b4d2 e1d2 e7e6 a2c3 e6d5 e4d5 f8e8 b5a6 d7b6 g5f6 g7f6 a4a5 f6c3 b2c3 b6d5 f1b5 e8e5 c3c4 d5b4 g1e2 d6d5 e2c3 d5d4 c3e4 c8a6 b5a6 a8a6 h1b1');
INSERT INTO evaluations VALUES(36478,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36478,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(36478,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(36478,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36478,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(36478,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36478,36,'cp 30','h2h3 c7c5 d4d5 e7e6 g1f3 e6d5 c4d5 f8e8 f1d3 c5c4 d3c2 b7b5 e1g1 a7a6 f1e1 b8d7 a2a3 d8b6 d1d2 a6a5 g5e3 d7c5 e3d4 f6d7 d4g7 g8g7 d2e3 d7f6 e3d4 c8d7 c3e2 a8b8 f3d2 c5a6 d4b6 b8b6 f2f3 a5a4 g1f2 b6b8 e2d4 a6c5 d4c6 b8a8 a1b1');
INSERT INTO evaluations VALUES(36475,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(36475,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(36475,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(36475,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36475,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(36475,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36475,36,'cp 37','c7c5 d4d5 e7e6 g1f3 e6d5 c4d5 f8e8 f1d3 c5c4 d3c2 b7b5 e1g1 a7a6 a2a3 b8d7 f1e1 d8b6 g5e3 d7c5 e3d4 f6d7 d4g7 g8g7 d1d4 g7g8 f3d2 a8b8 f2f4 c5d3 d4b6 d7b6 c2d3 c4d3 e1e3 b6c4 e3d3 c4b2 d3d4 c8d7 a1b1');
INSERT INTO evaluations VALUES(42747,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42747,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42747,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42747,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(42747,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(42747,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42747,36,'cp 13','b4c3 d2c3 d7d6 f1d3 b8d7 c1e3 b7b6 f3d2 d7c5 e3c5 b6c5 e1g1 e8g8 b2b4 c8g4 d1e1 f6h5 e1e3 h5f4 b4c5 d6c5 f1e1 d8g5 d3f1 a8d8 g1h1 g5f6 d2b3 d8d6 f2f3 g4e6 b3c5 f8d8 c5e6');
INSERT INTO evaluations VALUES(42815,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42815,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(42815,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(42815,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(42815,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(42815,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42815,36,'cp 18','d2c3 d7d6 f1d3 b8d7 c1e3 b7b6 f3d2 d7c5 e3c5 b6c5 b2b4 c8e6 e1g1 e8g8 d1e2 g7g6 b4c5 f6d7 c5d6 c7d6 d3c4 d7c5 e2e3 d8c7 c4e6 c5e6 c3c4 f8c8 a1b1 h7h5 h2h4 c7c5 e3c5 c8c5 b1b7 g8g7 a3a4 e6d4 b7d7 d4c2 d7d6 c2d4');
INSERT INTO evaluations VALUES(42811,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(42811,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(42811,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42811,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(42811,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42811,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42811,36,'cp 0','d7d6 f1d3 b8d7 c1e3 e8g8 f3d2 b7b6 e1g1 c8b7 f1e1 a7a5 e3g5 h7h6 g5h4 f8e8 f2f3 d7f8 d2f1 f8g6 h4f2 d6d5 e4d5 f6d5 d3g6 f7g6 f1e3 d5e3 e1e3 e5e4 f3e4 e8e4 d1d8 a8d8 a1e1 e4e3 f2e3 d8e8 e3f2 e8e1 f2e1');
INSERT INTO evaluations VALUES(41921,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(41921,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(41921,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(41921,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41921,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41921,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(41921,36,'cp 6','f1d3 e8g8 d1e2 b8d7 c1e3 b7b6 f3d2 d7c5 e3c5 b6c5 b2b4 c8g4 e2e3 g4e6 e1g1 f6d7 f1e1 d8e7 d3b5 a8b8 e3e2 e7g5 d2f3 g5h5 b5d7 e6d7 b4c5 d7g4 c5d6 c7d6 c3c4 h7h6 e1d1 g4f3 e2f3 h5f3');
INSERT INTO evaluations VALUES(45163,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(45163,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(45163,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45163,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45163,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(45163,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45163,36,'cp 0','b1c3 e7e5 f1e2 f8e7 a2a3 b8c6 e1g1 e8g8 a1b1 a7a5 d2d3 c8d7 c1d2 a8b8 c3d5 h7h6 f1e1 f8e8 h2h3 b7b6 d5c3 e8f8 c3d5');
INSERT INTO evaluations VALUES(45157,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45157,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45157,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45157,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45157,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45157,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45157,36,'cp 2','e7e5 d2d3 b8c6 a1b1 f8e7 f1e2 c8d7 e1g1 e8g8 a2a3 a7a5 c1d2 a8b8 b2b3 f6e8 c3d5 e8c7 b3b4 c5b4 a3b4 c6b4 d2b4 a5b4 b1b4 b7b5 d3d4 c7d5 c4d5 d7g4 b4b5 b8b5 e2b5 e5d4 d1d3 g4f3 d3f3 d8c7 b5d3');
INSERT INTO evaluations VALUES(45904,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45904,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45904,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45904,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45904,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45904,36,'cp 10','d2d3 b8c6 f1e2 f8e7 e1g1 e8g8 a2a3 c8d7 c1e3 a7a5 b2b3 c6d4 e3d2 d4c6 a1b1 f6e8 c3d5 e8f6 h2h3 b7b6 d5c3 f6e8 f3h2 c6d4 e2g4 b6b5 g4d7 d8d7 a3a4 b5b4 c3b5 e8c7 f2f4');
INSERT INTO evaluations VALUES(4986,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(4986,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(4986,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(4986,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(4986,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(4986,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(4986,36,'cp 19','g8f6 f1d3 e8g8 c3a4 c5b6 e1g1 d7d5 e4d5 c6d5 a4b6 a7b6 b2b3 c7c5 c1b2 d5d4 c2c3 d4c3 b2c3 f6d5 c3e5 d5b4 d3c4 d8g5 f2f4 g5h6 a2a3 b4c6 d1d6 c6e5 d6h6 g7h6 f4e5 c8e6 c4e6 f7e6 a3a4 a8d8 f1d1 d8d5 d1d5 e6d5');
INSERT INTO evaluations VALUES(19463,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19463,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(19463,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(19463,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19463,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19463,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19463,36,'cp 0','g8f6 b1c3 d7d5 e4d5 f6d5 e2b5 g7g6 c3a4 d8d6 d2d4 d6e6 b5e2 c5d4 f3d4 e6e5 d4f3 e5f5 f3d4');
INSERT INTO evaluations VALUES(12841,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12841,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12841,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12841,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12841,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12841,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12841,36,'cp 0','b1c3 d7d5 e4d5 f6d5 e2b5 g7g6 c3a4 d8d6 d2d4 d6e6 b5e2 c5d4 f3d4 e6e5 d4f3 e5f5 f3d4');
INSERT INTO evaluations VALUES(12823,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12823,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12823,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12823,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12823,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12823,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12823,36,'cp 0','d7d5 e4d5 f6d5 e2b5 g7g6 c3a4 d8d6 d2d4 d6e6 b5e2 c5d4 f3d4 e6e5 d4f3 e5f5 f3d4');
INSERT INTO evaluations VALUES(21178,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(21178,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(21178,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(21178,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(21178,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(21178,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(21178,36,'cp 42','b1c3 g8f6 g1f3 c8f5 e2e3 c6b4 a1c1 c7c6 a2a3 b4d5 c3d5 f6d5 f4g3 d5f6 f1d3 e7e6 e1g1 a7a5 c2c4 f8e7 f1e1 f5d3 d1d3 d6d5 d3b3 d8c8 f3e5 e8g8 b3c2 a5a4 c4c5 f6h5 e5f3 c8d8 e1d1 g7g6 e3e4 h5g3 h2g3');
INSERT INTO evaluations VALUES(27912,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(27912,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(27912,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(27912,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(27912,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(27912,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(27912,36,'cp 137','d4d5 e6d5 c4d5 b7d5 c3d5 g8f6 e4e5 f6d5 d1d5 c7c6 d5d6 h7h6 c1f4 b8a6 e1g1 a6c5 f1e1 c5e6 f4g3 d8e7 d6d2 e8g8 a1d1 a8d8 c2c3 f8e8 g3h4 g6g5 h4g3 f7f5 e5f6 e7f6 e1e3 f6f5 d1e1 f5d5 d2d5 c6d5 h2h4 g8f7');
INSERT INTO evaluations VALUES(27907,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(27907,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(27907,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(27907,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(27907,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(27907,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(27907,36,'cp 104','e6d5 c3d5 g8f6 e4e5 f6d5 c4d5 c7c6 d5b3 e8g8 e1g1 c6c5 c1f4 d8c7 b3c4 b8c6 f1e1 c6a5 c4d5 b7d5 d1d5 c7c6 a1d1 a8d8 c2c3 f8e8 d5d3 h7h6 f4g3 c6a4 d3b1 a5c6 h2h3 a4c4 e1e4 c4b5 d1d6 e8e6 a2a4 b5a6 d6d5 c6e7 d5d2 d8e8 h3h4 e7f5');
INSERT INTO evaluations VALUES(15197,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(15197,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(15197,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(15197,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(15197,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(15197,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(15197,36,'cp -12','d2d3 f8d6 b1c3 h7h6 e2e4 d5e4 d3e4 e8g8 c1e3 f6g4 e3d2 f8e8 d1e2 g4f6 d2e3 c8g4 f1d1 d8c8 e2f1 d6b4 c3d5 f6d5 d1d5 g4e6 c2c3 b4a5 d5d2 a5b6 f1e1 e8d8 d2d8 c8d8 a1d1 d8e7 a2a3 e6b3 d1d2 a8d8 e3b6 c7b6');
INSERT INTO evaluations VALUES(24132,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(24132,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(24132,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(24132,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24132,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24132,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24132,36,'cp 23','h2h3 g4e6 c3d5 h7h6 c2c3 e6d5 e4d5 c6e7 c4b5 c7c6 d5c6 b7c6 d3d4 c5d4 c3d4 c6b5 d4e5 d6e5 f3e5 e8g8 a2a4 d8d1 f1d1 a7a6 a4b5 a6b5 c1e3 a8a1 d1a1 f6d5 e3c5 f8d8 e5d3 e7c6 a1c1');
INSERT INTO evaluations VALUES(24131,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(24131,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(24131,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(24131,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(24131,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(24131,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24131,36,'cp 26','g4h5 c3d5 f6d5 c4d5 e8g8 c1e3 c5e3 f2e3 g8h8 d1d2 a7a5 a2a3 d8d7 f3h4 f7f6 b2b4 f8b8 d3d4 c6e7 d5b3 a5b4 a3b4 h5e8 c2c4 a8a1 f1a1 e8f7 d2d3 e5d4 e3d4 b7b5 c4b5 f7b3 d3b3 d6d5 e4e5 f6e5 b3g3 d7b5 g3e5 b5b4');
INSERT INTO evaluations VALUES(24129,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(24129,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(24129,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(24129,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24129,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24129,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(24129,36,'cp 27','c3d5 f6d5 c4d5 a7a5 c1e3 c5e3 f2e3 e8g8 d1d2 g8h8 b2b3 d8c8 d5c6 b7c6 a2a4 f7f6 f3h4 d6d5 h4f5 c8d7 f5g3 h5g6 a1d1 h7h6 d2f2 f8e8 f2f3 a8d8 f1f2 g6h7 d1f1 d8a8 f3h5 a8d8 g1h2 d5e4 g3e4');
INSERT INTO evaluations VALUES(54861,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(54861,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(54861,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(54861,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(54861,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(54861,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(54861,36,'cp 64','c2c4 e7e6 g1f3 g8f6 f1e2 b7b6 e1g1 c8b7 d1c2 b8d7 f1d1 a7a6 b2b3 f8d6 b1d2 e8g8 c1b2 a8c8 f3e5 d8e7 a1c1 c7c5 e5d7 f6d7 d4c5 d6c5 c2b1 c5a3 b2a3 e7a3 d2f3 d5c4 d1d7 c4b3 c1c8 f8c8 b1b3');
INSERT INTO evaluations VALUES(54846,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(54846,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(54846,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(54846,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(54846,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(54846,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54846,36,'cp 44','e7e6 f1e2 b7b6 g1f3 c8b7 e1g1 g8f6 b2b3 b8d7 b1c3 f8e7 c1b2 e8g8 a1c1 a8c8 d1c2 c7c5 f1d1 c5d4 f3d4 a7a6 c2b1 d5c4 e2c4 e6e5 d4f3 b6b5 c4d5 d8b6 d5b7 b6b7 h2h3 f8e8 f3d2 e8d8 c3e4 c8c1 d1c1 f6e4 d2e4 d7f6 e4c5 b7c6 c5d3');
INSERT INTO evaluations VALUES(55375,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(55375,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(55375,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(55375,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(55375,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(55375,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(55375,36,'cp 59','g1f3 g8f6 f1e2 b7b6 e1g1 c8b7 b2b3 b8d7 b1c3 f8d6 c1b2 e8g8 a1c1 c7c6 d1c2 d8e7 f1d1 a8d8 e2f1 f8e8 g2g3 d6a3 b2a3 e7a3 f1g2 a3e7 c2b2 d8c8 h2h4 h7h6 c4d5 e6d5 c1c2 b7a6 a2a4 f6e4 a4a5');
INSERT INTO evaluations VALUES(19948,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19948,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19948,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19948,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19948,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19948,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19948,36,'cp 0','g1f3 g8f6 c2c3 e7e6 e1g1 f8e7 f1e1 e8g8 b1d2 c6a5 c4b5 c8d7 b5d7 d8d7 a2a4 a5c6 h2h3 h7h6 d1e2 f8d8 d2c4 d6d5 c4e5 c6e5 f3e5 d7c7 e5g4 f6g4 e2g4 g8h7 g4f3 h7g8');
INSERT INTO evaluations VALUES(25602,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(25602,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(25602,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(25602,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(25602,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(25602,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(25602,36,'cp 39','e1g1 g8f6 d2d3 f8c5 a2a4 e8g8 b1c3 c6a5 a4b5 a6b5 f3e5 a5b3 a1a8 b7a8 c2b3 b5b4 c3a4 c5e7 e5f3 c7c5 c1e3 d7d6 f3d2 f8e8 f2f3 f6d7 f1e1 d7f8 d3d4 c5d4 e3d4 d6d5 e4d5 d8d5 d4f2 a8c6 d1c2 e7d6 e1e8 c6e8 d2c4 d6c7 c2d2 d5d2 c4d2 e8c6 f2e3');
INSERT INTO evaluations VALUES(25601,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(25601,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25601,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(25601,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(25601,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(25601,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(25601,36,'cp 46','g8f6 d2d3 f8e7 c2c4 d7d6 b1c3 b5c4 b3c4 e8g8 a2a3 c6d4 f3d4 e5d4 c3d5 f6d5 e4d5 d8d7 f1e1 a6a5 c1d2 e7f6 b2b4 a5b4 a3b4 a8a1 d1a1 f8a8 a1d1 h7h6 d1f3 d7d8 h2h3 f6e5 d2f4 d8f6 f4e5 f6f3 g2f3 d6e5 e1e5');
INSERT INTO evaluations VALUES(13508,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(13508,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(13508,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(13508,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(13508,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(13508,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(13508,36,'cp -50','e6d5 c4b3 f8e7 e1g1 a7a6 f1e1 e8g8 h2h3 b7b5 a2a4 b5b4 c3e2 h7h6 c1f4 f8e8 c2c3 e7d6 f4d6 d8d6 a4a5 c8f5 b3a4 a8b8 d1d2 f5h7 a1d1 e8d8 a4c6 d6c6 e2g3 c6c7 d3d4 c5c4 d1a1 f6e4 g3e4 h7e4');
INSERT INTO evaluations VALUES(13736,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(13736,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(13736,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(13736,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(13736,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(13736,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(13736,36,'cp -34','c4b3 f8e7 e1g1 a7a6 f1e1 e8g8 h2h3 b7b5 a2a3 c5c4 b3a2 c8f5 c1g5 h7h6 g5h4 a8c8 d1d2 g7g5 f3g5 d5d4 g5f3 d4c3 d2h6 f6h7 b2c3 d8d6 h6d6 e7d6 d3c4 f5c2 c4b5 a6b5 a2b1 c2a4 b1d3 c8e8 e1e8 f8e8 a1b1 e8b8 c3c4 b5b4 a3b4 c6b4 d3e4 d6f4 e4f5 h7f8');
INSERT INTO evaluations VALUES(13762,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(13762,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(13762,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(13762,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(13762,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(13762,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(13762,36,'cp -41','f8e7 h2h3 b7b5 a2a3 e8g8 e1g1 a7a6 c3e2 c5c4 b3a2 f8e8 a2b1 c8f5 c2c3 f5g6 c1g5 d8c7 g5f4 c7b7 b1c2 c4d3 c2d3 f6e4 f4e3 e7d6 b2b4 g6h5 a3a4 b5a4 a1a4 b7d7 b4b5 c6e7 a4e4 d5e4 d3e4 a8d8 b5a6 d6h2 g1h2 d7d1 f1d1 d8d1');
INSERT INTO evaluations VALUES(8121,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(8121,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(8121,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(8121,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(8121,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(8121,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(8121,36,'cp -34','e1g1 a7a6 h2h3 b7b5 a2a3 e8g8 f1e1 h7h6 c1f4 c5c4 b3a2 f8e8 c3e2 c8f5 a2b1 e7c5 c2c3 d8d7 d3d4 c5b6 b1f5 d7f5 f4h2 c6a5 e2g3 f5h7 d1b1 e8e1 f3e1 g7g6 b1c2 a5b3 a1d1 a8e8 e1f3 b6c7 g3f1 c7h2 g1h2 g8g7 d1e1 e8e1');
INSERT INTO evaluations VALUES(56001,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(56001,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(56001,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(56001,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(56001,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(56001,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(56001,36,'cp 40','a7a6 d2d4 b7b5 c4d3 c8b7 d1e2 f8e7 c3e4 g8f6 e4f6 e7f6 a2a4 b8d7 e1g1 b5b4 f1d1 e8g8 c1f4 c7c5 f4d6 c5d4 d6f8 d8f8 d3e4 b7e4 e2e4 d7c5 e4e2 a8d8 f3e5 f8d6 e5d3 a6a5 b2b3 d6d5 d3c5 d5c5 a1c1 c5c7 g2g3');
INSERT INTO evaluations VALUES(50973,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(50973,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(50973,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(50973,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(50973,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(50973,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(50973,36,'cp 52','d2d4 b7b5 c4d3 c8b7 c3e4 b8d7 d1e2 f8e7 a2a4 g8f6 e4f6 e7f6 e1g1 b5b4 f1d1 e8g8 c1f4 h7h6 c2c3 c7c5 f4d6 b4c3 b2c3 f8e8 d4c5 d8c8 d3c2 d7c5 a1b1 f6e7 d6e5 c5d7 e5d4 e7f6 e2d3 g7g6 d4f6 d7f6 f3e5 c8c7 d1e1 g8g7 d3g3 a8c8 c3c4');
INSERT INTO evaluations VALUES(52180,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52180,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(52180,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52180,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52180,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(52180,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52180,36,'cp 14','g1f3 b8c6 c3d4 e5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4e7 e2f4 d5f4 c1f4 c8e6 f4e5 e6d5 d3a6 h7h5 e5c3 f8b8 f1c1 h5h4 b2b3 e7c5 c1e1 b8e8 e1e8 a8e8 a1d1 c5b6 g1f1');
INSERT INTO evaluations VALUES(11635,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(11635,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(11635,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(11635,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(11635,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(11635,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11635,36,'cp 34','a2a4 c8d7 f3g5 d6d5 a4b5 a6b5 a1a8 d8a8 b3d5 f6d5 e4d5 c6d4 c2c3 d4f5 d2d4 a8d5 d4e5 d7c6 d1d5 c6d5 f1d1 d5b3 d1e1 f8e7 g5h3 h7h6 b1d2 b3a2 h3f4 e8g8 d2e4 f8a8 g2g3 a2c4 g1g2 e7g5 h2h4 g5f4 c1f4 a8a2 f4c1 h6h5 e4c5 c7c6');
INSERT INTO evaluations VALUES(50314,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(50314,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(50314,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(50314,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(50314,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(50314,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(50314,36,'cp -28','d7d5 g1f3 e7e5 f3e5 f8b4 e2e3 e8g8 f1e2 b8d7 e1g1 d7e5 b2e5 f8e8 d2d4 f6e4 c2c4 c7c5 c4d5 b4c3 b1c3 e4c3 d1d3 c3e2 d3e2 d8d5 a1d1 c8e6 e3e4 d5a2 e2f3 e6c4 f1e1 f7f6 e5d6 c5d4 d1d4 c4b5 e4e5 b5c6 f3g3');
INSERT INTO evaluations VALUES(54807,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(54807,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(54807,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(54807,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(54807,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(54807,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(54807,36,'cp 39','g8f6 e4e5 d6e5 f3e5 e7e6 d2d4 f8g7 e1g1 b8d7 b1c3 a7a6 d1e2 e8g8 c1g5 h7h6 g5h4 d7b6 c3e4 g6g5 e4f6 g7f6 h4g3 b6c4 e2c4 b7b5 c4e2 c8b7 a1d1 d8d5 f2f3 c7c5 d4c5 d5c5 f1f2 f8d8 e5d7 c5e7 d7f6 e7f6 d1d8 a8d8 c2c3 f6g6 h2h3 b7c6');
INSERT INTO evaluations VALUES(46646,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46646,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(46646,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(46646,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(46646,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46646,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(46646,36,'cp 54','e4e5 d6e5 f3e5 e7e6 d2d4 f8g7 b1c3 e8g8 e1g1 f6d7 c1f4 d7e5 d4e5 b8c6 f1e1 d8d1 a1d1 b7b6 c4d3 c8b7 d3e4 a7a6 f4g5 h7h6 g5f6 g7f6 e5f6 f8d8 f2f4 c6a5 e4b7 a5b7 g2g4 b7d6 g4g5 h6g5 f4g5 d6c4 b2b3 c4a3 d1d8 a8d8 c3e4 a3c2 e1c1');
INSERT INTO evaluations VALUES(19922,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19922,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19922,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19922,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19922,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19922,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(19922,36,'cp 19','g8f6 d1e2 c8g4 d2d4 c5d4 c3d4 d6d5 e4d5 f6d5 e1g1 e7e6 b1c3 f8e7 h2h3 g4f3 e2f3 a7a6 b5a4 b7b5 c3d5 d8d5 f3d5 e6d5 a4b3 c6d4 b3d5 a8d8 d5b7 e8g8 c1e3 d4c2 a1d1 c2e3 f2e3 d8d1 f1d1 e7f6 b7a6 f6b2 a6b5');
INSERT INTO evaluations VALUES(13219,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13219,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(13219,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(13219,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(13219,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(13219,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(13219,36,'cp 19','d1e2 c8g4 d2d4 c5d4 c3d4 d6d5 e4d5 d8d5 b1c3 g4f3 c3d5 f3e2 d5f6 e7f6 e1e2 e8c8 b5c6 b7c6 c1e3 d8d5 h1c1 c8d7 c1c4 d5a5 e3d2 a5a6 d2b4 f8d6 b4d6 d7d6 c4b4 a6a5 b4b7 h8e8 e2d3 e8e7 b7e7 d6e7 g2g3 a5b5 b2b3 e7d6 h2h4 b5f5 d3e3 f5d5 a1c1 d5a5 a2a4');
INSERT INTO evaluations VALUES(43129,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(43129,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(43129,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(43129,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43129,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(43129,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(43129,36,'cp 14','b1c3 h7h6 g5d2 c7c5 d4d5 e7e6 e2e4 e6d5 e4d5 e8g8 f1d3 d7d6 h2h3 f8e8 g1e2 b8d7 f2f4 b7b5 c3b5 c8a6 a2a4 f6e4 e1g1 a6b5 a4b5 g7b2 a1a2 e4d2 d1d2 b2f6 g1h2 d8e7 e2g3 f6d4 d2e2 e7e2 g3e2 e8e3 f1f3 e3f3 g2f3');
INSERT INTO evaluations VALUES(52256,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(52256,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(52256,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(52256,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(52256,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(52256,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(52256,36,'cp 76','c2c3 f8g7 f2f4 g8h6 g1f3 d7d5 e4e5 f7f6 b1d2 e8g8 d2b3 c8g4 f1e2 b8d7 e1g1 e7e6 a2a4 h6f5 a4a5 d8c7 e5f6 g7f6 f3e5 g4e2 d1e2 a8e8 c1e3 b7b6 e3f2 d7e5 d4e5 f6g7 g2g4 f5e7 a5b6 a7b6 b3d4');
INSERT INTO evaluations VALUES(49726,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(49726,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(49726,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(49726,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(49726,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49726,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(49726,36,'cp -6','b1c3 b8c6 g1f3 d7d6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5c4 a8c8 a1c1 f8e8 c4a5 b7a8 c3a4 c8d8 g3c7 d8d1 c1d1 f6d5 c7g3 e7f6 g1f1 d5b6 a4b6 a7b6 a5c4 c6c5 c4b6 f6b2 g3c7 b2d4 b6a8 e8a8');
INSERT INTO evaluations VALUES(49724,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(49724,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(49724,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(49724,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(49724,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49724,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(49724,36,'cp -11','b8c6 g1f3 d7d6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5c4 a8c8 a1c1 b7a8 c4e3 f8e8 g3e5 g8h7 e3f5 e7f8 e5f6 h7g6 f6d4 c6c5 d4e3 g6f5 c3a4 a8c6 a4c5 c8d8 d1d8 e8d8');
INSERT INTO evaluations VALUES(16637,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(16637,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16637,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(16637,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(16637,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16637,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(16637,36,'cp -8','g1f3 d7d6 d1b3 d8d7 b3c2 f8e7 e1g1 e8g8 f1d1 d7e8 h2h3 a7a6 a2a3 c8e6 c3d5 e7d8 c1d2 e6d5 e4d5 c6e5 f3e5 e8e5 d1e1 e5h5 a1d1 f8e8 d2c3 d8e7 c4e2 h5g6 e2d3 g6h5');
INSERT INTO evaluations VALUES(16636,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(16636,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(16636,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(16636,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(16636,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(16636,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(16636,36,'cp -2','d7d6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5c4 a8c8 a1c1 f8e8 c4a5 b7a8 c3a4 c8d8 g3c7 d8d1 c1d1 f6d5 c7g3 c6c5 d1e1 e8d8 f2f3 e7f6 a5c4 d5b6 a4b6 a7b6 g3e5');
INSERT INTO evaluations VALUES(7332,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7332,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(7332,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(7332,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7332,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(7332,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7332,36,'cp 16','b7b5 a4b3 d7d6 c2c3 e8g8 c1g5 h7h6 g5h4 c8e6 f1e1 e6b3 a2b3 g7g5 h4g3 d8d7 b1a3 c5b6 a3c2 f8e8 h2h3 c6e7 b3b4 e7g6 f3h2 h6h5 h2f1 g5g4 f1e3 h5h4 g3h2 g4h3 d1f3 f6h7 e3f5 h3g2 g1g2 g8h8 e1g1 e8g8');
INSERT INTO evaluations VALUES(7465,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(7465,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(7465,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(7465,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7465,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(7465,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(7465,36,'cp 3','a4b3 d7d6 c2c3 e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 c8e6 f1e1 c5b6 b1d2 e6b3 d1b3 d8d7 d2f1 a8e8 a2a4 c6e7 a4b5 a6b5 d3d4 e7g6 b3c2 d7c6 a1d1 b5b4 h2h3 b4c3 b2c3 e5d4 f3d4 b6d4 d1d4 e8e6 g3h2 d6d5 e4d5 f6d5 e1e6 c6e6');
INSERT INTO evaluations VALUES(14484,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(14484,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(14484,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(14484,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(14484,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(14484,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(14484,36,'cp 95','d4d5 c6b8 h2h3 e7e5 c1e3 f8e7 d1d2 a7a6 a2a4 b8d7 g2g4 h7h6 g1e2 f6h7 e2g3 e7g5 f1e2 g5e3 d2e3 d8g5 c3d1 a6a5 h3h4 g5f4 g3h5 f4e3 d1e3 g7g6 h5g3 d7c5 f2f3 c8d7 b2b3 e8e7 e1d2 h7f6 e2d3 h6h5 g4g5 f6e8 g3e2 f7f6');
INSERT INTO evaluations VALUES(14464,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(14464,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(14464,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(14464,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(14464,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(14464,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(14464,36,'cp 93','c6b8 h2h3 e7e5 c1e3 f8e7 d1d2 h7h6 a2a4 c7c6 g2g4 b8a6 g1e2 c8d7 f1g2 a6b4 a4a5 f6h7 e2g3 e8g8 e1g1 c6c5 g3f5 d7f5 e4f5 e7h4 c3e4 h7f6 e4f6 d8f6 g2e4 h4g5 e3g5 f6g5 d2g5 h6g5 g1g2 a8c8 h3h4 g5h4 g4g5');
INSERT INTO evaluations VALUES(21596,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21596,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21596,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(21596,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(21596,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(21596,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21596,36,'cp 14','g1f3 g8f6 f1b5 f8b4 e1g1 e8g8 c3e2 c6e7 c2c3 b4d6 e2g3 e7g6 f1e1 a7a5 b5d3 c7c6 c1g5 h7h6 g5d2 a5a4 d1c2 g6f4 d3f1 f4h5 g3h5 f6h5 c3c4 c8e6 c4c5 d6c7 b2b4 a4b3 a2b3 a8a1 e1a1 f8e8');
INSERT INTO evaluations VALUES(38048,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(38048,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(38048,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(38048,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(38048,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(38048,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(38048,36,'cp 73','d1b3 a7a5 a2a3 b4c3 b3c3 b8d7 c1f4 g8e7 e2e3 e8g8 c4c5 b7b6 b2b4 d7f6 f3d2 f6e4 d2e4 d5e4 f4d6 a5b4 a3b4 e7d5 c3b2 a8a1 b2a1 f8e8 f1e2 d5b4 e1g1 b6c5 d4c5 b4d5 a1b2 d8g5 f1a1 h7h6 a1a8 g8h7 b2c2 e6e5 e2c4 g5h5 a8a7 c8f5');
INSERT INTO evaluations VALUES(52445,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(52445,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(52445,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(52445,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(52445,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(52445,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(52445,36,'cp -55','e7e6 d2d3 b8c6 f3f4 d7d5 g1f3 f8e7 f1e2 g8h6 e1g1 e8g8 c2c3 b7b6 b1a3 c8b7 a3c2 g8h8 c2e3 d5d4 e3c4 b6b5 c4e5 c6e5 f4e5 d4c3 b2c3 a7a5 g1h1 b5b4 c1h6 g7h6 d1e1 c5c4 a1d1 c4d3 e2d3 f8g8 c3b4');
INSERT INTO evaluations VALUES(51998,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(51998,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(51998,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(51998,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(51998,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(51998,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(51998,36,'cp -48','d2d4 c5d4 d1d4 b8c6 d4f2 g8f6 b1c3 d7d5 e4d5 f6d5 c3d5 d8d5 c1e3 f8e7 g1e2 d5a5 c2c3 e8g8 e2d4 f8d8 d4b3 a5a4 f2e2 c6e5 e3d4 e5g6 e2b5 a4b5 f1b5 e6e5 d4e3 g6f4 e3f4 e5f4 b3d4 e7c5 e1c1 c5d4 d1d4 d8d4 c3d4 c8e6 b5d3 e6a2');
INSERT INTO evaluations VALUES(19005,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19005,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19005,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(19005,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(19005,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19005,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19005,36,'cp -9','d7d6 c3d5 g8f6 c2c4 f8e7 f1e2 a8b8 e1g1 e8g8 c1d2 c8d7 a2a3 b7b5 b2b4 b5c4 d5f6 e7f6 d3c4 c5b4 a3b4 c6b4 d2b4 b8b4 d1d6 b4b2 f1d1 d7e6 e2f1 d8d6 d1d6 f8c8 a1a7 b2b1 c4c5 c8c5 a7a8 c5c8 a8c8 e6c8 f3e5 f6e7 d6c6 c8e6 f2f4 b1a1');
INSERT INTO evaluations VALUES(20615,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(20615,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(20615,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20615,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20615,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20615,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20615,36,'cp 0','c3d5 g8f6 c2c4 f8e7 f1e2 e8g8 e1g1 a7a6 h2h3 b7b5 b2b3 c8d7 a2a3 h7h6 c1d2 a8b8 a1b1 b8c8 b1c1 c8b8');
INSERT INTO evaluations VALUES(10222,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(10222,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(10222,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(10222,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(10222,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(10222,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(10222,36,'cp -27','c5b4 f3d2 d7d5 c3d5 b4d2 d1d2 f6d5 e4d5 d8d5 f2f3 c8e6 f1e2 e8c8 e1g1 f7f5 f3f4 g7g5 e2f3 d5d7 f4e5 f5f4 e3f2 c6e5 f3e4 e5c6 a1e1 e6d5 a2a4 h8e8 e4d5 d7d5 e1e8 d8e8 f1e1 e8e1 d2e1 b7b6 b2b3 g5g4 e1e8 c8b7 e8e4 d5e4 d3e4 c6b4 c2c3 b4d3 f2d4 f4f3 g2f3 g4f3');
INSERT INTO evaluations VALUES(47738,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(47738,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(47738,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(47738,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(47738,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(47738,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(47738,36,'cp 144','f8d6 f1e2 b8c6 e1g1 c8d7 e2d3 e6e5 d4d5 c6e7 a2a4 d7g4 h2h3 g4f3 d1f3 b6d7 a4a5 e8g8 c1e3 a7a6 d3c2 g8h8 c3e2 b7b6 a5b6 c7b6 f1d1 e7g8 c2d3 d7c5 e3c5 d6c5 a1a6 d8f6 e2c3 a8a6 f3f6 g8f6 d3a6 f8a8');
INSERT INTO evaluations VALUES(48376,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(48376,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(48376,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(48376,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(48376,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(48376,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(48376,36,'cp -59','c7c5 e4d5 f6d5 c3d5 e6d5 g1e2 b8c6 c1e3 c5c4 d1d2 f8d6 e2c3 c6e7 f1e2 c8e6 e3g5 e8g8 e1g1 f7f6 g5f4 d6b4 a2a3 b4a5 f4g3 d8d7 b2b4 a5b6 g3f2 a7a5 f3f4 a5b4 a3b4 f8e8 a1a8 e8a8 e2f3 e6g4 d2e3 a8e8 h2h3 g4f5 e3d2 h7h6');
INSERT INTO evaluations VALUES(46262,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(46262,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(46262,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(46262,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(46262,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(46262,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(46262,36,'cp -58','e4d5 f6d5 c3d5 e6d5 g1e2 b8c6 c1e3 c5c4 d1d2 b7b5 e2f4 f8e7 a2a3 c8f5 f1e2 e8g8 e1g1 f8e8 c2c3 c6a5 e2d1 a5b3 d1b3 c4b3 f1e1 d8d7 f4d3 f5d3 d2d3 a7a5 e3f4 a5a4 g1f2 a8c8 e1e5 e7f6 e5e8 c8e8 a1e1 e8e1 f2e1 h7h5');
INSERT INTO evaluations VALUES(6516,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(6516,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6516,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6516,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6516,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6516,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(6516,36,'cp 18','b1c3 c5b6 f1c4 d7d6 d2d3 g8f6 c3a4 h7h6 e1g1 e8g8 a4b6 a7b6 h2h3 c6e7 f1e1 c8e6 c4e6 f7e6 a3a4 f8f7 c2c3 e7g6 d3d4 f6d7 d4e5 g6e5 f3e5 d7e5 f2f4 e5d7 c1e3 d8h4 e1f1 a8f8 d1f3 d7c5 a4a5 b6a5 a1a5');
INSERT INTO evaluations VALUES(53657,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(53657,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(53657,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(53657,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(53657,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(53657,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(53657,36,'cp 28','a2a4 f8b4 e2e3 b7b5 c1d2 a7a5 a4b5 b4c3 d2c3 c6b5 b2b3 c8b7 b3c4 b5b4 c3b2 g8f6 f1d3 e8g8 e1g1 b8d7 f3d2 d8c7 d3c2 f8d8 f1e1 d7b6 c4c5 b6d7 h2h3 h7h6 c2a4 f6e4 a1c1 b7d5 d2f1 d7b8 f2f3 e4g3 e3e4 d5c6 a4b3 a5a4');
INSERT INTO evaluations VALUES(53670,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(53670,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53670,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(53670,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(53670,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(53670,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(53670,36,'cp 24','g8f6 e2e3 c6c5 f1c4 c5d4 e3d4 b8c6 e1g1 f8e7 c1e3 e8g8 d4d5 e6d5 c3d5 f6d5 c4d5 c8f5 d1b3 c6b4 f1d1 d8a5 f3e5 b4d5 d1d5 a5a6 e5d7 f5e6 d7f8 a8f8 a1d1 f8c8 h2h3 e6d5 d1d5 c8d8 d5d2 d8d2 e3d2 b7b6 d2c3');
INSERT INTO evaluations VALUES(45719,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(45719,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45719,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(45719,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(45719,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(45719,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(45719,36,'cp 32','e2e3 c6c5 f1c4 c5d4 e3d4 f8e7 d1e2 b8c6 e1g1 e8g8 f1d1 c6b4 a4a5 c8d7 f3e5 d7e8 c1f4 a7a6 e2f3 f6d5 f4d2 a8c8 c3d5 b4d5 f3b3 c8c7 a1c1 d5f6 d2f4 e7d6 g2g3 h7h6 c4e2 f6d5 f4d2 c7c1 d1c1 f7f6 e5c4 e8c6');
INSERT INTO evaluations VALUES(6733,30,'cp 370',NULL);
INSERT INTO evaluations VALUES(6733,31,'cp 359',NULL);
INSERT INTO evaluations VALUES(6733,32,'cp 337',NULL);
INSERT INTO evaluations VALUES(6733,33,'cp 352',NULL);
INSERT INTO evaluations VALUES(6733,34,'cp 344',NULL);
INSERT INTO evaluations VALUES(6733,35,'cp 385',NULL);
INSERT INTO evaluations VALUES(6733,36,'cp 367','d4d5 c6e7 f3e5 a5b6 e5c4 d7d6 a2a4 g8f6 a4a5 b6f2 e1f2 f6e4 f2g1 f7f5 d1h5 e7g6 b1d2 e4f6 h5e2 e8f7 d2f3 h8e8 e2f2 h7h6 f1d3 d8e7 h2h3 c8d7 g1h2 f7g8 f3d4 f6e4 d3e4 e7e4 f2f3 e4f3 d4f3 e8e2 f3d4 a8e8 d4e2 e8e2 h1f1 e2e4 c4e3 f5f4 e3g4');
INSERT INTO evaluations VALUES(6721,30,'cp 371',NULL);
INSERT INTO evaluations VALUES(6721,31,'cp 364',NULL);
INSERT INTO evaluations VALUES(6721,32,'cp 359',NULL);
INSERT INTO evaluations VALUES(6721,33,'cp 386',NULL);
INSERT INTO evaluations VALUES(6721,34,'cp 378',NULL);
INSERT INTO evaluations VALUES(6721,35,'cp 359',NULL);
INSERT INTO evaluations VALUES(6721,36,'cp 348','c6e7 f3e5 a5b6 e5c4 d7d6 a2a4 g8f6 a4a5 b6f2 e1f2 f6e4 f2g1 f7f5 b1d2 e4f6 c4e3 e8g8 d2f3 e7g6 e3c2 f5f4 c2d4 d8e7 f1d3 c8d7 d1e2 e7e2 d3e2 a8e8 c1d2 g6e5 d2f4 e5f3 e2f3 d7g4 f3g4 f6g4 h2h3');
INSERT INTO evaluations VALUES(7014,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(7014,31,'cp 367',NULL);
INSERT INTO evaluations VALUES(7014,32,'cp 359',NULL);
INSERT INTO evaluations VALUES(7014,33,'cp 365',NULL);
INSERT INTO evaluations VALUES(7014,34,'cp 387',NULL);
INSERT INTO evaluations VALUES(7014,35,'cp 390',NULL);
INSERT INTO evaluations VALUES(7014,36,'cp 367','f3e5 a5b6 e5c4 d7d6 a2a4 g8f6 a4a5 b6f2 e1f2 f6e4 f2g1 f7f5 d1h5 e7g6 b1d2 e4f6 h5e2 e8f7 d2f3 h7h6 f3d4 h8e8 e2f2 f6e4 f2c2 f7g8 c1e3 d8f6 c4d2 c8d7 a1e1 g6e5 d4f3 e4d2 f3e5 d6e5 e3d2 f5f4 c3c4 e5e4 e1e4 f4f3 e4e8 a8e8 g2f3 f6f3 f1g2');
INSERT INTO evaluations VALUES(7013,30,'cp 371',NULL);
INSERT INTO evaluations VALUES(7013,31,'cp 386',NULL);
INSERT INTO evaluations VALUES(7013,32,'cp 383',NULL);
INSERT INTO evaluations VALUES(7013,33,'cp 353',NULL);
INSERT INTO evaluations VALUES(7013,34,'cp 358',NULL);
INSERT INTO evaluations VALUES(7013,35,'cp 369',NULL);
INSERT INTO evaluations VALUES(7013,36,'cp 363','a5b6 e5c4 d7d6 a2a4 g8f6 a4a5 b6c5 b2b4 c5f2 e1f2 f6e4 f2g1 f7f5 d1d4 e8g8 b1d2 e4f6 d2f3 e7d5 c4e3 c7c5 b4c5 d6c5 d4c5 c8e6 e3d5 e6d5 c5d4 a8c8 c3c4 d5f7 c1a3 f6e4 a3f8 d8f8 f1d3 f8c5 a1e1 f7c4 d4c5 c8c5 d3e4');
INSERT INTO evaluations VALUES(9285,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(9285,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(9285,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(9285,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(9285,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(9285,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(9285,36,'cp 55','d2d4 e5d4 f3d4 c8d7 d4f5 e8g8 c1g5 d7e6 c3a4 e6f5 e4f5 c5b6 a4b6 a7b6 a2a4 d6d5 g5h4 f8e8 c2c3 d5d4 a1c1 d4c3 c1c3 d8d1 f1d1 e8d8 d1d8 a8d8 g2g3 d8d1 g1g2 f6d5 c3d3 d1d3 b5d3 c6d4 d3c4 c7c6 g3g4 b6b5 a4b5 d4b5 c4d5 c6d5 g2f3 b5d6 h4g3');
INSERT INTO evaluations VALUES(9281,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(9281,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(9281,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(9281,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(9281,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(9281,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(9281,36,'cp 58','e5d4 f3d4 c8d7 d4f5 e8g8 c1g5 d7e6 c3a4 e6f5 e4f5 c5b6 a4b6 a7b6 a2a4 d6d5 c2c3 c6a7 b5d3 h7h6 g5h4 f8e8 h2h3 c7c5 d1c2 d8d6 a1e1 a7c6 e1e8 a8e8 f1d1 d6e5 h4g3 e5e7 d3b5 e8d8 f2f3 e7e3 c2f2 e3e7 g3f4 e7d7 f2g3 d7f5 f4h6');
INSERT INTO evaluations VALUES(9303,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(9303,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(9303,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(9303,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(9303,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(9303,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(9303,36,'cp 62','f3d4 c8d7 d4f5 e8g8 c1g5 d7e6 c3a4 e6f5 e4f5 c5b6 a4b6 a7b6 a2a4 c6a7 b5d3 f8e8 f1e1 e8e1 d1e1 a7c6 e1d2 c6e5 b2b3 e5d3 d2d3 d8e7 a1d1 a8e8 h2h3 e7e2 g5f6 g7f6 c2c4 e2d3 d1d3 e8e1 g1h2 e1e2 h2g3 e2b2 g3f3 h7h5 g2g4');
INSERT INTO evaluations VALUES(9301,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(9301,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(9301,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(9301,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(9301,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(9301,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(9301,36,'cp 59','c8d7 d4f5 e8g8 c1g5 d7e6 c3a4 e6f5 e4f5 c5b6 a4b6 a7b6 a2a4 d6d5 c2c3 c6a7 b5d3 h7h6 g5h4 f8e8 d1b3 a7c6 f1d1 d5d4 d3c4 d8e7 c3d4 c6a5 b3a2 a5c4 a2c4 a8a5 d4d5 a5c5 c4d4 e8d8 d4b4 c5d5 d1d5 e7b4 d5d8 g8h7 h4f6 g7f6');
INSERT INTO evaluations VALUES(24179,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(24179,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(24179,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(24179,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(24179,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(24179,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(24179,36,'cp 49','d4f5 e8g8 c1g5 d7e6 c3a4 e6f5 e4f5 c5b6 a4b6 a7b6 a2a4 d6d5 c2c3 c6a7 b5e2 f8e8 d1c2 a7c8 f1d1 c7c6 c3c4 h7h6 g5e3 c8d6 b2b3 e8e5 e2d3 e5e3 f2e3 f6g4 c2e2 d8g5 c4d5 c6d5 d1e1 a8e8 a1d1 g4e3');
INSERT INTO evaluations VALUES(50316,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(50316,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(50316,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(50316,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(50316,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(50316,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(50316,36,'cp -110','f6e4 c4f7 e8f7 d1h5 f7g8 h5d5 e7e6 d5e4 d7d5 e4e3 e6e5 d2d4 e5d4 e3d4 c7c5 d4d1 c8f5 g1e2 b8c6 e1g1 h7h6 e2g3 f5g6 f2f4 g6f7 b1d2 h6h5 h2h3 d8d7 d2f3 a8e8 f1e1 h5h4 g3f1 e8e1 d1e1 h8h6 c1d2 f7h5 f4f5 h5f3 d2h6 g7h6 f1d2 f3e4 d2e4 d5e4 e1e4 d7d4 e4d4 c6d4');
INSERT INTO evaluations VALUES(50438,30,'cp -116',NULL);
INSERT INTO evaluations VALUES(50438,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(50438,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(50438,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(50438,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(50438,35,'cp -128',NULL);
INSERT INTO evaluations VALUES(50438,36,'cp -99','c4f7 e8f7 d1h5 f7g8 h5d5 e7e6 d5e4 d7d5 e4e3 e6e5 d2d4 e5d4 e3d4 c8f5 b1c3 c7c6 g1f3 f5c2 c1e3 c2g6 e1g1 b8d7 a1e1 h7h6 d4g4 g6f7 g4f5 h8h7 e3f4 g7g6 f5c2 g6g5 f4g3 d8f6 f3e5 d7e5 g3e5 f6g6 c2d2 f7e6 f2f4 a8e8 f4g5 h6g5 e5d4');
INSERT INTO evaluations VALUES(6545,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(6545,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(6545,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6545,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(6545,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(6545,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(6545,36,'cp 17','b1c3 c6d4 b2b4 d4f3 d1f3 c5d4 a1b1 g8f6 c3e2 d7d5 e2d4 d5c4 c1b2 e8g8 e1g1 f8e8 b1e1 d8d6 a2a4 c8d7 b2c3 h7h6 a4a5 a8d8 e4e5 e8e5 e1e5 d6e5 f3b7 c7c5 d4c6 d7c6 b7c6 e5e2 c6c5 f6e4 c5a7 e4c3 d2c3 e2c2 a7c7 d8d1 c7b8 g8h7 b8f4 d1f1 g1f1 c2d1 f1f2 d1c2 f2g1 c2c3 f4e4 h7h8 b4b5 c3a5 e4c4 a5e1 c4f1');
INSERT INTO evaluations VALUES(29134,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(29134,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(29134,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(29134,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(29134,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(29134,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(29134,36,'cp 63','d1b3 f5c8 c1g5 e7e6 e2e3 b8c6 f1d3 h7h6 g5f6 d8f6 e1g1 f8d6 e3e4 d5e4 c3e4 f6d8 f1e1 d6e7 d3b5 e8g8 b5c6 b7c6 b3c3 a7a5 e4c5 d8d5 a2a3 c8d7 f3e5 d7e8 a1c1 a8d8 c1d1 e7c5 c3c5 d8b8 e5c4 d5d8 h2h3');
INSERT INTO evaluations VALUES(29130,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(29130,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(29130,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(29130,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(29130,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(29130,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(29130,36,'cp 60','f5c8 c1f4 e7e6 e2e3 b8c6 f1e2 f8e7 f3e5 e8g8 e1g1 c6a5 b3d1 f6d7 e5d3 d7f6 a1c1 a5c4 b2b3 c4d6 c1c2 c8d7 d1b1 d6e4 a2a4 e4c3 c2c3 d7c6 h2h3 f6e4 c3c2 e7a3 b1a2 d8e7 f2f3 e4f6 f1a1 a3d6 f4d6 e7d6');
INSERT INTO evaluations VALUES(43545,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(43545,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(43545,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43545,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(43545,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43545,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(43545,36,'cp 38','f2f4 b8d7 a2a4 d7c5 e1g1 f8e7 c1e3 e6e5 d4e2 e8g8 g1h1 c8e6 f4f5 c5d3 d1d3 e6d7 e3g5 d7c6 g5f6 e7f6 c3d5 c6d5 d3d5 d8b6 d5d3 d6d5 e4d5 b6b2 a1b1 e5e4 d3e4 b2a2 e2f4 f8e8 e4f3 a8c8 d5d6 a2a4 f3b7');
INSERT INTO evaluations VALUES(43551,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(43551,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(43551,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(43551,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(43551,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43551,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(43551,36,'cp 17','g7g6 f4f5 f8g7 e1g1 e8g8 f5e6 c8e6 d4e6 d8b6 g1h1 f7e6 d3c4 b8c6 c4e6 g8h8 e6h3 a8e8 g2g3 f6e4 f1f8 g7f8 c3e4 e8e4 d1f1 f8g7 a1b1 e4e8 h3g2 c6e5 c1f4 e5g4 g2f3 g4f2 h1g2 f2e4 b1e1 d6d5 b2b3 b6a5 f1e2 a5c3 a2a3 e8c8 f3e4 d5e4 e2e4 c3c2 e1e2 c2e4 e2e4');
INSERT INTO evaluations VALUES(10613,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(10613,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(10613,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(10613,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(10613,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(10613,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(10613,36,'cp 112','c4f7 e8f7 f3g5 f7g8 d1g4 c6d4 g4d1 d7d6 h2h3 h7h6 g5f3 d4f3 d1f3 g8h7 c1e3 h8f8 f3e2 c5b4 e1g1 b4c3 b2c3 d8h4 f2f3 c8d7 c3c4 b7b6 g1h2 f8f7 e3f2 h4f6 f2g3 a8f8 a1e1 f6g5 f1f2 d7a4 e2d2 g5d2 f2d2 g7g5 g3f2 a4d7 e1a1 d7e6 f2g3 e6d7 a2a4');
INSERT INTO evaluations VALUES(19211,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(19211,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(19211,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19211,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19211,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19211,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19211,36,'cp 0','f8g7 d2d3 d7d6 e1g1 c8d7 f1e1 g8f6 h2h3 e8g8 a2a4 a7a6 b5c4 a8c8 c1f4 f6h5 f4h2 h5f6');
INSERT INTO evaluations VALUES(5267,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(5267,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5267,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(5267,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(5267,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(5267,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(5267,36,'cp 15','d2d3 d7d6 e4e5 d6e5 b5c6 b7c6 e1g1 d8c7 d1e2 c5c4 h2h3 c4d3 c2d3 g8h6 c1h6 g7h6 f1e1 c8e6 d3d4 a8d8 a1d1 h6g7 f3e5 e8g8 b2b3 d8d6 c3a4 f8d8 a4c5 e6d5 e2e3 a7a5 d1c1 c7a7 c1c2 d6f6 e1c1 f6d6 c5d3 a7a6 d3f4 g7h6 c2c5');
INSERT INTO evaluations VALUES(5268,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(5268,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(5268,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(5268,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5268,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(5268,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(5268,36,'cp 16','d7d6 e4e5 d6e5 b5c6 b7c6 e1g1 d8c7 d1e2 c5c4 h2h3 g8h6 c1h6 g7h6 f1e1 c4d3 c2d3 c8e6 f3e5 e8g8 d3d4 h6g7 b2b3 g7e5 e2e5 c7e5 e1e5 f8d8 a1d1 d8d7 f2f3 a8d8 e5a5 f7f6 g1f2 g8f7 d1e1 d7c7 c3e2 d8d5 a5d5 e6d5 e2f4 c7d7 e1c1');
INSERT INTO evaluations VALUES(20708,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(20708,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(20708,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20708,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(20708,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20708,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20708,36,'cp -6','e1g1 f8e7 d1e2 g8f6 f1d1 c8d7 c3b5 d8b8 c1f4 c6e5 b5d6 e7d6 d1d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 f2f4 g5a5 d6f8 d7f8 g2g3 a8d8 g1g2 g7g6 d1d8 a5d8 e2d3 d8a5 a2a3 a5c5 b2b4 c5c6 d3d4 b7b5 c4d3 c6c1 d3b5 c1a3 e4e5 a3c1');
INSERT INTO evaluations VALUES(20707,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(20707,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(20707,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(20707,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20707,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(20707,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(20707,36,'cp -2','g8f6 d1e2 f8e7 f1d1 c8d7 c3b5 d8b8 c1f4 c6e5 b5d6 e7d6 d1d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 d6f8 d7f8 g2g3 a8d8 d1d8 g5d8 e2e3 b7b6 h2h4 d8d6 e3c3 d6c5 g1g2 h7h6 f2f4 f8d7 c4b3 c5c3 b2c3 g8f8 g2f3 f8e7 e4e5 d7c5 f3e3');
INSERT INTO evaluations VALUES(13980,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(13980,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13980,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13980,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13980,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13980,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13980,36,'cp 0','d1e2 f8e7 f1d1 c8d7 c3b5 d8b8 c1f4 c6e5 b5d6 e7d6 d1d6 b8d6 f3e5 d6e7 a1d1 e8g8 e5d7 f6d7 f4d6 e7g5 d6f8 d7f8 g2g3 a8d8 d1d8 g5d8 e2e3 b7b6 h2h4 d8d6 e3c3 d6c6 c3d4 c6a4 d4d3 a4a5 d3d4');
INSERT INTO evaluations VALUES(13979,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13979,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13979,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(13979,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13979,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13979,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13979,36,'cp 0','c8d7 f1d1 h7h6 c1f4 e6e5 f4e3 f8e7 f3h4 c6a5 c4b5 a5c6');
INSERT INTO evaluations VALUES(43238,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(43238,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(43238,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43238,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43238,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43238,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43238,36,'cp 13','e3e4 e8g8 c2c3 d7d6 g1f3 e7e5 e1g1 e5d4 c3d4 c7c5 b1c3 c5d4 f3d4 d6d5 e4d5 f6d5 d3e4 d5c3 b2c3 d8a5 a1b1 b8d7 d1f3 a5a2 c1g5 h7h6 g5f4 d7f6 f4h6 f8e8 h6g7 g8g7 e4b7 a8b8 h2h3 b8b7 b1b7 c8b7 f3b7 a7a5 d4b5 a2d5 b7d5 f6d5');
INSERT INTO evaluations VALUES(47153,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(47153,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47153,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47153,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(47153,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(47153,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(47153,36,'cp -5','c7c5 b1d2 d8b6 g1f3 b8c6 d1b3 b6b3 d2b3 c5d4 f3d4 e7e5 d4c6 e5f4 c6d4 f8d6 g2g3 h7h5 g3f4 d6f4 e2e3 f4d6 h2h4 a7a6 f1h3 c8h3 h1h3 g7g6 h3h1 a8c8 e1e2 e8e7 d4f3 c8c4 a1d1 h8c8 d1d4 c4d4 f3d4 c8c4 d4f3');
INSERT INTO evaluations VALUES(51845,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(51845,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(51845,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(51845,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51845,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51845,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(51845,36,'cp 0','d7d5 d2d4 d5c4 d4e5 d8d1 c3d1 f8b4 d1c3 b8d7 g1f3 g8e7 c1d2 e8g8 f1g2 a7a5 a2a3 b4c3 d2c3 e7d5 c3d4 c6c5 f3d2 c5d4 g2d5 d7e5 d2c4 f8d8 c4e5 d8d5 e5d3 d5b5 e1d2 b5b3 a1b1 c8f5 d3c5 b3b6 c5d3 b6b3');
INSERT INTO evaluations VALUES(53754,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(53754,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53754,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53754,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(53754,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53754,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53754,36,'cp 0','d2d4 d5c4 d4e5 d8d1 c3d1 f8b4 d1c3 b8d7 g1f3 f7f6 c1f4 g8e7 e5f6 d7f6 a2a3 b4c3 b2c3 e7d5 a1c1 c8e6 f3d4 e6d7 d4f3');
INSERT INTO evaluations VALUES(16586,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(16586,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(16586,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(16586,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(16586,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(16586,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16586,36,'cp -40','d7d5 e4d5 f6d5 f1e2 g7g6 c2c4 d5e7 b2b4 f8g7 c1b2 e8g8 b1c3 e7f5 e1g1 a7a6 f1e1 f8e8 a1b1 h7h5 h2h3 b7b6 e2f1 c8b7 c3d5 d8d6 g2g3 a8d8 d1c2 h5h4 g3g4 f5d4 b2d4 c6d4 f3d4 b7d5 c4d5 e5d4');
INSERT INTO evaluations VALUES(15165,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(15165,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(15165,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(15165,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(15165,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(15165,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(15165,36,'cp -21','e4d5 f6d5 f1e2 g7g6 c2c4 d5e7 b2b4 f8g7 c1b2 e7f5 e1g1 e8g8 b1c3 f8e8 f1e1 b7b6 b4b5 c6d4 f3d4 f5d4 e2f3 d4f3 d1f3 a8b8 a3a4 c8b7 f3e3 c7c6 f2f3 d8d7 h2h3 f7f5 a1d1 b8d8 e3f2 e5e4 b5c6 b7c6 f3e4');
INSERT INTO evaluations VALUES(15016,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(15016,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(15016,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(15016,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(15016,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(15016,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(15016,36,'cp -26','f6d5 f1e2 g7g6 c2c4 d5e7 b2b4 f8g7 c1b2 e7f5 b1c3 e8g8 e1g1 b7b6 b4b5 c6d4 f3d4 f5d4 e2f3 d4f3 d1f3 a8b8 a3a4 c8b7 f3e2 c7c6 f2f3 d8d4 g1h1 d4d7 f1d1 f7f5 a4a5 c6c5 d1e1 b8e8 a1d1 e8d8 e2f2 f8e8 c3a4');
INSERT INTO evaluations VALUES(15365,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(15365,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(15365,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(15365,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(15365,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(15365,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(15365,36,'cp -37','f1e2 g7g6 c2c4 d5e7 b2b4 e7f5 c1b2 f8g7 b1c3 e8g8 e1g1 b7b6 b4b5 c6d4 f3d4 f5d4 e2f3 d4f3 d1f3 a8b8 a3a4 c8b7 f3e2 f8e8 f1e1 f7f5 a4a5 c7c6 a5b6 d8b6 h2h3 c6b5 c4b5 b8c8 e2d1 a7a6 b5a6 b6b2 a6b7');
INSERT INTO evaluations VALUES(48521,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(48521,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(48521,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(48521,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(48521,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(48521,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(48521,36,'cp 137','f8e7 f1b5 c7c6 b5d3 c6c5 d4d5 e6d5 e4e5 f6d7 c3d5 b8c6 d5e7 d8e7 e1g1 d7f8 c1g5 e7c7 g5e3 f8e6 d1a4 c8d7 a4h4 c6d4 f3d4 c5d4 e3d4 e6d4 h4d4 e8g8 a1c1 d7c6 d3c4 a8d8 d4e3 c7a5 f1d1 d8d1 c1d1 h7h6 e5e6 f7e6 e3e6 g8h8 e6d6 f8e8');
INSERT INTO evaluations VALUES(41050,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(41050,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(41050,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(41050,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(41050,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(41050,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(41050,36,'cp 140','d1c2 c7c5 d4c5 e7c5 e4e5 f6d5 c3d5 d8d5 a2a3 c8d7 f1c4 d5c6 e1g1 c5e7 c2e2 c6c7 c1f4 e8g8 a1c1 d7c6 b2b4 b8d7 f3d4 d7b6 d4c6 b7c6 c4b5 c6c5 b4c5 b6d5 f4d2 a8c8 c1c2 e7c5 f1c1 c7b6 a3a4 a7a5 g2g3 h7h6 d2a5 b6a5 c2c5 c8c5 c1c5 d5c3');
INSERT INTO evaluations VALUES(49614,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(49614,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(49614,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(49614,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(49614,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(49614,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(49614,36,'cp -112','f8c5 c3a4 c5e7 d2d4 e5d4 e4e5 f6g8 d1d4 b8c6 d4e3 g8h6 c1d2 h6f5 e3d3 d7d5 e5d6 f5d6 e1c1 e8g8 g1e2 a7a5 d3e3 f8e8 e2d4 c8d7 d4c6 d7c6 a4c3 d8d7 f1d3 e7f6 e3c5 b7b6 c5f2 b6b5 h1e1 b5b4 c3e4 d6e4 f3e4 a8d8');
INSERT INTO evaluations VALUES(42618,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(42618,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(42618,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(42618,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(42618,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(42618,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(42618,36,'cp -93','c3a4 c5e7 d2d4 e5d4 e4e5 f6g8 d1d4 b8c6 d4e3 g8h6 e3f2 c6e5 c1h6 g7h6 e1c1 e8g8 f3f4 e5g4 f2f3 d7d5 a4c3 c7c6 h2h3 g4f6 f1d3 b7b5 g2g4 b5b4 c3e2 a7a5 f3g2 b4b3 c2b3 a5a4 g1f3 a4b3 a2a3 f6d7 e2d4 d7c5 c1b1 e7f6 f3e5');
INSERT INTO evaluations VALUES(25305,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(25305,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(25305,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(25305,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(25305,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(25305,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25305,36,'cp 45','h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 e1g1 d6d5 e4d5 c6d5 c4b3 d4b3 a2b3 d5d4 c3b1 f8e7 f1e1 f6d7 f2f4 e8g8 f4e5 e7h4 e1e2 h4g3 b1d2 g3e5 d1e1 e5c7 d2f3 d7f6 e1h4 h7h6 c1f4 c7b6 a1e1 f6d5 h4g4 d5f6 g4f5 f6d5 f4e5');
INSERT INTO evaluations VALUES(25303,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(25303,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(25303,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(25303,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25303,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(25303,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(25303,36,'cp 42','g4f3 d1f3 c6d4 f3d1 c7c6 e1g1 d6d5 e4d5 c6d5 c4b3 d4b3 a2b3 d5d4 c3b1 f8e7 f1e1 f6d7 f2f4 e7h4 e1e2 e8g8 f4e5 h4g3 b1d2 g3e5 d2f3 e5f6 d1e1 d7c5 c1f4 c5e6 f4e5 f6e5 e2e5 d8c7 e1f2 a7a5 a1e1 a8c8 e5e2 c8d8 f3h4 d8d5 e2e4');
INSERT INTO evaluations VALUES(25321,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(25321,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25321,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(25321,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(25321,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(25321,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(25321,36,'cp 19','d1f3 c6d4 f3d1 c7c6 a2a4 f8e7 e1g1 e8g8 c1e3 d6d5 c4a2 d5e4 c3e4 f6d5 e3d4 e5d4 d1h5 d5f6 h5f3 e7b4 e4f6 d8f6 f3f6 g7f6 f1c1 f8e8 g1f1 a8d8 c2c3 b4f8 a2c4 f6f5 a4a5 e8e5 f2f4 e5e8 c1b1 f8g7 a1a4 g7f8 b1d1 d4c3 b2c3');
INSERT INTO evaluations VALUES(25319,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(25319,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(25319,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(25319,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(25319,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(25319,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(25319,36,'cp 18','c6d4 f3d1 c7c6 a2a4 f8e7 e1g1 e8g8 c1e3 d6d5 c4a2 d5e4 c3e4 f6d5 e3d4 e5d4 d1h5 d5f6 h5f3 e7b4 a2c4 a7a5 f1d1 d8e7 e4f6 e7f6 f3f6 g7f6 a1c1 g8g7 c2c3 b4c5 d1e1 f8e8 g1f1 f6f5 g2g4 e8e1 c1e1 f5g4 h3g4 d4c3 b2c3');
INSERT INTO evaluations VALUES(25403,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(25403,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25403,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(25403,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(25403,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(25403,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(25403,36,'cp 31','f3d1 c7c6 a2a4 f8e7 e1g1 e8g8 c1e3 d4e6 d1f3 d8d7 a4a5 f8d8 c4e6 f7e6 f3e2 d8f8 c3b1 d6d5 b1d2 e7d6 d2f3 d7e7 e3g5 e7e8 g5d2 f6d7 a1b1 a7a6 b2b4 h7h6 c2c4 e8e7 g1h2 f8f7 b4b5 a6b5 c4b5');
INSERT INTO evaluations VALUES(25404,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(25404,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(25404,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(25404,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(25404,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(25404,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(25404,36,'cp 18','c7c6 e1g1 d6d5 e4d5 c6d5 c4b3 d4b3 a2b3 d5d4 c3b1 f8e7 d1e2 d8d5 c2c4 d5e6 b1a3 a7a6 f1e1 e7b4 c1d2 b4d2 e2d2 e8g8 f2f4 f6d7 a3c2 f7f6 b3b4 a8d8 e1e4 e6d6 a1a5 d6c7 a5d5 d8c8 e4e1 f8f7 f4e5 f6e5 d2g5 h7h6 g5g3 g8h8 g3g4');
INSERT INTO evaluations VALUES(24745,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24745,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24745,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24745,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(24745,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(24745,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(24745,36,'cp 27','a2a4 f8e7 e1g1 e8g8 c1e3 d8c7 c4a2 d4e6 a4a5 b7b5 a5b6 a7b6 a2e6 f7e6 c3e2 h7h6 e2g3 c6c5 d1e2 b6b5 g3h5 g8h7 h5f6 e7f6 e2g4 c7f7 h3h4 h6h5 g4h3 b5b4 c2c3 b4c3 b2c3 f6e7 f1b1 a8a1 b1a1 d6d5 a1a7 d5d4 c3d4');
INSERT INTO evaluations VALUES(36588,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36588,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(36588,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(36588,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36588,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36588,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(36588,36,'cp 24','c7c5 d4d5 e7e6 f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 b8c6 e1g1 e8f8 g1h1 b7b6 h2h3 g4e5 c1g5 d8d7 f3e5 f8f1 d1f1 g7e5 c3e4 c8b7 g5f6 d7f7 f1f2 a8f8 a1f1 c6d4 f6e5 f7f2 e4f2 d4e2 d6d7 b7c6 f1e1 e2d4 f2g4 h7h5 g4f6 g8f7 e5d6 f8d8 d6c7 f7e7 c7d8 e7d8 b2b3 c6d7 f6d7 d8d7');
INSERT INTO evaluations VALUES(35956,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35956,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(35956,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35956,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35956,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35956,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(35956,36,'cp 16','d4d5 e7e6 f1e2 e6d5 c4d5 f8e8 e4e5 d6e5 f4e5 f6g4 e5e6 f7e6 d5d6 e8f8 e1g1 b8c6 g1h1 b7b6 c3e4 d8d7 f3g5 f8f1 d1f1 g4e5 e4f6 g7f6 f1f6 d7g7 f6g7 g8g7 b2b4 c6d4 c1b2 e5f7 g5f7 g7f7 b2d4 c5d4 e2f3 a8b8 f3c6 c8b7 a1f1 f7g7 b4b5 e6e5 d6d7 b7a8 f1e1 g7f6');
INSERT INTO evaluations VALUES(45249,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45249,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(45249,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45249,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(45249,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45249,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45249,36,'cp 0','e7e5 d4b3 f8e7 c2c4 a7a5 c1e3 a5a4 b3d2 f6d7 b1c3 d7c5 f1e2 e8g8 e1g1 b8c6 c3b5 e7g5 e3f2 f7f5 e4f5 c8f5 d2e4 f5e4 f3e4 f8f2 f1f2 g5e3 d1d5 g8h8 b5d6 e3f2 g1h1 d8f8 d6f7 h8g8 f7e5 g8h8 e5f7 h8g8');
INSERT INTO evaluations VALUES(45949,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45949,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45949,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(45949,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45949,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45949,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45949,36,'cp 14','d4b3 f8e7 c2c4 b8c6 b1c3 e8g8 c1e3 a7a5 f1e2 a5a4 b3d2 f6d7 e1g1 d7c5 g1h1 e7g5 e3g1 a4a3 b2b3 c6d4 d2b1 f7f5 b3b4 c5e6 c3d5 f5e4 f3e4 d4e2 d1e2 e6f4 d5f4 e5f4 b1d2 h7h6 d2f3 g5f6 a1d1 c8e6');
INSERT INTO evaluations VALUES(45950,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(45950,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(45950,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(45950,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(45950,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45950,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(45950,36,'cp 17','f8e7 c2c4 a7a5 c1e3 e8g8 b1c3 b8c6 f1e2 a5a4 b3d2 f6h5 c3d5 h5f4 e1g1 d8a5 e3b6 a5a6 b6f2 a6a5 g1h1 c8e6 g2g3 f4e2 d1e2 e6d5 c4d5 c6d4 f2d4 e5d4 f3f4 e7f6 e2d3 b7b5 d2f3 a4a3 b2b3 f8c8 f3d4');
INSERT INTO evaluations VALUES(54013,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(54013,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(54013,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(54013,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(54013,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(54013,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(54013,36,'cp 87','g1f3 e6d5 e4d5 b8d7 b1c3 g8f6 f1d3 f8e7 e1g1 e8g8 f1e1 h7h6 h2h3 f8e8 c1f4 a7a6 a2a4 d7b6 d3f1 c8f5 d1d2 e7f8 e1e8 d8e8 a4a5 b6d7 f1d3 f5d3 d2d3 e8e7 f3d2 d7e5 f4e5 e7e5 d2c4 e5d4 a1e1 g7g6 g2g4 d4d3 c2d3 a8e8 e1e8 f6e8');
INSERT INTO evaluations VALUES(54012,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(54012,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(54012,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(54012,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(54012,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(54012,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(54012,36,'cp 90','e6d5 e4d5 b8d7 b1c3 f8e7 f1d3 a7a6 a2a4 g8f6 e1g1 d7b6 d3e2 e8g8 a4a5 b6d7 f3d2 b7b5 a5b6 d7b6 d2b3 f8e8 b3a5 d8c7 e2f3 c8d7 b2b3 g7g6 h2h3 b6a4 c3a4 c7a5 c1d2 a5d8 a4b2 a6a5 f1e1 e7f8 e1e8 f6e8 b2c4');
INSERT INTO evaluations VALUES(26616,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(26616,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(26616,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(26616,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(26616,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(26616,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(26616,36,'cp 43','g8f6 b1d2 e7e6 a2a3 a7a5 b2b3 h7h6 c1b2 g4f5 c2c4 f8e7 e1g1 e8g8 b2c3 f6e4 d2e4 f5e4 d1d2 e7f6 d2b2 d8d7 a1c1 b7b6 f1d1 c6e7 f3e5 f6e5 d4e5 c7c6 c4d5 c6d5 f2f3 e4g6 a3a4 f8c8 e2b5 d7b7 c3e1 e7f5');
INSERT INTO evaluations VALUES(24950,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(24950,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(24950,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(24950,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(24950,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24950,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(24950,36,'cp 41','b1d2 e7e6 a2a3 a7a5 h2h3 g4f5 f3h4 f5e4 c2c4 f8d6 d2e4 f6e4 h4f3 e8g8 e1g1 b7b6 d1c2 c6e7 c1d2 c7c5 f1d1 f7f5 a1c1 g8h8 d2e1 h7h6 e2d3 d8e8 c2e2 e7g6 d3c2 e8f7 c2b3 a5a4 b3a2 e4f6 a2b1 f6e4 e2d3 g6e7 d3c2 f7f6 c4d5');
INSERT INTO evaluations VALUES(24952,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(24952,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(24952,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(24952,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(24952,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(24952,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24952,36,'cp 34','e7e6 c2c4 a7a5 e1g1 f8d6 b2b3 e8g8 c1b2 f6e4 d2e4 d5e4 f3d2 g4e2 d1e2 f7f5 a2a3 d8e8 f2f3 e8h5 g2g3 e4f3 d2f3 h5g4 a1e1 a8e8 c4c5 d6e7 e2b5 e8b8 b2c3 g4e4 c3a5 b7b6 c5b6 c6a5 b5a5 b8b6 a5a4 e4d5 b3b4 f8a8 a4c2 a8a3 c2c7 e7b4');
INSERT INTO evaluations VALUES(25224,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(25224,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(25224,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(25224,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(25224,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(25224,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25224,36,'cp 36','c2c4 a7a5 a2a3 h7h6 b2b3 f8e7 c1b2 g4f5 b2c3 e8g8 h2h3 f6e4 d2e4 f5e4 e1g1 d8d7 d1c1 e7f6 c1b2 c6e7 f1c1 f8b8 f3h2 f6g5 c3e1 c7c6 c4c5 b7b6 c5b6 b8b6 b3b4 d7b7 a3a4 a5b4 a4a5');
INSERT INTO evaluations VALUES(44592,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(44592,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(44592,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(44592,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(44592,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(44592,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(44592,36,'cp 63','d7d6 e2e4 e7e6 f1c4 e6d5 c3d5 f8e7 g1e2 f6d5 c4d5 b8a6 e1g1 e8g8 e2f4 a6c7 a2a4 c7d5 f4d5 f8e8 f2f3 c8e6 c2c3 d8d7 c1f4 e7f8 d1d3 f7f5 a4a5 f5e4 f3e4 e6d5 e4d5 g7g6 h2h3 b7b5 d3g3 b5b4 a1d1 b4c3 b2c3');
INSERT INTO evaluations VALUES(45234,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(45234,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(45234,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(45234,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(45234,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(45234,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(45234,36,'cp 70','e2e4 e7e6 f1c4 e6d5 c3d5 f8e7 g1e2 f6d5 c4d5 e8g8 e1g1 b8a6 e2c3 a6b4 c1f4 f8e8 d5b3 c8e6 d1d2 e6b3 a2b3 d8b6 a1d1 b6c6 f2f3 h7h6 f4e3 a8d8 c3d5 b4d5 d2d5 e7f6 c2c3 a7a5 d5c6 b7c6 d1d2 f6e5 f3f4');
INSERT INTO evaluations VALUES(45228,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(45228,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(45228,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(45228,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(45228,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(45228,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(45228,36,'cp 70','e7e6 f1c4 f8e7 g1e2 e6d5 c3d5 f6d5 c4d5 e8g8 e1g1 b8a6 e2f4 a6c7 a2a4 c7d5 f4d5 f8e8 c2c3 c8e6 f2f3 e6d5 d1d5 d8b6 d5c4 e7f6 f1f2 f6e5 c1g5 h7h6 g5e3 a7a6 f2d2 b6a5 f3f4 b7b5 c4d5 e5f6 d2d1');
INSERT INTO evaluations VALUES(3820,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(3820,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(3820,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(3820,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(3820,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(3820,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(3820,36,'cp 20','c3a4 c5b6 a2a3 c8e6 c4e6 f7e6 a4b6 a7b6 e1g1 h7h6 c2c3 c6e7 f1e1 e7g6 d3d4 f6d7 d1e2 d8f6 c1e3 f8f7 a1d1 a8f8 a3a4 f6e7 f3h2 e5d4 c3d4 e6e5 d4e5 d7e5 h2g4 e5g4 e2g4 g6f4 f2f3 h6h5 g4g3 e7e5 e3f4');
INSERT INTO evaluations VALUES(30437,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(30437,31,'cp 246',NULL);
INSERT INTO evaluations VALUES(30437,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(30437,33,'cp 251',NULL);
INSERT INTO evaluations VALUES(30437,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(30437,35,'cp 245',NULL);
INSERT INTO evaluations VALUES(30437,36,'cp 246','a4b5 c7c6 b5c6 b8c6 g1f3 g8f6 f1c4 e7e6 e1g1 f8b4 c1d2 e8g8 d2b4 c6b4 f3e5 d8b6 b1c3 f8c8 d1d2 d7e8 c4e2 b4d5 e2f3 a8b8 c3a4 e8a4 a1a4 d5b4 f1c1 c8c1 d2c1 b4d5 a4c4 d5e7 e5d3 h7h6 c4c5 b6a6 c5c3 e7d5 c3a3 a6b6 c1d2 d5b4 d3c5 b8c8');
INSERT INTO evaluations VALUES(30430,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(30430,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(30430,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(30430,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(30430,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(30430,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(30430,36,'cp 232','c7c6 b5c6 b8c6 g1f3 e7e6 f1c4 g8f6 e1g1 f8b4 e3e4 e8g8 d1d3 h7h6 b1c3 c6e7 f3e5 d7c8 f1d1 c8b7 f2f3 e7c6 c1e3 c6e5 d4e5 d8d3 d1d3 f8c8 e5f6 c8c4 f6g7 a7a5 e3h6 b7c6 a1d1 b4c5 g1h1 c5e7 h6f4 e7f6 h2h3 f6c3 b2c3 g8g7 f4e3 a5a4');
INSERT INTO evaluations VALUES(53414,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(53414,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(53414,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53414,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53414,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53414,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53414,36,'cp 16','d2d4 b8c6 c1f4 g8f6 e2e3 c8f5 a1c1 a7a6 g1f3 a8c8 f1e2 e7e6 e1g1 f8d6 f4d6 d8d6 c3a4 e8g8 a4c5 c8c7 d1b3 d6e7 b3a4 c6a7 c1c3 f6e4 c5e4 f5e4 f3d2 c7c3 b2c3 e4g6 f1c1 h7h6 d2b3 a7c8 c3c4 c8b6 a4a5 b6c4 e2c4 d5c4 c1c4 b7b5');
INSERT INTO evaluations VALUES(57507,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(57507,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(57507,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(57507,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(57507,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(57507,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(57507,36,'cp 120','g8f6 e4f5 e5e4 f3g5 d7d5 d2d3 e4e3 f2e3 c8f5 f1e2 b8c6 e3e4 d5e4 e1g1 d8d7 d3e4 d7d1 e2d1 f8c5 g1h1 f5d7 e4e5 c6e5 c1f4 e5f7 g5f7 e8f7 f4c7 a8c8 c7e5 f7g6 d1e2 d7f5 e5f6 g7f6 c3d5 f5c2 a1c1 c2e4 d5f6 e4f5 g2g4 c5d4 c1d1 d4f6 g4f5 g6g5');
INSERT INTO evaluations VALUES(49998,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(49998,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(49998,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(49998,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(49998,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(49998,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(49998,36,'cp 171','e4f5 b8c6 d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 d1e2 d8e7 e2e7 e8e7 a2a3 b4c3 b2c3 e7f7 f2f3 d7d5 c1f4 c8f5 e1c1 h8e8 f4c7 e8e6 c7g3 h7h5 f1a6 a8e8 d1e1 f6d7 e1e6 e8e6 c1d2 g7g5 h2h4 g5g4 f3g4 f5g4 a6d3 d7e5 g3e5 e6e5');
INSERT INTO evaluations VALUES(49986,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(49986,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(49986,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(49986,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(49986,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(49986,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(49986,36,'cp 131','b8c6 d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 d1e2 d8e7 a2a3 b4c3 b2c3 d7d6 e2e7 e8e7 f1d3 f6d7 g2g4 h7h5 f2f3 d7e5 d3e2 h5g4 c1g5 e7f7 f3g4 h8h3 g5f4 h3c3 e1d2 c3h3 e2f1 h3f3 f4e5 d6e5 a1e1 f3a3 e1e5 c8a6 f1a6 a3a6 h1e1 a6a4 e5e7 f7f8 h2h3 a8d8 d2c1 a4a1 c1b2 a1e1 e7e1');
INSERT INTO evaluations VALUES(17105,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(17105,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(17105,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(17105,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(17105,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(17105,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(17105,36,'cp 163','d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 a2a3 b4c3 b2c3 e8g8 f1d3 d8e7 c1e3 d7d6 d1f3 e7e5 e1g1 e5c3 a1b1 a7a5 g2g4 c8a6 b1b3 c3e5 d3a6 a8a6 f1d1 f8e8 h2h3 e5e4 f3e4 e8e4 b3b8 g8f7 b8c8 f6d5 e3d4 c6c5 d4b2 c7c6');
INSERT INTO evaluations VALUES(4197,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(4197,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(4197,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(4197,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4197,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4197,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4197,36,'cp 5','a7a6 c1e3 d7d6 c3d5 f6d5 c4d5 c6e7 d5a2 c5e3 f2e3 e7g6 d1e1 d8e7 e1g3 c8e6 a2e6 e7e6 f1f2 f7f5 e4f5 f8f5 a1f1 a8f8 g3g4 f5f6 g4e6 f6e6 f3d2 f8f2 f1f2 g6f8 g1f1 f8d7 f1e2 e6f6 f2f6 d7f6 e3e4 f6h5 e2e3 g8f7 d2c4 h5f4 g2g3');
INSERT INTO evaluations VALUES(2647,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(2647,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2647,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(2647,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(2647,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(2647,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(2647,36,'cp 7','c1e3 d7d6 c3d5 f6d5 c4d5 c5e3 f2e3 c6e7 d5b3 c8e6 b3e6 f7e6 d1e2 e7g6 h2h3 d8e7 e2d2 b7b6 d2e2 f8f7 f1f2 a8f8 a1f1 a6a5 e2d2 g6h4 f3h2 f7f2 f1f2 f8f2 d2f2 a5a4 f2g3 g8h7 g3g4 h4g6 h2f3 g6f8 b2b3 a4b3 c2b3 f8d7');
INSERT INTO evaluations VALUES(31852,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31852,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(31852,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(31852,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31852,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31852,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(31852,36,'cp 14','d1b3 b7b6 c1g5 h7h6 g5h4 g8f6 h4f6 e7f6 e2e3 f8e7 c3c4 d5c4 f1c4 e8g8 b1c3 b8c6 b3d1 c6a5 c4d3 f5d3 d1d3 c7c5 g1f3 c5d4 f3d4 e7b4 e1g1 a5c6 f1d1 c6d4 e3d4 d8d7 a1c1 f8d8 d4d5 a8c8 g2g3 b4d6 d3f3 d6b4 h2h4 b4d6 c3e4 c8c1 d1c1');
INSERT INTO evaluations VALUES(38834,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(38834,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(38834,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(38834,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(38834,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(38834,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(38834,36,'cp 58','d8e7 a2a3 a7a6 c2c4 g8f6 g3g4 d5c4 f1c4 e6e5 g4g5 f6e4 d4e5 c8f5 b1d2 e7e5 d2e4 f5e4 d1b3 e8g8 g1f3 e5a5 e1e2 b8d7 a1d1 a8d8 c4d5 e4g6 b3b4 a5b4 a3b4 c7c6 d5a2 f8e8 d1d6 d7f8 d6d8 e8d8 f3d4 d8e8 f2f3');
INSERT INTO evaluations VALUES(34356,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(34356,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(34356,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(34356,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(34356,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(34356,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(34356,36,'cp 2','g8f6 g1f3 e7e6 b1a3 b8c6 c1e3 c5d4 a3b5 d5d8 b5d4 d8c7 d1d2 f6d5 e1c1 f8e7 e3g5 f7f6 g5h4 e8g8 f1c4 f8d8 d4c6 b7c6 f3d4 c8d7 h1e1 e6e5 d4b5 c7b6 e1e5 f6e5 h4e7 d8e8 c4d5 c6d5 b5d6 d7e6 d6e8 a8e8');
INSERT INTO evaluations VALUES(33608,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(33608,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(33608,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(33608,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(33608,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(33608,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(33608,36,'cp 17','g1f3 e7e6 b1a3 b8c6 c1e3 c5d4 a3b5 d5d7 b5d4 f8e7 d4c6 d7c6 f1d3 e8g8 d1e2 b7b6 e3d4 c8b7 e1c1 f8d8 h1e1 c6c7 c1b1 a8c8 h2h4 e7c5 d4e5 c7e7 d3c2 h7h6 f3d4 f6d7 e5g3 d7f6 d4f5 e6f5 e2e7 c5e7');
INSERT INTO evaluations VALUES(22131,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22131,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(22131,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(22131,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(22131,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(22131,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(22131,36,'cp 90','d1d4 g8f6 b1c3 f8e7 c1g5 c7c6 e1c1 d8b6 g5f6 g7f6 d4d2 c8e6 c4e6 f7e6 d2h6 e8c8 h6h3 f6f5 f2f4 b6c5 h3h6 c5c4 c1b1 d8g8 h1e1 e7f8 h6h5 f8g7 d1d3 g8f8 a2a3 c8b8 g2g3 a7a6 e4f5 e6f5 e1e7 g7f6 e7h7 f6c3 d3c3 c4g8 h7h8 g8h8 h5h8');
INSERT INTO evaluations VALUES(49616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49616,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(49616,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(49616,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(49616,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(49616,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(49616,36,'cp -14','f8c5 f1g2 e8g8 g1e2 c7c6 e1g1 d7d5 e4d5 c6d5 d2d4 e5d4 e2d4 c8g4 d4f3 b8c6 c1g5 d5d4 c3e4 c5e7 g5f6 e7f6 h2h3 g4f5 f3e1 f8e8 e4f6 d8f6 e1d3 c6e5 f1e1 e5d3 e1e8 a8e8 c2d3 h7h5 g2b7 e8b8 d1f3 f5h3 b2b3 f6f3 b7f3 g7g6 f3d5 b8b4 a1e1 g8f8 d5g2 h3g2');
INSERT INTO evaluations VALUES(42621,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(42621,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(42621,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(42621,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(42621,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(42621,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(42621,36,'cp -12','f1g2 e8g8 g1e2 c7c6 e1g1 d7d5 e4d5 c6d5 d2d4 e5d4 e2d4 c8g4 d4f3 b8c6 c1g5 d5d4 c3e4 c5e7 g5f6 e7f6 h2h3 g4f5 f3e1 f8e8 e4f6 d8f6 d1d2 a8d8 e1d3 h7h6 a1e1 c6e5 d3e5 e8e5 g3g4 e5e1 f1e1 f5e6 a2a3 b7b6 g2e4 g7g5 g1g2 e6c4');
INSERT INTO evaluations VALUES(42622,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42622,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(42622,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(42622,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42622,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(42622,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(42622,36,'cp -13','e8g8 g1e2 c7c6 e1g1 d7d5 e4d5 c6d5 d2d4 e5d4 e2d4 c8g4 d4f3 b8c6 c1g5 a8c8 h2h3 g4e6 f3e1 c5d4 g5f6 d8f6 c3d5 e6d5 g2d5 d4b2 a1b1 f8d8 c2c4 c6b4 d1f3 b4d5 c4d5 c8c5 f3f6 b2f6 b1b7 c5d5 b7a7');
INSERT INTO evaluations VALUES(52452,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(52452,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52452,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52452,36,'cp 0','g8f6 b1c3 d7d5 e4d5 f6d5 c3d5 d8d5 d2d4 e7e5 d4e5 d5e5 c2c3 b8c6 c1f4 e5f6 d1c1 f8e7 e2g3 e8g8 f1e2 c6e5 g3h5 f6f5 e1g1 c8d7 h5g3 f5f6 g3h5');
INSERT INTO evaluations VALUES(40761,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(40761,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40761,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40761,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40761,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(40761,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(40761,36,'cp 10','b8c6 d2d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 e5d4 e1g1 e8g8 e4c6 b7c6 d1d4 e7f6 d4d8 f8d8 c1g5 f6g5 f3g5 c8f5 f1e1 f7f6 g5e4 d8e8 e4g3 f5g6 b2b3 a7a5 e1e8 a8e8 a1d1 h7h5 h3h4 a5a4 b3a4 g6c2 d1c1 c2a4 g3h5 e8e2');
INSERT INTO evaluations VALUES(8581,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(8581,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8581,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(8581,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(8581,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(8581,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(8581,36,'cp 30','e1g1 c6a5 c4b5 c7c6 b5d3 c6c5 a2a3 a5c6 c3d5 f6d5 e4d5 c6b8 b2b4 b7b6 d3b5 c8d7 b5d7 b8d7 b4c5 b6c5 d2d3 e8g8 c2c4 a8b8 c1d2 f7f5 f1e1 e7f6 a1b1 b8b1 d1b1 d8c7 b1c2 f8b8 d2c3 d7f8 a3a4 c7f7 f3d2');
INSERT INTO evaluations VALUES(8580,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(8580,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(8580,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(8580,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8580,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(8580,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(8580,36,'cp 6','c6a5 c4d3 c7c5 a2a3 a5c6 c3d5 f6d5 e4d5 c6b8 b2b4 b8d7 c2c4 e8g8 d3c2 b7b6 d2d3 f7f5 c2a4 e7f6 a1b1 a7a6 b4c5 b6c5 f1e1 h7h6 c1d2 d8e8 a4c6 a8a7 a3a4 f6d8 d2c3 a6a5 f3d2 e8g6 d1e2 d8g5 d2f3');
INSERT INTO evaluations VALUES(50213,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(50213,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(50213,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(50213,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(50213,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(50213,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50213,36,'cp -11','d7d5 g1f3 f8g7 e1g1 c7c5 d4c5 e8g8 c2c4 d5c4 d3c4 d8c7 b1c3 c7c5 d1d4 c5d4 f3d4 c8d7 f1d1 b8c6 d4c6 d7c6 f2f3 f8d8 d1d8 a8d8 e3e4 f6d7 c1e3 d7e5 c4e2 a7a6 a1d1 d8d1 c3d1 c6b5 e2b5 a6b5 d1c3 e5c4 e3c5 g7c3 b2c3 e7e5 h2h3 f7f5 e4f5 g6f5 g1f2 g8f7 c5b4 f7f6');
INSERT INTO evaluations VALUES(46792,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(46792,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(46792,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(46792,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(46792,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46792,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46792,36,'cp 0','g1f3 f8g7 e1g1 c7c5 d4c5 e8g8 c2c4 d5c4 d3c4 b8d7 d1c2 d8c7 b2b4 a7a5 c1b2 a5b4 a2a3 b4b3 c2b3 d7c5 b3a2 c8e6 c4e6 c5e6 b1d2 c7a5 b2e5 f8d8 a2b2 e6c5 e5d4 c5e6');
INSERT INTO evaluations VALUES(46794,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(46794,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46794,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46794,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(46794,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46794,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46794,36,'cp 0','f8g7 e1g1 c7c5 d4c5 e8g8 c2c4 b8a6 c4d5 a6c5 d3c4 b7b6 b1c3 a7a6 a2a3 c8b7 c4a2 a8c8 f1e1 c5e4 c3e4 f6e4 d1d3 e4c5 d3d1');
INSERT INTO evaluations VALUES(50504,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(50504,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50504,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(50504,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(50504,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(50504,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(50504,36,'cp -12','b7b5 c4d3 c5c4 d3e2 g8f6 e4e5 f6d5 c2c3 b8c6 b2b3 c6a5 e1g1 c8b7 f3d2 d8c7 d2e4 f8e7 c1a3 e7a3 b1a3 e8g8 b3c4 a5c4 a3c4 b5c4 d1d2 d5e7 f2f3 b7d5 e2d1 e7g6 d1c2 f7f6 e5f6 g7f6 d2h6 c7f4 h6f4 g6f4 g2g3 f4d3 f1b1 d5e4 f3e4');
INSERT INTO evaluations VALUES(51205,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(51205,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(51205,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(51205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51205,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(51205,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(51205,36,'cp 17','c4d3 c5c4 d3e2 g8f6 e4e5 f6d5 c2c3 b8c6 b2b3 c6a5 e1g1 c8b7 f3d2 d8c7 d2e4 f8e7 c1a3 e7a3 b1a3 e8g8 b3c4 a5c4 a3c4 b5c4 d1d2 d5e7 f2f3 b7d5 e2d1 f7f5 e5f6 g7f6 d2h6 e7g6 d1c2 f8f7 a1e1 a8f8 e4c5 f7g7 c5a6 c7f4 h6f4 g6f4 g2g3 f8a8 a6b4 a8a3 e1e3 d5b7 c2e4 b7e4 e3e4 f4h3 g1g2 h3g5 e4e1 a3c3 d4d5 e6d5 b4d5 c3c2 f1f2');
INSERT INTO evaluations VALUES(51208,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(51208,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(51208,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(51208,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51208,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51208,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51208,36,'cp 9','c5c4 d3e2 g8f6 e4e5 f6d5 c2c3 b8c6 b2b3 c6a5 e1g1 c8b7 f3d2 d8c7 d2e4 f8e7 c1a3 e7a3 b1a3 e8g8 b3c4 d5f4 e2f3 a5c4 a3c4 b5c4 a1b1 b7d5 d1d2 f4g6 g2g3 a8b8 h2h4 b8b1 f1b1 f8b8 b1b8 c7b8 d2c2 h7h6 f3g2 g6e7 e4d6');
INSERT INTO evaluations VALUES(55565,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55565,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(55565,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(55565,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55565,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(55565,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(55565,36,'cp 18','f7f5 b1c3 g8f6 g1f3 f8e7 e1g1 e8g8 b2b4 a7a6 d2d3 d8e8 a2a4 h7h6 c1a3 c8e6 b4b5 a6b5 c3b5 e8d7 d3d4 e5e4 d4d5 e6f7 f3d4 c7c6 b5c3 c6c5 d4c2 f6g4 g2h3 e7f6 a3b2 d7e7 h3g4 f5g4 c2e3 f7g6 d1c2 f6g5 e3g4 g6f5 g4e3 g5e3');
INSERT INTO evaluations VALUES(56743,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(56743,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(56743,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(56743,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(56743,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(56743,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(56743,36,'cp 44','b1c3 g8f6 d2d3 f8e7 b2b4 e8g8 g1f3 d8e8 a2a4 a7a6 e1g1 b8c6 a1b1 a6a5 b4a5 a8a5 f3d2 e7d8 e2e3 b7b6 f1e1 c6e7 f2f4 c8d7 c3b5 d7c6 d2f3 f6d7 c1a3 g8h8 a3b2 e7g6 b2c3 a5a8');
INSERT INTO evaluations VALUES(56742,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(56742,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(56742,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(56742,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(56742,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(56742,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(56742,36,'cp 29','g8f6 g1f3 f8e7 e1g1 e8g8 b2b4 d8e8 d2d3 a7a6 a2a4 b8c6 c1a3 f5f4 e2e3 f4e3 f2e3 e8h5 b4b5 a6b5 c4b5 c6a5 f3d2 h5h6 d1e2 c8g4 g2f3 g4f3 f1f3 h6h3 d2e4 h3e6 e4f6 f8f6 a3b4 c7c6 f3f6 e7f6 a1f1 d6d5 e2a2 h7h6 b4a5 a8a5 c3e4 a5a8 e4f6 g7f6');
INSERT INTO evaluations VALUES(49100,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(49100,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(49100,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(49100,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(49100,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(49100,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(49100,36,'cp 28','b2b4 f8e7 d2d3 e8g8 g1f3 a7a6 e1g1 h7h6 a2a4 d8e8 c1a3 c8e6 f3d2 c7c6 b4b5 d6d5 b5c6 b7c6 c4d5 c6d5 e2e4 e7a3 e4d5 a3b2 d5e6 b2a1 e6e7 e8e7 g2a8 a1c3 d1b3 g8h8 b3c3 e5e4 d3e4 b8d7 a8b7 d7c5 e4e5 c5b7 e5f6 f8f6 c3c8 h8h7');
INSERT INTO evaluations VALUES(8019,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(8019,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(8019,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(8019,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(8019,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(8019,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(8019,36,'cp 48','d2d4 c5d4 f3d4 e8g8 e1g1 d8a5 e2e3 d7d6 c1d2 a5c5 d4c6 b7c6 a1c1 c8g4 d1e1 a8b8 b2b3 c5h5 g2c6 g4h3 e1d1 h3f1 d1h5 f6h5 g1f1 g7c3 d2c3 f8c8 c6d5 g8f8 f1e2 e7e6 d5g2 f7f5 f2f4 f8f7 c1d1 f7e7 d1d2 h5f6 c3f6 e7f6');
INSERT INTO evaluations VALUES(8023,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(8023,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(8023,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(8023,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(8023,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(8023,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(8023,36,'cp 44','c5d4 f3d4 e8g8 e1g1 d8a5 e2e3 d7d6 c1d2 a5c5 d4c6 b7c6 a1c1 c8g4 d1e1 c5h5 f2f3 g4e6 c3e2 e6d7 b2b4 h5e5 a2a4 d6d5 e2d4 e5d6 f3f4 f6e4 c4c5 d6c7 b4b5 g7d4 e3d4 e4d2 e1d2 a8b8 d2e3 c6b5 a4b5 d7b5');
INSERT INTO evaluations VALUES(8047,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(8047,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(8047,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(8047,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(8047,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(8047,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(8047,36,'cp 51','f3d4 e8g8 e1g1 d8a5 e2e3 d7d6 c1d2 a5c5 d4c6 b7c6 b2b3 a8b8 a1c1 c8b7 f1e1 f6d7 c3a4 c5a3 d1c2 d7b6 a4c3 b6d7 c1d1 a3a6 e3e4 d7c5 c3e2 c5e6 d2c3 a6b6 c2d2 g7c3 e2c3 c6c5 c3d5 b6d8 e4e5 b7d5 d2d5 e6d4 e5d6 d8d6');
INSERT INTO evaluations VALUES(8041,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(8041,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(8041,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(8041,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(8041,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(8041,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(8041,36,'cp 44','e8g8 e1g1 d8a5 e2e3 d7d6 c1d2 a5c5 d4c6 b7c6 b2b3 a8b8 a1c1 c8b7 f1e1 c5a3 d1c2 f6d7 c1d1 a3a6 h2h3 d7c5 g1h2 a6b6 d2c1 a7a5 c1b2 a5a4 c3a4 c5a4 b2g7 g8g7 b3a4 b6c5 c2e4 e7e5 d1c1 c5b4 e4c2 c6c5 g2b7');
INSERT INTO evaluations VALUES(3064,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(3064,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(3064,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(3064,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(3064,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(3064,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(3064,36,'cp 44','e1g1 d8a5 e2e3 d7d6 c1d2 a5c5 d4c6 b7c6 b2b3 a8b8 a1c1 c8b7 f1e1 f6d7 c3a4 c5a3 d1c2 d7b6 a4c3 b6d7 c1d1 a3a6 h2h3 d7c5 g1h2 b8d8 c3b1 c5e6 d2c3 d6d5 c3g7 e6g7 c4d5 c6d5 b1c3 e7e6 c2d2 a6a5 c3a4 a5d2 d1d2 g7f5 e1d1');
INSERT INTO evaluations VALUES(3063,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(3063,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(3063,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(3063,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(3063,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(3063,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(3063,36,'cp 51','d8a5 e2e3 d7d6 c1d2 a5c5 d4c6 b7c6 b2b3 a8b8 a1c1 f6d7 d1c2 c8b7 f1e1 f8e8 h2h3 c5b6 c3a4 b6c7 c4c5 d6d5 e3e4 e7e6 d2c3 e6e5 b3b4 d5d4 c3d2 b7a6 a4b2 h7h5 a2a4 c7d8 c1b1 d8e7 e1d1 a6e2 d1c1 e8c8 b2c4');
INSERT INTO evaluations VALUES(17240,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17240,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(17240,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17240,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(17240,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(17240,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(17240,36,'cp 65','d2d4 e7e6 a2a3 d7d5 b1c3 f8e7 e2e3 e8g8 b2b4 b7b6 c1b2 c6b8 c4d5 e6d5 f1d3 b8d7 b4b5 a7a6 a3a4 c8b7 e1g1 e7d6 d1b3 a6b5 a4b5 d8e7 f1c1 f8b8 h2h3 b8e8 c3e2 f6e4 a1a8 e8a8 e2f4 e4g5 d3e2 g5f3 e2f3');
INSERT INTO evaluations VALUES(42151,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(42151,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(42151,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(42151,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(42151,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(42151,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(42151,36,'cp 58','c7c5 e1g1 e8g8 a2a3 b4c3 b2c3 b7b6 f1e1 c8b7 c4f1 d8c7 c3c4 f8d8 c1b2 b8d7 f3d2 d7f8 a3a4 f8g6 a4a5 a8c8 a5a6 b7c6 d1e2 h7h6 e1d1 c5d4 e3d4 g6h4 h2h3 c7f4 a1a3 f4g5 a3g3 g5f4 e2e3 f4e3 g3e3 h4f5 e3d3 f6e4 d2e4 c6e4');
INSERT INTO evaluations VALUES(14908,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(14908,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(14908,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(14908,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(14908,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(14908,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(14908,36,'cp 187','f8b4 c4d3 f6d7 a2a3 b4c3 b2c3 b7b6 e1g1 c8b7 f1e1 e8g8 c1g5 f7f6 g5h4 c6a5 e4e5 g7g5 f3g5 f6g5 d1h5 d8e7 h4g5 e7f7 h5h4 g8h8 e1e3 f8g8 e3g3 b7d5 f2f4 a8f8 f4f5 d5c4 f5f6 c4d3 g3d3 g8g6 a1f1 f8g8 h4h5 a5b3 d3g3 c7c5 g5h6 d7f8 h6g7 g8g7 f6g7 f7g7 g3g6');
INSERT INTO evaluations VALUES(42563,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(42563,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(42563,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(42563,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(42563,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(42563,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42563,36,'cp 4','b8c6 d2d3 d7d6 c2c3 a7a5 b1d2 e8g8 h2h3 h7h6 f1e1 c8e6 c4b5 c5a7 d2f1 c6e7 d3d4 e5d4 f3d4 d6d5 e4e5 f6e4 b5d3 e7f5 d3e4 d5e4 d4f5 e6f5 c1e3 a7e3 f1e3 f5e6 d1c2 f7f5 e5f6 d8f6 c2e4 f6f2 g1h2 f2f4 e4f4 f8f4');
INSERT INTO evaluations VALUES(51418,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(51418,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(51418,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(51418,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(51418,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(51418,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(51418,36,'cp 109','e2e4 c8b7 f1d3 c7c5 c2c3 g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 f8e7 a2a3 a7a5 a3a4 b7a6 d3b5 e8g8 c1f4 f8e8 g2g3 d7f8 h2h4 a6b5 a4b5 b8d7 b1a3 f8g6 f4c1 c5d4 c3d4 f7f6 e5f6 e7f6 h4h5 g6h8 h5h6 g7h6 c1h6 h8f7 h6f4');
INSERT INTO evaluations VALUES(16682,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(16682,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(16682,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(16682,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(16682,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(16682,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(16682,36,'cp -117','f6e4 b1c3 f8b4 d1b3 d7d5 f1d3 c8g4 f3e5 c6e5 d4e5 a7a5 d3e4 d5e4 e1g1 e8g8 b3c4 b4c3 c4c3 g4e6 h2h3 h7h6 c1e3 e6d5 a2a4 f8e8 f1d1 c7c6 e3f4 d8e7 d1d4 b7b6 b2b4 a5b4 d4b4 b6b5 b4d4 e7e6 g1h2 g8h7 a4b5 a8a1 c3a1');
INSERT INTO evaluations VALUES(16867,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(16867,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(16867,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(16867,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(16867,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(16867,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(16867,36,'cp -109','b1c3 f8b4 d1b3 d7d5 f1d3 c8g4 f3e5 c6e5 d4e5 a7a5 d3e4 d5e4 e1g1 e8g8 b3c4 b4c3 c4c3 g4e6 c3g3 d8d7 a2a4 e6d5 f1d1 d7e6 h2h3 c7c6 c1g5 g8h8 g5e3 h7h6 e3d4 h8h7 d4c3 e6f5 g1h2 g7g5 d1d2 h6h5 g3e3 f8g8 a1b1 h5h4 b1e1');
INSERT INTO evaluations VALUES(20481,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20481,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20481,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20481,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(20481,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20481,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20481,36,'cp 0','f8d6 d2d4 g8e7 d4c5 d6c5 b1d2 c8g4 d2b3 c5d6 h2h3 g4h5 f1e1 e8g8 c1e3 e7f5 e3c1');
INSERT INTO evaluations VALUES(5595,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5595,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5595,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5595,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5595,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5595,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5595,36,'cp 0','d2d4 g8e7 d4c5 d6c5 b1d2 c8g4 d2b3 c5d6 h2h3 g4h5 f1e1 e8g8 c1e3 e7f5 e3c1');
INSERT INTO evaluations VALUES(53898,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(53898,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(53898,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(53898,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(53898,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(53898,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(53898,36,'cp 114','d2d4 c5d4 c3d4 e5d4 f3d4 b8c6 b1c3 f8e7 f1c4 g8f6 e1g1 e8g8 c1e3 c8d7 f2f3 a7a6 g1h1 c6d4 d1d4 d7e6 f1d1 a8c8 c4e6 f7e6 e4e5 d6e5 d4e5 d8e8 e5e6 e8f7 e6b6 e7d8 b6b4 f8e8 e3f2 h7h6 h1g1 b7b5 b4b3 f7b3 a2b3');
INSERT INTO evaluations VALUES(53891,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(53891,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(53891,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(53891,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(53891,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(53891,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(53891,36,'cp 104','e5d4 c3d4 c5d4 f3d4 b8c6 f1c4 g8f6 e1g1 f8e7 b1c3 e8g8 c1e3 c8d7 a2a4 a7a6 h2h3 a8c8 c4a2 c6b4 a2b3 f8e8 d1e2 e7f8 f1d1 h7h6 d1d2 b4c6 f2f3 c6d4 d2d4 d7e6 b3e6 f7e6 a1d1 d8c7 e2d2 c7f7 a4a5 f8e7');
INSERT INTO evaluations VALUES(49642,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(49642,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(49642,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(49642,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(49642,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(49642,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(49642,36,'cp -46','d7d5 g1f3 f8d6 e4d5 f6d5 g2g3 h7h6 f1g2 c8e6 e1g1 c7c5 b1c3 d8d7 c3d5 e6d5 f1e1 b8c6 d3d4 c5d4 f3d4 d5g2 g1g2 c6e7 d4f3 d7c6 d1e2 f7f6 e2e4 c6e4 e1e4 a8c8 c2c3 d6c5 c1e3 c5e3 e4e3 e8f7 a1d1 h8d8 e3e1 b7b6 f3d2 d8d5 d2f1 c8d8 d1d5 d8d5');
INSERT INTO evaluations VALUES(48839,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48839,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(48839,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(48839,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(48839,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(48839,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(48839,36,'cp -39','g1f3 f8d6 e4d5 f6d5 g2g3 h7h6 f1g2 c8e6 e1g1 c7c5 b1c3 d8d7 f1e1 b8c6 c3d5 e6d5 d3d4 c5d4 f3d4 d5g2 g1g2 c6e7 d4f3 d7c6 d1d3 a8c8 c1d2 e8g8 a1d1 f8d8 d3c3 c6c3 d2c3 f7f6 e1e2 g8f7 e2d2 d6c7 c3b4 d8d2 d1d2 e7c6 b4c5 f7e6 c2c4');
INSERT INTO evaluations VALUES(53296,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(53296,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(53296,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(53296,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(53296,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(53296,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(53296,36,'cp -29','g8f6 b2b3 e7e6 c1b2 f8e7 a2a4 e8g8 g2g3 d8b6 f1g2 b8c6 e1g1 c5c4 b1c3 c4b3 c2b3 c8d7 f3e5 f8d8 g3g4 d7e8 g4g5 f6d7 e5d7 e8d7 c3e2 e7c5 h2h4 d5d4 a4a5 c6a5 b2d4 d7c6 d4c5 b6c5 d2d4 c5b5 e2c3 b5b3 d1b3');
INSERT INTO evaluations VALUES(57161,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(57161,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(57161,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(57161,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(57161,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(57161,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(57161,36,'cp -33','g8f6 b1c3 f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 f1e2 b4f8 c3e4 e8e4 e1g1 d7d5 e2f3 e4e8 f1e1 c8f5 e1e8 d8e8 d3e1 c7c6 d2d4 b8d7 g2g3 h7h6 e1g2 d7f6 c1f4 g7g5 f4e5 f6d7 g2e3 f5h7 e5c7 e8e7 c2c3 d7f6 c7e5 a8e8 e5f6 e7f6');
INSERT INTO evaluations VALUES(49668,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(49668,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(49668,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(49668,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(49668,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(49668,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(49668,36,'cp -38','b1c3 f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 c3e4 e8e4 f1e2 b4f8 e1g1 d7d5 e2f3 e4e8 f1e1 c7c6 e1e8 d8e8 d3e1 b8d7 d2d4 d7b6 c1f4 c8f5 c2c3 b6c4 b2b3 c4d6 e1c2 f5e4 f3g4 e8e7 c2e3 e4g6 f4d6 e7d6 g2g3 a8e8 g4f3 a7a5');
INSERT INTO evaluations VALUES(49619,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(49619,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(49619,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(49619,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(49619,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(49619,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(49619,36,'cp -29','f8b4 g1f3 e8g8 f3e5 f8e8 e5d3 f6e4 c3e4 e8e4 f1e2 b4f8 e1g1 d7d5 e2f3 e4e8 f1e1 c8f5 e1e8 d8e8 d3e1 c7c6 d2d4 b8d7 c1f4 e8e6 g2g3 h7h6 h2h4 a7a5 e1g2 a5a4 g2e3 d7f6 e3f5 e6f5 a2a3 a8e8 d1d3 f5d3 c2d3 g7g6 f4d2 h6h5 a1e1 e8d8');
INSERT INTO evaluations VALUES(34079,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34079,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(34079,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34079,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34079,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34079,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34079,36,'cp 0','d7d5 d2d3 e4f6 b1c3 c8e6 d3d4 b8c6 c1f4 e8c8 e1c1 a7a6 f3g5 h7h6 g5e6 e7e6 g2g3 g7g5 f4e5 c6e5 d4e5 f6d7 c3d5 c7c6 d5f6 d7f6 d1d8 c8d8 e5f6 e6a2 e2e5 a2a1 c1d2 f8b4 c2c3 a1b2 d2d3 b2b1 d3d2');
INSERT INTO evaluations VALUES(33842,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33842,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33842,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33842,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33842,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33842,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33842,36,'cp 0','d2d3 e4f6 c1f4 c8e6 d3d4 b8c6 b1c3 e8c8 e1c1 a7a6 f3g5 h7h6 g5e6 e7e6 g2g3 g7g5 f4e5 c6e5 d4e5 f6d7 c3d5 c7c6 d5f6 d7f6 d1d8 c8d8 e5f6 e6a2 e2e5 a2a1 c1d2 f8b4 c2c3 a1b2 d2d3 b2b1 d3d2');
INSERT INTO evaluations VALUES(33838,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33838,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33838,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33838,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33838,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33838,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33838,36,'cp 0','e4f6 c1f4 c8e6 d3d4 b8c6 b1c3 e8c8 e1c1 a7a6 f3g5 h7h6 g5e6 e7e6 g2g3 g7g5 f4e5 c6e5 d4e5 f6d7 c3d5 c7c6 d5f6 d7f6 d1d8 c8d8 e5f6 e6a2 e2e5 a2a1 c1d2 f8b4 c2c3 a1b2 d2d3 b2b1 d3d2');
INSERT INTO evaluations VALUES(33827,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33827,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33827,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33827,36,'cp 11','g2g3 e7e2 e1e2 f8d6 f1g2 e8g8 h1e1 f8e8 e2f1 e8e1 f1e1 c7c6 b1c3 a7a5 b2b3 b8d7 h2h3 d7e5 f3e5 d6e5 c1d2 h7h6 a1b1 e5d6 a2a4 f6e8 c3e2 c8d7 d2e3 f7f6 h3h4 g8f7 e2d4 g7g5 e1d2 e8g7 g2f3 g5g4 f3g2');
INSERT INTO evaluations VALUES(22532,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(22532,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(22532,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(22532,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(22532,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(22532,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(22532,36,'cp 57','d1b3 d8f6 b1c3 c6a5 b3a4 a5c4 a4c4 c7c6 e1g1 d7d6 f1e1 g8e7 c4a4 b7b5 a4c2 c8g4 e1e3 a8d8 h2h3 g4e6 c3e2 g7g5 e2g3 f8g7 g3h5 f6g6 h5g7 g6g7 c1d2 e8g8 a2a4 b5b4 d2b4 c6c5 b4c3 g7g6 c2e2 e7c6 a1e1 d6d5 f3e5 c6e5 c3e5');
INSERT INTO evaluations VALUES(22529,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(22529,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(22529,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(22529,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(22529,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(22529,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(22529,36,'cp 45','d8e7 b1c3 c6a5 b3b5 a5c4 b5c4 c7c6 e1g1 d7d6 c1f4 g8f6 a1d1 c8g4 f1e1 f6h5 f4d2 e7d7 c4e2 g7g5 h2h3 g4e6 d2c1 h5f4 c1f4 g5f4 c3d5 e8c8 d5f4 h8g8 g1h2 c8b8 b2b4 d6d5 e4d5 c6d5 f4e6 f7e6 e2e6 f8b4 e6d7 d8d7 e1e6 d7d6 e6d6 b4d6 g2g3 g8d8 f3d4 a7a6 f2f4 d8e8');
INSERT INTO evaluations VALUES(2020,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(2020,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(2020,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(2020,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(2020,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(2020,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(2020,36,'cp 49','b1c3 c6a5 b3b5 a5c4 b5c4 c7c6 e1g1 d7d6 c1f4 g8f6 a1d1 c8g4 f1e1 f6h5 f4d2 e7d7 c4e2 g7g5 h2h3 g4e6 d2c1 h5f4 c1f4 g5f4 c3d5 e8c8 d5f4 h8g8 g1h2 c8b8 b2b4 d7c7 g2g3 f8g7 e2d2 e6c8 f3d4 g8e8 b4b5 d6d5 e4d5 c6b5 d2d3 e8e1 d1e1');
INSERT INTO evaluations VALUES(6253,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(6253,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(6253,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(6253,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(6253,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(6253,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(6253,36,'cp 97','c2c3 d7d6 c3d4 c5b6 b1c3 g8e7 c1e3 e8g8 d1d2 c8g4 a2a4 e7g6 f3e1 c6a5 c4a2 c7c5 c3d5 g4e6 d4c5 d6c5 a1d1 e6d5 e4d5 d8d6 e1f3 c5c4 f1e1 f8e8 d2c3 b6e3 e1e3 e8e3 f2e3 b7b6 a2c4 a8c8 b2b3 g6f8 f3d4 a5c4 b3c4 f8d7 d4f5');
INSERT INTO evaluations VALUES(6251,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(6251,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(6251,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(6251,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(6251,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(6251,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(6251,36,'cp 112','d7d6 c3d4 c5b6 h2h3 g8e7 b1c3 e8g8 f1e1 c6a5 c4d3 a5c6 d3c2 f7f5 c2b3 g8h8 e4e5 f5f4 c3a4 b6a5 e5d6 c7d6 c1d2 e7g6 d2a5 c6a5 b3d5 c8f5 b2b4 a5c6 b4b5 c6b4 d5b7 a8b8 b7e4 b8b5 a4c3 b5b8 a1c1 g6h4 f3h4 d8h4 d1d2 f5e4');
INSERT INTO evaluations VALUES(16649,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(16649,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(16649,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(16649,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(16649,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(16649,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(16649,36,'cp -12','b1c3 f8b4 e1g1 e8g8 e4e5 d7d5 c4b3 f6e4 c3d5 b4c5 d2d4 c6d4 f3d4 c8e6 c2c4 c7c6 f1f4 c6d5 c1e3 d5c4 b3c2 f7f5 c2e4 c5d4 d1d4 f5e4 f4f8 g8f8 d4e4 d8d3 e4d4 f8g8 a1d1 d3d4 d1d4 c4c3 b2c3 a8c8 d4d6 g8f7 e3a7 c8c3 a7b6 c3a3 h2h3 a3a2 b6c5 a2a8 g1f2');
INSERT INTO evaluations VALUES(16645,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16645,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(16645,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(16645,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16645,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(16645,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(16645,36,'cp -5','f8b4 e1g1 e8g8 e4e5 f6g4 d2d4 d7d6 h2h3 g4e3 c1e3 f4e3 c3d5 c8e6 d5e3 d6e5 d4d5 b4c5 d5e6 c5e3 g1h1 f7e6 d1e1 e3h6 a1d1 d8e7 f3e5 f8f1 c4f1 c6e5 e1e5 g7g6 f1c4 a8e8 e5d4 c7c5 d4e4 a7a6 d1e1 b7b5 e4e6 e7e6 e1e6 g8f7 e6e8 f7e8 c4d5 h6g7 b2b3 e8e7 d5g8 h7h5 g2g3 a6a5 h1g2 e7f6 c2c4 b5c4');
INSERT INTO evaluations VALUES(10498,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(10498,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(10498,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(10498,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10498,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10498,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10498,36,'cp -17','e1g1 e8g8 e4e5 d7d5 c4b3 f6e4 c3d5 b4c5 d2d4 c6d4 f3d4 c7c6 c2c3 c6d5 c1f4 a7a5 f4e3 d8c7 g1h1 c8e6 b3c2 a8e8 e3f4 c7b6 a1b1 e8e7 f4e3 b6c7 d1h5 e6d7 e3f4 g7g6 h5h4 e7e5 f4e5 c7e5 c2e4 d5e4');
INSERT INTO evaluations VALUES(27165,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(27165,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(27165,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(27165,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(27165,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(27165,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(27165,36,'cp 25','d7c6 e1g1 g8f6 c2c4 f8e7 b1c3 e8g8 d4d3 f6d7 c1e3 d7e5 f3e5 d6e5 c3d5 c6d5 d3d5 d8d5 c4d5 f7f5 a1c1 e7d6 f2f3 f8f7 e4f5 f7f5 e3c5 f5f7 c1c3 a8f8 f1c1 a7a5 c5d6 c7d6 c3c7 h7h5 c7f7 f8f7 c1c8 f7f8 c8c4 f8f7 h2h4 g8h7');
INSERT INTO evaluations VALUES(26779,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26779,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(26779,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(26779,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(26779,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(26779,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(26779,36,'cp 38','b1c3 g8f6 c1f4 f8e7 e1c1 e8g8 h1e1 f8e8 d4c4 d8c8 f3g5 e8f8 h2h3 f6d7 g5f3 f8e8 f3d4 d7c5 c1b1 a7a6 f2f3 b7b5 c4e2 c6b7 d4f5 b5b4 f5e7 e8e7 c3d5 b7d5 d1d5 a6a5 f4e3 c5d7 e1d1 e7e8 e3d4 a5a4 e2d2 c8b7 g2g4 d7e5 d4e5 e8e5 d5e5');
INSERT INTO evaluations VALUES(26778,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(26778,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(26778,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(26778,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(26778,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(26778,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(26778,36,'cp 36','f8e7 d4g7 e7f6 g7g3 d8d7 c1g5 f6c3 b2c3 f7f6 g5h4 g8e7 h4f6 h8g8 g3h4 e8c8 f6e7 d8e8 f3d2 e8e7 e1c1 g8g2 h1g1 g2g1 d1g1 b7b6 c1b2 a7a5 g1g5 c8b7 f2f3 a5a4 a2a3 d7e6 g5h5 e6d7 d2c4 e7f7 c4e3 f7f3 h5h7 f3f7 h7f7');
INSERT INTO evaluations VALUES(29996,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(29996,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(29996,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(29996,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(29996,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(29996,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(29996,36,'cp 42','c4e6 f7e6 d2d4 e5d4 f3d4 d8d7 c1g5 f8e7 d1e2 b8c6 e1c1 c6d4 d1d4 h7h6 g5d2 e6e5 d4d3 e8g8 a2a4 c7c6 h1d1 d7g4 f2f3 g4h5 h2h3 h5f7 c1b1 f6d7 d2e3 g8h7 a4a5 b7b6 d3d6 e7d6 d1d6 a8d8 d6c6');
INSERT INTO evaluations VALUES(29992,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(29992,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(29992,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(29992,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(29992,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(29992,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(29992,36,'cp 38','f7e6 d2d4 e5d4 f3d4 d8d7 c1g5 f8e7 d1e2 b8c6 e1c1 c6d4 d1d4 h7h6 g5h4 e8g8 h1d1 a8e8 f2f3 e6e5 e2c4 g8h7 d4d3 f6h5 h4e7 e8e7 c3d5 e7f7 a2a4 a7a6 a4a5 h5f4 d5f4 f7f4 c4b3 d7b5 b3b5 a6b5 c1d2 f4f6 h2h3');
INSERT INTO evaluations VALUES(30276,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(30276,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30276,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(30276,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(30276,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(30276,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(30276,36,'cp 51','d2d4 e5d4 f3d4 d8d7 c1g5 f8e7 d1e2 b8c6 e1c1 c6d4 d1d4 h7h6 g5h4 e8g8 h1d1 e6e5 d4d2 c7c6 c1b1 d7e6 f2f3 a8d8 h4f6 e7f6 e2f2 b7b5 a2a3 d8d7 c3a2 f6d8 d2d3 d8a5 a2b4 a5b4 a3b4 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(30274,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(30274,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(30274,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(30274,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(30274,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(30274,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(30274,36,'cp 49','e5d4 f3d4 d8d7 c1g5 f8e7 d1e2 b8c6 e1c1 c6d4 d1d4 h7h6 g5h4 e8g8 h1d1 e6e5 d4d2 f8f7 h4f6 f7f6 c3d5 f6f7 c1b1 e7g5 d2d3 c7c6 d5e3 g5e3 d3e3 d7e6 a2a3 a8d8 f2f3 f7e7 h2h4 e7d7 g2g3 d8f8 e3d3 d6d5 e4d5 c6d5 d3e3');
INSERT INTO evaluations VALUES(19332,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19332,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19332,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19332,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19332,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19332,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19332,36,'cp 0','c2c4 f8g7 g1e2 d7d6 b1c3 c8e6 e2f4 g7c3 b2c3 g8f6 f2f3 d8a5 f4d5 c6e5 d1d4 a8c8 c1d2 e5c6 d5f6 e8d8 d4d3 c6e5 d3d4');
INSERT INTO evaluations VALUES(38418,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(38418,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(38418,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(38418,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(38418,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(38418,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(38418,36,'cp 52','b1c3 a7a6 g1f3 g8f6 d1d2 e8g8 h2h4 b7b5 f1d3 c7c5 d4c5 b8c6 e3h6 g7h6 d2h6 f6g4 h6f4 d6c5 d3e2 d8d6 f4d6 e7d6 a1d1 f8d8 c3d5 h7h5 f3g5 g8g7 c2c3 a8b8 e1g1 b5b4 c3b4 c6b4 d5b4 c5b4');
INSERT INTO evaluations VALUES(47039,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47039,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47039,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47039,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47039,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47039,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47039,36,'cp 0','g7g6 b1c3 c7c5 f1e2 f8g7 d3d4 c5d4 f3d4 b8c6 c1e3 e8g8 e1g1 d6d5 c3d5 f6e4 d4c6 b7c6 d5b4 g7b2 d1d8 f8d8 a1b1 e4c3 e2f3 c3b1 f1b1 b2f6 b4c6 c8e6 c6d8 a8d8 a2a4 d8c8 e3a7 c8c2 h2h3 c2a2 f3c6 a2c2');
INSERT INTO evaluations VALUES(46726,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46726,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46726,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(46726,36,'cp -7','b1c3 c7c5 f1e2 f8g7 d3d4 e8g8 e1g1 c5d4 f3d4 b8c6 c1e3 d6d5 c3d5 f6e4 d4c6 b7c6 d5b4 g7b2 d1d8 f8d8 a1b1 e4c3 e2f3 c3b1 f1b1 b2f6 b4c6 c8e6 c6d8 a8d8 b1b7 a7a5 b7a7 f6d4 a7a5 d4e3 f2e3 d8d2 a5a7 d2c2 h2h4 g8f8 a2a4 c2a2 a4a5 a2a1 g1h2');
INSERT INTO evaluations VALUES(46723,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46723,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46723,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46723,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(46723,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46723,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(46723,36,'cp 0','c7c5 f1e2 b8c6 c3d5 f8g7 c2c3 e8g8 e1g1 f6d7 c1g5 h7h6 g5h4 d7b6 f3d2 c8e6 d5f4 e6d7 f4d5');
INSERT INTO evaluations VALUES(44764,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44764,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44764,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44764,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44764,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44764,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44764,36,'cp 0','f1e2 b8c6 c3d5 f8g7 c2c3 e7e5 c1g5 h7h6 g5f6 g7f6 h2h4 h6h5 f3g5 f6g7 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(51491,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(51491,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(51491,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51491,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(51491,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51491,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51491,36,'cp 0','c8b7 g1f3 e7e6 f1d3 g8f6 e1g1 d7d5 c2c4 f8d6 b2b3 e8g8 c4d5 e6d5 c1a3 c7c5 b1c3 f8e8 a1c1 b8d7 a3b2 d8e7 c3b5 d6b8 b5c3 b8d6');
INSERT INTO evaluations VALUES(30802,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(30802,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(30802,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(30802,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(30802,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30802,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30802,36,'cp 6','g1f3 e7e6 f1d3 d7d5 c2c4 f8d6 e1g1 g8f6 b2b3 b8d7 c4d5 e6d5 c1a3 c7c5 b1c3 e8g8 a1c1 f8e8 c3b5 d6f8 a3b2 a7a6 b5c3 f8d6 c3e2 d8e7 f1e1 a8c8 e2g3 g7g6 d3f1 h7h5 d1d2 f6e4 g3e4 d5e4 d4c5 d6c5');
INSERT INTO evaluations VALUES(30803,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(30803,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(30803,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(30803,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30803,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(30803,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30803,36,'cp 0','e7e6 f1d3 g8f6 c2c4 d7d5 e1g1 f8d6 b2b3 b8d7 c4d5 f6d5 b1d2 e8g8 c1b2 c7c5 a1c1 a8c8 d3b1 d6b8 e3e4 d5f4 g2g3 f4h3 g1g2 h3g5 d1e2 f7f5 g2g1 g5h3 g1g2');
INSERT INTO evaluations VALUES(30663,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30663,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30663,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(30663,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(30663,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30663,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(30663,36,'cp 12','c2c4 g8f6 f1d3 d7d5 e1g1 f8d6 b2b3 e8g8 b1c3 b8d7 c4d5 e6d5 c3b5 d6b4 c1a3 b4a3 b5a3 d8e7 a3b5 f8c8 a1c1 c7c5 b5c3 f6e4 d4c5 e4c3 c1c3 b6c5 d3f5 c8c7 d1c2 d7f6 f5d3 a8c8 f1c1 g7g6 c2b2 a7a6 d3e2 c5c4');
INSERT INTO evaluations VALUES(30646,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30646,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30646,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(30646,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(30646,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(30646,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30646,36,'cp 0','d7d5 f1d3 d5c4 d3c4 g8f6 e1g1 a7a6 b1c3 b6b5 c4b3 c7c5 e3e4 c5d4 f3d4 b8c6 c1e3 c6d4 e3d4 f8e7 e4e5 f6e4 c3e4 b7e4 f1e1 e4g6 d1f3 e8g8 e1d1 d8b8 a1c1 a6a5 c1c6 b8b7 c6b6 b7f3 g2f3 b5b4 b6b7');
INSERT INTO evaluations VALUES(30575,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(30575,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(30575,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(30575,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(30575,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(30575,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(30575,36,'cp 9','d1a4 c7c6 c4d5 e6d5 f1d3 g8f6 e1g1 f8e7 a4c2 c6c5 b2b3 e8g8 d4c5 b6c5 b1d2 h7h6 c1b2 f8e8 a1d1 d8b6 f1e1 b8c6 a2a3 a7a5 d1c1 e8d8 e1d1 a5a4 b2f6 e7f6 c2c5 b6c5 c1c5 a4b3 d2b3 a8a3 b3d4 c6d4 f3d4 d8c8 c5c8 b7c8 d4c6 f6c3 d1c1 c3f6');
INSERT INTO evaluations VALUES(21843,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(21843,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(21843,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(21843,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(21843,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(21843,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(21843,36,'cp 140','f8d6 b1c3 a7a6 c1g5 g8f6 e2e3 h7h6 g5h4 e8g8 a2a3 f8e8 f1e2 b7b5 e1g1 c8b7 h4g3 d6g3 h2g3 d8e7 c3d5 f6d5 d1d5 a8d8 d5b3 c6e5 f1d1 e5f3 e2f3 c7c5 g1f1 b7f3 g2f3 e7b7 f1g2 c5c4 d1d8 c4b3');
INSERT INTO evaluations VALUES(5913,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(5913,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(5913,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(5913,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(5913,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(5913,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(5913,36,'cp 150','b1c3 g8f6 c3b5 e8g8 c2c3 f6e4 b5d6 c7d6 c1e3 f8e8 g2g3 c8g4 f1g2 d8d7 e1g1 d6d5 f3h4 d7e6 f1e1 h7h6 h4f3 e6d7 d1c2 g4f5 c2a4 b7b5 a4d1 c6a5 e3d4 a5c4 b2b3 c4d6 f3h4 f5h3 f2f3 e4g5 d1d2 h3g2 g1g2 d7h3 g2g1 e8e6');
INSERT INTO evaluations VALUES(5911,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(5911,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(5911,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(5911,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(5911,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(5911,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(5911,36,'cp 154','g8f6 c3b5 e8g8 c2c3 f6e4 b5d6 c7d6 c1e3 f8e8 g2g3 c8e6 f1g2 d8d7 e1g1 h7h6 f3h4 e4f6 f1e1 a8d8 e3d4 c6d4 d1d4 a7a5 h4f3 a5a4 d4b4 e6d5 b2b3 d5e4 a1d1 d7c7 f3d4 e4g2 g1g2 a4b3 a2b3 c7c5 b4b5');
INSERT INTO evaluations VALUES(9208,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(9208,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(9208,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(9208,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(9208,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(9208,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(9208,36,'cp 148','c3b5 e8g8 c2c3 c8f5 b5d6 c7d6 c1f4 d8a5 e2e3 f6d5 f1e2 d5f4 e3f4 f8e8 e1g1 d6d5 b2b4 a5c7 d1d2 a8d8 a1d1 h7h6 h2h3 f5e4 f1e1 a7a6 a2a4 c7d6 b4b5 c6a5 f3d4 a5c4 d2c1 a6b5 a4b5 d6f6 e2f1 c4d6 f2f3 e4h7 c1d2 e8e1 d1e1');
INSERT INTO evaluations VALUES(38835,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(38835,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(38835,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(38835,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(38835,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(38835,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(38835,36,'cp 59','h2g3 c8d7 b1c3 b8c6 f1d3 h7h6 g1f3 g8f6 e1g1 c6b4 d3e2 e8g8 f1e1 b4c6 d1d3 a8c8 a2a3 c6e7 a1d1 c7c5 d4c5 c8c5 e3e4 d5e4 c3e4 f6e4 d3e4 d7c6 e4e3 d8b6 e2d3');
INSERT INTO evaluations VALUES(23299,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(23299,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(23299,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(23299,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(23299,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(23299,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(23299,36,'cp 168','d2d4 f5e4 f3e5 f8b4 d1h5 g7g6 e5c6 g6h5 c6d8 e8d8 f1c4 c7c6 d4d5 g8e7 c1g5 h7h6 g5h4 d8c7 e1c1 e7f5 h4f6 h8e8 h1e1 d7d6 a2a3 b4c3 f6c3 h5h4 d5c6 b7c6 c4d3 d6d5 f2f3 e4e3 d3f5 c8f5 c3d4 e8e6 e1e3 e6e3 d4e3 h6h5 d1e1 a8g8 e1e2 a7a5 e3c5 a5a4 c1d2 c7d7');
INSERT INTO evaluations VALUES(23297,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(23297,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(23297,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(23297,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(23297,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(23297,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(23297,36,'cp 172','f5e4 f3e5 f8b4 d1h5 g7g6 e5c6 g6h5 c6d8 e8d8 f1c4 g8e7 c1g5 c7c6 d4d5 h7h6 g5h4 d8c7 e1c1 e7f5 h4f6 h8e8 h1e1 d7d6 a2a3 b4c3 f6c3 h5h4 d5c6 b7c6 c4d3 d6d5 f2f3 e4e3 d3f5 c8f5 c3d4 e8e6 e1e3 c7d6 e3e6 f5e6 d1g1 c6c5 d4f2 h4h3 g2g4 a8f8');
INSERT INTO evaluations VALUES(5366,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(5366,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(5366,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(5366,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(5366,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(5366,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(5366,36,'cp -5','a2a4 a7a6 d2d3 d7d6 f1e2 e7e6 e1g1 g8e7 d1e1 c8d7 c3d1 e8g8 c2c3 b7b5 d1e3 b5a4 e2d1 c6a5 d1a4 d7a4 a1a4 a5b3 a4a3 d8b8 e1d1 b3c1 d1c1 b8b5 c1d2 d6d5 g2g3 f8b8 a3a2 a6a5 f1a1 h7h5 d2e2');
INSERT INTO evaluations VALUES(17977,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(17977,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(17977,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(17977,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(17977,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(17977,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(17977,36,'cp 32','g8f6 a2a4 c8g4 c2c3 f8e7 h2h3 g4d7 d2d4 e8g8 f1e1 h7h6 b1d2 e5d4 c3d4 c6b4 e4e5 f6d5 a4b5 a6b5 a1a8 d8a8 d2e4 d7e6 e5d6 c7d6 e4g3 a8c6 d1e2 f8e8 c1d2 e7f8 e2e4 e6d7 e4b1 e8e1 b1e1 b4c2 e1b1 c2b4 g3h5');
INSERT INTO evaluations VALUES(25446,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(25446,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(25446,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(25446,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(25446,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(25446,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(25446,36,'cp -37','d8d7 e1g1 e8c8 d2d4 f6d5 d4e5 h7h6 d1h5 g7g6 e5e6 f7e6 h5h3 a5c6 b1c3 d5c3 b2c3 f8d6 g5e4 d6e7 h3g4 c6e5 g4e2 d7c6 c1f4 e5c4 e4d2 c4d2 f4d2 b7b6 f1e1 g6g5 a2a4 e7f6 a4a5 d8d2 e2d2 f6c3');
INSERT INTO evaluations VALUES(27469,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(27469,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(27469,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(27469,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(27469,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(27469,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(27469,36,'cp -38','e1g1 e8c8 d2d4 e5d4 d1d4 a5c6 d4f4 f6d5 f4f7 c6d4 b1c3 d4c2 a1b1 c2a3 b1a1 d5c3 b2c3 h7h6 f7b3 h6g5 c1a3 f8a3 b3a3 c8b8 a1e1 h8e8 h2h3 b7b6 c3c4 d7d4 a3b3 e8e1 f1e1 a7a5 e1b1 d4d3 c4c5 d3b3 b1b3 d8d1 g1h2 b8b7 c5b6 c7b6 b3g3');
INSERT INTO evaluations VALUES(27468,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(27468,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(27468,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(27468,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(27468,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(27468,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(27468,36,'cp -50','e8c8 d2d4 e5d4 d1d4 a5c6 d4f4 f6d5 f4f7 c6d4 b1c3 d4c2 a1b1 c2a3 c3d5 a3b1 d5f4 b1d2 c1d2 d7d2 f4e6 f8d6 g5e4 d8d7 f7d7 c8d7 e4d2 d7e6 f1e1 e6d5 e1d1 b7b5 g1f1 h8c8 d2f3 d5c6 f3g5 a7a5 g5h7 c8h8 h7g5 h8h2 b2b3 h2h1 f1e2 h1d1 e2d1 c6d5 f2f3 d5d4 g5e6 d4c3 e6g7 c3b2');
INSERT INTO evaluations VALUES(6281,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(6281,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(6281,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(6281,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(6281,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(6281,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(6281,36,'cp 198','b1c3 g8f6 d4d5 c6a5 c4d3 b4c3 b2c3 d7d6 e1g1 e8g8 c3c4 f8e8 d1d2 c7c6 f1e1 c6d5 c4d5 b7b5 d2b4 a7a6 c1b2 a5b7 a1c1 b7c5 d3b1 a6a5 b4d4 c5a4 b2a1 b5b4 h2h3 c8a6 f3h4 a6c8 e1e3 c8d7 e3g3 e8e5 h4f3 f6h5 f3e5 h5g3 f2g3 d6e5 d4e5');
INSERT INTO evaluations VALUES(6279,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(6279,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(6279,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(6279,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(6279,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(6279,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(6279,36,'cp 214','g8f6 d4d5 b4c3 b2c3 c6a5 c4d3 d7d6 e1g1 e8g8 c3c4 f8e8 d1d2 c7c6 f1e1 c6d5 c4d5 b7b5 d2b4 a7a6 c1b2 a5b7 a1c1 b7c5 d3b1 a6a5 b4d4 c5d7 d4d2 d7b6 f3d4 c8d7 b1d3 b6a4 d4c6 d7c6 d5c6 a4b2 d3b5 e8e4 e1e4 f6e4 d2b2');
INSERT INTO evaluations VALUES(9873,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(9873,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(9873,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(9873,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(9873,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(9873,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(9873,36,'cp 199','d4d5 b4c3 b2c3 c6a5 c4d3 d7d6 e1g1 e8g8 c3c4 f8e8 d1d2 c7c6 f1e1 b7b5 c4b5 c6b5 d2b4 a7a6 c1b2 a5b7 a1c1 c8d7 d3b1 a8c8 c1c8 d7c8 h2h3 b7c5 b1c2 a6a5 b4d4 d8b6 e1e3 c5a4 c2a4 b6d4 b2d4 b5a4 d4f6 g7f6 f3d4 c8d7 e3c3 g8g7 f2f4 e8c8 c3c8');
INSERT INTO evaluations VALUES(10218,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10218,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10218,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10218,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10218,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10218,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10218,36,'cp 1','d7d5 c1g5 d5e4 d3e4 d8d1 a1d1 c5b4 g5d2 c8g4 a2a3 b4c3 d2c3 f6e4 c3e5 c6e5 f3e5 g4e2 e1e2 e4d6 e2f3 e8c8 h1e1 h8e8 g2g3 e8e7 h2h4 d8e8 e5d3 e7e1 d3e1 b7b6 a3a4 g7g6 b2b3 h7h5 e1d3 a7a5 d3f4 d6f5 d1d5');
INSERT INTO evaluations VALUES(14274,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14274,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(14274,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(14274,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(14274,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(14274,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14274,36,'cp 29','g1f3 e7e6 c3c4 d5c4 b1c3 f8d6 f1c4 e8g8 e1g1 h7h6 e3e4 e6e5 d4d5 c6a5 c4d3 c7c6 f1e1 f8e8 c1d2 c8d7 a2a3 c6d5 c3d5 a5c6 d2c3 d7g4 h2h3 g4h5 a1c1 c6d4 g2g4 d4f3 d1f3 f6d5 e4d5 h5g6 d3g6 f7g6 c3b4 d6b4 a3b4');
INSERT INTO evaluations VALUES(14275,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14275,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14275,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(14275,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(14275,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(14275,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(14275,36,'cp 22','e7e6 c3c4 d5c4 f1c4 f8d6 b1c3 e8g8 e3e4 e6e5 d4d5 c6a5 c4d3 c7c6 c1g5 d6b4 e1g1 b4c3 b2c3 h7h6 g5f6 d8f6 h2h3 c8d7 f1e1 b7b6 e1e3 a5b7 c3c4 f8e8 a1b1 b7d6 f3d2 c6d5 e4d5 d7f5 d1e2 a8c8 b1e1 f5d3 e2d3');
INSERT INTO evaluations VALUES(14847,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(14847,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(14847,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(14847,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(14847,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(14847,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(14847,36,'cp 21','c3c4 d5c4 f1c4 f8d6 b1c3 e8g8 e3e4 e6e5 d4d5 c6e7 e1g1 a7a6 a2a4 c7c6 d5c6 e7c6 h2h3 d6c5 c1g5 h7h6 g5h4 d8d1 f1d1 f6h5 c3e2 g7g5 h4g3 g5g4 h3g4 c8g4 g3h2 a8c8 d1d5 c5a7 d5d6 g4f3 g2f3 c6d4 e2d4 a7d4 d6h6 c8c4 h6h5');
INSERT INTO evaluations VALUES(22034,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(22034,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(22034,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(22034,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(22034,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(22034,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(22034,36,'cp 96','c1f4 g8f6 b1c3 f8e7 a2a3 c8g4 d1d3 d6d5 e1c1 d5e4 c3e4 f6d5 f4d2 e8g8 c2c4 d5f6 d4d5 c6b8 d3c2 f6e4 c2e4 g4h5 e4c2 h5g6 f1d3 g6d3 c2d3 b8d7 c1b1 f8e8 h2h4 d7c5 d3c2 d8d6 d1e1 h7h6 g2g4 d6g3 c2f5 g7g6 f5f4 e7d6 f4g3 d6g3 e1e8 a8e8 d2h6 c5e4 g4g5 e4d6');
INSERT INTO evaluations VALUES(22032,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(22032,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22032,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(22032,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(22032,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(22032,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(22032,36,'cp 89','g8f6 b1c3 f8e7 a2a3 c8g4 d1d3 d6d5 e1c1 d5e4 c3e4 f6d5 f4d2 e8g8 c2c3 f8e8 d3c2 g4h5 f1d3 h5g6 g2g4 d8d7 h1g1 d5f6 e4f6 e7f6 h2h4 d7d5 h4h5 g6d3 c2d3 d5a2 g4g5 f6e7 h5h6 g7g6 d1e1 c6a5 f3e5 a5b3 c1c2');
INSERT INTO evaluations VALUES(15645,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(15645,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(15645,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(15645,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(15645,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(15645,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(15645,36,'cp 91','b1c3 f8e7 a2a3 c8g4 d1d3 d6d5 e1c1 d5e4 c3e4 f6d5 f4g3 e8g8 c2c4 d5f6 d4d5 c6b8 h2h3 g4h5 e4f6 e7f6 d3c2 b8d7 f1d3 h7h6 c4c5 f6e5 g3f2 d8f6 h1e1 h5f3 g2f3 b7b6 d3h7 g8h8 c5c6 d7c5 c1b1 a8e8 h7f5 c5a4 f3f4 e5f4 f2d4 f4e5 f5d7 e5d4 e1e8 f8e8 d7e8 a4b2');
INSERT INTO evaluations VALUES(15644,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(15644,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(15644,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(15644,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(15644,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(15644,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(15644,36,'cp 114','f8e7 d1e2 d6d5 e4e5 f6h5 f4e3 c8e6 e1c1 d8d7 e2e1 h7h6 c1b1 g7g5 h2h4 g5g4 f3d2 e8c8 d2b3 f7f6 f1b5 a7a6 b5a4 c8b8 a2a3 b7b6 h1f1 d8f8 g2g3 b8b7 e3g1 f6f5 g1e3 f5f4 e3f4 h5f4 g3f4 h6h5 c3e2 b7b8');
INSERT INTO evaluations VALUES(8656,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(8656,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(8656,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(8656,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(8656,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(8656,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(8656,36,'cp 120','d1e2 d6d5 e4e5 f6h5 f4e3 c8e6 h1g1 d8d7 g2g4 e6g4 c3d5 d7d5 g1g4 e8c8 c2c3 f7f6 f1g2 d5f7 e3d2 f6f5 f3g5 f7g6 g2c6 h7h6 g5e6 g6g4 e2g4 f5g4 e6d8 h8d8 c6b5 g4g3 h2g3 h5g3 b5c4 h6h5 e1f2 h5h4 d2f4 g3h5 f4h2 e7g5 a1f1');
INSERT INTO evaluations VALUES(35266,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(35266,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(35266,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(35266,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(35266,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(35266,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(35266,36,'cp 191','b8c6 c1f4 d7d6 e5d6 e7f6 d1c1 f8d6 f4d6 f6d6 c1d2 d6e7 b1c3 g8f6 e1c1 e8g8 d2f4 a7a6 e2e4 c8e6 f1d3 h7h6 a2a3 a8d8 h1e1 f6g4 f4g3 e7c5 e1e2 f7f6 e2d2 g4e5 d3e2 e5f3 g3f3 c6d4 f3e3 d4e2 e3e2 d8d2 d1d2 f8e8 h2h3 c5g5');
INSERT INTO evaluations VALUES(34601,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(34601,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(34601,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(34601,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(34601,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(34601,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(34601,36,'cp -13','g8f6 f3d5 f6d5 f1c4 d5b6 c4b3 b8c6 b1c3 c8d7 d2d3 c6a5 c1f4 a8c8 c3e4 a7a6 g1e2 h7h5 h2h4 e7e6 f4e3 a5b3 a2b3 b6d5 e3c5 d5e7 f2f3 f7f6 e1f2 e8f7 d3d4 c8a8 c5e7 f8e7 e4c5 d7c6 b3b4 e6e5 d4e5 f6e5 c5d3');
INSERT INTO evaluations VALUES(33695,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(33695,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(33695,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(33695,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(33695,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(33695,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(33695,36,'cp -17','f3d5 f6d5 f1c4 d5b6 c4b3 b8c6 b1c3 c6a5 d2d3 c8d7 c1f4 a8c8 c3e4 a7a6 g1e2 f7f6 f4e3 a5b3 a2b3 e8f7 e3c5 c8a8 f2f3 b6c8 d3d4 h7h5 e1f2 b7b6 c5a3 g7g6 c2c4 a6a5 e2c3 f8h6 d4d5 c8d6 c4c5 d6e4 c3e4 b6c5');
INSERT INTO evaluations VALUES(17239,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17239,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(17239,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(17239,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(17239,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(17239,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17239,36,'cp 57','d2d4 d7d5 g1f3 e7e6 a2a3 f6e4 d1c2 e4c3 c2c3 a7a5 c1f4 f8d6 f4g3 d5c4 c3c4 h7h5 g3h4 f7f6 e2e3 g7g5 h4g3 c6e7 f1d3 h5h4 g3d6 c7d6 h2h3 d6d5 c4c3 c8d7 e1g1 e8f7 b2b4 a5b4 c3b4 d7c6 f1c1 a8a4 b4c5 d8d7');
INSERT INTO evaluations VALUES(17244,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17244,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(17244,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17244,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(17244,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(17244,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(17244,36,'cp 71','d7d5 g1f3 e7e6 a2a3 f8e7 c1f4 f6e4 e2e3 e4c3 b2c3 e8g8 c4d5 e6d5 f1d3 e7d6 f4d6 d8d6 e1g1 c6a5 f1e1 b7b6 d1c2 h7h6 c2e2 c8g4 e3e4 g4f3 g2f3 d6f4 e2e3 f4e3 f2e3 d5e4 f3e4 c7c5 e4e5');
INSERT INTO evaluations VALUES(22642,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22642,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22642,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22642,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(22642,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22642,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22642,36,'cp 0','a7a6 b5a4 g8f6 g1f3 f8c5 e1g1 e8g8 f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 c3d5 f6e4 f1e1 e5h2 g1h2 d8h4 h2g1 h4f2 g1h1 f2h4 h1g1');
INSERT INTO evaluations VALUES(17888,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17888,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17888,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17888,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17888,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17888,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17888,36,'cp 0','b5a4 f8c5 g1f3 g8f6 e1g1 e8g8 f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 c3d5 f6e4 f1e1 e5h2 g1h2 d8h4 h2g1 h4f2 g1h1 f2h4 h1g1');
INSERT INTO evaluations VALUES(26684,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(26684,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(26684,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(26684,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(26684,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(26684,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(26684,36,'cp 85','d4d5 c6e5 b1c3 g8f6 c1e3 c7c6 d5c6 e5c6 f1c4 d8d1 a1d1 a8c8 e1g1 e7e6 d1c1 g4f3 g2f3 f8e7 c4b5 e8g8 b5c6 c8c6 e3a7 f8a8 a7d4 f6h5 f1d1 h5f4 g1f1 f4g6 d4e3 e7f6 f3f4 f6c3 c1c3 c6c3 b2c3 h7h6 d1d7 a8a2 d7b7 e6e5');
INSERT INTO evaluations VALUES(26516,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(26516,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(26516,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(26516,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(26516,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(26516,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(26516,36,'cp 62','c6e5 c1f4 e5g6 f4e3 g8f6 b1c3 e7e5 f1c4 f8d6 h2h3 g4d7 f3d2 e8g8 e1g1 d8e7 f1e1 c7c6 c4f1 h7h6 a2a3 b7b6 d2c4 c6d5 e4d5 a8d8 a3a4 d6b8 d1b3 e5e4 a1d1 g6e5 e3d4 e5c4 b3c4 d8c8 c4b3 f8d8 d4f6 e7f6');
INSERT INTO evaluations VALUES(26696,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(26696,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(26696,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(26696,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(26696,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(26696,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(26696,36,'cp 46','c1f4 e5g6 f4e3 e7e5 b1c3 a7a6 f1c4 g6h4 h1g1 g8f6 c4e2 h4g6 d1c2 g4d7 f3d2 f8d6 e1c1 b7b5 d2b3 b5b4 c3b1 a6a5 b3c5 a5a4 b1d2 e8g8 g1e1 g6h4 e2f1 f6g4 c1b1 g4e3 e1e3 h4g6 c5d7 d8d7 d2c4 f8b8');
INSERT INTO evaluations VALUES(50374,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(50374,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(50374,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(50374,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(50374,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(50374,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(50374,36,'cp -11','d7d5 g2g3 c7c5 c2c4 d5c4 d1a4 c8d7 a4c4 e7e6 f3e5 b8c6 e5d7 d8d7 f1g2 f8e7 c4a4 f6d5 b1c3 e8g8 e1g1 f8d8 c1d2 a7a6 f1c1 b7b5 a4d1 h7h6 c3d5 e6d5 d2e3 d5d4 e3d2 a8c8 d1b3 c6e5 h2h4 e5c6 a2a4 c8b8');
INSERT INTO evaluations VALUES(6448,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(6448,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(6448,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(6448,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(6448,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(6448,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(6448,36,'cp -79','d7d6 d2d3 g8e7 c1e3 c6d4 f1e2 e8g8 e3d4 c5d4 e1g1 f7f5 f3d4 e5d4 c3d5 e7g6 e4f5 c8f5 f2f4 f5e6 e2g4 e6d5 c4d5 f8f4 f1f4 g6f4 g2g3 f4g6 d1a4 g6e5 g4e6 g8h8 a1f1 e5d3 a4d4 d3e5 h2h4 d8e7 g1g2 a7a5 f1f2 a8f8 h4h5 h7h6 f2f8 e7f8 d4e4');
INSERT INTO evaluations VALUES(5996,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(5996,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(5996,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(5996,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(5996,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(5996,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(5996,36,'cp -67','d2d3 g8e7 c1e3 c6d4 f1e2 e8g8 e3d4 c5d4 f3d4 e5d4 c3d5 e7g6 d1d2 f7f5 d5f4 g6e5 e4f5 c8f5 e1g1 d8g5 a1d1 a8e8 f4h5 g5d2 d1d2 c7c5 a2a3 b7b6 h2h3 e5g6 e2g4 f5g4 h3g4 g6e5 f2f3 e5g6 h5g3 g6f4 g3f5 d6d5 c4d5 f4d5 f5d6 e8e6');
INSERT INTO evaluations VALUES(5997,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(5997,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(5997,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(5997,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(5997,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(5997,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(5997,36,'cp -61','g8e7 a2a3 a7a5 c1e3 f7f5 d3d4 c6d4 f3d4 f5f4 e3f4 c5d4 f4e3 d4e3 f2e3 c8e6 c4c5 e8g8 c5d6 c7d6 f1e2 e7c8 d1d2 c8b6 e1c1 b6c4 e2c4 e6c4 d2d6 d8g5 c1b1 b7b5 c3d5 b5b4 d1d2 g5h4 d6e5 a8e8 g2g3 e8e5 g3h4 b4a3 d5b6 c4e6 b6d7 e6d7 d2d7 e5e4 b2a3 e4e3');
INSERT INTO evaluations VALUES(6471,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(6471,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(6471,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(6471,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(6471,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(6471,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(6471,36,'cp -53','e5d4 c3d5 g8e7 f1c4 e8g8 c1g5 g8h8 g5e7 c6e7 b2b4 c5d6 e4e5 e7d5 e5d6 d5b4 e1g1 b7b6 f3d4 c8b7 c2c3 b4c6 d6c7 d8c7 d4b5 c7f4 b5d6 c6a5 d6b7 a5c4 d1d7 a8c8 a1d1 h8g8 f1e1 h7h6 g2g3 f4f6 b7d6 c4d6 d1d6 f6c3');
INSERT INTO evaluations VALUES(6567,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(6567,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(6567,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(6567,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(6567,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(6567,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(6567,36,'cp -71','c3d5 g8e7 f1c4 h7h6 e1g1 e8g8 b2b4 c6b4 d5b4 c5b4 d1d4 e7c6 d4e3 d7d6 c1b2 f8e8 g1h1 c8e6 c4b5 d6d5 b5c6 b7c6 b2g7 g8g7 e3d4 d8f6 d4b4 a8b8 b4a4 d5e4 a4e4 e6d5 e4d3 b8d8 f1g1 d5f3 d3f3 f6f3 g2f3 g7f6 a1d1 d8d1 g1d1 e8e2 d1d7 e2c2 d7c7 c2a2 c7c6 f6g7');
INSERT INTO evaluations VALUES(6580,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(6580,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(6580,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(6580,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(6580,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(6580,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(6580,36,'cp -102','g8e7 f1c4 h7h6 a2a3 a7a5 d5f4 e8g8 g2g4 g7g5 f4h5 d7d5 e4d5 e7g6 d5c6 b7b5 h2h4 b5c4 h4g5 d8e7 d1e2 c8g4 e2e7 c5e7 f3d4 f8e8 c1e3 h6g5 e1d2 a8d8 d2c3 f7f5 d4b5 g8f7 e3d4 g6h4 h5g3 f5f4 b5c7 f4g3 f2g3 h4f5');
INSERT INTO evaluations VALUES(14316,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(14316,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(14316,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(14316,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(14316,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(14316,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(14316,36,'cp 130','e4e5 f6g8 c3e2 f7f6 f2f4 c8g4 h2h3 g4f5 e2g3 g8h6 g1f3 e7e6 c1e3 f5e4 f1d3 f6f5 c2c3 f8e7 e1g1 e8g8 b2b4 g8h8 a2a4 a7a6 g3e2 d8e8 g1h1 f8g8 a4a5 g8f8 d1d2 h6g8 e2g3 a8c8 e3f2 e4d3 d2d3');
INSERT INTO evaluations VALUES(14349,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(14349,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(14349,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(14349,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(14349,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(14349,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(14349,36,'cp 102','f6g8 c3e2 f7f6 f2f4 c8g4 h2h3 g4f5 c1e3 e7e6 e2g3 g8h6 g1f3 f5e4 f1d3 h6f5 g3f5 e6f5 c2c3 f8e7 e1g1 e8g8 d3e2 g8h8 f3d2 a7a5 g1h2 a5a4 h2h1 f6e5 d4e5 g7g5 d2e4 d5e4 e2b5 d8e8 f4g5 e8g6 e5e6 g6e6 b5a4 e6g6 a4c6 b7c6 d1d4 h8g8 a1d1');
INSERT INTO evaluations VALUES(16129,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(16129,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(16129,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(16129,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(16129,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(16129,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(16129,36,'cp -60','f6e4 c4b5 e4f6 f3e5 c6e5 f1e1 f8e7 e1e5 c7c6 b5f1 d7d5 d2d3 e8g8 h2h3 d8c7 e5e1 e7d6 c1e3 c8d7 a2a4 a8e8 a4a5 c6c5 b1d2 e8e6 d2f3 a7a6 e3d2 e6e1 d2e1 f8e8 e1d2 d7b5 b2b4 c5c4 f3d4 d6h2 g1h1 h2e5 d4b5 a6b5 d3c4 d5c4');
INSERT INTO evaluations VALUES(16197,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(16197,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(16197,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(16197,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(16197,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(16197,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(16197,36,'cp -51','c4b5 f8e7 f1e1 e4f6 f3e5 c6e5 e1e5 c7c6 b5f1 d7d5 e5e1 e8g8 h2h3 f8e8 d2d3 d8c7 b1d2 c8f5 d2f3 e7d6 e1e8 a8e8 a2a4 f5g6 f3h4 g6h7 h4f3 a7a5 c1e3 c6c5 b2b3 b7b6 d3d4 f6e4 c3c4 c5d4 f3d4 d5c4 f1c4');
INSERT INTO evaluations VALUES(23117,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(23117,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(23117,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(23117,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(23117,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(23117,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(23117,36,'cp 205','f3e5 f8c5 e5d3 c5b6 c3d5 g8f6 d5b6 a7b6 c2c3 d4e6 f2f3 d7d5 d3f2 c7c5 f1b5 c8d7 b5d7 d8d7 e1g1 d5e4 f3e4 e8g8 d2d3 b6b5 c1d2 d7c6 d1e2 f6d7 a2a3 d7e5 d2e3 f7f6 f2g4 e5g6 a1e1 a8e8 g4f2 g6e5');
INSERT INTO evaluations VALUES(23064,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(23064,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(23064,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(23064,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(23064,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(23064,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(23064,36,'cp 222','d7d6 e5f3 c8g4 f1e2 g4f3 e2f3 g7g6 e1g1 c7c6 d2d3 f8h6 c1h6 g8h6 f3e2 d8g5 c3b1 e8g8 b1d2 f7f5 c2c3 d4e6 d2f3 g5e7 f1e1 a8e8 e4f5 h6f5 d1d2 f5h4 f3h4 e7h4 g2g3 h4h3');
INSERT INTO evaluations VALUES(30775,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(30775,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30775,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(30775,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(30775,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(30775,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30775,36,'cp 24','e7e6 g1f3 g8f6 e4e5 d7d5 c4f1 f6d7 d2d4 c7c5 c3e2 b6b5 a2a3 b8c6 c2c3 a7a5 e2f4 b5b4 h2h4 d8b6 a3b4 a5b4 a1a8 b7a8 f1e2 f8e7 e1g1 a8b7 c1e3 b7a6 d4c5 e7c5 e3c5 a6e2 c5b6 e2d1 f3d4 d1a4 b6c7 b4c3 b2c3 e8g8 f1a1 f8a8 f4d3 h7h6');
INSERT INTO evaluations VALUES(48487,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(48487,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48487,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(48487,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(48487,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48487,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(48487,36,'cp 16','a7a6 a2a4 c7c5 f1c4 b8c6 e1g1 f8e7 d1e2 c5d4 f1d1 e6e5 e3d4 e5d4 f3d4 c6d4 e2e3 e8g8 d1d4 d8e8 c3e4 e8c6 c1d2 c8f5 e4g3 f5e6 c4e6 c6e6 g3f5 e6e3 d2e3 e7c5 d4d2 c5e3 f5e3 f8d8 a1d1 d8d2 d1d2 g8f8 g1f1 a8e8 d2d4 e8e4 d4e4 f6e4');
INSERT INTO evaluations VALUES(43676,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(43676,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43676,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43676,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(43676,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43676,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(43676,36,'cp 10','a2a4 c7c5 f1c4 c5d4 e3d4 b8c6 e1g1 f8e7 d4d5 f6d5 c3d5 e6d5 c4d5 e8g8 c1e3 c8f5 h2h3 a8c8 d1b3 c6b4 f1d1 d8d5 d1d5 f5c2 b3b4 e7b4 d5d7 c2e4 d7d4 e4f3 g2f3 b4c5 d4d7 b7b5 a4b5 a6b5 d7b7 b5b4 e3c5 c8c5 b7b4');
INSERT INTO evaluations VALUES(52840,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(52840,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52840,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(52840,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(52840,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(52840,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52840,36,'cp 27','f1g2 g8f6 e1g1 c8g4 h2h3 g4f3 e2f3 b8d7 f3f4 e7e6 d2d3 a7a5 b1d2 f8d6 d2f3 e8g8 c1e3 e6e5 f4e5 d7e5 f1e1 f8e8 a2a4 e5f3 d1f3 f6d7 a1c1 d7c5 f3g4 d8f6 e3d4 f6g6 e1e8 a8e8 g4g6 f7g6');
INSERT INTO evaluations VALUES(39542,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(39542,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(39542,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(39542,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(39542,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(39542,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(39542,36,'cp -14','d2d4 d7d6 e5d6 d8d6 g1f3 c7c5 d4c5 d6c5 b1c3 a7a6 h2h3 c8e6 f1d3 g8f6 d1e2 b8c6 e1g1 e8g8 c1e3 c5a5 a2a3 a8d8 a1d1 f6d5 c3d5 e6d5 c2c3 a5c7 f1e1 b7b5 e3c5 f8e8 d3e4 d5b3 d1d8 c6d8 e2e3 b3c4');
INSERT INTO evaluations VALUES(19012,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(19012,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19012,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(19012,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19012,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(19012,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19012,36,'cp 0','g2g3 g8e7 c2c4 g7g6 b1c3 d7d6 f1g2 f8g7 a1b1 e8g8 e1g1 c8e6 c1d2 a7a5 a2a3 a8b8 c3b5 h7h6 b2b4 a5b4 a3b4 b7b6 b5a3 d8d7 b4b5 c6d8 d1e2 f7f5 e4f5 e6f5 a3c2 f5h3 e2e4 h3f5');
INSERT INTO evaluations VALUES(19013,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19013,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19013,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19013,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19013,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19013,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19013,36,'cp 0','g8e7 c2c4 g7g6 f1g2 d7d6 e1g1 f8g7 b1c3 e8g8 a1b1 c8e6 c1d2 a7a5 a2a3 a8b8 c3b5 h7h6 b2b4 a5b4 a3b4 b7b6 b5a3 d8d7 b4b5 c6d8 d1e2 f7f5 e4f5 e6f5 a3c2 f5h3 e2e4 h3f5');
INSERT INTO evaluations VALUES(50350,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(50350,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(50350,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(50350,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(50350,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(50350,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(50350,36,'cp -67','e7e5 d2d4 e5d4 d1d4 b8c6 d4f2 d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 d5e4 f2c5 e4f3 g2f3 d8d5 c5d5 f6d5 c3g7 h8g8 e1c1 c6b4 d1e1 c8e6 g7e5 b4a2 c1b1 a2b4 g1e2 b4c6 e2f4 d5f4 e5f4 e8c8 f1d3 h7h5 h2h4 c6d4 h1f1 d8e8 e1e5 e6h3 f1e1 h3d7 e1e4 g8g1 b1a2 e8e5 e4e5 d4f3 e5h5');
INSERT INTO evaluations VALUES(47976,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(47976,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(47976,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(47976,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(47976,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(47976,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(47976,36,'cp 119','d2d4 e5d4 f3d4 h6g4 b1c3 b8c6 c4b5 c8d7 d4c6 b7c6 b5c6 d7c6 d1g4 d8d7 g4e2 f8e7 e1g1 e8g8 f2f3 f7f5 c1e3 f5e4 f3e4 d7e6 e2d3 e7h4 f1f8 g8f8 a1f1 f8g8 e3f2 h4f2 f1f2 a7a5 h2h3 a8e8 d3a6 c6e4 a6a5 e4c6 a5a6 d6d5 a6d3 e6e1 f2f1 e1e5 b2b3 c6b7');
INSERT INTO evaluations VALUES(47978,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(47978,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(47978,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(47978,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(47978,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(47978,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(47978,36,'cp 142','e5d4 f3d4 h6g4 b1c3 b8c6 c4b5 c8d7 d4c6 b7c6 b5c6 d7c6 d1g4 d8d7 g4e2 f8e7 c1e3 e7f6 e2d2 f6c3 d2c3 c6e4 f2f3 e4b7 e1c1 e8c8 e3a7 d7a4 a7b6 a4c6 b6a5 c6c3 a5c3 h8g8 h2h4 d8e8 b2b3 g7g5 h4g5 g8g5 d1d2 h7h5 a2a4 e8g8');
INSERT INTO evaluations VALUES(49829,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(49829,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(49829,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(49829,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(49829,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(49829,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(49829,36,'cp 16','e4f6 d3d4 d7d6 e5f3 d6d5 f1d3 c7c5 c2c3 f8d6 d4c5 d6c5 e1g1 e8g8 c1f4 b8c6 b1d2 c8g4 d1c2 a8c8 h2h3 g4h5 f1e1 c5d6 f4d6 d8d6 a1c1 c6b4 c2b1 b4c6 a2a3 a7a6 d3c2 f8e8 e1e8 c8e8 c1e1 h5g6 c2g6 h7g6');
INSERT INTO evaluations VALUES(49492,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(49492,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(49492,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(49492,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(49492,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(49492,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(49492,36,'cp 19','d3d4 d7d6 e5f3 g7g6 f1e2 f8g7 c2c4 e8g8 e1g1 c8f5 b1c3 f8e8 e2d3 d8d7 d3f5 d7f5 d1b3 b7b6 c1e3 f5d7 a1e1 f6e4 c3e4 e8e4 b3c2 e4e8 d4d5 a7a5 b2b3 b8a6 h2h3 a6c5 f3d4 c5e4');
INSERT INTO evaluations VALUES(53018,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(53018,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(53018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53018,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(53018,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53018,36,'cp -16','g7g6 b1c3 b8c6 g2g3 f8g7 f1g2 a7a6 a2a4 g8f6 g1e2 c8g4 f2f3 g4d7 c1e3 e8g8 e1g1 a8b8 a1b1 f6e8 f3f4 e8c7 d3d4 c6d4 e2d4 c5d4 e3d4 g7d4 d1d4 b7b5 a4b5 a6b5 b2b3 c7e6 d4e3 b5c4 b3c4 b8b1 f1b1 d7c6 f4f5 e6c5');
INSERT INTO evaluations VALUES(19526,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(19526,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(19526,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(19526,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(19526,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(19526,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(19526,36,'cp -87','g8f6 d1e2 e7e6 c2c3 f6e4 e2e4 d7d5 c4d5 d8d5 e4d5 e6d5 f3d4 c6d4 c3d4 c8e6 c1e3 f8e7 e1d2 e8d7 b1c3 a8c8 a2a3 c8c4 c3e2 h8c8 h1c1 b7b6 e2f4 e6f5 c1c3 d7d6 f4e2 h7h5 a1c1 h5h4 b2b3 c4c6 c3c6 c8c6 f2f3 c6c1 d2c1 d6d7 c1b2 h4h3 g2h3 f5h3');
INSERT INTO evaluations VALUES(12867,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(12867,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(12867,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(12867,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(12867,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(12867,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(12867,36,'cp -89','d1e2 e7e6 c2c3 f6e4 e2e4 d7d5 c4d5 d8d5 e4d5 e6d5 f3d4 c6d4 c3d4 c8e6 c1e3 e8d7 b1c3 h7h5 h2h4 f8e7 a2a3 a8c8 e1d2 c8c4 g2g3 f7f6 h1c1 h8c8 c3e2 e6g4 e2f4 g4f3 c1c3 e7d6 a1c1 a7a5 f4d3 b7b6 e3f4 d6e7 f4e3 a5a4 d3e1 f3e4 f2f3 e4h7 e1d3 g7g5 h4g5 c4c3 c1c3');
INSERT INTO evaluations VALUES(54848,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(54848,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(54848,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(54848,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(54848,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(54848,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(54848,36,'cp 77','c4d5 g8f6 d1a4 c7c6 d5c6 b8c6 g1f3 f8g7 b1c3 c8f5 g2g3 e8g8 f1g2 f6d5 e1g1 d5b6 a4a3 a8c8 c1f4 b6c4 a3b3 c4a5 b3a4 c6d4 f3d4 c8c4 a4a3 c4d4 a1d1 e7e5 f4g5 d8g5 a3a5 f8d8 h2h4 g5e7 c3d5 e7d7 d1d4 e5d4 a5a7 d4d3 e2d3 f5e6 d5f4');
INSERT INTO evaluations VALUES(54841,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(54841,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(54841,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(54841,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(54841,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(54841,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(54841,36,'cp 72','g8f6 g1f3 f8g7 d1a4 c7c6 d5c6 b8c6 b1c3 c8f5 e2e3 e8g8 f1b5 a8c8 e1g1 a7a6 b5e2 f6e4 a4d1 e4d6 c1d2 b7b5 d2e1 e7e5 d4e5 c6e5 f3d4 f5d7 c3d5 f8e8 a1c1 e5c6 d4c6 d7c6 d5b4 c6b7 c1c8 d6c8 b2b3 a6a5 d1d8 e8d8 b4d3 b5b4 d3c5 b7d5 f2f3 c8b6 e3e4 d5c6');
INSERT INTO evaluations VALUES(46742,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(46742,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(46742,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(46742,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(46742,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(46742,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(46742,36,'cp 78','d1a4 c7c6 d5c6 b8c6 g1f3 f8g7 e2e3 e8g8 b1c3 c8f5 f1e2 a7a6 a4d1 b7b5 e1g1 a8c8 a2a3 f8e8 f3h4 f5e6 c1d2 c6a5 h4f3 e6b3 d1b1 b3e6 e2d1 e6c4 c3e2 a5b3 d1b3 c4b3 f1c1 c8c1 b1c1 d8a8 d2c3 b3d5 f3e1 d5c4');
INSERT INTO evaluations VALUES(56865,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(56865,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(56865,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(56865,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(56865,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(56865,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(56865,36,'cp 97','b1c3 f8g7 g1f3 d7d5 c1f4 g8f6 e2e3 e8g8 a1c1 b8d7 f1e2 d5c4 e2c4 a7a6 e1g1 b7b5 c4e2 c8b7 b2b4 d8e7 a2a3 f8c8 h2h3 d7b6 f4g5 a6a5 b4a5 a8a5 f3e5 c7c6 c3e4 a5a3 e4c5 b6d5 e2f3 h7h6 g5h4 e7e8 h4f6 d5f6');
INSERT INTO evaluations VALUES(56863,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(56863,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(56863,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(56863,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(56863,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(56863,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(56863,36,'cp 96','f8g7 g1f3 d7d5 c1f4 g8f6 e2e3 e8g8 a1c1 a7a6 f1e2 d5c4 e2c4 b7b6 e1g1 c8b7 c4b3 d8e7 f1e1 b8c6 a2a3 f8d8 b3a2 h7h6 d1c2 a8c8 h2h3 c6a5 e3e4 c7c5 d4d5 c5c4 d5d6 e7d7 e4e5 f6d5 f4g3 d5c3 b2c3 b7f3 g2f3 d7c6');
INSERT INTO evaluations VALUES(10668,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(10668,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(10668,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(10668,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(10668,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(10668,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(10668,36,'cp 75','b4c3 b2c3 d8e7 d1d5 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 h7h6 e1g1 c6e5 d3e3 e5f3 g2f3 e7f7 e3d4 c8g4 f4d6 f8d8 f3g4 d8d6 d4f4 f6g4 f4g4 d6c6 c4c5 c6c5 g4e4 c7c6 e4c2 f7c4 a1d1 c4c3 c2g6 c3f6 g6f6 g7f6 d1d7 a7a5 f1d1 b7b5 d1d6 c5d5 d6d5 c6d5 d7d5');
INSERT INTO evaluations VALUES(10674,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(10674,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(10674,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(10674,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(10674,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(10674,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(10674,36,'cp 67','b2c3 d8e7 d1d5 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 b7b6 f1g2 c8b7 e1g1 c6a5 c4c5 d6c5 e2e4 a8d8 d3b5 e7d7 b5d7 d8d7 f3h4 e8g8 f1e1 f6h5 f4g5 h7h6 g5e3 f8e8 f2f3 a5c4 g2f1 c4e5 f1e2 h5f6 e1d1 b7c6 e3f4 g7g5 f4e5 e8e5 h4f5 c6a4 d1d7 a4d7 f5e3 g8g7 a1d1 d7a4 d1d2');
INSERT INTO evaluations VALUES(10673,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(10673,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(10673,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(10673,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(10673,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(10673,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(10673,36,'cp 80','d8e7 d1d5 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 c8g4 e1g1 a8e8 a1e1 b7b6 h2h3 g4d7 c4c5 b6c5 e2e4 f6h5 f4g5 e7f7 e4e5 c5c4 d3d2 c6e5 f3e5 d6e5 g5e3 e8d8 d2d5 f7d5 g2d5 g8h8 d5g2 f8e8 e3a7 h5f6 a7c5 d7a4 g3g4');
INSERT INTO evaluations VALUES(599,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(599,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(599,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(599,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(599,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(599,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(599,36,'cp 49','d1d5 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 c8e6 g2b7 a8b8 b7f3 e2c2 f3d1 c2c3 d4d5 e6h3 f1e1 c3c4 d1b3 c4c3 f4e3 a7a5 e3d4 c3d3 d4f6 g7f6 a1c1 b8c8 b3c4 d3d2 c4a6 c8a8 c1c7 d2a2 a6f1 h3f5 e1e7 a5a4 e7g7 g8h8');
INSERT INTO evaluations VALUES(589,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(589,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(589,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(589,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(589,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(589,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(589,36,'cp 68','f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 f4e3 c8e6 e1g1 h7h6 a1b1 a8b8 e3d4 f6d7 f3d2 c6e5 d4e5 d7e5 d3e4 e5d7 e4d4 d7c5 b1e1 b8d8 f2f4 e6f7 e2e4 b7b6 g2f3 g8h8 f3d1 f7e6 d4e3 e7f7 d1c2 e6c4 d2c4');
INSERT INTO evaluations VALUES(580,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(580,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(580,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(580,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(580,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(580,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(580,36,'cp 66','e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 h7h6 f1e1 e2c2 c4c5 d6c5 d4c5 c2c5 e1e7 c8f5 g2b7 a8b8 f4e3 c5c4 b7f3 b8b1 a1b1 f5b1 a2a3 a7a5 e7e8 f6e8 f3d5 g8f8 d5c4 f8e7 c4a6 e7e6 g1f1 e8d6 f1e2 e6d5');
INSERT INTO evaluations VALUES(576,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(576,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(576,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(576,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(576,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(576,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(576,36,'cp 87','g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 c8e6 g2b7 a8b8 b7f3 e2c2 f3d1 c2c3 d4d5 e6h3 f1e1 c3c4 d1b3 c4e4 f4d2 h3g4 d2a5 b8b7 e1c1 e4e7 c1c3 g4h5 a1c1 h5f7 c1d1 h7h6 f2f3 b7b5 c3c7 f6d5 c7c8 g8h7 a5d8 e7e3 g1f2 e3c3 b3c2 f7g6 c8c3 d5c3 c2g6 h7g6');
INSERT INTO evaluations VALUES(577,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(577,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(577,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(577,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(577,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(577,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(577,36,'cp 68','d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 c8e6 g2b7 a8b8 b7f3 e2c2 f3d1 c2c3 d4d5 e6h3 f1e1 c3c4 d1b3 c4e4 f4d2 h3g4 d2a5 b8b7 g1g2 g4h5 e1e4 f6e4 a1e1 h5g6 f2f3 b7b5 a5c7 e4d2 b3d1 b5d5 g3g4');
INSERT INTO evaluations VALUES(578,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(578,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(578,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(578,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(578,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(578,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(578,36,'cp 60','d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 c8e6 g2b7 a8b8 b7f3 e2c2 f3d1 c2c3 d4d5 e6h3 f1e1 c3c4 d1b3 c4e4 f4d2 h3g4 d2a5 b8b7 e1c1 e4e2 b3c4 e2e7 c1c3 g4h5 c4b3 h7h6 a1c1 h5f7 c1d1 f7h5 d1d4 h5f7 f2f3 b7b5 c3c7 f6d5 c7c8 g8h7');
INSERT INTO evaluations VALUES(490,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(490,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(490,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(490,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(490,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(490,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(490,36,'cp 65','g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 c8e6 g2b7 a8b8 b7f3 e2c2 f3d1 c2c3 d4d5 e6h3 f1e1 c3c4 d1b3 c4e4 f4e3 a7a5 f2f3 e4b4 e3d2 b4b5 a1d1 h3d7 g3g4 b5c5 d1c1 a5a4 c1c5 d6c5 b3c4 b8b2 d5d6 g8f8 d2f4 b2c2 c4d3 c2c3 d3e2');
INSERT INTO evaluations VALUES(16216,30,'cp 310',NULL);
INSERT INTO evaluations VALUES(16216,31,'cp 319',NULL);
INSERT INTO evaluations VALUES(16216,32,'cp 328',NULL);
INSERT INTO evaluations VALUES(16216,33,'cp 306',NULL);
INSERT INTO evaluations VALUES(16216,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(16216,35,'cp 343',NULL);
INSERT INTO evaluations VALUES(16216,36,'cp 337','e4e5 f6h7 b1c3 f8e7 d4g4 h7g5 g4g3 d7d6 h2h4 g5e6 c1d2 c7c6 c4e6 c8e6 g3g7 e8d7 e5d6 e7f6 g7g3 h8g8 g3f3 e6g4 f3d3 g8e8 e1f1 d8b6 f2f3 g4e6 c3e4 b6d4 d2c3 d4d3 c2d3 f6c3 b2c3 e8b8 g2g4 e6d5 f1e2 b7b5 h1f1 a7a5 e2e3 b5b4 c3c4 d5e4 f3e4 d7d6 f1f6 d6c5 f6f7');
INSERT INTO evaluations VALUES(16208,30,'cp 339',NULL);
INSERT INTO evaluations VALUES(16208,31,'cp 322',NULL);
INSERT INTO evaluations VALUES(16208,32,'cp 322',NULL);
INSERT INTO evaluations VALUES(16208,33,'cp 324',NULL);
INSERT INTO evaluations VALUES(16208,34,'cp 315',NULL);
INSERT INTO evaluations VALUES(16208,35,'cp 317',NULL);
INSERT INTO evaluations VALUES(16208,36,'cp 331','f6h7 b1c3 f8e7 d4g4 h7g5 g4g3 d7d6 h2h4 g5e6 c1d2 c7c6 c4e6 c8e6 g3g7 e8d7 e5d6 e7f6 g7g3 h8g8 g3f3 e6g4 f3d3 g8e8 e1f1 d8b6 f2f3 g4e6 c3e4 b6d4 e4f6 d4f6 f1f2 b7b5 d2c3 f6f4 a1e1 f4d6 d3d6 d7d6 c3e5 d6d7 e5f4 e6d5 e1e8 a8e8');
INSERT INTO evaluations VALUES(29590,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29590,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29590,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(29590,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(29590,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(29590,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(29590,36,'cp 8','h7h6 c2c4 c7c6 b1c3 e7e6 d1b3 d8b6 c4c5 b6b3 a2b3 f8e7 b3b4 e8g8 e1g1 a7a6 h2h3 b8d7 g3g4 f5c2 f3e1 c2h7 c1f4 g7g5 f4h2 f6e4 c3e4 h7e4 e2e3 e4g6 a1a3 h6h5 f2f3 h5g4 f3g4 f8e8 e1f3 f7f6 f1a1');
INSERT INTO evaluations VALUES(29420,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29420,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(29420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29420,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(29420,36,'cp 3','c2c4 c7c6 b1c3 e7e6 d1b3 d8b6 c4c5 b6b3 a2b3 f8e7 e1g1 a7a6 b3b4 e8g8 c1d2 b8d7 a1a3 f6e4 f1a1 e7f6 e2e3 e4d2 f3d2 a8d8 f2f4 g7g5 c3e2 f6e7 g1f2 g8h8 h2h3 f5c2 d2f3 c2e4 g3g4 g5f4 e2f4');
INSERT INTO evaluations VALUES(44456,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(44456,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(44456,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(44456,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(44456,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(44456,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(44456,36,'cp -31','c2c4 e7e6 d5e6 d7e6 d1d8 e8d8 b1c3 b8d7 g1f3 e6e5 c1d2 f8d6 f3g5 h8f8 g5e4 f6e4 c3e4 d6e7 h2h3 a7a5 e1c1 d8c7 g2g4 d7c5 e4c5 e7c5 d2c3 f7f6 e2e3 c8e6 f1e2 b7b6 g4g5 c5b4 e2g4 e6g4 h3g4 b4c3 b2c3');
INSERT INTO evaluations VALUES(44454,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(44454,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(44454,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(44454,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(44454,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(44454,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(44454,36,'cp -45','c6d5 c4d5 d8a5 c1d2 a5d5 b1c3 d5a5 g1f3 g7g6 c3e4 a5b6 e4f6 e7f6 e2e3 d7d5 f1e2 f8b4 e1g1 c8e6 d2b4 b6b4 a2a3 b4b6 d1d4 b6d4 f3d4 b8c6 e2b5 e8d7 a1c1 h8c8 b2b4 d7d6 b5c6 b7c6 c1c5 a8b8 c5a5 b8b7');
INSERT INTO evaluations VALUES(44601,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(44601,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(44601,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(44601,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(44601,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(44601,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(44601,36,'cp -29','c4d5 d8a5 c1d2 a5d5 b1c3 d5a5 e2e4 e7e6 g1f3 b8c6 a1c1 a5c7 e4e5 c6e5 c3b5 e5f3 g2f3 c7e5 f1e2 f6d5 b5a7 e5b2 a7c8 b2a3 e1g1 f8c5 e2c4 d5f6 c1c3 a3b2 c4e6 a8c8 e6b3 e8g8 d1c1 b2c1 f1c1 b7b6 c3d3 d7d5 d2g5 d5d4 g5f6 g7f6 g1g2 c8e8 f3f4 e8e7 c1g1 f8e8 d3g3 g8f8 b3c4 f6f5 c4d3 c5d6');
INSERT INTO evaluations VALUES(44599,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(44599,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(44599,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(44599,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(44599,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(44599,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(44599,36,'cp -22','d8a5 c1d2 a5d5 b1c3 d5a5 e2e4 e7e6 g1f3 b8c6 a1c1 a5b6 f1c4 f8e7 e4e5 f6g4 d1e2 f7f6 e5f6 g4f6 c4d3 e8g8 e1g1 d7d5 c3b5 c8d7 d2e3 b6d8 b5d4 e7d6 d4c6 d7c6 e3d4 f6e4 f3e5 d6e5 d4e5 f8f5 e5d4 e6e5 d4e3 f5f7 d3e4 d5e4 f1d1 d8f6 e2g4');
INSERT INTO evaluations VALUES(33566,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(33566,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(33566,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(33566,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(33566,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(33566,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(33566,36,'cp -24','c1d2 a5d5 b1c3 d5a5 e2e4 b8c6 g1f3 e7e5 f1c4 f8b4 e1g1 d7d6 a2a3 b4c3 d2c3 a5c7 d1d3 e8g8 f1d1 h7h6 a1c1 a7a6 b2b4 c7e7 a3a4 f8e8 h2h3 c8e6 b4b5 a6b5 a4b5 c6d4 c3d4 e6c4 c1c4 e5d4 c4d4 f6e4');
INSERT INTO evaluations VALUES(46200,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(46200,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46200,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(46200,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46200,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46200,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46200,36,'cp 0','d8b6 d1c2 f6h5 f4e5 b8c6 d3e2 c6e5 d4e5 g7g6 e2h5 g6h5 g1f3 b6a6 b1d2 c8d7 c3c4 b7b5 b2b3 b5c4 b3c4 f8e7 e1g1 e8g8 h2h3 a8c8 f1b1 d7a4 c2c3 f8d8 g1h2 a4c6 c3c2 c6a4');
INSERT INTO evaluations VALUES(36327,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(36327,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(36327,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(36327,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(36327,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(36327,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(36327,36,'cp -9','d5c4 d3c4 a7a6 e1g1 b7b5 c4d3 c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 d7c5 d3b5 d8d1 f1d1 c5e4 b4e7 e4c3 e7f8 g8f8 a2a4 b7f3 g2f3 c3d1 a1d1 f6d5 d1d4 f8e7 d4c4 h7h6 b5c6 a8a5 f3f4 e7d6 c6b5 a5a7 c4d4 d6e7 d4c4 g7g5 f4g5');
INSERT INTO evaluations VALUES(36377,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(36377,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(36377,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(36377,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(36377,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(36377,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(36377,36,'cp -17','d3c4 a7a6 e1g1 b7b5 c4d3 c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 e7c5 a2a3 d8b8 a1c1 c5b4 a3b4 d7e5 f3e5 b8e5 c3b5 f8d8 d1e2 b7a6 d3c4 d8c8 b5a3 a6c4 c1c4 c8c4 a3c4 e5b5 f1c1 a8c8 e4e5 f6d5 g1h1');
INSERT INTO evaluations VALUES(36370,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(36370,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(36370,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(36370,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(36370,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(36370,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(36370,36,'cp -4','a7a6 e1g1 b7b5 c4d3 c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 d7c5 d3b5 d8d1 f1d1 c5e4 b4e7 e4c3 e7f8 g8f8 a2a4 b7f3 g2f3 c3d1 a1d1 f6d5 d1d4 f8e7 g1g2 a8c8 d4c4 c8c4 b5c4 e7d6 c4b5 d6c5 b5e8 f7f6 a4a5 d5c7 e8f7 c5b4 a5a6 c7a6 f7e6 b4b5 e6b3 b5c5');
INSERT INTO evaluations VALUES(35605,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(35605,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(35605,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(35605,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(35605,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(35605,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(35605,36,'cp -19','e1g1 b7b5 c4d3 c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 e7c5 d1b1 d8b8 d3b5 f8c8 b5d7 f6d7 f1e1 b7a6 b4c5 d7c5 b1b8 a8b8 e1b1 b8b1 a1b1 a6d3 b1d1 d3e4 c3e4 c5e4 h2h4 f7f6 g2g3 c8a8 d1d4 e4c5 d4d6 a8a2');
INSERT INTO evaluations VALUES(35603,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(35603,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(35603,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(35603,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(35603,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(35603,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(35603,36,'cp -12','b7b5 c4d3 c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 e7c5 d1b1 d8b8 d3b5 f8c8 b5d7 f6d7 f1d1 d7f6 b4c5 c8c5 b1b3 b7c6 b3b8 a8b8 d1c1 c6e4 c3e4 f6e4 c1c5 e4c5 h2h3 c5a4 a1c1 a4b6 f3d2 g7g6 c1c6 h7h5 d2c4 b6c4 c6c4 b8b1 g1h2');
INSERT INTO evaluations VALUES(35619,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(35619,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(35619,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(35619,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(35619,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(35619,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(35619,36,'cp -16','c4d3 c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 e7c5 d1b1 d8b6 a2a3 f8c8 b1b2 h7h6 h2h3 c5d6 b4d6 b6d6 d3b5 d7e5 f3d2 b7a6 f1d1 d6d4 d2b1 d4c5 a3a4 a6b5 b2b5 f6e4 b5c5 e4c5 b1d2 e5d3 d2c4 c5b3 c4b6 c8c3 b6a8 b3a1');
INSERT INTO evaluations VALUES(35620,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(35620,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(35620,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(35620,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(35620,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(35620,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(35620,36,'cp -11','c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 d7c5 d3b5 d8d1 f1d1 c5e4 b4e7 e4c3 e7f8 g8f8 a2a4 b7f3 g2f3 c3d1 a1d1 f6d5 d1d4 f8e7 g1g2 d5c3 d4d7 e7f8 d7d4 c3b5 a4b5 a8b8 d4h4 h7h6 h4b4 f8e7 g2g3 b8b7 h2h4 h6h5 g3f4 f7f6 f4e4 e7d6 e4d4 g7g5 h4g5');
INSERT INTO evaluations VALUES(27742,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(27742,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(27742,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(27742,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(27742,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(27742,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(27742,36,'cp -11','e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 d7c5 d3b5 d8d1 a1d1 c5e4 b4e7 e4c3 e7f8 g8f8 a2a4 b7f3 g2f3 c3d1 f1d1 f6d5 d1b1 f8e7 b1c1 h7h6 c1c4 f7f5 f3f4 g7g5 f4g5 h6g5 h2h4 g5h4 c4h4 a8a5 g1f1 d5c3 h4b4 e7d6 b4d4 d6e5 d4b4 e5d5 f1e1 c3b5 b4b5 a5b5 a4b5');
INSERT INTO evaluations VALUES(57159,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(57159,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(57159,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(57159,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(57159,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(57159,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(57159,36,'cp -133','d7d5 e4d5 c8h3 g2h3 d8d5 h1g1 b8c6 f1g2 d5d6 d2d3 e8c8 d1g4 c8b8 c1e3 c6d4 e3d4 e5d4 b1d2 d6h2 e1c1 c7c6 g4f3 g8f6 a2a3 f8d6 c1b1 h2f4 d2b3 h8e8 d1e1 g7g6 f3f4 d6f4 e1e8 d8e8 b3d4 f4c7 c2c3 c7b6');
INSERT INTO evaluations VALUES(32224,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(32224,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(32224,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(32224,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(32224,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(32224,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(32224,36,'cp 121','d2d4 b8d7 f3g5 g8f6 g5e6 f7e6 d4e5 d6e5 f1c4 f8b4 d1e2 d8e7 e1g1 a7a5 a2a4 b4c5 c1d2 d7b6 c4d3 c5b4 d2c1 b6d7 c3b5 d7c5 d3c4 c5e4 c4e6 c7c6 e6f5 e4c5 c1g5 c6b5 c2c3 e8g8 c3b4 c5b3 f5h7 g8h7 e2d3 e5e4 d3b3');
INSERT INTO evaluations VALUES(32217,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(32217,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(32217,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(32217,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(32217,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(32217,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(32217,36,'cp 117','b8d7 f3g5 g8f6 d4e5 d6e5 g5e6 f7e6 f1c4 f8b4 d1e2 d8e7 a2a4 d7b6 c4d3 b6d7 e1g1 a7a6 c3a2 b4d6 c1d2 a6a5 a1d1 d7c5 a2c1 e8g8 b2b3 g8h8 f2f3 b7b6 g1h1 a8d8 d3c4 f6d7 d2e3 d7f6 g2g3 h7h6 e3d2 f6d7');
INSERT INTO evaluations VALUES(51506,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(51506,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(51506,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(51506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51506,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51506,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(51506,36,'cp 0','c8b7 b1c3 e7e6 g1f3 d7d5 e4d5 e6d5 d2d4 g8f6 f1d3 c7c5 e1g1 b8c6 a2a3 c5d4 f1e1 f8e7 c3b5 e8g8 b5d4 f8e8 c2c3 c6d4 c3d4 f6e4 c1e3 e7d6 f3e5 d8e7 a1c1 a8c8 d1g4 f7f6 e5f3 a7a6 f3d2 b6b5 d2f3 g8h8 f4f5 c8c1 e1c1');
INSERT INTO evaluations VALUES(30832,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30832,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(30832,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(30832,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(30832,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30832,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30832,36,'cp 0','d1e2 c7c5 d2d3 b8c6 g2g3 d7d5 f1g2 d5e4 d3e4 c6d4 e2d1 d4f5 d1e2');
INSERT INTO evaluations VALUES(1440,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(1440,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(1440,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(1440,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(1440,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(1440,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(1440,36,'cp -14','c2c4 d5c4 a2a4 a7a6 b5c4 c6d4 f3d4 c5d4 b1d2 b6c7 e1g1 c8d7 d1e2 f8b4 d2f3 g8e7 f1d1 d7c6 f3d4 b4c5 d4c6 e7c6 c4d3 c5d4 c1e3 d4e3 e2e3 e8g8 f2f4 f8d8 a1c1 g7g6 b2b4 c7e7 b4b5 a6b5 a4b5 c6b4 d3e4 d8d1 c1d1 a8a2 h2h3 b4c2 e4c2 a2c2');
INSERT INTO evaluations VALUES(16845,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(16845,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(16845,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(16845,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(16845,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(16845,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(16845,36,'cp 99','d4e5 d7d6 c4f7 e8f7 d1d5 c8e6 d5e4 h7h6 b1c3 d6e5 f3e5 c6e5 e4e5 f8d6 e5h5 f7g8 c3e4 d6e7 e1g1 d8d5 h5d5 e6d5 f1e1 g8f7 c1f4 h8e8 a1d1 e7b4 c2c3 d5e4 c3b4 e4d5 f4e5 c7c6 e1e3 a8d8 a2a3 d5c4 d1c1 a7a6 f2f3 d8d3 e3d3 c4d3 c1d1 e8d8 g1f2 d8d5 e5c3 d3c4');
INSERT INTO evaluations VALUES(16789,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(16789,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(16789,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(16789,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(16789,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(16789,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(16789,36,'cp 92','d7d6 c4f7 e8f7 d1d5 c8e6 d5e4 h7h6 b1c3 d6e5 f3e5 c6e5 e4e5 f8d6 e5h5 f7g8 c1e3 g8h7 e1c1 h8f8 c3e4 e6f5 e4d6 c7d6 c1b1 a8c8 d1d2 d8c7 h1c1 a7a5 g2g4 f5g6 h5d5 c7c4 a2a3 a5a4 d5c4 c8c4 f2f4 c4e4');
INSERT INTO evaluations VALUES(15679,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(15679,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(15679,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(15679,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(15679,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(15679,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(15679,36,'cp 87','c4f7 e8f7 d1d5 c8e6 d5e4 h7h6 b1c3 d6e5 f3e5 c6e5 e4e5 f8d6 e5h5 f7g8 c1e3 g8h7 c3b5 d8f6 e1c1 d6e5 e3d4 e5d4 b5d4 h8d8 d4e6 f6e6 c1b1 d8d1 h5d1 a8e8 h1f1 e6e4 g2g3 a7a5 a2a3 a5a4 d1d2 e8e7 f1d1 e4e2 h2h4 c7c6 b1c1 e2f3 d1e1 e7e1 d2e1 h7g6 c1b1 g6f5 e1d2 f5g4 d2d7 g4h5 d7g7 f3f2 g7e5 h5g4');
INSERT INTO evaluations VALUES(41319,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41319,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(41319,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(41319,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41319,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41319,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(41319,36,'cp 0','c7c5 b1c3 c5d4 e3d4 e8g8 h2h3 d8b6 e1g1 b8c6 d4d5 b6b2 f4d2 c6e5 f3e5 d6e5 c3b5 f6d5 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(19723,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(19723,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(19723,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(19723,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19723,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19723,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(19723,36,'cp 26','e1g1 c8g4 c2c4 d5c4 h2h3 g4h5 b5c4 g8f6 d2d4 e7e6 b1c3 f8e7 d4d5 e6d5 c3d5 e8g8 f1e1 f6d5 d1d5 d8d5 c4d5 e7f6 g2g4 h5g6 g4g5 f6d8 a2a3 d8b6 c1e3 f8d8 d5c6 b7c6 f3e5 b6e3 e1e3 c6c5 e5c6 d8d2 b2b4 c5b4 a3b4 h7h6 g5h6 g7h6 a1a7 a8a7');
INSERT INTO evaluations VALUES(19721,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(19721,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(19721,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(19721,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19721,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19721,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(19721,36,'cp 8','c8g4 c2c4 d5c4 h2h3 g4h5 b5c4 e7e6 b1c3 f8d6 d2d4 g8e7 d4d5 e6d5 c3d5 e8g8 f1e1 e7d5 d1d5 h5f3 d5f3 c6e5 f3e2 e5c4 e2c4 d8f6 c4d5 b7b6 c1g5 f6g6 a1d1 d6c5 g5e7 c5e7 e1e7 g6f6 d5e5 f6e5 e7e5 a8d8 d1d8 f8d8 e5e7 h7h6 g2g4 d8d2 e7a7 d2b2 g1g2 b2d2');
INSERT INTO evaluations VALUES(26248,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26248,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(26248,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(26248,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(26248,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(26248,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(26248,36,'cp 4','c2c4 d5c4 h2h3 g4h5 b1c3 e7e6 b5c4 f8d6 d2d4 g8e7 d4d5 e6d5 c3d5 e8g8 f1e1 e7d5 d1d5 h5f3 d5f3 c6e5 f3e4 e5c4 e4c4 d8a5 e1e2 a8d8 c1d2 d6h2 g1h2 d8d2 a1e1 d2e2 e1e2 h7h6 b2b4 a5b6 c4c5 b6c5 b4c5 f8c8 e2e7 c8c5');
INSERT INTO evaluations VALUES(19875,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(19875,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(19875,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(19875,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(19875,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(19875,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(19875,36,'cp -27','d2d4 g8f6 h2h3 c8f5 g1f3 e7e6 c1f4 f8e7 c2c3 e8g8 b1d2 b7b5 a2a3 a7a5 e1g1 d8b6 d1e2 b5b4 b3a4 f8c8 f3h4 f5e4 f1c1 h7h6 h4f3 e4f5 d2b3 b4a3 b2a3 f6e4 c3c4 d5c4 c1c4 e7d6 f4d6 e4d6 c4c6 c8c6 a4c6 b6c6 b3c5 c6b5 e2b5');
INSERT INTO evaluations VALUES(19848,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(19848,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(19848,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(19848,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(19848,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(19848,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(19848,36,'cp -39','g8f6 h2h3 c8f5 c2c3 e7e6 c1f4 a8c8 g1f3 f6h5 f4e5 f8d6 e5d6 d8d6 f3h4 f5g6 e1g1 e8g8 h4g6 h7g6 b1d2 b7b5 a2a3 a7a5 d1e2 b5b4 a3b4 a5b4 f1c1 c8a8 d2f3 c6a5 f3d2 f8c8 e2e3 b4c3 b2c3 d6c7 b3d1 h5f6 d1e2 a5c6 a1a8 c8a8 c3c4');
INSERT INTO evaluations VALUES(13145,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(13145,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(13145,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(13145,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(13145,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(13145,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(13145,36,'cp -25','h2h3 c8f5 c2c3 a8c8 c1f4 e7e6 g1f3 f6h5 f4h2 f8d6 h2d6 d8d6 f3h4 f5g6 e1g1 b7b5 b1d2 e8g8 a2a3 a7a5 d1e2 b5b4 h4g6 h7g6 a3b4 a5b4 d2f3 b4c3 b2c3 c6e7 f1c1 h5f6 c3c4 d5c4 b3c4 e7f5 e2e5 d6c6 h3h4 c8c7 a1a6 c6a6 e5c7 a6a3 c1b1 a3c3 b1b8 f8b8 c7b8 g8h7 b8b3 c3c1 c4f1');
INSERT INTO evaluations VALUES(5899,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(5899,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(5899,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(5899,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(5899,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(5899,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5899,36,'cp 0','f7f6 e1g1 g8h6 c2c4 a7a6 c4d5 a6b5 d5c6 e8g8 d2d3 b7c6 b1d2 c6c5 a2a4 c8b7 d1c2 d8e8 a4b5 e8b5 c2c4 b5c4 b3c4 h6f7 f1b1 a8a1 b1a1 f8b8 b2c3 b7f3 g2f3 f6f5 h2h4 f7d8 h4h5 g8f7 g1g2 d8c6 g2f1 d6e7 f1e2');
INSERT INTO evaluations VALUES(6108,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6108,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6108,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6108,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6108,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6108,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(6108,36,'cp 0','e1g1 g8h6 c2c4 a7a6 c4d5 a6b5 d5c6 e8g8 d2d3 b7c6 b1d2 c6c5 d1c2 d8e8 a2a4 c8b7 a4b5 e8b5 c2c4 b5c4 b3c4 h6f7 f1b1 a8a1 b1a1 b7f3 g2f3 f8b8 b2c3 f6f5 g1f1 f7d8 f1e2 d8c6 a1a6 b8b6 a6a8 b6b8');
INSERT INTO evaluations VALUES(41207,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(41207,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(41207,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(41207,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(41207,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(41207,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41207,36,'cp 35','e2e4 h7h6 g5f4 e8g8 d1d2 e7e5 d4e5 d6e5 f4h6 f6e4 d2d8 f8d8 h6g7 e4c3 g7f6 d8d6 f6e5 d6e6 f2f4 f7f6 b2c3 f6e5 e1c1 b8c6 f4e5 c6e5 g1f3 e5f3 g2f3 g8f7 d1d8 e6d6 d8d6 c7d6 h2h4 c8f5 h4h5 g6h5 h1h5 f7g6 h5h4 g6f6 c1d2 a8g8 h4h6 f6e7 d2e3 e7d7 h6h5 g8f8 f1e2 b7b6 e3d4 f5b1 h5g5 b1a2 g5g7 d7c6 g7a7');
INSERT INTO evaluations VALUES(57404,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(57404,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(57404,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(57404,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(57404,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(57404,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(57404,36,'cp 215','d2d4 b8c6 g1f3 c6d4 f3d4 e5d4 d1d4 g8e7 b1c3 e7c6 d4e3 b7b6 e3g3 c6e5 c4b3 c8b7 c1e3 d8e7 f2f4 e5c6 a2a3 e8c8 e1c1 c8b8 h1e1 e7e8 c1b1 e8g6 g3h3 h7h5 d1d7 d8d7 h3d7 f8d6 g2g3 h8d8 d7h3 g6g4 h3g2 h5h4 g2f2 c6a5 b3a2 h4g3 h2g3 a5c6');
INSERT INTO evaluations VALUES(42051,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(42051,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(42051,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(42051,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(42051,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(42051,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(42051,36,'cp 30','d1a4 b8c6 e2e3 e8g8 a2a3 b4c3 d2c3 c6e7 a4c2 b7b6 g1f3 a7a5 b2b3 c8a6 f1e2 f6e4 e1g1 e4c3 c2c3 c7c5 a1c1 a8c8 d4c5 b6c5 f1d1 d8c7 e2d3 h7h6 f3e5 f8d8 h2h3 d5d4 e3d4 c5d4 c3e1 a6b7 b3b4 a5b4 a3b4 f7f6');
INSERT INTO evaluations VALUES(4983,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4983,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4983,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4983,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4983,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4983,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(4983,36,'cp -7','d8h4 d1e2 g8f6 e4e5 e8g8 e5f6 d7d5 f6g7 f8d8 g2g3 h4h3 b1c3 c8g4 e2f1 d5c4 f1h3 g4h3 c1e3 c5e7 e1e2 d8b8 c3a4 e7f6 a1c1 f6g7 h1e1 g7b2 a4b2 b8b2 e2d2 b2a2 c1a1 a2a1 e1a1 a7a5 a1a4 h3e6 h2h4 f7f6 d2c3 e6d5 h4h5 g8f7 h5h6 f7g6 g3g4');
INSERT INTO evaluations VALUES(105,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(105,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(105,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(105,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(105,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(105,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(105,36,'cp -21','d1e2 g8f6 e4e5 e8g8 e5f6 d7d5 e1g1 c5d6 g2g3 h4c4 e2c4 d5c4 f1e1 g7f6 c1d2 a7a5 d2c3 f6f5 b1d2 c8e6 d2f3 a5a4 f3d4 e6d7 a2a3 c6c5 d4e2 f7f6 f2f3 g8f7 g1f2 d7e6 e2f4 d6f4 g3f4 f8e8 a1d1 e8d8 d1d8 a8d8 e1e2 d8d1');
INSERT INTO evaluations VALUES(54476,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(54476,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(54476,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(54476,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(54476,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(54476,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(54476,36,'cp 127','f8c5 f1b5 b8d7 d1d5 d8e7 d5e4 a7a6 b5c4 g8f6 e4e7 c5e7 c1e3 b7b5 c4e2 c8b7 e2f3 e8c8 f3b7 c8b7 g1f3 e7c5 e1c1 c5e3 f2e3 h8e8 h1e1 d7c5 d1d4 b7c7 d4f4 c5e6 f4b4 f6d5 c3d5 d8d5 b4e4 c7d6 b2b4 e8c8 c1b2');
INSERT INTO evaluations VALUES(38190,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(38190,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(38190,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(38190,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(38190,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(38190,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(38190,36,'cp 149','f1b5 c8d7 d1d5 d8e7 d5e4 e7e4 c3e4 d7b5 e4c5 b7b6 c5b3 b8c6 g1f3 e8c8 c1e3 g8e7 f3d4 c6d4 b3d4 b5d7 b2b3 h8e8 e1c1 e7f5 h1e1 h7h5 a2a4 f5d4 d1d4 d7f5 d4b4 g7g6 c1b2 d8d5 c2c4 d5d3 a4a5 b6a5 b4a4 e8e5 e1a1 c8b7 a4a5 e5a5 a1a5');
INSERT INTO evaluations VALUES(9950,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9950,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9950,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9950,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(9950,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9950,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9950,36,'cp 0','d7c6 e1g1 f6d7 c2c3 c5d6 d3d4 e8g8 d1c2 a7a5 f1e1 a5a4 h2h3 f7f6 c1e3 d8e7 b1d2 f8e8 b2b3 e5d4 c3d4 a4b3 a2b3 a8a1 e1a1 d6b4 a1a8 d7b6 a8a1');
INSERT INTO evaluations VALUES(47934,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47934,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(47934,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47934,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47934,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47934,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47934,36,'cp 0','f6e4 d1e2 d8e7 b1c3 e4f6 c3b5 c8e6 b5d4 b8c6 d4c6 b7c6 c4a5 c6c5 a5c6 e7d7 e2a6 e6d5 f1b5 d7e6 e1f1 d5g2 f1g2 e6g4 g2f1 g4d1 f1g2');
INSERT INTO evaluations VALUES(48010,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48010,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48010,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48010,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48010,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48010,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48010,36,'cp 0','d1e2 d8e7 b1c3 e4f6 c3b5 c8e6 b5d4 b8c6 d4c6 b7c6 c4a5 c6c5 a5c6 e7d7 e2a6 e6d5 f1b5 d7e6 e1f1 d5g2 f1g2 e6g4 g2f1 g4d1 f1g2');
INSERT INTO evaluations VALUES(48011,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48011,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48011,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48011,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48011,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48011,36,'cp 0','d8e7 b1c3 e4f6 c3b5 c8e6 b5d4 b8c6 d4c6 b7c6 c4a5 c6c5 a5c6 e7d7 e2a6 e6d5 f1b5 d7e6 e1f1 d5g2 f1g2 e6g4 g2f1 g4d1 f1g2');
INSERT INTO evaluations VALUES(33810,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33810,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33810,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33810,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33810,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33810,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33810,36,'cp 0','b1c3 e4f6 c3b5 c8e6 b5d4 e6c8');
INSERT INTO evaluations VALUES(54010,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(54010,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(54010,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(54010,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(54010,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(54010,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(54010,36,'cp 82','e2e4 e6d5 e4d5 g7g6 b1c3 f8g7 f1d3 d8e7 d1e2 e7e2 e1e2 b8d7 g1f3 d7e5 f3e5 g7e5 h1e1 g8e7 e2f1 e8g8 c1h6 f8e8 h6g5 c8f5 d3f5 e7f5 c3e4 h7h6 e4f6 e5f6 g5f6 g8f8 a2a4 e8e1 a1e1 h6h5 h2h3 a8e8 e1e8 f8e8 f1e2 f5h6 a4a5 h6g8 f6g5');
INSERT INTO evaluations VALUES(54008,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(54008,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(54008,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(54008,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(54008,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(54008,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(54008,36,'cp 60','g7g6 b1c3 f8g7 f1d3 e6d5 e4d5 b8d7 g1f3 d8e7 d1e2 d7e5 f3e5 g7e5 c1d2 c8g4 f2f3 g4d7 e1c1 e8c8 d1e1 f7f5 f3f4 e5f6 e2d1 e7f7 d1b3 f6d4 h2h3 c8b8 d3c2 d8e8 e1e8 f7e8 g2g4 f5g4 h3g4 h7h5 h1e1 e8d8 c3e2 d4g7 d2c3 g7c3 b3c3');
INSERT INTO evaluations VALUES(54765,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(54765,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(54765,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(54765,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(54765,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(54765,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(54765,36,'cp 104','c1f4 e7e6 e2e3 f8d6 c2c4 g8f6 b1c3 e8g8 f4d6 d8d6 f1d3 f6e4 e1g1 c7c6 c3e2 b8d7 a2a4 a7a5 a1c1 d6e7 d1c2 d7b8 e2f4 g7g5 f4h5 c8d7 c4d5 e6d5 h5g3 b8a6 d3a6 e4g3 h2g3 a8a6');
INSERT INTO evaluations VALUES(54755,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(54755,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(54755,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(54755,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(54755,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(54755,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(54755,36,'cp 98','e7e6 e2e3 f8d6 c2c4 g8f6 b1c3 e8g8 a1c1 b8c6 g2g3 c6e7 c4c5 d6f4 e3f4 c8d7 f1d3 d7e8 d1e2 e7c6 d3b5 e8h5 b5c6 h5f3 e2f3 b7c6 e1g1 f8e8 f1e1 h7h6 f3e3 a7a5 c3a4 d8d7 e1e2 d7f7 f2f3 f6d7');
INSERT INTO evaluations VALUES(56766,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(56766,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(56766,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(56766,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(56766,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(56766,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(56766,36,'cp 118','e2e3 f8d6 c2c4 g8f6 b1c3 e8g8 f1e2 d6f4 e3f4 d8d6 c3b5 d6b4 d1d2 b4d2 e1d2 f6e8 c4c5 a7a6 b5a3 e8f6 a3c2 c8d7 a2a4 a6a5 f3e5 g7g5 g2g3 g5f4 g3f4 b8c6 f2f3 c6e5 f4e5 f6h5 d2e3 f5f4 e3f2 h5g7 h1g1 d7e8 e2d3 g8h8 g1g4 e8h5 g4h4 g7f5 h4f4 h5g6 a1e1');
INSERT INTO evaluations VALUES(56765,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(56765,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(56765,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(56765,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(56765,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(56765,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(56765,36,'cp 119','f8d6 c2c4 g8f6 b1c3 e8g8 f1e2 d6f4 e3f4 d8d6 f3e5 d5c4 e2c4 b8c6 c3b5 d6b4 d1d2 b4d2 e1d2 f6e8 e5c6 b7c6 b5c3 e8d6 c4d3 f8d8 d2e3 a7a5 c3a4 c8a6 d3a6 a8a6 a1c1 d6e4 c1c4 e4f6 b2b3 f6d5 e3f3 d8d6 h1c1 g8f7 g2g3 d5b4 a2a3 b4d5 h2h4 d5f6 f3e3 f6d5 e3e2 f7e7 e2d2 d5f6 d2e3 f6d5 e3e2');
INSERT INTO evaluations VALUES(5967,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5967,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(5967,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5967,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(5967,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5967,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5967,36,'cp 0','g8f6 a2a3 c8e6 d2d3 a7a6 c1e3 h7h6 d1e2 f6d7 c3d5 e8g8 e1g1 c5e3 f2e3 c6e7 d5e7 d8e7 c4e6 f7e6 f3h2 b7b5 f1f8 a8f8 a1f1 c7c5 f1f8 e7f8 b2b3 g8h7 h2g4 f8e7 e2d2 e7d8 d2e2');
INSERT INTO evaluations VALUES(49474,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(49474,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(49474,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(49474,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(49474,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(49474,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(49474,36,'cp -38','b1c3 c7c6 b5a4 f8c5 g1f3 a7a5 e1g1 e8g8 a2a3 d7d6 h2h3 b8d7 d2d3 h7h6 a4b3 b7b5 c1e3 c5e3 f2e3 d7c5 b3a2 c8e6 a2e6 f7e6 c3e2 d8b6 d1d2 c5d7 e2g3 c6c5 d2e2 g8h7 f1f2 c5c4 a1f1 a8e8 g1h2 d7c5 d3c4 b5c4 e2c4 b6b2');
INSERT INTO evaluations VALUES(49471,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(49471,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(49471,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(49471,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(49471,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(49471,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(49471,36,'cp -41','c7c6 b5a4 f8c5 d2d3 e8g8 a4b3 a7a5 g1f3 d7d6 a2a3 b8d7 e1g1 c5b6 b3a2 a5a4 h2h3 f8e8 c1e3 h7h6 e3b6 d8b6 a1b1 d7c5 f1e1 c8e6 a2e6 c5e6 c3e2 f6h5 d1d2 b6a5 d2c3 a5c3 b2c3 b7b5 b1b4 h5f6 e2g3');
INSERT INTO evaluations VALUES(44059,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(44059,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(44059,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(44059,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(44059,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(44059,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(44059,36,'cp -25','b5a4 f8c5 g1f3 a7a5 a2a3 d7d6 h2h3 e8g8 e1g1 c5b6 d2d3 b8a6 a4b3 a6c5 b3a2 a5a4 a1b1 f8e8 c1e3 c8e6 a2e6 c5e6 e3b6 d8b6 d1d2 h7h6 f1e1 e8d8 f3h4 b6a7 h4f5 b7b5 b1d1 e6d4 f5d4 a7d4 d2e3 d4e3 e1e3 g8f8 d3d4 e5d4');
INSERT INTO evaluations VALUES(44115,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(44115,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(44115,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(44115,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(44115,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(44115,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(44115,36,'cp -33','f8c5 g1f3 a7a5 a2a3 d7d6 a4b3 e8g8 d2d3 b7b5 e1g1 h7h6 h2h3 b8d7 b3a2 f8e8 c1e3 c5a7 e3a7 a8a7 f1e1 d8c7 b2b4 d7f8 d1d2 c8e6 a2e6 f8e6 a1d1 a5b4 a3b4 c7b6 d3d4 e6d4 f3d4 e5d4 d2d4 b6d4 d1d4');
INSERT INTO evaluations VALUES(15936,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15936,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(15936,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(15936,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(15936,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(15936,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(15936,36,'cp 25','h7h6 g5h4 f8e7 c4d5 f6d5 h4e7 d8e7 e2e4 d5c3 b2c3 c7c5 f1d3 c5d4 c3d4 e8g8 e1g1 f8d8 a2a4 d7f6 d1e2 c8d7 f1b1 d7c6 a4a5 b7b5 a5b6 a7b6 a1a8 d8a8 f3e5 e7b7 e5c6 b7c6 h2h3 c6c3 e2e3 c3a3 b1b6');
INSERT INTO evaluations VALUES(15886,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(15886,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(15886,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(15886,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(15886,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(15886,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(15886,36,'cp 19','g5h4 f8e7 e2e3 e8g8 d1c2 c7c6 a2a3 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3d2 d7f6 f3e5 e7d8 f2f3 f6d7 e5g4 a7a5 a1e1 b7c6 d3a6 c8c7 e1c1 c6b7 a6b7');
INSERT INTO evaluations VALUES(15887,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(15887,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(15887,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(15887,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(15887,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(15887,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(15887,36,'cp 22','f8e7 e2e3 e8g8 d1c2 c7c6 a2a3 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3e1 d7f6 f3e5 b7c6 f2f3 e7d8 a1c1 c6e8 d3a6 c8c7 h2h3 f6h5 a6d3 c7c1 e1c1 h5f6 b2b4');
INSERT INTO evaluations VALUES(9566,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9566,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(9566,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(9566,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(9566,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(9566,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9566,36,'cp 13','e2e3 e8g8 d1c2 c7c6 a2a3 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3e1 d7f6 h2h3 f6e4 a1c1 e7d8 f3e5 e4d6 c1c3 c8c7 e1c1 c7c8 c1d2 a7a5 d2c2 c8c3 c2c3 d8c8 c3b3 c8c7');
INSERT INTO evaluations VALUES(9565,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(9565,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9565,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(9565,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(9565,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(9565,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(9565,36,'cp 25','e8g8 d1c2 c7c5 d4c5 d8a5 c4d5 f6d5 h4e7 d5e7 f1e2 a5c5 e1g1 d7f6 a1c1 c8d7 a2a3 c5a5 c3e4 f6e4 c2e4 d7c6 e4e5 a5e5 f3e5 a8d8 e5c6 e7c6 e2f3 d8d2 b2b4 d2a2 g2g3 f8d8 b4b5 c6a5 c1c7 a2a3 f1c1');
INSERT INTO evaluations VALUES(3975,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(3975,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(3975,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(3975,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(3975,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(3975,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3975,36,'cp 31','d1c2 c7c6 a2a3 a7a6 c4d5 c6d5 f1d3 b7b5 e1g1 c8b7 f1c1 a8c8 c2d1 d7b6 b2b3 b6d7 c3e2 d8b6 b3b4 e7d6 h4g3 f6e4 f3e5 d6e5 g3e5 c8c1 a1c1 e4d6 e5g3 f8c8 c1c8 b7c8 g3d6 b6d6 e2c1 d7b6 c1b3 b6c4');
INSERT INTO evaluations VALUES(3976,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(3976,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(3976,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(3976,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(3976,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3976,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(3976,36,'cp 17','c7c5 d4c5 d7c5 c4d5 f6d5 h4e7 d5e7 f1e2 d8b6 e1g1 c8d7 f3e5 b6c7 e5d7 c5d7 a1c1 a8c8 c2b3 e7c6 f1d1 d7f6 h2h3 f8d8 d1d8 c8d8 c1c2 d8d7 b3a4 c7b6 a2a3 b6d8 c3e4 f6e4 a4e4 d8a5 c2c3 d7d2 b2b4');
INSERT INTO evaluations VALUES(47832,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(47832,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(47832,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(47832,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(47832,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(47832,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(47832,36,'cp 70','d2d4 e5d4 f3d4 d6d5 c1g5 d5c4 e2e3 h7h6 g5h4 b8c6 d4c6 d8d1 a1d1 b7c6 f1c4 a8b8 c3a4 f8d6 h4f6 g7f6 f2f4 c8d7 b2b3 e8e7 e1f2 h8d8 h2h3 f6f5 d1d3 h6h5 h1d1 h5h4 a4c3 d7e6 c4e6 f7e6 c3a4 c6c5 a4b2 a7a5 d1d2 d8g8 a2a3');
INSERT INTO evaluations VALUES(47840,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(47840,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(47840,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(47840,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(47840,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(47840,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(47840,36,'cp 61','e5d4 f3d4 d6d5 c1g5 d5c4 e2e3 h7h6 g5h4 b8c6 d4c6 d8d1 a1d1 b7c6 f1c4 a8b8 c3a4 f8d6 h4f6 g7f6 e1g1 e8e7 d1d2 h8d8 f1d1 c8d7 h2h3 d6b4 a4c3 d7e8 d2d8 b8d8 d1d8 e7d8 g1f1 f6f5 f1e2 b4d6 e2d3 a7a5 d3d4 d6b4');
INSERT INTO evaluations VALUES(47938,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(47938,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(47938,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(47938,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(47938,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(47938,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(47938,36,'cp 73','f3d4 d6d5 c1f4 c7c6 e2e3 f8b4 d1b3 d8a5 a2a3 b4c3 b2c3 b8a6 f1e2 e8g8 e1g1 f8e8 f1c1 d5c4 e2c4 f6d5 c4d5 a5d5 b3d5 c6d5 a3a4 a6c5 c1b1 c8d7 a4a5 b7b6 h2h3 c5a4 b1b4 a4c3 a5b6 a7b6 a1a8 e8a8 b4b6 f7f6 b6b7');
INSERT INTO evaluations VALUES(47936,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(47936,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(47936,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(47936,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(47936,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47936,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(47936,36,'cp 62','d6d5 c1g5 d5c4 e2e3 h7h6 g5h4 b8c6 d4c6 d8d1 a1d1 b7c6 f1c4 a8b8 c3a4 f8b4 e1e2 e8e7 h4f6 e7f6 d1d4 b4d6 b2b3 a7a5 h2h3 f6e7 h1c1 f7f5 e2f1 b8b4 d4d1 g7g5 c4d3 h6h5 c1c6 c8d7 c6c2 d7a4 b3a4 f5f4 e3f4 b4f4 d3c4');
INSERT INTO evaluations VALUES(6506,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(6506,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6506,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6506,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(6506,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(6506,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(6506,36,'cp 6','f1g2 d7d6 c2c3 c5b6 e1g1 g8f6 d2d3 a7a5 f1e1 e8g8 c1e3 b6e3 e1e3 f8e8 b1d2 a5a4 b2b4 a4b3 a2b3 c8e6 e3e1 h7h6 h2h3 a8a1 d1a1 d8c8 g1h2 c8a8 b3b4 a8a1 e1a1 e8b8 b4b5 c6e7 d3d4 c7c6 b5c6 e7c6');
INSERT INTO evaluations VALUES(6507,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(6507,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6507,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(6507,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6507,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6507,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6507,36,'cp 1','d7d6 c2c3 c5b6 e1g1 g8f6 d2d3 a7a5 f1e1 e8g8 c1e3 b6e3 e1e3 f8e8 b1d2 a5a4 b2b4 a4b3 a2b3 c8e6 h2h3 e6d7 e3e1 a8a1 d1a1 d8c8 g1h2 b7b5 d3d4 c8b7 a1b2 e8a8 h2g1 b7b6 e1a1 a8a1 b2a1 h7h6 g2f1 e5d4 c3d4 c6b4 a1a8 d7e8 d4d5');
INSERT INTO evaluations VALUES(6025,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(6025,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(6025,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(6025,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6025,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(6025,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(6025,36,'cp 0','e1g1 g8f6 c2c3 c5b6 d2d3 a7a5 f1e1 e8g8 c1e3 b6e3 e1e3 h7h6 a2a4 c8e6 b1d2 d8d7 d3d4 e6h3 g2h3 d7h3 e3e2 e5d4 c3d4 c6b4 a1c1 h3d7 h2h3 f8e8 g1h2 a8d8 f3h4 c7c6 f2f4 b7b5 a4b5');
INSERT INTO evaluations VALUES(6024,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6024,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(6024,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6024,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6024,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(6024,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(6024,36,'cp 5','g8f6 c2c3 c5b6 d2d3 a7a5 b1a3 e8g8 f1e1 h7h6 c1e3 b6e3 e1e3 f8e8 d3d4 a5a4 h2h3 e5d4 f3d4 c6d4 d1d4 c8d7 a3c2 d8c8 g3g4 c8d8 f2f4 d7c6 a1e1 f6d7 g1h2 d8e7 c2b4 e7h4 d4d2 d7c5 b4d5');
INSERT INTO evaluations VALUES(11946,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(11946,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(11946,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11946,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(11946,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(11946,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(11946,36,'cp 48','c2c4 d5b6 c1e3 d7d5 e5d6 c7d6 f1d3 d6d5 e1g1 f8e7 b1d2 e8g8 d1c2 h7h6 f1e1 c8e6 a1d1 c6c5 e3f4 e7d6 f4d6 d8d6 c4d5 b6d5 d2b3 d6b6 c2c5 d5b4 d3e4 a8d8 c5b6 a7b6 e4b1 e6b3 a2b3 b4c6 g2g3 g7g6 b1e4 c6d4 b3b4');
INSERT INTO evaluations VALUES(11945,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(11945,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(11945,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(11945,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(11945,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(11945,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(11945,36,'cp 39','d5b6 c1e3 d7d6 e5d6 c7d6 f1d3 d6d5 e1g1 f8e7 b1d2 e8g8 d1c2 h7h6 f1e1 c8e6 a1d1 c6c5 e3f4 e7d6 f4g3 e6g4 d1c1 d6g3 h2g3 a8b8 c4d5 b6d5 a2a3 d5f6 d3f1 g4e6 d2c4 d8c7 b2b3 f8d8 e1d1 f6g4');
INSERT INTO evaluations VALUES(11912,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(11912,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(11912,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(11912,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11912,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(11912,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(11912,36,'cp 45','c1e3 d7d6 e5d6 c7d6 f1d3 d6d5 e1g1 f8e7 b1d2 e8g8 d1c2 h7h6 f1e1 c8e6 a1d1 d5c4 d2c4 d8c7 g2g3 b6d5 d3h7 g8h8 h7f5 d5e3 c4e3 e7b4 e1e2 e6f5 c2f5 f8d8 d1d8 a8d8 e3c4 b4f8 b2b3 h8g8 g1g2');
INSERT INTO evaluations VALUES(50554,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(50554,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(50554,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(50554,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(50554,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(50554,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50554,36,'cp 26','g8f6 b1c3 f8b4 d1d3 d7d5 e4d5 e6d5 f1e2 e8g8 e1g1 b8c6 c4d5 f6d5 d4c6 b7c6 c3d5 d8d5 d3g3 b4d6 c1f4 d6f4 g3f4 d5f5 f4f5 c8f5 f1d1 f5e6 b2b3 e6d5 d1d4 g7g6 g1f1 f8e8 f2f3 e8e6 a1c1 a6a5 d4a4 e6e3 e2c4 h7h6 c4d5');
INSERT INTO evaluations VALUES(43321,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43321,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(43321,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(43321,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(43321,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(43321,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(43321,36,'cp 32','b1c3 f8b4 d1d3 d7d5 e4d5 e6d5 f1e2 b8c6 d4c6 b7c6 e1g1 e8g8 c4d5 f6d5 c3d5 d8d5 d3g3 b4d6 c1f4 d6f4 g3f4 d5f5 f4f5 c8f5 b2b3 f8e8 f1e1 g7g6 e2c4 a6a5 f2f3 g8f8 g2g4 f5d7 e1e8 a8e8 g1f2 e8b8 a1e1 h7h6 e1d1 f8e7 d1c1 b8d8 f2e2');
INSERT INTO evaluations VALUES(43320,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(43320,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(43320,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43320,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(43320,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43320,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(43320,36,'cp 31','f8b4 d1d3 d7d5 e4d5 e6d5 f1e2 e8g8 e1g1 b8c6 c4d5 f6d5 c3d5 d8d5 d4c6 b7c6 d3g3 b4d6 c1f4 d6f4 g3f4 d5f5 f4f5 c8f5 e2f3 a8b8 b2b3 f8c8 f1d1 c6c5 d1d6 a6a5 d6a6 c5c4 b3c4 c8c4 a6a5 g7g6 a2a4 f5e4 a5b5 b8b5 a4b5 e4f3 g2f3 g8g7 h2h3 c4b4');
INSERT INTO evaluations VALUES(39655,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(39655,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(39655,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(39655,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(39655,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(39655,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(39655,36,'cp 26','d1d3 d7d5 e4d5 e6d5 f1e2 b8c6 d4c6 b7c6 e1g1 e8g8 c4d5 f6d5 c3d5 d8d5 d3g3 b4d6 c1f4 d6f4 g3f4 d5f5 f4f5 c8f5 e2f3 a8b8 b2b3 c6c5 a1c1 f8c8 f3d5 b8b6 f1d1 g8f8 d5c4 f8e7 g1f1 a6a5 c1c3 f5e6 c4e6 e7e6');
INSERT INTO evaluations VALUES(16319,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16319,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(16319,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(16319,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16319,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16319,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16319,36,'cp -13','d8e7 g1e2 a7a6 b5c6 d7c6 d2d3 c8g4 h2h3 g4e6 b1d2 f6d7 e1g1 e8c8 f2f4 e7c5 g1h1 f7f6 f4e5 f6e5 d3d4 e5d4 e2d4 c5e7 d1f3 h8g8 a1e1 d8e8 e3e4 g7g5 d4f5 e6f5 f3f5 g5g4 e4e5 e7e6 f5e6 e8e6 h3g4 g8g4');
INSERT INTO evaluations VALUES(27060,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(27060,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(27060,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(27060,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(27060,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(27060,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(27060,36,'cp 50','c4e6 f7e6 d2d4 e5d4 f3d4 d8d7 d4c6 d7c6 c1g5 c6d7 d1e2 g8e7 g5e3 e7c6 a1d1 g7g6 e1g1 f8g7 c3b1 e8g8 c2c3 c6d8 c3c4 a7a6 e2c2 b7b5 b1c3 b5c4 c3e2 d8c6 c2c4 c6e5 c4c2 e5g4 e3c1');
INSERT INTO evaluations VALUES(27045,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(27045,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(27045,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(27045,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(27045,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(27045,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(27045,36,'cp 45','f7e6 d2d4 e5d4 f3d4 c6d4 d1d4 g8e7 c1e3 e7c6 d4c4 d8d7 a1d1 g7g6 f2f4 f8g7 e1g1 e8g8 b2b4 g8h8 b4b5 c6a5 c4d3 a7a6 a2a4 a6b5 a4b5 d7e7 b5b6 c7c6 e4e5 d6d5 c3a4 e7d7 e3c5 f8f7 a4b2 g7f8 c5f8 a8f8 d3e3');
INSERT INTO evaluations VALUES(27126,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(27126,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(27126,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(27126,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(27126,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(27126,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(27126,36,'cp 59','d2d4 c6d4 f3d4 e5d4 d1d4 g8e7 c1e3 d8d7 e1g1 e7c6 d4c4 g7g6 c3e2 f8g7 e2f4 c6d8 c4b3 e8g8 a2a4 b7b6 a1d1 f8e8 h2h4 d8f7 f2f3 a7a6 h4h5 g6g5 f4h3 b6b5 a4a5 d7e7 c2c3 h7h6 f3f4 g5f4 h3f4 f7e5 e3d4 a8d8 d4e5');
INSERT INTO evaluations VALUES(27121,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(27121,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(27121,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(27121,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(27121,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(27121,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(27121,36,'cp 61','c6d4 f3d4 e5d4 d1d4 g8e7 c1e3 d8d7 e1g1 e7c6 d4c4 g7g6 c3e2 f8g7 e2f4 c6d8 c4b3 e8g8 a2a4 b7b6 h2h4 a7a6 c2c3 d7f7 f4h3 e6e5 b3c2 a6a5 c2d3 f7e7 d3d5 d8f7 h3g5 h7h6 g5f7 e7f7 d5c6 a8e8 f2f3 g8h7');
INSERT INTO evaluations VALUES(28818,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(28818,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(28818,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(28818,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(28818,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(28818,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(28818,36,'cp 63','e7e6 a2a3 c7c5 d4d5 f8e7 d1f3 e8g8 g1e2 e6d5 e4d5 d7d6 g2g4 b8d7 c1f4 d7e5 f4e5 d6e5 e1c1 c5c4 d3e4 f6e4 c3e4 b6b5 c1b1 d8b6 e2g3 a8d8 g3f5 e7c5 e4c5 b6c5 f3e3 c5e3 f2e3 f7f6 e3e4 b7c8 f5e7 g8f7 e7c6 c8g4 c6d8 f8d8');
INSERT INTO evaluations VALUES(28759,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(28759,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(28759,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(28759,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(28759,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(28759,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(28759,36,'cp 75','a2a3 f8e7 h2h4 d7d6 g1f3 b8d7 c1f4 c7c5 e4e5 d6e5 d4e5 f6d5 c3d5 b7d5 d1e2 d8c7 c2c4 d5b7 h4h5 h7h6 d3e4 b7e4 e2e4 e8c8 h1h3 c7b7 e4b7 c8b7 h3g3 h8g8 a1d1 b7c7 e1e2 a7a6 b2b3 b6b5 c4b5 a6b5 a3a4 d7b6 d1d8 g8d8');
INSERT INTO evaluations VALUES(28765,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(28765,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(28765,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(28765,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(28765,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(28765,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(28765,36,'cp 76','d7d6 g1f3 b8d7 e1g1 f8e7 d1e2 e8g8 e4e5 b7f3 g2f3 f6h5 f3f4 g7g6 g1h1 h5g7 f1g1 c7c6 c3e4 d8c7 c2c3 f7f5 e4g5 e7g5 g1g5 c6c5 c1e3 a8c8 a1g1 f8f7 g5g2 d6d5 d3c2 c5d4 e3d4 g8h8 h2h4 d7c5');
INSERT INTO evaluations VALUES(31525,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(31525,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(31525,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31525,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(31525,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(31525,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31525,36,'cp 25','c6c5 f1e2 b8c6 e1g1 a7a6 a2a3 c5d4 c3d4 g8e7 c1e3 f5g4 b1d2 e7f5 h2h3 f5e3 f2e3 g4h5 b2b4 f8e7 d2b3 a6a5 b4b5 c6b8 a3a4 b8d7 f3e1 h5e2 d1e2 a8c8 e1d3 c8c3 e2d1 e8g8 a1c1 c3c4 b3d2 c4c1 d1c1');
INSERT INTO evaluations VALUES(31668,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(31668,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(31668,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31668,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(31668,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31668,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31668,36,'cp 23','f1e2 b8c6 e1g1 a7a6 a2a3 c5d4 c3d4 g8e7 c1e3 f5g4 b1d2 e7f5 h2h3 f5e3 f2e3 g4f5 f3e1 f8e7 e2d3 f5g6 d3g6 h7g6 e1d3 a8c8 d2b3 e8g8 g1h2 a6a5 a1c1 b7b6 b3d2 d8d7 d2f3 c6a7 d1b3 b6b5 b3d1 c8c1 d1c1');
INSERT INTO evaluations VALUES(31666,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(31666,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(31666,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(31666,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(31666,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(31666,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31666,36,'cp 29','b8c6 e1g1 c5d4 c3d4 g8e7 b1c3 h7h6 a2a3 a7a6 c1e3 f5h7 b2b4 e7f5 c3a4 f8e7 a1c1 a8c8 d1d2 e8g8 h2h3 f7f6 e5f6 e7f6 a4c5 d8e7 e2d3 f5d6 e3f4 d6e4 d2e3 f6d4 f3d4 c6d4 e3d4 f8f4 d3e4 d5e4 f1d1');
INSERT INTO evaluations VALUES(26412,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(26412,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(26412,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(26412,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26412,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26412,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(26412,36,'cp 27','e1g1 c5d4 c3d4 h7h6 a2a3 g8e7 b1c3 a7a6 c1e3 f5h7 b2b4 e7f5 c3a4 f8e7 e3c1 e8g8 c1b2 b7b5 a4c5 f5h4 f3h4 e7c5 d4c5 d8h4 f2f4 h7e4 e2d3 e4d3 d1d3 f8c8 f4f5 e6f5 f1f5 h4e4 d3e4 d5e4 a1f1');
INSERT INTO evaluations VALUES(53692,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(53692,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(53692,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(53692,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53692,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53692,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53692,36,'cp 3','c6d5 c4b5 c8d7 b5d7 b8d7 d1e2 d8c7 b1c3 g8f6 e1g1 f8b4 c3b5 c7c2 f3e5 c2e4 e2e4 d5e4 b5c7 e8d8 e5d7 d8c7 d7f6 g7f6 d2d4 e4d3 c1e3 h8d8 a1c1 c7d7 f1d1 d7e6 c1c4 b4a5 g1f1 a5b6 e3b6 a7b6 a2a3 d8d5 c4c7 a8b8 g2g4 b8g8 c7b7 g8g4 b7b6 e6e5 b6b7 e5f5 b7f7 d3d2');
INSERT INTO evaluations VALUES(54598,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54598,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54598,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54598,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54598,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54598,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54598,36,'cp 0','c4b5 c8d7 b5d7 b8d7 d1e2 d8c7 b1c3 g8f6 e1g1 f8b4 c3b5 c7c2 f3e5 c2e4 b5c7 e8d8 e2e4 d5e4 e5d7 d8c7 d7f6 g7f6 d2d3 e4d3 c1e3 a8d8 a1c1 c7b8 f1d1 d8d5 c1c4 b4a5 g1f1 a5b6 e3b6 a7b6 c4h4 f6f5 h4h3 d3d2 h3c3 h8d8 c3c2 b6b5 f1e2 d5e5 e2f1');
INSERT INTO evaluations VALUES(54461,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(54461,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54461,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54461,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(54461,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54461,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54461,36,'cp 0','c8d7 b5d7 b8d7 d1e2 d8c7 e1g1 g8f6 b1c3 f8b4 c3b5 c7c2 f3e5 c2e4 b5c7 e8d8 e2e4 d5e4 e5d7 d8c7 d7f6 g7f6 d2d3 e4d3 c1e3 a8d8 a1c1 c7b8 f1d1 d8d5 c1c4 b4a5 g1f1 h8e8 g2g3 a5b6 e3f4 b8a8 c4c3 e8d8 f4c7 d8d7 c7b6 a7b6 f1e1 d3d2 e1e2 f6f5 c3c2 d5e5 e2f1 e5d5');
INSERT INTO evaluations VALUES(30902,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30902,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30902,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30902,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30902,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30902,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30902,36,'cp 0','b5d7 b8d7 d1e2 d8c7 b1c3 g8f6 e1g1 f8b4 c3b5 c7c2 f3e5 c2e4 b5c7 e8d8 e2e4 d5e4 e5d7 d8c7 d7f6 g7f6 d2d4 e4d3 c1e3 a8d8 a1c1 c7b8 f1d1 d8d5 c1c4 b4a5 g1f1 a5b6 e3b6 a7b6 c4f4 f6f5 f4f3 d3d2 f3c3 h8e8 c3c2 e8d8 f1e2 d8e8');
INSERT INTO evaluations VALUES(30862,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30862,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30862,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30862,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30862,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30862,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30862,36,'cp 0','b8d7 d1e2 d8c7 e1g1 g8f6 b1c3 f8b4 c3b5 c7c2 f3e5 c2e4 b5c7 e8d8 e2e4 d5e4 e5d7 d8c7 d7f6 g7f6 d2d3 e4d3 c1e3 a8d8 a1c1 c7b8 f1d1 d8d5 c1c4 b4a5 g1f1 a5b6 e3b6 a7b6 c4f4 f6f5 f4f3 d3d2 f3c3 h8e8 c3c2 e8d8 f1e2 d5e5 e2f1');
INSERT INTO evaluations VALUES(26172,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26172,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26172,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26172,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26172,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26172,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26172,36,'cp 0','d1e2 d8c7 e1g1 g8f6 b1c3 f8b4 c3b5 c7c2 f3e5 c2e4 b5c7 e8d8 e2e4 d5e4 e5d7 d8c7 d7f6 g7f6 d2d3 e4d3 c1e3 a8d8 a1c1 c7b8 f1d1 d8d5 c1c4 b4a5 g1f1 a5b6 e3b6 a7b6 c4f4 f6f5 f4f3 d3d2 f3c3 b6b5 c3c2 h8d8 f1e2 d8e8 e2f1');
INSERT INTO evaluations VALUES(11585,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11585,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(11585,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(11585,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11585,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(11585,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11585,36,'cp 38','b5a4 e7g6 d2d4 e5d4 c3d4 f8b4 b1c3 d7d5 e4d5 b4c3 b2c3 d8d5 a4b3 d5e4 c1e3 e8g8 h2h3 c6a5 e1g1 c8f5 f1e1 a5b3 a2b3 c7c6 c3c4 e4d3 d1d3 f5d3 e1c1 f8e8 f3e1 d3e4 c1c3 a6a5 f2f3 e4f5 g1f2 h7h5 g2g4 f5e6 g4h5 g6e7');
INSERT INTO evaluations VALUES(11594,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(11594,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11594,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11594,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(11594,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(11594,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(11594,36,'cp 51','e7g6 d2d4 e5d4 c3d4 f8b4 b1c3 d7d5 e4d5 b4c3 b2c3 d8d5 a4b3 d5e4 c1e3 e8g8 e1g1 c8g4 f3g5 g4d1 g5e4 d1b3 a2b3 b7b6 f1e1 f8e8 f2f3 c6a5 a1a3 a5c6 e3d2 e8d8 g1f1 a6a5 h2h4 h7h5 e4g3 d8d7 g3h5 g6h4 f1f2');
INSERT INTO evaluations VALUES(12773,30,'cp -139',NULL);
INSERT INTO evaluations VALUES(12773,31,'cp -151',NULL);
INSERT INTO evaluations VALUES(12773,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(12773,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(12773,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(12773,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(12773,36,'cp -147','e7e6 b1c3 d7d5 c4b5 d5d4 b5c6 b7c6 c3e2 h7h6 g5f3 f6e4 d2d3 e4f6 e2g3 f8d6 e1g1 e8g8 f1e1 f6d7 f3d2 d7b6 d2e4 d6c7 e4c5 d8d6 c5e4 d6e7 e4d2 c6c5 d2f3 c8b7 f3e5 c7e5 e1e5 b6d7 e5e1 f7f5 f2f4 a8e8 b2b3 e6e5 f4e5');
INSERT INTO evaluations VALUES(12343,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(12343,31,'cp -152',NULL);
INSERT INTO evaluations VALUES(12343,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(12343,33,'cp -150',NULL);
INSERT INTO evaluations VALUES(12343,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(12343,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(12343,36,'cp -151','b1c3 d7d5 c4b5 d5d4 b5c6 b7c6 c3e2 h7h6 g5f3 f6e4 d2d3 e4f6 e2g3 f8d6 e1g1 e8g8 f1e1 d8c7 d1e2 a7a5 f3e5 c8a6 b2b3 f8e8 c1d2 a5a4 f2f4 f6d5 e2g4 a6b5 g3h5 d6f8 g4f3 f7f6 e5c4 c7f7');
INSERT INTO evaluations VALUES(12340,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(12340,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(12340,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(12340,33,'cp -164',NULL);
INSERT INTO evaluations VALUES(12340,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(12340,35,'cp -157',NULL);
INSERT INTO evaluations VALUES(12340,36,'cp -156','d7d5 c4b5 d5d4 c3e2 h7h6 b5c6 b7c6 g5f3 f6e4 e1g1 f8d6 d2d3 e4f6 e2g3 e8g8 f1e1 f6d7 f3d2 d6c7 d2c4 e6e5 d1h5 f8e8 b2b3 a7a5 a2a4 d7f6 h5h4 f6d5 h4d8 e8d8 c1d2 f7f6 g3e4 d5b4 d2b4 c5b4');
INSERT INTO evaluations VALUES(22574,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(22574,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(22574,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(22574,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22574,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(22574,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(22574,36,'cp -23','d7d6 e5d6 f8d6 d2d4 g8f6 f1b5 a7a6 b5c6 b7c6 c1e3 d8e7 e1g1 c6c5 d4c5 d6c5 e3c5 e7c5 f1e1 c8b7 f3e5 e8g8 b1c3 f8d8 d1e2 a6a5 e2b5 c5b5 c3b5 f6d5 e1d1 f7f6 e5d3 b7a6 c2c4 c7c6 d3c5 c6b5 c4d5 e6d5 a2a3');
INSERT INTO evaluations VALUES(22063,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(22063,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(22063,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(22063,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(22063,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(22063,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(22063,36,'cp -10','e5d6 f8d6 d2d4 g8f6 f1b5 a7a6 b5c6 b7c6 c1e3 d8e7 e1g1 c6c5 d4c5 d6c5 e3c5 e7c5 f1e1 e8g8 f3e5 c8b7 b1c3 a8d8 d1e2 f6d5 e5d3 c5b6 c3a4 b6c6 a4c5 d5f4 d3f4 c6c5 f4d3 c5c6 f2f3 c6b6 g1h1 a6a5 b2b3 b7a6 c2c4 d8d4 a1d1 a5a4 d3e5 d4d1');
INSERT INTO evaluations VALUES(21845,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(21845,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(21845,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(21845,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(21845,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(21845,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(21845,36,'cp -16','f8d6 d2d4 g8f6 f1d3 e6e5 d4e5 c6e5 f3e5 d6e5 h2h3 c7c6 e1g1 c8e6 d1f3 d8c7 f1e1 e8g8 b1c3 a8e8 c1d2 f6d7 f3h5 f7f5 e1e2 d7c5 a1e1 c5d3 c2d3 e6d7 d2f4 e5f4 e2e8 f4h2 g1h1 d7e8 e1e8 h2d6 e8f8 d6f8 h5f5 c7f7 f5f7 g8f7 a2a3 f8d6');
INSERT INTO evaluations VALUES(5928,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(5928,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(5928,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(5928,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(5928,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(5928,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(5928,36,'cp -4','d2d4 g8f6 f1d3 e6e5 d4e5 c6e5 f3e5 d6e5 e1g1 e8g8 h2h3 c7c6 f1e1 e5c7 b1d2 c8e6 d2f3 d8d6 c1e3 e6d5 d3f1 f8d8 c2c4 d5f3 d1d6 d8d6 g2f3 d6d8 b2b4 c7b6 c4c5 b6c7 a1d1 c7e5 b4b5 f6d5 e3g5 e5f6 g5f6 d5f6 b5c6 b7c6 d1d8 a8d8 e1b1 g8f8 b1b7 d8d1 b7a7');
INSERT INTO evaluations VALUES(20075,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(20075,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(20075,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(20075,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(20075,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20075,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(20075,36,'cp 0','g8f6 f1b5 c8d7 b1c3 g7g6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 d1d2 d8a5 e1c1 f8g7 e3h6 e8g8 h2h4 g7h6 d2h6 f8b8 d1d3 f6d7 c3d1 a5a2 d3a3 a2c4 h4h5 d7f6 a3c3 c4a4 c3a3');
INSERT INTO evaluations VALUES(13375,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13375,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(13375,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(13375,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(13375,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(13375,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(13375,36,'cp 23','f1b5 c8d7 b1c3 g7g6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 e1g1 f8g7 c3a4 e8g8 d1d2 d8c7 c2c4 f6d7 a1d1 f8d8 b2b3 d7c5 a4c5 d6c5 d2f2 d8d1 f1d1 a8d8 d1d8 c7d8 f2d2 g7d4 g1f1 d8b6 e3f4 a6a5 d2c2 b6b4 f1e2 g8g7 h2h3 e7e5 f4d2 b4a3 d2g5 f7f6 g5c1');
INSERT INTO evaluations VALUES(28948,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(28948,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(28948,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(28948,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28948,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(28948,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28948,36,'cp 13','d5c4 e2e3 f6d5 f4b8 a8b8 f1c4 e7e6 e1g1 f8e7 a1c1 d5c3 c1c3 e8g8 c4d3 f5g4 d3e2 e7d6 h2h3 g4f3 e2f3 d8e7 d1c2 g7g6 g2g3 e6e5 f1d1 e5d4 c3d3 f8d8 e3d4 g8g7 g1g2 b8c8 h3h4 h7h5 c2b3 d8d7 d3e3 e7f6');
INSERT INTO evaluations VALUES(37225,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(37225,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(37225,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(37225,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(37225,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(37225,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(37225,36,'cp 97','f3e5 c5d4 e5f3 d4c3 d2c3 f6e4 e1g1 h7h6 f1e1 e4f6 c1f4 d7d5 c4f1 c7c5 c3c4 c8e6 c4d5 d8d5 d1d5 e6d5 f4d6 f8c8 a1d1 b7b6 f3e5 b8c6 c2c4 d5e6 e5c6 c8c6 d6g3 c6c8 f2f3 c8d8 b2b3 d8d7 d1d7 f6d7 e1d1 f7f6 d1d6 d7f8 g1f2 g8f7 d6c6 a8c8');
INSERT INTO evaluations VALUES(37205,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(37205,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(37205,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(37205,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(37205,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(37205,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(37205,36,'cp 78','c5d4 e5f3 d4c3 d2c3 f6e4 e1g1 h7h6 f1e1 e4f6 b2b3 d7d5 c4d3 c7c5 c1f4 c8g4 d1d2 c5c4 b3c4 g4f3 g2f3 d5c4 d3c4 d8d2 f4d2 b8c6 d2f4 f8e8 e1e8 a8e8 a1d1 c6e5 c4e2 e5g6 f4e3 e8c8 e3a7 g6f4 e2f1 c8c3 d1d8 g8h7 a7e3 f6d5 f1d3 f4d3 d8d5');
INSERT INTO evaluations VALUES(38480,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(38480,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(38480,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(38480,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(38480,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(38480,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(38480,36,'cp 70','f7f6 c1f4 g8h6 c2c4 g6g5 f4e3 g5g4 e5f6 e7f6 f3g1 e8g8 b1c3 h6f5 g1e2 d5c4 e2f4 g7h6 f1c4 g8h8 d1d2 f5e3 f2e3 h6f4 e3f4 f8e8 e1f2 c8f5 h1e1 e8e1 a1e1 b8c6 c4d5 c6b4 d5b7 b4d3 f2g3 d3e1 b7a8 e1d3 a8d5 c7c6 d5c4 d8d4 c4d3 f5d3');
INSERT INTO evaluations VALUES(38325,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(38325,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38325,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(38325,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(38325,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(38325,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(38325,36,'cp 50','c1f4 g8h6 c2c4 g6g5 f4e3 g5g4 e5f6 e7f6 f3h4 f6f5 g2g3 f5f4 e3f4 e8g8 d1d2 h6f5 h4f5 c8f5 f1g2 d8e8 f4e3 d5c4 g2d5 g8h8 d5b7 b8c6 b7a8 e8a8 e1g1 f8d8 b1c3 c6e5 f2f4 e5f3 f1f3 a8f3 a1f1 f3b7 f1e1 h8g8 e3f2 c7c5 d4d5 g7c3 d2c3 b7d5 e1e5 d5d1 f2e1');
INSERT INTO evaluations VALUES(44858,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(44858,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(44858,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(44858,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(44858,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(44858,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(44858,36,'cp -46','g1f3 b8d7 b2b4 e7e5 e2e3 e5e4 f3d2 f8e7 b1c3 d7f8 f2f3 e4f3 d2f3 c8f5 f1d3 f5d3 d1d3 d8d7 e1g1 f8g6 c1d2 e8g8 b4b5 f8e8 a2a4 e7d8 a4a5 a7a6 b5b6 f6e4 d2e1 d8f6 e1g3 e4g3 h2g3 d7g4 c3e2 f6g5 f3g5');
INSERT INTO evaluations VALUES(44857,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(44857,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(44857,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(44857,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(44857,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(44857,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(44857,36,'cp -43','b8d7 e2e3 e7e5 b2b4 e5e4 f3d2 f8e7 b1c3 d7f8 f2f3 e4f3 d2f3 c8f5 f1d3 f5d3 d1d3 d8d7 e1g1 f8g6 a2a4 e8g8 b4b5 e7d8 a4a5 a7a6 b5b6 f8e8 c1d2 f6e4 d2e1 d8f6 f3d2 e4c3 d3c3 e8e6 e1f2 a8e8 d2f3 d7e7 a1e1 f6g5 e1e2 g6f8 f1e1 h7h5 f3g5 e7g5');
INSERT INTO evaluations VALUES(49775,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49775,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49775,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49775,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49775,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49775,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49775,36,'cp -9','e4e5 f6d5 d1d4 d5e7 f1c4 e7g6 g1f3 b8c6 d4e3 d8e7 e1g1 g6e5 f3e5 c6e5 c4b3 d7d6 b1a3 c8d7 c1d2 e5g6 e3g3 d7e6 a1e1 e7d7 f2f4 e8c8 b3e6 f7e6 g3e3 a7a6 e3e6 f8e7 e6d7 c8d7 g2g3 h7h5 a3c2 d8e8 c2e3 e7f6 e3d5 e8e1 d2e1');
INSERT INTO evaluations VALUES(49503,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(49503,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(49503,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(49503,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(49503,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(49503,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(49503,36,'cp -14','f6d5 d1d4 d5e7 g1f3 e7g6 d4e4 b8c6 f1d3 f8e7 d3c2 e8g8 e1g1 f8e8 e4d3 d7d6 e5d6 d8d6 d3d6 e7d6 c1e3 c6e7 b1d2 e7d5 e3d4 e8d8 d2e4 d6e7 f1e1 b7b6 f3e5 c7c5 e5g6 h7g6 d4e5 f7f6 e5g3 f6f5 e4d2 e7f6 a1d1 c8a6');
INSERT INTO evaluations VALUES(49811,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49811,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(49811,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49811,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49811,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(49811,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(49811,36,'cp 0','g1f3 b8c6 c3d4 d7d6 f1b5 d6e5 e1g1 e5d4 d1d4 d5b4 d4c3 d8d6 c1g5 c8e6 f3d4 a7a6 b5a4 b4d5 c3c2 d5b4');
INSERT INTO evaluations VALUES(31312,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(31312,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(31312,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(31312,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(31312,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(31312,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(31312,36,'cp 142','d1b3 b8c6 g1f3 a8c8 c1f4 g8f6 b3b7 f5d7 h2h3 e7e5 f3e5 c8b8 e5c6 b8b7 c6d8 b7b2 a1b1 b2b1 c3b1 f8b4 f4d2 b4d2 b1d2 e8d8 e2e3 d8e7 f1d3 h8b8 d2b3 f6e4 d3e4 d5e4 e1g1 d7e6 f1c1 e7d6 b3c5 e6d5 c1c2 b8b1 g1h2 b1a1 a2a4 g7g5 c2b2 f7f5 b2b8 a1a2 b8d8 d6c6 c5a6 d5b3 a6b4 c6c7');
INSERT INTO evaluations VALUES(31308,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(31308,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(31308,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(31308,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(31308,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(31308,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(31308,36,'cp 113','g8f6 b3b7 b8d7 g1f3 a8b8 b7a7 e7e6 e2e3 f8b4 c1d2 e8g8 a7a4 b4e7 f1b5 d7b6 a4d1 b6c4 d1c1 f5d3 b5c4 d5c4 f3e5 d8b6 e5d3 c4d3 a1b1 b6a6 c1d1 e7b4 e1g1 f8c8 a2a4 b4c3 d2c3 f6e4 b1c1 e4c3 b2c3 a6c4 d1d2 b8b3 g2g3 b3c3');
INSERT INTO evaluations VALUES(52709,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(52709,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(52709,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(52709,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(52709,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(52709,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52709,36,'cp 9','c6d5 d2d4 g8f6 c1f4 b8c6 e2e3 c8f5 d1b3 c6a5 f1b5 f5d7 b5d7 d8d7 b3b5 a5c4 g1f3 e7e6 e1g1 a7a6 b5b3 f8d6 f4g5 a8c8 g5f6 g7f6 e3e4 d5e4 c3e4 d6e7 e4c3 c4a5 b3d1 a5c4 d1e2 c4b6 a1c1 e8g8 f1d1 g8h8 a2a3');
INSERT INTO evaluations VALUES(22817,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22817,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22817,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22817,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22817,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(22817,36,'cp 0','g8f6 b1c3 f8c5 g1e2 a7a5 d2d3 e8g8 e1g1 f8e8 g1h1 f6g4 h1g1');
INSERT INTO evaluations VALUES(47051,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47051,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(47051,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(47051,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(47051,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(47051,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47051,36,'cp 0','b1c3 c7c5 g1f3 b8c6 f1b5 c8d7 d2d3 e7e6 e1g1 a7a6 b5c4 d8c7 d1e1 e8c8 c4b3 d6d5 e4e5 f6g4 b3d5 e6d5 c3d5 c7b8 d5b6 c8c7 b6d5');
INSERT INTO evaluations VALUES(47049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47049,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(47049,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47049,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(47049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47049,36,'cp 0','c7c5 g1f3 b8c6 f1b5 c8d7 d2d3 e7e6 e1g1 a7a6 b5c4 d8c7 c4b3 f8e7 d1e1 e8g8 e4e5 f6e8 a2a4 c6a5 b3a2 a5c6');
INSERT INTO evaluations VALUES(45220,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(45220,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(45220,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(45220,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45220,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45220,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45220,36,'cp 0','f1b5 c8d7 d2d3 b8c6 g1f3 e7e6 b5a4 a7a6 a4b3 f8e7 e1g1 d8c7 d1e1 e8g8 a2a4 c6a5 b3a2 a5c6');
INSERT INTO evaluations VALUES(47902,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(47902,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(47902,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47902,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47902,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(47902,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(47902,36,'cp 40','b8c6 f1b5 e5f4 d2d4 a7a6 b5e2 d6d5 e4d5 f6d5 c3d5 d8d5 c1f4 f8d6 f4d6 d5d6 e1g1 e8g8 c2c3 c6e7 f3d2 c8f5 d2c4 d6d7 c4e5 d7d6 e2c4 f5e6 d1b3 e6c4 e5c4 d6d5 a1e1 e7g6 c4e3 d5b3 a2b3 f7f6 b3b4 g6e7 e3c4 e7d5 c4a5 a8b8');
INSERT INTO evaluations VALUES(15607,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(15607,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(15607,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(15607,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(15607,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(15607,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(15607,36,'cp 62','f1b5 e5f4 d2d4 a7a6 b5e2 d6d5 e4d5 f6d5 c3d5 d8d5 e1g1 c8g4 c1f4 e8c8 c2c3 f7f6 h2h3 g4h5 b2b4 f8d6 f4d6 d5d6 a2a4 c6e7 f3d2 h5e2 d1e2 h8e8 b4b5 a6a5 e2g4 c8b8 g4g7 e7d5 d2c4 d6e6 f1e1 e6e1 a1e1 e8e1 g1h2 e1e7 g7g3 b7b6 g3d3 h7h6 d3g6 e7e6 g6h6');
INSERT INTO evaluations VALUES(15438,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(15438,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(15438,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(15438,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(15438,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(15438,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(15438,36,'cp 36','e5f4 d2d4 a7a6 b5e2 d6d5 e4d5 f6d5 c3d5 d8d5 e1g1 c8g4 c1f4 e8c8 c2c3 f7f6 h2h3 g4h5 b2b4 f8d6 f4d6 d5d6 a2a4 c6e7 f3d2 h5e2 d1e2 h8e8 d2e4 d6e6 a1e1 c8b8 e2d3 e7d5 d3g3 e6c6 b4b5 a6b5 a4b5 c6b5 g3g7 e8e7 g7g3 b5c6 g3d3 d8e8');
INSERT INTO evaluations VALUES(48827,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(48827,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(48827,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(48827,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(48827,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(48827,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(48827,36,'cp -25','e4d5 f6d5 g2g3 f8b4 c1d2 d5c3 b2c3 b4a5 g1f3 e8g8 f1g2 b8c6 e1g1 c8g4 h2h3 g4e6 d1e2 f7f6 d3d4 d8d7 d4e5 e6h3 e5f6 h3g2 g1g2 f8f6 a1d1 a8f8 e2c4 d7f7 c4f7 f8f7 f3g5 f7e7 f1e1 a5b6 f2f4 h7h6 e1e7 c6e7 g5f3 e7c6 c3c4 g8f7 d1e1');
INSERT INTO evaluations VALUES(48607,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(48607,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(48607,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(48607,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(48607,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(48607,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(48607,36,'cp -23','f6d5 g2g3 f8b4 c1d2 d5c3 b2c3 b4d6 g1f3 e8g8 f1g2 b8c6 a1b1 a8b8 e1g1 h7h6 f1e1 c8g4 h2h3 g4e6 c3c4 d8d7 d2c3 f7f6 d3d4 e5d4 f3d4 e6h3 g2d5 g8h8 b1b7 b8b7 d5c6 d7c8 d4e6 h3e6 c6b7 c8d7 b7d5 e6f7 d5f7 d7f7');
INSERT INTO evaluations VALUES(49022,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49022,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(49022,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(49022,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(49022,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49022,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(49022,36,'cp -12','g2g3 f8b4 c1d2 e8g8 f1g2 d5c3 b2c3 b4d6 g1f3 b8c6 a1b1 d8d7 e1g1 b7b6 f1e1 a8b8 d3d4 f7f6 d2e3 g8h8 d4e5 c6e5 f3e5 f6e5 c3c4 d7e7 d1d2 h7h6 d2c3 c8f5 c4c5 b6c5 b1b8 f8b8 c3a3 f5c2 a3a7 e7f8 a2a4 e5e4 a4a5');
INSERT INTO evaluations VALUES(35136,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(35136,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(35136,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(35136,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(35136,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(35136,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(35136,36,'cp 173','f4e5 f6e5 f1g2 d7d5 d2d4 e5e7 e4e5 c8f5 b1c3 c7c6 g1f3 b8a6 f3h4 f5e6 c3e2 h7h5 h4f3 a6c7 e1g1 e8c8 e2f4 e6g4 h2h3 g4d7 h3h4 d7g4 c2c3 g8h6 a2a4 e7d7 b2b4 f8e7 d1e1 c7e6 b4b5 c8b8 c1d2 d8e8 f3h2 e6f4 d2f4 g4e6 h2f3 h6f5');
INSERT INTO evaluations VALUES(35090,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(35090,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(35090,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(35090,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(35090,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(35090,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(35090,36,'cp 169','f6e5 f1g2 d7d5 d2d4 e5e7 e4e5 c8f5 b1c3 c7c6 g1f3 b8a6 f3h4 f5e6 e1g1 h7h5 d1f3 a6c7 c3e2 e6g4 f3d3 c7e6 c2c3 g7g6 h2h3 g4e2 d3e2 f8h6 g1h2 h6c1 a1c1 e8c8 b2b4 g8h6 b4b5 g6g5 h4f5 h6f5 f1f5 h5h4 g3h4 c6b5 h4g5 e6g5 e2b5');
INSERT INTO evaluations VALUES(35212,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(35212,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(35212,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(35212,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(35212,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(35212,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(35212,36,'cp 171','f1g2 d7d5 d2d4 e5e7 e4e5 c8f5 b1c3 c7c6 g1f3 b8a6 f3h4 f5e6 c3e2 h7h5 h4f3 e8c8 e1g1 a6c7 a2a4 e7d7 c2c3 f8e7 b2b4 c8b8 e2f4 h5h4 f4e6 c7e6 g3h4 e7h4 f3h4 h8h4 b4b5 c6c5 d4c5 g8h6 c5c6 d7e7 d1e1 h4g4 h2h3 g4g2 g1g2');
INSERT INTO evaluations VALUES(54021,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(54021,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(54021,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(54021,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(54021,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(54021,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(54021,36,'cp 78','c2c4 g8f6 b1c3 f8e7 c1e3 e8g8 f1e2 b8c6 e1g1 d6d5 e4d5 e6d5 f1e1 c6d4 e3d4 c8e6 c4d5 f6d5 e2f3 d5c3 b2c3 e7f6 f3b7 a8b8 b7f3 f6d4 c3d4 e6d5 f3d5 d8d5 d1a4 a7a6 e1e2 d5b5 a4b5 b8b5 e2d2 f8d8 a1d1 h7h6 h2h3');
INSERT INTO evaluations VALUES(54019,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(54019,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(54019,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(54019,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(54019,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(54019,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(54019,36,'cp 77','g8f6 b1c3 b8c6 c1e3 f8e7 f1e2 e8g8 e1g1 d6d5 e4d5 e6d5 f1e1 c6d4 e3d4 c8e6 c4d5 f6d5 e2f3 d5c3 b2c3 e7f6 f3b7 a8b8 b7e4 f6d4 c3d4 e6d5 e4d5 d8d5 e1e5 d5d6 d1d2 h7h6 e5a5 f8d8 a1d1 d6g6 h2h3 b8b1 d1b1 g6b1 g1h2');
INSERT INTO evaluations VALUES(54288,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(54288,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(54288,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(54288,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(54288,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(54288,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(54288,36,'cp -91','b8c6 d4c5 c6e5 c1e3 g8h6 g1f3 e5c6 e3g5 f8e7 d1d2 h6g4 f1b5 e8g8 g5e7 d8e7 h2h3 g4h6 c3a4 f8d8 e1c1 e6e5 c2c4 d5c4 d2c3 c8d7 h1e1 f7f6 b5c4 g8h8 b2b4 c6a5 a4b2 a5c4 b2c4 d7c6 c4e3 h6f7');
INSERT INTO evaluations VALUES(20330,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(20330,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(20330,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(20330,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(20330,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(20330,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(20330,36,'cp -79','d4c5 c6e5 c1e3 e5c6 g1f3 g8h6 e3g5 f8e7 d1d2 h6g4 g5e7 d8e7 d2g5 e7g5 f3g5 c8d7 g5f3 a8c8 a2a3 g4e5 f3e5 c6e5 b2b4 g7g5 e1c1 e8e7 f1b5 b7b6 h2h4 g5g4 b5d7 e7d7 h1e1 f7f6 c3a4 c8b8 c2c3 h7h5 c1c2 e5c4 c5c6 d7d6 a4b2 c4b2 c2b2');
INSERT INTO evaluations VALUES(6903,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(6903,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(6903,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(6903,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(6903,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(6903,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(6903,36,'cp 34','d8f6 e4e5 f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 a2a3 h8e8 e1d2 d7d6 a1e1 e8e1 d2e1 c8f5 e1d2 a8a5 f1e2 a5d5 d2c1 d5e5 h1e1 h7h5 g2g3 f5g4 c1d2 e5d5 e2d3 g4f5 e1e3 f5d3 e3d3 d5f5 f2f3 f5c5 b2b4 c5c6 b4b5 c6c5');
INSERT INTO evaluations VALUES(369,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(369,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(369,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(369,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(369,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(369,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(369,36,'cp 44','e4e5 f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 f2f3 f6f5 e1f2 d7d5 a2a3 d8e7 a1d1 c7c6 f1d3 c8d7 h1e1 e7f6 c2c3 b6b5 f3f4 a8a4 g2g3 a4a8 e1e5 a8e8 e5e8 d7e8 d1e1 e8d7 d3c2 h7h5 h2h4 d7e6 c2d1 e6f7 d1b3');
INSERT INTO evaluations VALUES(361,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(361,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(361,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(361,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(361,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(361,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(361,36,'cp 32','f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 a2a3 h8e8 e1d2 d7d6 f1d3 a8a5 a1e1 c8f5 e1e8 d8e8 h1e1 e8f8 d3e2 a5d5 d2c1 d5e5 f2f3 h7h5 c2c3 f5d3 c1d2 d3e2 e1e2 f6f5 d2e1 f8f7 e1f2 f7e6 e2d2 f5f4 d2d4 e6f5 g2g3 f4g3 h2g3 f5g6 d4h4 e5a5 f2e3 a5e5 e3f4');
INSERT INTO evaluations VALUES(233,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(233,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(233,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(233,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(233,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(233,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(233,36,'cp 40','d4c3 g8e7 b1d2 e7d5 c3b3 b6b3 d2b3 b7b6 e1c1 c8b7 h1g1 e8c8 f1c4 d5b4 a2a3 b4c6 f2f4 h8f8 g1e1 f7f6 e5f6 f8f6 g2g3 h7h6 h2h4 f6f8 c4d3 d8e8 e1e8 f8e8 c1d2 c8d8 d1e1 e8f8 d3e4 f8e8 d2e3 e8e7 e3f2');
INSERT INTO evaluations VALUES(16483,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(16483,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(16483,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(16483,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(16483,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(16483,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(16483,36,'cp 159','d4e5 c6e5 f2f4 e5g6 e4e5 f6g8 d1e2 g8h6 e2e4 c7c6 c1d2 d8e7 e1c1 d7d5 e4e1 f7f6 g1f3 f6e5 f4e5 c8g4 e1g3 g4f3 g2f3 h6f5 g3g4 g6h4 f1h3 g7g6 h1e1 f8g7 f3f4 e8g8 g4e2 a8e8 e2d3 e7b4 c1b1');
INSERT INTO evaluations VALUES(16355,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(16355,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(16355,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(16355,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(16355,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(16355,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(16355,36,'cp 137','c6e5 f2f4 e5g6 e4e5 f6g8 f1d3 d7d5 g1f3 g8h6 d3g6 h7g6 d1d5 f8e7 c1e3 d8d5 c3d5 e7d8 h2h3 h6f5 e3f2 c7c6 d5c3 d8a5 e1g1 a5c3 b2c3 f5e7 f3g5 e8g8 c3c4 b7b6 f1d1 c6c5 d1d6 c8f5 g2g4 f5c2 e5e6 f7f6 g5f7 c2a4 a1c1 a8c8 c1e1 f8e8 g1h2 a4c6');
INSERT INTO evaluations VALUES(16960,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(16960,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(16960,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(16960,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(16960,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(16960,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(16960,36,'cp 150','f2f4 e5g6 e4e5 f6g8 g1f3 g8h6 c1e3 c7c6 d1d2 d7d5 h2h3 h6f5 e3f2 h7h5 f1d3 f8b4 e1g1 d5d4 d3f5 c8f5 d2d4 d8d4 f3d4 f5d7 c3e4 e8g8 c2c3 b4e7 f2g3 f7f5 e5f6 e7f6 a1d1 h5h4 g3h2 a8e8 e4d6 f6d4 d1d4 e8e2 f4f5 g6e7 d4h4 e2b2 h4h5 g7g6 f5g6 f8f1 g1f1 e7g6 h5g5 g8g7 a2a3 b2b1 f1f2');
INSERT INTO evaluations VALUES(17294,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(17294,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17294,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17294,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(17294,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(17294,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(17294,36,'cp 36','d2d4 d7d5 f1g2 c8f5 e1g1 e7e6 c2c4 f8e7 f3e5 e8g8 b1c3 e7d6 e5c6 b7c6 d1a4 d8d7 c1g5 f8b8 g5f6 g7f6 c4d5 e6d5 e2e4 b8b4 a4c2 d5e4 a1d1 a8b8 b2b3 f5g4 c3e4 d6e7 d1d2 b4d4 d2d4 d7d4 c2c6 d4e5 e4d2');
INSERT INTO evaluations VALUES(17262,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(17262,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17262,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(17262,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(17262,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(17262,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(17262,36,'cp 48','d7d5 f1g2 c8f5 c2c4 e7e6 e1g1 f8e7 f3e5 e8g8 b1c3 h7h6 c4d5 e6d5 c1f4 f8e8 a1c1 e7f8 e5c6 b7c6 c3a4 d8d7 a4c5 d7c8 f1e1 a8b8 d1d2 g7g5 f4e5 f6d7 c5d7 c8d7 a2a3 b8b3 c1c3 b3c3 d2c3 f8d6 e5d6');
INSERT INTO evaluations VALUES(46925,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(46925,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(46925,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(46925,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(46925,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46925,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(46925,36,'cp 38','d6e5 f3e5 g7g6 b1c3 f8g7 c1e3 c7c5 d1d2 c8e6 b2b3 e8g8 a1d1 c5d4 e3d4 b6d7 e5d7 d8d7 d4g7 d7d2 e1d2 g8g7 f1e2 b8a6 e2f3 a6c5 c3b5 a7a5 h1e1 a8d8 d2c3 d8d1 f3d1 f8d8 d1c2 g7f6 b5c7 d8c8 c7e6 c5e6 a2a3 b7b5 g2g3 c8c5 e1d1 b5c4');
INSERT INTO evaluations VALUES(46938,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(46938,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(46938,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(46938,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(46938,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(46938,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46938,36,'cp 47','f3e5 g7g6 c1e3 f8g7 b1c3 c7c5 d1d2 c5d4 e3d4 g7e5 d4e5 d8d2 e1d2 e8g8 e5f4 b8c6 a1d1 f8d8 d2c1 d8d1 c3d1 c8f5 d1e3 f5e6 b2b3 a8d8 g2g3 b6d7 f1g2 d7c5 g2c6 c5d3 c1c2 d3b4 c2c3 b4c6 b3b4 f7f6 h1d1 d8d1 e3d1 a7a6 b4b5 a6b5 c4b5 c6a7 b5b6 a7b5 c3b4');
INSERT INTO evaluations VALUES(46934,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(46934,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(46934,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(46934,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(46934,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(46934,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(46934,36,'cp 38','g7g6 b1c3 f8g7 c1e3 c7c5 d1d2 c8e6 b2b3 b8c6 e5c6 b7c6 e1c1 c5d4 e3d4 g7d4 d2d4 d8d4 d1d4 b6d7 f1e2 e8c8 h1d1 a7a5 c1b2 c8c7 h2h4 h7h6 f2f4 d7c5 d4d8 h8d8 d1d8 c7d8 a2a3 e6f5 b3b4 a5b4 a3b4 c5d3 e2d3 f5d3 b2b3 f7f6 c3a4 e7e5 f4e5 f6e5 b3c3 d3f5 g2g3 d8c7 a4c5');
INSERT INTO evaluations VALUES(55741,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(55741,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(55741,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(55741,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(55741,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(55741,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(55741,36,'cp 62','c7c5 d2d4 c5d4 f3d4 a7a6 g2g4 b7b5 a2a3 g8e7 h2h4 e7c6 f2f4 h7h6 f1g2 c6d4 d1d4 b8c6 d4f2 a8b8 e1g1 f8e7 g4g5 h6g5 h4g5 c6a5 b2b3 d8c7 c3a2 a5c6 a2b4 c8b7 b4d3 b8c8 c1b2 c6d8 b2g7 h8g8 g7b2 c7c2');
INSERT INTO evaluations VALUES(21491,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(21491,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(21491,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(21491,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(21491,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(21491,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(21491,36,'cp 185','a7a6 e1g1 b7b5 c4d3 f8e7 b1c3 g8f6 f1e1 c8b7 a2a3 b5b4 c3a4 e8g8 d3c2 f6d7 d1d3 e6e5 d4d5 c6a7 a3b4 e7b4 c1g5 f7f6 g5d2 b4d6 f3h4 a7b5 h4f5 g7g6 f5d6 c7d6 d2e3 f6f5 f2f3 f8f7 c2b3 g8h8 e3f2 a8c8');
INSERT INTO evaluations VALUES(36972,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36972,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36972,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(36972,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(36972,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36972,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(36972,36,'cp 8','c2c4 c7c5 c4d5 d6g3 h2g3 c5d4 d5e6 d8a5 d1d2 a5d2 b1d2 d4e3 e6f7 f8f7 f2e3 f7e7 f1c4 c8e6 e1g1 b8c6 f1d1 a8e8 a1c1 e6d5 d2f1 d5c4 c1c4 h7h6 a2a3 a7a6 d1d3 f6e4 c4c2 e4f6 c2c1 f6g4 c1e1 g8h7 d3d2 g4e3 f1e3 e7e3');
INSERT INTO evaluations VALUES(22362,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(22362,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(22362,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22362,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(22362,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(22362,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(22362,36,'cp 33','d6e5 f1c4 f8b4 c2c3 b4c5 f3g5 g8h6 b2b4 c5d6 a2a3 e8g8 c4a2 g8h8 e1g1 a7a5 f1e1 d8e7 g5f3 a5b4 a3b4 d7f6 b1a3 h6g8 h2h3 c8e6 d1e2 b7b6 a2e6 e7e6 c1d2 h7h6 c3c4 g8e7 a3b5 a8a1 e1a1 f8c8');
INSERT INTO evaluations VALUES(22402,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(22402,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(22402,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(22402,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(22402,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(22402,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(22402,36,'cp 37','f1c4 f8b4 c2c3 b4c5 f3g5 g8h6 b2b4 c5d6 a2a3 a7a5 c4a2 e8g8 e1g1 g8h8 f1e1 d8e7 g5f3 h6g8 b1d2 b7b5 d2b3 a5b4 a3b4 g8f6 b3a5 d7b6 d1e2 c8d7 c1d2 h7h6 h2h3 h8g8 c3c4 b6c4 a5c4 b5c4 a2c4');
INSERT INTO evaluations VALUES(22400,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(22400,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(22400,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(22400,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(22400,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(22400,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(22400,36,'cp 23','f8b4 c2c3 b4c5 b2b4 c5d6 f3g5 g8h6 a2a3 e8g8 e1g1 d7b6 c4a2 g8h8 f1e1 d8f6 g5f3 f6g6 c3c4 c7c5 c1e3 d6e7 f3e5 g6f6 b4c5 h6g4 e5g4 f6a1 c5b6 a1a2 b6a7 a2c4 g4e5 c4e6 d1d5 f7f6 e5f3 e6d5 e4d5 e7d6 b1c3');
INSERT INTO evaluations VALUES(50238,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(50238,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(50238,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(50238,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(50238,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(50238,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(50238,36,'cp 57','d7d6 d2d4 f8g7 c1g5 c7c6 d1d2 c8g4 h2h3 g4f3 g2f3 b8d7 h3h4 b7b5 e1c1 d7b6 g5f6 g7f6 f3f4 h7h5 c1b1 b5b4 c3e2 a7a5 e2g3 d8c7 f4f5 g6f5 g3f5 e7e6 f5e3 f6e7 f2f4 e8c8 c2c3 d6d5');
INSERT INTO evaluations VALUES(55300,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55300,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(55300,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(55300,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(55300,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55300,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(55300,36,'cp 39','e7e5 g1f3 b8c6 f1g2 g8f6 b2b4 f8d6 b4b5 c6d4 d1a4 d4f3 g2f3 e8g8 e1g1 a7a6 c1a3 h7h6 f1d1 a8b8 a3d6 d8d6 b5a6 c8d7 a4c2 b7a6 d2d3 d6d4 e2e3 d4c5 d1c1 d7f5 c3e4 c5c2 c1c2 f8d8 e4c5 f5d3 c5d3 d8d3 c2c7 d3d7 c7c5 b8b5 c5b5 a6b5');
INSERT INTO evaluations VALUES(56628,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(56628,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(56628,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(56628,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(56628,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(56628,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(56628,36,'cp 41','f1g2 g8f6 g1f3 b8c6 b2b4 f8d6 b4b5 c6d4 d1a4 d4f3 g2f3 f6d7 e1g1 d7c5 a4c2 e8g8 a2a4 f7f5 d2d4 e5d4 c3d5 a7a6 b5b6 c7b6 c1b2 d6e5 f1d1 d4d3 e2d3 e5b2 c2b2 b6b5 a4b5 c8e6 d5f4 e6b3 b2c3 b3d1 a1d1');
INSERT INTO evaluations VALUES(56629,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(56629,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(56629,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(56629,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(56629,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(56629,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(56629,36,'cp 33','g8f6 g1f3 b8c6 b2b4 f8d6 b4b5 c6d4 d1a4 d4f3 g2f3 f6d7 a4c2 e8g8 e1g1 d7c5 c1b2 a7a6 f3g2 a6b5 c3b5 c5a4 a1b1 a4b2 b1b2 c7c6 b5d6 d8d6 a2a4 f8d8 f1a1 a8a5 c2c3 a5c5 c3b3 c5a5 b3c2 a5c5 c2b1 c5a5 d2d3 h7h6 h2h4 c8g4 b1c2 d6d4 c2b3');
INSERT INTO evaluations VALUES(48945,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(48945,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(48945,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(48945,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(48945,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(48945,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(48945,36,'cp 39','g1f3 b8c6 b2b4 f8d6 b4b5 c6d4 d1a4 d4f3 g2f3 e8g8 e1g1 a7a6 d2d3 h7h6 c1a3 c8d7 a3d6 a6b5 d6c7 d8c7 a4b3 c7c5 f1c1 d7c6 a1b1 c5a3 c3b5 a3b3 b1b3 a8a2 b5d6 f8a8 d6b7 c6f3 e2f3 f6d5 b7c5 a2a1 c1a1');
INSERT INTO evaluations VALUES(37462,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(37462,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(37462,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(37462,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(37462,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(37462,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(37462,36,'cp 178','e4e5 f6e8 h2h4 d7d6 h4h5 d6e5 h5g6 h7g6 c1h6 g7h6 h1h6 g8g7 d1d2 e8f6 f3e5 b8c6 e5c6 b7c6 e1c1 f8h8 h6h8 d8h8 f1c4 h8h2 d1e1 e7e6 g2g3 c8d7 c3a4 h2h6 d2h6 g7h6 a4c5 a8d8 c4f1 d7c8 c2c3 f6d7 c5b3 c8b7 b3a5 b7a8 b2b4 d8h8 f1g2 h6g7 g2c6');
INSERT INTO evaluations VALUES(37353,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(37353,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(37353,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(37353,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(37353,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(37353,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(37353,36,'cp 124','f6e8 h2h4 h7h5 g2g4 h5g4 f3h2 g4g3 h2f3 d7d6 h4h5 g6h5 h1h5 d6e5 f1d3 e5d4 c3e4 b8c6 e4g3 e8f6 h5g5 d8d6 d3f5 c8e6 d1d2 f6e8 d2d3 c6b4 d3e4 d4d3 f5h7 g8h8 e1f1 f7f6 g3f5 f6g5 f5d6 e8d6 e4e6');
INSERT INTO evaluations VALUES(57782,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(57782,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(57782,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(57782,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(57782,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(57782,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(57782,36,'cp 152','h2h4 h7h5 c3d5 d7d6 c1g5 b8c6 f1b5 f7f6 e5f6 e7f6 g5e3 c8g4 b5d3 c6e7 d5f4 d6d5 d3g6 e7g6 f4g6 f8f7 d1d3 d8d7 g6f4 g7f8 f3h2 g4f5 d3b3 c7c6 f4h5 a7a5 h5g3 f5h7 e1g1 a5a4 b3c3 e8g7 a1e1 g7f5 c3d2 f5h4 e3h6 f8e7 f2f3');
INSERT INTO evaluations VALUES(9351,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9351,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9351,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(9351,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(9351,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(9351,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(9351,36,'cp 17','c2c3 a7a5 e1g1 h7h6 b1d2 c8e6 c4b5 e8g8 f1e1 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 a2a4 e7g6 b5c4 e6c4 e3c4 a7a6 d1c2 a6c6 c4e3 a8e8 b2b4 a5b4 c3b4 c6c2 e3c2 g6f4 a1a3 f4e6 b4b5 c7c6 c2b4 c6b5');
INSERT INTO evaluations VALUES(9343,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(9343,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(9343,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9343,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9343,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9343,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9343,36,'cp 0','a7a5 e1g1 h7h6 b1d2 c8e6 a2a4 e8g8 f1e1 d8b8 f3h4 b8d8');
INSERT INTO evaluations VALUES(17721,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(17721,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(17721,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17721,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(17721,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(17721,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(17721,36,'cp 95','c2c4 e7e6 a2a3 g8f6 b2b3 d5c4 b3c4 e6e5 d4d5 e5e4 f3d2 c6b8 b1c3 c8f5 f1e2 f5g6 d1c2 b8d7 c1b2 f8d6 c3e4 f6e4 d2e4 e8g8 e2d3 d7e5 b2e5 d6e5 a1d1 c7c6 e1g1 c6d5 c4d5 a8c8 c2b3 b7b5 d1c1 g6e4 d3e4 e5h2 g1h2 d8h4 h2g1 h4e4 f1d1 e4g6 c1c8 f8c8 d5d6 c8d8');
INSERT INTO evaluations VALUES(17706,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(17706,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(17706,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(17706,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(17706,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(17706,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(17706,36,'cp 66','e7e6 a2a3 g8f6 b2b3 d5c4 b3c4 e6e5 d4d5 e5e4 f3d2 c6b8 b1c3 c8f5 f1e2 b8d7 g2g4 f5g6 g4g5 f6g8 h2h4 h7h6 h4h5 h6g5 h1g1 g6h7 c1b2 g5g4 d1a4 f7f5 d2b3 g8f6 b3d4 c7c6 d5c6 d7c5 c6c7 d8d7 a4d7 e8d7 c3d5 f6d5 c4d5 a8c8 d4e6 c5d3 e2d3');
INSERT INTO evaluations VALUES(17757,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(17757,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(17757,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(17757,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(17757,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(17757,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(17757,36,'cp 91','a2a3 g8f6 b2b4 d5c4 f1c4 f8d6 c1b2 d8e7 f3e5 d6e5 d4e5 f6d7 d1c2 d7e5 c4e2 e8g8 e1g1 f7f6 b1c3 e5f7 c3a4 c8d7 a4c5 f7d6 c5d7 e7d7 f1d1 g8h8 a1c1 d7f7 e2f3 a6a5 b4a5 a8a5 f3c6 b7c6 c2c6');
INSERT INTO evaluations VALUES(21132,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(21132,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(21132,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(21132,33,'cp -119',NULL);
INSERT INTO evaluations VALUES(21132,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(21132,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(21132,36,'cp -117','d8d4 f1d3 c8f5 g1f3 d4d7 e4c5 d7d5 d3f5 d5f5 c1e3 e7e5 f3h4 f5c8 d1d3 f8c5 e3c5 g8e7 d3c4 e8g8 e1c1 f8e8 h1e1 e7f5 h4f3 c8e6 c4e6 e8e6 d1d7 b7b6 c5a3 a8c8 f3g5 e6f6 g5f3 f5d6 e1d1 d6c4 a3e7 f6e6 e7d8 e6d6 d1d6 c7d6 d8e7 e5e4');
INSERT INTO evaluations VALUES(1634,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(1634,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(1634,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(1634,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(1634,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(1634,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(1634,36,'cp -90','f1d3 c8f5 g1f3 d4d7 e4c5 d7d5 d3f5 d5f5 c1e3 e7e5 f3h4 f5c8 d1d3 f8c5 e3c5 g8e7 d3c4 e8g8 e1c1 f8e8 h1e1 e7f5 h4f5 c8f5 g2g4 f5f6 c5e3 a8d8 g4g5 f6f3 b2b3 a7a6 d1d8 e8d8 c1b2 f3d5 c4g4 d5d7 g4e4 g7g6 e4c4 d7d5 c4c3 d5e6 h2h4 d8d5 a2a4 a6a5');
INSERT INTO evaluations VALUES(50895,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(50895,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(50895,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(50895,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(50895,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(50895,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(50895,36,'cp 99','f1d3 g8f6 h2h3 c7c5 d4d5 f8e7 e1g1 e6d5 e4d5 b8d7 a2a4 d7b6 d3e2 e8g8 f1e1 f8e8 d1d3 e7f8 c1f4 c5c4 d3d1 a6a5 e2f1 e8e1 f3e1 c8f5 f4e3 a8c8 g2g4 f5g6 f1g2 b6d7 e1f3 d7e5 f3e5');
INSERT INTO evaluations VALUES(50896,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(50896,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(50896,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(50896,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(50896,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(50896,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(50896,36,'cp 102','g8f6 h2h3 f8e7 e1g1 c7c5 d4d5 e6d5 e4d5 b8d7 a2a4 d7b6 d3e2 e8g8 f1e1 f8e8 d1d3 e7f8 c1f4 c5c4 d3d1 a6a5 e2f1 e8e1 f3e1 c8f5 f4e3 a8c8 g2g4 f5h7 f1g2 b6d7 e1f3 f6e4 c3e4 h7e4 f3d4 e4g2');
INSERT INTO evaluations VALUES(51523,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(51523,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(51523,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(51523,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(51523,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(51523,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51523,36,'cp 31','c8b7 c2c4 g8f6 f1g2 c7c5 b1c3 g7g6 d2d4 c5d4 d1d4 d7d6 e1g1 f8g7 b2b3 b8d7 d4h4 e8g8 c1h6 a7a6 f1d1 f8e8 h6g7 g8g7 a1c1 a8c8 h4d4 d8c7 d4d2 h7h6 g2h3 b7a8 f3d4 c7a7 h3g2 a8g2 g1g2 a7b7 f2f3 e8d8');
INSERT INTO evaluations VALUES(30841,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(30841,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(30841,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(30841,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(30841,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(30841,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(30841,36,'cp 32','c2c4 c7c5 f1g2 g7g6 d2d4 c5d4 d1d4 g8f6 b1c3 d7d6 e1g1 b8d7 b2b3 f8g7 f3e1 a8b8 c1b2 e8g8 g2b7 b8b7 a1c1 d7c5 e1g2 f6e4 d4e3 d8d7 b3b4 e4c3 b2c3 g7c3 e3c3 c5a4 c3d3 a7a5 a2a3 a5b4 a3b4 b6b5 g2e3 b5c4 c1c4');
INSERT INTO evaluations VALUES(30786,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(30786,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(30786,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(30786,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(30786,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(30786,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(30786,36,'cp 27','c7c5 f1g2 g7g6 d2d4 c5d4 d1d4 g8f6 e1g1 f8g7 b1c3 d7d6 b2b3 b8d7 c1b2 e8g8 f3e1 a8b8 g2b7 b8b7 e1d3 d7c5 d4e3 d8d7 f1d1 f8c8 a1c1 a7a6 d3c5 b6c5 g1g2 d7e6 e3f3 c8b8 c3d5 f6g4 b2c3 g7c3 d5c3 g4f6 c3d5 f6d5 d1d5');
INSERT INTO evaluations VALUES(43910,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43910,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(43910,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(43910,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43910,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43910,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43910,36,'cp 0','e7e6 d5e6 f7e6 b1d2 b5c4 b3c4 b8c6 g2g3 a8b8 f1g2 d8c7 g1f3 f8e7 e1g1 e8g8 a1b1 b8b1 d2b1 c8a6 f3d2 a6b7 d2f3');
INSERT INTO evaluations VALUES(43898,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43898,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43898,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43898,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43898,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43898,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43898,36,'cp -7','d5e6 f7e6 b1d2 b5c4 b3c4 b8c6 g2g3 a8b8 f1g2 d8c7 g1f3 f8e7 e1g1 e8g8 a1b1 b8b1 d2b1 d7d5 c1f4 c7a5 d1b3 d5c4 b3c4 a5a6 b1d2 f6d5 e2e3 c6b4 a2a3 a6a3 f4g5 a3d3 g5e7 d3c4 d2c4 d5e7 c4d6 f8d8 d6c8 e7c8 g2h3');
INSERT INTO evaluations VALUES(49626,30,'cp -145',NULL);
INSERT INTO evaluations VALUES(49626,31,'cp -140',NULL);
INSERT INTO evaluations VALUES(49626,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(49626,33,'cp -168',NULL);
INSERT INTO evaluations VALUES(49626,34,'cp -140',NULL);
INSERT INTO evaluations VALUES(49626,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(49626,36,'cp -140','d7d5 d2d4 e5d4 e4e5 f6h5 g2g3 d4c3 b1c3 c7c6 f3f4 g7g6 c1e3 b8d7 g1f3 f8c5 e3c5 d7c5 b2b4 c5e6 d1d2 e8g8 f1g2 d8b6 a1b1 h5g7 b4b5 b6c5 c3a4 c5a3 a4c3 e6c5 e1g1 c8f5 f3d4 f5b1 c3b1 a3d3 b5c6 d3d2 b1d2 a8b8 f1c1 c5e6 d4e6 g7e6 g2d5 b7c6');
INSERT INTO evaluations VALUES(48828,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(48828,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(48828,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(48828,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(48828,34,'cp -145',NULL);
INSERT INTO evaluations VALUES(48828,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(48828,36,'cp -146','d2d4 e5d4 e4e5 f6h5 g2g3 d4c3 b1c3 c7c6 f3f4 g7g6 c1e3 h5g7 g1f3 c8e6 f1g2 b8d7 e1g1 g7f5 e3f2 f8c5 c3a4 c5f2 f1f2 h7h5 g2h3 d8a5 f3g5 d7c5 a4c5 a5c5 d1d2 c5d4 g5e6 d4d2 e6c7 e8d8 f2d2 d8c7 h3f5 g6f5 a2a3 h8e8');
INSERT INTO evaluations VALUES(48808,30,'cp -145',NULL);
INSERT INTO evaluations VALUES(48808,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(48808,32,'cp -157',NULL);
INSERT INTO evaluations VALUES(48808,33,'cp -148',NULL);
INSERT INTO evaluations VALUES(48808,34,'cp -160',NULL);
INSERT INTO evaluations VALUES(48808,35,'cp -161',NULL);
INSERT INTO evaluations VALUES(48808,36,'cp -150','e5d4 e4e5 f6h5 d1d4 b8d7 f3f4 c7c6 f1e2 f8c5 d4d3 c5b6 e2h5 d8h4 d3g3 h4h5 g3g7 h8f8 b2b4 a7a5 g1e2 h5f5 g7g3 a5b4 c3b4 f7f6 c1e3 f6e5 e3b6 a8a2 g3c3 a2a1 c3a1 d7b6 e1g1 d5d4 e2g3 f5c2 b1a3 c2d2 f4f5 e8d8 b4b5 d2e3 g1h1 c6c5 a3c2 e3c3 a1c3 d4c3');
INSERT INTO evaluations VALUES(31843,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(31843,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(31843,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(31843,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(31843,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(31843,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(31843,36,'cp -12','g8f6 g1f3 e7e6 f1d3 a7a6 d3f5 e6f5 c3e2 c7c6 b2b3 f8d6 e1g1 e8g8 a2a4 d8e7 c1b2 f8c8 a4a5 b8d7 c2c4 c8e8 d1c1 d6c7 c1c2 g7g6 e2f4 e7d8 b2c3 h7h5 h2h3 d8e7 c3b2 c7d6 a1d1 f6e4 f4d3 d7f6 f3e5 h5h4');
INSERT INTO evaluations VALUES(38421,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(38421,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(38421,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(38421,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(38421,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(38421,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(38421,36,'cp 34','g8f6 b1c3 e8g8 c1e3 a7a6 a2a4 b8c6 a4a5 e7e5 d4d5 c6e7 f1d3 c8d7 e1g1 f6e8 c3e2 f7f5 c2c4 f5e4 d3e4 e7f5 e3c1 e8f6 e4c2 d8e7 f1e1 a8e8 e2c3 e7f7 a1a3 h7h6 c1e3 f5e3 e1e3 d7f5 c3e4 f6e4 c2e4 f5e4 e3e4');
INSERT INTO evaluations VALUES(41388,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(41388,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(41388,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(41388,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(41388,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41388,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(41388,36,'cp 59','f1d3 e7e5 d4e5 d6e5 f3e5 f6e4 e5g6 h7g6 d3e4 d8d1 e1d1 e8g8 c2c3 f8e8 h1e1 b8a6 c1g5 a6c5 b1d2 e8e5 g5f4 e5e7 a2a4 c8e6 e4c2 a8e8 e1e3 e6d5 e3e7 e8e7 f4g5 e7e6 g5e3 b7b6 e3c5 b6c5 c2b3 e6d6 b3d5 d6d5');
INSERT INTO evaluations VALUES(41378,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(41378,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(41378,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(41378,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(41378,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(41378,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(41378,36,'cp 65','e7e5 d4e5 d6e5 f3e5 f6e4 e5g6 h7g6 d3e4 d8d1 e1d1 b8d7 b1c3 e8g8 h1e1 c7c6 a2a4 a7a5 c1f4 f8e8 d1d2 d7c5 e4f3 c8f5 e1e8 a8e8 f4e3 e8d8 d2c1 g7d4 e3d4 d8d4 b2b3 g8g7 a1b1 g6g5 c3e2 d4d8 b3b4 a5b4 b1b4 d8a8 e2c3 g7f6 b4c4 a8a5 h3h4 g5h4');
INSERT INTO evaluations VALUES(44244,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(44244,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(44244,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(44244,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(44244,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(44244,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(44244,36,'cp 170','e4e5 f6d5 c2c4 d5e7 d4d5 d7d6 e5d6 e7g6 h2h4 f8d6 h4h5 g6e5 h5h6 g7g6 b1c3 e8g8 f3g5 e6d5 c4d5 f8e8 f1e2 e5g4 e1g1 g4h6 g5e4 h6f5 c1g5 d6e7 g5f4 f5d6 e2b5 e8f8 b5d3 a7a6 d1f3 d6e4 c3e4 c8f5 d5d6 e7f6 a1d1 f6d4 d3c4 b8d7 f1e1');
INSERT INTO evaluations VALUES(44161,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(44161,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(44161,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(44161,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(44161,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(44161,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(44161,36,'cp 128','f6d5 c2c4 d5e7 d4d5 d7d6 e5d6 e7g6 f1d3 f8d6 e1g1 e8g8 f3g5 d6e7 d1h5 e7g5 c1g5 f7f6 g5e3 e6d5 c4d5 b8c6 e3c5 c6e5 h5d1 c8g4 f2f3 d8d5 c5f8 e5d3 d1b3 d5b3 a2b3 g4e6 f8d6 a8d8 d6g3 a7a6 b3b4 d3b4 b1c3 e6b3 f1b1 b4d3 h2h4 g8f7 h4h5 g6f4');
INSERT INTO evaluations VALUES(44141,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(44141,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(44141,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(44141,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(44141,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(44141,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(44141,36,'cp 154','c2c4 d5e7 d4d5 d7d6 e5d6 e7g6 h2h4 f8d6 h4h5 g6e5 h5h6 g7g6 b1c3 e8g8 f3g5 e6d5 c4d5 c8g4 f1e2 g4e2 d1e2 c5c4 e1g1 e5d3 c1d2 b8d7 b2b3 f8e8 c3e4 d6e7 b3c4 e7g5 d2g5 d8g5 e2d3 g5h6 a1e1 b7b6 d3g3 h6g7 g3c7 g7d4 c7d7 e8e4 e1e4 d4e4');
INSERT INTO evaluations VALUES(44138,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(44138,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(44138,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(44138,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(44138,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(44138,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(44138,36,'cp 138','d5e7 d4d5 d7d6 e5d6 e7g6 h2h4 f8d6 h4h5 g6e5 h5h6 g7g6 b1c3 e6d5 c4d5 e8g8 f3g5 c8g4 f1e2 g4e2 d1e2 c5c4 e1g1 e5d3 c1d2 b8d7 b2b3 f8e8 c3e4 d6e7 b3c4 e7g5 d2g5 d8g5 e2d3 g5h6 a1e1 b7b6 d3g3 h6g7 g3d6 f7f5 e4g5 d7c5 g5e6 c5e6 e1e6 a8d8 d6c6 e8e6 d5e6');
INSERT INTO evaluations VALUES(46713,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(46713,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(46713,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(46713,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(46713,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(46713,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(46713,36,'cp 60','f1d3 f8g7 e1g1 b8c6 f1e1 e7e5 c2c3 e8g8 h2h3 a7a5 d3c2 h7h6 c1e3 a5a4 b1d2 c8d7 a1c1 b7b5 b2b4 a4b3 a2b3 e5d4 f3d4 a8b8 b3b4 b8a8 d1e2 c6d4 c3d4 c7c6 c2b3 f8e8 e2f3 d7e6 b3e6');
INSERT INTO evaluations VALUES(54061,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(54061,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(54061,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(54061,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(54061,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(54061,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54061,36,'cp 21','d5d4 c3b5 b8c6 g1f3 e6e5 b2b4 a7a6 b5a3 g8f6 a3c4 f6e4 c4e5 c6b4 f1c4 c8e6 c4e6 f7e6 a2a3 b4d5 c5c6 b7c6 d1d4 d8a5 f3d2 d5f6 e5c6 f8c5 d4e4 a5d2 c1d2 f6e4 f2f3 a8c8 c6b4 e4d2 e1d2 a6a5 b4d3 e8g8 a1a2 c5d4 h1b1 c8c4 b1b5 d4c3 d2e2 f8c8 f3f4 c3f6 b5a5 c4c2 a2c2');
INSERT INTO evaluations VALUES(57681,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(57681,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(57681,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(57681,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(57681,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(57681,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(57681,36,'cp 21','c1b2 g8f6 g1f3 f8g7 g2g3 b7b6 f1g2 e8g8 e1g1 e7e6 c2c4 c8b7 d2d4 d8e7 b1c3 d7d5 c4d5 e6d5 e2e3 b8d7 a1c1 c7c6 f3e1 a8c8 e1d3 f8d8 f1e1 f6e4 c1c2 d7f6 h2h4 e4c3 c2c3 f6e4 c3c2 b7a6 d3e5');
INSERT INTO evaluations VALUES(57682,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57682,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(57682,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(57682,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(57682,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(57682,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(57682,36,'cp 15','g8f6 g1f3 f8g7 g2g3 e8g8 f1g2 b7b6 c2c4 c8b7 d2d4 e7e6 e1g1 d8e7 b1c3 d7d5 c4d5 e6d5 e2e3 b8d7 a1c1 f8d8 c1c2 c7c6 f3e1 f6e4 e1d3 d8e8 d1c1 a8c8 b2a3 c6c5 d3f4 d7f6 d4c5 b6c5 f1d1 e4c3');
INSERT INTO evaluations VALUES(18755,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(18755,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(18755,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(18755,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(18755,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(18755,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(18755,36,'cp -44','b1c3 g8f6 c4a2 d7d5 e4d5 e6d5 d2d3 c8e6 g1e2 h7h6 e2f4 d8d7 f4e6 f7e6 c3e2 f8d6 e1g1 d7c7 e2g3 e8c8 c2c3 c6e5 b2b4 c5c4 d3c4 d5c4 d1e2 c8b8 g3e4 e5d3 e4d6 d8d6 a2b1 d3c1 f1c1 h8d8 b1e4');
INSERT INTO evaluations VALUES(22050,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(22050,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(22050,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(22050,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(22050,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(22050,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(22050,36,'cp 54','b1c3 f7f5 e4f5 c8f5 g1f3 g8f6 f1c4 d8e7 c1e3 f6g4 c3d5 g4e3 f2e3 e7e4 c4b3 e8c8 e1g1 e4g4 d1d2 d8e8 f3d4 c6d4 d2a5 d4e2 g1f2 c8b8 a5c7 b8a8 c7c4 g4c4 b3c4 a8b8 f2e2 f5c2 f1f2 c2e4 a1f1 a7a6 b2b4 f8e7 f2f7 e4g2 d5e7 g2f1 e2f1 h8f8');
INSERT INTO evaluations VALUES(22043,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(22043,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(22043,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(22043,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(22043,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(22043,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(22043,36,'cp 43','f7f5 e4f5 c8f5 g1f3 g8f6 c1g5 f8e7 f1c4 d8d7 e1g1 e8c8 f1e1 h7h6 g5f4 a7a6 f3d4 c6e5 d1e2 e5c4 e2c4 h8e8 b2b4 d6d5 c4b3 f6e4 b4b5 e4c5 b3b2 e7f6 b5a6 e8e1 a1e1 c5a6 c3b5 c8b8 f4g3 d8e8 e1b1 b7b6');
INSERT INTO evaluations VALUES(46743,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(46743,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(46743,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(46743,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46743,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(46743,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46743,36,'cp 11','f6e4 g5f4 e4c3 b2c3 c7c5 e2e3 f8g7 c4d5 d8d5 g1f3 e8g8 f1e2 c5d4 c3d4 d5a5 d1d2 a5d2 e1d2 b8c6 h1c1 f8d8 c1c5 c8e6 e2d3 e6d5 a1b1 e7e5 f3e5 c6e5 d4e5 d5g2 d2e2 g2c6 h2h4 g7f8 c5c3 f8e7 d3b5 c6b5 b1b5 b7b6');
INSERT INTO evaluations VALUES(46849,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46849,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(46849,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(46849,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46849,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(46849,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(46849,36,'cp 25','h2h4 c7c5 c4d5 e4c3 b2c3 d8d5 e2e3 h7h6 g5f4 f8g7 g1f3 c5d4 c3d4 e8g8 d1a4 c8e6 f1e2 b8c6 e1g1 d5a5 a4a5 c6a5 f4c7 a5c4 a1b1 c4b6 c7b6 a7b6 f1c1 e6a2 b1b6 a2d5 c1c5 e7e6 e2c4 d5e4 f3e5 f8b8 f2f3 e4c6 c4b3 b8c8');
INSERT INTO evaluations VALUES(51888,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(51888,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(51888,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(51888,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(51888,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(51888,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(51888,36,'cp -14','a7a6 b5e2 b8c6 d2d4 c5d4 f3d4 g8f6 b1c3 d8c7 d4c6 b7c6 d1d3 d7d5 d3g3 c7g3 h2g3 e6e5 e4d5 c6d5 f2f4 f8d6 f4e5 d6e5 c1f4 e5c3 b2c3 e8g8 e1c1 c8f5 e2f3 a8c8 f4e5 f5e4 h1e1 f8e8 e5d4 h7h6 a2a4 c8c6 g3g4');
INSERT INTO evaluations VALUES(50472,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(50472,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50472,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50472,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50472,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(50472,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(50472,36,'cp 0','b5e2 b8c6 d2d4 c5d4 f3d4 g8f6 b1c3 d8c7 d4c6 b7c6 d1d3 d7d5 d3g3 c7g3 h2g3 a8b8 b2b3 f8b4 c1d2 d5d4 c3d1 f6e4 d2b4 b8b4 e2f3 e4c5 a2a3 b4b8 b3b4 c5d7 f3c6 e8e7 h1h4 d7e5 c6f3 e5f3 g2f3 c8b7 d1b2 b7f3 h4f4 f3c6 f4d4 g7g5 b2c4 h7h5 c4e5 b8c8 e1c1');
INSERT INTO evaluations VALUES(50531,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(50531,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(50531,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(50531,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(50531,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(50531,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(50531,36,'cp -9','b8c6 d2d4 c5d4 f3d4 g8f6 b1c3 d8c7 d4c6 b7c6 d1d3 d7d5 d3g3 c7g3 h2g3 a8b8 b2b3 f8b4 c1d2 d5d4 c3d1 f6e4 d2b4 b8b4 e2f3 e4c5 d1b2 e6e5 e1c1 e8e7 a2a3 b4b7 h1e1 f7f6 d1d4 b7b3 d4c4 b3b5 f3c6 b5a5 b2a4 c5a4 c6a4 c8e6 c4c7 e7d6 c7c6 d6e7');
INSERT INTO evaluations VALUES(17393,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17393,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17393,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17393,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17393,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17393,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17393,36,'cp -14','d2d4 c5d4 f3d4 g8f6 b1c3 d8c7 d4c6 b7c6 d1d3 a8b8 b2b3 d7d5 d3g3 c7g3 h2g3 f8b4 c1d2 d5d4 c3d1 f6e4 d2b4 b8b4 e2f3 e4c5 a2a3 b4b8 b3b4 c5d7 f3c6 e8e7 d1b2 a6a5 b2d3 e7d6 c6d7 c8d7 e1d2 e6e5 c2c3 d4c3 d2c3 h8c8 c3d2 d7f5 b4a5');
INSERT INTO evaluations VALUES(52927,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(52927,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(52927,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(52927,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(52927,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(52927,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(52927,36,'cp 37','d3e4 d8d1 e1d1 g8f6 b1d2 g7g6 a2a4 f8h6 f1d3 c8g4 d1e2 b8d7 h2h3 g4f3 e2f3 a7a5 h1d1 d7e5 f3e2 f6d7 b2b3 h6g7 a1b1 d7c5 c1b2 g7f6 f2f4 e5d3 c2d3 f6b2 b1b2 h8g8 e2e3 f7f5 d2c4 e8c8 b3b4 c5a4');
INSERT INTO evaluations VALUES(52919,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52919,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(52919,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(52919,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(52919,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52919,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52919,36,'cp 17','d8d1 e1d1 g8f6 b1d2 g7g6 a2a4 c8g4 f1d3 f8h6 d1e2 b8d7 h2h3 g4f3 e2f3 e8c8 b2b4 c6c5 d2c4 h6c1 h1c1 c5b4 a1b1 d7b8 b1b4 b8c6 b4b2 d8d4 c1b1 h8d8 f3e3 f6e4 d3e4 d4c4 b2b7 c4c3 e4d3 d8d6 b7b3 c3b3 b1b3 c8c7 f2f4 d6e6 e3d2 e6d6 d2c3 d6e6 b3b5 e6e1 c3d2 e1h1');
INSERT INTO evaluations VALUES(34308,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(34308,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(34308,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(34308,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34308,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(34308,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(34308,36,'cp 29','e1d1 g8f6 b1d2 g7g6 a2a4 c8g4 f1e2 f8h6 a4a5 e8g8 e4e5 f6d5 h2h3 g4f5 f3d4 h6f4 d4f5 g6f5 e2d3 e7e6 g2g4 b8d7 d2f3 d7e5 f3e5 f4e5 g4f5 e6f5 d3f5 a8d8 d1e2 e5f4 a5a6 b7b6 e2f3 f4c1 h1c1 d5e7');
INSERT INTO evaluations VALUES(34307,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(34307,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(34307,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(34307,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(34307,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(34307,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(34307,36,'cp 28','g8f6 b1d2 g7g6 a2a4 f8h6 a4a5 c8g4 f1e2 e8g8 c2c3 c6c5 h1e1 b8c6 h2h3 g4d7 e2c4 a8d8 d1c2 b7b6 a5b6 a7b6 d2b3 h6c1 c2c1 d8a8 a1a8 f8a8 e4e5 f6h5 e5e6 d7e8 e6f7 e8f7 c4f7 g8f7 c1c2 h5f4 f3g5 f7e8');
INSERT INTO evaluations VALUES(33579,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(33579,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(33579,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(33579,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(33579,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(33579,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(33579,36,'cp 20','b1d2 g7g6 a2a4 f8h6 a4a5 c8g4 e4e5 f6d7 f1e2 h6g7 h1e1 e8g8 h2h3 g4e6 e2c4 e6d5 c2c3 b8a6 c4d5 c6d5 d2b3 e7e6 c1f4 f8c8 f3d4 a6b8 d1c2 b8c6 d4c6 b7c6 a5a6 a8b8 b3a5 c6c5 a5b7 c8c6 f2f3');
INSERT INTO evaluations VALUES(11434,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11434,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(11434,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(11434,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(11434,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11434,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(11434,36,'cp 19','b7b5 a4b3 f8e7 c2c3 e8g8 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 d4d5 c6b8 a2a4 b8d7 c3c4 c7c6 d5c6 b7c6 a4b5 a6b5 a1a8 c6a8 c4b5 d7c5 b3c2 d8b8 b2b4 c5d7 c2d3 d6d5 d1a4 h7h6 c1b2 d7b6 a4a5 d5e4 d2e4 a8e4');
INSERT INTO evaluations VALUES(11653,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(11653,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(11653,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(11653,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(11653,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(11653,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(11653,36,'cp 25','a4b3 f8e7 c2c3 e8g8 h2h3 c8b7 d2d4 f8e8 a2a3 e7f8 b1d2 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 c2d3 f6h5 d3f1 h5f4 d1c2 a8c8 a1d1 d8c7 c3c4 f4e6 b2c3 e5d4 f3d4 e6d4 c3d4 b5c4 c2c4 d6d5 e4d5 e8e1 d1e1 f8a3 d5c6');
INSERT INTO evaluations VALUES(11636,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11636,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(11636,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(11636,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(11636,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(11636,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(11636,36,'cp 26','f8e7 c2c3 e8g8 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 b3c2 g7g6 a2a3 c6b8 a3a4 b8d7 c2d3 c7c6 b2b3 d8c7 d1c2 f8g7 c1b2 f6h5 d3f1 h5f4 b3b4 d7b6 a4b5 c6b5 d4e5 b6a4 c3c4 a4b2 c2b2 d6e5 c4b5 a6b5 f1b5');
INSERT INTO evaluations VALUES(53052,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53052,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53052,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53052,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53052,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(53052,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(53052,36,'cp -16','d2d4 g7g6 c2c3 f8g7 c1b2 g8f6 b1d2 c5d4 c3d4 b8c6 g1f3 d6d5 e4e5 f6e4 f1b5 c8d7 a2a4 e8g8 e1g1 g7h6 b5c6 d7c6 f1e1 a8c8 d2e4 d5e4 f3d2 h6d2 d1d2 f8e8 d2e3 c6d5 a1c1 d8b6 c1c8 e8c8 e1c1 c8c1 b2c1 b6b3 e3b3 d5b3');
INSERT INTO evaluations VALUES(53280,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(53280,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(53280,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(53280,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(53280,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(53280,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(53280,36,'cp 121','d4c5 g8f6 e4d5 d8d5 d1d5 f6d5 f1c4 c8e6 g1f3 b8d7 f3d4 g7g6 d4e6 f7e6 c5c6 b7c6 b1d2 a7a5 a2a3 f8g7 e1g1 e8f7 f1e1 h8b8 a1b1 a5a4 h2h4 g7h6 d2e4 h6c1 b1c1 e6e5 e4g5 f7g7 c1b1 a8a7 c4a2 b8d8');
INSERT INTO evaluations VALUES(53165,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(53165,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(53165,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(53165,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(53165,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(53165,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(53165,36,'cp 117','g8f6 e4d5 d8d5 d1d5 f6d5 f1c4 c8e6 g1f3 b8d7 f3g5 d7c5 g5e6 f7e6 b2b3 g7g6 c1b2 h8g8 b1d2 a7a6 e1g1 f8g7 b2g7 g8g7 a2a3 b7b5 c4e2 e8f7 b3b4 d5c3 f1e1 a8d8 d2f3 c5a4 e2d3 f7f6 e1e3 c3d5 e3e5 d5f4 a1e1 a4b2 d3f1 b2c4 f1c4 b5c4');
INSERT INTO evaluations VALUES(54819,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(54819,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(54819,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(54819,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(54819,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(54819,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(54819,36,'cp 49','b1c3 f8g7 e2e4 b8c6 f1b5 c8d7 g1f3 g8f6 d1d2 a7a6 b5c6 d7c6 d4d5 c6d7 f4h6 g7h6 d2h6 e7e5 d5e6 d7e6 h2h3 d8e7 e1c1 e8c8 h1e1 f6d7 f3d4 d7b6 f2f3 h8e8 f3f4 e7f6 b2b3 c8b8 a2a4 e6d7 a4a5 b6c8 g2g4 d7c6 f4f5 e8e5 d4f3');
INSERT INTO evaluations VALUES(15039,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(15039,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(15039,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(15039,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(15039,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(15039,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(15039,36,'cp 75','c1g5 f8e7 e2e3 f6h5 g5e7 c6e7 h2h3 h5f6 f1d3 e8g8 g2g4 c8e6 d1c2 g7g6 e1c1 b7b6 c2d2 c7c5 d3c2 e7c6 d4c5 b6c5 g4g5 f6d7 c3d5 a8b8 h1g1 e6d5 d2d5 c6b4 d5b3 d8a5 c1b1 b4c2 b3c2');
INSERT INTO evaluations VALUES(15018,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(15018,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(15018,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(15018,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(15018,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(15018,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(15018,36,'cp 47','f8e7 e2e3 f6h5 g5e7 c6e7 b2b4 h5f6 f1d3 c8f5 f3e5 c7c6 a2a4 e8g8 e1g1 f8e8 h2h3 a8c8 d1f3 f5g6 f1c1 g6h5 f3g3 h5g6 b4b5 c6c5 d4c5 c8c5 e5g6 e7g6 c3e2 c5c1 a1c1 g6e5 c1d1 d8d6 g3g5 e5d3 d1d3');
INSERT INTO evaluations VALUES(8514,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(8514,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(8514,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(8514,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(8514,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(8514,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(8514,36,'cp 62','e2e3 f6h5 g5e7 c6e7 d1c2 e8g8 f1d3 h5f6 f3e5 g7g6 h2h3 c8e6 b2b4 c7c6 e1g1 f6d7 e5f3 a8c8 c2b2 e7f5 b4b5 f5h4 f3h4 d8h4 b5c6 b7c6 c3e2 c6c5 d4c5 d7c5 d3c2 c8b8 c2b3 f8c8 a1c1 h4b4 e2d4 e6d7 f1d1 c5e4 c1c8 b8c8');
INSERT INTO evaluations VALUES(8513,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(8513,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(8513,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(8513,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(8513,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(8513,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(8513,36,'cp 39','f6h5 g5e7 c6e7 f1d3 h5f6 e1g1 e8g8 h2h3 c8f5 b2b4 a7a6 a2a4 f5d3 d1d3 c7c6 b4b5 d8d6 f1b1 a6b5 a4b5 c6c5 d4c5 d6c5 d3d4 c5d6 c3e2 f6e4 e2f4 g7g5 f4e2 e7f5 d4d1 h7h6 a1a8 f8a8 d1c2 d6a3 f3d4 f5d6 g2g4');
INSERT INTO evaluations VALUES(28949,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(28949,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(28949,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(28949,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(28949,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(28949,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(28949,36,'cp 43','d1b3 b8d7 g1f3 d8b6 c4c5 b6b3 a2b3 h7h6 h2h4 g7g6 b3b4 f8g7 e2e3 e8g8 b4b5 f6g4 f1e2 e7e5 f4g3 f8e8 e1g1 e5d4 e3d4 d7c5 b5c6 b7c6 d4c5 d5d4 f3d4 g7d4 e2g4 f5g4 a1a4 d4c3 b2c3 g4f5 f1a1 a7a5 a4a5 a8a5 a1a5 e8e1 g1h2 e1e2 a5a7');
INSERT INTO evaluations VALUES(10983,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(10983,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(10983,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(10983,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(10983,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(10983,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(10983,36,'cp 79','e4e5 c7c5 d4h4 f6d5 g5e7 d8e7 h4e7 e8e7 b1a3 b7b6 e1c1 c8b7 h1g1 h7h5 h2h4 d5c7 a3c4 b7d5 f2f4 d5e6 g2g3 g7g6 c4d6 c7e8 d6b5 e8g7 f1g2 a8b8 a2a4 g7f5 g2e4 f7f6 e5f6 e7f6 b2b3 b8e8 e4f5 e6f5 b5a7 c5c4 c1b2 e8e2');
INSERT INTO evaluations VALUES(10970,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(10970,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(10970,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(10970,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(10970,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(10970,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(10970,36,'cp 83','c7c5 d4h4 f6d5 g5e7 d8e7 h4e7 e8e7 b1a3 b7b6 e1c1 c8b7 h1g1 h7h5 h2h4 d5c7 a3c4 b7d5 c4d6 d5e6 g2g3 f7f6 e5f6 g7f6 f1g2 a8f8 g1e1 f6f5 d6c4 f5f4 c4e5 f8g8 e5d7 f4g3 f2g3 g8g3 g2c6 h8d8 e1e5 g3g4');
INSERT INTO evaluations VALUES(7953,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(7953,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(7953,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(7953,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(7953,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(7953,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(7953,36,'cp 92','d4h4 f6d5 g5e7 d8e7 h4e7 e8e7 b1a3 b7b6 e1c1 c8b7 h1g1 h7h5 h2h4 d5c7 a3c4 b7d5 c4d6 d5e6 f2f4 g7g6 g2g3 c7e8 d6b5 e8g7 f1g2 a8b8 g2e4 g7f5 b2b3 f7f6 e5f6 e7f6 c1b2 h8e8 a2a4 b8d8 b5c7 e8f8 e4f5 f6f5 c7b5 d8a8 d1d3 f5g4 b5c7');
INSERT INTO evaluations VALUES(28799,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(28799,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(28799,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(28799,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(28799,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(28799,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(28799,36,'cp 42','d1c2 d7d5 c4d5 f6d5 g1f3 e7e6 e2e4 d5c3 b2c3 f8e7 c1f4 e8g8 f1d3 c7c5 d4d5 e6d5 e4d5 d8d5 e1c1 g8h8 h1e1 c5c4 e1e7 c4d3 c2d3 d5d3 d1d3 b7a6 d3d6 h8g8 c1b2 a6c4 f3e5 c4e6 f4e3 f8c8 a2a4 b8a6 e7b7 c8c7 b7c7 a6c7 e5c6 g8f8 b2a3 f7f6 f2f3 e6d5');
INSERT INTO evaluations VALUES(28800,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(28800,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(28800,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(28800,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(28800,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(28800,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(28800,36,'cp 40','d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 e7e6 g1f3 f8e7 c1f4 e8g8 f1d3 c7c5 d4d5 f7f5 c2e2 c5c4 d3c2 b8a6 a1d1 a6c5 e4f5 b7d5 e1g1 c5d3 f4g3 e6f5 f3e5 d8e8 e5d3 c4d3 c2d3 g8h8 d3b5 e8f7 g3e5 a7a6 b5d3 d5a2 d3a6 e7c5 a6b5 a2b3');
INSERT INTO evaluations VALUES(44241,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(44241,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44241,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44241,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44241,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44241,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(44241,36,'cp 33','c5d4 f3d4 b8c6 d4c6 b7c6 e4e5 f6d5 c3e4 d8c7 f2f4 c7a5 c1d2 a5b6 f1d3 c8a6 d3a6 b6a6 d1e2 a6e2 e1e2 f7f5 e4f2 a7a5 a2a4 c6c5 b2b3 d7d6 c2c4 d5b4 d2b4 c5b4 f2d3 e8c8 a1d1 f8e7 h2h4 c8c7 h4h5 d6e5 d3e5 h8g8 e5f3 e7f6 d1d8 g8d8 h5h6 g7g6 g2g3');
INSERT INTO evaluations VALUES(44304,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(44304,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(44304,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(44304,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(44304,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(44304,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(44304,36,'cp 37','f3d4 b8c6 d4c6 b7c6 e4e5 f6d5 c3e4 c8b7 f1e2 c6c5 e1g1 d8c7 e4d6 f8d6 e5d6 c7c6 f2f3 c5c4 d1d4 e8g8 e2c4 f8c8 b2b3 c6d6 c1e3 a7a5 f1d1 a5a4 c4d5 d6d5 d4f4 e6e5 f4g4 d5e6 g4e6 d7e6 c2c4 f7f6 a1c1 a4b3 a2b3 h7h6 b3b4 a8a4 e3c5 b7c6');
INSERT INTO evaluations VALUES(54948,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(54948,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(54948,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(54948,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54948,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54948,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(54948,36,'cp 5','g8f6 c2c4 g7g6 b1c3 f8g7 e2e4 e8g8 f1d3 b8d7 g1f3 d7c5 d3c2 e7e6 e1g1 e6d5 e4d5 f8e8 c1g5 h7h6 g5e3 f6e4 c3e4 c5e4 f3d4 e4f6 h2h3 f6h5 c2d3 c8d7 d3e2 d8h4 e2h5 g6h5');
INSERT INTO evaluations VALUES(52937,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(52937,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52937,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(52937,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(52937,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52937,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52937,36,'cp 0','e7e5 d4e5 d8d1 e1d1 c8e6 e2e4 f8c5 f2f4 g7g6 c1d2 b8a6 a4a5 e8c8 g1f3 g8e7 d1e1 c8c7 f3g5 d8d4 d2e3 d4d7 e1f2 d7d2 f2f3 c5e3 g5e6 f7e6 f3e3 h8d8 f1c4 a6b4 a1d1 b4c2 e3f3 c2d4 f3e3');
INSERT INTO evaluations VALUES(53793,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(53793,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(53793,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53793,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53793,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(53793,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(53793,36,'cp 12','d4e5 d8d1 e1d1 b8d7 f2f4 f7f6 g1f3 f6e5 e2e3 e5f4 e3f4 d7c5 f1c4 c8g4 c1e3 e8c8 d1e2 g8h6 h2h3 g4e6 c4e6 c5e6 g2g4 f8d6 f4f5 e6f4 e3f4 d6f4 a1d1 d8e8 e2f2 g7g6 f5g6 h7g6 f2g2 h6f7 h1e1 h8f8 e1e4 e8e4 c3e4 f8e8 d1e1 b7b6 b2b3 c8c7 h3h4');
INSERT INTO evaluations VALUES(17894,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(17894,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(17894,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(17894,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(17894,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(17894,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(17894,36,'cp 13','f8c5 e1g1 d7d6 b1c3 g8f6 a2a4 c6d4 f3d4 c5d4 c3e2 d4a7 e2g3 e8g8 h2h3 c7c6 d1f3 c8e6 a4a5 h7h6 g3h5 f6h5 c4e6 f7e6 f3h5 d8e7 c1e3 f8f6 h5e2 a7e3 f2e3 a8f8 f1f6 f8f6 a1f1 f6f1 g1f1 e7h4 c2c3 h4g3 e2f2 g3f2 f1f2 c6c5 f2e2 h6h5');
INSERT INTO evaluations VALUES(4863,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4863,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4863,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4863,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4863,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(4863,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4863,36,'cp 18','b1c3 g8f6 c3d5 f6d5 c4d5 e8g8 c2c3 d7d6 h2h3 c5a7 a2a4 d8f6 e1g1 c8e6 d5e6 f6e6 c1e3 a7e3 f2e3 c6e7 d1c2 h7h6 d3d4 f7f5 d4e5 d6e5 a1d1 a8e8 f1e1 f5f4 g1h2 f4e3 e1e3 e7g6 b2b4 e6c4 a4a5');
INSERT INTO evaluations VALUES(4858,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(4858,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4858,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4858,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4858,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(4858,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(4858,36,'cp 13','g8f6 c3d5 f6d5 c4d5 d7d6 e1g1 e8g8 c2c3 d8f6 a2a4 h7h6 c1e3 c6e7 d5b3 c5e3 f2e3 e7g6 a4a5 f6e7 f1f2 g6h4 f3h4 e7h4 h2h3 c8e6 d1c2 a8d8 a1f1 h4e7 b3e6 f7e6 b2b4 f8f2 c2f2 d8f8 f2f8 e7f8 f1f8 g8f8 g1f2 h6h5 c3c4 f8e7 f2e1');
INSERT INTO evaluations VALUES(31906,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(31906,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(31906,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31906,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(31906,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(31906,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(31906,36,'cp 31','b8d7 b1c3 c7c6 e2e4 e7e6 h2h3 g4h5 f1d3 d5e4 c3e4 g8f6 e4g3 h5f3 d1f3 f6d5 f4d2 d5b4 d2b4 f8b4 c2c3 b4e7 e1c1 d7f6 c1b1 d8c7 h3h4 e8c8 h1e1 c8b8 g3e4 f6e4 e1e4 e7f6 d3c4 b8a8 a2a3 h8e8 g2g3 a7a6 f3e2');
INSERT INTO evaluations VALUES(32128,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(32128,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(32128,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(32128,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(32128,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(32128,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(32128,36,'cp 88','g2f3 g8f6 c1e3 g7g6 d4e5 d6e5 d1d8 e8d8 b1d2 h7h5 e3g5 f8e7 e1c1 d8e8 d2c4 b8c6 c2c3 f6d7 g5d2 a7a6 c1c2 b7b5 c4e3 d7b6 h2h4 a8d8 b2b3 e8f8 a2a4 b5a4 b3a4 b6a4 e2a6 a4c5 a6c4 f8g7 e3d5 c6a5 c4a2');
INSERT INTO evaluations VALUES(41452,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(41452,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(41452,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(41452,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(41452,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(41452,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(41452,36,'cp 49','c4d5 e8g8 f1e2 f6d5 e1g1 b8c6 e3e4 d5b6 c1e3 f7f5 e4f5 g6f5 g2g3 f5f4 e3f4 c6d4 f3d4 e7e5 d1b3 g8h8 d4c6 b7c6 f4e3 d8d6 b1c3 c8e6 c3e4 e6b3 e4d6 b3d5 d6b7 b6a4 b7a5 a8b8 a1c1 b8b2 e2d1 a4b6 a5c6 d5a2 e3c5 f8a8 d1f3 e5e4 f3e4 a2c4 f1e1 a7a5');
INSERT INTO evaluations VALUES(41428,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(41428,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(41428,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(41428,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(41428,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41428,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(41428,36,'cp 36','e8g8 f1e2 f6d5 e3e4 d5b6 e1g1 c8g4 b1d2 e7e5 d4e5 b8c6 h2h3 g4e6 d2b3 c6e5 f3e5 d8d1 f1d1 g7e5 f2f4 e5f6 b3c5 a8d8 c5e6 f7e6 g2g3 d8d1 e2d1 f8d8 d1b3 g8f7 e4e5 f6e7 c1e3 d8d3 g1f2 b6d5 b3d5 d3d5 a1c1 c7c5 g3g4 h7h6 f2e2 b7b6 c1d1');
INSERT INTO evaluations VALUES(49885,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(49885,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(49885,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(49885,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(49885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49885,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(49885,36,'cp 12','d1h5 e4d6 h5e5 d8e7 e5e7 f8e7 c4e2 d6f5 g1f3 b8c6 c3d5 e7d8 c2c3 c6e7 d5e7 f5e7 a2a4 a7a5 e1g1 d7d6 f1e1 e8g8 b2b3 c8d7 g2g3 e7d5 f3d4 d8f6 e2f3 d5b4 c1a3 f6d4 a3b4 d4f6 b4a3 a8b8 g1f1 b7b6 d2d4 f8d8 h2h4 g7g6');
INSERT INTO evaluations VALUES(49879,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(49879,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(49879,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(49879,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49879,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(49879,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(49879,36,'cp 4','e4d6 h5e5 d8e7 e5e7 f8e7 c4e2 d6f5 g1f3 b8c6 c3d5 e7d8 c2c3 c6e7 d5e7 f5e7 e1g1 a7a5 a2a4 e8g8 f1e1 d7d6 f3d4 c7c6 d2d3 f8e8 c1f4 d8c7 g1f1 c8d7 h2h3 h7h6 f4h2 a8d8 d4c2 e7d5 e2f3 e8e1 a1e1 d7e6 c2d4');
INSERT INTO evaluations VALUES(53230,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(53230,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(53230,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(53230,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(53230,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(53230,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(53230,36,'cp -5','g8f6 b1c3 f6d5 c3d5 d8d5 d1f3 d5e6 f3e3 b8c6 g1f3 g7g6 f1b5 f8g7 f3e5 g7e5 f4e5 e6d5 b5c6 d5c6 e1g1 c8e6 d2d3 h7h5 h2h3 h5h4 a2a3 h8h5 c1d2 b7b6 g1h2 e8c8 a1e1 d8d5 d2c3 g6g5 e3f2 h5h8 h2h1 h8d8');
INSERT INTO evaluations VALUES(45328,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(45328,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(45328,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(45328,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(45328,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(45328,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(45328,36,'cp -1','b1c3 f6d5 c3d5 d8d5 d1f3 d5e6 f3e3 g7g6 g1f3 f8g7 f1b5 b8c6 f3e5 g7e5 f4e5 a7a6 b5c6 e6c6 e1g1 c8e6 d2d3 e8c8 e3f2 d8d5 f1e1 h8d8 c1g5 d8d7 b2b3 b7b6 h2h3 c8b7 a1b1 c6c7 g5f4 c7c6');
INSERT INTO evaluations VALUES(20208,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(20208,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(20208,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(20208,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(20208,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(20208,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(20208,36,'cp -35','f8c5 b1c3 a7a6 f1d3 f7f6 d1e2 d8c7 c1f4 g8e7 g2g3 f6f5 c3b1 h7h6 h2h4 e8g8 b1d2 c8d7 a2a3 d5d4 h4h5 b7b5 f3h4 d7e8 e1g1 e7d5 h4g6 e8g6 h5g6 c6e7 e2h5 d5f4 g3f4 e7d5 h5f3 c7d8 f3g3 c5b6 d2f3 g8h8 g1h1 a8c8 f1g1');
INSERT INTO evaluations VALUES(19252,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(19252,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(19252,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(19252,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(19252,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(19252,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(19252,36,'cp 51','d2d4 c5d4 f3d4 d7d6 b1c3 c6d4 d1d4 g8f6 f1e2 f8g7 c1e3 a7a5 e1g1 e8g8 d4d1 c8d7 a2a4 d7c6 d1c2 f6d7 a1d1 d8c7 b2b3 d7c5 c3d5 c7d8 e2f3 e7e6 d5c3 d8e7 c3b5 g7e5 g2g3 a8c8 c2d2 c6e4 f3e4 c5e4');
INSERT INTO evaluations VALUES(19265,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(19265,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(19265,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(19265,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(19265,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(19265,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(19265,36,'cp 55','c5d4 f3d4 g8f6 b1c3 c6d4 d1d4 d7d6 f1e2 f8g7 e1g1 e8g8 d4d3 f6d7 d3c2 d7c5 c1e3 g7c3 c2c3 c5e4 c3d4 e4f6 e2f3 d8b6 f1e1 b6d4 e3d4 e7e5 d4c3 c8e6 a1d1 e6c4 d1d6 e5e4 f3e2 f6e8 d6d7 c4a2 e1a1 a2e6 d7b7 a7a5 a1a4 e8d6');
INSERT INTO evaluations VALUES(19312,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(19312,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(19312,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(19312,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(19312,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(19312,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(19312,36,'cp 68','f3d4 f8g7 c1e3 d8b6 d4b3 b6d8 d1d2 g8f6 b1c3 f6g4 e3f4 e8g8 h2h3 g4f6 f1e2 d7d6 e1g1 c8e6 f1d1 a7a5 f4h6 g7h6 d2h6 d8b6 a1b1 a5a4 b3d2 b6a5 b1c1 c6d4 h6e3 d4e2 e3e2 f8c8 b2b3 a4b3 a2b3 f6d7 e2e3');
INSERT INTO evaluations VALUES(39021,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(39021,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(39021,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(39021,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(39021,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(39021,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(39021,36,'cp 87','b1c3 g8e7 g1f3 e8g8 f1d3 d7d6 h2h4 e6e5 d4e5 d6e5 h4h5 c8g4 h5g6 e7g6 d3e2 d8d1 c3d1 b8c6 d1e3 g4f3 e2f3 c6d4 f3g4 g6f4 e1f1 f8d8 e3d5 f4d5 c4d5 d8d6 h1h3 c7c6 d5c6 d6c6 c1d2 c6g6 g4d1 g6c6');
INSERT INTO evaluations VALUES(42344,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42344,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(42344,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(42344,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42344,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(42344,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(42344,36,'cp 18','g5h4 b7b6 d1c2 c8b7 e2e3 b7e4 c2c1 c7c5 d4c5 b6c5 a2a3 b4c3 c1c3 b8c6 g1f3 e8g8 f3d2 e4h7 f1e2 a8b8 e2f3 d8b6 a1a2 f6e8 e1g1 f7f6 h4g3 e6e5 f1d1 d7d6 d2e4 h7e4 f3e4 f6f5 e4c2 f5f4 e3f4 e5f4 g3h4 c6d4 c3d3 f8f5 f2f3 e8c7 g1h1 c7e6 h4e1 b8e8');
INSERT INTO evaluations VALUES(42346,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(42346,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(42346,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(42346,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(42346,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42346,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42346,36,'cp 0','c7c5 e2e3 b8c6 a1c1 c5d4 e3d4 d7d5 g1f3 e8g8 a2a3 b4e7 c4c5 f6e4 h4e7 d8e7 f1e2 f8d8 b2b4 e6e5 f3e5 c6e5 d4e5 a7a5 e1g1 a5b4 a3b4 e4c3 c1c3 e7e5 c3e3 e5b2 e3b3 b2f6 b3d3 f6b2');
INSERT INTO evaluations VALUES(39860,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(39860,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39860,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39860,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39860,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39860,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39860,36,'cp 0','e2e3 c5d4 e3d4 b8c6 a1c1 e8g8 g1f3 d7d5 a2a3 b4e7 c4c5 f6e4 h4e7 d8e7 f1e2 f8d8 b2b4 e6e5 f3e5 c6e5 d4e5 a7a5 e1g1 a5b4 a3b4 e4c3 c1c3 e7e5 c3e3 e5f4 d1d2 c8e6 e3d3 f4e5 d3e3');
INSERT INTO evaluations VALUES(39859,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39859,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39859,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(39859,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(39859,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(39859,36,'cp 6','c5d4 e3d4 b8c6 a1c1 e8g8 g1f3 d7d5 a2a3 b4e7 c4c5 f6e4 h4e7 d8e7 f1e2 f8d8 b2b4 e6e5 f3e5 c6e5 d4e5 e4c3 c1c3 d5d4 c3g3 e7e5 e1g1 a7a5 e2c4 a5b4 a3b4 c8f5 d1b3 d8d7 h2h3 f5e6 g3d3 e6c4 b3c4 a8d8 f1d1 d7d5 c4b5 e5e7 d3d2 g8h8 b5d3 b7b6 c5c6 e7b4');
INSERT INTO evaluations VALUES(5740,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(5740,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(5740,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(5740,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(5740,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(5740,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(5740,36,'cp 27','c2c3 g8f6 b2b4 c5b6 a2a4 a7a6 b1d2 e8g8 h2h3 c8e6 f1e1 b6a7 c4b3 e6b3 d1b3 f8e8 d2f1 d6d5 f1g3 d8d6 c1b2 d5e4 d3e4 d6e6 b3e6 e8e6 e1d1 f6e8 b2a3 e8d6 g1f1 f7f6 b4b5 c6a5 a3d6 e6d6 g3f5');
INSERT INTO evaluations VALUES(5736,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(5736,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(5736,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(5736,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(5736,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(5736,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(5736,36,'cp 20','g8f6 b2b4 c5b6 b1d2 e8g8 c4b3 c6e7 a2a4 c7c6 d3d4 e7g6 d1c2 b6c7 d4e5 d6e5 f1d1 f6h5 f3e1 d8f6 a4a5 h5f4 d2f1 b7b6 a5b6 c7b6 c1e3 c8g4 f2f3 g4e6 b3e6 f4e6 g2g3 c6c5 b4b5 f8d8 d1d8 a8d8 c2e2 h6h5 h2h4 f6e7 e1d3 d8d6 a1a6 e7d7 d3f2 g6e7');
INSERT INTO evaluations VALUES(18950,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(18950,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18950,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18950,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18950,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(18950,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18950,36,'cp 18','a7a6 b5a4 b7b5 a4c2 g8f6 a2a4 c8b7 e1g1 f8e7 d2d4 c5d4 c3d4 c6d4 f3d4 e5d4 e4e5 f6d5 d1d4 d8b6 d4g4 e8g8 a4b5 a6b5 a1a8 f8a8 g4d7 b6c6 d7c6 b7c6 f1d1 d5b4 c2b3 a8a1 b1c3 b4d3 d1d3 a1c1 d3d1 c1d1 c3d1 g8f8 g1f1 e7c5 d1c3 h7h6 c3d5 c6d5 b3d5 c5b4 d5e4 b4e7');
INSERT INTO evaluations VALUES(17452,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(17452,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(17452,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(17452,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(17452,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(17452,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(17452,36,'cp 21','b5a4 b7b5 a4c2 g8f6 e1g1 c8b7 f1e1 d8b6 d2d3 f8e7 f3h4 e8g8 h4f5 e7d8 f5d6 c6b4 d6b7 b4c2 d1c2 b6b7 a2a4 a8b8 a4b5 a6b5 c3c4 b5b4 b1d2 d7d6 d2f1 d8e7 c2e2 h7h6 e2f3 b8a8 c1d2 f8e8 e1d1 a8a1 d1a1 e8a8 a1a8 b7a8');
INSERT INTO evaluations VALUES(10125,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10125,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10125,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10125,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10125,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10125,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10125,36,'cp 0','c5e3 f2e3 d7d6 a2a4 c6e7 e1g1 e7g6 d1e1 c7c6 b1c3 e8g8 a4a5 c8d7 f3h4 g6h4 e1h4 b7b5 a5b6 a7b6 a1a8 d8a8 f1f6 g7f6 h4f6 d7e6 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(10437,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10437,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10437,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10437,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10437,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10437,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10437,36,'cp 0','f2e3 d7d6 a2a4 c6e7 c4b3 e7g6 e1g1 c7c6 d1e1 e8g8 b1c3 c8d7 e1g3 f6h5 g3g5 h5f6');
INSERT INTO evaluations VALUES(10436,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10436,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10436,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10436,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10436,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10436,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10436,36,'cp 0','d7d6 a2a4 c6e7 e1g1 e7g6 c4b3 e8g8 d1e1 c7c6 b1c3 c8e6 b3e6 f7e6 f3g5 f6g4 g5f3');
INSERT INTO evaluations VALUES(9445,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9445,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9445,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9445,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9445,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9445,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9445,36,'cp 1','a2a4 e8g8 e1g1 c6e7 d1e1 c7c6 b1d2 e7g6 f3h4 g6h4 e1h4 f8e8 h4g3 c8e6 c4e6 e8e6 d2f3 h7h6 f1f2 g7g6 f3h4 f6h5 g3h3 d8e7 a1f1 h5f6 h4f3 g8g7 h3h4 a8f8 a4a5 d6d5 e4d5 f6d5 e3e4 e7h4 f3h4 d5e3');
INSERT INTO evaluations VALUES(51980,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51980,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51980,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51980,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51980,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51980,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(51980,36,'cp 1','b8c6 g1e2 d7d5 e4d5 e6d5 d2d4 c8g4 f1g2 c6d4 h2h3 g4f3 e1f1 d8f6 d1d3 f3g2 f1g2 d4e2 d3e2 g8e7 c3d5 f6c6 c2c4 e8c8 h1d1 e7d5 e2g4 c6e6 c4d5 e6g4 h3g4 h7h5 c1e3 h5g4 a1c1 h8h5 e3c5 f8c5 c1c5 c8b8 d1d4 g7g6 c5b5 d8d7 a2a4 b8c7 d4g4 h5d5 b5d5 d7d5 g4f4 f7f5 f4c4 c7d6');
INSERT INTO evaluations VALUES(18814,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(18814,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18814,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18814,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(18814,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(18814,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18814,36,'cp 12','f1g2 g8f6 g1e2 d7d5 e4d5 e6d5 d2d4 c5d4 e2d4 c8g4 f2f3 g4d7 e1g1 f8e7 c1e3 e8g8 d4b3 f8e8 e3f2 d7f5 f1e1 h7h6 a2a3 d8d7 d1d2 e7f8 a1d1 a8d8 e1e8 d8e8 g3g4 f5g6 c3d5 d7d5 d2d5 f6d5 d1d5 e8e2 d5d2 e2d2 b3d2 g6c2 g2f1 f8e7');
INSERT INTO evaluations VALUES(18815,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18815,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(18815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18815,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18815,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18815,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(18815,36,'cp 0','g8f6 g1e2 d7d5 e4d5 e6d5 d2d4 c5d4 e2d4 c8g4 f2f3 g4d7 e1g1 f8e7 c1e3 e7b4 e3g5 b4e7 g5e3');
INSERT INTO evaluations VALUES(39554,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(39554,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(39554,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(39554,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(39554,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(39554,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(39554,36,'cp 32','c7c5 c2c3 c5d4 c3d4 d7d5 e4d5 b8d7 b1c3 d7b6 g1f3 g8f6 c4b3 b6d5 e1g1 e8g8 f1e1 d8d6 c1g5 c8f5 h2h3 d5c3 b2c3 f6e4 d1e2 e4g5 f3g5 g7f6 e2e3 h7h5 g5e4 f5e4 e3e4 e7e6 e4b7 d6b6 b7f3 g8g7 g1h1 a7a5');
INSERT INTO evaluations VALUES(44987,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(44987,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44987,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(44987,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(44987,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(44987,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44987,36,'cp 21','e1g1 c8g4 h2h3 g4f3 e2f3 e7e6 f3f4 a7a5 d2d3 b8d7 b1d2 f8d6 d2f3 e8g8 c1e3 f8e8 f1e1 h7h6 c2c4 e6e5 f4e5 d7e5 f3e5 d6e5 d1c2 d8d7 e1e2 e5d6 a2a3 e8e7 a1e1 a8e8 e3d4 e7e2 e1e2 e8e2 c2e2 d6e7 c4c5 a5a4 e2f3 f6e8');
INSERT INTO evaluations VALUES(44986,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(44986,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44986,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(44986,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(44986,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(44986,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(44986,36,'cp 20','c8g4 h2h3 g4f3 e2f3 e7e6 f3f4 f8d6 d2d3 e8g8 b1d2 a7a5 d2f3 b7b5 c1e3 c6c5 a2a4 b5b4 c2c4 b4c3 b2c3 b8c6 d1c2 d8c7 a1b1 a8b8 f1c1 b8b1 c1b1 f8b8 c2d1 h7h6 b1b5 c6a7 b5b8 c7b8 f3d2 f6d7 d1b3 b8b6 h3h4 a7c6 b3b6 d7b6');
INSERT INTO evaluations VALUES(29056,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(29056,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(29056,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29056,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29056,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29056,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(29056,36,'cp 27','h2h3 g4f3 e2f3 e7e6 f3f4 b8d7 d2d3 a7a5 b1d2 f8d6 d2f3 e8g8 c1e3 f8e8 a2a4 h7h6 c2c3 d8c7 f3d4 c7b6 d1c2 d6c5 a1e1 a8c8 e1e2 d7b8 c2b3 f6d7 b3b6 d7b6 b2b3 c5d4 e3d4 b6d7 d4e3 b8a6');
INSERT INTO evaluations VALUES(50127,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(50127,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(50127,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(50127,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(50127,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(50127,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(50127,36,'cp 86','g1f3 d7d6 g2g3 c7c6 f1g2 f8g7 e1g1 e8g8 b2b4 e7e5 d4e5 d6e5 e2e4 d8d1 f1d1 f8e8 b4b5 c8e6 e4f5 g6f5 g2f1 b8d7 f3g5 d7c5 b5c6 b7c6 a1b1 e6d7 d1d6 c5e6 g5e6 e8e6 d6d1 e6e8 b1b7 d7e6 d1d6 a8c8 c1e3 g7f8');
INSERT INTO evaluations VALUES(50125,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(50125,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(50125,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(50125,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(50125,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(50125,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(50125,36,'cp 54','d7d6 e2e3 a7a5 f1e2 f8g7 e1g1 e8g8 a1b1 b8a6 a2a3 e7e5 d4e5 d6e5 b2b4 a5b4 a3b4 d8e7 c3d5 f6d5 c4d5 h7h6 d1b3 g8h7 f1d1 a6b8 b3c2 b8a6 c1a3 e5e4 f3d2 c8d7 b4b5 e7a3 b5a6 b7a6 c2c7');
INSERT INTO evaluations VALUES(46609,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(46609,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(46609,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(46609,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(46609,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(46609,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(46609,36,'cp 70','g2g3 f8g7 f1g2 c7c6 e1g1 e8g8 d1c2 b8a6 a1b1 d8e8 b2b4 a6c7 b4b5 c6b5 c4b5 e8f7 a2a4 h7h6 f3h4 a8b8 b5b6 a7b6 b1b6 f6g4 h2h3 g7d4 b6b1 g4f2 f1f2 d4f2 g1f2 g6g5 h4f3 c7e6 b1b4 c8d7 h3h4 b8c8 h4g5 h6g5 c2d3 f7g7 f3d4 e6d4 b4d4');
INSERT INTO evaluations VALUES(50321,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50321,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50321,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(50321,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50321,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50321,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50321,36,'cp 0','e7e5 b1c3 b8c6 d1b3 f8c5 f1e2 c5b6 g1f3 e8g8 e1g1 d7d6 d2d3 a7a5 a2a3 h7h6 c1d2 f8e8 b3c2 c8f5 c3a4 b6a7 a4c3');
INSERT INTO evaluations VALUES(43424,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(43424,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43424,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(43424,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(43424,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(43424,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(43424,36,'cp 12','e7e6 f1e2 b8c6 f2f4 f8e7 c1e3 e8g8 e1g1 c8d7 d4b3 e6e5 d1d2 a8c8 g1h1 e5f4 e3f4 c6e5 a1d1 d7e6 d2e3 e6d7 e3g3 f8e8 b3d4 f6g4 e2g4 e7h4 g3e3 d7g4 f4e5 e8e5 d4f3 g4f3 e3f3 d8d7 f3f4 h4f6 c3d5 f6d8');
INSERT INTO evaluations VALUES(43556,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(43556,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(43556,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(43556,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43556,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43556,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43556,36,'cp 12','a4a5 b8c6 d4c6 b7c6 f1d3 f8e7 e1g1 e8g8 f1e1 c8b7 d1f3 f6d7 c1f4 d7e5 f4e5 d6e5 e1d1 d8c7 c3b1 e7b4 b1d2 b4a5 d2c4 a5b6 f3g3 f8d8 g3e5 c7e5 c4e5 b6c7 e5c4 a6a5 g2g3 c6c5 b2b3 b7a6 f2f4 a6c4 d3c4 d8d1 a1d1 a8d8 d1d8 c7d8 e4e5 g7g5 f4g5 d8g5 g1h1 g5e7');
INSERT INTO evaluations VALUES(2039,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(2039,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(2039,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(2039,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(2039,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(2039,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(2039,36,'cp 169','c6e5 c1f4 e5f3 g2f3 c7c6 d1d2 g8f6 e2e4 d7d6 e1c1 c8e6 d2d4 e7d8 h1g1 d8b6 f4d6 b6d4 d1d4 f8d6 d4d6 e8e7 d6d4 g7g6 c1d2 b7b5 d2e3 a7a5 f1e2 f6d7 f3f4 d7b6 a2a3 f7f5 f2f3 h8d8 d4d8 a8d8 e2d3 b5b4 a3b4 a5b4 c3e2 c6c5 b2b3 c5c4 e2d4 c4d3');
INSERT INTO evaluations VALUES(2083,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(2083,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(2083,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(2083,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(2083,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(2083,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(2083,36,'cp 146','c1f4 e5f3 g2f3 c7c6 d1d2 g8f6 e2e4 d7d6 e1c1 c8e6 d2d4 e7d8 h1g1 d8b6 f4d6 b6d4 d1d4 f8d6 d4d6 e8e7 d6d4 g7g6 c1d2 b7b5 d2e3 a7a5 f1e2 f6d7 g1d1 f7f5 a2a3 h8f8 f3f4 f5e4 c3e4 d7f6 e4f6 f8f6 e2f3 e6f5 f3e4 a8e8 e4f5 f6f5 e3f3 e7f7');
INSERT INTO evaluations VALUES(34381,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(34381,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(34381,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(34381,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(34381,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(34381,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(34381,36,'cp 30','g1f3 d5d6 b1c3 g8f6 d2d4 e7e6 f1c4 f8e7 e1g1 e8g8 d1e2 a7a6 c4b3 b7b5 c3e4 d6d8 c1g5 b8d7 a1c1 c8b7 e4c5 e7c5 d4c5 a8c8 f1d1 d8e7 g5f4 e6e5 f4d2 d7c5 d2b4 a6a5 b4c5 c8c5 c1c5 b7f3 e2e5 e7e5 c5e5 f3d1 b3d1 f8b8 d1f3 g8h8');
INSERT INTO evaluations VALUES(39589,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(39589,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(39589,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(39589,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(39589,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(39589,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(39589,36,'cp 51','d7d5 e2e4 d5e4 c3e4 g8f6 e4f6 e7f6 f1e2 e8g8 e1g1 f8e8 f1e1 c7c6 c1f4 g7f8 a2a3 a7a5 c2c4 b8a6 d1d2 c8f5 a1d1 a6c7 d4d5 c6d5 c4d5 f8d6 f4h6 f5e4 f3d4 e8e5 e2f1 e5h5 h2h3 e4d5 d4b5 c7b5 f1b5 d8b6 b5f1 a8d8');
INSERT INTO evaluations VALUES(38463,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(38463,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(38463,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(38463,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(38463,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38463,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(38463,36,'cp 45','e2e4 d5e4 c3e4 g8f6 e4f6 e7f6 f1e2 e8g8 e1g1 f8e8 f1e1 a7a5 c1f4 c8f5 c2c3 b8d7 d1b3 d7b6 c3c4 a5a4 b3c3 f5e4 a1d1 b6d7 c3d2 c7c6 d4d5 f6f5 f3g5 d8b6 g5e4 e8e4 f4e3 c6c5 e2d3 e4e8 e3f4 e8e1 d1e1 b6b2');
INSERT INTO evaluations VALUES(38471,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(38471,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(38471,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(38471,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(38471,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38471,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(38471,36,'cp 24','g8f6 e4e5 f6e4 f1d3 e4c3 b2c3 c7c5 e1g1 b8c6 h2h3 c5c4 d3e2 d8a5 d1d2 h7h5 f3h4 g7h6 f2f4 e7e6 h4f3 h6f8 e2d1 f8e7 d2e1 c8d7 f3g5 h8f8 c1e3 e8c8 a2a4 a5c7 g2g4 h5g4 h3g4 d8e8 a1b1');
INSERT INTO evaluations VALUES(50090,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(50090,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(50090,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(50090,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(50090,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(50090,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50090,36,'cp 44','g7g6 c2c4 f8g7 b1c3 e8g8 e2e3 d7d6 f1e2 f6e4 c3e4 f5e4 f3g5 e7e5 d4e5 d6e5 d1d8 f8d8 f4g3 c8f5 c4c5 g7f6 e2c4 g8g7 g5e6 f5e6 c4e6 b8a6 a1c1 h7h5 h2h4 a6b4 e1e2 a7a5 h1d1 b4d3 c1c2 c7c6 a2a3 a5a4 f2f3 f6e7 e6c4 d3c5 g3e5 g7h6');
INSERT INTO evaluations VALUES(50133,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(50133,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(50133,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(50133,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(50133,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(50133,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50133,36,'cp 49','b1c3 f8g7 d1d3 d7d5 c3b5 b8a6 e2e3 c7c6 b5c3 a6b8 a2a4 a7a5 d3d1 c8e6 h2h4 e6f7 f3e5 b8d7 f1e2 d7e5 f4e5 h7h5 f2f3 d8b6 d1d2 e8g8 e1c1 f7e6 c1b1 b6b4 h1g1 g7h6 e5f6 e7f6 g2g4 h5g4 f3g4 f5g4');
INSERT INTO evaluations VALUES(52477,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52477,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52477,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52477,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(52477,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(52477,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(52477,36,'cp 3','c2c4 e7e6 g2g3 f8e7 f1g2 d7d5 c4d5 e6d5 d3d4 g8f6 g1f3 e8g8 e1g1 b8c6 b1c3 c5c4 f3e5 c8e6 e5c4 d5c4 d4d5 f6d5 g2d5 e6h3 f1e1 e7c5 d5c4 d8b6 c3e4 c5b4 c1e3 b6a5 e4c3 b4c3 b2c3 c6e5 d1d5 e5c4 d5c4 a8c8 c4f4 c8c3');
INSERT INTO evaluations VALUES(52422,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52422,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52422,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52422,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52422,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52422,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52422,36,'cp 0','e7e6 g2g3 b8c6 g1f3 g8e7 b1c3 d7d5 c4d5 e7d5 f1g2 f8e7 d3d4 e8g8 e1g1 b7b6 d4c5 b6c5 d1a4 c8d7 a4d1');
INSERT INTO evaluations VALUES(4327,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(4327,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(4327,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(4327,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(4327,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(4327,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(4327,36,'cp -76','f6e4 d2d4 e5d4 c3d4 c5b4 e1e4 d7d5 c4d5 d8d5 e4h4 c8f5 b1c3 b4c3 b2c3 a8d8 h4h5 d5d7 h2h3 f8e8 c1f4 c6e7 d1b3 e7d5 f4g3 c7c6 f3e5 d7c8 e5c4 f5e6 c4d6 d8d6 g3d6 d5f6 b3d1 f6h5 d1h5 c8d7 d6e5 f7f6 e5g3 e6f7 h5f3 f7d5 f3d3');
INSERT INTO evaluations VALUES(39031,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(39031,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(39031,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(39031,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(39031,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(39031,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(39031,36,'cp 141','h2h4 d7d5 g1f3 d5e4 c3e4 h7h6 f1d3 b8d7 d1e2 g8f6 c1f4 f6e4 e2e4 h6h5 e1c1 e8g8 e4e2 d7f6 f4g5 d8d5 c1b1 f6g4 h1e1 c7c5 g5e7 c5d4 d3e4 d5a5 e7f8 g8f8 f3d4 a5b6 c2c3 e6e5 e4c2 f8g8 f2f3 e5d4 e2e8 g8h7 f3g4 d4c3');
INSERT INTO evaluations VALUES(4168,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(4168,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(4168,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(4168,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(4168,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(4168,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(4168,36,'cp 73','g2g4 d7d5 e4d5 c6d4 c3e4 f6e4 d3e4 b7b5 c4d3 c7c6 f3d4 c5d4 c2c3 d4c5 d5c6 d8b6 d1e2 a8b8 c6c7 b8b7 e1g1 a7a5 b2b4 a5b4 c3b4 c5b4 c1e3 b6a5 a1c1 b4d6 c1c6 d6c7 f1c1 c8d7 c6c5 c7b8 e2d2 a5d2 e3d2 f8d8 c5d5 f7f6 c1c5 b8c7 d2e3 d7e6');
INSERT INTO evaluations VALUES(19804,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19804,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(19804,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19804,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(19804,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19804,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(19804,36,'cp 0','g1f3 e7e6 f1b5 g8e7 e1g1 a7a6 b5c6 e7c6 d2d4 f8e7 d4c5 e8g8 d1d2 f7f6 b2a3 e6e5 b1c3 c8e6 a1d1 d8a5 c3a4 a5b5 f3e1 d5d4 e1d3 d4e3 f2e3 a8d8 d2c3 e6f7 a4b6 f7h5 d1d2 h5f7');
INSERT INTO evaluations VALUES(19806,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19806,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19806,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19806,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19806,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19806,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19806,36,'cp 0','e7e6 d2d4 c5d4 e3d4 f8d6 f1d3 b7b6 e1g1 g8f6 c2c4 e8g8 b1c3 f8e8 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(32484,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(32484,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(32484,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(32484,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(32484,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(32484,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(32484,36,'cp 114','f8e7 f2f3 b8c6 c1e3 e7h4 g2g3 h4f6 d1d2 c6d4 e3d4 f6d4 d2d4 d8f6 d4f2 g8e7 e1c1 a7a6 f3f4 f6h6 f1c4 d7e6 c4e2 e7c6 h2h4 e8c8 g3g4 e6d7 f2g3 h6e6 h4h5 c8b8 e2d3 h8e8 c3d5 f7f6 h5h6');
INSERT INTO evaluations VALUES(28172,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(28172,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(28172,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(28172,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(28172,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(28172,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(28172,36,'cp 127','f2f3 a7a6 c1e3 g8f6 d1d2 c7c5 d4e2 b7b5 e2f4 d7c6 e1c1 b8d7 g2g4 b5b4 c3d5 f6d5 f4d5 c6d5 d2d5 e8g8 c1b1 a6a5 h2h4 d7b6 d5f5 d8c8 f5c8 a8c8 g4g5 c8d8 c2c4 b4c3 b2c3 f7f5 e4f5 f8f5 f3f4 d6d5');
INSERT INTO evaluations VALUES(39725,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(39725,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(39725,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(39725,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39725,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39725,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39725,36,'cp 0','a2a4 e8g8 c1e3 b8c6 d1e2 c8d7 d4c6 d7c6 a4a5 f6d7 f1d1 d8c7 e3d4 e7f6 e2e3 a8c8 d4f6 d7f6 h2h3 f8d8 d1d4 f6d7 c4e2 h7h6 d4d2 d7f6 d2d4');
INSERT INTO evaluations VALUES(39750,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39750,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39750,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39750,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39750,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39750,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(39750,36,'cp 0','b8c6 h2h3 c8d7 c1e3 e8g8 f2f4 d8c7 d1e2 c6d4 e3d4 d7c6 a1e1 e6e5 d4f2 e5f4 f2d4 a8e8 c4a2 f6d7 f1f4 e7f6 e2f2 d6d5 f4f6 d7f6 d4f6 g7f6 c3d5 c6d5 a2d5 f6f5 f2f5 c7c2 f5g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(7094,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7094,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7094,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7094,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7094,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7094,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7094,36,'cp 0','c1e3 e8g8 f1e1 c8d7 d4c6 d7c6 a4a5 f6d7 c4f1 a8c8 d1d2 h7h6 e1d1 d8c7 h2h3 f8e8 e3f4 d7e5 f4e3');
INSERT INTO evaluations VALUES(15949,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(15949,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(15949,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(15949,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(15949,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(15949,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(15949,36,'cp 23','c1f4 f8b4 c4d5 f6d5 f4d2 d5f6 a2a3 b4c3 d2c3 b7b6 e2e3 c8b7 f1d3 f6e4 d3e4 b7e4 e1g1 e8g8 d1e2 d8e7 f1d1 c7c5 d1d2 f8d8 a1d1 a8c8 h2h3 h7h6 f3e1 c5d4 d2d4 e4b7 d4d6 d7f6 d6d8 c8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(15945,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(15945,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(15945,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(15945,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(15945,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(15945,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(15945,36,'cp 22','f8b4 c4d5 f6d5 f4d2 d5f6 a1c1 e8g8 g2g3 b7b6 f1g2 c8b7 e1g1 d8e7 c3b5 b4d2 d1d2 c7c5 d2f4 a7a6 f4d6 e7d8 b5c3 c5c4 a2a4 a8c8 f1d1 d8c7 d6c7 c8c7 f3e1 b7g2 e1g2 f8b8 g2e3 b6b5 f2f3 h7h6 g1f2 b5b4');
INSERT INTO evaluations VALUES(26269,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(26269,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(26269,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(26269,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(26269,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(26269,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(26269,36,'cp 62','d4c5 e7e6 c1e3 g4f3 d1f3 a7a6 f3g3 g8e7 f1d3 e7g6 f2f4 d5d4 c3d4 f8c5 b1d2 c5d4 a1d1 d8a5 e1g1 e8g8 e3d4 c6d4 g3f2 a8d8 d2e4 d4f5 a2a3 d8d5 g2g4 a5d8 e4c5 f5h4 c5b7 d8a8 d3g6 d5d1 g6h7 g8h7 f2h4 h7g8 f1d1 a8b7 d1d2 b7c7');
INSERT INTO evaluations VALUES(18758,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(18758,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(18758,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(18758,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(18758,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(18758,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(18758,36,'cp -62','d7d5 c4b5 c8d7 e4d5 e6d5 b1c3 g8f6 b5c6 d7c6 d2d4 f8d6 d1e2 f6e4 c3e4 d5e4 f3e5 e8g8 e5c6 b7c6 d4c5 d6c5 c1e3 d8a5 e2d2 c5b4 c2c3 b4e7 e1g1 f8d8 d2e2 d8d3 e3d4 a5d5 a1d1 c6c5 d1d3 e4d3 e2e7 c5d4 e7e5 d5e5 f4e5 d4c3 b2c3 a8c8 f1c1 g8f8 c3c4 a7a6 c1d1 c8c4 d1d3 c4c1 g1f2 c1c2 f2g3');
INSERT INTO evaluations VALUES(20247,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(20247,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(20247,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(20247,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(20247,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(20247,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(20247,36,'cp -51','c4b5 c8d7 e4d5 e6d5 b1c3 g8f6 b5c6 d7c6 d2d4 d8c7 e1g1 f8e7 f3e5 e8g8 c1e3 f8d8 a2a4 f6e4 f1e1 a8c8 d1f3 e4c3 b2c3 f7f6 e5c6 c7c6 e3f2 e7f8 h2h3 b7b6 a4a5 b6b5 d4c5 f8c5 g1h2 c5f2 f3f2 d5d4 e1e7 c6c3 a1d1 c3c2 f2c2');
INSERT INTO evaluations VALUES(32400,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32400,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32400,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(32400,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(32400,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(32400,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(32400,36,'cp 22','g8f6 h2h3 g4f3 e2f3 c7c6 e1g1 f8e7 c2c4 d5c4 b2b3 c4b3 d1b3 d8d4 b3b7 e8g8 b7a8 d4a1 a8a7 a1b1 a7e7 b1a2 c1f4 a2e6 e7e6 f7e6 f1e1 f6d5 f4g3 g8f7 e1a1 c6c5 g3d6 f8d8 d6c5 g7g6 c5d4 d8d7 a1a8 b8c6 d4b2 c6e7 a8h8 d5f6 b2e5');
INSERT INTO evaluations VALUES(30247,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(30247,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(30247,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(30247,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(30247,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(30247,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(30247,36,'cp 22','e1g1 f8e7 h2h3 g4f3 e2f3 c7c6 c1e3 e8g8 b1d2 b8a6 c2c4 a6c7 d1b3 a8b8 a1d1 c7e6 a2a3 f8e8 f3e2 e7d6 g2g3 d6c7 d2f3 d5c4 b3c4 d8d6 g1g2 b8d8 c4c2 f6d5 e3d2 c7b6 e2d3 e6d4 f3d4 b6d4 d3h7 g8h8');
INSERT INTO evaluations VALUES(30246,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(30246,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(30246,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(30246,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(30246,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(30246,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(30246,36,'cp 25','f8d6 h2h3 g4e6 e2d3 e8g8 b1c3 c7c5 d4c5 d6c5 a2a3 f8e8 b2b4 c5f8 c3b5 b8c6 c1e3 h7h6 b5d4 e6d7 c2c3 f8d6 f1e1 c6d4 e3d4 e8e1 d1e1 d8c7 e1d1 f6h5 d1c2 a8e8 d3f1 h5f4 d4e3 f4e6 c2b3 e6f4 a1e1');
INSERT INTO evaluations VALUES(28441,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(28441,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(28441,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(28441,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(28441,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(28441,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(28441,36,'cp 39','h2h3 g4e6 b1c3 e8g8 c3b5 d6b4 c1f4 b4a5 e2d3 b8c6 b5c3 c6b4 d3e2 b4c6 c3a4 a5b4 a2a3 b4d6 f4d6 d8d6 a4c5 e6c8 f1e1 c6e7 a1c1 e7g6 e2d3 b7b6 d3g6 h7g6 c5d3 c8b7 b2b4 a7a5 f3e5 f6e4 d1f3 e4d2 f3f4');
INSERT INTO evaluations VALUES(22507,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(22507,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(22507,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(22507,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(22507,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(22507,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(22507,36,'cp 101','e1g1 g8f6 f1e1 d7d6 d2d4 c8d7 b5c6 d7c6 d1d3 f6d7 c1e3 e5d4 f3d4 f8e7 d4c6 b7c6 b1d2 e8g8 a1d1 a7a5 a2a4 d7e5 d3e2 f8e8 f2f4 e5g6 e1f1 d8d7 d2c4 g6f8 e4e5 d7f5 e5d6 c7d6 c4d6 e7d6 d1d6 a8b8');
INSERT INTO evaluations VALUES(22505,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(22505,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(22505,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(22505,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(22505,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(22505,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(22505,36,'cp 85','g8f6 d2d4 e5d4 e4e5 f6h7 f3d4 f8c5 d4b3 c5b6 d1g4 c6e5 g4g7 d8f6 g7g3 e8f8 b1c3 h8g8 c3d5 g8g3 d5f6 g3g6 f6d5 d7d6 b5e2 c8e6 d5f4 g6f6 c2c3 f8e7 b3d4 e7d7 d4e6 f7e6 f4h5 f6g6 g2g3 a8g8 g1g2 h7f6 a2a4 a7a5 h5f6 g6f6 h2h3 d6d5 f2f4 e5f7 g3g4');
INSERT INTO evaluations VALUES(16105,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(16105,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(16105,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(16105,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(16105,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(16105,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(16105,36,'cp 114','d2d4 e5d4 e4e5 f6h7 f3d4 f8c5 d4b3 c5b6 d1g4 c6e5 g4g7 d8f6 g7g3 e8f8 b1c3 h8g8 c3d5 g8g3 d5f6 g3g6 f6d5 d7d6 b5e2 c8e6 d5b6 a7b6 e2h5 g6f6 b3d4 e5c6 d4e6 f7e6 f2f4 f8e7 c2c3 f6f8 h5d1 h7f6 f4f5 f6e4 d1c2 d6d5 c1h6 f8f5 f1f5 e6f5 h6f4 e7e6 f4c7 b6b5 c2b3 c6e5');
INSERT INTO evaluations VALUES(46000,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(46000,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(46000,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46000,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46000,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(46000,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46000,36,'cp 12','f1e1 b8c6 c4b3 c6a5 d2d4 c5d4 d1d4 a7a6 b1c3 f8e7 e4e5 a5c6 d4f4 c6e5 f3e5 d6e5 f4e5 c8d7 c1f4 e8g8 a1d1 d8c8 h2h3 c8c5 c3a4 c5e5 f4e5 d7c6 a4b6 a8d8 e5c7 d8d1 e1d1 e7c5 g1f1 h7h6 c2c3 f6e8 b6d7 e8c7 d7c5 c7d5 b3d5');
INSERT INTO evaluations VALUES(46001,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(46001,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46001,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(46001,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46001,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(46001,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(46001,36,'cp 16','b8c6 c4b5 c8d7 c2c3 d6d5 e4d5 f6d5 d2d4 c5d4 b5c4 f8e7 c3d4 e8g8 b1c3 c6b4 c4b3 d5f6 f3e5 d7c6 c1g5 b4d5 a1c1 h7h6 g5h4 a8c8 e5c6 c8c6 d1f3 d8d7 e1d1 a7a6 h4g3 b7b5 a2a4 d7b7 b3d5 f6d5 c3d5 e6d5');
INSERT INTO evaluations VALUES(13939,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(13939,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(13939,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13939,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13939,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13939,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(13939,36,'cp 24','c4b5 c8d7 c2c3 d6d5 e4d5 f6d5 b5c4 f8e7 d2d4 c5d4 c3d4 e8g8 b1c3 e7f6 c3d5 e6d5 c4d5 d8a5 d5e4 c6d4 f3d4 f6d4 c1e3 d4e3 e1e3 d7c6 e4c6 b7c6 g2g3 g7g6 d1c2 a8c8 a2a3 f8d8 e3c3 d8d6 b2b4 a5d8 c3f3 d8d7 h2h4 c8d8 a1c1 d6d1 c1d1 d7d1 c2d1 d8d1 g1h2 d1d2 h2h3');
INSERT INTO evaluations VALUES(47326,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(47326,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47326,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(47326,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(47326,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47326,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(47326,36,'cp 65','d2d4 a7a6 c1e3 b8d7 d1f3 d7b6 c4b3 c7c6 h2h3 b6d5 c3d5 c6d5 g1e2 e7e6 e1g1 f8e7 c2c3 e8g8 e2f4 b7b5 f4d3 a6a5 f1c1 b5b4 c3b4 a5b4 e3f4 c8a6 f4c7 d8e8 d3c5 a6b5 f3g3 a8c8 c7d6 e7d6 g3d6 f6e4');
INSERT INTO evaluations VALUES(47331,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(47331,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(47331,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(47331,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(47331,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(47331,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(47331,36,'cp 43','a7a6 c1e3 b8d7 d1f3 c7c6 f3g3 d7b6 c4d3 g7g6 g1f3 f8g7 e1g1 c8e6 b2b3 b6d5 c3d5 e6d5 f3e5 e8g8 f1e1 d5e4 a1d1 e4d3 d1d3 f6e4 g3f4 e4d6 e3d2 d6f5 c2c4 g7h6 f4e4 h6d2 d3d2 f5d6');
INSERT INTO evaluations VALUES(30416,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(30416,31,'cp 258',NULL);
INSERT INTO evaluations VALUES(30416,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(30416,33,'cp 252',NULL);
INSERT INTO evaluations VALUES(30416,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(30416,35,'cp 249',NULL);
INSERT INTO evaluations VALUES(30416,36,'cp 267','b1c3 c7c6 b2b3 a7a5 b3c4 b5a6 c1a3 e7e6 a3f8 e8f8 f1e2 g7g6 h2h4 h7h5 g1f3 f8g7 e1g1 b8d7 c3e4 g8f6 e4d6 f6e8 d6f7 g7f7 f3g5 f7g8 e2d3 d7f8 d1f3 a8a7 f3c6 e8d6 c6c5 a7d7 f1b1 d8c7 a1a5 a6c4 d3c4 d6c4 a5a8 c7c5 d4c5 c4e5');
INSERT INTO evaluations VALUES(30415,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(30415,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(30415,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(30415,33,'cp 262',NULL);
INSERT INTO evaluations VALUES(30415,34,'cp 233',NULL);
INSERT INTO evaluations VALUES(30415,35,'cp 235',NULL);
INSERT INTO evaluations VALUES(30415,36,'cp 261','c7c6 b2b3 a7a5 b3c4 b5a6 c1a3 e7e6 a3f8 e8f8 g1f3 g7g6 h2h4 g8f6 h4h5 f6h5 g2g4 h5f6 g4g5 f6e8 d4d5 e6d5 d1d4 f8g8 c4d5 c6c5 d4f4 b8d7 e3e4 h7h5 g5h6 a6f1 h6h7 h8h7 h1h7 d8f6 f4f6 e8f6 h7h4 f1c4 f3d2 c4d3 f2f4 c5c4 e4e5 f6h5 d5d6 d3f5 d2c4');
INSERT INTO evaluations VALUES(30466,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(30466,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(30466,32,'cp 215',NULL);
INSERT INTO evaluations VALUES(30466,33,'cp 216',NULL);
INSERT INTO evaluations VALUES(30466,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(30466,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(30466,36,'cp 240','b2b3 a7a5 b3c4 b5a6 c1a3 e7e6 a3f8 e8f8 f1e2 g7g6 g1f3 f8g7 h2h4 h7h5 c3e4 b8d7 f3g5 d8e7 e1g1 g8f6 d1c2 f6e4 c2e4 c6c5 f1b1 c5d4 e3d4 h8e8 e4c2 e8c8 a1a5 a6b7 a5a8 b7a8 c2d2 d7f6 b1b6 a8c6 d2b2 g7g8 c4c5 c6d5');
INSERT INTO evaluations VALUES(30465,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(30465,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(30465,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(30465,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(30465,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(30465,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(30465,36,'cp 235','a7a5 b3c4 b5a6 c1a3 e7e6 a3f8 e8f8 g1f3 g7g6 f1e2 f8g7 h2h4 h7h5 c3e4 b8d7 f3g5 d8e7 e1g1 g8f6 e4f6 d7f6 d1d2 c6c5 a1a5 c5d4 d2d4 h8d8 d4e5 a6b7 f1a1 a8a5 a1a5 e7b4 a5a7 d8d7 e5a1 e6e5 a1e5 b4e1 e2f1 d7d1 a7b7 e1f1 g1h2');
INSERT INTO evaluations VALUES(39020,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(39020,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(39020,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(39020,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(39020,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(39020,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(39020,36,'cp 60','d7d5 b1c3 d5e4 c3e4 b8c6 c1g5 f7f6 g5e3 g8e7 d1d2 e7f5 e1c1 d8e7 h2h4 h7h5 f1d3 f5e3 f2e3 e8g8 c4c5 b7b6 c1b1 c8b7 h1g1 f6f5 e4g5 b6c5 d3c4 c6d8 g1e1 g8h8 e3e4 c5d4 e4f5 f8f5 f3d4');
INSERT INTO evaluations VALUES(57163,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(57163,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(57163,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(57163,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(57163,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(57163,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(57163,36,'cp -25','b8c6 g1f3 g8f6 d2d3 d7d5 e4d5 f6d5 e1g1 f8d6 a2a3 a7a5 b1c3 d5c3 b2c3 e8g8 a1b1 f7f5 f3d2 b7b6 d2c4 c8e6 e2f3 d8d7 c4d6 d7d6 b1b5 c6a7 b5b1 a8e8 f1e1 e6d5 h2h3 h7h6 f3h5 d5f7 h5f3 e8d8 c3c4');
INSERT INTO evaluations VALUES(22822,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(22822,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(22822,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(22822,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(22822,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(22822,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(22822,36,'cp -8','d2d3 g8f6 c2c4 f8c5 h2h3 c6d4 g1f3 d4f3 e2f3 c7c6 c1e3 c5e3 f2e3 e8g8 d1d2 d7d5 e4d5 c6d5 b1c3 c8e6 c4d5 f6d5 f3d5 e6d5 e1g1 b7b5 d3d4 e5d4 e3d4 d5c6 a2a3 a7a5 a1d1 b5b4 a3b4 a5b4 c3e2 f8e8 e2g3 c6b5');
INSERT INTO evaluations VALUES(22801,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(22801,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(22801,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(22801,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22801,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(22801,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(22801,36,'cp -8','g8f6 c2c4 f8c5 g1f3 d7d6 h2h3 h7h6 a2a3 c6d4 f3d4 c5d4 b1c3 c7c6 e1g1 a7a6 c1e3 d4e3 f2e3 b7b5 d3d4 e8g8 d1d2 d8e7 f1f2 a8b8 a1f1 c8e6 d4d5 e6d7 d5c6 d7c6 c4b5 a6b5 e2d3 f8c8 g1h2 c6d7 c3d5 f6d5 e4d5');
INSERT INTO evaluations VALUES(16560,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(16560,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(16560,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(16560,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(16560,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16560,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(16560,36,'cp 0','c2c4 f8c5 g1f3 d7d6 h2h3 c6d4 c1e3 d4f3 e2f3 c5e3 f2e3 c7c6 d1d2 b7b5 c4b5 c6b5 a2a4 b5a4 a1a4 e8g8 e1g1 c8d7 a4a6 d7b5 a6a3 d8b6 b1c3 a7a6 f1a1 b5d7 g1h2 f8b8 c3a4 b6b4 d2b4 b8b4 a4c3 b4b2 a3a6 a8a6 a1a6');
INSERT INTO evaluations VALUES(2463,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(2463,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(2463,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(2463,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(2463,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(2463,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(2463,36,'cp 243','c4d5 e6d6 f3d1 c7c6 d5e4 h7h5 d2d3 d6c7 e1g1 c8g4 f2f3 g4e6 c1e3 a8c8 d1d2 c7b8 a1e1 f8d6 e4g6 h5h4 h2h3 e6f5 g6f5 d4f5 c3e4 f5e3 e1e3 d6c7 g1h1 d8e7 f1e1 a7a5 e3e2 e7f7 b2b3 f7f4 d2c3 h8d8 a2a3 d8d4 c3c5 c8e8 b3b4');
INSERT INTO evaluations VALUES(2450,30,'cp 239',NULL);
INSERT INTO evaluations VALUES(2450,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(2450,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(2450,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(2450,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(2450,35,'cp 243',NULL);
INSERT INTO evaluations VALUES(2450,36,'cp 249','e6d6 f3d1 c7c6 d5e4 d6c7 e1g1 c8e6 d2d3 d8d7 c3b1 a8d8 c1g5 f8e7 g5e7 d7e7 c2c3 d4f5 b1d2 h7h5 d1e2 e7f6 b2b4 g7g5 a2a4 c7b8 b4b5 c6c5 a4a5 g5g4 b5b6 a7a6 a1e1 f5d6 d2c4 e6c4 d3c4 d6e4 e2e4');
INSERT INTO evaluations VALUES(2446,30,'cp 238',NULL);
INSERT INTO evaluations VALUES(2446,31,'cp 239',NULL);
INSERT INTO evaluations VALUES(2446,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(2446,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(2446,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(2446,35,'cp 251',NULL);
INSERT INTO evaluations VALUES(2446,36,'cp 246','f3d1 c7c6 d5e4 d6c7 d2d3 h7h5 e1g1 c8g4 f2f3 g4e6 c1e3 a8c8 d1d2 c7b8 a1e1 f8d6 e4g6 h5h4 h2h3 e6f5 g6f5 d4f5 c3e4 f5e3 e1e3 d6c7 g1h1 a7a6 e3e2 d8d5 a2a4 d5a2 b2b3 b8a8 f1e1 h8f8 d2b4 c8d8');
INSERT INTO evaluations VALUES(14344,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(14344,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(14344,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(14344,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(14344,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(14344,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(14344,36,'cp -19','e7e5 d2d3 c8e6 a2a3 a7a5 c2c4 d5d4 e2e3 f8e7 e3d4 e5d4 c1g5 h7h6 g5f6 e7f6 f1e1 e8g8 b1d2 a5a4 d1c2 d8d7 d2e4 f6e7 e4d2 a8a5 e1e2 e6g4 a1e1 e7f6 h2h4 f8a8 e2e4 g4f5 e4e2 a8d8 b2b4 a4b3 d2b3 a5a8');
INSERT INTO evaluations VALUES(990,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(990,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(990,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(990,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(990,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(990,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(990,36,'cp -52','f1e2 c8f5 a2a3 e8c8 e1g1 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 h8e8 a1d1 c5e3 f2e3 f5c5 g1f2 f7f6 d1d8 c6d8 b2b4 c5e7 e2d3 a7a6 c2c4 c8b8 f1e1 h7h6 h2h3 d8c6 d3c3 e7d6 f3d4 c6d4 c3d4 d6e6 e1d1 e6e7');
INSERT INTO evaluations VALUES(989,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(989,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(989,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(989,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(989,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(989,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(989,36,'cp -59','c8f5 a2a3 e8c8 e1g1 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 c5e3 f2e3 h8e8 a1d1 d8d1 f1d1 c6e5 d1f1 f5h5 e2d2 f7f6 d2d4 a7a6 f3e5 h5e5 d4e5 e8e5 f1e1 b7b6 g1f2 e5b5 b2b3 b5c5 c2c4 c8d7 e1d1 d7e6');
INSERT INTO evaluations VALUES(56745,30,'cp 289',NULL);
INSERT INTO evaluations VALUES(56745,31,'cp 311',NULL);
INSERT INTO evaluations VALUES(56745,32,'cp 297',NULL);
INSERT INTO evaluations VALUES(56745,33,'cp 288',NULL);
INSERT INTO evaluations VALUES(56745,34,'cp 280',NULL);
INSERT INTO evaluations VALUES(56745,35,'cp 284',NULL);
INSERT INTO evaluations VALUES(56745,36,'cp 297','f1c4 b8c6 f3g5 g8h6 d4d5 c6e7 b1c3 a7a6 g5e6 c8e6 d5e6 g7g6 c1g5 c7c6 g5f6 h8g8 f6g5 d8c7 d1d2 f5e4 g5h6 f8h6 d2h6 d6d5 e1c1 e8c8 h6h7 g8h8 h7g7 d8g8 g7f6 g8f8 f6g5 f8f2 h2h4 f2f4 h4h5 g6h5 c4e2 c7d6 h1h5 h8h5 g5h5 c8c7 h5h3 f4f8 c1b1');
INSERT INTO evaluations VALUES(1351,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(1351,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(1351,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(1351,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(1351,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(1351,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(1351,36,'cp 76','d4b3 e7e6 a2a3 f8e7 c1e3 b6c7 b1c3 g8f6 f2f4 e8g8 d1f3 d7d5 e4e5 f6e8 e1c1 f7f6 e5f6 e7f6 c3b5 c7f7 e3c5 f6e7 g2g3 e7c5 b3c5 a7a6 b5c3 a8b8 f1g2 b7b5 f3e3 e8c7 h1e1 a6a5 g2h3 b5b4 c3a4 e6e5 h3c8 b8c8 f4e5 b4a3 e3a3');
INSERT INTO evaluations VALUES(1352,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(1352,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(1352,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(1352,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(1352,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(1352,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(1352,36,'cp 72','b6c7 c2c4 e7e6 b1c3 g8f6 a2a3 f8e7 f1e2 e8g8 e1g1 b7b6 f2f4 c8a6 c1e3 f8d8 e4e5 f6e8 c3b5 c7c8 d1e1 d7d5 a1c1 c8b7 c4c5 a6b5 e2b5 b6c5 b5c6 b7c6 e3c5 c6b7 b3a5 b7d7 c5e7 d7e7 b2b4 d8c8');
INSERT INTO evaluations VALUES(51427,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(51427,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(51427,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(51427,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(51427,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(51427,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(51427,36,'cp 50','c8b7 d2d4 f8b4 f1d3 g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 f2f4 d7d6 a2a4 e6e5 d2b3 b4c3 b2c3 d8c7 d1e1 f8e8 f4e5 c6e5 b3d4 d6d5 d3b5 f6e4 b5e8 a8e8 c1f4 f7f6 e1e3 b7a6 f1e1 a6c4 d4f5 c7c6 e3d4 h7h6 f4e5 f6e5');
INSERT INTO evaluations VALUES(51708,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51708,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(51708,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(51708,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(51708,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(51708,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51708,36,'cp 23','f8e7 a2a4 g8f6 a4b5 c6b5 d2d4 b8d7 e1g1 e8g8 b1c3 c8b7 c3b5 f6e4 c1e3 a7a6 b5a3 e5d4 f3d4 d7c5 b3a2 f8e8 d4f5 e7f8 d1g4 g7g6 f5h6 f8h6 e3h6 d6d5 a1d1 d8b6 a2d5 b7d5 d1d5 b6b2 a3c4 b2c3 c4d6 e4d6 d5d6 c5e4 d6d7 c3c6');
INSERT INTO evaluations VALUES(46030,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(46030,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(46030,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(46030,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(46030,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(46030,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(46030,36,'cp 88','e2e4 g7g6 g1f3 f8g7 f1d3 e8g8 h2h3 e6d5 e4d5 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1c2 e4f6 e3d2 b8a6 a2a3 a6c7 b2b3 d8d7 d2c3 b7b5 f3g5 a7a5 c2b2 b5c4 b3c4 a8b8 b2c2 h7h6 g5f3 f6h5 f1e1 g7f6 c2d2 f6c3 d2c3');
INSERT INTO evaluations VALUES(46027,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(46027,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(46027,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(46027,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(46027,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(46027,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(46027,36,'cp 93','g7g6 f1d3 f8g7 g1f3 e6d5 e4d5 e8g8 h2h3 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1b3 d8b6 b3c2 e4f6 a1e1 b8d7 e3d2 e8e1 f1e1 a8e8 g2g4 b6d8 g1g2 e8e1 f3e1 b7b5 b2b3 h7h6 f2f4 d8e7 e1f3 a7a6 c2d1 d7b6 d2c1 b5c4 b3c4 f6e4 d1e2');
INSERT INTO evaluations VALUES(45532,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(45532,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(45532,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(45532,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(45532,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(45532,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(45532,36,'cp 101','g1f3 f8g7 h2h3 e6d5 e4d5 e8g8 f1d3 f8e8 c1e3 a7a6 e1g1 b8d7 d1d2 b7b5 c4b5 d7b6 b5a6 c8a6 d3a6 a8a6 b2b3 a6a3 e3g5 d8a8 g5f6 g7f6 c3b5 f6a1 f1a1 a8a5 b5a3 a5a3 a1d1 a3b4 g2g3 b4d2 d1d2 c5c4 g1g2 c4b3 a2b3');
INSERT INTO evaluations VALUES(47609,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(47609,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(47609,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(47609,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47609,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(47609,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(47609,36,'cp 23','c7c5 d2d3 f8e7 f3h2 c8e6 h2f1 e8g8 f1e3 f6d7 a2a4 e7g5 e1g1 b8c6 c3d5 d7b6 b2b3 b6c4 b3c4 d8d7 e3g4 f7f6 f2f4 e5f4 c1f4 e6g4 d1g4 d7g4 h3g4 g5f4 f1f4 a8e8 a4a5 e8e5 d5e3 c6d4 f4f2 f8f7');
INSERT INTO evaluations VALUES(45566,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(45566,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(45566,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(45566,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(45566,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(45566,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(45566,36,'cp 28','d2d3 f8e7 f3h2 b8c6 a2a4 e8g8 h2g4 c8e6 c3d5 c6a5 g4f6 e7f6 e1g1 f6g5 c1d2 a5c4 d3c4 g5h4 d1h5 e6d5 c4d5 h4g5 d2g5 d8g5 h5g5 h6g5 a1a3 g8h7 a4a5 f7f5 a3g3 h7h6 g3b3 a8b8 e4f5 f8f5 g2g3 f5f6 f2f3 b7b5 a5b6 a7b6');
INSERT INTO evaluations VALUES(45568,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(45568,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(45568,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45568,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(45568,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(45568,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(45568,36,'cp 27','f8e7 f3h2 c8e6 h2g4 b8c6 c3d5 a7a6 c2c3 b7b5 d5f6 e7f6 c4d5 a8c8 d1b3 c6a5 g4f6 d8f6 b3a3 f6d8 d5e6 f7e6 c1e3 c8a8 b2b4 c5b4 c3b4 a5c6 e1g1 e8g8 f1c1 d8d7 a3b2 g8h7 c1c2 a8c8 a1c1 c6e7 b2b3 c8c2 c1c2 f8c8 a2a3 c8c2 b3c2 e7c6 c2d1');
INSERT INTO evaluations VALUES(26643,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26643,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(26643,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26643,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(26643,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(26643,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26643,36,'cp 24','d2d4 g4f3 g2f3 e7e6 c2c3 g8e7 h2h4 h7h5 c1g5 d8d7 f1d3 e7g6 f3f4 f8e7 d1f3 e7g5 h4g5 c6e7 b1d2 e8c8 d2b3 b7b6 a2a4 a7a5 b3c1 c7c5 c1e2 c8b8 h1h5 h8h5 f3h5 d8h8 h5f3 h8h4 d3b5 d7c7 e1c1 b8a7 b5d3 e7f5');
INSERT INTO evaluations VALUES(26640,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(26640,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(26640,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(26640,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26640,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26640,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26640,36,'cp 21','g4f3 g2f3 e7e6 c2c3 g7g6 b1d2 c6e7 f1d3 e7f5 h2h4 h7h5 d1b3 d8c8 d2f1 f8e7 d3f5 g6f5 c1e3 e7d8 f1g3 g8e7 b3b5 c7c6 b5e2 e7g6 f3f4 d8h4 g3h5 h4e7 e1c1 c8c7 a2a3 e8c8 d1g1 c8b8 c1b1 d8c8 g1g3 c7a5');
INSERT INTO evaluations VALUES(26641,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(26641,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26641,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(26641,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26641,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26641,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(26641,36,'cp 24','g2f3 e7e6 c2c3 g7g6 b1d2 g8e7 f3f4 e7f5 f1b5 a7a6 b5a4 d8h4 d2f3 h4h3 a4c2 f8e7 c2d3 h3g2 h1g1 g2h3 c1d2 e8f8 d3f1 h3h5 g1h1 f5h4 f3h4 h5d1 e1d1 e7h4 f1d3 a6a5 a2a4 f8e7 d1e2 b7b6 a1g1 e7d7 h2h3 h4e7');
INSERT INTO evaluations VALUES(32124,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(32124,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(32124,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(32124,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(32124,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(32124,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(32124,36,'cp 81','e5d4 d1d4 b8c6 d4a4 g7g6 e4e5 g8e7 c1g5 f8g7 b1c3 g7e5 e1c1 f7f6 g5h6 d8d7 c3d5 e8c8 d5e7 c6e7 a4a7 d7b5 c2c3 e7c6 a7a3 d6d5 a3a8 c6b8 h6e3 c7c6 h2h3 h8e8 h1e1 b5a6 a8a6 b8a6 a2a4 d8d6 c1c2 d6e6 g2g4 e5d6 b2b4 a6c7 f3e2 f6f5 g4f5');
INSERT INTO evaluations VALUES(32195,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(32195,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(32195,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(32195,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(32195,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(32195,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(32195,36,'cp 87','d1d4 b8c6 d4a4 g7g6 e4e5 g8e7 e5d6 d8d6 e1g1 f8g7 c2c3 e8g8 b1a3 a8b8 a3b5 d6f6 b5c7 b8d8 a4e4 c6e5 f3e2 f6b6 c7b5 e7f5 g2g4 f5d6 b5d6 d8d6 a1b1 f7f5 g4f5 g6f5 e4a4 g7h6 c1h6 d6h6 a4f4 b6f6 g1h1 h6h4 f1g1 g8h8 f4g5 h4e4 g5f6 f8f6 e2b5 e5g4');
INSERT INTO evaluations VALUES(32190,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(32190,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(32190,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(32190,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(32190,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(32190,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(32190,36,'cp 90','b8c6 d4a4 g7g6 e4e5 g8e7 e5d6 d8d6 e1g1 f8g7 c2c3 e8g8 b1d2 a8b8 d2e4 d6e5 g2g3 b7b5 a4c2 h7h6 c1e3 f8e8 f3g2 e7d5 e3c5 b5b4 a1d1 b4c3 b2c3 e5f5 g1h1 d5f6 e4f6 f5f6 d1d7 f6c3');
INSERT INTO evaluations VALUES(27041,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(27041,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(27041,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(27041,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(27041,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(27041,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(27041,36,'cp 91','d4a4 g7g6 e4e5 g8e7 e5d6 d8d6 e1g1 f8g7 c2c3 e8g8 b1a3 a8b8 f1d1 d6e6 c1f4 c6e5 f4e5 e6e5 a4a7 e7c6 a7a4 e5e7 a4e4 e7f6 a3c4 f8e8 e4c2 h7h5 a2a4 c6e5 c4e5 e8e5 d1d7 b8e8 c2d2 c7c6 a1d1 e8e7 d7d6 f6g5 g2g3');
INSERT INTO evaluations VALUES(27044,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(27044,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(27044,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(27044,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(27044,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(27044,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(27044,36,'cp 91','g7g6 e4e5 g8e7 e5d6 d8d6 e1g1 f8g7 c2c3 e8g8 b1a3 a8b8 f1d1 d6f6 c1f4 c6e5 f4e5 f6e5 a4a7 e7c6 a7a4 e5e7 a4e4 c6e5 a3c4 e5f3 e4f3 e7e6 c4e3 c7c6 f3f4 b8d8 f4c4 e6e7 d1e1 f8e8 g2g3 h7h5 e1e2 e7f6 a1e1 e8e5 e3g4');
INSERT INTO evaluations VALUES(42017,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(42017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42017,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42017,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(42017,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(42017,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(42017,36,'cp 0','a2a3 b4c3 b2c3 c7c5 c3c4 b8c6 d4c5 f6e4 c4d5 e6d5 f1d3 d8a5 c1d2 a5c5 e1g1 c8g4 d1e1 g4f3 d2b4 c6b4 a3b4 c5c6 g2f3 e4c3 a1a5 c6f6 a5c5 f6f3 c5c3 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(42025,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42025,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(42025,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42025,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42025,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42025,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42025,36,'cp 0','b4c3 b2c3 c7c5 c3c4 b8c6 d4c5 f6e4 c4d5 e6d5 f1d3 d8f6 a1b1 f6c3 f3d2 e4c5 d1e2 e8g8 e1g1 c5d3 c2d3 c3f6 d2f3 f6g6 f3h4 g6f6');
INSERT INTO evaluations VALUES(42116,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42116,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42116,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42116,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(42116,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42116,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(42116,36,'cp 5','b2c3 c7c5 c3c4 b8c6 f1d3 d5c4 d3c4 b7b5 c4e2 c5d4 e1g1 a7a6 f3d4 c6d4 d1d4 d8d4 e3d4 c8d7 c1d2 e8g8 a3a4 f8d8 d2a5 d8c8 f1c1 h7h6 c2c4 b5c4 c1c4 c8c4 e2c4 d7c8 f2f3 f6d5 c4d5 e6d5 a1c1 a8b8 c1c5 c8d7 c5d5 d7a4');
INSERT INTO evaluations VALUES(42115,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42115,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42115,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42115,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42115,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42115,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(42115,36,'cp 5','c7c5 c3c4 b8c6 f1d3 d5c4 d3c4 b7b5 c4e2 c5d4 e1g1 a7a6 f3d4 c6d4 d1d4 d8d4 e3d4 e8g8 a3a4 c8d7 c1d2 f8d8 f1c1 d7e8 c2c4 b5a4 d2e3 a8b8 e2d1 b8b4 e3d2 b4b8 d1a4 d8d4 d2e3 d4d3 a4e8 b8e8 a1a6 e8c8 c4c5 f6d5 c5c6 d5e3 f2e3');
INSERT INTO evaluations VALUES(53145,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53145,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(53145,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53145,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(53145,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53145,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53145,36,'cp 0','c8d7 b5e2 d7f5 c2c3 e7e6 d1b3 d8d7 g1f3 f8d6 f3h4 g8f6 c1g5 f6e4 g5e3 e4f6');
INSERT INTO evaluations VALUES(30930,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30930,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30930,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30930,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30930,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30930,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30930,36,'cp 0','b5e2 d7f5 c2c3 e7e6 d1b3 d8d7 g1f3 f8d6 f3h4 g8f6 h4f5 e6f5 c1g5 f6e4 g5e3 e4f6');
INSERT INTO evaluations VALUES(4350,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4350,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4350,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4350,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(4350,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(4350,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(4350,36,'cp 9','d2d3 a7a6 c1g5 h7h6 g5h4 c5e7 h2h3 d7d6 h4g3 c8e6 c3d5 e6d5 e4d5 b7b5 c4a2 c6d4 f3d4 e5d4 e1g1 f6d7 d1g4 a6a5 a1b1 e7f6 g3f4 g8h7 f4d2 g7g6 c2c3 f6e5 f1e1 f7f5 g4e2 f8e8 g2g3 f5f4 c3d4 e5d4 e2e8 d8e8 e1e8');
INSERT INTO evaluations VALUES(4352,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4352,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(4352,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4352,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4352,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4352,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4352,36,'cp 13','h7h6 h2h3 d7d6 c3a4 c5b6 e1g1 c6e7 a4b6 a7b6 f1e1 c8e6 c4e6 f7e6 c2c3 e7g6 d3d4 e5d4 c3d4 d8d7 a3a4 d6d5 e4d5 f6d5 a1a3 f8f5 d1e2 a8e8 f3h2 d5f4 e2e4 e8f8 h2g4 b6b5 c1f4 g6f4 g4e5 d7d5 e4d5 f4d5');
INSERT INTO evaluations VALUES(4198,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(4198,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(4198,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(4198,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(4198,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4198,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4198,36,'cp 19','h2h3 d7d6 g2g4 f6h7 g4g5 h7g5 f3g5 h6g5 c3d5 c6d4 h1g1 g5g4 h3g4 c7c6 d5e3 d6d5 e4d5 c6d5 c4d5 d4c2 d1c2 c5e3 d5e4 e3c1 a1c1 c8e6 g1h1 a8c8 e4h7 g8h8 c2d2 d8e7 e1f1 c8c1 d2c1 f8c8 c1e3 e7f6 f1g2 f6f4 e3f4 e5f4 h7e4 h8g8 g2f3 c8c2 d3d4 c2b2 d4d5 e6d7 e4f5 d7f5 g4f5 b2a2 d5d6 a2a3 f3f4');
INSERT INTO evaluations VALUES(4196,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(4196,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(4196,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(4196,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(4196,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4196,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(4196,36,'cp 21','d7d6 c3a4 c5b6 e1g1 c8e6 c4e6 f7e6 c2c3 c6e7 a4b6 a7b6 f1e1 e7g6 d3d4 f6d7 c1e3 f8f7 d1e2 d8e7 d4e5 d7e5 f3e5 g6e5 f2f4 e5d7 a1d1 e7h4 g1h2 d7f6 g2g3 h4h5 e2h5 f6h5 e4e5 d6e5 f4e5 a8a5 d1d8 g8h7 d8b8 a5e5');
INSERT INTO evaluations VALUES(3627,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(3627,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(3627,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(3627,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(3627,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(3627,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3627,36,'cp 10','c3a4 c5b6 e1g1 c6e7 f1e1 c8e6 c4e6 f7e6 a4b6 a7b6 c2c3 d8e8 d3d4 e5d4 c3d4 e8g6 d1e2 f6h5 c1d2 h5f4 d2f4 f8f4 e2d3 a8f8 e1e3 c7c5 d4c5 d6c5 a1d1 e7c6 d1d2 g6f6 d3b3 c6e5 f3e5 f4f2 e5f3 f2d2 f3d2');
INSERT INTO evaluations VALUES(3638,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(3638,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(3638,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(3638,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(3638,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(3638,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3638,36,'cp 7','c5b6 e1g1 c8e6 c4e6 f7e6 a4b6 a7b6 c2c3 d8e8 a3a4 b6b5 c1e3 b5a4 a1a4 a8a4 d1a4 c6d4 a4e8 d4f3 g2f3 f8e8 f1a1 g8f7 h3h4 f7e7 a1a7 e8b8 f3f4 e5f4 e3f4 e6e5 f4d2 d6d5 e4d5 f6d5 c3c4 d5f4 d2f4 e5f4 g1g2 g7g5 h4g5 h6g5 a7a5 e7f6 g2f3');
INSERT INTO evaluations VALUES(295,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(295,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(295,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(295,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(295,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(295,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(295,36,'cp 29','c1e3 d7d6 f1e2 f6g6 e1g1 g8f6 f1e1 f6g4 e2g4 c8g4 f2f3 g4h3 d1e2 e8g8 b1d2 c5b6 d2c4 h3e6 c4b6 a7b6 a2a4 c6d4 e3d4 d6d5 d4e5 f8c8 e2f2 f7f6 e5f4 g6f7 f2g3 d5e4 f3e4 h7h6 h2h3 c8d8 f4c7 d8d2');
INSERT INTO evaluations VALUES(11720,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(11720,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(11720,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(11720,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(11720,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(11720,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(11720,36,'cp 104','f1e1 e4c5 b3d5 c5e6 f3e5 c8b7 d2d4 g7g6 d5f3 d7d5 f3g4 c6e5 g4e6 f7e6 e1e5 d8d7 b1d2 f8g7 e5e1 e8g8 d2f3 f8f5 c2c3 c7c5 a2a4 b5a4 a1a4 b7c6 a4a1 c5d4 c3d4 c6b5 a1a3 a6a5 a3e3 a8e8 f3e5 g7e5 e3e5 f5e5 e1e5');
INSERT INTO evaluations VALUES(11721,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(11721,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(11721,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(11721,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(11721,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(11721,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(11721,36,'cp 119','e4c5 b3d5 c5e6 f3e5 c8b7 d2d4 g7g6 d5f3 d7d5 f3g4 c6e5 g4e6 f7e6 e1e5 d8d7 b1d2 f8g7 d2f3 e8g8 c2c3 c7c5 e5g5 f8f3 g2f3 e6e5 c1e3 e5d4 c3d4 a8e8 g5g3 c5c4 d1d2 d7e7 a2a4 b5b4 e3g5 e7f8 a1e1 e8e1 d2e1 g7d4 g5e7 f8f5 e1b4 f5b1 g1g2');
INSERT INTO evaluations VALUES(11725,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(11725,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(11725,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(11725,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(11725,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(11725,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(11725,36,'cp 125','b3d5 c8b7 d2d4 c5e6 f3e5 g7g6 d5f3 f8g7 e5c6 d7c6 c2c3 e8g8 b1d2 a6a5 a2a4 d8c8 d2b3 c6c5 b3c5 e6c5 d4c5 f8d8 d1e2 b7f3 e2f3 a8b8 c5c6 d8e8 e1d1 e8d8 d1d8 c8d8 a4b5 b8b5 g2g3 h7h5 g1g2 h5h4 a1a4 b5f5');
INSERT INTO evaluations VALUES(18386,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18386,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18386,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18386,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18386,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(18386,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(18386,36,'cp 0','d7d5 f1d3 g8f6 e1g1 f8e7 f1e1 e8g8 c1g5 f8e8 d1e2 h7h6 g5f4 f6d7 e2g4 d7f6');
INSERT INTO evaluations VALUES(18486,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18486,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18486,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18486,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18486,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18486,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18486,36,'cp 0','f1d3 g8f6 e1g1 f8e7 d1e2 e8g8 f1e1 d8c7 e4d5 c6d5 c3b5 c7b8 c2c3 f6e4 c3c4 e4f6 b2b3 c8b7 c4d5 b7d5 c1b2 b8b7 b5d6 e7d6 b2f6 g7f6 e2g4 g8h8 g4h4 f6f5 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(52429,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(52429,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(52429,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(52429,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(52429,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(52429,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(52429,36,'cp -82','c5d4 f4b8 a8b8 d1d4 d8a5 b1d2 e7e5 d4e3 f8e7 g1f3 d7d6 b2b4 a5b4 e3a7 b4c3 a1b1 c3c7 e2e3 g8f6 f1b5 c8d7 b5d7 f6d7 c2c4 e8g8 a7a4 c7c6 a4c6 b7c6 e1g1 f7f5 b1b8 f8b8 f1b1 b8b1 d2b1 e7d8 f3d2 d7c5 d2b3 c5a4 c4c5 d6d5 g2g3 g8f7 f2f4 e5f4 e3f4 d8e7');
INSERT INTO evaluations VALUES(52521,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(52521,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(52521,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(52521,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(52521,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(52521,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(52521,36,'cp -67','c2c3 d4c3 b1c3 d7d6 e2e4 e7e5 f4e3 g8f6 f1c4 f8e7 g1e2 b8c6 e1g1 c8e6 d1d3 e8g8 a2a3 a7a6 c3d5 e6d5 c4d5 f6d5 d3d5 a8c8 a1d1 e7g5 e3g5 d8g5 d5b3 g5e7 d1d5 f8d8 f1d1 d8d7 d5d2 c8d8 b3d3 h7h6 e2c3 c6d4 c3d5 e7h4 d3e3 d4e6');
INSERT INTO evaluations VALUES(12015,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12015,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(12015,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(12015,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(12015,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(12015,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(12015,36,'cp -6','d7d5 e4d5 c6d5 c1f4 a7a6 d1d4 e7e6 g2g4 f8e7 g4g5 f6h5 f4e5 e8g8 h1g1 f7f6 g5f6 e7f6 f1d3 f6e5 d4e5 d8h4 c3e2 g7g6 e1c1 h5f4 f2f3 h4f6 e5f6 f8f6 e2f4 f6f4 d3e2 g8f7 g1g4 f4f5 b2b3 f7e7 c1b2 a6a5 g4h4 f5f7 f3f4');
INSERT INTO evaluations VALUES(12028,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12028,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12028,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12028,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(12028,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12028,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(12028,36,'cp 13','f1d3 e7e6 d1e2 f8e7 e1g1 e8g8 f1e1 a7a5 b2b3 f8e8 c1d2 e7b4 a2a3 b4c5 e4e5 f6d7 c3a4 c5f8 d2f4 c8b7 c2c4 b7a6 e2g4 f7f5 e5f6 d8f6 f4g3 f6f7 g4e2 e6e5 c4d5 a6d3 e2d3 c6d5 a1d1 f8a3 d3d5 f7d5');
INSERT INTO evaluations VALUES(37023,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(37023,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(37023,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(37023,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(37023,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(37023,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(37023,36,'cp 37','e2e3 h7h6 g5h4 c7c5 c4d5 e6d5 d4c5 b8d7 d1d4 b4c5 d4d2 d8b6 c3a4 b6d6 h4g3 d6e7 a4c5 d7c5 f1e2 c5e4 d2d4 e4g3 h2g3 c8f5 e1g1 f8c8 f1c1 a7a5 a2a3 e7d8 d4a4 c8c1 a1c1 a8c8 c1c8 f5c8 e2d3 c8g4 f3d4 g4d7 a4b3 b7b6');
INSERT INTO evaluations VALUES(37022,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(37022,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(37022,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(37022,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(37022,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(37022,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(37022,36,'cp 36','h7h6 g5h4 c7c5 c4d5 e6d5 d4c5 b8d7 a1c1 d8a5 f3d2 g7g5 h4g3 d7c5 f1e2 b4c3 b2c3 f8e8 e1g1 c8d7 d2b3 c5b3 a2b3 a5b6 e2d3 a7a6 d1c2 f6e4 c3c4 e4g3 c4c5 g3e2 c2e2 b6f6 b3b4 a8d8 c1d1 d7c6');
INSERT INTO evaluations VALUES(36912,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(36912,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(36912,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(36912,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36912,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(36912,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(36912,36,'cp 35','g5h4 c7c5 c4d5 e6d5 d4c5 b8d7 a1c1 d8a5 f3d2 g7g5 h4g3 d7c5 f1e2 b4c3 b2c3 f6e4 g3e5 a5a2 c1c2 a2a5 e1g1 f8e8 e5d4 c5e6 d2e4 d5e4 e2c4 e6d4 c3d4 c8e6 h2h3 e6c4 c2c4 a5b6 c4c5 a7a5 h3h4 g5h4 d1g4 b6g6 g4h4 e8e6 c5a5 a8a5 h4d8 g8g7 d8a5 e6d6 a5c5');
INSERT INTO evaluations VALUES(36915,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(36915,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(36915,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(36915,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36915,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36915,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(36915,36,'cp 37','c7c5 c4d5 e6d5 d4c5 b8d7 a1c1 d8a5 f3d2 g7g5 h4g3 d7c5 f1e2 b4c3 b2c3 f8e8 e1g1 c8f5 d2b3 c5b3 a2b3 a8c8 d1d4 a5b6 g3e5 f6d7 d4b6 a7b6 e5d4 d7c5 e2d1 f5d3 f1e1 c8c6 c1a1 c5e6 d1g4 e6d4 c3d4 c6c3 e1c1 c3b3 c1c7 g8g7 h2h3 e8b8 g4h5 d3g6 h5f3 g6e4 f3e4');
INSERT INTO evaluations VALUES(55433,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(55433,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(55433,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(55433,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(55433,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(55433,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(55433,36,'cp 132','b1c3 g8f6 c1e3 a7a6 d1d2 c7c5 d4b3 f8e7 e1c1 e8g8 f2f3 b7b5 h2h4 b5b4 c3a4 c8d7 e4e5 d6e5 a4c5 d7c6 d2b4 a6a5 d1d8 a5b4 d8f8 e7f8 c1b1 e5e4 f1e2 e4f3 e2f3 c6f3 g2f3 b8c6 c5d3 a8e8 e3f4 f6h5 f4d2 f8d6');
INSERT INTO evaluations VALUES(55432,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(55432,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(55432,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(55432,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(55432,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(55432,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(55432,36,'cp 114','f8e7 c1e3 g8f6 d1d2 b8c6 d4c6 b7c6 e1c1 a8b8 h2h3 e8g8 h1g1 c8e6 g2g4 d6d5 e4e5 f6e4 c3e4 d5e4 d2c3 e6d5 f1c4 f8e8 c1b1 e7b4 c3d4 d8e7 c4d5 c6d5 d4d5 b4c3 b2b3 e7a3 e3c1 a3b4 a2a3 b4b6 d5e4 c3e5');
INSERT INTO evaluations VALUES(4849,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(4849,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4849,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(4849,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(4849,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4849,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4849,36,'cp 13','b5c6 d7c6 f3e5 c5f2 e1f2 d8d4 f2e1 d4e5 d1f3 g8e7 d2d3 e7g6 f3g3 e5g3 h2g3 c8g4 c1f4 e8d7 a2a3 a6a5 e1d2 h7h5 b2b4 g6f8 b4b5 c6b5 c3d5 f8e6 a1b1 d7c6 d5c3 b7b6 b1b5 a8d8 f4e3 f7f6 a3a4 e6d4 b5d5 d4e6 d5d8 h8d8');
INSERT INTO evaluations VALUES(4843,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(4843,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4843,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(4843,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(4843,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4843,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4843,36,'cp 11','d7c6 f3e5 c5f2 e1f2 d8d4 f2e1 d4e5 d1f3 g8e7 d2d3 e7g6 c1e3 c8e6 e1d2 b7b6 b2b4 a6a5 a2a3 e5d6 f3g3 d6d7 h2h4 h7h5 g3g5 e6g4 b4b5 c6b5 g5b5 d7b5 c3b5 e8c8 e3f2 h8e8 h1e1 f7f6 b5c3 g4d7 a3a4 d7c6 f2g3 c8b7 a1b1');
INSERT INTO evaluations VALUES(4782,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(4782,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(4782,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4782,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4782,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4782,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(4782,36,'cp 11','f3e5 c5f2 e1f2 d8d4 f2e1 d4e5 d2d3 g8e7 d1f3 e7g6 f3g3 e5g3 h2g3 c8g4 c1f4 e8d7 a2a3 g6f8 b2b4 a6a5 a1b1 a5b4 a3b4 h7h5 e1d2 f8e6 b4b5 a8a5 b5c6 b7c6 h1f1 f7f6 b1b7 h8a8 f4e3 c6c5 f1b1 c5c4 d3c4 a8d8 c3d5 a5a4');
INSERT INTO evaluations VALUES(4775,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(4775,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4775,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4775,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(4775,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4775,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4775,36,'cp 8','c5f2 e1f2 d8d4 f2e1 d4e5 d1f3 g8e7 d2d3 e7g6 f3g3 e5g3 h2g3 c8g4 c1f4 e8d7 e1d2 a6a5 a2a3 g6f8 b2b4 f8e6 f4e3 a5b4 a3b4 h7h5 c3a4 b7b6 c2c4 f7f5 e4f5 g4f5 a4c3 f5g6 h1f1 h8e8 f1b1 a8d8 a1a7 d7c8 a7a8 c8d7 a8d8 d7d8');
INSERT INTO evaluations VALUES(15091,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(15091,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(15091,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(15091,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(15091,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(15091,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(15091,36,'cp 124','e4d5 f6d5 d2d4 c8g4 h2h3 g4f3 d1f3 c6d4 f3d5 d8d5 c4d5 d4c2 c1e3 c2a1 d5e4 e8c8 b1c3 h7h5 f1a1 g7g6 e4d5 f7f5 a1e1 f8h6 d5e6 c8b8 e6f7 h6e3 e1e3 d8d6 e3e5 c7c6 c3e2 h8f8 e5e7 d6d1 g1h2 d1d2 f7g6 d2b2 f2f4 h5h4 e2d4 b2a2 d4f5');
INSERT INTO evaluations VALUES(14989,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(14989,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(14989,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(14989,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(14989,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(14989,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(14989,36,'cp 116','f6d5 d2d4 c8g4 h2h3 g4f3 d1f3 c6d4 f3d5 d8d5 c4d5 d4c2 d5b7 a8b8 b7c6 e8d8 b1c3 c2a1 f1d1 f8d6 c6e4 g7g6 g2g4 h7h5 c1e3 h5g4 h3g4 b8b2 d1a1 b2b4 a1c1 b4c4 e4f3 f7f5 f3e2 c4b4 c3d5 b4b2 e3g5 d8c8 e2a6 c8b8 d5e3 b2b6 a6f1 b6b2 a2a4 f5g4 g5f6 h8e8 a4a5 b2b4 a5a6 g4g3 f2g3 e5e4 f6c3');
INSERT INTO evaluations VALUES(15321,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(15321,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(15321,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(15321,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(15321,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(15321,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(15321,36,'cp 121','d2d4 d5b6 c4b5 e5d4 f1e1 f8e7 b5c6 b7c6 f3d4 c8d7 d1f3 e8g8 d4c6 d7c6 f3c6 e7f6 b1c3 d8d6 c6d6 c7d6 c3e4 f6e5 c2c3 f7f5 e4d2 e5f6 a2a4 a7a5 e1d1 a8b8 g2g3 f8e8 g1f1 e8c8 h2h4 g8f7 a1a3 f7e6 a3a2 f5f4 g3f4 f6h4 d2e4 b6c4 f4f5 e6f5 e4d6 c4d6 d1d6');
INSERT INTO evaluations VALUES(15329,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(15329,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(15329,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(15329,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(15329,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(15329,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(15329,36,'cp 122','d5b6 c4b5 e5d4 f1e1 f8e7 f3d4 c8d7 b5c6 b7c6 d1e2 c6c5 d4b5 e8g8 e2e7 d8e7 e1e7 f8e8 e7e8 a8e8 c1e3 d7b5 b1c3 b5d7 e3c5 d7f5 c3b5 b6a4 c5a3 f5c2 b2b3 a4b6 b5c7 e8c8 c7b5 a7a5 a1c1 a5a4 b5d4 a4b3 a2b3 c2e4 c1c8 b6c8 b3b4 f7f6 b4b5 e4b7 a3c5');
INSERT INTO evaluations VALUES(42547,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(42547,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(42547,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42547,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(42547,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42547,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42547,36,'cp -12','d7d6 g1f3 e8g8 e1g1 c7c6 d2d4 e5d4 f3d4 f8e8 f1e1 a7a5 c1e3 b8d7 d4f5 d7e5 e3c5 c8f5 c4f1 f6e4 c3e4 d6c5 e4c5 b7b6 c5e4 f5g6 d1e2 e5d7 e2f3 d8c7 e1e2 b6b5 a1e1 a8d8 e4c5 e8e2 f3e2 h7h6 c5d7 c7d7 c2c4 b5b4 e2e3 d7c7 c4c5');
INSERT INTO evaluations VALUES(41857,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(41857,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(41857,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(41857,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(41857,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41857,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41857,36,'cp 0','g1f3 e8g8 e1g1 c8e6 c4b3 e6b3 a2b3 f8e8 d2d3 a7a5 c1e3 c5e3 f2e3 b8d7 d1e1 c7c6 e1g3 f6h5 g3f2 h5f6');
INSERT INTO evaluations VALUES(41853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41853,36,'cp 0','e8g8 e1g1 b8d7 a2a4 c7c6 d2d3 h7h6 c1e3 c5b4 c3a2 b4c5');
INSERT INTO evaluations VALUES(36989,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36989,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36989,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36989,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36989,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36989,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36989,36,'cp 0','e1g1 c8e6 c4b3 e6b3 a2b3 f8e8 d2d3 a7a5 c1e3 c5e3 f2e3 b8d7 d1e1 c7c6 e1g3 f6h5 g3f2 h5f6');
INSERT INTO evaluations VALUES(36988,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36988,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36988,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36988,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36988,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36988,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36988,36,'cp -10','c7c6 d2d4 e5d4 f3d4 f8e8 f1e1 a7a5 c1e3 b8d7 d4f5 d7e5 e3c5 c8f5 c4f1 f6e4 c3e4 d6c5 e4c5 b7b6 c5e4 f5g6 e1e3 e8e6 d1d8 a8d8 a1e1 e5d7 f1c4 e6e4 e3e4 g6e4 e1e4 b6b5 c4e2 g8f8 e4d4 f8e7 a2a4 b5b4 b2b3 g7g6');
INSERT INTO evaluations VALUES(55927,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(55927,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(55927,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(55927,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55927,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(55927,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(55927,36,'cp -34','c7c5 b2b3 f8e7 c1b2 g8f6 f1b5 b8d7 b5d3 b7b6 c2c4 c8b7 e1g1 e8g8 b1c3 d5d4 c3e2 d8c7 e3d4 c5d4 b2d4 b7f3 g2f3 e7c5 d4e3 e6e5 g1h1 f8e8 d1e1 a8d8 a1d1 c5e3 d2e3 d7c5 d3c2 d8d1 c2d1 c7e7 e3e4 c5d3 e1h4 a7a6 d1c2 f6e4 h4e7 e4f2 h1g1 e8e7');
INSERT INTO evaluations VALUES(54263,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(54263,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(54263,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(54263,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(54263,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(54263,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(54263,36,'cp -28','b2b3 f8e7 c1b2 g8f6 g2g3 b8c6 a2a4 d8b6 d1c1 b6c7 f1g2 e8g8 e1g1 c6b4 d2d3 b7b6 b1a3 c8b7 c1d2 a7a6 b2e5 c7d7 d2e2 a8d8 e5b2 d7c7 b2e5 c7c8 e5b2 f6e8 h2h3 b4c6 f3e5 e8d6 e5c6 b7c6 c2c4 c8b7 a3c2 d5c4 g2c6 b7c6');
INSERT INTO evaluations VALUES(16218,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(16218,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(16218,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(16218,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(16218,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(16218,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(16218,36,'cp 151','d1d4 d7d6 c1f4 f8e7 h2h4 c8e6 e1c1 e8g8 c4e6 f7e6 e4e5 d6e5 d4e5 d8c8 e5e2 f8f7 g2g3 c7c6 h1e1 f6d5 f4d2 e6e5 e2e5 e7f6 e5e2 f6c3 d2c3 d5c3 b2c3 c8f8 e2c4 g8h8 f2f4 f7e7 c1b2 a8e8 e1e7 f8e7 a2a3 a7a6 d1d4 e7e2 c4e2 e8e2');
INSERT INTO evaluations VALUES(54372,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(54372,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(54372,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54372,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(54372,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(54372,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(54372,36,'cp 22','g1f3 b8c6 c1f4 f8d6 f4d6 d8d6 f1d3 g8e7 e1g1 e8g8 f1e1 a7a6 c3e2 f7f6 h2h3 c8d7 d1d2 d7e8 d2f4 a8d8 f4d6 d8d6 e2c1 e8g6 d3g6 e7g6 c1d3 g8f7 h3h4 f8c8 c2c3 c6a5 h4h5 g6e7 d3c5 d6b6 b2b3 h7h6');
INSERT INTO evaluations VALUES(56694,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(56694,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(56694,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(56694,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(56694,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(56694,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(56694,36,'cp 195','c8e6 f1b5 b8d7 d1e2 g7g6 b5a4 f8h6 c1h6 g8h6 e2e3 h6f7 h2h3 e6c4 f3d2 c4a6 e1c1 c7c6 a4b3 h8f8 h3h4 d8b6 d2f3 e8c8 c3a4 b6a5 d4e5 d6e5 e3a7 c8c7 a7e3 a6b5 a4c3 b5a6 h4h5 g6g5 f3d2 a5b6 e3b6 d7b6 f2f3 f7d6');
INSERT INTO evaluations VALUES(39629,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39629,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39629,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39629,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39629,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39629,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39629,36,'cp 0','c7c5 f1e2 b8c6 d2d3 d7d6 c2c3 g8f6 a2a4 e8g8 e1g1 b7b6 d1e1 c5c4 b1d2 c4d3 e2d3 e7e6 d2b3 c8b7 c1e3 c6e7 e3d4 e7c6');
INSERT INTO evaluations VALUES(45222,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(45222,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(45222,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(45222,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45222,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(45222,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(45222,36,'cp 13','b1c3 b8c6 f1b5 c8d7 d2d3 e7e6 e1g1 a7a6 b5c4 d8c7 d1e1 c6d4 c4b3 d4f3 f1f3 f8e7 e1e2 e8c8 f4f5 e6f5 e4f5 d7c6 b3f7 c6f3 e2f3 c8b8 c1f4 h8f8 f7e6 c7c6 c3e4 b8a8 a1f1 f8e8 e4g5 c6f3 f1f3 d6d5 c2c3');
INSERT INTO evaluations VALUES(6889,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(6889,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(6889,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(6889,33,'cp 235',NULL);
INSERT INTO evaluations VALUES(6889,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(6889,35,'cp 245',NULL);
INSERT INTO evaluations VALUES(6889,36,'cp 231','d7d6 c1e3 g8f6 e1c1 c8d7 f2f3 d8e7 f1e2 e5c6 d4d2 e8c8 h1e1 h8e8 e3g5 e7f8 g5h4 c6e7 c3b5 e7c6 a2a3 c8b8 g2g4 h7h6 b5c3 c6e7 h4f2 e7c8 a3a4 d7c6 c3b5 c6b5 e2b5 c7c6 b5f1 d6d5 e4d5 e8e1 d2e1 f6d5');
INSERT INTO evaluations VALUES(6102,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(6102,31,'cp 263',NULL);
INSERT INTO evaluations VALUES(6102,32,'cp 247',NULL);
INSERT INTO evaluations VALUES(6102,33,'cp 268',NULL);
INSERT INTO evaluations VALUES(6102,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(6102,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(6102,36,'cp 240','c1e3 g8e7 d4d2 e8g8 e1c1 e5g4 e3d4 e7c6 h2h3 g4f6 d4e3 f8e8 f2f3 a7a6 g2g4 b7b5 g4g5 f6h5 h1g1 c8b7 f1e2 g7g6 f3f4 b5b4 c3d5 c6e7 d5f6 h5f6 g5f6 e7c6 f4f5 e8e4 e2d3 c6e5 f5g6 h7g6 d3e4 b7e4 d2b4 d6d5 b4e7 d8e8 e7e8 a8e8 e3d4 e5f3 g1g3 e8e6');
INSERT INTO evaluations VALUES(26620,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(26620,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(26620,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(26620,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(26620,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(26620,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(26620,36,'cp 71','f1g2 g4f3 g2f3 g8f6 c2c4 d5c4 b1c3 e7e6 f3c6 b7c6 d1a4 d8d7 a4c4 f6d5 e1g1 d5b6 c4a6 f8d6 e2e4 e6e5 d4e5 d6e5 f1d1 d7e6 c3a4 e8g8 a4c5 e6c4 c1e3 c4a6 c5a6 b6c4 e3d4 e5d4 d1d4 c4b2');
INSERT INTO evaluations VALUES(26621,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(26621,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(26621,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(26621,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(26621,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(26621,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(26621,36,'cp 58','g4f3 g2f3 d8d7 e1g1 g8f6 f3g2 e8c8 c2c3 e7e5 d4e5 c6e5 b1d2 h7h5 h2h4 e5g4 d2b3 c8b8 c1g5 d8e8 d1d3 f8d6 a1d1 f6e4 g5f4 e4f6 f4d6 d7d6 c3c4 g4e5 d3c2 e5c4 g2d5 f6d5 c2c4 d5b6 c4c2 d6e6');
INSERT INTO evaluations VALUES(26632,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(26632,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(26632,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(26632,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(26632,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(26632,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(26632,36,'cp 62','g2f3 g8f6 c2c4 d5c4 b1c3 e7e6 f3c6 b7c6 d1a4 d8d7 a4c4 f6d5 e1g1 d5b6 c4a6 f8d6 f1d1 e8g8 c1d2 f7f5 a1c1 a8b8 b2b3 b8e8 d2f4 d6f4 g3f4 f8f6 e2e3 f6h6 g1h1 e8f8 f2f3 d7e7 a6e2 a7a5 e2d3');
INSERT INTO evaluations VALUES(26631,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(26631,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26631,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(26631,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(26631,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(26631,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(26631,36,'cp 60','g8f6 c2c4 e7e5 c4d5 c6d4 f3g2 f8b4 c1d2 b4d2 b1d2 e8g8 e2e3 d4b5 a1c1 f8e8 a2a4 b5d6 e3e4 h7h5 e1g1 h5h4 f1e1 a7a5 b2b3 g7g6 d1c2 e8e7 c2b2 d8f8 f2f4 f8g7 d2f3 h4g3 f4e5 f6d7 h2g3 d7e5 f3e5 e7e5 c1c7 a8e8');
INSERT INTO evaluations VALUES(48939,30,'cp 247',NULL);
INSERT INTO evaluations VALUES(48939,31,'cp 270',NULL);
INSERT INTO evaluations VALUES(48939,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(48939,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(48939,34,'cp 252',NULL);
INSERT INTO evaluations VALUES(48939,35,'cp 274',NULL);
INSERT INTO evaluations VALUES(48939,36,'cp 268','d1d8 e8d8 f3e5 c8e6 f2f3 f6d7 e5d3 b8c6 d3f4 a7a5 a2a4 f8c5 b1a3 h7h6 c1d2 c6e5 f4e6 f7e6 a3c4 e5c4 f1c4 d8e7 e1c1 c5d6 d2c3 h8f8 h2h4 f8f7 c1b1 d7e5 c4b5 c7c6 b5e2 b7b6 b2b3');
INSERT INTO evaluations VALUES(35397,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(35397,31,'cp 277',NULL);
INSERT INTO evaluations VALUES(35397,32,'cp 276',NULL);
INSERT INTO evaluations VALUES(35397,33,'cp 253',NULL);
INSERT INTO evaluations VALUES(35397,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(35397,35,'cp 270',NULL);
INSERT INTO evaluations VALUES(35397,36,'cp 244','e8d8 f3e5 c8e6 f2f3 f6d7 e5d3 b8c6 d3f4 a7a5 f4e6 f7e6 a2a4 f8c5 b1a3 h7h6 a3c4 c6e5 c4e5 d7e5 c1f4 c5d6 f4e3 d8e7 h2h4 e5d7 e1c1 h8f8 c1b1 a8d8 h4h5 b7b6 e3f2 d6f4 f1b5 d7e5 d1d8 f8d8 f2h4 f4g5 h4g3');
INSERT INTO evaluations VALUES(35430,30,'cp 267',NULL);
INSERT INTO evaluations VALUES(35430,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(35430,32,'cp 269',NULL);
INSERT INTO evaluations VALUES(35430,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(35430,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(35430,35,'cp 257',NULL);
INSERT INTO evaluations VALUES(35430,36,'cp 267','f3e5 c8e6 f2f3 f6d7 e5d3 b8c6 d3f4 a7a5 a2a4 f8c5 f4e6 f7e6 c2c3 c6e5 b1a3 h7h6 a3b5 c5b6 h2h4 d7c5 b5d4 d8e7 e1d2 g7g5 d2c2 g5g4 f1e2 g4f3 g2f3 a8g8 c1f4 e5f7 h1g1 g8g1 a1g1 c5a4 g1g6 b6d4 c3d4 h8d8 f4e3 a4b6 h4h5 a5a4 c2c3');
INSERT INTO evaluations VALUES(35421,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(35421,31,'cp 259',NULL);
INSERT INTO evaluations VALUES(35421,32,'cp 252',NULL);
INSERT INTO evaluations VALUES(35421,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(35421,34,'cp 241',NULL);
INSERT INTO evaluations VALUES(35421,35,'cp 271',NULL);
INSERT INTO evaluations VALUES(35421,36,'cp 257','c8e6 f2f3 f6d7 e5d3 b8c6 d3f4 a7a5 f4e6 f7e6 a2a4 h7h6 b1a3 c6d4 e1d1 g7g5 a3c4 f8c5 c1d2 b7b6 d2e1 c5d6 c4d6 c7d6 e1f2 e6e5 c2c3 d4b3 a1b1 b3c5 f1b5 g5g4 d1c2 d8c7 b1f1 g4f3 g2f3 a8g8 h1g1 g8g1');
INSERT INTO evaluations VALUES(27722,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(27722,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(27722,32,'cp 256',NULL);
INSERT INTO evaluations VALUES(27722,33,'cp 265',NULL);
INSERT INTO evaluations VALUES(27722,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(27722,35,'cp 271',NULL);
INSERT INTO evaluations VALUES(27722,36,'cp 245','f2f3 f6d7 e5d3 b8c6 d3f4 a7a5 a2a4 f8c5 f4e6 f7e6 c2c3 d8e7 f1b5 c6e5 b5d7 e7d7 e1e2 g7g5 h1d1 d7c8 b1d2 h8d8 d2b3 d8d1 e2d1 c5e7 d1e2 b7b6 g2g3 h7h6 f3f4 g5f4 g3f4 e5c4 e4e5 c8b7 b3d4');
INSERT INTO evaluations VALUES(57441,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(57441,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(57441,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(57441,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(57441,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(57441,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(57441,36,'cp 93','b8c6 f1c4 f8c5 b1c3 d7d6 c3a4 d8e7 a4c5 d6c5 d2d3 c8e6 c4b5 e6d7 e1g1 e8c8 d1e1 c6d4 f3d4 e5d4 b5d7 d8d7 f4f5 h7h5 e1a5 c8b8 b2b4 c5c4 a5b5 c4c3 c1f4 g8h6 h2h3 h5h4 a2a4 h6f7 a4a5 a7a6 b5a4 f7d6 f4d6 d7d6 b4b5 a6b5 a4b5 e7e5 b5e5');
INSERT INTO evaluations VALUES(22723,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(22723,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(22723,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(22723,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(22723,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(22723,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(22723,36,'cp -46','f8c5 g1f3 d7d6 c1e3 g8e7 a2a3 a7a5 d3d4 e5d4 f3d4 e8g8 b1c3 c6d4 e3d4 b7b6 f1e2 c8b7 e1g1 e7c6 d4c5 d6c5 f2f3 c6d4 c3b5 f8e8 b5d4 d8d4 d1d4 c5d4 e2d3 c7c5 a1e1 f7f6 h2h4 b7c6 g2g4');
INSERT INTO evaluations VALUES(6452,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(6452,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(6452,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(6452,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(6452,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(6452,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(6452,36,'cp -44','g1f3 d7d6 c1e3 g8e7 a2a3 a7a5 d3d4 e5d4 f3d4 e8g8 b1c3 c6d4 e3d4 b7b6 f1e2 e7c6 d4c5 b6c5 e1g1 f8e8 c3b5 c8d7 e2d3 a5a4 f2f4 c6d4 b5d4 c5d4 c4c5 d6c5 a1c1 a8b8 c1c5 b8b2 c5d5 c7c6 d5d4 d8b6');
INSERT INTO evaluations VALUES(40380,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40380,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40380,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40380,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(40380,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(40380,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40380,36,'cp 7','e8g8 f1e1 b8c6 c4b3 c6d4 d1d4 a7a6 a2a4 f6d7 d4b4 a6a5 b4a3 d7c5 c1e3 g7c3 b2c3 c5e4 e3h6 e4c5 h6f8 d8f8 b3c4 c8d7 c4b5 d7b5 a4b5 a5a4 h2h3 e7e6 c3c4 f8e7 a3f3 g8g7 e1d1 a8d8 f3c3 g7g8 f2f3 e7c7 c3d2 g8g7 f3f4 d8d7 d2d4 g7g8');
INSERT INTO evaluations VALUES(35994,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(35994,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(35994,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(35994,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(35994,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(35994,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(35994,36,'cp 12','c4b3 b8c6 f1e1 c6d4 d1d4 c8d7 d4d3 d7c6 c1e3 f6d7 a1d1 a8c8 d3e2 a7a5 e3d4 g7d4 d1d4 b7b5 a2a3 d8b6 e2d2 d7e5 h2h3 c6b7 e1d1 c8c5 b3d5 f8c8 g1h1 b7d5 c3d5 b6d8 d5e3 e5c4 e3c4 c5c4 c2c3 c4d4');
INSERT INTO evaluations VALUES(53543,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(53543,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53543,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(53543,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53543,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53543,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(53543,36,'cp 12','e6d5 c1f4 g8f6 e2e3 c8f5 f1d3 f5d3 d1d3 f6h5 g1f3 b8d7 e1g1 f8e7 h2h3 e8g8 f4h2 f8e8 d3c2 h5f6 f1e1 e7b4 a2a3 b4f8 h2f4 a7a5 a1c1 h7h6 f4h2 d8b6 f3e5 f8d6 e5d3 d6h2 g1h2 f6e4 c3e4 e8e4 b2b3');
INSERT INTO evaluations VALUES(50836,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50836,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(50836,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(50836,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50836,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(50836,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(50836,36,'cp 36','g1f3 g8f6 b1d2 e7e6 c2c3 c7c5 e2e3 f8d6 f4g3 d8c7 d4c5 c7c5 f1d3 c5b6 d1c2 c8d7 c3c4 e8g8 c4d5 e6d5 a1c1 f8e8 a2a3 d6g3 h2g3 h7h6 e1g1 b8c6 b2b4 a8c8 c2b1 d7g4 d2b3 f6d7 b3c5 e8e7 c5d7');
INSERT INTO evaluations VALUES(50835,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(50835,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(50835,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(50835,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(50835,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(50835,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(50835,36,'cp 28','e7e6 e2e3 c7c5 b1d2 g8f6 c2c3 f8d6 f4g3 d8b6 f3e5 d6e7 d1b3 b6b3 a2b3 f6h5 f1e2 h5g3 h2g3 b8d7 e5d3 h7h6 b3b4 c5b4 d3b4 e7b4 c3b4 e8e7 g3g4 d7f6 g4g5 f6e8 d2f3 c8d7 b4b5 a6a5 f3e5 e8d6 g5h6 h8h6 h1h6');
INSERT INTO evaluations VALUES(50958,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50958,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(50958,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(50958,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(50958,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(50958,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(50958,36,'cp 33','e2e3 c7c5 b1d2 g8f6 c2c3 b8d7 a2a4 f8e7 h2h4 e8g8 f1d3 d8b6 d1b1 e7d6 f3e5 c5d4 c3d4 d6b4 e5f3 d7b8 e1g1 b8c6 b1d1 a6a5 f3e5 f6d7 d2f3 d7e5 f3e5 c6e5 f4e5 c8d7 d1h5 g7g6 h5f3 b4e7');
INSERT INTO evaluations VALUES(50957,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(50957,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(50957,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(50957,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(50957,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(50957,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(50957,36,'cp 28','g8f6 c2c3 c7c5 b1d2 b8d7 a2a4 f8e7 h2h4 e8g8 f1d3 d8b6 d1c2 b6c6 d3e2 e7d6 f3e5 c6c7 d2f3 f8e8 a4a5 f6e4 e2d3 c5d4 e3d4 f7f6 e5d7 c8d7 f4d6 c7d6 e1g1 d7b5 d3b5 a6b5 c2e2 b5b4 f3d2 b4c3 d2e4 d5e4 b2c3 d6c7 e2e4 a8a5 a1a5');
INSERT INTO evaluations VALUES(52788,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(52788,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(52788,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(52788,33,'cp -129',NULL);
INSERT INTO evaluations VALUES(52788,34,'cp -133',NULL);
INSERT INTO evaluations VALUES(52788,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(52788,36,'cp -86','d5e4 f3g5 c8f5 g2g4 f5g6 b1c3 g8f6 f1g2 h7h5 h2h3 h5g4 h3g4 h8h1 g2h1 d8a5 c1d2 e4e3 d2e3 e7e6 e3d2 a5b6 d1e2 g6c2 d2c1 c2g6 g5e6 f7e6 e2e6 f8e7 g4g5 b8d7 g5f6 d7f6 e6b3 b6b3 a2b3 g6f7 d4d5 f6d5 h1d5 f7d5 c3d5 c6d5 c1e3');
INSERT INTO evaluations VALUES(52903,30,'cp -112',NULL);
INSERT INTO evaluations VALUES(52903,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(52903,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(52903,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(52903,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(52903,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(52903,36,'cp -98','f3g5 c8f5 g2g4 f5g6 b1c3 g8f6 f1g2 h7h5 h2h3 h5g4 h3g4 h8h1 g2h1 d8c7 h1e4 f6e4 c3e4 b8a6 c2c3 e8c8 f2f4 e7e6 d1f3 c7d7 e4g3 f7f6 f4f5 e6f5 g4f5 f6g5 f5g6 d8e8 e1d1 g5g4 f3f7 c6c5 c1d2 c5d4 f7d7 c8d7 d1c2 f8d6 g3f5 e8e6 f5d4 e6g6 d4f5 g4g3 a1g1 g6g4 f5d6 d7d6');
INSERT INTO evaluations VALUES(52862,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(52862,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(52862,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(52862,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(52862,34,'cp -113',NULL);
INSERT INTO evaluations VALUES(52862,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(52862,36,'cp -97','c8f5 g2g4 f5g6 b1c3 g8f6 f1g2 h7h5 h2h3 b8d7 d1e2 d8b6 g5e4 e8c8 g4g5 f6e4 g2e4 g6e4 e2e4 e7e5 d4d5 f7f5 e4f3 g7g6 d5c6 b6c6 f3c6 b7c6 c3a4 f8b4 c2c3 b4e7 h3h4 c8b7 c1e3 f5f4 e3d2 d7b6 a4b6 a7b6 a2a4 h8e8');
INSERT INTO evaluations VALUES(57733,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(57733,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(57733,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(57733,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(57733,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(57733,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(57733,36,'cp 135','e2e4 e7e6 b1c3 d7d5 g1f3 f8b4 e4d5 e6d5 f1d3 b8c6 a2a3 b4c3 b2c3 g8e7 e1g1 c8f5 d3f5 e7f5 f1e1 c6e7 f3e5 h5h4 d1g4 e8g8 a3a4 f5d6 g4h4 e7f5 h4h5 f8e8 c1a3 d8f6 a4a5 d6e4 a5a6 b7b6 e5d7 f6g6 h5g6');
INSERT INTO evaluations VALUES(48180,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48180,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48180,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(48180,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48180,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(48180,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(48180,36,'cp 0','e6d5 c1f4 b8c6 f1b5 f8b4 g1e2 e8g8 e1g1 c6e7 f4g5 d8d6 g5f4');
INSERT INTO evaluations VALUES(56374,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(56374,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(56374,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(56374,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(56374,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(56374,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(56374,36,'cp 32','e5d4 d1d4 b8c6 d4d5 d8d5 e4d5 c6b4 f1b5 c8d7 b5d7 e8d7 e1d1 b4d5 c2c4 d5b6 b2b3 a8e8 h1e1 e8e1 d1e1 f7f6 b1c3 g8e7 c1f4 a7a6 a1d1 d7c8 f3d4 e7g6 f4d2 f8c5 c3e2 h8e8 e1f1 b6d7 d2c3 g6e5 e2f4 c5d4 d1d4 g7g6 d4d1 b7b5 c4b5 a6b5 d1e1 c7c6 f2f3 c8c7 a2a3');
INSERT INTO evaluations VALUES(56244,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(56244,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(56244,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(56244,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(56244,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(56244,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(56244,36,'cp 40','d1d4 b8c6 d4d5 d8d5 e4d5 c6b4 f1b5 c8d7 b5d7 e8d7 e1d1 b4d5 c2c4 d5b6 b2b3 a8e8 h1e1 e8e1 d1e1 d7c8 e1f1 b6d7 b1a3 f8e7 c1b2 e7f6 a1e1 g8h6 b2f6 d7f6 h2h3 h6f5 f3e5 h8f8 a3c2 f5d6 c4c5 d6b5 c5c6 b7c6 e5c6 f8e8 e1e8 f6e8 c6e5');
INSERT INTO evaluations VALUES(56239,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(56239,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(56239,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56239,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(56239,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(56239,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(56239,36,'cp 24','b8c6 d4d5 d8d5 e4d5 c6b4 f1b5 c8d7 b5d7 e8d7 e1d2 b4d5 c2c4 d5b6 b2b3 a8e8 c1b2 f8b4 b2c3 b4c3 b1c3 g8f6 h1e1 h8f8 e1e8 f8e8 a1e1 e8e7 f3e5 d7d8 e1e2 c7c6 e5d3 b6d7 f2f3 d7f8 e2e7 d8e7 d2e3 f8d7 g2g4 f6e8 c3e4 g7g6 d3c5 d7c5 e4c5');
INSERT INTO evaluations VALUES(20667,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(20667,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(20667,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(20667,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(20667,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(20667,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(20667,36,'cp -13','f8e7 e1g1 g8f6 d3d4 c5d4 f3d4 e8g8 c1e3 d8c7 c3b5 c7b8 b5c3 a7a6 a2a4 c8d7 h2h3 b8c7 d1e2 a8c8 c4d3 c6e5 d4f3 e5c4 d3c4 c7c4 e2c4 c8c4 e4e5 f6d5 e5d6 e7d6 c3d5 e6d5 f1d1 d6c7 d1d5 c4a4 a1a4 d7a4 c2c3 a4c6 d5c5');
INSERT INTO evaluations VALUES(5440,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(5440,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(5440,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(5440,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(5440,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(5440,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(5440,36,'cp -7','d3d4 c5d4 f3d4 g8f6 e1g1 e8g8 c1e3 d8c7 h2h3 c6d4 d1d4 c8d7 c4d3 a7a6 a2a4 d7c6 a4a5 a8c8 f1d1 f8d8 d4b6 c7b8 b6a7 b8a7 e3a7 d6d5 a7b6 d5e4 b6d8 e7d8 d3e2 d8e7 g2g4 h7h6 c3a4 c6a4 a1a4 c8c2');
INSERT INTO evaluations VALUES(54756,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(54756,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(54756,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(54756,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(54756,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(54756,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(54756,36,'cp 119','e2e3 e7e6 g1f3 f8d6 c2c4 g8f6 b1c3 e8g8 a1c1 b8c6 a2a3 c8d7 f1e2 c6e7 e1g1 d7e8 d1b3 a8b8 e2d3 d6f4 e3f4 c7c6 f1e1 f6e4 f3e5 e7g6 c4d5 e6d5 c3e2 g6e5 d4e5 d8b6 b3b6 a7b6 e2d4 c6c5 d4e6 c5c4');
INSERT INTO evaluations VALUES(54754,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(54754,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(54754,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(54754,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(54754,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(54754,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(54754,36,'cp 114','g8f6 c2c4 e7e6 b1c3 f8d6 a1c1 b8c6 g1f3 e8g8 a2a3 c8d7 f1e2 c6e7 e1g1 d7e8 f1e1 g8h8 c4c5 d6f4 e3f4 f6e4 b2b4 a7a6 f3e5 e7c6 e5c6 e8c6 f2f3 e4c3 c1c3 f8e8 g2g3 d8f6 d1d2 c6b5 c3e3 h7h6 e2f1 b5f1 g1f1');
INSERT INTO evaluations VALUES(46595,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(46595,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(46595,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(46595,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(46595,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(46595,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(46595,36,'cp 109','c2c4 e7e6 g1f3 f8d6 b1c3 e8g8 a1c1 b8d7 d1c2 d6f4 e3f4 c7c5 c4d5 e6d5 f1e2 c5d4 f3d4 d8b6 c3b5 f6e4 c2c7 b6f6 c1d1 a7a6 b5c3 e4c3 c7c3 f6b6 a2a4 a6a5 d4b5 d7c5 d1d5 c5a4 c3d4 b6d4 d5d4 a4b2 d4d2');
INSERT INTO evaluations VALUES(12780,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(12780,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(12780,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(12780,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(12780,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(12780,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(12780,36,'cp -17','d2d3 d7d5 e4d5 f6d5 g1f3 a7a6 e1g1 c8g4 d3d4 c5d4 c3d4 e7e6 b1c3 f8e7 c3d5 e6d5 c4b3 e8g8 h2h3 g4h5 d1d3 h5g6 d3d1 a8c8 f1e1 g6e4 c1f4 e7b4 e1e3 c6e7 a2a3 e7g6 f4g5 b4e7 g5e7 d8e7 a1c1');
INSERT INTO evaluations VALUES(12781,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(12781,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(12781,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(12781,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(12781,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(12781,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(12781,36,'cp -13','d7d5 e4d5 f6d5 g1f3 e7e6 e1g1 f8e7 d3d4 c5d4 f3d4 e8g8 d4c6 b7c6 f1e1 a7a5 c4d3 h7h6 a2a4 e7d6 b1d2 d8c7 h2h3 f8d8 d1c2 c6c5 d2c4 c8a6 b2b3 a6b7 c4d6 d8d6 d3f1 a8d8 c1d2 b7a6 f1a6 d6a6');
INSERT INTO evaluations VALUES(13080,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(13080,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13080,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(13080,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(13080,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(13080,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(13080,36,'cp -8','e4d5 f6d5 g1f3 e7e6 e1g1 f8d6 d3d4 c5d4 c3d4 e8g8 b1c3 h7h6 f1e1 d5c3 b2c3 b7b6 c4b3 c8b7 b3c2 c6e7 f3e5 d6e5 e1e5 e7g6 c2g6 f7g6 d1d3 d8c7 e5e6 c7f7 e6e2 b7d5 d3d2 f8e8 e2e8 a8e8 c1a3 d5c4 a1e1 e8e1 d2e1 c4a2');
INSERT INTO evaluations VALUES(13077,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(13077,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13077,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(13077,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(13077,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(13077,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(13077,36,'cp -13','f6d5 g1f3 g7g6 e1g1 f8g7 d1b3 e7e6 b3a3 d8a5 a3a5 c6a5 c4d5 e6d5 d3d4 c5d4 f3d4 g7d4 c3d4 c8e6 c1f4 a8c8 b1c3 e8d7 a1c1 a5c4 c3a4 b7b6 f2f3 h7h5 f1d1 c8c6 b2b3 c4a5 c1c3 c6c3 a4c3 a5c6 c3b5 h5h4 g1f2 a7a6 b5c7 c6b4 a2a3 b4c2');
INSERT INTO evaluations VALUES(13199,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(13199,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(13199,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(13199,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(13199,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(13199,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(13199,36,'cp -16','g1f3 g7g6 e1g1 f8g7 d1b3 e7e6 b3a3 d8a5 a3a5 c6a5 c4d5 e6d5 d3d4 c5d4 f3d4 g7d4 c3d4 c8e6 b1c3 e8d7 c1g5 h8c8 f1c1 a5c4 f2f3 h7h5 c3d1 c4d6 g1f2 a7a5 c1c8 a8c8 a1c1 c8c1 g5c1 d6f5 c1d2');
INSERT INTO evaluations VALUES(19225,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(19225,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(19225,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(19225,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(19225,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(19225,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(19225,36,'cp -93','e7e6 d2d3 f8g7 e1g1 g8e7 c2c3 d7d5 c4b5 e8g8 f2f4 e6e5 g5f3 e5f4 c1f4 a7a6 b5c6 e7c6 b1d2 d5d4 c3c4 f7f6 h2h3 a6a5 a2a3 a5a4 d1e2 d8e7 a1e1 g6g5 f4h2 h7h6 e2d1 c8e6 b2b3 a4b3 d1b3 g8h8 h2g3 e7d7 g3h2 c6a5');
INSERT INTO evaluations VALUES(18613,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(18613,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(18613,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(18613,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(18613,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(18613,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(18613,36,'cp -83','d2d3 g8e7 c2c3 f8g7 f2f4 d7d5 c4b5 e8g8 e1g1 e6e5 g5f3 a7a6 b5c6 e7c6 f3e5 c6e5 f4e5 d5e4 d3e4 g7e5 b1d2 e5g7 d1e2 c8e6 d2f3 d8c7 f3g5 c7e7 g5e6 e7e6 e2f2 b7b6 c1g5 e6e4 a1e1 e4d5 g5f6 g7f6 f2f6');
INSERT INTO evaluations VALUES(18611,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(18611,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(18611,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(18611,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(18611,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(18611,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(18611,36,'cp -85','g8e7 e1g1 f8g7 b1c3 e8g8 c4b3 h7h6 g5f3 b7b6 f1e1 d7d6 c3e2 e6e5 c2c3 c8g4 e2g3 d6d5 d1e2 d8d7 c1d2 g4e6 h2h3 f8d8 g3f1 a7a5 b3a4 d5d4 a2a3 d4c3 d2c3 d7c7 f1e3 e7c8 e3c4 c8d6');
INSERT INTO evaluations VALUES(48443,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(48443,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48443,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(48443,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(48443,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48443,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(48443,36,'cp 22','e2e3 c7c5 f1c4 c5d4 e3d4 f8e7 g1f3 e8g8 e1g1 b8c6 a2a3 a7a6 c4a2 b7b5 d1d3 c8b7 a1d1 b5b4 a2b1 g7g6 c3a4 f6d7 g5h6 f8e8 d4d5 e6d5 d3d5 d7f8 d5c4 c6a5 c4g4 d8c8 g4c8 a8c8 a3b4 a5c4 a4c5 c4b2 c5b7 b2d1 f1d1');
INSERT INTO evaluations VALUES(48441,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(48441,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48441,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(48441,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48441,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48441,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(48441,36,'cp 23','c7c5 g5f6 d8f6 d4d5 a7a6 g1f3 b8d7 d5e6 f6e6 d1a4 a8b8 f1c4 e6f6 c4d5 f8e7 e1c1 e8g8 a4e4 d7b6 g2g4 c8e6 h2h4 b6d5 c3d5 e6d5 d1d5 b8d8 g4g5 f6c6 d5e5 c6e4 e5e4 f8e8 h1d1 g8f8 e4a4 f7f6 a4a5 f6g5 h4g5 f8g8 b2b3 g7g6 d1d8');
INSERT INTO evaluations VALUES(51013,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(51013,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(51013,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(51013,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(51013,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(51013,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(51013,36,'cp 122','f1d3 d7d6 h2h3 b8d7 e1g1 b7b6 b1c3 c8b7 d4d5 e6e5 c1e3 g8e7 d1d2 e7g6 c3e2 f8e7 e2g3 e8g8 g3f5 f8e8 c2c4 b7c8 d3c2 d7f6 b2b3 a6a5 g1h2 c8d7 a1d1 b6b5 c4b5 d7b5 c2d3 b5d7 f5g3 a5a4 b3a4');
INSERT INTO evaluations VALUES(51012,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(51012,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(51012,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(51012,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(51012,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(51012,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(51012,36,'cp 117','d7d6 h2h3 b8d7 e1g1 g8e7 b1c3 e6e5 a2a4 e7g6 a4a5 f8e7 f1e1 e5d4 f3d4 e8g8 c3d5 e7h4 g2g3 c7c6 d5c3 h4f6 d3f1 g6e7 d4b3 d8c7 c3a4 c6c5 c1f4 d7e5 f4e5 f6e5 f2f4 e5f6 a4b6 a8b8 f1g2 f6b2');
INSERT INTO evaluations VALUES(57000,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(57000,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(57000,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(57000,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(57000,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(57000,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(57000,36,'cp -27','d7d5 g1f3 g8f6 b4b5 f8d6 e2e3 e8g8 a2a4 c7c6 c2c4 b8d7 f1e2 e6e5 d2d4 e5e4 f3d2 d6c7 b1c3 f8e8 b5c6 b7c6 c4d5 c6d5 c3b5 c7b8 b2a3 d7f8 d1c2 c8d7 b5d6 b8d6 a3d6 d7a4 c2c7 a4c2 c7d8 e8d8 d6c5 f8e6 c5a7 c2d3 e2d3');
INSERT INTO evaluations VALUES(55792,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(55792,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55792,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(55792,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(55792,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(55792,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55792,36,'cp -15','g1f3 g8f6 b4b5 f8d6 a2a4 b8d7 e2e3 e8g8 c2c4 c7c6 d2d4 d6b4 b2c3 a7a5 b5a6 b4c3 b1c3 d8a5 d1d2 b7b6 c4d5 e6d5 c3b1 a5d2 f3d2 c6c5 d2b3 c8a6 f1a6 a8a6 b1c3 f8c8 d4c5 d7c5 b3c5 c8c5 e1d2 f6g4 h1c1 g4f2 d2e2 f2e4 c3e4');
INSERT INTO evaluations VALUES(55791,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(55791,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(55791,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(55791,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(55791,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(55791,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(55791,36,'cp -10','g8f6 b4b5 f8d6 c2c4 e8g8 e2e3 c7c6 a2a4 b8d7 d2d4 d8e7 c4c5 d6c7 f1e2 e6e5 f3e5 d7e5 d4e5 f6e4 e1g1 e4c5 b5c6 b7c6 b1d2 c7e5 b2e5 e7e5 d1c2 e5e7 f1c1 c5e6 c2c6 c8b7 c6c3 d5d4 c3a3 e7a3 a1a3 a8c8 a3a1');
INSERT INTO evaluations VALUES(48222,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(48222,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(48222,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48222,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48222,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(48222,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(48222,36,'cp -13','b4b5 f8d6 c2c4 e8g8 e2e3 c7c6 a2a4 b8d7 c4d5 e6d5 f1e2 c6c5 e1g1 f8e8 d2d3 d7f8 f1e1 h7h6 b1d2 f8g6 e2f1 c8d7 b2c3 a7a6 d1b3 d7e6 b3b2 b7b6 g2g3 d8d7 b5a6 a8a6 c3f6 g7f6');
INSERT INTO evaluations VALUES(48162,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(48162,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(48162,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(48162,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48162,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(48162,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(48162,36,'cp -7','f8d6 c2c4 e8g8 e2e3 c7c6 a2a4 b8d7 c4d5 e6d5 f1e2 f8e8 e1g1 c6c5 d2d3 d7f8 f1e1 f8g6 b1d2 h7h6 e2f1 c8d7 b2c3 b7b6 g2g3 a7a6 d1b3 d7e6 b3b2 d8d7 b5a6 a8a6 c3f6 g7f6 e1b1 e8a8 b2f6 d6c7');
INSERT INTO evaluations VALUES(57024,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(57024,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(57024,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(57024,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57024,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(57024,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(57024,36,'cp 14','g8f6 c2c4 d7d5 g1f3 f8b4 c1d2 b4e7 f1g2 c7c6 e1g1 e8g8 d1c2 b7b6 f3e5 c8b7 f1d1 b8d7 d2f4 f6h5 c4d5 c6d5 f4e3 h5f6 e5c6 b7c6 c2c6 d8b8 d1c1 f8c8 c6b5 a7a6 c1c8 b8c8 b5b3 b6b5 a2a4 b5a4 a1a4 a6a5');
INSERT INTO evaluations VALUES(49431,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(49431,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(49431,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(49431,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(49431,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(49431,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(49431,36,'cp 27','c2c4 f8b4 c1d2 b4e7 f1g2 e8g8 d4d5 c7c6 b1c3 d7d6 e2e4 b8d7 g1f3 c6d5 c4d5 e6d5 e4d5 d7e5 f3e5 d6e5 e1g1 e7c5 d2g5 c8f5 d1f3 f5g6 g5f6 d8f6 f3f6 g7f6 c3e4 c5d4 e4f6 g8g7 f6e4 f8d8 a1c1');
INSERT INTO evaluations VALUES(53972,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(53972,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53972,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53972,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53972,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(53972,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53972,36,'cp 0','g8f6 c4d3 f8e7 d3c2 b7b5 e1g1 c8b7 f1e1 b8c6 a2a3 c5c4 d2d3 c4d3 c2d3 a7a6 c1f4 e8g8 d1e2 d8c7 b1d2 c6e5 f4e5 d6e5 b2b4 c7c3 e1c1 c3b2 c1b1 b2c3');
INSERT INTO evaluations VALUES(45991,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45991,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45991,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45991,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45991,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45991,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45991,36,'cp 10','c4d3 f8e7 d3c2 b7b5 e1g1 c8b7 f1e1 b8c6 a2a3 c5c4 d2d3 c4d3 c2d3 a7a6 c1f4 e8g8 d1e2 d8c7 b1d2 c6e5 f3e5 d6e5 f4e3 f8d8 h2h3 c7d7 d3c2 a8c8 a3a4 h7h6 e1d1 d7c6 a4b5 a6b5 d1e1 f6d7 c2d3 b5b4 c3b4');
INSERT INTO evaluations VALUES(20807,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(20807,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(20807,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(20807,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(20807,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(20807,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(20807,36,'cp 30','c7c6 d5c6 b7c6 b5a4 d8c7 c2c4 e7g6 b1c3 c8d7 h2h4 h7h5 e1g1 f8e7 b2b4 g8f6 c1g5 e8g8 d1d3 a8c8 a1c1 d7g4 f3h2 g6f4 g5f4 e5f4 b4b5 c6b5 c4b5 c7c4 d3c4 c8c4 h2g4 h5g4 c3d5 c4e4 a4b3 f6d5 b3d5 e4e2');
INSERT INTO evaluations VALUES(19622,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(19622,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(19622,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(19622,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(19622,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(19622,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(19622,36,'cp 44','d5c6 b7c6 b5a4 d8c7 b1c3 g8f6 c1e3 a7a5 e1g1 c8d7 e3g5 e7g6 f3d2 f6g4 d2c4 h7h6 g5c1 f8e7 b2b3 e8g8 c1a3 d7e6 a3d6 e7d6 d1d6 c7a7 a4c6 e6c4 b3c4 a8d8 d6a3 d8d2 c4c5 g6f4 c3d5 d2c2 a3f3 f4e2 g1h1 h6h5 h2h3 e2d4 f3g3 d4c6 h3g4 h5g4');
INSERT INTO evaluations VALUES(19621,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(19621,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19621,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(19621,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(19621,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(19621,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(19621,36,'cp 25','b7c6 b5a4 g8f6 b1c3 d8c7 c1g5 f6d7 b2b4 c8a6 g5d2 e7g6 b4b5 a6b5 c3b5 c6b5 a4b5 f8e7 e1g1 e8g8 a1b1 f8c8 d1e2 a8b8 b5a4 b8b1 f1b1 g6f4 d2f4 e5f4 a4b3 d7e5 e2a6 e5f3 g2f3 e7h4');
INSERT INTO evaluations VALUES(18441,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(18441,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(18441,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(18441,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(18441,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(18441,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(18441,36,'cp 35','b5a4 g8f6 b1c3 d8c7 c1g5 c8e6 e1g1 f6d7 b2b4 e7g6 h2h4 e6g4 d1d3 f8e7 g5e7 e8e7 g2g3 d7f6 a4b3 h7h6 b4b5 c6b5 c3b5 c7c5 a2a4 a7a6 b5a3 g6f8 a3c4 f8e6 c4e3 h8b8 f3d2 e6d4 e3g4 f6g4 d2c4');
INSERT INTO evaluations VALUES(18442,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(18442,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18442,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(18442,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(18442,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(18442,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(18442,36,'cp 40','g8f6 b1c3 d8c7 c1g5 c8e6 a4b3 h7h6 g5f6 g7f6 f3h4 d6d5 d1f3 e8c8 e4d5 c6d5 e1c1 e5e4 c3e4 d5e4 b3e6 f7e6 d1d8 c7d8 f3f6 h8g8 h1d1 d8b6 f6f7 f8g7 c2c3 g7c3 b2c3 g8d8 d1d8 c8d8 f7f6 b6a6 f6h8 d8d7 h8d4 d7e8 d4e4 a6a2 f2f4 a2a3 c1d2 a3b2 e4c2 b2c2 d2c2 e7d5 f4f5 d5e3 c2d3 e3f5 h4f5 e6f5 d3c4 e8d7 c4b5 d7e6');
INSERT INTO evaluations VALUES(8855,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8855,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8855,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8855,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8855,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8855,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8855,36,'cp 0','a7a6 c1e3 e8g8 c4a2 c8e6 e3c5 d6c5 a2e6 f7e6 e1g1 d8e8 c3b1 c5c4 b1d2 c4d3 c2d3 a8d8 d1b3 f6h5 g2g3 e8g6 f3h4 g6f6 h4f3');
INSERT INTO evaluations VALUES(7147,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7147,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7147,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7147,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7147,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7147,36,'cp 0','c1e3 e8g8 e3c5 d6c5 e1g1 d8d6 a3a4 c8e6 f3d2 b7b6 c4e6 d6e6 d2c4 a8d8 c4e3 e6d6 e3f5 d6e6');
INSERT INTO evaluations VALUES(7150,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7150,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7150,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7150,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7150,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7150,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7150,36,'cp 0','e8g8 c4a2 c8e6 e3c5 d6c5 a2e6 f7e6 e1g1 d8e8 c3b1 c5c4 b1d2 c4d3 c2d3 a8d8 d1b3 f6h5 g2g3 e8g6 f3h4 g6f6 h4f3');
INSERT INTO evaluations VALUES(11856,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(11856,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(11856,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(11856,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(11856,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(11856,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(11856,36,'cp -22','c6c5 d2d3 f8e7 f3d2 e8g8 d2c4 c7c6 b2b3 f8e8 h2h3 c8e6 a1b1 f6d7 f2f4 e5f4 c1f4 d7f8 f4g3 f8g6 d1d2 d6d5 c4e5 g6e5 g3e5 f7f6 e5g3 d8a5 d2e1 e7f8 c3a4 a5e1 f1e1 c5c4 b3c4 d5c4 d3d4 e8d8');
INSERT INTO evaluations VALUES(38876,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(38876,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(38876,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(38876,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(38876,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(38876,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(38876,36,'cp 77','f1c4 c7c5 a2a3 b4c3 b2c3 d8c7 d1d3 g8f6 g1e2 e8g8 e1g1 c8d7 c4a2 a7a6 f1e1 f8d8 e3e4 c5d4 e2d4 d7b5 d3h3 b8c6 e4e5 c6d4 c3d4 f6d5 c1g5 c7c3 h3c3 d5c3 g5d8 a8d8 h2h3 h7h6 g1h2 c3a2 a1a2 d8d4 a2c2 b5c4 f2f3 b7b5 e1c1');
INSERT INTO evaluations VALUES(38867,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(38867,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(38867,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(38867,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(38867,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(38867,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(38867,36,'cp 75','c7c5 a2a3 b4c3 b2c3 d8c7 c4a2 g8f6 g1e2 e8g8 e1g1 b7b5 e2g3 c8b7 d1e2 b8d7 f1d1 f8c8 e3e4 a7a5 h2h3 h7h6 c1b2 d7b6 c3c4 b6a4 d4d5 a4b2 e2b2 b5b4 e4e5 f6d7 f2f4 e6d5 b2e2 d5d4 g3f5 c8e8 e2g4 g7g6 f5h6 g8f8');
INSERT INTO evaluations VALUES(52894,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(52894,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(52894,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(52894,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(52894,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(52894,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(52894,36,'cp 85','f1c4 g8f6 g1f3 b7b5 c4d3 c8g4 c3e4 b8d7 h2h3 g4h5 c1f4 f6d5 f4h2 e7e6 a2a4 a7a6 d1e2 f8b4 c2c3 b4e7 e1g1 e8g8 f1e1 d7f6 e4c5 f6d7 c5d7 d8d7 g2g4 h5g6 d3g6 h7g6 f3e5 d7b7 e5d3 a6a5 h2g3 b7d7 h3h4 e7d6');
INSERT INTO evaluations VALUES(52871,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(52871,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(52871,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(52871,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(52871,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(52871,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(52871,36,'cp 94','g8f6 g1f3 b7b5 c4d3 c8g4 a2a4 b5b4 c3e4 b8d7 h2h3 g4h5 c1f4 e7e6 d1e2 a7a5 e1c1 h5f3 e2f3 f6d5 f4h2 d7b6 b2b3 f7f5 e4d2 f8d6 h2d6 d8d6 h1e1 e8g8 d2c4 b6c4 d3c4 a8e8 e1e5 d5c3 d1d3 c3d5 h3h4 g8h8 h4h5');
INSERT INTO evaluations VALUES(45006,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(45006,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(45006,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(45006,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(45006,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(45006,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(45006,36,'cp 99','g1f3 b7b5 c4d3 c8g4 a2a4 b5b4 c3e4 b8d7 h2h3 g4e6 c1d2 a7a5 c2c4 b4c3 b2c3 f6e4 d3e4 e6d5 d1c2 e7e6 c3c4 d5e4 c2e4 d8c7 c4c5 f8e7 a1b1 e8g8 e1g1 d7f6 e4e5 c7c8 e5e2 f6d5 b1b3 a8b8 b3b8 c8b8 d2a5 b8b3 f1a1 e7f6');
INSERT INTO evaluations VALUES(45004,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(45004,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(45004,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(45004,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(45004,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(45004,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(45004,36,'cp 99','b7b5 c4d3 c8g4 c3e4 b8d7 a2a4 b5b4 h2h3 g4h5 c1f4 f6e4 d3e4 a8c8 e4d3 d8a5 e1g1 e7e6 c2c3 b4c3 b2c3 f8e7 a1b1 d7b6 c3c4 e8g8 f1e1 h5g6 f3e5 e7f6 e5g4 f6e7 c4c5 b6d5 f4e5 g6d3 d1d3 c8d8 d3b3 d8d7 g4e3 f8d8 e3c4 a5c3 b3a2 h7h6 e1d1');
INSERT INTO evaluations VALUES(23413,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(23413,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(23413,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(23413,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(23413,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(23413,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(23413,36,'cp 92','d2d4 f8g7 d4d5 c6e5 f3e5 g7e5 f1d3 d7d6 e1g1 g8f6 g1h1 f6d7 b1c3 e8g8 c1h6 e5g7 d1d2 g7h6 d2h6 d7e5 d3e2 f7f5 e4f5 c8f5 h6d2 c7c6 f2f4 e5f7 a1e1 c6d5 c3d5 e7e6 d5e3 d6d5 e3f5 e6f5');
INSERT INTO evaluations VALUES(23409,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(23409,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(23409,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(23409,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(23409,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(23409,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(23409,36,'cp 77','f8g7 d4d5 c6e5 f3e5 g7e5 c2c4 d7d6 f1e2 c7c6 h2h3 c6d5 c4d5 c8d7 e1g1 g8f6 d1d3 d8b6 b1d2 b6d4 d2b3 d4d3 e2d3 f6h5 f2f4 e5g7 g2g4 h5f6 f1e1 e8g8 c1e3 f8c8 a1b1 a7a6');
INSERT INTO evaluations VALUES(22788,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(22788,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(22788,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(22788,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(22788,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(22788,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(22788,36,'cp -63','g8f6 f1c4 c6a5 c4b5 a7a6 b5a4 b7b5 a4c2 d7d5 d2d3 f8e7 f3e2 d5e4 d3e4 c7c5 g1f3 d8c7 c1g5 e8g8 e1g1 c8e6 f1d1 h7h6 g5f6 e7f6 h2h3 a8d8 b1d2 a5c4 d2c4 d8d1 e2d1 e6c4 a2a4 f8d8 d1e1 c4e6 a4b5 a6b5 e1e2 b5b4 e2b5 b4b3 c2b3 c5c4 b3a2 d8b8');
INSERT INTO evaluations VALUES(16535,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(16535,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(16535,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(16535,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(16535,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(16535,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(16535,36,'cp -49','f1c4 c6a5 c4b5 a7a6 b5a4 b7b5 a4c2 c7c5 d2d3 d7d5 f3g3 a5c6 g1f3 d5e4 d3e4 d8c7 e1g1 h7h6 f3d2 c8e6 a2a4 b5b4 c2b3 e6b3 d2b3 f6e4 g3f3 e4f6 c1e3 e5e4 f3e2 f8d6 b3c5 d6h2 g1h1 h2e5 c5a6 c7b7 a6b4 c6b4 c3b4');
INSERT INTO evaluations VALUES(1439,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(1439,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(1439,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(1439,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(1439,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(1439,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(1439,36,'cp -19','b5a4 c8d7 g1e2 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 f8e7 c1e3 e8g8 a4c2 f5e3 f2e3 a8c8 f1f3 g7g6 d1d2 f7f6 e5f6 e7f6 a1f1 f6g7 f3f8 c8f8 f1f8 g7f8 c2a4 f8d6 h2h3 a7a6 e2f4 b6c7 a4c6 c7c6 f4d3 c6b6 d3e5 b6c7 e5d7 c7d7 g1h1');
INSERT INTO evaluations VALUES(13516,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(13516,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13516,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(13516,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(13516,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13516,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13516,36,'cp 12','e6d5 f1b5 c5d4 e3d4 f8d6 c1g5 e8g8 e1g1 c8g4 b5e2 d6e7 h2h3 g4h5 f1e1 h7h6 g5h4 f6e4 h4e7 c6e7 d1b3 e4c3 b2c3 b7b6 f3e5 h5e2 e1e2 a8c8 a2a4 c8c7 a4a5 d8c8 a5b6 a7b6 b3b6 c7c3 a1a7 e7f5 e2e1 c3c2');
INSERT INTO evaluations VALUES(33975,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(33975,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(33975,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(33975,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(33975,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(33975,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(33975,36,'cp 80','f1e2 d7d5 e1g1 e4b4 d2d4 f8e7 c2c4 d5c4 b1c3 c8g4 h2h3 g4f3 e2f3 e8g8 a2a3 b4b6 d1e2 e7d6 e2c4 c7c6 b2b4 b8d7 g2g3 f8e8 g1g2 a7a6 f1d1 h7h6 c4b3 b6c7 a3a4 d7f8 a1b1 f8g6 b4b5 a6b5 a4b5 g6e7 c3e4 f6e4 f3e4 c7d7 d4d5 c6b5 b3b5');
INSERT INTO evaluations VALUES(33976,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(33976,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(33976,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(33976,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(33976,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(33976,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(33976,36,'cp 72','d7d5 e1g1 e4b4 d2d4 f8e7 c2c4 d5c4 b1c3 c8g4 h2h3 g4f3 e2f3 e8g8 a2a3 b4b6 d1e2 e7d6 e2c4 c7c6 g2g3 b8d7 b2b4 a7a6 g1g2 b6c7 c4b3 d7b6 c1g5 d6e7 a3a4 c7d6 b4b5 a6b5 a4b5 d6d4 a1a8 f8a8 b5c6 b7c6 f1b1 h7h6 g5f4 b6d5 c3d5');
INSERT INTO evaluations VALUES(35096,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(35096,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(35096,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(35096,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(35096,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(35096,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(35096,36,'cp -22','b8c6 f1c4 g7g6 h5g5 f8h6 g5f6 g8f6 b1c3 d7d6 d2d3 h6c1 a1c1 a7a6 h2h4 c8g4 f3g1 g4e6 c4b3 e8g8 g1e2 e6b3 a2b3 a8d8 e1d2 g8g7 f2f3 d6d5 e4d5 f6d5 c3d5 d8d5 e2c3 d5d7 c3e4 b7b6 h4h5 f7f5 h5g6 g7g6 e4f2 c6d4');
INSERT INTO evaluations VALUES(1856,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(1856,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(1856,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(1856,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(1856,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(1856,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(1856,36,'cp -16','b1c3 g7g6 h5g5 f8h6 g5f6 g8f6 f1c4 d7d6 d2d3 h6c1 a1c1 a7a6 h2h4 e8g8 c3d5 f6d5 c4d5 c6b4 d5c4 c8g4 f3d2 b7b5 c4b3 a8b8 a2a3 b4c6 f2f3 g4d7 b3d5 c6e7 e1f2 g8g7 b2b4 f7f5 g2g3 e7d5 e4d5 c7c6 d5c6 d7c6 h1e1 c6b7 d2b3');
INSERT INTO evaluations VALUES(37430,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(37430,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(37430,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(37430,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(37430,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(37430,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(37430,36,'cp -17','c7c5 c2c3 d8b6 d1b3 d7d6 b1d2 c5d4 e3d4 c8e6 d3c4 e6c4 d2c4 b6c6 a2a4 f8c8 c4e3 b8d7 e1g1 d7b6 f1c1 b6d5 e3d5 f6d5 f4d2 d5b6 c1e1 c8c7 a4a5 b6d5 h2h4 h7h6 e1c1 e7e5 c3c4 d5e7 d2c3 c6c4 b3c4 c7c4 d4e5 d6e5 c3e5 c4e4 e5g7 g8g7 a5a6');
INSERT INTO evaluations VALUES(48679,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(48679,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(48679,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48679,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(48679,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48679,36,'cp 0','f8d6 f1d3 h7h6 e1g1 e8g8 b1c3 b8c6 c3b5 c6b4 b5d6 d8d6 f3e5 b4d3 d1d3 f6e4 c1f4 c8f5 a1e1 d6b4 a2a3 b4b2 e1b1 b2a2 b1a1 a2b2');
INSERT INTO evaluations VALUES(41719,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(41719,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41719,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(41719,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(41719,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(41719,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(41719,36,'cp 11','f1d3 e8g8 e1g1 b8c6 b1c3 h7h6 c3b5 d6e7 a2a3 a7a6 b5c3 e7d6 f1e1 c8d7 c1d2 d8c8 c3e2 f8e8 e2g3 d6g3 e1e8 f6e8 f2g3 d7f5 d2f4 c8d7 g3g4 f5d3 c2d3 e8d6 d1b3 c6e7 a1e1 a6a5 f4e5 d7c6 e5d6 c7d6');
INSERT INTO evaluations VALUES(41709,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(41709,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(41709,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41709,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41709,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41709,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(41709,36,'cp 17','e8g8 e1g1 b8c6 b1c3 h7h6 c3b5 d6e7 a2a3 a7a6 b5c3 e7d6 f1e1 c6e7 f3e5 e7f5 c3e2 c7c5 c2c3 f8e8 c1f4 f6e4 f2f3 e4f6 d1d2 f5h4 f4h6 g7h6 d2h6 h4f5 h6g5 f5g7 e2g3 c5d4 c3d4 d8b6 g1h1 e8e6 d3f5');
INSERT INTO evaluations VALUES(36943,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36943,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36943,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(36943,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36943,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36943,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36943,36,'cp 16','e1g1 b8c6 b1c3 h7h6 c3b5 d6e7 c1f4 a7a6 b5c3 c6b4 f1e1 b4d3 d1d3 c7c6 c3e2 f8e8 e2g3 e7f8 f4e5 c8d7 d3b3 b7b5 e1e3 f6e4 g3e4 d5e4 e3e4 d7e6 b3c3 f7f6 e5f4 d8d7 e4e1 e6f7 a2a3 e8e1 a1e1 a8e8 f4d2 f7d5 c3d3 e8e1 f3e1 d5f7');
INSERT INTO evaluations VALUES(36942,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(36942,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(36942,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36942,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36942,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36942,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(36942,36,'cp 10','b8c6 b1c3 h7h6 c3b5 d6e7 c1f4 a7a6 b5c3 c6b4 f1e1 b4d3 d1d3 c7c6 e1e2 a6a5 a1e1 f8e8 a2a3 e7f8 f3e5 a5a4 g2g4 c8e6 f4g3 d8b6 h3h4 f6e4 c3e4 d5e4 d3e4 f7f6 e5d3 e6f7 e4f4 b6b5 f4f5 e8e2 e1e2 b5c4 g4g5 h6g5 h4g5 f6g5 f5g5 c4d4');
INSERT INTO evaluations VALUES(13893,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(13893,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(13893,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(13893,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(13893,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(13893,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(13893,36,'cp 47','c1g5 f8e7 g5f6 e7f6 f1c4 e8g8 e1g1 c6a5 c4b3 c8e6 c3d5 f6e7 d1d3 a8c8 c2c3 a7a6 a1d1 c8c6 h2h3 d8d7 d3e2 a5b3 a2b3 b7b5 f1e1 d7b7 f3h2 c6c8 e2d2 h7h6 d5e7 b7e7 d2d6 e7d6 d1d6');
INSERT INTO evaluations VALUES(13879,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(13879,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(13879,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(13879,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(13879,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(13879,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(13879,36,'cp 44','f8e7 g5f6 e7f6 c3d5 e8g8 f1c4 c8e6 c4b3 c6e7 d5f6 g7f6 d1d3 d8b6 e1g1 a8d8 c2c4 e7g6 a1d1 d8c8 g2g3 b6c6 h2h4 h7h5 f1e1 g8g7 f3h2 a7a5 d3d6 e6c4 d6c6 c8c6 b3c4 c6c4 d1d7 f8c8 d7b7 a5a4');
INSERT INTO evaluations VALUES(8142,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(8142,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(8142,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(8142,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(8142,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(8142,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8142,36,'cp 50','g5f6 e7f6 f1c4 e8g8 e1g1 c8e6 c4b3 g7g6 c3d5 f6g7 d5e3 c6d4 b3e6 d4f3 d1f3 f7e6 f3g4 d8d7 g2g3 g7h6 e3g2 b7b5 c2c3 f8f7 a1d1 a8f8 g4e2 d7c6 d1d3 a7a5 g2e1 b5b4 c3b4 a5b4 d3b3 c6a4 h2h4 a4a2 b3b4 a2a7 g1g2 f7b7 b4b7 a7b7');
INSERT INTO evaluations VALUES(38458,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(38458,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38458,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38458,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38458,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(38458,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38458,36,'cp 21','g8f6 b1d2 e8g8 h2h3 c7c5 c2c3 f6d7 f1e2 b8c6 e1g1 f8e8 e3e4 d5e4 d2e4 c5d4 c3d4 d7b6 e2b5 c8e6 b5c6 b7c6 b2b3 a7a5 f1e1 e6d5 f3e5 d5e4 e1e4 g7e5 d4e5 d8d1 a1d1 b6d5 f4d2 e8d8 d2g5 f7f6 g5h4 d8c8 e5f6');
INSERT INTO evaluations VALUES(38934,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(38934,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(38934,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(38934,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(38934,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(38934,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(38934,36,'cp -7','g8f6 a2a3 b4c3 d2c3 f6e4 f1d3 e4c3 b2c3 e8g8 d1f3 c7c5 d4c5 c8e6 g1e2 b8d7 c5c6 b7c6 e2d4 d7c5 e1g1 a8c8 f1e1 d8d6 f3g3 d6g3 h2g3 f8e8 g3g4 e6d7 f2f3 g8f8 c3c4 c5d3 c2d3 c6c5 e1e8 f8e8 a1e1 e8d8 d4f5 d7f5 g4f5 d5c4 d3c4 c8b8 e1e5');
INSERT INTO evaluations VALUES(42240,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(42240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42240,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(42240,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(42240,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(42240,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(42240,36,'cp 0','a2a3 b4c3 d2c3 f6e4 f1d3 e4c3 b2c3 e8g8 d1f3 c7c6 g1e2 b8d7 a3a4 d8a5 e1g1 d7f6 h2h3 f8e8 e2g3 h7h6 f1b1 b7b6 g1h2 e8e7 d3f5 c8a6 f5d3');
INSERT INTO evaluations VALUES(42248,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(42248,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(42248,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(42248,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42248,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(42248,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(42248,36,'cp -12','b4c3 d2c3 f6e4 f1d3 e4c3 b2c3 e8g8 d1f3 c7c5 d4c5 c8e6 g1e2 b8d7 c5c6 b7c6 e1g1 d7c5 f1e1 d8d6 e2d4 f8e8 f3g3 d6g3 h2g3 a8c8 f2f3 h7h5 c3c4 g7g6 g1f2 g8g7 c4d5 c5d3 c2d3 e6d5 e1e8 c8e8 a1c1 g7f6 c1c5 f6e5 d4b3');
INSERT INTO evaluations VALUES(43726,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43726,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(43726,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(43726,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(43726,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(43726,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(43726,36,'cp 47','c4d5 a8a7 c1e3 a7c7 d5b3 e7e6 f2f3 b5b4 c3a4 f8e7 d1d2 c7b7 c2c3 b4c3 a4c3 e8g8 e1g1 c8d7 b3c4 b7c7 c4d3 d8c8 d4b3 c8b7 b3a5 b7b4 c3e2 b4d2 e3d2 d7b5 d3b5 a6b5 f1c1 f8c8 e2d4 d6d5 e4e5 c7c1 a1c1 c8c1 d2c1 f6d7 d4b5 d7e5 c1e3');
INSERT INTO evaluations VALUES(43725,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(43725,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(43725,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43725,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(43725,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(43725,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(43725,36,'cp 45','a8a7 d4c6 b8c6 d5c6 c8d7 d1d4 a7c7 c6d7 d8d7 a2a4 b5b4 c3d5 f6d5 e4d5 e7e5 d4b4 f8e7 e1g1 e8g8 c2c4 d7c8 b2b3 c8b7 b4a5 c7c5 a5c3 b7d5 c1e3 c5c8 c4d5 c8c3 b3b4 f7f5 b4b5 a6b5 a4b5 f8b8 f1b1 f5f4 e3a7 b8b7 f2f3 c3d3 a7f2 d3d5 b5b6');
INSERT INTO evaluations VALUES(38931,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(38931,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38931,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38931,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(38931,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(38931,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(38931,36,'cp 28','c1f4 c7c5 d4c5 g8f6 e2e3 e8g8 f1d3 b8c6 g1e2 b4c5 e1g1 d5d4 e3d4 c6d4 c3a4 d4e2 d3e2 c5e7 e2f3 d8a5 a4c3 a5b4 f4e5 f8d8 d1a4 b4a4 c3a4 f6d5 f1c1 e7f6 c1e1 f6e5 e1e5 d5f4 g2g3 f4e6 e5d5 d8d5 f3d5 c8d7 a4c3 g7g6 a1d1 a8b8 f2f3 e6c5 g1f2 d7e6 d5e6 c5e6 c3e4');
INSERT INTO evaluations VALUES(38927,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(38927,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38927,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(38927,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(38927,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(38927,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(38927,36,'cp 24','c7c5 d4c5 g8f6 e2e3 e8g8 f1d3 b8c6 g1e2 b4c5 e1g1 d5d4 e3d4 c6d4 c3a4 d4e2 d3e2 c5e7 e2f3 d8a5 a4c3 a5b4 f4e5 f8d8 d1a4 b4a4 c3a4 c8d7 b2b3 d7c6 f3c6 b7c6 f1c1 c6c5 e5c7 d8c8 c7a5 f6d5 c1e1 e7f6 a1c1 f6d4 g1f1 d5b4 a5b4 c5b4 c1c8 a8c8 e1d1 c8d8');
INSERT INTO evaluations VALUES(45769,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(45769,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45769,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45769,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(45769,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(45769,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45769,36,'cp 25','f1d3 f8d6 d1e2 d6e7 h2h3 e8g8 e1g1 f8e8 c1f4 e7f8 e2d1 h7h6 f1e1 e8e1 d1e1 c8e6 c3e2 b8d7 e2g3 a7a5 c2c3 b7b5 a2a3 d7b6 e1e2 b6c4 f3e5 c4e5 f4e5 f6d7 f2f4 d7e5 d4e5 d8h4 e2f2 e6d7 a1e1 a8e8 f4f5 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(45774,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(45774,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(45774,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(45774,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45774,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(45774,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(45774,36,'cp 13','f8d6 d1e2 d6e7 h2h3 e8g8 e1g1 f8e8 c1f4 e7f8 e2d1 h7h6 a2a3 b8d7 f1e1 e8e1 d1e1 d7b6 a3a4 a7a5 b2b3 c8e6 c3e2 b6d7 e2g3 d8b6 e1d2 a8e8 a1e1 b6d8 c2c3 f6e4 g3e4 d5e4 d3e4 e6b3 d2b2 d7f6 b2b3 f6e4 f4e5 e4d6');
INSERT INTO evaluations VALUES(40523,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40523,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(40523,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(40523,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(40523,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(40523,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40523,36,'cp 11','d1e2 d6e7 h2h3 e8g8 e1g1 f8e8 c1f4 e7f8 e2d1 h7h6 c3e2 c6c5 c2c3 b8c6 f1e1 d8b6 a1b1 c5d4 e2d4 e8e1 d1e1 c6d4 f3d4 c8d7 e1d2 a8e8 f4e3 f6e4 d2e2 a7a6 b1e1 f8c5 e2c2 c5d6 d4f3 d6c5 b2b4 c5e3 e1e3 e4f6 e3e8 d7e8');
INSERT INTO evaluations VALUES(22931,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(22931,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(22931,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(22931,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(22931,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(22931,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(22931,36,'cp -31','d4d1 d7d5 f1e2 g8f6 g1f3 f8c5 e1g1 e8g8 c2c4 d5c4 d1d8 f8d8 e2c4 c8f5 b1c3 a7a6 a2a3 c5d6 f3d4 c6d4 e3d4 d8e8 c1g5 f6e4 c3e4 f5e4 f1e1 h7h6 g5e3 g8f8 a1c1 e4c6 g2g3 a8d8 b2b4 c6f3 c4e2 f3e2 e1e2');
INSERT INTO evaluations VALUES(22974,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(22974,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(22974,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(22974,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(22974,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(22974,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(22974,36,'cp -47','g8f6 g1f3 d7d5 a2a3 f8e7 c2c4 e8g8 b1c3 d5c4 d1d8 f8d8 f1c4 f6g4 c3d5 e7d6 c1d2 g4e5 f3e5 c6e5 c4e2 c8e6 d5f4 e6b3 d2c3 a7a5 f4h5 f7f6 h5g3 a5a4 a1c1 c7c6 g3f5 d6f8 f5d4 b3f7 e1g1 c6c5 d4b5 e5d3 c1c2 f7b3 c2d2 c5c4 e2f3 d3c5 c3b4 d8d2 b4d2');
INSERT INTO evaluations VALUES(16723,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(16723,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16723,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16723,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(16723,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(16723,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(16723,36,'cp -23','f1e2 d7d5 g1f3 f8e7 e1g1 c8e6 b1d2 e8g8 c2c4 a7a5 c4d5 d8d5 b2b3 f8d8 c1b2 f6e4 d2e4 d5e4 d1c1 e4g6 f1d1 d8d1 c1d1 a8d8 d1c1 h7h5 g1h1 e6d5 h2h3 e7f6 a1b1 d5e4 b2f6 g6f6 b1b2 b7b6 b2d2 d8d2 c1d2 c6e5 f3e5 f6e5 d2d8 g8h7');
INSERT INTO evaluations VALUES(50646,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(50646,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50646,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50646,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(50646,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(50646,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(50646,36,'cp -15','b7b5 c4d5 a8a7 a2a3 g8f6 d5a2 e7e6 d2d3 a7d7 c3e2 b8c6 c2c3 d8b6 e2g3 c6e5 h2h4 f8e7 c1f4 e5f3 d1f3 c8b7 f3e2 h7h6 e1g1 e8g8 b2b4 e6e5 b4c5 b6c7 f4e3 d6d5 a1d1 e7c5 e3c5 c7c5');
INSERT INTO evaluations VALUES(51258,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(51258,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(51258,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(51258,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(51258,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(51258,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51258,36,'cp 0','c4d5 a8a7 d2d3 g8f6 d5b3 e7e6 e1g1 b8c6 a2a3 f8e7 c3e2 a7d7 f1e1 e8g8 h2h3 c8b7 e2g3 c6e5 f3e5 d6e5 d1e2 a6a5 a3a4 c5c4 d3c4 b5b4 c1g5 h7h6 g5e3 d8c7 a1d1 d7d1 e1d1 b7e4 g3e4 f6e4 e2g4 e4f6 g4g3 f6e4 g3g4');
INSERT INTO evaluations VALUES(51259,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51259,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51259,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51259,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51259,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51259,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51259,36,'cp 0','a8a7 d2d3 e7e6 d5b3 g8f6 e1g1 f8e7 c3e2 d8c7 a2a4 e8g8 e2g3 c8d7 c2c3 c5c4 d3c4 b5c4 b3a2 e6e5 f1e1 d7e6 f3g5 e6g4 g5f3');
INSERT INTO evaluations VALUES(6259,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(6259,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(6259,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(6259,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(6259,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(6259,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(6259,36,'cp 68','c3d5 g8f6 e1g1 e8g8 c2c3 b4e7 f1e1 d7d6 a2a4 a7a5 h2h3 f6d5 e4d5 c6b8 d2d4 e5d4 f3d4 b8d7 c4b3 e7g5 b3c2 g5c1 a1c1 d7f6 c3c4 c8d7 d1d2 f8e8 b2b3 e8e5 c1d1 e5e8 c2f5 e8e1 d1e1 c7c5 f5d7 c5d4 d7b5 d8b6 d2d3');
INSERT INTO evaluations VALUES(6255,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(6255,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(6255,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(6255,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(6255,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(6255,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(6255,36,'cp 64','g8f6 d5b4 c6b4 f3e5 e8g8 e1g1 f6e4 c2c3 b4c6 d2d4 d7d5 c4b3 c6e5 d4e5 c7c6 b3c2 f7f5 f2f3 e4g5 c1e3 b7b6 d1d2 g5e6 f1d1 d8e7 d1e1 e7h4 a2a4 c8b7 a4a5 c6c5 b2b4 a8d8 e3f2 h4e7 c2d3 e7f7 f3f4 b7c6 a5b6 a7b6 f2h4 d8d7');
INSERT INTO evaluations VALUES(14194,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(14194,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(14194,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(14194,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(14194,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14194,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(14194,36,'cp 15','e2e3 c8f5 f1b5 a7a6 b5c6 b7c6 g1f3 h7h6 g5h4 d8b8 b2b3 f6d7 e1g1 e7e6 a2a3 b8b7 c3a4 c6c5 c2c4 g7g5 h4g3 d5c4 b3c4 f8g7 a4c5 d7c5 d4c5 e8g8 f3d4 a8d8 c5c6 b7a7 c4c5 e6e5 d4f5 d8d1 f1d1 a7c5 a1c1 c5b6 d1d7');
INSERT INTO evaluations VALUES(14192,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(14192,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(14192,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(14192,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(14192,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(14192,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(14192,36,'cp 15','c8f5 f1b5 e7e6 g1f3 h7h6 g5f6 g7f6 b5d3 f5d3 d1d3 d8d7 e3e4 e8c8 e1c1 c6b4 d3e3 d5e4 e3e4 b4d5 c3d5 d7d5 e4d5 d8d5 g2g3 d5a5 a2a3 a5f5 d1d3 h6h5 c1d2 h5h4 h1e1 a7a5 d2e2 h4g3 h2g3 a5a4 c2c4 h8h5 c4c5 f8h6 b2b4 a4b3 d3b3 h5h3');
INSERT INTO evaluations VALUES(43868,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(43868,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(43868,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(43868,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(43868,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(43868,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(43868,36,'cp 44','c8b7 d1c2 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 e7e6 g1f3 f8e7 c1f4 e8g8 f1d3 c7c5 d4d5 e6d5 e4d5 d8d5 e1c1 d5h5 d3e4 b7c6 h2h4 f7f5 c2b3 g8h8 e4d5 b6b5 h1e1 c5c4 b3b2 e7f6 f4g5 c6d5 d1d5 h5g6 b2b5 b8a6 b5a5 a8b8 g5f6 g6f6 f3e5 f6h4');
INSERT INTO evaluations VALUES(10387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10387,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(10387,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10387,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(10387,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10387,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10387,36,'cp 0','g1f3 d7d6 e1g1 c6a5 c4b5 c7c6 b5a4 b4c3 b2c3 b7b5 a4b3 h7h6 d3d4 d8c7 d4e5 d6e5 c1a3 c6c5 c3c4 c8d7 c4b5 d7b5 c2c4 b5d7 a3b2 a5c6 b2a3');
INSERT INTO evaluations VALUES(21587,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(21587,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(21587,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(21587,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(21587,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21587,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(21587,36,'cp 64','d4c6 b7c6 e4d5 d8d5 f1d3 d5e5 d1e2 e5e2 e1e2 f8d6 h1e1 g8e7 b1c3 e8g8 c3e4 c8f5 e2f1 f8e8 h2h3 f5g6 c1d2 a8b8 b2b3 d6a3 g2g4 e7d5 f2f4 f7f5 g4f5 g6f5 e4f6 d5f6 d3f5 g8f7 f5d3 f6d5 d3h7 a3b4 e1e8 b8e8 d2b4 d5b4');
INSERT INTO evaluations VALUES(21339,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(21339,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21339,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(21339,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(21339,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(21339,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(21339,36,'cp 49','b7c6 e4d5 d8d5 d1e2 d5e6 b1c3 g8f6 c1d2 f8d6 e1c1 e8g8 g2g3 a8b8 e2e6 c8e6 f1g2 f6g4 d2e1 g4e5 c3e4 e6d5 e1a5 e5g4 e4d6 d5g2 a5c7 g4f2 c7b8 g2h1 d1d2 f8b8 d2f2 b8d8 d6b7 d8b8 f2d2 h7h5 b7c5 b8e8 b2b3 g7g5 d2d7 e8e2 a2a4 g8f8 d7a7 e2h2');
INSERT INTO evaluations VALUES(18261,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(18261,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(18261,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(18261,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(18261,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(18261,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(18261,36,'cp 53','e4d5 d8d5 f1d3 d5e5 d1e2 e5e2 e1e2 g8e7 h1e1 c8f5 d3f5 e7f5 e2f1 e8d7 b2b3 h7h5 c1b2 f8d6 b1d2 a8e8 h2h3 c6c5 d2c4 d7c6 a1d1 f7f6 b2c3 e8e1 d1e1 h8b8 c3b2 c6d7 e1d1 b8e8 d1d5 d7e6 d5d3 h5h4 b2c3 g7g6 d3d1 e6f7 a2a4 e8e6');
INSERT INTO evaluations VALUES(18259,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(18259,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(18259,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(18259,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(18259,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(18259,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(18259,36,'cp 46','d8d5 d1e2 d5e6 b1c3 g8f6 c1d2 f8d6 e2e6 c8e6 f1a6 a8b8 e1c1 e8g8 h1e1 h7h6 h2h3 f8e8 a2a3 g8f8 f2f3 f6d5 c3e4 d6e7 a6d3 d5b6 e4c3 e7d6 d2e3 f7f5 a3a4 a7a5 b2b3 b6d7 e3d2 d7c5 d3f1 e6f7 e1e8 b8e8 c3b1 f5f4');
INSERT INTO evaluations VALUES(32169,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32169,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32169,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32169,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32169,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32169,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32169,36,'cp 0','d2d4 d8f6 b1c3 g8e7 h2h4 b8c6 c3e2 e8c8 c2c3 d6d5 e4e5 f6e6 e2f4 e6d7 c4b5 e7f5 d1d3 f5g3 e5e6 d7e7 b5c6 g3h1 c6b7 c8b7 d3b5 b7c8 b5a6 c8b8 a6b5 b8c8');
INSERT INTO evaluations VALUES(32172,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(32172,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(32172,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(32172,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32172,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32172,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(32172,36,'cp 6','d8f6 b1c3 g8e7 c3b5 b8a6 h2h3 g4f3 d1f3 c7c6 c1f4 e7g6 e1g1 g6f4 e4e5 d6e5 d4e5 f6g5 b5d4 e8c8 c4a6 g5g2 f3g2 f4g2 g1g2 d8d4 a6d3 f7f6 e5f6 g7f6 f1f6 d4d8 a1e1 h8g8 g2f3 h7h5 f6f7 h5h4 f7h7 f8d6 e1f1 d6g3 f3e2 g8e8 e2d2 e8f8 f1f7 f8f7 h7f7 d8d7 f7d7 c8d7 a2a3 d7e7 c2c3 a7a6');
INSERT INTO evaluations VALUES(47820,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(47820,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(47820,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(47820,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(47820,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(47820,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(47820,36,'cp 23','e5d4 d1d4 f8e7 c1f4 b8c6 d4d3 c8e6 b1d2 d8d7 e1g1 e6c4 d3c4 d7g4 f4g3 d6d5 c4b3 e8c8 e4d5 d8d5 d2c4 e7c5 c4e5 c6e5 f3e5 g4h5 a1e1 h8d8 e5d3 c5d6 g3d6 d5d6 e1e7 d8d7 e7d7 d6d7 f1e1 d7d8 g2g3 a7a6');
INSERT INTO evaluations VALUES(47453,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47453,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(47453,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(47453,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47453,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(47453,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(47453,36,'cp 14','c8f5 f1b5 c7c6 c3d5 d8d5 b5e2 e7e6 e1g1 f8e7 c2c4 d5d6 d2d4 b8d7 c1e3 e8g8 e2d3 f5d3 d1d3 a8d8 d3c2 d6c7 f1d1 f8e8 a1b1 h7h6 h2h3 e7f6 b2b4 d7f8 a2a4 f8g6 g2g3 g6e7 a4a5 e7f5 c2e2 f5e3 e2e3');
INSERT INTO evaluations VALUES(55550,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(55550,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(55550,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(55550,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(55550,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(55550,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(55550,36,'cp 10','e5d4 g1f3 g8f6 d1d4 f8e7 c1f4 b8c6 d4d3 c8e6 b1d2 d8d7 e1g1 e6c4 d3c4 d7e6 c4a4 f6d7 c2c3 e8g8 f1e1 e7f6 f4e3 d7e5 f3e5 c6e5 a4c2 f8e8 a2a4 h7h6 a1d1 a7a6 f2f4 e5g4 h2h3 g4e3 e1e3 f6e7 a4a5');
INSERT INTO evaluations VALUES(55666,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(55666,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55666,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(55666,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(55666,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(55666,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(55666,36,'cp 15','g1f3 g8f6 d1d4 f8e7 c1f4 b8c6 d4d3 c8e6 b1d2 e6c4 d3c4 d6d5 e4d5 d8d5 f4c7 c6b4 e1c1 a8c8 c4d5 f6d5 c7g3 b4c2 c1b1 e8g8 d2e4 f8d8 d1c1 c2b4 c1c8 d8c8 a2a3 b4c6 h1d1 c8d8 e4d6 e7d6 d1d5 d6g3 d5d8 c6d8 f2g3 a7a6 b2b4 d8c6');
INSERT INTO evaluations VALUES(21201,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(21201,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(21201,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(21201,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(21201,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(21201,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(21201,36,'cp 143','f2f4 e5d7 g1f3 c7c6 c2c4 g7g6 d1d4 g8f6 e4e5 f6g8 e5e6 d7f6 e6f7 e8f7 f3g5 f7g7 b1c3 g8h6 f1d3 h6f5 d4f2 d8b6 h2h3 h7h6 g5f3 h6h5 c1d2 c6d5 c3d5 f6d5 c4d5 g7g8 e1g1 f8g7 d2c3 g7c3 b2c3 b6e3 f2e3 f5e3 f1e1 e3d5 d3c4 e7e6');
INSERT INTO evaluations VALUES(21203,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(21203,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(21203,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(21203,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(21203,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(21203,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(21203,36,'cp 150','e5d7 g1f3 c7c6 c2c4 e7e5 d5e6 f7e6 f3g5 d7c5 b1c3 f8e7 b2b4 c5a6 a1b1 e6e5 f1d3 g8f6 e1g1 a6c7 f4f5 e8g8 g1h1 c8d7 a2a4 d7e8 b4b5 f6d7 g5f3 e8f7 c1e3 g8h8 a4a5 c6b5 c4b5 d7c5');
INSERT INTO evaluations VALUES(5180,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(5180,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(5180,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(5180,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(5180,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(5180,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(5180,36,'cp 71','d4b5 g8f6 c1f4 e8g8 f4c7 d8e7 c7d6 c5d6 d1d6 e7d8 e1c1 a7a6 b5d4 c6e7 d4b3 b7b5 h2h4 c8b7 f2f3 h7h5 f1d3 b7c6 c1b1 a8c8 g2g4 e7g6 b3c5 b5b4 g4h5 f6h5 c3e2 a6a5 d3a6 c8a8 e2d4 d8e7 d6e7 g6e7 d4c6 d7c6 c5d7 f8e8 a6c4 e7g6 e4e5');
INSERT INTO evaluations VALUES(5171,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(5171,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(5171,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(5171,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(5171,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(5171,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(5171,36,'cp 52','g8f6 c1f4 e8g8 f4c7 d8e7 c7d6 c5d6 d1d6 e7d8 b5c7 a8b8 f1e2 b7b5 e4e5 f6e8 c7e8 f8e8 e1g1 b5b4 c3e4 d8a5 e4c5 b8b6 f1d1 c6e7 c5b3 b6d6 b3a5 d6d5 a5c4 c8b7 a2a3 d5d1 a1d1 b4a3 c4a3 b7c6 d1a1 e8c8 b2b4 c6e4 c2c4 e7g6 a3b5 g6e5 b5a7');
INSERT INTO evaluations VALUES(42035,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42035,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42035,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(42035,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(42035,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(42035,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(42035,36,'cp 6','c4d5 f6d5 f4d2 e8g8 g1f3 c7c5 c3d5 b4d2 d1d2 d8d5 d4c5 d5c5 e2e3 b8c6 f1d3 f8d8 e1g1 c8d7 a1c1 c5b6 d2c2 h7h6 c2b1 c6e7 h2h3 d7b5 d3b5 b6b5 c1c2 a8c8 f1c1 c8c2 c1c2 d8c8 c2c8 e7c8 b1c2 c8e7 c2c7 e7c6 c7d7 g7g6 b2b3 g8g7 f3d4 b5e5 d4c6 e5a1 g1h2 b7c6 d7c6 a1a2');
INSERT INTO evaluations VALUES(44669,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44669,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44669,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44669,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44669,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44669,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44669,36,'cp 0','d7d6 g1f3 b8c6 f1c4 d6e5 d4e5 d5b4 e1g1 d8d1 f1d1 c8f5 b1c3 b4c2 a1b1 c2b4');
INSERT INTO evaluations VALUES(45281,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45281,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45281,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45281,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45281,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45281,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45281,36,'cp 17','g1f3 d6e5 d4e5 b8c6 f1c4 d5b4 e1g1 d8d1 f1d1 c8f5 b1c3 b4c2 a1b1 c2b4 e5e6 f7e6 f3g5 f5b1 g5e6 a8c8 c3b1 c6d8 b1a3 a7a6 c1e3 d8e6 c4e6 c8d8 e6f7 e8f7 d1d8 g7g6 d8d7 b7b5 e3c5 b4a2 d7a7 a2c1 a7a6 c1d3 c5e3 d3b2 a3b5 f8g7');
INSERT INTO evaluations VALUES(45280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45280,36,'cp 0','d6e5 d4e5 b8c6 f1c4 d5b4 e1g1 d8d1 f1d1 c8f5 b1c3 b4c2 a1b1 c2b4');
INSERT INTO evaluations VALUES(47739,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(47739,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(47739,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(47739,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(47739,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(47739,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(47739,36,'cp 147','g1f3 f8d6 f1e2 c8d7 e1g1 b8c6 e2d3 e6e5 d4d5 c6e7 a2a4 e8g8 a4a5 b6c8 c1e3 h7h6 f3d2 c7c6 d5c6 e7c6 d2c4 d6c7 d1h5 c8e7 f1d1 e7g6 d3f1 g6f4 e3f4 e5f4 a5a6 a8b8 a6b7 b8b7 c3d5 d7e6 d5f4 d8f6 f4d5 f6d8 d5c7 d8c7');
INSERT INTO evaluations VALUES(6690,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6690,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(6690,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(6690,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(6690,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(6690,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(6690,36,'cp -18','b4d2 d1d2 d7d6 f3g5 g8h6 b1c3 e8g8 f2f4 c6a5 c4d3 d6d5 e4e5 f7f6 g5h7 f8f7 d3g6 a5c4 d2d3 f6e5 d4e5 f7f4 d3d5 d8d5 c3d5 f4g4 e1g1 g4g6 h7f8 c8f5 f8g6 f5g6 e5e6 a8e8 a1c1 c4e5 c1c7 e8e6 c7b7 e6a6 b7e7 e5c4 e7c7 c4d6 d5e7 g8h7 e7g6 h7g6 f1f2 d6e8 c7c8 a6e6 h2h3 e8f6');
INSERT INTO evaluations VALUES(6704,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(6704,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(6704,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(6704,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(6704,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(6704,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(6704,36,'cp -7','d1d2 d7d6 f3g5 g8h6 b1c3 e8g8 f2f4 c6e7 e1g1 c7c6 e4e5 d6d5 c4d3 c8f5 c3e2 d8d7 e2g3 f5g6 a1e1 a8e8 h2h3 h6f5 g3f5 e7f5 g2g4 f5g3 f1f3 g3e4 d3e4 d5e4 g5e4 g6e4 e1e4 f7f5 g4f5 f8f5 d2c2 e8e7 g1h2 g7g6 f3c3 d7d5 a2a4');
INSERT INTO evaluations VALUES(6703,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6703,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6703,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6703,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6703,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6703,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(6703,36,'cp 0','d7d6 f3g5 g8h6 b1c3 e8g8 f2f4 c6a5 c4d3 d6d5 e4e5 f7f6 g5h7 f8f7 d3g6 a5c4 d2d3 f6e5 d4e5 f7f4 d3d5 d8d5 c3d5 f4g4 e1g1 g4g6 h7f8 g6a6 e5e6 c8e6 d5e7 g8h8 e7g6 h8g8');
INSERT INTO evaluations VALUES(6074,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6074,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6074,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6074,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6074,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6074,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6074,36,'cp 0','f3g5 g8h6 b1c3 e8g8 f2f4 c6e7 e1g1 c7c6 e4e5 e7g6 a1e1 d6d5 c4b3 f7f6 g5e4 f6e5 d4e5 h6f5 f1f2 d8b6 e4g5 b6d4 c3d5 c6d5 b3d5 g8h8 g5f7 h8g8 f7d6 g8h8');
INSERT INTO evaluations VALUES(12784,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(12784,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(12784,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(12784,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12784,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(12784,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12784,36,'cp 0','g1f3 e7e6 e1g1 d7d5 c4b5 c8d7 f1e1 f8e7 c2c3 a7a6 b5c6 d7c6 e4e5 f6d7 d3d4 a6a5 b1d2 a5a4 d2f1 d8a5 c1d2 a5b6 d2c1');
INSERT INTO evaluations VALUES(15627,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15627,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(15627,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(15627,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(15627,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(15627,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(15627,36,'cp 46','f8e7 c1f4 e8g8 e1g1 c6e5 c4b3 c8g4 f2f3 g4d7 a2a4 a7a6 d1e2 c7c5 f4e5 d6e5 d4f5 d7f5 e4f5 d8c7 g2g3 a8d8 b3c4 d8d4 c3e4 f8d8 c2c3 d4d7 a1d1 f6e4 f3e4 d7d1 f1d1 c7c6 d1d8 e7d8 b2b3 c6c7 e2h5 d8f6 c4d5');
INSERT INTO evaluations VALUES(8642,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(8642,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(8642,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(8642,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(8642,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(8642,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(8642,36,'cp 52','e1g1 e8g8 c1f4 c6e5 c4b3 c8g4 f2f3 g4d7 d1e2 c7c6 a2a4 a7a6 a4a5 c6c5 d4f5 d7f5 f4e5 d6e5 e4f5 d8c7 c3e4 a8d8 e4d2 f6e8 d2c4 e7f6 g2g4 e8d6 c4b6 c5c4 b6c4 e5e4 f3e4 d6c4 e2c4 c7c4 b3c4 d8d2 c4d5 d2c2 a1c1 c2d2 c1c7 d2b2 c7b7 b2b7 d5b7');
INSERT INTO evaluations VALUES(31157,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(31157,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31157,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(31157,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(31157,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31157,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(31157,36,'cp 29','h7h5 f1d3 f5d3 d1d3 d8a5 b1d2 e7e6 g1f3 g8h6 e1g1 h6f5 c2c3 f8e7 g2g3 a5a6 d3a6 b8a6 b2b4 b7b6 a2a4 e8g8 d2b3 c6c5 b4b5 a6b8 a4a5 b8d7 c1g5 f8c8 a5b6 a7b6 a1a6 c5c4 b3d2 a8a6 b5a6 c8a8 f1a1 e7g5 h4g5 f5e7 a6a7 e7c8');
INSERT INTO evaluations VALUES(31003,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(31003,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(31003,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31003,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(31003,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(31003,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31003,36,'cp 30','f1d3 f5d3 d1d3 e7e6 g1f3 d8a5 b1d2 g8e7 c2c3 e7f5 a2a4 a5a6 d3a6 b8a6 b2b4 b7b6 e1g1 f8e7 g2g3 e8g8 d2b3 c6c5 d4c5 b6c5 b4b5 a6b8 c1d2 f8c8 f1b1 b8d7 b3a5 c5c4 a5c6 e7c5 g1g2 f7f6 b1e1 a7a6 d2f4 g8f7 f3d4 a6b5 d4b5 f6e5');
INSERT INTO evaluations VALUES(31002,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31002,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(31002,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31002,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(31002,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(31002,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(31002,36,'cp 32','f5d3 d1d3 e7e6 g1f3 c6c5 d4c5 f8c5 c2c4 b8c6 e1g1 g8e7 b1c3 c5b4 c1g5 d5c4 d3c4 b4c3 c4c3 e7d5 c3c6 b7c6 g5d8 a8d8 a1c1 d5e7 b2b3 d8d5 f1d1 d5d1 c1d1 e7d5 g2g3 e8d7 f3g5 f7f6 e5f6 g7f6 g5e4 f6f5 e4g5 h8f8 g5f3 f8g8');
INSERT INTO evaluations VALUES(30998,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(30998,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(30998,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(30998,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(30998,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(30998,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(30998,36,'cp 37','d1d3 e7e6 g1f3 d8a5 b1d2 g8e7 c2c3 e7f5 a2a4 c6c5 d4c5 a5c5 e1g1 b8d7 d2b3 c5c4 d3c4 d5c4 b3d2 f8e7 d2c4 a8c8 c4a3 d7c5 g2g3 e8g8 a3c2 f8d8 c2e3 f5h6 a4a5 d8d7 g1g2 h6g4 e3c4 c5d3 b2b3 d7d5 c1g5 e7g5 h4g5 g4e5');
INSERT INTO evaluations VALUES(30997,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(30997,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(30997,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(30997,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(30997,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(30997,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30997,36,'cp 36','e7e6 g1f3 d8a5 b1d2 g8e7 c2c3 e7f5 a2a4 c6c5 d4c5 a5c5 e1g1 b8d7 d2b3 c5c4 d3c4 d5c4 b3d2 f8e7 d2c4 a8c8 c4a3 d7c5 g2g3 e8g8 a3c2 c5b3 a1b1 f8d8 c1g5 e7g5 h4g5 b3c5 c2d4 f5d4 f3d4 c5a4 g5g6 a7a6 f1e1 a4c5 b1d1 f7g6');
INSERT INTO evaluations VALUES(31400,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(31400,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31400,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31400,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31400,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(31400,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31400,36,'cp 30','g1f3 d8a5 b1d2 g8e7 c2c3 a5a6 d3a6 b8a6 b2b4 e7g6 g2g3 f8e7 e1g1 a6b8 a2a4 b8d7 a4a5 f7f6 f1e1 f6e5 d4e5 e8g8 g1g2 f8f5 d2b3 g6e5 f3d4 e5d3 e1e2 f5e5 e2e5 d7e5 d4e6 g8f7 e6g5 e7g5 c1g5 e5c4 a5a6 b7b6 b3d4 a8c8 a1d1');
INSERT INTO evaluations VALUES(6464,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(6464,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(6464,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(6464,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(6464,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6464,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(6464,36,'cp 23','g8f6 g1f3 e8g8 e1g1 c5b6 d2d3 d7d5 c1g5 d5c4 d3c4 d8d1 a1d1 c8e6 c3d5 f6e4 g5e3 b6e3 d5e3 f7f5 f3d2 e4d2 d1d2 a8d8 f1d1 d8d2 d1d2 e5e4 f2f3 f5f4 g3f4 e4f3 g2f3 f8f4 b2b3 f4d4 d2d4 c6d4 f3b7');
INSERT INTO evaluations VALUES(21413,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21413,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21413,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21413,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21413,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(21413,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21413,36,'cp 9','g8f6 f1d3 c6b4 e1g1 b4d3 c2d3 f8e7 b2b4 c7c6 d1b3 a7a5 b4b5 a5a4 b3b2 e8g8 c1d2 c8d7 a1b1 h7h6 b5c6 d7c6 f1c1 a4a3 b2b3 a8a5 f3e5 d8a8 c3e2 a5b5 b3c2 f8b8 b1b5 c6b5 c2c7');
INSERT INTO evaluations VALUES(22894,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(22894,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(22894,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(22894,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(22894,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(22894,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22894,36,'cp -40','g1f3 f8b4 b1c3 g8f6 e1g1 b4c3 b2c3 e8g8 e4e5 f6e4 c3b2 d7d6 f1e1 e4c5 e5d6 c7d6 a1c1 c8e6 f3d4 c6d4 d1d4 d8g5 e1e3 e6c4 e3g3 c5e6 g3g5 e6d4 b2d4 f7f6 d4f6 f8f6 c1c4 h7h6 g5b5 a8e8 f2f3 e8e7 a2a4 f6e6 h2h4 e6e1 g1h2 e1e5 b5e5 d6e5 h4h5 g8f7 h2g3 b7b6 c4c6 e5e4 f3e4 e7e4 c6c7 e4e7');
INSERT INTO evaluations VALUES(30071,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(30071,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(30071,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30071,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(30071,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(30071,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(30071,36,'cp 15','d1b3 b7b6 b1c3 h7h6 c4d5 e6d5 e3e4 f6e4 b3d5 d8d5 c3d5 f8d6 f1d3 e8g8 c1f4 b8c6 e1g1 d6f4 d5f4 c6b4 d3b1 e4d6 a2a3 b4c2 b1c2 f5c2 f1c1 c2e4 f3e5 f7f6 f4e6 f6e5 e6f8 g8f8 d4e5 d6b5 a3a4 b5d4 c1c7 e4f5 a1d1');
INSERT INTO evaluations VALUES(52607,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(52607,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(52607,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(52607,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(52607,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(52607,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(52607,36,'cp -76','g8f6 g1e2 f8g7 d2d3 b8c6 e1g1 e8g8 b1c3 a7a6 c4b3 a8b8 c3d5 f6d5 b3d5 d8d7 f3g3 b7b5 c2c3 a6a5 a2a3 e7e6 d5c6 d7c6 g3h4 f7f5 e2g3 e6e5 e4f5 c8f5 g3f5 f8f5 c1h6 g7h6 h4h6 b5b4');
INSERT INTO evaluations VALUES(44750,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(44750,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(44750,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(44750,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(44750,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(44750,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(44750,36,'cp -71','d2d3 f8g7 g1e2 a7a6 e1g1 e8g8 a2a4 b8c6 h2h3 c8d7 f3g3 g8h8 g3e3 c6b4 e3d2 b7b5 c4b3 a8c8 a4b5 a6b5 b1a3 c8b8 d2d1 b4c6 c2c3 b5b4 a3b1 d8c8 a1a2 h7h6 b1d2 b4c3 b2c3 h8h7 f2f4');
INSERT INTO evaluations VALUES(44745,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(44745,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(44745,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(44745,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(44745,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(44745,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(44745,36,'cp -80','f8g7 g1e2 e8g8 e1g1 b8c6 b1c3 a7a6 a2a4 c6b4 c4b3 c8g4 f3f4 g4e6 b3e6 b4c2 e6f7 f8f7 a1b1 c2b4 f4d2 e7e5 d2d1 h7h6 c1e3 d8d7 f2f3 b7b6 d1b3 a6a5 e3f2 g8h7 f2h4 g6g5 h4e1 a8d8 c3b5 d6d5 e1b4 a5b4');
INSERT INTO evaluations VALUES(40336,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(40336,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(40336,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(40336,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(40336,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(40336,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(40336,36,'cp -65','b1c3 a7a6 g1e2 b8c6 e1g1 e8g8 a2a4 c8g4 f3f4 c6b4 c4b3 g4e6 b3e6 b4c2 e6f7 f8f7 a1b1 e7e5 f4d2 c2b4 d2d1 h7h6 c1e3 g8h7 d1b3 d8d7 f2f3 a8f8 g1h1 b7b6 c3a2 a6a5 b1a1 b4c6 a2c3 f8d8 c3d5 f6d5 e4d5 c6b4 e2c3 f7f8 a1d1 d7f7 b3c4');
INSERT INTO evaluations VALUES(55616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55616,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55616,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55616,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55616,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55616,36,'cp 0','f1c4 g8f6 g1e2 f8e7 b1c3 e8g8 e1g1 c7c6 a2a4 c8g4 f3g3 g4h5 g3h3 h5g4');
INSERT INTO evaluations VALUES(1724,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(1724,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(1724,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(1724,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(1724,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(1724,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(1724,36,'cp -24','d2d4 g8f6 c1e3 f8d6 f1e2 e8g8 b1d2 e5d4 c3d4 f8e8 e1g1 h7h6 h2h3 a7a6 a1c1 b7b5 a2a3 c8e6 d1c2 c6e7 f3e5 e6f5 c2d1 a8d8 e2f3 d5e6 e5c6 e7c6 f3c6 f5h3 c6e8 d8e8 f1e1 h3g4 d2f3 e6f5 e3d2 e8e1 d2e1 f6d5 e1d2 d5f4 d2f4 f5f4');
INSERT INTO evaluations VALUES(1713,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(1713,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(1713,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(1713,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(1713,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(1713,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(1713,36,'cp -40','g8f6 c1e3 e5d4 c3d4 f8b4 b1c3 d5a5 d1b3 e8g8 f1c4 a5h5 c4e2 c8e6 b3c2 f8e8 e1g1 b4d6 f3e5 f6g4 e2g4 e6g4 e5g4 h5g4 a1d1 a7a6 h2h3 g4h5 f1e1 h7h6 c2b3 a8b8 a2a3 h5g6 c3e2 c6e7 e3f4 d6f4 e2f4 g6d6 f4h5 e7f5 e1e8 b8e8 b3b7');
INSERT INTO evaluations VALUES(7487,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(7487,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(7487,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(7487,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7487,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(7487,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(7487,36,'cp 25','a2a4 b5b4 a4a5 e8g8 e1g1 d7d6 c2c3 d6d5 e4d5 f6d5 d3d4 e5e4 f3e5 c6e5 d4e5 c7c6 c3c4 d5c7 c1e3 c8f5 b3a4 d8e8 f1e1 c7e6 e3b6 e6f4 b1d2 f4d3 d2e4 f5e4 e1e4 d3b2 d1g4 b2a4');
INSERT INTO evaluations VALUES(12538,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12538,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12538,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12538,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12538,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12538,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12538,36,'cp 0','d2d3 a7a6 f1e2 f8e7 a2a4 e8g8 e1g1 d7d6 f3e1 c8e6 c1d2 c6d4 e1c2 d4b3 a1b1 a8b8 c2e3 b3d4 e3c2');
INSERT INTO evaluations VALUES(12542,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12542,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12542,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12542,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12542,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12542,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12542,36,'cp 0','a7a6 f1e2 f8e7 e1g1 e8g8 a2a4 d7d6 f3e1 c8e6 c1d2 c6d4 e1f3 e6d7 h2h3 a8b8 b2b3 d4c6 f3e1 c6d4');
INSERT INTO evaluations VALUES(21670,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(21670,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(21670,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(21670,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(21670,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(21670,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(21670,36,'cp 52','c4d5 d8d5 b1c3 f8b4 a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 e5e4 f3d4 c7c6 d5c6 e7c6 f1e2 c6e5 e1g1 e8g8 a3a4 c8g4 e2g4 f6g4 c1a3 f8e8 h2h3 g4f6 c4c5 d8d7 d1b3 b7b6 f1c1 e5f3 d4f3 e4f3 c5c6 d7f5 e3e4 f5e4 c1c4 e4e2');
INSERT INTO evaluations VALUES(21571,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(21571,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(21571,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(21571,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(21571,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(21571,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(21571,36,'cp 51','d8d5 b1c3 f8b4 a2a3 b4c3 b2c3 g8f6 c3c4 d5d8 d4d5 c6e7 g1f3 e5e4 f3d4 e8g8 f1e2 c7c6 d5c6 e7c6 e1g1 c6e5 a3a4 c8g4 c1a3 f8e8 e2g4 f6g4 h2h3 g4f6 d1b3 d8d7 f1d1 e5d3 c4c5 h7h6 a1b1 a8b8 d4b5 b7b6 f2f3 a7a6 f3e4 a6b5 b3d3 d7d3');
INSERT INTO evaluations VALUES(48307,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(48307,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(48307,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(48307,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(48307,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(48307,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48307,36,'cp -33','c7c5 c2c3 f8e7 f1d3 b7b6 e1g1 e8g8 d1e2 c8b7 b2b3 b8d7 c1b2 f6e4 b1d2 c5d4 c3d4 e7b4 a2a3 b4d6 f3e5 e4d2 e2d2 d7f6 f1c1 d8e7 a3a4 f8c8 c1c8 a8c8 a1c1 e7d8 c1c8 d8c8 d2c3 c8e8 d3b5 e8a8 c3c2 f6e4 b5d3 f7f6 e5f3');
INSERT INTO evaluations VALUES(46221,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(46221,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(46221,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(46221,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(46221,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(46221,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(46221,36,'cp -32','c2c3 f8e7 f1d3 e8g8 e1g1 b7b6 d1e2 f6e4 b2b3 c8b7 c1b2 b8d7 b1d2 c5d4 c3d4 d7f6 f3e5 e4d2 e2d2 f6e4 d2e2 a8c8 f1c1 d8d6 a2a3 a7a5 d3e4 d5e4 e2b5 d6d5 b5d5 c8c1 a1c1 b7d5 e5d7 f8d8 c1c7 d5b3 d7b6');
INSERT INTO evaluations VALUES(46218,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(46218,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(46218,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(46218,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(46218,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(46218,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(46218,36,'cp -23','f8e7 f1d3 e8g8 e1g1 b7b6 d1e2 c8b7 b2b3 b8d7 b1d2 f6e4 c1b2 c5d4 c3d4 d7f6 f3e5 e4d2 e2d2 f6e4 d2e2 a8c8 a1c1 d8d6 a2a3 f7f6 e5g4 d6d7 g4f2 c8c1 f1c1 f8c8 c1c8 d7c8 f2e4 d5e4 d3c4 c8d7 f4f5 b7d5 f5e6 d5e6 d4d5');
INSERT INTO evaluations VALUES(49831,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(49831,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(49831,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(49831,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(49831,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(49831,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(49831,36,'cp -45','e4g5 f3e3 f8e7 f2f4 f7f6 e5f3 g5f7 f1e2 d7d5 e1g1 e8g8 e3f2 f8e8 b1c3 c7c6 h2h3 b8d7 f1e1 d7f8 d2d3 d8c7 f3d4 e7d6 c1d2 c8d7 a2a4 f8e6 d4e6 e8e6 e2h5 g7g6 e1e6 d7e6 h5d1 c7a5 c3b1 a5b6 f2b6 a7b6 d2e3');
INSERT INTO evaluations VALUES(49819,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(49819,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(49819,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(49819,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(49819,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(49819,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(49819,36,'cp -27','f3e2 g5e6 e5f3 d7d5 d2d4 c7c6 c2c3 f8d6 e2c2 b8d7 f1d3 d7f6 e1g1 e8g8 h2h3 h7h6 c3c4 d5c4 d3c4 e6c7 b1c3 c8e6 c4e6 c7e6 c1e3 f8e8 a1d1 e6c7 f1e1 a7a6 e3d2 e8e1 d1e1 c7d5 c3d5 f6d5 c2b3 d8b6 b3c2 b6c7 a2a3');
INSERT INTO evaluations VALUES(57211,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(57211,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(57211,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(57211,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(57211,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(57211,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(57211,36,'cp -34','d7d5 f1g2 g8f6 g1f3 f8d6 e2e4 d5e4 d3e4 e8g8 d1e2 b8d7 e1g1 d7c5 f3h4 g7g6 f1d1 d8e7 c2c4 c7c6 h2h3 c5e6 h4f3 a7a5 b1c3 f6d7 c1e3 a5a4 a1b1 d6c5 b2b3 a4b3 a2b3 c5e3 e2e3');
INSERT INTO evaluations VALUES(56458,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(56458,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(56458,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(56458,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(56458,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(56458,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(56458,36,'cp -7','g1f3 f8d6 f1g2 g8f6 c2c4 c7c6 c4d5 c6d5 e1g1 e8g8 b1c3 h7h6 e2e4 d5d4 c3b5 c8g4 d1d2 g4f3 g2f3 d6e7 a2a4 b8d7 b5a3 a8c8 a4a5 e7a3 a1a3 c8c6 f3d1 b7b5 a5b6 d8b6 d1a4 c6c7 a4b3');
INSERT INTO evaluations VALUES(56456,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56456,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(56456,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(56456,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(56456,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(56456,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(56456,36,'cp -37','f8d6 e2e4 d5e4 d3e4 g8f6 f1g2 e8g8 d1e2 b8d7 f3h4 g7g6 e1g1 c7c6 a2a4 a7a5 b1a3 f8e8 a3c4 d6f8 c1d2 b7b6 b2b3 c8a6 f1e1 d8c8 g2h3 c8c7 h3g2 f8b4 a1d1 a6c4 e2c4 b4d2 d1d2 d7c5 h4f3 g8g7');
INSERT INTO evaluations VALUES(27694,30,'cp 324',NULL);
INSERT INTO evaluations VALUES(27694,31,'cp 318',NULL);
INSERT INTO evaluations VALUES(27694,32,'cp 340',NULL);
INSERT INTO evaluations VALUES(27694,33,'cp 334',NULL);
INSERT INTO evaluations VALUES(27694,34,'cp 331',NULL);
INSERT INTO evaluations VALUES(27694,35,'cp 340',NULL);
INSERT INTO evaluations VALUES(27694,36,'cp 336','d5c6 b7c6 f3c6 f8d6 b1c3 a7a5 c3e4 f6e7 d2d3 h7h6 e1g1 h8e8 a2a4 f7g8 c1d2 e8d8 f2f3 e7f7 c6c3 e6d5 d2e3 f7e7 e4d6 c7d6 c3d2 e7b7 b2b3 d8f8 h2h3 f8f6 g1h2 d5c6 a1e1 f6e6 f1f2 e6g6 d3d4 e5e4');
INSERT INTO evaluations VALUES(27691,30,'cp 339',NULL);
INSERT INTO evaluations VALUES(27691,31,'cp 330',NULL);
INSERT INTO evaluations VALUES(27691,32,'cp 320',NULL);
INSERT INTO evaluations VALUES(27691,33,'cp 317',NULL);
INSERT INTO evaluations VALUES(27691,34,'cp 348',NULL);
INSERT INTO evaluations VALUES(27691,35,'cp 340',NULL);
INSERT INTO evaluations VALUES(27691,36,'cp 311','b7c6 f3c6 f8b4 a2a3 b4a5 e1g1 a8d8 d2d3 h8e8 b1c3 e6f5 c6a4 a5c3 a4c4 f7f8 c4c3 e5e4 d3d4 c7c6 c1e3 f8g8 f1e1 h7h5 b2b4 f6g6 e3f4 h5h4 h2h3 f5e6 c3c6 d8c8 c6d6 c8c2 d4d5');
INSERT INTO evaluations VALUES(27688,30,'cp 343',NULL);
INSERT INTO evaluations VALUES(27688,31,'cp 337',NULL);
INSERT INTO evaluations VALUES(27688,32,'cp 340',NULL);
INSERT INTO evaluations VALUES(27688,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(27688,34,'cp 328',NULL);
INSERT INTO evaluations VALUES(27688,35,'cp 335',NULL);
INSERT INTO evaluations VALUES(27688,36,'cp 336','f3c6 f8d6 b1c3 h8f8 c3e4 f6g6 e1g1 f7g8 d2d3 a7a5 f2f3 a5a4 c1d2 e6f7 b2b3 d6a3 c6g6 f7g6 d2e3 f8b8 g1f2 g6e4 d3e4 a3b2 a1d1 a4b3 a2b3 a8a2 f2e2 b2c3 e2d3 c3b4 e3d2 a2b2 d2b4 b8b4');
INSERT INTO evaluations VALUES(27687,30,'cp 340',NULL);
INSERT INTO evaluations VALUES(27687,31,'cp 348',NULL);
INSERT INTO evaluations VALUES(27687,32,'cp 352',NULL);
INSERT INTO evaluations VALUES(27687,33,'cp 345',NULL);
INSERT INTO evaluations VALUES(27687,34,'cp 354',NULL);
INSERT INTO evaluations VALUES(27687,35,'cp 348',NULL);
INSERT INTO evaluations VALUES(27687,36,'cp 353','f8d6 b1c3 h8b8 e1g1 a7a5 d2d3 f6e7 c3e4 h7h6 a2a4 e6d7 c6c4 d7e6 c4c3 f7g8 b2b3 e6d5 c3e1 d6b4 e1e2 a8a6 c1e3 b8f8 f2f3 a6c6 a1d1 b4a3 g1h1 f8f7 h2h3 a3b4 h1h2 b4c3 e3d2 d5e4 e2e4');
INSERT INTO evaluations VALUES(27718,30,'cp 347',NULL);
INSERT INTO evaluations VALUES(27718,31,'cp 334',NULL);
INSERT INTO evaluations VALUES(27718,32,'cp 350',NULL);
INSERT INTO evaluations VALUES(27718,33,'cp 342',NULL);
INSERT INTO evaluations VALUES(27718,34,'cp 363',NULL);
INSERT INTO evaluations VALUES(27718,35,'cp 348',NULL);
INSERT INTO evaluations VALUES(27718,36,'cp 359','b1c3 h8e8 d2d3 f7g8 c3e4 f6d8 e1g1 d6f8 c1e3 a7a5 c6c3 e6d5 b2b3 f8b4 c3b2 d8d7 a2a3 b4d6 f2f3 e8e6 g1h1 h7h6 b3b4 d5c6 b2c3 d6e7 f1b1 c6e4 f3e4 a5b4 a3b4 a8b8 b4b5 b8b5 b1b5 d7b5 c3c7 b5c6 c7c6 e6c6');
INSERT INTO evaluations VALUES(47903,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47903,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(47903,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(47903,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(47903,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(47903,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(47903,36,'cp 33','g1f3 b8c6 f1e2 g7g6 f4e5 d6e5 e2b5 c8d7 d2d3 c6d4 b5d7 f6d7 e1g1 c7c6 f3d4 e5d4 d1f3 d7e5 f3g3 d8d6 c1f4 f8g7 c3b1 e8g8 b1d2 a8e8 b2b3 d6d7 h2h3 h7h6 d2c4 e5c4 b3c4 e8e6 a2a4 f7f5 e4f5 f8f5 g3g4 g8h7 f4g3 f5f1 a1f1 d7e7 g4h4 e7h4');
INSERT INTO evaluations VALUES(20865,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(20865,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(20865,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(20865,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(20865,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(20865,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(20865,36,'cp 28','c8g4 f1b5 d8d7 e1g1 e7e6 b1d2 a7a6 b5e2 g8h6 h2h3 g4f5 c2c3 f6e5 f3e5 c6e5 d4e5 h6f7 f2f4 g7g5 d2b3 g5f4 b3d4 f5e4 e2g4 f7e5 g4e6 d7g7 d1h5 g7g6 h5g6 e5g6 c1f4 g6f4 f1f4 f8d6 f4f2 h8f8 f2f8 e8f8 a1f1 f8e7 f1f7 e7e8 f7g7');
INSERT INTO evaluations VALUES(26471,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26471,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(26471,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(26471,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(26471,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(26471,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(26471,36,'cp 22','f1b5 d8d7 e1g1 e7e6 b1d2 a7a6 b5e2 g8h6 h2h3 g4h5 c2c4 h5f3 d2f3 h6f5 b2b4 d5c4 e2c4 e8c8 c1e3 f5d4 f3d4 c6d4 e5f6 g7f6 e3d4 d7d4 c4e6 c8b8 a2a3 d4d1 f1d1 f8d6 g2g3 d6e5 a1b1 c7c6 h3h4 h7h6 d1d8 h8d8 f2f4 e5c3 b1b3 c3d4 g1g2 d4b6 h4h5 d8d4 g2h3 d4d2');
INSERT INTO evaluations VALUES(26457,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26457,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(26457,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(26457,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(26457,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26457,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(26457,36,'cp 5','f6e5 d4e5 e7e6 b1d2 g8h6 h2h3 g4h5 g2g4 h5g6 d2b3 a7a6 b5c6 b7c6 e1g1 c6c5 f3g5 d8d7 c2c4 h6f7 g5f7 g6f7 c4d5 h7h5 c1e3 h5g4 d5d6 c7d6 d1g4 c5c4 g4c4 d6d5 c4g4 f7h5 g4g3 f8b4 b3d4 e8g8 f2f4 b4c5 g1h2 c5d4');
INSERT INTO evaluations VALUES(29704,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(29704,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(29704,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(29704,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(29704,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(29704,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(29704,36,'cp 52','e7e6 g1f3 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 e8g8 c3c4 c6c5 a1b1 c5d4 e3d4 b8d7 c1f4 f8e8 e1d1 b7b6 a2a4 d8c8 d3b3 c8a6 b1c1 a8c8 b3b5 a6b5 a4b5 f7f6 f4e3 e6e5 d4e5 d7e5');
INSERT INTO evaluations VALUES(30194,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(30194,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(30194,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(30194,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(30194,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(30194,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(30194,36,'cp 59','g1f3 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 b8d7 c3c4 e8g8 c1b2 d8c7 e1d1 a8d8 e3e4 f8e8 a1c1 h7h6 h2h3 e7f8 d3c2 e6e5 d4e5 d7c5 g2g3 c5e6 g1g2 d8d1 c1d1 e8d8 f3d4 e6c5 d4f5 d8d1 c2d1 c5e4');
INSERT INTO evaluations VALUES(45419,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45419,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45419,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45419,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45419,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45419,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45419,36,'cp 0','g7g6 c4d5 f6d5 d1b3 f8g7 f1c4 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(44807,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44807,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44807,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44807,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44807,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44807,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44807,36,'cp 0','c4d5 f6d5 d1b3 f8g7 f1c4 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(1072,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(1072,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(1072,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(1072,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(1072,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(1072,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(1072,36,'cp 87','d7d5 c4d5 f6d5 e4d5 e7f7 d5c6 f8c5 f2f3 e8g8 b1c3 c5d4 d1e2 b7c6 d2d3 c8f5 c3d1 h7h5 c1e3 h5h4 c2c3 d4b6 d1f2 a8d8 h2h3 f7g6 e1c1 f5e6 f2g4 e6a2 g4e5 g6e6 e3b6 a7b6 d1e1 a2b3 c1d2 c6c5 h1g1 c5c4 e2e4 c4d3 g2g3 h4g3 g1g3 e6h6 e4e3 h6e3 e1e3 b3d5 c3c4 d5e6 e3d3 e6f5 d3d8');
INSERT INTO evaluations VALUES(6890,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(6890,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(6890,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(6890,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(6890,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(6890,35,'cp 257',NULL);
INSERT INTO evaluations VALUES(6890,36,'cp 239','d1d4 d7d6 c1e3 g8f6 e1c1 c8e6 f2f4 e5c6 d4d2 d8e7 f1b5 e8g8 f4f5 e6d7 e3g5 e7e5 g5f6 e5f6 c3d5 f6h4 g2g3 h4e4 h1e1 e4g4 f5f6 a8e8 e1e8 f8e8 d5c7 e8c8 c7d5 g8h8 b5c6 d7c6 d5e7 c8e8 e7c6 b7c6 f6g7 h8g7 d2d6 g4g5 c1b1 g5e5');
INSERT INTO evaluations VALUES(27205,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(27205,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(27205,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(27205,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(27205,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(27205,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(27205,36,'cp 26','g8f6 d3d4 e5d4 f3d4 c6d4 b5d7 d8d7 d1d4 f8e7 c1f4 d7g4 f4e3 e8g8 e1g1 a7a6 f2f3 g4c8 a1d1 c7c6 f1e1 f8e8 c3a4 f6d7 a4b6 d7b6 d4b6 d6d5 e4d5 c6d5 c2c3 c8c6 a2a4 c6a4 d1d5 e7f6 b6c7 a4c6 c7c6 b7c6');
INSERT INTO evaluations VALUES(25507,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(25507,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(25507,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25507,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25507,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(25507,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(25507,36,'cp 28','d3d4 e5d4 f3d4 f8e7 d4c6 d7c6 b5c6 b7c6 e1g1 e8g8 f2f4 d8c8 d1f3 c8g4 f3d3 g4e6 c1d2 a7a5 f1e1 a5a4 a1d1 f8b8 b2b3 a4b3 a2b3 b8e8 h2h3 f6d7 d3f3 e7h4 e1e2 h7h6 g1h2 g7g6 d2e3 a8b8 e3d4 h4f6 d4f6 d7f6');
INSERT INTO evaluations VALUES(54958,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(54958,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(54958,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(54958,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54958,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(54958,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(54958,36,'cp 35','e7e5 g2g3 f8e7 b1c3 f7f5 f1g2 g8f6 e1g1 e8g8 b2b4 a7a6 d2d3 d8e8 a2a4 b8c6 a1b1 a6a5 b4a5 a8a5 e2e3 a5a8 f3d2 c6d8 f2f4 c7c6 d2f3 d8f7 c1a3 f6g4 d1d2 g4f6 h2h3 e7d8 g1h2 d8c7 d2c2 g8h8 f1g1 e5e4 d3e4');
INSERT INTO evaluations VALUES(55563,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(55563,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(55563,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(55563,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(55563,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(55563,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(55563,36,'cp 36','b1c3 f7f5 g2g3 f8e7 f1g2 g8f6 b2b4 e8g8 d2d3 a7a6 e1g1 d8e8 a2a4 b8c6 a1b1 a6a5 b4a5 a8a5 e2e3 a5a8 f3d2 c6d8 f2f4 c7c6 d2f3 d8f7 c1a3 f6g4 d1d2 g4f6 h2h3 g8h8 f1e1 e5e4 f3d4 e4d3');
INSERT INTO evaluations VALUES(57456,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(57456,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(57456,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(57456,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(57456,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(57456,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(57456,36,'cp 232','f3d4 g8e7 c1e3 b8c6 a2a3 d7d6 b1c3 c6d4 d1d4 c8e6 e1c1 e7c6 d4d3 f8e7 f1e2 e8g8 f2f4 g8h8 g2g4 c6a5 c1b1 c7c6 h2h4 a5c4 e3c1 b7b5 g4g5 a7a5 d3g3 b5b4 c3a4 e6g8 e2c4 g8c4 b2b3 c4f7 h4h5');
INSERT INTO evaluations VALUES(57453,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(57453,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(57453,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(57453,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(57453,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(57453,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(57453,36,'cp 236','g8e7 c1e3 b8c6 a2a3 c6d4 d1d4 e7c6 d4d2 f8d6 h2h4 b7b6 b1c3 c8b7 e1c1 c6e5 h4h5 d8e7 c3b5 e8c8 b5d6 e7d6 d2d6 c7d6 f2f3 c8c7 f1e2 d8e8 e3f2 b7c6 f2g3 h7h6 b2b4 b6b5 d1d2 e5f7');
INSERT INTO evaluations VALUES(13350,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(13350,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(13350,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(13350,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(13350,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(13350,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(13350,36,'cp 13','e4e5 f6g4 e5d6 e7d6 e1g1 f8e7 h2h3 g4f6 f1e1 e8g8 c1f4 c8f5 f3d4 f5d7 d4b5 c6e5 c4b3 d7c6 b5d4 f8e8 f4e5 d6e5 d4c6 b7c6 e1e5 d8c7 d1d4 a8d8 d4f4 e8f8 e5f5 c7f4 f5f4 d8d2 f4c4 d2b2 a1d1 f8d8 c4c6 b2d2 g2g3 d8d7 d1d2 d7d2 c6c7 d2d7 c7d7 f6d7');
INSERT INTO evaluations VALUES(13347,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13347,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(13347,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13347,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(13347,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(13347,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(13347,36,'cp 9','f6g4 e5d6 e7d6 e1g1 f8e7 h2h3 g4f6 f1e1 e8g8 c1f4 a7a6 a1c1 c8f5 f3d4 c6d4 d1d4 a8c8 c3d5 f6d5 c4d5 c8c1 e1c1 f5e6 d5b7 e7f6 d4d6 d8d6 f4d6 f8d8 d6c7 f6b2 c1c2 d8d1 g1h2 b2d4 b7a6 d1a1 a2a3 a1a3');
INSERT INTO evaluations VALUES(44969,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(44969,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(44969,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(44969,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(44969,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(44969,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(44969,36,'cp 40','e4e5 f6e4 d2d4 e4c3 b2c3 e7e6 c3c4 c6c5 c4d5 e6d5 c2c3 b8c6 f1e2 c5d4 c3d4 f8e7 e1g1 c8f5 e2d3 f5d3 d1d3 e8g8 c1d2 d8d7 a2a4 f8c8 h2h4 c8c7 a1b1 a8c8 f1c1 h7h6 c1c3 d7g4 b1b5 g4e4 h4h5 e4d3 c3d3');
INSERT INTO evaluations VALUES(44994,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(44994,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(44994,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(44994,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(44994,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44994,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(44994,36,'cp 26','f6e4 d2d4 e4c3 b2c3 e7e6 c3c4 c6c5 c4d5 e6d5 c2c3 b8c6 f1d3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 c8g4 d3e2 d8a5 d2a5 c6a5 f3h4 g4d7 f2f4 e8g8 a1b1 a5c6 b1b7 d7c8 b7b5 c6d4 b5d5 d4e2 e1e2 g7g6 e2e3 a8b8 h4f3 c8e6 d5d6 f8c8');
INSERT INTO evaluations VALUES(45073,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(45073,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(45073,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(45073,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45073,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(45073,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(45073,36,'cp 40','d2d4 e4c3 b2c3 e7e6 c3c4 c6c5 c4d5 e6d5 c2c3 b8c6 f1d3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 c8g4 d3e2 d8a5 d2a5 c6a5 f3h4 g4e2 e1e2 e8g8 f2f4 f7f6 g2g3 f6e5 d4e5 g7g5 h4g2 a8c8 h1c1 a5c6 a1b1 b7b6 h2h4 g5f4 g3f4 c8c7 e2e3 f8c8 h4h5');
INSERT INTO evaluations VALUES(53267,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(53267,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53267,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(53267,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53267,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53267,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53267,36,'cp 15','d4c5 g8f6 a2a3 a7a5 b2b4 b7b6 c5b6 e7e5 g1f3 a5b4 f3e5 d8b6 e2e3 b8d7 e5d3 b4a3 b1a3 b6c6 a3c2 a8a1 c2a1 c6c3 c1d2 c3c7 a1c2 f8e7 f1e2 e8g8 c2d4 f6e4 e1g1 e7d6 h2h3 c8a6 d2b4 a6d3 b4d6 d3e2 d6c7 e2d1 f1d1 f8a8');
INSERT INTO evaluations VALUES(53168,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53168,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(53168,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(53168,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(53168,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(53168,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(53168,36,'cp 13','g8f6 b2b4 a7a5 a2a3 b7b6 c5b6 e7e5 g1f3 a5b4 f3e5 d8b6 e2e3 b8d7 e5d3 b4a3 c1a3 c8b7 d1c2 f8a3 b1a3 b7a6 f1e2 e8g8 e1g1 f8c8 d3f4 a6e2 f4e2 b6c5 c2c1 f6e4 a3c2 e4c3 e2c3 c5c3 a1a8 c8a8');
INSERT INTO evaluations VALUES(3226,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(3226,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(3226,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(3226,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(3226,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(3226,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(3226,36,'cp 37','c8d7 e1c1 a8c8 c4e2 a7a6 h2h4 b7b5 c3d5 f8e8 d4c6 d7c6 d5f6 e7f6 h4h5 d6d5 e4d5 d8d5 d2d5 c6d5 d1d5 e8e3 e2d3 f6f5 c2c3 g7f8 c1c2 g8g7 h1a1 c8c6 c2b3 g7h6 a2a4 b5a4 a1a4 h6h5 b3c2 c6d6 d5d6 f8d6 a4a6 d6f4 b2b4 h5h4 a6c6');
INSERT INTO evaluations VALUES(23661,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(23661,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(23661,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(23661,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(23661,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(23661,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(23661,36,'cp 58','e1c1 a8c8 d4c6 d7c6 c4b3 b7b5 e3a7 b5b4 c3d5 f6d5 e4d5 d8a5 d5c6 a5a7 h1e1 c8c6 e1e4 c6a6 e4b4 e7e6 c1b1 f8a8 a2a3 d6d5 b4b5 h7h5 g2g3 g7f6 f3f4 g8g7 c2c3 a6b6 b5b6 a7b6 b1a2 a8b8 d2c2 h5h4 a3a4 b6e3 g3h4 b8h8 h4h5 h8h5 c2d2 e3c5 d2e2');
INSERT INTO evaluations VALUES(23660,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(23660,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(23660,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(23660,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(23660,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(23660,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(23660,36,'cp 24','a8c8 d4c6 d7c6 c4b3 a7a5 a2a4 d8e8 d2e2 f6d7 h2h4 d7c5 e3c5 d6c5 e4e5 e7e6 f3f4 e8e7 b3c4 f8d8 g2g3 c6h1 d1h1 h7h5 c3b5 g7h6 c1b1 e7d7 b5d6 d7a4 d6c8 d8c8 g3g4 b7b5 c4b5 a4b4 g4h5 c8b8 h5g6 b4b5 g6f7 g8f8');
INSERT INTO evaluations VALUES(47,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(47,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(47,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(47,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(47,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(47,36,'cp 32','d4c6 d7c6 c4b3 a7a5 a2a4 d8e8 d2e2 f6d7 h2h4 d7c5 e3c5 d6c5 e4e5 e7e6 b3c4 e8e7 g2g3 f8d8 c3b5 d8d1 h1d1 c8d8 d1f1 h7h5 f3f4 g7f8 c1b1 e7d7 g3g4 d7d2 g4h5 d2e2 c4e2 d8d2 e2d3 g6h5 f4f5 d2h2 f5e6 f7e6 d3c4 c6d7 f1f6 h2h4 c4e6 d7e6 f6e6 g8g7');
INSERT INTO evaluations VALUES(16385,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(16385,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(16385,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(16385,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(16385,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(16385,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(16385,36,'cp -86','c6d4 f3d1 f8c5 g1f3 c7c6 e1g1 d7d6 c4e2 d4e2 d1e2 c5b6 d2d3 e8g8 c1e3 b6a5 e2d1 c8g4 c3e2 g4f3 g2f3 f6h5 f3f4 a5b6 f4e5 d6e5 e2g3 h5g3 e3b6 g3e2 d1e2 d8g5 g1h1 a7b6 e2e3 g5e3 f2e3 b6b5 a2a3');
INSERT INTO evaluations VALUES(16990,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(16990,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(16990,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(16990,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(16990,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(16990,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(16990,36,'cp -82','f3d1 c7c6 g1f3 f8c5 e1g1 d7d6 c4e2 d4e2 d1e2 e8g8 d2d3 h7h6 h2h3 c5b6 f1e1 b6c7 d3d4 e5d4 f3d4 d6d5 e4d5 f6d5 c3d5 d8d5 d4f3 c8e6 e2d1 d5c5 c2c3 e6d5 c1e3 c5c4 d1b3 c4a6 b3c2 d5f3 g2f3');
INSERT INTO evaluations VALUES(16991,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(16991,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(16991,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(16991,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(16991,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(16991,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(16991,36,'cp -98','c7c6 g1f3 f8c5 e1g1 d7d6 c4e2 d4e2 d1e2 e8g8 d2d3 c5b6 c1e3 b6a5 h2h3 f8e8 e2d1 d6d5 e3g5 h7h6 g5f6 d8f6 e4d5 a5c3 b2c3 c6d5 f1e1 c8d7 f3d2 d7c6 c3c4 d5c4 d2c4 e5e4 c4a5 e4d3 a5c6 e8e1 d1e1 f6c6 c2d3');
INSERT INTO evaluations VALUES(36645,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(36645,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(36645,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(36645,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36645,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(36645,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(36645,36,'cp 34','c2c4 c8f5 c1e3 d6d5 c4d5 f6d5 e3g5 h7h6 g5d2 c7c5 b1c3 d5c3 b2c3 f5e4 d1c1 h6h5 d2h6 c5d4 f1d1 e4f3 h6g7 g8g7 g2f3 b8c6 f3c6 b7c6 c3d4 d8d5 c1c3 g7g8 a2a4 d5e4 e2e3 h5h4 d1c1 f8c8');
INSERT INTO evaluations VALUES(47571,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(47571,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(47571,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(47571,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(47571,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(47571,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(47571,36,'cp -64','b8c6 f3d1 f8g7 b1c3 c6a5 g1f3 e8g8 h2h3 a5c4 d3c4 c7c6 a2a4 c8e6 d1e2 d8a5 f3g5 e6c8 e1g1 f6e8 g5f3 h7h6 f1d1 c8e6 c1d2 a5c7 b2b3 f7f5 d2c1 f5f4 c1a3 b7b6 a4a5 a8d8 d1d3 g6g5 a1d1 g5g4 h3g4 e6g4 a5b6');
INSERT INTO evaluations VALUES(14529,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(14529,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(14529,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(14529,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(14529,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(14529,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(14529,36,'cp -32','a2a4 f8g7 h2h3 e8g8 g1e2 c6b4 b1a3 d6d5 e4d5 b4d5 e1g1 c7c6 f1e1 c8e6 c1d2 f6h5 c4b3 d8c7 e2c3 d5c3 b2c3 f8e8 b3e6 e8e6 a3c4 f7f5 a4a5 h5f6 f3e3 f6d5 e3c5 e5e4 d3e4 e6e4 e1e4 f5e4');
INSERT INTO evaluations VALUES(17157,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(17157,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(17157,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(17157,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(17157,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(17157,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(17157,36,'cp 66','c2c3 f8g7 d2d4 a7a6 b5c6 e7c6 d4d5 c6a7 d5d6 a7b5 d6c7 b5c7 d1d6 c7b5 d6b4 d8e7 b4e7 e8e7 c3c4 b5c7 a2a4 d7d6 a4a5 c8d7 b1c3 d7c6 b2b4 f7f5 f1d1 c7e6 c1e3 h8c8 e4f5 e5e4 f3d4 g7d4 e3d4 g6f5 d4e3');
INSERT INTO evaluations VALUES(17153,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(17153,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(17153,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(17153,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(17153,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(17153,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17153,36,'cp 79','f8g7 d2d4 a7a6 b5c4 e5d4 c3d4 b7b5 c4b3 c6a5 c1d2 a5b3 d1b3 d7d5 e4e5 c7c5 d4c5 e8g8 d2e3 c8e6 f3d4 g7e5 b1d2 e7f5 d2f3 e5g7 h2h3 f5e3 b3e3 d8f6 b2b4 a8e8 a1e1 e6d7 e3d3 d7c6 d4c6 f6c6 e1e8 f8e8 f1e1 e8e6 e1e6 c6e6');
INSERT INTO evaluations VALUES(10997,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(10997,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(10997,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(10997,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(10997,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(10997,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(10997,36,'cp 66','d2d4 a7a6 b5c4 b7b5 c4b3 e5d4 c3d4 c6a5 c1d2 a5b3 d1b3 d7d6 d2c3 d6d5 e4e5 e8g8 c3b4 f8e8 f1c1 c8d7 b1d2 c7c6 b4c5 a6a5 c1e1 e7f5 h2h3 g7h6 a2a3 h6f4 b3c3 h7h5 d2f1 f4h6 f1e3 f5g7 b2b4 h6f4 c5d6 g7e6');
INSERT INTO evaluations VALUES(10993,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(10993,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(10993,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(10993,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(10993,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(10993,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(10993,36,'cp 72','a7a6 b5c4 b7b5 c4b3 e5d4 c3d4 c6a5 c1d2 a5b3 d1b3 d7d6 d2c3 d6d5 e4e5 e8g8 a2a4 c8d7 a4b5 d7b5 f1e1 d8d7 b1d2 e7c6 b3c2 f8e8 h2h3 g7f8 d2b3 d7f5 c2f5 g6f5 g2g3 a8b8 b3a5 c6d8 f3h4 b5d7 h4g2 c7c5 d4c5 f8c5 e1d1 d8e6 d1d5 d7b5 g2f4 c5f8');
INSERT INTO evaluations VALUES(21392,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21392,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(21392,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21392,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21392,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(21392,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21392,36,'cp 33','g8f6 e2e3 f8b4 f1d3 e8g8 e1g1 b4d6 f1e1 b7b6 f4g5 h7h6 g5h4 c6b4 d3f1 g7g5 h4g3 d6g3 h2g3 c7c5 a2a3 b4c6 d4c5 b6c5 e3e4 f6e4 c3e4 d5e4 e1e4 c8b7 d1d8 f8d8 e4c4 d8d5 b2b4 b7a6 c4c5 d5c5 b4c5 a6f1 g1f1');
INSERT INTO evaluations VALUES(36562,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36562,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(36562,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(36562,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(36562,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(36562,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(36562,36,'cp 43','e7e5 d4d5 b8a6 h2h3 f6h5 a2a3 f7f5 a1c1 h5f6 b2b4 c7c5 c1b1 f6h5 d1c2 b7b6 f1d3 d8e7 e1g1 f5f4 e3d2 h7h6 c2d1 h5f6 f3h4 g8f7 d3c2 f6h7 h4f3 f7g8 f3h2 h6h5 g1h1 g7f6 d1e2 g8g7 c2a4 a8b8 a4c6 f6g5');
INSERT INTO evaluations VALUES(36862,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(36862,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(36862,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(36862,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(36862,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(36862,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(36862,36,'cp 64','d4d5 f6g4 e3g5 f7f6 g5h4 b8a6 h2h3 g4h6 f3d2 h6f7 f1e2 c7c6 a2a3 g7h6 e1g1 c6c5 d1c2 b7b6 a1b1 h6f4 b2b4 a6c7 h4g3 c7e8 b4c5 d6c5 a3a4 f4g3 f2g3 e8d6 a4a5 c8d7 b1a1 b6a5 d2b3 d6b7 g3g4 f7d6 g1h1 d8e7');
INSERT INTO evaluations VALUES(36836,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(36836,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(36836,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(36836,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(36836,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(36836,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(36836,36,'cp 54','f6g4 e3g5 f7f6 g5h4 b8a6 h2h3 g4h6 f3d2 h6f7 f1e2 g7h6 a2a3 c7c6 d1c2 c8d7 e1g1 c6c5 h4g3 d8e7 f2f4 e5f4 g3f4 f6f5 f4h6 f7h6 e4f5 h6f5 d2e4 f5d4 c2d3 a8e8 a1e1 f8f1 e1f1 e8f8 f1f8 g8f8 e2g4 d7g4 h3g4 e7e5');
INSERT INTO evaluations VALUES(10209,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(10209,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(10209,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(10209,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(10209,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(10209,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(10209,36,'cp 47','f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 c3b5 a7a6 f2f4 e5b2 c1b2 a6b5 e4e5 f6e4 e1g1 d7d5 e5d6 d8d6 e2b5 c8d7 d1d6 c7d6 b5d3 d7c6 f1e1 a8a4 c2c4 f7f5 d3c2 a4a8 b2g7 h8g8 g7d4 e8f8 e1e2 g8g4 d4e3 e4c3 e2d2 c3e4 c2e4 c6e4 g2g3 g4g6 c4c5 d6c5 e3c5 f8g8');
INSERT INTO evaluations VALUES(10046,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(10046,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(10046,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(10046,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(10046,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(10046,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(10046,36,'cp 59','c6e5 d2d4 c5d6 d4e5 d6e5 c3b5 a7a6 f2f4 e5b2 c1b2 a6b5 e4e5 f6e4 e1g1 d7d5 e5d6 e8g8 d1d4 e4f6 e2b5 d8d6 d4d6 c7d6 c2c4 f6e4 a2a4 c8f5 f1d1 f5e6 h2h3 f8d8 d1d4 e4c5 a1a3 b7b6 d4d1 f7f6 b2d4 g8f7 d4e3 g7g6 g1h2 d6d5 c4d5 d8d5 d1d5 e6d5');
INSERT INTO evaluations VALUES(10766,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(10766,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(10766,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(10766,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(10766,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(10766,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(10766,36,'cp 53','d2d4 c5d6 d4e5 d6e5 c3b5 a7a6 f2f4 e5b2 c1b2 a6b5 e4e5 f6e4 e1g1 d7d5 e5d6 e8g8 d1d4 e4f6 e2b5 d8d6 d4d6 c7d6 c2c4 c8f5 a2a4 f8c8 f1d1 f5e6 b2f6 g7f6 a1c1 c8d8 g1f2 g8f8 d1d4 a8c8 c1d1 c8c5 f2e3 f8e7 g2g3 f6f5 e3f3 c5c8 a4a5 c8c5 h2h3');
INSERT INTO evaluations VALUES(45733,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(45733,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45733,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(45733,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45733,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(45733,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45733,36,'cp 7','e5e4 f3d4 c6d5 c3d5 f8c5 d4b5 e8g8 d5f6 d8f6 d2d4 e4d3 e1g1 d3c2 d1c2 b8a6 c2b3 c8e6 c4e6 f6e6 b3e6 f7e6 c1e3 c5e3 f2e3 a6b4 a1d1 b4d5 f1f8 a8f8 b5a7 d5e3 d1c1 f8d8 c1c8 d8f8 c8f8 g8f8 a7b5 e3d5 g1f2 f8e7 a2a3');
INSERT INTO evaluations VALUES(25104,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(25104,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(25104,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(25104,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(25104,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(25104,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(25104,36,'cp 87','g5e6 f7e6 d2d3 d8d7 c1e3 e8c8 b1d2 a7a6 e1c1 f8b4 c2c3 b4e7 f3g4 h7h5 g4e4 h8f8 d2f3 f8f5 c4b3 d5f6 e4a4 f6d5 c1b1 d7e8 a4e4 c8b8 h2h3 d5f6 e4a4 f6d5 b3c2 f5f8 h1e1 d5f4 g2g3 f4h3 a4e4');
INSERT INTO evaluations VALUES(25098,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(25098,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(25098,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(25098,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(25098,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(25098,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(25098,36,'cp 76','f7e6 d2d3 d8d7 c1e3 f8d6 b1d2 h8f8 f3h3 h7h6 h3h5 d7f7 h5g4 a7a6 e1c1 d6e7 d1e1 h6h5 g4e4 e8c8 d2f3 f7f5 e3d2 f5g4 h1g1 e7c5 d2e3 d5e3 f2e3 g4f5 c2c3 f5e4 c4e6 c8b8 d3e4 d8d3 c1c2 d3e3 e6d5 e3e1 g1e1');
INSERT INTO evaluations VALUES(25419,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(25419,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(25419,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(25419,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(25419,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(25419,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(25419,36,'cp 87','d2d3 d8d7 c1e3 f8d6 b1d2 h8f8 f3h3 h7h6 h3h5 d7f7 h5g4 a7a6 e1c1 h6h5 g4e4 e8c8 d2f3 d5e3 f2e3 f7f5 f3d2 c8d7 a2a4 f5e4 d2e4 c6e7 d1e1 e7d5 c2c3 h5h4 c1c2 f8f5 e1e2 d5f6 e4d2 e5e4 h1d1 d7e7 d3e4 f5g5 d2f3');
INSERT INTO evaluations VALUES(34380,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(34380,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(34380,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(34380,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(34380,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(34380,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(34380,36,'cp 37','d5d6 d2d4 g8f6 g1f3 e7e6 f1c4 f8e7 e1g1 e8g8 d1e2 a7a6 c4b3 b8c6 f1d1 h7h6 a2a3 c8d7 d4d5 e6d5 c3d5 f6d5 b3d5 d6c7 c1e3 e7f6 e2c2 d7e6 d5e6 f7e6 e3d4 a8d8 d4f6 f8f6 d1d8 c7d8 a1d1 d8b6 c2e2 b6b5 e2b5 a6b5 d1d7');
INSERT INTO evaluations VALUES(48461,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(48461,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(48461,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(48461,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(48461,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(48461,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(48461,36,'cp 76','d2d4 f8e7 e1g1 a7a6 a2a4 b8c6 f1e1 e8g8 c1f4 c6b4 d1e2 b7b6 a1d1 c8b7 c4b3 b4d5 c3d5 f6d5 f4g3 e7d6 f3e5 d8g5 h2h4 g5e7 c2c3 a6a5 b3c2 f7f6 e5c4 d6g3 f2g3 a8e8 e2e4 f6f5 e4f3 g7g5 h4g5 e7g5 c4e5');
INSERT INTO evaluations VALUES(48467,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(48467,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(48467,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(48467,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(48467,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(48467,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(48467,36,'cp 67','a7a6 a2a4 b8c6 e1g1 f8e7 c1f4 e8g8 f1e1 h7h6 f4g3 e7d6 g3h4 d6e7 d1d2 b7b6 a1d1 f6d5 c3d5 e6d5 h4e7 c6e7 c4d3 c8f5 e1e3 f5d3 d2d3 e7g6 d1e1 d8d7 h2h4 g6f4 d3b3 a8e8 f3e5 d7d6 g2g3 f4g6 e5f3 e8e3 e1e3 g6e7 h4h5');
INSERT INTO evaluations VALUES(23460,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(23460,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23460,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(23460,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(23460,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23460,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(23460,36,'cp 25','d2d4 d7d5 g1f3 c8f5 a2a3 e7e6 c2c4 g8f6 e2e3 c6b8 c4c5 b8d7 b1d2 h7h6 b3b4 g7g5 f1e2 g5g4 f3e5 d7e5 d4e5 f6e4 d2e4 f5e4 e1g1 h6h5 e2d3 c7c6 d3e4 d5e4 d1a4 f8g7 c1b2 d8c7 b4b5 c6b5 a4b5 c7c6 b5c6 b7c6');
INSERT INTO evaluations VALUES(10401,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(10401,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10401,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(10401,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(10401,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10401,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(10401,36,'cp 14','d2d3 b4c5 f1e2 d7d6 e1g1 e8g8 a2a3 a7a5 c1e3 f6g4 d1d2 c8e6 g1h1 g4e3 f2e3 c5b6 c3d5 e6d5 e4d5 c6e7 d3d4 e5e4 f3g5 e7d5 g5f7 d8e7 f7h6 g7h6 c4d5 f8f1 a1f1 a8f8 d2c2 f8f1 e2f1 g8g7 g2g3 h6h5 f1g2 h5h4 g2e4');
INSERT INTO evaluations VALUES(10403,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10403,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10403,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(10403,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(10403,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(10403,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10403,36,'cp 15','b4c5 f3e5 c6e5 d3d4 c5b4 d4e5 f6e4 d1d4 e4c3 b2c3 b4e7 f1e2 e8g8 e1g1 d7d6 c1f4 d6e5 f4e5 e7f6 d4d8 f6d8 e2f3 c7c6 f1e1 f8e8 e5d4 e8e1 a1e1 c8d7 c4c5 d8f6 e1b1 a8b8 a2a4 d7f5 b1b2 f6e7 h2h3 f5e6 g2g4 e7f6');
INSERT INTO evaluations VALUES(23195,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(23195,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(23195,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(23195,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(23195,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(23195,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(23195,36,'cp 69','f8c5 c3d5 g8e7 e1g1 a7a6 b5c4 d7d6 c2c3 e7d5 c4d5 c6e7 d5c4 c5a7 d2d4 b7b5 c4d3 e8g8 a2a4 b5a4 a1a4 a7b6 c1e3 g8h8 d1e2 a6a5 b2b4 c8d7 b4b5 c7c6 c3c4 d7g4 c4c5 g4f3 g2f3 e5d4 c5b6 d4e3 e2e3 c6b5 d3b5');
INSERT INTO evaluations VALUES(6915,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(6915,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(6915,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(6915,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(6915,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(6915,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(6915,36,'cp 72','b5c4 g8e7 a2a4 d7d6 h2h3 a7a6 d2d3 c8d7 e1g1 d8c8 c1e3 d7e6 c3d5 c5e3 f2e3 e6d5 e4d5 c6d8 e3e4 e8g8 d1d2 d8f7 c4b3 e7g6 f1f2 c8d7 a1f1 g6f4 g1h2 a6a5 c2c3 b7b5 a4b5 d7b5 b3c4 b5d7 b2b4 a5b4 c3b4');
INSERT INTO evaluations VALUES(27494,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(27494,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(27494,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(27494,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(27494,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(27494,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(27494,36,'cp -50','d2d3 g8e7 h2h3 g4d7 b1c3 e7g6 h1e1 f7f6 e2f1 e8f7 a2a4 a6a5 b2b3 b7b6 c1a3 f8b4 a3b4 a5b4 c3e4 g6f4 a4a5 a8a5 a1a5 b6a5 e4c5 d7e6 f3d4 e6d5 g2g3 f4h3 d4f5 h8e8 e1e8 f7e8 f5g7 e8f7 g7f5 h7h5 f2f4');
INSERT INTO evaluations VALUES(47199,30,'cp -116',NULL);
INSERT INTO evaluations VALUES(47199,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(47199,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(47199,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(47199,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(47199,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(47199,36,'cp -106','c7c5 b1c3 b8c6 a2a3 e7e6 f3f4 f8d6 g1f3 e8g8 f1d3 c5c4 d3e2 a8b8 f3e5 c6e7 e1g1 b7b5 b2b3 a7a6 d1e1 f6e8 c1d2 d6c7 c3d1 f7f6 e5f3 e8d6 d2b4 d6b7 b3c4 b5c4 a3a4 a6a5 b4a3 g8h8 d1c3 b7d6');
INSERT INTO evaluations VALUES(21777,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21777,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21777,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(21777,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(21777,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(21777,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(21777,36,'cp 14','f8b4 c1d2 d4e3 f2e3 b4d2 d1d2 d8d2 b1d2 g8e7 d2b3 c8g4 a1d1 e7g6 f1e2 e8g8 h2h3 g4c8 b3d4 c6e5 e1f2 c7c6 b2b4 e5d7 d4f5 d7f6 f5d6 c8e6 d1d4 b7b6 h1d1 c6c5 b4c5 b6c5 d4d2 h7h6 e2d3 a8b8 d3g6 f7g6 f2g3 b8d8 f3e5 d8d6 d2d6 f6e4 g3h2 e4d6');
INSERT INTO evaluations VALUES(5887,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5887,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(5887,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(5887,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(5887,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5887,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(5887,36,'cp 5','c1d2 d4e3 f2e3 b4d2 d1d2 d8d2 b1d2 g8e7 d2b3 e7g6 f3d4 c8d7 e5e6 d7e6 d4c6 b7c6 b3c5 g6e5 b2b3 e5d7 c5a4 e8c8 f1e2 d7e5 a1d1 d8d6 h1f1 h8d8 a4c5 d6d1 e2d1 e5d3 c5d3 d8d3 e1e2 d3d6 e2e1');
INSERT INTO evaluations VALUES(5886,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5886,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5886,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5886,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5886,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5886,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5886,36,'cp 0','d4e3 f2e3 b4d2 d1d2 d8d2 b1d2 g8e7 d2b3 e7g6 f3d4 c8d7 e5e6 d7e6 d4c6 b7c6 e1c1 a7a5 b3c5 a5a4 f1e2 e8e7 d1d4 g6e5 h2h4 a8a5 c5b7 a5a7 b7c5');
INSERT INTO evaluations VALUES(5885,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5885,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5885,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5885,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5885,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5885,36,'cp 0','f2e3 b4d2 d1d2 d8d2 b1d2 g8e7 d2b3 e7g6 f3d4 c8d7 e5e6 d7e6 d4c6 b7c6 e1c1 a7a5 b3c5 a5a4 f1e2 e8e7 d1d4 g6e5 h2h4 a8a5 c5b7 a5a7 b7c5');
INSERT INTO evaluations VALUES(5884,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5884,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5884,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5884,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5884,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5884,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5884,36,'cp 0','b4d2 d1d2 d8d2 b1d2 g8e7 d2b3 e7g6 f3d4 c8d7 e5e6 d7e6 d4c6 b7c6 e1c1 a7a5 b3c5 a5a4 f1e2 e8e7 d1d4 g6e5 h2h4 a8a5 c5b7 a5a7 b7c5');
INSERT INTO evaluations VALUES(52533,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(52533,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(52533,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(52533,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(52533,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(52533,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(52533,36,'cp 73','f1d3 b8c6 c3c4 e7e5 a2a3 f8c5 b2b4 c5d4 a1a2 d7d6 g1e2 d4b6 b1c3 g8f6 c3a4 c8g4 a4b6 g4e2 a2e2 d8b6 c1e3 c6d4 e2d2 a7a5 b4b5 b6c5 d1a1 e8g8 e1g1 a8c8 h2h3 f8e8 f1e1 h7h6 f2f3 f6h5 d3f1 h5f4 e3d4 e5d4 a1d4 c5a3 d4d6');
INSERT INTO evaluations VALUES(52531,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(52531,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(52531,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(52531,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(52531,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(52531,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(52531,36,'cp 92','b8c6 c3c4 e7e5 a2a3 f8c5 b2b4 c5b6 b1c3 d7d6 c3d5 g8e7 d5b6 a7b6 c1b2 e7g6 g1e2 d8g5 e1g1 c8h3 e2g3 h3g4 f2f3 g4e6 g1h1 e8g8 g3f5 g6f4 f5e3 a8d8 g2g3 f4d3 d1d3 g5g6 e3f5 g6f6 a1d1 e6f5 e4f5 f8e8 d1d2 c6e7');
INSERT INTO evaluations VALUES(19561,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(19561,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(19561,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(19561,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(19561,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(19561,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(19561,36,'cp 65','c3c4 e7e6 g1f3 f8c5 a2a3 a7a5 d1e2 g8f6 b1c3 d7d6 c1e3 e6e5 d3c2 c8e6 a1d1 c5e3 f2e3 d8b6 c3a4 b6c7 b2b3 e8g8 e2d2 a8d8 e1g1 c6b8 a4c3 c7c5 h2h3 b8a6 g1h2 c5a3 f3e5 a6c5 e5f3 c5b3 c2b3 a3b3 e4e5 d6e5 d2d8 b3c3');
INSERT INTO evaluations VALUES(19528,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(19528,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(19528,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(19528,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(19528,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(19528,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(19528,36,'cp 49','e7e6 g1f3 f8c5 a2a3 a7a5 b1c3 d7d6 d1e2 e6e5 d3c2 c8e6 c3d5 a5a4 e1g1 e6g4 e2d3 g8e7 d5c3 e8g8 c3a4 f7f5 a4c5 d6c5 e4f5 d8d3 c2d3 g4f3 g2f3 e7f5 d3e4 f5h4 c1e3 h4f3 e4f3 f8f3 e3c5 f3f7 c5d6 c6d4 d6e5');
INSERT INTO evaluations VALUES(55897,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(55897,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(55897,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55897,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(55897,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(55897,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(55897,36,'cp -42','d5e4 d3e4 g8f6 e4d3 b7b6 e1g1 c8b7 b2b3 c7c5 b1a3 f8e7 c1b2 b8c6 f1e1 e8g8 a3c4 d8c7 c4e5 c6b4 d3c4 a8d8 f3g5 b7d5 c4d5 f6d5 g5f3 h7h6 a2a3 b4c6 d1e2 e7f6 e5c6 c7c6 b2f6 d5f6 a1d1 f6d5');
INSERT INTO evaluations VALUES(56082,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(56082,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(56082,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(56082,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(56082,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(56082,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(56082,36,'cp -39','d3e4 g8f6 e4d3 b7b6 e1g1 c8b7 b2b3 c7c5 b1a3 b8c6 c1b2 f8e7 a3c4 e8g8 f1e1 d8c7 c4e5 c6b4 d3c4 b4d5 a2a4 a7a6 c4f1 d5b4 c2c3 b4c6 e5c6 b7c6 d1e2 a8d8 f3e5 c6b7 d2d4 f6d7 e5c4 f8e8 h2h3 e7f8 a1d1 b6b5 a4b5');
INSERT INTO evaluations VALUES(56066,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56066,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(56066,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(56066,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(56066,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(56066,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(56066,36,'cp -31','g8f6 e4d3 b7b6 e1g1 c8b7 f1e1 c7c5 a2a4 a7a6 d3f1 b8c6 d2d3 f8e7 c1f4 e8g8 b1a3 f6h5 f4d2 d8c7 g2g3 h5f6 f1g2 f6d5 c2c3 e7f6 a3c4 h7h6 d1e2 a8b8 f3e5 c6e5 c4e5 d5e7 g2b7 c7b7 e5c4');
INSERT INTO evaluations VALUES(48533,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(48533,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(48533,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(48533,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(48533,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(48533,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(48533,36,'cp -39','e4d3 b7b6 e1g1 c8b7 f1e1 c7c5 a2a4 a7a6 d3f1 b8c6 b1a3 f8e7 c2c3 d8c7 d2d3 e8g8 c1g5 f6d5 g5e7 d5e7 a3c4 a8d8 d1e2 e7f5 g2g3 h7h6 f3d2 c6e7 e2e5 f5d6 c4d6 e7g6 e5e3 c7d6 d2c4 d6c7 f1g2');
INSERT INTO evaluations VALUES(44709,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(44709,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(44709,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(44709,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(44709,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(44709,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44709,36,'cp 57','f8g7 f1e2 f6e4 e1g1 e8g8 d1b3 b7b6 f1d1 e4c3 b3c3 c8a6 b2b3 c6c5 c1b2 b8d7 a1c1 d5c4 b3c4 c5d4 e3d4 e7e6 b2a3 f8e8 a3d6 d7f6 d6e5 d8e7 c3b3 e8c8 a2a4 a6b7 a4a5 b6a5 c4c5 b7d5 b3a3 f6d7 e5g7 g8g7 a3a5');
INSERT INTO evaluations VALUES(56280,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(56280,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(56280,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(56280,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(56280,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(56280,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(56280,36,'cp -19','g8f6 g2g3 f8b4 c2c3 b4d6 f1g2 d8e7 d1e2 e7e2 e1e2 e8g8 h1e1 c7c6 b1a3 b8a6 c1e3 f6g4 e3d4 d6a3 b2a3 c6c5 d4e3 f8e8 e2d2 g4e3 e1e3 e8e3 f2e3 a6c7 a3a4 g8f8 a1b1 f8e7 f3g5 h7h6 g5h3 e7d6 h3f4 g7g6 f4e2');
INSERT INTO evaluations VALUES(48742,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(48742,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(48742,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(48742,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48742,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48742,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(48742,36,'cp -32','g2g3 f8b4 c2c3 b4d6 f1g2 d8e7 d1e2 e7e2 e1e2 e8g8 h1e1 h7h6 b1a3 d6a3 b2a3 c7c5 c1f4 f8d8 f4c7 d8e8 e2d2 c8e6 c7b8 a8b8 f3e5 c5c4 a1b1 c4d3 e5d3 b7b6 a3a4 a7a6 d3f4 g8f8 h2h4 g7g6 f4e6 f7e6');
INSERT INTO evaluations VALUES(23136,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(23136,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(23136,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23136,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(23136,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(23136,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(23136,36,'cp 50','d8f6 f3f6 g8f6 d2d3 f8c5 b1d2 d7d6 d2b3 c5b6 a2a4 a7a5 c2c3 d4c3 b2c3 c7c6 c1g5 c8d7 g5f6 g7f6 a1b1 e8c8 b3d4 b6d4 c3d4 c6c5 c4b5 c5d4 b5d7 d8d7 e1e2 d6d5 h1c1 c8b8 b1b5 d5e4 d3e4 d4d3 e2d2');
INSERT INTO evaluations VALUES(1965,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(1965,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(1965,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(1965,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(1965,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(1965,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(1965,36,'cp 52','f3f6 g8f6 d2d3 f8b4 b1d2 d7d6 e1g1 b4d2 c1d2 c8e6 f2f3 a7a6 c2c3 d4c3 d2c3 e8g8 c4e6 f7e6 a2a4 a6a5 f1e1 b7b6 d3d4 a8e8 c3d2 d6d5 e4d5 f6d5 a1c1 e8e7 c1c6 f8e8 g1f2 g8f7 e1c1');
INSERT INTO evaluations VALUES(1932,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(1932,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(1932,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(1932,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(1932,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(1932,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(1932,36,'cp 40','g8f6 d2d3 f8b4 b1d2 d7d6 e1e2 b4d2 c1d2 c8e6 c4e6 f7e6 b2b4 e6e5 a2a4 e8d7 c2c3 d4c3 d2c3 h8c8 c3d2 b7b6 g2g3 c7c5 h1c1 c5b4 d2b4 g7g5 f2f3 h7h6 b4d2 c8c5 f3f4 g5f4 g3f4 e5f4 d2f4 a8c8 c1c5 b6c5 h2h3 c5c4 f4h6 c4d3 e2d3');
INSERT INTO evaluations VALUES(1058,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(1058,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(1058,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(1058,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(1058,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(1058,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(1058,36,'cp 50','d2d3 f8b4 b1d2 d7d6 e1e2 b4d2 c1d2 c8e6 c4e6 f7e6 c2c3 d4c3 b2c3 b7b6 f2f3 e6e5 d3d4 e8c8 a2a4 h8e8 d4d5 a7a5 d2g5 e8f8 h2h4 h7h6 g5f6 f8f6 h4h5 c7c6 c3c4 c8b7 h1h4 f6f7 a1c1 b7c7 h4g4 d8f8');
INSERT INTO evaluations VALUES(1057,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(1057,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(1057,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(1057,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(1057,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(1057,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(1057,36,'cp 40','f8b4 b1d2 d7d6 e1g1 b4d2 c1d2 c8e6 f2f3 e8g8 f1d1 d6d5 e4d5 e6d5 d2b4 f8d8 a2a4 a7a5 b4e7 d8d7 e7f6 d5c4 d3c4 g7f6 d1d3 c7c5 a1e1 b7b6 d3b3 a8b8 b3b5 g8g7 g1f2 d4d3 c2d3 d7d3 e1e2 d3d4 b2b3 f6f5');
INSERT INTO evaluations VALUES(35070,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(35070,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(35070,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(35070,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(35070,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(35070,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(35070,36,'cp 137','g1f3 f8b4 b1d2 d7d6 e2e3 b8d7 f1d3 f6e7 e1g1 b4d2 c1d2 e6e5 e3e4 g8f6 f1e1 e8g8 d3c2 a7a5 d2c3 b7b6 d1d2 c8a6 f3h4 g7g6 b2b3 f6h5 g2g3 h5g7 a1d1 f7f6 f2f4 a6b7 f4f5 e5d4 c3d4 d7e5 d4c3');
INSERT INTO evaluations VALUES(35068,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(35068,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(35068,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(35068,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(35068,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(35068,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(35068,36,'cp 136','f6d8 e2e4 f8b4 b1c3 b4c3 b2c3 g8e7 f1d3 d7d6 e1g1 b8d7 f3d2 e8g8 f2f4 b7b6 d2f3 c8b7 d1e1 f8e8 e1g3 f7f5 g3h3 h7h6 f3h4 d7f6 e4e5 f6e4 d4d5 e6d5 c4d5 d6e5 d3e4 f5e4 c3c4 b7a6 c1b2 e7d5 c4d5 d8d5 f4e5 a6f1 a1f1');
INSERT INTO evaluations VALUES(41265,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(41265,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(41265,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41265,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(41265,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(41265,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41265,36,'cp 0','c8f5 e2e3 f6h5 f4g5 h7h6 g5h4 g6g5 h4g3 c7c5 f1d3 h5g3 h2g3 c5d4 f3d4 f5d7 a1c1 b8c6 d3e2 c6d4 e3d4 a8c8 e1g1 h6h5 e2h5 c8c4 d1f3 d7e6 c3e4 d8c7 e4g5 c4c1 h5f7 e6f7 f3f7 e8d7 f7f5 d7e8 f5f7');
INSERT INTO evaluations VALUES(55831,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(55831,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(55831,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(55831,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(55831,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(55831,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(55831,36,'cp -37','g8f6 c2c3 f8e7 e2e3 e8g8 f1d3 b7b6 d1e2 a7a5 e1g1 c8a6 d3a6 b8a6 b1d2 c7c5 e3e4 d5e4 d2e4 f6e4 e2e4 a6c7 f4f5 e6f5 e4f5 d8d5 f5c2 c5d4 f3d4 e7c5 c1e3 a8e8 c2f2 f7f5 a1e1 a5a4 a2a3 e8e5 e3f4 e5e1 f1e1');
INSERT INTO evaluations VALUES(48308,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48308,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(48308,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(48308,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48308,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(48308,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(48308,36,'cp -25','e2e3 c7c5 c2c3 f8e7 f1d3 b7b6 e1g1 e8g8 d1e2 c8b7 b2b3 a7a6 b1d2 b6b5 c1b2 b8d7 f3e5 f6e4 d3e4 d5e4 c3c4 f7f6 e5d7 d8d7 f1d1 c5d4 b2d4 d7c6 a1c1 f8d8 c4c5 a6a5 h2h3 b5b4 d2c4 e7c5 d4c5 c6c5');
INSERT INTO evaluations VALUES(5184,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(5184,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(5184,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(5184,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(5184,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5184,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5184,36,'cp 0','f1d3 g8e7 e1g1 e7g6 b1d2 c5e7 f1e1 e8g8 d2b3 b7b6 c2c3 d8c7 b3d4 c8b7 d4b5 c7b8 b5d4');
INSERT INTO evaluations VALUES(12795,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12795,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12795,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12795,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12795,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12795,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12795,36,'cp 0','e7e5 f1e2 f8e7 e1g1 e8g8 d2d3 a7a6 a2a4 d7d6 f3e1 c8e6 c1d2 c6d4 e1c2 d4b3 a1b1 a8b8 c2e3 b3d4 e3c2');
INSERT INTO evaluations VALUES(31793,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(31793,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(31793,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(31793,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(31793,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(31793,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(31793,36,'cp 180','e2e4 c7c6 d4d5 c6d5 e4d5 e6f5 f1c4 g8f6 g1f3 e7e5 f3e5 f8d6 e1g1 e8g8 c1f4 a7a6 f1e1 d8c7 h2h3 b8d7 d1d4 b7b5 e5d7 f6d7 f4d6 c7d6 c4f1 a8d8 a1c1 d7f6 e1e5 f5g6 a2a3 h7h5 c1e1 f8e8 e5e8 f6e8 d4b4 d6b4 a3b4');
INSERT INTO evaluations VALUES(31800,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(31800,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(31800,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(31800,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(31800,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(31800,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(31800,36,'cp 176','c7c6 d4d5 c6d5 e4d5 e6f5 g1f3 g8f6 f1c4 e7e5 f3e5 f8d6 e1g1 e8g8 c1f4 a7a6 f1e1 d8c7 h2h3 b8d7 d1d4 b7b5 e5d7 f6d7 f4d6 c7d6 c4e2 f8d8 a1c1 a8c8 e2f3 c8c4 d4d2 h7h6 b2b3 c4c8 f3e4 f5g6 e4g6 d6g6 c3e4 c8c1 d2c1');
INSERT INTO evaluations VALUES(31023,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(31023,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(31023,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(31023,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(31023,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(31023,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(31023,36,'cp 170','d4d5 c6d5 e4d5 e6f5 g1f3 g8f6 f1c4 e7e5 e1g1 f8d6 f3e5 e8g8 c1f4 a7a6 f1e1 d8c7 g2g3 b8d7 e5d7 f5d7 f4d6 c7d6 d1d4 a8c8 a1d1 f8d8 c4d3 b7b5 c3e4 f6e4 d3e4 g7g6 b2b3 d7h3 a2a4 b5a4 b3a4 h3g4 d1d2 d8e8 a4a5 d6c5 d4c5');
INSERT INTO evaluations VALUES(31014,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(31014,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(31014,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(31014,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(31014,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(31014,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(31014,36,'cp 181','c6d5 e4d5 e6f5 g1f3 g8f6 f1c4 e7e5 e1g1 f8d6 f3e5 e8g8 c1f4 a7a6 f1e1 d8c7 c4f1 b8d7 e5d7 c7d7 d1d4 f8e8 f4d6 d7d6 a2a4 h7h6 e1e8 a8e8 h2h3 e8d8 a1d1 f5c2 d1d2 c2b3 g2g3 d8c8 g3g4 a6a5 f1e2 g8f8 g1g2 c8d8 e2f3 d6b4 d5d6');
INSERT INTO evaluations VALUES(31264,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(31264,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(31264,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(31264,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(31264,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(31264,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(31264,36,'cp 182','e4d5 e6f5 g1f3 g8f6 f1c4 e7e5 f3e5 f8d6 e1g1 e8g8 c1f4 a7a6 f1e1 d8c7 c4f1 b8d7 e5d7 c7d7 d1d4 b7b5 f4e5 d6e5 e1e5 f6g4 e5e1 d7d6 g2g3 a8e8 f1g2 e8c8 h2h3 g4f6 g3g4 f5d7 g4g5 f6h5 c3e4 c8c4 e4d6 c4d4');
INSERT INTO evaluations VALUES(31263,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(31263,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(31263,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(31263,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(31263,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(31263,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(31263,36,'cp 165','e6f5 g1f3 g8f6 f1c4 e7e5 f3e5 f8d6 c4b5 b8d7 e5c4 e8g8 c4d6 d8e7 c1e3 e7d6 d1d4 d7b6 d4c5 f6e8 e1g1 a8c8 c5d4 e8c7 a1d1 f8d8 b5c4 b6c4 d4c4 a7a6 c4d4 d6g6 d1d2 f5e6 d5d6 h7h6 f1d1 c7e8 d4e4 e6f5 e4b7 d8d6 b7f3 d6d2 d1d2');
INSERT INTO evaluations VALUES(46965,30,'cp -201',NULL);
INSERT INTO evaluations VALUES(46965,31,'cp -201',NULL);
INSERT INTO evaluations VALUES(46965,32,'cp -184',NULL);
INSERT INTO evaluations VALUES(46965,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(46965,34,'cp -190',NULL);
INSERT INTO evaluations VALUES(46965,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(46965,36,'cp -189','d6e5 d4e5 d8d1 e1d1 f6g4 d1e1 g4e5 b1c3 e5d7 g1f3 c7c6 c1e3 d7f6 e1d2 c8g4 f1e2 b8d7 a1e1 e7e6 d2c1 a7a5 f3d2 g4e2 e1e2 a5a4 a2a3 b7b5 h1e1 e8c8 f2f3 c8c7 e3f4 c7b7 c3e4 d7b6 e4f6 g7f6 d2e4');
INSERT INTO evaluations VALUES(47309,30,'cp -170',NULL);
INSERT INTO evaluations VALUES(47309,31,'cp -179',NULL);
INSERT INTO evaluations VALUES(47309,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(47309,33,'cp -183',NULL);
INSERT INTO evaluations VALUES(47309,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(47309,35,'cp -166',NULL);
INSERT INTO evaluations VALUES(47309,36,'cp -173','d4e5 d8d1 e1d1 f6g4 d1e1 g4e5 b1c3 e5d7 c3d5 e8d8 c1f4 e7e5 f4g3 b8c6 a1d1 f8d6 f2f4 e5f4 g3f4 c6e5 f4g3 h8e8 f1e2 f7f6 g1f3 e5f7 g3d6 f7d6 d5c3 d7c5 e1f2 c8d7 h1e1 a7a5 f3d4 c5e4 c3e4 e8e4 e2f3 e4e1 d1e1 a8a6 b2b3');
INSERT INTO evaluations VALUES(47307,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(47307,31,'cp -181',NULL);
INSERT INTO evaluations VALUES(47307,32,'cp -176',NULL);
INSERT INTO evaluations VALUES(47307,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(47307,34,'cp -174',NULL);
INSERT INTO evaluations VALUES(47307,35,'cp -184',NULL);
INSERT INTO evaluations VALUES(47307,36,'cp -175','d8d1 e1d1 f6g4 d1e1 g4e5 b1c3 b8c6 f2f4 e5g6 c3b5 e8d8 c1e3 a7a6 a1d1 c8d7 b5c3 d8e8 g1f3 d7f5 d1d2 a8d8 d2d8 e8d8 h2h3 f7f6 g2g4 f5d7 f4f5 g6e5 f3e5 c6e5 f1g2 d8c8 e1f2 h7h5 e3d4 h5g4 d4e5 g4h3 e5g3 d7f5 g2e4 e7e6 e4f5 e6f5');
INSERT INTO evaluations VALUES(33698,30,'cp -174',NULL);
INSERT INTO evaluations VALUES(33698,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(33698,32,'cp -169',NULL);
INSERT INTO evaluations VALUES(33698,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(33698,34,'cp -171',NULL);
INSERT INTO evaluations VALUES(33698,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(33698,36,'cp -162','e1d1 f6g4 d1e1 g4e5 b1c3 e5d7 g1f3 c7c6 c1e3 d7f6 f1d3 e7e6 a2a4 a7a5 e1e2 b8d7 h1d1 f8c5 e3f4 e8g8 e2f1 c5e7 f4g3 f8e8 f3d2 f6d5 d2c4 b7b6 c4d6 e8d8 d6c8 a8c8 d3a6 c8a8 a6b7 d5c3 b7a8 c3d1');
INSERT INTO evaluations VALUES(33697,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(33697,31,'cp -183',NULL);
INSERT INTO evaluations VALUES(33697,32,'cp -176',NULL);
INSERT INTO evaluations VALUES(33697,33,'cp -177',NULL);
INSERT INTO evaluations VALUES(33697,34,'cp -183',NULL);
INSERT INTO evaluations VALUES(33697,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(33697,36,'cp -172','f6g4 d1e1 g4e5 b1c3 e5d7 g1f3 c7c6 c1e3 d7f6 a1d1 b8d7 f1e2 e7e6 e3d4 f8c5 e1d2 c5d6 h1e1 e8g8 d2c1 f6d5 e2f1 b7b6 c3d5 e6d5 g2g3 f8d8 f3e5 d7e5 d4e5 g8f8 a2a4 d6c5 e5d4 c5d4 d1d4 c8e6');
INSERT INTO evaluations VALUES(33763,30,'cp -171',NULL);
INSERT INTO evaluations VALUES(33763,31,'cp -183',NULL);
INSERT INTO evaluations VALUES(33763,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(33763,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(33763,34,'cp -163',NULL);
INSERT INTO evaluations VALUES(33763,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(33763,36,'cp -171','d1e1 g4e5 b1c3 e5d7 g1f3 c7c6 c1e3 d7f6 a1d1 b8d7 f1e2 e7e6 h1g1 f6d5 c3d5 e6d5 g2g4 f8e7 f3d4 g7g6 g4g5 d7b6 e1d2 c8d7 d2c1 e8c8 h2h4 d8e8 d4b3 e7d6 e2d3 c8c7 e3d4 d6f4 b3d2 f4e5 d4e3 d5d4');
INSERT INTO evaluations VALUES(39621,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(39621,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(39621,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(39621,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(39621,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(39621,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(39621,36,'cp 11','d3d4 d7d6 g1f3 g8f6 f1d3 e7e5 d4e5 d6e5 e1g1 d8e7 c2c3 e8g8 f1e1 f8d8 d1c2 c7c5 a2a4 b8c6 d3b5 h7h6 h2h3 c8d7 b1a3 c6a5 b5f1 d7c6 f3d2 a7a6 b2b3 a8b8 a3c4 a5c4 d2c4 b7b5 a4b5 a6b5 c4a5 c6a8 f2f3 e7c7');
INSERT INTO evaluations VALUES(4850,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(4850,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(4850,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(4850,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(4850,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(4850,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(4850,36,'cp 86','b5c6 d7c6 f3e5 d8g5 e1g1 g5e5 d2d4 c5d4 c3d4 e5d6 b2b3 c8e6 c1a3 d6f4 b1c3 e8c8 d1c2 f7f5 a3c1 f4g4 d4d5 c6d5 f2f3 g4h4 e4d5 e6d5 c2f5 c8b8 c1f4 g8f6 f1d1 d5c6 f4g3 h4b4 f5e5 b4b6 g3f2 b6b4 a1c1 d8d1 c3d1');
INSERT INTO evaluations VALUES(4844,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(4844,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(4844,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(4844,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(4844,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(4844,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(4844,36,'cp 85','d7c6 f3e5 d8g5 e1g1 g5e5 d2d4 e5d6 d4c5 d6c5 c1e3 c5e7 f1e1 h7h6 f2f4 g8f6 f4f5 b7b6 e3f2 f6d7 e4e5 d7e5 d1h5 e8g8 f5f6 g7f6 b1d2 g8g7 f2h4 e7c5 g1h1 e5g4 h5c5 b6c5 h2h3 g4e5 d2e4 e5d7 e4g3 d7b6 g3h5 g7g6 h5f6 b6d5 f6e4 c5c4 e4d2 c8f5 d2c4 f7f6 c4e3');
INSERT INTO evaluations VALUES(4783,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(4783,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(4783,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(4783,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(4783,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(4783,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(4783,36,'cp 105','f3e5 d8e7 e1g1 e7e5 d2d4 e5d6 d4c5 d6c5 c1e3 c5e7 f1e1 h7h6 b1d2 g8f6 h2h3 b7b6 c3c4 f6d7 d1h5 a6a5 a1d1 c8a6 e3d4 e8g8 e1e3 c6c5 d4c3 e7g5 h5g5 h6g5 e4e5 f8d8 e5e6 d7f6 c3f6 g7f6 e6e7');
INSERT INTO evaluations VALUES(4776,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(4776,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(4776,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(4776,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(4776,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(4776,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(4776,36,'cp 103','d8g5 e1g1 g5e5 d2d4 e5d6 d4c5 d6c5 c1e3 c5d6 d1d6 c7d6 f1d1 g8f6 e3g5 e8e7 b1d2 h7h6 g5f6 g7f6 d2c4 h8d8 c4e3 a6a5 f2f4 b7b5 g1f2 a5a4 d1d4 c8e6 f4f5 c6c5 d4d2 e6c4 h2h3 h6h5 f2g3 e7d7 g3f4 d7c7 a2a3 d8e8 g2g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(19949,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(19949,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19949,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19949,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(19949,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19949,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19949,36,'cp 0','g8f6 b1c3 e7e6 d2d4 f6e4 c3e4 d6d5 d4c5 d5e4 d1d8 c6d8 f3d2 f7f5 d2b3 d8c6 c1e3 c8d7 e1c1 c6e5 b3d2 a7a5 e3d4 e5c6 d4e3');
INSERT INTO evaluations VALUES(13254,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13254,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13254,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(13254,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13254,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(13254,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(13254,36,'cp 0','b1c3 e7e6 d2d4 c5d4 f3d4 f8e7 e1g1 e8g8 c1e3 a7a6 a2a4 c8d7 c4d3 d8c7 f2f4 c6d4 e3d4 d7c6 d1e1 a8e8 a4a5 f6d7 e1g3 e7f6 d4f6 d7f6 f1f2 f6h5 g3e3 h5f6');
INSERT INTO evaluations VALUES(1890,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(1890,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(1890,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(1890,33,'cp 269',NULL);
INSERT INTO evaluations VALUES(1890,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(1890,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(1890,36,'cp 285','b1c3 b4a5 e2e4 c6e5 g5d2 c7c6 c3d5 e5f3 d1f3 a5d8 d5f4 d7d6 e1c1 g8f6 f1c4 d8e7 h1e1 f6g4 e4e5 g4e5 f3b3 g7g6 f4d3 c8e6 d3e5 d6e5 e1e5 e8c8 b3e3 f8g7 c4e6 c8b8 e5e4 f7e6 e4e6 e7f8 e3f4 f8f4 d2f4 b8c8 d1d8 c8d8 f4e3 d8d7 e6e4 h8f8 e3a7 g7h6 c1b1 f8f4 e4f4 h6f4 g2g3 f4e5 a7b6');
INSERT INTO evaluations VALUES(1888,30,'cp 290',NULL);
INSERT INTO evaluations VALUES(1888,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(1888,32,'cp 294',NULL);
INSERT INTO evaluations VALUES(1888,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(1888,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(1888,35,'cp 275',NULL);
INSERT INTO evaluations VALUES(1888,36,'cp 304','b4b2 g5d2 f8b4 a1b1 b2a3 c3d5 b4a5 b1b5 a5d2 d1d2 e8d8 e2e4 a7a6 b5b3 a3a2 f1e2 a2a1 e2d1 a1a5 c2c3 h7h6 d2f4 g8e7 f4f7 e7d5 e4d5 c6e7 d5d6 c7d6 e5d6 h8e8 e1g1 e7f5 d1c2 f5d6 f7g7 d8c7 c2g6 a5d5 c3c4 d5g8 g7h6 e8e6');
INSERT INTO evaluations VALUES(1892,30,'cp 302',NULL);
INSERT INTO evaluations VALUES(1892,31,'cp 302',NULL);
INSERT INTO evaluations VALUES(1892,32,'cp 312',NULL);
INSERT INTO evaluations VALUES(1892,33,'cp 304',NULL);
INSERT INTO evaluations VALUES(1892,34,'cp 308',NULL);
INSERT INTO evaluations VALUES(1892,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(1892,36,'cp 299','g5d2 f8b4 a1b1 b2a3 c3d5 b4a5 b1b5 a5d2 d1d2 e8d8 e2e4 a7a6 b5b3 a3a2 f1e2 a2a1 e2d1 a1a5 c2c3 h7h6 d2f4 g8e7 f4f7 e7d5 e4d5 c6e7 d5d6 c7d6 f7g7 h8g8 g7f6 d6e5 f3e5 a5c5 e1g1 d8c7 f1e1 a6a5 e5d3 c5d6 f6f3 e7c6 b3b5 b7b6 d3f4');
INSERT INTO evaluations VALUES(19827,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19827,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19827,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(19827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19827,36,'cp 0','d2d4 a7a6 f1e2 c5d4 c3d4 c8f5 g1f3 e7e6 b1c3 g8e7 e1g1 f5e4 f3h4 e7g6 h4f3');
INSERT INTO evaluations VALUES(19817,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19817,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19817,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19817,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19817,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19817,36,'cp 0','a7a6 f1e2 c5d4 c3d4 c8f5 g1f3 e7e6 e1g1 f5e4 b1c3 g8e7 f3h4 e7g6 h4f3');
INSERT INTO evaluations VALUES(22676,30,'cp -728',NULL);
INSERT INTO evaluations VALUES(22676,31,'cp -728',NULL);
INSERT INTO evaluations VALUES(22676,32,'cp -717',NULL);
INSERT INTO evaluations VALUES(22676,33,'cp -706',NULL);
INSERT INTO evaluations VALUES(22676,34,'cp -721',NULL);
INSERT INTO evaluations VALUES(22676,35,'cp -717',NULL);
INSERT INTO evaluations VALUES(22676,36,'cp -713','d8h4 g2g3 h4e4 d1e2 e4h1 g1f3 d7d5 e5d6 f8e7 d6e7 c8g4 e2g2 g4f3 g2h1 f3h1 b1a3 g8e7 c2c3 e8c8 d2d4 e7d5 e1f2 h7h6 a3c4 h1e4 c1d2 d5f6 f2g1 h8e8 c4e3 c8b8 b2b4 c6e7 a2a4 e7f5 e3f5 e4f5 f1b5 c7c6 d2f4 b8c8 b5c4 f6e4');
INSERT INTO evaluations VALUES(25728,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(25728,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(25728,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25728,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(25728,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(25728,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(25728,36,'cp 46','g8f6 f1e1 g7g6 d2d4 f8g7 h2h3 e8g8 a4c2 f8e8 d4d5 c6e7 c3c4 b7b5 b1c3 c7c6 d5c6 d7c6 b2b3 b5b4 c3d5 e7d5 c4d5 c6b5 c1g5 h7h6 g5e3 a8c8 a2a4 b4a3 a1a3 d8d7 d1d2 c8c7 c2b1 g8h7 b1d3 b5d3 d2d3 d7b5');
INSERT INTO evaluations VALUES(24329,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(24329,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(24329,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(24329,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(24329,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(24329,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(24329,36,'cp 50','f1e1 b7b5 a4b3 f8e7 d2d4 e8g8 b1d2 a6a5 a2a4 b5a4 b3a4 e5d4 f3d4 c6d4 c3d4 c7c5 d4c5 d6c5 d1c2 d7e6 d2f3 c5c4 e1d1 d8c7 h2h3 f8d8 c1e3 d8d1 a1d1 a8d8 d1d8 e7d8 f3d4 e6d7 e4e5 f6d5 a4d7 d5e3 f2e3 c7d7 c2c4');
INSERT INTO evaluations VALUES(24330,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24330,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(24330,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(24330,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(24330,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(24330,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(24330,36,'cp 40','b7b5 a4c2 c6e7 d2d4 e7g6 b1d2 f8e7 h2h3 e8g8 d2f1 f8e8 f1g3 h7h6 c1e3 e7f8 d1d2 c7c6 b2b3 d7e6 c2d3 a6a5 e1c1 a5a4 c3c4 a4b3 a2b3 b5c4 b3c4 e5d4 f3d4 a8a1 c1a1 g6e5 d4e6 e8e6 a1a7 e5d3 d2d3 f6d7');
INSERT INTO evaluations VALUES(24350,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(24350,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(24350,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(24350,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(24350,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(24350,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(24350,36,'cp 52','a4b3 f8e7 h2h3 e8g8 b3c2 h7h6 d2d4 f8e8 b1d2 a6a5 d2f1 a5a4 c1d2 e7f8 f1g3 d8c8 d2e3 c8b7 a1c1 c6a5 f3d2 c7c5 d4d5 a5c4 d2c4 b5c4 c1b1 g7g6 d1d2 g8h7 f2f4 e5f4 e3f4 b7b8 a2a3 f8g7 b1a1 b8b2 f4d6 h7g8 d6c5');
INSERT INTO evaluations VALUES(24346,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(24346,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(24346,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(24346,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(24346,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(24346,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(24346,36,'cp 46','f8e7 d2d4 e8g8 b1d2 a6a5 b3c2 h7h6 d2f1 f8e8 c1d2 e7f8 f1g3 a5a4 h2h3 d8c8 d2e3 c8b7 a1c1 c6a5 f3d2 c7c5 d4d5 a5c4 d2c4 b5c4 c1b1 g7g6 d1f3 f6h7 g3f1 b7a6 g1h2 e8b8 f1d2 f8g7 f3e2 d7b5 a2a3 h7f6 e1g1');
INSERT INTO evaluations VALUES(56105,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(56105,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(56105,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(56105,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(56105,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(56105,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(56105,36,'cp -56','c7c6 b5e2 f8d6 g1f3 g8f6 d2d4 e8g8 h2h3 f8e8 e1g1 h7h6 f1e1 c8f5 c2c4 d5c4 e2c4 e8e1 d1e1 b8d7 b1c3 d8c7 c4b3 a8e8 c1e3 d7b6 e1f1 f5e6 b3e6 e8e6 a1e1 c7c8 e3d2 e6e1 f3e1 a7a6 f1e2 b6d5 c3d5 f6d5 b2b3 c8f5 e1d3');
INSERT INTO evaluations VALUES(53679,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(53679,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(53679,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(53679,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(53679,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(53679,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(53679,36,'cp -53','b5e2 g8f6 d2d4 f8d6 g1f3 e8g8 e1g1 h7h6 c2c4 c8f5 e2d3 f5d3 d1d3 d5c4 d3c4 f8e8 h2h3 b8d7 b1c3 d7b6 c4d3 b6d5 c1d2 d5b4 d3b1 d8d7 a2a3 b4d5 b1c2 d5c7 f1e1 a8d8 e1e8 d8e8 a1d1 f6d5 c2d3 d5c3 d2c3');
INSERT INTO evaluations VALUES(44744,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(44744,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(44744,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44744,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(44744,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(44744,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(44744,36,'cp 54','e4e5 f6g4 e5d6 d8d6 b1c3 f8g7 h2h3 g4f6 c3b5 d6b6 c1f4 f6d5 c4d5 b6b5 e1g1 b5d7 d5b3 b8c6 f1e1 b7b6 b3a4 c8b7 f3e5 g7e5 f4e5 e8g8 e5g3 d7d5 a4b3 d5g5 d1f3 g5f6 f3f6 e7f6 e1e3 a8d8 a1e1 c6d4 e3e7 b7d5 b3d5 d8d5');
INSERT INTO evaluations VALUES(2416,30,'cp 350',NULL);
INSERT INTO evaluations VALUES(2416,31,'cp 330',NULL);
INSERT INTO evaluations VALUES(2416,32,'cp 338',NULL);
INSERT INTO evaluations VALUES(2416,33,'cp 317',NULL);
INSERT INTO evaluations VALUES(2416,34,'cp 372',NULL);
INSERT INTO evaluations VALUES(2416,35,'cp 310',NULL);
INSERT INTO evaluations VALUES(2416,36,'cp 372','d2d4 c7c6 d4e5 d8e8 c1e3 e8f7 f3h3 f7f5 g2g4 f5e5 f2f4 e5d6 f4f5 e6f7 h1f1 h7h5 e1c1 f7g8 g4g5 h8h7 h3f3 c8f5 c3d5 c6d5 c4d5 g8h8 d5b7 d6c7 b7e4 f5e4 f3e4 c7c6 e4c6 e7c6 g5g6 h7h6 e3h6 g7h6 d1d7 f8g7 f1f7 g7e5 h2h3 h8g8 c2c3 a8e8 g6g7 e5g3 c1c2 a7a6 f7f6 c6e5 f6f8 e8f8 g7f8q g8f8 d7a7 e5c4 a7a6 f8f7 a6a7 f7g6');
INSERT INTO evaluations VALUES(2419,30,'cp 306',NULL);
INSERT INTO evaluations VALUES(2419,31,'cp 370',NULL);
INSERT INTO evaluations VALUES(2419,32,'cp 357',NULL);
INSERT INTO evaluations VALUES(2419,33,'cp 356',NULL);
INSERT INTO evaluations VALUES(2419,34,'cp 335',NULL);
INSERT INTO evaluations VALUES(2419,35,'cp 378',NULL);
INSERT INTO evaluations VALUES(2419,36,'cp 342','c7c6 d4e5 d8e8 c1e3 e8f7 f3h3 f7f5 g2g4 f5e5 f2f4 e5d6 f4f5 e6f7 e1c1 h7h5 h1f1 f7g8 g4g5 h8h7 h3f3 c8f5 c3d5 c6d5 c4d5 g8h8 d5b7 d6c7 b7e4 f5e4 f3e4 c7c6 e4c6 e7c6 g5g6 h7h6 e3h6 g7h6 d1d7 f8g7 f1f7 a8g8 d7d6 g8c8 c2c3 g7e5 d6d7 e5h2 f7h7 h8g8 c1c2 h2e5 h7h6 h5h4 h6h4 c8f8 h4g4 f8f2 c2b3 e5f6 g6g7 c6e5');
INSERT INTO evaluations VALUES(2395,30,'cp 378',NULL);
INSERT INTO evaluations VALUES(2395,31,'cp 371',NULL);
INSERT INTO evaluations VALUES(2395,32,'cp 367',NULL);
INSERT INTO evaluations VALUES(2395,33,'cp 363',NULL);
INSERT INTO evaluations VALUES(2395,34,'cp 367',NULL);
INSERT INTO evaluations VALUES(2395,35,'cp 376',NULL);
INSERT INTO evaluations VALUES(2395,36,'cp 348','d4e5 d8e8 c1e3 e8f7 f3h3 f7f5 g2g4 f5e5 f2f4 e5d6 f4f5 e6f7 h1f1 h7h5 e1c1 f7g8 g4g5 h8h7 h3f3 c8f5 c3d5 c6d5 c4d5 g8h8 d5b7 d6c7 b7e4 f5e4 f3e4 c7c6 e4c6 e7c6 g5g6 h7h6 e3h6 g7h6 d1d7 f8g7 f1f7 a8g8 d7d6 g8c8 d6e6 h8g8 c2c3 g7e5 h2h3 e5g3 c1b1 a7a5 a2a4 c6e5 f7a7 e5c4 b1a2 c8b8');
INSERT INTO evaluations VALUES(43192,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(43192,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(43192,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(43192,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(43192,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(43192,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(43192,36,'cp 45','f1g2 e8g8 b1c3 d7d6 g1f3 c7c5 d4d5 e7e6 e1g1 f8e8 c1f4 e6e5 f4g5 h7h6 g5f6 d8f6 e2e4 f6e7 g1h1 b8d7 g2h3 a7a6 a2a4 a8b8 d1e2 h6h5 a4a5 g7h6 a1a3 e8f8 f1a1 e7f6 f3h4 h6g7 e2c2 g7h6 h1g1 f6d8 c3d1');
INSERT INTO evaluations VALUES(35279,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(35279,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(35279,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(35279,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(35279,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(35279,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(35279,36,'cp 103','g1f3 g8f6 e1g1 d7d6 d2d4 b8c6 b1c3 c8g4 c4b5 a7a6 b5c6 b7c6 d1d3 g4f3 d3f3 e7e6 d4e5 d6e5 c1g5 f6d7 f1d1 a6a5 h2h3 f8d6 g5e3 e8g8 c3a4 d6b4 a2a3 b4e7 f3f5 e6f5 e4f5 d7f6 d1d3 a8d8 a1d1 d8d3 d1d3');
INSERT INTO evaluations VALUES(34951,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(34951,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(34951,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(34951,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(34951,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(34951,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(34951,36,'cp 24','g1f3 b8c6 b1c3 d5d8 f1g2 g8f6 e1g1 h7h6 d2d3 f8e7 c1e3 e8g8 d3d4 e5d4 f3d4 c6d4 e3d4 c7c6 d1a4 d8c7 f1d1 f8d8 a1c1 c8e6 b2b4 a7a6 a2a3 c7c8 e2e4 e6g4 f2f3 g4e6 a4c2 c8d7 d4e3');
INSERT INTO evaluations VALUES(34949,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(34949,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(34949,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(34949,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(34949,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(34949,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(34949,36,'cp 16','b8c6 b1c3 d5d6 f1g2 g8f6 e1g1 f8e7 d2d4 c6d4 f3e5 d6e5 c1f4 d4e2 c3e2 e5a5 d1c2 c7c5 e2c3 e8g8 c2b3 f8d8 a1d1 c8g4 b3b7 g4d1 f4c7 a5b4 c3d1 c5c4 b7b4 e7b4 g2a8 d8a8 d1e3 c4c3 b2c3 b4c3 c7f4 c3a5 f1c1 a5b6 e3c4');
INSERT INTO evaluations VALUES(1728,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(1728,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(1728,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(1728,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(1728,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(1728,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(1728,36,'cp 26','f1g2 g8f6 d2d3 f8d6 b1d2 e8g8 e1g1 d5e6 d2c4 a7a5 b2b3 f8e8 c1b2 e6f5 a2a3 f5h5 b3b4 a5b4 a3b4 a8a1 d1a1 b7b5 c4d6 c7d6 b2c3 h7h6 a1a8 c8d7 a8b7 h5g4 f1a1 c6e7 a1a7 g4e6 f3d2 e8c8 d2e4 f6e4 g2e4 d7c6 e4c6 e7c6');
INSERT INTO evaluations VALUES(47009,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(47009,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(47009,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(47009,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(47009,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(47009,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(47009,36,'cp 35','d6d5 d1d3 b8c6 e2e4 d5e4 c3e4 c8g4 f1e2 e7e6 e4f6 g7f6 c1e3 d8d5 d3b3 d5b3 a2b3 c6b4 e1d2 b4d5 c2c3 a7a6 b3b4 h8g8 g2g3 h7h5 f3e1 g4e2 d2e2 a8d8 e1g2 c7c6 e3d2 e6e5 d4e5 f6e5 g2e3 f7f6 h1d1 d5e3 d2e3 d8d1');
INSERT INTO evaluations VALUES(4177,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4177,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4177,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4177,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(4177,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4177,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4177,36,'cp 6','c5e3 f2e3 d7d6 a2a4 c8e6 b2b3 c6e7 h2h3 e7g6 c3e2 e6c4 b3c4 c7c6 f3d2 d6d5 e4d5 c6d5 e2g3 a8c8 d1e1 c8c6 e3e4 d5e4 d2e4 g6f4 a1b1 b7b6 a4a5 c6e6 e4f6 e6f6 a5b6 d8d4 g1h1 f6b6 b1b6 a7b6 g3e2 f4e2 e1e2');
INSERT INTO evaluations VALUES(4223,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4223,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4223,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4223,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(4223,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4223,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(4223,36,'cp 0','f2e3 d7d6 d3d4 c8g4 d1d2 g4h5 a1e1 h5g6 c4d3 f8e8 d2f2 f6g4 f2d2');
INSERT INTO evaluations VALUES(4221,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4221,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4221,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4221,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4221,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4221,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4221,36,'cp 0','d7d6 d3d4 c8g4 h2h3 g4h5 d1d3 c6b4 d3d2 b4c6');
INSERT INTO evaluations VALUES(3651,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3651,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3651,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3651,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3651,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3651,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3651,36,'cp 0','d3d4 c8g4 h2h3 e5d4 e3d4 g4h5 d1d2 f6e4 c3e4 d6d5 e4g3 h5f3 f1f3 d5c4 c2c3 c6e5 f3e3 e5g6 a1f1 d8g5 f1f5 g5d8');
INSERT INTO evaluations VALUES(3653,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3653,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3653,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3653,36,'cp 0','c8g4 h2h3 g4h5 d1d3 f8e8 a2a3 e5d4 e3d4 f6e4 c4f7 h5f7 c3e4 f7g6 d3b3 g6f7');
INSERT INTO evaluations VALUES(23715,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23715,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23715,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23715,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23715,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23715,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23715,36,'cp 0','h2h3 g4h5 d1d3 f8e8 a2a3 e5d4 e3d4 f6e4 c4f7 h5f7 c3e4 f7g6 d3b3 g6f7');
INSERT INTO evaluations VALUES(39483,30,'cp 903',NULL);
INSERT INTO evaluations VALUES(39483,31,'cp 901',NULL);
INSERT INTO evaluations VALUES(39483,32,'cp 913',NULL);
INSERT INTO evaluations VALUES(39483,33,'cp 913',NULL);
INSERT INTO evaluations VALUES(39483,34,'cp 913',NULL);
INSERT INTO evaluations VALUES(39483,35,'cp 905',NULL);
INSERT INTO evaluations VALUES(39483,36,'cp 919','e5h8 e8f8 d2d4 e7g5 c1g5 d8g5 b1c3 h7h5 g1f3 g5f6 h8h7 c7c6 e4e5 f6g7 h7g7 f8g7 c3e4 d7d5 e5d6 c8f5 f3d2 b8d7 f2f3 g8f6 e1f2 b7b6 f1d3 f6d5 c2c3 d5f4 d3f1 c6c5 d4c5 b6c5 f1c4 f4e6 a2a4 e6f4 b2b3 a8d8');
INSERT INTO evaluations VALUES(37504,30,'cp 902',NULL);
INSERT INTO evaluations VALUES(37504,31,'cp 905',NULL);
INSERT INTO evaluations VALUES(37504,32,'cp 904',NULL);
INSERT INTO evaluations VALUES(37504,33,'cp 906',NULL);
INSERT INTO evaluations VALUES(37504,34,'cp 909',NULL);
INSERT INTO evaluations VALUES(37504,35,'cp 909',NULL);
INSERT INTO evaluations VALUES(37504,36,'cp 913','e8f8 d2d4 e7g5 c1g5 d8g5 b1c3 h7h5 g1f3 g5f6 h8f6 g8f6 e4e5 f6e8 f3g5 b8c6 f1c4 c6d8 e1c1 a7a6 c4b3 d7d6 h1e1 b7b5 c3e4 a6a5 a2a3 a5a4 b3a2 f8g7 d1d3 d8e6 a2e6 c8e6 g5e6 f7e6');
INSERT INTO evaluations VALUES(35554,30,'cp 904',NULL);
INSERT INTO evaluations VALUES(35554,31,'cp 909',NULL);
INSERT INTO evaluations VALUES(35554,32,'cp 913',NULL);
INSERT INTO evaluations VALUES(35554,33,'cp 910',NULL);
INSERT INTO evaluations VALUES(35554,34,'cp 921',NULL);
INSERT INTO evaluations VALUES(35554,35,'cp 917',NULL);
INSERT INTO evaluations VALUES(35554,36,'cp 919','d2d4 e7g5 c1g5 d8g5 b1c3 h7h5 g1f3 g5f6 h8f6 g8f6 e4e5 f6e8 f3g5 c7c6 e1c1 d7d5 e5d6 e8d6 d4d5 c6c5 c3b5 d6b5 f1b5 a7a6 b5c4 f7f6 g5f3 h5h4 f3h4 f8g7 c4d3 f6f5 h4f3 b8d7 d5d6 g7f7');
INSERT INTO evaluations VALUES(50184,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(50184,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50184,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(50184,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(50184,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50184,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(50184,36,'cp 2','f8g7 b1c3 e8g8 g1f3 d7d5 a1c1 d5c4 e2e3 c8e6 f3g5 e6d5 e3e4 h7h6 e4d5 h6g5 f4g5 f6d5 f1c4 d5b6 c4b3 d8d4 d1d4 g7d4 c3b5 d4b2 c1c2 b2e5 b5c7 e5c7 c2c7 e7e6 c7b7 b8c6 b7c7 c6d4 b3d1 e6e5 h2h4 d4e6 c7b7 e6c5 b7e7 b6d5 d1f3 d5e7 f3a8 f8a8');
INSERT INTO evaluations VALUES(43176,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43176,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(43176,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(43176,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(43176,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43176,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(43176,36,'cp 3','b1c3 e8g8 e2e4 d7d6 g1f3 c8g4 f1e2 b8c6 d4d5 e7e5 f4g5 g4f3 e2f3 c6d4 e1g1 c7c5 d5c6 b7c6 b2b4 d8c7 a1b1 f8b8 a2a4 a7a6 d1d2 f6d7 f3g4 b8b7 g4d7 c7d7 b4b5 c6b5 a4b5 a6b5 c4b5 d4b5 c3d5 d7e6');
INSERT INTO evaluations VALUES(54692,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(54692,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(54692,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(54692,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(54692,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(54692,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(54692,36,'cp 129','b1c3 g8f6 f1d3 f5e4 c3e4 f6e4 d3e4 g7g6 e4d5 f8g7 c2c3 e7e6 d5b3 b8c6 g1f3 d8f6 e1g1 e8g8 f1e1 c8d7 c1g5 f6f5 b3c2 f5f7 d1d2 c6e7 c2e4 a8b8 g5h6 g7h6 d2h6 f7g7 h6d2 g7f6 f3g5 g8h8 c3c4 e7f5');
INSERT INTO evaluations VALUES(54691,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(54691,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(54691,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(54691,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(54691,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(54691,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(54691,36,'cp 152','g8f6 f1d3 f6e4 c3e4 f5e4 d3e4 g7g6 c2c3 d6d5 e4c2 f8g7 g1f3 b8c6 e1g1 e8g8 h2h3 c8f5 c2b3 h7h6 f3h4 e7e6 h4f5 e6f5 d1e2 d8d6 e2b5 c6e7 f1e1 f5f4 c1d2 c7c6 b5b7 e7f5 b3a4 a8c8 a4c2 c8c7 b7a6 f4f3 c2f5 f8f5 g2g3 d6d7 a6d3');
INSERT INTO evaluations VALUES(46579,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(46579,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(46579,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(46579,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(46579,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(46579,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(46579,36,'cp 162','f1d3 f5e4 c3e4 f6e4 d3e4 g7g6 c2c3 d6d5 e4c2 f8g7 g1f3 d8d6 h2h3 e8g8 e1g1 c8f5 c1g5 b8d7 g5h4 c7c5 f1e1 e7e6 a1c1 c5d4 c2f5 g6f5 c3d4 d7f6 h4f6 g7f6 d1b3 b7b6 e1e2 a8c8 c1e1 c8e8 h3h4 g8h8 f3g5 f6g5');
INSERT INTO evaluations VALUES(20766,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(20766,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(20766,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(20766,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(20766,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(20766,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(20766,36,'cp 242','d4a4 c8d7 a4b3 d8c8 f1c4 d7e6 c4b5 e6d7 b5d7 c8d7 f2f3 f8e7 h2h4 a8c8 c1e3 a7a6 e1c1 h7h5 c1b1 c8c6 a2a3 b7b5 g2g3 g8f6 e3g5 e8g8 g5f6 e7f6 c3d5 f8c8 d1d2 f6d8 b3d3 c6c5 h1d1 c5c6 d3e2 d7h3 d5b4 c6b6 d2d3 h3g3');
INSERT INTO evaluations VALUES(20768,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(20768,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(20768,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(20768,33,'cp 229',NULL);
INSERT INTO evaluations VALUES(20768,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(20768,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(20768,36,'cp 236','c8d7 a4b3 d8c8 f1c4 d7e6 c4b5 e6d7 b5d7 c8d7 f2f3 f8e7 h2h4 g8f6 c1g5 e8g8 e1c1 f8c8 c1b1 c8c6 g5f6 e7f6 c3d5 a8c8 d1d2 f6d8 g2g4 c6c5 g4g5 b7b5 c2c3 a7a5 a2a3 d7b7 b3d1 b5b4 d1b3 c8b8 c3b4 a5b4 a3b4 b8c8 h1d1');
INSERT INTO evaluations VALUES(25988,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(25988,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(25988,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(25988,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(25988,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(25988,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(25988,36,'cp 231','a4b3 d8c8 f1c4 d7e6 c4b5 e6d7 b5d7 c8d7 f2f3 f8e7 h2h4 a7a6 c1e3 d7c6 e1c1 b7b5 c1b1 g8f6 e3g5 e8g8 g5f6 e7f6 g2g4 f8c8 g4g5 f6d8 a2a3 a8b8 d1d3 a6a5 h1d1 b5b4 d3d6 c6e8 c3d5 a5a4 b3e3 d8e7 d5e7 e8e7 a3b4 b8b4 e3d2 b4b8');
INSERT INTO evaluations VALUES(24823,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24823,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24823,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(24823,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24823,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24823,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24823,36,'cp 21','e7e6 d3f5 e6f5 e1g1 d8d7 d1d3 f8e7 c2c4 c6b4 d3e2 d5c4 e2c4 c7c6 f1c1 b4d5 f4g3 e7d6 g3d6 d7d6 b1c3 e8g8 f3e5 g7g6 c4b3 d6e7 a2a4 a7a5 c3d5 f6d5 c1c5 g8g7 h2h3 f8e8 e5d3 f5f4 d3f4 d5f4 e3f4 e7e4 a1d1');
INSERT INTO evaluations VALUES(25160,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(25160,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(25160,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(25160,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(25160,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(25160,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(25160,36,'cp 33','d3f5 e6f5 e1g1 f6e4 b1c3 d8d7 c3e4 f5e4 f3d2 f8e7 c2c4 e8c8 c4d5 d7d5 d2b1 g7g5 f4g3 f7f5 b1c3 d5d7 a1c1 c8b8 c3b5 f5f4 e3f4 a7a6 d4d5 g5f4 g3f4 a6b5 d5c6 d7f5 d1b3 d8d3 c1c3 h8d8 g2g3 b7c6 b3c2 e7d6 f4e3');
INSERT INTO evaluations VALUES(25134,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(25134,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25134,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(25134,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(25134,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(25134,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(25134,36,'cp 16','e6f5 e1g1 f8d6 f4d6 c7d6 b1c3 a8c8 f3e1 a7a6 e1d3 e8g8 d3f4 c6e7 f4h5 f6h5 d1h5 c8c4 a2a4 d8d7 f1b1 d7e6 h5d1 h7h6 a4a5 f8c8 a1a4 c4a4 c3a4 c8c7 a4b6 g7g5 c2c3 c7c6');
INSERT INTO evaluations VALUES(25244,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(25244,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25244,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(25244,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(25244,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(25244,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(25244,36,'cp 23','e1g1 f8d6 f4g5 h7h6 g5f6 d8f6 b1d2 c6e7 c2c4 d5c4 d2c4 e8g8 d1b3 a8b8 g2g3 f6e6 c4e5 e7d5 b3a4 b8a8 a4d7 e6d7 e5d7 f8e8 a1c1 g7g5 d7c5 d6c5 c1c5 c7c6 f3d2 a8d8 c5a5 a7a6 g1g2 g5g4 f1c1 e8e7 d2c4');
INSERT INTO evaluations VALUES(44467,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(44467,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44467,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44467,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(44467,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(44467,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(44467,36,'cp 34','d2d4 d7d5 g1f3 e7e6 e2e3 b8d7 d1c2 f8d6 f1d3 e6e5 c4d5 c6d5 c3b5 d6b8 d4e5 d7e5 f3e5 b8e5 c1d2 e8g8 d2c3 e5c3 c2c3 f6e4 c3d4 c8d7 e1g1 a7a6 b5c3 e4c3 b2c3 d7c6 a2a4 d8a5 h2h3 f8e8 f1c1');
INSERT INTO evaluations VALUES(49769,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49769,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(49769,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(49769,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(49769,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(49769,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(49769,36,'cp -24','b8c6 d4a4 f8b4 c2c3 b4c5 f1d3 d7d6 a4c2 a7a5 c1g5 h7h6 g5h4 d8e7 b1d2 g7g5 h4g3 f6h5 d2b3 c5b6 e1c1 h5g3 h2g3 c8d7 d3b5 e8c8 b3d4 g5g4 b5c6 b7c6 f3d2 e7g5 c1b1 d6d5 d2b3 d5e4 c2e4 g5d5 e4d5 c6d5 h1h5 a5a4');
INSERT INTO evaluations VALUES(16689,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(16689,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16689,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(16689,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(16689,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(16689,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(16689,36,'cp -18','d4a4 f8b4 c2c3 b4c5 f1d3 d7d6 a4c2 d8e7 b2b4 c5b6 e1g1 a7a5 b4b5 c6e5 f3e5 d6e5 c3c4 e8g8 c1a3 b6c5 a3c5 e7c5 b1c3 f8d8 f1d1 c8g4 d3e2 g4e2 c2e2 d8d4 c3d5 f6d5 c4d5 d4c4 h2h3 c4c2 d1d2 c2d2 e2d2 c5b5 a1d1 a8d8');
INSERT INTO evaluations VALUES(21010,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(21010,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(21010,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(21010,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(21010,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(21010,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(21010,36,'cp -36','e7e5 d4e5 c6e5 g1f3 e5d3 c2d3 f8d6 e3e4 g8e7 b1c3 e8g8 e1g1 c7c6 h2h3 c8e6 f3g5 d6c7 f2f4 d8d6 e4e5 c7b6 g1h2 d6c7 d3d4 e6f5 g5f3 f7f6 e5f6 f8f6 f3e5 c7d6 g2g4 f5e4 c3e4 d5e4');
INSERT INTO evaluations VALUES(21687,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(21687,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(21687,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(21687,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(21687,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(21687,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(21687,36,'cp -40','d4e5 c6e5 g1f3 e5d3 c2d3 c7c6 e1g1 f8c5 h2h3 g8e7 e3e4 e8g8 b1c3 c5b6 c1g5 f7f6 g5f4 c8e6 c3a4 e6f7 a2a3 b6c7 f4c7 d8c7 f1e1 e7g6 a4c3 a8d8 d1c2 g6f4 c3e2 f4e2 c2e2 f8e8 a1c1 c6c5 b2b4 b7b6 b4c5 b6c5 e2c2');
INSERT INTO evaluations VALUES(21650,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(21650,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(21650,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(21650,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(21650,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(21650,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(21650,36,'cp -39','c6e5 g1f3 e5d3 c2d3 c7c6 e1g1 f8c5 h2h3 g8e7 e3e4 e8g8 b1c3 c5b6 c1d2 h7h6 f1e1 d5e4 d3e4 e7g6 d2e3 c8e6 d1c2 d8f6 e4e5 f6e7 e3b6 a7b6 a1d1 a8d8 f3d4 e7g5 d4f3 g5f4 e1e4 f4f5 f3d4 f5g5 c2d2 g5d2 d1d2');
INSERT INTO evaluations VALUES(22185,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(22185,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22185,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(22185,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(22185,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(22185,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22185,36,'cp -37','g1f3 e5d3 c2d3 c7c6 e1g1 f8c5 e3e4 g8e7 b1c3 e8g8 h2h3 c5b6 f1e1 h7h6 c1d2 d5e4 d3e4 e7g6 d2e3 c8e6 d1c2 d8f6 e3b6 a7b6 e4e5 f6e7 a1d1 a8d8 c2e4 b6b5 f3d4 e7g5 d4e6 f7e6 d1d6 b5b4 c3e2 g5e5 d6e6 e5e4 e6e4 c6c5');
INSERT INTO evaluations VALUES(35092,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(35092,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35092,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35092,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(35092,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35092,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(35092,36,'cp -4','f8c5 g1f3 d7d6 d2d3 g7g6 h5h4 f6h4 f3h4 b8c6 h4f3 h7h6 b1c3 c5b6 c1e3 c8e6 c3d5 e8d7 e1g1 e6d5 c4d5 f7f5 f3h4 g8e7 d5f7 f5f4 e3b6 a7b6 h4g6 e7g6 f7g6 c6b4 f1c1 b4a2 c1d1 a2b4 a1c1 a8a2 c2c3 b4a6 d3d4 e5d4');
INSERT INTO evaluations VALUES(57492,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(57492,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(57492,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(57492,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(57492,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(57492,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(57492,36,'cp 155','d2d4 e5d4 g1h3 g8f6 e4e5 f6g4 e1g1 f8e7 c4b3 b8c6 f1e1 g7g5 c2c3 d7d5 c3d4 c8e6 b1c3 e8g8 f2f3 g4h6 f3f4 g5g4 h3g5 e7g5 f4g5 h6f7 d1d2 g8h8 c3e2 f8g8 g5g6 h7g6 e2f4 d8d7 b3c2 g6g5 f4h5 d7e7 a2a3 f5f4 g2g3 f4f3 h5f6 g8g7 b2b4 a7a6 f6h5');
INSERT INTO evaluations VALUES(57587,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(57587,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(57587,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(57587,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(57587,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(57587,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(57587,36,'cp 52','e7e6 g1f3 f8e7 c2c4 g8f6 f1g2 d7d5 d1c2 c7c6 b1c3 f6e4 c1f4 b8d7 a1d1 e8g8 e1g1 a7a5 h2h4 h7h6 c3e4 d5e4 f3e5 g7g5 e5d7 c8d7 h4g5 h6g5 f4e5 f8f7 c2b3 d7e8 f2f3 e4f3 e2f3 a5a4 b3c3 e7f6 f1e1 f7h7');
INSERT INTO evaluations VALUES(57362,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(57362,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(57362,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(57362,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(57362,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(57362,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(57362,36,'cp 59','c2c4 g8f6 g1f3 f8e7 b1c3 e8g8 f1g2 c7c6 e1g1 d7d5 d1c2 b8d7 f3g5 d7b8 f1d1 a7a5 a2a3 f6e4 g5e4 f5e4 f2f3 e4f3 e2f3 b8d7 c2e2 e7f6 f3f4 g8h8 c1e3 d7b6 c4c5 b6c4 e3f2 b7b5 c5b6 d8b6');
INSERT INTO evaluations VALUES(49771,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(49771,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(49771,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(49771,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(49771,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(49771,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49771,36,'cp 0','c1d2 b8c6 d4e3 f8e7 b1c3 d7d5 c3d5 f6d5 e4d5 d8d5 g1e2 c8f5 e2c3 d5c5 e3c5 e7c5 f1b5 e8c8 b5c6 b7c6 e1c1 h8e8 h1e1 c8b7 c3a4 c5d6 d2e3 h7h5 h2h3 a7a6 a4c5 d6c5 e3c5 d8d1 e1d1 e8e2 d1d2 e2e1 d2d1');
INSERT INTO evaluations VALUES(10430,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(10430,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10430,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(10430,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10430,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10430,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(10430,36,'cp 0','d7d6 e1g1 h7h6 f1e1 e8g8 c4b3 c8d7 h2h3 b7b5 a2a4 a7a6 c3c4 b5c4 b3c4 f8e8 f3h4 f6g4 h4f3');
INSERT INTO evaluations VALUES(9443,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9443,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9443,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9443,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9443,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(9443,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9443,36,'cp 3','e1g1 h7h6 f1e1 e8g8 d3d4 c8g4 c4b5 e5d4 c3d4 g4f3 g2f3 f6h5 e4e5 d6e5 b5c6 b7c6 d4e5 d8d1 e1d1 f8e8 f3f4 f7f6 e5f6 h5f6 c1b2 f6d5 b2e5 g7g5 d1d4 g5f4 e5f4 d5f4 d4f4 e8e2 a1b1 a8d8 b1b7 d8d7 b7a7 e2c2 g1g2 d7g7 g2h3 c2d2 a7a6 d2c2 a2a4 c2c3 f2f3');
INSERT INTO evaluations VALUES(22148,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(22148,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(22148,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(22148,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(22148,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(22148,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(22148,36,'cp 122','d1d4 g8f6 c1f4 f8e7 e1c1 e8g8 c1b1 c8e6 f2f3 f6d7 d4d2 e7f6 h2h4 f6e5 f4g5 f7f6 g5e3 c7c6 f3f4 e5c3 d2c3 d8e7 h4h5 h7h6 f1e2 e6f7 h1e1 d7c5 e3c5 d6c5 e2c4 f7c4 c3c4 g8h8 e4e5 f6e5 f4e5 a8e8 e5e6 e7f6 c4e4 f6g5');
INSERT INTO evaluations VALUES(2311,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(2311,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(2311,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(2311,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(2311,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(2311,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(2311,36,'cp 33','f8d6 b1c3 c8d7 b5c4 d8e8 a2a3 a7a6 c1e3 g8f6 c4a2 b7b5 h2h3 a8b8 e1g1 a6a5 f3g5 c6d8 f2f4 d8e6 f4e5 d6e5 f1f5 e5c3 b2c3 h7h6 g5e6 d7e6 a2e6 f7e6 f5c5 e8d7 a1f1 d7c8 c3c4 f6e4 c5b5 e4d6 b5c5 b8b2 c5a5 d6c4 a5a8 b2b8 a8b8 c8b8');
INSERT INTO evaluations VALUES(2343,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(2343,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(2343,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(2343,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(2343,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(2343,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(2343,36,'cp 40','b1c3 g8e7 c1e3 a7a6 b5e2 f7f6 e1c1 d8e8 c3a4 c8e6 a2a3 b7b5 a4c5 d6c5 e3c5 e8f7 b2b3 a8d8 d1d2 d8d2 f3d2 h8d8 c5e3 c6d4 e2d1 e6c8 h2h4 d4e6 h4h5 c8b7 d1g4 e7c8 f2f3 c8d6 c1b2 a6a5');
INSERT INTO evaluations VALUES(44960,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44960,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44960,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(44960,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(44960,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44960,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(44960,36,'cp 6','g7g6 e1g1 f8g7 c2c4 e8g8 c4d5 c6d5 b1c3 f6e4 f3d2 e4c3 b2c3 c8e6 e2e4 d5e4 d2e4 e6d5 f1e1 b8d7 h2h4 d7f6 c1g5 a8c8 a1c1 d5a2 h4h5 a2d5 h5h6 g7h8 d1e2 b7b6 g5f6 h8f6 e4f6 e7f6 g2d5 d8d5 c3c4 d5d4 e1d1');
INSERT INTO evaluations VALUES(44718,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44718,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(44718,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44718,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44718,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44718,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(44718,36,'cp 19','c2c4 f8g7 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b6b3 a2b3 b8c6 e5c6 b7c6 a1a6 f8c8 e1g1 h7h6 b3b4 f6e4 g2e4 d5e4 f4e3 c8b8 a6a4 g6g5 f1a1 b8b7 d4d5 c6d5 c3d5 f5e6 d5c3 a8c8 a4a7');
INSERT INTO evaluations VALUES(44713,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44713,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44713,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(44713,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(44713,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44713,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44713,36,'cp 0','f8g7 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 e7e6 e1g1 f8c8 f1c1 h7h6 h2h4 c6e5 f4e5 g6g5 h4g5 h6g5 f2f3 g5g4 e3e4 g4f3 e4f5 f3g2 g1g2 f6e4 e5g7 g8g7 c3e4 d5e4 f5e6 f7e6 c1c8 a8c8 g2f2 c8c2 f2e3');
INSERT INTO evaluations VALUES(18615,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(18615,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(18615,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(18615,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(18615,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(18615,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(18615,36,'cp 137','d2d4 c5d4 f3d4 g8f6 d4b5 a7a6 b5d6 f8d6 d1d6 d8e7 d6g3 f6h5 g3c7 h5f6 c4b3 d7d6 c7b6 f6d7 b6e3 d7c5 e3h6 b7b5 c1e3 f7f6 e1c1 c8d7 c1b1 c6e5 h6h4 c5b3 a2b3 b5b4 f2f4 e5f7 c3a4 d7a4 b3a4 e8g8 b2b3 f8e8 e3d4 e6e5 d4b2 a8c8');
INSERT INTO evaluations VALUES(18627,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(18627,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(18627,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(18627,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(18627,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(18627,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(18627,36,'cp 141','c5d4 f3d4 g8f6 d4b5 a7a6 b5d6 f8d6 d1d6 d8e7 d6g3 f6h5 g3c7 h5f6 f2f3 d7d6 c7e7 e8e7 c1e3 b7b5 c4e2 f6d7 e1c1 f7f6 d1d2 c6e5 h1d1 e5f7 d2d4 c8b7 a2a4 b5a4 d4a4 h8c8 h2h4 h7h6 a4a5 g6g5 c3a4 g5g4 d1d4 g4f3 g2f3 c8c7 c1b1 a8c8');
INSERT INTO evaluations VALUES(18641,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(18641,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(18641,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(18641,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(18641,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(18641,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(18641,36,'cp 141','f3d4 g8f6 d4b5 a7a6 b5d6 f8d6 d1d6 d8e7 d6g3 f6h5 g3c7 h5f6 c4b3 d7d6 c7b6 f6d7 b6e3 b7b5 c1d2 c8b7 e1c1 d7c5 h2h4 h7h5 f2f3 c6e5 e3d4 f7f6 d2e1 a8d8 c1b1 e8g8 c3e2 e5c6 d4e3 c5b3 a2b3 d6d5 e1c3 d5e4 e2f4 e7f7 d1d8 c6d8 f3e4 e6e5 f4d5 b7c6');
INSERT INTO evaluations VALUES(18637,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(18637,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(18637,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(18637,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(18637,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(18637,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(18637,36,'cp 118','g8f6 d4b5 a7a6 b5d6 f8d6 d1d6 d8e7 d6g3 f6h5 g3c7 h5f6 f2f3 d7d6 c7e7 e8e7 c1e3 b7b5 c4e2 f6d7 e1c1 f7f6 d1d2 c6e5 h1d1 e5f7 a2a4 b5b4 c3a2 a6a5 e2b5 d7e5 b2b3 c8d7 b5d7 e5d7 c2c3 b4c3 a2c3 h8c8 c1b2 a8b8 c3b5 d7c5 e3c5 c8c5 b5d6 f7d6 d2d6 c5c7 d6a6 c7b7');
INSERT INTO evaluations VALUES(54850,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(54850,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54850,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54850,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54850,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54850,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54850,36,'cp 23','g8f6 b1d2 f8g7 c2c3 e8g8 f1e2 b8d7 h2h3 c7c5 g1f3 b7b6 e1g1 c8b7 a2a4 a7a6 d1b3 a8c8 f1d1 f8e8 c3c4 a6a5 a1c1 e7e6 f4g3 g7f8 c4d5 b7d5 d2c4 f6e4 g3f4 d8e7 f3e5 c5d4 e3d4 d7e5 d4e5 e7b4 b3e3');
INSERT INTO evaluations VALUES(16495,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(16495,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(16495,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(16495,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(16495,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(16495,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(16495,36,'cp -130','d7d5 b1d2 f8b4 c1b2 d5e4 a2a3 b4d2 f3d2 c8g4 d1c1 e4d3 f1d3 f6d5 e1g1 d8g5 d2e4 g5c1 a1c1 d5f4 f2f3 g4f5 d3b5 f7f6 c1d1 a7a6 b5c6 b7c6 d1d2 f5e4 f3e4 f4e6 g1f2 a8d8 d2d8 e6d8 a3a4 d8b7 f2e3 e8d7 f1d1 b7d6 b2a3 h7h5');
INSERT INTO evaluations VALUES(15126,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(15126,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(15126,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(15126,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(15126,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(15126,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(15126,36,'cp -122','b1d2 f8b4 c1b2 d5e4 a2a3 b4d2 f3d2 c8g4 d1c1 e4d3 f1d3 f6d5 e1g1 d8g5 f1e1 e8c8 g2g3 f7f6 b3b4 g4h3 b4b5 c6a5 f2f3 h8e8 a3a4 h3e6 d3f1 b7b6 d2e4 g5c1 a1c1 c8b8 b2a3 c7c5 b5c6 a5c6 e4d6 c6d4 d6e8 d4f3 g1f2 f3e1');
INSERT INTO evaluations VALUES(42063,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42063,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42063,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42063,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42063,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42063,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42063,36,'cp 12','b4c3 b2c3 e8g8 e2e3 c7c5 f1e2 d5c4 e2c4 d8c7 c4e2 b7b6 e1g1 c8b7 c3c4 h7h6 c1b2 f8d8 a1c1 b8c6 d1c2 c5d4 f3d4 c6d4 e3d4 a8c8 c2b3 c7f4 f1d1 h6h5 b3g3 f4g3 f2g3 f6g4');
INSERT INTO evaluations VALUES(42093,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(42093,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(42093,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(42093,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(42093,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42093,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42093,36,'cp 14','b2c3 e8g8 e2e3 c7c5 f1e2 d5c4 e2c4 d8c7 c4e2 b7b6 c1b2 b8c6 e1g1 c8b7 c3c4 f8d8 a1c1 a8c8 f1e1 c6a5 f3e5 c7e7 e2d3 b7e4 d3f1 h7h6 h2h3 e4b7 b2c3 a5c6 e5c6 b7c6 c3b2 c6e4 d1b3 e4g6 f2f3');
INSERT INTO evaluations VALUES(42088,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(42088,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42088,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42088,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42088,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42088,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42088,36,'cp 0','e8g8 e2e3 c7c5 f1e2 d5c4 e2c4 d8c7 c4e2 b7b6 c1b2 c8b7 c3c4 f8d8 e1g1 b8c6 a1c1 c6a5 f3e5 a8c8 f1e1 h7h6 e2d3 b7e4 d3f1 c7e7 h2h3 f6d7 e5f3 d7f6');
INSERT INTO evaluations VALUES(37068,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37068,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37068,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37068,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37068,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37068,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37068,36,'cp 0','e2e3 c7c5 f1e2 d5c4 e2c4 d8c7 c4e2 b7b6 e1g1 f8d8 f1e1 c8b7 c1b2 b8c6 c3c4 c6a5 a1c1 a8c8 f3e5 h7h6 e2d3 b7e4 d3f1 f6d7 e5f3 d7f6');
INSERT INTO evaluations VALUES(37067,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(37067,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37067,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37067,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37067,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37067,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37067,36,'cp 0','c7c5 f1e2 d5c4 e2c4 d8c7 c4e2 b7b6 c1b2 c8b7 e1g1 f8d8 c3c4 b8c6 a1c1 a8c8 f1e1 c6a5 f3e5 c7e7 h2h3 h7h6 e2d3 b7e4 d3f1 e4h7 a3a4 f6d7 e5f3 d7f6');
INSERT INTO evaluations VALUES(36157,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36157,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36157,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36157,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36157,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36157,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36157,36,'cp 0','f1e2 b8c6 e1g1 c6a5 c4d5 e6d5 d4c5 f6e4 f3d4 e4c5 a3a4 f8e8 c1a3 c5e4 a3b4 c8d7 f2f3 e4f6 d1d2 a8c8 f1b1 a5c4 e2c4 c8c4 a4a5 d8b8 b1e1 b8c8 d4e2 c8c7 e2d4');
INSERT INTO evaluations VALUES(44488,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(44488,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(44488,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(44488,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(44488,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(44488,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(44488,36,'cp 110','d2d4 g7g6 f1e2 d7d6 f2f4 f8g7 g1f3 c6c5 d4d5 e8g8 e1g1 e7e6 d5e6 f7e6 e4e5 d6e5 d1d8 f8d8 f4e5 f6g4 c3e4 b8c6 e4c5 g4e5 f3e5 c6e5 c1g5 d8d5 c5e4 c8d7 e4f6 g7f6 g5f6 a8c8 a1d1 e5c4 f6d4 e6e5 d4c3 d5d1 f1d1 d7e6');
INSERT INTO evaluations VALUES(44485,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(44485,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(44485,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(44485,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(44485,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(44485,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(44485,36,'cp 113','g7g6 f1e2 f8g7 f2f4 d7d6 g1f3 c6c5 d4d5 e8g8 e1g1 e7e6 d5e6 c8e6 f3g5 b8c6 c3b5 d6d5 e4e5 f6e4 g5e4 d5e4 c2c3 d8b6 c1e3 f8d8 d1c2 g7f8 c2e4 c6e7 g2g4 e6d5 e4c2 d5c6 c3c4 c6b5 c4b5');
INSERT INTO evaluations VALUES(33899,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(33899,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(33899,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(33899,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(33899,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(33899,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(33899,36,'cp 100','h7h6 f1d3 b8c6 c2c3 f8d6 d1e2 c8d7 e1g1 c6e7 f3e5 d6e5 d4e5 f6h4 b2b4 a7a6 a2a4 e8g8 c1e3 e7d5 g2g3 h4h3 e3d4 f8d8 d3e4 c7c5 e4d5 e6d5 b4c5 d8e8 f1b1 d7c6 f2f4 h3f5 e2g2 h6h5 b1f1 e8e7 h2h3 c6d7 g3g4 h5g4 h3g4 f5g4');
INSERT INTO evaluations VALUES(33863,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(33863,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(33863,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(33863,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(33863,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(33863,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(33863,36,'cp 102','f1d3 b8c6 c2c3 f8d6 d1e2 a7a5 e1g1 e8g8 f1e1 c6e7 f3e5 e7g6 e2h5 d6e5 d4e5 f6e7 d3g6 f7g6 h5g6 e7f7 g6f7 f8f7 c1e3 c8d7 c3c4 d7c6 a1d1 f7f5 e3d2 a5a4 f2f3 g7g5 d2c1 h6h5 d1d4 a8a5 f3f4 g5f4 c1f4 a4a3 b2b3 g8f7 e1d1');
INSERT INTO evaluations VALUES(33862,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(33862,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(33862,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(33862,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(33862,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(33862,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(33862,36,'cp 90','f8d6 d1e2 b8c6 c2c3 a7a5 e1g1 e8g8 f1e1 c6e7 g2g3 e7g6 h2h4 f8d8 d3e4 a8b8 h4h5 g6f8 c3c4 b7b6 c1d2 f8d7 a2a3 f6e7 e4c2 d7f6 a1d1 c8b7 d2c3 d8e8 e2d3 b8d8 d4d5 b6b5 c3f6 e7f6 d3h7 g8f8 d5e6 e8e6 f3h4 b5c4');
INSERT INTO evaluations VALUES(33406,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(33406,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(33406,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(33406,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(33406,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(33406,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(33406,36,'cp 90','d1e2 a7a5 e1g1 b8c6 c2c3 e8g8 f1e1 c6e7 f3e5 e7g6 e2h5 d6e5 d4e5 f6e7 d3g6 f7g6 h5g6 e7f7 g6f7 f8f7 e1d1 c8d7 c1e3 d7c6 d1d4 g8h7 c3c4 g7g5 a1e1 f7f5 e3d2 f5f8 f2f3 h7g6 d2c3 a5a4 e1d1 f8f5 d4d8 f5f8 d8a8 f8a8 g1f2 h6h5 b2b4 a4b3 a2b3');
INSERT INTO evaluations VALUES(45159,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(45159,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(45159,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45159,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(45159,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45159,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(45159,36,'cp 17','g1e2 e7e5 g2g3 a7a6 f1g2 b8c6 e1g1 a8b8 a2a4 g7g6 d2d3 f8g7 h2h3 e8g8 c1e3 c8d7 a1b1 f6e8 f2f4 e8c7 f4f5 b7b5 b2b3 c6d4 g3g4 h7h5 a4b5 a6b5 c4b5 c7b5 c3b5 d7b5 g4h5 g6f5 e4f5 d8h4 e2d4 c5d4');
INSERT INTO evaluations VALUES(51966,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(51966,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(51966,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(51966,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(51966,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(51966,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(51966,36,'cp 19','d7d5 c2c4 g8f6 c4d5 e6d5 f1b5 b8d7 d4c5 a7a6 b5d7 c8d7 b1c3 f8c5 e1g1 e8g8 b2b3 f8e8 c1b2 d7c6 f3d4 c6d7 h2h3 a8c8 d4f3 d7c6 a1c1 h7h6 f3d4 c6d7 d4e2 d7f5 e2g3 f5g6 c3e2 f6e4 g3e4 g6e4');
INSERT INTO evaluations VALUES(18937,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(18937,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(18937,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(18937,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(18937,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(18937,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(18937,36,'cp 7','f1d3 g8f6 g1f3 f8c5 e1g1 d7d6 c2c3 a7a6 d1e2 e8g8 c1e3 c5e3 e2e3 f6g4 e3e2 c8d7 b1d2 d8c7 f1d1 a8d8 h2h3 g4e5 f3e5 c6e5 d2f3 d7c6 f3e5 d6e5 e2e3 d8d7 d3e2 f8d8 d1d7 d8d7 a1d1 f7f6 d1d7 c6d7 e3a7 d7c6 f2f3');
INSERT INTO evaluations VALUES(25520,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(25520,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(25520,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(25520,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(25520,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(25520,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(25520,36,'cp 66','c2c3 a7a6 b5c4 b7b5 c4f1 c6e7 d1b3 e7g6 f3g5 d8e7 d2d4 h7h6 g5f3 e7d8 c3c4 c7c6 a2a4 b5a4 b3c2 f8e7 c4c5 e8g8 d4e5 g6e5 f3e5 d6e5 a1a4 a6a5 h2h3 d7e6 c1e3 f6d7 e1d1 e7g5 e3g5 d8g5 d1d6 e6h3 d6c6 h3e6 c6d6 f8c8 c5c6 d7b6');
INSERT INTO evaluations VALUES(32187,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(32187,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(32187,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(32187,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(32187,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(32187,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(32187,36,'cp 141','c1f4 b8c6 f1e2 g8f6 b1d2 f8e7 c2c3 e8g8 e1g1 g4h5 d1b3 d8d7 e2d3 h5g6 f3h4 c6a5 b3a3 a5c6 h4g6 h7g6 a3a4 d6d5 a1e1 f6e4 d2e4 d5e4 d3e4 e7d6 f4g5 a8e8 a4b5 d7g4 e4f3 g4f5 a2a4 e8e1 f1e1 c6d8 g5e7 f5b5 a4b5');
INSERT INTO evaluations VALUES(32185,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(32185,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(32185,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(32185,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(32185,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(32185,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(32185,36,'cp 117','b8c6 b1c3 g8f6 d1d3 f8e7 e1c1 d6d5 h2h3 g4e6 d3e3 e8g8 e4e5 f6e4 c3e4 d5e4 e3e4 e6d5 e4e1 d5f3 g2f3 d8d5 e1c3 c6b4 c3b3 b4a2 c1b1 d5b3 c2b3 a2b4 d4d5 a8d8 d5d6 e7h4 f1c4 c7d6 e5d6 d8d7 c4b5 b4c6 d1c1 f8c8 c1c4 c8d8 h1d1 h4f2 b5a4 a7a6 b3b4 f2b6 a4b3 a6a5 b4b5');
INSERT INTO evaluations VALUES(27039,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(27039,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(27039,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(27039,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(27039,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(27039,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(27039,36,'cp 121','b1c3 g8f6 d1d3 d6d5 e1c1 f8e7 h2h3 g4e6 d3e3 e8g8 e4e5 f6e4 c3e4 d5e4 e3e4 e6d5 e4e1 d8d7 a2a3 f8b8 f1d3 b7b5 f3g5 h7h6 g5h7 b5b4 h7f6 e7f6 e5f6 b4a3 b2a3 d5a2 e1e4 g7g6 f4e5 a7a5 c1d2 a5a4 h1f1 a8a5 d1e1 c6e5 d4e5 d7e6');
INSERT INTO evaluations VALUES(39334,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(39334,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(39334,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(39334,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(39334,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(39334,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(39334,36,'cp 23','b4c3 d2c3 d7d6 f1d3 a7a6 c1e3 g8f6 d1e2 d8e7 h2h3 b8d7 e1c1 b7b5 g2g4 d7c5 e3c5 d6c5 c3c4 b5b4 e2e3 h7h6 d3e2 c8b7 f3d2 a6a5 h3h4 f6d7 g4g5 e8c8 a2a4 b4a3 b2a3 a5a4 e2g4 c8b8 g4d7 d8d7 e3g3');
INSERT INTO evaluations VALUES(39390,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(39390,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39390,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39390,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(39390,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(39390,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(39390,36,'cp 33','d2c3 d7d6 f1d3 g8f6 c1e3 a7a6 d1e2 d8e7 h2h3 b8d7 e1c1 b7b5 f3d2 d7c5 e3c5 d6c5 c3c4 b5b4 d2f1 g7g6 g2g3 a6a5 f1e3 c7c6 f2f4 a5a4 e2f3 c8d7 e3g4 f6g4 h3g4 e8c8 b2b3 e5f4 g3f4');
INSERT INTO evaluations VALUES(39389,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(39389,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(39389,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(39389,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(39389,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(39389,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(39389,36,'cp 22','d7d6 f1d3 g8f6 c1e3 a7a6 d1e2 d8e7 h2h3 b8d7 f3d2 d7c5 e3c5 d6c5 a2a4 c8e6 a4a5 e8g8 d2f1 f8d8 f1e3 g7g6 g2g4 b7b5 a5b6 c7b6 f2f3 b6b5 c3c4 b5c4 d3c4 a6a5 e1g1 h7h5 c4b5 e6d7 b5d7 d8d7 e3c4 h5g4');
INSERT INTO evaluations VALUES(40842,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(40842,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(40842,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(40842,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(40842,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(40842,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(40842,36,'cp 34','e8g8 e1g1 f8e8 h2h3 c7c6 f1e1 e7f8 a2a4 b8d7 a4a5 h7h6 d3f1 d7e5 c1f4 c8d7 f4h2 b7b5 a5b6 a7b6 a1a8 d8a8 f2f4 e5g6 e4e5 d6e5 f4e5 f6h7 e5e6 d7e6 d4e6 e8e6 e1e6 f7e6 f1d3 f8c5 g1h1 h7f8 c3e4 a8a2 e4c5 b6c5 b2b3 a2a5 d1f3 a5b4 h2g1 g6f4');
INSERT INTO evaluations VALUES(36266,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(36266,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(36266,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(36266,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36266,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(36266,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(36266,36,'cp 33','e1g1 f8e8 h2h3 a7a6 f1e1 h7h6 c1f4 c8d7 d4f3 b8c6 a2a4 e7f8 f4g3 c6e5 a4a5 e5d3 d1d3 f6h5 g3h2 h5f6 a1d1 d7e6 f3d4 c7c6 e4e5 d6e5 e1e5 e6d5 e5e8 d8e8 c3d5 f6d5 d4f3 e8e7 h2g3 a8e8');
INSERT INTO evaluations VALUES(36265,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(36265,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(36265,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(36265,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(36265,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(36265,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(36265,36,'cp 37','f8e8 c1f4 c8d7 f1e1 b8c6 d3c4 a7a6 a2a4 e7f8 a4a5 c6d4 d1d4 d7e6 c4f1 c7c6 d4b4 d8e7 f4g5 a8b8 a1d1 h7h6 g5f4 b8d8 f4e3 d6d5 b4e7 f8e7 e3b6 d8d7 e4d5 f6d5 c3d5');
INSERT INTO evaluations VALUES(47317,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(47317,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(47317,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(47317,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(47317,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(47317,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(47317,36,'cp -95','c8f5 g5f6 e7f6 d1e2 b8c6 e1c1 f8e7 c3e4 e8g8 d4d5 c6e5 g1f3 f5g4 h2h3 g4f3 g2f3 f6f5 e4g3 e7d6 g3f5 d8g5 f5e3 g5f4 f1g2 a8e8 c1b1 e5g6 c2c3 d6c5 h1e1 g6h4 a2a3 g7g6 b1a2 g8g7 e2d3 h4f3 e3f5 f4f5 d3f3');
INSERT INTO evaluations VALUES(29726,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(29726,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(29726,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(29726,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(29726,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(29726,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(29726,36,'cp -63','g5f6 e7f6 d1e2 b8c6 e1c1 f8d6 f2f3 e8g8 f3e4 f8e8 g2g4 f5g6 f1g2 d6f4 c1b1 c6d4 e2c4 c7c5 h2h4 h7h6 h4h5 g6h7 c4c5 d4e6 d1d8 e6c5 d8a8 e8a8 g1e2 f4e5 h1d1 a8c8 b1c1 g8f8 a2a3 a7a6 d1d5 b7b6 d5d1 f8e8 e2d4 c8d8 d4e2 d8d1 c1d1');
INSERT INTO evaluations VALUES(50909,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(50909,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(50909,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(50909,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(50909,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(50909,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(50909,36,'cp 45','c7c5 d2d3 h7h6 c3d5 g8f6 e1g1 f8e7 a2a3 e8g8 b2b4 b8c6 c2c3 b7b5 c4a2 c8e6 c1e3 f6g4 e3d2 g4f6 h2h3 f8e8 d5f6 e7f6 a2e6 e8e6 d1b3 e6e8 a3a4 c5b4 c3b4 d8b6 f1c1 e8c8 d2e3 b6b7 a4b5 b7b5');
INSERT INTO evaluations VALUES(50714,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(50714,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(50714,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(50714,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(50714,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(50714,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(50714,36,'cp 65','d2d3 h7h6 a2a4 g8f6 f3d2 b8c6 c4d5 f6d5 c3d5 c8e6 d2c4 f8e7 a4a5 e7g5 e1g1 g5c1 d1c1 e8g8 f2f4 e5f4 c1f4 e6d5 e4d5 c6b4 c2c3 b4d5 f4g3 a8c8 g3f3 d5f6 f3b7 d6d5 c4b6 d8c7 b7c7 c8c7 b2b4 c5b4 c3b4 f8e8 b4b5 a6b5 f1b1 e8e2 b1b5');
INSERT INTO evaluations VALUES(55724,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(55724,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55724,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(55724,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(55724,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(55724,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(55724,36,'cp 33','b8d7 g1f3 g8f6 h2h3 b7b6 a2a4 f8e7 f1d3 c8b7 b1d2 e8g8 e1g1 a7a6 c2c3 c7c5 d1b1 d6d5 f3e5 a8c8 d3c2 d8e8 b1d1 e7d6 d1e2 e8e7 a1d1 f6e4 e5d7 e7d7 f4d6 d7d6');
INSERT INTO evaluations VALUES(41099,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(41099,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(41099,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(41099,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(41099,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(41099,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(41099,36,'cp 26','h2h3 e8g8 e1g1 h7h6 c1f4 c7c5 d4c5 e7c5 c2c3 f8e8 b1d2 b8c6 f1e1 c8e6 d2b3 c5b6 f4e3 e6d7 e3c5 d8c7 d1c2 a7a5 c5b6 c7b6 b3d4 a5a4 a1b1 c6d4 f3d4 b6d6 e1e8 a8e8 b1d1 f6e4 c2c1 d6b6 d1e1 b6c7 a2a3 c7b6');
INSERT INTO evaluations VALUES(41096,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(41096,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41096,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(41096,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(41096,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(41096,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(41096,36,'cp 22','e8g8 e1g1 h7h6 c1f4 c7c5 d4c5 e7c5 b1d2 f8e8 c2c3 b8c6 f1e1 c8e6 d2b3 c5b6 f4e3 e6d7 e3c5 e8e1 d1e1 b6c5 b3c5 d8b6 b2b4 b6c7 a1d1 a8e8 e1d2 c6e5 f3e5 e8e5 d1e1 e5e1 d2e1 d7c6 c5b3 c6d7 e1d2 a7a6 b3d4 c7b6 a2a3 b6c7');
INSERT INTO evaluations VALUES(36413,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(36413,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36413,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36413,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36413,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(36413,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(36413,36,'cp 30','e1g1 c7c5 d4c5 e7c5 b1d2 b8c6 f1e1 h7h6 d2b3 c5b6 c2c3 c8d7 c1e3 f8e8 e3c5 e8e1 d1e1 d8c7 e1d2 f6e4 c5b6 c7b6 d2c2 e4f6 a1e1 a8e8 e1e8 d7e8 b3d4 c6d4 f3d4 g7g6 a2a3 g8g7 f2f3 e8c6 c2e2 b6c7 e2e3 c7b6');
INSERT INTO evaluations VALUES(36412,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36412,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(36412,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36412,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(36412,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36412,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(36412,36,'cp 25','c7c5 d4c5 e7c5 b1d2 b8c6 f1e1 c5b6 c2c3 h7h6 d2b3 f8e8 e1e8 d8e8 b3d4 c8d7 c1e3 e8e7 d1e2 a8e8 a1e1 e7d8 e2d2 f6e4 d2d1 e4f6 e1e2 a7a6 d4c6 d7c6 e3b6 d8b6 d1d2 e8e2 d2e2 f6e4 e2c2 e4c5 b2b3 c5d3 c2d3');
INSERT INTO evaluations VALUES(44835,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(44835,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(44835,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(44835,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44835,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(44835,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(44835,36,'cp 69','a2a4 a7a5 b1c3 b8c6 g1f3 g7g6 d4d5 c6b4 f1e2 f8g7 c1e3 e8g8 e1g1 b6d7 a1c1 b7b6 f3d4 d7c5 d4c6 d8e8 c3b5 b4c6 d5c6 e8c6 e2f3 c5e4 d1c2 c8b7 f1d1 a8d8 c2e4 c6e4 f3e4 b7e4 e3b6');
INSERT INTO evaluations VALUES(54031,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(54031,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54031,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54031,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54031,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54031,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54031,36,'cp -10','d2d4 g7g6 f1c4 f8g7 g1f3 d8c7 c4b3 g8f6 c2c4 e8g8 e1g1 b8c6 b1c3 c6a5 c1f4 a5c4 b3c4 f6h5 c4e6 h5f4 e6c8 a8c8 g2g3 f4e6 d1b3 e6d4 f3d4 g7d4 c3b5 c7b6 b5d4 b6d4 a1d1 d4c5 b3b7 c8b8 b7a6 b8b2 d1d6 c5c2 d6d7 b2a2 a6b7 c2c3');
INSERT INTO evaluations VALUES(54022,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(54022,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54022,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(54022,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(54022,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(54022,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(54022,36,'cp 0','g7g6 b1c3 g8f6 d4d5 e6d5 f1b5 b8c6 c3d5 f8g7 d1e2 e8f8 d5f6 d8f6 c2c3 a7a6 b5d3 c6e5 c1e3 c8e6 g1f3 e5f3 e2f3 f6f3 g2f3 f8e7 h2h4 h7h5 a2a4 a8c8 a4a5 e6d5 d3e4 d5c6 e1c1 e7e6 e4c6 c8c6 h1e1 e6d7');
INSERT INTO evaluations VALUES(20319,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(20319,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(20319,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(20319,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(20319,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(20319,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(20319,36,'cp -67','g8h6 g1f3 d8b6 f1d3 c5d4 c3d4 c8d7 d3c2 c6b4 b1a3 b6a6 c2b1 a8c8 e1f2 f8e7 h1e1 h6f5 d1e2 a6e2 f2e2 f5h4 f3h4 e7h4 g2g3 h4e7 c1d2 d7a4 e1c1 e8d7 d2c3 a7a5 a3c2 b4c6 e2f2 h7h5 h2h4 f7f6 f2g2 f6e5 d4e5');
INSERT INTO evaluations VALUES(13453,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(13453,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(13453,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(13453,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(13453,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(13453,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(13453,36,'cp -70','g1f3 d8b6 f1d3 c5d4 c3d4 c8d7 d3c2 c6b4 b1a3 b6a6 c2b1 a8c8 e1f2 f8e7 h1e1 d7a4 d1e2 h6g4 f2f1 a6e2 f1e2 h7h5 h2h3 g4h6 c1d2 a7a5 g2g3 g7g6 e2f2 e8d7 b2b3 a4c6 e1c1 c8a8 c1c3 h8c8 f3g5 a5a4 g5f3 a4b3 a2b3 c8c7 c3c1 h6f5 b1f5 e6f5');
INSERT INTO evaluations VALUES(13452,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(13452,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(13452,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(13452,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(13452,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(13452,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(13452,36,'cp -47','d8b6 b1a3 c5d4 c3d4 f8a3 b2a3 h6f5 e1f2 c6d4 f3d4 f5d4 c1e3 b6b2 d1d2 b2d2 e3d2 f7f6 e5f6 g7f6 d2c3 d4c6 a1d1 e8g8 f1e2 c8d7 h2h4 a8c8 h1e1 b7b6 g2g4 f8f7 d1c1 d5d4 c3d2 e6e5 e2a6 c8c7 f4e5 f6e5 f2g3');
INSERT INTO evaluations VALUES(817,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(817,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(817,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(817,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(817,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(817,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(817,36,'cp -45','b1a3 c5d4 c3d4 f8a3 b2a3 h6f5 e1f2 c8d7 c1d2 c6d4 a1b1 b6c7 f3d4 f5d4 d2b4 d4f5 d1c1 c7c1 b1c1 d7c6 g2g4 f5e7 f1e2 a8d8 h1d1 g7g6 b4c5 h7h5 g4h5 a7a6 d1g1 d5d4 g1d1 d4d3 d1d3 d8d3 e2d3 h8h5 f2g3 g6g5 c5e7 e8e7 f4f5 e6f5 d3f5 h5h4 f5g4 c6d7 h2h3 d7g4 h3g4 e7d7 c1d1 d7e6 d1d6 e6e5');
INSERT INTO evaluations VALUES(20965,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(20965,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(20965,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(20965,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(20965,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(20965,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(20965,36,'cp -41','f7f6 g5c1 e7e5 c2c4 e5e4 f3d2 f6f5 e2e3 g8f6 b1c3 c8e6 c4d5 f6d5 f1c4 d8d7 c3d5 e6d5 c4d5 d7d5 d1b3 d5b3 d2b3 b7b6 c1d2 a7a5 a2a4 e8c8 e1e2 h7h5 d2c3 h8h6 b3d2 h6e6 h2h4 g7g6 a1c1 c6e7 b2b4 e7d5 b4a5');
INSERT INTO evaluations VALUES(20829,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(20829,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(20829,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(20829,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(20829,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(20829,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(20829,36,'cp -36','g5c1 e7e5 c2c4 e5e4 f3d2 f6f5 b1c3 g8f6 e2e3 c8e6 c4d5 f6d5 f1c4 d8d7 c4d5 e6d5 c3d5 d7d5 d1b3 d5b3 d2b3 b7b6 c1d2 e8d7 e1e2 f8d6 d2c3 c6e7 f2f3 e4f3 g2f3 e7d5 e2d3 g7g5 c3d2 a8d8 e3e4 d5f4 d2f4 d6f4 a1f1 f5e4 f3e4');
INSERT INTO evaluations VALUES(52786,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52786,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(52786,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52786,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52786,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52786,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52786,36,'cp 0','d5e4 d3e4 g8f6 b1c3 f6e4 c3e4 d8d5 d1e2 c8f5 e4g3 d5g2 g3f5 g2h1 f5d6 e8d8 d6f7 d8e8 f7d6 e8d8');
INSERT INTO evaluations VALUES(52901,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(52901,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52901,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52901,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52901,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52901,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52901,36,'cp 0','d3e4 g8f6 b1c3 f6e4 c3e4 d8d5 d1e2 c8f5 e4g3 d5g2 g3f5 g2h1 f5d6 e8d8 d6f7 d8e8 e2g4 e8f7 g4f5 f7e8 f5c8 e8f7');
INSERT INTO evaluations VALUES(52895,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52895,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52895,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52895,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52895,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(52895,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(52895,36,'cp -6','g8f6 e4f3 b8d7 c1f4 d8b6 b2b3 b6a5 d1d2 a5d2 b1d2 f6d5 f3d5 c6d5 g1e2 e7e6 c2c4 d5c4 b3c4 b7b6 e2c3 f8e7 e1g1 e8g8 a2a4 c8b7 f1b1 a7a5 c3b5 e7b4 d2f1 b7c6 f4d2 b4d2 f1d2 f8d8 a1a3 d7f6 d2f3 f6e4 b1b3 a8c8 f3e5 f7f6 e5c6 c8c6');
INSERT INTO evaluations VALUES(45035,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(45035,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(45035,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(45035,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(45035,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45035,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(45035,36,'cp 0','b1c3 f6e4 c3e4 d8d5 d1e2 c8f5 e4g3 d5g2 g3f5 g2h1 f5d6 e8d7 e2g4 f7f5 d6f5 e7e6 f5g7 h1d5 c2c4 f8b4 c1d2 h7h5 g4g6 b4d2 e1d2 d5d4 d2e1 d7c8 g1e2 d4d7 a1d1 d7e7 g7e6 b8d7 e1f1 e7f6 g6e4 f6e5 e4g6');
INSERT INTO evaluations VALUES(45034,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45034,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45034,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45034,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45034,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45034,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45034,36,'cp 0','f6e4 c3e4 d8d5 d1e2 c8f5 e4g3 d5g2 g3f5 g2h1 f5d6 e8d7 e2g4 f7f5 d6f5 e7e6 f5g7 h1d5 c2c4 d5c4 g1f3 f8d6 g7f5 d6b4 c1d2 d7c7 g4g3 c7c8 f5e3 c4b5 e1c1 b8d7 d2b4 b5b4 a2a3 b4f8 f3g5 h8g8 e3c4 h7h6 c4d6 c8b8 d6f7 b8c8');
INSERT INTO evaluations VALUES(45081,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45081,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45081,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45081,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45081,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45081,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45081,36,'cp 0','c3e4 d8d5 d1e2 c8f5 e4g3 d5g2 g3f5 g2h1 f5d6 e8d7 e2g4 f7f5 d6f5 e7e6 f5g7 h1d5 c2c4 d5c4 g1f3 f8d6 g7f5 d6b4 c1d2 d7c7 g4g3 c7c8 f5e3 c4b5 e1c1 b8d7 d2b4 b5b4 a2a3 b4f8 f3g5 h8g8 e3c4 h7h6 c4d6 c8b8 d6f7 b8c8');
INSERT INTO evaluations VALUES(45078,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45078,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45078,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45078,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45078,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45078,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45078,36,'cp 0','d8d5 d1e2 c8f5 e4g3 d5g2 g3f5 g2h1 f5d6 e8d7 e2g4 f7f5 d6f5 e7e6 f5g7 h1d5 c2c4 d5c4 g1f3 f8d6 g7f5 d6b4 c1d2 d7c7 g4g3 c7c8 f5e3 c4b5 e1c1 b8d7 d2b4 b5b4 a2a3 b4f8 f3g5 h8g8 e3c4 h7h6 c4d6 c8b8 d6f7 b8c8');
INSERT INTO evaluations VALUES(14812,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14812,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14812,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14812,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14812,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(14812,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14812,36,'cp 10','f8d6 c2c3 e8g8 b1d2 d8e7 f3e5 f6d7 d2f3 f7f6 e5c6 b7c6 f4d6 c7d6 c3c4 c8a6 a1c1 f8c8 b2b3 d7b6 f3d2 c6c5 f1e2 c5d4 e3d4 c8c7 e1g1 a8c8 f1e1 g7g6 a2a4 g8g7 c4c5 a6e2 d1e2 d6c5 d4c5 c7c5 c1c5 c8c5 e2e6 e7e6');
INSERT INTO evaluations VALUES(9242,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9242,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9242,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9242,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(9242,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(9242,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9242,36,'cp 21','c2c3 e8g8 b1d2 d8e7 f3e5 f6d7 d2f3 f7f6 e5c6 b7c6 f4d6 c7d6 c3c4 c8a6 c4d5 a6f1 e1f1 c6d5 g2g3 f6f5 f1g2 d7f6 d1e2 e7b7 h1c1 a8c8 b2b3 h7h6 f3e1 a7a5 e1d3 b7a6 c1c8 f8c8 a1c1 c8c1 d3c1 a6e2 c1e2 g8f7 f2f3 f6g8 e2c3 f7e8 c3b5 e8d7 g2f2 g8e7');
INSERT INTO evaluations VALUES(15708,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(15708,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(15708,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(15708,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(15708,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15708,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(15708,36,'cp -15','d1e2 d8d5 d2d3 d5e5 b1c3 c8d7 e2e4 f8d6 e4e5 d6e5 c1d2 e8g8 e1c1 a8e8 d1e1 b7b6 f2f4 e5d6 c3e4 f7f6 e4d6 c7d6 e1e8 f8e8 h1e1 e8e1 d2e1 h7h5 a2a4 c6c5 a4a5 g8f7 b2b4 d7c6 g2g3 f7e6 e1f2 c5b4 a5b6 a7b6 f2b6');
INSERT INTO evaluations VALUES(15709,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(15709,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(15709,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(15709,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(15709,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15709,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(15709,36,'cp -18','d8d5 d2d3 d5e5 b1c3 c8e6 e2e4 e5a5 e1g1 e8c8 a2a3 a5f5 e4f5 e6f5 f1e1 f8c5 c1e3 c5d6 f2f3 h8e8 e1e2 f7f6 a1e1 f5g6 e3d2 e8e2 e1e2 d6e5 f3f4 e5d4 d2e3 d4c3 b2c3 g6h5 e2e1 b7b6 e3f2 d8d7 c3c4 c6c5 a3a4 c8b7 e1a1 b7c6 a1e1');
INSERT INTO evaluations VALUES(1018,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(1018,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(1018,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(1018,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(1018,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(1018,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(1018,36,'cp -12','d2d3 d5e5 b1c3 c8e6 e2e4 e5a5 c1d2 e8c8 e1g1 f8e7 f1e1 h8e8 e4f3 e7f6 h2h3 a5f5 f3f5 e6f5 b2b3 b7b6 g1f1 f5d7 e1e8 d8e8 a1e1 e8e1 f1e1 f6e7 c3e4 f7f6 c2c4 h7h6 d2f4 g7g5 f4d2 f6f5 e4c3 c6c5');
INSERT INTO evaluations VALUES(38221,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(38221,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(38221,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38221,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38221,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38221,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38221,36,'cp 1','g8f6 d2d3 e8g8 e1g1 c8g4 b1d2 b8d7 f1e1 c7c6 d2f1 f8e8 h2h3 g4h5 f1g3 h5g6 f3h4 d8c7 c4b3 d7c5 h4f5 d6d5 f5e7 c7e7 b3c2 a8d8 d1e2 c5e6 f2f3 e7c7 e2f2 c6c5 c2a4 e8f8 c1e3 d5d4 c3d4 e6d4 a1b1');
INSERT INTO evaluations VALUES(40769,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40769,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40769,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(40769,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40769,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(40769,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(40769,36,'cp 22','d2d3 c7c6 e1g1 c8g4 b1d2 e8g8 f1e1 b8d7 h2h3 g4h5 d2f1 f8e8 f1g3 h5g6 f3h4 d8c7 c4b3 a7a5 g3f5 e7f8 a2a4 d7c5 b3a2 c5e6 h4g6 h7g6 f5e3 a8d8 e3g4 f6g4 h3g4 e6c5 c1e3 d6d5 e4d5 c6d5 e3c5 f8c5 d3d4 c5b6 e1e5 e8e5 d4e5 c7e5 d1f3');
INSERT INTO evaluations VALUES(40770,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40770,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(40770,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40770,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(40770,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40770,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40770,36,'cp 12','c7c6 e1g1 e8g8 f1e1 f8e8 d3d4 e5d4 c3d4 d6d5 e4d5 c6d5 c4b3 c8g4 h2h3 g4f3 d1f3 b8c6 c1e3 e7b4 b1c3 h7h6 a1c1 a7a6 e1d1 b4c3 c1c3 f6e4 c3c2 e8e6 d1c1 d8d7 b3a4 a8e8 f3g4 e8e7 a4c6 b7c6 e3f4 a6a5 f4e5 f7f6 e5f4 e4g5 f4g5 f6g5 g4g3');
INSERT INTO evaluations VALUES(50351,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(50351,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(50351,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(50351,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(50351,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(50351,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(50351,36,'cp -65','b8c6 b1c3 e7e5 f1d3 c6d4 f3g3 d7d6 g1e2 h7h5 e2d4 e5d4 c3d5 h5h4 d5f6 d8f6 g3f3 f6f3 g2f3 c7c6 b2b3 g7g6 c1b2 f8g7 c2c3 d4c3 b2c3 g7c3 d2c3 a7a5 f3f4 e8e7 h1g1 h4h3 e1d2 c8d7 d2e3 h8h4 c3c4 a5a4');
INSERT INTO evaluations VALUES(17282,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(17282,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(17282,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(17282,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(17282,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(17282,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(17282,36,'cp -57','b1c3 e7e5 f1d3 c6d4 f3g3 f8d6 g1e2 e8g8 e1g1 f8e8 g3h4 c7c5 e2d4 c5d4 c3d5 f6d5 h4d8 e8d8 e4d5 a7a5 f1e1 b7b6 a2a4 f7f6 d3c4 d6c5 d2d3 c5b4 e1e2 c8a6 b2b3 g7g5 f2f3 g8f7 g2g3 d8e8 c1d2 a6c4 b3c4 b4d2 e2d2');
INSERT INTO evaluations VALUES(54851,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(54851,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54851,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(54851,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(54851,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(54851,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(54851,36,'cp 13','g8f6 b1d2 f8g7 e2e3 e8g8 h2h3 c7c5 c2c3 b7b6 a2a4 c8b7 f1e2 a7a6 d1b3 b8d7 e1g1 a8c8 f1c1 f8e8 c3c4 a6a5 c1d1 e7e6 a1c1 d8e7 f4g3 b7a6 b3a3 c8a8 e2d3 a6b7 g3h2 a8d8 c4d5 b7d5');
INSERT INTO evaluations VALUES(14852,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(14852,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14852,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(14852,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(14852,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(14852,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(14852,36,'cp 21','b1d2 c6b4 d3e2 c7c5 d4c5 a7a5 b2b3 f8c5 c1b2 e8g8 e1g1 d8e7 c2c4 f8d8 a2a3 b4c6 d1c2 h7h6 f1d1 c5d6 d2f1 c8d7 c4d5 f6d5 f1g3 a8c8 c2b1 d6g3 h2g3 d5f6 e2b5 d7e8 d1d8 e7d8 b5f1 d8e7 e3e4 e6e5');
INSERT INTO evaluations VALUES(14853,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(14853,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(14853,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(14853,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(14853,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(14853,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(14853,36,'cp 12','c6b4 d3e2 c7c5 d4c5 a7a5 b2b3 f8c5 e1g1 e8g8 c1b2 d8e7 a2a3 b4c6 c2c4 f8d8 d1c2 c5d6 f1e1 h7h6 e1d1 c8d7 c4d5 e6d5 d2f1 a8c8 c2b1 d7g4 f1g3 e7e6 f3d4 c6d4 e2g4 f6g4 b2d4 d6g3 h2g3 g4f6 b1b2');
INSERT INTO evaluations VALUES(33320,30,'cp 279',NULL);
INSERT INTO evaluations VALUES(33320,31,'cp 295',NULL);
INSERT INTO evaluations VALUES(33320,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(33320,33,'cp 270',NULL);
INSERT INTO evaluations VALUES(33320,34,'cp 278',NULL);
INSERT INTO evaluations VALUES(33320,35,'cp 282',NULL);
INSERT INTO evaluations VALUES(33320,36,'cp 298','d2d4 g8f6 f1c4 e8g8 e1g1 c5b6 f1e1 d7d6 e5f3 c8g4 c1g5 b8d7 a2a4 a7a6 b1c3 c7c6 a4a5 b6c7 h2h3 g4h5 g2g4 h5g6 f3h4 g8h8 h4g6 f7g6 g5h4 h7h6 h4g3 a8e8 c4d3 f6h7 f2f4 g6g5 f4f5 e7f6 d3e2 f6d8 d1d3 h7f6 g1h1 c6c5');
INSERT INTO evaluations VALUES(33319,30,'cp 298',NULL);
INSERT INTO evaluations VALUES(33319,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(33319,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(33319,33,'cp 265',NULL);
INSERT INTO evaluations VALUES(33319,34,'cp 287',NULL);
INSERT INTO evaluations VALUES(33319,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(33319,36,'cp 301','g8f6 f1c4 e8g8 e1g1 c5b6 f1e1 d7d6 e5f3 c8g4 c1g5 b8d7 a2a4 c7c6 a4a5 b6c7 c4d3 f8e8 b1d2 g4h5 b2b4 h7h6 g5h4 e7d8 c2c3 h5g6 d3c2 b7b5 d2f1 a7a6 f3d2 c6c5 b4c5 d6c5 e4e5 c5d4 c3d4 d7e5 d4e5 c7e5 a1b1 g6h5 f2f3 e5c3 e1e8 d8e8');
INSERT INTO evaluations VALUES(42968,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42968,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42968,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(42968,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42968,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42968,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(42968,36,'cp 13','g1f3 h7h6 g5h4 b7b6 e2e3 c8b7 f1e2 e8g8 e1g1 c7c5 b1c3 b8c6 d4d5 e6d5 h4f6 e7f6 c4d5 c6e7 e2c4 d7d6 d1b3 b6b5 c3e4 a8b8 e4f6 g7f6 c4b5 b7d5 b3a4 e7g6 b2b3 d5f3 g2f3 d6d5 g1h1 f6f5 a1d1 d5d4 e3d4 c5d4 d1d4 d8g5 f1d1');
INSERT INTO evaluations VALUES(42967,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(42967,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(42967,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(42967,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(42967,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42967,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42967,36,'cp 0','h7h6 g5h4 b7b6 e2e3 e8g8 f1e2 c7c5 b1c3 b8c6 d4d5 e6d5 h4f6 e7f6 c4d5 c6e7 d1b3 d7d6 c3e4 e7g6 f3d2 f6e7 e4c3 e7f6');
INSERT INTO evaluations VALUES(16141,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(16141,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(16141,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(16141,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(16141,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(16141,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(16141,36,'cp -23','d7d5 e4d5 f6d5 e1g1 c8e6 b1d2 f8d6 f1e1 e8g8 c2c3 a7a6 c4b3 f8e8 d3d4 e5d4 e3d4 c6a5 b3c2 c7c5 d4e5 d6e5 f3e5 d8c7 d1f3 a8d8 f3g3 a5c6 e5c6 c7c6 a2a4 d5b4 c2e4 c6d6 g3d6 d8d6 a1d1 e8d8 c3b4 d6d2 d1d2 d8d2');
INSERT INTO evaluations VALUES(14568,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(14568,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(14568,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(14568,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(14568,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(14568,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(14568,36,'cp -16','e4d5 f6d5 e1g1 c8e6 d1e1 a7a6 b1c3 d8d7 d3d4 d5e3 c4e6 d7e6 f2e3 e5d4 f3d4 c6d4 e3d4 e8c8 e1e6 f7e6 a1d1 e6e5 d4e5 f8c5 g1h1 d8d1 f1d1 h8e8 c3e4 c5a7 d1d5 e8e6 g2g4 c7c6 d5d2 e6e5 d2e2 c8d7 h1g2 e5a5 e2d2 d7e6 a2a3 a5e5 g2f3');
INSERT INTO evaluations VALUES(14561,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(14561,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(14561,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(14561,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14561,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14561,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14561,36,'cp -10','f6d5 e1g1 c8e6 d1e1 d8f6 c4b5 f8d6 c2c4 d5f4 e3f4 f6f4 b5c6 b7c6 e1c3 e8g8 c4c5 d6e7 b1d2 f8d8 a1e1 d8d5 b2b4 a7a5 a2a3 a5b4 a3b4 e7f6 d2e4 d5d7 e4f6 g7f6 e1e4 f4f5 h2h3 a8a4 f3h4 f5h5 g2g3');
INSERT INTO evaluations VALUES(14593,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(14593,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(14593,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(14593,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(14593,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14593,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(14593,36,'cp -21','e1g1 c8e6 d1e1 d8f6 c4b5 f8d6 c2c4 d5f4 e3f4 f6f4 e1c3 e8g8 b5c6 b7c6 c4c5 d6e7 c3e5 f4b4 b1c3 b4c5 e5c5 e7c5 f3e5 a8d8 a1d1 e6d7 c3a4 c5b6 f1e1 f7f6 e5d7 d8d7 e1e6 c6c5 e6c6 f8b8 g2g3 d7d6 c6d6 c7d6 a4b6 b8b6 b2b3 g8f7');
INSERT INTO evaluations VALUES(27979,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(27979,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(27979,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(27979,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(27979,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(27979,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(27979,36,'cp 22','f7f5 g1e2 f5e4 d3c2 g8f6 c1g5 e8g8 e1g1 b4e7 a2a3 b8c6 g5f6 e7f6 c3e4 g7g6 e4f6 d8f6 c2e4 a8b8 d1d3 b7a6 e4c6 d7c6 d3c2 e6e5 c2a4 a6b7 c4c5 b8e8 d4e5 e8e5 e2g3 a7a5 g3e4 f6f5 a4c4 f5e6 c4e6 e5e6');
INSERT INTO evaluations VALUES(28001,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(28001,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(28001,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(28001,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(28001,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(28001,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(28001,36,'cp 37','g1e2 f5e4 d3c2 g8f6 c1g5 b4e7 g5f6 e7f6 e1g1 e8g8 c2e4 b7e4 c3e4 d8e8 e4f6 f8f6 e2c3 e8g6 d1d2 b8c6 a1e1 f6f3 e1e3 f3f4 c3e2 f4g4 g2g3 g4e4 b2b3 a8f8 d2d3 e4e3 d3g6 h7g6 f2e3 f8f1 g1f1 c6b4 f1e1 d7d6 e2f4');
INSERT INTO evaluations VALUES(31662,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(31662,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(31662,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(31662,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(31662,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(31662,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31662,36,'cp 38','c1e3 c5c4 c3e2 b8c6 c2c3 h7h5 h2h4 g8h6 e2g3 f5g6 f1e2 h6g4 e3f4 f8e7 f3g5 d8b6 d1c1 c6e5 f4e5 f7f6 g5e6 b6e6 e5f4 e7d6 e1g1 e8g8 f1e1 e6d7 e2f3 d6f4 c1f4 a8e8 a2a4 a7a5 f4d2 g8h7 e1e2 b7b5 a4b5 d7b5 f3g4 h5g4 d2d1 e8e2 d1e2 f8e8 e2g4 b5b2');
INSERT INTO evaluations VALUES(31661,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(31661,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31661,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(31661,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31661,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(31661,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(31661,36,'cp 43','c5c4 c3e2 b8c6 c2c3 g8e7 e3d2 f5g4 h2h4 h7h6 h4h5 e7f5 e2f4 f8e7 g2g3 d8b6 a1b1 b6a6 f1h3 g4h3 h1h3 e7g5 g3g4 g5f4 d2f4 f5e7 a2a3 e8c8 e1f1 d8g8 f1g2 c8b8 d1e2 b8a8 f3e1 g7g6 f4e3 a6a4 h5g6 e7g6');
INSERT INTO evaluations VALUES(54369,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54369,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(54369,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54369,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54369,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(54369,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(54369,36,'cp 25','b8c6 c1f4 f8d6 f4d6 d8d6 e2e3 g8f6 f1d3 e8g8 e1g1 c6b4 d3b1 b4c6 c3b5 d6b8 b1d3 c8d7 b5c3 b8d6 a1c1 a8c8 d1e2 a7a6 h2h3 h7h6 f1d1 f8d8 a2a3 d7e8 f3d2 c6e7 d2b3 f6d7 c3a4 c8c1 d1c1 e6e5 d4e5 d7e5');
INSERT INTO evaluations VALUES(20389,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(20389,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(20389,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(20389,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20389,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(20389,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20389,36,'cp 25','c1f4 f8d6 f4g3 g8e7 e2e3 e7f5 f1d3 f5g3 h2g3 c8d7 a1c1 a8c8 a2a3 d6e7 c1c2 g7g6 e1g1 e8g8 b2b4 e7f6 d1d2 h7h5 f1c1 f6g7 c3e2 d8b6 c2c5 a7a6 b4b5 c6a5 b5a6 a5b3 d2b2 b7a6 e2f4 c8c5 c1c5 f8c8 c5c8 d7c8');
INSERT INTO evaluations VALUES(20381,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(20381,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(20381,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(20381,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(20381,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(20381,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(20381,36,'cp 21','f8d6 f4d6 d8d6 e2e3 g8f6 f1d3 c8d7 a1c1 a7a6 h2h3 e8g8 e1g1 a8c8 d1e2 c8c7 f3d2 h7h6 f1d1 f8c8 d2b3 d7e8 e2d2 d6d8 b3c5 f6d7 c3a4 c6b8 b2b4 d7c5 a4c5 b8d7 c5d7 c7c1 d1c1 c8c1 d2c1 e8d7 c1c5 d7e8');
INSERT INTO evaluations VALUES(22664,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(22664,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(22664,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(22664,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(22664,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(22664,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(22664,36,'cp 123','e4d5 c6e7 g1f3 d7d6 b1c3 a7a6 c3e4 c8f5 e4g3 f5d7 c2c4 g8f6 f1d3 e7g6 e1g1 f8e7 b2b3 e8g8 c1b2 f8e8 d1d2 c7c6 d5c6 d7c6 f3d4 d6d5 a1d1 e7c5 d4c6 b7c6 c4d5 c6d5 g3f5 d8b6 d1e1 e8e1 f1e1 a8e8 e1e8 f6e8 h2h3 e8f6');
INSERT INTO evaluations VALUES(22656,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(22656,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(22656,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(22656,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(22656,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(22656,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(22656,36,'cp 85','c6e5 d1e2 d8e7 h2h3 d7d6 b1c3 g7g6 c1e3 g8f6 e1c1 f8g7 f2f4 e5c4 e3d4 c4b6 g2g4 e7e2 g1e2 c8d7 f1g2 e8c8 b2b3 a7a5 a2a4 h7h6 g2f3 d8e8 h1g1 h8g8 c1b2 e8e7 h3h4 f6e8 d4g7 e8g7 e2d4 g7e8 h4h5 g6h5 g4h5 g8g1 d1g1 e8f6 b2c1 c8d8');
INSERT INTO evaluations VALUES(23040,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(23040,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(23040,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(23040,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(23040,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(23040,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(23040,36,'cp 101','d1e2 d8e7 d5d6 e7e6 d6c7 f8d6 b1c3 g8e7 c1g5 f7f6 g5e3 d6c7 g1f3 d7d5 e2b5 e6c6 f3d4 c6b5 d4b5 c7b8 e3c5 e5c6 b5d6 b8d6 c5d6 e8f7 e1c1 c8g4 d1d2 h8d8 h2h3 g4e6 d6c5 d8d7 f1d3 d5d4 c3b5 b7b6 b5d6 f7f8 c5a3 a8d8 d6b5 e6a2');
INSERT INTO evaluations VALUES(23039,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(23039,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(23039,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(23039,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(23039,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(23039,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(23039,36,'cp 86','d8e7 d5d6 e7e6 d6c7 f8d6 b1c3 g8e7 c1g5 f7f6 g5e3 d6c7 g1f3 d7d5 e2b5 e6c6 f3d4 c6b5 d4b5 c7b8 e3c5 e5c6 e1c1 c8g4 f2f3 g4e6 b5d6 b8d6 c5d6 e8f7 d6c5 a7a6 c3a4 h7h5 c5f2 a8d8 a4c5 e6c8 f1e2 g7g5 h2h4 e7g6 g2g3 g5h4 g3h4');
INSERT INTO evaluations VALUES(42869,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(42869,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(42869,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(42869,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(42869,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(42869,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(42869,36,'cp 118','e1g1 d7d6 c3a4 c5b6 a4b6 a7b6 d2d4 e5d4 f3d4 e8g8 f2f3 g4e5 c4e2 e5g6 d4b5 c8d7 b5c3 f7f5 e4f5 f8f5 c1e3 g6f4 e2c4 g8h8 d1d2 d8h4 c4b3 d7c6 a2a4 b8d7 c3b5 c6b5 a4b5 a8f8 g2g3 f4h3 g1g2 h4h5 b3e6 f5f3 f1f3 h5f3 g2h3 d7e5');
INSERT INTO evaluations VALUES(42868,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(42868,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(42868,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(42868,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(42868,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(42868,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(42868,36,'cp 107','d7d6 c3a4 c5b6 d2d4 e5d4 f3d4 e8g8 a4b6 a7b6 f2f3 g4e5 c4e2 e5g6 d4b5 c8d7 b5c3 f7f5 e4f5 d7f5 d1d5 g8h8 d5b7 b8d7 b7d5 f5c2 c1e3 c2f5 d5d2 d7e5 a1d1 d8d7 b2b3 a8e8 e3f2 d7f7 f2g3 f5d7 f3f4 e5g4 e2c4');
INSERT INTO evaluations VALUES(41924,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(41924,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(41924,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(41924,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(41924,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(41924,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(41924,36,'cp 121','c3a4 c5b6 d2d4 e5d4 f3d4 b6d4 d1d4 e8g8 a4c3 b8c6 d4d1 g4f6 f1e1 c8e6 c4d3 h7h6 f2f4 c6b4 d3f1 f8e8 d1d2 c7c5 a2a3 b4c6 d2f2 a7a6 h2h3 b7b5 b2b3 b5b4 a3b4 c5b4 c3a4 a8c8 c1b2 d6d5 e4d5 e6d5 a1d1 e8e1');
INSERT INTO evaluations VALUES(30842,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(30842,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(30842,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(30842,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(30842,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(30842,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(30842,36,'cp 30','c7c5 c2c4 g8f6 b1c3 g7g6 d2d4 c5d4 d1d4 d7d6 e1g1 b8d7 d4h4 a8c8 f1d1 f8g7 c1h6 e8g8 b2b3 a7a6 a1c1 c8c5 h6g7 g8g7 a2a4 c5c7 e2e4 f8e8 h4f4 e7e5 f4e3 d7c5 f3d2 a6a5 f2f3 d8e7 g1h1 f6d7 c1b1');
INSERT INTO evaluations VALUES(30502,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(30502,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(30502,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(30502,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(30502,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(30502,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(30502,36,'cp 36','c2c4 g7g6 d2d4 c5d4 d1d4 g8f6 e1g1 f8g7 b1c3 d7d6 d4h4 h7h6 f1d1 b8d7 c1e3 a8c8 b2b3 d7c5 a1c1 c5e4 c3e4 b7e4 f3e1 e4g2 e1g2 d8d7 h4d4 e8g8 d4d3 f6g4 e3d4 g7d4 d3d4 g4f6 f2f3 f8e8 d4d2 g8g7 g2f4 e7e5 f4d5');
INSERT INTO evaluations VALUES(49858,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49858,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49858,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49858,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49858,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49858,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49858,36,'cp 0','f8c5 f1c4 e4f2 c4f7 e8f7 d1d5 f7g6 d5c5 f2h1 b1c3 h7h6 c5c4 g6h7 c4e4 h7g8 e4d5 g8h7');
INSERT INTO evaluations VALUES(42867,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42867,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42867,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42867,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42867,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42867,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42867,36,'cp 0','f1c4 e4f2 c4f7 e8f7 d1d5 f7g6 d5c5 f2h1 b1c3 h7h6 c5c4 g6h7 c4e4 h7g8 e4d5 g8h7');
INSERT INTO evaluations VALUES(12162,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12162,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12162,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12162,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(12162,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(12162,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12162,36,'cp 13','d5c4 d3c4 b7b5 c4d3 b5b4 c3e4 f6e4 d3e4 c8b7 f3d2 a8c8 a2a3 a7a5 d1e2 f8e7 e1g1 e8g8 a3b4 a5b4 b2b3 d7f6 e4f3 f6d5 c1b2 c6c5 a1d1 c5d4 b2d4 e7f6 d2c4 f6d4 d1d4 d5c3');
INSERT INTO evaluations VALUES(12184,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(12184,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(12184,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12184,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12184,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(12184,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(12184,36,'cp 19','d3c4 b7b5 c4d3 b5b4 c3e4 f6e4 d3e4 c8b7 f3d2 a8c8 e1g1 b7a6 f1e1 f8e7 a2a3 a6b5 a3b4 e7b4 a1a7 c6c5 d1b3 e8g8 e4f3 d8b6 a7a1 b5c6 e1d1 c5d4 e3d4 b4e7 f3c6 c8c6 d4d5 e6d5 b3b6 d7b6 d2b3 e7f6 b3d4 c6c4 c1e3 f8c8 d4f5 g7g6 e3b6 g6f5 d1d5 f6b2');
INSERT INTO evaluations VALUES(12182,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(12182,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12182,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(12182,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12182,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(12182,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(12182,36,'cp 15','b7b5 c4d3 b5b4 c3e4 f6e4 d3e4 c8b7 f3d2 f8e7 b2b3 d7f6 e4f3 e8g8 e1g1 a8c8 c1b2 b7a6 f3e2 a6e2 d1e2 c6c5 f1c1 c5d4 c1c8 d8c8 a1c1 c8b7 b2d4 f8c8 c1c8 b7c8 e2c4 c8b7 h2h3 h7h6 d4b2 g8h8 b2e5 h8g8 c4c2 b7a6 e5f6 e7f6 d2c4');
INSERT INTO evaluations VALUES(11958,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(11958,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(11958,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(11958,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(11958,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(11958,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(11958,36,'cp 29','c4d3 b5b4 c3e4 f6e4 d3e4 c8b7 e1g1 f8e7 f3d2 e8g8 b2b3 d7f6 e4f3 a8c8 a2a3 a7a5 a3b4 e7b4 c1b2 b7a6 f3e2 a6b5 e2c4 d8b6 d1e2 c8b8 f1d1 f8d8 h2h3 h7h6 b2a3 c6c5 a3b2 b4d2 d1d2 f6e4 d2c2 b5c4 b3c4 b6c7 e2d3 c7b7');
INSERT INTO evaluations VALUES(11960,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(11960,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(11960,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(11960,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(11960,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11960,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(11960,36,'cp 13','b5b4 c3e4 f6e4 d3e4 c8b7 f3d2 a8c8 e1g1 b7a6 f1e1 f8e7 a2a3 a6b5 a3b4 e7b4 a1a7 d8b6 a7a1 c6c5 d1b3 e8g8 e4f3 b5c6 e1d1 c5d4 e3d4 b4d2 b3b6 d7b6 f3c6 c8c6 c1d2 b6d5 b2b4 f8b8 b4b5 c6b6 d1b1 h7h6 a1a5 g7g5 h2h3 d5c7 f2f4 c7b5 f4g5 h6g5 d2g5');
INSERT INTO evaluations VALUES(29936,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(29936,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(29936,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(29936,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(29936,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(29936,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(29936,36,'cp 146','d4e5 g4f3 d1f3 d6e5 c1d2 f8b4 e1c1 d8e7 h2h4 b8c6 c3d5 f6d5 e4d5 b4d2 d1d2 c6d4 f3g4 g7g6 g4h3 d4f5 f1d3 f5d6 h4h5 e5e4 d2e2 e7g5 c1b1 g5h5 h3h5 g6h5 d3e4 e8g8 e4f3 a8e8 e2e8 f8e8 b1c1 g8g7 f3h5 e8e4 h5f3 e4f4 c1d2 f4d4 d2c3 d4f4 c3d3 b7b6 b2b3');
INSERT INTO evaluations VALUES(29879,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(29879,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(29879,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(29879,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(29879,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(29879,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(29879,36,'cp 129','g4f3 g2f3 d6e5 d1d8 e8d8 f3f4 f8d6 f4e5 d6e5 c1d2 b8d7 h1g1 c7c6 f1h3 e5h2 g1g7 d8e7 g7g2 h2e5 f2f4 e5c3 d2c3 a8g8 c3d4 c6c5 d4f2 g8g2 h3g2 f6h5 f2e3 h8g8 e1f2 h5f6 g2f3 h7h5 a1h1 f6g4 f3g4 h5g4 f2g3 b7b6 h1h5');
INSERT INTO evaluations VALUES(29881,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(29881,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(29881,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(29881,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(29881,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(29881,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(29881,36,'cp 139','g2f3 d6e5 d1d8 e8d8 f3f4 b8d7 f4e5 d7e5 c1e3 c7c6 f2f4 e5g6 h1g1 d8c7 e4e5 a8e8 f1c4 f8b4 e1c1 b4c3 b2c3 f6d5 c4d5 c6d5 d1d5 g6e7 d5b5 b7b6 e3f2 e7f5 c3c4 e8e6 g1d1 f7f6 c4c5 f6e5 f4e5 h8e8 d1d5 f5e7 d5d3 e7g6 c5b6 a7b6 d3c3 c7b7 c3b3 g6e5 f2b6 b7c6 a2a4');
INSERT INTO evaluations VALUES(39211,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(39211,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(39211,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(39211,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(39211,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(39211,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(39211,36,'cp -54','d7d6 d2d3 b8c6 c1e3 g8e7 d3d4 c5b4 b1c3 e5d4 f3d4 e8g8 f1e2 f7f5 e4f5 c6d4 e3d4 e7f5 e1g1 d8h4 c3b5 f5d4 d1d4 h4d4 b5d4 a7a5 a1d1 c7c6 a2a3 b4c5 d4b3 c5b6 d1d6 a5a4 b3c1 c8f5 d6d2 a8d8 f1d1 d8d2 d1d2 b6a5 d2d1 g7g6');
INSERT INTO evaluations VALUES(38710,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(38710,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(38710,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(38710,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(38710,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(38710,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(38710,36,'cp -25','d2d3 b8c6 c1e3 g8e7 d3d4 e5d4 f3d4 c5b4 b1c3 e8g8 f1e2 f7f5 e4f5 c6d4 d1d4 e7f5 d4d5 f8f7 e1g1 c7c6 d5d2 f5e3 d2e3 d8e7 e3d2 c8f5 a2a3 b4a5 a1d1 a8f8 f1e1 e7h4 g2g3 h4h3 b2b4 a5c7 e2f1 h3h5 f1e2 h5g6 g1h1 g6f6 f2f4 c7b6 g3g4 f5e6 f4f5');
INSERT INTO evaluations VALUES(16286,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(16286,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(16286,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(16286,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(16286,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(16286,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(16286,36,'cp 43','e2e3 h7h6 h2h4 d7d6 c2c4 g7g6 b1c3 f8g7 d1d2 d8e7 a1d1 c8d7 f4h2 e8g8 f1e2 a8d8 e1g1 e6e5 c3d5 f6d5 c4d5 c6b8 d2c2 d7g4 d4e5 d6e5 c2c4 e5e4 f3d4 g4e2 d4e2 b8a6 e2c3 a6c5');
INSERT INTO evaluations VALUES(56351,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(56351,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(56351,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(56351,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(56351,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(56351,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(56351,36,'cp -28','d5d4 g1f3 b8c6 d2d3 f8c5 e1g1 g8e7 e2e3 e8g8 e3d4 c5d4 f3d4 d8d4 b1c3 c8g4 c1e3 d4d7 d1a4 g4h3 g2h3 d7h3 c3d5 h3d7 e3c5 f8d8 d5e7 c6e7 a4d7 d8d7 c5e7 d7e7 a1e1 g8f8 e1e3 e7e8 g1g2 f7f6 f2f4 e5f4 e3e8 a8e8 f1f4 c7c5 f4e4 e8e4 d3e4 f8e7 g2f3 e7e6 f3f4 g7g5 f4e3 e6e5 g3g4 h7h6');
INSERT INTO evaluations VALUES(56546,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(56546,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(56546,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(56546,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(56546,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(56546,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(56546,36,'cp -32','g1f3 b8c6 d2d3 f8c5 e1g1 g8e7 e2e3 c8g4 h2h3 g4h5 e3d4 h5f3 g2f3 c5d4 b1a3 e8g8 a3c2 a7a5 c2d4 d8d4 c1g5 d4d6 g5e7 d6e7 f3c6 b7c6 d1f3 e7d6 a1e1 a5a4 e1e2 f7f6 f1e1 a8b8 h3h4 f8d8 e2d2 c6c5 h4h5 b8b4');
INSERT INTO evaluations VALUES(9149,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(9149,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(9149,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(9149,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(9149,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(9149,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(9149,36,'cp 172','b1c3 c8e6 g5e6 f7e6 c3e4 c5e7 c2c3 d8d7 e1g1 e8g8 d1g4 g8h8 b2b4 f8f5 a2a4 e7d6 a4a5 c6e7 c4b3 c7c6 g2g3 a8f8 g4e2 f5f3 b3d1 h7h6 a5a6 b7b6 e2d2 f3f7 d1h5 f7f5 h5g4 d5f6 d2d1');
INSERT INTO evaluations VALUES(9146,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(9146,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(9146,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(9146,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(9146,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(9146,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(9146,36,'cp 187','c8e6 g5e6 f7e6 d1g4 d8d7 c3d5 e6d5 g4d7 e8d7 c4d5 c6b4 d5b3 a8f8 c1e3 c5e3 f2e3 c7c6 e1e2 b4d5 a1f1 d7d6 f1f3 f8f3 g2f3 g7g6 h1d1 h8e8 b3d5 c6d5 d3d4 b7b5 d4e5 e8e5 d1d4 e5h5 h2h4 a7a5 c2c3 a5a4 e3e4 h5h4 d4d5 d6c6 e2d3');
INSERT INTO evaluations VALUES(9622,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(9622,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(9622,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(9622,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(9622,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(9622,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(9622,36,'cp 91','f1e2 e7e5 e1g1 e8g8 d1c2 e5d4 f3d4 c7c6 c1e3 f8e8 a1d1 a7a6 h2h3 d8e7 e3g5 h7h6 g5h4 d7c5 e2f3 c8e6 b2b3 a6a5 f1e1 e7f8 h4g3 f6d7 d4e6 e8e6 f3g4 f7f5 e4f5 e6e1 d1e1 h6h5 g4f3 f8f5 c2f5 g6f5 c3e2 a5a4 g3d6 a4b3 a2b3 c5b3 f3h5 b3a5 e2g3 a5c4 g3f5 c4d6');
INSERT INTO evaluations VALUES(9621,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(9621,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(9621,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(9621,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(9621,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(9621,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(9621,36,'cp 81','e7e5 e1g1 e8g8 d1c2 e5d4 f3d4 c7c6 c1e3 f8e8 a1d1 a7a6 h2h3 d8e7 e3g5 h7h6 g5h4 d7c5 e2f3 c8e6 b2b3 a6a5 f1e1 e7f8 h4g3 f6d7 f3e2 f7f5 c2c1 f8f6 e4f5 e6f5 d4f5 g6f5 c3a4 c5a4 b3a4 d7c5 g3d6 c5a4 c4c5 a4c3 e2h5 e8e4 d1d2 a5a4 h5f3');
INSERT INTO evaluations VALUES(9551,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(9551,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(9551,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(9551,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(9551,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(9551,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(9551,36,'cp 76','c1e3 e8g8 e1g1 e5d4 f3d4 c7c6 d1c2 f8e8 a1d1 a7a6 h2h3 d8e7 e3g5 h7h6 g5h4 d7c5 e2f3 c5d7 f1e1 c6c5 d4b3 d7e5 f3e2 g6g5 h4g3 c8d7 d1d2 a8d8 e1d1 d7c6 b3c5 d6c5 d2d8 e8d8 d1d8 e7d8 g3e5 d8e7 e5f6 g7f6 c3d5');
INSERT INTO evaluations VALUES(11313,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(11313,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(11313,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11313,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(11313,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(11313,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11313,36,'cp 39','e2e3 c8g4 h2h3 g4f3 d1f3 e7e6 e1c1 f8d6 g2g4 d6f4 f3f4 h7h6 h3h4 d8e7 c1b1 e8c8 f4g3 c8b8 a2a3 f6d7 f1e2 d7b6 f2f4 f7f6 h4h5 d8d6 g3g2 h8d8 b1a2 e6e5 h1g1 e5d4 e3d4 d8e8 e2f3');
INSERT INTO evaluations VALUES(11311,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11311,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(11311,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(11311,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(11311,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(11311,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(11311,36,'cp 25','c8f5 f3e5 c6e5 d4e5 f6e4 c3e4 d5e4 f1e2 e7e6 h2h4 d8d1 a1d1 h7h5 f4g5 f8e7 g5e7 e8e7 f2f4 g7g6 e1f2 h8d8 d1d8 a8d8 h1d1 d8d5 c2c4 d5d1 e2d1 c7c5 g2g4 h5g4 d1e2 e7f8 f2g3 f8g8 e2g4 g8g7 h4h5 f5g4 g3g4 g7h6 h5g6 h6g6 a2a3 b7b5 b2b3 b5c4 b3c4 a6a5 a3a4');
INSERT INTO evaluations VALUES(24263,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24263,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(24263,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(24263,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(24263,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(24263,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24263,36,'cp 37','f1d3 e7e6 d3f5 e6f5 c3e2 f8e7 c2c3 f6e4 d1c2 g7g6 a1c1 d8d7 e1g1 e8g8 f3e1 c6d8 e1d3 d8e6 f4e5 c7c6 b2b3 f7f6 e5g3 f8e8 c3c4 e7a3 c1a1 a3f8 a2a4 a6a5 g3f4 g6g5 f2f3 d5c4 b3c4 g5f4 f3e4');
INSERT INTO evaluations VALUES(24264,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24264,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24264,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(24264,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24264,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(24264,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24264,36,'cp 28','e7e6 d3f5 e6f5 e1g1 f6e4 c3e4 d5e4 f3e5 c6e5 f4e5 h7h5 h2h3 f8d6 e5d6 d8d6 c2c4 g7g6 d1c2 e8g8 f1d1 a8d8 d4d5 d8d7 a1b1 f8d8 b2b4 d6f6 a2a4 h5h4 c2b3 c7c6 d5c6 f6c6 d1d7 c6d7 c4c5 d7d2 b3c4 d8d3 b4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(17259,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(17259,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(17259,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(17259,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(17259,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(17259,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(17259,36,'cp 56','g1f3 e7e6 a2a3 d7d5 c2c4 d5c4 b1d2 b7b5 a3a4 b5a4 f1c4 c8d7 c4b5 f6d5 d2e4 f8d6 d1a4 c6b4 e1g1 a7a5 c1d2 d5b6 b5d7 d8d7 a4d7 e8d7 e4c5 d6c5 d4c5 b6d5 f3e5 d7e7 e3e4 d5f6 a1d1 h8d8 d2b4 d8d1 f1d1 a5b4 f2f3 a8a2 e5c6 e7f8 d1d8 f6e8 b2b3 a2c2 c6b4 c2c1 g1f2 c1c5');
INSERT INTO evaluations VALUES(17260,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(17260,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(17260,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(17260,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(17260,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(17260,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(17260,36,'cp 59','e7e6 c2c4 d7d5 a2a3 d5c4 b1d2 e6e5 d4e5 f6g4 d1a4 c8d7 e5e6 d7e6 f1c4 e6d7 e1g1 f8d6 c4e2 e8g8 a4b3 d8e7 d2c4 c6d4 e3d4 e7e2 c4d6 c7d6 h2h3 g4f6 b3b7 d7f5 f1e1 e2d3 b7b4 f5e4 f3h4 f8e8 b4d6 a8d8 d6g3 d3g3 f2g3');
INSERT INTO evaluations VALUES(16290,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(16290,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(16290,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(16290,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(16290,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(16290,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(16290,36,'cp 59','a2a3 d7d5 b2b4 a7a5 b4b5 c6b8 c2c4 c7c5 c1b2 b8d7 b1d2 f8e7 f1d3 c5d4 b2d4 e8g8 e1g1 b7b6 c4d5 f6d5 d1c2 h7h6 d2c4 c8b7 f1d1 a8c8 c2b2 e7f6 c4d6 c8c7 d6b7 c7b7 a1c1 b7c7 c1c6 f6d4 b2d4 d8f6 c6c7 f6d4 f3d4');
INSERT INTO evaluations VALUES(56317,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(56317,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(56317,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(56317,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(56317,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(56317,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(56317,36,'cp 218','e2e4 g8e7 f2f4 b8c6 b1c3 d5e4 d1d8 c6d8 c3e4 e7d5 c1d2 d5b4 d2b4 f8b4 c2c3 b4e7 e1c1 d8e6 f4f5 e6c5 e4c5 e7c5 f1d3 g7g6 f5f6 c8e6 d3e4 c7c6 c1c2 e8g8 g1f3 a7a5 a2a4 a8e8 h2h3 h7h5 h1e1 c5f2 e1e2 f2g3 d1d4');
INSERT INTO evaluations VALUES(9582,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(9582,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(9582,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(9582,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(9582,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(9582,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(9582,36,'cp 32','c4d5 f6d5 g5e7 d8e7 e2e4 d5c3 b2c3 c7c5 f1d3 c5d4 c3d4 e6e5 d4d5 e7b4 f3d2 d7c5 d1e2 e8g8 e1g1 c8d7 a1b1 b4c3 d3c4 a8c8 d2f3 f7f6 h2h4 h7h6 f1c1 c3a3 c1c2 d7a4 c2b2 a4d7 f3d2 b7b6 h4h5 g8h8 c4b5 d7b5 e2b5');
INSERT INTO evaluations VALUES(9579,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(9579,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(9579,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(9579,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(9579,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(9579,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9579,36,'cp 26','e6d5 e2e3 c7c6 f1d3 e8g8 e1g1 f6e4 g5f4 d7f6 d1c2 e7d6 f4d6 e4d6 b2b4 d8e7 a2a4 a7a5 b4b5 c6b5 c3b5 d6b5 d3b5 c8e6 c2c5 e7c5 d4c5 f8c8 f1c1 f6e4 c5c6 b7c6 c1c6 c8c6 b5c6 a8b8 c6b5 e4d6 f3d4 g7g6 h2h3 b8c8 b5c6');
INSERT INTO evaluations VALUES(9598,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(9598,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9598,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(9598,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(9598,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(9598,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9598,36,'cp 43','e2e3 c7c6 f1d3 d7f8 h2h3 f8g6 d1c2 e8g8 a2a3 f8e8 e1g1 d8d6 f1e1 c8d7 a1d1 a8d8 b2b4 d8c8 c3e2 e7d8 e2g3 d8c7 d3f5 d7f5 g3f5 d6e6 g5f6 e6f6 e3e4 d5e4 e1e4 g6e7 d1e1 f6f5 e4e7');
INSERT INTO evaluations VALUES(9597,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(9597,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(9597,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(9597,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9597,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(9597,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(9597,36,'cp 35','d7f8 f1d3 c7c6 d1c2 f8e6 g5h4 g7g6 a1b1 e8g8 e1g1 f6e8 h4e7 d8e7 b2b4 a7a6 a2a4 e6g5 f3g5 e7g5 b4b5 a6b5 a4b5 c8d7 b5c6 b7c6 c3e2 e8d6 c2c5 g5e7 e2f4 f8b8 b1b8 a8b8 h2h3 g8g7 f1c1 b8a8');
INSERT INTO evaluations VALUES(30776,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30776,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30776,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30776,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30776,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30776,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30776,36,'cp 0','e7e6 g1f3 c7c5 e1g1 g8e7 c2c3 e7g6 f1e1 f8e7 d3d4 d7d5 e4d5 b7d5 c4d5 d8d5 d4c5 d5c5 c1e3 c5c6 a2a4 b8d7 b1d2 e8g8 d1e2 f8d8 g2g3 a8c8 h2h4 h7h6 a4a5 e6e5 h4h5 g6f8 a5b6 a7b6 a1a7');
INSERT INTO evaluations VALUES(30625,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(30625,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(30625,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30625,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(30625,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(30625,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(30625,36,'cp 12','g1f3 f8e7 e1g1 d7d6 c1f4 g8f6 c2c3 e8g8 f1e1 d6d5 e4d5 f6d5 f4g3 c7c5 b1d2 e7d6 d3d4 d6g3 h2g3 c5d4 f3d4 b8c6 d4c6 b7c6 d1e2 d5f6 a1d1 d8c7 d2f3 f8d8 e2e5 c7b7 c4e2 a8c8 d1d8 c8d8 e5f4 h7h6 e1d1 d8d1 e2d1 f6d7');
INSERT INTO evaluations VALUES(30623,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30623,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30623,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(30623,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(30623,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(30623,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30623,36,'cp 0','f8e7 e1g1 d7d6 c2c3 g8f6 d1e2 c7c5 c1f4 e8g8 b1d2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(48292,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48292,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(48292,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48292,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48292,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48292,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(48292,36,'cp 13','c7c5 e2e3 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 c5d5 f1c4 d5d6 b1c1 f8d8 e1g1 d6e7 a2a3 h7h6 d1c2 b7b6 c4b5 c8b7 c2c7 e7c7 c1c7 b7a6 b5c4 f6e8 c4a6 b8a6 c7c1 a6c5 b2b4 c5d3 c1c3 e6e5');
INSERT INTO evaluations VALUES(46210,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(46210,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(46210,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46210,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(46210,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(46210,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(46210,36,'cp 18','e2e3 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 f6d5 b1c1 c5e7 f1c4 d5b6 c4d3 f8d8 d1c2 h7h6 e1g1 b8c6 a2a3 c8d7 f1d1 a8c8 c2b1 d7e8 h2h3 c6b8 c1c8 d8c8 d1c1 c8c1 b1c1 e7d6 d3e4 b8c6 c1c3');
INSERT INTO evaluations VALUES(57147,30,'cp -201',NULL);
INSERT INTO evaluations VALUES(57147,31,'cp -199',NULL);
INSERT INTO evaluations VALUES(57147,32,'cp -193',NULL);
INSERT INTO evaluations VALUES(57147,33,'cp -180',NULL);
INSERT INTO evaluations VALUES(57147,34,'cp -190',NULL);
INSERT INTO evaluations VALUES(57147,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(57147,36,'cp -198','d7d5 e4d5 d8d5 d1f3 d5e6 f1b5 b8c6 d2d3 c8d7 b1c3 e6g4 f3g4 d7g4 b5c6 b7c6 f2f3 g4h5 c3e4 f8d6 e4g3 h5g6 h2h4 h7h5 c1d2 f7f6 e1c1 g8e7 g1e2 g6f5 d1f1 e8f7 g3e4 a8e8 e2g3 g7g6 f3f4 f5g4 h1g1 e5f4 e4d6 c7d6 f1f4');
INSERT INTO evaluations VALUES(56380,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(56380,31,'cp -171',NULL);
INSERT INTO evaluations VALUES(56380,32,'cp -192',NULL);
INSERT INTO evaluations VALUES(56380,33,'cp -193',NULL);
INSERT INTO evaluations VALUES(56380,34,'cp -197',NULL);
INSERT INTO evaluations VALUES(56380,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(56380,36,'cp -181','e4d5 d8d5 d1f3 d5e6 f1b5 b8c6 d2d3 c8d7 b1c3 e8c8 b5c4 e6g4 f3g4 d7g4 c4f7 c6d4 f7b3 g8f6 f2f3 g4f3 g1f3 d4f3 e1f1 a7a6 f1g2 f3d4 c1e3 f8d6 h1f1 d4b3 a2b3 d8f8 c3e4 f6e4 d3e4 c8d7 f1f3 f8f6 f3f6 g7f6 a1g1 h7h5 g2f3 d7e6 h2h3 d6f8 e3d2 e6f7 b3b4 f8d6 f3e2 b7b6');
INSERT INTO evaluations VALUES(14613,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(14613,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(14613,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(14613,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(14613,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14613,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(14613,36,'cp 0','c6a5 c4b3 c7c6 e1g1 f8e7 d3d4 d8c7 d4e5 d6e5 f3h4 a5b3 a2b3 a7a5 h4f5 c8f5 e4f5 e8g8 d1f3 f8d8 f1e1 e7d6 c3e4 d6e7');
INSERT INTO evaluations VALUES(14679,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14679,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14679,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(14679,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(14679,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14679,36,'cp 0','c4b3 c7c6 e1g1 f8e7 d3d4 d8c7 d4e5 d6e5 f3h4 a5b3 a2b3 a7a5 h4f5 c8f5 e4f5 e8g8 d1f3 a8d8 c1e3 g8h7 f1d1 d8d1 a1d1 f8d8 d1d8 c7d8 g2g3 f6d5 e3d2 d5c3 d2c3 f7f6 g1f1 e7c5 f3g4 c5d4 f1e2 d8d5 g4g6 h7h8 g6e8 h8h7');
INSERT INTO evaluations VALUES(14806,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(14806,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(14806,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(14806,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(14806,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14806,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(14806,36,'cp 13','f8d6 g1f3 d8e7 b1d2 e6e5 d4e5 c6e5 f4e5 d6e5 e1g1 c7c6 f3e5 e7e5 a2a4 a7a5 d1c2 e8g8 b2b4 f8e8 b4a5 a8a5 f1c1 e5g5 d2f3 g5h6 h2h3 c8h3 g2h3 h6h3 f3h2 f6g4 h2g4 h3g4 g1f1 g4h5');
INSERT INTO evaluations VALUES(9238,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9238,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9238,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(9238,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(9238,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(9238,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9238,36,'cp 13','g1f3 d8e7 e1g1 e6e5 d4e5 c6e5 f4e5 d6e5 f1e1 c7c6 b1d2 e5c7 e3e4 d5e4 d3e4 f6e4 e1e4 c8e6 f3d4 e8c8 d4e6 f7e6 d1e1 c7h2 g1h2 e7d6 h2g1 d6d2 e4e6 h8f8 e6e7 d8d7 f2f3 d2e1 a1e1 g7g6 g1f2 c8d8 e7e2 f8f5 e2e8 d8c7 g2g4 f5a5 a2a3 h7h5');
INSERT INTO evaluations VALUES(9239,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(9239,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(9239,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9239,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(9239,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9239,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(9239,36,'cp 0','d8e7 e1g1 e8g8 b1d2 e6e5 d4e5 c6e5 f4e5 d6e5 f1e1 e5d6 e3e4 d5e4 d3e4 e7d8 d1a4 g7g6 e4c2 g8g7 a1d1 c7c6 h2h3 d8c7 d2e4 d6e7 e4g3 c8e6 f3d4 e6c8');
INSERT INTO evaluations VALUES(17302,30,'cp 244',NULL);
INSERT INTO evaluations VALUES(17302,31,'cp 253',NULL);
INSERT INTO evaluations VALUES(17302,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(17302,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(17302,34,'cp 241',NULL);
INSERT INTO evaluations VALUES(17302,35,'cp 235',NULL);
INSERT INTO evaluations VALUES(17302,36,'cp 234','c2c4 d5b6 d4d5 c6e5 c4c5 b6d5 d1d5 e5c6 c1e3 b7b6 d5c4 e7e6 g1f3 b6c5 e3c5 c8a6 c4c1 a6f1 e1f1 f8c5 c1c5 a8b8 b2b3 d8e7 c5e3 e8g8 b1c3 c6b4 a1d1 f8e8 h2h4 c7c5 e3e2 d7d5 f3e5 d5d4 c3e4 e7c7 e5d3 b4d3 e2d3 e6e5 g2g4 c5c4 b3c4 b8b4 e4d2 b4b2');
INSERT INTO evaluations VALUES(17300,30,'cp 241',NULL);
INSERT INTO evaluations VALUES(17300,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(17300,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(17300,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(17300,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(17300,35,'cp 251',NULL);
INSERT INTO evaluations VALUES(17300,36,'cp 254','d5b6 d4d5 c6e5 c4c5 b6d5 d1d5 e5c6 g1f3 e7e6 d5c4 b7b6 c1e3 f8c5 e3c5 b6c5 c4c5 a8b8 b2b3 d8e7 c5c3 e8g8 f1d3 c8b7 e1g1 c6b4 b1d2 c7c5 f1e1 e7f6 e1c1 f6c3 c1c3 f8d8 d3f1 d7d6 a2a3 b4d5 c3c1 g7g6 f3e1 b8c8 h2h3 d5f4 d2c4');
INSERT INTO evaluations VALUES(17213,30,'cp 246',NULL);
INSERT INTO evaluations VALUES(17213,31,'cp 242',NULL);
INSERT INTO evaluations VALUES(17213,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(17213,33,'cp 265',NULL);
INSERT INTO evaluations VALUES(17213,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(17213,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(17213,36,'cp 256','d4d5 c6e5 c4c5 b6d5 d1d5 e5c6 g1f3 e7e6 d5c4 b7b6 c1e3 b6c5 e3c5 c8a6 c4c1 a6f1 e1f1 f8c5 c1c5 c6e7 h2h4 e8g8 h1h3 h7h6 b1c3 d7d6 c5e3 c7c5 e3d2 d6d5 a1e1 d8b6 f1g1 a8b8 b2b3 f8d8 h4h5 b8c8 f3e5 e7f5 g2g4 f5d4');
INSERT INTO evaluations VALUES(37247,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(37247,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(37247,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(37247,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(37247,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(37247,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(37247,36,'cp 27','g1f3 b8c6 e1g1 d7d6 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 e2e3 d4b6 a4b6 a7b6 f1e1 f8e8 e3e4 c8h3 a2a4 h7h6 b2b4 d8d7 f3h1 a8a7 b4b5 e8a8 h1g2 c7c6 c1b2 c6b5 c4b5 h3g2 g1g2 d7e6 f2f4 e6e7 h2h3 e5f4 g3f4');
INSERT INTO evaluations VALUES(37246,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(37246,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(37246,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(37246,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(37246,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(37246,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(37246,36,'cp 36','b8c6 e1g1 d7d6 e2e3 f8e8 d2d3 c5b6 b2b3 a7a5 c1b2 c8f5 d1d2 d8d7 f3g5 c6b4 g5e4 f6e4 d3e4 f5e6 c3a4 b6a7 a2a3 b4c6 a1d1 e6g4 d1e1 h7h6 a4c3 c6e7 f2f3 g4e6 f3f4 c7c6 f4f5');
INSERT INTO evaluations VALUES(52105,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52105,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52105,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52105,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(52105,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(52105,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(52105,36,'cp 12','d7d6 g2g3 b8c6 f1g2 h7h5 h2h4 g7g6 g1e2 f8g7 d2d3 g8e7 a1b1 e8g8 e1g1 c8g4 a2a3 a7a6 b2b4 b7b5 c1g5 c6d4 f2f3 g4d7 b4c5 d6c5 e2d4 c5d4 c3d5 f7f6 g5d2 b5c4 d3c4 e7d5 c4d5 a8c8 a3a4 g8h7 g1h2 f8f7 d1b3 a6a5 g2h3');
INSERT INTO evaluations VALUES(53880,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53880,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53880,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53880,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53880,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53880,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53880,36,'cp 5','g2g3 b8c6 d2d3 a7a6 f1g2 a8b8 g1e2 b7b5 b2b3 f8e7 e1g1 g8f6 f2f4 e8g8 c1d2 c8g4 h2h3 g4e2 c3e2 f6d7 e2c3 e7f6 a1b1 b5b4 c3d5 e5f4 g3f4 f6d4 d5e3 g8h8 g1h1 d4e3 d2e3 c6d4 h1g1 f7f5 e3d4 c5d4 e4f5');
INSERT INTO evaluations VALUES(55130,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(55130,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(55130,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(55130,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(55130,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(55130,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(55130,36,'cp -59','e7e5 d2d4 b8c6 d4e5 c6e5 f1g2 g8f6 g1f3 e5f3 g2f3 c7c6 e1g1 f8c5 b1d2 e8g8 c2c4 c8h3 f1e1 f8e8 c4d5 f6d5 d2b3 c5b6 c1d2 a7a5 a2a4 g7g6 f3g2 h3g2 g1g2 d8f6 d1c2 a8d8 a1d1 f6e6 b3d4 b6d4');
INSERT INTO evaluations VALUES(19235,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(19235,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(19235,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(19235,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(19235,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(19235,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(19235,36,'cp 29','f8g7 c2c3 e7e6 c4b5 e6e5 d2d4 c5d4 c3d4 e5d4 c1f4 a7a6 b5a4 g8e7 f4d6 b7b5 a4b3 c8b7 f1e1 e7c8 e4e5 c8d6 e5d6 e8f8 h2h4 g7f6 b1d2 f6h4 d2e4 h4f6 e4c5 f8g7 c5b7 d8b6 f3d2 a8e8 e1e8 h8e8 d1f3 b6b7 d2e4 e8e4');
INSERT INTO evaluations VALUES(32076,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(32076,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(32076,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(32076,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(32076,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(32076,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(32076,36,'cp 119','b8c6 f1b5 d6e5 d1d8 a8d8 b5c6 b7c6 b1d2 g8e7 d2b3 e7g6 c1e3 a7a6 e1e2 f8e7 b3a5 d8d6 a1d1 c6c5 d1g1 d6b6 a5c4 b6b8 g1d1 g6f8 d1d3 f7f6 d3a3 b8a8 h1d1 f8e6 c2c3 g7g5 b2b3 e8f7 a3a5 h7h5 d1d5 h8d8 c4b2 d8d5 e4d5');
INSERT INTO evaluations VALUES(26941,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(26941,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(26941,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(26941,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(26941,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(26941,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(26941,36,'cp 124','f1b5 d6e5 d1d8 a8d8 b5c6 b7c6 b1d2 g8e7 d2b3 e7g6 c1e3 a7a6 e1e2 f8e7 b3a5 d8d6 h1g1 e8g8 a5c4 d6d8 b2b3 f7f6 c4a5 d8d6 a5b7 d6e6 g1d1 f8c8 d1d7 c6c5 b7c5 e6c6 b3b4 a6a5 a1g1 e7c5 b4c5 c8b8 e2d3 b8b2 d3c3 b2a2 g1b1 a5a4');
INSERT INTO evaluations VALUES(1621,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(1621,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(1621,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(1621,33,'cp 216',NULL);
INSERT INTO evaluations VALUES(1621,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(1621,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(1621,36,'cp 219','g1f3 e5a5 d2d4 g8f6 e1g1 c8g4 d4d5 c6b4 h2h3 g4d7 f3e5 c7c6 d5c6 d7c6 c1f4 e7e6 a2a3 b4d5 c3d5 a5d5 e5c6 b7c6 c2c4 d5d1 f1d1 a8c8 b2b4 f8e7 g2g3 e8g8 e2f3 f8d8 d1d8 e7d8 a1d1 d8e7 g1f1 g8f8 f4e3 a7a6 f3g2 f8e8 d1b1 f6d7 b1b3 e7f6 c4c5');
INSERT INTO evaluations VALUES(1619,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(1619,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(1619,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(1619,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(1619,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(1619,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(1619,36,'cp 206','e5a5 d2d4 c8g4 d4d5 a8d8 c1d2 g4f3 e2f3 c6d4 e1g1 g8f6 f1e1 d8d7 f3e4 c7c5 e4d3 a5d8 d2e3 a7a6 c3e4 f6d5 e4c5 d5e3 f2e3 d7c7 d3a6 b7a6 d1d4 d8d4 e3d4 e7e6 b2b4 h7h5 c2c3 f8e7 c5a6 c7c3 a2a4 e7d6');
INSERT INTO evaluations VALUES(38796,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(38796,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38796,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38796,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(38796,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(38796,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(38796,36,'cp 5','d5e4 d3e4 g8f6 e4f3 e8g8 g1e2 e6e5 e1g1 b4c3 e2c3 e5d4 c3e2 c7c5 b2b4 d4d3 d1d3 d8d3 c2d3 c5b4 a2a3 b4a3 c1a3 f8d8 a3e7 d8e8 e7f6 g7f6 f1c1 b8c6 c1c6 b7c6 f3c6 e8e2 c6a8 c8f5 d3d4 e2d2 d4d5 f5e4 h2h4 e4d5 a8d5 d2d5 a1a7 h7h5 g2g3 g8g7 g1g2 g7g8 a7e7 d5d2 g2f3 d2d4 e7c7 d4d8');
INSERT INTO evaluations VALUES(38864,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38864,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38864,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38864,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38864,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38864,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38864,36,'cp 0','d3e4 g8f6 e4f3 e8g8 g1e2 e6e5 e1g1 b4c3 e2c3 e5d4 c3e2 c7c5 b2b4 d4d3 d1d3 d8d3 c2d3 c5b4 a2a3 b4a3 c1a3 f8d8 a3e7 d8e8 e7f6 g7f6 f1c1 b8c6 c1c6 b7c6 f3c6 e8e2 c6a8 e2d2 g1f1 d2d3 a1a7 c8e6 f1e2 d3b3 a7b7 b3a3 b7b8 g8g7 e2d2 a3a2 d2e3 h7h5 a8f3 a2a3 e3d4 a3a4 d4c3 a4a3');
INSERT INTO evaluations VALUES(38857,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38857,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38857,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38857,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38857,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38857,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38857,36,'cp 0','g8f6 e4f3 e8g8 g1e2 e6e5 e1g1 b4c3 e2c3 e5d4 c3e2 c7c5 b2b4 d4d3 d1d3 d8d3 c2d3 c5b4 a2a3 b4a3 c1a3 f8d8 a3e7 d8e8 e7f6 g7f6 f1c1 b8c6 c1c6 b7c6 f3c6 e8e2 c6a8 e2d2 g1f1 d2d3 a1a7 c8e6 f1e2 d3b3 a7a5 b3b2 e2f3 b2b3');
INSERT INTO evaluations VALUES(42142,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42142,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42142,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42142,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42142,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42142,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42142,36,'cp 0','e4f3 e8g8 g1e2 e6e5 e1g1 b4c3 e2c3 e5d4 c3e2 c7c5 b2b4 d4d3 d1d3 d8d3 c2d3 c5b4 a2a3 b4a3 c1a3 f8d8 a3e7 d8e8 e7f6 g7f6 f1c1 b8c6 c1c6 b7c6 f3c6 e8e2 c6a8 e2d2 g1f1 d2d3 a1a7 c8e6 f1e2 d3b3 a7a5 b3b2 e2f3 b2b3 f3e2');
INSERT INTO evaluations VALUES(42138,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42138,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42138,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42138,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42138,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42138,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42138,36,'cp 0','e8g8 g1e2 e6e5 e1g1 b4c3 e2c3 e5d4 c3e2 c7c5 b2b4 d4d3 d1d3 d8d3 c2d3 c5b4 a2a3 b4a3 c1a3 f8d8 a3e7 d8e8 e7f6 g7f6 f1c1 b8c6 c1c6 b7c6 f3c6 e8e2 c6a8 e2d2 g1f1 d2d3 a1a7 c8e6 f1e2 d3b3 a7a5 b3b2 e2f3 b2b3');
INSERT INTO evaluations VALUES(54823,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(54823,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(54823,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(54823,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(54823,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(54823,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(54823,36,'cp 70','e2e4 g8f6 b1c3 f8g7 c1g5 c7c6 d1d2 d8a5 f1c4 b8d7 c4b3 h7h6 g5e3 f6g4 e3f4 e7e5 f4g3 e8g8 e1g1 g4f6 h2h3 e5d4 d2d4 d6d5 e4e5 f6e8 d4d2 a5d8 f1e1 e8c7 f3d4 d7c5 f2f4 c7e6 g3f2 e6d4 d2d4');
INSERT INTO evaluations VALUES(29759,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(29759,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(29759,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(29759,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(29759,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(29759,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(29759,36,'cp 120','f2f3 g4e6 g1e2 e6c8 e2f4 c7c5 d4d5 e7e5 f4e2 f8e7 e2g3 e8g8 f1c4 f6e8 a2a4 e8d6 b2b3 a7a6 d1e2 f8e8 c1d2 b8d7 e1g1 d7f8 a4a5 c8d7 e2f2 f8g6 d2e3 a8c8 f1d1 h7h6 g3f1 d8c7 e3d2 g6f4 f1e3 g8h8 c3a4 d7a4 a1a4 e7g5 e3f5 d6f5 e4f5');
INSERT INTO evaluations VALUES(29757,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(29757,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(29757,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(29757,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(29757,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(29757,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(29757,36,'cp 140','g4e6 g1e2 e6c8 e2f4 c7c5 d4d5 e7e5 f4e2 f8e7 e2g3 e8g8 f1c4 f6e8 a2a4 e8d6 b2b3 a7a6 e1g1 b7b6 d1e2 e7g5 f1d1 h7h6 g3f1 b8d7 c4d3 d7f6 c1g5 h6g5 f1e3 g7g6 d1e1 f6h5 g2g3 h5f6 e1c1 g8g7 c3d1 a8b8 d1f2 a6a5');
INSERT INTO evaluations VALUES(42626,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(42626,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(42626,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42626,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42626,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(42626,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42626,36,'cp 0','f2f4 d7d6 c3a4 c5g1 h1g1 b7b5 a4c3 b5b4 c3e2 b8c6 a2a3 b4a3 a1a3 e5f4 e2f4 e8g8 g2g4 c6d4 c2c3 d4e6 g4g5 e6f4 c1f4 f6d7 f1e2 a7a5 d1a1 f7f5 g5f6 d8f6 g1f1 f6h4 f4g3 f8f1 e2f1 h4g5 e1d1 d7f6 d1c2 c8e6 f1e2');
INSERT INTO evaluations VALUES(52890,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(52890,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(52890,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(52890,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(52890,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(52890,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(52890,36,'cp 88','c8g4 h2h3 g4f3 d1f3 g8f6 c1e3 e7e6 f1d3 d8a5 e1g1 b8d7 c3e2 f8e7 c2c4 e8g8 a2a3 a5c7 a1d1 c6c5 e3f4 e7d6 f4d6 c7d6 b2b4 c5b4 a3b4 d6b4 d1b1 b4a4 f3b7 f8b8 b7c7 b8b1 f1b1 a4a3 c7b7 a8d8 b1b3 a3a1 g1h2 a1e1 b7c7 d8f8 b3b7 e1f2 b7a7');
INSERT INTO evaluations VALUES(54869,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(54869,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(54869,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(54869,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(54869,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(54869,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(54869,36,'cp 95','e4d5 d8d5 g1f3 f8g7 f1e2 g8h6 b1c3 d5d8 c1f4 h6f5 d1d2 b8c6 d4d5 c6d4 f3d4 f5d4 e1c1 c8f5 e2d3 f5d3 d2d3 e8g8 h1e1 c7c5 f4e5 g7e5 e1e5 d8c7 d1e1 e7e6 d5e6 f7e6 c3b5 d4b5 d3b5 f8f2 e5c5 c7h2 b5b7');
INSERT INTO evaluations VALUES(54842,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(54842,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(54842,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(54842,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(54842,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(54842,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(54842,36,'cp 81','d8d5 f1e2 f8g7 g1f3 d5d8 e1g1 e7e6 c2c3 g8e7 c1f4 e8g8 a2a4 b7b6 a4a5 a7a6 b1d2 c8b7 f1e1 b8d7 d2c4 d7f6 e2f1 b6b5 c4e5 e7f5 h2h3 b7d5 e5d3 f8e8 d1e2 c7c6 f4h2 f6e4 f3d2 e4d2 e2d2 g7f8');
INSERT INTO evaluations VALUES(34512,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(34512,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(34512,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(34512,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(34512,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(34512,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(34512,36,'cp 107','g1f3 f8g7 f1e2 d5d8 e1g1 e7e6 a2a4 g8e7 c2c3 b7b6 a4a5 a7a6 c1f4 c8b7 b1d2 e8g8 f1e1 b8d7 d1c2 e7d5 f4g3 d5e7 h2h3 e7f5 g3f4 b6b5 b2b4 b7d5 a1d1 d8e7 e2d3 a8c8 d2e4 f5d6 e4c5 d7c5 d4c5');
INSERT INTO evaluations VALUES(12796,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12796,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12796,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(12796,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(12796,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12796,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(12796,36,'cp 4','g1e2 e7e6 d2d4 d7d5 c4d5 e6d5 c1g5 c5d4 e2d4 f8e7 f1b5 c8d7 b5c6 b7c6 e1g1 f6e4 g5e7 d8e7 f1e1 e8g8 f2f3 e4c3 e1e7 c3d1 e7d7 d1b2 d4c6 g7g6 d7d5 g8g7 d5d7 a7a6 g1f2 a8c8 a1c1 h7h5 d7d2 b2a4 c1c2 f8e8 d2d4 a4b6 d4d6 b6a4 f2g3 c8c7 h2h4 c7c8');
INSERT INTO evaluations VALUES(17586,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(17586,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(17586,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(17586,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17586,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(17586,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(17586,36,'cp 41','g8f6 h2h3 e7e5 d4c6 b7c6 f2f4 e5f4 e3f4 f8e7 e4e5 d6e5 d1d8 e7d8 f4e5 e8g8 e1c1 f8e8 e5d4 c8b7 a2a3 d8e7 d4f2 a8d8 d1d8 e7d8 h1g1 h7h5 g2g3 f6d5 c3d5 c6d5 f2d4 d8g5 c1d1 b7c8 c2c3 g7g6 b2b4 g5e3 d4e3 e8e3 d1d2 e3f3');
INSERT INTO evaluations VALUES(48306,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48306,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(48306,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(48306,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(48306,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(48306,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48306,36,'cp -34','c2c3 c7c5 g1f3 f8e7 f1d3 e8g8 e1g1 b7b6 d1e2 f6e4 b2b3 c8b7 c1b2 b8d7 b1d2 c5d4 c3d4 e7b4 g2g4 a8c8 f1c1 d8e7 d2f1 f7f6 f1g3 g7g6 c1c8 f8c8 a2a3 b4d6 e2d1 a7a5 a1c1 d6a3 b2a3 e7a3 c1c8 b7c8 g3e4');
INSERT INTO evaluations VALUES(48304,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(48304,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(48304,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(48304,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(48304,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(48304,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(48304,36,'cp -37','f8e7 g1f3 b7b6 f1d3 a7a5 b1d2 e8g8 e1g1 c8a6 d3a6 b8a6 d1e2 c7c5 e3e4 d5e4 d2e4 c5d4 f3d4 d8d5 e4f2 a6c5 c1e3 a8d8 a1d1 f8e8 c3c4 d5b7 f4f5 e6f5 d4f5 e7f8 d1d8 e8d8 f1d1 d8d1 e2d1 b7c6 b2b3 c5d7 f2g4 c6e6 g4f6 d7f6');
INSERT INTO evaluations VALUES(16162,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16162,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16162,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16162,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16162,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16162,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16162,36,'cp -1','c3d5 f8c5 c2c3 a7a5 f3e5 c6e5 d3d4 f6d5 e4d5 d8e7 f1e2 c5d6 d4e5 d6e5 e1g1 e8g8 f1e1 f8e8 g2g3 b7b6 c1e3 a5a4 e3d4 d7d6 e2b5 c8d7 d1d3 d7b5 d3b5 f7f5 d4e5 d6e5 a1d1 e7d6 b2b4 a4b3 a2b3 a8d8');
INSERT INTO evaluations VALUES(44878,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(44878,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(44878,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44878,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(44878,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(44878,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(44878,36,'cp -40','d5c4 b1c3 h7h6 g5h4 b7b5 a2a4 e7e6 h4f6 g7f6 a4b5 c6b5 c3b5 d8b6 b5a3 c8b7 a3c2 b8c6 e2e4 c6a5 f1e2 b7e4 e1g1 h8g8 c2e3 f8b4 d4d5 e8c8 e2c4 a5c4 e3c4 b6b7 d1c1 e4f3 c4e5 c8b8 e5f3 d8d5 g1h1 d5d2 c1c4 e6e5 a1d1 d2d1 f1d1 g8c8');
INSERT INTO evaluations VALUES(45000,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(45000,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(45000,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(45000,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(45000,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(45000,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(45000,36,'cp -30','b1c3 h7h6 g5h4 b7b5 a2a4 e7e6 h4f6 g7f6 e2e3 f8b4 g2g3 b8d7 f1g2 c8b7 e1g1 a7a6 d1e2 d8c7 f3d2 h6h5 f1c1 h5h4 b2b3 c6c5 g2b7 c7b7 b3c4 b4c3 c1c3 e8e7 c3d3 c5d4 e3d4 b5c4 d2c4 a8c8 c4e3 h4g3 h2g3');
INSERT INTO evaluations VALUES(44998,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(44998,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(44998,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(44998,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(44998,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(44998,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(44998,36,'cp -54','h7h6 g5h4 b7b5 e2e4 g7g5 h4g3 f8g7 f1e2 b8d7 f3e5 d7e5 g3e5 e8g8 d1c2 c8e6 h2h4 g5g4 a1d1 d8b6 f2f4 a8d8 f4f5 e6c8 e1g1 f6h5 e5g7 g8g7 c2d2 e7e5 f5e6 c8e6 d2e3 d8e8 a2a3 f7f5 g2g3 g7h7 e4e5 b6d8 a3a4');
INSERT INTO evaluations VALUES(15722,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(15722,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(15722,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(15722,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(15722,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15722,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15722,36,'cp 0','d8d6 c4f7 e8e7 f7b3 d4b3 a2b3 h7h6 g5f3 e5e4 f3g1 e7f7 b1c3 c8f5 g1e2 a7a6 e1g1 a8e8 e2g3 f5g6 f1e1 d6c6 d2d3 e4d3 e1e8 c6e8 c2d3 f8d6 c1e3 e8e6 g3e4 h8d8 e4d6 d8d6 d3d4 g6h5 d1c2 h5g6');
INSERT INTO evaluations VALUES(1020,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(1020,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(1020,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(1020,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1020,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(1020,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1020,36,'cp 11','c4f7 e8e7 f7b3 d4b3 a2b3 h7h6 g5f3 e5e4 f3g1 e7f7 b1c3 c8f5 g1e2 a7a6 e1g1 a8e8 h2h3 d6d7 a1a4 b7b5 a4d4 d7c8 d2d3 f8c5 d4e4 f5e4 d3e4 e8d8 d1e1 h8e8 c1e3 c5b6 e2g3 c8d7 f2f3 b6e3 e1e3 d7d2 e3d2 d8d2 f1f2 e8d8 g3f1 d2f2 g1f2 f7e6 f2e3 g7g5 c3e2 e6e5 f1g3 d8d7 e2c3');
INSERT INTO evaluations VALUES(931,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(931,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(931,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(931,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(931,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(931,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(931,36,'cp 0','e8e7 f7b3 d4b3 a2b3 h7h6 g5f3 e5e4 f3g1 e7f7 b1c3 c8f5 g1e2 a7a6 e1g1 a8e8 e2g3 f5g6 f1e1 d6c6 d2d4 e4d3 e1e8 c6e8 c2d3 f8d6 c1d2 e8d7 d1f3 c7c6 c3e4 d6g3 e4f6 g3h2 g1h2 g7f6 a1e1 h8e8 e1e8 d7e8 d3d4 e8e4 f3g3 e4e2 g3f4 e2h5 h2g1 h5d1 g1h2');
INSERT INTO evaluations VALUES(2146,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2146,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(2146,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(2146,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(2146,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(2146,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2146,36,'cp 14','f7b3 d4b3 a2b3 h7h6 g5f3 e5e4 f3g1 e7f7 g1e2 b7b5 e1g1 c8b7 d2d3 e4d3 e2f4 d3c2 d1c2 b7e4 c2e2 a7a6 b1c3 d6c6 c3e4 c6e4 e2e4 f6e4 c1e3 f8d6 g2g3 e4f6 f4d3 h8d8 f1d1 a6a5 g1g2 f6d5 e3d4 d8d7 d4c5 f7g8 c5d6 c7d6 d3e1 d5b4 a1c1 a8e8 e1d3 b4d3');
INSERT INTO evaluations VALUES(43128,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(43128,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(43128,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(43128,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(43128,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(43128,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(43128,36,'cp -28','f6e4 g5d2 d7d6 e2e3 e8g8 b1c3 e4d2 d1d2 c7c6 f1e2 b8d7 e1g1 e7e5 d4d5 c6c5 e3e4 b7b6 g2g3 d7f6 a2a3 g8h8 b2b4 f6g8 f3g5 c8d7 b4c5 b6c5 a1b1 a7a6 b1b7 g7h6 f2f4 d8e7 g1g2 e5f4 g3f4');
INSERT INTO evaluations VALUES(1486,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1486,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1486,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1486,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1486,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1486,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1486,36,'cp 0','c5d4 c3d4 f8b4 c1d2 g8e7 a3b5 b4d2 d1d2 e8g8 b5c3 f7f6 e5f6 f8f6 f1e2 c8d7 e1g1 d7e8 a1c1 e7f5 f1d1 e8h5 f3e5 h5e8');
INSERT INTO evaluations VALUES(1526,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1526,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1526,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1526,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1526,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1526,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1526,36,'cp 0','c3d4 f8b4 c1d2 g8e7 a3b5 b4d2 d1d2 e8g8 b5c3 f7f6 e5f6 f8f6 f1e2 c8d7 e1g1 d7e8 a1c1 e7f5 f1d1 e8h5 f3e5 h5e8');
INSERT INTO evaluations VALUES(1513,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1513,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1513,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1513,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1513,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1513,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1513,36,'cp 0','f8b4 c1d2 g8e7 a3b5 b4d2 d1d2 e8g8 b5c3 f7f6 e5f6 f8f6 f1e2 c8d7 e1g1 d7e8 h2h3 e8h5 f3e5 h5e8');
INSERT INTO evaluations VALUES(145,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(145,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(145,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(145,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(145,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(145,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(145,36,'cp 0','c1d2 g8e7 a3b5 b4d2 d1d2 e8g8 b5c3 f7f6 e5f6 f8f6 f1e2 c8d7 e1g1 d7e8 a1c1 e7f5 f1d1 e8h5 f3e5 h5e8');
INSERT INTO evaluations VALUES(144,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(144,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(144,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(144,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(144,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(144,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(144,36,'cp 0','g8e7 a3b5 b4d2 d1d2 e8g8 b5c3 f7f6 e5f6 f8f6 f1e2 c8d7 e1g1 d7e8 c3a4 b6b4 d2b4 c6b4 a4c3 b4c6 f1d1 a8c8 a1c1 h7h6 h2h3 a7a6 d1e1 e7f5 e2d3 f5e7');
INSERT INTO evaluations VALUES(43059,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(43059,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(43059,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(43059,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(43059,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(43059,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(43059,36,'cp 72','h2h4 d7d6 h4h5 h7h6 h5g6 e7g6 e5d6 d8d6 b1c3 b8c6 c1e3 b7b6 d1d2 c8b7 e1c1 e8c8 e3h6 c8b8 f1e2 g7h6 h1h6 h8h6 d2h6 d6f4 h6f4 g6f4 e2f1 c6e7 f3h4 f4d5 c3d5 e7d5 h4f3 d5b4 f1c4 b7f3 g2f3 d8h8 d1g1 h8h2 c2c3 b4d5 g1g7 a7a5 g7f7 h2f2 c4d3 d5e3 f7e7 f2f3 e7e6');
INSERT INTO evaluations VALUES(48825,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(48825,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(48825,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(48825,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(48825,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(48825,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48825,36,'cp -33','d5d4 c3b1 f6d7 f1e2 f8e7 e1g1 e8g8 b1d2 a7a5 a2a4 b8c6 d2c4 b7b6 g2g3 e7b4 f3d2 d7c5 e2g4 f7f6 d2b3 c5b3 c2b3 c8g4 d1g4 d8c8 g4c8 a8c8 g1g2 h7h5 f2f4 e5f4 g3f4 c8e8 f1g1 f6f5 e4e5 c6e7');
INSERT INTO evaluations VALUES(48922,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48922,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(48922,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(48922,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(48922,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(48922,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48922,36,'cp -48','c3b1 f8d6 b1d2 e8g8 f1e2 f8e8 d2c4 d6b4 c1d2 b4d2 d1d2 b8c6 e1g1 b7b6 b2b3 a7a6 a2a4 d8e7 h2h3 c8d7 f1c1 h7h6 c2c3 b6b5 c4b2 d4c3 d2c3 f6h7 d3d4 e5d4 c3c5 b5a4 c5e7 e8e7 a1a4 e7e4 e2a6 h7g5 f3g5 h6g5');
INSERT INTO evaluations VALUES(40793,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(40793,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(40793,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(40793,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(40793,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(40793,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(40793,36,'cp 53','d2d4 e5d4 f3d4 e8g8 f1d3 f8e8 e1g1 b8d7 c1f4 e7f8 f1e1 h7h6 d3f1 a7a5 a2a4 c7c6 e1e3 d7e5 f4g3 d8b6 b2b3 e5g6 d1d2 c8d7 f1c4 a8d8 a1d1 b6c5 c4e2 d6d5 e4e5 f6e4 e3e4 d5e4 c3e4 c5b6 e4d6 g6e5 d6e8 d8e8 e2f1 e5g6');
INSERT INTO evaluations VALUES(40790,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40790,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(40790,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(40790,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(40790,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(40790,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(40790,36,'cp 49','e5d4 f3d4 e8g8 f1d3 f8e8 e1g1 b8d7 f1e1 c7c6 c1f4 e7f8 d3f1 h7h6 a2a4 a7a5 d4b3 d7e5 g1h1 c8e6 f4h2 d8b6 f2f4 e5d7 h2g1 b6b4 b3d4 d7c5 d1f3 e6c4 e4e5 d6e5 f4e5 f6d7 f1c4 b4c4 f3g3');
INSERT INTO evaluations VALUES(40854,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(40854,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(40854,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(40854,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(40854,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(40854,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(40854,36,'cp 49','f3d4 e8g8 f1d3 f8e8 e1g1 b8d7 f1e1 e7f8 c1f4 h7h6 d3f1 c7c6 a2a4 a7a5 d4b3 d7e5 f4g3 d8b6 d1d4 b6b4 d4b4 a5b4 c3b1 c8e6 b1d2 e5d7 f2f3 e6b3 d2b3 d6d5 e4d5 e8e1 g3e1 f6d5 e1f2 f8d6 f2d4');
INSERT INTO evaluations VALUES(40839,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(40839,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(40839,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(40839,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(40839,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(40839,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(40839,36,'cp 28','e8g8 f1d3 f8e8 e1g1 c7c6 f1e1 b8d7 d3f1 e7f8 c1f4 a7a5 a2a4 h7h6 d4b3 d7e5 g1h1 c8e6 f4h2 e6b3 c2b3 e5d7 f1c4 d7c5 f2f3 f6d7 h2d6 d7b6 d6f8 e8f8 d1d8 f8d8 a1d1 b6c4 b3c4 d8d3 d1d3 c5d3');
INSERT INTO evaluations VALUES(36261,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(36261,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(36261,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(36261,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36261,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(36261,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(36261,36,'cp 34','f1d3 f8e8 e1g1 b8d7 f1e1 c7c6 c1f4 e7f8 d3f1 a7a5 a2a4 h7h6 e1e3 d7c5 d1e2 c8d7 b2b3 d6d5 e4d5 c6d5 e3e8 d8e8 e2e8 d7e8 c3b5 a8c8 f4e3 c5e6 c2c4 f8b4 a1d1 e6d4 d1d4 e8b5 a4b5 d5c4');
INSERT INTO evaluations VALUES(36264,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36264,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36264,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(36264,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(36264,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(36264,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(36264,36,'cp 23','f8e8 e1g1 b8d7 c1f4 e7f8 f1e1 c7c6 d3f1 h7h6 d4b3 d7e5 a2a4 a7a5 f4g3 d8b6 d1d4 b6b4 c3a2 b4d4 b3d4 g7g5 f2f3 f6h5 g3f2 f8g7 a1d1 e5g6 a2c1 h5f4 c1e2 f4e2 e1e2 g7e5 g2g3 h6h5 d4f5 c8f5 e4f5 g6e7');
INSERT INTO evaluations VALUES(17230,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(17230,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(17230,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(17230,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(17230,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(17230,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(17230,36,'cp -22','e7e6 c4b3 d7d5 e4d5 e6d5 d2d4 f8b4 g1e2 c6a5 e2g3 e8g8 e1g1 c7c6 c3e2 f8e8 c2c3 a5b3 a2b3 b4f8 e2f4 a7a5 f1e1 e8e1 d1e1 b7b6 e1d1 c8g4 d1d3 d8d6 h2h3 g4c8 d3f3 c8a6 f4d3 a6d3 f3d3 a8e8 c1d2 h7h6 d3f3 f6e4 d2f4 d6d7 a1e1');
INSERT INTO evaluations VALUES(16256,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(16256,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(16256,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(16256,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(16256,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(16256,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(16256,36,'cp -22','c4b3 d7d5 e4d5 e6d5 d2d4 f8b4 g1e2 c6a5 e2g3 c7c6 e1g1 e8g8 c3e2 f8e8 c2c3 a5b3 a2b3 b4f8 e2f4 a7a5 f4d3 b7b6 c1e3 f6e4 f1e1 e4g3 h2g3 c8f5 e3f4 e8e1 d3e1 h7h6 e1d3 g7g5 f4e3 d8e8 b3b4 a5b4 a1a8 e8a8 d3b4');
INSERT INTO evaluations VALUES(47159,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(47159,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47159,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47159,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47159,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47159,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47159,36,'cp 0','c7c5 e2e3 b8c6 g1f3 d8b6 d4c5 b6b2 a1b1 b2c3 f1b5 e7e6 e1g1 f8e7 f3e5 c8d7 e5d7 f6d7 e3e4 a7a6 e4d5 a6b5 d5c6 b7c6 d2e4 c3c4 e4d6 e7d6 d1d6 c4c5 d6c7 c5b6 c7d6');
INSERT INTO evaluations VALUES(49731,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(49731,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(49731,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(49731,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(49731,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(49731,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(49731,36,'cp -111','f6e4 d1e2 d8e7 d2d4 e4f6 b1c3 e7e2 c4e2 f8b4 c1f4 d7d6 e1g1 b4c3 b2c3 e8g8 e2d3 b8d7 c3c4 b7b6 a1e1 h7h6 d4d5 c8a6 a2a4 f8e8 e1a1 f6e4 f3d4 e4c5 a4a5 d7e5 f4e5 e8e5 d4c6 e5e3 f1e1 e3e1 a1e1 b6a5 c6a5');
INSERT INTO evaluations VALUES(49911,30,'cp -116',NULL);
INSERT INTO evaluations VALUES(49911,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(49911,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(49911,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(49911,34,'cp -119',NULL);
INSERT INTO evaluations VALUES(49911,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(49911,36,'cp -110','d1e2 d8e7 d2d4 e4f6 b1c3 e7e2 c4e2 c7c6 c1f4 d7d5 e1c1 f8e7 e2d3 e8g8 d1e1 f8e8 e1e2 e7b4 e2e8 f6e8 h1e1 e8f6 h2h3 b8d7 g2g4 h7h6 a2a3 b4f8 a3a4 f8b4 h3h4 d7f8 g4g5 f6h5 f4d2 h6g5 h4g5 c8e6 f3e5 h5g3 c3e2 g3e2 e1e2 b4d6 a4a5 a8e8');
INSERT INTO evaluations VALUES(57070,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(57070,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57070,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57070,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57070,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57070,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57070,36,'cp 0','g2g3 g8f6 f1g2 f8e7 g1f3 d7d5 c2c4 c7c5 e1g1 e8g8 d3d4 d5c4 d4c5 e7c5 b1d2 c4c3 d2b3 c5e7 b2c3 b8c6 f3d4 c6e5 d1c2 e5c4 c2d3 c4e5');
INSERT INTO evaluations VALUES(32107,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32107,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32107,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32107,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32107,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32107,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32107,36,'cp 0','e5f4 d2d4 d8f6 b1c3 g8e7 c3b5 b8a6 h2h3 g4f3 d1f3 c7c6 c1f4 e7g6 b5d6 f8d6 f4d6 f6d6 f3f7 e8d8 e1g1 d6d4 g1h1 h8f8 f7b7 f8f1 a1f1 a6c7 b7c6 g6e5 f1f8 d8e7 c6c7 e7f8 c4d5 a8e8 c7d6 e8e7 d6d8 e7e8');
INSERT INTO evaluations VALUES(40377,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(40377,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(40377,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40377,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(40377,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(40377,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(40377,36,'cp -15','f6g4 c4b5 e8f8 e1g1 g7e5 d1d2 e5h2 g1h1 h2e5 a1d1 e5g7 f2f4 b8c6 d4c6 b7c6 b5c6 a8b8 e4e5 b8b2 e3a7 d8c7 d2c1 b2a2 c3a2 c7a7 e5d6 e7d6 c1d2 g7f6 a2b4 g4e3 d2d6 f6e7 d6e5 f7f6 e5c3 e3d1 f1d1 a7c5 b4d5 c5c3 d5c3 f8g7 c6f3 f6f5 d1d3 h8e8 c3b5 e8d8 d3d8 e7d8');
INSERT INTO evaluations VALUES(40966,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(40966,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40966,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40966,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40966,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40966,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(40966,36,'cp 16','c7c6 e1g1 e8g8 d1c2 b7b6 f1d1 b8d7 b1d2 c8a6 e2e4 f6e4 d2e4 d5e4 f3e5 d7e5 d4e5 d8c7 c2e4 a8b8 b2b3 f8d8 c1e3 a6b7 e4g4 c6c5 g2b7 c7b7 h2h4 b7c7 e3f4 d8d4 d1d4 c5d4 g4f3 b8d8 a1c1 h7h6 c1d1 c7c8 f3e4');
INSERT INTO evaluations VALUES(40086,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40086,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40086,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(40086,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40086,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40086,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(40086,36,'cp 21','b2b3 e8g8 e1g1 b7b6 d1c2 b8d7 c1b2 c8b7 f1d1 c6c5 c4d5 f6d5 b1c3 c5d4 c3d5 b7d5 f3d4 a8c8 c2b1 d5g2 g1g2 c8c7 b1d3 d7f6 d3f3 d8d5 f3d5 f6d5 a1c1 f8c8 c1c7 c8c7 d1c1 c7c1 b2c1 d5b4 c1a3 e7c5 a3b4 c5d4 e2e3 d4e5 e3e4 e5d4 f2f4 f7f5 g2f3 g8f7 g3g4');
INSERT INTO evaluations VALUES(56354,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(56354,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56354,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(56354,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56354,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(56354,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(56354,36,'cp 0','d5c4 e2e3 e5d4 e3d4 g8f6 f1c4 f8d6 g1f3 e8g8 e1g1 b8c6 c1g5 h7h6 g5h4 c8g4 h2h3 g4f3 d1f3 c6d4 f3b7 a8b8 b7a7 b8a8 a7b7 a8b8 b7a6 b8a8');
INSERT INTO evaluations VALUES(14320,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(14320,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(14320,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(14320,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(14320,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(14320,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(14320,36,'cp 34','e4d5 f6d5 d2d4 c8g4 f1e2 d5c3 b2c3 e7e6 h2h3 g4h5 a1b1 a8b8 f3e5 h5e2 d1e2 f8d6 e5c6 b7c6 b1b3 e8g8 e1g1 d8f6 c1e3 b8e8 b3b7 a7a5 b7a7 f6f5 c3c4 e8a8 a7a8 f8a8 c4c5 d6e7 f1b1 e7f6 b1b7 e6e5 c2c3 f5e4 b7c7 e5d4');
INSERT INTO evaluations VALUES(14187,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(14187,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(14187,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14187,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(14187,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(14187,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(14187,36,'cp 22','f6d5 d2d4 c8g4 f1b5 a7a6 b5e2 e7e6 e1g1 d5c3 b2c3 f8e7 a2a4 e8g8 a4a5 f8e8 c1f4 e7d6 d1d2 e6e5 f3e5 g4e2 d2e2 d6e5 f4e5 c6e5 d4e5 d8d5 f1e1 d5c6 e2e3 b7b5 h2h3 a8d8 a1b1 h7h6 f2f4 d8d5');
INSERT INTO evaluations VALUES(14429,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(14429,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(14429,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(14429,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(14429,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(14429,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(14429,36,'cp 17','d2d4 c8g4 f1b5 a7a6 b5e2 d5c3 b2c3 e7e6 e1g1 f8d6 a1b1 a8b8 h2h3 g4h5 c1g5 d6e7 g5e7 c6e7 c3c4 e8g8 f1e1 h5f3 e2f3 b7b5 d4d5 e6d5 c4b5 a6b5 a2a4 c7c6 a4b5 b8b5 b1b5 c6b5 d1e2 e7c6 e2b5 c6d4 b5d3 d4f3 d3f3 d8c7 f3d5');
INSERT INTO evaluations VALUES(14419,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(14419,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14419,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(14419,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(14419,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(14419,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(14419,36,'cp 21','c8g4 f1b5 a7a6 b5e2 e7e6 e1g1 d5c3 b2c3 f8d6 a1b1 a8b8 h2h3 g4h5 c1g5 d6e7 g5e3 e8g8 c3c4 b7b6 c2c3 d8d7 f1e1 b8d8 b1b2 e6e5 c4c5 e5e4 f3d2 h5e2 e1e2 b6c5 d2e4 c5d4 c3d4 c6b4 d1b3 b4d5 e3d2');
INSERT INTO evaluations VALUES(20178,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(20178,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(20178,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(20178,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(20178,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20178,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(20178,36,'cp 8','d5e4 c3e4 c8d7 e1g1 c6d4 f3d4 c5d4 d1e2 d7b5 e2b5 d8d7 b5d7 e8d7 f1d1 e6e5 c2c3 g8f6 e4g5 f8c5 c3d4 e5d4 b2b4 c5b6 a2a4 a7a5 c1b2 h8c8 b2d4 b6d4 d1d4 d7e7 g2g3 e7f8 h2h4 c8c7 b4a5 g7g6 a1b1 a8a5 d4b4 h7h6 g5e4 f6e4 b4e4 f8g7 g1g2 g6g5');
INSERT INTO evaluations VALUES(35080,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(35080,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(35080,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(35080,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(35080,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(35080,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(35080,36,'cp 164','b1c3 f6d8 f1d3 d7d6 g1f3 a7a6 d1e2 b8c6 a2a3 e6e5 c1e3 c8g4 d4e5 d6e5 e1c1 f8d6 h2h3 g4f3 e2f3 d8f6 f3e2 g8e7 c3d5 e7d5 e4d5 c6e7 c2c4 e7f5 c4c5 f5e3 e2e3 d6e7 h1e1 e8c8 d3c2 f6g5 e3g5 e7g5 c1b1 g5f6 g2g4');
INSERT INTO evaluations VALUES(35078,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(35078,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(35078,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(35078,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(35078,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(35078,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(35078,36,'cp 164','f8b4 g1e2 d7d6 c1e3 b4a5 d1d2 f6e7 f2f3 c8d7 e1c1 b8c6 c1b1 g8f6 h2h4 a7a6 e2c1 h7h5 c1b3 a5b6 e3f2 c6a5 c3a4 a5b3 a2b3 d7a4 b3a4 c7c6 f1d3 d6d5 e4e5 f6d7 a4a5 b6c7 g2g4 e8c8');
INSERT INTO evaluations VALUES(47028,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(47028,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47028,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47028,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(47028,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47028,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(47028,36,'cp 33','e7e5 d4e5 d6e5 d1d8 e8d8 f1c4 d8e8 h2h4 a7a5 a2a3 c7c6 h4h5 b8a6 b1c3 a6c5 c1d2 b7b5 c4a2 c8e6 a2e6 c5e6 a3a4 e6d4 e1c1 f8b4 a4b5 c6b5 g1e2 d4e2 c3e2 b4d2 d1d2 e8e7 e2c3 b5b4 c3a4 h8d8 a4b6 d8d2 c1d2 a8d8 d2e3 d8d4 b2b3');
INSERT INTO evaluations VALUES(47857,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(47857,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(47857,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(47857,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(47857,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(47857,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(47857,36,'cp 44','d4e5 d6e5 d1d8 e8d8 f1c4 d8e8 h2h4 f6d7 a2a3 h7h5 b1c3 c7c6 f3f4 e5f4 c1f4 d7c5 g1f3 c8e6 c4e2 e6g4 e1c1 f8e7 f3e5 g4e2 c3e2 b8d7 e2d4 d7e5 f4e5 f7f6 e5f4 g7g6 h1e1 a8d8 b2b4 c5a4 e4e5 f6f5 d4e6 d8d1 e1d1');
INSERT INTO evaluations VALUES(21576,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(21576,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(21576,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(21576,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(21576,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(21576,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(21576,36,'cp -31','d8d5 d2d4 c8f5 f1e2 e8c8 d4e5 c6e5 d1d5 d8d5 e1g1 g8f6 b1d2 f5d3 f1e1 d3e2 e1e2 e5f3 d2f3 d5d1 e2e1 d1e1 f3e1 f8c5 c1f4 h8d8 g1f1 h7h6 f1e2 g7g5 f4g3 d8e8 e2f1 f6e4 a1d1 h6h5 h2h4 e4g3 f2g3 c5d6 h4g5 d6g3 e1f3 c7c6 f3d4 c8c7 d4f3 h5h4 d1d2');
INSERT INTO evaluations VALUES(41844,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(41844,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(41844,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(41844,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(41844,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(41844,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(41844,36,'cp 179','e5f3 b8c6 d2d4 c5b4 d4d5 c6e7 f1d3 c7c6 d5c6 b7c6 e1g1 e8g8 f1e1 e7g6 h2h3 f8e8 c1d2 c8b7 a2a3 b4a5 c3a4 a5c7 c2c4 h7h6 a4c3 f6d7 d2e3 d8f6 d3f1 d7c5 d1d4 f6d4 e3d4 c5b3 a1d1 b3d4 f3d4');
INSERT INTO evaluations VALUES(41906,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(41906,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(41906,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(41906,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(41906,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(41906,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(41906,36,'cp 174','e8g8 f1e2 f8e8 d2d3 d6d5 e4d5 c5b4 e1g1 b4c3 b2c3 f6d5 c1d2 d8f6 d3d4 b8c6 h2h3 h7h6 a2a4 c8f5 a4a5 a8d8 f1e1 b7b6 a5b6 a7b6 e2f1 e8e1 f3e1 f6g6 e1d3 d5f6 d1f3 f5e4 f3e3 d8e8 e3f4 f6d5 f4g4 g6d6 a1e1 d5f6 d2f4');
INSERT INTO evaluations VALUES(37006,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(37006,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(37006,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(37006,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(37006,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(37006,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(37006,36,'cp 175','f1e2 f8e8 d2d3 a7a6 e1g1 h7h6 a2a4 c5a7 f1e1 f6g4 d3d4 c7c5 d4d5 c5c4 e1f1 g4f6 e2c4 f6e4 c3e4 e8e4 c4d3 e4e8 f1e1 b8d7 c2c4 e8e1 d1e1 d7c5 d3c2 c8g4 e1c3 g4f3 c3f3 d8h4 b2b3 h4d4 a1b1 a8e8 c1f4 c5d7 g2g3 a7c5 g1g2 d7e5 f3e2 g7g5 f4e3 d4c3 e3c5');
INSERT INTO evaluations VALUES(38457,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(38457,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38457,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(38457,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(38457,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(38457,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(38457,36,'cp 16','b1d2 g8f6 c2c3 e8g8 f1e2 b8d7 h2h3 c7c5 a2a4 a7a6 g1f3 b7b6 e1g1 c8b7 f1e1 f8e8 f3e5 d7e5 f4e5 g7f8 d2f3 e7e6 h3h4 d8e7 e5f4 f6e4 f3g5 e4d6 d1b3 e7d8 b3c2 f7f6 g5f3 c5c4');
INSERT INTO evaluations VALUES(17630,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(17630,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(17630,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(17630,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(17630,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(17630,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(17630,36,'cp 54','f1d3 c8d7 a2a3 d8b6 d3c2 a8c8 d4c5 f8c5 e1g1 b6c7 c1f4 g8e7 f4g3 c5a7 b1d2 c7d8 d2b3 e7f5 g3f4 h7h6 f1e1 g7g5 c2f5 e6f5 f4e3 a7e3 e1e3 e8g8 e3e1 f8e8 b3c5 g5g4 d1d5 d7e6 d5d8 e8d8 f3h4 c6a5 b2b4 a5c4 c5b7');
INSERT INTO evaluations VALUES(17631,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(17631,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(17631,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(17631,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(17631,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17631,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(17631,36,'cp 43','c8d7 a2a3 d8b6 d3c2 a8c8 d4c5 f8c5 e1g1 b6c7 c1f4 g8e7 f4g3 c5a7 b1d2 c6a5 a1c1 h7h6 c2b1 e8g8 f1e1 e7f5 b1f5 e6f5 e5e6 f5f4 g3f4 c7f4 e6d7 c8d8 d2b3 a5b3 d1b3 d8d7 c1d1 f8d8 f3d4 f4f6 g2g3 h6h5 h2h4 d7e7 e1e7 f6e7 g1g2 g7g6 d4f3 e7c5 d1d4 c5c6 d4b4 b7b5 a3a4 b5a4 b3a4');
INSERT INTO evaluations VALUES(20952,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(20952,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(20952,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(20952,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(20952,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(20952,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(20952,36,'cp 26','c6e5 g1f3 e5f3 d1f3 e7e6 f1c4 e6d5 e4d5 f8d6 e1g1 d8h4 c4b5 e8f8 h2h3 a7a6 b5d3 g8f6 f1e1 h7h5 d3f1 h4d4 f3b3 c8d7 b3b7 a8e8 e1e8 d7e8 b7b3 d4e5 g2g3 h5h4 c1f4 e5d4 f4d6 c7d6 b3c3 d4d5 g3g4 e8c6 f2f3 d5f3 c3f3 c6f3 f1a6 f3c6 b1c3 f8e7 a2a4 d6d5 a1d1 e7d6');
INSERT INTO evaluations VALUES(21215,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(21215,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(21215,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(21215,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(21215,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(21215,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(21215,36,'cp 19','g1f3 e5f3 d1f3 e7e6 f1c4 e6d5 c4d5 g8f6 b1c3 f8d6 d5b3 d8e7 c1g5 c8e6 e1c1 e8c8 c3d5 e6d5 b3d5 c7c6 f3f5 d8d7 d5f7 e7f7 d1d6 f7c4 c1b1 c4e4 f5e4 f6e4 d6d7 c8d7 g5e3 a7a6 h1d1 d7e6 b1c2 e4f6 a2a4 f6d5 e3c5 h7h5 d1e1 e6f7 c2b3 h5h4 a4a5 h8h5');
INSERT INTO evaluations VALUES(21214,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(21214,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21214,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21214,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21214,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(21214,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(21214,36,'cp 37','e5f3 d1f3 e7e6 f1c4 e6d5 e4d5 f8d6 e1g1 d8h4 c4b5 e8f8 g2g3 c8g4 f3b3 h4h5 f1e1 g8f6 b1d2 h5d5 b5c4 d5f5 c4d3 f5d5 d2c4 g4e6 d3f1 h7h5 c4d6 d5b3 a2b3 c7d6 f1g2 d6d5 c1e3 f8e7 e3a7 h8c8 b3b4 e7d7 g2f1 f6e4 f1b5 d7e7 f2f3 e4d6 b5d3 g7g6 a7c5 a8a1 e1a1 e7d7 g1f2');
INSERT INTO evaluations VALUES(16546,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(16546,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(16546,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(16546,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(16546,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(16546,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(16546,36,'cp -32','d7d5 b1d2 f8c5 c2c3 a7a5 f1e2 e8g8 e1g1 c8e6 e4d5 e6d5 f1e1 h7h6 d2f1 c5d6 f1e3 d5e6 d3d4 e5d4 c3d4 d6b4 e1f1 a5a4 e2d3 a4a3 d3c2 a3b2 c1b2 b4a3 d1c1 a3d6 f1e1 f8e8 f3e5 f6d5 a2a3 d5e3 e1e3');
INSERT INTO evaluations VALUES(15154,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(15154,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(15154,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(15154,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(15154,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(15154,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(15154,36,'cp -19','b1d2 f8c5 f1e2 h7h6 e1g1 e8g8 c2c3 a7a5 f1e1 c8e6 e4d5 e6d5 d2f1 f8e8 c1e3 c5d6 f1g3 d8d7 d1c2 d5e6 g3e4 f6d5 e4d6 d5e3 f2e3 c7d6 d3d4 d6d5 f3e5 c6e5 d4e5 e6f5 e5e6 f7e6 e2d3 f5d3 c2d3 a8d8 e3e4 d5d4 a1d1');
INSERT INTO evaluations VALUES(17295,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(17295,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(17295,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(17295,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(17295,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(17295,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(17295,36,'cp 14','e7e5 c2c4 d7d5 c4d5 f6d5 e1g1 d5b6 d2d3 f8e7 b1c3 e8g8 a2a3 a7a5 c1e3 c8e6 e3b6 c7b6 e2e3 a8c8 f1e1 h7h6 d3d4 e5d4 f3d4 c6d4 e3d4 e7f6 d4d5 e6d7 a1c1 c8c4 c3e4 c4d4 d1f3 f6e5 c1d1 d8c7 d1d4 e5d4 d5d6');
INSERT INTO evaluations VALUES(22549,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(22549,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(22549,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(22549,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(22549,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(22549,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(22549,36,'cp 75','c4b3 g8f6 b1c3 f8b4 f2f4 e5c6 d4c6 d7c6 d1d8 e8d8 c1d2 c8e6 e1c1 d8e8 d1e1 e6b3 c2b3 f6d7 a2a3 b4e7 c1c2 a8d8 h1f1 f7f6 c3d1 e8f7 d1e3 d7c5 d2c3 h8e8 e3f5 e7f8 e4e5 f6e5 f4e5 f7g8 f5d4 c5e6 d4e6 e8e6 g2g3');
INSERT INTO evaluations VALUES(22550,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(22550,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(22550,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(22550,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(22550,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(22550,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(22550,36,'cp 80','g8f6 b1c3 f8b4 f2f4 e5c6 d4c6 d7c6 d1d8 e8d8 c1d2 c8e6 e1c1 e6b3 a2b3 d8e7 e4e5 f6d7 c3d5 c6d5 d2b4 c7c5 b4e1 e7e6 g2g4 d5d4 e1g3 f7f5 e5f6 d7f6 h1e1 e6f7 b3b4 f6g4 b4c5 a8d8 e1e4 h6h5 h2h3 g4e3 d1d4 d8d4 e4d4 e3f5 d4d7 f7e6 d7d3 h8e8 g3f2');
INSERT INTO evaluations VALUES(16211,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(16211,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(16211,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(16211,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(16211,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(16211,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(16211,36,'cp 75','b1c3 f8b4 f2f4 e5c6 d4c6 d7c6 d1d8 e8d8 c1d2 c8e6 e1c1 e6b3 a2b3 d8e7 e4e5 f6d7 c3e4 b4d2 c1d2 a7a5 d2e3 h8d8 g2g4 f7f6 e5f6 d7f6 e4f2 f6d5 e3f3 a5a4 b3a4 a8a4 h1e1 e7f8 f2e4 d8e8 c2c3 a4a2 d1d2 g7g5 f4g5 h6g5 e1h1 e8e5 h2h3 f8e7');
INSERT INTO evaluations VALUES(16212,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(16212,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(16212,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(16212,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(16212,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(16212,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(16212,36,'cp 67','f8b4 f2f4 e5c6 d4c6 d7c6 d1d8 e8d8 c1d2 c8e6 e1c1 e6b3 a2b3 d8e7 e4e5 f6d7 c3d5 c6d5 d2b4 c7c5 b4e1 e7e6 g2g4 d5d4 b3b4 h6h5 b4c5 h5g4 b2b4 a7a5 d1d4 a5b4 f4f5 e6e7 d4b4 d7e5 e1c3 f7f6 b4b7 e7f8 h1g1 h8h4 g1g2 a8c8 g2d2 e5c4');
INSERT INTO evaluations VALUES(9904,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(9904,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(9904,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(9904,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(9904,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(9904,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(9904,36,'cp 67','f2f4 e5c6 d4c6 d7c6 d1d8 e8d8 c1d2 c8e6 e1c1 e6b3 a2b3 d8e7 e4e5 f6d7 d2e3 h8d8 c3e2 e7e8 c2c3 b4c5 e3d2 a7a5 c1c2 d7b6 e2g3 b6d5 g3f5 c5f8 h1e1 a5a4 f5g3 a4a3 b2a3 a8a3 d1a1 d8a8 a1a3 a8a3');
INSERT INTO evaluations VALUES(51052,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51052,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(51052,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(51052,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(51052,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51052,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(51052,36,'cp 63','g1f3 d7d6 d2d4 e5d4 f3d4 g8f6 b1c3 f8e7 e1g1 e8g8 c1f4 f8e8 a2a4 b8c6 f1e1 e7f8 d4c6 b7c6 e4e5 c8g4 f2f3 f6h5 f3g4 h5f4 d1f3 d8g5 g2g3 f4e6 e1f1 d6d5 f3f7 g8h8 c4a6 e6c5 a6d3 g5e5 g1g2 c5d3 c2d3 f8d6 f7f5 e5e3 f5f2 d6b4 g2h1 e3f2 f1f2');
INSERT INTO evaluations VALUES(51049,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51049,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(51049,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(51049,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(51049,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(51049,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(51049,36,'cp 50','d7d6 e1g1 f8e7 d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 c1f4 f8e8 a2a4 b8c6 d4c6 b7c6 f1e1 c8g4 d1d3 a6a5 f2f3 g4h5 a1d1 f6d7 d3d2 e7h4 e1f1 d8e7 b2b3 d7b6 c4d3 d6d5 g1h1 d5e4 c3e4 h5g6 f4g3 h4g3 h2g3 g6e4');
INSERT INTO evaluations VALUES(3804,30,'cp -230',NULL);
INSERT INTO evaluations VALUES(3804,31,'cp -211',NULL);
INSERT INTO evaluations VALUES(3804,32,'cp -218',NULL);
INSERT INTO evaluations VALUES(3804,33,'cp -217',NULL);
INSERT INTO evaluations VALUES(3804,34,'cp -230',NULL);
INSERT INTO evaluations VALUES(3804,35,'cp -233',NULL);
INSERT INTO evaluations VALUES(3804,36,'cp -225','c8g4 d1d2 h7h6 h2h3 g4h5 g2g4 h6g5 g4h5 f6h5 d2g5 d8g5 c1g5 c6d4 g1g2 c7c6 c4b3 a7a5 a1e1 b7b5 a2a4 d4b3 c2b3 b5a4 c3a4 c5d4 e1c1 a8a6 c1a1 a6a7 a1a3 a7b7 a4c3 f8b8 g2f3 b7b3 a3a5 b3b2 c3e2 d4b6 a5a6 d6d5 e4d5 c6d5 e2c3 f7f6');
INSERT INTO evaluations VALUES(23727,30,'cp -211',NULL);
INSERT INTO evaluations VALUES(23727,31,'cp -202',NULL);
INSERT INTO evaluations VALUES(23727,32,'cp -213',NULL);
INSERT INTO evaluations VALUES(23727,33,'cp -220',NULL);
INSERT INTO evaluations VALUES(23727,34,'cp -217',NULL);
INSERT INTO evaluations VALUES(23727,35,'cp -201',NULL);
INSERT INTO evaluations VALUES(23727,36,'cp -200','d1d2 h7h6 h2h3 g4h5 g2g4 h6g5 g4h5 f6h5 d2g5 d8g5 c1g5 c6d4 g1g2 c7c6 c4b3 a7a5 a1e1 g7g6 a2a4 h5f4 g5f4 e5f4 c3a2 a8a6 b3c4 b7b5 c2c3 f4f3 g2g3 b5c4 c3d4 c5d4 d3c4 a6b6 e1b1 b6b3 a2c1 d4e5 g3h4 b3b4 b2b3 f8b8 h4g4 b4c4 c1d3 c4d4 f1d1 d4e4 g4f3 e4d4 d3e5 d4d1 b1d1 d6e5 d1d6 b8b3 f3e4 b3h3 e4e5 h3a3 d6c6 a3a4');
INSERT INTO evaluations VALUES(19089,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(19089,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19089,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19089,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19089,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19089,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19089,36,'cp 0','d7d5 d1d5 d8d5 e4d5 c8f5 d5c6 b7c6 f1d3 f5d3 c2d3 g8f6 e1e2 f6d5 b1c3 e8d7 c1d2 a8b8 c3a4 d5b6 a4c3');
INSERT INTO evaluations VALUES(20557,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20557,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20557,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20557,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20557,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20557,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20557,36,'cp 0','d1d5 d8d5 e4d5 c8f5 d5c6 b7c6 f1d3 f5d3 c2d3 g8e7 b1c3 a8b8 c3a4 e7d5 e1e2 e8d7 c1e3 d5b6 a4c3 b6d5');
INSERT INTO evaluations VALUES(50970,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50970,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(50970,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50970,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(50970,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(50970,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(50970,36,'cp 28','c7c5 c2c3 b8c6 a2a3 g8e7 f1d3 e7f5 h2h4 f8e7 d4c5 e7c5 d1e2 e8g8 g2g4 f5e7 g4g5 e7f5 b1d2 c5a7 d2f1 c8d7 f1e3 a7e3 c1e3 c6e7 e1c1 d7a4 d1d2 a4b5 d3b5 a6b5 f3d4 d8c7 h4h5 c7e5 d4b5 e7c6 e3c5');
INSERT INTO evaluations VALUES(50790,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(50790,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(50790,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(50790,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(50790,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(50790,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(50790,36,'cp 52','c2c3 c8d7 f1d3 b8c6 a2a3 g8h6 e1g1 h6f5 d4c5 f8c5 b1d2 h7h6 d2b3 c5a7 f1e1 e8g8 c1f4 d8b6 d1c2 c6e7 a3a4 e7g6 f4g3 b6d8 c2d2 g6h4 f3h4 f5h4 g1h1 h4f5 d3f5 e6f5 a4a5 a8c8 f2f3 f8e8 g3f2 a7f2 d2f2 d8c7');
INSERT INTO evaluations VALUES(9849,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9849,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9849,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(9849,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9849,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(9849,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(9849,36,'cp 21','d7d6 a2a3 c8e6 h1f1 a7a6 d1e2 e6c4 d3c4 c6b8 e1c1 b8d7 f3d2 d8e7 c3d5 e7d8 g2g4 f6d5 c4d5 d8e7 f1f2 e8c8 d1f1 d8f8 b2b4 g7g6 a3a4 h6h5 a4a5 h5g4 e2g4 h8h4 g4g2 h4h8 c1b2 c8b8 b2b3');
INSERT INTO evaluations VALUES(8887,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8887,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(8887,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(8887,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(8887,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8887,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(8887,36,'cp 0','a2a3 c8e6 h1f1 e6c4 d3c4 c6b8 f3h4 b8d7 d1f3 e8g8 h4f5 g8h7 f3h3 f6g8 e1c1 d7c5 f5h4 g7g6 h4f3 d8e7 g2g4 c7c6 g4g5 h6h5 b2b4 c5e6 h3g3 a8d8 c4c5 d6c5 b4b5 d8d1 f1d1 a7a6 b5a6 b7a6 g3e5 f7f6 g5f6 e7f6 e5f6 g8f6 e4e5');
INSERT INTO evaluations VALUES(28823,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(28823,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(28823,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(28823,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(28823,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(28823,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(28823,36,'cp 80','e4e5 f6d5 c3d5 b7d5 d2d4 e7e6 c1f4 f8e7 h2h4 b8c6 c2c3 f7f6 f1d3 e8g8 e5f6 e7f6 f3g5 f6g5 f4g5 d8e8 d1c2 h7h6 d3h7 g8h8 h7g6 f8f7 g6f7 e8f7 f2f3 e6e5 g5e3 d5a2 e1g1 e5d4 e3d4 c6d4 c3d4 c7c6 f1e1 a2d5 c2d2 f7f6 e1e5 f6h4 a1e1 h4f6 e5e7 a8d8');
INSERT INTO evaluations VALUES(28786,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(28786,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(28786,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(28786,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(28786,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(28786,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(28786,36,'cp 89','f6d5 c3d5 b7d5 d2d4 e7e6 c1f4 f8e7 h2h4 d7d6 f1d3 b8c6 a2a3 f7f5 d1e2 e8g8 h4h5 h7h6 e1c1 d6e5 d4e5 d8e8 c1b1 c6a5 f3h4 e7h4 h1h4 a8d8 h4h3 e8a4 f4e3 a5c6 h3g3 c6d4 e3d4 a4d4 g3g6 d4f4 g2g3');
INSERT INTO evaluations VALUES(28834,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(28834,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(28834,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(28834,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(28834,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(28834,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(28834,36,'cp 90','c3d5 b7d5 d2d4 e7e6 c1f4 b8c6 f1e2 d7d6 e1g1 d6e5 d4e5 c6e7 c2c3 e7g6 f4g3 f8e7 d1c2 e8g8 a1d1 d8c8 h2h4 f8d8 h4h5 g6f8 h5h6 f8g6 c2c1 c8b7 h6g7 d5a2 c1h6 d8d1 f1d1 a8d8 d1c1 a2d5 g3h4 e7h4 f3h4 d5e4 f2f3 e4c6 h4g6 f7g6');
INSERT INTO evaluations VALUES(28832,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(28832,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(28832,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(28832,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(28832,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(28832,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(28832,36,'cp 85','b7d5 d2d4 e7e6 c1f4 b8c6 f1e2 c6a5 e1g1 c7c5 d4c5 b6c5 d1d2 f8e7 a1d1 e8g8 b2b3 d8c7 c2c3 c7b6 c3c4 d5f3 e2f3 a8d8 d2e2 a5c6 f3c6 d7c6 f4d2 d8d4 d2c3 d4d1 e2d1 f8d8 d1f3 h7h6 g2g3 b6b7 f1d1 d8d1 f3d1 b7a6 d1d7 a6a2 d7e7');
INSERT INTO evaluations VALUES(28606,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(28606,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(28606,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(28606,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(28606,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(28606,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(28606,36,'cp 84','d2d4 e7e6 c1f4 b8c6 f1e2 c6a5 e1g1 c7c5 d4c5 b6c5 b2b3 f8e7 d1d2 a5c6 c2c3 a7a5 d2e3 e8g8 a1d1 a5a4 f3d2 c6a5 e3g3 e7h4 g3g4 f7f5 e5f6 h4f6 b3a4 d8e7 g4g3 a5b7 d2c4 d5c4 e2c4 a8a4 c4b3 a4e4');
INSERT INTO evaluations VALUES(28604,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(28604,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(28604,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(28604,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(28604,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(28604,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(28604,36,'cp 77','e7e6 c1f4 b8c6 f1e2 d7d6 e1g1 d6e5 f4e5 c6e5 f3e5 f8d6 e2b5 e8f8 c2c4 d5b7 d1h5 g7g6 h5h6 f8g8 a1d1 d8f8 h6e3 h7h5 f1e1 d6e5 e3e5 a7a6 b5a4 f8d6 e5d6 c7d6 d4d5 e6d5 c4d5 b6b5 a4b3 g8g7 e1e7 b7c8 d1c1');
INSERT INTO evaluations VALUES(28601,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(28601,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(28601,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(28601,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(28601,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(28601,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(28601,36,'cp 59','c1f4 d7d6 f1b5 c7c6 b5d3 d6e5 d4e5 b8d7 e1g1 b6b5 c2c3 f8e7 f1e1 d7f8 f4g3 f8g6 d1e2 h7h5 h2h3 h5h4 g3h2 e8g8 a1d1 a7a6 f3d4 e7c5 d4c6 d8b6 c6b4 c5b4 c3b4 a6a5 a2a3 a5b4 a3b4 b6b7 e2g4 a8a4 d1d2 b7e7 d3g6 f7g6 g4g6');
INSERT INTO evaluations VALUES(5315,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(5315,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(5315,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(5315,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(5315,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(5315,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(5315,36,'cp -13','d2d3 d7d6 f1e2 c8g4 f3g1 g4d7 g1f3 g8f6 h2h3 e8g8 c1e3 e7e5 a2a3 f6e8 b2b4 b7b6 a1b1 e8c7 c3d5 c7e6 h3h4 h7h6 d1d2 g8h7 g2g3 e6c7 b4c5 d6c5 e1g1 f7f5 f1e1 d7c8 d2c2 a8b8 b1d1 c7d5 c4d5 c6d4 e3d4 c5d4');
INSERT INTO evaluations VALUES(5319,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(5319,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(5319,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(5319,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(5319,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(5319,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(5319,36,'cp -22','d7d6 f1e2 c8g4 f3g1 g4d7 g1f3 g8f6 h2h3 e8g8 c1e3 e7e5 a2a3 b7b6 e1g1 f6e8 b2b4 e8c7 b4c5 b6c5 a1b1 c6d4 e3d2 a8b8 b1b8 d8b8 f3d4 e5d4 c3d5 c7d5 e4d5 b8b2 d2f4 g7e5 f4e5 d6e5 e2f3 b2a3 d5d6');
INSERT INTO evaluations VALUES(5484,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(5484,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(5484,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(5484,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(5484,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(5484,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(5484,36,'cp -19','f1e2 e7e6 e1g1 g8e7 c1e3 c6d4 d1d2 e8g8 e3h6 e7c6 h6g7 g8g7 f3d4 c6d4 a1e1 c8d7 f2f4 a8b8 e2d1 b7b5 f1f2 d7c6 c3e2 d8f6 e2g3 h7h6 e1f1 b5b4 f4f5 f6g5 f5g6 f7g6 d2g5 h6g5 f2f8 b8f8 f1f8 g7f8 g3e2 e6e5 e2d4 e5d4 a2a3 a7a5 a3b4');
INSERT INTO evaluations VALUES(5483,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(5483,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(5483,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(5483,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(5483,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(5483,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(5483,36,'cp -14','c8g4 f3g1 g4d7 g1f3 e7e6 c1f4 d7c8 e1g1 g8e7 f4e3 c6d4 d1d2 e8g8 e3h6 e7c6 h6g7 g8g7 f3d4 c6d4 a1e1 c8d7 f2f4 a8b8 e2d1 d7c6 c3e2 d8f6 e2g3 h7h6 f1f2 b7b5 e1f1 b5b4 f4f5 f6g5 d2g5 h6g5 f5g6 f7g6 f2f8 b8f8 f1f8');
INSERT INTO evaluations VALUES(32293,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32293,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(32293,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32293,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32293,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(32293,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32293,36,'cp 0','f5d3 c2d3 c7c5 d4c5 f8c5 g1f3 b8c6 e1g1 g8f6 d3d4 c5e7 h2h3 e8g8 b1c3 a8c8 a1c1 a7a6 f3e1 h7h6 d1e2 f8e8 e1d3 c6b4 d3e5 b4c6');
INSERT INTO evaluations VALUES(32339,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32339,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32339,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32339,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32339,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32339,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32339,36,'cp 0','c2d3 c7c5 d4c5 f8c5 g1f3 b8c6 e1g1 g8f6 d3d4 c5e7 b1c3 e8g8 d1e2 h7h6 a1c1 a8c8 f3e1 d8a5 e1d3 f8d8 h2h3 c6b4 d3e5 b4c6');
INSERT INTO evaluations VALUES(33900,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(33900,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(33900,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(33900,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(33900,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(33900,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(33900,36,'cp 73','g1f3 h7h6 f1d3 f8d6 d1e2 b8c6 c2c3 c8d7 e1g1 e8g8 f1e1 c6e7 f3e5 d6e5 d4e5 f6h4 g2g3 h4h3 e2e4 f7f5 e5f6 f8f6 e4h7 g8f7 d3f1 h3g4 h7e4 g4e4 e1e4 d7c6 e4e1 g7g5 f2f4 g5g4 c1e3 e7f5 a1d1 h6h5 f1g2 f5e3 e1e3 c6g2 g1g2');
INSERT INTO evaluations VALUES(50956,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(50956,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(50956,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(50956,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(50956,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(50956,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(50956,36,'cp 42','c2c3 c7c5 b1d2 g8f6 g1f3 b8d7 a2a4 f8e7 h2h4 e8g8 f1d3 d8b6 d1b1 e7d6 f3e5 c5d4 c3d4 d6b4 e5f3 b4d6 a4a5 b6b4 f4d6 b4d6 e1g1 h7h6 f1c1 e6e5 e3e4 e5d4 a1a4 b7b5 a5b6 d6b6 a4d4 d5e4 d3e4 c8b7 e4b7 b6b7');
INSERT INTO evaluations VALUES(50954,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(50954,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(50954,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(50954,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(50954,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(50954,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(50954,36,'cp 23','g8f6 b1d2 f8e7 g1f3 c7c5 h2h3 b8d7 a2a4 b7b6 f1d3 c8b7 d3c2 e8g8 e1g1 f6e4 d1e2 f7f5 f3e5 d7e5 f4e5 e4d2 e2d2 a6a5 e5f4 e7d6 a1b1 b7a6 f1e1 d8e7 c2d3 a6d3 d2d3 h7h6 d3e2 d6c7 e2f3 a8c8 f3g3 g8h8 f4c7 c8c7');
INSERT INTO evaluations VALUES(43655,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43655,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(43655,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(43655,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(43655,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43655,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(43655,36,'cp 25','b1d2 c7c5 g1f3 b8d7 f3e5 d7e5 d4e5 f6d7 e3e4 c5c4 f1e2 f8e7 e4d5 e6d5 d2f3 d7c5 e1g1 c8f5 f3d4 f5e4 f4e3 e8g8 f2f3 e4g6 f3f4 g6e4 e2f3 e4d3 f1e1 f8e8 a2a4 a6a5 f4f5 e7f8 e3f4 d8b6 f3d5 b6b2 d4b5 d3f5 d5c4');
INSERT INTO evaluations VALUES(43654,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(43654,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43654,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43654,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(43654,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(43654,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(43654,36,'cp 21','c7c5 g1f3 b8d7 f1d3 d8b6 d1c2 f8d6 e1g1 e8g8 a2a4 d6f4 e3f4 a6a5 g2g3 b6c7 f3e5 b7b6 d2f3 c8a6 d3a6 a8a6 a1d1 c5d4 e5d7 c7d7 d1d4 a6a8 f3e5 d7b7 c2e2 f8c8 f1e1 c8c7 h2h4 a8c8');
INSERT INTO evaluations VALUES(39283,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(39283,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(39283,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(39283,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(39283,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(39283,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(39283,36,'cp 3','c2c4 c7c6 b1c3 g8f6 c3a4 c5e7 g1f3 e5e4 f3d4 d7d5 d2d3 e4d3 d1d3 d5c4 d3c4 e8g8 a4c3 b8d7 e1g1 d7e5 c4a4 d8b6 d4b3 c8d7 a4d4 b6d4 b3d4 e5c4 f1d1 f8e8 b2b3 e7b4 c3a4 c4b6 c1d2 b4f8 a4c3 h7h6 e2e3 a8d8 h2h3 c6c5');
INSERT INTO evaluations VALUES(46504,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(46504,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(46504,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(46504,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(46504,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(46504,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(46504,36,'cp -40','d2d3 d5d4 c3e4 f6e4 g2e4 f8d6 f2f4 e8g8 f4e5 d6e5 g1f3 b8d7 f3e5 d7e5 c1f4 f8e8 a1c1 g7g5 f4e5 e8e5 e1g1 c8e6 f1f2 a8c8 c1c8 d8c8 d1d2 c8d7 h2h4 h7h6 a2a3 d7e7 f2f1 a7a5 f1f2 a5a4 h4g5 h6g5');
INSERT INTO evaluations VALUES(46505,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(46505,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(46505,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(46505,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(46505,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(46505,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(46505,36,'cp -65','d5d4 c3e4 f6e4 g2e4 h7h6 d1b3 b8c6 f2f4 f8d6 g1f3 c6a5 b3a4 c8d7 a4d1 e5f4 c1f4 d7c6 e1g1 e8g8 a1c1 d6f4 g3f4 d8b6 e4c6 a5c6 d1b3 f8e8 f1f2 b6b3 a2b3 a8d8 c1c5 d8d7 g1f1 f7f6 f1e1 e8d8 c5b5 g8f7 f4f5 c6e7 f3d2 a7a6');
INSERT INTO evaluations VALUES(51325,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(51325,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(51325,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(51325,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(51325,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(51325,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(51325,36,'cp 64','a2a4 c7c6 a4b5 c6b5 b1c3 d8b6 b2b3 e7e6 b3c4 f8b4 c1d2 b5c4 e2e3 g8f6 f1c4 a7a5 e1g1 e8g8 c3a4 b6c7 d2b4 a5b4 d1b3 b8c6 f1c1 c8d7 b3b2 f8c8 a4c5 a8a1 b2a1 f6d5 c4d5 e6d5 f3e1 d7f5 h2h3 c7a5 a1b2 h7h6 e1d3');
INSERT INTO evaluations VALUES(51334,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(51334,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(51334,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(51334,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(51334,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(51334,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(51334,36,'cp 45','c7c6 a4b5 c6b5 b2b3 e7e6 b3c4 b5c4 e2e3 g8f6 f1c4 f8d6 e1g1 e8g8 c1a3 c8b7 b1d2 a7a5 d1b3 a8a7 f1b1 d6a3 a1a3 b8d7 b3a2 d8c7 c4d3 f8c8 h2h3 h7h6 a2a1 b7c6 b1c1 a5a4 d2c4 c7b7 c4d6');
INSERT INTO evaluations VALUES(51662,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(51662,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51662,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(51662,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(51662,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(51662,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(51662,36,'cp 52','b2b3 e7e6 a4b5 c6b5 b3c4 b5c4 e2e3 g8f6 f1c4 f8e7 f3e5 d8c7 d1a4 c8d7 e5d7 b8d7 e1g1 e8g8 c4a6 d7b6 a4a2 b6d5 c1d2 e7b4 f1c1 c7d6 a6b7 a8b8 b7d5 f6d5 a2a7 b8c8 c1c8 f8c8 a7b7 c8b8 a1a8');
INSERT INTO evaluations VALUES(6475,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(6475,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(6475,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(6475,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(6475,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(6475,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(6475,36,'cp -60','g8f6 b1c3 e8g8 f1d3 a7a5 c3a4 c5d4 b2d4 e5d4 e1g1 d7d5 e4d5 d8d5 h2h3 c8f5 d3c4 d5d6 a4b2 f6d5 c4d3 f5g6 d3g6 h7g6 a2a3 d6f6 b2d3 f8e8 f1e1 a8d8 e1e8 d8e8 d1f1 f6d6 a1e1 e8e1 f1e1');
INSERT INTO evaluations VALUES(10202,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(10202,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(10202,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(10202,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(10202,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(10202,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(10202,36,'cp -58','b1c3 e8g8 f1d3 a7a6 c3e2 d7d5 e2g3 d5e4 g3e4 c5b6 e1g1 c8f5 e4f6 d8f6 d3f5 f6f5 d2d3 f7f6 f1e1 g7g5 d1e2 a8d8 a1d1 f8f7 c2c3 g5g4 f3h4 f5g5 g2g3 f6f5 d3d4 e5d4 h4g2 g5e7 e2e7 f7e7 e1e7 c6e7 c3d4 g8f7 g2f4 c7c6 a2a4 e7d5 f4d3 b6c7 d1c1');
INSERT INTO evaluations VALUES(51172,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(51172,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(51172,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(51172,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(51172,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(51172,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(51172,36,'cp 125','d2d4 e7e6 c2c3 c8b7 f1d3 c7c5 e1g1 g8f6 e4e5 f6d5 d4c5 f8c5 a2a4 b5b4 b1d2 f7f5 e5f6 d5f6 f3g5 e8g8 g5h7 f6h7 d3h7 g8h7 d1h5 h7g8 h5c5 b8c6 d2c4 f8f5 c5b6 d8b6 c4b6 a8d8 c1d2 b4c3 d2c3 c6e7 f2f3 e7d5 b6d5 f5d5 f1d1 b7c6 d1d5');
INSERT INTO evaluations VALUES(55710,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(55710,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55710,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(55710,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(55710,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(55710,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(55710,36,'cp 12','d2d4 d6d5 e4d5 e6d5 c4d3 f8d6 g1f3 g8e7 e1g1 e8g8 c1g5 b8c6 c2c3 f7f6 g5h4 g7g5 h4g3 e7g6 d1b3 g8g7 b1d2 c6e7 f1e1 g5g4 f3h4 g6h4 g3h4 e7g6 d3g6 h7g6 d2f1 f8h8 h4g3 d6g3 f2g3 c7c6 e1e2 h8e8 a1e1 e8e2 e1e2 b7b6 e2e1');
INSERT INTO evaluations VALUES(55711,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(55711,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(55711,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(55711,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(55711,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(55711,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55711,36,'cp 14','d6d5 e4d5 e6d5 c4b3 f8d6 g1f3 g8f6 d1e2 d8e7 e2e7 e8e7 e1g1 c8e6 f1e1 a7a5 c2c3 h7h6 b3c2 a5a4 f3h4 e7d8 h4f5 e6f5 c2f5 h8e8 e1e8 f6e8 g2g3 b8d7 c1e3 d7b6 b1d2 d6e7 f5d3 e8d6 a1b1 a4a3 b2b3 c7c6 g1h1 d8d7 e3f4');
INSERT INTO evaluations VALUES(22568,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22568,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22568,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22568,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22568,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22568,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22568,36,'cp 0','a7a6 b5c6 d7c6 e1g1 c6c5 c2c3 b7b6 d2d4 g8f6 e4e5 f6d5 d4c5 f8c5 f3g5 e8g8 d1h5 h7h6 g5e4 f7f5 e4c5 b6c5 c1h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(17881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17881,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17881,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17881,36,'cp 0','b5c6 d7c6 e1g1 c6c5 c2c3 b7b6 d2d4 g8f6 e4e5 f6d5 d4c5 f8c5 f3g5 e8g8 d1h5 h7h6 g5e4 f7f5 e4c5 b6c5 c1h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(17877,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17877,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17877,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17877,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17877,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17877,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17877,36,'cp 0','d7c6 e1g1 c6c5 c2c3 b7b6 d2d4 g8f6 e4e5 f6d5 d4c5 f8c5 f3g5 e8g8 d1h5 h7h6 g5e4 f7f5 e4c5 b6c5 c1h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(17339,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(17339,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(17339,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(17339,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(17339,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(17339,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(17339,36,'cp 78','b5c6 b7c6 d2d3 d7d5 e1g1 g8e7 c3a4 e7g6 f1e1 f7f6 b2b3 e6e5 c1a3 g6f4 a3c5 f8c5 a4c5 e8g8 d3d4 c8g4 h2h3 g4h5 e4d5 c6d5 d4e5 f6e5 e1e5 f4h3 g2h3 h5f3 d1d4 d8f6 g1h2 f6d6 c5d3 f3e2 a1g1 e2d3 c2d3 a8e8 f2f4 e8e5 f4e5');
INSERT INTO evaluations VALUES(17326,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(17326,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(17326,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(17326,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(17326,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(17326,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(17326,36,'cp 75','b7c6 d2d3 d7d5 e1g1 g8e7 b2b3 e7g6 c1a3 f8e7 f1e1 e8g8 a1c1 e6e5 e4d5 c6d5 f3e5 g6e5 e1e5 c8e6 d3d4 e7f6 e5e1 c5d4 c3e2 f8e8 a3c5 e6g4 f2f3 g4d7 d1d2 f6g5 f3f4 d4d3 c2d3 g5h4 e2g3 e8e1 c1e1 a6a5 e1e5 a5a4 e5d5');
INSERT INTO evaluations VALUES(18143,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(18143,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(18143,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(18143,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(18143,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(18143,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(18143,36,'cp 93','d2d3 g8e7 h2h4 d7d6 h4h5 e6e5 f3d2 c8e6 d2c4 f7f6 c3a4 e7c8 h5h6 g7g6 c1d2 f8e7 d2a5 d8d7 a4b6 c8b6 c4b6 d7c7 d1d2 a8b8 b6d5 c7d7 d5e7 b8b2 a5c3 b2b7 e7c6 d7c6 f2f4 b7f7 e1g1 e8g8 f1f2 e5f4 d2f4 d6d5 a1f1 d5e4 d3e4 e6a2');
INSERT INTO evaluations VALUES(18144,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(18144,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(18144,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(18144,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(18144,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(18144,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(18144,36,'cp 67','d7d5 d1e2 g8f6 c1g5 h7h6 g5f4 f8e7 f3e5 c8b7 e1g1 e8g8 b2b3 c5c4 b3c4 e7d6 f4g3 d8a5 c3d1 a5c7 e5g4 f6g4 e2g4 f7f5 g3d6 c7d6 e4f5 e6f5 g4f3 f5f4 a1b1 b7c8 d1c3 d5c4 d3c4 c8e6 b1b6 a8c8 f1d1 d6c5 b6a6 c5c4');
INSERT INTO evaluations VALUES(18229,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(18229,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(18229,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(18229,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(18229,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(18229,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(18229,36,'cp 67','e1g1 g8e7 b2b3 e7g6 c1a3 f8e7 c3a4 e8g8 a3c5 e7c5 a4c5 e6e5 f1e1 d8f6 d3d4 c8g4 e4d5 c6d5 d4e5 f6e7 d1d4 g4f3 g2f3 g6h4 d4g4 e7c5 g4h4 c5c2 e1d1 a8d8 h4d4 c2g6 g1h1 f8e8 f3f4 f7f6 d1d3 f6e5 f4e5 g6f5');
INSERT INTO evaluations VALUES(18228,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(18228,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(18228,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(18228,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(18228,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(18228,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(18228,36,'cp 85','g8f6 c1g5 h7h6 g5h4 f8e7 e4e5 f6d7 h4e7 d8e7 c3a4 e8g8 d1e1 c5c4 d3c4 a6a5 e1e3 c6c5 f3d2 c8b7 b2b3 d5d4 e3g3 a8e8 a1e1 e7g5 f2f4 g5g3 h2g3 b7c6 a4b2 f7f5 e5f6 g7f6 b2d3 e6e5 d2e4 c6e4 e1e4 f6f5 e4e2');
INSERT INTO evaluations VALUES(26907,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(26907,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(26907,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(26907,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(26907,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(26907,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(26907,36,'cp 140','b1c3 e5d4 d1d4 g8e7 d4e3 a7a6 b5a4 g4d7 a4b3 c6a5 c1d2 a5b3 a2b3 e7c6 c3d5 f8e7 d2c3 e8g8 e3f4 a6a5 h2h4 f7f6 h4h5 b7b5 e1g1 b5b4 c3d2 f8f7 f1e1 e7f8 f4g3 c6e7 c2c4 e7c6 d2e3 c6e5 f3h4 a5a4 b3a4 e5c4 b2b3');
INSERT INTO evaluations VALUES(51910,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(51910,31,'cp -148',NULL);
INSERT INTO evaluations VALUES(51910,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(51910,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(51910,34,'cp -141',NULL);
INSERT INTO evaluations VALUES(51910,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(51910,36,'cp -143','b8c6 c2c3 g8f6 h5e2 f6e4 e2e4 d7d5 c4d5 d8d5 e4d5 e6d5 d2d4 c5d4 g1f3 d4c3 b1c3 c8e6 c1e3 f7f6 e1g1 e8f7 f1d1 a8d8 d1d2 f8b4 a1d1 d8d7 f3d4 c6d4 d2d4 b4c3 b2c3 h8c8 d4d3 b7b6 e3d4 c8c4 d1a1 e6f5 d3e3 c4a4 a2a3 d7d8 f2f3 d8c8');
INSERT INTO evaluations VALUES(234,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(234,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(234,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(234,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(234,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(234,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(234,36,'cp -115','c5e3 f2e3 d7c6 d1d4 f6h4 g2g3 h4g5 b1d2 g8f6 e1c1 e8g8 h2h3 f8e8 f1c4 c8e6 c4e6 e8e6 g3g4 g5e5 h1f1 e5d4 e3d4 f6e4 d2e4 e6e4 d4d5 c6d5 d1d5 e4e7 g4g5 c7c6 d5d6 a8e8 f1d1 e7e5 d6d7 b7b6 b2b3 e5g5 d7a7 g5g2');
INSERT INTO evaluations VALUES(241,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(241,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(241,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(241,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(241,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(241,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(241,36,'cp -101','f2e3 d7c6 d1d4 c8g4 b1d2 a8d8 d4f6 g8f6 h2h3 g4e6 f1e2 h7h5 e2d3 e8e7 d2f3 f6d7 e1c1 c6c5 d3e2 h5h4 d1d2 d7b6 d2d8 e7d8 f3g5 d8e7 h1f1 b6d7 e2g4 d7f6 g4e6 f7e6 a2a4 a7a5');
INSERT INTO evaluations VALUES(240,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(240,31,'cp -140',NULL);
INSERT INTO evaluations VALUES(240,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(240,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(240,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(240,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(240,36,'cp -129','d7c6 d1d4 f6h4 g2g3 h4g5 e4e5 g8e7 f1c4 e7g6 e5e6 c8e6 c4e6 f7e6 b1d2 g5h5 h1f1 g6e5 f1f4 h8f8 d4e4 e8c8 f4h4 h5g5 h4f4 b7b6 e1c1 c8b7 d1f1 e5g6 f4f8 g6f8 f1f7 g5h5 e4f3 h5a5 f3e2 a5a2 f7g7 a2a1 d2b1 d8d7 g7g8 f8g6');
INSERT INTO evaluations VALUES(206,30,'cp -127',NULL);
INSERT INTO evaluations VALUES(206,31,'cp -139',NULL);
INSERT INTO evaluations VALUES(206,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(206,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(206,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(206,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(206,36,'cp -122','d1d4 f6h4 g2g3 h4g5 e4e5 g8e7 d4f4 g5h5 f1e2 h5h3 b1d2 e7g6 f4d4 e8g8 e1c1 h3e6 d2c4 b7b5 c4a5 g6e5 a2a4 a7a6 h1f1 e6e7 f1f4 c8h3 f4e4 a8e8 d4c3 e7f6 e4f4 f6e6 b2b3 e5g6 f4f2 e6e5 c3e5 g6e5 d1d4 c6c5 d4d5 c5c4 b3c4 b5a4');
INSERT INTO evaluations VALUES(31507,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31507,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(31507,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31507,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31507,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31507,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31507,36,'cp 0','f8b4 b1d2 d5c4 f1c4 g8e7 e1g1 h7h6 a2a3 b4d2 c1d2 b8d7 f1e1 d7b6 c4f1 a7a5 d2c3 f5g4 h2h3 g4h5 a1c1 e8g8 d1d3 h5g6 d3e2 g6h5');
INSERT INTO evaluations VALUES(28090,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28090,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28090,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28090,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28090,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28090,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28090,36,'cp 0','b1d2 g8e7 d1b3 a7a5 a2a3 d5c4 f1c4 b4d2 c1d2 b8d7 b3b7 a8b8 b7a7 b8a8 a7b7');
INSERT INTO evaluations VALUES(38814,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(38814,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(38814,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38814,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(38814,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(38814,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(38814,36,'cp 26','b4c3 b2c3 g8f6 e2e3 e8g8 c4d5 e6d5 f1d3 c7c5 g1e2 b7b6 e1g1 c8a6 d3a6 b8a6 d1d3 a6c7 d4c5 b6c5 c3c4 a8b8 f1d1 d5c4 d3c2 d8e7 c2c4 c7e6 c1d2 e7b7 d2e1 b7b3 a1c1 h7h6 h2h3 b3c4 c1c4 b8b2 e2f4 e6f4 c4f4 f8c8 e1c3 b2a2 c3f6 g7f6 d1c1 a2a3 f4f6');
INSERT INTO evaluations VALUES(38828,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(38828,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(38828,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38828,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(38828,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(38828,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(38828,36,'cp 37','b2c3 g8f6 e2e3 e8g8 c4d5 e6d5 f1d3 c7c5 g1e2 b7b6 e1g1 c8a6 d3a6 b8a6 d1d3 a6c7 d4c5 b6c5 c3c4 a8b8 f1d1 d5c4 d3c2 d8e8 c2c4 c7e6 a3a4 e8c6 c1d2 c6d5 c4d5 f6d5 d2a5 d5b6 h2h3 b6c4 a5e1 f8d8 e2c3 d8d1 a1d1');
INSERT INTO evaluations VALUES(38827,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(38827,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38827,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38827,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(38827,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(38827,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(38827,36,'cp 30','g8f6 e2e3 b7b6 c4d5 e6d5 f1d3 e8g8 g1e2 c7c5 e1g1 c8a6 d3a6 b8a6 d1d3 a6c7 d4c5 b6c5 c3c4 a8b8 f1d1 d5c4 d3c2 d8e7 c2c4 f8d8 c1d2 c7e6 d2e1 h7h6 e2c3 b8b2 d1d8 e7d8 c3a4 b2b8 h2h3 d8d5 c4d5 f6d5');
INSERT INTO evaluations VALUES(42090,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(42090,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(42090,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(42090,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(42090,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(42090,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42090,36,'cp 26','e2e3 c7c6 a3a4 e8g8 g1f3 f8e8 f1d3 b7b6 c4d5 c6d5 d1e2 b8c6 e1g1 c8b7 c1a3 h7h6 f1c1 c6a5 f3d2 e6e5 d4e5 e8e5 a3b4 a5c6 d2b3 e5e8 a4a5 c6e5 a5b6 a7b6 b3d4 e5d3 e2d3 a8a1 c1a1 d8c7 h2h3 b7c8 a1a8 c8d7 d3a6 e8a8 a6a8 g8h7');
INSERT INTO evaluations VALUES(42087,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(42087,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42087,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(42087,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42087,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(42087,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42087,36,'cp 25','c7c6 a3a4 e8g8 g1f3 f8e8 c1a3 h7h6 c4d5 c6d5 f1d3 b7b6 d1e2 c8d7 a3b2 b8c6 e1g1 d8c7 a1a3 c6a5 f3e5 d7c6 e5c6 a5c6 f2f3 c6a5 e3e4 e6e5 a3a1 a5c4 d3c4 d5c4 b2a3 e5d4 c3d4 c4c3 e2e3');
INSERT INTO evaluations VALUES(39120,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(39120,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(39120,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(39120,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(39120,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(39120,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(39120,36,'cp 143','f1c4 c7c5 c2c3 b8c6 e1g1 d6e7 h2h3 d7d6 c3d4 c5d4 f3d4 g8f6 b1c3 e8g8 f1e1 c6d4 d1d4 c8e6 c1e3 e6c4 d4c4 a7a6 c4b3 b7b5 a1d1 d8c8 e3d4 c8c4 b3c4 b5c4 d1d2 a8b8 e1c1 f8d8 c1d1 d6d5 d4f6 e7f6');
INSERT INTO evaluations VALUES(52998,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(52998,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(52998,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(52998,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(52998,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(52998,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(52998,36,'cp -26','g1f3 b8c6 f1b5 c8d7 e5d6 e7e6 d2d4 f8d6 d4c5 d6c5 e1g1 d8c7 b1d2 a7a6 b5d3 c6b4 d3c4 b7b5 a2a3 b5c4 a3b4 c5b4 d2e4 g8f6 e4f6 g7f6 c2c3 b4d6 d1e2 h8g8 g2g3 d7c6 f1d1 c6b7 a1a4');
INSERT INTO evaluations VALUES(52997,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(52997,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(52997,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(52997,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(52997,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52997,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(52997,36,'cp -17','b8c6 f1b5 c8d7 e5d6 e7e6 d2d4 f8d6 d4c5 d6c5 e1g1 d8c7 b1d2 g8f6 a2a3 a8d8 b5d3 e8g8 f1e1 b7b5 h2h3 a7a6 b2b4 c5b6 c1b2 d7c8 b2f6 g7f6 d2e4 g8g7 d1e2 b6a7 c2c4 b5c4 d3c4 c6d4 f3d4 a7d4 a1b1');
INSERT INTO evaluations VALUES(1540,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(1540,31,'cp -156',NULL);
INSERT INTO evaluations VALUES(1540,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(1540,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(1540,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(1540,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(1540,36,'cp -141','a1c1 c3a5 d4c5 g8e7 f1d3 e7g6 d3g6 h7g6 e1g1 f8e7 f1e1 c8d7 d2b3 a5a4 d1d2 a4b4 d2d1 a8c8 h2h3 h8h5 b3d4 c6e5 c1b1 b4a4 d1a4 d7a4 f3e5 h5e5 b1b7 e7c5 d4f3 e5e3 f2e3 a4c6 b7b3 f7f6 f3d4 c6d7 b3c3 e6e5 d4b3 c5b6 c3c8 d7c8');
INSERT INTO evaluations VALUES(1539,30,'cp -139',NULL);
INSERT INTO evaluations VALUES(1539,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(1539,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(1539,33,'cp -159',NULL);
INSERT INTO evaluations VALUES(1539,34,'cp -156',NULL);
INSERT INTO evaluations VALUES(1539,35,'cp -147',NULL);
INSERT INTO evaluations VALUES(1539,36,'cp -171','c3a5 d4c5 g8e7 f1d3 e7g6 d3g6 h7g6 e1g1 f8e7 f1e1 c8d7 d2b3 a5a4 d1d2 f7f6 e3f4 a4b4 d2e3 g6g5 f4g3 e8g8 e5f6 g7f6 g3d6 f8f7 e3d3 a8f8 d6e7 f7e7 c1b1 b4c4 d3d2 e6e5 b1c1 c4f4 d2d5 d7e6 d5d2 e7d7 d2f4 g5f4 f3d2 f8d8 d2e4 g8g7 e4d6');
INSERT INTO evaluations VALUES(1553,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(1553,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(1553,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(1553,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(1553,34,'cp -157',NULL);
INSERT INTO evaluations VALUES(1553,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(1553,36,'cp -151','d4c5 g8e7 f1d3 e7g6 d3g6 h7g6 e1g1 f8e7 f1e1 c8d7 d2b3 a5a4 d1d2 f7f6 e3f4 a4b4 c1c3 g6g5 e5f6 e7f6 f4e5 g5g4 e5f6 g7f6 f3d4 e6e5 d4c6 d7c6 b3d4 e8c8 d4c6 b7c6 d2c2 c8c7 e1b1 b4e4 c2e4 d5e4 c3b3 d8b8 b3b8 h8b8 b1b8 c7b8 g1f1 b8b7');
INSERT INTO evaluations VALUES(1550,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(1550,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(1550,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(1550,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(1550,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(1550,35,'cp -163',NULL);
INSERT INTO evaluations VALUES(1550,36,'cp -157','g8e7 f1d3 e7g6 d3g6 h7g6 e1g1 f8e7 h2h3 c8d7 d2b3 a5a4 c1b1 e8g8 b3d4 a4d1 f1d1 a8b8 d4c6 d7c6 f3d4 f8c8 d1c1 g6g5 g2g4 c6a4 c5c6 b7b6 d4b5 a4b5 b1b5 c8c7 a2a4 b8c8 a4a5 c7c6 c1c6 c8c6 a5b6 a7b6 e3b6 c6c4 g1g2 d5d4 b6a5 e7c5 b5b7 c4c2 b7d7 g8f8 a5c7');
INSERT INTO evaluations VALUES(15468,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(15468,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15468,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(15468,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15468,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15468,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15468,36,'cp -1','f8e7 g2g3 h7h6 g5e3 d6d5 c3d5 f6d5 e4d5 d8d5 f1g2 c8g4 h2h3 g4e6 e1g1 d5d7 d3d4 e6h3 f3e5 c6e5 g2h3 d7h3 d4e5 h6h5 d1d5 c7c6 d5e4 h3e6 a1d1 f7f5 e4d3 a8d8 d3b3 e6b3 a2b3 e8f7 e3a7 h5h4 g1g2 h4g3 g2g3 f7e6 d1d8 e7d8 a7d4 h8h4 d4c3 g7g5 f1e1');
INSERT INTO evaluations VALUES(8577,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8577,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8577,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8577,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8577,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8577,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8577,36,'cp 0','g2g3 h7h6 g5e3 d6d5 c3d5 f6d5 e4d5 d8d5 f1g2 c8g4 h2h3 g4e6 e1g1 d5d7 d3d4 e6h3 g2h3 d7h3 f3e5 e7d6 e5c6 d6g3 f2g3 h3g3 g1h1 g3h3 h1g1');
INSERT INTO evaluations VALUES(38699,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(38699,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38699,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(38699,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(38699,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38699,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(38699,36,'cp 0','g8f6 c2c3 e8g8 e1g1 h7h6 h2h3 b8c6 f1e1 a7a5 b1d2 c8e6 c4b5 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 a2a4 e7g6 b5c4 e6c4 e3c4 a7a6 b2b3 f8e8 d1c2 a6c6 d3d4 e5d4 f3d4 c6d7 a1d1 d6d5 e4d5 e8e1 d1e1 d7d5 d4f3 b7b6');
INSERT INTO evaluations VALUES(3050,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(3050,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(3050,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(3050,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(3050,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(3050,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(3050,36,'cp 28','d2d4 d7d5 d4c5 d5c4 d1a4 d8a5 a4a5 c6a5 c1e3 c8d7 a1d1 d7c6 f3d4 c6g2 g1g2 f6g4 e3g5 g7f6 d4f3 f6g5 f3g5 g4f6 c3d5 f6d5 d1d5 f8d8 f1d1 d8d5 d1d5 f7f6 g5e4 g8f8 f2f4 f8e8 g2f3 a5c6 e4c3 a8c8');
INSERT INTO evaluations VALUES(3054,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(3054,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(3054,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(3054,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(3054,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(3054,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(3054,36,'cp 20','d7d6 d4d5 c6a5 f3d2 e7e6 a1b1 f8e8 b2b4 c5b4 b1b4 h7h6 c1b2 e6e5 e2e4 b7b6 b2c1 a5b7 b4b1 c8d7 f1e1 h6h5 d2b3 h5h4 c1g5 h4g3 h2g3 g7f8 g2f1 a8c8 f2f3 g8g7 d1d2 f8e7 b1c1 f6g8 g5e3');
INSERT INTO evaluations VALUES(3197,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(3197,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(3197,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(3197,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(3197,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(3197,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(3197,36,'cp 33','d4d5 c6a5 f3d2 e7e6 a1b1 e6d5 c3d5 c8f5 e2e4 f5d7 f1e1 f8e8 b2b3 a5c6 d5f6 g7f6 d2f1 c6d4 f1e3 f6g7 c1b2 b7b5 c4b5 d7b5 h2h4 h7h5 e3c2 d8b6 d1d2 a8d8 c2d4 c5d4 b1d1 d4d3 b2g7 g8g7');
INSERT INTO evaluations VALUES(12806,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(12806,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12806,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(12806,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12806,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(12806,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(12806,36,'cp 31','g1f3 d7d5 d2d4 e7e6 c4d5 e6d5 f1g2 c5d4 f3d4 f8c5 d4c6 b7c6 c3a4 d8a5 c1d2 c5b4 a4c3 a5a6 e1g1 e8g8 a2a3 b4e7 d2e3 c8f5 b2b4 f8d8 e3d4 f6e4 e2e3 c6c5 b4c5 e7c5 d1e2 a6c6 c3e4 c5d4 a1c1 c6b6 e4c3 d4c3 c1c3 a8c8 c3c8 f5c8');
INSERT INTO evaluations VALUES(21168,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(21168,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(21168,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(21168,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(21168,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(21168,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(21168,36,'cp 33','g8f6 b1c3 c8g4 d2d4 e7e5 c4b5 f6d7 d4d5 c6d4 b5e2 d4e2 d1e2 f8e7 h2h3 g4f3 e2f3 e7g5 c1g5 d8g5 h3h4 g5g6 a2a4 a7a6 a4a5 h7h5 h1h3 g6f6 f3e3 a8c8 b2b4 g7g6 f2f3 c8a8 e3g5 f6g7 e1c1 d7f6 h3g3');
INSERT INTO evaluations VALUES(14467,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(14467,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(14467,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14467,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(14467,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(14467,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(14467,36,'cp 30','b1c3 c8g4 d2d4 e7e5 c4b5 f6d7 d4d5 c6d4 b5e2 d4e2 d1e2 f8e7 h2h3 g4f3 e2f3 e7g5 c1g5 d8g5 h3h4 g5g6 e1c1 h7h5 f3e2 e8c8 f2f3 c8b8 a2a4 a7a6 a4a5 f7f5 e4f5 g6f5 c3e4 d7f6 e4g5 h8e8 c1b1 e5e4 f3e4 f6e4 e2h5 e4g3');
INSERT INTO evaluations VALUES(14465,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(14465,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(14465,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14465,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14465,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14465,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(14465,36,'cp 29','c8g4 c4b5 f6d7 d2d4 e7e5 d4d5 c6d4 b5e2 d4e2 d1e2 f8e7 h2h3 g4f3 e2f3 e7g5 c1g5 d8g5 h3h4 g5g6 h1h3 e8g8 h3g3 g6h6 f3g4 d7f6 g4g5 h6g6 f2f3 f6h5 g5g6 h7g6 g3g4 g8h7 e1c1 a7a6 a2a4 f7f5 g4g5 f5e4 c3e4 f8f4 g5g4 f4g4 f3g4 h5f6 e4g5 h7g8 h4h5 f6g4');
INSERT INTO evaluations VALUES(25027,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(25027,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(25027,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(25027,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(25027,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25027,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(25027,36,'cp 40','d2d4 e7e5 c4b5 f6d7 d4d5 c6d4 b5e2 d4e2 d1e2 f8e7 h2h3 g4f3 e2f3 e7g5 c1g5 d8g5 h3h4 g5g6 h1h3 e8g8 h3g3 g6h6 f3g4 d7f6 g4g5 h6g6 a2a4 f6h5 g3g4 a7a6 f2f3 g6g5 h4g5 f7f6 e1f2 g7g6 a1h1 g8g7 f2e3 f6g5 g4g5 h7h6 g5g4 h5f6');
INSERT INTO evaluations VALUES(23229,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(23229,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(23229,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(23229,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(23229,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(23229,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(23229,36,'cp 103','f8b4 f1c4 d7d6 e1g1 b4c3 b2c3 c6a5 c4d3 c8g4 d1e2 g8e7 h2h3 g4h5 e2e3 e8g8 f3h4 c7c5 f2f4 e7g6 h4g6 h5g6 f4f5 g6e8 e3g3 g8h8 c1e3 c5c4 d3e2 b7b5 g3h4 a5c6 d4d5 c6e7 a2a4 b5a4 e2c4');
INSERT INTO evaluations VALUES(6934,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(6934,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(6934,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(6934,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(6934,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(6934,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(6934,36,'cp 104','f1c4 d7d6 a2a4 c8g4 e1g1 b4c3 b2c3 c6a5 c4a2 d8d7 d1d3 b7b6 c1a3 g8e7 a3b4 e7g6 b4a5 g6f4 d3e3 b6a5 a2c4 d7c8 f1b1 g4e6 c4f1 e8g8 c3c4 c7c5 d4c5 c8c5 e3c5 d6c5 b1b5 e6g4 f3h4 f4e2 f1e2 g4e2 h4f5 e2c4 b5a5 f8d8 a5c5 c4e6 f5e7 g8f8');
INSERT INTO evaluations VALUES(39317,30,'cp 563',NULL);
INSERT INTO evaluations VALUES(39317,31,'cp 559',NULL);
INSERT INTO evaluations VALUES(39317,32,'cp 553',NULL);
INSERT INTO evaluations VALUES(39317,33,'cp 540',NULL);
INSERT INTO evaluations VALUES(39317,34,'cp 578',NULL);
INSERT INTO evaluations VALUES(39317,35,'cp 548',NULL);
INSERT INTO evaluations VALUES(39317,36,'cp 587','d7d5 d2d4 d5e4 b1c3 b8d7 d1h5 g7g6 h5e2 g8f6 f2g1 e8g8 h2h3 f8e8 e5d7 c8d7 e2f2 f6h5 g2g4 h5g7 f1c4 g8h8 c1h6 f7f5 f2e3 c7c6 a1f1 b7b5 c4b3 a7a5 a2a4 b5a4 c3a4 a8b8 h6g5 d8c7 g5f4 c7d8 h1h2 d7e6 b3e6 e8e6 f4b8 d8b8 a4c5');
INSERT INTO evaluations VALUES(50541,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(50541,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(50541,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(50541,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(50541,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(50541,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(50541,36,'cp -36','d7d5 e4d5 e6d5 c4e2 f8d6 e1g1 g8f6 d2d4 c5d4 f3d4 e8g8 c1e3 b8c6 b1d2 d8c7 h2h3 f8e8 c2c3 c8d7 e2d3 h7h6 d1c2 d6h2 g1h1 h2f4 e3f4 c7f4 d2f3 f6e4 h1g1 e8e7 f1e1 a8e8 a4a5 f4d6 d4c6 b7c6 d3e4 d5e4 a1d1 d6f6');
INSERT INTO evaluations VALUES(50785,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(50785,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(50785,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(50785,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(50785,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(50785,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(50785,36,'cp -34','e4d5 e6d5 c4e2 g8f6 e1g1 f8d6 d2d4 c5d4 f3d4 e8g8 c1e3 b8c6 b1d2 d8c7 h2h3 h7h6 c2c3 c8d7 e2d3 a8e8 d3f5 d6h2 g1h1 h2e5 d2f3 c6d4 c3d4 e5d6 a1c1 c7d8 f5d3 f6e4 f3e5 d6e5 d4e5 e8e5 d3e4 e5e4 d1d5 d7c6 d5d8 f8d8');
INSERT INTO evaluations VALUES(50774,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(50774,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(50774,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(50774,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(50774,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(50774,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(50774,36,'cp -31','e6d5 c4e2 g8f6 e1g1 f8e7 d2d4 e8g8 d4c5 e7c5 c1g5 c8e6 g5h4 h7h6 b1d2 g7g5 h4g3 f6e4 c2c4 b8c6 c4d5 e4g3 h2g3 e6d5 a1c1 c5a7 e2c4 f8e8 f1e1 e8e1 d1e1 d5c4 d2c4 d8f6 e1c3 f6e7 c1e1 e7c5 c4e3 c5c3 b2c3');
INSERT INTO evaluations VALUES(50806,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(50806,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(50806,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(50806,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(50806,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(50806,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(50806,36,'cp -29','c4e2 g8f6 e1g1 f8d6 d2d4 c5d4 f3d4 e8g8 c1e3 f8e8 b1d2 b8c6 f1e1 d8c7 d2f1 c8d7 c2c3 h7h6 d1c2 e8e7 d4f5 d7f5 c2f5 a8e8 f5c2 e7e6 a1d1 c6e5 h2h3 e5c4 e3d4 c7e7 d4f6 e6e2 f6e7 e2c2 e7d6 e8e1 d1e1 c4d6 f1e3 c2b2 e3d5');
INSERT INTO evaluations VALUES(38007,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(38007,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(38007,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(38007,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(38007,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(38007,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(38007,36,'cp 114','b1c3 b8c6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 c8d7 c1e3 a7a6 b5c6 d7c6 d1d2 a8c8 e3h6 c6d7 h6f4 d7c6 a1c1 b7b5 f4h6 c6b7 d4d5 e6e5 a2a4 b5b4 c3e2 a6a5 d2e3 g7h6 e3h6 f7f6 c1c8 b7c8 e2g3 c8d7 b2b3 g8h8 e1c1');
INSERT INTO evaluations VALUES(38004,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(38004,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(38004,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(38004,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(38004,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(38004,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(38004,36,'cp 89','c8g4 f1b5 b8d7 c1e3 e7e6 d4d5 e6e5 h2h3 g4f3 d1f3 a7a6 b5d7 d8d7 f3d1 g8f6 a2a4 f6h5 e1g1 h5f4 a4a5 e8g8 a1a4 g7f6 d1b3 a8b8 g1h2 h7h6 f1d1 f8c8 a4b4 f6d8 b3a4 d7a4 b4a4 b7b5 a5b6 d8b6 a4a6 b6e3 f2e3 f4h5 a6d6 b8b2');
INSERT INTO evaluations VALUES(49687,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(49687,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(49687,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(49687,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49687,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49687,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(49687,36,'cp -16','d2d4 e5d4 e3d4 d7d5 c1g5 f8e7 g1f3 e8g8 f1d3 f8e8 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 c7c6 f2f4 g6d3 d1d3 b8d7 a1e1 d7e5 f4e5 f6d7 g5e7 e8e7 g4g5 d8b6 c3a4 b6a5 a4c3 d7f8 h3h4 f8e6 c3e2 a5b6 c2c3 b6b2 e2g3 g7g6 e1b1 b2a2');
INSERT INTO evaluations VALUES(49583,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(49583,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(49583,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49583,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49583,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(49583,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(49583,36,'cp 0','e5d4 e3d4 d7d5 f1d3 b8c6 c1e3 c6b4 g1f3 h7h6 e1g1 b4d3 d1d3 f8d6 c3b5 e8g8 b5d6 c7d6 f1e1 f8e8 f3d2 c8d7 d2f1 a7a6 f1g3 d8b6 d3b3 b6c6 c2c3 e8e6 b3d1 f6e4 g3e4 e6e4 f2f3 e4e6 e3d2 e6e1 d1e1 a8e8');
INSERT INTO evaluations VALUES(49772,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49772,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49772,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49772,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49772,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49772,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(49772,36,'cp 0','e3d4 d7d5 f1d3 b8c6 c1e3 c6b4 g1f3 h7h6 e1g1 b4d3 d1d3 f8d6 c3b5 e8g8 b5d6 c7d6 f1e1 d8b6 f3d2 c8d7 d3b3 b6c6 b3d3');
INSERT INTO evaluations VALUES(49756,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49756,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49756,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49756,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49756,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49756,36,'cp 0','d7d5 f1d3 b8c6 c1e3 c6b4 g1f3 h7h6 e1g1 b4d3 d1d3 f8d6 c3b5 e8g8 b5d6 c7d6 f1e1 d8b6 f3d2 c8d7 d3b3 b6c6 b3d3');
INSERT INTO evaluations VALUES(12360,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12360,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(12360,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(12360,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(12360,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(12360,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12360,36,'cp 0','e4e5 d7d5 c4b3 c5c4 e5f6 c4b3 f6g7 f8g7 a2b3 e6e5 d2d4 c8g4 e1g1 e5e4 d1e1 g4f3 g2f3 d8h4 g1h1 h4h3 f1g1 g7e5 d4e5 h3f3 g1g2 h8g8 e1g1 g8g2 g1g2 f3d1 g2g1 d1f3');
INSERT INTO evaluations VALUES(12328,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(12328,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12328,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12328,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(12328,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(12328,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12328,36,'cp 0','d7d5 c4b3 f6d7 d2d4 b7b5 e1g1 a7a5 c1e3 f8e7 f1e1 c8a6 a2a3 c5c4 b3c2 b5b4 a3b4 a5b4 d1e2 a6b7 b1d2 b4b3 a1a8 d8a8 c2b1 a8a1 d2f1 h7h6 b1h7 a1a4 h7b1 d7b6 h2h3 a4a1 b1g6 a1a5 g6b1');
INSERT INTO evaluations VALUES(15835,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(15835,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(15835,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(15835,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(15835,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(15835,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(15835,36,'cp -32','h7h6 g5f3 e5e4 f3d4 f8c5 c2c3 e8g8 e1g1 f6d5 d4b3 a5b3 a2b3 d5f6 h2h3 c7c6 d3e4 f6e4 d1f3 e4g5 f3g3 c5d6 g3d3 d8e7 c1g5 h6g5 d3e2 f8e8 e2e7 e8e7 f1d1 d6c7 b1d2 g5g4 h3g4 c8g4');
INSERT INTO evaluations VALUES(14684,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(14684,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(14684,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(14684,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(14684,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(14684,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(14684,36,'cp -45','g5f3 e5e4 f3d4 f8c5 c2c3 e8g8 e1g1 c8g4 d1c2 e4d3 c4d3 d8d5 h2h3 g4h5 d4e2 c5d6 c3c4 d5e6 b1c3 a5c6 a2a3 f8e8 c1f4 h5e2 c3e2 d6f4 e2f4 e6e5 f4e2 a8d8 f1e1 e5g5 a1d1 c6e5 d3e4 f6e4 c2e4 f7f5 e4f4 e5d3 f4g5 h6g5');
INSERT INTO evaluations VALUES(14685,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(14685,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(14685,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(14685,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(14685,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(14685,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(14685,36,'cp -30','e5e4 f3d4 f8c5 c1e3 f6d5 c4d5 c5d4 e3d4 d8d5 d3e4 d5e4 d1e2 e4e2 e1e2 f7f6 b1a3 c8f5 d4c3 a5c6 h1d1 a7a6 f2f3 a8d8 c3e1 e8f7 e1g3 c6d4 e2f1 d4e6 f1e2 g7g5 d1d8 h8d8 a1d1 d8c8 d1d2 h6h5 h2h4 b7b5 c2c3 g5h4 g3h4 e6f4 e2e3 f4g6 h4g3');
INSERT INTO evaluations VALUES(28028,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(28028,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(28028,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(28028,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(28028,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(28028,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(28028,36,'cp 120','h2h4 h7h5 f1e2 d7d6 g1f3 b8d7 e1g1 g8h6 a1b1 e8g8 d4d5 d8e8 c1d2 a7a6 f1e1 c7c6 d2g5 c6c5 f3h2 e7e5 d5e6 e8e6 c3d5 a8b8 g2g4 b8e8 g4h5 b7d5 c4d5 e6e4 d1d2 h6f5 e2a6 e4b4 d2b4 c5b4 h5g6');
INSERT INTO evaluations VALUES(53185,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53185,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53185,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53185,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(53185,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(53185,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(53185,36,'cp 26','a2a4 b8c6 b1a3 d8c7 f1b5 a7a6 b5c6 c7c6 e1g1 d4d3 c2d3 c8g4 d3d4 g4f3 g2f3 c5d4 d2d3 e7e6 a3c4 b7b5 c4a5 c6d7 c1d2 g8e7 d1e2 e7d5 e2e4 f8e7 e4g4 g7g5 a4b5 h7h5 g4e4 a6b5 a5b3 e8g8 a1a8 f8a8 b3d4');
INSERT INTO evaluations VALUES(18688,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(18688,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(18688,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(18688,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(18688,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(18688,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18688,36,'cp 0','c5d4 e1g1 a7a6 b5c6 d7c6 f3d4 e6e5 d4f3 d8c7 d1e2 g8f6 f1d1 c8e6 a2a4 f8c5 h2h3 e8g8 f3g5 e6d7 c1e3 c7b6 e3c5 b6c5 b1d2 c5c2 d1c1 c2b2 a1b1 b2a2 b1a1');
INSERT INTO evaluations VALUES(18683,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18683,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18683,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18683,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18683,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18683,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18683,36,'cp 0','e1g1 a7a6 b5c6 d7c6 f3d4 e6e5 d4f3 d8c7 d1e2 g8f6 f1d1 c8e6 a2a4 f8c5 h2h3 e8g8 f3g5 e6d7 c1e3 c7b6 e3c5 b6c5 b1d2 c5c2 d1c1 c2b2 a1b1 b2a2 b1a1');
INSERT INTO evaluations VALUES(19460,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(19460,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(19460,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(19460,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(19460,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(19460,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(19460,36,'cp -58','g8f6 b1c3 d7d5 e4d5 f6d5 c3d5 d8d5 g1e2 e7e5 e2c3 d5d8 f1e2 f8e7 e2f3 c6d4 c3d5 e7d6 c2c4 d4f3 d1f3 e8g8 e1g1 c8d7 a2a3 a7a5 f3g3 f7f6 d3d4 c5d4 e3d4 a8c8 a1d1 c8c4 d4b6 d8b8 b6a5 c4c6 d5e3 d7e6');
INSERT INTO evaluations VALUES(39318,30,'cp 544',NULL);
INSERT INTO evaluations VALUES(39318,31,'cp 575',NULL);
INSERT INTO evaluations VALUES(39318,32,'cp 590',NULL);
INSERT INTO evaluations VALUES(39318,33,'cp 582',NULL);
INSERT INTO evaluations VALUES(39318,34,'cp 575',NULL);
INSERT INTO evaluations VALUES(39318,35,'cp 599',NULL);
INSERT INTO evaluations VALUES(39318,36,'cp 586','e1f2 g8f6 b1c3 d7d6 e5f3 c8g4 d2d4 e8g8 h2h3 g4f3 g2f3 b8c6 h1g1 d6d5 c1e3 g8h8 f2g2 c6e7 g2h2 c7c6 d1d2 d8c7 e4e5 f6h5 f3f4 g7g6 f1d3 f8g8 f4f5 e7f5 d3f5 g6f5 a1f1 f5f4 e3f4 g8g6 c3e2 h5f4 d2f4');
INSERT INTO evaluations VALUES(52463,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52463,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52463,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(52463,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52463,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52463,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(52463,36,'cp -9','e7e5 c2c4 g7g6 b1c3 g8f6 c1b2 b8c6 c3d5 f8g7 d5f6 g7f6 g1e2 e8g8 e2c3 d7d6 g2g3 f6g7 d2d3 c8e6 f1g2 d8d7 d1d2 a8d8 e1g1 c6e7 e3e4 e6h3 g2h3 d7h3 a1e1 a7a6 c3d5 e7d5 e4d5 h3d7 f2f4 e5f4 b2g7 g8g7');
INSERT INTO evaluations VALUES(52139,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(52139,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(52139,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52139,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52139,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(52139,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(52139,36,'cp 0','c1b2 b8c6 c2c4 g7g6 b1c3 g8f6 f1e2 d7d5 c4d5 f6d5 c3d5 d8d5 f2f4 c8f5 e2f3 d5d7 f4e5 f8g7 e5e6 f7e6 f3c6 b7c6 b2g7 d7g7 g1f3 f5d3 a1c1 e8g8 c1c5 f8f3 d1f3 a8f8 f3c6 g7a1 c5c1 a1f6 c6f3 f6b2 f3c6');
INSERT INTO evaluations VALUES(53592,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(53592,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(53592,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(53592,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(53592,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(53592,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(53592,36,'cp 39','g1f3 g8f6 b1d2 c6c5 c2c3 d8b6 d1c2 f8d6 f4d6 b6d6 e3e4 d6c7 d4c5 c7c5 f1d3 b8d7 e1g1 d5e4 d2e4 c5c7 a1d1 e8g8 e4d6 h7h6 d3e2 b7b6 d1d4 c8b7 d6b7 c7b7 c2d2 a8c8 f1d1 f8d8');
INSERT INTO evaluations VALUES(53593,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(53593,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(53593,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(53593,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(53593,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(53593,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(53593,36,'cp 36','g8f6 b1d2 c6c5 c2c3 d8b6 d1c2 f8d6 f4d6 b6d6 e3e4 d6c7 d4c5 c7c5 e4d5 e6d5 d2b3 c5e7 f1e2 e8g8 h2h3 f8e8 b3d4 h7h6 e1g1 c8d7 e2d3 b8c6 f1e1 e7d6 a1d1 c6d4 f3d4 a7a6 c2d2 e8e1 d1e1 a8e8 e1e8 f6e8 d4c2 e8f6 d3e2');
INSERT INTO evaluations VALUES(9483,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9483,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9483,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9483,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9483,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(9483,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(9483,36,'cp 0','d2d3 f6d7 b1d2 e8g8 a2a4 f8e8 d2c4 f7f6 f3h4 c5f8 c4e3 d7c5 e3f5 c5e6 d1g4 e6d4 f5h6 g8h8 h6f7 h8g8');
INSERT INTO evaluations VALUES(9480,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9480,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9480,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9480,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9480,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9480,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9480,36,'cp 0','f6d7 f1e1 a7a5 c2c3 e8g8 d1c2 a5a4 h2h3 c5d6 c1e3 d8e7 d3d4 f8e8 b1d2 f7f6 b2b3 e5d4 c3d4 a4b3 a2b3 a8a1 e1a1 d7b6 a1a7 e7f7 f3h4 f7h5 h4f3');
INSERT INTO evaluations VALUES(25372,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(25372,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(25372,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(25372,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(25372,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(25372,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(25372,36,'cp 253','g5e6 f7e6 c4e6 c6d4 e6h3 g7g5 c3e2 g5g4 h3g4 f6g4 e2d4 d8h4 g2g3 h4h3 d4e6 g4f2 e6c7 e8d8 e1f2 d8c7 d1f1 h3g4 f1c4 c7d7 c4e2 g4h3 e2f1 h3g4 d2d3 a8c8 c2c3 f8e7 f2g2 c8f8 f1d1 g4g6 h1f1 h7h5 f1f8 h8f8 d1e2 h5h4 c1e3 h4g3 h2h3 g6e6 e2g4 e6g4 h3g4 d7e6 g2g3');
INSERT INTO evaluations VALUES(25368,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(25368,31,'cp 251',NULL);
INSERT INTO evaluations VALUES(25368,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(25368,33,'cp 245',NULL);
INSERT INTO evaluations VALUES(25368,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(25368,35,'cp 242',NULL);
INSERT INTO evaluations VALUES(25368,36,'cp 251','f7e6 c4e6 c6d4 e6h3 g7g5 c3e2 g5g4 h3g4 f6g4 e2d4 d8h4 g2g3 h4h3 d4e6 g4f2 e6c7 e8d7 e1f2 d7c7 d1f1 h3g4 f1c4 c7d7 c4e2 g4h3 e2f1 h3g4 d2d3 a8c8 c2c3 f8e7 f2g2 c8f8 f1d1 g4g6 h1f1 h7h5 f1f8 h8f8 d1e2 h5h4 c1e3 d6d5 e3f2 d5e4 d3e4 d7c8 a1e1 f8g8 e1d1');
INSERT INTO evaluations VALUES(25430,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(25430,31,'cp 265',NULL);
INSERT INTO evaluations VALUES(25430,32,'cp 248',NULL);
INSERT INTO evaluations VALUES(25430,33,'cp 260',NULL);
INSERT INTO evaluations VALUES(25430,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(25430,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(25430,36,'cp 272','c4e6 c6d4 e6h3 g7g5 c3e2 g5g4 h3g4 f6g4 e2d4 d8h4 g2g3 h4h3 d4e6 g4f2 e6c7 e8d8 e1f2 d8c7 d1f1 h3g4 f1c4 c7d7 c4e2 h7h5 d2d3 g4e6 h2h3 e6g6 c1e3 a8c8 a1f1 h8g8 h1g1 f8e7 c2c4 c8f8 f2e1 e7g5 f1f8 g8f8 e1d1 a7a6 d1c2 g5e3 e2e3 g6f6 c2c3 b7b5 g3g4 h5g4 h3g4 b5c4 g4g5 f6f7 g5g6');
INSERT INTO evaluations VALUES(25422,30,'cp 271',NULL);
INSERT INTO evaluations VALUES(25422,31,'cp 255',NULL);
INSERT INTO evaluations VALUES(25422,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(25422,33,'cp 254',NULL);
INSERT INTO evaluations VALUES(25422,34,'cp 255',NULL);
INSERT INTO evaluations VALUES(25422,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(25422,36,'cp 241','c6d4 e6h3 g7g5 c3e2 g5g4 h3g4 f6g4 e2d4 d8h4 g2g3 h4h3 d4e6 g4f2 e6c7 e8d7 e1f2 d7c7 d1f1 h3g4 f1c4 c7d7 c4e2 g4h3 e2f1 h3g4 d2d3 a8c8 c2c3 f8e7 f2g2 c8f8 f1d1 g4g6 h1f1 h7h5 f1f8 h8f8 d1e2 h5h4 g3g4 e7g5 c1g5 g6g5 g2h3 d7c8 a1f1 f8f1 e2f1 g5e3 h3h4');
INSERT INTO evaluations VALUES(25439,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(25439,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(25439,32,'cp 251',NULL);
INSERT INTO evaluations VALUES(25439,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(25439,34,'cp 269',NULL);
INSERT INTO evaluations VALUES(25439,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(25439,36,'cp 235','e6h3 g7g5 c3e2 g5g4 h3g4 f6g4 e2d4 d8h4 g2g3 h4h3 d4e6 g4f2 e6c7 e8d8 e1f2 d8c7 d1f1 h3e6 f2g2 f8e7 b2b3 h7h5 f1c4 c7d7 c4e6 d7e6 d2d3 b7b5 c1e3 h5h4 g3g4 b5b4 a2a3 h4h3 g2f3 a8c8 a1a2 b4a3 c2c4 c8f8 f3e2 h8g8 h1g1 f8b8 a2a3 a7a5 e3d2 e7h4 g4g5 g8f8 d2e3 a5a4 g1g4 f8h8 g5g6 b8b3 a3b3 a4b3');
INSERT INTO evaluations VALUES(36547,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36547,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(36547,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(36547,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(36547,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(36547,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(36547,36,'cp 49','g1f3 c7c5 d4d5 e7e6 e1g1 f8e8 c1f4 e6d5 c4d5 c8f5 f3h4 f5c8 d1d2 a7a6 a2a4 d8e7 h4f3 c8f5 f4g5 e7f8 f3h4 f5c8 f2f4 h7h6 g5f6 g7f6 e2e4 b8d7 e4e5 d6e5 c3e4 f8g7 f4f5 g6g5 h4f3 g5g4 f3e1 f6g5');
INSERT INTO evaluations VALUES(37390,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(37390,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37390,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37390,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37390,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37390,36,'cp 0','d7d5 a1c1 d5c4 e2e3 c8e6 f3g5 e6d5 e3e4 h7h6 e4d5 h6g5 f4g5 f6d5 f1c4 d5b6 c4b3 d8d4 d1d4 g7d4 c3b5 d4b2 c1c2 b2e5 b5c7 e5c7 c2c7 e7e6 c7b7 b8c6 b3d1 f8c8 e1g1 c6a5 b7e7 a5c6');
INSERT INTO evaluations VALUES(36722,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36722,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36722,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36722,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36722,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36722,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36722,36,'cp 11','a1c1 d5c4 e2e3 c8e6 f3g5 e6d5 e3e4 h7h6 e4d5 h6g5 f4g5 f6d5 f1c4 d5b6 c4b3 d8d4 d1d4 g7d4 c3b5 d4b2 c1c7 b8c6 e1g1 e7e6 f1b1 b2g7 c7b7 c6a5 b7e7 a5b3 b1b3 a7a5 b5d6 b6d5 e7d7 d5f6 g5f6 g7f6 b3b7 a8d8 d7d8 f8d8 d6f7');
INSERT INTO evaluations VALUES(38879,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38879,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38879,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38879,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38879,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(38879,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(38879,36,'cp 36','a2a3 b4c3 b2c3 c7c5 e2e3 d8c7 f3d2 g8f6 f1c4 b8c6 d1e2 e8g8 c1b2 b7b6 e1g1 c8b7 f2f3 a8c8 a1c1 f8d8 e2f2 c6a5 c4d3 h7h6 e3e4 e6e5 f1e1 c5c4 d3f1 c7e7 a3a4 f6h5 g2g3 h5f6 d4d5 f6d7 f3f4');
INSERT INTO evaluations VALUES(57207,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(57207,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(57207,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(57207,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(57207,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(57207,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(57207,36,'cp -20','b8c6 g1f3 d7d5 e2e4 g8f6 b1d2 f8d6 f1e2 e8g8 e1g1 h7h6 e4d5 f6d5 f1e1 f8e8 h2h3 a7a5 e2f1 c8f5 d2c4 d8f6 d1b3 d5e7 d3d4 a5a4 b3d1 e5d4 c4d6 f6d6 f3d4 c6d4 d1d4 d6d4 c3d4 f5e6 c1d2');
INSERT INTO evaluations VALUES(874,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(874,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(874,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(874,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(874,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(874,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(874,36,'cp 195','d2d4 e6d7 e2b5 d7d6 f3e5 c8d7 e5d7 f6d7 d4d5 c6e5 f2f4 e5g6 e1g1 e8c8 f4f5 g6e5 g1h1 g7g6 c1e3 g6f5 d1d4 d7b6 f1f5 f7f6 a1d1 f8g7 e3g1 a7a6 b5e2 e7e6 f5f1 f6f5 a2a4 h8g8 d4f4 c8b8 d5e6 d6e6 f4f5 e6f5 f1f5 d8d1 e2d1 e5c6');
INSERT INTO evaluations VALUES(872,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(872,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(872,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(872,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(872,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(872,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(872,36,'cp 156','e6d7 e2b5 d7d6 f3e5 c8d7 e5d7 f6d7 d4d5 c6e5 f2f4 e5g6 e1g1 e8c8 c1e3 e7e6 f4f5 g6e5 d5e6 f7e6 d1d6 f8d6 f5e6 d7c5 e3g5 d8g8 e6e7 c7c6 a1d1 c8c7 g5f4 g8e8 b2b4 c5d7 d1d4 d6e7 b5d3 h8f8 d3h7 g7g6 h7g6 f8f4 d4d7 e5d7 f1f4 e8f8 f4f8 e7f8 b4b5 c6b5 c3b5 c7b6 b5d4 f8c5 c2c3 d7e5 g6f5');
INSERT INTO evaluations VALUES(22537,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(22537,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(22537,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22537,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22537,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(22537,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(22537,36,'cp -11','d2d4 g7g5 d4d5 c6e7 h2h4 f8g7 d5d6 c7d6 b1c3 d6d5 c4d5 e7d5 d1d5 g8e7 d5d6 d8b6 d6b6 a7b6 h4g5 h6g5 h1h8 g7h8 c3b5 e8f8 f3g5 d7d5 b5c7 a8a5 c1f4 h8b2 a1b1 b2c3 e1d1 d5e4 g5e4 c3e5 f4e5 a5e5 e4d6 e5c5 d6c8 e7c8 c7b5 c5d5 d1c1 d5g5 b1b3 g5g2 b3c3 c8e7');
INSERT INTO evaluations VALUES(22539,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(22539,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(22539,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(22539,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22539,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(22539,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(22539,36,'cp -7','g7g5 d4d5 c6e7 h2h4 f8g7 d5d6 c7d6 b1c3 d6d5 c4d5 e7d5 d1d5 g8e7 d5d6 d8b6 d6b6 a7b6 h4g5 h6g5 h1h8 g7h8 c3b5 e8f8 f3g5 d7d5 b5d6 c8e6 c1f4 h8b2 g5e6 f7e6 a1b1 a8a2 e1d2 d5e4 f4h6 f8g8 d6e4 e7f5 h6e3 b6b5 d2d3 f5e3 d3e3 b5b4 g2g3 b7b5 e3d3 g8f7 e4d6 f7g6 d6e4 b2e5 b1b4');
INSERT INTO evaluations VALUES(22499,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(22499,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(22499,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22499,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(22499,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(22499,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(22499,36,'cp -9','d4d5 c6e7 h2h4 f8g7 d5d6 c7d6 b1c3 d6d5 c4d5 e7d5 d1d5 g8e7 d5d6 d8b6 d6b6 a7b6 h4g5 h6g5 h1h8 g7h8 c3b5 e8f8 f3g5 d7d5 b5c7 a8a5 c1f4 h8b2 a1b1 b2c3 e1d1 d5e4 g5e4 c3e5 f4e5 a5e5 e4d6 c8g4 d1c1 e5e1 c1b2 e1e2 b2c3 e2g2 b1b6 e7f5 d6b7');
INSERT INTO evaluations VALUES(45072,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(45072,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(45072,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(45072,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(45072,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(45072,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(45072,36,'cp 146','c3e4 d5e4 f1c4 e7e6 g1e2 b7b5 c4b3 c8b7 c2c3 c6c5 e1g1 b8c6 c1e3 f8e7 e2g3 e8g8 d1g4 c5d4 c3d4 g8h8 g3e4 a7a5 a1c1 c6b4 e4c5 b7d5 b3d5 b4d5 f1d1 h7h6 h2h3 a8c8 e3d2 e7c5 d4c5 d8c7 g4e4');
INSERT INTO evaluations VALUES(45071,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(45071,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(45071,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(45071,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(45071,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(45071,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(45071,36,'cp 154','d5e4 c2c3 e7e6 c1e3 c6c5 d4c5 d8d1 a1d1 a7a5 g1e2 b8d7 e2g3 f8c5 e3f4 f7f5 e5f6 d7f6 f1b5 e8e7 e1e2 a5a4 f4g5 a8a5 c3c4 a5b5 c4b5 c8d7 g3e4 d7b5 e2d2 c5b6 e4c3 b6a5 d2c2 b5c6 f2f3 a4a3 d1d4 a3b2 c2b2 h8a8 g5d2 a5b6 d4c4 h7h6 h1e1 g7g5 a2a4 a8d8 b2c1 e7f7 c1c2 d8c8 d2e3 b6a5 c4c5');
INSERT INTO evaluations VALUES(45076,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(45076,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(45076,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(45076,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(45076,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(45076,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(45076,36,'cp 150','c2c3 e7e6 c1e3 c6c5 d4c5 d8d1 a1d1 c8d7 g1e2 b7b6 e2g3 d7c6 c5b6 a7b6 e3b6 b8d7 b6d4 h7h5 h2h4 g7g6 f1c4 f8g7 e1g1 d7e5 f1e1 e8g8 d4e5 g7e5 g3e4 a8a4 e4d2 e5f6 g2g3 f8b8 c4b3 a4a8 d2c4 g8g7 c4e3 f6e7 d1d2 a8a5 e3c4');
INSERT INTO evaluations VALUES(45075,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(45075,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(45075,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(45075,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(45075,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(45075,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(45075,36,'cp 159','e7e6 c1e3 c6c5 d4c5 d8c7 d1d4 c8d7 b2b4 b7b6 d4e4 d7c6 e4f4 b8d7 b4b5 c6d5 c3c4 g7g5 f4g5 d5e4 g1f3 e4f3 g2f3 f8c5 f3f4 c5b4 e1d1 a7a6 a1c1 a6b5 f1g2 a8a2 c4b5 b4c5 g2c6 c5e3 f2e3 c7a7 g5g7 h8f8 g7h7');
INSERT INTO evaluations VALUES(53324,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(53324,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(53324,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(53324,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53324,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(53324,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(53324,36,'cp 36','g1f3 g8f6 e1g1 b8c6 d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 c2c4 c8e6 b1c3 a8c8 c4d5 c6d5 d1a4 d8d7 a4d7 e8d7 f1d1 f8c5 c3a4 h7h6 a4c5 c8c5 c1d2 e6f5 f2f4 d7e6 a1c1 h8c8 c1c5 c8c5 f4e5 f6e4 g2e4 f5e4 d2c3 h6h5 d1d4 c5c4 d4c4 d5c4');
INSERT INTO evaluations VALUES(16306,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(16306,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(16306,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(16306,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(16306,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(16306,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16306,36,'cp -7','e5d4 e1g1 a7a6 b5a4 f8e7 e4e5 f6e4 f3d4 e8g8 d4f5 d7d5 e5d6 c8f5 d6e7 c6e7 b1c3 e4c3 b2c3 c7c6 a4b3 e7d5 d1d4 f8e8 c3c4 d5b4 d4c3 c6c5 c1e3 b7b6 f1e1 d8c7 c3d2 a8d8 e3f4 e8e6 f4c7 d8d2 e1e6 f5e6 c7b6 b4c2 b3c2 d2c2 b6c5');
INSERT INTO evaluations VALUES(16326,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(16326,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(16326,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(16326,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16326,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(16326,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(16326,36,'cp 0','e1g1 a7a6 b5a4 f8e7 e4e5 f6e4 f3d4 e8g8 d4f5 d7d5 e5d6 c8f5 d6e7 d8e7 a4c6 b7c6 b1d2 e4f6 d2c4 a8d8 d1e1 e7c5 b2b3 f8e8 e1a5 c5a5 c4a5 f5c2 a5c6 d8d6 c6a5 c2d3 f1d1 d3c2');
INSERT INTO evaluations VALUES(11566,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(11566,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(11566,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(11566,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(11566,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(11566,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(11566,36,'cp 0','e4c5 a4c6 d7c6 f3e5 f8e7 d2d4 c5e6 c1e3 e8g8 c2c4 e7d6 f2f4 d8h4 g2g3 h4h3 b1c3 f8d8 c3e4 d6e5 e4f2 h3h6 f2g4 h6h3');
INSERT INTO evaluations VALUES(11552,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11552,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11552,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11552,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11552,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11552,36,'cp 0','a4c6 d7c6 f3e5 f8e7 d2d4 c5e6 c1e3 e8g8 c2c4 e7d6 f2f4 d8h4 g2g3 h4h3 b1c3 c6c5 d4d5 e6g5 f4g5 d6e5 d1d2 b7b6 e3f4 e5d4 f4e3');
INSERT INTO evaluations VALUES(11500,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11500,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11500,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11500,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11500,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11500,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11500,36,'cp 0','d7c6 f3e5 f8e7 d2d4 c5e6 c1e3 e8g8 c2c4 e7d6 f2f4 d8h4 g2g3 h4h3 b1c3 c6c5 d4d5 e6g5 f4g5 d6e5 d1d2 b7b6 e3f4 e5d4 f4e3');
INSERT INTO evaluations VALUES(11462,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11462,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11462,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11462,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11462,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11462,36,'cp 10','f3e5 f8e7 d2d4 c5e6 c1e3 e8g8 c2c4 f7f6 e5f3 f8e8 b1c3 e6f8 d1b3 c8g4 f3d2 d8d7 d4d5 c6d5 c4d5 g4h5 e3f4 b7b5 f4c7 h5f7 c7a5 f8g6 h2h3 g6f4 a1d1 f7d5 c3d5 d7d5 b3d5 f4d5 d2b1 e7b4 e1e8 a8e8 a5b4 d5b4');
INSERT INTO evaluations VALUES(11461,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11461,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(11461,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(11461,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(11461,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11461,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(11461,36,'cp 12','f8e7 d2d4 c5e6 c1e3 f7f6 e5f3 e8g8 c2c4 f8e8 b1c3 e6f8 d4d5 c8g4 e3f4 d8d7 h2h3 g4h5 d5c6 d7c6 d1d5 h5f7 d5c6 b7c6 f3d4 e7b4 f4c7 f8e6 d4e6 f7e6 e1e4 e6f7 a1e1 e8e4 e1e4 b4c3 b2c3 a8e8 e4e8 f7e8 c4c5 g8f7 c7d6 f7e6 f2f3 a6a5 c3c4 e8g6 h3h4 g6c2 a2a3 e6d7');
INSERT INTO evaluations VALUES(7130,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7130,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(7130,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(7130,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7130,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(7130,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(7130,36,'cp 17','d2d4 c5e6 c1e3 e8g8 c2c4 f7f6 e5f3 f8e8 b1c3 e6f8 d4d5 c8g4 e3f4 d8d7 f4g3 e7c5 d1b3 g4f3 g2f3 b7b6 d5c6 d7c6 c3d5 f8e6 e1e6 e8e6 d5c7 a8d8 c7e6 c6e6 a1d1 d8d1 b3d1 e6c4 b2b3 c4f7 d1d3 a6a5 g1g2 f7f8 d3e4 g8h8 e4c6 f8f7 c6a8 c5f8');
INSERT INTO evaluations VALUES(7129,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7129,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(7129,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7129,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(7129,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(7129,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(7129,36,'cp 12','c5e6 c1e3 e8g8 c2c4 f8e8 b1c3 e6f8 d1b3 f8g6 a1d1 d8d6 c3e4 d6b4 b3b4 e7b4 e3d2 b4d2 d1d2 c8f5 e5g6 f5g6 d2e2 e8e4 e2e4 g6e4 e1e4 f7f5 e4e7 a8d8 g2g3 d8d4 e7c7 f5f4 c7b7 d4c4 g1g2 f4g3 h2g3 c4c2 b7b6 g8f7 b6a6');
INSERT INTO evaluations VALUES(7126,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7126,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7126,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(7126,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7126,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(7126,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(7126,36,'cp 13','c1e3 e8g8 c2c4 f8e8 b1c3 e6f8 d1b3 f8g6 a1d1 e7f8 c3e4 g6e5 d4e5 d8h4 e3c5 c8f5 c5f8 f5e4 g2g3 h4g4 d1d4 g4f3 e1e4 f3b3 a2b3 g8f8 f2f4 e8d8 b3b4 f8e7 g1f2 d8d4 e4d4 a6a5 c4c5 b7b5 c5b6 c7b6 d4d6 a8c8 b4a5 b6a5 f2f3 c8b8 d6c6');
INSERT INTO evaluations VALUES(17256,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(17256,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(17256,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(17256,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(17256,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(17256,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(17256,36,'cp 81','e2e4 e7e6 a2a3 d7d6 g1f3 f8e7 c1e3 e8g8 f1d3 e6e5 d4d5 c6b8 c3e2 f6g4 e3d2 g4f6 c2c4 c8g4 d2e3 c7c6 e2c3 b8d7 h2h3 g4h5 g2g4 h5g6 f3d2 c6d5 c4d5 a8c8 d1e2 d7c5 d3c2 f6e8 e1c1 a7a6 c1b1');
INSERT INTO evaluations VALUES(26610,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(26610,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(26610,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(26610,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(26610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26610,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(26610,36,'cp 11','c1f4 e7e6 d1d2 g8f6 f3e5 c6e5 d4e5 f6e4 c3e4 d5e4 d2c3 c7c6 h2h4 h7h5 a2a3 f8e7 g2g3 d8d5 a1d1 d5b5 f4e3 g4f5 d1d2 a7a5 f1g2 a5a4 e1g1 b5a5 c3a5 a8a5 c2c4 e8g8 f1c1 a5e5 d2d7 e7c5 e3c5 e5c5 d7b7 f8d8');
INSERT INTO evaluations VALUES(26600,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(26600,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(26600,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(26600,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26600,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26600,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26600,36,'cp 1','e7e6 d1d2 g8f6 f3e5 c6e5 d4e5 f6e4 c3e4 d5e4 e1c1 d8d2 d1d2 a8d8 g2g3 d8d2 c1d2 e8d7 f1g2 g4f5 g3g4 f5g6 h2h4 h7h5 g4h5 h8h5 f2f3 f8c5 f3e4 c5f2 f4g5 f2d4 g2f3 h5h8 h4h5 g6h7 h5h6 d4e5 h6g7 e5g7 c2c3 d7c8 d2e3 c7c6');
INSERT INTO evaluations VALUES(31071,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(31071,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31071,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31071,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(31071,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(31071,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(31071,36,'cp 15','e7e6 f1d3 f5d3 d1d3 b8d7 b1d2 f8e7 e1g1 g8f6 f1d1 e8g8 b2b3 d8c7 c1b2 a7a5 a2a4 f8e8 a1c1 d7f8 e3e4 d5e4 d2e4 a8d8 e4f6 e7f6 d3e2 b7b6 d1d3 d8d7 c1d1 c6c5 d4d5 f6b2 d5d6 c7c6 e2b2 f7f6 h2h4');
INSERT INTO evaluations VALUES(34632,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(34632,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(34632,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(34632,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(34632,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(34632,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(34632,36,'cp 203','g1f3 g5a5 d2d4 g8f6 e1g1 c7c6 c1d2 c8f5 f3h4 f5e6 f1e1 a5d8 h4f3 b8d7 e2f1 e6g4 d4d5 f6d5 c3d5 c6d5 h2h3 g4f3 d1f3 e7e6 f3d5 d8b6 d2a5 b6c6 d5b3 a7a6 a5c3 c6b6 g2g3 e8c8 f1g2 b6b3 a2b3 h8g8 e1d1 f8c5');
INSERT INTO evaluations VALUES(34629,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(34629,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(34629,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(34629,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(34629,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(34629,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(34629,36,'cp 208','g5a5 e1g1 c7c6 d2d4 g8f6 c1d2 a5c7 f1e1 g7g6 e2c4 f8g7 d1e2 e7e6 c3e4 f6e4 e2e4 e8g8 d2f4 c7d8 a2a4 b8d7 a4a5 d7f6 e4d3 d8e7 c2c3 c8d7 b2b4 f8d8 f4c7 d8c8 c7e5 c8d8 f3d2 d7e8 h2h3 f6d5 e5g7 g8g7 d2e4 d5f6 d3g3 f6e4 e1e4');
INSERT INTO evaluations VALUES(57047,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(57047,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(57047,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(57047,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(57047,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(57047,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(57047,36,'cp -13','c7c5 g2g3 b8c6 f1g2 g7g6 c2c3 d7d6 g1f3 e6e5 d2d3 f8g7 e1g1 g8f6 a2a3 c8e6 b2b4 e8g8 b4b5 c6a5 b1d2 c5c4 d3d4 a8c8 d4d5 e6g4 a1b1 f6d7 h2h3 g4f3 g2f3 f7f5 b1b4 d7c5 d2c4 f5e4 f3e4 c5e4 e2e4 a5c4 b4c4 c8c4 e4c4');
INSERT INTO evaluations VALUES(52005,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(52005,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(52005,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52005,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(52005,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(52005,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52005,36,'cp 0','g2g3 b8c6 d2d3 d7d6 f1g2 g7g6 c2c3 f8g7 b1a3 g8f6 g1f3 e8g8 e1g1 e6e5 c1g5 h7h6 g5f6 d8f6 a3c2 f6d8 c2e3 a8b8 a2a4 c6e7 f3h4 g7f6 h4f3');
INSERT INTO evaluations VALUES(6969,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6969,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6969,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6969,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6969,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6969,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(6969,36,'cp 9','d2d4 c6a5 c4d3 d7d6 d4e5 d6e5 f3e5 g8f6 e1g1 e8g8 d1c2 c7c5 b1d2 d8c7 f2f4 c8e6 d2f3 c5c4 d3e2 a8e8 g1h1 e7d6 c1e3 e6c8 e3d4 b7b6 e5c4 a5c4 e4e5 f6d5 e2c4 c7c4 e5d6 f7f6 f4f5 d5e3 d4e3 e8e3 f1e1 f8e8 e1e3 e8e3 c2f2');
INSERT INTO evaluations VALUES(6975,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(6975,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6975,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(6975,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(6975,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(6975,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(6975,36,'cp 0','c6a5 c4d3 d7d6 d4e5 d6e5 f3e5 g8f6 e1g1 e8g8 d1c2 d8d6 e5f3 f8d8 d3e2 d6c6 b1d2 h7h6 e4e5 f6d5 d2e4 d5b4 c2b1 b4d5');
INSERT INTO evaluations VALUES(7003,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7003,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7003,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7003,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7003,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7003,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7003,36,'cp 0','c4d3 d7d6 d4e5 d6e5 f3e5 g8f6 e1g1 e8g8 d1c2 e7d6 e5f3 c8g4 e4e5 g4f3 e5f6 d8f6 g2f3 f6h4 f3f4 h4g4 g1h1 g4f3 h1g1');
INSERT INTO evaluations VALUES(42615,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42615,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42615,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(42615,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42615,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(42615,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42615,36,'cp 0','d7d6 f1c4 e8g8 e1g1 b8c6 d2d3 c6a5 c3a4 a5c4 a4c5 d6c5 d3c4 f6e4 f3e5 f7f6 e5d3 d8d4 c1e3 d4c4 b2b3 c4c3 f1e1 f8e8 e3d2 c3d4 d2e3');
INSERT INTO evaluations VALUES(41892,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41892,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(41892,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41892,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41892,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41892,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41892,36,'cp 0','f1c4 e8g8 e1g1 a7a5 d2d3 c7c6 a2a4 h7h6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1d2 b8d7 f1f2 a8b8 c3e2 d7c5 b2b3 b8a8 e2c3 f8f7 a1f1 d8e7 f3h4 g8h7 h4f3');
INSERT INTO evaluations VALUES(50496,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(50496,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(50496,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(50496,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(50496,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(50496,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(50496,36,'cp -28','b7b5 c4a2 b8c6 d2d3 g8f6 g1f3 c8b7 e1g1 f8d6 a3a4 d8b6 f1e1 c6d4 c1g5 d4f3 d1f3 d6e5 g5f4 e5f4 f3f4 d7d6 c3e2 e8g8 a4a5 b6c7 e2g3 f6d7 c2c3 d7e5 f4e3 c5c4 d3c4 b5c4 a2b1 a8b8 f2f4 e5g6 b1c2 b7c6 f4f5 e6f5 e4f5');
INSERT INTO evaluations VALUES(51202,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(51202,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(51202,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(51202,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(51202,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(51202,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(51202,36,'cp -27','c4a2 b8c6 d2d3 g8f6 g1f3 f8d6 e1g1 c8b7 a3a4 d8b6 f1e1 c6d4 h2h3 d4f3 d1f3 e8g8 c1f4 d6f4 f3f4 a8c8 a4b5 a6b5 e4e5 f6e8 c3e4 c5c4 d3c4 b5c4 e4d2 b6b2 a2c4 b2c2 f4d4 c2g6 c4f1 c8c2 d2e4 b7e4 e1e4');
INSERT INTO evaluations VALUES(51221,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(51221,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(51221,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(51221,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(51221,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(51221,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(51221,36,'cp -38','c8b7 d2d3 b8c6 g1h3 g8f6 e1g1 f8e7 f2f4 d8b6 g1h1 e8g8 c1e3 c6d4 e3g1 g7g6 c3b1 b5b4 b1d2 b4a3 d2c4 b6c6 b2a3 c6c7 c4e3 d7d5 e4d5 e6d5 f4f5 d4b5 f5g6 h7g6 e3g4 b5c3 d1f3 f6g4 f3g4 c3a2 a1a2 d5d4');
INSERT INTO evaluations VALUES(30337,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(30337,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(30337,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(30337,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(30337,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(30337,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(30337,36,'cp -37','d2d3 b8c6 g1h3 g8f6 e1g1 f8e7 f2f4 e8g8 c3e2 a6a5 e2g3 d7d5 e4e5 f6e8 f4f5 e6f5 g3f5 c6e5 f5e7 d8e7 f1e1 e8d6 c1f4 f7f6 d3d4 c5d4 d1d4 d6f5 d4b6 b5b4 h3g5 f8e8 g5e6 a8c8 f4e5 f6e5 e1e5 c8c2 a1e1 c2b2 e5f5 b2a2');
INSERT INTO evaluations VALUES(30338,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(30338,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(30338,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(30338,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(30338,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(30338,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(30338,36,'cp -31','b8c6 g1h3 f8e7 e1g1 g8f6 f2f4 d8b6 g1h1 e8g8 c1e3 c6d4 e3g1 g7g6 c3b1 b5b4 b1d2 b4a3 d2c4 b6c6 b2a3 d7d5 c4a5 c6c7 a5b7 d5e4 c2c3 d4b5 c3c4 b5d4 a1b1 f8b8 b7c5 c7c5 f4f5 e6f5 b1b8 a8b8 d3e4 c5a3 d1d4 a3a2 e4f5 b8b1 f1b1 a2b1 f5g6 h7g6');
INSERT INTO evaluations VALUES(25597,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(25597,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(25597,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(25597,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(25597,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(25597,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(25597,36,'cp -39','g1h3 f8e7 e1g1 g8f6 f2f4 e8g8 c3e2 d8b6 g1h1 d7d5 e4e5 f6h5 c2c3 g7g6 d3d4 h5g7 c1e3 a6a5 a2b1 b5b4 b1d3 c5d4 e2d4 b4a3 b2a3 c6d4 e3d4 b6c6 g2g4 f7f6 f1e1 f6f5 a1b1 f5g4 d1g4');
INSERT INTO evaluations VALUES(55414,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(55414,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(55414,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(55414,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(55414,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(55414,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(55414,36,'cp 63','d2d4 b8d7 d1d2 c7c6 b2b3 g8f6 c1b2 f8e7 e1c1 d8c7 f1d3 d7f8 c3e2 f8g6 d2c3 c8g4 d4e5 d6e5 f3e5 c7e5 c3e5 g6e5 b2e5 g4e2 d3e2 f6e4 h1e1 e4f2 e2f3 f2d1 e5g7 e8d7 f3g4 d7d6 e1d1 d6c5 d1d7 e7g5 c1b2 c5b6 g7d4 b6a6 d4h8 a8h8 d7f7 h8d8 g2g3 d8d2');
INSERT INTO evaluations VALUES(55409,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(55409,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(55409,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(55409,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(55409,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(55409,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(55409,36,'cp 59','b8d7 b2b3 c7c6 c1b2 f8e7 d1d2 g8f6 e1c1 d8c7 h1g1 d7f8 c1b1 c8g4 h2h3 g4f3 g2f3 f8g6 f3f4 g6f4 g1g7 f4e6 d4e5 d6e5 g7g1 a8d8 f1d3 e7c5 d2e2 h8g8 h3h4 g8g1 d1g1 e6f4 e2f3 c5d4 c3e2 d4b2 b1b2 f4d3 c2d3 e8e7');
INSERT INTO evaluations VALUES(10592,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(10592,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(10592,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(10592,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(10592,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(10592,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(10592,36,'cp 63','c1f4 e8g8 h2h3 g4h6 f4h6 g7h6 c2c3 d4c3 b1c3 d7d6 c3e4 c8f5 e4c5 d6c5 d1e2 c6d4 f3d4 d8d4 a1d1 d4f4 e5e6 f7e6 c4e6 g8h8 e6f5 f4f5 e2e3 a8e8 e3h6 f5f4 h6f4 f8f4 d1d7 f4b4 d7c7 b4b2 a2a3 b2b3 f2f3 b3a3 c7b7');
INSERT INTO evaluations VALUES(10588,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(10588,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(10588,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(10588,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(10588,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(10588,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(10588,36,'cp 62','e8g8 h2h3 g4h6 f4h6 g7h6 c2c3 d7d6 c3d4 c6d4 f3d4 d8h4 d4b5 h4c4 b5c7 a8b8 b1d2 c4d4 d2f3 d4d1 a1d1 d6e5 f3e5 c5b6 c7d5 b6d8 d1d4 f8e8 e5c4 b7b5 c4d6 e8e2 d5f4 e2b2 f1e1 c8d7 d6f7 d8b6 d4d7 b6f2 g1f1 f2e1 f1e1 a7a5 f7h6 g8h8 h6f7 h8g8 f4d5 b8f8 f7e5 f8f2 e5g4 f2g2 g4f6 g8f8');
INSERT INTO evaluations VALUES(4489,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(4489,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(4489,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(4489,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(4489,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(4489,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(4489,36,'cp 60','h2h3 g4h6 f4h6 g7h6 c2c3 d4c3 b1c3 d7d6 c3e4 c8f5 e4c5 d6c5 d1b3 d8e7 f1e1 a8d8 a1c1 c6d4 f3d4 c5d4 b3b7 d4d3 c1d1 e7c5 b7b3 d3d2 e1e3 d8d4 c4d3 f5d3 e3d3 f8d8 d1d2 d4d3 d2d3 d8d3 b3d3 c5e5 d3c3 e5d5 c3g3 g8f8 g3a3 c7c5 a3a7 d5d1 g1h2 d1d6 g2g3');
INSERT INTO evaluations VALUES(4488,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(4488,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(4488,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(4488,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(4488,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(4488,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(4488,36,'cp 52','g4h6 f4h6 g7h6 c2c3 d4c3 b1c3 d7d6 c3e4 c8f5 e4c5 d6c5 d1b3 d8e7 f1e1 c6a5 b3c3 a5c4 c3c4 f5g6 a1c1 a8d8 a2a3 d8d3 e5e6 b7b6 b2b4 d3d6 e6f7 e7f7 c4f7 f8f7 b4c5 d6d5 e1e5 d5e5 f3e5 f7f5 c5b6 c7b6 c1c8 g8g7 c8c7 g6f7 f2f4 f5f4 c7a7 g7g8 a7f7 f4f7 e5f7 g8f7 g1h2 f7e6 h2g3 e6d5 g3g4');
INSERT INTO evaluations VALUES(4486,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(4486,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(4486,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(4486,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(4486,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(4486,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(4486,36,'cp 62','f4h6 g7h6 c2c3 d7d6 c3d4 c6d4 f3d4 d8h4 d4b5 h4c4 b5c7 a8b8 b1d2 c4d4 d2f3 d4d1 f1d1 d6e5 f3e5 c5e7 e5d7 c8d7 d1d7 e7f6 c7d5 f6b2 a1b1 b2e5 b1b7 g8g7 d5e3 b8b7 d7b7 f8c8 b7a7 c8c1 e3f1 e5d6 a7d7 d6c5 g2g3 g7g6 g1g2 h6h5 a2a4 c1a1 d7c7');
INSERT INTO evaluations VALUES(4485,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(4485,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(4485,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(4485,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(4485,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(4485,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(4485,36,'cp 55','g7h6 c2c3 d7d6 c3d4 c6d4 f3d4 d8h4 d4b5 h4c4 b5c7 a8b8 b1d2 c4d4 d2f3 d4d1 f1d1 d6e5 f3e5 c5e7 e5d7 c8d7 d1d7 e7f6 c7d5 f6b2 a1b1 b2e5 d7b7 g8g7 g1f1 b8b7 b1b7 e5d4 b7b4 d4c5 b4g4 g7h8 g4c4 c5b6 d5e7 h8g7 c4c6 f7f6 c6d6');
INSERT INTO evaluations VALUES(4112,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(4112,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(4112,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(4112,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(4112,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(4112,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(4112,36,'cp 54','c2c3 d7d6 c3d4 c6d4 f3d4 d8h4 d4b5 h4c4 b5c7 a8b8 b1d2 c4d4 d2f3 d4d1 f1d1 d6e5 f3e5 c5e7 e5d7 c8d7 d1d7 e7f6 c7d5 f6b2 a1b1 b2e5 b1b7 g8g7 d5e3 a7a5 g2g3 b8b7 d7b7 f8b8 b7a7 b8b5 e3c4 b5c5 c4a5 c5c2 a5b3 h6h5 a2a4 g7g6 a4a5 h5h4 g3h4 c2a2 h4h5 g6h5 a7f7 h5g6');
INSERT INTO evaluations VALUES(19381,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(19381,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(19381,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(19381,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(19381,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(19381,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(19381,36,'cp -31','e7e6 b1c3 g8e7 d2d3 g7g6 c4b3 f8g7 g1f3 d7d5 e1g1 e8g8 e4e5 b7b6 c3e2 a7a5 c2c3 c8a6 e2g3 b6b5 b3c2 d5d4 d1e1 d4c3 b2c3 c5c4 d3c4 b5b4 f1f2 a6c4 g3e4 d8c7 h2h4 e7d5 h4h5 b4c3 e4c3 c6b4 c3d5 c4d5 c2e4 f8c8 c1d2 b4c2 e4c2 c7c2 h5g6');
INSERT INTO evaluations VALUES(18759,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(18759,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(18759,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(18759,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(18759,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(18759,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(18759,36,'cp -25','b1c3 a7a6 a2a3 b7b5 c4a2 c8b7 d2d3 d8h4 g2g3 h4d8 g1f3 g8f6 h1f1 d7d5 e4d5 e6d5 g3g4 f6g4 d1e2 f8e7 c3d5 e8g8 d5e7 d8e7 e2e7 c6e7 h2h3 g4f6 c1e3 a8c8 e1c1 e7f5 e3d2 b7d5 c2c4 d5f3 f1f3 f8e8 f3f2 h7h6 c1b1 c8b8 c4b5 a6b5 d2c3 e8e3');
INSERT INTO evaluations VALUES(18756,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(18756,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(18756,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(18756,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(18756,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(18756,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(18756,36,'cp -33','a7a6 d2d3 b7b5 c4b3 c8b7 g1f3 g8f6 e4e5 f6g4 d1e2 d7d5 e5d6 f8d6 c3e4 e8g8 c2c3 g4f6 f3g5 f6e4 d3e4 h7h6 e4e5 h6g5 e5d6 d8d6 f4g5 d6e5 e2e5 c6e5 e1g1 c5c4 b3c2 f8d8 c1e3 e5d3 c2d3 d8d3 e3f4 a6a5');
INSERT INTO evaluations VALUES(17372,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(17372,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(17372,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(17372,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(17372,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(17372,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(17372,36,'cp -31','a2a3 b7b5 c4a2 b5b4 a3b4 c5b4 c3a4 d7d5 e4d5 e6d5 g1f3 g8f6 e1g1 f8d6 f3e5 d8c7 d2d4 e8g8 c2c3 b4c3 b2c3 c6e7 c1e3 a6a5 a4c5 a5a4 c3c4 d5c4 e3f2 d6c5 d4c5 c4c3 d1d6 c7d6 c5d6 e7f5 f1c1 f5d6 f2c5 f8d8 c1c3 c8e6 a2e6');
INSERT INTO evaluations VALUES(20099,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(20099,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20099,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(20099,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(20099,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(20099,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(20099,36,'cp 17','f1e2 g8f6 g1f3 g7g6 e1g1 f8g7 f1d1 e8g8 b1c3 c8g4 c3d5 g4f3 e2f3 f6d7 c2c3 a8c8 e3e2 c6e5 c1g5 f8e8 h2h3 h7h6 g5f4 a7a6 a2a4 d7c5 a1a3 e7e6 d5e3 d8e7 b2b4 c5d7 a4a5 e5f3 e2f3 d7e5 f3e2');
INSERT INTO evaluations VALUES(32118,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(32118,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(32118,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(32118,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(32118,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(32118,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(32118,36,'cp 86','g4f3 g2f3 g8f6 b1c3 c7c6 c1e3 b8d7 h3h4 f8e7 d1d2 e8g8 h1g1 d8c7 e1c1 f8d8 c3e2 f6h5 f3f4 d7f6 f1g2 b7b5 d2c3 b5b4 c3b4 d6d5 b4d2 e5d4 e3d4 f6e4 g2e4 d5e4 d2e3 e7f6 e3e4 d8e8 e4f3 f6d4 e2d4 h5f4');
INSERT INTO evaluations VALUES(32127,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(32127,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(32127,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(32127,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(32127,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(32127,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(32127,36,'cp 95','g2f3 g8f6 b1c3 c7c6 c1e3 b8d7 a2a3 f8e7 f3f4 e8g8 d1f3 d8a5 f1e2 b7b5 e1g1 a5c7 e2d3 a7a6 c3e2 e5d4 e2d4 g7g6 f3g2 f8e8 d4f5 e7f8 a1d1 d7c5 f5g3 a6a5 g1h1 f6d7 f1g1 f8h6 g2f3 a8d8 b2b4 c5d3');
INSERT INTO evaluations VALUES(32125,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(32125,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(32125,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(32125,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(32125,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(32125,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(32125,36,'cp 74','b8d7 c1e3 c7c6 b1c3 g8f6 d1d2 e5d4 e3d4 d7c5 e1c1 c5e6 d4e3 d8a5 f3f4 d6d5 e4e5 f6d7 f4f5 e6c7 f2f4 e8c8 a2a3 f7f6 e5e6 d7b6 h3h4 f8e7 c3e4 a5d2 e4d2 g7g6 c2c3 g6f5 f1h3 c7b5 h1g1 d5d4 c3d4 b6d5 d2c4 b5d6 c4d6 e7d6 e3f2');
INSERT INTO evaluations VALUES(23216,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(23216,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(23216,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(23216,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(23216,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(23216,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(23216,36,'cp 97','g8e7 d3d4 e5d4 f3d4 c6d4 d1d4 b7b6 c1f4 c8b7 b1c3 e7g6 f4g3 f8c5 d4d2 g6e5 c4b3 d8e7 a2a3 e7d6 d2e2 e8c8 c3b5 d6e7 e1c1 c8b8 b5c3 d7d6 c1b1 h7h5 h2h4 a7a5 b3a2 e7d7 f2f3 h8e8 c3d5 b8a7 h1e1 b7a6 e2d2 a6c4 a2c4 e5c4');
INSERT INTO evaluations VALUES(17139,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(17139,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(17139,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(17139,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(17139,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(17139,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17139,36,'cp 87','d3d4 e5d4 f3d4 c6d4 d1d4 e7c6 d4d1 b7b6 a2a3 c8b7 b1c3 d8e7 c4d5 e8c8 c1e3 g7g6 d1d2 f8g7 e1c1 d7d6 g2g4 h7h5 h2h3 h5g4 h3g4 h8h1 d1h1 c6e5 d2e2 e7d7 h1g1 d8e8 f2f4 b7d5 c3d5 e5c6 e2d3 c8b8 c1b1 d7e6 g1d1 e6e4 d3a6');
INSERT INTO evaluations VALUES(46186,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46186,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(46186,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(46186,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(46186,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(46186,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(46186,36,'cp 0','d4c5 f8c5 c4d5 f6d5 c3d5 e6d5 g1f3 e8g8 e2e3 b8c6 f1d3 d8f6 h2h3 c5b6 e1g1 f8e8 a1b1 c8f5 b2b4 f5e4 c1b2 f6g6 f3h4 g6e6 h4f3 a8d8 b1c1 h7h6 f1e1 e6g6 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(46093,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(46093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46093,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(46093,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(46093,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(46093,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(46093,36,'cp 0','f8c5 c4d5 f6d5 c1d2 e8g8 g1f3 d5c3 d2c3 d8d1 a1d1 c8d7 g2g3 d7a4 d1c1 f8d8 f1g2 b8c6 f3e5 c5d4 g2c6 d4c3 b2c3 a4c6 e5c6 b7c6 c1b1 a8b8 e1g1 g7g6 f2f4 b8b1 f1b1 d8d2 b1b8 g8g7 g1f2 d2a2 b8c8 a2a3 c8c6 a7a5 c6a6 a3c3 a6a5 g7h6 a5b5 h6g7');
INSERT INTO evaluations VALUES(40607,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40607,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40607,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(40607,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(40607,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(40607,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(40607,36,'cp -8','g1f3 e8g8 c4d5 f6d5 c1d2 d5c3 d2c3 d8d1 a1d1 c8d7 g2g3 d7a4 d1c1 f8d8 f1g2 b8c6 f3e5 c5d4 g2c6 d4c3 b2c3 a4c6 e5c6 b7c6 c1b1 a8b8 e1g1 g7g6 f2f3 g8g7 a3a4 b8b1 f1b1 d8d5 g1f2 d5c5 b1b7 a7a5 b7a7 g7f6 h2h4 h7h5 f2g2 c5c3 a7a5');
INSERT INTO evaluations VALUES(46484,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(46484,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(46484,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(46484,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(46484,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(46484,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(46484,36,'cp 66','c4b5 b8d7 f3e5 d5d4 d1e2 f8e7 e2c4 e8g8 e5d7 f6d7 c4d4 e7c5 d4f4 f8e8 b5e2 d7f8 d2d3 f8e6 f4d2 b7b5 e1g1 c8b7 c3e4 c5b6 c2c3 b5b4 f1e1 b7e4 d3e4 d8f6 e2f3 a8d8 d2c2 b4c3 b2c3 e6f4 c1e3 f4d3 e1d1 b6e3 d1d3 d8d3 c2d3 e3b6 d3c4 g7g6 a1d1');
INSERT INTO evaluations VALUES(46383,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46383,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(46383,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(46383,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(46383,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(46383,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(46383,36,'cp 64','b8d7 f3e5 d5d4 d1e2 f8e7 e2c4 e8g8 e5d7 f6d7 c4d4 e7c5 d4f4 f8e8 b5e2 d7e5 e1g1 e5g6 f4a4 c8d7 a4c4 a8c8 d2d4 d7e6 c4a4 d8d4 e2b5 e8e7 c1g5 f7f6 g5e3 d4a4 b5a4 c5e3 f2e3 g8f8 e3e4 e6c4 f1d1 b7b5 a4b3 c4b3 a2b3 b5b4 c3d5 e7e4');
INSERT INTO evaluations VALUES(50498,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(50498,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(50498,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(50498,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(50498,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(50498,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(50498,36,'cp -25','d7d5 e4d5 e6d5 c4a2 f8e7 e1g1 g8f6 h2h3 e8g8 d2d3 h7h6 b1c3 b8c6 f1e1 f8e8 c1f4 c8e6 c3e2 e7f8 e2g3 d8b6 a1b1 a8d8 c2c3 c6e7 f3e5 f6d7 d3d4 c5d4 d1d4 b6d4 c3d4 d7e5 f4e5 d8c8 g3e2 e7g6 e2f4 g6f4 e5f4');
INSERT INTO evaluations VALUES(50778,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(50778,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(50778,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(50778,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(50778,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(50778,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(50778,36,'cp -19','e4d5 e6d5 c4a2 f8e7 e1g1 g8f6 f1e1 e8g8 h2h3 b8c6 b1c3 c8f5 d2d3 h7h6 c1f4 f8e8 f3e5 e7d6 e5c6 e8e1 d1e1 b7c6 e1d2 f6h5 f4d6 d8d6 c3e2 a8e8 a1e1 d6e5 d2c3 e5c3 b2c3 h5f4 g1f1 f4e2 e1e2 g8f8 e2e8 f8e8 f2f4 e8d7 f1e1');
INSERT INTO evaluations VALUES(50771,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(50771,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(50771,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(50771,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(50771,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(50771,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(50771,36,'cp -24','e6d5 c4a2 f8e7 e1g1 g8f6 d2d4 c5c4 c1g5 e8g8 c2c3 c8g4 b1d2 b8d7 g5f4 f8e8 f1e1 g4h5 a2b1 d8b6 b1f5 e7d6 f4d6 e8e1 d1e1 b6d6 f3e5 a8e8 f2f4 h5g6 f5d7 f6d7 e1f2 g6f5 d2f1 f5e4 e5d7 d6d7 a1e1 h7h6 h2h3 g7g6 f1d2 g8g7');
INSERT INTO evaluations VALUES(50802,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(50802,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(50802,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(50802,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(50802,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(50802,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(50802,36,'cp -14','c4a2 g8f6 e1g1 f8e7 f1e1 e8g8 b1c3 b8c6 h2h3 h7h6 d2d4 c5c4 c1f4 f8e8 c3e2 e7d6 f4d6 d8d6 c2c3 c8f5 e2g3 e8e1 f3e1 f5g6 a2b1 c6e7 e1f3 g6b1 a1b1 e7g6 d1c2 a8e8 b1e1 e8e1 f3e1 d6e6 e1f3 f6e4 c2c1 b7b5 c1e3 e4g3 e3e6 f7e6 f2g3 g8f7 g1f2 f7e7 f2e2 e7d6 e2f2 e6e5');
INSERT INTO evaluations VALUES(50805,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(50805,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(50805,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(50805,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(50805,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(50805,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(50805,36,'cp -23','f8e7 e1g1 g8f6 f1e1 e8g8 b1c3 h7h6 d2d4 c5c4 c1f4 b8c6 h2h3 f8e8 c3e2 e7d6 f4d6 d8d6 e2g3 e8e1 f3e1 c6e7 c2c3 e7f5 g3f1 f5h4 f1e3 c8d7 e1f3 h4f3 d1f3 a8e8 a1e1 b7b5 a2b1 g7g6 f3e2 g8g7 e2d2 h6h5');
INSERT INTO evaluations VALUES(50606,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(50606,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(50606,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(50606,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(50606,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(50606,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(50606,36,'cp 47','d7d5 e4d5 d8d5 d2d4 e7e6 c1e3 c5d4 c3d4 g8f6 b1c3 d5d8 f1d3 b8c6 a2a3 f8e7 e1g1 e8g8 c3e4 f6g4 a1c1 e6e5 f3e5 c6e5 d4e5 g4e3 f2e3 d8b6 d1e2 c8d7 b2b4 a8c8 c1c8 f8c8 d3c4 d7e6 c4e6');
INSERT INTO evaluations VALUES(50695,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(50695,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(50695,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50695,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(50695,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(50695,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(50695,36,'cp 44','e4d5 d8d5 d2d4 e7e6 c1e3 b8d7 g2g3 c5d4 e3d4 d5e4 d4e3 g8f6 b1d2 e4c6 f1g2 f8c5 e3c5 c6c5 f3d4 c5c7 d1e2 e8g8 e1g1 a8b8 e2c4 c7d8 a2a4 e6e5 d4b3 b7b5 a4b5 a6b5 c4e2 h7h6 f1e1 c8b7');
INSERT INTO evaluations VALUES(50694,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(50694,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(50694,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(50694,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(50694,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(50694,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(50694,36,'cp 40','d8d5 f1e2 g8f6 b1a3 e7e6 e1g1 f8d6 a3c4 d6c7 d2d4 e8g8 d4c5 d5c5 c1e3 c5e7 c4b6 c7b6 e3b6 f6d5 d1b3 e7d6 b6d4 b8c6 f1d1 c6d4 d1d4 b7b5 a2a4 c8d7 e2d3 d6c7 a4b5 a6b5 a1e1 d7c6 f3e5 d5e7');
INSERT INTO evaluations VALUES(34113,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(34113,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(34113,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(34113,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(34113,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(34113,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(34113,36,'cp 55','f1e2 e7e6 b1a3 g8f6 e1g1 f8d6 d2d4 c5d4 a3c4 d6c7 d1d4 d5d4 f3d4 e8g8 a2a4 b8d7 e2f3 d7c5 c1e3 c8d7 a4a5 f8e8 c4b6 c7b6 a5b6 e6e5 d4c6 c5b3 a1a3 e5e4 f3d1 d7c6 d1b3 f6d5 c3c4 d5e3 f2e3');
INSERT INTO evaluations VALUES(54412,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(54412,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(54412,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(54412,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(54412,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54412,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(54412,36,'cp 69','b8c6 b2b3 b7b5 f1e2 a7a6 e1g1 g8e7 f1e1 c8d7 c1a3 e7g6 a3f8 g6f8 h2h4 h7h6 b1d2 f8g6 g2g3 e8g8 d2f1 g6e7 f1e3 c4b3 a2b3 a6a5 h4h5 d8b6 d1d2 b5b4 c3c4 d5c4 e3c4 b6d8 e2d3 e7d5 c4d6 c6e7');
INSERT INTO evaluations VALUES(20410,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(20410,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(20410,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(20410,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(20410,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(20410,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(20410,36,'cp 82','f1e2 b7b5 b2b3 a7a6 e1g1 g8e7 b1a3 e7g6 a3c2 f7f6 e5f6 d8f6 c1g5 f6f7 c2e3 f8d6 f3h4 g6h4 g5h4 e8g8 e2f3 f7c7 h4g3 c8d7 f1e1 c6e7 b3c4 b5c4 d1d2 e7g6 f3d1 g6f4 d1c2 h7h5 g3f4 d6f4 g2g3 f4h6 f2f4 a8b8 e3g2 c7d8 a1b1');
INSERT INTO evaluations VALUES(20409,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(20409,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(20409,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(20409,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(20409,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(20409,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(20409,36,'cp 93','b7b5 b2b3 a7a6 e1g1 g8e7 b1a3 e7g6 a3c2 f7f6 e5f6 d8f6 c1g5 f6f7 c2e3 f8d6 b3c4 b5c4 f3h4 g6h4 g5h4 e8g8 h4g3 f7c7 d1e1 c8d7 e2d1 a8e8 e1d2 e8b8 d1c2 d7e8 d2e2 e8f7 a1e1 d6g3 h2g3 e6e5 d4e5 c7e5 e2d2 g8h8');
INSERT INTO evaluations VALUES(9997,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9997,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(9997,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9997,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9997,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9997,36,'cp 0','c6d4 b5a4 b7b5 a4b3 d7d6 h2h3 a7a6 e1g1 d4b3 a2b3 c8b7 c1e3 e8g8 c3a2 b4a5 a2c3');
INSERT INTO evaluations VALUES(10749,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10749,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10749,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10749,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10749,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10749,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10749,36,'cp 0','b5a4 b7b5 a4b3 d7d6 h2h3 a7a6 e1g1 d4b3 a2b3 c8b7 c1e3 e8g8 c3a2 b4a5 a2c3');
INSERT INTO evaluations VALUES(12906,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12906,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12906,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12906,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12906,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12906,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12906,36,'cp 0','b1c3 e7e6 g1f3 d8c7 c3b5 c7b8 b5d4 e6e5 d4f5 d7d5 e4d5 f6d5 f1d3 b8c7 e1g1 g7g6 f5e3 d5b6 a4h4 f8e7 h4h6 e7f8');
INSERT INTO evaluations VALUES(12905,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12905,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12905,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12905,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12905,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12905,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12905,36,'cp 0','e7e6 g1f3 d8c7 c3b5 c7b8 b5d4 e6e5 d4f5 d7d5 e4d5 f6d5 f1d3 b8c7 e1g1 g7g6 f5e3 d5b6 a4h4 f8e7 h4h6 e7f8');
INSERT INTO evaluations VALUES(12508,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12508,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12508,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12508,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12508,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12508,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12508,36,'cp 0','g1f3 d8c7 c3b5 c7b8 b5d4 e6e5 d4f5 d7d5 e4d5 f6d5 f1d3 b8c7 e1g1 g7g6 f5e3 d5b6 a4h4 f8e7 h4h6 e7f8');
INSERT INTO evaluations VALUES(3790,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(3790,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(3790,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(3790,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(3790,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(3790,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(3790,36,'cp 113','c3d5 c8e6 g5f6 g7f6 d1d2 g8g7 e1g1 f8g8 b2b4 c5b6 a2a4 a7a5 d5b6 c7b6 b4a5 a8a5 c4e6 f7e6 f1b1 d8c7 c2c4 g8a8 d2d1 g7h8 g2g3 c6b8 d3d4 b8d7 g1g2 a8g8 b1b5 e5d4 d1d4 e6e5 d4d3 d7c5 d3d5 c7d7 b5b6 a5a4 a1a4');
INSERT INTO evaluations VALUES(3873,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(3873,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(3873,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(3873,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(3873,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(3873,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(3873,36,'cp 115','c8e6 d5f6 g7f6 g5h6 f8e8 c4e6 f7e6 h2h4 g8h8 g2g4 a7a5 d1e2 d6d5 e1c1 a5a4 d1g1 a4a3 b2b3 e8g8 c1b1 d5e4 d3e4 c6d4 e2d3 d4c6 d3c4 c5f2 g1g2 f2a7 c4e6 d8e7 e6e7 c6e7 g4g5 a7e3 g2e2 e3f4 g5f6 f4h6 f6e7');
INSERT INTO evaluations VALUES(23744,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(23744,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(23744,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(23744,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(23744,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(23744,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(23744,36,'cp 96','g5f6 g7f6 d1d2 g8g7 e1g1 c6d4 f3d4 c5d4 c2c3 d4b6 a2a4 e6d5 c4d5 c7c6 d5b3 f6f5 e4f5 d8f6 a4a5 b6c7 d3d4 f6f5 d4e5 d6e5 b3c2 e5e4 d2d4 f5e5 d4e5 c7e5 c2e4 a8d8 a1d1 d8d1 f1d1 e5c7 a5a6 f8e8 e4d3 b7a6 d3a6 e8d8 d1d8 c7d8 a6d3 d8b6 g2g3 h7h6 d3e2 f7f5');
INSERT INTO evaluations VALUES(4144,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(4144,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(4144,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4144,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(4144,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(4144,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(4144,36,'cp 10','g5h4 d7d6 c2c3 a7a5 b1d2 g7g5 h4g3 c5a7 f1e1 g5g4 f3h4 f6h5 d2f1 d8g5 f1e3 a7e3 e1e3 c6e7 d3d4 h5f4 f2f3 h6h5 d1d2 e7g6 h4g6 f4g6 f3g4 h5h4 g3f2 c8g4 e3d3 g5h5 f2e3 g4e6 c4e6 f7e6 d2d1 h5d1 d3d1 e5d4 c3d4 g6f4 d1d2');
INSERT INTO evaluations VALUES(4160,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4160,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4160,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(4160,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(4160,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4160,36,'cp 20','d7d6 c2c3 a7a5 b1d2 g7g5 h4g3 c5a7 f1e1 g5g4 f3h4 f6h5 d2f1 d8g5 f1e3 a7e3 e1e3 c6e7 d3d4 h5f4 d1d2 c8e6 c4e6 f7e6 a1f1 f8f7 d4e5 d6e5 f2f3 a8f8 f1f2 h6h5 e3e1 e7g6 h4g6 g5g6 g1h1 g6g5 f3g4 g5g4 f2f3 h5h4 g3f4 f7f4 e1g1 f8f7 h2h3 g4h5 f3f4 e5f4');
INSERT INTO evaluations VALUES(3600,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(3600,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(3600,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(3600,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3600,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(3600,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(3600,36,'cp 17','b1c3 c8g4 h2h3 g4e6 c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 c2c3 d8e8 f1e1 e8g6 e1e3 f6h5 g1h2 f8f7 h4g3 h5f4 d3d4 g6h5 h2g1 a8f8 g3f4 f7f4 d1e2 e5d4 c3d4 h5g6 a1c1 d6d5 e4d5 e6d5 g2g3 f4e4 e3e4 g6e4 e2e4 d5e4 f3e5 c6e5 d4e5');
INSERT INTO evaluations VALUES(44209,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(44209,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(44209,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(44209,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(44209,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44209,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(44209,36,'cp 54','d4d5 g7g6 b1c3 f8g7 g1f3 d7d6 h2h3 e8g8 f1d3 e6d5 c4d5 b8a6 e1g1 a6c7 e3e4 b7b5 f1e1 f8e8 a2a3 a7a5 c1f4 c5c4 d3c2 c8b7 d1d2 b5b4 a3b4 a5b4 a1a8 b7a8 c3a4 b4b3 c2b1 c7a6 a4c3 a6c5 f4h6 g7h6 d2h6 a8b7 f3g5 d8b6 g1h1 c5d3 b1d3 c4d3');
INSERT INTO evaluations VALUES(11863,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(11863,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(11863,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(11863,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11863,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(11863,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(11863,36,'cp -9','e1g1 c6c5 b1c3 f8e7 f3d2 e8g8 d2c4 c8e6 b2b3 h7h6 f2f4 e5f4 c1f4 c7c6 e4e5 d6e5 c4e5 a8c8 g1h1 f6d5 c3d5 c6d5 d1f3 e7f6 c2c4 f6e5 f4e5 f7f6 e5c3 d8d7 h2h3 c8e8 a1e1 e6f7 f3f5 e8e1 c3e1 d7f5 f1f5 f7e6 f5f1 d5d4 e1d2 f8e8');
INSERT INTO evaluations VALUES(11862,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(11862,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(11862,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(11862,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(11862,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(11862,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(11862,36,'cp -8','f8e7 b1c3 c6c5 f3d2 c8e6 d2c4 e8g8 b2b3 d8d7 f2f4 e5f4 c1f4 a8e8 h2h3 h7h6 d1f3 f6h7 c3d5 e6d5 e4d5 f7f5 a1e1 e7f6 e1e8 d7e8 f4d2 f5f4 g1h1 f6d4 c2c3 d4e5 f3e2 e8g6 c4e5 d6e5 f1e1 f4f3 g2f3 g6g3 e2g2 g3f3 g2f3 f8f3');
INSERT INTO evaluations VALUES(7535,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(7535,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(7535,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(7535,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7535,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7535,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7535,36,'cp 0','b1c3 c8g4 h2h3 g4h5 c3e2 h5f3 g2f3 d8d7 g1g2 f6h5 e2g3 h5f4 c1f4 e5f4 g3h5 e8g8 h5f4 f7f5 f4e2 d7e6 e2d4 e6g6 g2h2 e7g5 d4f5 g5f4 h2h1 f8f5 e4f5 g6f5 h1g2 f5g5 g2h1 g5h4 h1g2');
INSERT INTO evaluations VALUES(7531,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7531,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7531,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7531,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7531,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7531,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7531,36,'cp 0','c8g4 h2h3 g4h5 c3e2 h5f3 g2f3 d8d7 g1h2 d6d5 f3f4 e5f4 e4e5 f4f3 e2d4 c6c5 d4f3 f6h5 d3d4 e8g8 f3e1 g7g6 c1h6 f8b8 d1h5 g6h5 f1g1 g8h8 h6g7 h8g8');
INSERT INTO evaluations VALUES(52884,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(52884,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(52884,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(52884,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(52884,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(52884,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(52884,36,'cp 89','f1c4 c6c5 g1f3 g8f6 e1g1 c5d4 b1c3 e7e6 e3d4 f8e7 a2a3 e8g8 f1e1 c8d7 d1e2 d7c6 f3e5 c6e8 c1e3 b8d7 e5d7 e8d7 d4d5 e6d5 c3d5 f8e8 a1d1 d8c8 d1c1 d7c6 d5e7 e8e7 f2f3 c8f5 e2f2 c6d5 c4f1 d5c6 e3a7 e7e1 c1e1');
INSERT INTO evaluations VALUES(52872,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(52872,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(52872,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(52872,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(52872,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(52872,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(52872,36,'cp 75','g8f6 g1f3 c6c5 e1g1 c5d4 e3d4 e7e6 b1c3 f8e7 f1e1 e8g8 a2a3 c8d7 d1e2 d7c6 f3e5 c6e8 c1e3 b8d7 e5d7 e8d7 d4d5 e6d5 c3d5 f6d5 c4d5 e7f6 a1b1 d7c6 d5c6 b7c6 b2b4 f8e8 e1d1 d8c8 e2f3 e8e6 d1c1 c8e8 c1c5 a8c8 b1c1 f6e7 c5a5 e7f6');
INSERT INTO evaluations VALUES(45007,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(45007,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(45007,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(45007,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(45007,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(45007,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(45007,36,'cp 90','g1f3 c6c5 e1g1 c5d4 e3d4 e7e6 b1c3 f8e7 d1e2 e8g8 f1d1 c8d7 d4d5 e6d5 c3d5 b8c6 c1e3 f6d5 c4d5 d8e8 d5e4 g8h8 a1c1 a8d8 h2h3 f7f5 e4b1 e7f6 a2a3 e8h5 e3f4 f8e8 e2c2 d7e6 b2b4 h7h6 b4b5 d8d1 c1d1');
INSERT INTO evaluations VALUES(45008,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(45008,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(45008,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(45008,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(45008,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(45008,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(45008,36,'cp 81','c6c5 e1g1 e7e6 d1e2 b8c6 b1c3 c5d4 f1d1 f8d6 e3d4 e8g8 c1g5 d6e7 a1c1 c6b4 h2h3 h7h6 g5f6 e7f6 c4b3 c8d7 a2a3 b4c6 d4d5 e6d5 d1d5 d8e8 c3e4 f6e7 e2d3 d7e6 d5b5 e6b3 d3b3');
INSERT INTO evaluations VALUES(57511,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57511,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(57511,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(57511,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57511,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(57511,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(57511,36,'cp 29','b8c6 d3d4 f5e4 f3e5 g8f6 f1c4 d7d5 c4b5 c8d7 c2c4 f8d6 e5d7 d8d7 c4d5 d6b4 c1d2 b4d2 b1d2 d7d5 d1a4 e8c8 b5c6 d5c6 a4c6 b7c6 d2b3 f6d5 e1g1 d5f4 f1e1 h8e8 e1e3 f4d3 a1f1 d3b2 f1c1 b2d3 c1c6 d3b4 c6c4 b4d5 e3h3 e4e3 f2e3 d5e3 c4c5 e3d5 h3h7');
INSERT INTO evaluations VALUES(23302,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(23302,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(23302,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23302,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(23302,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(23302,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(23302,36,'cp 23','d3d4 f5e4 f3e5 g8f6 f1c4 d7d5 c4b5 c8d7 c2c4 f8d6 e5d7 d8d7 c4d5 f6d5 e1g1 d7f5 b1c3 d5c3 b5c6 b7c6 b2c3 e8g8 a1b1 a8e8 d1e2 e8e6 c3c4 c6c5 d4c5 d6c5 c1e3 a7a6 f1d1 c5e3 e2e3 h7h6 d1d2 g8h7 b1e1 f5g6 e3g3 f8e8 g3c7 e4e3 f2e3 e6e3 e1e3 e8e3 c7f4');
INSERT INTO evaluations VALUES(55870,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55870,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55870,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(55870,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55870,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(55870,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(55870,36,'cp 0','f8e7 g1f3 g8f6 f1e2 e8g8 e1g1 b8d7 b1d2 b7b6 b2b3 c8b7 c1b2 a7a5 c2c4 e7d6 a1c1 a8c8 f3e5 d8e7 a3a4 d6b4 e5d3 b4d6');
INSERT INTO evaluations VALUES(56982,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(56982,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(56982,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(56982,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(56982,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(56982,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(56982,36,'cp 14','b1c3 b8c6 f2f4 d7d6 g1f3 g7g6 c1e3 f8g7 d3d4 e5d4 f3d4 g8f6 f1e2 e8g8 e2f3 c6d4 e3d4 c7c5 d4f2 d8b6 a1b1 f6g4 c3d5 b6a5 c2c3 g4f2 e1f2 a5d8 g2g3 c8e6 f2g2 b7b5 d1d3 a8b8 h1d1 d8d7 h2h4 f8e8 b1c1 a7a5 d5e3 e6a2 d3d6 d7d6 d1d6');
INSERT INTO evaluations VALUES(56980,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(56980,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(56980,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(56980,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(56980,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(56980,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(56980,36,'cp 21','b8c6 f2f4 d7d6 g2g3 g7g6 f1g2 f8g7 c1e3 e5f4 g3f4 d8h4 e1d2 c8e6 g1e2 e8c8 d3d4 c6e7 d1g1 d6d5 e3f2 h4h5 e4e5 e7f5 h2h4 g8e7 a1e1 c8b8 g2h3 h5f3 d2c1 h6h5 c1b1 a7a6 a2a3 g7h6 g1h2 f5e3 f2e3');
INSERT INTO evaluations VALUES(12681,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(12681,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(12681,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(12681,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12681,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(12681,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(12681,36,'cp 40','d2d4 d7d5 e4d5 e7d5 e1g1 c5d4 c3d5 e6d5 c4b5 f8e7 f3d4 c8d7 d4b3 d7f5 f1e1 e8g8 c2c3 f8e8 c1f4 d8d7 h2h3 h7h6 d1f3 a7a6 b5f1 e7f8 e1e8 a8e8 a1d1 f5e4 f3g4 d7e6 g4e6 f7e6 f1e2 g8f7 f4e3 c6e5 b3c5 f8c5');
INSERT INTO evaluations VALUES(12690,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(12690,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12690,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(12690,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(12690,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(12690,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(12690,36,'cp 37','d7d5 e4d5 e7d5 e1g1 c5d4 c3d5 e6d5 c4b5 f8e7 f3d4 c8d7 d4e2 a7a6 b5a4 c6a5 c2c3 e8g8 a4c2 d7c6 f1e1 a5c4 e2d4 f8e8 b2b3 c4d6 d4c6 b7c6 d1f3 g7g6 g2g3 e7g5 e1e8 d6e8 c1b2 d8f6 g1g2 f6f3 g2f3 g5f6 a1d1 f6g7 h2h3 e8d6 c2d3');
INSERT INTO evaluations VALUES(27667,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(27667,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(27667,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(27667,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(27667,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(27667,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(27667,36,'cp 50','g8e7 b1c3 e7c6 d4a4 f8e7 c3d5 e8g8 a4b3 a8b8 f1e1 e7d8 c1f4 c6e7 a1d1 e7d5 d1d5 d8f6 c2c4 d7c6 f2f3 a7a6 e1c1 f8e8 f4e3 h7h6 e3a7 b8c8 a7f2 b7b6 b3c2 f6e5 b2b3 c6d7 c1d1 d7e6 h2h3 e5f4');
INSERT INTO evaluations VALUES(35485,30,'cp -213',NULL);
INSERT INTO evaluations VALUES(35485,31,'cp -217',NULL);
INSERT INTO evaluations VALUES(35485,32,'cp -209',NULL);
INSERT INTO evaluations VALUES(35485,33,'cp -209',NULL);
INSERT INTO evaluations VALUES(35485,34,'cp -211',NULL);
INSERT INTO evaluations VALUES(35485,35,'cp -226',NULL);
INSERT INTO evaluations VALUES(35485,36,'cp -218','d6d5 c4b3 c7c5 e4d5 d8e8 d1e2 e8e2 e1e2 b7b5 d2d3 c8a6 b1d2 b8d7 h1d1 a8e8 e2f1 f8d6 a2a4 b5b4 a4a5 d7e5 f1g1 g7g5 b3a4 e8e7 d2e4 f6e4 d3e4 h7h6 f2f3 f7g6 g1h1 h8f8 d1e1 e5c4 b2b3 d6e5 b3c4 e5a1 c1e3 a1c3');
INSERT INTO evaluations VALUES(35493,30,'cp -228',NULL);
INSERT INTO evaluations VALUES(35493,31,'cp -226',NULL);
INSERT INTO evaluations VALUES(35493,32,'cp -223',NULL);
INSERT INTO evaluations VALUES(35493,33,'cp -223',NULL);
INSERT INTO evaluations VALUES(35493,34,'cp -217',NULL);
INSERT INTO evaluations VALUES(35493,35,'cp -223',NULL);
INSERT INTO evaluations VALUES(35493,36,'cp -207','c4b3 c7c5 e4d5 d8e8 d1e2 e8e2 e1e2 b7b5 d2d3 b8d7 b1c3 c8a6 c1e3 c5c4 d3c4 b5c4 b3a4 f8b4 h1d1 a6b7 c3b5 d7b6 a2a3 b4f8 e3b6 a7b6 b5c3 f6d5 c3d5 b7d5 d1d5 a8a4 d5d7 f8e7 e2f1 c4c3 b2b3 a4e4 a3a4 e4b4 d7b7 h8e8 b7c7');
INSERT INTO evaluations VALUES(46242,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(46242,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(46242,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(46242,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(46242,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(46242,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46242,36,'cp -1','b7b6 b1d2 f8e7 e3e4 c5d4 f3d4 e6e5 d4f3 e8g8 e1g1 c8b7 d1e2 b8d7 e4d5 b7d5 d2e4 d5e4 d3e4 f6e4 e2e4 d8c7 c1f4 e7d6 f4g3 f8e8 a1d1 a8d8 e4d3 d7f8 d3e4 f7f6 d1d2 c7f7 f1d1 d6e7 a2a3 d8d2 d1d2');
INSERT INTO evaluations VALUES(28947,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(28947,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(28947,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(28947,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(28947,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(28947,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(28947,36,'cp 24','e7e6 d1b3 b7b6 g1f3 f8d6 f4d6 d8d6 f3h4 f5e4 c4d5 e6d5 a1c1 e8g8 c3e4 f6e4 f1d3 g7g6 h4f3 b8d7 e1g1 d7f6 b3a4 c6c5 d3a6 g8g7 b2b3 a8b8 c1c2 f8e8 f3e5 d6c7 f1c1 e8e7 a6d3');
INSERT INTO evaluations VALUES(29226,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(29226,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29226,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(29226,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(29226,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(29226,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(29226,36,'cp 31','d1b3 b7b6 g1f3 f8d6 f4d6 d8d6 f3h4 f5e4 c4d5 e6d5 c3e4 f6e4 f1d3 g7g6 b3c2 e8g8 e1g1 e4f6 b2b3 c6c5 d4c5 b6c5 a1c1 b8d7 h4f3 a8c8 f1d1 d7e5 f3e5 d6e5 h2h3 g8g7 c2c3 e5c3 c1c3 c8c7 c3c2 f8c8 d1c1 f6d7');
INSERT INTO evaluations VALUES(29225,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(29225,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(29225,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(29225,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29225,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(29225,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(29225,36,'cp 50','b7b6 c4d5 e6d5 g1f3 f8d6 f1e2 e8g8 e1g1 b6b5 a1c1 a7a5 b3d1 f8e8 h2h3 h7h6 f4d6 d8d6 f3e5 b8d7 e2d3 d7e5 d3f5 e5d7 c3e2 a5a4 d1c2 e8c8 e2f4 c8c7 f4d3 d7b6 c2c5 d6c5 d3c5 b6c4 b2b3');
INSERT INTO evaluations VALUES(28610,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(28610,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(28610,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(28610,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(28610,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(28610,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(28610,36,'cp 37','c4d5 f6d5 c3d5 e6d5 g1e2 f8d6 e2g3 f5e6 f1e2 e8g8 e1g1 d6f4 e3f4 d8f6 g3h5 f6h4 b3f3 f7f5 h5g3 h4f6 f3c3 g7g6 f1d1 e6f7 g3f1 f8e8 f1e3 f6d6 g2g3 b8d7 e2a6 a8b8 a1c1 c6c5 a6b5 e8c8 b5d7 d6d7 h2h4 c5c4 c3a3');
INSERT INTO evaluations VALUES(19106,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(19106,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(19106,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(19106,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(19106,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(19106,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19106,36,'cp 13','d7d5 e4d5 c6d5 c4b5 c8d7 b5d7 d8d7 b1d2 f8c5 d2b3 b7b6 e1g1 g8e7 c1g5 f7f6 g5d2 e8g8 a2a4 a7a5 d1h5 f8e8 h2h3 e7c6 h5h4 h7h6 f1e1 g8h7 g1f1 a8c8 e1e8 c8e8 h4f4 e8e5 a1e1 h7g8 f4h4 g8h7 e1e2 e5e2 f1e2');
INSERT INTO evaluations VALUES(20738,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(20738,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(20738,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(20738,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(20738,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(20738,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(20738,36,'cp 39','e4d5 c6d5 d1e2 f8e7 c4b3 a7a5 a2a4 a8a6 e1g1 a6g6 c1f4 g8h6 e2e5 c8e6 b1a3 h6f5 a3b5 e8g8 a1e1 f5h4 f4g3 h4f5 e5c7 d8a8 g3f4 e7b4 c2c3 b4e7 b5d4 f5d4 c3d4 e7b4 e1e3 h7h6 h2h3 b4d2 e3e2 d2f4 c7f4 a8a6 e2e5 a6b6 b3d5 e6d5 e5d5 b6b2');
INSERT INTO evaluations VALUES(20736,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(20736,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(20736,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(20736,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(20736,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(20736,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(20736,36,'cp 32','c6d5 d1e2 f8e7 c4b3 a7a5 a2a4 a8a6 e1g1 a6g6 c1f4 h7h5 b1a3 c8h3 f4g3 h3g4 f2f3 g4h3 f1f2 h5h4 g3e5 g8h6 a3b5 e8g8 e5d4 f8e8 e2e5 h6f5 e5d5 d8d5 b3d5 f5d4 b5d4 e7c5 c2c3 e8d8 d5e4 f7f5 e4b7 h3g2 f2g2 g6b6 b7c8 c5d4 c3d4 d8c8 g2e2 b6b3');
INSERT INTO evaluations VALUES(20783,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(20783,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(20783,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20783,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(20783,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(20783,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(20783,36,'cp 29','d1e2 f8e7 c4b3 g8f6 e1g1 a7a5 b3a4 c8d7 a4d7 d8d7 b1d2 e8g8 d2b3 f8e8 e2d2 d7a4 d2f4 a8c8 c2c3 a4b5 b3d4 b5d3 d4f5 e7f8 c1e3 a5a4 f5h6 g7h6 f4f6 c8c6 f6f3 d3e4 f3e4 d5e4 a1c1 a4a3 b2a3 e8c8 g2g3 c6c3 a3a4 c3c1 f1c1 c8c1 e3c1 f7f5 c1e3 f8g7');
INSERT INTO evaluations VALUES(31194,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(31194,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(31194,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(31194,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(31194,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(31194,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(31194,36,'cp -9','g8f6 e4f6 e7f6 f1d3 f8e7 g1e2 e8g8 e1g1 f5g6 c1e3 f8e8 e3f2 d8b6 e2f4 b8a6 f4g6 h7g6 a1b1 a6b4 d3c4 b4d5 c2c3 b6c7 d1b3 e7d6 g2g3 d5e3 f2e3 e8e3 g1f2 a8e8 f1e1 e3e1 b1e1 e8e7 e1e7 c7e7 f2g2 f6f5');
INSERT INTO evaluations VALUES(29071,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(29071,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(29071,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(29071,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(29071,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(29071,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(29071,36,'cp -4','e4f6 e7f6 f1d3 f5d3 d1d3 f8e7 g1e2 e8g8 e1g1 b8a6 c2c4 b7b5 c4b5 a6b4 d3b3 c6b5 a2a3 b4c6 d4d5 c6e5 f1d1 a7a6 c1f4 e7d6 a3a4 b5a4 b3a4 d8b6 g1h1 b6b2 e2g3 f8d8 g3e4 b2b4 e4d6 b4a4 a1a4 d8d6');
INSERT INTO evaluations VALUES(31485,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(31485,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(31485,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(31485,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(31485,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31485,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31485,36,'cp 0','c6c5 b1d2 b8c6 c2c3 c5d4 f3d4 g4e2 d1e2 c6d4 c3d4 g8e7 e1g1 a8c8 d2f3 d8b6 h2h4 e7f5 h4h5 f8e7 f1d1 c8c4 b2b3 f5d4 f3d4 c4d4 d1d4 b6d4 e2b5 e8d8 b5b7 d4d1 g1h2 d1h5 h2g1 h5e5 b7a8 d8d7 a8a7 e5c7 a7a4 c7c6 a4a7');
INSERT INTO evaluations VALUES(31644,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31644,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31644,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31644,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31644,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31644,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31644,36,'cp 0','b1d2 b8c6 c2c3 c5d4 f3d4 g4e2 d1e2 c6d4 c3d4 g8e7 e1g1 a8c8 d2f3 e7c6 h2h4 h7h6 c1e3 f8e7 g2g3 e8g8 f1c1 d8d7 f3e1 a7a6 g1g2 c6a7 e2g4 g8h8 g4f3 h8g8');
INSERT INTO evaluations VALUES(49951,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(49951,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(49951,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(49951,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(49951,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(49951,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(49951,36,'cp 65','b7b6 f1d3 f8e7 c2c4 c8b7 b1c3 e8g8 e1g1 f6e4 d3e4 f5e4 f3d2 d7d5 d1g4 f8f6 c4d5 f6g6 g4d1 e6d5 f2f3 e4f3 d2f3 b8c6 a2a3 e7f6 b2b4 d8d7 d1b3 c6d8 a3a4 d8e6 f4e5 a8f8 b4b5 c7c5 b5c6 b7c6');
INSERT INTO evaluations VALUES(43860,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43860,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(43860,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(43860,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(43860,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(43860,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(43860,36,'cp 58','a2a3 f8e7 f1c4 c8b7 b1c3 d7d5 c4d3 e7d6 e1g1 d8e7 a1c1 e8g8 c3e2 b8d7 c2c4 g8h8 d1b3 a8b8 f1e1 d5c4 d3c4 b7f3 f4d6 e7d6 g2f3 f6d5 e2f4 d5f4 e3f4 d6f4 b3e3 f4d6 e3e6 f8f6 e6d6 f6d6 e1e7 d6d4 c4e6 d7c5 e6f5');
INSERT INTO evaluations VALUES(32539,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(32539,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(32539,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(32539,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(32539,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(32539,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(32539,36,'cp 28','a4c4 g8f6 g1f3 e7e6 a2a3 f8d6 b1d2 d6e7 b2b3 e8g8 c1b2 b8a6 e2e3 c7c5 f1e2 c5d4 c4d4 a6c5 a1c1 d8b6 e1g1 a8c8 b3b4 c5a4 d4b6 a4b6 d2b3 f6e4 e2d3 e4d6 b3c5 d7e8 b2d4 e8b5 d3b5 d6b5 c5b7');
INSERT INTO evaluations VALUES(32533,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32533,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(32533,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(32533,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(32533,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(32533,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(32533,36,'cp 40','e7e6 a2a3 g8f6 g1f3 f8d6 g2g3 e8g8 f1g2 d7c6 b1c3 b8d7 e1g1 f6d5 f1e1 d7b6 c4d3 d5c3 b2c3 f7f5 c3c4 c6e4 d3b3 c7c5 a3a4 b6d7 c1a3 b7b6 e2e3 d8e7 a3b2 a8b8 a1d1 h7h6 f3d2 e4g2 g1g2 e6e5 d4e5 d6e5 d2f3 e5b2 b3b2');
INSERT INTO evaluations VALUES(32509,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(32509,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(32509,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(32509,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(32509,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(32509,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(32509,36,'cp 37','a2a3 g8f6 g1f3 b8c6 b1d2 a7a5 b2b3 a8c8 c1b2 f8e7 e2e3 e8g8 c4d3 c6b8 f1e2 c7c5 d4c5 c8c5 d2c4 d7b5 d3d8 f8d8 a3a4 b5c4 e2c4 c5c8 e1e2 b8d7 h1d1 e7b4 a1c1 f6e4 c4d3 d7c5 d3e4 c5e4 d1d8 c8d8 c1c7 e4c5 b2d4 c5b3 c7b7 b3d4 f3d4');
INSERT INTO evaluations VALUES(34932,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(34932,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(34932,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(34932,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(34932,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(34932,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(34932,36,'cp 34','g1e2 e5d4 d1d4 d5e6 c1f4 b8c6 d4d2 g8f6 b1c3 f8b4 e1c1 f6e4 d2e3 e4c3 e2c3 e6e3 f4e3 b4c3 b2c3 c8e6 f1b5 e8g8 a2a4 a7a6 b5e2 f8e8 e2f3 a8d8 d1d8 e8d8 e3f4 d8c8 h1e1 h7h6 f3c6 b7c6 f2f3 g8f8 f4d2 f8g8 e1e4 g8f8');
INSERT INTO evaluations VALUES(1818,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(1818,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(1818,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(1818,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(1818,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(1818,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(1818,36,'cp -24','d7c6 b1c3 f8d6 d2d4 c8g4 c1e3 e8c8 e3g5 f6g5 f3g5 g4d1 a1d1 e5d4 g5f7 d4c3 f7d8 g8h6 d8e6 h8e8 e6g5 c3b2 e1e2 d6f4 h2h4 h6f5 e2f3 f4c1 d1c1 b2c1q h1c1 f5h4 f3g4 h4g2 g4g3 h7h6 g5f3 e8e4 g3g2 b7b6 c1d1 c6c5 f3d2 e4g4 g2h2 g4d4');
INSERT INTO evaluations VALUES(1751,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(1751,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1751,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1751,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1751,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(1751,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(1751,36,'cp -2','b1c3 f8d6 d2d4 e5d4 d1d4 f6d4 f3d4 g8e7 c1e3 c8d7 h2h4 h7h5 f2f3 b7b6 a1d1 c6c5 d4e2 e7g6 e1f2 e8c8 g2g3 g6e5 e3f4 f7f5 f4g5 d8e8 h1e1 d7c6 e2f4 c8b7 a2a4 a7a6 f4d5 f5e4 c3e4 c6a4');
INSERT INTO evaluations VALUES(9783,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(9783,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9783,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9783,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9783,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9783,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(9783,36,'cp 9','d2d3 a7a6 c1e3 d7d6 e1g1 e8g8 c3d5 f6d5 c4d5 c6e7 d5b3 c5e3 f2e3 e7g6 d1e1 d8e8 e1g3 c8e6 b3e6 e8e6 f1f2 f7f5 e4f5 f8f5 a1f1 a8f8 g3g4 f5f6 g4e6 f6e6 f3d2 f8f2 f1f2 e6e8 d2e4 e8f8 c2c4 f8f2 g1f2 g8f7 d3d4 e5d4 e3d4 f7e6');
INSERT INTO evaluations VALUES(21161,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(21161,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21161,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21161,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(21161,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(21161,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(21161,36,'cp 37','g1f3 g8f6 a4c4 c8e6 c4a4 e6d7 a4b3 e7e6 a2a3 f8d6 g2g3 e6e5 d4e5 c6e5 f1g2 e5f3 g2f3 e8g8 e1g1 d7c8 b1c3 h7h6 f1d1 d8e7 c1e3 c7c6 b3c2 c8e6 c3a4 e6d5 a1c1 f8e8 f3d5 f6d5 e3c5 d6c5');
INSERT INTO evaluations VALUES(21160,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(21160,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21160,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(21160,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(21160,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(21160,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(21160,36,'cp 16','g8f6 a4c4 c8e6 c4a4 e6d7 a4b3 e7e6 a2a3 h7h6 b1c3 f8d6 g2g3 e6e5 d4e5 c6e5 f1g2 e5f3 g2f3 e8g8 e1g1 d7c8 f1d1 d8e7 c1e3 c7c6 e3d4 c8e6 b3c2 f6d7 c3a4 d6e5 d4e3 f8d8 a4c5 d7c5 e3c5 d8d1 a1d1 e7f6 b2b4 e5b2 h2h4');
INSERT INTO evaluations VALUES(14403,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(14403,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(14403,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(14403,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(14403,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(14403,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(14403,36,'cp 16','a4c4 c8e6 c4a4 e6d7 a4b3 e7e6 a2a3 h7h6 b1c3 f8d6 g2g3 e6e5 d4e5 c6e5 f1g2 e5f3 g2f3 e8g8 e1g1 d7c8 f1d1 d8e7 c1e3 c7c6 e3d4 c8e6 d4f6 g7f6 b3c2 f8d8 f3e4 d6e5 e4f5 e6f5 c2f5 e5c3 b2c3 d8d1 a1d1 a8d8 f5g4 g8f8 d1d8 e7d8 g4a4 a7a6 a4b3 d8d2 b3b7 d2e2');
INSERT INTO evaluations VALUES(14371,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14371,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(14371,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(14371,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(14371,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(14371,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(14371,36,'cp 12','c8e6 c4a4 e6d7 a4b3 e7e6 a2a3 h7h6 b1c3 f8d6 g2g3 e8g8 f1g2 b7b5 e1g1 a8b8 b3c2 a7a5 f1d1 d8e7 c3e4 f6e4 c2e4 a5a4 f3e5 c6a5 e5d7 e7d7 a1b1 a5c4 e4c2 f8d8 c1d2 c7c5 d4c5 d6c5 d2c3 d7e7 b1c1 c5b6');
INSERT INTO evaluations VALUES(27190,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(27190,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(27190,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(27190,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(27190,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(27190,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(27190,36,'cp 112','b5c6 d7c6 b2e5 h7h5 g1f3 g8h6 e5b2 h6f5 f3e5 d8f6 d2d4 e8c8 d1e2 f6e6 c2c4 c6e8 b1c3 d5c4 e2c4 e6c4 e5c4 b7b5 c4d2 b5b4 c3e4 e8b5 a1c1 f7f6 e4c5 d8e8 d4d5 f8c5 c1c5 b5d3');
INSERT INTO evaluations VALUES(27158,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(27158,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(27158,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(27158,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(27158,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(27158,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(27158,36,'cp 121','d7c6 b2e5 h7h5 g1f3 g8e7 e5b2 e7g6 f3d4 c6d7 d1e2 c7c6 b1c3 d8c7 g2g3 e8c8 e1c1 c8b8 c1b1 d8e8 e2f1 f7f6 h2h3 f8c5 d4e2 a7a6 f1g2 c5a7 d2d4 d7c8 b2a3 g6e7 g2f3 e7f5 h3h4 f5d6 f3f4 d6e4 c3e4 e8e4 f4c7 b8c7');
INSERT INTO evaluations VALUES(26763,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(26763,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(26763,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(26763,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(26763,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(26763,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(26763,36,'cp 116','b2e5 d8d7 g1f3 g8e7 e5b2 e8c8 e1g1 h7h5 b1c3 d7e8 f3e5 f7f6 e5c6 e7c6 c3b5 e8d7 d1e2 a7a6 b5d4 c6d4 b2d4 f8d6 a2a4 c8b8 a1e1 d8g8 f2f4 g8e8 d4c3 e8e4 a4a5 c7c5 e2f3 d6c7 h2h3 h5h4 f3g4');
INSERT INTO evaluations VALUES(26762,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(26762,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(26762,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(26762,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(26762,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(26762,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(26762,36,'cp 117','d8d7 e5b2 e8c8 g1f3 g8e7 e1g1 h7h5 b1c3 d7e8 f3e5 f7f6 e5c6 e7c6 c3b5 e8d7 d1e2 c8b8 a1c1 a7a6 b5d4 c6d4 b2d4 f8d6 c2c4 c7c5 d4c3 d7c7 c4d5 d6h2 g1h1 h2d6 d2d4 c5d4 c3d4 c7d7 f1d1 d7f5 c1c4 d8e8 h1g1 g7g5 e2d3');
INSERT INTO evaluations VALUES(36519,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(36519,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(36519,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(36519,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(36519,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(36519,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36519,36,'cp 51','d2d4 c8f5 f3h4 f5c8 e1g1 e7e5 h4f3 b8c6 d4e5 d6e5 c1g5 h7h6 d1d8 f8d8 g5f6 g7f6 c3d5 g8g7 d5c7 a8b8 e2e4 f6e7 f1d1 c8g4 c7d5 e7c5 h2h3 g4e6 f3e1 a7a5 e1d3 c5d4 a1c1 h6h5 b2b3 h5h4 g3g4 d8h8 g2f3 b8c8 f3e2 f7f5 g4f5 g6f5 e4f5');
INSERT INTO evaluations VALUES(28756,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(28756,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(28756,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(28756,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(28756,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(28756,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(28756,36,'cp 77','e4e5 f6e4 c3e4 b7e4 f1c4 d8c8 c2c3 c8b7 h2h4 d7d6 d1e2 h7h6 h4h5 d6d5 h1h4 e4h7 c4b3 a7a5 b3a4 c7c6 c1e3 b7a6 e2a6 a8a6 e1d2 a6a7 h4g4 h7f5 g4g3 a7c7 e3f4 f5e4 a1c1 b8d7 a4c2 c6c5 a2a4 c5d4 f3d4 c7c8 d4b5 e4c2');
INSERT INTO evaluations VALUES(28687,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(28687,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(28687,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(28687,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(28687,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(28687,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(28687,36,'cp 68','f6e4 c3e4 b7e4 c2c3 d7d6 c1f4 b8d7 f1c4 f8e7 h2h4 h7h6 d1e2 d6d5 c4d3 e4d3 e2d3 d7f8 f4e3 c7c5 e1f1 h6h5 g2g3 d8d7 f1g2 f8g6 d3e2 a7a5 h1d1 d7c6 a1c1 a5a4 d1e1 a8c8 g2g1 c8a8 e1d1 a8c8 b2b4 a4b3 a2b3 c5c4 b3b4 c8c7');
INSERT INTO evaluations VALUES(32108,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(32108,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(32108,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(32108,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(32108,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(32108,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(32108,36,'cp 70','e5d4 c2c3 g8f6 c3d4 f8e7 b1c3 e8g8 h2h3 g4h5 c4b3 c7c5 e1g1 c5d4 d1d4 b8c6 d4e3 f6d7 c3d5 e7f6 e3f4 d7c5 d5f6 d8f6 f4f6 g7f6 b3d5 f8e8 c1e3 h5g6 f1d1 c6b4 a1c1 b4d5 e4d5 c5d3 c1c7 d3b2 d1d2 b2d3 c7b7 g6e4 b7d7 e4f3 d2d3 f3e4');
INSERT INTO evaluations VALUES(32177,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(32177,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(32177,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(32177,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(32177,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(32177,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(32177,36,'cp 65','c2c3 g8f6 c3d4 f8e7 h2h3 g4h5 b1c3 e8g8 c4b3 c7c5 e1g1 c5d4 d1d4 b8c6 d4d1 a7a6 g2g4 h5g6 d1e2 f6d7 f1d1 d7c5 b3c2 e7f6 c1e3 d8e8 d1d6 c5e4 c3e4 g6e4 a1d1 e4c2 e2c2 a8c8 c2f5 f6b2 d6d7');
INSERT INTO evaluations VALUES(32176,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(32176,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(32176,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(32176,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(32176,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(32176,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(32176,36,'cp 62','g8f6 c3d4 f8e7 b1c3 e8g8 c4b3 c7c5 d4c5 b8c6 c5d6 e7d6 e1g1 h7h6 h2h3 g4h5 g2g4 h5g6 f1e1 d6b4 e4e5 f6d7 d1e2 b4c3 b2c3 d7c5 c1a3 c5b3 a2b3 f8e8 a3d6 d8b6 f3h4 g6h7 h4f5 b6b3 e2f3 a8d8');
INSERT INTO evaluations VALUES(55156,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(55156,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(55156,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(55156,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55156,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(55156,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(55156,36,'cp -5','d2d4 g8f6 g1f3 e7e6 c1f4 c7c5 e2e3 b8c6 h2h3 a7a6 d4c5 f8c5 c2c4 e8g8 b1d2 d5c4 f1c4 b7b5 c4a2 c8b7 e1g1 c6e7 d2b3 c5b6 a1c1 a8c8 c1c8 d8c8 d1e2 f8d8 f1c1 c8a8 b3c5 b6c5 c1c5');
INSERT INTO evaluations VALUES(2090,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(2090,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(2090,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(2090,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(2090,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(2090,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(2090,36,'cp 200','b1c3 c7c6 c1f4 e7b4 d4e5 b4e7 e5c7 d7d6 c7e7 f8e7 e1c1 c8e6 f4d6 g8f6 f2f3 e7d6 d1d6 e8e7 d6d4 a7a5 b2b3 g7g6 f1e2 h8d8 d4d8 a8d8 h1d1 f6d7 c1d2 b7b5 d2e3 b5b4 c3a4 f7f5 e2d3 h7h5 h2h4 d7f6 e4f5 g6f5 e3f2');
INSERT INTO evaluations VALUES(2088,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(2088,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(2088,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(2088,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(2088,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(2088,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(2088,36,'cp 193','c7c6 f1e2 e7c5 d4d3 g8f6 c1e3 c5a5 e1c1 d7d6 f2f4 f8e7 c1b1 e8g8 e3d4 b7b5 e4e5 d6e5 f4e5 f6d5 c3d5 c6d5 d3b5 a5d8 h1f1 c8e6 b5b3 d8c7 b3c3 c7b7 c3d2 h7h6 b2b3 a7a5 a2a4 e7g5 d2e1 f8c8 e2d3 b7c7 d4b2 a8b8 h2h4 g5e7');
INSERT INTO evaluations VALUES(19591,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(19591,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(19591,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(19591,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(19591,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(19591,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(19591,36,'cp -31','f3e2 e5c4 e2c4 e7e6 g1f3 g8f6 c4e2 f8e7 c2c4 e8g8 e1g1 d7d5 e4e5 f6d7 c4d5 e6d5 d2d4 c5d4 f3d4 e7c5 d4f3 f7f6 e5f6 d7f6 c1e3 f8e8 h2h3 c5e3 f2e3 d8b6 f1e1 f6e4 b1c3 e4c3 b2c3 c8e6 f3g5 h7h6 g5e6 e8e6 e2d2 b6c6 a2a4 a8e8 d2d4');
INSERT INTO evaluations VALUES(19592,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(19592,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(19592,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(19592,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(19592,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(19592,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(19592,36,'cp -33','e5c4 e2c4 e7e6 g1f3 g8f6 b1c3 d7d5 c4b5 c8d7 b5b7 a8b8 b7a7 d8c8 a2a4 b8a8 a7b6 a8a6 b6b3 c5c4 b3a2 f6e4 e1g1 e4c3 b2c3 a6a4 a2b2 f7f6 a1a4 d7a4 d2d3 e8f7 c1e3 f8d6 f1a1 c8d7 d3c4 d5c4 f3d2 h8d8 b2a2 a4b5 d2c4 d6h2 g1h2 d7c7 h2h1 c7c4 a2c4 b5c4 a1a7 f7g6 a7c7 c4d5 c7e7 d8b8 f2f3 b8b1 h1h2');
INSERT INTO evaluations VALUES(19599,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(19599,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(19599,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(19599,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(19599,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(19599,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(19599,36,'cp -31','e2c4 e7e6 g1f3 g8f6 b1c3 d7d5 c4b5 c8d7 b5b7 a8b8 b7a7 d8c8 a2a4 b8a8 a7b6 a8a6 b6b3 c5c4 b3a2 f6e4 e1g1 e4c3 b2c3 a6a4 a2b2 f7f6 a1a4 d7a4 b2a2 a4d7 c1a3 c8a6 f1a1 f8a3 a2a3 a6a3 a1a3 e8e7 g1f1 h8b8 f1e2 e6e5 a3a5 e7d6 a5a6 d6c7 a6a7 b8b7 a7b7 c7b7 f3e1 b7c7 d2d4 g7g5 e2e3 c7d6 e1f3 e5e4 f3d2');
INSERT INTO evaluations VALUES(19596,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(19596,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(19596,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(19596,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(19596,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(19596,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(19596,36,'cp -35','e7e6 g1f3 g8f6 b1c3 d7d5 c4b5 c8d7 b5b7 a8b8 b7a7 d8c8 a2a4 b8a8 a7b6 a8a6 b6b3 c5c4 b3a2 f6e4 e1g1 e4c3 b2c3 a6a4 a2b1 f7f6 a1a4 d7a4 b1a2 a4d7 c1a3 c8a6 f1a1 f8a3 a2a3 a6a3 a1a3 e8e7 g1f1 h8b8 f1e2 b8b2 f3e1 e6e5 h2h4 d5d4 a3a7 e7d6 c3d4 e5d4 a7a6 d7c6 e1f3 d6d5 a6a5 d5e4');
INSERT INTO evaluations VALUES(19135,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(19135,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(19135,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(19135,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(19135,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(19135,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(19135,36,'cp -39','g1f3 g8f6 c4e2 f8e7 e1g1 e8g8 d2d4 b7b6 e4e5 f6e8 f1d1 c8b7 b1c3 f7f6 c1f4 f6e5 f4e5 b7f3 g2f3 f8f5 c3e4 d8c8 c2c4 c5d4 e4g3 f5f7 b2b3 d7d6 e5d4 e6e5 d4e3 c8h3 d1d3 e8c7 a1c1 c7e6 b3b4');
INSERT INTO evaluations VALUES(19134,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(19134,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(19134,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(19134,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(19134,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(19134,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(19134,36,'cp -37','a7a6 e1g1 b7b5 c4e2 c8b7 a2a4 g8f6 d2d3 f8e7 c1g5 e8g8 b1d2 h7h6 g5h4 d7d6 f1e1 d8c7 h2h3 f8c8 c2c3 e7d8 h4g3 c7c6 g3h4 a8b8 a1a2 c6d7 a4b5 a6b5 e1a1 b8a8 a2a8 c8a8 a1a8 b7a8 c3c4 d7c6 c4b5 c6b5 d2c4 b5a6 h4g3 d8e7 e4e5 a6a1 f3e1');
INSERT INTO evaluations VALUES(31533,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31533,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31533,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31533,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31533,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31533,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31533,36,'cp 0','g8e7 c2c3 b8d7 f3h4 f5e4 b1d2 c6c5 d2e4 d5e4 g2g3 e7c6 d4d5 c6e5 d5e6 f7e6 c1f4 d8f6 d1a4 e8c8 e1c1 g7g5 f4e5 d7e5 d1d8 c8d8 h4g2 f6f3 a4a7 d8c7 a7a5 c7b8 a5d8 b8a7 d8a5');
INSERT INTO evaluations VALUES(49473,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(49473,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(49473,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(49473,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(49473,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(49473,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(49473,36,'cp -98','f6e4 d1e2 a7a6 b5a4 e4c5 e2e5 d8e7 e5e7 f8e7 a4b3 b8c6 d2d3 c5b3 a2b3 b7b6 e1g1 c8b7 d3d4 c6b4 f3e1 b7e4 c2c3 b4c2 e1c2 e4c2 b3b4 d7d6 c1e3 e8d7 f1c1 c2d3 c1d1 d3e2 d1e1 e2c4 b1d2 c4d3 d4d5 e7f6 e3d4 f6d4 c3d4 h8e8');
INSERT INTO evaluations VALUES(49799,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(49799,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(49799,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(49799,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(49799,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(49799,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(49799,36,'cp -87','d2d4 e5d4 d1d4 e4f6 b1c3 f8e7 e1g1 e8g8 f1e1 d7d5 c1g5 c8e6 d4h4 b8d7 b5d3 h7h6 e1e6 f7e6 g5h6 d8e8 h6g5 e7d8 h4h3 c7c6 a1e1 d7c5 g5f6 f8f6 d3h7 g8f8 f3g5 f6h6 h3f3 d8f6 c3a4 c5e4 h7e4 d5e4 g5e4 e8h5 f3g3 h5a5 a4c3 f6e5 f2f4');
INSERT INTO evaluations VALUES(51945,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(51945,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(51945,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51945,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(51945,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(51945,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(51945,36,'cp 24','g2g3 g8f6 f1g2 b8c6 d2d4 d7d5 c4d5 e6d5 b1c3 c5d4 f3d4 f8c5 d4b3 c5b6 c3a4 e8g8 a4b6 a7b6 c1e3 h7h5 e1g1 f8e8 e3g5 d8d6 g5f4 d6e7 f1e1 c8g4 f2f3 g4c8 h2h4 c8d7 d1d2 c6e5 f4g5 e5c4 d2f4 b6b5 b3d4 a8c8');
INSERT INTO evaluations VALUES(51946,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(51946,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(51946,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(51946,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(51946,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(51946,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(51946,36,'cp 41','g8f6 f1g2 d7d5 c4d5 e6d5 d2d4 b8c6 e1g1 c5c4 b2b3 c4b3 d1b3 f8e7 c1f4 e8g8 b1c3 c6a5 b3b1 a7a6 f1c1 h7h6 e2e4 d5e4 c3e4 f6d5 f4d2 a5c6 e4c5 a8b8 b1d3 d5f6 c5a6 b8a8 a6c5 e7c5');
INSERT INTO evaluations VALUES(44202,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(44202,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(44202,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(44202,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44202,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(44202,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(44202,36,'cp 45','f1g2 d7d5 c4d5 e6d5 d2d4 b8c6 e1g1 c5c4 b2b3 c4b3 d1b3 f8e7 b1c3 e8g8 c1f4 c6a5 b3b1 c8g4 h2h3 g4d7 e2e4 d5e4 f3e5 a8c8 c3e4 f6d5 f4d2 d7f5 f1c1 c8c1 b1c1 f7f6 d2a5 d8a5 e5c4 a5d8 e4d6 f5e6 d6b7 d8c7 b7a5 f8c8 c1b2');
INSERT INTO evaluations VALUES(44203,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(44203,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(44203,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(44203,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(44203,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44203,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(44203,36,'cp 40','b8c6 e1g1 d7d5 c4d5 e6d5 d2d4 c5c4 b1c3 f8e7 b2b3 c4b3 d1b3 e8g8 c1f4 c6a5 b3b1 c8g4 h2h3 g4d7 f4e5 d7c6 e5f6 e7f6 f1c1 a5c4 a2a4 f6e7 e2e3 f8e8 f3e1 a8c8 b1b3 d8a5 e1d3 c4d2 b3a2 d2e4');
INSERT INTO evaluations VALUES(12405,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(12405,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(12405,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(12405,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(12405,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(12405,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(12405,36,'cp 52','e1g1 d7d5 c4d5 e6d5 d2d4 c5c4 b2b3 c4b3 d1b3 f8e7 b1c3 e8g8 c1f4 c6a5 b3b1 c8g4 h2h3 g4d7 e2e4 d5e4 f3e5 d7f5 c3e4 f6e4 g2e4 f5e4 b1e4 f8e8 f1d1 e7f6 a1b1 h7h6 e4f5 d8d5 f4h6 a8d8 f5g4 a5c4 h6g5 f6g5');
INSERT INTO evaluations VALUES(12404,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(12404,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(12404,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(12404,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(12404,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(12404,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(12404,36,'cp 42','d7d5 c4d5 e6d5 d2d4 c5c4 b1c3 f8e7 b2b3 c4b3 d1b3 e8g8 c1f4 c6a5 b3b1 c8g4 h2h3 g4d7 e2e4 d5e4 f3e5 d7f5 c3e4 f6e4 g2e4 f5e4 b1e4 f8e8 f1d1 e7f6 a1b1 b7b6 b1b5 d8c8 g1g2 a5c4 e4f3 f6e5 d4e5 c8e6 b5d5 e6c6 g2g1 a8c8');
INSERT INTO evaluations VALUES(13542,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(13542,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(13542,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(13542,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(13542,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(13542,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(13542,36,'cp 58','c4d5 e6d5 d2d4 f8e7 d4c5 e7c5 a2a3 f6e4 f3d2 e8g8 d2e4 d5e4 d1d8 f8d8 g2e4 c5d4 b1c3 c8g4 c1f4 d4c3 b2c3 g4e2 f1b1 d8e8 e4g2 e2a6 a3a4 a8d8 f4e3 d8d7 h2h4 e8d8 a4a5 h7h6 b1c1 d8e8 g1h2 e8d8 c3c4 c6e5 e3a7 a6c4');
INSERT INTO evaluations VALUES(13522,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(13522,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(13522,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(13522,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(13522,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(13522,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(13522,36,'cp 55','e6d5 d2d4 f8e7 d4c5 e7c5 a2a3 f6e4 f3d2 e8g8 d2e4 d5e4 d1d8 f8d8 g2e4 c5d4 b1c3 c8g4 c1f4 d4c3 b2c3 g4e2 f1b1 e2d3 e4d3 d8d3 b1b7 d3c3 f4e3 h7h6 b7c7 c6d4 c7c3 d4e2 g1g2 e2c3 g2f3 c3d5 e3d4 d5c7 a1c1 c7e6 d4e3 a8b8 a3a4 b8b3 f3e4 b3a3 c1c8 g8h7');
INSERT INTO evaluations VALUES(13773,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(13773,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(13773,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(13773,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(13773,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(13773,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(13773,36,'cp 62','d2d4 c5c4 b2b3 c4b3 d1b3 f8e7 c1g5 e8g8 f1c1 c6a5 b3d1 f6e4 g5e7 d8e7 b1c3 c8e6 f3e5 e4c3 c1c3 f8c8 a1c1 c8c3 c1c3 a8d8 e5d3 a5c4 a2a4 a7a6 d1b3 h7h6 c3c1 c4a5 b3a2 a5c4 d3c5 b7b6 c5e6 e7e6 e2e4 b6b5 a4b5 a6b5 e4d5');
INSERT INTO evaluations VALUES(13757,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(13757,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(13757,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(13757,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(13757,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(13757,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(13757,36,'cp 51','h7h6 b2b3 f8e7 c1b2 c8f5 d4c5 e7c5 b1c3 e8g8 a1c1 f8e8 e2e3 c5b6 c3a4 c6b4 a4b6 a7b6 c1a1 f5e4 d1d2 b4d3 b2d4 d8d6 a1d1 e8c8 d2e2 d3b4 d1d2 c8c2 d2c2 b4c2 a2a4 e4f3 e2c2 f3g2 g1g2 f6e4 f2f3 e4g5 h2h4 g5e6 d4b2');
INSERT INTO evaluations VALUES(4274,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(4274,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(4274,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(4274,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(4274,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(4274,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(4274,36,'cp 117','b5c6 b7c6 f3e5 d7d5 d3d4 c5b6 e4d5 d8d5 c1e3 c6c5 b1c3 d5e6 f1e1 c8b7 d4c5 e6e5 c5b6 c7b6 d1e2 a8d8 f2f3 h7h6 a1d1 f8e8 d1d8 e8d8 e3f2 e5f5 b2b3 g8h7 a2a4 d8e8 e2d2 f5g6 e1e8 f6e8 f2g3 e8f6 d2d3 g6d3');
INSERT INTO evaluations VALUES(4250,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(4250,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(4250,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(4250,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(4250,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(4250,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(4250,36,'cp 90','b7c6 f3e5 d7d5 d3d4 c5b6 e4d5 d8d5 c1e3 c6c5 b1c3 d5e6 f1e1 c8b7 d4c5 e6e5 c5b6 a7b6 d1d4 e5d4 e3d4 f6d5 c3d5 b7d5 a2a4 c7c5 d4e3 f7f6 b2b3 d5f7 e1e2 f8d8 g1f1 f7g6 f1e1 g8f7 e2d2 g6f5 a1d1 f7e8 d2d8 a8d8 d1d8 e8d8 e1d1 f5d7 f2f3 d8c8 a4a5 b6a5 e3c5 a5a4');
INSERT INTO evaluations VALUES(2867,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(2867,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(2867,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(2867,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(2867,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(2867,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(2867,36,'cp 102','f3e5 d7d5 d3d4 c5b6 e4d5 d8d5 c1e3 c6c5 b1c3 d5e6 f1e1 c8b7 d4c5 e6e5 c5b6 a7b6 a2a4 f8d8 d1e2 f6d5 c3d5 b7d5 b2b3 h7h6 f2f3 e5c3 a1c1 f7f6 e3f4 d5f7 e2e3 c3c6 h2h4 d8e8 e3f2 a8d8 e1e8 f7e8 f2g3 g8h7 f4c7 d8d2 c2c4 d2d3 c1b1 e8g6 g3f4 c6d7');
INSERT INTO evaluations VALUES(2864,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(2864,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(2864,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(2864,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(2864,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(2864,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(2864,36,'cp 94','d7d5 d3d4 c5b6 e4d5 d8d5 c1e3 c6c5 b1c3 d5e6 f1e1 c8b7 d4c5 e6e5 c5b6 a7b6 a2a4 f8d8 d1e2 h7h6 f2f3 f6d5 c3d5 b7d5 b2b3 e5c3 a1c1 f7f6 e3f2 d5f7 e2e3 c3c6 c2c4 d8d7 e3e2 a8e8 f2e3 d7e7 e2d3 e7d7 d3c2 e8d8 c1d1 c6e6');
INSERT INTO evaluations VALUES(50268,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(50268,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50268,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50268,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(50268,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(50268,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(50268,36,'cp 33','e2e4 d7d6 d2d4 f8g7 b1c3 b8c6 h2h3 e8g8 g1e2 e7e5 c1e3 e5d4 e2d4 f8e8 e1g1 a8b8 d4c6 b7c6 b2b3 c8b7 f1e1 c6c5 d1d3 f6d7 a1d1 d7e5 d3e2 e5c6 c3d5 c6d4 e2d3 b7d5 e4d5 d4b5 e3d2 e8e1 d1e1 d8f6 g3g4 b5c3 d3g3 f6d4 g3f4 d4f4 d2f4');
INSERT INTO evaluations VALUES(51524,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51524,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(51524,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(51524,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(51524,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(51524,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(51524,36,'cp 57','f1g2 d7d5 c2c4 e7e6 c4d5 e6d5 b1c3 g8f6 g1f3 f8e7 d2d4 e8g8 e1g1 c8b7 c1f4 b8d7 f3e5 c7c6 e2e4 d7e5 d4e5 f6e4 c3e4 d5e4 g2e4 d8d1 f1d1 a8d8 f4e3 d8d1 a1d1 f8d8 d1c1 f7f6 e5e6 d8d6 h2h4 e7f8 b2b4 d6e6 e4f5');
INSERT INTO evaluations VALUES(51525,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(51525,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51525,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(51525,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(51525,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51525,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(51525,36,'cp 57','d7d5 c2c4 e7e6 g1f3 g8f6 c4d5 e6d5 b1c3 f8e7 e1g1 e8g8 d2d4 c8b7 c1f4 b8d7 f3e5 f8e8 d1a4 d7f8 e5c6 d8d7 c6e7 d7e7 f4g5 e8d8 e2e3 f8e6 g5f6 e7f6 h2h4 c7c6 a1c1 h7h6 a4c2 g7g5 h4g5 h6g5 b2b3 a8c8');
INSERT INTO evaluations VALUES(51304,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(51304,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(51304,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(51304,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(51304,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(51304,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51304,36,'cp 42','c2c4 e7e6 c4d5 e6d5 b1c3 g8f6 g1f3 f8e7 e1g1 e8g8 d2d4 c8b7 c1f4 b8d7 a1c1 c7c6 f1e1 f8e8 d1b3 a7a5 f3e5 d7e5 f4e5 f6d7 e5f4 d7f6 e2e4 f6e4 c3e4 d5e4 g2e4 e7b4 e1e3 a8c8 c1d1 h7h6');
INSERT INTO evaluations VALUES(56808,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(56808,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(56808,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(56808,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(56808,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(56808,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(56808,36,'cp 63','d2d4 f6e5 f3e5 g8f6 b1d2 f8d6 d2f3 c7c5 d4c5 d8a5 c1d2 a5c5 f1d3 b8d7 d2c3 f6e4 c3d4 c5c7 e5d7 c8d7 f3e5 e8g8 e5d7 c7d7 d3e4 d5e4 c2c3 d7c6 e1g1 d6f4 d1e2 b7b6 a1e1 e6e5 d4e3 a8d8 e3f4 e5f4 e2e4 c6e4 e1e4 d8d2 f1b1 d2c2 a2a4');
INSERT INTO evaluations VALUES(13187,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(13187,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(13187,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(13187,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(13187,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(13187,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(13187,36,'cp -58','e7e5 d2d3 e5e4 d3e4 d8d1 e1d1 f6e4 c1e3 c8f5 b1c3 e8c8 d1e2 f8e7 c3d5 e7c5 a1d1 h8e8 e2e1 c6b4 f1d3 b4d3 d1d3 b7b5 e3c5 e4c5 d3e3 c5d3 e1d2 b5c4 e3e8 d8e8 d5e3 d3f2 h1f1 c4c3 b2c3 f2e4 d2c1 f5g6 e3d5 c8d7 f1d1 d7c6 d1d4 c6c5 d5e3');
INSERT INTO evaluations VALUES(13814,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(13814,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(13814,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(13814,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(13814,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(13814,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(13814,36,'cp -42','d2d3 e5e4 d3e4 d8d1 e1d1 f6e4 c1e3 c8f5 b1c3 e8c8 d1e2 f8e7 c3d5 e7c5 a1d1 h8e8 e2e1 c6b4 f1d3 b4d5 c4d5 d8d5 d3e4 f5e4 d1c1 e4f3 g2f3 c8d7 e1e2 e8e5 f3f4 e5f5 e3c5 d5c5 h1d1 d7e6 e2f3 f5h5 f3g3 c5c1 d1c1 h5d5 c1c7 d5d7 f4f5 e6e7 c7d7 e7d7 g3f4 f7f6 f4g4 d7d6');
INSERT INTO evaluations VALUES(13812,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(13812,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(13812,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(13812,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(13812,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(13812,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(13812,36,'cp -37','e5e4 d3e4 d8d1 e1d1 f6e4 c1e3 c8f5 b1c3 e8c8 d1e2 f8e7 c3d5 e7c5 a1d1 h8e8 e2e1 c6b4 f1d3 b4d5 c4d5 d8d5 d3e4 f5e4 d1c1 e4f3 g2f3 c8d7 e1e2 c5d6 h1d1 d5d1 c1d1 e8e5 e2f1 e5a5 e3f4 a5a6 d1d4 g7g6 b2b3 a6c6 f1e2 a7a6 e2e3 d7e6 f4d6');
INSERT INTO evaluations VALUES(13818,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(13818,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(13818,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(13818,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(13818,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(13818,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(13818,36,'cp -44','d3e4 d8d1 e1d1 f6e4 c1e3 c8f5 b1c3 e8c8 d1e2 f8e7 c3d5 e7c5 a1d1 h8e8 e2e1 c6b4 f1d3 b4d5 c4d5 d8d5 d3e4 f5e4 d1c1 e4f3 g2f3 c8d7 e1e2 c5d6 c1d1 d5d1 h1d1 e8e5 e2f1 e5a5 e3f4 a5a6 b2b3 g7g5 f4g5 a6a2 g5f4');
INSERT INTO evaluations VALUES(56002,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(56002,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(56002,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(56002,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(56002,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(56002,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(56002,36,'cp 59','d2d4 a7a6 c1e3 b8c6 g1f3 g8f6 e1g1 b7b5 c4d3 c6b4 d3e2 b4d5 c3d5 f6d5 e3g5 f8e7 g5e7 d5e7 a2a4 b5b4 d1d2 a8b8 f3e5 e8g8 f1d1 c8b7 a4a5 e7f5 e2f1 f7f6 e5c4 b7d5 c4e3');
INSERT INTO evaluations VALUES(56006,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(56006,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(56006,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(56006,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(56006,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(56006,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(56006,36,'cp 44','a7a6 g1f3 b7b5 c4d3 c8b7 d1e2 f8e7 a2a4 b5b4 c3e4 g8f6 e4f6 e7f6 e1g1 e8g8 c1f4 b8d7 f1d1 f8e8 c2c3 b4c3 b2c3 c7c5 f3e5 d7e5 d4e5 d8c7 a4a5 f6e7 d1b1 e8b8 b1b6 c5c4 d3c4 e7c5 b6b1 b8d8 c4d3 c7c6 b1b7 c6b7 d3e4 b7c8 e4a8 c8a8 a1b1');
INSERT INTO evaluations VALUES(35216,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(35216,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(35216,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(35216,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(35216,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(35216,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(35216,36,'cp -83','e1f1 d7d6 b1c3 g8f6 d2d4 f8e7 g1f3 h4h6 f1g1 c8e6 c4d3 b8d7 c3e2 f6h5 h2h3 d6d5 g2g4 g7g5 g4h5 h6h5 d1f1 d5e4 d3e4 e8c8 f3e1 h8g8 e1d3 d7f6 e4f3 h5g6 c1d2 e6d5 f3d5 d8d5 h1h2 e7d6 f1g2 g8e8 e2c3 d5d4 c3b5 d4c4 a1e1 e8e1 d2e1 f6e4 b5d6 e4d6');
INSERT INTO evaluations VALUES(35215,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(35215,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(35215,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(35215,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(35215,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(35215,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(35215,36,'cp -59','d7d6 b1c3 g8f6 d2d4 f8e7 g1f3 h4h6 f1g1 c8e6 c4d3 d6d5 g2g3 f6e4 c3e4 d5e4 d3e4 b8c6 c1f4 g7g5 f4e3 e8c8 c2c3 e6h3 d1d3 c8b8 a1e1 h8e8 e4h7 h6h5 h7e4 g5g4 f3d2 c6e5 d3e2 e5d7 e2f2 e7g5 e3g5 h5g5 f2f4 g5g7 f4f2');
INSERT INTO evaluations VALUES(34795,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(34795,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(34795,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(34795,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(34795,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(34795,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(34795,36,'cp -55','b1c3 c8e6 c3d5 h4d8 d1e2 c7c6 g1f3 b8d7 d2d4 g8f6 c1f4 f6e4 d5c7 d8c7 c4e6 f7e6 e2e4 e6e5 d4e5 e8c8 a1d1 d8e8 e4f5 c7a5 g2g3 a5b5 f1g2 g7g6 f5f7 e8e7 f7b3 b5b3 a2b3 d6e5 f4g5 e7e6 h1f1 f8c5 f3d2 e5e4 d2c4');
INSERT INTO evaluations VALUES(49615,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(49615,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(49615,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(49615,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(49615,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(49615,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(49615,36,'cp -56','b8c6 f1d3 f8c5 g1e2 e8g8 e1g1 c6b4 a2a3 b4d3 f3d3 c7c6 d3g3 d7d5 d2d3 c5d6 e4d5 c6d5 c1g5 c8e6 f2f4 f6h5 g3h4 d8b6 g1h1 h5f4 e2f4 e5f4 a1b1 f4f3 f1f3 d6e5 c3e2 d5d4 g5f4 f7f6 f4e5 f6e5 h4g3 e6f7 h2h3 a8e8 f3f2 a7a5 h1g1 e5e4 d3e4 e8e4');
INSERT INTO evaluations VALUES(16514,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(16514,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(16514,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(16514,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(16514,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(16514,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(16514,36,'cp -55','f1d3 c6d4 f3g3 d7d6 g1e2 c7c5 e2d4 c5d4 c3d5 f6d5 e4d5 g7g6 c2c3 f8g7 c3d4 c8d7 e1g1 e8g8 d4e5 g7e5 f2f4 e5d4 g1h1 d8a5 g3f3 f7f5 a2a3 a5b6 d3c4 a8c8 d2d3 f8e8 a1b1 d7b5 c1d2 b5c4 d3c4 c8c4 f3d3 c4c8 b2b3 d4f6 h2h3 b6d4 d3d4 f6d4 d2b4 c8c2 b4d6 g8f7');
INSERT INTO evaluations VALUES(57228,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(57228,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(57228,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(57228,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(57228,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(57228,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(57228,36,'cp -18','e5e4 f3h4 d7d5 d2d3 f8b4 c2c3 b4e7 d3e4 d5e4 d1d8 e7d8 h4g2 f7f5 c1f4 g8f6 h2h4 b8d7 b1a3 f6d5 f4g5 c7c6 g2f4 d7f6 e2e3 d5f4 g5f4 f6d5 a3c2 d5f4 e3f4 e8e7 a2a4 a7a5 c2e3 d8b6 f1c4 b6e3 f2e3 h8d8 a1d1');
INSERT INTO evaluations VALUES(57329,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(57329,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(57329,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(57329,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(57329,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(57329,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(57329,36,'cp -1','f3h4 d7d5 d2d3 f8b4 c2c3 b4e7 d3e4 d5e4 d1d8 e7d8 h4g2 f7f5 h2h4 g8f6 c1f4 c8e6 g2e3 b8d7 h4h5 f6d5 e3d5 e6d5 b1a3 c7c6 a3c2 d8f6 c2e3 d5e6 f2f3 d7c5 e1c1 a8d8 d1d8 f6d8 c1c2 d8f6 f3e4 c5e4 f1g2 e8f7 g2e4 f5e4 h1d1');
INSERT INTO evaluations VALUES(6474,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(6474,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(6474,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(6474,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(6474,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(6474,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(6474,36,'cp -14','f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 c2c3 c7c6 f1d3 d7d5 e4d5 d8d5 d1e2 g8e7 e1g1 c8g4 e2e3 e8c8 d3c4 d5d6 h2h3 g4f5 c1a3 d6c7 e3a7 e7d5 b1d2 c7b6 a7b6 d5b6 a3c5 c8c7 a1e1 e5c3 e1e7 d8d7 e7d7 b6d7 c5e3 d7e5 c4e2 h8d8 d2c4 e5c4 e2c4 f7f6 a2a4 c3e5');
INSERT INTO evaluations VALUES(50529,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(50529,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(50529,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(50529,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(50529,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(50529,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50529,36,'cp 41','d2d4 d7d5 e4d5 e6d5 f1d3 c5c4 d3c2 f8d6 e1g1 g8e7 c1g5 e8g8 f1e1 h7h6 g5h4 d8c7 b1d2 b8c6 h4g3 c8g4 d2f1 a8e8 h2h3 g4h5 g3d6 c7d6 a2a4 d6c7 f1g3 h5g6 f3h4 g6c2 d1c2 c7f4 h4f3 f4f6 e1e2 e7g6 g3h5 f6d6 a1e1 e8e2');
INSERT INTO evaluations VALUES(50870,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(50870,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(50870,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(50870,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(50870,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(50870,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(50870,36,'cp 88','d2d4 e5d4 f3d4 g8f6 c1f4 f8e7 d1f3 e8g8 e1c1 c8g4 f3e3 f6h5 g2g3 b8c6 f2f3 g4h3 e4e5 c6d4 e3d4 e7g5 f4g5 d8g5 f3f4 d6e5 d4c5 g5g6 c4d3 g6b6 c5b6 c7b6 f4e5 h3g4 d1e1 b6b5 c3e4 a8d8 h1f1 g4h3 f1f2 f7f6 e5f6 h5f6 e4f6 g7f6 d3e2 d8e8 e1d1');
INSERT INTO evaluations VALUES(50878,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(50878,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(50878,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(50878,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(50878,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(50878,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(50878,36,'cp 92','e5d4 f3d4 g8f6 c1f4 f8e7 d1f3 e8g8 e1c1 c8g4 f3e3 f6h5 g2g3 b8c6 f2f3 g4h3 e4e5 c6d4 e3d4 e7g5 f4g5 d8g5 f3f4 d6e5 d4c5 g5g6 c4d3 g6b6 c5b6 c7b6 f4e5 h3g4 d1e1 f8e8 d3e4 a8b8 e4g2 b6b5 h1f1 g7g6 f1f2 g4e6 g2f3 h5g7 c3e4 e8e7 e4f6 g8f8 e1d1 g7e8 f2d2');
INSERT INTO evaluations VALUES(50884,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(50884,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(50884,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(50884,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(50884,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(50884,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(50884,36,'cp 113','f3d4 g8f6 c1f4 f8e7 d1f3 e8g8 e1c1 c8g4 f3e3 f6h5 g2g3 b8c6 f2f3 g4h3 e4e5 c6d4 e3d4 e7g5 f4g5 d8g5 f3f4 d6e5 d4c5 g5g6 c4d3 g6b6 c5b6 c7b6 f4e5 h3g4 d1e1 b6b5 d3e4 a8b8 h1f1 f8e8 e4g2 g7g6 c3d5 g8f8 h2h3 g4e6 g3g4 h5g3 f1g1 e8c8 d5f6 b5b4 g2d5 e6d5');
INSERT INTO evaluations VALUES(50881,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(50881,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(50881,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(50881,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(50881,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(50881,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(50881,36,'cp 106','g8f6 c1f4 f8e7 d1f3 e8g8 e1c1 c8g4 f3e3 f6h5 g2g3 b8c6 f2f3 g4h3 e4e5 c6d4 e3d4 e7g5 f4g5 d8g5 f3f4 d6e5 d4c5 g5g6 c4d3 g6b6 c5b6 c7b6 f4e5 h3g4 d1e1 b6b5 d3e4 a8b8 h1f1 g4h3 f1f2 f8e8 c3d5 h3g4 a2a3 b8d8 f2d2 d8d7 e4g2 g4e6 d5e3 g7g6 g3g4 h5f4 d2d7 e6d7 g2b7');
INSERT INTO evaluations VALUES(43619,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(43619,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(43619,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(43619,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(43619,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(43619,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(43619,36,'cp 97','c1f4 f8e7 d1d2 c7c5 d4f5 c8f5 e4f5 b8d7 c4d5 e8g8 e1c1 d7b6 d5f3 d6d5 f3d5 b6d5 c3d5 f6d5 d2d5 d8b6 d5b3 b6c6 b3f3 c6f3 g2f3 f8e8 f4e3 c5c4 h1g1 g8f8 d1d7 b7b5 g1d1 e7f6 f3f4 c4c3 b2c3 f6c3 d7c7 a8c8 c7c8 e8c8 d1d6 a6a5 d6d5 b5b4');
INSERT INTO evaluations VALUES(45121,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(45121,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(45121,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(45121,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(45121,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(45121,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(45121,36,'cp -29','d6e5 f3e5 b8d7 d2d4 c5d4 d1d4 e7e6 f1b5 a7a6 e5d7 c8d7 b5d7 d8d7 d4d7 e8d7 b1d2 f6d5 d2c4 f8c5 e1e2 a8c8 h1d1 d7e7 c4e3 d5f4 e2f3 f4g6 f3e2 c5a7 a2a4 c8c7 e3f1 h7h5 g2g3 f7f6 c1e3 a7e3 f1e3 g6e5 a4a5');
INSERT INTO evaluations VALUES(45311,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(45311,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(45311,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(45311,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(45311,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(45311,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(45311,36,'cp -33','f3e5 b8d7 d2d4 c5d4 d1d4 e7e6 f1b5 f8e7 e5d7 c8d7 b5d7 d8d7 d4d7 e8d7 e1e2 e6e5 c1g5 d7e6 a2a4 a8d8 g5f6 g7f6 h1d1 d8d1 e2d1 h7h5 b1d2 f6f5 a4a5 h8d8 d1e1 h5h4 e1e2 f5f4 b2b4 f7f5 d2f3');
INSERT INTO evaluations VALUES(45306,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(45306,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(45306,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(45306,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(45306,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(45306,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(45306,36,'cp -28','b8d7 d2d4 c5d4 d1d4 e7e6 f1b5 f8e7 e5d7 c8d7 b5d7 d8d7 d4d7 e8d7 e1e2 e6e5 h1d1 d7e6 c1g5 h8d8 d1d8 a8d8 b1d2 h7h5 g2g3 f6d7 g5e7 e6e7 d2b3 d7b6 b3c5 b6c4 b2b3 c4d6 a1d1 g7g5 c3c4 f7f6 c5a4 d8c8 d1d5 b7b6 a4c3');
INSERT INTO evaluations VALUES(12266,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(12266,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(12266,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(12266,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(12266,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(12266,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(12266,36,'cp -17','d2d4 c5d4 d1d4 e7e6 e5d7 c8d7 b1d2 f8e7 d2c4 e8g8 c4e5 d7e8 d4d8 a8d8 f1e2 e8a4 c1e3 a7a6 e3b6 d8c8 e2f3 e7c5 e5c4 f6d7 b6c5 c8c5 c4d6 c5c7 b2b3 a4c6 f3c6 c7c6 d6b7 c6c3 e1g1 d7f6 f1c1 f8c8 c1c3 c8c3 b7d6');
INSERT INTO evaluations VALUES(16135,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(16135,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(16135,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16135,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(16135,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16135,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(16135,36,'cp 7','f8c5 e1g1 d7d6 c2c3 e8g8 b1d2 a7a6 f1e1 c8e6 b2b4 c5a7 c4b3 f8e8 a2a4 e6b3 d2b3 c6e7 a4a5 e7g6 d1c2 d8d7 c3c4 a8d8 g1h2 c7c6 g2g3 d6d5 c1e3 a7e3 e1e3 d7d6 b3c5 d6c7 h2g2 e8f8 a1e1 f8e8');
INSERT INTO evaluations VALUES(42984,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(42984,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(42984,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(42984,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(42984,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(42984,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(42984,36,'cp 112','e2e4 d7d6 h2h3 e6e5 d4d5 b8d7 c1e3 d7f8 g1f3 f8g6 f1e2 c7c6 f3d2 c8d7 a2a4 d8c8 e1g1 e8g8 a4a5 c6c5 f1e1 h7h6 d2f1 g8h8 e2d3 c8d8 f1g3 f6e8 a5a6 b7b6 g3f5 e7g5 e3g5 d8g5');
INSERT INTO evaluations VALUES(42994,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(42994,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(42994,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(42994,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(42994,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(42994,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(42994,36,'cp 111','d7d6 h2h3 e8g8 f1d3 e6e5 d4d5 c7c6 g1f3 b8d7 c1e3 f6h5 e1g1 h5f4 d3c2 g8h8 g1h2 c6d5 c4d5 d7b6 e3f4 e5f4 c3e2 e7g5 f3g5 d8g5 d1d2 f7f5 d2f4 g5f4 e2f4 f5e4 f4e6 c8e6 d5e6 a8e8 c2e4 d6d5 e4f3 e8e6 b2b3 e6d6 a1c1 g7g6 c1c7');
INSERT INTO evaluations VALUES(40885,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(40885,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(40885,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(40885,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(40885,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(40885,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(40885,36,'cp 101','g1f3 e8g8 f1d3 e6e5 d4d5 b8d7 h2h3 c7c6 c1e3 f6h5 d1d2 c6d5 c3d5 h5f6 d5c3 b7b6 e1g1 d7c5 f1d1 h7h6 d2e2 c8b7 f3d2 a8c8 a1c1 c5d3 e2d3 d8d7 b2b4 f6h7 c3d5 e7g5 c1c3 g5e3 d5e3 h7f6 a2a4');
INSERT INTO evaluations VALUES(40884,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(40884,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(40884,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(40884,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(40884,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(40884,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(40884,36,'cp 103','e8g8 f1d3 e6e5 d4d5 b8d7 h2h3 c7c6 c1e3 c6d5 c4d5 f6h5 d1d2 h5f4 e3f4 e5f4 d2f4 d8b6 a1b1 e7f6 e1g1 d7e5 f3e5 f6e5 f4d2 c8d7 g1h1 a8e8 c3e2 g7g6 e2g3 e5g7 b2b3 b6d4 g3e2 d4b6 a2a4 a7a6 f2f4 f7f5 e2g3 b6d4 e4f5');
INSERT INTO evaluations VALUES(36284,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(36284,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(36284,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(36284,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(36284,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(36284,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(36284,36,'cp 108','f1d3 e6e5 d4d5 b8d7 h2h3 f6h5 d3c2 e7f6 g2g3 g7g6 c1e3 f6g7 a2a3 a7a5 d1e2 b7b6 a1b1 d8e7 e1g1 g8h8 g1g2 d7f6 f1e1 h7h6 c3b5 f6g8 e3d2 h5f6 e1h1 f6h7 b1e1 h7f6 a3a4 f6h7 b5c3 g8f6');
INSERT INTO evaluations VALUES(36285,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(36285,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(36285,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(36285,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(36285,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(36285,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(36285,36,'cp 101','e6e5 d4d5 b8d7 h2h3 f6h5 d3c2 e7f6 a2a3 g7g6 e1g1 a7a5 f1e1 g8h8 a1b1 f8g8 g2g3 f6e7 b2b4 g8f8 c1h6 a5b4 a3b4 h5g7 d1d2 d7f6 g1g2 f6g8 h6e3 d8d7 f3g1 f7f5 c4c5 e7f6 b1a1 a8a1 e1a1');
INSERT INTO evaluations VALUES(51774,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(51774,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(51774,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(51774,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(51774,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(51774,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(51774,36,'cp -28','d7d5 c4d3 b8d7 e4d5 e6d5 e1g1 f8e7 f1e1 g8f6 c2c3 e8g8 d3c2 f8e8 d2d4 e7d6 e1e8 d8e8 c1e3 d7f8 h2h3 e8e7 b1d2 f8g6 a2a4 c8d7 a4a5 a8e8 c2g6 h7g6 f3e5 e7d8 d2f3 d8c7 e5d7 f6d7 a5a6 b7b6');
INSERT INTO evaluations VALUES(53556,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(53556,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(53556,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(53556,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(53556,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(53556,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(53556,36,'cp -23','c4d3 b8d7 e4d5 e6d5 e1g1 g8f6 c2c3 f8e7 d3c2 e8g8 d2d4 e7d6 f1e1 f8e8 c1g5 e8e1 d1e1 d8b6 b2b3 b6c7 b1d2 b7b6 c2d3 c8b7 e1f1 h7h6 g5h4 a8e8 a1e1 c6c5 h4g3 f6e4 c3c4 d7f6 c4d5 b7d5 d2e4 f6e4 d4c5 c7c5');
INSERT INTO evaluations VALUES(53622,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(53622,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(53622,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(53622,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(53622,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(53622,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(53622,36,'cp -24','f8d6 e1g1 b8d7 e4d5 c6d5 c2c4 d5c4 d3c4 a7a6 d2d4 b7b5 c4b3 g8f6 d4d5 d7c5 d5e6 c8e6 b3e6 c5e6 f3g5 e6g5 c1g5 e8g8 d1f3 d6e5 b1c3 d8a5 a2a3 a8c8 b2b4 c8c3 b4a5 c3f3 g2f3 e5a1 f1a1 f6d7 g1g2 f7f6 a1d1 d7e5 g5e3');
INSERT INTO evaluations VALUES(55757,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(55757,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(55757,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55757,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(55757,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(55757,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(55757,36,'cp -26','c7c6 b5d3 f8d6 e1g1 b8d7 e4d5 c6d5 c2c4 d5c4 d3c4 a7a6 d2d4 b7b5 c4b3 g8f6 d4d5 d7c5 d5e6 c8e6 b3e6 c5e6 f3g5 e6c5 f1e1 d6e7 d1e2 h7h6 g5f3 c5e6 a2a4 b5b4 c1e3 f6d5 f3d4 d8d7 d4e6 d7e6 g2g3 e8g8 e3d4 e6g6 b1d2 a8d8 e2f3');
INSERT INTO evaluations VALUES(53531,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(53531,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(53531,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(53531,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(53531,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(53531,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(53531,36,'cp -24','b5d3 f8d6 e1g1 b8d7 e4d5 c6d5 c2c4 d5c4 d3c4 a7a6 d2d4 b7b5 c4b3 g8f6 d4d5 d7c5 d5e6 c8e6 b3e6 c5e6 f3g5 e6g5 c1g5 e8g8 d1f3 d6e5 b1c3 d8b6 f1e1 a8e8 a1d1 h7h6 g5h4 e5d4 e1e8 f8e8 h4f6 b6f6 f3f6 d4f6 g1f1 f6e5 d1d7');
INSERT INTO evaluations VALUES(12392,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12392,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(12392,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(12392,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(12392,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12392,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(12392,36,'cp 11','f1e2 d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 e2c4 f8c5 d1c3 e8g8 e1g1 f8e8 c1g5 c5d4 a1e1 c8f5 h2h3 d4c3 e1e8 f6e8 b2c3 e8d6 c4b3 d6e4 b3c2 e4g5 c2f5 g7g6 f5d7 a8d8 f1d1 c6c5 g1f1 g5e6 d1d6');
INSERT INTO evaluations VALUES(38459,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(38459,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38459,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(38459,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(38459,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(38459,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38459,36,'cp 14','e2e3 g8f6 b1d2 e8g8 h2h3 c7c5 c2c3 b7b6 f1e2 c8b7 e1g1 e7e6 a2a4 b8c6 e2b5 d8e7 f3e5 f8c8 f1e1 a7a6 b5c6 b7c6 d1e2 c6e8 e2f3 f6h5 f4h2 h5f6 d4c5 c8c5 e3e4 f6e4 d2e4 d5e4 f3e4 c5d5 e4e2');
INSERT INTO evaluations VALUES(1484,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(1484,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(1484,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(1484,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(1484,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(1484,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(1484,36,'cp -8','b1c3 c8d7 f1b5 c5d4 f3d4 b6c7 f2f4 c6d4 b5d7 c7d7 d1d4 g8h6 d4d3 h6f5 e1c1 d7c6 c1b1 h7h5 c3b5 f8c5 e3c5 c6c5 b5d4 g7g6 d4f5 g6f5 d3e2 e8c8 d1d3 c8b8 h1d1 d8c8 e2d2 c5b5 d3b3 b5a6 b3a3 a6c6 a3d3 h5h4 a2a3 c6b6 d3b3 b6a6 b3b4');
INSERT INTO evaluations VALUES(1468,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(1468,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(1468,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(1468,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(1468,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(1468,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(1468,36,'cp -12','c8d7 f1b5 c5d4 f3d4 b6c7 f2f4 c6d4 b5d7 c7d7 d1d4 g8e7 e1c1 e7f5 d4d3 h7h5 c1b1 d7c6 c3e2 f8c5 e3c5 c6c5 e2d4 f5d4 d3d4 a8c8 d4d2 g7g6 g2g3 c5b6 h2h3 c8c4 h1f1 e8g8 f1f3 f8c8 c2c3 a7a6 f3d3 c8c6 f4f5 e6f5 d3d5');
INSERT INTO evaluations VALUES(10612,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(10612,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(10612,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(10612,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(10612,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(10612,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(10612,36,'cp 90','h7h6 h2h3 g4f6 c1e3 c5b6 e3b6 a7b6 d3d4 e5d4 f3d4 e8g8 f1e1 d7d6 f2f4 d6d5 e4d5 c6a5 c4b3 c7c6 d5c6 b7c6 a2a4 a5b3 c2b3 d8d6 d1d2 f8d8 a1d1 c8d7 d2f2 d8e8 e1e8 a8e8 d4f3 d6b4 f3e5 d7e6 e5c6 b4b3');
INSERT INTO evaluations VALUES(9880,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(9880,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(9880,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(9880,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(9880,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(9880,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(9880,36,'cp 109','h2h3 g4f6 c1e3 c5b6 e3b6 a7b6 d3d4 e5d4 f3d4 d7d6 d4b5 e8g8 f2f4 f8e8 a2a3 c8e6 c4d3 c6b8 d1e2 e6d7 a1e1 b8a6 e2f2 a6c5 e4e5 c5d3 c2d3 d7b5 c3b5 f6d5 e5d6 e8e1 f1e1 d8d7 b5c7 d5c7 d6c7 d7c7 d3d4 a8d8 e1d1 b6b5 d4d5');
INSERT INTO evaluations VALUES(9879,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(9879,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(9879,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(9879,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(9879,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(9879,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(9879,36,'cp 110','g4f6 c1e3 c5b6 e3b6 a7b6 d3d4 e5d4 f3d4 d7d6 d4b5 e8g8 f2f4 c8e6 d1e2 c6a5 c4d3 c7c6 b5d4 b6b5 g1h2 d8c7 b2b4 a5c4 a2a4 c4b2 a4a5 a8e8 e2f3 b2c4 f3g3 f6h5 g3h4 h5f6 a1e1 c4b2 h4g3 f6h5 g3f3 h5f6 c3e2 c6c5 d3b5 c5d4 b5e8 f8e8');
INSERT INTO evaluations VALUES(18434,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(18434,31,'cp 284',NULL);
INSERT INTO evaluations VALUES(18434,32,'cp 291',NULL);
INSERT INTO evaluations VALUES(18434,33,'cp 287',NULL);
INSERT INTO evaluations VALUES(18434,34,'cp 295',NULL);
INSERT INTO evaluations VALUES(18434,35,'cp 243',NULL);
INSERT INTO evaluations VALUES(18434,36,'cp 281','d4d5 c6a5 b1c3 g7g6 h2h4 h7h6 h4h5 g6g5 f3h2 e7e5 h2g4 d7d6 f1b5 e8e7 g4e3 a7a6 b5d3 e7f6 c3e2 f6g7 e2g3 g8f6 a2a4 c8d7 c1d2 b6b5 a4b5 a6b5 e1g1 g7h7 b2b4 c5b4 d2b4 h8g8 f1e1 a5c4 d3c4 a8a1 d1a1 b5c4 a1a7');
INSERT INTO evaluations VALUES(18437,30,'cp 249',NULL);
INSERT INTO evaluations VALUES(18437,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(18437,32,'cp 256',NULL);
INSERT INTO evaluations VALUES(18437,33,'cp 252',NULL);
INSERT INTO evaluations VALUES(18437,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(18437,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(18437,36,'cp 268','c6a5 b1c3 g7g6 e4e5 f8g7 d5d6 a7a6 h2h4 h7h6 c3d5 e7d6 e5d6 e8f8 c2c3 c8b7 d5c7 g8f6 c7a8 d8a8 f1d3 a8e8 d1e2 e8e2 d3e2 b6b5 e1g1 f6d5 f1e1 g7f6 f3d2 d5f4 e2f1 f4e6 h4h5 g6h5 d2e4 b7e4 e1e4 f8g7 c1e3 a5c6 f2f4 b5b4 a1c1 b4c3 b2c3 h8b8 f1c4 h5h4 c4e6 f7e6 e3c5');
INSERT INTO evaluations VALUES(18424,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(18424,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(18424,32,'cp 288',NULL);
INSERT INTO evaluations VALUES(18424,33,'cp 296',NULL);
INSERT INTO evaluations VALUES(18424,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(18424,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(18424,36,'cp 292','b1c3 g7g6 e4e5 f8g7 d5d6 a7a6 h2h4 h7h6 c3d5 e7d6 e5d6 e8f8 c2c3 c8b7 d5c7 g8f6 c7a8 d8a8 f1d3 a8e8 d1e2 e8e2 e1e2 f8g8 h1e1 g8h7 e2f1 b7d5 f3e5 d5e6 c1e3 f6h5 e3d2 g7e5 e1e5 a5c4 d3c4 e6c4 f1e1 h5g7 e5e7 c4e6 d2f4');
INSERT INTO evaluations VALUES(20661,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(20661,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(20661,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(20661,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(20661,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(20661,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(20661,36,'cp -96','h7h6 g5f3 d6d5 e4d5 e6d5 c4b3 g8f6 b1c3 c8e6 c3e2 e6g4 e2g3 d8e7 d1e2 e7e2 g3e2 f8d6 e2f4 c6e7 d3d4 g4f3 g2f3 c5c4 b3a4 e8f8 c2c3 g7g6 h2h4 f8g7 e1f1 h8b8 a4c2 b7b5 h4h5 g6g5 f4e2 a7a5 b2b3 b8e8 c1e3 a8d8 f3f4 g5g4 b3c4 d5c4');
INSERT INTO evaluations VALUES(20141,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(20141,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(20141,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(20141,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(20141,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(20141,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(20141,36,'cp -90','g5f3 d6d5 c4b3 d5e4 d3e4 d8d1 e1d1 b7b6 a2a4 c8a6 b1a3 e8c8 d1e1 g7g5 b3c4 a6b7 a4a5 c6a5 f3e5 h8h7 c1d2 a5c4 a3c4 f7f6 e5g6 f8d6 c4d6 d8d6 f2f3 f6f5 g6e5 f5e4 a1a7 e4f3 g2f3 g8f6 e5c4 d6d4 c4b6 c8c7 b6a4 c7b8 a7a5 b7f3 h1f1 d4e4 e1f2');
INSERT INTO evaluations VALUES(20142,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(20142,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(20142,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(20142,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(20142,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(20142,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(20142,36,'cp -91','d6d5 e4d5 e6d5 c4b5 f8d6 e1g1 g8e7 d3d4 c5d4 f3d4 e8g8 b1c3 a7a6 b5a4 d8c7 h2h3 c6d4 d1d4 c8e6 c3e2 d6h2 g1h1 h2e5 d4d1 e7f5 c2c3 a8d8 e2d4 e5d4 c3d4 e6d7 a4d7 c7d7 c1d2 f8e8 f1e1 e8e4 d2c3 d8e8 h1g1 e8e6 d1d2 e6g6 e1e4 d5e4 d2f4 f5d6 g1h1 d7c6 a1e1');
INSERT INTO evaluations VALUES(27246,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(27246,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(27246,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(27246,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(27246,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(27246,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(27246,36,'cp 136','c1e3 f8e7 d1d2 g8f6 f2f3 c6d4 e3d4 d7e6 e1c1 e8g8 c1b1 d6d5 d2f2 c7c6 g2g4 d8c7 g4g5 f6h5 e4d5 e6d5 c3d5 c6d5 d4e3 h5f4 h2h4 g7g6 f1b5 b7b6 f2d2 e7d6 c2c4 a7a6 b5a4 b6b5 c4b5 a6b5 a4b3 d6e5 e3f4 e5f4 d2d5 f8d8');
INSERT INTO evaluations VALUES(27244,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(27244,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(27244,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(27244,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(27244,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(27244,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(27244,36,'cp 118','g8f6 f2f3 f8e7 d1d2 e8g8 e1c1 c6d4 e3d4 c7c6 g2g4 d7e6 c1b1 d6d5 d2f2 d8c7 g4g5 f6h5 e4d5 e6d5 c3d5 c6d5 d4e3 h5f4 h2h4 g7g6 f1d3 b7b6 e3c1 e7d6 h4h5 f4d3 d1d3 d6f4 c1f4 c7f4 d3d4 f4g5 d4h4 g5e7 h5g6 f7g6 h4h7 e7h7 h1h7');
INSERT INTO evaluations VALUES(25544,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(25544,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(25544,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(25544,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(25544,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(25544,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(25544,36,'cp 147','f2f3 f8e7 d1d2 c6d4 e3d4 d7e6 e1c1 e8g8 g2g4 d6d5 d2f2 c7c6 c1b1 d8c7 g4g5 f6h5 e4d5 e6d5 c3d5 c6d5 d4e3 h5f4 h2h4 f8d8 f1b5 f4e6 e3c1 e7c5 f2g2 c7b6 b5a4 c5e3 c1e3 b6e3 a4b3 e6d4 h4h5 e3f4 h5h6 g7g6 h1e1 a7a5 g2g4 f4g4 f3g4');
INSERT INTO evaluations VALUES(29135,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(29135,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(29135,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(29135,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29135,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(29135,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(29135,36,'cp 32','f3e5 b8c6 f1b5 f5d7 e1g1 e7e6 f1e1 a7a6 b5c6 d7c6 c1f4 f8d6 e1e3 d8c7 e3g3 g7g6 g3e3 f6d7 d1d2 e8g8 f4g3 d7f6 e5d3 d6g3 h2g3 a8e8 a1e1 g8g7 d3e5 h7h5 f2f3 f6d7 c3e2 d7e5 d4e5 f8h8 e2d4 h5h4 g3h4 h8h4 c2c3');
INSERT INTO evaluations VALUES(55341,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(55341,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(55341,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(55341,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(55341,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(55341,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(55341,36,'cp 106','c1e3 b8c6 b1c3 g8f6 d1d2 c6d4 e3d4 f8g7 f2f3 c8e6 e1c1 c7c6 g2g4 e8g8 h2h4 b7b5 g4g5 f6h5 d4g7 h5g7 d2d6 d8b6 d6d4 b5b4 c3a4 b6a5 a4c5 a8b8 f1d3 a5a2 b2b3 a2a5 c5e6 f7e6 c1b1 b8d8 d4c4 a5b6 h4h5 g6h5');
INSERT INTO evaluations VALUES(10217,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(10217,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(10217,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(10217,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(10217,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(10217,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(10217,36,'cp -37','d7d5 c1g5 d5d4 c3d5 h7h6 d5f6 g7f6 g5h4 c5e7 f3d2 c8e6 a2a3 h8g8 d1h5 d8d6 f1e2 d6c5 a1c1 c5b5 c1b1 e8c8 e2g4 c8b8 g4e6 f7e6 e1g1 d8f8 d2c4 a7a6 g1h1 c6d8 b2b4 d8f7 b1a1 b5d7 a3a4 f7d6 b4b5 d6c4 d3c4 f6f5 b5a6');
INSERT INTO evaluations VALUES(9090,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(9090,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(9090,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(9090,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(9090,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(9090,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(9090,36,'cp -12','c1g5 d5d4 c3d5 c8e6 d5f6 g7f6 g5d2 c6e7 g2g3 d8d7 f3h4 f6f5 d1h5 c5d6 e4f5 e7f5 h4f5 e6f5 f1g2 e8c8 e1c1 h8g8 g2d5 f5e6 d5e6 d7e6 c1b1 e6g6 h5g6 g8g6 d1e1 d8e8 c2c3 d4c3 d2c3 g6g5 h1f1 f7f5 b1c2 c8d7 e1d1');
INSERT INTO evaluations VALUES(16470,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(16470,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(16470,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(16470,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(16470,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(16470,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(16470,36,'cp -21','e5d4 f3d4 f8b4 g2g3 b4c3 b2c3 c6a5 d1d3 e8g8 f1g2 d7d6 c1f4 f8e8 e1g1 d8e7 f4g5 h7h6 g5f6 e7f6 a1b1 a8b8 d4b3 a5c6 f2f4 f6e7 e2e4 c8d7 b1e1 b7b6 d3d2 g8h8 b3d4 c6a5 d2d3 b8d8 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(16667,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(16667,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(16667,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(16667,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(16667,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(16667,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(16667,36,'cp 0','f3d4 f8b4 g2g3 b4c3 b2c3 c6a5 d1d3 b7b6 f1g2 c8b7 g2b7 a5b7 c1g5 e8g8 d3f3 b7d6 d4b5 d6e4 f3e4 f6e4 g5d8 a8d8 b5c7 d8c8 c7d5 c8c4 f2f3 e4c5 e1c1 c4a4 c1b2 f8e8 d1d2 a4a5 e2e4 g8f8 h2h4 c5a4 b2b3 h7h5 h1d1 a4c5 b3b2 c5a4');
INSERT INTO evaluations VALUES(16662,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(16662,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16662,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(16662,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(16662,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16662,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(16662,36,'cp -9','f8b4 g2g3 b4c3 b2c3 c6a5 d1d3 e8g8 f1g2 d7d6 c1f4 f8e8 e1g1 d8e7 a1d1 h7h6 f1e1 f6d7 d4f5 e7f6 e2e4 d7c5 d3d4 c8f5 d4f6 g7f6 e4f5 g8g7 f4e3 a5c4 e3c5 d6c5 e1e8 a8e8 d1d7 e8e1 g2f1 c4e5 d7c7 e1a1 c7b7 a1a2');
INSERT INTO evaluations VALUES(10510,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(10510,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(10510,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(10510,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(10510,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(10510,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(10510,36,'cp -10','g2g3 b4c3 b2c3 c6a5 d1d3 e8g8 f1g2 d7d6 c1f4 d8e7 e1g1 f8e8 f1e1 f6d7 d4b3 a5c6 c4c5 d7c5 b3c5 d6c5 g2c6 b7c6 d3f3 a8b8 f3c6 b8b2 c6c7 e7c7 f4c7 h7h5 c7d6 b2e2 e1e2 e8e2 d6c5 a7a5');
INSERT INTO evaluations VALUES(24924,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24924,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24924,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24924,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(24924,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24924,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24924,36,'cp 29','e7e6 b1d2 f8d6 f4g3 e8g8 e1g1 g4f3 d2f3 d6g3 h2g3 d8e7 c2c4 c6b8 a2a3 a7a5 b2b4 a5b4 a3b4 b8d7 d1b3 c7c6 a1a5 d5c4 e2c4 b7b6 a5a8 f8a8 c4e2 b6b5 b3c2 a8b8 e2d3 e7d6 e3e4 d6b4 c2c6 g7g6');
INSERT INTO evaluations VALUES(25215,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(25215,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(25215,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(25215,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(25215,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(25215,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(25215,36,'cp 32','b1d2 f8d6 f4g3 e8g8 e1g1 d6g3 h2g3 c6b8 c2c4 c7c6 d1b3 d8b6 b3c3 a7a5 f1c1 b8d7 b2b3 f6e4 c3c2 e4d2 c2d2 b6b4 d2b2 b4e7 a2a3 h7h5 c4d5 e6d5 b3b4 a5b4 a3b4 a8a1 c1a1 g4f3 e2f3 g7g6 b4b5 d7f6 b5c6');
INSERT INTO evaluations VALUES(25217,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(25217,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25217,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25217,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(25217,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25217,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(25217,36,'cp 22','f8d6 f4g3 e8g8 e1g1 d6g3 h2g3 c6b8 b2b4 c7c6 c2c4 b8d7 d1b3 h7h5 a2a4 d8e7 f1e1 a8d8 b4b5 c6c5 b3a3 f8e8 a4a5 g4f3 d2f3 d5c4 e1c1 c5d4 a3e7 e8e7 e3d4 g7g6 c1c4 e6e5 d4e5 d7e5 f3e5 e7e5');
INSERT INTO evaluations VALUES(28757,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(28757,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(28757,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(28757,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(28757,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(28757,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(28757,36,'cp 70','f8b4 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 d2f3 b4c3 b2c3 d8c7 f1e1 c6e5 f3e5 c7e5 a1b1 a8a2 c3c4 b7c6 f2f4 e5a5 c1d2 a5a8 d1f3 c6a4 b1c1 a4c6 f3e3 f8b8 d2c3 b6b5 c3f6 g7f6 e3g3 g8f8 c4b5 c6b5 d3b5 b8b5 g3h4 f8e7');
INSERT INTO evaluations VALUES(30861,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(30861,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(30861,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(30861,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(30861,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(30861,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(30861,36,'cp 10','d8d7 e1g1 g8f6 f3e5 d7c8 d2d4 c5d4 f1e1 f8e7 d1d4 e8g8 b1c3 e7c5 d4d1 d5d4 c3b5 b8c6 e5c6 c8c6 b5d4 c5d4 d1d4 f8d8 d4h4 c6c2 c1g5 d8d6 h4g3 a8d8 a1c1 c2f5 h2h3 h7h5 e1e5 f5g6 c1c7 d8d7 c7c8 g8h7 g1h2 d6d5 g5f4 g6g3 h2g3 d5e5 f4e5 f6e4 g3f3');
INSERT INTO evaluations VALUES(32757,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32757,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32757,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32757,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(32757,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(32757,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(32757,36,'cp 5','e1g1 g8f6 f3e5 d7c8 f1e1 f8e7 d2d4 c5d4 d1d4 e8g8 b1c3 e7c5 d4d1 d5d4 c3b5 b8c6 e5c6 c8c6 b5d4 c5d4 d1d4 f8d8 d4b4 a7a5 b4b3 a5a4 b3f3 c6f3 g2f3 a8c8 c2c3 d8d3 c1e3 a4a3 a1b1 h7h6 b2a3 c8c3 a3a4 c3a3 e1e2 a3a4 b1b7 f6d5 e2d2 d3d2 e3d2 a4a2 b7b8 g8h7');
INSERT INTO evaluations VALUES(32756,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(32756,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(32756,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(32756,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(32756,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(32756,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(32756,36,'cp 11','g8f6 d2d4 c5d4 d1d4 b8c6 d4a4 f8e7 b1c3 e8g8 c1e3 e7b4 a4b3 f8e8 a1d1 a8d8 a2a3 b4c3 b3c3 h7h6 h2h3 e8e6 f1e1 d8e8 e3d4 c6d4 c3d4 a7a6 e1e6 e8e6 d4c3 e6b6 b2b3 d7c6 c3c6 b6c6 f3d4 c6c3');
INSERT INTO evaluations VALUES(32557,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32557,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(32557,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(32557,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(32557,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(32557,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(32557,36,'cp 17','d2d4 c5d4 d1d4 b8c6 d4a4 f8e7 b1c3 e8g8 c1e3 e7b4 a4b3 b4c3 b3c3 h7h6 a1d1 a8d8 f1e1 f8e8 e3d4 f6e4 c3d3 e8e6 a2a3 d8e8 h2h3 c6d4 f3d4 e4c5 d3d2 e6e1 d1e1 e8e1 d2e1 c5e6 d4e6 d7e7 e1e5 e7e6 e5b8 g8h7 b8b7 e6e1 g1h2');
INSERT INTO evaluations VALUES(43042,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(43042,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(43042,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(43042,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(43042,34,'cp -108',NULL);
INSERT INTO evaluations VALUES(43042,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(43042,36,'cp -140','e8g8 e4e5 f8e8 f1e2 f6d5 g5g3 b8c6 g1f3 h7h5 h2h4 d7d6 d2d4 d6e5 d4e5 c8g4 b1c3 d5c3 b2c3 d8d5 e2d3 c6e5 f3e5 e7f6 e1g1 d5e5 c1d2 e5g3 f2g3 g8f8 a1e1 b7b6 c3c4 f6d4 g1h2 e8e1 f1e1 a8e8 d2b4 c7c5 e1e8 f8e8 b4d2 e8d7 d2f4 f7f6 h2h1 d7e7 d3e4 g4e6 e4d3 f6f5');
INSERT INTO evaluations VALUES(37325,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(37325,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(37325,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(37325,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(37325,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(37325,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(37325,36,'cp -131','e4e5 f8e8 f1e2 f6d5 g5g3 b8c6 g1f3 h7h5 h2h4 d7d6 d2d4 d6e5 d4e5 c8g4 b1c3 d5c3 b2c3 d8d5 e2d3 e7f6 e1g1 c6e5 f3e5 d5e5 c1d2 e5g3 f2g3 b7b6 f1e1 e8d8 a2a4 a7a5 a1b1 d8d7 c3c4 f6d4 g1f1 d4c5 d2f4 a8d8 e1e5 g7g6 e5d5 g4e6 d5d7 d8d7 f1e1');
INSERT INTO evaluations VALUES(37312,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(37312,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(37312,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(37312,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(37312,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(37312,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(37312,36,'cp -146','f8e8 f1e2 f6d5 g5g3 b8c6 g1f3 h7h5 h2h4 d7d6 d2d4 d6e5 d4e5 c8g4 b1c3 d5c3 b2c3 d8d5 e2d3 c6e5 f3e5 e7f6 e1g1 d5e5 c1d2 b7b6 a1e1 e5g3 f2g3 g8f8 e1e8 a8e8 g1f2 a7a5 f1b1 e8e6 b1b5 g7g6 d3c4 e6e7 c4d3 f8e8 c3c4 f6d4 f2f1 g4e6 b5b1 e8d7 b1d1 d7d6 d2g5');
INSERT INTO evaluations VALUES(45026,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(45026,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45026,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45026,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45026,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(45026,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(45026,36,'cp 0','b7b5 f1e2 e7e6 e1g1 c8b7 a2a4 a7a6 b2b3 b5b4 e4e5 f6d5 c3d5 c6d5 b3c4 d5c4 e2c4 f8e7 c1e3 e8g8 f3d2 g8h8 a4a5 b8c6 d2b3 c6a7 b3c5 b7d5 a1c1 a7c6 d1a4 a8c8 c4d5 d8d5 f1d1 f8d8 c5b7 d8e8 b7c5 e8d8');
INSERT INTO evaluations VALUES(43939,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(43939,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(43939,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(43939,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(43939,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(43939,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(43939,36,'cp -8','a2a4 b5b4 c3a2 f6e4 f1c4 c8f5 a2b4 d8a5 c1d2 e4d2 d1d2 e7e6 b4d3 a5d2 e1d2 f8d6 f3e5 a7a5 a1e1 f5d3 e5d3 e8g8 e1c1 f8d8 c4a2 c6c5 d4d5 e6d5 a2d5 a8a7 g2g3 d6f8 d5c4 b8c6 h1e1 a7d7 d2e2 g7g6 b2b3 f8g7 e2f1 d7d3 c4d3 d8d3 c1c5');
INSERT INTO evaluations VALUES(43943,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(43943,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(43943,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(43943,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(43943,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(43943,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(43943,36,'cp -18','b5b4 c3a2 f6e4 f1c4 c8f5 a2b4 d8a5 c1d2 e4d2 d1d2 e7e6 b4d3 a5d2 e1d2 f8d6 f3e5 a7a5 f2f4 f5d3 e5d3 b8d7 a1c1 e8e7 h1e1 g7g6 g2g4 d7f6 h2h3 h7h5 f4f5 f6d5 f5e6 f7e6 e1e4 h5g4 e4g4 a8g8 c1g1 h8h3 c4d5 e6d5 g4g6 g8g6 g1g6');
INSERT INTO evaluations VALUES(43945,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(43945,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(43945,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(43945,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43945,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(43945,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(43945,36,'cp -17','c3a2 f6e4 f1c4 c8f5 a2b4 d8a5 c1d2 e4d2 d1d2 e7e6 b4d3 a5d2 e1d2 f8d6 f3e5 a7a5 f2f4 f5d3 d2d3 e8e7 d3e3 h8c8 b2b3 b8d7 e5d7 e7d7 a1f1 d7e7 f4f5 e6e5 d4e5 d6e5 e3f3 a8b8 f1e1 e7f6 e1e4 e5d6');
INSERT INTO evaluations VALUES(48933,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(48933,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(48933,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(48933,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(48933,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(48933,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(48933,36,'cp 28','d1d8 e8d8 g1f3 f8d6 c1e3 d8e7 f3d2 c8e6 f2f3 b8d7 f1c4 d6b4 c4e6 e7e6 c3d1 h7h5 a2a3 b4e7 a3a4 h5h4 a4a5 b7b5 d1f2 f6h5 e1c1 g7g5 d2b3 h5f4 g2g3 f4e2 c1b1 c7c5 g3h4 h8h4 d1d5 a8h8 f2d3 h4h2 h1h2 h8h2 e3c5');
INSERT INTO evaluations VALUES(35396,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(35396,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(35396,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(35396,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(35396,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(35396,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(35396,36,'cp 51','e8d8 g1f3 f8d6 c1e3 d8e7 h2h3 b8d7 e1c1 b7b6 f3h4 g7g6 g2g4 c8b7 f2f3 c7c6 h4g2 d7c5 h3h4 c5e6 f1c4 a8d8 h4h5 b7c8 g2e1 d6c5 e3d2 e6d4 c3e2 g6h5 g4g5 f6e8 h1h5 e8d6 e2d4 c5d4 c4b3 c8e6 b3e6 e7e6 e1d3 d6c4');
INSERT INTO evaluations VALUES(35429,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(35429,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(35429,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(35429,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(35429,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(35429,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(35429,36,'cp 37','g1f3 f8d6 c1e3 d8e7 f3d2 c8e6 f2f3 b8d7 e1c1 d6c5 e3c5 d7c5 f1c4 e6d7 b2b4 c5a4 c3a4 d7a4 c4b3 a4b3 d2b3 b7b6 b3d2 h8d8 d2c4 e7e6 c1b2 h7h5 b2b3 f6e8 a2a4 g7g6 c2c3 e8d6 c4d6 d8d6 d1d6 e6d6 h1d1 d6e6 b3c4 a7a6 g2g3 e6e7 a4a5');
INSERT INTO evaluations VALUES(35428,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(35428,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(35428,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35428,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(35428,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(35428,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(35428,36,'cp 44','f8d6 c1e3 b8d7 e1c1 f6g4 e3g5 f7f6 g5h4 h7h5 h2h3 g4h6 h4g3 g7g5 h3h4 g5g4 f3e1 h6f7 f1c4 d8e7 f2f3 d7b6 c4b3 g4f3 g2f3 c8e6 g3f2 a8g8 b3e6 e7e6 e1d3 g8g2 a2a4 h8g8 b2b3 b6c8 c3d5 b7b6 c2c4 f7d8 c4c5 b6c5 d5c7 d6c7 d3c5 e6f7 d1d7 f7e8 d7c7 g2f2 c7c8 f2f3 c1b2 g8g3');
INSERT INTO evaluations VALUES(35476,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(35476,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(35476,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(35476,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(35476,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(35476,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(35476,36,'cp 33','c1e3 d8e7 h2h3 c8e6 f3g5 b8d7 g5e6 e7e6 f2f3 c7c6 h3h4 e6e7 h4h5 a7a5 g2g4 d7f8 e1c1 f6d7 c1b1 d6c5 e3c1 f8e6 f1c4 a5a4 c4e6 e7e6 d1d2 h8d8 h1d1 b7b5 c3e2 g7g6 h5g6 f7g6 a2a3 c5e7');
INSERT INTO evaluations VALUES(14315,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(14315,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(14315,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(14315,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(14315,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(14315,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(14315,36,'cp -16','d5e4 d4d5 c6e5 f3e4 e7e6 g1f3 e5f3 d1f3 e6d5 c3d5 f6d5 e4d5 f8d6 f3e4 d8e7 e4e7 e8e7 f1d3 c8d7 c1d2 f7f6 e1g1 b7b5 a1c1 a8b8 d2f4 e7f7 f4d6 c7d6 c2c4 b5c4 c1c4 h8c8 f1c1 f6f5 c4c7 f7f6 b2b3 h7h5 c1c4 c8c7 c4c7 d7b5 d3b5');
INSERT INTO evaluations VALUES(14400,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(14400,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(14400,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(14400,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(14400,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(14400,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(14400,36,'cp -7','d4d5 c6e5 f3e4 e7e6 g1f3 e5f3 d1f3 e6d5 c3d5 f6d5 e4d5 f8d6 f1b5 c8d7 e1g1 e8g8 b5d3 d8e7 c1f4 a8e8 c2c4 g7g6 a2a3 a7a5 f3g3 d6f4 g3f4 e7c5 f4f2 c5d6 a1e1 f7f5 f2d2 e8e1 f1e1 b7b6 d2c3 f5f4 g1f2 f8e8 e1e8 d7e8 g2g3 f4g3 h2g3 e8d7');
INSERT INTO evaluations VALUES(41430,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(41430,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(41430,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(41430,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41430,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41430,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(41430,36,'cp 13','f6e4 g5f4 e4c3 b2c3 e8g8 e2e3 c7c5 c4d5 d8d5 f1e2 c5d4 c3d4 d5a5 d1d2 b8c6 d2a5 c6a5 h2h4 c8f5 h4h5 f8c8 h5h6 g7f8 f3d2 c8c2 e3e4 f5d7 e1g1 a8c8 f1d1 f7f6 e2d3 c2b2 d2b3 g6g5 f4e3 a5b3 a2b3 b2b3 a1a7 f8h6');
INSERT INTO evaluations VALUES(41558,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(41558,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(41558,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41558,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(41558,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(41558,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41558,36,'cp 19','g5f4 c7c5 c4d5 e4c3 b2c3 c5d4 c3d4 d8d5 e2e3 e8g8 f1e2 d5a5 d1d2 a5d2 e1d2 b8c6 h1b1 f8d8 d2e1 b7b6 h2h4 c8b7 a2a4 c6a5 g2g4 a8c8 h4h5 c8c3 f3d2 h7h6 h5g6 f7g6 b1b2 b7d5 g4g5 h6g5 f4g5 g7f6 g5f6');
INSERT INTO evaluations VALUES(44536,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(44536,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(44536,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(44536,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(44536,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44536,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44536,36,'cp 19','f1g2 d7d5 c4d5 f6d5 g1f3 b8c6 d2d4 c5d4 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 b1c3 c8f5 c1d2 e7e5 a1c1 f5e6 d2e3 f8e7 d1c2 e8c8 c2b1 c8b8 c3b5 e6d7 h1e1 a7a6 b5c3 d7f5 g2e4 c6d4 c1d1 f5e4 c3e4 f7f5 e3d4 e5d4');
INSERT INTO evaluations VALUES(44537,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44537,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(44537,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(44537,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(44537,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(44537,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(44537,36,'cp 12','d7d5 c4d5 f6d5 d2d4 c5d4 g1f3 b8c6 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 c1e3 e7e5 b1c3 c8e6 a1c1 f8e7 d1c2 e8c8 c2b1 c8b8 g2c6 b7c6 f2f3 f7f5 c3a4 e6d5 e3c5 e7f6 a4c3 h7h5 e2e4 d5e6 h1d1 h5h4 c5d6 b8b7 e4f5 e6f5 c3e4 f5e4 f3e4');
INSERT INTO evaluations VALUES(45344,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(45344,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(45344,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45344,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(45344,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(45344,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45344,36,'cp 11','c4d5 f6d5 d2d4 c5d4 g1f3 b8c6 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 b1c3 c8f5 c1d2 e7e5 a1c1 f5e6 d1e1 f7f5 g2c6 b7c6 f2f3 f8e7 b2b3 a8c8 e1f2 c6c5 c3b5 a7a6 b5a3 e5e4 a3c4 e8g8 h1d1 f8d8 d2e3 d8d1 c1d1 e4f3');
INSERT INTO evaluations VALUES(47617,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(47617,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47617,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(47617,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(47617,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(47617,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(47617,36,'cp 12','f8e7 b1c3 c8e6 e1g1 e8g8 f1e1 d8d7 a2a3 c7c6 c4e6 f7e6 b2b4 e7d8 c1e3 d8c7 c3e2 d7f7 c2c4 b8d7 d1c2 f6h5 a1d1 f7e7 f3d2 a7a5 d3d4 e5d4 e2d4 a5b4 a3b4 d7e5 d2f3');
INSERT INTO evaluations VALUES(40688,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(40688,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(40688,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40688,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(40688,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40688,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(40688,36,'cp 19','b1c3 c8e6 e1g1 d8d7 c4e6 f7e6 b2b4 e8g8 c1e3 a7a6 a2a4 d7e8 d3d4 b8d7 d4e5 d7e5 f3e5 d6e5 d1b1 a6a5 b4b5 f6d7 b1b3 e7c5 c3d1 b7b6 d1b2 e8g6 b2d3 c5e3 f2e3 f8f7 f1f7 g6f7 a1d1 a8f8');
INSERT INTO evaluations VALUES(40679,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(40679,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40679,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40679,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(40679,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(40679,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(40679,36,'cp 10','c8e6 e1g1 e8g8 a2a4 a7a6 c1d2 c7c6 f1e1 e6c4 d3c4 a6a5 d2e3 b8a6 d1d2 f6d7 a1d1 a6c5 f3h2 c5e6 h2f1 h6h5 f1g3 g7g6 g3e2 d7c5 f2f3 d8c7 e1f1 a8e8 b2b3 g8h7 g1h1 c5d7 g2g4 c7d8 f3f4 e5f4 e2f4 h5g4');
INSERT INTO evaluations VALUES(55327,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(55327,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(55327,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(55327,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(55327,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(55327,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(55327,36,'cp 183','d2d4 d8e7 b1c3 c7c6 a2a4 c8g4 d4d5 g8f6 b2b3 f8g7 h2h3 g4f3 d1f3 e8g8 a4a5 g8h8 e1g1 f6d7 d5c6 b8c6 c3d5 e7d8 c1a3 d7c5 a3c5 d6c5 a5a6 b7a6 a1a6 c6b4 d5b4 c5b4 c4d5 a8c8 a6a7 c8c2 a7f7 f8f7 f3f7 c2c7 f7f3 d8f6 f3g4 c7c3 f1a1');
INSERT INTO evaluations VALUES(55331,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(55331,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(55331,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(55331,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(55331,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(55331,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(55331,36,'cp 175','d8e7 b1c3 c7c6 a2a4 c8g4 d4d5 g8f6 b2b3 c6d5 c4b5 b8d7 c1g5 e7e6 b5d7 f6d7 c3d5 a8c8 h2h3 g4f3 d1f3 f8g7 c2c4 h7h6 g5e3 a7a6 e1g1 e8g8 f1d1 f7f5 a4a5 f5f4 e3b6 d7f6 f3e2 f8f7 e2d3 g7f8 d5f6 f7f6 f2f3 h6h5 g1f1 g6g5 d3d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(55949,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(55949,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55949,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55949,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(55949,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(55949,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(55949,36,'cp 13','b2b3 g8f6 c2c4 f8d6 b1c3 e8g8 c1b2 f8e8 d2d4 b7b6 f1e2 b8d7 e1g1 c8b7 a2a4 c7c6 d1c2 a7a5 e2d3 h7h6 a1c1 d8e7 f1d1 a8c8 c2d2 d6b8 d2e2 e6e5 d4e5 d7e5 c4d5 e5d3 e2d3 c6d5 c3b5 f6e4 c1c8');
INSERT INTO evaluations VALUES(55933,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(55933,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55933,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(55933,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55933,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(55933,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(55933,36,'cp 3','g8f6 c1b2 f8e7 d2d4 e8g8 f1d3 b7b6 e1g1 c8b7 b1d2 b8d7 a2a4 c7c5 a4a5 b6a5 b2c3 a5a4 a1a4 a7a5 d1a1 c5d4 e3d4 e7b4 f3e5 d8b6 d2f3 f8c8 c3e1 b7c6 a4a2 c6b7 e1b4 a5b4 e5d7 f6d7');
INSERT INTO evaluations VALUES(48407,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(48407,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(48407,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(48407,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48407,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48407,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(48407,36,'cp 12','c1b2 f8d6 c2c4 e8g8 b1c3 f8e8 d2d4 b7b6 f1d3 c8b7 e1g1 b8d7 d1c2 d5c4 b3c4 b7f3 g2f3 c7c5 c3b5 d6f8 d4c5 d7c5 d3e2 e6e5 a1d1 d8e7 b2a3 e8c8 c2f5 g7g6 f5h3 a7a6 b5c3 c8d8 a3b4 e7e6 h3e6 f7e6 d1d8');
INSERT INTO evaluations VALUES(48408,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(48408,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(48408,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(48408,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48408,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48408,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48408,36,'cp 0','f8e7 c2c4 e8g8 b1c3 c7c5 c4d5 f6d5 d1c2 b8c6 a2a3 b7b6 h2h4 f7f5 h4h5 h7h6 d2d4 c5d4 c3d5 d8d5 f1c4 d5e4 c4d3 e4d5');
INSERT INTO evaluations VALUES(41009,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41009,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41009,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(41009,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(41009,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41009,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(41009,36,'cp 0','d2d4 e8g8 f1d3 b7b6 e1g1 c8b7 b1d2 b8d7 d1e2 c7c5 f1d1 a8c8 a1c1 c5d4 e3d4 c8c7 f3e5 e7d6 c2c3 d8e7 f2f4 g7g6 d1f1 f6h5 d3b5 d7f6 b5d3');
INSERT INTO evaluations VALUES(40996,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40996,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40996,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(40996,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40996,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40996,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40996,36,'cp 6','e8g8 b1d2 b7b6 f1d3 b8d7 e1g1 c8b7 d1e2 c7c5 a1c1 f6e4 c2c4 e4d2 f3d2 c5d4 b2d4 e7f6 d4f6 d7f6 c4c5 b6c5 c1c5 a8c8 f1c1 c8c5 c1c5 e6e5 e2e1 f6d7 c5c2 d8a5 d2f3 a5e1 f3e1 f8c8 c2c8 b7c8 a2a3 d7c5 b3b4 c5d3 e1d3');
INSERT INTO evaluations VALUES(36355,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36355,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36355,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36355,36,'cp 9','f1d3 c7c5 e1g1 b8c6 d4c5 f6d7 c2c4 d7c5 b1c3 d5c4 d3c4 a7a6 d1b1 b7b5 f1d1 d8b6 c4f1 c8b7 c3e4 c6e5 b2e5 b7e4 b1b2 f7f6 e5d4 b6b7 d4c5 e7c5 b2e2 e4d5 f3d2 f6f5 a2a4 f8d8 a4b5 a6b5 e2b5 b7b5 f1b5 a8a1');
INSERT INTO evaluations VALUES(11471,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(11471,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(11471,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(11471,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11471,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(11471,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(11471,36,'cp 37','d3d4 f8d6 d4e5 c6e5 c4b3 e8g8 c1f4 d8e7 f4g3 e5f3 d1f3 d6g3 h2g3 d7d6 f3f4 c8e6 e1c1 a8d8 f2f3 f8e8 g3g4 e6b3 a2b3 e7e5 c3e2 g8f8 f4e5 e8e5 c2c4 f6d7 e2d4 d7c5 b3b4 c5e6 b2b3 e6d4 d1d4 f8e8 c1c2 f7f6 d4d1 d8d7');
INSERT INTO evaluations VALUES(14730,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14730,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14730,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14730,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14730,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(14730,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14730,36,'cp 0','f8d6 e1g1 e8g8 d1e2 a7a6 b5c6 b7c6 e3e4 f6e4 c3e4 d5e4 e2e4 a8b8 e4e3 d8e7 f3e5 c6c5 e5c6 e7h4 g2g3 h4h5 c6b8 c8b7 b8d7 c5d4 e3b3 b7c6 d7f8 h5h3 f2f3 d6g3 h2g3 h3g3 g1h1 g3h3 h1g1');
INSERT INTO evaluations VALUES(41260,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(41260,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(41260,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(41260,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(41260,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41260,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(41260,36,'cp 32','d2d4 e8g8 g1f3 c7c5 d4c5 d6c5 c1f4 b8c6 f3e5 c6d4 e1g1 f6h5 e2e3 d4f5 d1d8 f8d8 f1d1 c8e6 g2b7 h5f4 g3f4 d8d1 c3d1 a8d8 b7d5 g7e5 f4e5 e6d5 c4d5 f5h4 e3e4 h4f3 g1f1 f3e5 f1e2 g8f8 b2b3 g6g5 d1e3');
INSERT INTO evaluations VALUES(41276,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41276,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41276,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(41276,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(41276,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41276,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(41276,36,'cp 18','e8g8 h2h3 c7c5 e2e3 c8d7 g1e2 b8c6 e1g1 e7e5 b2b3 c5d4 e3d4 e5d4 e2d4 d6d5 c4d5 f6d5 c3d5 c6d4 c1b2 d7c6 b2d4 c6d5 g2d5 d8d5 d4g7 d5d1 f1d1 g8g7 d1d7 b7b5 a1e1 a7a5 e1e7 a8e8 e7e8 f8e8 d7a7 a5a4 b3a4 b5a4 a7a4 e8e2 g1g2 e2c2 a2a3 h7h5 a4b4');
INSERT INTO evaluations VALUES(46351,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(46351,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46351,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(46351,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(46351,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(46351,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(46351,36,'cp 24','f1d3 b8c6 e1g1 f8d6 c3b5 d6b8 c2c3 e8g8 f1e1 h7h6 b5a3 f6d7 a3c2 f8e8 h2h3 e6e5 d4e5 d7e5 f3e5 b8e5 c1e3 d8f6 e3d4 c8d7 d4e5 e8e5 d1d2 a8e8 d3b5 d7f5 e1e5 e8e5 b5d3 a7a6 a1d1 f5e4 d3e4 d5e4');
INSERT INTO evaluations VALUES(46354,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(46354,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46354,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(46354,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(46354,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(46354,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46354,36,'cp 21','b8c6 e1g1 f8d6 f1e1 e8g8 c3e2 d8c7 c2c3 h7h6 g2g3 e6e5 d4e5 d6e5 f3e5 c6e5 e2d4 e5d3 d1d3 c8d7 f2f3 f8e8 c1e3 e8e7 d3d2 a8e8 g1g2 f6h7 e3f2 h7g5 g3g4 e7e1 a1e1 e8e1 d2e1 g5e6 f2g3 c7b6 d4e6 b6e6 e1e6 f7e6');
INSERT INTO evaluations VALUES(10432,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(10432,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(10432,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10432,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10432,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10432,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10432,36,'cp 8','b2c3 h7h6 e1g1 e8g8 f1e1 f8e8 c1d2 d7d6 h2h3 b7b6 a2a4 a7a6 c4b3 c6a5 b3a2 c7c5 c3c4 c8d7 a2b1 a5c6 c2c3 f6h7 f3h2 e8e6 h2f1 e6g6 f1g3 d8h4 e1e3 g6f6 b1c2 a8f8 d1e2 f8e8 e3f3');
INSERT INTO evaluations VALUES(17714,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(17714,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(17714,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(17714,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(17714,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(17714,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(17714,36,'cp 50','c8f5 c2c4 g8f6 b1c3 e7e6 a1c1 f8b4 a2a3 b4c3 c1c3 f6e4 c3c1 g7g5 f3g5 e4g5 h2h4 g5e4 g2g4 e4f2 e1f2 f5e4 f1g2 e4g2 f2g2 d5c4 d1f3 d8d5 f4c7 a8c8 c7b6 h8g8 g4g5 g8g6 h1g1 f7f5 f3d5 e6d5 g2f3 e8d7 b6c5 d7e6 h4h5 g6g8 c1e1 c8e8');
INSERT INTO evaluations VALUES(25648,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(25648,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(25648,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(25648,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25648,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(25648,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(25648,36,'cp 65','c2c4 e7e6 c4d5 c6b4 b1c3 b4d5 f4e5 g8f6 f1d3 f5d3 d1d3 f8d6 e1g1 d5c3 d3c3 e8g8 c3b3 a8b8 f1d1 d8e7 e5f6 e7f6 a1c1 f6f5 b3c2 c7c6 a2a3 f5c2 c1c2 a6a5 g2g3 f7f6 f3d2 g8f7 d2c4 d6c7 d1c1 f8d8 g1g2 d8d5');
INSERT INTO evaluations VALUES(25639,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(25639,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(25639,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(25639,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(25639,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(25639,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(25639,36,'cp 66','g8f6 b1c3 e7e6 a1c1 h7h6 a2a3 f8d6 c4d5 f6d5 f4g3 d5c3 c1c3 c6e7 f3d2 d6g3 h2g3 e8g8 f1e2 c7c6 e1g1 a6a5 e2f3 d8c7 f1e1 f5h7 d2c4 e7d5 c3c1 f8d8 e3e4 d5b6 d1d2 b6c4 c1c4 d8d7 d2c3 c7b6 e1d1 a8d8 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(49364,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(49364,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(49364,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(49364,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(49364,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(49364,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(49364,36,'cp -10','b4b5 d7d5 e2e3 f8d6 g1f3 c7c6 a2a4 b8d7 f1e2 e8g8 c2c4 e6e5 b5c6 b7c6 c4d5 c6d5 e1g1 f8e8 b1c3 d5d4 e3d4 e5e4 f3e1 a7a6 e1c2 d7f8 c2e3 f8e6 d4d5 e6f4 f1e1 a8b8 b2a3 d6a3 a1a3 f4d5 a3b3 b8b3 d1b3 d5f4 e2c4 d8d2 c4f7 g8h8');
INSERT INTO evaluations VALUES(49343,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49343,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(49343,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(49343,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(49343,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(49343,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(49343,36,'cp -25','c7c6 a2a4 d7d5 g1f3 b8d7 e2e3 f8d6 d2d4 e8g8 c2c4 d6b4 b2c3 a7a5 b5a6 b4c3 b1c3 d8a5 d1d2 b7b6 c4d5 e6d5 c3b1 a5d2 f3d2 c6c5 b1c3 c5d4 e3d4 f8e8 e1d1 c8a6 f1a6 a8a6 h1e1 e8c8 a1c1 d7f8 d2f3 a6a8 e1e7 f8g6 e7b7 f6g4 b7b6 g4f2 d1d2 a8a4 c3a4');
INSERT INTO evaluations VALUES(3743,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(3743,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(3743,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(3743,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(3743,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(3743,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3743,36,'cp 25','a2a4 a7a5 f1e1 c8g4 b1d2 d5b6 c4b5 c5d6 d2e4 c6a7 h2h3 g4h5 d3d4 e5d4 d1d4 a7b5 a4b5 f8e8 e4d6 e8e1 f3e1 d8d6 e1c2 h5g6 c1f4 d6d4 c2d4 b6d5 f4g3 g6d3 f2f3 a5a4 c3c4 d3c4 b2b3 c4d3 b3a4 d5c3 g3c7 a8a4');
INSERT INTO evaluations VALUES(7588,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(7588,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(7588,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(7588,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(7588,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(7588,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(7588,36,'cp 21','f3e5 c5d4 e5g4 e8g8 g4f6 d8f6 e1g1 d7d5 e4d5 f8e8 d2d3 d4c3 b2c3 c6d5 f1e1 c8d7 e1e8 a8e8 c1d2 h7h6 a1b1 c7c5 b1b7 d7c8 b7b8 c8d7 b8e8 d7e8 h2h3 e8a4 d1g4 a4c2 g4c8 g8h7 c8c5 f6g6 c5a7 c2d3 f2f3 d3c4 d2f4 f7f6 g1h2 g6c2 f4g3 c2c3');
INSERT INTO evaluations VALUES(7580,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(7580,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(7580,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(7580,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(7580,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(7580,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(7580,36,'cp 33','c5d4 e1g1 e8g8 e5f3 d4c3 d2c3 f6e4 f1e1 f8e8 d1d4 d7d5 c3c4 c8g4 c1f4 g4f3 g2f3 d8f6 d4f6 e4f6 b2b3 d5c4 f4c7 e8e1 a1e1 a8e8 e1d1 f6d5 c7a5 f7f6 d1d4 c4b3 a2b3 g8f7 d4c4 e8e6 g1f1 d5b6 c4b4 c6c5 b4b5 e6c6 a5d2 b6d7 d2f4 a7a6 b5b7');
INSERT INTO evaluations VALUES(29540,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(29540,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29540,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(29540,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(29540,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29540,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29540,36,'cp 0','e7e6 g1f3 b8d7 c2c4 f8e7 f1e2 e8g8 e1g1 c7c5 c4d5 f6d5 f4g3 c5d4 f3d4 f5g6 d2c4 d7c5 a1c1 a8c8 a2a3 c5e4 d1b3 e4c5');
INSERT INTO evaluations VALUES(30094,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30094,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(30094,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(30094,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(30094,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(30094,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(30094,36,'cp 6','g1f3 f8d6 f3e5 b8d7 f1e2 h7h6 e1g1 f6e4 d2e4 f5e4 c2c4 d5c4 e2c4 e8g8 e5d7 d8d7 d1g4 e4f5 g4f3 d6f4 f3f4 d7d6 f4f3 d6b6 b2b3 c7c6 a2a4 a7a5 h2h4 b6b4 e3e4 b7b5 a4b5 c6b5 e4f5 b5c4 b3c4 b4c4 f5e6 c4e6');
INSERT INTO evaluations VALUES(31030,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(31030,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(31030,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(31030,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31030,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31030,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(31030,36,'cp 19','f1e2 e7e6 h2h4 h7h5 g4h5 g6f5 b1d2 c6c5 g1f3 b8c6 d4c5 f8c5 d2b3 c5b6 a2a4 d8c7 a4a5 c6a5 b3a5 b6a5 c2c3 a5b6 f3d4 g8e7 c1f4 b6d4 d1d4 e7c6 d4a4 e8g8 h5h6 g7g6 h1g1 g8h8 g1g3 a8c8 a1d1 a7a6 e1f1 b7b5 a4a3 c6e5');
INSERT INTO evaluations VALUES(31967,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(31967,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(31967,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(31967,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(31967,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(31967,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(31967,36,'cp 72','e2e3 g8f6 f1c4 e7e6 e1g1 c7c6 f1e1 b8d7 b1c3 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 f8e7 a1b1 e8g8 c3c4 b7b6 g2g3 d8c7 c1b2 a8d8 g1g2 c7b7 a2a4 f8e8 d3b3 b7c7 e3e4 e6e5 d4e5 d7c5 b3c2 d8d3 b1d1');
INSERT INTO evaluations VALUES(31966,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(31966,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(31966,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(31966,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(31966,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(31966,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(31966,36,'cp 64','e7e6 f1c4 g8f6 b1c3 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 e8g8 e3e4 b8d7 a2a4 d8c7 c3c4 e6e5 d4d5 a8c8 c1e3 e7b4 e1c1 c7d6 a4a5 c6d5 e4d5 b4c5 f3g5 d7f6 a1b1 h7h6 g5e4 f6e4 d3e4 f7f5 e4f3 e5e4 e3c5 c8c5 f3e3 b7b6 a5b6');
INSERT INTO evaluations VALUES(32370,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(32370,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(32370,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(32370,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(32370,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(32370,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(32370,36,'cp 61','b1c3 g8f6 f1c4 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 e8g8 a1b1 d8c7 c3c4 b8d7 e3e4 f8d8 e1d1 b7b6 c1b2 e6e5 d4d5 e7d6 b2a3 d7f6 a3d6 c7d6 d3e3 h7h6 h2h3 a8c8 b1c1 c6d5 c4d5 c8c5 f3h4 g7g6 h4f3 g8g7 a2a3 c5c1');
INSERT INTO evaluations VALUES(8043,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(8043,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(8043,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(8043,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(8043,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(8043,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(8043,36,'cp 79','d7d6 f1e2 e8g8 e1g1 c8d7 d4c2 a8c8 h2h3 a7a6 f2f4 c6a5 b2b3 a5c6 d1e1 b7b5 a1c1 b5b4 c3d5 f6e4 e3b6 d8e8 e2f3 f7f5 f3e4 f5e4 d5b4 e7e6 e1e4 a6a5 b4c6 d7c6 e4e3 g7h6 e3d2 e6e5 b6e3 e5f4 e3f4 f8f4 f1f4 e8e5');
INSERT INTO evaluations VALUES(8327,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(8327,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(8327,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(8327,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(8327,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(8327,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(8327,36,'cp 90','f1e2 e8g8 e1g1 a7a6 d1d2 c6d4 e3d4 c8d7 f2f3 d8a5 a1d1 f8e8 a2a3 d7e6 b2b3 a8c8 d2b2 f6d7 d4g7 g8g7 b3b4 a5d8 c3d5 g7g8 d5e3 b7b5 c4b5 a6b5 b2d4 c8b8 f3f4 d7f6 f4f5 e6b3 d1d3 b3a4 e3d5 f6d5 f5g6');
INSERT INTO evaluations VALUES(8326,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(8326,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(8326,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(8326,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(8326,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(8326,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(8326,36,'cp 75','e8g8 e1g1 a7a6 h2h3 c6d4 e3d4 c8d7 d1d3 d7c6 f1d1 a6a5 b2b3 f6d7 d4g7 g8g7 d3c2 d8b6 a2a3 h7h5 c3d5 b6d8 b3b4 a5b4 a3b4 a8a1 d1a1 d7e5 c2c3 c6d5 e4d5 g7g8 e2f1 d8b6');
INSERT INTO evaluations VALUES(3235,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(3235,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(3235,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(3235,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(3235,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(3235,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(3235,36,'cp 84','h2h3 c6d4 e3d4 c8d7 e1g1 d7c6 d1c2 a7a5 b2b3 f6d7 d4g7 g8g7 f1d1 d8c7 c3d5 c7d8 d1d2 d7c5 a1d1 f7f6 e2f1 f8f7 g2g3 d8f8 f1g2 a8d8 h3h4 e7e6 d5e3 h7h5 f2f3 g7h7 c2c3 b7b6 e3c2 c6b7 c3e3 f8e7 g2h3');
INSERT INTO evaluations VALUES(47041,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47041,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(47041,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47041,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(47041,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(47041,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47041,36,'cp 0','g2g3 c7c5 f1g2 g7g6 g1e2 f8g7 b1c3 e8g8 a2a4 b8c6 e1g1 a7a6 c3d5 f6d5 e4d5 c6d4 c2c3 d4f5 h2h3 a8b8 g2e4 f5h6 e4g2');
INSERT INTO evaluations VALUES(52267,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(52267,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52267,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52267,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52267,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(52267,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52267,36,'cp 7','b8c6 d2d3 e7e6 c4b3 d7d6 g1f3 f8g7 c1e3 g8f6 d3d4 c5d4 f3d4 e8g8 d1d2 c6d4 e3d4 b7b5 f2f3 c8d7 a2a4 b5a4 b3a4 a7a5 e1g1 d7a4 a1a4 d6d5 e4d5 f6d5 d4g7 g8g7 d2d4 d5f6 f1d1 d8d4 d1d4 f8c8 d4c4 f6d5 c4c8 a8c8 c3d5 e6d5 a4a5 c8c2 a5d5 c2b2 d5d1 g7h8');
INSERT INTO evaluations VALUES(19228,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19228,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19228,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19228,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(19228,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19228,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(19228,36,'cp 12','g1f3 f8g7 e1g1 d7d6 f1e1 a7a6 a2a4 e7e6 d2d3 b7b6 c3e2 g8f6 c2c3 e8g8 h2h3 c8b7 c4b3 c6e5 e2g3 b6b5 f3e5 d6e5 c1g5 d8c7 d1e2 h7h6 g5e3 g8h7 e1b1 f8d8 a4b5 a6b5 a1a8 d8a8 g3f1 h7g8 f1d2');
INSERT INTO evaluations VALUES(11169,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(11169,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(11169,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11169,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(11169,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(11169,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(11169,36,'cp 54','d4c6 b7c6 e4e5 f6d7 c1f4 d7b6 c4b3 d6d5 e1g1 e7e6 c3a4 a6a5 c2c4 c8a6 a4b6 d8b6 c4d5 e6d5 f1e1 f8c5 d1c2 e8g8 a1c1 c5d4 f4e3 d4e3 e1e3 f8e8 c2c5 e8b8 h2h3 b6c5 c1c5 a6b7 f2f4 g8f8 e3g3 b8e8 g1f2 g7g6 f2e3');
INSERT INTO evaluations VALUES(11136,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11136,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(11136,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(11136,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(11136,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(11136,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(11136,36,'cp 48','b7c6 e4e5 f6d7 c1f4 d7b6 c4b3 d6d5 c3a4 e7e6 e1g1 c8b7 f4e3 b6d7 f2f4 f8e7 c2c4 e8g8 a1c1 a8c8 d1e2 c6c5 c4d5 e6d5 e3f2 g8h8 c1d1 d5d4 b3c4 d8c7 b2b3 g7g6 f2g3 d7b6 a4b6 c7b6 g3f2 b6c6');
INSERT INTO evaluations VALUES(11787,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(11787,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11787,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(11787,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(11787,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(11787,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11787,36,'cp 51','e4e5 f6d7 c1f4 d7b6 c4b3 d6d5 e1g1 e7e6 c3a4 a6a5 c2c4 c8a6 a4b6 d8b6 c4d5 e6d5 f4e3 b6b7 f1e1 f8b4 e3d2 e8g8 d2b4 b7b4 a1c1 a6c8 c1c6 c8g4 d1b1 g4e6 b3c2 h7h6 a2a3 b4b5 c6c3 f8c8 c2h7 g8h8 h7d3 b5b6 b1c2 c8c3 b2c3');
INSERT INTO evaluations VALUES(23464,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(23464,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23464,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(23464,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(23464,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(23464,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(23464,36,'cp 32','d2d4 d7d5 c1f4 a7a6 e2e3 g8f6 g1f3 f6h5 f4e5 h5f6 e5f6 e7f6 f1d3 f8b4 e1g1 c6e7 c3e2 b4d6 c2c4 d5c4 d3c4 e8g8 a1c1 c7c6 h2h3 d8b6 c4b3 f8d8 d1c2 c8f5 e3e4 f5g6 f1e1');
INSERT INTO evaluations VALUES(23434,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(23434,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(23434,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(23434,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(23434,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(23434,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(23434,36,'cp 36','d7d5 c1f4 a7a6 e2e3 c8f5 f1d3 f5d3 d1d3 e7e6 g1f3 f8d6 f4g3 g8f6 e1g1 d8d7 a1d1 e8g8 g3h4 d6e7 f1e1 b7b5 h4f6 e7f6 e3e4 f8e8 g2g3 h7h6 d3e2 b5b4 e4e5 b4c3 e5f6 c3b2 f6g7');
INSERT INTO evaluations VALUES(44728,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44728,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(44728,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44728,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44728,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44728,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44728,36,'cp 0','c8d7 b5e2 f8g7 c1e3 b8c6 d1d2 a8c8 f2f3 a7a6 d4c6 d7c6 e1g1 d8c7 a2a4 e7e6 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(28863,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28863,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28863,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28863,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28863,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28863,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28863,36,'cp 0','b5e2 b8c6 c1e3 f8g7 f2f3 a8c8 d1d2 a7a6 d4c6 d7c6 e1g1 d8c7 a2a4 e7e6 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(5629,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(5629,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(5629,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(5629,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(5629,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(5629,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(5629,36,'cp 48','d2d4 e5d4 f3d4 g8f6 c1f4 e8g8 e1g1 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 f8e8 h2h3 d7c6 e2f3 c6c4 d4c4 e6c4 f1e1 c4e6 b2b3 a8d8 g1f1 f6d7 c3d5 e6d5 e4d5 e7f6 f4d2 e8e1 d1e1 g8f8 b3b4 h7h6 a4a5 b7b5');
INSERT INTO evaluations VALUES(5643,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(5643,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5643,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(5643,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(5643,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(5643,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(5643,36,'cp 50','e5d4 f3d4 g8f6 c1f4 e8g8 e1g1 f6d7 d4c6 b7c6 d1d2 e7f6 b2b3 a7a5 a1d1 f8e8 f1e1 d7e5 c4e2 g7g5 f4e3 d8e7 f2f3 h7h6 a2a4 f6g7 e2f1 f7f5 e4f5 c8f5 e3f2 f5g6 c3e2 e7f6 f3f4 e5g4 f2d4 f6f5 d4g7 g8g7 d2c3 g7h7 c3c6');
INSERT INTO evaluations VALUES(5658,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(5658,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(5658,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(5658,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(5658,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(5658,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(5658,36,'cp 57','f3d4 g8f6 c1f4 e8g8 e1g1 c6e5 c4b3 c8g4 f2f3 g4d7 a2a4 a7a6 f1e1 h7h6 a4a5 c7c5 d4f5 d7f5 e4f5 c5c4 f4e5 c4b3 e5d4 b3c2 d1c2 a8c8 c2b3 d8c7 g2g4 f8e8 g1g2 d6d5 d4f6 e7f6 c3d5 c7c2 b3c2 c8c2 g2g3 e8e1 a1e1 f6b2 e1e8 g8h7 d5b6 c2c1 e8d8 b2c3 b6d7 c3a5');
INSERT INTO evaluations VALUES(5652,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(5652,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(5652,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(5652,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(5652,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(5652,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(5652,36,'cp 39','g8f6 c1f4 e8g8 e1g1 c6e5 c4b3 c8g4 f2f3 g4d7 f1e1 c7c5 d4f5 d7f5 e4f5 d8c8 f4e5 d6e5 d1d3 c8c7 a1d1 a7a6 c3d5 f6d5 b3d5 a8d8 g2g3 e7f6 d3e4 d8d7 g1g2 f8d8 d1d3 g8f8 e1e2 b7b5 c2c4 h7h6 b2b3 c7a5 d3d1');
INSERT INTO evaluations VALUES(14903,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(14903,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14903,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(14903,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(14903,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(14903,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14903,36,'cp 45','f8d6 f1e1 e8g8 b1c3 e6e5 d4d5 c6e7 e3e4 a7a6 a2a4 h7h6 a4a5 c7c6 c1e3 c6d5 e4d5 b7b5 a5b6 e7f5 h2h3 a8b8 c4a6 f5e3 e1e3 b8b6 a6c8 d8c8 f3e5 b6b2 d1c1 b2b6 c1d2 f8e8 a1e1 e8e5');
INSERT INTO evaluations VALUES(9271,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(9271,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(9271,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(9271,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(9271,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(9271,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(9271,36,'cp 72','b1d2 e8g8 c4b5 c8d7 e3e4 d6e7 b5c4 a7a5 f1e1 a5a4 d2b1 c6a5 c4f1 c7c5 d4d5 e6d5 e4d5 f8e8 b1c3 c5c4 f3e5 e7b4 c1g5 a4a3 d1d2 d7a4 b2a3 b4c3 d2c3 f6d5 g5d8 d5c3 d8a5 a8a5 e5c4 a5a6 e1e3 e8e3 c4e3 a6a5 a1c1 a5c5');
INSERT INTO evaluations VALUES(31861,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(31861,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31861,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31861,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31861,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31861,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31861,36,'cp 9','e7e6 c2c4 h7h6 d1b3 b8c6 c1d2 a8b8 f1g2 g8f6 e1g1 f8e7 d2c3 d5c4 b3c4 e8g8 a2a4 f6d5 b1d2 d5c3 c4c3 e7b4 c3e3 d8e7 a1c1 f5h7 f1d1 f8d8 d2c4 a7a5 c4e5 c6e5 f3e5 f7f6 e5d3 c7c6 d3b4 e7b4 e3e6 g8h8 e6c4 b4b2');
INSERT INTO evaluations VALUES(32322,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(32322,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(32322,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(32322,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(32322,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(32322,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(32322,36,'cp 17','c2c4 g8f6 b1c3 h7h6 f1g2 c7c6 d1b3 d8b6 c4c5 b6b3 a2b3 f8e7 b3b4 e8g8 e1g1 a7a6 a1a3 b8d7 h2h3 f6e4 g3g4 f5h7 c1f4 f7f5 e2e3 e7f6 f1a1 a8e8 f4c7 f5g4 h3g4 e6e5 c3e2 e5d4 e2d4 h7g6 c7f4 e4g5 f3g5 h6g5 f4c7');
INSERT INTO evaluations VALUES(44525,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(44525,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(44525,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(44525,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44525,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44525,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44525,36,'cp 0','b1c3 b8c6 g1f3 e7e5 f1e2 f8e7 e1g1 e8g8 d2d3 a7a6 a2a4 d7d6 f3e1 c8e6 e1c2 f6d7 e2g4 d7f6');
INSERT INTO evaluations VALUES(44523,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44523,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44523,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44523,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(44523,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44523,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(44523,36,'cp 6','b8c6 g1e2 e7e6 d2d4 d7d5 c4d5 e6d5 c1g5 c5d4 e2d4 f8e7 f1b5 c8d7 b5c6 b7c6 e1g1 f6e4 g5e7 d8e7 f1e1 e8g8 f2f3 e4c3 e1e7 c3d1 e7d7 d1b2 d4c6 f8e8 d7d5 a8c8 a1c1 g8f8 c1c2 e8e6 c2b2 e6c6 d5a5 c6c1 g1f2 c1c7 f2g3 g7g6 g3f2 f8g8 f2g3');
INSERT INTO evaluations VALUES(46998,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(46998,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(46998,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(46998,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(46998,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(46998,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(46998,36,'cp 41','c7c6 e2e4 d8a5 f4d2 a5c7 h2h3 g7g6 g1f3 f8g7 a2a4 e8g8 f1e2 e7e5 d4e5 d6e5 a4a5 f6h5 e1g1 h5f4 f1e1 f4e2 d1e2 b8d7 e2c4 d7f6 c3a4 b7b5 a5b6 a7b6 c4b3 c8e6 b3b6 c7b6 a4b6 a8a1 e1a1 f6e4');
INSERT INTO evaluations VALUES(14226,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(14226,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(14226,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(14226,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(14226,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(14226,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(14226,36,'cp 5','c8f5 f1d3 f5d3 c2d3 e7e6 g1e2 f8d6 b1c3 d6f4 e2f4 e8g8 e1g1 d8d7 a1c1 b7b6 f4e2 f8c8 d1b3 c6e7 c1c2 d7d6 f1c1 c7c6 c3a4 h7h6 b3c3 d6d7 b2b3 f6e8 c3b2 c8c7 e2f4 c7c8 a2a3 c8c7 a4c3 c6c5 d4c5 b6c5 b3b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(24824,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(24824,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(24824,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(24824,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24824,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(24824,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(24824,36,'cp 5','f1d3 f5d3 c2d3 e7e6 b1c3 f8d6 g1e2 d6f4 e2f4 e8g8 f4e2 d8d7 a1c1 c6e7 e1g1 f8c8 d1b3 b7b6 c1c2 d7d6 c3a4 h7h6 f1c1 c7c6 b3c3 d6d7 c3a3 f6e8 a4c3 e8d6 b2b3 c6c5 a3b2 c8c7 d4c5 c7c5 b2a3 a8c8 a3b2');
INSERT INTO evaluations VALUES(39646,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(39646,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(39646,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39646,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(39646,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(39646,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(39646,36,'cp 17','d2d4 d7d5 f1g2 g8f6 e1g1 e8g8 c2c4 c7c6 f3e5 d5c4 e5c4 c8e6 b2b3 e6d5 c1b2 d5g2 g1g2 b8a6 b1d2 c6c5 d2f3 c5d4 d1d4 d8d4 b2d4 a6b4 a2a3 b4c6 d4b2 f8d8 f1d1 a8c8 f3e5 c6e5 b2e5 f6d5 e5g7 g8g7');
INSERT INTO evaluations VALUES(39598,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(39598,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(39598,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(39598,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(39598,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(39598,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(39598,36,'cp 14','d7d5 f1g2 c7c6 e1g1 g8f6 a2a4 c8f5 a4a5 e8g8 c1f4 b8d7 h2h3 h7h6 g3g4 f5e4 c2c3 e4f3 e2f3 c6c5 f1e1 a8c8 f4e3 e7e6 f3f4 c5d4 c3d4 b7b6 f4f5 g6f5 g4f5 b6a5 b1d2 e6f5 d1f3 c8c2 f3f5 c2b2 d2f3 f6e4');
INSERT INTO evaluations VALUES(38468,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(38468,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38468,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(38468,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(38468,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38468,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(38468,36,'cp 15','f1g2 c7c6 c2c4 g8f6 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 e7e6 e1g1 f8c8 f1c1 h7h6 h2h4 c6e5 f4e5 f6e4 g2e4 f5e4 c3e4 d5e4 e5c7 b6b5 g1g2 h6h5 a2a3 g7f8 c7f4 f7f6 f2f3 e4f3 g2f3 g8f7');
INSERT INTO evaluations VALUES(47800,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47800,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47800,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47800,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47800,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47800,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47800,36,'cp 0','c7c6 a2a3 f8e7 f2f4 e5f4 c1f4 d6d5 c4a2 e8g8 g1f3 d5e4 c3e4 f6e4 d3e4 b8d7 d1d2 d7c5 e1c1 d8d2 d1d2 f8e8 e4e5 c8f5 h1e1 a8d8 d2d8 e8d8 f4e3 h7h6 e3c5 e7c5 e5e6 f5e6 a2e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(45862,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45862,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45862,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45862,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45862,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45862,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45862,36,'cp 0','f2f4 e5f4 g1f3 d6d5 c4b3 d5e4 c3e4 f6e4 d1e2 b8a6 e2e4 d8e7 f3g5 f7f6 g5f7 e7e4 d3e4 h8g8 f7h6 g8h8 h6f7');
INSERT INTO evaluations VALUES(1471,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(1471,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(1471,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(1471,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1471,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(1471,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(1471,36,'cp -12','d1d2 g8h6 g1f3 c8d7 f1e2 h6g4 e3f4 c5d4 c3d4 f7f6 e1g1 f6e5 d4e5 f8c5 f4g3 e8g8 a2a3 g4h6 b2b4 c5d4 b1c3 d4c3 d2c3 h6f5 e2d3 d7e8 d3f5 f8f5 c3e3 b6e3 f2e3 a7a6 a1c1 h7h6 c1c3 e8h5 f3d4 c6d4 e3d4 a8f8 f1f5 f8f5 c3c8 g8h7 c8c7 f5f7 c7f7 h5f7');
INSERT INTO evaluations VALUES(1470,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(1470,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(1470,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1470,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(1470,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1470,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(1470,36,'cp -20','f7f6 f2f4 f6e5 d4e5 g7g5 b1a3 g8e7 f4g5 c6e5 e1c1 c8d7 g1f3 e5f3 g2f3 e8c8 h2h4 e7f5 f1h3 f5h4 d2f2 h4g6 f3f4 f8d6 f4f5 h8f8 c1b1 b6c7 f5f6 g6e5 a3c2 c8b8 e3c1 a7a6 b1a1 f8f7 a1b1 d8f8 f2e2 b8a8 c1f4');
INSERT INTO evaluations VALUES(5647,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(5647,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(5647,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(5647,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(5647,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(5647,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(5647,36,'cp 77','d4d5 c6b4 e1g1 g8f6 b1c3 e8g8 a2a3 b4a6 b2b4 a6b8 h2h3 b8d7 c1e3 a7a5 f3d2 a5b4 a3b4 a8a1 d1a1 f6h5 a1d1 h5f4 f1e1 h7h6 c4f1 d7b6 d2f3 f4g6 e3b6 c7b6 c3b1 f7f5 e4f5 c8f5 c2c4 g6h4 f3h4 e7h4 b1c3');
INSERT INTO evaluations VALUES(5623,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(5623,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(5623,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(5623,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(5623,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(5623,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(5623,36,'cp 69','c6b4 e1g1 g8f6 b1c3 e8g8 a2a3 b4a6 b2b4 a6b8 h2h3 b8d7 c4b3 a7a5 c1e3 d7b6 d1d3 h7h6 f3d2 f6h5 c3e2 g7g5 b4a5 a8a5 e3b6 c7b6 d2c4 a5c5 c4e3 b6b5 a3a4 h5f4 d3d1 b5a4 a1a4 b7b5 a4a1 c5c7 e2g3 g5g4');
INSERT INTO evaluations VALUES(16270,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(16270,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(16270,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(16270,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(16270,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(16270,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(16270,36,'cp 40','f8b4 d1c2 b4c3 c2c3 e8g8 b2b3 d7d6 g2g3 f8e8 f1g2 e6e5 d4d5 c6e7 e1g1 a7a5 c1b2 a5a4 a1e1 c7c6 e2e4 e7g6 f3d2 a4b3 a2b3 h7h5 e1a1 a8a1 f1a1 h5h4 a1a8 d8b6 d5c6 b6c6 c3d3 h4h3 g2f1 c6b6 b2a3 c8e6 a8e8 f6e8');
INSERT INTO evaluations VALUES(9978,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(9978,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(9978,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(9978,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(9978,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(9978,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(9978,36,'cp 53','c1d2 b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 d8f6 e2e3 d7d6 f1e2 b7b6 e1g1 e8g8 f3d2 c8b7 e2f3 c6d8 f3b7 d8b7 d2e4 f6e7 e4g3 c7c5 d4d5 e6d5 c4d5 f7f5 a1e1 a8e8 c3d2 e7f7 g3e2 e8e5 e1d1 f8e8 b2b3 h7h6 e2f4 b7d8 f1e1');
INSERT INTO evaluations VALUES(9977,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(9977,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(9977,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(9977,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(9977,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(9977,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(9977,36,'cp 40','b4c3 d2c3 f6e4 d1c2 e4c3 c2c3 d8f6 e2e3 d7d6 f1e2 b7b6 e1g1 e8g8 f3d2 c8b7 e2f3 c6a5 f3b7 a5b7 a1d1 c7c6 d1c1 f8c8 d2f3 c6c5 c3d2 c5d4 f3d4 b7c5 b2b3 f6g5 f2f3 g5e5 e3e4 a7a5 h2h3');
INSERT INTO evaluations VALUES(34062,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34062,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(34062,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(34062,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(34062,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(34062,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(34062,36,'cp 65','e1g1 e7e5 d2d4 e5f5 c4d3 f8e7 f2f3 f5h5 f3e4 h5d1 f1d1 d7d6 c1f4 e8g8 d1e1 b8c6 c2c3 f7f5 b1d2 f5e4 d3c4 g8h8 e1e4 c8f5 e4e3 f5d7 a1f1 a8e8 e3e1 e7h4 e1e8 f8e8 d2f3 h4e7 c4b5 e7f6 f4g5 e8f8 g5f6');
INSERT INTO evaluations VALUES(48274,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(48274,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48274,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48274,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(48274,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48274,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48274,36,'cp 0','d5c4 e2e3 b7b5 a3a4 c7c6 a4b5 c6b5 b2b3 a7a5 b3c4 b5c4 f1c4 f8b4 c1d2 b8c6 d2b4 c6b4 e1g1 e8g8 b1c3 c8b7 c3a4 f6e4 f3e5 d8g5 e5f3');
INSERT INTO evaluations VALUES(48497,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48497,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48497,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48497,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48497,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(48497,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(48497,36,'cp 0','e2e3 b7b5 a3a4 c7c6 a4b5 c6b5 b2b3 a7a5 b3c4 b5c4 f1c4 f8b4 c1d2 b8c6 e1g1 c8b7 d1e2 e8g8 f1c1 d8e7 b1c3 c6b8 c4d3 b8d7 e3e4 e6e5 d4d5 d7c5 d3c4 h7h6 h2h3 f8d8 c1d1 d8b8 a1c1 b8d8');
INSERT INTO evaluations VALUES(16597,30,'cp 310',NULL);
INSERT INTO evaluations VALUES(16597,31,'cp 322',NULL);
INSERT INTO evaluations VALUES(16597,32,'cp 325',NULL);
INSERT INTO evaluations VALUES(16597,33,'cp 310',NULL);
INSERT INTO evaluations VALUES(16597,34,'cp 319',NULL);
INSERT INTO evaluations VALUES(16597,35,'cp 334',NULL);
INSERT INTO evaluations VALUES(16597,36,'cp 340','f4e5 f6e4 d2d3 e4c5 d3d4 c5e4 d4d5 c6e5 d1e2 f8b4 c1d2 e5f3 g2f3 e8g8 f3e4 d8h4 e1d1 b4c5 d2e1 h4h6 b2b4 c5d4 e1c3 d4c3 b1c3 d7d6 e2d2 h6h4 h1g1 g7g6 d1c1 f7f5 c1b2 f5e4 d2d4 h4h2 f1c4 e4e3 c3e2 c8f5 a1f1 f5h3 f1f8 a8f8 d4e3 h2e5 e3e5 d6e5 e2c3');
INSERT INTO evaluations VALUES(16359,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(16359,31,'cp 317',NULL);
INSERT INTO evaluations VALUES(16359,32,'cp 310',NULL);
INSERT INTO evaluations VALUES(16359,33,'cp 312',NULL);
INSERT INTO evaluations VALUES(16359,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(16359,35,'cp 290',NULL);
INSERT INTO evaluations VALUES(16359,36,'cp 325','f6e4 d2d3 e4c5 d3d4 c5e4 d4d5 c6e5 d1e2 f8b4 c1d2 e5f3 g2f3 e8g8 f3e4 d8h4 e1d1 b4c5 d2e1 h4f4 f1h3 f4h6 h3f5 g7g6 f5g4 d7d6 g4c8 a8c8 c2c4 f7f5 b1c3 c8e8 e1d2 h6h3 d1c2 f5f4 e2f1 h3h5 f1d3 f4f3 c3a4 f3f2 a4c5 d6c5 a1f1');
INSERT INTO evaluations VALUES(16811,30,'cp 344',NULL);
INSERT INTO evaluations VALUES(16811,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(16811,32,'cp 293',NULL);
INSERT INTO evaluations VALUES(16811,33,'cp 293',NULL);
INSERT INTO evaluations VALUES(16811,34,'cp 315',NULL);
INSERT INTO evaluations VALUES(16811,35,'cp 305',NULL);
INSERT INTO evaluations VALUES(16811,36,'cp 336','d2d3 e4c5 d3d4 c5e4 d4d5 c6e5 d1e2 f8b4 c1d2 e5f3 g2f3 e8g8 f3e4 d8h4 e1d1 d7d6 h1g1 b4d2 b1d2 f7f5 e2g2 g7g6 g2g3 h4f6 c2c3 f5e4 f1e2 f6f7 h2h4 g8g7 h4h5 c8f5 d2c4 f7d5 d1c2 e4e3 c2c1 a8e8 c4e3 d5e5 g3e5 e8e5 c1d2 f5d7 h5g6 h7g6 a1f1 f8f1 e2f1 d7c6 f1e2 c6e4 b2b4 d6d5 g1f1 e5e8');
INSERT INTO evaluations VALUES(38121,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(38121,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(38121,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(38121,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(38121,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(38121,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(38121,36,'cp 138','e4d5 e6d5 f1b5 c8d7 d1d5 d8e7 d5e4 e7e4 c3e4 d7b5 e4c5 b7b6 c5b3 b8c6 g1f3 g8e7 c1e3 e7f5 e1c1 b5e2 d1d5 e2f3 g2f3 f5e7 d5d2 e7f5 h1d1 e8g8 d2d5 g7g6 b3d2 a8e8 d2c4 e8d8 d5d8 f8d8 d1d8 c6d8 c2c3 f7f6 c1c2 g8f7 a2a4 d8c6 b2b4');
INSERT INTO evaluations VALUES(38110,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(38110,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(38110,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(38110,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(38110,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(38110,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(38110,36,'cp 144','e6d5 f1b5 c8d7 d1d5 d8e7 d5e4 e7e4 c3e4 d7b5 e4c5 b7b6 c5b3 b8c6 g1f3 g8e7 b3d4 c6d4 f3d4 b5d7 c1e3 e8c8 e1c1 e7f5 h1e1 h8e8 b2b3 h7h5 a2a4 c8b7 h2h3 f7f6 d4b5 d7c6 d1d8 e8d8 g2g4 h5g4 h3g4 f5e3 e1e3 d8d7 b5c3');
INSERT INTO evaluations VALUES(48374,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(48374,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(48374,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(48374,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(48374,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(48374,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(48374,36,'cp -132','f6e4 c3e4 d5e4 f3g5 f7f5 f1c4 b8c6 c2c3 d8f6 d1b3 c6d8 g5h3 h7h6 c4b5 c7c6 b5e2 d8f7 e1g1 f8d6 f2f3 e4f3 e2f3 e6e5 c1e3 e8g8 a1e1 g7g5 d4e5 d6e5 f3h5 b7b6 h3f2 c8e6 b3c2 c6c5 f2d3 e5d6 e3d2 e6c4 b2b3');
INSERT INTO evaluations VALUES(29577,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29577,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(29577,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(29577,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(29577,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(29577,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29577,36,'cp 5','e7e6 d3f5 e6f5 d1d3 d8c8 b2b3 b8a6 e1g1 c7c6 c2c4 g7g6 c1a3 a6b4 d3d2 a7a5 a3b2 c8d8 b1c3 f8g7 d2e2 e8g8 f3e5 f8e8 c3a4 a8b8 a2a3 b4a6 a1d1 h7h5 e2c2 a6c7 a4c5 f6e4 c5d3 g6g5 a3a4');
INSERT INTO evaluations VALUES(30124,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30124,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(30124,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(30124,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(30124,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30124,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(30124,36,'cp 12','d3f5 e6f5 d1d3 d8c8 b2b3 b8a6 e1g1 c7c6 c2c4 g7g6 f3e5 c8e6 c1a3 f8a3 b1a3 e8g8 d3c3 a6c7 a1c1 a7a5 a3b1 a5a4 c3b4 a4b3 a2b3 f8b8 b1c3 f6e4 h2h3 e4d6 c1a1 e6e7 a1a8 b8a8 e5d3 g8g7');
INSERT INTO evaluations VALUES(43017,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(43017,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(43017,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43017,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(43017,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(43017,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(43017,36,'cp -5','b8c6 e5f4 d7d5 e4e5 g7g5 f4a4 f6d7 f1b5 c6e5 d2d4 e5g6 a4a5 c7c6 a5d8 e8d8 b5d3 d7f8 c3e2 f8e6 g1f3 d8c7 c1e3 a7a5 a2a4 c8d7 e1g1 e6f4 e2f4 g6f4 e3f4 g5f4 a1e1 e7d6 f3h4 a8e8 e1e8 h8e8 h4f5 f4f3 g2g3 b7b5 a4b5 c6b5 c2c3 a5a4 f5d6');
INSERT INTO evaluations VALUES(10869,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(10869,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(10869,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(10869,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(10869,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(10869,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(10869,36,'cp -4','e5f4 d7d5 e4e5 g7g5 f4a4 f6d7 f1b5 c6e5 d2d4 e5g6 a4a5 c7c6 a5d8 e7d8 b5d3 d7f8 c3e2 f8e6 c1d2 d8b6 e1c1 c8d7 c2c3 b6c7 g1f3 g5g4 f3e1 h7h5 h2h4 e8c8 g2g3 e6g7 e1g2 d7f5 e2f4 c8d7 h1e1 c7f4 g2f4 f5d3 f4d3');
INSERT INTO evaluations VALUES(53047,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53047,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(53047,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(53047,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(53047,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53047,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(53047,36,'cp 17','c5d4 d1d4 b8c6 d4d2 g7g6 b2b3 f8h6 f2f4 f7f5 c1b2 g8f6 f1d3 d8a5 e4f5 c8f5 g1e2 e8c8 e1c1 d6d5 c1b1 d5d4 c3b5 a5d2 d1d2 f5g4 b5d4 c6d4 b2d4 d8d4 e2d4 h6f4 d2e2 g4e2 d3e2 f6e4 b1b2 f4h6 e2f3 h6g7 f3e4 g7d4 c2c3');
INSERT INTO evaluations VALUES(53134,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53134,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53134,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53134,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(53134,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53134,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(53134,36,'cp 8','d1d4 b8c6 d4d3 g7g6 g1f3 f8g7 f1e2 g8f6 e1g1 e8g8 f1d1 c8g4 c3d5 g4f3 d3f3 f6d5 e4d5 c6d4 f3d3 d4e2 d3e2 b7b5 e2b5 a8b8 b5a4 g7b2 c1b2 b8b2 a4a7 b2c2 a2a4 c2c7 a7a5 c7c5 a5d8 f8d8 a4a5 d8a8 a5a6 a8a7 a1a4 g8f8 f2f3 c5b5 g1f2 b5b2 f2g3 b2b5');
INSERT INTO evaluations VALUES(53131,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53131,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53131,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53131,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53131,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(53131,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53131,36,'cp 10','b8c6 d4d2 e7e6 b2b3 g8f6 c1b2 d6d5 e4d5 e6d5 e1c1 c8e6 c1b1 d8a5 c3b5 a5d2 d1d2 f8b4 d2d3 e8g8 b2f6 g7f6 g1e2 a8c8 e2f4 f8d8 b5d4 c6d4 d3d4 b4c5 d4d2 c5b4 d2d1 d5d4 f1d3 b4c3 f4e2 f6f5 e2c3 d4c3 b1c1 g8g7 g2g3 c8c5 d3e2 e6d5 d1d3 d5h1 d3d8');
INSERT INTO evaluations VALUES(20083,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(20083,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(20083,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(20083,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(20083,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(20083,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(20083,36,'cp 14','d4e3 g7g6 h2h3 g8f6 f1c4 f8g7 g1f3 e8g8 e1g1 c6a5 c4d3 a7a6 f1d1 d8c7 a2a4 f8e8 e3f4 a5c4 a1b1 c8d7 b2b3 c4e5 f3e5 d6e5 f4d2 d7e6 c1b2 e8d8 d2e3 h7h5 d3f1 a8c8 d1d8 c7d8 b1d1 d8c7 f1e2 h5h4 e3d3');
INSERT INTO evaluations VALUES(20089,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20089,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20089,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20089,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20089,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(20089,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20089,36,'cp 0','g7g6 f1c4 g8f6 g1f3 f8g7 e1g1 e8g8 h2h3 b7b6 e3e2 c8b7 c1f4 d8c7 a1d1 a8c8 c3b5 c7b8 b5c3');
INSERT INTO evaluations VALUES(34211,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(34211,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(34211,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(34211,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(34211,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(34211,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(34211,36,'cp 101','h7h6 d3d4 e5d4 d1d4 f6d4 f3d4 g8f6 c1f4 d7d6 e1c1 g7g5 f4g3 f6h5 f1e2 h5g3 h2g3 a7a5 h1e1 f8e7 f2f4 e8g8 c3a4 e7d8 e4e5 d6e5 f4e5 d8c7 e2h5 f8e8 e5e6 c8e6 d4e6 e8e6 e1e6 f7e6 a4c5 c7b6 c5e6 b8a6 d1d7 a6c5 e6c5 b6c5 d7b7');
INSERT INTO evaluations VALUES(34187,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(34187,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(34187,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(34187,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(34187,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(34187,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(34187,36,'cp 107','c1e3 f8b4 d3d4 d7d6 d1d3 g8e7 f3d2 e8g8 e1c1 b8d7 a2a3 b4a5 d2b3 a5b6 d3d2 e5d4 e3d4 b6d4 d2d4 d6d5 e4d5 e7d5 c3d5 c6d5 f2f3 f6d4 d1d4 d7f6 d4b4 f8e8 f1b5 e8e7 h1d1 a7a5 b4d4 g7g6 d4d2 c8d7 b3d4 a8e8 b5d7 e7d7');
INSERT INTO evaluations VALUES(34188,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34188,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(34188,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(34188,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(34188,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(34188,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(34188,36,'cp 82','f8b4 d3d4 d7d6 d1d3 b8d7 e1c1 f6e7 c3e2 g8f6 e2g3 g7g6 c2c3 b4a5 h2h3 b7b5 d3d2 h6h5 f1d3 c8b7 e3g5 e8c8 d3c2 d8e8 h1e1 c6c5 b2b4 c5b4 c3b4 a5d8 c1b2 c8b8 c2b3 d8b6 d4e5 d7e5 d2f4 b6f2 f3e5 e7e5 f4e5 d6e5 g5f6 f2g3 f6h8 e8h8');
INSERT INTO evaluations VALUES(32976,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(32976,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(32976,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(32976,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(32976,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(32976,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(32976,36,'cp 115','d3d4 d7d6 d1d3 g8e7 e1c1 e8g8 f3d2 b8d7 a2a3 b4a5 d2b3 a5b6 d3d2 d6d5 g2g4 e5d4 b3d4 d5e4 g4g5 h6g5 c3e4 f6e5 e4g5 d7f6 d4f3 b6e3 f2e3 e5c5 e3e4 c8g4 e4e5 g4f3 g5f3 f6g4 d1e1 e7g6 e5e6 f7e6 d2g5 c5g5 f3g5 g4e5 f1h3');
INSERT INTO evaluations VALUES(47085,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(47085,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(47085,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(47085,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(47085,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(47085,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(47085,36,'cp -28','c7c6 b5d3 c6c5 d4c5 e7e6 g1f3 f8c5 e1g1 e8g8 c2c4 d5c4 d3c4 a7a6 d1d8 f8d8 b2b3 b8d7 b1d2 b7b6 c4e2 c8b7 c1b2 a8c8 a1c1 c5f8 f1d1 b6b5 a2a3 f6d5 a3a4 d7c5 a4b5 a6b5 b2d4 b7a6 c1c2');
INSERT INTO evaluations VALUES(44851,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(44851,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(44851,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(44851,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(44851,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(44851,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(44851,36,'cp -11','b5d3 c6c5 d4c5 e7e6 g1f3 f8c5 e1g1 e8g8 c2c4 d5c4 d3c4 a7a6 d1d8 f8d8 c1d2 b7b6 f1d1 c8b7 c4e2 c5e7 f3d4 b8c6 d4c6 b7c6 b1c3 b6b5 a2a3 a8c8 a1c1 f6d7 c3a2 c6d5 c1c8 d8c8 d2b4 e7f6 a2c3 d7b6 c3d5');
INSERT INTO evaluations VALUES(44949,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(44949,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(44949,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(44949,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(44949,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(44949,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(44949,36,'cp -14','c6c5 d4c5 e7e6 g1f3 f8c5 e1g1 e8g8 c2c4 d5c4 d3c4 a7a6 d1d8 f8d8 c1d2 b7b6 f1d1 c8b7 c4e2 b8d7 d2e1 a8c8 b1d2 c5f8 d2b3 c8c2 f3d4 c2c7 a1c1 d8c8 a2a3 g7g6 h2h3 e6e5 d4f3 c7c1 d1c1 c8c1 b3c1 f6d5 f3d2 f7f5');
INSERT INTO evaluations VALUES(25529,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(25529,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(25529,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(25529,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(25529,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(25529,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(25529,36,'cp 89','b1c3 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 d8d7 h2h3 a7a5 c1e3 a5a4 a1d1 d7e6 f1e1 f8b8 b2b3 a4b3 a2b3 b8e8 e3f2 a8a5 g2g4 h7h6 g1g2 e7f8 e1e2 e8a8 f2d4 f6d7 h3h4 c6c5 d4f2 a5a1 d1a1 a8a1 g4g5 h6g5 h4g5 c7c6');
INSERT INTO evaluations VALUES(36690,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(36690,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(36690,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(36690,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(36690,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(36690,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(36690,36,'cp 59','f1d3 b8c6 e1g1 e7e5 d4d5 c6d4 f4f5 c7c6 c1g5 h7h6 g5f6 g7f6 f3d2 d4b5 c3b5 c6b5 f1f2 c8d7 d1e2 a7a6 g2g3 g8g7 g1g2 f8h8 a1h1 g6g5 c2c3 h6h5 h2h3 h5h4 g3g4 f6e7 d2f1 f7f6 f1e3 h8h7 h1b1 d7e8 c3c4 b5c4 d3c4');
INSERT INTO evaluations VALUES(36691,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(36691,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(36691,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(36691,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(36691,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(36691,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(36691,36,'cp 59','b8c6 e1g1 e7e5 d4d5 c6d4 f4f5 c7c6 c1g5 h7h6 g5f6 g7f6 g1h1 g8g7 a2a4 c8d7 f3d2 f6g5 d1e1 h6h5 d2f3 d4f3 f1f3 h5h4 a4a5 a7a6 d5c6 d7c6 d3c4 g5f4 c3d5 d8g5 d5f4 e5f4 e1f2 g6f5 f3f4 c6e4 f4h4 d6d5 c4d3 a8e8');
INSERT INTO evaluations VALUES(3516,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(3516,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(3516,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(3516,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(3516,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(3516,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(3516,36,'cp 78','e1g1 e7e5 d4d5 c6d4 f4f5 c7c6 c1g5 h7h6 g5f6 g7f6 g1h1 c8d7 a2a4 g8g7 f3d2 f6g5 d1e1 h6h5 d2f3 g5f6 a4a5 a8c8 e1f2 g6g5 f3d2 g5g4 d2c4 d8c7 d5c6 b7c6 c4e3 c8e8 d3c4 e8b8 c3e2 d4e2 f2e2');
INSERT INTO evaluations VALUES(41806,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41806,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41806,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41806,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41806,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(41806,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41806,36,'cp 0','e8g8 g1f3 c7c5 b1d2 b7b6 f1d3 c8b7 d1e2 f6e4 f3e5 e4d2 e2d2 b8c6 d2e2 d8c7 e5c6 b7c6 e2h5 f7f5 a2a4 a7a5 h5g5 a8e8 e1g1 d6g3 h2g3 g7g6 g3g4 f5g4 g5g4 e6e5 d4c5 b6c5 d3g6 h7g6 g4g6 g8h8 g6h6 h8g8');
INSERT INTO evaluations VALUES(36962,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36962,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36962,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36962,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36962,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36962,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36962,36,'cp 0','g1f3 b7b6 b1d2 c7c5 f1d3 c8b7 d1e2 f6e4 f3e5 e4d2 e2d2 b8c6 d2e2 d8c7 e5c6 b7c6 e2h5 f7f5 a2a4 a7a5 h5g5 a8e8 e1g1 d6g3 h2g3 g7g6 g3g4 f5g4 g5g4 e6e5 d4c5 b6c5 d3g6 h7g6 g4g6 g8h8 g6h6 h8g8');
INSERT INTO evaluations VALUES(36963,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36963,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36963,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36963,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36963,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36963,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36963,36,'cp 4','c7c5 b1d2 d8c7 d4c5 c7c5 c3c4 b8c6 c4d5 d6g3 h2g3 e6d5 a2a3 c5e7 f1e2 c8g4 e1g1 a8c8 d2b3 h7h5 a1c1 g7g6 b3d4 f6e4 b2b4 a7a5 d4c6 c8c6 d1d4 a5b4 a3b4 f8c8 c1c6 b7c6 e2d3 c6c5 b4c5 e7c5 d3e4 d5e4');
INSERT INTO evaluations VALUES(36143,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36143,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(36143,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(36143,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36143,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36143,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36143,36,'cp 9','b1d2 d8c7 d4c5 c7c5 c3c4 b8c6 c4d5 d6g3 h2g3 e6d5 a2a3 c8g4 f1e2 a8c8 e1g1 c8c7 a1c1 c5e7 d1a4 f8c8 e2d3 a7a6 f1e1 g7g6 c1c2 g8g7 e1c1 g4f3 d2f3 c6e5 c2c7 c8c7 c1c7 e7c7 a4c2 c7c2 d3c2');
INSERT INTO evaluations VALUES(36142,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(36142,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(36142,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36142,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36142,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36142,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36142,36,'cp 0','d8c7 d4c5 c7c5 c3c4 d6g3 h2g3 b8c6 c4d5 e6d5 a2a3 f8d8 f1e2 c8g4 a1c1 c5e7 d2b3 a8c8 e1g1 g7g6 b3d4 f6e4 b2b4 e4g3 f2g3 e7e3 g1h1 c6d4 c1c8 d8c8 f3d4 e3h6 h1g1 h6e3');
INSERT INTO evaluations VALUES(53051,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(53051,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(53051,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(53051,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(53051,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(53051,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(53051,36,'cp -13','g7g6 d2d4 g8f6 b1c3 f8g7 d4c5 d8a5 d1d2 a5c5 c1b2 e8g8 f3d4 c8d7 e1c1 b8c6 f2f3 f8c8 c1b1 c6d4 d2d4 b7b5 d4c5 c8c5 c3d5 f6d5 d1d5 c5d5 e4d5 h7h5 f1d3 g7b2 b1b2 g8g7 h1e1 g7f6 h2h4 a8c8 c2c4 b5b4 g2g3 d7f5 d3f5 g6f5');
INSERT INTO evaluations VALUES(48595,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(48595,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(48595,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(48595,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(48595,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(48595,35,'cp -128',NULL);
INSERT INTO evaluations VALUES(48595,36,'cp -121','b7b5 c1g5 f8c5 c3b5 c5f2 e1f2 f6e4 f2e1 e4g5 d1e2 e8g8 h2h4 c7c6 b5c3 c6d5 h4g5 b8c6 g2g4 d8g5 c3d5 c8g4 e2e4 g4f5 e4g2 g5g2 f1g2 f5d3 d5f6 g7f6 g2c6 a8c8 c6a4 f6f5 a1c1 g8g7 g1f3 e5e4 f3d4 f5f4 h1h4 f8d8 d4c6 d8d6 h4f4 d6c6 a4c6');
INSERT INTO evaluations VALUES(43812,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(43812,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(43812,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(43812,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(43812,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(43812,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(43812,36,'cp -86','c1g5 f8c5 c3b5 c5f2 e1f2 f6e4 f2e1 e4g5 d1e2 e8g8 h2h4 c7c6 b5c3 c6d5 h4g5 b8c6 g2g4 d8g5 g1f3 g5d8 a1d1 c8g4 d1d5 d8f6 e2e4 h7h5 f1e2 c6b4 f3d2 f6f4 e2g4 f4e4 d2e4 h5g4 d5b5 a8b8 e1f1 b8b5 c3b5 f7f5 e4d6 a7a6 b5a3 g7g6 a3c4 e5e4 f1g2 b4d3 c4e3 d3b2 h1c1 b2d3 c1c7 d3f4 g2g3 f4h5 g3g2 f8d8');
INSERT INTO evaluations VALUES(52965,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(52965,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(52965,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(52965,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(52965,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(52965,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(52965,36,'cp 64','e2e4 g8f6 b1d2 g7g6 g1f3 f8g7 h2h3 e8g8 f1e2 b8d7 e1g1 f8e8 a2a4 e7e5 d4e5 d7e5 f3e5 d6e5 f4e3 c8e6 c2c3 d8c7 a4a5 b7b6 d1a4 a8b8 b2b4 b6b5 a4c2 g7f8 f1b1 c6c5 b4c5 f8c5 e3g5 f6d7 e2b5 e8c8 d2f3 f7f6 g5h6 c5f8 h6f8 d7f8');
INSERT INTO evaluations VALUES(55348,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(55348,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(55348,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(55348,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(55348,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(55348,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(55348,36,'cp 16','h2h4 h7h6 h4g5 f8g7 g5h6 h8h6 h1h6 g7h6 d1e2 b8c6 b1c3 c8g4 e2f2 h6g7 d2d3 g4f3 g2f3 g7d4 f2h2 d4c3 b2c3 c6e5 h2f4 e5c4 d3c4 g8e7 c4c5 e7g6 f4g5 d6c5 c1e3 d8g5 e3g5 e8d7 a1d1 d7e6 f3f4 f7f5 d1d5 f5e4 d5c5 a8f8 c5c7 g6f4 c7b7 e4e3 b7a7 f4g2 e1e2 f8f2 e2d3 f2d2 d3e4 d2c2 g5e3 g2e3');
INSERT INTO evaluations VALUES(55346,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(55346,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(55346,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(55346,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(55346,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(55346,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(55346,36,'cp 9','h7h6 h4g5 f8g7 g5h6 h8h6 h1h6 g7h6 b1c3 b8c6 d1e2 c8g4 e2f2 h6g7 d2d3 g4f3 g2f3 g7d4 f2h2 d4c3 b2c3 d8f6 c1d2 c6e5 h2f4 e5c4 d3c4 e8c8 f4g4 f6e6 g4e6 f7e6 c4c5 d6c5 e1e2 g8f6 d2g5 d8f8 a1h1 e6e5 g5f6 f8f6 h1h5 f6e6 e2e3 b7b6 c3c4 a7a6 f3f4 e5f4 e3f4 e6g6 f4f5');
INSERT INTO evaluations VALUES(9951,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9951,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9951,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9951,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(9951,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9951,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9951,36,'cp 0','d7c6 d2d3 c8g4 h2h3 g4h5 b1d2 f6d7 d2c4 f7f6 c2c3 e8g8 a2a4 c5e7 c1e3 d7c5 e3c5 e7c5 b2b4 c5e7 c4e3 f8e8 a4a5 e7f8 d1b3 h5f7 b3c2 f7h5');
INSERT INTO evaluations VALUES(33236,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(33236,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(33236,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(33236,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(33236,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(33236,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(33236,36,'cp 56','b1c3 g8e7 c1e3 c5e3 f2e3 e8g8 d1d2 d7d6 e1c1 a7a6 f1e2 b7b5 d1f1 b8d7 g2g4 c8b7 h2h4 b5b4 c3d1 a6a5 c1b1 f6e6 d1f2 f7f6 h1g1 d7c5 g4g5 f6f5 e4f5 f8f5 e3e4 f5f4 f3d4 e6f7 f2h3 f4f1 g1f1 f7e8 d4f5 e7f5 e4f5');
INSERT INTO evaluations VALUES(33232,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(33232,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(33232,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(33232,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(33232,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(33232,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(33232,36,'cp 42','g8e7 c1e3 d7d6 e3g5 f6g6 g5e7 e8e7 c3d5 e7d8 b2b4 c5b6 a2a4 c7c6 d5b6 a7b6 f1e2 c8e6 e1g1 d8e7 d1d2 b8d7 f3g5 h7h6 g5e6 f7e6 a4a5 b6a5 a1a5 b7b6 a5a1 h8f8 d2c3 c6c5 b4c5 b6c5 a1a5 g6f6 c3a3 a8a5 a3a5 f8b8 f1a1 e7f8 a5a7');
INSERT INTO evaluations VALUES(13431,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(13431,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(13431,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(13431,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(13431,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(13431,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(13431,36,'cp 133','c1g5 c8d7 e1c1 d8a5 f2f4 h7h6 g5h4 g7g5 h4e1 a5c5 d4c5 d6c5 e4e5 f6g4 c3b5 e8c8 b5a7 c8b8 e1a5 f8g7 d1d2 g5f4 a5d8 h8d8 f1b5 g4e5 b5d7 b8a7 d7b5 d8g8 h1f1 h6h5 c1b1 h5h4 a2a4 g7f6 c2c3 h4h3 g2h3 f4f3 d2d5 g8g2 d5c5 g2h2');
INSERT INTO evaluations VALUES(13426,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(13426,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(13426,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(13426,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(13426,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(13426,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(13426,36,'cp 136','c8d7 e1c1 d8a5 f2f4 h7h6 g5h4 g7g5 h4e1 g5f4 c3d5 a5a2 d5c7 e8d8 c7a8 d7g4 b2b3 a2a3 c1d2 g4d1 d2d1 e7e5 d4c4 a3c5 g2g3 d6d5 e4d5 f6g4 c4c5 f8c5 b3b4 c5d6 c2c4 b7b6 f1d3 d8e7 c4c5 b6c5 b4c5 d6c5 a8c7 c5b6 c7b5 f4f3 e1b4 e7f6 d1d2 g4e3 d3e4 f3f2 d2e2 f2f1q h1f1 e3f1');
INSERT INTO evaluations VALUES(15488,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(15488,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(15488,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(15488,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(15488,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(15488,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(15488,36,'cp 96','f3g5 d6d5 e4d5 c6a5 d1a4 f6d7 g5f7 d8h4 f7e5 f8d6 d2d4 a5c4 e5c4 e8g8 c1e3 h4e4 b1d2 e4g2 e1c1 g2d5 c4d6 c7d6 a4b3 d7b6 h1g1 c8f5 e3f4 a8d8 d1e1 f5d3 b3d5 b6d5 f4g3 b7b5 f2f3 d3g6 e1e6 d5f4 g3f4 f8f4 e6e7 a7a5 g1g4 f4f5 h2h4 a5a4');
INSERT INTO evaluations VALUES(15473,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(15473,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(15473,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(15473,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(15473,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(15473,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(15473,36,'cp 97','d6d5 e4d5 c6a5 d1a4 f6d7 g5f7 d8h4 f7e5 f8d6 d2d4 a5c4 e5c4 e8g8 c1e3 h4g4 b1d2 g4g2 e1c1 g2d5 c4d6 c7d6 a4b3 d7b6 h1g1 c8f5 e3f4 a8d8 f4g3 f5d3 b3d5 b6d5 d1e1 d3g6 a2a4 d5f4 g3f4 f8f4 f2f3 g8f8 g1g4 f4f5 g4g2 d8e8 e1e8 g6e8 d2e4 f5f3 e4d6 f3f1 c1d2 e8a4 c3c4');
INSERT INTO evaluations VALUES(26963,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(26963,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(26963,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(26963,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(26963,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(26963,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(26963,36,'cp 50','g4e6 c4e6 f7e6 c2c3 g8f6 b2b4 a7a6 e1g1 f8e7 f1e1 e8g8 a2a4 h7h6 b1d2 d8e8 d2f1 e8g6 c1e3 f6d7 f1h2 c6d8 a4a5 d8f7 c3c4 c7c6 b4b5 d6d5 e4d5 e6d5 b5c6 b7c6 d1a4 g6d6 c4d5 c6d5 a1b1 d5d4');
INSERT INTO evaluations VALUES(50288,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50288,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(50288,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50288,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50288,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(50288,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(50288,36,'cp 20','g5f6 e7f6 e2e3 d7d5 c2c4 d5c4 f1c4 f8d6 g1f3 e8g8 c4b3 c7c5 d4c5 d6c5 b1c3 b8c6 e1g1 c8f5 a2a3 a8c8 h2h3 c5b6 b3a2 b6a5 c3d5 f8e8 f3d4 c6d4 d1d4 a5b6 d5b6 d8d4 e3d4 a7b6');
INSERT INTO evaluations VALUES(50287,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50287,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(50287,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(50287,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50287,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50287,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(50287,36,'cp 12','e7f6 c2c4 f8b4 b1d2 d7d5 c4c5 e8g8 a2a3 b4d2 d1d2 b7b6 e2e3 f6f5 g1f3 f5f4 f1d3 b6c5 d4c5 f4e3 f2e3 a7a5 e1g1 a5a4 d3c2 c8a6 f1d1 a6c4 e3e4 d8e7 e4d5 e7c5 g1h1 b8d7 a1c1 c5a5 c2h7 g8h7 d2c2 h7g8 c2c4 d7f6 c4c7 f6d5 c7a5 a8a5');
INSERT INTO evaluations VALUES(49273,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(49273,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(49273,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(49273,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(49273,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(49273,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(49273,36,'cp 25','c2c4 f8b4 b1d2 d7d5 c4c5 e8g8 a2a3 b4d2 d1d2 b7b6 e2e3 f6f5 g1e2 b8d7 a1c1 d7f6 f2f3 f6h5 g2g3 d8e7 e1f2 f8e8 c1c3 g7g5 h2h4 g5g4 e2f4 h5f4 e3f4 b6c5 c3c5 a8b8 b2b4 g4f3 f1d3 e7f6 c5d5 f6b6 h4h5 c8a6 d3f5');
INSERT INTO evaluations VALUES(15496,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(15496,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(15496,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(15496,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(15496,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(15496,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15496,36,'cp 0','c6a5 c4b3 a5b3 a2b3 f8e7 h2h3 b7b5 d3d4 c8b7 d1e2 a7a6 b1c3 e5d4 f3d4 e8g8 e2f3 f8e8 c3d5 f6d5 e4d5 e7f6 c1e3 d8e7 c2c4 e7e4 f3e4 e8e4 f1d1 g8f8 a1c1 a8c8 c1a1');
INSERT INTO evaluations VALUES(21638,30,'cp -191',NULL);
INSERT INTO evaluations VALUES(21638,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(21638,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(21638,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(21638,34,'cp -168',NULL);
INSERT INTO evaluations VALUES(21638,35,'cp -195',NULL);
INSERT INTO evaluations VALUES(21638,36,'cp -190','g7g5 h2h4 g5g4 f3g1 f8h6 b1c3 f7f6 g1e2 f6e5 d4e5 f4f3 e2g3 d8e7 c1h6 g8h6 d1d5 h6f5 g3f5 c8f5 f1b5 a8d8 b5c6 e8f8 d5a5 b7c6 g2f3 g4f3 a1d1 d8e8 a5a3 e7a3 b2a3 e8e5 e1f2 e5c5 d1d8 f8g7 h1g1 f5g6 c3e4 h8d8 e4c5 d8d2 f2f3 d2c2 c5d7 g7h6 g1e1 h6h5 f3g3 c2c3 g3f4 c3a3 e1e7');
INSERT INTO evaluations VALUES(21245,30,'cp -223',NULL);
INSERT INTO evaluations VALUES(21245,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(21245,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(21245,33,'cp -152',NULL);
INSERT INTO evaluations VALUES(21245,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(21245,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(21245,36,'cp -190','h2h4 g5g4 f3g1 f8h6 b1c3 f7f6 g1e2 f6e5 d4e5 f4f3 e2g3 d8e7 c1h6 g8h6 d1d2 e7e5 e1f2 e8g8 a1e1 e5d4 d2d4 c6d4 c3d5 c8e6 d5c7 a8c8 e1e6 f3g2 f2g2 c8c7 e6h6 c7c2 f1e2 d4e2 g3e2 c2e2 g2g3 e2e4 h1d1 f8f3 g3g2 e4e2 g2g1 f3g3 g1h1 e2b2 h6h5 g3e3 h5g5 g8f7 g5g4 f7e6 d1f1 b2a2 g4g7 e3h3 h1g1 h3h4 g7b7 h4g4 g1h1');
INSERT INTO evaluations VALUES(22951,30,'cp -763',NULL);
INSERT INTO evaluations VALUES(22951,31,'cp -766',NULL);
INSERT INTO evaluations VALUES(22951,32,'cp -759',NULL);
INSERT INTO evaluations VALUES(22951,33,'cp -759',NULL);
INSERT INTO evaluations VALUES(22951,34,'cp -759',NULL);
INSERT INTO evaluations VALUES(22951,35,'cp -755',NULL);
INSERT INTO evaluations VALUES(22951,36,'cp -755','f8b4 c1d2 b4c3 b1c3 g8e7 e1c1 e8g8 g1e2 f8e8 c1b1 a7a6 d2g5 b7b5 e2d4 c6d4 d1d4 f7f6 g5f4 d7d6 f2f3 c8e6 c3d5 e7g6 f4g3 g6e5 d4d1 c7c6 g3e5 d6e5 d5e3 d8a5 a2a3 a8a7 b1c1 a5b6 e3f5 a6a5 h2h4 b6c5 f5d6');
INSERT INTO evaluations VALUES(6673,30,'cp -725',NULL);
INSERT INTO evaluations VALUES(6673,31,'cp -736',NULL);
INSERT INTO evaluations VALUES(6673,32,'cp -736',NULL);
INSERT INTO evaluations VALUES(6673,33,'cp -725',NULL);
INSERT INTO evaluations VALUES(6673,34,'cp -725',NULL);
INSERT INTO evaluations VALUES(6673,35,'cp -721',NULL);
INSERT INTO evaluations VALUES(6673,36,'cp -721','c1d2 b4c3 b1c3 g8e7 h2h4 e8g8 e1c1 d7d6 g1e2 c6e5 d2f4 f7f6 c3d5 e7d5 e4d5 d8e7 c1b1 a7a6 f2f3 a8b8 h4h5 h7h6 e2d4 e7f7 c2c4 c8d7 f1e2 f8e8 f4g3 b7b5 g3e5 f6e5 d4c6 d7c6 d5c6 b5c4 b1a1 f7f4 a1b1');
INSERT INTO evaluations VALUES(6674,30,'cp -740',NULL);
INSERT INTO evaluations VALUES(6674,31,'cp -748',NULL);
INSERT INTO evaluations VALUES(6674,32,'cp -748',NULL);
INSERT INTO evaluations VALUES(6674,33,'cp -751',NULL);
INSERT INTO evaluations VALUES(6674,34,'cp -751',NULL);
INSERT INTO evaluations VALUES(6674,35,'cp -755',NULL);
INSERT INTO evaluations VALUES(6674,36,'cp -751','b4c3 b1c3 g8e7 e1c1 e8g8 c1b1 a8b8 g1e2 b7b5 e2f4 b5b4 c3d5 a7a5 d2c1 a5a4 f2f3 a4a3 b2b3 c6e5 d5e7 d8e7 f4d5 e7d8 d1d4 c7c5 c1f4 c5d4 f4e5 d7d6 e5d4 f7f5 f1d3 f5e4 d3e4 c8e6 d4e3 h7h6 e3f4 d8d7 h1d1');
INSERT INTO evaluations VALUES(6718,30,'cp -728',NULL);
INSERT INTO evaluations VALUES(6718,31,'cp -736',NULL);
INSERT INTO evaluations VALUES(6718,32,'cp -732',NULL);
INSERT INTO evaluations VALUES(6718,33,'cp -728',NULL);
INSERT INTO evaluations VALUES(6718,34,'cp -721',NULL);
INSERT INTO evaluations VALUES(6718,35,'cp -721',NULL);
INSERT INTO evaluations VALUES(6718,36,'cp -728','b1c3 g8e7 h2h4 e8g8 e1c1 d7d6 c1b1 h7h6 g1e2 f7f5 e4f5 c8f5 e2f4 c6e5 d2c1 d8d7 f2f3 a8e8 f1e2 e7g6 f4g6 f5g6 h4h5 g6f5 c3b5 f5h7 b5d4 d7f7 g2g4 a7a6 g4g5 h6g5 c1g5');
INSERT INTO evaluations VALUES(56991,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(56991,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(56991,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(56991,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(56991,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(56991,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56991,36,'cp -19','a7a6 b5e2 d7d5 e4d5 e6d5 d2d4 f8d6 g1f3 g8f6 c1g5 e8g8 e1g1 h7h6 g5h4 f8e8 f1e1 c8f5 c2c4 d5c4 e2c4 e8e1 d1e1 b8d7 b1c3 d8f8 e1d2 a8e8 a1e1 c7c6 h2h3 d6c7 e1e8 f8e8 d4d5 c6d5 c3d5 f6d5 d2d5 f5e4 d5d2 e4f3 g2f3');
INSERT INTO evaluations VALUES(51020,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(51020,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(51020,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(51020,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(51020,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(51020,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(51020,36,'cp -18','b5e2 d7d5 e4d5 e6d5 d2d4 f8d6 g1f3 g8f6 e1g1 e8g8 c1g5 f8e8 f1e1 h7h6 g5h4 c8f5 c2c4 d5c4 e2c4 e8e1 d1e1 b8d7 b1c3 d8f8 e1d2 a8e8 a1e1 e8e1 d2e1 f8e8 e1d2 c7c6 h4g3 f5e6 c4d3 d6e7 h2h3 d7b6 d3b1 e8c8 d2c2 b6d5 c3d5');
INSERT INTO evaluations VALUES(36382,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(36382,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36382,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(36382,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(36382,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(36382,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(36382,36,'cp 25','f1d3 h7h6 g5h4 c7c6 d1c2 f8e8 g1e2 a7a5 e1g1 b8d7 c3a4 d7b6 a4b6 d8b6 e2c3 b6d8 h2h3 c8d7 f1e1 b7b6 h4g3 b6b5 a2a3 a8c8 a1c1 d7e6 g3h4 f6d7 h4e7 d8e7 a3a4 b5b4 c3e2 c6c5 d4c5 d7c5 d3b5 e8d8 e2d4 e6d7 b5d7 e7d7');
INSERT INTO evaluations VALUES(36393,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36393,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36393,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(36393,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(36393,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(36393,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36393,36,'cp 24','h7h6 g5h4 f8e8 g1e2 c7c6 d1c2 a7a5 e1g1 b8a6 a2a3 a6c7 h4g3 c7e6 h2h3 e7f8 g3e5 b7b5 f2f4 f6d7 e2g3 f7f6 c3e2 f6e5 f4e5 d7b6 g3h5 g8h8 e2g3 d8g5 c2f2 e8d8 h3h4 g5h4 f2f7 h4g5 g3f5 b6c4');
INSERT INTO evaluations VALUES(36205,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36205,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(36205,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(36205,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(36205,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(36205,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36205,36,'cp 32','g5h4 c7c6 g1e2 f8e8 e1g1 b8d7 a1c1 b7b5 c3b1 c8b7 a2a4 a7a6 b1d2 d8b6 d1c2 a6a5 h4f6 d7f6 d2b3 e7d6 b3c5 b7c8 a4b5 c6b5 e2c3 d6c5 d4c5 b6c5 d3b5 e8d8 f1d1 c8b7 d1d4 a8c8 b5f1 c5b6 h2h3 c8c7 b2b3 d8e8');
INSERT INTO evaluations VALUES(55510,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(55510,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55510,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55510,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55510,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(55510,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(55510,36,'cp -3','d6e5 d1d8 e8d8 b1c3 c8e6 b2b3 h7h5 g2g3 h5h4 f1h3 h4g3 h2g3 b8d7 e1f1 f8b4 c1b2 g8f6 f1g2 d8e7 h3e6 b4c3 h1h8 a8h8 b2c3 e7e6 f2f3 f6e8 g1h3 e8d6 a1h1 h8a8 h3f2 b7b6 f3f4 a7a5 a2a4 d6b7 h1h7 a8g8 g2f3 d7c5');
INSERT INTO evaluations VALUES(56544,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(56544,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(56544,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(56544,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(56544,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(56544,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(56544,36,'cp -5','d1d8 e8d8 b1c3 c8e6 b2b3 h7h5 g2g3 h5h4 f1h3 h4g3 h2g3 b8d7 e1f1 g8f6 f1g2 f8b4 c1b2 b4c3 b2c3 d8e7 h3e6 e7e6 h1h8 a8h8 f2f3 f6e8 g1h3 f7f6 a1h1 e6e7 c3b4 c7c5 b4e1 e8c7 g3g4 b7b5 c4b5 c7b5 h3f2 h8h1 f2h1 d7b6 h1g3 e7e6');
INSERT INTO evaluations VALUES(35406,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(35406,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(35406,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(35406,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(35406,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(35406,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(35406,36,'cp -6','e8d8 b1c3 c8e6 b2b3 h7h5 g2g3 h5h4 f1h3 h4g3 h2g3 g8f6 e1f1 f8b4 c1b2 b8d7 f1g2 d8e7 h3e6 b4c3 h1h8 a8h8 b2c3 e7e6 f2f3 f6e8 g1h3 e8d6 a1h1 e6e7 g3g4 f7f6 h3f4 h8h1 f4d5 e7d8 g2h1 c7c6 d5e3 g7g6 h1g2 b7b6 e3d1 d7c5 g2f2 a7a5 f2e3 a5a4 c3b4 a4b3');
INSERT INTO evaluations VALUES(35457,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35457,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35457,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35457,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35457,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35457,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35457,36,'cp -7','b1c3 c8e6 b2b3 h7h5 g2g3 h5h4 f1h3 h4g3 h2g3 b8d7 e1f1 f8b4 c1b2 g8f6 f1g2 b4c3 b2c3 d8e7 h3e6 e7e6 h1h8 a8h8 f2f3 f6e8 g1h3 f7f6 a1h1 e6e7 c3b4 c7c5 b4e1 d7f8 h3f4 h8h1 f4d5 e7d7 g2h1 f8e6 h1g2 e6d4 g2f2 b7b6 f2e3 f6f5 g3g4 f5g4 f3g4 e8d6');
INSERT INTO evaluations VALUES(35455,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(35455,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(35455,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(35455,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(35455,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(35455,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(35455,36,'cp -6','c8e6 b2b3 h7h5 g2g3 h5h4 f1h3 h4g3 h2g3 b8d7 e1f1 g8f6 f1g2 f8b4 c1b2 b4c3 b2c3 d8e7 h3e6 e7e6 h1h8 a8h8 f2f3 f6e8 g1h3 f7f6 a1h1 e6e7 c3b4 c7c5 b4e1 e8d6 h3f4 h8h1 f4d5 e7f7 g2h1 b7b5 c4b5 d6b5 h1g2 f7e6 d5c3 b5c3 e1c3 d7b6 g2f2 b6d5 c3d2 d5b4 d2b4');
INSERT INTO evaluations VALUES(27730,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(27730,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(27730,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(27730,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(27730,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27730,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(27730,36,'cp -6','b2b3 h7h5 g2g3 h5h4 f1h3 h4g3 h2g3 b8d7 e1f1 f8b4 c1b2 g8f6 f1g2 d8e7 h3e6 b4c3 b2c3 e7e6 h1h8 a8h8 f2f3 f6e8 g1h3 f7f6 a1h1 e6e7 g3g4 b7b6 c3b4 c7c5 b4e1 d7f8 h3f2 h8h1 f2h1 e8d6 h1g3 g7g6 g2f2 f6f5 g4f5 g6f5 e2e3 e7e6 f2e2 f5f4');
INSERT INTO evaluations VALUES(15985,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15985,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15985,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15985,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15985,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15985,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15985,36,'cp 0','f1c4 f8b4 e1g1 e8g8 d1e2 d8e7 c1d2 c7c6 a2a3 b4c5 c4a2 f8e8 f3g5 e8f8');
INSERT INTO evaluations VALUES(15982,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15982,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15982,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15982,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15982,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15982,36,'cp 0','f8b4 e1g1 e8g8 d1e2 c7c6 a2a4 d8c7 h2h3 a7a5 c3a2 b4e7 a2c3');
INSERT INTO evaluations VALUES(39209,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(39209,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(39209,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(39209,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(39209,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(39209,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(39209,36,'cp -65','d7d6 g1f3 b8c6 d2d3 g8e7 a2a3 a7a5 c1e3 f7f5 d3d4 c6d4 f3d4 f5f4 e3f4 c5d4 f4e3 d4e3 f2e3 c8e6 d1f3 e7g8 f3g3 g8f6 g3g7 h8g8 g7h6 d8e7 e1c1 f6g4 h6h5 e7f7 h5f7 e8f7 d1e1 c7c6 c1d2 a5a4 f1e2 g4f2 h1g1 f7e7');
INSERT INTO evaluations VALUES(38708,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(38708,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(38708,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(38708,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(38708,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(38708,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(38708,36,'cp -66','g1f3 b8c6 a2a3 a7a5 d2d3 g8e7 c1e3 f7f5 d3d4 c6d4 f3d4 f5f4 e3f4 c5d4 f4e3 d4e3 f2e3 e8g8 c4c5 c8e6 c5d6 c7d6 f1e2 e7c8 d1d2 c8b6 e1c1 e6b3 d1f1 f8f1 h1f1 g8h8 d2d3 b3c4 d3d2 c4g8 e2b5 d8c7 c1b1 b6c4 b5c4 c7c4 d2e2 c4e6 c3d5 h7h6 f1c1 e6f7 b1a1');
INSERT INTO evaluations VALUES(38707,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(38707,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(38707,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(38707,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(38707,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(38707,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(38707,36,'cp -80','b8c6 d2d3 g8e7 a2a3 a7a5 c1e3 f7f5 d3d4 c6d4 f3d4 f5f4 e3f4 c5d4 f4e3 d4e3 f2e3 e8g8 c4c5 c8e6 c5d6 c7d6 f1e2 e7c8 d1d2 c8b6 e1c1 e6b3 d1f1 f8f1 h1f1 g8h8 c1b1 b6c4 e2c4 b3c4 f1d1 c4b3 d1c1 a5a4 d2f2 d8g5 c3b5 a8d8 b5d6 g5g6 d6f5 h7h6 h2h3 b7b5');
INSERT INTO evaluations VALUES(40713,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(40713,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(40713,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(40713,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40713,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40713,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(40713,36,'cp 5','e8g8 g1f3 b7b6 f1e2 d5c4 e2c4 c7c5 e1g1 c5d4 f3d4 c8b7 d1e2 f6e4 h4e7 d8e7 c3e4 b7e4 c4d3 e4d3 e2d3 f8c8 f1c1 b8d7 h2h3 d7f6 d3a6 e7d7 a2a4 c8c7 a4a5 c7c1 a1c1 b6a5 a6a5 a8c8 c1c8 d7c8 a5a7 c8c1 g1h2');
INSERT INTO evaluations VALUES(36198,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36198,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36198,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36198,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(36198,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36198,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36198,36,'cp 18','g1f3 b7b6 f1e2 d5c4 e2c4 c7c5 e1g1 c5d4 f3d4 c8b7 a1c1 b8d7 d1e2 a7a6 f1d1 d8e8 a2a3 b6b5 c4a2 d7e5 h4g3 e5c6 d4c6 b7c6 e3e4 a8d8 e4e5 d8d1 c1d1 f6d7 a2b1 e8b8 c3e4 b8a8 e4d6 f8d8 e2c2 d7f8 b2b4 e7d6 e5d6 c6g2');
INSERT INTO evaluations VALUES(42718,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(42718,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(42718,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(42718,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(42718,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42718,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(42718,36,'cp 64','d2c3 d7d6 e5f3 f6e4 f1d3 e4c5 e1g1 e8g8 c1g5 d8d7 d3c4 h7h6 g5e3 b8c6 f1e1 d7f5 c4f1 f5f6 f3d4 c8d7 d1d2 c5e4 d2e2 e4c5 f2f3 c5e6 e2d2 a7a5 b2b3 a5a4 b3b4 c6e5 a1d1 f8e8 c3c4 f6g6 g1h1 e6d4 e3d4');
INSERT INTO evaluations VALUES(42717,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(42717,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(42717,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42717,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(42717,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42717,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(42717,36,'cp 45','d7d6 e5f3 f6e4 f1d3 e4c5 e1g1 c5d3 d1d3 b8c6 c1g5 f7f6 f1e1 c6e7 g5h4 e8g8 c3c4 e7c6 c4c5 d6c5 d3c4 g8h8 a1d1 c8d7 c4c5 f8e8 h4g3 e8e1 f3e1 d8e8 h2h3 b7b6 c5c3 a8c8 e1d3 c6e7 c3a3 e7c6 d3f4 d7f5 c2c4 h7h6 a3c3 c6e5 d1e1 e8d7 b2b3 e5g6 f4d5');
INSERT INTO evaluations VALUES(41915,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41915,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(41915,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(41915,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(41915,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(41915,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(41915,36,'cp 60','e5d3 f6e4 f1e2 e8g8 e1g1 f8e8 f2f3 e4f6 c1g5 c8f5 f1e1 b8c6 g5h4 c6e5 d1d2 f5d3 c2d3 e5g6 h4f2 a7a5 d3d4 a5a4 e2d3 d8d7 c3c4 c7c6 h2h3 h7h6 a1d1 b7b5 c4b5 c6b5 g1h1 f6d5 f2g3 e8e1 d1e1');
INSERT INTO evaluations VALUES(34524,30,'cp 759',NULL);
INSERT INTO evaluations VALUES(34524,31,'cp 765',NULL);
INSERT INTO evaluations VALUES(34524,32,'cp 759',NULL);
INSERT INTO evaluations VALUES(34524,33,'cp 762',NULL);
INSERT INTO evaluations VALUES(34524,34,'cp 766',NULL);
INSERT INTO evaluations VALUES(34524,35,'cp 766',NULL);
INSERT INTO evaluations VALUES(34524,36,'cp 771','f1b5 c8d7 b5c6 b8c6 g1e2 g8f6 e1g1 e8c8 a2a3 g7g6 b2b4 e7e5 d2d3 f8g7 b4b5 c6d4 a3a4 c8b8 c1g5 h7h6 g5d2 h6h5 d2g5 d8e8 f2f3 h5h4 h2h3 d7f5 a4a5 d4e2 d1e2 h8h5 b5b6');
INSERT INTO evaluations VALUES(34515,30,'cp 732',NULL);
INSERT INTO evaluations VALUES(34515,31,'cp 732',NULL);
INSERT INTO evaluations VALUES(34515,32,'cp 744',NULL);
INSERT INTO evaluations VALUES(34515,33,'cp 743',NULL);
INSERT INTO evaluations VALUES(34515,34,'cp 744',NULL);
INSERT INTO evaluations VALUES(34515,35,'cp 744',NULL);
INSERT INTO evaluations VALUES(34515,36,'cp 736','c8d7 b5c6 b8c6 g1e2 e8c8 a2a3 g7g6 e1g1 f8g7 h2h3 g8h6 f1e1 h6f5 d2d3 h7h5 b2b4 c6d4 e2d4 f5d4 c1d2 d7e6 a3a4 g7h6 d2h6 h8h6 d1c1 g6g5 c1g5 h6g6 g5e7 d8g8 c3b5 g6g2 g1h1 d4b5 a4b5 e6d5 c2c4 d5e6 a1a7 g2f2');
INSERT INTO evaluations VALUES(38811,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(38811,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38811,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38811,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(38811,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(38811,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38811,36,'cp 25','b1d2 b4e7 g1f3 g8f6 b2b3 e8g8 c1b2 b7b6 f1e2 a7a5 e1g1 a5a4 b2c3 d5c4 d2c4 f6d5 c3d2 a4b3 d1b3 c8a6 a2a4 c7c5 d4c5 e7c5 f3e5 b8d7 e5d7 d8d7 f1c1 d7e7 e2f3 f8b8 c1d1 b8c8 a1c1 a6b7 c4e5');
INSERT INTO evaluations VALUES(44434,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44434,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44434,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44434,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44434,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44434,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(44434,36,'cp 17','b8c6 d4a4 f6d5 a4e4 d5b6 b1c3 f8g7 c1f4 d7d5 e5d6 g7c3 b2c3 c8f5 d6e7 f5e4 e7d8q a8d8 f4h6 e8e7 a2a4 e4f3 g2f3 d8d5 f1d3 d5a5 d3b5 a7a6 b5c6 b7c6 e1g1 a5a4 a1a4 b6a4 f1a1 a4c3 h6g7 c3e2 g1f1 h8g8 g7e5 f7f6 f1e2 f6e5 a1a6');
INSERT INTO evaluations VALUES(12726,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(12726,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(12726,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(12726,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(12726,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12726,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(12726,36,'cp 12','d4a4 f6d5 a4e4 d5b6 b1c3 f8g7 c1f4 d7d5 e5d6 c8f5 d6e7 g7c3 b2c3 f5e4 e7d8r a8d8 f1b5 e8g8 e1g1 b6d5 b5c6 b7c6 f4d2 c6c5 f1c1 f8e8 c3c4 e4f3 c4d5 f3d5 d2e3 d8c8 c1d1 d5e4 a1c1 e4f5 d1d6 e8e6 d6e6 f5e6 a2a4 f7f6');
INSERT INTO evaluations VALUES(12729,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(12729,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12729,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(12729,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12729,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12729,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(12729,36,'cp 18','f6d5 a4b3 d5c7 f3g5 c7e6 g5e6 f7e6 f2f4 f8h6 b3h3 h6g7 c1d2 d7d6 b1c3 d6e5 f4e5 e8g8 e1c1 d8c7 d2h6 c7e5 f1b5 c6d4 b5d3 e5c5 h6g7 g8g7 h1e1 e6e5 h3g3 f8f4 g3e3 c8f5 d3f5 f4f5 g2g4 f5f7 h2h3 a8f8 e3e5 c5e5 e1e5 d4c6 e5e1 f8d8 d1d8 c6d8 c3e4 h7h5 g4h5 g6h5');
INSERT INTO evaluations VALUES(12748,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(12748,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(12748,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(12748,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(12748,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(12748,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(12748,36,'cp 12','a4b3 d5c7 f3g5 c7e6 g5e6 f7e6 f2f4 f8h6 c1e3 e8g8 g2g3 d7d6 e5d6 d8d6 f1g2 c6a5 b3b5 f8f5 b5e2 e6e5 b1c3 e5f4 a1d1 d6b4 d1d8 g8f7 e1g1 f4e3 f1f5 c8f5 d8a8 b4b2 g3g4 b2c3 g4f5 a5c6 e2f3 c6e5 f5g6 f7g7 f3f8 g7g6 f8f1 c3c2 g2d5 e3e2 a8g8 h6g7 f1g2 g6h5 d5f3 e5f3 g2f3 h5h4 f3f4 h4h5 f4f7 c2g6 f7g6 h5g6');
INSERT INTO evaluations VALUES(12746,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(12746,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(12746,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12746,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(12746,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(12746,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(12746,36,'cp 7','d5c7 f3g5 c7e6 g5e6 f7e6 f2f4 f8h6 b3h3 h6g7 c1d2 d7d6 b1c3 e8g8 e1c1 d6e5 f4e5 c6e5 d2h6 g7h6 h3h6 d8b6 h6g5 f8f5 g5e7 c8d7 f1d3 f5f7 e7g5 e5d3 d1d3 d7c6 h1d1 b6f2 d3d2 f7f5 g5h6 f2f4 h6f4 f5f4 d1e1 g8f7 a2a4 a7a6 b2b3 h7h5 d2e2 f4g4 c3e4 c6e4 e2e4 g4g2 e4e6 g2h2');
INSERT INTO evaluations VALUES(54501,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(54501,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(54501,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(54501,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(54501,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(54501,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(54501,36,'cp -16','f8e7 c4b5 b8c6 e1g1 g8f6 f1e1 c8e6 d2d4 e8g8 d4c5 e7c5 c1e3 c5b6 b1d2 h7h6 c2c3 d5d4 c3d4 c6d4 e3d4 b6d4 a1d1 d4b6 b5c4 e6f5 h2h3 f5c2 d1c1 c2g6 c4b3 g6d3 e2e7 f6d5 e7d8 a8d8 d2c4 b6c5');
INSERT INTO evaluations VALUES(37914,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(37914,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(37914,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(37914,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(37914,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(37914,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(37914,36,'cp -27','c4b5 b8c6 e1g1 g8f6 f1e1 c8e6 d2d4 e8g8 d4c5 e7c5 c1e3 c5b6 c2c3 f8e8 b1d2 a7a6 b5a4 b6e3 e2e3 d5d4 e3d3 d4c3 d3c3 a8c8 c3a3 d8b6 h2h3 h7h6 a4b3 e6b3 a3b3 b6b3 d2b3 c6b4 e1e8 f6e8 a1c1 c8c7 a2a3 b4d3 c1c7 e8c7 b3a5 b7b6');
INSERT INTO evaluations VALUES(37912,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(37912,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(37912,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(37912,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(37912,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(37912,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(37912,36,'cp -14','b8c6 e1g1 g8f6 f1e1 c8e6 d2d4 e8g8 d4c5 e7c5 c1e3 c5b6 b1d2 f8e8 c2c3 a7a6 b5a4 b6e3 e2e3 d5d4 e3d3 d4c3 d3c3 a8c8 c3a3 d8b6 h2h3 h7h6 a4b3 e6b3 a3b3 b6b3 d2b3 c6b4 e1e8 f6e8 a1c1 e8d6 a2a3 c8c1 b3c1 b4c6 g1f1 g8f8 c1b3 d6c4 b3c5 c4b2 c5b7');
INSERT INTO evaluations VALUES(5432,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(5432,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(5432,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(5432,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(5432,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(5432,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(5432,36,'cp -19','e1g1 g8f6 f1e1 c8e6 d2d4 e8g8 d4c5 e7c5 c1e3 c5b6 c2c3 f8e8 b1d2 a7a6 b5a4 b6e3 e2e3 d5d4 e3d3 d4c3 d3c3 a8c8 c3a3 h7h6 h2h3 d8c7 a1d1 e6d5 a4b3 e8e1 d1e1 c8d8 b3d5 f6d5 d2c4 d5f4 a3e3 f4e6 c4e5 c6e5 e3e5');
INSERT INTO evaluations VALUES(54538,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(54538,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(54538,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(54538,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54538,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54538,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(54538,36,'cp -27','d8e7 d1e2 e7e2 e1e2 g8e7 c2c3 b8c6 d2d4 c5c4 b3d1 a7a6 b1a3 f7f6 a3c2 e8f7 c2e3 b7b5 b2b3 c8e6 h1e1 h7h5 e2f1 g7g6 f3d2 a8d8 c1a3 e7f5 a3f8 d8f8 b3c4 b5c4 d1a4 c6e7 f1g1 f8b8 d2f1 b8b2 e3f5 e7f5');
INSERT INTO evaluations VALUES(3079,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(3079,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(3079,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3079,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(3079,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(3079,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3079,36,'cp 0','d7d5 e1c1 d5e4 d4c6 d8d2 e3d2 b7c6 c3e4 f6e4 f3e4 c8g4 d1e1 g4e6 f1a6 f8b8 b2b3 c6c5 d2a5 b8b6 a6d3 b6b8');
INSERT INTO evaluations VALUES(3154,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(3154,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(3154,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3154,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3154,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3154,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(3154,36,'cp 0','e1c1 d5e4 d4c6 d8d2 e3d2 b7c6 c3e4 f6e4 f3e4 c8g4 d1e1 g4e6 f1a6 a8b8 b2b3 c6c5 d2a5 g7e5 g2g3 f7f6 e1e3 c5c4 h1e1 g8f7 a5c3 e5d6 c3a5');
INSERT INTO evaluations VALUES(3152,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3152,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3152,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3152,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3152,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3152,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(3152,36,'cp 0','d5e4 d4c6 d8d2 e3d2 b7c6 c3e4 f6e4 f3e4 c8g4 d1e1 g4e6 f1a6 a8b8 b2b3 c6c5 d2a5 g7e5 g2g3 c5c4 c1d2 h7h5 e1b1 c4b3 a2b3 h5h4 g3h4 g8g7 a5c3 e5c3 d2c3 f8h8 b1a1 h8h4 a6d3 b8c8 c3b2 c8c7 a1a5 h4h3 d3f1 h3e3 f1d3 e3h3');
INSERT INTO evaluations VALUES(3156,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3156,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3156,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3156,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3156,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3156,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(3156,36,'cp -1','d4c6 d8d2 e3d2 b7c6 c3e4 f6e4 f3e4 c8g4 d1e1 g4e6 f1a6 a8b8 b2b3 c6c5 d2a5 g7e5 g2g3 c5c4 c1d2 h7h5 e1b1 c4b3 a2b3 h5h4 g3h4 g8g7 a5c3 b8d8 a6d3 e5c3 d2c3 f8h8 b1a1 d8d7 h4h5 h8h5 b3b4 f7f5 h1e1 f5f4 b4b5 h5c5 c3b2 d7c7 a1a5 c5e5 h2h4');
INSERT INTO evaluations VALUES(18271,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(18271,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(18271,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(18271,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(18271,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(18271,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(18271,36,'cp -19','g8f6 b1c3 g7g6 h2h3 f8g7 c1e3 e8g8 d1d2 c6c5 e3g5 c8e6 b2b3 d8b8 g5h6 g7h6 d2h6 b8b4 h6d2 f6h5 g2g4 h5f4 c3e2 b4d2 e1d2 f4e2 d2e2 f7f6 f3d2 a7a5 h1b1 f6f5 f2f3 g8f7 a2a4 f7f6 d2c4 f8b8 c4d2 f5f4 h3h4 f6g7');
INSERT INTO evaluations VALUES(19789,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(19789,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(19789,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(19789,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19789,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(19789,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(19789,36,'cp 32','c2c4 e7e6 c4d5 e6d5 f1e2 g8f6 e1g1 c5d4 f3d4 f8d6 b2b3 d8e7 b1c3 c6d4 d1d4 d6e5 e2b5 c8d7 b5d7 f6d7 d4d2 e8g8 c1b2 d7f6 a1c1 h7h6 f1d1 f8c8 h2h3 e7e6 d2e2 a7a6 e2d2 f6e4 c3e4 e5b2 d2b2 c8c1 b2c1 d5e4');
INSERT INTO evaluations VALUES(19771,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(19771,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19771,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(19771,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(19771,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19771,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(19771,36,'cp 10','e7e6 c4d5 e6d5 f1e2 c5d4 f3d4 g8f6 e1g1 f8c5 b1c3 e8g8 e2f3 d8d6 d4b3 f8d8 c3b5 d6e7 b3c5 e7c5 b5d4 c8d7 d4c6 d7c6 d1d4 c5d4 e3d4 f6e4 c1f4 a7a5 f1c1 a5a4 h2h4 f7f6 c1c2 a4a3 b2a3 a8a3 a1c1 g8f7');
INSERT INTO evaluations VALUES(9850,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9850,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(9850,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(9850,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(9850,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(9850,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(9850,36,'cp 34','f2e3 d7d6 a2a3 c8e6 h1f1 e6c4 d3c4 c6b8 c4c5 b8d7 c5d6 e8g8 d6c7 d8c7 d1e2 a8c8 e1c1 c7b6 e2b5 b6e3 c1b1 e3c5 b5c5 c8c5 d1d3 c5c8 f1d1 d7c5 d3e3 f8d8 d1d8 c8d8 b1c1 g8f8 f3e5 d8e8 e5f3 c5e4 c3e4 f6e4 c2c4 f7f6 c1c2 f8f7 b2b4');
INSERT INTO evaluations VALUES(25793,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(25793,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(25793,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(25793,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(25793,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(25793,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(25793,36,'cp 91','d1f3 g8f6 e1g1 d8d7 b1c3 d7g4 f3d3 f8e7 d3c4 g4d7 c1f4 e8g8 a1d1 f6h5 f4e3 f8e8 b2b3 e7f6 f2f3 f6e5 f1e1 h5f4 c3e2 f4e2 e1e2 a7a5 a2a4 e5f6 e3d4 f6d4 c4d4 h7h6 d4c4 e8e5 c4c3 e5c5 c3d3');
INSERT INTO evaluations VALUES(31935,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31935,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31935,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(31935,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(31935,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31935,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31935,36,'cp 18','f1g2 c7c6 e1g1 e7e6 h2h3 g4h5 c2c4 b8d7 c4d5 e6d5 d2d4 g8f6 f3h4 h5g6 h4g6 h7g6 d1b3 d8b6 b3d3 f8e7 b1c3 e8g8 c1d2 f8e8 a1c1 a7a5 h3h4 b6d8 c1c2 d7f8 f1c1 f8e6 e2e3 d8d7 d3e2 a8d8 c1d1 e7d6');
INSERT INTO evaluations VALUES(31936,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31936,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31936,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(31936,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31936,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31936,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31936,36,'cp 20','c7c6 e1g1 g8f6 h2h3 g4f3 e2f3 a7a5 f3f4 e7e6 d2d3 b8d7 b1d2 f8d6 d2f3 e8g8 a2a4 f8e8 c1e3 d8c7 d1e2 d6c5 d3d4 c5e7 f3e5 d7b6 b2b3 b6c8 c2c4 c8d6 e2d3 d6f5 e3d2 f5d6 c4c5');
INSERT INTO evaluations VALUES(41507,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(41507,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41507,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(41507,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(41507,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(41507,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(41507,36,'cp 16','e8g8 c2c4 c7c6 f3e5 f6d7 c4d5 c6d5 g2d5 d7e5 d4e5 b8c6 b1c3 c8h3 f1e1 c6e5 d5b7 a8b8 d1d8 f8d8 b7d5 d8c8 d5b3 e5c6 e1d1 c6d4 c1e3 d4c2 a1c1 c2e3 f2e3 h3e6 b3e6 f7e6 c3a4 c8c1 d1c1 g7b2 c1c7');
INSERT INTO evaluations VALUES(56453,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(56453,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(56453,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(56453,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(56453,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56453,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56453,36,'cp -14','g1f3 f8d6 e2e4 g8f6 c1g5 d5e4 d3e4 b8d7 b1d2 d8e7 d2c4 h7h6 c4d6 e7d6 d1d6 c7d6 g5f6 d7f6 f3d2 e8e7 f1c4 c8d7 a2a4 a8c8 f2f3 c8c7 b2b3 h8c8 g2g3 a7a6 a4a5 c7c5 e1e2 d6d5 e4d5 f6d5');
INSERT INTO evaluations VALUES(56449,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(56449,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(56449,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(56449,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(56449,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56449,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(56449,36,'cp -23','f8d6 e2e4 g8f6 c1g5 c8e6 b1d2 b8d7 d3d4 h7h6 e4d5 e6d5 g5f6 d8f6 f1c4 d5f3 d1f3 e5d4 c3d4 e8g8 f3f6 d7f6 e1g1 c7c6 d2f3 f8e8 g2g3 g8f8 a1e1 d6c7 g1g2 a8d8 e1e8 f6e8 f1d1 g7g5 h2h3 d8d7 b2b4 e8f6 d1e1 c7b6');
INSERT INTO evaluations VALUES(5298,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5298,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5298,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(5298,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5298,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5298,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5298,36,'cp 0','e1g1 g8f6 b1c3 e8g8 c1e3 b7b6 d1d2 c8b7 e3g5 d7d6 f1e1 f6g4 h2h3 g4e5 f3e5 d6e5 a2a4 a8b8 c4d5 c6d4 d5c4');
INSERT INTO evaluations VALUES(5297,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5297,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5297,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5297,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(5297,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5297,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5297,36,'cp 0','g8f6 b1c3 e8g8 h2h3 d7d6 c1e3 b7b6 d1d2 c8b7 e3g5 e7e6 f1e1 d8d7 c4b3 f6h5 a1d1 c6e5 f3h2 e5c6');
INSERT INTO evaluations VALUES(8008,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8008,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(8008,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8008,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(8008,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8008,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8008,36,'cp 0','b1c3 e8g8 h2h3 d7d6 c1e3 b7b6 d1d2 c8b7 e3g5 g8h8 g5h6 e7e6 c4b3 a7a6 a2a4 d8c7 a1d1 a8d8 f1e1 c6e5 f3h2 e5c6');
INSERT INTO evaluations VALUES(8002,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8002,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8002,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8002,36,'cp 0','e8g8 h2h3 d7d6 c1e3 b7b6 d1d2 c8b7 e3g5 g8h8 g5h6 e7e6 f1e1 a7a6 a2a4 d8c7 c4b3 c6e5 f3h2 a8d8 a1d1 e5c6 h2f3 c6e5');
INSERT INTO evaluations VALUES(3043,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3043,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3043,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3043,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3043,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3043,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3043,36,'cp 0','c1e3 b7b6 d1d2 c8b7 e3g5 d7d6 f1e1 g8h8 g5h6 a7a6 h2h3 b6b5 c4b3 e7e6 a1d1 d8c7 a2a4 b5b4 c3e2 a6a5 e2g3 b7a6 h6e3 h8g8 e3h6 g8h8');
INSERT INTO evaluations VALUES(52176,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52176,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(52176,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52176,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52176,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(52176,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(52176,36,'cp 43','d4b5 a7a6 b5c3 b8c6 g2g3 f8c5 f1g2 g8f6 e1g1 h7h6 b1d2 d7d6 d2e4 f6e4 c3e4 e8g8 e4c5 d6c5 c1e3 d8e7 c2c3 c8e6 d1a4 f8d8 g2c6 b7c6 a4c6 c5c4 e3c5 e7g5 c5b6 d8c8 c6d6 g5h5 f2f3 e5e4 g3g4 h5g6 a1d1 e4f3 e2f3 c8e8 d6f4');
INSERT INTO evaluations VALUES(53036,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(53036,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(53036,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(53036,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(53036,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(53036,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(53036,36,'cp 106','d4c5 d6c5 d1d8 e8d8 b1c3 g7g6 g2g3 f8g7 f1g2 g7c3 b2c3 g8f6 g1f3 b8d7 c1f4 f6h5 f4e3 h5g7 f3g5 d8e8 a2a4 g7f5 e3f4 h7h6 g5e4 g6g5 f4c1 a8b8 a4a5 b7b6 h2h4 g5h4 c1f4 e7e5 g3g4 e5f4 g4f5 e8e7 a5b6 a7b6 h1h4 c8b7');
INSERT INTO evaluations VALUES(52991,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(52991,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(52991,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(52991,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(52991,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(52991,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(52991,36,'cp 79','d6c5 d1d8 e8d8 b1c3 g7g6 g2g3 c8d7 f1g2 d7c6 e2e4 b8d7 f2f4 f8g7 g1e2 h7h5 h2h3 e7e6 c1e3 a7a6 e1g1 g8e7 a1d1 d8c7 e4e5 c6g2 g1g2 e7f5 e3f2 g6g5 c3e4 g5f4 g3f4 g7f8 g2h1 f8e7 e2g3 f5h4 d1d3 a8d8 f1d1 h4g6');
INSERT INTO evaluations VALUES(57676,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(57676,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57676,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(57676,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(57676,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57676,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57676,36,'cp 0','e2e4 d7d5 e4d5 g8f6 d2d4 d8d5 b1c3 d5a5 g1f3 f8g7 f3e5 e8g8 e5c4 a5a6 c4e5');
INSERT INTO evaluations VALUES(16431,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(16431,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(16431,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(16431,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(16431,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(16431,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(16431,36,'cp -26','f8c5 h2h3 d7d6 g1e2 a7a6 g2g3 b7b5 f1g2 a8b8 e1g1 e8g8 b2b3 c8d7 g1h2 c6d4 c1e3 b5c4 b3c4 h7h6 a1b1 b8b1 c3b1 d4e2 d1e2 c5e3 e2e3 c7c6 f1c1 c6c5 e3d2 d8c7 b1c3 f8b8 c1b1 b8b1 c3b1 c7b6 b1c3 d7e6 d2c2 b6a5 c3d5 f6d5');
INSERT INTO evaluations VALUES(10164,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(10164,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10164,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(10164,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(10164,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(10164,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(10164,36,'cp -12','h2h3 e8g8 g1f3 a7a6 f1e2 f8e8 e1g1 d7d6 a1b1 h7h6 c1e3 c8d7 a2a3 c6d4 b2b4 d4f3 e2f3 c5e3 f2e3 b7b5 f3e2 c7c6 d1d2 d8e7 a3a4 e8b8 a4b5 a6b5 c4b5 c6b5 b1a1 b8c8 d3d4 e5d4 d2d4 a8a1 f1a1 e7e5 d4e5 d6e5');
INSERT INTO evaluations VALUES(10163,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(10163,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10163,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(10163,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(10163,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(10163,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(10163,36,'cp -12','d7d6 g1f3 e8g8 f1e2 c8d7 e1g1 a7a6 a2a3 c6d4 f3d4 c5d4 a3a4 c7c6 c1e3 d4e3 f2e3 b7b5 a4b5 a6b5 a1a8 d8a8 c4b5 c6b5 d3d4 f8d8 d1b3 e5d4 e3d4 d7e6 b3b5 f6e4 b5d3 e4c3 b2c3 d8c8 e2g4 e6g4 h3g4 h7h6 d3f3');
INSERT INTO evaluations VALUES(26027,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26027,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26027,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(26027,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(26027,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(26027,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(26027,36,'cp 11','g8f6 b1c3 c6d4 b5d7 d8d7 d1d4 e7e6 a2a4 f8e7 f1d1 d7c6 c1g5 h7h6 g5f6 e7f6 d4d6 c6d6 d1d6 f6c3 b2c3 a8c8 a1d1 e8g8 d6d7 c8c3 d7b7 a7a5 b7a7 c3c2 g2g3 c2c5 d1d7 g7g5 g1g2 c5c4 a7a5 c4e4 a5a7 g8g7 a4a5 e4e2 a5a6 e2a2 g2f3 e6e5 d7d6 f8e8');
INSERT INTO evaluations VALUES(24471,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(24471,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24471,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(24471,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(24471,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(24471,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(24471,36,'cp 15','b1c3 c6d4 b5d7 d8d7 d1d4 e7e6 a2a4 f8e7 f1d1 d7c6 c1g5 h7h6 g5f6 e7f6 d4d6 c6d6 d1d6 f6c3 b2c3 a8c8 a1d1 e8g8 d6d7 c8c3 d7b7 a7a6 g2g3 c3c2 d1d6 f8a8 d6d7 a8f8 b7a7 c2e2 a7a6 e2e4 h2h4 g7g5 h4h5 g8g7 a6a7');
INSERT INTO evaluations VALUES(34141,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(34141,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34141,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(34141,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(34141,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34141,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(34141,36,'cp 27','g8f6 f1d3 b7b5 f2f4 c8b7 d1f3 f8b4 e1g1 b8c6 g1h1 c6d4 e3d4 b4c5 d4e5 c7d8 f4f5 e8g8 f3h3 d7d6 e5f6 d8f6 e4e5 f6h6 h3h6 g7h6 e5d6 a8d8 f5e6 f7e6 f1f8 g8f8 d3h7 d8d6 h7e4 b7c6 e4c6 d6c6 c3e2 e6e5 c2c3 h6h5 g2g3 c5b6 a1d1 f8e7');
INSERT INTO evaluations VALUES(33527,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33527,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33527,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33527,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33527,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33527,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33527,36,'cp 0','f1d3 b7b5 f2f4 c8b7 d1f3 b8c6 e1c1 c6a5 d4b3 f8b4 b3a5 c7a5 e3d4 a8c8 f3g3 b4c3 d4c3 c8c3 g3g7 e8e7 e4e5 f6d5 g7h8 a5a2 f4f5 e6f5 h8h7 a2a1 c1d2 a1b2 h7h4 e7e8 h1e1 c3d3 d2d3 b2c3 d3e2 c3c2 e2f1 d5e3 e1e3 c2d1 e3e1 d1d5 h4h8 e8e7 h8h4');
INSERT INTO evaluations VALUES(33535,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33535,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33535,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33535,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33535,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33535,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33535,36,'cp 0','b7b5 f2f4 c8b7 d1f3 b8c6 e1c1 c6a5 d4b3 f8b4 b3a5 c7a5 e3d4 a8c8 f3g3 b4c3 d4c3 c8c3 g3g7 e8e7 e4e5 f6d5 g7h8 a5a2 h8g7 c3d3 c2d3 a2a1 c1c2 d5b4 c2b3 a1a4 b3c3 a6a5 d1c1 b4d5 c3d2 a4d4 g7g3 d4b2 c1c2 b2b4 d2d1 a5a4 g3g5 e7e8 g5g8 e8e7');
INSERT INTO evaluations VALUES(51289,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(51289,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(51289,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(51289,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(51289,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(51289,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(51289,36,'cp 175','c4d5 g8f6 g1f3 c8b7 d1c2 b7d5 b1c3 d5b7 e2e4 e7e6 f1d3 a7a6 c1e3 b8d7 e1g1 f8e7 f1d1 c7c5 d4d5 e6d5 e4e5 f6g4 e3f4 d5d4 d3e4 b7e4 c3e4 e8g8 h2h3 g4h6 e4g5 e7g5 f3g5 g7g6 d1e1 a6a5 e5e6 d8f6 c2e4 a8e8 e6f7 h6f7 e4e8 f7g5');
INSERT INTO evaluations VALUES(51284,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(51284,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(51284,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(51284,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(51284,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(51284,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(51284,36,'cp 168','g8f6 g1f3 c8b7 d1c2 b7d5 b1c3 e7e6 e2e4 d5b7 f1d3 a7a6 c1e3 b8d7 e1g1 f8e7 f1d1 c7c5 d4d5 e6d5 e4e5 f6g4 e3f4 d5d4 c3e4 b6b5 h2h3 c5c4 d3c4 b5c4 d1d4 g4f2 c2f2 b7e4 d4e4 e8g8 e4c4 d7b6 c4c6 b6d5 f4d2');
INSERT INTO evaluations VALUES(14761,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(14761,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(14761,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14761,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(14761,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14761,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(14761,36,'cp 33','e2e4 f8e7 g5f6 e7f6 f1b5 e8g8 e1g1 c6b8 f1e1 c7c6 b5d3 c6c5 d4c5 f6c3 b2c3 b8d7 e4d5 e6d5 c5c6 b7c6 h2h3 d7c5 d1d2 c8d7 e1e3 f8e8 a1e1 e8e3 d2e3 d8f8 e3f4 c5e6 f4h4 h7h6 f3e5 e6c5 e1e3 c5d3 c2d3 f8d6 e5d7');
INSERT INTO evaluations VALUES(14765,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(14765,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(14765,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(14765,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(14765,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(14765,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14765,36,'cp 27','f8e7 g5f6 e7f6 f1b5 e8g8 e1g1 c6b8 f1e1 c7c6 b5d3 c6c5 d4c5 f6c3 b2c3 b8d7 e4d5 e6d5 c5c6 b7c6 d1d2 d7c5 d2f4 c8d7 d3f1 c5e4 c3c4 d8f6 f4f6 e4f6 c4d5 c6d5 f3d4 f8c8 e1b1 g7g6 f2f3 d7a4 b1b2 a8b8 a1b1 b8b2 b1b2 c8c7 g1f2');
INSERT INTO evaluations VALUES(25506,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(25506,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(25506,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(25506,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(25506,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(25506,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(25506,36,'cp 21','f8e7 a2a3 h7h6 c1e3 e8g8 h2h3 f8e8 b5c4 a7a6 c4a2 e7f8 c3e2 d7c8 c2c4 c6e7 c4c5 e7g6 c5d6 f8d6 e2g3 c7c5 d1c2 b7b6 b2b4 c5b4 a3b4 c8e6 f1b1 e6a2 a1a2 b6b5 g3f5 d8d7 e3c5 d6f8 f5e3 f8c5 b4c5');
INSERT INTO evaluations VALUES(24199,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24199,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(24199,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24199,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24199,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24199,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24199,36,'cp 30','a2a3 a7a6 b5c4 h7h6 c3d5 f6d5 c4d5 e8g8 c2c3 e7f6 c1e3 a6a5 d5b3 c6e7 a3a4 e7g6 d3d4 d8e7 b3c2 a8e8 h2h3 b7b6 d1d2 c7c5 f1d1 e5d4 c3d4 d7c6 d4d5 c6b7 c2d3 g6e5 f3e5 e7e5');
INSERT INTO evaluations VALUES(49390,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(49390,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49390,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49390,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(49390,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(49390,36,'cp 0','d7d5 c4d5 f6d5 f4g3 c7c5 d4c5 f8c5 e2e4 c5b4 b1d2 d5f6 g3b8 b4d2 d1d2 a8b8 d2d8 e8d8 f1d3 d8e7 g1f3 c8d7 f3e5 d7c6 f2f3 h8d8 e1e2 f6d7 e5d7 d8d7 a1c1 b8d8 h1d1 c6a4 b2b3 d7d3 d1d3 a4b5 c1c7 e7f6 c7c3 b5d3 c3d3 d8d3 e2d3 f6e5 g2g3 f7f6 b3b4 g7g5 d3e3');
INSERT INTO evaluations VALUES(50013,30,'cp 342',NULL);
INSERT INTO evaluations VALUES(50013,31,'cp 337',NULL);
INSERT INTO evaluations VALUES(50013,32,'cp 359',NULL);
INSERT INTO evaluations VALUES(50013,33,'cp 341',NULL);
INSERT INTO evaluations VALUES(50013,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(50013,35,'cp 318',NULL);
INSERT INTO evaluations VALUES(50013,36,'cp 322','d2d4 d7d5 c4d5 c7c6 d5b3 b8a6 c1h6 g7h6 b1c3 c8g4 d4d5 a6c5 d1e2 d8b6 d5c6 b7c6 a1d1 c5b3 a2b3 f8b4 h2h3 g4h5 e1g1 e8g8 g2g4 h5g6 f3h4 a8d8 h4f5 g6f5 e4f5 b4c3 b2c3 a7a5 d1d8');
INSERT INTO evaluations VALUES(50015,30,'cp 334',NULL);
INSERT INTO evaluations VALUES(50015,31,'cp 317',NULL);
INSERT INTO evaluations VALUES(50015,32,'cp 334',NULL);
INSERT INTO evaluations VALUES(50015,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(50015,34,'cp 330',NULL);
INSERT INTO evaluations VALUES(50015,35,'cp 335',NULL);
INSERT INTO evaluations VALUES(50015,36,'cp 329','d7d5 c4d5 c7c6 d5b3 b8a6 c1h6 g7h6 b1c3 d8c7 d4d5 a6c5 d5c6 b7c6 f3h4 c7g7 d1h5 e8d8 h4f5 c8f5 e4f5 d8c7 e1g1 f8e7 g2g3 h8d8 a1d1 a7a5 h2h4 d8d1 f1d1 a8d8 b3e6 d8d1 h5d1 g7f8 d1g4 f8b8 e6b3 b8f8 g1g2 c5d7 b3e6 e7b4');
INSERT INTO evaluations VALUES(14076,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(14076,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(14076,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(14076,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(14076,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(14076,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(14076,36,'cp 69','c2c3 a7a6 b5a4 c8d7 d2d4 e7g6 d4d5 c6b8 c3c4 f8e7 b1c3 e8g8 a4d7 d8d7 c1e3 a6a5 f3d2 f7f5 e4f5 d7f5 d1e2 b8d7 g2g3 f8f7 c3b5 a8c8 f2f3 d7f6 f1f2 g6f8 b5c3 c8e8 d2e4 f6e4 c3e4 f5g6 a1e1 f8d7 h2h4');
INSERT INTO evaluations VALUES(14071,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(14071,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(14071,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(14071,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(14071,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(14071,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(14071,36,'cp 68','e7g6 d2d4 a7a6 b5a4 c8d7 d4d5 c6b8 c3c4 f8e7 b1c3 e8g8 a4d7 d8d7 c1e3 a6a5 d1c2 b8a6 a2a3 f7f5 e4f5 f8f5 f3d2 e7g5 b2b4 a8f8 b4a5 g5e3 f2e3 f5f1 a1f1 a6c5 f1f8 g6f8 d2e4 d7f5 c2f2 f5g4 e4c5 d6c5 f2g3 g4g3 h2g3 g8f7 g3g4 f8d7');
INSERT INTO evaluations VALUES(15907,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15907,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15907,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(15907,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15907,36,'cp 0','d6e5 f1c4 f8b4 e1g1 d8e7 d1e2 e8g8 c1d2 d7b6 c4b3 a7a5 a2a3 b4c5 a3a4 c8e6 b3e6 e7e6 b2b3 b6d7 f3h4 e6c6 a1d1 f8e8 h4f5 c5d4 e2b5 d7c5 b5c6 b7c6 c3a2 c5e6 f5d4 e6d4 f2f3');
INSERT INTO evaluations VALUES(44867,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44867,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(44867,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44867,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(44867,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(44867,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44867,36,'cp 15','c6d5 b1c3 b8c6 c1f4 c8f5 e2e3 e7e6 f1b5 f8b4 e1g1 b4c3 b2c3 e8g8 b5c6 b7c6 a1c1 f6e4 f3d2 e4d2 d1d2 d8a5 f2f3 f8c8 f1d1 c6c5 c3c4 a5d2 d1d2 c5d4 d2d4 c8c4 c1c4 d5c4 d4c4 f5d3 c4c6 d3b5 c6c5 a7a6 e3e4 a8d8 f4e3 f7f6 c5c7 e6e5');
INSERT INTO evaluations VALUES(56901,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(56901,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(56901,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(56901,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(56901,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(56901,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(56901,36,'cp 107','d2d4 b8c6 d4d5 c6b8 b1c3 d7d6 a2a4 b8d7 h2h4 h7h5 c1g5 g8f6 f2f3 f8e7 d1d2 f6h7 g5e7 d8e7 g1e2 a7a6 e2g3 e8g8 c4e2 e7f6 g3f1 f6f4 a4a5 f7f5 e4f5 f4d2 e1d2 g6f5 g2g3 d7f6 f1e3 g8h8 f3f4 f6g4 e3g4 h5g4');
INSERT INTO evaluations VALUES(8018,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(8018,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(8018,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(8018,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(8018,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(8018,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(8018,36,'cp 14','e8g8 d2d4 d7d6 d4d5 c6a5 f3d2 e7e6 a1b1 e6d5 c3d5 c8f5 e2e4 f5d7 b2b3 a5c6 d5f6 d8f6 f1e1 f6d8 c1b2 c6d4 d2f3 d7g4 d1d3 g4f3 g2f3 d8a5 b2d4 g7d4 a2a4 a8e8 f3g2 e8d8 e1d1 f8e8');
INSERT INTO evaluations VALUES(10378,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(10378,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(10378,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(10378,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(10378,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(10378,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(10378,36,'cp -26','b4c3 b2c3 d7d5 f3d2 c8e6 a3a4 h7h6 f1e2 e8g8 e1g1 f8e8 f1e1 b7b6 e4d5 f6d5 c1b2 d5f4 e2f1 d8d7 d3d4 a8d8 h2h3 e5d4 c3d4 c6d4 d2f3 d4f5 d1d7 e6d7 f3e5 a7a5 e1d1 d7e6 g2g3 f4g6 e5g6 f7g6 f1b5 d8d1 a1d1 e8a8 b5d3 h6h5 d3b5 g8f7');
INSERT INTO evaluations VALUES(10469,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(10469,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(10469,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(10469,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(10469,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(10469,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(10469,36,'cp -25','b2c3 d7d5 f3d2 e8g8 a3a4 b7b6 f1e2 c8e6 e1g1 h7h6 f1e1 d8d7 e2f1 a8d8 e4d5 f6d5 c1b2 d5f4 a4a5 f7f6 a5b6 a7b6 b2c1 f8e8 h2h3 e6f7 d2e4 f4g6 e4g3 c6e7 c3c4 e7c6 c1b2 g6f4 d1g4 d7g4 h3g4 c6b4 e1c1 f4e6 g3f5 f7g6');
INSERT INTO evaluations VALUES(10468,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(10468,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(10468,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(10468,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(10468,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(10468,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(10468,36,'cp -28','d7d5 f3d2 c8e6 a3a4 e8g8 f1e2 b7b6 e1g1 h7h6 f1e1 d8d7 e4d5 f6d5 c1b2 d5f4 e2f1 f8e8 h2h3 a8d8 a4a5 f7f6 a5b6 a7b6 d2e4 c6e7 b2c1 e7g6 c1f4 g6f4 c3c4 e6f7 e4c3 f4e6 d1b1 e6d4 b1b2 d8a8 c3b5 d4e6 b5c3 a8a1 e1a1 g8h8 b2b5');
INSERT INTO evaluations VALUES(9125,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(9125,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(9125,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(9125,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(9125,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(9125,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(9125,36,'cp -36','f3d2 c8e6 a3a4 e8g8 f1e2 b7b6 e1g1 h7h6 f1e1 d8d7 e2f3 d5e4 d2e4 f6e4 f3e4 e6d5 e4d5 d7d5 d1h5 a8d8 a4a5 b6a5 h2h3 f8e8 h5g4 e8e6 g4e4 d5d7 e4g4 d7d6 g4e4 e6g6 a1a4 d6d7 a4c4 g6d6 c4c5 d6d5 e4d5 d7d5 c5d5 d8d5 f2f4 d5b5 c3c4');
INSERT INTO evaluations VALUES(20664,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20664,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20664,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(20664,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20664,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(20664,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20664,36,'cp 13','g8f6 d2d4 c5d4 f3d4 f8e7 a2a4 e8g8 c1e3 a7a6 d4c6 b7c6 a4a5 c8b7 f1e1 d6d5 c4f1 c6c5 e4d5 e6d5 e3f4 f8e8 d1f3 h7h6 f4e5 e7d6 e5f6 d8f6 f3f6 g7f6 e1d1 d5d4 c3a4 f6f5 a4b6 a8d8');
INSERT INTO evaluations VALUES(31998,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(31998,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(31998,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(31998,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(31998,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(31998,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(31998,36,'cp 97','f3e5 g4d7 b1c3 e7e6 e5c4 g8f6 e2e4 f8e7 c1e3 e8g8 a2a3 b8c6 f1e2 d7e8 e1g1 b7b5 c4e5 a7a6 e5c6 e8c6 d1c2 c6b7 h2h3 f6d7 a1d1 d7b6 e2d3 d8d7 e4e5 h7h6 c2e2 b6c4 c3e4 c4e3 f2e3 b7e4 d3e4 a8d8 e2g4 e7g5 g4f3 g8h8');
INSERT INTO evaluations VALUES(31950,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(31950,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(31950,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(31950,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(31950,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(31950,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(31950,36,'cp 132','g4d7 e2e4 b8c6 e5c4 b7b5 c4e3 e7e5 d4e5 c6e5 f1e2 a8b8 e1g1 f8d6 a2a4 a7a6 b1c3 g8e7 f2f4 e5g6 f4f5 b5b4 c3b1 g6e5 b1d2 e8g8 d2c4 d7c6 c4d6 d8d6 d1d6 c7d6 f5f6 g7f6 f1f6 c6e4 f6d6 a6a5 e3c4 e5c4 e2c4 f8d8 c1f4 d8d6 f4d6');
INSERT INTO evaluations VALUES(55709,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55709,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55709,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55709,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(55709,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(55709,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55709,36,'cp -7','d6d5 e4d5 e6d5 c4e2 f8d6 d2d4 h7h6 c2c4 d5c4 e2c4 g8f6 d1e2 d8e7 e2e7 e8e7 e1g1 h8d8 f1e1 e7f8 f3e5 f6g4 b1c3 g4e5 d4e5 d6b4 a2a3 b4c3 b2c3 b8c6 f2f4 c6a5 c4e2 c8e6 c1e3 b7b6 g1f2 a5c4 a3a4 f8e7 g2g4 c4e3 f2e3 a8b8 e1d1 d8d1 a1d1');
INSERT INTO evaluations VALUES(33588,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(33588,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(33588,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(33588,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(33588,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(33588,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(33588,36,'cp 63','c8f5 f1c4 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 c4b3 a7a5 g2g3 d7b6 h4g6 h7g6 a2a4 f8b4 d2f4 c7e7 c3e4 f6e4 e2e4 b6d5 h2h4 d5f4 e4f4 b4d6 f4e3 e8c8 h1e1 e7c7 e3f3 d8d7 c1b1 c8b8 c2c3 h8d8 b3c2 e6e5 d4d5 c6d5 d1d5 d6c5 d5d7 d8d7');
INSERT INTO evaluations VALUES(32574,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(32574,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(32574,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(32574,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32574,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(32574,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(32574,36,'cp 84','f1c4 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 c4b3 a7a5 g2g3 d7b6 h4g6 h7g6 a2a3 a8d8 c3e4 b6d5 e4g5 b7b5 h2h4 f8e7 c1b1 c7a7 c2c3 d8b8 d2f4 d5f4 g3f4 f6d5 b3d5 c6d5 h4h5 e7g5 f4g5 b5b4 h5g6 h8h1 d1h1');
INSERT INTO evaluations VALUES(32572,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(32572,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(32572,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32572,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(32572,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(32572,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(32572,36,'cp 63','b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 b7b5 c4d3 b5b4 c3e4 f6e4 d3e4 d7f6 e4f3 a7a5 g2g4 f6d5 g4g5 e8c8 h4h5 g6h5 g5g6 f7g6 e2e6 c8b8 h1e1 a5a4 e6g6 f8d6 f3d5 c6d5 e1e6 b8b7 d2g5 d8d7 d1e1');
INSERT INTO evaluations VALUES(49499,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(49499,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(49499,32,'cp -126',NULL);
INSERT INTO evaluations VALUES(49499,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(49499,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(49499,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(49499,36,'cp -100','d7d5 c4b3 f6e4 g1e2 c8g4 e1g1 f8c5 d1e1 c7c6 c2c4 d4c3 b1c3 e4c3 b2c3 g4e2 e1e2 d8e7 c1e3 b8a6 e3d4 a6c7 f2f4 e8c8 e2e3 c7e6 f4f5 e6d4 c3d4 c5b6 a2a4 e7b4 f1d1 h8e8 a1b1 f7f6 e5e6 b4a3 g2g3 g7g6 e3f4 g6f5 b3c2 a3d6 f4d6 d8d6 c2f5');
INSERT INTO evaluations VALUES(48765,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(48765,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(48765,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(48765,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(48765,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(48765,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(48765,36,'cp -91','c4b3 f6e4 g1e2 c8g4 e1g1 f8c5 d1e1 c7c6 c2c4 d4c3 b1c3 g4e2 e1e2 e4c3 b2c3 d8e7 c1e3 b8d7 e3d4 d7f8 e2e3 f8e6 f2f4 c5b6 d4b6 a7b6 e3b6 e7c5 b6c5 e6c5 a1d1 e8c8 b3c2 c8c7 a2a4 f7f6 d1b1 d8a8 b1a1 h8e8 e5f6 g7f6 c2h7');
INSERT INTO evaluations VALUES(55551,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(55551,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55551,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55551,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(55551,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(55551,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(55551,36,'cp 4','c7c5 d2d3 b8c6 g2g3 h7h5 h2h4 g7g6 f1g2 f8g7 g1e2 g8e7 e1g1 e8g8 a2a3 c8g4 f2f3 g4e6 c1g5 f7f6 g5d2 a8b8 g1h2 g8h7 c3d5 b7b5 b2b4 e6d7 a1c1 b5c4 d3c4 c6d4 e2d4 c5d4 a3a4 e7d5 c4d5 g7h6 f3f4');
INSERT INTO evaluations VALUES(16544,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(16544,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(16544,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(16544,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(16544,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(16544,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(16544,36,'cp -23','f8c5 b1c3 e8g8 h2h3 d7d6 d3c4 h7h6 a2a3 c6d4 d2d3 c7c6 c4a2 a7a5 a3a4 d4f3 d1f3 c8e6 c3e2 f6d7 e2g3 d8f6 f3f6 d7f6 a2e6 f7e6 c1e3 f6d7 f1d1 c5e3 f2e3 d7c5 g3f1 b7b5 b2b3 f8f7 f1d2');
INSERT INTO evaluations VALUES(10243,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(10243,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10243,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(10243,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(10243,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(10243,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(10243,36,'cp -21','b1c3 d7d6 h2h3 e8g8 d3c4 h7h6 a2a3 c8e6 d2d3 a7a5 c4e6 f7e6 c1e3 b7b6 c3e2 f6h5 d3d4 e5d4 f3d4 c6d4 e2d4 d8e8 d1d3 g8h8 a1d1 c5d4 d3d4 e6e5 d4c4 a8c8 c4e2 h5f6 f2f3 f6h5 g1h2 e8g6 f1f2 f8f7 e3d2 h5f6 d1f1');
INSERT INTO evaluations VALUES(31340,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(31340,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31340,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31340,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31340,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31340,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(31340,36,'cp 0','b8c6 h2h3 g4f3 d1f3 g8f6 c1e3 e7e6 f1d3 f8d6 g2g4 e8g8 g4g5 f6d7 a2a3 d8a5 f3g2 d6b4 a3b4 a5a1 e1d2 a1b2 h1b1 b2a3 b1b3 a3a1 b3b1');
INSERT INTO evaluations VALUES(41266,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41266,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(41266,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41266,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41266,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(41266,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(41266,36,'cp 26','g1f3 c8f5 e2e3 f6e4 c3e4 f5e4 f1d3 e4d3 d1d3 d8c8 e1g1 e8g8 a1d1 b8c6 e3e4 e7e5 f4e3 e5d4 f3d4 a7a5 f1e1 a5a4 d3c2 g7d4 e3d4 c6d4 d1d4 f8e8 h2h3 c8d7 f2f3 d7e6 e1d1 e6e5 d4d5 e5e7');
INSERT INTO evaluations VALUES(43715,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(43715,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(43715,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(43715,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(43715,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(43715,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(43715,36,'cp 67','e7e6 b1c3 e6d5 c3d5 f8e7 d5e7 d8e7 e2e3 e8g8 a2a3 d7d5 f1b5 c8a6 b5a6 a8a6 g1e2 f8d8 e1g1 b8c6 c1d2 h7h6 a1c1 f6e4 d2e1 e4g5 e1c3 g5e6 d1d3 a6b6 f1d1 d5d4 e3d4 c6d4 c3d4 e6d4 e2d4');
INSERT INTO evaluations VALUES(51499,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(51499,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(51499,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(51499,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(51499,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(51499,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51499,36,'cp 13','c7c5 g1f3 c8b7 c2c3 g8f6 b1d2 e7e6 f1e2 f8e7 e1g1 d7d6 f1e1 e8g8 e2f1 b6b5 a2a4 a7a6 a4b5 a6b5 a1a8 b7a8 d3d4 a8c6 e4e5 d6e5 d4c5 e5e4 f3d4 e7c5 d4c6 b8c6 d1e2 b5b4 d2e4 c5e7 c3b4 c6b4 e4f6 e7f6 e1d1');
INSERT INTO evaluations VALUES(16614,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(16614,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(16614,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(16614,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(16614,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(16614,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(16614,36,'cp -27','d7d5 e1g1 a7a5 b1c3 h7h6 e2e4 d5e4 d3e4 d8d1 f1d1 c8e6 c3b5 a8c8 f3e1 f8c5 e1d3 c5b6 b2b3 h6h5 c1a3 f6d7 h2h3 c6d4 b5d4 b6d4 a1c1 f7f6 g2f1 g7g5 c2c3 d4b6 d1d2 h5h4 g3g4 a5a4 b3a4 c8a8 d3b2 d7c5 f1b5 e8f7');
INSERT INTO evaluations VALUES(26702,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(26702,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(26702,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(26702,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(26702,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(26702,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(26702,36,'cp 56','f1b5 c7c6 b5a4 e7e6 a2a3 b4a6 b1d2 f8d6 f4d6 d8d6 c2c4 g8f6 e1g1 e8g8 a4c2 a6c7 a1c1 f5c2 d1c2 a7a5 f1e1 d6e7 h2h3 h7h6 f3e5 f6d7 e5d3 c7e8 c4c5 e8f6 f2f4 f8b8 d3e5 d7e5 d4e5 f6d7 d2f3 b7b6 f3d4 e7c5 d4c6 c5c2 c1c2 b8b7');
INSERT INTO evaluations VALUES(29228,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(29228,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(29228,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(29228,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(29228,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(29228,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(29228,36,'cp 49','d1b3 b7b6 c4d5 e6d5 e2e3 a7a5 f1e2 b6b5 e1g1 a5a4 b3d1 f8e7 f3e5 e8g8 a1c1 f6d7 e2d3 f5d3 e5d3 b8a6 c3e2 a8c8 b2b3 a4b3 a2b3 f8e8 d1d2 d7f6 f2f3 d8b6 b3b4 f6h5 f4e5 f7f6 e5g3 h5g3 e2g3 g7g6 c1c2 g8g7 h2h4 e7d6 g3e2 a6c7');
INSERT INTO evaluations VALUES(24829,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(24829,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24829,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(24829,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24829,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(24829,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(24829,36,'cp 33','e7e6 c2c4 f8d6 f4g5 e8g8 f1e2 h7h6 g5h4 d6e7 e1g1 a7a5 a2a3 f6e4 h4e7 c6e7 d2e4 f5e4 f3d2 e4h7 d1b3 a5a4 b3c3 e7f5 b2b4 a4b3 c3b3 d5c4 d2c4 f5d6 c4d2 d6e4 d2e4 h7e4 f1c1 e4d5');
INSERT INTO evaluations VALUES(25162,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(25162,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(25162,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(25162,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(25162,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(25162,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(25162,36,'cp 22','c2c4 f8d6 f4g5 e8g8 f1e2 h7h6 g5h4 a7a5 a2a3 d6e7 e1g1 c6b8 d1b3 a8a7 a1c1 a5a4 b3c3 b8d7 h4f6 d7f6 c4d5 f6d5 c3c4 c7c6 e2d3 d5b6 c4c2 f5d3 c2d3 b6d7 f1d1 a7a8 h2h3 d8b8 d2c4 b8c7 c4e5 d7e5 f3e5 f8d8');
INSERT INTO evaluations VALUES(25138,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(25138,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(25138,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(25138,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(25138,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(25138,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(25138,36,'cp 24','f8d6 f4g3 e8g8 f1e2 d5c4 e2c4 d8e7 c4b5 c6b4 e1g1 d6g3 h2g3 a7a6 b5e2 c7c5 d4c5 e7c5 f3d4 c5e7 a2a3 b4d5 d1b3 f5g6 a1c1 a8c8 e2f3 e6e5 f3d5 c8c1 f1c1 e5d4 e3d4 e7e2 d2f3 f6d5 b3d5 e2b2 c1c7 h7h6');
INSERT INTO evaluations VALUES(41283,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(41283,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(41283,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(41283,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(41283,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41283,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41283,36,'cp 25','e8g8 e1g1 c8f5 c1e3 d6d5 c4d5 f6d5 e3c1 b8c6 b1c3 d5c3 b2c3 e7e5 c1g5 d8d6 d4e5 c6e5 d1d6 e5f3 g2f3 c7d6 f3b7 a8b8 b7d5 g7c3 a1c1 f8c8 f1d1 g8f8 h2h4 c3e5 g5e3 b8b2 c1c8 f5c8 e3a7 b2e2 a7e3');
INSERT INTO evaluations VALUES(48192,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(48192,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48192,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48192,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(48192,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(48192,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(48192,36,'cp -13','e6d5 d2d4 f8d6 f1d3 e8g8 h2h3 f8e8 c3e2 b8d7 e1g1 c7c5 d4c5 d7c5 c1f4 d6f4 e2f4 d8b6 a1b1 c8d7 f1e1 h7h6 e1e8 a8e8 d3f1 b6d6 f4d3 f6e4 c2c3 d7f5 f3d4 f5g6 b1a1 d6f6 d1c1 a7a5 c1e3 e4d6 e3f4 f6f4');
INSERT INTO evaluations VALUES(5345,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(5345,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(5345,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(5345,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5345,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5345,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(5345,36,'cp 0','c1e3 c6d4 a2a4 d7d6 f1e2 c8d7 e1g1 d7c6 e3d2 d4e6 f3g5 e6d4');
INSERT INTO evaluations VALUES(5349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5349,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5349,36,'cp 0','c6d4 a2a4 d7d6 f1e2 c8d7 e1g1 d7c6 e3d2 d4e2 d1e2 g8f6 e4e5 f6d5 e5e6 f7e6 f3g5 d5c7 c3b5 g7b2 b5c7 d8c7 e2e6 b2a1 e6f7 e8d7 f7e6 d7e8 e6f7');
INSERT INTO evaluations VALUES(48166,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(48166,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(48166,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(48166,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(48166,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(48166,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(48166,36,'cp -49','b7b6 c5b6 a7b6 e2e3 f8b4 c1d2 e8g8 g1f3 b8d7 f1e2 b4c3 d2c3 f6e4 a2a4 c8a6 e1g1 c7c5 e2a6 a8a6 b2b3 d8a8 d1c2 c5c4 c3b4 f8c8 c2b2 a8b7 f1b1 b6b5 a4b5 a6a1 b2a1 b7b5 b3c4 d5c4 f3d2 b5c6 d2e4 c6e4 b1c1');
INSERT INTO evaluations VALUES(43788,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(43788,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(43788,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(43788,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(43788,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(43788,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(43788,36,'cp -49','c5b6 a7b6 e2e3 f8b4 c1d2 c7c5 f1e2 d8e7 g1f3 e8g8 e1g1 c8a6 e2a6 b8a6 d1e2 c5c4 a2a4 f8c8 f1b1 b4a5 f3e5 a6b4 h2h3 b4c2 a1a2 h7h6 c3b5 a5d2 e2d2 c2b4 a2a3 f6e4 d2d1 a8a5 b1a1 b4c6 e5c6 c8c6 d1e2 e4d6 b5c3');
INSERT INTO evaluations VALUES(897,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(897,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(897,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(897,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(897,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(897,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(897,36,'cp 4','c1e3 e7e5 g1f3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 d8e8 f1b5 f5d7 c2c3 c6e5 b5d7 c8d7 f3e5 e8e5 c1c2 f8d6 h1d1 e5e4 d1d5 h8e8 b2b4 e8e5 a1d1 b7b6 d5d2 d7c6 a2a4 a7a6 g2g3 g7g5 d2d4 e4d4 d1d4 h7h5 c2d3 g5g4 d4e4 e5e4 d3e4');
INSERT INTO evaluations VALUES(894,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(894,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(894,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(894,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(894,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(894,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(894,36,'cp 0','e7e5 g1f3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 c6b4 f3e1 g7g6 e3g5 d8e8 g5f6 f8h6 c1d1 h8f8 f1d3 f5d3 e1d3 b4d5 f6h4 d5f4 h4g3 f4d3 c2d3 e8e6 f2f4 f7f6 d1c2 e6c6 c2b3 c6b6 b3c2');
INSERT INTO evaluations VALUES(14920,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(14920,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(14920,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(14920,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(14920,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(14920,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(14920,36,'cp 178','f1c4 f8b4 c4d3 b4a5 e1g1 a5b6 c1e3 e8g8 e4e5 f6d5 d3h7 g8h7 f3g5 h7g6 d1d3 f7f5 e5f6 g6f6 a1e1 d5c3 b2c3 e6e5 d4e5 c6e5 g5h7 f6f7 d3e4 e5g4 h7g5 f7f6 e3b6 a7b6 g5e6 c8e6 e4e6 f6g5 e1d1 d8f6 d1d5 g4e5 d5e5 g5g6 e6g4 g6h7 g4e2 h7g8 h2h3 a8a5 e5e7 a5c5 c3c4 f6c6 e2g4 f8f7');
INSERT INTO evaluations VALUES(21028,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(21028,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(21028,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(21028,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(21028,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(21028,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(21028,36,'cp -35','b1c3 e7e5 d4e5 d5d4 c3e4 d8e7 e2e3 d4e3 d1d3 e7e5 d3b5 e5b5 f1b5 c8f5 c1e3 e8c8 g1e2 c6b4 a1c1 g8f6 e2g3 f5e4 g3e4 f6e4 f3e4 b4a2 c1d1 a7a6 d1d8 c8d8 b5a4 a2b4 e4e5 f8e7 e1g1 b7b5 a4b3 f7f6 c2c3 b4c6 e5f6 e7f6 e3c5');
INSERT INTO evaluations VALUES(20998,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(20998,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(20998,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(20998,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(20998,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(20998,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(20998,36,'cp -35','e7e5 d4e5 d5d4 c3e4 d8e7 e2e3 d4e3 d1d3 e7e5 d3b5 f7f5 b5e5 c6e5 e4c3 f5f4 g1e2 e5g6 g2g3 g8f6 e2f4 g6f4 g3f4 f8c5 c3a4 c5b4 a4c3 f6d5 a2a3 b4c3 b2c3 e8g8 c3c4 d5f4 c1e3 c8h3 e1c1 h3f1 h1f1 f4g6 d1d7 f8f7 d7f7 g8f7 f1d1 g6e5 d1d4 f7g6 d4e4 a8e8 c4c5 c7c6 e3f4');
INSERT INTO evaluations VALUES(13217,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(13217,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(13217,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(13217,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(13217,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(13217,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(13217,36,'cp 73','e4e5 f6d7 d2d4 c5d4 f3d4 d7e5 f2f4 a7a6 f4e5 a6b5 d4b5 c6a7 b5a7 a8a7 e1g1 d6e5 d1d8 e8d8 f1f7 a7a6 f7f2 g7g6 c1g5 c8f5 a2a4 d8e8 a4a5 e7e6 f2e2 f8c5 g5e3 c5e3 e2e3 e8e7 e3e5 f5c2 a1c1 c2d3 c3a4 h8d8 a4c5 a6c6 c5d3 d8d3 c1c6 b7c6 g1f2 d3d2 e5e2 d2d1 a5a6');
INSERT INTO evaluations VALUES(26592,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(26592,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(26592,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(26592,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(26592,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(26592,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(26592,36,'cp 82','c4d5 g4f3 g2f3 d8d5 e2e3 e7e5 b1c3 f8b4 c1d2 b4c3 b2c3 d5d6 h1g1 g7g6 d1b3 g8e7 e1c1 e8c8 f1g2 f7f5 f3f4 e5f4 c3c4 f4e3 f2e3 h8e8 d2c3 d6h2 d4d5 c6e5 c3d4 e5g4 c4c5 c7c6 d5d6 e7d5 g2d5 c6d5 c1b1 h2e2 d1c1 g4e3 c5c6 b7c6 c1c6 c8d7');
INSERT INTO evaluations VALUES(26520,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(26520,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(26520,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(26520,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(26520,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(26520,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(26520,36,'cp 69','g4f3 g2f3 d8d5 e2e3 e7e5 b1c3 f8b4 c1d2 b4c3 b2c3 d5d6 h1g1 g7g6 d1b3 g8e7 e1c1 e8c8 f1g2 h8e8 c1b1 d6e6 b3e6 f7e6 g2h3 d8d6 f3f4 e5d4 c3d4 e8d8 d1c1 c8b8 c1c5 e7d5 g1c1 c6e7 a2a4 d8f8 a4a5 a7a6 b1c2 g6g5 c2d3');
INSERT INTO evaluations VALUES(21270,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(21270,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(21270,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(21270,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(21270,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(21270,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(21270,36,'cp 59','c6e7 c4d3 g7g6 c2c4 f7f5 b1c3 g8f6 a2a4 a7a5 e1g1 f8g7 e4f5 c8f5 f3h4 f5d3 d1d3 g6g5 h4f5 e7f5 d3f5 e8g8 c1e3 d8d7 f5c2 a8e8 b2b3 g5g4 a1e1 h6h5 e3g5 f6h7 g5d2 b7b6 f2f3 g4f3 g2f3 d7h3 c2g6 h7f6 g6g2 h3g2 g1g2 g8h7 c3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(20805,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(20805,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(20805,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(20805,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(20805,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(20805,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(20805,36,'cp 62','c4d3 g7g6 c2c4 f7f5 b1c3 g8f6 a2a4 f8g7 a4a5 e8g8 f3d2 f8f7 f2f3 c7c5 d1e2 g8h8 b2b3 c8d7 c1b2 a7a6 g2g3 d8f8 c3a4 f5e4 d2e4 e7f5 a4b6 a8e8 e2f2 f8e7 b6d7 f6d7 e1g1 e8f8');
INSERT INTO evaluations VALUES(40971,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40971,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40971,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(40971,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40971,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40971,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(40971,36,'cp 26','c1f4 e8g8 e2e3 b7b6 c4d5 f6d5 c3d5 e6d5 d1f3 c8e6 f1d3 c7c5 g1e2 d8d7 h2h3 b8c6 a1d1 c5c4 d3b1 a8d8 e1g1 b6b5 e2c3 a7a6 f1e1 f7f5 f3g3 e6f7 b1c2 d8e8 d1d2 g8h8 h3h4 e8c8 c3e2 f7h5 f4e5 c6e5 d4e5 h5e2 e1e2');
INSERT INTO evaluations VALUES(40951,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40951,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(40951,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40951,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40951,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40951,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40951,36,'cp 17','e8g8 e2e3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c7c5 d1f3 c5d4 e3d4 e7d6 f4d6 d8d6 a1c1 c8a6 d3a6 b8a6 g1e2 a6c7 e1g1 a8c8 h2h4 c7e6 f1d1 h7h6 c1c3 f8d8 d1c1 c8c3 c1c3 g7g6 g2g3 d8e8 g1g2 d6d7 f3g4 h6h5 g4f3 g8g7 c3c1 e8e7 c1c2');
INSERT INTO evaluations VALUES(36317,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36317,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36317,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(36317,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36317,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36317,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36317,36,'cp 11','e2e3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c7c5 d1f3 c5d4 e3d4 e7d6 f4d6 d8d6 g1e2 c8a6 d3a6 b8a6 e1g1 a6c7 a1c1 d6d7 h2h4 f8e8 f1e1 h7h6 e2f4 a8d8 g2g3 c7b5 e1e8 d8e8 f3d5 d7d5 f4d5 b5d4 c1c7 e8e2 c7a7 e2b2');
INSERT INTO evaluations VALUES(47773,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(47773,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(47773,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(47773,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(47773,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(47773,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(47773,36,'cp 32','d6e5 d1d8 e8d8 g1f3 f8d6 a2a3 d8e7 f3d2 c8e6 d2c4 e6c4 f1c4 c7c6 f2f3 b8d7 c4a2 a7a5 c3d1 d7c5 c1e3 f6d7 d1f2 c5e6 e1c1 h8d8 a2e6 e7e6 f2d3 h7h5 d1d2 a5a4 c1b1 f7f6 h1d1 d6e7 h2h4 b7b6 d3f2 g7g6 b1a1 d7c5');
INSERT INTO evaluations VALUES(53007,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53007,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(53007,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(53007,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53007,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(53007,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53007,36,'cp -11','g8f6 d2d3 e7e6 b1d2 f8e7 g1f3 e8g8 e1g1 a7a6 a2a4 b7b6 f1e1 c8b7 c4b3 d8c7 b3c2 b8d7 d2f1 f8c8 h2h3 b6b5 c1d2 d7e5 d2f4 e5g6 f4h2 e7f8 f1e3 b7c6 f3d2 c8e8 f2f4 e6e5 f4f5 g6f4 d2f3 a8d8 g1h1');
INSERT INTO evaluations VALUES(45132,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45132,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(45132,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(45132,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45132,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(45132,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45132,36,'cp 0','d2d3 e7e6 g1f3 f8e7 e1g1 e8g8 f1e1 d6d5 c4b3 d5e4 d3e4 d8d1 b3d1 f8d8 d1c2 b8c6 a2a4 b7b6 b1a3 h7h6 h2h3 c8a6 c1f4 d8d7 a3b5 g7g5 f4h2 a8d8 h2c7 d8c8 c7h2 c8d8');
INSERT INTO evaluations VALUES(45133,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(45133,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(45133,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45133,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45133,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(45133,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(45133,36,'cp 0','e7e6 g1f3 f8e7 e1g1 e8g8 b1d2 a7a6 a2a4 b7b6 f1e1 c8b7 c4a2 b8d7 a2b1 b6b5 d2f1 d8c7 h2h3 f8b8 f1g3 d7e5 f3h2 e5g6 f2f4 b5a4 d1a4 a6a5 a4d1 e7f8 h2f3 c7b6 b1c2 b7c6 f4f5 e6f5 g3f5');
INSERT INTO evaluations VALUES(5852,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(5852,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(5852,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(5852,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(5852,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(5852,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(5852,36,'cp 192','g8f6 c4d3 h7h6 e1g1 b4a5 c1e3 a5b6 h2h3 e8g8 d3c2 e6e5 d4e5 d8d1 a1d1 f6d7 e3f4 d7c5 c3d5 c8e6 b2b4 c5d7 a2a3 f8d8 c2b3 d7f8 f1e1 a8c8 f4g3 f8g6 b3a2 d8d7 b4b5');
INSERT INTO evaluations VALUES(10206,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(10206,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(10206,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(10206,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(10206,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(10206,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(10206,36,'cp 60','f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 f1d3 d7d5 e4d5 e5c3 b2c3 d8d5 e1g1 c8e6 c3c4 d5c6 d1d2 e8c8 d2f4 f6d7 f1e1 h8e8 c1e3 d7b6 e3d4 g7g5 f4f6 h7h5 d4b6 a7b6 a2a4 g5g4 h3h4 c6d6 a4a5 g4g3 f2g3 d6g3 a5b6 c7b6 d3f5 c8c7 f5e6 f7e6');
INSERT INTO evaluations VALUES(21292,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(21292,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(21292,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(21292,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(21292,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(21292,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(21292,36,'cp 39','d2d3 e5f4 c1f4 c6a5 b1c3 g8e7 f4g3 e7g6 e1g1 a5c4 d3c4 f8e7 c4c5 c8e6 f3d4 e8g8 d4e6 f7e6 d1g4 f8f1 a1f1 g6f8 h2h4 d8d7 c5d6 c7d6 c3e2 f8h7 e4e5 d6e5 g3e5 e7c5 g1h1 a8f8 e2f4 h7f6 g4e6 d7e6 f4e6 f8e8 e6c5 e8e5 c5d3 e5e2 f1f2 e2e7 h1g1 e7c7 d3f4 c7c4 c2c3 a7a6 f4g6');
INSERT INTO evaluations VALUES(29722,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(29722,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(29722,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(29722,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(29722,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(29722,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(29722,36,'cp -52','g2g4 f5g6 g4g5 f6d5 c3e4 b8c6 f1b5 e7e5 b5c6 b7c6 g1e2 f7f5 g5f6 d5f6 c1g5 f8e7 g5f6 e7f6 e4f6 d8f6 d4e5 f6f3 h1f1 f3h5 d1d2 h5h4 f1f2 h8f8 d2d4 h4f2 d4f2 f8f2 e1f2 a8d8 e2f4 g6c2 a1c1 d8d2 f2e3 d2h2 f4e6 e8e7 e6d4 c2a4 d4f5 e7d7 f5g7 h2b2');
INSERT INTO evaluations VALUES(29720,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(29720,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(29720,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(29720,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(29720,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(29720,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(29720,36,'cp -47','f5g6 g4g5 f6d5 c3e4 b8c6 f1b5 e7e5 b5c6 b7c6 g1e2 f7f5 g5f6 d5f6 c1g5 f8b4 c2c3 b4e7 d4e5 f6e4 d1d8 e7d8 g5d8 a8d8 f3e4 g6e4 h1f1 h8f8 f1f8 e8f8 e2d4 d8e8 e5e6 c6c5 d4b3 c5c4 b3c5 e4f5 e1c1 f5e6 c5e6 e8e6 d1d4 e6c6 d4d7 c6h6');
INSERT INTO evaluations VALUES(29790,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(29790,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(29790,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(29790,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(29790,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(29790,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(29790,36,'cp -46','g4g5 f6d5 c3e4 b8c6 f1b5 e7e5 b5c6 b7c6 g1e2 f7f5 g5f6 d5f6 c1g5 f8b4 c2c3 b4e7 d4e5 f6e4 d1d8 e7d8 g5d8 a8d8 f3e4 g6e4 h1f1 h8f8 f1f8 e8f8 e2d4 d8e8 e5e6 c6c5 d4b3 c5c4 b3c5 e4d5 e1c1 d5e6 d1d4 e6f7 h2h4 h7h5 c1d2 g7g6 b2b4 c4b3 a2b3');
INSERT INTO evaluations VALUES(31358,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31358,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31358,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31358,36,'cp 0','h2h3 g4f3 d1f3 g8f6 d2d4 b8c6 c1e3 e7e6 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(31357,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31357,36,'cp 0','g4f3 d1f3 e7e6 d2d4 b8c6 c1e3 g8f6 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(47940,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(47940,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47940,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(47940,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(47940,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47940,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(47940,36,'cp 32','g7g6 c1g5 f8g7 e1g1 e8g8 c2c3 a7a6 a2a4 b8d7 f1e1 h7h6 g5h4 d7e5 d3f1 g6g5 h4g3 f6h5 b1d2 h5g3 h2g3 f8e8 d2c4 c8d7 d1c2 e5c4 f1c4 g7e5 d4f3 d8f6 c4d5 d7c6 f3d2 g8g7 d5c6');
INSERT INTO evaluations VALUES(48323,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(48323,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(48323,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(48323,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(48323,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(48323,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(48323,36,'cp -65','e2e4 c7c5 e4d5 f6d5 c3d5 e6d5 g1e2 b8c6 c1e3 c5c4 d1d2 f8d6 g2g4 b7b5 f1g2 b5b4 e1g1 e8g8 h2h3 c6e7 e3f4 e7g6 f4d6 d8d6 f3f4 f7f5 g4g5 f8e8 c2c3 a7a5 c3b4 a5b4 a2a3 b4b3 e2c3 c8e6 d2f2 h7h6 g5h6 g7h6 g1h1 e6f7 a3a4 g8h8');
INSERT INTO evaluations VALUES(5341,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5341,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(5341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5341,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5341,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(5341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5341,36,'cp 2','g1e2 a8b8 a2a4 d7d6 d2d3 g8f6 e1g1 e8g8 c3d5 f6d5 e4d5 c6d4 e2d4 c5d4 f1e1 f8e8 c1d2 c8d7 h2h3 b8c8 a4a5 d8c7 a1c1 b7b5 g1h2 d7f5 e1e2 a7a6 d1e1 h7h5 d2g5 e7e5 d5e6 f7e6 h3h4 c7d7 e1d2');
INSERT INTO evaluations VALUES(5343,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(5343,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5343,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(5343,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5343,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5343,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5343,36,'cp 4','e7e6 d2d3 g8e7 e1g1 e8g8 c1g5 a8b8 a2a4 b7b6 d1d2 d7d6 g5h6 g7h6 d2h6 c8b7 h6d2 d6d5 e4d5 e7d5 c3d5 e6d5 e2f4 d5d4 a1e1 d8d6 f4d5 b8e8 d2f4 d6f4 d5f4 g8g7 h2h4 e8c8 e1e2 c8e8 f1e1 e8e2 e1e2 f8b8 e2e1 c6b4 e1e7 b7g2 g1g2 b4c2 e7a7 c2b4');
INSERT INTO evaluations VALUES(5132,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5132,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5132,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(5132,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(5132,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5132,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(5132,36,'cp 0','e1g1 d7d6 d2d3 g8e7 c1g5 h7h6 g5e3 c6d4 e3d2 d4c6');
INSERT INTO evaluations VALUES(5131,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(5131,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5131,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5131,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5131,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5131,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5131,36,'cp 0','g8e7 d2d3 d7d6 c1g5 h7h6 g5e3 c6d4 d1d2 c8d7 a1e1 d8a5 e3f4 a5c7 f4e3');
INSERT INTO evaluations VALUES(8045,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(8045,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(8045,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(8045,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(8045,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(8045,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(8045,36,'cp 77','b1c3 d7d6 f1e2 e8g8 e1g1 c6d4 e3d4 c8d7 d1d3 a7a5 f1e1 d7c6 b2b3 f8e8 a1d1 f6d7 d4g7 g8g7 d3d2 d8b6 d2b2 g7g8 e2f1 d7c5 e1e3 b6b4 c3d5 c6d5 a2a3 b4b6 e4d5 e7e6 d5e6 e8e6 e3e6 c5e6 h2h3 a8e8 b2c3 h7h5 b3b4 a5b4 a3b4 e8a8');
INSERT INTO evaluations VALUES(17360,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17360,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17360,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17360,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17360,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17360,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17360,36,'cp 0','c4b3 c6a5 c2c4 d7d6 e1g1 f8e7 b1c3 g8f6 b3c2 e8g8 h2h3 f6d7 c1e3 a8b8 c2a4 d7b6 a4c2');
INSERT INTO evaluations VALUES(21075,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21075,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21075,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(21075,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21075,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(21075,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21075,36,'cp 3','d3d4 c8f5 e2e3 c6b4 f1b5 c7c6 b5a4 e7e6 a2a3 b4a6 d1e2 a6b8 e1g1 b8d7 c2c4 d7b6 a4b3 d5c4 b3c2 f5c2 e2c2 g8f6 b1d2 f8e7 d2c4 b6c4 c2c4 d8d5 c4c2 a7a5 f1e1 f6e4 f3e5 e4d6 e3e4 d5d4 e5c6 d4c4');
INSERT INTO evaluations VALUES(18737,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(18737,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(18737,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(18737,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(18737,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(18737,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(18737,36,'cp -107','c6e5 f3e2 e5c4 e2c4 d7d5 e4d5 e6d5 c4e2 f8e7 g1f3 g8f6 d2d4 e8g8 e1g1 f8e8 e2d1 c5c4 c1g5 h7h6 g5f6 e7f6 b1d2 c8f5 f1e1 e8e1 d1e1 d8b6 b2b3 c4b3 a2b3 b6c6 d2f1 a7a5 a1a2 f5e6 e1a1 a8c8 a2a5 c6c3 a1c3 c8c3 a5b5');
INSERT INTO evaluations VALUES(19127,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(19127,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(19127,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(19127,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(19127,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(19127,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(19127,36,'cp -105','f3e2 e5c4 e2c4 d7d5 e4d5 e6d5 c4e2 f8e7 g1f3 g8f6 d2d4 e8g8 e1g1 f8e8 e2d1 c8g4 h2h3 g4h5 d4c5 e7c5 c1g5 h7h6 g5f6 d8f6 d1d5 h5f3 d5f3 f6f3 g2f3 a8d8 b1a3 d8d3 b2b4 c5f8 f1e1 e8c8 a1c1 d3c3 c1c3 c8c3 a3b5 c3f3');
INSERT INTO evaluations VALUES(19128,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(19128,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(19128,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(19128,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(19128,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(19128,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(19128,36,'cp -115','e5c4 e2c4 d7d5 e4d5 e6d5 c4e2 f8e7 g1f3 g8f6 e1g1 e8g8 d2d4 f8e8 e2d1 c8f5 b1d2 c5c4 f1e1 e7d6 e1e8 d8e8 d2f1 f5d3 c1e3 e8b5 d1d2 f6e4 d2c1 f7f6 f3d2 e4d2 c1d2 a7a5 a2a4 b5b3 h2h4 a8e8 f1g3 g8f7 h4h5 d6c7 g3e2 b7b5 e2f4 c7f4 e3f4 b5a4');
INSERT INTO evaluations VALUES(19136,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(19136,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(19136,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(19136,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(19136,34,'cp -108',NULL);
INSERT INTO evaluations VALUES(19136,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(19136,36,'cp -107','e2c4 d7d5 e4d5 e6d5 c4e2 f8e7 g1f3 g8f6 d2d4 e8g8 e1g1 f8e8 e2d1 c8g4 c1e3 g4f3 d1f3 d8b6 d4c5 b6b2 b1d2 a8c8 d2b3 b2c3 f1d1 c3c4 a1c1 c4a6 h2h3 e8d8 e3g5 d8d7 g5e3 g7g6 d1e1 f6e4 e3d4 a6c6 d4b2');
INSERT INTO evaluations VALUES(19133,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(19133,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(19133,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(19133,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(19133,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(19133,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(19133,36,'cp -126','d7d5 e4d5 e6d5 c4e2 f8e7 d2d4 g8f6 g1f3 e8g8 e1g1 f8e8 e2d1 c8g4 c1e3 c5c4 b1d2 e7d6 e3g5 h7h6 g5h4 d6f4 f1e1 d8b6 d1c2 g4h5 f3e5 f4e5 d4e5 h5g6 c2d1 f6d7 d2c4 d5c4 d1d7 b6b2 d7d4 b7b5 a2a3 a7a5 h2h3 b5b4 a3b4 a5b4 a1a8');
INSERT INTO evaluations VALUES(20773,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(20773,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(20773,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(20773,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(20773,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(20773,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(20773,36,'cp -97','e4d5 e6d5 c4e2 f8e7 g1f3 g8f6 d2d4 e8g8 e1g1 f8e8 e2d1 c8g4 c1e3 g4f3 d1f3 d8b6 d4c5 b6b2 b1d2 a8c8 d2b3 b2c3 f1d1 c3b4 g2g3 h7h6 g1g2 b4a3 f3f5 a3a6 f5d3 a6c6 b3a5 c6c7 a5b3 b7b6 c5b6 a7b6 a2a4 c7b7');
INSERT INTO evaluations VALUES(45411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45411,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45411,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45411,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45411,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(45411,36,'cp 0','f6e4 g5d2 e4d2 d1d2 b8c6 e2e3 e7e6 f1d3 c8d7 f2f4 c6b4 d3b1 a8c8 g1f3 b4c6 b1d3 f8e7 e1g1 c6b4 d3b1 e8g8 g2g4 b4c6 b1d3 c6b4');
INSERT INTO evaluations VALUES(52906,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(52906,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(52906,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(52906,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(52906,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(52906,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(52906,36,'cp 125','c3e2 c6c5 c2c3 d4c3 b2c3 g8f6 e2g3 b8c6 f1c4 e7e6 e1g1 c6a5 c4e2 f8e7 d2d4 e8g8 e2d3 b7b6 d1e2 c8b7 f1d1 a8c8 c1f4 c5d4 c3d4 a5c6 d3c4 c6b4 f3e5 b4c6 e5c6 b7c6 d4d5 e6d5 e4e5 f6e8 g3f5');
INSERT INTO evaluations VALUES(52909,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(52909,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(52909,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(52909,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(52909,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(52909,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(52909,36,'cp 107','c6c5 c2c3 d4c3 b2c3 g8f6 e2g3 e7e6 f1c4 b8c6 e1g1 f8e7 d2d4 e8g8 f1e1 b7b6 a2a3 c6a5 c4a2 c8b7 c1f4 f8e8 d1e2 c5d4 c3d4 a8c8 e2d3 a7a6 a1d1 b6b5 d4d5 e6d5 e4d5 a5c4 d5d6 e7f8 e1e8 d8e8');
INSERT INTO evaluations VALUES(53217,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(53217,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(53217,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(53217,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(53217,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(53217,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(53217,36,'cp 111','c2c3 d4c3 b2c3 g8f6 e2g3 a7a6 d2d4 e7e6 f1d3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 b8c6 e1g1 e8g8 d3c2 c8d7 f1b1 d8a5 d2e3 c6b4 e3e1 a5c7 e1b4 c7c2 b1c1 c2a4 b4b7 f8b8 b7c7 b8c8 c7d6 h7h6 h2h3 d7c6 d6f4 c6e8 c1b1 c8c3 b1b8 a8b8 f4b8 c3c2');
INSERT INTO evaluations VALUES(53213,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(53213,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(53213,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(53213,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(53213,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(53213,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(53213,36,'cp 113','d4c3 b2c3 g8f6 e2g3 b8c6 a1b1 e7e5 f1b5 d8c7 d2d4 c5d4 c3d4 e5d4 e1g1 f8e7 f3d4 e8g8 d1c2 e7d6 d4c6 b7c6 b5d3 c6c5 f2f4 c5c4 d3e2 d6e7 e4e5 f6d5 c2c4 c8e6 c4c7 d5c7 a2a4 f8b8 c1e3 b8b1 f1b1 c7d5');
INSERT INTO evaluations VALUES(8529,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(8529,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(8529,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(8529,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(8529,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(8529,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(8529,36,'cp 62','b1c3 e7g5 c4d5 g5e7 d1h5 e8g8 d5c6 b7c6 h5e5 a8b8 e1g1 f7f5 e5e2 e7d6 a1b1 c6c5 c1d2 c8b7 e2e6 g8h8 e6c4 f8f6 f2f3 f6g6 g1h1 b7c8 b1e1 b8b4 c4f7 c8d7 c3d5 b4b2 d2f4 b2c2 f4d6 c7d6 e1e7 g6g2 d5f4 g2h2 h1g1 d8g8 e7d7 g8f7 d7f7 h8g8 f7a7 c2a2 a7a2 h2a2');
INSERT INTO evaluations VALUES(14388,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(14388,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(14388,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(14388,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(14388,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(14388,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(14388,36,'cp 117','f1b5 e7e6 e1g1 f8e7 f1e1 e8g8 a2a3 a7a5 b5d3 b7b6 c1f4 c8b7 c3b5 e7d6 b5d6 c7d6 f4g5 h7h6 g5h4 c6b8 a1c1 b8d7 h4g3 d8c7 f3d2 c7c6 d2e4 e6e5 e4f6 d7f6 f2f3 e5d4 g3f2 c6c5 d3f1 f8e8 e1e8 a8e8 d1d4 c5d4 f2d4');
INSERT INTO evaluations VALUES(14173,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(14173,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(14173,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(14173,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(14173,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(14173,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(14173,36,'cp 121','e7e6 e1g1 f8e7 a2a3 e8g8 f1e1 f8e8 c1f4 c8d7 d1d2 e7d6 f4g5 h7h6 g5h4 d6e7 a1d1 f6d5 h4g3 d5c3 d2c3 e7d6 b5d3 c6e7 f3e5 d6e5 g3e5 d7c6 c3a5 b7b6 a5d2 e7f5 d2f4 f5d6 f4g3 d8g5 g3h3 c6b5 d3b5 d6b5 h3d3');
INSERT INTO evaluations VALUES(17352,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(17352,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(17352,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(17352,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(17352,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(17352,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(17352,36,'cp -36','b7b5 c4b3 c6a5 e1g1 c8b7 f1e1 f8e7 c3e2 g8f6 e2g3 d7d6 c2c3 a5b3 a2b3 e8g8 d3d4 f8e8 d1e2 d8c7 h2h3 c5d4 f3d4 b5b4 c3c4 f6d7 c1e3 d7c5 e2g4 e7f6 a1d1 a8d8 h3h4 f6e5 h4h5 e5g3 g4g3 c5e4');
INSERT INTO evaluations VALUES(18110,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(18110,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(18110,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(18110,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(18110,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(18110,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(18110,36,'cp -26','c4b3 c6a5 e1g1 c8b7 f1e1 d7d6 c3e2 f8e7 e2g3 g8f6 h2h3 e8g8 c2c3 a5b3 a2b3 f8e8 c1d2 a6a5 d3d4 d8b6 d4c5 b6c5 d1e2 e8c8 d2e3 c5c7 e2b5 b7e4 g3e4 f6e4 e3b6 c7b7 f3d4 e4c5 b6c5 b7b5 d4b5 c8c5 c3c4 h7h6');
INSERT INTO evaluations VALUES(18125,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(18125,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(18125,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(18125,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(18125,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(18125,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(18125,36,'cp -31','c6a5 e1g1 c8b7 f1e1 f8e7 h2h3 d7d6 c3e2 g8f6 e2g3 e8g8 c2c3 a5b3 a2b3 d8c7 d1e2 f8e8 c1d2 f6d7 d3d4 a8c8 e1c1 c7b6 d2e3 d7f6 d4c5 d6c5 f3d2 e8d8 b3b4 b6c6 e3c5 e7c5 b4c5 c6c5 c1d1 c5g5 d2b3');
INSERT INTO evaluations VALUES(32462,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(32462,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(32462,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(32462,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(32462,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(32462,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(32462,36,'cp 27','b8d7 d2d3 g7g6 c2c4 g8f6 c1e3 f8g7 e1g1 f6g4 e3g5 g4f6 b1c3 e8g8 h2h3 c7c6 d1c2 d7c5 a1d1 c5e6 g5e3 f6h5 d3d4 e5d4 f3d4 g7d4 e3d4 h5f4 g1h1 d8g5 f2f3 e6d4 d1d4 f8e8 f1d1 a8d8 d4d6 d8d6 d1d6 e8d8 d6d8 g5d8');
INSERT INTO evaluations VALUES(27298,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(27298,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(27298,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(27298,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(27298,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(27298,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(27298,36,'cp 23','c2c4 g7g6 e1g1 g8f6 d2d3 f8g7 b1c3 e8g8 c1e3 c7c6 b2b4 d6d5 e4d5 c6d5 c4d5 d7b6 e3c5 f8e8 d5d6 b6d5 c3d5 f6d5 d1b3 d5f4 g2g3 d8d7 f3d2 f4e6 d2c4 b7b6 c5e3 b6b5 c4d2 d7d6 d2e4 d6d7 a2a4 e8d8 f1b1 b5a4 b3a4 d7a4 a1a4 d8d3');
INSERT INTO evaluations VALUES(42590,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(42590,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(42590,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(42590,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(42590,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(42590,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(42590,36,'cp 100','g1f3 d7d6 d2d4 e5d4 e3d4 c5b6 h2h3 e8g8 f1d3 f8e8 c1e3 c7c6 e1g1 h7h6 e3f4 d6d5 c4c5 b6c7 f4c7 d8c7 b2b4 b8d7 f1e1 e8e1 d1e1 d7f8 e1e3 f8e6 a1e1 c7f4 b4b5 c8d7 a2a4 b7b6 f3e5 d7e8 e3f4 e6f4 e1b1 f4d3 e5d3 f6e4 c3e4 d5e4');
INSERT INTO evaluations VALUES(42592,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(42592,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(42592,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(42592,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(42592,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(42592,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(42592,36,'cp 85','d7d6 d2d4 e5d4 e3d4 c5b6 h2h3 e8g8 f1d3 f8e8 c1e3 b8d7 e1g1 d7f8 d1c2 h7h6 b2b4 c7c6 a2a4 b6c7 a4a5 d6d5 c4d5 c6d5 f3e5 a7a6 c2b3 c7e5 d4e5 e8e5 e3d4 e5e8 d4f6 d8f6 c3d5 f6d8 d5b6 c8e6 b3a3 a8b8 b4b5 d8f6 f1e1 a6b5 d3b5 e8e7');
INSERT INTO evaluations VALUES(41884,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(41884,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(41884,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(41884,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(41884,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(41884,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(41884,36,'cp 92','d2d4 e5d4 e3d4 c5b6 h2h3 b8c6 c1e3 c6e7 e3g5 e7g6 d1e2 d8e7 e2e7 g6e7 f1d3 e8g8 e1g1 b6a5 f1c1 e7g6 a1b1 a5c3 c1c3 f8e8 g5f6 g7f6 d4d5 c8d7 g1f1 g6f4 d3c2 b7b5 b2b3 b5c4 b3c4 g8g7');
INSERT INTO evaluations VALUES(41890,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(41890,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(41890,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(41890,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(41890,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(41890,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(41890,36,'cp 93','e5d4 e3d4 c5b6 h2h3 b8c6 c1e3 c6e7 e3g5 e7g6 d1e2 d8e7 e2e7 g6e7 f1d3 e8g8 e1g1 b6a5 g5f6 g7f6 c3e2 c7c6 f1b1 a7a6 e2g3 e7g6 g3h5 f6f5 b2b4 a5d8 a2a4 a6a5 b4b5 c6c5 h5g3 c5d4 d3f5 d4d3 b1d1 d8f6 f5c8 f8c8');
INSERT INTO evaluations VALUES(41904,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(41904,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(41904,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(41904,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(41904,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(41904,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(41904,36,'cp 80','e3d4 c5b6 h2h3 e8g8 f1d3 f8e8 c1e3 h7h6 e1g1 b8d7 d1d2 d7f8 a1e1 c8d7 a2a3 c7c6 e3f4 b6c7 d3c2 a7a6 e1e8 d7e8 f1e1 f8g6 f4g3 b7b5 b2b3 e8d7 c4c5 d6c5 g3c7 d8c7 d4c5 a8e8 b3b4 e8e1 d2e1 c7f4 c3e2 f4c4 e2d4 c4d5 c2g6 f7g6');
INSERT INTO evaluations VALUES(41903,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(41903,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(41903,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(41903,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(41903,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(41903,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(41903,36,'cp 75','c5b6 h2h3 e8g8 f1d3 f8e8 c1e3 h7h6 e1g1 b8d7 d1d2 d7f8 a1e1 c7c6 e3f4 e8e1 f1e1 c8e6 b2b4 d8d7 f4g3 f6h5 g3h2 h5f6 a2a4 a8e8 d3f1 a7a6 d4d5 c6d5 c4d5 e6f5 e1e8 f6e8 b4b5 d7c8 f3d4 f5d7 b5a6 b7a6 d4c6 d7c6 d5c6 c8c6');
INSERT INTO evaluations VALUES(55296,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(55296,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(55296,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(55296,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(55296,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(55296,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(55296,36,'cp 87','c7c5 e2e4 a7a6 d2d4 c5d4 d1d4 d8d4 f3d4 e7e6 d4b3 b8c6 c1e3 g8f6 f2f3 e6e5 f1c4 b7b5 c4d5 c8d7 d5c6 d7c6 a1c1 f8e7 c3e2 a8c8 e1f2 e8g8 e3c5 e7c5 c1c5 f6d7 c5c2 d7b6 h1c1 b6c4 b3d2 c4d2 c2d2 b5b4 e2g3 c6b5 c1c8 f8c8 g3f5 g8h8 f5e7');
INSERT INTO evaluations VALUES(23475,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(23475,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23475,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23475,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(23475,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(23475,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(23475,36,'cp 13','e7e5 c2c4 g8f6 b1c3 f8c5 a2a3 d7d6 e2e3 c5b6 d2d3 e8g8 g1e2 c6e7 c3a4 c7c6 a4b6 d8b6 e1g1 d6d5 b2b3 c8e6 d1c2 f8e8 c1b2 f6d7 b2c3 a7a5 f1b1 b6c7 h2h3 h7h6 c2d2 b7b6 f2f4 a5a4 f4e5 a4b3 b1b3');
INSERT INTO evaluations VALUES(14393,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(14393,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(14393,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(14393,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(14393,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(14393,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(14393,36,'cp 210','e7e6 a2a3 a7a6 c1e3 f8e7 h2h3 e8g8 d1c2 d8e8 f1d3 g8h8 a1d1 b7b5 e4e5 f6d5 c3d5 e6d5 e1g1 c8d7 d1c1 c6a5 e3g5 e7g5 f3g5 h7h6 g5f3 a8c8 b2b3 a5c6 c2e2 c6e7 d3b1 d7e6 c1c5');
INSERT INTO evaluations VALUES(14927,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(14927,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(14927,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(14927,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(14927,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(14927,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(14927,36,'cp 209','a2a3 a7a6 h2h3 b7b5 f1d3 c8b7 c1e3 f6d7 d1c2 f8e7 e1g1 e8g8 a1d1 g8h8 c3e2 a8c8 b2b4 d8e8 e2g3 c6d8 f1e1 f7f6 c2e2 e7d6 d3b1 d8f7 d4d5 f7e5 f3e5 d7e5 d5e6 e8e6 g3f5 e6c4 f5d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(14929,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(14929,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(14929,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(14929,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(14929,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(14929,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(14929,36,'cp 216','a7a6 h2h3 b7b5 f1d3 c8b7 c1e3 f8e7 d1c2 f6d7 e1g1 e8g8 a1d1 g8h8 b2b4 d8e8 c2b3 c6d8 f1e1 f7f6 d4d5 d7e5 f3e5 f6e5 a3a4 b5a4 c3a4 a6a5 b4b5 e7b4 e1e2 e6d5 e4d5 d8f7 a4c5 b7c8 c5e6 c8e6 d5e6 f7d6');
INSERT INTO evaluations VALUES(46987,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46987,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(46987,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(46987,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(46987,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(46987,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(46987,36,'cp 57','d2d4 g7g6 e2e4 f8g7 f1e2 e8g8 h2h3 e7e5 d4d5 b8a6 c1g5 a6c5 d1c2 a7a5 g1f3 c8d7 f3d2 a5a4 e1g1 c5a6 c2d1 h7h6 g5e3 a4a3 b2b4 a6b4 a1b1 c7c5 d5c6 b4c6 d2f3 c6a5 c4c5 d6c5 e3c5 d7e6 c5f8 d8f8 f3e5');
INSERT INTO evaluations VALUES(6556,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(6556,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(6556,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(6556,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(6556,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(6556,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(6556,36,'cp -33','d7d6 c2c3 c8g4 d1b3 g4f3 c4f7 e8f8 g2f3 d4d3 f7g8 h8g8 c1e3 g7g5 b1d2 c6e5 g1h1 g8g6 f3f4 e5g4 f4g5 c5e3 f2e3 f8g7 b3b7 a8b8 b7a7 b8b2 a7d4 g4e5 f1f5 d8a8 f5e5 d6e5 d4e5 g7g8 d2f3 b2b8 e5c7 b8c8 c7e5 c8e8 e5d5 a8d5 e4d5 e8e3 h1g2 g6g7 a1d1 g7d7 c3c4 d7f7 f3d2 f7f5 h2h4 g8f7 d2f1 e3e2 g2g1 f5f4 d1d3 f4c4');
INSERT INTO evaluations VALUES(6043,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(6043,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(6043,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(6043,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(6043,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(6043,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(6043,36,'cp -14','c2c3 c8g4 d1b3 g4f3 c4f7 e8f8 g2f3 g8f6 g1h1 c6a5 b3a4 f8f7 a4a5 h8e8 c3d4 c5d4 a5a4 c7c5 c1e3 d6d5 b1c3 d4c3 b2c3 d8c8 f1g1 b7b6 e3g5 c8e6 a1d1 d5e4 g5f6 e6f6 d1d7 e8e7 d7e7 f6e7 g1g4 f7g8 g4e4 e7f7 h1g2 g8h8 e4f4 f7d5 a4c2 a8d8 c2e2 d5d3');
INSERT INTO evaluations VALUES(6041,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(6041,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(6041,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(6041,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(6041,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(6041,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(6041,36,'cp -16','c8g4 d1b3 g4f3 c4f7 e8f8 g2f3 d4d3 f7g8 h8g8 c1e3 c6e5 b1d2 g7g5 g1h1 g8g6 f3f4 e5g4 f4g5 c5e3 f2e3 f8g7 d2f3 g4e3 f1g1 d8f8 g1g3 f8f4 b3b4 a8e8 b4d4 g7g8 d4d3 e8e4 a1g1 c7c6 b2b3 b7b5 f3d2 e4e7 g3f3 f4e5 d2f1 e3f1 f3f1 e5e4 d3e4 e7e4');
INSERT INTO evaluations VALUES(23879,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(23879,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(23879,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(23879,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(23879,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(23879,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(23879,36,'cp -6','d1b3 g4f3 c4f7 e8f8 g2f3 g8f6 g1h1 c6a5 b3a4 f8f7 a4a5 h8e8 c3d4 c5d4 a5b4 c7c5 b4b7 e8e7 b7b3 d6d5 c1f4 d8b6 b1c3 b6b3 a2b3 d5e4 f3e4 f6e4 a1a5 a8d8 c3e4 e7e4 a5a7 f7g8 f4e3 d4e3 f1e1 g7g6 e1e3 e4e3 f2e3 d8d3 a7c7 d3e3 c7c8 g8f7 c8c5 e3e2 c5c7 f7e6 c7h7 e2b2');
INSERT INTO evaluations VALUES(10162,30,'cp -184',NULL);
INSERT INTO evaluations VALUES(10162,31,'cp -180',NULL);
INSERT INTO evaluations VALUES(10162,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(10162,33,'cp -176',NULL);
INSERT INTO evaluations VALUES(10162,34,'cp -183',NULL);
INSERT INTO evaluations VALUES(10162,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(10162,36,'cp -213','f6g4 d3d4 e5d4 c3a4 d7d6 a4c5 d6c5 h2h3 g4e5 f3e5 c6e5 d1h5 d8e7 c1g5 e7e6 g5f4 e5c6 f4c7 e6e7 c7g3 e8g8 f1d3 f7f5 f2f3 c8d7 e1c1 c6b4 d3b1 e7e6 b2b3 b7b5 c4b5 c5c4 d1d4 c4b3 d4b4 b3a2 c1b2 a2b1q b2b1 e6e7 b4b2');
INSERT INTO evaluations VALUES(10631,30,'cp -184',NULL);
INSERT INTO evaluations VALUES(10631,31,'cp -198',NULL);
INSERT INTO evaluations VALUES(10631,32,'cp -216',NULL);
INSERT INTO evaluations VALUES(10631,33,'cp -205',NULL);
INSERT INTO evaluations VALUES(10631,34,'cp -208',NULL);
INSERT INTO evaluations VALUES(10631,35,'cp -210',NULL);
INSERT INTO evaluations VALUES(10631,36,'cp -208','d3d4 e5d4 c3a4 d7d6 a4c5 d6c5 h2h3 g4e5 f3e5 c6e5 d1h5 d8e7 c1g5 e7e6 g5f4 e5c6 e1c1 e6e7 f4g3 e8g8 f1d3 f7f6 e4e5 f6f5 f2f4 c6b4 d3b1 c8e6 g3e1 e6c4 e1b4 c5b4 b1f5 g7g6 f5g6 h7g6 h5g6 e7g7 g6g7 g8g7 d1d4 c4a2 d4b4 a2d5 b4d4 d5e6');
INSERT INTO evaluations VALUES(4404,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(4404,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(4404,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(4404,33,'cp 229',NULL);
INSERT INTO evaluations VALUES(4404,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(4404,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(4404,36,'cp 192','c3d4 c5e7 e4e5 f6e8 d4d5 c6b8 f1e1 d7d6 e5e6 f7e6 d5e6 c7c6 d1c2 g8h8 h2h4 e7h4 b1d2 e8c7 f3h4 d8h4 d2f3 f8f3 g2f3 h4f6 e6e7 c7e8 c4d3 b8d7 f3f4 f6h4 c2e2 d7f6 e2f3 c8g4 f3g2 h4h5 g2h1 h5h1 g1h1 g4f3 h1g1 f3d5 f4f5 h8g8 b2b4 h7h6 c1f4');
INSERT INTO evaluations VALUES(4402,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(4402,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(4402,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(4402,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(4402,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(4402,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(4402,36,'cp 256','c5e7 d4d5 c6b8 e4e5 f6e8 f1e1 d7d6 e5e6 f7e6 d5e6 c7c6 d1c2 g8h8 h2h4 f8f3 g2f3 e7h4 c2f5 d8f6 f5g4 f6e7 g1g2 h4f6 b1d2 d6d5 c4d3 g7g6 f3f4 b8a6 e1h1 c8e6 g4g6 e6g8 d2f3 a6c5 d3c2 e7g7 g6g7 e8g7 c1e3 c5e4 f3g5 g7f5 g5e4 d5e4 c2e4');
INSERT INTO evaluations VALUES(16989,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(16989,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(16989,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(16989,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(16989,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(16989,35,'cp 235',NULL);
INSERT INTO evaluations VALUES(16989,36,'cp 229','f3e5 d7d5 c4d5 f6d5 c3d5 c7c6 d5e3 d8e7 f2f4 f7f6 d1h5 g7g6 e5g6 d4c2 e1f2 e7f7 e3c2 h8g8 h5f3 h7g6 d2d4 f8d6 d4d5 c6d5 e4d5 f7h7 c1d2 e8f7 a1c1 c8f5 d2b4 d6b4 c2b4 a8c8 c1c8 g8c8 d5d6 h7h4 g2g3 h4h3 f3b7 f5d7 h1e1 h3h2 b7g2 h2g2 f2g2 d7e6 e1h1 c8c4 h1h7 f7g8 h7e7 c4b4 e7e6');
INSERT INTO evaluations VALUES(16937,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(16937,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(16937,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(16937,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(16937,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(16937,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(16937,36,'cp 233','d8e7 e5f3 d7d5 f3d4 d5c4 d2d3 c8e6 f2f3 a8d8 c1e3 c7c5 d4e6 e7e6 d1e2 c4d3 c2d3 f8e7 e1c1 e8g8 c1b1 d8d7 d3d4 c5d4 d1d4 d7d4 e3d4 a7a6 h1d1 f8d8 e2c2 b7b5 c2b3 e6c6 c3e2 c6c7 e2g3 c7f4 g3f5 e7f8 h2h3 f4h2 b3c2 f6e8 d1c1 h2f4 c2c6');
INSERT INTO evaluations VALUES(30811,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(30811,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(30811,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(30811,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(30811,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(30811,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(30811,36,'cp 50','d7d5 b1c3 e7e6 e4d5 b7d5 c3d5 d8d5 f3f4 g8f6 f1e2 b8d7 e2f3 f6e4 g1h3 f7f5 c1e3 f8e7 e1g1 d7f6 d1e2 e8g8 c2c4 d5d7 a1d1 a8d8 h3f2 d7c8 f3e4 f6e4 f2e4 f5e4 g2g3 c7c5 f1e1 c5d4 d1d4 e7b4 e1d1 b4e7 e3f2 c8c6 d1e1 d8d4 f2d4 f8d8 d4c3 e7f6 c3f6 g7f6 e2e4 c6e4');
INSERT INTO evaluations VALUES(30550,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(30550,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(30550,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(30550,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(30550,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(30550,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(30550,36,'cp 50','b1c3 e7e6 e4d5 b7d5 c3d5 d8d5 f3f4 g8f6 f1e2 b8d7 e2f3 f6e4 g1e2 f7f5 e1g1 d7f6 e2c3 d5d7 f1e1 e4c3 b2c3 a8d8 a2a4 e8f7 d1e2 c7c5 c1e3 f6d5 f3d5 d7d5 a4a5 c5d4 a1d1 f8c5 c3d4 c5e7 c2c4');
INSERT INTO evaluations VALUES(50077,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(50077,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(50077,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(50077,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(50077,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(50077,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(50077,36,'cp 59','d2d4 e7e6 a2a3 g7g6 g1f3 d7d6 b2b4 f8g7 e2e3 e8g8 f1e2 b7b6 c1b2 b8d7 a1c1 d8e7 e1g1 c7c5 d4d5 f6e4 d5e6 e4c3 b2c3 d7e5 b4c5 b6c5 c1b1 c8e6 f3e5 d6e5 b1b5 f8b8 d1c2 e6d7 b5b8 a8b8 f1b1 b8b6 c3a5 b6b1 c2b1');
INSERT INTO evaluations VALUES(8315,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(8315,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(8315,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(8315,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(8315,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(8315,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8315,36,'cp 42','f2f3 c8d7 d1d2 e8g8 e1c1 a8c8 c4b3 c6d4 e3d4 b7b5 a2a3 a7a5 e4e5 d6e5 d4e5 b5b4 e5f6 e7f6 d2d7 b4c3 b3f7 g8h8 d7d8 c8d8 f7d5 g7h6 c1b1 h6d2 d5c4 f6f5 b2c3 d8b8 b1a2 d2c3 d1b1 f8d8 b1b3 b8b3 a2b3 c3f6 h1e1 d8d2');
INSERT INTO evaluations VALUES(8317,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(8317,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(8317,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(8317,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8317,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(8317,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(8317,36,'cp 41','c8d7 d1d2 e8g8 e1c1 a8c8 c4b3 c6d4 e3d4 b7b5 a2a3 a7a5 e4e5 d6e5 d4e5 b5b4 e5f6 e7f6 d2d7 b4c3 b3f7 g8h8 d7d8 c8d8 d1d8 c3b2 c1b2 f8d8 h1e1 d8b8 b2a2 f6f5 e1e8 b8e8 f7e8 g7e5 a2b3 h8g7 c2c4 e5h2 b3a4 g7f8 e8b5 h2c7 b5a6 h7h5 a4b5 h5h4 b5c6 c7f4 c4c5 g6g5 c6b6 f8e7 b6a5');
INSERT INTO evaluations VALUES(24005,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(24005,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(24005,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24005,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(24005,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(24005,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(24005,36,'cp 50','d1d2 e8g8 e1c1 a8c8 c4b3 c6d4 e3d4 b7b5 a2a3 a7a5 e4e5 d6e5 d4e5 b5b4 e5f6 e7f6 d2d7 b4c3 b3f7 g8h8 d7d8 c3b2 c1b2 f8d8 d1d8 c8d8 b2b3 g7f8 f7c4 d8d2 g2g4 d2d4 c4b5 f6f5 g4f5 g6f5 h1e1 h8g7 e1e8 d4f4 b5c6 a5a4 c6a4 f4f3 c2c3');
INSERT INTO evaluations VALUES(24006,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(24006,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(24006,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(24006,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(24006,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(24006,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(24006,36,'cp 28','e8g8 e1c1 a8c8 c4e2 a7a6 h2h4 d8c7 a2a3 b7b5 h4h5 f6h5 d4c6 d7c6 e3h6 h5f6 g2g4 g7h8 h6f8 c8f8 c1b1 f8b8 g4g5 f6h5 c3d5 c6d5 e4d5 c7b6 h1h4 b6f2 h4e4 h8e5 e4e5 d6e5 d5d6 e7d6 d2d6 b8c8 d6e5 h5g3 e2d3 f2f3');
INSERT INTO evaluations VALUES(11522,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(11522,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(11522,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(11522,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(11522,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(11522,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(11522,36,'cp 29','e1g1 f8c5 c2c3 d7d6 a2a4 h7h6 b2b4 c5a7 c1e3 a7e3 f2e3 e8g8 b1d2 c6e7 d3d4 e7g6 c4d3 b7b6 d1c2 c8d7 a4a5 b6b5 c3c4 e5d4 e3d4 g6f4 c4c5 d8e7 a1e1 a8e8 h2h3 f6h5 f1f2 f4d3 c2d3 d6c5 b4c5 h5f4 d3e3');
INSERT INTO evaluations VALUES(11520,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(11520,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(11520,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(11520,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(11520,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(11520,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(11520,36,'cp 13','f8c5 a2a4 d7d6 c2c3 c5a7 c1e3 a7e3 f2e3 c6e7 b1d2 e8g8 h2h3 e7g6 f1f2 a6a5 d3d4 d8e7 c4d3 c8d7 d1c2 c7c6 c3c4 a8e8 a1e1 b7b6 c2b3 e8b8 b3c3 h7h6 b2b3 f8e8 d2f1 d6d5 e4d5 c6d5 d3g6 f7g6 c4d5');
INSERT INTO evaluations VALUES(7308,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7308,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(7308,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(7308,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(7308,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7308,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(7308,36,'cp 16','b1c3 d7d6 c3d5 f6d5 c4d5 e8g8 c2c3 d8f6 h2h3 h7h6 d5b3 c5a7 c1e3 c8e6 b3e6 f6e6 e3a7 c6a7 f1e1 a7c6 d3d4 c6e7 d1b3 e6b3 a2b3 f7f6 b3b4 g7g5 g1f1 g8f7 e1d1 e7c6 f1e2 a6a5 b4a5 c6a5 d4e5 d6e5 f3d2 h6h5');
INSERT INTO evaluations VALUES(7302,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7302,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(7302,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(7302,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(7302,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(7302,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(7302,36,'cp 6','d7d6 c3d5 f6d5 c4d5 e8g8 c2c3 d8f6 a2a4 h7h6 d5b3 c6e7 h2h3 c8e6 c1e3 c5e3 f2e3 f6g6 g1h2 a6a5 d1c2 e6b3 c2b3 b7b6 b3c4 a8c8 f1f2 g6e6 c4e6 f7e6 h2g1 f8f7 a1f1 f7f8 f3d2 e7g6 f2f8 c8f8 f1f8 g6f8');
INSERT INTO evaluations VALUES(14623,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(14623,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14623,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14623,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14623,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14623,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(14623,36,'cp 28','e1g1 f8e7 a2a4 e8g8 h2h3 c8e6 c4e6 f7e6 b2b4 a7a6 f1e1 f6d7 b1d2 d8e8 d2f1 d6d5 c1d2 e7d6 f1h2 d7f6 e4d5 e6d5 f3e5 d6e5 f2f4 e5d4 c3d4 e8g6 h2f3 g6d3 e1e3 d3f5 b4b5 a6b5 a4b5 a8a1 d1a1 c6d8 a1e1 f5d7 d2b4 f8e8 e3e8 f6e8 f3e5 d7b5');
INSERT INTO evaluations VALUES(14622,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14622,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14622,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(14622,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(14622,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(14622,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(14622,36,'cp 33','f8e7 a2a4 e8g8 h2h3 c8e6 c4e6 f7e6 b2b4 a7a6 b1d2 d8e8 d2c4 f6d7 c1d2 d6d5 c4e3 d7f6 e4d5 e6d5 b4b5 a6b5 a4b5 a8a1 d1a1 c6d8 f3e5 e7d6 d3d4 e8b5 c3c4 d5c4 f1b1 b5a6 a1a6 b7a6 e3c4 f8e8 e5f3');
INSERT INTO evaluations VALUES(8444,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(8444,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8444,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(8444,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(8444,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(8444,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(8444,36,'cp 30','a2a4 a7a5 d1b3 e8g8 h2h3 f6h5 c1e3 e7g5 b1d2 d8f6 d3d4 f6g6 f3g5 h6g5 f1e1 g6f6 c4f1 e5d4 c3d4 c6d4 e3d4 f6d4 d2f3 d4b4 b3b4 a5b4 e1c1 c7c6 f3g5 h5f4 a4a5 c8d7 g2g3 f4e6 g5e6 d7e6 f2f4 d6d5 e4e5');
INSERT INTO evaluations VALUES(21329,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(21329,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(21329,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(21329,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(21329,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(21329,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(21329,36,'cp 128','g8e7 d4c3 c8e6 c4e6 f7e6 c3b3 g7g5 b3b5 c7c6 b5b7 f8g7 b7b3 d8d7 c1e3 e8g8 c2c3 a8b8 b3c2 c6c5 b1d2 e7g6 d2c4 g6e5 c4e5 g7e5 f2f3 a7a5 f1f2 b8b7 a1f1 a5a4 f3f4 g5f4 e3f4 e5f4 f2f4 f8f4 f1f4 d7g7 c2d2 g7e5 h2h3');
INSERT INTO evaluations VALUES(29589,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29589,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29589,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(29589,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(29589,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(29589,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(29589,36,'cp 14','c2c4 e7e6 f1g2 c7c6 d1b3 d8b6 c4c5 b6b3 a2b3 h7h6 b3b4 a7a6 b1c3 f8e7 c1f4 e8g8 h2h4 b8d7 e2e3 f8c8 a1a4 f6h5 e1g1 h5f4 e3f4 f5g4 f1a1 g4f3 g2f3 e7f6 c3e2 g7g5 f4g5 h6g5 h4g5 f6g5 f3h5');
INSERT INTO evaluations VALUES(44635,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44635,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(44635,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44635,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44635,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44635,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44635,36,'cp 11','e4e5 f6d5 g1f3 b8c6 c3d4 d7d6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 b5c6 d7c6 e1g1 e7e6 e5c6 b7c6 c1e3 f8e7 d1f3 a8c8 f1c1 b6d5 b1c3 d5e3 f2e3 c6c5 d4d5 e8g8 d5e6 f7e6 f3e2 e7f6 c3e4 d8b6 e4f6 f8f6 c1c4 f6f5');
INSERT INTO evaluations VALUES(44604,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(44604,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(44604,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44604,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44604,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(44604,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44604,36,'cp 0','f6d5 g1f3 b8c6 c3d4 d7d6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 e5c6 d7c6 b5c6 b7c6 b1c3 g7g6 d1f3 b6d5 e1g1 f8g7 f1e1 e7e6 c3d5 c6d5 f3a3 d8e7 a3a4 e7d7 a4a3');
INSERT INTO evaluations VALUES(15429,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(15429,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(15429,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(15429,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(15429,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(15429,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(15429,36,'cp 17','g7g6 h2h3 f8g7 e1g1 a7a6 b5a4 h7h6 f1e1 e8g8 d3d4 b7b5 a4c2 f8e8 d4d5 c6e7 a2a4 c8b7 a4b5 a6b5 a1a8 d8a8 b2b3 c7c6 c3c4 b7a6 b1d2 a8b8 c1a3 c6d5 c4d5 b5b4 a3b2 b8c7 d2f1 a6f1 e1f1 e8a8 c2b1 g8h8 d1d2 h8h7');
INSERT INTO evaluations VALUES(22758,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(22758,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(22758,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(22758,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(22758,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(22758,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(22758,36,'cp 37','d4d5 c6e7 e2e4 e7g6 g1f3 g8f6 h2h4 h7h5 g2g3 a7a5 c1e3 f6g4 d1e2 a5a4 f1h3 g4e3 e2e3 b7b6 b1d2 f8c5 e3d3 c7c6 e1g1 c8a6 d2c4 c6d5 e4d5 e8g8 b2b4 a4b3 a2b3 a6c4 d3c4 d8f6 a1a8 f8a8');
INSERT INTO evaluations VALUES(22660,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(22660,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(22660,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(22660,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(22660,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(22660,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(22660,36,'cp 37','c6e7 e2e4 e7g6 g1f3 g8f6 h2h4 h7h5 g2g3 a7a5 c1e3 a5a4 f1d3 f8e7 b1d2 f6g4 d1e2 g4e3 e2e3 d7d6 d3b5 c8d7 b5d7 d8d7 e1c1 d7b5 c1b1 g6f8 f3e1 f8d7 e1d3 c7c6 c3c4 b5a5 f2f4 a4a3 b2b3 c6d5 e4d5 e8g8 d2e4 b7b5 b3b4');
INSERT INTO evaluations VALUES(29815,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(29815,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(29815,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(29815,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(29815,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(29815,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(29815,36,'cp 105','c2c4 g4f3 d1f3 d5b4 f3b3 e7e5 d4d5 a7a5 a2a3 b4a6 b3g3 b8d7 f1d3 d8e7 c1d2 g7g6 b1c3 a6c5 d3c2 f8g7 e1c1 d7b6 d2g5 f7f6 g5e3 f6f5 c2f5 b6c4 e3c5 e7c5 f5e6 h8f8 d5d6 c4d6 c1b1 c5f2 g3h3 e5e4 d1d5 e8e7 h1d1 g7c3 h3c3 e7e6');
INSERT INTO evaluations VALUES(29811,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(29811,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(29811,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(29811,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(29811,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(29811,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(29811,36,'cp 109','d5f6 f1e2 c7c6 b1c3 e7e6 h2h3 g4h5 c1f4 f8e7 e1g1 e8g8 f1e1 h5g6 f3e5 b8d7 e5g6 h7g6 e2f3 f8e8 d1e2 d7f8 a1d1 e7d6 f4e3 d8a5 g2g3 f8d7 e3d2 e6e5 d4d5 a5c7 d5c6 b7c6 c3e4 f6e4 e2e4');
INSERT INTO evaluations VALUES(29750,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(29750,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(29750,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(29750,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(29750,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(29750,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(29750,36,'cp 115','f1e2 c7c6 b1c3 e7e6 h2h3 g4h5 e1g1 f8e7 c1f4 e8g8 a2a3 b8d7 f1e1 h5g6 e2f1 f8e8 d1d2 a7a5 a1d1 h7h6 e1e3 a5a4 f1d3 g6d3 d2d3 d8a5 d3e2 a8d8 f4g3 e7f8 f3e5 d7e5 g3e5 f8d6 f2f4');
INSERT INTO evaluations VALUES(29749,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(29749,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(29749,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(29749,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(29749,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(29749,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(29749,36,'cp 115','c7c6 b1c3 e7e6 h2h3 g4f5 f3h4 f5e4 e1g1 f8e7 c3e4 f6e4 h4f3 d8c7 e2d3 e4f6 a2a3 e8g8 c1d2 c6c5 d4c5 e7c5 b2b4 c5e7 d1e2 a7a5 a1b1 a5b4 a3b4 f8d8 d2c3 b8d7 f1e1 e7f8 f3g5 d8e8 d3c2 h7h6 g5e4 f6e4 e2e4');
INSERT INTO evaluations VALUES(31165,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(31165,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(31165,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(31165,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(31165,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31165,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31165,36,'cp 0','e7e6 f1d3 f5d3 d1d3 c6c5 c2c3 b8c6 g1e2 d8b6 e1g1 g8e7 b1d2 c5d4 c3d4 c6b4 d3b3 b4c6');
INSERT INTO evaluations VALUES(31541,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31541,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31541,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31541,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31541,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31541,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31541,36,'cp 0','f1d3 f5d3 d1d3 c6c5 c2c3 b8c6 g1e2 d8b6 e1g1 g8e7 b1d2 c5d4 c3d4 c6b4 d3b3 b4c6');
INSERT INTO evaluations VALUES(1785,30,'mate -1',NULL);
INSERT INTO evaluations VALUES(1785,31,'mate -1',NULL);
INSERT INTO evaluations VALUES(1785,32,'mate -1',NULL);
INSERT INTO evaluations VALUES(1785,33,'mate -1',NULL);
INSERT INTO evaluations VALUES(1785,34,'mate -1',NULL);
INSERT INTO evaluations VALUES(1785,35,'mate -1',NULL);
INSERT INTO evaluations VALUES(1785,36,'mate -1','d4f3');
INSERT INTO evaluations VALUES(15495,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(15495,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(15495,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(15495,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(15495,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15495,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15495,36,'cp 0','c6a5 c4b3 b7b5 e1g1 f8e7 c1d2 a5b3 a2b3 e8g8 c2c4 c8d7 f1e1 d8b8 b1c3 b5c4 b3c4 b8b2 a1b1 b2a3 b1a1');
INSERT INTO evaluations VALUES(55592,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(55592,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(55592,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(55592,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55592,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(55592,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(55592,36,'cp -49','f7f5 d2d3 g8f6 b1c3 f8e7 g2g3 f5e4 d3e4 c7c6 f1g2 b7b5 e1g1 e8g8 f3e1 a7a5 e1d3 b5b4 c3a4 c6c5 c2c4 b8c6 f2f4 g8h8 c1b2 f6d7 b2c1 c6d4 a4b2 c8b7 d3f2 b7c6 c1e3 d7f6 f4f5');
INSERT INTO evaluations VALUES(15704,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(15704,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(15704,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(15704,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(15704,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(15704,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(15704,36,'cp 75','d1d8 e8d8 f2f3 f6d7 f1e2 f7f6 b1c3 a7a5 f3f4 h7h5 c1d2 d7b6 e1c1 c8g4 h2h3 g4e2 c3e2 d8e8 e2d4 e8f7 e4e5 f8c5 d4b3 c5b4 h1e1 h8d8 e5e6 f7e8 f4f5 b4d2 b3d2 d8d5 g2g4 a8d8 g4h5 d5f5 e1g1 f5g5 g1g5 f6g5 d1g1');
INSERT INTO evaluations VALUES(2252,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(2252,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(2252,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(2252,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(2252,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(2252,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(2252,36,'cp 65','e8d8 f2f3 f6d7 f1e2 a7a5 b1c3 f7f6 c1d2 f8d6 e1c1 h8e8 f3f4 d7c5 h1e1 c8d7 e2f3 d8c8 g2g3 a8a6 d2e3 b7b5 e3d4 c5e6 d4f2 a5a4 e4e5 f6e5 c3e4 c6c5 f4e5 d6e7 f3g4 h7h5 g4h5 e8h8 h5f3 e6d4 f2d4 c5d4 c1b1 a4a3 d1d4');
INSERT INTO evaluations VALUES(2285,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(2285,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(2285,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(2285,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(2285,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(2285,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(2285,36,'cp 70','b1c3 a7a5 f2f3 f6d7 c1d2 f8d6 e1c1 h8e8 h2h4 f7f6 h4h5 d7e5 f1e2 b7b5 a2a4 b5b4 c3b1 c8a6 e2a6 a8a6 d2e3 e5c4 e3f4 d8e7 f4d6 c7d6 b2b3 c4b6 h1h4 d6d5 e4d5 b6d5 d1e1 e7d7 e1e8 d7e8 b1d2 e8f7 h4d4 a6a8 d2c4 f7e6 d4g4 e6f7 c4d6 f7e6');
INSERT INTO evaluations VALUES(2284,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(2284,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(2284,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(2284,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(2284,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(2284,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(2284,36,'cp 59','a7a5 f2f3 f6d7 c1d2 f8d6 e1c1 f7f6 f3f4 h8e8 f1e2 d7c5 e2f3 c8d7 h1e1 a8a6 d2e3 d8c8 g2g3 b7b5 e3d4 a5a4 f3h5 g7g6 h5f3 b5b4 c3b1 d7e6 a2a3 c5d7 a3b4 d6b4 e1e3 e8e7 d4c3 b4c5 e3e1 e7e8 e4e5 f6e5 c3e5 d7e5 e1e5 c5d6');
INSERT INTO evaluations VALUES(43980,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(43980,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(43980,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(43980,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(43980,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(43980,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(43980,36,'cp -33','b8c6 d4c5 f8c5 e1g1 e8g8 c1f4 a7a6 b5c6 b7c6 c3a4 c5e7 c2c4 c8b7 a1c1 c6c5 c4d5 b7d5 b2b3 d5c6 a4b2 d7b6 f4e3 a8c8 d1d8 f8d8 f3d2 d8d5 d2c4 b6c4 b2c4 c8d8 f1e1 g7g5 c4a5 c6a8 a5c4 h7h5 c4b6 d5e5 b6a8 d8a8 e1d1 g5g4 d1d7 a8e8 d7c7 e5f5 h2h3 g4h3');
INSERT INTO evaluations VALUES(12107,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(12107,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(12107,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(12107,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(12107,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(12107,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(12107,36,'cp -31','d4c5 f8c5 e1g1 e8g8 c1f4 a7a6 b5c6 b7c6 c3a4 c5a7 c2c4 a6a5 f1e1 c8a6 c4d5 c6d5 f4e3 d8b8 a1c1 a7e3 e1e3 f8c8 c1c8 b8c8 h2h3 c8c4 a2a3 c4f4 b2b3 h7h6 d1d2 a8c8 d2a5 c8c1 e3e1 a6d3 e1c1 f4c1 g1h2 d3e4 a5d8 d7f8 a4b6 e4f3 b6d7 c1f1 d8f8 g8h7 g2f3 f1f2 h2h1 f2f1 h1h2');
INSERT INTO evaluations VALUES(23032,30,'cp 593',NULL);
INSERT INTO evaluations VALUES(23032,31,'cp 586',NULL);
INSERT INTO evaluations VALUES(23032,32,'cp 589',NULL);
INSERT INTO evaluations VALUES(23032,33,'cp 583',NULL);
INSERT INTO evaluations VALUES(23032,34,'cp 594',NULL);
INSERT INTO evaluations VALUES(23032,35,'cp 601',NULL);
INSERT INTO evaluations VALUES(23032,36,'cp 596','f1b5 a7a6 b5a4 f8c5 g1f3 b7b5 a4c2 d7d6 e1g1 c8d7 g1h1 g8f6 d1e1 e8g8 c1e3 c5e3 e1e3 f6g4 e3e2 d8f6 b1d2 f6h6 h2h3 a8b8 f1e1 g4f6 d2f1 d7g4 f1g3 g4d7 e2e3 h6e3 e1e3 g8h8 a2a4 b5a4');
INSERT INTO evaluations VALUES(23001,30,'cp 581',NULL);
INSERT INTO evaluations VALUES(23001,31,'cp 578',NULL);
INSERT INTO evaluations VALUES(23001,32,'cp 584',NULL);
INSERT INTO evaluations VALUES(23001,33,'cp 586',NULL);
INSERT INTO evaluations VALUES(23001,34,'cp 578',NULL);
INSERT INTO evaluations VALUES(23001,35,'cp 571',NULL);
INSERT INTO evaluations VALUES(23001,36,'cp 576','f8c5 g1f3 a7a6 b5a4 b7b5 a4c2 d7d6 e1g1 c8d7 b1d2 g8e7 a2a4 e8g8 g1h1 h7h6 d1e2 c5b6 f3g1 f7f5 e4f5 b5a4 d2c4 e7f5 c2a4 d7b5 a4b5 a6b5 a1a8 d8a8 c4b6 c7b6 f1d1 a8a2 f2f4 b5b4 f4e5 d6e5 d5d6 b4c3 b2c3 a2e2 g1e2 b6b5 d6d7 f8d8 g2g4 f5h4');
INSERT INTO evaluations VALUES(32140,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(32140,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(32140,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(32140,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(32140,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(32140,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(32140,36,'cp 72','h2h3 g4f3 d1f3 g7g6 g2g3 f8g7 c1e3 g8f6 h3h4 a7a5 b1a3 b8d7 e1c1 h7h5 f1h3 e8g8 c1b1 c7c6 e3g5 d7c5 a3c4 b7b5 c4e3 a5a4 g3g4 a4a3 b2b4 c5e6 g5f6 d8f6 f3f6 g7f6 g4h5 g6h5 e3f5 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(32138,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(32138,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(32138,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(32138,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(32138,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(32138,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(32138,36,'cp 53','g4f3 d1f3 g7g6 g2g3 g8f6 c2c3 f8g7 c1e3 c7c6 a2a4 e8g8 a4a5 b8a6 b1d2 d6d5 f1g2 a6c7 d2b3 c7e6 b3c5 d5d4 c5e6 f7e6 c3d4 f6d5 f3e2 d5e3 e2e3 d8d4 e3d4 e5d4 e1g1 g7f6 f2f4 e6e5 a1b1 f6d8 b2b4');
INSERT INTO evaluations VALUES(32158,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(32158,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(32158,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(32158,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(32158,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(32158,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(32158,36,'cp 66','d1f3 g7g6 g2g3 f8g7 c1e3 g8f6 c2c3 a7a5 b1d2 c7c6 h3h4 b8d7 f1h3 e8g8 a2a4 d8e7 e1g1 f8d8 d2c4 h7h5 f1e1 d6d5 e4d5 c6d5 c4b6 d7b6 e3b6 d8e8 b6e3 a8a6 g1g2 e5e4 d3e4 f6e4');
INSERT INTO evaluations VALUES(32157,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32157,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(32157,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(32157,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(32157,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(32157,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(32157,36,'cp 56','g7g6 g2g3 g8f6 f1g2 c7c6 a2a4 b8d7 a4a5 d7c5 e1g1 f8g7 h3h4 c5e6 c1e3 e8g8 g2h3 d8e7 b1d2 d6d5 c2c3 f8d8 d2b3 h7h5 f1e1 d8e8 g1h2 e8d8 h2g2 g7f8 h3e6 e7e6 e3g5 g8g7 g5f6 e6f6 f3f6 g7f6 a5a6 b7b6 d3d4 d5e4 e1e4 e5d4 b3d4');
INSERT INTO evaluations VALUES(21492,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(21492,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(21492,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(21492,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(21492,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(21492,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(21492,36,'cp 207','g1f3 f8b4 b1c3 g8f6 c4d3 b4a5 e1g1 a5b6 c1e3 f6g4 e3f4 e8g8 h2h3 g4f6 e4e5 f6h5 f4e3 f7f5 d3b1 d8e8 a2a3 h7h6 f1e1 c8d7 d4d5 c6e7 b1a2 e6d5 e3b6 a7b6 a2d5 e7d5 c3d5 a8d8 d5c7 e8f7 e5e6 d7e6 c7e6 d8d1 a1d1 g8h7 e6f8 f7f8');
INSERT INTO evaluations VALUES(23084,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(23084,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(23084,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(23084,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(23084,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(23084,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(23084,36,'cp 32','b1c3 f8b4 g1f3 d7d6 d2d4 e5f3 d1f3 d8h4 g2g3 h4g4 f1g2 g4f3 g2f3 c8e6 c1d2 e8c8 a2a3 b4a5 e1c1 h7h5 d1e1 g8f6 c3d5 a5d2 c1d2 e6d5 e4d5 d8e8 c2c4 h5h4 g3g4 f6h7 e1e2 h7g5 h1f1 f7f6 b2b4 g5f3 f1f3 e8e2 d2e2 h8e8 e2d2 c7c6 d5c6 b7c6 d2d3');
INSERT INTO evaluations VALUES(7253,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7253,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7253,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(7253,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(7253,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(7253,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(7253,36,'cp 13','c1e3 d7d6 c3d5 c5e3 f2e3 e8g8 d1d2 f6d5 c4d5 c6e7 d5b3 c8e6 b3e6 f7e6 f1f2 d8d7 a1f1 f8f7 b2b4 e7g6 h2h3 a8f8 g1h2 d7e7 d2e2 b7b6 c2c4 g6h4 f3d2 f7f2 f1f2 f8f2 e2f2 h4g6 f2e2 b6b5 c4b5 a6b5');
INSERT INTO evaluations VALUES(7255,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(7255,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(7255,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7255,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7255,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(7255,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(7255,36,'cp 9','d7d6 c3d5 f6d5 c4d5 c5e3 f2e3 e8g8 d1e1 c8e6 d5e6 f7e6 a2a3 d8e7 h2h3 f8f7 e1c3 a8f8 a1e1 e7d7 b2b4 c6e7 f3d2 f7f1 e1f1 f8f1 g1f1 g8f7 c3b3 e7g8 f1e1 g8f6 a3a4 b7b6 b3c3 b6b5 a4a5 f7e7 e1d1 e7d8');
INSERT INTO evaluations VALUES(26469,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(26469,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(26469,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(26469,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(26469,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(26469,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(26469,36,'cp 95','f1e2 e7e6 c2c4 f8b4 b1c3 d5c4 e1g1 g8e7 e2c4 b4c3 b2c3 e6e5 d4e5 d8d1 f1d1 f6e5 f4g3 e7f5 a1b1 f5g3 h2g3 e5e4 b1b7 a8d8 d1d4 c6d4 f3d4 a7a6 b7c7 d8c8 c7g7 c8c4 g7g4 c4c3 g4e4 e8f7 e4f4 f7g6 d4f3 h8e8 a2a4 g6g7 f4d4 c3c1 g1h2');
INSERT INTO evaluations VALUES(26468,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(26468,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(26468,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(26468,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(26468,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(26468,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(26468,36,'cp 74','e7e6 c2c4 f8b4 b1c3 g8e7 e1g1 b4c3 b2c3 c6a5 c4d5 e6d5 h2h3 g4h5 f3d2 h5e2 d1e2 e8g8 c3c4 f8e8 c4c5 e7c6 e2d3 e8e7 a1b1 d8d7 f4g3 a8f8 b1c1 c6b4 d3c3 b4c6 c3b2 g8h8 c1b1 h7h6 f1c1 f8c8');
INSERT INTO evaluations VALUES(41191,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41191,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(41191,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(41191,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41191,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(41191,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41191,36,'cp 0','g2g3 e8g8 f1g2 a7a5 a2a4 b8a6 g1e2 f8e8 c1e3 a6b4 e1g1 e7e6 d5e6 c8e6 e2f4 e6d7 e3d4 b4c6 d4e3');
INSERT INTO evaluations VALUES(51319,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51319,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(51319,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(51319,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51319,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51319,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51319,36,'cp 0','a7a6 a2a4 b5b4 f1c4 e7e6 c3e2 c8b7 g1f3 g8f6 e1g1 b8d7 e2g3 c7c5 d1e2 h7h5 h2h4 f8d6 f3g5 d6g3 f2g3 e8g8 f1f4 d8e7 e3e4 e6e5 f4f6 d7f6 d4e5 e7e5 c1f4 e5d4 f4e3 d4d6 e3f4');
INSERT INTO evaluations VALUES(18980,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18980,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(18980,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(18980,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(18980,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(18980,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(18980,36,'cp 11','d2d3 d7d6 g1e2 a7a6 g2g3 a8b8 f1g2 b7b5 e1g1 g8f6 b2b3 f8e7 c3d5 b5c4 b3c4 e8g8 d5e3 c8d7 e2c3 f8e8 c1d2 c6d4 a1b1 e7f8 b1b8 d8b8 d1a1 b8a7 f1b1 e8b8 h2h3 b8b1 a1b1 f8e7 g1h2 h7h6 b1a1 e7f8 a1b2 a7a8 c3d5');
INSERT INTO evaluations VALUES(18981,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(18981,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(18981,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(18981,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18981,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(18981,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18981,36,'cp 5','d7d6 g2g3 a8b8 g1e2 a7a6 f1g2 b7b5 e1g1 f8e7 b2b3 g8f6 c3d5 e8g8 d5e7 d8e7 h2h3 b5c4 b3c4 c8d7 g1h2 b8b7 f2f4 f8b8 f4f5 h7h6 g3g4 e7d8 d1e1 c6b4 e1d2 d7a4 a1b1 f6d7 b1b2 b4c6 b2b7 b8b7 e2c3 d8a5 c3d5 a5d2');
INSERT INTO evaluations VALUES(20605,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(20605,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(20605,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(20605,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(20605,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(20605,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(20605,36,'cp 13','g2g3 h7h5 h2h4 g7g6 f1g2 f8g7 g1e2 g8e7 c1g5 e8g8 e1g1 c8g4 a2a3 f7f6 g5d2 a8b8 f2f3 g4e6 g1h2 a7a5 a1b1 d8d7 d1e1 g8h7 b2b3 f6f5 d2g5 f5f4 c3d5 b7b5 g3f4 e5f4 e2f4 g7e5 h2h1 b5c4 d3c4');
INSERT INTO evaluations VALUES(49044,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(49044,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(49044,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(49044,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(49044,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(49044,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(49044,36,'cp -128','f7f6 c4b5 c7c6 f3h5 g7g6 e5g6 h7g6 h5h8 c6b5 h8h7 c8f5 h7b7 d8e7 b7e7 f8e7 d2d3 e4d6 b1c3 f5e6 c1e3 b5b4 c3e2 d6f5 h2h4 b8c6 e2f4 e6f7 e1c1 e7d6 g2g4 f5d4 c1b1 e8d7 g4g5 d4f3 h1h3 f6g5 h4g5 f3g5 h3h6 d5d4 e3c1 a8g8 f4g6 f7g6 c1g5 b4b3 c2b3');
INSERT INTO evaluations VALUES(49143,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(49143,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(49143,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(49143,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(49143,34,'cp -126',NULL);
INSERT INTO evaluations VALUES(49143,35,'cp -143',NULL);
INSERT INTO evaluations VALUES(49143,36,'cp -137','f3h5 g7g6 c4b5 c7c6 e5g6 h7g6 h5h8 c6b5 h8h7 c8f5 h7b7 d8e7 b7e7 f8e7 d2d3 e4d6 b1c3 f5e6 c1e3 b5b4 c3e2 d6f5 e1c1 b8c6 e2f4 e6f7 h2h4 e8d7 g2g4 f5d4 g4g5 e7d6 c1b1 d6f4 e3f4 f6f5 c2c3 d4e6 f4d2 a7a5 c3c4 d5d4 d1e1 a5a4 b1c2');
INSERT INTO evaluations VALUES(1646,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(1646,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(1646,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(1646,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(1646,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(1646,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(1646,36,'cp 123','f1b5 c8d7 d2d4 a7a6 b5c4 e8c8 h2h3 d7e8 e1g1 g8f6 a2a3 c6d4 f3d4 a5b6 d1f3 b6d4 c4a6 f6d5 c1g5 d4e5 f1d1 e8c6 f3g4 e5e6 g4e6 f7e6 a6c4 g7g6 c3b5 c6b5 c4b5 f8g7 c2c3 c7c6 b5e2 e6e5 a3a4 g7f6 g5d2 d5f4 e2f1 f4d3');
INSERT INTO evaluations VALUES(1649,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(1649,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(1649,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(1649,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(1649,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(1649,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(1649,36,'cp 88','a7a6 b5c6 b7c6 d2d4 c8g4 d1d3 g4f5 d3c4 a5b6 e1g1 e7e6 f1e1 g8f6 b2b3 b6b7 f3e5 f8d6 e5c6 e8g8 d4d5 e6d5 c3d5 f6d5 c4d5 b7b5 d5f3 f5c2 c1b2 c2g6 b2e5 d6a3 b3b4 a8e8 f3c3 f7f6 e5c7 g8h8 c3a3 b5c6 c7g3 e8e1 a1e1 f8d8');
INSERT INTO evaluations VALUES(10734,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(10734,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(10734,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(10734,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(10734,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(10734,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(10734,36,'cp 98','d4e5 c6e5 c3b5 e5f3 g2f3 d6e7 h1g1 e8g8 c1h6 f6e8 d1d2 g8h8 h6f4 d7d6 e1c1 a7a6 b5c3 c8e6 h2h4 e8f6 f4e3 b7b5 c3d5 f6d5 e4d5 e6f5 h4h5 e7f6 h5h6 f8g8 h6g7 f6g7 e3g5 d8d7 g1g3 a8e8 f3f4 c7c6 d5c6 d7c6 f1d3 f5d3 g3d3');
INSERT INTO evaluations VALUES(10722,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(10722,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(10722,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(10722,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(10722,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(10722,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(10722,36,'cp 101','c6e5 c3b5 e5f3 g2f3 d6e7 e4e5 f6h5 f1e2 a7a6 b5c3 d7d6 f3f4 g7g6 e2h5 g6h5 d1h5 d8d7 e5d6 e7d6 h1g1 d7h3 h5d5 h3e6 c1e3 e6d5 c3d5 e8d7 d5c3 d7c6 c3e4 d6e7 e4g5 h8g8 e1c1 c8e6 g5e6 f7e6 g1g3 e7d6 g3h3 a8f8 d1d4 f8f7 h3h5 g8g1 c1d2 b7b6 a2a4 g1f1 d2e2 f1b1 b2b3 c6d7');
INSERT INTO evaluations VALUES(10745,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(10745,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(10745,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(10745,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(10745,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(10745,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(10745,36,'cp 121','c3b5 e5f3 g2f3 d6b4 c2c3 b4c5 e4e5 f6h5 d1d5 c5b6 e5e6 h5f6 e6f7 e8f8 d5g5 f8f7 h1g1 g7g6 c1e3 d7d5 f1d3 b6e3 f2e3 c7c6 b5d4 f6h5 h2h4 d8g5 h4g5 h5g7 g1h1 g7e6 f3f4 f7g7 e1f2 c8d7 d3c2 a8e8 b2b4 e6d4 c3d4 e8a8 a2a4 a8e8 c2d3 h7h6 h1h6 h8h6');
INSERT INTO evaluations VALUES(18820,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(18820,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(18820,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(18820,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(18820,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(18820,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18820,36,'cp 0','g2g3 g8f6 f1g2 f8e7 g1e2 e8g8 e1g1 d7d6 d3d4 c5d4 e2d4 c8d7 d4c6 d7c6 a2a4 a7a6 c1e3 d8c7 a4a5 f6d7 e3d4 a8c8 d1d2 h7h6 f1d1 f8d8 d4e3 d7c5 e3d4');
INSERT INTO evaluations VALUES(18819,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18819,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18819,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18819,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18819,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(18819,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18819,36,'cp -8','g8f6 f1g2 f8e7 g1e2 e8g8 d3d4 c5d4 e2d4 d7d6 e1g1 h7h6 h2h3 a7a6 c1e3 c8d7 d4c6 d7c6 a2a4 d8c7 a4a5 f6d7 d1d2 a8c8 f1d1 f8d8 e3d4 e7f6 c3e2 f6e5 e2c3 e5d4 d2d4 d7f6 d4e3 d8d7 d1d3 d6d5 e4d5 f6d5 c3d5 c6d5');
INSERT INTO evaluations VALUES(28118,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(28118,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(28118,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(28118,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(28118,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(28118,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(28118,36,'cp 110','c1e3 b8d7 d1d2 c7c5 d4c5 d7c5 f1b5 e8f8 f3d4 a7a6 b5d3 g4d7 f2f4 g8f6 h2h3 f8g8 e1g1 a8c8 e4e5 f6e8 d2f2 c5d3 c2d3 h7h6 a1d1 e7e6 d4b3 g8h7 d3d4 e8c7 d4d5 c7d5 c3d5 e6d5 d1d5 d6e5 f4e5 d8e7 f2f7 e7f7 f1f7');
INSERT INTO evaluations VALUES(28115,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(28115,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(28115,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(28115,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(28115,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(28115,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(28115,36,'cp 96','b8d7 d1d2 c7c6 f3g1 g4e6 f2f3 a8c8 g1e2 e6c4 b2b3 c4a6 e1c1 d8a5 c1b1 c6c5 d4c5 d7c5 e3d4 g8f6 d4f6 g7f6 c3d5 a5d2 d5f6 e7f6 d1d2 a6e2 f1e2 e8e7 h1d1 h8d8 e2c4 b7b6 h2h4 h7h5 a2a3 c5e6 a3a4');
INSERT INTO evaluations VALUES(49590,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(49590,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(49590,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(49590,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(49590,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(49590,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(49590,36,'cp -47','f6e4 f1d3 d7d5 d4e5 b8c6 g1f3 c8g4 e1g1 g4f3 g2f3 e4c5 d3b5 g7g5 c1e3 d8d7 c3c4 d5d4 b5c6 d7c6 d1d4 a8d8 d4c3 c6f3 e3c5 f3g4 c3g3 g4g3 h2g3 f8c5 b1c3 e8d7 c3e4 c5e7 f2f4 g5f4 g3f4 d7e6 g1f2 h7h5 a1d1 h5h4 d1d8 h8d8');
INSERT INTO evaluations VALUES(49899,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(49899,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(49899,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(49899,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(49899,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(49899,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(49899,36,'cp -22','f1d3 d7d5 d4e5 b8c6 g1f3 c8g4 e1g1 g4f3 g2f3 e4c5 d3b5 g7g5 c1e3 d8d7 c3c4 d5d4 d1d4 e8c8 d4d7 c5d7 e5e6 f7e6 b5c6 b7c6 f1d1 f8d6 b1d2 h8g8 h2h3 d7e5 g1f1 d6b4 d2e4 d8f8 e3a7 f8f3 a7c5 f3f4 c5b4 f4e4 b4c3 e5c4 d1e1 e4e1 a1e1 c8d7 e1d1 d7e8');
INSERT INTO evaluations VALUES(57074,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57074,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(57074,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(57074,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(57074,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57074,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(57074,36,'cp 5','d7d5 b2b3 g8f6 c1b2 f8d6 d2d4 e8g8 f1d3 b7b6 e1g1 c8b7 b1d2 b8d7 c2c4 a8c8 c4d5 e6d5 f3e5 d8e7 d3f5 g7g6 f5h3 c8e8 d2f3 d7b8 a1c1 f6e4 f3d2 f7f5 a2a4 e4f6 f1e1 f6e4 d2e4');
INSERT INTO evaluations VALUES(22104,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(22104,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(22104,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(22104,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(22104,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22104,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22104,36,'cp -15','d8e7 c1f4 f7f6 e5e6 c8e6 b1c3 c7c6 e1c1 d6d5 f1e2 e7d7 h1e1 e8f7 g2g4 b7b6 f4g3 a8d8 d4a4 h7h5 g4h5 g8h6 e2f3 f8c5 a4f4 c5d6 f4a4 h8e8 g3d6 b6b5 a4d4 h6f5 d4g4 f5d6 g4g6 f7f8 g6h7 b5b4 c3a4 d6f7 e1g1');
INSERT INTO evaluations VALUES(1670,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(1670,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(1670,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(1670,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(1670,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(1670,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(1670,36,'cp 0','c1f4 f7f6 e5e6 c8e6 b1c3 c7c6 e1c1 d6d5 f1e2 e7d7 h1e1 e8f7 g2g4 b7b6 f4g3 g8e7 h2h4 h7h5 g4h5 a8e8 d4d3 e7f5 e2g4 f5g3 f2g3 e6g4 d3g6 f7g8 g6e8 d7e8 e1e8 g4d1 c1d1 h8h5 e8a8 g7g5 h4g5 f6g5 a8a7 h5h1 d1e2 h1h2 e2d3 f8d6 c3e2 d6g3 e2g3 h2h3 a7b7 h3g3 d3d2 g3g2 d2d3');
INSERT INTO evaluations VALUES(53029,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(53029,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(53029,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(53029,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(53029,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(53029,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(53029,36,'cp 57','d2d4 c5d4 f3d4 b8c6 b1c3 g8f6 e2e4 c6d4 d1d4 g7g6 f1e2 f8g7 e1g1 e8g8 d4d3 a7a5 c1e3 a5a4 e3d4 d8a5 b2b3 c8d7 a1b1 a4b3 a2b3 a5g5 f2f4 g5h6 d3d2 d7c6 e2f3 a8a3 d4e3 h6h4 e3f2 h4h6 c4c5 f8e8 c5d6 e7d6 d2d6 a3a8 d6c7 f6e4');
INSERT INTO evaluations VALUES(53035,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(53035,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(53035,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(53035,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(53035,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(53035,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(53035,36,'cp 68','c5d4 f3d4 g7g6 e2e4 b8c6 f1e2 c6d4 d1d4 g8f6 c1e3 f8g7 b1c3 e8g8 d4d3 a7a5 e1g1 a5a4 e3d4 d8a5 b2b3 a4b3 a2b3 a5a1 f1a1 a8a1 e2f1 a1a8 h2h3 c8d7 d3d2 d7c6 f2f3 a8a3 d2b2 f8a8 c3a4 a3b3 b2b3 a8a4 b3b2 f6e8 d4c3 g7c3');
INSERT INTO evaluations VALUES(53127,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(53127,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(53127,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(53127,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(53127,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(53127,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(53127,36,'cp 53','f3d4 g8f6 b1c3 b8c6 e2e4 g7g6 f1e2 c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 a7a5 c1e3 a5a4 e3d4 d8a5 b2b3 c8d7 a1b1 a4b3 a2b3 a5b4 d4f6 g7f6 c3d5 b4a3 f1d1 d7c6 d5f6 e7f6 f2f3 a3c5 g1f1 f6f5 e4f5 a8a2 f5g6 h7g6 d3d6');
INSERT INTO evaluations VALUES(53125,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(53125,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(53125,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(53125,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(53125,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(53125,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(53125,36,'cp 57','g8f6 b1c3 g7g6 e2e4 b8c6 f1e2 c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 c8e6 c1e3 f6d7 b2b3 d8a5 e3d4 g7d4 d3d4 a8c8 g1h1 a7a6 f2f4 d7f6 a2a4 a5c5 d4d3 e6d7 h2h3 d7c6 e2f3 f6d7 a1b1 e7e5 b3b4 c5d4');
INSERT INTO evaluations VALUES(53451,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(53451,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(53451,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(53451,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(53451,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(53451,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(53451,36,'cp 33','e7e6 g1f3 f8d6 e1g1 g8e7 f1e1 e8g8 b1c3 h7h6 c1e3 b8c6 a2a3 e7d5 e3d2 c6e7 f3e5 d5c3 b2c3 d6e5 e1e5 e7g6 e5e3 c8d7 c4d3 d8f6 e3f3 f6h4 f3g3 d7c6 d1e1 f8e8 d3g6 f7g6 g3g6 h4h5 g6g3 e6e5 d4d5 c6d5 a1d1 d5f7 c3c4');
INSERT INTO evaluations VALUES(54446,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(54446,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(54446,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(54446,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(54446,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(54446,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(54446,36,'cp 59','g1f3 f8d6 b1c3 g8e7 e1g1 e8g8 f1e1 h7h6 c1d2 b8c6 d1e2 c6b4 a1d1 c8d7 a2a3 b4d5 c4d3 d6f4 c3d5 e7d5 e2e4 f7f5 e4e2 f8e8 d3c4 f4d6 f3e5 d6e5 e2e5 d7c6 e5g3 d8f6 e1e5 d5c7 c4b3 a8d8 d2c3');
INSERT INTO evaluations VALUES(54445,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(54445,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(54445,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(54445,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(54445,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(54445,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54445,36,'cp 26','g8f6 b1c3 f8e7 e1g1 e8g8 a2a3 c8d7 c1e3 d7c6 f3e5 a7a5 e5c6 b7c6 c4d3 b8d7 a1c1 d7b6 g2g3 d8d7 c1c2 b6d5 c3d5 f6d5 d3e4 e7f6 d1e2 a8b8 h2h4 h7h6 f1c1 f8c8 c2c5 d5e3 f2e3 b8b3 c5c6 c8c6 c1c6');
INSERT INTO evaluations VALUES(46376,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(46376,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(46376,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46376,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(46376,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46376,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(46376,36,'cp 37','e1g1 f8e7 b1c3 e8g8 a2a3 b8c6 c1f4 a7a6 c4a2 b7b5 d4d5 f6d5 c3d5 e6d5 a2d5 c8b7 a1c1 c6a5 d5a2 a5c4 a2c4 b5c4 f3e5 d8d1 f1d1 f8d8 e5c4 d8d1 c1d1 a8d8 d1e1 e7f8 h2h3 h7h6 f4e3 d8c8 c4b6 c8d8 b2b4 b7c6');
INSERT INTO evaluations VALUES(46375,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(46375,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46375,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(46375,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(46375,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(46375,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(46375,36,'cp 39','f8e7 b1c3 e8g8 a2a3 b8c6 c1e3 b7b6 d1e2 c8b7 f1d1 c6a5 c4a2 f6d5 a1c1 d5c3 c1c3 a8c8 c3c8 d8c8 b2b4 b7a6 b4b5 a6b7 a3a4 c8d8 e3f4 b7d5 h2h3 d8a8 a2d5 a8d5 d1c1 e7d6 f4d6 d5d6 e2e3 h7h6 f3e5 f8d8 e5c6 a5c6 b5c6');
INSERT INTO evaluations VALUES(40243,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40243,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(40243,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(40243,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40243,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(40243,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(40243,36,'cp 44','b1c3 e8g8 a2a3 c8d7 c1e3 d7c6 f3e5 a7a5 a1c1 b8d7 e5c6 b7c6 c4d3 d7b6 e3f4 e7d6 f4g5 d6e7 d1f3 d8d4 f1d1 d4e5 g5f4 e5h5 f3c6 b6d5 d3e2 h5f5 f4g3 d5c3 c1c3 a8d8 c3d3 d8c8 c6b7 f5e4 e2f3 e4b7 f3b7 c8c2 d3d2 c2d2 d1d2 f8d8 d2d8 e7d8 g3d6 d8b6');
INSERT INTO evaluations VALUES(40241,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(40241,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(40241,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40241,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(40241,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(40241,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(40241,36,'cp 38','e8g8 a2a3 a7a6 d4d5 b7b5 c4b3 f6d5 c3d5 e6d5 d1d5 d8d5 b3d5 a8a7 f1d1 a7d7 c1f4 c8b7 d5b7 d7b7 a1c1 b7d7 d1d7 b8d7 c1c6 e7f6 b2b3 f8a8 g1f1 h7h6 f4d6 d7f8 g2g3 f8e6 h2h3 f6b2 a3a4 b5a4 b3a4 b2f6 a4a5');
INSERT INTO evaluations VALUES(35848,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(35848,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(35848,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(35848,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(35848,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(35848,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(35848,36,'cp 42','f1e1 a7a6 f3e5 b8c6 e5c6 b7c6 c1f4 a6a5 f4e5 f6d7 e5g3 d7b6 c4b3 c8b7 c3e4 a5a4 b3c2 b7a6 a1c1 a6c4 a2a3 g7g6 h2h3 c4b5 g3f4 b6c4 c1b1 f8e8 f4h6 c4d6 e4c3 e7f6 b1c1 d8b6 d1f3 f6d4');
INSERT INTO evaluations VALUES(6549,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(6549,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(6549,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(6549,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(6549,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(6549,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(6549,36,'cp -58','c6d4 f3d4 d8h4 e1f1 d7d5 c4d5 c8g4 d1d3 e8c8 b1c3 g8e7 c1f4 h4f6 d4e2 g4e2 c3e2 e7d5 e4d5 g7g5 d3h3 d8d7 h3c3 f6c3 e2c3 g5f4 a1d1 h8e8 d1e1 d7e7 e1e7 e8e7 h2h4 c8d7 h1h3 c5e3 h3f3 d7d6 c3e2 e7e4 b2b3 d6e5 c2c4 f7f5 g2g3 f4g3 e2g3 e4f4 f1g2 f4f3');
INSERT INTO evaluations VALUES(6858,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(6858,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(6858,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(6858,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(6858,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(6858,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(6858,36,'cp -42','f3d4 d8h4 e1f1 d7d5 c4d5 c8g4 d1d3 e8c8 b1c3 g8e7 c1f4 h4f6 d4e2 g4e2 c3e2 e7d5 d3h3 d8d7 e4d5 g7g5 h3c3 f6c3 e2c3 g5f4 a1d1 h8e8 d1e1 e8g8 e1e4 b7b5 e4f4 b5b4 f4c4 c5f8 c3e2 d7d5 g2g3 d5a5 e2c1 f7f5 f1e2 g8g4 c4g4 f5g4 c1d3 a5a2 h1f1 f8g7');
INSERT INTO evaluations VALUES(6857,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(6857,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(6857,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(6857,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(6857,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(6857,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(6857,36,'cp -47','d8h4 e1f1 d7d5 c4d5 c8g4 d1d3 e8c8 b1c3 g8e7 c1f4 h4f6 d4e2 g4e2 c3e2 e7d5 e4d5 g7g5 d3c3 f6c3 e2c3 g5f4 a1d1 h8e8 h2h4 f7f5 h1h3 e8g8 d1e1 g8g4 e1e5 d8f8 h4h5 c5d6 e5e6 f8g8 c3b5 d6c5 h3c3 c5b6 b5d6 c8d7 d6c4 g4g2 c4b6 a7b6 e6f6 g2h2 f6f5 g8g7 h5h6 h2h6 f5f4');
INSERT INTO evaluations VALUES(9295,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(9295,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(9295,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(9295,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(9295,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(9295,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(9295,36,'cp 13','b1c3 e8g8 c3a4 c5b6 a4b6 a7b6 c2c3 c6e7 f1e1 c7c6 b5a4 e7g6 a4c2 f8e8 h2h3 d6d5 e4d5 c6d5 d3d4 e5e4 f3e5 g6e5 d4e5 e8e5 d1d4 e5e8 c1g5 c8d7 c2b3 d7c6 e1d1 h7h6 g5f6 d8f6 d4f6 g7f6 b3d5 f6f5 d1d4 f5f4 d5c6 b7c6 d4d7 e8e6');
INSERT INTO evaluations VALUES(9287,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9287,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(9287,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(9287,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(9287,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(9287,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9287,36,'cp 18','e8g8 c3a4 c5b6 c2c3 h7h6 f1e1 c8d7 a4b6 a7b6 a2a4 c6e7 b5d7 d8d7 c3c4 d7e6 b2b3 f6d7 d3d4 f7f5 d1c2 a8e8 d4e5 d6e5 c1b2 f5e4 c2e4 e7c6 b2a3 d7f6 e4g6 f8f7 a3b2 e5e4 f3d4 c6d4 b2d4 f6d7 g6e6 e8e6 e1d1');
INSERT INTO evaluations VALUES(3779,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(3779,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3779,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3779,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3779,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(3779,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(3779,36,'cp 25','c3a4 c5b6 a4b6 a7b6 f1e1 h7h6 h2h3 c6a7 b5c4 b6b5 c4b3 a7c6 a2a3 b5b4 d3d4 c8d7 d4d5 c6e7 c1d2 b4a3 a1a3 a8a3 b2a3 d8a8 d1a1 d7a4 a1b2 a8a6 f3h4 f6d7 b3a4 a6a4 b2b7 a4c2 h4f3 e7g6');
INSERT INTO evaluations VALUES(32763,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(32763,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(32763,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(32763,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(32763,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(32763,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32763,36,'cp 5','b8c6 c2c3 d6d5 e4d5 d7d5 d2d4 e7e6 c1e3 c5d4 f3d4 g8f6 b1d2 a7a6 c3c4 d5d7 d4c6 d7c6 d1e2 f8e7 a2a3 e8g8 b2b4 f8d8 d2b3 c6e4 b3a5 d8d7 a1d1 d7d1 e2d1 a8d8 d1f3 e4f3 g2f3 d8d3 a5b7 d3a3 e3c5 e7c5 b4c5 a3f3 g1g2');
INSERT INTO evaluations VALUES(27576,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(27576,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(27576,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(27576,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(27576,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(27576,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(27576,36,'cp 7','c2c3 d6d5 e4d5 d7d5 d2d4 e7e6 c1e3 c5d4 f3d4 g8f6 b1d2 f8e7 d1a4 e8g8 d4c6 d5c6 a4c6 b7c6 d2c4 f6d5 a2a4 f7f6 g2g3 f8b8 a4a5 g8f7 f1e1 e6e5 g1f1 h7h5 a1a4 b8b3 a5a6 g7g5 e1d1 f7e6 f1e2 h5h4 g3h4 g5h4');
INSERT INTO evaluations VALUES(27575,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(27575,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(27575,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(27575,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(27575,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(27575,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(27575,36,'cp 13','d6d5 e4d5 d7d5 d2d4 e7e6 c1e3 c5d4 f3d4 g8f6 b1d2 f8e7 d1a4 e8g8 d4c6 b7c6 d2f3 c6c5 a1d1 d5b7 e3c1 a8d8 f3e5 e7d6 e5c4 d6e7 d1d8 f8d8 h2h3 f6d5 f1d1 d8c8 a4c2 h7h6 d1d3 d5b6 c4b6 a7b6 c1e3 b7e4 c2e2 c8d8 d3d8 e7d8');
INSERT INTO evaluations VALUES(56338,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(56338,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(56338,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(56338,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(56338,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(56338,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(56338,36,'cp 115','e4d5 c7c6 d5c6 b8c6 d2d3 g8f6 b1c3 f8c5 g1e2 e8g8 e1g1 c6a5 c1g5 h7h6 g5f6 d8f6 c4d5 c5b6 g1h1 a5c6 d5b3 c6a5 c3d5 f6d6 d5b6 a7b6 e2c3 c8d7 f1e1 a5b3 a2b3 d7c6');
INSERT INTO evaluations VALUES(56152,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(56152,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(56152,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(56152,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(56152,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(56152,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(56152,36,'cp 107','c7c6 d5c6 b8c6 d2d3 g8f6 b1c3 c8f5 g1f3 f8b4 c1d2 e8g8 e1g1 f5g4 h2h3 g4h5 f1e1 a8c8 a2a3 b4e7 d2g5 h7h6 g5h4 g8h8 d1e2 f6d7 h4e7 d8e7 g2g4 h5g6 c4b5 c8e8 e2e3 f7f5 b5c6 b7c6');
INSERT INTO evaluations VALUES(56348,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(56348,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(56348,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(56348,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(56348,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(56348,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(56348,36,'cp 54','c4d5 d8d5 g1f3 f8d6 b1c3 d5e6 d2d4 g8f6 d4e5 d6e5 d1b3 e5d6 c3b5 e6b3 a2b3 b8c6 c1d2 f6e4 d2c3 e8g8 b5d6 c7d6 f3d4 c8e6 b3b4 e4c3 b2c3 a8c8 e1d2 a7a6 b4b5 a6b5 f1b5 c8c7 h1b1 f8c8 d4e6 f7e6 f2f4 d6d5 b5d3 c6d8 f4f5 c7c3 f5e6 d8e6 b1b7 d5d4');
INSERT INTO evaluations VALUES(56166,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(56166,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(56166,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(56166,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(56166,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(56166,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(56166,36,'cp 34','g8f6 d1a4 c7c6 g1f3 d8d5 f1c4 d5d6 a4b3 d6e7 d2d4 e7b4 b1c3 b4b3 a2b3 e5d4 f3d4 f8c5 f2f3 e8g8 e1f2 b7b5 c4e2 c8e6 d4e6 f7e6 e2d3 b8d7 h1d1 a7a5 c3e4 c5b4 e4f6 d7f6 c1d2 b4d2 d1d2 f8d8 a1c1 a8c8 f2e2 d8d5 d2c2 c8d8 c1d1 d8d6');
INSERT INTO evaluations VALUES(4033,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(4033,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(4033,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(4033,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(4033,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(4033,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(4033,36,'cp 73','e7e5 e2e4 f8e8 h2h3 e5d4 f3d4 d7c5 f1e1 c8d7 a1b1 h7h6 b2b3 a7a6 c1b2 b7b5 b3b4 c5e6 d4e6 e8e6 c4c5 d7c6 c5d6 c7d6 c3e2 c6b7 e4e5 b7g2 e2f4 e6e5 b2e5 d6e5 d1d8 a8d8 f4g2 d8d2 e1e5 d2a2 e5e3 g7f8 g2f4 a6a5 f4d3 a5a4');
INSERT INTO evaluations VALUES(3948,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(3948,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(3948,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(3948,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(3948,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(3948,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(3948,36,'cp 61','e2e4 f8e8 h2h3 e5d4 f3d4 d7c5 f1e1 c8d7 a1b1 a7a5 b2b3 h7h5 c1g5 d8c8 g1h2 d7c6 f2f3 f6d7 g5e3 c5e6 h3h4 d7e5 a2a4 e6d4 e3d4 c8d8 f3f4 e5g4 h2g1 d8e7 d4g7 g8g7 d1d4 e7f6 d4d2 e8e7 c3d5 c6d5 c4d5');
INSERT INTO evaluations VALUES(3958,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(3958,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(3958,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(3958,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(3958,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(3958,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(3958,36,'cp 46','e5d4 f3d4 d7c5 h2h3 f8e8 f1e1 c8d7 a1b1 a7a5 b2b3 d8c8 g1h2 h7h5 c1g5 d7c6 f2f3 f6d7 e1f1 g7d4 d1d4 c5e6 d4d2 e6g5 d2g5 c8d8 g5d2 d8f6 h3h4 e8e7 g2h3 a8e8 h3d7 e7d7 b1c1 a5a4 b3b4 f6e5 f3f4');
INSERT INTO evaluations VALUES(20724,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(20724,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(20724,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(20724,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(20724,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(20724,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(20724,36,'cp 57','g2g4 h7h6 c1e3 g8f6 h2h3 f8e7 d1d2 c6d4 d2d4 a7a6 e1c1 e6e5 d4d3 c8e6 c1b1 a8c8 a2a3 e8g8 f2f3 d8c7 g4g5 h6g5 e3g5 f6e4 c3e4 f7f5 h3h4 f5e4 f3e4 e7g5 h4g5 c7c4 d3c4 e6c4 f1c4 c8c4 d1d6 c4e4 g5g6 e4g4 h1h5');
INSERT INTO evaluations VALUES(21378,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(21378,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(21378,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(21378,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(21378,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(21378,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21378,36,'cp 52','g8f6 g1f3 f8b4 c1d2 e8g8 d1c2 a7a5 a2a3 b4c3 d2c3 b7b6 b2b3 c8a6 c3b2 c6e7 a1c1 c7c5 d4c5 b6c5 f1e2 d8b6 e1g1 f8b8 b2f6 g7f6 c4d5 a6e2 c2e2 e6d5 e2c2 b8c8 f1d1 b6e6 d1e1 c5c4 b3c4 c8c4');
INSERT INTO evaluations VALUES(37198,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37198,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(37198,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(37198,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37198,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37198,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37198,36,'cp 0','f2f3 d7d5 c4d5 e6d5 e2e3 c8f5 g1e2 b8c6 e2g3 f5g6 c3c4 f8e8 e1f2 d5c4 f1c4 c6a5 c4a2 c7c5 c1b2 c5c4 e3e4 a8c8 a1c1 h7h5 h2h4 d8c7 d4d5 f6g4 f3g4 c7f4 d1f3 f4d2 f3e2 d2f4');
INSERT INTO evaluations VALUES(51827,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(51827,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(51827,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(51827,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(51827,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(51827,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(51827,36,'cp 132','g8f6 f1c4 d7d5 e4d5 f8d6 d2d4 c6d5 c4b5 b8d7 e1g1 e8g8 c1f4 d8b6 a2a4 f6e4 b1d2 d7e5 d4e5 d6c7 d2f3 a7a6 f4e3 b6e6 b5d3 c7e5 f3e5 e6e5 c2c3 f8e8 f1e1 e4f6 f2f3 e5d6 d1d2 c8d7 d3c2 a8d8 e3f2 e8e1 a1e1 d8e8 e1e8 d7e8 d2d3 d6c7');
INSERT INTO evaluations VALUES(44067,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(44067,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(44067,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(44067,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(44067,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(44067,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(44067,36,'cp 165','f1c4 d7d5 e4d5 f8d6 d2d4 c6d5 c4b5 b8d7 e1g1 e8g8 c1f4 d8b6 a2a4 a7a6 e5d7 c8d7 f4d6 b6d6 b5d7 d6d7 c2c3 b7b5 b1a3 b5a4 a3c2 a6a5 c2e1 d7b5 e1d3 f6e4 d1c2 f8c8 f1e1 b5b3 e1c1 c8b8 h2h3 e4f6 a1a4 b3c2 c1c2');
INSERT INTO evaluations VALUES(49659,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(49659,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(49659,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(49659,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(49659,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(49659,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(49659,36,'cp -155','b8c6 b1c3 d7d5 f1b5 c8d7 c3d5 f6d5 e4d5 c6b4 b5d7 d8d7 f3b3 e8c8 g1e2 d7f5 d2d3 e5e4 e1g1 e4d3 c2d3 f5d5 d3d4 d5b3 a2b3 a7a6 f1d1 f8e7 c1d2 d8d7 g2g3 h8d8 d2b4 e7b4 d1d3 b4e7 d3f3 e7f6 h2h4 d7d5 g1g2 c7c6 a1c1 a6a5 c1c2 h7h6 c2c4');
INSERT INTO evaluations VALUES(16576,30,'cp -147',NULL);
INSERT INTO evaluations VALUES(16576,31,'cp -140',NULL);
INSERT INTO evaluations VALUES(16576,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(16576,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(16576,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(16576,35,'cp -141',NULL);
INSERT INTO evaluations VALUES(16576,36,'cp -139','b1c3 d7d5 f1b5 c8d7 c3d5 f6d5 e4d5 c6b4 b5d7 d8d7 f3b3 e8c8 g1e2 d7f5 d2d3 e5e4 e1g1 e4d3 c2d3 f5d5 d3d4 d5b3 a2b3 a7a6 c1d2 d8d7 f1d1 f8e7 g2g3 h8d8 d2e3 b7b6 h2h4 e7f6 d1d2 c8b7 a1d1 a6a5 h4h5 c7c6 h5h6 g7g6 e2c3 b4d5 c3d5 d7d5 d1e1');
INSERT INTO evaluations VALUES(16513,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(16513,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(16513,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(16513,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(16513,34,'cp -138',NULL);
INSERT INTO evaluations VALUES(16513,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(16513,36,'cp -138','d7d5 f1b5 c8d7 c3d5 f6d5 e4d5 c6b4 b5d7 d8d7 f3b3 e8c8 g1e2 d7f5 d2d3 e5e4 e1g1 e4d3 c2d3 f5d5 d3d4 d5b3 a2b3 a7a6 f1d1 f8e7 c1d2 b4d5 g1f1 e7f6 d2c3 h8e8 d1d3 e8e6 d3f3 h7h5 e2f4 d5f4 f3f4 d8d5 a1a5 b7b5 g2g3 g7g6 h2h4 c8b7 f1g2');
INSERT INTO evaluations VALUES(52700,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(52700,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(52700,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(52700,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(52700,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(52700,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(52700,36,'cp -88','d5f4 e7e5 f4h5 g7g6 h5g3 h7h5 f1c4 b7b5 c4b3 h5h4 g3f1 a7a5 a2a4 b8a6 d2d3 a6c5 f1d2 g6g5 h2h3 g8e7 b3a2 e7g6 a4b5 c6b5 d2f3 f8e7 f3h2 e8g8 g1f3 g8g7 e1g1 f7f6 a2d5 a8a6 c2c3 d4c3 b2c3 g6f4');
INSERT INTO evaluations VALUES(52911,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(52911,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(52911,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(52911,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(52911,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(52911,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(52911,36,'cp -85','e7e5 f4h5 g7g6 h5g3 h7h5 h2h4 c8g4 g1f3 f8e7 c2c3 b8d7 c3d4 e5d4 d1b3 g8f6 f3d4 d7c5 b3e3 d8b6 f1c4 a8d8 b2b3 c5e4 g3e4 b6d4 e3d4 d8d4 e4g3 b7b5 c1b2 d4d6 f2f3 g4c8 g3e4 f6e4 f3e4 e8g8 c4e2 f8d8 d2d3 c6c5 a2a4 b5b4 b2e5 d6a6 e1c1 c8e6');
INSERT INTO evaluations VALUES(53790,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(53790,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(53790,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(53790,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(53790,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(53790,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(53790,36,'cp -87','f4h5 g7g6 h5g3 h7h5 h2h4 f8e7 g1f3 c8g4 c2c3 g8f6 d1b3 b8d7 c3d4 e5d4 d2d3 a7a5 b3c2 e8g8 f1e2 c6c5 e1g1 a5a4 c1f4 f6h7 f3e5 g4e6 e5d7 d8d7 e4e5 e7h4 g3e4 b7b6 e4d6 h4e7 e2f3 e7d6');
INSERT INTO evaluations VALUES(978,30,'cp -213',NULL);
INSERT INTO evaluations VALUES(978,31,'cp -226',NULL);
INSERT INTO evaluations VALUES(978,32,'cp -231',NULL);
INSERT INTO evaluations VALUES(978,33,'cp -217',NULL);
INSERT INTO evaluations VALUES(978,34,'cp -230',NULL);
INSERT INTO evaluations VALUES(978,35,'cp -228',NULL);
INSERT INTO evaluations VALUES(978,36,'cp -238','e1g1 c8f5 d2d3 e8c8 c1e3 d5d7 f2f3 h7h6 g5e4 b7b6 g1h1 f5e6 b1d2 f7f5 e4c3 g7g5 d2c4 f8g7 a2a4 d7f7 c4d2 a7a6 d1e2 h8e8 e3g1 h6h5 a4a5 b6b5 e2e3 g5g4 f3g4 h5g4 c3e4 e8h8 e4c5 e6d5 c2c4 b5c4');
INSERT INTO evaluations VALUES(977,30,'cp -240',NULL);
INSERT INTO evaluations VALUES(977,31,'cp -239',NULL);
INSERT INTO evaluations VALUES(977,32,'cp -249',NULL);
INSERT INTO evaluations VALUES(977,33,'cp -253',NULL);
INSERT INTO evaluations VALUES(977,34,'cp -257',NULL);
INSERT INTO evaluations VALUES(977,35,'cp -257',NULL);
INSERT INTO evaluations VALUES(977,36,'cp -249','c8f5 d2d3 e8c8 f2f3 d5d7 c1e3 h7h6 g5e4 b7b6 e3f2 f5e6 a2a4 g7g5 b1a3 f8b4 d1e2 f7f5 e4d2 b4d2 e2d2 f5f4 a4a5 h8g8 a5b6 a7b6 d2e2 g5g4 g1h1 g4f3 g2f3 e6d5 c2c4 d5f7 a3b5 d7d3 a1a8 c8b7 e2d3 d8d3 a8g8');
INSERT INTO evaluations VALUES(6605,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(6605,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(6605,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(6605,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(6605,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(6605,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(6605,36,'cp -38','g8f6 f1e2 d7d5 c1d2 b4c3 d2c3 d8d6 e4d5 f6d5 e1g1 c8f5 f3d2 e8c8 e2f3 f7f6 f3d5 d6d5 f2f4 e5e4 d3e4 f5e4 d2e4 d5e4 d1g4 d8d7 f1f2 h8d8 a1e1 e4d5 g4e6 d5c5 e6e3 c5e3 e1e3 c6e7 e3e1 e7d5 f2f3 b7b6 g1f2 d5c3 f3c3');
INSERT INTO evaluations VALUES(31593,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(31593,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(31593,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(31593,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(31593,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(31593,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(31593,36,'cp 81','h2h3 g4h5 d1e2 b8d7 c4b3 h5f3 e2f3 g8f6 d2d3 d7c5 c1e3 c5b3 a2b3 a7a5 d3d4 e5d4 e3d4 f8e7 e1c1 e8g8 h1e1 f6e8 f3g3 b7b5 c1b1 a5a4 b3b4 h7h6 g3g4 e8f6 g4f3 d8c7 e4e5 d6e5 d4e5 c7b7 c3e4 f6d5');
INSERT INTO evaluations VALUES(31591,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(31591,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(31591,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(31591,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(31591,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(31591,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(31591,36,'cp 89','g4h5 d1e2 b8d7 a2a4 h5g6 d2d3 g8f6 g2g4 h7h6 c1d2 d8e7 d2e3 d7c5 f3h4 f6e4 h4g6 e4c3 c4f7 e7f7 g6h8 f7d5 e2f1 c3a2 h1g1 a2b4 f1e2 d5e6 e3c5 d6c5 c2c3 b4d5 g4g5 d5f4 e2g4 e6g4 g1g4 f4d3 e1f1 f8e7 a1d1 a8d8 g5h6 g7h6');
INSERT INTO evaluations VALUES(51788,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51788,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(51788,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(51788,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(51788,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(51788,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51788,36,'cp 34','d7d5 e2e3 g8f6 b1d2 c6c5 c4d5 e6d5 b2b3 c5d4 f3d4 f8b4 f1b5 b8d7 e1g1 a7a6 b5d3 b4c3 a1b1 d7e5 d2f3 c8g4 d3e2 e5f3 d4f3 g4f5 e2d3 f5d3 d1d3 f6e4 f1d1 b7b5 d3d5 d8d5 d1d5 e8g8 c1a3 f8c8');
INSERT INTO evaluations VALUES(2029,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(2029,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(2029,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(2029,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(2029,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(2029,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(2029,36,'cp 141','e1g1 a7a6 b5a4 g8f6 b1c3 e7d6 a4b3 f8e7 d2d3 e8g8 c3e2 f6h5 d3d4 h5f6 d4d5 c6a5 e2g3 g7g6 c1g5 f8e8 d1d2 a5b3 a2b3 e7f8 f3e1 b7b6 e1d3 a6a5 c2c4 c8a6 h2h3 f8g7 f2f3 d6e7');
INSERT INTO evaluations VALUES(2028,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(2028,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(2028,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(2028,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(2028,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(2028,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(2028,36,'cp 150','a7a6 b5a4 g8f6 f1e1 d7d6 d2d4 c8d7 c1g5 h7h6 g5f6 e7f6 b1c3 g7g6 c3d5 f6d8 d4e5 d6e5 c2c4 f8g7 d1b3 a8a7 b3e3 a7a8 e3a3 b7b5 a4c2 b5b4 d5b4 d8e7 b4d5 e7a3 b2a3 a8a7 a1b1 e8g8 c2a4 g8h7 c4c5 d7e8 e1d1 a6a5 a4b5 f7f5 a3a4 c6d4 e4f5 d4f3 g2f3 g6f5 c5c6 e8h5');
INSERT INTO evaluations VALUES(19930,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(19930,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(19930,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(19930,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(19930,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(19930,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(19930,36,'cp 148','d8a5 b1c3 g8f6 c5d6 f6e4 d1d5 e4c3 d5a5 c6a5 c1d2 c3e4 d6e7 f8e7 d2a5 e8g8 f1d3 e4c5 d3c4 c8g4 e1c1 e7f6 h1e1 b7b6 a5b4 f8d8 b2b3 d8d1 e1d1 a8c8 h2h3 g4f3 g2f3 f6g5 c1b1 g5h4 b4e1 c8e8 b1b2 g8f8 b3b4 c5e6 b2b3 f8e7 a2a4');
INSERT INTO evaluations VALUES(1413,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(1413,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(1413,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(1413,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(1413,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(1413,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(1413,36,'cp 143','b1c3 g8f6 c5d6 f6e4 d1d5 e4c3 d5a5 c6a5 c1d2 c3e4 d6e7 f8e7 d2a5 e8g8 f1d3 e4c5 d3c4 e7f6 e1c1 c8g4 h1e1 b7b6 a5b4 f8d8 b2b3 d8d1 e1d1 a8c8 h2h3 g4f3 g2f3 f6g5 c1b1 g5h4 d1d4 h4f6 d4d6 f6h4 b4e1 g8f8 d6d5 c8e8 d5d1 e8c8 a2a4 f8e7 a4a5 b6a5 d1d5 c8d8 d5d8 e7d8');
INSERT INTO evaluations VALUES(1412,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(1412,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(1412,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(1412,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(1412,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(1412,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(1412,36,'cp 141','g8f6 c5d6 f6e4 d1d5 e4c3 d5a5 c6a5 c1d2 c3e4 d6e7 f8e7 d2a5 e8g8 f1d3 f8e8 e1g1 c8f5 f1e1 e7f6 c2c3 f5g6 d3b5 e8c8 b5d7 c8f8 a5b4 f8d8 a1d1 b7b6 d7c6 d8d1 e1d1 a8b8 b4a3 e4c5 c6b5 b8c8 g1f1 g6c2 d1e1 c2f5 a3c5 b6c5 b5c4 h7h6');
INSERT INTO evaluations VALUES(36839,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36839,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(36839,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(36839,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36839,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(36839,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36839,36,'cp 27','a7a5 h2h3 f6h5 c1g5 f7f6 g5e3 h5f4 e2f1 f6f5 g2g3 f5e4 f3d2 f4d3 f1d3 e4d3 d2e4 b7b6 d1d3 b8a6 g3g4 a6c5 e3c5 b6c5 e1c1 a8b8 f2f3 f8f4 c1b1 c8d7 c3e2 f4f7 d1f1 d8c8 f1f2 a5a4 a2a3 c8b7 e2c3 f7f4 d3e2 b7b3 b1a1 g8h8 h3h4 h7h6');
INSERT INTO evaluations VALUES(35612,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(35612,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(35612,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(35612,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(35612,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35612,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(35612,36,'cp 30','h2h3 f6h5 c1g5 f7f6 g5e3 b8a6 f3h2 h5f4 e2f1 a5a4 g2g3 a4a3 b2a3 f4h3 f1h3 c8h3 g3g4 f6f5 g4f5 a6c5 e3c5 d6c5 h2f1 h3g2 h1h2 g2f1 e1f1 d8g5 a1b1 a8a3 h2h3 b7b6 d5d6 c7d6 f5g6 h7h6 d1e2 g8h8 e2e3 g5g6 b1b6 g6f7 h3h6 g7h6 e3h6 h8g8');
INSERT INTO evaluations VALUES(39460,30,'cp 266',NULL);
INSERT INTO evaluations VALUES(39460,31,'cp 271',NULL);
INSERT INTO evaluations VALUES(39460,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(39460,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(39460,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(39460,35,'cp 261',NULL);
INSERT INTO evaluations VALUES(39460,36,'cp 254','d2d4 e5d4 f3d4 c5d4 d1d4 d7d6 b1c3 b8c6 d4d1 d8e7 e1g1 c6e5 c4b3 c8e6 f2f4 e6b3 a2b3 e5c6 c3d5 e7d7 d1h5 e8d8 b3b4 g8e7 b4b5 c6d4 b5b6 e7d5 h5d5 d4c6 c1e3 c7b6 e3b6 d8e7 b6e3 h8e8 a1d1 a8d8 d5b3');
INSERT INTO evaluations VALUES(39462,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(39462,31,'cp 266',NULL);
INSERT INTO evaluations VALUES(39462,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(39462,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(39462,34,'cp 255',NULL);
INSERT INTO evaluations VALUES(39462,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(39462,36,'cp 251','e5d4 f3d4 d7d6 b1c3 c5d4 d1d4 b8c6 d4e3 c6a5 c4e2 c8e6 e2h5 g7g6 h5e2 a5c4 e3d4 c4e5 b2b3 e5c6 d4d2 d8e7 c1b2 e8c8 e1c1 e7f7 f2f4 h7h5 h2h3 d8e8 c3d5 e6d5 e4d5 c6e7 e2b5 e8d8 h1e1 a7a6 b5c4 e7f5 d2a5 g8e7 e1e6');
INSERT INTO evaluations VALUES(55599,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(55599,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55599,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(55599,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(55599,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(55599,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(55599,36,'cp 8','b8c6 g1f3 g7g6 d3d4 e5d4 f3d4 f8g7 f1b5 c8d7 d4c6 b7c6 b5d3 g8f6 e1g1 e8g8 c1f4 f8e8 d1d2 a7a5 f1e1 a5a4 f4h6 g7h6 d2h6 d7e6 h6d2 d8b8 c3d1 b8a7 g1h1 f6d7 d1e3 a7a5 d2a5 a8a5 f2f4');
INSERT INTO evaluations VALUES(21963,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(21963,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(21963,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21963,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21963,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(21963,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(21963,36,'cp 14','g1f3 g7g6 d3d4 e5d4 f3d4 f8g7 d4c6 b7c6 f1d3 g8f6 e1g1 e8g8 c1f4 d6d5 d1c1 f8e8 f1e1 d5e4 d3e4 f6d5 e4d5 e8e1 c1e1 c6d5 f4e5 c7c6 e5g7 g8g7 e1d2 c8f5 c3e2 d8f6 c2c3 a7a5 a1d1 a5a4 h2h3 f5d7 b2b4 a4b3 a2b3 h7h5 c3c4 d5c4 d2d7 c4b3');
INSERT INTO evaluations VALUES(17355,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(17355,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(17355,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(17355,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(17355,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(17355,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(17355,36,'cp 17','c4b3 d7d5 e4d5 e6d5 e1g1 h7h6 d2d4 g8f6 c1f4 c8e6 f1e1 f8e7 d4c5 e7c5 f3d4 c6d4 c3d4 c5b4 b1c3 e8g8 e1e3 a8c8 h2h3 b4c3 b2c3 f6e4 d1e1 c8c6 a2a4 d8a5 f2f3 e4c3 b3c2 a5b4 f4e5 f7f6 e5g3 b4d4 g3f2');
INSERT INTO evaluations VALUES(32282,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(32282,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(32282,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(32282,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(32282,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(32282,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(32282,36,'cp 30','d1b3 b7b6 b1c3 g8f6 g1f3 h7h6 c4d5 e6d5 e3e4 f6e4 b3d5 d8d5 c3d5 f8d6 f3e5 c7c6 d5e3 f5h7 f2f3 e4g5 c1d2 g5e6 d2c3 e8g8 f1c4 b6b5 c4e6 f7e6 a2a3 a7a5 e1e2 d6e5 d4e5 c6c5 h1d1 b5b4 a3b4 c5b4');
INSERT INTO evaluations VALUES(47670,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(47670,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(47670,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(47670,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(47670,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(47670,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47670,36,'cp -18','c7c5 g1f3 c5d4 f3d4 e6e5 d4c2 d5d4 h4f6 g7f6 c3d5 b8c6 e2e3 c8e6 e3d4 e6d5 c4d5 d8d5 d4e5 d5a5 d1d2 a5d2 e1d2 e8c8 d2e1 f8c5 e5f6 h8e8 f1e2 c6b4 c2b4 c5b4 e1f1 d8d2 e2g4 c8c7 a1c1 c7b6 g2g3 b4c5 f1g2 d2f2 g2h3 e8e5 h1f1 f2b2 c1b1 b2b1 f1b1 b6c7');
INSERT INTO evaluations VALUES(56312,30,'cp -289',NULL);
INSERT INTO evaluations VALUES(56312,31,'cp -303',NULL);
INSERT INTO evaluations VALUES(56312,32,'cp -305',NULL);
INSERT INTO evaluations VALUES(56312,33,'cp -320',NULL);
INSERT INTO evaluations VALUES(56312,34,'cp -305',NULL);
INSERT INTO evaluations VALUES(56312,35,'cp -311',NULL);
INSERT INTO evaluations VALUES(56312,36,'cp -323','g7g5 h2h3 g8h6 b1c3 h6f5 d2d4 b8c6 c1d2 f8g7 f1d3 h7h5 d3b5 f5g3 h1g1 c8f5 c3e2 g3e2 d1e2 g5g4 e1c1 g4f3 e2f3 f5e4 f3f4 g7h6 f4f2 d8e7 g1f1 e8c8 b5c6 b7c6 a2a3 e7e6 f2f7 h6d2 d1d2 e6f7 f1f7 d8g8');
INSERT INTO evaluations VALUES(55391,30,'cp -296',NULL);
INSERT INTO evaluations VALUES(55391,31,'cp -290',NULL);
INSERT INTO evaluations VALUES(55391,32,'cp -317',NULL);
INSERT INTO evaluations VALUES(55391,33,'cp -278',NULL);
INSERT INTO evaluations VALUES(55391,34,'cp -301',NULL);
INSERT INTO evaluations VALUES(55391,35,'cp -256',NULL);
INSERT INTO evaluations VALUES(55391,36,'cp -322','h2h3 g8e7 b1c3 e7f5 d2d4 b8c6 f1b5 f5g3 h1h2 h7h6 h3h4 h8g8 h4g5 h6g5 f3g5 g8g5 c1f4 g3f5 h2h8 c8e6 d1d2 g5g4 e1c1 f5h4 g2g3 h4g6 h8h5 d8d7 d1f1 g6e7 c3e2 e8c8 c2c3 g4g8 f4g5 a7a6 b5d3 d8e8 e2f4 e6g4 g5h4 e7g6 h5h7 g6f4 d2f4');
INSERT INTO evaluations VALUES(1860,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(1860,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(1860,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(1860,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(1860,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(1860,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(1860,36,'cp 32','f8c5 c2c3 d7d6 b2b4 c5b6 a2a4 a7a6 e1g1 f6g6 b1d2 g8f6 f3h4 g6h5 d1h5 f6h5 f1e1 h5f6 c4a2 c8e6 d2f3 e6a2 a1a2 g7g6 f3d2 a6a5 b4b5 c6d8 d2c4 f6d7 h4f3 d8e6 d3d4 e5d4 c4b6 d7b6 c3d4 f7f6 c1e3 e8c8');
INSERT INTO evaluations VALUES(40476,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(40476,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40476,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40476,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40476,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(40476,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(40476,36,'cp 17','e8g8 f1d3 d5c4 d3c4 b7b6 e1g1 c8b7 a1c1 b8d7 d1e2 c6c5 f1d1 c5d4 e3d4 a8c8 c4d3 h7h6 h2h3 d8e7 a2a3 b4c3 d2c3 f6d5 c3d2 d7f6 d1e1 c8c1 e1c1 f8c8 e2e1 c8c1 e1c1 a7a5 f3e5 f6d7 e5d7');
INSERT INTO evaluations VALUES(36101,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36101,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(36101,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(36101,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(36101,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(36101,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36101,36,'cp 30','f1d3 d5c4 d3c4 b7b6 e1g1 c8b7 d1e2 c6c5 f1d1 c5d4 f3d4 b8d7 a2a3 b4e7 a1c1 a7a6 d2e1 b6b5 c4a2 d8b6 f2f3 d7e5 b2b4 f8d8 e1f2 e5c6 d4b3 d8d1 e2d1 a8d8 d1e1 f6d7 e3e4 b6c7 c3e2 d7e5 b3c5');
INSERT INTO evaluations VALUES(36102,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36102,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36102,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(36102,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(36102,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36102,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(36102,36,'cp 25','d5c4 d3c4 b7b6 d1e2 c8b7 e1g1 c6c5 f1d1 c5d4 f3d4 b8d7 a2a3 b4e7 a1c1 a7a6 c4a2 b6b5 e3e4 d8b8 f2f4 e7c5 d2e3 e6e5 d4f5 c5e3 e2e3 e5f4 e3d4 b8e5 d4e5 d7e5 c3d5 b7d5 f5e7 g8h8 e4d5 g7g6 d5d6 h8g7 c1c7 a8b8 e7c6 e5c6 c7c6 b5b4 c6a6');
INSERT INTO evaluations VALUES(43228,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(43228,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(43228,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(43228,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(43228,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(43228,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(43228,36,'cp -28','d7d5 a2a4 e8g8 a4a5 c7c6 f1e2 b8d7 e1g1 f8e8 f3e5 d7e5 d4e5 f6e4 c3e4 d5e4 d1d4 d8d5 a1a4 g7e5 d4e4 c8f5 e4c4 d5d6 f2f4 e5g7 e3e4 f5e6 c4d3 b7b5 a5b6 d6d3 e2d3 a7b6 a4a8 g7d4 g1h1 e8a8 c2c3 d4c5 b2b4 c5d6 h2h3 d6c7 c1e3 a8a3 e3d4 c6c5 b4c5 b6c5 d4c5 a3c3');
INSERT INTO evaluations VALUES(41487,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(41487,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(41487,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(41487,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(41487,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(41487,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(41487,36,'cp -24','f1e2 e8g8 e1g1 b8d7 b2b3 b7b6 c3a4 c8b7 c1b2 f6e4 a1c1 c7c5 d4c5 g7b2 a4b2 e4c3 d1e1 c3e2 e1e2 d7c5 f1d1 d8d6 c2c4 a8d8 h2h3 d6f6 c4d5 b7d5 b2c4 d5f3 e2f3 f6f3 g2f3 d8d1 c1d1 f8c8');
INSERT INTO evaluations VALUES(57590,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(57590,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(57590,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(57590,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57590,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57590,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57590,36,'cp 0','f5e4 b1c3 g8f6 c1g5 g7g6 d4d5 f8g7 d1d2 e8g8 e1c1 d7d6 g1e2 b8d7 e2d4 d7b6 h2h4 d8e8 d4b5 e8d8');
INSERT INTO evaluations VALUES(57616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57616,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57616,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57616,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57616,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57616,36,'cp 0','b1c3 g8f6 c1g5 g7g6 d4d5 f8g7 d1d2 d7d6 e1c1 b8d7 g1e2 e8g8 e2d4 d7b6 h2h4 d8e8 d4b5 e8d8');
INSERT INTO evaluations VALUES(57614,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57614,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57614,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57614,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57614,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57614,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57614,36,'cp 0','g8f6 c1g5 g7g6 d4d5 d7d6 d1d2 f8g7 e1c1 b8d7 g1e2 e8g8 e2d4 d7b6 h2h4 d8e8 d4b5 e8d8');
INSERT INTO evaluations VALUES(50120,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50120,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50120,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50120,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50120,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50120,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50120,36,'cp 0','c1g5 g7g6 d4d5 d7d6 d1d2 f8g7 e1c1 b8d7 g1e2 e8g8 e2d4 d7b6 h2h4 d8e8 d4b5 e8d8');
INSERT INTO evaluations VALUES(50119,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50119,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50119,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50119,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50119,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50119,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50119,36,'cp 0','g7g6 d4d5 d7d6 d1d2 b8d7 g1e2 f8g7 e1c1 e8g8 e2d4 d7b6 h2h4 d8e8 d4b5 e8d8');
INSERT INTO evaluations VALUES(21447,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(21447,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(21447,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(21447,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(21447,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(21447,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(21447,36,'cp 55','d2d4 g8f6 e4e5 f6e4 c3e2 f7f6 e2g3 f6e5 d4e5 f8e7 c2c3 e4c5 c1e3 e8g8 h2h4 d8e8 f1b5 c8d7 b5e2 a7a5 e1g1 h7h6 a1c1 b7b6 d1c2 e8f7 f1d1 d7e8 e2b5 a5a4 c3c4 c6b4 c2e2 a4a3 b5e8 f8e8 b2a3');
INSERT INTO evaluations VALUES(23055,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(23055,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(23055,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(23055,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(23055,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(23055,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(23055,36,'cp 77','d4f3 d1f3 a7a6 c1e3 d7d6 b1d2 g8f6 h2h3 f8e7 a2a4 e8g8 a4a5 f6e8 g2g4 e7g5 d2c4 g5e3 c4e3 b7b6 h3h4 b6a5 g4g5 a8b8 b2b3 f7f6 f3g3 f6f5 e4f5 c8f5 a1a5 c7c6 a5a6 c6d5 e3f5 f8f5 f1d3');
INSERT INTO evaluations VALUES(23059,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(23059,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(23059,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(23059,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(23059,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(23059,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(23059,36,'cp 80','d1f3 f8c5 f3g3 d8f6 b1c3 d7d6 c3a4 c5d4 f1b5 e8d8 f2f3 g8e7 c2c3 a7a6 c3d4 a6b5 a4c3 b5b4 c1g5 f6g6 c3e2 f7f6 g5d2 e5d4 g3g6 e7g6 e2d4 g6e5 e1g1 e5d3 a2a3 b4a3 a1a3 a8a3 b2a3 d3c5 f1b1 b7b6 b1c1 c8d7 d2b4');
INSERT INTO evaluations VALUES(23058,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(23058,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(23058,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(23058,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(23058,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(23058,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(23058,36,'cp 61','f8c5 f3g3 d8f6 b1c3 a7a6 h2h4 h7h6 f2f4 d7d6 f4e5 d6e5 f1e2 g8e7 c1e3 c5e3 g3e3 h6h5 e1c1 f6h6 d1d3 e7g6 e3h6 h8h6 d3g3 e8f8 e2d1 g6f4 c3e2 c8g4 e2f4 e5f4 g3b3 g4d1 h1d1 b7b6 d1f1 h6f6 e4e5 f6f5 d5d6 c7d6 e5d6 b6b5');
INSERT INTO evaluations VALUES(22601,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(22601,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(22601,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(22601,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(22601,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(22601,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(22601,36,'cp 33','g8f6 e2e3 h7h6 c2c4 f8b4 b1d2 d7d6 a2a3 b4d2 d1d2 e8g8 f4g3 f6e4 d2c2 e4g3 h2g3 e6e5 d4e5 c6e5 f3e5 d6e5 a1d1 d8e7 f1e2 c7c6 e1g1 c8e6 c2c3 f7f6 c4c5 f8d8 b2b4 e6f7 c3c2 a7a5 d1d8 a8d8 b4a5 d8a8 a5a6 b7a6 e2d3');
INSERT INTO evaluations VALUES(34563,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(34563,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(34563,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(34563,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(34563,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(34563,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(34563,36,'cp 55','e7e5 d1e2 b8c6 f3c6 g6c6 e2e5 g8e7 e5e4 c8d7 g1f3 e8c8 d2d3 f7f6 e1g1 e7f5 f1e1 d8e8 e4c6 e8e1 f3e1 d7c6 c3e4 f8e7 c1d2 h8d8 e4g3 f5g3 h2g3 b7b6 e1f3 h7h5 a1e1 e7c5 d2e3 c5d6 g1f1 d8e8 f3d2 d6b4 c2c3 b4d6 d2c4 d6f8 f2f3 c6d7');
INSERT INTO evaluations VALUES(42754,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(42754,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(42754,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(42754,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(42754,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42754,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(42754,36,'cp -13','c7c6 c1d2 d7d5 c4b3 d5e4 c3e4 f6e4 d2b4 d8b6 d1e2 b6b4 c2c3 e4c3 e2e5 b4e7 e5e7 e8e7 b2c3 h8d8 e1c1 a7a5 g1e2 a5a4 b3c2 b7b5 a2a3 b8a6 h1e1 a6c7 e2f4 e7d6 d3d4 c8e6 f4e6 c7e6 e1e4 e6c7 c1d2 g7g6 d1e1 d8e8 e4e8 c7e8 c2d3 e8c7 c3c4 b5c4 d3c4');
INSERT INTO evaluations VALUES(39879,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(39879,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(39879,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(39879,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(39879,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(39879,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(39879,36,'cp -13','c1d2 d7d5 c4b3 d5e4 c3e4 f6e4 d2b4 d8b6 d1e2 b6b4 c2c3 e4c3 e2e5 b4e7 e5e7 e8e7 b2c3 h8d8 e1c1 a7a5 g1e2 a5a4 b3c2 b7b5 h1e1 b8a6 e2f4 e7d6 d3d4 a6c7 c1d2 c8d7 a2a3 d8e8 f2f3 g7g6 h2h4 d7e6 f4e6 e8e6 e1e6 c7e6 d1e1 a8d8 g2g4 e6f4 e1e5 f7f6 e5c5 d6c7');
INSERT INTO evaluations VALUES(39880,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(39880,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(39880,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(39880,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(39880,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(39880,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(39880,36,'cp -20','d7d5 c4b3 d5e4 c3e4 f6e4 d2b4 d8d4 d1e2 d4b4 c2c3 e4c3 e2e5 b4e7 e5e7 e8e7 b2c3 h8d8 e1c1 a7a5 g1e2 a5a4 b3c2 b7b5 h1e1 b8a6 d3d4 a6c7 a2a3 e7d6 e2f4 c8d7 c1d2 d8e8 e1e3 d7e6 f4e6 e8e6 d1e1 e6f6 f2f3 c7d5 e3e5 d5b6 c2d3');
INSERT INTO evaluations VALUES(22438,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(22438,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(22438,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(22438,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(22438,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(22438,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(22438,36,'cp -62','h5d1 d7d5 d2d3 d5e4 d3e4 d8d1 e1d1 c8d7 g1f3 e8c8 b1d2 g8h6 b5d3 f7f5 h1e1 h6f7 c2c3 f8c5 d1e2 c5e7 b2b4 f7d6 e2f1 e7f6 a2a4 f5f4 d2b3 d7e6 b3c5 e6f7 h2h4 b7b6 c5a6 f7c4 d3c4 d6c4 f1e2');
INSERT INTO evaluations VALUES(41311,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41311,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(41311,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(41311,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41311,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(41311,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(41311,36,'cp 6','h2h3 f6d7 g1f3 b8c6 f4h2 e8g8 f1e2 e7e5 e1g1 g8h8 c3c4 e5e4 f3d2 f7f5 b1c3 g6g5 a2a4 d7f6 a4a5 a7a6 c3d5 h7h6 d1b3 f5f4 d5f6 d8f6 d2e4 f6e7 e4c3 f4e3 c3d5 c6d4 d5e7 d4b3 e7g6 h8g8 f2e3 b3a1 g6f8 a1c2');
INSERT INTO evaluations VALUES(41309,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41309,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41309,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41309,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(41309,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41309,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41309,36,'cp 0','e8g8 f4h2 f6d7 g1f3 e7e5 f1e2 b8c6 e1g1 g8h8 c3c4 e5e4 f3d2 f7f5 b1c3 g6g5 a2a4 d7f6 a4a5 a7a6 d1c2 a8b8 f1b1 c6b4 c2b3 b4d3 b3c2');
INSERT INTO evaluations VALUES(36599,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36599,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36599,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36599,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36599,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36599,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36599,36,'cp 11','g1f3 f6d7 f4h2 e7e5 f1e2 b8c6 a2a4 f7f5 d4e5 d6e5 b1d2 d8e7 a4a5 a7a6 d1c2 g8h8 b2b4 e5e4 f3d4 d7f6 e1g1 f6d5 d2b3 c6d4 b3d4 c8e6 f1d1 a8d8 b4b5 a6b5 e2b5 c7c6 b5f1 e6g8');
INSERT INTO evaluations VALUES(51508,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51508,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51508,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(51508,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51508,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(51508,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51508,36,'cp 0','e2e4 c8b7 b1c3 e7e6 g1f3 d7d5 e4d5 e6d5 d2d4 g8f6 f1b5 c7c6 d1e2 d8e7 b5d3 e7e2 e1e2 f8d6 f3e5 e8g8 h1e1 f6e8 e2f2 e8c7 c1e3 f7f6 e5f3 b7a6 d3f5 a6c8 f5d3');
INSERT INTO evaluations VALUES(15262,30,'cp -211',NULL);
INSERT INTO evaluations VALUES(15262,31,'cp -217',NULL);
INSERT INTO evaluations VALUES(15262,32,'cp -214',NULL);
INSERT INTO evaluations VALUES(15262,33,'cp -222',NULL);
INSERT INTO evaluations VALUES(15262,34,'cp -232',NULL);
INSERT INTO evaluations VALUES(15262,35,'cp -252',NULL);
INSERT INTO evaluations VALUES(15262,36,'cp -273','f3e5 d8g5 d2d4 g5g2 h1f1 f8d6 e5c6 c8d7 c6a7 c7c6 a7c6 b7c6 b5c4 e8g8 c1e3 d7g4 c4e2 g4e6 e1d2 f7f5 f2f4 e4f3 f1f3 g2h2 d1f1 f5f4 e3g1 h2h5 f3a3 h5e8 a3a8 e8a8 e2f3 f8b8 a2a4 b8b2 a4a5 a8b7 d2c1 b7b4 a5a6 e6f5 f1e2 b2c2 e2c2 f5c2');
INSERT INTO evaluations VALUES(15260,30,'cp -271',NULL);
INSERT INTO evaluations VALUES(15260,31,'cp -235',NULL);
INSERT INTO evaluations VALUES(15260,32,'cp -260',NULL);
INSERT INTO evaluations VALUES(15260,33,'cp -267',NULL);
INSERT INTO evaluations VALUES(15260,34,'cp -239',NULL);
INSERT INTO evaluations VALUES(15260,35,'cp -263',NULL);
INSERT INTO evaluations VALUES(15260,36,'cp -216','d8g5 d2d4 g5g2 h1f1 f8d6 e5c6 c8d7 c6a7 c7c6 a7c6 b7c6 b5c4 e8g8 c1e3 d7g4 d1d2 g2h2 a2a4 g4f3 a4a5 g8h8 c4e2 f3g2 e1c1 f7f5 f2f4 g2f1 e2f1 h2d2 e3d2 f8e8 f1e2 g7g5 f4g5 f5f4 e2g4 f4f3 d1h1 a8a7 d2c3 a7g7 h1f1 d6f4 c1b1 h8g8 d4d5 c6d5 c3g7 g8g7 g4f3 e4f3');
INSERT INTO evaluations VALUES(965,30,'cp -226',NULL);
INSERT INTO evaluations VALUES(965,31,'cp -255',NULL);
INSERT INTO evaluations VALUES(965,32,'cp -255',NULL);
INSERT INTO evaluations VALUES(965,33,'cp -256',NULL);
INSERT INTO evaluations VALUES(965,34,'cp -228',NULL);
INSERT INTO evaluations VALUES(965,35,'cp -229',NULL);
INSERT INTO evaluations VALUES(965,36,'cp -226','d2d4 g5g2 h1f1 f8d6 e5c6 c8d7 c6a7 c7c6 a7c6 b7c6 b5c4 e8g8 c1e3 d7g4 c4e2 g4e6 e1d2 f7f5 f2f4 e4f3 f1f3 g2h2 a2a3 f5f4 e3f2 a8b8 b2b4 b8e8 d1h1 h2h1 a1h1 e6d5 f3h3 d5h1 h3h1 e8a8 e2c4 g8h8 c2c3 a8a3 h1e1 a3a8 d4d5 c6d5 c4d5 a8d8 f2b6 d8b8 b6a7 b8b5 d5c6 b5b4 c3b4 d6b4 d2e2 f4f3 e2f2');
INSERT INTO evaluations VALUES(16499,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(16499,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(16499,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(16499,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(16499,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(16499,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(16499,36,'cp -54','f8c5 b1c3 e8g8 f1d3 a7a6 c3e2 f6h5 d3c4 d7d6 d2d4 e5d4 e2d4 c6d4 b2d4 c5d4 d1d4 c8g4 f3d2 b7b5 h2h3 g4d7 c4f1 d7c6 e1c1 c6b7 g2g3 f8e8 f1g2 c7c5 d4c3 a6a5 a2a4 b5a4 b3a4 h5f6 h1e1 a8b8 g2f3 d8e7 f3g2');
INSERT INTO evaluations VALUES(11439,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(11439,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(11439,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(11439,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(11439,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(11439,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(11439,36,'cp 173','f6d7 e5f3 f8d6 e1g1 e8g8 c1g5 f7f6 g5h4 b7b5 f1e1 d7e5 f3e5 d6e5 b1c3 c6c5 h4g3 e5d4 d1b1 f6f5 c3e2 d4f6 f2f3 d8e7 c2c3 f6g5 b1c2 e7f7 d3d4 f5e4 c2e4 c8f5 e4e5 g8h8 e5c5 a8e8 c5b4 f7g6 e2f4');
INSERT INTO evaluations VALUES(16593,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(16593,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(16593,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(16593,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(16593,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(16593,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(16593,36,'cp 161','f4e5 c6e5 d2d4 e5g6 e4e5 f6g8 g1f3 f8b4 h2h4 d7d5 d1e2 g6e7 e2f2 e7f5 a2a3 b4c3 b2c3 g8h6 c1h6 f5h6 f1d3 c8f5 d3f5 h6f5 f3g5 f5h6 e1g1 e8g8 f2g3 d8d7 a1b1 d7g4 g3g4 h6g4 b1b7 a8e8 g5h3 f7f6 e5f6 f8f6 b7c7 e8e4 c7a7 e4e2 f1f6 g7f6');
INSERT INTO evaluations VALUES(16356,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(16356,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(16356,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(16356,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(16356,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(16356,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(16356,36,'cp 163','c6e5 d2d4 e5g6 e4e5 f6g8 g1f3 f8b4 h2h4 d7d5 d1e2 g6e7 e2f2 e7f5 a2a3 b4c3 b2c3 g8h6 c1h6 f5h6 f1d3 c8f5 d3f5 h6f5 f3g5 f5h6 e1g1 e8g8 a1b1 d8d7 f2g3 c7c6 f1f4 a8e8 b1f1 b7b6 a3a4 d7e7 g3d3 f7f5 c3c4 e7c7 c4d5 c6d5 d3b3 c7d7');
INSERT INTO evaluations VALUES(16962,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(16962,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(16962,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(16962,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(16962,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(16962,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(16962,36,'cp 199','d2d4 e5g6 e4e5 f6g8 g1f3 f8b4 h2h4 d7d5 h4h5 g6f8 f1e2 b4e7 e1g1 g8h6 c1h6 g7h6 d1d3 c7c6 f3h2 h8g8 c3d1 e7h4 d1e3 g8g3 f1f3 d8g5 f3g3 g5g3 a1f1 c8e6 h2g4 e6g4 e3g4 g3d3 e2d3 h4g5 c2c3 a7a6 g1h1 e8e7 d3f5 f8e6 f5h7');
INSERT INTO evaluations VALUES(16957,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(16957,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(16957,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(16957,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(16957,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(16957,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(16957,36,'cp 199','e5g6 e4e5 f6g8 g1f3 f8b4 f1d3 d7d6 d1e2 g8e7 h2h4 d6e5 h4h5 g6f4 c1f4 e5f4 h5h6 c8e6 h6g7 h8g8 e1f2 b4c3 b2c3 g8g7 e2e5 e8f8 h1h7 g7h7 d3h7 e7d5 c3c4 d5f6 h7f5 e6f5 e5f5 d8d6 a1e1 a8e8 c4c5 d6d8 e1e8 f6e8 f5f4 d8d5 f4h6 f8e7 h6g5 e8f6 g5f4 d5a2 f4e5 a2e6 e5c7 e6d7');
INSERT INTO evaluations VALUES(17067,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(17067,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(17067,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(17067,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(17067,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(17067,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(17067,36,'cp 212','e4e5 f6g8 g1f3 f8b4 f1d3 d7d6 d1e2 g8e7 h2h4 d6e5 h4h5 g6f4 c1f4 b4c3 b2c3 e5f4 h5h6 g7g6 e2e5 h8f8 e5f4 f7f6 e1g1 d8d6 f4e3 c8f5 d3f5 g6f5 a1e1 f8f7 c3c4 f5f4 e3b3 e8c8 c4c5 d6d5 c2c4 d5h5 d4d5 c7c6 f3d4 c6d5 d4b5 d5c4 b3c4 h5d5 f1f4 d5c4 f4c4');
INSERT INTO evaluations VALUES(17065,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(17065,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(17065,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(17065,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(17065,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(17065,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(17065,36,'cp 207','f6g8 g1f3 f8b4 f1d3 d7d6 d1e2 g8e7 h2h4 d6e5 h4h5 g6f4 c1f4 e5f4 h5h6 c8e6 h6g7 h8g8 e1f2 b4c3 b2c3 g8g7 e2e5 e8f8 f3g5 e6f5 g5h7 f5h7 h1h7 g7h7 d3h7 e7g8 a1h1 d8f6 e5c7 a8e8 h7d3 b7b6 c7d7 e8e3 d7c8 e3e8 c8f5 e8e3 h1h3 f6f5 d3f5 g8e7 f5d3');
INSERT INTO evaluations VALUES(50372,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(50372,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(50372,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50372,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50372,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50372,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50372,36,'cp 0','d2d4 e7e6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(10937,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(10937,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(10937,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(10937,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(10937,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(10937,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(10937,36,'cp 32','e8g8 e1g1 e7b4 d4c6 d7c6 d1e2 d8e7 e4e5 b4c3 b2c3 f6d5 c4d3 f8e8 f1e1 f7f5 e5f6 e7e2 d3e2 g7f6 c1d2 c8f5 c3c4 d5b6 a1c1 a7a5 d2f4 f5e6 f4c7 b6c4 e2d3 e6d5 e1e8 a8e8 g1f1');
INSERT INTO evaluations VALUES(4606,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(4606,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(4606,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4606,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(4606,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(4606,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(4606,36,'cp 41','e1g1 e7b4 d4c6 d7c6 c4d3 f8e8 h2h3 f6d7 a2a3 b4c3 b2c3 d7e5 c1f4 d8e7 f4g3 e7c5 d1d2 c8e6 g1h2 a8d8 f2f4 e5c4 d2e2 c4b2 c3c4 e6c4 g3f2 c5b5 a1b1 b2d3 b1b5 c6b5 e2f3 d3f4 f1d1 f4e6 d1d8 e8d8');
INSERT INTO evaluations VALUES(4605,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(4605,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(4605,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(4605,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(4605,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(4605,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(4605,36,'cp 48','e7b4 d4c6 d7c6 d1e2 d8e7 e4e5 b4c3 b2c3 f8e8 f1e1 e7c5 c4d3 c8g4 e2e3 c5e3 e1e3 f6d7 f2f4 d7c5 h2h3 c5d3 c2d3 g4d7 g2g4 c6c5 f4f5 g7g6 f5f6 a7a6 g1f2 d7e6 c3c4 b7b5 c1a3 b5c4 a3c5 c4d3');
INSERT INTO evaluations VALUES(23029,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(23029,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(23029,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(23029,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(23029,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(23029,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(23029,36,'cp 201','e2e3 d4e6 b2e5 g8f6 g1f3 a7a6 b5a6 b7b6 a2a4 c7c6 f1e2 f8e7 e1g1 c8a6 e2a6 a8a6 c2c4 e8g8 b1c3 d7d5 c4d5 f6d5 d1e2 d8a8 c3d5 c6d5 f1b1 f8c8 h2h3 e7f8 b1b5 e6c5 a1c1 a6a4 b5b6 c5d7 c1c8 a8c8 b6b5 d7e5');
INSERT INTO evaluations VALUES(23028,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(23028,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(23028,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(23028,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(23028,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(23028,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(23028,36,'cp 196','d4e6 b2e5 a7a6 b5a6 b7b6 g1f3 g8f6 a2a4 c7c6 f1e2 c8a6 e1g1 f8e7 e2a6 a8a6 c2c4 e8g8 b1c3 d7d5 c4d5 f6d5 d1e2 d8a8 c3d5 c6d5 f1b1 f8c8 h2h3 e7f8 b1b5 e6c5 a1b1 c5a4 e5d4 a8c6 f3e5');
INSERT INTO evaluations VALUES(22978,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(22978,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(22978,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(22978,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(22978,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(22978,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(22978,36,'cp 201','b2e5 g8f6 g1f3 a7a6 b5a6 b7b6 a2a4 c7c6 f1e2 f8e7 a6a7 a8a7 c2c4 c8a6 e1g1 e8g8 b1c3 d7d5 d2d3 d5c4 d3c4 d8d1 f1d1 f8c8 f3d4 f6d7 e5g3 e6d4 e3d4 c8a8 a1b1 a8e8 e2f1 e7f6 g3d6 a6b7 d6c7 b7a6 c7b6 d7b6 b1b6 c6c5 c3b5 a6b5 c4b5');
INSERT INTO evaluations VALUES(22977,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(22977,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(22977,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(22977,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(22977,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(22977,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(22977,36,'cp 199','a7a6 b5a6 b7b6 g1f3 g8f6 a2a4 c7c6 f1e2 c8a6 e2a6 a8a6 e1g1 f8e7 c2c4 e8g8 b1c3 d7d5 c4d5 f6d5 d1e2 a6a5 f1b1 e6c5 e2c4 d8d7 e3e4 d5c3 e5c3 a5a6 f3e5 d7e8 d2d4 a6a4 c4e2 a4a1 c3a1 c5e6 e2c2 e7d6');
INSERT INTO evaluations VALUES(26244,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(26244,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(26244,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(26244,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(26244,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(26244,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(26244,36,'cp 19','c2c4 g8f6 e1g1 g7g6 b5c6 b7c6 d1a4 g4f3 a4c6 f6d7 g2f3 d5c4 c6c4 f8g7 f1d1 e8g8 b1c3 d7b6 c4e2 e7e6 g1g2 d8h4 e2e4 h4h5 c1f4 a8c8 a1c1 h5a5 e4e2 f8d8 f4e5 b6d7 e5g7 a5g5 g2h1 g8g7 e2e3 g5h4');
INSERT INTO evaluations VALUES(53079,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(53079,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53079,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(53079,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(53079,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(53079,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(53079,36,'cp 30','c2c3 e7e6 d2d4 g8f6 f1g2 b8c6 g1e2 f8e7 b1a3 e8g8 e1g1 c8d7 h2h3 a8c8 a3c2 c5d4 c3d4 f8e8 g1h2 d8b6 e2c3 h7h6 a1b1 c6a5 c1f4 e7f8 c2a3 d6d5 e4e5 f6h7 a3c2 e8d8 c3e2 a5c4 b2b3 c4a3 c2a3 f8a3 f4e3 h7f8 e2f4 a3b4 a2a3 b4e7 d1d2');
INSERT INTO evaluations VALUES(7617,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(7617,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(7617,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(7617,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(7617,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(7617,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(7617,36,'cp -21','d8e7 f2f3 f6d5 c3d5 c6d5 d1e2 d5e4 e2e4 e7e4 f3e4 b4d6 e1c1 c8b7 f1d3 e8c8 g2g3 h8e8 h1e1 e8e7 d2e3 c8b8 e3f4 d6e5 e1f1 f7f6 f4e5 e7e5 f1f5 e5e7 f5b5 d8e8 b5b4 a7a5 b4b5 e7e5 b5e5 e8e5 d1d2 g7g6 d3c4 d7d6');
INSERT INTO evaluations VALUES(10689,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10689,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10689,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10689,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10689,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10689,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10689,36,'cp 7','b2c3 d7d5 c3b4 d5c4 f1e1 c6e7 d1a4 c8d7 b4b5 e8g8 a4c4 d7g4 f3g5 h7h6 d4d5 g4h5 c1a3 f8e8 a3e7 e8e7 g5h3 e7e1 a1e1 h5g6 h3f4 g6f5 c4c5 a7a6 b5a6 a8a6 d5d6 d8g5 d6c7 a6c6 c5b4 c6c7 f4d5 c7c1 d5e7 g8h8 e7f5 c1e1 b4e1 g5f5');
INSERT INTO evaluations VALUES(10688,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10688,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10688,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10688,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10688,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10688,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10688,36,'cp 0','d7d5 f1e1 b4e7 c4d3 e8g8 h2h3 b7b6 a2a4 c8e6 a1a2 d8d7 a2e2 e7d6 d3b1 e6f5 b1a2 a8e8 a2d5 h7h6 f3h4 f5h7 h4f3 h7f5');
INSERT INTO evaluations VALUES(28042,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(28042,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(28042,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(28042,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(28042,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(28042,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(28042,36,'cp 107','d7d6 h2h4 h7h5 e1g1 c7c5 d4d5 b8d7 c1f4 g8f6 d3e2 e8g8 a2a4 d8c8 d1d3 f6g4 f3g5 d7e5 d3d2 b7a6 f2f3 a6e2 d2e2 g4f6 f4g3 f6h7 g5h3 a7a6 a1d1 c8b7 f1e1 c5c4 f3f4 e5g4 e4e5 a8c8');
INSERT INTO evaluations VALUES(27901,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(27901,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(27901,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(27901,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(27901,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(27901,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(27901,36,'cp 103','h2h4 h7h6 c1f4 e7e6 e4e5 g8e7 e5d6 c7d6 c3e4 b7e4 d3e4 b8c6 c2c3 e8g8 d1d2 h6h5 d2e2 f8e8 e1g1 d8d7 a2a4 a7a6 e4d3 e7d5 f4g5 d5c7 f1d1 c6e7 f3h2 e7c6 e2d2 e6e5 d4e5 c6e5 d3e4 d6d5 e4d5 c7d5 d2d5 d7d5 d1d5 e5c4');
INSERT INTO evaluations VALUES(31969,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(31969,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(31969,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(31969,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(31969,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(31969,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(31969,36,'cp 72','e7e6 f1d3 f5d3 d1d3 c7c6 c1f4 g8f6 e1c1 f6d5 f4d2 b8d7 c1b1 d5c3 d3c3 d7f6 f3e5 f6e4 c3e3 e4d2 d1d2 f8d6 f2f4 d8c7 h1d1 e8g8 g2g4 a8d8 h2h4 c6c5 g4g5 g7g6 a2a4 c5c4 d4d5 d6e5 f4e5 d8d5 d2d5 e6d5 d1d5');
INSERT INTO evaluations VALUES(32372,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(32372,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(32372,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(32372,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(32372,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(32372,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(32372,36,'cp 74','f1d3 f5d3 d1d3 c7c6 h2h4 d8c7 c1d2 b8d7 g2g4 h7h5 g4g5 g8e7 c3e4 g7g6 e1c1 e8c8 d3a3 c8b8 a3d6 e7f5 d6c7 b8c7 d2f4 c7c8 f4e5 h8h7 c2c3 d7e5 f3e5 f8d6 f2f4 h7g7 d1e1 d6c7 c1c2 a7a5 h1h3 a5a4 a2a3');
INSERT INTO evaluations VALUES(32373,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(32373,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(32373,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(32373,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(32373,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(32373,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(32373,36,'cp 83','f5d3 d1d3 c7c6 c1g5 g8f6 e1c1 f8e7 h2h4 b8d7 c1b1 e8g8 c3e4 f6e4 d3e4 d7f6 g5f6 e7f6 f3g5 g7g6 e4e3 d8d5 g5e4 f6g7 h4h5 f8d8 h5h6 g7h8 h1e1 d5f5 f2f3 e6e5 d4e5 f5e5 e3b3 d8d5 c2c4 d5d1 e1d1 b7b6 g2g4 e5f4 g4g5 h8e5 b3d3');
INSERT INTO evaluations VALUES(6865,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(6865,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(6865,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6865,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(6865,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(6865,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(6865,36,'cp 35','c5d4 d1d4 d8f6 d4b4 f6b6 b4b6 a7b6 f1d3 d7d6 c2c4 c8e6 b1c3 g7g5 b2b3 g5g4 e1d2 g8f6 c3d5 e6d5 e4d5 h7h5 a2a4 h5h4 a1e1 e8f8 e1e3 h8h5 h2h3 h5g5 h3g4 g5g4 g2g3 h4g3 h1h8 g4g8 h8g8 f6g8 e3g3 g8f6 d3f5 f8e7');
INSERT INTO evaluations VALUES(6904,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(6904,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(6904,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(6904,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(6904,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(6904,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(6904,36,'cp 35','d1d4 d8f6 e4e5 f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 f2f3 d7d5 a2a3 d8e7 e1f2 f6f5 f1d3 c8d7 c2c3 e7f6 a1d1 c7c6 d3c2 a8e8 d1d4 b6b5 c2d3 h7h5 d4h4 d7e6 h1e1 e8a8 d3c2 e6d7 c2d1 d7e8 f3f4 a8d8 e1e5');
INSERT INTO evaluations VALUES(46757,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(46757,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(46757,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(46757,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(46757,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(46757,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46757,36,'cp 52','c4d5 f6d5 g1f3 f8g7 e3e4 d5b6 a2a4 a7a5 f1b5 c7c6 b5e2 e8g8 c1e3 b8a6 e1g1 a6b4 f1e1 b6d7 b1c3 e7e5 d4d5 d7f6 d5c6 b7c6 f3e5 d8d1 e1d1 f6e4 e5c6 e4c3 c6e7 g8h8 b2c3 b4c2 a1b1 c2e3 f2e3 c8e6 c3c4 f8d8 e7d5 a8c8 g1f2 c8c5 b1b7 e6d5');
INSERT INTO evaluations VALUES(12933,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(12933,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(12933,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(12933,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(12933,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(12933,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(12933,36,'cp 108','d7d6 c1g5 c8d7 e1c1 d8a5 f2f4 h7h6 g5f6 g7f6 b2b4 a5d8 f1c4 f8g7 d1d3 a8c8 c4b3 a7a5 b4b5 d8c7 c3d5 c7c5 d5b6 c5d4 d3d4 c8c5 a2a4 h6h5 d4d5 c5c7 d5d3 d7e6 h1d1 e6b3 d3b3 h5h4 c1b1 h8g8 d1d5 g7h6 g2g3 h6g7 d5d1 c7c5 g3h4 f6f5');
INSERT INTO evaluations VALUES(18272,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(18272,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(18272,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18272,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(18272,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(18272,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(18272,36,'cp -32','c6c5 c2c3 f7f5 e4f5 c8f5 d2d4 c5d4 c3d4 e5e4 f3d2 g8f6 b1c3 f8e7 d2f1 d6d5 f1g3 f5e6 d1a4 e6d7 a4a5 c7c6 a5d8 a8d8 e1g1 e8g8 c3a4 f6e8 c1e3 e8c7 a4c5 e7c5 d4c5 c7e6 f1d1 a7a5 d1d2 f8f7 g3e2 a5a4 a1d1');
INSERT INTO evaluations VALUES(11452,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(11452,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(11452,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(11452,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(11452,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(11452,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(11452,36,'cp 196','f3e5 f8c5 e5f3 d8e7 d1e2 c8d7 b1c3 e8c8 a2a3 c5b6 c1e3 c6c5 e1g1 b6a5 e3d2 h8e8 c3d1 a5d2 f3d2 f6d5 e2f3 d5b6 d1c3 g7g6 a3a4 d7c6 a4a5 b6d7 f3e3 f7f5 a1e1 c8b8 f2f4 f5e4 d3e4 d7f6 h2h3 d8d4 d2f3');
INSERT INTO evaluations VALUES(35303,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(35303,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(35303,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(35303,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(35303,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(35303,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35303,36,'cp 41','d1e2 d7d5 e4d5 g8f6 b1c3 b8d7 d2d4 d7b6 c1f4 f6d5 c3d5 b6d5 f4d2 c7c6 e2e7 f8e7 e1c1 e8g8 c2c4 d5c7 d1e1 f8e8 d2a5 b7b6 a5c3 f7f6 f1d3 c8d7 h1f1 e7d6 c1c2 e8e1 f1e1 a8e8 e1f1 d7e6 a2a4 e6f7 b2b3 f7g6 d3g6 h7g6');
INSERT INTO evaluations VALUES(43032,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(43032,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(43032,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43032,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(43032,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(43032,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(43032,36,'cp 29','g1f3 d7d6 e1g1 e8g8 f1e1 c7c5 a2a4 c8e6 c2c3 b8c6 b1a3 h7h6 c4e6 f7e6 c1d2 f6d7 a3c4 d8e8 b2b4 c5b4 c3b4 e8f7 d2c3 a7a6 b4b5 c6d4 c3d4 e5d4 d1e2 a6b5 a4b5 e6e5 h2h3 g8h7 e1b1 d7c5 c4b6 a8d8 f3d2 e7g5 d2c4 f7g6');
INSERT INTO evaluations VALUES(43029,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43029,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43029,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43029,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(43029,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43029,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(43029,36,'cp 16','d7d6 e1g1 c7c5 b1c3 b8c6 f1e1 e8g8 c1g5 f6e8 g5d2 e8c7 c3d5 c7d5 c4d5 c8e6 c2c3 d8d7 a2a4 h7h6 b2b4 c5b4 c3b4 a7a6 h2h3 e6d5 e4d5 c6d4 f3d4 e5d4 d2f4 e7f6 e1e4 a8c8 d1g4 d7g4 h3g4 f8d8 g2g3');
INSERT INTO evaluations VALUES(6498,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6498,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(6498,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6498,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6498,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6498,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6498,36,'cp 0','g8f6 b1c3 d7d5 c1g5 d5e4 d3e4 d8d1 a1d1 c5b4 g5d2 b4d6 d2g5');
INSERT INTO evaluations VALUES(41312,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(41312,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41312,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(41312,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(41312,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41312,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41312,36,'cp 0','e8g8 h2h3 f6d7 f1e2 e7e5 f4h2 f7f5 a2a4 b8c6 d4e5 d7e5 f3e5 c6e5 e1g1 g8h8 b1d2 d8e7 d2f3 c8e6 f3d4 e6d7 d4b5 d7c8 b5d4');
INSERT INTO evaluations VALUES(42973,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(42973,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42973,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(42973,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(42973,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(42973,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(42973,36,'cp 7','c2c3 c7c5 e2e3 e8g8 f1d3 b7b6 b1d2 c8a6 d3a6 b8a6 d1e2 a6b8 e1g1 d7d5 a1d1 b8c6 f3e5 a8c8 h2h3 f6d7 e5c6 c8c6 g5e7 d8e7 e3e4 c5d4 e4d5 c6d6 c3d4 d6d5 d2e4 d7f6 e4c3 d5d7 d4d5 f6d5 c3d5 e6d5 e2e7 d7e7');
INSERT INTO evaluations VALUES(47234,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47234,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(47234,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(47234,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(47234,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(47234,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47234,36,'cp 0','e4e5 f6g8 d2d4 c7c5 d4c5 e7e6 b1d2 f8c5 d2b3 c5b6 f1d3 d8h4 g2g3 h4h5 f3f4 b6f2 e1d2 f2e3 d2e1 e3f2');
INSERT INTO evaluations VALUES(47297,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47297,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47297,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47297,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47297,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47297,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47297,36,'cp 0','f6g8 d2d4 c7c5 d4c5 e7e6 b1d2 f8c5 d2b3 c5b6 f1d3 d8h4 g2g3 h4h5 f3f4 b6f2 e1d2 f2e3 d2e1 e3f2');
INSERT INTO evaluations VALUES(49370,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(49370,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(49370,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(49370,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(49370,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(49370,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(49370,36,'cp 0','e4e5 d7d5 c4e2 f6d7 d2d4 c7c5 c2c3 b8c6 f2f4 d8a5 b1d2 a5b6 d2b3 a7a5 a2a4 c5d4 b3d4 c6d4 c3d4 f8b4 e1f1 d7b8 c1e3 b4e7 h2h4 b6b2 a1b1 b2a2 h4h5 b8c6 b1a1 a2b2 h5h6 g7g6 a1b1 b2a2 b1a1');
INSERT INTO evaluations VALUES(49348,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(49348,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(49348,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(49348,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(49348,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(49348,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(49348,36,'cp 15','d7d5 c4e2 f6d7 d2d4 c7c5 c2c3 b8c6 f2f4 d8a5 b1d2 a5b6 d2b3 a7a5 a2a4 c5d4 c3d4 f8b4 e1f1 b6c7 g1f3 d7b6 h2h4 b6c4 h1h3 b7b6 f3e1 g7g6 h4h5 g6h5 e1c2 h8g8 c2b4 a5b4 f1g1 c8a6 e2h5 e8c8');
INSERT INTO evaluations VALUES(5637,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(5637,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(5637,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(5637,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(5637,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5637,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(5637,36,'cp 28','a2a4 g8f6 e1g1 e8g8 c2c3 a7a5 b1a3 c8e6 f1e1 e6c4 a3c4 h7h6 d1b3 d8d7 c4e3 b7b6 c1d2 a8d8 e3d5 f6d5 b3d5 e7f6 d3d4 f8e8 h2h3 c6e7 d5b5 e5d4 c3d4 e7g6 a1c1 c7c5 b5d7 d8d7 e4e5 d6e5');
INSERT INTO evaluations VALUES(26691,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(26691,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(26691,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(26691,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(26691,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(26691,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(26691,36,'cp 147','d4d5 c6b8 h2h3 g4f3 d1f3 g7g6 f3b3 d8c8 b1c3 f8g7 c1e3 b8d7 g2g4 c7c6 e1c1 g8f6 g4g5 f6h5 f1e2 e8g8 e2g4 c8c7 c1b1 f8c8 h1e1 g7c3 b3c3 c6d5 c3c7 c8c7 d1d5 d7b6 d5d1 b6c4 e3c1 c4e5 g4e2 e5c4 a2a4 a8c8 e2g4 e7e6 b2b3 c4a5 f2f4');
INSERT INTO evaluations VALUES(26688,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(26688,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(26688,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(26688,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(26688,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(26688,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(26688,36,'cp 157','c6b8 h2h3 g4f3 d1f3 g7g6 f3b3 d8c8 b1c3 f8g7 c1e3 g8f6 g2g4 b8d7 g4g5 f6h5 f1e2 c7c5 e2h5 g6h5 c3e2 h5h4 e2f4 c5c4 b3a3 e8g8 f4h5 g7h8 f2f4 c4c3 b2b3 f7f5 g5f6 d7f6 h5f6 h8f6 e1c1 c8d7 a3a5 b7b5 a5a6 a8c8 c1b1');
INSERT INTO evaluations VALUES(21131,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(21131,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(21131,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(21131,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(21131,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(21131,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(21131,36,'cp 202','d4d5 c6b8 c1f4 g8f6 f1c4 a7a6 g1f3 g7g6 d1e2 f8g7 e1c1 e8g8 h2h3 c8d7 f3e5 d7b5 h1e1 b5c4 e2c4 f6e8 c4e2 e8d6 g2g4 f8e8 f4g3 b8d7 f2f4 d6b5 c3e4 b5d6 e2f3 a6a5 h3h4 a5a4 a2a3 d6e4 e1e4 d7c5 e4e1 d8d6 h4h5');
INSERT INTO evaluations VALUES(20959,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(20959,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(20959,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(20959,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(20959,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(20959,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(20959,36,'cp 205','c6b8 c1f4 g8f6 f1c4 a7a6 g1f3 g7g6 d1e2 f8g7 e1c1 e8g8 h2h3 b7b5 c4b3 b5b4 c3a4 c8d7 e2c4 a6a5 c4c7 d7a4 b3a4 b8d7 c7d8 f8d8 a4c6 a8c8 f4e3 c8b8 h1e1 d7b6 e3b6 b8b6 f3e5 g8f8 g2g4 g7h6 c1b1 b6b8 d1d4 e7e6 d5d6 h6g7 d6d7 f6d5 c6d5 e6d5');
INSERT INTO evaluations VALUES(24803,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(24803,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(24803,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(24803,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(24803,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(24803,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(24803,36,'cp 108','c4d5 f6d5 d1b3 d5c3 b2c3 d8d7 e2e3 c6a5 b3b4 b7b6 f3e5 d7c8 d4d5 c7c6 e3e4 f5d7 c1e3 e7e6 b4a4 e6d5 e4d5 f8d6 e5d7 c8d7 d5c6 a5c6 a4e4 d6e7 f1b5 a8c8 e1g1 d7e6 e4e6 f7e6 a1d1 e8g8 d1d7 e7f6 a2a4 f8d8 b5c4 d8e8 c4a6 c8d8 d7c7 c6e7 c7a7 e7d5 a6b5 e8e7 a7e7 f6e7 b5d7');
INSERT INTO evaluations VALUES(24789,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(24789,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(24789,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(24789,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(24789,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(24789,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(24789,36,'cp 144','f6d5 d1b3 d5c3 b2c3 d8d7 e2e3 c6a5 b3b4 b7b6 f3e5 d7c8 d4d5 e7e6 b4a4 c7c6 e3e4 f8d6 f2f4 d6e5 f4e5 f5g6 c1a3 e6d5 e4d5 c8f5 f1e2 f5e4 a4e4 g6e4 c3c4 a8c8 d5d6 e4f5 c4c5 f5e6 a1c1 e8d7 e1g1 c8b8 h2h4 b6b5 a3b2 a5c4 b2c3 a7a5 e2d3 h8d8 d3h7 g7g6 h4h5 c4e3');
INSERT INTO evaluations VALUES(25010,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(25010,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(25010,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(25010,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(25010,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(25010,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(25010,36,'cp 157','d1b3 d5c3 b2c3 d8d7 e2e3 c6a5 b3b4 b7b6 f3e5 d7c8 d4d5 e7e6 b4a4 c7c6 e3e4 f8d6 f2f4 d6e5 f4e5 f5g6 c1a3 e6d5 e4d5 c8f5 f1e2 f5e4 a4e4 g6e4 c3c4 a8c8 d5d6 c6c5 e1g1 e4c6 a3b2 c6d7 b2c3 a5c6 a2a4 e8g8 e2f3 a7a5 g1f2 f8e8 f1e1 c8d8 f3d5 c6d4 c3d4 c5d4');
INSERT INTO evaluations VALUES(51407,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51407,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(51407,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(51407,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(51407,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(51407,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(51407,36,'cp 20','g8f6 a2a3 c8b7 c1f4 c7c5 b1c3 c5d4 d1d4 b8c6 d4d3 f6h5 f4g5 f8e7 g5e7 d8e7 c3b5 e8g8 b5d6 c6d8 e2e3 h5f6 a1d1 b7c6 f3e5 d8b7 e5c6 d7c6 d6b7 e7b7 f1e2 a8c8 e2f3 b7c7 d3d6 b6b5 d6c7 c8c7 c4b5');
INSERT INTO evaluations VALUES(52167,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(52167,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(52167,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(52167,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(52167,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(52167,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(52167,36,'cp 152','e2e4 g8f6 b1d2 f8e7 a2a4 e8g8 f1d3 f6e8 g1e2 d7d6 e1g1 b7b6 d2c4 c8a6 e2c3 e7g5 c1g5 d8g5 c4e3 a6d3 d1d3 a7a6 g2g3 b8d7 g1h1 g5f6 f2f4 e8c7 d3e2 e5f4 g3f4 b6b5 e2g2 b5b4 c3d1 a8e8 e3c4');
INSERT INTO evaluations VALUES(52165,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(52165,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(52165,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(52165,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(52165,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(52165,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(52165,36,'cp 137','g8f6 b1d2 d7d6 a2a4 f8e7 h2h3 b8a6 g1e2 e8g8 g2g4 a6c7 e2g3 f6e8 f1d3 c8d7 d2c4 e7g5 c1g5 d8g5 c4e3 b7b5 a4b5 c7b5 h3h4 g5d8 c2c3 a7a5 g3f5 b5c7 d1f3 a5a4 e3c4 c7b5 g4g5 f7f6 f3e3 g8h8 e3h3');
INSERT INTO evaluations VALUES(44345,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(44345,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(44345,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(44345,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(44345,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(44345,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(44345,36,'cp 156','b1d2 f8e7 a2a4 e8g8 f1d3 d7d6 g1e2 f6e8 h2h3 b7b6 d2c4 b8a6 e2g3 e7g5 c1g5 d8g5 h3h4 g5f4 c2c3 a6c7 h4h5 c8d7 d1e2 d7g4 e2e3 g4d7 g3e2 f4f6 e3d2 f6e7 e2g3 h7h6 c4e3 g8h8');
INSERT INTO evaluations VALUES(53727,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(53727,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53727,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53727,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(53727,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(53727,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53727,36,'cp 12','f1d3 f8d6 e1g1 g8e7 c2c4 d5c4 d3c4 e8g8 b1c3 b8a6 c4a6 b7a6 f1e1 h7h6 c1d2 c8e6 c3e4 f8e8 d2c3 e6d5 e4d6 d8d6 f3e5 a6a5 a2a3 a8b8 d1h5 d6e6 e5d3 e6f5 h5f5 e7f5 e1e8 b8e8 c3a5 f5d4 a1e1 e8e1 a5e1 d4c2 g1f1 d5c4');
INSERT INTO evaluations VALUES(53722,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(53722,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(53722,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53722,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(53722,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53722,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(53722,36,'cp 21','f8d6 e1g1 g8e7 c1g5 h7h6 g5h4 d8c7 c2c4 d5c4 d3c4 e8g8 h4e7 d6e7 b1c3 c8g4 h2h3 g4f3 d1f3 b8d7 a1e1 a8e8 c4b3 d7f6 d4d5 f6d5 b3d5 c6d5 c3d5 c7d7 e1d1 e7d6 f1e1 e8e1 d1e1 f8e8 e1d1 d7c6');
INSERT INTO evaluations VALUES(38088,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(38088,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38088,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(38088,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(38088,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38088,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(38088,36,'cp 26','e1g1 g8e7 f1e1 e8g8 c1g5 h7h6 g5h4 d8c7 c2c4 d5c4 h4e7 d6e7 d3c4 e7d6 b1c3 b8d7 d1d3 d7f6 f3e5 c8d7 d3g6 d7e8 a1c1 a8d8 c4b3 c6c5 g6f5 e8d7 f5f3 d7c6 e5c6 c7c6 f3c6 b7c6 d4c5 d6c5 g2g3 d8d2 e1e2 f8d8 e2d2 d8d2');
INSERT INTO evaluations VALUES(38087,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(38087,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38087,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(38087,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38087,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(38087,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(38087,36,'cp 21','g8e7 c2c4 d5c4 d3c4 e8g8 b1c3 b8a6 c1g5 a6c7 f1e1 h7h6 g5h4 c8e6 c4e6 c7e6 d4d5 c6d5 c3d5 f8e8 h2h3 d8a5 d5f6 g7f6 d1d6 e7f5 d6d3 e6f4 e1e8 a8e8 d3c4 f5h4 c4f4 h4f3 f4f3 e8e1 a1e1 a5e1 g1h2 e1a1 f3b3 b7b5 a2a3 a7a6 b3g3 g8h7 b2b4 f6f5 g3f3 a1f6 f3e2 h7h8 e2d2 h8g7');
INSERT INTO evaluations VALUES(40254,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40254,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40254,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(40254,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40254,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(40254,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(40254,36,'cp 17','c2c4 d5c4 d3c4 e8g8 b1c3 b8a6 f1e1 a6c7 c1g5 c8e6 g5e7 d8e7 c4e6 c7e6 d4d5 c6d5 c3d5 e7d7 f3e5 d7d8 g2g3 h7h6 e5f3 f8e8 d1b3 d8a5 a2a3 a5a6 b3c2 a8d8 a1d1 d6f8 c2f5 d8d6 g1g2 b7b6 d1d2 e8d8 e1d1 a6b7 b2b4');
INSERT INTO evaluations VALUES(15721,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(15721,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(15721,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(15721,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(15721,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(15721,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(15721,36,'cp -34','f3e5 f8e7 d2d3 e8g8 h2h3 f8e8 c1f4 c8e6 b1d2 e7f8 d2e4 f7f6 e5f3 e6d5 f3d2 d8d7 d1g4 d7f7 e4d6 c7d6 a2a4 a8d8 g4g3 d5e6 d2e4 f7g6 g3f3 e6d5 f3g4 g6g4 h3g4 d5e6 f2f3 d6d5 e4g3 c6c5 f4d2');
INSERT INTO evaluations VALUES(15717,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(15717,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(15717,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(15717,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(15717,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(15717,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(15717,36,'cp -34','f8e7 d2d3 e8g8 c1f4 f8e8 h2h3 c8e6 b1d2 e7f8 e5g4 f7f6 g4e3 d6c8 d1f3 c8b6 a2a4 d8d7 a4a5 b6d5 e3d5 c6d5 d3d4 e6f7 h3h4 e8e1 a1e1 b7b6 f3b3 d7c6 e1a1 a8e8 d2f3 c6d7 c2c3 f7h5 a5b6 h5f3 g2f3');
INSERT INTO evaluations VALUES(8669,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(8669,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(8669,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(8669,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(8669,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(8669,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(8669,36,'cp -18','d2d4 e8g8 c1f4 f8e8 b1d2 e7f8 c2c3 f7f6 e5c4 e8e1 d1e1 c8g4 c4d6 c7d6 d4d5 d8a5 d5c6 b7c6 b2b4 a5f5 f4e3 f5d3 d2b3 a8e8 h2h3 g4d7 a1d1 d3g6 g1h1 d7e6 e1d2 e6b3 a2b3 g6f7 e3f4 e8d8 d1e1');
INSERT INTO evaluations VALUES(8668,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(8668,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(8668,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(8668,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(8668,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(8668,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(8668,36,'cp -26','e8g8 c1f4 f7f6 e5f3 f8e8 d4d5 c6c5 b1d2 e7f8 c2c4 e8e1 d1e1 d8e8 f4d6 e8e1 f3e1 c7d6 f2f4 c8f5 g2g3 a8e8 g1f2 g7g5 e1g2 f8g7 a1e1 e8e1 f2e1 f5d3 b2b3 g8f7 e1f2 g5f4 g2f4 d3c2 f2e1 f6f5 f4e6 g7c3 e6d8 f7f6 d8b7');
INSERT INTO evaluations VALUES(3426,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(3426,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(3426,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(3426,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(3426,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(3426,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(3426,36,'cp -25','c1f4 f8e8 b1d2 a7a5 a2a4 e7f8 c2c3 f7f6 e5d3 e8e1 d3e1 b7b6 e1c2 c8e6 c2e3 d8d7 d1f3 a8d8 h2h4 h7h6 d2f1 g8h8 f4d6 f8d6 f1g3 e6b3 g3f5 d6f8 h4h5 c6c5 f3g4 c5d4 f5d4 b3g8 g4d7 d8d7 g2g4 g8h7');
INSERT INTO evaluations VALUES(55822,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55822,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(55822,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(55822,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55822,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(55822,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55822,36,'cp 9','c7c5 e2e3 g8f6 g1f3 d8b6 d1c2 f8d6 f4g3 e8g8 b1d2 c8d7 a2a4 h7h6 f1e2 f8c8 a4a5 b6c7 g3d6 c7d6 e1g1 b8c6 d4c5 d6c5 e3e4 c8c7 e4d5 c6b4 c2b3 b4d5 f1e1 a8d8 e2f1 d7c6 b3a3 f6d7 f3d4 c5a3 a1a3 d7f6 d4c6 c7c6');
INSERT INTO evaluations VALUES(54191,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(54191,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54191,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54191,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(54191,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54191,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(54191,36,'cp 20','e2e3 g8f6 g1f3 f8e7 b1d2 e8g8 f3e5 c5c4 f1e2 b7b5 a2a4 b5b4 e2f3 c8a6 e1g1 d8c8 a1c1 b8c6 e3e4 b4c3 b2c3 a6b7 d1e2 c6e5 f4e5 b7c6 h2h3 h7h6 e4d5 e6d5 c1a1 f8e8 f1e1 c8a6 d2f1 e7d6');
INSERT INTO evaluations VALUES(2878,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2878,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2878,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(2878,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(2878,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(2878,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(2878,36,'cp 12','e1g1 d7d5 e4d5 c6d5 h2h3 c7c6 d1f3 f8e8 c1g5 h7h6 g5f6 d8f6 f3f6 g7f6 c3e2 a8b8 c2c3 b4d6 b2b4 c6c5 b4c5 d6c5 f1b1 c8e6 b1b3 c5b6 e2d4 b6d4 c3d4 b8b3 a2b3 e8b8 g1f1 b8b3 f1e2 b3b2 e2e3 f6f5 a1a7 f5f4 e3f4 b2f2 f4g3 f2d2 a7a8 g8g7');
INSERT INTO evaluations VALUES(2877,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(2877,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(2877,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(2877,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2877,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(2877,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(2877,36,'cp 9','d7d5 c1d2 b4c3 d2c3 d5e4 c3f6 d8f6 d3e4 f6b2 e4c6 a8b8 a1b1 b2b1 d1b1 b8b1 f1b1 f8d8 c6f3 g8f8 b1b8 d8e8 g2g4 c8a6 b8b3 e8e1 g1g2 a6e2 b3a3 e2f3 g2f3 e1c1 a3a7 c1c2 a2a4 f8e7 f3g3 e7d6 h2h4 c2c3 f2f3 c3a3 g3f4 c7c5 a7f7 a3a4 f4g5 a4a5 f7g7 c5c4 g5f4');
INSERT INTO evaluations VALUES(2853,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2853,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(2853,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(2853,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(2853,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(2853,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2853,36,'cp 10','c1d2 b4c3 d2c3 d5e4 c3f6 d8f6 d3e4 f6b2 e4c6 a8b8 a1b1 b2b1 d1b1 b8b1 f1b1 f8d8 c6f3 g8f8 b1b8 d8e8 g2g4 c8a6 b8b3 e8e1 g1g2 a6e2 b3a3 e2f3 g2f3 e1c1 a3a7 c1c2 a2a4 f8e7 a4a5 c2c3 f3g2 e7d7 h2h4 d7c6 f2f3 c6b5 g4g5 c3c5');
INSERT INTO evaluations VALUES(4717,30,'cp -292',NULL);
INSERT INTO evaluations VALUES(4717,31,'cp -283',NULL);
INSERT INTO evaluations VALUES(4717,32,'cp -275',NULL);
INSERT INTO evaluations VALUES(4717,33,'cp -258',NULL);
INSERT INTO evaluations VALUES(4717,34,'cp -248',NULL);
INSERT INTO evaluations VALUES(4717,35,'cp -288',NULL);
INSERT INTO evaluations VALUES(4717,36,'cp -271','b1c3 d7d5 c1g5 d5e4 c3e4 c5e7 e1g1 f6e4 g5e7 f7e7 d3e4 d8d1 f1d1 a7a5 f2f3 c8e6 g1f2 c6d4 d1d3 b7b6 h2h4 a5a4 f3f4 g7g6 a1d1 a8f8 g2g3 e6a2 c2c3 a2b3 d1d2 d4e6 d3d7 e7e8 d7h7 e5f4 g3f4 f8f4 f2e3 f4f6 h7h8 f6f8 h8f8 e6f8');
INSERT INTO evaluations VALUES(4714,30,'cp -294',NULL);
INSERT INTO evaluations VALUES(4714,31,'cp -303',NULL);
INSERT INTO evaluations VALUES(4714,32,'cp -301',NULL);
INSERT INTO evaluations VALUES(4714,33,'cp -307',NULL);
INSERT INTO evaluations VALUES(4714,34,'cp -310',NULL);
INSERT INTO evaluations VALUES(4714,35,'cp -284',NULL);
INSERT INTO evaluations VALUES(4714,36,'cp -292','d7d5 c1g5 d5e4 c3e4 c5e7 g5f6 g7f6 d1h5 f7g8 e1g1 f6f5 e4g3 f5f4 g3f5 e7f8 f5h6 f8h6 h5h6 d8f8 h6h4 c8f5 f1e1 f8d6 a1d1 g8g7 f2f3 d6f6 h4f2 a8d8 f2e2 f6d6 g1h1 d8d7 b2b4 a7a6 a2a3 d6f6 c2c3 f5g6 a3a4 b7b6 b4b5 a6b5');
INSERT INTO evaluations VALUES(19398,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19398,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19398,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(19398,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(19398,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(19398,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(19398,36,'cp 12','g8f6 g1f3 e7e5 f1e2 d7d5 d2d4 c5d4 e3d4 e5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 e2c4 f6d5 c1d2 c8e6 d1c3 d5c3 c4e6 c3e4 e6g4 e4d2 e1d2 f8c5 a1e1 e8f8 e1c1 c5f2 c1c6 h7h5 g4h3 a8e8 d2c2 f2b6 h1f1 h8h6 c6h6 g7h6 f1d1 e8e2 d1d2 e2d2 c2d2 f8e7 b2b4 e7d6');
INSERT INTO evaluations VALUES(12802,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12802,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(12802,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12802,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12802,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12802,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(12802,36,'cp 15','g1f3 e7e5 f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e3e4 c8e6 c1e3 f8e7 f3d2 e8f7 d2c4 h8d8 a2a4 d8d1 a1d1 a8d8 d1d8 c6d8 f2f3 f7e8 g1f2 e8d7 e3d2 d8f7 c4e3 f7d6 c3c4 e7d8 a4a5 e6f7 b2b3 b7b6 a5b6 a7b6 g2g3 f7g6');
INSERT INTO evaluations VALUES(12803,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12803,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(12803,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12803,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(12803,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(12803,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(12803,36,'cp 9','e7e5 f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e3e4 c8e6 c1e3 f8e7 f3d2 e8c8 e2c4 e6c4 d2c4 d8d1 a1d1 h8d8 d1d8 c6d8 a2a4 c8c7 e3d2 a7a6 c4e3 b7b5 e3d5 c7d6 a4b5 a6b5 b2b3 g7g6 d5e3 d6d7 f2f3 c5c4 b3c4 b5c4 e3c4');
INSERT INTO evaluations VALUES(12547,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12547,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(12547,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12547,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12547,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(12547,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(12547,36,'cp 15','f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e3e4 c8e6 c1e3 f8e7 f3d2 e8f7 h2h4 b7b6 f2f3 a8d8 g1f2 h7h5 d2c4 d8d1 a1d1 h8d8 d1d8 e7d8 a2a4 f7e7 c4d2 d8c7 e2a6 c6a5 f2e2 e7d6 a6b5 d6e7 e2d3 a5b7 d2c4');
INSERT INTO evaluations VALUES(26184,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26184,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(26184,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(26184,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(26184,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(26184,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(26184,36,'cp 32','f1d3 h7d3 d1d3 e7e6 c1d2 f8e7 e1c1 g8f6 g3e4 e8g8 c1b1 c6c5 g2g4 f6g4 d3e2 d8b6 f3e5 d7e5 d4e5 f7f5 e5f6 g4f6 d1g1 f6e4 e2e4 e7f6 b2b3 b6d6 d2h6 d6d4 e4d4 c5d4 h6g7 f6g7 h5h6 f8f7 g1g5 a8c8 h1g1 f7f2 h6h7 g8h8 g5g7 c8c2');
INSERT INTO evaluations VALUES(26183,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(26183,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(26183,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(26183,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26183,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(26183,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(26183,36,'cp 33','h7d3 d1d3 e7e6 c1d2 f8e7 e1c1 g8f6 g3e4 e8g8 c1b1 f6e4 d3e4 d7f6 e4e2 d8d5 f3e5 d5e4 e2e4 f6e4 d2e3 e7d6 h1g1 a8d8 c2c4 e4f6 g1h1 d8c8 b2b3 b7b5 e5d3 f8e8 b1c2 f6e4 a2a4 b5c4 b3c4 c6c5 d4c5 d6c5');
INSERT INTO evaluations VALUES(26180,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(26180,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(26180,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(26180,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(26180,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(26180,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(26180,36,'cp 45','d1d3 e7e6 c1d2 f8e7 e1c1 g8f6 g3e4 e8g8 c1b1 c6c5 g2g4 f6g4 d3e2 d8b6 f3e5 d7e5 d4e5 f7f5 e5f6 g4f6 d1g1 f6e4 e2e4 e7f6 b2b3 b6d6 d2h6 d6d4 e4d4 c5d4 h6g7 f6g7 h5h6 f8f7 h6g7 f7g7 g1e1 a8f8 f2f4 f8f4 e1e6 g7g2 e6e8 g8f7 e8b8 f4f2 b8b7 f7e6 h1d1 f2c2 b7a7 c2e2 d1d4');
INSERT INTO evaluations VALUES(26179,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(26179,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(26179,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26179,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(26179,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(26179,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(26179,36,'cp 23','e7e6 c1d2 g8f6 e1c1 f8e7 g3e4 e8g8 e4f6 d7f6 f3e5 c6c5 g2g4 d8d4 d3d4 c5d4 g4g5 h6g5 d2g5 f6d5 g5e7 d5e7 d1d4 a8d8 h1d1 d8d5 d4d5 e6d5 d1d4 g7g6 h5g6 f7g6 d4a4 a7a6 a4b4 b7b5 e5d3 g6g5 b4b3 e7g6 a2a4 b5a4 b3b4 g6f4 d3f4 f8f4 b4f4 g5f4 c1d2 g8f7 d2d3 a4a3');
INSERT INTO evaluations VALUES(26085,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(26085,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(26085,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(26085,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26085,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(26085,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(26085,36,'cp 42','c1d2 f8e7 e1c1 g8f6 g3e4 e8g8 e4f6 d7f6 f3e5 c6c5 g2g4 d8d4 d3d4 c5d4 g4g5 h6g5 d2g5 f6d5 g5e7 d5e7 d1d4 a8d8 h1d1 d8d5 d4d5 e7d5 c2c4 f8c8 c1b1 d5f6 b2b3 a7a6 a2a4 f6e4 b1b2 f7f6 e5g6 g8h7 f2f3 e4g3 d1d7 g3h5');
INSERT INTO evaluations VALUES(26086,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(26086,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(26086,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26086,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26086,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(26086,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(26086,36,'cp 32','g8f6 e1c1 f8e7 g3e4 e8g8 e4f6 d7f6 f3e5 c6c5 g2g4 d8d4 d3d4 c5d4 g4g5 h6g5 d2g5 a8d8 h5h6 f6d5 g5d2 e7b4 c2c3 d4c3 b2c3 b4d6 h6g7 g8g7 d2h6 g7f6 h6f8 d6e5 h1h8 f6f5 h8h5 f5f6 f8g7 f6g7 h5e5 d8c8 c1b2 g7f6 e5e1 d5c3 d1d7 c3a4 b2b1 a4c3 b1a1 b7b5 d7a7 c3a4 a1b1 c8c4 e1e3 a4c3 b1a1 c3d1 e3b3 c4c1 b3b1 c1b1 a1b1 d1f2 b1b2');
INSERT INTO evaluations VALUES(24492,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(24492,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(24492,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(24492,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(24492,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24492,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24492,36,'cp 25','e1c1 f8e7 g3e4 e8g8 e4f6 d7f6 f3e5 c6c5 g2g4 d8d4 d3d4 c5d4 g4g5 h6g5 d2g5 a8d8 h5h6 f6d5 g5d2 e7b4 c2c3 d4c3 b2c3 b4d6 h6g7 g8g7 d2h6 g7f6 e5g4 f6e7 h6f8 d8f8 d1d3 f8c8 c1b2 b7b5 h1h7 c8c4 g4e3 d5e3 f2e3 d6c5 e3e4 c4e4 d3f3 a7a5 f3f7 e7d6 f7d7 d6c6 d7c7 c6d5 h7h2 d5d6 c7h7 d6c6 b2c2 c5d6 h2h6 e4e2 c2d3 e2a2 h6e6 c6c5');
INSERT INTO evaluations VALUES(24491,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24491,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(24491,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24491,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24491,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(24491,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(24491,36,'cp 10','f8e7 g3e4 e8g8 e4f6 d7f6 c1b1 c6c5 g2g4 f6g4 d3e2 g4f6 h1g1 g8h8 d4c5 d8c7 d2c3 a8d8 d1d8 f8d8 b2b4 a7a5 a2a3 c7f4 c3e5 f4f5 c5c6 b7c6 f3d4 d8d4 e5d4 a5b4 a3b4 e7b4 e2e5 f5e5 d4e5 f6h5');
INSERT INTO evaluations VALUES(24016,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24016,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(24016,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24016,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24016,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(24016,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(24016,36,'cp 11','g3e4 e8g8 c1b1 c6c5 g2g4 f6g4 d3e2 d8b6 f3e5 d7e5 d4e5 f7f5 e5f6 g4f6 d1g1 f6e4 e2e4 e7f6 b2b3 c5c4 d2h6 c4b3 a2b3 f8f7 h6c1 b6f2 c1b2 f6b2 b1b2 a8d8 h5h6 f2d4 e4d4 d8d4 g1g6 d4d7 h6g7 f7g7 g6e6 g7e7 e6h6 e7h7 h6h7 d7h7 h1h7 g8h7 b2c3 h7g6 c3d4 g6f6 d4d5 f6e7 b3b4');
INSERT INTO evaluations VALUES(34587,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(34587,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(34587,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(34587,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(34587,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(34587,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(34587,36,'cp 160','b1c3 d5c6 g1f3 c8g4 d4d5 c6a6 b2b3 e7e5 f1c4 a6f6 c3e4 f8b4 c1d2 f6e7 e1g1 b8d7 h2h3 g4f5 e4g3 f5g6 e3e4 g8f6 c4b5 e8g8 b5d7 f6d7 d2b4 e7b4 d1e2 d7f6 f1e1 f8e8 a1c1 a8c8 c1c4 b4b6 f3h4 c7c6 e1c1 c8c7 e2b2 a7a6 g3f5 g6f5 h4f5 g7g6 f5d6');
INSERT INTO evaluations VALUES(34586,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(34586,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(34586,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(34586,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(34586,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(34586,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(34586,36,'cp 163','d5c6 g1f3 c8g4 d4d5 c6a6 b2b3 e7e5 f1c4 a6f6 c3e4 f6b6 d5d6 g4h5 g2g4 h5g6 d1d5 f8d6 f3e5 g8e7 d5b5 b6b5 c4b5 b8c6 e5g6 h7g6 c1b2 f7f6 e4d6 c7d6 e1c1 d6d5 h2h4 e8c8 b5e2 c8b8 h4h5 g6h5 g4h5 d5d4 h1g1 d4e3 d1d8 h8d8 g1g7 e7f5');
INSERT INTO evaluations VALUES(34516,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(34516,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(34516,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(34516,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(34516,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(34516,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(34516,36,'cp 154','g1f3 c8g4 d4d5 c6a6 b2b3 e7e5 f1c4 a6f6 c3e4 f8b4 c1d2 f6e7 e1g1 b8d7 h2h3 g4f5 e4g3 f5g6 e3e4 b4d2 d1d2 g8f6 f1e1 e8g8 c4b5 d7c5 f3h4 e7d6 d2e3 f8d8 f2f3 a7a6 b5f1 c5d7 h4f5 g6f5 g3f5 d6b6 a1c1 f6e8 e3b6 d7b6 f5e3 d8d7 a2a4 e8d6');
INSERT INTO evaluations VALUES(34517,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(34517,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(34517,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(34517,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(34517,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(34517,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(34517,36,'cp 170','c8g4 d4d5 c6a6 b2b3 e7e5 f1c4 a6f6 c3e4 f8b4 c1d2 f6e7 e1g1 b8d7 h2h3 g4f5 e4g3 f5g6 e3e4 b4d2 d1d2 g8f6 c4d3 e8g8 a1c1 a8c8 d2e3 f8d8 e3a7 f6d5 f1d1 d5b6 d3b1 c7c5 a7b7 c5c4 b3c4 c8b8 b7a6 d8c8 b1c2 b8a8 a6b5 a8a2 c2b3');
INSERT INTO evaluations VALUES(32820,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(32820,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(32820,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(32820,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(32820,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(32820,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(32820,36,'cp 172','d4d5 c6a6 b2b3 e7e5 f1c4 a6f6 c3e4 f8b4 c1d2 f6e7 e1g1 b8d7 h2h3 g4f5 e4g3 f5g6 e3e4 b4d2 d1d2 g8f6 d2e3 h7h5 h3h4 e8g8 a1c1 f8c8 f1d1 f6g4 e3d2 a7a5 f3g5 e7b4 c1c2 b4d2 d1d2 a5a4 b3b4 g4f6 c4b5 d7b6 a2a3 f6d7 b5d7 b6d7');
INSERT INTO evaluations VALUES(1928,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(1928,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(1928,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(1928,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(1928,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(1928,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(1928,36,'cp 115','f1d3 b7b6 d1e2 c8b7 b1c3 g8f6 c1d2 e8c8 e1c1 c8b8 h1e1 f8c5 e4e5 h8e8 f2f4 f6d5 c3e4 d5b4 d3b5 c6e4 e2e4 b7e4 e1e4 c7c6 b5e2 b4d5 e2f3 f7f6 e4e2 f6e5 f3d5 c6d5 f4e5 e8e6 d2e3 d8e8 e3c5 b6c5 d1d5 d7d6 e2d2 e6e5 d5e5 d6e5');
INSERT INTO evaluations VALUES(1927,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(1927,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(1927,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(1927,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(1927,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(1927,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(1927,36,'cp 105','b7b6 d1e2 c8b7 b1c3 g8f6 c1d2 e8c8 e1c1 c8b8 h1e1 f8c5 e4e5 h8e8 f2f4 f6d5 c3e4 d5b4 d3b5 c6e4 e2e4 b7e4 e1e4 c7c6 b5e2 b4d5 e2f3 f7f5 e4e1 d5c7 b2b4 c5e7 g2g3 c7e6 h2h4 b6b5 a2a3 b8c7 d2c3 a7a5 d1d3 g7g6 c1b1 d7d5 e5d6 e7d6 h4h5 a5b4 a3b4');
INSERT INTO evaluations VALUES(33111,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(33111,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(33111,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(33111,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(33111,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(33111,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(33111,36,'cp 87','g1f3 b7b6 f1d3 c8a6 e1g1 b4c3 b2c3 a6d3 d1d3 g8e7 c1a3 d5f5 d3f5 e7f5 d4d5 f5d6 d5e6 d6c4 e6f7 e8f7 a3c1 h8e8 f1d1 b8c6 c1f4 e8e7 f4g5 e7e8 d1d7 f7g8 d7c7 a8c8 c7c8 e8c8');
INSERT INTO evaluations VALUES(33107,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(33107,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(33107,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(33107,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(33107,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(33107,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(33107,36,'cp 77','b7b6 f1e2 b4c3 b2c3 g8f6 e1g1 c8a6 e2a6 b8a6 d1d3 a6b8 c3c4 d5e4 d3b3 e8g8 f1e1 e4b7 c1a3 f8e8 d4d5 b8d7 d5e6 e8e6 e1e6 f7e6 a1e1 a8e8 h2h3 h7h6 f3d4 c7c5 d4b5 b7b8 e1d1 e8d8 a3b2 d7f8 d1e1 d8d7 b2e5 b8d8 b3e3 d7d1 b5a7 d8d2 e3d2');
INSERT INTO evaluations VALUES(46296,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(46296,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(46296,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46296,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(46296,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(46296,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46296,36,'cp 12','h7h6 g5h4 c8d7 e2e3 d7c6 f1d3 b8d7 e1g1 f8d6 f3d2 e8g8 f2f4 d6e7 g2g4 f6e8 h4e7 d8e7 d1e2 e8d6 d2f3 a8c8 e2g2 c8c7 g1h1 b7b5 a2a3 d7f6 g2h3 f6e4 g4g5 d6f5 g5h6 f5h6 d3e4 d5e4 f3e5 c6b7');
INSERT INTO evaluations VALUES(6869,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(6869,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(6869,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(6869,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(6869,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(6869,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(6869,36,'cp 26','e3d4 c5d4 d1d4 d8f6 d4b4 f6b6 b4b6 a7b6 f1d3 d7d6 c2c4 g8f6 b1c3 c8e6 f2f3 e8e7 d3e2 c7c6 e1f2 h8d8 f3f4 h7h5 a2a4 g7g6 a1d1 f6g4 e2g4 h5g4 b2b3 a8a5 d1d2 d8h8 h1d1 h8h2 d2d6 a5h5 d6d4 h2h1');
INSERT INTO evaluations VALUES(24934,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(24934,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(24934,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(24934,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24934,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(24934,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(24934,36,'cp 0','e7e6 h2h3 g4f3 e2f3 f8b4 c1d2 e6e5 e1g1 e5d4 e3d4 c6d4 d2g5 b4c3 b2c3 d4f3 d1f3 e8g8 c3c4 c7c6 a1d1 h7h6 g5f6 d8f6 f3f6 g7f6 c4d5 c6d5 d1d5 a8c8 d5d7 c8c2 a2a3 c2a2 f1b1 a2a3 b1b7 a7a5 g2g3 g8g7 g1g2 h6h5 b7a7 g7g6 a7c7 a3a4 c7a7 a4a3');
INSERT INTO evaluations VALUES(25221,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25221,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25221,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25221,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25221,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25221,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(25221,36,'cp 0','h2h3 g4h5 e1g1 a7a6 b2b3 f8d6 c1b2 e8g8 a1c1 h5g6 f3h4 d8e7 h4g6 h7g6 e2f3 a8d8 g2g3 d6a3 b2a3 e7a3 c3a4 e6e5 c2c3 a6a5 d1e2 b7b6 e2b5 d8d6 b5e2');
INSERT INTO evaluations VALUES(25219,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25219,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25219,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(25219,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(25219,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(25219,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25219,36,'cp 0','g4h5 e1g1 a7a6 c3b1 h8g8 c2c4 g7g5 b1d2 g5g4 f3e5 f8d6 e5c6 g4h3 c6d8 g8g2 g1h1 g2h2 h1g1');
INSERT INTO evaluations VALUES(25190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25190,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25190,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25190,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25190,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25190,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25190,36,'cp 0','e1g1 a7a6 c3b1 h8g8 b1d2 g7g5 c2c4 g5g4 f3e5 f8d6 e5c6 g4h3 c6d8 g8g2 g1h1 g2h2 h1g1');
INSERT INTO evaluations VALUES(31974,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(31974,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(31974,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(31974,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(31974,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(31974,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(31974,36,'cp 145','d1f3 d8c8 f1c4 g8f6 c1g5 c7c6 c4b3 f5g6 g1e2 c8f5 f3f5 g6f5 e2g3 f5d7 f2f4 b8a6 e1g1 a6c7 f4f5 f6d5 c3d5 c6d5 a1e1 f7f6 g5f4 e8c8 f4c7 c8c7 b3d5 d7b5 c2c4 d8d5 c4d5 b5f1 g1f1 c7d7 e1e2 d7d6 e2e6 d6d5 g3e2 d5c4 e2f4');
INSERT INTO evaluations VALUES(55024,30,'cp -212',NULL);
INSERT INTO evaluations VALUES(55024,31,'cp -192',NULL);
INSERT INTO evaluations VALUES(55024,32,'cp -211',NULL);
INSERT INTO evaluations VALUES(55024,33,'cp -195',NULL);
INSERT INTO evaluations VALUES(55024,34,'cp -197',NULL);
INSERT INTO evaluations VALUES(55024,35,'cp -210',NULL);
INSERT INTO evaluations VALUES(55024,36,'cp -196','g8f6 h5e2 f6e4 b1c3 e4f6 d2d4 c8f5 c1g5 b8c6 e1c1 e7e6 g1f3 f8b4 e2e3 b4c3 e3c3 f6e4 g5d8 e4c3 b2c3 c6d8 d1e1 h7h5 f3h4 g7g6 h4f5 g6f5 c3c4 c7c6 h2h4 e8d7 c1d2 b7b6 c4d5 c6d5 c2c4 d5c4 f1c4 d8b7 c4b5 d7e7 g2g3 b7d6');
INSERT INTO evaluations VALUES(47124,30,'cp -195',NULL);
INSERT INTO evaluations VALUES(47124,31,'cp -196',NULL);
INSERT INTO evaluations VALUES(47124,32,'cp -182',NULL);
INSERT INTO evaluations VALUES(47124,33,'cp -206',NULL);
INSERT INTO evaluations VALUES(47124,34,'cp -188',NULL);
INSERT INTO evaluations VALUES(47124,35,'cp -191',NULL);
INSERT INTO evaluations VALUES(47124,36,'cp -191','h5e2 f6e4 b1c3 e4f6 d2d4 c8f5 c1g5 e7e6 h2h4 b8c6 e1c1 h7h6 g5f6 g7f6 g2g4 f5h7 f2f4 f8b4 h1h3 d8d6 e2f2 h8g8 f1e2 e8c8 a2a3 b4a5 f4f5 h6h5 f5e6 f7e6 g4h5 h7f5 h3f3 c8b8 c3b5 d6e7 e2d3 f5d3 f3d3 a5b6 g1f3');
INSERT INTO evaluations VALUES(19293,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(19293,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(19293,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(19293,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(19293,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(19293,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(19293,36,'cp -27','f8g7 d2d3 d7d6 g2g3 e7e6 c1e3 g8e7 d3d4 c5d4 f3d4 e8g8 d1d2 c6d4 e3d4 e6e5 d4f2 d6d5 e4d5 e5f4 f1g2 e7f5 e1g1 f4g3 h2g3 f5d6 d2f4 b7b6 a1d1 h7h5 c3e4 d6e4 g2e4 c8g4 d1d2 g7b2 d5d6 b2c3 e4a8 c3d2 f4d2 d8a8 f2d4 f8d8 d4f6 d8d7 f6c3');
INSERT INTO evaluations VALUES(51983,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51983,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(51983,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(51983,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(51983,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(51983,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(51983,36,'cp -12','g8f6 g2g3 b8c6 f1g2 f8e7 g1e2 e8g8 e1g1 d7d6 d3d4 c5d4 e2d4 h7h6 f1e1 c6d4 d1d4 c8d7 a2a4 a7a6 e1d1 d7c6 a4a5 a8c8 d4d3 d8c7 c1e3 f6d7 d3e2 f8d8 e2d2 c7b8 e3d4 e7f6 d4f6 d7f6 d2d4 b8c7 d4e3 c8b8 d1d2');
INSERT INTO evaluations VALUES(21274,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(21274,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21274,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(21274,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(21274,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(21274,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(21274,36,'cp 40','g7g5 c3d5 f8g7 c2c3 g8f6 d2d4 e8g8 d4e5 f6d5 c4d5 d6e5 a2a4 d8f6 c1e3 a7a5 f3e1 f8d8 d1b3 c6e7 d5c4 e7g6 e1d3 g6f4 a1d1 g7f8 f2f3 b7b6 d3f4 d8d1 f1d1 g5f4 e3f2 g8g7 c4d5 a8b8 b3b5 f8d6 b2b4 a5b4');
INSERT INTO evaluations VALUES(32290,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32290,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32290,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32290,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32290,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(32290,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32290,36,'cp 0','f8d6 f4d6 d8d6 g1f3 g8f6 f1e2 c7c5 c3c4 d5c4 b1d2 c5d4 f3d4 b8c6 d4f5 e6f5 d2c4 d6b4 d1d2 e8e7 e1g1 b4d2 c4d2 h8d8 f1d1 g7g6 a1c1 a8c8 d2b3 f6e4 f2f3 d8d1 e2d1 e4d6 d1e2 c6e5 c1c8 d6c8 g1f2 c8d6 b3d4 e7f6');
INSERT INTO evaluations VALUES(28143,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28143,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28143,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28143,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28143,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28143,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(28143,36,'cp 0','f4d6 d8d6 c3c4 g8f6 c4d5 e6d5 f1d3 f5d3 d1d3 c7c6 g1f3 e8g8 e1g1 d6e7 d3a3 e7a3 b1a3 f6e4 a3b1 e4d6 b1c3 a7a5 c3a4 f8e8 f1c1 f7f6 a4b6 a8a6 b6a4');
INSERT INTO evaluations VALUES(40395,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(40395,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40395,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(40395,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40395,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(40395,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(40395,36,'cp 29','f6g4 e3g5 h7h6 g5h4 b8c6 d4b3 c8e6 f2f3 g4e5 c3d5 a8c8 h4f2 e8g8 f1e2 e5d7 c2c3 e6d5 e4d5 c6a5 e1g1 a5c4 e2c4 c8c4 f1e1 f8e8 a2a4 c4c8 f2a7 b7b6 a4a5 d8c7 a5b6 d7b6 g1h1 b6c4 d2f2 c7b7 b3a5 c4a5 a1a5 c8a8');
INSERT INTO evaluations VALUES(40412,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(40412,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(40412,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(40412,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(40412,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40412,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(40412,36,'cp 29','e3g5 b8c6 d4b3 h7h6 g5h4 c8e6 f2f3 g4f6 e1c1 a8c8 c1b1 a7a6 f3f4 e8g8 f1e2 c6b4 e2f3 f6d7 h4f2 d7b6 f2d4 b6c4 d2e1 g7d4 b3d4 d8b6 c3a4 b6a5 a2a3 a5a4 e1b4 a4b4 a3b4 g6g5 f4g5 h6g5 d4e6 f7e6 f3g4 g8f7 h2h4 g5h4 h1h4');
INSERT INTO evaluations VALUES(40411,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40411,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(40411,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(40411,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40411,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(40411,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(40411,36,'cp 21','h7h6 g5h4 b8c6 d4b3 c8e6 f2f3 g4f6 e1c1 a8c8 c1b1 e8g8 f1e2 c6e5 h1e1 e5c4 e2c4 c8c4 c3e2 b7b5 e2d4 d8d7 h4g3 f8c8 c2c3 c8c7 e4e5 f6d5 d4e6 f7e6 e5d6 e7d6 d2d3 b5b4 d3g6 b4c3 g6e6 d7e6');
INSERT INTO evaluations VALUES(15838,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(15838,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(15838,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(15838,33,'cp -138',NULL);
INSERT INTO evaluations VALUES(15838,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(15838,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(15838,36,'cp -143','a5c4 e2c4 d8d5 c4d5 f6d5 b1c3 d5b4 e1d1 f7f6 g5e4 c8e6 c3b5 e8d7 d2d3 f8e7 b5c3 b4c6 a2a3 b7b6 b2b4 a7a6 h2h4 h8d8 c1e3 h7h6 e4g3 f6f5 e3d2 b6b5 c3e2 a6a5 b4a5 c6a5 a1b1 a5c6 d2c3 e7a3 b1b5 a3d6 f2f3 a8a2 h4h5 d7e8');
INSERT INTO evaluations VALUES(15807,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(15807,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(15807,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(15807,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(15807,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(15807,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(15807,36,'cp -133','e2c4 d8d5 c4d5 f6d5 d2d3 d5b4 e1d1 h7h6 g5e4 f7f5 e4g3 b4c6 h1e1 e8f7 c1d2 f8c5 f2f4 e5f4 d2f4 c5b6 f4d2 g7g5 a2a4 c8d7 a4a5 b6d4 d2c3 h8e8 b1d2 a8d8 d2f3 d4c5 f3e5 c6e5 c3e5 c5f2 e1f1 e8e5 f1f2');
INSERT INTO evaluations VALUES(15804,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(15804,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(15804,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(15804,33,'cp -151',NULL);
INSERT INTO evaluations VALUES(15804,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(15804,35,'cp -149',NULL);
INSERT INTO evaluations VALUES(15804,36,'cp -148','d8d5 c4d5 f6d5 d2d3 f7f6 g5e4 d5b4 e1d1 c8e6 a2a3 b4c6 h1e1 b7b6 b2b4 e8f7 b1d2 f8e7 c2c3 f6f5 e4g3 e7f6 d2f3 h7h6 d1c2 a7a6 c1d2 g7g5 h2h3 a8d8 g3h5 f5f4 h5f6 f7f6 f3h2 h6h5 h2f3 g5g4 h3g4');
INSERT INTO evaluations VALUES(1022,30,'cp -145',NULL);
INSERT INTO evaluations VALUES(1022,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(1022,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(1022,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(1022,34,'cp -143',NULL);
INSERT INTO evaluations VALUES(1022,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(1022,36,'cp -136','c4d5 f6d5 d2d3 f7f6 g5e4 d5b4 e1d1 c8e6 a2a3 b4c6 b2b4 f8e7 h1e1 a7a6 b1d2 a8d8 f2f3 e8f7 c1b2 h7h5 e4f2 g7g5 h2h3 h8g8 c2c4 d8d7 d1c2 b7b5 g2g4 b5c4 d3c4 g8d8 b2c3');
INSERT INTO evaluations VALUES(1021,30,'cp -154',NULL);
INSERT INTO evaluations VALUES(1021,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(1021,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(1021,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(1021,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(1021,35,'cp -138',NULL);
INSERT INTO evaluations VALUES(1021,36,'cp -158','f6d5 d2d3 d5b4 e1d1 b4c6 a2a3 h7h6 g5e4 f7f5 e4g3 f8c5 c1e3 c5d6 h1e1 c8e6 e3d2 g7g5 b2b4 e8f7 d2c3 h6h5 b1d2 h5h4 g3f1 a8d8 d2f3 f7f6 b4b5 g5g4 b5c6 g4f3 g2f3 e6d5 c6b7 d5f3 d1c1 f3b7 f1d2');
INSERT INTO evaluations VALUES(1025,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(1025,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(1025,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(1025,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(1025,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(1025,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(1025,36,'cp -150','d2d3 d5b4 e1d1 h7h6 g5e4 f7f5 e4g3 b4c6 h1e1 c8e6 f2f4 e8c8 f4e5 g7g6 b1d2 f8g7 d2f3 e6d5 c1d2 h8e8 d2c3 d5f3 g2f3 g7e5 e1f1 e5d6 a2a3 c6e7 c3f6 e8f8 f6g7 f8g8 g7c3 e7d5 c3d2 h6h5 f3f4 g8e8');
INSERT INTO evaluations VALUES(51093,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(51093,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(51093,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(51093,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(51093,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(51093,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(51093,36,'cp 98','e7e6 d2d4 d7d5 e4e5 c7c5 c2c3 g8e7 b1a3 e7f5 g2g4 f5h6 h2h3 c8d7 f1d3 h6g8 e1g1 c5c4 d3b1 b7b5 a3c2 b5b4 c1g5 f8e7 g5f4 b8c6 d1d2 f7f6 f4g3 f6e5 d4e5 d8b6 c2d4 c6d4 f3d4');
INSERT INTO evaluations VALUES(51017,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(51017,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(51017,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(51017,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(51017,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(51017,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(51017,36,'cp 82','d2d4 d7d5 e4e5 c7c5 c2c3 c8d7 f1d3 g8e7 e1g1 b8c6 b1a3 e7g6 g2g3 c5d4 c3d4 c6b4 d3e2 d8b6 a3b1 d7b5 b1c3 b5e2 d1e2 b4c6 c1e3 a5a4 h2h4 f8b4 h4h5 g6e7 h5h6 g7h6 c3b5 e7f5 e3f4 e8g8 g1g2 a8a5 b5d6 b4d6 e5d6 b6b5 e2b5');
INSERT INTO evaluations VALUES(51015,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(51015,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(51015,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(51015,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(51015,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(51015,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(51015,36,'cp 94','d7d5 e4e5 c7c5 c2c3 g8e7 b1a3 e7f5 g2g4 f5h6 h1g1 c8d7 c1g5 d8b6 d1b3 b6c6 b3c2 c5c4 h2h4 b7b5 c2d2 h6g8 a3c2 b5b4 h4h5 c6b6 f1e2 h7h6 g5f4 b8c6 e1f1 g7g5 h5g6 f7g6 g1h1 e8c8 f1g2 f8e7');
INSERT INTO evaluations VALUES(53855,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(53855,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(53855,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(53855,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(53855,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(53855,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(53855,36,'cp 137','b1c3 g8f6 g1f3 f8e7 c1f4 d8a5 f1d3 b8d7 e1g1 e6e5 f4e3 d7f8 h2h3 f8g6 a2a3 e8g8 b2b4 a5d8 f1e1 f6h5 d3f1 h5f4 a1c1 f7f5 d4e5 d6e5 e3f4 g6f4 d1d8 f8d8 f3e5 e7f6 e5f3 f4g6 c4c5');
INSERT INTO evaluations VALUES(53853,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(53853,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(53853,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(53853,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(53853,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(53853,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(53853,36,'cp 133','b8d7 g1f3 g8f6 h2h3 d8c7 f1d3 f8e7 e1g1 e8g8 c1e3 a7a6 a1c1 c6c5 d4d5 d7e5 f3e5 d6e5 d1f3 c8d7 f1d1 g8h8 b2b3 h7h6 f3e2 e7d6 e2d2 a8d8 a2a3 b7b6 d3c2 f6g8 d5e6 d7e6 c3d5 e6d5 c4d5 c7e7 c2d3 b6b5');
INSERT INTO evaluations VALUES(18542,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(18542,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(18542,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(18542,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(18542,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(18542,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(18542,36,'cp 138','g1f3 d8c7 f1d3 f8e7 h2h3 g8f6 e1g1 e8g8 c1e3 a7a6 a1c1 c6c5 d4d5 d7e5 f3e5 d6e5 d1f3 c8d7 b2b3 g8h8 f1d1 a8b8 d1d2 b7b5 c4b5 a6b5 a2a4 e6d5 e4d5 b5a4 b3a4 b8b4 c3b5 e5e4 b5c7 e4f3 e3c5 e7c5 c1c5 d7a4 g2f3');
INSERT INTO evaluations VALUES(18541,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(18541,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(18541,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(18541,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(18541,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(18541,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(18541,36,'cp 126','g8f6 h2h3 d8c7 f1d3 f8e7 c1e3 e8g8 e1g1 a7a6 a1c1 c6c5 d4d5 d7e5 f3e5 d6e5 d3b1 e7d6 e3g5 f6e8 b2b3 h7h6 g5e3 c8d7 d1d3 e8f6 f1d1 f8d8 d5e6 d7e6 d1d2 d8d7 c1d1 d6f8 c3d5 e6d5 c4d5 b7b5');
INSERT INTO evaluations VALUES(5403,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(5403,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(5403,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(5403,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(5403,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(5403,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(5403,36,'cp -17','f1c4 g8f6 d1e2 e8g8 e1g1 d7d5 e4d5 c6d4 c3d4 f6d5 b1c3 c8e6 f1d1 d5c3 b2c3 e6c4 e2c4 d8d7 c4b3 f8c8 c1e3 b7b5 a2a4 b5a4 a1a4 c8c3 a4a7 a8a7 b3c3 h7h5 d1c1 g7f6 c3c5 a7b7 d4d5 h5h4 c5c8 g8g7 c8c6 g6g5 g2g3 h4g3 h2g3');
INSERT INTO evaluations VALUES(5384,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(5384,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(5384,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(5384,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(5384,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(5384,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(5384,36,'cp -14','g8f6 d1e2 e8g8 e1g1 c6d4 c3d4 d7d5 e4d5 f6d5 b1c3 c8e6 f1d1 d5c3 b2c3 e6c4 e2c4 d8a5 c4b3 a8d8 b3b7 a5c3 c1e3 g7d4 b7e7 d4e3 e7e3 d8d1 a1d1 c3e3 f2e3 f8c8 a2a4 c8c4 a4a5 c4c5 a5a6 c5c6 d1a1 c6c2 h2h3 h7h6 g1f1 g8g7 a1a5 g7f6 a5b5');
INSERT INTO evaluations VALUES(8037,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(8037,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(8037,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8037,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(8037,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(8037,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(8037,36,'cp -13','d1e2 e8g8 e1g1 d7d5 e4d5 c6d4 c3d4 f6d5 f1d1 c8e6 b1c3 d5c3 b2c3 e6c4 e2c4 d8d7 c1e3 f8c8 c4b3 b7b5 a2a4 b5a4 a1a4 c8c3 a4a7 a8a7 b3c3 h7h5 d1c1 g7f6 g2g3 h5h4 c3c8 g8g7 d4d5 d7c8 c1c8 a7a3 g1g2 h4g3 h2g3 a3d3 c8a8 d3d1 g2f3 d1d5 a8a7');
INSERT INTO evaluations VALUES(33703,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(33703,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(33703,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(33703,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(33703,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(33703,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(33703,36,'cp 130','d2d4 e5d6 g1f3 a7a6 e1g1 c8f5 f3h4 f5c8 f1e1 e7e6 h4f3 f8e7 f3e5 b8c6 e5c6 d6c6 e2f3 c6d6 c3e4 f6e4 f3e4 e8g8 c1e3 e7f6 c2c4 g7g6 c4c5 d6d8 d1d3 c7c6 a1d1 d8c7 d3c2 f8d8 g2g3 a6a5 a2a3 c7e7 e4g2');
INSERT INTO evaluations VALUES(39121,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(39121,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(39121,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(39121,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(39121,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(39121,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(39121,36,'cp 126','d2d4 d6e7 c1f4 g8f6 b1c3 e8g8 f1d3 b8c6 e4e5 f6h5 f4e3 d7d6 d1e2 c6b4 d3c4 d6d5 a2a3 b4c2 e2c2 d5c4 d4d5 f7f5 c2e2 f5f4 e3f2 b7b5 e1c1 a7a6 e5e6 h5f6 h1e1 c8b7 f2h4 f6g4 h2h3 g4e3');
INSERT INTO evaluations VALUES(39118,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(39118,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(39118,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(39118,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(39118,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(39118,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(39118,36,'cp 130','c7c6 e4e5 d6c7 c1f4 d7d6 f1c4 d6e5 f3e5 g8h6 f4h6 c7e5 d1h5 d8d4 h5f7 e8d8 h6g5 e5f6 g5f6 d4f6 f7f6 g7f6 e1g1 b8d7 c4d3 d8e7 b1d2 d7e5 a1e1 c8e6 d2e4 a8d8 f1f6 e5d3 c2d3 e6a2 f6h6 e7d7 e4c5 d7c8 e1e7 b7b6 h6c6 c8b8 c5d7 b8a8');
INSERT INTO evaluations VALUES(47169,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(47169,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(47169,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(47169,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(47169,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(47169,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47169,36,'cp -9','g1f3 c7c5 d4c5 e7e6 c1b2 f8c5 a2a3 a7a6 c2c4 d5c4 d1d8 e8d8 f3e5 d8e7 b3b4 c5a7 e5c4 b7b5 c4a5 c8d7 e2e3 b8c6 a5b3 a7b8 b3c5 h8c8 f1e2 b8e5 b2e5 c6e5 e1g1 f6d5 f1c1 a6a5 b1d2 a5b4 a3b4 a8a1 c1a1 d5b4 c5d7 e5d7');
INSERT INTO evaluations VALUES(47168,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(47168,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(47168,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(47168,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(47168,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(47168,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(47168,36,'cp -20','c7c5 d4c5 e7e6 a2a3 f8c5 c1b2 e8g8 e2e3 a7a6 b1d2 b7b5 f1e2 b8d7 c2c4 b5c4 b3c4 a8b8 d1c2 c8b7 c4d5 f6d5 e1g1 d8e7 f3e5 d7e5 b2e5 c5d6 e5d6 e7d6 d2c4 d6c5 f1c1 f8c8 c2b2 b7c6 c4e5 b8b2 c1c5 b2e2 c5c6 c8c6 e5c6');
INSERT INTO evaluations VALUES(48790,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(48790,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(48790,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(48790,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48790,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(48790,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(48790,36,'cp 21','c4d5 f6d5 g1f3 d5c3 b2c3 f8d6 d2d4 d8e7 f3e5 d6e5 d4e5 e7e5 f1e2 e5c3 c1d2 c3f6 e1g1 b8c6 d1b3 e8g8 f1d1 a8b8 d2c3 f6g6 b3a4 c8e6 e2f3 h7h6 a4f4 g6g5 f4c7 f8c8 c7g3 g5g3 h2g3 c8c7 c3b2 b8c8 f3e4 c8d8 d1d8 c6d8');
INSERT INTO evaluations VALUES(48599,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(48599,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48599,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(48599,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48599,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(48599,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(48599,36,'cp 19','f6d5 g1f3 d5c3 b2c3 f8d6 d2d4 d8e7 d4e5 d6e5 f3e5 e7e5 f1e2 e5c3 c1d2 c3a3 d1c2 e8g8 e1g1 b8c6 d2c3 f8d8 e2f3 c8e6 f1b1 a8b8 b1b5 h7h6 a1b1 a3a2 c2a2 e6a2 b1b2 a2e6 b5b7 b8b7 b2b7 c6e7 b7a7 f7f6 a7a8 d8a8 f3a8 c7c6 e3e4 e7g6 a8b7');
INSERT INTO evaluations VALUES(49014,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(49014,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(49014,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(49014,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(49014,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(49014,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(49014,36,'cp 15','g1f3 d5c3 b2c3 f8d6 d2d4 d8e7 d4e5 d6e5 f3e5 e7e5 f1e2 e5c3 c1d2 c3a3 e1g1 e8g8 d1c2 b8c6 e2f3 a8b8 f3c6 b7c6 c2c6 a3d6 c6d6 c7d6 f1d1 d6d5 a2a4 c8e6 d2c3 f7f6 d1c1 f8c8 a4a5 c8c4 a5a6 b8b6 h2h3 h7h5 f2f3');
INSERT INTO evaluations VALUES(50212,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(50212,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50212,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(50212,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50212,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(50212,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50212,36,'cp 21','c1f4 d7d5 e2e3 f8g7 b1d2 e8g8 f1e2 c7c5 g1f3 f6h5 f4e5 h5f6 e1g1 b7b6 e5g3 b8c6 h2h3 c8b7 f1e1 e7e6 e2b5 a7a6 b5a4 d8e7 a4c2 a8e8 f3e5 e8c8 d1e2 f8e8 f2f4 e7c7 e5g4 f6h5 g3h2');
INSERT INTO evaluations VALUES(57685,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(57685,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(57685,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(57685,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(57685,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(57685,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(57685,36,'cp 51','d2d4 d7d5 e2e4 d5e4 c3e4 f8g7 g1f3 g8f6 e4f6 e7f6 f1c4 e8g8 e1g1 f8e8 h2h3 b8d7 f1e1 d7b6 e1e8 d8e8 c4f1 c8f5 c2c4 f5e4 c1e3 f6f5 a2a4 a8d8 e3g5 d8d7 a4a5 b6c8 g5f4 e4f3 d1f3 g7d4 f3b7 c7c5');
INSERT INTO evaluations VALUES(23214,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(23214,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(23214,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(23214,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(23214,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(23214,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(23214,36,'cp 149','g8e7 d2d4 d7d5 e4d5 e7d5 e1g1 d5b6 c4d3 c8g4 h2h3 g4h5 f1e1 d8d6 a2a4 e8c8 a4a5 b6d7 a5a6 c8b8 a6b7 h5f3 d1f3 e5d4 d3e4 c6e5 f3g3 d7c5 c3d4 d6d4 b1c3 a7a6 c1f4 f8d6 a1d1 e5d3 f4d6 d4d6 g3g7 h8e8 g7h7 d6f4 e4d3 e8e1 d1e1 c5d3');
INSERT INTO evaluations VALUES(17137,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(17137,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(17137,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(17137,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(17137,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(17137,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(17137,36,'cp 132','d2d4 d7d5 e4d5 e7d5 e1g1 d5b6 c4d3 f8d6 d4e5 d6e5 d1c2 d8e7 d3h7 c8e6 h7g6 e6f7 g6f7 e7f7 f3e5 c6e5 f2f4 e5c6 f1e1 c6e7 b1d2 e8c8 d2f3 b6c4 b2b3 c4d6 c1a3 e7f5 a3d6 f5d6 h2h3 h8e8 g1h2 f7h5 c2f2 c8b8 a1d1 a7a6 c3c4 e8e1 d1e1 d8e8 e1e8 h5e8');
INSERT INTO evaluations VALUES(17143,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(17143,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(17143,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(17143,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(17143,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(17143,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(17143,36,'cp 131','d7d5 e4d5 e7d5 e1g1 d5b6 c4d3 f8d6 d4e5 d6e5 d1c2 c8e6 d3h7 d8e7 h7g6 e6f7 g6f7 e7f7 f3e5 c6e5 f2f4 e5c6 f1e1 c6e7 b1d2 e8c8 d2f3 b6c4 b2b3 c4d6 c1a3 e7f5 a1d1 h8e8 c2f2 b7b6 g2g4 d6e4 d1d8 c8d8 f2c2 f5d6 c3c4 f7d7 h2h3 a7a5 a3d6 e4d6 e1e8 d7e8');
INSERT INTO evaluations VALUES(52779,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52779,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(52779,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52779,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52779,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(52779,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(52779,36,'cp 12','g7g6 f1g2 g8f6 c2c4 f8g7 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 e7e6 e1g1 h7h6 h2h4 f8d8 g2f3 d8c8 a2a3 c6e5 f4e5 c8c6 g1g2 g6g5 e5f6 g7f6 h4g5 h6g5 f3e2 c6c7 g3g4 f5e4 g2g3 e4g6 f1c1 a8c8');
INSERT INTO evaluations VALUES(55725,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(55725,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(55725,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(55725,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(55725,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(55725,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(55725,36,'cp 43','g8f6 c2c4 c7c5 d4c5 d6d5 d1c2 f8c5 e2e3 e8g8 c4d5 c5b4 b1d2 e6d5 f1d3 b8c6 e1g1 c8e6 a2a3 b4d6 f4d6 d8d6 a1c1 f8c8 c2b1 h7h6 d2b3 c8c7 h2h3 a8c8 f1d1 c6e5 f3e5 d6e5 c1c7 e5c7');
INSERT INTO evaluations VALUES(48051,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(48051,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(48051,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(48051,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(48051,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(48051,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(48051,36,'cp 35','c2c4 c7c5 d4c5 d6d5 e2e3 f8c5 d1c2 e8g8 c4d5 c5b4 b1d2 e6d5 f1d3 b8c6 e1g1 c8e6 a2a3 b4d6 f4d6 d8d6 a1c1 a8c8 c2a4 e6d7 h2h3 c8c7 a4h4 f8c8 d2b3 h7h6 f1d1 a7a6 f3d4 c6e7 c1c7 c8c7 d3b1');
INSERT INTO evaluations VALUES(57501,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(57501,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(57501,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(57501,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(57501,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(57501,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(57501,36,'cp 38','d2d4 d7d6 d4e5 d6e5 d1d8 e8d8 b1c3 b8c6 g1f3 f8b4 c1d2 g8f6 e1c1 f6g4 c3d5 b4d2 d1d2 g4f2 f3g5 f2h1 g5f7 d8e8 f7h8 a8b8 d5c7 e8e7 e2e4 c8e6 c7e6 e7e6 e4f5 e6e7 f1g2 h1g3 h2g3 b8h8 d2d5 b7b6 c1d2 h8d8 d5d8 c6d8 b2b4 e7d6 d2d3 d8f7 g3g4 h7h6 g2e4 d6c7 d3e3 a7a6 e4d5 f7d6');
INSERT INTO evaluations VALUES(43206,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(43206,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(43206,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(43206,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(43206,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(43206,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(43206,36,'cp -12','c7c5 g1f3 b8c6 b1c3 d8b6 e1g1 c5d4 e3d4 c6d4 f3d4 b6d4 d1f3 e8g8 a1e1 e7e6 c3b5 d4d5 b5c7 d5f3 g2f3 f6d5 c7d5 e6d5 c2c3 d7d6 f4d6 c8h3 d6f8 g8f8 e1d1 h3f1 d3f1 d5d4 c3d4 a8d8 b2b3 g7d4 g1g2 f7f6 f3f4 d4b6 d1d8 b6d8 f1c4 f8g7 c4d5 b7b6 f4f5 g6f5 d5e6 g7g6 g2g3 h7h5 h2h3');
INSERT INTO evaluations VALUES(19968,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19968,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19968,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19968,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(19968,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19968,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19968,36,'cp 0','b1c3 g7g6 g2g3 a7a6 f1g2 f8g7 a2a4 e7e5 g1e2 g8e7 c1e3 e8g8 e1g1 c8d7 d1d2 c6b4 c3a2 b4c6');
INSERT INTO evaluations VALUES(19966,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19966,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(19966,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(19966,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19966,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(19966,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(19966,36,'cp -20','g7g6 g2g3 f8g7 f1g2 a8b8 g1e2 a7a6 a2a4 c8g4 e1g1 g4e2 c3e2 g8f6 d3d4 c5d4 e2d4 f6d7 c1e3 e8g8 b2b3 d8a5 a1c1 c6d4 e3d4 g7d4 d1d4 a5b4 d4e3 b8c8 h2h4 a6a5 g2h3 e7e6 f1d1 f8d8 e3d2 b4d2 d1d2');
INSERT INTO evaluations VALUES(47312,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(47312,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(47312,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(47312,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(47312,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(47312,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(47312,36,'cp -64','c7c6 e2e3 b7b5 a2a4 d8b6 f1e2 a7a5 g1f3 b8d7 a4b5 c6b5 b2b3 e7e6 e1g1 f8b4 c3a2 c8a6 b3c4 b5c4 f3e5 e8g8 e5c4 b6b5 a2b4 a5b4 g5f6 d7f6 e2d3 f8c8 c4e5 b5b7 a1a6 a8a6 d3a6 b7a6 e5d3 f6d5 d1b3 c8b8 d3c5 a6b5');
INSERT INTO evaluations VALUES(51710,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(51710,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(51710,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(51710,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(51710,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(51710,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(51710,36,'cp 37','a4b3 f8e7 a2a4 g8f6 a4b5 c6b5 d2d4 b8d7 b1c3 c8b7 e1g1 e8g8 c3b5 f6e4 c1e3 a7a6 b5a3 e5d4 f3d4 d7c5 d4f5 c5b3 c2b3 d6d5 f5e7 d8e7 a3c2 e7f6 f2f3 e4d6 c2d4 f8e8 d1d2 d6f5 d4f5 f6f5 f1c1 a8c8 e3d4 f7f6 h2h3 c8c1 a1c1');
INSERT INTO evaluations VALUES(15526,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(15526,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(15526,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(15526,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(15526,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15526,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(15526,36,'cp 34','d2d4 c8g4 d4d5 c6d4 f1e2 d4e2 d1e2 f8e7 c1e3 e8g8 h2h3 g4d7 a2a4 h7h6 a4a5 a7a6 a1a3 f6h7 a3b3 a8b8 f3d2 f7f5 e4f5 d7f5 d2e4 e7h4 e1g1 h7f6 f2f3 d8e8 e4f6 h4f6 c3e4 e8a4 b3a3 a4e8 g1h2 e8g6 e3f2 f6g5 a3b3 f5e4 e2e4');
INSERT INTO evaluations VALUES(15542,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(15542,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(15542,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(15542,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(15542,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(15542,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(15542,36,'cp 43','c8g4 d4d5 c6d4 f1e2 d4e2 d1e2 g4d7 e1g1 f8e7 f3e1 e8g8 h2h3 c7c6 d5c6 b7c6 e1c2 d8c7 f1d1 d7e6 c2e3 a8b8 a1b1 a7a5 b2b3 g7g6 c1a3 f6h5 b1c1 h5f4 e2d2 b8d8 c3d5 e6d5 c4d5 c6c5 e3c4');
INSERT INTO evaluations VALUES(17166,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(17166,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(17166,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(17166,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(17166,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(17166,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(17166,36,'cp -23','e7c6 d2d4 e5d4 f3d4 f8c5 d4c6 b7c6 d1f3 c5d4 e1g1 d7d6 c2c3 d4f6 c1f4 e8g8 f1e1 c8e6 b1a3 f8e8 a3c2 d8b8 b2b3 a7a5 e4e5 d6e5 f4e5 e6b3 a2b3 f6e5 c2d4 e5d4 e1e8 b8e8 c3d4 e8e6 f3d3 e6d5 d3c3 d5d7 a1a5');
INSERT INTO evaluations VALUES(16851,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(16851,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(16851,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(16851,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(16851,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(16851,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(16851,36,'cp -27','b1c3 f8c5 f3e5 d8g5 e5g4 d7d6 h2h3 g5g6 d2d3 f7f5 e4f5 c8f5 d1e2 e8f7 e1g1 c6d4 e2d1 h7h5 g4e3 f5h3 c3e4 f7g8 c2c3 d4e6 d3d4 c5b6 e4g3 h5h4 g2h3 h4g3 d1g4 g3f2 f1f2 g6g4 h3g4 e6g5 g1g2 a8e8 a2a4 c7c6 a4a5 b6c7 e3f5 e8e4 f5e7 g8h7 c1g5 e4g4 g2f1 g4g5');
INSERT INTO evaluations VALUES(16848,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(16848,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(16848,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(16848,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(16848,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(16848,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(16848,36,'cp -21','f8c5 f3e5 d8g5 e5g4 d7d6 h2h3 g5g6 d2d3 e8g8 c1e3 f7f5 e4f5 c8f5 c3d5 c5e3 g4e3 a8e8 d1d2 g6g2 e1c1 g2f3 d5c7 e8e5 d3d4 e5e4 c7b5 d6d5 e3f5 f8f5 h1e1 f3g2 e1e2 h7h6 e2e4 d5e4 d4d5 f5f2 d2c3 g2g5 c1b1 c6e5 d5d6 e4e3 c3c8 g8h7 d6d7 e5c6 d7d8q c6d8 d1d8 g5b5');
INSERT INTO evaluations VALUES(10642,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(10642,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(10642,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(10642,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(10642,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(10642,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(10642,36,'cp -18','f3e5 d8g5 e5g4 d7d6 h2h3 g5g6 d2d3 e8g8 c1e3 f7f5 e4f5 c8f5 c3d5 c5e3 g4e3 a8e8 d1d2 g6g2 e1c1 g2f3 d5c7 e8e7 c7b5 d6d5 e3f5 f3f5 b5c3 f5f2 d2f2 f8f2 h1f1 f2f1 d1f1 e7e3 f1h1 e3e5 c3b5 e5g5 a2a3 h7h5 a3a4 g7g6 d3d4 g8g7 h1e1 g5g3');
INSERT INTO evaluations VALUES(17841,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(17841,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(17841,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(17841,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(17841,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(17841,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(17841,36,'cp 16','d2d4 e5d4 f3d4 c6c5 d4e2 d8d1 f1d1 c8d7 c1e3 e8c8 b1c3 g7g5 d1d2 f8d6 c3d5 d7c6 e2c3 b7b5 a2a3 a6a5 a1d1 b5b4 a3b4 a5b4 c3e2 c6d5 d2d5 g8e7 d5d2 h7h5 c2c3 e7c6 h2h4 g5h4 c3b4 c6b4 e2f4 h8e8 f4h5 e8e4 h5f6');
INSERT INTO evaluations VALUES(17832,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(17832,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(17832,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(17832,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(17832,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(17832,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(17832,36,'cp 16','e5d4 f3d4 c6c5 d4e2 c8e6 b1c3 d8d7 d1d7 e8d7 f1d1 d7c6 c3d5 f8d6 c1f4 e6d5 e4d5 c6d7 f4d6 c7d6 e2f4 b7b5 b2b3 g8e7 f4e6 g7g6 c2c4 h7h5 a1b1 a8b8 g2g3 g6g5 g1g2 h5h4 d1e1 b5c4 b3c4 e7f5 g3g4 f5d4 e6d4 c5d4 b1d1 b8b2 d1d4 b2a2');
INSERT INTO evaluations VALUES(17843,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(17843,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(17843,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(17843,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(17843,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(17843,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(17843,36,'cp 26','f3d4 c6c5 d4e2 c8e6 b1c3 d8d7 d1d7 e8d7 f1d1 d7c6 c3d5 f8d6 c1f4 e6d5 d1d5 g8e7 d5d2 h8d8 a1d1 b7b5 f2f3 a6a5 a2a4 b5a4 e2c3 a4a3 b2a3 e7c8 f4d6 d8d6 c3d5 c8b6 d5e7 c6d7 e7f5 d6d2 d1d2 d7c6 f5e3 a5a4 g1f2 g7g6 g2g3');
INSERT INTO evaluations VALUES(17842,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(17842,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(17842,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17842,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(17842,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(17842,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(17842,36,'cp 19','c6c5 d4e2 c8e6 b1c3 d8d7 d1d7 e8d7 f1d1 d7c6 c3d5 f8d6 c1f4 e6d5 d1d5 g8e7 d5d2 a8d8 a1d1 b7b5 h2h4 e7g6 f4g3 h8e8 f2f3 f6f5 e4f5 g6e7 g3d6 d8d6 e2g3 e8f8 f5f6 f8f6 g3e4 d6d2 d1d2 f6e6 g1f2 e7g6 h4h5');
INSERT INTO evaluations VALUES(17845,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(17845,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(17845,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(17845,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(17845,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(17845,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(17845,36,'cp 15','d4e2 c8e6 b1c3 d8d7 d1d7 e8d7 f1d1 d7c6 a2a4 b7b6 e2f4 e6f7 f4d5 g8e7 c1f4 e7d5 c3d5 f7d5 d1d5 f8e7 c2c4 h8d8 h2h4 e7d6 a1d1 a8b8 e4e5 f6e5 f4e5 d8d7 e5c3 b8e8 f2f3 b6b5 c4b5 a6b5 a4b5 c6b5 g1f2 b5c6 h4h5 d7e7');
INSERT INTO evaluations VALUES(20931,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(20931,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(20931,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(20931,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(20931,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(20931,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(20931,36,'cp 64','e2e3 e7e6 c2c4 g8f6 b1c3 f8d6 f4d6 d8d6 h2h3 e8g8 a1c1 c8d7 c4d5 e6d5 f1d3 c6e7 e1g1 c7c6 d1c2 a8c8 c3e2 b7b6 b2b3 c6c5 c2b2 d7f5 d3a6 c8c7 e2c3 f5d7 f1d1 e7f5 c3b5 d7b5 a6b5 c5d4 b5d3 c7c1 d1c1');
INSERT INTO evaluations VALUES(20929,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(20929,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(20929,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(20929,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(20929,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(20929,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(20929,36,'cp 50','e7e6 c2c4 g8f6 b1c3 f8d6 f4d6 d8d6 h2h3 e8g8 a1c1 c8d7 c4d5 e6d5 f1d3 c6e7 e1g1 c7c6 d1c2 a8c8 f1d1 f8d8 c3e2 b7b6 b2b3 c6c5 c2b2 d7e6 d3a6 c8c7 e2c3 e6d7 a2a3 d7e8 c3b5 e8b5 a6b5');
INSERT INTO evaluations VALUES(21259,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(21259,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(21259,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(21259,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(21259,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(21259,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(21259,36,'cp 67','b1d2 f8d6 f4g3 g8f6 c2c4 e8g8 f1d3 c6b4 d3e2 c7c5 d4c5 d6c5 a2a3 b4c6 b2b4 c5d6 e1g1 d6g3 h2g3 c8d7 d1c2 d8e7 f1d1 f8d8 a1c1 a8c8 c2b2 d7e8 c4d5 f6d5 e3e4 d5b6 e4e5 b6a4 b2b3 a4b6 d2e4 d8d1 c1d1 c8d8 b3c3 d8d1 e2d1 e8d7');
INSERT INTO evaluations VALUES(23024,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(23024,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(23024,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(23024,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(23024,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(23024,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(23024,36,'cp 116','b1c3 c7c6 b5a4 f8b4 a2a3 b4a5 d2d4 g8e7 b2b4 a5c7 c1g5 f7f6 g5h4 e8g8 e1g1 d7d5 e4d5 e7d5 c3d5 c6d5 a4c2 d8d7 f1e1 d7f7 d1d3 g7g6 d3f3 c8e6 e1e3 a8e8 a1e1 f7d7 h2h3 c7b6 c2a4 d7a4 e3e6 e8e6 e1e6 a4d7 e6f6');
INSERT INTO evaluations VALUES(23021,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(23021,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(23021,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(23021,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(23021,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(23021,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(23021,36,'cp 107','c7c6 b5a4 f8b4 a2a3 b4a5 d2d4 g8e7 b2b4 a5c7 c1g5 f7f6 g5h4 e8g8 e1g1 d7d5 e4d5 e7d5 c3d5 c6d5 a4c2 d8d7 f1e1 d7g4 h4g3 g4d1 c2d1 c7g3 h2g3 c8f5 a1c1 f8f7 c1c5 a8d8 d1b3 f7d7 b3a4 d7f7 f2f3 b7b6 c5c1 g8f8 c1c6 f7b7 g1f2 f5d3 a4b3 b7d7 e1d1 d3g6');
INSERT INTO evaluations VALUES(19101,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(19101,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(19101,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(19101,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(19101,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(19101,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(19101,36,'cp 103','b5a4 f8b4 a2a3 b4a5 d2d4 g8e7 b2b4 a5c7 c1g5 f7f6 g5h4 e8g8 e1g1 d7d5 e4d5 e7d5 c3d5 c6d5 a4c2 d8d7 d1d3 g7g6 d3f3 d7g4 f3d5 c8e6 d5b7 c7h2 g1h2 g4h4 h2g1 a8d8 c2e4 d8d4 e4f3 d4d7 b7c6 h4c4 f1c1 c4c6 f3c6 d7c7 b4b5 f8c8 c1d1 e6b3 d1d6 g8f7 a3a4');
INSERT INTO evaluations VALUES(55514,30,'cp -412',NULL);
INSERT INTO evaluations VALUES(55514,31,'cp -412',NULL);
INSERT INTO evaluations VALUES(55514,32,'cp -397',NULL);
INSERT INTO evaluations VALUES(55514,33,'cp -399',NULL);
INSERT INTO evaluations VALUES(55514,34,'cp -393',NULL);
INSERT INTO evaluations VALUES(55514,35,'cp -407',NULL);
INSERT INTO evaluations VALUES(55514,36,'cp -400','d8h4 e1e2 h4e4 e2f2 e4d4 f2e2 d6e5 g1f3 d4c5 b1c3 b8c6 d2d3 g8f6 d1e1 c8g4 c1e3 c6d4 e2d1 g4f3 g2f3 e8c8 e1f2 c5b6 d1c1 g7g6 c1b1 f8b4 c3a4 b6c6 a2a3 b4a5 c2c3 c6a4 c3d4 e5d4 f1h3 c8b8 e3g5 d8d6 f2c2 a4c2 b1c2 f6d5 b2b4 a5b6 c2b1 d5c3 b1b2');
INSERT INTO evaluations VALUES(16884,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(16884,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(16884,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(16884,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(16884,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(16884,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(16884,36,'cp -39','d8e7 e5f3 d4b5 c3b5 e7e4 d1e2 e4e2 e1e2 f6d5 d2d3 a7a6 b5c3 d5c3 b2c3 b7b6 c1e3 c8b7 h1b1 d7d6 g2g3 e8c8 a2a4 f8e7 f3d4 h8e8 e2d2 g7g6 h2h4 d8d7 b1e1 c7c5 d4e2 b7c6 c3c4 d6d5 c4d5 d7d5 c2c4 d5d7 e2c3 c8b7 a4a5');
INSERT INTO evaluations VALUES(1121,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(1121,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(1121,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(1121,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(1121,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(1121,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(1121,36,'cp -36','e5f3 d4b5 c3b5 e7e4 d1e2 e4e2 e1e2 f6d5 d2d3 a7a6 b5c3 d5c3 b2c3 b7b6 c1e3 d7d6 a2a4 c8b7 g2g3 f8e7 h1b1 e8c8 f3d4 h8e8 e2d2 e7f6 b1e1 g7g6 h2h4 e8e5 d4e2 f6g7 e3d4 g7h6 d4e3 h6f8 e3d4 e5a5 c3c4 b7c6 e2c3 d8e8 e1e8 c6e8 d4e3 e8d7 d3d4 h7h6 d2d3 d7f5 d3d2');
INSERT INTO evaluations VALUES(49638,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(49638,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(49638,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(49638,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(49638,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(49638,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(49638,36,'cp -141','d7d5 b1c3 b8c6 c1g5 d5d4 c3e2 c8e6 a2a3 h7h6 g5d2 g7g5 g2g3 d8d6 f1h3 g5g4 h3g2 h6h5 h2h3 h8g8 f3f4 e8c8 h3h4 e5f4 d2f4 d6c5 b2b4 c5b6 f4g5 f8e7 a1b1 c6e5 e2f4 b6d6 g1e2 e5f3 e1f1 f3g5 h4g5');
INSERT INTO evaluations VALUES(48834,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(48834,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(48834,32,'cp -126',NULL);
INSERT INTO evaluations VALUES(48834,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(48834,34,'cp -133',NULL);
INSERT INTO evaluations VALUES(48834,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(48834,36,'cp -130','b1c3 f8c5 c1g5 c7c6 d1d2 d5d4 c3e2 b8d7 d2c1 a7a5 g1h3 a5a4 h3f2 e8g8 g5d2 a4a3 b2b3 c5d6 e2g3 g7g6 f1e2 f6e8 e1g1 e8c7 b3b4 h7h5 f3f4 h5h4 g3h1 e5f4 d2f4 d6f4 c1f4 c7e6 f4h6 d7e5');
INSERT INTO evaluations VALUES(48826,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(48826,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(48826,32,'cp -126',NULL);
INSERT INTO evaluations VALUES(48826,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(48826,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(48826,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(48826,36,'cp -121','f8c5 c1g5 c7c6 d1d2 d5d4 c3e2 b8d7 d2c1 a7a5 g1h3 a5a4 a2a3 d8b6 g5d2 h7h5 h3f2 c5d6 h2h4 c6c5 a1b1 b6c7 c2c3 b7b5 e2g3 d7f8 c3d4 e5d4 f3f4 f8g6 e4e5 g6e5 f4e5 d6e5 g3e4 f6e4 d3e4 c5c4 d2b4 d4d3 f1d3 c4d3 c1c7 e5c7 f2d3');
INSERT INTO evaluations VALUES(55679,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(55679,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(55679,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(55679,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(55679,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(55679,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(55679,36,'cp 92','c1e3 b8c6 d4d2 f7f5 e4f5 c8f5 b1c3 d8d7 e1c1 g8f6 g1e2 a7a6 f2f3 h7h5 e2f4 e8c8 f1c4 g7g6 h2h3 f8g7 c4b3 c6e5 c3e4 d8e8 e4f6 g7f6 f4d5 f6d8 a2a3 h5h4 h1e1 e5c6 c1b1 f5e6 b3a4 d7f7 a4c6 b7c6 d5b4');
INSERT INTO evaluations VALUES(55817,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(55817,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(55817,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55817,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55817,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55817,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(55817,36,'cp 11','d5c4 g1f3 c7c5 f1g2 b8c6 d1a4 c8d7 a4c4 b7b5 c4b5 c6d4 b5d3 d7c6 b1d2 g8e7 e1g1 d4f3 d2f3 d8d3 e2d3 a8d8 b2b3 e7f5 c1b2 f8e7 f1d1 e8g8 a1c1 c6d5 f3e5 d8c8 g2d5 e6d5 c1c2 f7f6 e5f3 f8d8 b2a3');
INSERT INTO evaluations VALUES(6460,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(6460,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(6460,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(6460,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(6460,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(6460,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(6460,36,'cp 51','d7d6 a2a3 a7a6 b2b4 c5a7 c1b2 g8e7 f1e2 e8g8 d1c2 c8f5 d2d3 g8h8 a1d1 h7h6 e1g1 d8d7 f1e1 f5h7 c2b3 e7f5 h2h3 f7f6 a3a4 a6a5 b4b5 c6b4 d3d4 e5d4 f3d4 f5d4 e3d4 a8e8');
INSERT INTO evaluations VALUES(6000,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(6000,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(6000,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(6000,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(6000,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(6000,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(6000,36,'cp 51','a2a3 a7a6 d2d4 c5a7 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 b2b4 g8f6 f1e2 c7c6 f2f3 d8e7 e1f2 c8e6 a3a4 e5e4 a4a5 h8d8 c3a4 h7h5 h2h4 e7e8 c1b2 d8d2 b2c3 e4f3 g2f3 d2e2 f2e2 e6c4 e2f2 f6d5 c3d4 a7d4 e3d4');
INSERT INTO evaluations VALUES(45952,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45952,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45952,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(45952,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45952,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45952,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45952,36,'cp 13','f8e7 c1g5 e8g8 g5f6 e7f6 f1c4 a7a6 e1g1 c8e6 d1e2 b8c6 c3d5 b7b5 d5f6 d8f6 c4e6 f6e6 f1d1 f8d8 d1d2 a8c8 a1d1 c6e7 e2e3 h7h6 c2c3 d8d7 h2h3 c8c4 e3f3 e6g6 g2g3 f7f5 e4f5 e7f5 f3d5 g8h8 d5a8 h8h7 a8a6');
INSERT INTO evaluations VALUES(40161,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(40161,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(40161,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(40161,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40161,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(40161,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40161,36,'cp 29','c1g5 e8g8 g5f6 e7f6 f1c4 a7a6 e1g1 c8e6 d1d3 b8c6 c3d5 b7b5 d5f6 d8f6 c4e6 f6e6 f1d1 f8d8 d1d2 h7h6 a1d1 c6e7 c2c3 a8c8 d3e3 d8d7 h2h3 c8c4 a2a3 c4c6 b3a1 f7f5 a1c2 f5e4 e3e4 d6d5 e4f3 c6c4 f3e2 g8h8 c2b4');
INSERT INTO evaluations VALUES(37767,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(37767,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(37767,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37767,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(37767,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(37767,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37767,36,'cp 5','d7d6 d2d4 c5d4 f3d4 g8f6 b1c3 e8g8 h2h3 b8c6 f1e1 c6d4 d1d4 a7a6 a2a4 c8d7 d4d3 d7c6 c1g5 d8a5 d3g3 e7e6 g3h4 f6e4 c3e4 c6e4 e1e4 d6d5 c4d5 e6d5 e4b4 b7b5 c2c3 d5d4 g5f6 g7f6 h4f6 d4c3 f6c3 a8c8');
INSERT INTO evaluations VALUES(37961,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(37961,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(37961,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37961,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(37961,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(37961,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37961,36,'cp 12','d2d4 c5d4 f3d4 b8c6 d4c6 b7c6 b1c3 g8f6 e4e5 d6e5 d1d8 e8d8 f1e1 c8f5 e1e5 f6d5 e5e2 e7e6 c3a4 h7h5 c1d2 a7a5 c4b3 d8e7 a1e1 h8b8 h2h4 g7f6 g2g3 d5b6 a4b6 b8b6 d2e3 b6b7 e3c5 e7d8 e1d1 d8c7 e2d2 a8d8 c5d6 c7c8 f2f3 f6g7 g1g2 d8d7 d6c5 d7d2 d1d2');
INSERT INTO evaluations VALUES(21504,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(21504,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(21504,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(21504,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(21504,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(21504,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(21504,36,'cp 175','g1f3 b7b5 c3b5 f8b4 b5c3 g8f6 f1c4 f6e4 e1g1 e4d6 d1a4 b4c3 c4e6 c8d7 e6d7 d8d7 b2c3 e8g8 c1f4 c6e5 a4d7 e5d7 f1c1 d7b6 a1b1 a8c8 f3d2 f8e8 g1f1 d6e4 d2e4 e8e4 g2g3 c7c5 d4c5 c8c5 f4e3 c5c4 e3b6 a7b6 b1b6 e4e8 b6b3 e8c8 b3a3 g7g5 f1e2 c8c7 e2d3');
INSERT INTO evaluations VALUES(26961,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(26961,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(26961,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(26961,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(26961,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(26961,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(26961,36,'cp 152','d1b3 d8d7 b3b7 a8b8 b7a6 g4f3 g2f3 g8f6 d2d3 f8e7 c1e3 e8g8 c4b3 d6d5 b3a4 d5d4 a6c6 d7c6 a4c6 d4e3 f2e3 b8b2 b1d2 b2c2 e1e2 c2c3 h1c1 c3c1 a1c1 f8b8 c6a4 c7c5 d2c4 b8b4 a4c6 g7g6 e2d2 e7d8 c4e5 b4b2 c1c2 b2c2 d2c2 d8c7 d3d4 f6h5 c2c3 c5d4 e3d4 h5f4');
INSERT INTO evaluations VALUES(40061,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(40061,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(40061,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(40061,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(40061,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(40061,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(40061,36,'cp 35','c4d5 e6d5 f1d3 b8d7 g1f3 e8g8 e1g1 f6e4 g5f4 d7f6 d1c2 f8e8 a2a4 a7a5 f3e5 e7d6 h2h3 d8e7 f1c1 h7h6 c3e2 f6h5 f4h2 e7g5 e5f3 d6h2 g1h2 g5e7 d3e4 d5e4 f3e5 e7h4 h2g1 h5f6 c2c5 c8e6');
INSERT INTO evaluations VALUES(40058,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40058,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(40058,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40058,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40058,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40058,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(40058,36,'cp 29','e6d5 f1d3 e8g8 g1f3 f8e8 d1c2 h7h6 g5f4 c8e6 e1g1 e7f8 h2h3 b8d7 a1c1 a8c8 f1d1 a7a6 c3a4 f6e4 f3e5 d7e5 d4e5 d8a5 a2a3 c6c5 d3e4 d5e4 a4c3 a5b6 d1d2 e6b3 c2b1 c8d8 c3e4 d8d2 e4d2 b3e6');
INSERT INTO evaluations VALUES(40104,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(40104,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(40104,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(40104,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(40104,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(40104,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(40104,36,'cp 38','f1d3 e8g8 g1f3 f8e8 d1c2 h7h6 g5f4 c8e6 h2h3 b8d7 e1g1 a8c8 a1c1 e7f8 a2a3 b7b5 c3e2 a7a6 f1d1 c6c5 d4c5 d7c5 c2b1 c5d3 b1d3 f6e4 c1c8 d8c8 d1c1 c8a8 f3d4 e8c8 c1c7 c8c7 f4c7 e6d7 c7a5 a8b7 e2f4');
INSERT INTO evaluations VALUES(40113,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(40113,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(40113,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(40113,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40113,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40113,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40113,36,'cp 27','e8g8 g1f3 f8e8 e1g1 h7h6 g5f4 c8g4 d1b3 d8b6 b3c2 b8d7 a2a3 b6d8 f3e5 d7e5 f4e5 g4e6 h2h3 f6d7 e5g3 a7a5 c3e2 d7f6 a1c1 e7f8 g3e5 e6d7 e2c3 d7e6 g2g4 f6d7 e5g3 d7b6 c3e2 f8d6 e2f4 e6d7 f4h5 d6g3');
INSERT INTO evaluations VALUES(35773,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(35773,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(35773,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(35773,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(35773,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(35773,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35773,36,'cp 39','g1f3 f8e8 d1c2 h7h6 g5f4 c8e6 h2h3 b8d7 e1g1 a7a6 a2a3 a8c8 f1d1 b7b5 f4h2 c6c5 d4c5 d7c5 h2e5 d8d7 a1c1 c5d3 c2d3 d7b7 c3e2 c8c4 f3d2 c4c1 d1c1 e8c8 c1c8 b7c8 e5c3 e6d7 e2f4');
INSERT INTO evaluations VALUES(35775,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(35775,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(35775,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(35775,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(35775,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(35775,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35775,36,'cp 28','f8e8 d1c2 h7h6 g5f4 c8e6 e1g1 b8d7 h2h3 a8c8 a1c1 a7a6 f1d1 e7f8 a2a3 b7b5 c3e2 d8e7 f3e5 c6c5 d4c5 d7c5 e2d4 f6e4 c2e2 c5d3 e5d3 c8c4 d3e5 c4c1 d1c1 e8c8 c1c6 e7b7 e2c2 c8c6 e5c6 b7c8 b2b4 e4d6 f4d6 f8d6 d4e6 c8e6 c6d4 e6e4 c2c8 g8h7');
INSERT INTO evaluations VALUES(47512,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(47512,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(47512,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(47512,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(47512,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(47512,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(47512,36,'cp 82','c3e2 f7f6 d2d3 e4g5 c1g5 f6g5 h2h4 g5g4 d3d4 c7c5 d4c5 b8c6 e2d4 d8c7 d4c6 b7c6 d1d4 a8b8 a2a3 g7g6 h4h5 f8g7 f2f4 g4f3 g1f3 e8g8 h5g6 f8f3 g6h7 g8h8 g2f3 g7e5 d4g1 c8f5 f1d3 f5d3 c2d3 e5b2 a1b1 c7e5 e1f1 b8f8 g1f2 b2d4 b1e1 e5f6 f2g3 d4c5 g3e5 f6e5 e1e5 f8f3 f1e2 f3f2 e2d1 c5a3 e5g5 f2f8 h1h6 a3d6 g5g4 a7a5 d1e2 a5a4 g4a4');
INSERT INTO evaluations VALUES(53856,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(53856,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(53856,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(53856,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(53856,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(53856,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(53856,36,'cp 85','b8d7 a2a3 a7a6 h2h3 d8c7 f1d3 g8f6 e1g1 f8e7 a3a4 e8g8 f1e1 e6e5 a4a5 e5d4 f3d4 f8e8 c1f4 e7f8 f4g3 d7c5 b2b4 c5d3 c2d3 c6c5 b4c5 c7c5 d4e2 b7b5 a5b6 c5b6 a1b1 b6a5 d1d2 a5d8 g1h1 c8d7 e2d4 a6a5');
INSERT INTO evaluations VALUES(18544,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(18544,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(18544,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(18544,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(18544,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(18544,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(18544,36,'cp 106','a2a3 a7a6 e4e5 c6c5 e5d6 f8d6 d4c5 d6c5 c3e4 d8c7 e4c5 c7c5 c1e3 c5c7 d1d4 g8f6 e1c1 b7b5 e3f4 c7b6 f4d6 c8b7 a3a4 b6d4 d1d4 f6e4 d6a3 e4f2 h1g1 e6e5 d4b4 e8g8 a4b5 f8d8 b5b6 h7h6 b4b3 f2g4 h2h3 d7c5');
INSERT INTO evaluations VALUES(55221,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(55221,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(55221,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(55221,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(55221,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(55221,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(55221,36,'cp -18','g8f6 b1c3 c7c6 a2a4 f6d5 f4g3 d5c3 b2c3 b7b5 e2e4 e7e6 g1f3 f8d6 f1e2 a7a6 g3d6 d8d6 e4e5 d6e7 f3d2 b8d7 e2f3 c8b7 d2e4 e8g8 e4d6 d7b6 e1g1 b6a4 a1a4 b5a4 d1a4 a6a5 f1b1 f8b8 h2h3 b7a6 b1b8 a8b8 a4c6 e7g5 g1h2 g5f4 g2g3');
INSERT INTO evaluations VALUES(27529,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(27529,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(27529,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(27529,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(27529,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(27529,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(27529,36,'cp 65','e1g1 g8h6 d4c5 f8c5 b2b4 c5e7 a2a4 b6c7 c1f4 e8g8 b1d2 f7f6 e5f6 c7f4 f6e7 c6e7 c3c4 a8e8 d1b3 h6g4 h2h3 g4f6 c4d5 e6d5 f1e1 e7f5 e2d3 g8h8 e1e8 f8e8 b4b5 f5d4 b3d1 d4e6 d3f1 e6c5 d2b3 c5e4 d1c1 f4c1 b3c1 g7g5 f3d4 g5g4 h3g4 f6g4 c1d3');
INSERT INTO evaluations VALUES(27528,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(27528,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(27528,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(27528,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(27528,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(27528,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(27528,36,'cp 62','g8h6 d4c5 f8c5 b2b4 c5e7 a2a4 b6c7 c1f4 e8g8 b1d2 f7f6 e5f6 c7f4 f6e7 c6e7 c3c4 g8h8 d1b3 a8d8 e2d3 d7c6 b4b5 d5c4 d3c4 c6d5 c4d5 e6d5 g2g3 f4f6 a1e1 e7g6 b3e3 a7a6 e3e6 a6b5 e6f6 g7f6 a4b5 h6f5');
INSERT INTO evaluations VALUES(27359,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(27359,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(27359,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(27359,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(27359,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(27359,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(27359,36,'cp 69','d4c5 f8c5 b2b4 c5e7 a2a4 a7a6 c1f4 b6d8 f4h6 g7h6 f1e1 f7f6 b1d2 e8g8 e5f6 e7f6 a1c1 d8e7 b4b5 a6b5 a4b5 c6a5 e2f1 e7g7 c3c4 f6b2 c1c2 d7e8 c4d5 e6d5 d1b1 b2c3 b1c1 d5d4 f3e5 a8c8 c1a3 g7f6 d2f3 g8h8 f1d3 b7b6');
INSERT INTO evaluations VALUES(47306,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(47306,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47306,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(47306,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(47306,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(47306,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47306,36,'cp 39','e5f6 d4c3 f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 h7h5 g1f3 c8g4 d2g5 g7h6 g5h6 h8h6 f1e2 b8c6 h2h3 g4f3 e2f3 a8d8 d1d8 c6d8 h1e1 h5h4 e1e5 b7b6 c1d2 e7e6 e5h5 h6h5 f3h5 d8c6 g2g3 h4g3 f2g3 e8e7 h5f3 c6e5 f3e2 f7f5 h3h4 e7f6 d2e3 e5g6 c2c3 e6e5 e2f3 e5e4 f3e2 g6e5 a2a4 e5f7 e2c4');
INSERT INTO evaluations VALUES(47082,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(47082,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(47082,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47082,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47082,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(47082,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(47082,36,'cp 28','d4c3 f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 h7h5 d2f4 g7h6 f4h6 h8h6 g1e2 b8c6 e2c3 c8d7 c3d5 e8c8 f1b5 d8e8 h1e1 e7e6 d5b4 c6b8 b5f1 h5h4 d1d2 d7c6 b4c6 b8c6 f1b5 e8g8 b5c6 b7c6 g2g3 h6h5 e1d1 h4g3');
INSERT INTO evaluations VALUES(47083,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(47083,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(47083,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(47083,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(47083,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(47083,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(47083,36,'cp 30','f6g7 c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 h7h5 g1f3 c8g4 f1e2 b8c6 h2h3 g4f5 d2g5 a8d8 h1e1 d8d6 e2b5 d6d1 c1d1 e7e6 f3h4 f5h7 d1c1 a7a6 b5c6 b7c6 h4f3 f7f6 g5d2 e6e5 g2g4 h5g4 h3g4 h7g6 g4g5 h8h5 g5f6 g7f6');
INSERT INTO evaluations VALUES(46865,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(46865,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46865,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(46865,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46865,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46865,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(46865,36,'cp 37','c3d2 d1d2 d8d2 c1d2 f8g7 e1c1 h7h5 g1f3 c8g4 d2f4 g7h6 f4h6 h8h6 f1e2 b8c6 h2h3 g4d7 h1e1 e8c8 f3h4 h6f6 e2f3 f6f4 e1e4 f4e4 f3e4 e7e6 f2f4 c6b4 c2c3 d7c6 d1d8 c8d8 e4c6 b4c6 c1d2 c6e7 d2e3 e7d5 e3e4 d8e7 f4f5 d5b6 f5e6 e7e6 h4f5 b6a4 f5g7 e6f6 g7h5 f6g5 h5g3 a4b2 e4d4 b2a4');
INSERT INTO evaluations VALUES(46868,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(46868,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(46868,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46868,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(46868,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46868,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46868,36,'cp 37','d1d2 d8d2 c1d2 f8g7 e1c1 c8g4 f2f3 g4d7 f1c4 b8c6 g1e2 h8g8 e2f4 c6e5 c4e2 d7c6 h1f1 g7h6 g2g4 a8d8 f4h5 h6d2 d1d2 d8d6 h2h3 e5d7 f1d1 d6d2 c1d2 g8g5 b2b3 g5c5 f3f4 e7e6 b3b4 c5d5 d2e3 d7b6 e3f2 d5d1');
INSERT INTO evaluations VALUES(46867,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(46867,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(46867,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46867,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(46867,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(46867,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(46867,36,'cp 33','d8d2 c1d2 f8g7 e1c1 c8f5 g1f3 b8d7 d2e3 a7a6 f1e2 e8c8 f3g5 d7e5 d1d8 h8d8 f2f4 f7f6 f4e5 f6g5 h1f1 f5e6 e3g5 g7e5 g2g3 e5d6 a2a4 d8g8 g5h6 a6a5 e2d3 c8d7 f1e1 e6g4 c1d2 b7b6 c2c3 g8d8 d3c4 g4f3 c4b5 c7c6 b5d3 d8h8 e1f1');
INSERT INTO evaluations VALUES(33645,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(33645,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(33645,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(33645,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(33645,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(33645,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(33645,36,'cp 33','c1d2 f8g7 e1c1 c8f5 g1f3 b8d7 d2e3 a7a6 f1e2 e8c8 h1e1 d7b6 d1d8 h8d8 c2c3 d8g8 g2g3 e7e6 e2d1 b6c4 e3f4 c4d6 d1c2 f5c2 c1c2 h7h5 h2h3 g7f6 f3e5 a6a5 h3h4 b7b6 e5d3 a5a4 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(33644,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(33644,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(33644,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(33644,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(33644,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(33644,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(33644,36,'cp 36','f8g7 e1c1 c8f5 g1f3 b8d7 d2e3 a7a6 f1e2 e8c8 h1e1 d7b6 d1d8 h8d8 c2c3 e7e6 h2h3 f5e4 g2g4 e4d5 a2a3 b6c4 e3f4 f7f6 f3d2 c4e5 f4e3 e5g6 e2d3 d8g8 f2f3 g8f8 e3c5 f8e8 c3c4 d5c6 d3e4 c6e4 d2e4');
INSERT INTO evaluations VALUES(33402,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(33402,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(33402,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(33402,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(33402,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(33402,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(33402,36,'cp 39','e1c1 c8f5 g1f3 b8d7 d2e3 a7a6 f1e2 e8c8 h1e1 d7b6 d1d8 h8d8 c2c3 e7e6 h2h3 f5e4 f3g5 e4g6 g2g4 b6d5 e3d2 h7h6 g5f3 b7b5 e2d1 d5b6 d1c2 b6c4 d2f4 g6c2 c1c2 d8d5 f4c1 c4d6 f3d2 a6a5 d2e4 d6e4 e1e4 h6h5 c1d2 h5g4 h3g4');
INSERT INTO evaluations VALUES(33401,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(33401,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(33401,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(33401,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(33401,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(33401,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(33401,36,'cp 33','c8f5 g1f3 b8d7 d2e3 a7a6 f1e2 e8c8 f3g5 f5g6 g2g3 h7h5 f2f4 e7e6 h1e1 d7b6 g5f3 b6d5 e3d4 g7d4 d1d4 d5e3 d4d8 h8d8 c2c3 e3g4 a2a4 a6a5 e1d1 d8g8 d1d2 g6e4 f3g5 e4g6 b2b4 g4h2 b4a5 g8d8 a5a6 d8d2 c1d2 b7a6');
INSERT INTO evaluations VALUES(51735,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(51735,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(51735,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(51735,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(51735,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(51735,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(51735,36,'cp 114','d2d4 f8g7 d4c5 d8a5 c2c3 a5c5 b1a3 a7a6 c1e3 c5c7 c4e2 b7b5 e2b5 g8f6 b5e2 f6e4 a3c4 b8c6 f3g5 e4g5 c4d6 e8f8 e3g5 c8b7 g5f4 c6e7 e1g1 e6e5 f4e3 b7d5 c3c4 c7d6 b2b4 d6b4 a1b1 b4a3 b1b3 a3a2 c4d5 e7f5 b3b7 f5e3 f2e3 f8e8 e2g4 a8d8 d1f3');
INSERT INTO evaluations VALUES(51738,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(51738,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(51738,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(51738,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(51738,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(51738,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(51738,36,'cp 96','f8g7 d4c5 d8a5 c2c3 a5c5 b1a3 g8f6 b2b4 c5b6 e4e5 f6g4 d1d4 b6d4 c3d4 e8g8 c4e2 d7d5 h2h3 g4h6 h3h4 a7a6 h4h5 h6f5 g2g4 f5e7 b4b5 f7f6 e5f6 g7f6 c1h6 f6g7 h6g5 g7f6 g5f6 f8f6 h5g6 h7g6 f3e5 a6b5 a3b5 b8c6');
INSERT INTO evaluations VALUES(53262,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(53262,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(53262,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(53262,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(53262,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(53262,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(53262,36,'cp -26','b8c6 e2e3 g8f6 d4c5 e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3a4 c5d6 c2c3 e8g8 h2h3 f8e8 a2a3 h7h6 f1e1 c8d7 b2b4 a7a6 c3d4 c6d4 f3d4 e5d4 a4c5 d7c6 c1b2 d6c5 b4c5 f6e4 d3e4 c6e4 d1d4 d8d4 b2d4 e4c6');
INSERT INTO evaluations VALUES(19784,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(19784,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(19784,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(19784,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(19784,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(19784,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(19784,36,'cp -28','d4c5 g8f6 e2e3 e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 h2h3 h7h6 c1d2 a7a5 c2c3 d4c3 d2c3 d8d6 e2g3 f8d8 d3e2 c5d4 c3d4 c6d4 f3d4 d6d4 d1d4 d8d4 f1d1 d4d1 a1d1 c8e6 a2a3 a8c8 d1d6 g7g6 f2f3 c8c2 d6b6 f6d7 b6b7 d7c5 b7c7 a5a4 e2b5');
INSERT INTO evaluations VALUES(19752,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(19752,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(19752,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(19752,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(19752,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(19752,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(19752,36,'cp -21','g8f6 e2e3 e7e6 f1d3 f8c5 e1g1 e8g8 e3e4 d5d4 c3a4 c5d6 c2c3 e6e5 h2h3 f8e8 c1d2 h7h6 f1e1 d4c3 a4c3 c8e6 d2e3 a7a6 a1c1 a8c8 d3b1 b7b5 c3e2 e6c4 e2c3 d8e7 a2a3 c6a5');
INSERT INTO evaluations VALUES(56386,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(56386,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(56386,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(56386,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(56386,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(56386,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(56386,36,'cp -55','d5d4 c3e2 g8e7 f2f4 b8c6 c1d2 e5f4 e2f4 e7g6 g1f3 f8d6 g2g3 g6e5 f1g2 e8g8 e1g1 c8g4 a2a3 d8d7 f3e5 c6e5 d1e1 a8e8 h2h3 g4e6 a3a4 e5g6 f4e6 e8e6 b2b4 g6e5 e1f2 c7c5 b4c5 d6c5');
INSERT INTO evaluations VALUES(56576,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(56576,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(56576,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(56576,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(56576,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(56576,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(56576,36,'cp -41','c3e2 g8e7 c1d2 a7a5 f2f4 e5f4 e2f4 e7g6 g2g3 f8b4 f1g2 e8g8 a2a4 b8c6 g1e2 b4c5 e1g1 c6e5 g1h1 b7b6 h2h3 c7c6 e2g1 a8a7 f4g6 h7g6 g1f3 e5f3 d1f3 c5b4 f3f4 c6c5 d2b4 a5b4 h1g1');
INSERT INTO evaluations VALUES(56588,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(56588,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(56588,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(56588,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(56588,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(56588,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(56588,36,'cp -54','c7c5 f2f4 b8c6 g1f3 e5f4 e2f4 f8d6 g2g3 g8e7 f1e2 e7g6 e1g1 e8g8 f1f2 d6f4 c1f4 f7f6 h2h4 h7h6 h4h5 g6e5 f4e5 c6e5 f3e5 f6e5 f2f8 d8f8 a2a4 f8e7 a4a5 e7g5 g1g2 c8e6 d1c1 g5e7 c1d2 e7d7 g2g1 e6h3');
INSERT INTO evaluations VALUES(54489,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(54489,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(54489,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(54489,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(54489,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(54489,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(54489,36,'cp -58','f2f4 e5f4 e2f4 b8c6 g1f3 f8d6 g2g3 g8e7 f1e2 e7g6 e1g1 e8g8 f1f2 b7b5 c2c3 g6e5 c3d4 e5g4 f2g2 c5d4 g1h1 g4e5 f3e5 c6e5 d1g1 d6c5 f4d5 c8h3 g2f2 h3e6 c1f4 f7f6 a1c1 a8c8 f2f1 h7h6 g3g4 e6d5 e4d5 d8d5 g1g2 d5g2 h1g2');
INSERT INTO evaluations VALUES(21425,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(21425,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(21425,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(21425,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(21425,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(21425,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(21425,36,'cp 37','f8e7 e1g1 g8f6 c2c4 e8g8 b2b3 b7b6 c1b2 c8b7 b1d2 c6a5 c4d5 e6d5 e2e3 e7d6 a1c1 a8c8 d2b1 f8e8 b1c3 a5c6 a2a3 c6e7 f1e1 e7g6 c3b5 a7a6 b5a7 c8a8 a7c6 d8d7 d1c2 a6a5 e1d1 g6e7 c6e5 d7e6 f3g5');
INSERT INTO evaluations VALUES(45650,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(45650,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(45650,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45650,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45650,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(45650,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45650,36,'cp 29','b8d7 e1g1 f8e7 b2b3 b7b6 b1c3 c8a6 e2e4 e7b4 e4d5 c6d5 c3b5 e8g8 d1e2 a6b5 c4b5 a8c8 c1b2 f6e4 a1c1 e4c3 b2c3 b4c3 e2e3 c3b2 c1b1 b2c3 f1c1 c3b4 c1c6 c8c6 b5c6 d7f6 f3e5 d8c7 b1c1 f8c8 e3e2');
INSERT INTO evaluations VALUES(12165,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(12165,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(12165,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(12165,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(12165,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(12165,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(12165,36,'cp 34','e1g1 f8e7 b2b3 b7b6 b1c3 e8g8 d1d3 c8b7 c1b2 d8c7 f1d1 a8d8 f3d2 f8e8 a1c1 c7b8 e2e4 d5e4 d2e4 c6c5 d4d5 e6d5 c4d5 d7e5 d3e2 e5g6 e2c4 g6e5 e4f6 e7f6 c4f4 b8a8 c3e4 e5f3 g2f3 f6b2 e4d6 d8d6 f4d6 b2c1 d1c1 h7h6');
INSERT INTO evaluations VALUES(12164,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(12164,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(12164,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(12164,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12164,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(12164,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(12164,36,'cp 25','f8e7 b2b3 e8g8 d1c2 b7b6 f1d1 c8b7 b1c3 d8c8 c1b2 b7a6 c3b1 f8d8 b1d2 a6b7 a1c1 d5c4 d2c4 c6c5 c4d2 c8b8 d4c5 e7c5 d2c4 c5f8 a2a4 d8c8 c2d2 d7c5 b2f6 g7f6 f3d4');
INSERT INTO evaluations VALUES(7782,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(7782,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(7782,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(7782,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(7782,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(7782,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(7782,36,'cp 33','b2b3 b7b6 d1c2 c8b7 b1c3 e8g8 f1d1 d8c8 e2e4 f6e4 c3e4 d5e4 c2e4 c6c5 d4d5 f8e8 e4c2 e7f6 a1b1 e6d5 c4d5 d7e5 f3e5 f6e5 c1b2 e5b2 c2b2 c8f5 b3b4 c5b4 b2d4 e8e5 f2f4 e5e8 d4b4 a8d8 b4a4 h7h6 a4a7');
INSERT INTO evaluations VALUES(16232,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(16232,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(16232,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(16232,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(16232,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(16232,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(16232,36,'cp -37','d7c6 f3e5 f8e7 d2d3 e8g8 h2h3 c8e6 e5f3 h7h6 c1f4 d6f5 b1c3 a7a5 c3e4 a5a4 d3d4 f8e8 c2c3 e7d6 f3e5 d6f8 d1f3 e6d5 f3g4 f5d6 e4c5 b7b6 c5d3 d8c8 g4c8 e8c8 d3b4 a4a3 b2a3 a8a3 e5c6 d5c6 b4c6 a3c3 c6e7 f8e7 e1e7 d6f5 e7e4 c7c6 a2a4 c3b3 a1e1');
INSERT INTO evaluations VALUES(54551,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(54551,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(54551,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54551,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(54551,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(54551,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(54551,36,'cp 21','f1b5 c8d7 b5d7 b8d7 e1g1 f8e7 d2d4 g8f6 d4c5 d7c5 f1e1 e8g8 c1e3 d8d7 b1a3 f8e8 a3c2 b7b6 e3d4 c5e6 d4f6 e7f6 c2e3 a8d8 e3g4 f6e7 g4e5 d7c8 d1a4 a7a6 a1d1 e6c5 a4c6 d8d6 c6c8 e8c8 g2g3 e7f6');
INSERT INTO evaluations VALUES(54469,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(54469,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54469,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54469,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(54469,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54469,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54469,36,'cp 29','c8d7 d1e2 d8e7 e1g1 e7e2 b5e2 b8c6 f1e1 g8e7 d2d4 c5d4 f3d4 c6d4 c3d4 d7e6 b1c3 e7c6 c3d5 e8c8 d5f4 c6d4 f4e6 d4e2 g1f1 f7e6 f1e2 f8c5 c1f4 d8d5 a1c1 c8d7 e1d1 c5d6 f4e3 b7b6 g2g3 h8d8 c1c4 d5d1 e2d1 d6e7 d1e2 e7f6 e3d4 e6e5 d4c3');
INSERT INTO evaluations VALUES(30910,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(30910,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(30910,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(30910,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(30910,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(30910,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(30910,36,'cp 32','d1e2 d8e7 b5d7 b8d7 d2d4 e7e2 e1e2 a8c8 d4c5 f8c5 h1e1 f7f6 c1e3 g8e7 e2f1 e8f7 e3c5 d7c5 b1a3 h8d8 e1e2 e7c6 a1d1 c5a4 a3b5 d8d7 h2h4 a7a6 b5d4 c8e8 e2c2 c6e5 f3e1 a4c5 b2b3 e5c6 d4e2 e8d8 b3b4 c5a4 c2d2');
INSERT INTO evaluations VALUES(56593,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(56593,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(56593,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(56593,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(56593,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(56593,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(56593,36,'cp -76','f7f6 e2g3 c8e6 f1b5 c7c6 b5a4 b8a6 a4b3 a6c5 d2d3 d8d7 h2h3 h7h5 b3e6 c5e6 e1g1 h5h4 g3f5 g7g6 f5h4 g8e7 c2c3 g6g5 h4f5 e8c8 c3d4 g5g4 f3h4 e7f5 h4f5 g4h3 g2g3 e6d4 f5d4 d7d4 c1e3 d4d3 d1d3 h3h2 g1h1 d8d3 e3a7');
INSERT INTO evaluations VALUES(10988,30,'cp 438',NULL);
INSERT INTO evaluations VALUES(10988,31,'cp 421',NULL);
INSERT INTO evaluations VALUES(10988,32,'cp 431',NULL);
INSERT INTO evaluations VALUES(10988,33,'cp 427',NULL);
INSERT INTO evaluations VALUES(10988,34,'cp 424',NULL);
INSERT INTO evaluations VALUES(10988,35,'cp 447',NULL);
INSERT INTO evaluations VALUES(10988,36,'cp 419','e4e5 e8g8 e5f6 e7f6 d4d3 c7c6 f1e2 d7d5 d3f3 d8d6 e1g1 c8d7 e2d3 f6e5 g1h1 f8e8 c1d2 e8e6 a1e1 a8e8 f3h5 g7g6 h5d1 b7b5 a2a3 e6f6 b2b3 e5h2 e1e8 d7e8 g2g3 h2g3 f2g3 f6f1 d1f1 d6g3 f1g2 g3g2 h1g2 e8d7 g2g3 h7h5 c3e2 d7g4 e2d4');
INSERT INTO evaluations VALUES(10972,30,'cp 440',NULL);
INSERT INTO evaluations VALUES(10972,31,'cp 423',NULL);
INSERT INTO evaluations VALUES(10972,32,'cp 428',NULL);
INSERT INTO evaluations VALUES(10972,33,'cp 416',NULL);
INSERT INTO evaluations VALUES(10972,34,'cp 423',NULL);
INSERT INTO evaluations VALUES(10972,35,'cp 416',NULL);
INSERT INTO evaluations VALUES(10972,36,'cp 428','e8g8 e5f6 e7f6 d4d3 c7c6 c3e4 f6e5 e4g5 g7g6 g5f3 e5f6 c2c3 d8b6 a2a4 f8e8 f1e2 a7a5 e1g1 d7d5 h2h3 c8f5 d3d1 a8d8 g1h1 f5e6 c1e3 b6b2 a1b1 b2a2 b1b7 f6c3 e2d3 c3g7 e3d2 c6c5 d3b5 c5c4 b5e8 d8e8 d2a5 c4c3 f3d4');
INSERT INTO evaluations VALUES(15256,30,'cp -269',NULL);
INSERT INTO evaluations VALUES(15256,31,'cp -225',NULL);
INSERT INTO evaluations VALUES(15256,32,'cp -241',NULL);
INSERT INTO evaluations VALUES(15256,33,'cp -241',NULL);
INSERT INTO evaluations VALUES(15256,34,'cp -251',NULL);
INSERT INTO evaluations VALUES(15256,35,'cp -250',NULL);
INSERT INTO evaluations VALUES(15256,36,'cp -238','d5e4 f3e5 d8g5 d2d4 g5g2 h1f1 f8d6 e5c6 c8d7 c6a7 c7c6 b5e2 a8a7 c1e3 e8g8 c2c3 g8h8 d1c1 g2h2 e1d2 f7f5 f1h1 h2g2 h1g1 g2h3 g1h1 d6h2 d4d5 a7a5 d5c6 d7c6 c1f1 h3f1 h1f1 h2e5 e3b6 a5a8 f2f4 e4f3 e2f3 c6f3 f1f3 g7g6 a2a4 h8g7 f3d3 g7f6 d2c2 h7h5 a4a5');
INSERT INTO evaluations VALUES(21762,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(21762,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(21762,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(21762,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(21762,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(21762,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(21762,36,'cp -28','e5e4 f3e1 g8f6 d2d3 c8f5 d3e4 f5e4 e1f3 c6b4 b1c3 e4c2 d1d2 c2g6 a2a3 b4a6 f3e5 c7c6 b2b4 a6c7 c1b2 f8e7 b4b5 c6b5 e5g6 h7g6 a1d1 e8g8 e2e4 e7d6 c3d5 c7d5 e4e5 d6c5 e5f6 d5f6 g2b7 a8b8 b7f3 a7a6 g1g2 d8b6 b2e5 b8d8 d2c3 d8d1 f1d1 c5f2');
INSERT INTO evaluations VALUES(21624,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(21624,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(21624,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(21624,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(21624,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(21624,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(21624,36,'cp -27','f3e1 g8f6 d2d3 c8f5 d3e4 f5e4 e1f3 c6b4 b1c3 e4c2 d1d2 c2g6 a2a3 b4a6 b2b4 c7c6 c1b2 a6c7 f3e5 f8e7 b4b5 c6b5 f1d1 e8g8 e5g6 h7g6 e2e4 d5e4 d2e3 d8e8 c3e4 a8d8 e3a7 f6e4 g2e4 e7f6 b2f6 d8d1 a1d1 e8e4 a7d4 e4e2 f6g7');
INSERT INTO evaluations VALUES(21625,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(21625,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(21625,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(21625,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(21625,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(21625,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(21625,36,'cp -17','g8f6 d2d3 c8g4 c1g5 h7h6 g5f6 d8f6 d3e4 f6b2 b1d2 d5e4 d2e4 c6d4 a1b1 g4e2 b1b2 e2d1 c2c3 d4e2 g1h1 f7f5 e1c2 d1c2 b2c2 e2f4 g3f4 f5e4 g2e4 e8c8 f1g1 c7c6 c2e2 c8b8 f4f5 b8c7 g1g6 d8d7 h1g2 f8c5 h2h3 h8e8');
INSERT INTO evaluations VALUES(49651,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(49651,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(49651,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(49651,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(49651,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(49651,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(49651,36,'cp -28','b8c6 g2g3 f8c5 f1g2 e8g8 e1g1 d7d5 e4d5 d8d5 b1c3 d5d6 c3b5 d6e7 d2d3 c5b6 c1e3 f8e8 b5a3 c8g4 h2h3 g4e6 e3b6 a7b6 a3b5 e7d7 f1e1 a8d8 e2f1 e6d5 a2a3 h7h6 a1d1 f6h7 d3d4 d5f3 g2f3 c6d4 b5d4 e5d4 f3b7 e8e1 d1e1 h7f6');
INSERT INTO evaluations VALUES(16570,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(16570,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(16570,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(16570,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(16570,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(16570,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(16570,36,'cp -31','g2g3 f8c5 f1g2 e8g8 e1g1 d7d5 e4d5 d8d5 b1c3 d5d6 c3b5 d6e7 d2d4 a7a6 b5c7 c6d4 f3d4 e7c7 d4b3 c8g4 e2e1 c5d6 c1e3 a6a5 e1c3 c7e7 e3g5 a5a4 b3d2 f8d8 g5f6 g7f6 d2e4 a8c8 c3e3 f6f5 e4d6 e7d6 h2h3 g4h5 a1e1 e5e4 c2c3');
INSERT INTO evaluations VALUES(55659,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55659,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55659,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55659,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55659,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55659,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55659,36,'cp 0','g1f3 g8f6 e1g1 f8e7 e4e5 d6e5 d1b3 c8e6 c4e6 f7e6 f3g5 d8d5 b3h3 d5d7 f2f4 e5f4 g5e6 e8f7 f1e1 e7b4 b1c3 h8e8 e6g5 f7g6 g5e6');
INSERT INTO evaluations VALUES(55657,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55657,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55657,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55657,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55657,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55657,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55657,36,'cp 0','g8f6 e1g1 f8e7 e4e5 d6e5 d1b3 c8e6 c4e6 f7e6 f3g5 d8d5 b3h3 d5d7 f2f4 e5f4 g5e6 e8f7 f1e1 e7b4 b1c3 h8e8 a1d1 b4d6 e6g5 f7g6 g5e6');
INSERT INTO evaluations VALUES(47926,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47926,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47926,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47926,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47926,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47926,36,'cp 0','e1g1 f8e7 e4e5 d6e5 d1b3 c8e6 c4e6 f7e6 f3g5 d8d5 b3h3 d5d7 f2f4 e5f4 g5e6 e8f7 f1e1 e7b4 b1c3 h8e8 a1d1 b4d6 e6g5 f7g6 g5e6');
INSERT INTO evaluations VALUES(56983,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(56983,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(56983,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(56983,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(56983,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(56983,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(56983,36,'cp 178','f3e5 d8e7 d2d4 d7d6 e5f3 e7e4 c1e3 e4e7 f1d3 g8f6 e1g1 d6d5 f1e1 c8e6 b1c3 b8c6 c3e2 c6b4 d3b5 c7c6 b5a4 b4a6 c2c3 e8c8 a4c2 e7c7 e2f4 e6d7 f3e5 f8d6 d1f3 d6e5 d4e5 c7e5 e3a7 e5d6 b2b4 d8e8 a2a4 e8e1 a1e1 h8e8 e1e8 f6e8 f4d3 d6f6 f3f6 e8f6 a7d4 a6c7 a4a5 d7g4');
INSERT INTO evaluations VALUES(56968,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(56968,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(56968,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(56968,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(56968,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(56968,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(56968,36,'cp 180','d8e7 d2d4 d7d6 e5f3 e7e4 c1e3 e4e7 f1d3 g8f6 e1g1 c8e6 f1e1 d6d5 b1c3 b8c6 c3e2 g7g5 c2c3 f6e4 f3e5 f8g7 e5c6 b7c6 e2g3 f7f5 f2f3 e4g3 h2g3 e8g8 d1a4 e7d6 e1e2 f5f4 g3f4 g5g4 a1e1 g4f3 g2f3 e6f5 d3f5 f8f5 a4c2');
INSERT INTO evaluations VALUES(35251,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(35251,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(35251,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(35251,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(35251,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(35251,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(35251,36,'cp 177','d2d4 d7d6 e5f3 e7e4 c1e3 e4e7 f1d3 g8f6 e1g1 c8e6 f1e1 d6d5 e3f4 b8c6 b1d2 e7d7 c2c3 a7a6 h2h3 f8e7 a2a4 d7d8 a4a5 e8g8 f4g3 f8e8 f3e5 d8c8 d3c2 e7d6 d1f3 c6e7 g3h4 e6f5 h4f6 f5c2 f6e7 d6e5 d4e5');
INSERT INTO evaluations VALUES(35250,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(35250,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(35250,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(35250,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(35250,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(35250,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(35250,36,'cp 188','d7d6 e5f3 e7e4 c1e3 e4e7 f1d3 g8f6 e1g1 c8e6 f1e1 d6d5 b1d2 e7d8 e3f4 f8d6 e1e6 f7e6 d3g6 e8e7 f3e5 b8c6 d2f3 f6e4 e5f7 d8d7 f4d6 e4d6 f7h8 a8h8 d1e2 e7d8 a1e1 h8f8 h2h4 d7e7 c2c3 d8d7 h4h5 a7a6 g6c2 e7f7');
INSERT INTO evaluations VALUES(20287,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(20287,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(20287,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(20287,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20287,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20287,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20287,36,'cp 0','g1f3 d5e4 d2e4 c5d4 c3d4 g8f6 f1d3 f8e7 e4f6 e7f6 e1g1 e8g8 c1e3 c6b4 d3e4 b4d5 f3e5 b7b6 d1d3 h7h6 e3d2 c8b7 a1c1 a8c8 c1c8 b7c8 e4h7 g8h8 h7e4');
INSERT INTO evaluations VALUES(20284,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20284,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20284,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20284,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20284,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20284,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(20284,36,'cp 0','d5e4 d2e4 c5d4 c3d4 g8f6 f1d3 f8e7 e1g1 e8g8 e4f6 e7f6 c1e3 c6b4 d3e4 b4d5 d1d3 h7h6 f3e5 b7b6 e3d2 c8b7 a1c1 a8c8 c1c8 b7c8 d3g3 f6h4 g3a3 a7a5 e5c6 d8d7 c6e5');
INSERT INTO evaluations VALUES(1792,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1792,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(1792,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1792,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1792,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(1792,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1792,36,'cp 0','f3d1 a7a6 g1f3 g8e7 d2d3 h7h6 c1e3 f8g7 a2a4 d7d6 e1g1 g6g5 b1c3 c8g4 c3d5 e7d5 c4d5 c6e7 d5b7 a8b8 b7a6 d6d5 a6b5 c7c6 b5a6 b8a8 a6b7 a8b8');
INSERT INTO evaluations VALUES(19539,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(19539,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19539,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(19539,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(19539,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(19539,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19539,36,'cp 18','b1c3 g8f6 e2e4 e7e5 d4b5 d7d6 c1g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 f6g5 a3c2 c6e7 c2b4 e8g8 a2a4 e7d5 b4d5 c8e6 f1e2 a8b8 a4b5 a6b5 e1g1 d8c8 d5b4 c8b7 e2f3 f8d8 f1e1 g7g6 g2g3 b7c7 f3g4 e6c4');
INSERT INTO evaluations VALUES(19536,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(19536,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19536,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(19536,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19536,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(19536,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19536,36,'cp 26','g8f6 e2e4 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 c2c4 a7a6 b5c3 f8e7 f1d3 e8g8 e1g1 f7f5 f2f3 b8d7 a1b1 d7c5 d3c2 e7g5 b2b4 c5d7 c1g5 d8g5 f1f2 b7b6 d1d2 g5d2 f2d2 a8a7 c2d3 a7c7 d3f1 g7g6 a2a3 g8h8');
INSERT INTO evaluations VALUES(12878,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(12878,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12878,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(12878,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12878,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(12878,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(12878,36,'cp 24','e2e4 e7e5 d4b5 d7d6 c1g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 a8b8 a3c2 f6g5 a2a4 b5a4 c2e3 b8b2 d1a4 c8d7 f1a6 e8g8 a4a3 d8b8 e1g1 g5e3 d5e3 b2b3 a3c1 d7e6 c1d2 c6e7 d2d3 b8c7 a6c4 e6c4 e3c4 f8c8');
INSERT INTO evaluations VALUES(6609,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(6609,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(6609,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(6609,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(6609,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(6609,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(6609,36,'cp 100','c2c3 b4a5 b1d2 d7d6 d2c4 g8f6 f1e2 e8g8 e1g1 f8e8 c1g5 h7h6 g5h4 c8e6 c4a5 c6a5 d3d4 e5d4 c3d4 e6g4 f3d2 g4e2 d1e2 a5c6 e2d3 c6b4 d3c3 c7c5 f2f3 b4c6 d4d5 c6e5 f3f4 e5g6 h4f6 d8f6 c3f6 g7f6 f1f3 a8d8 h2h3 b7b5 a2a4 b5b4 a1e1 f6f5');
INSERT INTO evaluations VALUES(6608,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(6608,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(6608,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(6608,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(6608,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(6608,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(6608,36,'cp 76','b4e7 d3d4 e5d4 c3d4 d7d5 e4d5 d8d5 b1c3 d5h5 c1f4 g8f6 f4c7 c8e6 f1e2 a8c8 c7f4 h5a5 e1g1 e8g8 a2a3 f8d8 f4e3 f6d5 d1d2 a5c7 f1c1 d5c3 c1c3 e7f6 a1c1 h7h6 h2h3 c7d7 e2b5 a7a6 b5c4 e6c4 c3c4 d7e6 b2b4');
INSERT INTO evaluations VALUES(43448,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(43448,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(43448,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(43448,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(43448,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(43448,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(43448,36,'cp 77','d4f5 c8e6 d1d3 g7g6 f5e3 b8c6 c3d5 f6h5 a2a4 f8h6 e1g1 e8g8 c2c3 c6e7 d5e7 d8e7 c4e6 e7e6 f1d1 h6e3 c1e3 f8d8 c3c4 f7f5 a1c1 h5f6 f2f3 f5e4 f3e4 e6g4 c4c5 g4e4 c5d6 e4d3 d1d3 g8f7 a4a5 d8d7 c1c7');
INSERT INTO evaluations VALUES(43477,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(43477,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(43477,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(43477,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(43477,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(43477,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(43477,36,'cp 94','c8e6 c4b3 h7h6 e1g1 b8c6 f5g3 c6d4 b3e6 d4e6 c3d5 h6h5 d5f6 d8f6 g3f5 a8d8 c2c4 f8e7 g1h1 g7g6 f5e3 f6h4 f2f3 e7g5 e3d5 g5c1 a1c1 e8g8 d1b3 d8b8 b3b4 h4d8 b4c3 b8c8 c3d2 g8g7 f1d1 b7b5 c4b5 a6b5 c1c8');
INSERT INTO evaluations VALUES(55161,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(55161,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(55161,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55161,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(55161,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(55161,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(55161,36,'cp -12','d2d4 g8f6 c1f4 c8f5 e2e3 e7e6 f1d3 f5d3 d1d3 f8d6 g1f3 e8g8 e1g1 c7c5 d4c5 d6c5 c2c4 b8c6 b1c3 d5c4 d3c4 d8b6 f1d1 f8d8 c4b5 a7a6 b5b6 c5b6 g1f1 b6a5 f3e5 d8d1 a1d1 a5c3 b2c3 c6e5 f4e5 a8c8 e5f6 g7f6 d1d7 b7b5 a4b5 a6b5 d7b7 c8c3 b7b5 c3c1 f1e2');
INSERT INTO evaluations VALUES(55157,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(55157,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(55157,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(55157,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55157,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(55157,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(55157,36,'cp -18','c8f5 e2e3 g8f6 g1f3 e7e6 f1d3 f5g6 b2b3 f8e7 c2c4 c7c5 c4d5 f6d5 e1g1 c5d4 f3d4 b8d7 c1b2 e8g8 d3g6 h7g6 b1d2 e7b4 d4f3 d8e7 d2c4 a8c8 e3e4 d5b6 c4b6 d7b6 d1e2 f8d8 f1d1 f7f6 d1d8 c8d8 a1d1 e6e5 d1d8 e7d8');
INSERT INTO evaluations VALUES(16531,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(16531,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(16531,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(16531,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(16531,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(16531,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(16531,36,'cp -38','d7d5 d2d3 f8d6 f1e2 h7h6 b1d2 e8g8 e1g1 a7a5 b2b3 f8e8 f1e1 c6e7 e2f1 e7g6 e4d5 f6d5 d3d4 e5d4 e1e8 d8e8 f3d4 a5a4 b3b4 e8e5 d2f3 e5h5 a2a3 g6e5 f3e5 d6e5 h2h3 e5d4 c3d4 c8f5 c2c4 c7c6 b4b5 h5g6 b5c6 b7c6 c1d2 f5h3');
INSERT INTO evaluations VALUES(15142,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(15142,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(15142,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(15142,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(15142,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(15142,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(15142,36,'cp -40','d2d3 f8d6 f1e2 h7h6 b1d2 e8g8 e1g1 f8e8 f1e1 a7a5 b2b3 c6e7 e2f1 e7g6 e4d5 f6d5 d3d4 e5d4 e1e8 d8e8 f3d4 a5a4 b3b4 e8e5 d2f3 e5h5 a2a3 c7c5 b4c5 d6c5 h2h3 c8d7 c3c4 d5f6 c1b2 a8e8 a1e1 e8e1 f3e1 g6f4 e1d3 f4d3 f1d3 h5e5 c2c3 f6e4 c3e1 c5d4 b2d4 e5d4 e1e4');
INSERT INTO evaluations VALUES(15148,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(15148,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(15148,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(15148,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(15148,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(15148,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(15148,36,'cp -29','f8d6 f1e2 h7h6 e1g1 e8g8 f1e1 a7a5 b1d2 f8e8 b2b3 c6e7 e2f1 e7g6 e4d5 f6d5 d3d4 e5d4 e1e8 d8e8 f3d4 a5a4 b3b4 e8e5 d2f3 e5f6 a2a3 c8g4 h2h3 g4f3 d4f3 f6c3 c2c3 d5c3 c1b2 c3d5 g2g3 c7c6 b4b5 d6f8 a1d1 d5f6 b5c6 b7c6');
INSERT INTO evaluations VALUES(44897,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(44897,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(44897,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(44897,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(44897,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(44897,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(44897,36,'cp -34','d5c4 g2g3 g7g6 f1g2 f6d5 f4d2 f8g7 a2a4 c6c5 d4c5 e8g8 e1g1 b8a6 d1c1 a6c5 c1c4 b7b6 b1c3 c8b7 f1d1 a8c8 a4a5 b6a5 c4a2 e7e6 a2a3 d8b6 a3a5 c5b3 a5b6 a7b6 a1a7 b3d2 a7b7 d5c3 b2c3 d2e4 c3c4 e4c3 d1d2 f8d8 d2d8 c8d8 b7b6 c3e2 g1f1');
INSERT INTO evaluations VALUES(45019,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(45019,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(45019,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(45019,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(45019,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(45019,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(45019,36,'cp -28','g2g3 f6d5 f1g2 g7g6 f4d2 f8g7 a2a4 e8g8 e1g1 a7a5 b1a3 d5b6 d2c3 c8f5 f1e1 f5e4 g2f1 c6c5 d4c5 d8d1 e1d1 g7c3 b2c3 b6a4 a3c4 a4c5 c4a5 e4c2 d1e1 b8d7 a5c4 c2b3 f3d2 b3c4 d2c4 a8a4 a1a4 c5a4 e1b1');
INSERT INTO evaluations VALUES(55206,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55206,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55206,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(55206,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(55206,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(55206,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(55206,36,'cp -16','b8c6 e3d4 c6d4 f1e2 g8h6 f3d4 d8d4 d2d3 e7e5 e1g1 h6f5 c1d2 f8c5 d2c3 d4d6 b1d2 c5d4 e2f3 e8g8 f1e1 c7c6 f3e4 a7a5 a2a4 f7f6 d2b3 d4b6 c4c5 b6c5 b3a5 f5d4 a1b1 d6d8 a5b3 c5a7 b3d4 a7d4 d1b3 g8h8 h2h3 f6f5 e4f3');
INSERT INTO evaluations VALUES(21107,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(21107,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(21107,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(21107,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(21107,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(21107,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21107,36,'cp -6','e3d4 c6d4 f1e2 e7e5 e1g1 g8e7 f3e5 c8f5 e2d3 d8c8 f1e1 f7f6 e5f3 f5d3 f3d4 c8d7 d4e6 d3c4 d1h5 g7g6 e6f8 h8f8 h5h7 c4e6 h7h6 e8c8 h6e3 e6d5 b1c3 e7c6 c3d5 d7d5 d2d3 g6g5 e3e4 d5f7 c1e3 c8b8 e4f3 d8d3');
INSERT INTO evaluations VALUES(35121,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(35121,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(35121,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(35121,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(35121,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(35121,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(35121,36,'cp 124','d7d6 d2d4 f6g6 b1d2 b8c6 f1b5 a7a6 b5a4 g6g2 h1g1 g2h3 d4e5 c8d7 f3g5 h3h2 d2f3 h2h5 a4b3 g8h6 e5e6 f7e6 g5e6 d7e6 b3e6 c6e5 f3d4 h5d1 e1d1 g7g6 d1c2 h6f7 f2f4 f7d8 e6h3 e5c6 c1e3 f8g7 a1f1 e8g8 c2d3 a8b8 d4f3 g7h6 f3d2 d8f7 d3c2');
INSERT INTO evaluations VALUES(34783,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(34783,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(34783,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(34783,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(34783,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(34783,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(34783,36,'cp 130','d2d4 f6e7 f1c4 g7g6 e1g1 f8g7 f1e1 g8f6 a2a4 a7a5 d4e5 d6e5 b2b3 e8g8 b1d2 f8d8 f3g5 d8f8 c1a3 c7c5 g5f3 f8d8 d1e2 b8c6 a1d1 b7b6 d2f1 d8d1 e1d1 g7h6 h2h3 h6f4 f3d2 c8e6 c4e6 e7e6 d2c4 a8d8 d1d8 c6d8 f1e3');
INSERT INTO evaluations VALUES(34780,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(34780,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(34780,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(34780,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(34780,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(34780,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(34780,36,'cp 110','f6e7 f1c4 g7g6 e1g1 f8g7 a2a4 g8f6 f1e1 a7a5 d4e5 d6e5 b2b3 e8g8 c1a3 c7c5 b1d2 f8d8 d1c2 b8d7 f3g5 d8f8 a1d1 d7b6 c4b5 f6e8 g5f3 e8c7 b5e2 f8d8 d2f1 d8d1 e1d1 c8d7 f1e3 d7c6 f3d2 g7f8 a3b2 f8g7 d2c4 b6c4 e2c4 a8d8');
INSERT INTO evaluations VALUES(45919,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(45919,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(45919,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(45919,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(45919,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(45919,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(45919,36,'cp 48','f8e7 f1b5 c8d7 b5d7 d8d7 d1d3 b8c6 c1g5 a8d8 a1d1 e8g8 e1g1 f8e8 f1e1 h7h6 g5h4 a7a6 h2h3 b7b5 a2a3 d7c8 b2b3 g7g5 h4g3 e7f8 c3d5 f6d5 e4d5 e8e1 d1e1 c6e7 c2c4 b5c4 b3c4 e7f5 f3d2 f8g7 a3a4 c8d7 g3h2 f5d4 d2e4 d7a4 e4d6');
INSERT INTO evaluations VALUES(40151,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(40151,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(40151,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(40151,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(40151,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(40151,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(40151,36,'cp 60','c1f4 b8c6 f1c4 e8g8 e1g1 c8e6 c3d5 f6e4 f1e1 e6d5 c4d5 e4f6 d5c6 b7c6 e1e7 d8e7 f4d6 e7e4 d6f8 a8f8 b2b3 f8e8 d1d2 e4d5 a1d1 e8d8 c2c4 d5d2 d1d2 d8d2 f3d2 g7g6 f2f4 g8f8 d2f3 f6d7 g1f2 h7h6 f2g3 f7f5 f3e5 d7e5 f4e5 f8f7 g3f4');
INSERT INTO evaluations VALUES(53200,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(53200,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53200,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(53200,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(53200,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(53200,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53200,36,'cp 25','f1c4 e7e6 g1f3 a7a6 e1g1 g8f6 d4c5 f8c5 d1d8 e8d8 b2b3 b8d7 c1b2 d8e7 c4e2 b7b6 b1d2 c8b7 f3e1 b6b5 e1d3 c5d6 e2f3 b7f3 d2f3 h8c8 a2a4 a8b8 a4b5 a6b5 f1c1 d7c5 d3c5 c8c5 c1c5 d6c5 g1f1');
INSERT INTO evaluations VALUES(53191,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53191,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(53191,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(53191,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53191,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(53191,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(53191,36,'cp 32','g8f6 g1f3 a7a6 e1g1 e7e6 b2b3 b7b6 d4c5 d8d1 f1d1 f8c5 c1b2 b8d7 f3d2 c8b7 b1c3 e8g8 c4e2 f8d8 d2c4 c5e7 a1c1 f6d5 c3d5 b7d5 f2f3 a8c8 g1f2 f7f5 b2d4 b6b5 c4e5 d7e5 d4e5 e7a3 c1c7 c8c7');
INSERT INTO evaluations VALUES(45313,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(45313,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(45313,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(45313,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45313,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(45313,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(45313,36,'cp 32','g1f3 e7e6 e1g1 a7a6 b2b3 b7b6 c1b2 c8b7 b1d2 b8d7 a1c1 f8e7 d4c5 e7c5 d1e2 d8e7 c4d3 e8g8 f1d1 a8c8 d3a6 b7a6 e2a6 c8a8 a6d3 a8a2 c1c2 e7d6 d3d6 c5d6 d2c4 d6b4 f3e5 d7e5 c4e5');
INSERT INTO evaluations VALUES(45314,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(45314,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(45314,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(45314,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(45314,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(45314,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(45314,36,'cp 28','a7a6 e1g1 e7e6 d4c5 f8c5 d1d8 e8d8 c4e2 d8e7 b2b3 b7b6 b1d2 b8d7 f3e1 a6a5 c1b2 h8d8 d2c4 c8a6 a1c1 e7f8 e1f3 a6b7 f1d1 a5a4 f3d4 b7d5 d4b5 a4b3 a2b3 c5e7 b5c3 d5c4 e2c4 d7c5 g2g3 f6e4 c3e4 c5e4');
INSERT INTO evaluations VALUES(14164,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(14164,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14164,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14164,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14164,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14164,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14164,36,'cp -3','c8g4 g1e2 g4d7 e1g1 e7e6 b5d3 f8d6 h2h3 e8g8 e3e4 f6e4 d3e4 d5e4 c3e4 c6e7 c2c4 d7c6 e4d6 d8d6 f1e1 e7f5 b2b3 a8d8 c1b2 f8e8 d1d3 f7f6 a1d1 f5h4 f2f3 b7b6 d3e3 h4f5 e3f2 h7h6 b2c3');
INSERT INTO evaluations VALUES(28761,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(28761,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(28761,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(28761,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(28761,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(28761,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(28761,36,'cp 142','e4e5 f6e4 d1e2 f7f5 b1d2 e4g5 h1g1 c7c5 d4c5 b8c6 f3g5 d8g5 c5b6 a7b6 c2c3 g5d8 f2f4 f8c5 g1f1 d7d5 e5d6 e8g8 d2f3 c5d6 c1e3 d8e7 d3b5 f8c8 a2a4 c6a5 e1f2 b7d5 g2g3 d6c5 e3c5 e7c5 f3d4 a5b3 a1d1');
INSERT INTO evaluations VALUES(28688,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(28688,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(28688,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(28688,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(28688,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(28688,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(28688,36,'cp 135','f6e4 d1e2 f7f5 b1d2 e4g5 h1g1 c7c5 d4c5 b8c6 f3g5 d8g5 c5b6 a7b6 c2c3 f8e7 d2f3 g5g4 h2h3 g4a4 g2g4 f5f4 d3b5 a4a5 a2a4 e8g8 c1d2 a5a7 g4g5 c6a5 a1d1 a5b3 g5g6 f8f5 g6h7 g8h8 g1g4 b3d2 d1d2 b7f3');
INSERT INTO evaluations VALUES(31874,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(31874,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31874,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31874,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(31874,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(31874,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(31874,36,'cp 0','g2g3 b8c6 d2d3 h7h5 f1g2 e7e6 c1e3 f8e7 e3f2 h5h4 f3h4 e7h4 g3h4 g8e7 h4h5 f5g4 d1d2 g4h5 b1c3 e7g6 h1g1 c6e7 e1c1 d8d6 c3b5 d6d7 b5c3 d7d6');
INSERT INTO evaluations VALUES(32009,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(32009,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(32009,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(32009,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(32009,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(32009,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(32009,36,'cp 48','b8c6 d2d4 e7e6 c1e3 g8f6 h2h3 g4f5 a2a3 f8d6 f3h4 c6e7 c3b5 e8g8 e1g1 e7d5 b5d6 d8d6 e3c1 d5f4 e2f3 c7c6 a3a4 f4g6 h4g6 f5g6 a4a5 h7h6 a5a6 b7b6 f1e1 f8e8 c1d2 a8d8 c2c3 g6d3 g2g3');
INSERT INTO evaluations VALUES(38224,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38224,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38224,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(38224,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38224,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38224,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(38224,36,'cp 17','g8f6 b1c3 e8g8 e1g1 c7c5 a2a4 b8c6 d2d3 c8e6 f1e1 a7a6 c1g5 c6b4 f3h2 f6d7 g5c1 a8b8 h2f1 g8h8 c1d2 e7g5 a4a5 g5d2 d1d2 d7f6 f1e3 b4c6 c3e2 e6c4 e3c4 d6d5 e4d5 f6d5 e2c3 f7f6 c3d5 d8d5 d2c3');
INSERT INTO evaluations VALUES(40777,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40777,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40777,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40777,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40777,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40777,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40777,36,'cp 21','b1c3 e8g8 e1g1 c7c5 a2a4 b8c6 d2d3 c8e6 f1e1 a7a6 c1g5 c6b4 d1e2 a8b8 g5f6 e7f6 c3d1 f6g5 d1e3 g5e3 f2e3 e6c4 d3c4 d8e7 e1f1 g7g6 f1f2 b8d8 a1f1 d8d7 a4a5 b4c6 f3h2 h7h5 e2d2 g8g7 b2b3 e7d8');
INSERT INTO evaluations VALUES(38552,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(38552,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(38552,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(38552,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(38552,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(38552,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(38552,36,'cp 62','b1c3 b8c6 c1e3 g8f6 e1g1 e8g8 f1e1 b7b6 d1d2 e6e5 a1d1 c8g4 d4e5 d6e5 c4b5 g4f3 g2f3 d8d2 d1d2 c6d4 e3d4 e5d4 d2d4 f6g4 d4d1 g7c3 b2c3 g4e5 g1g2 g6g5 h2h4 f7f6 h4h5 g8f7 g2g3 f8g8 d1d4 c7c5 d4d5 f7e7 e1d1 a8d8 a2a4 d8d5 d1d5');
INSERT INTO evaluations VALUES(38549,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(38549,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(38549,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(38549,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(38549,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(38549,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(38549,36,'cp 77','b8c6 e1g1 g8e7 a2a4 e8g8 h2h3 b7b6 f1e1 h7h6 c1f4 a7a6 c3e2 g8h7 f4h2 c8b7 c2c3 e6e5 d1c2 e5d4 e2d4 c6e5 f3e5 d6e5 d4f3 f7f6 b2b4 d8e8 c4d3 e8c6 f3d2 a8d8 a4a5 b6b5');
INSERT INTO evaluations VALUES(42577,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(42577,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(42577,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(42577,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(42577,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(42577,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(42577,36,'cp -35','f6g4 d2d4 e5d4 c3a4 c5b4 c1d2 d8e7 f1e2 g4f6 e1g1 b4d2 d1d2 d7d6 f3d4 f6e4 d2c2 e8g8 f1e1 e7h4 e2f1 e4c5 a4c5 h4d4 c5e4 c8f5 a1d1 d4e5 f1d3 b8c6 e4c3 f5d3 c2d3 e5c5 d3e4 c6e5 e4b7 a8b8 b7d5 c5d5 c3d5 b8b2 c4c5 f7f6 c5d6 c7d6 d5e7 g8f7 e7f5 f7g6 f5e7 g6h5 d1d5 b2a2 d5d6');
INSERT INTO evaluations VALUES(42874,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(42874,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(42874,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(42874,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(42874,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(42874,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(42874,36,'cp -29','d2d4 e5d4 c3a4 c5b4 c1d2 d8e7 f1e2 g4f6 e1g1 b4d2 d1d2 d7d6 f3d4 f6e4 d2c2 e8g8 f1e1 e7h4 e2f1 e4c5 a4c5 h4d4 c5e4 c8f5 a1d1 d4e5 f1d3 b8c6 c2c1 f5g6 c4c5 a8d8 c5d6 c7d6 c1d2 e5d4 e4c3 g6d3 d2d3 d4d3 d1d3 d8d7 e1d1 f8d8 c3e4 h7h6 d3d6 d7d6 e4d6 c6b4 a2a3 b4c6 g1h1');
INSERT INTO evaluations VALUES(42877,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(42877,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(42877,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(42877,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(42877,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(42877,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(42877,36,'cp -50','e5d4 c3a4 c5b4 c1d2 d8e7 f1e2 g4f6 e1g1 b4d2 d1d2 d7d6 f3d4 f6e4 d2c2 e8g8 f1e1 e4f6 c4c5 d6d5 e2b5 e7d8 c5c6 b7c6 b5d3 d8d6 h2h3 g7g6 a2a3 f8e8 e1e8 f6e8 a1e1 e8g7 b2b4 a7a5 b4b5 g7e6 d4e6 c8e6');
INSERT INTO evaluations VALUES(42882,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(42882,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(42882,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(42882,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(42882,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(42882,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(42882,36,'cp -24','c3a4 c5b4 c1d2 d8e7 f1e2 g4f6 e1g1 b4d2 d1d2 d7d6 f3d4 f6e4 d2c2 e8g8 f1e1 e7h4 e2f1 e4c5 a4c5 h4d4 c5e4 c8f5 a1d1 d4e5 f1d3 b8c6 c2c1 e5e7 c4c5 d6d5 e4c3 e7c5 d3f5 d5d4 c1g5 g7g6 c3e2 c5f5 g5f5 g6f5 e2d4 c6d4 d1d4 f8e8 e1e8 a8e8 g2g3 e8e2 d4d7 e2b2 d7c7 g8g7');
INSERT INTO evaluations VALUES(52712,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(52712,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52712,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52712,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(52712,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52712,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(52712,36,'cp -13','c6d5 d3d4 c8f5 c2c3 b8c6 g1f3 e7e6 c1f4 f8d6 f4d6 d8d6 f1d3 g8f6 e1g1 e8g8 d1c2 f5d3 c2d3 f8c8 d3e2 d6c7 b1d2 a7a5 a2a4 c7b6 d2b1 b6a6 f1e1 a6e2 e1e2 g8f8 b1a3 f6e4 a3b5 f8e7 f3e1 e4d6 e1d3 d6c4 g2g3');
INSERT INTO evaluations VALUES(53422,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(53422,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(53422,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(53422,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(53422,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(53422,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53422,36,'cp -4','d3d4 c8f5 c2c3 e7e6 c1f4 b8c6 g1f3 f8d6 f4d6 d8d6 f1d3 g8e7 e1g1 e8g8 f1e1 f5d3 d1d3 e7g6 g2g3 f7f6 c3c4 d5c4 d3c4 d6d5 c4d5 e6d5 e1e3 g6e7 e3b3 b7b6 b1c3 f8d8 a1d1 g8f7 b3b5 c6a5 f3d2 a8c8 d2f1');
INSERT INTO evaluations VALUES(53389,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(53389,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(53389,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(53389,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(53389,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(53389,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(53389,36,'cp -13','c8f5 c2c3 e7e6 c1f4 b8c6 g1f3 g8e7 f3h4 e7g6 h4f5 g6f4 f5e3 d8b6 d1b3 b6c7 b1d2 f8d6 g2g3 f4h5 f1d3 h5f6 e1g1 e8g8 d2f3 g7g6 b3c2 g8g7 a1e1 a8e8 c2d2 a7a6 e1e2 b7b5 f1e1 f8h8 d2c2 h7h6 a2a4 b5a4');
INSERT INTO evaluations VALUES(43513,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43513,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43513,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43513,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43513,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43513,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43513,36,'cp 0','f8e7 d1f3 b8d7 f3g3 d8b6 d4b3 b6c7 a2a4 b7b6 e1g1 c8b7 a1d1 e8g8 f2f4 f8e8 g1h1 a8c8 d1e1 f6h5 g3f2 h5f6');
INSERT INTO evaluations VALUES(39712,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39712,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39712,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39712,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39712,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39712,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39712,36,'cp 0','d1f3 b8d7 f3g3 d8b6 d4b3 b6c7 a2a4 f6h5 g3h4 h5f6 e1g1 d7e5 h4g3 f6h5 g3e3 h5f6');
INSERT INTO evaluations VALUES(39710,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39710,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39710,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39710,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39710,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39710,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39710,36,'cp 0','b8d7 f3g3 d8b6 d4b3 b6c7 a2a4 f6h5 g3h4 h5f6 e1g1 d7e5 h4g3 f6h5 g3e3 h5f6 e3g3');
INSERT INTO evaluations VALUES(48305,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(48305,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(48305,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(48305,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(48305,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(48305,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48305,36,'cp -25','c7c5 c2c3 f8d6 g1f3 e8g8 f3e5 f6e4 e1g1 b7b6 b1d2 c8b7 a2a4 f7f6 e5c4 d6e7 d2e4 d5e4 d3e2 b7d5 c4a3 f6f5 b2b3 a7a6 c1b2 c5d4 e3d4 b8d7 a3c4 d8c7 c4e3 d7f6 c3c4 d5b7 a1c1 a8d8 b2c3 b7c6');
INSERT INTO evaluations VALUES(46220,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(46220,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(46220,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(46220,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(46220,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(46220,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(46220,36,'cp -27','c2c3 f8e7 g1f3 e8g8 e1g1 b7b6 d1e2 f6e4 b2b3 b8d7 c1b2 c8b7 b1d2 d7f6 a1d1 a8c8 c3c4 e4d2 d1d2 c5d4 f3d4 e7c5 d4b5 c5b4 d2d1 a7a6 a2a3 b4c5 b5d4 a6a5 d4b5 d8e7 a3a4 d5c4 d3c4 f8d8 b2e5 d8d1 f1d1 c8d8');
INSERT INTO evaluations VALUES(15697,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(15697,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(15697,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(15697,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(15697,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(15697,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(15697,36,'cp -26','f6e4 d2d3 e4d6 f3e5 f8e7 f1e1 e8g8 h2h3 a7a5 a2a4 c8e6 c1d2 d6f5 b1c3 f8e8 d1h5 f5d4 h5d1 c6c5 c3b5 d4f5 e5c4 c7c6 b5c3 e7f8 c3e4 b7b5 a4b5 c6b5 d2a5 d8b8 c4d2 f5d4');
INSERT INTO evaluations VALUES(15715,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(15715,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(15715,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(15715,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(15715,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(15715,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(15715,36,'cp -32','d2d3 e4d6 f3e5 f8e7 c1f4 e8g8 b1c3 f7f6 e5g4 d6f5 f1e1 f8f7 a2a4 e7f8 h2h3 c6c5 a4a5 c8d7 g4e3 f5d4 c3e2 d4e6 f4g3 d7c6 c2c3 f6f5 d1b3 d8d7 e2f4 a8e8 e3c4 e6f4 g3f4 d7d5 f2f3 f7e7 e1e7 e8e7 g1f2 b7b5 c4d2 h7h6 b3d5 c6d5');
INSERT INTO evaluations VALUES(16721,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(16721,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16721,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16721,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(16721,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(16721,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(16721,36,'cp -9','f1c4 f8b4 e4e5 f6g4 d2d4 d7d6 c1f4 d6e5 d4e5 c8e6 c4b5 d8d1 a1d1 e6d7 b5c6 d7c6 d1d2 f7f6 e5e6 b4c5 f3d4 c5d4 d2d4 a8d8 d4d8 e8d8 e1g1 d8e7 f1e1 h8d8 h2h3 g4e5 f4e5 f6e5 e1e5 d8d2 e5e2 d2e2 c3e2 c6d5 e2f4 d5e6 f4e6 e7e6 g1f2 e6f5 f2f3');
INSERT INTO evaluations VALUES(51779,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(51779,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(51779,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(51779,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(51779,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(51779,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(51779,36,'cp 51','b1c3 d7d5 d2d4 d5e4 c3e4 c6c5 e4c5 b8c6 c1e3 e6e5 d4e5 d8a5 d1d2 c6b4 c5e4 c8f5 a1d1 b4c2 e1e2 a5e5 d2c2 f5e4 c2c3 e5c3 b2c3 g8e7 g1f3 e7f5 f3d4 f5e3 e2e3 e4c6 d4c6 b7c6 e3f3 f8c5 f1d3 e8g8 h1e1 a8e8 d3f5 h7h5 f5d7 e8e1 d1e1 f8d8 e1e8 d8e8 d7e8');
INSERT INTO evaluations VALUES(51777,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(51777,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(51777,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(51777,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(51777,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(51777,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(51777,36,'cp 45','d7d5 d2d4 d5e4 c3e4 c6c5 e4c5 b8c6 c1e3 e6e5 d4e5 d8a5 d1d2 c6b4 c5e4 c8f5 a1d1 b4c2 e1e2 a5e5 d2c2 f5e4 c2c3 e5c3 b2c3 g8e7 g1f3 e7f5 e3f4 f8c5 f3e5 b7b6 e5d3 c5d6 f4d6 f5d6 f2f3 e4g6 e2f2 e8c8 f1e2 d6c4 d3b4 c4a5 e2b5 d8d1 h1d1 h8d8');
INSERT INTO evaluations VALUES(53558,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(53558,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(53558,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(53558,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(53558,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(53558,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(53558,36,'cp 36','d2d4 d5e4 c3e4 c6c5 g1f3 c5d4 d1d4 f8b4 e4c3 d8d4 f3d4 c8d7 f1e2 g8f6 e1g1 b8c6 c1e3 e8g8 f1d1 b4c3 b2c3 b7b6 d4b3 f8c8 c4c5 f6d5 e3d2 h7h6 c3c4 d5f6 d2e3 d7e8 g1f1 f6e4 d1c1 e4c5 b3c5 b6c5 e3c5 c8d8 f2f3 c6d4');
INSERT INTO evaluations VALUES(14488,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14488,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(14488,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(14488,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(14488,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(14488,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(14488,36,'cp 75','d2d4 e7e5 f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 h2h3 d8d7 c1e3 f8b8 e4e5 f6d5 c3d5 c6d5 f3d5 d7e6 d5e4 d6e5 f4e5 e7d6 e3d4 f7f6 a1e1 f6e5 b2b3 a5a4 d4c3 a4b3 a2b3 b8e8 e1e2 a8d8 g1h2 d6f8 f1f5 f8d6 c3e5');
INSERT INTO evaluations VALUES(50103,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(50103,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(50103,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(50103,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50103,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(50103,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(50103,36,'cp 35','e7e6 f1e2 b7b6 c2c4 f8b4 b1d2 b4e7 e1g1 c8b7 d1c2 c7c5 b2b3 e8g8 c1b2 c5d4 f3d4 b8c6 d4c6 b7c6 e2f3 d8c7 h2h3 a8c8 a1c1 c6f3 d2f3 f6e4 c2e2 e7f6 f1d1 f8e8 b2f6 e4f6 e2b2 h7h6 f3e5 g8h8 e5g6 h8g8 f2f3 d7d5');
INSERT INTO evaluations VALUES(49962,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(49962,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(49962,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(49962,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(49962,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(49962,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(49962,36,'cp 56','f1e2 a7a5 c2c4 g7g6 b1c3 f8g7 b2b3 e8g8 c1a3 d7d6 a1c1 b8a6 e1g1 d8e7 c4c5 a6b4 c5d6 c7d6 f3d2 f6d5 d2b1 c8d7 e2f3 d7c6 c3d5 c6d5 f3d5 b4d5 b1c3 d5c3 c1c3 e7d7');
INSERT INTO evaluations VALUES(55428,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(55428,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55428,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(55428,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(55428,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(55428,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(55428,36,'cp 49','g8f6 b1c3 b8c6 d4d3 f8e7 c1f4 f6d7 d3e3 e7f6 e1c1 e8g8 g2g4 f6c3 e3c3 d8f6 c3e3 d7e5 f4e5 c6e5 f3e5 d6e5 h1g1 c8e6 g4g5 h6g5 c4e6 f7e6 g1g5 a8d8 d1g1 d8d7 g1g2 f6f4 b2b3 f4e3 f2e3 f8f3 g5e5 g8f7 e5b5 b7b6 b5g5 f7f8 g5g6 d7f7 g6e6 f3e3 c1d2 e3f3 d2e2 f3h3');
INSERT INTO evaluations VALUES(47647,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(47647,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(47647,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(47647,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(47647,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47647,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47647,36,'cp 42','b1c3 b8c6 d4d3 f8e7 c1f4 f6d7 c4b5 e8g8 e1c1 f8e8 h1e1 e7f8 e4e5 d7e5 f3e5 d6e5 b5c6 d8d3 d1d3 b7c6 e1e5 c8e6 d3e3 f8d6 e5e4 f7f5 f4d6 f5e4 d6c7 e8e7 c7d6 e7e8 c3e4 e6a2');
INSERT INTO evaluations VALUES(47646,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47646,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(47646,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(47646,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(47646,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(47646,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47646,36,'cp 30','b8c6 d4d3 f8e7 c1f4 f6d7 e1c1 e8g8 d3e3 e7f6 g2g4 f6c3 e3c3 d8f6 c3e3 d7e5 f4e5 c6e5 f3e5 d6e5 h1g1 f6g5 e3g5 h6g5 a2a4 f8e8 f2f3 c8e6 c4e6 e8e6 d1d7 a8c8 g1d1 g8f8 a4a5 e6e7 d7d5 f7f6 a5a6 b7a6 d5a5 f8f7');
INSERT INTO evaluations VALUES(14660,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(14660,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(14660,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14660,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(14660,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(14660,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(14660,36,'cp 42','d4d3 f8e7 c1f4 e8g8 e1c1 f6d7 d3e3 e7f6 g2g4 f6c3 e3c3 d8f6 c3e3 d7e5 f4e5 c6e5 f3e5 d6e5 h1g1 f6g5 d1d3 g5e3 d3e3 g7g5 b2b3 c7c6 a2a4 g8g7 e3d3 a8b8 f2f3 b7b5 a4b5 c6b5 c4d5 a7a5 d3c3 b8b6 c3c7 a5a4 b3a4 b5a4 c7a7 c8e6 d5e6');
INSERT INTO evaluations VALUES(11019,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(11019,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(11019,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(11019,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(11019,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(11019,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(11019,36,'cp 75','c2c3 e7g6 d2d4 c5b6 a2a4 a7a6 b5c6 d7c6 f3e5 g6e5 a4a5 b6a7 d4e5 d8e7 d1f3 c8d7 c1e3 a7e3 f3e3 e7e5 f2f4 e5e7 b1d2 e8g8 h2h3 f7f6 a1e1 c6c5 e4e5 a8d8 e5f6 e7e3 e1e3 f8f6 d2e4 f6e6 f1d1 b7b6 g1f2 c5c4 g2g4 b6a5 f4f5 e6e5 e4c5');
INSERT INTO evaluations VALUES(11015,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(11015,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(11015,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(11015,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(11015,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(11015,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(11015,36,'cp 75','e7g6 d2d4 c5b6 a2a4 a7a6 b5c6 d7c6 a4a5 b6a7 f3e5 g6e5 d4e5 d8e7 d1f3 c8d7 c1e3 a7e3 f3e3 e7e5 f2f4 e5e7 b1d2 e8g8 f4f5 f7f6 e3g3 d7e8 a1e1 a8d8 d2f3 d8d3 g3f2 c6c5 e4e5 e8c6 f2h4 d3d7 h4f4 c6f3 f4f3 f6e5 f3b7 e7d6 b7f3');
INSERT INTO evaluations VALUES(16592,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(16592,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(16592,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(16592,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(16592,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(16592,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(16592,36,'cp -25','e5f4 f1c4 f8b4 e4e5 f6h5 e1g1 e8g8 c3d5 d7d6 c2c3 c8e6 c4b3 b4a5 d2d4 d6e5 f3e5 c6e5 d1h5 e6d5 h5e5 d5b3 a2b3 a5b6 e5f5 d8c8 f5f4 c8d7 f4f5 d7f5 f1f5 f7f6 g1f2 f8e8 f5f3 c7c6 f3e3 a7a6 c1d2 g8f7 f2e2 b6c7 e2d3 h7h5 g2g3 h5h4');
INSERT INTO evaluations VALUES(29091,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(29091,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(29091,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(29091,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(29091,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(29091,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(29091,36,'cp 49','h2h4 f6h5 f1d3 h5g3 f2g3 e7e6 c1f4 f8d6 d3g6 h7g6 d1d2 d6f4 d2f4 d8a5 c2c3 b8d7 g3g4 c6c5 f3e5 d7e5 f4e5 e8c8 g4g5 b7b6 e1g1 c5d4 e5a5 b6a5 f1f7 d4d3 f7g7 h8h4 g7a7 h4e4 a7a5 e4e2 a5a8 c8c7 a8d8 c7d8 a1d1 e2b2 d1d3 d8c7 d3d4 b2a2');
INSERT INTO evaluations VALUES(29089,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(29089,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(29089,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(29089,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(29089,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(29089,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(29089,36,'cp 36','f6h5 f1d3 h5g3 f2g3 b8d7 c1g5 d7f6 d3g6 h7g6 e1g1 e7e6 f3e5 f8e7 c2c3 e8g8 d1e2 d8d5 a1d1 a8e8 g1h2 d5d6 g5f4 d6d8 f1f3 f6d7 e5d3 d7f6 a2a4 e7d6 f4d6 d8d6 d3e5 d6c7 d1f1 c6c5 d4c5 c7c5');
INSERT INTO evaluations VALUES(43270,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(43270,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(43270,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(43270,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(43270,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(43270,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(43270,36,'cp 164','e4e5 f6g8 d2d4 d7d6 d1e2 b8c6 h2h3 g8h6 c1f4 d6e5 d4e5 c6d4 e2e4 d4f3 e4f3 h6f5 f1b5 c7c6 a1d1 c8d7 b5c4 d8b6 e1g1 d7e6 c4e6 f7e6 c3e4 b6b2 f3d3 e8g8 d3c4 g8h8 c2c3 h7h6 c4e6 g6g5 f4c1 b2b5 f2f4 g5f4 c1f4 a8d8 c3c4 b5b6 c4c5 b6c7 e6g6 d8d1');
INSERT INTO evaluations VALUES(43125,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(43125,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(43125,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(43125,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(43125,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(43125,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(43125,36,'cp 154','f6g8 d2d4 d7d6 d1e2 b8c6 h2h3 g8h6 g2g4 h6g4 h3g4 c8g4 c1e3 d6e5 d4e5 c6e5 f1g2 h7h5 e3f4 e5f3 g2f3 g4f3 e2f3 c7c6 e1g1 d8d7 f1e1 d7g4 f3g4 h5g4 a1d1 g7c3 b2c3 a8d8 d1d8 e8d8 f4e3 b7b6 c3c4 h8h5 a2a4 d8c7 c4c5 b6c5 e3f4 c7d7 g1g2 f7f6 g2g3');
INSERT INTO evaluations VALUES(51628,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(51628,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(51628,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(51628,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(51628,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(51628,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(51628,36,'cp 277','d4c5 e7e6 f1b5 c8d7 b5d7 d8d7 c5b6 a7b6 d1e2 g8f6 c2c4 f8e7 e1g1 e8g8 b1c3 b8c6 f1d1 f8d8 c1f4 d7c7 h2h3 c6e5 b2b3 e5g6 f4g3 f6h5 g3h2 c7c5 c3b5 c5b4 a1b1 b4c5 e2d2 h5f6 b5d6 e7d6 h2d6 f6e4 d6c5');
INSERT INTO evaluations VALUES(51624,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(51624,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(51624,32,'cp 261',NULL);
INSERT INTO evaluations VALUES(51624,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(51624,34,'cp 266',NULL);
INSERT INTO evaluations VALUES(51624,35,'cp 259',NULL);
INSERT INTO evaluations VALUES(51624,36,'cp 262','e7e6 f1b5 c8d7 b5d7 d8d7 c5b6 a7b6 d1e2 g8f6 c2c4 f8e7 e1g1 b8c6 b1c3 e8g8 c1f4 f8d8 a1d1 d7b7 h2h3 h7h6 f1e1 d8d7 a2a3 a8d8 f4g3 f6h5 g3h2 h5f6 c3b5 e7f8 h2f4 b7a8 b2b3 c6a5 b3b4 a5c6 f4e3 a8b7 e3c1 c6e5 c1b2 f8e7 f3e5');
INSERT INTO evaluations VALUES(19812,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(19812,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(19812,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(19812,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(19812,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(19812,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(19812,36,'cp -29','e7e5 d2d3 f8e7 e2e4 d5d4 b1d2 c8e6 d2c4 d8c7 c1d2 h7h5 c2c3 h5h4 c3d4 e6c4 d3c4 h4h3 g2h1 c5d4 f3e1 d4d3 a1c1 g8f6 e1d3 e8g8 a2a3 a8d8 d2e3 c7d7 c1c3 d7e6 c4c5 a7a5 d1e2 c6d4 e2d1 a5a4 e3d4');
INSERT INTO evaluations VALUES(20547,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(20547,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(20547,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(20547,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(20547,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(20547,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(20547,36,'cp -10','d2d3 f8e7 e2e4 d5d4 b1a3 c8e6 a3c4 d8c7 c1d2 h7h5 c2c3 h5h4 c3d4 e6c4 d3c4 c6d4 f3d4 c5d4 d1a4 e8f8 a1c1 g8f6 c4c5 h4h3 g2f3 c7c6 a4c6 b7c6 c1c4 a7a5 c4a4 g7g6 d2a5 e7c5 b2b4 c5e7 f1c1 g6g5 c1c6 f8g7 b4b5');
INSERT INTO evaluations VALUES(20545,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(20545,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(20545,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(20545,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(20545,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(20545,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(20545,36,'cp -39','f8e7 e2e4 d5d4 b1d2 c8e6 d2c4 d8c7 c1d2 g8f6 f3g5 e6g4 f2f3 g4h5 g3g4 h7h6 g5h3 h5g6 f3f4 f6d7 f4e5 d7e5 b2b3 c7d7 a2a4 e7h4 d2e1 h4e1 d1e1 e5c4 b3c4 d7g4 h3f4 e8g8 a4a5 a8e8 a1b1 c6b4 e1f2 g6h5 f2g3 g4g3 h2g3 e8e5 f4d5 b4c2 b1b7 c2e3');
INSERT INTO evaluations VALUES(33149,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(33149,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(33149,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(33149,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(33149,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(33149,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(33149,36,'cp 52','c2c3 d7d6 d3d4 c5b6 a2a4 a7a6 a4a5 b6a7 e1g1 g8e7 c1e3 e8g8 f1e1 e7g6 c4f1 c8g4 b1d2 g6h4 f1e2 h4f3 d2f3 e5d4 f3d4 g4e2 e1e2 b8c6 d4f5 a7e3 f5e3 a8e8 d1b3 e8b8 e3d5 f6e6 h2h3 c6e7 a1e1 e7d5 e4d5 e6d7');
INSERT INTO evaluations VALUES(45135,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(45135,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(45135,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(45135,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(45135,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(45135,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(45135,36,'cp -30','e7e6 c4b3 b8c6 g1e2 f8e7 e1g1 e8g8 a2a4 g8h8 b1c3 a7a6 h2h3 a8b8 c1f4 c8d7 f4h2 h8g8 f3e3 b7b5 a4b5 a6b5 c3b1 d8b6 b1d2 b6b7 b3a2 b8a8 c2c3 a8a4 a2b3 a4a1 f1a1 f8a8 a1a8 b7a8 g1h1 b5b4');
INSERT INTO evaluations VALUES(51413,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(51413,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(51413,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51413,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(51413,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(51413,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(51413,36,'cp 38','e2e4 c8b7 f1d3 c7c5 c2c3 f7f5 f2f3 g8f6 e4f5 b8c6 f4g3 c5d4 g1e2 f6d5 f5e6 d7e6 e2d4 c6d4 c3d4 f8b4 b1d2 e8g8 d1e2 d5f4 g3f4 f8f4 e1c1 a8c8 c1b1 f4d4 d2b3 d4d6 a2a3 b4c5 d3c2 b7d5 h2h4 d8f6 h4h5 c8d8 b3c5 b6c5 h1e1 h7h6 e2e5 f6e5 e1e5');
INSERT INTO evaluations VALUES(55138,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(55138,31,'cp -102',NULL);
INSERT INTO evaluations VALUES(55138,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(55138,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(55138,34,'cp -130',NULL);
INSERT INTO evaluations VALUES(55138,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(55138,36,'cp -100','g8f6 f3d1 e7e5 d2d4 b8d7 g1f3 e5e4 f3d2 f8d6 c2c4 c7c6 b1c3 e8g8 a2a4 d6c7 c4d5 c6d5 b2b4 f8e8 f1b5 h7h5 b5e2 h5h4 h2h3 d7f8 d2b3 f6h7 e2b5 e8e6 d1h5 a7a6 b5e2 h7g5 b3c5 e6g6 a4a5 f8e6 h5h4 e6c5 b4c5 c7a5');
INSERT INTO evaluations VALUES(47282,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(47282,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(47282,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(47282,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(47282,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(47282,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(47282,36,'cp -101','f3d1 e7e5 d2d4 b8d7 g1f3 e5e4 f3d2 c7c6 a2a4 f8b4 c2c4 e8g8 b1c3 a7a5 f1e2 d5c4 d2c4 d7b6 c4b6 d8b6 e1g1 c8f5 e2c4 b6c7 c3e2 c7d7 e2g3 f5g6 c1d2 b4d6 f2f4 e4f3 d1f3 d6g3 f3g3 f8e8');
INSERT INTO evaluations VALUES(57288,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(57288,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(57288,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(57288,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(57288,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(57288,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(57288,36,'cp -43','d7d5 f2f3 g8f6 f3e4 f6e4 c3e4 d5e4 g1e2 f8d6 c2c4 c7c6 e2f4 b8d7 f1e2 d8h4 g2g3 h4e7 c1d2 d6f4 e3f4 e8g8 d1b3 f8e8 d2b4 e7d8 e1c1 a7a5 b4c3 d7f6 h1e1 e4e3 b3c2 b7b5 d4d5 c6d5 c4b5 f6e4 c2d3 c8b7 c1b1 a5a4');
INSERT INTO evaluations VALUES(56225,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(56225,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(56225,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(56225,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(56225,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(56225,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(56225,36,'cp -40','f2f3 g8f6 f3e4 f6e4 c3e4 d5e4 g1e2 f8d6 c2c4 c7c6 d1c2 e8g8 c1d2 b8d7 e1c1 f7f5 c1b1 d7f6 h2h3 d8e8 e2c3 a8b8 c4c5 d6c7 f1e2 c7g3 d2e1 g3e1 h1e1 c8e6 e1g1 e8e7 g2g4 f5g4 h3g4 f6d5 c3d5 e6d5');
INSERT INTO evaluations VALUES(56226,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(56226,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(56226,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(56226,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(56226,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(56226,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(56226,36,'cp -35','g8f6 f3e4 f6e4 c3e4 d5e4 g1e2 f8d6 c2c4 c7c6 e2f4 b8d7 f1e2 d8h4 g2g3 h4e7 c1d2 d6f4 e3f4 e8g8 e1g1 d7f6 f4f5 e4e3 d2c3 f8e8 f1f3 f6e4 f3e3 c8f5 e2f3 e7d7 d4d5 e4c3 e3c3 a8d8 d1d4 c6d5 c4d5 h7h5 d4a7');
INSERT INTO evaluations VALUES(48700,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48700,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(48700,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(48700,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(48700,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(48700,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(48700,36,'cp -48','f3e4 f6e4 f1d3 f7f5 d3e4 d5e4 d1h5 g7g6 h5e2 b8d7 b2b3 f8d6 c3b5 d6b4 c2c3 b4a5 b3b4 a5b6 c3c4 a7a6 b5c3 c7c6 c1b2 e8g8 e1c1 d7f6 g1h3 d8e7 c4c5 b6c7 d4d5 c6d5 c3d5 f6d5 d1d5 c8e6 d5d2 e6f7 h1d1 c7e5 b2e5 e7e5');
INSERT INTO evaluations VALUES(1949,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(1949,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(1949,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(1949,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(1949,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(1949,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(1949,36,'cp 45','f6b6 d4b6 a7b6 c1f4 e8d8 a2a3 g8e7 b1c3 e7g6 f4g3 f8b4 e1d2 h7h5 h2h4 f7f6 e5f6 g7f6 f1d3 b4c3 d2c3 g6e5 d3e2 d7d6 b2b3 c8g4 e2f1 g4d7 a3a4 d7c6 a1d1 d8d7 f2f3 h8g8 g3f2 g8g7 d1e1 a8g8 f2d4 g7g3 c3b2 g3g7 e1d1 d7e6 d4c3');
INSERT INTO evaluations VALUES(1817,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(1817,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(1817,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(1817,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(1817,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(1817,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(1817,36,'cp 52','d4b6 a7b6 c1f4 e8d8 a2a3 f7f6 b1c3 f6e5 f4e5 d7d6 e5d4 g8e7 c3e4 c8f5 e4g5 e7c6 d4c3 d8d7 e1c1 f8e7 g5f3 e7f6 c3f6 g7f6 f3h4 f5g4 f2f3 g4e6 f3f4 c6e7 f1d3 c7c6 d1d2 b6b5 g2g3 e6g4 c2c3 e7d5 h1e1');
INSERT INTO evaluations VALUES(1816,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(1816,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(1816,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(1816,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(1816,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(1816,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(1816,36,'cp 50','a7b6 c1f4 e8d8 a2a3 f7f6 b1c3 f6e5 f4e5 d7d6 e1c1 c8d7 e5d4 g8e7 c3e4 a8a4 c2c4 d7e6 e4c3 a4c4 f1c4 e6c4 h2h4 c4b3 d1d2 e7c6 d4e3 b3f7 h4h5 d8d7 h5h6 g7h6 e3h6 c6a5 h6f8 a5b3 c1c2 b3d2 c2d2 h8f8 h1h7 d7c6 f2f3 d6d5 h7g7');
INSERT INTO evaluations VALUES(1227,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(1227,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(1227,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(1227,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(1227,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(1227,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(1227,36,'cp 48','c1f4 e8d8 a2a3 f7f6 b1c3 f6e5 f4e5 d7d6 e5d4 g8e7 f2f3 e7c6 d4f2 c8e6 e1c1 g7g6 h2h4 f8h6 c1b1 d8d7 f1b5 h6g7 f2d4 g7d4 d1d4 a8f8 h1e1 f8e8 b1c1 h8f8 d4a4 f8f5 a4e4 f5f8 c1d2 e6f7 a3a4 d7d8 b2b3 d6d5 e4e8 f8e8 e1e8 d8e8 b5c6 b7c6 d2e3 f7e6 c3e2 e6d7 c2c4 d5c4 b3b4');
INSERT INTO evaluations VALUES(35048,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(35048,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(35048,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(35048,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35048,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(35048,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(35048,36,'cp 17','h4e4 f1e2 e4c6 c2c4 d7d6 b1c3 g8f6 h2h3 g7g6 c1g5 f8g7 d1d2 c6d7 e1g1 e8g8 a1d1 b8c6 d4d5 c6e7 d5e6 f7e6 c4c5 d6d5 f1e1 a7a6 f3e5 d7e8 g5f4 e7c6 e2f3 c6e5 f4e5 e8d8 e1e2 c8d7 d1e1 d7c6 d2e3 d8d7 b2b4');
INSERT INTO evaluations VALUES(35057,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(35057,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(35057,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(35057,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(35057,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(35057,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(35057,36,'cp 42','f1e2 e4c6 c2c4 g8f6 b1c3 d7d6 h2h3 f8e7 d1c2 a7a6 c1e3 b7b5 c4b5 c6d7 b2b4 a6b5 e2b5 c7c6 b5d3 c8a6 e1g1 a6d3 c2d3 a8a3 e3c1 a3a8 a1b1 e8g8 a2a4 d7b7 a4a5 b8a6 b4b5 c6b5 b1b5 b7c6 d4d5 f6d5 c3d5');
INSERT INTO evaluations VALUES(35055,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(35055,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(35055,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35055,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35055,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(35055,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(35055,36,'cp 3','e4c6 c2c4 g8f6 b1c3 d7d6 h2h3 g7g6 c1g5 f8g7 d1d2 c6d7 e1c1 h7h6 g5e3 a7a6 f3h2 b7b5 h2g4 b5b4 g4f6 g7f6 c3e4 f6g7 d2b4 f7f5 e4c3 b8c6 b4a4 a8b8 d4d5 c6b4 a4d7 c8d7 d5e6 d7e6 e3d4 g7d4 d1d4 b4c6 d4d1 e8f7 h1e1 g6g5 b2b3 f7f6 g2g4 f5g4 e2g4 e6f7 g4f3 c6e5');
INSERT INTO evaluations VALUES(50225,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50225,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(50225,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(50225,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(50225,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50225,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50225,36,'cp 23','c2c4 c7c6 b1c3 d7d5 c4d5 c6d5 g1f3 f8g7 f1g2 e8g8 f3e5 c8f5 c1f4 d8b6 e1g1 b6b2 c3d5 f6d5 g2d5 b8c6 d5c6 g7e5 f4e5 b7c6 f1e1 f7f6 e5c7 f8c8 c7a5 a8b8 e2e4 f5e6 d1d3 e6a2 a5c3 b2b5 d3e3 a2e6 a1a7 b8b7 a7a1 g8g7 h2h4 c8b8');
INSERT INTO evaluations VALUES(23866,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(23866,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23866,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(23866,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(23866,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(23866,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(23866,36,'cp 27','c3a4 c5b6 c2c3 g8f6 h2h3 g4e6 a4b6 a7b6 c4e6 f7e6 a2a4 e8g8 b2b4 d8e8 c1e3 e8g6 e1g1 h7h6 g1h2 f6h5 a1a2 h5f4 f3h4 g6f7 g2g3 f4h5 h2g2 h5f6 d1e2 c6e7 h4f3 f7h5 a2a3 f8f7 f3g1 h5e2 g1e2 f6d7 f1a1');
INSERT INTO evaluations VALUES(50330,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(50330,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(50330,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(50330,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(50330,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(50330,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(50330,36,'cp -42','c7c5 d4d5 e7e6 d5e6 f7e6 c1b2 d7d5 b1d2 b8c6 e2e4 f8e7 f1b5 c8d7 g1f3 f6e4 d2e4 d5e4 b5c6 d7c6 f3e5 d8d1 a1d1 e7f6 e1g1 f6e5 b2e5 e8g8 f1e1 f8f5 e5c7 a8c8 c7g3 f5d5 c2c4 d5d3 d1d3 e4d3 g3d6 e6e5 d6e5 b7b5 c4b5 c6b5');
INSERT INTO evaluations VALUES(39018,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(39018,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(39018,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(39018,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(39018,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(39018,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(39018,36,'cp 75','b1c3 d7d5 c1f4 g8f6 a1c1 e8g8 e2e3 a7a6 f1e2 d5c4 e2c4 b7b6 c4b3 c8b7 e1g1 b8d7 d4d5 e6d5 c3d5 b7d5 b3d5 f6d5 d1d5 d7e5 d5b7 e5f3 g2f3 c7c5 f1d1 d8f6 b2b3 a6a5 g1g2 a5a4 d1d6 f6b2 c1d1 b2a2 b3a4');
INSERT INTO evaluations VALUES(43836,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(43836,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(43836,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(43836,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43836,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(43836,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(43836,36,'cp 23','c8b7 h2h4 f8d6 f4d6 c7d6 b1c3 e8g8 f1d3 d6d5 d1e2 b8c6 a2a3 a7a6 h4h5 b6b5 h5h6 g7g6 e3e4 d8b6 e2e3 d5e4 d3e4 f6g4 e3f4 f7f5 e4c6 b7c6 f3e5 g4e5 f4e5 f8f7 d4d5 c6d5 c3d5 e6d5 e5d5 a8e8 e1f1');
INSERT INTO evaluations VALUES(23103,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(23103,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(23103,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(23103,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(23103,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(23103,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(23103,36,'cp 47','c7c6 e1g1 d4f3 d1f3 g8f6 f3g3 d7d6 d2d3 f6h5 g3f3 h5f6 c4b3 f8e7 c1e3 e8g8 h2h3 b7b5 a2a4 b5b4 c3e2 c6c5 e2g3 c8e6 b3c4 g8h8 a1d1 f6g8 b2b3 e7g5 f3h5 g5e3 f2e3 g8f6 h5e2 d8e7 d3d4 e5d4 e3d4 c5d4 d1d4 a8d8');
INSERT INTO evaluations VALUES(55466,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(55466,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(55466,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(55466,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(55466,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(55466,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(55466,36,'cp 37','e2e3 g8f6 c2c4 f8d6 f4d6 c7d6 b1c3 d5c4 f1c4 a7a6 e1g1 b7b5 c4d3 c8b7 f3d2 e8g8 a2a4 b5b4 c3e4 f6e4 d3e4 b7e4 d2e4 d6d5 e4d2 b8d7 d1e2 d8b6 d2b3 b6b7 f1c1 a8c8 h2h3 c8c1 a1c1 f8c8 c1c2 c8c2 e2c2');
INSERT INTO evaluations VALUES(55464,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55464,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(55464,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(55464,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(55464,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(55464,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(55464,36,'cp 34','g8f6 c2c4 f8d6 f4d6 c7d6 c4d5 f6d5 f1c4 d5f6 e1g1 a7a6 b1c3 e8g8 a2a4 d6d5 c4e2 b8c6 d1b3 d8d6 a4a5 c8d7 h2h3 f8c8 b3b6 d7e8 f3e5 c8c7 e5d3 a8b8 f1c1 f6d7 b6b3 d6e7 e2f1 b8c8 b3a3 e7a3 a1a3');
INSERT INTO evaluations VALUES(47686,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(47686,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(47686,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(47686,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(47686,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(47686,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(47686,36,'cp 33','c2c4 f8d6 f4d6 c7d6 b1c3 d5c4 f1c4 e8g8 e1g1 a7a6 a2a4 b8c6 d4d5 e6d5 c3d5 c8f5 f3d4 f5e4 d4e2 e4d5 c4d5 d8b6 e2f4 a8d8 a1a3 c6b4 d5f3 d6d5 f4d3 a6a5 d3b4 b6b4 a3b3 b4e7');
INSERT INTO evaluations VALUES(47669,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47669,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(47669,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(47669,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(47669,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(47669,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(47669,36,'cp 30','f8d6 f4d6 c7d6 b1c3 d5c4 f1c4 a7a6 e1g1 b7b5 c4d3 c8b7 f3d2 e8g8 a2a4 b5b4 c3e4 f6e4 d3e4 b7e4 d2e4 d6d5 e4d2 b8d7 d1e2 a6a5 d2b3 d7b6 b3c5 d8e7 f1c1 b6c4 b2b3 c4d6 g2g3 a8c8 c1c2 e6e5');
INSERT INTO evaluations VALUES(55820,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(55820,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(55820,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(55820,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(55820,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(55820,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(55820,36,'cp -95','d5e4 b1c3 f7f5 c1e3 g8f6 g1h3 b8c6 d1d2 f8e7 e1c1 e8g8 f2f3 e6e5 d4d5 c6a5 f3e4 f5f4 e3g1 c8h3 g2h3 b7b6 d2c2 a5b7 c3a4 d8e8 g1f2 e7c5 a4c5 b7c5 f2c5 b6c5 f1e2 e8d7 c2d3 a8e8 h1g1 d7e7');
INSERT INTO evaluations VALUES(56012,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(56012,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(56012,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(56012,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(56012,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(56012,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(56012,36,'cp -112','b1c3 f7f5 c1e3 g8f6 f2f3 b8c6 d1d2 f8b4 a2a3 b4c3 d2c3 b7b6 e1c1 e8g8 g1h3 d8e8 e3f4 c8b7 f1e2 f8f7 h1e1 a8d8 h3g5 f7e7 f3e4 h7h6 e4e5 h6g5 e5f6 g7f6 f4e3 f5f4 e3g1 e6e5 d4d5 c6d4 g1d4 e5d4 c3d2 c7c6 e2g4 c6d5 e1e7 e8e7 d2d4 e7e3 c1c2');
INSERT INTO evaluations VALUES(56011,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(56011,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(56011,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(56011,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(56011,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(56011,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(56011,36,'cp -110','f7f5 c1e3 g8f6 g1h3 b8c6 d1d2 f8e7 e1c1 e8g8 f2f3 e6e5 d4d5 c6a5 f3e4 f5f4 e3g1 c8h3 g2h3 b7b6 c1b1 g8h8 h3h4 a5b7 f1g2 d8d7 h2h3 a7a6 g2f3 b6b5 h4h5 b5c4 h5h6 g7g5 d2e2 b7d6 g1c5 f8b8 b1a1 a6a5 c5a3 a5a4 e2c2 b8b6 c2d2 a8a5 d2c2');
INSERT INTO evaluations VALUES(31059,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(31059,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(31059,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(31059,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(31059,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(31059,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31059,36,'cp 28','d8b6 d1d2 e7e6 f2f3 g8f6 g2g4 f5g6 h2h4 h7h6 e2e3 f8e7 f1e2 d5c4 e2c4 b6b4 c4b3 c6c5 d4d5 e6d5 b3d5 b8c6 e3e4 b4b6 g1h3 a8d8 f4e3 c6d4 e1g1 e8g8 h3f4 g6h7 g1g2 f6d7 a1d1 b6c7 e3d4 c5d4');
INSERT INTO evaluations VALUES(49914,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49914,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49914,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49914,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49914,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49914,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49914,36,'cp 0','d1d4 e4f6 b1c3 b8c6 d4h4 f8b4 c1g5 b4c3 b2c3 e8g8 e1c1 h7h6 g5h6 g7h6 h4h6 d7d5 d1d5 f6d5 c4d3 d5f6 f3g5 c6e5 d3h7 g8h8 h7e4 h8g8 h1d1 d8e7 e4h7 g8h8 h7f5 h8g8');
INSERT INTO evaluations VALUES(49912,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49912,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49912,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49912,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49912,36,'cp 0','e4f6 b1c3 b8c6 d4h4 f8b4 c1g5 b4c3 b2c3 e8g8 e1c1 h7h6 g5h6 g7h6 h4h6 d7d5 c4d3 d8d6 h6g5 g8h8 g5h6');
INSERT INTO evaluations VALUES(49733,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49733,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49733,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49733,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49733,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49733,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49733,36,'cp 0','b1c3 b8c6 d4h4 f8b4 c1g5 h7h6 e1c1 b4c3 b2c3 e8g8 g5h6 g7h6 g2g4 d8e7 g4g5 f6e4 h4h6 e7a3 c1b1 e4c3 b1a1 c3d1 h6g6 g8h8 g6f6 h8g8');
INSERT INTO evaluations VALUES(51964,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51964,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(51964,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(51964,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(51964,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(51964,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(51964,36,'cp 19','e2e4 d7d5 e4d5 e6d5 f1b5 c8d7 b5d7 b8d7 g1f3 g8f6 e1g1 f8e7 d4c5 d7c5 c1e3 e8g8 f1e1 a7a6 d1c2 a8c8 b1d2 f8e8 a1d1 b7b5 a2a3 d8d7 d2b3 c5e4 e3d4 h7h6 c2d3 g7g6 d4f6 e4f6 g2g3 d7h3');
INSERT INTO evaluations VALUES(8685,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(8685,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(8685,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(8685,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(8685,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(8685,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(8685,36,'cp 82','e8g8 f1e2 c7c6 f2f4 f8e8 e2f3 h7h6 g5h4 d8b6 d4d3 b6a5 c1b1 a8b8 h1e1 b7b5 e4e5 b5b4 e5f6 b4c3 b2b3 c8a6 d3d4 c6c5 d4e3 e7f8 f6g7 e8e3 g7f8q g8f8 e1e3 a6b7 f3b7 b8b7 d1d6 a5b5 e3c3 b7d7 d6d7 b5d7 h4f2 c5c4 c3c4 d7d2 f2a7 d2g2 a7b6 f8e8 c4c8 e8d7 c8c7 d7e6 f4f5 e6f5 c7f7 f5g6');
INSERT INTO evaluations VALUES(3437,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(3437,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(3437,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(3437,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(3437,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(3437,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(3437,36,'cp 90','f1e2 f8e8 h1e1 c8d7 e2c4 d7c6 e4e5 f6d7 g5e7 e8e7 e5d6 e7e1 d1e1 d7f6 c3b5 c6b5 c4b5 c7d6 e1d1 d8a5 b5c4 a8d8 h2h4 h7h5 a2a3 b7b6 f2f3 a5f5 d1d2 d8d7 b2b4 f5e5 c4b3 d7e7 c1b1 d6d5 b3d5 f6d5 d4d5 e5e1 b1b2 e1h4 d5d8 g8h7');
INSERT INTO evaluations VALUES(9782,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9782,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9782,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(9782,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9782,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9782,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9782,36,'cp 14','d7d6 d2d3 c5b6 c3a4 c8e6 c4e6 f7e6 h2h3 e8g8 c2c3 d8e8 b2b4 f6h5 a4b6 a7b6 a3a4 e8g6 f3h4 g6f7 a1a3 a8d8 g1h2 c6e7 a3a2 g7g5 h4f3 h5f4 c1f4 f7f4 h2g1 e7g6 f3h2 f4h4 a4a5 b6a5');
INSERT INTO evaluations VALUES(8854,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(8854,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8854,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8854,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(8854,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(8854,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(8854,36,'cp 12','d2d3 a7a6 c1e3 e8g8 e3c5 d6c5 c3d5 d8d6 d5e3 b7b5 c4a2 f8d8 h2h3 c8e6 a2e6 d6e6 a3a4 a8b8 a4b5 a6b5 d1e2 f6h5 e3f5 h5f4 e2e3 c6d4 f3d4 c5d4 e3f3 g8h7 g1h2 g7g6 g2g3 g6f5 g3f4 e5f4 e4f5 e6f5 f1g1 f5f6 a1a7 d8d7 a7a8 b8a8 f3e4 h7h8 e4a8 d7d8');
INSERT INTO evaluations VALUES(8856,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(8856,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(8856,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(8856,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(8856,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(8856,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(8856,36,'cp 7','a7a6 c1e3 e8g8 c3d5 f6d5 c4d5 c5e3 f2e3 c6e7 d5b3 c8e6 b3e6 f7e6 d1d2 b7b5 h2h3 c7c5 d2f2 d8d7 f2g3 f8f6 f1f2 a8f8 a1f1 d7d8 f3d2 f6f2 f1f2 f8f2 g3f2 e7g6 f2g3 g6f8');
INSERT INTO evaluations VALUES(7149,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(7149,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(7149,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(7149,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7149,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7149,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7149,36,'cp 12','c1e3 e8g8 c3d5 f6d5 c4d5 c5e3 f2e3 c6e7 d5b3 c8e6 b3e6 f7e6 a3a4 a6a5 b2b3 f8f7 d1e2 e7c6 f3d2 d8e7 f1f7 e7f7 a1f1 f7e7 h2h3 a8f8 f1f8 g8f8 g1f1 b7b6 f1e1 c6b8 e1d1 f8e8 e2g4 b8d7 d1c1 e8d8');
INSERT INTO evaluations VALUES(30778,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(30778,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(30778,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(30778,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(30778,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(30778,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(30778,36,'cp -47','e7e6 g1e2 f7f5 b1c3 f5e4 c3e4 d7d5 e4g3 d8f6 f3f6 g8f6 c4b3 c7c5 c2c3 b8c6 d2d4 c5d4 e2d4 c6d4 c3d4 f8d6 c1d2 e8d7 a1c1 h8c8 b3a4 b7c6 c1c6 c8c6 e1g1 a7a6 a4c6 d7c6 f1c1 c6d7 g3e2 f6g8 d2f4 g8e7 c1c3 a6a5 f4g3 h7h6 g3d6 d7d6 f2f4 e7f5');
INSERT INTO evaluations VALUES(30628,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(30628,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(30628,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(30628,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(30628,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(30628,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(30628,36,'cp -36','g1e2 f7f5 b1c3 f5e4 c3e4 d7d5 e4g3 d8f6 f3f6 g7f6 c4b3 e8f7 c2c3 c7c5 d2d4 b8c6 c1e3 c5d4 e2d4 c6d4 c3d4 a8c8 e1g1 f8d6 f1c1 g8e7 b3d1 b7a6 d1h5 f7g7 h5d1 h8g8 d1b3 g8f8 g3h5 g7g6 c1c8 f8c8 b3d1');
INSERT INTO evaluations VALUES(49574,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(49574,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49574,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(49574,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49574,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49574,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49574,36,'cp 0','b8c6 e2e4 f8c5 b2b4 c5d4 g1e2 d7d6 d2d3 d4b6 c3a4 b6d4');
INSERT INTO evaluations VALUES(16467,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16467,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16467,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16467,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16467,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16467,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16467,36,'cp 0','e2e4 f8c5 b2b4 c5d4 g1e2 d7d6 d2d3 d4b6 c3a4 b6d4');
INSERT INTO evaluations VALUES(16438,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16438,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16438,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16438,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16438,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16438,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16438,36,'cp 0','f8c5 b2b4 c5d4 g1e2 d7d6 d2d3 d4b6 c3a4 b6d4');
INSERT INTO evaluations VALUES(10172,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10172,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10172,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10172,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10172,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10172,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10172,36,'cp 0','b2b4 c5d4 g1e2 d7d6 d2d3 d4b6 c3a4 b6d4');
INSERT INTO evaluations VALUES(56385,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(56385,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(56385,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(56385,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(56385,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(56385,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(56385,36,'cp 154','c3d5 g8f6 d5f6 d8f6 g1f3 b8c6 d2d3 a7a5 c2c3 f6g6 c1e3 f8e7 h2h4 h7h5 f1e2 c8g4 d1a4 e8c8 a4b5 f7f6 g2g3 h8e8 f3d2 g4e2 e1e2 f6f5 h1g1 f5e4 d2e4 g6f7 f2f3 c8b8 g1d1 f7f5');
INSERT INTO evaluations VALUES(56130,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(56130,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(56130,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(56130,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(56130,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(56130,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(56130,36,'cp 130','g8f6 d5f6 d8f6 g1f3 b8c6 d2d3 f8e7 f1e2 f6g6 e1g1 a7a5 c1e3 e8g8 f3d2 c8h3 e2f3 a5a4 g1h1 h3e6 c2c3 f8d8 f3e2 e7g5 d1c1 f7f5 e4f5 e6f5 d2e4 g5e3 c1e3 c6e7 e3g3 g6g3 f2g3 a4a3 b2b4 f5g6 e4c5 b7b6');
INSERT INTO evaluations VALUES(13632,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(13632,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(13632,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(13632,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(13632,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(13632,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(13632,36,'cp 75','c4c5 f8e7 f1b5 e8g8 b5c6 b7c6 e1g1 f6e4 d1c2 f7f5 f1e1 e7h4 g2g3 h4f6 c1f4 c8d7 f4e5 d7e8 e5f6 g7f6 c3e2 e8h5 f3h4 h5e2 e1e2 d8e8 g1g2 e8f7 f2f3 e4g5 c2d2 a7a5 d2f4 f7d7 a1e1 g8h8 f4d6 d7e8 b2b3');
INSERT INTO evaluations VALUES(13494,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(13494,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(13494,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(13494,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(13494,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(13494,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(13494,36,'cp 68','f8e7 f1b5 e8g8 b5c6 b7c6 e1g1 f6e4 d1c2 f7f5 f1e1 e7h4 g2g3 h4f6 c1f4 a7a5 f4e5 d8e8 g1g2 e8h5 h2h4 c8a6 c3e2 a6e2 c2e2 f8f7 a1c1 h7h6 c1c2 a5a4 e5d6 f7b7 e2d3 h5e8 e1h1 e8g6 f3e5 f6e5 d6e5 b7b4 h1e1 a8f8 f2f3 e4f6 d3d2 b4c4 c2c4 d5c4');
INSERT INTO evaluations VALUES(8071,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(8071,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(8071,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(8071,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(8071,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(8071,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(8071,36,'cp 62','f1b5 e8g8 b5c6 b7c6 e1g1 f6e4 d1c2 f7f5 f1e1 e7h4 g2g3 h4f6 c1f4 a7a5 f4e5 c8a6 a1d1 d8e8 g1g2 e8h5 h2h4 h7h6 d1c1 f6e7 c3e2 a6e2 e1e2 f5f4 e2e4 d5e4 c2e4 f4g3 e5g3 a8c8 e4e6 g8h7 e6e4 h5f5 e4f5 f8f5 f3e5 e7f6 f2f3');
INSERT INTO evaluations VALUES(23481,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(23481,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(23481,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(23481,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(23481,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(23481,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(23481,36,'cp 140','e2e4 c7c6 c2c4 d7d6 f1e2 g7g6 b1c3 f8g7 g1f3 g8f6 e1g1 e8g8 c1e3 b4a6 h2h3 e7e5 a2a3 c6d5 c4d5 f6e8 f3d2 f7f5 f2f3 g7f6 d2c4 f6g5 e3f2 g5h4 f2h4 d8h4 d1e1 h4e1 f1e1 a6c5 c3b5 f5e4 b5d6 e8d6 c4d6');
INSERT INTO evaluations VALUES(23479,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(23479,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(23479,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(23479,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(23479,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(23479,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(23479,36,'cp 140','d7d6 c1e3 g8f6 d1d2 a7a5 f2f3 g7g6 a2a3 b4a6 b1c3 f8g7 f1a6 a8a6 e3h6 g7h6 d2h6 c7c6 g1e2 b7b5 e2d4 d8b6 e1c1 c8d7 d5c6 d7c6 d4c6 b6c6 c3d5 f6d5 e4d5 c6c5 h6g7 c5e3 c1b1 e3e5 g7e5 d6e5 h1e1 e8g8 e1e5');
INSERT INTO evaluations VALUES(21198,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(21198,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(21198,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(21198,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(21198,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(21198,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(21198,36,'cp 146','c1e3 c7c6 c2c3 b4a6 f1a6 b7a6 d1a4 d8c7 a4c6 c7c6 d5c6 e7e5 c3c4 g8e7 c6c7 c8e6 b2b3 f7f5 f2f3 e8d7 b1c3 d7c7 h2h4 f5e4 f3e4 e7g8 c3d5 c7d7 g1e2 g8f6 e2c3 f8e7 h1f1 e6d5 e4d5 h8f8 e1e2 f6g4 f1f8 a8f8 e3a7 e7h4 c3e4');
INSERT INTO evaluations VALUES(7486,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(7486,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(7486,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(7486,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(7486,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(7486,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(7486,36,'cp 18','d7d6 a2a4 c8d7 f1e1 e8g8 h2h3 c6a5 b3a2 c7c5 a4b5 a6b5 b1c3 b5b4 c3d5 f6d5 a2d5 d7c6 d5c6 a5c6 c1e3 d8d7 f3d2 f7f5 a1a8 f8a8 e4f5 d7f5 d2e4 a8a2 d1b1 a2a5 e4g3 f5f7 c2c4 h7h6 b1c2 a5a2 e1b1 f7g6 g3e4 g8h7 c2b3 a2a8 b3d1 h7g8 b1a1 a8a1 d1a1');
INSERT INTO evaluations VALUES(7388,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7388,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7388,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(7388,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(7388,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(7388,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(7388,36,'cp 23','a2a4 c8d7 h2h3 e8g8 f1e1 c6a5 b3a2 c7c5 a4b5 a6b5 b1c3 b5b4 c3d5 f6d5 a2d5 d7c6 d5c6 a5c6 c1e3 d8d7 f3d2 f7f5 a1a8 f8a8 e4f5 d7f5 d2f1 e7h4 c2c3 h7h6 f1d2 h4e7 d2e4 a8a2 d1b3 f5f7 c3c4 a2a8 b3d1 f7g6 d1e2 c6d8 g1h2 d8e6 e1b1 a8a2 g2g3 e7g5 h3h4 g5e3');
INSERT INTO evaluations VALUES(16630,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(16630,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(16630,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(16630,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(16630,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(16630,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(16630,36,'cp -82','h7h6 d4c6 b7c6 g5f4 f6e4 d1e2 d8e7 f2f3 e4d6 b1d2 e7e2 f1e2 e8d8 e1c1 a7a5 d2e4 d6b7 g2g4 d7d6 f4e3 c8d7 c2c4 a5a4 f3f4 c6c5 e3d2 b7a5 d2c3 a5c6 e2f3 a8a6 e4g3 d7e6 f3e2 a4a3 b2b3 c6b4 c1b1 e6d7');
INSERT INTO evaluations VALUES(16175,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(16175,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(16175,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(16175,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(16175,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(16175,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(16175,36,'cp -92','d4c6 b7c6 g5f4 f6e4 d1e2 d8e7 f2f3 e4c5 e2e7 f8e7 f4c7 d7d6 b1d2 c8f5 e1c1 c5b7 f1a6 a8c8 a6b7 c8c7 b7a6 e8f8 a6d3 f5c8 f3f4 d6d5 f4f5 h6h5 g2g3 e7d6 d2f3 f8e7 h1e1 e7f6 c2c4 d6b4 e1e3 b4c5 e3e5 c7e7 e5e7 c5e7 f3d4 c8d7');
INSERT INTO evaluations VALUES(20263,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20263,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(20263,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(20263,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(20263,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(20263,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20263,36,'cp 10','e2e3 g8f6 b1d2 f8d6 f4g3 e8g8 f1d3 g7g6 e1g1 b7b6 g3h4 d6e7 d1e2 c8b7 a2a4 a7a6 f1e1 f6h5 h4e7 d8e7 a1d1 f7f5 f3e5 c6e5 d4e5 h5g7 f2f4 f8d8 d2f3 d8d7 c3c4 b7c6 c4d5 e6d5 d3a6');
INSERT INTO evaluations VALUES(34935,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(34935,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(34935,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(34935,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(34935,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(34935,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(34935,36,'cp 51','b1c3 d5e6 f4e5 e6e5 f1e2 b8c6 g1f3 e5a5 d2d4 c8g4 h2h3 g4f3 e2f3 e8c8 f3c6 b7c6 e1g1 d8d4 d1f3 d4d7 c3e4 f8e7 c1d2 a5b5 a2a4 b5d5 d2c3 g8f6 c3f6 e7f6 e4f6 g7f6 f3d5 c6d5 f1f6 h8e8 a1f1 e8e7 f1f5 a7a5 f6a6 e7e2 a6a8 c8b7 a8a5 e2c2');
INSERT INTO evaluations VALUES(34933,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(34933,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(34933,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(34933,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(34933,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(34933,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(34933,36,'cp 29','d5e6 f4e5 e6e5 f1e2 b8c6 g1f3 e5a5 d2d4 c8g4 h2h3 g4f3 e2f3 e8c8 f3c6 b7c6 e1g1 d8d4 d1f3 d4d7 c3e4 g8f6 e4f6 g7f6 c1e3 f8c5 a1e1 c8b7 e3c5 a5c5 f3f2 d7d5 f2c5 d5c5 c2c3 f6f5 e1e7 h8f8 a2a4 c5d5 f1f2 h7h6 g1h2 a7a5 c3c4 d5d4 f2f5 d4c4 f5a5');
INSERT INTO evaluations VALUES(34869,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(34869,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(34869,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(34869,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(34869,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(34869,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(34869,36,'cp 49','f4e5 e6e5 f1e2 b8c6 g1f3 e5a5 d2d4 c8g4 h2h3 g4f3 e2f3 e8c8 f3c6 b7c6 e1g1 d8d4 d1f3 d4d7 c3e4 g8f6 e4f6 g7f6 c1e3 a5d5 f3d5 c6d5 a1d1 f8d6 f1f6 h8e8 e3f4 d6f4 f6f4 e8e2 f4f2 e2f2 g1f2 c7c5 g2g4 c8b7 f2e2 b7c6 d1f1 c5c4 f1f5 c6c5 h3h4 c5d4');
INSERT INTO evaluations VALUES(34865,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(34865,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(34865,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(34865,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(34865,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(34865,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(34865,36,'cp 50','e6e5 f1e2 e5a5 g1f3 b8c6 d2d4 g8f6 d4d5 c6e7 e2b5 c7c6 d5c6 b7c6 b5c4 e7d5 c1d2 a5c5 d1e2 f8e7 c3d5 f6d5 d2g5 c8e6 g5e7 e8e7 c4d5 c5d5 e1g1 h8d8 a1e1 d8e8 e1d1 d5c5 f3d4 e7f8 e2f2 f8g8 d4e6 c5f2 f1f2 e8e6 d1d7');
INSERT INTO evaluations VALUES(34967,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(34967,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(34967,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(34967,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(34967,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(34967,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(34967,36,'cp 54','f1e2 b8c6 g1f3 e5a5 d2d4 g8f6 d4d5 c6e7 e2b5 c7c6 d5c6 b7c6 b5c4 e7d5 c1d2 a5c7 c3d5 f6d5 d1e2 f8e7 e1c1 e8g8 e2e4 a8b8 c4b3 h7h6 h1e1 e7d6 c1b1 c8e6 f3d4 e6d7 h2h3 f8e8 e4d3 d6e5 g2g4 c7b6 d4f5 d7f5 d3f5');
INSERT INTO evaluations VALUES(34968,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(34968,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(34968,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(34968,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(34968,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(34968,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(34968,36,'cp 33','b8c6 g1f3 e5a5 d2d4 g8f6 d4d5 c6e7 c1d2 c8d7 d5d6 c7d6 c3e4 a5b6 e4f6 g7f6 a2a4 f8g7 a4a5 b6b2 e1g1 d7c6 e2d3 f6f5 d1e2 h8g8 g1h1 b2f6 a1e1 f6e6 e2e6 f7e6 e1e6 e8d7 f1e1 a8e8 d2b4 g7e5 f3e5 d6e5 e6e5 c6g2 h1g1 g2e4 g1f1 e4d3 c2d3');
INSERT INTO evaluations VALUES(39360,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39360,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39360,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(39360,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39360,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39360,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39360,36,'cp 0','b4c3 b2c3 d7d6 f1g2 f7f5 d2d4 g8f6 g1h3 e8g8 c4c5 h7h6 d1b3 f8f7 f2f4 b8c6 h3g5 h6g5 f4g5 c6a5 b3a4 a5c6');
INSERT INTO evaluations VALUES(39381,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39381,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(39381,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39381,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39381,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(39381,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(39381,36,'cp 12','b2c3 d7d6 f1g2 f7f5 c4c5 d6c5 c1a3 d8e7 g1h3 g8f6 e1g1 b8d7 e2e4 f5e4 h3g5 e8g8 g5e4 f6e4 g2e4 a8b8 d2d4 e5d4 e4d5 g8h8 c3d4 b7b6 f1e1 e7f6 d1c2 c8b7 d5b7 b8b7 a1d1 h7h6 f2f4 f6f7 d4c5 b6c5 a3c5 d7c5 c2c5 b7b2');
INSERT INTO evaluations VALUES(39377,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(39377,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(39377,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(39377,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(39377,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39377,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39377,36,'cp 0','d7d6 f1g2 b8c6 d2d3 g8e7 g1h3 c8e6 f2f4 d8c8 h3g5 e6d7 f4e5 d6e5 c1a3 f7f6 g5e4 d7h3 g2f3 h3g4 f3g2');
INSERT INTO evaluations VALUES(51394,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(51394,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(51394,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(51394,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(51394,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51394,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51394,36,'cp 0','d7d5 e4d5 e6d5 c4e2 f8d6 d2d4 h7h6 e1g1 g8e7 c2c4 e8g8 b1c3 c8e6 c4d5 e7d5 f1e1 b8d7 c3e4 d7f6 e4d6 d8d6 f3e5 a8d8 e5c4 d6e7 c4e5 e7d6');
INSERT INTO evaluations VALUES(9086,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(9086,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(9086,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(9086,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(9086,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(9086,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(9086,36,'cp 122','e4d5 f6d5 d1b3 d5f6 c4f7 e8f8 e1g1 d8d3 b1d2 e5e4 f7c4 d3d6 f3g5 d6e7 b3c2 c6e5 c4e2 e4e3 d2b3 e3f2 g1h1 c5b6 b3d4 e5g4 g5h3 b6d4 c3d4 c8f5 e2d3 f5d3 c2d3 e7e4 d3d2 e4d5 h3f2 g4f2 d2f2 f8g8 c1f4 f6e4');
INSERT INTO evaluations VALUES(9073,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(9073,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(9073,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(9073,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(9073,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(9073,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(9073,36,'cp 117','f6d5 d1b3 d5f6 c4f7 e8f8 f7c4 c6a5 b3a4 c7c6 c1e3 c5e3 f2e3 g7g6 f3e5 f8g7 c4b3 a5b3 a2b3 d8d5 a4f4 d5g2 h1f1 h8f8 b1d2 c8h3 e5f3 g2g4 f4c7 g7g8 f1g1 g4e6 c7e5 e6d7 e5d4 f6d5 e3e4 a8e8 e1d1 h3g4 g1g4 d7g4 e4d5');
INSERT INTO evaluations VALUES(9155,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(9155,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(9155,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(9155,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(9155,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(9155,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(9155,36,'cp 102','d1b3 d5f6 c4f7 e8f8 e1g1 d8d3 f7c4 d3d6 b1d2 e5e4 f3g5 d6e7 b3c2 c6e5 c4e2 e4e3 d2b3 e3f2 g1h1 c5d6 f1f2 e5g4 f2f3 d6h2 b3d4 c7c5 d4f5 c8f5 f3f5 h2d6 c1f4 d6f4 f5f4 g4e3 c2b3 a8e8 e2b5 h7h6 g5f3 c5c4 b3a4 f8f7 b5c4 e3c4 a4c4 e7e6 c4b5 g7g5 f4f5 f7g6 f3e5 g6g7 b5b7 e8e7 b7f3');
INSERT INTO evaluations VALUES(22842,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(22842,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(22842,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(22842,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(22842,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22842,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22842,36,'cp 0','c1b2 g8f6 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a1e1 b6a5 c2c3 c8d7 e1e6 d7e6 f1e1 a5c7 b2c1 d8f6 g2g3 a7a5 d2c2 h7h6 d3h7 g8h8 h7d3 h8g8');
INSERT INTO evaluations VALUES(23280,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(23280,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(23280,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(23280,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(23280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23280,36,'cp 0','e5e4 f3d4 g8f6 d4c6 b7c6 c4e2 d7d5 e2h5 f6h5 d1h5 e8d7 b1c3 d8f6 b2b3 d7d8 c1b2 f6f5 h5f5 c8f5 c3a4 d8d7 b2d4 f8a3 a4c5 d7d6 c5b7 d6d7');
INSERT INTO evaluations VALUES(23305,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23305,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23305,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23305,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23305,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23305,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23305,36,'cp -2','f3d4 g8f6 d4c6 b7c6 d2d3 d7d5 c4b3 c8f5 e1g1 f8e7 b1c3 e8g8 c1g5 e4d3 c2d3 g8h8 f1e1 e7d6 g5h4 d8d7 d3d4 f5g6 h4g3 f6h5 g3d6 c7d6 c3e2 d7g4 a1c1 g4g5 e2g3 a8e8 e1e8 g6e8 d1c2 g5f6 c2d2 h5f4 c1e1');
INSERT INTO evaluations VALUES(26535,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26535,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(26535,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(26535,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(26535,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(26535,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(26535,36,'cp 12','e2e3 e7e6 f1d3 f5d3 d1d3 f8d6 f4g3 g8f6 g1f3 e8g8 g3h4 d6e7 b1d2 c6b8 h4f6 e7f6 e3e4 b8d7 e4d5 e6d5 e1g1 c7c6 g2g3 f8e8 f1e1 d8c7 g1g2 g7g6 a2a4 f6g7 a4a5 e8e1 f3e1 c6c5 e1c2 c5d4 c2d4 d7c5 d3e3 c7d7 d2f3');
INSERT INTO evaluations VALUES(16406,30,'cp -541',NULL);
INSERT INTO evaluations VALUES(16406,31,'cp -544',NULL);
INSERT INTO evaluations VALUES(16406,32,'cp -523',NULL);
INSERT INTO evaluations VALUES(16406,33,'cp -536',NULL);
INSERT INTO evaluations VALUES(16406,34,'cp -546',NULL);
INSERT INTO evaluations VALUES(16406,35,'cp -545',NULL);
INSERT INTO evaluations VALUES(16406,36,'cp -533','c6d4 f3d1 d7d5 c4f1 c8g4 f2f3 g4h3 c2c3 d4c6 g2h3 f6e4 f1b5 d8h4 e1e2 h4f2 e2d3 e4c5 d3c2 c5e6 d1g1 f2f3 g1f1 f3h5 f1e2 h5e2 b5e2 f7f5 d2d3 e8c8 b1d2 f8e7 h1f1 h8f8 a2a4 e7g5 d2b3 g5c1 a1c1 f8f6 c1e1 d8d6 b3d2 a7a6');
INSERT INTO evaluations VALUES(17003,30,'cp -530',NULL);
INSERT INTO evaluations VALUES(17003,31,'cp -551',NULL);
INSERT INTO evaluations VALUES(17003,32,'cp -542',NULL);
INSERT INTO evaluations VALUES(17003,33,'cp -542',NULL);
INSERT INTO evaluations VALUES(17003,34,'cp -549',NULL);
INSERT INTO evaluations VALUES(17003,35,'cp -551',NULL);
INSERT INTO evaluations VALUES(17003,36,'cp -529','f3d1 d7d5 c4f1 c8g4 f2f3 g4h3 c2c3 d4c6 g2h3 f6e4 f1b5 d8h4 e1e2 h4h3 d1f1 h3h5 d2d3 e4c5 e2d1 f8d6 c1e3 e8g8 b5c6 b7c6 d1c2 c5e6 b1d2 a7a5 a2a4 c6c5 h1g1 f7f6 f3f4 e5f4 e3f2 a8b8 a1e1 h5f5 e1e6 f5e6 d2f3 g8h8');
INSERT INTO evaluations VALUES(26049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26049,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26049,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(26049,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(26049,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(26049,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(26049,36,'cp 0','g8f6 d2d4 c5d4 f3d4 c6d4 b5d7 d8d7 d1d4 e7e6 f1d1 f8e7 d1d3 d7c6 d3g3 f6h5 g3h3 h5f6');
INSERT INTO evaluations VALUES(24476,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24476,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24476,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24476,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24476,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24476,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24476,36,'cp 9','d1e2 e7e5 d2d3 f8e7 b5c4 e8g8 h2h3 a7a6 a2a4 g8h8 c3d1 c6d4 f3d4 c5d4 f2f4 d8c7 d1f2 a8e8 e2f3 b7b5 a4b5 a6b5 c4b3 d7e6 c1d2 e8c8 f3d1 e6b3 c2b3 c8a8 a1a8 f8a8 f2g4 f6g4 h3g4 f7f6 f4f5 d6d5 e4d5 c7d7 d1f3');
INSERT INTO evaluations VALUES(27102,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(27102,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(27102,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(27102,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(27102,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(27102,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(27102,36,'cp 79','e1g1 g8f6 b1c3 f8e7 c4e6 f7e6 e4e5 d6e5 d4e5 d8d6 e5b5 d6c6 c1f4 e7d6 b5c6 b7c6 f4d2 e8c8 a1d1 h8f8 c3a4 f6g4 g2g3 d6e7 d2c3 d8d1 f1d1 g4f2 d1e1 e7f6 c3f6 f8f6 a4c5 f6f5 c5e6 g7g6 e1f1 f5e5 e6d4 f2e4 d4c6 e5c5 c6b4 e4d2 f1f7');
INSERT INTO evaluations VALUES(27103,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(27103,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(27103,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(27103,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(27103,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(27103,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(27103,36,'cp 94','g8f6 b1c3 f8e7 c4e6 f7e6 e4e5 d6e5 d4e5 d8d6 e5b5 d6c6 f1e1 e8f7 b5c6 b7c6 g2g3 a7a5 c3a4 h8d8 c1e3 f6d5 c2c3 a8b8 e1e2 d5b6 a4b6 c7b6 a1e1 e6e5 e3d4 e5d4 e2e7 f7g8 c3d4 d8d4 e7a7 d4d6 e1e7 d6g6 f2f4 b8d8 g1g2 c6c5 a7b7 h7h5 f4f5 g6g5 h2h4 g5f5 e7g7 g8h8 g7h7 h8g8 b7g7 g8f8');
INSERT INTO evaluations VALUES(25411,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(25411,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(25411,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(25411,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(25411,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(25411,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(25411,36,'cp 103','b1c3 f8e7 c4e6 f7e6 e4e5 d6e5 d4e5 d8d6 e5b5 d6c6 b5c6 b7c6 g2g3 e8f7 f1e1 h8b8 b2b3 f6d5 c3a4 a7a5 a2a3 c6c5 e1e4 a8a6 c1b2 a6c6 b2e5 c5c4 e4c4 c6c4 b3c4 d5f6 g1g2 e7d6 e5d4 d6e7 a4b2 f6d7 d4c3 e7f6 c3f6 d7f6 b2d3 f6d7');
INSERT INTO evaluations VALUES(25410,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(25410,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(25410,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(25410,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(25410,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(25410,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(25410,36,'cp 98','f8e7 c4e6 f7e6 e4e5 d6e5 d4e5 d8d6 e5b5 d6c6 b5c6 b7c6 f1e1 e8f7 g2g3 a7a5 c3a4 c6c5 c1f4 e7d6 f4e5 h8b8 e1e2 a8a6 b2b3 c5c4 a4b2 d6e5 e2e5 c4c3 b2d3 a5a4 b3b4 a4a3 e5c5 f6d5 a1b1 f7e7 g1f1 h7h6 b1d1 d5b4 d3b4 b8b4 c5c7 e7f6 c7c3');
INSERT INTO evaluations VALUES(42222,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(42222,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(42222,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(42222,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(42222,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42222,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(42222,36,'cp 45','e2e3 e8g8 f1d3 c7c6 g1e2 f8e8 d1c2 h7h6 g5h4 b4e7 e1g1 b8d7 a2a3 a7a6 a1c1 f6h5 h4e7 d8e7 c1e1 a6a5 c3a4 e7d8 e1c1 d7b6 a4c3 d8g5 g1h1 h5f6 c2b3 b6d7 h2h3 g5h4 h1h2 h6h5');
INSERT INTO evaluations VALUES(42219,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(42219,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(42219,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(42219,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(42219,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(42219,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(42219,36,'cp 43','b8d7 f1d3 e8g8 g1e2 c7c6 e1g1 h7h6 g5h4 f8e8 d1c2 b4e7 f2f3 b7b5 a2a3 c8b7 h2h3 a7a5 h4f2 a8c8 f1d1 b5b4 a3b4 e7b4 a1c1 c6c5 d4c5 d7c5 d3b5 e8e5 e2f4 d8b6 f4d3 c5d3 c2d3 e5e6 c3a4 b6a7 d3d4 a7d4 c1c8 b7c8');
INSERT INTO evaluations VALUES(44206,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44206,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44206,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44206,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44206,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(44206,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(44206,36,'cp 11','d7d5 g1f3 d5c4 f1c4 a7a6 a2a4 c5d4 e3d4 f8b4 e1g1 b8c6 f1e1 e8g8 c1g5 h7h6 g5h4 b4e7 h2h3 c8d7 h4f6 e7f6 d4d5 e6d5 c3d5 d7f5 d1e2 f8e8 e2e8 d8e8 e1e8 a8e8 d5f6 g7f6 a1c1 e8c8 c1d1 c6e7 b2b3 f5e6 c4e6 f7e6 d1d7');
INSERT INTO evaluations VALUES(53016,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(53016,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(53016,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53016,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53016,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53016,36,'cp 0','c5d4 d1d4 g8f6 g1f3 g7g6 f3g5 e7e6 b1c3 b8c6 d4d1 d8c7 a2a4 a7a6 g5f3 h7h6 d1e2 b7b6 c1f4 g6g5 f4e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(53124,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53124,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53124,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53124,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53124,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53124,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53124,36,'cp 0','g1f3 g8f6 d1d4 g7g6 f3g5 e7e6 b1c3 b8c6 d4d1 d8c7 g5f3 a7a6 a2a4 h7h6 e1g1 b7b6 c1e3 c8b7 c4e2 f8g7 d1d2 f6g4 e3f4 g4e5 f4e3');
INSERT INTO evaluations VALUES(53123,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53123,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53123,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53123,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53123,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53123,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53123,36,'cp 0','g8f6 d1d4 g7g6 f3g5 e7e6 b1c3 b8c6 d4d1 a7a6 a2a4 d8c7 g5f3 h7h6 e1g1 b7b6 d1e2 c8b7 f1d1 g6g5 c4b3 f8g7 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(45245,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45245,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45245,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45245,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45245,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45245,36,'cp 7','d1d4 g7g6 f3g5 e7e6 e1g1 b8c6 d4d3 c6e5 d3e2 e5c4 e2c4 a7a6 a2a4 c8d7 f1d1 h7h6 g5f3 a8c8 c4d3 d7c6 b1c3 f8g7 f3d4 c6d7 h2h3 e8g8 c1e3 f8e8 d4b3 d7c6 b3d4');
INSERT INTO evaluations VALUES(45243,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45243,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45243,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45243,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45243,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45243,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45243,36,'cp 0','g7g6 f3g5 e7e6 b1c3 a7a6 a2a4 b8c6 d4d1 h7h6 g5f3 b7b6 e1g1 c8b7 d1e2 d8c7 f1d1 g6g5 c4b3 f8g7 c1e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(5153,30,'cp 330',NULL);
INSERT INTO evaluations VALUES(5153,31,'cp 308',NULL);
INSERT INTO evaluations VALUES(5153,32,'cp 331',NULL);
INSERT INTO evaluations VALUES(5153,33,'cp 339',NULL);
INSERT INTO evaluations VALUES(5153,34,'cp 354',NULL);
INSERT INTO evaluations VALUES(5153,35,'cp 340',NULL);
INSERT INTO evaluations VALUES(5153,36,'cp 343','d4b5 e8f8 b5d6 h7h5 f2f4 g8f6 e4e5 f6e8 c3b5 e8d6 b5d6 b7b6 f1b5 d8c7 d1d2 c6e7 e1c1 e7d5 b5e2 c8b7 d6b7 c7b7 e2f3 a8c8 f3d5 e6d5 d2d5 b7d5 d1d5 f8e7 h1d1 c8c7 c2c3 h8c8 c1c2 g7f8 g2g3 e7e6 a2a4 f8e7 d5b5 c7c6 c2b3 d7d6 a4a5 d6e5');
INSERT INTO evaluations VALUES(57444,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(57444,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(57444,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(57444,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(57444,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(57444,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(57444,36,'cp 51','d2d4 d7d5 d4e5 f6e5 e2e3 f8b4 c2c3 b4d6 d1d5 g8f6 d5d1 c8f5 b1a3 f6e4 g1f3 d8f6 a3c4 b8d7 f1d3 e4f2 c4d6 c7d6 e1f2 e5e4 d3b1 e4f3 d1f3 f5b1 f3f6 d7f6 h1b1 e8g8 f2g1 a8e8 c3c4 e8e3 b2a3 e3e6 b1d1 f8d8 d1d3 d6d5 a1d1 e6e2 c4d5 e2a2');
INSERT INTO evaluations VALUES(21264,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(21264,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(21264,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(21264,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(21264,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(21264,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(21264,36,'cp 203','d1d8 e8d8 c4f7 g8f6 c1e3 f8d6 b1c3 h8f8 f7c4 a7a6 h2h3 b7b5 c4d3 c8e6 e1c1 d8e8 a2a3 g7g5 h3h4 g5g4 f3d2 f6h5 g2g3 h5f4 d3f1 h6h5 h1h2 a8d8 c1b1 d6e7 f2f3 f4g6 c3d5 e7d6 f3g4 e6g4 f1e2 g4e2 h2e2');
INSERT INTO evaluations VALUES(2260,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(2260,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(2260,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(2260,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(2260,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(2260,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(2260,36,'cp 192','e8d8 c4f7 f8d6 c1e3 g8f6 b1c3 h8f8 f7c4 a7a6 h2h3 b7b5 c4d3 c8e6 a2a3 f6h5 a3a4 b5b4 c3b1 b4b3 c2c3 c6a5 b1d2 d8e8 f3h4 e8f7 h4f5 h5f6 h1f1 f7g8 f5d6 c7d6 f2f4 e5f4 e3f4 d6d5 f4d6 f8c8 e4d5 f6d5 e1c1');
INSERT INTO evaluations VALUES(2309,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(2309,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(2309,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(2309,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(2309,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(2309,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(2309,36,'cp 201','c4f7 f8d6 c1e3 g8f6 b1c3 h8f8 f7c4 a7a6 h2h3 b7b5 c4d3 c8e6 a2a3 f6d7 e1c1 d8c8 f3h4 c6e7 g2g4 d6c5 h4f5 f8f7 h3h4 c5e3 f2e3 c8b7 h1g1 a8h8 a3a4 d7c5 a4b5 c5d3 c2d3 a6b5 c3b5 e7c6 b5c3 e6b3 d1d2 f7d7 g4g5 h6g5 g1g5');
INSERT INTO evaluations VALUES(2307,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(2307,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(2307,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(2307,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(2307,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(2307,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(2307,36,'cp 205','g8f6 b1c3 f8d6 c1e3 h8f8 f7c4 f6g4 e1c1 d8e8 a2a3 g4e3 f2e3 c6a5 c4d5 c8g4 h1f1 e8e7 d5a2 a8d8 h2h3 g4f3 g2f3 g7g5 b2b4 a5c6 c1b2 e7d7 a2c4 d7c8 c4e6 c8b8 b2b3 c6e7 c3e2 e7g6 e2g3 f8f6 e6f5 g6h4 c2c3 d8f8 a3a4');
INSERT INTO evaluations VALUES(2274,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(2274,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(2274,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(2274,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(2274,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(2274,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(2274,36,'cp 199','c1e3 c8d7 b1c3 f8b4 e1c1 b4c3 b2c3 d8c8 f3d2 c6d8 f7c4 f6g4 e3c5 b7b5 c4e2 d8b7 d2b3 g4f6 f2f3 d7e6 c5f2 b7d6 a2a4 a7a6 b3c5 e6c4 e2d3 h8d8 f2g3 d8e8 h2h4 b5a4 c5a4 f6d7 h4h5 a6a5 c1d2 e8e6 d2e2 c4a6');
INSERT INTO evaluations VALUES(55584,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(55584,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(55584,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(55584,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(55584,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(55584,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(55584,36,'cp 25','e5d4 c3d4 d6d5 e4d5 g8f6 b1c3 f6d5 f1c4 c8e6 d1b3 f8b4 g1f3 e8g8 e1g1 d5c3 c4e6 c3e2 g1h1 e2d4 f3d4 d8d4 c1e3 d4h4 e6d5 c7c6 d5c4 b8a6 c4a6 b7a6 a1c1 b4d6 h2h3 h4e4 f1d1 e4e5 g2g3 d6e7 h1g1 a8b8 e3f4 b8b3 f4e5 b3b4 b2b3');
INSERT INTO evaluations VALUES(55684,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55684,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55684,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(55684,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(55684,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55684,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(55684,36,'cp 37','c3d4 d6d5 e4d5 g8f6 b1c3 f6d5 f1c4 c8e6 d1b3 f8b4 g1f3 e8g8 e1g1 d5c3 c4e6 c3e2 g1h1 e2d4 b3b4 d4e6 b4b7 b8d7 f1d1 d7c5 d1d8 c5b7 d8f8 g8f8 f3g5 f8e7 g5e6 e7e6 c1e3 c7c5 a1c1 a8d8 g2g3 d8d3 h1g2 e6d5');
INSERT INTO evaluations VALUES(55673,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(55673,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55673,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(55673,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(55673,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(55673,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55673,36,'cp 30','d6d5 e4d5 g8f6 b1c3 f6d5 f1c4 c8e6 d1b3 f8b4 g1f3 e8g8 e1g1 d5c3 c4e6 c3e2 g1h1 e2d4 b3b4 d4e6 b4b7 b8d7 f1d1 d7c5 d1d8 c5b7 d8f8 g8f8 c1e3 b7c5 a1d1 c5a4 f3d4 e6d4 e3d4 c7c5 d4e3 a4b2 e3c5 f8g8 d1d7 a8c8 c5e3 h7h6 g2g3 a7a5 h1g2 b2c4 e3f4 c4b6');
INSERT INTO evaluations VALUES(23868,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(23868,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(23868,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(23868,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(23868,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(23868,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(23868,36,'cp 62','b2b4 c5b6 a2a4 a7a6 d2d3 g8f6 h2h3 g4e6 c4e6 f7e6 b1d2 e8g8 a1b1 h7h6 d1b3 d8e7 d2c4 b6a7 b4b5 a6b5 a4b5 c6d8 b5b6 a7b6 c4b6 c7b6 b3b6 f6h5 c1e3 h5f4 e3f4 f8f4 b6e3 d8c6 b1b5 f4f7 f1b1 a8a2 b5b2 a2b2 b1b2 c6a5 f3d2');
INSERT INTO evaluations VALUES(51930,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(51930,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(51930,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(51930,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(51930,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(51930,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(51930,36,'cp -85','d7d5 c4b5 b8d7 e4e5 d8b6 b5e2 g7g5 c2c3 g8e7 g1f3 g5f4 d2d4 f8h6 g2g3 f7f6 e5f6 d7f6 d4c5 b6c5 d1d4 c5d4 c3d4 e7f5 g3f4 c8d7 b1c3 f6e4 e2d3 d7c6 c3e2 a7a6 h1g1 c6b5 d3b5 a6b5 e2g3 e8e7 e1e2 h8c8 g3f5 e6f5 e2d3 e4f2 d3e2 c8c2 c1d2 h6f4 e2f2');
INSERT INTO evaluations VALUES(54160,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(54160,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(54160,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(54160,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(54160,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(54160,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(54160,36,'cp -53','c4b5 b8d7 e4e5 d8b6 b5e2 g7g5 c2c3 g8e7 g1f3 g5f4 d2d4 f8h6 e1g1 e8g8 g2g3 f7f6 g3f4 c5d4 g1h1 d4c3 f1g1 h6g7 b1c3 f6e5 f4e5 e7g6 e2d3 b6f2 g1f1 f2c5 f1g1 f8f3 d3g6 h7g6 d1f3 d7e5 f3g2 c5c4 c3e2 c8d7 c1e3');
INSERT INTO evaluations VALUES(53244,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(53244,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(53244,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(53244,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(53244,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(53244,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(53244,36,'cp 39','g8f6 c4d5 f6d5 g1f3 d5c3 b2c3 g7g6 e2e3 f8g7 f1b5 c8d7 b5e2 c5d4 c3d4 e8g8 e1g1 d7c6 a2a4 f8e8 a4a5 b8d7 d1b3 a8c8 f1d1 e7e6 e2b5 b7b6 b5c6 c8c6 c1a3 b6a5 b3a4 d8b6 d1b1 b6c7 b1c1 e8c8 c1c6 c7c6 a4a5');
INSERT INTO evaluations VALUES(45348,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(45348,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(45348,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(45348,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(45348,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(45348,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(45348,36,'cp 36','c4d5 f6d5 g1f3 d5c3 b2c3 g7g6 e2e3 f8g7 f1b5 c8d7 b5e2 e8g8 e1g1 d7c6 c1a3 c5d4 c3d4 f8e8 a1b1 e7e6 f3e5 c6e4 b1b2 e4d5 e5c4 g7f8 a3f8 e8f8 c4d2 b8d7 e3e4 d5c6 d1b3 d8e7 f1c1 a8c8 b3e3 f8d8 h2h3 c8c7 b2b1 d8c8 e2d3');
INSERT INTO evaluations VALUES(45324,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(45324,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(45324,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(45324,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(45324,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(45324,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(45324,36,'cp 24','f6d5 g1f3 e7e6 e2e4 d5c3 b2c3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1d3 b8c6 d3c2 d8a5 d2a5 c6a5 e1d2 b7b6 c2d3 c8b7 a1c1 a8c8 h2h4 f7f6 d2e3 c8c1 h1c1 f8c8 c1c8 b7c8 f3d2 h7h6 f2f4 g8f8 d3b5 f8e7 g2g4 e6e5 d4e5 c8g4 e5f6 g7f6 d2b1 e7d6 b1c3 a5b7 a2a3 g4d7 b5d7');
INSERT INTO evaluations VALUES(45500,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45500,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(45500,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(45500,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(45500,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45500,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(45500,36,'cp 37','g1f3 e7e6 e2e4 d5c3 b2c3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1d3 b8c6 d3c2 d8a5 d2a5 c6a5 e1d2 b7b6 c2d3 c8b7 a1c1 a8c8 d2e3 f7f6 h2h4 c8c1 h1c1 f8c8 c1c8 b7c8 f3d2 g8f7 f2f4 c8d7 g2g4 e6e5 d4e5 d7g4 e5f6 g7f6 d2b1 a5c6 b1c3 c6b4 d3b5 f7g6');
INSERT INTO evaluations VALUES(14263,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(14263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14263,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14263,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(14263,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(14263,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(14263,36,'cp -12','c8f5 c1f4 e7e6 e2e3 f8d6 f1d3 d6f4 e3f4 e8g8 d3f5 e6f5 f3e5 a7a5 a2a4 f6e4 c3e2 d8h4 e1g1 c6d8 c2c4 d8c6 b2b3 d5c4 b3c4 a8d8 e5f3 h4e7 d4d5 c6b4 e2d4 f8e8 d4f5 e7f6 f3d4 g7g6 f5h6 g8g7 h6g4 f6f4');
INSERT INTO evaluations VALUES(24850,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24850,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24850,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24850,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24850,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24850,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24850,36,'cp 0','c1f4 e7e6 e2e3 f8d6 f1b5 e8g8 f3e5 c6b4 a1c1 c7c6 b5e2 c6c5 d4c5 d6c5 a2a3 b4c6 e5c6 b7c6 c3a4 c5d6 f4d6 d8d6 c2c4 a7a5 e1g1 f8b8 a4b6 a8a7 b6a4');
INSERT INTO evaluations VALUES(24810,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24810,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24810,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24810,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24810,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24810,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24810,36,'cp -2','e7e6 e2e3 f8d6 f1d3 d6f4 e3f4 e8g8 d3f5 e6f5 f3e5 a7a5 a2a4 f8e8 e1g1 f6e4 c3e2 d8h4 f2f3 e4d6 b2b3 e8e7 d1d2 a8e8 a1e1 g7g6 g1h2 c6d8 g2g3 h4h6 h3h4 f7f6 e5d3 b7b6 e2c3 c7c6 e1e7 e8e7 f1e1 e7e1 d2e1');
INSERT INTO evaluations VALUES(21883,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(21883,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(21883,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(21883,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(21883,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(21883,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(21883,36,'cp 117','e5d6 f8d6 f1d3 d8e7 b1c3 g8f6 c1d2 e8g8 d2g5 c8e6 g1f3 h7h6 g5h4 c6e5 e1g1 e5g6 h4g3 d6g3 h2g3 e7c5 d1e1 f8e8 a2a3 a8d8 b2b4 c5h5 f3h2 g6e5 d3e2 f6g4 h2g4 e5g4 e2g4 e6g4 e1e3 h5e5 a1e1 g4e6 f2f4');
INSERT INTO evaluations VALUES(21841,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(21841,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(21841,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(21841,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(21841,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(21841,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(21841,36,'cp 96','f8d6 b1c3 g8f6 f1d3 d8e7 c1d2 c8e6 h2h3 e8c8 g1f3 f6d7 d1e2 d7c5 d2e3 h8e8 a1d1 h7h6 e1g1 c5d3 c2d3 f7f5 e4f5 e6f5 d3d4 g7g5 d4d5 c6e5 f3d4 f5d7 c3e4 e5g6 e4d6 e7d6 e2c2 g6f4 d4f5 d7f5 c2f5 d6d7 f5f6 f4h3 g1h2 h3f4 e3f4 g5f4 f6f4 d7g7');
INSERT INTO evaluations VALUES(5909,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(5909,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(5909,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(5909,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(5909,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(5909,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(5909,36,'cp 92','f1d3 d8e7 b1c3 g8f6 c1d2 c8e6 g1f3 e8c8 h2h3 f6d7 d1e2 d7c5 a1d1 f7f6 d2e3 g7g5 d3c4 c8b8 c4d5 e6d7 e1g1 g5g4 h3g4 d7g4 d5c6 b7c6 e4e5 f6e5 e3c5 d6c5 d1d8 h8d8 e2e4 g4f3 e4f3 d8d2');
INSERT INTO evaluations VALUES(38855,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(38855,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(38855,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(38855,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(38855,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(38855,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(38855,36,'cp 127','f1d3 g8f6 e1g1 e8g8 f3e5 b4e7 c1e3 b8c6 e5c6 b7c6 c3a4 f6d5 d1e2 e7d6 f2f4 d8f6 b2b3 g7g6 a1d1 d5b6 a4c5 b6d5 e2f3 a7a5 d3e4 f6g7 c2c4 f7f5 e4c2 d5e3 f3e3 f8d8 g2g3 d6c5 d4c5 d8d1');
INSERT INTO evaluations VALUES(45351,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(45351,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(45351,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(45351,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45351,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(45351,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(45351,36,'cp 39','c4d5 c5d4 d1d4 d8d5 b1c3 d5d4 f3d4 a7a6 g2g3 e7e5 d4b3 a6a5 f1g2 b8c6 e1g1 a5a4 b3d2 c8e6 c3b5 a8c8 b2b3 a4b3 a2b3 f8b4 d2c4 e8g8 f1d1 f6d7 c1e3 d7c5 a1b1 c5b3 c4b6 c8d8 d1d8 f8d8 b5c7 b3d2 e3d2 b4d2 c7e6 f7e6 e2e3 g8f7 g2c6 b7c6 b6c4 d2c3');
INSERT INTO evaluations VALUES(45326,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(45326,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(45326,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45326,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45326,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(45326,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(45326,36,'cp 33','c5d4 f3d4 f6d5 e2e4 d5f6 b1c3 e7e5 f1b5 c8d7 b5d7 d8d7 d4c2 d7d1 e1d1 b8c6 d1e2 f8c5 h1d1 a8c8 c1g5 f6d7 g5e3 c5e3 c2e3 c6d4 e2f1 g7g6 a1c1 e8e7 c3d5 e7d6 c1c3 h8d8 f2f4 c8c3 b2c3 d4c6 f4f5 g6f5 e3f5 d6c5 f5h6 d7b6 h6f7 d8d7 f7g5 b6d5');
INSERT INTO evaluations VALUES(10020,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(10020,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(10020,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10020,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10020,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(10020,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10020,36,'cp -12','c6d4 f3e5 d8e7 e5g4 d4b5 c3b5 e7e4 d1e2 e8d8 e2e4 f6e4 d2d3 e4d6 c1e3 c5b4 b5c3 d6f5 e3d2 c7c6 c3e2 b4c5 e1c1 d7d5 h1e1 f7f6 g4e3 h7h5 a2a4 c5d6 c2c4 f5e3 f2e3 h5h4 c4d5 c6d5 d2c3 c8d7');
INSERT INTO evaluations VALUES(21986,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(21986,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21986,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21986,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21986,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(21986,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(21986,36,'cp 6','e5f4 d2d4 g7g5 d4d5 c6e5 f1b5 c8d7 b5d7 e5d7 h2h4 f8g7 h4g5 d8e7 d1e2 e8c8 c1f4 g7c3 b2c3 d8e8 e1c1 e7e4 e2e4 e8e4 d1e1 g8f6 g5f6 e4f4 h1h7 h8f8 h7f7 f4f6 f7h7 d7b6 e1e7 b6d5 e7g7 f6f4 f3d4 c8b8 d4e6 f4f1 c1b2 f8c8 c3c4 d5f6 h7h4 a7a6 g2g4 f1e1 e6f4 e1g1 f4e2 g1g2 e2d4 b8a7 d4f5 f6e4 h4h7 e4c5 g7c7 c8c7 h7c7');
INSERT INTO evaluations VALUES(22052,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(22052,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(22052,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(22052,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(22052,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(22052,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(22052,36,'cp 34','d2d4 g7g5 d4d5 c6e5 f1b5 c8d7 b5d7 e5d7 h2h4 f8g7 h4g5 d8e7 e1f2 f7f6 c1f4 e8c8 g5g6 h7g6 h1h8 g7h8 d1d4 g6g5 f4d2 g8h6 d4a7 f6f5 d2g5 h6g4 f2f1 h8f6 g5f6 g4f6 e4f5 d8h8 a1e1 h8h1 f1f2 h1e1 a7a8 d7b8 f3e1 f6g4 f2g3 g4e3 e1f3 e3f5 g3f4 f5h4 c3b5 h4f3 b5a7 c8d7 a8b8 e7g5 f4f3 g5d5 f3e3 d5e5 e3f2 e5d4 f2e2 d4e4 e2d1 e4d5 d1c1');
INSERT INTO evaluations VALUES(22033,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(22033,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22033,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22033,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(22033,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22033,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22033,36,'cp 0','g7g5 d4d5 c6e5 f1b5 c8d7 b5d7 e5d7 h2h4 f8g7 h4g5 d8e7 e1f2 e8c8 c1f4 g7c3 b2c3 e7e4 d1d2 f7f6 a1e1 e4g6 d2d4 d8f8 e1e6 g6c2 e6e2 c2g6 f4d6 c7d6 d4a7 f6g5 e2b2 b7b6 h1e1 g5g4 b2b6 d7b6 a7b6 g4f3 b6a6 c8c7 a6a7 c7c8');
INSERT INTO evaluations VALUES(31246,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(31246,31,'cp 265',NULL);
INSERT INTO evaluations VALUES(31246,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(31246,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(31246,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(31246,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(31246,36,'cp 245','c4f7 e8f7 f3g5 f7e8 d1g4 d8d7 g4d7 b8d7 f2f4 h7h6 g5f3 e7e6 b1c3 a7a6 d2d3 g8e7 a2a4 e7c6 c1d2 f8e7 h2h4 e8f7 a4a5 e7d8 c3d1 d8f6 d1f2 b7b5 a5b6 f6b2 a1b1 b2f6 e1g1');
INSERT INTO evaluations VALUES(54610,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(54610,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(54610,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54610,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54610,36,'cp 0','d2d4 e5d4 d1d4 g8f6 d4d3 b8c6 g1f3 f8b4 c2c3 d8e7 d3e3 b4c5 e3e7 e8e7 e1g1 h8e8 h2h3 c8f5 c1g5 a8d8 b1d2 f5d3 f1e1 e7f8 a1d1 e8e1 d1e1 d5d4 g5f6 g7f6 d2e4 d3e4 e1e4 d4c3 b2c3 d8d3 e4f4 f8e7 f4e4');
INSERT INTO evaluations VALUES(54604,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(54604,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54604,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54604,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54604,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(54604,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54604,36,'cp 0','e5d4 d1d4 g8f6 c1g5 f8e7 b1c3 b8c6 d4f4 e8g8 g1f3 d5d4 g5f6 e7f6 c3e4 d8a5 c2c3 f6d8 e1g1 d8c7 f4h4 c7d8');
INSERT INTO evaluations VALUES(4188,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(4188,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4188,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(4188,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(4188,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(4188,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4188,36,'cp 11','b1c3 a7a6 c3d5 d7d6 c2c3 c5a7 d5f6 d8f6 c4b3 c6e7 c1e3 c8e6 d3d4 f6g6 b3e6 g6e6 f1e1 e7g6 d1b3 e6b3 a2b3 f8e8 a1d1 a6a5 g2g3 e5d4 e3d4 a7d4 f3d4 g6e5 g1g2 e5c6 d4f3 g8f8 d1d5 c6e7 d5b5 b7b6');
INSERT INTO evaluations VALUES(15588,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(15588,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(15588,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(15588,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(15588,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(15588,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(15588,36,'cp 41','d3d4 e5d4 c3d4 d6d5 e4e5 f6e4 f1b5 f8b4 c1d2 e8g8 e1g1 b4e7 b1c3 a7a6 b5a4 c8g4 d2f4 d8d7 f4e3 e4c3 b2c3 b7b5 a4c2 g4f5 a2a4 f5c2 d1c2 c6a5 f1e1 a5c4 e3g5 c7c6 h2h3 a6a5 g5e7 d7e7 f3d2 c4d2 c2d2');
INSERT INTO evaluations VALUES(16906,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(16906,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(16906,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(16906,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(16906,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(16906,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(16906,36,'cp 126','e4e5 f6d5 e1g1 d5b6 f1e1 a7a6 b5c4 b6c4 d3c4 d7d5 e5d6 c8e6 d6c7 d8c7 d1d4 f8d6 b1c3 e8g8 c3d5 e6d5 c4d5 c7c2 c1e3 a8d8 g2g3 c2f5 a1d1 f8e8 g1g2 d8c8 h2h3 h7h5 e1e2 f7f6 a2a4 g8f7 d4b6 f5e4 g2g1 e4b4 b6b4 d6b4');
INSERT INTO evaluations VALUES(34841,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(34841,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(34841,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(34841,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(34841,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(34841,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(34841,36,'cp 153','c3b5 f8d6 c2c4 e5f6 d2d4 a7a6 b5d6 c7d6 g1f3 f6d8 d1a4 c8d7 a4b3 d6d5 b3b7 d7c6 b7b3 d5c4 e2c4 h7h6 b3c3 g8e7 e1g1 b8d7 d4d5 e6d5 c4e2 e8g8 f3d4 d7e5 c1f4 e5g6 f4e3 c6d7 a1d1 a8c8 c3a3 e7f5 d4f5 d7f5');
INSERT INTO evaluations VALUES(54237,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(54237,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(54237,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(54237,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(54237,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54237,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(54237,36,'cp 7','e4d5 e6d5 g1f3 c5c4 d3e2 f8e7 b2b3 c4b3 a2b3 g8f6 e1g1 e8g8 e2d3 b8c6 f1e1 c8g4 c2c3 g4h5 c1g5 f8e8 b1d2 a7a5 h2h3 h5g6 d2f1 f6e4 g5e7 c6e7 a1c1 e7c6 d1c2 d8b6 f1e3 e4f6 c1d1 a8c8 e3g4 f6g4 h3g4 g6d3 d1d3 h7h6');
INSERT INTO evaluations VALUES(54130,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54130,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54130,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54130,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(54130,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54130,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54130,36,'cp 8','e6d5 g1f3 c5c4 d3e2 g8f6 b2b3 c4b3 a2b3 f8e7 e1g1 e8g8 e2d3 b8c6 f1e1 c8g4 c2c3 a7a5 c1f4 g4h5 b1d2 h5g6 d2f1 f6e4 a1c1 f8e8 d1c2 e7d6 f4d6 d8d6 f1e3 e4f6 c2d2 g6d3 d2d3 d6d7 d3b5 e8c8 c1c2 c6e7 b5d7 f6d7 g2g3');
INSERT INTO evaluations VALUES(27063,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(27063,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(27063,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(27063,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(27063,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(27063,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(27063,36,'cp 41','c4e6 f7e6 c2c3 g8f6 b2b4 a7a6 e1g1 f8e7 c1e3 e8g8 b1d2 f6g4 a2a4 g4e3 f2e3 e7g5 d1e2 g5h6 h2h3 d8e7 f3h2 f8f1 a1f1 c6d8 d2f3 d8f7 c3c4 a8f8 a4a5 h6g5 f3g5 f7g5 f1f8 g8f8 b4b5 f8g8 e2g4 g5f7 h2f3 h7h6 g4g6');
INSERT INTO evaluations VALUES(27047,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(27047,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(27047,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(27047,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(27047,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(27047,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(27047,36,'cp 32','f7e6 c2c3 g8f6 b2b4 a7a6 e1g1 f8e7 b1d2 e8g8 f1e1 h7h6 a2a4 d8e8 d2c4 f6d7 c1d2 e8f7 b4b5 c6d8 d2e3 f7g6 c4d2 d8f7 b5a6 b7a6 a1b1 e7g5 b1b7 g5e3 e1e3 f8c8 d3d4 d7f6 d4e5 f6g4 e3e1 g4e5 f3e5 f7e5 f2f4');
INSERT INTO evaluations VALUES(27130,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(27130,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(27130,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(27130,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(27130,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(27130,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(27130,36,'cp 50','c2c3 g8f6 b2b4 a7a6 e1g1 f8e7 b1d2 e8g8 f1e1 h7h6 a2a4 d8e8 d2c4 f6d7 c1d2 e8f7 b4b5 c6d8 d2e3 f7g6 c4d2 a6b5 a4b5 d8f7 a1a8 f8a8 d1b3 f7g5 b3c4 g5f3 d2f3 e7d8 c4b3 d8e7 h2h3 g8h7 e1f1 e7d8 c3c4 g6e8 b3b2 b7b6 b2e2 e8f7 f3h2');
INSERT INTO evaluations VALUES(30810,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(30810,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(30810,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(30810,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(30810,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(30810,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(30810,36,'cp -84','b7e4 c1g5 g8f6 b1c3 e4b7 g5f6 g7f6 d4d5 c7c6 f1c4 c6d5 c4d5 b8c6 d1e2 e7e6 e1c1 f8g7 d5c6 b7c6 f3d4 f6f5 c1b1 d8f6 d4c6 d7c6 e2f3 a8c8 g2g4 e8g8 g4f5 f8d8 d1d8 c8d8 f3c6 f6f5 h1d1 d8d1 c3d1 g7f6 c6d6 g8g7');
INSERT INTO evaluations VALUES(30449,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(30449,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(30449,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(30449,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(30449,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(30449,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(30449,36,'cp -101','c2c4 e7e6 a2a3 g8f6 b1c3 e4b7 d4d5 g7g6 c1g5 f8g7 f1d3 e8g8 e1g1 h7h6 g5h4 e6d5 c4d5 d7d6 d3c4 f8e8 d1c2 g6g5 h4g3 f6h5 a1d1 b8d7 h2h3 h5g3 f2g3 d8e7 c4b5 a7a6 b5c6 b7c6 d5c6 d7e5 f3e5');
INSERT INTO evaluations VALUES(38140,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(38140,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(38140,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(38140,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(38140,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(38140,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(38140,36,'cp -41','f7f6 b1c3 a7a6 f1d3 b8c6 d1e2 d8c7 c1f4 g8e7 g2g3 f6f5 c3b1 h7h6 h2h4 e8g8 b1d2 c6b4 e1g1 c5a7 d2b3 b4d3 e2d3 c8d7 b3d4 e7g6 a2a4 a8c8 c2c3 g6f4 g3f4 c7b6 d3d2 b6d8 g1g2 a7d4 c3d4 c8c4 b2b3');
INSERT INTO evaluations VALUES(595,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(595,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(595,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(595,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(595,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(595,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(595,36,'cp 89','f4d2 c6e5 f3e5 e7e5 e2e3 g4f6 f1d3 c7c6 a2a3 b4c5 d1c2 e8g8 c3e2 e5e7 d2c3 h7h6 e1g1 f8d8 b2b4 c5b6 c2b2 b6c7 c3f6 e7f6 b2f6 g7f6 e2d4 c7e5 f2f4 d7d6 g1f2 e5d4 e3d4 f6f5 f1c1 c8e6 d4d5 c6d5 c4d5 e6d5 d3f5 d5e6 f5e6 f7e6');
INSERT INTO evaluations VALUES(30616,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(30616,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(30616,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(30616,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(30616,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(30616,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30616,36,'cp 27','c7c5 b1c3 c5d4 f3d4 a7a6 c1e3 b8c6 d4c6 b7c6 f1d3 b6b5 e1g1 f7f5 c3e2 d8c7 e2d4 c6b7 f1e1 g8e7 c2c3 e7g6 f2f4 f8c5 d1d2 e8g8 b2b4 c5b6 e3f2 h7h6 a2a4 b5a4 a1a4 a8c8');
INSERT INTO evaluations VALUES(30482,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(30482,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(30482,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(30482,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(30482,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30482,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(30482,36,'cp 24','b1c3 c5d4 f3d4 b8c6 d4c6 b7c6 f1d3 g8e7 d1e2 e7g6 f2f4 f8b4 h1f1 d7d6 c1d2 d6e5 e1c1 e8g8 f4e5 d8c7 f1e1 a8d8 g2g3 b4c5 c1b1 c7b7 c3e4 d8d5 c2c4 d5d7 d2c3 c6e4 d3e4 d7d1 e1d1 b7c7 b1a1 f8d8 d1d8 c7d8 e4c2');
INSERT INTO evaluations VALUES(41285,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(41285,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(41285,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(41285,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(41285,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(41285,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(41285,36,'cp 11','f6g4 e3c1 c7c5 d4d5 g7d4 g1h3 a7a6 f1e2 g4f6 d1d3 b8d7 d3c2 d7e5 h3g5 e8g8 f2f4 e5g4 c3d1 f6h5 e2g4 c8g4 d1f2 g4d7 g5f3 f7f5 f3d4 c5d4 c2d3 b7b5 d3d4 f5e4 e1g1 b5c4 c1d2 d7f5 a1e1 h5f6 h2h3 a6a5 d2c3 e4e3');
INSERT INTO evaluations VALUES(48510,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(48510,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(48510,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(48510,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(48510,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(48510,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(48510,36,'cp -81','c1f4 f8d6 e1c1 b8c6 f4g5 e8g8 g1e2 d6e7 g5f4 f6d5 c3d5 d8d5 f3d5 e6d5 e2c3 c8g4 d1d2 e7b4 f4c7 a8c8 c7g3 f8d8 d2f2 b4c3 b2c3 c6a5 c1b2 f7f6 f1d3 a5c4 d3c4 c8c4 h1e1 g8f7 h2h3 g4h5 g3f4 h5g6 g2g4 g6e4 g4g5');
INSERT INTO evaluations VALUES(48503,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(48503,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(48503,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(48503,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(48503,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(48503,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(48503,36,'cp -79','f8d6 e1c1 b8c6 f4g5 e8g8 g1e2 d6e7 h2h4 f6d5 f3g3 e7d6 g3f2 d8e8 c3e4 f7f6 g5d2 c6b4 e4d6 e8c6 d2b4 d5b4 e2c3 c6d6 c1b1 b4d5 c3d5 e6d5 f2f3 d6d7 a2a3 c7c6 f1d3 d7d6');
INSERT INTO evaluations VALUES(57748,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(57748,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(57748,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(57748,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(57748,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(57748,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(57748,36,'cp 101','d2d4 d7d6 g1f3 a7a6 f1d3 b8c6 b1d2 c6b4 d3c4 c8f5 c4b3 e7e6 e1g1 d6d5 a2a3 b4c6 b3a4 b7b5 a4b3 c6a5 b3a2 a5c4 f1e1 c4d2 c1d2 c7c5 a2b1 c5c4 a3a4 b5b4 c2c3 b4c3 b2c3 g8e7 b1f5 e7f5');
INSERT INTO evaluations VALUES(57746,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(57746,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(57746,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(57746,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(57746,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(57746,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(57746,36,'cp 107','d7d6 g1f3 d6e5 f3e5 c8e6 c1f4 b8d7 c2c4 g7g5 e5d7 d8d7 f4e3 f8g7 b1c3 e6f5 d1d2 g5g4 f1e2 g8f6 a1d1 e8g8 e1g1 c7c6 e3h6 a8d8 h6g7 g8g7 f2f3 g4f3 e2f3 f5g6 d2e3 d7d6 b2b3 b7b6 g1h1 f8e8 f1e1 e7e6 h2h3');
INSERT INTO evaluations VALUES(54949,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(54949,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(54949,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(54949,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(54949,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(54949,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(54949,36,'cp 121','g1f3 d6e5 f3e5 c8e6 b1c3 g8f6 f1d3 c7c6 c3e2 b8d7 e1g1 f6d5 a2a3 g7g6 c2c4 d7e5 d4e5 d5b6 d1c2 b6d7 e2f4 d7e5 f4e6 f7e6 d3e2 f8g7 c1e3 e8g8 a1d1 d8a5 f2f4 e5f7 e2g4 a8d8 g4e6 d8d1 f1d1 f8d8 b2b4 d8d1 c2d1 a5a3 e3c5 a3a1 d1a1 g7a1');
INSERT INTO evaluations VALUES(15434,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15434,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(15434,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(15434,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(15434,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(15434,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(15434,36,'cp 65','d2d4 e5d4 f3d4 c8d7 b1c3 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 e1g1 f8e8 d1f3 e7f8 c1d2 d8b8 a1b1 a7a5 g1h1 b8b7 f1e1 a5a4 e1e2 f6d7 b2b3 a4b3 a2b3 h7h6 d2e1 d7c5 e4e5 d6e5 f4e5 a8d8 e1g3 b7b4 b1f1');
INSERT INTO evaluations VALUES(18327,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(18327,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(18327,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(18327,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(18327,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(18327,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(18327,36,'cp 115','f1c4 f8c5 e1g1 g8f6 c3d5 e8g8 c2c3 c6a5 d5f6 d8f6 c4e2 a5c6 d2d4 c5d6 f1e1 h7h6 a2a3 f8e8 d4d5 c6e7 c1e3 c7c6 c3c4 c6d5 c4d5 d6c5 e3c5 b6c5 d1a4 e7g6 e2f1 e8e7 b2b4 c5b4 f3d2 f6g5 a4b4 d7d6 g2g3 c8d7');
INSERT INTO evaluations VALUES(18316,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(18316,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(18316,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(18316,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(18316,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(18316,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(18316,36,'cp 108','f8c5 c3d5 g8f6 e1g1 e8g8 c2c3 c6a5 d5f6 d8f6 c4e2 a5c6 d2d4 c5d6 f1e1 a7a5 e2c4 c8a6 c1g5 f6g6 c4d5 a8e8 a2a3 g8h8 g5h4 g6h6 h4g3 f7f6 b2b4 a6c8 f3h4 g7g6 d5a2 h6g7 d1d2 c8b7 a1d1 c6d8 b4a5');
INSERT INTO evaluations VALUES(28887,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(28887,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(28887,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(28887,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(28887,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(28887,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(28887,36,'cp 26','e1g1 e7f5 f1d1 f8e7 b1d2 b8d7 c2c3 d8b6 g2g4 f5h4 f3h4 e7h4 d3h3 h4e7 f2f4 c6c5 d2f3 c5d4 c3d4 a8c8 f4f5 e8g8 g1h1 c8c2 c1g5 e7g5 f3g5 h7h6 g5f3 c2f2 f5f6 g7f6 d1g1 f2f3 h3f3 b6d4 e5f6 d4f6 f3f6 d7f6 g4g5 f6d7 g5h6 g8h7 a1e1 h7h6 a2a3 e6e5');
INSERT INTO evaluations VALUES(28888,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(28888,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(28888,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28888,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(28888,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(28888,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(28888,36,'cp 38','b8d7 c2c4 d5c4 d3c4 e7d5 b1c3 f8e7 c4b3 d7b6 c3e4 a7a5 c1g5 e7g5 f3g5 e8g8 a2a3 d5e7 g5f3 a5a4 b3c2 e7f5 f1d1 h7h6 a1c1 d8e7 g2g3 f8d8 d1d3 d8d7 e4c5 d7d5 h2h4 a8d8 c2d2 d8a8 d2c3 d5d8 g1g2');
INSERT INTO evaluations VALUES(29451,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(29451,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(29451,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(29451,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(29451,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(29451,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(29451,36,'cp 157','f2f3 g4c8 c1e3 c7c6 d1d2 b8d7 a2a4 d8a5 g2g4 d7b6 g1h3 e7e5 d4e5 d6e5 c3b1 a5b4 d2b4 f8b4 c2c3 b4d6 b1d2 b6d7 d2c4 d6c7 a4a5 d7f8 e1c1 h7h6 h3f2 c8e6 h2h4 f6d7 c4d6 c7d6 d1d6 e8e7');
INSERT INTO evaluations VALUES(29449,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(29449,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(29449,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(29449,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(29449,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(29449,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(29449,36,'cp 127','g4c8 c1e3 c7c6 g2g4 g7g6 g1e2 f8g7 f1g2 b8a6 a2a4 e7e5 d1d2 d8a5 a1d1 e8g8 e1g1 c8e6 h2h3 e6c4 b2b3 c4e2 c3e2 a5d2 d1d2 f8e8 g1f2 e5d4 e3d4 d6d5 g4g5 f6h5 e4d5 g7d4 d2d4 c6d5 f1d1 a8c8 d1d2 a6c7 f3f4 c7e6');
INSERT INTO evaluations VALUES(35551,30,'cp 542',NULL);
INSERT INTO evaluations VALUES(35551,31,'cp 549',NULL);
INSERT INTO evaluations VALUES(35551,32,'cp 538',NULL);
INSERT INTO evaluations VALUES(35551,33,'cp 551',NULL);
INSERT INTO evaluations VALUES(35551,34,'cp 551',NULL);
INSERT INTO evaluations VALUES(35551,35,'cp 551',NULL);
INSERT INTO evaluations VALUES(35551,36,'cp 529','h5e5 e7f7 f1c4 d7d5 c4d5 f7g6 h2h4 h7h6 d5b7 f8d6 e5a5 g8e7 b7a8 b8d7 d2d4 c7c6 a5c3 d8b6 e1g1 c8a6 e4e5 d6c7 c3f3 h8a8 f3e4 g6f7 e5e6 f7f8 e6d7 a6f1 g1f1 a8d8 b1d2 d8d7 d2f3 c6c5 c2c3 f8f7 c1f4 c7f4 e4f4 f7g8 d4c5 b6b2 a1e1 b2a2 f4g4 a2d5 f1g1 e7f5');
INSERT INTO evaluations VALUES(35549,30,'cp 556',NULL);
INSERT INTO evaluations VALUES(35549,31,'cp 546',NULL);
INSERT INTO evaluations VALUES(35549,32,'cp 550',NULL);
INSERT INTO evaluations VALUES(35549,33,'cp 551',NULL);
INSERT INTO evaluations VALUES(35549,34,'cp 551',NULL);
INSERT INTO evaluations VALUES(35549,35,'cp 542',NULL);
INSERT INTO evaluations VALUES(35549,36,'cp 536','e7f7 f1c4 d7d5 c4d5 f7g6 h2h4 h7h6 d5b7 f8d6 e5a5 g8e7 b7a8 c7c6 a5c3 b8d7 d2d4 d8b6 e1g1 c8a6 e4e5 d6c7 c3a3 b6d4 a3e7 d4e5 e7e5 d7e5 f1e1 h8a8 c1f4 e5f3 g2f3 c7f4 b1c3 a8f8 e1e7 a6c8 a1d1 c8h3 c3e2 f4b8 h4h5 g6h7 e2d4 c6c5 d4c6 c5c4 d1d4 a7a6');
INSERT INTO evaluations VALUES(35538,30,'cp 566',NULL);
INSERT INTO evaluations VALUES(35538,31,'cp 558',NULL);
INSERT INTO evaluations VALUES(35538,32,'cp 559',NULL);
INSERT INTO evaluations VALUES(35538,33,'cp 535',NULL);
INSERT INTO evaluations VALUES(35538,34,'cp 557',NULL);
INSERT INTO evaluations VALUES(35538,35,'cp 576',NULL);
INSERT INTO evaluations VALUES(35538,36,'cp 586','f1c4 d7d5 c4d5 f7g6 h2h4 h7h6 d5b7 f8d6 e5a5 g8e7 b7a8 c7c6 a5d8 h8d8 d2d4 b8d7 d4d5 c6d5 b1c3 d5d4 c3d5 c8a6 d5e7 d6e7 a8c6 d7e5 c6d5 d8c8 d5b3 e5c4 h4h5 g6h7 e1g1 c8f8 g2g3 c4e5 f2f4 e7c5 c1d2 e5c4 b3c4 a6c4 g1g2 c4f7 f4f5 f7h5 e4e5 h5g4 f5f6 g7f6');
INSERT INTO evaluations VALUES(35536,30,'cp 572',NULL);
INSERT INTO evaluations VALUES(35536,31,'cp 576',NULL);
INSERT INTO evaluations VALUES(35536,32,'cp 553',NULL);
INSERT INTO evaluations VALUES(35536,33,'cp 578',NULL);
INSERT INTO evaluations VALUES(35536,34,'cp 586',NULL);
INSERT INTO evaluations VALUES(35536,35,'cp 575',NULL);
INSERT INTO evaluations VALUES(35536,36,'cp 578','d7d5 c4d5 f7g6 h2h4 h7h6 d5b7 f8d6 e5a5 g8e7 b7a8 c7c6 a5d8 h8d8 d2d4 b8d7 d4d5 c6d5 b1c3 d5d4 c3d5 c8a6 d5e7 d6e7 a8c6 d7e5 c6d5 d8c8 d5b3 e5c4 h4h5 g6h7 e1g1 c8f8 g2g3 c4e5 f2f4 e5g4 c1d2 a6b7 f1e1 g4f6 b3c4 f6h5 g1g2 h5f6 c4d3 f6g4 g2g1 h6h5 e4e5 g7g6 d3e4 f8b8');
INSERT INTO evaluations VALUES(35528,30,'cp 578',NULL);
INSERT INTO evaluations VALUES(35528,31,'cp 589',NULL);
INSERT INTO evaluations VALUES(35528,32,'cp 578',NULL);
INSERT INTO evaluations VALUES(35528,33,'cp 578',NULL);
INSERT INTO evaluations VALUES(35528,34,'cp 591',NULL);
INSERT INTO evaluations VALUES(35528,35,'cp 575',NULL);
INSERT INTO evaluations VALUES(35528,36,'cp 600','c4d5 f7g6 h2h4 h7h6 d5b7 f8d6 e5a5 g8e7 b7a8 c7c6 a5d8 h8d8 d2d4 b8d7 d4d5 c6d5 b1c3 d5d4 c3d5 c8a6 d5e7 d6e7 a8c6 d7e5 c6d5 d8c8 d5b3 e5c4 h4h5 g6h7 e1g1 e7b4 f1d1 b4c5 e4e5 c4e5 d1e1 e5c4 e1e6 c5b6 c1f4 c8f8 f4g3 f8f5 e6e7 f5g5 g3f4 g5h5 f4c7 b6c5 e7e6');
INSERT INTO evaluations VALUES(35527,30,'cp 582',NULL);
INSERT INTO evaluations VALUES(35527,31,'cp 586',NULL);
INSERT INTO evaluations VALUES(35527,32,'cp 582',NULL);
INSERT INTO evaluations VALUES(35527,33,'cp 582',NULL);
INSERT INTO evaluations VALUES(35527,34,'cp 586',NULL);
INSERT INTO evaluations VALUES(35527,35,'cp 578',NULL);
INSERT INTO evaluations VALUES(35527,36,'cp 548','f7g6 h2h4 h7h6 d5b7 f8d6 e5a5 g8e7 b7a8 b8c6 a8c6 e7c6 a5a4 c8d7 d2d3 d8e7 c2c3 c6e5 a4c2 e5g4 c2e2 g6h7 g2g3 h8f8 f2f4 d6c5 h1f1 f8e8 f4f5 g4f6 g3g4 f6e4 d3e4 e7e4 e2e4 e8e4 e1d1 e4g4 f5f6 h7g8 b2b4 c5d6 b1d2 g7f6 f1f6');
INSERT INTO evaluations VALUES(35529,30,'cp 582',NULL);
INSERT INTO evaluations VALUES(35529,31,'cp 578',NULL);
INSERT INTO evaluations VALUES(35529,32,'cp 580',NULL);
INSERT INTO evaluations VALUES(35529,33,'cp 576',NULL);
INSERT INTO evaluations VALUES(35529,34,'cp 587',NULL);
INSERT INTO evaluations VALUES(35529,35,'cp 574',NULL);
INSERT INTO evaluations VALUES(35529,36,'cp 590','h2h4 h7h6 d5b7 f8d6 e5a5 g8e7 b7a8 b8c6 a8c6 e7c6 a5a4 c8d7 d2d3 c6b4 a4b3 d8f6 c1e3 d7e6 b3c3 f6c3 b1c3 b4c2 e1e2 c2a1 h1a1 h8b8 b2b3 d6a3 c3a4 b8b5 e3a7 b5a5 a7e3 e6g4 f2f3 g4d7 a1b1 a5a6 e3d4 g6f7 g2g4 a6c6 d4c3 c6a6 h4h5 d7b5 a4b2 c7c5 g4g5 h6g5 b2c4 a3b4');
INSERT INTO evaluations VALUES(37453,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(37453,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(37453,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(37453,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(37453,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37453,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37453,36,'cp 4','e3e4 c7c5 e4e5 f6e8 d4c5 d8a5 b1d2 a5c5 e1g1 b8c6 d2c4 d7d6 e5d6 e8d6 c1e3 c5d5 c4d6 d5d6 c2c3 c8e6 d1e2 f8d8 f1d1 d6c7 h2h3 e6d5 d3c2 e7e6 c2b3 c6e7 e3g5 d5b3 a2b3 a7a6 b3b4 d8d5 g5e7 d5d1 a1d1 c7e7 b4b5 a6b5 e2b5');
INSERT INTO evaluations VALUES(6382,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(6382,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(6382,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(6382,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(6382,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(6382,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(6382,36,'cp 115','c2c3 b4a5 d1a4 a5b6 d2d4 e5d4 c3d4 c6b8 b1c3 g8e7 e1g1 e8g8 b5c4 h7h6 c1f4 d7d6 h2h3 c8e6 d4d5 e6c8 a4c2 b8d7 f4g3 f7f6 c3a4 d7e5 f3e5 f6e5 a4b6 a7b6 g3h4 d8d7 a2a4 e7g6 h4g3 g8h8');
INSERT INTO evaluations VALUES(6380,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(6380,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(6380,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(6380,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(6380,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(6380,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(6380,36,'cp 97','b4a5 d1a4 a5b6 d2d4 g8f6 e1g1 f6e4 b5d3 d7d5 f3e5 e8g8 e5c6 b7c6 a4c6 c8f5 d3e4 f5e4 c1f4 f8e8 b1d2 e8e6 c6b5 a8b8 b5a4 e4d3 f1e1 d8e8 a4d1 e6e1 d1e1 e8e1 a1e1 f7f6 h2h4 g8f7 d2f1 b8e8 f1e3 c7c6 e1d1 d3b5 b2b3 e8e4 f4g3 h7h5 a2a4 b5a6 d1d2 e4e8 b3b4 b6d8');
INSERT INTO evaluations VALUES(6749,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(6749,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(6749,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(6749,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(6749,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(6749,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(6749,36,'cp 101','d1a4 a5b6 d2d4 g8f6 e1g1 f6e4 b5d3 d7d5 f3e5 e8g8 e5c6 b7c6 a4c6 c8f5 d3e4 f5e4 c1f4 f8e8 b1d2 e8e6 c6b5 e4g6 f1e1 d8e8 b5e8 a8e8 e1e6 e8e6 f4e3 f7f6 a2a4 c7c6 a4a5 b6c7 d2b3 c7d6 a1e1 g6d3 b3c5 d6c5 d4c5 d3b5 h2h4 e6e4 b2b4');
INSERT INTO evaluations VALUES(16686,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(16686,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(16686,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(16686,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(16686,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(16686,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(16686,36,'cp 67','d1e2 g7g5 e4e5 f6d5 b1c3 f8b4 c1d2 b4c3 b2c3 g5g4 e2e4 g4f3 e4d5 f3g2 d5g2 d7d6 e5d6 d8d6 g2g7 d6e7 f1e2 h8f8 d2f4 c8e6 f4g5 f7f6 g7e7 c6e7 g5d2 e7f5 e1c1 e8c8 e2d3 f5d6 a2a4 e6c4 d1e1 f8g8 h1f1 c4d3 c2d3 g8g2 f1f6 g2h2 c3c4 d8g8 c4c5 g8g2');
INSERT INTO evaluations VALUES(30617,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(30617,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(30617,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(30617,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(30617,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(30617,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(30617,36,'cp 28','b1c3 d7d6 g1f3 b8c6 f1d3 d6e5 d4e5 f8b4 e1g1 g8e7 d1e2 b4c3 b2c3 d8d7 a2a4 h7h6 f1d1 e8c8 c1e3 c8b8 c3c4 c6a5 c4c5 g7g5 c5b6 c7b6 e3d2 d7d5 h2h3 e7f5 d2a5 d5a5 d3f5 e6f5');
INSERT INTO evaluations VALUES(30615,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(30615,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(30615,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30615,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(30615,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(30615,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(30615,36,'cp 30','d7d6 g1f3 b8c6 f1d3 d6e5 d4e5 c6b4 d3b5 c7c6 b5e2 d8d1 e2d1 e8c8 a2a3 b4d5 c3e4 c6c5 e4g5 d8d7 c2c4 d5c7 d1a4 d7e7 a4c2 h7h6 g5e4 e7d7 h2h4 f7f5 e4d6 f8d6 e5d6 b7f3 g2f3 d7d6 c1f4 d6d7 f4e5 a7a6 h1g1 g8f6 a1d1 d7d1 e1d1 c7e8 g1g6 f6d7 g6e6 d7e5');
INSERT INTO evaluations VALUES(46626,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46626,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(46626,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(46626,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(46626,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(46626,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(46626,36,'cp 1','f8g7 g2g3 e8g8 f1g2 a7a5 a2a4 e7e6 d5e6 c8e6 g1e2 b8a6 e1g1 c7c6 e2f4 e6g4 f2f3 g4c8 c1e3 f8e8 e3f2 d6d5 e4d5 a6b4 d5d6 g7f8 c3e4 f6e4 f3e4 d8d6 d1d6 f8d6 a1d1 d6e7 c2c3 b4a6 f4d3 e7g5 f2d4 c8e6 h2h3 a8d8 f1e1');
INSERT INTO evaluations VALUES(47834,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(47834,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(47834,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(47834,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47834,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(47834,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(47834,36,'cp 32','f1g2 c7c6 e2e4 c8e6 d2d3 f8e7 g1f3 e8g8 e1g1 a7a6 a2a4 a6a5 h2h3 f6d7 d1e2 c6c5 b2b3 b8c6 g1h2 d7b8 c3b5 b8a6 f3e1 c6d4 e2d1 e6d7 e1f3 d4c6 f3g1 a6b4 f2f4 e5f4 c1f4 c6e5 d3d4 c5d4 b5d4 e7f6 d4f5');
INSERT INTO evaluations VALUES(51513,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(51513,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(51513,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(51513,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(51513,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(51513,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51513,36,'cp 27','e2e4 c8b7 b1c3 e7e5 g1f3 g8f6 f1c4 f8b4 f3e5 d8e7 e5g4 b4c3 d2c3 e7e4 g4e3 b8c6 c4d3 e4e5 c1d2 f6e4 e1g1 e8c8 c3c4 e4d2 d1d2 e5f4 f1d1 c6e5 d3e2 f4g5 e3d5 g5d2 d1d2 d7d6 h2h4 b7d5 d2d5 c7c6 d5d4 c8c7 a1d1 h8e8 g1f1 c6c5 d4f4 h7h6');
INSERT INTO evaluations VALUES(51496,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(51496,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(51496,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(51496,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(51496,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(51496,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51496,36,'cp 25','c8b7 b1c3 e7e5 d2d4 g8f6 d4e5 f8b4 c1d2 b4c3 d2c3 f6e4 c3b2 d8e7 d1d4 f7f5 e5f6 e4f6 d4e5 e7e5 b2e5 b8c6 e5f6 g7f6 g1e2 e8c8 h2h4 d8e8 e1c1 e8e5 h4h5 c6d8 e2f4 d8e6 h1h4 f6f5 g2g3 b7f3 d1d3 e5e1 c1d2 e1f1 d3f3 h8f8 d2e3 f1a1 f4e6');
INSERT INTO evaluations VALUES(30814,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(30814,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(30814,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(30814,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(30814,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(30814,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30814,36,'cp 15','b1c3 e7e5 g1f3 g8f6 f3e5 d8e7 d2d4 f6e4 c3e4 b7e4 f1c4 b8c6 c4f7 e8d8 d1g4 c6e5 d4e5 e7e5 e1g1 f8d6 g2g3 e4c6 c1e3 h7h5 g4g5 d8c8 g5e5 d6e5 a1d1 c8b7 e3g5 a8f8 f7d5 e5c3 g1g2 f8f5 d5c6 b7c6 f2f4 h8e8 g2f3 b6b5');
INSERT INTO evaluations VALUES(55233,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(55233,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(55233,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(55233,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(55233,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(55233,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(55233,36,'cp 98','e7e5 g1f3 e5d4 d1d4 d8d4 f3d4 g8f6 d4b5 b8a6 c1f4 c7c6 b5d6 f8d6 f4d6 c8e6 e2e4 e8c8 f1a6 d8d6 e4e5 d6d8 e5f6 b7a6 f6g7 h8g8 f2f3 g8g7 e1g1 h7h5 f1d1 g7g8 g1f2 c8c7 b2b3 a6a5 h2h4 c6c5 c3e4 c5c4 d1d8 c7d8 b3c4 e6c4 a2a3 d8e7');
INSERT INTO evaluations VALUES(57767,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(57767,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(57767,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(57767,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(57767,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(57767,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(57767,36,'cp -44','e7e5 a3c4 e5e4 d2d3 d7d5 c4e3 d5d4 e3c4 c8e6 c4d2 e4d3 e2d3 g8f6 g1f3 b8c6 f1e2 f8e7 e1g1 e8g8 d2e4 h7h6 a2a3 a7a5 e4f6 e7f6 f3d2 d8d7 e2f3 a5a4 f1e1 f6e7 d2c4 a8e8 c1f4 e6c4 d3c4 e7d6 f4d6 e8e1 d1e1');
INSERT INTO evaluations VALUES(57241,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(57241,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(57241,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(57241,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(57241,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(57241,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(57241,36,'cp -51','a3c4 b8c6 e2e4 d7d5 e4d5 d8d5 c2c3 g8f6 c4e3 d5d7 f1c4 f8d6 d2d3 e8g8 c4b3 b7b6 g1f3 a7a5 e1g1 c8a6 b3a4 h7h6 f1e1 f8e8 a2a3 b6b5 a4c2 a5a4 f3d2 c6a5 d2e4 f6e4 d3e4 d7e6 e3f5 d6f8 d1g4 e6g6 g4g3 g6g3');
INSERT INTO evaluations VALUES(13369,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13369,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(13369,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(13369,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(13369,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(13369,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(13369,36,'cp 15','e7e5 d4f3 f8e7 e1g1 h7h6 e2c4 e8g8 f1e1 c6a5 c4f1 c8d7 h2h3 a7a6 b2b3 a8c8 c3d5 f6d5 e4d5 f7f5 c2c4 b7b6 c1e3 a5b7 b3b4 a6a5 a2a3 d7e8 a1b1 d8c7 f3d2 f5f4 e3b6 c7b6 c4c5 d6c5 b4c5 b6a7 d5d6');
INSERT INTO evaluations VALUES(13886,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(13886,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(13886,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(13886,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13886,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(13886,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(13886,36,'cp 15','d4f3 f8e7 e1g1 e8g8 c1g5 c8e6 g5f6 e7f6 c3d5 a8c8 c2c4 g7g6 d1d2 f6g7 a1d1 a7a6 b2b3 f8e8 d5e3 h7h5 h2h3 d8c7 e3c2 c7e7 g1h2 e8d8 c2e3 c8c7 g2g3 c7c8 h3h4 c6d4 f3d4 e5d4 e3c2 e6g4 e2g4 h5g4');
INSERT INTO evaluations VALUES(18691,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(18691,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(18691,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(18691,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(18691,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(18691,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18691,36,'cp 9','d7d5 e4d5 d8d5 e1g1 c8d7 d2d3 g8f6 c1f4 f6h5 f4e3 f8e7 d3d4 c5d4 f3d4 c6d4 b5d7 d5d7 e3d4 h5f6 d1e2 e8g8 b1d2 f6d5 a1e1 f8d8 d2f3 e7f6 d4f6 d5f6 e1d1 d7c6 g2g3 a8c8 d1d8 c8d8 f1d1 d8d1 e2d1 h7h6');
INSERT INTO evaluations VALUES(20184,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20184,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20184,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(20184,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20184,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(20184,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(20184,36,'cp 15','e4d5 d8d5 e1g1 c8d7 d2d3 g8f6 f1e1 d5d6 c1g5 a7a6 b5a4 f8e7 g5h4 d6c7 d1e2 a8d8 b1d2 e8g8 h4g3 e7d6 a1d1 h7h6 g3d6 c7d6 a4c2 c6e7 d3d4 d7c6 d4c5 d6c5 f3e5 c5b6 b2b4 e7f5 e5c6 b6c6 d2e4 f6e4');
INSERT INTO evaluations VALUES(22748,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(22748,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(22748,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(22748,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(22748,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(22748,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(22748,36,'cp 13','e5d4 f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 b1c3 c8e6 e2e4 d8c8 c1d2 a7a5 f1e2 f8b4 a2a3 b4c3 d2c3 f6e4 c3g7 h8g8 g7e5 a5a4 f2f3 a8a5 e5f4 e4d6 f4d6 c7d6 e1f2 c8c7 g2g3 a5c5 a1c1 g8e8 h1e1 e8b8 e1d1 b8d8 c1c3 d6d5 c4d5 c5d5 d1d5 d8d5 e2d3 c7d6 f2e3');
INSERT INTO evaluations VALUES(22909,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(22909,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(22909,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(22909,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22909,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22909,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(22909,36,'cp 0','f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 b1c3 c8e6 e2e4 d8c8 f1e2 f8b4 c1d2 a7a5 a2a3 b4c3 d2c3 f6e4 c3g7 h8g8 g7e5 a5a4 f2f3 a8a5 e5f4 e4d6 f4d6 c7d6 e1f2 a5c5 a1c1 g8d8 h1d1 d6d5 c4d5 c5c1 d1c1 d8d5 c1d1 d5h5 d1d4 b7b5 h2h4 c8c7 g2g3 h5c5 e2d3 h7h6 g3g4 c5c1 g4g5 h6g5 h4g5 c7b6 f3f4');
INSERT INTO evaluations VALUES(22904,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(22904,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(22904,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(22904,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22904,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22904,36,'cp 12','g8f6 d4c6 d7c6 d1d8 e8d8 b1c3 c8e6 e2e4 d8c8 f1e2 f8b4 c1d2 a7a5 a2a3 b4c3 d2c3 f6e4 c3g7 h8g8 g7e5 a5a4 f2f3 a8a5 e5f4 e4d6 f4d6 c7d6 e1f2 d6d5 c4d5 a5d5 h1d1 g8d8 d1d5 d8d5 a1d1 d5h5 h2h3 c8c7 f3f4 h5c5 d1d2 b7b5 g2g4 f7f6 h3h4 h7h6 f2e3 e6b3 h4h5 c5c1 g4g5 h6g5 f4g5 f6g5');
INSERT INTO evaluations VALUES(55908,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(55908,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(55908,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(55908,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(55908,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(55908,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(55908,36,'cp -37','e4d5 e6d5 d2d4 g8f6 c1f4 f8e7 d1d2 e8g8 b1c3 b8d7 f1d3 f8e8 g1e2 a7a6 e1g1 c7c5 d4c5 d7c5 f4e5 b7b5 e2f4 d5d4 e5f6 e7f6 c3e4 f6e7 e4c5 e7c5 f1e1 e8e1 a1e1 c8b7 g1h1 c5d6 d3e4 d8h4 g2g3 d6f4 d2f4 h4f4 g3f4 b7e4 f3e4');
INSERT INTO evaluations VALUES(1514,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(1514,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(1514,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(1514,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(1514,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(1514,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(1514,36,'cp 81','g8h6 b1c3 h6f5 c3a4 b6d8 b2b4 f8e7 h2h4 b7b6 a4c3 a7a6 h4h5 h7h6 g2g4 f5h4 f3h4 e7h4 c1e3 e8g8 a1c1 h4g5 f1d3 g5e3 f2e3 d8g5 d3b1 c8d7 d1d3 f7f5 e1d2 g5g4 c1g1 g4f3 d3e2 f3e2 c3e2 c6e7 h1h3 g8f7 h3g3 f8g8 e2f4 d7b5 b1c2 a8c8 a3a4');
INSERT INTO evaluations VALUES(827,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(827,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(827,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(827,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(827,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(827,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(827,36,'cp 87','b1c3 h6f5 c3a4 b6d8 b2b4 c8d7 c1b2 f5h4 f3h4 d8h4 f1d3 b7b5 g2g3 h4h3 d3f1 h3f5 a4c3 a7a6 h2h4 h7h5 f2f3 f5h7 f1d3 g7g6 e1f2 f8h6 f3f4 h6g7 c3e2 a6a5 b2c3 a5b4 c3b4 c6b4 a3b4 e8g8 a1c1 f8c8 d1d2 c8c1 h1c1 g7f8 e2g1 a8a4 c1c7 f8b4');
INSERT INTO evaluations VALUES(2663,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(2663,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(2663,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2663,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(2663,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(2663,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2663,36,'cp 21','c1g5 h7h6 g5h4 c5e7 a2a4 d7d6 h4f6 e7f6 c3d5 g7g6 h2h3 f6g7 c2c3 c6e7 d5e7 d8e7 f1e1 c8d7 a4a5 h6h5 d3d4 a8d8 d1b3 d7c8 b3b4 d8e8 d4e5 d6e5 e1d1 e7b4 c3b4 g7h6 d1d5 c8e6 d5c5 e6c4 c5c4 c7c6 a1d1 e8e7');
INSERT INTO evaluations VALUES(2654,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2654,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2654,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2654,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2654,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2654,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2654,36,'cp 31','h7h6 g5h4 c5e7 a2a4 d7d6 a4a5 g8h8 f1e1 c8g4 h2h3 g4f3 d1f3 f6g8 h4g3 c6d4 f3d1 f7f5 c3d5 f5f4 g3h2 e7g5 c2c3 d4c6 d3d4 g8e7 f2f3 e5d4 c3d4 e7d5 c4d5 c6e7 d5c4 d6d5 c4b3 d5e4 e1e4 e7f5 h2f4 g5f4 e4f4 d8d6 d1d2 a8e8 a1f1');
INSERT INTO evaluations VALUES(2640,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(2640,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(2640,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(2640,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2640,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(2640,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(2640,36,'cp 21','g5h4 c5e7 a2a4 d7d6 h4f6 e7f6 c3d5 g7g6 c2c3 f6g7 d5e3 g8h8 g2g3 c8h3 f1e1 d8d7 f3h4 g7f6 h4g2 a8b8 a4a5 f6g7 g2h4 c6e7 f2f4 c7c6 d1f3 d6d5 e4d5 e5f4 g3f4 c6d5 c4d5 g7f6 f4f5 e7d5 e3d5 f6h4 f3h3 d7d5 h3h4 d5f5');
INSERT INTO evaluations VALUES(2642,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2642,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(2642,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(2642,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(2642,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(2642,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2642,36,'cp 19','c5e7 a2a4 d7d6 h2h3 g7g5 h4g3 g5g4 h3g4 c8g4 d1d2 g8h7 f3h2 d8d7 c3d5 g4e6 c2c3 f8g8 a1e1 f6h5 d3d4 h5g3 f2g3 g8g3 d5e7 d7e7 d4d5 c6a5 d5e6 a5c4 d2f2 g3g7 b2b3 c4a5 e6f7 a8f8 f2f5 g7g6 h2g4 h7g7 g4e3 a5b3 f5h3 e7e6 h3h4 b3c5 e3f5 g7h7 f5e7');
INSERT INTO evaluations VALUES(16897,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(16897,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(16897,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(16897,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(16897,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(16897,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(16897,36,'cp 141','f3d4 e5d4 e1g1 c7c6 e4e5 c6b5 c1g5 f8e7 e5f6 e7f6 f1e1 e8f8 g5f6 d8f6 d1h5 b7b6 h5b5 c8b7 b5d7 b7c6 d7c7 g7g6 b1d2 f8g7 c7e5 f6e5 e1e5 a8c8 a2a4 g7f6 e5e2 h8e8 e2e8 c6e8 d2c4 e8c6 h2h4 h7h5 b2b3 c8e8 g1f1 e8e7 a1d1 f6f5 f2f3 f5f4 f1f2');
INSERT INTO evaluations VALUES(16892,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(16892,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(16892,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(16892,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(16892,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(16892,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(16892,36,'cp 120','e5d4 e1g1 c7c6 e4e5 c6b5 c1g5 f8e7 e5f6 e7f6 f1e1 e8f8 g5f6 d8f6 d1h5 b7b6 h5b5 c8b7 b5d7 b7c6 d7c7 g7g6 f2f3 f8g7 c7e7 a8c8 g1f2 h7h5 e7f6 g7f6 e1e2 h5h4 a2a4 h4h3 g2g3 f6g7 a4a5 b6b5 b1d2 c6d5 a1c1 c8c5 b2b4 c5c3 d2b1 c3c8 c1e1 a7a6 b1a3 d5e6');
INSERT INTO evaluations VALUES(17399,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(17399,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(17399,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(17399,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(17399,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(17399,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(17399,36,'cp -20','g1f3 g8f6 d2d3 d7d5 c4b3 d5e4 c3e4 f8e7 c2c3 e8g8 e4f6 e7f6 e1g1 b7b6 d1e2 c8b7 c1f4 c6e7 a1d1 a8a7 f3e5 e7g6 e5g6 h7g6 f4e3 d8a8 f2f3 b7d5 c3c4 d5c6 d3d4 c5d4 e3d4 a8b8 e2e3 f6d4 d1d4 a7d7 d4d7 c6d7 f1d1');
INSERT INTO evaluations VALUES(21996,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(21996,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21996,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(21996,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(21996,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(21996,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(21996,36,'cp 19','e2e3 g8f6 d2d3 f8e7 c2c4 e8g8 f1e2 f8e8 a2a3 e5e4 d3e4 f6e4 b1d2 c8f5 d2e4 f5e4 e2d3 e4g6 d1c2 g6d3 c2d3 e7f6 b2f6 d8f6 e1g1 a7a5 a1d1 c6e5 f3e5 e8e5 d3c3 a8e8 d1d4 h7h5 b3b4 e5e4 d4e4');
INSERT INTO evaluations VALUES(21995,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(21995,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21995,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(21995,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(21995,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(21995,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(21995,36,'cp 10','g8f6 d2d3 f8e7 f1e2 e8g8 c2c4 f8e8 e1g1 e7f8 b1c3 c6e7 c3e4 f6e4 d3e4 e7c6 f3d2 a7a5 d2b1 a5a4 b2a3 d8g5 b1c3 a4b3 a2b3 c6b8 c3d5 g5d8 d1c2 c7c6 d5c3 d8g5 g1h1 b8d7 a3b4 a8a1 f1a1 d7f6 c2d2 e8d8 f2f4 e5f4 e3f4');
INSERT INTO evaluations VALUES(15611,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(15611,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(15611,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(15611,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(15611,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(15611,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(15611,36,'cp 29','d2d3 f8e7 c2c4 e8g8 f1e2 f8e8 b1d2 a7a5 a2a3 h7h6 d1c2 c8f5 e1g1 e7f8 e3e4 f5d7 f1e1 f6h7 d3d4 c6d4 f3d4 e5d4 b2d4 h7g5 d4c3 f8e7 d2f1 e7f6 f2f3 g5e6 a1d1 f6e5 c2d2 d8f6 c3e5 f6e5 f1g3 e5g5');
INSERT INTO evaluations VALUES(29167,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(29167,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(29167,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(29167,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(29167,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(29167,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(29167,36,'cp -15','e7e6 f3e5 g4e2 c3e2 b8c6 e5d3 b7b5 c2c3 f8d6 c1g5 a7a5 a2a3 e8g8 g5h4 d6e7 e1g1 b5b4 d1a4 d8e8 a3b4 a5b4 a4a8 e8a8 a1a8 f8a8 d3e5 c6d8 c3b4 e7b4 f1c1 d8b7 g2g4 b7a5 f2f3 f6e8 h4g3 b4d6 c1a1 g7g5 e5d3 g8g7 g1f2 h7h6');
INSERT INTO evaluations VALUES(29394,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(29394,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(29394,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(29394,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(29394,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(29394,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(29394,36,'cp -17','f3e5 g4e2 c3e2 b8c6 e5d3 b7b5 c2c3 a7a5 a2a3 h7h6 c1f4 f8d6 e1g1 e8g8 f1e1 d6f4 e2f4 d8d6 f4h5 f6h5 d1h5 b5b4 a3b4 a5b4 d3c5 b4c3 b2c3 c6a5 g2g3 d6c6 h5e2 a5c4 e2g4 c4d6 g4e2 f8c8 h2h4 c8c7 g1g2 a8a1');
INSERT INTO evaluations VALUES(1511,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(1511,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(1511,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(1511,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(1511,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(1511,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(1511,36,'cp 87','b1c3 c8d7 a2a3 g8e7 c3a4 b6c7 c1f4 e7f5 g2g4 f5e7 a1c1 h7h5 h1g1 e7g6 f4d2 c7d8 g4h5 g6h4 a4c5 h4f3 d1f3 c6d4 f3g4 d4f5 c5d7 d8d7 f1d3 f5e7 e1f1 e7c6 d2f4 e8c8 h2h4 c8b8 g1g3 f8e7 b2b4 c6d4 f1g2 d4b3 c1b1 b3d4 f4e3 d4c6 b4b5 c6e5 g4d4 e5d3 d4a7 b8c7 b1b3');
INSERT INTO evaluations VALUES(1505,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(1505,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(1505,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(1505,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(1505,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(1505,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(1505,36,'cp 83','c8d7 a2a3 g8e7 c3a4 b6c7 c1f4 e7f5 g2g4 f5e7 a1c1 h7h5 h1g1 e7g6 f4d2 c7d8 g4h5 g6h4 g1g4 h4f5 a4c5 h8h5 f1d3 a8c8 e1f1 f8c5 c1c5 c6e7 c5c8 d8c8 d2g5 e7g6 h2h4 c8c6 f1g2 h5h8 d1b3 c6b6 b3b6 a7b6 g2h2 d7c6 b2b3 e8d7 h2g2 d7e8 g2h1 e8d7 a3a4 h8h5 d3f5');
INSERT INTO evaluations VALUES(15520,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(15520,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(15520,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(15520,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(15520,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(15520,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(15520,36,'cp 27','d2d3 e5f4 c1f4 f8e7 c4b3 e8g8 e1g1 c6a5 b1c3 c8d7 h2h3 b7b5 d1e2 a5b3 a2b3 a7a5 g1h1 c7c5 f4h2 b5b4 c3d1 a5a4 b3a4 a8a4 a1a4 d7a4 d1e3 a4d7 c2c3 b4c3 b2c3 f8e8 c3c4 h7h6 e2b2 e7f8 b2b7 d8c8 f1b1 c8b7 b1b7');
INSERT INTO evaluations VALUES(15517,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(15517,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(15517,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(15517,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(15517,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(15517,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(15517,36,'cp 14','e5f4 c1f4 f8e7 c4b3 e8g8 b1c3 c6a5 e1g1 c8d7 d1e1 a5b3 a2b3 c7c5 e1e2 b7b5 e4e5 d6e5 f3e5 b5b4 c3e4 f6e4 d3e4 d8e8 e5d7 e8d7 f1d1 d7b7 d1d3 a7a5 a1d1 a8d8 e4e5 b7c6 d3d8 f8d8');
INSERT INTO evaluations VALUES(23459,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(23459,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(23459,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(23459,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(23459,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(23459,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(23459,36,'cp -28','e7e5 c1b2 e5e4 f3d4 c6d4 b2d4 d7d5 e2e3 c8f5 d2d3 h7h5 d3e4 d5e4 d1d2 a7a6 d2c3 g8f6 b1d2 f8e7 f1e2 f6d5 c3b2 f7f6 c2c4 d5b4 e1g1 h5h4 c4c5 h4h3 g2g3 d8d7 a2a3 b4d3 b2c2 e8c8 b3b4 d7e6 c2c4 e6c4 d2c4 f5e6 c4a5 c8b8 f1d1');
INSERT INTO evaluations VALUES(22845,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(22845,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22845,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(22845,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(22845,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(22845,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(22845,36,'cp -19','c1b2 e5e4 f3d4 c6d4 b2d4 d7d5 e2e3 h7h5 d2d3 c8f5 d3e4 d5e4 d1d2 c7c6 d2c3 g8f6 b1d2 f8e7 f1e2 e8g8 c3b2 f8e8 h2h3 h5h4 e1g1 e7f8 d4f6 d8d2 c2c4 d2b2 f6b2 f5e6 b2d4 f8a3 f1d1 f7f6 a1b1 a3b4 d4b2 e8d8 a2a3 b4e7 d1d8 a8d8 b1d1 d8d1 e2d1 g8h8');
INSERT INTO evaluations VALUES(42690,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(42690,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42690,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(42690,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42690,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(42690,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(42690,36,'cp -27','d7d6 e2e3 e8g8 a2a3 c7c6 f1d3 a7a5 g1e2 b8a6 a1b1 c8d7 e1g1 c6d5 c4d5 d8e8 e2g3 b7b5 e3e4 b5b4 a3b4 a5b4 c3e2 h7h5 b1a1 a6c7 a1a8 e8a8 c1g5 f6h7 d1c1 h7g5 c1g5 a8d8 g5h5 c7e8 e2c1 d8b6 c1b3 g7g6');
INSERT INTO evaluations VALUES(41912,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(41912,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(41912,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(41912,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(41912,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(41912,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(41912,36,'cp -16','e2e3 a7a5 f1d3 e8g8 a2a3 c7c6 g1e2 c6d5 c4d5 b8a6 e1g1 a6c7 d3c2 b7b5 e2g3 c8g4 d1d3 b5b4 a3b4 a5b4 a1a8 d8a8 g3e4 c7e8 e4c5 b4c3 c5a4 c3b2 c1b2 a8d5 f2f4 d5d3 c2d3 f6d7 b2a3 g4e6 f1b1 f7f5 a4b6 d7b6 b1b6');
INSERT INTO evaluations VALUES(41911,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(41911,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(41911,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(41911,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(41911,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(41911,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41911,36,'cp 0','a7a5 f1d3 e8g8 a2a3 c7c6 g1e2 c6d5 c4d5 b8a6 a1b1 c8d7 e1g1 d8e8 e2g3 b7b5 b1a1 b5b4 a3b4 a6b4 d3c4 e8e7 b2b3 f8c8 h2h3 g7g6 c1d2 f6e8 c3a4 c5a7 a4c3 a7b6 c3a4');
INSERT INTO evaluations VALUES(54539,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(54539,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(54539,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(54539,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(54539,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(54539,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(54539,36,'cp -23','b1c3 g8f6 d2d3 f8e7 g1e2 e8g8 e1g1 c8e6 e2f4 b8c6 h2h3 d8d7 f4e6 f7e6 c3e2 c6a5 c2c3 a5b3 a2b3 a7a6 c1f4 f6h5 f4h2 e7d6 h2d6 d7d6 d3d4 c5d4 d1d4 e6e5 d4e3 a8d8 e3g5 h5f6 e2g3 h7h6 g5g6 g8h8 g3f5');
INSERT INTO evaluations VALUES(4435,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4435,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(4435,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4435,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4435,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(4435,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4435,36,'cp 22','e1g1 b4c3 b2c3 d7d5 e4d5 f6d5 c1d2 c8g4 f1e1 f8e8 h2h3 g4h5 c4b3 h7h6 a2a4 f7f5 a4a5 a7a6 a1a4 h5f7 d1a1 d8d6 c3c4 d5f6 d2c3 e5e4 d3e4 e8e4 e1d1 d6e7 c3f6 e7f6 a1f6 g7f6 d1d7 e4e7 d7e7 c6e7 c4c5 a8e8');
INSERT INTO evaluations VALUES(4434,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(4434,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4434,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4434,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4434,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(4434,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4434,36,'cp 13','b4c3 b2c3 h7h6 f1e1 d7d6 h2h3 c6a5 c4b3 c7c5 c1e3 b7b5 f3d2 f8e8 a2a4 a7a6 a4b5 a6b5 c3c4 c8d7 c4b5 a5b3 d2b3 d7b5 b3d2 d8d7 d1f3 a8a1 e1a1 d7e6 a1a7 e8b8 f3e2 b8e8 d2f1 e8b8');
INSERT INTO evaluations VALUES(4450,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4450,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(4450,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(4450,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(4450,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4450,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4450,36,'cp 0','b2c3 d7d5 e4d5 f6d5 f1e1 c8g4 c1d2 g4h5 h2h3 f8e8 a1b1 a7a6 c4b3 h7h6 g2g4 h5g6 c3c4 d5b6 c4c5 b6d7 b3d5 a8b8 b1b7 b8b7 d5c6 b7b8 d1a1 g8h7 a1c3 e8e6 c6d5 e6e8');
INSERT INTO evaluations VALUES(4448,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(4448,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(4448,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(4448,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4448,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(4448,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4448,36,'cp 7','d7d5 e4d5 f6d5 f1e1 c8g4 c1d2 f8e8 c4b3 g4h5 h2h3 h7h6 a2a4 d8d7 g2g4 h5g6 a4a5 a8d8 f3h4 g6h7 d1f3 a7a6 h4f5 h7f5 f3f5 d7f5 g4f5 d8d6 b3a4 d5e7 f2f4 e7f5 a1b1 f5h4 g1f2 d6g6 b1b7 g6g2 f2f1 g2d2 a4c6 e8f8 e1e5 d2c2 b7c7 c2c3 e5d5');
INSERT INTO evaluations VALUES(3730,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(3730,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3730,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(3730,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(3730,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(3730,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3730,36,'cp 12','e4d5 f6d5 c1d2 c8g4 h2h3 g4h5 f1e1 f8e8 c4b3 h7h6 a2a4 a7a6 g2g4 h5g6 a4a5 d8d7 f3h4 g6h7 d1f3 a8d8 f3g3 g7g5 b3d5 d7d5 h4g2 f7f6 g2e3 d5f7 h3h4 d8d7 f2f3 e8d8 a1b1 d8b8 c3c4 c6d4 g3f2 c7c6 h4g5');
INSERT INTO evaluations VALUES(3727,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(3727,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3727,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(3727,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(3727,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3727,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(3727,36,'cp 7','f6d5 c1d2 c8g4 f1e1 f8e8 c4b3 g4h5 h2h3 h7h6 a2a4 d8d7 g2g4 h5g6 f3h4 g6h7 d1f3 a8d8 b3d5 d7d5 f3d5 d8d5 f2f4 f7f6 f4f5 g8f7 g1f2 b7b6 c3c4 d5d7 d2c3 a7a5 a1b1 g7g6 f5g6 h7g6 b1b5 c6a7 b5b2 a7c6 h4g6 f7g6 b2b5 c6a7 b5d5 d7d5 c4d5');
INSERT INTO evaluations VALUES(3754,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(3754,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(3754,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3754,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(3754,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3754,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3754,36,'cp 16','c1d2 h7h6 f1e1 d8d6 d3d4 e5d4 f3d4 d5f6 d4b5 d6c5 c4f1 f8d8 d2e3 d8d1 e3c5 d1a1 e1a1 f6e8 b5d4 c8d7 g2g3 e8d6 f1g2 c6d4 c3d4 a8e8 c5d6 c7d6 g2b7 e8e2 c2c4 g8f8 b7d5 d7e6 d5e6 f7e6 c4c5 d6c5 d4c5 e2c2 a2a4 c2c5 a1b1 a7a5 b1b7 c5c2 g1g2 c2c3');
INSERT INTO evaluations VALUES(42624,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(42624,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(42624,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(42624,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(42624,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(42624,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(42624,36,'cp -26','f6g4 d3d4 e5d4 c3a4 c5b4 c2c3 d4c3 b2c3 b4d6 f3d4 g4f6 f1d3 e8g8 d4f5 f8e8 f5d6 c7d6 e1g1 b8c6 c3c4 b7b6 a4c3 c8a6 f1e1 c6e5 d3f1 a8c8 c3b5 a6b5 c4b5 e5g6 d1d6 e8e4 c1b2 e4e1 a1e1 c8c2 d6d4 g6f8 f1d3 c2c8 d4a4 c8c7 e1d1 f8e6');
INSERT INTO evaluations VALUES(42878,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(42878,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(42878,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(42878,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(42878,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(42878,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(42878,36,'cp 0','d3d4 e5d4 c3a4 c5b4 c2c3 d4c3 b2c3 b4d6 f3d4 h7h5 h2h3 d8h4 d1f3 g4e5 f3e3 e5g4');
INSERT INTO evaluations VALUES(45746,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(45746,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(45746,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(45746,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(45746,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(45746,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(45746,36,'cp -33','c6d5 d1b3 b8c6 c3d5 c6d4 d5f6 d8f6 b3d3 f8e7 e2e3 d4c6 g2e4 c8e6 a2a3 a8d8 d3e2 e6f5 e4f5 f6f5 e3e4 f5e6 g1f3 e6b3 e1g1 f7f6 d2d3 b3d3 e2d3 d8d3 c1e3 e8f7 f1c1 g7g6 g1f1 h8d8 c1c2 a7a5 a1b1 h7h5 b1c1 d3b3 f1e2');
INSERT INTO evaluations VALUES(47821,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(47821,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(47821,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(47821,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(47821,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(47821,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(47821,36,'cp 143','d4e5 d6e5 d1d8 e8d8 g1f3 b8c6 c4f7 f6e4 f7d5 c8f5 d5c6 b7c6 f3e5 f8c5 e5c6 d8d7 c6e5 d7c8 c1e3 a8b8 e5d3 c5e3 f2e3 e4f6 b2b3 h8d8 b1d2 f5d3 c2d3 d8d3 e1e2 d3d8 h1d1 b8b5 a1c1 b5d5 c1c2 d5g5 g2g3 g5h5 h2h4 g7g5 d1h1 d8e8 d2c4 g5h4 h1h4');
INSERT INTO evaluations VALUES(47770,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(47770,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(47770,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(47770,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(47770,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(47770,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(47770,36,'cp 150','d6e5 d1d8 e8d8 g1f3 b8c6 c4f7 f6e4 f7d5 c8f5 d5c6 b7c6 f3e5 f8c5 e5c6 d8d7 c6e5 d7c8 c1e3 a8b8 e5d3 c5e3 f2e3 e4f6 b2b3 h8d8 b1d2 b8b5 d2c4 f5d3 c2d3 d8d3 a1c1 b5g5 e1g1 g5d5 h2h3 d5d8 g1h2 d3d5 f1f4 d8e8 c1f1 e8e6 f4f5 d5f5 f1f5');
INSERT INTO evaluations VALUES(48908,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(48908,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(48908,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(48908,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(48908,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(48908,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(48908,36,'cp 160','d1d8 e8d8 g1f3 b8c6 f3g5 c6a5 c4f7 h7h6 g5f3 f8d6 b1c3 h8f8 f7g6 a5c6 h2h3 c6b4 e1d1 b4c6 c1e3 c6e7 f3h4 c8e6 f2f3 a7a5 d1e2 f6d7 h1d1 b7b5 g2g4 e6c4 e2f2 b5b4 c3d5 e7d5 e4d5 d7b6 g6e4 a5a4');
INSERT INTO evaluations VALUES(35394,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(35394,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(35394,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(35394,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(35394,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(35394,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(35394,36,'cp 159','e8d8 g1f3 b8c6 f3g5 d8e8 g5f7 c6a5 f7e5 f8d6 c4f7 e8e7 f2f4 a5c6 f7c4 c6e5 f4e5 d6e5 b1d2 c8e6 c4d3 f6g4 d2f3 e5f6 c1d2 f6b2 a1b1 b2f6 b1b7 h8b8 b7c7 e7d6 d2f4 g4e5 f3e5 d6c7 e5f3 c7d8 f4b8 f6c3 e1f2 a8b8 a2a3 d8e7 e4e5 e6d5 d3h7 d5f3 f2f3 c3e5 h1e1 e7d6 h7f5 e5h2');
INSERT INTO evaluations VALUES(35426,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(35426,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(35426,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(35426,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(35426,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(35426,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(35426,36,'cp 138','g1f3 b8c6 f3g5 d8e8 g5f7 c6a5 f7e5 f8d6 c4f7 e8e7 f2f4 a5c6 b1d2 d6e5 f4e5 e7f7 e5f6 g7f6 e1g1 c6d4 d2f3 d4f3 f1f3 c8g4 f3f2 h8e8 c1g5 e8e4 f2f6 f7g8 a1f1 a8e8 g5h6 e4e1 f1e1 e8e1 g1f2 e1e8 f2g3 g4h5 f6f2 h5g6 h2h3 g6e4 g3h4 c7c6 b2b3 e4g6 a2a3');
INSERT INTO evaluations VALUES(19647,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(19647,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(19647,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(19647,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(19647,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(19647,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19647,36,'cp 44','d2d4 c5d4 f3d4 f8g7 c1e3 c8d7 d1d2 g8f6 f2f3 a8c8 c4b3 e8g8 e1c1 c6d4 e3d4 b7b5 d4a7 b5b4 c3e2 d8a5 a7d4 c8c7 c1b1 f8a8 c2c4 d7e6 d1c1 a8c8 h1d1 h7h5 h2h3 g8h7 c1c2 e6c4 c2c4 c7c4 b3c4 c8c4 b2b3 c4c8 a2a4 f6d7 d4g7 h7g7 d1c1');
INSERT INTO evaluations VALUES(19659,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(19659,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19659,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(19659,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(19659,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(19659,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(19659,36,'cp 35','c5d4 f3d4 f8g7 c1e3 c8d7 d1d2 g8f6 f2f3 e8g8 e1c1 a8c8 c4e2 a7a6 h2h4 f6h5 h1e1 d8a5 c1b1 g7d4 e3d4 c6d4 d2d4 d7e6 c3d5 e6d5 e4d5 a5c5 d4d2 c5f2 g2g4 h5g3 e2d3 f2f3 e1e7 f3g4 e7b7 g4h4 b7a7 h4f6 a7a6');
INSERT INTO evaluations VALUES(19686,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(19686,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(19686,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(19686,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(19686,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(19686,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(19686,36,'cp 39','f3d4 f8g7 c1e3 g8f6 f2f3 c8d7 d1d2 e8g8 c4e2 c6d4 e3d4 d8a5 g2g4 a8c8 g4g5 f6h5 d4g7 h5g7 c3d5 a5d2 e1d2 c8e8 h2h4 d7c6 c2c4 f7f5 g5f6 e7f6 e2d3 f6f5 b2b4 f5e4 d3e4 g7h5 h1e1 h5g3 b4b5');
INSERT INTO evaluations VALUES(19682,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(19682,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(19682,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(19682,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19682,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(19682,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(19682,36,'cp 30','f8g7 c1e3 c8d7 f2f3 g8f6 d1d2 a8c8 c4b3 e8g8 h2h4 h7h5 e1c1 c6e5 e3h6 e5c4 b3c4 c8c4 h6g7 g8g7 h1e1 d8a5 d4b3 a5c7 e1e2 b7b5 c3d5 f6d5 e4d5 f8e8 b3d4 c7b6 d2f4 a7a5 b2b3 c4c5 d1e1 c5d5 e2e7 e8e7 e1e7 b6d4');
INSERT INTO evaluations VALUES(5522,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(5522,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(5522,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(5522,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(5522,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5522,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(5522,36,'cp 21','c1e3 c8d7 f2f3 a8c8 c4b3 d8b6 c3d5 b6a5 c2c3 g8f6 d1d2 f6d5 e4d5 c6d4 e3d4 g7d4 d2d4 e8g8 e1c1 a5c5 h1e1 c5d4 d1d4 f8e8 h2h4 a7a5 a2a4 c8c5 h4h5 g8g7 d4e4 g7f8 e1h1 d7f5 e4d4 g6h5 h1h5 f5g6 h5h4 h7h5 b3c2 e8c8 c2g6 f7g6');
INSERT INTO evaluations VALUES(5521,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(5521,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(5521,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(5521,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(5521,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(5521,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(5521,36,'cp 30','c8d7 f2f3 g8f6 d1d2 e8g8 h2h4 h7h5 c4b3 c6e5 e1c1 a8c8 e3h6 e5c4 b3c4 c8c4 h6g7 g8g7 c1b1 d8a5 h1e1 f8c8 d4b3 a5c7 e1e2 a7a6 g2g3 g7g8 b3d4 c7b6 c3d5 f6d5 e4d5 c8c5 e2e7 c5d5 e7d7');
INSERT INTO evaluations VALUES(25574,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(25574,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(25574,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(25574,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(25574,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(25574,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(25574,36,'cp 50','e1g1 e7e6 b2b3 f8d6 c1b2 e8g8 b5e2 c6e7 b1d2 e7g6 c2c4 b7b6 a1c1 d8e7 f3e5 d7c8 e5d3 c7c5 d4c5 b6c5 b2f6 g7f6 f1e1 f8d8 e2f3 c8a6 d1c2 a8c8 c4d5 e6d5 d3f4 d6f4 e3f4 e7d7 g2g3 g8g7 e1d1 d5d4 d2e4 a6b7 f3g2');
INSERT INTO evaluations VALUES(25573,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(25573,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(25573,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(25573,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(25573,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(25573,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25573,36,'cp 55','e7e6 b2b3 f8d6 c1b2 c6e7 b5e2 e8g8 b1d2 c7c5 d4c5 d6c5 a1c1 e7g6 c2c4 a8c8 c4d5 e6d5 a2a3 a7a5 d2b1 d7e6 b1c3 d8e7 c3b5 f6e4 a3a4 g6h4 f3h4 e7h4 e2f3 h4e7 b5d4 h7h6 d4e6 f7e6 g2g3 c8d8 f3g2');
INSERT INTO evaluations VALUES(25467,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(25467,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(25467,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(25467,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(25467,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(25467,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(25467,36,'cp 52','b2b3 f8d6 c1b2 e8g8 b5e2 c6e7 c2c4 c7c5 d4c5 d6c5 b1d2 e7g6 c4d5 e6d5 a1c1 a8c8 d2b1 d7e6 b2d4 d8e7 b1c3 f8d8 c3b5 c5d4 b5d4 e6d7 c1c8 d8c8 d1b1 a7a6 h2h3 h7h6 f1c1 c8c1 b1c1 e7d6 a2a3 g6e7 c1b2 e7c6 b3b4 c6d4 b2d4 d6e7');
INSERT INTO evaluations VALUES(32444,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(32444,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(32444,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(32444,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(32444,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(32444,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(32444,36,'cp 35','g8f6 c2c3 f8e7 e1g1 e8g8 d1b3 d8c8 c1e3 b7b6 a2a4 b8c6 b1d2 c8d7 f1e1 h7h6 a1d1 f8f7 d3d4 e5d4 f3d4 e6e5 d4f5 e7f8 f2f3 a7a5 d2f1 c6e7 f5e7 d7e7 e3f2 f6d7 b3b5 d7c5 f2c5 b6c5');
INSERT INTO evaluations VALUES(30279,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(30279,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(30279,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(30279,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30279,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(30279,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(30279,36,'cp 22','c2c3 f8e7 e1g1 e8g8 b1d2 a7a5 d1b3 d8c8 a2a4 b8c6 f1e1 c8d7 d2c4 b7b6 b3d1 d7e8 c1d2 f6d7 c4a3 d7c5 a3b5 e8g6 e1e3 a8c8 b2b4 c5d7 d1e2 h7h6 h2h3 a5b4 c3b4 d6d5 a1c1 c6b4 e4d5 b4d5 f3e5 d7e5');
INSERT INTO evaluations VALUES(48368,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48368,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(48368,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(48368,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(48368,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48368,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48368,36,'cp 0','f8e7 c2c4 e8g8 b1d2 b7b6 c4d5 e6d5 f1d3 b8d7 e1g1 f8e8 d1c2 c7c5 b2b3 c8b7 c1b2 a8c8 d4c5 b6c5 h2h3 e7d6 d3f5 c8c7 a1d1 d8e7 f1e1 d7f8 f5d3 f8g6 d3b5 e8d8 b5d3');
INSERT INTO evaluations VALUES(41007,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41007,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41007,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(41007,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41007,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41007,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41007,36,'cp 0','c2c4 e8g8 b1c3 b8d7 c4d5 e6d5 f1d3 f8e8 e1g1 c7c6 f1e1 e7d6 d1c2 d7b6 h2h3 h7h6 c1d2 c8e6 a1c1 a7a5 c1d1 b6d7 d1c1');
INSERT INTO evaluations VALUES(13394,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13394,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13394,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13394,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13394,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13394,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13394,36,'cp 0','g1f3 g7g6 h2h3 f8g7 f1c4 e8g8 e1g1 c6a5 c4d3 a7a6 a2a4 c8e6 f1e1 a8c8 a1b1 f8e8 c3e2 e6a2 b1a1 a2c4 a1b1');
INSERT INTO evaluations VALUES(13393,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13393,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13393,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13393,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13393,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13393,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13393,36,'cp 0','g7g6 f1c4 f8g7 h2h3 c6a5 c4d3 a7a6 a2a4 c8e6 e1g1 e8g8 f1e1 a8c8 a1b1 f6d7 c3e2 d7e5 f3d4 e6d7 b2b3 a5c6 d4c6 d7c6 e2d4 e5d3 d1d3 b7b5 a4b5 a6b5 d4c6 c8c6 d3b5 c6c2 c1e3 c2c7 b5d5 d8a8');
INSERT INTO evaluations VALUES(13044,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13044,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13044,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13044,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13044,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13044,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(13044,36,'cp 9','h2h3 f8g7 f1c4 c6a5 c4d3 e8g8 e1g1 a7a6 d1e1 a5c6 a2a4 d6d5 e4d5 f6d5 c3d5 d8d5 d3e4 d5d6 e1e3 f8d8 f1e1 c8f5 e4f5 g6f5 e3e2 e7e5 c1g5 d8e8 f3h4 c6d4 e2d1 h7h6 g5e3 d6f6 e3d4 e5d4 e1e8 a8e8 d1h5 e8e5 a1d1');
INSERT INTO evaluations VALUES(20026,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(20026,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(20026,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(20026,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(20026,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(20026,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20026,36,'cp 9','b1c3 g8f6 d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 f1d3 f8e7 e1g1 e8g8 d1e2 c8e6 b2b3 d8c7 c1g5 h7h6 g5f6 e7f6 d3c4 f6e7 a1d1 e6c8 b3b4 a8b8 d1b1 d6d5 e4d5 b8b4 f1d1 f8d8 b1b4 e7b4 c3e4 b4f8 d5c6 d8d1 e2d1 c7c6');
INSERT INTO evaluations VALUES(19997,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19997,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19997,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19997,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19997,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(19997,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19997,36,'cp 0','g8f6 d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 f1c4 f8e7 e1g1 e8g8 c4b3 a7a5 f1e1 d8c7 d1f3 c8e6 c1e3 a8b8 a1d1 h7h6 e3d2 g8h8 f3d3 h8g8');
INSERT INTO evaluations VALUES(27212,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(27212,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(27212,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(27212,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(27212,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(27212,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(27212,36,'cp 37','e1g1 a7a6 b5a4 b7b5 a4b3 f8e7 c2c3 g8f6 b3c2 e8g8 b1d2 f8e8 f1e1 e7f8 d2f1 h7h6 d3d4 d7g4 d4d5 c6a5 f1e3 g4d7 b2b4 a5b7 c1d2 a6a5 h2h3 c7c6 f3h2 c6d5 e4d5 a5b4 c3b4 h6h5 h2f1 g7g6 f1g3 f8h6 a2a4 a8a4 a1a4 b5a4');
INSERT INTO evaluations VALUES(27211,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(27211,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(27211,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(27211,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(27211,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(27211,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(27211,36,'cp 32','a7a6 b5a4 b7b5 a4b3 f8e7 c2c3 g8f6 b3c2 e8g8 f1e1 f8e8 b1d2 e7f8 d2f1 c6a5 c1g5 c7c5 f1e3 h7h6 g5h4 g7g6 h4f6 d8f6 h2h4 h6h5 e3d5 f6d8 a2a4 d7e6 a4b5 e6d5 e4d5 a6b5 b2b4 c5b4 c3b4 a5b7 a1a8 d8a8 g2g3 a8d8');
INSERT INTO evaluations VALUES(25713,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(25713,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(25713,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(25713,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(25713,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(25713,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25713,36,'cp 41','b5a4 b7b5 a4b3 f8e7 c2c3 g8f6 b3c2 e8g8 b1d2 d7g4 f1e1 d6d5 e4d5 d8d5 h2h3 g4f5 d1e2 f8e8 a2a4 b5b4 d2c4 e7c5 c4e5 b4c3 b2c3 c6e5 f3e5 c5d6 d3d4 f5c2 e2c2 d6e5 d4e5 e8e5 c1f4');
INSERT INTO evaluations VALUES(25730,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(25730,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(25730,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(25730,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(25730,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25730,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(25730,36,'cp 29','b7b5 a4b3 f8e7 c2c3 g8f6 b3c2 e8g8 f1e1 f8e8 b1d2 e7f8 d2f1 c6e7 c1g5 e7g6 g5f6 d8f6 f1e3 g6f4 h2h4 g7g6 g2g3 f4e6 d3d4 e5d4 c3d4 f8g7 e4e5 f6d8 c2e4 d6e5 d4e5 c7c6 h4h5 e6c5 d1c2 c5e4');
INSERT INTO evaluations VALUES(40840,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(40840,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(40840,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(40840,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(40840,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(40840,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(40840,36,'cp 47','e8g8 c1f4 b8c6 d1d2 f8e8 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 f6e4 f4d6 e4c3 d4c3 c7d6 e2f3 c8g4 f3g4 d8g5 d1d2 g5g4 c3g3 g4e6 g3b3 e6e4 f2f3 e4c6 h1d1 a8d8 b3b4 d6d5 b4d4 e8e6 c1b1 d8d6 g2g4 h7h6 d4f2 a6a5 a2a3 c6b5 h2h4');
INSERT INTO evaluations VALUES(36263,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(36263,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36263,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(36263,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(36263,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(36263,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(36263,36,'cp 31','c1f4 f8e8 d1d2 b8c6 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 f6e4 f4d6 e4c3 d4c3 c7d6 e2f3 c8g4 f3g4 d8g5 c1b1 g5g4 f2f3 g4g5 c3d2 g5g6 a2a3 h7h6 d2f2 e8e6 d1d2 a8e8 h1d1 g6g5 f3f4 g5h5 f2b6 h5h2');
INSERT INTO evaluations VALUES(48091,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(48091,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(48091,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(48091,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(48091,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(48091,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(48091,36,'cp 134','e2e4 f8b4 c1d2 b4d2 d1d2 d5b6 f1e2 b8c6 b1c3 c8d7 e1g1 e8g8 h2h3 d8e7 f1e1 e7f6 a1c1 f8d8 d2e3 d7e8 e1d1 a7a5 a2a3 a5a4 c3b5 d8d7 e2d3 f6e7 b5c3 c6a5 c3e2 f7f6 d1e1 a8d8 d3b5 c7c6 b5d3');
INSERT INTO evaluations VALUES(48094,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(48094,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(48094,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(48094,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(48094,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(48094,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(48094,36,'cp 131','f8b4 c1d2 b4d2 d1d2 d5b6 f1e2 c8d7 b1c3 b8c6 e1g1 e8g8 h2h3 h7h6 f1e1 c6e7 a1c1 e7g6 e2d3 d8e7 c3e2 f8c8 d2a5 c7c6 b2b3 d7e8 a2a4 b6d7 a5d2 c6c5 d4d5 b7b6 a4a5 b6a5 d2a5 a8b8');
INSERT INTO evaluations VALUES(41938,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(41938,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(41938,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(41938,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(41938,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(41938,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(41938,36,'cp 132','c1d2 b4d2 d1d2 d5b6 f1e2 b8c6 e1g1 e8g8 h2h3 c8d7 b1c3 d8f6 f1d1 d7e8 d2e3 f6e7 d4d5 e6d5 c3d5 b6d5 e4d5 e7e3 f2e3 c6d8 e3e4 f7f6 a1c1 c7c6 f3d4 e8f7 e2f3 f8e8 d4f5 g8f8 b2b4 e8e5 f5d6 f7g6 a2a4 a8b8 b4b5 c6d5 e4d5 d8f7 d6f7 f8f7');
INSERT INTO evaluations VALUES(41937,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(41937,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(41937,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(41937,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(41937,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(41937,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(41937,36,'cp 122','b4d2 d1d2 d5b6 f1e2 b8c6 e1g1 e8g8 d2e3 d8f6 f1d1 c8d7 b1c3 f8d8 a1c1 d7e8 h2h4 a7a6 e2f1 f6e7 f1d3 f7f6 g2g4 c6b4 d3e2 b4c6 a2a3 g8h8 h4h5 c6a5 f3d2 a5c6 d2b3 h7h6 d4d5 c6e5 f2f4 e5c4 e3c5 e7c5 b3c5 e6d5 c5b7');
INSERT INTO evaluations VALUES(41940,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(41940,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(41940,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(41940,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(41940,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(41940,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(41940,36,'cp 135','d1d2 d5b6 f1e2 e8g8 b1c3 b8c6 h2h4 f7f5 a1d1 f5e4 c3e4 d8d5 e4c3 d5f5 e1g1 f5f4 f1e1 h7h6 c3b5 b6d5 d2f4 f8f4 b5c3 c8d7 e2c4 c6e7 c3d5 e7d5 f3e5 d7e8 g2g3 f4f8 d1c1 c7c6 a2a3 a8d8 c1c2');
INSERT INTO evaluations VALUES(57087,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(57087,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(57087,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(57087,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(57087,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(57087,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(57087,36,'cp -79','a7a6 e2e3 a6b5 f1b5 e5e4 c1b2 c7c6 b5e2 d7d5 c2c4 d8g5 g2g3 d5c4 h2h4 g5d8 e2c4 b7b5 c4e2 g8f6 a2a4 b5b4 d2d3 e4d3 d1d3 d8d3 e2d3 b8d7 g1f3 d7c5 d3c4 f6d7 a4a5 c8a6 c4a6 a8a6 b1d2 c5d3 e1e2 d3b2 h1b1 f8e7 b1b2 e7f6 f3d4 c6c5');
INSERT INTO evaluations VALUES(30610,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(30610,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(30610,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(30610,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(30610,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(30610,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(30610,36,'cp 0','f8b4 c1d2 d8e7 f1e2 g8f6 d2b4 e7b4 d1d2 b4d2 b1d2 d7d6 g1f3 e6e5 e1c1 b8d7 f3e1 a7a5 b2b3 h7h5 h2h4 f6g4 h1f1 g4h2 f1h1');
INSERT INTO evaluations VALUES(27939,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(27939,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(27939,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(27939,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27939,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27939,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27939,36,'cp 0','c1d2 d8e7 f1e2 g8f6 d2b4 e7b4 d1d2 b4d2 b1d2 e8g8 f2f3 d7d6 a1d1 c7c6 d5e6 f7e6 d2b3 d6d5 c4d5 e6d5 e4d5 f6d5 g1h3 b7a6 b3d4 d5e3 d1d2 e3g2 e1f2 a6e2 d2e2 g2h4 h3g5 h7h6 g5e6 f8f7 e6d8 f7f8');
INSERT INTO evaluations VALUES(31119,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(31119,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31119,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31119,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31119,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(31119,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(31119,36,'cp 22','c2c4 e7e6 h2h3 g4h5 d1b3 d8b6 g2g4 h5g6 f3e5 g6e4 f2f3 e4b1 a1b1 f8d6 c1d2 g8e7 f3f4 h7h5 f1g2 h5g4 h3g4 h8h1 g2h1 b8d7 b3b6 d7b6 c4d5 c6d5 h1f3 f7f6 e5d3 b6c4 b2b3 c4d2 e1d2 e8d7 b1h1 a7a5 h1h7 a8g8 d3c5 d6c5 d4c5');
INSERT INTO evaluations VALUES(31108,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31108,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(31108,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31108,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(31108,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(31108,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(31108,36,'cp 18','e7e6 d1b3 d8b6 c1d2 b8d7 d2c3 g8f6 b1d2 b6c7 f1d3 a7a5 h2h3 g4h5 b3c2 h5g6 e1g1 f8e7 a1c1 e8g8 b2b3 g6d3 c2d3 h7h6 c1c2 f8c8 f1c1 d5c4 d2c4 a5a4 c4e5 e7a3 c3b2 a3b2 c2b2 d7e5 f3e5 c7e7');
INSERT INTO evaluations VALUES(42156,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(42156,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(42156,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(42156,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(42156,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(42156,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(42156,36,'cp 47','c1d2 b4d2 b1d2 e8g8 e1g1 b7b6 b2b4 c8b7 f1e1 b8d7 d1b3 c7c5 b4c5 b7f3 d2f3 b6c5 e1c1 d8b6 b3a3 c5d4 f3d4 f8c8 a1b1 b6c5 a3a6 d7b6 d4b3 c5e7 c4e2 f6d5 e2f3 g7g6 h2h3 g8g7 b3d4 c8c7 c1c7 e7c7 f3d5 b6d5 b1b7');
INSERT INTO evaluations VALUES(42155,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(42155,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(42155,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(42155,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(42155,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(42155,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(42155,36,'cp 50','b4d2 b1d2 e8g8 e1g1 b7b6 b2b4 c8b7 a1c1 b8d7 c4b5 a7a5 a2a3 a5b4 a3b4 a8a7 d1c2 d8e7 c2c7 f8a8 c7c4 b7d5 c4d3 e7b4 e3e4 f6e4 c1b1 b4c3 d2e4 c3d3 b5d3 a7a3 d3c2 a3a2 c2b3 d5e4 b3a2 a8a2 b1b3 h7h6 b3c3 e4f3 c3f3 d7f6');
INSERT INTO evaluations VALUES(42181,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(42181,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(42181,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(42181,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(42181,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(42181,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(42181,36,'cp 55','b1d2 e8g8 e1g1 b7b6 c4b5 a7a6 b5e2 b8d7 a1c1 a6a5 e2b5 c8b7 b5c6 b7c6 c1c6 f6d5 a2a3 d5e7 c6c2 c7c5 d2e4 a8c8 d4c5 d7c5 e4c5 c8c5 c2c5 b6c5 d1d8 f8d8 f1c1 d8b8 c1c5 b8b2 g2g3 g7g6 c5a5 e7c6 a5a8 g8g7 f3g5 b2a2 g1g2');
INSERT INTO evaluations VALUES(42179,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(42179,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(42179,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42179,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(42179,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(42179,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(42179,36,'cp 46','e8g8 a1c1 b7b6 e1g1 b8d7 c4e2 h7h6 d1a4 c8b7 f1d1 c7c5 d2b3 c5d4 f3d4 a7a6 d4c6 d8e8 a4d4 a8c8 d4d6 b7c6 c1c6 c8c6 d6c6 e8a8 d1c1 a8c6 c1c6 f8a8 g1f1 g8f8 e2f3 a8a7 f1e2 f8e7 b3d4 f6e8 b2b4 e8d6');
INSERT INTO evaluations VALUES(37089,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(37089,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(37089,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(37089,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(37089,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(37089,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(37089,36,'cp 43','c4e2 b8d7 a1c1 b7b6 b2b4 c8b7 e1g1 h7h6 d1c2 c7c6 d2c4 d8e7 c2b2 f6d5 a2a3 a7a5 b4a5 b6b5 c4e5 d7e5 f3e5 a8a5 c1c5 f8a8 e5c6 b7c6 c5c6 e7a3 b2a3 a5a3 e2b5 d5f6 b5e2 a3a2 e2f3 a8b8 h2h4 g7g6 g2g4 g8g7 c6c7 g6g5 f1c1 b8b2 h4g5 h6g5');
INSERT INTO evaluations VALUES(8689,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(8689,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(8689,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(8689,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(8689,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(8689,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(8689,36,'cp 39','e8g8 c1f4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 a8d8 f1e1 f8e8 h2h3 d6d5 e4d5 f6d5 c3d5 d7d5 d4c3 d5e4 d1d8 e8d8 f4c7 e7f6 c3f3 e4f3 e2f3 d8c8 c7g3 c8c2 f3b7 c2b2 b7a6 b2a2 a6b5');
INSERT INTO evaluations VALUES(3440,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(3440,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(3440,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(3440,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(3440,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(3440,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(3440,36,'cp 66','c1f4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 f8e8 h2h3 d7c6 e2f3 c6c4 d4d2 c4b4 b2b3 a8d8 f1e1 g8f8 d2d3 f6d7 c3d5 e6d5 d3d5 e7f6 f4d2 b4c5 d5b7 d8b8 b7a6 c5c2 d2e3 c2b3 d1b1 b3b1 e1b1 b8b1 g1h2 b1a1 a4a5 d7e5 f3h5 g7g6');
INSERT INTO evaluations VALUES(3442,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(3442,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(3442,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(3442,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(3442,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(3442,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(3442,36,'cp 48','a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 a8d8 f1e1 f8e8 h2h3 c7c6 a4a5 d7c7 d4a4 h7h6 e2f1 e7f8 a4a1 d8d7 b2b4 c7d8 f4e3 d6d5 e3b6 d8b8 e4e5 f6h5 c3a4 f8b4 c2c3 b4f8 a4c5 f8c5 b6c5 h5f4 g2g3');
INSERT INTO evaluations VALUES(14245,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14245,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(14245,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14245,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14245,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(14245,36,'cp 0','g1f3 c8g4 h2h3 g4f3 d1f3 e7e6 b1c3 h7h5 f1d3 c6b4 e1g1 c7c5 f4f5 c5c4 d3e2 f8d6 b2b3 c4b3 e2b5 b4c6 a2b3 e8g8 b5d3 d6b4 c3a2 b4d6');
INSERT INTO evaluations VALUES(14246,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14246,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14246,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14246,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14246,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14246,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14246,36,'cp 0','c8g4 h2h3 g4f3 d1f3 e7e6 b1c3 h7h5 f1d3 c6b4 e1g1 c7c5 f4f5 c5c4 d3e2 f8d6 b2b3 c4b3 e2b5 b4c6 a2b3 e8g8 b5d3 d6b4 c3a2 b4d6');
INSERT INTO evaluations VALUES(24932,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24932,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24932,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24932,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24932,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24932,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24932,36,'cp 0','h2h3 g4f3 d1f3 e7e6 b1c3 h7h5 c1d2 h5h4 f1d3 c6b4 e1g1 b4d3 c2d3 f8e7 a2a4 d8d7 c3e2 f6h5 b2b4 g7g6 e3e4 a7a6 g1h1 f7f5 e4e5 e8f7 f1c1 c7c6 f3f2 a8g8 a4a5 f7g7 f2f3 g7f7 f3f2');
INSERT INTO evaluations VALUES(18320,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(18320,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(18320,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(18320,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(18320,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(18320,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(18320,36,'cp 176','e1g1 c6a5 c4e2 a5c6 c2c3 a7a6 d2d4 d7d6 d1c2 h7h6 f1d1 g7g5 d4e5 g5g4 f3d4 c6e5 b1d2 d8f6 f2f4 f6f4 d2c4 f4f6 d4f5 h6h5 c4e5 f6e5 d1d5 e5e6 c2a4 c8d7 a4d4 e6f6 e4e5 f6e6 e5d6 g8f6 e2c4 f6d5 c4d5 e6f5 d5a8');
INSERT INTO evaluations VALUES(18319,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(18319,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(18319,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(18319,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(18319,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(18319,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(18319,36,'cp 173','c6a5 c4e2 a5c6 c2c3 g8f6 e2b5 f8e7 f1e1 e8g8 d2d4 e5d4 e4e5 f6e8 c3d4 c6b4 b1c3 c7c6 b5a4 c8a6 d4d5 b4d5 c3d5 c6d5 d1d5 a8c8 d5d7 d8d7 a4d7 c8d8 e5e6 e8c7 e6f7 f8f7 f3e5 f7f8 d7g4 e7c5 c1e3 c5e3 e1e3 d8d2');
INSERT INTO evaluations VALUES(51782,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(51782,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(51782,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(51782,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(51782,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(51782,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(51782,36,'cp 59','d2d4 d7d5 e2e4 d5e4 c3e4 c6c5 e4c5 b8c6 c1e3 e6e5 d4e5 d8a5 d1d2 c6b4 c5e4 c8f5 a1d1 b4c2 e1e2 a5e5 d2c2 f5e4 c2c3 e5c3 b2c3 g8h6 g1f3 h6f5 f3d4 f5e3 e2e3 e4c6 d4c6 b7c6 e3f3 f8c5 g2g3 e8g8 f1h3 a8b8 h1e1 b8b2 e1e2 f8b8 d1e1 b2e2 e1e2 b8d8 e2e5 c5f8 f3e2 g7g6');
INSERT INTO evaluations VALUES(19852,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(19852,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19852,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(19852,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(19852,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(19852,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(19852,36,'cp 25','e2e4 g8f6 e4e5 f6e4 g1f3 d8a5 d1b3 e7e6 f1d3 f8b4 e1g1 b4c3 b2c3 a5c3 b3b1 e8g8 c1e3 h7h6 a2a4 c8d7 f1c1 c3b4 d3e4 d5e4 b1e4 c6e7 c1b1 b4a5 e4b7 a5d8 b1c1 a7a5 f3d2 a8b8 b7a7 b8b4 d2c4 b4a4 a1a4 d7a4 c4a5 d8d5 a7e7 d5a5 e7a3 f8b8');
INSERT INTO evaluations VALUES(19867,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(19867,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19867,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(19867,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(19867,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(19867,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(19867,36,'cp 25','g8f6 e4e5 f6e4 f1b5 e4c3 b2c3 d8a5 d1b3 a7a6 b5c6 b7c6 g1e2 a5b5 b3c2 e7e6 e1g1 a6a5 e2f4 c8a6 f1e1 f8e7 c2d1 e8g8 e1e3 c6c5 a2a4 b5c4 e3g3 c5d4 c3d4 g7g6 f4h5 c4e2 h5f6 g8h8 d1e2 a6e2 g3e3 e2c4 e3h3 e7f6 e5f6');
INSERT INTO evaluations VALUES(39332,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(39332,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(39332,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(39332,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(39332,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(39332,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(39332,36,'cp 197','c2c3 b4f8 d4e5 b8c6 g1f3 d7d6 c1g5 d8d7 e5d6 f8d6 b1d2 h7h6 g5h4 g8e7 d2c4 e8g8 h4g3 d6g3 h2g3 d7d1 a1d1 c8e6 f1e2 f8e8 f3d2 b7b5 c4e3 b5b4 d2b3 b4c3 b2c3 a8b8 e3d5 e8d8 c3c4 e6d5 c4d5 c6b4 e1g1 b4a2');
INSERT INTO evaluations VALUES(39330,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(39330,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(39330,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(39330,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(39330,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(39330,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(39330,36,'cp 209','b4f8 d4e5 d7d6 g1f3 b8c6 f1b5 c8d7 e5d6 f8d6 e1g1 a7a6 b5a4 g8e7 b1d2 e7g6 d2c4 d6e7 c4e3 e8g8 a4c2 f8e8 a2a4 e7c5 e3d5 d7g4 h2h3 g4f3 d1f3 c6e5 f3e2 c7c6 d5f4 d8f6 f4h5 f6e6 g1h1 e6c4 e2c4 e5c4 b2b3 c4d6 f2f3 g8h8 a4a5 a8d8');
INSERT INTO evaluations VALUES(55037,30,'cp -152',NULL);
INSERT INTO evaluations VALUES(55037,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(55037,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(55037,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(55037,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(55037,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(55037,36,'cp -132','d5d4 c3b1 e7e5 d2d3 g8f6 g2g3 b8c6 f1g2 a7a5 g1f3 f8b4 c1d2 b4c5 e1g1 e8g8 d2g5 h7h6 g5f6 d8f6 b1d2 f6e7 d1c2 c6b4 c2c1 a5a4 d2e4 c5b6 f3h4 c8e6 a2a3 b4c6 c1c2 a8a5 e4d2 f8d8 g2c6');
INSERT INTO evaluations VALUES(55202,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(55202,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(55202,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(55202,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(55202,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(55202,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(55202,36,'cp -137','c3b1 e7e5 d2d3 f8b4 b1d2 g8f6 g2g3 a7a5 f1g2 b8c6 a2a3 b4d6 g1f3 h7h6 e1g1 e8g8 d1c2 d6c5 b2b3 d8e7 d2e4 f6e4 d3e4 b7b6 h2h3 f8e8 f3e1 c6d8 e1d3 c5d6 e2e3 d8c6 g1h2 g8h8 e3d4 c6d4 c2d1 c7c5 c1b2 f7f5');
INSERT INTO evaluations VALUES(15962,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(15962,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(15962,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(15962,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(15962,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(15962,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(15962,36,'cp 21','b1d2 f8e7 h2h3 c7c5 c2c3 d8b6 d1b3 e8g8 a2a4 b6b3 d2b3 f6e8 g2g4 f7f6 h1g1 c5c4 b3d2 e7d6 f4d6 e8d6 h3h4 e6e5 h4h5 f8e8 d4e5 f6e5 e3e4 d7c5 e4d5 e5e4 f3d4 c5d3 f1d3 e4d3 e1d1 c8d7 h5h6 g7g6 a4a5 a8c8 g4g5');
INSERT INTO evaluations VALUES(15963,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(15963,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(15963,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(15963,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(15963,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(15963,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(15963,36,'cp 12','f8e7 h2h3 e8g8 a2a4 b7b6 f1d3 c8b7 c2c3 a7a5 e1g1 f6e4 d1e2 e7d6 f4d6 c7d6 f1c1 a8c8 d3b5 d7f6 d2e4 d5e4 f3d2 d8c7 e2e1 c7e7 b2b4 a5b4 c3b4 c8c1 a1c1 d6d5 f2f3 e7b4 f3e4');
INSERT INTO evaluations VALUES(9591,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(9591,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(9591,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(9591,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(9591,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9591,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(9591,36,'cp 27','h2h3 e8g8 a2a4 b7b6 f1d3 c8b7 c2c3 a7a5 e1g1 f6e4 d1e2 e7d6 f1d1 d8e7 f3e5 f8d8 e5d7 d8d7 d2e4 d5e4 d3b5 d7d8 e2g4 b7d5 g4g3 d6f4 g3f4 c7c6 b5e2 c6c5 d4c5 b6c5 c3c4 d5b7 b2b3 a8c8 f4g3 b7c6');
INSERT INTO evaluations VALUES(26765,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(26765,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(26765,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(26765,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(26765,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(26765,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(26765,36,'cp 78','b5c6 b7c6 d1f3 d8d7 b1c3 g8f6 e1g1 d7g4 f3d3 f8e7 d3c4 g4d7 c1f4 e8g8 a1d1 f6h5 f4e3 e7f6 b2b3 f6e5 f1e1 f8e8 f2f3 h5f4 c3e2 f4e2 e1e2 a7a5 a2a4 g7g6 f3f4 e5f6 h2h3 d7e6 c4c6 e6e4 c6c7');
INSERT INTO evaluations VALUES(26756,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(26756,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(26756,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(26756,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(26756,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(26756,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(26756,36,'cp 79','b7c6 d1f3 d8d7 b1c3 g8f6 e1g1 d7g4 f3d3 f8e7 c1e3 e8g8 f2f4 a7a5 h2h3 g4e6 b2b3 f8e8 e3d4 f6d7 a1e1 e7f6 d4f6 d7f6 e1e2 f6d7 d3d4 f7f6 f1d1 d7b6 d1e1 a5a4 e4e5 f6e5 f4e5 d6d5 c3a4 b6a4');
INSERT INTO evaluations VALUES(31013,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31013,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31013,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(31013,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(31013,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(31013,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31013,36,'cp 25','g1f3 g8f6 f1d3 h7d3 d1d3 e7e6 c1d2 b8d7 e1c1 f8e7 g3e4 e8g8 c1b1 c6c5 e4f6 e7f6 d4d5 e6d5 d3d5 d7b6 d5d8 f8d8 d2e3 b6c4 e3c1 b7b5 c2c3 c4b6 d1d8 a8d8 b1c2 b6d5 h1e1 a7a5 f3e5 a5a4 e5g4 f6e7 c1d2 c5c4');
INSERT INTO evaluations VALUES(31012,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31012,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(31012,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(31012,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(31012,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31012,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31012,36,'cp 12','b8d7 f1d3 h7d3 d1d3 e7e6 c1d2 g8f6 e1c1 f8e7 g3e4 e8g8 c1b1 c6c5 e4f6 d7f6 h1g1 d8d5 g2g4 d5e4 d3e4 f6e4 d2e3 f8d8 c2c3 c5d4 d1d4 d8d4 f3d4 a8c8 b1c2 e4f6 a2a4 a7a6 c2b3 f6d5 d4f3 c8d8 g1e1 d5e3 f2e3 b7b5');
INSERT INTO evaluations VALUES(41327,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(41327,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41327,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(41327,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(41327,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41327,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(41327,36,'cp 0','f6h5 f4g5 h7h6 g5h4 c7c5 c2c3 c5d4 c3d4 b8c6 f1d3 e8g8 e1g1 c8e6 a2a3 a8c8 a1c1 d8d7 d2e4 f7f5 e4g3 h5g3 h4g3 g6g5 h2h3 e6d5 f1e1 e7e6 e3e4 f5e4 d3e4 d5e4 e1e4 c6e7 c1c8 f8c8');
INSERT INTO evaluations VALUES(47319,30,'cp -134',NULL);
INSERT INTO evaluations VALUES(47319,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(47319,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(47319,33,'cp -151',NULL);
INSERT INTO evaluations VALUES(47319,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(47319,35,'cp -149',NULL);
INSERT INTO evaluations VALUES(47319,36,'cp -137','c8g4 f1e2 g4f5 d2d3 e4d3 c2d3 h7h6 g5e4 f6e4 d3e4 d8d1 e2d1 f5h7 f2f3 e7e6 c3b5 b8a6 c1f4 c7c6 b5d6 f8d6 f4d6 e8c8 d6g3 f7f6 g3f2 b7b6 d1e2 a6b4 a1c1 c8b7 a2a3 b4d3 e2d3 d8d3 c1d1 d3d1 e1d1 e6e5 d1d2 h7g8 b2b4 g8b3 h1c1 h8d8 d2c3 b3e6 f2e3 b7c7');
INSERT INTO evaluations VALUES(52107,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(52107,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(52107,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(52107,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(52107,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52107,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52107,36,'cp 0','b8c6 d2d3 g7g6 b1c3 f8g7 a2a3 a7a5 g2g3 g8e7 f1g2 d7d6 a1b1 e8g8 e1g1 c8g4 h2h3 g4e6 c1d2 h7h6 g1h2 a8b8 b2b3 b7b6 f3h4 c6d4 h4f3');
INSERT INTO evaluations VALUES(57014,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(57014,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(57014,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(57014,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(57014,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(57014,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(57014,36,'cp -36','c7c5 c2c3 g8f6 e2e3 f8e7 g1f3 e8g8 f1d3 d7d5 e1g1 b7b6 d1e2 c8b7 c1d2 f6e4 d2e1 b8d7 b1d2 d7f6 e1h4 f6e8 h4e7 d8e7 f3e5 a8c8 a2a4 e8d6 a4a5 f7f6 d3e4 d6e4 e5d3 e4d6 a5b6 a7b6 a1a7 c8c7 d2f3 c5c4 d3b4');
INSERT INTO evaluations VALUES(36218,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(36218,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36218,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(36218,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36218,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(36218,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(36218,36,'cp 25','d2d4 e5d4 f3d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 c2c3 d7f6 e4f3 c7c6 a2a4 a7a5 f1e1 e7c5 c1g5 c5d4 c3d4 h7h6 g5h4 d8d6 d1d3 c8e6 h4f6 g7f6 e1e3 a8d8 a1d1 f8e8 d3c3 d6b4 c3b4 a5b4 d4d5 e6d5 e3e8 d8e8 f3d5 c6d5');
INSERT INTO evaluations VALUES(36232,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(36232,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36232,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(36232,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(36232,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(36232,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(36232,36,'cp 19','e5d4 f3d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 e4c2 e7c5 c1f4 d7f6 d1d3 c5d4 c3d4 g7g6 f4g5 c8f5 d3f3 h7h6 g5f6 d8f6 a1e1 g8g7 e1e8 a8e8 c2f5 f6f5 f3b7 f5c2 b7a7 c2b2 a7c7 e8e2 c7d7 e2c2 d4d5 b2a2 d5d6');
INSERT INTO evaluations VALUES(36253,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(36253,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36253,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36253,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(36253,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(36253,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36253,36,'cp 31','f3d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 b8d7 e4c2 d7f6 h2h3 d8d5 c2b3 d5c5 d4f3 e7d6 f3g5 e8f8 c1e3 c5a5 f1e1 c8d7 d1d3 a5f5 d3f5 d7f5 g5f3 f8e8 a1d1 h7h6 b3a4 f5d7 a4d7 f6d7 e3f4 e8e1 f3e1 d6f4 d1d7 a8e8 g1f1 f4d6 e1f3 g8f8 a2a4');
INSERT INTO evaluations VALUES(54959,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(54959,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(54959,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(54959,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(54959,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54959,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(54959,36,'cp 19','e7e5 f1g2 f7f5 b2b4 g8f6 b1c3 f8e7 g1f3 e8g8 e1g1 a7a6 d2d3 h7h6 a2a4 d8e8 c1a3 c8e6 b4b5 a6b5 a4b5 b8d7 e2e3 e8f7 f1e1 c7c6 d1c2 d6d5 a3e7 f7e7 a1a8 f8a8 b5c6 b7c6 c4d5 c6d5 g2h3 a8c8 e1b1 e7f7');
INSERT INTO evaluations VALUES(17477,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(17477,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(17477,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(17477,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(17477,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(17477,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(17477,36,'cp -26','b7b5 c4d5 c8b7 d5c6 b7c6 e1g1 e7e6 d2d4 c5d4 d1d4 g8f6 c1g5 f8e7 c3d5 c6d5 e4d5 h7h6 g5h4 d8c8 a2a4 e8g8 d5d6 e7d8 a4b5 a6b5 f3e5 f6d5 a1a8 c8a8 e5d7 d8h4 d4h4 f8d8 d7e5 d8d6 h4g3 d6d8 g3f3 a8b7 g2g3 b7c7 f1e1 d5f6 f3c6 c7c6 e5c6');
INSERT INTO evaluations VALUES(18168,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(18168,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(18168,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(18168,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(18168,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(18168,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(18168,36,'cp -4','c4d5 c8b7 e1g1 e7e6 d5c6 b7c6 d2d4 c5d4 d1d4 g8f6 c1g5 f8e7 c3d5 c6d5 e4d5 h7h6 g5h4 d8c8 a2a4 e8g8 d5d6 e7d8 f3e5 f6d5 a4b5 a6b5 a1a8 c8a8 e5d7 d8h4 d4h4 f8d8 d7e5 d8d6 h4h5 a8b7 h5e2 d6d8 g2g3 d8a8 f1d1 a8a2 d1b1 b7c7 h2h4 c7c5 c2c4 b5c4 e2c4 c5c4');
INSERT INTO evaluations VALUES(18173,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(18173,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(18173,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(18173,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(18173,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(18173,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(18173,36,'cp -16','c8b7 e1g1 e7e6 d5c6 b7c6 d2d4 c5d4 d1d4 g8f6 c1g5 f8e7 c3d5 c6d5 e4d5 h7h6 g5h4 d8c8 d5d6 e7d8 a2a4 e8g8 a4b5 a6b5 f3e5 f6d5 a1a8 c8a8 e5d7 d8h4 d4h4 f8d8 d7e5 d8d6 h4e4 f7f6 e5f3 a8a4 e4a4 b5a4 f1a1 d6c6 c2c3 d5b6 g1f1 c6c5 a1c1');
INSERT INTO evaluations VALUES(53011,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(53011,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(53011,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(53011,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(53011,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(53011,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53011,36,'cp -17','g8f6 d2d3 e7e6 g1f3 f8e7 e1g1 e8g8 f1e1 d6d5 e4d5 e6d5 c4b3 b8c6 b1c3 h7h6 c1f4 c8f5 f3e5 c6d4 c3e2 d4b3 a2b3 e7d6 d1d2 f8e8 e2g3 f5h7 e5g4 e8e1 a1e1 d6f8 d2e2 d8b6 e2f3 f6g4 f3g4 b6f6 g3h5 f6f5');
INSERT INTO evaluations VALUES(45147,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45147,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(45147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45147,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(45147,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(45147,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45147,36,'cp -2','d1e2 e7e6 g1f3 f8e7 c2c3 e8g8 d2d3 d6d5 c4b3 b7b5 e1g1 c8b7 f1d1 d8b6 e4e5 f6d7 a2a4 b6a6 c1f4 b5b4 c3c4 b8c6 b1d2 c6a5 b3a2 a6b6 d2b3 d5d4 b3a5 b6a5 a2b3 a5c7 f3d2 b7c6 d1e1 a7a5 d2e4');
INSERT INTO evaluations VALUES(14314,30,'cp -152',NULL);
INSERT INTO evaluations VALUES(14314,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(14314,32,'cp -147',NULL);
INSERT INTO evaluations VALUES(14314,33,'cp -154',NULL);
INSERT INTO evaluations VALUES(14314,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(14314,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(14314,36,'cp -144','d5e4 f3e5 c6e5 d4e5 d8d1 e1d1 f6g4 d1e1 c8f5 c1f4 g7g5 f4g5 e8c8 e5e6 h8g8 g5c1 f7e6 h2h3 g4f6 g2g4 f5g6 c1e3 f6d5 a1d1 a7a6 f1c4 c7c6 h3h4 f8g7 h4h5 g7c3 b2c3 g6f7 h1h4 b7b5 c4d5 e6d5 g4g5 e7e5 a2a3 f7e6 e1d2 d5d4 c3d4 e5d4 e3f4 c8b7');
INSERT INTO evaluations VALUES(14399,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(14399,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(14399,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(14399,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(14399,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(14399,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(14399,36,'cp -137','f3e5 c6e5 d4e5 d8d1 e1d1 f6g4 d1e1 c8f5 c1f4 g7g5 f4g5 e8c8 e5e6 h8g8 g5c1 f5e6 c3e4 e6f5 f2f3 f5e4 f3e4 d8d4 f1d3 g4e5 g2g3 e5d3 c2d3 d4d3 h1f1 d3d4 f1f7 d4e4 e1f1 f8g7 a1b1 g7d4 b2b3 h7h5 c1b2 d4c5 b1e1 e4e1 f1e1 c5g1 f7f3 b7b6 e1f1 g1c5 b2e5 c8d7 f3f5 h5h4 g3h4 g8g1 f1e2 g1g2 e2d3 g2a2');
INSERT INTO evaluations VALUES(14351,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(14351,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(14351,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(14351,33,'cp -143',NULL);
INSERT INTO evaluations VALUES(14351,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(14351,35,'cp -151',NULL);
INSERT INTO evaluations VALUES(14351,36,'cp -128','c6e5 d4e5 d8d1 e1d1 f6g4 d1e1 c8f5 c1f4 g7g5 f4g5 e8c8 e5e6 h8g8 g5c1 f5e6 c3e4 e6d5 e4c3 d5c6 f2f3 g4h6 c1f4 e7e5 f4g3 h6f5 a1d1 d8d1 e1d1 f5e3 d1c1 e3f1 h1f1 f7f5 g3e5 f8h6 c1b1 g8g2 a2a4 a7a6 h2h4 h6d2 b2b3 b7b6 h4h5 f5f4 c3e4 d2e3 e4f6 g2h2 f6g4 h2h5 g4e3');
INSERT INTO evaluations VALUES(24990,30,'cp 220',NULL);
INSERT INTO evaluations VALUES(24990,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(24990,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(24990,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(24990,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(24990,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(24990,36,'cp 243','d1b3 e7e6 b3b7 g4f3 g2f3 c6b4 c4b5 f6d7 b5d7 e8d7 b7b5 c7c6 b5e2 d8a5 e1f1 c6c5 d4c5 a5c5 b1c3 f8e7 h1g1 g7g6 a2a3 b4d5 c3d5 e6d5 f1g2 c5c4 e2d2 a7a5 g1d1 h8d8 e3e4 d7e8 e4d5 a5a4 d2f4 c4f4 c1f4 e7d6 f4d6 d8d6 a1c1 e8e7');
INSERT INTO evaluations VALUES(27541,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(27541,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(27541,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(27541,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(27541,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(27541,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(27541,36,'cp -24','c6d4 b1d2 f8c5 f3d4 c5d4 d1h5 g8e7 d2f3 e7g6 h2h4 e8g8 c1d2 f7f6 e5f6 d4f6 d3g6 h7g6 h5g6 d7e8 g6g4 e6e5 a1d1 a8d8 d2e3 d5d4 e3g5 d8d5 g5f6 b6f6 f1e1 e8a4 d1d2 a4c6 g4g3 f8e8 f3g5 f6g6 g3b3 e5e4');
INSERT INTO evaluations VALUES(27508,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27508,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27508,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(27508,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27508,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(27508,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(27508,36,'cp -10','b1d2 f8c5 f3d4 c5d4 d1h5 g8e7 d2f3 e7g6 h2h4 e8g8 c1d2 f7f6 e5f6 d4f6 d3g6 h7g6 h5g6 d7e8 g6g4 e6e5 a1d1 a8d8 d2e3 d5d4 e3g5 b6d6 g5f6 d6f6 g4e4 e8c6 e4e5 c6f3 e5f6 f8f6 g2f3 f6f3 g1g2 f3f5 d1d3 f5b5 b2b3 b5h5 g2g3 h5a5 f1d1 a5a2 d3d4 d8d4 d1d4 a2a1 g3g2');
INSERT INTO evaluations VALUES(27507,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(27507,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(27507,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(27507,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(27507,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(27507,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(27507,36,'cp -15','f8c5 f3d4 c5d4 d1h5 g8e7 d2f3 e7g6 h2h4 e8g8 c1d2 f7f6 e5f6 d4f6 d3g6 h7g6 h5g6 d7e8 g6g4 e6e5 a1d1 a8d8 d2e3 b6d6 f3g5 d5d4 g5e4 d6c6 e3h6 d8d5 g4g3 c6d7 d1c1 f6d8 e4c5 d7f7 c5e6 d8h4 g3g7 f7g7 h6g7 f8f5 e6c7 e8b5 c7d5 b5f1 c1c7 f1e2 d5e7 h4e7 c7e7');
INSERT INTO evaluations VALUES(47906,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(47906,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(47906,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(47906,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(47906,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(47906,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(47906,36,'cp 32','b1c3 f8e7 f1e2 e8g8 e1g1 c7c5 f4e5 d6e5 e2c4 c8e6 d2d3 h7h6 d1e1 e6c4 d3c4 b8c6 c3d5 f6d5 c4d5 c6d4 c2c3 d4f3 f1f3 f7f5 c3c4 f5f4 e1e2 e7d6 c1d2 f8f7 a1f1 d8e7 g2g3 a8f8 g1h1 a7a6 a2a4 e7e8 a4a5 f4g3 f3f7 f8f7 f1f7 e8f7 h2g3');
INSERT INTO evaluations VALUES(52743,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52743,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52743,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(52743,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(52743,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52743,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52743,36,'cp 25','d2d4 g7g6 f1g2 g8f6 c4d5 c6d5 g1f3 f8g7 e1g1 e8g8 b1c3 f6e4 f3d2 f7f5 d2e4 f5e4 f2f3 e4f3 g2f3 e7e6 c1e3 b8c6 d1d2 c8d7 f3g2 f8f1 a1f1 a8c8 f1d1 c6a5 b2b3 a5c6 c3a4 b7b6');
INSERT INTO evaluations VALUES(52750,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52750,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52750,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(52750,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(52750,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(52750,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52750,36,'cp 27','g7g6 g1f3 g8f6 f1g2 f8g7 e1g1 e8g8 c4d5 c6d5 b1c3 f6e4 f3d2 e4c3 b2c3 c8e6 e2e4 d5e4 d2e4 e6d5 f1e1 b8d7 a1b1 d7b6 e4c5 d5g2 g1g2 d8d5 d1f3 e7e6 c5b7 d5a2 c1f4 a2d5 f3d5 b6d5 f4d2 f8c8 e1c1 a7a5 c3c4 d5b4 b7d6');
INSERT INTO evaluations VALUES(992,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(992,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(992,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(992,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(992,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(992,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(992,36,'cp -66','b1c3 f8b4 c1d2 b4c3 d2c3 f7f6 d1e2 e8g8 a2a4 c8e6 e1g1 e6g4 f1e1 a8d8 h2h3 g4f3 e2f3 d5f3 g2f3 g7g5 e1e4 a7a6 h3h4 h7h6 g1f1 g8f7 f1e2 f7g6 a1g1 g6f5 e4c4 d8d7 a4a5 f8e8 e2d1 e8d8 h4g5 f6g5 b2b4 f5g6');
INSERT INTO evaluations VALUES(988,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(988,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(988,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(988,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(988,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(988,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(988,36,'cp -72','f8b4 c1d2 b4c3 d2c3 f7f6 d1e2 c8e6 e1g1 e6g4 f1e1 e8g8 a2a4 a7a6 e2e4 g4f3 g2f3 d5d7 e4g4 a8d8 g4d7 d8d7 e1e4 g7g5 h2h4 h7h6 g1f1 g8f7 f1e2 f8e8 h4g5 h6g5 a1h1 f7g6 e4c4 c6d8 c4b4 c7c6 a4a5 d8e6 e2d1 e8d8 b4b6 e6g7 c3d2 g7f5');
INSERT INTO evaluations VALUES(508,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(508,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(508,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(508,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(508,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(508,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(508,36,'cp -68','c1d2 b4c3 d2c3 f7f6 d1e2 c8f5 e1g1 f5g4 f1e1 g4f3 e2f3 d5f3 g2f3 e8f7 a2a4 g7g5 e1e4 h7h6 g1f1 a8d8 h2h4 a7a6 e4c4 f7g6 f1e2 h8h7 a1g1 g6f5 a4a5 d8d5 b2b4 c6d8 c3d2 d8e6 d2e3 h6h5 c4e4 d5d8');
INSERT INTO evaluations VALUES(509,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(509,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(509,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(509,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(509,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(509,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(509,36,'cp -78','b4c3 d2c3 f7f6 d1e2 a7a5 d3d4 e5e4 f3d2 f6f5 e2c4 c8e6 c4d5 e6d5 a2a4 e8c8 e1c1 h8e8 h1e1 b7b6 b2b3 f5f4 d2c4 e8e6 c4e5 e6h6 h2h3 g7g5 d1d2 d8e8 c1b1 c6e5 d4e5 d5c6 c3d4 h6h4');
INSERT INTO evaluations VALUES(20836,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(20836,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(20836,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(20836,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(20836,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(20836,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(20836,36,'cp -17','e7e5 f4g3 h7h5 h2h3 h5h4 g3h2 c8e6 d4e5 f6e5 f1b5 f8d6 g1f3 d8f6 c3c4 d5c4 b5c6 b7c6 b1d2 f6g6 d1b1 g6b1 a1b1 h8h5 e1g1 c4c3 b2c3 g8f6 f1d1 e8c8 d2b3 f6e4 b1c1 e6c4 f3d2 e4d2 d1d2 e5e4 d2d4 d6a3 d4d8 c8d8 c1d1 h5d5 d1d5 c4d5 h2e5 g7g5 b3d2 a3d6 e5d4');
INSERT INTO evaluations VALUES(22259,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22259,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(22259,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(22259,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(22259,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22259,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22259,36,'cp 0','f4g3 h7h5 h2h3 c8e6 b1d2 h5h4 g3h2 f8d6 d4e5 d6e5 g1f3 e5h2 f3h2 d8d6 b2b4 e8c8 b4b5 c6e5 a2a4 g8e7 h2f3 g7g5 a4a5 c8b8 f1e2 e6f5 f3d4 e5d3 e2d3 f5d3 d2f3 d3c4 f3d2');
INSERT INTO evaluations VALUES(22264,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22264,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22264,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22264,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22264,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22264,36,'cp 0','h7h5 h2h3 c8e6 b1d2 h5h4 g3h2 f8d6 d4e5 d6e5 g1f3 e5h2 f3h2 d8d6 b2b4 e8c8 b4b5 c6e5 a2a4 g8e7 h2f3 g7g5 a4a5 c8b8 f1e2 e6f5 f3d4 e5d3 e2d3 f5d3 d2f3 d3c4 f3d2');
INSERT INTO evaluations VALUES(34919,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34919,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(34919,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(34919,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(34919,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(34919,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(34919,36,'cp 13','b1c3 d5a5 g1f3 b8c6 f1e2 g8f6 e1g1 f8b4 a2a3 e8g8 h2h3 c8e6 a1b1 b4d6 d2d4 c6e7 c1d2 a5b6 d2g5 e6b3 d1d3 e7d5 c3b5 b3a4 b5d6 b6d6 b1c1 h7h6 g5d2 f8e8 f1e1 c7c6');
INSERT INTO evaluations VALUES(34918,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34918,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34918,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(34918,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(34918,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(34918,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34918,36,'cp 7','d5a5 d2d4 g8f6 f1c4 b8c6 g1e2 f8d6 e1g1 e8g8 a2a3 f8e8 h2h3 a5h5 c3b5 c6a5 b5d6 c7d6 c4d3 c8e6 d1a4 a5b3 e2f4 h5a5 a4a5 b3a5 f4e6 e8e6 c1e3 f6d5 a1e1 a8e8 d3b5 e8f8 g1h1 d5e3 e1e3 e6e3 f2e3 d6d5');
INSERT INTO evaluations VALUES(34986,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(34986,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(34986,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(34986,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(34986,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(34986,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(34986,36,'cp 14','d2d4 g8f6 f1c4 b8c6 g1e2 f8d6 e1g1 e8g8 a2a3 a5h5 c1f4 d6f4 e2f4 h5g5 f4d5 f8d8 d1c1 g5c1 a1c1 f6d5 c3d5 c6d4 d5c7 a8b8 f1d1 c8f5 c7d5 d4c6 d5e3 f5e6 c4e6 f7e6 f2f4 g8f8 g1f2 e6e5 f4e5 c6e5 f2g3 d8d1 c1d1 e5c6 e3c4');
INSERT INTO evaluations VALUES(45350,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45350,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45350,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45350,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45350,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45350,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45350,36,'cp 0','c5d4 e3d4 g7g6 c4d5 f6d5 d1b3 f8g7 f1c4 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(45421,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45421,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45421,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45421,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45421,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45421,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45421,36,'cp 0','e3d4 g7g6 c4d5 f6d5 d1b3 f8g7 f1c4 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(9088,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(9088,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(9088,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(9088,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(9088,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(9088,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(9088,36,'cp 72','e4d5 f6d5 f1e1 e8g8 f3e5 c6e5 e1e5 c7c6 d1f3 c8e6 b1d2 b7b5 c4b3 c5b6 d2e4 b6d4 e5h5 g7g6 f3g3 f8e8 h5h4 a7a5 e4d6 d4f6 h4e4 e8f8 a2a4 f6e7 e4e6 f7e6 d6e4 e7f6 c2c3 b5b4 d3d4 b4c3 b2c3 d8b8 a1b1 b8g3 h2g3 g8g7 c1d2 f8f7 b3c2');
INSERT INTO evaluations VALUES(9075,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(9075,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(9075,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(9075,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(9075,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(9075,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(9075,36,'cp 76','f6d5 f1e1 e8g8 f3e5 c6e5 e1e5 c7c6 d1f3 c8e6 b1d2 b7b5 c4b3 c5b6 a2a4 d5b4 f3e2 d8d6 e5e4 a8d8 e4e6 f7e6 d2f3 b4d5 c1d2 f8f3 g2f3 b6c7 g1f1 c7b6 b3d5 e6d5 f3f4 d6h6 e2f3 b5a4 f3g3 h6f6 d2c3 b6d4 a1a4 d4c3 b2c3 f6c3 f1g2 c3c2');
INSERT INTO evaluations VALUES(13054,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(13054,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13054,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(13054,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13054,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(13054,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13054,36,'cp 0','b7c6 h2h3 g7g6 f1c4 f8g7 e4e5 d6e5 d1d8 e8d8 c1g5 d8e8 e1c1 h7h5 c4b3 c8f5 g5e3 g7h6 e3h6 h8h6 g2g4 h5g4 h3g4 f6g4 f2f3 g4h2 d1d2 h2g4');
INSERT INTO evaluations VALUES(22947,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(22947,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(22947,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(22947,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(22947,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(22947,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(22947,36,'cp -3','f1c4 g8f6 g1f3 f8d6 e1g1 e8g8 f1e1 c6e5 f3e5 d6e5 f2f4 e5c3 b2c3 f8e8 c1a3 d7d5 e4d5 c8f5 d1b3 f6g4 h2h3 g4e3 c4b5 c7c6 d5c6 f5c2 b3b4 d8d5 e1e3 e8e3 b4c5 d5c5 a3c5 e3c3 c6b7 a8b8 c5d4 c3c7 d4e5 b8b7 e5c7 b7b5 a1e1 f7f6 e1e7 c2f5 c7d6');
INSERT INTO evaluations VALUES(48043,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(48043,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(48043,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(48043,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(48043,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(48043,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(48043,36,'cp 119','e2e4 f8e7 f1d3 e6e5 d4d5 e8g8 h2h3 a7a5 c1e3 c8d7 a2a3 b8a6 g1f3 c7c6 e1g1 a8c8 f3d2 c6d5 c4d5 a6c5 d3e2 d8e8 a3a4 e8d8 a1b1 h7h6 f1e1 f6h7 d2c4 h7g5 c4a5 g5e4 c3e4 c5e4 a5b7 d8e8 a4a5 f7f5 a5a6 f5f4 e3b6 e8g6');
INSERT INTO evaluations VALUES(48047,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(48047,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(48047,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(48047,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(48047,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(48047,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(48047,36,'cp 107','f8e7 h2h3 e8g8 f1d3 e6e5 d4d5 b8d7 g1f3 a7a6 a2a3 f6h5 d3c2 e7f6 g2g3 g7g6 d1e2 f6g7 c1e3 b7b6 e1g1 d8e8 f3h4 d7f6 g1g2 a8b8 b2b4 e8d7 f1h1 c7c6 a3a4 c8b7 h1d1 b8c8 h4f3 c6d5 c3d5 f6d5 c4d5');
INSERT INTO evaluations VALUES(48294,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48294,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48294,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48294,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(48294,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48294,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(48294,36,'cp -3','g1f3 f8d6 e2e3 d6f4 e3f4 e8g8 f1d3 b7b6 e1g1 c7c5 c2c3 d8c7 f3e5 c8a6 d3a6 b8a6 f1e1 a6b8 b1d2 b8c6 d2f1 f6e4 f1d2 e4d6 e5c6 c7c6 d2f3 d6e4 g2g3 c6b7 g1g2 a8c8 h3h4 b6b5 d1e2 b5b4 c3b4 b7b4 d4c5 b4c5');
INSERT INTO evaluations VALUES(22038,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(22038,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(22038,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(22038,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(22038,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(22038,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(22038,36,'cp 110','c3d4 d6d5 e4d5 f8b4 b1c3 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 e2d3 c8g4 f1e1 d8d7 h2h3 g4f3 d1f3 e8c8 c1e3 f7f5 c3b5 a7a6 b5d6 d7d6 d3f5 e7f5 f3f5 d6d7 f5c5 d7d5 a1d1 h8e8 c5c2 g7g6 f2f3 d5f7 c2c3 f7f6 f3f4 e8e4 d4d5 f6c3');
INSERT INTO evaluations VALUES(22030,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(22030,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(22030,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(22030,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(22030,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(22030,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(22030,36,'cp 108','d6d5 e4d5 f8b4 b1c3 d8d5 f1e2 g8e7 e1g1 d5d8 a2a3 b4d6 e2d3 e8g8 d4d5 c6b8 c1g5 f7f6 g5e3 b8d7 d1c2 h7h6 c3b5 g8h8 c2c4 b7b6 a1c1 c8a6 c1d1 a6b5 c4b5 f6f5 e3d4 d7f6 f1e1 f6d5 b5b3 d5f4');
INSERT INTO evaluations VALUES(21927,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(21927,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(21927,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(21927,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(21927,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(21927,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(21927,36,'cp 22','g7g6 a1b1 a7a5 e2e3 g8e7 g1e2 f8g7 e1g1 e8g8 b2b3 c8f5 e3e4 f5g4 d2d3 c6d4 h2h3 g4e2 c3e2 d4e6 c1e3 e7c6 d1d2 d8e7 h3h4 h7h5 g1h2 g8h7 g2h3 b7b6 a2a3 f8e8 h3e6 e7e6 b1e1 e8f8 b3b4 a5b4 a3b4 c6d4 e2d4 e5d4 e3f4');
INSERT INTO evaluations VALUES(26119,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(26119,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(26119,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(26119,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(26119,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(26119,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(26119,36,'cp 23','f1d3 f5d3 c2d3 g8e7 h2h4 h7h5 c3e2 e7f5 c1g5 f8e7 g2g3 d7f8 g5e7 d8e7 a2a3 f8g6 e1g1 e8g8 d1d2 a7a5 a1c1 f7f6 g1g2 e7d7 f1e1 a5a4 e2f4 g6f4 d2f4 d7f7 f4d2 f8e8 c1b1 f7g6 b2b4 e8f8 e5f6 f8f6 d2g5 a8f8 g5g6 f6g6');
INSERT INTO evaluations VALUES(26120,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26120,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(26120,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26120,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(26120,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26120,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(26120,36,'cp 13','f5d3 c2d3 g8e7 h2h4 h7h5 c3e2 e7f5 c1g5 f8e7 g2g3 d7f8 g5e7 d8e7 d1d2 f8g6 a1c1 f7f6 e1g1 e8g8 g1g2 a7a5 a2a3 e7c7 e2g1 a8e8 f1e1 c7b6 b2b4 a5b4 a3b4 f5d4 e5f6 d4f3 g1f3 f8f6 f3g5 e8f8 f2f4 e6e5 f4e5 b6f2 d2f2 f6f2 g2g1');
INSERT INTO evaluations VALUES(26105,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26105,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(26105,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26105,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(26105,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26105,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26105,36,'cp 33','c2d3 g8e7 h2h4 h7h5 c3e2 e7f5 c1g5 f8e7 g2g3 e8g8 e1g1 b7b6 a1c1 c6c5 g1g2 a8c8 d1d2 a7a5 f1e1 a5a4 g5e7 d8e7 e2f4 g7g6 f4e2 c8a8 c1c2 d7b8 e1c1 b8d7 g2g1 f8c8 d2f4 g8g7 e2c3 a8a5 g1g2 c8b8 c1h1 a5a8 c3b5');
INSERT INTO evaluations VALUES(33087,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(33087,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(33087,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(33087,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(33087,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(33087,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(33087,36,'cp 178','b1c3 f6d8 c3a4 b8d7 a4c5 d7c5 d2d4 c5e4 f1e1 d6d5 e1e4 d5e4 f3e5 c8e6 c4e6 f7e6 d1h5 g7g6 e5g6 g8f6 h5h3 h7g6 h3h8 e8f7 h8d8 a8d8 c1e3 f6g4 h2h3 g4e3 f2e3 e6e5 a1f1 f7e6 d4e5 d8d2 f1f6 e6e5 f6g6 d2c2 h3h4 c2b2 h4h5 b2b6 g6g7 c7c5 g2g4 c5c4 g1f2 b6b2 f2g3');
INSERT INTO evaluations VALUES(33079,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(33079,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(33079,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(33079,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(33079,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(33079,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(33079,36,'cp 198','c8e6 c3d5 e6d5 e4d5 h7h6 b2b4 c5b4 c2c3 b8d7 a1b1 b4a5 b1b7 f6d8 f1e1 g8e7 d1a4 a5b6 c1a3 e8g8 a3d6 e7f5 d6f8 g8f8 f3e5 d7e5 e1e5 f5d6 b7b6 a7b6 a4b4 d8h4 d2d3 h4g4 g2g3 g4d1 g1g2 a8a4 b4b3 d1b3 c4b3 a4a3 c3c4');
INSERT INTO evaluations VALUES(57497,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(57497,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(57497,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(57497,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(57497,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(57497,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(57497,36,'cp 38','d7d6 f1g2 f8e7 g1f3 g8f6 b2b4 e8g8 d2d3 d8e8 e1g1 a7a6 a2a4 b8c6 c1a3 a6a5 b4b5 c6b4 a3b4 a5b4 c3d5 f6d5 c4d5 e5e4 f3d4 e4e3 f2e3 e7g5 d4e6 c8e6 d5e6 e8e6 g1h1 c7c6 a1b1 d6d5 d1b3 g5e3 b3b4 e6h6 b4b3 f5f4 b5c6 b7c6 g3f4 e3f4');
INSERT INTO evaluations VALUES(18086,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(18086,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(18086,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(18086,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(18086,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(18086,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(18086,36,'cp 67','e1g1 d7d6 d2d4 e5d4 f1e1 a5b3 a2b3 c7c5 c2c3 g8f6 c3d4 f8e7 d4c5 d6c5 d1d8 e7d8 b1c3 c8b7 e4e5 f6d7 e5e6 f7e6 e1e6 e8f7 e6d6 d7f6 f3g5 f7g6 g5e6 a8c8 c1e3 d8c7 e6c7 c8c7 f2f3 h8e8 g1f2 g6f7 a1d1 h7h6 h2h4 b5b4 c3a4 c5c4 b3c4 c7c4 a4c5 b7c8 c5d3');
INSERT INTO evaluations VALUES(18085,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(18085,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(18085,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(18085,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(18085,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(18085,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(18085,36,'cp 71','d7d6 d2d4 e5d4 f1e1 a5b3 a2b3 c7c5 c2c3 f8e7 c3d4 g8f6 d4c5 d6c5 d1d8 e7d8 b1c3 e8g8 c3b5 c8e6 c1f4 e6b3 f3d2 b3e6 a1a3 a6a5 e1a1 a8a6 b5c7 a6b6 b2b3 f6d7 c7e6 b6e6 f4e3 d8c7 f2f3 f7f5 e3f2 f5e4 d2e4 f8e8');
INSERT INTO evaluations VALUES(18027,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(18027,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(18027,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(18027,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(18027,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(18027,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(18027,36,'cp 65','d2d4 e5d4 f1e1 a5b3 a2b3 c7c5 c2c3 g8f6 c3d4 c8b7 b1c3 f8e7 f3h4 b5b4 c3d5 f6d5 e4d5 e8g8 h4f5 e7f6 c1f4 b7c8 f5d6 f6d4 d1e2 a6a5 a1d1 d8f6 f4e5 f6e5 e2e5 d4e5 e1e5 f7f6 e5e7 f8d8 d6c4 c8f5 d1c1 f5d3 c4b6');
INSERT INTO evaluations VALUES(19307,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19307,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(19307,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(19307,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19307,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19307,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(19307,36,'cp 0','f3d4 f8g7 d4c6 b7c6 b1c3 d7d6 e1g1 g8f6 e4e5 d6e5 d1d8 e8d8 f1d1 d8c7 d1e1 e7e6 e1e5 f6d5 e5e4 a7a5 c4b3 d5c3 b2c3 h8e8 c1f4 e6e5 b3f7 e8f8 e4e5 g7e5 f4e5 c7b6 e5d4 b6c7');
INSERT INTO evaluations VALUES(49643,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(49643,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(49643,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(49643,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(49643,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(49643,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(49643,36,'cp -22','d7d5 g1f3 b8c6 b1d2 f8d6 e1g1 e8g8 f1e1 f8e8 e4d5 f6d5 e2f1 h7h6 c2c3 a7a5 h2h3 c8f5 d2c4 a5a4 c4d6 c7d6 c1d2 d8f6 a1c1 d5e7 d3d4 f5e6 d4e5 d6e5 c3c4 e7f5 d2c3 e8d8 f1d3 f5d4 f3d4 e5d4 c3d2');
INSERT INTO evaluations VALUES(48840,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(48840,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48840,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(48840,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(48840,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(48840,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(48840,36,'cp -13','g1f3 b8c6 b1d2 f8d6 e1g1 e8g8 f1e1 f8e8 e2f1 h7h6 e4d5 f6d5 h2h3 c8f5 c2c3 d6f8 a2a4 f5g6 a4a5 d8d7 d2c4 f7f6 d3d4 e5d4 e1e8 a8e8 f3d4 d5b4 c1e3 c6d4 d1d4 d7d4 e3d4 b4c2 a1c1 c2d4 c3d4 e8d8 a5a6 b7b6 c4e3 d8d4 c1c7');
INSERT INTO evaluations VALUES(2036,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(2036,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(2036,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(2036,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(2036,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(2036,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(2036,36,'cp 206','g1f3 d7d6 e5d6 e7f6 d1c1 f8d6 f4d6 f6d6 c1d2 d6d2 b1d2 c8d7 e1c1 g8f6 h2h3 e8c8 e2e3 h8e8 f1c4 d7e6 c4b5 e6d7 a2a3 h7h6 h1e1 a7a6 b5d3 g7g5 b2b3 e8e7 c1b2 c6e5 f3e5 e7e5 b3b4 d7b5 d2b3 f6e4 d3e4 e5e4 d1d8 c8d8');
INSERT INTO evaluations VALUES(30130,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(30130,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(30130,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(30130,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(30130,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(30130,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30130,36,'cp 10','c2c4 c7c6 b1c3 h7h6 d1b3 d8b6 c4c5 b6b3 a2b3 f8e7 e1g1 a7a6 b3b4 e8g8 a1a3 b8d7 h2h3 a8e8 g3g4 f5h7 c1f4 g7g5 f4g3 h6h5 f3g5 h5g4 g5h7 g8h7 h3g4 f6g4 e2e4 f7f5 e4d5 e6d5 c3d5 c6d5 g2d5 e7f6');
INSERT INTO evaluations VALUES(30072,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(30072,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(30072,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(30072,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30072,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(30072,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(30072,36,'cp 0','f8e7 b1c3 c7c6 d1b3 d8b6 c4c5 b6b3 a2b3 e8g8 e1g1 b8d7 b3b4 a7a6 a1a3 h7h6 c1e3 f6e4 f1a1 f8e8 g2f1 e7f6 b4b5 c6b5 c3b5 a8c8 b5a7 c8a8');
INSERT INTO evaluations VALUES(55169,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(55169,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(55169,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55169,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(55169,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(55169,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(55169,36,'cp -7','b8c6 d1e2 e7e6 d2d3 d8d7 a2a3 b7b6 b1d2 c8b7 f2f4 c6e7 g1f3 h7h5 d2e4 g8h6 e1g1 h6f5 e2f2 e7g6 b2b4 a7a6 c1d2 f8e7 c4b3 e8c8 a3a4 c8b8 a1b1 h5h4 f1e1 f5e3 d2e3 d4e3 f2e3 h4h3 g2g3 e7b4');
INSERT INTO evaluations VALUES(21091,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(21091,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(21091,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21091,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21091,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21091,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(21091,36,'cp -29','d1e2 e7e6 d2d3 d8d7 a2a3 b7b6 f2f4 c8b7 b1d2 c6e7 d2e4 h7h5 g1f3 g8h6 e1g1 h6f5 h2h3 h5h4 e2f2 c7c5 a3a4 b7c6 c4b5 e7d5 c2c4 d4c3 b2c3 c6b5 a4b5 d5c7 f1d1 d7b5 c3c4 b5c6 c1b2 a7a5 f3g5 f8e7 e4d6 e7d6 e5d6 c6d6 b2e5 d6c6 d3d4');
INSERT INTO evaluations VALUES(15009,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(15009,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(15009,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(15009,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(15009,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(15009,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(15009,36,'cp -70','d8d5 d1b3 a7a5 f1e2 a5a4 b3d1 f8c5 e1g1 e8g8 c1g5 f8d8 b1d2 c5e7 f1e1 h7h6 g5h4 c8f5 d2c4 e5e4 d3e4 d5c5 f3d2 f6e4 d2e4 f5e4 h4e7 c6e7 d1c1 e7g6 c1e3 c5e3 c4e3 g6f4 e2f1 a4a3 b2b3 g8f8 a1d1 f4d5 e3d5 e4d5 c3c4');
INSERT INTO evaluations VALUES(957,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(957,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(957,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(957,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(957,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(957,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(957,36,'cp -40','d1b3 a7a5 f1e2 a5a4 b3d1 c8f5 e1g1 f8c5 c1g5 e8g8 b1d2 f5d3 e2d3 d5d3 g5f6 g7f6 f1e1 c5e7 d2e4 d3d1 a1d1 f8d8 f3h4 d8d1 e1d1 a8d8 d1d8 c6d8 g1f1 d8e6 b2b3 a4a3 f1e2 g8f8 e2d2 e6g7 g2g4 h7h5 h2h3 f8e8 d2c2 e8d7');
INSERT INTO evaluations VALUES(28941,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(28941,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(28941,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(28941,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(28941,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(28941,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(28941,36,'cp 49','c6d5 d1b3 f5c8 c1g5 e7e6 e2e3 b8c6 a1c1 c8d7 f3e5 f8e7 h2h3 h7h6 g5h4 c6a5 b3d1 a8c8 f1d3 e8g8 e1g1 a5c4 c1c2 a7a6 d1f3 d7b5 c3b5 a6b5 b2b3 c4d6 c2c8 d8c8 e5g4 c8d8 g4f6 e7f6 h4g3 f6e7 g3d6 e7d6 d3b5 d8a5 b5d3 f8c8 f3e2');
INSERT INTO evaluations VALUES(36518,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(36518,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(36518,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36518,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(36518,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(36518,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(36518,36,'cp 23','c7c5 d2d4 b8c6 d4d5 c6a5 f3d2 e7e6 a1b1 e6d5 c4d5 c8f5 e2e4 f5d7 b2b3 b7b5 c3e2 a8c8 c1b2 f8e8 b2a1 b5b4 e2f4 c5c4 b3c4 a5c4 b1b4 d7g4 f4e2 c4d2 a1f6 d8f6 d1d2 c8b8 e2f4 e8c8 b4b8 c8b8 f4d3 g4d7 h2h3 f6c3 d2c3 g7c3 f1c1');
INSERT INTO evaluations VALUES(35938,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(35938,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(35938,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(35938,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(35938,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(35938,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(35938,36,'cp 30','d2d4 b8c6 d4d5 c6a5 f3d2 e7e6 a1b1 e6d5 c4d5 c8f5 e2e4 f5d7 b2b3 b7b5 c3e2 a8c8 c1b2 f8e8 h2h3 b5b4 f1e1 d7b5 b1c1 d8b6 e2f4 c5c4 b3c4 a5c4 d2c4 c8c4 c1c4 b5c4 b2d4 b6d8 d1c2 c4b5 c2b2 a7a5 a2a3 g6g5 f4e2 b5e2 e1e2');
INSERT INTO evaluations VALUES(35943,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(35943,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(35943,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(35943,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(35943,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(35943,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(35943,36,'cp 35','b8c6 d4d5 c6a5 f3d2 e7e6 a1b1 e6d5 c4d5 c8f5 e2e4 f5d7 b2b3 f8e8 d2c4 a5c4 b3c4 f6g4 c1d2 g4e5 d1e2 d8c8 c3b5 d7b5 b1b5 b7b6 d2f4 c8d7 a2a4 h7h5 h2h3 f7f5 a4a5 h5h4 f1b1 f5e4 e2e4 h4g3');
INSERT INTO evaluations VALUES(44247,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(44247,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44247,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44247,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(44247,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44247,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(44247,36,'cp 25','e4e5 f6d5 d2d4 c5d4 c3d4 d7d6 f1c4 f8e7 e1g1 e8g8 d1e2 d6e5 d4e5 c8d7 c4d5 e6d5 b1c3 d7e6 c1e3 b8c6 a1d1 d8d7 h2h3 a8d8 c3b5 a7a6 b5d4 e6f5 a2a3 f5e4 d4c6 b7c6 b2b4 a6a5 e3b6 a5b4 b6d8 f8d8 a3b4 e7b4 d1c1');
INSERT INTO evaluations VALUES(44163,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44163,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(44163,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(44163,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(44163,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(44163,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44163,36,'cp 12','f6d5 d2d4 c5d4 c3d4 d7d6 f1c4 f8e7 e1g1 e8g8 d1e2 b8d7 c4d5 e6d5 b1c3 d7b6 h2h3 c8e6 b2b3 a8c8 c3b5 a7a6 b5d6 e7d6 e5d6 d8d6 e2e5 d6c7 c1a3 f8d8 e5c7 c8c7 f1c1 c7c1 a1c1 d8c8 c1c5 c8c5 d4c5');
INSERT INTO evaluations VALUES(44143,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44143,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44143,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(44143,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44143,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44143,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44143,36,'cp 18','d2d4 c5d4 c3d4 d7d6 f1c4 f8e7 e1g1 e8g8 d1e2 b8d7 a2a3 a7a6 f1d1 b7b5 c4d5 e6d5 c1d2 d6e5 d4e5 f7f6 b1c3 d7e5 d2f4 c8g4 f4e5 f6e5 d1d5 d8c7 e2e5 c7e5 f3e5 g4e6 d5d2 a8c8 e5d7 e6d7 d2d7 e7f6 d7d6 f6c3 b2c3 c8c3 d6a6 c3b3 a1d1 b3b2 a6a7 b2f2');
INSERT INTO evaluations VALUES(44139,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44139,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(44139,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(44139,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(44139,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44139,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44139,36,'cp 19','c5d4 c3d4 d7d6 f1c4 f8e7 d1e2 e8g8 e1g1 b8d7 c4d5 e6d5 b1c3 d7b6 c1f4 f8e8 f1e1 d6e5 d4e5 c8f5 f3d4 f5g6 a1d1 a8c8 e2f3 a7a6 e5e6 e7c5 e1e5 c5d4 e6f7 g6f7 d1d4 e8e5 f4e5 d8g5 f3g3 g5g3 h2g3 h7h5 c3e2 b6c4');
INSERT INTO evaluations VALUES(44148,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44148,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(44148,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44148,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44148,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44148,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44148,36,'cp 30','c3d4 d7d6 f1c4 f8e7 e1g1 e8g8 d1e2 b8d7 f1e1 d6e5 d4e5 d7c5 e1d1 b7b6 c4d5 e6d5 b1c3 c8a6 e2c2 a6b7 c2f5 c5e6 c1e3 d8d7 c3e2 f8d8 e2d4 a8c8 d4e6 d7e6 f3d4 e6f5 d4f5 e7b4 a2a3 b4f8 f2f4 c8c4');
INSERT INTO evaluations VALUES(44146,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(44146,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44146,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(44146,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44146,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(44146,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(44146,36,'cp 12','d7d6 f1c4 f8e7 d1e2 d6e5 d4e5 e8g8 e1g1 c8d7 f1d1 d7c6 f3d4 e7c5 b1c3 d8b6 d4b3 d5c3 b2c3 b8d7 d1d3 a8c8 d3h3 g7g6 c4d3 c5e7 c1e3 b6c7 e3d4 f8d8 a1e1 e7f8 h3g3 f8g7 h2h4 h7h5 e2e3 b7b5 e3g5 d7f8 b3d2');
INSERT INTO evaluations VALUES(45962,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(45962,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(45962,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(45962,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(45962,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(45962,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(45962,36,'cp 29','f1c4 f8e7 e1g1 e8g8 d1e2 b8d7 a2a3 a7a6 h2h3 b7b5 c4d5 e6d5 c1f4 d7b6 f1e1 d6e5 f3e5 e7f6 b1c3 c8e6 b2b3 b6c8 e2d3 c8e7 c3e2 f8e8 a1c1 a8c8 d3d2 e7g6 c1c8 e6c8 b3b4 g6e5 f4e5 f6e5 d4e5 e8e5');
INSERT INTO evaluations VALUES(45960,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(45960,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45960,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(45960,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(45960,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(45960,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(45960,36,'cp 19','f8e7 e1g1 e8g8 d1e2 b8d7 c1d2 d6e5 d4e5 b7b6 c4d5 e6d5 b1c3 d7c5 f3d4 d8d7 e2f3 c5e4 c3e4 d5e4 f3e4 c8b7 e5e6 d7d5 e6f7 f8f7 e4d5 b7d5 a2a3 a8d8 d2c3 e7c5 a1d1 f7d7 d1d2 d5b7 f1d1 g8f7 b2b4 c5e7 d4f5 d7d2 d1d2 d8d2 c3d2 e7f8 f5d4 g7g6 h2h3 a7a6 d2e3');
INSERT INTO evaluations VALUES(45110,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45110,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(45110,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45110,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(45110,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(45110,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(45110,36,'cp 15','b8d7 b5d3 d7c5 b1c3 g7g6 e1g1 f8g7 f1e1 e8g8 d3f1 b7b6 a2a4 c8b7 f2f3 a7a6 c1e3 d8c7 d1d2 e7e6 e3h6 g7h6 d2h6 c5d7 h6g5 g8g7 a1d1 h7h6 g5e3 a8d8 d1d2 g7h7 h2h3 f6h5 f3f4 e6e5 d4e2 e5f4 e2f4 h5f4');
INSERT INTO evaluations VALUES(12261,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(12261,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(12261,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12261,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(12261,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(12261,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(12261,36,'cp 20','b5d3 d7c5 b1c3 g7g6 e1g1 f8g7 f1e1 e8g8 d3f1 b7b6 f2f3 c8b7 c1g5 a7a6 d1d2 d8c7 a2a4 e7e6 g5h6 g7h6 d2h6 c5d7 h6g5 g8g7 a1d1 a8c8 g5e3 h7h6 d1d2 d7e5 c3d1 e5d7 c2c4 d7c5 b2b3 g7h7 d1f2');
INSERT INTO evaluations VALUES(49697,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(49697,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(49697,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(49697,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(49697,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(49697,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(49697,36,'cp -32','d7d5 g1f3 b8c6 e1g1 c8e6 f3g5 e6g4 c2c4 d5d4 c4c5 a8b8 d1b3 g4h5 b3c4 f8e7 b2b4 f6d7 g5f3 f7f6 a2a3 h5f7 c4c2 a7a5 c1d2 b8a8 c2b2 e8g8 e2e3 d4e3 f2e3 e5e4 d3e4 a5b4 a3b4 a8a1 b2a1 f7c4 f1e1 d7e5 f3e5 c6e5 d2c3 d8d7 c3e5 f6e5');
INSERT INTO evaluations VALUES(57597,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(57597,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(57597,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(57597,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(57597,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(57597,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(57597,36,'cp 5','g8f6 c2c4 e7e6 b1c3 b7b6 e2e3 g7g6 f1e2 f8g7 a1b1 a7a5 e1g1 e8g8 a2a3 d7d6 b2b4 a5b4 a3b4 c7c5 d1b3 b8c6 b4b5 c6a5 b3c2 d6d5 d2d3 d5d4 c3d1 d4e3 d1e3 d8d6 f3e5 c8b7 c1b2 f6d7 e2f3 b7f3');
INSERT INTO evaluations VALUES(50110,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(50110,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(50110,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(50110,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(50110,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50110,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(50110,36,'cp 23','c2c4 c7c5 b1c3 e7e6 e2e3 b8c6 d2d4 d7d5 f1e2 c5d4 e3d4 f8b4 e1g1 e8g8 c1e3 b4c3 b2c3 b7b6 a1c1 c8a6 c4d5 a6e2 d1e2 e6d5 c3c4 a8c8 f3e5 f6e4 f1d1 d8d6 h2h3 h7h6 g1h2 f8e8 c4d5 d6d5 e2h5 d5e6 c1c6 c8c6 e5c6');
INSERT INTO evaluations VALUES(54833,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(54833,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(54833,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(54833,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(54833,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(54833,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(54833,36,'cp 53','d6d5 e4d5 d8d5 g1f3 f8g7 b1c3 d5d6 c1e3 g8f6 f3e5 b8c6 f1e2 c6b4 c3b5 d6d8 e3d2 b4c6 c2c3 a7a6 b5a3 c8f5 a3c4 f5e4 e2f3 e7e6 c4e3 c6e7 d1e2 e4f3 e2f3 f6d5 e1c1 d5e3 d2e3 d8d5 h1e1 d5f3');
INSERT INTO evaluations VALUES(49953,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(49953,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(49953,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(49953,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(49953,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(49953,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(49953,36,'cp 54','e2e3 f8e7 h2h3 e8g8 f1d3 b7b6 c2c4 c8b7 e1g1 f6e4 f3d2 e4d2 d1d2 d7d6 b1c3 b8d7 a1d1 e7h4 d2c2 d8e7 b2b4 d7f6 f4h2 a7a5 a2a3 a5b4 a3b4 g7g6 c3b5 f6e4 c4c5 b6c5 d4c5 b7d5 d3e4 d5e4 c2b3 e4c6 c5d6 c7d6');
INSERT INTO evaluations VALUES(51904,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(51904,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(51904,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(51904,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(51904,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(51904,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(51904,36,'cp -74','c5d4 c2c3 d8a5 f2f4 d7d6 g1f3 d6e5 f3e5 g8f6 d1d4 f8e7 b1d2 e8g8 f1d3 b8d7 d2f3 d7c5 d3c2 a5c7 d4f2 c5e4 f2e3 e4d6 f3d4 d6f5 d4f5 e6f5 e3e2 e7c5 c1e3 f6d5 e3d2 b7b5 e2b5 c8e6 e1c1 a8b8 b5a4 c5e7');
INSERT INTO evaluations VALUES(52032,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(52032,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(52032,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(52032,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(52032,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(52032,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(52032,36,'cp -66','c2c3 d8a5 f2f4 d7d6 g1f3 d6e5 f3e5 g8f6 d1d4 b8d7 b1d2 f8c5 d2b3 c5d4 b3a5 d4e5 f4e5 d7e5 c1f4 e5g6 f4d6 b7b6 a5c4 c8a6 a2a4 e8c8 d6g3 a6c4 f1c4 f6e4 a4a5 e4g3 h2g3 g6e5 c4e2 c8b7 h1h4 d8d5 a5b6 a7b6 e2f3 e5f3 g2f3 d5d7 e1e2 h7h5 h4a4 h8d8 a4a7 b7c6 a7d7 d8d7 a1a8 c6b7 a8h8 g7g6 e2e3 b7c6 f3f4 d7d5 h8h7');
INSERT INTO evaluations VALUES(56172,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(56172,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(56172,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(56172,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(56172,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(56172,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(56172,36,'cp 137','d5e6 c8e6 g1f3 g8f6 b1c3 f8d6 d2d4 e8g8 f1d3 b8c6 e1g1 h7h6 h2h3 c6b4 d3e2 c7c6 a2a3 b4d5 c3d5 c6d5 e2d3 f6e4 f3e5 d8c7 f1e1 f8e8 c2c3 f7f6 e5f3 e6f7 c1e3 f7h5 d1c2 f6f5 c3c4 d5c4 c2c4 c7c4 d3c4 h5f7');
INSERT INTO evaluations VALUES(3801,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(3801,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(3801,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(3801,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(3801,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(3801,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(3801,36,'cp -34','h2h3 c6a5 c4b3 a5b3 a2b3 a7a5 g5f3 f8e8 d2d3 h7h6 c1e3 c5e3 f2e3 c7c6 d1e1 c8e6 f3h4 b7b5 h4f5 e6f5 f1f5 e8e6 e1h4 d8e7 a1f1 f6h7 h4e1 a8f8 c3e2 g7g6 f5f2 b5b4 c2c3 d6d5 e4d5 c6d5 c3c4 h7f6 c4d5 f6d5 e1d2 d5f6 e3e4');
INSERT INTO evaluations VALUES(3799,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(3799,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(3799,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(3799,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(3799,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(3799,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(3799,36,'cp -38','c6a5 c4b3 h7h6 g5f3 a5b3 a2b3 a7a5 d2d3 f8e8 c1e3 c5e3 f2e3 c7c6 d1e1 c8e6 c3e2 b7b5 e1f2 d8b6 e2c3 e8f8 g1h1 a8e8 h1h2 b6c7 f2g3 c7d8 c3e2 c6c5 h2g1 g8h7 g3h4 d8b6 e2c3 b5b4');
INSERT INTO evaluations VALUES(16455,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(16455,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(16455,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(16455,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(16455,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(16455,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(16455,36,'cp 17','g2g3 d7d5 c4d5 f6d5 g1f3 f8e7 c3d5 d8d5 f1g2 e8g8 e1g1 d5d6 a2a3 a7a5 d1c2 c8e6 c1e3 a5a4 a1c1 d6d7 f3d2 f8b8 d2e4 a8a5 e4c5 e7c5 e3c5 e6b3 g2c6 d7c6 c2c3 a5a6 c5b4 c6c3 c1c3 c7c6 c3c5 f7f6 f2f3');
INSERT INTO evaluations VALUES(16454,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(16454,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(16454,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(16454,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(16454,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(16454,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16454,36,'cp 7','d7d5 c4d5 f6d5 g1f3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 a2a3 f7f6 d1a4 e7c5 b2b4 c5b6 c1e3 c6e7 a4d7 e8d7 f1c1 b6e3 f2e3 c7c6 f3d2 h7h5 h2h4 a8c8 g1f2 e7f5 g2h3 f5h6 h3e6 d7e6 e3e4 h8d8 c1c3 h6g4 f2f3 b7b6 a1c1 e6d7');
INSERT INTO evaluations VALUES(32365,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(32365,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(32365,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(32365,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(32365,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(32365,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(32365,36,'cp 82','b1c3 g8f6 g1f3 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 b8d7 c3c4 e8g8 a1b1 b7b6 e3e4 e6e5 d4e5 d8c7 c1b2 a8d8 d3e2 d7c5 f3d4 c5e6 d4f5 e7c5 h2h4 d8d7 h4h5 f8d8 e2g4 d7d2 e1f1 h7h6 b2c1');
INSERT INTO evaluations VALUES(32363,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(32363,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(32363,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(32363,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(32363,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(32363,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(32363,36,'cp 53','g8f6 g1f3 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 b8d7 c3c4 e8g8 a1b1 b7b6 e3e4 e6e5 d4e5 d8c7 c1b2 d7c5 d3c2 c5e6 f3d4 e7c5 d4b3 f8d8 e1d1 d8d1 b1d1 a8d8 g2g3 d8d1 c2d1 c5f8 g1g2 a7a5 a2a4 e6g5 d1e2 c7d7 h2h4 g5e6 h4h5 c6c5');
INSERT INTO evaluations VALUES(38692,30,'cp 221',NULL);
INSERT INTO evaluations VALUES(38692,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(38692,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(38692,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(38692,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(38692,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(38692,36,'cp 196','e5f3 c5b6 f1e2 c7c5 c2c4 b8c6 e1g1 g8e7 b1c3 e8g8 d2d3 f7f5 c3d5 h7h6 a2a3 c6e5 c1d2 e5g6 d5b6 a7b6 f3e1 f5f4 d2c3 e7c6 d3d4 g6h4 d4c5 b6c5 d1d5 g8h7 e2h5 g7g6 h5e2 f4f3 e2f3');
INSERT INTO evaluations VALUES(38751,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(38751,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(38751,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(38751,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(38751,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(38751,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(38751,36,'cp 205','c5b6 f1e2 g8f6 d2d3 b8d7 e1g1 c7c6 f1e1 e8g8 c1f4 d6d5 e4d5 f6d5 f4g3 d7f6 b1d2 f6h5 d2c4 c8g4 f3e5 g4e2 d1e2 h5g3 h2g3 f8e8 e2d2 b6c7 e5f3 h7h6 d3d4 d5b6 c4e3 b6d5 c2c3 d5e3 e1e3 e8e3');
INSERT INTO evaluations VALUES(45389,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(45389,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(45389,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(45389,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(45389,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(45389,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(45389,36,'cp 63','e4d5 f6d5 f1b5 c8d7 f3e5 d5f6 d1f3 d8c8 e5d7 b8d7 b2b3 a7a6 b5d7 c8d7 c1b2 a8c8 d2d3 e7e6 c3e4 f6e4 d3e4 d7c7 e1g1 f8d6 b2g7 d6h2 g1h1 h8g8 g7h6 g8g6 h6e3 h2e5 a1d1 c7c6 c2c4 h7h5 e3f4 e5f4 f3f4 g6g4 f4f6 c6e4 f2f3');
INSERT INTO evaluations VALUES(45330,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(45330,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(45330,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(45330,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(45330,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(45330,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(45330,36,'cp 49','f6d5 f1b5 c8d7 f3e5 d5f6 d1f3 d8c8 e5d7 b8d7 b2b3 a7a6 b5d7 c8d7 c1b2 a8c8 d2d3 e7e6 c3e4 f6e4 d3e4 d7c7 c2c4 c8d8 e1g1 f7f6 a1e1 h7h5 f3e2 c7f4 b2c1 f4g4 f2f3 g4g6 f3f4 f8d6 f1f3 h5h4 c1d2 b7b6 e2f2 g6h5 d2c3 e6e5 f4f5 h5h6 f3d3 d6e7 d3d8 e7d8 e1d1 e8g8');
INSERT INTO evaluations VALUES(45504,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(45504,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(45504,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(45504,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(45504,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(45504,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(45504,36,'cp 41','f1b5 c8d7 f3e5 d5f6 d1f3 d8c8 e5d7 b8d7 b2b3 a7a6 b5d7 c8d7 c1b2 a8c8 d2d3 e7e6 c3e4 f6e4 d3e4 c5c4 e1g1 d7c7 b3c4 f8d6 b2g7 d6h2 g1h1 h8g8 g7b2 h2e5 b2e5 c7e5 a1b1 b7b5 c4b5 g8g5 g2g3 a6b5 h1g2 c8c2 f1h1 c2c3 b1b5 c3f3 b5e5 g5e5 g2f3 e5a5 h1h7 a5a3 f3g4 a3a2');
INSERT INTO evaluations VALUES(16685,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(16685,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(16685,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(16685,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16685,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(16685,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(16685,36,'cp -19','f8b4 f1d3 d7d5 e4d5 f6d5 e1g1 b4c3 b2c3 e8g8 c1d2 c8g4 d1e1 g4h5 a1b1 b7b6 e1f2 f8e8 b1e1 d8d6 f3h4 e8e1 f1e1 h5g6 h4g6 h7g6 e1e4 a8f8 h2h4 d6d7 d3b5 f4f3 f2f3 d5f6 d2g5 c6d4 e4d4 d7b5 g5f6 g7f6 f3f6 b5h5 d4d8 f8d8');
INSERT INTO evaluations VALUES(10483,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(10483,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(10483,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(10483,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(10483,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(10483,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(10483,36,'cp -6','f1d3 d7d5 e4d5 f6d5 e1g1 b4c3 b2c3 e8g8 c1d2 c8g4 d1e1 g4h5 e1f2 f8e8 a1b1 b7b6 b1e1 d8d6 f3h4 h5g6 h4g6 h7g6 e1e8 a8e8 d2f4 d5f4 f2f4 d6f4 f1f4 c6e7 c3c4 f7f6 h2h4 g8f7 g1f2 c7c5 d4d5 e7c8 f4e4 e8e4 d3e4 a7a6 e4d3 c8d6 f2e3 f6f5 e3f4 f7f6 g2g4');
INSERT INTO evaluations VALUES(10484,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(10484,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10484,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(10484,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(10484,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(10484,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(10484,36,'cp -15','d7d5 e4d5 f6d5 e1g1 b4c3 b2c3 e8g8 c1d2 c8g4 d1e2 d8d7 a1b1 b7b6 e2f2 f7f6 f1e1 g4f5 c3c4 d5e3 d2e3 f5d3 c2d3 f4e3 f2e3 a8e8 e3f2 c6e7 d4d5 e7f5 e1e8 f8e8 b1e1 e8e1 f3e1 h7h5 h2h3 g8f7 f2f4 g7g6 e1f3 c7c6 f4b8 c6d5 c4d5 f5e7');
INSERT INTO evaluations VALUES(21037,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(21037,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21037,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(21037,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21037,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21037,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(21037,36,'cp -9','e7e5 g1f3 e5e4 f3d2 c6e7 c2c4 c7c6 b1c3 g8f6 f1e2 h7h5 c4d5 c6d5 f2f3 e7f5 d2e4 d5e4 f3e4 f5h6 e1g1 h5h4 e4e5 f6d5 e3e4 d5c3 b2c3 h4h3 g2g3 f8e7 c1h6 h8h6 d1b3 c8e6 d4d5 e6d7 d5d6 d8b6 b3b6 a7b6 d6e7 e8e7 e2c4 d7e6 c4b5 a8c8 a1c1');
INSERT INTO evaluations VALUES(21696,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21696,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21696,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21696,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21696,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21696,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21696,36,'cp 0','g1f3 e5e4 f3d2 c6e7 c2c4 c7c6 b1c3 g8f6 f1e2 a7a6 e1g1 e7f5 d1a4 c8e6 c3d5 f6d5 c4d5 e6d5 a4c2 f8d6 d2e4 d6h2 g1h2 d8h4 h2g1 d5e4 e2d3 e4g2 g1g2 h4g4 g2h2 g4h5 h2g1 h5g4');
INSERT INTO evaluations VALUES(28713,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(28713,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28713,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28713,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28713,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28713,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(28713,36,'cp 0','d7d5 c4d5 e6d5 f1b5 c7c6 b5d3 f8d6 e1g1 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 e4c2 b8d7 d1d3 d7f6 f3e5 d6e5 d4e5 d8d3 c2d3 f6d5 f1d1 a8d8 c1d2 b7c8 h2h3 d5e7 d2c3 e7d5');
INSERT INTO evaluations VALUES(28653,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28653,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28653,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28653,36,'cp 0','c4d5 e6d5 f1b5 c7c6 b5d3 f8d6 e1g1 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 e4c2 e8g8 d1d3 d7f6 f3e5 d6e5 d4e5 d8d3 c2d3 f6d5 f1d1 a8d8 c1d2 b7c8 d3e4 f8e8 f2f4 c8g4 d1e1 g4e6 e4f3 c6c5 g1f2 d5e7 d2c3 e7d5');
INSERT INTO evaluations VALUES(28648,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28648,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28648,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28648,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28648,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28648,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(28648,36,'cp 0','e6d5 f1b5 c7c6 b5d3 f8d6 e1g1 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 e4c2 b8d7 d1d3 d7f6 f3e5 d6e5 d4e5 d8d3 c2d3 f6d5 f1d1 a8d8 c1d2 b7c8 h2h3 d5e7 d2c3 e7d5');
INSERT INTO evaluations VALUES(28669,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28669,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28669,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28669,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28669,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28669,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28669,36,'cp 0','f1b5 c7c6 b5d3 f8d6 e1g1 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 e4c2 b8d7 d1d3 d7f6 f3e5 d6e5 d4e5 d8d3 c2d3 f6d7 f2f4 d7c5 d3f5 a8d8 c1e3 c5d3 a1b1 d3b4 f1d1 b4d5 e3d2 b7c8 f5e4 c8g4 d1e1 d5e7 d2c3 e7d5');
INSERT INTO evaluations VALUES(32925,30,'cp 322',NULL);
INSERT INTO evaluations VALUES(32925,31,'cp 301',NULL);
INSERT INTO evaluations VALUES(32925,32,'cp 335',NULL);
INSERT INTO evaluations VALUES(32925,33,'cp 301',NULL);
INSERT INTO evaluations VALUES(32925,34,'cp 335',NULL);
INSERT INTO evaluations VALUES(32925,35,'cp 312',NULL);
INSERT INTO evaluations VALUES(32925,36,'cp 331','f3b3 b7b6 b1c3 g8e7 d2d3 b8d7 c3d5 e7d5 c4d5 a8c8 b3a4 c7c6 d5c6 f6d8 c1e3 f8e7 d3d4 e5d4 e3d4 d8c7 c6b5 a7a6 a4a6 e8g8 b5d7 c7d7 a6d3 d6d5 e1c1 d7a4 d3b3 a4b3 a2b3 d5e4 d4b6 f7f5 d1d7 f8f7 c1b1 e7f6 d7f7 g8f7 c2c3 c8b8 b6a7 b8b7 a7e3 b7b3 b1c2 b3b8');
INSERT INTO evaluations VALUES(32922,30,'cp 340',NULL);
INSERT INTO evaluations VALUES(32922,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(32922,32,'cp 335',NULL);
INSERT INTO evaluations VALUES(32922,33,'cp 339',NULL);
INSERT INTO evaluations VALUES(32922,34,'cp 347',NULL);
INSERT INTO evaluations VALUES(32922,35,'cp 349',NULL);
INSERT INTO evaluations VALUES(32922,36,'cp 334','b7b6 b1c3 g8e7 d2d3 g7g5 c3b5 e8d8 c1e3 a7a6 b5c3 b8d7 e1c1 f8g7 d3d4 e5d4 e3d4 f6h6 d4e3 f7f5 f2f3 g7e5 h3h4 f5f4 e3f2 g5g4 f3g4 h6g7 c4e6 a8b8 h4h5 h7h6 h1e1 b6b5 c3e2 a6a5 e2d4 d7c5 b3f3 g7f6 e6d5 e7d5 e4d5 d8c8 c1b1 b5b4');
INSERT INTO evaluations VALUES(32728,30,'cp 321',NULL);
INSERT INTO evaluations VALUES(32728,31,'cp 318',NULL);
INSERT INTO evaluations VALUES(32728,32,'cp 326',NULL);
INSERT INTO evaluations VALUES(32728,33,'cp 346',NULL);
INSERT INTO evaluations VALUES(32728,34,'cp 334',NULL);
INSERT INTO evaluations VALUES(32728,35,'cp 338',NULL);
INSERT INTO evaluations VALUES(32728,36,'cp 339','b1c3 g8e7 d2d3 b8d7 c3d5 e7d5 c4d5 a8c8 b3a4 f6d8 a4c4 d8e7 d5c6 d6d5 c4d5 c8d8 c1d2 g7g6 d5b5 a7a5 e1g1 e7c5 a2a4 c5b5 a4b5 f8c5 d2g5 c5e7 g5e3 e8g8 e3h6 f8e8 g2g3 g8h8 h6e3 f7f5 g1g2 f5e4 d3e4');
INSERT INTO evaluations VALUES(32729,30,'cp 318',NULL);
INSERT INTO evaluations VALUES(32729,31,'cp 318',NULL);
INSERT INTO evaluations VALUES(32729,32,'cp 333',NULL);
INSERT INTO evaluations VALUES(32729,33,'cp 325',NULL);
INSERT INTO evaluations VALUES(32729,34,'cp 316',NULL);
INSERT INTO evaluations VALUES(32729,35,'cp 326',NULL);
INSERT INTO evaluations VALUES(32729,36,'cp 326','g8e7 d2d3 g7g5 c3b5 e8d8 c1e3 b8c6 c4f7 f8g7 f7d5 e7d5 e4d5 c6d4 b3c4 d4b5 c4b5 h7h6 a2a4 a7a5 b5c6 a8a7 c2c3 f6f7 c6c4 g7f6 e1d2 f7d7 d2c2 a7a8 h1e1 d8e7 e1e2 e7f7 f2f3 h8f8 a1e1 f7g8 e3d2 g8h8 e1a1 d7f7');
INSERT INTO evaluations VALUES(53174,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(53174,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(53174,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(53174,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(53174,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(53174,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(53174,36,'cp 25','e7e6 a2a3 f8c5 g1f3 c5b6 f1d3 b8c6 e1g1 g8e7 f1e1 e7g6 c1g5 d8c7 d3g6 f7g6 b1c3 e8g8 g5e3 b6e3 e1e3 a7a6 c3e2 h7h6 d1d2 c8d7 h2h4 f8f5 d2c3 a8f8 a1d1 g8h7 b2b4 f8c8 d1d2 h7g8 c3b2 c8f8 e3c3 g8h7');
INSERT INTO evaluations VALUES(54077,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(54077,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(54077,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54077,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54077,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(54077,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(54077,36,'cp 40','a2a3 f8c5 g1f3 g8e7 b2b4 c5b6 c1b2 a7a5 b1d2 a5b4 a3b4 a8a1 d1a1 e8g8 f1e2 b8a6 b2a3 c8d7 a1b2 e7g6 g2g3 f7f6 e1g1 f6e5 b4b5 a6c5 f3e5 d8f6 e5d7 f6b2 a3b2 c5d7 c2c4 g6e5 g1g2 f8a8 f1a1 a8a1 b2a1 g8f7 c4d5 e6d5 f2f4 e5g6 a1b2 g6e7 d2f3');
INSERT INTO evaluations VALUES(4266,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(4266,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(4266,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(4266,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(4266,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(4266,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(4266,36,'cp 0','c6d4 b5c4 a7a5 a2a4 d7d6 h2h3 c8e6 f3d4 c5d4 c4e6 f7e6 c3e2 d4b6 c1e3 c7c6 e2g3 b6e3 f2e3 d8b6 d1d2 f8f7 b2b3 a8f8 f1f3 f6d7 f3f7 f8f7 a1f1 b6b4 d2e2 b4c5 e2d2 f7f8 f1f8 g8f8 g3f1 h7h6 f1h2 b7b5 h2f3');
INSERT INTO evaluations VALUES(4524,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(4524,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4524,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4524,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(4524,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(4524,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(4524,36,'cp -4','b5c4 a7a5 h2h3 d7d6 a2a4 c8e6 f3d4 c5d4 c4e6 f7e6 c3e2 d4b6 c1e3 c7c6 e2g3 b6e3 f2e3 d8b6 d1e2 f8f7 b2b3 a8f8 f1f3 f6d7 f3f7 g8f7 a1f1 f7e7 f1f8 e7f8 g1f1 b6c5 e2d2 g7g6 f1e1 f8e7 e1d1 b7b5 a4b5 c6b5');
INSERT INTO evaluations VALUES(4798,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(4798,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4798,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(4798,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(4798,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(4798,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(4798,36,'cp 32','d2d4 e5d4 d1d4 f7f6 c3e2 g8e7 c1f4 c8g4 f4g3 d6g3 h2g3 d8d4 f3d4 c6c5 d4b3 b7b6 f2f3 g4e6 e1c1 a6a5 b3d2 e8c8 g3g4 e7c6 a2a4 e6f7 d2f1 d8d1 c1d1 h7h5 g4h5 h8h5 h1h5 f7h5 f1e3 c8d7 d1d2 c6e7 e2c3 h5f7 d2d3 e7g6 c3d5 d7d6 f3f4 d6c6');
INSERT INTO evaluations VALUES(4794,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(4794,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(4794,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(4794,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(4794,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(4794,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(4794,36,'cp 19','e5d4 d1d4 f7f6 c3e2 g8e7 c1f4 c8g4 f4g3 d6g3 h2g3 d8d4 f3d4 c6c5 d4b3 b7b6 f2f3 g4e6 e1c1 e7c6 b3d2 e8c8 g3g4 h7h6 b2b3 c6d4 e2g3 c8b7 c2c3 d4c6 g3h5 h8g8 c1c2 e6f7 d2f1 c5c4 f1e3 d8d1 h1d1 c4b3 a2b3');
INSERT INTO evaluations VALUES(4812,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(4812,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4812,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4812,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(4812,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(4812,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(4812,36,'cp 22','d1d4 f7f6 c3e2 g8e7 c1f4 c8g4 f4g3 d6g3 h2g3 d8d4 f3d4 c6c5 d4b3 b7b6 f2f3 g4e6 b3d2 e8c8 g3g4 e7c6 e1c1 c6d4 e2g3 d8d7 b2b3 a6a5 d2f1 a5a4 b3a4 e6a2 f1e3 a2e6 g3h5 h8d8 h5f4 h7h6 f4e6 d4e6 d1d7 c8d7');
INSERT INTO evaluations VALUES(4810,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(4810,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(4810,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(4810,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(4810,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(4810,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(4810,36,'cp 20','f7f6 c3e2 g8e7 c1f4 c8g4 f4g3 d6g3 h2g3 d8d4 f3d4 c6c5 d4b3 b7b6 f2f3 g4e6 e1c1 e7c6 b3d2 e8c8 g3g4 a6a5 a2a4 e6f7 d2f1 d8d1 c1d1 h7h5 g4h5 h8h5 h1h5 f7h5 f1e3 h5e8 d1d2 c8d7 e2c3 c6e7 c3d5 e7d5 e3d5 d7d8 d5c3');
INSERT INTO evaluations VALUES(10142,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10142,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(10142,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(10142,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10142,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10142,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10142,36,'cp 0','b1c3 c6d4 g1f3 c7c6 c4a2 e8g8 f3e5 d7d5 c1f4 f8e8 f4g3 d8c7 e5f3 c7d8');
INSERT INTO evaluations VALUES(10139,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10139,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10139,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10139,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10139,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10139,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(10139,36,'cp 0','c6d4 g1f3 c7c6 f3d4 c5d4 d1f3 d7d6 h2h3 c8e6 c3e2 d4b6 e1g1 d6d5 c4b3 d5e4 d3e4 e6b3 c2b3 f6d7 b3b4 d8f6 f3g3 f6g6 g3f3');
INSERT INTO evaluations VALUES(30307,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(30307,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(30307,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(30307,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(30307,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(30307,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(30307,36,'cp 62','b5c4 d7g4 f2f3 g4c8 b1c3 b8d7 g1e2 d7b6 c4b3 a7a5 a2a3 b6d5 c3d5 f6d5 c2c4 d5f6 d2d4 e7e6 e2c3 f8e7 c1f4 e8g8 d1d2 c7c6 e1g1 f6h5 f4e3 h5f6 b3c2 b7b6 f1e1 c8a6 d2d3 b6b5 b2b3 g7g6 a1d1 b5c4 b3c4 d8c7');
INSERT INTO evaluations VALUES(30311,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(30311,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(30311,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(30311,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(30311,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(30311,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(30311,36,'cp 50','d7g4 f2f3 g4c8 b1c3 b8d7 g1e2 d7b6 c4b3 a7a5 a2a3 b6d5 c3d5 f6d5 c2c4 d5f6 d2d4 g7g6 e1g1 f8g7 e2c3 e8g8 f1e1 f6e8 d4d5 e7e6 c4c5 e6d5 b3d5 d8f6 d5a2 c7c6 g1h1 e8c7 c3e4 f6d4 c1f4 d4d1 e1d1 c7d5 a2d5 c6d5 d1d5');
INSERT INTO evaluations VALUES(29499,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(29499,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(29499,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(29499,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(29499,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(29499,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(29499,36,'cp 53','f2f3 g4c8 b1c3 b8d7 d2d4 d7b6 c4b3 a7a5 a2a3 b6d5 c3d5 f6d5 c2c4 d5f6 g1e2 g7g6 e2c3 f8g7 e1g1 e8g8 c1e3 f6e8 d4d5 e7e6 d1d2 e6d5 a1d1 d5d4 c4c5 c7c6 f1e1 d4d3 d2d3 d8d3 d1d3 c8f5 d3d2 e8c7 e1d1 f5e6 b3e6 c7e6 d2d7');
INSERT INTO evaluations VALUES(29498,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(29498,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(29498,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(29498,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(29498,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(29498,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(29498,36,'cp 63','g4c8 b1c3 b8d7 g1e2 d7b6 c4b3 a7a5 a2a3 b6d5 c3d5 f6d5 c2c4 d5f6 d2d4 g7g6 c1g5 f8g7 d1d2 c7c6 e1c1 b7b5 c4c5 h7h6 g5h4 f6d5 h1e1 c8e6 e2c3 g6g5 c3d5 e6d5 b3d5 d8d5 h4g3 a5a4 d2b4 e8g8 e1e7');
INSERT INTO evaluations VALUES(47823,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(47823,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(47823,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(47823,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47823,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(47823,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(47823,36,'cp 27','d2d3 a7a6 f1e2 g7g6 g1f3 f8g7 d3d4 e5d4 f3d4 e8g8 c1g5 f8e8 f2f3 b8c6 d4c6 b7c6 d1d2 d8e7 e1g1 a8b8 b2b3 e7e5 a1c1 f6d7 f1d1 e5a5 g5h6 g7h6 d2h6 d7c5 h6d2 a5b4 e2f1 a6a5 h2h3 a5a4 c3a4 c5a4 b3a4 b4d2 d1d2 c6c5 f1e2 b8b4 a4a5');
INSERT INTO evaluations VALUES(47824,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(47824,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(47824,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(47824,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(47824,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47824,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47824,36,'cp 28','c7c6 g1f3 c6c5 g2g3 b8c6 a2a3 f8e7 f1g2 e8g8 a1b1 a7a6 e1g1 b7b5 c4b5 a6b5 b2b4 c5b4 a3b4 c8e6 d3d4 e6c4 f1e1 d8d7 d4d5 c6b8 f3h4 g7g6 c1h6 f8c8 h4f3 f6e8 d1d2 a8a3 e1c1 b8a6 h2h3 c8a8');
INSERT INTO evaluations VALUES(49759,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(49759,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(49759,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49759,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(49759,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(49759,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(49759,36,'cp 21','g1f3 d7d5 f1d3 c7c5 c2c3 c5c4 d3c2 c8g4 e1g1 f8e7 c1g5 e8g8 b1d2 b8d7 f1e1 e7d6 h2h3 g4h5 g2g4 h5g6 f3e5 g6c2 d1c2 d8c7 e5d7 f6d7 c2f5 c7c6 d2f1 h7h6 g5e7 f8e8 e7h4 g7g6 f5f3 b7b5 e1e8 a8e8 f1e3 e8e4 a1e1');
INSERT INTO evaluations VALUES(49757,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(49757,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(49757,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(49757,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(49757,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(49757,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(49757,36,'cp 3','d7d5 f1d3 c7c5 c2c3 c5c4 d3c2 c8g4 e1g1 f8e7 b2b3 c4b3 a2b3 e8g8 f1e1 f8e8 c1f4 b8c6 h2h3 g4h5 b3b4 e7d6 f4d6 d8d6 b1d2 d6c7 c2d3 e8e1 d1e1 a7a5 b4a5 h5g6 d3g6 h7g6 e1c1 a8a5 c1b2 f6h5 g2g3 a5a1 b2a1');
INSERT INTO evaluations VALUES(53947,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53947,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(53947,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(53947,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(53947,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(53947,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53947,36,'cp 15','g8f6 f1d3 f8e7 c2c4 b8c6 b1c3 a7a5 c1e3 e8g8 d3e2 c8e6 a1b1 f6d7 e1g1 d7c5 b3c5 d6c5 b2b3 d8d1 f1d1 c6d4 g1f1 e6d7 c3d5 e7d8 h2h3 g7g6 f2f4 f7f5 e4f5 d7f5 b1b2 e5f4 e3d4 c5d4 d1d4 g8h8');
INSERT INTO evaluations VALUES(45951,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(45951,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(45951,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45951,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45951,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(45951,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(45951,36,'cp 24','f1d3 f8e7 c2c4 e8g8 b1c3 b8c6 b3d2 a7a6 e1g1 c8g4 d3e2 g4e6 b2b3 c6d4 c1b2 a8b8 e2d3 f6d7 c3d5 d7c5 b2d4 e5d4 d2f3 e6d5 e4d5 e7f6 b3b4 c5d3 d1d3 b7b5 a1c1 d8b6 f1d1 b8c8 c4b5 c8c1 d1c1 a6b5');
INSERT INTO evaluations VALUES(39450,30,'cp 264',NULL);
INSERT INTO evaluations VALUES(39450,31,'cp 280',NULL);
INSERT INTO evaluations VALUES(39450,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(39450,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(39450,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(39450,35,'cp 260',NULL);
INSERT INTO evaluations VALUES(39450,36,'cp 262','d4e5 f6e5 f3e5 b8c6 e5d3 g8f6 b1c3 e8g8 g2g3 a7a5 a2a3 d6e7 f1g2 d7d6 e1g1 c6e5 h2h3 c8e6 f2f4 e5d3 c2d3 a5a4 g1h2 e6b3 d1e2 c7c6 c1e3 f6d7 e2f2 d8a5 f2d2 a8e8 f1f2 e7f6 a1c1 b3f7 d2d1 f7b3 d1e2 f6e7 d3d4 d7b6');
INSERT INTO evaluations VALUES(39446,30,'cp 273',NULL);
INSERT INTO evaluations VALUES(39446,31,'cp 259',NULL);
INSERT INTO evaluations VALUES(39446,32,'cp 257',NULL);
INSERT INTO evaluations VALUES(39446,33,'cp 273',NULL);
INSERT INTO evaluations VALUES(39446,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(39446,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(39446,36,'cp 270','f6e5 f3e5 b8c6 e5d3 g8f6 b1c3 e8g8 g2g3 a7a5 a2a3 d6e7 f1g2 d7d6 h2h3 c8e6 e1g1 c6e5 g1h2 a5a4 f2f4 e5d3 c2d3 d8d7 d3d4 e6b3 d1e2 c7c6 c1e3 e7d8 e4e5 f6d5 c3e4 d5e3 e2e3 d8c7 a1e1 b3d5 e4c3 d5g2 h2g2 d6e5');
INSERT INTO evaluations VALUES(39452,30,'cp 285',NULL);
INSERT INTO evaluations VALUES(39452,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(39452,32,'cp 266',NULL);
INSERT INTO evaluations VALUES(39452,33,'cp 270',NULL);
INSERT INTO evaluations VALUES(39452,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(39452,35,'cp 282',NULL);
INSERT INTO evaluations VALUES(39452,36,'cp 270','f3e5 b8c6 e5d3 g8f6 b1c3 e8g8 g2g3 a7a5 a2a3 d6e7 f1g2 d7d6 e1g1 a5a4 h2h3 c8e6 c3d5 d8d7 d5e7 c6e7 d3f4 e6c4 f1e1 a8e8 b2b3 c4f7 c1b2 e7g6 f4g6 h7g6 b3b4 f7e6 g1h2 b7b5 b2f6 g7f6 d1d2 e6c4 a1d1 g6g5 f2f3 d7f7 f3f4 g5f4 g3f4');
INSERT INTO evaluations VALUES(39499,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(39499,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(39499,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(39499,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(39499,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(39499,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(39499,36,'cp 61','d7d5 e4e5 c7c5 c2c3 c8d7 f1d3 b8c6 d1e2 c5d4 f3d4 e7c5 d4c6 b7c6 b1d2 g8e7 e1g1 e7g6 b2b4 c5e7 d2b3 a7a5 b3a5 c6c5 a2a3 e8g8 f2f4 f7f6 c1e3 c5b4 a3b4 f6e5 f4e5 g6e5 d3h7 g8h7 e2h5 h7g8 h5e5 f8f1 a1f1 e7f6');
INSERT INTO evaluations VALUES(38294,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(38294,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(38294,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(38294,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(38294,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(38294,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(38294,36,'cp 77','e4e5 c7c5 c2c3 c8d7 f1d3 b8c6 d1e2 c5d4 f3d4 c6d4 c3d4 d8b6 c1e3 e7d8 b1c3 g8e7 e1g1 e8g8 a1c1 f7f6 f2f4 a8c8 h2h4 f6f5 h4h5 f8f7 f1d1 a7a6 g1h1 b6a7 a2a3 d8a5 b2b4 a5b6 e2f2 h7h6 d3e2 a7a8 a3a4 d7e8 a4a5 b6d8');
INSERT INTO evaluations VALUES(38298,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(38298,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(38298,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(38298,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(38298,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(38298,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(38298,36,'cp 63','c7c5 c2c3 c8d7 f1d3 b8c6 d1e2 a7a6 e1g1 c5d4 f3d4 c6d4 c3d4 d8b6 c1e3 d7b5 f1d1 b5d3 d1d3 e7d8 e2g4 g7g6 d3b3 b6c6 b1d2 c6d7 b3c3 h7h5 g4d1 g8e7 d2b3 b7b6 a1c1 e7f5 d1d2 a6a5 c1c2 d7a4 a2a3 e8g8 c2c1 a4d7 e3g5 d8g5 d2g5');
INSERT INTO evaluations VALUES(37911,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(37911,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(37911,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(37911,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(37911,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(37911,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(37911,36,'cp 85','c2c3 c8d7 f1d3 b8c6 d1e2 c5d4 f3d4 c6d4 c3d4 d8b6 c1e3 e7d8 b1c3 g8e7 a1c1 e8g8 e1g1 a8c8 h2h4 f7f6 f2f4 f6f5 h4h5 a7a6 g1h2 b6b4 f1g1 f8f7 a2a3 b4b3 d3b1 g8h8 e3f2 f7f8 h5h6 g7h6 f2h4 e7g8 h4d8 f8d8 e2d2');
INSERT INTO evaluations VALUES(6569,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(6569,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(6569,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6569,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(6569,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6569,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(6569,36,'cp 21','c5b6 d1e2 g8f6 b1c3 e8g8 c1e3 d7d5 e1c1 d5d4 e3g5 a7a5 c3d5 a5a4 g5f6 g7f6 b3d2 a4a3 b2b3 g8h8 e2f3 b6a5 f3f6 d8f6 d5f6 h8g7 f6d5 c8e6 d2c4 e6d5 e4d5 c6b4 d1d4 b4a2 c1b1 a2c3 b1a1 a5b4 c4e3 b4d6 e3f5 g7h8 f5d6 c7d6 d4c4 a8c8 f1d3 f8e8 h2h3 c3d5 c4c8 e8c8 g2g4');
INSERT INTO evaluations VALUES(6346,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(6346,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(6346,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(6346,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(6346,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(6346,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(6346,36,'cp 28','d1e2 g8f6 b1c3 e8g8 c1g5 h7h6 g5h4 a7a5 a2a4 c6d4 e2d3 d4b3 c2b3 c7c6 e1c1 d7d5 e4d5 c6d5 d3c2 d5d4 f1c4 c8e6 c1b1 a8c8 c4e6 f7e6 c2e2 d8d7 h4f6 f8f6 d1d2 f6g6 f2f3 e6e5 b1a2 d7f7 c3e4 b6d8 d2c2 c8c2 e2c2 g6c6 c2d3 d8e7 e4f2');
INSERT INTO evaluations VALUES(12944,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(12944,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(12944,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(12944,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12944,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12944,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12944,36,'cp 0','c8d7 b5a4 d8a5 d4c5 a5c5 e1g1 e7g6 f1e1 f8e7 a4c2 c5a5 b1d2 a5c7 c2g6 h7g6 d2f1 b7b5 a2a3 a7a5 c1f4 c7d8 d1d2 b5b4 a3b4 a5b4 h2h4 b4c3 b2c3 a8a1 e1a1 e8g8 f4g5 e7g5 d2g5 d8g5 h4g5 f8b8 f1e3 d7e8 f3d4 c6d4 c3d4');
INSERT INTO evaluations VALUES(24450,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(24450,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24450,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24450,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24450,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24450,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(24450,36,'cp 0','b5a4 d8a5 d4c5 c6e5 a4d7 e5d7 b2b4 a5c7 b1a3 a7a6 c3c4 e7c6 a1b1 d7f6 e1g1 f8e7 c1b2 a8d8 d1b3 e8g8 a3c2 c7f4 c4d5 f6d5 a2a3 e7f6 c2e3 f6b2 b1b2 d5e3 f2e3 f4c7 b2d2 c6e5 d2d8 e5f3 g2f3 c7d8 f1d1 d8g5 g1f2 g5h4 f2g2 h4g5');
INSERT INTO evaluations VALUES(22597,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(22597,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22597,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22597,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22597,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(22597,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22597,36,'cp 0','d7d5 e4d5 f8b4 b1c3 e6d5 c4d5 d8d5 g1f3 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 a1c1 e7f5 h1d1 h8e8 c3a4 b4e7 g2g4 f5e3 f2e3 d8d6 a4c5 d6h6 d1f1 h6h2 e2d3 e7f6 c5a6 c8b7 a6c5');
INSERT INTO evaluations VALUES(21390,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21390,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21390,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(21390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21390,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21390,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21390,36,'cp 0','c4d5 e6d5 e4d5 d8d5 g1f3 f8b4 b1c3 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8d7 e2d3 a7a5 c3a4 e7d5 a1c1 h8e8 e3d2 a8b8 b2b3 e8e7 a4c5 d7d6 c5e4 d6d7 e4c5');
INSERT INTO evaluations VALUES(21367,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21367,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21367,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21367,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21367,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21367,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21367,36,'cp 0','e6d5 e4d5 d8d5 g1f3 c8g4 f1e2 g4f3 e2f3 f8b4 b1c3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8c8 h1d1 h8e8 a1c1 e7f5 c3a4 b4e7 c1c6 e7f6 e2f3 f5h4 f3g3 f6d4 g3h4 d4e3 d1d8 c8d8 f2e3 e8e4 h4g3 e4a4 a2a3 d8d7 c6c3 a4a5 h2h4 h7h6 c3d3 d7c8 d3c3');
INSERT INTO evaluations VALUES(54338,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(54338,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(54338,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(54338,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(54338,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(54338,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(54338,36,'cp 26','b1c3 g8f6 g1f3 f8b4 c4d5 f6d5 c1d2 c8d7 c3d5 b4d2 d1d2 e6d5 f3e5 e8g8 f1d3 f8e8 e1g1 b8c6 f1e1 d8c7 d2c2 g7g6 d3b5 e8e6 c2c5 c7d6 c5d6 e6d6 e5d7 d6d7 a1c1 a8b8 g2g3 d7d6 b5c6 d6c6 c1c6 b7c6 b2b3');
INSERT INTO evaluations VALUES(54336,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(54336,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(54336,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54336,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(54336,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(54336,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(54336,36,'cp 26','g8f6 g1f3 f8b4 c4d5 f6d5 c1d2 c8d7 c3d5 b4d2 d1d2 e6d5 f1e2 b8c6 e1g1 e8g8 f1c1 d8e7 h2h3 a8c8 e2b5 a7a6 b5a4 h7h6 c1c3 e7d6 a1c1 f8d8 a2a3 c6e7 a4d7 d6d7 f3e5 d7d6 a3a4 e7c6 e5c6 c8c6 c3c6 b7c6');
INSERT INTO evaluations VALUES(38274,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(38274,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(38274,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(38274,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(38274,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(38274,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(38274,36,'cp 68','g8f6 f1d3 b8c6 a2a3 e6e5 d4d5 c6b8 c3a4 c8d7 c2c4 d7a4 d1a4 b8d7 h2h3 a7a5 c1e3 e8g8 a4d1 a5a4 f3d2 c7c6 e1g1 d8a5 b2b4 a4b3 d2b3 a5a4 d1b1 f8c8 b3d2 e7d8 b1b7 c6d5 c4d5 d7c5 e3c5 c8c5 b7b2');
INSERT INTO evaluations VALUES(40889,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(40889,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(40889,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(40889,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(40889,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(40889,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(40889,36,'cp 61','f1d3 b8c6 a2a3 e6e5 d4d5 c6b8 c3e2 c8g4 e2g3 h7h5 h2h3 g4f3 d1f3 g7g6 a3a4 b8d7 a4a5 h5h4 g3e2 a7a6 c1e3 f6h7 e1g1 e7g5 b2b4 d8f6 f3f6 d7f6 e3g5 h7g5 f2f3 e8g8 c2c4 f6d7 f1b1 a8e8 e2c3 f7f5 d3c2 g8g7 c2a4');
INSERT INTO evaluations VALUES(40890,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(40890,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(40890,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(40890,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(40890,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(40890,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(40890,36,'cp 59','b8c6 a2a3 e6e5 d4d5 c6b8 c3a4 c8d7 c2c4 d7a4 d1a4 b8d7 c1e3 f6g4 e3d2 a7a5 a4d1 d7c5 d3c2 g4f6 d1e2 c7c6 e1g1 a5a4 d2b4 d8b6 g2g3 c6d5 c4d5 e8g8 f3d2 c5a6 b4c3 a6c5 a1c1 b6a6 e2a6 a8a6 c3b4 b7b5 f1d1 g7g6 c2b1 f8b8 g1g2 h7h5 h2h3 c5b3 d2b3');
INSERT INTO evaluations VALUES(47948,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(47948,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(47948,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(47948,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(47948,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(47948,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(47948,36,'cp 69','d6d5 e4d5 f6d5 c3d5 d8d5 f1d3 d5d6 d1e2 c8e6 e2e4 b8c6 c1f4 d6d7 d3c4 e8c8 c4e6 f7e6 c2c3 d7d5 e4c2 c8b8 e1g1 f8d6 f4d6 c7d6 f1f2 d8f8 h2h3 e6e5 d4e5 c6e5 f3e5 d5e5 f2e2 e5c5 g1h2 f8e8 e2e8 h8e8 c2h7 c5e5 h2g1 e5e3 g1h1 e3g3 a1f1');
INSERT INTO evaluations VALUES(16634,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(16634,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(16634,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(16634,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(16634,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(16634,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(16634,36,'cp -63','d7d5 e4d5 d8e7 e1f1 c6e5 d1d4 h7h6 g5e4 e5c4 e4f6 e7f6 d4c4 f8e7 b1c3 e8g8 c1e3 c8f5 f2f3 e7d6 f1f2 a8c8 e3d4 f6h4 g2g3 h4h3 a1e1 f5c2 d4c5 b7b5 c4d4 a7a6 e1c1 c2g6 c3e4 g6e4 d4e4 c8e8 e4g4 h3g4 f3g4 f7f5 c5d6 c7d6');
INSERT INTO evaluations VALUES(15021,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(15021,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(15021,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(15021,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(15021,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(15021,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(15021,36,'cp -53','e4d5 d8e7 e1f1 c6e5 d1d4 e5c4 d4c4 h7h6 g5f3 e7c5 c4e2 f8e7 c1e3 c5d5 b1c3 d5a5 e3f4 a5b4 e2b5 c7c6 b5b4 e7b4 a1e1 e8f8 h2h4 c8e6 f4e5 f6d7 a2a3 b4a5 h4h5 d7e5 f3e5 f8e7 h1h4 h8d8 h4a4 a5c3 b2c3 d8d5 a4b4 e7f6 b4f4 e6f5 g2g4 d5e5 e1e5 f6e5 f4f5 e5e6 f1g2 a8d8 c3c4');
INSERT INTO evaluations VALUES(14973,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(14973,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(14973,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(14973,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(14973,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(14973,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(14973,36,'cp -63','d8e7 e1f1 c6e5 d1d4 h7h6 g5e4 e5c4 e4f6 e7f6 d4c4 f8e7 b1c3 e8g8 c1e3 c8f5 f2f3 a8d8 f1f2 f8e8 h1e1 a7a6 c4d4 f6d6 g2g3 f5c2 f2g2 d6d7 e3f4 e7f8 e1e8 d8e8 a1c1 c2g6 d4c4 c7c6 d5c6 b7c6 c1d1 d7b7 d1d2 f8b4 h2h4');
INSERT INTO evaluations VALUES(51971,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51971,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51971,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51971,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51971,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51971,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51971,36,'cp 23','c5d4 d1d4 b8c6 d4e3 g8f6 g1f3 f8b4 a2a3 f6g4 e3f4 b4c3 b2c3 d8f6 f4f6 g4f6 f1d3 d7d6 a3a4 e6e5 e1g1 e8e7 a4a5 c8g4 c1e3 g4f3 g2f3 h8c8 f1b1 c8c7 f3f4 f6d7 f4f5 c6d8 a1a3 d7c5 b1b5 a8c8 c3c4 d8c6 e3c5 d6c5 b5c5 f7f6 g1g2');
INSERT INTO evaluations VALUES(52077,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(52077,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(52077,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(52077,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(52077,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52077,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(52077,36,'cp 25','d1d4 g8f6 g1f3 b8c6 d4e3 f8e7 e4e5 f6d5 c3d5 e6d5 c1d2 e8g8 e1c1 d7d6 e5d6 d8d6 e3f4 d6f4 d2f4 f8d8 f3d4 c6d4 d1d4 g7g6 d4d1 e7f6 f1d3 c8d7 h1e1 g8g7 h2h3 d8e8 g2g4 e8e1 d1e1 h7h6 f4e5 f6e5 e1e5 d7e6 e5e3 g7f6 e3f3 f6e7');
INSERT INTO evaluations VALUES(52067,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(52067,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(52067,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(52067,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52067,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(52067,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(52067,36,'cp 0','b8c6 d4e3 a7a6 e3g3 d7d5 c1g5 f8e7 g5e7 g8e7 e4d5 e6d5 e1c1 e8g8 g1f3 c8e6 f1d3 d8a5 c1b1 h7h6 c3e2 d5d4 a2a3 a8d8 e2f4 e6c8 f3d2 e7f5 g3f3 f5h4 f3g3');
INSERT INTO evaluations VALUES(18919,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18919,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18919,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18919,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18919,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18919,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(18919,36,'cp 0','d4e3 a7a6 e3g3 d7d5 c1g5 f8e7 g5e7 g8e7 e4d5 e6d5 e1c1 e8g8 g1f3 c8e6 f1d3 d8a5 c1b1 h7h6 c3e2 d5d4 a2a3 a8d8 e2f4 e6c8 h2h4 a5c5 h1e1 f8e8 e1e4 e7f5 g3h2 f5d6 e4e8 d8e8 f3d2 c5e5 d2f3');
INSERT INTO evaluations VALUES(18926,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18926,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18926,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18926,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18926,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18926,36,'cp 0','a7a6 f1d3 g8f6 g1f3 d7d6 c1d2 f8e7 h2h3 e8g8 e1g1 c8d7 f1d1 c6b4 a2a3 b4d3 c2d3 a8c8 a1c1 f8e8 f3d4 e7f8 d4e2 d7c6 e2d4');
INSERT INTO evaluations VALUES(56699,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(56699,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(56699,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(56699,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(56699,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(56699,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(56699,36,'cp 173','d2d4 c8e6 f1b5 b8d7 d1e2 g7g6 b5a4 f8h6 e1g1 h6c1 a1c1 g8h6 e2e3 h6f7 h2h3 e6c4 f1d1 e8g8 b2b3 c4a6 a4d7 d8d7 c3d5 d7d8 d5b4 a6b5 d4e5 f6e5 c2c4 b5d7 c4c5 d7e6 b4d5 e6d5 d1d5 a8c8 b3b4 d8f6 a2a4 a7a6 d5d1 c8d8 e3b3 c7c6 c5d6 d8d6 d1d6 f6d6 c1d1 d6e7');
INSERT INTO evaluations VALUES(14329,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(14329,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(14329,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(14329,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(14329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14329,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14329,36,'cp 0','d2d4 c8g4 h2h3 g4f3 d1f3 e7e6 b1c3 h7h5 c1d2 h5h4 f1b5 a7a6 b5c6 b7c6 e1g1 f8b4 d2e1 b4c3 b2c3 f6e4 c3c4 f7f5 f3e2 d8e7 a1b1 a6a5 c4d5 c6d5 c2c4 e8g8 c4d5 e6d5 e2b5 e7d6 b5d3 d6e7');
INSERT INTO evaluations VALUES(27256,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(27256,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(27256,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(27256,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(27256,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(27256,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(27256,36,'cp 144','f3e5 d6e5 d1d5 g8f6 d5b7 f8c5 b5d7 f6d7 e1g1 e8g8 b7b5 d8e7 b5e2 f8d8 b1c3 h7h6 b2b3 d7f6 c3a4 c5d4 c1b2 d4b2 a4b2 d8d4 f2f3 f6h5 g2g3 e7g5 e2g2 a8d8 b2c4 a7a5 a2a4 d4c4 b3c4 d8d2 f1f2 g5e3 g1f1 d2d8 f2e2 e3c3 a1e1');
INSERT INTO evaluations VALUES(27254,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(27254,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(27254,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(27254,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(27254,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(27254,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(27254,36,'cp 136','d6e5 d1d5 g8f6 d5b7 f8c5 b5d7 f6d7 e1g1 e8g8 b7b5 d8e7 b5e2 h7h6 b1d2 a7a5 a2a4 c5d4 d2c4 f8b8 a1a2 e7e6 b2b3 e6g6 f1e1 d7b6 c4b2 g6c6 b2d3 b6d7 c1b2 b8e8 a2a1 d7c5 b2d4 e5d4 d3c5 c6c5 e2d3 a8d8 h2h3');
INSERT INTO evaluations VALUES(51835,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(51835,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(51835,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(51835,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(51835,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(51835,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(51835,36,'cp -40','g8f6 b1c3 b7b5 c4b3 a7a5 a2a3 f8c5 g1e2 e8g8 d2d3 d7d6 e1g1 c8g4 f3g3 b8d7 g1h1 g4e6 b3e6 f7e6 g3h3 d8e7 f2f4 e5f4 c1f4 a8e8 a1e1 e7f7 f4d2 a5a4 e4e5 d6e5 c3e4 f7g6 e4c5 d7c5 h3g3 g6g3 e2g3 f6d7 h2h3 f8f7 e1d1 h7h6 h1g1 f7f1 d1f1');
INSERT INTO evaluations VALUES(13527,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(13527,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13527,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13527,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13527,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(13527,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(13527,36,'cp 14','e4d5 e6d5 c4b5 f8e7 d2d4 e8g8 d4c5 e7c5 c1g5 c8e6 d1d3 a8c8 a1e1 h7h6 g5h4 c6a5 f3d4 c5d4 d3d4 a7a6 b5d3 a5c6 d4a4 g7g5 h4g3 f6h5 c3e2 h5g3 e2g3 d8b6 g3f5 e6f5 d3f5 c8e8 f5d7 e8e1 f1e1 f8d8 d7c6 b6c6 a4c6');
INSERT INTO evaluations VALUES(13507,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(13507,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(13507,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(13507,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13507,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13507,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13507,36,'cp 0','e6d5 c4b5 f8d6 d2d4 e8g8 d4c5 d6c5 c1g5 c8e6 d1d3 a8c8 a1e1 h7h6 g5h4 c6b4 d3d2 b4c6');
INSERT INTO evaluations VALUES(13733,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13733,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13733,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13733,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13733,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13733,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13733,36,'cp 10','c4b5 f8e7 d2d4 c8e6 c1g5 e8g8 d4c5 e7c5 g5h4 h7h6 f1e1 f8e8 h4f6 d8f6 c3d5 f6b2 d5c7 a8d8 d1c1 b2c1 a1c1 e8e7 c7e6 e7e6 e1e6 f7e6 b5c6 b7c6 g1f1 d8b8 f3e5 b8b6 c2c3 b6a6 c1c2 c5d6 e5c4 d6c7 c2d2 a6a4 c4e3 c7h2 g2g3 g8f7 d2d7 f7f6');
INSERT INTO evaluations VALUES(13712,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(13712,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(13712,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(13712,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13712,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(13712,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13712,36,'cp 7','f8d6 d2d4 c8e6 c1g5 e8g8 d4c5 d6c5 g5h4 c5e7 f1e1 d8b6 b5c6 b7c6 f3d4 a8e8 c3a4 b6c7 c2c3 e6d7 b2b4 h7h6 h2h3 c7c8 h4f6 e7f6 d1f3 e8e1 a1e1 f8e8 e1d1 f6e7 a4c5 e7c5 b4c5');
INSERT INTO evaluations VALUES(14811,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(14811,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14811,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14811,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(14811,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(14811,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(14811,36,'cp 46','g1f3 f8d6 f4g3 e8g8 c2c4 f8e8 f3e5 c6e7 a1c1 c7c5 d4c5 d6c5 c4d5 d8d5 f1e2 c5b4 a2a3 d5d2 d1d2 b4d2 e1d2 f6e4 d2e1 f7f6 e5d3 e4g3 h2g3 e8d8 f2f3 b7b6 e1f2 c8a6 c1c7 e7d5 c7c6 a8c8 h1c1 c8c6 c1c6 a6d3');
INSERT INTO evaluations VALUES(31173,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31173,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31173,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31173,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31173,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(31173,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(31173,36,'cp 0','e7e6 d2d4 c6c5 c2c4 d5c4 b1c3 b8c6 d4d5 e6d5 d1d5 c6d4 d5b7 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(31552,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31552,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31552,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31552,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31552,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31552,36,'cp -16','d2d4 c6c5 c2c4 d5c4 b1c3 b8c6 d4d5 e6d5 d1d5 c6d4 f1c4 d8d5 c3d5 d4f3 g2f3 e8c8 d5e3 f5e6 c4e6 f7e6 f4f5 e6f5 e3f5 g7g6 f5g3 d8d5 f3f4 g8e7 c1d2 e7c6 a1d1 f8h6 h1f1 h8f8 d2e3 c6d4 f1f2 f8d8 e3d4 d5d4 d1d4 d8d4 g3e2 d4b4 b2b3 b7b6 h2h3 c8d7 f2f3 d7e6');
INSERT INTO evaluations VALUES(42976,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(42976,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(42976,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42976,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(42976,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(42976,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(42976,36,'cp 65','e4e5 f6d5 g5e7 d8e7 c2c4 d5b6 g1f3 d7d6 e5d6 c7d6 f1e2 b8c6 e1g1 e8g8 b1c3 e6e5 f1e1 c8f5 d1d2 f5g6 e2f1 g6h5 d4d5 h5f3 g2f3 c6b8 f3f4 b8d7 b2b3 a8e8 f1g2 e7h4 f4e5 e8e5 c3b5 e5h5 h2h3 h4f6 f2f4 d7c5 b5a7 f8a8');
INSERT INTO evaluations VALUES(42964,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(42964,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(42964,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(42964,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(42964,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(42964,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(42964,36,'cp 50','f6d5 g5e7 d8e7 c2c4 d5b6 g1f3 b8c6 b1c3 d7d6 e5d6 c7d6 f1e2 e8g8 e1g1 e6e5 f1e1 e5e4 e2f1 f7f5 d4d5 c6e5 f3e5 e7e5 d1d2 e5f6 b2b4 b6d7 c3b5 a7a6 b5c7 a8b8 c7e6 f8e8 e6d4 d7e5 d2c3 c8d7 c4c5 d6c5 b4c5 b8d8');
INSERT INTO evaluations VALUES(42960,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(42960,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(42960,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42960,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(42960,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42960,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(42960,36,'cp 47','g5e7 d5e7 b1c3 b8c6 d1g4 e8g8 e1c1 d7d6 g1f3 e7d5 h2h4 f7f5 g4g3 d5c3 b2c3 h7h6 h4h5 c6e7 f3h4 d6d5 c3c4 b7b6 c4d5 d8d5 g3b3 d5b3 a2b3 a7a5 h1h3 a5a4 h4g6 e7g6 h5g6 c8b7 h3c3 a4b3');
INSERT INTO evaluations VALUES(42926,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(42926,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(42926,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(42926,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(42926,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(42926,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(42926,36,'cp 45','d5e7 f2f4 d7d6 b1c3 e7f5 d1d3 d6e5 d4e5 d8d3 f1d3 c8d7 g1f3 b8c6 a1d1 e8c8 e1f2 c6b4 d3e4 d7c6 e4f5 e6f5 f3d4 c6d7 a2a3 b4c6 b2b4 c6d4 d1d4 d7e6 d4d3 e6c4 d3d8 h8d8 f2e3 d8e8 h1d1 b7b6 d1d4 c4e6 e3d2 c8b7 h2h4 g7g6 d2e3');
INSERT INTO evaluations VALUES(43091,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43091,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(43091,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(43091,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43091,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43091,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(43091,36,'cp 62','b1c3 b8c6 f2f4 e7f5 g1f3 d7d6 d1d3 d6e5 d4e5 d8d3 f1d3 c8d7 a1d1 e8c8 e1f2 c6b4 d3e4 d7c6 e4f5 e6f5 f3d4 c6d7 a2a3 b4c6 b2b4 c6d4 d1d4 d7e6 d4d3 e6c4 d3d8 h8d8 h1d1 d8d1 c3d1 c4d5 g2g3 b7b6 f2e3 d5e6 e3d4 a7a5 b4a5 c7c5 d4d3 b6a5 d1e3 c8c7 e3c4');
INSERT INTO evaluations VALUES(10913,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(10913,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(10913,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(10913,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(10913,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(10913,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(10913,36,'cp 3','f6e4 c3e4 d7d5 c4b5 d5e4 f3e5 d8d6 d1e2 e8g8 b5c6 b7c6 e2e4 c8a6 e4c6 d6d4 c6a6 d4e5 c1e3 e5b2 e1g1 f8b8 g2g3 b2c2 a1c1 c2g6 a6g6 h7g6 c1c7 e7d8 c7d7 b8b2 a2a3 b2b3 f1d1 d8b6 d1d3 b3d3 d7d3 a8e8 g1g2 e8e4 e3b6 a7b6 g2f3 e4a4 h2h4 g8f8 d3b3 a4a5 b3b6 a5a3 f3g2 f8e8 b6b4 a3a5');
INSERT INTO evaluations VALUES(43992,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(43992,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(43992,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(43992,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(43992,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(43992,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(43992,36,'cp -33','b8c6 d4c5 f8c5 e3c5 d7c5 f1d3 e8g8 e1g1 c8d7 f1e1 d7e8 d1d2 d8b6 a2a4 c5d3 c2d3 c6a5 a1a3 a5b3 d2d1 b3d4 f3d4 b6d4 a4a5 b7b6 a5b6 d4b6 d3d4 a8b8 d1d2 b6b2 d2b2 b8b2 a3a7 b2b4 c3e2 h7h6 f2f3 e8b5 a7a3');
INSERT INTO evaluations VALUES(12119,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(12119,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(12119,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(12119,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(12119,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(12119,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(12119,36,'cp -19','d4c5 f8c5 e3c5 d7c5 f1d3 e8g8 e1g1 c8d7 f1e1 d7e8 d3b5 c6e7 b5f1 d8b6 a1b1 h7h6 d1d4 e7f5 d4f4 e8c6 g2g4 f5e7 f4e3 f7f5 e5f6 f8f6 b2b4 c5a4 c3a4 c6a4 f1d3 a8c8 e3b6 a7b6 f3d4 e7c6 c2c3 c6d4');
INSERT INTO evaluations VALUES(52771,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52771,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(52771,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52771,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52771,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(52771,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52771,36,'cp 7','c1f4 c8f5 g1f3 b8d7 b1d2 h7h6 a2a4 g8f6 h2h3 a7a5 d1b3 d8b6 e2e3 b6b3 d2b3 e7e6 f3d2 g7g5 f4c7 b7b6 b3c1 h6h5 f1d3 f6e4 d3e4 f5e4 d2e4 d5e4 h3h4 f8e7 h4g5 e7g5 c3c4 g5e7 b2b3 h5h4');
INSERT INTO evaluations VALUES(52755,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52755,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(52755,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52755,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(52755,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(52755,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(52755,36,'cp 13','c8f5 e2e3 e7e6 d1b3 b7b6 g1f3 g8f6 f1e2 a7a5 a2a4 b8d7 b1d2 f8e7 e1g1 e8g8 h2h3 h7h6 f1d1 d8c8 c3c4 c8b7 a1c1 f8c8 f3e1 f6e4 d2e4 f5e4 e2d3 c6c5 d3e4 d5e4 d4d5 e6d5 c4d5 b7a6 e1c2 c5c4 b3b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(31081,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31081,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(31081,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31081,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31081,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31081,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(31081,36,'cp 12','g1f3 e7e6 d1b3 b7b6 e2e3 g8f6 f1e2 a7a5 a2a4 b8d7 b1d2 f8e7 e1g1 e8g8 h2h3 h7h6 a1c1 d8c8 c3c4 c8b7 f1d1 f8d8 f4h2 d8c8 f3e5 d7e5 h2e5 f6d7 e5g3 e7b4 e2d3 f5d3 b3d3 c6c5 d2f3 c5d4 e3d4 d5c4 c1c4 c8c4');
INSERT INTO evaluations VALUES(31080,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(31080,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(31080,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31080,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31080,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31080,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31080,36,'cp 13','e7e6 d1b3 b7b6 h2h3 g8f6 e2e3 f8d6 f4d6 d8d6 c3c4 e8g8 f1e2 c6c5 e1g1 b8c6 b1c3 c6a5 b3a4 a5c4 e2c4 d5c4 a4c4 a8c8 f1d1 d6b8 c4a6 f5e4 c3e4 f6e4 a6d3 e4f6 a1c1 h7h6 d3c3 f8d8 c3e1 f6e4 d4c5 c8c5 c1c5');
INSERT INTO evaluations VALUES(22854,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(22854,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(22854,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(22854,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(22854,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(22854,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(22854,36,'cp -50','d7d5 f1b5 f8d6 d2d3 a7a6 b5c6 b7c6 e3e4 g8e7 d3d4 e5d4 d1d4 f7f6 e1g1 c6c5 d4d1 d5d4 c3e2 c8g4 c2c3 d4d3 e2f4 d6f4 c1f4 e7g6 f4e3 d8d7 h2h3 g4f3 d1f3 d7c6 b2b3 g6e5 f3f4 c5c4 b3c4 e8g8 c4c5 a6a5 f1e1 a8d8');
INSERT INTO evaluations VALUES(21742,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(21742,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(21742,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(21742,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(21742,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(21742,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(21742,36,'cp -25','f1b5 f8d6 d2d3 g8e7 b5a4 f7f6 e1g1 c8g4 h2h3 g4h5 e3e4 d5d4 c3d5 e8g8 d5e7 d8e7 c2c3 d4c3 b2c3 c6a5 a4c2 c7c5 g2g4 h5f7 f3h4 g7g6 c1h6 f8d8 h4g2 b7b5 g2e3 b5b4 c3b4 c5b4 c2a4 d6c5 e3d5 d8d5 e4d5 f7d5 f1e1');
INSERT INTO evaluations VALUES(21538,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(21538,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(21538,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(21538,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(21538,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(21538,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(21538,36,'cp -37','f8d6 d2d3 g8e7 b5a4 e8g8 a4b3 h7h6 e1g1 d6b4 e3e4 b4c3 b2c3 d5e4 d3e4 e7g6 d1d8 f8d8 c1e3 c6a5 f1d1 d8e8 d1d5 b7b6 b3a4 e8e7 d5d8 g8h7 a4b5 a5b7 d8d1 g6f4 e3f4 e5f4 f3d4 c8d7 b5d3 a8e8 f2f3');
INSERT INTO evaluations VALUES(28989,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(28989,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(28989,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(28989,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(28989,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(28989,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(28989,36,'cp 47','c2c4 e7e6 d1b3 b7b6 f1e2 a7a5 e1g1 a5a4 b3d1 d5c4 e2c4 b8d7 c4d3 f5d3 d1d3 d8c8 b1c3 c8a6 d3a6 a8a6 a1c1 f8e7 e3e4 e8g8 e4e5 f6e8 f4g5 e7g5 f3g5 h7h6 g5e4 f7f5 e5f6 e8f6 f1e1 g8f7 a2a3 f8b8 f2f4 f6e4 c3e4 b6b5');
INSERT INTO evaluations VALUES(28951,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(28951,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(28951,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(28951,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(28951,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(28951,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(28951,36,'cp 34','e7e6 d1b3 b7b6 f1e2 a7a5 e1g1 a5a4 b3c3 f6e4 c3c1 f8d6 f3e5 e8g8 f2f3 e4f6 f4g3 h7h6 a2a3 f5h7 b1c3 f6h5 g3e1 h5f6 e1h4 d5c4 e5c4 d6e7 c4e5 f6d7 h4e7 d8e7 e5d3 h7d3 e2d3 d7f6 e3e4 c6c5 d4d5 e6d5 c3d5 f6d5 e4d5 e7d6 d3b5');
INSERT INTO evaluations VALUES(29232,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(29232,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(29232,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(29232,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(29232,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(29232,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(29232,36,'cp 50','d1b3 b7b6 f1e2 a7a5 e1g1 a5a4 b3d1 d5c4 e2c4 b8d7 c4d3 f5d3 d1d3 f8e7 b1c3 d8c8 a1c1 c8a6 d3a6 a8a6 c1c2 e8g8 f1c1 h7h6 e3e4 c6c5 c1d1 g7g5 f4e3 c5d4 f3d4 f8c8 d4b5 d7e5 e3d4 e5c6 a2a3 a6a5');
INSERT INTO evaluations VALUES(29224,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(29224,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(29224,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(29224,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(29224,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(29224,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(29224,36,'cp 45','b7b6 f1e2 a7a5 e1g1 a5a4 b3d1 d5c4 e2c4 b8d7 c4d3 f5d3 d1d3 d8c8 b1c3 f8e7 f1c1 e8g8 e3e4 c8a6 d3a6 a8a6 e4e5 f6e8 c3e4 c6c5 d4c5 d7c5 e4c5 e7c5 f4e3 c5e3 f2e3 a6a7 c1c6 e8c7 e3e4 c7a6 a1c1 a6b4 c6b6 b4a2');
INSERT INTO evaluations VALUES(32656,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32656,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(32656,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(32656,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(32656,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(32656,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(32656,36,'cp 53','h2h3 g4f3 e2f3 c7c6 d2d4 e7e6 e1g1 f8e7 c3e2 b8d7 c2c4 e8g8 g2g3 f8e8 d1c2 e6e5 a2a3 e5d4 e2d4 d7e5 f3e2 a5b6 d4f3 e5f3 e2f3 a7a5 g1g2 f6d7 f1d1 d7f8 b2b3 a8d8 d1d8 e8d8 c1b2 f8e6 a1d1 d8d1 f3d1');
INSERT INTO evaluations VALUES(32654,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(32654,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32654,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(32654,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(32654,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(32654,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(32654,36,'cp 50','g4f3 e2f3 c7c6 e1g1 e7e6 d2d4 b8d7 c1f4 f8e7 d1e2 a8d8 c3d1 f6d5 f4d2 a5c7 g2g3 e7f6 c2c3 e8g8 d1e3 f8e8 f1e1 d5e3 d2e3 e6e5 d4e5 d7e5 f3g2 b7b5 e3f4 g7g5 f4e3 e8e6 e2c2 e5c4 e3c5 e6e5 c5d4 e5e1 a1e1 f6d4 c3d4 c7d7 e1d1 d8e8 g2f1 c4d6 b2b3');
INSERT INTO evaluations VALUES(42054,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42054,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42054,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42054,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42054,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42054,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42054,36,'cp 21','b4e7 b1c3 e8g8 c4d5 e6d5 f1d3 b7b6 f3e5 c8b7 e1g1 c7c5 d1f3 g7g6 f3h3 d8c8 h3c8 f8c8 f1d1 b8c6 e5c6 b7c6 d3a6 c8b8 a2a4 c6b7 a6b7 b8b7 f2f3 a7a6 g2g4 a8d8 d2e1 c5c4 e1h4 f6e8 h4e7 b7e7');
INSERT INTO evaluations VALUES(52178,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(52178,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(52178,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52178,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52178,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52178,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(52178,36,'cp -13','e5d4 g1f3 f8c5 f3d4 g8f6 b1c3 e8g8 f1e2 d7d5 e4d5 f6d5 e1g1 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 e2f3 c8e6 f3c6 a8b8 c6e4 h7h6 e4c2 a7a5 h2h3 c5e7 c2b3 e6b3 a2b3 b8b3 a1a5 b3c3 g2g3 e7f6 c1f4');
INSERT INTO evaluations VALUES(52190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52190,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(52190,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(52190,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(52190,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(52190,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(52190,36,'cp 0','g1f3 f8c5 f3d4 g8f6 b1c3 e8g8 f1e2 d7d5 e4d5 f6d5 e1g1 d5c3 b2c3 b8c6 d4c6 b7c6 e2d3 d8a5 d1c2 h7h6 c1f4 c8e6 h2h3 a8d8 f1d1 f8e8 a2a4 c5b6 d1e1 e6d5 a1d1 e8e1 d1e1 d5e6 d3h7 g8f8 h7e4 f8g8');
INSERT INTO evaluations VALUES(52189,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(52189,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(52189,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(52189,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(52189,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52189,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52189,36,'cp 0','f8c5 f3d4 g8f6 b1c3 e8g8 f1e2 d7d5 e4d5 f6d5 e1g1 d5c3 b2c3 b8c6 d4c6 b7c6 d1a4 d8e7 e2f3 a8b8 f3c6 b8b6 c1f4 f8d8 a4e4 e7e4 c6e4 b6b2 f4g3 c8e6 a2a4 d8d2 a1b1 g7g6 b1b2 d2b2 f1a1 b2e2 e4d3 e2d2 d3e4 d2e2');
INSERT INTO evaluations VALUES(36624,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36624,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36624,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36624,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36624,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36624,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36624,36,'cp 0','c2c3 b8c6 f1e2 f6d7 d4d5 c6e5 f3e5 d7e5 e1g1 e7e6 d5e6 c8e6 d2f3 d8e7 f3d4 e6d7 e3e4 e5c6 e2f3 c6e5');
INSERT INTO evaluations VALUES(36604,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36604,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36604,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36604,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(36604,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(36604,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36604,36,'cp 0','b8c6 f1e2 f6d7 d4d5 c6e5 f3e5 d7e5 e1g1 e7e6 d5e6 c8e6 d1c2 a7a5 d2f3 e6f5 c2d2 f5d7 f3d4 f7f5 d4b5 g8h8 a1d1 d7c6 b5d4 c6d7');
INSERT INTO evaluations VALUES(3503,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3503,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3503,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3503,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3503,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3503,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3503,36,'cp 0','f1e2 f6d7 d4d5 c6e5 f3e5 d7e5 e1g1 e7e6 d5e6 c8e6 d1c2 a7a5 d2f3 e6f5 c2d2 f5d7 f3d4 f7f5 a1d1 d8e7 f4g3 a5a4 h2h3 a8e8 b2b3 g6g5 d4b5 d7c6 b5d4');
INSERT INTO evaluations VALUES(57508,30,'cp 242',NULL);
INSERT INTO evaluations VALUES(57508,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(57508,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(57508,33,'cp 271',NULL);
INSERT INTO evaluations VALUES(57508,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(57508,35,'cp 288',NULL);
INSERT INTO evaluations VALUES(57508,36,'cp 279','e4f5 d8e7 f1c4 g8f6 g1f3 d7d6 g2g4 c7c6 e1g1 f6g4 d2d4 c8f5 h2h3 b8d7 h3g4 f5g4 c4e2 e8c8 d4d5 e7e8 d5c6 b7c6 b2b4 e8g6 f3h4 g4e2 h4g6 e2d1 g6h8 d1h5 c3e4 h5f3 e4g5 f3h5 g5f7 d8e8 c1g5 d7f6 a1b1 f8e7 f2f4 e5e4 f1e1 e8f8 g5f6 g7f6 e1e4 c8d7');
INSERT INTO evaluations VALUES(57485,30,'cp 292',NULL);
INSERT INTO evaluations VALUES(57485,31,'cp 279',NULL);
INSERT INTO evaluations VALUES(57485,32,'cp 303',NULL);
INSERT INTO evaluations VALUES(57485,33,'cp 299',NULL);
INSERT INTO evaluations VALUES(57485,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(57485,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(57485,36,'cp 275','d8e7 f1c4 g8f6 g1f3 d7d6 g2g4 c7c6 e1g1 f6g4 d2d4 c8f5 h2h3 b8d7 h3g4 f5g4 d1d3 g4f3 d3f3 e5d4 c1f4 e8c8 c3e4 d7e5 f4e5 d6e5 f3g3 d8e8 f2f4 c8b8 f4f5 b8a8 a1e1 e7d8 e1e2 f8e7 c4d3 e7f6 e2g2 a7a6 b2b3 a8a7 g2h2 d8d5 g3g4');
INSERT INTO evaluations VALUES(30962,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(30962,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30962,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(30962,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(30962,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(30962,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(30962,36,'cp 27','g6h7 g1f3 g8f6 f1d3 h7d3 d1d3 e7e6 c1d2 f8e7 e1c1 b8d7 g3e4 e8g8 h1g1 f6e4 d3e4 d7f6 e4e2 d8d5 c1b1 d5b5 e2b5 c6b5 g2g4 f6d5 g4g5 g8h7 a2a3 h6g5 f3g5 e7g5 g1g5 f7f6 g5g1 a8c8 d1e1 f8e8 g1g3 c8c7 c2c3 a7a6 g3g6 c7c8');
INSERT INTO evaluations VALUES(38167,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(38167,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(38167,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(38167,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(38167,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(38167,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(38167,36,'cp 135','b1c3 g8e7 f1d3 b8c6 g1f3 c8d7 e1g1 a8c8 c1d2 e7g6 a1c1 b4e7 c3e2 e8g8 c1c3 a7a5 d3b1 b7b5 h2h4 b5b4 c3d3 f7f6 e5f6 e7f6 h4h5 g6h4 f3h4 f6h4 h5h6 g7g6 d2e3 f8f7 d1d2 h4e7 f1e1 d8f8 e2g3 a5a4 a2a3 b4a3 b2a3');
INSERT INTO evaluations VALUES(38164,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(38164,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(38164,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(38164,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(38164,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(38164,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(38164,36,'cp 107','g8e7 f1d3 b8c6 g1f3 c8d7 e1g1 a8c8 c1d2 e7g6 a1c1 b4e7 c3e2 e8g8 c1c3 b7b5 c3c1 b5b4 d3b1 f7f6 b1g6 h7g6 e5f6 e7f6 d2e3 d8e8 e3f4 e8e7 d1d2 a7a5 h2h3 a5a4 d2e3 c6a5 b2b3 d7b5 f1e1 c8c1 e2c1 f8c8');
INSERT INTO evaluations VALUES(40259,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(40259,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(40259,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(40259,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(40259,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(40259,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(40259,36,'cp 115','f1d3 b8c6 g1f3 c8d7 e1g1 a7a6 c1d2 h7h6 a1c1 c6a5 f1e1 a5c4 d3c4 b4c3 c4d3 c3d2 d1d2 a8c8 c1c8 d7c8 h2h4 c8d7 d3b1 d7b5 h4h5 d8b6 e1e3 e8d7 e3b3 b6a7 g2g4 e7c6 g1g2 d7c8 b3c3 a7b6 d2f4 b6c7 g4g5 h6g5 f3g5');
INSERT INTO evaluations VALUES(44479,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44479,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(44479,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(44479,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44479,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(44479,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(44479,36,'cp 23','c2c3 d7d5 g1f3 c8f5 f3h4 f5b1 a1b1 e7e6 e2e3 f8d6 f1d3 e8g8 e1g1 c6c5 d4c5 d6c5 h4f3 c5d6 f4d6 d8d6 c3c4 b8c6 c4d5 d6d5 a2a3 h7h6 d3e2 d5d1 f1d1 f6d5 f3e1 f8c8 g2g3 g7g5');
INSERT INTO evaluations VALUES(35746,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(35746,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(35746,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(35746,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(35746,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(35746,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(35746,36,'cp 24','h7h6 g5h4 c6c5 d4c5 d5c4 d3c4 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 b5d6 f8d8 d6e4 d8d1 e4f6 g7f6 f1d1 b4e7 a2a3 e6e5 c4d5 g8f8 f3d2 c8e6 d5e6 f7e6 f2f3 f8f7 h4e1 f6f5 d2c4 a8d8 d1d8 e7d8 c4d6 f7e7 d6b7 d8b6');
INSERT INTO evaluations VALUES(35720,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35720,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(35720,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35720,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(35720,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(35720,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(35720,36,'cp 19','g5h4 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 c4e2 b4e7 h4g3 a7a6 b5d6 f6h5 g3e5 h5f6 d6c4 b7b5 c4b6 a8a7 e5f4 g7g5 f3e5 c6e5 f4e5 f6d7 b6d7 a7d7 d1d7 c8d7 f1d1 d7c6 d1c1 c6d5 a2a3 f8d8 c1c7 e7d6 e5d6 d8d6');
INSERT INTO evaluations VALUES(54309,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(54309,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(54309,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(54309,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(54309,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(54309,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(54309,36,'cp -60','b8c6 f1b5 g8e7 e1g1 c8d7 f1e1 d5d4 c3e4 e7g6 d2d3 h7h6 h2h4 g6e5 f3e5 c6e5 b5d7 d8d7 e4c5 f8c5 e1e5 c5d6 e5e1 e8g8 c1d2 a8c8 c2c3 d4c3 d2c3 c8c5 d1g4 f7f5 g4f3 c5d5 d3d4 f8e8 a1c1 d6e7 g2g3 e7f6 c1d1 g8f7 h4h5 e8c8');
INSERT INTO evaluations VALUES(20355,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(20355,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(20355,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(20355,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(20355,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(20355,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(20355,36,'cp -28','f1b5 g8e7 e1g1 a7a6 b5c6 e7c6 c3e2 d5d4 d2d3 d8c7 c1f4 b7b6 c2c3 d4c3 b2c3 c8b7 e2g3 f7f5 d1b3 c6d8 d3d4 c5c4 b3b2 b7f3 g2f3 f8e7 b2d2 c7c6 f4g5 e7g5 d2g5 e8g8 g3h5 f8f7 g5g3 g8h8 h5f4 b6b5 h2h4 b5b4 c3b4');
INSERT INTO evaluations VALUES(20199,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(20199,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(20199,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(20199,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(20199,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(20199,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(20199,36,'cp -53','g8e7 e1g1 a7a6 b5c6 e7c6 c3e2 d5d4 d2d3 b7b6 e2g3 f8e7 g3e4 e8g8 c1f4 d8d5 f1e1 c8b7 a2a3 b6b5 d1d2 a8d8 b2b3 d8d7 f4g5 h7h6 g5h6 c6e5 f3e5 d5e5 h6g5 e7g5 e4g5 e5d5 f2f3 c5c4 b3c4 b5c4 g5e4 c4d3');
INSERT INTO evaluations VALUES(5381,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5381,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5381,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5381,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5381,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(5381,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5381,36,'cp 0','e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4g5 d8a5 c3d5 h6f5 g5d2 a5d8 d2g5');
INSERT INTO evaluations VALUES(5376,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5376,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5376,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5376,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5376,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5376,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5376,36,'cp 0','g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4g5 d8a5 c3d5 h6f5 g5d2 a5d8 d2g5');
INSERT INTO evaluations VALUES(15071,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(15071,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(15071,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(15071,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(15071,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(15071,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(15071,36,'cp 60','f6e4 d1d5 e4c3 d5d8 c6d8 b2c3 d8c6 c1f4 f8c5 f1c4 e8g8 a1d1 h7h6 f3d4 g7g5 d4c6 g5f4 c6b4 f8e8 e1g1 c5b4 c3b4 a7a5 b4b5 c8e6 c4e6 e8e6 f1e1 a8e8 d1d5 f7f6 g2g4 e6e5 e1e5 e8e5 d5e5 f6e5 f2f3 b7b6 g1g2 g8f7 h2h4 f7g6 c2c3 a5a4 c3c4');
INSERT INTO evaluations VALUES(15373,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(15373,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(15373,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(15373,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(15373,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(15373,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(15373,36,'cp 62','d1d5 e4c3 d5d8 c6d8 b2c3 d8c6 f1c4 f8c5 c1f4 e8g8 e1g1 h7h6 f1e1 c8f5 c4b3 c5b6 a1d1 f8d8 g1f1 d8d1 e1d1 a8e8 h2h3 g7g5 f4g3 c6a5 h3h4 g5g4 f3d4 e8d8 f1e2 f5g6 f2f3 a5b3 c2b3 c7c5 d4b5 d8d1 e2d1 g4f3');
INSERT INTO evaluations VALUES(15900,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15900,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15900,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15900,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15900,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15900,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15900,36,'cp 0','c7c6 f1e2 c6d5 e4d5 f8e7 a2a4 d8a5 e1g1 e8g8 a1b1 a7a6 h2h3 a5b4 b1a1 b4a5');
INSERT INTO evaluations VALUES(12197,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12197,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12197,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12197,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12197,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12197,36,'cp 5','f1e2 c6d5 e4d5 f8e7 a2a4 d8a5 e1g1 e8g8 a1b1 a7a6 h2h3 a5b4 f1e1 f6e4 c3e4 b4e4 e2a6 e4a4 a6f1 d7f6 b2b3 a4e8 c2c4 c8f5 f1d3 e8d7 d3f5 d7f5 c1e3 a8a2 b3b4 f8a8 c4c5 e5e4 f3h4 f5d5 d1d5 f6d5 h4f5');
INSERT INTO evaluations VALUES(22152,30,'cp 468',NULL);
INSERT INTO evaluations VALUES(22152,31,'cp 454',NULL);
INSERT INTO evaluations VALUES(22152,32,'cp 471',NULL);
INSERT INTO evaluations VALUES(22152,33,'cp 461',NULL);
INSERT INTO evaluations VALUES(22152,34,'cp 475',NULL);
INSERT INTO evaluations VALUES(22152,35,'cp 467',NULL);
INSERT INTO evaluations VALUES(22152,36,'cp 477','b2b4 c7c6 b4a5 g8f6 b1c3 d8a5 c1d2 a5c7 f1d3 f8e7 e1g1 c6d5 c3d5 f6d5 e4d5 e8g8 a1b1 f7f5 d3e2 b7b6 g1h1 e7f6 f3g1 c8d7 c2c4 e5e4 f2f3 a8e8 d1c1 f6d4 c1a3 d4c5 d2b4 e4f3 e2f3 c5b4 a3b4 g7g5 b1e1 e8e1 f1e1 g5g4');
INSERT INTO evaluations VALUES(22150,30,'cp 463',NULL);
INSERT INTO evaluations VALUES(22150,31,'cp 477',NULL);
INSERT INTO evaluations VALUES(22150,32,'cp 477',NULL);
INSERT INTO evaluations VALUES(22150,33,'cp 454',NULL);
INSERT INTO evaluations VALUES(22150,34,'cp 460',NULL);
INSERT INTO evaluations VALUES(22150,35,'cp 458',NULL);
INSERT INTO evaluations VALUES(22150,36,'cp 460','g8f6 b1c3 c7c6 b4a5 d8a5 c1d2 a5c7 a1b1 f8e7 f1d3 e8g8 e1g1 c6d5 c3d5 f6d5 e4d5 f7f5 d3e2 c8d7 c2c4 b7b6 g1h1 e7f6 f3g1 e5e4 d1b3 a8e8 b3a3 f6d4 d2f4 d4c5 a3g3 f8f6 g1h3 c5d4 f1d1 d4c5 f4e3 f6g6 g3f4 g6h6 e3c5 c7c5');
INSERT INTO evaluations VALUES(4056,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(4056,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(4056,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(4056,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(4056,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(4056,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(4056,36,'cp 63','e1g1 e7e5 h2h3 e5d4 f3d4 f8e8 c1e3 a7a6 d1c2 d7c5 a1d1 c5d3 c2d3 b7b5 d4c6 d8d7 c6a5 d7e6 c4c5 d6c5 e3c5 f6d7 c5d4 g7d4 d3d4 e6b6 d4b6 d7b6 b2b3 b5b4 c3d5 b6d5 e4d5 e8e2 a5c6 e2a2 f1e1 g8g7 c6b4 a2b2 b4c6 c8d7 e1e7 d7c6 d5c6 b2b3 e7c7');
INSERT INTO evaluations VALUES(4055,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(4055,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(4055,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(4055,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(4055,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(4055,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(4055,36,'cp 50','e7e5 c1e3 e5d4 f3d4 f8e8 h2h3 c7c6 f1e1 d7c5 d1c2 a7a5 a1d1 a5a4 d4f3 d8a5 e3f4 a4a3 b2b3 f6d7 f4d2 d7e5 f3e5 g7e5 c3a4 a5c7 a4c5 d6c5 d2e3 e5d4 e3d4 c5d4 c2d2 c6c5 f2f4 c7a5 d2f2 c8d7 e4e5 d7c6 d3e4 c6e4');
INSERT INTO evaluations VALUES(3961,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(3961,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(3961,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(3961,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(3961,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(3961,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(3961,36,'cp 51','h2h3 e5d4 f3d4 f8e8 c1e3 d7c5 f2f3 c7c6 f1e1 d6d5 d4c6 b7c6 e3c5 d5e4 d3e4 f6e4 e1e4 d8d1 a1d1 c8e6 c5d4 e6c4 d4g7 e8e4 c3e4 g8g7 b2b3 c4d5 g1f2 h7h5 d1d4 a7a5 e4c3 d5e6 d4d6 a8c8 f2e3 c8c7 c3e4 c6c5 d6a6 c5c4');
INSERT INTO evaluations VALUES(18225,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(18225,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(18225,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(18225,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(18225,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(18225,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(18225,36,'cp 69','c3d5 g8f6 c2c4 b5b4 a3c2 f6e4 g2g3 a8b8 f1g2 e4c5 e1g1 f8e7 a2a3 c5b3 c2b4 c6d4 c1e3 e8g8 a1c1 a6a5 e3d4 b3d4 d5e7 d8e7 b4d5 e7g5 f2f4 g5g4 f1f2 g4d1 c1d1 c8e6 d5e7 g8h8 f4e5 d6e5 c4c5 f8d8 d1c1 d8d7 e7c6 d4c6 g2c6 d7d4 c6g2 b8c8 g2b7 c8c7 c5c6 g7g6');
INSERT INTO evaluations VALUES(18224,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(18224,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(18224,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(18224,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(18224,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(18224,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(18224,36,'cp 48','g8f6 c2c4 b5b4 a3c2 f6e4 g2g3 a8b8 f1g2 e4c5 e1g1 f8e7 c1e3 e8g8 b2b3 a6a5 d1e2 c8b7 a1d1 d8e8 f2f4 g8h8 e3c5 d6c5 f4e5 e7d6 d5f6 g7f6 d1d6 e8e5 e2e5 c6e5 f1f6 b7g2 g1g2 f8d8 c2e3 a5a4 b3a4 h8g7 e3f5 g7f8 d6d8 b8d8 f6d6 d8d6 f5d6');
INSERT INTO evaluations VALUES(11799,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(11799,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(11799,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(11799,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11799,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11799,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(11799,36,'cp 52','f1d3 f8e7 e1g1 e8g8 f1e1 h7h6 c2c4 b5b4 a3c2 a6a5 b2b3 f6d7 c2e3 e7g5 d3c2 c6d4 c1b2 g5e3 d5e3 d4c2 d1c2 d7c5 a1d1 d8e7 b2c1 c8e6 e3f5 e6f5 e4f5 a5a4 c1e3 a4b3 a2b3 a8a3 e3c5 d6c5 d1d5 e7a7');
INSERT INTO evaluations VALUES(11796,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(11796,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(11796,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11796,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(11796,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11796,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(11796,36,'cp 28','f8e7 e1g1 e8g8 c2c4 f6d5 c4d5 c6d4 a3c2 d4c2 d1c2 c8d7 c1e3 e7g5 c2d2 g5e3 d2e3 a8c8 a2a3 c8c5 f1c1 c5c1 a1c1 d8b8 d3e2 g7g6 b2b4 h7h5 e2f1 g8g7 h2h4 b8d8');
INSERT INTO evaluations VALUES(50572,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(50572,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(50572,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(50572,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(50572,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(50572,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(50572,36,'cp 44','d4f3 g8f6 f1c4 f8b4 c1d2 d8e7 e1g1 b4d2 d1d2 f6e4 d2e3 e4f6 e3e5 e7e5 f3e5 d7d5 c4d3 b8c6 f1e1 c8e6 b1d2 e8e7 c2c3 h8e8 e5f3 e7f8 h2h3 a8c8 g1f1 c6a5 f3d4 e6d7 e1e8 d7e8 a1e1 g7g6 f2f3');
INSERT INTO evaluations VALUES(50574,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(50574,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(50574,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(50574,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(50574,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(50574,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(50574,36,'cp 44','g8f6 f1c4 f8b4 c1d2 d8e7 e1g1 b4d2 d1d2 f6e4 d2e3 e4f6 f3e5 d7d5 c4d3 e8g8 b1d2 f8e8 f1e1 b8c6 d2f3 e7c7 e3c5 c8d7 a1d1 a8c8 e5c6 d7c6 h2h3 g7g6 c5d4 e8e1 d1e1 c7d6 d3f1 c8e8 e1e8 f6e8 f3e5 e8f6 c2c3 f6d7 e5c6 d6c6');
INSERT INTO evaluations VALUES(54204,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(54204,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(54204,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(54204,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(54204,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(54204,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(54204,36,'cp -28','g8f6 d4c5 f8c5 f1d3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 c1e3 c5e3 f2e3 d8e7 d1e2 a8c8 d3c2 f8d8 a1d1 e7c5 c3a4 c5a7 a4c3 a7b6 c3a4 b6c7 d1d8 c7d8 f1d1 d8c7 a4c3 g8f8');
INSERT INTO evaluations VALUES(57557,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(57557,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(57557,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(57557,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(57557,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(57557,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(57557,36,'cp 106','f6f5 c2c4 g8f6 b1c3 e7e6 a2a3 g7g6 c1f4 f8g7 e2e3 e8g8 a1c1 d7d6 c4c5 b8c6 f1c4 g8h8 h2h4 h7h6 b2b4 a7a6 c3e2 f6h5 c5d6 c7d6 d4d5 e6e5 d5c6 e5f4 e2f4 h5f4 e3f4 d8e7 c4e2 b7c6 e1g1 c8d7 f1e1');
INSERT INTO evaluations VALUES(23080,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(23080,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(23080,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(23080,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(23080,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(23080,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(23080,36,'cp 43','d7d6 b1c3 g8f6 c1f4 e5g6 f4g3 f8e7 d1d2 e8g8 e1c1 h7h6 h2h4 f8e8 f1e2 a7a6 a2a3 b7b5 h4h5 g6f8 f3d4 c8d7 e2f3 a8b8 e4e5 d6e5 d4c6 d7c6 f3c6 d8d2 d1d2 e8d8 d2d8 b8d8 g3e5 f8e6 c6f3 e7d6 e5f6 g7f6 c3d5 f6f5 h1d1 d6c5 d5e3 d8d1 c1d1 e6d4 d1d2 a6a5 d2c3 d4f3 g2f3');
INSERT INTO evaluations VALUES(49355,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(49355,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(49355,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(49355,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(49355,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(49355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49355,36,'cp 0','f8e7 e2e4 f6e4 g5e7 e4c3 e7d8 c3d1 d8c7 d1b2 c7d6 b2a4 e1d2 a7a6 c2c4 b8c6 g1f3 f7f6 a1b1 c6e7 b1b3 e7f5 d6c7 f5e7');
INSERT INTO evaluations VALUES(42970,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42970,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42970,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42970,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42970,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42970,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42970,36,'cp 0','e2e4 f6e4 g5e7 e4c3 e7d8 c3d1 d8c7 d1b2 c7d6 b2a4 e1d2 b8c6 g1f3 a7a6 c2c4 b7b5 c4b5 c6e7 a1b1 a6b5 f1b5 a4b6 a2a4 f7f6 h1c1 e8f7 d6e7 f7e7 b5c6 a8b8 c6b5');
INSERT INTO evaluations VALUES(42975,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42975,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42975,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42975,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42975,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42975,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42975,36,'cp 0','f6e4 g5e7 e4c3 e7d8 c3d1 d8c7 d1b2 c7d6 b2a4 e1d2 b8c6 g1f3 a7a6 c2c4 c6e7 f1d3 b7b5 h1c1 b5c4 d3e4 c4c3 d2c2 a8a7 a1b1 a6a5 b1b8 a7a6 d6a3 a6b6 b8a8 b6a6');
INSERT INTO evaluations VALUES(8598,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(8598,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(8598,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(8598,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(8598,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(8598,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(8598,36,'cp 26','a2a4 e8g8 c2c3 a7a5 d1b3 c8g4 b1d2 f6d7 h2h3 d7c5 b3c2 g4e6 f1e1 h7h6 d3d4 e5d4 f3d4 e7g5 c4e6 f7e6 d2b3 c5b3 c2b3 c6d4 c3d4 d8f6 e1e2 f6d4 c1g5 h6g5 b3e6 f8f7 e6d5 d4d5 e4d5 f7f6 a1d1 a8f8 b2b3 f8f7 f2f3');
INSERT INTO evaluations VALUES(23320,30,'cp 366',NULL);
INSERT INTO evaluations VALUES(23320,31,'cp 382',NULL);
INSERT INTO evaluations VALUES(23320,32,'cp 385',NULL);
INSERT INTO evaluations VALUES(23320,33,'cp 382',NULL);
INSERT INTO evaluations VALUES(23320,34,'cp 386',NULL);
INSERT INTO evaluations VALUES(23320,35,'cp 390',NULL);
INSERT INTO evaluations VALUES(23320,36,'cp 373','f3d4 e5d4 d1d4 c7c5 d4a4 d8c7 f1e2 a7a6 e1g1 g8e7 a4b3 e7g6 a2a4 f8d6 g2g3 d6e7 f2f4 d7d6 b1d2 b7b6 d2c4 a8b8 c4e3 e8g8 c1d2 g6h8 c2c4 h8f7 d2c3 c8h3 f1e1 c7d7 b3c2 e7d8 b2b3 f7h6 c2d3 h6f7 c3b2 f8e8 e2f3 b6b5');
INSERT INTO evaluations VALUES(23319,30,'cp 366',NULL);
INSERT INTO evaluations VALUES(23319,31,'cp 373',NULL);
INSERT INTO evaluations VALUES(23319,32,'cp 377',NULL);
INSERT INTO evaluations VALUES(23319,33,'cp 376',NULL);
INSERT INTO evaluations VALUES(23319,34,'cp 374',NULL);
INSERT INTO evaluations VALUES(23319,35,'cp 393',NULL);
INSERT INTO evaluations VALUES(23319,36,'cp 389','e5d4 d1d4 c7c5 d4a4 d8c7 f1e2 a7a6 e1g1 g8e7 a4b3 e7g6 a2a4 f8d6 g2g3 d6e7 b1d2 b7b6 d2c4 a8b8 d5d6 e7d6 f1d1 d6e5 f2f4 d7d5 d1d5 e5d4 g1g2 c8e6 c2c3 g6e7 f4f5 e6f7 c3d4 e7d5 e4d5 f7d5 g2f2 c5d4 c1f4 c7c5');
INSERT INTO evaluations VALUES(23314,30,'cp 382',NULL);
INSERT INTO evaluations VALUES(23314,31,'cp 394',NULL);
INSERT INTO evaluations VALUES(23314,32,'cp 387',NULL);
INSERT INTO evaluations VALUES(23314,33,'cp 364',NULL);
INSERT INTO evaluations VALUES(23314,34,'cp 385',NULL);
INSERT INTO evaluations VALUES(23314,35,'cp 378',NULL);
INSERT INTO evaluations VALUES(23314,36,'cp 381','d1d4 f8d6 b1c3 d8e7 c3b5 d6e5 d5d6 e5d6 b5d6 e7d6 d4d6 c7d6 c1f4 g8h6 f4d6 h6f7 d6a3 b7b6 f2f3 c8b7 b2b3 e8c8 e1c1 f7e5 a3d6 h7h5 h2h4 d8e8 d6b4 e8d8 f1b5 b7c6 b5a6 c6b7 a6b7 c8b7 b4d6 h8h6 d1d2 h6g6 a2a4 d8c8 d6e5 f6e5');
INSERT INTO evaluations VALUES(23313,30,'cp 371',NULL);
INSERT INTO evaluations VALUES(23313,31,'cp 380',NULL);
INSERT INTO evaluations VALUES(23313,32,'cp 378',NULL);
INSERT INTO evaluations VALUES(23313,33,'cp 378',NULL);
INSERT INTO evaluations VALUES(23313,34,'cp 373',NULL);
INSERT INTO evaluations VALUES(23313,35,'cp 355',NULL);
INSERT INTO evaluations VALUES(23313,36,'cp 348','f8d6 b1c3 d8e7 c3b5 d6e5 d5d6 e5d6 b5d6 e7d6 d4d6 c7d6 c1f4 g8h6 f4d6 h6f7 d6a3 b7b6 e1c1 c8b7 b2b3 b7c6 h2h4 b6b5 h1h3 a7a5 f1e2 b5b4 a3b2 e8g8 d1d4 f8e8 h3e3 h7h6 e2g4 f7e5 g4h5 g7g6 h5e2 h6h5 e3g3 a5a4 f2f4 a4b3 f4e5 b3a2');
INSERT INTO evaluations VALUES(26531,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(26531,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(26531,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(26531,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(26531,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(26531,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(26531,36,'cp 23','e2e4 f5e4 c3e4 d5e4 d4d5 c6b4 c2c4 b4a6 d1a4 d8d7 a4d7 e8d7 f2f3 e7e5 d5e6 d7e6 e1c1 g8f6 f3e4 f8c5 g1h3 a8d8 h3g5 e6e7 f1e2 h7h6 e4e5 h6g5 f4g5 c5d4 e5f6 d4f6 g5f4 d8d1 h1d1 h8d8 d1d8 e7d8 h2h3 a6c5 b2b4 c5e6 f4d2 c7c5 b4b5 f6d4 e2g4');
INSERT INTO evaluations VALUES(27228,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(27228,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(27228,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(27228,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(27228,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(27228,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(27228,36,'cp 159','c3d4 g7g6 b1c3 a7a6 b5e2 f8g7 c1g5 g8e7 d4d5 c6e5 f3e5 g7e5 d1d2 e5g7 h2h4 h7h6 g5f4 h6h5 a1c1 e8g8 e1g1 c7c6 f4d6 f8e8 d6g3 c6d5 c3d5 e7d5 e4d5 d8b6 c1c7 a8d8 e2f3 b6b2 d2b2 g7b2 c7b7 b2e5 g3e5 e8e5 f1d1');
INSERT INTO evaluations VALUES(27224,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(27224,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(27224,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(27224,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(27224,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(27224,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(27224,36,'cp 163','g7g6 c1g5 f8e7 g5e3 a7a6 b5e2 d6d5 e4e5 e7f8 d1b3 g8e7 b1c3 f8g7 e1g1 c6a5 b3a3 a5c6 a1d1 e8g8 f1e1 d7e6 a3b3 b7b5 h2h3 e7f5 e3c1 f7f6 g2g4 f5e7 e5f6 f8f6 f3g5 d8d7 e2f3 a8f8 f3g2 e6f7 g5f7 f6f7 c3d5');
INSERT INTO evaluations VALUES(32151,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(32151,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(32151,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(32151,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(32151,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(32151,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(32151,36,'cp 69','e2f3 b8c6 c3e2 g7g6 c2c3 h7h5 d1b3 a8b8 d2d4 f8h6 c1h6 g8h6 b3a4 e8g8 d4d5 b7b5 a4a3 c6e7 a3a7 f7f5 a7e3 g8g7 e3d2 h6g8 e4f5 e7f5 e2g3 f5g3 h2g3 g8f6 e1g1 e5e4 f3d1 f8e8 d1b3 e8e5 c3c4 d8d7 c4b5 b8b5');
INSERT INTO evaluations VALUES(32147,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(32147,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(32147,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(32147,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(32147,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(32147,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(32147,36,'cp 52','b8c6 e1g1 c6d4 d2d3 g7g6 f3g4 h7h5 g4h3 f8h6 c1e3 h6e3 f2e3 d4e6 h3e6 f7e6 d1f3 g8h6 d3d4 d8g5 f3f6 g5f6 f1f6 e8e7 f6g6 a8g8 g6g8 h8g8 h2h3 h6f7 a1f1 f7g5 g1f2 h5h4 d4e5 d6e5 f2g1 c7c6 a2a4');
INSERT INTO evaluations VALUES(27018,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(27018,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(27018,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(27018,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(27018,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(27018,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(27018,36,'cp 56','e1g1 c6d4 d2d3 c7c6 c1e3 g7g6 a2a4 f8g7 f3g4 d8e7 c3e2 d4e2 g4e2 g8f6 c2c3 e8g8 a4a5 b7b5 g2g3 d6d5 e2f3 f8e8 b2b4 a8d8 d1e2 h7h5 f3g2 e7c7 h2h3 c7d7 a1d1 d8c8 f1e1 d5d4 c3d4 e5d4');
INSERT INTO evaluations VALUES(27017,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(27017,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(27017,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(27017,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(27017,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(27017,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(27017,36,'cp 49','c6d4 d2d3 c7c6 c1e3 g7g6 e3d4 e5d4 c3e2 f8g7 a1b1 a8b8 g2g3 g8e7 h2h4 e8g8 d1d2 d6d5 h4h5 d8d6 h5h6 g7h8 e2f4 d5e4 f3e4 h8f6 f4h3 e7d5 b1e1 d5c7 e4g2 b8e8 a2a4 e8e1 f1e1 f8e8 e1e8 c7e8');
INSERT INTO evaluations VALUES(34164,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(34164,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(34164,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(34164,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(34164,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(34164,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(34164,36,'cp 214','d2d4 c7c6 g1f3 c8a6 e1g1 a6e2 c3e2 g8f6 f1e1 g7g6 c1f4 f8g7 e2g3 e6c8 c2c4 b8d7 h2h4 e8g8 e1e7 b6b5 b2b3 f8e8 e7e8 c8e8 a1c1 b5c4 b3c4 e8e6 d1b3 a8e8 b3b7 c6c5 d4d5 e6g4 f4e3 d7b6 b7a6 f6d7 a6a7');
INSERT INTO evaluations VALUES(34161,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(34161,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(34161,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(34161,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(34161,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(34161,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(34161,36,'cp 208','c7c6 g1f3 c8a6 e1g1 a6e2 c3e2 g8f6 f1e1 g7g6 c1f4 f8g7 e2g3 e6c8 c2c4 b8d7 h2h4 e8g8 e1e7 b6b5 b2b3 c8b7 d1c2 a8e8 a1e1 e8e7 e1e7 b5c4 b3c4 f8e8 e7e8 f6e8 g3e4 b7b6 f4e3 b6a6 g2g3 d7b6 f3d2 a6a4 c2a4 b6a4');
INSERT INTO evaluations VALUES(57457,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(57457,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(57457,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(57457,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(57457,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(57457,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(57457,36,'cp 213','d2d4 d7d5 e4d5 g8e7 f1c4 e7d5 d1e2 f8e7 b1c3 c7c6 e1g1 e8g8 c1f4 g8h8 c4d5 c6d5 c3d5 d8d5 e2e7 f8g8 e7d6 d5d6 f4d6 c8e6 f1e1 e6f7 d6b8 g8b8 e1e7 f7d5 g1f2 h7h6 c2c3 h8h7 b2b3 b8e8 a1e1 e8e7');
INSERT INTO evaluations VALUES(57454,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(57454,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(57454,32,'cp 215',NULL);
INSERT INTO evaluations VALUES(57454,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(57454,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(57454,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(57454,36,'cp 201','d7d5 e4d5 c7c6 d5c6 b8c6 c1f4 g7g5 f4g3 g5g4 f3h4 c6d4 f1c4 d8e7 e1f1 c8e6 c4e6 d4e6 d1g4 g8h6 g4h5 e7f7 h5b5 f7d7 b5d7 e8d7 b1c3 d7c6 h2h3 a8e8 a1d1 e6g5 d1d3 f8c5 g3f2 c5f2 f1f2 h6f7 h4f5 f7e5');
INSERT INTO evaluations VALUES(20518,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(20518,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(20518,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(20518,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(20518,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(20518,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(20518,36,'cp 37','f1b5 f8d6 d4c5 d6c5 e1g1 g8e7 b1d2 e8g8 d2b3 c5d6 f1e1 c8g4 h2h3 g4h5 b5e2 f8e8 c1e3 d8c7 f3d4 h5e2 e1e2 a7a6 d1c2 a8d8 a1d1 h7h6 d4f3 e7g6 b3d4 c6a5 d4f5 a5c4 f5d6 c7d6 d1e1 c4e3 e2e3 e8e3 e1e3');
INSERT INTO evaluations VALUES(20460,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(20460,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(20460,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20460,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(20460,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(20460,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(20460,36,'cp 33','c5d4 f3d4 c8d7 e1g1 f8d6 f1e1 g8e7 c1e3 a7a6 b5a4 d8c7 h2h3 e8g8 b1d2 c6d4 e3d4 e7f5 a4d7 f5d4 d7g4 d4c6 d2b3 a8d8 g4f3 d6e5 d1c2 d5d4 c3d4 e5d4 b3d4 d8d4 a1d1 c7a5 c2b3 a5b4 b3b4 d4b4 b2b3 b4b6 d1d6 g7g6 f3c6 b6c6 d6c6 b7c6');
INSERT INTO evaluations VALUES(40913,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(40913,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(40913,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(40913,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40913,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(40913,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40913,36,'cp 13','f6d5 g5e7 d8e7 d1c2 d5c3 c2c3 b7b6 g1f3 c8b7 g2g3 e8g8 f1g2 c7c5 e1g1 b8d7 d4c5 d7c5 f1d1 a8c8 c3e3 f8d8 a1c1 d8d1 c1d1 c8d8 d1d4 h7h6 f3e5 b7g2 g1g2 d8d4 e3d4 f7f6 e5d3 e7d7 d4d7 c5d7 f2f4 g8f7 e2e4 e6e5 g2f2');
INSERT INTO evaluations VALUES(40892,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40892,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40892,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40892,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(40892,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40892,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(40892,36,'cp 21','g5e7 d8e7 d1c2 e8g8 g1f3 d5c3 c2c3 b7b6 g2g3 c8b7 f1g2 c7c5 e1g1 b8d7 d4c5 d7c5 a1c1 a8c8 c3e3 f8d8 b2b4 c5d7 b4b5 c8c5 e3a3 h7h6 a3a7 b7d5 c1c5 e7c5 f1d1 c5b5 f3d4 b5c5 g2d5 c5d5');
INSERT INTO evaluations VALUES(40669,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40669,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(40669,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40669,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(40669,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40669,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40669,36,'cp 15','d8e7 d1c2 e8g8 g1f3 b7b6 c3d5 e6d5 a1c1 b8d7 e2e3 c8b7 f1d3 g7g6 e1g1 c7c5 b2b3 f8c8 c2b2 a7a5 d3b5 b7c6 b5c6 c8c6 h2h3 a5a4 f1d1 d7f6 c1c2 a4b3 a2b3 a8c8 f3e5 c6c7 e5d3 e7d6 d4c5 b6c5 b3b4 c5c4');
INSERT INTO evaluations VALUES(33403,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(33403,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(33403,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(33403,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(33403,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(33403,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(33403,36,'cp 16','d1c2 e8g8 g1f3 d5c3 c2c3 b7b6 g2g3 c8b7 f1g2 c7c5 e1g1 b8d7 d4c5 d7c5 a1c1 a8c8 c3e3 f8d8 b2b4 c5d7 f3d4 b7g2 g1g2 e7d6 e3f4 d6d5 d4f3 d5b7 f1d1 h7h6 f4d6 d7f6 d6d8 c8d8 d1d8 g8h7 a2a3');
INSERT INTO evaluations VALUES(44547,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(44547,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(44547,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(44547,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(44547,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(44547,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(44547,36,'cp 86','d4d5 d7d6 e2e4 e7e6 d5e6 c8e6 g1e2 b8c6 e2f4 f8e7 g2g3 e8g8 f1g2 e6g4 f2f3 g4d7 e1g1 f8e8 c1e3 e7f8 e3f2 a8b8 d1d2 h7h6 b2b3 b7b5 a1d1 a7a5 f4d5 f6d5 c3d5 c6b4 d5c3 d7c6 f3f4');
INSERT INTO evaluations VALUES(46035,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46035,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(46035,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(46035,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(46035,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(46035,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(46035,36,'cp 12','a7a6 f2f4 d8b6 d4b3 b8d7 d1d2 b6c7 e1c1 b7b5 a2a3 c8b7 f1d3 f8e7 c1b1 h7h6 g5h4 e8g8 g2g4 d7c5 b3c5 d6c5 g4g5 f6h7 d2g2 h6g5 f4g5 b5b4 a3b4 c5c4 d3e2 e7b4 c3a2 a6a5 g5g6 h7f6 g2g3 c7g3 h4g3 f7g6 a2b4');
INSERT INTO evaluations VALUES(46169,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(46169,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(46169,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(46169,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(46169,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(46169,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(46169,36,'cp -13','c5d4 e3d4 f8b4 b1d2 d5c4 a2a3 b4d2 d1d2 b7b5 a3a4 b5a4 f1c4 b8d7 c4d3 e8g8 a1a4 c8b7 e1g1 b7f3 g2f3 f6d5 f4g5 d8b6 d3e4 f8b8 f1a1 d7f8 h2h4 h7h6 g5f4 b6b2 d2b2 b8b2 e4d5 e6d5 a4a7 a8a7 a1a7');
INSERT INTO evaluations VALUES(46313,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46313,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46313,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46313,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46313,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46313,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(46313,36,'cp 0','e3d4 f8b4 b1d2 d5c4 a2a3 b4d2 d1d2 b7b5 a3a4 b5a4 f1c4 b8d7 c4d3 e8g8 a1a4 c8b7 e1g1 b7f3 g2f3 f6d5 f4g5 d8c7 f1c1 c7b6 c1a1 f8b8 d3e4 d7f8 g5f4 b6b2 d2b2 b8b2 e4d5 e6d5 a4a7 a8a7 a1a7 f7f6 h2h4 b2b4 f4e3 f8g6 h4h5 g6h4');
INSERT INTO evaluations VALUES(46306,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(46306,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(46306,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(46306,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46306,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(46306,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46306,36,'cp 0','f8b4 b1d2 d5c4 a2a3 b4e7 d2c4 e8g8 f1d3 c8d7 a1c1 d7b5 e1g1 b8c6 h2h3 f6d5 f4g3 a8c8 d1d2 h7h6 f1d1 b5c4 c1c4 e7d6 g3e5 f7f6 e5d6 d8d6 d1c1 f8d8 d3e4 c6e7 c4c8 d8c8 g2g3 c8c1 d2c1');
INSERT INTO evaluations VALUES(55419,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(55419,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(55419,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(55419,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(55419,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(55419,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(55419,36,'cp 84','d2d4 g8f6 f1d3 b8c6 c2c3 a7a6 d4d5 c6e7 c3c4 e7g6 b1c3 c8d7 a2a4 f8e7 a4a5 e8g8 e1g1 f6h7 g1h2 h7g5 b2b4 g5f3 d1f3 e7g5 g2g3 g8h8 c1g5 d8g5 c4c5 g5e7 f3e3 f7f5 e4f5 d7f5 d3f5 f8f5 a1a3 f5f7 c3e4');
INSERT INTO evaluations VALUES(55411,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(55411,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(55411,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(55411,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(55411,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(55411,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(55411,36,'cp 67','g8f6 f1d3 b8c6 c2c3 g7g6 d4d5 c6e7 c3c4 f8g7 b1c3 e8g8 a2a4 f6h5 e1g1 a7a5 c4c5 d6c5 f1e1 c7c6 d3f1 c6d5 e4d5 f8e8 d5d6 e7c6 c3b5 h5f6 b5c7 e5e4 f3d2 c8f5 d2c4 c6d4 c1e3 f6d7 e3f4 d8h4 f4g3 h4g5 c7e8');
INSERT INTO evaluations VALUES(5205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5205,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(5205,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(5205,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(5205,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(5205,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5205,36,'cp 0','b1c3 g8f6 f1c4 e8g8 a2a3 d7d6 d1d3 c8e6 c1g5 a8c8 g5f6 d8f6 c3d5 f6d8 b2b4 c5b6 d5b6 d8b6 c4e6 f7e6 c2c3 c6e7 a1d1 e7g6 e1g1 d6d5 e4d5 g6f4 d3e4 e6d5 d1d5 f4d5 e4d5 g8h8 f3e5 c8c3 e5f7 h8g8 f7d6 g8h8');
INSERT INTO evaluations VALUES(5204,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(5204,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(5204,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(5204,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5204,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5204,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5204,36,'cp 0','d7d6 f1c4 g8f6 d1d3 e8g8 a2a3 c8e6 c1g5 c6d4 e1c1 h7h6 g5h4 e6c4 d3c4 g7g5 h4g3 a8c8 c4d3 d4e6 c1b1 d8b6 h2h4 g5g4 f3g5 e6g7 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(20002,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(20002,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20002,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20002,36,'cp 0','g7g6 g1e2 f8g7 d2d3 a8b8 e1g1 b7b5 f2f4 b5b4 c3d5 e7e6 d5e3 a7a5 a2a3 g8e7 f4f5 e6f5 e4f5 g6f5 e2f4 e8g8 f4h5 g7d4 d1d2 g8h8 c2c3 b4c3 b2c3 d4e5 e3c4 f8g8 d2h6 g8g6 h6d2 g6g8');
INSERT INTO evaluations VALUES(19668,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19668,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19668,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19668,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19668,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19668,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19668,36,'cp 0','d2d3 f8g7 g1e2 e7e6 c1e3 c6d4 e3d2 d4c6');
INSERT INTO evaluations VALUES(19672,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19672,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19672,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19672,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19672,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19672,36,'cp 0','f8g7 g1e2 e7e6 c1e3 c6d4 e3d2 d4c6');
INSERT INTO evaluations VALUES(5503,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5503,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5503,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5503,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5503,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5503,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5503,36,'cp 0','g1e2 e7e6 c1e3 c6d4 e3d2 d4c6');
INSERT INTO evaluations VALUES(5502,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5502,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5502,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5502,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5502,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5502,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5502,36,'cp 0','e7e6 e1g1 g8e7 c1g5 b7b6 d1d2 h7h6 g5e3 c8a6 a2a4 a8c8 f1e1 c6d4 c3b5 a6b5 a4b5 d8d7 e3d4 c5d4 c2c4 d4c3 e2c3 e8g8 g2h3 d6d5 e4e5 f8d8 d3d4 c8c7 h3f1 e7f5 f1h3');
INSERT INTO evaluations VALUES(50592,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(50592,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(50592,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(50592,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(50592,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(50592,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(50592,36,'cp -47','e7e6 c2c3 b7b5 c4e2 c8b7 d2d3 d7d5 e4d5 b7d5 g1f3 g8f6 a2a4 b8d7 c1f4 f8e7 h2h3 e8g8 b1d2 d8b6 e1g1 f8d8 f1e1 h7h6 e2f1 a8c8 a4b5 a6b5 c3c4 b5c4 d3c4 d5c6 d1c2 b6b7 f4h2 c8a8 a1a8 d8a8 f3e5 d7e5 h2e5');
INSERT INTO evaluations VALUES(50495,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(50495,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(50495,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(50495,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(50495,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(50495,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(50495,36,'cp -40','c2c3 b7b5 c4e2 c8b7 d2d3 d7d5 e4d5 b7d5 g1f3 b8d7 e1g1 g8f6 c1f4 f8e7 f1e1 d8b6 b1d2 e8g8 a2a4 f8d8 h2h3 a8c8 a4b5 a6b5 f3e5 d7e5 f4e5 f6d7 e5g3 h7h6 e2f3 d7f8 g3f4 f8g6 f4g3 c5c4 f3d5 d8d5 d3c4');
INSERT INTO evaluations VALUES(16427,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(16427,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(16427,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(16427,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(16427,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(16427,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(16427,36,'cp 58','d4d5 c6b8 g1f3 f8c5 f3e5 d8e7 e5d3 e7e4 d1e2 e4e2 e1e2 c5d4 c2c3 d4b6 a2a4 a7a5 c1g5 d7d6 g5f6 g7f6 h1e1 e8f8 b1d2 b8d7 e2f1 d7e5 c4b5 c8f5 d3f4 f8g7 h2h3 h7h5 h3h4 b6c5 d2b3 c5b6 b3d4 b6d4 c3d4 e5g6');
INSERT INTO evaluations VALUES(16334,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(16334,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(16334,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(16334,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(16334,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(16334,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(16334,36,'cp 56','c6b8 g1f3 f8c5 f3e5 d8e7 e5d3 e7e4 d1e2 e4e2 e1e2 c5d4 c2c3 d4b6 a2a4 a7a5 c1g5 f6e4 g5e3 b6e3 e2e3 e4d6 b1a3 e8d8 h1e1 h8e8 e3d2 d6c4 a3c4 d7d6 e1e8 d8e8 b2b4 a5b4 c3b4 b8d7 c4e3 g7g6 d2c3 b7b6 h2h4 c8b7 c3d4 c7c5 d5c6 b7c6 d3b2 d7f8 b2c4 f8e6 d4c3');
INSERT INTO evaluations VALUES(18779,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(18779,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18779,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(18779,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(18779,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(18779,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18779,36,'cp 10','g7g6 g1f3 f8g7 d2d4 c5d4 c3b5 d7d5 c4d5 e6d5 e1g1 g8f6 f3d4 e8g8 c1f4 a7a6 b5c3 f8e8 a1c1 c6d4 d1d4 f6h5 d4d5 h5f4 g3f4 c8e6 d5d8 e8d8 f1d1 d8d1 c1d1 g7c3 b2c3 a8b8 g2d5 e6d5 d1d5 b8c8 d5d3 c8c4 e2e3 g8g7 d3d7 c4c3 d7b7 c3a3 b7c7 a3a2');
INSERT INTO evaluations VALUES(35952,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(35952,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(35952,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(35952,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(35952,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(35952,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(35952,36,'cp 74','e1g1 c5d4 f3d4 b8c6 c1e3 c6d4 e3d4 c8d7 d1c2 a8c8 f1d1 d8a5 c2d2 d7c6 c3d5 a5d8 d5f6 g7f6 d4f6 e7f6 f2f3 d8b6 d2d4 b6c5 g1f1 c5e5 b2b4 c6a4 d4e5 d6e5 d1d5 b7b6 a1c1 f8d8 c4c5 d8d7 e2a6 c8c7 f1e2 b6c5 c1c5 c7c5 b4c5 g8f8 e2d2 f8e7 d2c3 a4c6 d5d7 e7d7 a6c4');
INSERT INTO evaluations VALUES(35951,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(35951,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(35951,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(35951,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(35951,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(35951,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(35951,36,'cp 76','c5d4 f3d4 b8c6 c1e3 c6d4 e3d4 c8d7 d1c2 d8c7 f1e1 a7a6 c2d3 d7c6 c3d5 f6d5 e4d5 g7d4 d3d4 c6d7 e2f1 a8e8 h2h3 a6a5 b2b3 c7c5 d4d2 e7e5 d5e6 e8e6 e1e6 d7e6');
INSERT INTO evaluations VALUES(36005,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(36005,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(36005,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(36005,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(36005,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(36005,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(36005,36,'cp 82','f3d4 b8c6 c1e3 c6d4 e3d4 c8d7 d1c2 d8a5 b2b3 d7c6 a2a3 a7a6 b3b4 a5c7 f1d1 f8c8 a3a4 b7b6 c2d2 a6a5 b4b5 c6b7 a1c1 f6d7 d4g7 g8g7 c3d5 c7d8 e2g4 b7d5 g4d7 d8d7 c4d5 c8c5 c1c5 d6c5 d2e3 a8d8 h2h3 f7f6 e3g3 d7b7 h3h4 h7h5');
INSERT INTO evaluations VALUES(36002,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(36002,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(36002,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(36002,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(36002,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(36002,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(36002,36,'cp 80','b8c6 c1e3 a7a6 d1d2 f6g4 e2g4 c8g4 b2b3 d8a5 a1b1 g4d7 a2a4 a8c8 h2h3 c6d4 e3d4 g7d4 d2d4 a5g5 d4d3 d7c6 f1e1 f7f5 f2f3 f5e4 c3e4 g5f4 e4c3 f4g5 d3e3 g5e3 e1e3 f8e8 b1d1 h7h5 a4a5 g8f7 b3b4 h5h4 g1f2 c8c7');
INSERT INTO evaluations VALUES(3239,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(3239,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(3239,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(3239,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(3239,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(3239,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(3239,36,'cp 77','c1e3 a7a6 d1d2 f6g4 e2g4 c8g4 b2b3 d8a5 d4c6 b7c6 e3d4 g4e6 a1d1 f7f6 d2d3 c6c5 d4e3 f8f7 e3d2 a8f8 d3c2 f6f5 c3d5 a5d8 d2c3 f5e4 c3g7 g8g7 c2e4 e6d5 d1d5 d8a5 e4c2 a5b4 c2b2 g7g8 h2h3 b4b7 d5d3 f7f6');
INSERT INTO evaluations VALUES(3234,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(3234,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(3234,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(3234,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(3234,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(3234,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(3234,36,'cp 72','a7a6 d1d2 f6g4 e2g4 c8g4 b2b3 d8a5 a1b1 a8c8 f1c1 g4e6 a2a4 g7d4 e3d4 c6d4 d2d4 a5g5 c1e1 e6h3 g2g3 g5c5 d4d2 c5b4 d2e3 h3e6 e3d3 c8c5 h2h3 h7h5 e1e3 f8e8 g1g2 b7b6 h3h4 g8g7 f2f3 e8c8 c3d5 e6d5 e4d5 c8c7 b1e1');
INSERT INTO evaluations VALUES(2546,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(2546,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(2546,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(2546,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(2546,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(2546,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(2546,36,'cp 77','d1d2 f6g4 e2g4 c8g4 b2b3 d8a5 d4c6 b7c6 a1c1 g4e6 e3d4 g7d4 d2d4 f7f6 d4e3 e6f7 h2h3 f8e8 c3a4 g8g7 f1d1 a8d8 a4c3 h7h5 d1d3 g6g5 c3e2 a5e5 e2d4 f7g6 d4c6 e5e4 e3e4 g6e4 c6d8 e4d3 d8e6 g7g6 e6c7');
INSERT INTO evaluations VALUES(2544,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(2544,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(2544,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(2544,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(2544,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(2544,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(2544,36,'cp 61','f6g4 e2g4 c8g4 b2b3 d8a5 a1b1 a8c8 h2h3 g4d7 d4c6 b7c6 e3d4 d7e6 b1c1 g7d4 d2d4 f7f6 c3a4 e6f7 f2f4 c8d8 d4e3 e7e5 f4f5 d6d5 c4d5 c6d5 c1c5 d5d4 e3d4 e5d4 c5a5 d4d3 a5a6 d3d2 f1d1 f8e8 f5g6 h7g6');
INSERT INTO evaluations VALUES(56206,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(56206,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(56206,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(56206,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(56206,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(56206,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(56206,36,'cp -34','g8f6 f1d3 f8d6 g1e2 e8g8 e1g1 f8e8 e2g3 b8d7 b1c3 a7a6 f1e1 c7c5 d3c2 b7b5 d4c5 d7c5 c3d5 f6d5 d1d5 c8b7 d5d1 g7g6 e3e4 a8c8 c1e3 d6e5 d1d8 e8d8 a1b1 h7h5 f2f3 c5d3 c2d3 d8d3 g3f1 f7f5 b1d1 f5e4');
INSERT INTO evaluations VALUES(262,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(262,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(262,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(262,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(262,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(262,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(262,36,'cp 88','c2c3 g8e7 f1e1 a7a6 d2d4 a6b5 d4c5 e8g8 b1a3 c6a7 c1e3 f6e6 d1e2 h7h6 a3b5 a7b5 e2b5 d7d5 c5d6 c7d6 b5b3 f7f5 e4f5 e7f5 f3d2 a8a6 b3e6 c8e6 a2a3 f8a8 f2f3 b7b5 a1d1 e6d5 e3f2 g8f7 h2h3 a6a5 g2g4 f5e7 h3h4 a8a6 h4h5');
INSERT INTO evaluations VALUES(258,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(258,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(258,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(258,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(258,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(258,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(258,36,'cp 87','g8e7 f1e1 a7a6 d2d4 a6b5 d4c5 e8g8 b1a3 c6a7 d1d3 h7h6 a3b5 a7b5 d3b5 d7d6 c5d6 c7d6 a2a4 f6e6 a4a5 e6g6 b2b4 c8h3 f3h4 g6g4 b5d3 g4h4 d3h3 h4h3 g2h3 f8c8 c1d2 f7f5 e4f5 e7f5 b4b5 f5h4 g1f1 c8c5 e1b1 h4f3 f1e2 f3d2 e2d2');
INSERT INTO evaluations VALUES(565,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(565,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(565,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(565,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(565,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(565,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(565,36,'cp 102','f1e1 a7a6 d2d4 a6b5 d4c5 e8g8 b1a3 c6a7 d1e2 h7h6 c1e3 f8d8 a3b5 a7b5 e2b5 d7d6 c5d6 c7d6 c3c4 f6g6 e3b6 c8d7 b5b3 d8c8 a2a4 d7c6 a4a5 c6e4 f3h4 g6g4 f2f3 g4h4 e1e4 h4h5 b6f2 c8c7 b3d3 h5g6 f2b6 c7c8 b2b4');
INSERT INTO evaluations VALUES(18071,30,'cp 319',NULL);
INSERT INTO evaluations VALUES(18071,31,'cp 327',NULL);
INSERT INTO evaluations VALUES(18071,32,'cp 319',NULL);
INSERT INTO evaluations VALUES(18071,33,'cp 320',NULL);
INSERT INTO evaluations VALUES(18071,34,'cp 279',NULL);
INSERT INTO evaluations VALUES(18071,35,'cp 281',NULL);
INSERT INTO evaluations VALUES(18071,36,'cp 277','f3d4 e5d4 b3f7 e8f7 d1h5 g7g6 h5d5 f7g7 d5a8 c7c5 d2d3 d8c7 a2a4 b5b4 b1d2 c8b7 a8a7 g8e7 d2b3 e7c8 c1f4 d7d6 a7c5 c7c5 b3c5 d6c5 f4e5 g7g8 e5h8 g8h8 b2b3 f8d6 e1c1 b7c6 h1f1 c6d7 f2f4 h8g7 f1f2 h7h5 h2h3 g7f7 d1g1 c8e7 g2g4 h5h4 g1f1 a6a5');
INSERT INTO evaluations VALUES(18070,30,'cp 312',NULL);
INSERT INTO evaluations VALUES(18070,31,'cp 296',NULL);
INSERT INTO evaluations VALUES(18070,32,'cp 273',NULL);
INSERT INTO evaluations VALUES(18070,33,'cp 280',NULL);
INSERT INTO evaluations VALUES(18070,34,'cp 309',NULL);
INSERT INTO evaluations VALUES(18070,35,'cp 288',NULL);
INSERT INTO evaluations VALUES(18070,36,'cp 324','e5d4 b3f7 e8f7 d1h5 g7g6 h5d5 f7g7 d5a8 c7c5 a8b8 d7d5 d2d3 d5e4 c1f4 e4d3 c2d3 f8e7 e1g1 g8f6 f1e1 h8f8 b8e5 f8f7 b1d2 c8f5 d2e4 e7f8 a2a4 f5e4 d3e4 b5b4 a1c1 d8e8 e5e8 f6e8 f4e5 g7g8 g1f1 e8f6 e5f6 f7f6 a4a5 g8f7 e4e5 f6f5 e5e6 f7e8');
INSERT INTO evaluations VALUES(18080,30,'cp 343',NULL);
INSERT INTO evaluations VALUES(18080,31,'cp 356',NULL);
INSERT INTO evaluations VALUES(18080,32,'cp 362',NULL);
INSERT INTO evaluations VALUES(18080,33,'cp 351',NULL);
INSERT INTO evaluations VALUES(18080,34,'cp 320',NULL);
INSERT INTO evaluations VALUES(18080,35,'cp 322',NULL);
INSERT INTO evaluations VALUES(18080,36,'cp 335','b3f7 e8f7 d1h5 g7g6 h5d5 f7g7 d5a8 c7c5 a8b8 d7d6 a2a4 b5b4 d2d3 g8e7 c1g5 d8d7 e4e5 e7d5 e1g1 h7h6 g5f6 d5f6 e5f6 g7f6 b1d2 d7b7 b8b7 c8b7 f1e1 b7c6 f2f4 h6h5 e1e2 f6f5 a4a5 h5h4 a1f1 h8h7 d2e4 b4b3 c2b3 h7f7 g2g4 f5g4 e2g2 g4h5 g2g5 h5h6 f4f5 g6f5 f1f5 c6e4');
INSERT INTO evaluations VALUES(12550,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(12550,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(12550,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(12550,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(12550,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(12550,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(12550,36,'cp 59','d2d3 f8e7 a2a3 d7d6 c1g5 f6g8 g5d2 g8f6 g1f3 h7h6 e1g1 a7a5 f3e1 c8e6 c3d5 e8g8 a3a4 e6d7 d5c3 f6e8 e1c2 f7f5 e2e3 e8f6 b2b3 a8b8 c3b5 c6b4 e3e4 g8h8 d2c3 d7c6');
INSERT INTO evaluations VALUES(12553,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(12553,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(12553,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(12553,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(12553,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(12553,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(12553,36,'cp 42','f8e7 c1g5 f6g8 g5d2 d7d6 a2a3 g8f6 a1b1 a7a5 d2g5 f6g8 g5c1 c8e6 h2h3 g8f6 e2e4 c6d4 b2b3 e8g8 g1e2 d8d7 e2d4 c5d4 c3d5 d7d8 e1g1 e6d5 c4d5 f6d7 g1h2 b7b5 f2f4 d7c5 f4e5 d6e5 a3a4');
INSERT INTO evaluations VALUES(23471,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(23471,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(23471,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(23471,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(23471,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23471,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(23471,36,'cp 19','e7e5 d2d4 e5d4 e3d4 d7d5 b1c3 g8f6 f1b5 f8b4 e1g1 e8g8 c1g5 b4c3 b2c3 d8d6 d1c1 f6e4 g5f4 d6f6 b5d3 c8g4 f3e5 g4f5 f1e1 f8e8 c1b2 e4d6 f4g3 c6a5 b2c1 f5d3 c2d3 f6f5 c1d2 f7f6 e5f3 e8e1 a1e1');
INSERT INTO evaluations VALUES(20479,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(20479,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(20479,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(20479,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(20479,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(20479,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(20479,36,'cp -52','f8d6 e1g1 g8e7 d3d4 c5d4 f3d4 e8g8 b1c3 a7a6 b5a4 d8c7 h2h3 c6d4 d1d4 c8e6 c3e2 d6h2 g1h1 h2e5 d4d1 e7f5 c2c3 a8d8 e2d4 e5d4 c3d4 c7b6 f1e1 h7h6 g2g4 f5d6 h1g2 d6e4 f2f3 f7f5 f3e4 d5e4 c1e3 f5g4 h3g4');
INSERT INTO evaluations VALUES(36660,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(36660,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(36660,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(36660,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(36660,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(36660,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(36660,36,'cp 46','e7e5 d4e5 d6e5 f3e5 f6e4 c3e4 d8d1 f1d1 g7e5 c2c3 f8e8 c1g5 g8g7 d1e1 b8c6 e2f3 h7h6 g5e3 a7a5 a1d1 a5a4 e4c5 e8e7 a2a3 e5d6 c5e4 d6e5 g2g3 g6g5 f3g2 f7f5 e4c5 e5d6 c5d3 c8e6 d3b4 e6b3 b4c6');
INSERT INTO evaluations VALUES(55742,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(55742,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(55742,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(55742,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(55742,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(55742,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(55742,36,'cp 79','d2d4 b8d7 g1f3 g8f6 h2h3 f8e7 c1f4 c7c6 f1d3 d8c7 e4e5 d6e5 f3e5 c7b6 d1d2 d7e5 f4e5 e8g8 e1c1 f6d5 c3d5 c6d5 d2e2 e7d6 h3h4 c8d7 g2g4 a8c8 e5d6 b6d6 e2e3 a7a6 g4g5 d7b5');
INSERT INTO evaluations VALUES(19829,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19829,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19829,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(19829,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19829,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19829,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(19829,36,'cp 0','f1b5 c8g4 b5c6 b7c6 d2d3 e7e6 b1d2 g4h5 e1g1 g8h6 h2h3 f7f5 c2c3 a8c8 d1a4 g7g5 a4a7 f8e7 d3d4 g5g4 h3g4 f5g4 f3h2 h6f5 d2b3 g4g3 f2g3 f5g3 c1f4 g3e2 g1h1 e8g8 f4h6 e2g3 h1g1 g3e2');
INSERT INTO evaluations VALUES(19744,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19744,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19744,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19744,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19744,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19744,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19744,36,'cp 0','c8g4 b5c6 b7c6 d2d3 e7e6 b1d2 g8e7 h2h3 g4h5 e1g1 e7c8 a2a4 c8b6 d1e1 b6d7 b2b3 h5g6 c1a3 d8b8 a1c1 f8e7 a3b2 d7b6 b2a3');
INSERT INTO evaluations VALUES(26255,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26255,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26255,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26255,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26255,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26255,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26255,36,'cp 0','b5c6 b7c6 d2d3 e7e6 b1d2 g4h5 h2h3 g8e7 e1g1 e7c8 d1e1 c8b6 a2a4 f8e7 f3h2 h5g6 b2b3 e8g8 c1a3 b6d7 a1c1 d8c7 a3b2 a7a5 e1e2 f8b8 g1h1 c5c4 d3c4 d7b6 b2c3 e7a3 c1a1 a3e7');
INSERT INTO evaluations VALUES(55268,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55268,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55268,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55268,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55268,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55268,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55268,36,'cp 0','f1c4 e7e5 g1e2 d8h4 e2g3 f8d6 e1g1 g8e7 a2a4 c7c5 d2d3 e8g8 c1d2 b8c6 b1a3 d6c7 a3b5 c7b8 b5a3');
INSERT INTO evaluations VALUES(14265,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14265,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14265,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(14265,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(14265,36,'cp 0','c8f5 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(24851,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24851,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24851,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24851,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24851,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(24851,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24851,36,'cp 0','f3h4 f5e6 h4f3 e6f5');
INSERT INTO evaluations VALUES(30869,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(30869,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(30869,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(30869,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(30869,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(30869,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(30869,36,'cp -5','b8d7 e1g1 g8f6 d2d3 a8c8 c2c3 e7e6 c1f4 f8e7 b1d2 b7b5 h2h3 a7a5 a2a3 e8g8 f1e1 d8b6 d1e2 c8a8 f3e5 d7e5 f4e5 f8c8 e5d4 b6b7 b2b4 f6d7 e2g4 e7f6 d2f3 f6d4 g4d4 a5b4 a3b4 a8a1');
INSERT INTO evaluations VALUES(26217,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(26217,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(26217,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26217,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(26217,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(26217,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26217,36,'cp -10','d2d3 d8c7 e1g1 g8f6 b2b3 e7e5 c2c4 d5c4 b3c4 f8e7 f1e1 e8g8 b1c3 f8e8 a1b1 e7f8 c1g5 h7h6 g5h4 e5e4 h4g3 c7c6 f3d4 c6c8 d3e4 c8c4 b1c1 a8c8 e4e5 f6d5 c3d5 c4d5 c1c8 e8c8 e5e6 f7e6 d4e6 d5d1 e1d1 f8e7 h2h3 d7f6 g3d6 e7d6');
INSERT INTO evaluations VALUES(43835,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(43835,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(43835,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43835,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(43835,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(43835,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43835,36,'cp 37','d1f3 b8c6 b1c3 c8b7 e1c1 a7a6 d4d5 e6d5 c3d5 f6d5 f3d5 d7d6 g1e2 d8d7 h2h4 c6d8 d5b3 d8e6 f2f3 d7b5 b3c3 b5c6 c3c6 b7c6 h4h5 f8e7 e2d4 c6b7 d4e6 f7e6 e3e4 h8f8 f4e3 e8c8 f1d3 e7f6 c2c3 d6d5 e4d5 d8d5');
INSERT INTO evaluations VALUES(51503,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(51503,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(51503,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(51503,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(51503,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(51503,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(51503,36,'cp 29','c8b7 f1g2 g8f6 d2d3 c7c5 g1e2 d7d6 e1g1 g7g6 b1c3 b8c6 c3d5 f8g7 c2c3 e8g8 d5f6 g7f6 c1h6 f8e8 a2a4 c6e5 d3d4 e5d7 f2f4 e7e5 f4e5 d6e5 d4d5 c5c4 a4a5 b7a6 g1h1 b6a5 d1a4 d7c5 a4a5 f6e7 h6e3 d8a5 a1a5');
INSERT INTO evaluations VALUES(30827,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(30827,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(30827,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(30827,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(30827,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(30827,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(30827,36,'cp 41','f1g2 g8f6 d2d3 g7g6 g1e2 d7d6 e1g1 c7c5 b1c3 b8c6 c3d5 f8g7 c2c3 e7e6 d5f6 d8f6 c1e3 e8g8 d3d4 f6e7 f1e1 a7a5 a2a4 b7a6 b2b3 a8c8 h2h4 h7h5 a1b1 c5d4 e2d4 c6d4 e3d4 g7d4 d1d4 e7c7 c3c4 a6b7 e1d1 f8d8');
INSERT INTO evaluations VALUES(30828,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(30828,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30828,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(30828,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30828,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(30828,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(30828,36,'cp 25','g8f6 d2d3 d7d6 g1e2 c7c5 e1g1 g7g6 b1c3 b8c6 c3d5 f8g7 c2c3 e7e6 d5f6 d8f6 c1e3 e8g8 d3d4 a8e8 d1d2 f6e7 a1e1 c5d4 c3d4 b7a6 e1d1 a6e2 d2e2 e6e5 d4e5 d6e5 d1d2 e8d8 f1c1 d8d2 e2d2 c6d4 b2b4');
INSERT INTO evaluations VALUES(28826,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(28826,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(28826,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(28826,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(28826,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(28826,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(28826,36,'cp 34','d2d3 d7d6 g1e2 c7c5 e1g1 g7g6 b1c3 b8c6 c3d5 f8g7 d5f6 g7f6 c2c3 f6g7 d3d4 e8g8 c1e3 a7a5 d1d2 b7a6 f1e1 a8c8 a1d1 c5d4 e2d4 c6e5 b2b3 e5d3 e1e2 d3c5 c3c4 a6b7 e3g5 f7f6 g5e3 b7e4 g2e4');
INSERT INTO evaluations VALUES(53920,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(53920,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(53920,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(53920,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(53920,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(53920,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(53920,36,'cp 99','g1f3 g8f6 b1c3 g7g6 c1f4 a7a6 a2a4 f8g7 h2h3 e8g8 e2e3 f6e8 f1d3 b8d7 e3e4 d7e5 e1g1 e5f3 d1f3 f7f5 f3g3 d8b6 f1b1 c8d7 d3c2 b6b4 a1a3 c5c4 a4a5 b7b5 a5b6 b4b6 a3a1 f5e4 c3e4 g7d4 g1h1 e8f6 f4d6 f6e4 c2e4 f8f2 e4f3');
INSERT INTO evaluations VALUES(52904,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(52904,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(52904,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(52904,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(52904,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(52904,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(52904,36,'cp 44','f3e4 e7e5 g1f3 c8e6 c2c3 b8d7 f1d3 f8d6 e1g1 g8f6 g1h1 h7h6 c1e3 e8g8 b1d2 f8e8 e3g1 d6c7 d1c2 d8e7 a1e1 f6h5 a2a4 h5f4 d3c4 a8d8 g1f2 e5d4 c3d4 d7f6 b2b3 f6g4 f2g3 e6c4 d2c4 f4e6 g3h4 e6d4 h4e7 d4c2');
INSERT INTO evaluations VALUES(52898,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(52898,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(52898,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(52898,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(52898,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(52898,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(52898,36,'cp 52','e7e5 g1f3 c8e6 c1g5 f8e7 g5e7 d8e7 b1c3 e5d4 d1d4 g8f6 e1c1 e8g8 e4e5 f6d5 c3e4 b8d7 f1d3 f8d8 e4g5 d7f8 a2a3 h7h6 g5e6 f8e6 d4e4 g7g6 d3c4 e6g5 f3g5 e7g5 c1b1 d5e3 d1d8 a8d8 e5e6 f7e6 h1e1 e3c4 e4e6 g8h8');
INSERT INTO evaluations VALUES(53785,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(53785,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(53785,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(53785,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(53785,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(53785,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(53785,36,'cp 51','g1f3 c8e6 c2c3 b8d7 f1d3 f8d6 e1g1 g8f6 g1h1 h7h6 c1e3 e8g8 b1d2 f8e8 e3g1 d6c7 d1c2 d8e7 a1e1 f6h5 d3c4 h5f4 a2a4 a8d8 c4e6 e7e6 b2b4 f4g6 d2b3 e6c4 f3d2 c4e6 b4b5 e6e7 d2f3 d7f6 b5c6 b7c6 a4a5 e5d4');
INSERT INTO evaluations VALUES(53784,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(53784,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(53784,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(53784,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(53784,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(53784,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(53784,36,'cp 39','c8g4 c2c3 g8f6 f1c4 g4h5 e1g1 b8d7 d1e1 d8c7 c1g5 h7h6 g5f6 g7f6 c4b3 h5g6 b1d2 e8c8 g1h1 c8b8 a1d1 d7b6 e1f2 e5d4 c3d4 f8d6 f3h4 g6h5 d1c1 h8g8 g2g3 d6b4 d2c4 g8e8 a2a3 b6c4 b3c4 b4d6');
INSERT INTO evaluations VALUES(15491,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(15491,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15491,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(15491,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(15491,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(15491,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(15491,36,'cp 24','f8e7 a2a4 e8g8 e1g1 a7a5 b1a3 d6d5 e4d5 f6d5 h2h3 e7a3 a1a3 c8f5 f1e1 d8d6 d3d4 e5d4 f3d4 c6d4 c3d4 c7c6 e1e5 d6d7 c4d5 c6d5 a3e3 f8e8 e5e8 a8e8 e3e8 d7e8 c1d2 b7b6 d1b3 e8e6 b3g3 f7f6 g3c7 f5g6 d2f4 g6h5 c7b8 h5e8');
INSERT INTO evaluations VALUES(8593,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(8593,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(8593,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(8593,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8593,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(8593,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(8593,36,'cp 22','e1g1 e8g8 a2a4 a7a5 d1b3 c8g4 b1d2 f6d7 h2h3 d7c5 b3c2 g4e6 f1e1 h7h6 d3d4 e5d4 f3d4 e7g5 d4e6 c5e6 d2f3 g5c1 a1c1 d8f6 c4e6 f7e6 e4e5 c6e5 f3e5 d6e5 e1e3 f8f7 c1e1 f6f5 c2f5 f7f5 e3e5 f5e5 e1e5');
INSERT INTO evaluations VALUES(8592,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(8592,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(8592,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(8592,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(8592,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(8592,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(8592,36,'cp 17','e8g8 a2a4 a7a5 b1d2 c8e6 f1e1 e6c4 d2c4 h7h6 h2h3 f8e8 d1b3 d8d7 c1d2 e7f8 c4e3 c6d8 b3c2 c7c6 e3f1 d8e6 f1g3 d7c7 d3d4 b7b5 b2b3 e8c8 e1c1 f6d7 c1d1 d7f6 a1a2 f6d7 a4b5 c6b5 d1a1 d7f6 d4d5');
INSERT INTO evaluations VALUES(3384,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(3384,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(3384,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(3384,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(3384,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(3384,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3384,36,'cp 35','a2a4 a7a5 f1e1 c8e6 c4e6 f7e6 b1a3 d8e8 d1b3 e8d7 a3b5 h7h6 d3d4 e5d4 f3d4 c6d8 e4e5 d6e5 e1e5 f6g4 e5e2 c7c6 h2h3 g4f2 e2f2 c6b5 f2f8 e7f8 a4b5 d7f7 c1e3 e6e5 b3f7 d8f7 d4e6 f8d6 a1d1 a5a4 e6c5 d6c5 e3c5');
INSERT INTO evaluations VALUES(19304,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19304,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19304,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19304,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(19304,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(19304,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19304,36,'cp 0','f8g7 e4e5 g8h6 e1g1 e8g8 h2h3 d7d6 e5d6 d8d6 d1d6 e7d6 c1f4 f8e8 f1e1 c8e6 c4e6 e8e6 e1e6 f7e6 f4d6 h6f5 d6a3 a8c8 f3g5 c6d4 a1d1 d4c2 a3d6 c2d4');
INSERT INTO evaluations VALUES(24916,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(24916,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24916,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(24916,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24916,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(24916,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24916,36,'cp 24','e7e6 b1d2 f8d6 f4g5 h7h6 g5h4 c6e7 f1b5 c7c6 b5d3 e7g6 d3g6 f7g6 h4g3 g6g5 g3d6 d8d6 h2h3 g4h5 g2g4 h5g6 f3e5 g6h7 h3h4 g5h4 h1h4 f6d7 e5d7 d6d7 f2f4 d7e7 h4h1 b7b5 d1e2 b5b4 c3c4 c6c5 d4c5 e7c5');
INSERT INTO evaluations VALUES(25208,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(25208,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(25208,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(25208,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(25208,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(25208,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(25208,36,'cp 27','b1d2 f8d6 f4g5 h7h6 g5h4 a7a6 f1e2 c6e7 e1g1 c7c6 e2d3 e8g8 d1b3 a8b8 b3c2 b8c8 h2h3 g4h5 b2b4 a6a5 b4a5 d8a5 f3e5 a5c7 g2g4 h5g6 d3g6 f7g6 e5g6 e7g6 c2g6 e6e5 d4e5 d6e5 g6c2 f6d7 f2f4 e5f6 d2f3');
INSERT INTO evaluations VALUES(25206,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25206,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(25206,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(25206,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(25206,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(25206,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(25206,36,'cp 24','f8d6 f4g5 h7h6 g5h4 a7a6 f1e2 c6e7 e1g1 c7c6 e2d3 e8g8 d1b3 a8b8 h4g3 e7f5 g3e5 g4f3 d2f3 d6e5 f3e5 f5d6 c3c4 d6c4 d3c4 d5c4 b3c4 f8e8 e5d3 d8e7 a1c1 f6d7 f1d1 e6e5 d4e5 d7e5 d3e5 e7e5 c4c3 e5c3 c1c3 f7f6 d1d7 a6a5');
INSERT INTO evaluations VALUES(24115,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24115,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(24115,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(24115,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24115,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24115,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(24115,36,'cp 19','f4g5 h7h6 g5h4 a7a6 f1e2 c6e7 e1g1 c7c6 e2d3 e8g8 d1b3 a8b8 h4g3 e7f5 g3e5 g4f3 d2f3 d6e5 f3e5 f6d7 e5d7 d8d7 c3c4 d5c4 b3c4 d7e7 f1c1 b8d8 c4c5 e7f6 c5c2 g7g6 g2g3 g8g7 a2a3 f5d6 h2h4 a6a5');
INSERT INTO evaluations VALUES(32414,30,'cp 338',NULL);
INSERT INTO evaluations VALUES(32414,31,'cp 353',NULL);
INSERT INTO evaluations VALUES(32414,32,'cp 342',NULL);
INSERT INTO evaluations VALUES(32414,33,'cp 330',NULL);
INSERT INTO evaluations VALUES(32414,34,'cp 357',NULL);
INSERT INTO evaluations VALUES(32414,35,'cp 323',NULL);
INSERT INTO evaluations VALUES(32414,36,'cp 346','d1d8 e8d8 f3e5 g4e6 b1c3 g8f6 f2f4 f6d7 e5d7 e6d7 c1e3 f8b4 e1c1 b4c3 b2c3 h8e8 f1c4 e8e4 c4d5 e4e3 d5b7 e3c3 b7a8 c7c6 h1e1 c3a3 c1b2 a3a4 g2g3 a4b4 b2a1 b4b6 d1d3 b8a6 e1d1 a6b8 g3g4 d8c7 f4f5');
INSERT INTO evaluations VALUES(27720,30,'cp 325',NULL);
INSERT INTO evaluations VALUES(27720,31,'cp 332',NULL);
INSERT INTO evaluations VALUES(27720,32,'cp 331',NULL);
INSERT INTO evaluations VALUES(27720,33,'cp 306',NULL);
INSERT INTO evaluations VALUES(27720,34,'cp 313',NULL);
INSERT INTO evaluations VALUES(27720,35,'cp 314',NULL);
INSERT INTO evaluations VALUES(27720,36,'cp 323','e8d8 f3e5 g4e6 b1c3 d8c8 c1d2 b8d7 e5d7 e6d7 e1c1 a7a6 d2e3 b7b5 g2g3 d7c6 f1g2 c8b7 c3e2 a8e8 f2f3 h7h5 h2h4 g8f6 e2d4 c6d5 a2a3 f8c5 h1e1 d5c4 e3g1 g7g6 e4e5 f6d5 f3f4 b7c8 c2c3 c5d4 g1d4 h8g8 g2h3 c8b8 d4f2 c7c6 e5e6');
INSERT INTO evaluations VALUES(27734,30,'cp 322',NULL);
INSERT INTO evaluations VALUES(27734,31,'cp 312',NULL);
INSERT INTO evaluations VALUES(27734,32,'cp 310',NULL);
INSERT INTO evaluations VALUES(27734,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(27734,34,'cp 341',NULL);
INSERT INTO evaluations VALUES(27734,35,'cp 315',NULL);
INSERT INTO evaluations VALUES(27734,36,'cp 324','f3e5 g4e6 b1c3 d8c8 c1d2 b8d7 e5d7 e6d7 e1c1 a7a6 f1e2 f8c5 h1e1 g8f6 f2f3 h8e8 e2c4 d7e6 c4e6 e8e6 e4e5 f6d7 c3e4 c5e7 f3f4 b7b5 d2a5 d7c5 e4c3 c8b7 c3d5 e7d8 a5b4 c5a4 d5c3 a4c3 b4c3 e6g6');
INSERT INTO evaluations VALUES(27731,30,'cp 317',NULL);
INSERT INTO evaluations VALUES(27731,31,'cp 316',NULL);
INSERT INTO evaluations VALUES(27731,32,'cp 316',NULL);
INSERT INTO evaluations VALUES(27731,33,'cp 313',NULL);
INSERT INTO evaluations VALUES(27731,34,'cp 321',NULL);
INSERT INTO evaluations VALUES(27731,35,'cp 320',NULL);
INSERT INTO evaluations VALUES(27731,36,'cp 329','g4e6 b1c3 d8c8 e5d3 c7c6 c1e3 b8d7 f2f3 c8c7 c3e2 f8d6 e1c1 a8e8 e2d4 g8f6 a2a4 h7h5 a4a5 h5h4 c2c3 e6c4 c1c2 g7g6 b2b3 c4a6 b3b4 c7b8 d3b2 a6f1 d1f1 d7e5 f1a1 d6c7 a5a6 h4h3 g2g3 b7b5 b2d3 f6d7');
INSERT INTO evaluations VALUES(27729,30,'cp 339',NULL);
INSERT INTO evaluations VALUES(27729,31,'cp 323',NULL);
INSERT INTO evaluations VALUES(27729,32,'cp 323',NULL);
INSERT INTO evaluations VALUES(27729,33,'cp 314',NULL);
INSERT INTO evaluations VALUES(27729,34,'cp 315',NULL);
INSERT INTO evaluations VALUES(27729,35,'cp 342',NULL);
INSERT INTO evaluations VALUES(27729,36,'cp 331','b1c3 d8c8 c1d2 b8d7 e5d7 e6d7 e1c1 a7a6 f1c4 d7e6 c4e6 f7e6 e4e5 g8e7 c3e4 e7c6 e4g5 c6e5 g5e6 f8d6 d2c3 e5g4 d1d2 h8g8 f2f3 g4f6 e6g7 g8g7 c3f6 g7f7 d2e2 c8d7 f6c3 a8g8 h1d1 d7c6 a2a3 b7b6');
INSERT INTO evaluations VALUES(27728,30,'cp 314',NULL);
INSERT INTO evaluations VALUES(27728,31,'cp 332',NULL);
INSERT INTO evaluations VALUES(27728,32,'cp 315',NULL);
INSERT INTO evaluations VALUES(27728,33,'cp 338',NULL);
INSERT INTO evaluations VALUES(27728,34,'cp 303',NULL);
INSERT INTO evaluations VALUES(27728,35,'cp 320',NULL);
INSERT INTO evaluations VALUES(27728,36,'cp 351','d8c8 c1d2 b8d7 e5d7 e6d7 e1c1 a7a5 d2e3 g8f6 f2f3 a5a4 a2a3 a8a5 h2h3 f6h5 c3e2 f8d6 e3d2 a5e5 g2g4 h5f6 d2f4 e5e6 f4d6 e6d6 d1d6 c7d6 e2c3 d7c6 f1b5 c6b5 c3b5 d6d5 e4e5 f6d7 f3f4 d7c5');
INSERT INTO evaluations VALUES(51428,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(51428,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(51428,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(51428,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(51428,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(51428,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(51428,36,'cp 96','d2d4 c8b7 f1d3 f8b4 g1e2 d7d5 e4e5 d8d7 e1g1 b4e7 c3b1 b7a6 d3a6 b8a6 c2c3 a6b8 b1d2 c7c5 d2f3 e7d8 c1e3 c5d4 c3d4 g8e7 h2h4 h7h6 a1c1 b8c6 h4h5 e7f5 e2f4 e8g8 g2g4 f5e3 f2e3 d8g5 d1e2 g5f4 e3f4');
INSERT INTO evaluations VALUES(53376,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(53376,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(53376,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(53376,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(53376,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(53376,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(53376,36,'cp 12','b8c6 g1f3 c8f5 c2c4 e7e6 b1c3 f8b4 c4d5 e6d5 a1c1 a8c8 a2a3 b4d6 f4e5 d6e5 f3e5 g8e7 f1b5 e8g8 b5c6 e7c6 e5c6 c8c6 e1g1 f8e8 d1d2 e8e6 f1e1 c6c4 e1e6 f5e6 h2h3 h7h6 c3e2 d8d6 f2f3 d6b6 c1c4');
INSERT INTO evaluations VALUES(19845,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19845,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19845,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(19845,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(19845,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19845,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19845,36,'cp 11','g1f3 c8f5 c2c4 e7e6 b1c3 d5c4 f1c4 f8d6 f4d6 d8d6 d4d5 e6d5 c3d5 g8e7 e1g1 e8g8 d1b3 c6a5 b3b5 a5c4 d5e7 d6e7 b5f5 a8d8 b2b3 c4d2 f3d2 d8d2 f5a5 e7d6 a5a7 b7b6 a2a4 f8d8 a4a5 b6a5 a7a5 g7g6 b3b4 d6f6 a1e1 d2d5');
INSERT INTO evaluations VALUES(19843,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(19843,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19843,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(19843,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(19843,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19843,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19843,36,'cp 12','c8f5 c2c4 e7e6 b1c3 d5c4 f1c4 f8d6 f4d6 d8d6 e1g1 g8e7 d4d5 e6d5 c3d5 e8g8 d1b3 a8e8 a1d1 c6a5 d5e7 d6e7 b3b5 a5c4 b5f5 g7g6 f5b5 c4e5 f3e5 e7e5 b5b7 e5a5 b7d5 a5d5 d1d5 e8e2 d5b5 f8c8 a2a3 c8d8 b5b7 a7a5 b2b3 e2c2 b7b5 d8d3 b5a5');
INSERT INTO evaluations VALUES(19245,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(19245,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19245,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(19245,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(19245,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(19245,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19245,36,'cp 0','c5d4 f3d4 f8g7 d4c6 b7c6 e1g1 d7d6 b1c3 g8f6 e4e5 d6e5 d1d8 e8d8 f1e1 c8f5 e1e5 f6d5 e5e2 e7e6 c3a4 d8e7 c4b3 h8b8 c1d2 h7h6 h2h4 h6h5 a1d1 g7f6 d2g5 f6g5 h4g5 d5f4 e2e1 a7a5 f2f3');
INSERT INTO evaluations VALUES(19857,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19857,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(19857,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19857,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19857,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19857,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19857,36,'cp -5','c8g4 c2c3 d8c7 b1d2 g4h5 d2f1 e7e6 d1e2 h5g6 f1g3 g6d3 e2d3 f8d6 g3h5 g7g6 h5g3 e8f8 e1g1 f8g7 f1e1 a8c8 c1d2 g8e7 d3e2 h7h5 a2a4 c8e8 g3f1 e7f5 f3e5 c6e5 d4e5 d6c5 d2g5 c5b6');
INSERT INTO evaluations VALUES(26301,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26301,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26301,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(26301,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26301,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26301,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26301,36,'cp 0','c2c3 d8c7 b1d2 e7e6 d2f1 g4h5 f1g3 h5g6 e1g1 f8d6 f1e1 g8e7 d3g6 h7g6 d1e2 e8g8 f3g5 c6b8 e2g4 b8d7 g4h3 d7f6 e1e6 f7e6 h3e6 g8h8 e6h3 h8g8');
INSERT INTO evaluations VALUES(26300,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26300,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26300,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26300,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26300,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26300,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26300,36,'cp 0','d8c7 b1d2 e7e6 d2f1 g4h5 f1g3 h5g6 e1g1 f8d6 f1e1 g8e7 d3g6 h7g6 d1e2 e8g8 c1d2 a8e8 f3g5 c6b8 e2g4 b8d7 e1e6 d7f6 g4h3 f7e6 h3e6 g8h8 e6h3 h8g8');
INSERT INTO evaluations VALUES(27681,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27681,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27681,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27681,36,'cp 0','b1d2 e7e6 d2f1 g4h5 f1g3 h5g6 e1g1 f8d6 f1e1 g8e7 d3g6 h7g6 d1e2 e8g8 f3g5 c6b8 e2g4 b8d7 e1e6 d7f6 g4h3 f7e6 h3e6 g8h8 e6h3 h8g8');
INSERT INTO evaluations VALUES(20856,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(20856,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(20856,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(20856,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(20856,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(20856,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(20856,36,'cp 126','c2c4 e7e6 a2a3 f6f5 b2b4 g8f6 c1b2 a7a6 b1c3 f8e7 f1d3 e8g8 e1g1 d8e8 d1c2 g8h8 c3e2 c8d7 f3e5 c6d8 e5d7 e8d7 c4d5 f6d5 e2c1 a6a5 b4a5 a8a5 d3c4 e7d6 c1d3 d8f7 c4b3 b7b6 f1e1 f7g5 f2f3 f8a8 a3a4 d5b4');
INSERT INTO evaluations VALUES(20831,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(20831,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(20831,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(20831,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(20831,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(20831,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(20831,36,'cp 116','e7e6 b1c3 f6f5 f1d3 g8f6 e1g1 f8e7 a2a3 e8g8 b2b4 a7a6 c1b2 d8e8 d1c2 g8h8 c3e2 c8d7 e2f4 e7d6 c4c5 d6f4 e3f4 f6e4 c2b3 b7b5 c5b6 c7b6 a3a4 e8e7 b4b5 a6b5 a4b5 c6a5 b3a4 f8c8 f1c1 c8c1 a1c1 h7h6 f3e5');
INSERT INTO evaluations VALUES(23461,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(23461,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(23461,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23461,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(23461,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(23461,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(23461,36,'cp -8','e7e5 e2e3 d7d5 f1b5 f8d6 f2f4 f7f6 d1h5 g7g6 h5f3 a7a6 b5c6 b7c6 g1e2 c8e6 c2c4 g8h6 b1c3 e6g4 f3f2 d5c4 f2h4 g4e2 c3e2 h6f5 h4g4 d6b4 b2c3 b4c3 e2c3 d8d3 f4e5 e8c8 e1c1 c4b3 a2b3 f6e5 g4e4 d3e4 c3e4 c8b7 e4c5 b7b6 c5a4 b6b5 g2g4 f5d6 h2h3');
INSERT INTO evaluations VALUES(32844,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(32844,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(32844,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(32844,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(32844,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(32844,35,'cp 265',NULL);
INSERT INTO evaluations VALUES(32844,36,'cp 286','b1c3 d5d8 e2e4 f5g6 d1b3 e7e6 b3b7 b8d7 f1b5 g8f6 f2f3 a8b8 b7a7 f8d6 e4e5 d6b4 e5f6 b8b5 a7a4 b4c3 b2c3 g6d3 f6g7 h8g8 a4d1 d3c4 g1e2 d8h4 g2g3 h4f6 e1f2 g8g7 e2f4 b5a5 a2a4 d7b6 h1e1 g7g8 h2h4 e8d7 c1e3 d7c8 d1b1 h7h5 b1h7');
INSERT INTO evaluations VALUES(32842,30,'cp 249',NULL);
INSERT INTO evaluations VALUES(32842,31,'cp 253',NULL);
INSERT INTO evaluations VALUES(32842,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(32842,33,'cp 246',NULL);
INSERT INTO evaluations VALUES(32842,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(32842,35,'cp 261',NULL);
INSERT INTO evaluations VALUES(32842,36,'cp 266','d5d8 e2e4 f5g6 h2h4 g8f6 f1d3 e7e6 h4h5 f6h5 h1h5 g6h5 d1h5 d8d4 h5e2 b8c6 g1f3 d4b6 c1d2 c6b4 e1c1 a8d8 d3b5 c7c6 b5c4 f8e7 c1b1 b6c7 a2a3 b4a6 c4a6 b7a6 e2a6 e8g8 d1c1 c6c5 c3a4 d8b8 b1a2 b8c8 d2e3 f8d8 e3c5 e7c5 a4c5');
INSERT INTO evaluations VALUES(31973,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(31973,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(31973,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(31973,33,'cp 308',NULL);
INSERT INTO evaluations VALUES(31973,34,'cp 276',NULL);
INSERT INTO evaluations VALUES(31973,35,'cp 284',NULL);
INSERT INTO evaluations VALUES(31973,36,'cp 292','e2e4 f5g6 h2h4 g8f6 f2f3 c7c6 g1e2 h7h6 e2f4 g6h7 d1b3 e7e6 b3b7 b8d7 b7c6 f8e7 c6a6 e8g8 c1e3 f6e4 f3e4 e7h4 e1e2 a8b8 f4d3 h4g5 e3g5 d8g5 a6d6 f8d8 h1h3 g5g4 e2e1 d7f6 d6e5 f6e4 c3e4 g4e4 e5e4 h7e4 d3f2 d8d4 h3b3 b8c8 b3c3 c8b8 f2d1 e4d5 c3c7 b8d8 c7a7');
INSERT INTO evaluations VALUES(31975,30,'cp 290',NULL);
INSERT INTO evaluations VALUES(31975,31,'cp 309',NULL);
INSERT INTO evaluations VALUES(31975,32,'cp 302',NULL);
INSERT INTO evaluations VALUES(31975,33,'cp 291',NULL);
INSERT INTO evaluations VALUES(31975,34,'cp 293',NULL);
INSERT INTO evaluations VALUES(31975,35,'cp 293',NULL);
INSERT INTO evaluations VALUES(31975,36,'cp 302','f5g6 h2h4 g8f6 f2f3 c7c6 g1e2 h7h6 e2f4 g6h7 d1b3 e7e6 b3b7 b8d7 b7c6 f8e7 c1e3 e8g8 c6a6 f6e4 c3e4 h7e4 f3e4 e7h4 e1d1 h4g5 a6e2 d7f6 e4e5 g5f4 e3f4 d8d4 e2d2 a8d8 d2d4 d8d4 d1e1 d4f4 e5f6 f8c8 f1d3 f4f6 h1h4 c8c5 h4c4 c5e5 d3e2 e5g5 c4g4 g5e5 a1c1 g7g6 g4c4 e5g5 e2f3 g5a5 a2a4 f6f5 b2b3 g8g7');
INSERT INTO evaluations VALUES(31811,30,'cp 284',NULL);
INSERT INTO evaluations VALUES(31811,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(31811,32,'cp 292',NULL);
INSERT INTO evaluations VALUES(31811,33,'cp 280',NULL);
INSERT INTO evaluations VALUES(31811,34,'cp 295',NULL);
INSERT INTO evaluations VALUES(31811,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(31811,36,'cp 333','h2h4 g8f6 f2f3 c7c6 g1e2 h7h6 e2f4 g6h7 d1b3 e7e6 b3b7 b8d7 b7c6 f8e7 f1a6 e8g8 c1e3 f6e8 e1c1 a8b8 f4d3 b8b6 c6a4 b6b8 d3c5 d7b6 a4a5 e7h4 c5b7 d8e7 a5c5 b8b7 c5e7 b7e7 h1h4 e7c7 c1d2 f7f5 e3f4 c7c6 a6b5 c6c8 d1c1');
INSERT INTO evaluations VALUES(39408,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39408,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39408,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39408,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39408,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39408,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(39408,36,'cp -11','f1c4 b8c6 g1f3 d7d6 f3g5 g8h6 h2h3 e8g8 e1g1 g8h8 f2f4 f7f6 g5f3 d8e7 d1e2 b4c3 b2c3 c8e6 c4d3 a8e8 a1b1 c6a5 c3c4 b7b6 c1b2 a5b7 b1e1 b7c5 d3c2 e7d7 e4e5 f6e5 f4e5 e6f5 b2c3 f5c2 e2c2 h6f7 f3h4');
INSERT INTO evaluations VALUES(39413,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(39413,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(39413,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(39413,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(39413,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(39413,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(39413,36,'cp -44','b8c6 g1f3 d7d6 f3g5 g8h6 e1g1 e8g8 f2f4 g8h8 a2a3 b4c3 b2c3 d8e7 f1e1 c8g4 g5f3 f7f5 h2h3 g4f3 d1f3 a8e8 c4d3 f5e4 e1e4 e7d8 e4e8 d8e8 d3f1 b7b6 c1d2 c6e5 f3d5 e5d7 a1e1 e8d8 f1d3 d7c5 d3c2 h6g8 c3c4 f8e8 d5h5 g8f6 e1e8 d8e8 h5e8 f6e8 f4f5');
INSERT INTO evaluations VALUES(6682,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(6682,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(6682,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(6682,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(6682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6682,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(6682,36,'cp 0','g1e2 g8f6 a2a3 b4c3 e2c3 d7d6 c1g5 c6e5 e1g1 h7h6 g5h4 e8g8 c4d3 f8e8 h2h3 a7a6 a1c1 e5d3 d1d3 c8d7 f1e1 d7c6 f2f3 e8e6 d3c2 d8d7 c3e2 f6h5 e2d4 e6g6 g1h2 h5f4 h4g3 f4h5');
INSERT INTO evaluations VALUES(48518,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(48518,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(48518,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(48518,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(48518,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(48518,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(48518,36,'cp 37','f6e4 d3e4 c7c5 g1f3 b8d7 c2c3 d7f6 e4c2 c5d4 f3d4 f8c5 e1g1 e8g8 a2a4 b7b6 d1f3 d8d5 f3d5 f6d5 f1d1 c8a6 d4b3 c5e7 c2e4 a8d8 e4d5 a6e2 d1d2 d8d5 d2d5 e6d5 b3d4 e2c4 a4a5 b6a5 a1a5 a7a6');
INSERT INTO evaluations VALUES(48145,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(48145,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(48145,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(48145,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48145,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(48145,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(48145,36,'cp 37','d3e4 c7c5 g1e2 f8e7 e1g1 e8g8 c1e3 c5d4 d1d4 f7f5 e4f3 b8c6 f3c6 b7c6 a1d1 c8a6 f1e1 a6e2 e1e2 e7f6 d4d6 d8d6 d1d6 f6b2 d6e6 f8e8 e3g5 g8f7 e6e7 f7f8 g1f1 e8e7 e2e7 h7h6 e7b7 h6g5 b7b2 a8d8 f1e2 d8d7 b2b8 f8f7');
INSERT INTO evaluations VALUES(48142,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(48142,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(48142,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(48142,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(48142,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(48142,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(48142,36,'cp 31','c7c5 g1f3 b8d7 c1e3 d8c7 c2c3 f8e7 e4c2 e8g8 d1d3 d7f6 d4c5 f8d8 d3c4 b7b6 c5c6 c8d7 e1c1 d7c6 f3d4 e7c5 d4c6 c7c6 e3c5 b6c5 d1d8 a8d8 h1d1 d8d1 c2d1 g7g6 d1f3 c6b6 a2a3 g8f8 c1b1 f6d7 c4e2 b6c7 h2h4 c7f4 g2g3 f4f6 e2e4 f8e7 b1a1');
INSERT INTO evaluations VALUES(20641,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(20641,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(20641,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(20641,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(20641,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(20641,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(20641,36,'cp 65','f1d3 g8f6 c2c4 g7g6 e1g1 f8g7 b1c3 e8g8 f1e1 f6d7 a1b1 d7c5 d3f1 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 g8h8 b2b4 c5e6 a2a4 a7a5 c1e3 a5b4 b1b4 e6c5 c3d5 h8g8 g2g3 d8a5 h3h4 h7h5 g1g2 f8b8 f1e2 c5e4 e3d4 e5d4 e2f3 e4c3');
INSERT INTO evaluations VALUES(28719,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28719,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(28719,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28719,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(28719,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28719,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(28719,36,'cp 0','f1d3 d7d5 e1g1 d5c4 d3c4 b8d7 b1c3 f8d6 d1e2 c7c5 d4c5 d6c5 e3e4 d8b8 c4b5 e8g8 f1d1 f8d8 c1g5 a7a6 b5d7 d8d7 g5f6 d7d1 a1d1 g7f6 e4e5 f6f5 f3g5 h7h6 g5f7 g8f7 e2h5 f7e7 h5h4 e7e8 h4h5 e8e7');
INSERT INTO evaluations VALUES(28717,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28717,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(28717,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(28717,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28717,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28717,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28717,36,'cp 0','d7d5 e1g1 d5c4 d3c4 a7a6 b1c3 b6b5 c4b3 c7c5 e3e4 c5d4 f3d4 b8c6 d4c6 d8d1 f1d1 b7c6 f2f3 f8c5 g1f1 e8e7 c3e2 h8c8 e2d4 c6d7 a2a4 b5a4 b3a4 d7a4 a1a4 f6d7 b2b3 c5b6 c1a3 b6c5');
INSERT INTO evaluations VALUES(2169,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(2169,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(2169,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(2169,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(2169,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(2169,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(2169,36,'cp 59','e1g1 g5e5 f7g8 h8g8 c2c3 d4e6 f2f4 e5e4 f4f5 e7d8 f5e6 e4e6 d2d4 f8e7 c1f4 d7d6 b1a3 c8d7 d4d5 e6g4 d1b3 b7b6 a1e1 e7h4 e1e3 g7g5 f4g3 d7a4 b3c4 g4c4 a3c4 a4b5 b2b3 b5c4 b3c4 g8e8 g3h4 g5h4 e3e8 d8e8');
INSERT INTO evaluations VALUES(2168,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(2168,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(2168,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(2168,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(2168,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(2168,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(2168,36,'cp 64','g5e5 f7g8 h8g8 c2c3 d4e6 f2f4 e5e4 f4f5 e7d8 f5e6 e4e6 d2d4 f8e7 c1f4 d7d6 b1d2 g8e8 f4g3 c8d7 d1c2 e6g6 c2b3 d8c8 a1e1 a7a5 b3f7 e7f6 f7g6 h7g6 d2e4 f6d8 d4d5 a8a6 f1f7 a6b6 b2b3 a5a4 g3f2 b6b5 c3c4 b5b4 e1e3 a4b3 a2b3 b7b5 f2e1 b4b3 e3b3 e8e4 e1c3 e4c4 f7f8 c4a4');
INSERT INTO evaluations VALUES(2172,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(2172,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(2172,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(2172,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(2172,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(2172,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(2172,36,'cp 65','f7g8 h8g8 c2c3 d4e6 d2d4 e5f6 f2f4 g7g6 d4d5 e6g7 c1e3 e7d8 e3d4 f6f7 c3c4 b7b6 b1c3 a7a5 d1d3 c8a6 a2a3 g7e8 e4e5 d8c8 b2b4 a6b7 b4b5 e8g7 c3e4 f8e7 a1e1 g7f5 d4f2 c8b8 g2g4 h7h5 g4f5 g6f5 e4g5 e7g5 f4g5 g8g5 g1h1 b8a7 f2e3 g5g4 f1f5');
INSERT INTO evaluations VALUES(2159,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(2159,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(2159,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(2159,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(2159,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(2159,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(2159,36,'cp 38','h8g8 c2c3 d4e6 d2d4 e5f6 f2f4 g7g6 d4d5 e6g7 c1e3 e7d8 e3d4 f6f7 c3c4 b7b6 b1c3 a7a5 d1b3 c8a6 a1d1 d8c8 a2a3 f8d6 c3a4 d6f4 f1f2 g8e8 a4c3 g7h5 g2g3 g6g5 e4e5 f7g6 g3f4 h5f4 g1h1 a6b7 b3a4 c8d8 d1g1 g6d3');
INSERT INTO evaluations VALUES(2176,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(2176,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(2176,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(2176,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(2176,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(2176,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(2176,36,'cp 74','c2c3 d4e6 d2d4 e5f6 f2f4 g7g6 d4d5 e6g7 c1e3 e7d8 e3d4 f6f7 c3c4 b7b6 b1c3 a7a5 d1f3 c8b7 c3a4 d8c8 a1c1 g7e8 f3b3 a8a6 c4c5 b6c5 d4c5 c8b8 c5d4 f8b4 a2a3 b4f8 b3b5 f8g7 d4f2 e8d6 b5c5 d6e4 c5c7 b8a8 a4b6 a6b6 c7b6');
INSERT INTO evaluations VALUES(2175,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(2175,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(2175,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(2175,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(2175,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(2175,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(2175,36,'cp 42','d4e6 d2d4 e5f6 f2f4 g7g6 d4d5 e6g7 c1e3 e7d8 e3d4 f6f7 c3c4 b7b6 b1c3 a7a5 d1b3 c8a6 c3a4 f8e7 f1f3 g7h5 g2g3 h5f6 d5d6 e7d6 c4c5 f7b3 d4f6 d8e8 a2b3 d6c5 a4c5 b6c5 a1a5 d7d6 f3e3 a6b7 a5a8 b7a8 g3g4 e8f7 f6c3 h7h5 c3a5 h5g4 a5c7 g8e8 c7d6 e8e4 d6c5');
INSERT INTO evaluations VALUES(14771,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(14771,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(14771,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14771,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(14771,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(14771,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(14771,36,'cp 46','b1c3 f8d6 f4g3 e8g8 g1f3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6b4 d3e2 c8f5 e1g1 a7a5 d1d2 b4c6 a1c1 c6e7 e2b5 a8c8 f3e5 c7c5 b2b3 f5e6 b5a6 c8c7 d4c5 b6c5 e5f3 a5a4 b3a4 e7f5 f1d1 f5g3 h2g3 c5c4');
INSERT INTO evaluations VALUES(14768,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14768,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14768,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14768,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(14768,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(14768,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(14768,36,'cp 45','f8d6 f4g3 e8g8 g1f3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 g7g6 e1g1 c6e7 d1a4 c8d7 a4c2 d7f5 f1c1 f5d3 c2d3 c7c6 g3d6 d8d6 d3a6 f8c8 a2a4 c8c7 a4a5 a8b8 h2h3 e7f5 c1c3 g8g7 a1c1 h7h5 c1c2 f5e7 h3h4 d6b4');
INSERT INTO evaluations VALUES(38869,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(38869,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(38869,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(38869,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(38869,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(38869,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(38869,36,'cp 76','b1c3 c7c5 a2a3 b4a5 g1e2 c5d4 e2d4 g8f6 b2b4 a5b6 e1g1 b6d4 e3d4 b8c6 d4d5 e6d5 c3d5 c8e6 d5f6 d8f6 c4e6 f7e6 d1h5 f6f7 h5e2 e8g8 c1b2 a8d8 a1e1 c6d4 b2d4 d8d4 e2e6 f7e6 e1e6 d4d3 e6e3 f8d8 a3a4 d3e3 f2e3 d8d3 f1c1');
INSERT INTO evaluations VALUES(55107,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(55107,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(55107,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(55107,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(55107,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(55107,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(55107,36,'cp -64','e7e5 d2d3 g8f6 f1g2 f8c5 e2e3 b8c6 a2a3 e8g8 g1e2 c8f5 e1g1 f8e8 b3b4 c5d6 b1c3 f5e6 e3e4 d5e4 c3e4 f6e4 g2e4 d8d7 f1e1 e6g4 c1b2 a8d8 d1d2 d6f8 b4b5 c6d4 e2d4 e5d4 a3a4 c7c5 c2c4 b7b6');
INSERT INTO evaluations VALUES(55702,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(55702,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(55702,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(55702,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(55702,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(55702,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(55702,36,'cp -80','d6e5 d4e5 d8d1 e1d1 c8d7 f2f4 b8c6 g1f3 e8c8 c2c3 c6e7 d1c2 g8h6 f1c4 d7c6 h1e1 c6f3 g2f3 e7d5 b1d2 h6f5 d2e4 f5h4 e1f1 f8e7 a2a4 a7a5 c1d2 d8d7 a1d1 h8d8 d2c1 h7h6 e4g3 g7g6 c4b5 c7c6 b5c4 e7c5 c4d5 c6d5');
INSERT INTO evaluations VALUES(55992,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(55992,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(55992,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(55992,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(55992,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(55992,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(55992,36,'cp -68','d4e5 d8d1 e1d1 b8c6 f2f4 c8d7 g1f3 e8c8 c2c3 c6e7 d1c2 g8h6 f1c4 d7a4 b2b3 a4c6 h1e1 c6f3 g2f3 e7d5 b1d2 h6f5 d2e4 f5h4 e1f1 f8e7 c1d2 h4g6 c4d5 d8d5 a1d1 g6h4 d2e1 h4g2 e1d2 g7g6 d2c1 g2h4 d1d2 d5d8 a2a4 b7b6 b3b4 d8d5 d2d5 e6d5 e4g5');
INSERT INTO evaluations VALUES(55990,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(55990,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(55990,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(55990,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(55990,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(55990,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(55990,36,'cp -94','d8d1 e1d1 b8c6 f2f4 c8d7 g1f3 e8c8 c2c3 c6e7 d1c2 g8h6 f1c4 d7c6 h1e1 c6f3 g2f3 h6f5 b1d2 e7d5 d2e4 f5h4 e1f1 f8e7 c1d2 h4g6 c4d5 d8d5 d2e3 g6h4 a1d1 b7b6 a2a4 h8d8 d1d5 d8d5 b2b4 a7a5 b4b5 h7h6 e4d2 h4g2 d2c4 e7c5 e3c5 d5c5');
INSERT INTO evaluations VALUES(34838,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(34838,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(34838,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(34838,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(34838,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(34838,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(34838,36,'cp -88','e1d1 b7b6 c1f4 c8b7 b1d2 f8c5 d1e1 g8e7 g1f3 b8d7 h2h4 h7h5 c2c3 a7a5 f1d3 e7g6 f4g3 c5e7 d3e4 e8c8 e4b7 c8b7 e1e2 h8f8 h1d1 d7c5 a1b1 f8e8 e2f1 b7a6 d2b3 c5e4 f1e2 e4g3 f2g3 c7c5 d1d8 e7d8 b3d2 d8c7');
INSERT INTO evaluations VALUES(34837,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(34837,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(34837,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(34837,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(34837,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(34837,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(34837,36,'cp -96','b7b6 c1f4 f8c5 g1f3 c8b7 f4g3 b8d7 b1d2 h7h5 f1d3 e8c8 d1e2 g8e7 h2h4 e7f5 g3f4 c5e7 d2e4 d7f8 g2g3 f8g6 a1d1 f7f6 a2a3 a7a5 h1f1 d8d7 f1g1 h8e8 d3b5 d7d1 g1d1 g6f4 g3f4');
INSERT INTO evaluations VALUES(14163,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(14163,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(14163,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(14163,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(14163,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(14163,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14163,36,'cp 30','e7e6 g1f3 f8d6 f4g3 e8g8 b1d2 c6e7 b5d3 b7b6 c2c4 d5c4 d2c4 d6b4 c4d2 c8b7 a2a3 b4d2 d1d2 c7c5 d4c5 b6c5 d2c3 a8c8 e1g1 c5c4 d3c4 b7d5 b2b3 f6e4 c3b2 d5c4 b3c4 c8c4 a1c1 d8d5 c1c4 d5c4 f1c1 c4a6');
INSERT INTO evaluations VALUES(15430,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(15430,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(15430,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(15430,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15430,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15430,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(15430,36,'cp 0','a7a6 b5c6 b7c6 e1g1 c6c5 b1c3 f8e7 f3h2 e8g8 f2f4 e5f4 c1f4 c8b7 f1f2 f6d7 d1g4 d7e5 f4e5 d6e5 g1h1 e7g5 g4e2 g5f4 h2f1 d8d4 a1e1 d4b4 e1b1 b4d4');
INSERT INTO evaluations VALUES(11410,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11410,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(11410,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11410,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11410,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11410,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(11410,36,'cp 18','b5c6 b7c6 b1c3 c6c5 e1g1 f8e7 f3d2 e8g8 d2c4 c8e6 b2b3 f6d7 d1e2 a6a5 a2a4 d7b8 c3d5 b8c6 c1b2 e7g5 f2f4 e5f4 d5f4 e6c4 d3c4 c6d4 b2d4 c5d4 a1d1 g5f4 f1f4 c7c5 e2d3 d8e7 f4f5 a8e8 d1e1 h7h6 d3g3');
INSERT INTO evaluations VALUES(20200,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(20200,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(20200,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(20200,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20200,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(20200,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(20200,36,'cp 0','g8e7 d2d4 d8a5 b5c6 b7c6 e1g1 c5d4 c3d4 e7f5 b1c3 c6c5 d4c5 f8c5 c1d2 e8g8 c3d5 c5f2 f1f2 a5d5 d2b4 f8e8 d1d5 e6d5 b4c5 a7a5 b2b3 h7h5 a1c1 c8e6 f2c2 e8d8 c2d2 d8c8 d2c2');
INSERT INTO evaluations VALUES(12948,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(12948,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12948,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12948,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12948,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12948,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12948,36,'cp 0','d2d4 d8a5 b5c6 b7c6 e1g1 c5d4 c3d4 e7f5 a2a3 a5b6 b1c3 c8a6 f1e1 a6c4 b2b4 a7a5 c3a4 b6b5 a4c3');
INSERT INTO evaluations VALUES(22366,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(22366,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(22366,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(22366,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(22366,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(22366,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(22366,36,'cp 73','e5d4 f3d4 g8f6 b1c3 d7b6 c4d3 f8e7 a2a4 a7a5 e1g1 e8g8 f1e1 f8e8 h2h3 f6d7 d3f1 d7c5 b2b3 e7f6 c1f4 c8d7 d1d2 d7c6 f2f3 c5e6 f4e3 d6d5 d4e6 e8e6 a1d1 d5e4 c3e4 c6e4 d2d8 a8d8 d1d8 f6d8 f3e4 e6e4');
INSERT INTO evaluations VALUES(18700,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(18700,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18700,36,'cp 0','f8c5 f1d3 g8e7 a2a3 e7g6 h2h4 h7h6 b1c3 e8g8 h4h5 g6e5 f3e5 c6e5 c1f4 d8f6 f4g3 e5d3 d1d3 c5d4 g3h4 f6e5 h4g3');
INSERT INTO evaluations VALUES(45117,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(45117,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(45117,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(45117,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(45117,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(45117,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(45117,36,'cp -32','d6e5 f3e5 e7e6 e1g1 d8c7 d2d4 c5d4 f1e1 f8d6 d1d4 e8g8 c4f1 d6e5 d4e5 c7c2 b1c3 b8c6 e5g3 c2g6 g3g6 h7g6 c1g5 f6d5 c3d5 e6d5 f1b5 c8e6 b5c6 b7c6 a1c1 d5d4 c1c6 e6a2 c6d6 f8b8 d6d4 b8b2 e1d1 a2e6 g5e3 b2a2 f2f3 a8c8 d4d6 c8c2');
INSERT INTO evaluations VALUES(45308,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(45308,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(45308,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(45308,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(45308,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(45308,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(45308,36,'cp -44','f3e5 e7e6 e1g1 d8c7 f1e1 f8d6 d2d4 c5d4 d1d4 e8g8 c4f1 f8d8 d4c3 b8c6 e5c6 b7c6 g2g3 a7a5 a2a3 c8b7 b1d2 c6c5 f1g2 b7g2 g1g2 f6d5 c3d3 d5b6 d3f3 a8c8 d2e4 d6e5 c1g5 f7f6 g5f4 e5f4 f3f4 c7f4 g3f4 b6d5 g2f3 c5c4 a1d1 h7h6 c2c3 a5a4 e4g3');
INSERT INTO evaluations VALUES(45304,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(45304,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(45304,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(45304,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(45304,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(45304,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(45304,36,'cp -44','e7e6 e1g1 d8c7 d2d4 c5d4 f1e1 f8d6 d1d4 e8g8 c4f1 f8d8 c1f4 f6h5 b1c3 h5f4 d4f4 a7a6 f4h4 b8d7 e5c4 d6f8 a1d1 b7b5 f1d3 d7f6 c3e4 h7h6 e4f6 g7f6 d3e4 d8d1 e1d1 a8b8 c4e3 f6f5 e4f3 g8g7 g2g3 f8e7 h4f4 c7f4 g3f4 c8b7 f3b7 b8b7');
INSERT INTO evaluations VALUES(46108,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46108,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(46108,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(46108,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(46108,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(46108,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(46108,36,'cp -27','e1g1 d8c7 d2d4 c5d4 f1e1 f8d6 d1d4 e8g8 c4f1 f8d8 d4c3 b8c6 e5c6 b7c6 g2g3 f6d5 c3f3 e6e5 b1c3 d5b4 e1e2 d6e7 f3e4 c8e6 a2a3 b4d5 e4e5 c7e5 e2e5 d5c3 b2c3 e7f6 e5e3 d8d1 c1b2 a8d8 e3e1 d1d2 e1e2 d2e2 f1e2');
INSERT INTO evaluations VALUES(46107,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(46107,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(46107,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(46107,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(46107,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(46107,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(46107,36,'cp -32','d8c7 f1e1 f8d6 d2d4 c5d4 d1d4 e8g8 c4f1 d6e5 d4e5 c7c2 b1c3 b8c6 e5g3 c2g6 g3g6 h7g6 c1g5 b7b6 a1c1 c6e7 c3b5 a7a6 b5c7 a8a7 h2h4 f8d8 g5f4 c8b7 f4e3 f6d5 e1d1 d8c8 c7d5 e7d5 a2a3 c8c1 e3c1 b6b5 f1e2');
INSERT INTO evaluations VALUES(45750,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(45750,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(45750,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(45750,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(45750,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(45750,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(45750,36,'cp -6','h7h6 g5f4 c8f5 a2a3 f8e7 h2h3 b8d7 e2e3 e8g8 f1d3 f5d3 d1d3 a7a5 d3c2 f8e8 e1g1 d8b6 f1c1 b6a6 a3a4 e7b4 f3e1 a8c8 e1d3 b7b6 c2d1 a6b7 d1b3 b4c3 c1c3 f6e4 c3c2 c6c5 b3b5 c5c4 d3e5 d7f6 a1c1');
INSERT INTO evaluations VALUES(45551,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45551,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45551,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(45551,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45551,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(45551,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(45551,36,'cp -9','g5f4 c8f5 e2e3 b8d7 a2a3 f8e7 f1d3 f5d3 d1d3 e8g8 e1g1 f8e8 h2h3 a7a5 d3c2 d7b6 a3a4 a8c8 f3e1 b6c4 e1d3 e7d6 c3e2 f6h5 f4d6 c4d6 a1c1 c8c7 c2c3 c7e7 d3e5 g7g6 c3c2 h5g7');
INSERT INTO evaluations VALUES(36669,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(36669,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36669,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(36669,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36669,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36669,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36669,36,'cp 28','b8c6 h2h3 e7e5 d4e5 d6e5 d3b5 d8e7 b5c6 b7c6 e1g1 a7a5 f3e1 f8d8 e1d3 c8a6 f1e1 g7f8 d1f3 a6d3 e3g5 d8d6 c2d3 h7h6 g5h4 f8g7 b2b3 d6d4 a1c1 e7d6 e1d1 a8d8 h4f6 d6f6 f3f6 g7f6 c3a4 d4d3 d1d3 d8d3 c1c6 f6d8 a4c3 d3d2 g2g3 g8g7 g1g2');
INSERT INTO evaluations VALUES(3513,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(3513,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(3513,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(3513,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(3513,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(3513,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3513,36,'cp 29','h2h3 e7e5 d4e5 d6e5 d3b5 d8e7 b5c6 b7c6 e1g1 a7a5 f3e1 c8a6 e1d3 f8d8 f1e1 g7f8 d1f3 a6d3 e3g5 d8d6 c2d3 h7h6 g5h4 g6g5 h4g3 a8d8 e1d1 f6d7 c3e2 f8g7 a1c1 d7f8 b2b3 f8e6 d3d4 a5a4 b3a4 e6d4 e2d4 d6d4 f3b3 e7d7 d1e1 d4d2');
INSERT INTO evaluations VALUES(47394,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(47394,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(47394,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(47394,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(47394,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(47394,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(47394,36,'cp 12','e4f6 e7f6 d2d4 f8d6 f1e2 e8g8 g1f3 c7c5 c1e3 d8e7 e1g1 c5d4 f3d4 b8c6 e2f3 c6d4 d1d4 f8d8 d4h4 c8e6 c2c3 e7c7 g2g3 d6c5 e3c5 c7c5 f3b7 a8b8 f1d1 h7h6 d1d8 b8d8 h4b4 c5b4 c3b4 d8b8 b7f3 b8b4 b2b3 b4d4 a1d1 d4b4 d1d8 g8h7');
INSERT INTO evaluations VALUES(47081,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(47081,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47081,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(47081,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47081,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47081,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47081,36,'cp 1','e7f6 d2d4 f8d6 f1e2 e8g8 g1f3 c7c5 c1e3 d8e7 e1g1 b8c6 d4d5 c6e5 b2b3 f8e8 e2b5 c8d7 b5d7 e7d7 c2c4 b7b6 f1e1 e8e7 h2h3 a8e8 a2a4 e5g6 a4a5 e7e4 a5b6 a7b6 a1a6 d7b7 a6a1');
INSERT INTO evaluations VALUES(48956,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(48956,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48956,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(48956,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(48956,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(48956,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(48956,36,'cp 8','d2d4 f8d6 f1e2 e8g8 g1f3 f8e8 e1g1 c7c6 c2c4 b8a6 c1e3 a6c7 d1c2 c7e6 e2d3 g7g6 h2h3 e6f4 f1e1 f4d3 c2d3 b7b5 c4b5 c8f5 d3b3 c6b5 d4d5 f5e4 f3d2 e4f5 a1c1 a7a6 c1d1 d8d7 d2f1');
INSERT INTO evaluations VALUES(49522,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(49522,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(49522,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(49522,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(49522,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(49522,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(49522,36,'cp 54','h5e5 f8e7 e5a5 e8g8 d2d3 f8e8 f1e2 d7d5 g1f3 e7d6 e1g1 d8e7 a5d2 c7c5 d2d1 b8c6 b1d2 d6c7 b2b3 c8f5 c1b2 a8d8 a2a3 a7a6 f1e1 h7h6 g2g3 c7a5 b3b4 c5b4 d2b3 b4a3 b2a3 a5b4 a3b4 c6b4 d1d2 f5h3 f3d4 d8c8');
INSERT INTO evaluations VALUES(49512,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(49512,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(49512,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(49512,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(49512,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(49512,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49512,36,'cp 37','f8e7 e5a5 e8g8 g1f3 d7d5 d2d3 f8e8 f1e2 e7d6 e1g1 d8e7 a5e1 c7c5 e1d1 b8c6 b1d2 d6c7 a2a3 h7h6 f1e1 c8f5 b2b4 c5b4 a3b4 c6b4 c1a3 c7d6 f3d4 f5d7 d2b3 a7a6 d1d2 b4c6 d4c6 d7c6 a3d6 e7d6 e1b1 f6d7 e2f1 d7c5 b3c5 d6c5 c2c3');
INSERT INTO evaluations VALUES(43022,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43022,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(43022,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(43022,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(43022,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43022,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(43022,36,'cp 62','e5a5 d7d5 d2d3 e8g8 g1f3 f8e8 f1e2 e7d6 e1g1 d8e7 a5e1 c7c5 e1d1 d6c7 a2a3 b8c6 b1d2 f6g4 f1e1 c8f5 d2f1 a8d8 b2b3 e7f6 a1b1 f5g6 c1b2 d5d4 h2h3 g4e5 f3e5 c6e5 e2h5 f6g5 h5g6 e5g6 d1f3 g6h4 f3g4');
INSERT INTO evaluations VALUES(31317,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31317,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31317,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31317,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31317,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31317,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31317,36,'cp 0','f5d3 d1d3 b8c6 e1g1 e7e6 c1f4 g8f6 c2c3 d8b6 d3e2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(31350,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31350,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31350,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31350,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31350,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31350,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31350,36,'cp 0','d1d3 b8c6 e1g1 e7e6 c1f4 g8f6 c2c3 d8b6 d3e2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(31349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31349,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31349,36,'cp 0','e7e6 e1g1 g8f6 c1f4 b8c6 b1d2 f8e7 c2c3 e8g8 d3e2 d8b6 f1b1 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(31721,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31721,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31721,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31721,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31721,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31721,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31721,36,'cp 0','e1g1 g8f6 c1f4 b8c6 b1d2 f8e7 c2c3 d8b6 a2a4 e8g8 f1b1 a7a5 h2h4 b6a6 f3e1 f8c8 d3f3 f6d7 f3g3 d7f6');
INSERT INTO evaluations VALUES(31722,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31722,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31722,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31722,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31722,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31722,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31722,36,'cp 0','b8c6 c1f4 g8f6 b1d2 f8e7 c2c3 d8b6 f1b1 f6h5 f4e3 e8g8 d3e2 h5f6 e3f4 f6h5');
INSERT INTO evaluations VALUES(42062,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(42062,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(42062,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(42062,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(42062,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42062,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(42062,36,'cp 8','b4e7 g1f3 e8g8 f1e2 b7b6 c4d5 f6d5 c3d5 e6d5 c1d2 c7c5 d4c5 b6c5 b2b4 e7f6 a1c1 c5c4 e1g1 f8e8 d1c2 b8c6 d2c3 f6c3 c2c3 d8b6 f1d1 c8e6 f3d4 c6d4 d1d4 a7a5 e2f3 e8d8 c1d1 a5b4 c3b4 b6b4 a3b4 d8b8 f3d5 e6d5 d4d5');
INSERT INTO evaluations VALUES(51940,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(51940,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(51940,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(51940,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(51940,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(51940,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(51940,36,'cp -41','d7d5 c4d5 e6d5 d2d3 g8f6 c1g5 f8e7 f1g2 d5d4 g5f6 e7f6 c3e4 f6e7 a1c1 b8d7 a2a3 e8g8 g1f3 a8b8 e1g1 b7b6 b2b4 c8b7 b4c5 b6c5 d1a4 b7d5 e4d2 d7b6 a4c2 d5c6 f3e5 c6g2 g1g2 d8d5 e5f3');
INSERT INTO evaluations VALUES(54165,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(54165,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(54165,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(54165,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(54165,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(54165,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(54165,36,'cp -41','c4d5 e6d5 d2d3 g8f6 c1g5 f8e7 f1g2 d5d4 g5f6 e7f6 c3e4 f6e7 a1c1 b8a6 a2a3 e8g8 g1f3 d8b6 e4d2 c8e6 d1a4 f8e8 e1g1 a8b8 f3e5 b6c7 e5f3 e6d7 a4d1 d7c6 b2b4 b8d8 b4c5 a6c5 c1c4 c5e6 d1a1 g8h8');
INSERT INTO evaluations VALUES(54141,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(54141,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(54141,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(54141,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(54141,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(54141,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(54141,36,'cp -44','e6d5 d2d3 g8f6 c1g5 f8e7 g5f6 e7f6 f1g2 d5d4 c3e4 f6e7 a1c1 d8a5 d1d2 a5b6 d2c2 b8a6 a2a3 e8g8 g1h3 c8d7 h3f4 d7c6 e1g1 a8d8 e2e3 h7h6 h2h4 f8e8 e3d4 d8d4 f1e1 d4d8 c2c3 e7f8 e1e2 a6c7 e4c5 e8e2 f4e2 c6g2 g1g2 c7d5 c3c2 f8c5 c2c5 b6b2');
INSERT INTO evaluations VALUES(54546,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(54546,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(54546,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(54546,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(54546,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(54546,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(54546,36,'cp -48','d2d3 g8f6 c1g5 f8e7 g5f6 e7f6 f1g2 d5d4 c3d5 f6e5 a1c1 b8a6 d1b3 e8g8 g1f3 e5d6 e1g1 a8b8 e2e4 d4e3 d5e3 c8e6 b3d1 e6a2 b2b3 d6e7 d3d4 c5d4 f3d4 e7c5 c1c2 d8d4 d1d4 c5d4 c2a2 a6b4 a2d2 d4e3 f2e3 a7a6 d2d7 b7b6 f1d1 b8e8 d7b7 e8e3 b7b6 e3b3');
INSERT INTO evaluations VALUES(52763,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(52763,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(52763,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52763,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(52763,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(52763,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52763,36,'cp 0','g8f6 f1d3 c6c5 g1f3 b8c6 e1g1 c5d4 f3d4 c8d7 d4b3 d7g4 f2f3 g4h5 e3e4 d5e4 c3e4 e7e6 e4f6 d8f6 c2c3 f8e7 d1e2 e8g8 c1e3 f8d8 f1d1 f6e5 e3f2 h5g6 d3b5 e5c7 f2g3 c7b6 g3f2 b6c7');
INSERT INTO evaluations VALUES(44939,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44939,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44939,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44939,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44939,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44939,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44939,36,'cp 0','f1d3 c6c5 g1f3 b8c6 e1g1 c5d4 f3d4 c8d7 d4b3 d7g4 f2f3 g4h5 e3e4 d5e4 c3e4 e7e6 e4f6 d8f6 c2c3 f8e7 d1e2 e8g8 c1e3 f8d8 f1d1 f6e5 e3f2 e5c7 f2g3 c7b6 g3f2');
INSERT INTO evaluations VALUES(44946,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44946,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44946,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44946,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44946,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44946,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44946,36,'cp 0','c6c5 g1f3 b8c6 e1g1 c5d4 f3d4 g7g6 d4c6 b7c6 b2b3 f8g7 c1b2 e8g8 c3a4 d8d6 f2f4 f6h5 b2g7 h5g7 d1f3 c8f5 a1d1 a8d8 f3f2 d6c7 a4c5 c7b6 c5a4');
INSERT INTO evaluations VALUES(2063,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(2063,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(2063,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(2063,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(2063,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(2063,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(2063,36,'cp 173','b1c3 g8f6 f1e2 d7d6 c1g5 c6d4 d1d4 e7e5 d4e3 f8e7 f2f4 e5a5 e1c1 h7h6 g5h4 f6g4 e2g4 c8g4 d1d5 a5b6 e3g3 e7h4 g3g4 h4f6 e4e5 d6e5 g4d7 e8f8 f4e5 b6e3 c1b1 f6e5 h1f1 f7f6 c3b5 e5h2 b5c7 h2c7 d7c7 a8e8 c7b7 e3b6 b7d7 b6e6 f1d1 e6d7 d5d7 a7a6 d7a7 h6h5 a2a4 h5h4 a7a6 f8f7');
INSERT INTO evaluations VALUES(2062,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(2062,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(2062,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(2062,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(2062,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(2062,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(2062,36,'cp 163','g8f6 c1e3 d7d6 f1e2 c8d7 e1g1 e7d8 f2f4 f8e7 d1e1 e8g8 a1d1 f8e8 e2f3 f6g4 d4c6 b7c6 e3d4 e7f6 h2h3 f6d4 d1d4 g4h6 e1d2 f7f6 d4d3 g8h8 c3e2 c6c5 e2c3 d7c6 b2b3 d8e7 f1e1 a8d8 c3d5 c6d5 d3d5 e7f7 d2a5');
INSERT INTO evaluations VALUES(1111,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(1111,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(1111,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(1111,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(1111,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(1111,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(1111,36,'cp 161','c1e3 d7d6 f1e2 c8d7 e1g1 e7d8 f2f4 f8e7 d1e1 e8g8 a1d1 f8e8 e2f3 f6g4 d4c6 b7c6 e3d4 e7f6 d4f6 d8f6 h2h3 g4h6 e4e5 f6e7 e1f2 d6e5 f4e5 e8d8 f1e1 d7e8 d1d8 a8d8 f2a7 h6f5 a7f2 e7b4 e1b1 h7h6 f3e4 f5d4 e4d3');
INSERT INTO evaluations VALUES(3225,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(3225,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(3225,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(3225,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(3225,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(3225,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(3225,36,'cp 43','d1d2 c8d7 e1c1 a8c8 c4e2 a7a6 h2h4 f6h5 c1b1 h5g3 h1e1 b7b5 d4c6 d7c6 c3d5 c6d5 e4d5 g3e2 e1e2 d8d7 e3h6 g7h6 d2h6 d7f5 d1d3 c8c5 g2g4 f5f6 h4h5 g6h5 h6f6 e7f6 g4h5 f8c8 c2c3 g8g7 e2e7 c8c7 e7e8');
INSERT INTO evaluations VALUES(24946,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(24946,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(24946,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(24946,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24946,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(24946,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(24946,36,'cp 23','d3e2 g4f3 e2f3 e7e5 e1g1 f8e7 c2c4 e5d4 e3d4 d5c4 c1e3 d8d7 b1d2 c6a5 b2b3 c4b3 d2b3 a5c6 a1b1 f6d5 d1c1 e8g8 b3c5 e7c5 c1c5 d5b6 f1c1 f8e8 f3c6 d7c6 c5c6 b7c6 c1c6 b6d5 h2h3 d5e3 f2e3 e8e3 c6c7 e3d3 c7a7');
INSERT INTO evaluations VALUES(40985,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40985,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40985,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(40985,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(40985,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40985,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40985,36,'cp 13','e8g8 e1g1 f6h5 f4e5 c7c5 d4c5 e7c5 h2h3 h5f6 b1d2 b8c6 e5g3 c5d6 g3d6 d8d6 e3e4 c6e5 f1e1 b7b6 f3e5 d6e5 c2c3 d5e4 d2e4 c8b7 d1f3 f6e4 d3e4 b7e4 e1e4 e5b5 b2b3 b5c5 e4c4 c5e7 c4a4 a8c8 a1d1 f8d8 d1d8 c8d8');
INSERT INTO evaluations VALUES(36350,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36350,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36350,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(36350,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(36350,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(36350,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(36350,36,'cp 25','e1g1 f6h5 f4e5 c7c5 c2c4 d5c4 d3c4 c5d4 f3d4 h5f6 b1c3 c8d7 d1b3 b8c6 d4c6 d7c6 c4b5 c6b5 b3b5 d8d7 f1d1 d7b5 c3b5 a7a6 b5d6 b7b5 a2a4 f6e8 d6e8 f8e8 h2h3 b5a4 a1a4 e7f6 e5f6 g7f6 g1f1 a6a5');
INSERT INTO evaluations VALUES(36349,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36349,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(36349,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(36349,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36349,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(36349,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(36349,36,'cp 0','f6h5 f4e5 c7c5 c2c4 d5c4 d3c4 b8c6 d4c5 c6e5 f3e5 h5f6 b1d2 e7c5 a1c1 c5e7 d2f3 f6d7 d1c2 d8a5 e5d3 d7f6 d3e5');
INSERT INTO evaluations VALUES(53167,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(53167,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(53167,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(53167,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(53167,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(53167,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(53167,36,'cp -13','g8f6 g1f3 b8c6 e2e3 e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 d1e2 a8c8 c1e3 c5e3 e2e3 f6h7 f1d1 h7g5 f3g5 d8g5 c3d5 g5h4 a1c1 c6d4 c1c8 f8c8 d1c1 c8c6 c1c6 b7c6');
INSERT INTO evaluations VALUES(45302,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(45302,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(45302,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(45302,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(45302,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(45302,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(45302,36,'cp -25','g1f3 b8c6 e2e3 e7e6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3a4 c5e7 c2c3 c8g4 c3d4 c6d4 d3e2 g4f3 e2f3 e8g8 c1e3 a8c8 a4c3 d8b6 e3d4 e5d4 c3e2 f8d8 e2f4 d4d3 e4e5 f6d7 f4d5 b6e6 d5e7 e6e7 d1d3 d7e5 d3e2 e5f3 e2f3 c8c2 a1e1 e7c7 b2b3 g8h8 f3g3 c7g3');
INSERT INTO evaluations VALUES(9822,30,'cp 416',NULL);
INSERT INTO evaluations VALUES(9822,31,'cp 425',NULL);
INSERT INTO evaluations VALUES(9822,32,'cp 434',NULL);
INSERT INTO evaluations VALUES(9822,33,'cp 432',NULL);
INSERT INTO evaluations VALUES(9822,34,'cp 452',NULL);
INSERT INTO evaluations VALUES(9822,35,'cp 463',NULL);
INSERT INTO evaluations VALUES(9822,36,'cp 458','c3d4 c5b6 e4e5 f6g4 b1c3 d7d6 h2h3 d6e5 h3g4 c8g4 d4e5 d8d1 f1d1 e8g8 c1f4 g4f3 g2f3 b6d4 c4d5 d4e5 f4g3 e5g3 f2g3 c6a5 d5e4 a8d8 d1d8 f8d8 a1d1 d8d1 c3d1 g7g6 f3f4 g8g7 d1e3 f7f5 e4g2 b7b5 b2b4 a5c4 e3c4');
INSERT INTO evaluations VALUES(9819,30,'cp 414',NULL);
INSERT INTO evaluations VALUES(9819,31,'cp 426',NULL);
INSERT INTO evaluations VALUES(9819,32,'cp 409',NULL);
INSERT INTO evaluations VALUES(9819,33,'cp 430',NULL);
INSERT INTO evaluations VALUES(9819,34,'cp 441',NULL);
INSERT INTO evaluations VALUES(9819,35,'cp 437',NULL);
INSERT INTO evaluations VALUES(9819,36,'cp 448','c5b6 e4e5 f6g4 b1c3 d7d6 h2h3 d6e5 h3g4 c8g4 d4e5 e8g8 c1f4 c6d4 c4e2 d4e2 d1e2 c7c6 e2e4 d8c8 f1e1 f8d8 f4e3 b6c7 a1d1 d8e8 e3f4 c8f5 e4f5 g4f5 c3e4 f5g4 e4d6 e8d8 f4g3 d8d7 d1c1 c7a5 e1e2 a8d8 g1h2');
INSERT INTO evaluations VALUES(9718,30,'cp 450',NULL);
INSERT INTO evaluations VALUES(9718,31,'cp 451',NULL);
INSERT INTO evaluations VALUES(9718,32,'cp 461',NULL);
INSERT INTO evaluations VALUES(9718,33,'cp 470',NULL);
INSERT INTO evaluations VALUES(9718,34,'cp 461',NULL);
INSERT INTO evaluations VALUES(9718,35,'cp 467',NULL);
INSERT INTO evaluations VALUES(9718,36,'cp 463','e4e5 f6g4 b1c3 d7d6 h2h3 d6e5 h3g4 c8g4 d4e5 e8g8 c1f4 f8e8 f1e1 d8d1 a1d1 a8d8 c3d5 b6a5 e1f1 g4f3 g2f3 a7a6 g1h1 c6e5 f4e5 e8e5 d5f6 g7f6 d1d8 g8g7 f3f4 e5h5 h1g1 a5b6 d8d7 h5h4 c4f7 g7h8 f7d5 h4f4 d5b7');
INSERT INTO evaluations VALUES(9706,30,'cp 461',NULL);
INSERT INTO evaluations VALUES(9706,31,'cp 440',NULL);
INSERT INTO evaluations VALUES(9706,32,'cp 469',NULL);
INSERT INTO evaluations VALUES(9706,33,'cp 451',NULL);
INSERT INTO evaluations VALUES(9706,34,'cp 444',NULL);
INSERT INTO evaluations VALUES(9706,35,'cp 449',NULL);
INSERT INTO evaluations VALUES(9706,36,'cp 447','f6g4 b1c3 d7d6 h2h3 d6e5 h3g4 c8g4 d4e5 e8g8 c1f4 f8e8 f1e1 c6d4 c4e2 g4f5 f3d4 d8d4 d1d4 b6d4 g2g4 f5e6 c3b5 d4b6 e1c1 e8e7 c1d1 c7c6 b5d6 e6d5 d1d2 a8d8 f4g3 e7d7 d2d5 c6d5 g1g2');
INSERT INTO evaluations VALUES(11933,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(11933,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(11933,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(11933,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(11933,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(11933,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(11933,36,'cp -20','d7d6 e1g1 f8e7 h2h3 e8g8 b1c3 c6c5 f3d2 c7c6 d2c4 c8e6 b2b3 f6d7 c4e3 f8e8 d1f3 h7h6 a2a4 e7g5 c3e2 a7a5 e2g3 g7g6 g3e2 h6h5 e2c3 d7f8 a1b1 e6c8 e3c4 g5c1 b1c1 f8e6 c3e2 e8f8');
INSERT INTO evaluations VALUES(44200,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(44200,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(44200,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(44200,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(44200,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(44200,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(44200,36,'cp -27','d7d5 d1b3 d5c4 b3c4 f8e7 c4b5 b8d7 b5b3 a8b8 d2d3 b7b6 c1f4 c8b7 g2f3 b7f3 g1f3 b8b7 c3b5 e8g8 b5d6 e7d6 f4d6 f8e8 e1g1 d8c8 f1c1 c8c6 d6e5 d7e5 f3e5 c6d6 e5c4 d6e7 a2a4 h7h6 c4e5 e7c7 e5c4 e8d8 a4a5');
INSERT INTO evaluations VALUES(46162,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(46162,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(46162,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(46162,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(46162,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(46162,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(46162,36,'cp -3','d1b3 d5c4 b3c4 f8e7 c4b5 b8d7 b5b3 a8b8 d2d3 b7b6 c1f4 c8b7 g2f3 b7f3 g1f3 b8b7 e2e4 e8g8 e4e5 f6e8 e1g1 e8c7 f4e3 d8a8 d3d4 b7b8 b3d1 c7d5 c3d5 a8d5 d4c5 b6c5 b2b3 d7b6 a1c1 f8c8 d1d5 e6d5 e3g5 f7f6 e5f6');
INSERT INTO evaluations VALUES(6297,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(6297,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(6297,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(6297,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(6297,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(6297,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(6297,36,'cp 172','f3e5 d6e5 d4e5 c6e5 c4e2 g8f6 c2c4 e8g8 b1c3 f8e8 e1g1 e5g6 f2f3 c7c6 d1d6 d8b6 c4c5 b6d8 f1d1 b7b6 c1e3 b6c5 d6c5 a7a5 d1d6 c8a6 e2a6 a8a6 a1d1 e8e6 d6e6 f7e6');
INSERT INTO evaluations VALUES(14581,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(14581,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(14581,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(14581,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(14581,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(14581,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(14581,36,'cp 63','g5f7 e8f7 d1f3 f7e6 b1c3 c6e7 d3d4 c7c6 d4e5 d8e8 c1e3 e8f7 f3e2 f7g6 c4d3 g6g2 e1c1 d5c3 b2c3 e7d5 h1g1 f8a3 c1d2 g2h3 g1g3 h3h4 g3g7 d5e3 d2e3 a3c5 e3d2 h4f4 d2e1 f4f2 e2f2 c5f2 e1f2 e6e5 h2h4 h8f8 f2e3 a7a5 c3c4 f8f4 d1h1 c8e6 h1e1 a8e8 e3d2 e5d6 g7b7 f4h4 b7a7 e6d7 e1e8 d7e8 a7a5');
INSERT INTO evaluations VALUES(14508,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(14508,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(14508,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(14508,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(14508,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(14508,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(14508,36,'cp 123','e8f7 d1f3 f7e6 b1c3 c6e7 d3d4 c7c6 d4e5 d8e8 c1e3 e8f7 f3e2 f7g6 c4d3 g6f7 c3e4 e7g6 e4d6 f8d6 e2g4 e6e7 e5d6 e7d6 g4g6 f7g6 d3g6 d5e3 f2e3 h8f8 a1d1 d6c7 d1d4 c8e6 a2a3 f8f6 g6d3 a8f8 h1f1 f6f1 d3f1 g7g5 d4a4 a7a6 f1d3 e6d5 e3e4 b7b5 a4d4 d5e6 e4e5 e6f5 d4d6 f8e8 d6h6 e8e5 e1d2 f5d3 c2d3');
INSERT INTO evaluations VALUES(2423,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(2423,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(2423,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(2423,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(2423,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(2423,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(2423,36,'cp 110','d1f3 f7e6 b1c3 c6e7 d3d4 c7c6 d4e5 d8e8 c1e3 e8f7 f3e2 f7g6 c4d3 g6f7 c3e4 e7g6 e4d6 f8d6 e2g4 e6e7 e5d6 e7d6 g4g6 f7g6 d3g6 d5e3 f2e3 h8f8 a1d1 d6e7 d1d4 c8e6 a2a3 a7a5 g6d3 f8f6 d4e4 b7b5 e4f4 g7g5 f4f3 a8f8 f3f6 f8f6 h2h4 g5g4 h1f1 f6f1 d3f1 e7d6 e1f2 b5b4 a3b4 a5b4 f2g3 d6e5');
INSERT INTO evaluations VALUES(2422,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(2422,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(2422,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(2422,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(2422,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(2422,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(2422,36,'cp 108','f7e6 b1c3 c6e7 d3d4 c7c6 d4e5 d8e8 c1e3 e8f7 f3e2 f7g6 c4d3 g6f7 c3e4 e7g6 e4d6 f8d6 e2g4 e6e7 e5d6 e7d6 g4g6 f7g6 d3g6 d5e3 f2e3 c8g4 e1g1 h8f8 f1f4 g4e6 a2a3 d6e7 g1f2 a8d8 f2e2 f8f6 g6d3 g7g5 f4e4 e7d6 e4d4 d6e7 d4d8 e6g4 e2d2 e7d8 a1f1 d8e7 f1f6 e7f6');
INSERT INTO evaluations VALUES(2441,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(2441,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(2441,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(2441,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(2441,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(2441,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(2441,36,'cp 106','b1c3 c6e7 d3d4 c7c6 d4e5 d8e8 c1e3 e8f7 f3e2 f7g6 c4d3 g6f7 c3e4 e7g6 e4d6 f8d6 e2g4 e6e7 e5d6 e7d6 g4g6 f7g6 d3g6 d5e3 f2e3 h8f8 a1d1 d6e7 d1d4 c8e6 a2a3 a7a5 g6d3 f8f6 d4e4 b7b5 a3a4 b5a4 e4a4 c6c5 e1e2 a8f8 e2d2 c5c4 d3c4 f6f2 d2c1 f2g2 c4e6 e7e6');
INSERT INTO evaluations VALUES(16004,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(16004,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(16004,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(16004,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(16004,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(16004,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(16004,36,'cp 90','e2e4 g7g6 f1e2 f8g7 g1f3 e7e5 e1g1 e8g8 d1c2 e5d4 f3d4 f8e8 c1e3 c7c6 a1d1 a7a6 f1e1 d8c7 d4b3 b7b5 c4b5 c6b5 c2d2 g7f8 c3d5 f6d5 e4d5 d7f6 d1c1 c7d8 b3d4 f6e4 d2c2 c8d7 d4c6 d8f6 h2h3 f6h4 e2f3');
INSERT INTO evaluations VALUES(16006,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(16006,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(16006,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(16006,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(16006,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(16006,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(16006,36,'cp 79','g7g6 f1e2 f8g7 d1c2 e7e5 g1f3 e5d4 f3d4 e8g8 c1g5 f8e8 a1d1 d7c5 f2f3 c7c6 e1g1 a7a5 g1h1 a5a4 c2c1 d8c7 g5h6 g7h6 c1h6 c8e6 h6d2 a8d8 f1e1 c7b6 d4c2 f6h5 c2e3 b6a5 e2f1 h5f6 h1g1 a4a3 b2b3');
INSERT INTO evaluations VALUES(15992,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(15992,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(15992,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(15992,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(15992,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(15992,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(15992,36,'cp 92','g1f3 e7e5 c1e3 f8g7 f1e2 e5d4 f3d4 c7c6 e1g1 e8g8 d1c2 a7a6 f1e1 f8e8 a1d1 d8c7 d4b3 b7b5 c4b5 c6b5 c2d2 g7f8 c3d5 c7d8 d5f6 d7f6 f2f3 c8e6 e3f2 e6b3 a2b3 d8c7 f2g3 a8d8 e2f1 c7c6 g3f2 d8c8 g2g3 c6c2 d2c2 c8c2');
INSERT INTO evaluations VALUES(15991,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(15991,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(15991,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(15991,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(15991,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(15991,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(15991,36,'cp 81','f8g7 f1e2 e8g8 e1g1 e7e5 c1e3 e5d4 f3d4 c7c6 d1c2 f8e8 a1d1 a7a6 h2h3 d8c7 d4f3 b7b5 a2a3 d7c5 e4e5 c8f5 c2d2 f6e4 e5d6 c7b8 c3e4 c5e4 d2c1 e4d6 c4c5 d6c4 e2c4 b5c4 e3d4 f5e4 d4g7 g8g7 f3d4 e4d5 f1e1 e8e5 f2f3 e5e1 d1e1');
INSERT INTO evaluations VALUES(47885,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47885,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(47885,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47885,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47885,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47885,36,'cp 0','c7c5 g2g3 b8c6 f1g2 f8e7 b1c3 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(48285,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(48285,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(48285,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48285,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48285,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(48285,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(48285,36,'cp -6','f8d6 g1f3 d6f4 e3f4 b7b6 c2c3 e8g8 f1d3 c7c5 e1g1 d8c7 f3e5 c8a6 d3a6 b8a6 f1e1 a6b8 b1a3 a7a6 a3c2 b8c6 d1e2 f8e8 g2g3 c6e7 c2e3 f6e4 h3h4 b6b5 h4h5 h7h6 d4c5 e4c5 g1g2 e7c6 e5c6 c7c6 e3c2');
INSERT INTO evaluations VALUES(41791,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41791,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41791,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41791,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41791,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(41791,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41791,36,'cp 0','f4d6 d8d6 a2a3 e8g8 g1f3 f8d8 c2c4 d5c4 f1c4 c7c5 b1c3 b8c6 d4c5 d6c5 d1a4 c8d7 c4e2 h7h6 e1g1 c6e5 a4b3 e5f3 e2f3 d7c6 f3c6 c5c6 f1d1 a8c8 d1d8 c8d8 a1d1 d8d1 b3d1 c6b6 d1c2 b6c6 c2d3 c6b6');
INSERT INTO evaluations VALUES(41674,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41674,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41674,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41674,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41674,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41674,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41674,36,'cp -2','d8d6 a2a3 e8g8 g1f3 f8d8 c2c4 d5c4 f1c4 c7c5 b1c3 b8c6 d4c5 d6c5 d1a4 c8d7 c4e2 c6e5 a4f4 e5f3 e2f3 d7c6 e1g1 c6f3 f4f3 c5b6 c3e4 f6e4 f3e4 b6b3 f1c1 a8c8 e4b4 b3b4 a3b4 c8c1 a1c1 g7g5 g1f1 d8d2 c1c7 d2b2 c7b7 h7h6 b4b5 g8g7 b7a7 b2b5 a7c7');
INSERT INTO evaluations VALUES(33419,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(33419,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(33419,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(33419,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33419,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33419,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(33419,36,'cp 0','a2a3 e8g8 g1f3 f8d8 c2c4 d5c4 f1c4 c7c5 b1c3 b8c6 d4c5 d6c5 c4d3 c8d7 e1g1 c6e5 f3e5 c5e5 d1b3 e5a5 f1d1 d7c6 d3b5 a5b6 a3a4 a7a5 d1d8 a8d8 a1d1 f6d5 b5c4 d5b4 c4b5');
INSERT INTO evaluations VALUES(50362,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(50362,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(50362,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(50362,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(50362,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50362,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(50362,36,'cp 0','e7e5 c1b2 b8c6 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a1e1 d8f6 a2a4 a7a5 g2g3 g7g6 c2c3 f6f3 d3e2 f3f6');
INSERT INTO evaluations VALUES(49681,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49681,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49681,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49681,36,'cp 0','c1b2 b8c6 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a1e1 d8f6 a2a4 a7a5 g2g3 g7g6 c2c3 f6f3 d3e2 f3f6');
INSERT INTO evaluations VALUES(49682,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49682,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49682,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49682,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49682,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49682,36,'cp 0','b8c6 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 c7c6 a1e1 d8f6 a2a4 b6c7 b2a3 a7a5 g2g3 g7g6 c2c3 f6f3 d3e2 f3f6');
INSERT INTO evaluations VALUES(52398,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(52398,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(52398,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(52398,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(52398,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(52398,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52398,36,'cp 34','e2e4 d7d5 d2d4 d5e4 c3e4 c8f5 e4g3 f5g6 h2h4 h7h6 g1f3 b8d7 f1d3 g6d3 d1d3 d8c7 d3e2 e7e6 e1g1 g8f6 f1e1 f8d6 g3f5 e8g8 f5d6 c7d6 a2a4 f8e8 b2b3 a8d8 c1a3 d6d5 c2c4 d5f5 a1d1 d7f8 g2g3 f8g6 f3e5 h6h5 e2e3 g6e5 d4e5 d8d1 e1d1');
INSERT INTO evaluations VALUES(55968,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(55968,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(55968,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(55968,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(55968,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(55968,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(55968,36,'cp -15','c7c5 g1f3 b8c6 c2c3 c8d7 f1e2 f7f6 e5f6 g8f6 c1f4 f8e7 e1g1 e8g8 f1e1 f6h5 f4e5 h5f4 e2f1 d8b6 b1d2 a7a5 a2a3 g8h8 a1b1 a8d8 h2h3 f4g6 e5g3 g6f4 d2b3 e6e5 f3e5 c6e5 e1e5 e7d6 e5e1 d7a4');
INSERT INTO evaluations VALUES(54315,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(54315,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(54315,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(54315,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(54315,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54315,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(54315,36,'cp -23','g1f3 b8c6 c2c3 c8d7 f1e2 f7f6 e5f6 g8f6 c1f4 d8b6 d1b3 f6h5 b3b6 a7b6 f4c7 a8c8 c7b6 h5f4 g2g3 f4e2 e1e2 f8e7 a2a4 e8g8 a4a5 e6e5 f3d2 f8f7 b1a3 c8f8 f2f3 g7g5 h2h3 h7h5 a1f1 g5g4 f3g4 h5g4 f1f7 f8f7 h3g4 d7g4 e2e1');
INSERT INTO evaluations VALUES(54314,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(54314,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(54314,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(54314,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(54314,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(54314,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(54314,36,'cp -29','b8c6 c2c3 c8d7 d3d4 d8b6 f1e2 g8h6 e1g1 c5d4 c1h6 d4c3 b2c3 g7h6 c3c4 f8g7 c4d5 c6e5 b1d2 e5f3 e2f3 a8d8 a1b1 b6a6 d2e4 e8g8 d5d6 d7c8 d1b3 b7b6 b1c1 f7f5 e4g3 d8d6 g3h5 a6d3 c1c7 c8d7 b3d3 d6d3 c7a7 d3d2 h5g7');
INSERT INTO evaluations VALUES(20360,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(20360,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(20360,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(20360,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(20360,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(20360,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(20360,36,'cp -13','c2c3 c8d7 f1e2 f7f6 e5f6 g8f6 c1f4 f6h5 f4e5 f8e7 e1g1 e8g8 f1e1 d8e8 b1d2 h5f4 a2a3 e8g6 e5f4 f8f4 b2b4 g6f6 g2g3 f4f5 b4b5 c6a5 c3c4 a8f8 f3h4 f5f2 d2f3 f2e2 d1e2 a5b3 a1b1 b3d4 f3d4 f6d4 e2e3 d5c4 d3c4 d4c4 h4f3 c4a2 e3b3 a2b3 b1b3 e7d6 a3a4 b7b6 f3e5');
INSERT INTO evaluations VALUES(20359,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(20359,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(20359,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(20359,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(20359,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(20359,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(20359,36,'cp -34','c8d7 f1e2 f7f6 e5f6 g8f6 c1f4 d8b6 d1b3 b6b3 a2b3 f6h5 f4e3 f8d6 f3e5 d6e5 e2h5 g7g6 h5f3 b7b6 g2g3 e5f6 e1g1 e8g8 f3g2 a7a5 b1d2 c6e5 d3d4 e5g4 e3f4 c5d4 c3c4 d4d3 h2h3 e6e5 g2d5 g8g7 h3g4 e5f4 d5a8 f8a8 f2f3 f4g3 g1g2 f6b2 a1d1 a5a4 g2g3 a4a3 f1e1');
INSERT INTO evaluations VALUES(19776,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19776,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19776,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19776,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19776,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(19776,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(19776,36,'cp -3','e2e3 g8f6 b1d2 c8f5 g1f3 d8b6 f3h4 f5d7 d1b3 e7e6 f1e2 f8e7 h4f3 e8g8 d4c5 e7c5 e1g1 f8c8 h2h3 a7a5 a2a4 h7h6 f3d4 b6b3 d4b3 c5b6 f1c1 b6c7 f4c7 c8c7 c3c4 g8f8 b3c5 a8c8 c4d5 f6d5 e2b5');
INSERT INTO evaluations VALUES(34455,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(34455,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(34455,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(34455,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(34455,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(34455,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(34455,36,'cp 198','d2d4 f6e7 c1e3 h7h6 d1d2 c8e6 d4d5 e6c8 c4b3 g8f6 h2h3 e7c7 g2g4 c6d5 b3d5 f6d5 c3d5 c7c6 e1c1 b8a6 f3h4 a6c7 f2f4 c7d5 e4d5 c6a4 c1b1 f8e7 h4f3 e8g8 f4e5 d6e5 f3e5 a4e4 d2d4 e4d4 e3d4 f8e8 h1f1 e7f6 e5c4 f6d4 d1d4');
INSERT INTO evaluations VALUES(34457,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(34457,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(34457,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(34457,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(34457,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(34457,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(34457,36,'cp 186','h7h6 c1e3 c8g4 c4e2 b8d7 d1d2 f6d8 h2h3 g4f3 g2f3 f8e7 e1c1 e7g5 h1g1 d8e7 h3h4 g5e3 d2e3 g7g6 d4e5 d6e5 f3f4 e5f4 e3f4 e8c8 d1d6 e7e5 f4d2 g8f6 f2f4 e5e7 e4e5 f6e8 d6d3 e8g7 e2g4 g7f5 c3e4 d7c5 e4c5 e7c5 g1d1 d8d3 d2d3 c8b8 g4f5 g6f5');
INSERT INTO evaluations VALUES(34415,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(34415,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(34415,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(34415,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(34415,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(34415,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(34415,36,'cp 198','c1e3 c8g4 c4e2 b8d7 d1d2 f6d8 h2h3 g4f3 g2f3 f8e7 h1g1 e7g5 e1c1 d8e7 h3h4 g5e3 d2e3 g8f6 d4e5 d6e5 g1g7 f6h5 g7g1 e7c5 e2f1 c5e3 f2e3 e8e7 f1h3 d7b6 b2b3 a8g8 a2a4 g8g1 d1g1 b6d7 a4a5 d7f6 b3b4 h8g8 g1g8 f6g8 c3a4 h5g3 a4c5');
INSERT INTO evaluations VALUES(57671,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(57671,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(57671,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(57671,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(57671,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(57671,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(57671,36,'cp -49','b8c6 c2c3 e7e5 f1c4 g8f6 a2a4 f8g7 d2d3 e8g8 c1g5 h7h6 g5f6 g7f6 b1d2 d7d6 f3d1 f6g7 g1f3 g8h8 a4a5 a8b8 h2h3 f7f5 b2b4 a7a6 c4b3 c6a7 a1c1 a7b5 e1g1 c7c6 d3d4 e5d4 c3d4 b5d4 f3d4 g7d4 d2c4');
INSERT INTO evaluations VALUES(14954,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(14954,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(14954,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(14954,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(14954,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(14954,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(14954,36,'cp 232','f3e5 d8d6 c1f4 g7g5 b5c6 b7c6 f4g3 d6b4 b1d2 f8g7 e1g1 e8g8 e5c6 b4c5 c6e5 d5e4 d2b3 c5b6 f1e1 f8e8 d3d4 f6d7 e1e4 d7e5 g3e5 c8b7 e4e2 f7f6 e5g3 b6b5 e2d2 b5d5 f2f3 g5g4 c2c4 d5c4 b3a5 c4d5 a5b7 d5b7 f3g4 g7h6 d2c2 a8d8 d1f3 h6e3 g3f2 b7f3 g2f3 d8d4 c2c7');
INSERT INTO evaluations VALUES(14944,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(14944,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(14944,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(14944,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(14944,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(14944,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(14944,36,'cp 243','d8d6 c1f4 c8d7 b5c6 d7c6 b1c3 d6e6 e1g1 d5e4 e5c6 e6c6 c3e4 f6d5 f4g3 f8e7 c2c4 d5b6 b2b3 e8g8 d1d2 a8e8 e4c3 e8d8 a1d1 e7f6 f1e1 d8d7 g3e5 f8d8 d2e3 c6e6 e5f6 e6f6 c3e4 f6d4 e3d4 d7d4');
INSERT INTO evaluations VALUES(35079,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(35079,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(35079,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(35079,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(35079,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(35079,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(35079,36,'cp 169','d7d5 b1c3 f8b4 c1g5 f6g6 e4d5 e6d5 f1d3 g6e6 f3e5 b8c6 e1g1 c6e5 c3b5 e6b6 d4e5 a7a6 g5e3 b4c5 b5c3 d5d4 c3a4 b6a5 e5e6 c8e6 a4c5 d4e3 c5e6 f7e6 d1g4 a5e5 a1e1 g8f6 g4a4 e8f7 e1e3 e5b2 a4f4 a8d8 f1e1 b2d4 f4c7 d8d7 c7c4 h8e8 e3e6 d4c4 d3c4 f7f8 e6e8 f6e8 c4d3 d7f7 a2a4 e8f6 a4a5');
INSERT INTO evaluations VALUES(5046,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(5046,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(5046,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(5046,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(5046,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(5046,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(5046,36,'cp -24','d7d5 f1d3 g8f6 e1g1 e8g8 a1b1 a7a5 b2b3 a5a4 d1e2 a4b3 a2b3 e7e5 b1a1 c8e6 a1a8 d8a8 c3a4 f6d7 d2b4 f8b8 b4d6 b8c8 e2e3 d5d4 e3e2 a8a5 a4b2 g7f8 d6f8 c8f8 d3c4 d7c5 c4e6 c5e6 b2d3 a5c7 e2f3 f8a8 h2h3');
INSERT INTO evaluations VALUES(5054,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(5054,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(5054,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(5054,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(5054,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(5054,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(5054,36,'cp -19','f1d3 g8f6 e1g1 e8g8 a1b1 a7a5 b2b3 a5a4 f1e1 a4b3 a2b3 c8b7 d1e2 e7e6 h2h3 d8c7 c3a4 d5e4 d3e4 f6e4 e2e4 c6c5 e4f4 c7c6 f2f3 f8d8 d2e3 d8d5 b1d1 g7d4 e3d4 c5d4 d1d4 c6c2 d4c4 c2d2 e1c1 d5d7 a4c5 a8a2 f4d2 d7d2 c5b7 d2g2 g1f1 g2f2 f1g1 f2h2 c1c3');
INSERT INTO evaluations VALUES(15490,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(15490,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(15490,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(15490,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(15490,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(15490,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(15490,36,'cp -31','h2h3 f8e7 d2d3 e8g8 a2a4 c6a5 c4a2 c8e6 a2e6 f7e6 b1d2 c7c5 f3e2 d8e8 g1f3 f6h5 g2g3 e8g6 f3h2 h5f6 e1g1 a5c6 d2c4 f8f7 c1d2 a8e8 h2g4 e8f8 a1e1 h7h5 g4f6 g7f6 a4a5 d6d5 c4e3 d5e4 d3e4 c6a5 e3g2 a5b3');
INSERT INTO evaluations VALUES(15489,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(15489,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(15489,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(15489,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(15489,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(15489,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(15489,36,'cp -40','f8e7 d2d3 e8g8 g1e2 c6a5 c4b5 d6d5 b1d2 a7a6 b5a4 b7b5 a4c2 c7c5 e1g1 c8b7 f3g3 f8e8 f2f4 e5f4 e2f4 d5e4 d2e4 f6e4 d3e4 e7h4 g3f3 a5c4 f4d5 c4e5 f3f5 g7g6 f5f4 b7d5 e4d5');
INSERT INTO evaluations VALUES(8590,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(8590,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(8590,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(8590,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(8590,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(8590,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(8590,36,'cp -34','d2d3 e8g8 g1e2 c6a5 c4b5 a7a6 b5a4 c7c5 e1g1 b7b5 a4c2 d6d5 b1d2 c8b7 e2g3 g7g6 f3e2 f8e8 d2f3 e7f8 h3h4 h7h5 c1g5 d8b6 e2d2 d5e4 g3e4 f6e4 d3e4 f7f6 g5h6 e8d8 d2c1 f8e7 f1d1 d8d1 c1d1 a8d8 d1e2 a5c4 a2a4 g8h7 a4b5 a6b5 b2b3 c4d6');
INSERT INTO evaluations VALUES(8594,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(8594,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(8594,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(8594,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(8594,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(8594,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(8594,36,'cp -38','e8g8 g1e2 c6a5 c4b5 c7c5 e1g1 d6d5 b1d2 a7a6 b5a4 b7b5 a4c2 c8b7 e2g3 g7g6 f3e2 f8e8 d2f3 e7f8 h3h4 h7h5 c1g5 d8b6 e2d2 d5e4 g3e4 f6e4 d3e4 f7f6 g5e3 a5c4 d2e2 b6c6 a2a4 c4e3 e2e3 c5c4 f1d1 a8d8 d1d8 e8d8 a4b5 a6b5 b2b4 g8g7 e3e2');
INSERT INTO evaluations VALUES(3386,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(3386,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(3386,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(3386,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(3386,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(3386,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(3386,36,'cp -41','g1e2 c6a5 c4b5 f6d7 e2g3 d7b6 b2b4 a5c6 a2a4 a7a5 b5c6 b7c6 b4b5 d6d5 e1g1 d5e4 d3e4 d8e8 f3e2 c8e6 f1e1 e6c4 e2g4 g8h8 g3f5 f8g8 f5e7 e8e7 b1d2 c6b5 d2c4 b5c4 c1e3 g8d8 e1b1 d8d3 e3b6 c7b6 b1b6 d3c3');
INSERT INTO evaluations VALUES(18952,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(18952,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(18952,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(18952,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(18952,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(18952,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(18952,36,'cp 57','e1g1 f8d6 c2c3 a7a6 b5a4 d8c7 d2d4 c5d4 c3d4 c6d4 c1e3 g8e7 b1c3 b7b5 e3d4 e5d4 d1d4 e8g8 a4b3 d6c5 d4d3 d7d6 a1d1 e7c6 b3d5 c8b7 d5c6 b7c6 a2a3 c6b7 c3d5 c7d8 b2b4 c5b6 d1d2 a8c8 d5b6 d8b6 f1d1 f8e8 d3d6 b6d6 d2d6 e8e4');
INSERT INTO evaluations VALUES(18951,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(18951,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(18951,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(18951,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(18951,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(18951,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(18951,36,'cp 57','f8d6 c2c3 a7a6 b5a4 g8e7 d2d4 c5d4 c3d4 e5d4 f3d4 e8g8 c1e3 b7b5 a4b3 c6a5 b1c3 a5b3 d1b3 c8b7 a2a3 d8c7 d4f3 b7c6 a1d1 f8e8 d1d2 h7h6 f1d1 d6f4 e3f4 c7f4 d2d4 f4f6 d1e1 f6g6 d4d1 a8d8 f3d4 g6f6 f2f3 g8h8');
INSERT INTO evaluations VALUES(21605,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(21605,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(21605,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(21605,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21605,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21605,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(21605,36,'cp 31','b1c3 c8f5 a2a3 g8f6 c1g5 f8e7 e2e3 f6e4 g5e7 c6e7 f1d3 c7c6 d3e4 f5e4 c3e4 d5e4 f3d2 f7f5 e1g1 d8d5 a1c1 e8g8 g2g3 d5e6 b2b4 a8d8 d1b3 e6b3 d2b3 d8d5 b3c5 b7b6 c5e6 f8c8 h2h4 g8f7 e6g5 f7g6 g5h3 d5d8 g1g2 g6f7 f1h1');
INSERT INTO evaluations VALUES(21592,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(21592,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(21592,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(21592,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(21592,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(21592,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(21592,36,'cp 36','c8f5 c1f4 f8d6 c3d5 f5e4 f4d6 d8d6 d5c3 e4f3 g2f3 d6d4 h1g1 d4f6 e2e3 a8d8 d1b3 c6e5 f1e2 e5f3 e2f3 f6f3 g1g7 g8e7 g7g1 h8g8 g1g8 e7g8 b3b5 e8f8 a1d1 d8d1 c3d1 b7b6 b5e5 f3c6 d1c3 h7h6 e1e2 g8f6 e3e4 c6c4 e2f3 c4c6 f3g3 c6e6 e5e6 f7e6 a2a3 e6e5');
INSERT INTO evaluations VALUES(35119,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(35119,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(35119,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(35119,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(35119,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(35119,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(35119,36,'cp 143','d2d4 e5d4 c3d5 f6d8 c1f4 d7d6 d1d4 g8f6 e1c1 b8c6 f1b5 c8d7 d4e3 a7a6 b5c6 b7c6 d5f6 d8f6 e4e5 f6e6 e5d6 e6e3 f2e3 c7d6 f4d6 f8d6 d1d6 e8c8 g1f3 h8e8 h1e1 e8e4 f3d2 e4g4 g2g3 d8e8 d6d3 d7e6 d3d4 h7h5 d2e4 c8c7 e4c5 e6d5 d4g4 h5g4 c5a6 c7b6');
INSERT INTO evaluations VALUES(12327,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(12327,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(12327,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(12327,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(12327,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(12327,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(12327,36,'cp -28','f6g4 d1e2 d8c7 c3b5 c7b8 d2d4 a7a6 b5d6 f8d6 e5d6 b8d6 d4d5 c6e7 d5e6 f7e6 c4d3 g4f6 b2b3 e8g8 c1b2 e7d5 e2d2 d6f4 b2e5 f4d2 f3d2 b7b6 d2c4 c8b7 e1g1 b7c6 a2a3 b6b5 c4a5 f8c8 c2c4 d5e7 e5d6 e7f5 d3f5 e6f5 d6c5 b5c4 b3c4 a8b8');
INSERT INTO evaluations VALUES(12297,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(12297,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(12297,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(12297,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(12297,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(12297,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(12297,36,'cp -65','d1e2 d8c7 c3b5 c7b8 d2d4 a7a6 b5d6 f8d6 e5d6 b8d6 d4d5 c6e5 d5e6 f7e6 f3e5 d6e5 f2f3 e5e2 c4e2 g4f6 c1f4 d7d5 f4d6 b7b6 g2g4 e8d7 d6e5 h8f8 h2h4 a6a5 f3f4 c8a6 e2f3 d7c6 e1c1 f6d7 h1e1 d7e5 e1e5 a8e8 d1e1 f8f4 f3d5 c6d6 e5e6 e8e6 d5e6 a6c4 e6c4 f4c4 e1g1 c4e4 b2b3 d6e5 c1b2 e5f4 a2a4 e4e2 b2c3');
INSERT INTO evaluations VALUES(12298,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(12298,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(12298,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(12298,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(12298,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(12298,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(12298,36,'cp -40','d8c7 c3b5 c7b8 d2d4 a7a6 b5d6 f8d6 e5d6 b8d6 d4d5 c6e5 d5e6 f7e6 f3e5 d6e5 f2f3 e5e2 c4e2 g4f6 c1f4 d7d5 f4d6 b7b6 g2g4 e8d7 d6e5 h8f8 h2h4 a6a5 f3f4 c8a6 e2f3 d7c6 e1c1 f6d7 h1e1 d7e5 e1e5 a8e8 f4f5 a6c8 d1e1 e8d8 c2c4 c6d6 e5e6 c8e6 e1e6 d6d7 e6b6 d5c4 b6b7 d7d6 b7g7 d6e5 g7h7');
INSERT INTO evaluations VALUES(50250,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50250,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(50250,36,'cp 0','c7c5 b2f6 e7f6 c2c4 d7d5 c4d5 d8d5 b1c3 d5d6 a1c1 b8c6 f1b5 c8d7 c3e4 d6e5 d1c2 a7a6 g1f3 e5f5 f3h4 f5e5');
INSERT INTO evaluations VALUES(51851,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51851,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(51851,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51851,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51851,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(51851,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(51851,36,'cp -28','d7d5 d2d4 d5c4 d4e5 d8d1 e1d1 c8e6 g1f3 b8a6 b1c3 e8c8 d1e1 h7h6 c1e3 g8e7 f3d4 e7d5 d4e6 f7e6 e3a7 f8b4 a1c1 d5c3 b2c3 b4c5 a7c5 a6c5 c1c2 g7g5 h2h4 h8f8 c2c1 f8f5 g2f3 f5e5 h4g5 h6g5 h1h7 b7b5 c1b1');
INSERT INTO evaluations VALUES(53755,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(53755,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(53755,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(53755,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(53755,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(53755,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(53755,36,'cp -15','d2d4 d5c4 d4e5 d8d1 e1d1 c8e6 g1f3 b8a6 f3g5 a8d8 c1d2 e6f5 b1c3 f7f6 g5f3 f8c5 d1e1 g8e7 a1d1 e8f7 e5f6 g7f6 f3h4 f5e6 c3e4 c5b6 e2e3 a6c5 e4c5 b6c5 e1e2 b7b5 a2a3 e7d5 d2a5 d8d7 g2e4 c5b6 a5b6 a7b6 e4c2 h8d8 c2h7 c4c3 h7g6 f7g7');
INSERT INTO evaluations VALUES(6241,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(6241,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(6241,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(6241,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(6241,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(6241,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(6241,36,'cp 97','c4f7 e8f7 d1d5 f7f8 d5c5 d7d6 c5c3 g8f6 b1d2 f8f7 b2b4 a7a6 a2a4 h8e8 c1b2 f7g8 b4b5 a6b5 a4b5 a8a1 f1a1 c6b8 a1c1 c7c6 h2h3 c8e6 f3d4 d8e7 c3d3 e6d7 c1e1 c6b5 d4b5 d7b5 d3b5 b8c6 b2f6 e7f6 b5b7 e8f8 b7b3 g8h7 b3f3 f6f3 d2f3 c6e5 f3d4');
INSERT INTO evaluations VALUES(6149,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(6149,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(6149,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(6149,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(6149,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(6149,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(6149,36,'cp 124','e8f7 d1d5 f7f8 d5c5 d7d6 c5c3 g8f6 b1d2 f8f7 b2b4 a7a6 c1b2 h8e8 a2a4 f7g8 f1e1 d8e7 a1c1 e7f7 b4b5 a6b5 a4b5 c6e5 f3e5 e8e5 c3c7 e5b5 c7f7 g8f7 d2c4 f6e8 b2a3 c8e6 c4d6 e8d6 a3d6 a8c8 f2f4 g7g6 c1c8 e6c8 e1c1 c8d7 c1c7 f7e6 d6a3 d7c6');
INSERT INTO evaluations VALUES(4655,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(4655,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(4655,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(4655,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(4655,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(4655,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(4655,36,'cp 109','d1d5 f7f8 d5c5 d7d6 c5c3 g8f6 f1e1 f8f7 c3b3 f7g6 b3c2 g6f7 b1c3 h8e8 h2h3 c8e6 b2b3 f7g8 c1b2 d8d7 e1e3 d7f7 a1e1 a8d8 c3b5 d8d7 b5d4 c6b4 c2c3 c7c5 d4e6 e8e6 a2a3 b4c6 b3b4 c5b4 a3b4 a7a6 f3h4 e6e8');
INSERT INTO evaluations VALUES(4656,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(4656,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(4656,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(4656,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(4656,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(4656,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(4656,36,'cp 101','f7f8 d5c5 d7d6 c5c3 g8f6 f1e1 f8f7 c3b3 f7g6 c1f4 g6h7 b3d3 c6b4 d3d2 b4c6 f3d4 c6d4 d2d4 d8e7 b1c3 h8e8 a1c1 e7f7 f2f3 a7a6 d4b4 b7b6 c3e2 h7g8 b4d2 f6h5 f4g3 c8b7 e2d4 a8d8 b2b3 c7c5 d4f5');
INSERT INTO evaluations VALUES(16054,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(16054,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(16054,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(16054,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(16054,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(16054,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(16054,36,'cp -125','c6d4 f3d1 c7c6 c4e2 f8g7 g1f3 e8g8 d2d3 d7d5 e1g1 f8e8 a2a4 a7a5 e4d5 d4e2 d1e2 f6d5 c3d5 c6d5 c2c3 d5d4 c1g5 f7f6 g5d2 c8f5 c3d4 e5d4 e2d1 d8d5 d2f4 g6g5 f4c7 f5d7 h2h3 e8e6 f1e1 e6c6 c7g3 a8c8 h3h4 h7h6 h4g5 h6g5 e1e7');
INSERT INTO evaluations VALUES(16078,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(16078,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(16078,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(16078,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(16078,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(16078,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(16078,36,'cp -112','f3d1 c7c6 d2d3 d7d5 c4b3 f8g7 c1g5 h7h6 g5f6 g7f6 h2h3 e8g8 g1f3 f8e8 e1g1 g8g7 e4d5 d4b3 a2b3 c6d5 f1e1 a7a5 b3b4 b7b6 d3d4 e5e4 f3e5 c8e6 c3b5 g7h7 c2c4 d5c4 b4a5 a8a5 a1a5 b6a5 b5c3 d8b6 c3e4 f6g7 d1d2 a5a4');
INSERT INTO evaluations VALUES(16079,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(16079,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(16079,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(16079,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(16079,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(16079,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(16079,36,'cp -131','c7c6 g1f3 f8g7 e1g1 e8g8 c4e2 d7d5 d2d3 f8e8 f1e1 d4e2 d1e2 h7h6 h2h3 a7a5 e2d1 a5a4 c1d2 d5d4 c3e2 c6c5 c2c3 b7b6 b2b3 c8a6 c3c4 a6b7 b3b4 c5b4 d2b4 f6d7 e2g3 b7c6 a1b1 d8c7 a2a3 h6h5 f3g5');
INSERT INTO evaluations VALUES(16271,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(16271,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(16271,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(16271,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(16271,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(16271,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(16271,36,'cp 74','e2e4 f8b4 f1d3 d7d5 e4e5 d5c4 d3c4 f6e4 g1e2 c6e5 d4e5 d8d1 e1d1 e4f2 d1c2 f2h1 c1e3 c8d7 a1h1 e8g8 h1f1 b4e7 e2d4 a7a6 a2a4 a8d8 c4e2 e7b4 e2f3 b4c3 c2c3 c7c5 d4b3 b7b6 b3d2 d7a4');
INSERT INTO evaluations VALUES(16279,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(16279,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(16279,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(16279,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(16279,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(16279,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(16279,36,'cp 72','d7d5 e4e5 f6e4 g1f3 f8b4 d1c2 f7f5 e5f6 d8f6 c1e3 f6g6 a2a3 b4c3 b2c3 c8d7 f3g5 c6e7 g5e4 d5e4 h2h4 d7c6 h4h5 g6f7 g2g4 f7f3 h1g1 e8c8 g1g3 f3f6 a3a4 g7g6 g3h3 h8g8 f1e2 g6g5 e1c1 h7h6 h3h1 f6f7 a4a5 g8f8 c1b2 a7a6 b2a1 c8b8 d1b1 f7e8 c2b3 b8a8 b3b4 f8f7');
INSERT INTO evaluations VALUES(23316,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(23316,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(23316,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(23316,33,'cp 247',NULL);
INSERT INTO evaluations VALUES(23316,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(23316,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(23316,36,'cp 218','b1c3 d7d6 c1e3 g8e7 d1e2 c7c6 e1c1 b4a6 f3h4 c6d5 d1d5 a6c7 d5d2 c8e6 f2f4 e7c6 g2g3 g7g6 f1g2 f8e7 f4f5 e6f7 h1d1 d8d7 g2f1 e8c8 c1b1 c8b8 e2f2 d8c8 h2h3 g6g5 h4g2 h7h6 e3c5 d6d5 e4d5 e7c5 f2c5');
INSERT INTO evaluations VALUES(30796,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(30796,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(30796,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(30796,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(30796,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(30796,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(30796,36,'cp 26','g8f6 c2c4 e7e6 a2a3 d7d5 c4d5 e6d5 b1c3 f8d6 f4d6 d8d6 e2e3 c7c5 f1e2 e8g8 e1g1 b8c6 d4c5 b6c5 a1c1 f8c8 d1a4 d6e7 f1d1 h7h6 e2d3 c6d8 b2b4 d8e6 d3f1 b7c6 b4b5 c6b7 f3e5');
INSERT INTO evaluations VALUES(28806,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(28806,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(28806,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(28806,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28806,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28806,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28806,36,'cp 8','c2c4 e7e6 a2a3 c7c5 e2e3 c5d4 f3d4 a7a6 b1c3 d7d6 d4f3 f8e7 h2h3 d8c7 f1e2 b8d7 e1g1 f6e4 c3e4 b7e4 a1c1 e8g8 f4g3 f8d8 f3d4 a8c8 b2b4 e6e5 e2f3 e4f3 d4f3 g7g6 g3h4 e7h4 f3h4 d7f6 h4f3 c7b7 d1e2 b6b5 f3d2 g8g7 f1d1 c8c7 c4b5 c7c1 d1c1 a6b5 g1h1');
INSERT INTO evaluations VALUES(34597,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34597,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34597,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34597,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34597,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34597,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34597,36,'cp 0','b8c6 b1c3 d5d6 f1e2 e7e5 g1f3 c8d7 e1g1 e8c8 f1e1 d6g6 f3h4 g6f6 h4f3');
INSERT INTO evaluations VALUES(1613,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1613,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1613,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1613,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1613,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1613,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1613,36,'cp 0','b1c3 d5d6 f1e2 e7e5 g1f3 c8d7 e1g1 e8c8 f1e1 d6g6 f3h4 g6f6 h4f3');
INSERT INTO evaluations VALUES(37408,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(37408,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(37408,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(37408,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(37408,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37408,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37408,36,'cp 27','b1c3 d7d5 c4d5 f6d5 f1g2 c7c5 e1g1 c5d4 f3d4 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 a1d1 d8f8 c3c4 c8a6 c6b5 a6b5 c4b5 b2b5 d1d2 f8a8 d2c2 g7e5 f4e3 b5a5 g1g2 a5a4 f1d1 e5d6 d1c1 a4a5');
INSERT INTO evaluations VALUES(37401,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(37401,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(37401,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(37401,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(37401,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(37401,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(37401,36,'cp 14','d7d5 c4d5 f6d5 f1g2 c7c5 e1g1 b8c6 c3d5 d8d5 c1e3 c5d4 f3d4 d5c4 d4c6 b7c6 d1d2 c8g4 f1c1 c4b5 c1c5 b5b2 d2b2 g7b2 a1b1 a8b8 g2c6 a7a6 c6f3 g4f3 e2f3 e7e5 b1d1 f8d8 d1d8 b8d8 c5c6 b2d4 c6a6 d4e3 f2e3 d8d1 g1g2 d1d2 g2h3 d2f2 f3f4 f2e2 e3e4 e5f4 g3f4 e2d2 a6a8 g8g7');
INSERT INTO evaluations VALUES(36732,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36732,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36732,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36732,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(36732,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36732,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36732,36,'cp 12','c4d5 f6d5 f1g2 c7c5 e1g1 b8c6 c3d5 d8d5 c1e3 c5d4 f3d4 d5c4 g2c6 b7c6 d1c1 c8a6 c1c4 a6c4 a1c1 g7d4 e3d4 c4e2 f1e1 e2f3 e1e7 f8e8 e7e3 f3d5 a2a3 a7a5 c1e1 e8e6 d4c3 a5a4 f2f4 a8e8 e3e5 e6e5 f4e5 e8e7 e1d1 e7e8 d1d4 e8e5 d4a4');
INSERT INTO evaluations VALUES(39075,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(39075,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(39075,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(39075,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(39075,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(39075,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(39075,36,'cp 167','d2d4 b8c6 d4e5 d6e5 f3e5 c6e5 c4e2 g8f6 c2c4 e8g8 b1c3 d7d6 e1g1 a7a5 f2f3 c8d7 d1d4 a5a4 c1e3 f8e8 a1d1 d7c6 c4c5 d6c5 d4c5 d8e7 c5e7 e8e7 d1d2 e7d7 f1d1 d7d2 d1d2 g7g5 d2d1 c6d7 f3f4 e5g4 d1d7 g4e3 d7c7 g5f4 c7b7');
INSERT INTO evaluations VALUES(39076,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(39076,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(39076,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(39076,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(39076,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(39076,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(39076,36,'cp 168','b8c6 d4e5 d6e5 f3e5 c6e5 c4e2 d7d6 b1c3 g8f6 b2b3 c8d7 c1b2 d8e7 d1d2 e8c8 e1c1 d7c6 f2f3 c8b8 h1e1 h8e8 c3b5 a7a6 b5d4 c6d7 e2f1 c7c5 d4e2 d7e6 e2g3 b8a8 c2c4 e7c7 c1b1 a8a7 h2h3 e5c6 g3e2 c6e5 d2e3 a7b8 e2c3 f6h5 b1a1 h5g3');
INSERT INTO evaluations VALUES(49993,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(49993,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(49993,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(49993,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49993,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(49993,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(49993,36,'cp 10','f8b4 d1c2 b8c6 c3d5 e8g8 a2a3 c6d4 c2d3 b4e7 e2e3 d4e6 d5f6 f8f6 b2b4 e5e4 d3b3 c7c6 c1b2 f6f7 g1e2 d7d5 c4d5 d8d5 b3d5 c6d5 e1g1 c8d7 f1c1 a7a6 g2f1 e7d6 a1b1 d7b5 e2d4 e6d4 b2d4 f7c7 c1c7 d6c7 f2f4 g8f7 f1b5 a6b5 b1c1 c7d6 c1c3 g7g6 c3b3 h7h6 d2d3 e4d3 b3d3');
INSERT INTO evaluations VALUES(6743,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(6743,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(6743,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(6743,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(6743,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(6743,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(6743,36,'cp -17','g8f6 c1a3 d7d6 e4e5 f6e4 e5d6 c7d6 d1e2 c8f5 b1d2 e8g8 d2e4 f8e8 c4d5 f5e4 d5e4 d6d5 e4h7 g8h7 e2d3 h7g8 c3d4 a8c8 a1b1 d8d7 h2h3 a5d8 a3c1 a7a6 c1g5 b7b5 g5d8 c6d8 f1c1 d8e6 g2g3 c8c4 f3e5 d7b7 h3h4 c4c1 b1c1 e8c8 c1e1 b5b4');
INSERT INTO evaluations VALUES(10545,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(10545,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(10545,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(10545,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(10545,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(10545,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(10545,36,'cp -5','c1a3 d7d6 e4e5 f6e4 e5d6 c7d6 d1e2 d6d5 f3d4 c6d4 c3d4 f7f5 c4b3 e8f7 e2b5 f7g6 b3d5 h7h6 d5e4 f5e4 a3b4 a5b6 b1c3 c8d7 b5d5 h8e8 d5b7 a8c8 c3d5 g6h7 d5b6 a7b6 b7d5 d7g4 d5d8 e8d8 a1c1 c8c1 f1c1 d8d4 b4c3 d4c4 h2h3 g4h5');
INSERT INTO evaluations VALUES(19246,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(19246,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(19246,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(19246,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(19246,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(19246,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(19246,36,'cp -17','f8g7 f1e2 d7d6 d2d3 e7e6 c1e3 c6d4 d1d2 g8e7 e1g1 e8g8 e3h6 e7c6 h6g7 g8g7 f3d4 c6d4 a1e1 a7a6 e2d1 b7b5 f2f4 a8b8 b2b3 c8b7 c3e2 d8f6 d2e3 h7h6 e2d4 f6d4 f4f5 d4e5 f1f2 e6f5 e4f5 e5e3 f5f6 g7h7 e1e3 b7c6 f2f4');
INSERT INTO evaluations VALUES(19921,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(19921,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(19921,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(19921,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(19921,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(19921,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(19921,36,'cp 33','b5c6 b7c6 f2f4 c5c4 b2b3 g7g6 b3c4 f8g7 g1e2 d8a5 a1b1 g8h6 e1g1 e8g8 b1b3 f7f5 d2d3 f5e4 c3e4 a5a2 h2h3 d6d5 e4d2 a2a5 c1b2 g7b2 b3b2 h6f5 d1e1 a5c7 e1f2 a7a5 b2a2 d5c4 d2c4 c8e6 f1a1 e6c4 d3c4 c7d6 a2a4 d6e6 e2d4 f5d4');
INSERT INTO evaluations VALUES(19710,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19710,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(19710,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(19710,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(19710,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(19710,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(19710,36,'cp 29','b7c6 f2f4 c5c4 b2b3 g7g6 b3c4 f8g7 g1e2 d8a5 a1b1 g8h6 e1g1 e8g8 b1b3 f7f5 d2d3 f5e4 c3e4 a5a2 h2h3 d6d5 e4d2 a2a5 c1b2 a5c5 d3d4 c5d6 c4c5 d6c7 g2g4 c8a6 b3e3 a8e8 g1g2 a6e2 d1e2 a7a5 f4f5 g6f5 g4g5 h6f7 f1f5 c7b7 b2c3 a5a4 d2f3');
INSERT INTO evaluations VALUES(18457,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(18457,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(18457,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(18457,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18457,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(18457,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(18457,36,'cp 43','f2f4 c5c4 b2b3 g7g6 d1e2 c8a6 g1f3 d8a5 b3c4 f8g7 e4e5 g8h6 a1b1 d6d5 e1g1 a6c4 d2d3 a5c3 d3c4 e8g8 h2h3 f8b8 b1b3 c3c4 e2c4 d5c4 b3c3 h6f5 c3c4 e7e6 g1f2 f5e7 f1d1 e7d5 f3g5 g7f8 g5e4 b8b4 c4b4 f8b4 g2g4');
INSERT INTO evaluations VALUES(18462,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(18462,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(18462,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(18462,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(18462,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(18462,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(18462,36,'cp 41','c5c4 b2b3 g7g6 b3c4 f8g7 g1e2 d8a5 a1b1 c8e6 e1g1 e6c4 d2d3 c4a2 b1a1 g7c3 e2c3 a5c3 a1a2 g8f6 h2h3 e8g8 c1b2 c3c5 g1h2 a7a5 d1a1 f6d7 b2d4 c5h5 f1b1 c6c5 d4c3 h5h4 b1f1 c5c4 a2a5 a8a5 a1a5 c4d3 c2d3 d7c5 f1f3 f8b8 a5a7');
INSERT INTO evaluations VALUES(29505,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(29505,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(29505,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(29505,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(29505,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(29505,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(29505,36,'cp 101','f2f3 g4d7 c2c4 c7c6 b1c3 c6d5 c4d5 e7e6 d5e6 d7e6 f1d3 b8c6 g1e2 f8b4 c1e3 f6d5 e3f2 d8g5 g2g3 e8g8 e1g1 a8d8 c3d5 g5d5 d1c2 d5f3 d3h7 g8h8 h7e4 f3g4 a2a3 b4a5 a1d1 e6c4 c2c4 g4e4 c4b5 a5b6 d4d5 c6e5 f2b6 a7b6');
INSERT INTO evaluations VALUES(29503,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(29503,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(29503,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(29503,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(29503,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(29503,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(29503,36,'cp 84','g4d7 c2c4 c7c6 b1c3 c6d5 c4d5 e7e6 d5e6 d7e6 f1d3 b8c6 g1e2 a8c8 c1e3 f8d6 e1g1 e8g8 c3e4 d6c7 g1h1 f6e4 d3e4 c6b4 e3g1 b4d5 d1d3 f7f5 e4d5 d8d5 f1e1 c8d8 e2c3 d5c4 d3d2 c7a5 a2a3 c4b3 e1e5 a5c7 e5e2 e6c4 e2e1 c7a5');
INSERT INTO evaluations VALUES(40965,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40965,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40965,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40965,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(40965,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(40965,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(40965,36,'cp 26','f1g2 e8g8 e1g1 d5c4 d1c2 a7a6 a2a4 c8d7 c2c4 d7c6 c1g5 b8d7 b1c3 a6a5 f1d1 h7h6 g5f6 d7f6 f3e5 c6g2 g1g2 c7c6 a1c1 e7b4 c3a2 b4e7 e2e3 d8c7 a2c3 a8d8 c4e2 c6c5 c3b5 c7b8 e5c4 c5d4 b5d4 e7b4 e2f3 d8d7 c4b6');
INSERT INTO evaluations VALUES(43175,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43175,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(43175,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43175,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(43175,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43175,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43175,36,'cp 0','c7c5 d4d5 b7b5 c4b5 a7a6 b5a6 c8a6 b1c3 d8a5 f4d2 e8g8 e2e4 a6f1 e1f1 b8a6 d1e2 f8b8 a1b1 d7d6 g2g3 a6b4 a2a4 a5a6 b2b3 a6e2 f1e2 b4c2 b1b2 c2a3 e4e5 d6e5 f3e5 f6d5 c3d5 g7e5 d2c3 e5c3 d5c3 a3c4 b2b1 c4a3');
INSERT INTO evaluations VALUES(50320,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(50320,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(50320,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50320,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(50320,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(50320,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(50320,36,'cp -13','e7e6 d3d4 b7b6 b1c3 c8b7 a2a3 d7d5 g1f3 d5c4 e2e3 c7c5 f1c4 b8c6 e1g1 f8e7 d1e2 e8g8 d4c5 b6c5 f1d1 d8c7 c1d2 f8d8 h2h3 c6a5 c3b5 c7b6 d2a5 b6a5 b5c3 a5c7 a1c1 d8d1 e2d1 a8d8 d1c2 c7c6 c1d1 d8d1 c2d1');
INSERT INTO evaluations VALUES(49381,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(49381,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(49381,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(49381,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49381,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49381,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(49381,36,'cp -9','g1f3 d7d5 d3d4 f8e7 b1c3 e8g8 c4d5 e6d5 c1g5 c7c5 d4c5 e7c5 e2e3 b8c6 f1e2 d5d4 f3d4 c6d4 e3d4 c5d4 e1g1 c8e6 e2f3 a8c8 f3b7 c8c4 b7f3 d8b6 g5f6 d4f6 c3d5 b6b2 d1b3 c4c5 a1b1 e6d5 b3b2 f6b2 f3d5 b2c3 f1c1 c5d5');
INSERT INTO evaluations VALUES(53303,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53303,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53303,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53303,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(53303,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53303,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53303,36,'cp 19','e2e3 g8f6 c1b2 b8c6 f1b5 c8d7 c2c4 e7e6 c4d5 e6d5 e1g1 f8d6 d2d4 c5d4 f3d4 e8g8 b1d2 d8e7 b5d3 a7a5 h2h3 a5a4 a2a3 a4b3 d1b3 c6d4 b2d4 d7a4 b3b2 a4c6 d2b3 d6a3 d4f6 e7f6 b2f6 g7f6 b3d4 f8c8 f1d1 a3b4 d4f5 a8a1 d1a1');
INSERT INTO evaluations VALUES(53302,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53302,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(53302,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53302,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(53302,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(53302,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(53302,36,'cp 5','b8c6 c1b2 e7e6 d2d4 g8f6 b1d2 c5d4 e3d4 d8c7 c2c4 f8d6 c4c5 d6f4 f1b5 e8g8 e1g1 a7a6 b5c6 b7c6 f1e1 a6a5 b2c3 f8e8 f3e5 c8a6 d1c2 a5a4 b3a4 f4d2 c3d2 f6e4 e5g4 f7f6 d2b4 e6e5 d4e5 f6e5 e1e4');
INSERT INTO evaluations VALUES(19805,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19805,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19805,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(19805,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19805,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(19805,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19805,36,'cp 8','f1b5 e7e6 c2c4 g8f6 c1b2 f8d6 c4d5 e6d5 d2d4 c5d4 f3d4 e8g8 e1g1 d8b6 d4c6 b7c6 b5e2 b6c7 h2h3 c7e7 b1d2 c8f5 a1c1 c6c5 e2b5 a8c8 d1f3 d6e5 b2e5 e7e5 f3e2 e5e6 b5a6 c8c6 a6d3 f5d3 e2d3 h7h6 d2f3');
INSERT INTO evaluations VALUES(54709,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(54709,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(54709,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(54709,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(54709,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(54709,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(54709,36,'cp -32','g8h6 f2f3 b8c6 h4f2 e7e5 e2e3 h6f5 g1e2 c8e6 b1c3 f8b4 a2a3 b4a5 e2g3 e5d4 g3f5 e6f5 e3d4 e8g8 f1e2 a5c3 b2c3 c6a5 e1g1 d8d7 a3a4 a7a6 f1e1 f8e8 d1d2 d7c6 e2f1 h7h5 f2g3 a5c4 d2f4 f5c2 f4c7 e8e1 c7c6 b7c6 a1e1');
INSERT INTO evaluations VALUES(56341,30,'cp -156',NULL);
INSERT INTO evaluations VALUES(56341,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(56341,32,'cp -184',NULL);
INSERT INTO evaluations VALUES(56341,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(56341,34,'cp -166',NULL);
INSERT INTO evaluations VALUES(56341,35,'cp -181',NULL);
INSERT INTO evaluations VALUES(56341,36,'cp -178','d5c4 b1c3 h7h5 g4h5 h8h5 d1a4 c7c6 a4c4 c8e6 c4a4 g8f6 d2d3 b8d7 a4c2 f6d5 g1f3 d7c5 c3d5 e6d5 h2h4 e5e4 d3e4 d5e4 c2c3 d8d5 c1e3 c5e6 c3b3 d5b3 a2b3 a7a5 e1f1 h5b5 f3d2 e4d5 h4h5 d5b3 d2b3 b5b3 a1b1 b3b5 g2e4 e6c5 e4h7 a5a4 f1g2 c5e6 f2f4 e8c8');
INSERT INTO evaluations VALUES(56548,30,'cp -161',NULL);
INSERT INTO evaluations VALUES(56548,31,'cp -168',NULL);
INSERT INTO evaluations VALUES(56548,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(56548,33,'cp -154',NULL);
INSERT INTO evaluations VALUES(56548,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(56548,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(56548,36,'cp -172','b1c3 h7h5 g4h5 h8h5 d1a4 c7c6 a4c4 c8e6 c4a4 g8f6 d2d3 b8d7 c1d2 d7c5 a4c2 f6d5 c3d5 c6d5 b2b4 c5d7 a1c1 f8d6 g1f3 d7b6 c2d1 f7f6 a2a4 d8d7 b4b5 a8c8 c1c8 d7c8 e2e3 h5h8 h2h4 b6d7 e1g1 e8f7 a4a5 e5e4 d3e4 d5e4');
INSERT INTO evaluations VALUES(1172,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(1172,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(1172,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(1172,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(1172,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(1172,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(1172,36,'cp -110','f7f6 e5c4 e7e4 c4e3 c8e6 b1c3 e4g6 c3e2 f8d6 c2c4 e8c8 c1d2 g6f7 c4c5 d6f8 e1g1 f6f5 e2f4 d8d4 f4e6 f7e6 d1c2 g8h6 f1e1 f5f4 e3f1 e6f7 d2c3 d4d5 b2b4 f8e7 e1e4 h8d8 a1e1 f7g6 a2a4 e7f6 c2b3 f4f3 f1g3 f6c3 b3c3 h6g4 g2f3 g4f6 e4e7 d5d2');
INSERT INTO evaluations VALUES(1168,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(1168,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(1168,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(1168,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(1168,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(1168,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(1168,36,'cp -110','e5g4 e7e4 g4e3 c8e6 b1c3 e4g6 c3e2 f8d6 c2c4 e8c8 c1d2 g8h6 c4c5 d6e7 e1g1 g6f7 d1c2 h8e8 b2b3 g7g5 f2f4 g5f4 f1f4 f6f5 a1f1 c8b8 h2h3 f7g7 f4f3 e6f7 d4d5 f7d5 e3d5 d8d5 b3b4 e7f6 e2f4 d5e5 d2c3 e5e4 c3f6 g7f6');
INSERT INTO evaluations VALUES(40400,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40400,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(40400,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40400,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(40400,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(40400,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40400,36,'cp 12','e1g1 b8c6 c1e3 e8g8 f1e1 c8d7 h2h3 a7a6 e2f1 b7b5 a2a4 b5a4 a1a4 c6d4 a4d4 d8c7 e3g5 d7e6 c3d5 e6d5 e4d5 f8b8 b2b3 a6a5 c2c4 b8b4 d4d3 a5a4 d3e3 b4b3 e3b3 a4b3 d1b3 g7f8 b3b5 a8a2 g5e3 f6d7 e1d1 d7c5');
INSERT INTO evaluations VALUES(40399,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40399,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(40399,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(40399,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40399,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40399,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40399,36,'cp 13','e8g8 c1e3 c8d7 h2h3 b8c6 f1e1 a7a6 e2f1 c6d4 e3d4 a8c8 a2a4 d8a5 c3d5 f6d5 d4g7 d5e3 e1e3 g8g7 e4e5 c8c6 d1d4 a5c5 d4c5 d6c5 f1c4 f8b8 b2b3 h7h5 c4d5 c6c7 e5e6 d7e6 d5e6 f7e6 e3e6 g7f7 e6b6 b8d8 c2c4 d8d2 a4a5 d2b2');
INSERT INTO evaluations VALUES(36016,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36016,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36016,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36016,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36016,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36016,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(36016,36,'cp 23','c1e3 b8c6 d1d2 c8d7 h2h3 c6d4 e3d4 d7c6 f1d1 a7a6 e2f3 f6d7 a2a4 g7d4 d2d4 d8b6 d4b6 d7b6 a4a5 b6d7 f3e2 g8g7 f2f3 f8c8 d1d4 b7b5 a5b6 a6a5 b6b7 c6b7 d4a4 c8c5 a4c4 c5c6 c4c6 b7c6 b2b3 d7c5');
INSERT INTO evaluations VALUES(36008,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36008,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36008,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(36008,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(36008,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36008,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36008,36,'cp 6','b8c6 d1d2 c8d7 d4b3 a8c8 h2h3 a7a6 f2f4 d7e6 a1d1 c6a5 b3a5 d8a5 e2f3 a5b4 e4e5 d6e5 f4e5 f6d7 c3d5 b4d2 d5e7 g8h8 d1d2 c8e8 e7d5 d7e5 f3e2 e6d5 d2d5 e5c4 e2c4 e8e3 d5d7 b7b5 c4f7 g7b2 f7b3 f8f1 g1f1 b2f6 a2a4 f6h4 f1g1');
INSERT INTO evaluations VALUES(3248,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(3248,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(3248,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3248,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(3248,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(3248,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(3248,36,'cp 24','d1d2 c8d7 f2f3 c6d4 e3d4 a7a6 a2a4 d8a5 f1d1 f8e8 c3d5 a5d2 d5f6 g7f6 d1d2 f6g5 d2d1 b7b5 a4b5 a6b5 g1f2 b5b4 e2d3 h7h5 e4e5 a8a1 d1a1 d6e5 d4e5 d7f5 f2e2 e8d8 f3f4 f5d3 c2d3 g5f6 e2e3 d8c8 e5f6 e7f6 e3d2 g8g7 d3d4');
INSERT INTO evaluations VALUES(3252,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3252,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(3252,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(3252,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(3252,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(3252,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(3252,36,'cp 23','c8d7 f2f3 c6d4 e3d4 d8a5 a1d1 f8d8 e2c4 d7c6 a2a3 f6h5 d2f2 g7d4 d1d4 a8c8 f1e1 d8d7 g1h1 d7c7 c4b3 a5g5 c3d1 h5f4 d1e3 h7h5 c2c3 g5f6 h2h4 g8g7 e1d1 f6e5 d4d2 e5b5 g2g3 b5b3 g3f4 b3e6');
INSERT INTO evaluations VALUES(52383,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(52383,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(52383,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(52383,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(52383,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(52383,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(52383,36,'cp -20','d7d5 e4d5 c6d5 d2d4 b8c6 c2c3 c8f5 g1f3 h7h6 f1d3 f5d3 d1d3 e7e6 c1f4 g8e7 f3e5 c6e5 f4e5 e7g6 b1d2 g6e5 d4e5 f8e7 e1g1 e8g8 g1g2 d8b6 a1b1 f8d8 d2b3 a8c8 f2f4 c8c4 b3d2 c4c7 a2a3 d8c8 d2f3 c7c4');
INSERT INTO evaluations VALUES(52814,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(52814,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(52814,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(52814,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(52814,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(52814,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(52814,36,'cp -19','e4d5 c6d5 d2d4 b8c6 c2c3 c8f5 g1f3 h7h6 f3h4 f5h7 f1d3 h7d3 d1d3 e7e6 h4g2 g8f6 e1g1 f8d6 b1d2 e8g8 d2f3 c6a5 c1f4 a5c4 f4d6 d8d6 d3e2 a8c8 g2f4 a7a5 f4d3 a5a4 a2a3 d6a6 f1e1 f6e4 g1g2');
INSERT INTO evaluations VALUES(52718,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(52718,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(52718,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(52718,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(52718,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(52718,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(52718,36,'cp -12','c6d5 d2d4 b8c6 c2c3 c8f5 g1f3 h7h6 f3h4 f5h7 h4g2 g8f6 f1d3 h7d3 d1d3 e7e6 e1g1 f8d6 b1d2 a7a6 d2f3 b7b5 a2a3 e8g8 c1f4 d6e7 f3e5 c6e5 f4e5 f6e4 g2e3 a8c8 d3e2 d8b6 f2f3 e4f6 g1h1 a6a5');
INSERT INTO evaluations VALUES(53433,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(53433,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(53433,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(53433,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(53433,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(53433,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53433,36,'cp -7','d2d4 b8c6 c2c3 c8f5 g1f3 h7h6 f3h4 f5h7 f1d3 h7d3 d1d3 e7e6 e1g1 g8f6 h4g2 f8d6 b1d2 a7a5 a2a4 d8b6 d2f3 b6a6 d3b5 e8g8 c1f4 d6f4 g2f4 f6e4 f4d3 f8c8 b5a6 a8a6 f1e1 a6b6 e1e2 f7f6 f3d2 e4d2 e2d2 g8f7 d2e2 g7g6');
INSERT INTO evaluations VALUES(53386,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(53386,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(53386,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(53386,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(53386,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(53386,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(53386,36,'cp -19','c8f5 g1f3 b8c6 c2c3 h7h6 f3h4 f5h7 f1d3 h7d3 d1d3 e7e6 e1g1 g8f6 h4g2 f8d6 b1d2 e8g8 d2f3 c6a5 c1f4 a5c4 d3e2 f6e4 f4d6 d8d6 a2a4 f7f6 g2e3 c4e3 e2e3 a8e8 e3d3 d6d7 f3h4 a7a6 h4g6 f8f7 f2f3 e4d6');
INSERT INTO evaluations VALUES(6949,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(6949,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(6949,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(6949,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(6949,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(6949,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(6949,36,'cp 79','d5h5 e8f8 h5c5 d8e7 c5e3 d7d5 e4d5 c6b4 e1g1 e7e3 f2e3 b4d5 b1c3 d5c3 b2c3 b7b6 e3e4 c8a6 f1d1 a8e8 a2a4 a6b7 c1a3 c7c5 a4a5 g8f6 a5b6 a7b6 d1d6 b7e4 d6b6 f8g8 f3d2 c5c4 a1e1 e4d3 e1e8 f6e8 b6b8 g8f7 d2f3 h7h6 a3c5 f7e6 c5d4 h8g8 b8b6 e8d6 h2h3');
INSERT INTO evaluations VALUES(6946,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(6946,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(6946,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(6946,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(6946,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(6946,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(6946,36,'cp 90','e8f8 h5c5 d8e7 c5e3 d7d5 e4d5 c6b4 e1g1 e7e3 f2e3 b4d5 b2c3 g8f6 c3c4 d5b6 e3e4 f6e4 f3e5 e4f6 b1c3 c8e6 c3e4 b6c4 e4f6 g7f6 f1f6 f8e7 c1g5 c4e5 f6h6 e7d6 a1d1 d6c5 h6e6 e5c4 e6h6 a7a5 h2h4 a8a6 h6h5 c5b4 d1b1 b4a4 b1b7 a6c6 b7a7 c4d6 g1h2 d6b5 a7b7 b5d6 h5h6 d6b7 h6c6');
INSERT INTO evaluations VALUES(30922,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(30922,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(30922,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(30922,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(30922,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(30922,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(30922,36,'cp 16','b5d7 d8d7 d2d4 c5d4 d1d4 b8c6 d4d3 g7g6 g1f3 g8f6 c3d5 f8g7 e1g1 e8g8 c1g5 f6d5 e4d5 c6e5 f3e5 g7e5 c2c3 f8e8 f1e1 b7b5 e1e2 e5f6 h2h4 a8c8 a1e1 f6g5 h4g5 c8c4 f2f3 a7a6 b2b3 c4h4 d3e3 d7b7 e1d1 a6a5 d1d4 h4d4 e3d4 a5a4 c3c4 a4b3 a2b3 b5c4 b3c4');
INSERT INTO evaluations VALUES(30863,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(30863,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(30863,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30863,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(30863,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(30863,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(30863,36,'cp 1','d8d7 d2d4 c5d4 d1d4 b8c6 d4d3 g7g6 g1f3 g8f6 e1g1 f8g7 c3d5 e8g8 c1g5 f6d5 e4d5 c6e5 f3e5 g7e5 c2c3 e5f6 h2h4 a8c8 g5f6 e7f6 d3f3 f6f5 f3f4 c8c5 f1d1 f8e8 h4h5 e8e4 f4d2 d7b5 h5h6 e4e2 d2f4 c5d5');
INSERT INTO evaluations VALUES(32762,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32762,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(32762,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(32762,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(32762,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(32762,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(32762,36,'cp 10','d2d4 c5d4 d1d4 b8c6 d4d3 g7g6 g1f3 g8f6 e1g1 f8g7 c3d5 e8g8 c2c3 f6h5 f1d1 e7e6 d5b4 h5f6 c1f4 f8d8 b4c6 d7c6 f4g5 h7h6 g5f6 g7f6 a2a4 a8c8 g2g3 b7b6 d3e3 c6c4 e3h6 c4e4 f3g5 f6g5 h6g5 e4e5 g5e5');
INSERT INTO evaluations VALUES(44179,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(44179,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(44179,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44179,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(44179,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(44179,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(44179,36,'cp -18','d7d5 e4d5 e6d5 c4b3 f8e7 g1e2 c8e6 e1g1 e8g8 e2f4 b8c6 f4e6 f7e6 c3e2 e7d6 c2c3 d8c7 h2h3 a8e8 d3d4 g8h8 c1g5 h7h6 g5h4 c5d4 c3d4 d6h2 g1h1 g7g5 h4g3 h2g3 e2g3 c7g7 b3c2 g5g4 d1d2 g4h3 g2h3 g7g5 d2e3 f8g8 e3g5 g8g5');
INSERT INTO evaluations VALUES(46155,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46155,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(46155,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(46155,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(46155,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(46155,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(46155,36,'cp -11','c4b5 b8d7 e4d5 f6d5 c3d5 e6d5 c2c3 a7a6 b5a4 d5d4 c3d4 c5d4 g1f3 f8e7 e1g1 e8g8 c1d2 d7c5 a4c2 a6a5 b2b4 a5b4 d2b4 b7b6 f1e1 c8b7 c2b3 e7d6 b4c5 d6c5 a2a4 c5b4 e1e2 b4c3 a1b1 d8d6 f3d2 c3d2 e2d2');
INSERT INTO evaluations VALUES(50224,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(50224,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(50224,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(50224,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(50224,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(50224,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(50224,36,'cp 42','e2e4 d7d5 e4e5 f6d7 f3f4 c7c5 g1f3 c5d4 f3d4 b8c6 c1e3 c6d4 e3d4 e7e6 b1c3 a7a6 f1e2 d7b8 d4f2 b7b5 e2d3 b8c6 a2a3 f8e7 h2h4 b5b4 a3b4 c6b4 h4h5 e8g8 h5g6 f7g6 c3e2 a8b8 d1d2 b4d3 c2d3 c8d7 e1g1 g8h8 e2d4');
INSERT INTO evaluations VALUES(50234,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(50234,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(50234,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(50234,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(50234,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(50234,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(50234,36,'cp 45','d7d5 e4e5 f6d7 f3f4 c7c5 g1f3 c5d4 f3d4 b8c6 c1e3 e7e6 c2c3 c6d4 c3d4 d7b6 b2b3 c8d7 f1d3 b6c8 e1g1 d8a5 a2a4 c8e7 b1d2 e7c6 d2f3 a5b6 d1e2 f8e7 g2g4 h7h5 h2h3 h5g4 h3g4 c6b4 g1g2 b4d3 e2d3');
INSERT INTO evaluations VALUES(57365,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(57365,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(57365,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(57365,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(57365,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(57365,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(57365,36,'cp 208','e4f5 e6f5 f1d3 g8f6 g1f3 f8e7 e1g1 e8g8 f1e1 f6e4 c2c4 d7d6 b1d2 e4d2 c1d2 e7f6 d1c2 g8h8 e1e2 b8c6 d2c3 g7g5 d4d5 g5g4 d5c6 g4f3 c6b7 c8b7 e2e6 f6c3 c2c3 h8g8 a1e1 d8g5 g2g3 f5f4 c4c5 g5c5 c3c5 d6c5 e6e7 a8d8 d3h7 g8h8 e7c7');
INSERT INTO evaluations VALUES(57338,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(57338,31,'cp 228',NULL);
INSERT INTO evaluations VALUES(57338,32,'cp 215',NULL);
INSERT INTO evaluations VALUES(57338,33,'cp 216',NULL);
INSERT INTO evaluations VALUES(57338,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(57338,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(57338,36,'cp 202','e6f5 f1d3 d7d6 g1f3 g8f6 e1g1 f8e7 f1e1 e8g8 c2c4 c7c5 d3c2 b8c6 b1c3 c6b4 h2h3 c8d7 c2b1 f8e8 d4c5 d6c5 a2a3 b4c6 c3d5 f6d5 d1d5 g8h8 c1e3 e8f8 e3c5 e7c5 d5c5 d8f6 b1c2 b7b6 c5d5 a8d8 a1d1 f6b2 c2a4 b2f6 d5d6 f6d6 d1d6');
INSERT INTO evaluations VALUES(57524,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(57524,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(57524,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(57524,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(57524,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(57524,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(57524,36,'cp 210','f1d3 f8e7 g1f3 d7d6 e1g1 g8f6 c2c4 e8g8 b1c3 c7c5 f1e1 b8c6 d3c2 c6b4 h2h3 c8d7 c2b3 d7e8 f3g5 e8h5 d1d2 h5f7 d4c5 d6c5 d2e2 b4c6 g5f7 f8f7 c1e3 g8h8 a1d1 d8b6 c3d5 f6d5 d1d5 a8d8 g1f1 d8d5 c4d5 c6d4 e2h5 f7f8');
INSERT INTO evaluations VALUES(30295,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(30295,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(30295,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(30295,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(30295,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(30295,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(30295,36,'cp 135','f2f3 b8c6 c1e3 f8e7 d1d2 c6d4 e3d4 d7e6 g2g4 d6d5 e1c1 e8g8 d2f2 c7c6 c1b1 d8c7 g4g5 f6h5 e4d5 c6d5 c3b5 c7d7 f3f4 e6g4 d1e1 f7f6 f1e2 g4e2 f2e2 e7b4 e1f1 d7f7 e2d3 a7a6 b5c3 a8e8 h1g1 b4d6 f4f5 f6g5 g1g5 d6h2');
INSERT INTO evaluations VALUES(52271,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(52271,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(52271,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(52271,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(52271,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(52271,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(52271,36,'cp 0','e7e6 d2d3 b8c6 g1e2 b7b5 c4b3 c8b7 h2h4 h7h6 c1f4 a7a6 a2a4 g8f6 a4b5 a6b5 a1a8 b7a8 b1c3 f6h5 f4h2 d8a5 g2g4 h5g7 f3f6 h8h7 f6f4 d7d6 e1g1 f8e7 f4e3 h6h5 g4g5 h7h8 d3d4 c5c4 b3a4 b5a4 f1a1 a5b4 d4d5 e6d5 a1a4 b4b7 e4d5');
INSERT INTO evaluations VALUES(51737,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51737,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51737,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(51737,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(51737,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(51737,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(51737,36,'cp -20','g1e2 b8c6 d2d3 f8g7 c4b3 g8e7 b1c3 d7d6 e1g1 e8g8 f3h3 a8b8 c1h6 g7h6 h3h6 b7b5 a2a3 g8h8 a1e1 a7a5 h6d2 a5a4 b3a2 f7f5 f2f3 e6e5 f3f4 c6d4 f4e5 d6e5 e2d4 c5d4');
INSERT INTO evaluations VALUES(772,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(772,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(772,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(772,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(772,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(772,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(772,36,'cp 50','d4c5 b6c5 b2b4 c5b6 b4b5 c6a5 e1g1 c8d7 a2a4 a8c8 e2d3 e7f5 f1e1 f8c5 d1e2 h7h6 b1d2 e8g8 a1b1 b6d8 d3c2 d8c7 c1b2 c5e7 c2d3 e7c5 b1d1 f8d8 d3b1 f5e7 e2d3 e7g6 h2h4 a7a6 b2c1 a6b5 a4b5 c5b6 c1a3 c7c3');
INSERT INTO evaluations VALUES(23209,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(23209,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(23209,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(23209,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(23209,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(23209,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(23209,36,'cp 179','e2e4 f6e5 f1c4 g8f6 f3g5 d7d5 e4d5 c6a5 c4d3 d8d5 e1g1 c8g4 f2f3 g4h5 b1c3 d5d7 c1e3 f8b4 d1e2 b4c3 b2c3 e8g8 e3c5 f8e8 c5b4 h7h6 g5e4 a5c6 e4f6 g7f6 e2e4 d7g7 b4c5 g8h8 f1f2 e8g8 a1b1 a8b8 c5e3 h5g6 e4h4 h8h7 d3c4 g8e8');
INSERT INTO evaluations VALUES(51995,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(51995,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(51995,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(51995,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(51995,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(51995,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51995,36,'cp 14','d7d5 e4d5 e6d5 d2d4 c8e6 c1e3 c5c4 e1g1 g8f6 b2b3 c4b3 a2b3 b8c6 e2b5 f8d6 c2c4 e8g8 b1c3 d6b4 c3a2 a7a5 a2b4 a5b4 a1a8 d8a8 d1c2 f6e4 h2h3 a8c8 c4c5 e6f5 f3h4 f5g6 f1e1 e4c3 b5d3 f8e8');
INSERT INTO evaluations VALUES(54256,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(54256,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(54256,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54256,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(54256,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(54256,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(54256,36,'cp 28','e4d5 e6d5 d2d4 c8e6 e1g1 b8c6 c1g5 f8e7 g5e7 d8e7 b1d2 g8f6 d4c5 e7c5 d2b3 c5b6 c2c3 e8g8 a2a4 a7a5 e2b5 f8e8 b3d4 e6d7 d4c2 h7h6 h2h3 a8d8 d1d2 d7f5 a1d1 f5h7 f1e1 e8e1 f3e1');
INSERT INTO evaluations VALUES(54146,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54146,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(54146,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(54146,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(54146,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(54146,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(54146,36,'cp 22','e6d5 d2d4 c8e6 e1g1 b8c6 f1e1 g8f6 b1d2 c5d4 d2b3 f8b4 c1d2 b4d2 d1d2 f6e4 d2d3 e8g8 f3d4 f8e8 c2c3 e6d7 d4c6 b7c6 e2f3 d7f5 f3e4 e8e4 e1e4 f5e4 d3e3 d8b6 a1e1 b6e3 e1e3 e4b1 a2a4 a8b8 a4a5');
INSERT INTO evaluations VALUES(54558,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54558,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(54558,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(54558,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(54558,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54558,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54558,36,'cp 21','d2d4 c8e6 e1g1 b8c6 f1e1 g8f6 b1d2 c5d4 d2b3 f8b4 c1d2 b4d2 d1d2 f6e4 d2d3 e8g8 f3d4 f8e8 c2c3 e6d7 f2f3 c6e5 d3c2 e4d6 b3c5 d8c7 c5d7 e5d7 e2f1 h7h6 a1d1 a7a6 c2f2 e8e1 d1e1 d7f6 d4c2');
INSERT INTO evaluations VALUES(54526,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(54526,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54526,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54526,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(54526,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54526,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(54526,36,'cp 20','c8e6 e1g1 b8c6 f1e1 g8f6 b1d2 c5d4 d2b3 f8b4 c1d2 b4d2 d1d2 f6e4 d2d3 e8g8 f3d4 f8e8 a1d1 c6e5 d3e3 e6d7 b3d2 e4f6 e3b3 d8c7 h2h3 h7h6 d2f1 e5c4 a2a3 a7a6 e2f3 e8e1 d1e1 c7d6 f1g3 b7b5 b3c3 a8c8 g3h5');
INSERT INTO evaluations VALUES(6351,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(6351,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6351,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(6351,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(6351,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(6351,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(6351,36,'cp 0','d7c6 d2d4 c8g4 d4e5 d6e5 d1d8 a8d8 b1d2 e5f4 h2h3 g4c8 d2c4 f4c1 a1c1 g8e7 c1d1 d8d1 e1d1 c6c5 h3h4 b7b6 h4h5 c8b7 f3d2 f7f5 e4f5 e8g8 f2f3 e7f5 h1e1 f5g3 e1e7 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(6099,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(6099,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(6099,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6099,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(6099,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6099,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(6099,36,'cp 0','d2d4 c8g4 d4e5 d6e5 d1d8 a8d8 b1d2 e5f4 h2h3 g4c8 d2c4 f4c1 a1c1 g8e7 c1d1 d8d1 e1d1 e8g8 d1e2 b7b6 h1d1 f7f5 e4f5 e7f5 b2b3 c6c5 e2f1 c8a6 d1d5 a6b7 d5d7 b7f3 g2f3 f8f7 d7d8 f7f8');
INSERT INTO evaluations VALUES(6096,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(6096,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(6096,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(6096,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(6096,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(6096,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6096,36,'cp 0','f7f6 c1e3 d8e7 d1d3 c8e6 b1d2 e8c8 e1c1 g8h6 d2c4 h6g4 c4d6 d8d6 h2h3 e5d4 h3g4 d4e3 d3e3 d6d1 h1d1 b7b6 g4g5 c8b7 a2a3 e6g4 g5f6 g7f6 d1e1 h8e8 f3h4 g4e6 f2f3 c6c5 e3f4 e8d8 e1d1 d8d1 c1d1');
INSERT INTO evaluations VALUES(7865,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(7865,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(7865,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(7865,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(7865,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(7865,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(7865,36,'cp -54','b4c3 b2c3 e8g8 e1g1 d7d6 f1e1 c8e6 f3d2 d8c7 a2a4 a8c8 d2f1 c6a5 c1g5 f6d7 f1g3 a7a6 d3f1 d7c5 g5c1 f7f6 c1e3 f8f7 f2f3 f7d7 d1b1 a5c4 f1c4 e6c4 b1b4 d6d5 e4d5 c4d5 c3c4 d5f7 a1d1 c5e6 a4a5 f7g6 d1d7');
INSERT INTO evaluations VALUES(7876,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(7876,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(7876,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(7876,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(7876,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(7876,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(7876,36,'cp -57','b2c3 d7d6 e1g1 e8g8 a1b1 b7b6 f3h4 h7h6 c1a3 d8c7 d1f3 c6e7 f3e3 c8e6 c3c4 e7g6 h4g6 f7g6 b1b4 a8c8 f2f3 g8h7 f1e1 f6d7 e1d1 d7c5 d3f1 f8f6 d1d2 g6g5 d2d1 c8d8');
INSERT INTO evaluations VALUES(7874,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(7874,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(7874,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(7874,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(7874,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(7874,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(7874,36,'cp -39','d7d6 e1g1 e8g8 h2h3 h7h6 f3h2 d6d5 e4d5 d8d5 c3c4 d5a5 c1d2 a5a6 f1e1 c8e6 c4c5 e6c4 h2g4 f6g4 d3c4 a6c4 d1g4 c4g4 h3g4 f8d8 d2e3 a8c8 a1b1 c8c7 b1b2 d8d7 e1b1 f7f6 f2f3 c6e7 g1f2 e7d5 b2b3 d5e3 f2e3 c7c5 b3b7 d7b7');
INSERT INTO evaluations VALUES(38467,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(38467,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38467,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(38467,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38467,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38467,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(38467,36,'cp 41','c2c4 c7c6 f1e2 g8f6 e1g1 e8g8 b1c3 f6e4 d1b3 b7b6 f1d1 e4c3 b3c3 c8a6 b2b4 c6c5 b4b5 a6b7 a2a4 a7a6 c1b2 a6b5 a4b5 b8d7 c3c1 a8a1 c1a1 d8c7 d4c5 f8a8 a1c1 c7c5 c4d5 c5c1 b2c1 a8a2 e2c4 a2a1');
INSERT INTO evaluations VALUES(38450,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(38450,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(38450,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(38450,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(38450,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(38450,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(38450,36,'cp 36','c7c6 f1e2 g8f6 b1c3 e8g8 e1g1 f6e4 d1b3 b7b6 f1d1 e4c3 b3c3 c8a6 b2b3 c6c5 c3c2 d5c4 b3c4 c5d4 f3d4 d8c8 c1a3 b8c6 a1c1 a6b7 d4b5 f8d8 d1d8 c8d8 c1d1 d8c8 h2h3 h7h5 a3b2 g7b2 c2b2 a8b8');
INSERT INTO evaluations VALUES(37923,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(37923,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(37923,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(37923,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37923,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(37923,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(37923,36,'cp 57','f1e2 g8f6 e1g1 e8g8 b1c3 c8e6 b2b3 e6g4 h2h3 g4f3 e2f3 e7e6 c1a3 f8e8 d1c2 d8c7 a3b2 a7a5 a1d1 b8d7 e3e4 f6e4 f3e4 d5e4 c3e4 a8d8 f1e1 h7h6 b2c3 b7b6 d1d3 d7f6 e4d2 f6d7 d2f3 b6b5 a2a4');
INSERT INTO evaluations VALUES(54919,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(54919,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(54919,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(54919,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(54919,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(54919,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54919,36,'cp 34','g8f6 c2c4 e7e6 b1c3 f8d6 f4d6 d8d6 a1c1 e8g8 c4d5 f6d5 a2a3 d5c3 c1c3 b8d7 g1f3 f8d8 d1c2 c7c6 f1e2 e6e5 e1g1 e5d4 f3d4 d7f8 f1d1 d6f6 h2h3 f8e6 d4b3 d8d1 c2d1 f6e7 d1c2 e6g5 e2d3 c8e6 b3d4 a8d8 b2b4 e7f6 d4e6');
INSERT INTO evaluations VALUES(46890,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(46890,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(46890,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46890,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(46890,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(46890,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(46890,36,'cp 33','c2c4 e7e6 b1c3 f8d6 f4d6 d8d6 a1c1 e8g8 c4d5 f6d5 a2a3 b7b6 c3d5 e6d5 d1c2 c7c6 f1d3 c8b7 g1e2 b8d7 e1g1 f8c8 c2d2 a7a5 e2g3 d7f6 f1e1 c6c5 d3b1 g7g6 d4c5 c8c5 h2h3 c5c1 e1c1 a8c8 c1c8 b7c8 g3e2 c8a6 e2d4 d6e5');
INSERT INTO evaluations VALUES(22442,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(22442,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(22442,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(22442,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(22442,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(22442,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(22442,36,'cp 79','c2c3 a7a6 b5c6 b7c6 d2d4 e5d4 c3d4 g8e7 e1g1 f8g7 c1e3 d7d6 d1d2 h7h6 b1c3 g6g5 f3e1 f7f5 f2f4 g5f4 e3f4 f5e4 c3e4 c8e6 e1c2 e7f5 a1e1 d8d7 g2g4 f5e7 h2h3 e7d5 f4g3 e8c8 d2d3 c8b8 d3a6 h8f8');
INSERT INTO evaluations VALUES(53053,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(53053,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(53053,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(53053,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(53053,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(53053,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(53053,36,'cp -23','g8f6 f1b5 b8d7 b2f6 e7f6 b1c3 g7g6 d2d4 a7a6 b5d7 c8d7 g1e2 f6f5 d1d3 f8g7 a1d1 e8g8 d4c5 f5e4 d3d6 d8a5 d6d7 g7c3 e2c3 a5c3 d7d2 c3c5 e1g1 b7b5 f1e1 a8d8 d2e2 f7f5 d1d8 f8d8 e1d1 d8d4 d1d4 c5d4 c2c4 g8g7 c4b5 a6b5 g1f1');
INSERT INTO evaluations VALUES(45183,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(45183,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(45183,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(45183,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(45183,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(45183,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(45183,36,'cp -23','f1b5 b8d7 d2d3 g7g6 a2a4 f8g7 a4a5 e8g8 a5a6 d7b8 a6b7 c8b7 g1f3 d8b6 b5a4 d6d5 e4e5 f6d7 e1g1 b7c6 h2h4 c6a4 a1a4 b8c6 f1e1 e7e6 h4h5 a7a5 b1d2 b6c7 d1e2 h7h6 h5g6 f7g6 c2c4 d7b6 a4a1 a5a4 b3a4 b6a4 b2a3 g8h7 e2e3 c6b4 a3b4');
INSERT INTO evaluations VALUES(53710,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(53710,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(53710,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(53710,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53710,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(53710,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53710,36,'cp 22','b1c3 g8f6 f1d3 b8a6 c4d5 f6d5 g1f3 a6b4 d3e4 d5f6 e4b1 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 f8e8 h2h3 b4d5 c3d5 c6d5 b1d3 c8d7 d1b3 d7c6 a2a3 g7g5 h4g3 d6g3 f2g3 f6e4 g3g4 d8d6 f3e5 f7f6 e5c6');
INSERT INTO evaluations VALUES(53708,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(53708,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(53708,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53708,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53708,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(53708,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(53708,36,'cp 16','g8f6 f1d3 d5c4 d3c4 f8b4 a2a3 b4c3 b2c3 e8g8 g1e2 b8d7 c4d3 c6c5 e1g1 c5d4 c3d4 d7b6 c1f4 c8g4 f4e5 g4h5 a3a4 a8c8 d1b3 h5g6 e5f6 g7f6 a4a5 d8d5 b3b5 d5b5 d3b5 b6c4 f1e1 c4d2');
INSERT INTO evaluations VALUES(45758,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(45758,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(45758,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45758,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(45758,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45758,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(45758,36,'cp 13','f1d3 d5c4 d3c4 f8b4 a2a3 b4c3 b2c3 e8g8 g1e2 b8d7 c4d3 c6c5 e1g1 c5d4 c3d4 d7b6 f1e1 f8e8 c1f4 c8d7 e2g3 e8e1 d1e1 d7e6 f4e5 b6d7 a1d1 d8e7 a3a4 f6d5 g3f5 e6f5 d3f5 a8e8 f5e4 d5f6 e4f3 h7h6 g1f1 d7e5 d4e5 e7e5 e1e5 e8e5');
INSERT INTO evaluations VALUES(20750,30,'cp 280',NULL);
INSERT INTO evaluations VALUES(20750,31,'cp 284',NULL);
INSERT INTO evaluations VALUES(20750,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(20750,33,'cp 283',NULL);
INSERT INTO evaluations VALUES(20750,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(20750,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(20750,36,'cp 290','d4a4 c8d7 a4b3 d8b6 b3b6 a7b6 c3d5 e8c8 d5b6 c8c7 b6d7 d8d7 f2f3 f7f5 e4f5 d6d5 c1f4 f8d6 f4d6 d7d6 e1f2 g8e7 g2g4 e7c6 a2a4 c5c4 a1e1 c6b4 c2c3 b4a6 f2g3 a6c5 h2h4 c5a4 h1h2 a4c5 g4g5 d6d7 f1h3 h7h6 g5g6 h8f8 h2e2 b7b5 h3g4 c5d3 e2e8 f8e8 e1e8 d3b2');
INSERT INTO evaluations VALUES(20757,30,'cp 303',NULL);
INSERT INTO evaluations VALUES(20757,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(20757,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(20757,33,'cp 287',NULL);
INSERT INTO evaluations VALUES(20757,34,'cp 256',NULL);
INSERT INTO evaluations VALUES(20757,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(20757,36,'cp 269','c8d7 a4b3 d8b6 b3b6 a7b6 c3d5 e8c8 d5b6 c8c7 b6d7 d8d7 f2f3 g7g6 f1b5 d7e7 e1f2 f8h6 a2a4 h6c1 h1c1 e7e5 c1e1 g6g5 c2c3 f7f6 a4a5 g8e7 b5c4 e7c6 a5a6 h8a8 a6b7 a8a1 e1a1 c7b7 c4d5 h7h5 b2b4 c5b4 c3b4 b7b6 d5c6 b6c6');
INSERT INTO evaluations VALUES(25986,30,'cp 296',NULL);
INSERT INTO evaluations VALUES(25986,31,'cp 255',NULL);
INSERT INTO evaluations VALUES(25986,32,'cp 296',NULL);
INSERT INTO evaluations VALUES(25986,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(25986,34,'cp 292',NULL);
INSERT INTO evaluations VALUES(25986,35,'cp 279',NULL);
INSERT INTO evaluations VALUES(25986,36,'cp 274','a4b3 d8b6 b3b6 a7b6 c3d5 e8c8 d5b6 c8c7 b6d7 d8d7 f2f3 g7g6 c1d2 f8g7 f1b5 d7e7 e1c1 g7h6 a2a4 h6d2 d1d2 g8h6 h1e1 f7f6 c2c3 h8d8 d2d5 h6f7 b2b4 f6f5 b4c5 f5e4 e1e4 e7e4 f3e4 d6c5 d5c5 c7b6 c5c4 d8f8 c4d4 f7e5 d4d6 e5c6 e4e5 b6c5 d6d7 c6e5 d7b7');
INSERT INTO evaluations VALUES(24923,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24923,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(24923,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24923,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24923,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(24923,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(24923,36,'cp 20','g4f3 d1f3 e7e6 c2c3 f8d6 b1d2 d6f4 f3f4 d8e7 f4g3 e8g8 f1d3 e6e5 e1g1 a7a6 f1e1 f8e8 a2a4 f6h5 g3f3 h5f6 f3d1 g7g6 d1b3 a8b8 e3e4 d5e4 d3e4 c6a5 b3b4 e7b4 c3b4 f6e4 d2e4');
INSERT INTO evaluations VALUES(24907,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(24907,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(24907,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(24907,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24907,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24907,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(24907,36,'cp 33','d1f3 e7e6 b1c3 f8d6 e1c1 d6f4 f3f4 d8e7 e3e4 d5e4 c3e4 f6e4 f4e4 e8c8 c2c3 e7g5 c1b1 g5d5 f1d3 d5e4 d3e4 c6e7 e4f3 h7h5 h1e1 g7g6 a2a4 c7c6 b2b4 c8d7 b1b2 e7f5 e1e5 d7e7 b2b3 f5h4 d1d3 b7b6 a4a5 b6b5 e5c5 h4f3');
INSERT INTO evaluations VALUES(24905,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(24905,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(24905,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24905,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24905,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24905,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24905,36,'cp 28','e7e6 c2c3 f8d6 b1d2 d6f4 f3f4 d8e7 f4g3 e8g8 f1d3 e6e5 e1g1 a7a6 a2a4 f6h5 g3f3 h5f6 f3d1 g7g6 b2b4 c6d8 d4e5 e7e5 d1b3 e5e7 c3c4 d5c4 d3c4 d8c6 f1b1 f8e8 b3b2 c6e5 c4f1 c7c6 b4b5 a6b5 a4b5 a8a1');
INSERT INTO evaluations VALUES(25204,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(25204,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(25204,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(25204,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25204,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(25204,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(25204,36,'cp 25','c2c3 f8d6 b1d2 d6f4 f3f4 d8e7 f4h2 e8g8 f1d3 e6e5 e1g1 a7a6 a2a4 h7h6 b2b4 f8e8 f1e1 e5e4 d3e2 c6a7 e1b1 a7c8 b4b5 a6b5 e2b5 c7c6 b5e2 b7b6 h2g3 g7g6 c3c4 g8g7 c4d5 f6d5 e2f1 e7f6 b1c1 e8e7 f1c4 h6h5 c4d5 c6d5');
INSERT INTO evaluations VALUES(29919,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(29919,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(29919,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(29919,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(29919,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(29919,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(29919,36,'cp 75','f8e7 b1c3 b8c6 c1e3 c6a5 e1c1 c7c6 g2g4 a5c4 d3c4 f6d7 c1b1 d8a5 e3d2 a5d8 f3g2 e7g5 f2f3 g5d2 g2d2 d7c5 f3f4 d8e7 f4f5 e8c8 d2e3 c8b8 h3h4 c5d7 g4g5 h7h6 b2b3 h6g5 h4g5 f7f6 g5g6');
INSERT INTO evaluations VALUES(28386,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(28386,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(28386,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(28386,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(28386,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(28386,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(28386,36,'cp 91','a2a3 b8c6 c1e3 c6d4 e3d4 e5d4 c4a2 e8g8 b1d2 f6d7 e1g1 c7c6 f3g3 g8h8 a1e1 g7g6 e1e2 d8b6 d2c4 b6c7 f2f4 a8e8 g3f2 d6d5 c4d2 e7c5 e4d5 c6d5 a2d5 d7f6 d5a2 f6h5 e2e8 f8e8 b2b4 c5b6 d2c4 h8g7 c4b6 c7b6');
INSERT INTO evaluations VALUES(31232,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(31232,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(31232,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(31232,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(31232,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(31232,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(31232,36,'cp 121','h2h3 g4h5 c2c4 g8f6 b1c3 e7e6 c1e3 f8e7 d1b3 d8b6 b3b6 a7b6 g2g4 h5g6 f3d2 h7h5 g4g5 f6d7 f1e2 b8a6 f2f4 a6c7 a2a4 f7f5 g5f6 g7f6 h1g1 g6f7 f4f5 e6f5 e4f5 e8c8 g1g7 d8f8 e1c1 c7e8 g7g1 e8c7 d4d5 c6d5 c4d5 c7d5 c3d5 f7d5 d2c4 d5c6 c4d6 e7d6 d1d6');
INSERT INTO evaluations VALUES(31230,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(31230,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(31230,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(31230,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(31230,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(31230,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(31230,36,'cp 106','g4h5 c2c4 g8f6 b1c3 b8d7 d1e2 h5f3 e2f3 g7g6 f1e2 f8g7 e1g1 e8g8 f1d1 a7a6 c1e3 e7e6 f3g3 e6e5 d4e5 d6e5 g3f3 d8e7 c3a4 f8e8 a2a3 a6a5 c4c5 d7f8 e2c4 a8d8 b2b4 f8e6 c4e6 e7e6 d1d8 e8d8 b4a5');
INSERT INTO evaluations VALUES(31228,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(31228,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(31228,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(31228,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(31228,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(31228,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(31228,36,'cp 108','c2c4 b8d7 c1e3 g8f6 b1c3 e7e5 g2g4 h5g6 f1d3 d6d5 c4d5 e5d4 e3d4 f8b4 e1g1 e8g8 f1e1 f8e8 d5c6 b7c6 d1c2 d8c7 d4f6 d7f6 e4e5 b4c3 b2c3 f6d5 a1d1 d5f4 d3g6 f4g6 c2f5 a8d8 h3h4 g6f8 d1d8 e8d8 f5e4 c6c5 g4g5 f8e6 h4h5 c7e7 e4g4 g8h8');
INSERT INTO evaluations VALUES(32863,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(32863,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(32863,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(32863,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(32863,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(32863,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(32863,36,'cp 39','g4f5 c3b5 b8a6 d2d4 e5f6 c2c3 c7c6 b5a3 e7e5 e2a6 b7a6 a3c4 e5d4 d1d4 c6c5 d4f6 g8f6 g1e2 f5e6 c4a5 e8c8 c1f4 f6d5 e1c1 d8d7 h1e1 h7h5 f4g5 f8d6 e2g3 d6f4 g5f4 d5f4 d1d7 c8d7 e1d1 d7c7 g3e4 f4g2 e4c5 g2f4 b2b4 h8d8 c5a6 c7c8');
INSERT INTO evaluations VALUES(32875,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(32875,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(32875,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(32875,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(32875,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(32875,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(32875,36,'cp 40','c3b5 b8a6 d2d4 e5f6 c1e3 c7c6 b5a3 a6c7 g2g4 f5g6 d1d2 h7h6 a3c4 e7e6 e3f2 f6g5 c4e3 g5d8 h2h4 g8f6 e3c4 f8e7 e1c1 c7d5 g1h3 g6h7 d1g1 f6d7 h3f4 e8g8 e2d3 b7b5 d3h7 g8h7 c4e5 d7e5 d4e5 d5f4');
INSERT INTO evaluations VALUES(41524,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(41524,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(41524,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(41524,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(41524,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(41524,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(41524,36,'cp -99','d8d1 c3d1 f6g4 c1f4 b8c6 d1c3 c8d7 c3b5 e8c8 h2h3 g4e5 b5a7 c6a7 f4e5 f7f6 e5f4 e7e5 f4e3 a7b5 f1e2 b5d6 e1g1 d6f5 f1d1 f5e3 f2e3 d7c6 d1d8 h8d8 a1d1 f6f5 d1d8 c8d8 f3g5 c6d5 e3e4 f5e4 g5h7 d8e7 h7g5 e4e3 g1f1 d5a2 g5e4 a2d5 e4c3 d5c6');
INSERT INTO evaluations VALUES(33393,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(33393,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(33393,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(33393,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(33393,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(33393,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(33393,36,'cp -74','c3d1 f6g4 c1f4 b8c6 d1c3 c8d7 c3b5 e8c8 h2h3 g4e5 b5a7 c6a7 f4e5 f7f6 e5c3 e7e5 e1c1 d7c6 f1c4 a7b5 c3d2 e5e4 f3e1 f6f5 c2c3 g7e5 e1c2 b5d6 c4e6 c6d7 e6d7 d8d7 d2e3 h8e8 e3d4 d6c4 a2a4 e5d4 d1d4 d7d4 c2d4 e8e5 d4b3 c7c6');
INSERT INTO evaluations VALUES(33394,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(33394,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(33394,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(33394,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(33394,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(33394,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(33394,36,'cp -78','f6g4 c1f4 b8c6 d1c3 c8d7 c3b5 e8c8 h2h3 g4e5 b5a7 c6a7 f4e5 f7f6 e5d4 a7c6 d4c3 e7e5 e1c1 d7e6 f3d2 g7h6 c1b1 c6e7 f1c4 e6d5 c4d5 e7d5 d2e4 h8f8 c3d2 h6d2 e4d2 f6f5 d2f3 e5e4 f3g5 d5f6 f2f3 d8d1 h1d1 c7c6 f3e4 f5e4');
INSERT INTO evaluations VALUES(33400,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(33400,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(33400,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(33400,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(33400,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(33400,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(33400,36,'cp -73','c1f4 b8c6 d1c3 c8d7 c3b5 e8c8 h2h3 g4e5 b5a7 c6a7 f4e5 f7f6 e5d4 a7b5 d4c5 b5d6 f1e2 b7b6 c5e3 e7e5 e1g1 c8b7 a2a4 d8a8 b2b3 d7c6 e3d2 e5e4 f3d4 f6f5 d2c3 c6d7 f1e1 h8g8 a1d1 f5f4 e2g4 d7g4 h3g4 h7h6 c3b4');
INSERT INTO evaluations VALUES(33399,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(33399,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(33399,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(33399,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(33399,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(33399,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(33399,36,'cp -75','b8c6 d1c3 c8d7 c3b5 e8c8 h2h3 g4e5 b5a7 c6a7 f4e5 f7f6 e5c3 e7e5 e1c1 d7c6 f1c4 g7h6 c3d2 h6d2 d1d2 d8d2 c1d2 c8d7 f3e1 a7c8 c4b3 d7e7 f2f3 e5e4 f3e4 c6e4 h1g1 f6f5 e1d3 c8d6 a2a4 h7h5 d3c5 e7f6 c5e4 d6e4 d2c1 h8e8 g1e1 g6g5 b3d5 c7c6 d5e4');
INSERT INTO evaluations VALUES(489,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(489,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(489,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(489,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(489,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(489,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(489,36,'cp -67','d1c3 c8d7 c3b5 e8c8 h2h3 g4e5 b5a7 c6a7 f4e5 f7f6 e5c3 e7e5 e1c1 d7c6 f1c4 a7b5 c3d2 b7b6 d2b4 e5e4 c4b5 c6b5 f3d4 b5d7 b4c3 h8e8 h1e1 g7h6 c3d2 h6f8 d4e2 c8b7 d2e3 f8d6 c2c4 f6f5 b2b3 b6b5 e3g5 d8a8 g5f4 a8a2 f4d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(2057,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(2057,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(2057,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(2057,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(2057,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(2057,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(2057,36,'cp 159','b1c3 e5d4 f3d4 g8f6 f1e2 d7d6 c1g5 c6d4 d1d4 e7e5 d4e3 h7h6 g5f6 e5f6 c3d5 f6d8 e1c1 c7c6 e4e5 c8e6 d5f4 d6d5 f4e6 f7e6 f2f4 d8e7 c1b1 e8c8 e3a7 e7c5 a7a4 c8c7 h1f1 c5b4 a4b4 f8b4 c2c3 b4c5 g2g3 g7g5 f4f5 e6f5 f1f5');
INSERT INTO evaluations VALUES(9007,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(9007,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(9007,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(9007,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(9007,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(9007,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(9007,36,'cp 51','d8d6 f1b5 c8d7 b1d2 h7h6 b5c6 d7c6 f3e5 f6d7 d2f3 d7e5 f3e5 e8g8 d1h5 d6e7 h5g4 g8h7 g4h4 f7f6 e5c6 b7c6 e1g1 c6c5 h4g4 f8e8 a1c1 c5c4 b2b3 c4b3 a2b3 a7a5 c3c4 c7c6 g4e2 h7h8 e2d3 a5a4 b3a4 a8a4 c4d5 c6d5 c1c5 e7b7 f1c1');
INSERT INTO evaluations VALUES(29764,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(29764,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(29764,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(29764,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(29764,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(29764,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(29764,36,'cp 62','e7e6 h2h3 g4f3 e2f3 c7c6 e1g1 b8d7 c1f4 d7b6 d1d3 f8d6 f4d2 e8g8 a1d1 b6d5 c3e4 f6e4 d3e4 d5f6 e4d3 f8e8 g2g3 d8c7 g1g2 a8d8 d2c1 d8d7 d3e2 e8d8 h3h4 h7h6 f1e1 d6e7 c2c3 c6c5 c1f4 e7d6');
INSERT INTO evaluations VALUES(30215,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(30215,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(30215,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(30215,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(30215,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(30215,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(30215,36,'cp 68','h2h3 g4f3 e2f3 c7c6 e1g1 f8d6 f1e1 e8g8 c3e4 f6e4 f3e4 b8d7 c2c3 d7f6 e4c2 d6c7 g2g3 d8d5 d1e2 a8d8 c1e3 f8e8 a1d1 h7h6 c3c4 d5d6 e2d2 e6e5 d4e5 d6e5 e3d4 e5d6 e1e8 d8e8 c4c5 d6d5 d4f6 d5d2 d1d2 g7f6 d2d7');
INSERT INTO evaluations VALUES(30213,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(30213,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(30213,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(30213,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(30213,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(30213,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(30213,36,'cp 60','g4f3 e2f3 c7c6 e1g1 f8d6 c3e4 f6e4 f3e4 b8d7 c2c3 e8g8 f1e1 d7f6 e4c2 d6c7 a2a4 a7a5 c2b3 d8e7 d1f3 a8d8 g2g3 d8d7 c1e3 f6d5 e3d2 f8d8 a1d1 c7b6 d2c1 d5f6 h3h4 c6c5 d4c5 d7d1 e1d1 d8d1 b3d1 b6c5 c1f4');
INSERT INTO evaluations VALUES(43413,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(43413,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(43413,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(43413,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(43413,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(43413,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43413,36,'cp 32','c2c4 b8c6 d4c6 b7c6 b1c3 f6d7 d3c2 g7g6 d1d4 h8g8 e1g1 f8g7 d4d2 d8a5 a1b1 g7c3 b2c3 g6g5 c4c5 a5c5 c2a4 a6a5 d2d4 c8a6 d4c5 d6c5 a4c6 a8c8 c6b5 a6b5 b1b5 f7f6 b5a5 e8f7 f2f4 g5f4 c1f4 g8d8 f4e3 d7e5 e3c5');
INSERT INTO evaluations VALUES(43402,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(43402,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(43402,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43402,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(43402,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(43402,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43402,36,'cp 30','b8c6 d4c6 b7c6 e1g1 g7g6 d1a4 c8d7 a4c2 c6c5 h2h3 f8g7 c1g5 e8g8 b1c3 d7c6 a1d1 f6d7 f2f4 g7d4 g1h1 f8e8 b2b3 d8c8 g5h4 e7e6 d3e2 c8b7 e2f3 f7f6 h4e1 a8d8 c3e2 e6e5 e1a5 d8b8 a5c3 e8e7 e2g3 b8e8 c2d2');
INSERT INTO evaluations VALUES(52149,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(52149,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(52149,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(52149,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(52149,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(52149,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(52149,36,'cp 176','d4e5 d8e7 e5e7 f8e7 c1f4 b8c6 b1c3 g8f6 f2f3 a7a6 e1c1 d7d6 f4d6 e7d6 d1d6 e8e7 d6d2 c8e6 g1e2 h8d8 d2d8 a8d8 h2h4 h7h6 h4h5 b7b5 e2f4 c5c4 b2b3 b5b4 c3d1 c4c3 d1e3 a6a5 f1b5 c6d4 b5c4 e6c4 e3c4 e7f8 c4a5 d8a8 h1d1 a8a5 d1d4 a5a2 d4b4');
INSERT INTO evaluations VALUES(52096,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(52096,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(52096,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(52096,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(52096,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(52096,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(52096,36,'cp 163','d8e7 e5e7 f8e7 c1f4 g8f6 b1c3 b8c6 f2f3 b7b6 e1c1 c8b7 g1e2 e8c8 h2h4 d7d6 f4g3 f6h5 g3e1 h5f6 h4h5 h7h6 e2f4 h8e8 f1c4 c6e5 c4b5 b7c6 a2a4 c8b7 e1g3 a7a6 b5e2 e8g8 f4d5 f6d5 c3d5 c6d5 d1d5 e7g5 f3f4');
INSERT INTO evaluations VALUES(34275,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(34275,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(34275,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(34275,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(34275,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(34275,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(34275,36,'cp 171','e5e7 f8e7 b1c3 g8f6 c1f4 b8c6 f2f3 b7b6 c3b5 e8g8 b5c7 f6h5 f4e3 a8b8 e1c1 h5f6 g1e2 d7d6 e2c3 f8d8 g2g4 d8d7 c7d5 f6d5 c3d5 e7f8 f1b5 c8b7 g4g5 d7d8 b5c6 b7c6 h2h4 c6d5 d1d5 d8e8 h4h5 a7a6 a2a3 b8b7 h5h6 g7h6 g5h6');
INSERT INTO evaluations VALUES(57007,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57007,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57007,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(57007,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57007,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57007,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(57007,36,'cp 6','g8f6 g1f3 d7d5 b2b3 f8e7 c1b2 e8g8 b1c3 c7c5 c4d5 f6d5 f1e2 b8c6 e1g1 d5c3 b2c3 b7b6 d2d4 c5d4 f3d4 c6d4 c3d4 c8b7 e2f3 b7f3 d1f3 e7f6 d4f6 d8f6 f3f6 g7f6 g2g3 a8c8 a1c1 f8d8 c1c8 d8c8 f1d1 c8c2 a2a4');
INSERT INTO evaluations VALUES(49382,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(49382,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(49382,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(49382,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(49382,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(49382,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(49382,36,'cp 14','g1f3 d7d5 b2b3 f8e7 d2d4 e8g8 b1c3 c7c5 d4c5 d5c4 f1c4 d8d1 e1d1 e7c5 c1b2 b8c6 d1e2 b7b6 h1d1 c8b7 e2f1 c5e7 c4e2 f8d8 d1d8 a8d8 a1c1 d8c8 c3a4 c6b4 c1c8 b7c8 a4c3 c8b7 a2a3 b4d5 c3a4 a7a6 b3b4');
INSERT INTO evaluations VALUES(49383,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(49383,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(49383,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(49383,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(49383,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(49383,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(49383,36,'cp 6','d7d5 b2b3 b7b6 c1b2 c8b7 g2g3 f8e7 c4d5 d8d5 f1g2 c7c5 e1g1 e8g8 b1c3 d5d6 d2d4 f8d8 a1c1 b8d7 d4c5 d7c5 d1d6 e7d6 f1d1 a8c8 c3b5 d6f8 f3e5 d8d1 c1d1 b7g2 g1g2 c5e4 e5c4 a7a6 b5d6 e4d6 c4d6');
INSERT INTO evaluations VALUES(9750,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(9750,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9750,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9750,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(9750,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9750,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9750,36,'cp 0','d7d6 d2d3 g7g5 f3h2 c6a5 c4d5 h8g8 c1e3 c7c6 e3c5 d6c5 d5b3 d8e7 c3e2 g5g4 h3h4 c5c4 d3c4 f6e4 d1e1 e7c5 e2c1 a5b3 c1b3 c5c4 b3a5 c4c2 a1c1 c2d3 c1d1 d3c2');
INSERT INTO evaluations VALUES(8821,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8821,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8821,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8821,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8821,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8821,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8821,36,'cp 0','d2d3 g7g5 f3h2 c6a5 c4d5 h8g8 c1e3 c7c6 e3c5 d6c5 d5b3 d8e7 c3e2 g5g4 h3h4 c5c4 d3c4 f6e4 d1e1 e7c5 e2c1 a5b3 c1b3 c5c4 b3a5 c4c2 a1c1 c2d3 c1d1 d3c2');
INSERT INTO evaluations VALUES(11430,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(11430,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(11430,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(11430,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(11430,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(11430,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(11430,36,'cp 13','c2c3 g7g6 h2h3 f8g7 e1g1 h7h6 f1e1 e8g8 a4c6 b7c6 d3d4 f6h5 d4e5 d6e5 b1d2 f8e8 d1a4 d8f6 f3h2 a8b8 d2c4 f6e7 c4a5 c8d7 a5b3 c6c5 a4a6 h5f4 a6c4 d7h3 c1f4 h3e6 c4c5 e7c5 b3c5 e5f4 e1b1 e6c8 h2f3 g6g5 f3d4 g5g4 a2a4 g4g3');
INSERT INTO evaluations VALUES(11427,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11427,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(11427,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(11427,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(11427,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(11427,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(11427,36,'cp 8','g7g6 e1g1 f8g7 f1e1 e8g8 h2h3 h7h6 d3d4 b7b5 a4c2 c8b7 d4e5 c6e5 f3e5 d6e5 d1d8 f8d8 b1d2 b7c6 a2a4 f6d7 d2b3 d7b6 b3a5 c6e8 b2b3 g7f8 c1e3 d8b8 e1d1 g8g7 f2f3 f7f6 a4b5 a6b5 c2d3 f8d6 d3e2 h6h5 b3b4 b6a4 a1a3');
INSERT INTO evaluations VALUES(36623,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36623,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36623,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(36623,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36623,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(36623,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36623,36,'cp 13','h2h3 c7c5 e1g1 d8b6 d1c1 b8c6 c2c3 c8e6 b1d2 c5d4 e3d4 a8c8 f1e1 e6d5 f4g5 b6c7 d2f1 c7a5 c1f4 f6e4 f1e3 e4g5 f3g5 h7h6 e3d5 a5d5 g5e4 d5f5 f4f5 g6f5 e4d2 f8d8 e2b5 e7e6 b5c6 c8c6');
INSERT INTO evaluations VALUES(44790,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(44790,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(44790,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(44790,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(44790,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(44790,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(44790,36,'cp 49','a7a6 c1e3 h7h5 f1c4 b7b5 c4b3 c8b7 d1d2 f8g7 e1g1 b8d7 a2a4 b5a4 c3a4 d8c7 a1c1 e8g8 c2c4 a8b8 f1d1 f8c8 c4c5 d7c5 e4e5 f6e8 e5e6 f7f5 b3a2 c7d8 a4c3 g7f6 e3g5 d8b6 g5f6 e8f6 d2g5');
INSERT INTO evaluations VALUES(43346,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(43346,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(43346,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(43346,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(43346,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(43346,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(43346,36,'cp 70','c1e3 h7h5 f1e2 e7e5 d4b3 c8e6 d1d2 b8d7 e1c1 d8c7 c1b1 b7b5 f3f4 f8e7 h2h3 a8c8 h1e1 e8g8 a2a3 c7b7 b3a5 b7c7 f4f5 g6f5 e4f5 e6f5 g2g4 f5h7 a5b3 b5b4 a3b4 d6d5 g4g5 f6e4 c3e4 h7e4 e2d3 e4d3 d2d3 e7b4 c2c3 b4c3');
INSERT INTO evaluations VALUES(43345,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(43345,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(43345,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(43345,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(43345,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(43345,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(43345,36,'cp 54','h7h5 d1d2 e7e5 d4b3 c8e6 f1e2 b8d7 e1c1 f8e7 c1b1 a8c8 h1f1 b7b5 f3f4 e8g8 a2a3 e5f4 e3f4 f8e8 h2h3 d8b6 f4e3 b6b7 g2g4 h5g4 b3d4 e6c4 h3g4 e7f8 f1h1 b5b4 a3b4 c4e2 d4e2 c8c4 g4g5');
INSERT INTO evaluations VALUES(46640,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46640,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46640,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(46640,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(46640,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(46640,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(46640,36,'cp 28','g1f3 f8g7 d2d4 e8g8 c4b3 c8g4 h2h3 g4f3 d1f3 b8c6 f3d1 a7a6 c1e3 e7e5 d4e5 d6e5 d1d8 f8d8 f2f3 c6d4 e1c1 d8d7 d1d2 a8d8 h1d1 b7b5 c3d5 f6d5 b3d5 d7d5 e4d5 d4f5 d1e1 f5e3 e1e3 g7h6 e3e5 h6f4 e5e2 d8d5 c1d1 f4d2 e2d2 d5c5 d2d7 g8g7 c2c3 h7h6');
INSERT INTO evaluations VALUES(46717,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(46717,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(46717,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(46717,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(46717,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46717,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(46717,36,'cp 39','c7c6 d1d2 d8a5 g1f3 b8d7 f1e2 f8g7 e1g1 e8g8 f4h6 e7e5 h6g7 g8g7 a2a4 e5d4 f3d4 d6d5 d4b3 a5b4 e4d5 f6d5 c3d5 b4d2 b3d2 c6d5 d2b3 f8e8 f1e1 d7e5 f2f4 e5c6 e2f3 c8f5 c2c3 f5e4 f3e4 d5e4');
INSERT INTO evaluations VALUES(43275,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43275,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(43275,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(43275,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43275,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43275,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43275,36,'cp 0','d2d4 e8g8 f1e2 b7b6 g1f3 c7c5 e1g1 c5d4 f3d4 c8b7 c2c4 b8c6 d4c6 b7c6 b1c3 d7d5 a1c1 d5c4 e2c4 a7a6 b2a3 b6b5 c4e2 d8c7 c3d5 f6d5 d1d5 c7h2 g1h2 c6d5 a3e7 f8e8 e7b4 e8c8 f1d1 c8c1 d1c1 d5e6 e2f3 a8c8 c1c8 e6c8 f3d5');
INSERT INTO evaluations VALUES(43222,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43222,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43222,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43222,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(43222,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43222,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43222,36,'cp 0','b7b6 c2c4 c8b7 g1f3 c7c5 f1e2 c5d4 d1d4 e8g8 e1g1 b8a6 d4d2 d7d5 f1d1 d5c4 d2e1 d8c8 e2c4 b7f3 g2f3 a6c5 b1c3 c8f5 e1e2 f5g5 g1h1 g5h4 h1g1');
INSERT INTO evaluations VALUES(47808,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(47808,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(47808,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(47808,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(47808,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47808,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(47808,36,'cp 13','f8e7 f1e1 c7c5 b1c3 e8g8 d2d3 b8c6 a2a4 c8e6 c1g5 c6b4 h2h3 d8d7 f3h2 h7h6 g5d2 b4c6 c3d5 e6d5 e4d5 c6b4 d3d4 e5e4 d4c5 d6c5 d2c3 a8d8 c3f6 e7f6 c2c3 b4d5 e1e4 d5b6 d1e2 b6c4 e2c4 b7b6 a1e1 f8e8 h2f1 e8e4');
INSERT INTO evaluations VALUES(56278,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(56278,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(56278,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(56278,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(56278,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(56278,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56278,36,'cp -18','f8d6 e1g1 g8e7 f1e1 h7h6 d3f1 e8g8 d2d4 c8e6 b2b3 c7c6 c1a3 d6a3 b1a3 d8d6 a3b1 b8d7 c2c4 e6g4 b1d2 f8e8 d1c2 g4f5 c2b2 e7g6 g2g3 d7f8 e1e8 a8e8 f1g2 e8d8 a1e1 d6d7 h2h4 f5g4 a2a3 b7b6 b2c3');
INSERT INTO evaluations VALUES(38631,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(38631,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(38631,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(38631,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(38631,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(38631,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(38631,36,'cp -12','e1g1 g8e7 f1e1 e8g8 d3f1 h7h6 d2d4 c8e6 b2b3 c7c6 c1a3 d6c7 c2c4 f8e8 b1c3 b8d7 a1c1 d7f6 h2h3 e7f5 f1d3 c7b6 c4d5 f6d5 c3d5 d8d5 a3c5 a8d8 d3e4 d5d7 d1c2 b6c5 d4c5 f5d4 f3d4 d7d4 e1e2 e6d5 e4d5 e8e2 c2e2 d4d5');
INSERT INTO evaluations VALUES(47462,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(47462,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(47462,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(47462,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(47462,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(47462,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(47462,36,'cp 100','d2d4 c8f5 b1d2 d5b6 e2e4 f5g6 h2h4 h7h6 f1d3 b8d7 h4h5 g6h7 e1g1 e7e6 a2a4 a7a5 d2b3 f8b4 c1d2 d8e7 d1e1 e8g8 b3a5 c7c5 a5b7 b6a4 b2b3 d7f6 d2b4 c5b4 b3a4 e7b7 a4a5 h7e4 a5a6 b7d5 d3e4 f6e4 e1b4 f8b8 b4a5 d5a5 a1a5');
INSERT INTO evaluations VALUES(49534,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(49534,31,'cp -178',NULL);
INSERT INTO evaluations VALUES(49534,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(49534,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(49534,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(49534,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(49534,36,'cp -136','d7d5 e4d5 f6d5 c4d5 d8d5 d2d3 f8c5 d1e2 b8c6 b1c3 d5e6 c1e3 c6d4 e3d4 c5d4 c3b5 d4b6 e1c1 c8d7 b5c3 e8c8 e2e1 f7f5 c1b1 h8e8 g1e2 h7h6 h2h4 c8b8 a2a3 a7a5 h4h5 e6f7 e1h4 b6e3 h4g3 d7c6 f3f4 e3d4 d1e1 d4c3 e2c3 e5f4 g3f4');
INSERT INTO evaluations VALUES(48785,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(48785,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(48785,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(48785,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(48785,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(48785,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(48785,36,'cp -135','e4d5 f6d5 c4d5 d8d5 d2d3 f8c5 d1e2 b8c6 b1c3 d5e6 c1e3 c6d4 e3d4 c5d4 c3b5 d4b6 b5c3 e8g8 e1c1 b6d4 e2e1 e6a6 g1e2 c7c5 c1b1 c8e6 a2a3 a8c8 c3e4 h7h6 e2d4 c5d4 e1b4 a6c6 c2c4 d4c3 e4c3 f8d8 b4b5 e6f5 b5c6 c8c6 h1e1 f5d3 b1a1');
INSERT INTO evaluations VALUES(51840,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(51840,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(51840,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(51840,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(51840,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51840,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51840,36,'cp 16','f8c5 g1f3 d7d6 d2d4 e5d4 f3d4 g8f6 f1e2 e8g8 e1g1 f8e8 c1g5 b8d7 d4b3 c5b4 f2f3 a7a5 c3a4 d6d5 c4d5 c6d5 e2b5 d5e4 f3e4 e8e4 g5f6 g7f6 d1d3 d8e7 a1d1 d7f8 a2a3 c8g4 d3g3 f8g6 a3b4 g4d1 f1d1 a5b4');
INSERT INTO evaluations VALUES(37656,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(37656,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(37656,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(37656,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(37656,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(37656,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(37656,36,'cp 30','g1f3 d7d6 d2d4 e5d4 f3d4 g8f6 d4b3 c5b4 f1d3 e8g8 e1g1 f8e8 c1g5 b8d7 c3a4 h7h6 g5h4 d7e5 a2a3 b4c5 d3c2 c8g4 d1e1 c5b6 a4b6 a7b6 f2f3 g4d7 e1e2 d6d5 c4d5 c6d5 b3d4 e5g6 h4g3 f6h5 e2f2 g6f4 g3f4 h5f4');
INSERT INTO evaluations VALUES(3373,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3373,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(3373,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(3373,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(3373,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3373,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3373,36,'cp 0','c6a5 c4b3 a7a6 c1d2 b7b5 c3d5 a5b3 d5e7 d8e7 a2b3 f6d7 f3h2 f7f5 e4f5 f8f5 f2f4 c8b7 d1e2 e7f7 f4e5 d7e5 f1f5 f7f5 a1f1 f5g6 h2g4 a8e8 g4e3 c7c5 f1f4 b5b4 e2f2 a6a5 e3f5 h7h6 f5h4 g6e6 h4f5');
INSERT INTO evaluations VALUES(3448,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3448,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3448,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3448,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3448,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3448,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3448,36,'cp 0','c4b3 c7c6 c3e2 a5b3 a2b3 f6h5 e2c3 h5f6');
INSERT INTO evaluations VALUES(3450,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3450,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3450,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3450,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3450,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3450,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3450,36,'cp 0','a5b3 a2b3 c7c6 c3e2 f6h5 e2c3');
INSERT INTO evaluations VALUES(3432,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3432,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3432,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3432,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3432,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3432,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3432,36,'cp 0','a2b3 c7c6 c3e2 f6h5 e2c3');
INSERT INTO evaluations VALUES(3431,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3431,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3431,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3431,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3431,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3431,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3431,36,'cp 0','c7c6 c3e2 f6h5 e2c3');
INSERT INTO evaluations VALUES(10329,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(10329,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(10329,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(10329,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(10329,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(10329,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(10329,36,'cp -68','f6e4 a2a3 b4c3 b2c3 d7d5 c3c4 c8e6 c4d5 d8d5 d1d5 e6d5 c1e3 e8c8 e1c1 h8e8 d1e1 h7h6 h2h4 c6e5 f3e5 e8e5 e3a7 b7b6 f2f3 c8b7 a7b6 b7b6 f3e4 f7f5 f1d3 f5e4 e1e3 d8a8 d3e2 b6c5 h1f1 c7c6 f1d1 g7g6 e3c3 c5d6 c1d2 a8a4 d2e3 e5e8');
INSERT INTO evaluations VALUES(10663,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(10663,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(10663,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(10663,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(10663,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(10663,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(10663,36,'cp -47','a2a3 b4c3 b2c3 d7d5 c3c4 c8e6 c4d5 d8d5 d1d5 e6d5 c1e3 e8c8 e1c1 h8e8 h1g1 e4c3 d1d3 c3a4 f1e2 d8d7 g1e1 a4b6 e3f4 d7e7 f3d4 c6e5 d3e3 e5d7 d4b5 e7e3 f2e3 d7e5 b5a7 c8b8 a7b5 g7g6 e3e4 d5c4 e2c4 b6c4 e1d1 b7b6 d1d4 e8c8 a3a4 f7f6 f4e5 c4e5 b5c3 b8b7');
INSERT INTO evaluations VALUES(30394,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(30394,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(30394,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(30394,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(30394,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(30394,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(30394,36,'cp 71','c2c4 d5c4 f1c4 e7e6 b1c3 f8d6 g1e2 g8f6 f2f3 b8c6 e3e4 d6f4 e2f4 d8d6 d1d2 f5g6 e4e5 d6b4 c4e6 f7e6 e5f6 c6d4 f6g7 h8g8 f4g6 h7g6 e1c1 e8c8 d2g5 g8g7 d1d2 b4e7 g5e5 d4c6 d2d8 c8d8 e5e3 d8c8 c3e4 c8b8 h1d1 b7b6 e3g5 c6b4 g5e7');
INSERT INTO evaluations VALUES(30389,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(30389,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(30389,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(30389,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(30389,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(30389,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(30389,36,'cp 62','d5c4 f1c4 e7e6 b1c3 f8d6 g1e2 g8f6 f2f3 b8c6 e3e4 d6f4 e2f4 d8d6 d1d2 f5g6 e4e5 d6b4 c4e6 f7e6 e5f6 e8c8 f6g7 h8g8 f4e6 c6d4 e6d8 d4c2 e1f1 c2a1 d8e6 a1c2 d2f4 g6d3 f1f2 b4b6 e6d4 b6d4 f4d4 c2d4 h1d1 d4f3 f2f3 d3c4 h2h4 g8g7 g2g4 c4e6 g4g5 g7f7 f3g3 f7f5 d1e1 c8d7 c3e4');
INSERT INTO evaluations VALUES(37434,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37434,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37434,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37434,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37434,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37434,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37434,36,'cp 1','c7c5 f1e2 d8b6 d1c1 b8c6 c2c3 d7d5 e1g1 c8f5 b1d2 a8c8 d2b3 f6e4 f1d1 c5d4 e3d4 a7a5 f4e3 b6d8 f3h4 e4d6 a2a4 f5e4 e3g5 d8c7 g5f4 c7b6 b3c5 f8d8 c1e3 e7e5 d4e5 c6e5 c5e4 b6e3');
INSERT INTO evaluations VALUES(35709,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(35709,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(35709,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(35709,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35709,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35709,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35709,36,'cp 22','f1e2 b8c6 c2c3 d8b6 d1c1 d7d6 e1g1 c8e6 b1d2 c5d4 e3d4 a8c8 f1e1 e6d5 f4g5 b6d8 c1d1 h7h6 g5f4 d8b6 b2b3 f8e8 f4e3 b6c7 c3c4 d5f3 d2f3 d6d5 e2d3 e7e6 a1c1 c7a5 a2a4 e8d8 c4c5 f6e4');
INSERT INTO evaluations VALUES(38559,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(38559,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(38559,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(38559,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(38559,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(38559,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(38559,36,'cp 90','f1d3 a7a6 c1g5 g8e7 d1d2 h7h6 g5e3 b8d7 e1g1 c7c5 a2a4 b7b6 d4d5 e6d5 e4d5 d7f6 h2h3 e8f8 a1e1 f8g8 e3f4 g8h7 e1e2 c8b7 d3c4 h8e8 f1e1 d8d7 g2g4 g6g5 d2d3 e7g6 f4g3 e8e2 e1e2 h7h8 e2e1 h8g8 f3d2 a8f8 e1e3 g6f4 g3f4 g5f4');
INSERT INTO evaluations VALUES(54920,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(54920,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(54920,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54920,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(54920,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(54920,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(54920,36,'cp 35','e7e6 e2e3 g8f6 c2c4 f8d6 f4d6 c7d6 b1c3 d5c4 f1c4 a7a6 e1g1 e8g8 a2a4 b8c6 d4d5 e6d5 c3d5 c8e6 d5f6 d8f6 c4e6 f6e6 a1a3 f8d8 h2h3 d6d5 a3d3 e6e4 f3d4 a8c8 d1d2 c6e5 d3b3 d8d7 b3c3 d7c7 c3c7 c8c7 f1d1 e5c6');
INSERT INTO evaluations VALUES(55125,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(55125,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(55125,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(55125,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(55125,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55125,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(55125,36,'cp -10','e7e5 f1g2 b8c6 g1f3 c8e6 e1g1 g8f6 a2a3 a7a5 b2b3 f8d6 c1b2 e8g8 e2e3 f8e8 b1d2 e6f5 f1e1 f5g6 h2h3 h7h6 e3e4 d5e4 d3e4 d8d7 f3h4 g6h7 d2c4 d6f8 h4f5 h7f5 e4f5 d7f5 g2c6');
INSERT INTO evaluations VALUES(5537,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(5537,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(5537,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(5537,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(5537,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(5537,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(5537,36,'cp 68','g8f6 f2f3 e8g8 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 d5c3 d2c3 g7h6 d4e3 h6e3 c3e3 d8b6 e3e7 c8e6 e7a3 a8d8 f1d3 d8d5 b2b3 f8d8 d1e1 c6c5 d3c4 d5d2 c4e6 f7e6 e1e4 d2g2 c1b1 g2d2 a3c1 d8d4 e4d4 d2d4 h2h4 d4d5 h4h5 d5h5 h1h5 g6h5 c1g5 g8h8');
INSERT INTO evaluations VALUES(8336,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(8336,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(8336,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(8336,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(8336,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(8336,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(8336,36,'cp 75','f2f3 e8g8 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 c8e6 f1c4 d8c7 c4d5 c6d5 h2h4 a8c8 h4h5 g6g5 h1e1 h7h6 e1e5 f8d8 d1d2 c7d7 d2e2 d7d6 d4d3 g8h8 c3b5 d6a6 e5e6 f7e6 d3g6 a6b5 g6h6 h8g8 h6e6 g8f8 e6e7 f8g8');
INSERT INTO evaluations VALUES(29483,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(29483,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(29483,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(29483,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(29483,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(29483,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(29483,36,'cp -15','b8a6 a2a3 c7c6 b5c3 e7e6 e2e3 a6b8 f1d3 f8e7 d3f5 e6f5 d1d3 g7g6 g1f3 b8d7 c3e2 e8g8 e1g1 f6e4 h2h3 f8e8 c2c4 d5c4 d3c4 d7b6 c4c2 b6d5 f4e5 a8c8 f1d1 f7f6 e5h2 e7d6 h2d6 e4d6 a3a4 g8g7 e2c3 a7a5 c3d5 c6d5');
INSERT INTO evaluations VALUES(25045,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(25045,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(25045,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25045,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(25045,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25045,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(25045,36,'cp 0','a2a3 e7e6 e2e3 c7c6 b5c3 a6b8 f1d3 f8e7 d3f5 e6f5 g1f3 b8d7 h2h3 a7a5 f3d2 h7h5 e1g1 e8g8 b2b3 h5h4 c3e2 f6h5 f4e5 h5f6');
INSERT INTO evaluations VALUES(40698,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(40698,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(40698,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(40698,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(40698,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(40698,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(40698,36,'cp 54','e8g8 c1f4 b8c6 h2h3 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 f6d7 c4e6 f7e6 f4e3 e7f6 d4c4 f8e8 f2f4 f6c3 c4c3 d7f6 e4e5 f6d5 c3d4 d8e7 d1d3 d5e3 d3e3 a8d8 f1e1 d6d5 e3g3 c7c5 d4d3');
INSERT INTO evaluations VALUES(36190,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(36190,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(36190,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(36190,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(36190,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(36190,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(36190,36,'cp 59','h2h3 b8c6 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 d7c6 f1e1 f6d7 c3d5 e7d8 b2b3 e6d5 d4d5 c6d5 d1d5 f8e8 f2f3 d7b6 d5d1 b6d7 g1f2 d8h4 g2g3 h4g5 a4a5 g8f8 h3h4 g5f4 g3f4 d7f6 f2e3 f6h5 b3b4 a8d8 d1d3 f7f5 e1d1 h5f6');
INSERT INTO evaluations VALUES(36188,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(36188,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(36188,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(36188,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(36188,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(36188,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(36188,36,'cp 57','b8c6 c1f4 f6d7 f4e3 d7b6 c4b3 c6d4 e3d4 c8e6 b3e6 f7e6 d1g4 e6e5 d4b6 a7b6 a1d1 d8c8 g4e2 c8e6 a2a4 f8f7 b2b3 a8f8 d1d3 c7c6 e2g4 e6g4 h3g4 f7f6 f2f3 f6e6 c3d1 d6d5 e4d5 c6d5 d3d5 e5e4 d5d4 e4f3');
INSERT INTO evaluations VALUES(48352,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48352,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48352,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48352,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48352,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48352,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48352,36,'cp 0','b7b6 b2b3 f8d6 c2c4 b8d7 e1g1 e8g8 a2a4 a7a5 b1c3 c8b7 c1b2 d6b4 c3a2 b4d6');
INSERT INTO evaluations VALUES(55252,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(55252,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(55252,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(55252,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(55252,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(55252,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(55252,36,'cp 71','g1f3 c7c5 e2e3 g7g6 b3b4 c5b4 f3d4 g8f6 f1e2 f8g7 a2a3 b4a3 b1a3 e8g8 d4b3 b8c6 d2d4 e7e5 d4e5 d8d1 a1d1 f6g4 a3b5 g7e5 b2a3 f8d8 d1d8 c6d8 h2h3 a7a6 b5d4 g4f6 e1g1 b7b5 c2c4 b5c4 e2c4 c8b7 d4f3 b7f3 g2f3');
INSERT INTO evaluations VALUES(55251,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(55251,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(55251,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(55251,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(55251,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(55251,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(55251,36,'cp 63','c7c5 e2e3 g7g6 b3b4 c5b4 a2a3 b4a3 b2d4 g8f6 b1a3 f8g7 f1e2 e8g8 c2c4 c8f5 e1g1 b8c6 d4b2 f5d3 a3b5 a7a6 b5d4 d3e2 d1e2 d8c7 d4c6 c7c6 f1c1 f8c8 a1b1 b7b6 b2a3 c6b7 d2d4 b6b5 c4b5 c8c1 a3c1 a6b5 e2b5');
INSERT INTO evaluations VALUES(53211,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(53211,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(53211,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(53211,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(53211,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(53211,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(53211,36,'cp 81','e2e3 g7g6 b3b4 c5b4 f3d4 g8f6 f1e2 f8g7 e1g1 e8g8 a2a3 b4a3 b1a3 f8e8 d4f3 b7b6 c2c4 c8b7 d2d4 f6e4 f3d2 e4d2 d1d2 e7e6 f1c1 a7a6 a3b1 b8d7 b1c3 d8g5 e2f1 b7f3 g1h1 f3c6 d2e2 c6b7 f2f3 g5h4 e2d2 h4g5 c3d1 g5e7 b2a3');
INSERT INTO evaluations VALUES(53210,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(53210,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(53210,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(53210,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(53210,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(53210,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(53210,36,'cp 71','g7g6 b3b4 c5b4 b2d4 g8f6 a2a3 b4a3 a1a3 b8c6 d4b2 a7a6 c2c4 c8f5 f1e2 e7e5 f3e5 f8a3 e5c6 b7c6 b2a3 d8a5 d1b3 f6d7 a3d6 a5b6 b3b6 d7b6 b1c3 e8d7 d6e5 h8e8 e5f4 a6a5 d2d3 a5a4 e1d2 a4a3 e3e4 f5e6 h1a1 b6a4 a1a3 a4c5 a3a8 c5b3 d2e3 e8a8 d3d4');
INSERT INTO evaluations VALUES(12517,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(12517,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(12517,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(12517,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(12517,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(12517,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(12517,36,'cp 46','c1g5 f8c5 b1c3 d7d6 g5f6 d8f6 b5c7 e8d8 f2f3 a8b8 d1d2 h7h5 c7d5 f6h6 f3f4 e5f4 e1c1 f4f3 g2f3 h6d2 d1d2 a7a6 f3f4 d8e8 a2a3 e8f8 d2d3 c8e6 f1h3 h8h6 h1d1 e6g4 d1d2 g4e6 c1b1 c5a7 f4f5 e6d5 c3d5 b8e8 d3b3 e8b8 h3f1 c6e5 f1e2');
INSERT INTO evaluations VALUES(12514,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(12514,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(12514,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(12514,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12514,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(12514,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(12514,36,'cp 44','f8c5 b1c3 d7d6 g5f6 d8f6 b5c7 e8d8 f2f3 a8b8 d1d2 h7h5 c7d5 f6h6 f3f4 e5f4 e1c1 f4f3 g2f3 h6d2 d1d2 a7a6 f3f4 h8h6 f1g2 c5a7 d2d3 d8e8 g2h3 e8f8 a2a3 h5h4 h3c8 b8c8 d3h3 c6e7 d5e7 f8e7 h1d1 c8c4 h3d3 e7f8 f4f5 b7b5 b2b4 h4h3 d3g3 g7g6');
INSERT INTO evaluations VALUES(37730,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(37730,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(37730,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(37730,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(37730,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(37730,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(37730,36,'cp 107','e2e4 d7d6 f1e2 g8f6 g1f3 e8g8 e1g1 e7e5 d4d5 a7a5 c1e3 f6g4 e3g5 f7f6 g5d2 b8a6 h2h3 g4h6 a2a3 h6f7 b2b4 f6f5 d1b3 a8b8 f1b1 a5b4 a3b4 c6c5 b4c5 a6c5 b3b6 f5e4 b6d8 f8d8 f3g5 e4e3 d2e3 c8f5 g5e6 f5b1 e6d8');
INSERT INTO evaluations VALUES(37732,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(37732,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(37732,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(37732,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(37732,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(37732,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(37732,36,'cp 78','d7d6 f1e2 g8f6 g1f3 e8g8 e1g1 e7e5 d4d5 a7a5 c1g5 h7h6 g5e3 f6g4 e3d2 b8a6 f3e1 f7f5 e2g4 f5g4 d2e3 g8h7 c3a4 c6c5 f2f3 g4f3 e1f3 c8d7 d1b3 a8b8 a4c3 a6b4 b3d1 b4a6 d1e2 a6c7 f3d2 f8f1 a1f1 d8e7 a2a4 b8f8 c3b5 c7b5');
INSERT INTO evaluations VALUES(37938,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(37938,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(37938,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(37938,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(37938,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(37938,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(37938,36,'cp 90','g1f3 g8f6 f1e2 e8g8 e1g1 e7e5 d4d5 a7a5 c1g5 h7h6 g5e3 f6g4 e3d2 b8a6 f3e1 f7f5 e2g4 f5g4 d2e3 a6c7 e1d3 c6d5 c4d5 c8d7 a2a4 b7b5 e3b6 d8b8 b6c7 b8c7 a4b5 f8b8 a1c1 c7b6 b2b4 a5b4 d3b4 d7b5 c3b5');
INSERT INTO evaluations VALUES(37937,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(37937,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(37937,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(37937,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(37937,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(37937,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(37937,36,'cp 75','g8f6 f1e2 e8g8 e1g1 e7e5 d4d5 a7a5 h2h3 c8d7 a2a3 b8a6 c1e3 a6c5 d1c2 c6d5 e3c5 d6c5 c4d5 f6e8 a3a4 e8d6 e2b5 c5c4 c2e2 a8c8 f1c1 f8e8 c1c2 d8c7 a1c1 g7f8 b5d7 c7d7 c1d1 f8h6 h3h4 f7f5 c3b5 f5e4 f3d2 d6b5 a4b5 h6d2 d1d2 d7b5 e2e4');
INSERT INTO evaluations VALUES(40304,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(40304,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(40304,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(40304,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(40304,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(40304,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(40304,36,'cp 85','f1e2 e8g8 e1g1 e7e5 d4d5 b8a6 c1e3 f6g4 e3g5 f7f6 g5h4 c6c5 g1h1 g4h6 f3e1 b7b6 g2g4 c8d7 f2f3 h6f7 e1g2 a6c7 a2a3 a8b8 b2b4 g7h6 h4f2 g8h8 b4c5 d6c5 a3a4 c7a6 h2h4 a6b4 a4a5 f6f5 a5b6');
INSERT INTO evaluations VALUES(40303,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(40303,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(40303,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(40303,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(40303,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(40303,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(40303,36,'cp 78','e8g8 e1g1 e7e5 d4d5 b8a6 c1e3 f6g4 e3g5 f7f6 g5h4 c6c5 f3e1 g4h6 g1h1 b7b6 g2g4 h6f7 e1g2 c8d7 f2f3 g7h6 h4e1 f7g5 d1c1 g8g7 e1d2 g5f7 d2h6 f7h6 a2a3 a6c7 h2h4 h6f7 c1e3 g7h8 e2d3 g6g5 b2b4 a7a5 b4a5');
INSERT INTO evaluations VALUES(35911,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(35911,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(35911,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(35911,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(35911,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(35911,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(35911,36,'cp 67','e1g1 e7e5 d4d5 b8a6 f3d2 f6e8 d2b3 c8d7 d5c6 d7c6 c1e3 b7b6 d1d2 a6c7 a2a4 f7f5 f2f3 c7e6 a4a5 g8h8 f1d1 a8b8 a5b6 a7b6 a1a7 b8a8 a7a8 d8a8 e4f5 g6f5 c3d5 e6c5 b3c5 b6c5 b2b4 c6d5 c4d5 c5b4 d2b4 a8d8 b4b1 d8d7 e2d3');
INSERT INTO evaluations VALUES(35910,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(35910,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(35910,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(35910,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(35910,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(35910,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(35910,36,'cp 70','e7e5 d4d5 b8a6 c1e3 f6g4 e3g5 f7f6 g5h4 c6c5 g1h1 b7b6 f3e1 g4h6 g2g4 a6c7 e1g2 h6f7 e2d3 c8d7 h4g3 a7a6 f2f3 b6b5 a2a4 b5b4 c3e2 c7e8 b2b3 g7h6 h2h4 e8g7 g3h2 a8b8 d1c2 b8b7 f1g1 g8h8 a1e1');
INSERT INTO evaluations VALUES(44940,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(44940,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(44940,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(44940,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(44940,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(44940,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(44940,36,'cp -12','c8f5 f1d3 e7e6 d1e2 f5d3 c2d3 c6c5 d4c5 f8c5 d3d4 c5d6 e2b5 d8d7 b5d7 b8d7 e1e2 a7a6 c1d2 e8e7 f3e1 f6e4 e1d3 h7h5 h2h4 f7f6 a1c1 a8c8 d2e1 e4c3 e1c3 e6e5 c3b4 d6b4 d3b4 e7d6 e2d2 c8c1 h1c1 g7g5 d4e5 d7e5 b4d3 g5h4 d3f4 e5c4 d2d3 c4b2 d3d4 b2c4 f4d5');
INSERT INTO evaluations VALUES(28992,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(28992,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(28992,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(28992,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(28992,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(28992,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(28992,36,'cp 0','f1d3 e7e6 d1e2 f5d3 c2d3 c6c5 d4c5 f8c5 d3d4 c5d6 e2b5 d8d7 b5d7 e8d7 f3e5 d6e5 d4e5 f6e4 c3e4 d5e4 e1e2 b8c6 h1d1 d7e7 c1d2 h8d8 d2c3 d8d1 a1d1 a8d8 d1d8 e7d8 f2f3 e4f3 g2f3 d8c7 e2d3 a7a6 f3f4 c7b6 d3c4 c6e7 e3e4 b6c6 c4d4 c6b5 b2b3 g7g6 d4e3 e7c6 c3d2 c6e7');
INSERT INTO evaluations VALUES(28995,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(28995,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(28995,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(28995,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28995,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28995,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28995,36,'cp -11','e7e6 d1e2 f5d3 c2d3 c6c5 d4c5 f8c5 d3d4 c5d6 e2b5 d8d7 b5d7 b8d7 c3b5 d6b8 c1d2 a7a6 a1c1 e8g8 b5c3 b8d6 e1e2 a8c8 h2h3 d7b6 c1c2 f6d7 h1c1 f7f6 f3e1 e6e5 b2b3 e5d4 e3d4 f8e8 d2e3');
INSERT INTO evaluations VALUES(52354,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(52354,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(52354,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(52354,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(52354,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(52354,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(52354,36,'cp -76','g8f6 h5e2 e7e5 g1f3 d7d5 f3e5 f8d6 d2d4 e8g8 b1c3 c6c5 c1e3 f8e8 f2f4 b8c6 e1c1 c5d4 e3d4 d6e5 d4e5 c6e5 f4e5 e8e5 e2f2 f6e4 c3e4 e5e4 f1d3 e4e5 h1e1 e5e1 d1e1 c8e6 c1b1 a7a6 a2a3 d8d6 g2g4 h7h6 f2h4');
INSERT INTO evaluations VALUES(44460,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(44460,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(44460,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(44460,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(44460,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(44460,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(44460,36,'cp -77','h5h4 e7e5 f2f4 e5f4 h4f4 f8e7 b1c3 d7d5 e4e5 f6g4 f1e2 g4h6 g1f3 e8g8 d2d4 f7f6 e1g1 b8d7 e5f6 d7f6 e2d3 c6c5 f4g3 h6f5 d3f5 c8f5 c1e3 f6g4 e3f4 c5d4 f3d4 e7c5 c3e2 d8b6 c2c3 g4f6 b2b4 c5d4 e2d4 f6e4 g3e3 a8e8');
INSERT INTO evaluations VALUES(55168,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(55168,31,'cp -175',NULL);
INSERT INTO evaluations VALUES(55168,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(55168,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(55168,34,'cp -153',NULL);
INSERT INTO evaluations VALUES(55168,35,'cp -150',NULL);
INSERT INTO evaluations VALUES(55168,36,'cp -161','b8d7 e5c4 g8f6 d2d4 g7g6 b1c3 f8g7 a2a4 d7b6 a4a5 b6d5 c3d5 f6d5 c2c3 e8g8 f1e2 c8e6 e1g1 a8c8 c4e3 d5f4 e3c2 f4e2 d1e2 f7f5 f1d1 d8d6 g2g3 c7c6 c1f4 d6d7 c2e3 e6b3 d1e1 c6c5 d4c5 c8c5');
INSERT INTO evaluations VALUES(53594,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(53594,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(53594,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53594,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(53594,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(53594,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(53594,36,'cp 38','b1d2 g8f6 e2e3 c6c5 c2c3 d8b6 d1c2 f6h5 f4e5 b8c6 a2a4 c5d4 a4a5 b6d8 e3d4 a7a6 f1e2 c8d7 d2b3 f8e7 b3c5 e7c5 d4c5 c6e5 f3e5 h5f4 e5d7 d8d7 e1g1 f4e2 c2e2 h7h6 b2b4 e8g8 a1d1 f8e8 f1e1 e6e5 e2h5 f7f5 h5g6 e5e4 g6d6 d7d6');
INSERT INTO evaluations VALUES(12777,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(12777,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(12777,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(12777,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(12777,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(12777,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(12777,36,'cp -14','e7e6 g1f3 f8e7 e1g1 e8g8 f1e1 d7d6 a2a4 a7a6 h2h3 c8d7 c1f4 h7h6 e4e5 d6e5 f3e5 c6e5 f4e5 d7c6 a4a5 e7d6 d1e2 d6e5 e2e5 a8c8 e5g3 d8d4 e1e5 f8d8 c4b3 d8d7 b3a4 d7d6 a4c6 d6c6');
INSERT INTO evaluations VALUES(15273,30,'cp -225',NULL);
INSERT INTO evaluations VALUES(15273,31,'cp -223',NULL);
INSERT INTO evaluations VALUES(15273,32,'cp -231',NULL);
INSERT INTO evaluations VALUES(15273,33,'cp -223',NULL);
INSERT INTO evaluations VALUES(15273,34,'cp -230',NULL);
INSERT INTO evaluations VALUES(15273,35,'cp -241',NULL);
INSERT INTO evaluations VALUES(15273,36,'cp -232','d8d5 e1g1 c8f5 d2d3 h7h6 g5e4 e8c8 c1e3 f5e6 b1c3 d5d7 f2f3 b7b6 e4d2 g7g5 g1h1 f7f5 d2c4 f8g7 a2a4 d7f7 c4d2 a7a6 a4a5 b6b5 c3e2 g7f8 e2c1 h8g8 e3g1 h6h5 c2c4 b5c4 d1a4 c6b4 d2c4 e6c4 d3c4 f7c4 f1d1 d8d1 a4d1 g5g4 f3g4 c4g4');
INSERT INTO evaluations VALUES(17980,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(17980,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(17980,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(17980,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(17980,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(17980,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(17980,36,'cp 26','f8e7 d2d4 g8f6 e1g1 e8g8 h2h3 c8b7 f1e1 f8e8 b1d2 e7f8 b3c2 g7g6 d4d5 c6b8 b2b3 c7c6 c3c4 b8d7 d2f1 a8c8 f1g3 a6a5 c1e3 b7a6 c2d3 b5c4 d3c4 a6c4 b3c4 c6d5 c4d5 d7c5 f3d2 c5d3 e1f1 f6d7 a1b1');
INSERT INTO evaluations VALUES(37736,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(37736,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(37736,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(37736,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(37736,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(37736,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(37736,36,'cp 41','d7d5 f1e2 d5e4 c3e4 g8f6 e4f6 e7f6 e1g1 e8g8 c1e3 f8e8 c2c4 f6f5 d1c2 a7a5 h2h3 b8d7 a1d1 d7f6 f1e1 a5a4 e2d3 c8e6 e3f4 d8a5 f4d2 a5d8 a2a3 d8d7 d2f4 f6h5 f4e5 h5f6 f3g5 h7h6 g5f3 f6e4 d3e4');
INSERT INTO evaluations VALUES(37927,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(37927,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(37927,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(37927,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(37927,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(37927,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(37927,36,'cp 41','f1e2 d5e4 c3e4 g8f6 e4f6 e7f6 c2c4 e8g8 e1g1 f8e8 c1e3 a7a5 f1e1 b8a6 d1d2 c8f5 e1d1 g7f8 d4d5 f8b4 d2c1 c6d5 f3d4 f5d7 e2f3 a6c7 a2a3 b4d6 c1d2 d7c6 a1c1 a5a4 d4c6 b7c6 c4d5 c7d5 f3d5 c6d5 d2d5');
INSERT INTO evaluations VALUES(47203,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(47203,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47203,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47203,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(47203,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47203,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47203,36,'cp 0','c7c5 e2e3 c5d4 e3d4 d8c7 c2c3 b8c6 f1b5 a7a6 b5c6 b7c6 e1g1 c8f5 f3e5 e7e6 d1a4 a6a5 d2b3 a8a6 f1e1 f8d6 b3c5 d6c5 d4c5 e8g8 a4d4 c7b7 g2g4 f5g6 b2b3 f6d7 c1f4 d7e5 e1e5 f8e8 a1e1 g6e4 e5g5 e4g6');
INSERT INTO evaluations VALUES(57591,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(57591,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(57591,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(57591,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(57591,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(57591,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(57591,36,'cp 32','f5e4 d2d3 e4d3 f1d3 g8f6 c1g5 e7e6 c3e4 b8c6 e4f6 g7f6 d1h5 e8e7 g5h4 f8g7 f2f4 d8e8 h5f3 d7d6 g1e2 e8f7 e1g1 e7f8 f4f5 c6e5 f3g3 e5d3 f5e6 c8e6 c2d3 h8g8 e2d4 a8e8 g3f2 e6d7 d4f5 f7g6 a1c1 d7f5 f2f5 c7c6 h4f6 g6f5 f1f5 f8f7 f6c3 f7g6');
INSERT INTO evaluations VALUES(57618,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(57618,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(57618,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(57618,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(57618,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(57618,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(57618,36,'cp 46','c3e4 b8c6 d2d4 d7d5 e4g5 e7e5 d4e5 c6e5 d1e2 d8e7 c1d2 e5g6 e1c1 e7e2 g1e2 f8c5 e2f4 g6f4 d2f4 h7h6 g5f3 c7c6 f3e5 g8e7 e5d3 e8g8 f4h6 c5f2 h6g5 e7g6 d3f2 f8f2 g5e3 f2f7 e3c5 c8g4 d1e1 g4f5 f1e2 b7b6 c5d6 f7f6 d6g3 f5e4 h1f1 e4g2 f1f6 g7f6 e2h5');
INSERT INTO evaluations VALUES(57617,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(57617,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(57617,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(57617,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(57617,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(57617,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(57617,36,'cp 29','b8c6 d2d4 d7d5 e4g5 e7e5 d4e5 c6e5 d1e2 d8e7 c1d2 e5g6 h2h4 e7e2 g1e2 f8d6 h4h5 g6e5 e2f4 h7h6 f4g6 e5g6 h5g6 g8f6 g5f7 e8g8 f7d6 c7d6 f1d3 f6g4 f2f4 c8f5 e1g1 f5d3 c2d3 d5d4 f1e1 f8f6 a1c1 f6g6 c1c7 a8f8 c7b7 g4e3 d2e3 d4e3 e1e3 f8f4 b7a7 f4b4 b2b3 b4g4');
INSERT INTO evaluations VALUES(57665,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(57665,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(57665,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(57665,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(57665,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(57665,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(57665,36,'cp -56','f8g7 b1c3 c7c5 g2g3 d7d6 g1e2 b8c6 f1g2 a8b8 e1g1 b7b5 a1b1 e7e6 a2a3 g8e7 b3b4 c5b4 a3b4 a7a5 c3a2 e6e5 b4a5 d8a5 a2c3 e8g8 c3d5 e7d5 e4d5 c6e7 d2d3 c8f5 c1g5 a5c7 h2h3 h7h6 g5d2 f8c8 g3g4 f5d7 e2c3 b5b4 c3e2 c7c2 d1c2 c8c2');
INSERT INTO evaluations VALUES(39611,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(39611,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(39611,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(39611,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(39611,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(39611,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(39611,36,'cp -46','b1c3 c7c5 g2g3 b8c6 f1g2 d7d6 g1e2 a8b8 c1b2 b7b5 d1c1 e7e5 c3d5 g8e7 d5e7 c6e7 d2d3 e8g8 c1d2 b5b4 e1g1 d8b6 g1h1 f7f5 f2f4 c8b7 c2c4 b4c3 e2c3 e5f4 g3f4 g7d4 a1e1 b8e8 c3d1 d4b2');
INSERT INTO evaluations VALUES(39612,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(39612,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(39612,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(39612,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(39612,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(39612,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(39612,36,'cp -49','c7c5 g2g3 b8c6 f1g2 d7d6 g1e2 a8b8 a2a4 a7a6 e1g1 b7b5 a4b5 a6b5 a1a2 h7h5 h2h3 e7e6 d2d3 g8e7 c1e3 d8b6 e3g5 b6c7 d1d2 c6d4 d2e1 e7c6 e2d4 c5d4 c3e2 e8g8 g5d2 c8d7 f2f4 b8a8 a2a8 f8a8 f4f5 e6f5 e4f5');
INSERT INTO evaluations VALUES(5864,30,'cp 269',NULL);
INSERT INTO evaluations VALUES(5864,31,'cp 279',NULL);
INSERT INTO evaluations VALUES(5864,32,'cp 274',NULL);
INSERT INTO evaluations VALUES(5864,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(5864,34,'cp 279',NULL);
INSERT INTO evaluations VALUES(5864,35,'cp 271',NULL);
INSERT INTO evaluations VALUES(5864,36,'cp 269','b2b4 a5b6 f1d3 a7a6 a2a4 f7f6 d1c2 g8e7 d3h7 e7f5 h7g6 e8f8 g6f5 e6f5 e1g1 f6e5 d4e5 h8h5 a4a5 b6a7 b4b5 a6b5 b1a3 b5b4 a3b5 a7c5 c3b4 c5b4 c1d2 b4d2 c2d2 h5h6 f1c1 d8e7 h2h3 h6g6 d2d5 c8e6 d5d3 a8d8 d3a3 d8d7 a3e7 f8e7 a5a6 b7a6 a1a6 c6b4 a6a7 b4d3 c1c7 d3f4 g2g3 f4h3 g1g2 h3f4 g2f1 f4h3');
INSERT INTO evaluations VALUES(5863,30,'cp 295',NULL);
INSERT INTO evaluations VALUES(5863,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(5863,32,'cp 295',NULL);
INSERT INTO evaluations VALUES(5863,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(5863,34,'cp 297',NULL);
INSERT INTO evaluations VALUES(5863,35,'cp 278',NULL);
INSERT INTO evaluations VALUES(5863,36,'cp 287','a5b6 a2a4 a7a6 f1d3 f7f6 d1c2 g8e7 d3h7 e7f5 h7g6 e8f8 g6f5 e6f5 e1g1 f6e5 d4e5 h8h5 a4a5 b6a7 b4b5 a6b5 b1a3 b5b4 a3b5 a7c5 c3b4 c5b4 c1d2 b4d2 c2d2 f8g8 f1c1 a8a6 h2h3 h5h6 d2g5 d8g5 f3g5 g8f8 b5c7 a6a5 a1a5 c6a5 c7d5 h6c6 c1e1 a5b3 e5e6 c6c1 e6e7 f8e8');
INSERT INTO evaluations VALUES(5762,30,'cp 292',NULL);
INSERT INTO evaluations VALUES(5762,31,'cp 286',NULL);
INSERT INTO evaluations VALUES(5762,32,'cp 290',NULL);
INSERT INTO evaluations VALUES(5762,33,'cp 299',NULL);
INSERT INTO evaluations VALUES(5762,34,'cp 299',NULL);
INSERT INTO evaluations VALUES(5762,35,'cp 281',NULL);
INSERT INTO evaluations VALUES(5762,36,'cp 272','f1d3 a7a6 a2a4 f7f6 d1c2 g8e7 d3h7 e7f5 h7g6 e8f8 g6f5 e6f5 e1g1 c8e6 f1e1 d8d7 b1d2 a8e8 b4b5 a6b5 a4b5 c6a7 d2f1 a7b5 f1g3 c7c6 h2h4 b6c7 g3e2 e6f7 e2f4 f8g8 c1d2 f7h5 f4h5 h8h5 g2g3 c7b8 c2d3 b5c7 e5f6 g7f6 e1e8 d7e8 a1e1 e8d7 e1b1 c7e6 c3c4 f5f4 d3g6 e6g7');
INSERT INTO evaluations VALUES(17501,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17501,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(17501,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(17501,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(17501,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(17501,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(17501,36,'cp 44','d2d4 d7d5 e4d5 d8d5 c1e3 c5d4 c3d4 e7e6 b1c3 d5d8 a1c1 g8f6 g2g3 f6d5 e3d2 f8b4 f1g2 e8g8 e1g1 h7h6 f3e5 d5c3 d2c3 b4c3 e5c6 b7c6 b2c3 c8d7 d1h5 a8b8 c3c4 d8g5 h5g5 h6g5');
INSERT INTO evaluations VALUES(17495,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(17495,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(17495,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(17495,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(17495,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17495,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(17495,36,'cp 65','d7d5 e4d5 d8d5 d4c5 d5d1 e1d1 g8f6 c1e3 f6d5 f1c4 d5e3 f2e3 e7e6 b2b4 a6a5 c4b5 c8d7 f3d4 a5b4 b5c6 d7c6 c3b4 c6g2 h1g1 g2e4 b1d2 e4g6 a2a4 a8d8 d2c4 g6d3 c4e5 d3e4 d1e2 f8e7 e5c4 d8d5 g1g7 e6e5 g7g4 e4g6 c4b6 d5d8 d4f3 f7f5 g4g1 e5e4 f3d4 f5f4 d4e6 g6h5 e2f1');
INSERT INTO evaluations VALUES(27917,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(27917,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(27917,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(27917,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(27917,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(27917,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(27917,36,'cp 147','c1g5 g8e7 d4d5 b8a6 d1d2 d7d6 f1d3 b7c8 e1c1 a6c5 d3c2 e6e5 h1g1 h7h6 g5e3 f7f5 e4f5 c8f5 g2g4 f5c2 d2c2 e8g8 f3d2 a8c8 h2h4 c7c6 e3c5 b6c5 d5c6 f8f2 c2d3 c8c6 d2e4 f2h2 e4d6 h2h4 d3g3 c6d6 g3h4');
INSERT INTO evaluations VALUES(36335,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(36335,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(36335,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(36335,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36335,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36335,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36335,36,'cp 30','d5c4 d1c2 a7a6 a2a4 c8d7 c2c4 d7c6 c1g5 c6d5 c4d3 c7c5 b1c3 d5c6 f1d1 c5d4 f3d4 c6g2 g1g2 b8d7 d3f3 h7h6 g5f6 e7f6 c3e4 f6e7 d4f5 a8c8 f5e7 d8e7 f3a3 e7a3 a1a3 d7e5 a3b3 c8c7 d1d4 f8c8');
INSERT INTO evaluations VALUES(36378,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36378,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36378,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36378,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(36378,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(36378,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36378,36,'cp 28','d1c2 a7a6 a2a4 c8d7 c2c4 d7c6 c1g5 c6d5 c4c2 d5e4 c2d1 h7h6 g5f6 e7f6 b1d2 e4c6 e2e3 a6a5 d1e2 b8a6 b2b3 a6b4 f1d1 b7b6 a1c1 a8c8 f3e1 c6g2 g1g2 c7c5 d4c5 c8c5 c1c5 b6c5 d2c4 d8c7 e2f3 f8d8 d1d8 f6d8');
INSERT INTO evaluations VALUES(36379,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36379,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(36379,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36379,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36379,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36379,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(36379,36,'cp 24','a7a6 a2a4 c8d7 c2c4 d7c6 c1g5 c6d5 c4c2 d5e4 c2c1 b8c6 b1d2 e4g6 f1d1 h7h6 g5f6 e7f6 d2b3 g6e4 e2e3 e4d5 b3d2 c6b4 e3e4 d5c6 a4a5 f6e7 c1c3 d8e8 f3e5 c6b5 d2b3 b7b6 c3c7 f7f6 e5f3 b5a4 a1a4 e8a4 c7e7 a4b3');
INSERT INTO evaluations VALUES(44792,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44792,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(44792,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(44792,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(44792,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(44792,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44792,36,'cp 15','f8g7 c1e3 e8g8 e1g1 b8c6 f1e1 a7a6 h2h3 c8d7 a2a4 a8c8 d4f3 d7e6 f3g5 e6d7 d1d2 h7h6 g5f3 g8h7 e2f1 d7e6 f3d4 c6d4 e3d4 d8a5 a1d1 f6d7 f2f4 g7d4 d2d4 a5c5 f1e2 g6g5 f4f5 e6c4 c3d5 c5d4 d1d4 c4e2');
INSERT INTO evaluations VALUES(2144,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(2144,31,'cp 302',NULL);
INSERT INTO evaluations VALUES(2144,32,'cp 314',NULL);
INSERT INTO evaluations VALUES(2144,33,'cp 298',NULL);
INSERT INTO evaluations VALUES(2144,34,'cp 283',NULL);
INSERT INTO evaluations VALUES(2144,35,'cp 309',NULL);
INSERT INTO evaluations VALUES(2144,36,'cp 319','f7g5 g8h8 c4f7 e8d8 f7b3 d8e8 b1c3 d7d6 d2d3 c6d4 c1e3 c8g4 f2f3 d4b3 a2b3 g4d7 d1d2 h7h6 g5h3 d7h3 g2h3 e7e6 d2g2 f8e7 d3d4 e5d4 e3d4 f6h5 e1c1 h8f8 c3d5 e7g5 g2g5 h6g5 d5c7 e8f7 c7e6 f7e6 h1g1 f8f3 g1g5');
INSERT INTO evaluations VALUES(2142,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(2142,31,'cp 292',NULL);
INSERT INTO evaluations VALUES(2142,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(2142,33,'cp 269',NULL);
INSERT INTO evaluations VALUES(2142,34,'cp 285',NULL);
INSERT INTO evaluations VALUES(2142,35,'cp 318',NULL);
INSERT INTO evaluations VALUES(2142,36,'cp 314','g8h8 c4f7 e8d8 f7b3 d8e8 e1g1 h7h6 g5f3 e8d8 c2c3 d7d6 d2d4 c8d7 b1d2 g7g5 d2c4 a7a5 f1e1 f8g7 a2a4 h8f8 d4d5 c6b8 b3c2 e7f7 b2b4 a5b4 c3b4 g5g4 f3d2 f6h5 e1f1 h5f4 c4e3 h6h5 d2c4 f7g6 e3f5 f8f5 e4f5 d7f5 c2f5 g6f5 f2f3 f5g6 f3g4 g6g4 d1g4 h5g4 c4e3 b8d7 e3g4');
INSERT INTO evaluations VALUES(14816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14816,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14816,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14816,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14816,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14816,36,'cp 0','c2c3 f8d6 e2e3 e8g8 b1d2 d8e7 f3e5 f6d7 e5d7 c8d7 f4d6 e7d6 f1e2 e6e5 e1g1 f8e8 a1c1 a7a5 a2a3 d6g6 g1h1 g6h6 h1g1');
INSERT INTO evaluations VALUES(14802,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14802,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14802,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14802,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14802,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14802,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14802,36,'cp 0','f8d6 e2e3 e8g8 b1d2 d8e7 f3e5 f6d7 e5d7 c8d7 f4d6 e7d6 f1e2 e6e5 e1g1 a8e8 a1c1 d6g6 g1h1 g6h6 h1g1');
INSERT INTO evaluations VALUES(21994,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(21994,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21994,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21994,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(21994,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21994,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(21994,36,'cp 25','g1f3 g8f6 d2d3 c8f5 a2a3 g7g6 c2c4 f8g7 f1e2 e8g8 e1g1 f8e8 b1d2 d8e7 h2h3 f5c8 d1c2 b7b6 a1d1 c8b7 b3b4 c6d8 d2b1 c7c5 b4c5 b6c5 f3d2 a8b8 b2c3 d8e6 e2f3 b7f3 d2f3 e7b7');
INSERT INTO evaluations VALUES(53526,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(53526,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(53526,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(53526,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(53526,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(53526,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(53526,36,'cp 216','c6b7 c8b7 g1f3 b8c6 b1c3 g8f6 f1c4 f8d6 d2d3 e8g8 e1g1 c6d4 f3d2 a8c8 c3e4 f6e4 d3e4 d8f6 c4d3 d4e6 d2c4 f8d8 d1g4 d6c5 c1d2 c5d4 d2a5 d8d7 a1d1 g7g6 c2c3 d4c5 b2b4 d7d3 d1d3 b7a6 c4b2 h7h5 g4d1 a6d3 b2d3 c5b6 a5b6 a7b6');
INSERT INTO evaluations VALUES(51721,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(51721,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(51721,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(51721,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(51721,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(51721,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(51721,36,'cp 186','c8b7 g1f3 b8c6 b1c3 g8f6 f1c4 f8d6 d2d3 e8g8 e1g1 c6d4 f3d2 a8c8 b2b3 d6e7 c1b2 d8c7 c3e4 f6d5 d2f3 d5b6 f3d4 e5d4 b2d4 b6c4 b3c4 f8d8 c2c3 b7e4 d3e4 e7c5 d1d3 c5d6 h2h3 d6h2 g1h1 h2e5 d4e5 d8d3 e5c7 c8c7 f1e1 d3c3');
INSERT INTO evaluations VALUES(30519,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(30519,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(30519,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(30519,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(30519,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(30519,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(30519,36,'cp 206','g1f3 b8c6 b1c3 g8f6 f1c4 f8d6 d2d3 e8g8 e1g1 c6d4 f3d2 a8c8 c3e4 f6e4 d3e4 h7h6 c2c3 d4e6 c4e6 f7e6 d1e2 a7a5 f1d1 d6c5 d2f3 d8b6 b2b3 a5a4 b3b4 c5d6 c1d2 b6c6 a1c1 c6e4 e2e4 b7e4 d2h6 g7h6 d1d6 e4f3 g2f3 f8f3');
INSERT INTO evaluations VALUES(5401,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(5401,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(5401,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(5401,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(5401,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(5401,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(5401,36,'cp -50','g8f6 d4c6 b7c6 b1d2 d7d5 f1d3 a8b8 a1b1 d8c7 h2h3 d5e4 d2e4 f6d5 e3d2 e8g8 e1g1 c6c5 f1e1 c5c4 d3f1 c8f5 b2b3 d5f6 e4f6 g7f6 b1b2 f8d8 d1c1 f5d3 d2f4 e7e5 f4e3 a7a5 b3c4 d3f1 e1f1 c7c4 e3b6 d8d3 b6a5 b8b2 c1b2 g8g7 a5b6');
INSERT INTO evaluations VALUES(8057,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(8057,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(8057,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(8057,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(8057,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(8057,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(8057,36,'cp -39','b1d2 d7d5 d4c6 b7c6 f1d3 d8c7 h2h3 e8g8 e1g1 a8b8 a1b1 f8d8 d1f3 d5e4 d3e4 f6d5 d2b3 d5e3 f3e3 g7e5 b3d4 c8d7 d4c6 d7c6 e4c6 b8b6 c6a4 e5f4 e3f3 b6f6 g2g3 f4h6 f3e4 d8d2 b1d1 d2b2 d1d7 c7c5 e4d4 c5d4 c3d4 b2a2 d7a7 g8f8 f2f4 a2b2 g1h1 b2d2');
INSERT INTO evaluations VALUES(8056,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(8056,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(8056,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(8056,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(8056,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(8056,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(8056,36,'cp -31','d7d5 f1d3 e8g8 e1g1 d8c7 d4c6 c7c6 f2f3 d5e4 d2e4 f6d5 e3f2 d5f4 d3c2 c8e6 f1e1 e6d5 f2e3 e7e5 e3f4 e5f4 c2b3 a8d8 d1c2 d5b3 c2b3 d8d5 g1h1 d5b5 b3c2 f8e8 e4f2 b5e5 e1e5 g7e5 a1d1 g8g7 c3c4 c6c7 f2d3');
INSERT INTO evaluations VALUES(10177,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(10177,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(10177,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(10177,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(10177,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(10177,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(10177,36,'cp 106','f3e5 c6e5 d2d4 c5b4 d4e5 f6e4 d1d4 e4c3 b2c3 b4e7 c4c5 b7b6 c1a3 e8g8 a1d1 b6c5 a3c5 f8e8 f1c4 e7c5 d4d5 c8b7 d5b7 e8e5 d1d7 d8d7 b7a8 e5e8 a8f3 e8d8 e1g1 d7e7 h2h4 e7f6 f3f6 g7f6 f1b1 c5b6 h4h5 g8g7 g2g4 d8d2 c4b3 d2d6 g1f1 d6d3 f1g2 b6c5 h5h6 g7h6 b3f7');
INSERT INTO evaluations VALUES(20127,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(20127,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(20127,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(20127,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(20127,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(20127,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(20127,36,'cp 118','d1d4 g8f6 c1g5 c8d7 e1c1 d8a5 f2f4 h7h6 g5f6 g7f6 f1c4 e7e6 d4f6 h8g8 e4e5 d6d5 c4b3 a5c5 g2g3 c5e7 f6e7 f8e7 d1d3 h6h5 b3a4 d7a4 c3a4 a8b8 h1d1 h5h4 a4c3 b8c8 a2a3 c8c4 c1b1 c4c8 b2b3 a7a6');
INSERT INTO evaluations VALUES(54255,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(54255,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(54255,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(54255,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(54255,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(54255,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(54255,36,'cp 0','g2g3 g8f6 f1g2 b8c6 g1f3 f8e7 e1g1 e8g8 f1e1 b7b5 e4d5 f6d5 c2c4 b5c4 d3c4 d5b4 f3e5 c8b7 a2a3 c6e5 g2b7 b4d3 b7a8 d3e1 d1e1 e5d3 e1e2 d3c1 a1c1 d8a8 c1d1 f8d8 d2f3 d8d1 e2d1 h7h6 b2b3 a8e4 d1d7 e4f3 d7e7 f3b3 e7c5 b3b1 g1g2 b1e4 g2f1 e4d3 f1g1');
INSERT INTO evaluations VALUES(54254,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54254,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54254,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54254,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54254,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(54254,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54254,36,'cp 0','f8e7 f1g2 g8f6 g1f3 e8g8 e1g1 b8c6 f1e1 b7b5 e4d5 f6d5 c2c4 b5c4 d3c4 d5b4 f3e5 c8b7 a2a3 c6e5 g2b7 b4d3 b7a8 d3e1 d1e1 e5d3 e1e2 d3c1 a1c1 d8a8 c1d1 f8d8 d2f3 d8d1 e2d1 h7h6 b2b3 a8e4 d1d7 e4f3 d7e7 f3b3 e7c5 b3b1 g1g2 b1e4 g2f1 e4d3 f1g1');
INSERT INTO evaluations VALUES(4582,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4582,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4582,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4582,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4582,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4582,36,'cp 0','d7d5 e4d5 f6d5 a2a4 a7a5 b1d2 d5b6 c4b5 f7f6 d2c4 b6d5 c4e3 d5b6');
INSERT INTO evaluations VALUES(3344,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3344,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3344,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3344,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3344,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3344,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3344,36,'cp 0','e4d5 f6d5 a2a4 a7a5 b1d2 d5b6 c4b5 f7f6 d2c4 b6d5 c4e3 d5b6');
INSERT INTO evaluations VALUES(3341,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3341,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3341,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3341,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3341,36,'cp 0','f6d5 a2a4 a7a5 b1d2 d5b6 c4b5 f7f6 d2c4 b6d5 c4e3 d5b6');
INSERT INTO evaluations VALUES(3345,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3345,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3345,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3345,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3345,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3345,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3345,36,'cp 0','a2a4 a7a5 b1d2 d5b6 c4b5 f7f6 d2c4 b6d5 c4e3 d5b6');
INSERT INTO evaluations VALUES(8049,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(8049,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8049,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8049,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8049,36,'cp 0','e8g8 e1g1 d7d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 f2f4 c8e6 g1f2 a7a5 a2a3 a8c8 c2c3 c8c6 a3a4 h7h6 d1d4 c6c5 a1d1 g6g5 g2g3 b7b6 e2b5 e6b3 d1d2 b3e6');
INSERT INTO evaluations VALUES(3073,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3073,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3073,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(3073,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(3073,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(3073,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(3073,36,'cp 8','e1g1 d7d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 f2f4 c8e6 g1f2 a7a5 a2a3 a8c8 c2c3 c8c6 a3a4 c6c5 d1d4 h7h6 a1d1 g6g5 g2g3 b7b6 e2b5 e6b3 d1d2 g7f6 f2e3 b3e6 d2d1 f6g7 d4d8 g5f4 e3f4 c5f5 f4e3 f5e5 e3f2 e5f5 f2g1 f8d8');
INSERT INTO evaluations VALUES(3072,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(3072,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(3072,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(3072,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(3072,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(3072,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(3072,36,'cp 0','d7d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 f2f4 c8e6 d1d4 f8c8 c2c3 h7h5 h2h4 a7a5 a2a3 c8c6 a1d1 a8c8 d1d2 c6b6 a3a4 b6b3 g1f2 g7f6 e2f3 b3b6 f2e3 b6b3');
INSERT INTO evaluations VALUES(3150,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(3150,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(3150,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(3150,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(3150,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3150,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3150,36,'cp 0','e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 f2f4 c8e6 a2a3 b7b6 d1d4 e6f5 c2c3 a8e8 d4d5 f5e6 d5d2 e6b3 e2d1 b3e6 d1f3 e8c8 d2d4 h7h6 g1f2 c8c5 a1d1 f8c8 f3e2 g6g5 g2g3 g7f6 d4b4 e6f5 d1d4 g5f4 g3f4');
INSERT INTO evaluations VALUES(3145,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(3145,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(3145,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(3145,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3145,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(3145,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(3145,36,'cp 3','f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 f2f4 b7b6 a2a4 a8b8 g1f2 c8f5 c2c3 h7h5 d1d2 a7a5 h2h4 f5e6 a1e1 f8c8 e2a6 c8c5 e1e4 g7f6 g2g3 e6f5 e4d4 e7e5 f4e5 c5e5 d4d6 f5e6 a6c4 f6e7 c4e6 e5e6 d6d7 e7f6 d2d4 b8c8');
INSERT INTO evaluations VALUES(3165,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(3165,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(3165,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3165,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3165,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3165,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(3165,36,'cp 5','c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 f2f4 c8e6 a2a3 h7h5 h2h4 b7b6 d1d4 e6f5 c2c3 f8e8 e2f3 a8c8 a1e1 g7f6 g1f2 c8c5 e1d1 e8c8 d1d2 e7e5 f4e5 c5e5 d4d8 c8c7 d2d5 e5d5 d8d5 f6e7 d5b5 f5e6 a3a4');
INSERT INTO evaluations VALUES(3163,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3163,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(3163,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3163,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3163,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(3163,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(3163,36,'cp 0','c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 f2f4 c8e6 g1f2 a7a5 a2a3 h7h5 g2g3 f8c8 c2c3 c8c6 a3a4 c6c7 d1d4 a8c8 a1e1 g7f6 e1d1 e6b3 d1d2 e7e6 e2f3 f6e7 h2h4 c8h8 f3e2 h8c8 d4e4 c7c5 e4d4');
INSERT INTO evaluations VALUES(34302,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(34302,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(34302,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(34302,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(34302,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(34302,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(34302,36,'cp 93','c2c4 d5a5 c1d2 a5d8 g1f3 c8f5 b1c3 e7e6 h2h3 h7h6 d1b3 d8b6 g2g4 f5h7 f1g2 g8f6 e1g1 f8e7 f1e1 b8a6 b3b6 a7b6 a2a3 e8g8 f3e5 a6c7 d2f4 e7d6 a1d1 h7c2 d1d2 c2h7 f4g3 f8e8 h3h4 e8d8 h4h5 f6d7 e5d7');
INSERT INTO evaluations VALUES(34297,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(34297,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(34297,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(34297,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(34297,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(34297,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(34297,36,'cp 102','d5a5 c1d2 a5d8 g1f3 c8f5 d1e2 e7e6 b1c3 b8d7 e1c1 f8e7 d4d5 c6d5 c4d5 e6d5 c3d5 a8c8 d2c3 f5e6 d5e7 d8e7 f3d4 g8f6 d4e6 e7e6 c1b1 e8g8 e2e6 f7e6 c3d4 d7c5 f1c4 g8f7 f2f4 c5e4 c4e2 f8d8 h1e1 b7b6 g2g4 f6d5 e2d3 e4c5 f4f5 c5d3');
INSERT INTO evaluations VALUES(50258,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(50258,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(50258,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(50258,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(50258,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(50258,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(50258,36,'cp 43','e2e4 d7d6 d2d4 f8g7 c1e3 a7a6 d1d2 b7b5 e3h6 e8g8 f1d3 b8c6 d4d5 g7h6 d2h6 f6g4 h6d2 c6e5 f3e5 g4e5 d3e2 e7e6 a2a3 e6d5 d2d5 c8e6 d5d2 f7f5 f2f4 e5c4 e2c4 e6c4 e1c1 d8f6 c3d5 c4d5 d2d5 f6f7 e4f5 f7d5 d1d5 f8f5 d5f5');
INSERT INTO evaluations VALUES(22735,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(22735,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(22735,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(22735,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(22735,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(22735,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22735,36,'cp 14','g8e7 g2g3 d7d5 c4d5 e7d5 g1f3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 d1a4 e8g8 c1e3 f8d8 a1c1 f7f6 f1e1 c6b4 a4d7 d8d7 a2a3 b4d5 e3c5 e7c5 c1c5 c7c6 f3d2 d5c7 g1f1 g8f7 d2e4 a7a5 c5c3 c7a6 f2f4 e5f4 g3f4 a8d8 e1c1 h7h6 f1f2 a5a4 e4c5');
INSERT INTO evaluations VALUES(32622,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(32622,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(32622,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(32622,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(32622,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(32622,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(32622,36,'cp 63','e1g1 b8c6 h2h3 g4f3 e2f3 d5d7 f1e1 e8c8 d2d3 h8g8 b1d2 g7g5 c2c3 g5g4 h3g4 f6g4 d2c4 h7h5 d1a4 e7e6 d3d4 h5h4 c1f4 c6e5 a4d7 e5d7 c4a5 c7c6 a2a4 d7f6 a5c4 f6d5 f4d2 f8h6 d2h6 g4h6 a4a5 a7a6 e1e4 h6f5 c4e5 g8g7 a1e1 d5f6 e4f4 d8g8 g1f1');
INSERT INTO evaluations VALUES(32621,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(32621,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(32621,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(32621,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(32621,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(32621,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(32621,36,'cp 65','b8c6 h2h3 g4f3 e2f3 d5d7 f1e1 e8c8 d2d3 h7h5 c1g5 d7f5 g5f6 e7f6 f3c6 b7c6 b1d2 f8c5 d1f3 f5f3 d2f3 h8e8 g1f1 c5d6 e1e8 d8e8 a1e1 e8e1 f1e1 c8d7 e1d2 d7e6 d3d4 e6d5 d2d3 c6c5 c2c4 d5e6 d4d5 e6f5 f3d2 g7g5 d2e4 g5g4 h3g4 h5g4 e4d6 c7d6 d3c3 f5e4 b2b4 c5b4 c3b4 f6f5 b4b5 f5f4 c4c5');
INSERT INTO evaluations VALUES(27405,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(27405,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(27405,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(27405,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(27405,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(27405,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(27405,36,'cp 63','h2h3 g4f3 e2f3 d5d7 f1e1 e8c8 d2d3 h7h5 c1g5 d7f5 g5f6 e7f6 f3c6 b7c6 b1d2 f8c5 d1f3 f5f3 d2f3 h8e8 e1e8 d8e8 g1f1 c5b4 d3d4 b4d6 a1d1 c8d7 f3d2 c6c5 d4c5 d6c5 d2e4 c5d6 d1d4 a7a5 e4d6 c7d6 b2b3 e8b8 f1e2');
INSERT INTO evaluations VALUES(47305,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(47305,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(47305,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(47305,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(47305,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(47305,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(47305,36,'cp 37','f6d5 f1c4 b8c6 g1f3 c8f5 b1c3 d5b6 c4e2 e7e6 a2a3 d8d7 c1e3 f8e7 e1g1 e8g8 a1c1 a8d8 f1e1 a7a6 d1b3 d8a8 e1d1 f8d8 h2h3 h7h6 f3e1 e7f8 e2f3 f5h7 c3e4 a6a5 e4c5 f8c5 c1c5 a5a4 b3c3 b6d5');
INSERT INTO evaluations VALUES(47474,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(47474,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(47474,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(47474,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47474,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(47474,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(47474,36,'cp 44','g1f3 c8f5 b1c3 e7e6 f1c4 d5b6 c4b3 b8c6 e1g1 f8e7 a2a3 f5g4 c1e3 d8d7 h2h3 g4f3 d1f3 c6d4 f3b7 d4b3 e3b6 e8g8 a1d1 d7c8 b7c8 f8c8 b6e3 a7a6 d1d7 g8f8 f2f4 a8b8 c3e4 f8e8 f1d1 h7h6 g1f2 b8b5 g2g3 c8d8 d7d8 e7d8');
INSERT INTO evaluations VALUES(47473,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(47473,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47473,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(47473,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47473,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(47473,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(47473,36,'cp 27','c8f5 f1c4 e7e6 b1c3 d5b6 c4e2 f8e7 e1g1 e8g8 c1e3 c7c6 h2h3 b6d5 e2d3 f5d3 d1d3 b8a6 a2a3 a6c7 d3e2 d5c3 b2c3 b7b5 f3d2 c7d5 d2e4 a7a5 c3c4 b5c4 e2c4 d8d7 c4e2 h7h6 f1b1 d5e3 f2e3 d7c7 e2c2');
INSERT INTO evaluations VALUES(52118,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52118,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52118,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52118,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(52118,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52118,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(52118,36,'cp 7','c5d4 g1f3 b8c6 c3d4 e5d4 f3d4 f8b4 b1c3 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4e7 e2f4 d5f4 c1f4 e7f6 a1c1 f6b2 c1c6 c8e6 f4e3 e6a2 c6c2 a8b8 e3a7 b8a8 a7e3 b2f6 d3b5 a2e6 b5c6 a8c8');
INSERT INTO evaluations VALUES(6410,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(6410,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(6410,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(6410,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(6410,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(6410,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(6410,36,'cp 200','b1c3 d7d6 c1f4 g8e7 e5d6 c7d6 d1d2 e8g8 a2a3 d8a5 a1b1 c5b6 e2e3 d6d5 c3b5 a7a6 b5d4 c6d4 f3d4 a5a4 c2c3 b6d4 d2d4 a4d4 e3d4 c8f5 b1c1 f8e8 e1d2 e7c6 b2b4 e8e6 f1e2 a8e8 f4e3 f5e4 h1g1 c6e7 f2f3 e4g6');
INSERT INTO evaluations VALUES(6408,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(6408,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(6408,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(6408,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(6408,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(6408,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(6408,36,'cp 205','f7f6 c1f4 f6e5 f3e5 g8f6 e5c6 b7c6 c3a4 c5e7 e2e3 d7d5 f1e2 e8g8 e1g1 c6c5 b2b3 c8f5 c2c4 d5d4 e2f3 a8b8 d1d2 f5g4 f3g4 f6g4 a1d1 d4e3 f2e3 d8d2 d1d2 b8d8 d2d8 f8d8 h2h3 g4f6 f4c7 d8d2');
INSERT INTO evaluations VALUES(12800,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(12800,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12800,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12800,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12800,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(12800,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(12800,36,'cp 0','b1c3 e7e5 d2d3 d7d6 f1e2 f8e7 e1g1 e8g8 a2a3 a7a5 c1e3 c8e6 f3d2 f6d7 e2g4 d7f6');
INSERT INTO evaluations VALUES(54978,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(54978,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(54978,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(54978,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(54978,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54978,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(54978,36,'cp -12','c7c5 f1b5 c8d7 b5d7 d8d7 g1e2 b8c6 c2c3 g7g6 d2d4 f8g7 f3d3 d6d5 e4d5 d7d5 e2f4 d5d6 d4d5 c6e5 d3e2 g8f6 e1g1 f6d5 f1d1 d5f4 c1f4 d6a6 e2a6 b7a6 f4e5 g7e5 b1d2 e5c7 d2e4 a8b8 b2b3 c7b6 a2a4 f7f5 e4g5 e7e5 g5f3 e5e4 f3e5 e8g8 d1d7');
INSERT INTO evaluations VALUES(6345,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(6345,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(6345,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(6345,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(6345,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(6345,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(6345,36,'cp 105','b1c3 d7d6 f1b5 c8d7 c1e3 g8e7 e1g1 e8g8 h2h3 e7g6 b5c4 c6a5 c4d3 a5c6 a1c1 c6e7 a2a4 c7c6 f1e1 d6d5 e4d5 e7d5 c3d5 c6d5 e3g5 d8b8 d1b3 d7c6 h3h4 g6f4 d3b1 f4e6 b3c2 e6g5 f3g5 g7g6 h4h5');
INSERT INTO evaluations VALUES(6344,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(6344,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(6344,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(6344,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(6344,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(6344,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(6344,36,'cp 98','d7d6 f1b5 c8d7 e1g1 g8e7 h2h3 e8g8 c1e3 e7g6 b5c4 c6a5 c4d3 a5c6 a1c1 c6e7 f1e1 c7c6 b2b4 d6d5 e4d5 e7d5 c3d5 c6d5 e3g5 d8b8 f3e5 b8d6 d3g6 h7g6 e5d7 d6d7 e1e7 d7f5 e7e5 f5d7 d1f3 a8c8 e5d5 c8c1 g5c1');
INSERT INTO evaluations VALUES(5785,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(5785,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(5785,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(5785,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(5785,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(5785,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(5785,36,'cp 96','f1b5 c8d7 e1g1 g8e7 c1e3 e8g8 h2h3 e7g6 b5c4 c6a5 c4d3 a5c6 a1c1 c6e7 a2a4 c7c6 f1e1 b6a5 d3b1 d6d5 e4d5 c6d5 e3g5 h7h6 b1g6 h6g5 g6c2 g5g4 f3e5 g4h3 d1h5 g7g6 h5h6 d7f5 g2g4 f5c2 c1c2 a5b6 h6h3');
INSERT INTO evaluations VALUES(11469,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11469,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(11469,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(11469,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(11469,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(11469,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(11469,36,'cp 39','d2d4 e5d4 f3d4 c6e5 c4b3 f8c5 f2f4 d7d6 g1h1 c8g4 d1d2 e5d7 h2h3 f6h5 d2f2 c5d4 f2d4 d8h4 f1f2 e8g8 c1d2 a8e8 h1g1 g4e6 g2g4 h5f6 f2f3 f6h7 a1d1 d7c5 d2e3 c5b3 a2b3 e6d7 e3f2 h4f6 d4d2 d7c6 d1e1 f6g6 f3e3 h7f6 d2d4 f6d7 f2h4 f7f6 f4f5 g6h7');
INSERT INTO evaluations VALUES(11487,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(11487,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(11487,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(11487,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(11487,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(11487,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(11487,36,'cp 43','e5d4 f3d4 c6e5 c4b3 f8c5 d4f3 e5f3 d1f3 d7d6 e4e5 d6e5 f3g3 e8f8 g3e5 c5d6 e5e2 d8e8 e2e8 f8e8 c1d2 e8f8 f1e1 c8f5 a1d1 a8e8 c3e2 f5d7 f2f3 d7e6 e2g3 f8g8 d2c3 g8h7 b3e6 e8e6 e1e6 f7e6 d1e1 h8e8 g3e4 f6d5 c3d2 d5f6 h2h3 h7g6');
INSERT INTO evaluations VALUES(11492,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(11492,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(11492,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(11492,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(11492,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(11492,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(11492,36,'cp 25','f3d4 c6e5 c4b3 f8c5 d4f3 e5f3 d1f3 d7d6 h2h3 e8g8 f3g3 g8h8 c1e3 c5e3 g3e3 d8e7 a1d1 c8e6 b3e6 e7e6 d1d4 a8d8 f1d1 d8d7 a2a3 f8d8 b2b4 b7b5 d1d3 c7c5 b4c5 d6c5 d4d7 f6d7 f2f4 c5c4 d3d4 d8e8 e4e5 d7c5 d4d1 e6c6 c3d5');
INSERT INTO evaluations VALUES(11491,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(11491,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(11491,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(11491,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11491,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(11491,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(11491,36,'cp 37','c6e5 c4b3 f8c5 g1h1 d7d6 f2f4 c8g4 d1d2 e5d7 d4f3 e8g8 h2h3 g4f3 g2f3 g8h8 d2g2 f6h5 c1d2 a6a5 c3e2 a5a4 b3c4 d8e7 a1d1 c7c6 c4d3 b7b5 g2g4 h5f6 g4h4 f6d5 h4e7 d5e7 h3h4 d6d5 a2a3 f7f5 e4e5 d7b6 d2c3 b6c4 d3c4');
INSERT INTO evaluations VALUES(22524,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(22524,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(22524,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(22524,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(22524,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(22524,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(22524,36,'cp 93','d2d4 e5d4 f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 c8e6 c1d2 f8c5 c3e2 d8e7 e2f4 a7a5 h2h4 h8d8 f4e6 f7e6 g2g4 c5b4 c2c3 b4d6 h1g1 d8h8 e1c1 a8f8 f1e2 f6d7 g4g5 h6h5 g1g2 d6f4 c1c2 f4d2');
INSERT INTO evaluations VALUES(22522,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(22522,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(22522,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(22522,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(22522,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(22522,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(22522,36,'cp 85','e5d4 f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 c8e6 a2a3 f8c5 c3e2 d8e7 h2h4 a8d8 e2f4 h8e8 f1e2 e7f8 c1d2 c5f2 e1f2 d8d2 a1c1 c6c5 g2g3 d2d4 c1d1 d4d1 h1d1 g7g5 f4e6 e8e6 d1d8 e6e8 d8e8 f6e8 h4g5 h6g5 e4e5 f7f6 f2e3 f6e5');
INSERT INTO evaluations VALUES(22544,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22544,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(22544,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(22544,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(22544,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(22544,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(22544,36,'cp 101','f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 f6d7 f1e2 f7f6 f3f4 a7a5 c1d2 f8c5 e1c1 d8e7 a2a4 d7b6 b2b3 c5a3 c1b1 a3b4 h2h4 c8e6 b1b2 a8d8 e2f3 d8d7 d2e3 h8d8 d1d7 b6d7 c3e2 b4c5 e2d4 c5d4 e3d4 b7b6 d4c3 e6f7 g2g4');
INSERT INTO evaluations VALUES(22543,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22543,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(22543,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(22543,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(22543,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(22543,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(22543,36,'cp 84','g8f6 d4c6 d7c6 d1d8 e8d8 f2f3 f8d6 c1e3 d8e7 g2g3 h8e8 c3e2 f6d7 e2d4 e7f8 e1f2 d7e5 h2h3 f7f6 f1e2 c8d7 g3g4 a7a5 d4f5 d7f5 g4f5 a5a4 h1g1 e5f7 a1b1 e8e5 g1d1 h6h5 a2a3 a8e8');
INSERT INTO evaluations VALUES(23008,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(23008,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23008,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23008,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23008,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23008,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23008,36,'cp 0','d4b5 c3b5 c7c6 b5c3 d8c7 d2d4 d7d6 c3e2 g8f6 e2g3 h7h5 c1g5 f6h7 g5e3 h7f6');
INSERT INTO evaluations VALUES(23035,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23035,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23035,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23035,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23035,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23035,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23035,36,'cp 0','c3b5 c7c6 b5c3 d8c7 d2d4 d7d6 c3e2 g8f6 e2g3 h7h5 c1g5 f6h7 g5e3 h7f6');
INSERT INTO evaluations VALUES(23025,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(23025,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23025,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23025,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(23025,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23025,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(23025,36,'cp 0','d7d6 d2d4 c7c6 b5c3 d8c7 e1g1 g8f6 c1g5 f8e7 g5f6 e7f6 d4d5 f6e7 d1d3 e8g8 f1d1 c7b6 a2a4 a7a5 a1a3 f7f5 d3c4 g8h8 a3b3 b6c7 c3b5 c7d7 b5c3');
INSERT INTO evaluations VALUES(25949,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(25949,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(25949,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(25949,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(25949,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25949,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25949,36,'cp 0','c5d4 c3d4 d8b6 d3e2 g8e7 b1c3 e7f5 c3a4 b6a5 a4c3 f8b4 e1g1 b4c3 b2c3 a5c3 c1d2 c3a3 d2c1');
INSERT INTO evaluations VALUES(25959,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25959,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25959,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25959,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25959,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25959,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25959,36,'cp 0','c3d4 d8b6 d3e2 g8h6 b1c3 h6f5 c3a4 b6a5 a4c3 f8b4 e1g1 b4c3 b2c3 a5c3 c1d2 c3a3 d2c1');
INSERT INTO evaluations VALUES(25952,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25952,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25952,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25952,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25952,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25952,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25952,36,'cp 0','d8b6 d3e2 g8h6 b1c3 h6f5 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(47749,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47749,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(47749,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47749,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47749,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47749,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47749,36,'cp 0','f8d6 f1d3 e4c5 b1c3 c5d3 d1d3 e8g8 c1e3 d6b4 e1c1 d8d3 d1d3 c8f5 d3d5 f5e6 d5d3');
INSERT INTO evaluations VALUES(41762,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41762,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(41762,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(41762,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(41762,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(41762,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41762,36,'cp 1','f1d3 e4c5 d3e2 e8g8 c1g5 d8e8 e1g1 c5e6 g5e3 b7b6 b1c3 c8b7 c3b5 d6e7 c2c3 b8c6 f1e1 a8d8 d1c2 a7a6 b5d4 c6d4 f3d4 e6d4 e3d4 c7c5 d4e3 e8c6 e2f3 c6c8 f3b7 c8b7 g2g3 b6b5 a1d1 b7c6 c2e2 h7h6 e3f4 d8d1');
INSERT INTO evaluations VALUES(41761,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(41761,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41761,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41761,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41761,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41761,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41761,36,'cp 0','e4c5 d3e2 d8f6 b1a3 c5e4 a3b5 b8c6 c2c3 h7h6 d1a4 c8f5 f3d4 f5d7 d4f3');
INSERT INTO evaluations VALUES(48442,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(48442,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(48442,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(48442,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(48442,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48442,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(48442,36,'cp -19','a7a6 a2a4 b8c6 e2e3 c6a5 f3d2 f8b4 f1e2 c7c5 d4c5 c8d7 e1g1 d7c6 d1c2 b4c5 e2c4 a8c8 c4a2 c5e7 a1b1 e8g8 b2b4 c6d7 c2b2 a5c6 b4b5 a6b5 c3b5 h7h6 g5h4 c6b4 a2c4 d7b5 a4b5 b4d3 b2a2 d3e5 h4f6 e7f6 f1d1 e5c4 d2c4');
INSERT INTO evaluations VALUES(43667,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(43667,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(43667,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(43667,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43667,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(43667,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(43667,36,'cp -12','a2a4 b8c6 e2e3 c6a5 f3d2 c7c5 d4c5 f8c5 f1c4 e8g8 e1g1 c5e7 c4e2 c8d7 a1b1 a8c8 d2f3 d8c7 g5f4 c7c5 f4e5 c5b4 e2a6 b7a6 e5f6 e7f6 d1d7 a5b3 d7d3 b3c5 d3c2 f8d8 c2e2 c5a4 e2a6 a4b2 c3e2');
INSERT INTO evaluations VALUES(6088,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(6088,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(6088,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(6088,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(6088,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(6088,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(6088,36,'cp -11','f7f6 d2d3 a7a5 c1e3 c5e3 f2e3 g8h6 b1c3 b7b6 a2a3 d8e7 f3h4 c8g4 d1d2 g4d7 h2h3 e8d8 h4f5 d7f5 e4f5 d8c8 e3e4 c8b7 a1e1 h6f7 g1h2 f7d6 c3d1 a5a4 d1e3 d6b5 c2c3 h8d8 e1d1 h7h6 d2e2 b5d6 e3g4 d8f8 g4e3');
INSERT INTO evaluations VALUES(6118,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(6118,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(6118,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(6118,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(6118,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(6118,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(6118,36,'cp -9','d2d3 a7a5 c1e3 c5e3 f2e3 g8h6 b1c3 d8e7 a2a3 b7b6 f3h4 c8d7 d1d2 a5a4 d2f2 e8d8 h4f5 d7f5 e4f5 d8c8 e3e4 c8b7 c3d1 h6f7 d1e3 f7d6 a1d1 h8g8 c2c3 g7g6 f5g6 g8g6 g1h1 a8g8 f2c2 g6g5 c2a4');
INSERT INTO evaluations VALUES(43470,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43470,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43470,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43470,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43470,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43470,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43470,36,'cp 0','f8e7 d1e2 b8c6 e1g1 e8g8 d3c4 c8g4 c1e3 c6a5 c4d3 a8c8 f1d1 g4e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(39685,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39685,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39685,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39685,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39685,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39685,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39685,36,'cp 0','h2h3 c8e6 f3g5 e6d7 e1g1 e8g8 g5f3 d7e6 f3g5');
INSERT INTO evaluations VALUES(39683,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39683,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39683,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39683,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39683,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39683,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39683,36,'cp 0','c8e6 f3g5 e6d7 e1g1 e8g8 g5f3 d7e6 f3g5');
INSERT INTO evaluations VALUES(52314,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(52314,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(52314,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(52314,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(52314,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52314,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(52314,36,'cp -17','b8c6 g1f3 f8g7 a2a4 b7b6 f1e2 g8h6 d2d3 d7d6 e1g1 e8g8 c3b5 a7a6 b5a3 b6b5 d1e1 e7e6 h2h3 c8b7 c2c3 b5b4 a3c4 d6d5 c4e5 d8c7 e5c6 c7c6 e4e5 b4c3 b2c3 d5d4 c3c4 h6f5 a1b1 a8b8 a4a5 f5e3');
INSERT INTO evaluations VALUES(13030,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(13030,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(13030,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(13030,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(13030,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(13030,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(13030,36,'cp 71','d4c6 b7c6 e4e5 f6g4 e5d6 d8d6 h2h3 d6d1 c3d1 g4e5 c4b3 c8a6 d1e3 e5d7 b3c4 a6b7 e1g1 d7b6 c4e2 f8g7 a2a4 a7a5 f1e1 e8g8 e3c4 b6c4 e2c4 f8e8 c1g5 b7a6 c4a6 a8a6 e1e7 e8e7 g5e7 g7b2 a1b1 a6a7 e7c5 a7b7 g1f1 f7f6 f1e2 g8f7 g2g4 h7h5');
INSERT INTO evaluations VALUES(12987,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(12987,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(12987,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(12987,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(12987,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(12987,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(12987,36,'cp 47','b7c6 e4e5 f6g4 e5d6 d8d6 h2h3 g4e5 c4b3 d6d1 c3d1 c8a6 d1c3 f8g7 c1d2 e5c4 b3c4 a6c4 e1c1 g7d4 b2b3 c4e6 f2f3 h7h5 h1e1 h5h4 e1e4 e8c8 d1e1 h8h5 e1e2 d8d7 c3a4 d4f6 e2e1 h5f5 e4e2 d7d8 a4c3 e6d5 c3e4 f6g7 e4c5 g7d4 c5d3');
INSERT INTO evaluations VALUES(12024,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(12024,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(12024,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(12024,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(12024,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(12024,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(12024,36,'cp 62','e4e5 f6g4 e5d6 d8d6 h2h3 g4e5 c4b3 d6d1 c3d1 f8g7 c1d2 c8a6 d1c3 e5c4 b3c4 a6c4 e1c1 c4d5 f2f3 g7e5 h1e1 f7f6 b2b3 h7h5 c3a4 h5h4 c2c4 d5f7 e1e4 e8c8 d2b4 d8d1 c1d1 c8d7 a4c5 d7d8 c5d3 e5d6 b4c5 d6c5 d3c5 h8h5 b3b4 g6g5');
INSERT INTO evaluations VALUES(6505,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(6505,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(6505,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(6505,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(6505,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(6505,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(6505,36,'cp 128','f3e5 c6e5 d2d4 c5b4 c2c3 b4d6 d4e5 d6e5 f2f4 e5f6 e2f3 g8h6 e1g1 e8g8 b1a3 f6e7 a3c4 d7d6 g2g4 f7f6 c4e3 h6f7 g1h1 a8b8 d1e2 b7b5 c1d2 c8b7 h2h4 a7a6 a1d1 f8e8 g4g5 e7f8 e2g2 g8h8 g5g6 f7h6 f4f5 h7g6 f5g6');
INSERT INTO evaluations VALUES(14565,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(14565,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(14565,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(14565,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(14565,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(14565,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(14565,36,'cp 121','e4d5 f6d5 d1b3 c6a5 b3a4 a5c6 f3e5 d8d6 d3d4 c8e6 c4b5 d5b6 a4c2 d6d5 b5c6 b7c6 e1g1 f8d6 f1e1 e8g8 b1d2 b6d7 d2f3 d7e5 d4e5 d6c5 c1e3 c5e3 e1e3 c6c5 b2b3 a7a5 c3c4 d5c6 a1d1 f8d8 a2a4 e6g4 d1d8 a8d8 h2h3 g4f3 e3f3 d8d4 f3e3');
INSERT INTO evaluations VALUES(14558,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(14558,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(14558,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(14558,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(14558,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(14558,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(14558,36,'cp 119','f6d5 d1b3 c6a5 b3a4 a5c6 f3e5 d8d6 d3d4 c8e6 c4b5 d5b6 a4c2 d6d5 b5c6 b7c6 e1g1 f8d6 f1e1 e8g8 b1d2 f8e8 c3c4 d5d4 d2f3 d4c5 c1e3 c5b4 e3d2 b4a4 b2b3 a4a6 e5c6 b6c4 e1e6 e8e6 c2c4 a6c4 b3c4 d6h2 g1h2 e6c6 a1c1 c6a6 c1c2 a8e8 d2e3 f7f6 g2g3 a6a4 h2g2 g8f7');
INSERT INTO evaluations VALUES(14587,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(14587,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(14587,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(14587,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(14587,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(14587,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(14587,36,'cp 132','d1b3 c6a5 b3a4 a5c6 f3e5 d8d6 d3d4 c8e6 c4b5 d5b6 a4c2 d6d5 b5c6 b7c6 e1g1 f8d6 f1e1 e8g8 b1d2 b6d7 d2f3 d7e5 d4e5 d6c5 c1e3 c5e3 e1e3 c6c5 b2b3 d5c6 c3c4 a7a5 a2a4 a8d8 a1d1 e6g4 d1d8 f8d8 h2h3 g4f3 e3f3 d8d4 c2c3 c6e6');
INSERT INTO evaluations VALUES(14584,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(14584,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(14584,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(14584,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(14584,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(14584,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(14584,36,'cp 124','c6a5 b3a4 a5c6 f3e5 d8d6 d3d4 c8e6 c4b5 d5b6 a4c2 d6d5 b5c6 b7c6 e1g1 f8d6 f1e1 e8g8 b1d2 f8e8 d2f3 d6e5 e1e5 d5d7 c1f4 f7f6 e5e1 e6f5 c2d2 b6c4 d2d1 a7a5 e1e8 a8e8 f4g3 d7e6 b2b3 c4b2 d1d2 b2d3 g3c7 e6e2 c3c4 e2d2 f3d2 e8e2 c7a5 e2f2 d2f3 f2c2');
INSERT INTO evaluations VALUES(19778,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19778,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19778,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19778,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19778,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19778,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19778,36,'cp 0','g8f6 f1e2 c5d4 e3d4 c8f5 f3h4 f5e4 c2c3 h7h6 d1b3 c6a5 b3a4 a5c6');
INSERT INTO evaluations VALUES(21357,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(21357,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(21357,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(21357,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(21357,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(21357,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(21357,36,'cp -13','e2e3 a7a6 d1c2 g7g6 g1f3 f8g7 a2a3 g8e7 b2b4 e6e5 d4e5 c8f5 f1d3 c6e5 f3e5 g7e5 a1a2 e8g8 e1g1 c7c6 b1d2 f8e8 c1b2 f5d3 c2d3 e5b2 a2b2 e7f5 b2b3 d8f6 d3c2 f5g7 c2c3 f6c3 b3c3 a6a5 f1b1 g7e6 b4a5 a8a5');
INSERT INTO evaluations VALUES(21355,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(21355,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(21355,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(21355,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(21355,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21355,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(21355,36,'cp -22','a7a6 d1c2 g7g6 g1f3 f8g7 a2a3 e6e5 d4e5 c6e5 f3e5 g7e5 b1d2 e5g7 d2f3 c8f5 f1d3 f5d3 c2d3 g8f6 b2b4 f6e4 a1b1 e8g8 e1g1 c7c6 c1b2 g7b2 b1b2 f8e8 f3d4 e4g5 d3c3 h7h5 f1d1 d8f6 b2c2 e8e7 d4b3 f6c3 c2c3 g5e6 b3d4 e6d4');
INSERT INTO evaluations VALUES(27852,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(27852,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(27852,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(27852,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(27852,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(27852,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(27852,36,'cp 73','c1e3 c7c5 d4d5 e7e6 d5e6 g4e6 e1g1 b8c6 h2h3 d8b6 b2b3 f8e8 a1c1 e6d7 e2d3 c6d4 d3b1 d4f3 d1f3 e8e6 c1d1 d7c6 c3d5 c6d5 e4d5 e6e8 g2g4 f6d7 e3d2 a7a6 g1g2 b6d8 d1e1 b7b5 e1e8 d8e8 f1e1 e8d8 g4g5 b5c4 b3c4');
INSERT INTO evaluations VALUES(27848,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(27848,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(27848,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(27848,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(27848,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(27848,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(27848,36,'cp 60','c7c5 d4d5 e7e6 d5e6 g4e6 e1g1 d8b6 b2b3 f6g4 e3d2 b6d8 f3g5 g4f6 g5e6 f7e6 f2f4 b8c6 g1h1 a7a6 a1c1 d8d7 e2d3 c6d4 c3e2 d4c6 e2g3 a8b8 d1e1 b7b5 c1d1 c6d4 g3e2 b5c4 b3c4 d4c6 d2c3');
INSERT INTO evaluations VALUES(29266,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(29266,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(29266,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(29266,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(29266,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(29266,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(29266,36,'cp 29','f3h4 f5g6 e2e3 b8d7 h4g6 h7g6 c4d5 e6d5 f1d3 f8d6 h2h4 d8c7 d1f3 f6h7 e1c1 d6b4 c3e2 h7g5 h4g5 h8h1 d1h1 d7f8 f3h3 c7e7 f2f4 b4d6 a2a3 f8e6 h3h8 e7f8 h8f8 e6f8 c1d2 e8d7 e2c1 a8c8 d3e2 f8e6 e2g4 c8b8 h1h7 d6f8');
INSERT INTO evaluations VALUES(37868,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(37868,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(37868,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(37868,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(37868,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(37868,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(37868,36,'cp 123','b1c3 d7d6 g1f3 b8c6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 a7a6 b5c6 e7c6 c1g5 c6e7 d1d2 b7b5 a1d1 c8b7 d4d5 e6e5 c3e2 f7f6 g5h6 a8c8 h6g7 g8g7 a2a3 f8f7 d1c1 h7h6 e2g3 c8c1');
INSERT INTO evaluations VALUES(37865,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(37865,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(37865,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(37865,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(37865,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(37865,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(37865,36,'cp 92','d7d6 g1f3 b8c6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 c8d7 c1e3 a7a6 b5c6 d7c6 d1d2 a8c8 e3h6 c6d7 f3h2 b7b5 a1d1 b5b4 c3e2 a6a5 e2g3 f7f5 e4f5 g7h6 d2h6 e7f5 g3f5 f8f5 d1c1 c8c1 e1c1 e6e5 d4e5 f5e5 h6d2');
INSERT INTO evaluations VALUES(38005,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(38005,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(38005,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(38005,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(38005,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(38005,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(38005,36,'cp 108','g1f3 b8c6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 c8d7 c1e3 a7a6 b5c6 d7c6 d1d2 b7b5 e3h6 c6b7 d4d5 e6e5 b2b4 g8h8 h6g7 h8g7 a2a4 b5a4 c3a4 e7g8 a4b2 a8c8 d2e3 c8c2 b2d3 c2c7 e1c1 g8f6 c1c7 d8c7');
INSERT INTO evaluations VALUES(39561,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(39561,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(39561,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(39561,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(39561,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(39561,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(39561,36,'cp 10','c7c5 g2g3 b8c6 f1g2 d7d6 e1g1 c8f5 h2h3 e7e5 d2d3 g8e7 a1b1 e8g8 a2a3 a7a5 b2b3 h7h6 f3d2 f5e6 c3d5 a8b8 c1b2 b7b5 e2e3 d8d7 g1h2 f7f5 d1e2 b5b4 a3a4 f8f7 f2f4 b8d8 e2f2 d8f8 f4e5');
INSERT INTO evaluations VALUES(37782,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(37782,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(37782,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37782,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(37782,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(37782,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(37782,36,'cp 6','g2g3 b8c6 f1g2 d7d6 a2a3 g8f6 e1g1 e8g8 d2d3 a8b8 a1b1 a7a6 b2b4 c5b4 a3b4 b7b5 c4b5 a6b5 e2e4 c8e6 f1e1 d8c8 d1d2 e6g4 d3d4 e7e5 d4d5 g4f3 g2f3 c6d4 f3g2 f8d8 d2d3 c8c4 g2f1 c4d3 f1d3 f6g4');
INSERT INTO evaluations VALUES(37783,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(37783,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37783,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37783,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37783,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(37783,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(37783,36,'cp 12','b8c6 f1g2 d7d6 e1g1 c8f5 a2a3 d8d7 b2b4 g8f6 c1b2 e8g8 b4b5 c6d4 d2d3 f5g4 a3a4 a7a6 b5b6 a8c8 a4a5 g4f3 e2f3 d4c6 f3f4 f8d8 d1d2 c6d4 a1a2 e7e6 c3e2 d6d5 e2d4 c5d4 f1c1 d5c4 d3c4 f6g4 d2e2');
INSERT INTO evaluations VALUES(5325,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(5325,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(5325,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(5325,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(5325,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(5325,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(5325,36,'cp 10','f1g2 d7d6 e1g1 c8f5 a2a3 d8d7 b2b4 g8f6 c1b2 e8g8 b4b5 c6d4 d2d3 f5g4 a3a4 a7a6 b5b6 a8d8 a4a5 d4c6 e2e3 h7h6 d1b3 g4f3 g2f3 c6e5 b3d1 e5f3 d1f3 e7e6 c3a4 f6h7 b2c3 g7c3 a4c3 h7g5 f3g2 f7f5');
INSERT INTO evaluations VALUES(50097,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(50097,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(50097,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(50097,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(50097,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(50097,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(50097,36,'cp 34','e7e6 a2a3 d7d5 c1f4 c7c5 e2e3 a7a6 f1e2 c8d7 d4c5 f8c5 c3b1 d7a4 b2b3 a4b5 b3b4 b5e2 d1e2 c5e7 b1d2 e8g8 e1g1 f6e4 f4b8 d8b8 c2c4 e7f6 a1c1 g8h8 f3d4 f6d4 e3d4 e4d2 e2d2 d5c4 c1c4 b8d8');
INSERT INTO evaluations VALUES(49959,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(49959,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(49959,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(49959,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(49959,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(49959,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(49959,36,'cp 43','a2a3 d7d5 c1f4 a7a6 e2e3 c7c5 f1e2 c8d7 e1g1 b8c6 h2h3 c5d4 e3d4 f8e7 c3a2 f6e4 c2c4 e8g8 a2c3 g7g5 f4h2 d7e8 c4d5 e6d5 a1c1 g8h8 c3e4 f5e4 f3e5 e7d6 d1d2 d8e7 g1h1 c6e5 d4e5 d6e5 d2d5 e5h2');
INSERT INTO evaluations VALUES(53642,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(53642,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(53642,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(53642,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(53642,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(53642,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(53642,36,'cp 70','c2c3 c6c5 d2d4 b8c6 f1d3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f5e3 f2e3 c8d7 b1c3 f8e7 a2a3 e8g8 d1c2 g7g6 a1c1 f7f5 c2f2 g6g5 h2h3 d7e8 g2g4 e8g6 g4f5 g6f5 d3f5 f8f5 f2g2 d8b6 h3h4 a8f8 h4g5 b6d8 c3e2 e7g5 f3g5 d8g5 e2f4 g8h8 g2g5');
INSERT INTO evaluations VALUES(54260,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54260,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54260,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54260,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(54260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54260,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54260,36,'cp 0','f1b5 c8d7 b5d7 d8d7 d2d3 g8e7 g1f3 b8c6 e1g1 e8c8 f3g5 d7e8 g5f3 e8d7');
INSERT INTO evaluations VALUES(54043,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54043,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54043,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54043,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(54043,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(54043,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(54043,36,'cp -8','c8d7 b5d7 d8d7 d2d3 b8c6 g1f3 g8e7 e1g1 e8c8 a1b1 d5e4 c3e4 e7f5 d1e1 d7c7 e1e2 c6d4 f3d4 c5d4 c2c4 d4c3 b2c3 c8b8 c1d2 b8a8 d2e1 f8d6 e2f3 d8d7 e1f2 h8d8 f1d1 d6e7 d3d4 c7c6 h2h3 h7h6 d1e1 d8c8');
INSERT INTO evaluations VALUES(9372,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(9372,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(9372,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9372,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9372,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(9372,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9372,36,'cp 0','c5b6 h2h3 e8g8 f1e1 f8e8 a2a4 e5d4 c3d4 d6d5 e4d5 e8e1 d1e1 c6d4 f3d4 b6d4 b1c3 c8d7 c1f4 d8f8 e1d2 f8c5 c4a2 a8e8 a1c1 c5b4 f4c7 d7a4 c7d6 b4d6 d2d4 a4d7 d4a7 d6b4 a7a3 b4h4 c1d1 h4g5 g1h1 g5h4');
INSERT INTO evaluations VALUES(8915,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(8915,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(8915,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(8915,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(8915,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8915,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8915,36,'cp 0','h2h3 e8g8 f1e1 f8e8 a2a4 e5d4 c3d4 d6d5 e4d5 e8e1 d1e1 c6d4 f3d4 b6d4 b1c3 c8d7 c1e3 d4e3 e1e3 d8f8 c4b5 f8d6 b5d7 d6d7 e3c5 b7b6 c5c6 d7c6 d5c6 a7a6 a1d1 g8f8 g2g4 a8e8 g4g5 f6e4 d1d7 e4c3 b2c3 e8e7 d7d8 e7e8');
INSERT INTO evaluations VALUES(14233,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(14233,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(14233,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(14233,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(14233,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(14233,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(14233,36,'cp 18','f6h5 f4g3 h5g3 h2g3 g7g6 c2c4 f8g7 e2e3 e8g8 a2a3 c8g4 c4d5 d8d5 f1c4 d5d8 c4e2 g4d7 d1b3 e7e5 d4d5 c6e7 f3g5 h7h6 g5e4 a8b8 a1d1 e7c8 e4c5 h6h5 c5d7 d8d7 d2e4 c8d6 e4c5');
INSERT INTO evaluations VALUES(14610,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(14610,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(14610,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14610,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(14610,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(14610,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14610,36,'cp 65','d2d4 f8e7 e1g1 e8g8 d4d5 c6b8 d1e2 c7c6 a2a4 c6d5 c4d5 f6d5 c3d5 b8c6 f1d1 f7f5 e4f5 c8f5 b2b4 e7f6 c1b2 a8c8 c2c4 a7a5 b4b5 c6b4 b2a3 b4c2 a1a2 c2a3 a2a3 d8e8 a3e3 f5g6 d5f6 f8f6 f3d2 e8f7 d2e4 g6e4 e3e4 c8f8');
INSERT INTO evaluations VALUES(14643,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14643,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(14643,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(14643,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(14643,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(14643,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(14643,36,'cp 57','e5d4 f3d4 f8e7 c1f4 e8g8 e1g1 f6d7 f4e3 d7b6 c4b3 c6d4 e3d4 c8e6 b3e6 f7e6 d1g4 e6e5 d4b6 a7b6 a2a4 c7c6 g4e6 g8h8 e6b3 f8f6 c3e2 d8c7 a1d1 a8f8 d1d2 h6h5 c2c4 h5h4 b3d1 d6d5 c4d5 e7c5 d5c6 c5f2 g1h1 b7c6');
INSERT INTO evaluations VALUES(14661,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(14661,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14661,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14661,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(14661,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(14661,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(14661,36,'cp 50','f3d4 f8e7 e1g1 e8g8 f1e1 c6d4 d1d4 c7c6 a2a4 d8b6 d4d3 c8e6 c1e3 e6c4 d3c4 b6c7 e3d4 f8e8 c3d5 f6d5 e4d5 c7d7 d4c3 e7f8 c4d3 c6d5 d3d5 d7c6 d5c6 b7c6 g1f1 e8e1 c3e1 d6d5 a1b1 a8e8 b2b3 f8d6 b1d1 a7a6 d1d3');
INSERT INTO evaluations VALUES(14655,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(14655,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(14655,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(14655,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(14655,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(14655,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(14655,36,'cp 56','f8e7 e1g1 c6d4 d1d4 e8g8 c1f4 f6d7 a1d1 e7f6 d4d2 d7b6 c4d3 c8d7 b2b3 d7c6 a2a4 f8e8 f2f3 a7a6 a4a5 b6d7 d3c4 d7e5 c4d5 d8e7 b3b4 e5g6 f4g3 a8d8 f1e1 f6h4 g3h4 e7h4 d5b3 g6f4 c3e2 f4e2 d2e2 g8f8 e2d2 e8e5 c2c4 h4f6');
INSERT INTO evaluations VALUES(31982,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(31982,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(31982,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(31982,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(31982,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(31982,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(31982,36,'cp 85','d2d4 e7e6 f1d3 f5d3 d1d3 c7c6 h2h4 b8d7 c1d2 g8f6 e1c1 f8e7 c1b1 e8g8 h1g1 b7b5 g2g4 b5b4 g4g5 f6d5 c3e4 c6c5 d4c5 d7c5 d3d4 c5e4 d4e4 d8c7 f3e5 a8c8 h4h5 f8d8 d2c1 e7d6 e5g4 d6f4 h5h6 f4c1 d1c1 a7a5 h6g7');
INSERT INTO evaluations VALUES(50953,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50953,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(50953,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(50953,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50953,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(50953,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(50953,36,'cp 8','e2e4 g8f6 f1d3 c7c5 d4c5 f8c5 g1f3 d5e4 c3e4 f6e4 d3e4 d8d1 a1d1 b8d7 e1g1 d7f6 f3d2 f6e4 d2e4 c5f8 f1e1 c8d7 e4d6 f8d6 f4d6 e8c8 d6e7 d8e8 e7b4 d7c6 b4a5 e8e7 f2f4 h8e8 g1f2 f7f6 d1d6 e6e5 f4e5 f6e5 e1d1 e7d7 d6d7');
INSERT INTO evaluations VALUES(57581,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57581,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(57581,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(57581,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57581,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(57581,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57581,36,'cp 0','g8f6 e2e3 b7b6 c2c4 e7e6 g1f3 c8b7 f1e2 f8b4 c1d2 b4d2 d1d2 f6e4 d2c2 d7d6 e1g1 e8g8 f3d2 b8d7 e2f3 e4d2 b1d2 b7f3 d2f3 d7f6 a1d1 d8e7 f1e1 c7c5 h2h3 a8e8 d4c5 b6c5 f3d2 f6h5 d2f3');
INSERT INTO evaluations VALUES(50095,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50095,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50095,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50095,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50095,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50095,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(50095,36,'cp 8','e2e3 b7b6 f1e2 c8b7 g1f3 e7e6 e1g1 g7g6 c2c4 c7c5 b1c3 f8g7 a2a3 c5d4 e3d4 f6e4 c3b5 b8a6 c1e3 e8g8 d1b3 a6c7 a3a4 c7b5 a4b5 d8c7 a1a3 a7a6 b5a6 b7a6 f1a1 a6b7 f3e5 a8a3 a1a3 d7d6 e5f3 f8a8 a3a8 b7a8 d4d5 e6d5');
INSERT INTO evaluations VALUES(50093,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50093,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(50093,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(50093,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50093,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(50093,36,'cp 8','b7b6 f1e2 g7g6 g1f3 c7c5 c2c4 f8g7 b1c3 e8g8 e1g1 c8b7 c1d2 e7e6 d2e1 f6e4 a2a3 d8e7 e1h4 g7f6 h4f6 e4f6 d1d3 b7e4 d3d2 b8c6 a1d1 a8d8 f3g5 d7d5 c4d5 e6d5 c3e4 f5e4 d4c5 b6c5 d2c3 d5d4 c3c4 g8h8 g5e6');
INSERT INTO evaluations VALUES(27263,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(27263,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(27263,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(27263,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(27263,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(27263,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(27263,36,'cp 100','b5d7 d8d7 d1d4 g8e7 c1e3 e7c6 d4d3 c6e5 d3e2 e5g4 e3f4 e8c8 e1c1 d7e6 c1b1 c8b8 h1e1 g7g6 e4e5 f8h6 f4h6 g4h6 e5d6 d8d6 d1d6 e6e2 e1e2 c7d6 c3d5 h8f8 c2c4 h6f5 b2b3 b8c8 b3b4 f5d4 e2e7 d4c6 e7c7 c8b8 b4b5');
INSERT INTO evaluations VALUES(27145,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(27145,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(27145,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(27145,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(27145,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(27145,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(27145,36,'cp 109','d8d7 d1d4 g8e7 c1e3 e7c6 d4c4 g7g6 e3d4 c6d4 c4d4 h8g8 e1c1 b7b6 c3d5 f8g7 d4c4 e8c8 a2a4 c8b8 d5b4 d8e8 a4a5 c7c5 b4d5 b6b5 c4d3 c5c4 d3f3 c4c3 d5c3 g7c3 b2c3 d7c7 a5a6 e8e5 c1b2 g8e8');
INSERT INTO evaluations VALUES(27668,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(27668,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(27668,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(27668,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(27668,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(27668,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(27668,36,'cp 102','d1d4 g8e7 c1e3 e7c6 d4d3 c6e5 d3e2 e5c6 e1c1 e8c8 c3d5 g7g6 e3g5 f8e7 d5e7 c6e7 e2c4 d8e8 e4e5 d6d5 c4d3 c8b8 h1e1 e8d8 f2f4 d7e6 g2g4 d8d7 f4f5 g6f5 g5e7 e6e7 g4f5 d7d8 d3h3 h8e8 h3h7 e8h8 h7g7 d8g8 f5f6 e7e6');
INSERT INTO evaluations VALUES(27665,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(27665,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(27665,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(27665,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(27665,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(27665,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(27665,36,'cp 107','g8e7 c1e3 e7c6 d4d3 c6e5 d3e2 e5g4 e3f4 e8c8 f2f3 g4e5 e1c1 d8e8 c3d5 f7f5 h1e1 c8b8 e2d2 f5e4 e1e4 e5c6 e4e8 d7e8 b2b4 c6e5 d2c3 c7c6 d5e3 e8e6 b4b5 c6b5 e3d5 e6d7 f4e5 d6e5 c3e5 f8d6 d5f6 g7f6 d1d6 f6e5 d6d7');
INSERT INTO evaluations VALUES(44552,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44552,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44552,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44552,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44552,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44552,36,'cp 0','d7d5 c2c4 c5d4 e3d4 g7g6 c4d5 f6d5 d1b3 f8g7 f1c4 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(56203,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(56203,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(56203,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(56203,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(56203,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(56203,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(56203,36,'cp -20','g8f6 h2h3 f8d6 f1d3 e8g8 g1e2 b8d7 e1g1 f8e8 b1d2 d7f8 f1e1 c7c6 e2g3 e8e1 d1e1 c8e6 d2f3 d8c7 f3e5 f6d7 f2f4 f7f6 e5d7 e6d7 e1f2 g7g6 c1d2 a8e8 b2b3 f8e6 a1f1 f6f5 c3c4 c7c8 g3e2 c8d8 f1e1 d8f6 a2a4');
INSERT INTO evaluations VALUES(48667,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48667,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(48667,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(48667,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48667,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48667,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(48667,36,'cp -6','f1d3 f8d6 g1e2 e8g8 b1d2 f8e8 e1g1 c8g4 f1e1 g4h5 d2f1 h5g6 g2g3 b8d7 c1f4 d7f8 f4d6 d8d6 e2f4 e8e1 d1e1 a8e8 e1d2 g6e4 a1e1 f8e6 f4e6 e8e6 f1e3 e4d3 d2d3 c7c6 e3g2 e6e1 g2e1 h7h6 d3f3 f6e4 e1d3');
INSERT INTO evaluations VALUES(48668,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(48668,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(48668,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(48668,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(48668,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(48668,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(48668,36,'cp -11','f8d6 g1e2 e8g8 e1g1 f8e8 b1d2 c8g4 f1e1 g4h5 d2f1 h5g6 g2g3 b8d7 c1f4 d7f8 f4d6 d8d6 e2f4 f8e6 f4g6 h7g6 f1d2 e6g5 e1e8 a8e8 h2h4 g5e4 g1g2 a7a6 d3e4 d5e4 d2c4 d6e6 c4e3 f6d5 e3d5 e6d5 d1b3 d5b3 a2b3 c7c6 b3b4');
INSERT INTO evaluations VALUES(41708,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(41708,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(41708,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(41708,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(41708,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(41708,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(41708,36,'cp 0','g1e2 e8g8 e1g1 f8e8 b1d2 c8g4 f1e1 c7c6 d2f1 b8d7 c1g5 d8c7 d1c2 h7h6 g5h4 g4e2 e1e2 e8e2 c2e2 a8e8 e2f3 c7b6 b2b4 e8e6 d3f5 e6e8');
INSERT INTO evaluations VALUES(20030,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(20030,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20030,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20030,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(20030,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(20030,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(20030,36,'cp 0','g7g6 d2d4 c6d4 f3d4 c5d4 d1d4 g8f6 e1g1 f8g7 b1c3 e8g8 a2a4 d8c7 d4d1 c8d7 c1e3 f8c8 e3d4 d7g4 d1d2 g4e6 f1c1 c7a5 c1d1 e6g4 f2f3 g4e6 c3e2 a5d2 d1d2 f6d7 f3f4 a7a5 d4g7 g8g7 e2d4 d7c5 e4e5 d6e5 d4e6 f7e6');
INSERT INTO evaluations VALUES(19677,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19677,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(19677,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(19677,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19677,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19677,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19677,36,'cp 4','d2d4 c5d4 f3d4 c6d4 d1d4 g8f6 e1g1 f8g7 b1c3 e8g8 a2a4 d8c7 d4d1 c8g4 c3d5 f6d5 d1g4 d5b4 c2c3 b4d3 g4e2 d3c1 f1c1 a8b8 c1c2 a7a6 a4a5 f8c8 g2h3 e7e6 c2d2 c8d8 h3g2 h7h5 h2h4 b8c8 g2h3 c7c4 e2d1 c4c7 f2f4 g7f6 g1h2');
INSERT INTO evaluations VALUES(42983,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(42983,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(42983,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42983,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42983,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42983,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(42983,36,'cp 23','d7d5 c1f4 e8g8 e2e3 b7b6 f1d3 c7c5 e1g1 b8c6 d4c5 b6c5 a1c1 c8b7 c4d5 f6d5 c3d5 e6d5 f3e5 c6e5 f4e5 g7g6 b2b3 e7d6 e5d6 d8d6 h2h4 a8e8 d1c2 d5d4 e3e4 d6f4 c2c5 b7e4 c5d4 f4h4 d3e4 e8e4');
INSERT INTO evaluations VALUES(57039,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(57039,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(57039,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(57039,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(57039,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(57039,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(57039,36,'cp -52','d7d5 e4d5 e6d5 d2d4 g8f6 b1c3 f8e7 c1e3 e8g8 d1d2 f8e8 g1e2 c7c6 e2g3 b8d7 f1e2 a7a5 e1g1 a5a4 a2a3 e7f8 e2d3 d7b6 a1e1 h7h6 e3f2 e8e1 f1e1 f6e8 c3d1 e8d6 d1e3 c8e6 g3e2 b6c4 d3c4 d6c4 e3c4 d5c4');
INSERT INTO evaluations VALUES(23133,30,'cp 242',NULL);
INSERT INTO evaluations VALUES(23133,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(23133,32,'cp 247',NULL);
INSERT INTO evaluations VALUES(23133,33,'cp 259',NULL);
INSERT INTO evaluations VALUES(23133,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(23133,35,'cp 251',NULL);
INSERT INTO evaluations VALUES(23133,36,'cp 250','b1c3 d7d5 c3d5 c7c6 d5c3 b7b5 c4b3 a7a5 a2a3 a5a4 b3a2 g8f6 e1g1 f8e7 d2d4 e8g8 c1e3 c8e6 a2e6 f7e6 a1c1 d8e8 d1e2 a8c8 c1d1 c8d8 f2f3 f6d7 e3f2 d7b6 d4d5 b6c4 d5e6 e8g6 d1d8 f8d8');
INSERT INTO evaluations VALUES(23130,30,'cp 252',NULL);
INSERT INTO evaluations VALUES(23130,31,'cp 241',NULL);
INSERT INTO evaluations VALUES(23130,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(23130,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(23130,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(23130,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(23130,36,'cp 266','c7c6 e1g1 g8f6 f1e1 f8e7 e4e5 f6d5 c3d5 c6d5 c4d5 e8g8 d2d4 d8a5 d5b3 d7d6 c1f4 a5b4 d1f3 d6e5 f4e5 e7d6 e5d6 b4d6 a1c1 a7a5 e1e8 c8e6 e8a8 f8a8 b3e6 d6e6 f3b7 a8d8 d4d5 e6e4 b7b3 a5a4 b3c4 e4c4 c1c4 g7g6 g1f1 d8d5 c4a4 d5d2 a4a8 g8g7 b2b3');
INSERT INTO evaluations VALUES(38799,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38799,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38799,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38799,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38799,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38799,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38799,36,'cp 0','b4c3 b2c3 d5e4 d1g4 g8f6 g4g7 h8g8 g7h6 c8d7 g1e2 d7a4 a1b1 a4c2 b1b7 c2d3 e2g3 g8g6 h6f4 f6d5 f4e5 b8c6 e5h8 e8d7 h8h7 d8f6 f1d3 e4d3 g3e4 f6f5 e4g3');
INSERT INTO evaluations VALUES(38845,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38845,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38845,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38845,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38845,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38845,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38845,36,'cp 0','b2c3 d5e4 d1g4 g8f6 g4g7 h8g8 g7h6 g8g6 h6h4 g6g4 h4h6');
INSERT INTO evaluations VALUES(38844,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38844,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38844,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38844,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38844,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38844,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38844,36,'cp 0','d5e4 d1g4 g8f6 g4g7 h8g8 g7h6 g8g6 h6h4 g6g4 h4h6');
INSERT INTO evaluations VALUES(38907,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38907,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38907,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38907,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38907,36,'cp 0','d1g4 g8f6 g4g7 h8g8 g7h6 g8g6 h6h4 g6g4 h4h6');
INSERT INTO evaluations VALUES(38904,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38904,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38904,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38904,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38904,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38904,36,'cp 0','g8f6 g4g7 h8g8 g7h6 c8d7 g1e2 d7a4 a1b1 a4c2 b1b7 c2d3 e2g3 g8g6 h6f4 f6d5 f4e5 b8c6 e5h8 e8d7 h8h7 d8f6 g3h5 f6f5 h5g7 f5f6');
INSERT INTO evaluations VALUES(42190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42190,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42190,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42190,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42190,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42190,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42190,36,'cp 0','g4g7 h8g8 g7h6 c8d7 g1e2 d7a4 a1b1 a4c2 b1b7 c2d3 e2g3 g8g6 h6f4 f6d5 f4e5 b8c6 e5h8 e8d7 h8h7 d8f6 g3h5 f6f5 h5g3');
INSERT INTO evaluations VALUES(41580,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41580,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41580,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41580,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41580,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41580,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41580,36,'cp 0','h8g8 g7h6 c8d7 g1e2 d7a4 a1b1 a4c2 b1b7 c2d3 e2g3 g8g6 h6f4 f6d5 f4e5 b8c6 e5h8 e8d7 h8h7 d8f6 g3h5 f6f5 h5g3');
INSERT INTO evaluations VALUES(39654,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39654,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39654,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39654,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39654,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39654,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39654,36,'cp 0','g7h6 c8d7 g1e2 d7a4 a1b1 a4c2 b1b7 c2d3 e2g3 g8g6 h6f4 f6d5 f4e5 b8c6 e5h8 e8d7 h8h7 d8f6 g3h5 f6f5 h5g3');
INSERT INTO evaluations VALUES(39651,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39651,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39651,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39651,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39651,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39651,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39651,36,'cp 0','g8g6 h6h4 g6g4 h4h6');
INSERT INTO evaluations VALUES(51076,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(51076,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(51076,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(51076,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(51076,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(51076,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(51076,36,'cp 105','d7d6 c2c4 e7e5 b1c3 e5d4 f3d4 f8e7 c1e3 g8f6 f1e2 b8c6 e1g1 e8g8 d4c2 f8e8 f2f3 c6e5 a2a4 e7f8 a4a5 c8e6 b2b3 c7c5 d1d2 b7b6 a5b6 d8b6 a1b1 e5c6 g1h1 a6a5 f1d1 f6d7 e3f2 g8h8 f3f4 f7f5 e2f3 d7f6 e4f5');
INSERT INTO evaluations VALUES(50810,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(50810,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(50810,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(50810,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(50810,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(50810,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(50810,36,'cp 110','c2c4 e7e5 b1c3 e5d4 f3d4 g8f6 f1e2 f8e7 c1e3 b8c6 d4c2 e8g8 f2f3 c6e5 e1g1 b7b5 c4b5 a6b5 e2b5 c7c6 b5e2 d6d5 e4d5 f6d5 c3d5 c6d5 g1h1 c8a6 e2a6 a8a6 a2a4 e5c6 d1d3 d8a8 b2b3 e7f6 a1d1 f8d8 c2d4 d8e8 d4c6 a6c6');
INSERT INTO evaluations VALUES(9695,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(9695,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(9695,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(9695,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(9695,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(9695,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(9695,36,'cp -42','g1e2 e8g8 a2a3 h7h6 c1d2 d7d6 b1c3 g8h7 c4a2 c8g4 f3g3 g4e6 a2e6 f7e6 f2f3 d8e7 g3h3 d6d5 e1g1 f8f7 b2b4 a8f8 g1h1 e7d7 b4b5 c6d8 a1b1 b7b6 a3a4 c7c6 c3d1 c6b5 a4b5 d5d4 d1f2');
INSERT INTO evaluations VALUES(9694,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(9694,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(9694,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(9694,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(9694,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(9694,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(9694,36,'cp -40','d7d6 h2h3 e8g8 a2a3 h7h6 b1c3 c8e6 c3d5 e6d5 e4d5 e5e4 f3e3 c6e5 c4a2 e4d3 c2d3 f8e8 e1g1 d8d7 d3d4 e5c6 e3d3 c6e7 e2c3 a7a6 c1d2 h6h5 a1e1 a8b8 e1c1 b7b5 b2b4 e7f5 a2b1 b8a8 f1e1 e8e1 c1e1 a8e8 g1f1 e8e1 d2e1 d7e8 d3d2 f5e7 d2e2');
INSERT INTO evaluations VALUES(16022,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(16022,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(16022,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(16022,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(16022,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(16022,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(16022,36,'cp 52','a2a4 c7c6 d2d4 e7e5 f1c4 f8e7 e1g1 e8g8 h2h3 h7h6 f1e1 f8e8 a4a5 e5d4 f3d4 e7f8 c1f4 d7c5 d1f3 c5e6 d4e6 c8e6 c4d3 f6d7 f3g3 d8f6 f4d6 f8d6 g3d6 d7e5 e1e3 a8d8 d6b4 b7b5 a5b6 a7b6 a1d1 b6b5 d3f1 e6c4 d1d8 e8d8');
INSERT INTO evaluations VALUES(26679,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(26679,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(26679,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(26679,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(26679,34,'cp 293',NULL);
INSERT INTO evaluations VALUES(26679,35,'cp 308',NULL);
INSERT INTO evaluations VALUES(26679,36,'cp 297','d4d5 c6a5 d1a4 c7c6 b2b4 b7b5 a4a5 d8a5 b4a5 b5b4 c3a4 c6d5 e2e4 e7e6 f3e5 g8f6 e5g4 f6g4 f1e2 f7f5 e4d5 e6d5 c1e3 e8d7 e1c1 d7c6 d1d5 g4e3 f2e3 c6d5 e2f3 d5e6 f3a8 b4b3 a8f3 b3a2 c1b2 f8d6 b2a2 h8b8 h1c1 b8b3 c1c4 b3a3 a2b1 a3e3');
INSERT INTO evaluations VALUES(26519,30,'cp 278',NULL);
INSERT INTO evaluations VALUES(26519,31,'cp 297',NULL);
INSERT INTO evaluations VALUES(26519,32,'cp 303',NULL);
INSERT INTO evaluations VALUES(26519,33,'cp 292',NULL);
INSERT INTO evaluations VALUES(26519,34,'cp 303',NULL);
INSERT INTO evaluations VALUES(26519,35,'cp 273',NULL);
INSERT INTO evaluations VALUES(26519,36,'cp 246','c6b8 f3e5 g8f6 e5g4 f6g4 e2e4 g4e5 f2f4 e5d3 f1d3 c4d3 c1e3 e7e6 d1d3 d8d7 e1g1 e6d5 a1c1 f8e7 f1d1 c7c6 e4d5 e8g8 d5d6 e7h4 c3e4 b8a6 a2a3 f8e8 e3d4 a8d8 g2g3 f7f5 e4f2 h4f6 d4f6 g7f6 b2b4 e8e6 d3f5 e6d6 f5h5 d7e7 d1d6 e7d6');
INSERT INTO evaluations VALUES(34580,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(34580,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(34580,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(34580,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(34580,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(34580,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(34580,36,'cp 143','d2d4 c5d6 g1f3 g8f6 f1c4 a7a6 e1g1 e7e6 d1e2 f8e7 f1e1 e8g8 c3e4 f6e4 e2e4 b8d7 c4d3 g7g6 c1f4 d6b6 f4h6 f8e8 h2h3 c7c5 d4c5 e7c5 e4h4 c5e7 f3g5 d7f6 c2c4 b6c7 a1d1 c8d7 c4c5 c7c5');
INSERT INTO evaluations VALUES(34577,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(34577,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(34577,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(34577,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(34577,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(34577,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(34577,36,'cp 137','c5d6 f1c4 g8f6 g1f3 a7a6 e1g1 e7e6 c1g5 b8d7 d1e2 f8e7 d4d5 d7b6 d5e6 c8e6 c4b3 e6b3 c2b3 e8g8 a1d1 d6e6 e2e6 f7e6 f1e1 a8d8 e1e6 d8d1 c3d1 g8f7 e6e2 h7h6 g5d2 e7d6 g1f1 b6d5 g2g3 g7g5 d2c3 d5c3 b2c3 g5g4 f3d2');
INSERT INTO evaluations VALUES(34539,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(34539,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(34539,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(34539,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(34539,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(34539,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(34539,36,'cp 149','g1f3 g8f6 f1c4 a7a6 e1g1 e7e6 d1e2 f8e7 a2a4 e8g8 c3e4 d6d8 e4f6 e7f6 f1d1 b7b6 d4d5 d8e7 c1f4 e6d5 e2e7 f6e7 c4d5 a8a7 a4a5 b6b5 f3e5 c8b7 d5b3 b7c8 f4g3 c8e6 b3e6 f7e6 d1d2 e7f6 h2h3 f6g5 d2d1 g5f6 c2c3 f8e8 h3h4 f6e5 g3e5 b8c6 e5g3');
INSERT INTO evaluations VALUES(20198,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(20198,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(20198,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(20198,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(20198,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(20198,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(20198,36,'cp -53','c8d7 b5c6 d7c6 d2d4 g8e7 b1c3 e7f5 c3e2 c6b5 c2c3 a8c8 e2g3 f5h4 f3h4 d8h4 d1b3 b5a6 b3a4 c8c6 c1e3 c5d4 c3d4 h4e7 a1c1 e7b4 a4b4 f8b4 e1f2 h7h5 c1c6 b7c6 g3e2 e8d7 h1d1 h8b8 b2b3 b4a3 e2g3 g7g6 g3e2 a6b5 e3c1 a3e7 e2c3 a7a5 c1e3 b5a6 h2h3 h5h4 c3a4');
INSERT INTO evaluations VALUES(25936,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(25936,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(25936,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(25936,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(25936,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(25936,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(25936,36,'cp -40','b5c6 d7c6 d2d4 g8h6 b1c3 h6f5 c3e2 c6b5 c2c3 a8c8 e1f2 h7h5 c1e3 f8e7 h1e1 a7a5 e2g3 f5e3 e1e3 b5c4 b2b3 c5d4 f3d4 c4a6 d1d2 e7c5 f2g1 h5h4 g3f1 h4h3 g2g3 a5a4 e3f3 d8b6 f1e3 c5d4 d2d4');
INSERT INTO evaluations VALUES(25902,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(25902,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(25902,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(25902,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(25902,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(25902,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(25902,36,'cp -54','d7c6 d2d4 g8h6 b1c3 h6f5 c3e2 c6b5 c2c3 a8c8 e1f2 h7h5 c1e3 f8e7 h1e1 g7g6 e2g3 f5e3 e1e3 d8b6 g3e2 b5c4 d1d2 c4e2 e3e2 c5d4 c3d4 e8g8 a1c1 g8g7 g2g3 c8c4 c1c4 d5c4 h2h3 f8c8 g3g4 h5g4');
INSERT INTO evaluations VALUES(26360,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(26360,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(26360,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(26360,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(26360,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(26360,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(26360,36,'cp -49','d2d4 g8e7 b1c3 e7f5 c3e2 c6b5 c2c3 a8c8 e2g3 f5h4 f3h4 d8h4 d1b3 b5a6 b3a4 c8c6 c1e3 c5d4 c3d4 h4e7 a1c1 e7b4 a4b4 f8b4 e1f2 h7h5 c1c6 b7c6 g3e2 e8d7 h1d1 h8b8 e2c1 b4e7 b2b3 a6b5 h2h3 a7a5 a2a4 b5a6 g2g4 h5g4 h3g4 e7a3 e3d2 c6c5 d2a5');
INSERT INTO evaluations VALUES(41357,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(41357,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(41357,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(41357,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(41357,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(41357,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(41357,36,'cp 41','a7a6 d1d2 b7b5 e1c1 b8d7 h2h4 c8b7 g1h3 c7c5 d4c5 d6c5 e3h6 g7h6 d2h6 d8c7 h3g5 h8f8 g5h7 f6h7 h6h7 b5b4 c3e2 e8c8 h4h5 g6h5 h7h5 d7e5 d1d8 f8d8 c1b1 c5c4 e2c1 c8b8 f1e2 a6a5 h1d1 d8d1 e2d1 c4c3 b2b3 c7c5 h5h6');
INSERT INTO evaluations VALUES(47922,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(47922,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(47922,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(47922,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(47922,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(47922,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47922,36,'cp -38','h7h6 g5f4 d6d5 f1e2 f8d6 f4d6 d8d6 e1g1 e8g8 d3d4 c8f5 c2c3 b8d7 f1e1 f8e8 b1d2 a7a5 a2a4 c7c6 e2f1 d6c7 h2h3 e8e1 f3e1 a8e8 e1c2 f5h7 b2b4 d7f8 b4a5 c7a5 c2b4 h7g6 d2f3 f6e4 d1b3 e4d6');
INSERT INTO evaluations VALUES(53844,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(53844,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(53844,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(53844,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(53844,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(53844,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(53844,36,'cp 165','d1d4 d8b6 c1f4 b6d4 f3d4 c8e6 e1c1 b8d7 f2f3 h7h5 a2a4 d7e5 f4e3 h5h4 h2h3 e6d7 f3f4 e5g6 f1c4 g8f6 d4f3 d7e6 c4e6 f7e6 f3g5 e8d7 h1e1 a8e8 e4e5 f6d5 c3d5 c6d5 f4f5 e6f5 e5e6 e8e6 g5e6 d7e6 e3a7 g6e5 a7b6');
INSERT INTO evaluations VALUES(53843,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(53843,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(53843,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(53843,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(53843,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(53843,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(53843,36,'cp 172','d8b6 c1f4 b6d4 f3d4 c8e6 e1c1 b8d7 f1e2 g8f6 f4d6 f8d6 d4e6 f7e6 d1d6 e8e7 h1d1 d7e5 h2h4 h7h6 h4h5 a7a5 a2a4 e5f7 d6d2 a8d8 d2d8 h8d8 d1d8 e7d8 c1d2 e6e5 d2e3 f6d7 b2b3 d7c5 f2f4 d8e7 g2g3 f7d6 e2f3 b7b6 c3e2 d6f7 e2c1 c5e6 c1d3');
INSERT INTO evaluations VALUES(47236,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(47236,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(47236,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(47236,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(47236,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(47236,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(47236,36,'cp 60','e4e5 f6g8 d2d4 e7e6 c2c3 c7c5 f1d3 c5d4 c3d4 b8c6 g1e2 c6b4 b1c3 b4d3 f3d3 c8d7 e1g1 a8c8 e2g3 h7h5 h2h3 d8b6 c1g5 h5h4 g3e2 f8e7 d3d2 b6d8 g5e3 g8h6 e3h6 h8h6 c3d1 h6h8 d1e3 d8b6 f1c1 c8c1 a1c1 b6d8 e3g4 e7g5');
INSERT INTO evaluations VALUES(47298,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(47298,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(47298,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(47298,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(47298,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(47298,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(47298,36,'cp 45','f6g8 d2d4 e7e6 c2c3 c7c5 f1d3 b8c6 g1e2 c5d4 c3d4 c6b4 b1c3 b4d3 f3d3 c8d7 e1g1 a8c8 c1d2 a7a6 a1c1 g8e7 h2h4 e7f5 h4h5 d8h4 e2f4 c8c4 c3e2 h4d8 g2g4 d7b5 c1c4 b5c4 d3c3 f5h4 b2b3 c4e2 f4e2 f8e7 f1c1 e8g8');
INSERT INTO evaluations VALUES(6683,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(6683,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(6683,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(6683,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(6683,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(6683,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(6683,36,'cp -15','b4a5 e1g1 a5b6 c1a3 g8e7 e4e5 e8g8 f1e1 h7h6 c4b3 f8e8 b1d2 c6a5 d2c4 a5b3 a2b3 e7c6 d1c2 d7d6 e5d6 c8g4 e1e8 d8e8 a1e1 g4e6 h2h3 e8d8 c4b6 a7b6 d6c7 d8c7 a3b2 c7f4 b2c1 f4f6 e1e3 b6b5 c2e4 f6f5 e4f5 e6f5');
INSERT INTO evaluations VALUES(6738,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(6738,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(6738,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(6738,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(6738,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(6738,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(6738,36,'cp -11','e1g1 a5b6 e4e5 g8e7 c1a3 e8g8 f1e1 c6a5 c4d3 d7d5 e5d6 c7d6 a3d6 f8e8 b1a3 c8f5 a3c4 a5c4 d3c4 f5g6 d6e7 e8e7 d1d8 b6d8 c4d5 g8f8 e1e7 d8e7 f3e5 a8d8 e5g6 h7g6 d5b7 e7f6 g2g3 d8d2 a1e1 f6e7 b7c6 d2a2 e1e3 a2c2 g1g2 c2a2');
INSERT INTO evaluations VALUES(6737,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(6737,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(6737,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(6737,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(6737,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(6737,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(6737,36,'cp -16','a5b6 e4e5 g8e7 c1a3 e8g8 f1e1 c6a5 c4d3 d7d5 e5d6 c7d6 a3d6 f8e8 b1a3 c8f5 a3c4 a5c4 d3c4 f5g6 d1d2 e7f5 d6f4 d8d2 f3d2 e8d8 d2e4 h7h6 c4b3 d8d7 e1d1 a8d8 d1d7 d8d7 a1e1 d7d8 h2h3 d8d7 e4d2');
INSERT INTO evaluations VALUES(18726,30,'cp -199',NULL);
INSERT INTO evaluations VALUES(18726,31,'cp -185',NULL);
INSERT INTO evaluations VALUES(18726,32,'cp -176',NULL);
INSERT INTO evaluations VALUES(18726,33,'cp -183',NULL);
INSERT INTO evaluations VALUES(18726,34,'cp -182',NULL);
INSERT INTO evaluations VALUES(18726,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(18726,36,'cp -201','d7d5 e4d5 g8f6 h5h4 e6d5 c4b5 d8e7 e1d1 e7d8 f3e5 d8c7 h1e1 f8e7 e5c6 b7c6 h4g3 c7b7 e1e7 e8e7 b5f1 h7h5 b2b3 h5h4 g3e3 e7d8 e3c5 b7b6 c5b6 a7b6 h2h3 f6e4 d1e1 e4c5 a2a4 c8f5 d2d3 c5e6 f1e2 e6d4 e1d1 f7f6 b1c3');
INSERT INTO evaluations VALUES(22898,30,'cp -110',NULL);
INSERT INTO evaluations VALUES(22898,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(22898,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(22898,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(22898,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(22898,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(22898,36,'cp -114','g8f6 g1f3 f6e4 d1b3 d8f6 e1g1 e4c5 b3c2 d4d3 c2d1 f8e7 f1e1 e8g8 b2b4 c5e6 c4d3 d7d5 d1c2 h7h6 b1d2 e6f4 d3h7 g8h8 b4b5 g7g6 b5c6 h8h7 c1b2 c8f5 c2b3 f4d3 e1e7 f6e7 b2a3 d3c5 b3b7 f8b8 a3c5 e7c5 b7c7 c5c3 c7f7 c3g7 f7g7 h7g7 f3d4 b8b2 d4f5 g6f5 d2f3 g7f6');
INSERT INTO evaluations VALUES(16653,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(16653,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(16653,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(16653,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(16653,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(16653,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(16653,36,'cp -98','g1f3 f6e4 d1b3 d8f6 e1g1 e4c5 f1e1 f8e7 b3c2 d4d3 c2d1 h7h6 b2b4 c5e6 d1d3 e8g8 c4b3 f8e8 b3c2 e6f8 b4b5 c6a5 c1e3 d7d6 b1d2 f6g6 f3d4 g6d3 c2d3 f8e6 d4b3 a5b3 a2b3 c8d7 d3e4 e7f6 e1c1 c7c6 b5c6 b7c6 e4d3 d6d5 a1a7 a8a7 e3a7');
INSERT INTO evaluations VALUES(16652,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(16652,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(16652,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(16652,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(16652,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(16652,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(16652,36,'cp -115','f6e4 d1b3 d8f6 e1g1 f8e7 f1e1 e4c5 b3c2 d4d3 c2d1 e8g8 b2b4 c5e6 c4d3 d7d6 d1c2 g8h8 c1e3 a7a5 b4b5 c6e5 f3e5 d6e5 b1d2 b7b6 a1d1 e6f4 e3f4 f6f4 g2g3 f4f6 d3e4 a8b8 e4c6 e7c5 d2e4 f6g6');
INSERT INTO evaluations VALUES(16860,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(16860,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(16860,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(16860,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(16860,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(16860,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(16860,36,'cp -105','d1b3 d8f6 e1g1 e4c5 f1e1 f8e7 b3c2 d4d3 c2d1 e8g8 b2b4 c5e6 b4b5 c6a5 c4d3 d7d5 e1e5 g7g6 e5d5 c7c6 b5c6 a5c6 d3e4 e6c5 c1g5 f6e6 e4c2 e7g5 f3g5 e6e7 b1a3 h7h6 g5f3 c8e6 d5d2 a7a6 f3d4 e7f6 d1f3 f6f3 d4f3 a8c8 f3d4 c6a5 f2f4 f8d8 h2h3 b7b5 a1d1 e6a2 f4f5 d8d6 f5g6 f7g6 c2g6 c5a4 a3b5 d6g6 d2a2 a6b5');
INSERT INTO evaluations VALUES(40393,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40393,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40393,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40393,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40393,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(40393,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40393,36,'cp 21','e8g8 d1d2 b8c6 e1g1 c8d7 h2h3 c6d4 e3d4 d7c6 f1d1 e7e5 d4e3 f6e4 c3e4 c6e4 c2c3 e4c6 d2d6 d8h4 d6b4 h4b4 c3b4 a7a6 b2b3 f8c8 a2a4 g7f8 b4b5 a6b5 a4b5 a8a1 d1a1 c6d5 a1d1 d5e6 e2f3 c8b8 e3a7 b8a8 a7b6 a8a3 f3b7 a3b3');
INSERT INTO evaluations VALUES(36009,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(36009,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(36009,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36009,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36009,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36009,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36009,36,'cp 19','e1g1 b8c6 a2a4 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 a4a5 c8e6 f2f4 a8c8 c2c3 c8c5 e2f3 f8b8 d1d2 b7b6 a5b6 a7b6 d2d4 c5b5 b2b4 b8c8 a1c1 h7h5 h2h3 c8c7 g1f2 e6c4 f3e2 c4e2 f2e2 f7f5 e2d3 h5h4');
INSERT INTO evaluations VALUES(43234,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43234,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43234,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43234,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43234,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(43234,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43234,36,'cp 12','g2g3 d7d5 f1g2 a7a5 a2a4 e8g8 e1g1 b8d7 f1e1 c7c6 b1d2 c6c5 e2e3 b7b6 e3e4 d5e4 f3g5 c5d4 c3d4 c8b7 d2e4 f6e4 g5e4 b7d5 c1g5 h7h6 g5e3 a8c8 d1d2 d7f6 e4f6 e7f6 g2d5 d8d5 e3h6');
INSERT INTO evaluations VALUES(43235,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43235,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43235,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43235,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43235,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43235,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43235,36,'cp 0','d7d5 f1g2 e8g8 e1g1 a7a5 a2a4 c7c6 b1d2 b8d7 f1e1 c6c5 e2e4 f6e4 d2e4 d5e4 e1e4 c5d4 c3d4 b7b6 c1g5 d7f6 e4e1 c8b7 a1c1 e7e6 f3e5 b7g2 g1g2 d8d5 d1f3 d5f3 g2f3 f6d5 g5d2 g7e5 d4e5 f8b8 c1c2 b6b5 e1a1 d5e7 f3e4 e7d5 e4f3');
INSERT INTO evaluations VALUES(29745,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(29745,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(29745,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(29745,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(29745,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(29745,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(29745,36,'cp 55','e7e6 b1c3 c7c5 d1a4 b8d7 f3e5 c5d4 e3d4 a7a6 e5g4 f6g4 c4e2 b7b5 a4b3 g4f6 e2f3 a8b8 c1f4 b8b6 e1g1 f8d6 f4e3 e8g8 a1c1 h7h6 g2g3 b6b8 c3e4 f6e4 f3e4 d7f6 e4f3 a6a5 c1c6 d8d7 f1c1 a5a4 b3d3 f6d5 f3d5 e6d5 g1g2 f8e8 d3e2');
INSERT INTO evaluations VALUES(30209,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(30209,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(30209,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(30209,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(30209,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(30209,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(30209,36,'cp 65','b1c3 c7c5 d1a4 b8d7 f3e5 c5d4 e3d4 a7a6 e5g4 f6g4 c4e2 b7b5 a4b3 g4f6 e2f3 a8b8 c1f4 b8b6 f4e3 f8d6 g2g3 e8g8 e1g1 h7h6 f1c1 b6b8 c3e4 f6e4 f3e4 d7b6 c1c6 b8c8 a1c1 c8c6 c1c6 d6e7 b3d3 b6d5 a2a3 d5e3 f2e3 b5b4 a3b4 e7b4 e4h7 g8h8');
INSERT INTO evaluations VALUES(30206,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(30206,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(30206,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(30206,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(30206,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(30206,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(30206,36,'cp 70','c7c5 d1a4 b8d7 f3e5 c5d4 e3d4 a7a6 e5g4 f6g4 c4e2 b7b5 a4b3 g4f6 e2f3 a8b8 c1f4 b8b6 f4e3 f8d6 e1g1 e8g8 a1c1 b6b8 g2g3 d7b6 d4d5 e6d5 f1d1 d6e5 e3c5 f8e8 c5b6 b8b6 c3d5 b6d6 g1g2 g7g6 d5f6 e5f6 d1d6 d8d6 f3d5 e8e2 d5f7 g8g7 b3f3 d6e7 c1c6 f6d4');
INSERT INTO evaluations VALUES(31912,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31912,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(31912,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31912,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31912,36,'cp 0','g4f3 g2f3 e7e6 e2e4 g8f6 b1c3 f8b4 d1d3 c7c5 d4c5 d8a5 c1d2 b8c6 d3b5 a5b5 f1b5 e8c8 b5c6 b7c6 h1g1 f6h5 a2a3 b4c5 c3a4 c5d6 d2e3 d8d7 e1c1 g7g6 b2b4 h5f4 g1h1 f4g2 e3c5 g2f4 c5e3');
INSERT INTO evaluations VALUES(31925,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31925,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(31925,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31925,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(31925,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(31925,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(31925,36,'cp 16','g2f3 g8f6 b1c3 c7c5 e2e4 e7e6 c1g5 b8c6 e4d5 c6d4 d5e6 d8e7 f1b5 d4b5 c3b5 a7a6 d1e2 e8c8 b5c3 e7e6 e2e6 f7e6 h1g1 f8e7 a1d1 h8f8 d1d8 c8d8 c3e4 f6e4 f3e4 g7g6 g5e3 d8d7 e1e2 d7c6 g1h1 e7f6 c2c3 f6e5 h3h4 e5f4 h4h5 f4e3');
INSERT INTO evaluations VALUES(31922,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31922,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(31922,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(31922,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31922,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(31922,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(31922,36,'cp -6','e7e6 e2e4 g8f6 b1c3 f8b4 a2a3 b4c3 b2c3 d5e4 a1b1 b8d7 f1g2 f6d5 b1b3 e4e3 f2e3 d8h4 e1f1 b7b6 f3f4 a8d8 d1g4 h4h6 g4g5 h6g5 f4g5 c7c5 a3a4 d5e7 f1f2 d8c8 c1d2 h7h6 g5h6 h8h6 a4a5 h6f6 f2e2 e6e5 h1g1 c5d4 c3d4 c8c2');
INSERT INTO evaluations VALUES(48391,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(48391,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(48391,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(48391,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(48391,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(48391,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(48391,36,'cp 35','e4e5 f6d7 c2c3 c7c5 d3d4 b8c6 f1d3 f8e7 e1g1 f7f6 e5f6 d7f6 f1e1 e8g8 a2a3 d8c7 h2h3 c8d7 b2b4 c5c4 d3c2 e7d6 b1d2 b7b5 a3a4 b5a4 c2a4 a7a5 a4c6 d7c6 b4a5 a8a5 a1a5 c7a5 e1e6 a5c7 e6e1 c6e8 f3e5 e8h5 g2g4 h5g6 e5g6');
INSERT INTO evaluations VALUES(48434,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(48434,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(48434,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(48434,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(48434,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(48434,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(48434,36,'cp 13','f6d7 d3d4 c7c5 c2c3 b8c6 f1d3 f7f6 e5f6 d7f6 e1g1 f8d6 d4c5 d6c5 c3c4 e8g8 c1g5 d8c7 g5f6 f8f6 b1c3 c6e5 f3e5 c7e5 d1e2 e5e2 d3e2 d5c4 a1d1 f6f4 g2g3 f4d4 d1d4 c5d4 f1d1 d4f6 c3e4 g8f8 e4f6');
INSERT INTO evaluations VALUES(49174,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(49174,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(49174,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(49174,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(49174,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(49174,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(49174,36,'cp 27','d3d4 c7c5 c2c3 f7f6 f1d3 b8c6 e5f6 d7f6 c1e3 f8e7 e1g1 e8g8 h2h3 d8c7 f1e1 b7b6 a2a3 e7d6 b2b4 c5c4 d3c2 b6b5 b1d2 a7a5 e3g5 a5b4 a3b4 a8a1 d1a1 h7h6 g5f6 g7f6 f3h4 c6e7 d2f3 e6e5 h4g6 e7g6 c2g6 c7g7');
INSERT INTO evaluations VALUES(49165,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(49165,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(49165,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(49165,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(49165,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(49165,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(49165,36,'cp 31','c7c5 c2c3 b8c6 f1d3 f7f6 e5f6 d7f6 c1e3 c5d4 c3d4 f8d6 b1c3 c8d7 e1g1 e8g8 e3g5 d8b6 f1e1 g8h8 f3e5 d7e8 e5c6 b6c6 h2h3 h7h6 d3b5 c6b6 b5e8 f6e8 c3a4 b6a6 g5h4 e8f6 a4c5 d6c5 d4c5 a8c8 a1c1 a6a2');
INSERT INTO evaluations VALUES(44008,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44008,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44008,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(44008,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44008,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44008,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(44008,36,'cp 17','c2c3 f7f6 f1d3 b8c6 e5f6 d7f6 c1e3 c5d4 c3d4 f8d6 e1g1 e8g8 h2h3 f6h5 f3g5 h5f6 b1d2 c8d7 f1e1 h7h6 g5f3 d7e8 a1c1 d8d7 d1e2 e8h5 d3b1 a8e8 e2d3 c6b4 d3c3 e8c8 c3b3 b4c6 f3h4 g7g5 h4g6 h5g6 b1g6 d7g7');
INSERT INTO evaluations VALUES(44000,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(44000,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44000,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(44000,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(44000,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(44000,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44000,36,'cp 27','f7f6 f1d3 b8c6 e5f6 d7f6 e1g1 f8d6 d4c5 d6c5 b2b4 c5d6 b4b5 c6e5 f3e5 d6e5 c1a3 d8c7 h2h3 e5h2 g1h1 h2f4 b1d2 e6e5 c3c4 c7a5 a3c1 a5c3 d1c2 c3c2 d3c2 c8d7 c2b3 e8c8 d2f3 d5c4 b3c4 h8e8 c1f4 e5f4 a2a4 c8b8');
INSERT INTO evaluations VALUES(50487,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(50487,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(50487,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(50487,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(50487,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(50487,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(50487,36,'cp -25','d7d5 e4d5 e6d5 c4b3 g8f6 e1g1 f8e7 f1e1 e8g8 h2h3 b8c6 c2c3 c8f5 c1g5 b7b5 d3d4 c5c4 b3c2 f5c2 d1c2 f6e4 g5e7 d8e7 b1d2 f7f5 b2b3 e7a3 b3c4 b5c4 d2e4 f5e4 f3d2 a8b8 a1b1 b8b5 d2f1 h7h6 f1g3 g8h8 c2d2');
INSERT INTO evaluations VALUES(50775,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(50775,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(50775,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(50775,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(50775,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(50775,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(50775,36,'cp -16','e4d5 e6d5 c4b3 g8f6 e1g1 f8e7 f1e1 e8g8 c2c3 c8f5 c1g5 b8d7 d3d4 c5c4 b3c2 f5c2 d1c2 e7d6 b1d2 d8b6 g5h4 a8e8 h4g3 d6g3 h2g3 h7h6 a2a4 g7g6 d2f1 g8g7 f1e3 h6h5 a1d1 b6c6 b2b3 b7b5 a4b5');
INSERT INTO evaluations VALUES(15501,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(15501,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(15501,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(15501,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(15501,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(15501,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(15501,36,'cp 62','f3g5 d6d5 e4d5 c6a5 d2d3 a5c4 d3c4 b7b5 e1g1 b5c4 f1e1 f8e7 d1f3 d8d5 b1c3 d5f3 g5f3 f6d7 f3e5 d7e5 e1e5 c8e6 c3b5 e7d6 b5d6 c7d6 e5a5 e8d7 c1d2 h8c8 d2b4 a7a6 a1d1 c8c6 f2f3 g7g6 b4c3 h7h5 a2a4 a8e8 d1d2 e8b8 d2d4 b8e8 h3h4 e8b8 g2g4');
INSERT INTO evaluations VALUES(15475,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(15475,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(15475,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(15475,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(15475,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(15475,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(15475,36,'cp 40','d6d5 e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5d3 f6d5 g5f3 f8d6 e1g1 d5b4 f1e1 b4d3 c2d3 e8g8 f3e5 c6c5 d3d4 c5d4 d2d3 c8a6 b1a3 f8e8 e5c4 e8e1 d1e1 d6a3 c4a5 a3f8 a5c4 d8f6 c1d2 f6g6 e1f1 a8e8 b2b3 h7h6 a1e1 e8e1 d2e1 g6f5 f1e2 f5d5 e2f3 d5f3 g2f3');
INSERT INTO evaluations VALUES(30143,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30143,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(30143,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(30143,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(30143,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(30143,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(30143,36,'cp 16','c7c6 f3h4 f5g6 e2e3 d8b6 d1b3 f8e7 c4c5 b6c7 g5f4 c7c8 f1e2 b8d7 e1g1 e8g8 b3d1 b7b6 b2b4 a7a5 h4g6 h7g6 a2a3 a5b4 a3b4 c8b7 d1c2 f6h5 e2h5 g6h5 c3a4 b6b5 a4c3 a8a1 f1a1 f8a8 c2b2 g7g5 f4d6 e7d6 a1a8 b7a8');
INSERT INTO evaluations VALUES(45606,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45606,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(45606,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45606,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45606,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45606,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45606,36,'cp 0','e6d5 d1c2 g7g6 c1g5 f8e7 e2e3 c8f5 f1d3 f5d3 c2d3 b8d7 g5h6 e7f8 h6f4 f8e7');
INSERT INTO evaluations VALUES(48343,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48343,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48343,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48343,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(48343,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48343,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(48343,36,'cp -17','c7c5 c2c3 b7b6 g1f3 c8b7 e1g1 f8e7 d1e2 e8g8 b2b3 b8d7 c1b2 a8c8 a1c1 c8c7 c3c4 d5c4 d2c4 d8a8 d4c5 c7c5 b2d4 c5c6 e2b2 f8c8 d3e2 e7f8 c4e5 d7e5 f3e5 c6c7 f2f3 f6d5 c1c7 c8c7');
INSERT INTO evaluations VALUES(46243,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(46243,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(46243,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(46243,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46243,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(46243,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46243,36,'cp 0','b2b3 c5d4 e3d4 b7b6 c1b2 c8a6 c2c4 f8d6 g1f3 e8g8 e1g1 b8c6 f1e1 a8c8 f3e5 a6b7 d1e2 c6b4 d3b1 b4c6');
INSERT INTO evaluations VALUES(56928,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(56928,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(56928,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(56928,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(56928,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(56928,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(56928,36,'cp -85','d7d6 g2g3 f8g7 b1c3 b8c6 d4d5 c6e5 g3f4 g5f4 c1f4 c8g4 f1e2 e5f3 e2f3 d8f6 f3g4 f6f4 d1f3 g7c3 b2c3 f4f3 g4f3 g8f6 e1d2 h8g8 a1e1 f6d7 h1g1 e8e7 e1f1 a7a5 f3e2 d7f6 e2d3 g8g1 f1g1 a8g8 g1g8 f6g8 d2e3 e7f6 e3f4 g8e7 d3b5 e7g6 f4g4');
INSERT INTO evaluations VALUES(19444,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(19444,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(19444,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(19444,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(19444,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(19444,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(19444,36,'cp -69','g8f6 f1b5 d8b6 b1a3 g7g6 g1e2 f8g7 e1g1 e8g8 h2h3 d7d6 d2d3 a7a6 b5c6 b6c6 a3c4 c8e6 c4e3 d6d5 e4d5 e6d5 e3d5 c6d5 f3d5 f6d5 f1d1 f8d8 d3d4 c5d4 e2d4 a8c8 a2a4 e7e5 d4b3 f7f5 c1g5 d8d7 b3a5');
INSERT INTO evaluations VALUES(12831,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(12831,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(12831,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(12831,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(12831,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(12831,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(12831,36,'cp -49','f1b5 d8b6 b1a3 g7g6 g1e2 f8g7 e1g1 e8g8 h2h3 f8d8 c3c4 a7a6 b5c6 b6c6 e2c3 a8b8 d2d3 b7b5 c3d5 d8e8 c1d2 f6d5 c4d5 c6c7 a1b1 e7e6 b2b4 c5b4 f1c1 c7d8 d2b4 e6d5 e4d5 d7d6 b4c3 d8g5 c3g7 g8g7 a3c2 g5f5 c2e3 f5f3');
INSERT INTO evaluations VALUES(40330,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40330,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40330,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(40330,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40330,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40330,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40330,36,'cp 0','e1g1 e8g8 h2h3 a7a6 a2a4 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2g4 g1h1 g4f2');
INSERT INTO evaluations VALUES(40329,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40329,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40329,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40329,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40329,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40329,36,'cp 0','e8g8 a2a4 a7a6 h2h3 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2g4 g1h1 g4f2');
INSERT INTO evaluations VALUES(35931,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35931,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35931,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35931,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35931,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35931,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35931,36,'cp 0','a2a4 a7a6 h2h3 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2g4 g1h1 g4f2');
INSERT INTO evaluations VALUES(29611,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(29611,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(29611,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(29611,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(29611,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(29611,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(29611,36,'cp -8','b8d7 e2e3 e7e6 f1d3 f5g6 f3e5 f8e7 f2f4 d7e5 f4e5 f6d7 g5e7 d8e7 d1e2 g6d3 e2d3 c7c6 e3e4 e8g8 e1g1 f7f6 e5f6 f8f6 f1f6 d7f6 e4e5 f6d7 c3e2 a8f8 a1f1 f8f1 g1f1 e7f8 d3f3 f8f3 g2f3 d7b6 f1f2 g8f7 h2h4 b6c4 b2b3 c4a3 c2c3 a3b1 f2e3 f7g6');
INSERT INTO evaluations VALUES(32973,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(32973,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(32973,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(32973,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(32973,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(32973,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(32973,36,'cp 118','c1e3 c5b6 d3d4 d7d6 h2h3 b6a5 d1d3 g8e7 e1c1 e8g8 d4e5 d6e5 f3h2 b7b5 c4b3 b8a6 a2a4 a5b4 a4b5 a6c5 e3c5 b4c5 h2g4 c8g4 h3g4 c6b5 c1b1 b5b4 c3a4 c5d4 h1h5 f6f2 g4g5 f2e3 d3e3 d4e3 g5h6 e3h6 h5e5 e7g6 e5f5');
INSERT INTO evaluations VALUES(32974,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(32974,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(32974,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(32974,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(32974,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(32974,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(32974,36,'cp 101','d7d6 d3d4 c5b6 d1d2 c8g4 d4e5 b6e3 d2e3 d6e5 f3d2 g8e7 f2f3 g4e6 e1c1 e7c8 c4e6 f6e6 f3f4 e5f4 e3f4 b8d7 d2f3 e8g8 h1e1 c8b6 e4e5 d7c5 f3d4 e6g6 d4f5 a8e8 c3e4 c5e4 e1e4 b6d5 f4f3 d5e7 g2g4 e7f5 g4f5 g6g5 c1b1 g7g6');
INSERT INTO evaluations VALUES(34624,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(34624,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(34624,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(34624,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(34624,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(34624,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(34624,36,'cp 152','d2d4 g7g6 g1f3 f8g7 f1c4 c8e6 c4d3 f5a5 e1g1 g8f6 c1d2 e8g8 h2h3 e6d5 c3d5 a5d5 f1e1 e7e6 b2b4 f8d8 a2a4 b8c6 c2c3 a7a5 b4b5 c6e7 c3c4 d5d6 c4c5 d6d7 d1b3 f6d5 d2g5 h7h6 g5h4 d5b4 d3e4 g6g5 h4g3 c7c6 b5b6 e7d5 g3e5 g7e5 f3e5');
INSERT INTO evaluations VALUES(34620,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(34620,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(34620,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(34620,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(34620,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(34620,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(34620,36,'cp 149','g7g6 g1f3 f8g7 f1c4 c8e6 c4d3 f5a5 f3g5 b8c6 g5e6 f7e6 e1g1 e8c8 c3e4 g7d4 e4g5 g8f6 g5f7 c6e5 f7h8 e5d3 c2d3 d8h8 d1e2 a5b6 c1f4 f6d7 a1b1 h8d8 f4e3 d7f6 e3d4 d8d4 b1c1 d4d5 f1e1 e6e5 h2h3 c7c6 a2a3 b6d4 c1c3 a7a6 c3c4 d4d3');
INSERT INTO evaluations VALUES(34959,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(34959,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(34959,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(34959,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(34959,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(34959,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(34959,36,'cp -2','g8f6 f1b5 c8d7 b1c3 d7b5 c3b5 b8a6 f3e5 f6e4 d1e2 c7c6 b5d4 f8c5 h1d1 e8c8 e5f7 d8e8 f7h8 e4c3 e2d2 c3d1 d2d1 c5d4 h8f7 e8f8 f7d6 c8c7 d6e4 d4f2 g2g3 a6c5 e4c5 f2c5 c1f4 c5d6 f4d6 c7d6 d1e2 d6c5 a1d1 f8e8 e2f3 e8e7 d1d8 b7b5 d8a8 e7f7 f3e3 b5b4 a8b8 f7f1 b8a8 f1h1 a8a7');
INSERT INTO evaluations VALUES(42807,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42807,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(42807,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42807,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42807,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42807,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42807,36,'cp 10','d7d5 e4d5 d8d5 f1e2 b8c6 e1g1 e8g8 c3c4 d5d6 a2a4 f8e8 c1a3 d6d8 f1e1 h7h6 f3d2 c6d4 a3b2 b7b6 e2f3 a8b8 a4a5 c8f5 d2b3 d8d6 a5b6 a7b6 b3d4 e5d4 e1e8 b8e8 h2h3 f5d7 a1a8 c7c5 a8e8 d7e8 c2c3 d4c3 b2c3 e8c6 f3e2 f6d7');
INSERT INTO evaluations VALUES(47254,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(47254,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(47254,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(47254,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(47254,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(47254,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(47254,36,'cp -54','e7e5 d2d3 f8d6 g1f3 e8g8 e1g1 f8e8 c1b2 c7c5 b1c3 b8c6 e2e4 d5d4 c3d5 c8g4 h2h3 g4e6 d5f6 d8f6 b2c1 h7h6 f3h2 f6e7 f2f4 e5f4 g3f4 f7f5 e4e5 d6c7 c1d2 g8h7 h2f3 a8d8');
INSERT INTO evaluations VALUES(53867,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(53867,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(53867,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(53867,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(53867,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(53867,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(53867,36,'cp 53','c8d7 b5d7 d8d7 d4e2 b8c6 e1g1 g8f6 b1c3 e8c8 e2g3 f8e7 f1e1 h8e8 c1f4 g7g6 c3d5 f6d5 e4d5 c6b4 c2c4 f7f5 d1d2 e7f6 h2h3 b4a6 a1b1 e8e1 b1e1 a6c7 e1b1 b7b6 f4e3 c8b7 b2b4 c5b4 d2b4 c7a6 b4a5 a6c5 e3c5 d6c5 a5c5 d8c8');
INSERT INTO evaluations VALUES(30874,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(30874,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(30874,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(30874,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(30874,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(30874,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(30874,36,'cp 58','b5d7 d8d7 d4e2 b8c6 e1g1 g8f6 b1c3 e8c8 e2g3 f8e7 f1e1 h8e8 c1f4 g7g6 c3d5 f6d5 e4d5 c6b4 d1d2 f7f5 c2c4 e7f6 h2h3 b4a6 a1b1 e8e1 b1e1 a6c7 b2b3 d8e8 g3e2 b7b5 f4e3 c8b7 e2f4 e8e4 f2f3 e4e7 e1b1 b7a8 f4d3 e7e8 b1d1 b5c4');
INSERT INTO evaluations VALUES(30851,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(30851,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(30851,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(30851,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(30851,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(30851,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(30851,36,'cp 75','d8d7 d4e2 b8c6 e1g1 g8f6 b1c3 e8c8 c1g5 f8e7 d1d3 c6b4 d3d2 h7h6 g5f6 e7f6 a2a3 b4c6 c3d5 d7e6 f2f3 h8e8 a1e1 e6e5 b2b3 c6e7 d5f6 e5f6 e2c3 c8b8 f1f2 g7g5 b3b4 f6g6 f2e2 f7f5 e4f5 e7f5 c3d5 a7a6 c2c3 e8e2 d2e2 g6f7');
INSERT INTO evaluations VALUES(32747,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(32747,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(32747,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(32747,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(32747,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(32747,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(32747,36,'cp 60','d4e2 b8c6 e1g1 g8f6 b1c3 e8c8 c1g5 f8e7 d1d3 c6b4 d3d2 h7h6 g5f6 e7f6 a2a3 b4c6 c3d5 d7e6 f2f3 h8e8 a1e1 e6e5 d5f6 e5f6 c2c3 c6e7 e2g3 c8b8 f3f4 d6d5 e4d5 d8d5 d2f2 f6b6 f4f5 c5c4 f2b6 a7b6 f5f6');
INSERT INTO evaluations VALUES(54698,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(54698,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(54698,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(54698,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(54698,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(54698,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(54698,36,'cp -18','g8f6 g1f3 e7e6 d2d4 c7c5 f1b5 c8d7 d1e2 d7b5 e2b5 d8d7 b5d7 f6d7 f3g5 e8e7 d4c5 h7h6 g5f3 b8c6 b1c3 e7f7 c1e3 g7g5 e1c1 g5g4 f3e1 f8c5 e3c5 d7c5 e1d3 c5d3 d1d3 h6h5 h1e1 a8d8 h2h3 f7f6 a2a3 h8g8 h3g4 g8g4 g2g3 c6e5 d3e3');
INSERT INTO evaluations VALUES(46606,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(46606,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46606,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(46606,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(46606,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(46606,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(46606,36,'cp -20','g1f3 e7e6 d2d4 c7c5 f1b5 c8d7 d1e2 d7b5 e2b5 d8d7 b5d7 b8d7 e1g1 h7h6 c1e3 a8c8 c2c3 e8f7 f1e1 g7g5 h2h3 c5c4 f3e5 d7e5 d4e5 f6d7 e3a7 f8c5 a7c5 d7c5 b1a3 c8c6 a3b5 c6b6 b5d6 f7g6 b2b3 c4b3 a2b3 b6b3 e1e3 h8f8 g1f1 b3b2 e3e2 b2e2 f1e2');
INSERT INTO evaluations VALUES(5580,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(5580,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5580,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(5580,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(5580,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(5580,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(5580,36,'cp 53','b4d2 d1d2 d8b6 b1c3 g8e7 f1d3 c8d7 d3c2 c6a5 b2b3 a8c8 a1c1 d7b5 c3b5 b6b5 c1b1 h7h6 c2d3 b5b6 e1g1 a5c6 a2a3 e8g8 b3b4 e7f5 d3f5 e6f5 f1c1 c6e7 c1c5 b6e6 h2h4 b7b6 c5c8 f8c8 b4b5 c8c4 h4h5 e6c8 b1e1 c4c3 a3a4 c3c4 d2d1 c8e6 d1a1 e6d7 a1a2 c4c8 e1b1 c8c4 g2g3');
INSERT INTO evaluations VALUES(5584,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(5584,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(5584,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(5584,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(5584,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5584,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(5584,36,'cp 36','d1d2 g8e7 f1d3 e8g8 e1g1 f7f6 e5f6 f8f6 b1c3 h7h6 c3e2 c8d7 a1c1 d7e8 f3e5 f6f8 e5c6 e7c6 d3b1 f8f6 f1e1 d8b6 c1c3 e8h5 e2f4 h5f7 e1c1 a8e8 c3c5 b6c7 f4d3 f7g6 d3e5 g6b1 c1b1 c7d6 b1e1 e8f8 f2f3 c6e5 e1e5 d6b6 c5c3');
INSERT INTO evaluations VALUES(5583,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(5583,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(5583,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(5583,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(5583,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(5583,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(5583,36,'cp 45','g8e7 f1d3 e8g8 b1c3 f7f6 e5f6 f8f6 e1g1 h7h6 c3e2 c8d7 a1c1 d7e8 f3e5 f6f8 e5c6 e7c6 d3b1 f8f6 f1e1 d8b6 c1c5 e8f7 a2a3 a8f8 b2b4 f7h5 f2f3 h5g6 b1g6 f6g6 e2f4 g6f6 g2g3 c6e7 d2e3 f8f7 e3e2 e7c6 e2d2 c6e7 d2e3 b6d6 e1c1 d6b6 e3e5 f7f8');
INSERT INTO evaluations VALUES(8227,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(8227,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(8227,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(8227,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(8227,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(8227,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(8227,36,'cp 60','f1d3 e8g8 e1g1 f7f6 e5f6 f8f6 d2e3 c8d7 b1d2 d7e8 a2a3 e8g6 a1e1 g6d3 e3d3 e7g6 e1e3 d8d7 g2g3 h7h6 b2b4 a7a6 d2b3 a8f8 d3e2 d7f7 g1g2 g6e7 b3c5 e7f5 e3e6 c6d4 e6f6 f7f6 f3d4 f5d4 e2d1 f8f7 c5d3 f6g5 f1e1 d4f3 e1e3');
INSERT INTO evaluations VALUES(21517,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(21517,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(21517,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(21517,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(21517,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(21517,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(21517,36,'cp 141','d8d5 b1c3 d5d6 e1g1 g8e7 d2d4 e5d4 f3d4 c8d7 b5c6 d7c6 d4c6 d6c6 f1e1 a8d8 d1h5 g7g6 h5e2 d8d7 c1g5 f7f6 g5f4 e8f7 c3b5 f7g7 b5a7 c6a4 e2b5 a4b5 a7b5 c7c6 b5c7 g7f7 a1d1 d7d1 e1d1 g6g5 f4d6 e7g6 d6f8 h8f8 d1d7 f7g8 c7e6 f8e8');
INSERT INTO evaluations VALUES(1694,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(1694,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(1694,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(1694,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(1694,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(1694,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(1694,36,'cp 169','b1c3 d5d6 e1g1 g8e7 d2d4 e5d4 f3d4 c8d7 b5c6 d7c6 d4c6 e7c6 f1e1 f8e7 d1d6 c7d6 c1f4 e8c8 a1d1 d8d7 c3d5 e7d8 c2c4 c6e7 g2g3 e7d5 d1d5 d8f6 b2b3 h8d8 e1d1 c8c7 d1d3 b7b6 a2a4 f6e5 f4e3 d7e7 a4a5 e7e8 h2h3 e5f6 b3b4 b6a5 d5a5');
INSERT INTO evaluations VALUES(1693,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(1693,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(1693,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(1693,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(1693,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(1693,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(1693,36,'cp 164','d5d6 e1g1 g8e7 d2d4 e5d4 f3d4 c8d7 b5c6 d7c6 d4c6 e7c6 f1e1 f8e7 d1d6 c7d6 c1f4 e8c8 a1d1 d8d7 c3d5 e7d8 c2c4 h7h6 b2b4 d8g5 b4b5 c6e7 f4g5 e7d5 d1d5 h6g5 d5g5 f7f6 g5g4 c8c7 h2h4 c7b6 g2g3 h8h5 e1e8 d6d5 c4d5 h5d5 e8g8 g7g5');
INSERT INTO evaluations VALUES(48169,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(48169,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(48169,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(48169,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(48169,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(48169,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(48169,36,'cp -52','a2a3 b7b6 b2b4 a7a5 c1b2 a5b4 a3b4 a8a1 b2a1 e6e5 e2e3 e5e4 g1e2 c7c6 e2c3 f8e7 f1e2 b6b5 c3b5 c6b5 e2b5 e8f8 b5e2 h7h5 b1c3 h8h6 b4b5 h6g6 g2g3 h5h4 d1a4 f8g8 a4a8 c8e6 a1b2 g8h7 b5b6 f6g4 b2a3 g4h6 a8a4');
INSERT INTO evaluations VALUES(48170,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(48170,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(48170,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(48170,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(48170,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(48170,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(48170,36,'cp -36','b7b6 b2b4 a7a5 c1b2 a5b4 a3b4 a8a1 b2a1 e6e5 e2e3 e5e4 g1e2 f8e7 e2c3 c7c6 f1e2 h7h5 b4b5 b6c5 d4c5 e7c5 d1c2 e8g8 e1g1 c5b6 b1d2 c6b5 c3b5 f6g4 a1d4 b6d4 b5d4 d8g5 c2b3 g5e5 e2g4 c8g4 h2h3 g4e6 f1a1 f8c8 b3b2 b8c6');
INSERT INTO evaluations VALUES(43791,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(43791,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(43791,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(43791,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(43791,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(43791,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(43791,36,'cp -37','b2b4 a7a5 c1b2 a5b4 a3b4 a8a1 b2a1 e6e5 e2e3 e5e4 g1e2 f8e7 e2c3 c7c6 f1e2 h7h5 e1g1 h5h4 h2h3 b6b5 b1d2 h8h6 f2f3 h6g6 f3e4 c8h3 f1f2 f6g4 e2g4 h3g4 d1f1 h4h3 e4d5 c6d5 g2g3 e8f8 f2f7 f8g8 c3d5 d8d5');
INSERT INTO evaluations VALUES(43787,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(43787,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(43787,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(43787,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(43787,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(43787,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(43787,36,'cp -55','a7a5 c1b2 a5b4 a3b4 a8a1 b2a1 e6e5 e2e3 e5e4 g1e2 f8e7 e2c3 c7c6 f1e2 h7h5 b4b5 b6c5 d4c5 e7c5 d1c1 d5d4 e3d4 c5d4 e1g1 e8g8 b1a3 d8e7 a3c2 d4e5 c2e3 c8e6 f1d1 e7c7 h2h3 c6b5 c3b5 c7c1 d1c1 b8d7 a1e5 d7e5 b5d4 f8b8 d4e6');
INSERT INTO evaluations VALUES(2317,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(2317,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(2317,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(2317,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(2317,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(2317,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(2317,36,'cp 36','d8e8 b1c3 f8d6 c1e3 g8f6 a2a3 a7a6 h2h3 h7h6 c4a2 c8e6 a2e6 f7e6 f3d2 f6d7 d2b3 d7b6 c3d1 b6c4 c2c3 e8f7 b3d2 c6a5 b2b4 c4d2 e3d2 a5b3 a1b1 b3d2 e1d2 a6a5 d1b2 h8f8 b2d3 a5b4 a3b4 f7e7 d2e2 a8a2 b1b2 a2b2 d3b2');
INSERT INTO evaluations VALUES(29002,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(29002,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(29002,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(29002,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(29002,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(29002,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(29002,36,'cp 37','d1b3 b8d7 c4d5 c6d5 g1f3 e7e6 e2e4 f5e4 f3e5 f8e7 f1b5 a7a6 g5f6 a6b5 f6e7 d8e7 b3b5 f7f6 b5b7 a8b8 b7d7 e7d7 e5d7 e8d7 c3e4 d5e4 b2b3 h8c8 e1d2 b8b4 h1c1 b4d4 d2e3 d4d3 e3e2 c8a8 c1d1 d3d1 e2d1 d7c6 d1c2 c6b7 a2a4 a8c8 c2b2 c8d8 a1c1 d8d2 c1c2 d2d3');
INSERT INTO evaluations VALUES(16317,30,'cp -524',NULL);
INSERT INTO evaluations VALUES(16317,31,'cp -515',NULL);
INSERT INTO evaluations VALUES(16317,32,'cp -555',NULL);
INSERT INTO evaluations VALUES(16317,33,'cp -564',NULL);
INSERT INTO evaluations VALUES(16317,34,'cp -562',NULL);
INSERT INTO evaluations VALUES(16317,35,'cp -565',NULL);
INSERT INTO evaluations VALUES(16317,36,'cp -555','c6d4 f3d3 c7c6 b5a4 d7d5 f2f3 f6d7 c2c3 d7c5 d3f1 c5a4 c3d4 e5d4 d2d3 f8d6 g1e2 a4c5 e2f4 e8g8 b2b4 c5e6 f4e6 c8e6 b1d2 a7a5 f1e2 d8h4 e2f2 h4f2 e1f2 a5b4 c1b2 c6c5 a2a3 d6f4 e4d5 e6d5 d2e4 b4a3 b2a3 b7b6 h1e1 f8e8 a3c5 b6c5 e4c5 e8c8 a1a8 f4e3 e1e3 d4e3 f2e3 c8a8');
INSERT INTO evaluations VALUES(16900,30,'cp -559',NULL);
INSERT INTO evaluations VALUES(16900,31,'cp -549',NULL);
INSERT INTO evaluations VALUES(16900,32,'cp -549',NULL);
INSERT INTO evaluations VALUES(16900,33,'cp -579',NULL);
INSERT INTO evaluations VALUES(16900,34,'cp -569',NULL);
INSERT INTO evaluations VALUES(16900,35,'cp -575',NULL);
INSERT INTO evaluations VALUES(16900,36,'cp -574','f3d3 c7c6 b5a4 d7d5 f2f3 f6d7 c2c3 d7c5 d3f1 c5a4 c3d4 e5d4 d2d3 f8d6 g1e2 a4c5 e2f4 e8g8 b2b4 c5e6 f4e6 c8e6 b1d2 a7a5 f1e2 d8h4 e2f2 h4f2 e1f2 a5b4 c1b2 c6c5 a2a3 d6f4 h1d1 b4a3 b2a3 b7b6 g2g3 f4e3 f2g2 h7h5 e4d5 e6d5 d2c4 d5c4 d3c4 a8a4 a3c5 a4a1 d1a1 b6c5 a1d1 f8e8');
INSERT INTO evaluations VALUES(16898,30,'cp -581',NULL);
INSERT INTO evaluations VALUES(16898,31,'cp -559',NULL);
INSERT INTO evaluations VALUES(16898,32,'cp -571',NULL);
INSERT INTO evaluations VALUES(16898,33,'cp -580',NULL);
INSERT INTO evaluations VALUES(16898,34,'cp -579',NULL);
INSERT INTO evaluations VALUES(16898,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(16898,36,'cp -571','c7c6 b5a4 d7d5 f2f3 f6d7 c2c3 d7c5 d3f1 c5a4 c3d4 e5d4 d2d3 f8d6 e4e5 d6e5 f3f4 e5f6 f1e2 d8e7 e2e7 f6e7 g1f3 a4c5 e1d2 c8f5 f3d4 f5d3 b1c3 e7f6 h1e1 e8d7 d4f3 d3f5 f3e5 f6e5 e1e5 f5g6 b2b3 c5d3 c1a3 d3e5 f4e5 h8e8 a1e1 f7f6 e5e6 e8e6 e1e6 d7e6');
INSERT INTO evaluations VALUES(12619,30,'cp -580',NULL);
INSERT INTO evaluations VALUES(12619,31,'cp -566',NULL);
INSERT INTO evaluations VALUES(12619,32,'cp -564',NULL);
INSERT INTO evaluations VALUES(12619,33,'cp -563',NULL);
INSERT INTO evaluations VALUES(12619,34,'cp -582',NULL);
INSERT INTO evaluations VALUES(12619,35,'cp -567',NULL);
INSERT INTO evaluations VALUES(12619,36,'cp -557','b5a4 d7d5 f2f3 f6d7 c2c3 d7c5 d3f1 c5a4 c3d4 e5d4 d2d3 f8d6 e4e5 d6e5 f3f4 e5d6 f1e2 d8e7 e2e7 d6e7 e1d2 c8f5 g1f3 a4c5 f3d4 f5d3 b1c3 e7f6 h1e1 e8d7 d4f3 d3g6 g2g4 c5d3 f4f5 d3e1 d2e1 h7h5 h2h3 a8e8 e1f2 g6h7 c1f4 h5g4 h3g4');
INSERT INTO evaluations VALUES(20993,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(20993,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(20993,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20993,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(20993,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20993,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(20993,36,'cp 22','g1f3 c8f5 a2a3 g8f6 e2e3 e7e6 c2c4 c6b8 b1d2 b8d7 c4c5 h7h6 b3b4 g7g5 h2h3 f8g7 a3a4 f6e4 d2e4 f5e4 a1a3 c7c6 b4b5 e6e5 f1e2 e8g8 e1g1 e4g6 f3e5 d7e5 d4e5 g7e5 b5c6 b7c6 a3b3 e5g7 c1a3 d8a5');
INSERT INTO evaluations VALUES(20992,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(20992,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20992,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(20992,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(20992,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(20992,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20992,36,'cp 23','c8f5 a2a3 e7e6 c2c4 g8f6 e2e3 c6b8 c4c5 f8e7 b3b4 e8g8 c1b2 c7c6 f1d3 b8d7 e1g1 f8e8 d3f5 e6f5 b1c3 a7a6 c3e2 f6e4 f3e1 g7g6 e1d3 e4g5 a3a4 h7h5 d1b3 h5h4 e2f4 e7f6 b4b5 h4h3 g2h3 g5f3 g1h1');
INSERT INTO evaluations VALUES(26552,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26552,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(26552,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(26552,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(26552,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(26552,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(26552,36,'cp 21','a2a3 g8f6 e2e3 e7e6 c2c4 c6b8 c4c5 f8e7 c1b2 e8g8 b3b4 c7c6 f1d3 b8d7 e1g1 d8c7 d3c2 b7b6 c2f5 e6f5 f3e1 c7b7 e1d3 a7a5 b1c3 a5b4 a3b4 a8a1 b2a1 f8a8 a1b2 g7g6 d1c2 f6e4 f1c1 b6c5');
INSERT INTO evaluations VALUES(28821,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(28821,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(28821,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(28821,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(28821,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(28821,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(28821,36,'cp 124','d1e2 e7e6 g1f3 d7d5 e4e5 f6d7 c2c4 b8c6 c4d5 c6b4 b1c3 b4d5 e1g1 f8e7 f1d1 a8c8 d3c4 d5c3 b2c3 e8g8 c4b3 c7c5 d4d5 b7d5 b3d5 e6d5 d1d5 d8c7 c1f4 c7c6 a1d1 f8d8 f4g5 f7f6 e2c4 c6e6 e5f6 g7f6 g5h4 d7f8 h4g3 d8d5 d1d5 e6f7 h2h4');
INSERT INTO evaluations VALUES(28820,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(28820,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(28820,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(28820,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(28820,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(28820,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(28820,36,'cp 117','b8c6 c2c3 d7d6 g1f3 e7e5 e1g1 f8e7 d4d5 c6b8 c3c4 b8d7 a2a3 b7c8 b1c3 e8g8 d3c2 a7a5 h2h3 f6h5 f3e5 h5g3 f2g3 d7e5 g1h2 e7f6 c1d2 c8d7 a3a4 f6g5 d2f4 g5f6 a1e1 d8e7 f4d2 a8b8 b2b3 f8e8 e2f2 f6g5 c3b5 d7b5 c4b5');
INSERT INTO evaluations VALUES(29991,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(29991,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(29991,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(29991,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(29991,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(29991,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(29991,36,'cp 129','f7e6 g5e6 d8c8 e6f8 f6e4 f8h7 h8h7 d2d3 e4f6 c1g5 c8e6 b1c3 b8c6 d1e2 e8c8 e1c1 d6d5 c1b1 d5d4 c3e4 c6b4 c2c4 d4c3 e4c3 h7h8 h2h4 d8d7 f2f4 f6h7 a2a3 b4c6 g2g4 c6d4 e2g2 h7g5 f4g5 h8f8 d1f1 f8f4 h4h5 e6a6');
INSERT INTO evaluations VALUES(30272,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(30272,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(30272,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(30272,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(30272,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(30272,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(30272,36,'cp 140','g5e6 d8d7 e6f8 h8f8 b1c3 b8c6 d2d3 b7b6 c1e3 e8c8 a2a3 c8b7 e1g1 h7h6 f2f3 g7g5 b2b4 d6d5 c3d5 f6d5 e4d5 d7d5 f1e1 f8g8 e3f2 h6h5 e1e4 g5g4 a1c1 b7b8 a3a4 d5d7 a4a5 d8f8 f3g4 g8g4 a5a6 c6d4 f2d4 g4e4 d3e4');
INSERT INTO evaluations VALUES(30269,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(30269,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(30269,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(30269,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(30269,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(30269,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(30269,36,'cp 139','d8d7 e6f8 h8f8 b1c3 b8c6 d2d3 b7b6 c1e3 e8c8 e1g1 c8b7 a2a3 h7h6 f2f3 d6d5 c3d5 f6d5 e4d5 d7d5 f1e1 a7a5 e3f2 a5a4 d1e2 g7g5 h2h3 h6h5 e2e4 d5b5 f2e3 b5b2 e4c4 c6d4 a1a2 b2b5 c4b5 d4b5 e3g5 d8d5 e1e4 f8f5');
INSERT INTO evaluations VALUES(32711,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(32711,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(32711,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(32711,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(32711,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(32711,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(32711,36,'cp 138','e6f8 h8f8 d2d3 b8c6 f2f3 e8c8 c2c4 f6h5 b1c3 h5f4 c1f4 f8f4 c3e2 f4f7 d3d4 e5d4 e2d4 c6d4 d1d4 c8b8 h2h4 b7b6 e1c1 g7g6 d4c3 d8f8 c1c2 d7e7 h4h5 g6g5 d1d5');
INSERT INTO evaluations VALUES(32553,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(32553,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(32553,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(32553,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(32553,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(32553,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(32553,36,'cp 152','h8f8 d2d3 b8c6 c2c4 f8f7 e1g1 e8c8 b1a3 c6d4 a3c2 d4c2 d1c2 f6h5 b2b4 g7g5 a2a4 h5f4 c1f4 g5f4 f2f3 f7g7 g1h1 d8g8 f1f2 h7h5 c2d1 g7g6 b4b5 d7g7 d1d2 h5h4 h2h3 g6g3 a4a5');
INSERT INTO evaluations VALUES(53080,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53080,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53080,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(53080,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(53080,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53080,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(53080,36,'cp 24','g8f6 g1e2 e7e5 b1c3 b8c6 d2d3 f8e7 e1g1 e8g8 f2f4 c8d7 a2a4 c6d4 h2h3 f8e8 g1h2 h7h6 c3b1 e7f8 e2d4 c5d4 c2c3 e5f4 g3f4 d4c3 b1c3 g7g6 c1e3 f8g7 d1d2 f6h5 d3d4 d8h4 d2f2 h4f2 f1f2 a8c8');
INSERT INTO evaluations VALUES(45215,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(45215,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(45215,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(45215,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(45215,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(45215,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45215,36,'cp 36','c2c3 e7e6 d2d4 b8c6 d4c5 d6c5 d1d8 c6d8 b1a3 f8e7 c1f4 e8g8 a3c4 d8c6 f4d6 e7d6 c4d6 f6e8 d6c8 a8c8 f2f4 e6e5 g1h3 e8f6 e1c1 c8d8 g2f3 d8d1 h1d1 f8d8 d1d8 c6d8 h3f2 f6d7 f3g4 d7f6 c1c2 f6g4 f2g4 e5f4 g3f4 g8f8');
INSERT INTO evaluations VALUES(54862,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(54862,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(54862,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(54862,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(54862,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(54862,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(54862,36,'cp 50','g8f6 c2c4 e7e6 f1e2 b7b6 b1c3 c8b7 e1g1 b8d7 b2b3 f8d6 c1b2 e8g8 a1c1 f8e8 c4d5 f6d5 d1c2 d8e7 f1d1 d5c3 c2c3 a8c8 f3d2 d6b4 c3c2 b4a3 d2c4 a3b2 c2b2 e8f8 f2f3 c7c5 d4c5 c8c5 c4d6 c5c1 b2c1');
INSERT INTO evaluations VALUES(15514,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(15514,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(15514,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15514,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(15514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15514,36,'cp -8','f6e4 c3e4 d6d5 c4b5 d5e4 f3e5 c8d7 d1e2 c6e5 f4e5 c7c6 b5c4 d7f5 e1g1 f8c5 g1h1 d8d7 d2d3 e4d3 e2f3 f5e6 c4d3 h7h6 c1e3 d7e7 a1e1 e8c8 d3f5 c8b8 e3c5 e7c5 f5e6 f7e6 f3g4 c5e7 e1e2 h8f8 f1f8 e7f8 h2h3 b8c7 h1h2');
INSERT INTO evaluations VALUES(21485,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(21485,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(21485,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(21485,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(21485,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(21485,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(21485,36,'cp 99','a7a6 g1f3 b7b5 c4d3 c6b4 d3e2 g8f6 e3e4 c7c5 e1g1 c8b7 a2a4 c5d4 f3d4 f8d6 a4b5 e8g8 b5a6 b4a6 d4b5 d6e7 d1d8 f8d8 f2f3 a6b4 a1a8 e7c5 g1h1 d8a8 f1d1 a8a1 c1e3 a1d1 c3d1 f6d7 h1g1 b7a6 g1f1');
INSERT INTO evaluations VALUES(54282,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(54282,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(54282,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(54282,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(54282,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(54282,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(54282,36,'cp -23','b8c6 g1f3 d8b6 f1d3 g8h6 e1g1 c8d7 g1h1 h6g4 a2a3 g7g6 h2h3 h7h5 d4c5 f8c5 d1e2 a7a5 b1c3 c6e7 c1d2 g4f2 h1h2 a5a4 b2b4 a4b3 c2b3 f2d3 e2d3 b6a6 d3a6 b7a6 f1c1 c5b6 a3a4 e8g8 a4a5 b6a7 c3e2 e7c6 b3b4 a8b8 a1b1 f8c8 h2h1');
INSERT INTO evaluations VALUES(20322,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(20322,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(20322,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(20322,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(20322,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(20322,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(20322,36,'cp -19','g1f3 c8d7 c2c3 g8h6 b1a3 c5d4 c3d4 h6f5 a3c2 d8b6 g2g3 a8c8 f1h3 f8e7 h3f5 e6f5 e1g1 e8g8 f1f2 c8c7 c2e1 f8c8 c1e3 b6b5 f2c2 c6a5 c2c7 c8c7 b2b3 b5a6 e3d2 d7b5 d1b1 a5c6 d2e3 b5e2 b1f5 c6b4 a2a4');
INSERT INTO evaluations VALUES(20321,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(20321,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(20321,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(20321,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(20321,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(20321,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(20321,36,'cp -25','c8d7 c2c3 d8b6 d1b3 b6b3 a2b3 c5d4 f3d4 c6d4 c3d4 a7a6 c1d2 g8e7 d2c3 f7f6 b1d2 f6e5 f4e5 e7c6 d2f3 f8e7 f1d3 e8g8 e1f2 c6a7 h1c1 d7b5 d3b5 a7b5 f2g3 g7g6 c3e1 a8c8 c1c8 f8c8');
INSERT INTO evaluations VALUES(16707,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(16707,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16707,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16707,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16707,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16707,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16707,36,'cp -12','b1c3 f8c5 c1f4 d7d6 d3d2 c8e6 a2a3 e8g8 g1f3 a7a6 f1d3 f6g4 f4g3 h7h6 e1g1 g4e5 f3e5 c6e5 d3e2 d8g5 a1d1 g5d2 d1d2 e5c4 e2c4 e6c4 f1e1 f8e8 g3f4 b7b5 b2b3 c4e6 b3b4 c5a7 c3d5 a7b8 f4g3 a6a5 f2f4');
INSERT INTO evaluations VALUES(40397,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(40397,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(40397,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(40397,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(40397,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(40397,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(40397,36,'cp 45','c1e3 b8c6 d1d2 e8g8 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4h5 f6h5 d4g7 g8g7 e1c1 f8h8 f3f4 c8g4 d1e1 a8d8 f4f5 h5f6 f1d3 h8h5 f5g6 f7g6 h1g1 d8h8 e1f1 a5e5 f1f6 e7f6 g1g4 h5h4 g4h4 h8h4 d2d1 e5f4 c1b1 f4h2 a2a3 h2h1 d1h1 h4h1 b1a2 h1h2 a2b3 h2h1');
INSERT INTO evaluations VALUES(47950,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(47950,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(47950,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47950,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(47950,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(47950,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(47950,36,'cp 28','b8c6 f1b5 f8e7 e4e5 f6g4 e5d6 d8d6 d4d6 e7d6 c3e4 e8g8 e4d6 c7d6 e1g1 g4e5 b5e2 f8e8 f3e5 d6e5 c2c3 c8e6 f1e1 f7f6 c1e3 c6e7 e2b5 e7c6 f2f4 e5f4 e3f4 a7a6 b5f1 e6f7 e1d1 f7e6 d1d6 a8d8 c3c4 d8d6 f4d6 g8f7 a1d1 e8d8 a2a3');
INSERT INTO evaluations VALUES(15647,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(15647,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(15647,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(15647,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(15647,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(15647,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(15647,36,'cp 48','f1b5 f8e7 c1f4 e8g8 d4d2 c8d7 e1c1 a7a6 b5d3 f8e8 h1e1 d7g4 h2h3 g4f3 g2f3 c6d4 d3e2 d4e2 c3e2 f6d7 e2c3 e7f6 f4g3 d7b6 d2d3 d8e7 c1b1 a8d8 f3f4 f6c3 d3c3 b6a4 c3a3 b7b5 f4f5 a4c5 e4e5 d6e5 d1d8 e8d8 g3e5');
INSERT INTO evaluations VALUES(15445,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(15445,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(15445,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(15445,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(15445,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15445,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(15445,36,'cp 43','f8e7 c1f4 e8g8 d4d2 c8d7 e1c1 a7a6 b5d3 f8e8 c1b1 b7b5 c3d5 d7g4 h2h4 f6d5 e4d5 c6e5 f3g5 e5c4 d3h7 g8f8 d2d3 e7f6 f4c1 g4d1 d3d1 g7g6 d1d3 d8e7 f2f4 f6g5 h4g5 e7e4 d3e4 e8e4 h7g6 f7g6 h1h8 f8e7 h8a8 c4b6 a8a6 e4e1 a6a7 b6d5 g2g3 e7e6 a7a8 e6d7 a8g8');
INSERT INTO evaluations VALUES(13322,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13322,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(13322,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13322,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13322,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13322,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13322,36,'cp 0','f1e2 e7e5 e1g1 c8d7 d1b3 d8c7 f3g5 c6d8 c3c4 h7h6 g5f3 g7g6 b1c3 f8g7 a2a4 d8c6 c3d5 c7d8 b3b7 f6d5 c4d5 c6a5 b7a6 d7c8 a6b5 c8d7');
INSERT INTO evaluations VALUES(13321,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13321,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13321,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13321,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13321,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13321,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13321,36,'cp -9','e7e5 e1g1 c8d7 d1b3 d8c7 f3g5 d7c8 f2f4 h7h6 g5f3 e5f4 c1f4 g7g5 f4d2 f8g7 c3c4 e8g8 b1c3 c7d8 b3d1 f6e8 d2e3 f7f5 e4f5 c8f5 d1d2 e8c7 h2h3 d8e7 a1e1 a8e8 e2d1 c7e6 c3d5 e7d8 d1a4 g8h8 b2b4');
INSERT INTO evaluations VALUES(13873,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(13873,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(13873,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(13873,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13873,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(13873,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13873,36,'cp 0','e1g1 c8d7 d1b3 d8c7 f3g5 d7c8 c3c4 h7h6 g5f3 f8e7 b1c3 e8g8 c3d5 c7d8 c1d2 a8b8 h2h3 c8d7 b3c2 a7a5 a2a3 b8c8 c2d1 c8b8');
INSERT INTO evaluations VALUES(13872,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13872,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13872,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13872,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13872,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13872,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13872,36,'cp 0','c8d7 d1b3 d8c7 f3g5 d7c8 f2f4 h7h6 g5f3 e5f4 c1f4 g7g5 f4g3 f6h5 g3f2 h5f4 b3c2 f8g7 f2g3 f4h5 g3f2');
INSERT INTO evaluations VALUES(15758,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15758,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(15758,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(15758,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(15758,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(15758,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(15758,36,'cp 60','f8e7 c1f4 a7a6 a2a4 c8e6 b1c3 d8d7 a1d1 e8g8 c4e2 f8e8 h2h3 d7c6 e2f3 c6c4 d4c4 e6c4 f1e1 c4e6 f4g3 f6d7 e4e5 d7e5 f3b7 a8b8 b7a6 b8b2 c3b5 b2c2 f2f4 e5d7 b5d4 c2c3 a6b5 c3g3 d4e6 f7e6 b5d7 e8a8 d7e6 g8f8 e6d5 a8a4');
INSERT INTO evaluations VALUES(8691,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(8691,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(8691,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(8691,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(8691,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(8691,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(8691,36,'cp 55','b1c3 e8g8 c1f4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 f8e8 h2h3 d7c6 e2f3 c6c4 d4d2 c4b4 b2b3 h7h6 f1e1 a8b8 e1e3 b7b5 e3d3 f6d7 a4b5 a6b5 c3d5 b4d2 d5e7 e8e7 d3d2 f7f5 e4f5 e6f5 f3e2 d7f6 f2f3 b5b4 g1f2');
INSERT INTO evaluations VALUES(42809,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(42809,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(42809,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(42809,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42809,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(42809,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(42809,36,'cp 13','b2c3 d7d5 e4d5 d8d5 f1e2 e8g8 e1g1 b8c6 c3c4 d5d6 a2a4 f8e8 c1a3 d6d8 f1e1 c8f5 f3h4 f5d7 a3b2 c6d4 h4f3 c7c5 f3d4 e5d4 c2c3 d7c6 c3d4 c5d4 d1d2 d8d6 b2a3 d6c7 a4a5 h7h6 h2h3 e8e5 a3b4 a8e8 e2f1 f6d7 e1e5 c7e5 a1e1 e5e1 d2e1 e8e1 b4e1 d7c5 e1b4 c5b3 f2f3 f7f6');
INSERT INTO evaluations VALUES(47027,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(47027,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(47027,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(47027,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(47027,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(47027,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(47027,36,'cp -111','f6e4 f1d3 e4f6 e1g1 g7g6 c2c4 f8g7 f1e1 e8g8 b1c3 e7e6 c1g5 h7h6 g5h4 b8d7 e1e3 b7b6 d1d2 c8b7 a1e1 a7a6 h2h3 d8c8 a2a4 b7f3 e3f3 c7c5 d4c5 c8c5 e1d1 a8c8 d3f1 g6g5 h4g3 d7e5 g3e5 d6e5 f3d3 c5c4 d3e3 c4h4');
INSERT INTO evaluations VALUES(47077,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(47077,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(47077,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(47077,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(47077,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(47077,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(47077,36,'cp -90','f1d3 e4f6 c2c4 g7g6 b1c3 f8g7 h2h3 b7b6 e1g1 e8g8 f1e1 c8b7 d4d5 c7c6 c1g5 b8d7 f3d4 c6d5 c4d5 a7a6 a2a4 d7e5 d3f1 a8c8 a1c1 f6d5 c3d5 c8c1 d5e7 d8e7 d1c1 e7d7 b2b3 h7h5 c1d2 f8c8 g5e3 d7d8 e1d1 c8c5 b3b4');
INSERT INTO evaluations VALUES(47076,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(47076,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(47076,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(47076,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(47076,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(47076,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(47076,36,'cp -109','e4f6 h2h3 g7g6 c2c4 b7b6 e1g1 f8g7 b1c3 c8b7 f1e1 e8g8 d4d5 c7c6 c1g5 c6d5 c4d5 b8d7 f3d4 a7a6 a2a4 h7h6 g5h4 a8c8 d1e2 d7c5 d3c4 d8d7 e2e7 f8e8 e7d7 e8e1 a1e1 f6d7 d4c6 g7c3 b2c3 b7c6 d5c6 d7e5 c4f1 c8c6 h4g3 c5a4 f1a6 a4c3 g3e5 d6e5 e1e5 g8g7');
INSERT INTO evaluations VALUES(47013,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(47013,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(47013,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(47013,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(47013,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(47013,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(47013,36,'cp -84','c2c4 g7g6 h2h3 f8g7 b1c3 b8d7 e1g1 b7b6 d4d5 c8b7 f1e1 e8g8 c1g5 c7c6 f3d4 c6d5 c4d5 d7e5 d3f1 a8c8 a1c1 a7a6 a2a4 f6d5 c3d5 c8c1 d5e7 d8e7 g5c1 e7c7 c1e3 f8c8 d1b3 e5d7 e1c1 d7c5 b3a3 c7d8 b2b4');
INSERT INTO evaluations VALUES(57172,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(57172,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(57172,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(57172,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(57172,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(57172,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(57172,36,'cp -107','e5f4 e3f4 f8e7 d2d4 g8f6 f1e2 d7d5 g1f3 e8g8 c1e3 f6g4 e3g1 f8e8 a2a3 e7d6 d1d2 c7c6 b1c3 c8f5 h2h3 g4f6 g1h2 b8d7 e1g1 f6e4 c3e4 f5e4 f3e5 e8e7 a1e1 d8c7 c2c4 d5c4 e2c4 d7f6 c4d3 e4d3 d2d3');
INSERT INTO evaluations VALUES(57332,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(57332,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(57332,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(57332,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(57332,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(57332,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(57332,36,'cp -92','e3f4 g8f6 g1f3 f8e7 d2d4 e8g8 f1e2 d7d5 c1e3 f6g4 e3g1 f8e8 a2a3 e7d6 d1d2 c7c6 b1c3 c8f5 h2h3 g4f6 g1h2 b8d7 e1g1 f6e4 c3e4 f5e4 f3e5 e8e7 g1h1 d7f6 a1e1 d8c7 c2c4 d5c4 e2c4 a8e8 e5g4 f6g4 h3g4 h7h6 h2g3 c7b6');
INSERT INTO evaluations VALUES(57331,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(57331,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(57331,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(57331,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(57331,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(57331,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(57331,36,'cp -104','g8f6 g1f3 f8e7 d2d4 e8g8 f1e2 d7d5 c1e3 f8e8 e1g1 f6g4 e3c1 c7c5 f3e5 g4h6 e2b5 e8f8 c2c3 d8b6 b5d3 b8c6 g1h1 g7g6 d4c5 e7c5 b2b4 c5e3 c1e3 b6e3 d1d2 d5d4 c3d4 e3d4 e5f3 d4b4');
INSERT INTO evaluations VALUES(13572,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13572,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13572,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13572,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(13572,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13572,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13572,36,'cp 0','b1d2 f8d6 d4c5 d6c5 g1f3 e8g8 e1g1 f8e8 b2b4 c5f8 b4b5 e6e5 b5c6 e5f4 c6b7 c8b7 e3f4 d8c7 f3e5 f8d6 d2f3 f6e4 c3c4 a8c8 a1c1 f7f6 c4d5 e4c3 c1c3 c7c3 d3h7 g8h7 f3g5 f6g5 d1h5 h7g8 h5f7 g8h7');
INSERT INTO evaluations VALUES(13571,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13571,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13571,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13571,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13571,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13571,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13571,36,'cp 0','f8d6 d4c5 d6c5 g1f3 e8g8 e1g1 f8e8 b2b4 c5f8 b4b5 e6e5 b5c6 e5f4 c6b7 c8b7 e3f4 d8c7 f3e5 f8d6 d2f3 f6e4 c3c4 a8c8 a1c1 f7f6 c4d5 e4c3 c1c3 c7c3 d3h7 g8h7 f3g5 f6g5 d1h5 h7g8 h5f7 g8h7');
INSERT INTO evaluations VALUES(8388,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8388,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8388,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8388,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8388,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8388,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8388,36,'cp 0','d4c5 d6c5 g1f3 e8g8 e1g1 f8e8 b2b4 c5f8 b4b5 e6e5 b5c6 e5f4 c6b7 c8b7 e3f4 d8c7 f3e5 f8d6 d2f3 f6e4 c3c4 a8c8 a1c1 f7f6 c4d5 e4c3 c1c3 c7c3 d3h7 g8h7 f3g5 f6g5 d1h5 h7g8 h5f7 g8h7');
INSERT INTO evaluations VALUES(26982,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(26982,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(26982,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(26982,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(26982,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(26982,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(26982,36,'cp -48','e5f4 d2d4 g7g5 c2c3 d8d7 b1d2 h7h6 b2b4 f8g7 d1b3 g8f6 h2h3 g4f3 d2f3 e8c8 c4f7 d6d5 f7d5 f6d5 e4d5 d7d5 c1d2 c6e7 a1e1 e7f5 e1e6 h8f8 f1e1 d5b3 a2b3 a7a6 g1f2 g7f6 f2e2 h6h5 e2d1 g5g4 h3g4 h5g4 f3e5 f5g7 e6f6 f8f6 e5g4');
INSERT INTO evaluations VALUES(27029,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(27029,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(27029,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(27029,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(27029,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(27029,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(27029,36,'cp -28','d2d4 g7g5 c2c3 d8d7 b1d2 f8g7 h2h3 g4h5 d4d5 c6e5 f3e5 h5d1 e5d7 d1h5 d2f3 h7h6 g2g3 f4g3 c4b5 e8c8 c1e3 a7a6 b5a4 b7b5 d7b6 c7b6 a4c2 f7f5 e3b6 d8e8 a2a4 h5f3 f1f3 c8b7 b6d4 g7d4 c3d4 f5e4 f3f7 g8e7 a4b5 a6b5 c2e4 b7b6 g1g2 h8f8 a1f1 f8f7 f1f7');
INSERT INTO evaluations VALUES(27032,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(27032,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(27032,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(27032,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(27032,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(27032,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(27032,36,'cp -74','g7g5 c2c3 d8d7 b1d2 h7h6 b2b4 f8g7 d1b3 g4h5 a2a4 g8e7 a4a5 e8g8 b4b5 c6d8 a5a6 b7a6 a1a6 e7c8 h2h3 c8b6 c4d3 f8e8 c1b2 h5g6 c3c4 d8e6 f1a1 h6h5 c4c5 d6c5 d4d5 e6d4 b2d4 g7d4 f3d4 c5d4 b3b4 d7e7 b4e7 e8e7 g1f2 g5g4 a6a5 b6d7 h3g4');
INSERT INTO evaluations VALUES(26723,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(26723,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(26723,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(26723,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(26723,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(26723,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(26723,36,'cp -51','c2c3 d8d7 h2h3 g4h5 d4d5 c6e7 d1b3 h5f3 b3b7 d7c8 b7c8 a8c8 g2f3 f7f5 b1a3 f5e4 f3e4 f8g7 c4a6 c8a8 a3b5 e8d7 b5d4 a8b8 a6b5 b8b5 d4b5 g7e5 b5a7 g8f6 a2a4 f6e4 a7c6 e7c6 d5c6 d7c6 a4a5 c6b7 a5a6 b7a7 f1e1 e4c5 a1a3 h7h5 b2b4 c5d3 e1d1 d3c1 d1c1 g5g4 b4b5 f4f3 g1f2 e5f4 c1e1');
INSERT INTO evaluations VALUES(26722,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(26722,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(26722,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(26722,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(26722,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(26722,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(26722,36,'cp -49','d8d7 h2h3 g4h5 d4d5 c6e7 d1b3 h5f3 b3b7 d7c8 b7c8 a8c8 g2f3 f7f5 e4f5 f8g7 f1e1 g8f6 b1d2 c8b8 d2e4 f6e4 e1e4 g7e5 g1g2 e8d7 f5f6 e7g8 c4b3 g8f6 e4a4 b8b7 h3h4 h7h6 c1d2 f6d5 a1h1 c7c6 a4a6 d5e3 d2e3 f4e3 h1e1');
INSERT INTO evaluations VALUES(43417,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43417,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(43417,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(43417,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(43417,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43417,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(43417,36,'cp 0','e7e5 d4f3 b8d7 a2a4 f8e7 f1c4 d8a5 c1d2 e5f4 d2f4 d7e5 c4d5 e8g8 e1g1 c8e6 g1h1 e6d5 c3d5 f6d5 e4d5 a8d8 f4g3 f8e8 f3e5 d6e5 c2c4 f7f6 d1c2 a5c7 a1d1 e7d6 h2h3 b7b5 b2b3 b5c4 b3c4 d8c8');
INSERT INTO evaluations VALUES(43458,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43458,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43458,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43458,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43458,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43458,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43458,36,'cp 0','d4f3 b8d7 a2a4 f8e7 f1c4 d8a5 c1d2 e5f4 d2f4 d7e5 c4d5 f6d5 d1d5 a5b4 f3e5 d6e5 f4e5 b4b2 c3b5 e7b4 e1f1 b2c2 b5c7 e8e7 e5f4 b4c5 d5g5 e7d7 g5d5');
INSERT INTO evaluations VALUES(43473,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43473,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43473,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43473,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43473,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43473,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43473,36,'cp 0','b8d7 a2a4 f8e7 f1c4 d8a5 c1d2 e5f4 d2f4 d7e5 c4d5 f6d5 d1d5 a5b4 f3e5 d6e5 f4e5 b4b2 c3b5 e7b4 e1f1 b2c2 b5c7 e8e7 e5f4 b4c5 d5g5 e7d7 g5d5');
INSERT INTO evaluations VALUES(46715,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(46715,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(46715,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(46715,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(46715,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(46715,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(46715,36,'cp 55','c1e3 f8g7 b1c3 a7a6 a2a4 e8g8 d1d2 b8c6 g1e2 a8b8 e2c1 b7b6 c1b3 c6b4 c3d1 c7c5 c2c3 b4c6 f1e2 c5d4 c3d4 c6a5 b3a5 b6a5 d1f2 c8d7 e1g1 b8b4 e2d1 d8b6 b2b3 h7h5 g1h1 f8c8 f2d3');
INSERT INTO evaluations VALUES(47442,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(47442,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47442,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(47442,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47442,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47442,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47442,36,'cp 0','d5c3 b2c3 g7g6 f1e2 c7c5 a1b1 f8g7 g1f3 c5d4 c3d4 d8a5 c1d2 a5a2 c2c4 b8c6 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(47535,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47535,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47535,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47535,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47535,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47535,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47535,36,'cp 1','b2c3 g7g6 g1f3 f8g7 f1e2 c7c5 e1g1 e8g8 a1b1 b8c6 c1e3 d8c7 d4d5 f8d8 c3c4 e7e6 e3c5 c6e7 c5b4 e6d5 b4e7 c7e7 c4d5 e7a3 e2c4 a3c5 d1e2 c8f5 b1b5 c5c7 c4d3 d8e8 e2d2 g7c3 d2g5 f5d3 c2d3 c7d7 b5b1 a8d8 g5c1');
INSERT INTO evaluations VALUES(47532,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47532,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47532,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47532,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47532,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47532,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47532,36,'cp 5','g7g6 f1e2 f8g7 g1f3 c7c5 e1g1 e8g8 a1b1 b8c6 c1e3 b7b6 d4c5 b6c5 e2b5 c8b7 d1d8 f8d8 e3c5 g7c3 b5a4 a8b8 h2h3 c3f6 b1b5 b7a8 b5b8 d8b8 f1d1 c6b4 c5a7 a8f3 g2f3 b8c8 d1d7 b4c2');
INSERT INTO evaluations VALUES(48846,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(48846,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(48846,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(48846,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48846,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(48846,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48846,36,'cp 33','e4d5 c7c6 g1e2 c6d5 d2d4 e5e4 e1g1 f8e7 c2c4 d5c4 b1c3 e8g8 c3e4 b8c6 c1e3 f6e4 g2e4 c8g4 f2f3 g4h3 f1e1 d8d7 d1c2 c6b4 c2c4 h3e6 d4d5 b4d5 a1d1 d5e3 d1d7 e3c4 d7e7 f8e8 e7b7 c4d6 b7c7 d6e4 f3e4 e6a2 e1a1 a2e6');
INSERT INTO evaluations VALUES(48617,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(48617,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(48617,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(48617,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(48617,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(48617,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(48617,36,'cp 12','c7c6 g1e2 c6d5 d2d4 e5e4 e1g1 f8e7 c2c4 d5c4 b1c3 e8g8 c3e4 b8c6 c1e3 c8f5 e4d2 f5e6 d1a4 a8b8 e2f4 b7b5 f4e6 f7e6 a4d1 c6d4 d2c4 d4f5 c4e5 d8d1 a1d1 f5e3 f2e3 e7c5 g1h1 c5e3 d1d6 b8e8 g2c6 e8e7 c6b5 e3c5 d6d1 h7h6 d1c1 c5d6 e5c6');
INSERT INTO evaluations VALUES(55452,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(55452,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(55452,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(55452,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(55452,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(55452,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(55452,36,'cp 111','a7a6 f1d3 g8f6 d1e2 f8e7 e1g1 b8c6 e4e5 f6d5 a2a3 d6e5 d4e5 d5c3 b2c3 c8d7 f1d1 c6a5 e2e4 d7c6 e4g4 g7g5 c1e3 h6h5 g4g3 g5g4 f3d4 d8d5 d4c6 d5c6 g3f4 e8c8 f4f7 e7c5 e3c5 c6c5 f7e6 c8b8 a1b1 h8f8');
INSERT INTO evaluations VALUES(55534,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55534,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(55534,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(55534,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(55534,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(55534,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(55534,36,'cp -5','g8f6 c2c4 f8e7 a2a3 e8g8 f1e2 c7c6 d2d3 a7a5 g1f3 a5b4 a3b4 a8a1 b2a1 b8a6 a1c3 d6d5 c4d5 a6b4 d5c6 e5e4 f3e5 e4d3 e5d3 b4c6 e1g1 c8f5 d3f4 d8d1 f1d1 g7g5 c3f6 e7f6 f4d5 f6e5 b1c3 g8g7 g2g4 f5c2 d1d2 c2g6 e2c4 f8d8');
INSERT INTO evaluations VALUES(47790,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(47790,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(47790,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(47790,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(47790,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(47790,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(47790,36,'cp -10','f1e2 f8e7 a2a3 e8g8 d2d3 a7a5 g1f3 a5b4 a3b4 a8a1 b2a1 c7c6 b1d2 d8b6 d1b1 b8a6 c2c3 c8f5 e3e4 f5g6 g2g3 a6c7 e1g1 f8a8 c3c4 c7e6 a1c3 f6d7 b1b3 b6c7 b3b2 c6c5 f1a1 a8a1 b2a1 c5b4 c3b4 e7f8 b4a3 g6h5');
INSERT INTO evaluations VALUES(8583,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(8583,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8583,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8583,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8583,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8583,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(8583,36,'cp 0','c6a5 c4b3 c7c6 c3e2 e8g8 e1g1 a5b3 a2b3 f6h5 e2c3 h5f6');
INSERT INTO evaluations VALUES(8700,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8700,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8700,36,'cp 0','c4b3 c7c6 e1g1 e8g8 c3e2 a5b3 a2b3 f6h5 e2c3 h5f6');
INSERT INTO evaluations VALUES(22262,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(22262,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(22262,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(22262,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(22262,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(22262,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(22262,36,'cp 159','d4e5 g7g5 f4g3 g5g4 f3d4 c6e5 b1c3 f8b4 f2f4 e5f7 f4f5 h7h5 d4b5 b4c3 b2c3 f7e5 d1d5 d8d5 b5c7 e8f7 c7d5 c8f5 g3e5 f6e5 f1d3 g8f6 d5f6 f5d3 c2d3 f7f6 h1f1 f6e6 a1b1 a8c8 e1d2 h8f8 e3e4 b7b5 f1f5 b5b4 b1b4 f8f5 e4f5 e6f5 b4b7 a7a6 c3c4 e5e4 b7b1 e4d3');
INSERT INTO evaluations VALUES(22254,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(22254,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(22254,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(22254,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(22254,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(22254,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(22254,36,'cp 170','g7g5 f4g3 g5g4 f3d4 c6e5 b1c3 h7h5 d1d2 h5h4 g3f4 a7a6 f1e2 d8d6 e1c1 g8e7 c1b1 c8d7 f2f3 h4h3 g2g3 e8c8 f3g4 d6b6 g4g5 f8g7 d4b3 e5c4 d2d4 d8e8 h1e1 f6g5 d4g7 g5f4 c3d5 e7d5 e2c4 b6f6 g7f6 d5f6 g3f4 d7g4 d1d4 c8b8 b3c5 g4f5 e3e4');
INSERT INTO evaluations VALUES(33702,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(33702,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(33702,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(33702,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(33702,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(33702,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(33702,36,'cp 131','e5d6 d2d4 a7a6 e1g1 e7e6 d1d3 f8e7 f1d1 c7c5 c3e4 f6e4 d3e4 c5d4 d1d4 d6b6 f3e5 b8c6 e5c6 b6c6 e4c6 b7c6 e2f3 e8g8 f3c6 a8b8 d4d1 c8b7 c6a4 b7d5 a4b3 e7f6 c2c3 d5b3 a2b3 b8b3 a1a6 h7h5 g1f1 g7g5 d1d7 g5g4 a6a7 h5h4 a7b7 b3b7 d7b7');
INSERT INTO evaluations VALUES(33663,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(33663,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(33663,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(33663,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(33663,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(33663,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(33663,36,'cp 135','d2d4 a7a6 e1g1 e7e6 d1d3 c7c5 f1d1 c5d4 f3d4 b8d7 c3e4 d6e5 e4f6 e5f6 d3g3 f8e7 d4b3 e8g8 c1e3 e6e5 d1d2 f8d8 a1d1 d7f8 g3f3 d8d2 b3d2 f6f3 e2f3 h7h6 d2c4 c8f5 f3b7 f5c2 b7a8 c2d1 c4e5');
INSERT INTO evaluations VALUES(33656,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(33656,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(33656,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(33656,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(33656,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(33656,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(33656,36,'cp 123','e7e6 e1g1 f8e7 c3b5 d6d8 c2c4 a7a6 b5c3 b8d7 c1f4 e8g8 c4c5 d7b8 f1e1 b8c6 a2a3 f6d5 f4g3 e7f6 e2d3 c8d7 d1c2 g7g6 d3e4 f6g7 c3d5 e6d5 e4d5 d7f5 d5e4 f5e4 e1e4 d8d5 g3e5 c6e5 d4e5 a8d8 a1e1');
INSERT INTO evaluations VALUES(55066,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(55066,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(55066,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55066,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55066,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(55066,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55066,36,'cp -8','c8f5 g1f3 e7e6 c2c4 g8f6 d1b3 d8c8 f3h4 f5e4 f2f3 d5c4 d2c4 e4c6 e2e4 c8d8 g2g3 d8d4 c1e3 d4d7 c4a5 f8d6 a1c1 e8g8 a5b7 a7a5 b7c5 d7e8 f1e2 b8d7 c5d7 c6d7 e1f2 e8b8 h1e1 a5a4 b3b8 f8b8 c1c2 d7b5');
INSERT INTO evaluations VALUES(20029,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(20029,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20029,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(20029,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(20029,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20029,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20029,36,'cp 0','f1g2 g8f6 d2d4 c5d4 f3d4 c6d4 d1d4 g7g6 b1c3 f8g7 a2a4 e8g8 e1g1 f6d7 d4d1 b7b6 e4e5 c8a6 e5d6 a6f1 d6e7 d8e7 g2a8 g7c3 g1f1 f8a8 b2c3 a8e8 c1f4 e7e4 f1g1 d7e5 f4e5 e4e5 d1d7 g8f8 d7a7 e5c3 a1c1 e8e1 c1e1 c3e1 g1g2 e1e4 g2f1 e4h1 f1e2 h1e4');
INSERT INTO evaluations VALUES(56134,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(56134,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(56134,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56134,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(56134,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(56134,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(56134,36,'cp -19','d8d5 d4b5 d5d7 d2d4 a7a6 b5c3 f7f5 f2f3 b8c6 f3e4 d7d4 d1d4 c6d4 f1d3 c8d7 e1g1 e8c8 c1g5 f8e7 g5e7 g8e7 b1d2 e7c6 d2b3 h8f8 b3d4 c6d4 a1d1 d7e6 e4f5 e6f5 d3f5 f8f5 f1f5 d4f5 d1d8 c8d8 c3d1 d8e7 g1f2 e7d6 c2c3 d6e5 f2f3 f5d6 d1f2 e5d5 f2d3 a6a5 b2b3 c7c5 d3f4 d5e5');
INSERT INTO evaluations VALUES(34896,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(34896,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(34896,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(34896,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(34896,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(34896,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(34896,36,'cp -21','d4b5 d5d7 d2d4 a7a6 b5a3 g8f6 f1e2 b8c6 c2c3 b7b5 c1e3 d7f5 f2f3 f5g6 f3e4 g6e4 a3c2 c8g4 e1g1 g4e2 d1e2 f8d6 e2f3 e4f3 f1f3 c6e7 e3f4 d6f4 f3f4 e8g8 f4f1 f8e8 a2a4 e7d5 a4b5 e8e2 b1a3 a6b5');
INSERT INTO evaluations VALUES(3617,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(3617,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(3617,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3617,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(3617,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(3617,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3617,36,'cp 10','f1e1 a7a6 a2a4 c5a7 b2b4 c8e6 c4e6 f7e6 c1e3 a7e3 e1e3 f6d7 b1d2 d8f6 d1e2 f8f7 f3h2 c6e7 e3f3 f6g6 f3f7 g6f7 a4a5 g8h8 d2c4 d7f6 c4e3 a8d8 a1f1 f7g6 h2g4 d8f8 e2a2 f6g4 h3g4 g6f7 g2g3');
INSERT INTO evaluations VALUES(3618,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(3618,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(3618,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(3618,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(3618,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3618,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(3618,36,'cp 4','a7a5 d3d4 c5a7 c1e3 f8e8 b1d2 e5d4 f3d4 a7d4 e3d4 c6d4 c3d4 d6d5 e4d5 e8e1 d1e1 f6d5 e1e5 c8e6 d2e4 b7b6 e4c3 c7c6 c4d5 e6d5 c3d5 c6d5 a1c1 a8c8 c1c8 d8c8 b2b3 b6b5 h3h4 c8d7 g1h2 g8h7 g2g3 b5b4 e5f4 h7g8 f4f3 g8h7 h2g1 h7g8 g1g2 g8h7 h4h5');
INSERT INTO evaluations VALUES(15280,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(15280,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(15280,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(15280,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(15280,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(15280,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(15280,36,'cp -67','d8d5 f1e2 c8f5 e1g1 e8c8 a2a3 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 h8e8 a1d1 c5e3 f2e3 f5c5 g1f2 c5e7 d1d8 c6d8 e2d3 g7g6 f1d1 f7f6 d3d4 a7a6 d1d3 d8c6 d4d5 f6f5 b2b4 c6d8 c2c4 d8c6 c4c5 c8b8');
INSERT INTO evaluations VALUES(47970,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(47970,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(47970,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(47970,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(47970,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47970,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(47970,36,'cp 42','d2d3 g7g6 h5f3 d8e7 h2h4 f8g7 b1c3 c7c6 a2a4 c8e6 c4e6 e7e6 b2b4 a7a6 f3h3 e6h3 g1h3 b8d7 a4a5 f7f5 h3g5 g7f6 c1d2 h6f7 g2g3 f7d8 h4h5 f6g5 d2g5 d8e6 g5d2 e8c8 c3a4 f5f4 g3f4 e6d4 e1c1 d4e2 c1b1 e2f4');
INSERT INTO evaluations VALUES(47969,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47969,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47969,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(47969,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(47969,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(47969,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(47969,36,'cp 39','c8g4 h5g5 d8g5 c1g5 b8d7 b1d2 g4e6 g5h6 g7h6 g1f3 a7a5 c4e6 f7e6 g2g3 f8e7 e1e2 h8f8 c2c3 h6h5 a1c1 a5a4 b2b4 a4b3 a2b3 d7c5 c1a1 e8d7 h1c1 e7f6 c1b1 f6e7 h2h3 b7b5 a1a8 f8a8');
INSERT INTO evaluations VALUES(29983,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(29983,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(29983,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(29983,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(29983,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(29983,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(29983,36,'cp 46','h5g5 d8g5 c1g5 h6g8 b1c3 f8e7 g5e3 g4e6 c4b3 g8f6 g1e2 b8a6 f2f4 c7c6 f4e5 f6g4 c3d1 e6b3 a2b3 d6e5 e3d2 a6b4 d2b4 e7b4 c2c3 b4c5 e1d2 g4f6 b3b4 c5b6 e2g3 g7g6 g3f1 h7h5 f1e3 f6g4 d2e2 b6e3 d1e3');
INSERT INTO evaluations VALUES(57245,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(57245,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(57245,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(57245,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(57245,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(57245,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(57245,36,'cp -43','d7d5 e2e3 c8e6 d2d4 b8c6 d4e5 c6e5 g1f3 f8d6 f3e5 d6e5 b1d2 g8f6 c2c4 c7c6 d2f3 e5c7 d1c2 d8e7 b2b4 a7a5 b4b5 f6e4 f1d3 e4c5 d3e2 e8g8 b5c6 b7c6 e1g1 a5a4 c1b2 f8b8 b2e5 c7e5 f3e5 e7d6 c4d5 c6d5 e5f3');
INSERT INTO evaluations VALUES(56500,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(56500,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(56500,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(56500,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(56500,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(56500,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(56500,36,'cp -39','e2e3 c8e6 d2d4 b8c6 d4e5 c6e5 b2b4 f8d6 b1d2 g8f6 c1b2 d8e7 g1f3 e5f3 d1f3 a7a5 b4b5 e8c8 f1d3 f6d7 f3e2 h8e8 d2b3 a5a4 b3d4 d7b6 e1g1 e6d7 f1b1 g7g6 e2d2 c8b8 b2c3 h7h5 c3b4 g6g5 d2c3 e7e5 b4d6');
INSERT INTO evaluations VALUES(56497,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(56497,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(56497,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(56497,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(56497,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(56497,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(56497,36,'cp -46','c8e6 d2d4 b8c6 d4e5 c6e5 b1d2 f8d6 b2b4 g8f6 c1b2 e8g8 g1f3 e5f3 d1f3 f6d7 f3h5 a7a5 b4b5 c7c6 f1e2 d7c5 a1d1 f7f5 e1g1 d8e7 b5c6 b7c6 c2c4 e6f7 h5f3 c5e4 c4d5 f7d5 e2c4 d6a3 c4d5 c6d5 b2a3 e7a3 d2e4 d5e4 f3e2 a3e7 d1d4');
INSERT INTO evaluations VALUES(6540,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(6540,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(6540,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(6540,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(6540,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(6540,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(6540,36,'cp 133','c4f7 e8f7 d1d5 f7f8 d5c5 d7d6 c5c3 h7h6 b1d2 d8f6 e4e5 c6e5 c3c7 e5c6 d2c4 f6e7 c7d6 e7d6 c4d6 g8e7 f1e1 f8g8 a1b1 b7b6 c1a3 c8g4 f3h4 g8h7 h2h3 g4h5 b1c1 h8f8 d6b5 a8c8 b5a7 c6a7 e1e7 a7c6 e7e6 c6d4 c1c8 f8c8 e6b6 c8c2');
INSERT INTO evaluations VALUES(6153,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(6153,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(6153,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(6153,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(6153,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(6153,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(6153,36,'cp 128','e8f7 d1d5 f7f8 d5c5 d7d6 c5c3 h7h6 c1b2 g8f6 b1d2 f8f7 c3b3 c8e6 b2f6 d8d7 b3b2 g7f6 f3h4 c6e7 f2f4 d7a4 a1c1 h8g8 f4f5 e6d7 f1e1 a4b5 b2b5 d7b5 c1c7 g8c8 c7b7 b5c6 b7b4 a7a5 b4d4 d6d5 e4e5 f6e5 e1e5 c6e8 e5e1 c8c2 d2f3 c2a2 f3e5 f7f8');
INSERT INTO evaluations VALUES(4662,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(4662,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(4662,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(4662,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(4662,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(4662,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(4662,36,'cp 127','d1d5 f7f8 d5c5 d7d6 c5c3 d8f6 c3a3 f6g6 b1c3 c8h3 f3e1 h3e6 f2f4 e6c4 f1f3 h7h5 f3g3 g6e8 e1f3 h5h4 g3g5 h8h5 a3a4 c4f7 c1a3 h5g5 f3g5 g8h6 a4d1 c6e7 f4f5 e8c6 d1d2 f8g8 a1e1 c6b6 g1h1 b6a5 a3b2 a5e5');
INSERT INTO evaluations VALUES(4665,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(4665,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(4665,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(4665,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(4665,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(4665,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(4665,36,'cp 126','f7f8 d5c5 d7d6 c5c3 g8f6 b1d2 h7h6 c1b2 f8f7 c3b3 c8e6 b2f6 d8d7 b3b2 g7f6 a1c1 c6e7 f3h4 h8g8 b2b7 c7c5 b7d7 e6d7 d2c4 e7c8 f1d1 g8g4 c4d6 c8d6 d1d6 d7e6 h4f5 g4e4 f2f3 e4e5 f5h6 f7g7 h6g4 e6g4 f3g4 a8c8 c1c4 e5e2 d6d7 g7g6 h2h4 e2a2 d7d5 g6h6 g1h2');
INSERT INTO evaluations VALUES(32867,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(32867,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(32867,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(32867,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(32867,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(32867,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(32867,36,'cp 163','d2d4 e5a5 h2h3 g4e2 d1e2 b8c6 d4d5 c6b4 e2b5 a5b5 f1b5 c7c6 d5c6 b4c6 c1e3 e7e6 e1c1 g8f6 b5e2 a8c8 e2f3 a7a6 c3a4 f6d7 d1d2 d7e5 f3e2 c8d8 c2c3 f8e7 h1d1 d8d2 d1d2 g7g5 a4c5 e7c5 e3c5 h7h5 b2b4 g5g4 a2a4 f7f5 b4b5');
INSERT INTO evaluations VALUES(32859,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(32859,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(32859,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(32859,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(32859,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(32859,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(32859,36,'cp 136','e5a5 h2h3 g4e2 d1e2 b8c6 d4d5 c6b4 e2b5 a5b5 f1b5 c7c6 d5c6 b4c6 c1e3 g8f6 e1c1 e7e6 b5e2 a7a6 e2f3 a8c8 c3a4 f6d7 h1e1 f8e7 c2c3 h7h6 a4b6 c8c7 b6d7 c7d7 c1c2 e7g5 e3b6 g5d8 f3c6 b7c6 d1d7 e8d7 e1d1 d7c8 d1d8 h8d8 b6d8 c8d8 c2b3 d8c7 b3b4 c7b6 b4c4 e6e5 g2g4 f7f6 f2f3');
INSERT INTO evaluations VALUES(43838,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(43838,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43838,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(43838,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(43838,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43838,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43838,36,'cp 34','e2e3 c8b7 h2h4 f8d6 f4d6 c7d6 b1c3 a7a6 a2a3 b6b5 f1d3 d6d5 f3d2 b8c6 g2g4 b5b4 c3a4 d7d6 g4g5 f6d7 f2f4 b4a3 a1a3 c6b4 d3e2 a6a5 e1g1 e8g8 f1f2 e6e5 d2f1 f7f5 f1g3 g7g6 h4h5 d8e7 c2c3');
INSERT INTO evaluations VALUES(43843,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43843,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(43843,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(43843,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43843,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43843,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(43843,36,'cp 0','f1d3 d7d5 e1g1 c8b7 c2c4 f8d6 b2b3 b8d7 c4d5 f6d5 b1d2 e8g8 d2c4 d5c3 d1e1 b7f3 g2f3 d8g5 g1h1 g5h5 c4d6 c7d6 e1c3 f8c8 c3b2 d7e5 d4e5 h5f3 h1g1 f3g4 g1h1');
INSERT INTO evaluations VALUES(43841,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43841,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43841,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43841,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43841,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43841,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43841,36,'cp 0','d7d5 e1g1 f8d6 b2b3 c8b7 c2c4 b8d7 c4d5 f6d5 b1d2 e8g8 d2c4 d5c3 d1c2 b7f3 g2f3 d8g5 g1h1 g5h5 f3f4 h5f3 h1g1 d6b4 d3h7 g8h8 h7d3 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(43806,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43806,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(43806,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43806,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(43806,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43806,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(43806,36,'cp 11','e1g1 c8b7 c2c4 f8d6 b2b3 e8g8 c4d5 f6d5 c1b2 b8d7 b1d2 c7c5 d2e4 d6e7 e4g3 a7a6 a2a4 c5d4 f3d4 d5b4 d3c4 d7f6 d1e2 b7d5 f1c1 d5c4 c1c4 a8c8 b2a3 c8c4 e2c4 a6a5 a1c1 d8d5 c4d5 f6d5 a3b4 d5b4 c1c7 b4d5 c7a7 e7f6 d4c6');
INSERT INTO evaluations VALUES(55150,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(55150,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(55150,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(55150,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(55150,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(55150,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(55150,36,'cp 1','d2d4 c7c5 e2e3 e7e6 g1f3 g8f6 a2a3 a7a6 d4c5 f8c5 b2b4 c5d6 c2c4 d5c4 f1c4 b7b5 c4b3 e8g8 e1g1 a6a5 b1c3 a5b4 c3b5 d6e7 c1b2 b4a3 b5a3 c8b7 b3c4 d8d1 f1d1 b8d7 a3b5 f8c8 a1a8 c8a8 b5d4 d7b6 c4b3');
INSERT INTO evaluations VALUES(21366,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(21366,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21366,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(21366,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21366,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(21366,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21366,36,'cp 13','e6d5 b1c3 g8f6 f1b5 f8b4 b5c6 b7c6 e1g1 e8g8 f3e5 c8a6 f1e1 f8e8 c1g5 d8d6 g5f6 d6f6 a2a3 b4c3 b2c3 e8e7 d1h5 f6d6 h2h3 f7f6 e5g4 a6c8 e1e7 d6e7 g4e3 c8e6 h5f3 a8e8 a3a4 e6f7 a4a5 a7a6 f3g4 e7d6 a1b1 d6a3');
INSERT INTO evaluations VALUES(26910,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(26910,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(26910,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(26910,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(26910,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(26910,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(26910,36,'cp 132','d2d4 e5d4 d1d4 g8e7 d4e3 a7a6 b5a4 g4d7 a4b3 c6a5 h2h4 a5b3 a2b3 d7e6 c1d2 d8d7 c3d5 e7c6 f3d4 e8c8 e1g1 c6d4 e3d4 c7c5 d4d3 e6d5 e4d5 f8e7 b3b4 e7f6 b4c5 d6c5 c2c4 d7g4 d2e3 f6d4 e3d4 g4d4 d3d4 c5d4 f1d1');
INSERT INTO evaluations VALUES(29041,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(29041,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(29041,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(29041,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(29041,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(29041,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(29041,36,'cp 40','c2c4 b8d7 b1d2 f6h5 f4e5 e7e6 f1e2 h5f6 e5f4 f6h5 f4g5 f8e7 g5e7 d8e7 h2h3 g4f3 e2f3 h5f6 d1b3 a8b8 e1g1 h7h5 a2a4 a7a5 c4d5 e6d5 f1b1 e8g8 b3c3 b8a8 f3e2 f8c8 c3a3 e7a3 a1a3 c6c5 d4c5 d7c5 b2b4 a5b4 b1b4');
INSERT INTO evaluations VALUES(29011,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(29011,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(29011,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(29011,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(29011,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(29011,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(29011,36,'cp 34','b8d7 d1b3 f6h5 f4e5 f7f6 e5g3 d8b6 b1c3 e7e6 c4c5 b6b3 a2b3 g4f3 g2f3 a7a6 b3b4 a8d8 f3f4 g7g6 f1d3 f8e7 c3e2 e8f7 e2c1 h5g3 h2g3 h7h5 c1b3 f6f5 d3e2 d8a8 e1f1 h5h4 g3h4 h8h4 h1h4 e7h4 f1g1');
INSERT INTO evaluations VALUES(33208,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(33208,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(33208,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(33208,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(33208,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(33208,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(33208,36,'cp 116','b8c6 d2d4 e5d4 e4e5 f6g6 c3d4 d7d5 c4d5 c5b4 b1c3 g8e7 e1g1 b4c3 f3h4 g6g4 d1g4 c8g4 d5c6 b7c6 b2c3 e7d5 h2h3 g4e6 c1d2 e8g8 f2f4 f7f5 e5f6 d5f6 f1e1 a8e8 h4f3 e6d5 e1e8 f6e8 f3e5 e8d6 a1e1 d6e4');
INSERT INTO evaluations VALUES(280,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(280,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(280,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(280,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(280,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(280,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(280,36,'cp 136','d2d4 e5d4 e4e5 f6g6 c3d4 d7d5 c4d5 c5b4 b1c3 g8e7 e1g1 b4c3 f3h4 g6g4 d5c6 b7c6 d1g4 c8g4 b2c3 e7d5 h2h3 g4e6 c1d2 d5b6 g1h2 a8b8 f2f4 f7f5 g2g4 g7g6 f1g1 e8f7 g4f5 g6f5 g1g5 h7h6 g5g6 b6d5 c3c4 b8b2 c4d5 b2d2 h2g3 d2d3 g3f2 e6d5 g6f6 f7e8 h4f5 h8g8');
INSERT INTO evaluations VALUES(285,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(285,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(285,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(285,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(285,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(285,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(285,36,'cp 110','e5d4 e4e5 f6g6 d1e2 c5e7 c3d4 d7d6 e5d6 g6d6 b1c3 g8f6 c3b5 d6d8 c1f4 e8g8 f4c7 e7b4 e1f1 d8e8 a2a3 b4a5 e2e8 f6e8 c7a5 c6a5 c4a2 a7a6 b5c3 c8f5 f3e5 a8d8 f1e2 f5c2 e2e3 a5c6 e5c6 b7c6 h1e1 e8f6 f2f3 f8e8 e3f2 d8d4 e1e8 f6e8');
INSERT INTO evaluations VALUES(293,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(293,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(293,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(293,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(293,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(293,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(293,36,'cp 111','e4e5 f6g6 c3d4 d7d5 c4d5 c5b4 b1c3 g8e7 e1g1 b4c3 f3h4 g6g4 d1g4 c8g4 d5c6 e7c6 b2c3 g4e6 h4f3 h7h6 f3d2 e8c8 f2f3 d8d5 h2h4 h8d8 a2a4 d5a5 a1a3 a5a6 g2g4 c6a5 d2e4 a6b6 g1h2 b6b1 e4d2 b1b6 f1f2 a5c4 a4a5 b6b5 d2c4 e6c4');
INSERT INTO evaluations VALUES(33816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33816,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33816,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(33816,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33816,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(33816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33816,36,'cp 0','g7g6 g2g3 b8c6 f1g2 c8g4 c2c3 e4f6 e2e7 c6e7 e1g1 h7h6 f1e1 e8c8 a2a4 d8e8 b1a3 f8g7 a3c2 g4e6 a4a5 a7a6 f3d2 e7f5 d2f1 e6d5 c1d2 d5g2 g1g2 c8d7 c2b4 e8e1 a1e1 h8e8 e1e8 f6e8 f1e3 f5e3 d2e3 c7c6 d4d5 c6c5 b4d3 e8f6');
INSERT INTO evaluations VALUES(37750,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(37750,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(37750,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(37750,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(37750,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(37750,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(37750,36,'cp 80','d8a5 c2c3 a5c5 b1a3 g8f6 a3b5 e8g8 c1e3 c5c6 f3d4 c6e4 b5c7 b7b6 d1f3 e4f3 d4f3 c8b7 c7a8 b7a8 f1e2 f6d5 e3d2 a8b7 h2h4 h7h5 e1g1 b8a6 a1d1 a6c5 c3c4 d5f6 f3g5 f6e4 d2e3 d7d6 g5e4 b7e4 f2f3 e4c6 e3d4 a7a5 d4g7 g8g7 d1d2');
INSERT INTO evaluations VALUES(33014,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(33014,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(33014,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(33014,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(33014,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(33014,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(33014,36,'cp 87','c2c3 a5c5 b1a3 g8f6 a3b5 e8g8 c1e3 c5c6 f3d4 c6e4 b5c7 b7b6 d1f3 e4f3 d4f3 c8b7 c7a8 b7a8 f1e2 f6d5 e3d2 d7d6 e1g1 f8c8 f1e1 a8b7 a1c1 b8d7 e2b5 d7c5 b2b4 c5e6 c3c4 d5f6 d2e3 f6g4 b5d7 c8d8 d7e6 f7e6 e1d1 g4e3 f2e3 d8b8');
INSERT INTO evaluations VALUES(33013,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(33013,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(33013,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(33013,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(33013,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(33013,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(33013,36,'cp 80','a5c5 b1a3 g8f6 a3b5 e8g8 c1e3 c5c6 f3d4 c6e4 b5c7 b7b6 d1f3 e4f3 d4f3 c8b7 c7a8 b7a8 f1e2 f6d5 e3d2 d7d6 e1g1 b8d7 f1e1 a8b7 a1c1 e7e5 b2b4 d5e7 e1d1 f8c8 a2a4 d7f6 c3c4 e7f5 d2e1 g7h6 c1c2 h6g7');
INSERT INTO evaluations VALUES(33010,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(33010,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(33010,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(33010,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(33010,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(33010,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(33010,36,'cp 80','b1a3 g8f6 a3b5 e8g8 c1e3 c5c6 f3d4 c6e4 b5c7 b7b6 d1f3 e4f3 d4f3 c8b7 c7a8 b7a8 f1b5 f6d5 e3d2 d7d6 e1g1 f8c8 a1c1 a8b7 f1d1 b8a6 f3d4 a6c5 d4b3 g8f8 b3c5 b6c5 b5e2 e7e6 c1b1 f8e7 c3c4 d5b4 d2f4 d6d5 c4d5 b7d5 a2a3');
INSERT INTO evaluations VALUES(41704,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(41704,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(41704,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(41704,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(41704,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(41704,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(41704,36,'cp -21','e8g8 h2h3 f8e8 c3e2 f6e4 e1g1 c8f5 c2c3 b8d7 c1f4 d8f6 f4d6 e4d6 d3f5 f6f5 e2g3 f5f4 d1b3 c7c6 a1e1 g7g6 g3e2 f4f5 b3a3 d6b5 a3b3 a7a5 e2g3 f5f4 b3c2 b5d6 e1e8 a8e8 f1e1 e8e1 f3e1 f4f6 e1d3 h7h5 g3f1 d7f8 c2a4 f6d8 d3c5');
INSERT INTO evaluations VALUES(36938,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(36938,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(36938,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(36938,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(36938,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(36938,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(36938,36,'cp -12','h2h3 f8e8 c3e2 f6e4 e1g1 c8f5 c2c3 b8d7 c1f4 d8f6 f4d6 f6d6 f1e1 c7c6 d3e4 f5e4 e2g3 h7h6 f3d2 e4h7 d2f3 a7a5 d1d2 d7f8 f3e5 a5a4 e1e2 f7f6 e5d3 a4a3 b2b4 e8e2 g3e2');
INSERT INTO evaluations VALUES(45928,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(45928,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(45928,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(45928,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(45928,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(45928,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(45928,36,'cp 89','g7g6 c1f4 a7a6 a2a4 f8g7 h2h3 e8g8 e2e3 f6e8 f1d3 b8d7 e1g1 d7e5 f4e5 g7e5 f3e5 d6e5 a4a5 e8d6 c3e4 d6e4 d3e4 d8d6 d1c2 c8d7 f1c1 a8c8 e4f3 c5c4 c2c3 f8d8 e3e4 d7b5 c1c2 h7h5 c3e3 d6b4 f3e2 b4c5 e3g5 c5d4 e2f3');
INSERT INTO evaluations VALUES(45527,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(45527,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(45527,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(45527,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(45527,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(45527,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(45527,36,'cp 84','c1f4 f8g7 d1a4 c8d7 a4b3 b7b5 c3b5 d8a5 b5c3 c5c4 b3c4 f6d5 f4h6 g7h6 c4d5 a5b4 d5d4 b4d4 f3d4 b8c6 d4b5 c6b4 e2e3 b4c2 e1d2 c2a1 b5c7 e8d8 c7a8 d7c6 f1a6 c6a8 h1a1 a8g2 a1c1 d8c7 d2e2 h8b8 b2b3 c7c6 a6d3 c6d7 f2f3 b8c8 a2a4 d7d8');
INSERT INTO evaluations VALUES(12380,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(12380,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(12380,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(12380,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(12380,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(12380,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(12380,36,'cp -39','d7d5 e4d5 e6d5 c4a2 c8e6 g1e2 h7h6 e2f4 d8d7 e1g1 e8c8 f4e6 f7e6 c3e2 d7c7 c1f4 f8d6 f4d6 c7d6 d1d2 c8b8 c2c3 h8e8 f1e1 a7a6 h2h3 g7g5 b2b4 e6e5 e2g3 c5b4 a3b4 d5d4 a1b1 d4c3 d2c3 d6d3 c3d3 d8d3 b4b5 a6b5 b1b5');
INSERT INTO evaluations VALUES(13533,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(13533,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(13533,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(13533,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(13533,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(13533,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(13533,36,'cp -33','e4d5 e6d5 c4a2 c8e6 g1e2 h7h6 e2f4 d8d7 e1g1 e8c8 a1b1 c8b8 c1d2 e6g4 d1c1 c6d4 f4d5 f6d5 a2d5 d7d5 c3d5 d4e2 g1h1 e2c1 d5e3 c1b3 c2b3 g4e6 b3b4 d8d3 d2c3 c5b4 a3b4 f7f6 b1d1 d3d1 f1d1 f8e7 d1e1 e7d6 e3c2 e6d7 c2d4 b8c7 h1g1 a7a6 d4e6 d7e6 e1e6 h8d8 h2h4 c7c6 g2g3');
INSERT INTO evaluations VALUES(13513,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(13513,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(13513,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(13513,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(13513,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(13513,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(13513,36,'cp -40','e6d5 c4a2 f8e7 g1e2 c8e6 e1g1 e8g8 e2f4 g8h8 c3e2 e7d6 c2c3 d8c7 f4e6 f7e6 e2g3 a7a5 d1e2 a8e8 c1e3 b7b5 a1e1 c5c4 a2b1 c6e5 d3d4 e5g4 e3g5 e6e5 d4e5 d6e5 e2d2 b5b4 b1f5 b4b3 f5g4 f6g4 d2d5 a5a4');
INSERT INTO evaluations VALUES(13741,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(13741,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(13741,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(13741,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(13741,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(13741,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(13741,36,'cp -40','c4a2 c8e6 g1e2 h7h6 e2f4 d8d7 e1g1 e8c8 a1b1 c8b8 c1d2 e6g4 d1c1 c6d4 f4d5 f6d5 a2d5 d7d5 c3d5 d4e2 g1h1 e2c1 d5e3 c1b3 c2b3 g4e6 b3b4 d8d3 d2c3 f7f6 b1d1 d3d1 f1d1 c5b4 a3b4 f8e7 e3c2 h8e8 c2d4 e6d7 d1e1 a7a6 h1g1 g7g6 d4c2 d7a4 e1c1 e7d6 c2d4');
INSERT INTO evaluations VALUES(13775,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(13775,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(13775,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(13775,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(13775,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(13775,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(13775,36,'cp -33','h7h6 g1e2 c8e6 e2f4 d8d7 e1g1 e8c8 a1b1 c8b8 c1d2 e6g4 d1c1 c6d4 f4d5 f6d5 a2d5 d7d5 c3d5 d4e2 g1h1 e2c1 d5e3 c1b3 c2b3 g4e6 b3b4 d8d3 d2c3 c5b4 a3b4 f7f6 f1d1 d3d1 b1d1 f8e7 d1d4 b8c7 h2h3 h8d8 d4e4 c7d7 g2g4 e7d6 h1g1 a7a6 f2f3 h6h5 g1g2 e6f7 e4d4 h5g4');
INSERT INTO evaluations VALUES(42542,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(42542,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(42542,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(42542,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(42542,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(42542,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(42542,36,'cp -64','e8g8 b1c3 c7c6 d2d3 d7d5 c4b3 c8g4 g5f3 d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 c5f2 h2h3 g4f3 g2f3 f2b6 d1e2 a7a5 a2a4 f8d8 c1g5 d8d7 a1d1 g8f8 h3h4 f7f6 g5d2 b8a6 b3c4 a6c5 b2b3 a8d8 d2c1 d7d1 h1d1 d8d1 e2d1 f8e7 c1a3 e7d7 c4g8 h7h6 h4h5 c5e6 g8e6 d7e6');
INSERT INTO evaluations VALUES(37220,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(37220,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(37220,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(37220,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(37220,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(37220,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(37220,36,'cp -77','d2d3 c7c6 c4b3 d7d5 b1c3 c8g4 g5f3 d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 c5f2 h2h3 g4f3 g2f3 f2b6 d1e2 a7a5 a2a4 f8d8 c1g5 d8d7 g5h4 g8f8 a1d1 f8e8 b3c4 f7f6 c4e6 d7c7 e6g4 b8d7 d1d2 d7f8 h1d1 e8e7 c2c3 f8e6 g4e6 e7e6 d2d8 a8d8 d1d8');
INSERT INTO evaluations VALUES(37218,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(37218,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(37218,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(37218,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(37218,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(37218,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(37218,36,'cp -68','c7c6 b1c3 d7d5 c4b3 c8g4 g5f3 d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 c5f2 h2h3 g4f3 g2f3 f2b6 d1e2 a7a5 a2a4 f8d8 c1g5 d8d7 a1d1 g8f8 h3h4 f7f6 g5d2 b8a6 b3c4 a6c5 d2c1 d7d1 h1d1 a8d8 d1d8 b6d8 b2b3 h7h5 c1e3 d8b6 e3d2 f8e8 c4g8 e8e7 b3b4 c5a4 b4a5 b6d4 d2b4 c6c5 b4a3');
INSERT INTO evaluations VALUES(44068,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(44068,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(44068,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(44068,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(44068,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(44068,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(44068,36,'cp -38','d7d5 e4d5 c6d5 c4b3 b8c6 g1e2 d8a5 b1d2 f8e7 e1g1 h7h6 g5f6 e7f6 f2f4 c8g4 h2h3 a5c5 g1h1 h6h5 f4e5 f6e5 d2f3 e5c7 e2c3 e8c8 d3d4 c5e7 d1e1 e7e6 a1d1 g7g5 e1e6 f7e6 f3g5 g4d1 g5e6 d1c2 b3c2 d8d7 e6c5 c6d4 c5d7 d4c2 d7c5');
INSERT INTO evaluations VALUES(45796,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(45796,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(45796,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(45796,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(45796,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(45796,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(45796,36,'cp -27','e4d5 c6d5 c4b3 b8c6 g1e2 d8a5 b1d2 f8e7 e1g1 h7h6 g5f6 e7f6 f2f4 c8g4 f4e5 c6d4 d2f3 g4f3 g2f3 d4b3 a2b3 a5b6 g1h1 f6e5 e2c3 e8c8 f3f4 e5c3 b2c3 c8b8 d1e1 h8e8 e1f2 d5d4 c3d4 d8d4 a1e1 e8c8 e1e2 b6c6 f2g2 c6f6 f4f5 a7a6 f1g1 f6f5 g2g7 f5d5 g7g2');
INSERT INTO evaluations VALUES(45728,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(45728,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(45728,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(45728,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(45728,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(45728,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(45728,36,'cp -27','c6d5 c4b3 b8c6 g1e2 d8a5 b1d2 f8e7 e1g1 h7h6 g5f6 e7f6 f2f4 c8g4 f4e5 c6d4 d2f3 g4f3 g2f3 d4b3 a2b3 a5b6 g1h1 f6e5 e2c3 e8c8 f3f4 e5c3 b2c3 c8b8 d1e2 h8e8 e2f2 d5d4 c3d4 d8d4 a1e1 e8c8 e1e2 b6f6 h1g1 a7a6 f2f3');
INSERT INTO evaluations VALUES(46479,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(46479,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(46479,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(46479,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(46479,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(46479,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(46479,36,'cp -34','c4b3 b8c6 g1e2 d8a5 b1d2 f8e7 e1g1 h7h6 g5f6 e7f6 f2f4 c8g4 f4e5 c6d4 d2f3 g4f3 g2f3 d4b3 a2b3 a5b6 g1h1 f6e5 e2c3 e8c8 f3f4 e5c3 b2c3 c8b8 d1e2 h8e8 e2f2 d5d4 c3d4 d8d4 a1e1 e8c8 e1e2 b6f6 f2e3 b8a8 h2h3 d4d8 f1g1 a7a6 e3e5 f6c6 g1g2 d8d7 e5g7');
INSERT INTO evaluations VALUES(46480,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(46480,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(46480,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(46480,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(46480,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(46480,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(46480,36,'cp -42','b8c6 g1e2 d8a5 b1d2 f8e7 e1g1 h7h6 g5f6 e7f6 f2f4 c8g4 h2h3 a5c5 g1h1 h6h5 d1e1 e8c8 c2c3 c5e3 f4e5 f6e5 e2d4 e5d4 c3d4 e3e1 a1e1 g4e6 d2f3 h8e8 b3a4 f7f6 a4c6 b7c6 b2b4 c8d7 h1g1 d8b8 a2a3 a7a5 b4a5 b8a8 f3d2 a8a5 d2b3 a5a3 b3c5 d7d6');
INSERT INTO evaluations VALUES(19513,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(19513,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(19513,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(19513,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(19513,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(19513,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(19513,36,'cp -78','g8f6 d5d3 d8a5 c2c3 g7g6 b1d2 f8g7 f1e2 e8g8 d2c4 a5c7 g1f3 b7b5 c4e3 b5b4 e1g1 c8b7 e3d5 f6d5 e4d5 c6e5 f3e5 c7e5 c3c4 e7e6 e2f3 e6d5 f3d5 b7d5 c4d5 a7a5 a1b1 f8e8 c1e3 e5e4 d3e4 e8e4 b2b3 g7c3 g2g3 e4e5 a2a4 e5d5 f1c1 g6g5');
INSERT INTO evaluations VALUES(12857,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(12857,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(12857,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(12857,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(12857,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(12857,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(12857,36,'cp -53','d5d3 e7e6 g1f3 d8b6 b1c3 f8c5 d3d2 f6g4 c3d1 e8g8 f1d3 f7f5 e4f5 d7d5 h2h3 g4e5 f3e5 c6e5 e1g1 d5d4 c2c3 e6f5 f1e1 d4c3 b2c3 e5d3 d2d3 c8e6 c1e3 f8e8 e3c5 b6c5 d3g3 a8d8 e1e5 c5c8 g1h2 b7b6 d1e3 c8c3 e3f5 c3g3 f5g3 d8d2 g3e4 d2d5');
INSERT INTO evaluations VALUES(37954,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(37954,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(37954,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(37954,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(37954,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(37954,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(37954,36,'cp 31','d2d4 c5d4 f3d4 g8f6 f2f3 d8b6 c4b5 c8d7 c1e3 b6c7 g2g4 e8g8 d1d2 a7a6 b5d3 b8c6 e1c1 c6d4 e3d4 e7e5 d4e3 d7e6 c1b1 d6d5 c3d5 f6d5 e4d5 e6d5 d3e4 d5e4 f3e4 a8d8 d2g2 d8d1 h1d1 f8d8 g2f3 d8d1 f3d1 g7f6 d1d5 g8g7 g4g5 f6e7 c2c3');
INSERT INTO evaluations VALUES(37965,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(37965,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(37965,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(37965,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(37965,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(37965,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(37965,36,'cp 25','c5d4 f3d4 g8f6 c4b3 b8c6 d4c6 b7c6 e4e5 d6e5 d1d8 e8d8 c1e3 e7e6 e1c1 d8c7 h1e1 h8d8 e3d2 a7a5 c3a4 f6d7 d2g5 d8e8 d1d2 g7f8 e1d1 h7h6 g5e3 g6g5 f2f3 f8e7 b3c4 e8d8 a2a3 c8b7 c1b1 c6c5 a4c3');
INSERT INTO evaluations VALUES(37994,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(37994,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37994,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(37994,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(37994,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(37994,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(37994,36,'cp 33','f3d4 g8f6 c4b3 b8c6 d4c6 b7c6 e4e5 d6e5 d1d8 e8d8 c1g5 d8e8 e1c1 c8f5 g5e3 a7a6 f2f3 h7h5 b3a4 a8c8 c3d5 h5h4 d5b6 c8c7 h1e1 f6d7 d1d2 f7f6 b6d7 f5d7 e1d1 d7e6 d2d8 e8f7 d8h8 g7h8 d1d8 h8g7 a4b3 e6b3 a2b3 f6f5 d8a8 e5e4 f3e4 f5e4 a8a6');
INSERT INTO evaluations VALUES(37990,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(37990,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37990,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(37990,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(37990,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37990,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(37990,36,'cp 22','g8f6 f2f3 d8b6 c4b5 c8d7 c1e3 b6c7 d1d2 a7a6 b5d3 b8c6 e1c1 h7h5 c1b1 b7b5 d4c6 d7c6 h2h3 a8c8 c3e2 c6b7 g2g4 f6d7 g4h5 h8h5 e2f4 h5h4 e3f2 h4h6 h3h4 d7e5 d3e2 g7f6 h4h5 g6g5 f4d5 b7d5 e4d5 c7b7 f2e3 e5c4 e2c4 b5c4');
INSERT INTO evaluations VALUES(57645,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(57645,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(57645,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57645,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(57645,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(57645,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(57645,36,'cp 24','e7e5 d2d4 g8f6 g1f3 e5d4 e4e5 f8b4 c1d2 d8e7 d2b4 e7b4 d1d2 b4d2 b1d2 f6h5 f3d4 b8c6 d2f3 b7b6 d4c6 d7c6 g2g3 c6c5 f1g2 c8b7 e1g1 a8d8 a1d1 e8e7 f3g5 b7g2 g1g2 d8d4 g5f3 d4d1 f1d1 h5g7 f3g5 h7h6 g5e4 g7f5 d1d3 h8d8 d3d8 e7d8');
INSERT INTO evaluations VALUES(56902,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(56902,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(56902,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(56902,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(56902,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(56902,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(56902,36,'cp 13','d2d4 g8f6 g1f3 e5d4 e4e5 f8b4 c1d2 d8e7 d2b4 e7b4 d1d2 b4d2 b1d2 f6h5 f3d4 b8c6 d2f3 b7b6 d4b5 e8d8 e1c1 h8e8 g2g3 a7a6 b5d4 c8b7 d4c6 b7c6 f1h3 f7f5 h3g2 g6g5 h1e1 f5f4 f3h4 c6g2 h4g2 d8e7 d1d3 e8f8 g3g4 h5g7 h2h4 g7e6 h4g5 e6g5');
INSERT INTO evaluations VALUES(10905,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(10905,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10905,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(10905,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10905,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10905,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(10905,36,'cp 0','f6e4 e1e4 d7d5 c4d5 d8d5 b1c3 d5d6 c3b5 d6g6 d2d3 e7d6 f3e5 c6e5 f2f4 c8g4 d1f1 d6c5 g1h1 e8g8 f4e5 c7c6 h2h3 g4h3 g2h3 c6b5 f1g2 f7f5 e4e2 g6g2 h1g2 a8d8 c1f4 c5e7 g2f3 d8d5 a1e1 f8c8 e5e6 e7f6 f4e5 f6e7');
INSERT INTO evaluations VALUES(12413,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(12413,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12413,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12413,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(12413,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12413,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(12413,36,'cp 1','h7h6 b1d2 f6h5 f4g3 h5g3 h2g3 d7d5 f1d3 d8b6 a1b1 f8e7 d1e2 a7a5 e1g1 c8d7 d4c5 e7c5 e3e4 e8g8 d3c2 c6e7 e4d5 e7d5 e2d3 d5f6 d2e4 f6e4 d3e4 f7f5 e4e2 d7b5 c2d3 b5d3 e2d3 f8e8 d3e2 c5d6 f1e1 d6g3 f3e5');
INSERT INTO evaluations VALUES(24802,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(24802,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(24802,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(24802,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(24802,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24802,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24802,36,'cp 28','e7e6 f1e2 h7h6 e1g1 f8e7 b2b3 e8g8 e2d3 f5d3 d1d3 c6b8 f1d1 b8d7 h2h3 f8e8 e3e4 d5e4 c3e4 c7c5 c1b2 c5d4 d3d4 d8c7 e4f6 d7f6 f3e5 e8d8 d4e3 f6d7 e5d7 d8d7 b2e5 c7c6 e3g3 f7f6 e5f4 a8d8 d1d7 c6d7 f4h6');
INSERT INTO evaluations VALUES(25141,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(25141,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25141,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(25141,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(25141,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(25141,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(25141,36,'cp 34','a2a3 h7h6 b2b4 c6e7 c4c5 f6d7 f1d3 f5d3 d1d3 g7g5 e1g1 f8g7 a3a4 e8g8 c3e2 a7a5 c1a3 a5b4 a3b4 e7c6 b4c3 b7b6 c5b6 c7b6 d3b5 d8c7 f1c1 f8c8 h2h3 c7b7 c3d2 c6a5 c1c8 b7c8 d2a5 a8a5 a1c1 c8e8');
INSERT INTO evaluations VALUES(30534,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(30534,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(30534,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(30534,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(30534,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(30534,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(30534,36,'cp 113','g1f3 a7a6 f1c4 e7e6 d4d5 e6e5 a2a4 f8e7 g2g4 h7h6 h1g1 e7g5 f3g5 h6g5 c1e3 f7f6 g1h1 b8d7 d1d2 d7f8 e1c1 g8e7 f2f3 f8g6 c1b1 b7c8 b2b3 g6h4 d2e2 e7g6 c3a2 b6b5 a4b5 a6b5 c4b5 c8d7 b5d7 d8d7 a2c3 g6f4 e3f4 g5f4 b1b2 d7e7');
INSERT INTO evaluations VALUES(30533,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(30533,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(30533,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(30533,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(30533,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(30533,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(30533,36,'cp 124','a7a6 f1c4 e7e6 d4d5 e6e5 a2a4 f8e7 g2g3 b8d7 d1e2 g8f6 f3h4 g7g6 h4g2 h7h5 h2h4 c7c6 d5c6 b7c6 f2f3 e8g8 g2e3 b6b5 a4b5 a6b5 a1a8 c6a8 c4b5 d7b6 e1g1 d6d5 c3d5 b6d5 e3d5 f6d5 e4d5 d8d5 b5c4 d5c6 c4d3 e7f6 g1g2 f6g7 f1d1 c6b6');
INSERT INTO evaluations VALUES(32090,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(32090,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(32090,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(32090,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(32090,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(32090,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(32090,36,'cp 22','b8d7 d2d4 g8f6 d1b3 d6d5 c4d5 f6d5 b3d5 g4f3 g2f3 d8f6 h1g1 c7c6 d5b3 e5d4 c3d4 f6d4 c1e3 d4f6 e1e2 f8c5 b3c3 f6e7 g1g7 c5e3 f2e3 e7h4 g7g2 e8c8 b1d2 h8g8 a1g1 g8g2 g1g2 h4h3 g2f2 h3e6 c3c4 e6e5 c4f7 e5b2 f7b3 b2f6 b3c2');
INSERT INTO evaluations VALUES(36463,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36463,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36463,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(36463,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(36463,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(36463,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36463,36,'cp 0','f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 c7c5 c2c3 b8d7 e1g1 d7f6 g5f6 g7f6 d4c5 d8c7 b2b4 f6c3 a1b1 a7a5 a2a3 a5b4 a3b4 c3f6 d1c1 a8a4 e4c2 a4a2 c2d3 a2a4');
INSERT INTO evaluations VALUES(6768,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(6768,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(6768,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(6768,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(6768,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(6768,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(6768,36,'cp 108','d2d4 g8f6 f3e5 d6e5 d4e5 c6e5 c4e2 e8g8 b1c3 f8e8 e1g1 d7d6 f2f3 h7h6 d1d4 c8e6 b2b3 e5g6 f1d1 d6d5 e4d5 e6d5 e2f1 c7c6 d4f2 d8e7 c1b2 a7a5 d1e1 e7f8 a2a4 e8e1 a1e1 g6f4 c3d5 f6d5 c2c4');
INSERT INTO evaluations VALUES(6772,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(6772,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(6772,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(6772,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(6772,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(6772,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(6772,36,'cp 73','c6d4 f3d4 e5d4 e1g1 g8e7 d1d4 e8g8 c1e3 b7b6 b1c3 c8b7 a2a3 e7g6 f2f4 d8e7 a1e1 a8e8 g2g3 d6c5 d4d2 c5e3 e1e3 e7c5 c4a2 a7a5 f1e1 d7d6 d2f2 g6e7 a2b3 b7a6 h2h4 g8h8 g3g4 f7f6 f4f5 a6c4 b3c4 c5c4 e3d3');
INSERT INTO evaluations VALUES(18990,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(18990,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(18990,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(18990,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18990,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(18990,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(18990,36,'cp 45','d7d6 a2a3 a7a5 d2d3 g7g6 c3d5 h7h6 c1d2 g8f6 e2e3 f6d5 c4d5 c6e7 b2b4 b7b6 b4c5 b6c5 g1e2 f8g7 a1b1 e8g8 d1c2 f7f5 e2c3 f5f4 e3f4 e5f4 d2f4 g6g5 f4e3 e7f5 e1g1 f5e3 f2e3 f8f1 b1f1 d8e7 c3e4 a8b8');
INSERT INTO evaluations VALUES(20610,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(20610,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(20610,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(20610,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(20610,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(20610,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(20610,36,'cp 37','a2a3 a7a5 d2d3 g7g6 c3d5 h7h6 a1b1 g8f6 d5f6 d8f6 e2e3 f8g7 g1e2 e8g8 e2c3 f6d8 e1g1 c8e6 c1d2 g8h7 d1c2 f7f5 f1e1 h6h5 h2h4 g7h6 c3e2 h6g7 b2b4 a5b4 a3b4 c5b4 d2b4');
INSERT INTO evaluations VALUES(42379,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(42379,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(42379,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(42379,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(42379,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(42379,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(42379,36,'cp 200','e5c4 f6e4 d2d3 b7b5 d3e4 b5c4 f1c4 e8g8 e1g1 a7a5 b1c3 c8a6 c4d3 d6e5 h2h3 e5c3 b2c3 d7d6 c1e3 b8d7 d1e2 a6b7 f2f3 d7e5 d3b5 a5a4 f1e1 c7c6 b5d3 a8b8 c3c4 b7a6 f3f4 e5d3 c2d3 a4a3 a1b1');
INSERT INTO evaluations VALUES(42396,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(42396,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(42396,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(42396,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(42396,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(42396,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(42396,36,'cp 207','f6e4 d2d3 b7b5 d3e4 b5c4 f1c4 e8g8 e1g1 a7a5 c1e3 c8a6 b1d2 f8e8 c2c3 b8c6 c4a6 a8a6 f2f4 d6f8 d1a4 d7d5 e4e5 d8d7 a4c2 a6a8 c2d3 c6e7 d2f3 e7f5 e3f2 g7g6 b2b3 c7c6 a1d1 f5g7 d3d2 g7e6 h2h3 a5a4 b3b4');
INSERT INTO evaluations VALUES(56606,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(56606,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56606,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(56606,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56606,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56606,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(56606,36,'cp -12','f7f5 e4g3 e5e4 f3e5 d8d6 d2d4 b8c6 c1f4 g8e7 f1e2 c6d4 e2h5 g7g6 e5g6 d4f3 h5f3 d6d1 a1d1 h7g6 f3e2 c8e6 f2f3 e7d5 f4e5 h8h7 f3e4 d5e3 e4f5 e6f5 g3f5 e3d1 f5d4 d1b2 d4b5 a7a6 b5c7 h7c7 e5c7 a8c8 c7f4 c8c2 f4d2 b2c4 e2c4 c2c4 h2h4 c4c2 h4h5 c2a2');
INSERT INTO evaluations VALUES(20117,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(20117,31,'cp 241',NULL);
INSERT INTO evaluations VALUES(20117,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(20117,33,'cp 247',NULL);
INSERT INTO evaluations VALUES(20117,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(20117,35,'cp 251',NULL);
INSERT INTO evaluations VALUES(20117,36,'cp 248','f3e5 d6e5 d1h5 g7g6 h5e5 g8f6 f1e2 f8g7 e1g1 e8g8 b1c3 e7e6 d5e6 c8e6 e5g3 d8d4 e2f3 f6d7 f1d1 d4b4 a2a3 b4b6 f3e2 f8d8 d1d6 b6a5 c1f4 d7b6 d6d8 a8d8 h2h4 h7h6 f4e5 g7e5 g3e5 b6d7 e5g3 a5b6 a1b1 g8h7 e4e5 e6f5');
INSERT INTO evaluations VALUES(20115,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(20115,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(20115,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(20115,33,'cp 251',NULL);
INSERT INTO evaluations VALUES(20115,34,'cp 234',NULL);
INSERT INTO evaluations VALUES(20115,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(20115,36,'cp 256','d6e5 d1h5 g7g6 h5e5 g8f6 b1c3 f8g7 f1e2 e8g8 e1g1 e7e6 d5e6 c8e6 e5g3 d8b6 f2f4 g8h8 f1e1 a8e8 a2a4 f6h5 g3f2 g7d4 c1e3 h5f6 e2f3 e8d8 e4e5 f6e8 e3d4 c5d4 c3b5 e8g7 c2c3 g7f5 g2g4 f5e3 c3d4 e3g4');
INSERT INTO evaluations VALUES(20130,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(20130,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(20130,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(20130,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(20130,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(20130,35,'cp 242',NULL);
INSERT INTO evaluations VALUES(20130,36,'cp 240','d1h5 g7g6 h5e5 g8f6 b1c3 f8g7 f1e2 e8g8 e1g1 e7e6 d5e6 c8e6 e5g3 f8e8 c1f4 d8b6 a1b1 h7h6 a2a3 g8h7 h2h3 a8c8 f1e1 a7a6 h3h4 e8d8 f4e5 d8d7 h4h5 g6g5 e5f6 g7f6 c3d5 e6d5 e4d5 c8d8 c2c4 h7g8 e2g4 d7e7 e1e7 f6e7 g3f3 b6f6');
INSERT INTO evaluations VALUES(52208,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(52208,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(52208,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(52208,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(52208,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(52208,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(52208,36,'cp 173','f8b4 f1c4 d7d5 e4d5 c6d5 c4b3 b8c6 e1g1 b4c3 b2c3 e5e4 f3h4 g8e7 d1h5 g7g6 h5h6 c8e6 f2f3 d8d7 f3e4 e8c8 e4d5 e6d5 c1a3 g6g5 h4f3 e7g6 a1e1 h8e8 b3d5 d7d5 a3c1 d5d7 h6h5 c8b8 e1e8 d8e8 f1e1 e8e1 f3e1 g6f4 c1f4 g5f4 e1d3');
INSERT INTO evaluations VALUES(5801,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(5801,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(5801,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(5801,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(5801,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(5801,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(5801,36,'cp 117','e4e5 c6b8 f1d3 b7b6 f3d2 c8a6 d3a6 b8a6 c3e2 b4e7 c2c3 a6b8 d2f3 c7c5 c1e3 b8c6 e1g1 d8d7 a1c1 c6a5 b2b3 a8c8 h2h4 e7d8 d1d2 c5d4 c3d4 c8c1 f1c1 g8e7 c1c3 h7h6 e2f4 e7f5 h4h5 a5c6 g2g4');
INSERT INTO evaluations VALUES(5832,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(5832,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(5832,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(5832,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(5832,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(5832,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(5832,36,'cp 116','b7b6 f1b5 g8e7 e1g1 b4c3 b2c3 h7h6 f3d2 e8g8 f2f4 c6a5 b5d3 c7c5 d2f3 f7f5 e5f6 f8f6 g2g4 f6f8 a2a4 c8d7 d1e1 a5c4 h2h4 c4d6 h4h5 a8c8 c1a3 d6e4 e1e3 d7a4 a1e1 e7c6 d3e4 d5e4 e3e4 d8d5 e4e6 d5e6 e1e6 a4c2');
INSERT INTO evaluations VALUES(18681,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(18681,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(18681,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(18681,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(18681,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(18681,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(18681,36,'cp -16','g8f6 b1c3 d8c7 b5e2 f8c5 d4b3 c5b4 d1d3 e8g8 e1g1 b4c3 b2c3 d7d5 e4d5 f6d5 c3c4 d5f4 c1f4 c7f4 a1d1 e6e5 d3f3 c8f5 f3f4 e5f4 e2d3 f5g6 d1d2 b7b6 f1d1 f8c8 h2h3 f7f6 d3g6 h7g6 c4c5 b6c5 b3c5 c6b4 c5d3 b4d3 d2d3 c8c2');
INSERT INTO evaluations VALUES(12317,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(12317,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(12317,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(12317,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(12317,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(12317,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(12317,36,'cp -20','b1c3 d8c7 b5e2 f8c5 d4b3 c5b4 d1d3 e8g8 e1g1 b4c3 b2c3 c7e5 e2f3 f8e8 c1e3 d7d5 e4d5 e6d5 b3d4 e5d6 d4c6 b7c6 e3d4 f6g4 f3g4 c8g4 f2f3 g4e6 d3e3 e6f5 e3f2 f5g6 f1e1 h7h6 e1e8 a8e8 a1e1');
INSERT INTO evaluations VALUES(12316,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(12316,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(12316,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(12316,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(12316,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(12316,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(12316,36,'cp -19','d8c7 b5e2 f8c5 d4b3 c5b4 d1d3 e8g8 e1g1 b4c3 b2c3 c7e5 e2f3 d7d5 e4d5 e6d5 b3d4 f8e8 c1e3 c8g4 d4c6 b7c6 e3d4 e5d6 f3g4 f6g4 d3g3 d6g3 h2g3 f7f6 f1e1 g8f7 c3c4 d5c4 e1e8 a8e8 d4a7 e8a8 a7d4 a8a4 f2f3 c4c3 a1d1 g4h6 d4c3 a4a2 d1d2');
INSERT INTO evaluations VALUES(30921,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(30921,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(30921,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(30921,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(30921,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(30921,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(30921,36,'cp -21','d7b5 c3b5 g8f6 b5c3 a7a6 e1g1 b8c6 d2d4 c5d4 f3d4 c6d4 d1d4 e7e6 c3d1 f8e7 c2c4 b7b5 d1c3 b5c4 d4c4 e8g8 c1d2 d8b6 b2b3 f8c8 c4d3 b6b7 f1c1 c8c6 c1c2 a8c8 a1c1 h7h6 h2h3 a6a5 c1d1 f6d7 d3e2 a5a4 b3a4');
INSERT INTO evaluations VALUES(31238,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(31238,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(31238,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(31238,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(31238,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(31238,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(31238,36,'cp 0','c3b5 g8f6 b5c3 b8c6 d2d4 c5d4 f3d4 g7g6 d4c6 b7c6 d1f3 d8c7 e1g1 f8g7 c1d2 e8g8 b2b3 f6d7 a1d1 d7c5 h2h3 a8b8 f3e2 c7b7 f1e1 c5e6 e2d3 e6c5');
INSERT INTO evaluations VALUES(31237,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31237,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(31237,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31237,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(31237,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(31237,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(31237,36,'cp -18','g8f6 b5c3 b8c6 d2d4 c5d4 f3d4 g7g6 d4c6 b7c6 d1f3 d8c7 e1g1 f8g7 c1d2 f6d7 a1d1 e8g8 b2b3 d7c5 f3e3 a7a5 a2a4 e7e6 h2h3 c7a7 c3e2 d6d5 e4e5 f8e8 f1e1 g7f8 f2f3 c5d7 e3a7 a8a7 e2d4 c6c5');
INSERT INTO evaluations VALUES(35272,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(35272,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(35272,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(35272,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(35272,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(35272,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(35272,36,'cp -53','b8c6 f1c4 g8f6 h5h4 c6d4 f3d4 e5d4 e1g1 f6e4 h4f4 c7c6 d2d3 e4c5 f4d4 d7d5 c4b3 c5e6 d4e5 e7d6 e5d6 f8d6 c1e3 a7a5 b1c3 b7b5 a2a4 b5b4 c3e2 c8a6 e2d4 e8d7 f1c1 h8c8 c1e1 c8e8 d4f3 e6c5 f3d2 g7g6 h2h4 h7h5 g2g3 c5e6 d2f3');
INSERT INTO evaluations VALUES(2049,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(2049,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(2049,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(2049,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(2049,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(2049,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(2049,36,'cp -53','b1c3 g8f6 h5h4 c6d4 f3d4 e5d4 c3e2 d7d5 d2d3 e7b4 c2c3 d4c3 b2c3 b4a4 c1g5 d5e4 g5f6 g7f6 h4f6 h8g8 f6e5 c8e6 e5e4 a4e4 d3e4 e8c8 e2d4 f8b4 e1c1 e6g4 f2f3 b4c3 d4b5 c3b2 c1b2 d8d1 b5c3 d1e1 f3g4 g8g6 g4g5 g6g5 g2g3 c8d8 f1g2 e1h1 g2h1 c7c6 h1f3 b7b5 b2c2 d8e7 c3e2 g5c5 c2d3 a7a5 e2d4 b5b4');
INSERT INTO evaluations VALUES(52375,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(52375,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(52375,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(52375,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(52375,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(52375,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(52375,36,'cp -9','d7d5 e4e5 c6c5 g1f3 d8c7 d2d4 c8g4 f1b5 b8c6 c2c4 c5d4 e1g1 d5c4 b5c6 c7c6 f3d4 g4d1 d4c6 d1h5 c6a5 b7b6 a5c4 h5e2 f1e1 e2c4 b3c4 e7e6 e1d1 f8b4 c1a3 b4e7 c4c5 e7c5 a3c5 b6c5 d1c1 g8e7 c1c5 e8g8 b1a3 f8d8 a3b5 e7d5 a1d1 h7h6 b5d6');
INSERT INTO evaluations VALUES(52793,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(52793,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(52793,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(52793,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(52793,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(52793,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(52793,36,'cp -16','e4d5 c6d5 c1b2 g8f6 g1f3 g7g6 f1b5 c8d7 b5d7 d8d7 d1e2 f8g7 b1c3 e8g8 d2d4 g7h6 e1g1 f8c8 f1e1 e7e6 f3e5 d7d8 a1d1 b8c6 c3a4 d8c7 c2c4 d5c4 b3c4 c6e5 d4e5 f6h5 a4c3 c7c4 e2c4 c8c4 d1d7 h6f8');
INSERT INTO evaluations VALUES(52708,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(52708,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52708,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(52708,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52708,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(52708,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(52708,36,'cp -25','c6d5 g1f3 b8c6 c1b2 c8g4 h2h3 g4f3 d1f3 e7e6 f1b5 g8e7 c2c4 a7a6 b5a4 d5c4 b3c4 d8b6 b2c3 e7f5 a4c6 b7c6 e1g1 a8d8 f1e1 f5d4 c3d4 d8d4 b1c3 f8e7 a1b1 b6c7 d2d3 e8g8 b1b3 f8d8 c3e2 d4d7 e1b1 e7f6 f3e4 c6c5 b3b7 c7d6 b7b3 g8h8 h3h4 g7g6');
INSERT INTO evaluations VALUES(53407,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(53407,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(53407,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(53407,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(53407,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(53407,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(53407,36,'cp -11','c1b2 b8c6 g1f3 c8g4 h2h3 g4f3 d1f3 e7e6 f1b5 g8e7 c2c4 a7a6 b5a4 d5c4 b3c4 d8b6 b2c3 e7f5 a4c6 b7c6 e1g1 f8c5 d2d3 c5d4 c3d4 f5d4 f3e3 e8g8 b1d2 c6c5 a1b1 b6c7 d2b3 f8d8 f1e1 h7h6 e3e5 c7e5 e1e5 d4b3 b1b3 a8c8 g2g3 g8f8 g1g2');
INSERT INTO evaluations VALUES(53408,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(53408,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(53408,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(53408,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(53408,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(53408,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(53408,36,'cp -20','g8f6 g1f3 g7g6 f1b5 c8d7 b2f6 e7f6 d1e2 f8e7 b1c3 d7c6 b5c6 b7c6 e1g1 e8g8 f1e1 f8e8 b3b4 d8b6 a1b1 b8d7 b4b5 e7c5 e2f1 c6b5 e1e8 a8e8 f1b5 b6e6 d2d4 e8b8 b5a4 b8b1 c3b1 c5b6 b1c3 g8g7 g2g3 g6g5 c3d1 e6e2 a4d7 e2d1 g1g2 d1c2 d7d5 c2d3 d5b3');
INSERT INTO evaluations VALUES(9920,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(9920,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(9920,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(9920,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(9920,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(9920,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9920,36,'cp 0','d2d3 d7d6 a2a4 a7a5 b4b5 c6e7 e1g1 e8g8 b1d2 e7g6 c4b3 c7c6 d2c4 b6c7 f1e1 h7h6 d3d4 e5d4 c3d4 c6b5 a4b5 f8e8 c4d2 c7b6 h2h3 c8e6 b3e6 e8e6 d1b3 d8e8 b3d3 g6f4 d3f1 f4g6');
INSERT INTO evaluations VALUES(9922,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9922,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9922,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9922,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9922,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9922,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9922,36,'cp -5','d7d6 e1g1 e8g8 a2a4 a7a5 b4b5 c6e7 b1d2 e7g6 c4b3 c7c6 d2c4 b6c7 c1g5 h7h6 g5f6 d8f6 c4e3 c8e6 d3d4 c6b5 a4b5 a8c8 e3d5 e6d5 b3d5 c7b6 a1a3 c8c7 g2g3 g6e7 d5a2 f8c8 d1d3 e5d4 c3d4 d6d5 f3e5 d5e4 d3e4 e7f5 e5f7 c7f7');
INSERT INTO evaluations VALUES(8912,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8912,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8912,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8912,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(8912,36,'cp 6','a2a4 a7a5 b4b5 c6e7 b1d2 e7g6 e1g1 e8g8 c4b3 c7c6 d2c4 b6c7 f1e1 h7h6 d3d4 c8e6 c1a3 c6b5 a4b5 a5a4 b3a2 e5d4 f3d4 e6c4 a2c4 c7a5 a1c1 g6e5 c4f1 a5b6 h2h3 d8d7 d1c2 b6c5 c1a1 f8e8 a3c5 d6c5 d4f5 c5c4 f5d4');
INSERT INTO evaluations VALUES(8916,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8916,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(8916,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8916,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8916,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8916,36,'cp 0','a7a5 b4b5 c6e7 e1g1 e8g8 b1d2 e7g6 c4b3 c7c6 d2c4 b6c7 f1e1 h7h6 d3d4 e5d4 c3d4 c6b5 a4b5 f8e8 c4d2 c7b6 h2h3 c8e6 b3e6 e8e6 d1b3 d8e8 b3d3 g6f4 d3f1 f4g6');
INSERT INTO evaluations VALUES(19756,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(19756,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(19756,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(19756,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(19756,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(19756,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(19756,36,'cp 91','e2e4 g8f6 e4d5 d8d5 d1d5 f6d5 f1c4 e7e6 c4d5 e6d5 c1e3 c8g4 c2c3 g4f3 g2f3 e8c8 b1d2 f8e7 d2b3 c6e5 e1c1 e5f3 b3d4 f3d4 e3d4 h8e8 b2b4 e7h4 c1c2 g7g6 d1d2 e8e6 h1d1 h4g5 d2d3 g5f6 d4e3 e6e4 d3d5 d8d5 d1d5 e4h4 a2a4 h7h5 c5c6 b7c6 d5a5 h4h2 a5a7');
INSERT INTO evaluations VALUES(19748,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(19748,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(19748,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(19748,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(19748,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(19748,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(19748,36,'cp 76','g8f6 e4d5 d8d5 d1d5 f6d5 f1c4 e7e6 c4d5 e6d5 c1e3 c8g4 c2c3 g4f3 g2f3 e8c8 b1d2 d5d4 c3d4 c6d4 a1c1 f8e7 h1g1 d4e6 d2c4 d8d5 b2b4 d5h5 e1e2 h8d8 f3f4 g7g6 g1d1 d8d1 c1d1 h5h2 c4e5 f7f6 e5c4 h2h4 e2f3 e6g7 b4b5 g7f5 c5c6 b7c6 b5c6 f5e3');
INSERT INTO evaluations VALUES(48388,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(48388,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(48388,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(48388,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(48388,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(48388,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(48388,36,'cp -20','d5d4 c3b1 c7c5 g2g3 b8c6 b1d2 e6e5 d2c4 d8c7 a2a4 h7h6 f3h4 c8g4 f1e2 g4e6 b2b3 c7d7 e1g1 g7g5 h4f5 e8c8 c1d2 f6e8 c4a5 c6a5 d2a5 b7b6 a5d2 e8g7 a4a5 b6b5 c2c4 b5b4 f5g7 f8g7 e2f3 f7f6 a5a6 g5g4 f3g2');
INSERT INTO evaluations VALUES(52781,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(52781,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52781,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52781,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52781,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(52781,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52781,36,'cp 7','c8f5 c2c4 e7e6 b1c3 d5c4 g1f3 b8d7 e2e4 f5g6 d1e2 g8f6 e1g1 f8e7 f3h4 g6h5 e2c4 e8g8 g2f3 b7b5 c4e2 h5f3 h4f3 b5b4 c3a4 a8c8 c1f4 d8a5 b2b3 a5b5 f1e1 b5e2 e1e2 c6c5 d4c5 d7c5 a4c5 c8c5 g1g2 f8c8 a1d1 c5c3');
INSERT INTO evaluations VALUES(31095,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31095,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(31095,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31095,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31095,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31095,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(31095,36,'cp 12','c2c4 e7e6 b1c3 d5c4 g1f3 b8d7 e2e4 f5g6 d1e2 g8f6 e1g1 f8e7 f3h4 g6h5 e2c4 e8g8 g2f3 b7b5 c4e2 h5f3 h4f3 b5b4 c3a4 d8a5 b2b3 a5b5 f1e1 b5e2 e1e2 c6c5 c1b2 c5d4 b2d4 f8c8 g1g2 c8c7 a4b2 a8c8 b2c4 d7c5 d4c5 e7c5 a1c1 h7h6 c4e5 g7g5 c1c4');
INSERT INTO evaluations VALUES(19275,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19275,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(19275,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(19275,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(19275,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19275,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(19275,36,'cp -12','f8g7 c1e3 c6d4 a2a4 d7d6 f1e2 b7b6 e1g1 c8b7 a4a5 e7e5 a5b6 a7b6 a1a8 b7a8 d1d2 g8f6 f1a1 a8c6 e3g5 d4e6 h2h4 e8g8 g2g3 d8d7 g5f6 g7f6 c3d5 c6d5 e4d5 e6d4 f3d4 c5d4 e2f3 h7h5 c2c3 d4c3 b2c3 b6b5');
INSERT INTO evaluations VALUES(48795,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48795,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48795,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(48795,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48795,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48795,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(48795,36,'cp 12','c4d5 f6d5 g1f3 b8c6 g2g3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 a2a3 a7a5 d1a4 e8g8 c1e3 f8d8 a1c1 e6d5 c1c2 h7h6 f1c1 a8b8 a4b5 d7e6 f3d2 f7f5 g2d5 d8d5 b5b3 a5a4 b3a4 f5f4 e3c5 e7c5 c2c5 c6d4 c5d5');
INSERT INTO evaluations VALUES(48606,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(48606,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48606,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(48606,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(48606,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(48606,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(48606,36,'cp 4','f6d5 g2g3 b8c6 g1f3 f8e7 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 d1a4 e8g8 c1e3 f8d8 f1e1 f7f6 e1c1 c6b4 a4d7 d8d7 a2a3 b4d5 e3c5 c7c6 f3d2 e7c5 c1c5 d5c7 g1f1 c7b5 c5c1 b5d4 e2e3 d4b3 d2b3 e6b3 c1c3 b3d5 g2d5 d7d5 a1b1');
INSERT INTO evaluations VALUES(49021,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49021,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49021,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(49021,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49021,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(49021,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(49021,36,'cp 10','g2g3 f8e7 g1f3 b8c6 c3d5 d8d5 f1g2 a7a5 e1g1 a5a4 c1e3 d5b5 b2b3 e8g8 b3a4 b5a4 d1a4 a8a4 f1c1 e7f6 a1b1 a4a2 f3d2 c6d4 c1c7 d4e2 g1f1 e2d4 g2d5 c8h3 f1g1 a2a4 e3d4 a4d4 d5e4 f6d8 c7c4 d8a5 c4d4 e5d4 d2c4');
INSERT INTO evaluations VALUES(49020,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49020,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(49020,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49020,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49020,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49020,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49020,36,'cp 0','f8e7 g1f3 b8c6 c3d5 d8d5 f1g2 c8e6 e1g1 d5d7 d1a4 e8g8 c1e3 f8d8 f1e1 f7f6 a1c1 c6b4 a4d7 d8d7 a2a3 b4d5 e3c5 e7c5 c1c5 c7c6 e1c1 d5c7 c5a5 a7a6 g1f1 a8d8 f3d2 d7d4 b2b4 g8f7 f1e1 c7b5 e1f1');
INSERT INTO evaluations VALUES(50947,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(50947,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(50947,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(50947,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(50947,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(50947,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50947,36,'cp 34','c4d5 e6d5 g1f3 g8f6 c1g5 f8e7 e2e3 h7h6 g5h4 c8e6 f1d3 e8g8 e1g1 b8d7 h4g3 c7c5 d4c5 d7c5 a1c1 a8c8 c3e2 d8a5 d3b1 c5e4 g3e5 c8c1 e2c1 e7d6 e5d6 e4d6 c1b3 a5b6 d1d4 d6c4 d4b6 c4b6 b3c5 e6c8 b2b3');
INSERT INTO evaluations VALUES(50944,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50944,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(50944,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(50944,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(50944,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(50944,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50944,36,'cp 34','e6d5 g1f3 g8f6 c1g5 f8e7 e2e3 h7h6 g5h4 e8g8 f1d3 c8e6 h2h3 b8d7 e1g1 c7c6 h4g3 e7b4 a1c1 f8e8 g3h2 b4c3 c1c3 f6e4 c3c1 e6f5 d1b3 d8b6 b3b6 d7b6 b2b3 e4g3 d3f5 g3f5 f3e1 b6d7 a2a4 g7g5 a4a5 f7f6 e1d3 g8g7');
INSERT INTO evaluations VALUES(50991,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(50991,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50991,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(50991,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(50991,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(50991,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(50991,36,'cp 44','g1f3 g8f6 c1g5 f8e7 e2e3 e8g8 f1d3 h7h6 g5h4 c8e6 h2h3 c7c6 e1g1 b8d7 h4g3 e7b4 a1c1 f8e8 g3h2 d8e7 a2a3 b4c3 c1c3 f6e4 c3c1 e6f5 d1c2 a6a5 f3e5 f5h7 g1h1 a5a4 f2f3 d7e5 d4e5 e4g5 d3h7 g5h7 f3f4');
INSERT INTO evaluations VALUES(50989,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(50989,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(50989,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50989,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(50989,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(50989,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(50989,36,'cp 31','g8f6 c1g5 f8e7 e2e3 e8g8 f1d3 h7h6 g5h4 c8e6 e1g1 b8d7 h2h3 c7c6 h4g3 e7b4 a1c1 a8c8 g3h2 f8e8 d3b1 b4f8 f3d2 c6c5 d4c5 d7c5 h2e5 f6e4 e5d4 e6f5 c3e2 f8d6 f2f3 e4d2 d1d2 f5b1');
INSERT INTO evaluations VALUES(43678,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(43678,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(43678,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(43678,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(43678,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(43678,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(43678,36,'cp 49','c1g5 c8e6 e2e3 b8d7 f1d3 h7h6 g5f4 c7c5 d4c5 d7c5 a1c1 f8d6 f4d6 d8d6 e1g1 e8g8 d3b1 a8c8 h2h3 f8d8 d1d4 c5e4 f1d1 c8c4 d4c4 d5c4 d1d6 e4d6 f3d4 g7g6 f2f3 g8g7');
INSERT INTO evaluations VALUES(43677,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(43677,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(43677,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43677,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(43677,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(43677,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(43677,36,'cp 34','f8e7 e2e3 h7h6 g5h4 c7c6 f1d3 e8g8 e1g1 f6e4 h4e7 d8e7 d1c2 e4c3 c2c3 b8d7 a2a4 d7f6 f3d2 c8g4 f1c1 g4h5 c3c5 e7e6 c5c2 f6d7 d3f5 e6d6 c2b3 d6c7 c1c2 h5g6 f5g6 f7g6 h2h3 g8h7 a1c1 d7f6 d2f3 f6e4');
INSERT INTO evaluations VALUES(16028,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(16028,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16028,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(16028,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(16028,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(16028,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(16028,36,'cp -7','g1f3 e7e6 e2e4 h7h6 g5f6 d7f6 d1d3 c7c6 f1e2 f8e7 e1g1 e8g8 e4e5 f6d7 e2d1 f7f6 e5f6 e7f6 f1e1 d8c7 d3e3 e6e5 d4e5 f8e8 c3e4 d5e4 e5f6 d7f6 f3d4 c8d7 d1e2 a7a5 a2a4 a8d8 a1d1 d7g4 e3b3 g8h8 e2g4 f6g4 b3g3 c7g3 f2g3 g4f6 c2c3 c6c5');
INSERT INTO evaluations VALUES(16027,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(16027,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(16027,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(16027,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(16027,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(16027,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(16027,36,'cp -4','h7h6 g5f4 e7e6 c3b5 f8b4 c2c3 b4a5 e2e3 a7a6 b5a3 e8g8 f1e2 c7c6 e1g1 a5c7 f4c7 d8c7 c3c4 b7b5 c4b5 a6b5 a3c2 e6e5 c2b4 e5e4 f3e5 d7e5 d4e5 c7e5 b4c6 e5b2 a1b1 b2a2 c6b4 a2a4 b4d5 f6d5 d1d5 c8e6 d5b5 f8c8 b5a4 a8a4 b1b2 c8c6 f1b1 a4a8 e2h5 g7g6 b2b8 a8b8 b1b8 g8g7');
INSERT INTO evaluations VALUES(15996,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15996,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15996,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(15996,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(15996,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(15996,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(15996,36,'cp 0','g5f4 e7e6 c3b5 f8b4 c2c3 b4a5 e2e3 a7a6 b5a3 d8e7 f1d3 c7c5 a3b1 c5c4 d3c2 b7b5 a2a3 c8b7 b1d2 g7g5 f4g3 f6h5 g3e5 h5f6');
INSERT INTO evaluations VALUES(16564,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(16564,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(16564,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(16564,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(16564,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(16564,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(16564,36,'cp -52','d7d5 b1d2 d5d4 e3g5 h7h6 g5h4 f8c5 f1e2 a7a5 e1g1 c8e6 h4g3 f6d7 c2c3 a5a4 c3d4 c5d4 f3d4 c6d4 d2f3 d4f3 e2f3 c7c5 f3e2 e8g8 f2f4 f7f6 d1c2 d8e7 c2c3 e5f4 g3f4 f8d8 f4c7 d8e8 f1d1 b7b5 d3d4 b5b4 c3e3 c5d4');
INSERT INTO evaluations VALUES(15162,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(15162,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(15162,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(15162,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(15162,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(15162,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(15162,36,'cp -35','b1d2 c8e6 c2c3 f8e7 f1e2 a7a5 e1g1 e8g8 f3g5 e6c8 g5f3 h7h6 d1a4 e7d6 f1e1 c6e7 a4c2 f8e8 e2f1 c7c5 d3d4 e5d4 c3d4 d5e4 d4c5 e4f3 c5d6 e7d5 e3d4 f3g2 f1c4 d8d6 d2e4 f6e4 e1e4 e8e4 c2e4 d5f6 d4f6 d6f6 e4e8 g8h7 e8f7');
INSERT INTO evaluations VALUES(17049,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(17049,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(17049,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(17049,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(17049,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(17049,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(17049,36,'cp 96','c4e2 a5c6 e1g1 d7d6 d2d4 f8e7 d4d5 c6b8 a2a4 c7c6 f3d2 e8g8 d2c4 b8a6 b2b3 c6d5 e4d5 c8f5 e2d3 f5g6 c1a3 a6c7 f1e1 f8e8 d1d2 a8c8 d3f1 f6h5 d2e3 g6c2 e3a7 c2b3 a7b7 b3c4 f1c4 h5f6 c3b5 c7b5 b7b5');
INSERT INTO evaluations VALUES(26070,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(26070,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(26070,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(26070,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(26070,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(26070,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(26070,36,'cp 76','c1e3 g7g6 f1e2 f8g7 d1d2 c6d4 e3d4 g7d4 d2d4 g8f6 e1c1 d7c6 f2f4 d8b6 d4b6 a7b6 d1d4 h8g8 h1e1 e8f8 b2b4 g6g5 b4b5 c6e8 e4e5 d6e5 f4e5 f6d7 e2f3 d7c5 e1e3 e7e6 a2a4 f8e7 c1b2 a8a5 g2g3 g8h8 d4d6');
INSERT INTO evaluations VALUES(26069,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(26069,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(26069,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(26069,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(26069,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(26069,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(26069,36,'cp 67','g7g6 f2f3 h7h5 f1c4 g8f6 d1d2 c6d4 e3d4 f8h6 d2d3 e8g8 a1d1 e7e6 e1g1 d7c6 d4f2 d6d5 e4d5 e6d5 c3d5 f6d5 c4d5 d8d5 d3d5 c6d5 d1d5 f8e8 f1b1 b7b6 g2g3 e8e2 c2c3 a8e8 d5d7 a7a5 f2b6');
INSERT INTO evaluations VALUES(43253,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(43253,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(43253,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(43253,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43253,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(43253,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(43253,36,'cp 34','c2c4 e8g8 b1c3 d7d6 g1f3 c7c5 d4d5 e7e6 e1g1 f8e8 c1g5 h7h6 g5f6 d8f6 f3e1 e6e5 e2e4 a7a6 a2a4 h6h5 h2h4 g7h6 d1e2 f6e7 e1c2 c8g4 f2f3 g4c8 a4a5 b8d7 g2h3 a8b8 c3d1 b7b5 c2e3 b5c4 e3c4 b8b4 d1e3');
INSERT INTO evaluations VALUES(45649,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(45649,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(45649,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(45649,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(45649,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(45649,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(45649,36,'cp 32','f1g2 f8e7 e1g1 e8g8 b2b3 b7b6 c1b2 b8d7 f3d2 c8b7 b1c3 c6c5 d4c5 b6c5 c4d5 e6d5 d2c4 d7b6 c4a5 b7a6 a5c6 d8d6 g2d5 b6d5 c3d5 f6d5 d1d5 a8c8 c6e7 d6e7 f1e1 a6e2 a1c1 c8d8 d5e5 e7e5 b2e5 d8d2 a2a3 f8c8 b3b4 c5c4 e5c3 d2a2');
INSERT INTO evaluations VALUES(47956,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(47956,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(47956,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(47956,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47956,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47956,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(47956,36,'cp 16','f6e4 d1e2 d8e7 d3f4 e4f6 b1c3 e7e2 f1e2 c7c6 h2h4 f8e7 f2f3 h7h5 e1f2 a7a5 a2a4 c8f5 e2d1 b8a6 c3e2 a6b4 c2c3 b4d5 d2d4 e8d7 g2g3 d5b6 f4g2 f6d5 d1b3 g7g6 h1e1 f5e6 c1g5 e7g5 h4g5 d7c7 b3d5 b6d5 e2f4 h8e8 f4d5 e6d5 e1e8');
INSERT INTO evaluations VALUES(48017,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(48017,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48017,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(48017,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48017,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(48017,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(48017,36,'cp 10','d1e2 d8e7 d3f4 b8c6 f4d5 c6d4 d5e7 d4e2 e7d5 e2d4 b1a3 d4e6 f2f3 e4c5 d2d4 c7c6 d4c5 c6d5 c5d6 f8d6 c1e3 d6a3 b2a3 c8d7 e1c1 d7c6 a3a4 e8g8 a4a5 f8c8 d1d2 b7b6 e3f2 b6a5 f1d3 c8b8 h1e1 g7g6 h2h4 c6b5 d3b5 b8b5 a2a4 b5b4 d2d5');
INSERT INTO evaluations VALUES(48018,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(48018,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(48018,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48018,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(48018,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48018,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(48018,36,'cp 8','d8e7 d3f4 b8c6 f4d5 c6d4 d5e7 d4e2 e7d5 e2d4 b1a3 d4e6 f2f3 e4c5 d2d4 c7c6 d4c5 c6d5 c5d6 f8d6 c1e3 d6a3 b2a3 c8d7 e1c1 d7c6 f1d3 e8g8 h1e1 f8c8 d1d2 a7a6 c1b2 c6a4 e3f2 c8c6 e1e5 e6f4 d3f5 f7f6 e5e7 g7g6 f2e3 g6f5 e3f4');
INSERT INTO evaluations VALUES(33821,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(33821,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(33821,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33821,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(33821,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(33821,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(33821,36,'cp 11','d3f4 b8c6 f4d5 c6d4 d5e7 d4e2 e7d5 e2d4 b1a3 d4e6 f2f3 e4c5 d2d4 c7c6 d4c5 c6d5 c5d6 f8d6 c1e3 d6a3 b2a3 c8d7 e1c1 d7c6 f1d3 e8g8 h1e1 f8c8 d1d2 c6a4 e3f2 g7g6 e1e5 e6f4 d3f1 c8c7 f2g3 f7f6 e5e1 g6g5 g3f4 g5f4 f1d3 a8e8 e1e8 a4e8 c1b2 e8f7 d3f5');
INSERT INTO evaluations VALUES(19800,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(19800,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(19800,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(19800,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(19800,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(19800,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(19800,36,'cp -12','b2b3 a7a6 c1b2 e7e6 g2g3 g8f6 a2a4 f8e7 f1g2 e8g8 e1g1 b7b6 d1e2 f6d7 b1c3 c8b7 c3d1 d8c7 d1f2 a8d8 a1c1 d5d4 e3d4 e7f6 f3e5 c6d4 b2d4 c5d4 g2b7 c7b7 e5d7 d8d7 f2d3 b6b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(19798,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(19798,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(19798,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(19798,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(19798,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(19798,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(19798,36,'cp -27','a7a6 c1b2 e7e6 g2g3 g8f6 f1g2 f8e7 a2a4 b7b6 e1g1 c8b7 d1e2 e8g8 f3e5 a8c8 e5c6 b7c6 b1a3 f6d7 f4f5 e6e5 d2d3 e7f6 e3e4 d5d4 a3c4 b6b5 a4b5 a6b5 c4a5 f6g5 h2h4 g5e3 g1h2 d7f6 g2f3 c6d7 b2c1 e3c1 f1c1 c5c4 b3c4');
INSERT INTO evaluations VALUES(41097,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(41097,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41097,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41097,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41097,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(41097,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41097,36,'cp 14','c8g4 c2c3 e8g8 c1f4 e7d6 f4d6 d8d6 b1d2 b8d7 f1e1 c7c6 d1b3 d6c7 e1e3 a8e8 a1e1 e8e3 e1e3 g7g6 b3a3 a7a5 h2h3 g4f3 d2f3 f8e8 e3e8 f6e8 a3e7 d7f6 e7e2 a5a4 g2g3 a4a3 b2b4 b7b5 g1g2 e8d6 e2e5 c7d8 e5f4 g8g7');
INSERT INTO evaluations VALUES(28405,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(28405,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(28405,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(28405,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(28405,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(28405,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28405,36,'cp 25','c1f4 e8g8 b1d2 e7d6 f4g3 d6g3 h2g3 b8d7 f1e1 f8e8 e1e8 f6e8 c2c3 c7c6 d1c2 g7g6 a1e1 d8b6 d2b3 g4f3 g2f3 e8f6 g1g2 a8e8 c2d2 e8e1 d2e1 b6d8 a2a4 h7h5 e1e3 g8g7 a4a5 f6h7 f3f4 h7f6 a5a6 b7a6 d3a6');
INSERT INTO evaluations VALUES(43199,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43199,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43199,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43199,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43199,36,'cp 0','d7d6 a2a4 e8g8 g1f3 b8d7 h2h3 b7b6 f1c4 c8b7 e1g1 a7a6 f1e1 d6d5 c4f1 c7c5 c3b1 f8e8 b1d2 c5c4 c2c3 b6b5 b2b3 b7c6 a4b5 a6b5 d1c2 e7e6 b3c4 b5c4 a1a8 d8a8 c2b2 a8b7 b2b7 c6b7 e1a1 e8a8 a1a8 b7a8');
INSERT INTO evaluations VALUES(41303,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41303,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41303,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(41303,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41303,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41303,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41303,36,'cp 0','a2a4 b8d7 g1f3 e8g8 h2h3 b7b6 f1c4 a7a6 e1g1 c8b7 f1e1 d6d5 c4f1 c7c5 c3b1 e7e6 b1d2 d8e7 c2c3 f8c8 d1c2 b7c6 f4h2 c6b7 h2f4');
INSERT INTO evaluations VALUES(55639,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(55639,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(55639,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(55639,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(55639,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(55639,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(55639,36,'cp 70','d2d4 b8d7 e2e4 g8f6 f1c4 f8e7 e1g1 c7c6 a2a4 e8g8 h2h3 f6e4 c3e4 d6d5 f3e5 d5e4 c1f4 d7b6 c4b3 b6d5 f4h2 e7d6 f1e1 d8g5 e1e4 c8h3 d1f3 h3e6 e4e2 a8e8 a1e1 d5f6 h2g3 d6c7 b3e6 e8e6 c2c3 h7h6');
INSERT INTO evaluations VALUES(55574,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(55574,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(55574,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(55574,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(55574,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(55574,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(55574,36,'cp 61','b8d7 e2e4 g8f6 a2a4 f8e7 f1c4 e8g8 e1g1 h7h6 h2h3 f8e8 f1e1 c7c6 a4a5 d8c7 c4f1 d7f8 d4d5 c8d7 f3d2 c6d5 e4d5 e8b8 d2c4 f8g6 c1e3 g6h4 b2b4 h4f5 b4b5 f5e3 e1e3 e7f8 b5b6 c7d8 b6a7 a8a7');
INSERT INTO evaluations VALUES(56496,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(56496,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(56496,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(56496,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(56496,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(56496,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(56496,36,'cp -63','f8d6 g1f3 c7c6 c2c4 h7h6 b1c3 g8e7 b2b4 b8d7 c1b2 e8g8 d1c2 f8e8 f1e2 a7a6 e1g1 d7f6 c3b1 e7g6 b1d2 c8e6 f1e1 d8e7 b2c3 f6d7 e2f1 e6g4 c2b3 a8c8 a1c1 g4e6 a3a4 f7f5 h2h3 g8h8 b3b2 b7b5 c4d5');
INSERT INTO evaluations VALUES(56523,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(56523,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(56523,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(56523,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(56523,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(56523,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56523,36,'cp 0','f8b4 c1d2 d4e3 f2e3 d8h4 g2g3 h4e4 g1f3 b4d2 b1d2 e4e3 d1e2 e3e2 e1e2 c8g4 e2d3 g8e7 d3c3 b8d7 a1e1 a8d8 f3d4 e8g8 f1g2 b7b6 b2b4 c7c5 b4c5 d7c5 d2b3 c5a4 c3b4 a4b2 b4c3');
INSERT INTO evaluations VALUES(38965,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(38965,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38965,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38965,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38965,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(38965,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38965,36,'cp 0','c1d2 d4e3 f2e3 d8h4 g2g3 h4e4 g1f3 b4d2 b1d2 e4e3 d1e2 e3e2 e1e2 g8e7 e2d3 c8g4 d3c3 b8d7 a1e1 a8d8 f3d4 e8g8 f1g2 b7b6 b2b4 c7c5 b4c5 d7c5 d2b3 c5a4 c3b4 a4c5');
INSERT INTO evaluations VALUES(38964,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(38964,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(38964,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(38964,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38964,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38964,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38964,36,'cp 0','d4e3 f2e3 d8h4 g2g3 h4e4 g1f3 b4d2 b1d2 e4e3 d1e2 e3e2 e1e2 c8g4 e2d3 b8d7 d3c3 g8e7 a1e1 e8g8 f1g2 a8d8 f3d4 b7b6 b2b4 c7c5 b4c5 d7c5 d2b3 c5a4 c3b4 a4b2 b4c3');
INSERT INTO evaluations VALUES(38963,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38963,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38963,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38963,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38963,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38963,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38963,36,'cp 0','f2e3 d8h4 g2g3 h4e4 g1f3 b4d2 b1d2 e4e3 d1e2 e3e2 e1e2 a7a5 e2d3 g8e7 d3c3 b8c6 f1d3 c8f5 d3f5 e7f5 d2b3 e8c8 a1e1 b7b6 g3g4 f5e7 f3g5 d8d7 g5f7 c6b4 b3d4 h8e8 e5e6 b4a2 c3d3 a2b4 d3c3');
INSERT INTO evaluations VALUES(38962,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38962,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(38962,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38962,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38962,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(38962,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(38962,36,'cp -19','d8h4 g2g3 h4e4 g1f3 b4d2 b1d2 e4e3 d1e2 e3e2 e1e2 b8c6 e2d3 g8e7 d3c3 a7a5 f1d3 e8g8 d2b3 e7g6 h1e1 c8g4 b3d4 c6b4 d3f5 g4f5 d4f5 f8e8 e1e3 b7b5 b2b3 b5c4 b3c4 g6f8 a2a3 b4a6 a1b1 a6c5 b1b5 f8d7 f5d4 h7h6 d4b3 c5b3 c3b3 e8d8');
INSERT INTO evaluations VALUES(33137,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(33137,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(33137,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(33137,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(33137,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(33137,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(33137,36,'cp -4','g2g3 h4e4 g1f3 b4d2 b1d2 e4e3 d1e2 e3e2 e1e2 b8c6 e2d3 a7a5 d3c3 g8e7 f1d3 e8g8 d2b3 e7g6 h1e1 c8g4 b3d4 c6b4 d3f5 g4f5 d4f5 f8d8 e1e3 d8d7 a2a3 g6f8 a1e1 b4a6 e3d3 d7d3 c3d3 a6c5 d3c3 f8e6 f3d4 g8f8 d4e6 f7e6 f5d4 a5a4 e1f1 f8e7 d4b5 a8c8 g3g4 b7b6 f1e1');
INSERT INTO evaluations VALUES(33136,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(33136,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(33136,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(33136,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(33136,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(33136,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(33136,36,'cp -8','h4e4 g1f3 b4d2 b1d2 e4e3 d1e2 e3e2 e1e2 a7a5 e2d3 b8c6 d3c3 g8e7 f1d3 e8g8 d2b3 e7g6 h1e1 c8g4 b3d4 c6b4 d3f5 g4f5 d4f5 f8d8 e1e3 d8d7 a2a3 g6f8 a1e1 b4a6 e3d3 d7d3 c3d3 a6c5 d3c3 f8e6 e1d1 g8f8 f3d4 e6d4 f5d4 a8e8 d4b5 c7c6 b5d6 e8e5 b2b4');
INSERT INTO evaluations VALUES(11310,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(11310,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(11310,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(11310,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(11310,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(11310,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(11310,36,'cp 38','f1d3 c6b4 a2a3 b4d3 c2d3 e7e6 g1f3 f8d6 c3e2 e8g8 d1c2 c8d7 e1g1 a8c8 f1c1 d7c6 h2h3 f6h5 f4h2 h5f6 g1h1 h7h6 f3e5 c6e8 b2b4 f6d7 f2f4 d7e5 f4e5 d6e7 h2f4 f7f6 e5f6 f8f6');
INSERT INTO evaluations VALUES(11314,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(11314,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(11314,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(11314,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(11314,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(11314,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(11314,36,'cp 18','c6b4 a2a3 b4d3 c2d3 e7e6 a1c1 f8e7 c3a4 e7d6 f4g3 c8d7 g1f3 e8g8 a4c5 d7c8 e1g1 h7h6 c5a4 c8d7 f1e1 d6g3 h2g3 d7a4 d1a4 d8e7 c1c2 f8c8 b2b4 c7c6 e1c1 c8b8 g1h1 f6e8 a4b3 e8d6 a3a4 f7f6 a4a5 e7d7');
INSERT INTO evaluations VALUES(13229,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(13229,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(13229,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(13229,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(13229,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(13229,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(13229,36,'cp -72','e7e6 d2d3 h7h6 g5f3 a7a6 c4b3 b7b5 e1g1 c8b7 a2a4 b5b4 c3e2 d6d5 e4d5 f6d5 f1e1 f8e7 e2f4 d5f4 c1f4 e8g8 f3e5 c6d4 c2c3 d4b3 d1b3 a6a5 a1d1 e7d6 f4g3 d6e5 e1e5 d8b6 d3d4 c5d4 d1d4 b6c6 f2f3 f8d8');
INSERT INTO evaluations VALUES(13916,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(13916,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(13916,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(13916,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(13916,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(13916,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(13916,36,'cp -69','d2d3 h7h6 g5f3 a7a6 e1g1 b7b5 c4b3 c8b7 c3e2 d6d5 e4d5 f6d5 c2c4 d5b4 d3d4 c6a5 a2a3 a5b3 d1b3 b5c4 b3c4 d8d5 c4c3 c5d4 e2d4 a8c8 c3d2 b4c6 d4c6 b7c6 d2d4 d5d4 f3d4 c6e4 f2f3 e4d5 b2b4 f8e7 c1e3 e8g8 f1c1 e7f6 e3f2');
INSERT INTO evaluations VALUES(13917,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(13917,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(13917,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(13917,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(13917,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(13917,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(13917,36,'cp -67','h7h6 g5f3 a7a6 e1g1 d6d5 c4b3 b7b5 a2a3 f8e7 b3a2 e8g8 c1f4 d5e4 c3e4 f6d5 f4d2 d8b6 d1e2 f8d8 f1e1 b6c7 a1c1 a6a5 c2c3 b5b4 a2c4 c8b7 e4g3 d5f4 e2e3 e7d6 d3d4 c5d4 c3d4 c7b6 h2h3');
INSERT INTO evaluations VALUES(13441,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(13441,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(13441,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(13441,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(13441,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(13441,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(13441,36,'cp -53','g5f3 a7a6 c4b3 b7b5 e1g1 d6d5 a2a3 f8e7 h2h3 d5e4 c3e4 f6e4 d3e4 c5c4 b3a2 d8c7 d1e2 e8g8 c1e3 f8d8 f1d1 d8d1 a1d1 e6e5 c2c3 c8e6 a2b1 a8d8 b1c2 d8d1 e2d1 c6a5 a3a4 b5b4 c3b4 e7b4');
INSERT INTO evaluations VALUES(13442,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(13442,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(13442,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(13442,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(13442,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(13442,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(13442,36,'cp -75','a7a6 e1g1 b7b5 c4b3 c8b7 c3e2 d6d5 e4d5 f6d5 c2c3 f8e7 d3d4 d5f6 c1e3 c5d4 e2d4 c6a5 f3e5 a5b3 d1b3 d8d6 d4f3 e8g8 c3c4 b7f3 e5f3 b5c4 b3c4 f8c8 c4e2 f6d5 f1d1 d6c6 g2g3 c6b5 e2b5 a6b5 a1c1 d5e3 c1c8 a8c8 f2e3');
INSERT INTO evaluations VALUES(22693,30,'cp -161',NULL);
INSERT INTO evaluations VALUES(22693,31,'cp -155',NULL);
INSERT INTO evaluations VALUES(22693,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(22693,33,'cp -157',NULL);
INSERT INTO evaluations VALUES(22693,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(22693,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(22693,36,'cp -124','g8f6 h5e2 d7d5 d2d3 f8c5 g1f3 c8e6 h2h3 h7h6 b2b4 c5b6 e2c2 a7a6 f1e2 g7g5 f3h2 h8g8 b1d2 h6h5 a2a4 d5d4 a4a5 b6a7 c3c4 c6b4 c2a4 b4c6 a1b1 a7c5 d2f3 d8d6 f3g5 c5b4 e1f1 e8c8 g5e6 f7e6 g2g3 h5h4 f1g2 f6d7 h2f3 d7c5 a4a2 h4g3 f2g3 b4a5 h3h4');
INSERT INTO evaluations VALUES(16376,30,'cp -134',NULL);
INSERT INTO evaluations VALUES(16376,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(16376,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(16376,33,'cp -134',NULL);
INSERT INTO evaluations VALUES(16376,34,'cp -145',NULL);
INSERT INTO evaluations VALUES(16376,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(16376,36,'cp -125','h5e2 d7d5 d2d3 c8e6 g1f3 f8c5 h2h3 c5b6 e2c2 h7h6 b2b4 a7a6 f1e2 g7g5 f3h2 h8g8 b1d2 h6h5 a2a4 d8d7 a1b1 e8c8 b4b5 a6b5 a4b5 c6a5 e4d5 d7d5 d2e4 g5g4 h3g4 h5g4 h2f1 f6e4 d3e4 d5d6 c1e3 b6e3 f1e3 g4g3 b1d1 d6c5 d1d8 g8d8');
INSERT INTO evaluations VALUES(22747,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22747,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22747,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22747,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22747,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22747,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22747,36,'cp 0','f8b4 b1c3 g8f6 g1f3 e5d4 e3d4 d7d5 f1e2 d5c4 e2c4 e8g8 e1g1 c8g4 c1e3 d8d6 a1c1 a8e8 h2h3 g4h5 c3b5 d6d7 b5c3');
INSERT INTO evaluations VALUES(6631,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6631,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6631,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6631,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6631,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6631,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6631,36,'cp 0','b1c3 g8f6 g1f3 e5d4 e3d4 d7d5 f1e2 e8g8 e1g1 d5c4 e2c4 c8g4 c1e3 d8d6 a1c1 a8e8 h2h3 g4h5 c3b5 d6d7 b5c3');
INSERT INTO evaluations VALUES(34182,30,'cp 220',NULL);
INSERT INTO evaluations VALUES(34182,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(34182,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(34182,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(34182,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(34182,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(34182,36,'cp 234','d2d4 d7d6 c1e3 c8g4 d1d2 g4f3 g2f3 f6f3 h1g1 g8f6 c4e2 f3h3 e1c1 h3c8 f2f4 b8d7 f4e5 d6e5 d4d5 c8c7 a2a3 e8c8 d5c6 b7c6 e2c4 f8c5 d2e2 g7g6 c4f7 c5e3 e2e3 g6g5 c3a4 c8b7 f7e6 h8e8 e6f5 b7b8 e3b3 b8a8 b3b4 d8b8');
INSERT INTO evaluations VALUES(34184,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(34184,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(34184,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(34184,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(34184,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(34184,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(34184,36,'cp 193','d7d6 c1e3 c8g4 c4e2 b8d7 d1d2 f6d8 h2h3 g4f3 g2f3 f8e7 h1g1 e7g5 e1c1 d8e7 f3f4 e5f4 e3f4 g5f4 d2f4 g7g5 f4f5 e8c8 f2f4 g8f6 g1f1 g5f4 f5f4 d8e8 e2d3 a7a6 c1b1 c8b8 f1f3 f6h7 h3h4 h7f6 d1f1 h6h5 a2a4 h8g8 a4a5 b8a8 f4e3 a8b8');
INSERT INTO evaluations VALUES(36534,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36534,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36534,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36534,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36534,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36534,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36534,36,'cp 31','b1c3 f6h5 f4g5 h7h6 g5h4 f7f5 f3d2 h5f6 f2f4 c7c5 d4d5 d8e8 f1e2 e7e5 c3b5 e5f4 e3f4 e8d7 e1g1 b8a6 f1e1 a6c7 b5c3 d7f7 e2d3 f8e8 d1c2 f6h5 g2g3 c8d7 d2f3 e8e1 a1e1 a8e8 g1g2 e8e1 f3e1 h5f6 h2h3');
INSERT INTO evaluations VALUES(50226,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50226,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(50226,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(50226,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(50226,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50226,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(50226,36,'cp 14','d7d5 g1f3 f8g7 c2c4 c7c6 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 e1g1 b6b2 c3d5 f6d5 g2d5 b8c6 d5c6 g7e5 f4e5 b7c6 f1e1 f7f6 e5c7 f8c8 c7a5 a8b8 e2e4 f5h3 d1d3 c6c5 a5c3 b2b5 d3b5 b8b5 d4d5 g8f7 f2f3 e7e6 a1d1 e6d5 d1d5');
INSERT INTO evaluations VALUES(46803,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46803,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(46803,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46803,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46803,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(46803,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(46803,36,'cp 17','g1f3 f8g7 c2c4 c7c6 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 e1g1 b6b2 c3d5 f6d5 g2d5 b8c6 d5c6 g7e5 f4e5 b7c6 f1e1 f7f6 e5c7 a8c8 c7a5 c8b8 a2a3 g8f7 e2e4 f5h3 a5b4 f8d8 b4c5 d8d7 e4e5 b2b3 e5f6 b3d1 e1d1 e7f6 a1b1 b8b1');
INSERT INTO evaluations VALUES(46802,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(46802,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46802,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46802,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46802,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46802,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46802,36,'cp 7','f8g7 c2c4 c7c6 c4d5 c6d5 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 e8g8 e1g1 d8e7 h2h4 a8c8 f1c1 f8d8 c1c2 c8c6 a1c1 d8c8 a2a3 e7d8 e2e3 h7h5 c3e2 c6c2 c1c2 c8c2 d2c2 d7c6 c2b3 d8d7 e2f4 c6b5 e5f6 g7f6');
INSERT INTO evaluations VALUES(1469,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1469,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1469,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1469,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1469,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1469,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1469,36,'cp 0','d4c5 f8c5 d1d2 c5e7 f1d3 f7f6 d2e2 f6e5 f3e5 g8f6 e1g1 c6e5 e2e5 e7d6 e5g5 e8g8 c3b5 d6c5 c1f4 c8d7 f4c7 b6c6 c7e5 c6b6');
INSERT INTO evaluations VALUES(14769,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(14769,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(14769,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(14769,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14769,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(14769,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14769,36,'cp 28','f6e4 e2e3 g7g5 c3e4 g5f4 e4c3 f4e3 f2e3 f8h6 d1d2 c8d7 c4d5 e6d5 f1b5 c6e7 b5d3 c7c6 e1g1 e7g6 a1e1 f7f5 c3e2 d8f6 b2b4 e8c8 e2f4 h8e8 a2a4 c8b8 g1h1 h6f8 e1b1 f6f7 f1e1 f8d6 b4b5 d8c8 g2g3 f7f6 f4h5 f6f7 b5b6 a7a6 a4a5 f7e7 h5f4');
INSERT INTO evaluations VALUES(51857,30,'cp 221',NULL);
INSERT INTO evaluations VALUES(51857,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(51857,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(51857,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(51857,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(51857,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(51857,36,'cp 219','d4e5 d7d5 g1f3 f8c5 e4d5 c6d5 b1c3 g8e7 c3a4 b8c6 a4c5 d8a5 c2c3 a5c5 c1e3 c5a5 h2h3 e8g8 f1d3 f8e8 e1g1 c8f5 d3f5 e7f5 e3f4 c6d8 d1c2 f5e7 c2b3 a5c7 f3d4 a7a6 f1e1 d8e6 f4e3 c7e5 b3b7 e5d6 b7b4 d6b4 c3b4 e6d4 e3d4 e7c6 e1e8 a8e8');
INSERT INTO evaluations VALUES(51828,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(51828,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(51828,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(51828,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(51828,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(51828,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(51828,36,'cp 200','d7d5 g1f3 f8c5 e4d5 c6d5 b1c3 g8e7 c3a4 b8c6 a4c5 d8a5 c2c3 a5c5 h2h3 d5d4 f3d4 c5e5 d1e2 e5f6 d4c6 f6c6 c1g5 c8e6 e2b5 f7f6 g5e3 e8c8 b5c6 e7c6 f1e2 e6f5 b2b3 c6e5 e1g1 f5d3 a1e1 d3e2 e1e2 e5c6 f1e1');
INSERT INTO evaluations VALUES(52598,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(52598,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(52598,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(52598,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(52598,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(52598,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52598,36,'cp 37','f8g7 d2d4 c5d4 f3d4 g8f6 f2f3 d8b6 c4b5 c8d7 c1e3 b6c7 b5e2 b8c6 g2g4 e8g8 d1d2 a7a6 e1c1 b7b5 c3d5 f6d5 e4d5 c6d4 e3d4 g7d4 d2d4 a8c8 d4d2 f7f5 h2h4 f5g4 f3g4 f8f2 d1f1 f2g2 h1g1 g2g1 f1g1 c7c5 g1e1 c8f8 h4h5 c5f2 e2d3 d7g4 h5g6 f2d2 c1d2 h7g6 d3g6 g8g7 g6d3 g7f7 e1e4 g4f3 e4f4 f7g7 f4f8');
INSERT INTO evaluations VALUES(18466,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(18466,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(18466,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18466,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18466,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18466,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(18466,36,'cp 8','g7g6 b2b3 g8f6 f1d3 f8g7 c1b2 e8g8 e1g1 d6d5 f1e1 e7e5 d1d2 f8e8 a1d1 c8g4 f2f3 g4e6 g1h1 d5d4 c3a4 g7f8 b2c1 a7a5 c2c3 c6c5 d3b5 e6d7 b5c4 d8c7 a4b2 d7e6 c4f1 f8g7 d2c2 f6d7 c3d4 e5d4 f1b5 e8b8 a2a4 d7b6 f3f4 e6g4');
INSERT INTO evaluations VALUES(18445,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(18445,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(18445,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(18445,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18445,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(18445,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18445,36,'cp 0','h2h3 g8f6 f1c4 f8g7 e4e5 d6e5 d1d8 e8d8 c1d2 d8e8 e1c1 h7h5 h1e1 f6d7 f2f4 e5f4 d2f4 g7c3 b2c3 d7f8 f4c7 f8e6 c7e5 h8h7 h3h4 e6d8 d1d2 c8e6 c4a6 e6c8');
INSERT INTO evaluations VALUES(32794,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(32794,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(32794,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(32794,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(32794,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(32794,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(32794,36,'cp 111','h2h3 g4f3 d1f3 b8d7 c1d2 g8f6 e1c1 e7e6 f1c4 a5b6 h1e1 f8e7 f3d3 e8g8 f2f4 e7b4 a2a3 b4c3 d3c3 f6d5 c3d3 b6c7 f4f5 b7b5 c4b3 e6f5 d3f5 d7b6 f5d3 f8e8 c1b1 c7d6 e1e2 e8e2 d3e2 a7a5 d1e1 h7h6 g2g4 d6d7 h3h4 d5f6 g4g5');
INSERT INTO evaluations VALUES(32792,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(32792,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(32792,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(32792,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(32792,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(32792,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(32792,36,'cp 138','g4f3 d1f3 e7e6 c1d2 b8d7 e1c1 f8b4 f1c4 g8e7 a2a3 b4c3 d2c3 a5g5 c1b1 b7b5 c4d3 e7d5 c3d2 g5h4 f3g4 h4g4 h3g4 d7b6 b2b3 e8c8 g4g5 d5e7 b1b2 c8b7 d2b4 e7f5 c2c3 b6d5 b4c5 f5d6 d3e2 h7h6 g5h6 h8h6 h1h6 g7h6');
INSERT INTO evaluations VALUES(35281,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(35281,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(35281,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(35281,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(35281,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(35281,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(35281,36,'cp 111','g1f3 e5d4 f1e2 b8c6 e1g1 g8f6 f1e1 d7d6 c1f4 e7d8 f3d4 f8e7 d4c6 b7c6 e2f3 e8g8 e4e5 f6d5 f3d5 c6d5 e5d6 e7d6 d1d5 c8e6 d5e4 d6c5 b1d2 d8f6 c2c3 c5b6 d2f3 h7h6 a2a4 a7a5 f4e5 f6f5 e4f5 e6f5 e5d4 c7c5 d4e5 f7f6 e5g3');
INSERT INTO evaluations VALUES(35280,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(35280,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(35280,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(35280,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(35280,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(35280,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(35280,36,'cp 89','e5d4 f1e2 b8c6 e1g1 g8f6 f1e1 d7d6 e2b5 f6d7 a2a4 g7g6 c2c3 d4d3 c1g5 f7f6 g5e3 f8g7 b5d3 d7e5 f3e5 f6e5 d3c4 c8e6 c4e6 e7e6 d1d3 a7a6 b2b4 c6d8 b1d2 e8g8 a4a5 e6f7 d3d5 a8b8 h2h4 d8e6 a1d1');
INSERT INTO evaluations VALUES(35300,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(35300,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(35300,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(35300,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(35300,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(35300,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(35300,36,'cp 107','f1e2 g8f6 e1g1 b8c6 f1e1 d7d6 e2b5 f6d7 c1f4 e7f6 d1d2 f8e7 f4g5 f6e6 g5e7 e6e7 b5c6 b7c6 f3d4 c8b7 b1c3 e8g8 a1d1 e7f6 f2f3 d7e5 b2b3 a7a5 c3a4 f8e8 e1f1 e5g6 c2c4 h7h6 f1f2 f6g5 f3f4 g5d8 a4c3 c6c5');
INSERT INTO evaluations VALUES(44319,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(44319,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44319,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(44319,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(44319,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(44319,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(44319,36,'cp 30','f1e2 b8c6 d2d3 f8d6 g1f3 d6e7 e1g1 a7a6 a2a4 e8g8 f3e1 d7d6 f2f4 e5f4 c1f4 c8e6 e1c2 f6d7 a1b1 d7e5 b2b4 c5b4 c3d5 e7g5 f4g3 g5h4 d3d4 e5g6 c2b4 h4g3 h2g3 d8g5 g3g4 c6b4 b1b4');
INSERT INTO evaluations VALUES(52241,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(52241,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(52241,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(52241,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(52241,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(52241,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(52241,36,'cp 194','b1c3 e7e6 d2d4 c5d4 f3d4 a7a6 f1c4 b7b5 c4b3 c8b7 c1e3 b8c6 f2f4 c6d4 d1d4 g8e7 f4f5 e7c6 d4d1 d8e7 d1h5 g7g6 f5g6 e8c8 e1c1 e7g7 c1b1 h7g6 h5e2 f8b4 h2h4 b4c3 b2c3 c6e5 e3b6 b7c6 h4h5 h8h5 h1h5 g6h5 b6d8');
INSERT INTO evaluations VALUES(52240,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(52240,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(52240,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(52240,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(52240,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(52240,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(52240,36,'cp 185','e7e6 d2d4 c5d4 f3d4 a7a6 f1c4 b8c6 d1h5 g7g6 d4c6 d7c6 h5f3 g8h6 c4b3 h6f7 c1e3 f8e7 h2h4 e8g8 f3g3 c6c5 a2a4 g8h8 g3h3 d8d6 a4a5 a8b8 a1d1 d6c6 e1g1 h8g8 h4h5 b7b5 a5b6 b8b6 h5g6 h7g6');
INSERT INTO evaluations VALUES(54763,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(54763,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(54763,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(54763,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(54763,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(54763,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(54763,36,'cp 69','f1e2 e7e6 g1f3 c7c6 c2c4 g8f6 e1g1 f8d6 b2b3 e8g8 c1a3 d6a3 b1a3 d8e7 a3b1 c8d7 d1c2 d7e8 b1d2 e8g6 c4d5 e6d5 f3e5 b8d7 d2f3 a8e8 e2d3 f6g4 e5d7 e7d7 b3b4 g8h8 a2a4 d7c7 c2b2');
INSERT INTO evaluations VALUES(3722,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(3722,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3722,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(3722,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(3722,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(3722,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3722,36,'cp 21','e4d5 f6d5 a2a4 d5b6 c4b5 a7a5 f1e1 c5d6 c1e3 b6d5 b1d2 g8h8 d1b3 d5e3 e1e3 f7f6 d2e4 c8g4 h2h3 g4h5 b5c6 b7c6 d3d4 h5f3 e3f3 f6f5 e4d6 c7d6 d4e5 d6e5 a1d1 d8e8 f3d3 e5e4 d3d6 a8b8 b3c2 f5f4 d1e1 e4e3 f2e3 f4e3 c2e2 h8g8 e2e3 e8e3 e1e3');
INSERT INTO evaluations VALUES(3714,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(3714,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3714,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(3714,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(3714,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(3714,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3714,36,'cp 20','f6d5 a2a4 d5b6 c4b5 a7a5 f1e1 c5d6 c1e3 b6d5 b1d2 g8h8 d1b3 d5e3 e1e3 f7f6 d2e4 c8g4 h2h3 g4h5 b5c6 b7c6 d3d4 h5f3 e3f3 f6f5 e4d6 c7d6 d4e5 d6e5 a1d1 d8e8 b3c4 e5e4 f3e3 e8g6 f2f3 g6g5 c4e2 a8e8 f3e4 f5e4');
INSERT INTO evaluations VALUES(4147,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(4147,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(4147,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(4147,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(4147,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(4147,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(4147,36,'cp -33','g5f3 d7d6 h2h3 c6a5 c4b3 a5b3 a2b3 f8e8 c3a4 c5d4 e1g1 a7a5 a4c3 d4b6 c1e3 b6e3 f2e3 c7c6 d1e1 c8e6 e1h4 b7b6 h4f2 b6b5 f2g3 e6d7 g1h2 g8h7 f3h4 d7c8 h2g1 e8f8 g3f3 f6e8 f3g3 c8e6 g3f2 e8f6 h4f3');
INSERT INTO evaluations VALUES(4172,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(4172,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(4172,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(4172,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(4172,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(4172,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(4172,36,'cp -32','d7d6 h2h3 c6a5 c4b3 a5b3 a2b3 f8e8 c1e3 c5e3 f2e3 c7c6 e1g1 a7a5 d1d2 c8e6 d2f2 b7b5 f2g3 a8a7 f3h4 f6h5 g3e1 g8h7 h4f3 h5f6 e1f2 d8b6 g1h1 h7h8 g2g4 b6d8 c3e2 e8g8 e2g3 c6c5 d3d4 c5d4 e3d4 b5b4');
INSERT INTO evaluations VALUES(3610,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(3610,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(3610,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(3610,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(3610,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(3610,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(3610,36,'cp -27','h2h3 c6a5 c4b3 a5b3 a2b3 a7a5 e1g1 f8e8 c1e3 c5e3 f2e3 c8e6 d1e1 c7c6 f3h4 b7b5 e1g3 f6h5 g3f2 a8b8 h4f3 h5f6 f2g3 b8b7 g3e1 b7a7 e1g3 g8h8 f3h4 f6h5 g3f2 a7b7 h4f3 h5f6 c3e2 b7a7 e2g3 a7a8 f3h4 h8h7 h4f5 f6g8 d3d4');
INSERT INTO evaluations VALUES(6735,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(6735,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(6735,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(6735,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(6735,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(6735,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(6735,36,'cp 48','b2b4 a5b6 b4b5 c6d4 f3e5 g8f6 e5c4 d7d5 c4b6 c7b6 d2d3 c8g4 f2f3 a8c8 c1e3 d5e4 d3e4 d4f3 g2f3 c8c3 d1d8 e8d8 e1c1 d8e8 d1d3 c3d3 c2d3 g4f3 h1g1 f6g4 e3f4 f7f5 h2h3 g4f6 g1g7 f5e4 g7b7 h8g8 b7b8 e8d7 b8g8 f6g8 h3h4 g8f6 d3e4 f6e4 c1c2 e4f6 f4b8 d7c8 b8g3 c8d7 f1h3 d7e7');
INSERT INTO evaluations VALUES(6725,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(6725,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(6725,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(6725,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(6725,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(6725,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(6725,36,'cp 42','a5b6 b4b5 c6d4 f3e5 g8f6 e5c4 d7d5 c4b6 a7b6 d2d3 e8g8 c1e3 c7c5 b5c6 d4c6 e3g5 c8e6 e4d5 e6d5 g5f6 d8f6 c3d5 f6e5 f1e2 e5d5 e1g1 d5c5 a3a4 c5c3 e2f3 a8a5 g2g3 g7g6 a1b1 c6d4 b1b6 d4f3 d1f3 c3c2 b6b4 f8e8 f3b7 a5a4 b4a4 c2a4 b7d5 a4a8 d5a8 e8a8 g1g2 a8a2 f1d1');
INSERT INTO evaluations VALUES(6315,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(6315,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(6315,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(6315,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(6315,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(6315,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(6315,36,'cp 48','b4b5 c6d4 f3e5 g8f6 e5c4 d7d5 c4b6 a7b6 d2d3 e8g8 c1e3 c7c5 b5c6 d4c6 e3g5 c8e6 e4d5 e6d5 g5f6 d8f6 c3d5 f6e5 f1e2 e5d5 e1g1 c6d4 c2c3 d4e2 d1e2 d5b3 c3c4 f8e8 e2d2 e8d8 f1d1 b6b5 c4b5 b3b5 a1b1 b5d5 d2b4 b7b5 h2h3 g7g6 d3d4 a8a4 b4e7 a4d4 d1d4 d5d4 b1b5 d4d1 g1h2 d1d6 e7d6 d8d6');
INSERT INTO evaluations VALUES(15076,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(15076,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(15076,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(15076,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(15076,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(15076,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(15076,36,'cp -31','f8b4 f3d2 d5e4 d3e4 b4c3 b2c3 h7h6 g5e3 d8e7 f2f3 b7b6 f1b5 c8d7 e1g1 e8g8 a2a4 c6a5 d1e2 a5b7 d2c4 f8d8 b5d7 f6d7 e3f2 f7f6 a4a5 d7f8 a5b6 a7b6 c4e3 e7f7 a1a8 d8a8 e3d5 b7d6 f1d1 a8a2 d5b4');
INSERT INTO evaluations VALUES(46706,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(46706,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46706,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(46706,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(46706,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(46706,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(46706,36,'cp 30','b8c6 g1f3 f8g7 d4d5 c6b4 d3e2 e8g8 e1g1 e7e6 a2a3 b4a6 e2a6 b7a6 c1e3 e6d5 e4d5 c8d7 d1d2 f8e8 f1e1 c7c5 d5c6 d7c6 f3d4 c6b7 h2h3 d8c7 e3h6 g7h6 d2h6 e8e5 a1d1 c7b6 h6d2 a8e8 d4b3 a6a5 e1e5 d6e5 d2d6');
INSERT INTO evaluations VALUES(49320,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(49320,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(49320,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(49320,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(49320,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(49320,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(49320,36,'cp 47','d2d4 d7d6 d1e2 f8e7 e2g4 g7g6 g1f3 c7c5 e1g1 c5d4 g4d4 e8g8 c1h6 b8c6 d4d2 f8e8 e5d6 d8d6 b1c3 d5f6 d2d6 e7d6 f1d1 d6f8 h6f8 g8f8 c4b5 c8d7 c3e4 f6e4 d1d7 e8c8 b5c6 c8c6 f3e5 c6c5 e5f7 c5c2 d7b7 a8c8 a1e1 c2f2 f7g5 c8c2 g5e6 f8g8 b7g7 g8h8 e1e4');
INSERT INTO evaluations VALUES(49321,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(49321,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(49321,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(49321,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49321,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(49321,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(49321,36,'cp 25','d7d6 d1e2 f8e7 e2g4 g7g6 g1f3 c7c5 e1g1 c5d4 g4d4 e8g8 c1h6 b8c6 d4d3 f8e8 e5d6 d8d6 b1c3 d5f6 d3d6 e7d6 f1d1 d6e7 c3b5 e8d8 d1d8 c6d8 b5c7 a8b8 h6f4 f6h5 f4e3 b7b6 c7b5 a7a6 b5c7 b6b5 c4e2 c8b7 a2a4 b5a4 a1a4 b7f3 e2f3');
INSERT INTO evaluations VALUES(48028,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(48028,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(48028,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(48028,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(48028,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(48028,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(48028,36,'cp 36','g1f3 c7c5 e1g1 c5d4 c4d5 e6d5 c1g5 f8e7 g5e7 d8e7 e5d6 e7d6 f1e1 c8e6 b1a3 e8g8 d1d2 a7a6 f3d4 b8d7 c2c3 d7f6 a3c2 f8e8 a2a4 e6d7 a4a5 e8e1 a1e1 b7b6 a5b6 d6b6 c2e3 a8e8 e1a1 f6e4 d2e2 e4c5 e2c2 c5e6 d4e6 e8e6');
INSERT INTO evaluations VALUES(48027,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(48027,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(48027,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(48027,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(48027,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(48027,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(48027,36,'cp 36','c7c5 e1g1 c5d4 c4d5 e6d5 c1g5 f8e7 g5e7 d8e7 e5d6 e7d6 f1e1 c8e6 d1d3 e8g8 f3d4 b8c6 b1d2 f8e8 c2c3 e6d7 d2f3 a7a6 h2h3 c6d4 d3d4 d7c6 a2a4 h7h6 a4a5 c6b5 f3h4 d6d7 a1d1 a8d8 e1e8 d7e8 h4f3 f7f6');
INSERT INTO evaluations VALUES(57439,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(57439,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(57439,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(57439,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(57439,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(57439,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(57439,36,'cp 108','f1c4 b8c6 g1e2 f8c5 b1c3 d7d6 f3h5 g7g6 h5f3 h7h5 h2h3 e8f8 c3a4 f8g7 a4c5 d6c5 f3c3 d8d6 d2d3 c6d4 c3d2 d4e2 d2e2 g8e7 e1g1 g6g5 h3h4 g5h4 f2f4 c8g4 e2f2 e7g6 f4f5 g6f4 c1f4 e5f4 f2f4 d6f4 f1f4 h4h3 g1h2 h3g2 h2g2 a8e8');
INSERT INTO evaluations VALUES(57403,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(57403,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(57403,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(57403,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(57403,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(57403,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(57403,36,'cp 127','b8c6 g1e2 g8e7 a2a3 d7d6 f3b3 d8d7 c4f7 e8d8 b1c3 g7g6 d2d3 f8g7 h2h4 h8f8 h4h5 g6g5 h5h6 g7h8 f7h5 d7e6 c3d5 e6g8 c1d2 f6f5 e1c1 b7b6 d5e7 d8e7 b3g8 f8g8 e4f5 g5g4 d1e1 c8f5 d3d4 e7d7 d4d5 c6e7 e2g3 a8f8 g3f5 f8f5 f2f3 f5h5 h1h5 g4f3 g2f3 e7d5');
INSERT INTO evaluations VALUES(23221,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(23221,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(23221,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(23221,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(23221,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(23221,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(23221,36,'cp 121','g1e2 d7d6 d2d4 g7g6 e1g1 f8g7 b1c3 g8e7 a2a4 h8f8 f1d1 f6f5 c1g5 f5e4 f3e4 c8f5 e4e3 d8d7 c3d5 e7d5 c4d5 h7h6 g5h4 g6g5 d4e5 g7e5 f2f4 g5h4 f4e5 e8c8 e5e6 d7e7 d1f1 d8e8 d5c6 b7c6 e2d4 f5e6 f1f8 e7f8 d4e6');
INSERT INTO evaluations VALUES(37760,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(37760,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(37760,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(37760,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37760,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(37760,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(37760,36,'cp 39','h2h4 h7h5 d1f3 e7e6 d2d4 c5d4 c3b5 d7d5 e4d5 e6d5 f3d5 d8e7 c4e2 g8f6 d5d6 e7d6 b5d6 e8e7 d6c8 h8c8 g1f3 f6g4 e2d3 b8c6 e1g1 g4e5 f3e5 c6e5 d3e4 c8c7 c1f4 e7f8 g2g3 f7f5 e4d3 c7c5 f4e5 c5e5 a2a4 b7b6 b2b4 a8c8 a4a5 g7f6 a5b6 a7b6 a1a7');
INSERT INTO evaluations VALUES(42507,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(42507,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(42507,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(42507,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(42507,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(42507,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(42507,36,'cp 34','d1d2 d7d5 e2e3 d8e7 c4d5 e6d5 f1d3 b7b6 e1g1 c7c5 b2b3 e8g8 h2h3 f8d8 b1c3 b8c6 d2b2 c6b4 d3e2 c8a6 e2a6 b4a6 f1d1 a8c8 d4c5 b6c5 a1b1 h7h6 a2a3 c8b8 c3e2 e7b7 b2c2 f6e4 e2f4');
INSERT INTO evaluations VALUES(42506,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(42506,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(42506,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(42506,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(42506,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42506,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(42506,36,'cp 34','d7d5 e2e3 d8e7 c4d5 e6d5 f1d3 b7b6 e1g1 c7c5 b1c3 e8g8 h2h3 b8c6 b2b3 f8d8 d2b2 f6e4 d3b5 c8b7 a1c1 a8c8 c3e2 e7f6 f1d1 a7a6 b5d3 a6a5 a2a3 c6e7 c1c2 e4d6 f3e5 e7g6 e5g6 h7g6 d4c5 b6c5');
INSERT INTO evaluations VALUES(57219,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(57219,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(57219,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(57219,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(57219,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(57219,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(57219,36,'cp -17','b8c6 g1f3 d7d5 d2d4 e5e4 f3d2 c6e7 c2c4 c7c6 b1c3 f7f5 f1e2 g8f6 e1g1 a7a6 c4d5 e7d5 d2c4 c8e6 a2a4 f8e7 c3d5 f6d5 c1d2 e8g8 f2f3 e4f3 e2f3 d5b6 c4e5 e7d6 d2a5 d6e5 d4e5 d8d1 f3d1 b6c4 d1b3 a8e8 a5b4 f8f7 b3c4 e6c4 f1d1 c4d5 b4d6');
INSERT INTO evaluations VALUES(22871,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(22871,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(22871,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(22871,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(22871,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(22871,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(22871,36,'cp -14','d2d4 d7d5 g1f3 e5e4 f3d2 c6e7 c2c4 c7c6 b1c3 g8f6 c4d5 c6d5 f1b5 c8d7 d1b3 d7c6 f2f3 e4f3 d2f3 a7a6 b5d3 e7g6 e3e4 d5e4 f3g5 d8c7 d3e4 f8d6 e4c6 b7c6 e1g1 e8g8 g5e4 f6e4 c3e4 c7e7 e4d6 e7d6 c1e3 a8b8 b3c2 d6d5 a1e1 h7h5 b2b3 f8e8 e3f2 e8e1 f1e1');
INSERT INTO evaluations VALUES(17349,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(17349,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(17349,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(17349,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(17349,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(17349,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(17349,36,'cp -34','b7b5 c4d5 g8e7 f1e1 c8b7 d5c6 e7c6 d2d4 c5d4 f3d4 c6e5 d1h5 f8d6 c1f4 e8g8 a1d1 d8b8 d4b3 e5f3 g2f3 d6f4 d1d7 b7c6 d7d3 b5b4 c3e2 f4e5 e2d4 c6a4 b3c5 a4e8 f3f4 e5f4 e4e5 b8c7 h5f3 f4e5 f3a8 c7c5 a8e4 e5f6 e4e3 a6a5');
INSERT INTO evaluations VALUES(18109,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(18109,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(18109,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(18109,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(18109,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(18109,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(18109,36,'cp -35','c4d5 g8e7 f1e1 c8b7 d5c6 e7c6 d2d4 c5d4 f3d4 c6e5 c1f4 d8f6 f4g3 f8e7 d1e2 e5c4 e4e5 f6g6 a2a4 b5b4 e2c4 b4c3 c4c3 a8c8 c3d3 g6d3 c2d3 b7d5 e1b1 c8b8 f2f3 e8g8 g3e1 e7c5 e1f2 f8c8 g1f1 f7f6 e5f6 g7f6 d4e2 d5b3 b1c1 c5f2 c1c8 b8c8 a1c1 c8b8 f1f2 b3a4 c1a1 b8b4');
INSERT INTO evaluations VALUES(25750,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(25750,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(25750,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(25750,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(25750,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(25750,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(25750,36,'cp 196','e1g1 g4f3 d1f3 d6e5 f3c3 d8d6 c1e3 g8f6 f2f3 f6d7 b1d2 a6a5 d2b3 a5a4 a1d1 d6e6 b3a5 c6c5 d1d3 f8d6 a5c4 e8g8 f1d1 a8a6 c4d6 a6d6 d3d6 c7d6 c3d3 d7b6 b2b3 a4b3 c2b3 b6c8 a2a4 c8e7 d3c4 e6c4');
INSERT INTO evaluations VALUES(25749,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(25749,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(25749,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(25749,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(25749,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(25749,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(25749,36,'cp 191','g4f3 d1f3 d6e5 f3c3 d8d6 c1e3 g8f6 f2f3 a6a5 b1d2 f6d7 d2b3 a5a4 a1d1 d6e6 b3a5 f8d6 a5c6 e8g8 c6a5 h7h5 b2b4 a4b3 a2b3 f8b8 a5c6 b8f8 h2h3 f7f5 e4f5 f8f5 f1e1 a8e8 e3f2 e6f7 c3d3 d7f8 b3b4 f8g6 d3e4 f5f4');
INSERT INTO evaluations VALUES(40928,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(40928,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(40928,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(40928,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(40928,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(40928,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(40928,36,'cp -29','e8g8 f1e2 c7c5 e1g1 b7b6 c3b1 b8d7 b1d2 c8b7 c2c3 a7a6 a2a4 c5c4 d1c2 b6b5 b2b3 d8c7 g5f4 e7d6 f4d6 c7d6 a1a2 d6c7 f1a1 b7c6 a4b5 a6b5 c2b2 h7h6 a2a8 f8a8 a1a8 c6a8 b2a3 a8c6 h2h3 d7b6 f3e5 c6e8 e2f3 g7g5 b3c4 b5c4');
INSERT INTO evaluations VALUES(36303,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(36303,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(36303,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(36303,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(36303,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(36303,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(36303,36,'cp -17','f1e2 c7c5 e1g1 b8c6 d4c5 e7c5 c3b5 a7a6 b5d4 c8d7 d4c6 d7c6 f3d4 a8c8 a1c1 c6a4 b2b3 a4d7 c2c4 d5c4 c1c4 e6e5 d4f3 d7e6 c4c1 h7h6 g5h4 d8a5 h4f6 g7f6 e2c4 f8d8 d1e2 c5e7 c4e6 c8c1 f1c1');
INSERT INTO evaluations VALUES(55170,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(55170,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(55170,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(55170,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(55170,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(55170,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(55170,36,'cp -51','e7e6 g1f3 g8e7 f1d3 b8c6 d3e4 d8d7 e1g1 b7b6 c2c4 d4d3 b2b4 c8b7 b4b5 c6a5 e4b7 a5b7 b1c3 e7g6 g2g3 f8c5 g1g2 a7a6 b5a6 b7a5 d1a4 d7a4 c3a4 c5e7 a4b2 a8a6 a1b1 f7f6 e5f6 g7f6 c4c5 b6c5 b2d3 a5c6 c1b2');
INSERT INTO evaluations VALUES(55984,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55984,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(55984,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(55984,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(55984,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(55984,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55984,36,'cp -32','g1f3 g8e7 f1d3 b8c6 d3e4 d8d7 e1g1 b7b6 c2c4 c8b7 b2b4 d4d3 c1b2 e7g6 a2a3 f8e7 g2g3 f7f5 e5f6 g7f6 f3d4 c6d4 e4b7 a8d8 g1h1 c7c6 b2d4 d7b7 d4e3 h7h5 h2h4 b7d7 b1c3 f6f5 c4c5 b6c5 b4c5 e8f7 d1f3 e6e5');
INSERT INTO evaluations VALUES(55983,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(55983,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(55983,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(55983,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(55983,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(55983,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(55983,36,'cp -26','g8e7 f1c4 a7a6 d1e2 b8c6 a2a3 b7b5 c4d3 c8b7 d3e4 e7d5 d2d3 d8d7 e1g1 e8c8 c2c4 d4c3 b1c3 f8c5 g1h1 c8b8 f1d1 c6a5 e4d5 b7d5 c3d5 d7d5 b2b4 a5b3 a1b1 c5b6 c1e3 b6e3 e2e3 b3d4 f3d4 d5d4 e3f3 f7f5 e5f6 g7f6 f3e2 e6e5 f4e5 f6e5');
INSERT INTO evaluations VALUES(55334,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(55334,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(55334,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(55334,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(55334,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(55334,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(55334,36,'cp 161','d2d4 b8d7 f1c4 h7h6 e1g1 f8g7 d4e5 d6e5 d1e2 c7c6 f1d1 d8a5 c1d2 a5d8 a2a4 g8f6 a4a5 d8e7 b2b3 e8g8 d2c1 f8d8 c1a3 e7e8 a3b4 f6h5 e2d2 h5f4 h2h3 g8h7 d2e3 b7b5 a5b6 d7b6 c4f1 d8d1 a1d1 c8e6 b4a3 a8c8 a3d6');
INSERT INTO evaluations VALUES(55332,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(55332,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(55332,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(55332,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(55332,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(55332,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(55332,36,'cp 158','b8d7 f1c4 h7h6 e1g1 g8f6 a2a4 f8g7 d4e5 d6e5 d1e2 c7c6 f1d1 d8a5 c1d2 a5d8 a4a5 e8g8 h2h3 d8e7 b2b3 f6h5 d2c1 h5f4 e2e3 f8d8 c1a3 e7e8 a3d6 b7b5 a5b6 d7b6 c4f1 b6d5 e4d5 d8d6 d5c6 d6d1 a1d1 e8c6 f3e5 c6f6 e5c4 f4h3 g2h3 f6c3 e3e8 g8h7 d1d8');
INSERT INTO evaluations VALUES(57703,30,'cp 281',NULL);
INSERT INTO evaluations VALUES(57703,31,'cp 277',NULL);
INSERT INTO evaluations VALUES(57703,32,'cp 276',NULL);
INSERT INTO evaluations VALUES(57703,33,'cp 289',NULL);
INSERT INTO evaluations VALUES(57703,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(57703,35,'cp 283',NULL);
INSERT INTO evaluations VALUES(57703,36,'cp 283','c1g5 f8g7 e2e3 c7c5 c2c3 d8b6 b1d2 d7d5 g1e2 g8f6 d1c2 e8g8 d4c5 b6c5 e2g3 e7e5 f1d3 h7h6 g5f6 g7f6 a1d1 b8c6 d2b3 c5d6 e1g1 f6g7 c2e2 f7f5 c3c4 f5f4 c4d5 c6b4 e3f4 b4d3 e2d3 e5f4 g3h5 c8g4');
INSERT INTO evaluations VALUES(57699,30,'cp 265',NULL);
INSERT INTO evaluations VALUES(57699,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(57699,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(57699,33,'cp 261',NULL);
INSERT INTO evaluations VALUES(57699,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(57699,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(57699,36,'cp 262','c7c5 e2e3 f8g7 c2c3 d7d5 b1d2 d8b6 g1e2 g8f6 g5f6 g7f6 e2f4 e7e6 a1b1 b6d6 f4h5 b8d7 f1e2 e8g8 e1g1 f6h8 g1h1 d6e7 b1a1 c5d4 e3d4 d7f6 f2f4 c8d7 h5f6 h8f6 a2a3');
INSERT INTO evaluations VALUES(6988,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(6988,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(6988,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(6988,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6988,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(6988,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(6988,36,'cp 9','c6a5 c4e2 d4c3 e1g1 d7d6 b1c3 a5c6 d1a4 g8f6 e4e5 d6e5 f3e5 d8d4 e5c6 d4a4 c3a4 b7c6 e2f3 f6d5 c1d2 c8d7 a1c1 a8b8 b2b3 e7a3 c1c4 e8g8 f1d1 f8d8 a4c5 d7f5 d2c1 d5b6 d1d8 b8d8 c1a3 b6c4 b3c4 d8d4');
INSERT INTO evaluations VALUES(14926,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(14926,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(14926,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(14926,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(14926,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(14926,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(14926,36,'cp 89','f1d3 c6b4 e4f6 g7f6 d3c4 b7b6 e1g1 c8b7 a2a3 b4d5 f1e1 d8d7 d1e2 e8c8 c4b5 c7c6 b5d3 c6c5 d4c5 f8c5 d3b5 d7c7 b5a6 h8g8 a6b7 c7b7 g2g3 d5e7 b2b4 c5d6 c2c4 d6e5 f3e5 f6e5 e2e5 e7c6 e5e4 c6d4 e4b7 c8b7 c1f4 d4c2');
INSERT INTO evaluations VALUES(14924,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(14924,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(14924,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(14924,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(14924,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(14924,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(14924,36,'cp 84','c6b4 e4f6 g7f6 d3e2 b7b6 a2a3 b4c6 e1g1 c6e7 f1e1 f8g7 a3a4 c8b7 a4a5 e8g8 a5a6 b7e4 f3d2 e4c6 c2c3 d8d5 d2f3 a8d8 c1f4 d8d7 h2h3 d5f5 f4g3 f8d8 b2b4 c6e4 d1c1 c7c6 a1a3 c6c5 b4c5 b6c5 a3a5');
INSERT INTO evaluations VALUES(35072,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(35072,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(35072,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(35072,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(35072,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(35072,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(35072,36,'cp 105','e2e3 c7c5 g1f3 c5d4 f4g5 f6f5 e3d4 b8c6 c2c3 f5a5 g5f4 g8f6 b1d2 d7d5 h2h4 a5d8 f1d3 f8d6 f3e5 d8c7 d1e2 f6d7 d2f3 d6e5 f3e5 d7e5 e1g1 e8g8 d4e5 f7f6 f1e1 c6e5 f4e5 c7e5 e2e5 f6e5 e1e5 c8d7 g2g3 g7g6 f2f4 g8g7 g1g2 a7a5 a2a3 a5a4');
INSERT INTO evaluations VALUES(35071,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(35071,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(35071,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(35071,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(35071,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(35071,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(35071,36,'cp 97','c7c5 g1f3 c5d4 f4g5 f6f5 e3d4 b8c6 c2c3 f5a5 g5f4 g8f6 b1d2 d7d5 h2h4 a5d8 f1d3 f8d6 f3e5 d8c7 d1e2 f6d7 d2f3 d6e5 f3e5 d7e5 d4e5 f7f6 h4h5 c6e5 e1g1 c8d7 f1e1 e8g8 f4e5 f6e5 e2e5 a8c8 e5c7 c8c7 e1e2 f8c8 f2f4 h7h6');
INSERT INTO evaluations VALUES(34227,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(34227,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(34227,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(34227,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(34227,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(34227,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(34227,36,'cp 92','g1f3 c5d4 f4g5 f6f5 e3d4 g8f6 b1d2 f5a5 f1d3 f8e7 e1g1 d7d6 f1e1 e8g8 d2e4 b8d7 d1e2 h7h6 g5h4 a5d8 c2c4 f6h5 h4e7 d8e7 e2d2 b7b6 e4c3 e7d8 d3f1 h5f6 h2h3 c8b7 d4d5 e6e5 b2b4 f6e8 a2a3 f7f5 a1d1 d7f6');
INSERT INTO evaluations VALUES(19233,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(19233,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(19233,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(19233,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19233,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(19233,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19233,36,'cp 0','f8g7 c2c3 c6a5 c4b3 a5b3 a2b3 g8f6 b3b4 c5b4 c3b4 e8g8 b1c3 a7a6 e1g1 d7d6 h2h3 b7b5 f1e1 c8b7 f3d4 a8c8 d4b3 d8d7 b3a5 f6h5 c1d2 b7a8 a5b3 a8b7');
INSERT INTO evaluations VALUES(19699,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(19699,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(19699,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(19699,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(19699,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(19699,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(19699,36,'cp 54','d2d3 f8g7 a2a4 a7a5 h2h3 b7b6 e1g1 g8f6 f1e1 e8g8 d1e2 d8c7 e4e5 f6d5 c1d2 c8e6 b1a3 a8d8 f3h4 e6c8 a1d1 d5b4 f2f4 f7f5 d2e3 b4d5 h4f3 c8e6 f3g5 e6d7 e2d2 g7h6 h3h4 h6g5 h4g5');
INSERT INTO evaluations VALUES(19697,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(19697,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(19697,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(19697,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(19697,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(19697,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(19697,36,'cp 47','f8g7 a2a4 a7a5 e1g1 g8f6 c1d2 e8g8 b1a3 f6h5 d2c3 f7f6 a3c4 h5f4 d1d2 g7h6 c3a5 d8d7 g1h1 b7b5 c4b6 d7g4 f3e1 f4g2 f2f4 a8a5 e1g2 a5a6 a4a5 e7e5 d2f2 e5f4 b6c8 g4c8 g2f4 c8g4 f2g2 g4g2 h1g2 h6f4 f1f4 g8g7 b2b3 f8d8');
INSERT INTO evaluations VALUES(5550,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(5550,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(5550,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(5550,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(5550,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(5550,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(5550,36,'cp 42','h2h3 b7b6 e1g1 g8f6 b1c3 d8c7 e4e5 f6d5 c3d5 c6d5 d3d4 c8f5 d4c5 c7c5 c2c3 f5e4 c1f4 e4f3 d1f3 e8g8 a1d1 e7e6 h3h4 h7h6 d1d4 b6b5 f1d1 a7a5 a2a3 c5c7 f3e2 c7c6 h4h5 a8c8 f4g3 g6g5 f2f4 g5f4');
INSERT INTO evaluations VALUES(5548,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(5548,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(5548,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(5548,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(5548,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(5548,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(5548,36,'cp 40','d8c7 a2a4 a7a5 b1a3 b7b6 e1g1 g8f6 c1d2 e8g8 e4e5 f6d5 d1e2 h7h6 a1e1 c8d7 e2e4 d7f5 e4h4 g6g5 h4g3 c7d7 h3h4 f7f6 a3c4 b6b5 c4a5 d5b4 e5e6 f5e6 a5b7 d7b7 e1e6 b4c2 h4g5 h6g5 d2g5 f6g5 g3g5');
INSERT INTO evaluations VALUES(35083,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(35083,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(35083,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(35083,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(35083,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(35083,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(35083,36,'cp 178','d2d4 f6d8 b1c3 d7d6 f1d3 a7a6 c1f4 g8e7 h2h4 h7h6 f4g3 b8d7 d1e2 g7g6 e4e5 d6d5 c3d1 c7c5 c2c3 e7c6 h4h5 g6g5 d4c5 d7c5 d3c2 c5e4 c2e4 d5e4 e2e4 d8d5 e4d5 e6d5 d1e3 c8e6');
INSERT INTO evaluations VALUES(36217,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36217,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36217,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(36217,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(36217,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(36217,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36217,36,'cp 30','e1g1 c7c5 d2d3 b8c6 a2a4 c8e6 f1e1 a7a6 c1g5 c6b4 d1e2 a8c8 g5f6 e7f6 c3d1 b4c6 d1e3 c6d4 e2d1 b7b5 c4d5 d4f3 d1f3 f6g5 d5e6 f7e6 f3e2 d8d7 a4b5 a6b5 c2c3 g5e3 e2e3 d7f7');
INSERT INTO evaluations VALUES(36216,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(36216,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36216,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36216,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36216,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36216,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36216,36,'cp 12','c7c5 f1e1 c8e6 d2d3 b8c6 c1g5 a7a6 a2a4 c6b4 g5d2 d8c7 f3h2 a8d8 h2f1 d6d5 e4d5 f6d5 c3d5 e6d5 c4d5 d8d5 f1e3 d5d8 d1g4 f8e8 d2c3 e7f8 h3h4 c7d7 g4d7 d8d7 h4h5 e5e4 d3e4 e8e4 a1b1');
INSERT INTO evaluations VALUES(47113,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(47113,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(47113,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(47113,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(47113,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(47113,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(47113,36,'cp 85','g1f3 c7c6 d5c6 b8c6 e2e3 e7e5 d2d3 e5e4 d3e4 d8d1 e1d1 f6e4 d1e1 f8e7 b1c3 e4c3 b2c3 e8g8 f3d4 c8d7 a1b1 a8b8 c1d2 f8c8 f2f3 a7a6 f1d3 c6e5 d3e4 b7b5 e1e2 e5c4 e4d5 c4a3 b1c1 a3c4 c1c2 c4d2 e2d2');
INSERT INTO evaluations VALUES(12526,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12526,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12526,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12526,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(12526,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(12526,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12526,36,'cp 37','f8e7 c1g5 d7d6 g5f6 e7f6 c3d5 e8g8 a2a4 c8e6 c2c3 f6g5 f3g5 d8g5 e1g1 a8b8 d1f3 g5d8 d5e3 c6e7 f3h5 h7h6 a4a5 d8c7 h5d1 b8d8 d1b3 c7d7 f1e1 a7a6 b3a4 d7c7 a4a2 c7d7 f2f3 e6c4 d3c4 f7f5 e4f5');
INSERT INTO evaluations VALUES(7939,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(7939,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(7939,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(7939,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(7939,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(7939,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(7939,36,'cp 37','c1g5 d7d6 g5f6 e7f6 c3d5 e8g8 a2a4 c8e6 c2c3 f6g5 f3g5 d8g5 d5e3 a8b8 d1f3 g5d8 h2h4 g8h8 f3g3 a7a6 e1g1 h7h6 c4d5 c6e7 f2f4 e6d5 e4d5 e7g6 e3f5 d8f6 f4e5 d6e5 f5h6 g6f4 h6g4 f6g6 f1f4 e5f4 g3f4 g6d3');
INSERT INTO evaluations VALUES(7935,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(7935,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(7935,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(7935,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(7935,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(7935,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(7935,36,'cp 28','d7d6 g5f6 e7f6 c3d5 e8g8 a2a4 c8e6 c2c3 f6g5 f3g5 d8g5 e1g1 a8b8 d1f3 g5d8 d5e3 c6e7 f3h5 h7h6 a4a5 d8c7 h2h3 b8d8 f1e1 c7d7 e1d1 d7c7 h5f3 c7d7 d1d2 d6d5 e4d5 e7d5 e3d5 e6d5 f3d5 d7d5');
INSERT INTO evaluations VALUES(8129,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(8129,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(8129,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(8129,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(8129,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(8129,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(8129,36,'cp 31','g5f6 e7f6 a2a4 c6b4 c3d5 b4d5 c4d5 e8g8 a4a5 a8b8 e1g1 c8e6 d5e6 f7e6 d1e2 d8c7 c2c3 h7h6 f1b1 b8c8 c3c4 f6d8 b2b4 c5b4 b1b4 c7d7 b4b2 b7b6 a5a6 d8e7 h2h3 f8f7 e2d1 g8h7 d1b3 d7c7 a1b1 f7f8');
INSERT INTO evaluations VALUES(17927,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(17927,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(17927,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(17927,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(17927,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(17927,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(17927,36,'cp 126','b5a4 c7c6 d2d3 f8c5 b1d2 g8e7 f2f4 d7d5 f4f5 a6a5 d1h5 b7b5 a4b3 a5a4 f5f6 g7f6 e4d5 e7d5 d2e4 c5e7 b3d5 d8d5 e4f6 e7f6 h5d5 c6d5 f1f6 h8g8 c1d2 h7h5 d2f4 c8e6 a1e1 e8d7 a2a3 a8c8 e1e2 c8e8 h2h3 g8g6 f4e5 g6f6 e5f6 e6f5 e2f2 e8e1 g1h2 f5e6 f6d4');
INSERT INTO evaluations VALUES(17941,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(17941,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(17941,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(17941,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(17941,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(17941,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(17941,36,'cp 135','f8c5 d2d3 c7c6 b1d2 g8e7 f2f4 e8g8 f4f5 f7f6 d1h5 a6a5 a4b3 d7d5 a2a4 b7b6 g2g4 g8h8 h5h3 a8a7 g1h1 d8e8 d2f3 d5e4 d3e4 e7g8 c1d2 c5b4 c2c3 d4c3 b2c3 b4c5 e4e5 f6e5 a1e1 e5e4 f3g5 g8f6 e1e4 a7e7 f1e1 e7e4 e1e4 e8d8 d2f4 h7h6');
INSERT INTO evaluations VALUES(34363,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(34363,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(34363,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(34363,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(34363,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(34363,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(34363,36,'cp 26','b1c3 d5e6 f1e2 b8c6 g1f3 c6d4 f3d4 c5d4 c3b5 e6d7 b5a3 d7d8 c2c3 c8d7 a3c2 d4c3 d2c3 e7e6 c1e3 g8f6 e2f3 f8e7 d1e2 e8g8 e3d4 d8a5 e1c1 f8d8 a2a3 d7b5 e2e5 a8c8 h1e1 a5a6 f3e2 b5e2 e5e2 a6e2 e1e2 g8f8');
INSERT INTO evaluations VALUES(34361,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(34361,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(34361,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(34361,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(34361,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(34361,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(34361,36,'cp 12','d5d8 g1f3 b8c6 f1b5 c8d7 f3e5 c6e5 f4e5 g8h6 d1f3 d7b5 c3b5 d8d7 a2a4 a7a6 b5a3 d7d4 d2d3 d4e5 e1d1 g7g6 f3b7 e5h5 d1e1 h5e5 b7e4 e5e4 d3e4 a8d8 c1h6 f8h6 a1d1 f7f5 e4f5 g6f5 d1d8 e8d8 a3c4 e7e6 e1f2 d8e7 h1d1 h8d8 d1d8 e7d8 c2c3 d8e7');
INSERT INTO evaluations VALUES(53222,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(53222,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(53222,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53222,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(53222,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(53222,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(53222,36,'cp 19','b2b3 g7g6 f1b5 c8d7 d1e2 f8g7 g1f3 g8h6 b5d7 b8d7 c1b2 h6f5 e1c1 e8g8 h1e1 a8c8 c1b1 e7e6 c3e4 f5d4 e2f2 d4f3 f2f3 g7b2 b1b2 d7f6 f3c3 f6e4 e1e4 d8d5 c3e5 d5e5 f4e5 c8c7 d2d3 b7b6 g2g4 g6g5 a2a4 c7d7 b2c3 a7a5 d1f1 f8d8');
INSERT INTO evaluations VALUES(53015,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(53015,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(53015,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(53015,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(53015,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(53015,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(53015,36,'cp -3','g8f6 d2d3 d6d5 e4d5 f6d5 d1e2 b8c6 g1f3 g7g6 f3e5 f8g7 e5c6 b7c6 e1g1 a7a5 g1h1 a5a4 b1d2 e8g8 a1b1 a4a3 c4b3 a3b2 c1b2 g7d4 d2c4 c8e6 e2f3 d8c7 a2a4 d5c3 b2c3 d4c3 c4e5 c3e5 f4e5 e6b3 c2b3 c7e5 f3c6');
INSERT INTO evaluations VALUES(45154,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45154,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(45154,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45154,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(45154,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45154,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45154,36,'cp -10','d2d3 e7e6 g1f3 f8e7 b1c3 d6d5 c4b5 c8d7 b5d7 d8d7 e1g1 e8g8 e4e5 f6g4 c3e2 b8c6 c2c3 f7f6 d3d4 c5d4 c3d4 g4h6 e5f6 e7f6 c1e3 h6f5 e3f2 f6e7 h2h3 f5d6 f3e5 c6e5 d4e5 d6e4 d1d3 e4f2');
INSERT INTO evaluations VALUES(45152,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(45152,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(45152,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45152,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(45152,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(45152,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(45152,36,'cp -7','e7e6 g1f3 d6d5 c4b5 b8d7 b1c3 a7a6 b5d7 c8d7 e1g1 d8c7 f3e5 d5e4 e5d7 c7d7 d3e4 f8e7 d1f3 d7c6 c1e3 a8d8 a2a4 e8g8 g2g4 c5c4 g4g5 f6e8 f4f5 e7c5 a1e1 c5e3 e1e3 c6c5 g1h1 d8d2 f1f2 d2f2 f3f2');
INSERT INTO evaluations VALUES(54991,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54991,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54991,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(54991,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(54991,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(54991,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54991,36,'cp 16','d2d4 g8f6 c2c4 g7g6 b1c3 f8g7 c1b2 e8g8 g1f3 b8a6 g2g3 e7e5 f1g2 e5e4 f3d2 e4e3 f2e3 c7c6 e1g1 d8e7 f1f3 d6d5 h2h3 f6e4 d2e4 d5e4 f3f4 f7f5 d1c1 f8f7 g3g4 g7h6 b2a3 a6b4 d4d5 h6f4');
INSERT INTO evaluations VALUES(1504,30,'cp -155',NULL);
INSERT INTO evaluations VALUES(1504,31,'cp -159',NULL);
INSERT INTO evaluations VALUES(1504,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(1504,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(1504,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(1504,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(1504,36,'cp -126','f8b4 c1d2 c6d4 f1b5 d4b5 d2b4 a7a5 b4d2 b5d4 e1g1 g8e7 d2e3 d4f3 d1f3 b6c7 e3f4 c8d7 b1d2 e8g8 f1c1 d7c6 f3c3 f8c8 d2f3 a5a4 h2h3 e7g6 f4g3 c7d8 c3e3 h7h6 f3d4 c6d7 f2f4 d8a5 c1c8 a8c8 a1c1 c8c1 e3c1 g6e7 g3f2 e7c6 d4c6 d7c6 c1e3 h6h5');
INSERT INTO evaluations VALUES(137,30,'cp -136',NULL);
INSERT INTO evaluations VALUES(137,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(137,32,'cp -136',NULL);
INSERT INTO evaluations VALUES(137,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(137,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(137,35,'cp -141',NULL);
INSERT INTO evaluations VALUES(137,36,'cp -128','c1d2 c6d4 f1b5 d4b5 d2b4 a7a5 b4d2 b5d4 e1g1 d4f3 d1f3 g8e7 d2e3 b6c7 e3f4 c8d7 b1d2 e8g8 a1c1 d7c6 f3c3 e7g6 f4g3 c7b6 d2f3 f8c8 f3d4 c6d7 c3e3 a5a4 f1d1 b6b4 f2f3 h7h6 h2h4 g6e7 g3e1 b4a3 e3d2 a4b3 a2b3 e7c6 d4c6 d7c6 d2e3 c6b5 e1f2 c8c1');
INSERT INTO evaluations VALUES(136,30,'cp -126',NULL);
INSERT INTO evaluations VALUES(136,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(136,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(136,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(136,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(136,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(136,36,'cp -126','c6d4 f1b5 d4b5 d2b4 a7a5 b4d2 b5d4 e1g1 d4f3 d1f3 g8e7 d2e3 b6c7 e3f4 c8d7 b1d2 e8g8 a1c1 d7c6 f3c3 e7g6 f4g3 f8c8 d2f3 c7b6 f3d4 g6e7 c3e3 c6b5 f1d1 e7c6 g3f4 c6d4 c1c8 a8c8 d1d4 b6c5 d4d1 c5b4 h2h3 b5d7 d1d4');
INSERT INTO evaluations VALUES(123,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(123,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(123,32,'cp -136',NULL);
INSERT INTO evaluations VALUES(123,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(123,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(123,35,'cp -130',NULL);
INSERT INTO evaluations VALUES(123,36,'cp -136','f1b5 d4b5 d2b4 a7a5 b4d2 b5d4 e1g1 g8e7 d2c1 d4f3 d1f3 e8g8 c1a3 f8e8 f1c1 b6d4 f3c3 e7f5 c3d4 f5d4 b1c3 c8d7 a3c5 d4f5 c3e2 f7f6 f2f4 a5a4 e2d4 f6e5 f4e5 f5d4 c5d4 e8c8 g1f2 g8f7 g2g4 d7b5 f2g3 a4a3 d4c5 b5d3 c1c3 d3e4');
INSERT INTO evaluations VALUES(18276,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(18276,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18276,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18276,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18276,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18276,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18276,36,'cp 0','c6c5 d4e2 g7g6 b1c3 f8g7 c1e3 c8b7 e1g1 g8e7 d1d3 e7c6 a1d1 c6e5 d3d2 e5c4 d2d3');
INSERT INTO evaluations VALUES(18282,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18282,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18282,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18282,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18282,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18282,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18282,36,'cp 0','d4e2 g7g6 b1c3 f8g7 c1e3 c8b7 d1d3 g8e7 e1g1 e7c6 a1d1 c6e5 d3d2 e5c4 d2d3');
INSERT INTO evaluations VALUES(19731,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(19731,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(19731,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(19731,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(19731,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(19731,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(19731,36,'cp 60','d8d5 b5c6 d5c6 g1f3 e7e6 e1g1 f8d6 b1c3 g8e7 f1e1 b7b6 c3e4 e8g8 d2d4 c8b7 d4d5 e6d5 f3d4 c6c7 e4d6 c7d6 d1g4 b7c8 g4g5 e7g6 g5g3 d6g3 h2g3 c8d7 a2a4 f8e8 e1e8 a8e8 c1d2 g6e7 a4a5 e7f5 d4f3 e8c8 a5b6 a7b6 a1a6 d5d4 a6b6 c8c2');
INSERT INTO evaluations VALUES(31976,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(31976,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(31976,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(31976,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(31976,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(31976,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(31976,36,'cp -71','g8f6 g2g4 f5g6 g4g5 f6d5 c3e4 e7e6 h2h4 f8e7 h4h5 g6f5 g1e2 b8c6 e2g3 h7h6 f3f4 h6g5 f4g5 d8d7 c2c3 e8c8 c1d2 f5h7 d1e2 f7f5 g5g6 h7g6 h5g6 h8h1 g3h1 f5e4 e1c1 e6e5 d4e5 d7f5 h1f2 c8b8 f2e4 c6e5 e4g3 f5f6 f1h3 f6h4 e2e5 h4h3');
INSERT INTO evaluations VALUES(34649,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(34649,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(34649,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(34649,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(34649,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(34649,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(34649,36,'cp 152','g1f3 d4d8 d2d4 g8f6 f1c4 b8c6 e1g1 e7e6 d1e2 f8e7 f1d1 e8g8 c3e4 a7a5 c2c3 a5a4 c4d3 f6e4 d3e4 f7f5 e4c2 e7d6 a2a3 d8f6 c1g5 f6f7 g5d2 b7b6 f3g5 f7f6 d1e1 c8d7 g5f3 h7h6 a1d1 c6a5 c3c4 d7c6 c4c5 b6c5 d4c5');
INSERT INTO evaluations VALUES(34648,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(34648,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(34648,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(34648,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(34648,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(34648,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(34648,36,'cp 159','d4d8 d2d4 g8f6 f1c4 b8c6 e1g1 e7e6 f1e1 f8e7 a2a3 a8b8 d1d3 e8g8 c1d2 h7h6 a1d1 a7a6 d4d5 f6d5 c3d5 e6d5 c4d5 c8e6 d5e6 d8d3 c2d3 f7e6 e1e6 e7f6 d2c3 b8e8 e6e3 f6c3 b2c3 b7b5 d3d4 e8e3 f2e3 c6a5 e3e4');
INSERT INTO evaluations VALUES(55298,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(55298,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(55298,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(55298,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(55298,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(55298,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(55298,36,'cp 147','f1c4 b8c6 e1g1 g8f6 d2d4 e7e6 d1e2 f8e7 f1d1 e8g8 c3e4 a7a5 c2c3 a5a4 d1e1 h7h6 c1d2 e7d6 a1d1 f6e4 e2e4 c8d7 f3e5 d7e8 d2f4 g8h8 e4f3 c6e7 f3h3 h8h7 c4d3 f7f5 h3f3 e7d5 f4g3 a4a3 c3c4 d6e5 g3e5 a3b2');
INSERT INTO evaluations VALUES(45256,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(45256,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45256,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(45256,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(45256,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(45256,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45256,36,'cp 3','b8c6 d4e3 g7g6 f1e2 c8g4 c3d5 f8g7 e1g1 e8g8 f1d1 f6d7 c2c3 d7c5 f3d4 g4e2 e3e2 e7e6 d5f4 a8c8 d4c6 b7c6 f4d3 c5d3 d1d3 d6d5 c1f4 d8b6 f4e3 b6a6 e2d2 c8d8 b2b3 d8d7 e4d5 c6d5 a1d1');
INSERT INTO evaluations VALUES(13387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13387,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13387,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13387,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13387,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13387,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13387,36,'cp 0','d4e3 g7g6 f1c4 f8g7 h2h3 b7b6 e1g1 e8g8 e3e2 c8b7 c1f4 a8c8 a1d1 d8c7 c3b5 c7b8 b5c3');
INSERT INTO evaluations VALUES(13392,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13392,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13392,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13392,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13392,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13392,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13392,36,'cp 0','g7g6 f1c4 f8g7 h2h3 e8g8 e1g1 b7b6 e3e2 c8b7 c1f4 a8c8 a1d1 d8c7 c3b5 c7b8 b5c3');
INSERT INTO evaluations VALUES(12611,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12611,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12611,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12611,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12611,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12611,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12611,36,'cp 0','d7d5 e4d5 c8f5 d5c6 b7c6 d1f3 d8d7 f1c4 f8e7 e1g1 e8g8 f1d1 d7c8 c4a6 c8e6 a6c4');
INSERT INTO evaluations VALUES(13807,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13807,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13807,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13807,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13807,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13807,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13807,36,'cp 0','e4d5 c8f5 d5c6 b7c6 d1f3 d8d7 f1c4 f8e7 e1g1 e8g8 f1d1 d7c8 c4a6 c8e6 a6c4');
INSERT INTO evaluations VALUES(13788,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13788,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13788,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13788,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13788,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13788,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13788,36,'cp 0','c8f5 d5c6 b7c6 d1f3 d8d7 f1c4 f8e7 e1g1 e8g8 f1d1 d7c8 c4a6 c8e6 a6c4');
INSERT INTO evaluations VALUES(24735,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24735,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24735,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24735,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24735,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24735,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24735,36,'cp 0','d5c6 b7c6 d1f3 d8d7 f1c4 f8e7 e1g1 e8g8 f1d1 d7c8 c4a6 c8e6 a6c4');
INSERT INTO evaluations VALUES(24685,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24685,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24685,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24685,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24685,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24685,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24685,36,'cp 0','b7c6 d1f3 d8d7 f1c4 f8e7 e1g1 e8g8 f1d1 d7c8 c4a6 c8e6 a6c4');
INSERT INTO evaluations VALUES(19226,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(19226,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(19226,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(19226,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(19226,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(19226,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(19226,36,'cp 57','h5c5 e7e6 c5e3 f8h6 e3e2 g8e7 g1f3 d7d5 c4b3 e8g8 e1g1 d5e4 e2e4 h6g7 b1a3 h7h6 d2d3 g6g5 e4e2 e7g6 c2c3 e6e5 c1e3 c8f5 d3d4 e5d4 f3d4 c6d4 c3d4 d8f6 a1d1 f8e8 e2d2 g6h4 f2f3 a8d8 a3c4 f5e6 f3f4 h4f5 f4g5 h6g5 e3g5 f6d4 d2d4 d8d4 c4d2 e6b3 a2b3 d4d5 d2c4 d5d1 f1d1');
INSERT INTO evaluations VALUES(19220,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(19220,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(19220,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(19220,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(19220,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(19220,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(19220,36,'cp 46','e7e6 c5e3 f8h6 e3e2 g8e7 c4b3 a7a5 c2c3 d7d5 g1f3 e6e5 e1g1 d5e4 e2e4 h6g7 d2d4 a5a4 b3c4 e5d4 f3d4 c6d4 c3d4 e8g8 c1g5 f8e8 f1e1 c8f5 e4b7 d8d4 b1a3 d4b2 a3b5 a8b8 c4f7 g8f7 b5d6 f7g8 b7c7 b2a1 e1a1 g7a1 d6e8 b8b1 g5c1 e7d5');
INSERT INTO evaluations VALUES(20001,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20001,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20001,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20001,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20001,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20001,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20001,36,'cp 12','g1f3 g7g6 d2d4 c5d4 f3d4 g8f6 d4c6 b7c6 e4e5 d6e5 d1d8 e8d8 f1g2 d8c7 e1g1 f8g7 f1e1 f6d7 c3b5 c7b7 f2f4 a7a6 b5a3 d7b6 f4e5 c8e6 b2b3 a6a5 a3b5 a5a4 c1f4 a4b3 a2b3 e6d5 b5d4 d5g2 g1g2 b6d5 h2h3 h8b8 a1d1 b7c7 d4f3 a8a2');
INSERT INTO evaluations VALUES(20000,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(20000,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(20000,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(20000,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(20000,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(20000,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(20000,36,'cp -13','e7e5 f1c4 h7h6 d2d3 g8f6 a2a4 f8e7 h2h3 c8e6 c3d5 d8d7 c2c3 e7d8 d5f6 d8f6 c4e6 d7e6 f3d2 e8g8 d2f1 d6d5 d1g4 c6a5 g4e6 f7e6 e1e2 b7b6 f1d2 a7a6 c3c4 a8d8 h3h4 d5c4 d2c4 a5c4 d3c4');
INSERT INTO evaluations VALUES(47616,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(47616,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(47616,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(47616,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(47616,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(47616,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(47616,36,'cp 15','f8e7 e1g1 e8g8 c4b3 f8e8 h2h3 c7c6 b3c2 e7f8 f1e1 b8d7 d3d4 a7a5 a2a4 d6d5 d4e5 f6e4 b1d2 d7e5 d2e4 e5f3 d1f3 d5e4 e1e4 e8e4 f3e4 g7g6 c2b3 c8f5 e4e2 d8f6 c1e3 f6e7 b3c2 f5c2 e2c2 g8h7 e3b6 a8a6 b6e3');
INSERT INTO evaluations VALUES(40686,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40686,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40686,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(40686,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(40686,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40686,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(40686,36,'cp 18','e1g1 e8g8 f1e1 c7c6 c4b3 b8d7 d3d4 f8e8 c3c4 e5d4 f3d4 d7c5 b1c3 c5b3 a2b3 d6d5 c4d5 c6d5 e4e5 f6e4 f2f3 e4c3 b2c3 e7c5 d1d3 d8h4 c1d2 c5d4 c3d4 c8e6 d2b4 e8c8 b4d6 h4g5 e1d1 g5h4 d6c5 a7a6 h2h3 h4g3');
INSERT INTO evaluations VALUES(40685,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(40685,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40685,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(40685,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40685,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40685,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40685,36,'cp 14','c7c6 f1e1 e8g8 h2h3 b8d7 c4b3 f8e8 c3c4 f6h7 d3d4 h7g5 d4d5 c6d5 c4d5 d7c5 b3c2 g5f3 d1f3 a7a5 c1e3 b7b6 a2a4 e7g5 b1a3 e8f8 a1b1 g5e3 e1e3 a8b8 b2b3 f7f5 e4f5 d8g5 f3g4 c8f5 c2f5 f8f5 g4g5');
INSERT INTO evaluations VALUES(48756,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48756,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48756,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48756,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48756,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48756,36,'cp 0','d5d4 f1c4 d4c3 c4f7 e8e7 f7b3 d8d4 e5f7 c3b2 c1b2 d4b2 f7h8 b8c6 c2c3 c8g4 f2f3 g4e6 a1b1 e6b3 a2b3 b2a2 e1g1 g7g6 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(48897,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(48897,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(48897,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(48897,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48897,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48897,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48897,36,'cp 0','f1c4 d4c3 c4f7 e8e7 f7b3 d8d4 e5f7 c3b2 c1b2 d4b2 f7h8 b8c6 c2c3 c8g4 f2f3 g4e6 a1b1 e6b3 a2b3 b2a2 e1g1 g7g6 e4e5 f6h5 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(50551,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(50551,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(50551,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(50551,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50551,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(50551,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50551,36,'cp 15','g8f6 c4d3 d8c7 e1g1 f8c5 d4b3 c5a7 c2c4 h7h5 b1c3 f6g4 g2g3 b8c6 d3e2 d7d6 c1f4 e6e5 c3d5 c7d8 f4d2 c8e6 a1c1 a8c8 c1c3 c6d4 b3d4 a7d4 c3b3 d8d7 d5b6 d4b6 b3b6 e6c4 e2c4 c8c4 d1b3 c4c7 f2f3 g4f6 d2g5 e8g8 f1d1 f8d8 g5f6 g7f6 b3d5 g8g7 b6d6');
INSERT INTO evaluations VALUES(43319,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43319,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(43319,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(43319,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(43319,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43319,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43319,36,'cp 8','c4d3 d8c7 d1e2 d7d5 e4e5 f6d7 f2f4 b8c6 c1e3 c6d4 e3d4 f8c5 e2f2 c7b6 d4c5 d7c5 e1g1 c8d7 b1d2 e8g8 d2b3 f8c8 a1d1 c8c7 d1d2 g7g6 h2h4 d7a4 b3c5 b6c5 h4h5 g6h5 c2c3 c5f2 g1f2 a4b5 d3b5 a6b5 a2a3 c7c4');
INSERT INTO evaluations VALUES(18851,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(18851,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(18851,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(18851,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(18851,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(18851,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(18851,36,'cp -20','d7d5 d2d3 g8h6 f1e2 f8e7 f3e5 e8g8 e1g1 c6a5 h2h3 b7b6 e5g4 h6g4 h3g4 d5d4 c3c4 e6e5 f4f5 e7g5 g2g3 g5c1 d1c1 f7f6 b1d2 a5b7 d2f3 h7h6 f3h4 b7d6 h4g6 f8e8 e2f3 d6f7 c1c2 d8d6 c2e2 c8d7 a2a4 a7a6 b2b3 b6b5');
INSERT INTO evaluations VALUES(20305,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(20305,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(20305,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(20305,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(20305,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20305,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(20305,36,'cp -11','d2d3 g8h6 f1e2 f8e7 e1g1 e8g8 f3e5 c6a5 h2h3 b7b6 e5g4 h6g4 h3g4 d5d4 c3c4 e6e5 f4f5 e7g5 g2g3 g5c1 d1c1 f7f6 b1d2 a5b7 d2f3 h7h6 g1g2 b7d6 f3h4 f8e8 b2b3 d6f7 h4g6 c8d7 a2a4 d8c7 f1h1 c7d6 e2f3 a8b8 c1e1 a7a6');
INSERT INTO evaluations VALUES(19826,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(19826,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(19826,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(19826,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(19826,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(19826,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(19826,36,'cp -31','c8g4 f1b5 d8b6 b5c6 b6c6 h2h3 g4h5 e1g1 h7h6 d2d4 e7e6 f1e1 c5d4 c3d4 a8c8 b1c3 g8e7 c1f4 c6b6 a1c1 e7c6 c3a4 b6b5 e1e3 f8e7 a2a3 e8g8 e3c3 b7b6 g2g4 h5g6 f4e3 f7f6 a4b6 b5b6 d1a4 b6b2 e5f6 f8f6');
INSERT INTO evaluations VALUES(26274,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(26274,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(26274,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(26274,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(26274,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(26274,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(26274,36,'cp -32','f1b5 d8b6 h2h3 g4h5 b5c6 b6c6 e1g1 e7e6 d2d4 g8e7 d4c5 c6c5 c1e3 c5c7 b1d2 e7c6 d1a4 f8e7 c3c4 d5c4 f3d4 c4c3 d4c6 b7c6 b2c3 e8g8 d2c4 h5e2 f1c1 e2c4 a4c4 c7e5 c4c6 a8c8 c6f3 e7d6 g2g3 c8c3 c1c3 e5c3');
INSERT INTO evaluations VALUES(26254,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(26254,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(26254,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(26254,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(26254,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(26254,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(26254,36,'cp -20','d8b6 b5c6 b6c6 h2h3 g4h5 e1g1 d5d4 c3d4 h5f3 d1f3 c6f3 g2f3 c5d4 d2d3 e7e6 b1d2 g8e7 f3f4 a8d8 d2f3 e7f5 c1d2 f8e7 f1c1 e8d7 g1f1 d8c8 f1e2 c8c1 a1c1 f7f6 f3e1 h8f8 d2a5 b7b6 a5d2 f8c8 c1c8 d7c8 e2f3');
INSERT INTO evaluations VALUES(20101,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20101,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20101,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20101,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20101,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20101,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20101,36,'cp 0','g8f6 b1c3 e7e6 c1f4 f8e7 e4e5 c6e5 f3e5 d6e5 d1d8 e7d8 f4e5 c8d7 f1e2 a8c8 e1c1 a7a6 e2f3 d7c6 f3c6 c8c6 d1d3 d8c7 h1d1 e8e7 e5f6 e7f6 c3e4 f6g6 e4d6 h7h5 d3g3 g6h7 g3d3');
INSERT INTO evaluations VALUES(13401,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13401,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13401,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13401,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13401,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13401,36,'cp 0','b1c3 e7e6 c1f4 f8e7 e4e5 c6e5 f3e5 d6e5 d1d8 e7d8 f4e5 c8d7 f1e2 a8c8 e1c1 a7a6 e2f3 d7c6 f3c6 c8c6 d1d3 d8c7 h1d1 e8e7 e5f6 e7f6 c3e4 f6g6 e4d6 h7h5 d3g3 g6h7 g3d3');
INSERT INTO evaluations VALUES(21467,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(21467,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(21467,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(21467,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(21467,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(21467,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(21467,36,'cp 75','g1f3 f7f6 f1b5 c8d7 b1c3 a7a6 b5c6 d7c6 h2h4 g8h6 c1h6 g7h6 d1d2 f8g7 e1c1 e8g8 d1e1 c6d7 c1b1 a8c8 h1h3 g8h8 g2g4 f6e5 f3e5 c7c5 f2f4 c5d4 c3e2 d4d3 c2d3 d7e8 h4h5 g7e5 f4e5 d8g5 d2g5 h6g5 e2d4 f8f4 d4e6');
INSERT INTO evaluations VALUES(46487,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46487,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(46487,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46487,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46487,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(46487,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46487,36,'cp 0','c4b3 b8c6 e1g1 f8d6 c1g5 c8e6 f1e1 e8g8 b1c3 d6c7 b3a4 d8b8 a4b3');
INSERT INTO evaluations VALUES(46494,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46494,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46494,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46494,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46494,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46494,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46494,36,'cp 0','b8c6 e1g1 f8d6 c1g5 c8e6 f1e1 e8g8 b1c3 d6c7 b3a4 d8b8 a4b3');
INSERT INTO evaluations VALUES(13802,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13802,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13802,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13802,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13802,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13802,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13802,36,'cp 0','e1g1 f8d6 c1g5 d6c7 f1e1 e8g8 b1c3 c8e6 b3a4 d8b8 a4b3');
INSERT INTO evaluations VALUES(13801,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(13801,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(13801,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(13801,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13801,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13801,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13801,36,'cp 0','f8d6 c1g5 c8e6 g5h4 e8g8 f1e1 f8e8 b1d2 a8c8 h2h3 h7h6 c2c4 d5c4 d2c4 d6c7 a1c1 a7a6 h4g3 f6d7 g3h4');
INSERT INTO evaluations VALUES(1127,30,'cp 475',NULL);
INSERT INTO evaluations VALUES(1127,31,'cp 450',NULL);
INSERT INTO evaluations VALUES(1127,32,'cp 485',NULL);
INSERT INTO evaluations VALUES(1127,33,'cp 458',NULL);
INSERT INTO evaluations VALUES(1127,34,'cp 498',NULL);
INSERT INTO evaluations VALUES(1127,35,'cp 498',NULL);
INSERT INTO evaluations VALUES(1127,36,'cp 513','c1e3 f6g8 b1c3 c7c6 e1c1 f7f6 f2f4 g8h6 d4d2 f6e5 f4e5 e7e6 f1d3 h6g4 e3d4 f8e7 h1f1 h8f8 f1f8 e8f8 d2f4 f8g8 f4e4 g7g6 h2h3 g4e5 e4e5 e6e5 d4e5 e7g5 c1b1 d7d5 a2a3 c8d7 c3e4 g5e7 e4f6 e7f6 e5f6 a8e8 f6c3 e8e3 b1c1 e3d3 d1d3 c6c5 c1d2 d7b5 d3e3 d5d4');
INSERT INTO evaluations VALUES(1125,30,'cp 499',NULL);
INSERT INTO evaluations VALUES(1125,31,'cp 499',NULL);
INSERT INTO evaluations VALUES(1125,32,'cp 498',NULL);
INSERT INTO evaluations VALUES(1125,33,'cp 497',NULL);
INSERT INTO evaluations VALUES(1125,34,'cp 488',NULL);
INSERT INTO evaluations VALUES(1125,35,'cp 488',NULL);
INSERT INTO evaluations VALUES(1125,36,'cp 484','f6g8 b1c3 c7c6 e1c1 f7f6 f2f4 g8h6 d4d2 f6e5 f4e5 e7e6 f1d3 h6g4 e3d4 f8e7 h1f1 h8f8 f1f8 e8f8 d2f4 f8g8 f4e4 g7g6 h2h3 g4e5 d4e5 d7d5 e4d4 c8d7 c1b1 b7b6 d1e1 a8f8 e5g3 e6f7 d4a4 e7f6 a4b3 b6b5 a2a3 a7a5 b1a1 f6g7 c3b1 b5b4 a3b4');
INSERT INTO evaluations VALUES(11593,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(11593,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11593,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11593,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(11593,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(11593,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(11593,36,'cp 57','d7d6 e1g1 c8d7 f1e1 e7g6 c3d5 b7b5 a4b3 c6a5 d2d4 c7c6 d5e3 f8e7 d4e5 d6e5 e3f5 e8g8 c1e3 a5b3 a2b3 f8e8 d1d2 e7f6 f5d6 e8e7 e1d1 d7g4 d6f5 e7e8 d2d8 e8d8 d1d8 f6d8 f3d2 g6f4 g2g3 f4e6 b3b4 g4f5 e4f5 e6d4 e3d4 e5d4 d2e4 d8e7');
INSERT INTO evaluations VALUES(21174,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(21174,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(21174,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(21174,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(21174,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(21174,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21174,36,'cp 98','d4d5 c6b8 e2e4 e7e5 f1e2 g7g6 b1c3 b8d7 h2h4 f8h6 c1h6 g8h6 d1d2 h6g8 g2g4 h7h6 e1c1 g8f6 g1h3 d7c5 f2f3 a7a5 h3f2 f6d7 d1f1 a5a4 h4h5 g6g5 f2d1 d7b6 c1b1 d8f6 d1e3 f6f4 e2d1 f4f6 b2b4 a4b3 a2b3');
INSERT INTO evaluations VALUES(21165,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(21165,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(21165,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(21165,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(21165,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(21165,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(21165,36,'cp 95','c6b8 b1c3 e7e5 e2e4 b8d7 f1e2 g7g6 h2h4 f8h6 c1h6 g8h6 d1d2 h6g8 g2g4 h7h6 e1c1 g8f6 g1h3 d7c5 f2f3 a7a5 h3f2 f6d7 d1g1 d8e7 c1b1 d7b6 h4h5 g6g5 f2d1 c8d7 d1e3 b6a4 c3b5 e8d8 e2d1 e7f6 g1f1 d7b5 c4b5 a4b6');
INSERT INTO evaluations VALUES(30013,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(30013,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(30013,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(30013,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(30013,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(30013,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(30013,36,'cp 65','g1f3 b8c6 e1g1 f8b4 f3g5 d8g5 c3d5 b4d6 d2d4 g5h5 d4e5 d6e5 d1a4 e8g8 g2f3 h5f5 a4c6 b7c6 d5e7 g8h8 e7f5 e6f5 f3c6 a8b8 a2a4 e5b2 a1a2 b2c3 c1f4 b8b6 c6b5 c7c6 b5c4 c6c5 f1c1 b6b1 g1g2 b1c1 f4c1 h7h6 c1e3 f5b1 a2a3 c3b4 a3b3 b1a2');
INSERT INTO evaluations VALUES(30012,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(30012,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(30012,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(30012,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(30012,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(30012,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(30012,36,'cp 55','b8c6 e1g1 f8b4 f3g5 d8g5 c3d5 b4d6 d2d4 g5h5 d1a4 e8g8 d4e5 d6e5 g2f3 h5f5 a4c6 b7c6 d5e7 g8h8 e7f5 e6f5 c1e3 e5b2 a1d1 f5h3 f3g2 h3g2 g1g2 h8g8 d1d7 f8c8 a2a4 a8b8 f1b1 a7a5 h2h4 b2e5 b1c1 e5d6 c1c6 b8b2 c6a6 b2e2 a6a5 c8e8 g2f3');
INSERT INTO evaluations VALUES(25112,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(25112,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(25112,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(25112,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(25112,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(25112,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(25112,36,'cp 51','e1g1 f8b4 f3g5 d8g5 c3d5 b4d6 d2d4 g5h5 d4e5 d6e5 g2f3 e6g4 d5f4 e5f4 f3c6 b7c6 c1f4 e8g8 f1e1 h5b5 d1c2 f8b8 b2b3 b8e8 f2f3 g4e6 a1c1 a7a5 c2c6 b5c6 c1c6 a5a4 c6c3 a4b3 a2b3 a8a3 g1f2 a3b3 c3c7 h7h6 g3g4 b3b4 f4e3');
INSERT INTO evaluations VALUES(25111,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25111,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(25111,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(25111,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(25111,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(25111,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(25111,36,'cp 56','f8b4 f3g5 d8g5 c3d5 b4d6 d2d4 g5h5 d4e5 d6e5 g2f3 e6g4 d5f4 e5f4 f3c6 b7c6 c1f4 e8g8 f1e1 h5b5 d1c2 f8b8 a1d1 b8e8 d1d2 h7h6 g1g2 c6c5 f2f3 g4e6 e1a1 b5b6 f4e3 e6a2 e3c5 b6e6 c5d4 a7a5 c2c7 a5a4 c7c3 f7f6 a1c1 a8b8 c3c6 b8b4 c6e6 a2e6 c1c7');
INSERT INTO evaluations VALUES(32096,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(32096,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(32096,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(32096,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(32096,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(32096,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(32096,36,'cp 100','d1f3 g8f6 d2d3 b8c6 c1e3 c6a5 g2g4 a5c4 d3c4 h7h5 g4g5 f6d7 h2h4 c7c6 e1c1 d8e7 b2b3 e7e6 f3g3 f8e7 h1f1 e8g8 c1b1 f8e8 f2f4 e5f4 f1f4 d7e5 e3d4 e7f8 d1f1 e8e7 f4f5 a8e8 d4e5 d6e5 g3f2 a7a6 f2f3 e7d7 a2a4 f8b4 c3a2 g7g6 f5f6');
INSERT INTO evaluations VALUES(32095,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(32095,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(32095,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(32095,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(32095,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(32095,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(32095,36,'cp 95','g8f6 c3e2 c7c6 d2d3 b8d7 a2a3 h7h5 h2h3 g7g6 c4a2 d8e7 c1g5 f8h6 g5h6 h8h6 f3e3 h6h7 h1f1 d7b6 f2f4 f6d7 e1c1 e8c8 c1b1 c8b8 f4e5 d6e5 d3d4 f7f6 d4d5 c6d5 a2d5 b6d5 d1d5 d7b6 e2c3 b6d5 c3d5 d8d5 e4d5 f6f5');
INSERT INTO evaluations VALUES(29914,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(29914,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(29914,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(29914,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(29914,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(29914,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(29914,36,'cp 101','c3e2 c7c6 d2d3 b8d7 c1e3 h7h5 h2h3 g7g6 a2a3 f8h6 e3h6 h8h6 e1c1 h6h7 c4a2 d8e7 f3e3 a7a6 f2f4 e8c8 h1f1 f6e8 e3a7 e8c7 e2c3 e5f4 f1f4 e7g5 d1f1 f7f6 a7f2 c8b8 c1b1 d8f8 c3e2 h7e7 h3h4 g5h6 f2g3 b8a8');
INSERT INTO evaluations VALUES(37757,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(37757,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(37757,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(37757,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(37757,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(37757,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(37757,36,'cp -89','e7e6 d2d3 g8e7 e1g1 e8g8 b1c3 d7d5 c4b3 b8c6 e4d5 e6d5 h2h3 b7b6 g5f3 a7a6 f1e1 c8e6 f3g5 e6f5 c3e2 b6b5 c2c3 c5c4 d3c4 d5c4 b3c2 d8d1 c2d1 b5b4 g5f3 a8c8 c3b4 e7d5 a2a3 f5d3 d1a4 f8e8 a4c6 c8c6 e2g3 e8e1 f3e1 h7h6 e1d3 c4d3');
INSERT INTO evaluations VALUES(37613,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(37613,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(37613,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(37613,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(37613,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(37613,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(37613,36,'cp -81','g5f3 b8c6 e1g1 g8e7 c4b3 d7d5 d2d3 e8g8 b1c3 b7b6 e4d5 e7d5 c3d5 e6d5 b3a4 d8d6 c2c3 d5d4 a4c6 d6c6 c3d4 c5d4 f1e1 c8b7 c1f4 a8e8 d1d2 c6d7 e1e8 f8e8 a1e1 e8e1 d2e1 d7b5 e1d2 b7f3 g2f3 b5d5 d2e2 h7h6 g1g2 d5a2 f4b8');
INSERT INTO evaluations VALUES(41470,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(41470,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(41470,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41470,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(41470,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(41470,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(41470,36,'cp 0','c7c5 e1g1 e8g8 h2h3 d8b6 b1c3 c5c4 d3e2 b6b2 d1d2 b8c6 a1b1 b2a3 f3e5 e7e6 c3b5 a3a5 d2a5 c6a5 b5c7 a8b8 c7a6 b8a8');
INSERT INTO evaluations VALUES(41860,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(41860,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41860,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(41860,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41860,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41860,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(41860,36,'cp 13','c3a4 c5b4 c2c3 d6d5 e4d5 b4e7 d3d4 e5d4 c3d4 f6d5 a4c3 c7c6 d1b3 e8g8 c4d5 c6d5 g1e2 b8c6 e1g1 b7b6 b3d5 d8d5 c3d5 c8a6 d5e7 c6e7 f1e1 e7c6 c1d2 a6e2 e1e2 c6d4 e2e7 f8d8 d2c3 d4b5 c3e1 a8c8 a2a4 g8f8 e7b7 c8c7 b7c7 b5c7 a4a5 c7e6 h2h3 f8e8 e1c3 f7f6 a5b6 a7b6');
INSERT INTO evaluations VALUES(12813,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(12813,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(12813,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(12813,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12813,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(12813,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(12813,36,'cp 30','b1c3 e7e6 g1f3 d8b6 e1g1 f8e7 b2b3 d7d5 c4d5 f6d5 c3d5 e6d5 d2d4 c5d4 c1b2 e7f6 b2a3 b6a5 a3d6 c8e6 b3b4 a5d8 d6c5 a8c8 f3d4 f6d4 c5d4 e8g8 e2e3 c6d4 d1d4 c8c4 d4a7 c4b4 a7a3');
INSERT INTO evaluations VALUES(15690,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(15690,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(15690,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(15690,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(15690,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(15690,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(15690,36,'cp 66','a2a3 f8e7 f1d3 e6e5 d4d5 c6b8 c3a4 e8g8 c2c4 c8g4 h2h3 g4h5 a4c3 b8d7 g2g4 h5g6 d1e2 d7c5 d3c2 c7c6 h3h4 c6d5 c4d5 a8c8 h4h5 g6e4 c3e4 f6e4 c2e4 c5b3 e4h7 g8h7 e2d3 e5e4 d3b3 e4f3 b3f3 e7g5 e1g1 h7g8 c1e3 c8c2');
INSERT INTO evaluations VALUES(18302,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(18302,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(18302,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(18302,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(18302,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(18302,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(18302,36,'cp 62','d8f6 d4d3 f8b4 c2c3 b4e7 e1g1 f6g6 b1d2 c8b7 f1e1 g8f6 d2f3 e8g8 e4e5 g6d3 c4d3 f6h5 d3f5 b7f3 g2f3 a8d8 f5g4 g7g6 g4h5 g6h5 c1h6 f8e8 a1d1 f7f6 h6f4 e7f8 g1f1 d7d6 e5e6 e8e7 e1e4 d8e8 e4c4 c7c5 c4a4 a7a5 d1d6 e7e6 d6e6 e8e6 f4e3 g8f7');
INSERT INTO evaluations VALUES(30157,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(30157,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(30157,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(30157,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(30157,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(30157,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(30157,36,'cp 58','d1b3 b8d7 b3b7 g4f3 g2f3 c7c5 c4d5 f6d5 b1c3 d5c3 b2c3 a8b8 b7a7 c5d4 a7d4 f8e7 f1e2 d8a5 e1g1 e7f6 d4d3 e8g8 a2a4 f6c3 a1a2 d7c5 d3c2 f8c8 e2b5 c5a6 c1a3 a6b4 a3b4 c3b4 c2e4 g7g6 e4f4 b4e7 f1b1 c8d8 f4e4 d8c8 f3f4 c8c5 a2c2 b8c8 c2c5');
INSERT INTO evaluations VALUES(41563,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(41563,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(41563,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(41563,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(41563,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(41563,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41563,36,'cp 31','c7c5 g1e2 b8c6 c1e3 e8g8 e1g1 b7b6 a1c1 e7e6 c4b5 c8d7 d4c5 c6e5 c3c4 d7b5 c4b5 e5d3 c1c4 d3c5 d1c2 a7a6 f1d1 d8c7 a2a4 a6b5 a4b5 a8a3 g2g3 f8a8 e3c5 b6c5 c4c5 c7b7 d1b1 a3a2 c2c4 a2a4 c5c7 a4c4 c7b7 c4e4');
INSERT INTO evaluations VALUES(40430,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(40430,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(40430,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40430,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(40430,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40430,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(40430,36,'cp 27','g1e2 e8g8 e1g1 b8c6 c1e3 b7b6 a1c1 c8b7 d4d5 c6e5 c4b3 f7f5 e4f5 c5c4 b3c2 d8d5 d1d5 b7d5 f5g6 h7g6 e2f4 d5f7 f1e1 a8d8 e3d4 e5d3 c2d3 c4d3 f4d3 g7d4 c3d4 d8d4 d3e5 d4a4 e5f7 g8f7 c1c7 f8e8 e1e2 a7a5 g2g3 a4d4 f2f3 d4b4');
INSERT INTO evaluations VALUES(40429,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(40429,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40429,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40429,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40429,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(40429,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(40429,36,'cp 28','b8c6 c1e3 e8g8 e1g1 b7b6 c4b5 c8b7 a1c1 a8c8 b5c6 c8c6 d4d5 c6d6 c3c4 e7e6 d1a4 e6d5 e4d5 d6d7 e3g5 d8c8 a4b3 h7h6 g5d2 d7e7 e2g3 h6h5 h2h3 f8e8 f1d1 g7d4 d2c3 h5h4 g3f1 e7e2 c3d4 c5d4 d1d4 c8f5');
INSERT INTO evaluations VALUES(8237,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(8237,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(8237,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(8237,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8237,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(8237,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(8237,36,'cp 10','c1e3 e8g8 e1g1 b7b6 a1b1 c8b7 d1a4 c5d4 c3d4 a7a6 d4d5 b6b5 a4d1 c6a5 c4d3 a8c8 a2a4 a5c4 a4b5 c4e3 f2e3 a6b5 b1b5 b7a6 b5b3 a6c4 b3b4 c4d3 d1d3 d8a5 b4b7 c8c7 b7c7 a5c7 f1c1 c7d6 e2d4');
INSERT INTO evaluations VALUES(8235,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(8235,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(8235,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8235,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(8235,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8235,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(8235,36,'cp 15','e8g8 e1g1 b7b6 d1d2 c8b7 a1c1 e7e6 f1d1 a8c8 e3g5 d8d7 h2h4 c5d4 c3d4 c6a5 c4d3 c8c1 d1c1 a5c6 d3b5 a7a6 b5c6 b7c6 e2g3 f8c8 d4d5 c6b5 c1c8 d7c8 d2f4 h7h5 g5f6 e6d5 f6g7 g8g7 e4d5 c8c4 f4d2 c4h4 d5d6');
INSERT INTO evaluations VALUES(3158,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(3158,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(3158,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(3158,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3158,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(3158,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3158,36,'cp 23','e1g1 b7b6 a1c1 e7e6 c4b5 c8d7 b5d3 d7c8 e4e5 c8b7 d1d2 d8d7 f1d1 a8d8 f2f3 f7f6 e5f6 g7f6 d3e4 d7e7 d2e1 c6a5 e4b7 a5b7 e2g3 c5d4 e3d4 b7d6 d4f6 e7f6 d1d3 d6f5 d3d8 f8d8 g3e4 f6e7 c1d1 f5d6 e4f2 d8c8 h2h3 d6f5 f2e4');
INSERT INTO evaluations VALUES(3157,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(3157,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(3157,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(3157,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3157,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(3157,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3157,36,'cp 26','b7b6 a1c1 e7e6 c4b5 c8d7 b5d3 d7c8 e4e5 c8b7 d1d2 d8d7 f1d1 a8d8 f2f3 f7f6 d3e4 c5d4 e5f6 g7f6 c3d4 c6e7 e3g5 e7d5 g5f6 d5f6 e2c3 f6d5 e4d5 b7d5 d2e3 d7d6 c3b5 d6d7 a2a4 d5c6 h2h4 c6b5 a4b5 d7b5 e3e6 f8f7 c1c8 d8c8 e6c8 g8g7');
INSERT INTO evaluations VALUES(54227,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54227,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54227,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54227,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54227,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54227,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54227,36,'cp 0','e6e5 c3d5 c5d4 d1h5 d4e3 h5e5 f8e7 e1c1 e8f8 d5c7 b8d7 e5f4 g7g5 c7e6 f8e8 e6c7');
INSERT INTO evaluations VALUES(9662,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9662,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9662,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9662,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9662,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9662,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9662,36,'cp 0','d8h4 g2g3 e4g3 f7h8 c6d4 h2g3 h4g3 g1f1 g3f4 f1e1 f4e4 e1f2 e4f5 f2e1');
INSERT INTO evaluations VALUES(534,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(534,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(534,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(534,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(534,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(534,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(534,36,'cp 0','g2g3 e4g3 f7h8 c6d4 h2g3 h4g3 g1f1 g3f4 f1e1 d4f3 e1e2 f3d4');
INSERT INTO evaluations VALUES(532,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(532,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(532,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(532,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(532,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(532,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(532,36,'cp 0','e4g3 f7h8 c6d4 h2g3 h4g3 g1f1 g3f4 f1e1 f4e4 e1f2 e4f5 f2e1');
INSERT INTO evaluations VALUES(538,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(538,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(538,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(538,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(538,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(538,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(538,36,'cp 0','f7h8 c6d4 h2g3 h4g3 g1f1 g3f4 f1e1 f4g3');
INSERT INTO evaluations VALUES(14394,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(14394,31,'cp 348',NULL);
INSERT INTO evaluations VALUES(14394,32,'cp 371',NULL);
INSERT INTO evaluations VALUES(14394,33,'cp 373',NULL);
INSERT INTO evaluations VALUES(14394,34,'cp 366',NULL);
INSERT INTO evaluations VALUES(14394,35,'cp 396',NULL);
INSERT INTO evaluations VALUES(14394,36,'cp 388','d4d5 c6b8 e4e5 f6d7 e5e6 f7e6 d5e6 d7f6 d1a4 c7c6 f1c4 g7g6 c1e3 f8g7 a1d1 d8c7 g1f3 e8g8 c4b3 a7a5 f3g5 b8a6 a4h4 c7e5 e1g1 h7h6 g5f7 e5h5 h4h5 f6h5 f7h6 g7h6 e3h6 h5g7 c3a4 g8h7 a4b6 h7h6 b6a8 a5a4 b3a4 c8e6 a8b6 e6a2 d1a1 a6b4 f1e1 g7f5 a1d1 a2g8 d1d7 f8b8 e1d1 b4d5 b6d5 g8d5 a4c2');
INSERT INTO evaluations VALUES(14186,30,'cp 394',NULL);
INSERT INTO evaluations VALUES(14186,31,'cp 371',NULL);
INSERT INTO evaluations VALUES(14186,32,'cp 389',NULL);
INSERT INTO evaluations VALUES(14186,33,'cp 378',NULL);
INSERT INTO evaluations VALUES(14186,34,'cp 372',NULL);
INSERT INTO evaluations VALUES(14186,35,'cp 355',NULL);
INSERT INTO evaluations VALUES(14186,36,'cp 377','c6b8 e4e5 f6d7 e5e6 f7e6 d5e6 d7f6 d1a4 c7c6 f1c4 g7g6 c1e3 f8g7 a1d1 d8c7 g1f3 e8g8 c4b3 a7a5 a4h4 a5a4 c3a4 c7a5 a4c3 b8a6 e1g1 a6c7 f1e1 c8e6 b3e6 c7e6 e3d2 e6d8 e1e7 d8f7 e7b7 a5a6 h4b4 c6c5 b4b3 c5c4 b3b4 f8e8 d2f4 f6h5 f4e3 g7c3 b2c3 a6a4 b4a4');
INSERT INTO evaluations VALUES(18715,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(18715,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(18715,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(18715,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(18715,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(18715,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(18715,36,'cp -58','g1f3 d7d6 b2b3 c8d7 b1a3 g8h6 f1d3 c6b4 d3e2 f8e7 e5d6 e7d6 c1b2 e8g8 g2g3 f7f6 e1g1 d6c7 a3c4 f8e8 f3h4 d7c6 a2a3 b4d5 e2h5 e8e7 h5f3 b7b5 c4e3 d8d7 a3a4 b5a4 b3a4 a8b8 b2a3 c7b6 e3c4 d5b4 f3c6 b4c6');
INSERT INTO evaluations VALUES(18714,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(18714,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(18714,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(18714,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(18714,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(18714,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(18714,36,'cp -71','d7d6 b2b3 c8d7 b1a3 g8h6 f1d3 c6b4 d3f1 f8e7 c2c3 b4d5 g2g3 e8g8 e5d6 e7d6 a3c4 d6c7 a2a4 f8e8 f1g2 d7c6 e1g1 b7b6 c1b2 f7f6 d1c2 h6f5 a1d1 d8d7 c4e3 a8d8 e3f5 e6f5 f1e1 e8e4 f3h4 e4e1 d1e1 g7g6 d2d4 d8e8 e1e8 d7e8');
INSERT INTO evaluations VALUES(20658,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(20658,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(20658,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(20658,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(20658,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(20658,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(20658,36,'cp -64','b2b3 c8d7 b1a3 g8h6 f1d3 c6b4 d3e2 f8e7 e1g1 e8g8 c1b2 d7e8 c2c3 b4d5 g2g3 d6e5 f3e5 d8c7 e2f3 a8d8 d1e2 h6f5 a1d1 e7f6 e5g4 e8c6 a3c4 h7h5 g4e5 h5h4 e2f2 b7b6 e5c6 c7c6 f3g2');
INSERT INTO evaluations VALUES(35989,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(35989,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(35989,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(35989,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(35989,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(35989,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(35989,36,'cp 33','f2f3 c8d7 d1d2 b8c6 e1c1 a8c8 c4e2 a7a6 h2h4 h7h5 c1b1 c6e5 c3d5 f6d5 e4d5 c8c5 d4e6 c5d5 e6d8 d5d2 e3d2 f8d8 d2g5 d8e8 h1g1 d7c6 a2a3 b7b5 d1f1 e5c4 c2c3 f7f6 g5c1 g8f7 f1e1 c6d7 g2g4 c4e5 g4h5');
INSERT INTO evaluations VALUES(35991,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(35991,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(35991,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(35991,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(35991,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(35991,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(35991,36,'cp 44','b8c6 d1d2 c8d7 e1c1 a8c8 c4b3 c6d4 e3d4 b7b5 a2a3 a7a5 e4e5 d6e5 d4e5 b5b4 e5f6 e7f6 d2d7 d8d7 d1d7 b4c3 b2c3 c8c3 c1d2 f8c8 b3f7 g8f8 f7b3 g7h6 d2e2 a5a4 b3a4 c8e8 e2f2 h6e3 f2g3 c3a3 a4b3 a3a7 d7d6 f8g7 h2h4 a7c7 h1d1 f6f5 d6d7 c7d7 d1d7 g7h6 b3d5 e3c5 c2c4');
INSERT INTO evaluations VALUES(40268,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(40268,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(40268,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(40268,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(40268,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(40268,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(40268,36,'cp 62','c4d5 c6d5 e2e3 b8c6 f1d3 e8g8 h2h3 a7a6 a1c1 b7b5 a2a4 c6b4 d3b1 b5a4 e1g1 a6a5 g5h4 c8d7 f3e5 a8c8 e5d7 d8d7 d1a4 d7a4 c3a4 e7e6 c1c8 f8c8 a4c5 c8a8 f1c1 a8c8');
INSERT INTO evaluations VALUES(44823,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(44823,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(44823,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(44823,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44823,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(44823,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(44823,36,'cp 73','c4d5 c6d5 c1f4 e7e6 e2e3 b8c6 f1d3 f8d6 f3e5 d6e5 d4e5 f6d7 d1h5 e8g8 e1g1 f7f5 e5f6 d7f6 h5e2 e6e5 f4g3 c8e6 h2h3 g8h8 a1d1 e6g8 a2a3 d8e8 g3h4 a8c8 d3f5 c8d8 f5b1 c6a5 b1a2 d5d4 a2g8 f8g8 c3e4 f6e4');
INSERT INTO evaluations VALUES(44817,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(44817,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(44817,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(44817,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(44817,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(44817,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44817,36,'cp 56','c6d5 c1f4 e7e6 e2e3 b8c6 f1d3 f8d6 f3e5 e8g8 e1g1 d6e5 d4e5 f6d7 d1h5 f7f5 e5f6 d7f6 h5e2 e6e5 f4g3 c8e6 a1d1 g8h8 d3b1 d8e8 f2f3 e6g8 a2a3 a7a6 b2b4 a8d8 c3a4 e8e7 g3h4 g8h7 b1h7 h8h7 a4c5 h7g8 h2h3 d8e8 f1e1 e7f7 e2d2 g8h8 d2b2');
INSERT INTO evaluations VALUES(44832,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(44832,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(44832,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(44832,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(44832,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(44832,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(44832,36,'cp 66','c1f4 e7e6 e2e3 b8c6 f1d3 f8d6 f3e5 d6e5 d4e5 f6d7 d1h5 e8g8 e1g1 f7f5 e5f6 d7f6 h5e2 e6e5 f4g3 g8h8 a1d1 c8e6 d3b1 d8e7 h2h3 a8d8 g3h4 d8d7 f2f3 e6g8 a2a3 f8d8 f1e1 g8h7 b1a2 h7g8 b2b4 d5d4 a2g8 d4d3 e2d2 d8g8 h4f6 g7f6 c3e4');
INSERT INTO evaluations VALUES(44831,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(44831,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(44831,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(44831,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(44831,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(44831,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(44831,36,'cp 63','e7e6 e2e3 b8c6 f1d3 f8d6 f3e5 e8g8 e1g1 d6e5 d4e5 f6d7 d1h5 f7f5 e5f6 d7f6 h5e2 e6e5 f4g3 c8e6 h2h3 g8h8 a1d1 d8e8 a2a3 a8d8 f2f3 e6g8 g3h4 a7a6 e2e1 d8d7 d3f5 d7c7 f5b1 d5d4 e3d4 e5d4 e1e8 f8e8');
INSERT INTO evaluations VALUES(46252,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(46252,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46252,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(46252,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(46252,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(46252,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(46252,36,'cp 9','c5d4 f3d4 e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 h7h6 d2e4 d8b6 e4d6 f8d6 g2d5 e6h3 c1e3 b6c7 d5g2 h3g2 g1g2 e8g8 a1c1 c7e7 b3d2 a8d8 d1b3 g8h8 d2e4 c6d4 b3c4 d6b8 f1d1 b7b5 c4d3 f8e8 e3d4 e5d4 e4c5 a7a6 c5b3 e7e2 d3e2 e8e2 d1d4 d8d4 b3d4');
INSERT INTO evaluations VALUES(46369,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46369,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46369,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(46369,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(46369,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46369,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(46369,36,'cp 16','f3d4 e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 h7h6 d2e4 d8b6 e4d6 f8d6 g2d5 e6h3 c1e3 b6c7 d5g2 h3g2 g1g2 e8g8 a1c1 c7e7 b3d2 a8d8 d1b3 g8h8 d2e4 c6d4 e3d4 e5d4 e4d6 d8d6 c1c2 f8e8 f1d1 b7b6 d1c1 h8g8 c2d2 e7d7 c1d1 d7c6 g2g1 c6c5 b3d3 g7g6');
INSERT INTO evaluations VALUES(46330,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(46330,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(46330,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(46330,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46330,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46330,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(46330,36,'cp 0','e6e5 d4b3 c8e6 c2c4 b8c6 c4d5 f6d5 b1d2 b7b6 d2e4 f8e7 e4c3 d5b4 d1d8 a8d8 a2a3 b4d5 c3d5 e6d5 c1e3 d5g2 g1g2 d8d5 b3d2 e8d7 d2e4 h8c8 e4c3 d5d6 c3e4');
INSERT INTO evaluations VALUES(51282,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(51282,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(51282,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(51282,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(51282,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(51282,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(51282,36,'cp 110','g1f3 e7e6 f1d3 c8b7 e1g1 b8d7 c2c4 f8e7 b1c3 e7f6 c1e3 g8e7 d3c2 e8g8 d1d2 e7g6 h2h3 d8e7 a1d1 a8e8 f1e1 c7c5 d4d5 f6c3 d2c3 e6e5 b2b4 f7f6 a2a3 b7c8 e3d2');
INSERT INTO evaluations VALUES(51281,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(51281,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(51281,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(51281,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(51281,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(51281,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(51281,36,'cp 88','c8b7 b1c3 a7a6 f1c4 b6b5 c4b3 e7e6 d4d5 e6e5 a2a4 b5b4 c3a2 a6a5 c2c3 b4c3 a2c3 g8f6 d1e2 b8a6 e1g1 f6d7 c1e3 f8e7 a1c1 a6c5 b3c4 e8g8 e2c2 g8h8 c3e2 f7f5 e4f5 c5a6 c4a6 b7a6 f1d1 a6e2 c2e2 f8f5 b2b3 d7f6 e2b5 f6g4 e3d2');
INSERT INTO evaluations VALUES(30537,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(30537,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(30537,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(30537,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(30537,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(30537,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(30537,36,'cp 107','f1d3 b8d7 c2c4 g8f6 b1c3 g7g6 a2a3 f8g7 e1g1 e8g8 f1e1 e7e5 d4d5 g8h8 b2b4 f6g8 d1c2 b7c8 h2h3 a7a5 c3a4 g8e7 c1d2 f7f5 b4a5 b6a5 f3g5 d7f6 a4c3 h7h6 g5e6 c8e6 d5e6 f5f4 a1b1');
INSERT INTO evaluations VALUES(54955,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(54955,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(54955,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(54955,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(54955,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54955,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54955,36,'cp 9','e7e5 b1c3 c7c5 d2d3 b8c6 a2a3 f8e7 f1e2 g8f6 f2f4 e5f4 c1f4 e8g8 g1f3 c8e6 e1g1 f6d7 g1h1 d7e5 c3d5 a7a5 f3e5 d6e5 f4e3 c6d4 e2g4 e7g5 e3d4 e5d4 a3a4 d8d6 g2g3 a8b8 g4e6 f7e6 d1g4 e6d5');
INSERT INTO evaluations VALUES(26555,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26555,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26555,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26555,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26555,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26555,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26555,36,'cp 0','e7e6 f1d3 g8f6 e1g1 h7h5 a2a3 a7a6 f3g5 f6g4 g5f3');
INSERT INTO evaluations VALUES(26822,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26822,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26822,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26822,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26822,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26822,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26822,36,'cp 0','f1d3 h7h5 e1g1 g8f6 f3g5 a7a6 f1e1 f8e7 a2a3 f6g4 g5f3 g4f6 f3g5');
INSERT INTO evaluations VALUES(34840,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(34840,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(34840,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(34840,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(34840,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(34840,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(34840,36,'cp 118','e5d6 d2d4 g8f6 e1g1 f8e7 c3b5 d6d8 c2c4 a7a6 b5c3 e8g8 c1f4 b8d7 c4c5 f6d5 f4g3 d7f6 f1e1 f6h5 d1b3 h5g3 h2g3 d5c3 b2c3 a8b8 f3e5 c8d7 a1d1 e7f6 e2f3 f6e5 e1e5 d7c6 f3c6 b7c6 b3a3 d8f6 a3a6 b8b2');
INSERT INTO evaluations VALUES(34814,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(34814,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(34814,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(34814,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(34814,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(34814,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(34814,36,'cp 134','d2d4 g8f6 e1g1 f8e7 c3b5 d6d8 c2c4 a7a6 b5c3 b8d7 c1f4 f6h5 f4e3 h5f6 f1e1 e8g8 d1c2 b7b6 f3e5 c8b7 a1d1 d8c8 e2d3 h7h6 e3f4 f8e8 h2h3 c7c5 d4d5 d7e5 f4e5 e6d5 c3d5 f6d5 c4d5 c8d8 c2c3 e7f6 e5f6 e8e1 d1e1 d8f6 c3f6 g7f6');
INSERT INTO evaluations VALUES(42018,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(42018,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42018,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(42018,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(42018,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42018,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(42018,36,'cp -5','b7b6 f1b5 c7c6 b5d3 c8b7 e1g1 b8d7 d1e1 b4e7 e3e4 d5e4 c3e4 e8g8 a1d1 d8c7 e4f6 d7f6 e1e2 c6c5 d4c5 c7c5 f1e1 a8c8 c2c3 c5h5 d3a6 b7f3 e2f3 h5f3 g2f3 c8d8 d2f4 d8d1 e1d1 f8d8 d1d8 e7d8 f4g3 f6d5 c3c4');
INSERT INTO evaluations VALUES(52932,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(52932,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(52932,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(52932,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(52932,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(52932,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(52932,36,'cp 95','d2d4 c8g4 h2h3 g4f3 d1f3 e7e6 c1e3 g8f6 a2a3 f8e7 f1d3 b8d7 e1c1 a8c8 c1b1 f6d5 c3d5 c6d5 f3e2 d7b6 e3c1 e8g8 e2h5 g7g6 h5e2 b6c4 f2f4 d8a5 f4f5 c4a3 b2a3 e7a3 c1a3 a5a3 e2e5 f8e8 f5g6 h7g6');
INSERT INTO evaluations VALUES(54416,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(54416,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(54416,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(54416,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(54416,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(54416,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(54416,36,'cp 198','b2b3 c4b3 a2b3 c8d7 f1d3 a7a6 e1g1 b8c6 c2c3 a8c8 b1a3 g8e7 c1g5 d8c7 f1e1 c6a7 g5d2 a7b5 a3b5 d7b5 d3c2 e7c6 e1e3 g7g6 f3e1 f8h6 e3f3 h6d2 d1d2 c6e7 h2h4');
INSERT INTO evaluations VALUES(54409,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(54409,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(54409,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(54409,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(54409,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(54409,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(54409,36,'cp 203','c4b3 a2b3 c8d7 c2c3 g8h6 b1a3 h6f5 f1d3 b8c6 e1g1 f8e7 g1h1 a7a6 a3c2 c6a5 c2e1 d8b6 a1b1 a8c8 b3b4 a5c4 g2g4 f5h6 f1g1 b6d8 c1h6 g7h6 d1c1 e7g5 f3g5 h6g5 f2f4 h7h6 e1f3 d7b5 g1g3 f7f6 e5f6 d8f6 f4g5');
INSERT INTO evaluations VALUES(1461,30,'cp -136',NULL);
INSERT INTO evaluations VALUES(1461,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(1461,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(1461,33,'cp -138',NULL);
INSERT INTO evaluations VALUES(1461,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(1461,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(1461,36,'cp -148','c5d4 e3d4 c6d4 d1d4 b6d4 f1b5 c8d7 b5d7 e8d7 f3d4 g7g5 e1g1 g8e7 b1d2 e7g6 c2c4 f8c5 d4c2 d5c4 d2c4 b7b5 c4d2 g6e5 a2a4 b5a4 a1a4 h8c8 c2e3 h7h6 h2h4 f7f5 d2c4 e5c4 e3c4 a8b8 c4e5 d7d6 f1e1 b8b3 h4g5 h6g5 e5f7 d6d7 f7g5 b3b2 a4f4 c8e8 g5f7 b2c2 f7e5 d7c7 e1d1');
INSERT INTO evaluations VALUES(1516,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(1516,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(1516,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(1516,33,'cp -129',NULL);
INSERT INTO evaluations VALUES(1516,34,'cp -130',NULL);
INSERT INTO evaluations VALUES(1516,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(1516,36,'cp -138','e3d4 c6d4 d1d4 b6d4 f1b5 c8d7 b5d7 e8d7 f3d4 g7g5 e1g1 g8e7 b1d2 e7g6 c2c4 f8c5 d4c2 d5c4 d2c4 b7b5 c4d2 g6e5 a2a4 b5a4 a1a4 d7e7 d2c4 e5c4 a4c4 h8c8 c2d4 c5d6 d4c6 e7d7 f1c1 c8c7 g1f1 a7a6 g2g3 a8c8 c6a5 c7c5 c4c5 c8c5 c1a1 d6c7 a5b7 c5c6 f1g2 f7f5');
INSERT INTO evaluations VALUES(1497,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(1497,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(1497,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(1497,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(1497,34,'cp -153',NULL);
INSERT INTO evaluations VALUES(1497,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(1497,36,'cp -151','c6d4 d1d4 b6d4 f1b5 c8d7 b5d7 e8d7 f3d4 g7g5 e1g1 g8e7 b1d2 e7g6 c2c4 f8c5 d4c2 d5c4 d2c4 b7b5 c4d2 g6e5 a2a4 b5a4 a1a4 d7e7 d2c4 e5c4 a4c4 h8c8 c2d4 c5d6 d4c6 e7d7 c6a5 c8c4 a5c4 d6c5 f1c1 d7e7 c4e5 c5d6 e5c4 a8c8 c1a1 d6c5 g1f1 h7h5 a1a6 c8c7 h2h3 h5h4 f1e2');
INSERT INTO evaluations VALUES(1427,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(1427,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(1427,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(1427,33,'cp -153',NULL);
INSERT INTO evaluations VALUES(1427,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(1427,35,'cp -138',NULL);
INSERT INTO evaluations VALUES(1427,36,'cp -140','d1d4 b6d4 f1b5 c8d7 b5d7 e8d7 f3d4 g7g5 e1g1 g8e7 b1d2 e7g6 c2c4 f8c5 d4c2 d5c4 d2c4 b7b5 c4d2 g6e5 a2a4 b5a4 a1a4 d7e7 d2c4 e5c4 a4c4 h8c8 c2d4 c5d6 d4c6 e7d7 f1c1 c8c7 h2h3 h7h5 g2g4 h5h4 g1g2 f7f6 c6a5 c7c4 a5c4 d6b4 c1d1 d7e7');
INSERT INTO evaluations VALUES(29613,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(29613,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(29613,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29613,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29613,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(29613,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(29613,36,'cp 0','e2e3 e7e6 f1d3 f5g6 e3e4 d5e4 d3e4 g6e4 c3e4 d8d5 e4f6 g7f6 g5f6 d5g2 d1f3 h8g8 f6h4 g2f3 g1f3 b8c6 c2c3 c6e7 e1c1 e7f5 h4g3 c7c6 d1e1 a8d8 c1c2 d8d5 a2a4 f8d6 h1g1 f5g3 f2g3 d5a5 c2b3 a5f5 f3d2 h7h5 d2e4');
INSERT INTO evaluations VALUES(29610,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(29610,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(29610,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(29610,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29610,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(29610,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(29610,36,'cp -7','e7e6 f1d3 f5g6 e3e4 d5e4 d3e4 g6e4 c3e4 d8d5 e4f6 g7f6 g5f6 d5g2 d1f3 h8g8 f6h4 g2f3 g1f3 b8c6 e1c1 f8e7 h1g1 g8g1 d1g1 a8d8 g1g8 e8d7 g8d8 e7d8 h4g3 f7f6 c2c3 c6e7 c1c2 e7g6 c3c4 d8e7 c2d3 e7d6 c4c5 d6e7 b2b4 a7a6 a2a3');
INSERT INTO evaluations VALUES(30145,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30145,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30145,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30145,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30145,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30145,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30145,36,'cp -6','f1d3 f5g6 e3e4 d5e4 d3e4 g6e4 c3e4 d8d5 e4f6 g7f6 g5f6 d5g2 d1f3 h8g8 f6h4 g2f3 g1f3 b8c6 c2c3 c6e7 h4e7 e8e7 e1e2 f8h6 a1g1 h6f4 a2a4 f7f6 h2h3 h7h5 f3e1 a7a5 e1g2 f4h6 g2e3 b7b6 g1g3 a8d8 g3g8');
INSERT INTO evaluations VALUES(46893,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(46893,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46893,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(46893,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(46893,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46893,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(46893,36,'cp 27','c2c4 e7e6 b1c3 f8d6 f4d6 d8d6 e2e3 e8g8 a1c1 a7a6 c4d5 e6d5 f1d3 c8g4 e1g1 f8e8 d1c2 b8d7 f3h4 c7c5 d4c5 d7c5 c3b5 a6b5 c2c5 d6c5 c1c5 b5b4 c5b5 d5d4 e3d4 a8a2 b5b4 g4e2 d3e2 e8e2 b4b7 a2b2 b7b2 e2b2');
INSERT INTO evaluations VALUES(53133,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(53133,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53133,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53133,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(53133,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(53133,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(53133,36,'cp 36','f2f3 b8c6 d4f2 g8f6 b1c3 d6d5 e4d5 f6d5 c1d2 c8f5 e1c1 d5b4 d2e1 d8b6 d1d2 a8d8 f2b6 a7b6 d2d8 e8d8 e1f2 d8c8 f2b6 g7g6 f3f4 h7h5 f1b5 b4c2 g1f3 f8h6 g2g3 e7e5 f3g5 e5f4 g5f7 f4g3 f7h6 h8h6 h2g3 c2d4');
INSERT INTO evaluations VALUES(54809,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(54809,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(54809,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(54809,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(54809,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(54809,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(54809,36,'cp 39','d2d4 g8f6 d1e2 b8c6 g1f3 c8g4 c2c3 e7e5 d4d5 c6b8 h2h3 g4d7 c4b3 f8g7 c3c4 e8g8 b1c3 a7a5 b3c2 b8a6 a2a3 f6h5 c1e3 h5f4 e2f1 c7c6 g2g3 c6d5 c4d5 f4h5 f1e2 a8c8 f3d2 a6c5 a3a4 d7e8 e2f3 f7f5 e4f5 e5e4 d2e4 f8f5 f3d1 c5e4 c2e4 c8c3 b2c3 g7c3 e3d2 c3a1 d1a1');
INSERT INTO evaluations VALUES(54810,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(54810,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(54810,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54810,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(54810,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(54810,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54810,36,'cp 44','g8f6 d1e2 b8c6 g1f3 c8g4 c2c3 e7e5 d4d5 c6b8 h2h3 g4d7 c4b3 f8g7 c3c4 a7a5 b1c3 e8g8 b3c2 b8a6 a2a3 c7c6 d5c6 b7c6 e1g1 d8e7 c1e3 f8b8 a1b1 b8b7 e2d2 a6c7 f1d1 g7f8 b2b4 a5b4 a3b4 a8b8 c3a4');
INSERT INTO evaluations VALUES(46650,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(46650,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(46650,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(46650,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46650,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(46650,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(46650,36,'cp 37','d1e2 b8c6 g1f3 c8g4 c2c3 e7e5 d4d5 c6b8 h2h3 g4d7 c4b3 f8g7 c3c4 e8g8 b1c3 a7a5 b3c2 b8a6 c1e3 a6b4 c2b1 b4a6 g2g4 c7c6 f3d2 a5a4 a2a3 a8c8 b1d3 d8a5 e1g1 h7h5 g4g5 f6h7 h3h4 a6c5 f2f3 c6d5');
INSERT INTO evaluations VALUES(38762,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(38762,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(38762,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(38762,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(38762,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(38762,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(38762,36,'cp 16','e6d5 c1f4 c7c5 d4c5 g8f6 e2e3 e8g8 f1d3 b8c6 g1f3 b4c5 e1g1 d5d4 e3d4 c6d4 h2h3 c8e6 f1e1 a8c8 f4e5 d4c6 e5f6 d8f6 c3e4 f6e7 e4c5 e7c5 a1c1 c5d5 a2a3 g7g6 d3e4 d5d1 e1d1 f8e8 f3g5 h7h6 e4c6 c8c6');
INSERT INTO evaluations VALUES(41190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41190,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41190,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41190,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41190,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(41190,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(41190,36,'cp -4','e8g8 f1e2 c7c6 e1g1 b7b6 c1g5 c8b7 f3d4 a7a6 e2f3 c6d5 e4d5 b8d7 d4c6 b7c6 d5c6 d7e5 g5f6 g7f6 c3d5 a8c8 f3e2 b6b5 a2a4 e5c6 d5f6 e7f6 a4b5 a6b5 e2b5 d8b6 c2c4 c6d4 a1a6 b6c5 b2b4 c5e5 d1d2 d4b5 c4b5 c8c4 d2d6 e5d6 a6d6 f8b8 d6f6 b8b5 f1d1 c4b4 g1f1 b4b2');
INSERT INTO evaluations VALUES(36454,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36454,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36454,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36454,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36454,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(36454,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(36454,36,'cp 1','f1e2 c7c6 e1g1 b7b5 d5c6 b5b4 c3d5 b8c6 a2a3 b4a3 a1a3 f6e4 e2b5 c8b7 d1e2 e7e6 e2e4 e6d5 e4d5 d8b6 b5c6 b6c6 d5c6 b7c6 a3a5 f8b8 c2c3 a7a6 f1d1 b8b5 a5a1 c6f3 g2f3 d6d5 a1a2 a6a5 g1g2 g7f6 c1e3 a8d8 h2h3 g8f8 d1a1 d5d4 c3d4 f6d4 a2a5 b5a5 a1a5 d4b2 e3h6 f8e7 h6e3');
INSERT INTO evaluations VALUES(51550,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(51550,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(51550,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(51550,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(51550,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(51550,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(51550,36,'cp 224','d4d5 a7a6 b1c3 d7d6 d5e6 f7e6 f3g5 f8e7 d1f3 e7g5 f3h5 g7g6 h5g5 d8g5 c1g5 b8c6 e1c1 e8d7 g5f4 e6e5 c3d5 a8b8 f4g3 g8e7 f2f4 b6b5 f4e5 c6e5 f1e2 e7c6 h1f1 c8b7 f1f6 b8e8 g3e5 c6e5 d5f4 d7c7 d1d6 e5d7 f6e6 e8e6 d6e6 c7d8 c1d2 h8e8 e6e8 d8e8 d2e3 b7c6 f4e6 e8e7 e6c7');
INSERT INTO evaluations VALUES(51553,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(51553,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(51553,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(51553,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(51553,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(51553,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(51553,36,'cp 217','a7a6 b1c3 b6b5 c1f4 g8f6 e4e5 f6d5 c3d5 e6d5 d1d5 b8c6 e1c1 d8e7 f1d3 e7e6 d3e4 e6d5 e4d5 f8e7 f3g5 e8g8 g5f7 f8f7 e5e6 d7e6 d5c6 a8a7 f4b8 c8b7 c6e8 f7f2 b8a7 b7e4 d1d7 e7g5 c1d1 f2g2 h2h4 e4f3 d1e1 g5f6 e8f7 g8h8 a7c5 g2c2');
INSERT INTO evaluations VALUES(52404,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52404,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52404,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52404,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(52404,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(52404,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52404,36,'cp 16','d7d5 f1g2 c8g4 e1g1 g8f6 h2h3 g4h5 c2c4 e7e6 d2d4 b8d7 c4d5 e6d5 f3h4 h5g6 h4g6 h7g6 d1b3 d8b6 b3d3 f8e7 b1c3 e8g8 c1g5 f8e8 h3h4 d7f8 e2e3 f8h7 a1c1 h7g5 h4g5 f6d7 f2f4 e7b4 g2f3');
INSERT INTO evaluations VALUES(14390,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(14390,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(14390,32,'cp -126',NULL);
INSERT INTO evaluations VALUES(14390,33,'cp -129',NULL);
INSERT INTO evaluations VALUES(14390,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(14390,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(14390,36,'cp -142','f1b5 c8d7 g1e2 a7a6 b5a4 g7g6 e1g1 f8g7 c1f4 b7b5 a4b3 e8g8 a1d1 c6a5 h2h3 d7c6 d4d5 c6b7 f4e5 d8d7 c3e4 f6e4 e5g7 g8g7 f3e4 d7d6 e4e3 d6b6 e3b6 c7b6 d5d6 a5b3 a2b3 f8e8 e2f4 b7c6 f4d3 e7d6 d3b4 c6b7 d1d6 e8e2');
INSERT INTO evaluations VALUES(14174,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(14174,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(14174,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(14174,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(14174,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(14174,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(14174,36,'cp -147','c8d7 g1e2 a7a6 b5a4 g7g6 c1f4 b7b5 a4b3 f8g7 h2h3 e8g8 a2a3 a8b8 e1g1 c6a5 b3a2 a5c4 b2b4 d7e6 a1d1 e6d5 c3d5 d8d5 f4c7 b8c8 c7f4 d5f3 f1f3 e7e6 c2c3 c4a3 a2b3 a3c4 g1h1');
INSERT INTO evaluations VALUES(25578,30,'cp -142',NULL);
INSERT INTO evaluations VALUES(25578,31,'cp -151',NULL);
INSERT INTO evaluations VALUES(25578,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(25578,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(25578,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(25578,35,'cp -138',NULL);
INSERT INTO evaluations VALUES(25578,36,'cp -128','g1e2 a7a6 b5a4 g7g6 c1f4 b7b5 a4b3 f8g7 e1g1 e8g8 a1d1 c6a5 h2h3 a5b3 a2b3 b5b4 c3e4 f6e4 f3e4 d7f5 e4e3 a6a5 d1d2 d8d5 e3e7 f8e8 e7c5 d5c5 d4c5 g7b2 f4c7 f5e4 f1d1 h7h6 c7d6 b2g7 e2d4 g7d4 d2d4 e4c2');
INSERT INTO evaluations VALUES(32441,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32441,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(32441,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(32441,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(32441,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(32441,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(32441,36,'cp 43','e5d4 f3d4 g8f6 c1g5 e6e5 d4f5 g7g6 f5e3 f8e7 e1g1 f6h5 g5h6 e7f8 h6f8 h8f8 a2a4 b8c6 a4a5 d8d7 f2f3 e8c8 b1c3 c6d4 a5a6 b7b6 b2b4 h5f4 c3d5 f4d5 e3d5 c7c6 d5e3 c8b8 g1h1 d4e6 c2c3 h7h5 c3c4 e6d4 h2h3 g6g5 d1d2 b8a8');
INSERT INTO evaluations VALUES(32450,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32450,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(32450,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(32450,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(32450,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(32450,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(32450,36,'cp 37','f3d4 g8f6 c1g5 e6e5 d4f5 g7g6 f5e3 f8e7 e1g1 f6h5 g5h6 e7f8 h6f8 h8f8 b1c3 b8c6 a2a4 d8d7 a4a5 c6d4 a5a6 b7b6 f2f3 f8f7 c3d5 c7c6 c2c3 d4e6 d5b4 a8d8 d1d2 e8f8 a1d1 d7c7 g2g3 h5f6 e3c4 f8g7 g1h1 d6d5 b4c6 c7c6');
INSERT INTO evaluations VALUES(32448,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(32448,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(32448,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32448,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(32448,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(32448,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(32448,36,'cp 27','g8f6 d1e2 e6e5 d4f3 b8c6 b1c3 d8d7 c1g5 f8e7 e2c4 c6a5 c4d3 a5c6 e1g1 e8g8 a1d1 g8h8 c3d5 a8e8 g5f6 e7f6 d3b5 e8b8 b2b4 d7f7 a2a4 a7a6 b5c4 b8c8 c2c3 c6d8 d5b6 c8b8 c4f7 f8f7 b6d5 d8e6 h2h4 f6e7 h4h5 c7c6 d5e3 e6f4');
INSERT INTO evaluations VALUES(35130,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(35130,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(35130,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(35130,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(35130,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(35130,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(35130,36,'cp 30','b1c3 b8c6 c3d5 f6d6 c2c3 g8f6 d5f6 d6f6 f3f6 g7f6 g1e2 h7h5 h2h4 d7d5 e4d5 c6e7 c3c4 c7c6 e2c3 c8f5 g2g3 a8d8 f1g2 c6d5 c3d5 e7d5 g2d5 f8c5 e1e2 f5g4 f2f3 g4e6 d5e6 f7e6 d2d3 h8g8 h1h3 b7b6 c1e3 a7a5 a2a3 a5a4 a1b1 e8e7');
INSERT INTO evaluations VALUES(43212,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43212,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43212,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43212,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43212,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43212,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43212,36,'cp 0','c7c5 e2e3 d8b6 d1c1 e8g8 f1e2 b8c6 c2c3 d7d5 a2a4 c8f5 b1d2 c5d4 e3d4 c6b4 c3b4 f8c8 c1d1 f5c2 d1c1');
INSERT INTO evaluations VALUES(55370,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(55370,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(55370,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(55370,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(55370,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(55370,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(55370,36,'cp 107','f1d3 f8g7 b1c3 a7a6 c1g5 g8e7 d1d2 h7h6 g5e3 b7b6 e1g1 b8d7 c3e2 c8b7 c2c4 d7f6 d4d5 f6g4 e3d4 e6e5 d4c3 g4f6 a2a4 c7c6 e2g3 f6d7 a4a5 c6d5 c4d5 b6b5 f1c1 a8c8 b2b4 e8g8 c3b2 c8c1 a1c1 f7f5 e4f5 e7f5');
INSERT INTO evaluations VALUES(14139,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(14139,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(14139,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(14139,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(14139,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(14139,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(14139,36,'cp 70','b1d2 e7e6 c2c4 f8d6 f4g3 e8g8 f1d3 c6b4 d3e2 c7c5 d4c5 d6c5 a2a3 b4c6 b2b4 c5d6 e1g1 d6g3 h2g3 d8e7 d1c2 a7a6 a1d1 d5c4 d2c4 e6e5 c4b6 a8b8 f3d2 c8e6 e2d3 f8d8 d2e4 e6g4 d1c1 f6d7 b6d7 g4d7 e4c5 d7g4');
INSERT INTO evaluations VALUES(14140,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(14140,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14140,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(14140,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(14140,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(14140,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(14140,36,'cp 51','e7e6 f4g3 f8d6 c2c4 e8g8 a2a3 c6e7 c4c5 d6g3 h2g3 e7c6 f1b5 c8d7 e1g1 a7a6 b5c6 d7c6 f3e5 d8e8 d1c2 f6e4 f1d1 a6a5 b2b3 f7f6 e5c6 e8c6 d2e4 d5e4 d1b1 b7b5 b3b4 c6d5 b4a5 a8a5 a3a4');
INSERT INTO evaluations VALUES(14554,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(14554,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14554,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(14554,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(14554,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14554,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14554,36,'cp 55','c2c4 f8d6 f4g3 e8g8 f1d3 c6b4 d3e2 c7c5 d4c5 d6c5 a2a3 b4c6 d1c2 c5d6 b2b4 d6g3 h2g3 d8e7 a1d1 c8d7 e3e4 a8c8 e4e5 f6g4 c2b2 f7f6 e5f6 e7f6 b2f6 g4f6 e1g1 c6e7 d1c1 c8c7 c4d5 c7c1 f1c1 f6d5 f3e5 f8c8 c1c8 d7c8 d2b3');
INSERT INTO evaluations VALUES(16807,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16807,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(16807,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16807,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(16807,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(16807,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(16807,36,'cp 14','e4c5 d3e2 d7d6 c1g5 d8d7 e5d6 f8d6 b1c3 c5e6 g5e3 d6b4 d1d7 c8d7 e1c1 e8c8 c3d5 b4c5 e2c4 c5e3 d5e3 h7h6 h1e1 e6g5 f3d4 c6d4 d1d4 b7b5 c4b3 h8e8 d4f4 f7f6 h2h4 g5e4 g2g3 e4d6 h4h5 a7a5 e1d1');
INSERT INTO evaluations VALUES(16763,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(16763,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(16763,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(16763,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(16763,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(16763,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(16763,36,'cp 12','b1c3 c5d3 d1d3 d7d6 c1f4 d6e5 f3e5 d8d3 e5d3 c8f5 f4c7 a8c8 c7f4 c6b4 e1c1 b4d3 c2d3 f7f6 c1d2 e8f7 d2e2 f8e7 d1c1 a7a6 c3e4 h8e8 f2f3 f5d7 e2f2 f6f5 e4d6 e7d6 f4d6 d7b5 h1d1 e8d8 c1c8 d8c8 d6f4 c8c2 d1d2 c2d2 f4d2 b5d3 a2a3 h7h5 d2c3');
INSERT INTO evaluations VALUES(43207,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(43207,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43207,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(43207,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(43207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43207,36,'cp -6','c7c5 g1f3 e8g8 f1e2 d8b6 d1c1 b8c6 c2c3 d7d5 e1g1 c8f5 b1d2 a8c8 d2b3 f6e4 f1d1 c5d4 e3d4 a7a5 f4e3 b6d8 a2a4 e4d6 f3h4 f5e4 b3c5 b7b6 c5e4 d5e4 e2a6 c8a8 a6b5 a8c8');
INSERT INTO evaluations VALUES(53802,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(53802,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(53802,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(53802,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(53802,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(53802,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(53802,36,'cp 67','c2c4 f8e7 b1c3 g8f6 h2h3 b8a6 c1e3 e8g8 g1f3 c8d7 f1e2 c6d5 c4d5 a8c8 f3d2 a6c5 a1c1 a7a5 e1g1 a5a4 b2b4 a4b3 a2b3 d8a5 d1c2 a5b4 c3a2 b4a3 c1a1 b7b5 c2b1 c8a8 f1c1 b5b4 e2f1 f8b8 e3c5 d6c5 d2c4 a3a7 b1b2 d7b5 b2e5 b5c4 f1c4');
INSERT INTO evaluations VALUES(53799,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(53799,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(53799,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(53799,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(53799,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(53799,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(53799,36,'cp 71','f8e7 b1c3 g8f6 g1f3 e8g8 h2h3 c8d7 c1e3 b8a6 f1e2 c6d5 c4d5 a8c8 f3d2 a6c5 e1g1 f6e8 b2b4 c5a6 d1b3 e7g5 c3d1 g5e3 d1e3 d8g5 g1h1 g5f4 a2a4 f7f5 e4f5 f4b4 b3b4 a6b4 d2e4 b4c2 e3c2 c8c2 e2d3 c2b2 g2g4 b7b6 f1e1 e8f6 e4d6 b2f2 e1e5');
INSERT INTO evaluations VALUES(12498,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(12498,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(12498,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(12498,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(12498,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(12498,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(12498,36,'cp 54','b1c3 d7d5 e4e5 f6e4 f1d3 d8a5 e1g1 e4c3 b2c3 a5c3 a1b1 h7h6 d1e2 c3a5 c1d2 a5d8 f1c1 a7a5 c1c3 f8b4 c3b3 b4d2 e2d2 a5a4 b3b6 e8g8 d3c2 f7f6 e5f6 d8f6 b1e1 c6a5 d2d3 f6f5 d3e2 f5f4 b6b1 a5c4 c2d3 c4a3 b1c1 a3c4 d3c4 d5c4 c1c4');
INSERT INTO evaluations VALUES(12495,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(12495,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(12495,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(12495,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(12495,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(12495,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(12495,36,'cp 73','d7d5 e4e5 f6e4 f1d3 d8a5 e1g1 e4c3 b2c3 a5c3 a1b1 h7h6 d1e2 c3a5 c1d2 a5d8 f1c1 a7a5 b1b3 f8b4 h2h4 c8d7 c1b1 b4d2 e2d2 c6b4 d3f1 d7a4 b3a3 b7b5 a3c3 a8c8 a2a3 b4c6 f1b5 a4b5 b1b5 e8g8 c3c5 c8c7 g2g3 d8a8 g1g2 c7b7 h4h5 f8b8 b5b7 a8b7 d2d3');
INSERT INTO evaluations VALUES(14645,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(14645,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(14645,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(14645,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(14645,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(14645,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(14645,36,'cp 181','d4e5 d6e5 d1d8 e8d8 c4f7 f8b4 c1d2 c8g4 f7c4 a7a6 f3h4 c6d4 a1c1 b7b5 c4d5 c7c6 h4g6 h8h7 d5b3 b4d6 f2f3 g4h5 g6h4 c6c5 b3d5 a8a7 h4f5 a7d7 f5d6 d7d6 a2a3 h7h8 d2e3 f6d5 c3d5 h5f7 d5c3 d8c7 e1f2 h8e8 b2b4 c5b4 a3b4 c7b7');
INSERT INTO evaluations VALUES(14604,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(14604,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(14604,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(14604,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(14604,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(14604,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(14604,36,'cp 173','d6e5 d1d8 e8d8 c4f7 f8b4 c1d2 c8g4 f7c4 a7a6 f3h4 c6d4 a1c1 b7b5 c4d5 c7c6 h4g6 h8h7 d5f7 b4d6 f2f4 e5f4 e4e5 a8a7 f7g8 f6g8 e5d6 g4f5 g6e5 g7g5 g2g4 f5e6 e1g1 d8c8 c1e1 a7d7 c3e4 e6d5 c2c3 d4c2 e1e2 d5e4 e5d7 c8d7 e2e4 g8f6');
INSERT INTO evaluations VALUES(14572,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(14572,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(14572,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(14572,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(14572,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(14572,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(14572,36,'cp 180','d1d8 e8d8 c4f7 f8b4 c1d2 c8g4 f7c4 a7a6 e1c1 d8c8 h1e1 b4d6 c4e2 d6c5 d2e3 c5e3 f2e3 b7b5 e1f1 c8b7 h2h3 g4e6 a2a3 a8f8 f3h4 h8g8 h4g6 f8d8 d1d8 g8d8 g6h4 d8g8 h4f5 f6d7 f1d1 d7f6 g2g4 f6d7 b2b3 c6d8 c3d5');
INSERT INTO evaluations VALUES(2248,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(2248,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(2248,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(2248,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(2248,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(2248,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(2248,36,'cp 187','e8d8 c4f7 f8b4 c1d2 c8g4 f7c4 a7a6 e1c1 d8c8 h2h3 g4f3 g2f3 b7b5 c4d3 b4c3 b2c3 f6h5 d2e3 h8f8 a2a4 h5f4 d3f1 f8d8 d1d8 c6d8 h1g1 d8e6 e3f4 e6f4 g1g7 b5a4 g7e7 c8b7 e7e5 a8g8 e5f5 g8g1 f5f4 g1f1 c1b2 f1f2 f4f5');
INSERT INTO evaluations VALUES(19794,30,'cp 305',NULL);
INSERT INTO evaluations VALUES(19794,31,'cp 301',NULL);
INSERT INTO evaluations VALUES(19794,32,'cp 288',NULL);
INSERT INTO evaluations VALUES(19794,33,'cp 301',NULL);
INSERT INTO evaluations VALUES(19794,34,'cp 305',NULL);
INSERT INTO evaluations VALUES(19794,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(19794,36,'cp 308','e4d5 c6d4 f3d4 c5d4 d1d4 g8f6 b1c3 g7g6 c1f4 f8g7 f4e5 e8g8 f1e2 h7h5 h2h3 d8b6 d4b6 a7b6 e1c1 c8f5 h1e1 f8d8 e2f3 b6b5 d1d4 a8c8 d4b4 g7h6 e5f4 g6g5 f4e3 g5g4 e3h6 g4f3 g2f3 f6d5 b4b5 e7e6 c3e4 c8c7 h6d2 d8c8 c2c3');
INSERT INTO evaluations VALUES(19767,30,'cp 295',NULL);
INSERT INTO evaluations VALUES(19767,31,'cp 297',NULL);
INSERT INTO evaluations VALUES(19767,32,'cp 306',NULL);
INSERT INTO evaluations VALUES(19767,33,'cp 296',NULL);
INSERT INTO evaluations VALUES(19767,34,'cp 309',NULL);
INSERT INTO evaluations VALUES(19767,35,'cp 312',NULL);
INSERT INTO evaluations VALUES(19767,36,'cp 305','c6d4 f3d4 c5d4 d1d4 g8f6 b1c3 g7g6 c1f4 f8g7 f4e5 e8g8 f1e2 h7h5 h2h3 g7h6 g2g4 f6d7 e5f4 h6g7 d4e3 d7f6 e1c1 d8b6 g4h5 b6e3 f4e3 f6h5 e2h5 g6h5 d5d6 e7d6 c3d5 g8h7 d5c7 a8b8 e3a7 c8f5 a7b8 f8b8 c7d5 b8c8 c2c3 c8e8 d5e3 f5e4 h1g1');
INSERT INTO evaluations VALUES(22647,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(22647,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(22647,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(22647,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(22647,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(22647,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(22647,36,'cp 29','g1f3 g8f6 e1g1 f6e4 f1e1 e4d6 f3e5 f8e7 b5f1 c6e5 e1e5 e8g8 d2d4 e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 a7a5 f1d3 g7g6 b1d2 d7d5 d2f3 e8g7 a2a4 c8f5 d3f5 g7f5 e1e2 c7c6 a1e1 f5g7 g2g3 d8d7 f3e5 f6e5 f4e5 a8e8 g1g2 g7e6 b2b3 d7e7');
INSERT INTO evaluations VALUES(32742,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(32742,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(32742,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(32742,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(32742,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(32742,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(32742,36,'cp 44','f1e2 d7b5 e2b5 b6b5 b1a3 b5d7 a3c2 c5d4 c3d4 g8e7 c1d2 e7f5 a2a4 a7a5 g2g4 f5e7 c2a3 e7c6 h2h4 f8b4 a3b5 e8g8 g4g5 c6e7 h4h5 b8c6 h5h6 g7g6 e1f1 e7f5 h1h3 b4d2 d1d2 a8c8 f3h4 c6e7 h4f5 e7f5 f1g1');
INSERT INTO evaluations VALUES(32741,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(32741,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(32741,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(32741,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(32741,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(32741,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(32741,36,'cp 33','d7b5 e2b5 b6b5 b1a3 b5d7 a3c2 b8c6 e1g1 c5d4 c3d4 f8e7 h2h4 e7d8 h4h5 h7h6 a2a4 a7a5 c2a3 c6a7 d1b3 g8e7 c1d2 e7c6 a3b5 d8e7 b5a7 a8a7 b3b5 e8g8 f1c1 e7b4 d2b4 a5b4 b2b3 a7a5 b5d3 a5a8 c1c2 f8c8 a1c1 c6a5 g2g4 c8c2');
INSERT INTO evaluations VALUES(32803,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(32803,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(32803,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(32803,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(32803,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(32803,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(32803,36,'cp 53','e2b5 b6b5 b1a3 b5d7 a3c2 c5d4 c3d4 b8c6 e1g1 g8e7 c1d2 h7h6 c2e3 e7c8 a1c1 f8e7 h2h4 e7d8 h4h5 a7a5 a2a4 c8e7 g2g3 e8g8 g1g2 a8c8 f1h1 c6b4 c1c8 d7c8 d2b4 a5b4 b2b3 c8c3');
INSERT INTO evaluations VALUES(47220,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(47220,31,'cp -123',NULL);
INSERT INTO evaluations VALUES(47220,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(47220,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(47220,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(47220,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(47220,36,'cp -133','d5e4 f3g5 c8g4 f1e2 g4f5 d2d3 e4d3 c2d3 h7h6 g5e4 b8c6 c1e3 e7e6 d1b3 a8b8 e4f6 d8f6 e2f3 c6e5 f3e4 f8d6 e3a7 b8a8 a7d4 e8g8 e1g1 f5e4 d3e4 e5f3 g2f3 f6d4 c3b5 d4f6 b5d6 c7d6 g1h1 a8a5 b3c3 f6c3 b2c3 f8c8 f1c1 b7b5 h1g2 g8f8 a1b1 c8c5 b1b2 f8e7 h2h4 e7f6 c1d1');
INSERT INTO evaluations VALUES(47404,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(47404,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(47404,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(47404,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(47404,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(47404,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(47404,36,'cp -126','f3g5 c8g4 f1e2 g4f5 d2d3 e4d3 c2d3 h7h6 g5e4 f6e4 d3e4 d8d1 e2d1 f5h7 f2f3 b8d7 c3b5 e8d8 c1e3 c7c6 b5a3 e7e5 d1b3 h7g8 e1c1 d8c7 d1d2 f8b4 d2c2 a7a6 h1d1 f7f6 b3g8 a8g8 a3b1 b7b5 h2h4 a6a5 a2a3 b4d6 h4h5');
INSERT INTO evaluations VALUES(47920,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(47920,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47920,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47920,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(47920,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47920,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(47920,36,'cp 0','f8e7 f1d3 d6d5 e4e5 f6e4 g5e3 e8g8 e1g1 c7c5 d4b3 b8c6 d3e4 d5e4 d1d8 e7d8 b3c5 c6e5 c5e4 e5c4 e3d4 c8f5 f1e1 d8a5 d4c3 f8e8 c3a5 f5e4 b1d2 c4d2 a5d2 e4c2 a1c1 e8e1 c1e1 a8d8 d2e3 a7a6 e3b6 d8d1 e1d1 c2d1 f2f3 f7f6 a2a3 g8f8 b6e3 d1b3 g1h1 h7h6 e3f4 b7b6');
INSERT INTO evaluations VALUES(40807,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(40807,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40807,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(40807,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40807,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40807,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40807,36,'cp 0','f1d3 d6d5 e4e5 f6e4 g5e3 e8g8 e1g1 c7c5 d4b3 b8c6 d3e4 d5e4 d1d8 e7d8 b3c5 c6e5 c5e4 e5c4 e3d4 c8f5 f1e1 d8a5 d4c3 f8e8 c3a5 f5e4 b1d2 c4d2 a5d2 e4c2 e1e8 a8e8 a1c1 c2e4 c1c7 e4c6 f2f3 e8d8 d2e1 f7f6 g1f2 d8d7 c7c8 g8f7 e1b4 a7a6 c8h8 f7g6 h8c8');
INSERT INTO evaluations VALUES(40806,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40806,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40806,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40806,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40806,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40806,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40806,36,'cp 0','d6d5 e4e5 f6e4 g5e3 e8g8 e1g1 c7c5 d4b3 b8c6 d3e4 d5e4 d1d8 e7d8 b3c5 c6e5 c5e4 e5c4 e3d4 c8f5 f1e1 d8a5 d4c3 a8e8 c3a5 e8e4 e1e4 f5e4 b1d2 c4d2 a5d2 e4c2 f2f3 f7f6 a1c1 c2a4 d2a5 a4c6 g1f2 g8f7 c1e1 a7a6 e1d1 c6b5 d1c1 b5c6');
INSERT INTO evaluations VALUES(49536,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(49536,31,'cp -174',NULL);
INSERT INTO evaluations VALUES(49536,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(49536,33,'cp -191',NULL);
INSERT INTO evaluations VALUES(49536,34,'cp -197',NULL);
INSERT INTO evaluations VALUES(49536,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(49536,36,'cp -190','f6e4 b1c3 e4c3 d2c3 c7c6 d1h5 d8f6 c1g5 f6g6 h5g4 h7h5 g4e2 d7d6 c4d3 c8f5 d3f5 g6f5 e1c1 b8d7 h1e1 f7f6 g5e3 g7g5 f2f3 d7b6 h3f2 e8c8 c1b1 d6d5 e3b6 a7b6 c3c4 d5d4 e1h1 h8g8 g2g4 h5g4 f3g4 f5f4 h1f1 f4e3 e2e3 d4e3 d1d8 c8d8 f2e4');
INSERT INTO evaluations VALUES(49892,30,'cp -177',NULL);
INSERT INTO evaluations VALUES(49892,31,'cp -180',NULL);
INSERT INTO evaluations VALUES(49892,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(49892,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(49892,34,'cp -180',NULL);
INSERT INTO evaluations VALUES(49892,35,'cp -169',NULL);
INSERT INTO evaluations VALUES(49892,36,'cp -169','b1c3 e4c3 d2c3 c7c6 d1h5 d8f6 c1g5 f6g6 h5g4 b8a6 c4d3 d7d5 g4g3 g6e6 e1g1 f7f6 g5e3 e6g4 g3g4 c8g4 f2f4 g4h3 g2h3 f8e7 f4e5 f6e5 f1f5 e7f6 d3e2 a6c7 e2g4 e8g8 f5f2 f6e7 f2f8 g8f8 c3c4 d5c4 a1f1 f8g8 c2c3 c7d5');
INSERT INTO evaluations VALUES(54715,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(54715,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(54715,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(54715,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(54715,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(54715,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(54715,36,'cp 81','e7e6 c2c4 f6f5 b1c3 g8f6 d1c2 f6e4 f1d3 c7c6 c4c5 g7g5 e1g1 f8g7 c3e2 b7b6 b2b4 a7a5 c5b6 a5b4 a1b1 d8b6 d3e4 f5e4 f3g5 h7h6 g5h3 e6e5 d4e5 c8h3 g2h3 g7e5 f2f4 e5g7 g1h1 e8g8 c1b2 a8a2');
INSERT INTO evaluations VALUES(56800,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(56800,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(56800,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(56800,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(56800,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(56800,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(56800,36,'cp 92','c2c4 f6f5 f1e2 a7a5 b1c3 g8f6 e1g1 c7c6 c4c5 b7b6 c5b6 d8b6 b2b3 f8d6 c3a4 b6c7 c1b2 e8g8 a1c1 c8a6 c1c2 a6e2 d1e2 b8a6 a2a3 c7b7 f1c1 b7b3 c2c6 d6a3 c6a6 a8a6 e2a6 b3a4 b2a3 a4a3');
INSERT INTO evaluations VALUES(56791,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(56791,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(56791,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(56791,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(56791,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(56791,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(56791,36,'cp 89','f6f5 f1e2 g8f6 e1g1 a7a5 b1c3 c7c6 b2b3 f8d6 c1b2 e8g8 c3a4 b8d7 f3e5 d8c7 f2f4 f6e4 e2d3 c7d8 d3e4 f5e4 d1e2 a8b8 a1c1 d8e8 c4c5 d6c7 b2c3 e8d8 f1f2 d8e7 c1f1 e7e8 g2g4 d7e5 d4e5 b7b5 c5b6');
INSERT INTO evaluations VALUES(57473,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(57473,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(57473,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(57473,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(57473,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(57473,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(57473,36,'cp 127','c7c6 b1c3 d7d6 f1c4 d8e7 e1g1 g8f6 b2b4 c8e6 c4b3 b8a6 b4b5 a6c5 b5c6 b7c6 c1e3 c5b3 a2b3 a7a5 d1d3 e7b7 c3a4 f8e7 a4b6 a8a6 f3g5 e6g8 b6c4 g8c4 b3c4 h7h6 f1b1 b7c8 g5f3 e8g8 c4c5 d6d5 f3e5 d5e4 d3c4 f6d5 c4e4');
INSERT INTO evaluations VALUES(52218,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(52218,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(52218,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(52218,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(52218,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(52218,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(52218,36,'cp 164','b1c3 d8c7 f1c4 g8f6 a2a3 a7a5 c1e3 b7b5 c4a2 b8a6 f3g5 f8c5 e1g1 d7d6 e3c5 a6c5 g5f7 e8g8 d1d6 c7d6 f7d6 c8e6 f2f3 e6a2 c3a2 f8d8 d6f5 g8f7 a1d1 g7g6 f5e3 f7e7 a2c1 d8d4 c1e2 d4d7 d1d7 f6d7 e2c1 a5a4 f1d1 a8f8 c1a2 e7d8 a2b4 d8c7');
INSERT INTO evaluations VALUES(57553,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(57553,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(57553,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(57553,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(57553,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(57553,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(57553,36,'cp 89','e7e5 e2e3 d7d6 b2b3 c7c6 c1b2 g8h6 b1c3 b8a6 d1d2 f8e7 a1d1 e8g8 f1d3 e5d4 e3d4 f8e8 g1e2 a6b4 d4d5 b4d3 d2d3 c8f5 d3d2 a7a6 d5c6 b7c6 e1g1 d8d7 e2f4 h6f7 c3e2 e7f8 h2h3 f5g6 f4g6 h7g6 e2g3');
INSERT INTO evaluations VALUES(57412,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(57412,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(57412,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(57412,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(57412,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(57412,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(57412,36,'cp 89','e2e3 d7d6 b1c3 f6f5 f1e2 f8e7 g1f3 e5e4 f3d2 g8f6 f2f3 e4f3 d2f3 e8g8 e1g1 c7c6 c1d2 g7g6 e2d3 d6d5 d1b3 d8b6 b3c2 b6d8 c3e2 a7a5 e2f4 e7d6 g1h1 a5a4 b2b4 a4b3 c2b3 a8a3 b3b2 f6e4 c4d5 c6d5 d3e4 f5e4 f3e5');
INSERT INTO evaluations VALUES(57410,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(57410,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(57410,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(57410,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(57410,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(57410,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(57410,36,'cp 88','d7d6 f1e2 f6f5 b1c3 a7a5 a2a3 c7c6 b2b3 g8f6 c1b2 b8d7 d1c2 g7g6 d4e5 d6e5 g1f3 d8e7 c3a4 f8g7 c4c5 e8g8 e1g1 e5e4 f3d4 g8h8 f1d1 f6d5 e2c4 f5f4 c4d5 c6d5 e3f4 f8f4 d4e2 f4f8 b2g7 e7g7 c2d2 d7e5 a4b6');
INSERT INTO evaluations VALUES(5727,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(5727,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(5727,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(5727,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(5727,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(5727,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(5727,36,'cp 21','d2d3 c6a5 c3a4 c5b6 c4d5 g8e7 a4b6 a7b6 a2a4 a5c6 d5b3 e8g8 c2c3 e7g6 c1e3 c8e6 b3e6 f7e6 b2b4 d8f6 d3d4 f6f7 b4b5 c6a5 d1e2 a5b3 a1a3 b3a5 c3c4 e5d4 f3d4 e6e5 d4f5 f7c4 e2g4 c4f7 h2h4');
INSERT INTO evaluations VALUES(16243,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(16243,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(16243,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(16243,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(16243,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(16243,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(16243,36,'cp 167','c2c4 d5b6 f1d3 d7d6 e5d6 c7d6 e1g1 f8e7 f1e1 d6d5 c4c5 b6d7 b1c3 b7b6 d3b5 c8b7 b2b4 a7a6 b5c6 b7c6 c1f4 e7f6 a1c1 e8g8 a2a4 c6b7 f4d6 f6e7 d6e7 d8e7 a4a5 b6a5 b4a5 b7c6 c3a4 a8c8 a4b6 c8c7 h2h3 e7f6 d1d2 c6b5 b6d7 b5d7');
INSERT INTO evaluations VALUES(16241,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(16241,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(16241,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(16241,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(16241,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(16241,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(16241,36,'cp 144','d5b6 f1d3 d7d6 e5d6 c7d6 e1g1 f8e7 f1e1 e6e5 d4e5 e8g8 b1c3 c8g4 h2h3 g4f3 d1f3 c6e5 f3e4 f7f5 e4e2 a8c8 b2b3 f5f4 d3e4 f4f3 g2f3 e7g5 c1e3 g5f4 e2f1 d8h4 a1d1 f8f6 e3f4 h4f4 e1e3 c8f8 c3e2 f4h4 f3f4 e5g6 e4b7 g6f4 e2f4 h4f4 f1g2');
INSERT INTO evaluations VALUES(16225,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(16225,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(16225,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(16225,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(16225,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(16225,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(16225,36,'cp 162','f1d3 d7d6 e5d6 c7d6 e1g1 f8e7 f1e1 e7f6 c1e3 e6e5 b1c3 e8g8 d4e5 c6e5 f3e5 f6e5 c4c5 d6c5 d3h7 g8h7 d1h5 h7g8 h5e5 b6c4 e5h5 c4e3 e1e3 d8f6 c3d5 f6h6 d5e7 g8h7 h5c5 c8e6 a1e1 a8d8 c5b4 f8e8 e7f5 h6g5 f5d6 e6d5 e3g3 e8e1 b4e1 g5h5 g3h3');
INSERT INTO evaluations VALUES(31963,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(31963,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(31963,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(31963,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(31963,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(31963,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(31963,36,'cp 64','g8f6 e2e3 e7e6 f1c4 f8e7 e1g1 c7c6 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 b8d7 c3c4 e8g8 a1b1 b7b6 e3e4 e6e5 d4e5 d8c7 c1b2 d7c5 d3e2 c5e6 f3d4 a8d8 d4f5 e7c5 g2g3 d8d7 e1d1 f8d8 g1g2 h7h6 b1c1 d7d1 c1d1 d8d1 e2d1 b6b5 c4b5 c6b5 d1d5 b5b4 f5d6');
INSERT INTO evaluations VALUES(29715,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(29715,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(29715,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(29715,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(29715,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(29715,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(29715,36,'cp 59','e2e3 c7c6 f1c4 e7e6 e1g1 f8e7 f3h4 f5g6 h4g6 h7g6 d1e2 b7b5 c4d3 b8d7 f1d1 e8g8 b2b3 a8c8 c1b2 a7a6 a1c1 c6c5 d4c5 c8c5 a2a4 b5b4 c3b1 c5c1 b2c1 a6a5 d3b5 d8c7 c1b2 d7c5 b1d2 f6e4 d2e4 c5e4');
INSERT INTO evaluations VALUES(29714,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(29714,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(29714,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(29714,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(29714,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(29714,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(29714,36,'cp 61','e7e6 f1c4 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 e8g8 a1b1 b8d7 c3c4 b7b6 e1d1 c6c5 d4d5 e6d5 c1b2 e7f6 d3d5 f6b2 d5d7 b2f6 g2g3 d8e8 d7e8 f8e8 d1d7 g7g6 g1g2 g8g7 b1e1 e8e7 e1d1 e7d7 d1d7');
INSERT INTO evaluations VALUES(42113,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(42113,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(42113,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(42113,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(42113,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(42113,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42113,36,'cp 20','h2g3 b8d7 c2c4 h7h6 b1c3 e8g8 a2a4 a7a5 d1c2 c7c6 f1d3 d5c4 d3c4 d8e7 e1g1 e6e5 a1d1 e5d4 d1d4 d7c5 c3e2 c8g4 e2f4 c5e6 c4e6 g4e6 f4e6 e7e6 c2d3 f8e8 b2b3 e8e7 f1c1 a8e8 c1d1');
INSERT INTO evaluations VALUES(45093,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(45093,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(45093,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(45093,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(45093,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(45093,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(45093,36,'cp 91','e2e4 e7e5 g1f3 c8g4 c1e3 b8d7 f1e2 f8e7 d4d5 e8g8 h2h3 g4h5 f3d2 h5g6 e1g1 a8c8 a2a4 a7a5 d1b3 d8c7 e2f3 h7h6 g2g3 g8h8 b3d1 c6d5 c4d5 d7c5 d1e2 c7b6 e2b5 b6c7 a1d1 f8e8 f3g2 g6h7 e3c5 c7c5');
INSERT INTO evaluations VALUES(45095,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(45095,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(45095,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(45095,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(45095,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(45095,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(45095,36,'cp 75','e7e5 g1f3 c8g4 c1e3 b8d7 d4d5 f8e7 h2h3 g4h5 f1e2 h5g6 f3d2 e8g8 e1g1 a8c8 e2f3 a7a6 d1e2 c6d5 c4d5 f6e8 a2a4 e7g5 f3g4 g5f4 a4a5 e8f6 g4f3 c8c7 g2g3 f4e3 e2e3 d8e7 f3e2 f8a8 f1c1 a8c8 f2f3 h7h6 g1h2 d7c5 b2b4');
INSERT INTO evaluations VALUES(14804,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(14804,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(14804,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14804,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(14804,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(14804,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(14804,36,'cp 32','e2e3 f8d6 f4g3 e8g8 b1d2 b7b6 f1d3 c8b7 e1g1 c6e7 b2b4 c7c5 b4c5 b6c5 d4c5 d6c5 a1b1 b7c6 f3d4 c6d7 d1e2 d7a4 d3a6 c5d6 g3h4 e7g6 h4f6 g7f6 c3c4 d6h2 g1h2 d8d6 h2g1 d6a6 c4d5 a6e2 d4e2 a4c2 b1b2 c2d3 f1e1 d3e2 e1e2 e6d5');
INSERT INTO evaluations VALUES(21735,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(21735,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(21735,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(21735,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(21735,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(21735,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(21735,36,'cp -27','e5e4 f3d4 c6d4 b2d4 h7h5 d2d3 c8f5 d3e4 d5e4 d1d2 a7a6 d2c3 g8f6 b1d2 f8e7 f1e2 e8g8 h2h3 f6d5 c3b2 f7f6 g2g4 f5h7 c2c4 d5b4 g4h5 d8d7 h5h6 g7g5 c4c5 b4d3 e2d3 e4d3 b2c3 d7e6 e1c1 b7b6 c1b1 f8d8 h1g1 b6c5 d4c5');
INSERT INTO evaluations VALUES(21621,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(21621,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(21621,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(21621,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(21621,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(21621,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(21621,36,'cp -14','f3d4 c6d4 b2d4 h7h5 d2d3 c8f5 d3e4 d5e4 d1d2 a7a6 d2c3 g8f6 b1d2 f8e7 f1e2 e8g8 h2h3 h5h4 e1g1 f8e8 a1d1 f6d5 c3a1 e7f8 a2a3 c7c5 d4b2 d8g5 g1h1 a8d8 c2c4 d5b6 d2b1 f8d6 b1c3 d6c7 d1d8 e8d8 f1d1 g5e7 c3d5 b6d5 d1d5 d8d5');
INSERT INTO evaluations VALUES(21588,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(21588,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(21588,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(21588,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(21588,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(21588,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(21588,36,'cp -24','c6d4 b2d4 c8f5 d2d3 h7h5 d3e4 d5e4 d1d2 a7a6 d2c3 g8f6 b1d2 f8e7 f1e2 f6d5 c3b2 e7f6 a1d1 d8e7 d4f6 d5f6 d2c4 e8g8 b2e5 e7e5 c4e5 g7g6 e1g1 g8g7 a2a4 f8e8 e5c4 b7b5 c4a5 f5d7 a5b7 b5a4 b7c5');
INSERT INTO evaluations VALUES(22177,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(22177,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(22177,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(22177,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(22177,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(22177,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(22177,36,'cp -19','b2d4 c8f5 d2d3 h7h5 d3e4 d5e4 d1d2 c7c6 d2c3 g8f6 b1d2 f8e7 f1e2 e8g8 h2h3 h5h4 c3b2 d8d6 d4e5 d6e6 e2c4 e6d7 e1c1 d7e8 h1g1 b7b5 c4e2 e7a3 b2a3 e8e5 a3b2 e5h2 g1h1 h2c7 h1g1 f8d8 g2g3 h4g3 g1g3 f5g6 h3h4 a7a5 a2a3 a5a4');
INSERT INTO evaluations VALUES(6008,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(6008,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(6008,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(6008,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(6008,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(6008,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(6008,36,'cp -34','f1b5 g8f6 b5c6 b7c6 d2d3 e8g8 b1d2 a7a5 e1g1 a5a4 h2h3 c5b6 d2c4 b6a7 c4e3 d8e8 a1c1 c8e6 c1a1 h7h6 g1h2 a7e3 f2e3 e8d8 b2c3 c6c5 d1e1 f6h7 h2h1 h7g5 e1e2 d8e7 f3g5 h6g5 e2f3 f7f6 f1f2 a4b3 a2b3 a8a1 c3a1');
INSERT INTO evaluations VALUES(5938,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(5938,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(5938,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(5938,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(5938,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(5938,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(5938,36,'cp -39','g8f6 b5c6 b7c6 d2d3 e8g8 b1d2 a7a5 e1g1 a5a4 h2h3 c5b6 d2c4 b6a7 c4e3 a7e3 f2e3 c6c5 d1e1 f8e8 e1g3 h7h6 f3h4 g8h8 g1h2 c8e6 h4f5 e8g8 g3h4 f6d7 h4f2 d8g5 b2c3 a8a6 b3b4 c5b4 c3b4 g7g6 f5g3 c7c5');
INSERT INTO evaluations VALUES(48446,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48446,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48446,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(48446,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48446,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(48446,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48446,36,'cp 11','b7b5 c3b5 f8b4 b5c3 c7c5 d4c5 d8a5 g5f6 g7f6 g1f3 a5c5 f1e2 e8e7 e1g1 c8b7 a1c1 h8d8 f3d2 b8c6 d1e1 b7a6 a2a3 b4c3 c1c3 c6e5 d2c4 a6c4 e2c4 e5c4 b2b3 a8c8 c3c4 c5a3 e4e5 c8c4 e5f6 e7f6 b3c4 d8d4 e1e3 a3c5 e3h6 f6e7 h6h7 c5c4 h7h8 c4c3 h8h6 c3d2');
INSERT INTO evaluations VALUES(48714,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(48714,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(48714,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(48714,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(48714,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(48714,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48714,36,'cp -50','d8e7 d1e2 c8g4 b1c3 c7c6 d2d4 g4f3 g2f3 b8a6 c1g5 e8c8 e1c1 a6c7 c1b1 h7h6 g5h4 e7e2 c3e2 f8e7 c2c3 h8e8 b3c2 g7g6 h4g3 c7e6 h1e1 e7d6 f3f4 e6g7 g3h4 e8e6 e2g3 e6e1 d1e1 d8e8 e1e8 f6e8 f4f5 d6g3 h4g3 g7f5 c2f5 g6f5 g3e5 b7b6 b1c1');
INSERT INTO evaluations VALUES(46424,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(46424,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46424,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46424,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46424,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(46424,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(46424,36,'cp 13','c5c4 c1d2 a7a6 a2a4 b8c6 b2b3 c4b3 d1b3 c8f5 f1e2 c6a5 b3a2 a5c6 a4a5 f8b4 f3e5 e8g8 e1g1 a8c8 e5c6 c8c6 f1c1 f8e8 e2f3 b4c3 c1c3 c6c3 d2c3 f5d3 c3e1 d8d7 a2b2 e8c8 f3d1 d3b5 a1c1 c8c7 h2h3 c7c8 f2f3 h7h6 g1h1 c8c6 c1c6 d7c6');
INSERT INTO evaluations VALUES(46469,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46469,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(46469,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(46469,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(46469,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46469,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(46469,36,'cp 25','c1d2 a7a6 a2a4 b8c6 b2b3 c4b3 d1b3 c8f5 f1e2 c6a5 b3a2 a5c6 a4a5 f8b4 e1g1 e8g8 f3e5 a8c8 e5c6 c8c6 f1c1 f8e8 e2f3 b4c3 c1c3 c6c3 d2c3 f5d3 c3e1 d8d7 f3d1 d3b5 a1c1 e8c8 a2b2 h7h5 h2h3 d7f5 c1c8 f5c8 d1b3 h5h4 f2f3 c8e6 b2c3');
INSERT INTO evaluations VALUES(21882,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(21882,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21882,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21882,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(21882,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21882,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(21882,36,'cp 8','d6e5 f1b5 f8c5 b5c6 b7c6 b1c3 g8e7 c3a4 c5d6 d2d3 e8g8 c1e3 a7a5 d1e2 f7f5 e1c1 e7g6 h1f1 d8e7 e4f5 c8f5 d1e1 a8e8 a4c3 e7d7 h2h3 g6f4 e2d2 f4d5 c3e4 f5e4 d3e4 d6b4 c2c3');
INSERT INTO evaluations VALUES(21920,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(21920,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(21920,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(21920,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(21920,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(21920,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(21920,36,'cp 31','d2d4 c8g4 d4d5 c6d4 b1d2 g7g6 h2h3 d4f3 d2f3 g4f3 d1f3 f8h6 c4c5 h6c1 a1c1 g8f6 f1b5 c7c6 d5c6 d8a5 c1c3 a5b5 f3f6 e8g8 c6b7 b5b7 e1g1 b7b2 f1c1 d6c5 f6e5 c5c4 e5a5 f8c8 c3c4 b2c1 c4c1 c8c1 g1h2 c1c2 h2g3 h7h5 f2f3 h5h4 g3h2');
INSERT INTO evaluations VALUES(21935,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(21935,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(21935,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(21935,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21935,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21935,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21935,36,'cp 31','c8g4 d4d5 c6d4 f1e2 d4e2 d1e2 f8e7 b1c3 g8f6 h2h3 g4d7 c1e3 e8g8 a2a4 h7h6 a4a5 a7a6 a1a3 f6h7 a3b3 a8b8 f3d2 d8e8 e1g1 f7f5 e4f5 d7f5 c3e4 e8g6 f2f3 f5h3 e3a7 h3f5 a7b8 f8b8 b3c3 b8a8 c4c5 d6c5 d2c4');
INSERT INTO evaluations VALUES(46653,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(46653,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(46653,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(46653,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(46653,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(46653,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(46653,36,'cp 59','d2d4 f8g7 f1e2 e8g8 c1e3 a7a5 g1f3 f6g4 e3d2 e7e5 d2g5 f7f6 g5h4 b8c6 d4d5 c6b4 a2a3 b4a6 f3d2 g4h6 f2f3 c8d7 d1c2 b7b6 e1c1 a6c5 g2g4 h6f7 h4f2 g7h6 c1b1 a5a4 h2h4 h6f4 d2f1 c5b3 f1e3 b3d4');
INSERT INTO evaluations VALUES(49640,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(49640,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(49640,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(49640,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(49640,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(49640,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(49640,36,'cp -37','d7d5 e4d5 d8d5 g1f3 e5e4 d1e2 b8c6 f1g2 c8g4 d3e4 d5e4 e2e4 f6e4 e1g1 e8c8 h2h3 g4h5 g3g4 h5g6 f3h4 e4c5 g2c6 b7c6 h4g6 h7g6 g1g2 c5e6 c1e3 f7f5 b1c3 f5f4 e3c1 f4f3 g2f3 h8h3 f3g2 h3h4 f1e1 e6g5 c1g5 h4g4 g2f1 g4g5 a1d1 f8b4 d1d8 c8d8 c3e4 g5a5 c2c3 b4e7');
INSERT INTO evaluations VALUES(48837,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(48837,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(48837,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(48837,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(48837,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48837,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(48837,36,'cp -20','e4d5 d8d5 g1f3 c8g4 f1g2 e5e4 d1e2 b8c6 d3e4 d5e4 e2e4 f6e4 e1g1 e8c8 h2h3 g4h5 g3g4 h5g6 f3h4 e4c5 h4g6 h7g6 g2c6 b7c6 g1g2 c5e6 c1e3 f7f5 b1c3 f5f4 e3c1 f4f3 g2f3 h8h3 f3g2 h3h4 f1e1 e6g5 c1g5 h4g4 g2f1 g4g5 a1d1 f8b4 d1d8 c8d8 c3e4 g5a5 c2c3 b4e7 a2a3 a5b5');
INSERT INTO evaluations VALUES(52716,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(52716,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(52716,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(52716,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(52716,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(52716,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(52716,36,'cp -39','c6d5 d2d4 b8c6 c2c3 g8f6 f3d1 e7e5 d4e5 f6g4 h2h3 g4e5 g1f3 e5f3 d1f3 f8d6 f1e2 e8g8 e1g1 f8e8 f1d1 d6c7 d1d5 c6e5 f3e4 c8d7 c1e3 d8f6 e4f4 f6c6 d5c5 e5d3 c5c6 c7f4 e2d3 e8e3 f2e3 f4e3 g1f1 d7c6 b1a3 e3b6 d3b5 c6d5 a3c4 b6c7 a2a4 d5e6 a1d1 g7g6');
INSERT INTO evaluations VALUES(53430,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(53430,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(53430,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(53430,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(53430,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(53430,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(53430,36,'cp -40','d2d4 b8c6 f1b5 g8f6 h2h3 e7e6 g1e2 f8e7 e1g1 e8g8 f1d1 d8b6 a2a4 c6b4 f3b3 c8d7 b5d7 f6d7 c1f4 b4c6 b3b6 d7b6 b1d2 c6a5 c2c3 f8c8 e2c1 b6c4 d2c4 c8c4 c1d3 a5b3 d3e5 c4c8 a1b1 b3a5 f4h2 g7g5 e5d3');
INSERT INTO evaluations VALUES(9382,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(9382,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(9382,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(9382,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(9382,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(9382,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(9382,36,'cp 50','g1e2 e8g8 e1g1 c5b6 b2b3 c8f5 d2d4 f8e8 d4d5 c6e7 e3e4 f5d7 d1d3 h7h6 h2h3 e7g6 a2a4 a7a5 a1a2 b6c5 c1e3 c5b4 a2c2 c7c6 g1h2 a8c8 f1d1 c6d5 c3d5 f6d5 d3d5 b4c5 e3c5');
INSERT INTO evaluations VALUES(9383,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(9383,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(9383,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(9383,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(9383,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(9383,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(9383,36,'cp 41','c5b6 h2h3 e8g8 e1g1 c6e7 d2d4 c7c6 d4e5 d6e5 b2b3 b6c5 c3a4 c5b4 a2a3 b4a5 c1b2 e7g6 b2c3 a5c3 d1d8 f8d8 e2c3 g6e7 f1d1 c8f5 b3b4 f5c2 d1c1 c2g6 g3g4 h7h6 c1d1 g6c2 d1d8 a8d8 a1c1 c2g6 a4c5 b7b6 c5b3 g6d3');
INSERT INTO evaluations VALUES(19850,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(19850,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(19850,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(19850,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(19850,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19850,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(19850,36,'cp 7','g8f6 c1f4 c8f5 e2e3 e7e6 f1e2 f8b4 f3d2 e8g8 a1c1 a8c8 d1b3 d8d7 e1g1 b4e7 a2a3 a7a6 b3a2 f5g4 e2g4 f6g4 h2h3 g4f6 b2b4 f8d8 d2b3 b7b6 f1d1 c6a7 a2e2 a7b5 c3b5 a6b5 b3d2 c8c1 d1c1 d8c8 c1c8 d7c8 e2b5 c8c1 d2f1 c1a3 b5b6 a3b4 b6b4');
INSERT INTO evaluations VALUES(41369,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(41369,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(41369,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(41369,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(41369,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(41369,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(41369,36,'cp 59','f2f4 e8g8 g1f3 b8c6 e1g1 e7e5 d4d5 c6d4 f4f5 c7c6 c1g5 h7h6 g5f6 g7f6 f3d2 f6g5 g1h1 c8d7 a2a4 g8h8 a4a5 f8g8 f5g6 f7g6 a5a6 b7a6 d5c6 d7c6 a1a6 g8f8 c3d5 g5d2 f1f8 d8f8 d1d2 c6d5 e4d5 a8b8 h2h3 b8b2');
INSERT INTO evaluations VALUES(48016,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48016,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(48016,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48016,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(48016,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48016,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48016,36,'cp 11','e4f6 c2c3 f8e7 h2h3 e8g8 d3c2 f8e8 e1g1 h7h6 d2d4 e7f8 c1f4 c8e6 b1d2 c7c5 f1e1 b8c6 d4c5 d6c5 c2b3 d8d7 d2c4 a8d8 d1d7 d8d7 c4a3 e6b3 e1e8 f6e8 a2b3 g7g5 f4g3 f8g7 a3c2 e8f6 c2e3');
INSERT INTO evaluations VALUES(47963,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(47963,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(47963,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(47963,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47963,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(47963,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(47963,36,'cp 13','e1g1 f8e7 f1e1 e8g8 d3f1 d6d5 d2d4 c8e6 h2h3 f8e8 b1c3 b8d7 f1d3 c7c5 d4c5 e7c5 c1e3 a7a6 d1d2 a8c8 e3d4 h7h6 e1e2 c5d4 f3d4 d7c5 a1e1 d8b6 d4b3 c5b3 a2b3 c8d8 d2f4 f6h5 f4h4 h5f6 b3b4 e6d7 e2e8 d8e8 e1e8 d7e8');
INSERT INTO evaluations VALUES(47962,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47962,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(47962,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(47962,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(47962,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(47962,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(47962,36,'cp 13','f8e7 f1e1 e8g8 d3f1 d6d5 d2d4 h7h6 c2c4 b8c6 b1c3 d5c4 f1c4 e7d6 d4d5 c6a5 c4f1 c7c6 d5c6 a5c6 h2h3 d6c5 a2a3 d8d1 e1d1 a7a5 c3a4 c5a7 c1e3 a7e3 f2e3 c8f5 a1c1 a8d8 f1b5 d8d1 c1d1 f8c8 a4b6 c8d8');
INSERT INTO evaluations VALUES(40861,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40861,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(40861,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(40861,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(40861,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40861,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40861,36,'cp 28','c2c3 e8g8 h2h3 f8e8 d3c2 b7b6 d2d4 h7h6 f1e1 e7f8 e1e8 d8e8 c1f4 b8c6 b1d2 c8b7 c3c4 c6e7 d1f1 e8d7 a1e1 a8e8 d4d5 b7a6 c2d3 c7c6 d5c6 e7c6 d2e4 c6e5 e4f6 g7f6 d3e4 a6c4 f3e5 d6e5 f1c4');
INSERT INTO evaluations VALUES(7864,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(7864,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(7864,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(7864,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(7864,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(7864,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(7864,36,'cp -66','b4c3 d2c3 f6e4 f3e5 e4c3 e5c6 c3d1 c6d8 d1b2 d8b7 c8b7 a1b1 b7g2 f1g2 a8b8 e1g1 e8g8 f1e1 b2a4 b1b8 f8b8 e1e5 a4c3 e5e3 c3a2 e3a3 b8b1 g2f1 a2c1 g1g2 g8f8 a3a7 f8e7 h2h4 b1b2 a7c7 c1a2 h4h5 e7d8 c7a7 a2c3 f1d3 g7g6 a7a1 d8e7 g2f3');
INSERT INTO evaluations VALUES(7875,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(7875,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(7875,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(7875,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(7875,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(7875,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(7875,36,'cp -66','d2c3 f6e4 f3e5 e4c3 e5c6 c3d1 c6d8 d1b2 d8b7 c8b7 a1b1 b7g2 f1g2 a8b8 e1g1 e8g8 f1e1 b2a4 b1b8 f8b8 e1e5 a4c3 g2f3 c3a2 e5a5 a2c3 a5a7 b8d8 f3g4 d7d5 h2h4 c3b5 a7a5 b5d4 c2c3 d4b3 a5a7 g7g6 h4h5 b3d2 h5h6 g8f8 a7c7 d5d4 c3d4 d8d4 c7c8 f8e7');
INSERT INTO evaluations VALUES(7873,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(7873,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(7873,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(7873,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(7873,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(7873,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(7873,36,'cp -57','f6e4 f3e5 e4c3 e5c6 c3d1 c6d8 d1b2 d8b7 c8b7 a1b1 b7g2 f1g2 a8b8 e1g1 e8g8 f1e1 b2a4 b1b8 f8b8 e1e5 a4c3 e5e3 c3a2 e3a3 b8b1 g2f1 a2c1 g1g2 g8f8 a3a7 f8e7 a7a4 b1b2 a4e4 e7d8 e4f4 f7f6 f4c4 g7g6 c4h4 h7h5 h4c4 f6f5 h2h4 c1a2 f1d3 a2b4 g2f3 d8e7');
INSERT INTO evaluations VALUES(7882,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(7882,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(7882,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(7882,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(7882,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(7882,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(7882,36,'cp -58','f3e5 e4c3 e5c6 c3d1 c6d8 d1b2 d8b7 c8b7 a1b1 b7g2 f1g2 a8b8 e1g1 e8g8 f1e1 b2a4 b1b8 f8b8 e1e5 a4c3 g2f3 c3a2 e5a5 a2c3 a5a7 b8d8 f3g4 d7d5 a7c7 c3e4 g4f3 e4f6 c2c4 d5c4 c7c4 g7g6 g1g2 g8f8 h2h3 f6e8 c4c5 f8g7 c5d5 d8d6 g2g3 d6c6 d5d7 c6e6 f3d5 e6f6 d7b7 f6f5');
INSERT INTO evaluations VALUES(16059,30,'cp -464',NULL);
INSERT INTO evaluations VALUES(16059,31,'cp -447',NULL);
INSERT INTO evaluations VALUES(16059,32,'cp -498',NULL);
INSERT INTO evaluations VALUES(16059,33,'cp -492',NULL);
INSERT INTO evaluations VALUES(16059,34,'cp -474',NULL);
INSERT INTO evaluations VALUES(16059,35,'cp -496',NULL);
INSERT INTO evaluations VALUES(16059,36,'cp -483','c6d4 f3d1 d7d5 e4d5 c8g4 f2f3 g4h3 g2h3 f6d5 c4d5 d8d5 e1g1 d5c5 c2c3 d4e6 g1h1 c5b5 d2d3 e8c8 a2a4 b5d3 d1d3 d8d3 b1a3 f8a3 a1a3 h8d8 a3a1 d8d5 c1h6 c8d7 a1e1 f7f6 h6e3 b7b6 h3h4 a7a5 e3h6 d7c6 f3f4 e6f4 h6f4 e5f4 f1f4 f6f5 e1e7 d3d1 h1g2');
INSERT INTO evaluations VALUES(16084,30,'cp -463',NULL);
INSERT INTO evaluations VALUES(16084,31,'cp -480',NULL);
INSERT INTO evaluations VALUES(16084,32,'cp -458',NULL);
INSERT INTO evaluations VALUES(16084,33,'cp -458',NULL);
INSERT INTO evaluations VALUES(16084,34,'cp -462',NULL);
INSERT INTO evaluations VALUES(16084,35,'cp -482',NULL);
INSERT INTO evaluations VALUES(16084,36,'cp -463','f3d1 d7d5 e4d5 c8g4 f2f3 g4h3 g2h3 f6d5 c4d5 d8d5 e1g1 d5c5 c2c3 d4e6 g1h1 e8c8 b1a3 d8d3 a3c2 c5b5 a2a4 b5c4 c2e3 c4h4 d1e1 h4h5 e1g3 f8d6 e3c4 f7f5 c4d6 c7d6 h3h4 e5e4 a4a5 h8e8 b2b4 e4f3 f1g1 e6c7 b4b5 e8e2 a5a6 b7b6 c1a3 c8d7 a1e1');
INSERT INTO evaluations VALUES(49469,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(49469,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(49469,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(49469,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49469,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(49469,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(49469,36,'cp 29','d7d5 d2d4 f8b4 c2c3 b4e7 c3c4 e8g8 e1g1 d5c4 d1c2 b7b5 a2a4 b5b4 f3d2 c7c6 d2c4 d8d4 b1d2 c8a6 b2b3 b8d7 c1b2 d4c5 a1c1 a8c8 f1d1 f8d8 h2h4 h7h6 d2e4 f6e4 b2d4 e4f2 g1f2 c5f5 c2f5 e6f5 d4a7 d7f6 d1d8 c8d8 g2c6');
INSERT INTO evaluations VALUES(9789,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9789,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9789,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9789,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9789,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9789,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9789,36,'cp 0','d7d6 b1c3 a7a6 c1e3 c5e3 f2e3 c8e6 c3d5 e8g8 d5f6 d8f6 c4e6 f6e6 f3h4 a8d8 d1f3 d6d5 e4d5 e6d5 f3g4 c6e7 f1f2 d5c5 g4e4 c5d5');
INSERT INTO evaluations VALUES(8861,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(8861,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8861,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(8861,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(8861,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(8861,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8861,36,'cp 0','b1c3 a7a6 c1e3 c5e3 f2e3 c8e6 c3d5 e8g8 d5f6 d8f6 c4e6 f6e6 f3h4 c6e7 f1f2 a8d8 d1f3 d6d5 a1f1 c7c6 a3a4 f7f6 h2h3 d8d7 b2b4 f8f7 g1h2 d7d8 a4a5 d8d7 h2g1 g8h7 g1h2');
INSERT INTO evaluations VALUES(21286,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(21286,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(21286,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(21286,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(21286,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(21286,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(21286,36,'cp 80','d2d4 e5d4 f3d4 f8e7 b1c3 c6d4 d1d4 e7f6 d4d3 g8e7 c1e3 c8e6 e1c1 e6c4 d3c4 d8d7 h1e1 d7c6 c4c6 b7c6 e3d4 f6d4 d1d4 e7g6 g2g3 h6h5 h3h4 g6e5 d4a4 g7g5 h4g5 e5f3 e1h1 f3g5 a4a6 e8d7 f2f4 g5e6 e4e5 d6e5 h1d1 d7e7 a6c6 e5f4');
INSERT INTO evaluations VALUES(21295,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(21295,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(21295,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(21295,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(21295,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(21295,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(21295,36,'cp 61','e5d4 f3d4 f8e7 e1g1 g8f6 b1c3 e8g8 c1f4 c6d4 d1d4 c8e6 a1d1 f6d7 c4e6 f7e6 f4g3 d8e8 d4c4 e7d8 f2f4 a7a6 a2a4 e8f7 f1f3 a8c8 a4a5 f8e8 g3f2 d8f6 c4b3 f6c3 b2c3 c8b8 f2a7 b8a8 b3b7');
INSERT INTO evaluations VALUES(26359,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(26359,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(26359,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(26359,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26359,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26359,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(26359,36,'cp 2','e1g1 g8e7 f1e1 e7f5 b1c3 d8c7 a2a4 f5d4 f3d4 c5d4 d1d4 c7b6 c3e2 b6d4 e2d4 a7a5 e1e3 f8c5 c2c3 c6d7 e3g3 c5f8 f2f4 a8c8 h2h4 h7h5 c1e3 h8g8 g3f3 f8e7 f4f5 e7h4 f3h3 h4e7 h3h5 e6f5');
INSERT INTO evaluations VALUES(26358,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(26358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26358,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(26358,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(26358,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(26358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26358,36,'cp 0','g8e7 c2c3 e7f5 f1e1 f8e7 b1d2 c5d4 f3d4 f5d4 c3d4 e8g8 e1e3 c6e8 d1g4 f7f5 e5f6 f8f6 d2f3 e7d6 e3b3 d8c7 b3c3 c7b6 c3b3 b6c7');
INSERT INTO evaluations VALUES(41088,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(41088,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(41088,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(41088,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41088,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41088,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41088,36,'cp 10','c1g5 e8g8 e2e3 h7h6 g5f6 e7f6 a1c1 c7c6 f1d3 b8d7 h2h3 f8e8 e1g1 a7a5 a2a3 d7f8 d1c2 f6e7 f1e1 e7d6 e3e4 d5e4 d3e4 c8e6 d4d5 c6d5 e4d5 d8b6 d5e6 e8e6 c3d5 b6b5 c1d1 e6e1 f3e1 f8e6 e1f3 b5c6 d5c3 a8c8');
INSERT INTO evaluations VALUES(44708,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(44708,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(44708,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(44708,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(44708,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(44708,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(44708,36,'cp 55','f1e2 f8g7 g1f3 e8g8 e1g1 f8e8 a2a4 a7a5 c4d5 c6d5 c1d2 b8c6 d1b3 e7e6 f1c1 f6e4 d2e1 e4g5 f3e5 c6e5 d4e5 g7e5 f2f4 e5c3 e1c3 g5e4 c3d4 d8h4 e2f3 e4d2 b3b5 d2f3 g2f3 e8d8 d4b6 e6e5 b6d8 h4d8 b2b4 a5b4 a4a5 d8h4 b5d5');
INSERT INTO evaluations VALUES(51830,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(51830,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(51830,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(51830,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(51830,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(51830,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51830,36,'cp 0','d8e7 h5e2 f7f5 d2d3 g8f6 c4b3 a7a5 a2a4 b8a6 b1c3 d7d6 g1f3 f5e4 d3e4 a6c5 b3a2 c8g4 e1g1 g4h5 a1b1 e7c7 f1e1 c5a6 a2e6 a6c5');
INSERT INTO evaluations VALUES(34269,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34269,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34269,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34269,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34269,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34269,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34269,36,'cp 0','h5e2 f7f5 d2d4 f5e4 e2h5 g7g6 h5e5 g8f6 c1g5 f8g7 e5c7 e7d8 c7e5');
INSERT INTO evaluations VALUES(54188,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54188,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(54188,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(54188,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(54188,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54188,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54188,36,'cp 6','e2e4 c5d4 d1d4 b8c6 d4a4 f8b4 e4d5 e6d5 f1b5 b4c3 b2c3 c8d7 g1f3 g8f6 a4a3 d8e7 a3e7 c6e7 b5d3 f6e4 e1g1 e8g8 c3c4 f8e8 f1e1 a8c8 f3e5 d7f5 g2g4 f7f6 g4f5 f6e5 c4d5 e7d5 f4e5 e8e5 e1e4 e5e4 d3e4 d5f6 e4d3 f6d5 a1e1 g8f8 e1e4 b7b6 f2f3');
INSERT INTO evaluations VALUES(54242,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(54242,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(54242,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(54242,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(54242,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(54242,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(54242,36,'cp 17','c5d4 d1d4 b8c6 d4a4 f8b4 e4d5 e6d5 f1b5 b4c3 b2c3 c8d7 g1e2 g8e7 e1g1 e8g8 f1e1 f8e8 e2g3 a7a6 b5f1 e7g6 f4e3 d7e6 a4a3 d8c7 a3c5 g6e5 a2a4 a8c8 h2h3 c7a5 c5a3 e5c4 f1c4 d5c4 g3e4 c8d8 a1b1 e6d5 e4d6 e8e7');
INSERT INTO evaluations VALUES(14793,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(14793,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(14793,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(14793,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(14793,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(14793,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(14793,36,'cp 26','g1f3 f8d6 f3e5 e8g8 d1f3 d8e7 h2h4 f6d7 e5d7 c8d7 e1c1 a7a6 e3e4 d5e4 f3e4 c6b4 a2a3 d7c6 f4d6 e7d6 e4e5 b4d5 c3d5 d6d5 e5d5 c6d5 h1h3 a8d8 h4h5 d8d7 f2f4 f8d8 d1e1 f7f5 h3e3 g7g6 h5g6 h7g6 g2g3 g8f7 f1e2');
INSERT INTO evaluations VALUES(17803,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(17803,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17803,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(17803,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(17803,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(17803,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(17803,36,'cp 22','f8e7 d2d4 b7b5 a4b3 c8b7 a2a4 c6d4 f3d4 e5d4 d1d4 g8f6 a4b5 a6b5 a1a8 b7a8 c3b5 a8e4 f2f3 e4f5 d4a7 e8g8 e1g1 d8d7 a7c7 d7c7 b5c7 f8c8 g2g4 f5c2 b3c2 c8c7 c2b1 g7g6 f1d1 f6d7 b1a2 e7f6 a2d5 d7b6 d5b3 d6d5 g4g5 f6g7 b3d5 b6d5');
INSERT INTO evaluations VALUES(30990,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(30990,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(30990,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(30990,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(30990,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(30990,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(30990,36,'cp 86','f3h4 f5h7 e5e6 d8d6 e6f7 e8f7 f1d3 e7e5 d1h5 g7g6 h5e5 d6e5 d4e5 g6g5 h4f3 f8g7 d3h7 h8h7 c1g5 h6g5 f3g5 f7g6 g5h7 g7e5 b1d2 g6h7 e1c1 b8d7 h1e1 a8f8 e1e2 e5g7 d2b3 h7g6 c2c4 g8f6 c4d5 f6d5 g2g3 g6f7 b3a5 d7c5 f2f4 f8d8 a5c4 f7g8 e2e1 g7f6 c1b1 f6g7 h2h4');
INSERT INTO evaluations VALUES(10138,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10138,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10138,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10138,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10138,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10138,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10138,36,'cp 0','d2d3 c6d4 g1f3 d7d6 f3d4 c5d4 d1f3 c7c6 c3e2 d4b6 f3g3 d8e7 h1f1 d6d5 e4d5 c6d5 c4a2 e8g8 c1h6 f6h5 g3g5 e7g5 h6g5 c8e6 e1c1 h7h6 g5d2 f8e8 f2f3 d5d4 a2e6 e8e6 d1e1 a8c8 c1b1 h5f6 e2g3 f6d5 g3f5 g8h7 g2g4');
INSERT INTO evaluations VALUES(19569,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(19569,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(19569,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(19569,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(19569,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(19569,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19569,36,'cp 0','g8f6 b1c3 e7e6 g1f3 d7d5 e4d5 e6d5 f1d3 c8g4 d2e2 f8e7 h2h3 c6d4 e2e5 d4c6');
INSERT INTO evaluations VALUES(12903,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12903,36,'cp 0','b1c3 e7e5 a2a3 f8c5 f1c4 d7d6 g1f3 c5d4 f3d4 e5d4 c3d5 f6e4 d2f4 f7f5 e1g1 c6e5 c4b5 c8d7 f4f5 d7b5 f2f4 b5f1 f4e5 f1c4 f5e6 e8f8 e6f5 f8g8 f5e6');
INSERT INTO evaluations VALUES(12892,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12892,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12892,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12892,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(12892,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(12892,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(12892,36,'cp -12','e7e5 a2a3 f8c5 f1c4 d7d6 g1f3 a7a6 d2d3 b7b5 c4a2 h7h6 c3d5 c8e6 e1g1 e8g8 c2c3 e6d5 a2d5 f6d5 d3d5 d8c7 d5d3 c6e7 c1e3 c5e3 d3e3 f7f5 e4f5 e7f5 e3e4 c7c4 f1e1 c4e4 e1e4 g8f7 h2h3 f8c8 a1d1 c8c6 a3a4 a8c8');
INSERT INTO evaluations VALUES(43047,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(43047,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43047,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(43047,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(43047,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(43047,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(43047,36,'cp -16','e8g8 d2d3 d7d5 b1c3 b8c6 f1e2 c6d4 e2d1 d5e4 d3e4 b7b5 g1e2 b5b4 e2d4 d8d4 c1e3 d4d8 e4e5 b4c3 e5f6 e7f6 b2b3 c8f5 e1g1 c7c6 g3f4 f5g6 a1c1 d8a5 d1f3 a5a2 f3c6 a8d8 f4a4 a2a4 c6a4 d8d2 b3b4 f8d8 g1h1 d2e2 f1e1 e2e1 c1e1 h7h6 e3c5');
INSERT INTO evaluations VALUES(37327,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(37327,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(37327,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(37327,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(37327,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(37327,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(37327,36,'cp -23','d2d3 d7d5 b1c3 b8c6 f1e2 c6d4 e2d1 d5e4 d3e4 b7b5 g1f3 b5b4 f3d4 d8d4 c1e3 d4d8 e4e5 b4c3 e5f6 e7f6 b2b3 c8a6 d1e2 a6e2 e1e2 d8c8 h1d1 c8a6 e2e1 a8e8 g3g4 f6e5 g4e2 a6g6 e1f1 e5h2 e2d3 g6h5 d3e2 h5h4 e2f3 h2e5 f3h3 h4b4 d1d5 b4b7 h3f3 b7a6 f1g1 a6g6 e3c5 g6c2 c5f8 c2b2');
INSERT INTO evaluations VALUES(37326,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(37326,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(37326,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(37326,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(37326,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(37326,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(37326,36,'cp -10','d7d5 b1c3 b8c6 f1e2 c6d4 e2d1 d5e4 d3e4 b7b5 g1e2 b5b4 e2d4 d8d4 c1e3 d4d8 e4e5 b4c3 e5f6 e7f6 b2b3 c8a6 d1e2 a6e2 e1e2 d8c8 h1d1 c8a6 e2e1 f8e8 a2a4 h7h6 d1d7 a6e6 a1d1 c7c5 g3f3 f6d4 d7c7 a8c8 c7c8 e6c8 d1d3 c8c7 h2h3 c5c4');
INSERT INTO evaluations VALUES(22830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22830,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22830,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(22830,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(22830,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22830,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22830,36,'cp 0','b1c3 g8f6 g1f3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 c8g4 c1e3 c6e7 h2h3 g4h5 f1e1 e7d5 e3d2 d8f6 g2g4 h5g6 c2c4 d5f4 c4c5 f4h3 g1g2 h3f4 g2g1');
INSERT INTO evaluations VALUES(22825,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(22825,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(22825,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(22825,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22825,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(22825,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(22825,36,'cp -2','g8f6 g1f3 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 h7h6 b5d3 d8f6 h2h3 c8f5 d3f5 f6f5 c2c4 f8d8 c1e3 a7a6 d1b1 f5b1 f1b1 b7b6 g2g3 c6a5 f3d2 d6e7 a1a2 d8d7 g1g2 a8e8 d4d5 a5b7 d2b3 e7f6 a3a4 b7d6');
INSERT INTO evaluations VALUES(32355,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32355,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(32355,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(32355,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(32355,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(32355,36,'cp -15','c7c5 c2c3 b8d7 f1d3 f5d3 d1d3 g8e7 b1a3 e7c6 a3c2 f8e7 e1g1 e8g8 c1d2 a7a6 h2h3 d8b6 d2g5 b6b5 d3d2 e7g5 f3g5 c5d4 c2d4 c6d4 c3d4 a8c8 g5f3 c8c4 b2b3 c4c6 h3h4 f8c8 h4h5 h7h6 a1c1 b5b6 g2g4 c6c1 f1c1 c8c1 d2c1 b6a5 c1c8 d7f8');
INSERT INTO evaluations VALUES(31675,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31675,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31675,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(31675,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31675,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31675,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31675,36,'cp 0','c2c3 b8d7 b1d2 g8e7 d4c5 d7c5 d2b3 c5d7 f1e2 e7c6 f3d4 d7e5 d4f5 e6f5 c1e3 g7g6 e1g1 f8e7 e3h6 e7f8');
INSERT INTO evaluations VALUES(31667,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(31667,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(31667,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(31667,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(31667,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(31667,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(31667,36,'cp -7','b8d7 b1d2 g8e7 d4c5 d7c5 d2b3 c5d7 f1e2 f5g6 c1f4 e7c6 h2h4 h7h6 h4h5 g6e4 b3d2 e4h7 d2b3 f8e7 e1g1 e8g8 e2d3 h7d3 d1d3 e7g5 f4g3 a7a5 a2a4 a8c8 d3e2 d8b6 e2b5 g5e7 g3f4 f8d8 f1e1 c8a8 g2g3 e7c5 b3c5 b6c5 b5e2 c5c4');
INSERT INTO evaluations VALUES(45738,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(45738,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45738,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45738,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45738,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45738,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45738,36,'cp 0','c6d5 c4b3 b8c6 e1g1 c8e6 c1g5 f8d6 f1e1 e8g8 b1c3 d6c7 b3a4 e6g4 a4b3');
INSERT INTO evaluations VALUES(51489,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(51489,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(51489,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(51489,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(51489,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(51489,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51489,36,'cp 46','d7d5 e2e4 d5e4 c3e4 c8b7 f1d3 e7e6 g1f3 g8f6 e4f6 g7f6 d1e2 b8c6 c2c3 d8d7 c1f4 e8c8 e1c1 h7h5 d3e4 f8d6 f4d6 d7d6 c1b1 c6a5 g2g3 b7e4 e2e4 c8b8 e4e2 d6c6 h1e1 c6d5 b2b3 a5c6 d1d3 a7a5 b1b2 e6e5 d3d1 d5e6 d4e5 d8d1 e1d1 f6e5');
INSERT INTO evaluations VALUES(51294,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(51294,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(51294,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51294,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(51294,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(51294,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(51294,36,'cp 49','e2e4 d5e4 c3e4 c8b7 f1d3 g8f6 d1e2 e7e6 e4f6 g7f6 g1f3 b8c6 c2c3 d8d7 c1e3 e8c8 e1c1 c8b8 h1e1 h8g8 g2g3 c6b4 d3c4 b6b5 c4b3 b4d5 c1b1 h7h5 e3c1 a7a5 b3c2 a5a4 a2a3 f8d6 b1a1 d5b6 f3h4 f6f5 e2h5');
INSERT INTO evaluations VALUES(51300,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(51300,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(51300,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51300,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(51300,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(51300,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51300,36,'cp 57','d5e4 c3e4 c8b7 f1d3 e7e6 d1e2 g8f6 e4f6 g7f6 g1f3 b8c6 c1e3 d8d5 d3c4 d5h5 e1c1 e8c8 h2h3 f8h6 g2g4 h5g6 c4d3 g6g7 d3e4 h6e3 f2e3 h7h5 h1g1 h5g4 h3g4 g7f8 a2a3 f8e7 f3d2 c8b8 g1e1 c6a5');
INSERT INTO evaluations VALUES(54801,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54801,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54801,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54801,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54801,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(54801,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(54801,36,'cp 5','f8g7 c2c4 e7e6 g1f3 g8f6 b1c3 e8g8 f1d3 b8a6 e1g1 a6c5 d3c2 e6d5 c4d5 f8e8 f1e1 a7a5 f3d4 c8d7 c1f4 f6h5 f4e3 h5f6 f2f3 c7c6 d5c6 b7c6 d4b3 c5b3 c2b3 d6d5 e4d5 f6d5 c3d5 c6d5 d1d5 d7e6 d5d8 a8d8 e3g5');
INSERT INTO evaluations VALUES(38337,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(38337,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38337,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(38337,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(38337,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38337,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38337,36,'cp 13','c2c4 g8f6 b1c3 e7e6 g1f3 e8g8 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 c1f4 f6h5 f4e3 b8d7 d3e2 f8e8 f1e1 h5f6 h2h3 g4f3 e2f3 a7a6 d1b3 a8b8 b3a3 d7e5 f3e2 b8c8 a1c1 e5c4 e2c4 c8c4 a3b3 d8c8 a2a4 h7h5 a4a5');
INSERT INTO evaluations VALUES(38333,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(38333,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38333,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(38333,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(38333,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38333,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38333,36,'cp 0','g8f6 b1c3 e7e6 g1f3 e8g8 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 c1g5 h7h6 g5f4 b7b5 a1d1 d8d7 d3f3 h6h5 f4g5 f6h7 g5f4');
INSERT INTO evaluations VALUES(41187,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41187,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41187,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41187,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(41187,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41187,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(41187,36,'cp 0','b1c3 e7e6 g1f3 e8g8 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 c1g5 d8b6 g5e3 b6b4 a2a3 b4b3 e3f4 f8e8 d3d2 b3b6 f4e3 b6b3');
INSERT INTO evaluations VALUES(41186,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41186,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41186,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41186,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41186,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(41186,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41186,36,'cp 12','e7e6 g1f3 e8g8 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 f8e8 c1f4 d8b6 d3f1 a7a6 e1c1 a8c8 a1b1 f6h5 f4e3 b6d8 a2a4 d7e5 f3d1 d8h4 a4a5 e5c4 f1c4 c8c4 d1f3 c4b4 b2b3');
INSERT INTO evaluations VALUES(3497,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(3497,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(3497,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(3497,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(3497,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(3497,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(3497,36,'cp 98','d4d5 c6b8 c1g5 h7h6 g5e3 e7e6 d5e6 c8e6 f3d4 b8c6 d4e6 f7e6 d1d2 g8h7 e1g1 e6e5 a1d1 f6d7 g2g3 c6d4 c3b5 d4e2 d2e2 b7b6 b2b4 d7f6 b5c3 a7a5 b4b5 d8e8 f2f3 f8f7 c3d5 h6h5 g3g4 f6d5 c4d5 e8d7 g4h5 g6h5');
INSERT INTO evaluations VALUES(3473,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(3473,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(3473,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(3473,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(3473,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(3473,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(3473,36,'cp 89','c6b8 c1g5 h7h6 g5e3 e7e6 d5e6 c8e6 f3d4 b8c6 d4e6 f7e6 d1d2 g8h7 a1d1 f6d7 e1g1 e6e5 g2g3 c6d4 c3b5 d4c6 e2d3 d7c5 d3b1 c5e6 f2f4 e5f4 g3f4 d8h4 g1h1 h4h3 f4f5 g7e5 e3h6 h3h6 d2h6 h7h6 f5e6 f8f1 d1f1 a8e8 b5c7');
INSERT INTO evaluations VALUES(11765,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(11765,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(11765,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(11765,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(11765,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(11765,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(11765,36,'cp 70','a4b3 c8b7 e1g1 e7g6 d2d4 f8e7 d4d5 c6a5 b3c2 a5c4 b2b3 c4d6 b1d2 e8g8 c3c4 d6e8 c1b2 d7d6 f1e1 c7c6 d2f1 c6d5 c4d5 b7c8 a2a4 b5a4 b3a4 f7f5 f1e3 f5e4 f3d2 g6f4 d2e4 e8f6 e4g3 g7g6 b2a3 c8d7');
INSERT INTO evaluations VALUES(11761,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(11761,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(11761,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(11761,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(11761,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(11761,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(11761,36,'cp 87','e7g6 d2d4 c6a5 b3c2 d7d6 a2a4 c8d7 c1g5 f7f6 g5c1 c7c5 b1d2 f8e7 d2f1 c5d4 c3d4 e8g8 f1e3 a5c4 e3f5 d8a5 f3d2 d6d5 f5e7 g6e7 e1g1 d7e6 b2b3 a5c3 b3c4 c3a1 c4d5 e6f7 d4e5 f6e5 d2f3 a8c8 c2b3 b5a4 b3a4 f7h5 d5d6 e7g6 d1d5 g8h8');
INSERT INTO evaluations VALUES(14179,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(14179,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(14179,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(14179,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(14179,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(14179,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(14179,36,'cp -28','e2e3 e7e5 b1c3 d8e7 d4e5 c6e5 c3d5 f6d5 d1d5 c7c6 d5d4 e7c5 d4c5 f8c5 c1d2 c8f5 d2c3 e5d3 f1d3 f5d3 g1e2 f7f6 e2d4 e8f7 e1c1 d3a6 b2b4 c5b6 d4b3 a8d8 b3c5 b6c5 b4c5 a6c4 a2a3 h8e8 d1d8 e8d8 h1d1 d8d1 c1d1 c4f1 g2g3 f7e6 c3d4 h7h5 d1c2');
INSERT INTO evaluations VALUES(14177,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(14177,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(14177,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(14177,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(14177,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(14177,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(14177,36,'cp -21','e7e5 b1c3 d8e7 d4e5 c6e5 c3d5 f6d5 d1d5 c7c6 d5d4 e7c5 d4c5 f8c5 c1d2 c8f5 d2c3 e5d3 f1d3 f5d3 g1e2 f7f6 e2d4 d3c4 b2b3 c4f7 e1c1 e8g8 d4f5 f7e6 f5d4 e6d7 f2f3 a7a5 a2a4 h7h5 d1d2 f8e8 e3e4 h5h4 h1d1 c5b6 c1c2 a8d8');
INSERT INTO evaluations VALUES(14964,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(14964,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(14964,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(14964,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(14964,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(14964,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(14964,36,'cp -28','b1c3 d8e7 d4e5 c6e5 c3d5 f6d5 d1d5 e7c5 d5c5 f8c5 g1f3 c5b4 c1d2 b4d2 f3d2 c8e6 d2b3 e8e7 b3c5 e6d5 e3e4 b7b6 c5d3 e5d3 f1d3 d5e6 e1e2 h8d8 e2e3 c7c5 h1d1 d8d4 d3e2 d4d1 e2d1 a8d8 d1e2 d8d4 b2b3 f7f5 e4f5');
INSERT INTO evaluations VALUES(21600,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(21600,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21600,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21600,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(21600,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(21600,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21600,36,'cp 26','c2c3 g8f6 g1e2 f8d6 e1g1 h7h6 e2g3 e8g8 b1d2 c6e7 f1e1 e7g6 d2f3 f8e8 c1d2 e8e1 d1e1 d8e7 e1e7 g6e7 a1e1 c8e6 f3h4 c7c6 h4f5 e7f5 g3f5 d6c7 f2f3 f6e8 h2h4 g8f8 h4h5 e6f5 d3f5 e8d6 f5d3 d6c4 d2c1 a8e8');
INSERT INTO evaluations VALUES(21598,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(21598,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(21598,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(21598,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(21598,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(21598,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(21598,36,'cp 14','f8d6 g1e2 g8f6 e1g1 h7h6 b1d2 e8g8 d2f3 f8e8 e2g3 a7a6 c1d2 c6a5 b2b3 a5c6 f1e1 e8e1 d1e1 d8e7 h2h3 c8d7 e1e7 c6e7 a1e1 b7b6 f3e5 d7b5 d3c2 g8f8 a2a4 b5e8 g3f5 e7f5 c2f5 b6b5 a4a5 c7c5 g2g4 c5d4 c3d4');
INSERT INTO evaluations VALUES(5902,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(5902,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(5902,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(5902,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(5902,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(5902,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(5902,36,'cp 21','g1e2 g8f6 e1g1 e8g8 e2g3 h7h6 b1d2 c6e7 f1e1 e7g6 d2f3 f8e8 c1d2 e8e1 d1e1 d8e7 e1e7 g6e7 a1e1 c8e6 f3h4 c7c6 h4f5 e7f5 g3f5 d6c7 f2f3 e6d7 f5e3 h6h5 g2g3 a8e8 e3g2 e8e1 g2e1 g7g6 e1g2 f6e8 d3c2 e8g7 d2f4');
INSERT INTO evaluations VALUES(44856,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(44856,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(44856,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(44856,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(44856,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(44856,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(44856,36,'cp -58','e7e5 b1c3 f8e7 b2b4 e8g8 g1f3 c8g4 d4e5 f6d7 c1b2 a7a5 b4b5 d7c5 f1e2 a5a4 a1c1 b8d7 e1g1 a4a3 b2a1 f8e8 f3e1 g4e2 c3e2 d7b6 b5c6 b7c6 e2f4 b6c4 e1d3 c5d3 f4d3 e7f8 d1g4 d8c7 f1e1 c7a5 a1c3');
INSERT INTO evaluations VALUES(47272,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(47272,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(47272,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(47272,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(47272,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(47272,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(47272,36,'cp -24','e7e5 g1f3 f8d6 e1g1 e8g8 c2c4 c7c6 c4d5 c6d5 b1c3 h7h6 e2e4 d5d4 c3b5 c8g4 h2h3 g4f3 g2f3 d6b4 a2a3 b4c5 a3a4 d8e7 f3g2 b8c6 h3h4 a7a6 b5a3 b7b5 g2h3 f8b8 c1d2 c5b4 a4b5 a6b5 d1c2 c6a5 a3b1 g8h8 d2b4 e7b4');
INSERT INTO evaluations VALUES(48536,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(48536,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(48536,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(48536,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(48536,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(48536,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(48536,36,'cp 96','e4f6 g7f6 d2d4 b8c6 f1c4 h8g8 c1f4 g8g4 g2g3 g4f4 g3f4 d8d6 d4d5 c6a5 c4d3 d6f4 d1e2 c8d7 d5e6 d7e6 e2e4 f4e4 d3e4 e8c8 a1d1 d8e8 e1g1 a5c4 f3d4 e6d7 e4d5 c4e5 f1e1 f8c5 g1h1 e8g8 d5e4 g8g4 c2c3 g4f4 f2f3');
INSERT INTO evaluations VALUES(48153,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(48153,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(48153,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(48153,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(48153,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(48153,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(48153,36,'cp 87','g7f6 d2d4 b8c6 f1c4 h8g8 c1f4 g8g4 g2g3 g4f4 g3f4 d8d6 d4d5 d6f4 d1e2 c6a5 c4d3 c8d7 d5e6 d7e6 e2e4 f4e4 d3e4 e8c8 a1d1 d8e8 e1g1 a5c4 f3d4 e6d7 e4d5 c4b2 d1e1 f8c5 e1e8 d7e8 f1e1 e8d7 d4b3 c5b4 e1e3');
INSERT INTO evaluations VALUES(21087,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(21087,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(21087,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(21087,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(21087,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(21087,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(21087,36,'cp -12','d5d4 g1f3 e7e6 a2a3 h7h5 f1d3 b7b6 b2b4 c8b7 c1b2 d8d7 d3e4 e8c8 b4b5 c6a5 e4b7 c8b7 f3d4 b7b8 d1f3 f8c5 d4c6 a5c6 b5c6 d7e8 c2c3 c5f8 d2d4 g8e7 b1d2 e8c6 f3c6 e7c6 d2e4 f8e7 e1g1 h5h4 f4f5 h4h3 f5e6 f7e6 f1f7 g7g5 a1f1 h3g2 g1g2 h8h4');
INSERT INTO evaluations VALUES(21093,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(21093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21093,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21093,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21093,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(21093,36,'cp -5','g1f3 e7e6 a2a3 b7b6 f1c4 g8e7 d2d3 e7f5 d1e2 h7h5 b1d2 c6e7 d2e4 c8b7 e1g1 d8d7 c1d2 e7d5 c2c3 e8c8 a3a4 f8e7 f3d4 f5d4 c3d4 d5b4 d2c3 b4c6 c4b5 h5h4 h2h3 c8b8 e2e3 a7a6 b5c6 b7c6 a4a5 b6b5 b2b4 f7f5 e4c5 e7c5 b4c5');
INSERT INTO evaluations VALUES(39193,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(39193,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(39193,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(39193,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(39193,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(39193,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(39193,36,'cp 197','d7d5 e5d6 g8f6 c1f4 c5d6 f4d6 c7d6 e2e3 e8g8 b1d2 b8c6 f1e2 d6d5 e1g1 c8f5 c2c3 h7h6 d1a4 f5g6 f1d1 d8e7 h2h3 f8d8 f3d4 a8c8 a1c1 f6e4 d2e4 g6e4 d1d2 g8h8 d4c6 b7c6 c1d1 e7g5');
INSERT INTO evaluations VALUES(43727,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43727,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43727,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(43727,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(43727,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43727,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43727,36,'cp 9','e7e6 c1e3 f8e7 a2a3 e8g8 f2f3 c8b7 d1e2 b8c6 e1c1 d8c7 c1b1 f8b8 h2h4 c6d4 e3d4 b5b4 a3b4 b7c6 h4h5 b8b4 e4e5 f6d7 e5d6 e7d6 h5h6 g7g6 c3e4 d6e5 d4e5 c7e5 e4c3 e5c7 b3e6 a8b8 e6b3 d7c5 c3d5 c6d5 d1d5 c5b3');
INSERT INTO evaluations VALUES(13277,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(13277,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13277,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(13277,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(13277,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(13277,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(13277,36,'cp -22','c8g4 f1e2 g7g6 c1e3 f8g7 h2h3 g4f3 e2f3 e8g8 d1d2 a7a6 e1g1 a8b8 a2a4 g8h8 a1e1 e7e6 e3g5 d8c7 f3d1 f6g8 f2f4 c6d4 g1h1 g8e7 g2g4 e7c6 f4f5 g7e5 c3e2 f7f6 g5h6 f8g8 e2d4');
INSERT INTO evaluations VALUES(24672,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(24672,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(24672,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(24672,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(24672,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(24672,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(24672,36,'cp -18','f1e2 g7g6 e1g1 f8g7 f3e1 g4e2 d1e2 a7a6 e1c2 f6d7 a2a4 e7e6 a1b1 e8g8 f2f4 c6d4 e2d1 f7f5 e4f5 d4f5 d1e2 e6e5 g2g4 f5h4 c3e4 e5f4 c1f4 d7e5 f4e5 g7e5 f1f8 d8f8 b1f1 f8g7 a4a5');
INSERT INTO evaluations VALUES(24671,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(24671,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(24671,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(24671,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(24671,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(24671,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(24671,36,'cp -21','g7g6 e1g1 a7a6 c1e3 f8g7 a1b1 e8g8 a2a4 e7e6 f3e1 g4e2 d1e2 c6d4 e2d1 f6d7 e1c2 d4c2 d1c2 d7e5 b2b4 b7b6 b4b5 a6b5 c3b5 e5c6 e3f4 c6d4 b5d4 g7d4 h2h3 d8c7 f4e3 d4e3 f2e3 c7d8 c2b2 a8a4 b2b6 d8b6');
INSERT INTO evaluations VALUES(47782,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(47782,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(47782,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(47782,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(47782,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(47782,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(47782,36,'cp -7','f8e7 c2c4 e8g8 b1c3 c7c5 g2g3 b8c6 f1g2 f6e8 g5e7 d8e7 e1g1 e8c7 f3e1 a8b8 f2f4 b7b5 d1d2 f7f6 f4f5 c6d4 b2b3 a7a5 e1c2 c8d7 c2e3 g8h8 h2h4 b5b4 c3e2 a5a4 a1b1 a4b3 a2b3 b8a8 g3g4 a8a3 e2d4 e5d4 e3d5 c7d5 e4d5');
INSERT INTO evaluations VALUES(40757,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(40757,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(40757,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(40757,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(40757,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(40757,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(40757,36,'cp -8','c2c4 e8g8 b1c3 f6e8 g5e7 d8e7 f1e2 c7c5 b2b4 c5b4 c3d5 e7d8 d5b4 e8c7 e1g1 b8c6 b4c2 c7e6 a1b1 b7b6 f1e1 c8b7 e2f1 d8f6 c2e3 g8h8 g2g3 g7g6 f1g2 f6g7 g2h3 e6c5 h3g2 a8d8 g2h3 c6d4 f3d4 e5d4');
INSERT INTO evaluations VALUES(48759,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(48759,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(48759,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(48759,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(48759,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(48759,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(48759,36,'cp 163','e4d5 d8d5 d1e2 f8e7 b1c3 d5e6 b2b3 e8g8 c1b2 b8c6 e5c6 e6c6 e1c1 f8e8 e2b5 c8d7 f2f3 a7a6 b5c6 d7c6 f1d3 g8f8 c3e2 h7h5 e2d4 c6d7 d3f5 g7g6 f5d7 f6d7 f3f4 a8d8 d4f3 b7b5 d1e1 e7f6 e1e8 d8e8 b2f6 d7f6 h1e1 e8e1 f3e1 h5h4 e1f3 f6d5');
INSERT INTO evaluations VALUES(48584,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(48584,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(48584,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(48584,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(48584,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(48584,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(48584,36,'cp 151','d8d5 d1e2 f8e7 b1c3 d5d6 b2b3 e8g8 c1b2 b8c6 e5c6 d6c6 e1c1 f8e8 f2f3 a7a6 e2f2 e7c5 d2d4 c5b4 d4d5 c6c5 f2c5 b4c5 f1d3 c8d7 c1b1 b7b5 h2h3 c5d6 c3e4 f6e4 d3e4 a6a5 a2a3 h7h5 g2g4 h5h4 h1e1 d6g3 e1e3 g3f4 e3e2 f4d6 d1e1 a5a4');
INSERT INTO evaluations VALUES(33908,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(33908,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(33908,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(33908,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(33908,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(33908,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(33908,36,'cp 156','d1e2 f8e7 b1c3 d5e6 b2b3 e8g8 c1b2 b8c6 e5c6 e6c6 e1c1 f8e8 e2b5 c8d7 f2f3 a7a6 b5c6 d7c6 f1d3 h7h5 c3e2 g8f8 e2d4 c6d7 d3f5 g7g6 f5d7 f6d7 f3f4 e7f6 d1e1 c7c5 d4f3 b7b5 b2f6 d7f6 e1e5 c5c4 b3c4 b5c4 e5e8 a8e8');
INSERT INTO evaluations VALUES(6650,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(6650,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(6650,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(6650,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(6650,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(6650,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(6650,36,'cp -19','d7d5 e4d5 d8d5 d2d4 g8f6 d4e5 d5d1 e1d1 f6e4 c1e3 e4c3 d1d2 c3d5 a1e1 c8g4 d2c1 e8c8 e3g5 d8d7 f1b5 g4f3 g2f3 a7a6 b5c6 b7c6 g5d2 d5b6 e1e4 g7g6 d2c3 h8e8 h1e1 c6c5 e5e6 f7e6');
INSERT INTO evaluations VALUES(42046,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42046,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(42046,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42046,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42046,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(42046,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(42046,36,'cp -5','c7c5 c4d5 e6d5 c1g5 h7h6 g5h4 g7g5 h4g3 f6e4 d4c5 d8a5 f3d2 c8f5 d2b3 e4c5 c2f5 b4c3 b2c3 a5c3 e1d1 c5b3 f5e5 c3e5 g3e5 f7f6 e5f6 e8g8 a2b3 f8f6 g2g4 f6f2 d1e1 f2f4 f1g2 b8c6 g2d5 g8g7 a1c1 f4g4 h1f1 a8b8 d5c6 b7c6 c1c6');
INSERT INTO evaluations VALUES(51842,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(51842,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(51842,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(51842,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(51842,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(51842,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(51842,36,'cp 63','g1f3 g8f6 b1c3 d8a5 f1e2 d7d6 e1g1 c6c5 f3e1 b8c6 d2d3 f8e7 f2f4 a5d8 g1h1 a7a6 a2a4 c8d7 a1b1 e8g8 c1d2 c6d4 e1f3 e5f4 b2b4 d4e2 d1e2 c5b4 b1b4 d6d5 b4b7 d5c4 d3c4 a8b8 b7b8 d8b8 f1b1 b8c7 e4e5');
INSERT INTO evaluations VALUES(51841,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(51841,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(51841,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(51841,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(51841,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(51841,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(51841,36,'cp 68','g8f6 b1c3 d8a5 d2d3 d7d6 f1e2 c6c5 e1g1 a5d8 f3e1 b8c6 f2f4 f8e7 g1h1 h7h6 e1f3 e8g8 d1e1 a7a6 a2a4 c8d7 c3d5 d7g4 f4e5 f6d5 c4d5 g4f3 g2f3 e7h4 e1d1 c6e5 f3f4 e5g6 c1e3 h4f6 a1b1 d8e7 e2f3');
INSERT INTO evaluations VALUES(51963,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(51963,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51963,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51963,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51963,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51963,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51963,36,'cp 7','g8f6 c1f4 d7d5 e2e3 d8b6 d1b3 f8e7 h2h3 c8d7 b1d2 d7c6 b3c2 e8g8 a2a4 f8c8 a4a5 b6d8 d4c5 e7c5 f1d3 c5d6 e1g1 a7a6 f1c1 d6c7 f4c7 c8c7 b2b4 d8e7 c2b2 c6b5 d3b5 a6b5 d2b3 b8c6');
INSERT INTO evaluations VALUES(44232,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44232,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(44232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44232,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44232,36,'cp 0','c1f4 b8c6 e2e3 h7h6 b1d2 f6h5 d4c5 h5f4 e3f4 d8c7 b2b4 b7b6 c5b6 a7b6 f4f5 f8d6 d2c4 d6e7 d1d2 e8g8 c4e3 a8a3 e3c2 a3a8 c2e3');
INSERT INTO evaluations VALUES(44220,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44220,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44220,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44220,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44220,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44220,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44220,36,'cp 0','h7h6 e2e3 b8c6 b1d2 f6h5 d4c5 h5f4 e3f4 d8c7 b2b4 b7b6 c5b6 a7b6 f4f5 f8e7 f5e6 d7e6 f1d3 e8g8 e1g1 a8a3 d2b1 a3a8');
INSERT INTO evaluations VALUES(55026,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(55026,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(55026,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(55026,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(55026,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(55026,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(55026,36,'cp -98','e7e5 a2a3 e5d4 g1f3 g8f6 f3d4 f6e4 e2e3 d8f6 d1f3 f6g6 h2h3 a7a5 b4a5 b8d7 f3g4 g6f6 g4f3 f6e5 f3f5 e5f5 d4f5 a8a5 c1b2 f7f6 f5d4 d7b6 b1c3 e4c5 d4b3 a5a8 b3c5 f8c5 f1d3 e8d8 a3a4 c5b4');
INSERT INTO evaluations VALUES(17474,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(17474,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(17474,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(17474,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(17474,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(17474,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(17474,36,'cp 50','b5c6 d7c6 d1h5 c5c4 g1f3 g8f6 h5c5 c8e6 f3g5 d8d6 c5e3 f6d7 b2b3 e6g4 b3c4 e7e5 e1g1 b7b5 e3g3 g4h5 d2d3 b5c4 d3c4 f8e7 g3h3 d7f6 g5f3 d6c5 a1b1 e8g8 b1b3 h5g4 c1e3 c5e3 f2e3 g4h3 g2h3 h7h6 f3e5');
INSERT INTO evaluations VALUES(17471,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(17471,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(17471,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17471,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(17471,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(17471,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(17471,36,'cp 64','d7c6 f2f4 e7e5 f4e5 g8e7 g1f3 c8g4 e1g1 e7g6 d1e1 d8d7 d2d3 c5c4 d3c4 g4f3 g2f3 g6e5 c1f4 e5g6 a1d1 f8c5 g1h1 d7e7 f4g3 e8g8 f3f4 a8d8 d1d8 f8d8 e4e5 c5d4 b2b3 f7f6 e5f6 e7e1 g3e1 d4f6 c3e4');
INSERT INTO evaluations VALUES(17598,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(17598,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(17598,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17598,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(17598,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(17598,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(17598,36,'cp 78','g1f3 d8c7 d2d4 g8f6 e1g1 c5d4 d1d4 c6c5 d4d3 e7e6 f1d1 f8e7 e4e5 f6d5 c1g5 e7g5 f3g5 h7h6 g5e4 c7e5 e4c5 e8g8 c3d5 e6d5 c2c3 c8e6 d1d2 a8c8 c5b3 e5f6 d3g3 f8e8 a1d1 f6g5 a2a3 e6g4 f2f3 g4e6 b3a5 g5g3 h2g3 g8h8 a5b7');
INSERT INTO evaluations VALUES(17597,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(17597,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(17597,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(17597,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(17597,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(17597,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(17597,36,'cp 62','d8c7 d2d4 g8f6 e1g1 c5d4 d1d4 c6c5 d4d3 e7e6 e4e5 f6d7 c1f4 b7b5 a2a4 c5c4 d3e2 b5b4 c3e4 c8b7 a1d1 a8c8 d1d4 b7e4 d4e4 d7b6 f3d2 c4c3 b2c3 b4c3 d2b3 f8e7 e2g4 g7g6 f4h6 e7f8 h6f8 e8f8 a4a5 b6d5 f1e1');
INSERT INTO evaluations VALUES(27462,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(27462,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(27462,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(27462,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(27462,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(27462,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(27462,36,'cp 77','f8e7 f1d1 b7b6 e4e5 d6e5 d4e5 d7e6 e5c7 e8g8 h2h3 e7b4 c7g3 a8c8 c1d2 e6f5 a1c1 f8e8 g3d3 f5h5 a2a3 c8d8 g2g4 h5h4 d3f3 b4c5 d2f4 h7h5 g4g5 f6h7 d1d8 e8d8 c3e4 c5e7 f3g3 h4g3 f2g3 f7f5 g5f6 h7f6 c1e1 h5h4 g3h4 f6e4 e1e4 e7f6 c2c3 f6h4 g1f1');
INSERT INTO evaluations VALUES(27345,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(27345,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(27345,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(27345,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(27345,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(27345,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(27345,36,'cp 89','f1d1 e7d8 a2a4 d7e6 f2f3 e8g8 a4a5 f8e8 c1e3 e6c8 h2h3 h7h6 d4d2 d8e7 d2f2 a7a6 c3e2 f6d7 f2g3 g8h8 e3d4 e7f6 e2f4 c8d8 f4d5 f6d4 d1d4 c7c6 d5f4 d7f6 f4d3 d6d5 e4e5 f6d7 f3f4 d7c5 d3c5 d8e7 a1d1 e7c5');
INSERT INTO evaluations VALUES(36611,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(36611,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(36611,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(36611,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(36611,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(36611,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(36611,36,'cp -18','b8c6 e1g1 f6h5 f4g5 f7f6 g5h4 e7e5 f1e1 g8h8 b1d2 d8e8 e3e4 c8g4 d1c2 h5f4 d3f1 g6g5 h4g3 e8h5 h2h3 g4d7 b2b4 c6d8 a2a4 d8f7 c2d1 h5g6 g3f4 e5f4 f3h2 a8e8 f1d3 h7h5 d2f3 g5g4 f3h4 g6g5 h4f5 d7f5 e4f5 e8e1 d1e1 g4h3');
INSERT INTO evaluations VALUES(3505,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(3505,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(3505,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(3505,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(3505,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(3505,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(3505,36,'cp -19','e1g1 f6h5 f4g5 f7f6 g5h4 e7e5 b1d2 d8e8 e3e4 a7a5 a2a4 g8h8 d3b5 e8f7 b5c4 f7e7 f1e1 g7h6 d2f1 e7g7 f1e3 c8d7 d4e5 c6e5 f3e5 f6e5 e3d5 d7e6 c4e2 f8f7 e2g4 e6g4 d1g4 h6f4 f2f3 a8f8 g1h1 f4d2 e1e2 d2h6 h4g5 h6g5 g4g5 h7h6');
INSERT INTO evaluations VALUES(45436,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45436,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45436,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45436,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45436,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45436,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45436,36,'cp 0','a7a6 g1f3 e7e6 f1d3 b7b5 c1d2 c8b7 e1g1 f8d6 a2a3 b8c6 e3e4 d5e4 c3e4 d6e7 d2c3 e8g8 a1c1 f6d5 c3d2 d5f6');
INSERT INTO evaluations VALUES(43432,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43432,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43432,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43432,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43432,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43432,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43432,36,'cp 0','f1d3 e7e6 c1d2 b7b5 g1f3 c8b7 e1g1 f8d6 a2a3 b8c6 e3e4 d5e4 c3e4 d6e7 d2c3 f6d5 a1c1 e8g8 c3d2 d5f6 d2c3');
INSERT INTO evaluations VALUES(52740,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(52740,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52740,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(52740,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52740,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(52740,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(52740,36,'cp 12','g8f6 d2d4 c8f5 b1c3 e7e6 f3h4 f5g6 h4g6 h7g6 d1b3 d8c7 g2g3 b8d7 f1g2 a7a5 e1g1 f8e7 a2a4 e8g8 c1d2 d5c4 b3c4 e6e5 c3e4 d7b6 c4b3 e5d4 e3d4 b6d5 e4f6 e7f6 g2d5 c6d5 d2e3 c7d7 f1c1 a8c8 b3b5 d7b5');
INSERT INTO evaluations VALUES(44891,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(44891,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44891,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44891,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(44891,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44891,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(44891,36,'cp 27','d2d4 c8f5 f1d3 f5g6 e1g1 e7e6 d1c2 d8c7 b1d2 g6d3 c2d3 f8e7 b2b3 e8g8 e3e4 d5e4 d2e4 f8d8 c1g5 f6e4 d3e4 h7h6 g5f4 e7d6 f4g3 d6g3 h2g3 b8d7 e4e3 d8e8 f1d1 e6e5 d4d5 c6d5 d1d5 e5e4 f3d4 d7f6');
INSERT INTO evaluations VALUES(57570,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(57570,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(57570,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(57570,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(57570,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(57570,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(57570,36,'cp 62','d7d6 g1f3 b8c6 f1b5 d6e5 b5c6 b7c6 f3e5 d8d5 e5f3 d5e6 d1e2 e6e2 e1e2 e7e6 d2d3 c6c5 h1e1 f8d6 b1d2 g8e7 b2b3 h7h6 d2c4 h8g8 e2f1 g7g5 h2h3 e7d5 c1b2 e8e7 a1d1 a7a5 d3d4 c8b7 d4c5 d6c5 f3e5 a5a4 e5d3');
INSERT INTO evaluations VALUES(54681,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(54681,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(54681,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(54681,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(54681,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(54681,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(54681,36,'cp 80','g1f3 b8c6 f1b5 d6e5 f3e5 d8d5 b5c6 b7c6 e5f3 d5e6 d1e2 e6e2 e1e2 e7e6 d2d3 f8d6 b1d2 c6c5 h1e1 g8e7 b2b3 e8g8 c1b2 e7c6 e2d1 f8e8 d2c4 c8a6 c4e5 a6b7 g2g3 f5f4 e5c6 b7c6 f3e5 c6d5 e1e2 e8f8 d1d2 f8f5 g3g4 f5f8 a1e1 a7a5');
INSERT INTO evaluations VALUES(54680,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(54680,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(54680,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(54680,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(54680,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(54680,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(54680,36,'cp 74','b8c6 f1b5 d6e5 b5c6 b7c6 f3e5 d8d5 e5f3 e7e6 e1g1 g8f6 b2b3 f8d6 f1e1 e8g8 c1b2 c6c5 d1e2 f8e8 b1c3 d5b7 f3e5 b7b4 c3b5 b4f4 b5a3 c8b7 a3c4 a8d8 d2d3 b7d5 f2f3 f4h6 e2e3 h6e3 c4e3 d5b7 e3c4 f6d5 g1f1 b7a8 g2g3 d5b6 c4a5 h7h6 e1e2 g7g5');
INSERT INTO evaluations VALUES(20899,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(20899,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(20899,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(20899,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(20899,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(20899,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(20899,36,'cp 52','f1b5 d6e5 b5c6 b7c6 f3e5 d8d5 e5f3 g7g5 d2d4 d5e4 c1e3 f5f4 b1c3 e4f5 f3e5 f8g7 d1f3 g7e5 f3c6 f5d7 c6a8 f4e3 f2e3 e5g7 a8a7 g8h6 e1c1 e8g8 e3e4 h6g4 d1d2 g4e3 h1e1 e3f1 d2d3 f1h2 c1b1 e7e5 a2a4 e5d4 e4e5 h7h6 e5e6 d7d6 e6e7 f8e8 b2b3');
INSERT INTO evaluations VALUES(20897,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(20897,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(20897,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(20897,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(20897,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(20897,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(20897,36,'cp 59','d6e5 f3e5 d8d5 b5c6 b7c6 e5f3 g7g5 d2d4 d5e4 c1e3 f5f4 b1c3 e4f5 f3e5 f8g7 d1f3 g7e5 f3c6 f5d7 c6a8 f4e3 f2e3 e5g7 a8a7 g8f6 e1c1 f6d5 c3d5 d7d5 a7c7 e8g8 c7e7 d5a2 e7a3 a2c4 a3d3 c4b4 d3c3 b4b5 c3b3 b5b3 c2b3 c8g4 d1d3 g4f5 d3c3 f5e4 b3b4 f8a8 c1d2 g7f8 c3c4');
INSERT INTO evaluations VALUES(20875,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(20875,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(20875,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(20875,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(20875,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(20875,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(20875,36,'cp 64','b5c6 b7c6 f3e5 d8d5 e5f3 g7g5 b1c3 d5e6 d1e2 e6e2 c3e2 g5g4 f3e5 f8g7 d2d4 c6c5 c1f4 g8f6 d4c5 f6h5 e5d3 h5f4 e2f4 e8f7 e1g1 e7e5 f4h5 e5e4 h5g7 e4d3 g7h5 d3c2 f1c1 c8a6 c1c2 a8d8 h5f4 d8d4 g2g3 h8d8 h2h4 g4h3 f4h3 a6b5 a1e1 d4d2 h3g5 f7g6 c2d2 d8d2 g5e6');
INSERT INTO evaluations VALUES(19967,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(19967,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(19967,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(19967,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(19967,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(19967,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(19967,36,'cp -27','c8g4 f1e2 g7g6 c1g5 f8g7 b1c3 a7a6 e1g1 g8f6 f3d2 g4e2 d1e2 b7b5 f2f4 c6d4 e2d1 e8g8 g1h1 b5b4 c3e2 h7h6 g5h4 d4c6 d2f3 a6a5 h2h3 a5a4 g2g4 d8d7 d1d2 f6e8 a1e1 e7e6 f4f5 e6f5 g4f5');
INSERT INTO evaluations VALUES(26340,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(26340,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(26340,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(26340,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(26340,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(26340,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(26340,36,'cp -12','f1e2 g7g6 c1g5 g8f6 b1c3 f8g7 e1g1 a7a6 f3d2 g4e2 d1e2 c6d4 e2d1 b7b5 f2f4 e8g8 f4f5 b5b4 c3d5 e7e6 d5f6 g7f6 g5f6 d8f6 f5e6 f6e6 d2f3 e6f6 d1d2 f6g7 f1f2 a8e8 a1e1 a6a5 f3d4 g7d4 g1h1 g8g7 h2h3 a5a4 a2a3 f7f6 a3b4 c5b4');
INSERT INTO evaluations VALUES(26339,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(26339,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(26339,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(26339,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(26339,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(26339,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(26339,36,'cp -31','g7g6 c1g5 f8g7 b1c3 g8f6 e1g1 e8g8 d1d2 a7a6 h2h3 g4f3 e2f3 a8b8 f3d1 b7b5 a1c1 e7e6 b2b3 d8a5 g5f4 f8d8 f4g5 c6d4 f2f4 h7h6 g5h4 d8e8 g1h1 b8a8 d1f3 f6h7 h4f2 d4f3 g2f3 g8h8 f2e3 g7c3 d2c3 a5c3');
INSERT INTO evaluations VALUES(45828,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(45828,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(45828,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(45828,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45828,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45828,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45828,36,'cp 19','f8d6 f1d3 e8g8 e1g1 b8a6 c2c3 c8g4 h2h3 g4h5 d3f5 a6c7 d1d3 h5g6 f3h4 f8e8 d3f3 c7d5 h4g6 h7g6 f5d3 f6f5 g2g3 f5f4 c1d2 f4g3 f2g3 d8c7 g1g2 e8e6 d3c4 b7b5 c4b3 a8e8 a1e1 e6e1 f1e1 e8e1 d2e1 d5f6 e1f2 c7d7 a2a4 a7a6 a4b5 a6b5');
INSERT INTO evaluations VALUES(40533,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40533,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40533,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40533,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40533,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(40533,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40533,36,'cp 17','f1d3 e8g8 e1g1 c8g4 f1e1 f8e8 e1e8 d8e8 h2h3 g4h5 c1e3 b8d7 g2g4 h5g6 f3h4 g6d3 d1d3 a8d8 c2c4 d7c5 d3c3 c5e6 a1d1 d6e5 d1d2 e5c7 g1g2 g7g6 h4f3 g8g7 b2b3 e6f4 e3f4 c7f4 d2d1 e8e4 c3d3 e4d3 d1d3 f4c7');
INSERT INTO evaluations VALUES(40531,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40531,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40531,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40531,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(40531,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(40531,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(40531,36,'cp 20','e8g8 e1g1 c8g4 f1e1 f8e8 e1e8 d8e8 c1e3 b8a6 h2h3 g4h5 c2c4 a6c7 d1b3 h5f3 g2f3 b7b5 c4b5 c6b5 f3f4 e8d7 d3f1 a8d8 d4d5 d7f5 f1g2 d6f4 e3a7 f4d6 a1d1 f5e5 g1f1 g7g6 b3c3 e5c3 b2c3 f6f5 a7b6 d8c8');
INSERT INTO evaluations VALUES(36130,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36130,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36130,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36130,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(36130,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(36130,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36130,36,'cp 20','e1g1 b8a6 c2c3 f8e8 h2h3 c8e6 f1e1 d8d7 c1d2 a6c7 c3c4 e6f5 d3f5 d7f5 d1b3 e8e1 a1e1 b7b5 e1c1 a7a5 a2a4 h7h5 b3c2 f5c2 c1c2 b5a4 c4c5 d6f8 c2c4 a4a3 b2a3 a5a4 g1f1 g7g6 f1e2 c7b5 d2b4 a8e8 e2d3 f6f5 g2g3 b5c7');
INSERT INTO evaluations VALUES(36129,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36129,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36129,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(36129,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36129,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36129,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36129,36,'cp 15','c8g4 h2h3 g4h5 d3f5 b8a6 c2c3 a6c7 d1d3 h5g6 f3h4 f8e8 d3f3 c7d5 h4g6 h7g6 f5d3 f6f5 g2g3 f5f4 g1g2 d6c7 c1d2 f4g3 f2g3 d8d7 a2a4 e8e6 a1e1 a8e8 e1e6 e8e6 f1f2 a7a6 d3c4 b7b5 c4b3 e6f6 f3g4 f6e6 a4b5 a6b5 f2e2');
INSERT INTO evaluations VALUES(27811,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(27811,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(27811,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(27811,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(27811,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(27811,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(27811,36,'cp 26','h2h3 g4h5 d3f5 f8e8 d1d3 h5g6 f3h4 b8a6 c2c3 a6c7 d3f3 c7e6 g2g3 g6f5 h4f5 d6f8 h3h4 g7g6 f5h6 g8g7 h6g4 f6f5 g4e5 d8c7 e5d3 h7h5 f1e1 a8d8 a2a4 a7a5 c1d2 c7b6 g1g2 b6b3 d3f4 e6f4 d2f4 b3c4 e1e8 d8e8');
INSERT INTO evaluations VALUES(27810,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(27810,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(27810,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(27810,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(27810,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27810,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(27810,36,'cp 19','g4h5 d3f5 b8a6 c2c3 f8e8 d1d3 h5g6 f3h4 a6c7 d3f3 c7e6 g2g3 g6f5 h4f5 d6f8 f1d1 g7g6 f5h6 g8g7 d4d5 c6d5 d1d5 d8c7 h6g4 f6f5 c1h6 g7h8 h6f8 e8f8 g4e5 a8d8 a1d1 d8d5 d1d5 h8g7 c3c4 f8d8 f3e3 e6g5 e5f3 d8d5 c4d5 g5f3 e3f3 c7c2 g1g2');
INSERT INTO evaluations VALUES(46261,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(46261,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(46261,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(46261,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(46261,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(46261,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(46261,36,'cp 68','e4d5 f6d5 c3d5 d8d5 c1e3 c5d4 d1d4 d5d4 f3d4 a7a6 c2c3 f8e7 f1e2 e6e5 d4b3 b8c6 e2c4 c8f5 a1d1 e8g8 e1g1 a8c8 c4d5 f8d8 d1d2 h7h6 f1d1 b7b5 e3b6 d8e8 h2h3 e7g5 b6e3 g5e7 d5f3 e5e4 f3e2 c6e5 b3a5');
INSERT INTO evaluations VALUES(55548,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(55548,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(55548,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(55548,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(55548,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(55548,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(55548,36,'cp -64','e5f4 g1f3 h7h6 h2h4 b8c6 d2d3 g8f6 c1f4 f6h5 f4e3 c6a5 b1c3 a5c4 d3c4 f8e7 d1d2 c8g4 e1c1 e8g8 c4c5 f8e8 e4e5 d6c5 d2f2 d8c8 e3c5 c8f5 c5e7 e8e7 h1e1 g4f3 f2f3 f5f3 g2f3 a8e8 c3d5 e7e5 e1e5 e8e5 d5c7');
INSERT INTO evaluations VALUES(55661,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(55661,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(55661,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(55661,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(55661,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(55661,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(55661,36,'cp -69','g1f3 h7h6 h2h4 g8f6 b1c3 b8c6 d2d4 f6h5 c3d5 h5g3 h1h2 c6e7 d5f4 g3e4 d1e2 e4f6 c1d2 d6d5 c4b3 c8g4 e1c1 c7c6 d1e1 d8c7 e2d3 g4f3 d3f3 e8c8 h2h1 c8b8 f4d3 e7c8 h1f1 f8d6 d3e5 d8e8 c2c3 f6e4 e5f7 e4d2 c1d2 e8e1 d2e1 h8f8 e1d1 d6e7 h4h5 c8d6');
INSERT INTO evaluations VALUES(57667,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(57667,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(57667,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(57667,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(57667,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(57667,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(57667,36,'cp 37','g8f6 e4e5 f6d5 c3c4 d5b6 c4c5 b6d5 b1c3 e7e6 d2d4 d7d6 c3e4 d6e5 d4e5 h7h6 g1f3 f8g7 b2b3 c8d7 c1b2 d7c6 d1d2 b8d7 f1c4 g6g5 e1c1 g5g4 f3d4 d7e5 f2f4 e5c4 b3c4 d5e7 e4d6 c7d6 d4c6 e7c6 b2g7');
INSERT INTO evaluations VALUES(15935,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(15935,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(15935,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15935,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(15935,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(15935,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(15935,36,'cp 19','f8e7 g1f3 e8g8 a1c1 h7h6 g5h4 c7c5 c4d5 f6d5 h4e7 d5e7 f1e2 b7b6 d4c5 d7c5 e1g1 c8b7 b2b4 c5d7 c3b5 d7f6 a2a3 e7d5 d1d4 a7a6 b5c7 a8c8 c7d5 f6d5 d4b2 d8e7 f3e5 d5f6 f1d1 b6b5 h2h3 b7d5 c1c8 f8c8 d1c1 c8c7 e5d3 d5c4 d3c5 c4e2 b2e2');
INSERT INTO evaluations VALUES(9580,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(9580,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(9580,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(9580,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(9580,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9580,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(9580,36,'cp 17','a1c1 c7c6 c4d5 f6d5 g5e7 d8e7 d1c2 e8g8 g1f3 d5c3 c2c3 f8e8 c1d1 e6e5 f3e5 d7e5 d4e5 c8e6 f2f4 e6a2 f1d3 a7a5 e1g1 e7b4 c3b4 a5b4 g1f2 a2b3 d1d2 e8d8 f1c1 g8f8 e3e4 a8a5 h2h4 f8e7 h4h5 c6c5 f2e3 d8d4 f4f5');
INSERT INTO evaluations VALUES(24891,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(24891,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(24891,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(24891,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(24891,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(24891,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(24891,36,'cp 151','c4d5 f6d5 e2e4 d5c3 b2c3 a7a6 a1b1 e7e5 b1b7 e5d4 d1a4 g4d7 c3d4 f8d6 a4c2 e8g8 a2a3 c6a5 b7b1 c7c5 d4c5 d8c7 c1e3 f8c8 b1c1 d6f4 c2d2 f4e3 d2e3 a5b7 f1c4 c7a5 e3d2 c8c5 d2a5 c5a5 c4d5 a8c8 c1c8 d7c8 e1d2 b7d6 h1b1 g8f8 f3d4 a5a3 b1b8 a3a4 d4c6 f8e8');
INSERT INTO evaluations VALUES(24793,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(24793,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(24793,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(24793,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(24793,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(24793,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(24793,36,'cp 146','f6d5 e2e4 g4f3 g2f3 d5c3 b2c3 e7e5 d4d5 c6b8 h1g1 b8d7 a1b1 g7g6 b1b7 f8d6 c1g5 f7f6 g5e3 a8b8 b7b8 d8b8 d1c2 e8g8 h2h4 b8b7 h4h5 g6g5 f1h3 d7c5 h3f5 f8b8 e1d2 h7h6 d2e2 b7b5 c3c4 b5b2 g1c1 b2a3 c1b1 b8b1 c2b1 a3a4 e2f1 a4c4 f1g2 c5d3 b1b8 g8g7');
INSERT INTO evaluations VALUES(25012,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(25012,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(25012,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(25012,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(25012,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(25012,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(25012,36,'cp 140','e2e4 g4f3 g2f3 d5c3 b2c3 e7e5 d4d5 c6b8 h1g1 g7g6 a1b1 b8d7 b1b7 f8d6 c1g5 f7f6 g5h6 e8f7 h6e3 d8e7 b7b1 a8b8 f1b5 d7c5 e1f1 b8b6 b5c6 h8b8 b1b6 a7b6 f1g2 f7g7 h2h4 h7h5 d1c2 g7h7 a2a4 b8c8 g1b1 h7h8');
INSERT INTO evaluations VALUES(25262,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(25262,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(25262,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(25262,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(25262,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(25262,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(25262,36,'cp 56','f1e1 d8d6 b1c3 d5c3 b2c3 f7f6 h2h3 g4h5 a2a4 e8c8 c1a3 d6d7 a3f8 h8f8 f3e5 c6e5 d1h5 e5c4 d3c4 g7g6 h5c5 b7b6 c5e3 d7c6 e3e7 f8h8 a4a5 b6a5 e7e3 h8e8 e3a7 e8e1 a1e1 d8e8 e1a1 e8e5 a1a5 e5a5 a7a5 c6c4 a5b4 c4e6 g1h2 f6f5 b4d4 c8b7 c3c4 e6e2 d4b2 b7c8 b2h8 c8b7 h8h7 e2c2 h7g6 c2f2');
INSERT INTO evaluations VALUES(25263,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(25263,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(25263,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(25263,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(25263,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(25263,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(25263,36,'cp 56','d8d6 b1c3 d5c3 b2c3 f7f6 h2h3 g4h5 a2a4 e8c8 a4a5 a7a6 c1a3 d6d7 a3f8 d8f8 f3e5 c6e5 d1h5 e5c4 d3c4 f8e8 e1e8 d7e8 h5c5 c8b8 c5d5 e8e5 d5e5 f6e5 a1e1 h8e8 g1f1 b8a7 f1e2 e8d8 h3h4 b7b6 e1a1 a7b7 e2e3 d8d6 c4c5 d6c6 c5b6 c7b6 a5b6 c6c3 e3d2 c3c6');
INSERT INTO evaluations VALUES(39923,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(39923,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(39923,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(39923,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(39923,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(39923,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(39923,36,'cp 37','d4c5 e8g8 a2a3 b4c5 g1f3 d7d5 c4d5 e6d5 c1g5 c8e6 e2e3 h7h6 g5h4 b8d7 f1d3 a8c8 e1g1 c5d6 a1d1 d8c7 d1c1 c7b6 c2e2 a7a6 e2d2 g7g5 h4g3 d6g3 h2g3 d7c5 f3d4 f6e4 d2c2 c5d3 c2d3 b6b2 d4e6 f7e6 c3e4');
INSERT INTO evaluations VALUES(39894,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(39894,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(39894,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(39894,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(39894,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(39894,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(39894,36,'cp 21','e8g8 a2a3 b4c5 g1f3 d7d5 c4d5 e6d5 c1g5 c8e6 e2e3 h7h6 g5h4 b8c6 a1d1 a8c8 f1e2 g7g5 h4g3 d8e7 e1g1 f8d8 h2h3 c5b6 c2b1 d5d4 f3d4 c6d4 e3d4 d8d4 e2f3 c8d8 b1c2 d4d1 f1d1 d8d7 g3e5 d7d1 c2d1 f6d7 e5d4');
INSERT INTO evaluations VALUES(37352,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37352,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37352,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37352,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(37352,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(37352,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37352,36,'cp 15','f6e8 c1f4 d7d6 g1f3 c8g4 h2h3 d6e5 f4e5 g4f3 d1f3 g7e5 d4e5 c7c6 f3e3 e8g7 f1d3 b8d7 e1c1 d8b6 e3b6 d7b6 b2b3 g7e6 d3c2 a7a5 g2g3 f8d8 f2f4 b6d7 c3e4 g8g7 d1d2 d7c5 e4c5 e6c5 h3h4 b7b6 h1e1 g7g8 e1d1 d8d2 c1d2 a8d8 d2e3 d8d1 c2d1 f7f6 e5f6');
INSERT INTO evaluations VALUES(57781,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(57781,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(57781,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57781,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57781,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(57781,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(57781,36,'cp 8','c1f4 d7d6 g1f3 c8g4 h2h3 d6e5 f4e5 g4f3 d1f3 b8c6 e1c1 c6e5 d4e5 d8b8 f3e3 c7c6 d1e1 f7f6 e5f6 e7f6 h3h4 e8d6 f1d3 f8e8 e3g3 e8e1 h1e1 d6f7 f2f4 b8d6 g3e3 f6f5 c4c5 d6d4 e3d4 g7d4 e1e7 d4c5 e7b7 f7d6 b7c7 a8c8 d3c4 g8h8 c7c8 d6c8 c1d2');
INSERT INTO evaluations VALUES(44794,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(44794,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(44794,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44794,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44794,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44794,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44794,36,'cp 0','b8c6 d4c6 b7c6 e1g1 f8g7 h2h3 e8g8 c1g5 d8a5 d1d2 a8b8 a1b1 f8e8 b2b3 c8e6 c3a4 a5d2 g5d2 f6d7 f1e1 d7e5 d3f1 c6c5 f2f4 e5c6 c2c3 e8d8 b1d1 c6a5 f1e2 e6d7 a4b2 d7c6 e2f1 g7f6 d1b1 f6g7');
INSERT INTO evaluations VALUES(5701,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(5701,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(5701,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(5701,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(5701,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(5701,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(5701,36,'cp 102','c1e3 g8f6 d1d2 a7a6 d4d5 c6b8 e3h6 g7h6 d2h6 e7e5 h2h3 b8d7 f1e2 b7b5 a2a3 d8e7 e1c1 d7b6 f3d2 f6d7 d2b3 e7f6 h3h4 b6a4 c3a4 b5a4 b3a5 f6f2 e2f3 f2g3 d1f1 d7c5 h4h5 g3f4 h6f4 e5f4 h5g6 f7g6 e4e5 d6e5 a5c4 e8g8 c4e5');
INSERT INTO evaluations VALUES(5700,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(5700,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(5700,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(5700,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(5700,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(5700,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(5700,36,'cp 90','g8f6 d1d2 a7a6 d4d5 c6b8 e3h6 g7h6 d2h6 e7e5 a2a4 b8d7 f1e2 f6g4 h6d2 f7f5 f3g5 d8e7 e4f5 g6f5 e1c1 d7c5 f2f4 h7h6 e2g4 f5g4 g5e4 c5e4 c3e4 c8d7 e4g3 e8c8 h1e1 h6h5 f4e5 h5h4 g3f1 d6e5 d2c3 e7g5 c1b1 c8b8 e1e5');
INSERT INTO evaluations VALUES(8748,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(8748,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(8748,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(8748,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(8748,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(8748,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(8748,36,'cp 101','d1d2 a7a6 d4d5 c6b8 e3h6 g7h6 d2h6 e7e5 h2h3 d8e7 f1e2 b8d7 e1c1 b7b5 a2a3 e7f8 h6d2 d7c5 d2e3 h7h5 f3g5 f8h6 f2f4 c5d7 h1f1 e5f4 e3f4 e8g8 g2g4 h5g4 e2g4 f6h7 h3h4 d7e5 g4c8 a8c8 c1b1 c8e8 f4g3 h7g5 h4g5 h6g7');
INSERT INTO evaluations VALUES(17562,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(17562,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(17562,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(17562,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(17562,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17562,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(17562,36,'cp 43','d2d4 c5d4 f3d4 e7e6 d4c6 b7c6 d1e2 f8e7 c1f4 g8f6 e1c1 d6d5 h1e1 c8b7 g2g4 e7b4 g4g5 b4c3 b2c3 f6e4 c4d3 e4c5 e2e3 d8a5 f4d6 c5a4 e3f4 e8c8 d6b4 a5c7 f4c7 c8c7 b4a5 c7d6 a5d8 h8d8 c1d2 e6e5 f2f3 a4c5 h2h4 c5e6 d3h7 d8h8 d1b1 h8h7 b1b7 h7h4');
INSERT INTO evaluations VALUES(17570,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(17570,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(17570,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17570,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(17570,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(17570,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(17570,36,'cp 58','c5d4 f3d4 e7e6 d4c6 b7c6 d1e2 g8f6 c1g5 f8e7 e1c1 d8c7 h1e1 h7h6 g5h4 e8g8 e4e5 d6e5 e2e5 c7e5 e1e5 g7g5 h4g3 f6h5 c3a4 h5g3 f2g3 a8b8 c4e2 g8g7 a2a3 f8d8 d1d8 e7d8 e5e4 d8f6 e4c4 c6c5 c4c5 f6d4 c5c4 d4e3 c1b1 c8b7');
INSERT INTO evaluations VALUES(34773,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(34773,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(34773,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(34773,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(34773,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(34773,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(34773,36,'cp 75','c8e6 b1c3 e6c4 d3c4 c7c6 d1e2 f6e6 b2b3 g8f6 c1a3 b8d7 e1c1 f8e7 h2h4 d7c5 c1b1 e6g4 a3c5 d6c5 e2e3 g4f4 e3f4 e5f4 e4e5 f6g4 c3e4 a8d8 b1c1 d8d1 h1d1 h7h5 c1b1 f7f6 e5f6 e7f6 c2c3 f6e7 f3g5');
INSERT INTO evaluations VALUES(55505,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(55505,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(55505,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55505,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(55505,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55505,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(55505,36,'cp 35','e2e4 f8e7 c1e3 g8f6 f1d3 b8a6 f2f3 a6b4 d3f1 c7c6 c2c4 e8g8 a2a3 b4a6 b1c3 c6d5 c4d5 f6h5 g2g3 c8d7 g1h3 g7g6 e3h6 d8b6 b2b4 f8c8 d1d2 b6d4 a1c1 d4d2 e1d2 f7f5 f1b5 c8c3 b5d7 c3a3 d7e6 g8h8');
INSERT INTO evaluations VALUES(10304,30,'cp 612',NULL);
INSERT INTO evaluations VALUES(10304,31,'cp 621',NULL);
INSERT INTO evaluations VALUES(10304,32,'cp 613',NULL);
INSERT INTO evaluations VALUES(10304,33,'cp 609',NULL);
INSERT INTO evaluations VALUES(10304,34,'cp 621',NULL);
INSERT INTO evaluations VALUES(10304,35,'cp 619',NULL);
INSERT INTO evaluations VALUES(10304,36,'cp 617','d4c6 b7c6 e3c5 d7d6 c5d4 e8g8 b1c3 c6c5 d4f6 d8f6 d1f3 f6h6 f3e3 h6e3 f2e3 f8e8 e1c1 e8e5 f1e2 c8b7 h1f1 a8f8 e2c4 g8h8 f1f2 f7f6 a2a3 b7c8 c1b1 c8g4 d1f1 c7c6 b1a1 g4e6 c4e6 e5e6 f1d1 d6d5 e4d5 c6d5 c3d5');
INSERT INTO evaluations VALUES(9960,30,'cp 594',NULL);
INSERT INTO evaluations VALUES(9960,31,'cp 574',NULL);
INSERT INTO evaluations VALUES(9960,32,'cp 594',NULL);
INSERT INTO evaluations VALUES(9960,33,'cp 594',NULL);
INSERT INTO evaluations VALUES(9960,34,'cp 603',NULL);
INSERT INTO evaluations VALUES(9960,35,'cp 597',NULL);
INSERT INTO evaluations VALUES(9960,36,'cp 597','b7c6 e3c5 d7d6 c5d4 e8g8 b1c3 c6c5 d4f6 d8f6 d1f3 f6h4 f3g3 h4h6 g3e3 h6e3 f2e3 f8e8 e1c1 e8e5 f1e2 c8b7 h1f1 a8f8 f1f2 b7e4 e2f3 e4f3 f2f3 f8e8 c3d5 e8b8 c2c4 e5d5 c4d5 f7f6 b2b3 g8f7 d1f1 b8e8 c1b2 e8e5');
INSERT INTO evaluations VALUES(7603,30,'cp 615',NULL);
INSERT INTO evaluations VALUES(7603,31,'cp 612',NULL);
INSERT INTO evaluations VALUES(7603,32,'cp 620',NULL);
INSERT INTO evaluations VALUES(7603,33,'cp 598',NULL);
INSERT INTO evaluations VALUES(7603,34,'cp 613',NULL);
INSERT INTO evaluations VALUES(7603,35,'cp 613',NULL);
INSERT INTO evaluations VALUES(7603,36,'cp 617','e3c5 d7d6 c5d4 e8g8 b1c3 c6c5 d4f6 d8f6 d1f3 f6h6 f3e3 h6e3 f2e3 a8b8 e1c1 c8g4 d1d2 f8e8 f1b5 e8e5 c1b1 g8h8 b1a1 f7f6 h1f1 g4e6 b5e2 b8e8 a2a3 e6g8 e2f3 h7h5 f1d1 h5h4 g2g4 g8h7 f3g2 e5e4 g2e4 h7e4 c3b5 h4h3 b5c7');
INSERT INTO evaluations VALUES(7579,30,'cp 608',NULL);
INSERT INTO evaluations VALUES(7579,31,'cp 605',NULL);
INSERT INTO evaluations VALUES(7579,32,'cp 595',NULL);
INSERT INTO evaluations VALUES(7579,33,'cp 601',NULL);
INSERT INTO evaluations VALUES(7579,34,'cp 598',NULL);
INSERT INTO evaluations VALUES(7579,35,'cp 605',NULL);
INSERT INTO evaluations VALUES(7579,36,'cp 590','d7d6 c5d4 e8g8 b1c3 c6c5 d4f6 d8f6 d1f3 f6h6 f3e3 h6e3 f2e3 a8b8 e1c1 c8g4 d1d2 f8e8 c1b1 e8e5 f1b5 g4e6 b5e2 g8h8 h1f1 f7f6 b1a1 h7h5 a2a3 e6g8 d2d1 g8f7 f1f2 f7g8 g2g3 g8e6 e2f1 b8b6 f1g2 e6g4 d1d2 c7c6 d2d6 c5c4');
INSERT INTO evaluations VALUES(53088,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(53088,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(53088,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53088,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(53088,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(53088,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53088,36,'cp 8','g1f3 b8c6 d2d4 c5d4 f3d4 g7g6 b1c3 g8f6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d2 c8e6 e1g1 a7a6 a1d1 b7b5 h2h3 a8c8 e3d4 e6c4 f1e1 c4e2 d2e2 d8c7 d4f6 g7f6 c3d5 c7c2 e2c2 c8c2 d5f6 e7f6 d1d6 c2b2 d6a6 b2a2 a6f6');
INSERT INTO evaluations VALUES(53086,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53086,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(53086,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53086,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53086,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(53086,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(53086,36,'cp 2','b8c6 d2d4 c5d4 f3d4 g7g6 f1e2 f8g7 c1e3 g8f6 b1c3 e8g8 e1g1 a7a6 d1d2 c8d7 f1e1 a8c8 h2h3 b7b5 a1d1 d8c7 d4f3 d7e6 g1h1 f8e8 f3g5 e6c4 e2c4 b5c4 b2b4 c4b3 c2b3 c7b7 b3b4 c6e5 c3d5 f6d5');
INSERT INTO evaluations VALUES(20036,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(20036,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(20036,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(20036,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(20036,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(20036,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(20036,36,'cp 12','b1c3 g8f6 d2d4 c5d4 f3d4 g7g6 f1e2 f8g7 c1e3 e8g8 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 a1d1 g8g7 c2c3 c8e6 d1d4 b7b6 f1d1 a8c8 f2f3 g7f6 g1f2 a7a5 a3a4 c8c5 g2g3 f8a8 h2h4 e6b3 d1d2 a8c8 e2b5 b3e6');
INSERT INTO evaluations VALUES(20033,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(20033,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(20033,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(20033,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(20033,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(20033,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(20033,36,'cp 11','g8f6 d2d4 c5d4 f3d4 g7g6 f1e2 f8g7 c1e3 e8g8 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 c2c3 c8e6 d1d4 g7f6 a1d1 a7a5 a3a4 a8c8 e2b5 h7h6 f2f3 c8c5 h2h4 b7b6 c3c4 g6g5 b2b3 g5h4 d4h4');
INSERT INTO evaluations VALUES(13337,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13337,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13337,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(13337,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13337,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13337,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(13337,36,'cp 14','d2d4 c5d4 f3d4 g7g6 f1e2 f8g7 c1e3 e8g8 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 a1d1 g8g7 c2c3 c8e6 d1d4 a8b8 g2g4 e6b3 d4b4 b3e6 f2f4 b7b6 b4d4 b8c8 f4f5 e6b3 d4d7 f8d8 d7e7 d8d2 f5g6');
INSERT INTO evaluations VALUES(14199,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(14199,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(14199,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(14199,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(14199,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(14199,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(14199,36,'cp 26','e2e3 f6e4 g5h4 h7h6 c2c4 g7g5 h4g3 e7e5 b1c3 e5d4 f3d4 f8b4 a1c1 e8g8 d4c6 b7c6 f1d3 b4c3 b2c3 c8f5 c4d5 d8d5 d3b1 d5d1 c1d1 f8d8 d1c1 e4g3 h2g3 f5b1 c1b1 g8g7 e1e2 a8b8 a2a4 f7f5 e2f3 b8b1 h1b1 g7f6 e3e4 h6h5 e4f5');
INSERT INTO evaluations VALUES(14197,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(14197,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(14197,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(14197,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(14197,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(14197,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(14197,36,'cp 30','f6e4 g5h4 h7h6 c2c4 g7g5 h4g3 e7e6 f3d2 e4g3 h2g3 f8g7 c4d5 e6d5 d2f3 c6e7 f1d3 d8d6 b1c3 c7c6 a1c1 c8g4 d1b3 b7b6 b3c2 g4d7 e1g1 e8g8 e3e4 d5e4 c3e4 d6c7 f1e1 d7e6 d3c4 e6c4 c2c4 a8d8 g3g4 c7d7 e4g3 f8e8 f3e5 g7e5 d4e5');
INSERT INTO evaluations VALUES(21599,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21599,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21599,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(21599,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(21599,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(21599,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21599,36,'cp 11','c6b4 d3b5 c7c6 b5a4 f8d6 c2c3 b4a6 e1g1 c8g4 f1e1 g8e7 b1d2 a6c7 d2f1 e8g8 f1g3 e7g6 h2h3 g4f3 d1f3 g6h4 f3h5 g7g6 h5g5 c7e6 g5d8 a8d8 g3f1 d6f4 f1d2 e6g5 g1f1 g5e6 d2b3 f4c1 a1c1 f8e8 b3c5 e6c5 d4c5');
INSERT INTO evaluations VALUES(34961,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34961,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34961,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34961,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34961,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34961,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34961,36,'cp 0','b1d2 a7a5 a2a4 h7h5 h2h4 f8c5 f2f3 f7f6 f1d3 g8e7 g1e2 c8e6 d3c4 e6d7 c2c3 e7c8 d1c2 c8d6 c4d5 d7c6 d5c6 b8c6 d2f1 c6d8 f1e3 d8e6 e3f5 b7b6 h1d1 d6b7 f5e3 b7d6');
INSERT INTO evaluations VALUES(37850,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37850,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37850,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(37850,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37850,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(37850,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37850,36,'cp 0','d8a5 c2c3 g8f6 d1e2 e8g8 e1g1 d7d6 f2f3 a7a6 a2a4 a5c7 c1e3 b7b6 c4a2 c8b7 c3c4 f6d7 b1c3 b8c6 d4c6 b7c6 e2c2 a6a5 a1d1 d7c5 c3b5 c6b5 c4b5 e7e6 f3f4 a8c8 g1h1 f8e8 e3g1 c5d7 c2e2 d7c5 e2c2');
INSERT INTO evaluations VALUES(50939,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(50939,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(50939,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(50939,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(50939,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(50939,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(50939,36,'cp 74','b7b5 f1d3 c8b7 d4d5 g8f6 e1g1 b8d7 d5e6 f7e6 f3g5 d7c5 b2b4 c5d3 c2d3 d8d7 d1b3 b7c8 c1e3 f8e7 d3d4 h7h6 g5h3 e8g8 h3f4 f6g4 f4g6 f8e8 g6e7 d7e7 e3f4 e7f7 f4g3 g4f6 f1c1 c8d7 f2f3 f6h5 c3e2 h5g3 h2g3 e8c8 e4e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(19143,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(19143,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(19143,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(19143,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(19143,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(19143,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(19143,36,'cp 38','a7a6 b1c3 b7b5 c4b3 c8b7 c1e3 g8e7 e1c1 e7c6 d4d3 f8e7 c1b1 c6a5 h1e1 d8c7 f2f4 a5b3 a2b3 b7c6 c3d5 e6d5 e4d5 e8g8 e3d2 f8e8 d5c6 d7c6 d3e4 a8d8 f4f5 c7c8 e4f3 e7f6 d2c3 d8d1 e1d1 f6c3 b2c3 c6c5 f5f6 c5c4');
INSERT INTO evaluations VALUES(21957,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21957,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(21957,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(21957,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21957,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21957,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(21957,36,'cp 49','c6d4 f1c4 g8f6 d2d3 d4f3 d1f3 f8e7 c1e3 e8g8 d3d4 c7c6 c4b3 e5d4 e3d4 c6c5 d4e3 c8e6 e1c1 c5c4 b3a4 a7a6 e4e5 b7b5 e5f6 e7f6 c3e4 f6e5 e3c5 d8a5 c5d6 e5d6 d1d6 a5a4 c1b1 a8e8 e4g5 h7h6 g5e6 e8e6 d6e6 f7e6 f3c6');
INSERT INTO evaluations VALUES(41827,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41827,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(41827,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41827,36,'cp 0','e8g8 f1e2 c7c5 d4c5 d8a5 b1c3 d6c5 e1g1 c5e7 a2a3 b8c6 d1d2 c6e5 f4e5 f6d7 e5g3 d7f6');
INSERT INTO evaluations VALUES(36976,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36976,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36976,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36976,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36976,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36976,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36976,36,'cp 0','f1e2 c7c5 d4c5 d6c5 e1g1 d8b6 b1c3 c5e7 d1d4 b6d8 f1d1 b8c6 d4d2 f6d7 e5f3 d7f6');
INSERT INTO evaluations VALUES(16792,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(16792,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(16792,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(16792,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(16792,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(16792,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(16792,36,'cp 21','e1g1 f8e7 f1e1 d7d6 e5d6 c7d6 c1g5 g4e5 g5e7 d8e7 c4e2 e8g8 f3d4 d6d5 c2c3 c8e6 b1a3 c6d4 d1d4 e5c6 d4a4 d5d4 c3d4 e7b4 a4b4 c6b4 e1d1 f8d8 b2b3 a7a6 e2f3 e6d5 f3d5 d8d5 d1d2 b7b5 g2g3 a8c8 a1d1 f7f6 a3b1 c8c2 g1g2 d5d8 a2a4 g8f7 d2c2 b4c2 b1c3 c2d4 a4b5');
INSERT INTO evaluations VALUES(16791,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(16791,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(16791,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16791,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16791,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(16791,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(16791,36,'cp 21','f8e7 f1e1 d7d6 e5d6 c7d6 c1g5 g4e5 g5e7 d8e7 c4e2 e8g8 f3d4 d6d5 c2c3 c8e6 b1a3 c6d4 d1d4 e5c6 d4a4 d5d4 c3d4 f8d8 a1d1 e7b4 a4b4 c6b4 e2f3 d8d7 d4d5 e6d5 d1d4 a7a5 f3d5 d7d5 d4d5 b4d5 g2g3 d5b4 e1e7 b7b6 e7b7 b4a2');
INSERT INTO evaluations VALUES(10959,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(10959,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10959,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(10959,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(10959,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(10959,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(10959,36,'cp 26','f1e1 d7d6 e5d6 c7d6 c1g5 g4e5 g5e7 d8e7 c4e2 e8g8 f3d4 c8e6 b1c3 d6d5 e2f1 e7f6 c3e2 a8c8 c2c3 f8e8 h2h3 h7h6 d4c6 b7c6 d1d4 c6c5 d4f4 e5g6 f4f6 g7f6 a1d1 a7a5 b2b3 f6f5 g2g3 g8g7 f2f4 c8c7 f1g2 e8d8 h3h4');
INSERT INTO evaluations VALUES(10960,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(10960,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10960,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(10960,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(10960,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10960,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(10960,36,'cp 11','d7d6 e5d6 c7d6 c1g5 g4e5 g5e7 d8e7 c4e2 e5f3 e2f3 c8e6 f3c6 b7c6 d1d4 e8g8 b1c3 f8e8 e1e3 c6c5 d4e4 e7d7 a1e1 h7h6 c3d5 a8c8 b2b3 e6d5 e4d5 e8e3 e1e3 c8e8 e3e8 d7e8 g1f1 e8b5 d5d3 b5c6 c2c4 g7g6 f1g1 c6e8 g2g3 e8e5 d3d2 g8g7 h2h4');
INSERT INTO evaluations VALUES(34582,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34582,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34582,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34582,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34582,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34582,36,'cp 0','d2d3 b8c6 b1c3 d5d8 c1e3 h7h6 g5e4 e7e5 f2f4 e5f4 e3f4 g8f6 e4f6 d8f6 c3d5 f6d4 d1e2 e8d7 e2g4 d7e8 g4e2');
INSERT INTO evaluations VALUES(34581,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34581,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34581,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34581,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34581,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34581,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34581,36,'cp 0','b8c6 b1c3 d5d8 c1e3 h7h6 g5e4 e7e5 f2f4 e5f4 e3f4 g8f6 e4f6 d8f6 c3d5 f6d4 d1e2 e8d7 e2g4 d7e8 g4e2');
INSERT INTO evaluations VALUES(12122,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(12122,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(12122,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(12122,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(12122,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(12122,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(12122,36,'cp 62','f1d3 c5d4 c3d4 d8b6 e1g1 a7a5 f1e1 f8e7 h2h3 a5a4 a1b1 h7h6 b2b4 a4b3 b1b3 b6a7 a2a3 c6d4 f3d4 a7d4 c1b2 d4a7 d1g4 g7g6 b2d4 d7c5 b3c3 b7b6 d2f3 c5d3 c3d3 a7a4 d3c3 c8a6 g4f4 g6g5 f4g4 a8c8 c3c8 a6c8 e1c1 c8d7 h3h4 a4a3 c1c7 e7d8');
INSERT INTO evaluations VALUES(12126,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(12126,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(12126,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(12126,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(12126,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(12126,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(12126,36,'cp 42','f8e7 e1g1 g7g5 d1e2 d8c7 e2e3 g5g4 f3e1 a7a6 b2b4 c5c4 d3c2 f7f6 f2f4 f6f5 a2a4 b7b6 f1f2 c8b7 e3e2 h7h5 d2f1 a6a5 b4b5 c6d8 f1e3 d8f7 e2d2 d7f8 g2g3 e8c8 c1a3 e7a3 a1a3');
INSERT INTO evaluations VALUES(7764,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(7764,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(7764,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(7764,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7764,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(7764,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(7764,36,'cp 63','e1g1 g7g5 a2a4 h7h5 b2b4 c5d4 c3d4 c6b4 d3e2 d7b8 f3e1 b7b6 d2b3 c8a6 f2f4 a6e2 d1e2 g5f4 c1f4 b8c6 f4d2 a7a5 e1f3 d8d7 h2h4 a8c8 d2g5 c6d8 g5e7 c8c2 e2e3 d7e7 f3g5 e8d7 f1f6 h8g8 a1c1 c2c1 e3c1 g8g6');
INSERT INTO evaluations VALUES(7763,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(7763,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(7763,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(7763,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(7763,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(7763,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(7763,36,'cp 39','g7g5 d4c5 d7e5 d3c2 e5g6 b2b4 e8g8 d2b3 f7f5 b4b5 c6e5 f3e5 g6e5 f1e1 e5g6 d1h5 e6e5 c1e3 a7a6 b5b6 a6a5 a1d1 f8f7 a2a4 e7f6 e3c1 g5g4 g2g3 d8d7 c1g5 f6g7 h2h3 g4h3 f2f4 d7a4 f4e5 g7e5 d1d5 c8e6 e1e5 g6e5 d5e5 e6b3 c2f5 a4a1 g1h2 a1c3');
INSERT INTO evaluations VALUES(39163,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(39163,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(39163,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(39163,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(39163,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(39163,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(39163,36,'cp 52','a2a3 b4c3 d2c3 b8c6 f2f4 e6e5 g1f3 d7d6 f1d3 c8g4 e1g1 e5f4 c1f4 g8e7 d1d2 g4f3 g2f3 d8d7 a3a4 a7a6 f4g3 c6e5 d3e2 f7f5 g3e5 d6e5 d2d7 e8d7 e4f5 e7f5 e2d3 d7e6 a1e1 g7g5 f3f4 g5f4 f1f4 f5d6');
INSERT INTO evaluations VALUES(39166,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(39166,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(39166,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(39166,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(39166,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(39166,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(39166,36,'cp 63','b4c3 d2c3 b8c6 g1e2 d8h4 d1d3 d7d6 g2g3 h4e7 f1g2 b7b6 e2d4 e7d7 c1d2 c8b7 e1c1 e8c8 h1e1 c8b8 b2b3 g8e7 c1b2 e6e5 d4e2 h7h5 h2h4 h8f8 e1f1 f7f5 f2f3 d8e8 c3c4 d7e6 d2g5 e6g6 e2c3');
INSERT INTO evaluations VALUES(39181,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(39181,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(39181,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(39181,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(39181,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(39181,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(39181,36,'cp 56','d2c3 b8c6 g1e2 d8h4 d1d3 c6e5 d3g3 h4g3 h2g3 d7d6 f2f3 c8d7 b2b3 f7f6 e2f4 g8e7 a3a4 a7a5 g3g4 h7h6 f1e2 e8c8 f4h5 d8g8 c1e3 f6f5 g4f5 e6f5 e4f5 e7f5 e3f2 d7c6 e1c1 b7b6 c1b2 e5d7 c3c4 g7g5');
INSERT INTO evaluations VALUES(42602,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(42602,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(42602,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42602,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(42602,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(42602,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(42602,36,'cp -15','d7d5 c4d5 f6d5 g1f3 b8c6 e1g1 e8g8 d2d4 e5d4 e3d4 c5b6 b1c3 h7h6 c3d5 d8d5 f3e5 d5d4 e5c6 b7c6 d1d4 b6d4 f1d1 d4b6 c1e3 c8e6 g2c6 a8d8 b2b3 g7g5 a1c1 d8d1 c1d1 f8b8 d1d2 g8f8 e3d4 b8d8 d4c3 d8d2 c3d2 f8e7');
INSERT INTO evaluations VALUES(45700,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(45700,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(45700,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(45700,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(45700,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(45700,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(45700,36,'cp 84','c6c5 e1g1 a7a6 c4b3 b8c6 d1e2 c5d4 f1d1 f8e7 e3d4 e8g8 f3e5 f6d5 e2f3 d5c3 e5c6 c3e2 f3e2 b7c6 b3c2 e7f6 e2e4 g7g6 c1h6 f6g7 h6g7 g8g7 e4c6 a8b8 c6c3 d8f6 c2e4 c8b7 e4b7 b8b7 a1c1 b7b5 d1d2 a6a5 b2b3 f8d8');
INSERT INTO evaluations VALUES(53841,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(53841,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(53841,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(53841,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(53841,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(53841,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(53841,36,'cp 49','g8f6 c1f4 f8e7 d1d2 e8g8 e1c1 d6d5 f2f3 d5e4 d2e1 f6d5 d4b5 e7g5 c3d5 c6d5 e1d2 g5f4 d2f4 b8c6 b5c7 a8b8 c7d5 e4f3 g2f3 c8e6 f1c4 e6d5 d1d5 d8e7 h1g1 g7g6 h2h4 b8d8 h4h5 d8d5 c4d5 e7e5 f4e5 c6e5 h5g6 e5g6 c2c4 g8g7');
INSERT INTO evaluations VALUES(45885,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(45885,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(45885,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(45885,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(45885,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(45885,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(45885,36,'cp 81','c1f4 f8e7 d1d2 e8g8 e1c1 d6d5 f2f3 d5e4 f1c4 e4f3 g2f3 c6c5 d4b5 d8d2 d1d2 f6e8 c3d5 e7d8 h1e1 b8c6 b5d6 e8d6 f4d6 c8f5 d6f8 g8f8 c2c3 d8h4 e1e2 g7g6 c4d3 f5h3 e2e4 h4d8 f3f4 h3e6 d3c4 c6a5 c4b5 a7a6 b5f1 a8c8 e4e1 d8h4 e1d1 c8e8 d5c7 e6a2 c7e8 a5b3 c1c2 b3d2 d1d2');
INSERT INTO evaluations VALUES(54783,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(54783,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(54783,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(54783,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54783,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(54783,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(54783,36,'cp -13','e7e5 d4e5 d8d1 e1d1 g8e7 f3e4 f5e4 g1e2 c8f5 e2d4 b8c6 d4f5 e7f5 c3e4 e8c8 c1d2 c6e5 d1c1 d8d4 e4f2 f8e7 c2c3 d4d5 c1c2 e7c5 f2e4 c5e3 a1e1 e3d2 e4d2 h8d8 d2c4 e5g4 f1e2 g4e3 c4e3 f5e3 c2c1 e3g2 e2g4 c8b8 e1e7 d5d2 e7g7');
INSERT INTO evaluations VALUES(9311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9311,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9311,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9311,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(9311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9311,36,'cp 0','b1c3 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 c6e7 d5e3 e8g8 e1g1 c5e3 f2e3 c7c6 d3d4 d6d5 c4d3 f7f5 d1b3 f5e4 d3e4 d8b6 e4c2 e5e4 f3e5 f8f1 a1f1 e7f5 e5g6 g8h7 g6f8 h7g8');
INSERT INTO evaluations VALUES(11949,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(11949,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(11949,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(11949,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(11949,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(11949,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(11949,36,'cp 59','e1g1 f8e7 f1e1 e4d6 b1c3 e8g8 d2d4 d6f5 d1d3 f5h4 c1d2 h4g6 a1d1 c8b7 d3f3 d7d5 c3a4 g6e5 d4e5 d8c8 f3g3 c8f5 d2h6 f5g6 g3g6 h7g6 h6e3 b7c8 e3c5 e7c5 a4c5 a8b8 b2b3 b8b5 c5d3 b5a5 a2a4 c6c5 d3f4 c7c6 e5e6 c5c4 f2f3 c8e6 f4e6 f7e6');
INSERT INTO evaluations VALUES(11948,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(11948,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(11948,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11948,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(11948,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(11948,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11948,36,'cp 43','f8e7 f1e1 e4d6 b1c3 e8g8 d2d4 d6f5 d1d3 f5h4 c1d2 h4g6 h2h3 c8b7 c3e4 d7d6 e5g6 h7g6 d2a5 d8d7 e1e2 a8e8 a1e1 e7d8 a5d2 e8e6 c2c4 d8h4 e4g5 h4g5 d2g5 f8e8 e2e6 e8e6 e1e6 d7e6');
INSERT INTO evaluations VALUES(19326,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19326,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19326,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19326,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19326,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19326,36,'cp 0','e4e5 g8h6 f1c4 f8g7 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 h2h3 c6e5 f3e5 d6e5 d1d8 f8d8 f4g5 d8d4 c4b3 c8d7 g5e3 d4h4 e3g5');
INSERT INTO evaluations VALUES(30752,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(30752,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(30752,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(30752,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(30752,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(30752,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(30752,36,'cp 106','g1f3 f8g7 b1c3 d7d6 h2h4 h7h6 e4e5 d8d7 d3b5 b7c6 d1e2 c6b5 c3b5 a7a6 b5c3 d6d5 e2d3 e7e6 c3e2 c7c5 c2c3 g8e7 e1g1 c5d4 c3d4 b8c6 c1e3 e8g8 g2g4 a6a5 a1d1 a5a4 e2f4');
INSERT INTO evaluations VALUES(30751,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(30751,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(30751,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(30751,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(30751,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(30751,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(30751,36,'cp 106','d7d6 h2h4 h7h6 b1c3 f8g7 c1f4 e7e6 e4e5 b8c6 d1e2 d8e7 e1c1 d6d5 c1b1 e8c8 f4c1 c8b8 g2g4 e7d7 e2e1 c6b4 d3b5 d7c8 a2a3 b4c6 c3e2 g8e7 b5d3 b7a6 d3a6 c8a6 h4h5 g6g5 f3h2 f7f6 f2f4 g5f4');
INSERT INTO evaluations VALUES(40402,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40402,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40402,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40402,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40402,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40402,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40402,36,'cp 0','d3e2 b8c6 c1e3 e8g8 e1g1 c6d4 e3d4 c8e6 d1d2 a8c8 a1d1 a7a6 f1e1 f8e8 a2a3 d8a5 d2e3 f6d7 f2f4 g7d4 d1d4 a5b6 c3a4 b6a5');
INSERT INTO evaluations VALUES(54337,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(54337,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(54337,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(54337,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(54337,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(54337,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(54337,36,'cp 14','g8f6 b1c3 f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 d8d6 b5c6 b7c6 g5f6 d6f6 h2h3 c8f5 f1e1 f6d6 f3e5 c6c5 d1d2 c5d4 d2d4 f8d8 e1e3 f5e4 a1e1 a8c8 f2f3 f7f6 f3e4 f6e5 d4d5 d6d5 e4d5 d8d5 e3e5 d5e5 e1e5 c8c3');
INSERT INTO evaluations VALUES(46309,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46309,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(46309,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46309,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46309,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(46309,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(46309,36,'cp 24','b1c3 f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 d8d6 b5c6 d6c6 f3e5 c6e6 g5f6 e6f6 f1e1 f6d6 d1f3 c8e6 e5d3 f8e8 a2a4 f7f6 d3c5 d6c6 c5e6 e8e6 e1e6 c6e6 a1b1 a8e8');
INSERT INTO evaluations VALUES(38703,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(38703,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(38703,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(38703,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(38703,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38703,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38703,36,'cp 0','g8f6 c2c3 c5b6 a2a4 a7a5 f1e1 e8g8 d2d3 b8c6 b1a3 h7h6 c4b3 c8e6 a3c4 b6a7 c1e3 a7e3 c4e3 d8d7 d1c2 e6b3 c2b3 f8e8 e3c4 b7b6 a1d1 d7e6 b3b5 c6a7 b5b3');
INSERT INTO evaluations VALUES(41871,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41871,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(41871,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(41871,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(41871,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(41871,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(41871,36,'cp 13','c2c3 c5b6 a2a4 a7a5 f1e1 e8g8 d2d3 h7h6 h2h3 c7c6 b1d2 b6c7 d3d4 b8d7 d4e5 d7e5 f3e5 d6e5 b2b3 c7d6 d1c2 d8e7 d2f3 d6c5 c1e3 c5e3 e1e3 e7c7 a1d1 f8e8 f3h4 c8e6 c4e6 e8e6 h4f5 a8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(45900,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(45900,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(45900,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(45900,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(45900,34,'cp 245',NULL);
INSERT INTO evaluations VALUES(45900,35,'cp 227',NULL);
INSERT INTO evaluations VALUES(45900,36,'cp 250','f3g5 d6d5 e4d5 h7h6 g5e4 f6e4 c3e4 d8h4 d1e2 f7f5 e4c3 b8d7 c3b5 e8d8 d5d6 d7b6 d2d3 c8d7 b5c7 f8d6 c7a8 b6a8 c1d2 d7c6 e1c1 h8e8 f2f3 b7b5 c4b3 a8c7 h1e1 h4h2 e2f2 f5f4 e1h1 h2g3 f2g3 f4g3 d2e1 c7e6 e1g3 a7a5 b3e6 e8e6 g3f2 a5a4 c1b1 d8c8');
INSERT INTO evaluations VALUES(45896,30,'cp 239',NULL);
INSERT INTO evaluations VALUES(45896,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(45896,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(45896,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(45896,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(45896,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(45896,36,'cp 253','d6d5 e4d5 h7h6 g5e4 f6e4 c3e4 d8h4 d1e2 f7f5 e4c3 b8d7 c3b5 e8d8 d5d6 d7b6 d2d3 c8d7 b5c7 b6c4 d3c4 a8c8 e2e5 f8d6 e5d6 c8c7 e1g1 h4e7 d6g3 g7g5 c1d2 d8c8 a1e1 e7f7 g3a3 a7a6 e1e5 h8e8 e5c5 c7c5 a3c5 d7c6 f2f3 f7g6 f1e1 e8e1 d2e1 f5f4 e1c3 g6c2 h2h3 c2e2 c5f8 c8c7 f8h6');
INSERT INTO evaluations VALUES(47886,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47886,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47886,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47886,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47886,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47886,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(47886,36,'cp 0','g7g6 c2c4 f8g7 b1c3 b8c6 e1g1 c8g4 a1b1 a7a5 f3e1 g4e2 c3e2 e8g8 a2a3 d8d7 h2h3 a5a4 e2c3 c6d4 e1f3 d4b3 c1e3 f8e8 f1e1 b7b6 f3d2 b3c5 d2f3');
INSERT INTO evaluations VALUES(48473,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(48473,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(48473,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(48473,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(48473,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(48473,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(48473,36,'cp 86','e4e5 f6d5 b1c3 c7c5 g1f3 b8c6 e1g1 c5d4 f3d4 d5c3 b2c3 c8d7 d1e2 f8e7 d4b5 e8g8 c4d3 d8a5 c1f4 a7a6 b5d6 e7d6 e5d6 a5c3 a1b1 c3d4 f4g3 e6e5 b1b7 d4d6 f1d1 c6d4 e2e5 d6e5 g3e5 d7g4 d1d2 d4e6 b7b4 g4h5 e5d6 f8d8 d3f1 d8c8 f2f3 a6a5 b4h4 h5g6 h4c4 c8c4 f1c4');
INSERT INTO evaluations VALUES(48437,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(48437,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(48437,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(48437,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(48437,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(48437,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(48437,36,'cp 69','f6d5 b1c3 c7c5 g1f3 b8c6 e1g1 c5d4 f3d4 d5c3 b2c3 c8d7 d1e2 f8e7 f1d1 c6d4 c3d4 d7a4 d1d2 a4c6 d2d3 c6d5 c4d5 d8d5 e2g4 g7g6 c1h6 e8c8 h6e3 c8b8 g4f4 d8d7 f4f7 e7d8 f7f3 d5f3 g2f3 d7d5 a1d1 d8b6 g1g2 h8c8 f3f4 c8c2');
INSERT INTO evaluations VALUES(48103,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(48103,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(48103,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(48103,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(48103,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(48103,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(48103,36,'cp 71','g1e2 c7c5 e1g1 b8c6 b1c3 f8e7 d4c5 d5c3 e2c3 d8d1 f1d1 e7c5 c4b5 c8d7 c1f4 a7a6 b5a4 b7b5 a4c2 c5d4 c2e4 d4c3 b2c3 a8c8 d1d6 c6e7 d6a6 e8g8 a6d6 d7c6 e4d3 h7h6 h2h4 f8d8 a1d1 e7d5 f4d2 d5b6 d2e3 b6c4');
INSERT INTO evaluations VALUES(13265,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(13265,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(13265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13265,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13265,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13265,36,'cp 0','e7e6 d2d3 f8e7 c4b3 a7a6 a2a4 d6d5 e1g1 e8g8 e4e5 f6d7 d1e1 f7f5 g1h1 c6a5 b3a2 a5c6');
INSERT INTO evaluations VALUES(27146,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(27146,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(27146,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(27146,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(27146,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(27146,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(27146,36,'cp 70','d8d7 d1d4 g8e7 b1c3 e7c6 d4c4 f8e7 c3d5 e8g8 c4b5 a8b8 b2b3 f8e8 c1b2 e7d8 b5d3 c6e5 d3d2 c7c6 d5e3 f7f6 a1d1 e5f7 f2f3 d8b6 b2d4 b6d4 d2d4 a7a6 a2a4 b7b5 a4b5 a6b5 d4d3 b8d8 c2c4 b5b4 e3f5 d7c7');
INSERT INTO evaluations VALUES(27669,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(27669,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(27669,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(27669,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(27669,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(27669,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(27669,36,'cp 78','d1d4 g8e7 b1c3 e7c6 d4c4 f8e7 c1f4 e8g8 a1d1 f8e8 f1e1 e7f8 c3d5 a7a6 c4b3 b7b5 b3g3 g8h8 f4g5 f7f6 g5d2 c6e7 d5e7 e8e7 g3d3 d7c6 f2f3 a6a5 g1f1 e7e5 d2e3 c6a6 h2h3 a5a4 b2b4 a4b3 a2b3 e5e8 e3f2');
INSERT INTO evaluations VALUES(41810,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(41810,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41810,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41810,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(41810,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(41810,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(41810,36,'cp 20','b1d2 c7c5 d4c5 d6c5 e1g1 e8g8 c2c4 b7b6 d1e2 b8c6 f1d1 d8e7 a2a3 a7a5 c4d5 e6d5 b2b3 d5d4 d2f1 d4e3 c1e3 c5a3 d3b5 c6a7 b5c4 a3b4 e3b6 e7e2 c4e2 a7c6 d1c1 c6e7 f1e3 c8a6 e2a6 a8a6 b6c5 f8c8 c5b4 c8c1 a1c1 a5b4');
INSERT INTO evaluations VALUES(41811,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41811,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(41811,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41811,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(41811,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41811,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41811,36,'cp 1','c7c5 d4c5 d6c5 e1g1 e8g8 e3e4 b8c6 h2h3 c5d6 f1e1 d6c7 c2c3 h7h6 d3c2 f8e8 e4d5 f6d5 d2c4 d5f6 c2a4 a7a6 f3d4 c8d7 d4c6 d7c6 d1d8 e8d8 a4c6 b7c6 g2g3 f6d5 g1g2 f7f6 a2a4 g8f7 e1d1 a8b8');
INSERT INTO evaluations VALUES(55740,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(55740,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(55740,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(55740,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(55740,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(55740,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(55740,36,'cp 48','d6d5 b1c3 d5e4 c3e4 b8d7 g1f3 d7f6 e4f6 g8f6 f3e5 f8e7 c1e3 f6d5 e3d2 d5f6 c2c3 d8d5 h1g1 f6d7 g2g3 d5e4 d1e2 e4e2 f1e2 d7e5 f4e5 c7c5 d2e3 c5d4 e3d4 c8d7 a2a4 e8g8 a4a5 d7c6 e1f2 a7a6 g1d1 h7h5 d4b6 h5h4 b2b4 h4g3 h2g3');
INSERT INTO evaluations VALUES(56233,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56233,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56233,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56233,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56233,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56233,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56233,36,'cp 0','c2c4 c7c6 a2a3 f7f5 b1c3 f8e7 c4d5 c6d5 h2h4 c8e6 g1h3 b8c6 h3g5 e6c8 g5h3');
INSERT INTO evaluations VALUES(56187,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56187,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56187,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56187,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56187,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56187,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56187,36,'cp 0','c7c6 a2a3 f8e7 b1c3 f7f5 c4d5 c6d5 h2h4 c8e6 g1h3 b8c6 h3g5 e6c8 g5h3');
INSERT INTO evaluations VALUES(53712,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53712,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53712,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53712,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53712,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53712,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53712,36,'cp 0','a2a3 f7f5 b1c3 f8e7 c4d5 c6d5 h2h4 c8e6 g1h3 b8c6 h3g5 e6c8 g5h3');
INSERT INTO evaluations VALUES(56804,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(56804,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(56804,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(56804,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(56804,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(56804,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(56804,36,'cp 14','c7c5 f1d3 g8h6 c2c3 b8c6 g1f3 f8e7 e1g1 e8g8 g1h1 b7b6 e5f6 e7f6 c1e3 c5d4 c3d4 h6f5 d3f5 e6f5 b1c3 c6a5 f1e1 c8e6 e3f2 f8e8 f3e5 a8c8 h1g1 a5c4 d1e2 c4d6 c3b5 f6e5 d4e5 d6b5 e2b5 c8c4');
INSERT INTO evaluations VALUES(54662,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(54662,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(54662,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(54662,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(54662,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(54662,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54662,36,'cp 36','c2c3 b8c6 g1f3 f6e5 f4e5 g8e7 f1d3 e7f5 e1g1 c5d4 c3d4 f8e7 d3f5 e6f5 b1c3 c8e6 d1b3 d8d7 c1g5 e7b4 g5d2 e8g8 f3g5 h7h6 g5e6 d7e6 f1f2 b4c3 d2c3 f8f7 a1f1 a8f8 h2h3 f5f4');
INSERT INTO evaluations VALUES(54661,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(54661,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(54661,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(54661,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54661,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54661,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54661,36,'cp 11','f6e5 d4e5 c8d7 f1d3 g8h6 c1e3 d8b6 d1e2 b8c6 g1h3 h6g4 e3g1 g4h6 b2b4 b6c7 g1c5 f8c5 b4c5 c6e7 b1d2 c7c5 h3g5 c5c3 e1g1 h6f5 d2b3 c3b4 g2g4 f5d4 b3d4 b4d4 g1h1 h7h6 a1d1 d4c5 d1c1 c5b4 c1b1 b4a3 g5f3 e8g8 b1b7');
INSERT INTO evaluations VALUES(47649,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(47649,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(47649,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(47649,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(47649,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(47649,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(47649,36,'cp 94','f3d4 f8e7 c1f4 b8c6 d4c6 b7c6 e4e5 c8g4 d1d4 d6e5 f4e5 d8d4 e5d4 e8c8 d4a7 c8b7 a7e3 g4f5 c4f7 f5c2 f7b3 c2b3 a2b3 f6d5 e1e2 d5e3 f2e3 h8e8 a1a4 e7c5 a4e4 c5d6 g2g3 e8e5 h1d1 d8b8 e4e5 d6e5 d1d7 e5f6 c3e4 f6b2 e4c5 b7c8');
INSERT INTO evaluations VALUES(50243,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(50243,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(50243,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(50243,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(50243,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(50243,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(50243,36,'cp 157','e4e5 f6h5 d2d4 c7c6 b1c3 d7d6 f1c4 d6d5 c4d3 h5g7 c1h6 c8f5 h6g7 f8g7 d3f5 g6f5 e1g1 e7e6 c3e2 c6c5 c2c3 b8c6 e2f4 c5d4 c3d4 g7f8 f4h5 d8b6 a1b1 e8c8 f3g5 b6d4 d1d4 c6d4 g5f7');
INSERT INTO evaluations VALUES(50150,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(50150,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(50150,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(50150,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(50150,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(50150,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(50150,36,'cp 138','f6h5 d2d4 c7c6 c2c3 a7a5 a2a4 h5g7 f1d3 d7d5 e1g1 c8f5 b1a3 e7e6 c1h6 b8d7 a3c2 f5d3 d1d3 g7f5 h6f8 h8f8 d3d2 f8h8 h2h3 h7h6 g2g3 d8c7 h3h4 g6g5 c2e3 f5e3 d2e3 g5h4 f3h4 e8c8 b2b4 d8g8');
INSERT INTO evaluations VALUES(51514,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(51514,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(51514,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(51514,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(51514,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(51514,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(51514,36,'cp 14','g8f6 c2c4 e7e6 g2g3 c8b7 g1f3 f8e7 f1g2 e8g8 e1g1 d7d5 c4d5 e6d5 e2e3 f8e8 b1c3 b8d7 a1c1 c7c6 d2d3 e7f8 f1e1 b6b5 e3e4 b5b4 c3a4 d5e4 d3e4 e8e4 f3e5 e4e1 d1e1 d7e5 e1e5 h7h6 a4c5 f8c5 e5c5 d8d2 c5d4 d2d4 b2d4 a8d8 d4e3 f6d5');
INSERT INTO evaluations VALUES(51955,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(51955,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(51955,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(51955,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(51955,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51955,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(51955,36,'cp 0','d8b6 g1f3 g8f6 d1b3 f8e7 e2e3 e8g8 b1d2 d7d5 f1e2 b8c6 b3b6 a7b6 f4c7 e7d8 c7d6 d8e7');
INSERT INTO evaluations VALUES(52451,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(52451,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(52451,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(52451,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52451,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(52451,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(52451,36,'cp 0','b8c6 g1f3 g8f6 b1c3 e7e5 e1g1 d7d6 e2c4 f8e7 f1e1 e8g8 d2d3 h7h6 a2a3 a8b8 c1d2 a7a6 a3a4 g8h8 h2h3 c8d7 c3d5 f6d5 c4d5 f7f5 c2c3 a6a5 d3d4 f5e4 d5e4 c5d4 c3d4 c6d4 f3d4 e5d4 e4d3');
INSERT INTO evaluations VALUES(19471,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19471,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19471,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19471,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19471,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(19471,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19471,36,'cp 9','g1f3 g8f6 b1c3 e7e5 e1g1 f8e7 f1e1 e8g8 e2c4 d7d6 d2d3 h7h6 a2a3 a8b8 h2h3 a7a6 a3a4 c8d7 c1d2 c6b4 c3e2 b7b5 a4b5 a6b5 c4b3 b4c6 c2c3 b5b4 e2g3 d7e6 b3c4 d8d7 c4e6 f7e6 d1e2');
INSERT INTO evaluations VALUES(54037,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(54037,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(54037,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(54037,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(54037,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(54037,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(54037,36,'cp 157','c8d7 e4d5 d7b5 c3b5 d8a5 b5c3 e6d5 g1f3 b8c6 e1g1 e8c8 d1d3 g8f6 c1g5 c8b8 f3e5 c5d4 e5c6 b7c6 g5f6 g7f6 d3d4 f8d6 d4f6 h7h5 a1e1 b8a8 h2h3 a5c5 g1h1 h5h4 e1e2 d8b8 f1e1 d6c7 e2e7 c7d8 f6h8 d8e7 h8g7');
INSERT INTO evaluations VALUES(30882,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(30882,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(30882,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(30882,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(30882,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(30882,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(30882,36,'cp 158','e4d5 e6d5 d1e2 f8e7 c3d5 d7c6 d5e7 g8e7 b5c6 b8c6 d4c5 d8a5 c2c3 a5c5 g1f3 e8g8 e1g1 c5f5 e2c4 a8d8 c1g5 h7h6 g5h4 d8d7 f1d1 f8d8 d1d7 d8d7 h4g3 f5c2 c4b3 c2b3 a2b3 f7f6 g1f1 a7a6 f1e2 g7g5 h2h3 g8f7 b3b4 e7g6');
INSERT INTO evaluations VALUES(30880,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(30880,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(30880,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(30880,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(30880,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(30880,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(30880,36,'cp 155','e6d5 d1e2 f8e7 c3d5 d7c6 b5c6 b8c6 d5e7 g8e7 d4c5 d8a5 c2c3 e8g8 g1f3 a5c5 e1g1 c5f5 f1e1 f8e8 c1d2 a8d8 h2h3 h7h6 e2e4 f5e4 e1e4 f7f6 a1e1 g8f7 g1f1 d8d7 d2e3 e8d8 e3c5 b7b6 c5e7 c6e7 e4d4 e7c6 d4d7 d8d7 f1e2 b6b5 b2b3 b5b4');
INSERT INTO evaluations VALUES(30903,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(30903,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(30903,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(30903,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(30903,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(30903,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(30903,36,'cp 172','d1e2 f8e7 c3d5 d7c6 b5c6 b8c6 d5e7 g8e7 d4c5 d8a5 c2c3 a5c5 g1f3 e8g8 e1g1 c5f5 f1e1 h7h6 c1e3 e7d5 f3h4 f5e4 e2f3 f8e8 f3e4 e8e4 h4f3 a8e8 e3d2 e4e1 a1e1 e8d8 g1f1 f7f6 e1e4 g8f7 f3d4 c6e5 b2b3 h6h5');
INSERT INTO evaluations VALUES(57736,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(57736,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(57736,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(57736,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(57736,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(57736,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(57736,36,'cp 129','c7c5 c2c3 g8f6 e4e5 f6d5 d2d4 c5d4 c3d4 d7d6 f1c4 d5b6 c4b3 d6d5 c1g5 b8c6 f3h4 c8g4 f2f3 g4e6 b1c3 g7g6 d1d2 d8d7 e1g1 e8c8 f3f4 b6c4 d2f2 c8b8 a1c1 d8c8 b3a4 a7a6 h2h3 f8h6 g5h6 h8h6 b2b3 c4a3 a4c6 c8c6');
INSERT INTO evaluations VALUES(20254,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(20254,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(20254,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(20254,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(20254,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(20254,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(20254,36,'cp 32','c4d5 e6d5 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 e3d4 f3d1 d4b6 d1g4 b6e3 g4e6 f1d1 f8c8 c4e6 f7e6 g1f1 f6d5 e3d4 a7a6 f1e2 g8f7 h2h4 d5b4 a2a3 c8c1 d1c1 b4c6 d4c3 a8d8 e2e3');
INSERT INTO evaluations VALUES(20227,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(20227,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(20227,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(20227,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(20227,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(20227,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20227,36,'cp 27','e6d5 d4c5 d5d4 c3a4 f8c5 a4c5 d8a5 c1d2 a5c5 a1c1 c5b6 e2e3 g8f6 f1c4 e8g8 e1g1 c8g4 e3d4 c6d4 d2e3 g4f3 e3d4 f3d1 d4b6 d1g4 f2f3 g4e6 c4e6 f7e6 b6d4 f6d5 f1e1 g8f7 e1e4 a7a6 c1e1 f8e8 h2h4 a8c8 h4h5 c8d8 d4e5 d5f6 e4f4 d8d2 e5c3 d2d5 g2g4 e8d8');
INSERT INTO evaluations VALUES(20792,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(20792,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(20792,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(20792,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(20792,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(20792,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(20792,36,'cp 65','f1b5 c8d7 b5c6 d7c6 e1g1 g8e7 d4c5 e7g6 e5f6 d8f6 c1g5 f6f5 b2b4 d5d4 f1e1 f8c5 b4c5 e8g8 c3d4 c6f3 d1f3 f5g5 f3g3 g5f6 b1c3 f6d4 c3b5 d4c4 b5c7 a8c8 c7e6 f8e8 g3b3 c4b3 a2b3 a7a5 h2h4 g6f8 e6f8 e8e1 a1e1 g8f8 e1c1 c8e8 c1c3 f8f7 g1h2');
INSERT INTO evaluations VALUES(20789,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(20789,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(20789,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(20789,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(20789,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(20789,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(20789,36,'cp 72','c8d7 e1g1 c6e5 f3e5 f6e5 d1h5 g7g6 h5e5 d7b5 e5h8 b5f1 h8h7 f1d3 h7g8 d8f6 g8h7 e8c8 h7h3 f6f5 h3f5 e6f5 c1e3 c5c4 g2g3 d8e8 b1d2 f8d6 d2f3 d3e4 f3g5 e4d3 g1g2 c8d7 h2h4 a7a5 g2f3 d6c7 a1g1 d7c6 g1e1 c7d6 e3d2 e8e1 d2e1 c6d7 e1d2 a5a4 g5f7 d3e4 f3e3');
INSERT INTO evaluations VALUES(25995,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(25995,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(25995,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(25995,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(25995,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(25995,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(25995,36,'cp 72','e1g1 c6e5 f3e5 f6e5 d1h5 g7g6 h5e5 d7b5 e5h8 b5f1 h8h7 f1d3 h7g8 d8f6 g8h7 e8c8 h7h3 f6f5 h3f5 e6f5 c1e3 d8e8 b1d2 c5c4 d2f3 f8d6 f3e5 d6e5 d4e5 e8e5 e3a7 e5e2 b2b4 c4b3 a2b3 e2b2 b3b4 b2b1 a1b1 d3b1 f2f3 c8d7 g1f2 f5f4 h2h4 d7e6 f2e2 e6f6 a7d4 f6f5 b4b5 b1c2');
INSERT INTO evaluations VALUES(38772,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(38772,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(38772,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38772,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(38772,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38772,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38772,36,'cp 0','d1d3 g8f6 a2a3 b4d6 f4d6 c7d6 e2e4 f6e4 c3e4 d5e4 d3e4 b8d7 f1d3 d7f6 e4e3 e8g8 g1f3 b7b5 e1g1 c8d7 f1e1 d7c6 f3d2 g7g6 e3h6 f6g4 h6f4 g4f6');
INSERT INTO evaluations VALUES(38774,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38774,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38774,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38774,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38774,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38774,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38774,36,'cp 0','g8f6 a2a3 b4d6 f4d6 c7d6 g2g4 f6g4 d3g3 g4f6 g3g7 h8g8 g7h6 b8c6 e2e3 g8g6 h6h4 g6g4 h4h6');
INSERT INTO evaluations VALUES(47160,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(47160,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(47160,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(47160,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(47160,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(47160,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(47160,36,'cp -101','c8f5 g1f3 e7e6 e2e3 c7c5 f3e5 b8c6 f1b5 a8c8 e1g1 f8d6 c3a4 c5d4 e3d4 e8g8 b5c6 b7c6 a4c5 f6e4 c5b7 d8c7 b7d6 e4d6 c1e3 a7a5 d1e2 a5a4 b2b4 a4a3 f1e1 f7f6 e5d3 d6c4 g2g4 f5d3 c2d3');
INSERT INTO evaluations VALUES(29549,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(29549,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(29549,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(29549,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(29549,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(29549,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(29549,36,'cp -98','e2e3 c7c5 g1f3 b8c6 f1d3 e7e6 e1g1 a7a6 c1d2 c5d4 e3d4 f8e7 h2h3 a8c8 d2e3 e8g8 g2g4 f5g6 f3e5 g6d3 c2d3 f6d7 e5f3 e7d6 a1c1 c6e7 d1b3 b7b5 c3e2 d8b6 c1c2 b5b4 f1c1 b6b7 c2c8');
INSERT INTO evaluations VALUES(29547,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(29547,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(29547,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(29547,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(29547,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(29547,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(29547,36,'cp -90','c7c5 f1d3 e7e6 g1f3 b8c6 e1g1 a7a6 c1d2 c5d4 e3d4 f8e7 d2e3 e8g8 f3e5 a8c8 a1c1 e7d6 c3a4 c6e7 a4c5 c8c7 c2c3 f5d3 d1d3 d8c8 a2a4 e7f5 e3f2 h7h5 d3e2 b7b6 c5a6 c7a7 a6b4 a7a4 b4d3 c8a6');
INSERT INTO evaluations VALUES(4433,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4433,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(4433,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4433,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(4433,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(4433,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4433,36,'cp 28','c3d5 f6d5 e4d5 e5e4 d5c6 e4f3 d1f3 d7c6 d2d3 a7a5 a2a4 h7h6 h2h3 b4d6 f1e1 d8d7 c1d2 d7f5 f3f5 c8f5 g2g4 f5d7 g1g2 f8e8 e1e8 d7e8 f2f4 g8f8 g2f3 e8d7 a1e1 b7b6 d2c3 a8e8 e1e4 f7f6 h3h4 c6c5 c4b5 c7c6');
INSERT INTO evaluations VALUES(30035,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(30035,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(30035,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30035,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30035,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30035,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(30035,36,'cp 20','f8b4 c1d2 b4d2 d1d2 d8e7 d2e3 b8c6 f1e2 e8c8 d4d5 f6g4 e3d2 e7c5 g1h3 g4f6 h3f4 e6g4 e1g1 g4e2 d2e2 h8e8 e2c2 c6d4 c2d3 b7b5 c4b5 f6d5 d3h3 c8b8 f4d5 d4e2 g1h1 c5d5 b1c3 d5d3 h3g4 d3g6 g4c4 d8d4 c4c5 g6d6 c5h5 d6g6 h5g6 h7g6 f1e1 d4d2 b2b3 e8e6 e1e2');
INSERT INTO evaluations VALUES(28482,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(28482,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(28482,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(28482,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(28482,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(28482,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(28482,36,'cp 28','c1d2 b4d2 d1d2 d8e7 d2e3 b8c6 f1e2 e8c8 d4d5 f6g4 e3d2 e7c5 g1h3 g4f6 h3f4 e6g4 b2b3 g4e2 f4e2 f6e4 d2f4 c5a5 e1f1 h8e8 g2g3 c8b8 f1g2 c6b4 b1c3 e4c3 e2c3 b4c2 b3b4 a5b4 c3b5 b4c5 a1c1 g7g5 f4g5 c5c4 h1d1 c4b5 c1c2 b5a4 c2d2 a4e4 f2f3 e4e5 g5e5 e8e5 g3g4 f7f5 h2h4 f5g4 f3g4 d8e8 g2g3 a7a6 h4h5 e5g5');
INSERT INTO evaluations VALUES(28481,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(28481,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(28481,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(28481,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(28481,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(28481,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(28481,36,'cp 15','b4d2 d1d2 d8e7 d2e3 b8c6 f1e2 e8c8 d4d5 f6g4 e3c3 e7c5 g1h3 c6d4 b2b4 d4e2 b4c5 e2c3 b1c3 e6f5 e1g1 g4e5 h3g5 h8f8 g5f3 e5c4 f3d4 f5g6 f1e1 f8e8 f2f4 g6d3 c5c6 b7b6 d4f3 c4d6 f3e5 d3a6 a2a4 f7f6 e5d7 e8e1 a1e1 d8e8 e1e8 d6e8 c3b5 e8d6 b5d6 c7d6 g1f2 c8c7 f4f5');
INSERT INTO evaluations VALUES(56274,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(56274,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(56274,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(56274,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(56274,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(56274,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(56274,36,'cp -54','f8e7 d2d4 g8f6 e2d1 e8g8 f1e2 f8e8 g1f3 b8d7 e1g1 d7f8 h2h3 c8f5 e2d3 f5g6 d3g6 h7g6 f1e1 c7c6 c1f4 b7b5 f4e5 a7a5 c3e2 f8d7 c2c3 f6e4 e2g3 f7f6 e5f4 g6g5 g3e4 d5e4 e1e4 g5f4 d1b3 g8h8');
INSERT INTO evaluations VALUES(38320,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(38320,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(38320,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(38320,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(38320,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(38320,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(38320,36,'cp -44','d2d4 g8f6 g1f3 e8g8 e2d1 f8e8 f1e2 c7c6 e1g1 c8f5 f3h4 f5e6 e2d3 e6g4 f2f3 g4d7 h4f5 d7f5 d3f5 g7g6 f5h3 b7b5 c3e2 a7a5 e2g3 a5a4 c1g5 a4a3 b2b3 d8a5 g1h1 b8d7 d1d3 e7d6 d3d2 a5d2');
INSERT INTO evaluations VALUES(38311,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(38311,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(38311,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(38311,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(38311,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(38311,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(38311,36,'cp -60','g8f6 e2d1 c7c5 g1f3 e8g8 f1e2 b8c6 e1g1 h7h6 h2h3 c8f5 c1e3 c5c4 f3e5 e7b4 e2f3 f8e8 e3f4 e8e6 a2a3 b4a5 f1e1 a8c8 f4h2 f5h7 h2f4 a5b6 e5c6 b7c6 e1e6 f7e6 f4g3 h7g6');
INSERT INTO evaluations VALUES(41093,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(41093,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(41093,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(41093,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(41093,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(41093,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(41093,36,'cp -49','e2d1 e8g8 f1e2 e7b4 g1f3 f6e4 c1d2 b8c6 h2h3 d8f6 e1g1 e4d2 d1d2 c8e6 a1d1 c6e7 f3e5 a8e8 e2g4 c7c6 g4e6 f6e6 f1e1 e6c8 a2a3 b4a5 e1e2 f7f6 e5f3 e7g6 d1e1 c8d7 e2e8 f8e8 e1e8 d7e8 d2e2 g8f7 e2e8 f7e8 c3e2');
INSERT INTO evaluations VALUES(30822,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(30822,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30822,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(30822,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(30822,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(30822,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(30822,36,'cp 36','g2g3 g7g6 f1g2 f8g7 d3d4 g8f6 d4d5 c7c6 c2c4 e8g8 g1e2 e7e6 e1g1 c6d5 c4d5 e6d5 e4d5 b8a6 b1c3 f8e8 c1f4 a6c5 b2b4 c5e4 a1c1 a8c8 d1d3 e4c3 e2c3 f6g4 c3e4 g4e5 d3d1 c8c1 f4c1 b7a6');
INSERT INTO evaluations VALUES(39341,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(39341,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(39341,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(39341,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(39341,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(39341,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(39341,36,'cp 53','c2c3 b4a5 b1a3 d7d6 a3c4 c7c5 c4a5 d8a5 f1e2 g8f6 f2f4 b8c6 g1f3 e8g8 f4e5 d6e5 c1g5 c8e6 e1g1 a5b6 d1d2 c5c4 g1h1 c4d3 e2d3 a8d8 d2e2 d8d7 d3b5 e6g4 g5f6 g7f6 h2h3 g4h5 a2a4 a7a6 b5c4 c6e7 a1d1 d7d1 f1d1 f8d8 d1d8 b6d8 c4b3 g8g7');
INSERT INTO evaluations VALUES(39340,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(39340,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(39340,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(39340,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(39340,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(39340,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(39340,36,'cp 50','b4a5 b1a3 d7d6 a3c4 c7c5 c4a5 d8a5 f2f4 g8f6 f4e5 d6e5 g1f3 b8d7 f1e2 e8g8 c1e3 c5c4 d3c4 a5c7 f3d2 d7c5 e1g1 f8d8 f1f6 g7f6 d1f1 c7c6 f1f3 d8d2 e3d2 c5e4 d2e3 c8e6 e2d3 f6f5 d3e4 c6e4');
INSERT INTO evaluations VALUES(39442,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(39442,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(39442,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(39442,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(39442,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(39442,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(39442,36,'cp 60','b1a3 d7d6 a3c4 c7c5 c4a5 d8a5 f2f4 g8f6 f4e5 d6e5 g1f3 b8c6 f1e2 e8g8 c1g5 c8e6 d1c1 a5b6 b2b3 f6d7 g5e3 a7a5 f3g5 a8e8 e1g1 h7h6 g5e6 e8e6 c1e1 e6g6 e2h5 g6d6 a1d1 b6d8 e1g3 b7b6 h5e2 d7f6');
INSERT INTO evaluations VALUES(40771,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40771,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40771,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(40771,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40771,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40771,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(40771,36,'cp 24','e8g8 e1g1 c7c5 f1e1 c8e6 b1c3 b8c6 c1g5 f6d7 g5e7 d8e7 c3d5 e7d8 c2c3 a8c8 d5e3 d7f6 a2a4 d8d7 b2b3 h7h6 d3d4 c5d4 c3d4 e5d4 e3d5 e6d5 e4d5 c6e7 d1d4 a7a6 f3d2 e7g6 d2e4 f6e4 e1e4');
INSERT INTO evaluations VALUES(36228,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36228,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36228,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36228,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36228,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36228,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(36228,36,'cp 22','e1g1 c7c5 f1e1 c8e6 b1c3 b8c6 c1g5 f6e8 g5d2 e8f6 a2a4 a8c8 d2g5 f6d7 g5e7 d8e7 c3d5 e7d8 c2c3 d7f6 d5f6 d8f6 f3h2 g8h8 h2g4 f6g6 e1e3 f7f5 e4f5 f8f5 e3g3 g6f7 g3f3 c8f8 f3f5 f7f5 d1e2 f5f7 c4e6 f7e6 g4e3 h7h6 a1d1 d6d5 e2g4 e6g4');
INSERT INTO evaluations VALUES(36227,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(36227,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36227,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(36227,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36227,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36227,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36227,36,'cp 10','c7c5 f1e1 c8e6 b1c3 b8c6 c1g5 a8c8 g5f6 e7f6 c3d5 f6g5 c2c3 g5h6 a2a3 c8b8 a3a4 g7g6 a1b1 a7a6 b2b4 b7b5 c4a2 g8h8 a2b3 d8c8 b4c5 d6c5 a4b5 a6b5 d1e2 c8d8 d5e3 e6b3 b1b3 h6e3 e2e3 d8d6 e3h6');
INSERT INTO evaluations VALUES(52802,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(52802,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(52802,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(52802,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(52802,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52802,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(52802,36,'cp -7','c8g4 h2h3 g4h5 e1g1 e7e6 f1e1 b8d7 e4d5 c6d5 c2c4 d5c4 d3c4 g8f6 b1c3 a8c8 c4e2 f8e7 d2d4 e8g8 d4d5 d7c5 d5e6 c5e6 f3g5 d8d1 e2d1 h5d1 e1d1 e7c5 d1d3 f8d8 d3d8 c8d8 g5e6 f7e6 g1f1 e6e5 f1e2 g8f7 c1e3 c5e3 e2e3 d8d4 a1e1 d4b4 b2b3 f7e7 f2f3');
INSERT INTO evaluations VALUES(31129,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(31129,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31129,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31129,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(31129,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(31129,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(31129,36,'cp 0','h2h3 g4h5 e1g1 e7e6 f1e1 b8d7 e4d5 c6d5 c2c4 d5c4 d3c4 a8c8 c4d5 d7c5 d5c4');
INSERT INTO evaluations VALUES(57095,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(57095,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(57095,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(57095,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(57095,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(57095,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(57095,36,'cp 100','g1f3 d7d6 c1e3 g8f6 f1d3 g7g6 c2c4 f8g7 b1c3 e8g8 e1g1 b8a6 h2h3 f6e8 a2a3 f7f5 e3g5 d8d7 e4f5 g6f5 d3c2 h7h6 g5e3 d7d8 f1e1 c7c5 e3d2 a6c7 g2g3 b7b5 c4b5 a8b8 a3a4 a7a6 f3h4 a6b5 a4b5 c7b5 c3b5 b8b5 a1a7 b5b2');
INSERT INTO evaluations VALUES(57093,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(57093,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(57093,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(57093,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(57093,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(57093,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(57093,36,'cp 82','d7d6 h2h3 g8f6 f1d3 g7g6 c2c4 f8g7 e1g1 b8d7 b1c3 e8g8 a2a3 a7a5 a1b1 h7h6 b2b4 f6h5 d3c2 h5f4 f1e1 a5b4 a3b4 d7f6 c4c5 f6h5 c3e2 f4e2 d1e2 b7b6 c5b6 c7b6 c1e3 c8d7 b1a1 h5f4 e2d2 g8h7 a1a8 d8a8 e3f4');
INSERT INTO evaluations VALUES(14527,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(14527,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(14527,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(14527,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(14527,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(14527,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(14527,36,'cp -76','d2d3 f8g7 a2a4 c8g4 f3g3 e8g8 g3h4 d8d7 f2f3 g4e6 c4e6 d7e6 g1h3 d6d5 e1g1 d5d4 b1a3 a8d8 h3f2 c6a5 c1g5 a5b3 a1e1 d8d7 f3f4 e5f4 c3d4 b3d4 g5f4 f6h5 f4g5 e6b3 e4e5 h7h6 g5d2 b3a4 a3c4');
INSERT INTO evaluations VALUES(14528,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(14528,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(14528,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(14528,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(14528,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(14528,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(14528,36,'cp -78','f8g7 a2a4 c8g4 f3g3 e8g8 h2h3 g4d7 c1g5 h7h6 g5f6 d8f6 b1d2 h6h5 g3f3 f6g5 d2f1 g7h6 f1e3 g8g7 f3e2 a8e8 g1f3 g5d8 b2b4 c6e7 e1g1 f7f5 f3d2 f5f4 e3c2 e7g8 d3d4 g8f6 d4e5 e8e5 f1d1 g7h8 f2f3 e5g5');
INSERT INTO evaluations VALUES(8774,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(8774,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(8774,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(8774,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(8774,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(8774,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(8774,36,'cp -74','c1g5 c6a5 c4b5 c7c6 b5a4 e8g8 b1d2 h7h6 g5f6 g7f6 f3e3 f6g5 e3e2 h6h5 g1f3 g5h6 e1g1 b7b5 a4c2 c6c5 f1d1 c8e6 d2f1 a8c8 f1e3 a7a6 g2g3 b5b4 h2h4 c8b8 e3g2 b4c3 b2c3 b8b2 g2e3 d8e7');
INSERT INTO evaluations VALUES(8770,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(8770,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(8770,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(8770,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(8770,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(8770,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(8770,36,'cp -72','c6a5 c4b5 c7c6 b5a4 e8g8 b1d2 h7h6 g5f6 g7f6 f3e3 f6g5 e3e2 h6h5 g1f3 g5h6 a4c2 c6c5 e1g1 b7b5 f1e1 a8b8 d2f1 c8e6 f1e3 d8c7 h2h4 b5b4 c2a4 b4c3 b2c3 b8b7 e2c2 f8b8 a1b1 h6e3 f2e3 g8g7 f3g5 b7b1 e1b1 b8b1 c2b1');
INSERT INTO evaluations VALUES(15426,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(15426,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(15426,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(15426,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(15426,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(15426,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(15426,36,'cp 42','f6e4 f1e1 e4g5 f3g5 d8g5 d2d4 g5d8 d4e5 d6e5 e1e5 c8e6 d1a4 d8d7 e5e1 a7a6 b5c4 e8c8 c4e6 f7e6 b1d2 f8e7 d2f3 e6e5 c1e3 d7f5 h2h3 h8e8 a1d1 d8d1 e1d1 h7h6 f3d2 e8d8 d1e1 f5g6 d2e4 g6f7 e4g3 c8b8 e1e2 e7g5 e3g5 h6g5');
INSERT INTO evaluations VALUES(15677,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(15677,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(15677,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(15677,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(15677,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(15677,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(15677,36,'cp 42','f1e1 e4f6 d2d4 c8d7 b5c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 c7c6 b1d2 d8c7 e5d3 f8d6 g5f6 g7f6 d2e4 a8e8 e4f6 e8e1 d3e1 d5e6 e1f3 d6e7 f6h5 h8d8 f3d4 e6g4 h5f4 c6c5 d4c2 e7g5 c2e3 g5f4 e3g4 d8d2 g2g3 f7f5 g3f4 f5g4 f4f5 c7d8 f5f6 d8e8 a1e1 e8f8 e1e7 d2b2 e7h7 b2a2 h7b7');
INSERT INTO evaluations VALUES(26149,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(26149,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(26149,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(26149,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(26149,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(26149,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(26149,36,'cp -21','c2c3 g8e7 g1e2 e7c6 e1g1 f8e7 b1d2 a7a5 a2a4 d8b6 d2b1 a8c8 b1a3 e8g8 a3b5 c6a7 b5a7 b6a7 c1e3 a7b6 a1b1 b6c6 e2f4 c5d4 c3d4 b7b5 d1g4 b5a4 f4h5 g7g6 e3h6 c6b6 h6f8 g8f8 h5g3 c8c4 g3e2 c4c2 h2h4 c2b2 b1b2 b6b2');
INSERT INTO evaluations VALUES(26144,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(26144,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(26144,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(26144,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(26144,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(26144,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(26144,36,'cp -11','g8e7 g1e2 d8b6 e1g1 e7f5 d1d3 f8e7 b1d2 c5d4 c3d4 a8c8 d2f3 d7b8 c1d2 b6a6 d3a6 b8a6 a1c1 e8d7 g2g4 f5h4 f3h4 e7h4 g1g2 a6b8 b2b4 a7a6 a2a4 c8c1 f1c1 b8c6 b4b5 a6b5 a4b5 c6e7 c1c5 h8c8 d2e3');
INSERT INTO evaluations VALUES(47740,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(47740,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(47740,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(47740,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(47740,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(47740,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(47740,36,'cp 138','b1c3 f8d6 f1e2 b8c6 e1g1 c8d7 e2d3 e6e5 d4d5 c6e7 a2a4 e8g8 a4a5 b6c8 c1e3 h7h6 f3d2 c7c6 d5c6 e7c6 d1h5 d7e6 f1d1 c6a5 e3h6 a5b3 d2b3 e6b3 h6e3 d8f6 d1c1 c8e7 a1a7 a8a7 e3a7 d6b4 a7e3 f8c8 h5e2 e7g6 g2g3 b4c3 b2c3');
INSERT INTO evaluations VALUES(33746,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(33746,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(33746,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(33746,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(33746,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(33746,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(33746,36,'cp 32','g1f3 e7e5 g2g3 f8d6 f1g2 e8g8 e1g1 c7c6 f1e1 f8e8 a2a4 h7h6 h2h3 a5d8 g3g4 b8d7 d3d4 e5d4 e1e8 d8e8 d1d4 d6c7 c1d2 a7a5 a1e1 e8f8 d2e3 c7d6 c3e4 f6e4 d4e4 d7f6');
INSERT INTO evaluations VALUES(33745,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(33745,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(33745,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(33745,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(33745,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(33745,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(33745,36,'cp 27','g7g6 f1e2 f8g7 e1g1 e8g8 a2a4 b8c6 f1e1 e7e5 c1d2 a5c5 c3e4 f6e4 d3e4 c6d4 f3d4 c5d4 d2e3 d4b4 a4a5 a7a6 f2f3 c8e6 d1c1 f8d8 e3g5 b4c5 g1h1 f7f6 g5e3 c5c6 c2c3 g7f8 b2b4 d8d7 c1c2 a8d8');
INSERT INTO evaluations VALUES(34714,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(34714,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34714,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(34714,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34714,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34714,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34714,36,'cp 0','h5e2 g8f6 g1f3 g7g6 b1c3 f8g7 d2d4 e5d4 f3d4 b8c6 d4c6 b7c6 e1g1 e8g8 c1f4 d6d5 e4d5 e7e2 c4e2 f6d5 c3d5 c6d5 e2b5 c8e6 a1b1 c7c6 b5a6 e6f5 a6d3 f5e6');
INSERT INTO evaluations VALUES(39415,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39415,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39415,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39415,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(39415,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(39415,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(39415,36,'cp -39','b1c3 b8c6 g1f3 d7d6 f3g5 g8h6 e1g1 e8g8 f2f4 g8h8 a2a3 b4c3 b2c3 d8e7 f1e1 c8g4 g5f3 f7f5 h2h3 g4h5 c4d5 a8e8 d1d3 h5f3 d3f3 c6d8 c1d2 c7c6 e4f5 e7f6 d5b3 h6f5 g2g4 f5h4 f3g3 d6d5 b3c2 f6h6 e1e8 f8e8');
INSERT INTO evaluations VALUES(46645,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46645,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46645,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46645,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(46645,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46645,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(46645,36,'cp 5','g1f3 f8g7 c2c3 c7c5 a2a4 e8g8 e1g1 b8c6 f1e1 h7h6 h2h3 d6d5 e4d5 f6d5 a4a5 e7e6 a5a6 d8c7 a6b7 c8b7 c4d5 e6d5 d3d4 f8e8 e1e8 a8e8 b1a3 c5c4 a3b5 c7b8 b5a3');
INSERT INTO evaluations VALUES(46642,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46642,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46642,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46642,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(46642,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(46642,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46642,36,'cp 0','f8g7 e1g1 c7c5 a2a4 b8c6 c2c3 e8g8 f1e1 d6d5 e4d5 f6d5 a4a5 e7e6 a5a6 d8c7 h2h3 d5b6 a6b7 c8b7 c4a6 b7a6 a1a6 f8d8 d1e2 b6d5 e2e4 c7c8 a6a3 c8c7');
INSERT INTO evaluations VALUES(41234,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41234,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41234,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41234,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41234,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41234,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41234,36,'cp 0','c2c3 c7c5 e1g1 e8g8 a2a4 b8c6 f1e1 d6d5 e4d5 f6d5 a4a5 e7e6 a5a6 d8c7 h2h3 d5b6 a6b7 c8b7 c4a6 b7a6 a1a6 f8d8 d1e2 b6d5 e2e4 c7b7 a6a3 b7c7');
INSERT INTO evaluations VALUES(41232,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41232,36,'cp 0','c7c5 e1g1 e8g8 a2a4 b8c6 f1e1 d6d5 e4d5 f6d5 a4a5 e7e6 a5a6 d8c7 h2h3 d5b6 a6b7 c8b7 c4a6 b7a6 a1a6 f8d8 d1e2 b6d5 f3g5 c6e5 b1a3 c7b7 a6a5 b7c7');
INSERT INTO evaluations VALUES(40333,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40333,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40333,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40333,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40333,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40333,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40333,36,'cp 0','e1g1 e8g8 f1e1 b8c6 a2a4 d6d5 e4d5 f6d5 a4a5 e7e6 a5a6 d8c7 h2h3 d5b6 a6b7 c8b7 c4a6 b7a6 a1a6 f8d8 d1e2 b6d5 f3g5 c6e5 b1a3 c7b7 a6a5 b7c7');
INSERT INTO evaluations VALUES(40332,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40332,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40332,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40332,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40332,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40332,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40332,36,'cp 0','e8g8 f1e1 b8c6 a2a4 d6d5 e4d5 f6d5 a4a5 e7e6 a5a6 d8c7 h2h3 d5b6 a6b7 c8b7 c4a6 b7a6 a1a6 f8d8 d1e2 b6d5 f3g5 c6e5 b1a3 c7b7 a6a5 b7c7');
INSERT INTO evaluations VALUES(57658,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(57658,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57658,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(57658,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(57658,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(57658,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57658,36,'cp 9','g8f6 c1f4 d7d6 b1d2 f8g7 e2e4 b8d7 g1f3 e7e5 d4e5 d6e5 f3e5 f6h5 e5d3 h5f4 d3f4 e8g8 g2g3 a7a5 d1c2 c7c6 f1e2 a5a4 e1g1 a4a3 b2b4 d7b6 d2f3 b6a4 a1d1 d8e7 f3d4 c8d7 e2f3 a4b2 d1e1 e7d6 f3g2 f8d8 f4d3 b2d3 c2d3');
INSERT INTO evaluations VALUES(14616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14616,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14616,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14616,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14616,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14616,36,'cp 0','c6a5 c4b3 f8e7 e1g1 a5b3 a2b3 c7c5 f3d2 e8g8 h2h3 a7a6 f2f4 e5f4 e3f4 b7b5 d1f3 c8b7 c3e2 f6h7 f3g3 g8h8 d2f3 h7f6 f4d2 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(41720,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(41720,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(41720,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(41720,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(41720,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(41720,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(41720,36,'cp -1','e8g8 c1g5 c8e6 b1d2 c7c6 c2c4 h7h6 g5h4 b8d7 d1b3 d6f4 h4g3 f4g3 h2g3 d8c7 c4c5 e6f5 e2d3 f5d3 b3d3 a7a5 f1e1 f8b8 a1c1 b8e8 e1e8 a8e8 c1e1 e8e1 f3e1 d7f8 e1f3 f8e6 d3c3 b7b5 c5b6 c7b6 d2b3 a5a4 b3c5 e6c5 d4c5 b6a6');
INSERT INTO evaluations VALUES(36949,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(36949,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(36949,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(36949,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(36949,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(36949,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(36949,36,'cp -7','c1g5 c8e6 b1d2 b8d7 c2c4 h7h6 g5h4 f8e8 f1e1 c7c6 e2d3 d5c4 d3c4 e6d5 c4d5 e8e1 f3e1 c6d5 d1b3 d8c7 h4g3 d6g3 h2g3 a8c8 e1d3 c7b6 b3b6 d7b6 d3c5 g7g6 f2f3 f6e8 d2f1 e8d6 f1e3 h6h5 g1f2 b6c4 e3c4 d6c4');
INSERT INTO evaluations VALUES(50824,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(50824,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(50824,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(50824,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(50824,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(50824,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(50824,36,'cp 87','g1f3 g8f6 f1d3 b8c6 c2c3 e7e5 e1g1 g7g6 a2a4 e5d4 c3d4 f8g7 b1c3 e8g8 h2h3 c6b4 d3b1 c7c5 d4d5 f8e8 c1f4 b7b6 d1d2 d8e7 f4h6 g7h6 d2h6 c8d7 h6d2 e7f8 f3g5 f8g7 f2f4 b6b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(50823,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(50823,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(50823,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(50823,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(50823,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(50823,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(50823,36,'cp 79','g8f6 f1d3 b8c6 e1g1 e7e5 c2c3 f8e7 c1e3 f6g4 e3d2 g4f6 d4d5 c6b8 c3c4 c8g4 h2h3 g4h5 d2e3 b8d7 b1d2 g7g5 d3e2 h7h6 f3h2 h5g6 e2f3 a6a5 f1e1 h6h5 a1c1 d7c5 e3c5 d6c5 h2f1 e7d6 f1e3 d8d7 d1e2 e8g8 c1c3 a5a4');
INSERT INTO evaluations VALUES(3545,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(3545,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(3545,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(3545,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(3545,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(3545,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(3545,36,'cp 81','d4d5 c6e7 b2b4 a7a5 b4a5 a8a5 a2a4 a5a8 f3d2 f6d7 d2b3 f7f5 f2f3 b7b6 a1a3 d7c5 b3c5 b6c5 a4a5 f5f4 c1d2 g6g5 d1a1 c8a6 c3d1 c7c6 d1c3 h7h5 a1b1 c6d5 c4d5 a6e2 c3e2 g5g4 a5a6 a8a7 b1b3 d8e8 f1a1 e8g6');
INSERT INTO evaluations VALUES(3528,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(3528,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(3528,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(3528,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(3528,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(3528,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(3528,36,'cp 68','c6e7 b2b4 a7a5 b4a5 a8a5 f3d2 a5a8 a2a4 f6d7 d2b3 f7f5 f2f3 b7b6 a1a3 d7c5 b3c5 b6c5 a4a5 f5f4 c1d2 g6g5 d1a1 c8a6 f1b1 h7h5 b1b2 g7f6 c3d1 g8h7 d2e1 e7g6 d1f2 f8g8 a1c1 g8h8 a3b3 d8e8 c1c2 e8d7 c2d1');
INSERT INTO evaluations VALUES(3461,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(3461,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(3461,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(3461,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(3461,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(3461,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(3461,36,'cp 60','b2b4 a7a5 b4a5 a8a5 a2a4 a5a8 c1e3 f6g4 e3d2 b7b6 h2h3 g4f6 a1a3 f6d7 a4a5 b6a5 c3b5 d7c5 a3a5 c8d7 a5a8 d8a8 d1c2 f7f5 f3g5 a8d8 d2e3 c5e4 g5e4 f5e4 g2g4 g8h8 c2e4 e7g8 f1a1 g8f6 e4c2 d7b5 c4b5 f6d5');
INSERT INTO evaluations VALUES(3456,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(3456,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(3456,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(3456,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(3456,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(3456,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(3456,36,'cp 57','a7a5 b4a5 a8a5 a2a4 a5a8 a1a3 f6e8 f1e1 b7b6 e2d3 c8g4 h2h3 g4f3 d1f3 f7f5 c3b5 g8h8 c1d2 e8f6 e1a1 f5e4 d3e4 e7f5 e4f5 g6f5 a4a5 b6a5 d2a5 f6e8 a5b4 a8a3 f3a3 e5e4 b4c3 f5f4 c3g7 e8g7 c4c5 g7f5 c5d6 f5d6 b5d6 d8d6');
INSERT INTO evaluations VALUES(49281,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(49281,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(49281,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(49281,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(49281,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(49281,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(49281,36,'cp 158','b1c3 b8c6 f2f4 f8b4 g1f3 b4c3 b2c3 d7d5 c4c5 b6c4 f1d3 h7h6 a1b1 b7b6 e1g1 b6c5 d1e1 c5d4 e1g3 g7g6 f3d4 d8d7 f4f5 c6d4 c3d4 e6f5 e5e6 f7e6 g3g6 d7f7 d3f5 c4d6 g6f7 e8f7 f5d3 f7e7 c1f4 c8d7 b1c1 c7c6 f1f3 a8f8 c1f1 d6c4 h2h4 e6e5 d4e5 d7f5 d3c4 d5c4');
INSERT INTO evaluations VALUES(49279,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(49279,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(49279,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(49279,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(49279,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(49279,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(49279,36,'cp 168','b8c6 f2f4 f8b4 g1f3 b4c3 b2c3 d7d5 c4d5 b6d5 c3c4 d5e7 f1e2 b7b6 e1g1 d8d7 f4f5 c8b7 d4d5 e6d5 f5f6 g7f6 e5f6 e7f5 c4d5 e8c8 d5c6 d7c6 c1d2 f5e3 d1c1 e3f1 d2f4 f1d2 c1c6 b7c6 f3d2 h8g8 g2g3 g8g6 a1c1 g6f6 d2f3 d8e8 g1f2');
INSERT INTO evaluations VALUES(57221,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(57221,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(57221,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(57221,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(57221,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(57221,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(57221,36,'cp -12','d7d5 d2d4 b8c6 d4e5 c6e5 g1f3 f8d6 b1d2 g8f6 f3e5 d6e5 e1g1 c7c6 c2c4 e8g8 d1c2 f8e8 d2f3 e5d6 c4d5 f6d5 b2b3 d8f6 c1b2 c8f5 b2f6 f5c2 f6b2 a8d8 f1c1 c2g6 c1d1 d6e7 h2h3 d5b4 e2c4 h7h6 f3e5 e7f6 e5g6 f6b2 a1b1 b2f6 d1d8');
INSERT INTO evaluations VALUES(56476,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(56476,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(56476,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(56476,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(56476,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(56476,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(56476,36,'cp -11','d2d4 b8c6 d4e5 c6e5 g1f3 f8d6 e1g1 c7c6 b1d2 g8f6 f3e5 d6e5 c2c4 e8g8 d1c2 f8e8 d2f3 e5d6 c4d5 f6d5 b2b3 d8f6 c1b2 c8f5 b2f6 f5c2 f6b2 a8d8 f1c1 c2g6 c1d1 d6e7 e2c4 d5b6 c4f1 f7f6 b2d4 e7a3 d4c3 b6d5 c3d4 d5b4 d4a7 d8d1 a1d1');
INSERT INTO evaluations VALUES(1808,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(1808,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(1808,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(1808,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(1808,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(1808,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(1808,36,'cp 163','d4b5 f8b4 c1d2 b4d2 b1d2 e8d8 e1g1 g7g5 f2f4 g5f4 d1f3 a7a6 b5c3 b7b5 c4d5 g8e7 d5c6 e7c6 c3d5 f6d4 g1h1 d7d6 a1d1 d4g7 f3f4 c8e6 d2f3 d8c8 b2b3 h8g8 f1f2 c8b7 d5f6 g8d8 d1e1 g7g6 c2c4 c6b8 e4e5 b8d7 f6d7 e6d7 e5d6 g6d6 f3e5 d6c5 c4b5 d7b5 f4f3 b5c6 e5c6 c5c6 f3c6 b7c6 f2f7 d8d6 h1g1 c6b7');
INSERT INTO evaluations VALUES(7452,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7452,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7452,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(7452,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(7452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7452,36,'cp 0','d7d6 c2c3 e8g8 d2d4 c5b6 a2a4 c8g4 c1g5 e5d4 a4b5 a6b5 a1a8 d8a8 g5f6 g7f6 b3d5 a8e8 d5c6 g4f3 d1f3 e8c6 f3g4 g8h8 g4f4 c6c5 e4e5 f6e5 f4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(7414,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7414,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7414,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7414,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7414,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7414,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7414,36,'cp 0','c2c3 e8g8 d2d4 c5b6 a2a4 c8g4 c1g5 e5d4 a4b5 a6b5 a1a8 d8a8 g5f6 g7f6 b3d5 a8e8 d5c6 g4f3 d1f3 e8c6 f3f6 c6c5 e4e5 f8e8 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(7419,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7419,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7419,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7419,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7419,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7419,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7419,36,'cp 0','e8g8 d2d4 c5b6 a2a4 c8g4 c1g5 e5d4 a4b5 a6b5 a1a8 d8a8 g5f6 g7f6 b3d5 a8e8 d5c6 g4f3 d1f3 e8c6 f3f6 c6c5 e4e5 f8e8 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(2757,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2757,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2757,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2757,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2757,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2757,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2757,36,'cp 0','d2d4 c5b6 a2a4 c8g4 c1g5 e5d4 a4b5 a6b5 a1a8 d8a8 g5f6 g7f6 b3d5 a8e8 d5c6 g4f3 d1f3 e8c6 f3f6 c6c5 e4e5 f8e8 f6g5 g8f8 g5h6 f8g8');
INSERT INTO evaluations VALUES(2753,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2753,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2753,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2753,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2753,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2753,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2753,36,'cp 0','c5b6 a2a4 c8g4 c1g5 e5d4 a4b5 a6b5 a1a8 d8a8 g5f6 g7f6 b3d5 a8e8 d5c6 g4f3 d1f3 e8c6 f3f6 c6c5 e4e5 f8e8 f6g5 g8f8 g5h6 f8g8');
INSERT INTO evaluations VALUES(14420,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14420,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14420,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(14420,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(14420,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(14420,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(14420,36,'cp 33','g1f3 e7e5 f3e5 d5c3 e5c6 c3d1 c6d8 d1b2 c1b2 f8b4 e1d1 e8d8 e2e3 c7c6 f1d3 c8e6 d1e2 b7b5 a2a3 b4e7 e3e4 f7f5 d4d5 f5e4 d3e4 c6d5 h1d1 d8c7 e4d3 e7c5 d3b5 c7b6 b5a4 h8f8 f2f3 a8d8 b2e5 d5d4 a1c1');
INSERT INTO evaluations VALUES(14418,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14418,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(14418,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(14418,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(14418,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(14418,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(14418,36,'cp 23','e7e5 f3e5 d5c3 e5c6 c3d1 c6d8 d1b2 c1b2 f8b4 e1d1 e8d8 e2e3 c8e6 f1d3 c7c6 a2a3 b4e7 d1e2 d8c7 h2h4 f7f5 a3a4 h8f8 g2g3 g7g6 h1c1 e6d5 b2a3 e7a3 a1a3 a8e8 e2f1 d5f3 a3c3 c7d6 d3e2 f3e2 f1e2 f5f4 g3g4 f4e3');
INSERT INTO evaluations VALUES(14567,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(14567,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(14567,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(14567,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(14567,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(14567,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(14567,36,'cp 151','e4d5 f6d5 f1e1 c8e6 b1d2 d8d7 c4b5 f8d6 f3e5 d6e5 e1e5 e8c8 d2b3 d5e7 b3c5 d7d6 c5e6 c6e5 e6d8 h8d8 b5a4 e7f5 c1f4 d6d4 f4e5 d4a4 d1d2 f7f6 e5c3 b7b6 h2h3 d8e8 a1e1 e8e1 d2e1 f5d6 e1e7 d6b5 e7g7 b5c3 b2c3 a4c2 g7h6 c2a2 h6f6');
INSERT INTO evaluations VALUES(14560,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(14560,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(14560,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(14560,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(14560,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(14560,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(14560,36,'cp 132','f6d5 f1e1 c8e6 b1d2 d8d7 c4b5 f8d6 f3e5 d6e5 e1e5 e8c8 d2e4 h8e8 c1d2 e6g4 e5d5 d7d5 d1g4 f7f5 b5c6 b7c6 g4h3 e8e4 d3e4 d5d2 h3f5 c8b7 f5g4 d2c2 e4e5 d8d5 a1f1 d5e5 g4g7 c2b2 h2h4 a7a5 f2f4 b2b6 g1h2 e5h5 g7g4 b6b5 f1f3 h5f5 g4g7 b5e2 g7h6');
INSERT INTO evaluations VALUES(14590,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(14590,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(14590,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(14590,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(14590,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(14590,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(14590,36,'cp 135','f1e1 c8e6 b1d2 f8c5 c4b5 e8g8 b5c6 b7c6 f3e5 f7f6 e5c4 f8e8 h2h3 c5f8 d2f1 d8d7 c1d2 a8d8 f1g3 c6c5 b2b3 e6f7 a2a3 e8e1 d2e1 d5b6 c4e3 b6d5 d1f3 d5e3 f2e3 f8d6 e1c3 a7a6 g3e4 d8e8 e4d6 c7d6 e3e4');
INSERT INTO evaluations VALUES(14592,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(14592,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(14592,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(14592,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(14592,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(14592,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(14592,36,'cp 124','c8e6 b1d2 f8c5 c4b5 e8g8 b5c6 b7c6 f3e5 f7f6 e5c4 f8e8 h2h3 c5f8 d2e4 a7a5 c1d2 a5a4 d1f3 d8d7 e1e2 d5b6 c4e3 b6d5 a2a3 d5e3 e2e3 e6d5 f3f4 c6c5 a1e1 g8h8 d2c3 d7f7 e4g3 e8e3 e1e3 f8d6 f4g4 d5c6 e3e6 a8e8 e6e8 c6e8');
INSERT INTO evaluations VALUES(17851,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(17851,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(17851,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(17851,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(17851,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(17851,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(17851,36,'cp 75','d2d4 d7d6 d1a4 d8d7 a4c2 g8f6 e1g1 f8e7 f1e1 d7g4 b1d2 e8g8 d4d5 c6b8 d2f1 g4g6 f1g3 f6h5 g3f5 c8f5 e4f5 g6h7 b2b4 b8d7 a2a4 h5f6 a4a5 a8e8 h2h3 e7d8 c1e3 e5e4 f3d4 d7e5 c4b3 g7g6 b3a4 g6f5 a4e8 f8e8');
INSERT INTO evaluations VALUES(17861,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17861,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(17861,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(17861,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(17861,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(17861,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(17861,36,'cp 60','d7d6 d1a4 d8d7 a4c2 g8f6 e1g1 f8e7 f1e1 d7g4 b1d2 e8g8 d4d5 c6b8 d2f1 g4g6 f1g3 f6h5 g3f5 c8f5 e4f5 g6h7 c1e3 h5f6 a1d1 b8d7 h2h3 e5e4 f3d4 a8e8 a2a3 d7e5 c4f1 g7g6 f5g6 h7g6 f2f4 e4f3 c2g6 f7g6 e3h6 f3g2');
INSERT INTO evaluations VALUES(17741,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(17741,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(17741,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(17741,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(17741,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(17741,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(17741,36,'cp 81','d1a4 d8d7 a4c2 g8f6 e1g1 f8e7 f1e1 d7g4 c4d5 f6d5 e4d5 c8f5 c2b3 e5e4 f3d2 c6d8 c3c4 e8g8 b1c3 e7f6 c3e4 f6d4 e4g3 g4g6 g3f5 g6f5 d2f3 d4f6 c1e3 b7b6 e3d4 d8b7 b3c3 f6d4 f3d4 f5f6 b2b4 f8e8 c3d2 g8f8 a2a3 e8e1 a1e1 a8e8 e1e8 f8e8 g2g3 e8f8 g1g2 b7d8 h2h3 f6e5 d2d3 g7g6 d4f3 e5g7 d3d2 g6g5');
INSERT INTO evaluations VALUES(32445,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(32445,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(32445,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(32445,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(32445,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(32445,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(32445,36,'cp 49','b8c6 c2c3 g8f6 d1b3 d8c8 d2d4 f8e7 f1e1 e8g8 d4e5 c6e5 f3e5 d6e5 b3c4 c8d7 b1d2 b7b5 c4e2 d7c6 d2f3 e7d6 c1g5 f6d7 g5e3 a7a5 a1c1 d7c5 e3c5 c6c5 b2b3 f8f7 c1d1 h7h6 c3c4 b5b4 d1d3 a8f8');
INSERT INTO evaluations VALUES(27132,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(27132,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(27132,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(27132,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(27132,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(27132,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(27132,36,'cp 63','c2c3 g8f6 d1b3 d8c8 f1e1 f8e7 d2d4 e8g8 d4e5 c6e5 f3e5 d6e5 b3c4 c8d7 b1d2 b7b5 c4e2 e7d6 d2f3 d7c6 c1g5 f6d7 a1c1 h7h6 g5h4 a7a5 b2b3 a5a4 b3b4 d7b6 f3d2 c6d7 h4g3 c7c5 b4c5 d6c5 c1b1 b6c4 d2c4 b5c4 e2c4 a8c8');
INSERT INTO evaluations VALUES(49611,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(49611,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(49611,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(49611,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(49611,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(49611,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(49611,36,'cp -91','f6e4 g1f3 d7d5 f3e5 f8d6 e5f3 e8g8 f1e2 c7c5 e1g1 b8c6 d2d3 e4f6 d3d4 f8e8 b1d2 c8f5 c2c4 f5g6 c4d5 f6d5 e2b5 d5f4 b5c6 b7c6 d2c4 g6d3 f1e1 e8e1 d1e1 c5d4 b2d4 f4e2 g1h1 e2d4 f3d4 d3c4 b3c4');
INSERT INTO evaluations VALUES(50942,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(50942,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(50942,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(50942,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(50942,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(50942,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(50942,36,'cp 96','f1d3 c7c5 d4d5 g7g6 e1g1 f8g7 h2h3 g8e7 c2c4 b8d7 a2a4 e8g8 a4a5 d7e5 f3e5 g7e5 a1a3 e5g7 b1d2 b7b5 a5b6 d8b6 b2b3 a6a5 d2f3 g7c3 c1d2 c3b4 a3a2 e6e5 d1a1 b4d2 f3d2');
INSERT INTO evaluations VALUES(50941,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(50941,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(50941,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(50941,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(50941,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(50941,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(50941,36,'cp 87','c7c5 d4d5 g7g6 c2c4 f8g7 h2h3 g8e7 a2a4 e8g8 a4a5 b8d7 e1g1 d7e5 f3e5 g7e5 b1d2 e5g7 d2f3 b7b5 a5b6 d8b6 f1e1 a6a5 c1f4 b6d8 d1b3 c8a6 f4d2 e6d5 e4d5 a6c8 b3a3 a5a4 b2b4 c8f5 d3f1 g7a1 a3a1');
INSERT INTO evaluations VALUES(54024,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(54024,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(54024,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(54024,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(54024,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(54024,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(54024,36,'cp 133','b8c6 b1c3 f8e7 f1d3 g8f6 e1g1 a7a6 h2h3 e8g8 d1e2 c8d7 c1f4 a8c8 a1d1 d6d5 e4e5 f6h5 f4c1 f7f5 e5f6 h5f6 d3b1 e7d6 a2a3 b7b5 f3e5 d7e8 c1g5 h7h6 g5h4 g7g5 h4g3 e8h5 e5g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(55121,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(55121,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(55121,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(55121,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55121,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(55121,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(55121,36,'cp -41','g8f6 g1f3 c7c5 d3d4 b8c6 e2e3 e7e6 b1d2 d8c7 f1d3 f8d6 e1g1 e8g8 d4c5 d6c5 a2a3 e6e5 e3e4 f8d8 e4d5 f6d5 d1c2 h7h6 f1e1 a7a5 d2c4 f7f6 c1e3 d5e3 c4e3 c6e7 a1d1 c8e6 d3c4 e6c4 e3c4 b7b5 c4e3 c5e3 e1e3 d8d1 c2d1');
INSERT INTO evaluations VALUES(55308,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(55308,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(55308,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55308,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(55308,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55308,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55308,36,'cp 0','b8c6 g1f3 e7e5 c1e3 f8b4 d3d4 e5e4 f3d2 g8f6 c3e2 e8g8 c2c3 b4e7 g2g3 c8f5 f1g2 f8e8 h2h3 e7d6 e1g1 h7h6 g3g4 f5g6 e2g3 d6g3 f2g3 a7a5 d1e2 a5a4 a2a3 c6e7 c3c4 d8d6 e2f2');
INSERT INTO evaluations VALUES(21157,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21157,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(21157,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(21157,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21157,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(21157,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21157,36,'cp 0','c1e3 e7e5 g1f3 f8b4 d3d4 e5e4 f3d2 g8f6 c3e2 e8g8 c2c3 b4e7 g2g3 f8e8 f1g2 c8f5 e1g1 d8d7 f1e1 a7a5 e3g5 f6g4 g5f4 g4f6');
INSERT INTO evaluations VALUES(29713,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(29713,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(29713,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(29713,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(29713,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(29713,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(29713,36,'cp 75','g1f3 e7e6 f1c4 c7c6 e1g1 f8e7 f1e1 b8d7 e3e4 f5g6 c1f4 e8g8 a1c1 a8c8 c4d3 a7a6 d1e2 f8e8 e1d1 b7b5 d3b1 d8b6 f3e5 d7e5 d4e5 g6h5 e2e3 b6e3 f2e3 h5d1 e5f6 e7f6 c3d1 e8d8 h2h3 c6c5 e4e5');
INSERT INTO evaluations VALUES(30033,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(30033,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(30033,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(30033,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(30033,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(30033,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(30033,36,'cp 47','d8e7 d1e2 e6f5 e2e7 f8e7 d2d4 b8c6 a2a3 c6a5 b1d2 c7c5 f1e2 e8c8 e1g1 c5d4 f1e1 d4d3 e2f1 a5c6 b2b4 f6e8 c1b2 e7f6 b2f6 e8f6 d2b3 f5e4 f3d2 e4f5 f2f3 c6d4 b3d4 d8d4 a1c1 h8d8 g2g3 c8b8 g1f2 h7h5 h2h4 d8d7 e1e5 g7g6');
INSERT INTO evaluations VALUES(32682,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(32682,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(32682,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(32682,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(32682,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(32682,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(32682,36,'cp 47','d1e2 e6f5 d2d4 b8c6 a2a3 c6a5 b1d2 c7c5 e2e7 f8e7 f1e2 e8c8 e1g1 c5d4 f1e1 a5c6 b2b4 d4d3 e2f1 f6h5 c1b2 e7f6 b2f6 h5f6 d2b3 f5e4 f3d2 e4f5 f2f3 c6d4 b3d4 d8d4 g1f2 c8b8 a1c1 h7h5 g2g3 h5h4 g3g4 f5g6 e1e7 h8d8 h2h3 d8c8 f2e3 c8d8 c1d1');
INSERT INTO evaluations VALUES(39242,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(39242,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(39242,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(39242,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(39242,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(39242,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(39242,36,'cp 82','d2d4 e5d4 c3d4 c5b4 b1c3 d7d5 d1b3 a7a5 e4d5 g8f6 f1b5 b8d7 a2a3 b4e7 g1f3 e8g8 c1g5 f6d5 b5d7 d5c3 d7c8 a8c8 g5d2 c3d5 d2a5 d5f4 e1g1 c8a8 a5d2 d8d5 b3d5 f4d5 a1c1 c7c6 f1e1 f7f6 g1f1 g8f7 h2h3 f8d8 g2g3 e7d6 f1g2');
INSERT INTO evaluations VALUES(39229,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(39229,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(39229,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(39229,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(39229,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(39229,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(39229,36,'cp 58','e5d4 c3d4 c5b4 b1c3 d7d5 d1b3 a7a5 b3d5 d8d5 e4d5 b8d7 a2a3 b4d6 f1c4 d7b6 c4a2 c8f5 g1f3 g8e7 f3e5 f7f6 e5c4 b6c4 a2c4 e7c8 c4b5 e8f7 c1d2 h8d8 a1c1 c8b6 h2h4 a5a4 b5e2 h7h6 e1g1 f7f8 f1e1 f5d7 g2g3 a8a5 c3e4 a5d5 e4d6 c7d6');
INSERT INTO evaluations VALUES(39301,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(39301,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(39301,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(39301,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(39301,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(39301,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(39301,36,'cp 65','c3d4 c5b4 b1c3 d7d5 d1b3 a7a5 b3d5 d8d5 e4d5 b8d7 a2a3 b4d6 g1f3 g8e7 f1c4 d7b6 c4a2 c8f5 f3e5 a5a4 c1d2 f5c2 e5c4 b6c4 a2c4 e8d7 a1c1 c2b3 c4b3 a4b3 e1e2 a8a5 e2d3 e7d5 d3c4 c7c6 c4b3 d5c3 c1c3 a5f5 d2e3 h8e8 h2h3 h7h5 h1c1 d6f4 c3d3 f4e3 f2e3 f7f6');
INSERT INTO evaluations VALUES(39298,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(39298,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(39298,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(39298,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(39298,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(39298,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(39298,36,'cp 59','c5b4 b1c3 d7d5 d1b3 a7a5 b3d5 d8d5 e4d5 b8d7 a2a3 b4d6 g1f3 g8e7 f1c4 d7b6 c4a2 c8f5 f3e5 f7f6 e5c4 b6c4 a2c4 e7c8 c3a4 e8f7 e1g1 c8b6 a4b6 c7b6 c1d2 a8c8 f1c1 g7g5 c1c3 h7h5 a1c1 c8c7 c4e2 c7c3 c1c3 f7e7 c3b3 d6c7 h2h3 f5d7 b3c3 c7d6 e2d3 h8c8 c3c8');
INSERT INTO evaluations VALUES(39400,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(39400,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(39400,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(39400,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(39400,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(39400,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(39400,36,'cp 66','b1c3 d7d5 e4d5 d8d5 g1f3 g8f6 f1e2 e8g8 e1g1 d5d6 c1g5 b4a5 a1c1 b8c6 d4d5 c6e7 c3b5 d6d5 b5c7 a5c7 c1c7 d5d1 f1d1 e7d5 c7c1 h7h6 g5d2 c8g4 h2h3 f8e8 d1e1 g4d7 f3d4 e8e4 d4b3 a8e8 e2f3 e4e1 c1e1 e8e1 d2e1 b7b6 b3d4 g8f8 g1h1 d5f4 a2a3 f6d5');
INSERT INTO evaluations VALUES(11858,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(11858,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(11858,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(11858,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(11858,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(11858,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(11858,36,'cp -14','g7g6 h2h3 f8g7 c1e3 e8g8 d1d2 c6c5 e1c1 c8e6 e3h6 a7a5 h6g7 g8g7 f3g5 f6h5 g5e6 f7e6 c1b1 d8b8 b1a1 b8b6 f2f3 a8b8 d1b1 b6b4 h3h4 b4d4 h1d1 h7h6 a2a3 g7h7 d1f1 h5g3 f1f2 c7c6 f3f4 e5f4 f2f4');
INSERT INTO evaluations VALUES(14335,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(14335,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(14335,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14335,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14335,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(14335,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14335,36,'cp 38','e2e3 c8g4 h2h3 g4h5 d2d3 e7e6 a2a3 a7a5 b1d2 h7h6 f1e2 f8d6 c2c4 h5g6 e1g1 e8g8 d3d4 g6h7 c4c5 d6e7 b2c3 f6e4 d2e4 d5e4 f3d2 e6e5 d4e5 e7c5 e2b5 c6e7 b5c4 c7c6 b3b4 c5b6 e5e6 e7d5 d1b3 a5b4 a3b4 d5c3');
INSERT INTO evaluations VALUES(14334,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(14334,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(14334,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(14334,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14334,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(14334,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(14334,36,'cp 33','c8g4 h2h3 g4h5 d2d3 e7e6 b1d2 f8d6 a2a3 d8e7 c2c4 e8g8 f1e2 a7a5 d3d4 b7b6 a1c1 c6a7 a3a4 f6e4 e1g1 f7f5 c4d5 e6d5 f3e5 h5e2 d1e2 e7e6 d2e4 f5e4 f2f4 e4f3 f1f3 f8f3 e2f3');
INSERT INTO evaluations VALUES(24972,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(24972,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(24972,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24972,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(24972,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(24972,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(24972,36,'cp 45','h2h3 g4h5 d2d3 e7e6 b1d2 f8e7 f1e2 e8g8 e1g1 a7a5 a2a3 h7h6 c2c4 c6b8 d3d4 h5g6 f3e5 g6h7 e2d3 b8d7 d3h7 f6h7 e5d3 c7c6 e3e4 d5e4 d2e4 d8b6 c4c5 b6c7 e4d2 h7f6 d2c4 b7b5 c4e3 f6d5 e3d5');
INSERT INTO evaluations VALUES(15742,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(15742,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(15742,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(15742,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(15742,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(15742,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(15742,36,'cp 13','d6e5 d4e5 f8e7 c1d2 e8g8 e1c1 e7d6 e5d4 d6e7 d2e3 c8g4 f2f3 d8d4 e3d4 g4e6 g2g3 a7a6 f1g2 a8d8 f3f4 c7c6 g2f3 d8d7 c3a4 f8d8 d4c5 d7d1 h1d1 d8d1 c1d1 g8f8 c5e7 f8e7 a4c5 e6c8 d1d2 e7d6 c5e4 f6e4 f3e4 h7h6 b2b4 c8e6 a2a3 d6c7 d2e3');
INSERT INTO evaluations VALUES(15830,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(15830,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(15830,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(15830,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(15830,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(15830,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(15830,36,'cp 8','d4e5 f8e7 c1d2 e8g8 e1c1 e7d6 e5d4 d6e7 d4e3 c7c6 f1d3 c8e6 a2a3 f8e8 h1e1 d8c7 e3g3 c7g3 h2g3 a8d8 f2f3 h7h5 d2g5 g8f8 e1e2 f6d5 g5e7 e8e7 c3e4 e6f5 d3c4 e7e5 d1e1 b7b5 c4d5 d8d5 e4c3 e5e2 c3e2');
INSERT INTO evaluations VALUES(15826,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(15826,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(15826,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(15826,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(15826,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(15826,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(15826,36,'cp 12','f8e7 c1d2 e8g8 e1c1 e7d6 e5d4 d6e7 d4e3 c7c6 f1d3 c8e6 h1e1 d8a5 e3g3 f8e8 e1e5 a5c7 e5e2 c7g3 h2g3 h7h6 f2f3 f6d7 g3g4 a8d8 a2a4 d7c5 d3f5 e7f6 b2b3 g7g6 f5e6 c5e6 c3e4 f6g7 e4f2 b7b6');
INSERT INTO evaluations VALUES(8709,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(8709,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8709,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(8709,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8709,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(8709,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(8709,36,'cp 23','c1d2 e8g8 e1c1 e7d6 e5d4 d6e7 d4f4 e7d6 f4a4 c8f5 d2g5 d8e8 a4e8 a8e8 f1b5 e8d8 b5d3 f5d3 d1d3 h7h6 g5f6 d6f4 c1b1 d8d3 c2d3 g7f6 c3d5 f4e5 b1c2 g8g7 h2h3 c7c6 d5e3 e5f4 a2a4 f8d8 h1d1 f4e3 f2e3');
INSERT INTO evaluations VALUES(21424,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(21424,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(21424,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(21424,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(21424,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(21424,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(21424,36,'cp 37','f1g2 g8f6 e1g1 f8e7 c2c4 e8g8 b2b3 b7b6 c1b2 c8b7 c4d5 e6d5 b1c3 f8e8 a1c1 e7d6 e2e3 a7a6 a2a3 c6b8 f3d2 b8d7 f1e1 d6f8 b3b4 c7c5 d4c5 b6c5 b4c5 d7c5 d2f3 h7h6 c3e2 a8c8 e2f4 a6a5 f3e5');
INSERT INTO evaluations VALUES(21460,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(21460,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(21460,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(21460,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(21460,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(21460,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(21460,36,'cp 69','b7b6 c2c3 a7a5 f1b5 c8d7 b5d3 c6e7 c1e3 h7h5 e3f2 d7c8 d1f3 c8a6 d3a6 a8a6 g1e2 e7f5 b1d2 a5a4 e2g3 f5h4 f3e2 a6a7 e1g1 g7g6 b2b4 f8e7 a1d1 g8h6 c3c4 c7c6 h2h3 h4f5 g3f5 h6f5 c4c5 a7a8 d2f3 e8f8 b4b5 c6b5 e2b5');
INSERT INTO evaluations VALUES(27264,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(27264,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(27264,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(27264,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(27264,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(27264,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(27264,36,'cp 56','b5d7 d8d7 d1d4 g8e7 b1c3 e7c6 d4c4 f8e7 c1f4 e8g8 a1d1 f8e8 f1e1 e7f8 b2b3 c6e5 c4d4 e5c6 d4d5 a7a6 d5d2 b7b5 f4g3 c6e7 a2a4 d7c6 d2d3 b5a4 c3a4 g7g6 f2f4 a6a5 h2h3 c6a6 g3f2 f8g7 d3a6 a8a6');
INSERT INTO evaluations VALUES(37944,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(37944,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(37944,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(37944,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(37944,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(37944,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(37944,36,'cp 79','h2h3 g8f6 c1e3 e8g8 a2a4 b8d7 f1e2 d8c7 d1d2 e7e5 d4e5 d7e5 e1g1 e5f3 e2f3 c8e6 e3f4 f8d8 a4a5 f6e8 f4g5 d8d7 f1e1 g7e5 f3e2 d6d5 e4d5 e6d5 d2e3 e8g7 e2g4 a8e8 e3a7 f7f5 c3d5 c6d5 g4f3 g7e6 g5e3 c7c2 a1c1');
INSERT INTO evaluations VALUES(12467,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12467,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12467,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12467,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12467,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12467,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12467,36,'cp 0','f8c5 d4c6 d7c6 d1d8 e8d8 c1f4 e6e5 e1c1 d8e8 f4e5 f6g4 e5g7 h8g8 g7d4 c5d4 d1d4 g4f2 h1g1 f2g4 g1h1');
INSERT INTO evaluations VALUES(34304,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(34304,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(34304,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(34304,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(34304,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(34304,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(34304,36,'cp 94','d2d4 c8f5 c2c4 d5d8 b1c3 e7e6 h2h3 h7h6 g2g4 f5h7 f1d3 h7d3 d1d3 f8b4 c1d2 b8d7 c3e4 b4e7 e1c1 g8f6 c1b1 f6e4 d3e4 d8c7 h1e1 e8c8 e4c2 h8e8 h3h4 e7d6 g4g5 h6g5 h4g5 d6f4 d2c3 e8h8 b1a1 d7b6 c2e2');
INSERT INTO evaluations VALUES(34300,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(34300,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(34300,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(34300,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(34300,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(34300,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34300,36,'cp 90','c8f5 c2c4 d5a5 c1d2 a5d8 h2h3 e7e6 b1c3 h7h6 d1b3 d8b6 g2g4 f5h7 f1g2 g8f6 e1g1 f8e7 f1e1 e8g8 b3b6 a7b6 d4d5 f6d7 g4g5 d7c5 g5h6 b8d7 d2e3 e7f6 e1d1 g7h6 e3h6 f6g7 h6e3 g7c3 b2c3 e6e5 d5c6 b7c6');
INSERT INTO evaluations VALUES(38232,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(38232,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(38232,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(38232,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(38232,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(38232,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(38232,36,'cp 236','d4e5 c8g4 c1f4 b8c6 f1b5 d6e5 d1d8 a8d8 f4e5 g4f3 g2f3 e7f6 e5f6 g8f6 b1c3 e8g8 b5c6 b7c6 a1d1 d8d6 d1d6 c7d6 e1e2 f8e8 h1d1 d6d5 d1d4 g7g6 d4a4 e8e7 e2f1 g8g7 b2b3 h7h6 a4a6 d5e4 c3e4 f6e4 f3e4 e7e4 a6a7 e4h4 f1g2 h4g4 g2f3 g4d4 a7c7 d4h4 f3g3');
INSERT INTO evaluations VALUES(38217,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(38217,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(38217,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(38217,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(38217,34,'cp 232',NULL);
INSERT INTO evaluations VALUES(38217,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(38217,36,'cp 235','c8g4 c1f4 b8c6 f1b5 d6e5 d1d8 a8d8 f4e5 g4f3 g2f3 e7f6 e5f6 g8f6 b1c3 f6h5 c3d5 e8g8 e1c1 c6e5 f3f4 e5g4 b5e2 h5f6 d5f6 g4f6 e2f3 c7c6 d1d8 f8d8 h1d1 d8d1 c1d1 f6d7 d1e2 g8f8 e2d3 g7g6 d3e3 a7a5 f3g4 f8e7 g4d7 e7d7 e3d4 d7d6 b2b3 f7f6 a2a4 b7b6 c2c3');
INSERT INTO evaluations VALUES(49468,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(49468,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(49468,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(49468,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(49468,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(49468,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(49468,36,'cp 36','f1g2 d7d5 d2d4 c7c5 c2c4 b8c6 c4d5 e6d5 b1c3 c5d4 f3d4 f8c5 d4b3 c5b4 c1d2 e8g8 a2a3 b4c3 d2c3 f6e4 e1g1 d8b6 b3d4 e4c3 b2c3 f8d8 a1b1 b6c5 d1b3 c6d4 c3d4 c5d6 b3b4 d6b4 a3b4 b7b6 b4b5 c8e6 e2e3 a8c8 f1c1 h7h6');
INSERT INTO evaluations VALUES(11446,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(11446,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(11446,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(11446,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(11446,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(11446,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(11446,36,'cp -33','c8g4 d2d3 f6d7 c1e3 f7f6 c3b1 d7c5 f1e1 c5e6 b1d2 c6c5 h2h3 g4h5 c2c3 d8e7 d1a4 c7c6 a4c2 e7d7 a2a4 a8d8 a4a5 h5f7 c2b3 e6d4 b3d1 d4f3 d1f3 d7d3 e1d1 f7e6 d2f1 d3b5 f1g3 d8d1 f3d1 f8e7 g3f5 e6f5');
INSERT INTO evaluations VALUES(24301,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(24301,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(24301,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(24301,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(24301,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(24301,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(24301,36,'cp -32','d2d3 g4h5 c3b1 f8d6 c1e3 d8d7 b1d2 f6g4 d2c4 g4e3 f2e3 f7f6 c4d6 c7d6 a2a4 h5f7 a4a5 e8g8 f3h4 d6d5 h4f5 a8d8 d1f3 f7e6 f5g3 d7e7 h2h3 e7b4 f3f2 h7h6 c2c3 b4e7 b2b4 c6c5 b4c5 e7c5 f1c1 d5e4 g3e4');
INSERT INTO evaluations VALUES(24302,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(24302,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(24302,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(24302,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(24302,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(24302,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(24302,36,'cp -42','f6d7 c1e3 f7f6 c3b1 d7c5 f1e1 c5e6 b1d2 f8e7 h2h3 g4h5 c2c3 d8d3 f3e5 h5d1 e5d3 d1c2 d3b4 c2a4 b2b3 a6a5 b4c2 a4b5 c2d4 e6d4 c3d4 b5d3 f2f3 e7b4 a1c1 a5a4 e1d1 a4b3 a2b3 e8g8 d4d5 a8a2 d5c6 b7c6 d2c4 d3e2 c1a1 e2d1');
INSERT INTO evaluations VALUES(17987,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17987,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(17987,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17987,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(17987,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(17987,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(17987,36,'cp 52','a4b3 g8f6 a2a4 c8g4 c2c3 f8e7 h2h3 g4d7 f1e1 e8g8 d2d4 h7h6 b1d2 f8e8 d2f1 e7f8 f1g3 c6a5 b3c2 c7c6 a4b5 a6b5 b2b3 d8c7 c1e3 g7g6 f3d2 f8g7 d4e5 d6e5 b3b4 a5c4 d2c4 a8a1 d1a1 b5c4');
INSERT INTO evaluations VALUES(38136,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(38136,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(38136,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38136,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38136,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(38136,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(38136,36,'cp 14','d4b3 c5b6 d1g4 e8f8 f1d3 b8c6 g4e2 g8e7 b1c3 c8d7 c3b5 c6b4 b5d6 e7c8 a2a4 b4d3 c2d3 a7a5 c1e3 d7c6 d6c8 a8c8 b3d4 b6d4 e3d4 c6a4 e1g1 a4b5 e2d2 f8g8 a1a5 b5d3 f1a1 d3g6 a5a7 h7h6 h2h3 g8h7 a7b7 c8c2 d2e3');
INSERT INTO evaluations VALUES(39617,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(39617,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39617,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39617,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(39617,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(39617,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(39617,36,'cp 32','e7e5 d2d4 e5d4 c3d4 d7d5 e4e5 f7f6 b1c3 c7c6 e5f6 g8f6 h2h3 e8g8 f1d3 d8e7 c1e3 g7h6 d1e2 h6e3 e2e3 e7e3 f2e3 b8d7 h1f1 f6e8 g2g4 e8d6 e1d2 b7b5 f3g5 d7f6 a2a3 c8d7 e3e4 d5e4 g5e4 d6e4 c3e4 f6e4 d3e4 a8e8 e4f3 e8e6 a1c1');
INSERT INTO evaluations VALUES(47528,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(47528,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(47528,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(47528,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(47528,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47528,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(47528,36,'cp 29','d1f3 e7e6 d2c3 f8d6 f3e2 d8e7 g1f3 b8d7 c1e3 d7b6 f3d2 e8g8 e1c1 b6c4 d2c4 e6e5 b2b3 b7b6 c4d6 c7d6 h1e1 c8b7 f2f4 a8e8 g2g3 h7h6 c1b2 f7f5 e2f2 b7e4 d1d2 e8c8 c3c4 f8d8 e1d1 e7c7');
INSERT INTO evaluations VALUES(47527,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(47527,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(47527,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(47527,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(47527,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(47527,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47527,36,'cp 25','e7e6 d2c3 f8d6 f3e2 d8e7 g1f3 b8d7 c1e3 d7b6 c4b3 a7a5 e3b6 c7b6 b3a4 c8d7 a4d7 e7d7 e1c1 e8g8 d1d3 d7c7 h1d1 a8d8 e2e4 d6e7 f3d4 e7f6 f2f4 f6d4 d3d4 d8d4 d1d4 b6b5 g2g4 h7h6 a2a3 g7g6 c1b1 c7c6 e4d3 c6h1 b1a2 h1c6');
INSERT INTO evaluations VALUES(48136,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(48136,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(48136,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(48136,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(48136,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(48136,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(48136,36,'cp 26','d2c3 f8d6 f3e2 b8d7 g1f3 d8e7 c1e3 d7b6 f3d2 b6c4 d2c4 e6e5 e1c1 e8g8 h1e1 c8e6 b2b3 a7a6 c4d6 c7d6 f2f4 e7c7 c1b2 a8c8 c3c4 b7b5 f4e5 d6e5 c4c5 f8d8 g2g4 f7f6 d1d8 c8d8 g4g5');
INSERT INTO evaluations VALUES(47822,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47822,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47822,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47822,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47822,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47822,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47822,36,'cp 0','c7c5 a1b1 b8c6 a2a3 a7a5 d2d3 f8e7 f1e2 e8g8 e1g1 c8d7 c1d2 f6e8 c3d5 e8f6');
INSERT INTO evaluations VALUES(10457,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(10457,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(10457,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(10457,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10457,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(10457,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(10457,36,'cp 0','d7d5 e4d5 f6d5 d1d2 c8e6 f1e2 d8e7 c3d5 e6d5 c2c4 d5f3 e2f3 a8d8 a2a3 f7f5 f3c6 b7c6 e1g1 e8g8 b2b4 c6c5 d2c3 f5f4 e3f4 e5f4 a1e1 e7d6 e1e4 d6d3 c3d3 d8d3 e4e7 d3a3 b4b5 f4f3 f1f3 a3f3 g2f3 f8f7 e7e8 f7f8');
INSERT INTO evaluations VALUES(38837,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38837,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(38837,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(38837,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(38837,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38837,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38837,36,'cp 25','h2g3 g8f6 c2c4 h7h6 b1c3 e8g8 e2e3 b8d7 a2a4 c7c5 c4d5 f6d5 c3d5 e6d5 b2b3 d7f6 d4c5 d8a5 d1d2 a5c5 d2d4 c5c7 f1e2 c8d7 a4a5 f8c8 e1g1 c7c3 f1b1 c8c7 b3b4 d7f5 b1d1 f5g4 b4b5 c3d4 f3d4 g4e2 d4e2 a8c8');
INSERT INTO evaluations VALUES(38836,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(38836,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(38836,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38836,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(38836,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(38836,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38836,36,'cp 20','g8f6 e2e3 b8d7 c2c4 e8g8 b1c3 h7h6 a1c1 d5c4 f1c4 c7c5 e1g1 c5d4 d1d4 d7b6 c4e2 d8d4 f3d4 c8d7 e3e4 e6e5 d4c2 f8d8 c2e3 a8c8 f2f3 d7e6 f1d1 g8f8 g3g4 f8e7 b2b3 d8d1 e3d1 a7a6 g1f2 c8d8 f2e1');
INSERT INTO evaluations VALUES(42114,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(42114,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(42114,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(42114,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(42114,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(42114,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(42114,36,'cp 21','e2e3 b8d7 c2c4 h7h6 b1c3 e8g8 d1c2 c7c6 f1e2 d5c4 e2c4 d8e7 e1g1 c6c5 f1d1 d7b6 c4e2 c8d7 a2a4 f8c8 a4a5 b6d5 c3d5 f6d5 d4c5 c8c5 c2b3 d7c6 f3d4 c6e8 b3a3 a8c8 e3e4 d5f6 f2f3 e6e5');
INSERT INTO evaluations VALUES(38872,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(38872,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(38872,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(38872,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(38872,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(38872,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(38872,36,'cp 83','b1c3 g8f6 e4e5 f6d5 c1d2 e8g8 d1g4 f7f5 e5f6 d5f6 g4h4 f6d5 h4d8 f8d8 g1f3 c7c5 d4c5 b4c5 e1g1 h7h6 a2a3 b8c6 c4a2 c6d4 f3d4 c5d4 f1d1 d4f6 c3d5 e6d5 d2c3 c8e6 c3f6 g7f6 h2h4 a8c8 d1d2 g8f7 a1d1 f7e7 a2d5 e6d5 d2d5 d8d5');
INSERT INTO evaluations VALUES(43378,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43378,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43378,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43378,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43378,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43378,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43378,36,'cp 0','e1g1 e7e6 a2a4 f8e7 c1f4 e8g8 f1e1 b7b6 b1d2 f6h5 f4e3 c8b7 c2c3 h5f6 e3f4 f6h5');
INSERT INTO evaluations VALUES(43377,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43377,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43377,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43377,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43377,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43377,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(43377,36,'cp 0','e7e6 a2a4 f8e7 c2c3 b7b6 c1f4 c8b7 b1d2 e8g8 d1e2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(43493,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43493,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43493,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43493,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43493,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43493,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43493,36,'cp 0','a2a4 f8e7 c2c3 b7b6 c1f4 c8b7 b1d2 e8g8 f1e1 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(44542,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(44542,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(44542,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(44542,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(44542,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(44542,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(44542,36,'cp -12','c5d4 c3d4 d8b6 d1c1 b8c6 e2e3 d7d6 b1c3 e7e5 f4g5 e5d4 e3d4 f8e7 g5e3 b6d8 a2a3 c8f5 c1d1 d6d5 f1d3 f5d3 d1d3 e8g8 g1f3 e7d6 e1g1 h7h6 f1e1 f8e8 e3d2 d8d7 h2h3 a8c8 a1d1 d6b8 e1e8 c8e8 d1e1 a7a6 e1e8 f6e8 d2e3 e8f6');
INSERT INTO evaluations VALUES(44627,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(44627,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(44627,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(44627,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(44627,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(44627,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(44627,36,'cp -15','c3d4 b8c6 e2e3 d8b6 d1c1 d7d6 b1c3 e7e5 f4g5 e5d4 e3d4 f8e7 g5e3 b6d8 a2a3 e8g8 f1d3 f6g4 g1f3 g4e3 c1e3 c8g4 e1g1 g4f3 e3f3 g7g6 a1d1 e7f6 d3e4 c6d4 f3e3 d4c6 f1e1 f8e8 g2g3 d8d7 g1g2 a8d8 e3d2 a7a6 c3d5 f6g7 e4f3 e8e1 d1e1 c6e5');
INSERT INTO evaluations VALUES(44617,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(44617,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(44617,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(44617,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(44617,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(44617,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(44617,36,'cp -13','b8c6 e2e3 d8b6 d1c1 d7d6 b1c3 e7e5 f4g5 e5d4 e3d4 f8e7 g5e3 b6d8 g1e2 e8g8 e2f4 c8f5 c1d1 d6d5 f1e2 e7d6 e1g1 d6f4 e3f4 d8b6 d1d2 c6d4 f4e3 d4e2 c3e2 b6d6 f2f3 f8e8 e3d4 a8c8 a1c1 c8c4 d4f2 d6d7 a2a3 c4c1 f1c1');
INSERT INTO evaluations VALUES(48882,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(48882,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(48882,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(48882,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(48882,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(48882,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(48882,36,'cp -74','b8c6 h2h3 c8f5 d2d3 d8d7 b1d2 f8d6 b2b3 e8c8 a2a3 h7h5 c1b2 h8e8 b3b4 c8b8 d2b3 f5e6 d1d2 d5d4 e1c1 e6b3 c2b3 d4e3 f2e3 d7e6 d2c2 f6d5 c2d2 f7f5 c1b1 a7a6 h1f1 c6a7 e3e4 f5e4 d3e4 d5b6 d2e3 d6e7 d1d8 e7d8 b1a1 e6g6 a3a4');
INSERT INTO evaluations VALUES(15196,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(15196,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(15196,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(15196,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(15196,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(15196,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(15196,36,'cp -77','h2h3 c8f5 d2d3 d8d7 b2b3 e8c8 b1d2 h7h5 a2a3 h5h4 g3g4 f6g4 h3g4 f5g4 c1b2 g7g5 b3b4 d5d4 b4b5 c6e7 e3d4 e5d4 c2c4 d4c3 b2c3 h8h6 d1a4 g4e2 a4a7 d7b5 a1b1 b5d3 h1h3 e2f3 h3f3 h6e6 f3e3 d3c3 a7b7 c8d7 b7b5 e7c6 e3e6 f7e6 b5g5 f8a3 g2c6 c3c6 g5h4');
INSERT INTO evaluations VALUES(50045,30,'cp 898',NULL);
INSERT INTO evaluations VALUES(50045,31,'cp 895',NULL);
INSERT INTO evaluations VALUES(50045,32,'cp 883',NULL);
INSERT INTO evaluations VALUES(50045,33,'cp 901',NULL);
INSERT INTO evaluations VALUES(50045,34,'cp 890',NULL);
INSERT INTO evaluations VALUES(50045,35,'cp 900',NULL);
INSERT INTO evaluations VALUES(50045,36,'cp 905','e5h8 d7d5 d2d4 c8e6 g1f3 b8d7 b1c3 c7c6 f3g5 d8b6 f1e2 e8c8 e1g1 f8h6 h8h7 h6g5 c1g5 d8e8 g5e7 e8e7 h7h8 b6d8 h8d8 c8d8 f2f3 d5e4 f3e4 f7f6 g1f2 c6c5 d4d5 e6g8 h2h4 e7h7 g2g3 d8e7 f2g2 h7h8 d5d6 e7e6 e2c4 e6e5 c4g8 h8g8');
INSERT INTO evaluations VALUES(43307,30,'cp 888',NULL);
INSERT INTO evaluations VALUES(43307,31,'cp 894',NULL);
INSERT INTO evaluations VALUES(43307,32,'cp 905',NULL);
INSERT INTO evaluations VALUES(43307,33,'cp 892',NULL);
INSERT INTO evaluations VALUES(43307,34,'cp 894',NULL);
INSERT INTO evaluations VALUES(43307,35,'cp 905',NULL);
INSERT INTO evaluations VALUES(43307,36,'cp 904','d7d5 d2d4 c8g4 c2c4 d5e4 c1h6 b8d7 h2h3 g4f5 g2g4 c7c5 d4d5 d8a5 h6d2 a5b6 d2c3 h7h5 g4f5 e7f5 b1a3 e8c8 h8h7 f5d6 e1c1 f8e7 a3b5 d6b5 c4b5 d8f8 h7h6 e7d6 h3h4 d6e7 a2a4 b6d6 h6d2 f7f5 a4a5 d7e5 d2e3');
INSERT INTO evaluations VALUES(43304,30,'cp 926',NULL);
INSERT INTO evaluations VALUES(43304,31,'cp 921',NULL);
INSERT INTO evaluations VALUES(43304,32,'cp 925',NULL);
INSERT INTO evaluations VALUES(43304,33,'cp 928',NULL);
INSERT INTO evaluations VALUES(43304,34,'cp 928',NULL);
INSERT INTO evaluations VALUES(43304,35,'cp 926',NULL);
INSERT INTO evaluations VALUES(43304,36,'cp 926','d2d4 c8e6 g1f3 d8d6 c2c4 c7c6 e4d5 c6d5 c4c5 d6c7 b1c3 a7a6 f3g5 b8d7 h8h7 d7f6 h7h8 f6d7 f1d3 e8c8 h8h7 d7f6 g5e6 f6h7 e6c7 c8c7 c1f4 c7c6 d3c2 b7b6 c5b6 c6b6 c2b3 h7f6 f4g5 f8g7 g2g4 e7c6 e1c1 c6a5 b3c2 a5c4 h2h3 d8e8');
INSERT INTO evaluations VALUES(7774,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(7774,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(7774,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(7774,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7774,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(7774,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(7774,36,'cp 12','h7h6 g5h4 d5c4 d3c4 b7b5 c4e2 a7a6 e1g1 c8b7 f3e5 e8g8 h4f6 d7f6 e5d3 d8c7 e2f3 a8d8 d1c2 e7d6 g2g3 e6e5 d4e5 d6e5 c3e4 f6e4 f3e4 e5d6 a1c1 b7c8 d3c5 c7b6 b2b4 d6c5 c2c5 b6c5 c1c5');
INSERT INTO evaluations VALUES(7756,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(7756,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(7756,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(7756,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(7756,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7756,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(7756,36,'cp 12','g5h4 d5c4 d3c4 b7b5 c4d3 c8b7 a1c1 a8c8 a2a3 a7a5 e1g1 e8g8 f3e5 d7e5 d4e5 f6d7 h4e7 d8e7 f2f4 d7c5 d3b1 b5b4 d1c2 g7g6 c3e2 c5d7 f1d1 c6c5 d1d6 d7f6 c2d2 f6e4 b1e4 b7e4 e2g3 e7b7 h2h3 b4a3 b2a3 c5c4 g3e4 b7e4 g1f2 c8b8 d2d4 e4d4 d6d4 b8b2 f2f3');
INSERT INTO evaluations VALUES(7758,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(7758,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(7758,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(7758,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(7758,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(7758,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(7758,36,'cp 28','d5c4 d3c4 e8g8 a2a4 b7b6 e1g1 c8b7 d1e2 f6d5 h4g3 d5c3 b2c3 d7f6 a4a5 c6c5 a5b6 a7b6 e2b2 f6e4 g3e5 d8d7 c4b5 d7c8 a1a8 b7a8 f1b1 a8c6 b5f1 b6b5 d4c5 e4c5 c3c4 f7f6 e5d4 b5c4 f1c4 g8h8 b1c1 c6b7 d4c5 e7c5 c4d3 b7f3 g2f3');
INSERT INTO evaluations VALUES(42988,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42988,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(42988,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(42988,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(42988,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(42988,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(42988,36,'cp 33','g2g3 d7d5 f1g2 d5c4 e1g1 e8g8 d1c2 a7a6 a2a4 b8c6 c2c4 d8d5 c4d3 d5f5 d3f5 e6f5 c1d2 f8d8 e2e3 c6b4 f1c1 c7c6 d2b4 e7b4 f3e5 a6a5 e5d3 b4e7 b1d2 g7g5 d2b3 g8f8 h2h4 h7h6 b3c5 g5h4 g3h4 e7d6');
INSERT INTO evaluations VALUES(42989,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(42989,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(42989,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(42989,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(42989,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(42989,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(42989,36,'cp 34','d7d5 f1g2 e8g8 e1g1 d5c4 d1c2 c7c6 c2c4 b7b5 c4b3 c8b7 b1c3 b8d7 f3e5 d7e5 d4e5 f6d5 c3d5 e6d5 c1e3 d8d7 f1d1 d7e6 f2f4 a7a5 a1c1 a5a4 b3d3 b5b4 e3f2 b7a6 d3d2 a4a3 b2b3 h7h6');
INSERT INTO evaluations VALUES(47257,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(47257,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(47257,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(47257,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(47257,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(47257,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47257,36,'cp -33','d2d4 e7e6 g1f3 c7c5 d4c5 f8c5 f1d3 b8c6 e1g1 e8g8 a2a3 e6e5 e3e4 d5d4 c3e2 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 d1e2 a8c8 f1e1 d8d6 c3b5 d6e7 c1e3 f8d8 a1c1 c5e3 e2e3 f6h5 d3f1 a5a4 g2g3 h5f6 e1d1 d8d7 d1d7 e7d7');
INSERT INTO evaluations VALUES(52039,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(52039,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(52039,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(52039,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(52039,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(52039,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(52039,36,'cp -49','f8b4 g1f3 g8e7 c1d2 b8c6 a2a3 b4a5 e1g1 a7a6 e4e5 e8g8 c4d3 h7h6 d1c2 b7b5 f1e1 c8b7 c3e4 a8c8 c2d1 a5c7 e4c5 b7a8 c5a6 c7b6 a1c1 e7g6 d2e3 b6e3 e1e3 c6e7 c1c8 e7c8 a6b4 c8b6 d3e4 a8e4');
INSERT INTO evaluations VALUES(37697,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(37697,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(37697,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(37697,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(37697,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(37697,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(37697,36,'cp -33','g1f3 b8c6 e1g1 g8e7 c1e3 e8g8 d1b3 d8a5 a1c1 b4c3 c1c3 f8d8 c4d3 h7h6 f1c1 d7d5 e3c5 d5e4 d3e4 e7d5 c3c2 c8d7 f3d4 d5f6 e4c6 d7c6 d4c6 b7c6 c5e7 d8b8 b3a3 a5f5 e7d6 b8b7 c2c6 a8d8 a3g3');
INSERT INTO evaluations VALUES(37696,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(37696,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(37696,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(37696,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(37696,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(37696,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(37696,36,'cp -27','b8c6 e1g1 g8e7 c1e3 e8g8 d1b3 d8a5 a1c1 b4c3 c1c3 f8d8 f1d1 d7d5 e4d5 e6d5 c4e2 c8g4 e3c5 d5d4 c5e7 d4c3 e7d8 a8d8 b2c3 d8d1 e2d1 g4c8 d1c2 g7g6 c2e4 a5c7 e4d5 g8g7 b3a4 c6e7 a4d4 f7f6 h2h4 e7d5 d4d5 c7c3');
INSERT INTO evaluations VALUES(5196,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(5196,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(5196,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(5196,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(5196,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(5196,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(5196,36,'cp -18','e1g1 g8e7 c1e3 e7g6 d1d2 e8g8 a1d1 b7b6 a2a3 b4c3 d2c3 d8e7 h2h4 g6h4 f3h4 e7h4 f2f4 h4g4 f1f3 d7d5 c4e2 c8b7 f3f2 g4h4 e4d5 c6e7 d5e6 f7e6 d1d7 b7d5 c3e5 e7f5 e3d4 f8f7 d7f7 g8f7 d4c3 a8c8 e2d3 h4f6 e5f6 g7f6 d3f5 e6f5');
INSERT INTO evaluations VALUES(5195,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(5195,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(5195,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(5195,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(5195,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(5195,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(5195,36,'cp -44','g8e7 c3b5 d7d5 e4d5 e6d5 c4d5 e7d5 d1d5 e8g8 d5d8 f8d8 c1e3 c8g4 b5d4 c6d4 f3d4 g4d7 f1d1 a8c8 h2h3 b4e7 g2g4 g7g6 a1c1 c8c1 d1c1 e7f6 c1d1 d7e8 d1d2 a7a6 f2f3 h7h6 d2d1 f6g7 g1f2 g8h7 d1d2 d8d7 b2b3 d7d8');
INSERT INTO evaluations VALUES(7992,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(7992,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(7992,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(7992,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(7992,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(7992,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(7992,36,'cp -47','c1g5 h7h6 g5h4 e8g8 a1c1 a7a6 a2a3 b4c3 c1c3 d7d5 e4d5 e6d5 c4a2 f8e8 f1e1 c8e6 h2h3 d8b6 c3b3 b6c7 b3d3 a8c8 h4e7 c6e7 f3d4 e6d7 d3e3 c7d6 d1d3 c8d8 d3d2 g7g6 d4f3 e7f5 d2d5 d6d5 e3e8 d7e8 a2d5 g8g7');
INSERT INTO evaluations VALUES(52473,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52473,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(52473,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52473,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(52473,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(52473,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52473,36,'cp 4','e2e4 g8f6 e4e5 f6d5 g1f3 b8c6 d2d4 c5d4 f1c4 d5b6 c4b3 d7d5 e5d6 d8d6 f3d4 c6d4 c3d4 c8e6 b1c3 e6b3 a2b3 g7g6 d1f3 f8g7 f3b7 e8g8 e1g1 f8b8 b7e4 d6d4 e4e7 b6d5 c3d5 d4d5 a1a7 a8a7 e7a7 d5b3 a7a3 b3e6 a3a5 e6e5 a5e5 g7e5 g1h1 e5b2 c1b2 b8b2');
INSERT INTO evaluations VALUES(7454,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(7454,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(7454,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(7454,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(7454,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7454,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7454,36,'cp 0','e8g8 d2d3 d7d6 c3d5 f6d5 b3d5 d8e8 c2c3 a8b8 a2a4 c6e7 d5b3 e7g6 a4b5 a6b5 f3g5 h7h6 d1h5 g6h8 g5f3 c8e6 b3e6 e8e6 f3h4 h8g6 h4f5 g6e7 f5e7 e6e7 h5g4 g8h7 g4f5 h7g8');
INSERT INTO evaluations VALUES(2783,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2783,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2783,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2783,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2783,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2783,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2783,36,'cp 0','d2d3 d7d6 c3d5 f6d5 b3d5 d8e8 a2a4 b5a4 c1e3 c5e3 f2e3 a8b8 f3g5 c6d8 d1h5 h7h6 f1f6 h6g5 f6h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g5 g8h7 g5h5 h7g7 h5g5');
INSERT INTO evaluations VALUES(2784,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(2784,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(2784,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2784,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2784,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2784,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2784,36,'cp 6','d7d6 c3d5 f6d5 b3d5 d8e8 c2c3 a8b8 h2h3 c5b6 a2a4 c6e7 d5b3 e7g6 f1e1 c8e6 d3d4 h7h6 a4b5 a6b5 c1e3 e6b3 d1b3 e8e7 b3c2 e7e6 a1a6 b8a8 e1a1 a8a6 a1a6 e6c4 a6a1 b5b4 g2g3 b6d4 f3d4 e5d4 e3d4');
INSERT INTO evaluations VALUES(2755,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2755,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(2755,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(2755,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(2755,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2755,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2755,36,'cp 0','c3d5 h7h6 c2c3 a8b8 d3d4 c5a7 c1e3 f6d5 b3d5 c6e7 d5b3 e7g6 d4e5 a7e3 f2e3 b8b6 e5d6 b6d6 d1c2 d8e7 a1d1 d6d1 f1d1 c8g4 d1f1 g4f3 g2f3 e7g5 g1f2 g5h4 f2g2 h4g5');
INSERT INTO evaluations VALUES(17710,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(17710,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(17710,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(17710,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(17710,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(17710,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(17710,36,'cp 34','c8f5 g1f3 e7e6 f1d3 f5d3 d1d3 f8d6 f4g3 c6e7 c3c4 g8f6 e1g1 e8g8 b1d2 c7c6 g3d6 d8d6 c4c5 d6c7 d3c2 e7g6 h2h3 c7e7 f1e1 a6a5 a2a3 f8e8 b2b4 a5b4 a3b4 h7h6 e1b1 e6e5 d4e5 g6e5 f3d4 e5d7 d2b3 f6e4 a1a8 e8a8');
INSERT INTO evaluations VALUES(25644,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(25644,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(25644,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25644,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(25644,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(25644,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(25644,36,'cp 28','f1d3 f5d3 d1d3 e7e6 b1d2 f8d6 f4g3 g8f6 g1f3 c6e7 c3c4 e8g8 e1g1 c7c6 g3d6 d8d6 a2a3 e7g6 d3c2 f8e8 b2b4 d6e7 d2b3 h7h6 f1c1 e6e5 c4d5 f6d5 f3e5 g6e5 d4e5 e7e5 b3c5 e5e7 g2g3 g8h8 c2b3 a6a5');
INSERT INTO evaluations VALUES(25646,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(25646,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(25646,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(25646,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(25646,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25646,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(25646,36,'cp 35','f5d3 d1d3 e7e6 b1d2 f8d6 f4g3 g8f6 g1f3 c6e7 c3c4 e8g8 e1g1 c7c6 a2a3 d6g3 h2g3 e7f5 c4c5 f8e8 b2b4 f5e7 d3c3 e7g6 a3a4 d8e7 f1e1 a8d8 c3b3 e7c7 b3b2 e6e5 d4e5 g6e5');
INSERT INTO evaluations VALUES(30695,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(30695,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(30695,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(30695,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(30695,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(30695,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(30695,36,'cp 29','g2g3 c7c5 c2c4 d7d6 f1g2 f8e7 e1g1 g8f6 b2b3 e8g8 b1c3 f8e8 c1b2 h7h6 f3e1 b8c6 e1c2 a7a6 d3d4 c5d4 c2d4 d8b8 f1e1 c6d4 d1d4 b7c6 a2a4 f6d7 e1d1 e7f8 d4e3 b8b7 b3b4 a8c8 a1c1 b7b8 e3e2');
INSERT INTO evaluations VALUES(30696,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(30696,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(30696,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30696,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(30696,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30696,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(30696,36,'cp 21','c7c5 c2c4 f8e7 f1g2 d7d6 e1g1 g8f6 b2b3 f6d7 c1b2 e8g8 b1c3 d7e5 d3d4 e5f3 g2f3 b8c6 c3e2 e6e5 d4d5 c6d4 e2d4 c5d4 f3g2 b7c8 a2a4 a7a6 f2f4 c8d7 b2a3 a8b8 d1e2 f7f6 a3b4 f8f7 b4d2 a6a5');
INSERT INTO evaluations VALUES(40107,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40107,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40107,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(40107,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(40107,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40107,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40107,36,'cp 13','e2e3 c8f5 f3h4 f5e6 f1d3 f6e4 g5e7 d8e7 h4f3 e4d6 e1g1 e8g8 f1e1 b8d7 a1c1 g7g6 d1c2 d7b6 h2h3 f8e8 b2b4 a7a5 b4b5 c6b5 c3b5 d6b5 d3b5 e8c8 c2b1 c8c1 e1c1 a8c8 c1c8 b6c8 b5d3 e7b4 b1b4 a5b4 g1h1 c8d6 f3d2 e6f5 d3f5 d6f5');
INSERT INTO evaluations VALUES(40106,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(40106,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40106,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(40106,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40106,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40106,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40106,36,'cp 0','c8f5 f3h4 f5e6 f1d3 f6e4 g5e7 d8e7 h4f3 e4d6 e1g1 b8d7 f1e1 e8g8 a1c1 h7h6 d1c2 e7f6 b2b4 a7a6 f3d2 f8e8 a2a4 f6g5 d2f3 g5f6');
INSERT INTO evaluations VALUES(28346,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28346,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28346,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28346,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28346,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28346,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(28346,36,'cp 0','f3h4 f5e6 f1d3 f6e4 g5e7 d8e7 h4f3 e4d6 e1g1 b8d7 b2b4 e8g8 d1b1 h7h6 b4b5 c6b5 c3b5 a7a6 b5d6 e7d6 a2a4 d7b6 f1c1 b6c4 f3e5 f8c8 b1b7 a8b8 b7a7 b8a8');
INSERT INTO evaluations VALUES(56101,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56101,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(56101,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(56101,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(56101,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(56101,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(56101,36,'cp 0','c7c6 b5d3 g8f6 g1f3 f8d6 d1e2 d6e7 c2c3 e8g8 e1g1 c8g4 c1f4 f8e8 e2d1 e7d6 f4d6 d8d6 b1d2 b8d7 d1c2 g7g6 f1e1 g8g7 h2h3 g4f3 d2f3 d6f4 f3d2 f6h5 d2f3');
INSERT INTO evaluations VALUES(53674,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53674,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53674,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53674,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53674,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53674,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53674,36,'cp 0','b5d3 g8f6 g1f3 f8d6 d1e2 d6e7 c2c3 e8g8 e1g1 c8g4 c1f4 f8e8 e2d1 b8d7 b1d2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(53723,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53723,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53723,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53723,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53723,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53723,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53723,36,'cp 0','g8f6 g1f3 f8d6 d1e2 d6e7 c2c3 c8g4 e1g1 e8g8 c1f4 f8e8 e2d1 b8d7 b1d2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(56736,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(56736,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(56736,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(56736,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(56736,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(56736,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(56736,36,'cp 156','f8e7 b1c3 b8c6 a2a4 f5f4 a4a5 c8g4 a5a6 b7b6 c3d5 g8f6 c2c3 c6b8 h2h3 g4h5 g2g4 f4g3 d5f6 e7f6 f2g3 b6b5 c4b5 c7c6 b5a4 d8b6 d1e2 e8g8 g3g4 h5f7 c1e3 b6a6 e1g1 b8d7 g1h1 a6b7 g4g5 f6e7 f3h4 g7g6');
INSERT INTO evaluations VALUES(14729,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(14729,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(14729,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(14729,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(14729,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(14729,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(14729,36,'cp 13','f8d6 f4g3 e8g8 b1d2 c6e7 b5d3 b7b6 e3e4 d5e4 d2e4 e7f5 d1e2 c8b7 e1c1 b7e4 d3e4 f6e4 e2e4 d8d7 g3e5 a8e8 h1e1 f7f6 g2g4 d7b5 g4f5 e6f5 e4f4 f6e5 d4e5 d6c5 f4h4 c5e7 h4g3 e7c5 c1b1 f5f4 g3g4 c5f2 d1d7');
INSERT INTO evaluations VALUES(18418,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18418,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18418,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(18418,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(18418,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(18418,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18418,36,'cp 8','f8g7 f1c4 d8c7 e1g1 g8f6 f1e1 d7d6 c1f4 f6h5 f4e3 e8g8 d1d2 c8e6 c4e6 f7e6 a1b1 c6c5 d2e2 a8b8 e2c4 c7c8 a2a3 h5f4 g2g3 f4h5 g1g2 h7h6 f2f3 g6g5 b2b3 f8f7 h2h3 g7e5 c3e2 h5g7 c4d3 c8b7 d3d2');
INSERT INTO evaluations VALUES(19934,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(19934,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(19934,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(19934,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(19934,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(19934,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(19934,36,'cp -68','g1f3 g7g6 f1b5 c8d7 d1e2 f8g7 e1g1 d6e5 f4e5 g8h6 d2d3 e8g8 b1a3 a7a6 b5c6 d7c6 c2c3 h6f5 c1f4 b7b5 a3c2 d8d5 c2e3 f5e3 f4e3 a8c8 a2a3 a6a5 d3d4 c5d4 e3d4 b5b4 a3b4 a5b4 e2e3 b4c3 b2c3 c8a8 d4c5 f8e8 a1a8');
INSERT INTO evaluations VALUES(19933,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(19933,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(19933,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(19933,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(19933,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(19933,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(19933,36,'cp -50','g7g6 f1b5 c8d7 d1e2 g8h6 e5d6 h6f5 e1g1 a8c8 d6e7 f8e7 b1c3 e8g8 c3d5 f8e8 c2c3 f5d6 b5d3 e7f8 e2f2 c5c4 d3c2 d7f5 c2f5 d6f5 d5e3 d8d3 g2g4 f5e3 d2e3 f8c5 f3e5 c6e5 f4e5 e8e5 f2f7 g8h8 f7f2');
INSERT INTO evaluations VALUES(42457,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42457,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42457,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42457,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42457,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42457,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42457,36,'cp 0','a2a3 b4c3 b2c3 c7c5 e2e3 e8g8 c1b2 d7d5 c3c4 b8c6 d4c5 f6e4 c4d5 e6d5 f1d3 d8a5 f3d2 e4c5 d3h7 g8h7 d1h5 h7g8 b2g7 g8g7 h5g5 g7h7 g5h5');
INSERT INTO evaluations VALUES(48363,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(48363,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(48363,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(48363,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(48363,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(48363,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(48363,36,'cp -30','c7c5 c1g5 c5d4 d1d4 f8e7 e1c1 h7h6 g5h4 b8c6 d4d3 f6g4 h4g3 e7d6 h2h3 d6g3 h3g4 g3f2 e2e4 d5d4 d3e2 f2g3 c1b1 c8d7 c3b5 a7a6 b5d4 d8c7 g4g5 e8c8 g5h6 g7h6 e2e3 g3f4 e3f2 c8b8 f1e2 h6h5 d4c6 d7c6 d1d8 h8d8 e2d3 d8h8 f2h4 h8g8 h4h5 g8g2 f3d4 c6a4 b1a1 b8a8');
INSERT INTO evaluations VALUES(46256,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(46256,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(46256,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(46256,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(46256,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(46256,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(46256,36,'cp -21','d4c5 f8c5 e2e3 b8c6 b2b4 c5d6 c3b5 d6e7 b5d4 d8b6 c2c4 c6d4 d1d4 b6d4 f3d4 a7a5 b4a5 a8a5 c1b2 e8g8 c4d5 f6d5 d4b3 a5a7 f1b5 d5c7 b5e2 c8d7 e1g1 f8a8 e2f3 c7d5 h2h3 b7b6 f1c1 a7a4 f3d5');
INSERT INTO evaluations VALUES(55284,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(55284,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(55284,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(55284,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(55284,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(55284,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(55284,36,'cp -40','e4d3 f1d3 b8c6 d1e2 c6d4 e2e5 d8d6 c1f4 d4e6 e1c1 c7c6 d3c4 d6e5 f4e5 g8f6 g1f3 e6c5 f3g5 e7e6 h1e1 f8e7 f2f4 b7b5 c4e2 e8g8 e2f3 f6d5 e5d4 c5b7 f3d5 c6d5 c3b5 c8d7 b5c3 f8b8 g5f3 f7f6 c3d5 e7d6 d5e3 d6f4');
INSERT INTO evaluations VALUES(55309,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(55309,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(55309,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(55309,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55309,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(55309,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(55309,36,'cp -49','f1d3 b8c6 d1e2 e7e5 d3b5 f8d6 f2f4 d8h4 g2g3 h4e7 f4e5 e7e5 c1f4 e5e2 g1e2 c8g4 f4d6 c7d6 e1c1 e8c8 h2h3 g4e2 c3e2 g8f6 e2d4 c8c7 h1f1 c6d4 d1d4 a7a6 b5a4 b7b5 a4b3 d8d7 g3g4 h8e8 h3h4 h7h6 g4g5 h6g5 h4g5 f6e4 f1f7 e4g5 f7d7 c7d7 d4g4 e8e5 a2a4 d6d5 a4b5');
INSERT INTO evaluations VALUES(55306,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(55306,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(55306,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(55306,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(55306,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(55306,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(55306,36,'cp -50','b8c6 d1e2 e7e5 d3b5 f8d6 f2f4 d8h4 g2g3 h4e7 f4e5 e7e5 c1f4 e5e2 g1e2 c8g4 f4d6 c7d6 h2h3 g4e2 c3e2 e8c8 e1c1 g8f6 e2d4 c8c7 h1f1 c6d4 d1d4 a7a6 b5a4 b7b5 a4b3 d8d7 a2a4 h8e8 a4b5 a6b5 f1f5 e8e1 c1d2 e1h1 f5g5 d7e7 g5g7 f6e4 d4e4 e7e4 b3f7');
INSERT INTO evaluations VALUES(16997,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(16997,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(16997,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(16997,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(16997,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(16997,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(16997,36,'cp -116','f3d1 d7d5 e4d5 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 g1f3 f8d6 e1g1 e8g8 b1d2 f7f5 b3b4 a5a4 d2b3 f5f4 f1e1 c8g4 h2h3 g4h5 b3d2 f8e8 d2e4 d5f6 b2b3 f6e4 d3e4 a4b3 a1a8 d8a8 d1b3 h5f7 b3b1 h7h6 c1b2 a8a4 e1d1 f7c4 d1c1 e8a8 f3d2 g8h7 c1e1 a8d8 d2c4 b5c4 b1d1 a4a2 d1c2');
INSERT INTO evaluations VALUES(16998,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(16998,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(16998,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(16998,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(16998,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(16998,35,'cp -149',NULL);
INSERT INTO evaluations VALUES(16998,36,'cp -142','d7d5 e4d5 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 g1f3 f8d6 b1d2 e8g8 e1g1 f8e8 d2e4 d6f8 d3d4 e5d4 d1d4 c8f5 e4g3 c7c5 d4d1 f5g6 f1e1 d5b6 d1d8 e8d8 c1f4 b6d5 f4d2 d5c7 e1e2 f7f6 c3c4 a5a4 b3a4 b5a4 d2c3 g6f7 e2e4 h7h6 a1e1');
INSERT INTO evaluations VALUES(15825,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(15825,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(15825,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(15825,33,'cp -144',NULL);
INSERT INTO evaluations VALUES(15825,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(15825,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(15825,36,'cp -140','e4d5 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 g1f3 f8d6 e1g1 e8g8 b1d2 f8e8 d2e4 d6f8 d3d4 e5d4 d1d4 c8f5 e4g3 c7c5 d4d1 f5g6 f1e1 e8e1 f3e1 h7h6 h2h3 d8d6 e1f3 d5b6 d1e2 d6d3 e2d3 g6d3 c1e3 d3c2 f3d2 f7f5 g3f1');
INSERT INTO evaluations VALUES(15813,30,'cp -136',NULL);
INSERT INTO evaluations VALUES(15813,31,'cp -146',NULL);
INSERT INTO evaluations VALUES(15813,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(15813,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(15813,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(15813,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(15813,36,'cp -137','b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f6d5 g1f3 f8d6 e1g1 e8g8 b1d2 f8e8 d2e4 d6f8 d3d4 e5d4 d1d4 c8f5 e4g3 c7c5 d4d1 f5g6 h2h3 d5e7 c1d2 e7c6 d1c1 h7h6 f1d1 d8b6 d2f4 e8d8 d1d8 a8d8 h3h4 g8h8 f3e5 c6e5 f4e5 f7f6 e5f4 b6c6 c3c4');
INSERT INTO evaluations VALUES(46153,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(46153,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46153,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(46153,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(46153,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(46153,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(46153,36,'cp 18','e4d5 e6d5 c4b5 b8c6 e1g1 c8e6 d2d4 f8e7 d4c5 e7c5 c1g5 e8g8 g5h4 h7h6 f1e1 a7a6 b5c6 b7c6 h2h3 g7g5 h4g3 f8e8 f3d4 e6d7 d1d3 c5f8 d4f3 d7e6 a1d1 f6h5 g3e5 f7f6 d3g6 f8g7 g6h5 e6f7 e5f6 d8f6 h5g4 f7g6 e1e8 a8e8 f3d4');
INSERT INTO evaluations VALUES(46128,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(46128,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(46128,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(46128,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46128,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(46128,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(46128,36,'cp 6','e6d5 c4b5 c8d7 d2d4 c5d4 b5d7 d8d7 f3d4 b8c6 e1g1 f8e7 c1e3 e8g8 d1d3 a8d8 a1d1 f8e8 d3f5 d7f5 d4f5 e7b4 c3b5 a7a6 b5d4 f6g4 e3f4 c6d4 f5d4 b4c5 c2c3 f7f6 f4c7 d8d7 c7g3 g4e5 f1e1 g8f7 g1f1 c5a7 b2b3');
INSERT INTO evaluations VALUES(46413,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(46413,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46413,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(46413,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(46413,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(46413,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(46413,36,'cp 12','c4b5 b8c6 d2d4 c5d4 f3d4 c8d7 e1g1 f8e7 b5e2 e8g8 e2f3 d7e6 c1e3 d8d7 h2h3 f8e8 c3e2 e7d6 e2f4 c6d4 d1d4 a8c8 c2c3 d6c5 d4d2 c5e3 f2e3 d7d6 a1d1 h7h6 f4d5 e6d5 f3d5 f6d5 d2d5 d6d5');
INSERT INTO evaluations VALUES(46392,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(46392,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46392,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46392,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(46392,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(46392,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46392,36,'cp 0','b8c6 d2d4 c5d4 f3d4 c8d7 e1g1 f8e7 c1e3 e8g8 b5e2 d7e6 e2f3 a8c8 c3b5 f8e8 f1e1 e6d7 a2a3 c6d4 b5d4 d8b6 d4b3 b6c7 b3d4');
INSERT INTO evaluations VALUES(13714,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13714,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(13714,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(13714,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13714,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(13714,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(13714,36,'cp 3','e1g1 f8e7 d2d4 c5d4 f3d4 c8d7 c1e3 e8g8 b5e2 e7b4 e2f3 b4c3 d4c6 d7c6 b2c3 h7h6 f3e2 f6e4 e2d3 f8e8 e3d4 d8g5 f2f3 e4d6 d1c1 g5c1 f1c1 c6b5 a2a4 b5d3 c2d3 e8e2 a4a5 f7f6 c1b1 e2d2');
INSERT INTO evaluations VALUES(51029,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(51029,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(51029,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(51029,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(51029,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(51029,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51029,36,'cp 40','g1f3 d7d5 c1g5 f8e7 g5e7 g8e7 e2e3 e8g8 d1c2 b8d7 b1c3 c7c5 d4c5 d8a5 c4d5 e7d5 a1c1 a5c5 f1e2 d7f6 e1g1 c8d7 f3e5 f8c8 e5d7 f6d7 c2b1 d5f6 e2f3 a8b8 c3e4 c5b4 e4f6 g7f6 f1d1 d7e5 d1d4 e5f3 g2f3 c8c1 b1c1');
INSERT INTO evaluations VALUES(51028,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(51028,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(51028,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(51028,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(51028,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(51028,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(51028,36,'cp 28','d7d5 c1g5 f8e7 g5e7 d8e7 b1d2 g8f6 e2e3 b8d7 a1c1 c7c6 f1d3 e8g8 e1g1 e6e5 f3e5 d7e5 d4e5 e7e5 d1c2 h7h6 c4d5 f6d5 a2a3 d5f6 h2h3 c8e6 d2f3 e5c7 f3d4 f8e8 d4e6 e8e6 f1d1 a8d8 b2b4 c7e7 d3e2');
INSERT INTO evaluations VALUES(50950,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50950,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(50950,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(50950,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50950,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(50950,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50950,36,'cp 34','c4d5 e6d5 b1c3 c7c6 e2e4 d5e4 f3g5 h7h6 g5e4 c8e6 f1e2 g8f6 e4f6 d8f6 e1g1 f8d6 d4d5 c6d5 c3d5 e6d5 d1d5 b8c6 d5e4 f6e5 e4e5 d6e5 a1b1 e8c8 g2g3 f7f6 g1g2 g7g5 c1e3 e5d4 e3d2 c8b8 f1d1 h6h5 e2f3 h5h4 f3e4 h4h3 g2f1 d4a7 e4c6');
INSERT INTO evaluations VALUES(19273,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(19273,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19273,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19273,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19273,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19273,36,'cp 0','f8g7 g1e2 d7d6 d2d3 e7e6 c1e3 g8e7 h2h4 b7b5 a2a3 a8b8 h4h5 c6d4 e3c1 d4c6');
INSERT INTO evaluations VALUES(22449,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(22449,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(22449,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(22449,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(22449,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(22449,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(22449,36,'cp -94','h5f3 f7f5 b1c3 g8f6 c3d5 c6a5 c4e2 e5e4 f3g3 f8d6 d5f6 d8f6 f2f4 e4f3 g3f3 a5c6 d2d4 b7b6 f3f2 c8b7 e2f3 e8c8 c1d2 f6f7 g1e2 h8e8 e1g1 d6f8 b2b3 d7d5 c2c4 c6b4 d2b4 f8b4 f3d5 b7d5 c4d5 f7d5 e2f4 d5f7');
INSERT INTO evaluations VALUES(22463,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(22463,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(22463,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(22463,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(22463,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(22463,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(22463,36,'cp -94','f7f5 b1c3 g8f6 c3d5 c6a5 c4e2 e5e4 f3g3 f8d6 d5f6 d8f6 f2f4 e4f3 g3f3 a5c6 d2d4 b7b6 e2c4 c8b7 c4d5 e8c8 c1d2 b7a6 e1c1 c8b8 g1e2 c6e7 d5b3 a6b7 d4d5 h8f8 d2c3 f6h4 c3g7 f8f7 g7d4 e7g8 a2a4 g8f6 a4a5 f6g4 e2c3 g4f2 a5b6 a7b6');
INSERT INTO evaluations VALUES(24890,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(24890,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(24890,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(24890,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24890,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(24890,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(24890,36,'cp 38','e7e6 h2h3 g4h5 c4d5 e6d5 d1a4 d8d7 c1d2 a7a6 a2a3 f8d6 f3h4 h5g6 h4g6 h7g6 g2g4 c6e7 a4d7 f6d7 e1c1 e8c8 h3h4 f7f5 f1e2 d7f6 f2f3 d6g3 d2e1 g3e1 d1e1 f5g4 f3g4 c8b8 c1d2 c7c6 e1g1 f6e4 c3e4 d5e4 g1f1 d8f8 f1f8 h8f8 h1f1 f8f1 e2f1');
INSERT INTO evaluations VALUES(25194,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(25194,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(25194,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(25194,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(25194,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25194,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(25194,36,'cp 35','h2h3 g4f3 d1f3 f8b4 c1d2 e8g8 c4d5 e6d5 f1d3 f8e8 e1g1 c6e7 a2a3 b4c3 b2c3 c7c5 d3b5 e7c6 d4c5 c6e5 f3f4 e8e7 f1d1 a8c8 d2e1 a7a6 b5f1 c8c5 c3c4 e5c4 f1c4 e7e4 c4d5 c5d5 d1d5 d8d5 f4f3 e4e8 f3d5 f6d5 a1d1 d5f6 d1d4');
INSERT INTO evaluations VALUES(31863,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31863,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(31863,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(31863,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(31863,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31863,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(31863,36,'cp 28','c2c4 e7e6 b1c3 f8b4 f1g2 g8f6 g1f3 e8g8 e1g1 b8c6 f3h4 d5c4 h4f5 e6f5 c1g5 d8d4 g2c6 b4c3 b2c3 d4d1 f1d1 b7c6 g5f6 g7f6 d1d7 f8c8 a1d1 a8b8 g1g2 a7a5 d7e7 c8d8 d1d8 b8d8 e7c7 d8d2 a2a4 d2c2 c7c6 c2c3 c6f6 c3a3 f6f5 a3a4 f5c5 g8g7 h2h4 a4b4 c5a5 c4c3');
INSERT INTO evaluations VALUES(46199,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46199,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(46199,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46199,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46199,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46199,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46199,36,'cp 6','e2e3 d8b6 d1c2 f8d6 f4g3 e8g8 b1d2 h7h6 f1d3 b8d7 e1g1 d6g3 h2g3 b6c7 c3c4 c5d4 f3d4 d5c4 c2c4 c7b6 c4c2 d7e5 d3e2 f8d8 d2e4 e5g4 e4f6 g4f6 f1d1 c8d7 c2b3 e6e5 b3b6 a7b6 d4b5 d7b5 e2b5 d8d1 a1d1 a8a2');
INSERT INTO evaluations VALUES(5560,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(5560,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(5560,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(5560,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(5560,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(5560,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(5560,36,'cp -69','g8e7 e1g1 e7g6 b1d2 e8g8 d2b3 c5b6 d1e2 f7f6 e5f6 d8f6 c2c4 a7a6 b5c6 b7c6 c1e3 b6c7 a1d1 h7h6 b3c5 f8d8 e2c2 g6e5 f3d4 e5g4 h2h3 f6e5 d4f3 g4e3 c2e2 e5f6 f2e3 f6e7 b2b4 a6a5 a2a3 a8b8 c5d3 e6e5 c4d5 c6d5 d3e5 c7e5 f3e5');
INSERT INTO evaluations VALUES(8219,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(8219,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(8219,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(8219,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(8219,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(8219,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(8219,36,'cp -89','e1g1 e7g6 b1d2 c5b6 d2b3 e8g8 d1e2 f7f6 e5f6 d8f6 c2c4 a7a6 c1g5 f6f5 b5c6 b7c6 g5e3 b6c7 a1d1 f5h5 f1e1 f8f3 e2f3 h5h2 g1f1 a6a5 f3h3 h2h3 g2h3 g6h4 f2f4 d5c4 b3c5 e6e5 c5d7 c4c3 b2c3 h4f3 f4e5 f3e1 f1e1 c8d7 d1d7 c7e5 a2a4 e5c3 e1e2');
INSERT INTO evaluations VALUES(47993,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(47993,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(47993,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(47993,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(47993,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(47993,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(47993,36,'cp 183','e5f7 e8f7 e2e4 c7c6 d2d4 d8f6 d4d5 c8f5 e4a4 f8e7 b1c3 b8d7 d5c6 d7b6 a4b3 f5e6 c3e4 e6b3 e4f6 b3c2 c6b7 a8b8 f6d5 b6d5 f1c4 f7f6 c4d5 f6e5 d5f3 c2e4 f3e4 e5e4 c1e3 b8b7 e1c1 e7f6 b2b3 a7a5 h1e1 a5a4 e3g5 e4f5 g5f6 g7f6 c1b2 a4b3 a2b3 b7b6 e1e7 h8b8 d1d5 f5g6');
INSERT INTO evaluations VALUES(47557,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(47557,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(47557,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(47557,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(47557,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(47557,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(47557,36,'cp 190','e8f7 e2e4 c7c6 d2d4 d8f6 d4d5 c8f5 e4a4 b8d7 b1c3 f8e7 d5c6 b7c6 a4c6 a8c8 c6d5 f5e6 d5d1 c8c3 b2c3 f6c3 c1d2 c3c6 d1f3 c6f3 g2f3 h8c8 e1d1 e6d5 f1g2 d7b6 a1b1 e7f6 h1g1 d5a2 b1b5 a2e6 f3f4 c8c7 d2e3 e6c4 b5a5 f6c3 a5a3 c3b4 a3a1');
INSERT INTO evaluations VALUES(35491,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(35491,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(35491,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(35491,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(35491,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(35491,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(35491,36,'cp 171','e2e4 c7c6 d2d4 d8f6 d4d5 c8f5 e4a4 b8d7 b1c3 f8e7 c1e3 b7b5 a4f4 g7g5 f4d4 c6c5 d4f6 d7f6 f1b5 f5c2 e3g5 a8g8 g5f4 g8g2 f4g3 h8b8 b5f1 g2g3 h2g3 b8b2 f1e2 f7g8 h1h4 f6d7 h4g4 g8h8 g4f4 c2g6 c3d1 b2b7 d1e3 e7f6 a1c1');
INSERT INTO evaluations VALUES(35490,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(35490,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(35490,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(35490,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(35490,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(35490,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(35490,36,'cp 168','c7c6 d2d4 d8f6 d4d5 c8f5 e4a4 f8e7 b1c3 b8d7 c1e3 b7b5 a4f4 g7g5 f4d4 c6c5 d4f6 d7f6 f1b5 f5c2 e1g1 h7h6 b5e2 a8b8 b2b3 b8b4 e2c4 h8b8 a1c1 b4c4 b3c4 c2d3 f1d1 d3c4 f2f3 b8b4 g2g4 b4b2 d1d2 b2d2 e3d2 f6d5');
INSERT INTO evaluations VALUES(4126,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(4126,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(4126,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(4126,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(4126,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(4126,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(4126,36,'cp -67','c1g5 c6a5 b1d2 h7h6 g5f6 g7f6 a2a4 a5c4 d2c4 d7d6 g1e2 f6g7 e1g1 d8e8 f3e3 c8e6 c4d2 d6d5 e4d5 e6d5 c3c4 d5e6 e2c3 e8c6 d2f3 c6b6 e3b6 a7b6 f1e1 f7f6 c3d5 f8f7 d3d4 e6d5 c4d5 e5d4 f3d4 a8d8 d4e6 d8d5 a1c1 c7c6 f2f4');
INSERT INTO evaluations VALUES(4119,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(4119,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(4119,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(4119,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(4119,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(4119,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(4119,36,'cp -97','c6a5 b1d2 h7h6 g5f6 g7f6 a2a4 a5c4 d2c4 d7d6 g1e2 f6g7 e1g1 h6h5 c4e3 a7a5 f3g3 h5h4 g3f3 c8e6 h2h3 f7f5 g1h1 c7c6 e4f5 g6f5 f3h5 f5f4 e3g4 d8e7 d3d4 f8f5 g4h6 g7h6 h5h6 e7f6 h6f6 f5f6 f2f3 g8f7 f1e1 f6g6 d4e5 d6e5 a1d1');
INSERT INTO evaluations VALUES(14673,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(14673,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(14673,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(14673,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(14673,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(14673,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(14673,36,'cp 64','f8e7 c1f4 e8g8 a1d1 f6d7 f1e1 e7f6 d4d2 d7e5 c4e2 f6g5 f4e3 e5g4 e2g4 c8g4 f2f3 g4e6 h2h3 f8e8 f3f4 g5h4 e3f2 h4f6 e1e3 f6c3 d2c3 f7f6 c3b4 b7b6 b4d2 e6f7 a2a4 a7a6');
INSERT INTO evaluations VALUES(8458,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(8458,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(8458,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(8458,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(8458,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(8458,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(8458,36,'cp 59','c1f4 e8g8 a1d1 f6d7 b2b3 d7b6 c4d3 c8g4 d3e2 e7f6 d4d3 g4d7 f4e3 d7c6 f2f3 d8e7 a2a4 a7a6 a4a5 b6d7 g2g3 e7e5 c3d5 c6d5 d3d5 e5c3 d5d2 c3d2 e3d2 f8e8 h2h3 c7c6 g1g2 d6d5');
INSERT INTO evaluations VALUES(22387,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(22387,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(22387,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(22387,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(22387,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(22387,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(22387,36,'cp 112','c2c4 c7c5 b1c3 c5d4 f3d4 a7a6 c1e3 g8f6 f1e2 b7b6 g2g4 c8b7 f2f3 h7h6 d1d2 a8c8 e1c1 d7e5 b2b3 d6d5 e4d5 b7d5 c1b1 d5b7 h2h4 f8b4 a2a3 b4c3 d2c3 f6d5 c3d2 d5e3 d2e3 d8c7 h1g1 e5c6 d4c6 b7c6 f3f4 c6b7 b1b2 c7c5 e3g3 c8d8 d1d8 e8d8 g1d1 d8c8 h4h5');
INSERT INTO evaluations VALUES(22382,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(22382,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(22382,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(22382,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(22382,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(22382,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(22382,36,'cp 112','c7c5 d4d5 e6d5 c4d5 g7g6 b1c3 g8f6 f1e2 a7a6 e1g1 f8g7 c1f4 d8e7 a2a4 e8g8 f1e1 f6g4 f4g5 g7f6 g5f6 e7f6 d1d2 d7e5 h2h3 e5f3 g2f3 g4e5 f3f4 e5d7 d2e3 f6g7 e3g3 b7b6 g1h2 f7f5 a1d1 g7f6 e2f1 f5e4 c3e4 f6f4 g3f4');
INSERT INTO evaluations VALUES(37388,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(37388,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(37388,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(37388,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(37388,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37388,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(37388,36,'cp 24','d2d4 c7c6 b1c3 d7d5 d1b3 d8b6 e1g1 f8d8 h2h3 c8f5 f1e1 f5e4 b3b6 a7b6 c4d5 e4f3 g2f3 f6d5 c3d5 c6d5 e2e3 b8c6 e1d1 e7e6 a2a3 d8c8 c1d2 c6a5 d2c3 a5c4 d1c1 b6b5 g1g2 g7f8 e3e4 c4b6 h3h4 h7h5 c1e1 f8g7 e4d5 b6d5 f3d5 e6d5 a1b1 c8c4');
INSERT INTO evaluations VALUES(56682,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(56682,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(56682,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(56682,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(56682,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(56682,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(56682,36,'cp 114','c1e3 f6f5 e4f5 c8f5 f1d3 f5d3 d1d3 f8e7 g1h3 g8h6 e3h6 g7h6 e1g1 b8d7 b1d2 d7f6 g1h1 d8d7 f2f3 e8g8 a1e1 b7b5 d2e4 f6e4 e1e4 g8h8 d3d2 c7c6 c2c3 a8c8 d5c6 d7c6 f1e1 c6c7 f3f4 e7f6 f4e5 f6e5');
INSERT INTO evaluations VALUES(17926,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(17926,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(17926,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(17926,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(17926,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(17926,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(17926,36,'cp 101','b5c4 g8f6 e1g1 d7d6 c2c3 d4c3 b1c3 f8e7 d3d4 e8g8 c4d3 c7c5 d4c5 c8e6 c5d6 d8d6 d3e2 a8c8 c1e3 d6c6 e2f3 f8d8 d1c1 e7c5 h2h3 b7b5 e3c5 c6c5 f1d1 d8d1 c1d1 f6d7 d1d2 d7e5 f3e2');
INSERT INTO evaluations VALUES(17934,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(17934,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(17934,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(17934,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(17934,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(17934,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(17934,36,'cp 101','g8f6 e1g1 d7d6 c2c3 d4c3 b1c3 f8e7 d3d4 e8g8 d1c2 c8e6 c4d3 h7h6 f2f4 b7b5 h2h3 f8e8 c1e3 c7c6 a1d1 e6c4 d3c4 b5c4 f1f3 a8b8 e3f2 d8c8 e4e5 f6d5 c3e4 c8c7 b2b3 f7f5 e4c3 c4b3 a2b3');
INSERT INTO evaluations VALUES(34856,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(34856,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(34856,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(34856,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(34856,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(34856,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(34856,36,'cp 148','g1f3 g8f6 c1d2 a7a6 f1c4 a5h5 c3e2 f8d6 e2g3 d6g3 f2g3 e8g8 e1g1 b7b5 c4e2 h5d5 d2f4 d5d8 c2c3 c8b7 a2a4 b5a4 a1a4 a6a5 d1a1 b8c6 f1e1 h7h6 b2b3 f6h5 f4d2 h5f6 e2d3 c6e7 f3e5 f6d7 e5d7 d8d7 a4a5');
INSERT INTO evaluations VALUES(34853,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(34853,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(34853,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(34853,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(34853,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(34853,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(34853,36,'cp 124','g8f6 c1d2 b8d7 f1d3 f8b4 e1g1 e8g8 f1e1 a5h5 c3e4 b4d2 d1d2 h5h6 d2h6 g7h6 f3e5 f8d8 a1d1 g8g7 c2c3 f6e4 d3e4 a8b8 f2f4 d7e5 f4e5 b7b6 g1f2 c8b7 e4b7 b8b7 f2e3 c7c6 e1f1 c6c5 d4c5 d8d1 f1d1 b6c5 b2b3 g7g6');
INSERT INTO evaluations VALUES(33892,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(33892,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(33892,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(33892,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(33892,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(33892,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(33892,36,'cp 136','c1d2 b8d7 f1d3 f8b4 e1g1 e8g8 f1e1 a5h5 c3e4 b4d2 d1d2 h5h6 d2h6 g7h6 f3e5 f8d8 a1d1 g8g7 c2c3 b7b6 e4g3 c7c5 g3e4 c5d4 e5c6 d8f8 c6d4 d7c5 e4f6 g7f6 d3h7 c8b7 h7c2 f8g8 g2g3 g8d8 f2f4 a7a5 d4b5 b7f3 d1d4 d8d4 b5d4');
INSERT INTO evaluations VALUES(33893,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(33893,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(33893,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(33893,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(33893,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(33893,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(33893,36,'cp 125','b8d7 f1d3 f8b4 e1g1 e8g8 f1e1 a5h5 c3e4 b4d2 d1d2 h5h6 d2h6 g7h6 f3e5 f8d8 a1d1 a7a5 c2c3 f6e4 d3e4 a5a4 e5c4 a8a6 f2f3 d7f6 a2a3 f6e4 f3e4 f7f6 h2h3 c8d7 d1d2 g8g7 e4e5 f6f5 c4e3 d7c6 g1f2');
INSERT INTO evaluations VALUES(49363,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(49363,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(49363,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(49363,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(49363,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(49363,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(49363,36,'cp -37','h5e2 d7d5 e4e5 f6d7 f2f4 c7c5 g1f3 b8c6 c2c3 c5c4 d2d4 c4d3 e2d3 f8c5 b1d2 f7f6 e5f6 d8f6 d2b3 c5d6 f3g5 f6e7 f1e2 d7f6 e1g1 e8g8 g5f3 e7c7 g2g3 c8d7 c1e3 a8e8 a1e1 b7b6 e2d1 e6e5 f4e5 c6e5 f3e5 d6e5');
INSERT INTO evaluations VALUES(52255,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(52255,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(52255,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(52255,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(52255,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(52255,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(52255,36,'cp 85','d7d6 f1d3 g8f6 e1g1 b8d7 c2c3 f8g7 c1e3 e8g8 b1d2 f8e8 h2h3 a7a5 a2a4 b7b6 d1b3 c8a6 f3g5 e7e6 d3a6 a8a6 f2f4 a6a8 a1e1 b6b5 a4b5 a5a4 b3c2 c6b5 e4e5 f6d5 g5e4 d5e3 e1e3 d6e5 f4e5 d8b6 g1h1 e8f8');
INSERT INTO evaluations VALUES(52590,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(52590,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(52590,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(52590,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(52590,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(52590,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(52590,36,'cp 88','f1d3 b8d7 c2c3 e7e5 c1e3 g8f6 e1g1 f8g7 h2h3 e8g8 b1d2 f8e8 f1e1 d8c7 a2a4 d6d5 d4e5 d5e4 e5f6 d7f6 d3e4 f6e4 e3a7 a8a7 d2e4 e8d8 f3d4 a7a5 d1f3 c8f5 e4g3 g7d4 c3d4 a5d5 a1d1 c7b6 f3c3 f5e6 g3e4 d5d4');
INSERT INTO evaluations VALUES(1841,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(1841,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(1841,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(1841,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(1841,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(1841,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(1841,36,'cp 87','g7g5 d2d4 g5g4 f3e5 c6e5 d4e5 f6e5 b5d3 h7h5 f2f4 g4f3 d1f3 e5f6 b1d2 d7d6 e1g1 f6f3 d2f3 c8e6 f3d4 f8h6 h2h3 a8d8 g1h2 h6c1 a1c1 g8e7 d4e6 f7e6 f1f6 e8d7 c1f1 d8g8 d3c4 d6d5 e4d5 e7d5 f6f7 d7d6 f1f2 h5h4 c4e2 d5e3 e2f3');
INSERT INTO evaluations VALUES(15486,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(15486,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(15486,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(15486,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(15486,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(15486,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(15486,36,'cp 24','f2f4 f8e7 g1f3 e5f4 c1f4 e8g8 c4b3 c6a5 e1g1 a5b3 a2b3 a7a5 h2h3 c8d7 d1e2 b7b5 f4h2 c7c5 a1e1 d8b6 g1h1 b6b7 e4e5 d6e5 f3e5 f8e8 e2f3 b7f3 f1f3 b5b4 e5d7 f6d7 c3e4 f7f6 g2g4 e7f8 f3f5 a5a4 b3a4 a8a4');
INSERT INTO evaluations VALUES(15516,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(15516,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(15516,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(15516,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(15516,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(15516,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(15516,36,'cp 20','e5f4 c1f4 f8e7 g1f3 e8g8 e1g1 c6a5 c4b3 c8d7 h2h3 b7b5 d1e2 a5b3 a2b3 a7a5 g1h1 c7c5 f4h2 b5b4 c3d1 a5a4 b3a4 a8a4 a1a4 d7a4 d1e3 a4d7 c2c3 b4c3 b2c3 f8e8 f1b1 d8a8 c3c4 e8b8 e2c2 b8b1 c2b1 d7e6 e3f5');
INSERT INTO evaluations VALUES(44924,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44924,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(44924,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44924,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(44924,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(44924,36,'cp 6','c8f5 d1b3 d8b6 b1d2 b8d7 f1e2 a7a5 a2a4 b6b3 d2b3 h7h5 h2h4 f5c2 b3c1 d7b6 f4c7 f6d7 b2b3 e7e6 e1d2 c2f5 e2d3 f5d3 c1d3 a8c8 c7b6 d7b6 g1f3 f7f6 a1c1 e8f7 c3c4 f8e7 c4c5 b6d7');
INSERT INTO evaluations VALUES(28981,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(28981,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(28981,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(28981,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(28981,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(28981,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(28981,36,'cp 20','d1b3 d8b6 h2h4 b8d7 b1d2 h7h5 f2f3 a7a5 a2a4 e7e6 b3b6 d7b6 g1e2 f8e7 e2c1 e8g8 c1d3 f8c8 f1e2 f5h7 g2g3 b6d7 e1f2 c6c5 d3e5 c5d4 e3d4 d7e5 f4e5 h7g6 e5f4 f6e8 d2f1 e7d6 f4d6 e8d6 f1e3');
INSERT INTO evaluations VALUES(47221,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(47221,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(47221,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(47221,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(47221,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(47221,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(47221,36,'cp -86','d5d4 c3e2 e7e5 d2d3 b8c6 g2g3 h7h6 c1d2 c8e6 f1h3 d8d7 h3e6 d7e6 f3f4 e8c8 g1f3 g7g5 f4f5 e6d7 e1g1 g5g4 f3h4 h6h5 a1b1 a7a5 d2g5 f8e7 g1h1 a5a4 a2a3 f6e8 f5f6 e7f8');
INSERT INTO evaluations VALUES(47386,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(47386,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(47386,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(47386,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(47386,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(47386,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(47386,36,'cp -108','c3e2 e7e5 d2d3 f8b4 c1d2 b4d2 d1d2 c7c5 d2g5 d8a5 e1d1 e8g8 g5e5 a5b6 b2b3 a7a5 a2a4 b8c6 e5g5 c6b4 g5d2 c8d7 g1h3 b6c7 d1e1 d7h3 g2h3 f6d7 f3f4 f7f5 e4e5 b4d5 f1g2 d5e3 e1f2 a8e8 h3h4 g8h8 f2g1 d7e5 f4e5 c7e5 g2f3');
INSERT INTO evaluations VALUES(53875,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(53875,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53875,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(53875,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(53875,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(53875,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(53875,36,'cp 31','f8e7 a2a3 g8f6 b2b4 e8g8 e1g1 b8c6 b4c5 d6c5 a3a4 h7h6 b1c3 e7d6 c3b5 d6e7 c1b2 a7a6 b5c3 e7d6 c3d5 f6d5 e4d5 c6d4 f3d2 a8b8 a4a5 d6c7 b2a3 c7a5 a3c5 f8e8 d2e4 b7b5 c4a2 a5b6 c2c3 d4f5 c5a3 a6a5');
INSERT INTO evaluations VALUES(37902,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(37902,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37902,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(37902,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(37902,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(37902,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(37902,36,'cp 39','b1c3 g8f6 c1g5 e8g8 g5f6 e7f6 c3d5 b8c6 c2c3 a7a6 a2a4 a8b8 h2h4 b7b5 a4b5 a6b5 c4a2 h7h6 d5e3 c6e7 g2g3 c8e6 a2e6 f7e6 e1g1 d8d7 d1e2 e7c6 a1a6 f8f7 f3h2 b8f8 a6b6 b5b4 g1g2 f6e7 b6a6');
INSERT INTO evaluations VALUES(54985,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(54985,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(54985,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54985,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54985,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54985,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54985,36,'cp 0','g8f6 d2d3 d6d5 g2g3 e7e6 b1c3 f8e7 e2e4 d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 e8g8 c1e3 b7b6 f1d3 c8b7 d1e2 b8d7 h1d1 d7c5 e3c5 e7c5 d1d2 g7g6 a1d1 a7a5 c2c3 f8d8 h2h4 g8g7 d3c2 b7a6 c2d3 a6b7');
INSERT INTO evaluations VALUES(47053,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(47053,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47053,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47053,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47053,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47053,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47053,36,'cp 0','d2d3 c7c5 e2e4 b8c6 a2a4 e7e6 g2g3 d6d5 d1e2 d5e4 d3e4 c6d4 e2d3 b7b6 b1c3 c8b7 f1g2 d4c6 d3e2 c6d4');
INSERT INTO evaluations VALUES(56652,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(56652,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(56652,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(56652,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(56652,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(56652,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(56652,36,'cp 22','f1c4 f8d6 g1f3 g8f6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 f1e1 e7d6 c1g5 b8d7 c4b3 h7h6 g5h4 c7c6 b1c3 a7a5 h2h3 g7g5 h4g3 d6g3 f2g3 d7b6 a2a4 c8d7 e1e7 a8e8 a1e1 g8g7 e7e5 b6d5 b3d5 c6d5 c3d5 f6d5 e5d5 e8e1 f3e1 d7a4 d5a5 f8e8 a5a4 e8e1 g1f2 e1b1');
INSERT INTO evaluations VALUES(17854,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(17854,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(17854,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(17854,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(17854,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(17854,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(17854,36,'cp 49','d3d4 e5d4 f3d4 c6e5 c4e2 d7d5 e4d5 d8d5 e1g1 f8c5 d4b3 d5d1 e2d1 c5d6 f1e1 g8e7 b1c3 e8g8 c3e4 f8e8 e4d6 c7d6 c1d2 c8f5 b3d4 f5g6 b2b3 e5c6 d4c6 e7c6 a1c1 e8e1 d2e1 a8e8 g1f1 d6d5 e1d2 g6f5 f2f3 d5d4 h2h4 h6h5 f1f2 c6e5 d1e2 e5g6 d2g5 e8c8');
INSERT INTO evaluations VALUES(26998,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(26998,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(26998,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(26998,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(26998,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(26998,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(26998,36,'cp 164','d1b3 d8d7 b1d2 f7f5 e4f5 e5d4 f1b5 d4c3 b2c3 a7a6 b5c6 d7c6 e1g1 c6b5 b3e6 g8e7 f3d4 b5d5 h2h3 g4h5 c1b2 d5e6 d4e6 e8d7 f1e1 h5f7 g2g4 f7e6 f5e6 d7e8 c3c4 h7h5 g4g5 e7g6 a1d1 b7b5 g1h2 a8d8 d1c1 b5c4 d2c4 g6f4 c1d1 d6d5 h2g3 f4h3 g3h3 d5c4 f2f4 f8b4 d1d8 e8d8 e1d1 d8e7 b2g7');
INSERT INTO evaluations VALUES(42659,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(42659,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(42659,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42659,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42659,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42659,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42659,36,'cp 0','g1f3 d7d6 d2d4 e5d4 f3d4 e8g8 d4b3 c5b6 c1g5 h7h6 g5h4 f8e8 d1e2 c7c6 f2f3 d6d5 e1c1 b8d7 c1b1 a7a5 a3a4 d8c7 h4g3 c7d8');
INSERT INTO evaluations VALUES(42657,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42657,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42657,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42657,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42657,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42657,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42657,36,'cp 0','d7d6 d2d4 e5d4 f3d4 e8g8 d4b3 c5b6 c1g5 f8e8 d1e2 h7h6 g5h4 d6d5 e1c1 c7c6 f2f3 b8d7 c1b1 d7f8 e2e1 f8g6 h4f2 c8e6 b3c5 e6c8');
INSERT INTO evaluations VALUES(9080,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(9080,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(9080,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(9080,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(9080,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(9080,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(9080,36,'cp 47','e4d5 c6a5 c4b5 c7c6 d5c6 e8g8 c6b7 c8b7 e1g1 h7h6 g5f3 a7a6 b5e2 e5e4 f3h4 a5c6 d2d3 c6d4 c1e3 f8e8 e3d4 d8d4 d3e4 d4b4 e2d3 f6e4 c3e4 b7e4 d1g4 e4d3 g4b4 c5b4 c2d3 e8e2 h4f3 a8d8 a1b1 b4c5 f1c1 c5d4');
INSERT INTO evaluations VALUES(9067,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(9067,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(9067,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(9067,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(9067,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(9067,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(9067,36,'cp 45','c6a5 c4b5 c7c6 d5c6 e8g8 e1g1 a5c6 b5c6 b7c6 d2d3 h7h6 c3e4 f6e4 g5e4 c5b6 c1e3 f7f5 e3b6 d8b6 e4d2 b6b2 d1b1 b2b1 f1b1 f8e8 f2f3 c8e6 b1b7 a8b8 b7c7 e6d5 d2b3 e5e4 b3d4 e4d3 c2d3 f5f4 d4f5');
INSERT INTO evaluations VALUES(9542,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(9542,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(9542,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(9542,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(9542,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(9542,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(9542,36,'cp 59','d2d3 e8g8 e1g1 a5c4 d3c4 c8f5 f1e1 f8e8 g5e4 f5e4 c3e4 f6e4 e1e4 f7f5 e4e2 c5d6 c1e3 b7b6 f2f3 d8h4 b2b3 h4h5 d1d2 e5e4 f3f4 d6b4 d2d1 a7a5 e3d4 b4d6 g2g3 a5a4 e2e1 h5h3 c2c3 a4b3 a2b3 a8a1 d1a1 h7h6 a1a4 g8f8 b3b4');
INSERT INTO evaluations VALUES(10727,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(10727,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(10727,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(10727,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(10727,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(10727,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(10727,36,'cp 53','e1g1 e8g8 f1e1 h7h6 c3d5 f6d5 e4d5 c6d4 f3d4 e5d4 d2d3 d8f6 c1d2 b7b6 a2a4 c7c6 d1g4 a8b8 c4b3 c6d5 e1e2 a7a5 h2h3 c8b7 g4d7 b7c6 d7g4 b8e8 a1e1 e8e2 e1e2 f8d8 h3h4 g8f8 f2f4 d6c5 g4h5 c6e8 f4f5 e8c6');
INSERT INTO evaluations VALUES(10726,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(10726,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(10726,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(10726,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(10726,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(10726,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(10726,36,'cp 37','e8g8 f1e1 h7h6 c3d5 f6d5 e4d5 c6d4 f3d4 e5d4 d2d3 d8f6 c1d2 b7b6 e1e4 c8b7 d1h5 c7c6 a1e1 c6d5 c4d5 d6h2 g1h2 f6d6 h5e5 d6d5 e5g3 d5h5 e4h4 h5g6 h4d4 g6g3 h2g3 b7c6 c2c4 f8e8 f2f3 e8e1 d2e1 a8e8 g3f2 f7f6 b2b3 e8e6 b3b4');
INSERT INTO evaluations VALUES(18673,30,'cp 320',NULL);
INSERT INTO evaluations VALUES(18673,31,'cp 317',NULL);
INSERT INTO evaluations VALUES(18673,32,'cp 300',NULL);
INSERT INTO evaluations VALUES(18673,33,'cp 312',NULL);
INSERT INTO evaluations VALUES(18673,34,'cp 303',NULL);
INSERT INTO evaluations VALUES(18673,35,'cp 310',NULL);
INSERT INTO evaluations VALUES(18673,36,'cp 322','d4d5 d8f6 e1g1 g8e7 d5c5 e7c6 c5h5 f8b4 c2c3 b4e7 c1e3 e8g8 h5e2 f6g6 b1d2 g8h8 f2f4 f7f5 e4f5 g6f5 c4d3 f5f7 d2f3 d7d6 a1d1 c8f5 d3f5 f7f5 f1e1 a8d8 e2d3 f5f7 d3b5 d8d7 g2g3 a7a6 b5b3 f7b3 a2b3');
INSERT INTO evaluations VALUES(18674,30,'cp 317',NULL);
INSERT INTO evaluations VALUES(18674,31,'cp 324',NULL);
INSERT INTO evaluations VALUES(18674,32,'cp 317',NULL);
INSERT INTO evaluations VALUES(18674,33,'cp 312',NULL);
INSERT INTO evaluations VALUES(18674,34,'cp 311',NULL);
INSERT INTO evaluations VALUES(18674,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(18674,36,'cp 315','d8f6 e1g1 g8e7 d5c5 e7c6 c5h5 f8b4 c2c3 b4e7 f2f4 d7d6 h5e2 c8e6 c4d3 e8g8 c1e3 f6h4 b1d2 h4g4 f1f3 f7f5 e3f2 a8e8 a1e1 e6a2 d2f1 e7h4 f1e3 g4h5 e4f5 d6d5 b2b4 a2b3 g2g4 h4f2 e2f2 h5f7 b4b5 d5d4 e3g2 e8e1 g2e1');
INSERT INTO evaluations VALUES(1320,30,'cp 305',NULL);
INSERT INTO evaluations VALUES(1320,31,'cp 305',NULL);
INSERT INTO evaluations VALUES(1320,32,'cp 300',NULL);
INSERT INTO evaluations VALUES(1320,33,'cp 313',NULL);
INSERT INTO evaluations VALUES(1320,34,'cp 306',NULL);
INSERT INTO evaluations VALUES(1320,35,'cp 313',NULL);
INSERT INTO evaluations VALUES(1320,36,'cp 332','b1c3 f8d6 c1e3 d6e5 e1c1 g8e7 d5d2 b7b6 g2g4 a8b8 h1g1 b6b5 c4b5 e8g8 b5c4 e7g6 e3c5 e5h2 c5a7 c8b7 g1f1 h2f4 a7e3 f4e5 e3d4 e5d4 d2d4 f6d4 d1d4 g6e5 c4b3 e5g4');
INSERT INTO evaluations VALUES(24913,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24913,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(24913,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(24913,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(24913,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(24913,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24913,36,'cp 15','f3e5 c6e5 d4e5 d5d4 c3b1 f6h5 f4c1 g7g6 c2c3 d4c3 d1d8 a8d8 b1c3 f7f6 e5f6 e7f6 e2e4 f8d6 c1e3 a7a6 f2f3 g4e6 g2g3 f6f5 e1c1 h5f6 e4f5 g6f5 f1d3 e8f7 h1e1 h8e8 d3c2 d6b4 d1d8 e8d8 a2a3 b4a5 b2b4 a5b6 e3b6 c7b6 c1b2 d8d2');
INSERT INTO evaluations VALUES(34694,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(34694,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(34694,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(34694,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(34694,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(34694,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(34694,36,'cp 46','c7c6 d2d4 g8f6 f1d3 c8g4 g1f3 g4h5 h2h3 e7e6 d1e2 h5g6 e1g1 g6d3 e2d3 b8d7 f1e1 f8e7 c3e2 b7b5 c1d2 a5c7 c2c4 b5c4 d3c4 e8g8 a1c1 c7b7 b2b4 f8c8 e2f4 f6d5 f4d3 h7h6 d3c5 d7c5 d4c5 e7f6 e1d1 b7b5 c4b5 c6b5');
INSERT INTO evaluations VALUES(34332,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(34332,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(34332,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(34332,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(34332,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(34332,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(34332,36,'cp 54','d2d4 g8f6 f1d3 c8g4 g1f3 g4h5 d1e2 e7e6 h2h3 b8d7 c1d2 h5f3 e2f3 a5b6 f3f4 g7g6 e1c1 f8g7 c3a4 b6d8 f4d6 b7b5 a4c5 d7c5 d6c5 d8b6 h1e1 e8c8 c5b6 a7b6 c2c3 c8b7 c1c2 f6d5 c2b3');
INSERT INTO evaluations VALUES(48395,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48395,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48395,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(48395,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48395,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(48395,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48395,36,'cp 0','g1f3 c7c5 g2g3 b8c6 f1g2 f8e7 e1g1 e8g8 f1e1 h7h6 c2c3 d8c7 d1e2 b7b5 d2f1 d5e4 d3e4 b5b4 e4e5 f6d7 c1f4 a7a5 a1d1 c8a6 e2e4 a6b7 f4h6 g7h6 e4g4 g8h7 g4e4 h7g8');
INSERT INTO evaluations VALUES(48392,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48392,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48392,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48392,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48392,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48392,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(48392,36,'cp -17','f8e7 c2c3 e8g8 e4e5 f6d7 d3d4 c7c5 f1d3 c5d4 c3d4 f7f6 e5f6 e7f6 e1g1 b8c6 d2b3 e6e5 f3e5 d7e5 d4e5 c6e5 h2h3 c8e6 d3c2 e6f7 c1e3 g8h8 a1b1 f8e8 f1e1 e5c6 e3f4 e8e1 d1e1 b7b6 e1d2 f7g8 c2f5 d5d4 b1e1 d8d5');
INSERT INTO evaluations VALUES(19370,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(19370,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(19370,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(19370,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(19370,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(19370,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(19370,36,'cp -40','g8f6 b4b5 c6d4 c2c3 d4e6 e4e5 f6d5 d1b3 d5b6 a3a4 d7d6 a4a5 b6d7 g1f3 g7g5 d2d4 c5d4 c3d4 d6e5 d4d5 e6d4 f3d4 e5d4 f1c4 f8g7 e1g1 d7e5 c1b2 d8d6 b2a3 d6f6 f1e1 e8g8 c4f1 d4d3 b1d2 c8f5 a1d1 f8c8 d2e4 f6g6 e4c5 h7h6 c5d3 e5d3 f1d3 c8c3 d3f5 c3b3 f5g6 b3a3 g6d3');
INSERT INTO evaluations VALUES(19413,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(19413,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(19413,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(19413,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19413,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(19413,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19413,36,'cp 22','c5d4 f3d4 g8f6 b1c3 e7e6 a2a3 f8c5 d4b3 c5b6 c4c5 b6c7 g2g3 e8g8 f1g2 b7b6 e1g1 h7h6 c3b5 c8a6 b5c7 d8c7 c1e3 a8c8 f1e1 f8d8 a1c1 c7b8 c5b6 a7b6 b3d4 c6a5 d1a4 a6b7 f2f3 d8e8 c1c8 e8c8 e1c1 e6e5 c1c8 b8c8');
INSERT INTO evaluations VALUES(19533,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(19533,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19533,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(19533,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(19533,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19533,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(19533,36,'cp 21','f3d4 g8f6 b1c3 e7e6 a2a3 f8c5 d4b3 c5b6 c4c5 b6c7 g2g3 e8g8 f1g2 b7b6 e1g1 h7h6 c3b5 c8a6 b5c7 d8c7 c1e3 a8c8 a1c1 f6g4 e3d4 c6d4 d1d4 g4f6 c5b6 c7b6 d4b6 a7b6 b3d4 f8d8 f1d1 g8f8 e2e4 f6g4 b2b4 c8c1');
INSERT INTO evaluations VALUES(19530,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19530,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19530,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(19530,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(19530,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(19530,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(19530,36,'cp 0','g8f6 b1c3 e7e6 d4c6 d7c6 d1d8 e8d8 c1f4 f6d7 e1c1 f7f6 e2e4 e6e5 f4e3 d8c7 f1e2 h7h5 h2h4 d7c5 f2f4 b7b6 f4f5 a7a5 c1c2 c8d7 g2g4 h5g4 e2g4 d7e8 h4h5 h8h7 d1g1 e8f7 b2b3 a8d8 h1h3 d8c8 h3h1');
INSERT INTO evaluations VALUES(12872,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12872,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(12872,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12872,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(12872,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(12872,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(12872,36,'cp 12','b1c3 e7e6 d4c6 d7c6 d1d8 e8d8 c1f4 f6d7 e1c1 f7f6 g2g3 e6e5 f4e3 d8c7 f1g2 d7b6 b2b3 c8e6 c1b2 f8e7 f2f4 b6d7 d1d2 h8d8 c3d5 c6d5 c4d5 e6g4 h2h3 g4f5 d5d6 e7d6 h1c1 d7c5 e3c5 d6c5 c1c5 c7b6 c5d5 d8d5 d2d5 a8c8 g2f3 c8c2 b2a3 f5h3 f4e5 f6e5 d5e5');
INSERT INTO evaluations VALUES(12870,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12870,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(12870,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(12870,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(12870,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12870,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12870,36,'cp 0','e7e6 d4c6 d7c6 d1d8 e8d8 c1f4 f6d7 e1c1 f7f6 e2e4 e6e5 f4e3 d8c7 f1e2 b7b6 c3a4 f8e7 h2h4 d7f8 g2g3 c8e6 d1d2 e7b4 d2d1');
INSERT INTO evaluations VALUES(12464,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(12464,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12464,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12464,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12464,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(12464,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(12464,36,'cp 19','d4c6 d7c6 d1d8 e8d8 c1f4 f6d7 e1c1 f7f6 g2g3 e6e5 f4e3 d8c7 f1g2 f8e7 h2h4 d7b6 b2b3 c8e6 h4h5 b6d7 g2h3 e6h3 h1h3 f6f5 f2f3 d7f6 c1c2 h8d8 h3h1 d8d1 c3d1 c7d7 d1f2 d7e6 g3g4 b7b6 g4g5 f6d7 h1g1 f5f4 e3d2 a8d8 d2c3 e6f5 h5h6');
INSERT INTO evaluations VALUES(18593,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(18593,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(18593,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(18593,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(18593,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(18593,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(18593,36,'cp 100','c2c4 g8f6 b1c3 g7g6 f1e2 f8g7 e1g1 e8g8 c1e3 f6g4 e3d2 g4f6 d1c2 e7e5 d4d5 a7a5 d2e3 c6d5 c4d5 f6g4 e3d2 d7c5 h2h3 g4f6 d2e3 b7b6 f3d2 c8d7 a1c1 a8c8 a2a4 c5a4 d2c4 a4c3 b2c3 f6e4 c2e4 d7b5');
INSERT INTO evaluations VALUES(18588,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(18588,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(18588,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(18588,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(18588,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(18588,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(18588,36,'cp 102','e7e5 f1e2 g8f6 b1c3 f8e7 d1c2 e8g8 e1g1 f8e8 d4d5 a7a5 h2h3 d7c5 c1e3 d8c7 f3d2 c8d7 f1c1 c6d5 c4d5 e8c8 a2a4 c7d8 c2d1 f6e8 e2b5 e7g5 b5d7 c5d7 e3g5 d8g5 c3b5 g5d8 d2c4 d7b6 d1e2 b6c4 c1c4 c8c4');
INSERT INTO evaluations VALUES(41348,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(41348,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(41348,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(41348,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(41348,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(41348,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41348,36,'cp 39','c2c4 e8g8 b1c3 c7c5 d4d5 e7e6 e1g1 e6d5 c4d5 f8e8 c1f4 c8f5 f3h4 f5c8 d1d2 d8b6 e2e4 b8d7 h2h3 c5c4 f4e3 d7c5 d2e2 c8d7 e2c4 b6b2 e3c5 d6c5 a1b1 b7b5 c3b5 d7b5 c4c5 a8c8 c5b5 b2b5 b1b5 f6e4');
INSERT INTO evaluations VALUES(46187,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(46187,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46187,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(46187,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(46187,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46187,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46187,36,'cp 0','g1f3 d5c4 f1c4 a7a6 c4e2 b8d7 e1g1 b7b5 b2b3 c8b7 c1b2 c5d4 f3d4 f8e7 b1d2 e8g8 a3a4 b5b4 e2f3 f6d5 d2c4 a8c8 a1c1 h7h6 a4a5 c8c5 d1e2 d8c7 c1c2 e7f6 f1c1 f8c8 h2h3 f6e7 c1d1 e7f6');
INSERT INTO evaluations VALUES(46188,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(46188,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(46188,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(46188,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46188,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46188,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46188,36,'cp 0','d5c4 f1c4 a7a6 e1g1 b7b5 c4e2 b8d7 a3a4 b5b4 b2b3 f8e7 b1d2 c8b7 c1b2 e8g8 a1c1 a8c8 d2c4 d8c7 f3e5 f8d8 e5d7 f6d7 e2f3 h7h6 f3b7 c7b7 d1e2 c5d4 b2d4 d7c5 c4a5 b7b6 a5c4');
INSERT INTO evaluations VALUES(52579,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(52579,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(52579,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(52579,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(52579,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(52579,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(52579,36,'cp 118','e4e5 c8f5 g2g4 f5e6 c1e3 h7h5 g4h5 h8h5 f1e2 h5h8 d1d2 b8d7 h2h4 g8h6 g1f3 h6g4 e1c1 d7b6 e2d3 g4e3 f2e3 f8h6 d1g1 d8d7 f3g5 h6g5 g1g5 e8c8 b2b3 h8h6 a2a4 d7c7 d2f2 d8h8 a4a5');
INSERT INTO evaluations VALUES(52581,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(52581,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(52581,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(52581,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(52581,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(52581,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(52581,36,'cp 102','c8f5 g2g4 f5e6 c1e3 h7h5 g4h5 h8h5 f1e2 h5h8 d1d2 g8h6 e1c1 h6g4 g1f3 e6f5 h2h4 g4e3 f2e3 b8d7 e2d3 e7e6 a2a3 f5d3 d2d3 d7b6 d1f1 f8e7 c1b1 b6c4 f3g5 d8b6 c3a4 b6b5 g5f7 h8g8 b2b3 c4a3 b1b2 b5d3 c2d3 b7b6 f7g5 e8d7 f1f7');
INSERT INTO evaluations VALUES(56037,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(56037,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(56037,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(56037,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(56037,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(56037,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(56037,36,'cp 134','e2e4 f8b4 a2a3 b4c3 b2c3 g8e7 g1f3 c7c5 f1c4 e8g8 e1g1 h7h6 f1e1 b7b6 d4d5 e6d5 e4d5 e7g6 h2h3 b8d7 c4f1 d7f6 c3c4 f8e8 a3a4 c8f5 a4a5 d8d6 e1e8 a8e8 a5b6 a7b6 c1b2 f5e4 f3d2 g6f4 d2e4 f6e4');
INSERT INTO evaluations VALUES(56046,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(56046,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(56046,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(56046,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(56046,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(56046,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(56046,36,'cp 151','c7c5 d4d5 e6d5 d1d5 d8d5 e4d5 g8f6 c1f4 f8e7 e1c1 e8g8 h2h3 f8d8 g1f3 b8d7 f1d3 h7h6 h1e1 e7f8 a2a4 f6e8 g2g4 e8d6 f3d2 d7f6 d2c4 d6c4 d3c4 f6e8 c3e4 c8d7 b2b3 b7b6 f4h2 a7a6 d5d6 d7c6 c4d5 c6d5 d1d5 e8f6 e4f6 g7f6');
INSERT INTO evaluations VALUES(12237,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(12237,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(12237,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(12237,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(12237,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(12237,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(12237,36,'cp 66','c4d5 c6d5 c1f4 e7e6 e2e3 f6h5 f4g5 d8b6 f3d2 h7h6 g5h4 h5f6 f1d3 f8e7 a1c1 e8g8 d2b3 a7a5 a2a3 b6d8 e1g1 d7b6 b3c5 f6d7 h4e7 d8e7 c3a4 b6a4 c5a4 a8b8 d1b3 e7d8 c1c2 d7f6 b3b6 c8d7 b6d8 f8d8 a4c5 b7b6 c5d7 d8d7');
INSERT INTO evaluations VALUES(12226,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(12226,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(12226,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(12226,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(12226,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(12226,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(12226,36,'cp 64','c6d5 c1f4 e7e6 e2e3 f8e7 h2h3 e8g8 f1d3 d7b6 e1g1 c8d7 d1b3 a7a5 f1c1 a5a4 b3c2 a8c8 c2e2 d7e8 c1c2 b6c4 a1c1 h7h6 c3b1 c8a8 f3e5 b7b5 b2b3 e7d6 b3c4 d5c4 e5c4 b5c4 c2c4 d6f4 e3f4 f6d5');
INSERT INTO evaluations VALUES(16360,30,'cp -127',NULL);
INSERT INTO evaluations VALUES(16360,31,'cp -138',NULL);
INSERT INTO evaluations VALUES(16360,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(16360,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(16360,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(16360,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(16360,36,'cp -90','f6g4 e3e2 d7d6 f2f3 g4h6 c1h6 g7h6 e5d6 c8e6 b1c3 f8d6 e1c1 d8g5 c1b1 e8c8 g1h3 g5g7 e2d2 d6h2 f1d3 h2e5 h3f4 e6c4 d2e1 e5f4 d3c4 d8d1 e1d1 g7g2 c4f7 h8d8 d1f1 g2f1 h1f1 c6e5 f7b3 c7c6 c3e4 c8c7 b3e6 c7b6 c2c3 e5g6 e6g4');
INSERT INTO evaluations VALUES(16813,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(16813,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(16813,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(16813,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(16813,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(16813,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(16813,36,'cp -109','e3e2 d7d6 f2f3 g4h6 c1h6 d8h4 g2g3 h4h6 e5d6 c8e6 d6d7 e8d7 e2d2 h6d2 b1d2 a8e8 e1c1 e6a2 f1b5 d7c8 b5c6 b7c6 g1h3 f8d6 d1e1 e8e1 h1e1 h8d8 c2c3 a2d5 c1c2 c8b7 h3g5 h7h6 g5e4 b7b6 b2b3 a7a5 d2c4 d5c4 b3c4 d6f8 e4f2 a5a4 f2d3 a4a3 c4c5 b6b5 e1b1 b5c4 b1b4 c4d5 b4a4 f8c5');
INSERT INTO evaluations VALUES(17183,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(17183,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(17183,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(17183,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(17183,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(17183,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(17183,36,'cp 113','b1c3 e7g6 h2h4 f8e7 h4h5 g6h4 f3h4 e7h4 c3d5 e8g8 d2d3 d7d6 h5h6 g7g6 g2g3 h4e7 a2a4 c6d4 c2c3 d4e6 d5e7 d8e7 c1e3 c7c6 a4a5 f8d8 d1f3 d8f8 c4b3 c8d7 f3e2 e6c5 b3c2 d7e6 e1g1 e6h3 f1e1 c5e6 c2b3 e6g5 a5a6 a8e8 a6b7 e7b7');
INSERT INTO evaluations VALUES(17181,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(17181,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(17181,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(17181,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(17181,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(17181,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(17181,36,'cp 92','e7g6 h2h4 f8e7 h4h5 g6h4 f3h4 e7h4 h5h6 g7g6 c3d5 e8g8 d2d3 d7d6 g2g3 h4e7 a2a4 c6d4 c1e3 c7c6 d5e7 d8e7 c2c3 d4e6 a4a5 e6g5 a5a6 b7b6 c4b3 c8e6 f2f4 e6b3 d1b3 g5e6 f4f5 g6f5 e4f5 e6c5 b3c4 g8h8 e1c1 d6d5 c4g4 f8g8');
INSERT INTO evaluations VALUES(17278,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(17278,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(17278,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17278,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17278,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17278,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17278,36,'cp 0','f2f4 e7e5 f4e5 c6e5 g1f3 e5f3 d1f3 c7c6 b1c3 f8c5 f1e2 d7d5 e4e5 f6g4 f3f4 d8h4 g2g3 h4h3 d3d4 c5b6 e2f1 h3h5 f1e2');
INSERT INTO evaluations VALUES(30736,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(30736,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(30736,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(30736,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(30736,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(30736,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(30736,36,'cp 117','e2e4 d7d6 c1g5 h7h6 g5e3 e7e5 g1f3 b8d7 f1e2 f8g7 d4e5 d6e5 d1c2 d8e7 e1g1 c7c6 a1b1 g8f6 b2b4 e8g8 h2h3 a7a5 a2a3 a5b4 a3b4 a8a7 f3d2 f8a8 d2b3 b7a6 b4b5 c6b5 c4b5 a6b7 f1d1');
INSERT INTO evaluations VALUES(30739,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(30739,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(30739,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(30739,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(30739,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(30739,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(30739,36,'cp 94','d7d6 c1e3 b8d7 g1f3 f8h6 e3h6 g8h6 d1d2 h6g4 f1d3 e7e5 e1g1 e8g8 a1d1 g4f6 h2h3 g8g7 f1e1 c7c6 d3f1 d8e7 a2a3 f8e8 b2b3 a7a6 d2e3 h7h6 e3d2 a6a5 d4d5 c6d5 e4d5 e8c8 d2b2 d7c5 f3d2 f6h5');
INSERT INTO evaluations VALUES(35027,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(35027,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(35027,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(35027,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(35027,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(35027,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(35027,36,'cp 92','c1e3 b8c6 g2g3 f8b4 c2c3 b4a5 b1d2 a5b6 d2c4 d7d6 f1g2 g8e7 e1g1 e8g8 d3d4 e5d4 c3d4 d6d5 c4b6 a7b6 e4e5 f6g6 f3h4 g6h7 f2f4 e7f5 h4f5 h7f5 g2d5 c6b4 d5b3 f8d8 d1f3 c8e6 b3e6 f7e6 g3g4 f5g6 g1h1 b4c2 f4f5 g6f7 a1d1 a8a2');
INSERT INTO evaluations VALUES(43796,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(43796,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(43796,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(43796,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(43796,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(43796,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(43796,36,'cp 74','c4d5 e6d5 f3e5 f8e7 c1f4 e8g8 e2e3 c7c6 f1d3 f6d7 e5f3 c8b7 a1c1 d7f6 f3e5 f8e8 e1g1 b8d7 f4g3 g7g6 f2f4 f6h5 g3f2 d7e5 d4e5 h5g7 d1e2 d8d7 f1d1 a8d8 e3e4 d5e4 c3e4 d7c8 b2b4 c6c5 b4c5 b7e4 e2e4 e7c5 f2c5 b6c5 d3c4 c8g4 d1d8 e8d8');
INSERT INTO evaluations VALUES(43792,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(43792,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(43792,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(43792,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(43792,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(43792,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(43792,36,'cp 60','e6d5 f3e5 f8e7 c1f4 e8g8 e2e3 c7c6 f1d3 f6d7 e5f3 d7f6 e1g1 c8b7 f3e5 b8d7 d1c2 g7g6 f4h6 d7e5 d4e5 f6g4 h6f4 d8b8 e5e6 g4e5 e6f7 f8f7 d3e2 e7d6 f4g3 b8f8 e3e4 a8d8 h2h3 e5c4 e4d5 d6g3 f2g3 c6d5');
INSERT INTO evaluations VALUES(43810,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(43810,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(43810,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(43810,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(43810,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(43810,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(43810,36,'cp 72','f3e5 c7c6 g2g3 f8b4 f1g2 h7h6 e1g1 e8g8 c1f4 c8b7 d1b3 b4c3 b3c3 d8e7 a1c1 e7e6 c3a3 b8d7 e5d7 e6d7 c1c2 d7e6 f1c1 f8e8 a3d6 e6f5 e2e3 g7g5 f4e5 f6g4 g2h3 e8e6 d6a3 f5e4 e5d6 a8e8 a3a7');
INSERT INTO evaluations VALUES(45139,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(45139,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(45139,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(45139,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(45139,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(45139,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(45139,36,'cp -94','f6e4 d2d4 c5d4 d1d4 e4f6 f1d1 e7e6 c1f4 b8c6 d4d3 d6d5 c4b3 f8e7 c2c4 d5c4 d3e2 c8d7 b3c4 e8g8 b1c3 d8a5 a2a3 a5f5 f4g3 a8c8 c4d3 f5h5 c3e4 f8d8 b2b4 a7a6 h2h3 f6e4 d3e4 d7e8 d1d8 e7d8 a1d1 d8f6 d1c1 h7h6');
INSERT INTO evaluations VALUES(45269,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(45269,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(45269,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(45269,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(45269,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(45269,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(45269,36,'cp -82','d2d4 c5d4 d1d4 e4f6 f1d1 e7e6 c1f4 d6d5 c4b3 b8c6 d4d3 f8e7 c2c4 d5c4 d3c4 d8b6 b1c3 c8d7 c4e2 e8g8 c3e4 a8d8 a1c1 d7c8 e4d6 h7h6 h2h3 b6c7 d6c8 d8d1 b3d1 c7c8 f3e5 f6d5 e5c6 b7c6 f4e5 f8d8 d1c2');
INSERT INTO evaluations VALUES(52116,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52116,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(52116,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52116,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52116,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(52116,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(52116,36,'cp 36','b1c3 g7g6 a2a3 a7a5 d2d3 d7d6 c3d5 h7h6 e2e3 f8g7 g1e2 b8c6 e2c3 g8e7 d5e7 c6e7 f1g2 a8b8 a1b1 e7c6 c3d5 e8g8 e1g1 c6e7 d5c3 e7c6 c1d2 c8f5 c3d5 d8d7 d1b3 b7b6 f1e1 f8d8 e1d1 f5h3 g2h3 d7h3');
INSERT INTO evaluations VALUES(53220,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(53220,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(53220,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(53220,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(53220,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(53220,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(53220,36,'cp 201','d1d8 e8d8 b1c3 b8d7 a2a4 e7e6 f2f4 a7a6 g1f3 g8e7 c1e3 e7c6 e1c1 f8e7 f1e2 b7b6 d1d2 d8c7 h1d1 h7h6 f4f5 h8d8 c1b1 c7b7 f5e6 f7e6 h2h3 g7g5 e3f2 h6h5 f2g3 g5g4 h3g4 h5g4 f3h4 d7f6 d2d8 e7d8 h4g6');
INSERT INTO evaluations VALUES(35401,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(35401,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(35401,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(35401,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(35401,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(35401,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(35401,36,'cp 210','e8d8 b1c3 b8c6 c1e3 e7e6 e1c1 c8d7 g1f3 g8f6 d1d2 d8e8 h2h3 a7a6 e4e5 f6d5 c3d5 e6d5 d2d5 d7e6 d5d2 a8d8 f1d3 c6b4 d3e4 b4a2 c1b1 a2b4 h1d1 d8d2 d1d2 b4c6 f3g5 f8e7 g5e6 f7e6 f2f4 g7g6 g2g4 h8f8 d2d3 c6b4');
INSERT INTO evaluations VALUES(35440,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(35440,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(35440,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(35440,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(35440,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(35440,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(35440,36,'cp 199','b1c3 b8c6 c1e3 e7e6 e1c1 c8d7 f2f4 d8c7 c3b5 c7c8 g1f3 g8f6 b5c3 f8e7 f1e2 h7h6 d1d2 f6g4 e3g1 g4f6 h2h3 h8d8 e4e5 f6e8 g1f2 e8c7 c3e4 c7d5 g2g3 b7b6 h1d1 a7a5 h3h4 a5a4 c2c4 d5c7 e4d6 e7d6 e5d6 c7e8 f3e5 c6e5 f4e5');
INSERT INTO evaluations VALUES(35437,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(35437,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(35437,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(35437,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(35437,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(35437,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(35437,36,'cp 211','c8d7 c1e3 e7e6 g1f3 a7a6 f3e5 d8e8 f2f3 h7h5 e5c4 d7c6 a2a4 b8d7 f1e2 a8d8 h2h4 f8e7 e1c1 g8h6 c4d6 e8f8 d1d2 f7f6 h1d1 d7e5 e3c5 h6f7 f3f4 e5g4 d6f5 e7c5 d2d8 f7d8 d1d8 c6e8 f5d6 f8e7 d6b7 c5e3 c1b1 e3f4 d8a8');
INSERT INTO evaluations VALUES(28950,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(28950,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(28950,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(28950,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(28950,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(28950,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(28950,36,'cp 69','d1b3 d8c8 c4d5 f6d5 f4g3 e7e6 b1d2 d5f6 d2c4 f6e4 f2f3 e4g3 h2g3 h7h6 g3g4 f5h7 f1d3 h7d3 b3d3 f8e7 f3f4 c6c5 g1f3 b8c6 d4c5 c6b4 d3e4 c8c5 e1e2 e8g8 a2a3 c5d5 e4d5 b4d5 a1c1 a8c8 g4g5 b7b5 c4d2');
INSERT INTO evaluations VALUES(28946,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(28946,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(28946,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(28946,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(28946,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(28946,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(28946,36,'cp 58','d8c8 c4d5 f6d5 f4g3 a7a5 b1d2 a5a4 b3d1 d5f6 f1e2 b8d7 e2f3 e7e6 a2a3 c8d8 g1e2 d8b6 d2c4 b6b5 c4d6 f8d6 g3d6 d7b6 e1g1 b5b3 a1c1 b3d1 f1d1 f5e4 d6c5 e4f3 g2f3 f6d7 c5d6 f7f5 e2f4 e8f7 f4d3 b6d5 d1e1 h8d8');
INSERT INTO evaluations VALUES(46204,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46204,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(46204,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(46204,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46204,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46204,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46204,36,'cp 18','g1f3 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 c4d5 f6d5 b1c1 c5e7 f1c4 d5f6 e1g1 f8d8 d1e2 h7h6 f1d1 b8c6 h2h3 c8d7 c4d3 a8c8 a2a3 d7e8 d2e4 a7a6 e4f6 e7f6 d3e4 d8d1 c1d1 a6a5 d1c1 f6e7 g1h1 g8h8 e2d3 c8d8 d3c2 e8d7 c2c5 e7c5');
INSERT INTO evaluations VALUES(52540,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(52540,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(52540,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(52540,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(52540,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(52540,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(52540,36,'cp -40','d2d4 d7d6 b1a3 g8f6 f1d3 g7g6 g1e2 f8g7 c1g5 b8c6 c2c3 e8g8 e1g1 e7e5 f2f4 h7h6 g5h4 e5f4 h4f6 g7f6 e2f4 c6e7 a3c4 f6g7 c4e3 a7a6 h2h4 a8b8 d3c4 b7b5 c4a2 c8b7 h4h5 b7e4 f4g6 e7g6');
INSERT INTO evaluations VALUES(6751,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(6751,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(6751,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6751,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6751,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(6751,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(6751,36,'cp 13','g8f6 e1g1 e8g8 d2d3 a7a6 b5c4 h7h6 a2a4 d6c5 c1e3 c5e3 f2e3 d7d6 c3d5 f6d5 c4d5 c6e7 d5c4 e7g6 f3d2 d8g5 d1f3 c8e6 c4e6 f7e6 f3e2 g5e7 a4a5 f8f1 a1f1 a8f8 f1f8 g6f8 d2f3 f8d7 h2h3 b7b5 a5b6 c7b6 d3d4 a6a5 g1h1 e5d4 e3d4');
INSERT INTO evaluations VALUES(12320,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(12320,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(12320,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(12320,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(12320,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(12320,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(12320,36,'cp 92','b5c6 b7c6 e4e5 f6d5 c3e4 c5c4 d2d4 c4d3 d1d3 f7f5 e4d6 f8d6 e5d6 e8g8 e1g1 c6c5 f3e5 d8b6 f1e1 d5b4 d3c3 f5f4 a2a3 b4c6 e5d3 f8f5 b2b4 c5b4 a3b4 c8b7 h2h4 b6d4 c1b2 d4c3 b2c3 a7a5 b4a5 b7a6 d3e5 c6e5 e1e5 f5e5 c3e5 a8c8 e5f4 c8c2 f4e3 g8f7 f2f3');
INSERT INTO evaluations VALUES(12290,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(12290,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(12290,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(12290,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(12290,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(12290,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(12290,36,'cp 84','b7c6 e4e5 f6d5 c3e4 c5c4 d2d3 c4d3 d1d3 f7f5 e4d6 f8d6 e5d6 c6c5 f3e5 e8g8 e1g1 d5b4 d3c3 d8b6 c1e3 b4d5 c3c5 d5e3 c5e3 b6d6 f1d1 d6c7 c2c4 d7d6 e5f3 c7e7 a1c1 c8b7 b2b4 f5f4 e3b3 a8c8 h2h3 h7h6 g1h2 f8d8 d1e1 e7f7 c4c5 e6e5 c5d6 f7b3 a2b3 c8c1 e1c1 d8d6 f3e5');
INSERT INTO evaluations VALUES(11970,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(11970,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(11970,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(11970,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(11970,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(11970,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(11970,36,'cp 85','e4e5 f6d5 c3e4 c5c4 d2d3 c4d3 d1d3 f7f5 e4d6 f8d6 e5d6 e8g8 e1g1 c6c5 f3e5 d5b4 d3c3 d8b6 c1e3 b4d5 c3c5 d5e3 c5b6 a7b6 f2e3 a8a5 e5c4 a5b5 f1d1 c8b7 a2a4 b5c5 b2b3 b7c6 d1d4 b6b5 a4b5 c6b5 a1a7 e6e5 d4d2 b5c4');
INSERT INTO evaluations VALUES(12455,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(12455,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12455,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12455,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(12455,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(12455,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12455,36,'cp 0','e1g1 d7d6 d1e2 c8d7 f1d1 h7h6 c1f4 e6e5 f4e3 f8e7 f3h4 a8c8 h2h3 a7a6 a2a4 e8g8 h4g6 f8e8 a1c1 d7e6 c4e6 f7e6 e3c5 e7f8 g6f8 e8f8 d1d6 c6d4 e2d3 d8e7 d6c6 e7d7 c6d6');
INSERT INTO evaluations VALUES(12454,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12454,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(12454,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(12454,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(12454,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(12454,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(12454,36,'cp -9','d7d6 d1e2 a7a6 c1f4 f8e7 f1d1 d8c7 a1c1 e8g8 c4b3 c7b8 c3a4 b7b5 c1c6 b5a4 b3a4 c8d7 c6c4 d7a4 c4a4 b8b5 e2b5 a6b5 a4a8 f8a8 f4d6 e7d6 d1d6 h7h6 a2a3 b5b4 e4e5 b4a3 b2a3 a8a3 h2h3 f6d5 d6d8 g8h7 d8d7 f7f5 e5f6 d5f6 d7d1 h7g6 d1c1 f6d5 c1c4');
INSERT INTO evaluations VALUES(12827,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(12827,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(12827,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(12827,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(12827,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(12827,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(12827,36,'cp -38','d7d5 e4d5 f6d5 c3d5 d8d5 f1e2 b7b6 c2c3 e7e5 d1a4 c8b7 a4e4 d5e6 f3g5 e6d7 g5h7 f8e7 e2g4 d7d6 h7g5 c6d8 e4e2 b7g2 h1g1 g2d5 g5e4 d6c7 c1e3 g7g6 g4f3 d5c6 e4g5 e7g5 e3g5 c6f3 e2f3 c7c6 f3c6 d8c6 e1c1 h8h5 g5f6 e8f8');
INSERT INTO evaluations VALUES(50549,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(50549,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(50549,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(50549,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(50549,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(50549,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(50549,36,'cp -32','b7b5 c4d3 g8e7 e1g1 e7g6 f1e1 f8c5 h2h4 g6h4 e4e5 c8b7 c3e4 b7e4 d3e4 a8a7 c1d2 d7d5 e5d6 a7d7 g2g3 h4f3 d1f3 c5d6 a2a4 e8g8 a4b5 a6b5 d2e3 f7f5 e4c2 g8h8 e1d1 d8e7 a1a8 d6c5 d1d7 b8d7 f3b7 f8a8 b7a8 d7f8 e3c5 e7c5 c2b3 e6e5 b3d5 c5e7 a8c6 e7d7');
INSERT INTO evaluations VALUES(51176,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(51176,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(51176,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(51176,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(51176,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(51176,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(51176,36,'cp -36','c4d3 g8e7 e1g1 e7g6 f1e1 b8c6 c1e3 a8b8 d3f1 f8e7 a2a4 b5a4 c3d5 e6d5 e4d5 e8g8 d5c6 d7c6 d1a4 b8b2 a1d1 d8c7 d1c1 c6c5 a4a3 c7b6 e3c5 e7c5 a3c5 h7h6 c5b6 b2b6 f3e5 g6e5 e1e5 c8e6 c1a1 f8c8 e5a5 e6c4 a1c1 b6c6 c1c4 c6c4 f1c4');
INSERT INTO evaluations VALUES(51389,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51389,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(51389,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(51389,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(51389,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(51389,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51389,36,'cp 20','c8b7 b1c3 d7d6 g1f3 a7a6 f1d3 b8d7 e1g1 d6e5 f3e5 g8f6 d1e2 f8e7 f1d1 e8g8 c1f4 d8c8 c3e4 f6e4 d3e4 b7e4 e2e4 d7e5 f4e5 e7d6 d1d3 c8d7 b2b3 a8d8 a1d1 d7e7 c2c4 d6e5 e4e5 d8d7 g1f1 f8d8 d4d5 e6d5 e5e7');
INSERT INTO evaluations VALUES(51997,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51997,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51997,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51997,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51997,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(51997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51997,36,'cp 25','b8c6 d2d4 c5d4 f3d4 d7d5 f1g2 d5e4 d4c6 d8d1 e1d1 b7c6 g2e4 c8b7 d1e2 e8c8 b1d2 g8f6 e4d3 c6c5 h1e1 f6g4 d2f3 f8d6 h2h3 b7f3 e2f3 g4e5 f3e2 e5d3 c2d3 c8d7 c1e3 d8b8 b2b3 b8b5 e3d2 b5b6 e1c1 h8b8 c1c4 b6a6 c4c2 a6a3 a1c1 g7g6');
INSERT INTO evaluations VALUES(18841,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(18841,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18841,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(18841,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18841,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(18841,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(18841,36,'cp 17','d2d4 c5d4 f3d4 d7d5 f1g2 d5e4 d4c6 d8d1 e1d1 b7c6 g2e4 c8b7 d1e2 g8f6 e4d3 e8c8 b1d2 f6g4 h2h3 g4e5 d2c4 e5d3 c2d3 c6c5 h1d1 f8e7 c1e3 b7a6 a1c1 d8d5 c1c3 h8d8 d1c1 e7f6 c3a3 c8b7 a3b3 b7a8 c1c2 d8d7 h3h4 h7h6 b3a3 a8b7 a3a5 d5d3 a5c5');
INSERT INTO evaluations VALUES(6492,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6492,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(6492,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(6492,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(6492,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(6492,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(6492,36,'cp 0','d7d5 d1a4 d5e4 d3e4 d8e7 a4c2 g8f6 b2b4 c5d6 b1d2 e8g8 f1e2 a7a5 b4b5 c6d8 e1g1 b7b6 f1e1 d8b7 d2c4 c8g4 c4d6 e7d6 e1d1 d6e7 f3h4 g4e6 h4f3');
INSERT INTO evaluations VALUES(7847,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(7847,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(7847,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(7847,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(7847,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(7847,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(7847,36,'cp 25','e1g1 d7d6 c1g5 h7h6 g5h4 e8g8 d4c6 b7c6 d1f3 e6e5 a2a3 b4a5 b2b4 a5b6 c3a4 b6c7 c4e2 d8e7 c2c4 g7g5 h4g3 h6h5 f3e3 f6g4 e3d2 h5h4 h2h3 h4g3 e2g4 g3f2 d2f2 c8g4 h3g4 g8g7 f2f5 f8h8 f1f3 c7b6 a4b6 a7b6 a1d1 f7f6 f3d3');
INSERT INTO evaluations VALUES(7846,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(7846,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(7846,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(7846,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(7846,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(7846,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(7846,36,'cp 8','d7d6 c1g5 h7h6 d4c6 b7c6 g5f6 d8f6 a2a3 b4c5 d1d3 f6d4 d3e2 e8g8 c3a4 d4e5 a4c5 e5c5 b2b4 c5e5 e2e3 d6d5 e3c5 e5b8 c4d3 b8b6 a3a4 c8a6 e4d5 b6c5 b4c5 a6d3 c2d3 e6d5 h2h3 f8e8 f1b1 e8e2 b1b7 e2c2 d3d4 c2d2 b7c7 d2d4');
INSERT INTO evaluations VALUES(10465,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(10465,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(10465,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10465,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(10465,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(10465,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(10465,36,'cp 27','c1e3 d8e7 f1d3 d7d5 e3g5 c8e6 d3b5 e7c5 b5c6 c5c6 f3e5 c6a4 e1g1 f6e4 g5e3 e8g8 f2f3 e4d6 e5d3 f8e8 f1e1 f7f6 b2b3 a4d7 d3c5 d7c6 c5e6 e8e6 a3a4 a8e8 d1d3 b7b6 a1d1 d6f7 e3f2 e6e1 f2e1 e8d8 e1f2 f7e5 d3d2 h7h6');
INSERT INTO evaluations VALUES(23429,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(23429,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23429,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(23429,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(23429,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(23429,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(23429,36,'cp 25','d7d5 c4d5 d8d5 g1f3 e7e5 f1g2 c8g4 b1c3 d5d7 e1g1 g8f6 d2d3 f8e7 c1g5 e8g8 d1a4 f8d8 f1c1 g4f3 g2f3 c6d4 a4d7 d4f3 g1g2 d8d7 g2f3 c7c6 g5f6 e7f6 c3e4 f6e7 e4c5 d7d5 b2b4 b7b6 c5e4 f7f5 e4c3 e5e4 d3e4 f5e4 c3e4');
INSERT INTO evaluations VALUES(24978,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(24978,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(24978,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(24978,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(24978,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(24978,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(24978,36,'cp 64','e7e6 e1g1 f8e7 d1e2 f6e4 c4d3 e4c3 b2c3 f5d3 e2d3 e8g8 f1d1 d8d7 c3c4 f8d8 c1a3 b7b6 a1c1 c6b4 a3b4 e7b4 d3e4 f7f6 e4c2 d7e7 g2g3 b4a3 c1b1 a7a5 c2e4 a3d6 g1g2 d6b4 e4c2 a8c8 h2h4');
INSERT INTO evaluations VALUES(25232,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(25232,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(25232,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(25232,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(25232,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(25232,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(25232,36,'cp 86','e1g1 f8e7 h2h3 e8g8 f3h4 c6a5 h4f5 a5c4 f5e7 d8e7 d1b3 c4a5 b3b5 b7b6 b2b3 f8d8 c1b2 c7c6 b5e2 f6d5 c3e4 a5b7 f1c1 a8c8 c1c2 a7a5 a2a3 e7e8 a1c1 f7f6 e4d2 c6c5 e3e4 d5f4 e2f3 e6e5 d4e5 f4d3 e5f6 d3c1 b2c1');
INSERT INTO evaluations VALUES(25231,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(25231,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(25231,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(25231,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(25231,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(25231,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(25231,36,'cp 74','f8e7 d1e2 f6e4 f1d1 e8g8 c1d2 e7d6 a1c1 d8e7 d2e1 a8d8 c4b3 a7a6 a2a3 e4c3 e1c3 f5e4 b3c2 e4c2 e2c2 f7f6 f3d2 e7f7 c2b3 d8b8 d2e4 h7h6 e4d6 c7d6 d4d5 e6d5 b3d5 f8d8 g1f1 d8d7 f1e2 f7d5 d1d5 c6e7 d5d2 d6d5 b2b3');
INSERT INTO evaluations VALUES(29889,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(29889,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(29889,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(29889,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(29889,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(29889,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(29889,36,'cp 94','a2a3 c7c6 d2d3 b8a6 g2g4 h5g6 d1e2 b7b5 c4a2 b5b4 a3b4 a6b4 d3d4 b4a2 a1a2 f6d7 e1g1 a7a5 f1d1 d8c7 f3h4 f8e7 h4f5 g6f5 e4f5 d6d5 d4e5 c7e5 e2e5 d7e5 c1f4 e5d7 d1e1 e8d8 c3a4 h7h5 g4g5 e7b4 c2c3');
INSERT INTO evaluations VALUES(36561,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(36561,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(36561,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(36561,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(36561,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(36561,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(36561,36,'cp 62','h2h3 b8a6 g1f3 e7e5 f1d3 e5d4 f3d4 f8e8 e1g1 a6c5 d1c2 b7b6 f2f3 c8b7 a1d1 c7c6 d3e2 f6h5 f1e1 g7e5 e2f1 h5g3 f3f4 e5g7 e3f2 g3f1 g1f1 d8f6 f4f5 c5d7 f1g1 d7e5 c2e2 a7a5 d4f3 b7a6 f3e5 f6e5 f5g6');
INSERT INTO evaluations VALUES(49684,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(49684,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(49684,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(49684,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(49684,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(49684,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(49684,36,'cp -71','d7d5 d2d3 c7c5 g1f3 b8c6 e1g1 c8e6 f3g5 e6g4 h2h3 g4h5 g5f3 f8d6 c1g5 h7h6 g5f6 d8f6 b1c3 c6e7 c3b5 a8d8 c2c4 e8g8 c4d5 e7d5 b5d6 f6d6 a2a3 b7b6 d1c2 f7f5 a1c1 f8e8 e2e4 h5f3 g2f3 d5e7 e4f5 e7f5 f3e4 f5d4 c2b2');
INSERT INTO evaluations VALUES(52882,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52882,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52882,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52882,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52882,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52882,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52882,36,'cp 0','b7b5 g2g3 g8f6 f3e5 a7a6 f1g2 e7e6 e1g1 c8b7 b2b3 c4b3 d1b3 d8d4 c1f4 f8e7 f1d1 d4a7 a2a4 b5b4 c3e4 f6e4 g2e4 c6c5 e4f5 b7c8 f5e4');
INSERT INTO evaluations VALUES(51649,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51649,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51649,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51649,36,'cp 0','g2g3 g8f6 f1g2 e7e6 f3e5 a7a6 e1g1 c8b7 b2b3 c4b3 a2b3 f8e7 c1b2 e8g8 c3e4 f6e4 g2e4 a6a5 d1d3 h7h6 f1c1 d8c7 e4h7 g8h8 h7e4');
INSERT INTO evaluations VALUES(29016,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(29016,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(29016,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(29016,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(29016,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(29016,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(29016,36,'cp 77','d1b3 g4c8 g1f3 e7e6 f1d3 d5c4 d3c4 b7b5 c4d3 b8d7 e1g1 d7c5 d4c5 d8d3 f3d4 c8b7 a2a4 f8c5 c3b5 d3b3 b5c7 e8d7 d4b3 d7c7 b3c5 f6d7 c5d7 c7d7 f1d1 d7e7 e3e4 h8d8 d1d8 a8d8 c1e3 a7a6 f2f3 a6a5 g1f2 f7f6 e3c5 e7e8 b2b4 a5b4 c5b4');
INSERT INTO evaluations VALUES(42764,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42764,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(42764,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(42764,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42764,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42764,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42764,36,'cp 0','c7c6 f2f4 d7d6 f1e2 b4c5 c3a4 c5b4');
INSERT INTO evaluations VALUES(39881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39881,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39881,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39881,36,'cp 0','f2f4 d7d6 f1e2 b4c5 c3a4 c5b4');
INSERT INTO evaluations VALUES(46485,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(46485,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(46485,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(46485,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(46485,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(46485,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(46485,36,'cp -41','c4b3 f8b4 c1d2 e8g8 g1e2 h7h6 h2h3 b8c6 e1g1 c8e6 f2f4 d5d4 c3e4 f6e4 d3e4 d8b6 g1h1 a8d8 f1f3 e6b3 f3b3 b6c5 f4e5 b4d2 d1d2 c6e5 e2g3 e5c4 d2f2 g7g6 b3b7 c4e3 c2c3 d4c3 b2c3 c5c3 a1g1 e3d1 f2f4 c3e3 f4e3 d1e3 b7e7 d8d2');
INSERT INTO evaluations VALUES(46500,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(46500,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(46500,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(46500,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(46500,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(46500,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(46500,36,'cp -69','f8b4 c1d2 b8c6 g1e2 e8g8 e1g1 h7h6 h2h3 b4c5 g1h1 g7g5 e2g3 c8e6 h1g1 a8c8 f1e1 a7a6 c3a4 c5e7 c2c4 b7b5 a4c3 d5c4 d3c4 b5c4 b3c2 e7c5 c3e4 c5d4 d2c3 f6e4 c2e4 d4c3 b2c3 d8d1 a1d1 c6e7');
INSERT INTO evaluations VALUES(40658,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(40658,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(40658,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(40658,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(40658,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(40658,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(40658,36,'cp -47','c1d2 e8g8 g1e2 b8c6 e1g1 h7h6 h2h3 f8e8 f2f4 e5e4 d3e4 b4c5 g1h1 d5d4 c3a4 c5f8 e4e5 f6e4 d2e1 c8f5 c2c4 d8d7 b3c2 f7f6 g2g4 f5h7 f4f5 e8e5 c2e4 e5e4 a4c3 e4e8 c3d5 g8h8 e2d4 c6d4 d1d4 b7b5 b2b3 h7g8 e1f2 b5c4 b3c4 a8c8 h1g1 f8c5 d4d3 d7c6 a1e1 c5f2 g1f2 g8d5');
INSERT INTO evaluations VALUES(40659,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(40659,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(40659,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(40659,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(40659,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(40659,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(40659,36,'cp -53','e8g8 g1e2 b8c6 e1g1 h7h6 h2h3 c8e6 f2f4 d5d4 c3e4 f6e4 d3e4 d8b6 g1h1 a8d8 f1f3 g8h7 d2b4 b6b4 e2c1 e5f4 c1d3 b4e7 f3f4 e6b3 a2b3 f7f6 d1f3 f8e8 f3h5 a7a6 h5f5 h7g8 f4g4 c6e5 g4g3 e7d6 a1d1');
INSERT INTO evaluations VALUES(1623,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(1623,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(1623,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(1623,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(1623,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(1623,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(1623,36,'cp 114','e2e5 c6e5 d2d4 e5d7 c3d5 e8d8 c1f4 c7c6 d5e3 d7f6 h2h4 f6d5 e3d5 c6d5 h4h5 g8h6 e1c1 c8d7 f1d3 a8c8 c1b1 d7f5 g1f3 f5d3 d1d3 h6g4 d3b3 b7b6 f4g3 e7e6 a2a4 d8d7 a4a5 f8d6 g3d6 d7d6 a5b6 a7b6');
INSERT INTO evaluations VALUES(1617,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(1617,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(1617,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(1617,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(1617,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(1617,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(1617,36,'cp 100','c6e5 d2d4 e5d7 c3d5 e8d8 c1f4 c7c6 d5e3 d7f6 h2h4 f6d5 e3d5 c6d5 h4h5 c8f5 c2c3 a8c8 f1e2 f5d7 e2d3 g8h6 e1d2 h6g4 g1h3 h7h6 f2f3 g4f6 g2g4 e7e6 h1g1 f6e8 f4e5 f8d6 a1e1 d6e5 e1e5 e8d6 h3f4 b7b5 g1g3');
INSERT INTO evaluations VALUES(1656,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(1656,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(1656,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(1656,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(1656,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(1656,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(1656,36,'cp 107','c3d5 e8d7 d2d4 e5g6 h2h4 e7e6 d5e3 f8d6 h4h5 g6e7 f1b5 e7c6 g1f3 a7a6 b5e2 h7h6 c2c3 b7b6 e3c4 g8e7 e1g1 c8b7 f1e1 d7e8 c1d2 e7f5 b2b4 c6e7 a2a4');
INSERT INTO evaluations VALUES(39739,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(39739,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(39739,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(39739,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(39739,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(39739,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(39739,36,'cp 75','g2g4 f6d7 d1d2 e7h4 e3f2 h4f2 d2f2 d8b6 e1c1 b8c6 f2g1 c6d4 d1d4 b6c5 f1e2 b7b5 g1d1 d7b6 d1d2 e8g8 b2b4 c5c7 d4d6 c8b7 d2d4 b6d5 c3d5 e6d5 d4c5 c7e7 d6d5 e7e6 d5d6 e6a2 c5d4 a6a5 c1d2 a5b4 d4b4 b7c8 e2b5 a2a7 b5e2 a8b8 b4c3 c8e6 c3d4 a7a5 d2e3 a5c7 h1c1 f8c8 e2d3');
INSERT INTO evaluations VALUES(46181,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(46181,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(46181,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(46181,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(46181,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(46181,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(46181,36,'cp 38','f1g2 c5d4 e1g1 f8c5 f3d4 e8g8 c4d5 f6d5 d4b3 c5e7 c1d2 b8c6 b1c3 d5c3 d2c3 d8b6 b3d2 b6a6 d1b3 f8d8 b3c4 a6c4 d2c4 c8d7 c4a5 a8b8 a5c6 d7c6 g2c6 b7c6 f1c1 c6c5 c1c2 c5c4 g1f1 e7f6 c3f6 g7f6 c2c4 b8b2 c4c7 d8d5 c7a7 b2c2');
INSERT INTO evaluations VALUES(9469,30,'cp -463',NULL);
INSERT INTO evaluations VALUES(9469,31,'cp -464',NULL);
INSERT INTO evaluations VALUES(9469,32,'cp -451',NULL);
INSERT INTO evaluations VALUES(9469,33,'cp -470',NULL);
INSERT INTO evaluations VALUES(9469,34,'cp -463',NULL);
INSERT INTO evaluations VALUES(9469,35,'cp -475',NULL);
INSERT INTO evaluations VALUES(9469,36,'cp -473','d8d4 d1e2 d4e5 b1c3 c8d7 c1e3 e8c8 e3c5 e5c5 e2d2 h8e8 e1c1 c5f8 h1e1 c8b8 h2h3 c6c5 a2a3 d7c6 f2f4 f6d7 d2f2 d7b6 g2g4 f8d6 c1b1 a7a6 g4g5 b6a4 c3a4 c6a4 e4e5 d6d4 f2d4 d8d4 f4f5 d4f4 e1e4 f4e4 d3e4');
INSERT INTO evaluations VALUES(518,30,'cp -440',NULL);
INSERT INTO evaluations VALUES(518,31,'cp -468',NULL);
INSERT INTO evaluations VALUES(518,32,'cp -467',NULL);
INSERT INTO evaluations VALUES(518,33,'cp -459',NULL);
INSERT INTO evaluations VALUES(518,34,'cp -455',NULL);
INSERT INTO evaluations VALUES(518,35,'cp -440',NULL);
INSERT INTO evaluations VALUES(518,36,'cp -467','d1e2 d4e5 b1c3 c8d7 c1e3 e8c8 e3c5 e5c5 f2f3 c5e7 e1c1 c6c5 h1e1 d7c6 c1b1 c8b8 e2f2 h8e8 g2g4 f6d7 f3f4 d7b6 h2h3 e7d6 f2h4 d6f4 e1f1 f4e5 h4f2 b8a8 h3h4 b6c8 a2a3 f7f6 g4g5 f6g5 b1a1 a8b8 d1e1 g5h4 f2h4');
INSERT INTO evaluations VALUES(31763,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(31763,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(31763,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(31763,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(31763,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(31763,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(31763,36,'cp 97','e2e4 g8f6 h2h3 g4h5 d1e2 h5g6 b1c3 e7e6 h3h4 h7h5 e2b5 b8d7 b5b7 f6e4 f1d3 d6d5 d3e4 a8b8 b7a7 g6e4 c3e4 d8c8 a7a4 b8b4 a4a5 d5e4 f3g5 c8b7 c2c3 b4d4 e1g1 d4d6 c1f4 d6b6 f1e1 f8d6 f4d6 c7d6 g5e4 e8g8 b2b4 d6d5 e4c5 d7c5 a5c5 f8a8');
INSERT INTO evaluations VALUES(32986,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(32986,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(32986,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(32986,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32986,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(32986,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(32986,36,'cp 85','d7d6 e1g1 c8g4 c1g5 f6g5 f3g5 g4d1 c4f7 e8e7 c3d1 h7h6 f7g8 h6g5 g8b3 h8h7 d1e3 b8d7 c2c3 c5e3 f2e3 a8h8 f1f7 e7e8 a1f1 h7h2 g1f2 h2h1 f1h1 h8h1 f7g7 h1b1 f2f3 b1b2 d3d4 d7f6 g7g5 b2f2 f3f2 f6e4 f2f3 e4g5 f3g4 g5e4 d4e5 d6e5 c3c4 e4d6 g4g5 e8e7 g2g4 b7b5 c4b5 c6b5 g5h6 e5e4 b3c2 b5b4 g4g5 d6f5 h6h7 f5e3 c2e4 e3g4 h7g7 g4e5');
INSERT INTO evaluations VALUES(33034,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(33034,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(33034,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(33034,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(33034,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(33034,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(33034,36,'cp 53','e1g1 c8g4 c1g5 f6g5 f3g5 g4d1 c4f7 e8e7 c3d1 h7h6 f7g8 h6g5 g8b3 b8d7 d1e3 h8h7 c2c3 c5e3 f2e3 a8h8 f1f7 e7e8 a1f1 h7h2 g1f2 h2h1 f1h1 h8h1 f7g7 h1b1 f2f3 b1b2 d3d4 d7f6 g7g5 b2f2 f3f2 f6e4 f2f3 e4g5 f3g4 g5e4 c3c4 e5d4 e3d4 e8e7 g4f4 e4c3 g2g4');
INSERT INTO evaluations VALUES(33033,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(33033,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(33033,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(33033,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(33033,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(33033,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(33033,36,'cp 56','c8g4 c1g5 f6g5 c4f7 e8e7 f3g5 g4d1 c3d1 h7h6 f7g8 h6g5 g8b3 b8d7 d1e3 h8h7 c2c3 c5e3 f2e3 a8h8 h2h3 g5g4 h3g4 d7f6 g1f2 f6g4 f2e2 g4f6 f1f5 h7h1 a1h1 h8h1 f5f1 h1f1 e2f1 a7a5 a2a4 f6d7 b3d1 d7f6 f1f2 b7b5 f2g3 g7g5 g3f3 e7d7 f3e2 d7d8 d1b3 d8d7 a4b5');
INSERT INTO evaluations VALUES(38659,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(38659,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(38659,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(38659,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(38659,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38659,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(38659,36,'cp 15','e2e3 d6f4 e3f4 g8e7 h2h4 h7h5 g1f3 b7b6 g2g3 b8d7 f1h3 d7f6 e1g1 c7c5 f1e1 e8g8 c3b1 e7c6 b1d2 a7a5 c2c3 d8d6 a1c1 c5d4 c3d4 c8d7 e1e3 f8c8 e3c3 c8c7 h3f1 a8c8 f1a6 c8b8 f3g5 d7e8 d2f3 f6e4');
INSERT INTO evaluations VALUES(38658,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(38658,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38658,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(38658,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(38658,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(38658,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38658,36,'cp 10','d6f4 e3f4 g8e7 g1f3 b7b6 h2h4 h7h5 g2g3 b8d7 f1h3 d7f6 e1g1 e8g8 c3b1 c7c5 f1e1 e7c6 c2c3 a7a5 b1d2 d8d6 a2a3 c5d4 c3d4 c8d7 d1e2 f8c8 e2d3 c8c7 f3g5 c6e7 d2f3 d7e8 a1c1 c7c1 e1c1 b6b5 f4f5 e6f5 h3f5');
INSERT INTO evaluations VALUES(54876,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(54876,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(54876,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(54876,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(54876,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(54876,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54876,36,'cp 51','f8g7 f2f4 g8h6 g1f3 e8g8 f1e2 c7c5 d4c5 b8c6 b1c3 e7e6 c3b5 b7b6 c5b6 d8b6 d1d3 c8a6 d3e3 a6b5 e3b6 a7b6 e2b5 c6b4 e1d2 b4a2 b5c6 a8a5 f3d4 a2c1 a1a5 b6a5 h1c1 f8b8 b2b3 h6f5 d4f5');
INSERT INTO evaluations VALUES(38481,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(38481,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(38481,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(38481,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(38481,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(38481,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(38481,36,'cp 59','g1f3 f7f6 h2h3 f6e5 d4e5 g8h6 c1f4 h6f7 b1c3 e7e6 d1d2 d8e7 e1c1 b8c6 f1b5 c8d7 d2e3 a7a6 b5c6 d7c6 h3h4 b7b6 h4h5 e8c8 c3e2 g6h5 h1h5 h7h6 c1b1 c8b7 f3h4 e7e8 g2g4 c6b5 h4g6 h8g8 e2c3 g7f8 g6f8 e8f8 f4h6 f7h6 h5h6 g8g4');
INSERT INTO evaluations VALUES(56854,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(56854,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(56854,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(56854,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(56854,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(56854,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(56854,36,'cp 44','d7d6 d1e2 c7c5 d4c5 d6e5 b1c3 f8c5 e2e5 c5d4 e5g3 g8e7 f1d3 b8c6 g1f3 d4g7 e1g1 e8g8 f1d1 d8a5 c3e4 a5h5 c2c3 e6e5 d3e2 e7f5 g3g5 h5g5 c1g5 h7h6 g5f6 g7f6 e4f6 g8g7 f6e4 f5e7 a2a4 b7b6 b2b4 f7f5 e4d6 e5e4 f3d4 c6d4 d1d4 c8e6 b4b5');
INSERT INTO evaluations VALUES(16404,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(16404,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(16404,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(16404,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16404,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16404,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16404,36,'cp 0','f8c5 e1g1 d7d6 d2d3 e8g8 a2a4 c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 d8e8 h2h3 e8g6 f1f2 f8f7 b1d2 a8f8 a1f1 b7b6 f3h2 a7a6 c2c3 a6a5 f2f3 c6e7 d2c4 e7c6');
INSERT INTO evaluations VALUES(10130,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10130,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10130,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(10130,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(10130,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(10130,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10130,36,'cp -2','e1g1 e8g8 d2d3 d7d6 a2a4 c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 d8e8 b1d2 e8g6 b2b4 c6b4 f1b1 a7a5 c2c3 b4c6 b1b7 f8f7 a1f1 h7h6 d2c4 g6h5 b7b5 f6d7 d3d4 a8f8 h2h3 h5g6 c4d2 d7b6 f1f2 g6g3 d2f1 g3g6');
INSERT INTO evaluations VALUES(10131,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(10131,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(10131,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10131,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10131,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10131,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(10131,36,'cp -7','e8g8 d2d3 c6a5 c1e3 c5e3 e2e3 a5c4 d3c4 b7b6 b1c3 d7d6 a1d1 c8b7 b2b3 a7a5 f3e5 f8e8 e5f3 f6e4 f1e1 e4c3 e3c3 d8d7 f3d4 h7h6 f2f3 e8e1 d1e1 a8d8 e1d1 d8e8 a2a4 h6h5 d4b5 d7d8 g1f1 h5h4 h2h3 e8e5 d1e1 d8e7 c3d2 b7c6');
INSERT INTO evaluations VALUES(32895,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(32895,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(32895,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(32895,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(32895,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(32895,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(32895,36,'cp 120','b2b4 a5b6 c3d5 b6d6 d5e3 f5g6 e3c4 d6d8 a2a3 e7e6 g2g3 b8d7 f1g2 f8e7 e1g1 g8f6 d2d4 e8g8 f1e1 c7c6 c4a5 d8c8 f3h4 g6h5 d1d2 e7d8 c2c4 d8a5 b4a5 f8e8 d2c3 b7b5 a5b6 a7b6 h2h3 h5g6 c1f4 a8a7 a1c1 g8h8 g3g4 c6c5 d4d5 e6d5 e1e8 c8e8 h4g6 h7g6 c4d5');
INSERT INTO evaluations VALUES(35149,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(35149,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(35149,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(35149,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(35149,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(35149,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(35149,36,'cp 26','b8c6 d4c4 f8b4 c2c3 b4a5 c1g5 f6g6 h2h4 d7d6 h4h5 g6e6 g5h4 h7h6 c4e6 c8e6 b1d2 g8f6 h4f6 g7f6 a2a4 a5b6 d2c4 e8g8 h1h4 f8e8 f1e2 a8d8 f3d2 d6d5 e4d5 d8d5 b2b4 a7a5 c4b6 c7b6 d2c4 a5b4 c4e3 d5d7 c3b4');
INSERT INTO evaluations VALUES(1881,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(1881,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(1881,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(1881,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(1881,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(1881,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(1881,36,'cp 29','d4c4 f8b4 c2c3 b4a5 c1g5 f6g6 h2h4 d7d6 h4h5 g6e6 c4e6 c8e6 h5h6 g8h6 g5h6 g7h6 h1h6 e8c8 b1d2 h8e8 f3g5 d6d5 g5e6 e8e6 h6e6 f7e6 e4d5 e6d5 f1d3 h7h5 e1e2 a5b6 a2a4 a7a5 a1e1 c6e5 d3f5 c8b8 e2f1 e5c4 d2c4');
INSERT INTO evaluations VALUES(39186,30,'cp 248',NULL);
INSERT INTO evaluations VALUES(39186,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(39186,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(39186,33,'cp 242',NULL);
INSERT INTO evaluations VALUES(39186,34,'cp 234',NULL);
INSERT INTO evaluations VALUES(39186,35,'cp 230',NULL);
INSERT INTO evaluations VALUES(39186,36,'cp 221','a2a4 a7a6 b1a3 d7d5 b2b4 a5b6 a4a5 b6a7 e4e5 f7f6 f1d3 d8d7 f2f4 b8c6 d1e2 g8h6 g1f3 e8g8 e2c2 f6f5 h2h3 d7e8 e1g1 c8d7 g1h1 g8h8 c2e2 c6e7 c1e3 e7g6 f1g1 d7c6 g2g3 e8d7 b4b5 a6b5 a3b5 c6b5 d3b5 c7c6');
INSERT INTO evaluations VALUES(39696,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(39696,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(39696,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(39696,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(39696,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(39696,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(39696,36,'cp -32','b7b5 c4d3 c8b7 d1e2 b5b4 c3b1 b8d7 a2a3 b4a3 b1a3 d7e5 a3c4 e5c4 d3c4 h7h6 g5c1 f6e4 c4a6 b7a6 a1a6 a8a6 e2a6 e8g8 d4c6 d8c7 c6e7 c7e7 c1e3 e7c7 c2c4 f8b8 f1d1 e6e5 b2b3 f7f5 f2f3 b8b3 f3e4 b3e3 a6d6 c7d6 d1d6 e3e4');
INSERT INTO evaluations VALUES(39798,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(39798,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(39798,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(39798,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(39798,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(39798,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(39798,36,'cp -26','c4b3 e8g8 f2f3 b5b4 c3a4 c8d7 c2c3 d7a4 b3a4 b4c3 b2c3 d8c7 g1h1 f8c8 g5f6 e7f6 f3f4 c7c3 a1c1 c3c1 d1c1 c8c1 f1c1 b8d7 a4d7 f6d4 c1c6 d4c5 e4e5 c5e3 f4f5 d6e5 f5e6 g8f8 e6f7 f8f7 d7e6 f7g6 e6c8 g6h5 c8a6 e5e4 a6e2 h5g5 g2g3');
INSERT INTO evaluations VALUES(39799,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(39799,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(39799,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(39799,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(39799,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(39799,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(39799,36,'cp -31','e8g8 f2f3 d8c7 g1h1 b8d7 a2a3 h7h6 g5e3 a8b8 d1e1 d7c5 b3a2 e6e5 d4e2 c5a4 c3a4 b5a4 b2b4 a4b3 c2b3 c8e6 b3b4 e6a2 a1a2 d6d5 e1g3 g8h7 a2d2 f8d8 f1c1 c7d6 e4d5 f6d5 e3c5 d6f6 g3e1 b8c8 c5e7 d5e7 d2d8');
INSERT INTO evaluations VALUES(50394,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(50394,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(50394,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(50394,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(50394,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(50394,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(50394,36,'cp 142','d2d4 c7c6 c2c4 d7d5 e4d5 c6d5 c4d5 d8d5 b1c3 d5d8 f1c4 h6f5 g1f3 e7e6 e1g1 f8e7 d1e2 e8g8 f1d1 e7f6 d4d5 d8e7 c4d3 e6d5 c3d5 e7e2 d5f6 g7f6 d3e2 b8c6 b2b3 g8g7 c1b2 f8e8 e2b5 a7a6 b5c6 b7c6 a1c1 c8e6 c1c6 e8d8 c6c1 a8c8 b2a3 d8d1 c1d1 c8c2');
INSERT INTO evaluations VALUES(50392,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(50392,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(50392,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(50392,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(50392,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(50392,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(50392,36,'cp 128','c7c6 c2c4 d7d5 c4d5 c6d5 e4d5 d8d5 b1c3 d5d8 c1f4 g7g6 f4e5 f7f6 e5g3 f8g7 f1c4 b8c6 g1e2 h6f5 e1g1 c6a5 c4b5 c8d7 d1a4 a5c6 d4d5 c6e5 a4b3 e8g8 a1d1 d7b5 c3b5 g8h8 e2f4 f5g3 h2g3 d8b6 f4e6 f8c8 d5d6 e7d6 d1d6 e5c6 a2a3 a7a6');
INSERT INTO evaluations VALUES(8584,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8584,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8584,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8584,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(8584,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(8584,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8584,36,'cp -6','c6a5 c4b3 c7c6 h2h3 a5b3 a2b3 e8g8 c3e2 f8e8 e2g3 h7h6 c1d2 e7f8 f1e1 c6c5 f3h2 d6d5 c2c4 d5d4 f2f4 e5f4 d2f4 a7a5 e1f1 a8a6 d1f3 f8d6 g3h5 d6e5 f4e5 e8e5 h5f4 c8d7 f4d5 g8h7 a1e1 f6d5 c4d5 a6f6');
INSERT INTO evaluations VALUES(19944,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(19944,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(19944,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(19944,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(19944,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19944,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19944,36,'cp 31','d2d4 e7e6 d4d5 e6d5 e4d5 c6e5 c4b5 c8d7 a2a4 a7a6 b5d7 d8d7 g1f3 d7f5 f3e5 f5e5 d1e2 e5e2 e1e2 g8f6 c3c4 e8c8 h2h3 f6d7 b2b3 d8e8 e2d3 f8e7 c1b2 f7f5 h1f1 e7h4 b1d2 e8e7 a1e1 h8e8 e1e7 e8e7 g2g3 h4f6 b2f6 d7f6 d2f3 f6d7 a4a5 d7e5 f3e5 e7e5');
INSERT INTO evaluations VALUES(19961,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19961,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(19961,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19961,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(19961,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(19961,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(19961,36,'cp 22','e7e6 d4d5 e6d5 e4d5 c6e5 c4b5 c8d7 a2a4 a7a6 b5d7 d8d7 g1f3 d7f5 f3e5 f5e5 d1e2 e5e2 e1e2 g8f6 c3c4 e8c8 h2h3 f6d7 f2f4 d8e8 e2d3 f7f5 b1c3 g7g6 c1d2 f8g7 h1e1 g7d4 g2g4 d7f6 g4g5 f6h5 b2b4 c5b4 d3d4 b4c3 d4c3 e8e4 e1e4 f5e4 c3d4 h5g3 a4a5');
INSERT INTO evaluations VALUES(55038,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(55038,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(55038,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(55038,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(55038,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(55038,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(55038,36,'cp -71','d5c4 d1a4 b8c6 a4c4 e7e5 b1d2 c8e6 c4c2 c6b4 c2d1 c7c5 g1f3 b4c6 b2b3 g8h6 c1b2 f7f6 e2e3 d8d7 f1e2 f8e7 e1g1 e8g8 a2a3 h6f7 d1c2 a8c8 a1c1 b7b6 c2b1 f8d8 b2c3 f7d6 c1c2 h7h6 f1c1 e7f8 d2c4 d6c4 b3c4 g8h8');
INSERT INTO evaluations VALUES(55215,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(55215,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(55215,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(55215,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(55215,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(55215,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(55215,36,'cp -62','d1a4 b8c6 a4c4 e7e5 g1f3 g8f6 b1d2 c8e6 c4c2 c6d4 c2d1 c7c6 e2e4 e6g4 f1e2 g4f3 d2f3 d4f3 e2f3 f8c5 e1g1 d8d6 f3e2 a7a5 c1d2 c5d4 d1c2 e8g8 g2g3 a5a4 a1c1 f6d7 a2a3 f8d8 g1g2 d6f6 d2c3 d7f8 e2g4 f8e6 g4e6 f6e6 c3d4');
INSERT INTO evaluations VALUES(1232,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(1232,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(1232,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(1232,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(1232,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(1232,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(1232,36,'cp 109','c3d5 f6d8 b5a4 g8f6 e1g1 f8e7 f1e1 b7b5 a4b3 d7d6 c2c3 e8g8 a2a4 c8b7 d2d3 h7h6 h2h3 d8d7 d5f6 e7f6 f3h2 d7e7 h2f1 c6d8 f1g3 f6g5 g3f5 e7f6 c1g5 h6g5 f5e3 d8e6 b3d5 b7d5 e3d5 f6h6 d5e7 g8h7 b2b4 a6a5 e7f5 h6f6 a4b5 a5b4 c3b4 a8a1 d1h5 h7g8');
INSERT INTO evaluations VALUES(1230,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(1230,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(1230,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(1230,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(1230,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(1230,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(1230,36,'cp 122','f6d8 b5a4 g8f6 e1g1 f8e7 f1e1 d7d6 d5b4 c8d7 b4c6 d7c6 a4c6 b7c6 d2d4 f6d7 d4e5 d7e5 f3e5 d6e5 d1g4 e8g8 c1h6 e7f6 e1d1 d8c8 d1d7 f8d8 d7d8 c8d8 a1d1 d8c8 g4c8 a8c8 h6e3 c8d8 d1d8 f6d8 g1f1 g7g6 f2f3 h7h5 f1e2 f7f5 e2d3 g8f7 d3c4 f7e6 c2c3 d8e7 b2b3 e7h4 a2a4 h4f6 c4c5 e6d7 e4f5 g6f5');
INSERT INTO evaluations VALUES(5977,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(5977,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5977,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(5977,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(5977,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5977,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5977,36,'cp 9','g8f6 c2c3 a7a5 e1g1 h7h6 b1d2 e8g8 f1e1 c8e6 c4b5 f6d7 d2c4 d7f6 c4e3 c5b6 e3f1 c6e7 f1g3 e7g6 d3d4 e5d4 f3d4 d6d5 e4e5 f6e4 c1e3 e6c8 d1h5 d8e7 d4f5 c8f5 g3f5 e7e5 e3b6 c7b6 h5g4 e5f4 f5e7 g8h8 e7g6 f7g6 g4f4');
INSERT INTO evaluations VALUES(12363,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(12363,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(12363,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(12363,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(12363,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(12363,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(12363,36,'cp -64','d7d5 e4d5 e6d5 c4b5 f8d6 e1g1 d8b6 b5c6 b7c6 f1e1 c8e6 b2b3 e8g8 c1g5 f6d7 g5h4 f8e8 h4g3 d6e7 b1d2 a7a5 h2h3 b6a6 d3d4 a5a4 a1c1 e7f8 d1c2 f7f6 c3c4 g7g6 g3f4 e6f7 e1e8 a8e8 b3a4 d7b6 f4e3 b6c4 d2c4');
INSERT INTO evaluations VALUES(13529,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(13529,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(13529,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(13529,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(13529,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(13529,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(13529,36,'cp -55','e4d5 e6d5 c4b5 f8d6 e1g1 h7h6 d1e2 c8e6 d3d4 e8g8 d4c5 d6c5 c1e3 c5e3 e2e3 f8e8 b5c6 b7c6 b1d2 d8b6 e3b6 a7b6 f1e1 c6c5 d2f1 g8f8 a2a3 e6d7 e1e8 f8e8 f3e1 d7e6 a1d1 g7g5 f1e3 e8e7 f2f3 h6h5 g2g3 a8a4 g1f2 g5g4 e1g2 g4f3 f2f3 d5d4');
INSERT INTO evaluations VALUES(13509,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(13509,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(13509,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(13509,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(13509,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(13509,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(13509,36,'cp -57','e6d5 c4b5 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 d8b6 c3c4 g7g5 h4g3 c6e7 f1e1 f8d8 b5a4 d5c4 g3d6 b6d6 d3c4 d6d1 a4d1 c8e6 b1c3 e7c6 f3e5 c6d4 f2f3 d8d6 d1a4 a8d8 g1f2 a7a6 a1b1 e6f5 b1d1 h6h5 c3e2 g8g7 g2g3 h5h4 e2d4 d6d4 g3h4');
INSERT INTO evaluations VALUES(13737,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(13737,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(13737,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(13737,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(13737,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(13737,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(13737,36,'cp -59','c4b5 f8d6 e1g1 d8b6 f1e1 c8e6 b1a3 e8g8 c1g5 f6d7 b5a4 a7a6 a4b3 h7h6 g5h4 d7e5 f3e5 d6e5 d1f3 a8e8 h4g3 e5g3 h2g3 e8d8 a3c2 b6b5 a1b1 f8e8 c2a3 b5b6 b1d1 e8e7 a3c2 d8e8 b3d5 e6d5 f3d5 e7e1 d1e1 e8e1 c2e1');
INSERT INTO evaluations VALUES(13718,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(13718,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(13718,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(13718,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(13718,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(13718,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(13718,36,'cp -66','f8d6 e1g1 d8b6 b5c6 b7c6 f1e1 c8e6 b2b3 e8g8 c1g5 f6d7 g5h4 f8e8 h4g3 d6e7 b1d2 a7a5 h2h3 h7h6 d1c2 e6f5 c3c4 a5a4 b3a4 b6a5 a1b1 e7f6 e1e8 a8e8 b1b7 e8a8 d2f1 f5e6 f1e3 a5a4 c2a4 a8a4 e3g4');
INSERT INTO evaluations VALUES(16730,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16730,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(16730,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(16730,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(16730,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(16730,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(16730,36,'cp 12','g7g6 h5g5 d7d6 b1c3 c8e6 c3d5 f8g7 g1f3 h6g8 g5d8 e8d8 a2a4 h7h6 a4a5 c6e7 d5e7 d8e7 c4e6 e7e6 e1e2 g8e7 h2h4 b7b5 h4h5 g6g5 g2g4 c7c5 c1e3 e7c6 f3d2 c6b4 c2c3 b4c6 a5a6 h8f8 b2b3 b5b4 c3b4 c6b4 h1b1');
INSERT INTO evaluations VALUES(16072,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(16072,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(16072,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(16072,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(16072,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(16072,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(16072,36,'cp 18','h5g5 d7d6 b1c3 f8g7 g1e2 c8e6 h2h4 c6b4 g5d2 h6g4 c4e6 f7e6 c3b5 b4a6 d2g5 g4f6 b5a3 d8e7 c1d2 a6b8 e1c1 b8c6 f2f3 d6d5 c2c3 e8c8 a3c2 h8f8 c1b1 c8b8 g5g3 f6h5 g3f2 e7f7 c2b4 c6b4 c3b4 d8c8 g2g4');
INSERT INTO evaluations VALUES(48713,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(48713,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(48713,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(48713,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(48713,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(48713,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(48713,36,'cp -65','f8d6 g1e2 e8g8 e1g1 c8f5 h2h3 f8e8 c1f4 a7a5 b1c3 d6f8 e2d4 f5g6 d1d2 d8d7 a2a4 f8b4 d2d1 b8a6 d4f3 c7c6 c3e2 a6c5 f3e5 d7d8 c2c3 g6h5 b3c2 c5e6 f4h2 b4d6 g2g4 d6e5 h2e5 h5g6');
INSERT INTO evaluations VALUES(41723,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(41723,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(41723,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(41723,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(41723,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(41723,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(41723,36,'cp -70','g1e2 e8g8 e1g1 f8e8 h2h3 c8f5 c1f4 a7a5 b1c3 d6f4 e2f4 c7c6 a2a3 b8d7 b3a2 a5a4 f1e1 d8b6 a1b1 h7h6 c3e2 f5h7 e2g3 b6a5 f4e2 d7c5 e2d4 a8d8 d1c1 a5b6 d4f5 b6c7 c1d2 c5e6 b2b4 a4b3 a2b3 h7g6 d2b4');
INSERT INTO evaluations VALUES(9310,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9310,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(9310,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(9310,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9310,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(9310,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9310,36,'cp 14','h7h6 g5h4 g7g5 h4g3 c8g4 b1d2 a7a6 c2c3 c5a7 a2a4 f6h5 g1h1 d8f6 b2b4 c6e7 b4b5 a6b5 a4b5 e8g8 h2h3 g4f3 d2f3 a7b6 a1a8 f8a8 d3d4 h5f4 g3f4 f6f4 d1d3 e7g6 g2g3 f4f6 h1g2 a8a3 f1d1 b6d4 f3d4 e5d4 d3d4 f6d4');
INSERT INTO evaluations VALUES(8812,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(8812,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8812,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(8812,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(8812,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(8812,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(8812,36,'cp 16','g5h4 a7a6 c2c3 g7g5 h4g3 c5a7 b1d2 c8g4 g1h1 e8g8 h2h3 g4h5 b2b4 b7b5 c4b3 a6a5 a2a3 a5b4 a3b4 d8d7 f1e1 a7b6 h1g1 g8g7 a1c1 h5g6 d2f1 a8a3 d1c2 f6h5 c2b2 f8a8 g3h2 h5f4 h2f4 g5f4 b3d5 a8a4 d3d4 e5d4 d5c6 d7c6');
INSERT INTO evaluations VALUES(8819,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(8819,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(8819,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(8819,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8819,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8819,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8819,36,'cp 0','a7a6 b1d2 c5a7 c2c3 g7g5 h4g3 c8g4 b2b4 f6h5 g1h1 d8f6 a2a4 c6e7 b4b5 e8g8 h2h3 g4c8 c4b3 e7g6 d2c4 h5f4 c4e3 h6h5 f3h2 g8g7 b5a6 b7a6 e3d5 f6d8 d5e3');
INSERT INTO evaluations VALUES(20339,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(20339,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(20339,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(20339,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(20339,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(20339,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(20339,36,'cp 0','c2c3 d8b6 d1d2 g8h6 g1f3 c8d7 f1e2 c5d4 c3d4 h6f5 b1c3 f5e3 f2e3 f7f6 e5f6 g7f6 e1g1 e8c8 b2b4 f8b4 a2a3 b4c3 d2c3 b6a5 c3b2 a5b6');
INSERT INTO evaluations VALUES(30769,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(30769,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(30769,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(30769,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(30769,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(30769,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30769,36,'cp 27','e7e6 b1c3 d7d6 g1f3 a7a6 f1d3 d6e5 f3e5 g8f6 d3e2 a8a7 c1e3 b8d7 e2f3 d7e5 f3b7 a7b7 d4e5 f6d5 e3d2 d5c3 d2c3 d8d1 a1d1 b7b8 e1e2 b8d8 c3d4 d8d5 c2c4 d5d8 f2f4 h7h5 d4f2 f8e7 d1d8 e8d8');
INSERT INTO evaluations VALUES(47804,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(47804,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(47804,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(47804,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(47804,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(47804,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47804,36,'cp -33','b8c6 c1e3 c6a5 h2h3 a5c4 d3c4 c8e6 b2b3 c7c6 g1e2 a7a6 b1d2 f8e7 f3g3 b7b5 c4c5 e8g8 e1g1 d8c7 e3h6 f6e8 f2f4 f7f6 f4e5 d6e5 h6e3 a6a5 c2c4 e6f7 g3g4 b5c4 d2c4 f7c4 b3c4 a5a4');
INSERT INTO evaluations VALUES(15498,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(15498,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(15498,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(15498,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(15498,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(15498,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(15498,36,'cp -26','c1e3 c6a5 h2h3 a5c4 d3c4 c8e6 b2b3 c7c6 g1e2 a7a6 b1d2 f8e7 e1g1 h7h5 c4c5 d6d5 e4d5 c6d5 c2c4 d8c7 c4d5 f6d5 d2e4 e8c8 a1c1 f7f5 e4d6 c8b8 f1d1 f5f4 e3d2 e7d6 c5d6 c7d6 d2a5 b7b6 a5d2 g7g5 d2c3 g5g4 f3e4 d5c3 d1d6 c3e4 d6b6 b8a7');
INSERT INTO evaluations VALUES(24864,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(24864,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(24864,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(24864,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(24864,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(24864,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(24864,36,'cp 28','e7e6 c2c4 f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 c6e5 d4e5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c7c6 d5c4 b7b5 c4d3 f5d3 e2d3 f8d8 c1e3 d8d3 a1d1 d3d8 d1d8 a8d8 f1c1 d8c8 g1g2 f7f5 e5f6 g7f6 g2f3 g8f7 f3e4 f7e6 e3c5 e7c5 c1c5 e6d6 c5h5 c8e8 e4d3 e8e1 h5h7 e1b1');
INSERT INTO evaluations VALUES(25179,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(25179,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(25179,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(25179,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(25179,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(25179,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25179,36,'cp 40','c2c4 f8e7 b1c3 d5c4 d1a4 e8g8 a4c4 f6e4 f3e5 e4c3 e5c6 c3e2 c4e2 b7c6 c1e3 a8b8 g2c6 e7d6 b2b3 d8f6 h2h4 d6a3 a1d1 h7h6 e3f4 a3d6 f4e5 f6g6 c6f3 a7a5 d4d5 b8e8 e5d6 c7d6 d5e6 e8e6 e2d2 f5g4 f3g4 g6g4 d2a5 g4h3');
INSERT INTO evaluations VALUES(25143,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(25143,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(25143,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(25143,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25143,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25143,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(25143,36,'cp 47','f8e7 f3e5 e8g8 b1c3 a7a5 c4d5 c6e5 d4e5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c7c6 d5c4 b7b5 c4d3 f5d3 e2d3 a8d8 c1e3 d8d3 a1d1 d3d1 f1d1 f8d8 d1c1 d8c8 g1g2 f7f6 e5f6 e7f6 b2b3 f6e7 c1d1 c8d8 d1d8 e7d8 g2f3 g8f7 f3e4 f7e6 a2a4 g7g6 g3g4 d8e7');
INSERT INTO evaluations VALUES(29639,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29639,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29639,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29639,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29639,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29639,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(29639,36,'cp -11','c7c5 g1f3 e7e6 e1g1 b8c6 d4c5 f8c5 c2c4 e8g8 b1c3 c5b4 a1c1 h7h6 h2h3 d5c4 d3c4 a8c8 a2a3 b4c3 c4c3 f6d5 c3b3 d8b6 b3b6 d5b6 f1d1 b6d5 f4g3 f7f6 e3e4 d5b6 g3f4 f8d8 d1d8 c6d8 c1c8 b6c8 f4c7 d8c6 g1f1 g8f7');
INSERT INTO evaluations VALUES(31822,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(31822,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(31822,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31822,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(31822,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31822,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31822,36,'cp 23','e7e6 b1c3 g8f6 d1b3 b8d7 e2e3 d5c4 f1c4 d7b6 g1f3 a7a5 f3e5 f8b4 e1g1 b6c4 b3c4 e8g8 f1c1 b4c3 b2c3 f6h5 f4g3 f8e8 c4e2 h5g3 h2g3 f7f6 e5f3 d8d7 e3e4 f5g4 a2a4 a8d8 c1e1 b7b6 e2e3 g4h5 e4e5 f6f5 f3g5 h5g6');
INSERT INTO evaluations VALUES(32276,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32276,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(32276,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32276,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(32276,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(32276,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(32276,36,'cp 14','b1c3 g8f6 d1b3 b8c6 b3b7 c6b4 a1c1 a8b8 b7a7 f8d6 f4d6 c7d6 a7a4 e8e7 g1f3 d8b6 c4c5 d6c5 d4c5 b6c5 e2e3 b8a8 a4d1 c5a5 a2a3 b4a2 b2b4 a5a3 c3a2 a3a2 f1b5 h8c8 e1g1 c8c1 d1c1 a2c2 f3d4 c2c1 f1c1 f5g6 f2f3 a8a2 h2h4 e6e5 d4c6 e7d6 g2g4 a2c2 c1c2 g6c2 g4g5');
INSERT INTO evaluations VALUES(32273,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32273,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(32273,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32273,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(32273,36,'cp 12','g8f6 g1f3 f8b4 e2e3 e8g8 a1c1 c7c5 d4c5 b8d7 f4d6 d7c5 d6c5 b4c5 c4d5 a8c8 d5e6 d8d1 c1d1 f5e6 f1d3 c5b4 d1c1 e6a2 e1e2 a2e6 f3d4 e6d7 h1d1 a7a6 d3f5 d7f5 d4f5 g7g6 f5d6 c8c7 c3e4 b4d6 e4d6 c7d7 e2f3 f8d8');
INSERT INTO evaluations VALUES(30062,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30062,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30062,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(30062,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(30062,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30062,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(30062,36,'cp 33','d1b3 b8d7 e2e3 d5c4 f1c4 d7b6 g1f3 a7a5 f3e5 f8b4 e1g1 b6c4 b3c4 e8g8 a1c1 b4c3 b2c3 f6e4 c4e2 f7f6 f2f3 f6e5 f3e4 f5e4 f4e5 d8e7 f1f4 f8f4 e3f4 e4c6 c1e1 b7b5 h2h3 e7d7 a2a3 a5a4 c3c4 b5c4 e2c4');
INSERT INTO evaluations VALUES(31919,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(31919,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(31919,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31919,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(31919,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(31919,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(31919,36,'cp 21','d1d3 c7c6 g1f3 e7e6 e1g1 h7h6 c2c4 g8f6 b1d2 f8e7 b2b3 e8g8 f1d1 b8d7 e3e4 d5e4 d2e4 d8a5 c1f4 a8d8 a2a3 f6e4 d3e4 f8e8 b3b4 d7f6 e4e2 a5f5 f4e5 f6d7 h2h3 d7b6 e5c7 d8d7 c7g3 e7f6 d1e1 e8d8 a1d1');
INSERT INTO evaluations VALUES(31918,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(31918,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(31918,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(31918,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31918,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31918,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(31918,36,'cp 19','c7c6 g1f3 e7e6 c2c4 g8f6 e1g1 f8e7 b1c3 e8g8 b2b3 b8d7 e3e4 d5e4 c3e4 h7h6 c1f4 d8a5 f1e1 f6e4 d3e4 f8e8 f4d2 a5b6 a1d1 c6c5 d2c3 e7f6 d1d2 a8d8 h2h4 c5d4 f3d4 a7a6 e1d1 d7e5 e4e3 b6c7');
INSERT INTO evaluations VALUES(56127,30,'cp -388',NULL);
INSERT INTO evaluations VALUES(56127,31,'cp -359',NULL);
INSERT INTO evaluations VALUES(56127,32,'cp -381',NULL);
INSERT INTO evaluations VALUES(56127,33,'cp -389',NULL);
INSERT INTO evaluations VALUES(56127,34,'cp -381',NULL);
INSERT INTO evaluations VALUES(56127,35,'cp -379',NULL);
INSERT INTO evaluations VALUES(56127,36,'cp -361','c7c6 f3e5 c6d5 f1b5 c8d7 e5d7 b8d7 e4d5 a7a6 d1e2 d8e7 b5d3 g8f6 c2c4 d4c3 d2c3 e8c8 e1g1 e7e2 d3e2 f8c5 c3c4 h8e8 e2d3 f6e4 c1f4 f7f6 g2g3 e4d6 h2h4 h7h6 a1e1 e8e1 f1e1 g7g5 f4d6 c5d6 h4g5 f6g5 d3f5 c8b8');
INSERT INTO evaluations VALUES(53684,30,'cp -371',NULL);
INSERT INTO evaluations VALUES(53684,31,'cp -364',NULL);
INSERT INTO evaluations VALUES(53684,32,'cp -377',NULL);
INSERT INTO evaluations VALUES(53684,33,'cp -361',NULL);
INSERT INTO evaluations VALUES(53684,34,'cp -386',NULL);
INSERT INTO evaluations VALUES(53684,35,'cp -375',NULL);
INSERT INTO evaluations VALUES(53684,36,'cp -396','f3e5 c6d5 f1b5 c8d7 e5d7 b8d7 e4d5 d8e7 d1e2 e8c8 e1g1 g8f6 e2d3 e7e4 d3e4 f6e4 b5e2 f8d6 d2d3 e4f6 g2g3 f6d5 e2f3 d5c7 f3g2 h8e8 f1d1 g7g6 c1g5 f7f6 g5d2 f6f5 d2g5 d6e7 g5d2 e7f6 d1e1 d7e5 g1f1 e5g4 h2h4 e8e1 a1e1 d8e8 e1e8 c7e8');
INSERT INTO evaluations VALUES(56953,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(56953,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(56953,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(56953,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(56953,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(56953,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(56953,36,'cp 38','b1c3 d7d5 c4d5 e6d5 c1f4 c7c6 e2e3 f8d6 f4d6 d8d6 f1d3 g8e7 d1f3 b8d7 g1e2 e8g8 a1c1 d7f6 h2h3 a7a5 e1g1 g7g6 f3g3 d6g3 e2g3 h6h5 h3h4 a5a4 f1e1 f6d7 f2f3 b7b5 e3e4 d7b6 g3e2 b5b4');
INSERT INTO evaluations VALUES(56951,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(56951,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(56951,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(56951,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(56951,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(56951,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(56951,36,'cp 39','d7d5 c4d5 e6d5 c1f4 c7c6 e2e3 f8d6 f4d6 d8d6 g1f3 g8e7 d1b3 b7b6 f1d3 e8g8 a1c1 c8g4 f3e5 g4f5 e1g1 b8d7 d3f5 e7f5 e5f3 f5e7 h2h3 f8e8 f1d1 a7a5 f3e1 b6b5 e1d3 a5a4 b3c2 d6g6 b2b3 a4b3 a2b3 d7f6');
INSERT INTO evaluations VALUES(55459,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(55459,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(55459,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(55459,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(55459,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(55459,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(55459,36,'cp 41','c1f4 f8d6 f4d6 d8d6 c4d5 e6d5 e2e3 g8e7 f1d3 e8g8 g1e2 c7c6 d1c2 b7b6 e1g1 f8e8 h2h3 c8a6 e2f4 a6d3 f4d3 b8d7 a1c1 c6c5 d4c5 b6c5 f1d1 a8d8 b2b4 c5c4 d3f4 d7f6 a2a3 d6b6 c2a4 g8h8 c3e2');
INSERT INTO evaluations VALUES(50211,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(50211,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(50211,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(50211,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50211,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50211,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50211,36,'cp 0','f8g7 g2g3 d7d5 f1g2 e8g8 e1g1 b8d7 c1f4 b7b6 a2a4 a7a5 f1e1 c8b7 b1a3 f6e4 a1c1 e7e6 a3b5 a8c8 b5a7 c8a8');
INSERT INTO evaluations VALUES(56411,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(56411,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(56411,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(56411,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(56411,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56411,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(56411,36,'cp -42','e5f4 e4d5 g8f6 c2c4 c7c6 d2d4 f8b4 b1c3 c6d5 f1e2 b8c6 e1g1 c8e6 d1b3 e8g8 c4c5 f6e4 c1f4 e4c5 d4c5 d5d4 b3c2 d4c3 f3g5 g7g6 g5e6 f7e6 f4d6 f8f1 a1f1 d8g5 g1h1 b4c5 d6f4 g5d5 b2c3 a8f8 f1d1 d5f5 c2f5 f8f5 g2g3 g6g5 f4d6 c5d6 d1d6 f5c5 d6d7 c5c3 d7b7');
INSERT INTO evaluations VALUES(18739,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(18739,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(18739,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(18739,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(18739,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(18739,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(18739,36,'cp -50','d7d5 c4b5 c8d7 e4d5 e6d5 g1f3 f8d6 e1g1 g8e7 d3d4 c5d4 c3d4 e8g8 b1c3 d7g4 b5e2 h7h6 h2h3 g4e6 e2d3 e7g6 c1e3 d8f6 c3b5 d6b8 b5c3 f8d8 f1e1 b8d6 c3b5 d6b4 e3d2 b4d2 d1d2 g6f4 f3e5 f4d3 d2d3');
INSERT INTO evaluations VALUES(20242,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(20242,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(20242,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(20242,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(20242,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(20242,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(20242,36,'cp -35','c4b5 c8d7 e4d5 e6d5 g1f3 f8d6 d3d4 c5d4 e1g1 g8e7 c3d4 e8g8 b1c3 d7g4 c1g5 f7f6 g5e3 e7f5 a2a3 g8h8 b5e2 f5e3 f2e3 c6e7 d1b3 g4h5 g2g3 a8c8 c3d5 e7d5 b3d5 h5f7 d5e4 f7g6 e4d5');
INSERT INTO evaluations VALUES(31034,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(31034,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(31034,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(31034,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(31034,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(31034,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(31034,36,'cp 178','d3g6 h7g6 e5e6 g8h6 g1f3 d8d6 c1h6 h8h6 e6f7 e8f7 e1g1 f7g8 b1d2 b8d7 f1e1 g6g5 d2f1 d6g6 d1d2 g5g4 f3e5 d7e5 e1e5 e7e6 e5g5 g6h7 g5g4 c6c5 c2c3 c5d4 c3d4 a8c8 a1c1 c8c1 d2c1 h6f6 f1g3 h7h6 c1h6 f6h6 g3e2 f8e7 g2g3 e7f6 h2h3 g8f7 g1g2 h6h8 h3h4 h8e8 h4h5');
INSERT INTO evaluations VALUES(31026,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(31026,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(31026,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(31026,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(31026,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(31026,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(31026,36,'cp 157','h7g6 e5e6 g8h6 g1f3 d8d6 c1h6 h8h6 e6f7 e8f7 e1g1 f7g8 b1d2 b8d7 f1e1 g6g5 d2f1 d6f4 f1e3 g5g4 f3h4 e7e5 g2g3 f4d4 e3g4 h6h5 d1d4 e5d4 h4f3 c6c5 g1g2 d7f6 g4f6 g7f6 e1e6 g8f7 a1e1 a8d8 g3g4 h5h8 h2h4 c5c4 f3d4 h8h4 g2g3 h4h7 f2f4 d8d6 e6e8 d6b6 b2b3');
INSERT INTO evaluations VALUES(30971,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(30971,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(30971,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(30971,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(30971,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(30971,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(30971,36,'cp 169','e5e6 g8h6 g1f3 d8d6 c1h6 h8h6 e6f7 e8f7 e1g1 f7g8 f1e1 b8d7 b1d2 g6g5 d2f1 d6f4 f1e3 g5g4 f3h4 e7e5 g2g3 f4d4 e3g4 h6h5 d1d4 e5d4 h4f3 c6c5 e1e6 h5f5 g1g2 f5f7 a1e1 d7f6 g4f6 f7f6 h2h4 f6e6 e1e6 a8d8 g2f1 g8f7 f3g5 f7g8 f2f4');
INSERT INTO evaluations VALUES(16331,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(16331,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(16331,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(16331,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(16331,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(16331,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(16331,36,'cp -26','f6e4 d2d4 e5d4 f1d3 e4c5 e1g1 h7h6 f1e1 c5e6 c1d2 g7g6 c2c3 f8g7 c3d4 e8g8 d2e3 b7b6 d3e4 c8b7 d5c3 d7d5 e4d5 a8b8 h2h4 c6d4 f3d4 b7d5 c3d5 d8d5 d4e6 d5e6 e3h6 e6f6 h6g7 g8g7 d1c2 f6h4 c2c7 f8c8 c7e5 h4f6 e5f6 g7f6 a1d1 c8c2 d1d7');
INSERT INTO evaluations VALUES(16764,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(16764,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(16764,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(16764,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(16764,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(16764,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16764,36,'cp -18','d2d4 d7d6 f1b5 e4f6 c1g5 f8e7 d5e7 d8e7 b5c6 b7c6 d4e5 h7h6 g5h4 d6e5 e1g1 e8g8 f1e1 c8g4 d1d3 g4f3 d3f3 e7e6 h4f6 e6f6 f3f6 g7f6 e1e4 f8d8 a1e1 d8d5 b2b3 d5c5 e1e2 a7a5 f2f4 a8e8 g1f2 g8g7 e4a4 g7g6 f2f3 h6h5');
INSERT INTO evaluations VALUES(26230,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26230,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26230,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(26230,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26230,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(26230,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(26230,36,'cp 0','e7e6 c1g5 g8f6 e1c1 f8e7 h1e1 e8g8 e4e5 d6e5 f3e5 d8c7 d4c4 h7h6 e5c6 b7c6 g5f4 c7b6 f4e3 b6a5 c4a4 a5c7 a4f4 c7b7 e3d4 f8d8 f4g3 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(45435,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(45435,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(45435,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(45435,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(45435,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(45435,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(45435,36,'cp -28','c8g4 g1e2 e7e6 h2h3 g4f5 c2c3 b8c6 e1g1 f8e7 c1g5 e8g8 b3c2 f5g6 e2f4 g6c2 d1c2 d8c7 c2e2 h7h6 g5f6 e7f6 g2g3 f8d8 b1a3 f6e7 a3c2 e7d6 f4g2 e6e5 d4e5 d6e5 f2f4 e5d6 a1e1');
INSERT INTO evaluations VALUES(29162,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(29162,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(29162,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(29162,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(29162,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(29162,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(29162,36,'cp -25','g1e2 e7e6 h2h3 g4f5 e1g1 f8e7 c2c3 b8c6 b3c2 f5c2 d1c2 b7b5 a2a3 e8g8 c1f4 d8b6 b1d2 f8c8 c2d3 c6a5 b2b4 a5c4 d2c4 d5c4 d3g3 f6d5 f4e5 f7f6 e5f4 b6c6 f1e1 c6d7 h3h4 g8h8 h4h5 h7h6 g3g6 e7d6 f4d6 d7d6');
INSERT INTO evaluations VALUES(49531,30,'cp -186',NULL);
INSERT INTO evaluations VALUES(49531,31,'cp -173',NULL);
INSERT INTO evaluations VALUES(49531,32,'cp -172',NULL);
INSERT INTO evaluations VALUES(49531,33,'cp -189',NULL);
INSERT INTO evaluations VALUES(49531,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(49531,35,'cp -191',NULL);
INSERT INTO evaluations VALUES(49531,36,'cp -196','f6e4 d2d3 e4f6 d1e2 b8c6 g1f3 d7d6 d3d4 e5e4 d4d5 c6a5 c4b5 c7c6 d5c6 a5c6 f3d4 c8d7 f2f3 c6d4 c3d4 d6d5 b1c3 f8e7 e1g1 e8g8 c1f4 e4f3 e2f3 d7b5 c3b5 d8b6 a2a4 a7a6 b5c7 a8c8 g1h1 b6d4 a1e1 e7d8 f4e5');
INSERT INTO evaluations VALUES(49886,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(49886,31,'cp -201',NULL);
INSERT INTO evaluations VALUES(49886,32,'cp -192',NULL);
INSERT INTO evaluations VALUES(49886,33,'cp -200',NULL);
INSERT INTO evaluations VALUES(49886,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(49886,35,'cp -159',NULL);
INSERT INTO evaluations VALUES(49886,36,'cp -169','d2d3 e4f6 d1e2 b8c6 g1f3 d7d6 d3d4 e5e4 d4d5 c6a5 c4d3 d8e7 d3b5 e8d8 f3d4 a7a6 b5a4 f6d5 e1g1 d5b6 a4c2 c8d7 c2e4 d6d5 e4c2 e7e2 d4e2 f8d6 e2f4 c7c5 f1d1 d7g4 f2f3 g4e6 f4d3 a8c8 c1e3 b6d7 d3f4 d7f6 f4e6 f7e6 e3f2 d8c7 b1d2 a5c6 d2b3');
INSERT INTO evaluations VALUES(50160,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(50160,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50160,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(50160,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50160,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(50160,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(50160,36,'cp 11','e2e3 f8g7 f1e2 d7d6 g1f3 e8g8 e1g1 c7c5 b4c5 d6c5 c2c4 b7b6 d2d4 c5d4 e3d4 c8b7 b1c3 b8c6 d4d5 c6a5 a1c1 a8c8 c3b5 a7a6 b5d4 b6b5 c4b5 c8c1 d1c1 d8b6 b5a6 b7a6 e2a6 b6a6 c1c5 a5c4 b2c1 c4b6 c5e7 f6d5 e7e2 a6e2 d4e2');
INSERT INTO evaluations VALUES(50158,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50158,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50158,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(50158,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(50158,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(50158,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(50158,36,'cp 0','f8g7 g1f3 e8g8 f1e2 d7d6 d2d4 c7c5 b4c5 d6c5 c2c3 b7b6 b1d2 c8b7 e1g1 b8d7 f1e1 a8c8 a2a4 d8c7 d1b3 e7e5 f3e5 d7e5 d4e5 c7e5 c3c4 e5e6 a4a5 c8d8 e1d1 d8d7 a5b6 a7b6 a1a7 f8d8 d2f3 d7d1 e2d1');
INSERT INTO evaluations VALUES(43140,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43140,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(43140,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(43140,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(43140,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(43140,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(43140,36,'cp 1','g1f3 d7d6 f1e2 e8g8 e1g1 e7e5 d2d3 f6d5 b4b5 d5b6 a2a4 a7a6 c2c4 a6b5 a4b5 a8a1 b2a1 c8f5 d1c2 b8d7 b1c3 d7c5 h2h3 d8e7 e3e4 f5e6 a1b2 f8a8 f1a1 a8a1 b2a1 c5d7 c3d5 e6d5 c4d5');
INSERT INTO evaluations VALUES(51799,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(51799,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(51799,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(51799,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(51799,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(51799,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(51799,36,'cp 92','e2e4 d7d5 b1d2 g8f6 f1d3 c6c5 e4e5 f6d7 c2c3 b8c6 e1g1 a7a5 d3c2 c5d4 c3d4 b7b5 f1e1 a5a4 d2f1 d7b6 f1g3 b6c4 g3h5 c6b4 c2b1 h7h6 d1e2 d8b6 c1e3 a4a3 b2b3 c4e3 f2e3 c8a6 e1c1 e8d8 e2d2 f8e7');
INSERT INTO evaluations VALUES(54790,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(54790,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54790,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(54790,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(54790,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54790,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(54790,36,'cp 58','f1c4 e7e6 g5e6 c8e6 c4e6 d8f6 e6c8 b8c6 c8b7 a8b8 b7c6 f6c6 e1g1 g8f6 d2d3 f8d6 d3e4 f5e4 b1a3 d6a3 b2a3 e8g8 d1e2 b8e8 a1b1 f6d5 b1b3 d5f4 c1f4 f8f4 e2e3 e8f8 e3a7 c6c2 b3b8 c2a2 b8f8 f4f8 a7c7');
INSERT INTO evaluations VALUES(54788,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(54788,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(54788,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(54788,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(54788,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(54788,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(54788,36,'cp 49','e7e6 g5e6 c8e6 c4e6 f8c5 d2d4 d8d4 d1d4 c5d4 c2c3 d4b6 e6f5 g8f6 b1d2 e8g8 d2e4 f6e4 f5e4 b6f2 e1d1 c7c6 d1c2 b8d7 h1d1 d7e5 a2a4 a8e8 a4a5 e5c4 e4d3 c4e5 d3f1 b7b5 a5b6 a7b6 h2h3 h7h6 b2b4 b6b5 c2b3 f2b6 d1e1');
INSERT INTO evaluations VALUES(57200,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(57200,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(57200,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(57200,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(57200,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(57200,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(57200,36,'cp -47','d7d5 e2e4 d5d4 c3e2 g8e7 f2f4 b8c6 g2g3 e7g6 g1f3 e5f4 e2f4 f8b4 c1d2 a7a5 f1g2 e8g8 e1g1 g6e5 f3e5 c6e5 a2a4 b4d2 d1d2 c7c6 h2h3 h7h6 d2f2 d8d6 b2b3 c8e6 g1h1 f8e8 f4e2 c6c5 h1g1 e6d7 e2f4');
INSERT INTO evaluations VALUES(56440,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(56440,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(56440,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(56440,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(56440,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(56440,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(56440,36,'cp -47','e2e4 d5d4 c3e2 g8e7 f2f4 e5f4 g1f3 c8g4 e2f4 e7g6 a2a3 f8d6 g2g3 e8g8 h2h3 g4d7 f1g2 c7c5 e1g1 b8c6 d1e1 d8c7 f4d5 c7b8 c1f4 g6f4 g3f4 d6f4 d5f4 b8f4 f3d4 f4c7 d4c6 d7c6 f1f2 b7b6 e1d2 f7f6 a1f1');
INSERT INTO evaluations VALUES(3628,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3628,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(3628,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(3628,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3628,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(3628,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3628,36,'cp 13','c3a4 c5b6 h2h3 c8e6 c4e6 f7e6 b2b4 d8d7 c2c3 f8f7 a4b6 a7b6 a3a4 c6e7 c1e3 e7g6 f3h2 d6d5 h2g4 f6g4 d1g4 g6f4 e3f4 f7f4 g4h5 d5e4 h5e5 d7d6 e5d6 c7d6 f1d1 e4d3 d1d3 d6d5 d3e3 d5d4 c3d4 f4d4 e3e6 d4b4');
INSERT INTO evaluations VALUES(10863,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(10863,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(10863,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(10863,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(10863,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(10863,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(10863,36,'cp 58','f1e1 d7d6 c2c3 e8g8 d2d4 c8d7 b5a4 d7g4 a4c6 b7c6 d4e5 g4f3 d1f3 d6e5 e1d1 d8b8 b1d2 f8d8 d1e1 f6d7 a2a4 a7a5 d2c4 b8b3 f3e2 f7f6 c1e3 e7c5 e3c5 d7c5 a1a3 b3b7 c4e5 f6e5 e2c4 g8h8 c4c5 b7b2 a3a1 d8d2 e1f1 h8g8 c5c4 g8h8 c4c6 a8f8 c6c5 h8g8 c5a5 d2f2 a5d5 g8h8');
INSERT INTO evaluations VALUES(10864,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(10864,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(10864,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(10864,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(10864,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(10864,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(10864,36,'cp 54','d7d6 c2c3 e8g8 d2d4 c8g4 b5c6 b7c6 d4e5 g4f3 d1f3 d6e5 b1d2 d8d6 d2c4 d6e6 b2b3 a7a5 c1g5 a5a4 g2g3 g7g6 h2h4 f8b8 g5f6 e7f6 a1b1 a4b3 a2b3 h7h5 f3e2 g8g7 g1g2 f6e7 e1d1 g7g8 d1d2 b8d8 d2d8 e7d8 c4e3 d8f6 b3b4 a8a3 b1c1 g8g7 e2c4');
INSERT INTO evaluations VALUES(14247,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(14247,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(14247,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(14247,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(14247,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14247,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14247,36,'cp 0','e2e3 c8g4 a2a3 e7e6 f1d3 a7a6 b1c3 g4f5 h2h3 f8e7 c1d2 f6e4 e1g1 e4g3 f1e1 g3e4 d1e2 e4g3');
INSERT INTO evaluations VALUES(29968,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(29968,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(29968,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(29968,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(29968,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(29968,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(29968,36,'cp 165','f2f3 g4e6 c1e3 b8c6 g2g4 c6d4 d1d4 h7h6 h2h3 c7c5 f1b5 f6d7 d4d3 a7a6 b5d7 d8d7 e1c1 b7b5 c3d5 d7b7 c1b1 e6d5 d3d5 b7d5 d1d5 f7f6 c2c3 e8f7 e3f4 a8d8 h1d1 g7g5 f4d6 f8d6 d5d6 f7e7 d6d8 h8d8 d1d8 e7d8 b1c2 a6a5 c2d3 d8d7 d3e3 a5a4 f3f4 d7e8 e3f3 e8d7 e4e5 f6e5');
INSERT INTO evaluations VALUES(29966,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(29966,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(29966,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(29966,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(29966,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(29966,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(29966,36,'cp 150','g4e6 c1e3 b8c6 g2g4 c6d4 d1d4 h7h6 h2h3 c7c5 f1b5 f6d7 d4d3 a7a6 b5d7 d8d7 e1c1 d7c6 f3f4 b7b5 c3d5 e6d5 e4d5 c6d7 h1e1 e8c8 b2b4 f8e7 d3a3 c5b4 a3a6 d7b7 a6b7 c8b7 e3d4 e7h4 e1e3 h6h5 g4g5 d8e8 e3b3 e8e4 b3b4 e4f4 b4b5 b7c7 c1b1 h4g5');
INSERT INTO evaluations VALUES(39502,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(39502,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(39502,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(39502,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(39502,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(39502,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(39502,36,'cp 74','d2d4 d7d5 e4e5 c7c5 c2c3 c8d7 f1d3 b8c6 e1g1 c5d4 c3d4 d8b6 d3c2 c6b4 c2b3 d7b5 f1e1 b4d3 e1e3 d3c1 d1c1 g8h6 b3c2 a8c8 c1d2 b5c4 b1c3 e8g8 h2h3 g7g6 g2g4 g8g7 g1g2 h6g8 a1b1 b6d8 a2a3 h7h6 b1h1 e7g5 f3g5');
INSERT INTO evaluations VALUES(48642,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(48642,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48642,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48642,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(48642,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48642,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48642,36,'cp 0','c7c5 e2e3 c5d4 d1d4 b8c6 d4d2 f8c5 g1f3 e8g8 f1e2 d8e7 e1g1 f8d8 f3d4 c6d4 e3d4 f6e4 d2e3 c5d4 e3d4 e4c3 d4c3 e7e2 f1e1 e2g4 f4c7 d8f8 c7d6 f8d8 d6c7');
INSERT INTO evaluations VALUES(46419,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46419,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(46419,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(46419,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46419,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(46419,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(46419,36,'cp 2','e2e3 c5d4 d1d4 b8c6 d4d2 f8c5 f1d3 e8g8 g1e2 d5d4 e3d4 c6d4 e2d4 d8d4 e1g1 f8d8 f4e3 d4d3 d2d3 d8d3 e3c5 c8f5 a1d1 d3d1 f1d1 f6e4 c5b4 e4c3 b4c3 f7f6 a2a4 a8e8 f2f3 e8e2 d1d5 f5e6 d5d6 g8f7 a4a5 e2c2 a5a6');
INSERT INTO evaluations VALUES(53665,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(53665,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(53665,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(53665,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(53665,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(53665,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(53665,36,'cp 108','b8d7 f1d3 g8f6 d1e2 f6e4 e2e4 d7f6 e4e2 c6c5 c1f4 f8d6 f4e5 c5d4 e5d4 d8a5 c2c3 c8d7 f3e5 d7c6 e1g1 e8g8 e5c6 b7c6 d4f6 g7f6 d3h7 g8g7 h7c2 f6f5 f1d1 a5c7 g2g3 d6e7 d1d2 f8d8 a1d1 d8d2 d1d2 a8d8 d2d8 c7d8 e2c4');
INSERT INTO evaluations VALUES(1492,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(1492,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(1492,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(1492,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(1492,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(1492,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(1492,36,'cp -13','c8d7 f1e2 c5d4 c3d4 g8e7 b1a3 e7f5 a3c2 c6a5 c1d2 a5c4 e2c4 d5c4 d2c3 d7c6 d1e2 f8e7 e1g1 e8g8 c2e3 c6f3 e2f3 f5d4 f3g4 a8d8 f1d1 d4b3 a1b1 b3a5 e3c4 a5c4 g4c4 d8c8 c4g4 f8d8 g4f3 b6a6 d1d8 c8d8 b1d1 d8d1 f3d1');
INSERT INTO evaluations VALUES(27536,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27536,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27536,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(27536,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(27536,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27536,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27536,36,'cp 0','f1e2 c5d4 c3d4 g8e7 b1a3 e7f5 a3c2 a8c8 e1g1 c6a5 g2g4 f5h6 f3d2 a5c6 d2b1 c6a5');
INSERT INTO evaluations VALUES(4744,30,'cp 315',NULL);
INSERT INTO evaluations VALUES(4744,31,'cp 325',NULL);
INSERT INTO evaluations VALUES(4744,32,'cp 315',NULL);
INSERT INTO evaluations VALUES(4744,33,'cp 337',NULL);
INSERT INTO evaluations VALUES(4744,34,'cp 308',NULL);
INSERT INTO evaluations VALUES(4744,35,'cp 290',NULL);
INSERT INTO evaluations VALUES(4744,36,'cp 318','c3d5 g7e5 f2f4 e5c7 d1f3 c7a5 c2c3 d7d6 e1c1 c8d7 d4b3 a5c7 e4e5 d7e6 f1c4 g8h6 h2h3 h6f5 e3f2 e8g8 g2g4 f5g7 f2h4 g8h8 h1e1 d6e5 d5e7 d8e8 e7c6 e8c6 f3c6 b7c6 c4e6 g7e6 h4f6 h8g8 f4e5 f8e8 b3d4 e6d4 d1d4 c7d8 g4g5 h7h6 f6d8 a8d8 d4d8');
INSERT INTO evaluations VALUES(19864,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19864,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(19864,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(19864,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19864,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19864,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19864,36,'cp 24','b1c3 g8f6 c1f4 c8f5 e2e3 e7e6 f1e2 f8b4 e1g1 e8g8 a1c1 a8c8 h2h3 b4e7 a2a3 h7h6 b2b4 e7d6 f4d6 d8d6 d1d2 f5e4 f3h2 e4g6 d2b2 d6e7 h2f3 g6h5 f1d1 f8d8 g2g4 h5g6 c3a4 b7b6 e2a6');
INSERT INTO evaluations VALUES(35171,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(35171,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(35171,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(35171,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(35171,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(35171,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(35171,36,'cp 47','g8f6 b1c3 d7d6 d2d3 h4g4 f3g4 c8g4 f2f4 e5f4 c1f4 f8e7 c4b3 e8g8 h2h3 g4e6 e1c1 b8d7 g1f3 d7c5 b3e6 c5e6 f4h2 f8e8 h1e1 h7h6 a2a3 a7a6 a3a4 a8d8 d3d4 e6f8 c1b1 c7c6 a4a5 d6d5 e4e5 f6h7 h2g3 e7b4');
INSERT INTO evaluations VALUES(33967,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(33967,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(33967,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33967,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(33967,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(33967,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(33967,36,'cp 50','b1c3 d7d6 d2d3 h4g4 f3g4 c8g4 f2f4 e5f4 c1f4 f8e7 c4b3 e8g8 h2h3 g4e6 e1c1 b8a6 d3d4 e6b3 a2b3 a6b4 g1f3 f6h5 f4d2 f7f5 e4e5 d6e5 d4e5 a8d8 c3b5 h5g3 h1e1 g3e4 d2b4 e7b4 c2c3 d8d1 e1d1 c7c6 c3b4 c6b5 d1d7');
INSERT INTO evaluations VALUES(46679,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(46679,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(46679,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(46679,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(46679,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(46679,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(46679,36,'cp 51','e2e4 c7c6 d1d2 d8a5 g1f3 c8g4 f1e2 b8d7 h2h3 g4f3 e2f3 f8g7 e1c1 e8g8 c1b1 f8d8 c3d5 a5d2 d5f6 g7f6 d1d2 d7b6 h1d1 e7e6 b2b3 d6d5 e4e5 f6e7 b1b2 a7a5 a2a4 g8g7 g2g3 h7h5 h3h4 d8d7 f3g2 a8b8 b2b1 d7c7 d2d3 c6c5 d4c5 e7c5 d3c3');
INSERT INTO evaluations VALUES(6430,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(6430,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(6430,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(6430,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(6430,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6430,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(6430,36,'cp 17','d2d4 e5d4 c4f7 e8f7 d1h5 g7g6 h5c5 g8f6 g1f3 d7d5 c3d4 f6e4 c5a3 d8d6 e1g1 d6a3 b1a3 c8g4 f3e5 c6e5 f2f3 e5d3 f3e4 f7g7 c1g5 h8f8 e4d5 h7h6 g5h4 d3f4 a3b5 g6g5 h4g3 f4e2 g1h1 f8f1 a1f1 e2g3 h2g3 c7c6 b5c7 a8d8 c7e6 g4e6 d5e6 d8d4 f1f7 g7g6 f7b7 g6f6');
INSERT INTO evaluations VALUES(36336,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(36336,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(36336,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36336,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(36336,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(36336,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(36336,36,'cp 24','e1g1 d5c4 d1c2 c7c6 c2c4 b7b5 c4d3 b8d7 c1g5 c8b7 f3e5 d7e5 d4e5 f6d5 g5e7 d8e7 b1c3 d5c3 d3c3 f8c8 b2b4 a7a5 a2a3 h7h6 f1c1 a5b4 a3b4 a8a4 a1b1 c6c5 g2b7 e7b7 b4c5 b7c6 c3b3 a4c4 c1c4 b5c4 b3c4 c6c5 c4e4 c5c2 e4c2');
INSERT INTO evaluations VALUES(38348,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(38348,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(38348,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(38348,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(38348,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(38348,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(38348,36,'cp 25','g8f6 d2d4 e8g8 c4b3 c8g4 c1e3 b8c6 h2h3 g4f3 d1f3 e7e5 d4e5 d6e5 c3d5 f6d5 b3d5 c6d4 e3d4 e5d4 e1g1 a8b8 a2a4 c7c6 d5c4 d4d3 a1d1 g7b2 d1d3 d8f6 f1b1 b7b5 f3f6 b2f6 a4b5 c6b5 b1b5 b8b5 c4b5 f8d8 f2f4 d8d3 c2d3 f6d4 g1h2 g8g7 h2g3 f7f6 h3h4 d4c3 b5c4 h7h5 g3f2 a7a5 c4b3 g6g5 f4g5');
INSERT INTO evaluations VALUES(43646,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(43646,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(43646,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43646,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(43646,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(43646,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(43646,36,'cp 42','c4d5 e6d5 c1g5 c7c6 e2e3 h7h6 g5h4 c8e6 d1b3 b7b5 f1d3 b8d7 e1g1 f8d6 c3e2 e8g8 f1c1 d8b6 h4g3 d6g3 e2g3 a6a5 b3a3 b5b4 a3a4 f8c8 g3f5 e6f5 d3f5 c8c7 a2a3 c6c5 a3b4 c5b4 f5d3 g7g6 c1c7 b6c7');
INSERT INTO evaluations VALUES(43638,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(43638,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(43638,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(43638,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(43638,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43638,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(43638,36,'cp 31','e6d5 c1g5 f8e7 e2e3 h7h6 g5h4 e8g8 f1d3 c8e6 e1g1 b8d7 h2h3 c7c6 h4g3 e7b4 a1c1 f8e8 d3b1 b4c3 b2c3 f6e4 g3h2 e6f5 c3c4 d5c4 c1c4 a8c8 c4c1 b7b5 f1e1 c6c5 f3e5 c5d4 d1d4 d7e5 c1c8 d8c8 h2e5 f7f6');
INSERT INTO evaluations VALUES(45195,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(45195,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45195,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45195,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45195,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45195,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45195,36,'cp 0','g2g3 b8c6 f1g2 e7e6 f2f4 a8b8 a2a4 g7g6 g1h3 f8g7 e1g1 a7a6 h3f2 e8g8 c3e2 b7b5 a4b5 a6b5 g3g4 h7h6 c1e3 b5b4 d1d2 c8b7 c2c3 c6a5 d2c2 a5c6');
INSERT INTO evaluations VALUES(51010,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(51010,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(51010,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(51010,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(51010,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(51010,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(51010,36,'cp 102','d7d6 f1d3 g8f6 d1e2 f8e7 e1g1 b8c6 e4e5 f6d5 a2a3 d5c3 b2c3 d6d5 f3d2 c6a5 d2b3 a5c4 e2g4 g7g6 a3a4 a6a5 g4f3 c8d7 d3c4 d5c4 b3d2 d7c6 f3e2 d8d7 e2c4 e7g5 c4d3 g5d2 c1d2 b7b6 f2f3 c6b7 d2e3 b7a6 d3e4 d7d5 e4d5 e6d5 f1c1 e8c8 g1f2');
INSERT INTO evaluations VALUES(51469,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(51469,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(51469,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(51469,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(51469,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(51469,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(51469,36,'cp 123','f8g7 b1c3 d7d6 e4e5 d6e5 f3e5 c7c6 d1f3 g8f6 c1f4 c8e6 h2h4 f6d5 h4h5 d5f4 f3f4 d8c7 c3b5 c6b5 f1b5 e8d8 e5f7 d8c8 f4c7 c8c7 f7h8 g6g5 e1c1 g7h8 d1e1 c7d6 b5d3 e6d5 d3h7 d5g2 h1g1 g2d5 c2c3 h8f6 h5h6 b8d7 h7f5 e7e6');
INSERT INTO evaluations VALUES(54834,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(54834,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(54834,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(54834,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(54834,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54834,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(54834,36,'cp 46','g8f6 d2d4 f8g7 f1e2 c7c6 a2a4 e8g8 c1e3 d8c7 e1g1 e7e5 d4e5 d6e5 a4a5 f6h5 g2g3 h5f6 g1g2 f8d8 d1c1 b8a6 e2a6 b7a6 e3g5 a8b8 b2b3 c6c5 f1d1 c8b7 g5f6 g7f6 c1e3 d8d1 a1d1');
INSERT INTO evaluations VALUES(55762,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(55762,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(55762,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(55762,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(55762,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(55762,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(55762,36,'cp -32','c7c5 a2a4 b8c6 f1c4 a7a6 b5a3 f8d6 d2d3 g8e7 e1g1 e8g8 c1d2 e7g6 g2g3 c8d7 h2h4 g6e7 h4h5 h7h6 f3h4 c6a7 c4b3 b7b5 a4b5 a6b5 d1e2 g8h8 c2c3 e7c6 c3c4 b5b4 a3b5 d6e7 h4f3 e6e5 b5a7 a8a7 a1a7 c6a7');
INSERT INTO evaluations VALUES(54052,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(54052,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(54052,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(54052,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(54052,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(54052,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(54052,36,'cp -14','a2a4 b8c6 f1c4 a7a6 b5a3 f8d6 d2d3 g8e7 e1g1 c8d7 c1d2 e7g6 c4b3 d6c7 c2c3 e8g8 c3d4 c5d4 a3c2 e6e5 c2b4 c6e7 b4d5 c7d6 a4a5 e7d5 b3d5 a8b8 d1e2 h7h6 f1c1 d8e7 e2d1 d7g4 h2h3 g4d7 d5c4 e7f6 c4d5 d7e6 d5e6 f6e6');
INSERT INTO evaluations VALUES(7000,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(7000,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(7000,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(7000,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(7000,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(7000,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(7000,36,'cp 85','g8f6 e4e5 d7d5 e5f6 d5c4 d4d8 e7d8 f6g7 h8g8 c1h6 d8f6 b1c3 c8e6 a1d1 f6g7 h6g7 g8g7 e1g1 g7g5 f2f3 a8d8 g1f2 h7h5 d1d8 e8d8 f1e1 g5c5 e1e4 a7a5 a2a3 e6f5 e4d4 d8e8 d4d2 f5e6 h2h4 e8e7 d2d4 b7b5 c3e2 b5b4 a3b4 a5b4 e2f4 b4b3 c2c3 c5a5');
INSERT INTO evaluations VALUES(10985,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(10985,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(10985,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(10985,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(10985,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(10985,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(10985,36,'cp 57','b1c3 c7c6 a2a4 e8g8 e1g1 d8c7 h2h3 c7d6 d4d6 e7d6 c4d3 f8e8 a4a5 d6e5 f1e1 d7d6 g2g4 c8d7 f2f4 e5c3 b2c3 c6c5 g4g5 c5c4 g5f6 c4d3 c2d3 d6d5 e4e5 g7f6 g1f2 d7h3 e1g1 g8h8 e5f6 e8e6 c1e3 h3f5 e3d4 f5d3 g1g7');
INSERT INTO evaluations VALUES(10984,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(10984,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(10984,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(10984,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(10984,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(10984,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(10984,36,'cp 58','c7c6 e1g1 e8g8 a2a4 d8c7 h2h3 f8e8 c4b3 d7d6 f1d1 c8e6 b3e6 f7e6 c1f4 e8d8 d4c4 c7c8 d1e1 d6d5 e4d5 c6d5 c4e2 e7b4 f4g5 d8f8 e2e6 c8e6 e1e6 b4c3 b2c3 a8c8 e6e7 f8f7 g5f6 g7f6 e7e3 f7c7 a1d1 c7c3 e3c3 c8c3 d1d5 b7b6 d5d7');
INSERT INTO evaluations VALUES(12838,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(12838,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(12838,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(12838,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(12838,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(12838,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(12838,36,'cp -2','c2c3 g7g6 b1d2 f8g7 d3d4 c5d4 c3d4 d7d5 e4e5 f6e4 f1d3 e4d2 c1d2 c8g4 d2c3 f7f6 e5f6 g7f6 e1g1 f6d4 c3d4 g4f3 d1f3 c6d4 f3e3 d4c6 d3b5 d8d6 f1e1 e8g8 b5c6 b7c6 e3e7 d6e7 e1e7 f8f7 e7e5 g8f8 f2f3 f7b7 b2b3');
INSERT INTO evaluations VALUES(12834,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(12834,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(12834,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(12834,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(12834,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(12834,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(12834,36,'cp -13','g7g6 b1d2 e7e5 d2c4 d7d6 c1g5 h7h6 g5f6 d8f6 a2a4 c8e6 g2g3 h6h5 f1g2 a8d8 d1e2 f8h6 c4e3 e8g8 e1g1 b7b6 f3d2 c6e7 a1e1 g8g7 h2h4 a7a6 e1a1 f8e8 e3c4 b6b5 a4b5 a6b5 c4e3 d8b8 g1h2 h6e3');
INSERT INTO evaluations VALUES(21678,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(21678,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(21678,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(21678,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(21678,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(21678,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(21678,36,'cp 137','d4e5 c8e6 g1f3 a7a6 c3e2 e6g4 e2f4 g8e7 f1e2 g4f3 e2f3 c6e5 f3d5 e7d5 d1d5 d8d5 f4d5 e8c8 d5f4 e5c4 f4d3 g7g6 e1e2 f8g7 a1b1 f7f5 b2b3 c4d6 c1b2 g7b2 b1b2 h8e8 c2c4 g6g5 g2g3 d6e4 b2c2 h7h6 b3b4 c7c6 h2h4 d8d7 b4b5 a6b5 c4b5 e8d8 h1d1 c8c7');
INSERT INTO evaluations VALUES(31933,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(31933,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(31933,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(31933,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(31933,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(31933,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(31933,36,'cp 38','c2c4 e7e6 h2h3 g4f3 d1f3 g8f6 b1c3 c7c6 d2d4 f8b4 f1d3 e8g8 e1g1 b8d7 f3e2 f8e8 c1d2 d7b6 c4c5 b6d7 a2a3 b4c3 d2c3 e6e5 d4e5 d7e5 d3c2 d8e7 b2b4 e5d7 c3d4 a7a5 e2d3 d7f8 d3c3 a5b4 a3b4 a8a1 f1a1 f8e6 d4f6 e7f6 c3f6');
INSERT INTO evaluations VALUES(31890,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(31890,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(31890,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(31890,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(31890,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(31890,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(31890,36,'cp 29','e7e6 h2h3 g4f3 d1f3 g8f6 b1c3 c7c6 d2d4 f8b4 f1d3 e8g8 e1g1 b8d7 c4d5 e6d5 a2a3 b4d6 g2g4 d6c7 f3g2 f8e8 g1h1 d7f8 c1d2 f8g6 f2f4 c7a5 b2b4 a5c7 g4g5 f6e4 c3e4 d5e4 d3c4 g6h4 g2g4 h4f3 d2c3 d8d6 h1g2 a8d8');
INSERT INTO evaluations VALUES(35971,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(35971,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(35971,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(35971,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(35971,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(35971,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(35971,36,'cp 71','e7e6 h2h3 e6d5 e4d5 c8f5 g2g4 f5e4 e1g1 e4f3 e2f3 b8d7 c1g5 h7h6 g5h4 f8e8 d1c2 g6g5 h4g3 d7e5 f3e2 f6d7 c3e4 d7f6 e4d2 b7b5 a1e1 a8b8 b2b3 b5c4 e2c4 f6d7 g1g2 d8c7 f2f4 g5f4 f1f4 e5c4 e1e8 b8e8 b3c4 d7e5 d2e4');
INSERT INTO evaluations VALUES(36088,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(36088,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(36088,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(36088,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(36088,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(36088,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(36088,36,'cp 74','c1e3 f8e8 f3d2 c8d7 e1g1 e6d5 c4d5 b7b5 d1c2 b5b4 c3b5 d7g4 f2f3 f6d5 e4d5 e8e3 d2c4 g4f5 c2d2 e3e7 b5d6 e7d7 d6f5 g6f5 d5d6 b8c6 g1h1 c6d4 e2d3 d4b5 d3f5 b5d6 a1d1 d6c4 d2d7 d8d7 d1d7 c4b2 d7b7 a8e8 g2g3 c5c4 b7b4 b2d3');
INSERT INTO evaluations VALUES(55650,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(55650,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(55650,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(55650,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(55650,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(55650,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(55650,36,'cp 34','e2e4 f8e7 g1e2 c7c5 h2h3 b8c6 e1g1 g8f6 c2c4 a7a6 a2a4 e8g8 b1c3 a8b8 f2f4 c8d7 d2d3 f6e8 b2b3 e8c7 f4f5 b7b5 a4b5 a6b5 g1h2 b5c4 b3c4 e7g5 f1f2 c6b4 c1d2 g5e7 d2e3 e7g5 e3g5 d8g5 a1a7 g5d8 f5f6');
INSERT INTO evaluations VALUES(55618,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55618,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55618,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(55618,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(55618,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55618,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(55618,36,'cp 27','f8e7 g1e2 c7c5 e1g1 g8f6 d2d3 b8c6 b1c3 c6d4 h2h3 c8d7 a2a4 e8g8 f2f4 d7c6 g1h2 f8e8 c1d2 b7b6 e2d4 c5d4 c3e2 a7a6 g3g4 f6d7 g4g5 b6b5 a4b5 a6b5 h3h4 a8a1 d1a1 h7h6 g5h6 e7h4');
INSERT INTO evaluations VALUES(55828,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(55828,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55828,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(55828,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55828,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(55828,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55828,36,'cp 0','g8f6 e2e3 c7c5 g1f3 b8c6 h2h3 c5d4 e3d4 d8b6 b1c3 c8d7 f1e2 c6e7 a1b1 e7g6 f4h2 f8e7 e1g1 e8g8 f1e1 d7c6 e2d3 a7a6 d1e2 a8c8 e2d2 b6d8 d2e3 d8b6');
INSERT INTO evaluations VALUES(48300,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48300,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48300,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(48300,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(48300,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(48300,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(48300,36,'cp 0','e2e3 c7c5 g1f3 c5d4 e3d4 d8b6 b1c3 c8d7 a1b1 b8c6 f1e2 c6e7 e2d3 e7g6 f4g3 f8e7 h2h4 f6h5 g3h2 e8g8 c3e2 h5f6 h2g3 f6h5');
INSERT INTO evaluations VALUES(48297,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48297,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48297,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(48297,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48297,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48297,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48297,36,'cp 0','c7c5 g1f3 b8c6 h2h3 a7a6 f1e2 b7b5 d4c5 f8c5 c2c3 e8g8 e1g1 f8e8 b2b4 c5f8 a3a4 b5a4 d1a4 c8d7 a4b3 d8b6 f4h2 c6e7 f3d4 e7c6');
INSERT INTO evaluations VALUES(46214,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46214,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46214,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46214,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46214,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46214,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(46214,36,'cp 12','g1f3 b8c6 h2h3 b7b6 c2c3 c8b7 f1b5 a7a6 b5d3 f8d6 b1d2 d6f4 e3f4 d8c7 g2g3 e8g8 e1g1 c6e7 d1e2 e7f5 f1e1 f5d6 f3e5 a6a5 d2f1 c7c8 e2d1 b7a6 d3c2 a6f1 g1f1 c5d4 c3d4 d6c4 b2b3 c4d6 c2d3 f6e4');
INSERT INTO evaluations VALUES(9261,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(9261,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(9261,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(9261,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(9261,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(9261,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(9261,36,'cp 50','c2c4 e8g8 b2b3 b7b6 b1c3 c8b7 c1g5 d6e7 g5f6 e7f6 c4d5 e6d5 e2e3 f8e8 h2h4 c6e7 a1c1 e7f5 b3b4 a7a6 a2a4 c7c6 f3e1 f5d6 e1d3 f6e7 d1c2 h7h6 c2d1 d6c4 b4b5 a8c8 b5c6 b7c6 d3f4 e7a3 c1c2');
INSERT INTO evaluations VALUES(9227,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(9227,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(9227,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(9227,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(9227,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(9227,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(9227,36,'cp 57','e8g8 b2b3 b7b6 c1g5 d6e7 b1c3 c8b7 c4d5 f6d5 c3d5 e6d5 g5e7 d8e7 a2a3 a7a5 d1c2 c6a7 f1c1 c7c6 a3a4 a8c8 h2h4 h7h6 f3e1 f8e8 e2e3 c8a8 e1d3 a7c8 c2d1 c8d6 c1c2 d6e4 d3f4 e4d6 a1c1');
INSERT INTO evaluations VALUES(17225,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(17225,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(17225,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(17225,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(17225,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(17225,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(17225,36,'cp 109','f6g8 b1c3 d7d6 f2f4 d6d5 c1e3 c8f5 f1e2 h7h6 g1f3 e7e6 a2a3 h6h5 e3f2 g8e7 d1d2 e7g6 e1g1 f8e7 f3e1 a7a6 a1c1 c6a7 e2d3 f5d3 e1d3 g6h4 d3c5 b7b6 f2h4 b6c5');
INSERT INTO evaluations VALUES(46994,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(46994,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(46994,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(46994,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46994,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(46994,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(46994,36,'cp 19','g7g6 b1c3 f8g7 f1e2 b8d7 g1f3 e8g8 e1g1 e7e5 e3e4 f8e8 d4d5 d7c5 d1c2 a7a5 c1e3 f6g4 e3c5 d6c5 g2g3 g4h6 g1g2 f7f5 a1e1 d8e7 e2d3 f5f4 h2h3 h6f7 g3g4 g8h8 f1h1 e8f8 a2a3 g7f6 h1h2 f7g5 d3e2 c7c6 c3a4');
INSERT INTO evaluations VALUES(46667,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46667,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46667,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(46667,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(46667,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(46667,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(46667,36,'cp 40','f1e2 f8g7 b1c3 e8g8 g1f3 b8d7 e1g1 e7e5 e3e4 f8e8 d4d5 d7c5 d1c2 a7a5 c1g5 b7b6 f3d2 h7h6 g5e3 c8g4 e2g4 f6g4 e3c5 b6c5 d2f3 h6h5 a1e1 g7h6 g2g3 d8d7 f3h4 a5a4 h4g2 f7f5 f2f3 g4f6 e4f5 g6f5 g1h1 e8f8');
INSERT INTO evaluations VALUES(55867,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(55867,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55867,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55867,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55867,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55867,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55867,36,'cp 0','c2c4 c7c5 g1f3 g8f6 c4d5 e6d5 g2g3 f8e7 f1g2 b8c6 e1g1 c8e6 d4c5 e7c5 f3g5 e6c8 b1c3 h7h6 g5f3 e8g8 c3a4 c5d6 a4c3');
INSERT INTO evaluations VALUES(36301,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36301,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36301,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36301,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36301,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(36301,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36301,36,'cp 2','b1c3 h7h6 g5f6 e7f6 a1c1 c7c6 f1d3 b8d7 e1g1 d5c4 d3c4 c6c5 d4c5 d7c5 b2b4 c5d7 c3e4 f6e7 b4b5 d7b6 c4e2 e7a3 c1c2 d8e7 d1a1 e6e5 a1e5 e7e5 f3e5 c8e6 e4c5 a8c8 c5e6 c8c2 e6f8 c2e2');
INSERT INTO evaluations VALUES(31479,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(31479,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(31479,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(31479,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(31479,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(31479,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(31479,36,'cp -5','c6c5 c1e3 d8b6 b1c3 b8c6 g1f3 g8h6 d4c5 f8c5 e3c5 b6c5 e1c1 h6g4 d3b5 c5b5 c3b5 g4f2 b5c7 e8d7 c7a8 f2h1 d1h1 h8a8 h1e1 b7b5 a2a3 a7a5 c1d2 f7f6 g2g3 h7h5 h2h3 b5b4 a3b4 a5b4 g3g4 a8f8 f4f5 f6e5 f3e5 c6e5 e1e5 e6f5 e5f5 f8f5 g4f5 d7d6 c2c3 b4c3 d2c3');
INSERT INTO evaluations VALUES(31641,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(31641,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(31641,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(31641,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(31641,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(31641,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(31641,36,'cp -5','c1e3 d8b6 b1c3 b8c6 d4c5 f8c5 e3c5 b6c5 g1f3 g8h6 e1c1 h6g4 d3b5 c5b5 c3b5 g4f2 b5c7 e8d7 c7a8 f2h1 a8b6 a7b6 d1h1 h8a8 a2a3 b6b5 c1d2 b5b4 a3b4 c6b4 c2c3 b4c6 d2d3 a8a2 d3c2 b7b5 g2g3 a2a8 h1d1 h7h6 f3d4 c6d4 d1d4 f7f6 c2d3 a8a1 d4b4 f6e5 f4e5 d7c6');
INSERT INTO evaluations VALUES(44216,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44216,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(44216,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44216,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44216,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(44216,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(44216,36,'cp 10','d4d5 g7g6 b1c3 f8g7 e2e4 d7d6 d1c2 a7a6 a3a4 e8g8 g1f3 a6a5 f1d3 b8a6 e1g1 a6b4 c2d1 b7b6 d3b1 e6d5 c4d5 c8a6 f1e1 f6d7 c1f4 d8e7 d1d2 d7e5 f3e5 d6e5 f4e3 a8d8 a1a3 h7h6 h2h3 f7f5 f2f3 g8h8 c3b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(44262,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44262,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(44262,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44262,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44262,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44262,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(44262,36,'cp 0','g7g6 b1c3 f8g7 g1f3 e8g8 e2e4 d7d6 d1c2 e6d5 c4d5 d8e7 f1d3 b8d7 e1g1 a7a6 a3a4 d7e5 f3e5 e7e5 f2f3 c8d7 g1h1 f6h5 c1d2 f7f5 e4f5 d7f5 d3f5 e5f5 c2f5 g6f5 g2g4 f5g4 f3g4 h5f6 d2f4 f6g4 f4d6 f8f1 a1f1 g7c3 b2c3');
INSERT INTO evaluations VALUES(11586,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(11586,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(11586,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11586,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11586,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(11586,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(11586,36,'cp 35','b5a4 e7g6 a2a3 f8c5 c1g5 c5e7 g5e7 g6e7 d3d4 e5d4 f3d4 e8g8 b1c3 d7d6 d4c6 e7c6 e1g1 c8e6 a4c6 b7c6 d1d4 a6a5 a1d1 c6c5 d4d3 c5c4 d3d4 a8b8 f2f4 f7f6 c3d5 e6d5 e4d5 f8e8 f1e1 h7h6 h2h3 a5a4');
INSERT INTO evaluations VALUES(11595,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(11595,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11595,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(11595,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(11595,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11595,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(11595,36,'cp 18','e7g6 a2a3 d7d5 e4d5 d8d5 a4b3 d5d6 f3g5 c6d8 b1c3 f8e7 g5e4 d6c6 d3d4 e5d4 d1d4 d8e6 d4d5 c6d5 c3d5 e7d8 c1e3 g6h4 e1g1 h4f5 f1d1 e8g8 d1d3 b7b5 a1e1 f5e3 d5e3 c7c6 b3e6 c8e6 e4c5 f8e8 c5e6 e8e6 e1d1 d8e7 g1f1 a8e8 b2b3 e7c5');
INSERT INTO evaluations VALUES(19255,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(19255,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(19255,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(19255,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(19255,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(19255,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(19255,36,'cp 32','e2e3 g8f6 d2d4 c5d4 e3d4 d7d5 c4d5 f6d5 d1b3 e7e6 f1b5 f8g7 e1g1 e8g8 b5c6 b7c6 f1e1 d8b6 b3c2 d5c3 b2c3 b6a5 a2a4 c6c5 c1a3 c8b7 a3c5 b7f3 g2f3 f8c8 a1b1 a5d8 c2e4 g7f8 c5f8 g8f8 e1c1 c8c4 e4e5 d8d5 e5d5 e6d5');
INSERT INTO evaluations VALUES(19254,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(19254,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(19254,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(19254,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(19254,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(19254,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(19254,36,'cp 43','g8f6 d2d4 c5d4 e3d4 d7d5 c4d5 f6d5 d1b3 e7e6 f1b5 f8g7 b5c6 b7c6 e1g1 e8g8 f1e1 d8c7 f3e5 f7f6 e5f3 a8b8 b3d1 d5c3 b2c3 c7a5 c1f4 b8b2 d1c1 b2a2 a1a2 a5a2 h2h4 c6c5 h4h5 g6h5 f4d6 f8d8 d6c5 a2d5 c3c4 d5f5 c1e3 d8e8 e3b3 e6e5 d4e5 f6e5 c5a7');
INSERT INTO evaluations VALUES(12718,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(12718,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(12718,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(12718,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(12718,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(12718,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(12718,36,'cp 51','d2d4 c5d4 e3d4 d7d5 c4d5 f6d5 d1b3 e7e6 f1b5 f8g7 b5c6 b7c6 e1g1 e8g8 f1e1 d8c7 h2h3 c6c5 c3d5 e6d5 b3d5 c8e6 d5c5 c7c5 d4c5 f8c8 c1e3 g7b2 a1d1 a7a5 e1e2 b2a3 f3g5 e6c4 e2c2 c4b5 d1d5 b5a4 c2c4 a4c6 d5d1 h7h6 g5f3 c6b5 c4g4 a3c5 e3h6 c5f8 h6f8 g8f8 g4f4 c8c2 a2a3');
INSERT INTO evaluations VALUES(12723,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(12723,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(12723,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12723,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12723,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(12723,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(12723,36,'cp 39','c5d4 e3d4 d7d5 c4d5 f6d5 d1b3 e7e6 f1b5 f8g7 e1g1 e8g8 b5c6 b7c6 f1e1 d8c7 h2h4 a8b8 c3d5 c6d5 b3e3 f8e8 b2b3 f7f6 c1a3 c7d7 a1c1 c8a6 e3f4 b8c8 c1c8 e8c8 f4d2 a6b5 a3c5 e6e5 a2a4 e5e4 a4b5 e4f3 g2f3 d7f5 b3b4 f5f3 e1e7 f3g4 g1h2');
INSERT INTO evaluations VALUES(41268,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(41268,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(41268,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(41268,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(41268,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(41268,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(41268,36,'cp -19','f6h5 f4g5 h7h6 g5h4 c7c5 b1c3 g6g5 h4g3 g5g4 f3h4 c5d4 e3d4 b8c6 d4d5 h5g3 h2g3 d8b6 a1b1 c6e5 f1e2 c8d7 e1g1 a8c8 b2b3 h6h5 a2a4 a7a6 f1e1 e5f3 g2f3 g7c3 e1f1 b6d4 d1d4 c3d4 f3f4 e8d8 a4a5 d4f6 b3b4 e7e5 d5e6 d7e6');
INSERT INTO evaluations VALUES(53458,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(53458,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(53458,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(53458,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(53458,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(53458,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(53458,36,'cp 62','f1c4 e7e6 b1c3 g8f6 g1f3 f8e7 e1g1 e8g8 f1e1 a7a6 f3e5 b8c6 e5c6 b7c6 c1f4 a6a5 a1c1 c8a6 c4b3 e7d6 f4e5 d6e5 d4e5 d8d1 e1d1 f6d5 c3a4 a6b5 d1d4 f8b8 a4c5 h7h6 g2g3 a5a4 b3a4 b5a4 d4a4 a8a4 c5a4 b8b4');
INSERT INTO evaluations VALUES(16053,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(16053,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(16053,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(16053,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(16053,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(16053,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(16053,36,'cp 175','f3g5 d7d5 e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5a4 f6d7 d2d3 h7h6 g5f3 f8g7 e1g1 e8g8 f1e1 d8c7 b2b4 a5b7 c1b2 a7a5 a2a3 b7d6 a4b3 f8e8 f3d2 d7f6 c3a4 f6d5 a4c5 g8h7 h2h3 d6b7 d3d4 b7c5 b4c5 c8f5 g2g4 f5d7 d4e5 g7e5');
INSERT INTO evaluations VALUES(16048,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(16048,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(16048,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(16048,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(16048,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(16048,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(16048,36,'cp 155','d7d5 e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5a4 f6d7 d2d3 h7h6 g5f3 f8g7 e1g1 e8g8 f1e1 d8c7 b2b4 a5b7 a2a3 b7d6 a4b3 a7a5 c1b2 f8e8 f3d2 g8h7 c3e4 d6e4 d3e4 d7f8 b4a5 a8a5 a3a4 f8e6 b2c3 a5a7 a4a5 e6d4 b3c4 d4b5 c3b4 e8d8 c2c3 b5d6 d1a4 d6c4 d2c4');
INSERT INTO evaluations VALUES(14514,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(14514,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(14514,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(14514,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(14514,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(14514,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(14514,36,'cp 172','e4d5 c6a5 c4b5 c7c6 d5c6 b7c6 b5a4 f6d7 d2d4 f8e7 g5f3 c8a6 f3e5 d7e5 d4e5 d8c7 f2f4 a8d8 d1f3 e8g8 a4b3 a5c4 e1g1 c4d2 c1d2 a6f1 f3f2 f1a6 d2e3 a6c8 c3e4 c8f5 e4c5 f8e8 h2h3 e7c5 e3c5 d8d7 c5d6 c7b6 c2c4 e8d8 c4c5 b6b5 a1d1 f5e6 b3e6 f7e6 d1d4 h7h6');
INSERT INTO evaluations VALUES(29899,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(29899,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(29899,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(29899,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(29899,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(29899,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(29899,36,'cp 36','f8e7 h2h3 g4e6 e1g1 e8g8 d1e2 b8c6 a2a4 c6d4 f3d4 e5d4 c4e6 f7e6 c3b1 d8e8 f2f4 e6e5 f4e5 e8h5 e2h5 f6h5 f1f8 a8f8 b1d2 d6e5 d2f3 e7d6 c1d2 h7h6 g1f2 g7g5 f2e2 f8f6 c2c3 d4c3 b2c3 h5g3 e2e1');
INSERT INTO evaluations VALUES(28380,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(28380,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(28380,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(28380,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(28380,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(28380,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(28380,36,'cp 29','h2h3 g4e6 d1e2 e8g8 a2a4 b8c6 e1g1 d8d7 c1g5 c6a5 c4a2 e6a2 c3a2 a5c6 a2c3 a8e8 f1e1 a7a6 e2d1 e7d8 g5f6 d8f6 c3d5 f6d8 c2c3 c6e7 d5e7 e8e7 d1b3 c7c6 a1d1 f8e8 d3d4 e5d4 d1d4 h7h6 e1d1 d7e6 b3e6 e7e6 d4d6');
INSERT INTO evaluations VALUES(28378,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(28378,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(28378,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(28378,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(28378,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28378,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(28378,36,'cp 26','g4e6 d1e2 e8g8 e1g1 b8c6 f1e1 f8e8 c1d2 c6d4 f3d4 e5d4 c3d5 f6d5 c4d5 e6d5 e4d5 e7f6 e2f3 d8d7 b2b3 a7a6 a2a4 e8e5 d2f4 e5e1 a1e1 b7b5 a4a5 a8e8 e1e8 d7e8 f4d2 f6e5 g1f1 e5f6 f3g4 g7g6');
INSERT INTO evaluations VALUES(38362,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(38362,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(38362,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(38362,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(38362,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(38362,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(38362,36,'cp 32','g1f3 g8f6 d1e2 e8g8 e1g1 c8g4 e4e5 d6e5 d4e5 f6d5 b1d2 d5b6 h2h3 b6c4 e2c4 g4f5 f3d4 f5d7 e5e6 b8c6 d2f3 c6d4 e6f7 f8f7 f3d4 g7d4 c4d4 d7b5 f1d1 d8d4 d1d4 a8e8 c1h6 e7e5 d4d5 a7a6 a1d1 f7f5 d5c5 b5c6 d1e1 e5e4');
INSERT INTO evaluations VALUES(43542,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(43542,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43542,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43542,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(43542,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43542,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(43542,36,'cp 0','f8e7 f2f4 e8g8 c1e3 e6e5 f4e5 d6e5 d4f5 c8f5 f1f5 b8d7 g1h1 a8c8 e2d3 e7a3 e3c1 a3b4 c1d2 d7c5 d2g5 b4c3 g5f6 g7f6 b2c3 c5d3 c2d3 c8c3 d1h5 g8h8 f5e5 c3d3 e5d5 d3d5 e4d5 d8d6 a1d1 b7b5 h2h3 h8g8 d1d3 g8h8');
INSERT INTO evaluations VALUES(39741,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(39741,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(39741,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39741,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39741,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39741,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(39741,36,'cp 3','c1e3 e8g8 g1h1 d8c7 a2a4 b7b6 f2f4 c8b7 e2f3 b8c6 f4f5 c6d4 d1d4 e6f5 e4f5 b6b5 f3b7 c7b7 d4b4 d6d5 b4b3 f8d8 a1d1 a8c8 a4b5 a6b5 e3d4 f6e4 c3e4 d5e4 f5f6 e7f6 d4f6 d8d1 f1d1 g7f6 b3g3 g8h8');
INSERT INTO evaluations VALUES(39734,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39734,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(39734,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(39734,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(39734,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(39734,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(39734,36,'cp 0','e8g8 a2a4 b8c6 d4b3 b7b6 f2f4 c8b7 d1e1 d8c7 e1g3 f6d7 f4f5 e7f6 f5e6 f7e6 a1d1 f6e5 g3h3 a8e8 e2h5 f8f1 d1f1 e8e7 h5g4 d7f8 e3g5 e7e8 g4h5 f8g6 h5g4');
INSERT INTO evaluations VALUES(35577,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35577,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(35577,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(35577,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(35577,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(35577,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35577,36,'cp 12','g1h1 d8c7 a2a4 b7b6 f2f4 c8b7 e2f3 b8c6 f4f5 c6d4 d1d4 e6f5 e4f5 b6b5 f3b7 c7b7 d4b4 d6d5 b4b3 b5b4 a1d1 a8d8 e3d4 f6h5 f5f6 h5f6 d4f6 e7f6 c3d5 f6b2 c2c3 a6a5 c3b4 a5b4 b3b4 b7b4 d5b4 d8d1 f1d1 f8a8 b4d5 f7f6');
INSERT INTO evaluations VALUES(44197,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44197,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(44197,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(44197,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44197,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44197,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44197,36,'cp 12','g2g3 b8c6 f1g2 d7d5 c4d5 e6d5 d2d4 c5d4 f3d4 f8c5 d4b3 c5b6 c3a4 e8g8 a4b6 a7b6 b3d4 f8e8 e1g1 c6d4 d1d4 e8e2 c1e3 c8f5 f1c1 a8c8 g2f1 e2c2 c1c2 c8c2 a1c1 c2c1 e3c1 d8d6 f1g2 f5e6 c1e3 f6d7 h2h3 h7h6 a2a3 d6c6 a3a4 b6b5 a4a5 c6c4 g3g4 c4d4 e3d4 b5b4');
INSERT INTO evaluations VALUES(44198,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(44198,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(44198,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(44198,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(44198,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(44198,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(44198,36,'cp 34','f8e7 f1g2 d7d5 c4d5 e6d5 d2d4 c5c4 e1g1 e8g8 b2b3 c4b3 d1b3 b8c6 c1f4 c6a5 b3b1 c8g4 e2e4 d5e4 c3e4 f6d5 f4d2 a8c8 d2a5 d8a5 f3e5 g4e6 b1b7 a5a4 f1d1 c8b8 b7c6 a4a3 c6c1 f8c8 c1a3 e7a3');
INSERT INTO evaluations VALUES(45736,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(45736,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(45736,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(45736,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(45736,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(45736,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(45736,36,'cp -64','c6d5 c4b3 f8b4 c1d2 e8g8 g1e2 b8c6 e1g1 h7h6 h2h3 c8e6 f2f4 d5d4 c3e4 f6e4 d3e4 d8b6 g1h1 a8d8 f1f3 b4d2 d1d2 e6b3 f3b3 b6c7 f4e5 f8e8 b3d3 c7e5 e2g3 e5g5 d2g5 h6g5 a1d1 g7g6 d3b3 d8d7 b3b5 f7f6 b5d5');
INSERT INTO evaluations VALUES(46783,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(46783,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(46783,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(46783,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(46783,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(46783,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(46783,36,'cp -12','f8g7 f1e2 e8g8 a2a4 c7c5 e1g1 b7b6 a4a5 c8a6 f3e5 b6a5 c1d2 c5d4 e3d4 d8b6 e2a6 b6a6 d1e2 a6e2 c3e2 f8c8 f1c1 f6e4 d2e3 e7e6 b2b3 e4c3 e2c3 c8c3 a1a5 g7e5 d4e5 b8c6 e3d2 c6a5 d2c3 a5c6 f2f4 a7a6 c3b2 a8c8');
INSERT INTO evaluations VALUES(56684,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(56684,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(56684,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(56684,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(56684,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(56684,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(56684,36,'cp 169','b8c6 a2a4 g8h6 d2d4 h6f7 a4a5 f8e7 e1g1 e8g8 c3d5 c6d4 f3d4 e5d4 d1d4 c7c6 d5f4 g8h8 f4e6 c8e6 c4e6 d8e8 c1d2 e7d8 e6f5 g7g6 f5g4 e8e5 d4b4 e5e7 a1e1 b7b6 a5a6 b6b5 f2f4 d8b6 g1h1 b6c5 b4c3 a8d8 c3h3');
INSERT INTO evaluations VALUES(22299,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(22299,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(22299,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(22299,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(22299,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(22299,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(22299,36,'cp 184','a2a4 g8h6 e1g1 h6f7 d2d4 f8e7 c3d5 c6d4 f3d4 e5d4 d1d4 e8g8 a4a5 d8d7 c1f4 e7d8 d5e3 g8h8 f4g3 d7e8 f2f4 c7c6 g1h1 c8e6 c4d3 b7b6 f4f5 e6c8 g3d6 f7d6 d4d6 b6a5 b2b3 d8b6 e3c4 c8a6 c4a5');
INSERT INTO evaluations VALUES(10256,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(10256,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(10256,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(10256,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(10256,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(10256,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(10256,36,'cp 36','f3e5 c6e5 d2d4 d7d5 d4c5 f6e4 c1f4 e5g6 f4e3 e8g8 f2f3 e4g5 e1g1 c7c6 d1d2 h7h6 b1c3 g5e6 c3e2 f8e8 f1e1 b7b6 e2d4 e6d4 e3d4 e8e1 a1e1 g6f8 d4f2 c8f5 c2c3 f8d7 c5b6 a7b6 b2b3 f5e6 a2a4 c6c5 g2f1 d8f6 g3g4');
INSERT INTO evaluations VALUES(10875,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10875,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10875,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10875,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(10875,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(10875,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10875,36,'cp 4','b1c3 d7d6 a2a4 f6e4 c3e4 d6d5 g5e7 c6e7 c4b3 d5e4 d3e4 e7c6 b3d5 c6d4 g1e2 c7c6 d5a2 d4e6 d1d8 e6d8 f2f4 e5f4 e2f4 e8e7 a4a5 c8e6 a2e6 d8e6 f4e6 e7e6 e1f2 a8d8 a1d1 h7h5 b2b4 h5h4 f2e3 d8d1 h1d1 b7b6 a5a6');
INSERT INTO evaluations VALUES(10873,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10873,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(10873,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(10873,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(10873,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(10873,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(10873,36,'cp -10','h7h6 g5f6 e7f6 c3d5 c6a5 d5f6 d8f6 g1e2 e8g8 e1g1 a5c4 d3c4 d7d6 b2b3 c8e6 f2f3 a7a5 d1d3 b7b6 a1e1 f6g5 e2c3 a8c8 f1f2 c8e8 d3d2 g5g6 a2a4 f7f5 c3d5 f8f7 e4f5 e6f5 f3f4 f5e4 f4e5 f7f2 d2f2');
INSERT INTO evaluations VALUES(15277,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(15277,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(15277,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(15277,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(15277,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(15277,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(15277,36,'cp -76','d8d5 b1c3 f8b4 c1d2 b4c3 d2c3 f7f6 d1e2 a7a5 a2a3 b7b5 e1g1 c8g4 d3d4 e8g8 d4e5 f6e5 f1e1 b5b4 a3b4 a5b4 a1a8 f8a8 e1d1 d5f7 c3e1 c6d4 e2e4 d4f3 g2f3 f7f3 e4d5 f3d5 d1d5 a8a2 e1b4 a2b2 b4c3 b2c2 d5d8 g8f7 c3e5 c2c4 d8h8');
INSERT INTO evaluations VALUES(16185,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(16185,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(16185,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16185,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(16185,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(16185,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(16185,36,'cp -39','f6e4 d4d5 f8b4 b1d2 c6e7 e1g1 e4d6 c4d3 e8g8 d3c2 b7b6 a2a3 b4d2 d1d2 c8b7 d2d3 f7f5 c2b3 e7g6 f1e1 g8h7 c1d2 d8f6 d2c3 f6f7 c3b4 f8e8 a1c1 g6f4 d3d2 b7d5 d2f4 e8e1 c1e1 d5b3 b4d6 c7d6 h2h4 b3c4 f4d6 a8e8 e1e8 f7e8 f3e5 c4e6 g1h2 e8h5 e5d7 h5h4 h2g1');
INSERT INTO evaluations VALUES(16204,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(16204,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(16204,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(16204,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(16204,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(16204,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(16204,36,'cp -33','d4d5 f8b4 b1d2 e4d2 c1d2 b4d2 d1d2 c6e7 d5d6 c7d6 e1g1 e8g8 a1d1 d6d5 c4d5 e7d5 d2d5 a7a5 h2h3 a8a6 f1e1 d7d6 e1e3 a6c6 d5d2 d8c7 f3d4 c6c5 d4b3 c5c2 d2d4 c2c4 d4d2 a5a4 b3d4 c8d7 e3g3 g8h8 b2b3 c4c5 b3b4 c5d5 g3d3 c7b6 a2a3 f8d8 d2f4 d5e5');
INSERT INTO evaluations VALUES(41858,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(41858,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41858,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41858,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41858,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41858,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(41858,36,'cp -14','c7c6 e1g1 e8g8 c3e2 b8d7 c2c3 c5b6 e2g3 h7h6 c4b3 f8e8 b3c2 d7f8 h2h3 d6d5 d3d4 f6e4 g3e4 d5e4 f3e5 f7f6 e5c4 b6c7 f2f4 e4f3 d1f3 c8e6 c4e3 d8d6 e3f5 e6f5 c2f5 a8d8 c1e3 g7g6 f5c2 g8g7 f3f4 d6f4 e3f4 c7f4 f1f4');
INSERT INTO evaluations VALUES(47065,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(47065,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47065,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47065,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(47065,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47065,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47065,36,'cp 24','d6e5 d4e5 c7c6 b1c3 e7e6 g1f3 d5c3 d1d8 e8d8 b2c3 c8d7 c4d3 d7e8 a2a4 a7a5 h2h4 b8d7 h4h5 h7h6 h1h3 d8c7 h3g3 d7c5 c1e3 c5d3 c2d3 b7b5 f3d2 b5a4 d2e4 a4a3 c3c4 a5a4 e1f1 g7g6 h5g6 f7g6 e4c5 f8c5 e3c5');
INSERT INTO evaluations VALUES(53421,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(53421,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(53421,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(53421,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(53421,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(53421,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(53421,36,'cp -31','b8c6 d3d4 c8g4 c2c3 e7e6 h2h3 g4h5 c1f4 d8f6 f4g3 h5f3 d1f3 f6f3 g2f3 g8f6 b1d2 f8e7 d2b3 e8g8 b3c5 b7b6 c5d3 g7g6 e1d1 a7a6 a2a4 f8c8 d3e5 h7h5 e5c6 c8c6 g3h4 g8g7 h4g5 e7d6 f3f4 c6c8');
INSERT INTO evaluations VALUES(19877,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(19877,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(19877,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(19877,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(19877,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(19877,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(19877,36,'cp -16','d3d4 c8g4 c2c3 e7e6 c1f4 d8f6 f4g3 g4f3 d1f3 f6f3 g2f3 g7g6 a2a4 h7h5 h2h4 g8e7 f1d3 e7f5 b1d2 f7f6 e1f1 e8f7 f1g2 f8e7 a1e1 a8c8 d2b3 f5g7 e1e2 h8d8 h1a1 e7d6 a4a5 a7a6 g3d6 d8d6 e2e1 d6d8');
INSERT INTO evaluations VALUES(6469,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(6469,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(6469,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(6469,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(6469,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6469,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(6469,36,'cp 21','b1c3 g8f6 a2a3 d7d6 b2b4 c5b6 e2e3 e8g8 d2d3 c6e7 g1f3 c7c6 d1c2 e7g6 a3a4 d6d5 e1g1 c8e6 a4a5 b6c7 c3a4 b7b6 a5b6 a7b6 c1b2 a8b8 f1c1 h7h6 c4c5 b6b5 a4b6 f6d7 b6d7 d8d7 a1a7 b8d8');
INSERT INTO evaluations VALUES(7458,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(7458,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(7458,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(7458,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(7458,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(7458,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(7458,36,'cp 130','d2d4 c5d6 e1g1 c8b7 f3g5 e8g8 f2f4 e5f4 e4e5 c6e5 d4e5 d6e5 d1e1 d8e7 c1f4 a8e8 f4e5 e7e5 e1e5 e8e5 g5f7 f8f7 b3f7 g8f7 b1d2 e5e2 f1f2 e2e7 g1f1 b7d5 a2a4 f7g8 h2h3 e7e3 a4b5 a6b5 a1e1 e3g3 f1g1 h7h6 d2f1 g3g6 g1h2 d5f7 f1e3 g6g5');
INSERT INTO evaluations VALUES(7446,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(7446,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(7446,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(7446,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(7446,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(7446,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(7446,36,'cp 141','c5d6 e1g1 c8b7 f3g5 e8g8 f2f4 e5f4 e4e5 c6e5 d4e5 d6e5 d1e1 d8e7 c1f4 a8e8 e1e5 e7e5 f4e5 e8e5 g5f7 f8f7 b3f7 g8f7 b1d2 e5e2 f1f2 e2e3 d2b3 e3e5 b3d4 b7d5 f2e2 e5h5 a1e1 c7c6 d4b3 h5g5 b3d2 f7f8 e1a1 g5g4 b2b3 d5e6 a1f1 f8e7 c3c4 g4d4 d2f3 d4d3 f1e1 e7d6 c4b5');
INSERT INTO evaluations VALUES(13550,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13550,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13550,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13550,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13550,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13550,36,'cp 0','d5c4 e1g1 a7a6 e2c4 b7b5 c4d3 c5d4 e3d4 c6b4 d3e2 f8d6 f3e5 c8b7 a2a4 e8g8 a4b5 a6b5 a1a8 d8a8 c3b5 d6e7 b5c7 a8d8 c7b5');
INSERT INTO evaluations VALUES(13500,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13500,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13500,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13500,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13500,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13500,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13500,36,'cp 0','e1g1 c5d4 e3d4 f8e7 e2c4 e8g8 a2a3 b7b6 c4a2 c8b7 c1f4 e7d6 f4g5 d6e7');
INSERT INTO evaluations VALUES(18753,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(18753,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(18753,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(18753,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(18753,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(18753,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(18753,36,'cp -45','g8f6 b1c3 d7d5 e4d5 e6d5 c4a2 d5d4 d1e2 f8e7 c3e4 e8g8 g1f3 f6d5 e1g1 h7h6 e4g3 c8e6 f3e5 c6e5 e2e5 e7d6 e5h5 d5f6 h5f3 d6g3 f2g3 e6a2 a1a2 d8d5 a2a1 d5f3 g2f3 f6d5 c1d2 f7f6 c2c3 d4c3 b2c3 a8d8 c3c4 d5c7 d2e3 d8d3');
INSERT INTO evaluations VALUES(12384,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(12384,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(12384,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(12384,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(12384,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(12384,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(12384,36,'cp -25','b1c3 d7d5 e4d5 e6d5 c4a2 c8e6 g1e2 f8e7 e1g1 e8g8 f1e1 e7d6 e2f4 e6g4 d1d2 d5d4 c3e4 f6e4 d3e4 d6f4 d2f4 g4e6 a2e6 f7e6 f4g4 d8d7 c1d2 c5c4 a3a4 a7a6 f2f4 f8f7 h2h3 a8f8 g1h1 b7b6 g4e2 b6b5 a4b5 a6b5 e1f1 f8e8');
INSERT INTO evaluations VALUES(36666,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36666,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(36666,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(36666,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(36666,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36666,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36666,36,'cp 31','c8g4 h2h3 g4f3 d1f3 b8c6 c3e2 c6b4 c1g5 c7c5 d4c5 d6c5 a1d1 d8c7 g5f4 c7c6 e2c3 c5c4 d3e2 a8d8 d1d8 f8d8 f1c1 f6d7 f4g5 c6e6 f3g4 g7c3 g4e6 f7e6 b2c3 b4a2 c1a1 a2c3 e2c4 g8f8 g5e3 f8f7 a1a7 c3e4 a7b7');
INSERT INTO evaluations VALUES(27866,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(27866,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(27866,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(27866,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(27866,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(27866,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(27866,36,'cp 29','h2h3 g4f3 d1f3 b8c6 c3e2 c6b4 c2c3 b4d3 f3d3 e7e5 e2g3 f8e8 c1e3 h7h5 d4d5 d8e7 g3e2 g8h7 a1e1 e8f8 c3c4 a7a6 a2a4 g7h6 e3h6 h7h6 a4a5 h6h7 e2c3 f6d7 b2b4 a8e8 d3e3 f7f5 e4f5 g6f5 f2f4 e5e4');
INSERT INTO evaluations VALUES(45655,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(45655,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(45655,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(45655,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(45655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45655,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(45655,36,'cp 0','c4d5 e6d5 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(48283,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48283,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(48283,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48283,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(48283,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48283,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48283,36,'cp 0','e2e3 f8d6 f1d3 c7c5 g1f3 e8g8 f4d6 d8d6 f3e5 d6b6 d1e2 f6d7 f2f4 b8c6 e1g1 g8h8 e5d7 c8d7 d4c5 b6c5 b1d2 e6e5 d2b3 c5e7 a1d1 f7f6 e3e4 d5e4 e2e4 f6f5 e4d5 e5f4 d5c5 e7f6 f1f4 c6e5 d3c2');
INSERT INTO evaluations VALUES(56475,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(56475,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(56475,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(56475,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(56475,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(56475,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(56475,36,'cp -4','d2d4 b8c6 g1f3 e5e4 f3d2 c6e7 c2c4 c7c6 b1c3 f7f5 f1e2 c8e6 f2f3 g8f6 e1g1 a7a6 d1b3 d8d7 c3a4 e7c8 c4d5 e6d5 b3d1 e4f3 d2f3 f8d6 e2d3 f6e4 f3e5 d7e6 a4c3 e8g8 d3e4 f5e4 f1f8 d6f8 d1c2 f8d6 c3e4 d6e5');
INSERT INTO evaluations VALUES(3811,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(3811,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(3811,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(3811,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(3811,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(3811,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(3811,36,'cp -13','a2a4 c6e7 c1e3 b7b6 b1c3 h7h6 g5f3 c8e6 c4e6 f7e6 a4a5 e7g6 a5b6 a7b6 d1e2 f8f7 a1a8 d8a8 h2h3 f7f8 f3h2 a8c6 h2g4 c5d4 c3d1 f6g4 e2g4 g8h7 c2c3 d4e3 f2e3 c6b5 g4e2 d6d5 d1f2 b5b3 f2g4 f8f1 g1f1 d5e4');
INSERT INTO evaluations VALUES(11820,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(11820,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(11820,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(11820,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(11820,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(11820,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(11820,36,'cp 57','c4f1 f6d5 f1b5 c8b7 d2d3 f8e7 g5f3 f7f6 b1c3 d5b6 c1e3 a7a6 b5c6 b7c6 f3h4 g7g6 h4f3 b6d5 e1g1 d5c3 b2c3 e8g8 f1e1 f8e8 a2a4 a6a5 d3d4 e7f8 d4e5 d8d1 e1d1 c6f3 g2f3 e8e5 d1d7 f8c5 e3d4 c5d4 c3d4');
INSERT INTO evaluations VALUES(12534,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(12534,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(12534,32,'cp 238',NULL);
INSERT INTO evaluations VALUES(12534,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(12534,34,'cp 232',NULL);
INSERT INTO evaluations VALUES(12534,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(12534,36,'cp 222','f3g5 d7d5 e4d5 c6d4 c2c3 d4f5 b1d2 f5d6 c4b3 f8e7 g5e4 b7b6 e1g1 e8g8 f1e1 c8f5 b3a4 f5d7 e4d6 d7a4 d1a4 d8d6 c3c4 f6d7 d2f3 f7f5 c1d2 e7f6 d2c3 f8e8 a1d1 e5e4 d3e4 f6c3 b2c3 e8e4 e1e4 f5e4 f3g5');
INSERT INTO evaluations VALUES(12522,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(12522,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(12522,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(12522,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(12522,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(12522,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(12522,36,'cp 231','d7d5 e4d5 c6d4 c2c3 d4f5 b1d2 f5d6 e1g1 f8e7 g5f3 d6c4 d3c4 e7d6 f3e5 e8g8 d2f3 f8e8 f1e1 d8c7 c1f4 c8f5 f4g3 h7h6 f3h4 f5h7 f2f4 c7b6 e1e2 f6h5 e5d7 b6c7 e2e8 a8e8 d1h5 c7d7 f4f5 b7b5 h5f3 b5c4 a1d1 e8d8 f3g4 g7g5 g3d6 d7d6');
INSERT INTO evaluations VALUES(13783,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(13783,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(13783,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(13783,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(13783,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(13783,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(13783,36,'cp 232','e4d5 c6d4 c2c3 d4f5 b1d2 f5d6 e1g1 d6c4 d2c4 h7h6 f1e1 f8e7 d5d6 e7d6 d1a4 e8f8 g5f3 f8g8 c4e5 d8c7 d3d4 g8h7 a4c2 g7g6 c2d2 h7g7 b2b3 c8f5 e5c4 d6f8 c4e3 f5e4 f3e5 f8d6 e3c4 d6f8 c1b2 a8e8 f2f3 e4d5 c4e3 f8d6 a1d1 c5d4 c3d4');
INSERT INTO evaluations VALUES(13730,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(13730,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(13730,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(13730,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(13730,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(13730,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(13730,36,'cp 244','c6d4 c2c3 d4f5 b1d2 f5d6 e1g1 d6c4 d2c4 h7h6 f1e1 f8e7 d5d6 e7d6 d1a4 e8f8 g5f3 c8e6 c4e5 d6e5 f3e5 f8g8 c1e3 d8c8 d3d4 c5c4 a4c2 f6d5 c2e2 f7f6 e5g6 h8h7 e3f4 g8f7 e2h5 f7g8 b2b3 d5c3 h5f3 c4b3 a2b3 e6b3');
INSERT INTO evaluations VALUES(41887,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41887,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(41887,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(41887,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(41887,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(41887,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(41887,36,'cp 32','b2b4 c5b6 g1f3 b8c6 b4b5 c6d4 f3d4 b6d4 d1c2 h7h5 h2h4 e8g8 e1g1 c8e6 d2d3 c7c6 b5c6 b7c6 e2e3 d4c3 c2c3 d8c7 c1b2 f6d7 c3d2 a8b8 f1b1 e6f5 b2a3 b8b1 a1b1 f8b8 b1b8 d7b8 d2c3 b8d7 e3e4 f5e6 d3d4 g8h8');
INSERT INTO evaluations VALUES(42949,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(42949,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(42949,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(42949,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(42949,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(42949,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(42949,36,'cp 58','e2e3 e8g8 d1c2 c7c5 g1f3 b8c6 d4c5 h7h6 g5h4 e7c5 a2a3 b7b6 b2b4 c5d6 a1d1 d6b8 f1e2 c8b7 e1g1 d8e7 h4f6 f8f6 d1d2 c6e5 f3e5 b8e5 f1d1 a8d8 c3b5 b7e4 c2b3 f6g6 g2g3 d7d6');
INSERT INTO evaluations VALUES(42947,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(42947,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(42947,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(42947,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(42947,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(42947,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(42947,36,'cp 50','e8g8 d1c2 c7c5 a1d1 c5d4 e3d4 b7b6 g1f3 c8b7 f1e2 f6e4 g5e7 d8e7 e1g1 d7d6 f1e1 b8d7 f3d2 e4c3 c2c3 h7h6 e2f1 f8e8 b2b4 g8h8 h2h3 d7f6 c3d3 b7e4 d3b3 a8c8 f2f3 e4b7 b3d3');
INSERT INTO evaluations VALUES(37294,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(37294,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(37294,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(37294,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(37294,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(37294,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(37294,36,'cp 59','d1c2 c7c5 f1e2 h7h6 g5f4 c5d4 e3d4 d7d5 a1d1 d5c4 e2c4 b8c6 g1f3 c6a5 c4e2 c8d7 e1g1 a8c8 c2b1 d7c6 f3e5 e7d6 f1e1 f6d5 e5c6 a5c6 f4d6 d8d6 e2c4 g8h8 c4b3 a7a6 c3d5 e6d5 b1d3 g7g6 h2h4 c8d8 h4h5 g6h5 d3h3 d6g6');
INSERT INTO evaluations VALUES(43587,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(43587,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(43587,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(43587,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43587,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(43587,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(43587,36,'cp -24','e7e6 e2e3 c7c5 d4c5 f8c5 f1d3 b8d7 e1g1 b7b5 f1e1 c8b7 e3e4 f6e4 d3e4 d5e4 c3e4 b7e4 e1e4 d7f6 e4e2 d8d1 a1d1 e8e7 f4c7 h8c8 c7a5 h7h6 g1f1 e7f8 f3e5 f6d5 e5d7 f8g8 d7c5 c8c5 c2c3 c5c8 g2g3');
INSERT INTO evaluations VALUES(43652,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(43652,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(43652,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(43652,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(43652,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(43652,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(43652,36,'cp -17','e2e3 c7c5 d4c5 f8c5 f1d3 b8d7 e3e4 e6e5 f4g5 d8b6 e1g1 d5e4 c3e4 f6e4 d3e4 h7h6 g5h4 e8g8 e4d5 d7f6 h4f6 b6f6 d1e2 f8e8 f3d2 f6g6 a1e1 a8b8 e2c4 c5f8 c4c7 c8f5 e1e5 f5c2 e5e8 b8e8 c7b7 e8e7 b7a8 g6f6 d2f3 c2d3 f1e1 e7e1 f3e1');
INSERT INTO evaluations VALUES(43651,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(43651,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(43651,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(43651,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(43651,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(43651,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(43651,36,'cp -9','c7c5 d4c5 f8c5 f1d3 b8d7 e3e4 e6e5 f4g5 d8b6 e1g1 d5e4 c3e4 f6e4 d3e4 h7h6 g5h4 e8g8 d1d3 b6b2 f1b1 b2a3 e4b7 a3d3 c2d3 a8a7 b7d5 d7f6 d5c4 c8b7 b1e1 b7f3 g2f3 c5d4 a1c1 f6h5 h4g3 h5g3 h2g3 g7g6 e1e2 g8g7 g1g2 f8d8 c1b1 h6h5 a2a4 h5h4 g3h4');
INSERT INTO evaluations VALUES(43563,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(43563,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(43563,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43563,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(43563,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(43563,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(43563,36,'cp 0','d4c5 f8c5 f1d3 b8d7 e3e4 d8b6 e1g1 b6b2 f4d2 d5e4 a1b1 b2a3 c3e4 f6e4 d3e4 a3a4 f1e1 h7h6 d2c3 e8g8 f3d2 c5d4 c3b4 d4c5');
INSERT INTO evaluations VALUES(49293,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(49293,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(49293,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(49293,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(49293,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(49293,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(49293,36,'cp 78','c4b3 f8e7 d2d4 e8g8 e1g1 b8c6 d1d3 d6e8 f1e1 c6e5 d4e5 d7d6 c1f4 d6e5 d3d8 e7d8 f4e5 c8d7 b1c3 e8f6 a1d1 d7c6 g1f1 f8e8 c3d5 a7a5 d5f6 d8f6 e5f6 g7f6 e1e8 a8e8 f2f3 g8g7 d1d4 f6f5 b3c4 g7f6 f1f2 e8e5 b2b3 b7b5 c4d3 h7h6 c2c4');
INSERT INTO evaluations VALUES(49295,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(49295,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(49295,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(49295,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(49295,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(49295,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(49295,36,'cp 77','b8c6 d2d4 f8e7 d1d3 e8g8 e1g1 d6e8 f1e1 c6e5 d4e5 d7d6 c1f4 d6e5 d3d8 e7d8 f4e5 c8d7 b1c3 e8f6 a1d1 d7c6 g1f1 f8e8 c3d5 a7a5 d5c7 d8c7 e5c7 a5a4 b3c4 e8e1 f1e1 c6g2 c4b5 g2d5 c2c4 d5c6 b5c6 b7c6 d1d6 c6c5 f2f3 h7h6 e1d1 a8e8');
INSERT INTO evaluations VALUES(53631,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(53631,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(53631,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53631,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(53631,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53631,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(53631,36,'cp 23','c6c5 g1f3 g8h6 c2c3 b8c6 b1a3 c8d7 a3c2 a7a5 a2a3 d8b6 f1d3 c5d4 c3d4 h6f5 d3f5 e6f5 e1g1 d7e6 a1b1 f8e7 h2h3 e8g8 c1d2 a5a4 g1h2 f8c8 d2c3 b6b5 b1c1 c6a5 c3b4 e7b4 c2b4 c8c4 c1c4 a5c4');
INSERT INTO evaluations VALUES(54283,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(54283,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(54283,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(54283,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(54283,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(54283,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(54283,36,'cp 39','g1f3 g8h6 c2c3 b8c6 b1a3 f8e7 a3c2 c8d7 d4c5 e7c5 c1e3 c5e3 c2e3 d8c7 d1d2 f7f6 e5f6 g7f6 g2g3 e8c8 f1g2 c8b8 e1c1 h6f7 h1e1 f7d6 f4f5 d6e4 d2c2 h8e8 c1b1 h7h5 f5e6 d7e6 b1a1 c7a5 f3d4 c6d4 d1d4');
INSERT INTO evaluations VALUES(16908,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(16908,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(16908,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(16908,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(16908,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(16908,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(16908,36,'cp 113','e4e5 f6d5 d2d3 d5b6 f1e1 a7a6 b5c4 b6c4 d3c4 d7d5 e5d6 c8e6 d6c7 d8c7 d1d4 f8d6 b1c3 e8g8 c3d5 e6d5 c4d5 c7c2 c1d2 d6c5 d4c3 c2c3 d2c3 f7f6 g2g3 a8d8 e1d1 d8d7 a1c1 b7b6 d1d2 f8d8 c1d1 h7h5 g1g2 g8f7 g2f3 d8c8 d2d3 c5d6 a2a3 c8c4 d3d4 c4c8 d4e4');
INSERT INTO evaluations VALUES(16886,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(16886,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(16886,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(16886,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(16886,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(16886,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(16886,36,'cp 131','f6d5 d2d3 d5b6 f1e1 a7a6 b5c4 b6c4 d3c4 d7d5 e5d6 c8e6 c1f4 f8d6 d1d4 e8g8 f4d6 d8d6 d4d6 c7d6 b2b3 a8c8 e1d1 f8d8 b1c3 d6d5 c4d5 e6f5 c3e2 f5c2 d1d4 g8f8 a1c1 d8e8 g1f1 f8e7 f2f3 e7d6 f1f2 b7b5 h2h4 a6a5 g2g4 a5a4 b3a4 b5a4');
INSERT INTO evaluations VALUES(17073,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(17073,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(17073,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(17073,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(17073,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(17073,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(17073,36,'cp 135','d2d3 d5b6 f1e1 a7a6 b5c4 b6c4 d3c4 d7d5 e5d6 c8e6 c1f4 f8d6 d1d4 e8g8 f4d6 d8d6 d4d6 c7d6 b1a3 f8d8 a1d1 d6d5 c4d5 d8d5 d1d5 e6d5 b2b3 d5e6 e1d1 a8c8 d1d4 g8f8 c2c4 c8c5 a3c2 b7b5 c2e3 b5c4 e3c4 h7h5 g1f1 g7g5 f1e2 h5h4 g2g3 h4g3 h2g3 f8e7 c4e3');
INSERT INTO evaluations VALUES(18884,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(18884,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(18884,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(18884,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(18884,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(18884,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(18884,36,'cp -25','g8f6 c4d3 d7d5 e4d5 e6d5 g1f3 f8b4 c2c3 b4c5 e1g1 e8g8 c1f4 h7h6 b1d2 f6h5 f4g3 h5g3 h2g3 c5b6 d2b3 c8g4 d1d2 h6h5 a1e1 d8f6 f3h2 g4d7 h2f3 g7g6 d3b5 a8d8 d2d5 g8g7 d5e4');
INSERT INTO evaluations VALUES(12457,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(12457,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(12457,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(12457,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(12457,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(12457,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(12457,36,'cp -27','c4d3 d7d5 e4d5 f6d5 g1f3 d5b4 d3e2 e6e5 a2a3 d8d1 e2d1 b4d5 c2c4 d5e7 b2b4 c8e6 b1c3 f7f6 c4c5 h7h5 e1g1 e7f5 h2h4 f8e7 d1c2 e8c8 f1e1 c6d4 f3d4 f5d4 c2d1 c8b8 c1b2 d4f5 g2g3 g7g6 a1c1 f5d4 g1g2 d4b3 d1b3 e6b3 c3b5 a7a6');
INSERT INTO evaluations VALUES(20505,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(20505,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(20505,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(20505,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(20505,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(20505,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(20505,36,'cp -7','f8e7 c4f1 g8f6 d2d4 e8g8 b1c3 c5c4 h2h3 h7h6 b2b3 c4b3 a2b3 c8f5 c3a4 e7d6 c2c3 d8c7 c1a3 d6a3 a1a3 f8e8 b3b4 b7b6 a4b2 a7a5 e1e8 a8e8 b4a5 c6a5 d1a1 f5e4 f3e5 e4c2 a1c1 c2b3 b2d1 b3d1 c1d1');
INSERT INTO evaluations VALUES(5436,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(5436,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(5436,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(5436,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(5436,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(5436,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(5436,36,'cp -15','c4f1 g8f6 d2d4 e8g8 b1c3 h7h6 d4c5 e7c5 h2h3 c8f5 f1d3 f5d3 c2d3 d8d7 d3d4 c5b4 c1d2 f8e8 d1b3 f6e4 a1d1 a8d8 f3e5 e4d2 d1d2 d7d6 e5f3 e8e1 f3e1 c6a5 b3c2 d8c8 d2e2 b4c3 b2c3 d6f6 e1f3 a5c4');
INSERT INTO evaluations VALUES(31484,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(31484,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(31484,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(31484,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(31484,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(31484,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(31484,36,'cp -28','g4f3 d1f3 c6c5 f1b5 b8c6 c2c3 c5d4 c3d4 g8e7 e1g1 e7f5 f3d3 f8e7 b1d2 e8g8 d2f3 c6b4 d3d1 a8c8 a2a3 b4c6 g1h1 a7a6 b5e2 d8b6 e2d3 c6d4 f3d4 b6d4 d3f5 d4d1 f1d1 e6f5 d1d5 e7c5 h1g1 c5a7 c1f4 c8c4 f4e3 a7e3 f2e3 c4c2');
INSERT INTO evaluations VALUES(31499,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(31499,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(31499,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(31499,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(31499,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(31499,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(31499,36,'cp -12','d1f3 c6c5 c2c3 b8c6 f1b5 d8b6 f3d3 a7a6 b5c6 b6c6 e1g1 a8c8 c1e3 c5d4 e3d4 g8e7 b1d2 e7f5 d2b3 c6a4 d4a7 f8e7 b3d4 f5d4 a7d4 e8g8 b2b3 a4c6 a1c1 f8d8 f1d1 e7c5 d3e3 h7h6 h3h4 c5d4 d1d4 c6b6 h4h5');
INSERT INTO evaluations VALUES(31498,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(31498,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(31498,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(31498,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(31498,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(31498,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(31498,36,'cp -14','c6c5 f1b5 b8c6 c2c3 d8b6 f3d3 a7a6 b5c6 b6c6 e1g1 g8e7 d4c5 c6c5 c1e3 c5c7 e3d4 e7f5 b1d2 a8c8 d2b3 h7h5 d3e2 f8e7 f1e1 g7g6 e1d1 e8g8 d1d3 g8g7 b3d2 c7d7 d2f3 d7b5 e2d2 c8c4 b2b3 f5d4');
INSERT INTO evaluations VALUES(31657,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(31657,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(31657,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31657,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(31657,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(31657,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(31657,36,'cp -19','f1b5 b8c6 c2c3 c5d4 c3d4 g8e7 b1c3 e7f5 c1e3 f8e7 e1g1 e8g8 c3e2 c6b4 b5a4 a8c8 e3d2 c8c4 d2c3 b4c6 a4c6 c4c6 f3d3 d8b6 a2a4 c6c4 a4a5 b6b5 e2g3 f5g3 f2g3 e7b4 f1f3 h7h6 g3g4 b4c3 b2c3 b5c6 a1f1');
INSERT INTO evaluations VALUES(43162,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(43162,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(43162,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43162,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43162,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43162,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(43162,36,'cp 16','d7d5 c4d5 f6d5 f1g2 c7c5 e1g1 b8c6 d2d4 c5d4 f3d4 d5c3 b2c3 c6d4 c3d4 d8d4 a1b1 e8g8 c1e3 d4c4 g2d5 c4a6 d1b3 a6e2 e3c5 c8e6 d5e6 e2e6 b3e6 f7e6 b1b7 f8b8 f1b1 b8c8 c5a7 c8c2 b7e7 g7f6 e7b7 c2a2 a7e3');
INSERT INTO evaluations VALUES(46630,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(46630,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(46630,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(46630,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(46630,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(46630,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(46630,36,'cp -31','d6e5 d4e5 d8d1 c3d1 f6d5 g1f3 f8g7 d1e3 d5e3 c1e3 b8c6 f1b5 c8d7 e3f4 a7a6 b5a4 e8c8 e1c1 h7h6 h2h4 b7b5 a4b3 e7e6 h1e1 c6a5 g2g3 d7c6 d1d8 h8d8 f3d2 a5b3 d2b3 g7f8 e1d1 c6d5 f4e3 h6h5 b3c5 a6a5 b2b3 f8e7 c1b2 d5f3 d1d8 c8d8 c5d3');
INSERT INTO evaluations VALUES(46817,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46817,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(46817,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(46817,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(46817,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(46817,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(46817,36,'cp -35','d4e5 d8d1 c3d1 f6d5 d1e3 d5e3 c1e3 b8c6 g1f3 f8g7 f1b5 c8d7 e3f4 a7a6 b5c6 d7c6 e1c1 e7e6 c2c4 h7h6 h2h4 e8e7 b2b3 h8d8 h1g1 b7b5 d1d8 a8d8 g1d1 d8d1 c1d1 c6f3 g2f3 e7d7 d1c2 h6h5 c2d3 c7c5 d3e4 b5c4 b3c4 d7c7');
INSERT INTO evaluations VALUES(46816,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(46816,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(46816,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(46816,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(46816,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(46816,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(46816,36,'cp -24','d8d1 c3d1 f6d5 d1e3 d5e3 c1e3 b8c6 g1f3 f8g7 f1b5 c8d7 e3f4 a7a6 b5a4 h7h6 h2h4 e8c8 e1c1 b7b5 a4b3 e7e6 c2c3 c6e7 h1e1 d7c6 d1d8 h8d8 b3c2 c6f3 g2f3 e7d5 f4d2 g7f8 d2e3 h6h5 c2e4 c7c6 c1c2 c8c7 f3f4 a6a5 a2a3 a5a4');
INSERT INTO evaluations VALUES(33637,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(33637,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(33637,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(33637,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(33637,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(33637,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(33637,36,'cp -22','c3d1 f6d5 d1e3 d5e3 c1e3 b8c6 g1f3 f8g7 f1b5 c8d7 e3f4 a7a6 b5a4 h7h6 h2h4 e8c8 e1c1 b7b5 a4b3 e7e6 c2c3 c6e7 b3c2 d7c6 h1e1 d8d1 c2d1 h8d8 d1c2 c6f3 g2f3 e7d5 f4d2 g7f8 d2e3 h6h5 c2e4 c7c6 c1c2 c8c7 f3f4 f8h6');
INSERT INTO evaluations VALUES(33638,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(33638,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(33638,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(33638,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(33638,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(33638,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(33638,36,'cp -27','f6d5 d1e3 d5e3 c1e3 b8c6 g1f3 f8g7 f1b5 c8d7 e3f4 a7a6 b5a4 e8c8 e1c1 h7h6 h2h4 b7b5 a4b3 e7e6 c2c3 c6e7 b3c2 d7c6 d1d8 h8d8 h1e1 c6f3 g2f3 e7d5 f4d2 g7f8 a2a4 h6h5 a4b5 a6b5 c2e4 f8e7 e1h1 c7c5 c1c2 c8c7 f3f4 c7c6 e4g2 c6b6 b2b3');
INSERT INTO evaluations VALUES(33639,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(33639,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(33639,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(33639,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(33639,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(33639,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(33639,36,'cp -17','d1e3 d5e3 c1e3 b8c6 g1f3 f8g7 f1b5 c8d7 e3f4 a7a6 b5a4 e8c8 e1c1 h7h6 h2h4 b7b5 a4b3 e7e6 c2c3 c6a5 b3c2 d7c6 h1e1 d8d1 c2d1 c6d5 b2b3 c7c5 g2g3 a5c6 e1e3 c8c7 d1e2 b5b4 c3c4 d5f3 e2f3 c6d4 f3d1 h8d8 e3d3');
INSERT INTO evaluations VALUES(47233,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(47233,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(47233,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(47233,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(47233,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(47233,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(47233,36,'cp 87','e4e5 f6d7 d2d4 c7c5 c2c4 e7e6 c4d5 e6d5 f1b5 a7a6 b5a4 b8c6 e1g1 c5c4 a4c2 d7b6 h2h3 c6b4 c1g5 f8e7 g5e7 d8e7 b1c3 b4c2 d1c2 e8g8 a2a4 c8e6 a4a5 b6d7 c2d2 b7b5 a5b6 d7b6 f3e1 a6a5 f2f4 e7d7 g2g4 f7f5 e5f6 f8f6 e1f3');
INSERT INTO evaluations VALUES(47296,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(47296,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(47296,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(47296,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(47296,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(47296,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(47296,36,'cp 85','f6d7 d2d4 e7e6 f1d3 c7c5 c2c3 f8e7 c1e3 d8b6 b2b3 e8g8 e1g1 b8c6 b1d2 a7a5 a2a3 b6c7 f1e1 c5d4 c3d4 f7f6 d1c2 f6e5 d3h7 g8h8 h7g6 e5d4 f3d4 d7e5 d2f3 e5f3 d4f3 e6e5 a1d1 c8e6 c2e2 h8g8 h2h4 e5e4 f3g5 e7g5 h4g5 e6f5 e2h5 f5g6 h5g6');
INSERT INTO evaluations VALUES(51102,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(51102,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(51102,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(51102,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(51102,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(51102,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(51102,36,'cp 41','e7e6 b1c3 d7d5 c4d5 e6d5 c1f4 g8f6 e2e3 c7c5 g1e2 b8c6 g2g3 c5d4 e2d4 f8b4 f1g2 b4c3 b2c3 e8g8 e1g1 h7h6 d4c6 b7c6 d1a4 c8d7 c3c4 f8e8 c4d5 f6d5 a4a3 d5f4 e3f4 d8b6 a1c1 a6a5 a3c5 b6c5 c1c5 a5a4 f1c1 e8e2 h2h3');
INSERT INTO evaluations VALUES(56394,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(56394,31,'cp -148',NULL);
INSERT INTO evaluations VALUES(56394,32,'cp -159',NULL);
INSERT INTO evaluations VALUES(56394,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(56394,34,'cp -145',NULL);
INSERT INTO evaluations VALUES(56394,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(56394,36,'cp -153','g8f6 c1e3 f8e7 b1d2 c7c5 g2g3 b8c6 f1g2 c8e6 g1e2 h7h5 h2h3 h5h4 g3g4 d8c7 f3f4 d5d4 e3f2 e5f4 c2c3 d4c3 b2c3 e8c8 d3d4 g7g5 e4e5 f6d5 e1g1 f7f5 e5f6 d5f6 d2e4 f6g4 h3g4 e6g4 g1h1 g4h5 g2h3 c8b8');
INSERT INTO evaluations VALUES(5976,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(5976,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(5976,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(5976,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(5976,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(5976,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5976,36,'cp 0','g8f6 e1g1 e8g8 h2h3 h7h6 f1e1 a7a6 b1d2 c8e6 b2b4 c5a7 c4b3 f8e8 a2a4 d8d7 b3c2 e6h3 g2h3 d7h3 e1e2 e8e6 d1f1 h3h5 f1g2 f6d5 e4d5 e6g6 d5c6 g6g2 g1g2 h5g4 g2h2 g4h5');
INSERT INTO evaluations VALUES(15175,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(15175,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(15175,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(15175,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(15175,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(15175,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(15175,36,'cp -65','e5e4 f3d4 c6d4 b2d4 h7h5 c2c4 h5h4 b1c3 h4h3 g2f1 d5c4 e2e3 c4b3 d1b3 f8d6 f1e2 e8g8 e1g1 f8e8 e2c4 c8e6 c4e6 e8e6 b3b7 d6e5 d4e5 e6e5 c3e2 a8c8 e2f4 d8d7 a1c1 e5a5 c1c2 d7f5 c2c7 c8f8 b7b2 f6d5 c7c4 d5f4 e3f4 f8d8 f1d1');
INSERT INTO evaluations VALUES(15064,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(15064,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(15064,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(15064,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(15064,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(15064,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(15064,36,'cp -67','f3d4 c6d4 b2d4 h7h5 c2c4 h5h4 b1c3 h4h3 g2f1 d5c4 e2e3 c4b3 d1b3 f8d6 b3c2 e8g8 c3e4 f6e4 c2e4 f8e8 e4d5 c8g4 d4c3 a7a5 f1e2 g4e6 d5d4 d6f8 d4d8 e8d8 h1g1 c7c6 a2a4 f8b4 g3g4 b7b5 a4b5 c6b5 e2b5 a8c8 e1e2 b4c3 d2c3 c8c3');
INSERT INTO evaluations VALUES(15030,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(15030,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(15030,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(15030,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(15030,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(15030,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(15030,36,'cp -73','c6d4 b2d4 h7h5 c2c4 h5h4 b1c3 h4h3 g2f1 f6g4 b3b4 d5c4 e2e3 c8e6 f1e2 g4f6 a1b1 f8e7 b4b5 a7a6 e1g1 e8g8 d1c2 a6b5 b1b5 b7b6 a2a4 d8d7 d4f6 g7f6 b5h5 a8a5 h5a5 b6a5 c3b5 c7c5 e2c4 e6c4 c2c4 f6f5 b5c3 f8d8 f1d1');
INSERT INTO evaluations VALUES(15818,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(15818,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(15818,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(15818,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(15818,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(15818,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(15818,36,'cp -91','b2d4 h7h5 c2c4 h5h4 b1c3 h4h3 g2f1 d5c4 e2e3 c4b3 d1b3 f8d6 c3d5 f6d5 b3d5 e8g8 d5e4 f8e8 e4h4 c8f5 h4d8 a8d8 d4c3 c7c6 a2a4 d6e5 h1g1 e5c3 d2c3 e8e5 g3g4 f5e6 g4g5 e5c5 g1g3 c5c3 f1h3 e6h3 g3h3 d8d5 h3g3 c3c2 g5g6 f7g6 g3g6');
INSERT INTO evaluations VALUES(24854,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24854,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(24854,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24854,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(24854,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(24854,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24854,36,'cp 8','e7e6 e1g1 f5g6 b2b3 f8d6 c3b5 g6d3 d1d3 a7a6 b5d6 c7d6 c1b2 e8g8 c2c4 d8d7 f1c1 f8c8 h2h3 c8c7 a2a4 a8c8 f3d2 c6b4 d3e2 h7h6 b2a3 b4c6 e2d3 c6a5 f2f3 a5c6 e3e4 c6e7 c4d5 e6d5 c1c7 d7c7');
INSERT INTO evaluations VALUES(25175,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(25175,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(25175,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25175,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(25175,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25175,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(25175,36,'cp 10','d3f5 e6f5 c3e2 h7h5 h2h4 g7g6 b2b3 f8e7 c2c4 e8g8 e1g1 f8e8 g2g3 a7a5 a2a3 c6b8 c4c5 b8d7 c1d2 f6e4 b3b4 b7b6 e2f4 d7f6 d1b3 e7f8 g1g2 f8h6 f4d3 a5b4 a3b4 f6g4 a1a8 d8a8 c5b6 c7b6 f1c1');
INSERT INTO evaluations VALUES(44855,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(44855,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(44855,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(44855,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(44855,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(44855,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(44855,36,'cp -60','e7e5 e2e3 f8e7 b2b4 e8g8 g1f3 e5e4 f3d2 b7b6 d2b3 c8a6 f1a6 b8a6 a2a3 a6c7 e1g1 d8d7 c1d2 c7e6 a3a4 f8b8 c3e2 h7h5 h2h3 b8e8 c5b6 a7b6 a4a5 b6a5 a1a5 h5h4 b3c5 e7c5 d4c5 a8a5 b4a5');
INSERT INTO evaluations VALUES(55778,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55778,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55778,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55778,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55778,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(55778,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55778,36,'cp -4','e6d5 d2d4 c7c6 c1g5 f8e7 g5e7 d8e7 e2e3 c8f5 b1c3 g8f6 f3h4 f5e4 d1b3 e8g8 f1e2 a7a5 e1g1 b8a6 a2a3 f8b8 a1d1 a6c7 f2f3 e4g6 e3e4 f6e8 h4g6 h7g6 c3a4 b7b5 a4c5 e8d6 b3c2 d6c4 a3a4 c7e6 c5e6 f7e6');
INSERT INTO evaluations VALUES(56285,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56285,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56285,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(56285,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(56285,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56285,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56285,36,'cp 0','b1c3 c7c6 d2d4 c8f5 c1f4 b8d7 a2a3 g8f6 e2e3 f6h5 f4g5 h5f6');
INSERT INTO evaluations VALUES(56268,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56268,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56268,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56268,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56268,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56268,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56268,36,'cp 0','c7c6 d2d4 c8f5 c1f4 g8f6 e2e3 b8d7 a2a3 f6h5 f4g5 h5f6');
INSERT INTO evaluations VALUES(56567,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(56567,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(56567,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(56567,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(56567,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(56567,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(56567,36,'cp 92','f3e5 b8d7 d1e2 g8f6 e2c4 d7e5 d4e5 c8e6 c4b5 c7c6 b5b7 a8b8 b7c6 e6d7 c6a6 f6d5 a6c4 d5b4 b1a3 d8a5 c1d2 a5e5 d2c3 e5f5 c4d4 f8c5 d4e5 f5e5 c3e5 b8c8 e5g7 h8g8 g7e5 g8g6 e1c1 b4a2 c1b1 c5a3 b1a2 a3c5 f2f3 c5e7 c2c3 d7e6 a2b1 e6f5 g2g4 f5d7 f3e4 d7g4 d1d4');
INSERT INTO evaluations VALUES(56512,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(56512,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(56512,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(56512,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(56512,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(56512,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(56512,36,'cp 55','b8d7 d1e2 g8f6 e2c4 d7e5 d4e5 c8e6 c4b5 c7c6 b5b7 a8b8 b7c6 e6d7 c6a6 f6d5 a6c4 d5b4 b1a3 d8a5 c1d2 a5e5 d2c3 e5f5 c4d4 f8c5 d4e5 f5e5 c3e5 b8c8 e1c1 b4a2 c1b1 c5a3 f1a6 c8c5 e5g7 h8g8 g7d4 c5a5 a6c4 a3c5 c4a2 c5d4 d1d4 g8g2 a2d5 d7f5 b2b4 a5a6 d5e4 f5e4 d4e4 a6e6');
INSERT INTO evaluations VALUES(12559,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(12559,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12559,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12559,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12559,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12559,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12559,36,'cp 0','g2g3 d7d6 f1g2 g7g6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 h6h5 e1g1 f8h6 d1e1 e8g8 f3h4 h6g5 h4f3');
INSERT INTO evaluations VALUES(14268,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14268,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14268,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14268,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14268,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14268,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(14268,36,'cp 0','c8f5 c1g5 e7e6 e2e4 d5e4 c3e4 c6d4 c2c3 f5e4 g2e4 d4c6 d1b3 a8b8 a1d1 f8d6 c3c4 d8c8 g5f6 g7f6 b3e3 c6e5 c4c5 d6e7 g1f3 e5f3 e4f3 e8g8 e1g1 c7c6 e3f4 b7b6 f1e1 b6c5 f3e4 c8b7 f4g4 g8h8 e4h7 h8h7 g4h4 h7g7 h4g4');
INSERT INTO evaluations VALUES(24853,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(24853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24853,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(24853,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(24853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24853,36,'cp 0','c1g5 e7e6 e2e4 d5e4 c3e4 c6d4 c2c3 f5e4 g2e4 d4c6 d1b3 a8b8 a1d1 f8d6 c3c4 d8c8 g5f6 g7f6 b3e3 c6e5 c4c5 d6e7 g1f3 e5f3 e4f3 e8g8 e1g1 c7c6 e3f4 b7b6 f1e1 b6c5 f3e4 c8b7 e4h7 g8h7 f4h4 h7g8 h4g4 g8h7');
INSERT INTO evaluations VALUES(21412,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(21412,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(21412,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(21412,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(21412,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(21412,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21412,36,'cp 23','g1f3 g8f6 f1d3 c6b4 e1g1 c7c5 e3e4 b4d3 c2d3 c5d4 d1a4 c8d7 a4d4 d5e4 d3e4 d7c6 d4c4 d8b6 f3e5 b6c5 c4c5 f8c5 e5c6 b7c6 c1f4 c5d4 a1c1 a8d8 b2b3 e8g8 h2h3 h7h6 f1e1 d8d7 g1f1 e6e5 f4e3 f8d8 f2f3 f6h5 e3d4 d7d4 c3e2');
INSERT INTO evaluations VALUES(30630,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(30630,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30630,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(30630,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30630,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(30630,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(30630,36,'cp 0','f8b4 c4d3 f7f5 f2f3 d8h4 g2g3 h4h5 e4f5 b8c6 g1e2 e8c8 e1g1 e6f5 c3d5 b4d6 d5e3 g8h6 c2c3 c6e7 e3c4 h6f7 c4d6 f7d6 e2f4 h5f7 d1b3 f7f6 f4h5 f6g6 h5f4');
INSERT INTO evaluations VALUES(27970,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27970,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27970,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27970,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27970,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27970,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27970,36,'cp 0','c4d3 f7f5 f2f3 d8h4 g2g3 h4h5 e4f5 b8c6 g1e2 e8c8 e1g1 e6f5 c3d5 b4d6 d5e3 g8h6 c2c3 c6e7 e3c4 h6f7 c4d6 f7d6 e2f4 h5f7 d1b3 f7f6 f4h5 f6g6 h5f4');
INSERT INTO evaluations VALUES(31190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31190,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31190,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31190,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31190,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(31190,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(31190,36,'cp 6','e7e6 g1f3 f8e7 f1d3 f5d3 d1d3 g8f6 g2g4 e8g8 c1d2 c6c5 g4g5 f6d5 d4c5 d5c3 d3d8 f8d8 d2c3 e7c5 e1e2 b8c6 a1d1 c5d6 d1d3 d6b4 c3d4 c6d4 d3d4 d8d4 f3d4 h7h6 c2c3 b4c5 g5h6 g7h6 d4b5 c5e7');
INSERT INTO evaluations VALUES(31565,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(31565,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(31565,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(31565,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(31565,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(31565,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(31565,36,'cp 13','g1f3 b8d7 f1d3 f5d3 d1d3 g8f6 c3e4 f8e7 e1g1 h7h6 c2c4 e8g8 e4f6 e7f6 c1e3 c6c5 f1d1 d8c7 d3e2 a8d8 a1c1 c5d4 f3d4 a7a6 d4f3 d7e5 f3e5 c7e5 b2b3 d8d1 c1d1 f8d8 d1d8 f6d8 g2g3 e5e4 h2h4 d8e7 h4h5 g8h8 g3g4');
INSERT INTO evaluations VALUES(31564,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(31564,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(31564,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(31564,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31564,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31564,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(31564,36,'cp 0','g8f6 f1d3 f5d3 d1d3 f8e7 e1g1 e8g8 c3e4 h7h6 c1f4 b8d7 f1d1 d8a5 e4f6 e7f6 d3b3 a5a6 a2a4 f8d8 f4c7 d8c8 c7f4');
INSERT INTO evaluations VALUES(29302,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29302,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29302,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29302,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29302,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29302,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(29302,36,'cp 0','f1d3 f5d3 d1d3 f8e7 g2g4 f6d5 c3e4 d8a5 c1d2 d5b4 d2b4 a5b4 c2c3 b4b2 a1b1 b2a2 b1b7 a2a1 b7b1 a1a5 b1b7');
INSERT INTO evaluations VALUES(41789,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(41789,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(41789,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41789,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41789,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41789,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41789,36,'cp 0','g1f3 d6f4 e3f4 d8d6 d1d2 b7b6 b1a3 e8g8 e1g1 c8a6 a3b5 a6b5 d3b5 a7a6 b5d3 c7c5 c2c3 b8c6 d2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(41790,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41790,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41790,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41790,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41790,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41790,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41790,36,'cp 0','d6f4 e3f4 d8d6 d1d2 b7b6 e1g1 e8g8 b1a3 c8a6 a3b5 a6b5 d3b5 c7c6 b5d3 c6c5 c2c3 b8c6 d2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(42102,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42102,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42102,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42102,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42102,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42102,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42102,36,'cp 0','e3f4 d8d6 d1d2 b7b6 b1a3 e8g8 e1g1 c7c5 c2c3 c8a6 a3b5 a6b5 d3b5 a7a6 b5d3 b8c6 d2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(42099,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42099,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42099,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42099,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42099,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42099,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42099,36,'cp 0','d8d6 d1d2 b7b6 b1a3 e8g8 e1g1 c7c5 c2c3 c8a6 a3b5 a6b5 d3b5 a7a6 b5d3 b8c6 d2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(42675,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(42675,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(42675,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(42675,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(42675,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42675,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(42675,36,'cp 4','e4e5 d7d5 e5f6 d5c4 d1e2 c8e6 f6g7 h8g8 c1g5 c5e7 g5e7 e8e7 e2e4 b8c6 e4h7 e7d7 b1c3 d8f6 e1c1 g8g7 h7h5 a8g8 c3d5 e6d5 h5d5 d7c8 f3d4 f6f4 d1d2 g7g2 d5f5 f4f5 d4f5 g8d8 d2e2 c6d4 f5d4 d8d4 h2h4 d4d6 h4h5 d6h6 h1h4 g2g5 h4c4 g5h5 e2e8 c8d7 e8b8 h5h1 c1d2 h6d6 d2e2 d6e6 e2d2');
INSERT INTO evaluations VALUES(42535,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42535,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42535,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(42535,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(42535,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(42535,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42535,36,'cp 12','d7d5 e5f6 d5c4 d1e2 c8e6 f6g7 h8g8 c1g5 c5e7 g5e7 e8e7 e2e4 b8c6 e4h7 e7d7 b1c3 d8f6 c3e4 f6g7 h7g7 g8g7 a1d1 e6g4 e4f6 d7e7 f6g4 g7g4 e1g1 g4f4 f1e1 e7d6 c2c3 d4d3 f3d2 c6e5 g1f1 a8h8 g2g3 f4f5 f2f4 e5g4 d2c4 d6c5 c4e3 g4e3 e1e3 h8h2 d1d3 h2b2 e3e2 b2e2 f1e2 c5b6 c3c4 b6c5 d3d7');
INSERT INTO evaluations VALUES(47135,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(47135,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(47135,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(47135,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47135,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(47135,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47135,36,'cp 10','c4d5 f6d5 b1c3 c7c5 g1f3 e7e6 c3d5 e6d5 f1b5 b8c6 e1g1 f8d6 d2d4 c5d4 f3d4 e8g8 d4c6 b7c6 b5c6 a8b8 h2h3 b8b6 c6d5 c8a6 f1e1 d6b4 e3e4 b4e1 d1e1 a6b7 c1e3 b6b2 e1c3 b2b5 c3d3 b5d5 e4d5 d8d5 d3d5 b7d5 e3a7 f8a8 a7e3 a8a2 a1a2 d5a2');
INSERT INTO evaluations VALUES(47106,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(47106,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47106,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47106,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(47106,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(47106,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(47106,36,'cp 6','f6d5 g1f3 c7c5 b1c3 e7e6 d2d4 f8e7 f1d3 e8g8 c3d5 d8d5 d1c2 h7h6 d3e4 d5d6 e1g1 b8c6 d4c5 d6c5 c2c5 e7c5 f1d1 f7f6 a2a3 f8d8 d1d8 c6d8 h2h4 c8d7 c1d2 a8c8 d2c3 c5b6 a1c1 d7a4');
INSERT INTO evaluations VALUES(47461,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(47461,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47461,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(47461,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(47461,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(47461,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(47461,36,'cp 8','g1f3 c7c5 b1c3 e7e6 d2d4 f8e7 f1d3 b8c6 e1g1 e8g8 a2a3 c5d4 e3d4 e7f6 f1e1 c8d7 c3d5 e6d5 c1e3 d7g4 h2h3 g4h5 a1c1 a7a6 c1c5 f6e7 c5c3 e7d6 g2g4 h5g6 e3g5 d8b6 d1e2 g6d3 c3d3 b6c7 e1d1');
INSERT INTO evaluations VALUES(50702,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(50702,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(50702,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(50702,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(50702,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(50702,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(50702,36,'cp 28','c2c3 b8c6 f1d3 c8g4 c1f4 g8f6 e1g1 e7e6 b1d2 f8d6 f4d6 d8d6 h2h3 g4h5 d1e2 d6c7 e2e3 e8g8 f1e1 a8c8 e1e2 h5g6 d3g6 h7g6 a2a3 a6a5 a1e1 b7b5 d2b3 f6e4 f3g5 e4g5 e3g5 b5b4 b3c5 b4a3 b2a3 c7d8');
INSERT INTO evaluations VALUES(53649,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(53649,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(53649,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(53649,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(53649,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(53649,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(53649,36,'cp 126','b7b5 c4b3 b5b4 c3e2 a7a5 g1h3 c8a6 e1g1 a5a4 b3c2 a6e2 d1e2 d8d4 h3g5 g8f6 e4e5 d4g4 f2f3 g4d4 g1h1 f6d5 f3f4 a4a3 a1b1 g7g6 f4f5 g6f5 e2h5 a8a7 g5e6 d4g4 h5g4 f5g4 e6g5 a3b2 c1b2 f8g7 b2d4 a7d7 g5e4 e8g8 d4c5 g7e5 c5f8 g8f8 c2b3');
INSERT INTO evaluations VALUES(51693,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(51693,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(51693,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(51693,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(51693,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(51693,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(51693,36,'cp 137','c4b3 b5b4 c3a4 g8f6 e4e5 f6d5 g1e2 f8e7 e2g3 e8g8 e1g1 c8a6 f1e1 b8d7 g3e4 a6b5 c1d2 a7a5 d1g4 g8h8 a1d1 b5a4 b3a4 c6c5 a4c6 a8c8 c6d5 e6d5 e4g3 f8e8 d4c5 d7c5 d2e3 e7f8 g4f3 e8e5 f3f7 d8f6 f7f6 g7f6');
INSERT INTO evaluations VALUES(56872,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(56872,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(56872,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(56872,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(56872,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(56872,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(56872,36,'cp 128','d7d5 h2h4 d5e4 c3e4 c8d7 c1g5 f8e7 g1f3 d7c6 d1e2 c6e4 e2e4 d8d5 e4e2 g8f6 g2g3 c7c6 g5e3 b8d7 f1g2 d5b5 c2c4 b5a5 f3d2 e8g8 e1g1 f8e8 d2b3 a5c7 e3d2 e7d6 a2a4 e6e5 c4c5 d6e7 d2c3 a8d8');
INSERT INTO evaluations VALUES(55383,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(55383,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(55383,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(55383,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(55383,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(55383,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(55383,36,'cp 135','h2h4 d5e4 c3e4 c8d7 c1g5 f8e7 d1e2 d7c6 g1f3 c6e4 e2e4 d8d5 e4d5 e6d5 e1c1 h7h6 g5f4 e7d6 f1b5 e8f8 f3e5 g8f6 g2g3 a7a6 b5a4 b8c6 a4c6 b7c6 h1e1 f8g7 e5c6 d6f4 g3f4 h8e8 f2f3 f6h5 f4f5 h5f4 e1e5 e8e5 d4e5 g6f5');
INSERT INTO evaluations VALUES(37530,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(37530,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(37530,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(37530,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(37530,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(37530,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(37530,36,'cp 86','f1c4 b7b5 c4b3 b8c6 e4e5 g8h6 e1g1 e8g8 b3d5 d8d7 f1e1 e7e6 d5e4 a8b8 e4c6 d7c6 c3e4 h6f5 c1g5 c8b7 c2c3 h7h6 g5f6 c6d7 d1d2 b7e4 e1e4 f5e7 f6g7 g8g7 a1e1 e7d5 h2h3 b5b4 c3b4 b8b4 a2a3 b4c4 b2b4 d7c6');
INSERT INTO evaluations VALUES(37521,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(37521,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(37521,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(37521,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(37521,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(37521,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(37521,36,'cp 67','b8c6 e1g1 g8f6 h2h3 b7b5 c4b3 e7e6 d4d5 c6e7 f1e1 e8g8 d5e6 c8e6 c1f4 f6h5 f4e3 h5f6 e4e5 d6e5 b3e6 f7e6 f3e5 d8d1 a1d1 a8d8 e5d3 f6d5 c3d5 e7d5 e3c5 f8e8 c2c3 d5f6 a2a4 f6d5 d3c1 b5b4 c3b4 g7b2');
INSERT INTO evaluations VALUES(49732,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(49732,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(49732,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(49732,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(49732,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(49732,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(49732,36,'cp -53','b1c3 c7c6 d2d4 d7d5 e4d5 c6d5 c4b5 b8c6 c1f4 f8e7 g1e2 c8g4 e1g1 e8g8 g1h1 a7a6 b5d3 c6d4 d3h7 f6h7 d1d4 e7f6 d4d3 a8c8 c3d5 g4e2 d5f6 d8f6 d3e2 f6b2 e2f3 b2c2 f3b7 h7f6 f4e5 f6e4 e5d4 c8e8 b7d5');
INSERT INTO evaluations VALUES(49730,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(49730,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(49730,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(49730,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(49730,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(49730,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(49730,36,'cp -36','c7c6 d2d4 d7d5 e4d5 c6d5 c4b5 b8c6 c1f4 f8e7 g1f3 e8g8 e1g1 f6e4 d1e2 c8g4 b5c6 e4c3 e2e1 c3e2 e1e2 b7c6 h2h3 g4f5 f3e5 a8c8 e2a6 d8b6 a6b6 a7b6 g2g4 f5e6 f1e1 c6c5 e5f3 c5d4 f3d4 e7c5 f4e3 c5d4 e3d4 c8c2 d4b6 c2d2 b6c5 f8e8 a2a4 d5d4');
INSERT INTO evaluations VALUES(44119,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(44119,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(44119,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(44119,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(44119,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(44119,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(44119,36,'cp -43','d2d4 d7d5 e4d5 c6d5 c4b3 f8b4 g1e2 e8g8 e1g1 f4f3 f1f3 c8g4 f3f6 d8f6 c3d5 f6h4 h2h3 g4e2 d1e2 b8c6 c1e3 b4d6 e2g4 h4g4 h3g4 c6a5 g4g5 f8e8 g1f2 b7b5 f2f3 a8c8 c2c3 a5c4 b3c4 c8c4 e3d2 a7a6 a1h1 c4a4 d5e3');
INSERT INTO evaluations VALUES(12460,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(12460,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(12460,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12460,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(12460,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12460,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(12460,36,'cp 15','f3d4 d8c7 e1g1 c6d4 d1d4 f8c5 d4d2 e8g8 d2e2 d7d6 c1g5 f6d7 c4b3 a7a6 a2a4 h7h6 g5f4 b7b6 a1d1 c8b7 g1h1 d7e5 f4g3 e5c6 b3c4 c7e7 f2f3 c6e5 c4b3 e5c6 g3f2 e7f6 e2d2 a8c8 f3f4 f8d8 f2g3 c5b4 f4f5 b4c3 b2c3');
INSERT INTO evaluations VALUES(13119,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(13119,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(13119,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(13119,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(13119,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(13119,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(13119,36,'cp 40','d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 c2c4 c8e6 c4d5 c6d5 b1c3 d8d7 c1g5 a8d8 e2e4 d5d4 c3d5 f8e7 d5e7 d7e7 d1a4 e7d7 a4a3 d7d6 a3d6 d8d6 f1c1 e8g8 g5d2 f8b8 b2b4 d6d8 a2a4 d8c8 f2f3 f6d7 a4a5 f7f6 g2f1 g8f7 c1c8 b8c8 f1d3');
INSERT INTO evaluations VALUES(19494,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19494,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19494,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19494,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19494,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19494,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(19494,36,'cp 9','g1f3 g8f6 e2e3 g7g6 d2d4 c5d4 f3d4 f8g7 d4c6 b7c6 f1e2 e8g8 c2c4 d7d5 e1g1 d5c4 e2c4 d8c7 d1d4 a7a5 d4e5 c7e5 b2e5 f6e4 e5g7 g8g7 f1c1 f8d8 c4e2 c8e6 c1c6 a5a4 b3a4 a8a4 c6c2 e6b3 c2b2 b3e6 f2f3 e4c5 b1d2 c5d3 e2d3 d8d3');
INSERT INTO evaluations VALUES(19491,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19491,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19491,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19491,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19491,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19491,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19491,36,'cp 4','g8f6 e2e3 g7g6 d2d4 c5d4 f3d4 f8g7 d4c6 b7c6 f1e2 d8c7 c2c4 e8g8 d1c1 c8b7 e2f3 a8c8 e1g1 d7d5 f1d1 f8d8 h2h3 h7h5 b1c3 e7e5 c3a4 f6h7 c4d5 c6d5 c1c7 c8c7 a1c1 c7c1 d1c1 h7g5 f3e2 d5d4 e3d4 e5d4 a4c5 b7d5 f2f3');
INSERT INTO evaluations VALUES(12853,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(12853,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(12853,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(12853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12853,36,'cp 0','c2c4 d7d5 c4d5 f6d5 b1c3 d5c3 b2c3 f7f6 e2e3 e7e5 f1e2 c8e6 e1g1 d8d7 a2a3 e8c8 b3b4 c8b8 d1b1 e6f5 b1b3 f5e6');
INSERT INTO evaluations VALUES(12801,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(12801,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(12801,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(12801,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12801,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12801,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12801,36,'cp 0','d7d5 c4d5 f6d5 b1c3 d5c3 b2c3 f7f6 e2e3 e7e5 d1c2 c8e6 f1d3 d8d7 a1d1 e8c8 d3e4 e6d5 e4f5 d5e6');
INSERT INTO evaluations VALUES(29511,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(29511,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(29511,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(29511,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(29511,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(29511,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(29511,36,'cp 108','c2c4 e7e6 d5e6 g4f3 d1f3 b8c6 e6f7 e8f7 c4c5 g7g6 f3c3 d8e7 e1d1 e7c5 b2b3 c5f5 c1b2 a8d8 f1b5 c6d4 c3c7 f8e7 h1e1 f6d5 c7c4 d4b5 e1e5 f5d7 c4d5 d7d5 e5d5 d8d5 b2h8 d5f5 a2a4 b5c7 h8d4 e7f6 d4f6 f7f6 d1e2 c7e6 e2e3 e6c5');
INSERT INTO evaluations VALUES(51976,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(51976,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(51976,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(51976,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(51976,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(51976,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(51976,36,'cp 14','a7a6 c1b2 b8c6 g2g3 d7d5 e4d5 e6d5 d1e2 c8e6 f1g2 d5d4 e1g1 f8e7 f3e5 c6e5 e2e5 a8b8 c2c3 b7b6 f1e1 e7f6 e5e2 g8e7 e2a6 e8g8 b1a3 e7d5 c3d4 d5b4 a6e2 f6d4 b2c3 g8h8 c3d4 d8d4 e2e5 d4d2 e1e2 d2d4 e5d4 c5d4 a3b5');
INSERT INTO evaluations VALUES(50520,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(50520,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(50520,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(50520,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50520,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(50520,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50520,36,'cp 0','c1b2 b8c6 g2g3 d7d5 e4d5 e6d5 d1e2 c8e6 f1g2 d5d4 g2h3 d8e7 h3e6 e7e6 c2c3 e6e2 e1e2 a8d8 c3d4 c5d4 h1e1 f7f6 e2f1 e8f7 e1e4 f8c5 b2a3 c5e7 a3b2');
INSERT INTO evaluations VALUES(50521,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50521,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50521,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50521,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(50521,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(50521,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(50521,36,'cp 0','d7d5 e4d5 e6d5 d1e2 c8e6 g2g3 b8c6 d2d3 h7h6 f1h3 d8e7 h3g2 e6g4 h2h3 e7e2 e1e2 g4e6 h1e1 g8f6 e2f1 f8d6 b2f6 g7f6 b1c3 e8c8 c3e2 c6b4 e1c1 h8e8 a2a3 b4c6 c1d1 c6e5 f3e1 e5c6');
INSERT INTO evaluations VALUES(52969,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(52969,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(52969,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(52969,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(52969,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(52969,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(52969,36,'cp 99','c2c4 g8f6 b1c3 c8f5 f3h4 f5d7 e2e4 e7e5 h4f3 f8e7 h2h3 e8g8 f1d3 e5d4 f3d4 f8e8 e1g1 b8a6 d3c2 e7f8 f1e1 a6b4 c2b1 b4a6 a2a3 g7g6 c1f4 a6c5 b1c2 f6h5 f4e3 c5e6 d4e6 d7e6 d1d3 f8g7 a1d1');
INSERT INTO evaluations VALUES(52960,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(52960,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(52960,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(52960,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(52960,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(52960,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(52960,36,'cp 62','f7f5 b1c3 g8f6 d1c2 g7g6 e2e3 b8d7 b2b4 f8g7 a2a4 e7e5 f1e2 a7a5 b4b5 e5e4 f3g5 d8e7 g5h3 e8g8 c1a3 c6c5 h3f4 b7b6 a1d1 c8b7 d4c5 d7c5 e1g1 g6g5 f4d5 b7d5 c3d5 f6d5 d1d5 f5f4 e3f4 g5f4');
INSERT INTO evaluations VALUES(16572,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(16572,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(16572,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(16572,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(16572,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(16572,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(16572,36,'cp -36','f6e4 e1g1 d7d6 f1e1 f7f5 d2d4 f8e7 d4e5 d6e5 d1d8 e8d8 b1d2 e4d2 c1d2 e5e4 f3g5 d8e8 a1d1 c6b4 e1e2 b4c2 d2f4 c2b4 f2f3 h7h6 g5h3 e8f7 f3e4 h8d8 d1d8 e7d8 e4f5 c8f5 g2b7 a8b8 b7g2 d8f6 h3f2 b8e8 g2e4 f5e4 e2e4 e8e4 f2e4 b4d5 e4f6');
INSERT INTO evaluations VALUES(16855,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(16855,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(16855,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(16855,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(16855,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(16855,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(16855,36,'cp -26','e1g1 d7d6 d2d4 f7f5 f1e1 f8e7 d4e5 d6e5 d1d8 e8d8 b1d2 e4d2 c1d2 e5e4 f3g5 d8e8 a1d1 h7h6 g5h3 e8f7 f2f3 h8d8 f3e4 f5e4 h3f2 e7c5 d2e3 c5e3 e1e3 d8d1 f2d1 c8f5 g2e4 f5e4 e3e4 a8d8 e4f4 f7g6 d1e3 d8d2 f4f2 d2f2 g1f2 c6e5 f2f1 c7c6 b2b3 g6f6 f1f2 f6e6 f2e2 e5d7 e3c4');
INSERT INTO evaluations VALUES(16854,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(16854,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(16854,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(16854,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(16854,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(16854,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(16854,36,'cp -53','d7d6 d2d4 f7f5 f1e1 f8e7 d4e5 d6e5 d1d8 e8d8 b1d2 e4d2 c1d2 e5e4 f3g5 d8e8 a1d1 c8d7 f2f3 h7h6 g5h3 c6d4 f3e4 d4c2 e1f1 f5e4 d1c1 c2d4 h3f4 d7c6 c1e1 e7c5 d2e3 c5b6 f4g6 h8g8 e3f2 d4c2 e1e2 e4e3 g2c6 b7c6 e2c2 e3f2 g1g2 e8f7 g6e5 f7e6 c2e2 a8e8 e5c6 e6f6');
INSERT INTO evaluations VALUES(19149,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(19149,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(19149,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(19149,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(19149,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(19149,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(19149,36,'cp 105','d1d4 a7a6 c1f4 b7b5 f1e2 c8b7 e1c1 g8f6 e2f3 f8e7 h1e1 e8g8 e4e5 b7f3 g2f3 f6h5 f4e3 d8a5 c1b1 a8c8 c3e4 a5c7 c2c3 c8d8 e1g1 g8h8 f3f4 f7f6 e4d6 e7d6 d4d6 c7b7 b1a1 h8g8 d6b6 b7c8 b6a7 f8f7 d1d6 f6e5 f4e5 d8f8');
INSERT INTO evaluations VALUES(36753,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(36753,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(36753,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(36753,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(36753,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36753,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36753,36,'cp 0','b1d2 c7c5 c2c3 b8c6 h2h3 f6d7 e1g1 e7e5 d4e5 d7e5 f3e5 c6e5 d3e2 c8e6 d2f3 e5c6 a1c1 a7a6 b2b4 c5c4 f3d4 c6d4 e3d4 b7b5 e2f3 a6a5 a2a3 h7h5 f1e1 g8h7 d1d2 a5b4 a3b4 d8d7 f4e5 g7h6 e5f4');
INSERT INTO evaluations VALUES(36754,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36754,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36754,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36754,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(36754,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(36754,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36754,36,'cp -14','f6h5 f4g5 c7c5 c2c3 c5d4 c3d4 h7h6 g5h4 b8c6 e1g1 d8d6 a2a3 c8d7 d2b3 e7e5 d4e5 c6e5 d3e2 d7g4 a1b1 f8c8 b3d4 e5f3 e2f3 g4f3 d4f3 c8c5 h4g3 h5g3 h2g3 a7a6 d1d2 c5c4 f1d1 a8d8 b2b3 c4c3');
INSERT INTO evaluations VALUES(44520,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(44520,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(44520,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(44520,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(44520,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(44520,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(44520,36,'cp -70','b8c6 d2d3 c6d4 f3d1 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 e7e6 g1f3 d7d6 b1a3 d8b6 c1e3 b6b8 e1g1 f8e7 c3c4 b5b4 a3c2 e8g8 f3d2 c8b7 f1e1 h7h6 f2f4 b7c6 e3f2 a5a4 b3a4 a8a4 a1a4 c6a4');
INSERT INTO evaluations VALUES(51611,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(51611,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(51611,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(51611,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(51611,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(51611,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(51611,36,'cp 173','g1f3 e7e6 f1d3 d7d6 e1g1 g8f6 d1e2 f8e7 a2a4 b5b4 a4a5 d6d5 b1d2 d8c7 f1e1 e8g8 e4e5 f6d7 d2b3 h7h6 c1f4 c6c5 d4c5 d7c5 b3c5 e7c5 e2d2 b8d7 f4h6 f7f6 h6e3 c5e3 d2e3 d7e5 f3e5 c7e5 e3e5 f6e5 e1e5 c8d7');
INSERT INTO evaluations VALUES(51610,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(51610,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(51610,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(51610,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(51610,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(51610,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(51610,36,'cp 157','d7d6 f1d3 g8f6 e1g1 g7g6 a2a4 b5b4 c2c3 a7a5 c3b4 a5b4 c1d2 c6c5 e4e5 d6e5 f3e5 d8d4 d3b5 c8d7 e5d7 b8d7 d2e3 d4d1 f1d1 e7e6 b1d2 f6d5 d2c4 e8c8 e3g5 f7f6 g5h4 f8e7 a4a5 c8b7 h4g3 d7b8 d1e1 e6e5 f2f4 d5c7 b5a4 d8d4 a4b3 d4d3 b3d1 b8c6 f4e5 c6e5');
INSERT INTO evaluations VALUES(55083,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(55083,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(55083,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(55083,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(55083,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(55083,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(55083,36,'cp -84','d5e4 e2e4 g8f6 e4d3 c8d7 g1f3 b8c6 d3e3 e7e5 f3e5 c6e5 e3e5 f8e7 f1b5 e8g8 b5d7 d8d7 e1g1 a8e8 h2h3 e7d6 e5c3 f6e4 c3d3 d7c6 d3f3 c6c2 d2d3 e4c5 d3d4 c5e6 b1a3 c2a4 c1e3 e6d4 f3d1 e8e4 d1a4 d4e2 g1h1 e4a4 a3b5 d6e5 f1e1 a4b4 b5a7 b4b2 a1b1 b2b1 e1b1 f8b8 a2a4 e2d4 e3d4 e5d4');
INSERT INTO evaluations VALUES(55292,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(55292,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(55292,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(55292,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(55292,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(55292,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(55292,36,'cp -73','e2e4 g8f6 e4d3 c8d7 g1f3 b8c6 d3e3 e7e5 f3e5 c6e5 e3e5 f8e7 f1b5 e8g8 b5d7 d8d7 e1g1 a8e8 b1c3 e7d6 e5b5 c7c6 b5d3 e8e5 h2h3 f8e8 b2b3 e5g5 f2f4 g5f5 c1b2 h7h5 g1h1 f5f4 f1f4 d6f4 d3d7 f6d7 a1d1 f7f6 h1g1 g7g5 g1f1 g8f7 c3e2 f4d6 b2d4 a7a5 d2d3 h5h4 d4f2');
INSERT INTO evaluations VALUES(56162,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(56162,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(56162,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(56162,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(56162,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(56162,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(56162,36,'cp 113','g8f6 f1c4 f8d6 d2d3 e8g8 g1e2 c7c6 e1g1 g8h8 d5c6 b8c6 c3e4 d6e7 e4f6 e7f6 e2c3 c8f5 a2a4 a8c8 c1e3 h7h6 c3d5 f5e6 d5f6 d8f6 c4e6 f6e6 e3d2 f8d8 f1e1 e6d7 d2c3 f7f6 h2h3 b7b6 a4a5 d7c7 a5b6 a7b6');
INSERT INTO evaluations VALUES(48604,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(48604,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(48604,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(48604,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(48604,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(48604,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(48604,36,'cp 141','f1c4 f8d6 d2d3 e8g8 g1e2 c7c6 e1g1 f6d5 c3d5 c6d5 c4d5 b8c6 c1e3 g8h8 c2c4 c6e7 d5f3 f7f5 e2c3 e7c6 c3d5 c8e6 f1e1 a8c8 a2a3 d8h4 b2b4 f5f4 e3c5 d6c5 b4c5 c6d4 e1e5 c8c5 a1b1 b7b6 e5e1 e6f7 e1e7');
INSERT INTO evaluations VALUES(48587,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(48587,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(48587,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(48587,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(48587,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(48587,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(48587,36,'cp 136','f8d6 d2d3 e8g8 g1e2 g8h8 e1g1 c7c6 d5c6 b8c6 c3e4 d6e7 e4f6 e7f6 e2c3 c8f5 c1e3 a8c8 a2a4 d8d7 h2h3 h7h6 d1h5 f5g6 h5g4 g6f5 g4g3 c6b4 a1c1 b7b6 c4b3 f8d8 g3f3 f5e6 b3e6 d7e6 f3g4 e6g4 h3g4 f6g5 e3g5 h6g5 f1d1');
INSERT INTO evaluations VALUES(8816,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(8816,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(8816,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(8816,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(8816,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(8816,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(8816,36,'cp -22','g7g5 h4g3 c8g4 c4b5 f6d7 c3d5 a7a6 b5a4 b7b5 a4b3 c6d4 c2c3 d4b3 a2b3 c7c6 d5e3 c5e3 f2e3 e8g8 e1g1 f8e8 h2h3 g4h5 b3b4 a6a5 d1c2 d8b6 c2f2 a5b4 a1a8 e8a8 c3b4 h5g6 h3h4 f7f6 f3h2 a8f8 f2d2 g8g7 f1a1 h6h5 h4g5 f6g5 h2f3');
INSERT INTO evaluations VALUES(8758,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(8758,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(8758,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(8758,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(8758,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(8758,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(8758,36,'cp -27','h4g3 c8g4 c4b5 f6d7 c3d5 g4e6 d5e3 c6d4 b5c4 d4f3 d1f3 c5e3 f3e3 e6c4 d3c4 d8f6 f2f3 h6h5 h2h3 g5g4 h3g4 h5g4 h1h8 f6h8 e1c1 g4f3 g2f3 b7b6 g3f2 e8c8 b2b4 h8h3 c1b2 h3e6 e3e2 d8f8 a2a3 f8g8 d1g1 g8g1 f2g1');
INSERT INTO evaluations VALUES(8762,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(8762,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(8762,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(8762,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(8762,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(8762,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(8762,36,'cp -19','c8g4 c4b5 f6d7 c3d5 a7a6 b5a4 b7b5 a4b3 c6d4 c2c3 d4b3 a2b3 c7c6 d5e3 c5e3 f2e3 d8b6 d3d4 f7f6 e1g1 c6c5 c3c4 b5b4 h2h3 g4e6 f1f2 e8c8 f2d2 c8b7 f3h2 h6h5 h2f1 g5g4 d4d5 e6f7 h3g4 f7g6 g3h4 h5g4 d1g4');
INSERT INTO evaluations VALUES(24068,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(24068,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(24068,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(24068,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(24068,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(24068,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(24068,36,'cp -6','c4b5 f6d7 c3d5 a7a6 b5a4 b7b5 a4b3 c6d4 c2c3 d4b3 a2b3 c7c6 d5e3 c5e3 f2e3 f7f6 e1g1 d8b6 d3d4 c6c5 b3b4 c5b4 c3b4 g4e6 g3f2 e8e7 d4d5 e6f7 f3d2 a6a5 b4a5 a8a5 a1c1 a5a2 c1c6 b6a7 d1c2 a7a4 c2c3 b5b4 c3d3 d7c5 c6c7 e7d8 c7c5');
INSERT INTO evaluations VALUES(20670,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(20670,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20670,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(20670,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(20670,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(20670,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(20670,36,'cp 14','c4b3 g8f6 d1e2 d8b6 e1g1 f8e7 f1d1 e6e5 d2d3 h7h6 b1a3 e8g8 a3c4 b6a6 c4e3 c8e6 e3d5 e6d5 e4d5 c6a5 c1d2 b7b6 c3c4 a5b3 a2b3 a6b7 f3h4 b7d7 e2f3 h6h5 f3g3 f8e8 a1a6 f6h7 h2h3 f7f5 f2f4 e5f4 g3f4 e7h4 f4h4');
INSERT INTO evaluations VALUES(29426,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(29426,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(29426,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(29426,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(29426,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(29426,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(29426,36,'cp 83','e5d6 e7d6 h2h3 g4f3 d1f3 f8e7 b1c3 e8g8 c1e3 c7c6 e1c1 d6d5 c4d5 b6d5 c3d5 d8d5 f3d5 c6d5 f1e2 b8c6 c1b1 f8e8 d1d3 c6a5 d3c3 a7a6 g2g4 a5c6 e2f3 a8d8 h3h4 h7h6 g4g5 h6g5 h4g5 g7g6 f3e2 c6a5 c3c1 e7f8');
INSERT INTO evaluations VALUES(29423,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(29423,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(29423,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(29423,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(29423,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(29423,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(29423,36,'cp 73','e7d6 c1e3 f8e7 h2h3 g4f3 d1f3 e8g8 b1c3 c7c6 e1c1 d6d5 c4d5 b6d5 c3d5 d8d5 f3d5 c6d5 f1e2 b8c6 d1d3 c6a5 c1b1 f8e8 e2f3 a8d8 e3f4 e7d6 f4c1 d6f8 h1d1 h7h6 d3e3 e8e3 c1e3 a5c4 e3c1 c4a5 h3h4 f8e7 h4h5 e7g5 c1g5 h6g5');
INSERT INTO evaluations VALUES(29864,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(29864,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(29864,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(29864,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(29864,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(29864,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(29864,36,'cp 84','b1c3 f8e7 h2h3 g4f3 d1f3 e8g8 c1e3 c7c6 e1c1 d6d5 c4d5 b6d5 c3d5 d8d5 f3d5 c6d5 f1e2 b8c6 d1d3 c6a5 e3f4 f8e8 e2f3 a8d8 h1d1 e7b4 a2a3 b4d6 f4d2 a5c4 d3b3 b7b6 d2c3 d6e7 c1c2 a7a6 a3a4 e7g5 d1d3');
INSERT INTO evaluations VALUES(44961,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(44961,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44961,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44961,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44961,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44961,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(44961,36,'cp 14','g1f3 g7g6 e1g1 f8g7 c2c4 e8g8 c4d5 c6d5 b1c3 f6e4 f3e5 c8f5 c3e4 d5e4 g3g4 f5e6 g2e4 b8c6 e5c6 b7c6 e4c6 a8c8 c6f3 d8d4 d1d4 g7d4 f1d1 d4g7 a1b1 e6a2 b1a1 a2b3 d1d3 b3e6 c1g5 g7b2 a1a7 b2f6 g5e3 c8b8 g4g5 e6h3');
INSERT INTO evaluations VALUES(52664,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(52664,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(52664,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(52664,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(52664,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(52664,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(52664,36,'cp 71','g1f3 d5e4 c3e4 g8f6 f1d3 f6e4 d3e4 c8g4 h2h3 g4f3 e4f3 e7e6 c1e3 b8d7 d1e2 d8a5 c2c3 f8d6 e1g1 e8g8 g2g3 a8d8 a2a4 f8e8 e2c2 a5c7 a4a5 a7a6 h3h4 f7f5 g1g2 f5f4 g3f4 d6f4 f1e1');
INSERT INTO evaluations VALUES(52663,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(52663,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(52663,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(52663,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(52663,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(52663,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(52663,36,'cp 64','g8f6 f1d3 f6e4 c3e4 d5e4 d3e4 c8g4 h2h3 g4f3 e4f3 e7e6 c1e3 b8d7 d1e2 d8a5 c2c3 f8d6 h3h4 d7f6 a2a3 f6d5 e1c1 e8c8 c1b1 d5e3 f2e3 c8b8 e3e4 a5b6 b1a2 d8d7 e4e5 d6e7 e2c4 h8d8 c4a4 a7a6 f3h5');
INSERT INTO evaluations VALUES(3788,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(3788,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3788,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3788,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3788,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(3788,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3788,36,'cp 0','c6a5 g5f6 d8f6 c3d5 f6d8 b2b4 c7c6 b4c5 a5c4 d3c4 c6d5 c4d5 d8a5 f3d2 a5c5 a2a4 c8d7 c2c4 a7a6 e1g1 b7b5 a4b5 a6b5 c4b5 d7b5 f1e1 a8a1 d1a1 c5b4 e1b1 b4d2 b1b5 f7f5 e4f5 f8f5 b5b8 g8f7 b8b7 f7g6 a1a7 d2c1 g1h2 c1f4 h2g1');
INSERT INTO evaluations VALUES(45653,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(45653,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(45653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45653,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(45653,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(45653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45653,36,'cp 0','f8d6 f1d3 e8g8 g1f3 b8d7 e1g1 e6e5 c4d5 c6d5 e3e4 d5e4 c3e4 f6e4 d3e4 e5d4 d1d4 d7f6 e4d3 c8g4 f1d1 d6h2 g1h2 d8d4 d3h7 f6h7 d1d4 g4f3 g2f3 a8d8 c1e3 d8d4 e3d4 a7a6 h2g3 f8d8 d4e3 h7f6 a1c1 f6d5 g3g4 d5f6');
INSERT INTO evaluations VALUES(40550,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40550,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40550,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(40550,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40550,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40550,36,'cp 0','c1d2 b8d7 c4c5 d6e7 f2f4 b7b6 b2b4 e8g8 f1e2 a7a5 g1f3 c8a6 e1g1 d8c7 h2h3 f8b8 e2a6 a8a6 d1c2 b8a8 a1b1 a5b4 a3b4 a6a3 g2g4 h7h6 d2c1 a3a7 c1d2');
INSERT INTO evaluations VALUES(10002,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(10002,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(10002,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(10002,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(10002,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(10002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10002,36,'cp 6','b4c3 d2c3 a7a6 b5c6 b7c6 f3e5 d8e7 e5d3 e7e4 d1e2 e4e2 e1e2 d7d6 h1e1 e8g8 c1e3 c8e6 b2b3 a6a5 a3a4 f8e8 c3c4 c6c5 e2d2 f6d7 d3f4 f7f6 h2h4 g8f7 d2c3 d7e5 f2f3 e5c6 a1d1 a8c8 e3f2 c6b4 g2g4 e6d7 e1e8 d7e8 f4d5 e8c6');
INSERT INTO evaluations VALUES(10010,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(10010,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(10010,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10010,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(10010,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10010,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(10010,36,'cp 7','d2c3 a7a6 b5c6 b7c6 f3e5 d8e7 e5d3 e7e4 d1e2 e4e2 e1e2 d7d6 h1e1 e8g8 c1e3 f8e8 e2d2 c8e6 d3f4 a6a5 f2f3 f6d7 b2b3 c6c5 c3c4 f7f6 h2h4 g8f7 a3a4 d7e5 e1e2 e5g6 f4g6 f7g6 a1e1 g6f7 e3f4 e6d7 d2c1 e8e2 e1e2 h7h5 f4d2');
INSERT INTO evaluations VALUES(10008,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(10008,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10008,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(10008,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10008,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10008,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10008,36,'cp 0','a7a6 b5c6 b7c6 f3e5 d8e7 e5d3 e7e4 d1e2 e4e2 e1e2 d7d6 h1e1 e8g8 c1e3 a6a5 b2b3 f6d7 e2d2 f7f6 c3c4 c6c5 f2f3 f8e8 d2c3 g8f7 h2h4 d7f8 a3a4 f8e6 e1e2 c8b7 h4h5 g7g5 h5g6 h7g6 a1e1 e6g7 d3f4 g7f5 e3d2 e8e2');
INSERT INTO evaluations VALUES(24915,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(24915,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(24915,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(24915,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24915,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(24915,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(24915,36,'cp 54','d1b3 a8b8 c3c4 a7a6 f4g3 e7e6 b1c3 g4f5 g1f3 f8d6 g3h4 d8d7 h4f6 d5c4 f1c4 g7f6 c4e2 f5g6 f3h4 f6f5 g2g3 b7b5 e1g1 e8g8 a1c1 c6e7 f1d1 c7c6 b3c2 e7d5 c3d5 c6d5 c2c6 d7e7 h4g6 h7g6 c6a6');
INSERT INTO evaluations VALUES(43544,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(43544,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43544,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43544,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43544,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43544,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(43544,36,'cp 29','b8d7 a2a4 b7b6 f1e1 d8c7 c3d5 e6d5 e4d5 d7e5 f2f4 f8e7 f4e5 d6e5 d4c6 e7c5 c1e3 c5e3 e1e3 f6d5 e3e5 c8e6 d3c4 d5f4 d1f3 e8g8 f3f4 c7c6 c4e6 f7e6 f4e4 a8c8 c2c3 c6e4 e5e4 c8c6 a1d1 h7h6 h2h4 a6a5 g2g3 f8f7 g1g2 g8f8 h4h5 f7c7 e4g4 e6e5 d1d5 c7e7 d5b5 f8f7');
INSERT INTO evaluations VALUES(49190,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(49190,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(49190,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(49190,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(49190,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(49190,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(49190,36,'cp 100','e2e4 d7d6 c1g5 f8g7 h2h3 e8g8 f2f4 d8d7 f1e2 b8c6 e4e5 f6e8 g1f3 f7f6 e5f6 e8f6 e1g1 h7h6 g5f6 g7f6 d1d2 d7g7 a1d1 c8d7 c3e4 c6e7 e4f6 g7f6 f3e5 a8d8 e5g4 f6g7 e2f3 b7b6 b2b3 b6b5 c4b5 d7b5 f1e1');
INSERT INTO evaluations VALUES(49193,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(49193,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(49193,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(49193,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(49193,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(49193,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(49193,36,'cp 93','d7d6 g1f3 f8g7 f1e2 e8g8 e1g1 e6e5 c1e3 c8g4 d4e5 d6e5 f3e5 d8d1 e2d1 g4d1 a1d1 f6e4 e5g6 h7g6 c3e4 f8e8 e4c3 g7c3 b2c3 b8c6 h2h3 a8c8 c4c5 f7f6 g2g4 c6e5 g1g2 e5c4 e3d4 g8f7 f1h1 e8e2 h3h4 e2a2 d1b1 b7b6 g4g5 f6g5 h4g5');
INSERT INTO evaluations VALUES(47587,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(47587,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(47587,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(47587,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(47587,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(47587,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(47587,36,'cp 125','f2f4 f8g7 g1f3 e8g8 e4e5 d6e5 f4e5 f6d7 c1g5 f7f6 e5f6 d7f6 d1d2 c7c5 d4d5 e6d5 c4d5 a7a6 a2a4 c8g4 f1e2 g4f3 e2f3 f8e8 f3e2 h7h6 g5f4 f6h5 f4h6 d8h4 e1d1 h5g3 h6g5 h4h7 h1e1 g3e2 e1e2 h7h5 h2h4 g8h7 d2d3 b8d7 d3f3 e8e2 f3e2 a8e8 e2h5 g6h5 a1c1 e8f8 d5d6 f8f2 c3e4 f2g2 e4c5');
INSERT INTO evaluations VALUES(54665,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(54665,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(54665,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(54665,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(54665,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(54665,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(54665,36,'cp 87','c2c3 b8c6 f1b5 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 c8d7 b1c3 g8h6 e5f6 d8f6 e1g1 e8g8 a1e1 a8c8 h2h3 h6f5 b5c6 d7c6 e1e2 c6d7 f3e5 d7c6 f1e1 f6h6 d2h6 g7h6 e5c6 c8c6 e2e6 f5d4 e6e7 c6b6 b2b3 b6g6 g1h1 f8f2 e7b7');
INSERT INTO evaluations VALUES(54663,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(54663,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(54663,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(54663,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(54663,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(54663,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(54663,36,'cp 75','c5d4 c3d4 b8c6 f1b5 f8b4 c1d2 b4d2 d1d2 c8d7 b1c3 g8h6 e1g1 e8g8 h2h3 f6e5 b5c6 d7c6 f3e5 d8h4 f1e1 a8e8 a1c1 h6f7 e5f3 h4f6 b2b3 f7d6 c3e2 d6e4 d2e3 e8c8 c1c2 f6h6 e3h6 g7h6 e2g3 e4g3 f2g3');
INSERT INTO evaluations VALUES(55142,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(55142,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(55142,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(55142,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(55142,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(55142,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(55142,36,'cp -66','f3f4 c7c5 g1f3 b8c6 e2e3 a7a6 g2g3 g8f6 f1g2 e7e6 b2b3 b7b5 e1g1 f8e7 d2d3 c8b7 d1e2 e8g8 b1d2 a8c8 c2c4 d5d4 e3e4 f6g4 a2a4 g4e3 c4b5 c6a5 a1a2 e3f1 g2f1 a6b5 a4b5 f7f6');
INSERT INTO evaluations VALUES(55655,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(55655,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(55655,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(55655,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(55655,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(55655,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(55655,36,'cp -26','g8f6 e4e5 c3b2 c1b2 d6e5 d1d8 e8d8 f3e5 f8b4 b1c3 d8e7 e5f7 h8e8 e1c1 e7f8 a2a3 b4c5 c3d5 c8e6 d5f6 g7f6 c4e6 e8e6 f7d8 e6b6 h1e1 c5d6 b2d4 d6a3 c1c2 b6d6 d8b7 d6c6 d4c3 b8a6 e1e3 a8e8 b7a5 c6e6 a5c4 a3c5 e3h3 e6e2 d1d2 e2d2 c2d2 f8g7 h3f3 e8e6 c4e3 c5e3 f3e3');
INSERT INTO evaluations VALUES(47925,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(47925,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(47925,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(47925,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(47925,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(47925,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47925,36,'cp -28','e4e5 c3b2 c1b2 d6e5 d1d8 e8d8 f3e5 f8b4 b1c3 d8e7 e5f7 h8e8 e1c1 e7f8 c3b5 c8e6 c4e6 e8e6 f7d8 e6e7 a2a3 b8d7 d8b7 a8b8 a3b4 b8b7 b5c3 d7e5 b4b5 e5c4 b2a1 c4d6 h1e1 e7e1 d1e1 d6b5 c3b5 b7b5 a1f6 g7f6 e1e6 b5b6 e6e4 b6c6 c1d2 a7a5 e4h4 f8g8 h4g4 g8f7 g4h4 f7g6 h4g4 g6h5 g4g8 h5h6 g8a8 c6c5 h2h3 h6g7 g2g3 h7h5 a8a7 g7g6 d2d3');
INSERT INTO evaluations VALUES(2375,30,'cp 223',NULL);
INSERT INTO evaluations VALUES(2375,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(2375,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(2375,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(2375,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(2375,35,'cp 227',NULL);
INSERT INTO evaluations VALUES(2375,36,'cp 225','f8b4 f2f3 b4c3 b2c3 f6d7 e5d3 b7b6 d3f4 d7c5 c1e3 d8e6 e1c1 e6f4 e3f4 c8a6 f1a6 c5a6 d1d5 e8g8 h2h4 f8e8 h1d1 f7f6 h4h5 h7h6 c3c4 a8c8 d5d7 g8f8 f4g3 a6c5 d7c7 c8c7 g3c7 e8c8 c7d6 f8e8 d1d5 c5b7 c4c5 b6c5 d6f4 c8c6 d5d3 c6a6 d3b3');
INSERT INTO evaluations VALUES(2367,30,'cp 235',NULL);
INSERT INTO evaluations VALUES(2367,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(2367,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(2367,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(2367,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(2367,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(2367,36,'cp 233','f2f3 f6d7 e5d3 b4c3 b2c3 b7b6 d3f4 d7c5 c1e3 d8e6 e1c1 e6f4 e3f4 c8a6 f1a6 c5a6 d1d5 e8g8 h2h4 f8e8 h1d1 h7h6 g2g4 a8c8 h4h5 g8f8 f4e3 f7f6 a2a4 a6b8 c3c4 b8c6 e3f4 c6a5 c4c5 a5b7 c5c6 b7a5 d5d7 a5c6');
INSERT INTO evaluations VALUES(9035,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(9035,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(9035,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(9035,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(9035,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(9035,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(9035,36,'cp 76','e8g8 f1e1 b4d6 c4b5 a7a6 b5c6 b7c6 e3e4 f6d7 c1g5 f7f6 g5h4 c6c5 e4e5 c5d4 e5f6 d7f6 f3d4 d8e8 h4g3 a8b8 b2b3 e6e5 d4f3 e8c6 a1c1 c8g4 g3e5 g4f3 g2f3 d6e5 e1e5 b8d8 d1e2 f6d7 e5e7 f8f3 e2e4 c6g6 e4g6 h7g6 c3e4 f3f4');
INSERT INTO evaluations VALUES(3697,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(3697,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(3697,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(3697,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(3697,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(3697,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(3697,36,'cp 79','f1e1 b4d6 c4b5 a7a6 b5c6 b7c6 e3e4 f6d7 e4e5 d6e7 d1c2 c6c5 d4c5 d7c5 e1d1 c5d7 c1f4 c7c5 d1d3 d8c7 a1d1 d7b6 f3g5 e7g5 f4g5 c7e5 g5e7 f8e8 d3e3 e5h5 c3e4 b6d5 e3h3 h5f5 h3f3 f5h5 e7c5 c8b7 f3b3 b7c6 b3g3 g8h8 e4d6 d5f4 d1e1');
INSERT INTO evaluations VALUES(40678,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(40678,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(40678,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(40678,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(40678,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(40678,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(40678,36,'cp 70','d2d4 e5d4 f3d4 e8g8 f1e1 f8e8 c1f4 e7f8 h2h3 b8d7 a2a4 c7c6 a4a5 d7c5 d1f3 c5d7 c4b3 d7c5 b3a2 c5e6 a2e6 c8e6 a5a6 b7a6 a1a6 e6c8 a6c6 c8b7 c6c4 d6d5 c4c7 d5e4 f3e3 e8e7 d4b5 e7c7 b5c7 a8c8 e1d1 d8e7 f4d6 e7d7 d6f8 d7c7');
INSERT INTO evaluations VALUES(40694,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(40694,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(40694,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(40694,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(40694,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(40694,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(40694,36,'cp 52','e5d4 f3d4 e8g8 h2h3 b8c6 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 d7c6 e2f3 c6c4 d4d2 a8d8 b2b3 c4b4 f4e3 b7b5 f1e1 f8e8 a4b5 a6b5 g2g4 c7c5 f3g2 c5c4 e3d4 c4b3 c2b3 b4b3 d1b1 b3c4 b1b5');
INSERT INTO evaluations VALUES(40703,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(40703,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(40703,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(40703,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(40703,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(40703,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(40703,36,'cp 54','f3d4 e8g8 h2h3 f8e8 c1f4 b8d7 f1e1 e7f8 a2a4 c7c6 a4a5 d7c5 d1f3 c5d7 c4a2 d7e5 f3e2 c8d7 f4g3 b7b5 a5b6 d8b6 e2d2 e5g6 a2c4 b6b4 c4b3 a7a5 b3a4 b4b6 g3h2 a8d8 g1h1 b6b7 a1d1 c6c5 a4d7 d8d7');
INSERT INTO evaluations VALUES(18656,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(18656,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(18656,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(18656,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(18656,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(18656,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(18656,36,'cp 58','c3d5 f8e7 c2c3 g8f6 e1g1 d7d6 d2d4 e5d4 c3d4 f6d5 c4d5 e8g8 c1e3 c6b4 d5b3 d6d5 d4c5 d5e4 f3e5 b4d3 e5d3 e4d3 d1f3 e7g5 a1d1 g5e3 f2e3 d8e7 d1d3 c8e6 d3c3 a8c8 f3e4 c8c5 c3c5 e7c5 b3e6 f7e6 e4e6 g8h8 f1f8 c5f8 e6e5 f8c8 h2h3 b7b6');
INSERT INTO evaluations VALUES(20121,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(20121,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(20121,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(20121,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(20121,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(20121,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(20121,36,'cp 80','g8f6 c1g5 a7a6 b1c3 e7e6 c4b3 f8e7 e1c1 d8c7 f2f4 b7b5 c1b1 h7h6 g5f6 e7f6 d4d6 c7d6 d1d6 f6e7 d6d3 c8b7 f4f5 b5b4 c3e2 e6f5 e4f5 a8d8 b3a4 e8f8 d3d8 e7d8 e2f4 f8e7 h1d1 e7f6 d1d7 b7e4 a4b3 f6f5 g2g3 f5e5 d7f7 g7g5 f4d3 e4d3 c2d3');
INSERT INTO evaluations VALUES(13428,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(13428,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(13428,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(13428,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(13428,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(13428,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(13428,36,'cp 82','c1g5 a7a6 b1c3 e7e6 c4b3 f8e7 e1c1 d8c7 f2f4 b7b5 g5h4 e8g8 h1f1 c7a7 h4f6 e7f6 d4a7 a8a7 d1d6 f6e7 d6d3 a7d7 e4e5 f8d8 d3d7 c8d7 f1d1 b5b4 c3e2 a6a5 g2g3 g7g5 b3c4 g5f4 g3f4 a5a4 d1d3 h7h6 b2b3');
INSERT INTO evaluations VALUES(1911,30,'cp -635',NULL);
INSERT INTO evaluations VALUES(1911,31,'cp -651',NULL);
INSERT INTO evaluations VALUES(1911,32,'cp -652',NULL);
INSERT INTO evaluations VALUES(1911,33,'cp -650',NULL);
INSERT INTO evaluations VALUES(1911,34,'cp -675',NULL);
INSERT INTO evaluations VALUES(1911,35,'cp -663',NULL);
INSERT INTO evaluations VALUES(1911,36,'cp -660','f8b4 c3b4 c6b4 b1a3 b2a3 a1b1 a7a5 g2g3 a3a2 f1g2 a2c2 d1c2 b4c2 e1d2 c2b4 f3d4 b4c6 d4b5 e8d8 h1c1 g8e7 c1c3 b7b6 g2e4 h8e8 h2h4 c8a6 f2f4 a6b5 b1b5 a5a4 c3a3 d8c8 b5b1 h7h6 b1b3');
INSERT INTO evaluations VALUES(327,30,'cp -632',NULL);
INSERT INTO evaluations VALUES(327,31,'cp -628',NULL);
INSERT INTO evaluations VALUES(327,32,'cp -649',NULL);
INSERT INTO evaluations VALUES(327,33,'cp -640',NULL);
INSERT INTO evaluations VALUES(327,34,'cp -632',NULL);
INSERT INTO evaluations VALUES(327,35,'cp -640',NULL);
INSERT INTO evaluations VALUES(327,36,'cp -640','e2e4 b4c3 b1c3 b2c3 f3d2 c6e5 f1b5 g8f6 e1g1 e8g8 f1e1 d7d6 b5f1 c8g4 d1c1 g4d7 e1e3 c3c5 d2b3 c5b6 a1b1 a7a5 c2c3 b6a7 b3d4 f8e8 h2h3 e5g6 c1b2 b7b5 b2c2 b5b4 a2a4 b4c3 d4b3 c7c6 c2c3');
INSERT INTO evaluations VALUES(19202,30,'cp -401',NULL);
INSERT INTO evaluations VALUES(19202,31,'cp -423',NULL);
INSERT INTO evaluations VALUES(19202,32,'cp -389',NULL);
INSERT INTO evaluations VALUES(19202,33,'cp -398',NULL);
INSERT INTO evaluations VALUES(19202,34,'cp -412',NULL);
INSERT INTO evaluations VALUES(19202,35,'cp -409',NULL);
INSERT INTO evaluations VALUES(19202,36,'cp -425','e8f7 d1h5 g7g6 h5c5 g8f6 g1f3 e7e5 c5b5 d7d5 d2d3 f7g7 e4d5 d8d5 b5d5 f6d5 b1c3 d5b4 e1d1 c8g4 a2a3 b4a6 h2h3 g4f3 g2f3 a6c5 f3f4 e5f4 c1f4 a8d8 d1d2 f8e7 a1e1 g7f7 b2b4 c5e6 f4e3 d8d7 c3e4 h8d8 d2c1 c6d4 c2c3 d4f3 e1d1 d7d3 d1d3 d8d3');
INSERT INTO evaluations VALUES(2494,30,'cp -419',NULL);
INSERT INTO evaluations VALUES(2494,31,'cp -435',NULL);
INSERT INTO evaluations VALUES(2494,32,'cp -425',NULL);
INSERT INTO evaluations VALUES(2494,33,'cp -427',NULL);
INSERT INTO evaluations VALUES(2494,34,'cp -414',NULL);
INSERT INTO evaluations VALUES(2494,35,'cp -432',NULL);
INSERT INTO evaluations VALUES(2494,36,'cp -417','d1h5 g7g6 h5c5 g8f6 g1f3 e7e5 c5b5 f7g7 e1g1 d7d5 d2d3 f8e7 e4d5 d8d5 b5d5 f6d5 f1e1 e7f6 c1g5 f6g5 f3g5 c8f5 b1c3 a8d8 h2h3 h8e8 c3d5 d8d5 g5f3 e8e7 a1d1 h7h6 g1f1 d5b5 g2g4 f5c8 c2c3 b5b2 f3d2 b2a2');
INSERT INTO evaluations VALUES(2493,30,'cp -412',NULL);
INSERT INTO evaluations VALUES(2493,31,'cp -425',NULL);
INSERT INTO evaluations VALUES(2493,32,'cp -424',NULL);
INSERT INTO evaluations VALUES(2493,33,'cp -423',NULL);
INSERT INTO evaluations VALUES(2493,34,'cp -423',NULL);
INSERT INTO evaluations VALUES(2493,35,'cp -417',NULL);
INSERT INTO evaluations VALUES(2493,36,'cp -437','g7g6 h5c5 g8f6 g1f3 e7e5 c5b5 f7g7 d2d3 d7d5 e1g1 f8e7 e4d5 d8d5 b5d5 f6d5 f1e1 e7f6 c1g5 f6g5 f3g5 h8e8 b1a3 c8f5 a1d1 a8d8 f2f3 a7a6 g5e4 d5b4 d1d2 b7b5 e1d1 b4a2 e4c5 f5c8 c2c3 b5b4 a3b1 a6a5 d1e1 d8d5 d3d4');
INSERT INTO evaluations VALUES(2492,30,'cp -405',NULL);
INSERT INTO evaluations VALUES(2492,31,'cp -397',NULL);
INSERT INTO evaluations VALUES(2492,32,'cp -403',NULL);
INSERT INTO evaluations VALUES(2492,33,'cp -400',NULL);
INSERT INTO evaluations VALUES(2492,34,'cp -416',NULL);
INSERT INTO evaluations VALUES(2492,35,'cp -414',NULL);
INSERT INTO evaluations VALUES(2492,36,'cp -416','h5c5 g8f6 g1f3 e7e5 c5b5 f7g7 d2d3 d7d5 e4d5 d8d5 b5d5 f6d5 b1c3 d5b4 e1d1 c8g4 a2a3 b4a6 h2h3 g4f3 g2f3 a6c5 f3f4 e5f4 c1f4 a8d8 d1d2 f8e7 a1e1 h8f8 f4e3 d8d7 d2c1 g7f7 h1f1 c5e6 c3e4 b7b6 e1d1 f8d8 f2f4 e6g7 e4g5 f7g8 f1e1 d8e8 e3f2 g7h5 g5e6 e7f6 e1e4 h7h6 d3d4 e8c8');
INSERT INTO evaluations VALUES(2491,30,'cp -415',NULL);
INSERT INTO evaluations VALUES(2491,31,'cp -428',NULL);
INSERT INTO evaluations VALUES(2491,32,'cp -405',NULL);
INSERT INTO evaluations VALUES(2491,33,'cp -413',NULL);
INSERT INTO evaluations VALUES(2491,34,'cp -411',NULL);
INSERT INTO evaluations VALUES(2491,35,'cp -424',NULL);
INSERT INTO evaluations VALUES(2491,36,'cp -411','g8f6 g1f3 e7e5 c5b5 d7d5 d2d3 f7g7 e4d5 d8d5 b5d5 f6d5 b1c3 d5b4 e1d1 c8g4 a2a3 b4a6 h2h3 g4f3 g2f3 a6c5 f3f4 e5f4 c1f4 f8e7 c3d5 e7h4 f4e3 c5e6 d5c3 a8d8 c3e4 d8d7 b2b4 h4e7 b4b5 c6d4 a3a4 h8c8 c2c3 d4b3 a1a2 b3c5 e4c5 e6c5 d1e2 c5e6 h1c1 c8f8 d3d4 e6f4 e3f4 f8f4 a4a5 f4f5');
INSERT INTO evaluations VALUES(21445,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(21445,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(21445,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(21445,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(21445,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(21445,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(21445,36,'cp 96','g1f3 g8f6 e4e5 f6d7 d2b3 f8e7 c1f4 e8g8 c2c3 a7a5 f1b5 a5a4 b3d2 f7f6 e5f6 e7f6 e1g1 e6e5 b5c6 b7c6 d4e5 f6e7 f3d4 d7b8 d4e2 c6c5 f4g3 b8c6 h2h3 a4a3 b2a3 a8a3 f1e1 c8e6 e2f4 d8d7 f4e6 d7e6 d2b1 a3a8 a2a4 d5d4 b1d2 d4c3 d2e4 c6d4 e4c3 e6b3 d1b3 d4b3');
INSERT INTO evaluations VALUES(31078,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(31078,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(31078,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31078,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(31078,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(31078,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(31078,36,'cp 50','f2f3 g8f6 g2g4 f5d7 d1d2 d8a5 a2a3 e7e6 e2e4 c6c5 d4c5 f8c5 e4d5 f6d5 c3d5 a5d2 e1d2 e6d5 a1e1 e8d8 h2h4 b8c6 h4h5 h7h6 d2c1 c6d4 f1d3 a8c8 c1b1 h8e8 e1d1 d7c6 f4d2 d4e6 f3f4 f7f6');
INSERT INTO evaluations VALUES(31077,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(31077,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(31077,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(31077,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(31077,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(31077,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(31077,36,'cp 40','g8f6 g2g4 f5g6 h2h4 h7h6 e2e3 e7e6 d1d2 b8d7 g1h3 g6h7 h3f2 a8c8 f1d3 h7d3 f2d3 d7b6 b2b3 f8a3 c3e2 b6d7 e1f2 a3e7 a2a4 c6c5 a1b1 f6h7 f4g3 e8g8 f2g2 h7f6 d3f4 g7g6 h4h5');
INSERT INTO evaluations VALUES(34162,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(34162,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(34162,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(34162,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(34162,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(34162,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(34162,36,'cp 203','c8a6 d2d4 a6e2 c3e2 g8f6 e1g1 g7g6 c1f4 c7c5 d4c5 f8g7 e2d4 e6c8 c5b6 a7b6 f1e1 e8g8 c2c3 b8d7 d1d2 f6d5 f4g3 c8c4 a2a3 e7e6 h2h3 f8d8 a1d1 d8e8 d4c2 b6b5 c2e3 d5e3 d2e3 d7c5');
INSERT INTO evaluations VALUES(32726,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(32726,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(32726,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(32726,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(32726,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(32726,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(32726,36,'cp 207','d2d4 g8f6 e1g1 a6e2 c3e2 g7g6 c1f4 c7c5 d4c5 f8g7 e2d4 e6c8 c5b6 a7b6 d1e2 e8g8 c2c3 b8c6 f1d1 f6d5 f4g3 c6d4 f3d4 f8d8 a2a4 e7e5 d4b5 c8c6 h2h3 d5f4 g3f4 e5f4 d1d8 a8d8 a1d1 d8d1 e2d1 g7f6 d1e2 g8g7');
INSERT INTO evaluations VALUES(41474,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41474,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(41474,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(41474,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41474,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41474,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41474,36,'cp 0','c7c5 c2c3 b8c6 f1e2 d8b6 d1c1 c8f5 b1d2 e8g8 a2a4 c5d4 e3d4 c6b4 a4a5 b6e6 c3b4 f8c8 c1d1 f5c2 d1c1 c2f5');
INSERT INTO evaluations VALUES(49735,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(49735,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(49735,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(49735,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(49735,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(49735,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(49735,36,'cp -93','b8c6 d4e3 f8b4 c2c3 b4a5 g1f3 e8g8 e1g1 f8e8 e3d3 f6e4 f1e1 e4f6 e1e8 d8e8 c1g5 e8e4 d3f1 d7d5 b1d2 e4g6 g5f6 d5c4 f6e5 c8f5 f1c4 f5e6 c4a4 a8d8 d2f1 a5b6 e5g3 g6d3 a1e1 d3d5 a2a3 f7f6 g3f4 c6a5 h2h4 a5b3 f4e3 b3c5 e3c5 b6c5 h4h5');
INSERT INTO evaluations VALUES(16651,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(16651,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(16651,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(16651,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(16651,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(16651,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(16651,36,'cp -79','d4e3 f8b4 c2c3 b4a5 e3e2 e8g8 c1g5 c6d4 e2d1 h7h6 g5f6 d8f6 g1e2 d4e2 c4e2 c7c6 e1g1 f8d8 b1d2 d7d5 e4d5 d8d5 e2f3 d5d8 d1e2 c8e6 a2a3 d8d7 f1d1 a5c7 d2e4 f6e5 e4g3 e5e2 g3e2 a8d8 d1d7 e6d7 e2d4 g8f8 h2h4 c7b6 a1e1 b6d4 c3d4');
INSERT INTO evaluations VALUES(16641,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(16641,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(16641,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(16641,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(16641,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(16641,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(16641,36,'cp -100','f8b4 c2c3 b4a5 e3e2 e8g8 c1g5 c6d4 e2d1 h7h6 g5f6 d8f6 g1e2 d4e2 c4e2 f8d8 e1g1 d7d5 e4d5 c7c6 b1d2 d8d5 e2f3 d5d8 d1e2 a5c7 a1d1 c8e6 a2a3 d8e8 f1e1 e8e7 d2e4 f6e5 e4d6 e5e2 e1e2 a8d8 e2d2 e6b3 d6f5 e7e8 d2d8 c7d8 d1a1 d8b6 f5d4 b6d4 c3d4');
INSERT INTO evaluations VALUES(52340,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52340,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52340,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52340,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(52340,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52340,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52340,36,'cp 0','g1f3 f8g7 f1c4 b8c6 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4d6 f8e8 d6f4 c8g4 c4d5 a8c8 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(52339,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52339,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52339,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52339,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52339,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52339,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52339,36,'cp 0','f8g7 f1c4 b8c6 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4d6 f8e8 d6f4 c8g4 c4d5 a8c8 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(37875,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37875,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37875,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37875,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37875,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37875,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37875,36,'cp 0','f1c4 b8c6 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4d6 f8e8 d6f4 c8g4 c4d5 a8c8 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(37848,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37848,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(37848,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(37848,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37848,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37848,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(37848,36,'cp -3','b8c6 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4d6 f8e8 d6f4 c8g4 c4d5 a8c8 d1d2 g4f3 d5f3 d8d2 f4d2 h6f5 f1e1 c6e5 f3b7 c8b8 b7d5 b8b2 d2f4 e5f3 g2f3 e8e1 a1e1 g7c3 e1e8 g8g7 e8b8 b2b8 f4b8 a7a5 b8c7 f5d4 g1f1 d4e6 c7d6 c3d2 d5c6 e6f4 d6f4 d2f4 c6b7 f4h2');
INSERT INTO evaluations VALUES(54310,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(54310,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(54310,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(54310,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(54310,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(54310,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(54310,36,'cp -47','g1f3 b8c6 f1b5 g8e7 e1g1 c8d7 f1e1 d5d4 c3e4 e7g6 d2d3 h7h6 c2c3 a7a6 b5a4 c6e5 f3e5 g6e5 a4d7 d8d7 e4c5 f8c5 e1e5 c5d6 e5e1 d4c3 d1g4 e8g8 b2c3 a8c8 c3c4 f7f5 g4e2 e6e5 c1b2 c8e8 a1d1 f8f6 e2f3 d7f7 b2c3 e8c8 c3b2 b7b5 c4b5 f7a2');
INSERT INTO evaluations VALUES(55009,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(55009,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(55009,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55009,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(55009,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(55009,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(55009,36,'cp 0','c7c6 b5c4 c8f5 g2g4 f5g6 g1e2 g8f6 g4g5 f6d5 d2d4 e7e6 h2h4 g6h5 d1d3 h5g6');
INSERT INTO evaluations VALUES(52697,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(52697,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52697,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52697,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52697,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52697,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(52697,36,'cp 2','b5c4 g8f6 d2d4 c8f5 g1f3 e7e6 e1g1 b8d7 h2h3 h7h6 c1f4 f8e7 c4d3 f5d3 d1d3 e8g8 f1d1 d8a5 a2a3 f8e8 f4h2 a8d8 a3a4 a7a6 f3d2 a5b4 c3e4 f6e4 d2e4 d7c5 e4c5 e7c5 c2c3 b4b3 d3e2 c5d6 h2d6 d8d6 a4a5 b3d5 c3c4');
INSERT INTO evaluations VALUES(52869,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52869,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52869,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52869,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52869,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52869,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(52869,36,'cp 0','c8f5 g1f3 e7e6 d2d4 g8f6 e1g1 h7h6 c4d3 f5d3 d1d3 b8d7 f1d1 f8e7 c1f4 e8g8 h2h3 d8a5 a2a3 f8e8 f4h2 a8d8 a3a4 d7b6 h2e5 b6d7');
INSERT INTO evaluations VALUES(55654,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55654,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(55654,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(55654,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(55654,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(55654,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55654,36,'cp -11','b8c6 g1f3 g8f6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 e7d6 f3e5 d7e6 c4b5 e8f8 e5c4 d6e7 b5b3 a7a6 c1f4 b7b5 c4a5 c7c5 b3e6 c8e6 a5c6 a8c8 c6e7 f8e7 f4d6 e7e8 a1c1 c8c6 f2f3 f6d5 c3d5 c6d6 c1c5 e6d5 c5d5 d6d5 d1d5 e8e7 g1f2 h8c8 f2e2 c8c6 e2d3 c6c1 d3d2 c1f1 d2e2');
INSERT INTO evaluations VALUES(56627,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(56627,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(56627,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(56627,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(56627,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(56627,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(56627,36,'cp 208','f1c4 f8d6 e1g1 b8c6 f1e1 c8e6 c4e6 f7e6 d2d4 e5d4 e1e6 d6e7 c3b5 g8f6 c1g5 e8g8 b5d4 d8d5 e6e1 c6d4 d1d4 d5d4 f3d4 f8f7 d4f3 e7d6 g1f1 a7a5 h2h3 a5a4 a2a3 f7e7 e1e7 d6e7 a1d1 c7c6 g5e3 a8e8');
INSERT INTO evaluations VALUES(56540,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(56540,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(56540,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(56540,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(56540,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(56540,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(56540,36,'cp 200','c8e6 c4e6 f7e6 e1g1 b8c6 f1e1 d8f6 c3e4 f6g6 f3h4 g6f7 e4g5 f7f6 d1h5 g7g6 h5g4 c6d4 h4f3 g8h6 g4e4 e8c8 f3d4 d8d4 e4e5 f6e5 e1e5 d4d5 g5f3 d5e5 f3e5 h6f5 e5f3 f8g7 d2d3 h8e8 g1f1 f5d6 f3d2 g6g5 f1e2 g5g4 c2c3 h7h5 a2a4 c7c5 d2c4 d6c4 d3c4 e8f8');
INSERT INTO evaluations VALUES(12013,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(12013,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(12013,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(12013,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(12013,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(12013,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(12013,36,'cp 55','e4e5 f6d5 e1g1 d7d6 f1e1 c8g4 h2h3 g4h5 d2d3 e7e6 b1d2 f8e7 g2g4 h5g6 d2c4 e8g8 e5d6 e7d6 f3e5 d6c7 d1f3 f7f5 f3g3 g6e8 e1e2 d5b6 c1d2 b6c4 d3c4 f5g4 h3g4 c7e5 g3e5 d8d4 e5d4 c5d4 e2e6 c6c5 a1e1 e8d7 e6e4 d7c6');
INSERT INTO evaluations VALUES(12009,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(12009,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(12009,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(12009,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(12009,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(12009,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(12009,36,'cp 64','f6d5 e1g1 d5c7 b1c3 g7g6 d2d4 c5d4 d1d4 c7e6 d4h4 c8b7 f1e1 c6c5 c3e4 a8b8 b2b3 b7e4 h4e4 f8g7 c2c4 e8g8 h2h4 d7d6 c1b2 b8b6 a1d1 b6a6 h4h5 a6a2 b2a1 a2a6 e4h4 f8e8 h5h6 g7h8 f3g5 e6g5 h4g5 d8d7 e5d6 a6a1 d1a1');
INSERT INTO evaluations VALUES(12018,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(12018,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(12018,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(12018,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(12018,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(12018,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(12018,36,'cp 80','e1g1 d7d6 f1e1 c8g4 h2h3 g4h5 d2d3 e7e6 b1d2 f8e7 d2c4 e8g8 g2g4 h5g6 e5d6 e7d6 f3e5 d6e5 c4e5 d8d6 e5c4 d6e7 d1e2 e7h4 e2f3 f7f5 e1e2 f5g4 f3g4 h4h5 c4e5 f8f5 e5g6 h7g6 g4h5 f5h5 e2e6 d5b4 e6e2 h5h3 g1g2 h3h4 c1e3 h4g4 g2f3');
INSERT INTO evaluations VALUES(21980,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21980,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21980,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(21980,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(21980,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(21980,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(21980,36,'cp 32','f1g2 f7f5 e4f5 c8f5 d2d4 f8e7 c2c3 g8f6 d4d5 c6b8 f3g5 c7c6 c3c4 h7h6 g5e6 f5e6 d5e6 b8a6 e1g1 e8g8 b1c3 a6c7 g2h3 d6d5 c4d5 c6d5 d1e2 d8d6 f2f4 e5e4 f4f5 d6c6 h3g2 e7d6 g1h1 a8e8 c1e3 d6c5 a1d1');
INSERT INTO evaluations VALUES(21981,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(21981,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(21981,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(21981,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(21981,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(21981,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(21981,36,'cp 31','f7f5 e4f5 c8f5 d2d4 f8e7 c2c3 g8f6 d4d5 c6b8 f3g5 e8g8 e1g1 b8d7 c3c4 a7a5 g5e6 f5e6 d5e6 d7c5 c1e3 c5e6 g2b7 a8b8 b7g2 b8b2 b1c3 d8e8 c3d5 g8h8 d1c1 b2b8 a1b1 b8b1 c1b1 f6d5 c4d5 e6d4 e3d4 e5d4 b1c2 e7f6 g2e4 g7g6 f1b1 f6g7 b1b7 d4d3 c2d3');
INSERT INTO evaluations VALUES(1953,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(1953,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(1953,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(1953,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(1953,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(1953,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(1953,36,'cp 63','f3g3 d7d6 d2d4 c8d7 e1g1 f6g6 f2f3 g6g3 h2g3 f7f6 a2a4 g8e7 a4a5 d6d5 e4d5 e5d4 c3d4 e7f5 f1d1 f8d6 c1f4 e8c8 a5a6 b7b6 f4d6 f5d6 c4f1 d7b5 b1c3 b5f1 g1f1 d8e8 f1f2 h7h5 a1c1 c8d7 d1e1 e8e1 c1e1');
INSERT INTO evaluations VALUES(1954,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(1954,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(1954,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(1954,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(1954,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(1954,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(1954,36,'cp 56','d7d6 d2d4 h7h5 h2h4 c8d7 c1e3 g8e7 b1d2 f6g6 g3g6 e7g6 g2g3 f7f5 e4f5 d7f5 e1g1 c7c6 f1e1 e8d7 c4f7 e5d4 e3d4 c6c5 d4e3 g6e5 f7d5 f8e7 e3c5 e5d3 e1e7 d7e7 c5d4 d3b2 d4g7 h8e8 d5b7 a8b8 b7f3 f5g4 g7d4');
INSERT INTO evaluations VALUES(4348,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4348,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4348,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4348,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4348,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4348,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4348,36,'cp 0','d7d6 d2d3 a7a5 c1e3 c5d4 e1g1 d4c3 b2c3 c8e6 c4b5 c6e7 c3c4 e7g6 b5a4 d8e7 a1b1 b7b6 f3d2 e6d7 a4d7 f6d7 d1g4 e7e8 f1e1 g6e7 e3h6 e7g6');
INSERT INTO evaluations VALUES(14127,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(14127,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(14127,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14127,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(14127,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(14127,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14127,36,'cp 15','g5h4 g7g5 h4g3 f6h5 e2e3 h5g3 h2g3 f8g7 f1b5 d8d6 d1d3 c8d7 e1c1 e8c8 b5c6 d6c6 f3e5 g7e5 d4e5 d7e6 d3d4 c6b6 d4b6 a7b6 h1h2 h8h7 b2b3 c7c6 a2a4 h6h5 f2f3 c8c7 c1b2 e6f5 c3e2 f5g6 d1h1 d8h8 e2d4 c6c5 d4b5 c7c6 b5c3 e7e6 e3e4');
INSERT INTO evaluations VALUES(14142,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(14142,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(14142,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14142,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(14142,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(14142,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14142,36,'cp 0','g7g5 h4g3 f6h5 e2e4 d5e4 c3e4 g5g4 d4d5 c6b4 c2c4 c7c6 d1a4 g4f3 a4b4 c6d5 c4d5 f3g2 f1g2 d8b6 b4b6 a7b6 e1g1 h5g3 h2g3 f8g7 f1d1 g7b2 a1b1 b2e5 d5d6 e5d6 e4d6 e7d6 b1b6 e8e7 d1d6 a8a2 d6d4 h8d8 d4d8 e7d8 b6h6 c8e6');
INSERT INTO evaluations VALUES(16935,30,'cp 462',NULL);
INSERT INTO evaluations VALUES(16935,31,'cp 443',NULL);
INSERT INTO evaluations VALUES(16935,32,'cp 446',NULL);
INSERT INTO evaluations VALUES(16935,33,'cp 440',NULL);
INSERT INTO evaluations VALUES(16935,34,'cp 463',NULL);
INSERT INTO evaluations VALUES(16935,35,'cp 467',NULL);
INSERT INTO evaluations VALUES(16935,36,'cp 474','f3e5 d7d6 e5c4 c7c5 f1d3 f8e7 c4e3 c8d7 e1g1 e8g8 c2c3 d4b5 f2f4 b5c7 c3c4 b7b5 e4e5 b5c4 d3c4 c7b5 b1d2 f6e8 d2f3 b5d4 f3d4 c5d4 d1d4 d8b6 d4d3 b6c7 d3c3 e8f6 e5f6 e7f6 c3d3 c7b6 a1b1 f8e8 g1h1 a8c8 b2b3 b6d4 a2a3 g7g6 d3d4 f6d4 e3c2');
INSERT INTO evaluations VALUES(16923,30,'cp 468',NULL);
INSERT INTO evaluations VALUES(16923,31,'cp 482',NULL);
INSERT INTO evaluations VALUES(16923,32,'cp 477',NULL);
INSERT INTO evaluations VALUES(16923,33,'cp 462',NULL);
INSERT INTO evaluations VALUES(16923,34,'cp 476',NULL);
INSERT INTO evaluations VALUES(16923,35,'cp 488',NULL);
INSERT INTO evaluations VALUES(16923,36,'cp 481','d7d6 e5c4 c7c5 f1d3 f8e7 c4e3 c8d7 b1a3 a7a6 e1g1 d4b5 a3c4 b5a7 b2b3 b7b5 c4d2 e8g8 c1b2 a7c8 g1h1 c8b6 f2f4 a8b8 d1f3 c5c4 b3c4 b5c4 d2c4 b6c4 e3c4 d7b5 a1b1 b5c4 d3c4 f6d7 e4e5 b8b2 b1b2 d6e5 d5d6 e7d6 b2b7 e5f4');
INSERT INTO evaluations VALUES(25872,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(25872,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(25872,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(25872,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(25872,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(25872,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(25872,36,'cp 154','d4d5 c6e7 d1e2 g8f6 c1d2 e6d5 e4d5 f6d5 e1c1 g7g6 c3d5 b7d5 d2c3 f8h6 c1b1 e8g8 h2h4 f7f6 h4h5 g6g5 f3d4 c7c6 e2d2 f8f7 d1e1 b6b5 a2a3 a7a6 f2f3 h6f8 c3a5 d8b8 f1d3 b8f4 d2c3 f8g7 h5h6 g7f8 e1e2 e7g6 d4f5 g8h8 h1e1 d5e6 g2g4 f4f3');
INSERT INTO evaluations VALUES(25867,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(25867,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(25867,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(25867,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(25867,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(25867,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(25867,36,'cp 171','c6e7 d1e2 a7a6 c1d2 g8f6 e1c1 b6b5 d2g5 d8b8 f3e5 e7g6 f2f4 f8e7 h2h4 c7c6 d5e6 d7e6 e5g6 h7g6 e4e5 f6d5 c3e4 c6c5 e2f2 b8c7 g5e7 e8e7 e4c5 a8c8 c5b7 c7b7 f1d3 b7c7 d3e4 c7c5 f2c5 c8c5 e4d5 e6d5 d1d4 e7e6 g2g3 c5c4');
INSERT INTO evaluations VALUES(27996,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(27996,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(27996,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(27996,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(27996,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(27996,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(27996,36,'cp 110','d7d6 a2a4 b8d7 a4a5 g8f6 d1e2 a7a6 e1g1 e8g8 f1e1 e6e5 a5b6 f8e8 b6c7 d8c7 d4e5 d6e5 f3d2 c7c3 d2b3 c3c7 e2d2 b7c6 d2a5 c7c8 c1a3 f6h5 g2g3 h5f6 b3d2 c8d8 a5d8 e8d8 a3e7 d8e8 e7f6 d7f6 a1a6 a8a6');
INSERT INTO evaluations VALUES(33716,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(33716,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(33716,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33716,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(33716,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(33716,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(33716,36,'cp 8','f2f3 e7e5 a2a4 a7a5 c1e3 f6d7 g1e2 f8c5 e3d2 d7b6 e2c1 c8e6 c1d3 c5d6 b1a3 f7f5 b2b3 f5e4 f3e4 b8c6 a3b5 b6d7 d1c1 d7f6 b5d6 c7d6 d3f2 c6d4 c1b2 h8f8 d2e3 a8c8 f1b5 e8e7 b5d3 f6d7 h1d1');
INSERT INTO evaluations VALUES(33714,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(33714,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(33714,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(33714,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(33714,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(33714,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(33714,36,'cp 6','e7e5 a2a4 a7a5 c1e3 c8e6 b1d2 f6d7 f1c4 f8c5 c4e6 c5e3 e6d7 b8d7 d2c4 e3d4 g1e2 d7b6 c4b6 d4b6 e2c1 h7h5 d1e2 e8c8 c1d3 f7f6 h2h4 h8h6 h1h3 h6g6 g2g3 d8d6 d3c1 b6d4 c2c3 d4c5 c1b3 b7b6 b3d2 c8b7 a1c1 d6d7 d2c4 d7d8 c1a1 b7c6');
INSERT INTO evaluations VALUES(42955,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(42955,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(42955,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(42955,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(42955,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(42955,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(42955,36,'cp 50','e7f6 g7h6 d7d5 g1f3 d8e7 f1b5 c7c6 b5e2 h8g8 e1g1 c8g4 f1e1 e4c5 g1f1 c5e6 d2d4 b8d7 b1d2 e8c8 e2d3 f6g7 h6e3 e7f6 d2b3 g4f3 e3f3 f6f3 g2f3 e6d4 b3d4 g7d4 c1f4 d4b2 a1b1 b2c3 e1e7 d8e8 e7f7 g8g7 f7g7 c3g7 d3h7');
INSERT INTO evaluations VALUES(42312,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(42312,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(42312,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(42312,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(42312,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(42312,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(42312,36,'cp 47','g7h6 d7d5 g1f3 d8e7 f1b5 c7c6 b5e2 h8g8 e1g1 c8g4 f1e1 e4c5 h6e3 c5e6 c2c3 c6c5 d2d4 b8c6 d4c5 e8c8 b1d2 e7d7 e2f1 d8e8 g1h1 h7h5 h2h3 g4f5 d2b3 f6g5 f3g5 e6g5 e3g5 g8g5 c1g5 e8e1 a1e1');
INSERT INTO evaluations VALUES(47136,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47136,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(47136,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47136,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47136,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47136,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47136,36,'cp 13','e7e6 b2b3 f8e7 b1c3 e8g8 c1b2 b7b6 c4d5 e6d5 d2d4 e7d6 f1d3 f8e8 a1c1 a7a6 f3e5 c7c5 e5f3 c8b7 e1g1 b8d7 c3e2 d8e7 d4c5 b6c5 e2g3 g7g6 d1d2 h7h5 f1e1 f6g4 h2h3 g4e5 f3e5 d6e5 b2e5');
INSERT INTO evaluations VALUES(53147,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(53147,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(53147,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(53147,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(53147,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(53147,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(53147,36,'cp -12','c8d7 b5d7 d8d7 g1f3 b8c6 e1g1 d5d4 c3e4 g8f6 d2d3 a8d8 c1d2 f6e4 d3e4 d7c7 d1e2 e7e5 f1d1 f8d6 c2c3 e8g8 a1c1 c7b6 c1b1 d6e7 c3d4 c6d4 f3d4 e5d4 e2d3 b6e6 a2a3 f8e8 b2b4 e7f8 f2f3 h7h6 h2h3 e6g6 d1e1');
INSERT INTO evaluations VALUES(30933,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(30933,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(30933,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(30933,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(30933,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(30933,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(30933,36,'cp -7','b5d7 d8d7 g1f3 e7e6 f3e5 d7c7 d2d4 f8b4 e1g1 b4c3 b2c3 g8f6 c1a3 b8d7 e5d7 c7d7 d1d3 d7a4 a3b4 a7a5 b4c5 b7b6 c5d6 f6e4 d6e5 a4c4 d3c4 d5c4 f1b1 e4c3 b1b6 e8g8 a2a4 f7f6 e5d6 f8d8 d6c5 c3e2 g1f1 e2d4 c5d4 d8d4 b6e6');
INSERT INTO evaluations VALUES(19529,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(19529,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(19529,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(19529,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(19529,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(19529,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(19529,36,'cp -21','g8f6 e2e3 d7d5 b1c3 e7e6 c4d5 e6d5 g1f3 f8b4 f1e2 e8g8 c1d2 a7a6 e1g1 b4d6 d1c2 c8e6 f1d1 a8c8 a1c1 d6b8 f3g5 h7h5 d2e1 d8d6 f2f4 b8a7 e1f2 c6b4 c2b1 e6g4 h2h3 g4e2 c3e2 c8c1 d1c1 b4c6');
INSERT INTO evaluations VALUES(12869,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(12869,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(12869,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(12869,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(12869,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(12869,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(12869,36,'cp -26','b1c3 e7e6 g1f3 d7d5 e2e3 f8b4 f1e2 e8g8 e1g1 d5c4 e2c4 d8e7 c1d2 b4d6 a2a3 h7h6 h2h3 f8d8 d1e2 c8d7 e3e4 c6e5 f3e5 d6e5 f2f4 e5d4 g1h2 e6e5 c3d5 f6d5 e4d5 d8e8 d2e3 e5f4 e3d4 e7e2 c4e2 e8e2 a1c1 g7g5 c1c7 e2d2 c7b7 d2d4 b7d7 d4d2');
INSERT INTO evaluations VALUES(12868,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(12868,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(12868,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(12868,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(12868,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(12868,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(12868,36,'cp -23','e7e6 g1f3 f8b4 c1d2 d7d5 c4d5 e6d5 e2e3 e8g8 f1e2 a7a6 e1g1 b4d6 d1c2 c8e6 f1d1 a8c8 d2e1 d8d7 e2f1 b7b5 a2a3 d6b8 h2h3 c6e5 f3e5 b8e5 c2d2 f8d8 a1c1 d5d4 e3d4 d7d4 d2d4 d8d4');
INSERT INTO evaluations VALUES(12462,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(12462,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(12462,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(12462,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(12462,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(12462,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(12462,36,'cp -12','e2e3 f8b4 c1d2 d7d5 c4d5 e6d5 f1e2 a7a6 g1f3 e8g8 e1g1 c8e6 a1c1 a8c8 e2d3 b4d6 h2h3 d6b8 f1e1 d8d6 c3e2 f6e4 e2f4 e6f5 d2c3 f8d8 d3e4 d5e4 f3d4 c6d4 c3d4 d6d7 d4c3 h7h6 d1d7 f5d7 e1d1 d7a4 d1d8 c8d8 a2a3 g8h7');
INSERT INTO evaluations VALUES(19787,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(19787,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(19787,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(19787,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(19787,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(19787,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(19787,36,'cp 61','d4c5 e7e6 b2b4 a7a5 b4b5 c6a7 e2e3 f8c5 f1e2 g8f6 e1g1 c8d7 c3c4 e8g8 a2a4 a7c8 c1b2 d8e7 b1d2 f8d8 f3e5 c8b6 d1c2 a8c8 f1c1 d5c4 d2c4 b6c4 e5c4 f6d5 c2e4 b7b6 c4e5 d7e8 b2d4 c5d4 e4d4 c8c1 a1c1');
INSERT INTO evaluations VALUES(19753,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(19753,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(19753,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(19753,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(19753,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(19753,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(19753,36,'cp 55','e7e6 b2b4 a7a5 b4b5 c6a7 e2e3 f8c5 f1e2 g8f6 e1g1 c8d7 c3c4 a7c8 a2a4 c8b6 b1d2 c5b4 c1a3 b4a3 a1a3 a8c8 a3c3 d8e7 d1a1 d5c4 d2c4 b6c4 e2c4 h7h6 f1c1 e8g8 h2h3 b7b6 c4f1 c8c3 c1c3 f8c8 f3e5');
INSERT INTO evaluations VALUES(23264,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(23264,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(23264,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(23264,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(23264,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(23264,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(23264,36,'cp -27','e5e4 f3g1 d8g5 g2g4 g5h4 d1e2 f8e7 b5c6 d7c6 b1c3 g8f6 c3e4 f6e4 e2e4 h7h5 g1e2 h4g4 e4g4 h5g4 e2g3 c8d7 d2d3 e7f6 c2c3 c6c5 c1f4 e8c8 e1c1 b7b6 h1g1 a7a6 d1e1 h8h3 e1e3 d8h8 g3e4 d7f5 e4f6 g7f6 f4g3');
INSERT INTO evaluations VALUES(23269,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(23269,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(23269,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(23269,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(23269,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(23269,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(23269,36,'cp -32','d1e2 d8e7 b5c6 b7c6 f3d4 g8f6 e1g1 e7c5 c2c3 a7a5 f1e1 c8a6 d2d3 e8f7 c1g5 a6d3 e2d1 c5d5 g5f6 f7f6 b1d2 a8e8 d2f1 d3f1 e1f1 h7h5 d1e2 f8d6 a1d1 f6f7 a2a3 h8h6 b2b4 a5b4 c3b4 f7g8 e2c2 c6c5 b4c5 d5c5 c2b3 g8f8 f1e1 h5h4');
INSERT INTO evaluations VALUES(23270,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(23270,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(23270,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(23270,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(23270,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(23270,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(23270,36,'cp -22','d8e7 b5c6 b7c6 f3d4 g8f6 e1g1 e7c5 c2c3 a7a5 d2d3 c8a6 f1e1 f8e7 e2d1 a6d3 f2f3 d3b1 a1b1 e4f3 c1e3 c5d5 e3f4 e8f7 d4f3 h8e8 d1d5 f6d5 f4d2 f7g8 g1f1 a5a4 c3c4 d5f6 e1e2 e7d6 b1e1 e8e2 e1e2 a8b8 d2c3 d6b4 c3d4 c6c5 d4f6');
INSERT INTO evaluations VALUES(2015,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(2015,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(2015,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(2015,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(2015,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(2015,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(2015,36,'cp -14','b5c6 b7c6 f3d4 g8f6 e1g1 e7c5 c2c3 a7a5 d2d3 c8a6 f1e1 f8e7 e2d1 a6d3 f2f3 d3b1 a1b1 e4f3 c1e3 c5d5 e3f4 f3f2 g1f2 e8f7 f4c7 a5a4 f2g1 e7c5 c7e5 h8e8 e5f6 e8e1 d1e1 g7f6 b1d1 a8e8 e1h4 d5f5 g1h1 c5d4 h4d4 e8e4 d4a7 f7g6 h2h3 d7d5 a7a6 f5e6 d1g1 e4e2 a6a4 e2b2 g1a1 c6c5 a1g1 b2d2');
INSERT INTO evaluations VALUES(2011,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(2011,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(2011,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(2011,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(2011,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(2011,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(2011,36,'cp -31','b7c6 f3d4 g8f6 e1g1 e7c5 c2c3 a7a5 f1e1 c8a6 d2d3 f8e7 e2d1 a6d3 f2f3 d3b1 a1b1 e4f3 c1e3 c5d5 e3f4 f3f2 g1f2 e8f7 f4c7 a5a4 f2g1 h8e8 c7e5 e7c5 e5f6 e8e1 d1e1 g7f6 e1h4 a8g8 h4h7 f7e8 h7h5 e8d8 h5e2 d5a2 b1e1 d8c7 g1h1');
INSERT INTO evaluations VALUES(29459,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(29459,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(29459,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(29459,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(29459,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(29459,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(29459,36,'cp 89','f1e2 b8c6 e1g1 d6e5 f3e5 g4e2 d1e2 c6d4 e2e4 d5f6 e4b7 d8b8 b7a6 b8b5 a6b5 d4b5 e5c6 e7e6 b1d2 a7a6 d2b3 b5d6 b3a5 f6d5 f1e1 d6f5 c1d2 d5e7 c6b4 e7g6 a2a3 f8d6 a1d1 e8g8 c2c4 g6e7 b4c6 e7c6 a5c6 f5e7 c6e7 d6e7 g1f1');
INSERT INTO evaluations VALUES(29458,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(29458,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(29458,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(29458,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(29458,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(29458,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(29458,36,'cp 77','b8c6 c2c4 d5b6 e5e6 f7e6 b1d2 g4f5 e1g1 g7g6 d2b3 d6d5 c4d5 d8d5 c1f4 f8g7 a1c1 e8g8 f4g3 e6e5 c1c5 d5d8 e2b5 b6d7 c5c1 e5e4 f3g5 e7e5 d4d5 d8g5 d5c6 b7c6 c1c6 d7b6 c6c7 a7a6 b5e2 a8d8 d1c1 g5c1 c7c1 g7h6');
INSERT INTO evaluations VALUES(25031,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(25031,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(25031,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(25031,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(25031,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(25031,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(25031,36,'cp 91','c2c4 d5b6 e5e6 f7e6 b1d2 g4f5 e1g1 g7g6 d2b3 d6d5 c4d5 d8d5 c1f4 f8g7 a1c1 e8g8 f4g3 e6e5 f3e5 c6e5 c1c5 d5f7 g3e5 c7c6 e5g7 f7g7 e2f3 g7f6 b3a5 a8b8 f1e1 b6d5 d1d2 f8d8 a2a3 f5e6 h2h3 d5f4 e1e4');
INSERT INTO evaluations VALUES(38627,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(38627,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(38627,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(38627,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(38627,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(38627,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(38627,36,'cp -12','g8f6 c1g5 e8g8 e1g1 c8e6 b1d2 b8d7 f1e1 f8e8 e2d3 c7c6 c2c3 d8c7 g5h4 e6g4 d1c2 g4h5 a2a3 h5g6 g2g3 e8e1 a1e1 g6d3 c2d3 a8e8 e1e8 f6e8 d3e3 d7f6 h4f6 e8f6 c3c4 d5c4 d2c4 h7h6 e3e2 c7d7 f3e5 d6e5 d4e5 d7e6 e2d3');
INSERT INTO evaluations VALUES(42119,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(42119,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(42119,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42119,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(42119,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(42119,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(42119,36,'cp 23','b8c6 e2e3 e8g8 c1d2 f8e8 a2a3 b4f8 f1d3 h7h6 e1g1 d5c4 d3c4 f8d6 c3e2 d8e7 d2c3 e6e5 d4e5 c6e5 f3e5 d6e5 c3e5 e7e5 a4b3 e5e7 a1c1 c7c6 e2g3 b7b5 c4e2 c8e6 b3c2 a8d8 e3e4 e6g4');
INSERT INTO evaluations VALUES(9021,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(9021,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(9021,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(9021,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(9021,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(9021,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9021,36,'cp 13','e2e3 e8g8 a4c2 f8e8 c1d2 b4f8 a2a3 h7h6 c4d5 e6d5 f1d3 c8g4 e1g1 a8c8 c3e2 f6e4 e2g3 e4d2 f3d2 c6b8 b2b4 b8d7 h2h3 g4e6 d2b3 c7c6 g3e2 b7b6 e2f4 c6c5 d4c5 b6c5 b4c5 d7c5 b3c5 c8c5 c2e2 e6d7 a1c1 c5c1');
INSERT INTO evaluations VALUES(9019,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(9019,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(9019,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9019,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(9019,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(9019,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(9019,36,'cp 23','e8g8 a4c2 f8e8 c1d2 b4f8 f1e2 e6e5 d4e5 c6e5 c4d5 f6d5 a2a3 d5c3 d2c3 e5f3 e2f3 c7c6 e1g1 c8e6 f1d1 d8e7 b2b4 e8d8 g2g3 h7h6 h2h4 d8d1 a1d1 a8d8 c3d4 a7a6 d4c5 d8d1 c2d1 e7d7 d1d7 e6d7 g1f1 f8c5 b4c5 d7e6 h4h5 a6a5 f1e2');
INSERT INTO evaluations VALUES(4481,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(4481,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(4481,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(4481,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(4481,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(4481,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(4481,36,'cp 60','d4d5 c6e7 e1g1 d7d6 f1e1 c7c6 d5c6 e7c6 a1c1 c6e5 h2h3 h7h6 a2a3 f8e8 c4d5 f6d5 e4d5 e5d3 e1e8 d8e8 c1c3 d3f4 c3e3 e8b5 f3d4 b5d5 e3e8 g8h7 d2f3 b7b6 d1c2 g7g6 c2d2 c8b7 e8a8 b7a8 d2f4');
INSERT INTO evaluations VALUES(4415,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(4415,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(4415,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(4415,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(4415,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(4415,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(4415,36,'cp 36','c6e7 e1g1 d7d6 a1c1 c7c6 d5c6 e7c6 d1b3 c6a5 b3a3 a5c4 c1c4 c8e6 c4d4 d6d5 h2h3 f8e8 e4e5 f6d7 f1d1 d8c7 d1e1 h7h6 d2f1 c7b6 d4b4 b6c5 b4b7 c5a3 b2a3 e8c8 f1e3 a7a5 f3d4 d7e5 e3f5 e6f5 d4f5 e5g6 b7d7 c8c3 f5e7 g6e7 e1e7 c3a3 e7f7');
INSERT INTO evaluations VALUES(16486,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(16486,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(16486,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(16486,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(16486,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(16486,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(16486,36,'cp 38','d4d5 c6b8 f1d3 c7c6 c3c4 f8b4 c1d2 b8a6 a2a3 b4d2 b1d2 a6c5 d3c2 d7d6 b2b4 c5d7 g1e2 a7a5 e1g1 e8g8 f1e1 d8b6 a1b1 a5b4 a3b4 b6c7 c2d3 c6c5 b4b5 d7b6 e2c3 c8d7 d1b3 h7h5 b1a1 g7g6 h2h3 a8a1 e1a1 f8a8 a1a8 b6a8');
INSERT INTO evaluations VALUES(16337,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(16337,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(16337,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(16337,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(16337,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(16337,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(16337,36,'cp 30','c6b8 f1d3 c7c6 c3c4 f8b4 b1c3 b4c3 b2c3 b8a6 g1e2 a6c5 e2g3 e8g8 e1g1 d7d6 h2h3 b7b6 c1a3 c5d3 d1d3 c6c5 a3c1 f6d7 a2a4 f8e8 a1a3 d7f8 c1d2 f8g6 f1a1 a7a5 a1b1 f7f6 b1e1 g8h8 g1h2 c8d7 g3f5');
INSERT INTO evaluations VALUES(27546,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(27546,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(27546,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(27546,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(27546,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(27546,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(27546,36,'cp -23','c6b4 e1g1 b4c2 d1c2 g8e7 b1c3 e7f5 c2d3 f8b4 c1d2 a8c8 a2a4 h7h6 f1b1 a7a5 b2b3 e8g8 c3e2 b4d2 d3d2 c8c7 b3b4 a5b4 a4a5 b6a6 e2g3 f5g3 h2g3 d7b5 d2b4 b5d3 b1b2 f8c8 f3e1 h6h5 g1h2 d3g6 f2f3 g6f5');
INSERT INTO evaluations VALUES(27502,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(27502,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(27502,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(27502,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(27502,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(27502,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(27502,36,'cp -12','e1g1 b4c2 d1c2 g8e7 b1c3 a8c8 c2d3 e7c6 f1d1 h7h6 h2h3 b6a6 d3a6 b7a6 f3e1 f8e7 c1e3 e8g8 b2b3 c6b4 d1c1 c8c7 a2a3 f8c8 a3b4 c7c3 c1c3 c8c3 e3d2 c3b3 a1a6 e7b4 d2b4 b3b4 a6a7 b4b1 a7d7');
INSERT INTO evaluations VALUES(27501,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(27501,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(27501,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(27501,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(27501,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(27501,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(27501,36,'cp -8','b4c2 d1c2 a8c8 b1c3 g8e7 c2d3 e7c6 f1d1 b6a6 d3a6 b7a6 c1e3 c8b8 d1d2 f8e7 e3g5 e7b4 f3e1 e8g8 e1d3 f8c8 g5e3 b4e7 a1c1 h7h6 d2c2 c6b4 d3b4 e7b4 c3e2 b4a5 e2f4 c8c2 c1c2 d7b5 b2b3 a5b6 c2c3 a6a5 f4d3 b5d3');
INSERT INTO evaluations VALUES(27506,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(27506,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(27506,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(27506,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(27506,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(27506,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(27506,36,'cp -10','d1c2 a8c8 b1c3 g8e7 c2d3 e7c6 c1d2 c6b4 d3e2 f8e7 f1c1 e8g8 e2e3 b6a6 f3e1 f7f6 e5f6 e7f6 c3e2 c8c1 e2c1 a6c4 c1b3 b4c2 e1c2 c4c2 d2b4 f8f7 e3c3 c2c3 b4c3 b7b6 b3d2 f7f8 d2f3 f8c8 a1e1 g8f7 h2h3 f7e7 g2g4 h7h6 f3e5 a7a5 e1e3 d7a4 f2f4 a4c2');
INSERT INTO evaluations VALUES(27505,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(27505,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(27505,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(27505,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(27505,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(27505,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(27505,36,'cp -21','a8c8 b1c3 g8e7 c1d2 e7f5 c2d3 f8b4 a2a4 c8c4 a4a5 b6a6 c3e2 b4e7 b2b3 c4c7 d3a6 b7a6 f1c1 c7b7 g2g4 f5h4 f3h4 e7h4 a1b1 e8g8 e2f4 d7b5 d2b4 f8d8 b1b2 h7h5 g4h5 h4g5 b4d2 b5c4 d2e3 d8b8 b3c4 b7b2 c4d5');
INSERT INTO evaluations VALUES(34260,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(34260,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(34260,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(34260,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(34260,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(34260,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(34260,36,'cp 60','a5c5 e2e4 g8f6 b1c3 g7g6 d2e3 c5a5 g1f3 f8g7 f3d2 a5c7 c3b5 c7a5 f1e2 e8g8 e1g1 b8c6 c2c4 d7d6 b5c3 a5d8 a1c1 b7b6 a2a3 a7a5 d1e1 f6d7 f2f4 c6d4 e2d3 d7c5 d3b1 c5b3 c1d1 b3d2 e1d2 d4b3 d2d5 c8e6');
INSERT INTO evaluations VALUES(34254,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(34254,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(34254,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(34254,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(34254,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(34254,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(34254,36,'cp 50','b1c3 g8f6 e2e4 g7g6 g1f3 b8c6 d2e3 c5a5 f3d2 f8g7 d2c4 a5c7 e3g5 c7b8 a2a4 d7d6 f1d3 e8g8 e1g1 c6b4 c4e3 b8c7 g5f6 g7f6 c3d5 b4d5 e3d5 c7d8 d5f6 e7f6 f1e1 c8e6 c2c4 d8a5 b2b3 f8d8 d1f3 g8g7 d3f1 a7a6 e1d1 a8c8 f3g3 a5e5 f2f4');
INSERT INTO evaluations VALUES(34250,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(34250,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(34250,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(34250,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(34250,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(34250,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(34250,36,'cp 36','g8f6 e2e4 g7g6 g1f3 f8g7 d2e3 c5a5 f3d2 b8c6 d2c4 a5c7 e3g5 c7b8 a2a4 e8g8 d1d2 d7d6 c4e3 h7h6 g5f6 g7f6 f2f4 f6g7 f1c4 c8e6 c4e2 e6d7 e1g1 b8d8 g1h1 e7e6 c3b5 d6d5 e4e5 a7a6 b5d6 f7f6 c2c4 d5c4 e3c4');
INSERT INTO evaluations VALUES(38285,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(38285,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(38285,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38285,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(38285,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38285,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(38285,36,'cp 34','c1f4 g8f6 g1f3 e8g8 e2e3 b7b6 f1d3 c7c5 e1g1 b8c6 c4d5 f6d5 c3d5 e6d5 d3b5 c8b7 b2b3 a7a6 b5e2 f8e8 a1c1 a8c8 d1d2 c5d4 f3d4 c6d4 e3d4 e7a3 c1c8 b7c8 e2d3 d8f6 f4e5 f6e7 h2h3 c8e6 f1d1');
INSERT INTO evaluations VALUES(38283,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(38283,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38283,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(38283,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(38283,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(38283,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38283,36,'cp 34','g8f6 e2e3 e8g8 g1f3 b7b6 c4d5 f6d5 c3d5 d8d5 f1e2 d5a5 f3d2 c8a6 e1g1 c7c5 d2c4 a6c4 e2c4 c5d4 e3d4 b8c6 d4d5 e6d5 d1d5 a5d5 c4d5 a8c8 f1d1 c6b4 d5b3 f8d8 d1d8 c8d8 a1c1 g7g5 f4e3 b4d5 e3d2 d5f4 d2f4 g5f4 c1c7 g8f8 c7a7 d8d2 a7b7 d2b2 b7b6');
INSERT INTO evaluations VALUES(40947,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(40947,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40947,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(40947,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(40947,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40947,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(40947,36,'cp 31','g1f3 e8g8 e2e3 b8d7 c4c5 f6h5 a2a3 h5f4 e3f4 c7c6 f1d3 b7b6 b2b4 a7a5 e1g1 c8a6 d1c2 a6d3 c2d3 d8c7 g2g3 e7f6 f1c1 f8b8 g1g2 h7h6 c3a4 a8a7 d3d1 a5b4 a3b4 b6c5 a4c5 d7c5 b4c5 a7b7 c1b1 b7b1 a1b1');
INSERT INTO evaluations VALUES(42389,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(42389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42389,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(42389,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(42389,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(42389,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42389,36,'cp 20','e5d4 f3d4 e8g8 f1d3 b8c6 c1e3 d6e5 d4e2 d7d5 e4d5 c6e7 e3g5 d8d6 g5f6 d6f6 d1d2 f6b6 e1g1 f8e8 d2e3 b6e3 f2e3 a7a6 a2a4 c8d7 g1f2 g7g6 f1e1 c7c6 d5c6 e7c6 a1d1 e5h2 d3g6 f7g6 d1d7 c6e5 d7b7');
INSERT INTO evaluations VALUES(42400,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(42400,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(42400,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(42400,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(42400,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(42400,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42400,36,'cp 18','f3d4 e8g8 f1d3 b8c6 c1e3 d6e5 d4e2 d7d5 e4d5 c6e7 e3g5 d8d6 g5f6 d6f6 d1d2 f8e8 e1g1 f6b6 d2g5 e5f6 g5e3 b6e3 f2e3 e7g6 d3g6 h7g6 g1f2 c8f5 e2d4 f6d4 e3d4 f5c2 f1e1 c2d3 e1e5 f7f6 e5e8 a8e8 a1e1 e8e1');
INSERT INTO evaluations VALUES(42397,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(42397,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42397,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(42397,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42397,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42397,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42397,36,'cp 0','e8g8 f1d3 b8c6 d4e2 d6c5 e1g1 c6e5 c3a4 c5e7 c2c4 c7c6 a4c3 e7c5 c3a4 c5b4 a4c3');
INSERT INTO evaluations VALUES(37159,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37159,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37159,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37159,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(37159,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(37159,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(37159,36,'cp 19','c1g5 b8c6 d4f5 d6b4 f5e3 b4c3 b2c3 h7h6 g5h4 f8e8 f1d3 c6e5 f2f4 e5g6 h4f6 d8f6 e3d5 f6d6 e1g1 b7b5 a1b1 a7a6 d1d2 c8b7 c3c4 b7d5 c4d5 c7c6 d5c6 d7c6 d2f2 a8d8 e4e5 d6d4 g2g3 g6e7 a2a4 d4c3 f2f3 e7d5 f3e4 g7g6 g1h1');
INSERT INTO evaluations VALUES(43183,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(43183,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43183,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43183,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(43183,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43183,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(43183,36,'cp 12','d7d5 c4d5 f6d5 f1g2 d5c3 b2c3 c7c5 g1f3 b8c6 e1g1 c5d4 f3d4 c6d4 c3d4 d8d4 a1b1 e8g8 c1g5 d4c4 g5e7 f8e8 g2d5 c4g4 e7c5 g4e2 d1e2 e8e2 c5e3 e2b2 b1b2 g7b2 f1e1 c8f5 d5b7 a8b8 b7g2 a7a6 a2a4 b8c8 a4a5 c8d8');
INSERT INTO evaluations VALUES(41449,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(41449,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(41449,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(41449,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(41449,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41449,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41449,36,'cp 12','c4d5 f6d5 f1g2 d5c3 b2c3 c7c5 g1f3 b8c6 a1b1 c5d4 c3d4 c6d4 f3d4 d8d4 e1g1 e8g8 c1e3 d4c4 g2d5 c4a6 d1b3 a6e2 e3c5 c8e6 d5e6 e2e6 b3e6 f7e6 b1b7 f8f5 b7c7 f5d5 c5e7 g7d4 e7b4 a8b8 b4c3 d4c3 c7c3 d5a5 c3c2 b8b7 f1e1 g8f7 c2e2 b7d7 e1c1 h7h5');
INSERT INTO evaluations VALUES(13330,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(13330,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(13330,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(13330,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(13330,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(13330,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(13330,36,'cp -12','g7g6 e1g1 f8g7 a2a4 e8g8 b1a3 h7h6 f1e1 e7e5 c1d2 c8e6 c2c4 a8b8 a3c2 a7a5 c2e3 b7b6 e1f1 e6c8 f3e1 c6d4 a1b1 g8h7 e1c2 d4b3 c2a3 b3d2 d1d2 h6h5 a3b5 h5h4 f2f4 g7h6 b2b3 f6h5 f4f5 h4g3 h2g3');
INSERT INTO evaluations VALUES(13024,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(13024,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13024,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(13024,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(13024,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(13024,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(13024,36,'cp 0','a2a4 f8g7 b1a3 h7h6 c1d2 e8g8 e1g1 a8b8 c2c3 c8e6 d2e3 f6g4 e3c1 g4f6 c1e3');
INSERT INTO evaluations VALUES(38442,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38442,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38442,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(38442,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(38442,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38442,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(38442,36,'cp 24','d2d4 c7c5 d4d5 g8f6 b1c3 e8g8 f1d3 b7b5 c3b5 c5c4 d3c4 f6e4 e1g1 d8b6 b5d4 b8d7 f1e1 d7f6 c4f1 e7e5 d5e6 c8e6 c2c3 e6d5 a2a4 a8b8 d1c2 f8c8 a4a5 b6b7 a5a6 b7b6 f4f5 g6f5 c1e3 b6b2 c2b2 b8b2 d4f5 g7f8 g1h1 f6g4 e3a7');
INSERT INTO evaluations VALUES(40270,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(40270,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40270,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40270,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(40270,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40270,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(40270,36,'cp 31','e8g8 a1c1 c8e6 f3g5 e6g4 d1b3 h7h6 g5f3 d8b6 f3e5 g4e6 f1e2 d5c4 e2c4 b6b3 a2b3 f6d5 f4g3 d5c3 c1c3 e6f5 e1e2 a7a5 f2f3 b8d7 e3e4 d7e5 d4e5 f5c8 h1c1 g6g5 e2e3 b7b6 c1d1 c6c5 c3c1 a5a4 b3a4 a8a4 b2b3');
INSERT INTO evaluations VALUES(35867,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(35867,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(35867,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(35867,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(35867,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(35867,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(35867,36,'cp 56','f1e2 d5c4 e2c4 b8d7 e1g1 f6h5 f4g5 h7h6 g5h4 g6g5 h4g3 h5g3 h2g3 d7b6 c4d3 c8g4 d1c2 b6d5 a1c1 d5b4 d3h7 g8h8 c2b1 d8d7 a2a3 b4d5 h7d3 f7f5 c3d5 c6d5 f1e1 e7e6 c1c3 a8c8 e1c1 c8c3 b2c3 g4h5 f3e5 d7e7 b1b4 e7b4 c3b4 g7e5');
INSERT INTO evaluations VALUES(35866,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(35866,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(35866,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(35866,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(35866,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(35866,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(35866,36,'cp 32','d5c4 e2c4 b8d7 e1g1 f6h5 f4g5 h7h6 g5h4 g6g5 h4g3 h5g3 h2g3 e7e6 d1c2 d8e7 a2a4 c6c5 a1d1 d7f6 d4d5 e6d5 c4d5 c8e6 d5e6 e7e6 c3b5 f8c8 b5d6 c8c7 d6c4 e6e4 c2e2 e4c6 a4a5 a8e8 c4d6 e8e7 f1e1 f6e4 d6f5 e7e8 e2c2 c6e6');
INSERT INTO evaluations VALUES(50227,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(50227,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(50227,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(50227,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(50227,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(50227,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(50227,36,'cp 43','c2c4 f8g7 b1c3 d7d5 g1f3 e8g8 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 f1e2 c5d4 c3d4 b7b6 c1e3 c8b7 d1d3 b7a6 d3c2 d8d7 e1g1 f8c8 c2d2 a6e2 d2e2 b8c6 a1d1 e7e6 e2d2 c8d8 e3g5 f7f6 g5f4 a8c8 d4d5 e6d5 e4d5 c6a5 d5d6');
INSERT INTO evaluations VALUES(52449,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52449,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52449,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52449,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52449,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52449,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(52449,36,'cp -2','g7g6 c2c3 g8f6 e4e5 f6d5 d2d4 c5d4 d1b3 d5b6 a2a4 d7d6 a4a5 b6d7 e5d6 f8g7 g1f3 d4c3 b1c3 e8g8 d6e7 d8e7 c1e3 d7e5 f3e5 e7e5 f1c4 b8c6 b3a4 c8e6 c4e6 e5e6 e1g1 a8c8');
INSERT INTO evaluations VALUES(4713,30,'cp -268',NULL);
INSERT INTO evaluations VALUES(4713,31,'cp -262',NULL);
INSERT INTO evaluations VALUES(4713,32,'cp -271',NULL);
INSERT INTO evaluations VALUES(4713,33,'cp -273',NULL);
INSERT INTO evaluations VALUES(4713,34,'cp -280',NULL);
INSERT INTO evaluations VALUES(4713,35,'cp -276',NULL);
INSERT INTO evaluations VALUES(4713,36,'cp -262','d2d3 h7h6 c3a4 c5f8 c2c4 f8b4 a4c3 d7d6 h2h3 a7a5 e1g1 b4c5 c1e3 f7g8 c3d5 c8e6 a2a3 e6d5 c4d5 c6d4 b2b3 g8h8 g1h2 c5b6 f2f4 e5f4 f1f4 d8e7 f4f1 f6d7 a1b1 d7e5 b3b4 a5a4 h2h1 d4b3 e3b6');
INSERT INTO evaluations VALUES(5833,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(5833,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(5833,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(5833,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(5833,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(5833,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(5833,36,'cp 135','a2a3 b4c3 b2c3 d8d7 f1b5 a7a6 b5d3 b7b6 d1g4 f7f5 g4e2 c8b7 g1h3 d7f7 h3g5 f7d7 h2h4 c6a5 h4h5 e8c8 a3a4 a5c4 d3c4 d5c4 e2c4 h7h6 g5h3 b7g2 h1g1 g2h3 c4a6 c8b8 a4a5 d7c8 a5b6 c8a6 b6c7 b8c7 a1a6 c7b7 a6e6 h3g4 f2f3');
INSERT INTO evaluations VALUES(5838,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(5838,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(5838,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(5838,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(5838,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(5838,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(5838,36,'cp 133','b4c3 b2c3 d8d7 f1b5 a7a6 b5d3 b7b6 d1g4 f7f5 g4h5 g7g6 h5h4 a6a5 h4g3 c8a6 d3a6 a8a6 h2h4 d7f7 h4h5 g6h5 g1e2 a6a8 e2f4 e8d7 a3a4 g8e7 f4h5 e7g6 g3g5 d7c8 g5f6 f7d7 f6g7 c6e7 c1g5 c8b7 g5e7 a8g8 g7f6 d7e7 f6e7 g6e7 g2g3 e7g6 e1e2 b7a6 h1h2 g8c8 h5f6 h7h6');
INSERT INTO evaluations VALUES(31568,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(31568,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(31568,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(31568,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(31568,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(31568,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(31568,36,'cp 34','g1f3 g8f6 e4f6 d8f6 c1g5 f6g6 d3f5 g6f5 d1e2 f8b4 c2c3 b4d6 e1g1 e8g8 f1e1 b8d7 g5h4 d6c7 c3c4 h7h6 a1d1 f8e8 h4g3 c7f4 b2b3 a8d8 g3f4 f5f4 g2g3 f4c7 g1g2 c6c5 d1d3 a7a5 d4d5 d7f6 d5e6 e8e6 d3d8 c7d8');
INSERT INTO evaluations VALUES(31567,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(31567,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(31567,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(31567,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(31567,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(31567,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(31567,36,'cp 34','g8f6 e4f6 d8f6 c1g5 f6g6 d3f5 g6f5 d1e2 f8b4 c2c3 b4d6 e1g1 e8g8 f1e1 b8d7 g5h4 d6c7 c3c4 f8e8 h4g3 c7f4 b2b3 f4g3 h2g3 h7h6 a1d1 f5a5 e2b2 a5c7 b2c3 a7a5 c3e3 d7f6 d1d3 a5a4 b3b4 b7b5 e1c1 b5c4 c1c4');
INSERT INTO evaluations VALUES(29305,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(29305,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(29305,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(29305,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(29305,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(29305,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(29305,36,'cp 27','e4f6 d8f6 c1g5 f6g6 d3f5 g6f5 d1e2 f8b4 c2c3 b4d6 e1g1 b8d7 g5h4 e8g8 a1d1 h7h6 f1e1 d6c7 d1d3 f8e8 h4g3 c7g3 h2g3 a8d8 e1d1 f5b5 b2b3 b5a5 c3c4 a7a6 d3d2 a5f5 e2e3 d7f6 d1e1 d8d7 f3e5 f6g4');
INSERT INTO evaluations VALUES(51484,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51484,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(51484,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51484,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(51484,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(51484,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(51484,36,'cp 27','g8f6 g2g3 c7c5 b1c3 c8b7 f1g2 g7g6 d2d4 c5d4 d1d4 f8g7 e1g1 d7d6 d4h4 h7h6 c1e3 b8a6 a1c1 d8d7 g2h3 d7d8 h4d4 e8g8 d4d2 h6h5 f1d1 a6c5 e3d4 c5e4 c3e4 f6e4 d2e3 e4f6 h3g2 a8c8 b2b3 f6g4');
INSERT INTO evaluations VALUES(43867,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(43867,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43867,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(43867,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(43867,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(43867,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(43867,36,'cp 36','g2g3 c8b7 f1g2 e7e6 d2d4 f8b4 c1d2 b4d2 d1d2 d7d5 c4d5 e6d5 b1c3 e8g8 e1g1 b8d7 b2b3 f8e8 d2b2 a7a5 a1c1 b7a6 f1e1 c7c6 f3d2 h7h6 e2e4 d5e4 d2e4 f6d5 b2d2 d5b4 e4d6 e8e1 c1e1 d7f6 d6f5');
INSERT INTO evaluations VALUES(54835,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(54835,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(54835,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(54835,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(54835,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(54835,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(54835,36,'cp 65','d2d4 g8f6 c1e3 c7c6 a2a4 f8g7 d1d2 e8g8 h2h3 b8d7 g1f3 e7e5 d4e5 d7e5 f1e2 d8a5 e1g1 e5f3 e2f3 f8d8 f1d1 c8e6 e3d4 d6d5 e4d5 f6d5 f3d5 e6d5 d4g7 g8g7 d2d4 f7f6 d4f4 d5f7 c3e4 a5e5 f4e5 f6e5 a4a5 d8d5 e4d6 d5d1 a1d1 a8d8 d1d3');
INSERT INTO evaluations VALUES(55754,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(55754,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(55754,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55754,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55754,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55754,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(55754,36,'cp -8','b8d7 g1f3 g8f6 b1d2 f8d6 b2b3 c7c6 b5e2 e6e5 d4e5 d7e5 c1b2 e5f3 e2f3 d8e7 e1g1 e8g8 c2c4 f8d8 c4d5 f6d5 f3d5 d6b4 a2a3 b4d2 d1d2 d8d5 d2c3 f7f6 f1d1 d5g5 h2h4 g5g4 d1d4 g4d4 c3d4 h7h6 b3b4 a7a5 d4c4 g8h8 c4c3 a5b4 a3b4 a8a1 b2a1 e7e6');
INSERT INTO evaluations VALUES(5857,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(5857,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(5857,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(5857,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(5857,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(5857,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(5857,36,'cp 210','f1c4 g8e7 a2a3 b4c3 b2c3 b7b6 d1e2 e8g8 c1f4 c6a5 c4a2 c8b7 h2h4 c7c5 e1g1 d8c8 f4d6 f8e8 d4c5 b7a6 c3c4 b6c5 a1d1 c8b7 h4h5 h7h6 f1e1 a8c8 e2c2 e8d8 c2c3 a5c6 d6c5 b7c7 d1d8 c8d8 c3e3 a6b7 e1c1 c7c8 a3a4 c8c7 a2b1 b7a6 e3c3');
INSERT INTO evaluations VALUES(27012,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(27012,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(27012,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(27012,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27012,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(27012,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(27012,36,'cp 46','f1e2 f8e7 h2h3 g4f3 e2f3 c6d4 c1e3 d4e6 d1d2 c7c6 e1c1 g8f6 c1b1 b7b5 c3e2 e8g8 d3d4 d8c7 g2g4 f6d7 d4d5 c6d5 e4d5 e6c5 h3h4 a7a5 e2c3 b5b4 c3b5 c7b7 f3e2 d7b6 f2f4 e5f4 e3f4 c5e4 d2d3');
INSERT INTO evaluations VALUES(27011,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(27011,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(27011,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(27011,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(27011,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(27011,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(27011,36,'cp 30','f8e7 h2h3 g4f3 e2f3 c6d4 c1e3 d4e6 d1d2 c7c6 e1c1 h7h5 c1b1 e7g5 d1g1 g5e3 f2e3 d8b6 h3h4 g8f6 d2f2 e8c8 g2g4 h5g4 f3g4 f6g4 g1g4 d8f8 g4g3 c8b8 h4h5 a7a6 a2a3 h8h6 g3g1 b6d8 c3e2 g7g6 h5g6 f7g6 f2g3 h6h1 g1h1');
INSERT INTO evaluations VALUES(28241,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28241,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28241,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28241,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(28241,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28241,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28241,36,'cp 0','f6e4 e1g1 e4c3 b2c3 b4c3 a1b1 b8c6 d3e4 d7d5 e5d6 c7d6 b1b3 d6d5 e4h7 c3d4 f3d4 h8h7 f1e1 c6d4 d1d4 a8c8 h2h3 c8c4 d4e5 d8c7 e5g5 e8d7 b3f3 g7g6 g5f6 d5d4 f3d3 d7c8 c1f4 c7c5 e1e6 f7e6 f6e6 h7d7 e6e8 d7d8 e8e6');
INSERT INTO evaluations VALUES(28232,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28232,36,'cp 0','e1g1 e4c3 b2c3 b4c3 a1b1 b8c6 d3e4 d7d6 e5d6 c7d6 b1b3 d6d5 e4h7 c3d4 f3d4 h8h7 h2h3 c6d4 d1d4 a8c8 b3a3 a7a6 a3g3 c8c4 d4d3 g7g6 f1e1 e8d7 c1e3 d7c8 e3d4 b7c6 c2c3 c8b7 g3f3 d8c7 d3e3 c6b5 d4e5 c7d7 e5d4');
INSERT INTO evaluations VALUES(42545,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(42545,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(42545,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42545,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(42545,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(42545,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(42545,36,'cp -12','c7c6 d2d3 d7d6 g1e2 b7b5 c4b3 c8e6 b3e6 f7e6 f3h3 d8e7 f2f4 e8g8 c1d2 e5f4 e2f4 f8e8 g2g4 e6e5 g4g5 e5f4 g5f6 e7f6 h1f1 f6e6 h3e6 e8e6 a2a4 b5b4 c3e2 b8d7 f1f4 d6d5 e1c1 a8e8 e2g3 c5d6 f4f1 a7a5 d2f4 d6f4 f1f4 d5e4');
INSERT INTO evaluations VALUES(32087,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(32087,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(32087,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(32087,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(32087,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(32087,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(32087,36,'cp 157','d2d4 g8f6 d4e5 h5f3 d1f3 d6e5 c1d2 f8b4 e1c1 d8e7 h3h4 b8c6 c3d5 b4d2 d1d2 f6d5 e4d5 c6d4 f3g4 g7g6 h4h5 e7d7 g4e4 e8c8 h5g6 f7g6 e4e5 d4f5 e5e6 c8b8 e6d7 d8d7 d2d1 d7d8 c2c3 f5d6 f1d3');
INSERT INTO evaluations VALUES(41983,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41983,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41983,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41983,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(41983,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(41983,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(41983,36,'cp 18','e6d5 c1g5 h7h6 g5h4 e8g8 e2e3 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 c7c6 f3d2 f6e4 h4d8 e4c3 b2c3 f8d8 f2f3 c6c5 e1f2 a8c8 h1c1 c8c7 d2b3 d8c8 a2a4 g7g5 a4a5 g8g7 e2b5 c5d4 b3d4 f5h7 b5d7 c7d7 f2e2 c8c5 a1a3 d7c7 e2d2 a7a6 c1a1 c5c4 g2g3');
INSERT INTO evaluations VALUES(53304,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(53304,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(53304,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(53304,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(53304,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(53304,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(53304,36,'cp -16','f7f6 d2d4 c5d4 d1d4 e7e5 d4d2 c8e6 e2e3 b8c6 c2c4 d5c4 f1c4 e6c4 b3c4 a8c8 b2a3 f8e7 d2b2 d8b6 b2b6 a7b6 e1e2 c6a5 f3d2 a5c4 h1c1 b6b5 d2c4 c8c4 c1c4 b5c4 a3e7 g8e7 b1a3 e8d7 a1d1 d7e6 a3c4 h8c8 c4d6 c8c2 e2f3 b7b6');
INSERT INTO evaluations VALUES(53662,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(53662,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(53662,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(53662,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(53662,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(53662,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(53662,36,'cp 161','f1c4 b7b5 c4b3 b5b4 c3e2 g8f6 e4e5 f6d5 g1h3 c6c5 e1g1 b8c6 h3f4 d5b6 c1e3 c5c4 d1c2 c4b3 c2c6 c8d7 c6f3 b3a2 a1a2 a7a5 d4d5 b6d5 f4d5 e6d5 f3d5 d7e6 d5c6 e6d7 c6e4 a8a6 e2f4 f8e7 a2a1 e8g8 a1c1 d7b5 f1d1 d8e8 f4d5');
INSERT INTO evaluations VALUES(8222,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(8222,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(8222,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(8222,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(8222,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(8222,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(8222,36,'cp 100','f1d3 c8d7 e1g1 a7a6 c1d2 h7h6 a1c1 c6a5 f1e1 a5c4 d3c4 b4c3 c4d3 c3d2 d1d2 d8b6 c1c3 e7c6 c3b3 b6a5 d2e2 c6b4 d3b1 e8c8 a2a3 b4c6 e1c1 c8b8 b1d3 a5a4 c1c3 d8c8 h2h3 b8a7 e2d2 g7g5 d3c2 a4a5 f3e1 c8c7 e1d3 h8c8 d2e3 a7b8');
INSERT INTO evaluations VALUES(8223,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(8223,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(8223,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(8223,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(8223,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(8223,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(8223,36,'cp 115','c8d7 e1g1 a7a6 c1d2 h7h6 a1c1 e8g8 d3b1 f7f6 d1c2 f6f5 c3e2 d8b6 d2e3 e7g6 c2d1 c6a5 b2b3 d7b5 h2h4 g6e7 g2g3 f8c8 g1g2 c8c1 e3c1 a5c6 f3g1 b4c3 e2c3 b5f1 g2f1 b6d4 d1d4 c6d4 c1b2 a8c8 c3e2 d4e2 g1e2');
INSERT INTO evaluations VALUES(17908,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(17908,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(17908,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(17908,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17908,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(17908,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(17908,36,'cp 47','f1c4 g8f6 d1e2 f8d6 e1g1 e8g8 d2d4 h7h6 c4d3 f8e8 a2a3 d6f8 d4d5 c6e7 c3c4 e7g6 b1c3 d7d6 f1e1 g6f4 c1f4 e5f4 e2d2 g7g5 e4e5 d6e5 e1e5 g5g4 e5e8 f6e8 f3d4 d8g5 a1e1 e8f6 b2b4 c8d7 c4c5 a8e8 e1e8 f6e8 c5c6 b7c6 d4c6');
INSERT INTO evaluations VALUES(17893,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(17893,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(17893,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(17893,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17893,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(17893,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(17893,36,'cp 40','g8f6 d1e2 f8d6 e1g1 h7h6 d2d4 e8g8 b1d2 f8e8 e2d1 b7b5 c4d3 c8b7 f1e1 d6f8 f3e5 c6e5 d4e5 e8e5 d2f3 e5e4 d3e4 f6e4 c1e3 a6a5 a2a3 a5a4 d1d3 e4d6 e3c5 d8f6 e1e2 b7f3 d3f3 f6f3 g2f3 d6c4 c5f8 g8f8 g1g2 d7d6 a1d1 c7c6 f3f4 d6d5 h2h3 a8d8 f2f3');
INSERT INTO evaluations VALUES(11516,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(11516,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(11516,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11516,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(11516,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11516,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(11516,36,'cp 38','d1e2 h7h6 e1g1 d7d6 d2d4 c8g4 f1d1 g4h5 a2a4 g7g5 b1d2 e5d4 d2f1 d8e7 f1g3 h5g6 f3d4 c6d4 d1d4 f8g7 f2f4 g5f4 c1f4 f6d7 d4d1 g7e5 d1f1 e5f4 f1f4 e8c8 a1f1 d7e5 g3f5 g6f5 f4f5 d8e8 c4f7 e5f7 f5f7 e7e4 e2e4');
INSERT INTO evaluations VALUES(29607,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29607,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29607,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(29607,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(29607,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29607,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(29607,36,'cp 0','f6e4 g5h4 g7g5 c3e4 d5e4 h4g5 h8g8 g5e3 e7e5 d1a4 f5d7 a4c2 f8b4 e3d2 b4d2 c2d2 e5d4 d2d4 d8e7 d4c3 b8c6 g2g3 e8c8 g1h3 e7b4 h3f4 b4c3 b2c3 c6e5 f4d5 g8f8 d5f6 e4e3 f6d7 c8d7 a1b1 d7c8 f1h3 c8b8 f2e3 e5c4 e1f2 c4d6 f2f3');
INSERT INTO evaluations VALUES(47841,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(47841,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(47841,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(47841,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(47841,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(47841,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(47841,36,'cp 71','g1f3 e5d4 f3d4 d6d5 c1g5 d5c4 e2e3 h7h6 g5h4 b8c6 d4c6 d8d1 a1d1 b7c6 f1c4 a8b8 c3a4 f8d6 h4f6 g7f6 f2f3 e8e7 b2b3 h6h5 e1f2 h5h4 d1d4 c8f5 c4e2 a7a5 f3f4 d6a3 d4c4 h8d8 g2g4 h4g3 h2g3 d8d2 c4c6 a3b4 c6c7 e7d6 c7f7');
INSERT INTO evaluations VALUES(28689,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(28689,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(28689,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28689,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28689,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(28689,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28689,36,'cp 0','h7h6 g5h4 f8e7 e2e3 e8g8 f1e2 d7d5 h4f6 e7f6 c4d5 e6d5 e1g1 c7c6 a1c1 b8d7 b2b4 a7a5 b4b5 c6c5 d4c5 d7c5 f3d4 d8d6 d4f5 d6e5 f5d4');
INSERT INTO evaluations VALUES(28670,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(28670,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(28670,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28670,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28670,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28670,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28670,36,'cp 0','g5h4 f8e7 e2e3 e8g8 f1e2 f6e4 c3e4 b7e4 h4e7 d8e7 e1g1 c7c5 a1c1 d7d6 f3d2 e4b7 d2b1 b8c6 d4d5 e6d5 c4d5 c6e5 d1d2 f7f5 c1e1 e5d7 b1c3 a7a6 a2a4 a8e8 e2d3 d7e5 d3e2');
INSERT INTO evaluations VALUES(28671,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28671,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28671,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28671,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28671,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28671,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28671,36,'cp 0','f8e7 e2e3 e8g8 f1d3 c7c5 e1g1 b8c6 d4d5 e6d5 h4f6 e7f6 c4d5 c6e5 f3e5 f6e5 d1b3 d8e7 h2h3 d7d6 a1e1 a8c8 d3c4 c8d8 c4d3');
INSERT INTO evaluations VALUES(32648,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(32648,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(32648,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(32648,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(32648,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(32648,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(32648,36,'cp 34','e7e6 h2h3 g4h5 f3e5 h5e2 d1e2 f8b4 e1g1 e8g8 c3e4 f6e4 e2e4 a5d5 e4d5 e6d5 c1f4 b4d6 f1e1 f8d8 e5d3 d6f4 d3f4 g8f8 g2g4 c7c6 f2f3 b8d7 a2a4 a8c8 a1d1 d8e8 e1e8 f8e8 d1e1 e8f8 g1f2 c8e8 f4e2 a7a5 h3h4 f8e7 b2b3 e7d6');
INSERT INTO evaluations VALUES(32701,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(32701,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(32701,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(32701,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(32701,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(32701,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(32701,36,'cp 36','h2h3 g4h5 a2a3 b8d7 e1g1 c7c6 c1f4 f8e7 f4h2 e8g8 f1e1 h5g6 e2f1 b7b5 c3e2 f8d8 c2c3 a5b6 e2f4 e7d6 d1e2 a7a5 g1h1 d8e8 f3e5 d6e5 d4e5 f6d5 f2f3 d5f4 h2f4 b6c7 e1d1 d7b6 e2f2 b6d5 f4g3 e8d8 f2c5 a8c8');
INSERT INTO evaluations VALUES(46967,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(46967,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(46967,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(46967,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(46967,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(46967,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(46967,36,'cp -45','d6e5 f3e5 b8d7 e5c4 b7b5 c4e3 a8b8 f1e2 g7g6 c2c4 b5c4 b1c3 d7e5 e1g1 f8g7 e3c4 e5c6 d1a4 c8d7 d2d3 e8g8 c1e3 e7e6 e2f3 c6d4 a4d1 d7c8 a1c1 c8a6 b2b3 f6d5 c3d5');
INSERT INTO evaluations VALUES(47311,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(47311,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(47311,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(47311,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(47311,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(47311,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(47311,36,'cp -20','f3e5 b8d7 e5c4 b7b5 c4e3 a8b8 f1e2 e7e6 e1g1 f8e7 a2a4 b5b4 e3c4 e8g8 f1e1 f6d5 d2d3 c7c5 e2f1 e7f6 b1d2 c8b7 d1g4 d5e7 d2e4 f6d4 g4h3 d7b6 c4b6 b7e4 d3e4 d8b6 a4a5 b6c7 h3g3 e6e5 a5a6');
INSERT INTO evaluations VALUES(47303,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(47303,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(47303,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(47303,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(47303,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(47303,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(47303,36,'cp -20','b8d7 e5c4 e7e5 f1e2 f8c5 e1g1 e8g8 d2d3 a7a5 e2f3 c7c6 a2a4 d8c7 f1e1 f8e8 b1d2 b7b6 d2e4 f6e4 f3e4 d7f6 c1e3 c5e3 c4e3 c8e6 d1f3 f6e4 d3e4 a8d8 b2b3 h7h6 a1d1 b6b5 e3f5 e6f5 f3f5 d8d1 e1d1 e8d8 f5f3 b5a4 b3a4 d8d4');
INSERT INTO evaluations VALUES(16035,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(16035,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16035,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(16035,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(16035,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(16035,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(16035,36,'cp -24','e5c4 e7e5 f1e2 f8c5 e1g1 e8g8 d2d3 f8e8 c1g5 d7b6 b1d2 b6c4 d2c4 c8f5 c4e3 f5e6 g5h4 c7c6 e3g4 e6g4 e2g4 c5e7 g4f3 f6d7 h4e7 e8e7 f1e1 d7f6 d1d2 g7g6 d2b4 d8c7 g2g3 a8d8 b4c3 a7a5 e1e2 f6d5 f3d5 c6d5');
INSERT INTO evaluations VALUES(49673,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(49673,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(49673,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(49673,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(49673,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(49673,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(49673,36,'cp -42','d7d5 e4d5 f6d5 g1f3 b8c6 f1e2 g7g6 c2c4 d5e7 b2b4 f8g7 c1b2 e7f5 e1g1 e8g8 b1c3 f8e8 f1e1 c6d4 a3a4 d4e2 d1e2 b7b6 a4a5 c8b7 a5a6 b7f3 e2f3 f5d4 f3d1 c7c6 b4b5 c6b5 c3b5 d4b5 c4b5 d8d5 d1a4 a8d8 a1d1 d8d7 d1c1 h7h5 h2h3 d5d3 c1d1 d3f5 d1d7');
INSERT INTO evaluations VALUES(48849,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(48849,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(48849,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(48849,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(48849,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(48849,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48849,36,'cp -27','e4d5 f6d5 g1f3 b8c6 f1e2 g7g6 c2c4 d5e7 b2b4 f8g7 c1b2 e7f5 b1c3 e8g8 e1g1 f8e8 f1e1 b7b6 b4b5 c6d4 f3d4 f5d4 e2f3 d4f3 d1f3 a8b8 a3a4 c8b7 f3e3 a7a6 h2h3 f7f5 c4c5 a6b5 a4b5 f5f4 e3e2 f4f3 e2e3 b6c5');
INSERT INTO evaluations VALUES(12636,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(12636,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(12636,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(12636,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(12636,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(12636,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(12636,36,'cp 264','e4e5 f6g8 c1f4 a7a6 e1c1 b7b5 c3e4 c8b7 f4g5 d8c8 f1d3 a8b8 f2f4 b7e4 d3e4 g8e7 c1b1 c8c7 d1d3 e7c6 d4e3 h7h6 g5h4 f8e7 h4e7 c6e7 h1d1 b8d8 d3d6 a6a5 d6a6 e7c8 e4d3 b5b4 f4f5 e6f5 e3g3 e8g8 a6h6');
INSERT INTO evaluations VALUES(12631,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(12631,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(12631,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(12631,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(12631,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(12631,35,'cp 242',NULL);
INSERT INTO evaluations VALUES(12631,36,'cp 241','f6g8 c1f4 b7b6 c3e4 c8b7 e1c1 f8c5 e4c5 b6c5 d4c5 a8c8 c5d4 g8e7 c2c4 d8c7 h2h4 e7d5 f4g3 h7h5 f2f3 a7a5 c1b1 b7a6 d1c1 e8g8 g3f2 d5b4 f1e2 c8b8 h1d1 a5a4 d4d6 f8c8 d6c7 c8c7 d1d2 b4c6');
INSERT INTO evaluations VALUES(32211,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32211,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(32211,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(32211,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(32211,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(32211,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(32211,36,'cp 17','g8f6 c4e6 f7e6 e1g1 f8e7 c2c3 e8g8 d3d4 e5d4 f3d4 d8c8 c1g5 e6e5 d4f5 e7d8 b1d2 g8h8 g5f6 d8f6 d2c4 b8d7 a2a4 d7c5 f2f3 c8e6 f5e3 e6d7 a4a5 f6g5 b2b4 c5e6 g2g3 f8f7 d1e2 g5e3 c4e3 e6g5');
INSERT INTO evaluations VALUES(29998,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(29998,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29998,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(29998,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(29998,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(29998,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(29998,36,'cp 25','c4e6 f7e6 c2c3 f8e7 e1g1 e8g8 d1b3 d8c8 f1e1 b8c6 a2a4 c8d7 b1d2 a7a5 d2c4 b7b6 b3d1 d7e8 c1d2 e8h5 d3d4 e5d4 f3d4 h5d1 a1d1 c6d4 c3d4 f8e8 f2f3 d6d5 c4e5 c7c5 e5c6 c5d4 e4d5 f6d5 c6e7 e8e7 d2g5');
INSERT INTO evaluations VALUES(29993,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(29993,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29993,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(29993,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(29993,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(29993,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(29993,36,'cp 31','f7e6 c2c3 f8e7 e1g1 e8g8 d1b3 d8c8 f1e1 b8c6 c1e3 f6g4 b1d2 g4e3 f2e3 e7f6 d3d4 c6d8 e1f1 d8f7 a2a4 a7a5 f1f2 b7b6 a1f1 c8d7 d4d5 e6d5 e4d5 f6g5 c3c4 a8e8 h2h3 g5h6 d2e4 d7e7 b3d3 e7d7 f3d2 d7a4');
INSERT INTO evaluations VALUES(36545,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(36545,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(36545,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(36545,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(36545,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(36545,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(36545,36,'cp 38','f1g2 c7c5 d4d5 e7e6 e1g1 f8e8 f3g5 e6d5 c3d5 b8d7 c1f4 d7b6 d5f6 d8f6 a1b1 c8f5 e2e4 f5d7 b2b3 d7c6 d1d6 f6d6 f4d6 h7h6 g5h3 g7d4 f1e1 g6g5 g1f1 b6c8 e4e5 c8d6 e5d6 e8d8 g2c6 b7c6 e1e2 d8d6 h3g1 d4c3 g1f3 g8g7 h2h3 a7a6');
INSERT INTO evaluations VALUES(49052,30,'cp -186',NULL);
INSERT INTO evaluations VALUES(49052,31,'cp -180',NULL);
INSERT INTO evaluations VALUES(49052,32,'cp -194',NULL);
INSERT INTO evaluations VALUES(49052,33,'cp -191',NULL);
INSERT INTO evaluations VALUES(49052,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(49052,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(49052,36,'cp -193','d8g5 d1e2 g5g2 h1f1 c8e6 d2d3 e4f6 b1c3 f8d6 c1d2 b8c6 e5c6 b7c6 e1c1 e8g8 f1g1 g2h3 f2f3 f6h5 e2f2 h5f4 d1e1 a8e8 c3e2 f4e2 e1e2 e6d7 e2e8 f8e8 f3f4 d7g4 d2c3 h7h5 b3a4 e8e2 f2d4 h3e3 d4e3 e2e3 a4c6 d6f4 c3d2 d5d4 g1h1 g7g6 c6g2 g8g7 d2e3 d4e3 h2h3 g4d7 g2f3 e3e2 c1b1');
INSERT INTO evaluations VALUES(33925,30,'cp -185',NULL);
INSERT INTO evaluations VALUES(33925,31,'cp -181',NULL);
INSERT INTO evaluations VALUES(33925,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(33925,33,'cp -178',NULL);
INSERT INTO evaluations VALUES(33925,34,'cp -192',NULL);
INSERT INTO evaluations VALUES(33925,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(33925,36,'cp -194','d1e2 g5g2 h1f1 c8e6 d2d3 e4f6 b1c3 f8d6 c1f4 b8c6 e1c1 c6d4 e2e3 d4f5 e3e1 g2h3 d3d4 c7c6 c3e2 f6d7 c1b1 e8g8 e2g3 a8e8 e1a5 d6b8 e5d7 e6d7 f4b8 e8b8 c2c4 d5c4 b3c2 f5g3 h2g3 b7b6 a5c3 d7f5 c3c4 f5c2 c4c2 h3d7 f1h1 h7h6');
INSERT INTO evaluations VALUES(12421,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(12421,31,'cp 287',NULL);
INSERT INTO evaluations VALUES(12421,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(12421,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(12421,34,'cp 281',NULL);
INSERT INTO evaluations VALUES(12421,35,'cp 290',NULL);
INSERT INTO evaluations VALUES(12421,36,'cp 296','d4d5 e6d5 e4d5 c6d4 f3d4 c5d4 d1d4 f8d6 c1f4 d6f4 d4f4 e8g8 e1c1 d7d6 f1e2 f8e8 h1e1 h7h6 h2h3 e8e5 e2d3 f6d7 g2g4 d8e7 e1e4 a7a6 f4e3 d7c5 e4e5 e7e5 e3e5 d6e5 d3f1 f7f5 b2b4 c5d7 a2a4 f5g4 h3g4 d7f6 f2f3 c8d7 f1d3');
INSERT INTO evaluations VALUES(12447,30,'cp 301',NULL);
INSERT INTO evaluations VALUES(12447,31,'cp 273',NULL);
INSERT INTO evaluations VALUES(12447,32,'cp 272',NULL);
INSERT INTO evaluations VALUES(12447,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(12447,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(12447,35,'cp 278',NULL);
INSERT INTO evaluations VALUES(12447,36,'cp 297','e6d5 e4d5 c6d4 f3d4 c5d4 d1d4 f8d6 c1f4 d6f4 d4f4 e8g8 e1c1 d7d6 f1e2 f8e8 h1e1 h7h6 a2a3 e8e5 e2d3 f6d7 c1b1 a7a6 c3e4 d8e7 e1e3 e5d5 d1e1 d7e5 e4c3 d5d3 c2d3 c8e6 f4e4 e7d7 c3d5 e5c6 f2f4 g8h8 d5b6 d6d5');
INSERT INTO evaluations VALUES(12572,30,'cp 315',NULL);
INSERT INTO evaluations VALUES(12572,31,'cp 318',NULL);
INSERT INTO evaluations VALUES(12572,32,'cp 317',NULL);
INSERT INTO evaluations VALUES(12572,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(12572,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(12572,35,'cp 301',NULL);
INSERT INTO evaluations VALUES(12572,36,'cp 314','e4d5 c6d4 f3d4 c5d4 d1d4 f8d6 c1f4 d6f4 d4f4 e8g8 f1e2 d7d6 e1c1 f8e8 h1e1 h7h6 a2a3 a7a6 g2g4 e8e2 c3e2 c8g4 d1d2 d8e7 e1g1 g4e2 g1e1 a8e8 e1e2 e7e2 d2e2 e8e2 b2b3 e2e1 c1b2 f6e8 a3a4 g8f8 f4b4 e1e7 a4a5 e7d7 b4e4 d7e7 e4d4 e7d7 f2f4');
INSERT INTO evaluations VALUES(12571,30,'cp 292',NULL);
INSERT INTO evaluations VALUES(12571,31,'cp 305',NULL);
INSERT INTO evaluations VALUES(12571,32,'cp 304',NULL);
INSERT INTO evaluations VALUES(12571,33,'cp 310',NULL);
INSERT INTO evaluations VALUES(12571,34,'cp 313',NULL);
INSERT INTO evaluations VALUES(12571,35,'cp 307',NULL);
INSERT INTO evaluations VALUES(12571,36,'cp 298','c6d4 f3d4 c5d4 d1d4 f8d6 c1f4 d6f4 d4f4 e8g8 e1c1 d7d6 f1e2 f8e8 h1e1 h7h6 a2a3 e8e5 e2d3 f6d7 c1b1 d8f8 e1e5 d7e5 f4b4 c8g4 d1d2 b7b6 d3b5 g4f5 h2h3 g7g5 b5e2 e5d7 h3h4 f5g6 h4g5 h6g5 e2b5 d7f6');
INSERT INTO evaluations VALUES(29942,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(29942,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(29942,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(29942,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(29942,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(29942,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(29942,36,'cp 63','d2d4 b8d7 h2h3 g4f3 e2f3 f8e7 a2a4 c7c6 e1g1 e8g8 g2g3 d8a5 f3g2 f8e8 c1e3 h7h6 d1b1 e5d4 e3d4 f6h7 b2b4 a5c7 a4a5 e7f6 b1d1 h7f8 d1d2 f6d4 d2d4 f8e6 d4d2 a8d8 d2e3 b7b5 a5b6 a7b6 b4b5 d7e5 b5c6 c7c6 a1b1 d8c8 c3d5 c6c2 b1b6');
INSERT INTO evaluations VALUES(29935,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(29935,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(29935,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(29935,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(29935,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(29935,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(29935,36,'cp 59','b8d7 h2h3 g4f3 e2f3 c7c6 e1g1 f8e7 d4d5 e8g8 c1e3 d8a5 d1d2 f8c8 d5c6 b7c6 f1d1 a5c7 d2e2 h7h6 b2b3 d7f8 g2g3 f8e6 d1d3 c8b8 f3g2 b8b4 a1d1 a7a5 h3h4 a8e8 c3a4 e7f8 g2f3 e8b8 c2c3 b4b5 g1g2 c7b7');
INSERT INTO evaluations VALUES(30898,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(30898,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(30898,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(30898,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(30898,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30898,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(30898,36,'cp 34','c2c3 b8c6 f1e2 c5d4 c3d4 g8e7 e1g1 e7f5 b1c3 a7a6 c1g5 d8b6 c3a4 b6a7 g5e3 f8e7 a1c1 c6a5 a4c3 a7b6 b2b3 f5e3 f2e3 a8c8 e2d3 b6d8 d1d2 a5c6 e3e4 c6b4 d3b1 d7b5 c3b5 c8c1 f1c1 a6b5');
INSERT INTO evaluations VALUES(39261,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(39261,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(39261,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(39261,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(39261,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(39261,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(39261,36,'cp -93','b1c3 g8f6 c3a4 c5e7 d2d4 e5d4 e4e5 f6g8 d1d4 b8c6 d4e3 g8h6 e3f2 c6e5 c1h6 g7h6 e1c1 d7d5 f2d4 e5g6 d4d5 d8d5 d1d5 c8d7 a4c3 e7g5 c1b1 e8c8 g1h3 d7h3 d5d8 c8d8 g2h3 c7c6 f1d3 h8e8 c3e4 d8c7 c2c3 g6e5 b1c2 e5f3 h1f1');
INSERT INTO evaluations VALUES(39234,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(39234,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(39234,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(39234,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(39234,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(39234,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(39234,36,'cp -117','g8f6 c3a4 c5e7 d2d4 e5d4 e4e5 f6d5 d1d4 d5b6 d4g4 g7g6 a4b6 a7b6 c1h6 a8a5 g4f4 b8c6 e1c1 a5a2 f1c4 a2a1 c1d2 a1d1 d2d1 d7d5 e5d6 d8d6 f4d6 e7d6 g1e2 c6e5 c4b3 c8e6 b3e6 f7e6 e2c3 e5f7 h6g7 h8g8 g7d4 e8d7 c3e4 d6e7 d4c3 g8a8 d1e2 f7d6');
INSERT INTO evaluations VALUES(44686,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(44686,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(44686,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(44686,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(44686,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(44686,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(44686,36,'cp 37','g1f3 b8c6 b1a3 c8d7 f1d3 c5d4 c3d4 c6b4 d3b1 e7f5 e1g1 f8e7 c1d2 a7a5 b1f5 e6f5 d2b4 e7b4 d1b3 d7e6 f1c1 e8g8 a3c2 b4e7 c2e1 d8d7 e1d3 a5a4 b3d1 f8c8 h2h4 h7h6 b2b3 c8c1 a1c1 a4b3 a2b3 a8a3 d3f4 d7b5 c1c3 e7b4 c3c7 b4a5 f4e6 f7e6');
INSERT INTO evaluations VALUES(55006,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(55006,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(55006,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(55006,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(55006,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(55006,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(55006,36,'cp -66','c7c6 b5f1 e7e5 d2d4 e5e4 c2c4 f8d6 b1c3 g8f6 g1e2 b8a6 c4d5 a6b4 e2f4 b4d5 f4d5 c6d5 d1b3 e8g8 c1d2 d6b8 f1e2 f8e8 a1c1 e8e6 c3a4 b7b6 a2a3 h7h6 e1g1 e6e8 d2b4 c8d7 a4c3 d7c6 c3b5 d8d7');
INSERT INTO evaluations VALUES(52694,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(52694,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(52694,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(52694,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(52694,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(52694,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(52694,36,'cp -71','b5f1 e7e5 d2d4 e5e4 c2c4 g8f6 b1c3 f8d6 g1e2 b8a6 c4d5 a6b4 e2f4 d6f4 e3f4 b4d5 f1c4 c8e6 c4d5 e6d5 e1g1 d8d7 h2h3 e8g8 c1e3 d5c4 f1e1 c4d3 b2b3 b7b6 a1c1 f8e8 f2f3 f6d5 d1d2 d5c3 d2c3 a8c8 c3d2 h7h6 c1c3 c8d8 e1c1 d7f5 a2a4');
INSERT INTO evaluations VALUES(55509,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(55509,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(55509,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(55509,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(55509,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(55509,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(55509,36,'cp 176','e5f3 d6d5 d2d4 g8f6 c1g5 f8e7 e2e3 h7h6 g5f4 c7c5 d4c5 e8g8 f1e2 b8c6 b1d2 e7c5 c2c3 f8e8 d2b3 c5b6 f3d4 a7a5 a2a4 b6d4 b3d4 c6d4 d1d4 e8e4 d4d1 d8b6 d1c2 c8d7 e1g1 e4e6 h2h3 a8c8 e2d3 e6c6 f4e5');
INSERT INTO evaluations VALUES(55696,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(55696,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(55696,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(55696,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(55696,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(55696,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(55696,36,'cp 171','d6d5 d2d4 g8f6 c1g5 f8e7 e2e3 e8g8 b1d2 h7h6 g5f6 e7f6 c2c3 c8e6 f1e2 b8d7 e1g1 b7b6 f1e1 c7c5 e2d3 f8e8 d2f1 d8c7 f1g3 f6e7 d3f5 e6f5 g3f5 d7f6 f5e7 e8e7 d1e2 g7g6 c3c4 d5c4 e2c4 a8c8 c4b3');
INSERT INTO evaluations VALUES(56985,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(56985,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(56985,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(56985,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(56985,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(56985,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(56985,36,'cp -83','g7g5 d2d4 f8g7 h2h4 d7d6 d1d3 b8c6 b1c3 g5g4 f3g1 f4f3 g2f3 g7d4 c1f4 d4e5 f4e5 c6e5 d3f1 g8f6 e1c1 d8e7 f3f4 e5c4 f1c4 c8d7 g1e2 e8c8 c4d4 c8b8 e2g3 h8e8 h1e1 f6g8 c3d5 e7h4 e1e3 d8c8 e3b3 h4d8 g3h5 d7c6 h5g7 e8f8 a2a4 a7a6 a4a5 b8a8 c2c4 g8e7 d5f6');
INSERT INTO evaluations VALUES(56943,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(56943,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(56943,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(56943,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(56943,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(56943,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(56943,36,'cp -58','h2h4 f8g7 b1c3 b8c6 d2d4 d7d6 d1d3 g5g4 f3g1 f4f3 g2f3 g7d4 c1f4 d4e5 f4e5 c6e5 d3f1 g8f6 e1c1 d8e7 f3f4 e5c4 f1c4 c8d7 g1e2 e8c8 c4d4 c8b8 e2g3 h8e8 a2a4 a7a6 h1e1 h6h5 g3f5 d7f5 e4f5 c7c5 e1e7 c5d4 e7f7 d4c3 f7f6 e8e4 b2c3 e4f4 f6f7 f4a4 f5f6 a4a1 c1d2 a1d1 d2d1 g4g3 f7g7 d8f8 f6f7');
INSERT INTO evaluations VALUES(8334,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(8334,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(8334,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8334,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(8334,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(8334,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(8334,36,'cp 12','d1d2 e8g8 e1g1 c8d7 a1d1 a7a6 f2f4 c6d4 e3d4 d7c6 d2e3 b7b5 a2a3 d8d7 h2h3 d7b7 e2f3 f8e8 g1h1 e7e5 f4e5 d6e5 d4b6 e8e7 b6a5 a8c8 a5b4 e7e8 d1d6 c8d8 f1d1 d8d6 d1d6 b7c7 e3d2');
INSERT INTO evaluations VALUES(8337,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(8337,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(8337,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8337,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(8337,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(8337,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(8337,36,'cp 14','e8g8 e1g1 c8d7 f2f3 c6d4 e3d4 f8e8 f1d1 a7a6 a2a4 d8a5 c3d5 a5d2 d5f6 g7f6 d1d2 f6d4 d2d4 e8c8 c2c3 b7b5 a4b5 d7b5 g1f2 a8b8 d4d2 g8g7 a1a5 h7h5 h2h4 b5e2 f2e2 c8c6 e2d3 b8b5 a5b5 a6b5 d3e3 c6c4 d2d5 b5b4 e3d3 b4c3 b2c3 c4c6 c3c4 c6c5');
INSERT INTO evaluations VALUES(3251,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3251,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(3251,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(3251,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(3251,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(3251,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3251,36,'cp 17','e1g1 c8d7 f2f3 a7a6 a2a4 c6d4 e3d4 d8a5 f1d1 f8e8 c3d5 a5d2 d5f6 g7f6 d1d2 f6d4 d2d4 e8c8 c2c3 b7b5 a4b5 d7b5 g1f2 a8b8 e2d1 g8g7 d4d2 h7h5 h2h4 c8c5 a1a3 b8a8 a3a5 b5c4 a5c5 d6c5 f2e3 a8b8 f3f4');
INSERT INTO evaluations VALUES(16743,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(16743,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(16743,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(16743,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(16743,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(16743,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(16743,36,'cp 130','e5d4 e1g1 h6g4 f3d4 g4e5 c4b3 f8c5 c1e3 d7d6 d4c6 e5c6 e3c5 d6c5 d1h5 d8e7 b1c3 e8g8 f2f4 c8e6 f4f5 c5c4 f5e6 c4b3 e6f7 f8f7 h5f7 e7f7 f1f7 g8f7 a2b3 c6b4 c3e2 c7c6 c2c3 b4d3 a1d1 d3c5 b3b4 c5e4 d1d7 f7g8 d7b7 a7a5 h2h3 a5b4 b7b4 a8a1 g1h2 e4d6 b4d4');
INSERT INTO evaluations VALUES(16747,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(16747,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(16747,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(16747,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(16747,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(16747,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(16747,36,'cp 110','e1g1 h6g4 f3d4 g4e5 c4b3 f8c5 c1e3 d7d6 d4c6 e5c6 e3c5 d6c5 d1h5 d8e7 b1c3 g7g6 h5h6 c8d7 c3d5 e7e5 h6e3 e8c8 c2c3 h8e8 f1e1 g6g5 e3c5 f7f5 f2f3 b7b6 c5f2 f5f4 a1d1 e5g7 d5b4 c6e5 b4d3 g5g4 d3e5 e8e5 d1d4 d8e8 b3d5 c7c6 d5b3 a7a5 e1d1 g4f3 f2f3');
INSERT INTO evaluations VALUES(19405,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(19405,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(19405,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(19405,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(19405,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(19405,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19405,36,'cp 35','d2d4 c5d4 e3d4 d7d5 c4d5 d8d5 g1f3 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 f1e2 e8g8 e1g1 f8e8 f1e1 c8d7 e2c4 d5c4 d1d7 e7d5 d7b7 a8c8 a1c1 c8c7 b7b3 h7h6 g2g3 c4b3 a2b3 d5c3 c1c3 c7c3 b2c3 e8e5 g1f1 e5c5 c3c4 f7f5');
INSERT INTO evaluations VALUES(19412,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(19412,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19412,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(19412,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19412,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(19412,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(19412,36,'cp 40','c5d4 e3d4 d7d5 c4d5 d8d5 g1f3 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 d1a4 c8d7 a4a3 d5e6 a3b4 a8c8 b4e4 d7c6 e4h4 e7g6 h4g3 e8g8 f1d3 f8d8 d3g6 h7g6 e1g1 e6f5 f2f3 a7a6 f1e1 d8d3 a1c1 f5d7');
INSERT INTO evaluations VALUES(1643,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(1643,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(1643,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(1643,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(1643,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(1643,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(1643,36,'cp 80','c8g4 f1b5 e8c8 b5c6 b7c6 d1d3 c6c5 c1e3 g4f3 g2f3 e7e6 e1c1 g8e7 d3c4 c5d4 d1d4 d8d4 c4d4 e7c6 d4f4 f8a3 e3d2 h8d8 b2a3 a5a3 c1b1 c6d4 c3d1 a3c5 c2c3 d4f3 f4f3 d8d2 f3a8 c8d7 a8e4 g7g6 h1e1 c5b5 b1a1 b5d3 e4d3 d2d3 a1b2 d3h3 e1h1 h3h4 b2a1 d7d6 d1e3');
INSERT INTO evaluations VALUES(27635,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(27635,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(27635,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(27635,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(27635,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(27635,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(27635,36,'cp 64','f1b5 e8c8 b5c6 b7c6 d1d3 c6c5 d4d5 e7e6 e1g1 c5c4 d3e3 f8c5 e3f4 g4f3 f4f3 e6d5 b2b3 c4b3 f3g4 f7f5 g4f5 c8b7 a1b1 g8f6 b1b3 c5b6 c1e3 h8e8 g2g3 b7a8 e3g5 d5d4 f5f3 a8b8 c3b5 a7a6 b5c7 b8c7 f1b1 a5g5 b3b6 g5d5');
INSERT INTO evaluations VALUES(7064,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(7064,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(7064,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(7064,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7064,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(7064,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(7064,36,'cp 72','a3c4 c8e6 g5f6 e7f6 c4d6 e8f8 d6c4 d8d4 c4e3 f6g5 d1d2 h7h5 h2h4 g5h6 e1c1 g7g6 g2g3 f8g7 f1h3 e6h3 h1h3 h8d8 d2e2 d4b4 a2a3 d8d1 e2d1 b4d4 d1f3 a8d8 c1b1 d4d7 h3h1 c6d4 f3g2 h6e3 f2e3 d4b5 c3d5 b5d6 h1d1 d7g4 d5c3');
INSERT INTO evaluations VALUES(7062,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(7062,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(7062,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(7062,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(7062,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(7062,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(7062,36,'cp 63','c8e6 g5f6 e7f6 c4d6 e8f8 d6c4 d8d4 c4e3 f6g5 e3d5 h7h5 d5c7 a8d8 c7e6 f7e6 f1d3 d4b4 a1b1 g7g6 e1g1 f8g7 d1e1 h5h4 h2h3 h8f8 a2a3 b4e7 c3e2 b7b5 g1h1 b5b4 a3b4 e7b4 f2f3 b4e1 f1e1 c6b4');
INSERT INTO evaluations VALUES(7146,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7146,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7146,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7146,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7146,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7146,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7146,36,'cp 0','e8g8 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 f1f2 c6b8 c3b1 d8e7 b1d2 b8d7 d1e2 f8f7 a1f1 a8f8 f3h4 g8h7 h4f3');
INSERT INTO evaluations VALUES(30244,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(30244,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(30244,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30244,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(30244,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30244,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(30244,36,'cp 24','e1g1 f8e7 c1g5 e8g8 f1e1 b8d7 h2h3 g4h5 c2c4 d5c4 d3c4 c7c6 b1c3 h7h6 g5h4 f8e8 g2g4 h5g6 h4g3 e7b4 e1e8 d8e8 d1b3 b4c3 b2c3 b7b5 c4f1 f6e4 a1e1 e8f8 g3f4 a7a5 a2a4 b5b4 c3b4 f8b4 b3c4');
INSERT INTO evaluations VALUES(30243,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(30243,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(30243,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(30243,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(30243,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(30243,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30243,36,'cp 14','f8e7 h2h3 g4h5 c1f4 e8g8 b1d2 e7d6 f4g5 c7c5 c2c4 b8c6 d4c5 d6c5 d1b3 d5c4 b3c4 c5e7 f1e1 f8e8 a1d1 d8d5 g2g4 h5g6 d3g6 h7g6 c4d5 f6d5 d2c4 e7g5 e1e8 a8e8 d1d5 g5f6 b2b3 e8d8 d5d8 f6d8 g1g2');
INSERT INTO evaluations VALUES(39043,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(39043,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(39043,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(39043,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(39043,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(39043,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(39043,36,'cp 112','d2d4 a7a6 f1d3 d7d6 c1g5 g8e7 d1d2 h7h6 g5e3 b8d7 e1g1 c7c5 a2a4 b7b6 a1d1 g6g5 d4c5 b6c5 d3c4 a8b8 c4b3 g5g4 f3e1 d8c7 d2d6 c7d6 d1d6 c5c4 b3c4 b8b2 c3e2 c8b7 e2g3 b2b1 e1d3 b1f1 g3f1 e7c8');
INSERT INTO evaluations VALUES(56612,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(56612,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56612,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(56612,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(56612,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(56612,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56612,36,'cp -11','d8d1 e1d1 g8f6 f1b5 c8d7 b1c3 a7a6 b5d7 b8d7 a2a3 d7c5 f3e5 c5e4 c3e4 f6e4 d1e2 f7f6 e5d3 e8c8 h1e1 e4c5 d3c5 f8c5 e2f1 h8e8 e1e8 d8e8 h2h3 c8d7 a3a4 d7d6 c1d2 b7b5 a4b5 a6b5 a1d1 d6c6 b2b3 h7h5 d1e1 e8a8 d2e3 c5e3 e1e3 a8a1 f1e2');
INSERT INTO evaluations VALUES(34960,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(34960,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(34960,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(34960,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(34960,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(34960,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(34960,36,'cp -7','e1d1 g8f6 f1b5 c8d7 b5d3 b8c6 d1e2 h7h6 b1d2 g7g5 c2c3 d7e6 h1d1 a7a5 d2f1 e8c8 f1g3 f8c5 c1e3 c5e3 e2e3 g5g4 f3h4 c6e7 b2b3 c7c6 d3e2 c8c7 d1d8 h8d8 a1d1 d8d1 e2d1 b7b5 f2f3 e7c8 f3g4 e6g4');
INSERT INTO evaluations VALUES(1457,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(1457,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(1457,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(1457,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(1457,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(1457,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(1457,36,'cp -59','g8h6 b1a3 c5d4 c3d4 f8a3 b2a3 h6f5 e1f2 c6d4 f3d4 f5d4 c1e3 b6b2 d1d2 b2d2 e3d2 d4c6 a1b1 f7f6 e5f6 g7f6 f1d3 a8b8 f4f5 c8d7 f5e6 d7e6 d3e2 e8g8 h1c1 d5d4 e2f3 c6e5 c1c7 b7b6 c7a7 f8c8 b1d1 e5f3 g2f3 c8c2 f2g3 c2a2 d2b4 b8d8 a7c7');
INSERT INTO evaluations VALUES(1605,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(1605,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(1605,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1605,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(1605,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(1605,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(1605,36,'cp 18','e7e5 b1c3 f8b4 d4e5 d5d1 e1d1 g8e7 e2e4 c8g4 c1e3 c6e5 f1e2 e5c6 d1c2 f7f5 h2h3 g4f3 e2f3 f5f4 e3f4 c6d4 c2d3 a8d8 d3c4 b4c3 c4c3 d4f3 g2f3 h8f8 f4e3 f8f3 a1g1 e8f7 a2a3 h7h6 g1g3 f3g3 f2g3 e7c6 e3f4 d8e8 f4c7');
INSERT INTO evaluations VALUES(32064,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(32064,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(32064,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(32064,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32064,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(32064,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(32064,36,'cp 81','b1c3 c7c6 a2a3 g8f6 d2d3 b8d7 f3h4 d6d5 h4g6 h7g6 e4d5 f6d5 d1f3 d7b6 c3d5 b6d5 c1d2 d8d7 d2a5 b7b6 a5d2 e8c8 e1c1 c8b7 h3h4 f7f6 h4h5 g6g5 d2e3 d7e6 c1b1 b6b5 c4a2 a7a5 c2c4 d5e3 f2e3 b5b4 d3d4 b4a3 b2a3');
INSERT INTO evaluations VALUES(32062,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(32062,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(32062,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(32062,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(32062,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(32062,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(32062,36,'cp 84','c7c6 a2a3 b8d7 d2d3 g8f6 f3h4 d6d5 h4g6 h7g6 e4d5 c6d5 c4d5 f6d5 c3d5 d8a5 d5c3 f8b4 c1d2 b4c3 d2c3 a5d5 f2f3 d7c5 d1e2 f7f6 e1c1 d5a2 f3f4 c5a4 c1d2 a4c3 b2c3 a2a3 f4e5 e8g8 d1a1 a3c5 e2e3 c5c6 e5e6 f8c8 c3c4 b7b5 c4c5 c6c5 e3c5 c8c5 h1e1 a8c8 a1a7 c5c2 d2e3');
INSERT INTO evaluations VALUES(31583,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(31583,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(31583,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(31583,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(31583,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(31583,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(31583,36,'cp 63','a2a3 b8d7 d1e2 h7h5 h1g1 h5g4 h3g4 g8f6 d2d3 f8e7 c1e3 d6d5 e4d5 c6d5 c4d5 f6d5 c3d5 d8a5 d5c3 a8c8 e1f1 a5a6 a1e1 f7f6 f3d4 e8f8 d4f5 a6c6 f5e7 f8e7 f2f3 e7f7 f1g2 d7f8 g1h1 f8e6 h1h8 c8h8 e1h1 h8h1 g2h1 f7g8 e3a7 b7b6');
INSERT INTO evaluations VALUES(42143,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(42143,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(42143,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(42143,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(42143,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(42143,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(42143,36,'cp 184','f1d3 c7c5 a2a3 b4a5 e1g1 c5d4 c3b5 b8c6 b2b4 a5c7 b5c7 d8c7 d1c2 a7a6 f3d4 e8g8 d4c6 c7c6 c2c6 b7c6 f1d1 a6a5 b4a5 a8a5 a3a4 a5a8 d3c2 c8a6 f2f3 f8b8 d1d6 b8c8 c1b2 g8f8 a4a5 a6b5 g1f2 f8e7 d6d2 c6c5 b2c3 c8b8 f2e3 b8d8 d2d8 e7d8 c2d3 b5d3 e3d3');
INSERT INTO evaluations VALUES(42145,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(42145,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(42145,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(42145,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(42145,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(42145,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(42145,36,'cp 178','c7c5 a2a3 b4a5 e1g1 c5d4 c3b5 b8c6 b2b4 a5c7 b5c7 d8c7 d1c2 a7a6 f3d4 e8g8 d4c6 c7c6 c2c6 b7c6 f1d1 a6a5 b4a5 a8a5 a3a4 a5a8 f2f3 c8a6 c1e3 h7h6 a4a5 f8c8 d3a6 a8a6 e3c5 c8b8 c5b6 b8b7 d1d8 g8h7 a1d1 e6e5 d8f8 c6c5 d1d6 f6h5 d6c6 h5f4 c6c5');
INSERT INTO evaluations VALUES(44560,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(44560,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(44560,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(44560,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(44560,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(44560,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(44560,36,'cp 143','e4e5 f6d5 d4c5 e7e6 c2c4 d5e7 g1f3 e7g6 c1e3 f8e7 b1c3 b8c6 f1e2 b7b6 c5b6 a7b6 e1g1 c8a6 c3b5 e8g8 b2b3 a8b8 a2a4 c6e5 f3e5 g6e5 d1d2 a6b5 a4b5 d7d5 a1a7 b8c8 c4d5 e7c5 f1d1 d8d5 d2d5 e6d5 e3c5 c8c5 a7b7 f8c8 b7b6 c5c1 h2h4 d5d4 d1c1 c8c1 g1h2');
INSERT INTO evaluations VALUES(44510,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(44510,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(44510,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(44510,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(44510,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(44510,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(44510,36,'cp 166','f6d5 d4c5 e7e6 c2c4 d5e7 b1c3 e7g6 c1e3 b7b6 c5b6 a7b6 g1f3 b8c6 f1e2 c8a6 e1g1 c6e5 f3e5 g6e5 c3b5 a6b5 c4b5 d8b8 a2a4 d7d5 d1d2 f8e7 b2b4 e8g8 a4a5 e5c4 e2c4 d5c4 e3b6 b8e8 b6d4 e8b5 d4c3 e7f6 f1d1 f6c3 d2c3');
INSERT INTO evaluations VALUES(44661,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(44661,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(44661,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(44661,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(44661,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(44661,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(44661,36,'cp 189','d4c5 e7e6 c2c4 d5e7 b1c3 e7g6 c1e3 b7b6 c5b6 a7b6 g1f3 b8c6 f1e2 f8e7 e1g1 a8b8 d1d2 c6e5 f3e5 g6e5 c3b5 e8g8 f1d1 c8b7 f2f4 e5g6 d2d7 e7c5 e3c5 b6c5 d7d8 f8d8 g2g3 e6e5 f4e5 g6e5 a2a4 b7f3 d1d8 b8d8 e2f3 e5f3 g1g2 f3e5 a4a5 d8d2 g2g1 d2b2');
INSERT INTO evaluations VALUES(47139,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(47139,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(47139,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(47139,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(47139,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(47139,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(47139,36,'cp 48','c4d5 d8d5 g1f3 b8c6 b1c3 d5a5 d2d3 e7e5 f1g2 c6d4 e1g1 d4f3 g2f3 c7c6 c1d2 a5d8 a1b1 a7a5 a2a3 f8d6 b2b4 e8g8 d1c2 h7h6 b4a5 d6c7 c3a4 c7a5 d2a5 a8a5 a4b6 c8e6 c2c3 a5a6 b6c4 e6c4 c3c4 d8c7 f1c1 a6a3 c4b4 f8a8 b4b7 c7b7 b1b7');
INSERT INTO evaluations VALUES(47110,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47110,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47110,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(47110,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(47110,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(47110,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(47110,36,'cp 38','d8d5 g1f3 e7e5 b1c3 d5a5 f1g2 b8c6 d2d3 c6d4 e1g1 d4f3 g2f3 c7c6 c1d2 a5d8 d1c2 f8e7 a1b1 a7a5 a2a3 e8g8 b2b4 a5b4 a3b4 c8d7 d2e3 d7h3 f1c1 h7h6 e3c5 e7c5 b4c5 d8d7 c2d2 f8b8 d2b2 d7e7 c3e4 f6e4 f3e4 g7g6');
INSERT INTO evaluations VALUES(33696,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(33696,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(33696,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(33696,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(33696,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(33696,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(33696,36,'cp 46','g1f3 e7e5 b1c3 d5a5 f1g2 b8c6 e1g1 c6d4 d2d3 d4f3 g2f3 c7c6 c1d2 a5d8 a1b1 c8h3 f1e1 f8b4 d1a4 a7a5 a2a3 b4c5 c3e4 f6e4 a4e4 c5d4 e2e3 f7f5 e4h4 d8h4 g3h4 d4b6 b2b4 e8g8 b4b5 c6b5 b1b5');
INSERT INTO evaluations VALUES(33694,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(33694,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(33694,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(33694,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(33694,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(33694,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(33694,36,'cp 40','e7e5 b1c3 d5a5 d2d3 b8c6 f1g2 c6d4 e1g1 d4f3 g2f3 c7c6 c1d2 a5d8 d1c2 f8e7 a1b1 a7a5 a2a3 e8g8 b2b4 a5b4 a3b4 c8d7 d2g5 h7h6 g5f6 e7f6 c3e4 d8e7 f1c1 f6g5 c2c5 f8e8 e4g5 h6g5 b4b5 e7c5 c1c5 c6b5 f3b7');
INSERT INTO evaluations VALUES(33911,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(33911,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(33911,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33911,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(33911,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(33911,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(33911,36,'cp 50','b1c3 d5a5 f1g2 b8c6 d2d3 c6d4 e1g1 d4f3 g2f3 c7c6 c1d2 a5d8 a1b1 a7a5 a2a3 f8d6 b2b4 e8g8 b4a5 d6c7 c3e4 f6e4 f3e4 c7a5 d2e3 d8e7 d1c2 a5c7 a3a4 h7h6 e4g2 c7d6 e3d2 c8e6 a4a5 f8d8 c2c3 e6d5 g2d5');
INSERT INTO evaluations VALUES(56860,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(56860,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(56860,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(56860,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(56860,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(56860,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(56860,36,'cp 42','b1c3 g8f6 d1e2 c7c5 g1f3 b8c6 e4e5 d7d5 e5f6 d5c4 e2c4 d8f6 d2d3 f6e7 c1e3 b7b6 d3d4 f8g7 d4c5 c8b7 h2h4 h7h6 e1c1 e8g8 d1d2 c6a5 c4f4 e6e5 f4b4 e5e4 f3d4 e7c5 b4c5 b6c5 d4b3 a5c4 d2d1 c4e3 f2e3 a8c8 d1d7 b7a8 d7a7 c5c4 b3a5 g7c3 b2c3 a8d5');
INSERT INTO evaluations VALUES(1432,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(1432,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(1432,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(1432,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(1432,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(1432,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(1432,36,'cp -80','b7c6 g1e2 c5d4 c3d4 g8e7 h2h4 c6c5 d4c5 b6c5 b1c3 e7f5 g2g4 d5d4 g4f5 d4c3 b2c3 e6f5 h1h3 f8e7 d1d4 c5c6 c1g5 f7f6 e5f6 g7f6 g5f4 e8f7 d4d3 h8g8 a1d1 a7a5 d3f3 c6f3 h3f3 g8g4 f4d6 g4h4 d6e7 f7e7 f3g3 f5f4 g3g8 h4h1 e1d2 h1d1 d2d1 e7f7');
INSERT INTO evaluations VALUES(1309,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(1309,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(1309,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(1309,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(1309,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(1309,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(1309,36,'cp -56','g1e2 c5d4 c3d4 g8e7 h2h4 c6c5 d4c5 b6c5 b1c3 e7c6 e1g1 f8e7 c1e3 c5a5 h4h5 h7h6 f2f4 a8b8 e2d4 c6d4 d1d4 b8b4 d4a7 a5a7 e3a7 c8a6 f1f2 e7h4 f2c2 b4f4 a2a4 e8g8 a7e3 f4g4 c3b5 f8b8 e3a7 b8a8 c2c7 h4d8 c7c6 a6b5 a4b5 g4b4 b5b6');
INSERT INTO evaluations VALUES(5182,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5182,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5182,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(5182,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5182,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5182,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5182,36,'cp 0','a7a6 f1d3 g8f6 e1g1 d7d6 c3a4 c5a7 c2c4 e8g8 a4c3 f6g4 c1f4 g4e5 d3e2 e5f3 e2f3 e6e5 f4e3 c8e6 f3e2 d8a5 e3d2 a5d8');
INSERT INTO evaluations VALUES(4726,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4726,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4726,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(4726,36,'cp 0','f1d3 d7d6 e1g1 g8f6 c1f4 b7b5 e4e5 d6e5 f3e5 c8b7 c3e4 c5e7 d1f3 f6e4 e5c6 e4c5 d3e4 d8b6 f4e3 b6c7 e3f4');
INSERT INTO evaluations VALUES(15101,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(15101,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(15101,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(15101,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(15101,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(15101,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(15101,36,'cp 103','d4e5 c6e5 c4d5 f8d6 e3e4 d8e7 f1e2 e5g6 g1f3 e8g8 c1g5 h7h6 g5f6 e7f6 e1g1 f8e8 d1d4 f6e7 a1d1 c8d7 a2a3 d6c5 d4d2 c7c6 f1e1 c5b6 h2h3 e7f6 d5c6 d7c6 c3d5 c6d5 e4d5 a8d8 d5d6 g6f4 e2f1 e8e1 d2e1 d8d6 e1e8 g8h7');
INSERT INTO evaluations VALUES(15069,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(15069,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(15069,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(15069,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(15069,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(15069,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(15069,36,'cp 71','c6e5 c4d5 f8d6 e3e4 d8e7 f1e2 e5g6 g1f3 e8g8 c1g5 h7h6 g5f6 e7f6 e1g1 f8e8 d1d4 f6e7 f1e1 c8d7 g2g3 d6c5 d4d2 a8d8 g1g2 c7c6 a1c1 e7f6 e2d1 c5b6 a2a3 a7a6 d2c2 b6a7 c2e2 b7b5 d1b3 c6d5 c3d5 f6e6 g2h1 g8h8 c1d1 d7c6 e2c2 d8c8 h1g2 c6d5 e4d5 c8c2');
INSERT INTO evaluations VALUES(15820,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(15820,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(15820,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(15820,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(15820,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(15820,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(15820,36,'cp 63','c4d5 f8d6 e3e4 d8e7 f1e2 e5g6 g1f3 e8g8 c1g5 h7h6 g5f6 e7f6 e1g1 f8e8 d1d4 f6e7 d4d2 c8d7 f3d4 g6f4 f1d1 d7h3 g2g3 f4e2 d2e2 f7f5 e4f5 e7e2 c3e2 h3f5 e2c3 f5d7 d1d2 a8c8 a1d1 d7h3 d4b5 d6b4 d5d6 c7d6 b5d6 b4d6 d2d6');
INSERT INTO evaluations VALUES(15795,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(15795,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(15795,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(15795,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(15795,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(15795,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(15795,36,'cp 68','f8d6 e3e4 d8e7 f1e2 e5g6 c1g5 e8g8 g1f3 h7h6 g5f6 e7f6 e1g1 c8d7 d1d4 f6e7 g2g3 a8d8 a1d1 d6c5 d4c4 c7c6 d5c6 b7c6 c3a4 c5d6 g1g2 f8e8 c4c2 e7e6 a2a3 d6c7 a4c5 e6h3 g2g1 d7g4 d1d8 c7d8 f3d4');
INSERT INTO evaluations VALUES(20291,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(20291,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(20291,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(20291,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(20291,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(20291,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(20291,36,'cp 37','e4d5 e6d5 f1b5 d8e7 b5e2 e7c7 e1g1 g8f6 f1e1 c5d4 d2b3 f8e7 f3d4 e8g8 h2h3 a7a6 e2d3 f8e8 c1g5 c8d7 d1f3 h7h6 g5h4 f6e4 h4e7 e8e7 d3e4 a8e8 e4h7 g8h7 e1e7 e8e7 c2c3 c7e5 d4c6 b7c6 f3d3 d7f5 d3a6 f5e4 a1e1');
INSERT INTO evaluations VALUES(20232,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(20232,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20232,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(20232,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(20232,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(20232,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(20232,36,'cp 24','e6d5 f1b5 d8e7 b5e2 e7c7 e1g1 g8f6 f1e1 c5d4 d2b3 f8e7 b3d4 e8g8 h2h3 f8e8 d4b5 c7b8 c1g5 a7a6 b5d4 c8d7 e2d3 h7h6 g5h4 b8c7 d4c6 d7c6 f3d4 f6e4 h4e7 e8e7 f2f3 e4d6 d1d2 e7e1 a1e1 c7b6 c2c3 d6c4');
INSERT INTO evaluations VALUES(20522,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20522,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(20522,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20522,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(20522,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(20522,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(20522,36,'cp 30','f1b5 d8e7 b5e2 g8f6 e1g1 e7c7 f1e1 c5d4 d2b3 f8e7 b3d4 e8g8 c2c3 c8d7 c1e3 a7a6 e2d3 f8e8 d1c2 e7d6 h2h3 h7h6 a1d1 c6e7 e1e2 e7c6 d4c6 d7c6 d1e1 a8d8 c2d1 e8e7 e3d4 f6e4 d1c2 c6d7 d3e4 d5e4 e2e4');
INSERT INTO evaluations VALUES(20465,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(20465,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(20465,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(20465,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(20465,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(20465,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20465,36,'cp 35','f8d6 d4c5 d6c5 d2b3 c5d6 e1g1 g8e7 f1e1 e8g8 h2h3 h7h6 b3d4 c6d4 f3d4 d6c7 c2c3 e7f5 d4f3 c8e6 b5d3 f8e8 c1d2 d8d6 d1a4 e6d7 a4g4 d6f6 e1e8 a8e8 a1e1 e8e1 f3e1 d7e6 e1c2 a7a6 d2f4 f5d6 g4f3');
INSERT INTO evaluations VALUES(31930,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(31930,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31930,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31930,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(31930,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(31930,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31930,36,'cp 0','e2e3 b8d7 h2h3 g4f3 d1f3 c7c6 d2d4 e7e6 f1d3 f7f5 g2g4 g7g6 g4g5 h7h6 h3h4 d8b6 b1c3 e8c8 c3a4 b6b4 a4c3');
INSERT INTO evaluations VALUES(31929,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31929,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31929,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(31929,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(31929,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(31929,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(31929,36,'cp -12','b8d7 h2h3 g4f3 d1f3 c7c6 d2d4 e7e6 f1d3 f7f5 g2g4 g7g6 g4g5 g8e7 c1d2 e7c8 h3h4 h7h5 f3e2 f8g7 c2c4 e8g8 e1g1 f8e8 d2c3 c8d6 b1d2 c6c5 d4c5 d7c5 c3g7 g8g7 c4d5 e6d5 a1c1 d8b6 d2f3 c5d3 e2d3 e8e3 d3d5');
INSERT INTO evaluations VALUES(33683,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(33683,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(33683,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(33683,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(33683,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(33683,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(33683,36,'cp 59','c8g4 c2c4 g4f3 g2f3 d5f5 b1c3 e7e5 d4d5 f8c5 c1e3 c5e3 f2e3 f5g5 d1e2 g5h4 e2f2 h4f2 e1f2 b8d7 f1h3 a7a5 h3d7 f6d7 a1g1 g7g6 h2h4 e8e7 f2e2 c7c6 e2d3 a5a4 b2b3 a4b3 a2b3 h8g8 d3c2 d7f6 e3e4 f6h5 b3b4');
INSERT INTO evaluations VALUES(32615,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32615,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(32615,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(32615,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(32615,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(32615,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(32615,36,'cp 51','c2c4 d5d7 c1e3 e7e6 b1c3 b8c6 d4d5 e6d5 c4d5 c6b4 f1c4 c7c6 d5c6 b4c6 e1g1 f8e7 h2h3 g4f3 d1d7 f6d7 g2f3 e8g8 c3d5 e7d8 a1c1 d7b6 f3f4 b6d5 c4d5 g7g6 g1g2 d8b6 d5c6 b7c6 c1c6 f8d8 b2b4 d8d3 e3c5 b6c5 b4c5 d3a3 f1d1 a3a2');
INSERT INTO evaluations VALUES(39289,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(39289,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(39289,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(39289,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(39289,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(39289,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(39289,36,'cp 14','g8f6 d2d4 c5b6 b1c3 e8g8 c4b3 d7d6 c1f4 b8c6 e1g1 c8g4 f4e3 c6a5 g1h1 g4h5 e3g5 a5b3 a2b3 h7h6 g5f6 d8f6 c3d5 f6e6 d5f4 e6g4 f4h5 g4h5 b3b4 h5b5 c2c3 a7a5 b4a5 a8a5 a1a5 b6a5 f3h4 c7c5 d4c5 d6c5');
INSERT INTO evaluations VALUES(42671,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42671,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(42671,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(42671,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(42671,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42671,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(42671,36,'cp 12','d2d4 c5b6 a2a4 b8c6 c1f4 d7d5 e4d5 f6d5 f4g5 d8d6 b1c3 d5c3 b2c3 c8e6 d1e2 e8g8 c4e6 b6a5 g5d2 a8e8 e1g1 e8e6 e2d3 e6e8 a1e1 d6g6 e1b1 a5b6 f1e1 e8e1 b1e1 f7f6 g1f1 g6d3 c2d3 f8d8 e1b1 g8f7 h2h3 d8e8 d4d5 c6e5 f1e2 e5f3 e2f3 e8e5');
INSERT INTO evaluations VALUES(42673,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(42673,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42673,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(42673,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(42673,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(42673,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(42673,36,'cp 9','c5b6 a2a4 b8c6 c1f4 d7d5 e4d5 f6d5 f4g5 d8d6 b1c3 d5c3 b2c3 c8e6 d1e2 e8g8 c4e6 b6a5 g5d2 a8e8 f3g5 f7e6 e2d3 f8f5 g5f3 e8f8 e1g1 e6e5 a1b1 e5d4 c3d4 a5d2 d3d2 b7b6 c2c3 c6e7 b1e1 h7h6 d2c2 c7c5 h2h3 e7g6 e1e4 c5d4');
INSERT INTO evaluations VALUES(42371,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(42371,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(42371,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42371,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(42371,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42371,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(42371,36,'cp 19','a2a4 b8c6 c1f4 d7d5 e4d5 f6d5 f4g5 d8d6 b1c3 d5c3 b2c3 c8e6 d1e2 e8g8 c4e6 b6a5 g5d2 a8e8 e1g1 d6e6 e2e6 f7e6 f1e1 e6e5 d4d5 c6b8 c3c4 a5b6 d2e3 b8d7 a4a5 b6e3 e1e3 f8f4 e3c3 h7h6 a1e1 b7b6 a5b6 a7b6 g2g3 f4g4 f3d2 g4d4 d2b3 d4g4 h2h3');
INSERT INTO evaluations VALUES(48354,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(48354,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(48354,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(48354,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(48354,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(48354,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48354,36,'cp 0','c7c5 b2b3 c5d4 e3d4 f8d6 f1d3 e8g8 e1g1 b8c6 c1b2 b7b6 a2a3 c8b7 d1e2 a8c8 f3e5 c8c7 a1e1 c6e7 f2f4 g7g6 d2f3 f6e4 f3d2');
INSERT INTO evaluations VALUES(46250,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46250,36,'cp 1','b2b3 c5d4 e3d4 f8d6 f1d3 e8g8 e1g1 b8c6 c1b2 b7b6 a2a3 c8b7 d1e2 a8c8 f3e5 c8c7 a1e1 c6e7 f2f4 e7f5 g2g4 f5h4 f1f2 b6b5 b3b4 f6e4 d2e4 d5e4 d3e4 d6e5 e4b7 e5d4 e1d1 d4f2 g1f2 c7b7 d1d8 f8d8 f2e1');
INSERT INTO evaluations VALUES(51059,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(51059,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(51059,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(51059,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(51059,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(51059,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(51059,36,'cp 161','f3e5 d8e7 d2d4 d7d6 e5f3 e7e4 c1e3 g8f6 f1d3 e4e7 e1g1 e7d8 h2h3 f8e7 c2c4 e8g8 f1e1 f8e8 b1c3 b8d7 a2a4 a6a5 d1c2 d7f8 a1d1 c8d7 d4d5 f8g6 f3d4 c7c6 d3f5 c6d5 c4d5 a8c8 c2b3 d7f5 d4f5 e7f8 f5g3 f6d7');
INSERT INTO evaluations VALUES(51045,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(51045,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(51045,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(51045,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(51045,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(51045,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(51045,36,'cp 148','d8e7 d2d4 d7d6 e5f3 e7e4 c1e3 e4e7 f1d3 e7d8 c2c4 f8e7 b1c3 g8f6 h2h3 e8g8 d1c2 f8e8 e1g1 h7h6 f1e1 c7c6 e3f4 b8d7 a2a4 a6a5 a1d1 d7b6 c2b3 c8e6 e1e6 f7e6 d3g6 e8f8 d1e1 d6d5 c4c5 b6c4 e1e6 d8d7');
INSERT INTO evaluations VALUES(34129,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(34129,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(34129,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(34129,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(34129,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(34129,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(34129,36,'cp 166','d2d4 d7d6 e5f3 e7e4 c1e3 e4e7 f1d3 g8f6 e1g1 e7d8 f1e1 f8e7 c2c4 e8g8 b1c3 f8e8 h2h3 h7h6 d1c2 e7f8 a1d1 b8d7 g2g4 b7b5 c4b5 c8b7 d4d5 a6b5 d3b5 d7e5 c2e2 c7c6 f3e5 c6b5 e5c6 d8d7 e2b5 b7a6 b5b3 f6g4 e3d4 e8e1 d1e1');
INSERT INTO evaluations VALUES(52177,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(52177,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(52177,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(52177,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(52177,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(52177,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(52177,36,'cp 93','d4b5 d7d6 c2c4 g8f6 b5c3 b8c6 f1e2 c8e6 c1e3 f8e7 e1g1 a8c8 h2h3 e8g8 b1a3 a7a6 d1d2 d8d7 f1d1 e7d8 a3c2 d8a5 f2f3 f6h5 d2d6 d7d6 d1d6 a5c3 b2c3 c6a5 c4c5 e6c4 e2d3 c4d3 d6d3 a5c4 e3f2 h5f4 d3d7 f8d8');
INSERT INTO evaluations VALUES(52087,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(52087,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(52087,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(52087,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(52087,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(52087,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(52087,36,'cp 88','d7d6 c2c4 b8c6 b5c3 g8f6 f1e2 c8e6 e1g1 f8e7 c1e3 a8c8 b1a3 a7a6 a1c1 e8g8 a3c2 f6d7 b2b3 g8h8 e2g4 h7h6 h2h3 b7b5 c4b5 a6b5 d1d2 c8a8 c1a1 d7f6 g4f3 d8a5 a2a4 b5a4 a1a4 a5d8 a4a8 d8a8 b3b4 a8b7 b4b5');
INSERT INTO evaluations VALUES(7262,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7262,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7262,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7262,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(7262,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(7262,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(7262,36,'cp 23','c3d5 d7d6 c1e3 c5e3 f2e3 c8e6 d1e2 e8g8 d5f6 d8f6 c4e6 f6e6 f3d2 f7f5 e4f5 f8f5 e1c1 a8f8 d1f1 f5f7 f1f7 f8f7 h1f1 f7f1 e2f1 e6f7 f1e2 b7b5 b2b3 a6a5 d2e4 c6e7 e2d2 a5a4 b3a4 b5a4 c1b2 e7d5 c2c4 d5b6 d2b4');
INSERT INTO evaluations VALUES(45613,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(45613,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(45613,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(45613,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45613,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(45613,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(45613,36,'cp 26','e2e3 f8e7 b1c3 h7h6 g5h4 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 c8d7 f3e5 d7e8 c3b5 b8c6 h4f6 g7f6 e5d3 c5e7 b5c7 a8c8 c7e8 f8e8 c4b5 c8c7 b5c6 c7c6 d1c1 e8c8 c1c6 c8c6 h1c1 c6c1 d3c1 f6f5 c1b3 e7d6 h2h3');
INSERT INTO evaluations VALUES(45612,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(45612,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45612,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(45612,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45612,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(45612,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(45612,36,'cp 25','h7h6 g5h4 f8e7 b1c3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 b8c6 a2a3 c5e7 c4b5 g7g5 h4g3 c8d7 f3e5 c6e5 g3e5 f8d8 b5d7 d8d7 h2h3 a8d8 d1d4 g8h7 h1d1 d7d4 d1d4 d8d4 e5d4 a7a6 e2d3 h7g6 f2f3 f6d7 b2b4 e7d6');
INSERT INTO evaluations VALUES(52852,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(52852,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(52852,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(52852,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(52852,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(52852,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(52852,36,'cp -25','d5d4 c3e2 c6c5 e2g3 b8c6 f2f4 f7f5 f1b5 d8c7 a2a4 e7e6 b5c4 g8e7 d2d3 c7d7 g1f3 c6a5 c4a2 e7d5 c1d2 a5c6 a4a5 b7b5 a5b6 a7b6 a2d5 a8a1 d1a1 d7d5 a1a8 c6e7 a8a7 d5b7 a7b7 c8b7 e1g1 b7c6 f3g5 c6d5 g3h5 e7c6');
INSERT INTO evaluations VALUES(18156,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(18156,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(18156,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(18156,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(18156,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(18156,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(18156,36,'cp 31','f1d3 d8c7 d1f3 g8f6 f3g3 c7g3 h2g3 d7d5 f2f3 f6d7 c1e3 f8c5 e3f4 e8e7 e1c1 h7h6 g3g4 c5d6 f4e3 a6a5 e4d5 c6d5 c3b5 d6c5 e3f4 c8b7 a2a4 f7f6 f4g3 b7c6 c1b1 d7e5 d3e2 c5b6 b2b3 e5f7 c2c4 a8d8 c4d5 c6d5');
INSERT INTO evaluations VALUES(18157,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(18157,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(18157,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(18157,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(18157,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(18157,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(18157,36,'cp 36','d7d5 e1g1 g8f6 d1f3 f8e7 f3g3 f6h5 g3g4 g7g6 c1h6 d5d4 c3a4 e6e5 g4f3 e7g5 d3c4 a8a7 h6g5 d8g5 f3a3 g5e7 a3c5 e7c5 a4c5 e8e7 c5d3 e7d6 a1d1 a6a5 f2f3 c8e6 c4e6 d6e6 g1f2 f7f6 c2c3');
INSERT INTO evaluations VALUES(31555,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(31555,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(31555,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(31555,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(31555,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(31555,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(31555,36,'cp -44','d8b6 b2b3 c6c5 b1c3 c5d4 c3a4 b6c7 f1b5 b8c6 f3d4 c7e5 g5e3 a8c8 e1g1 g8f6 d4f5 e5f5 e3a7 f8d6 a7c5 d6c5 a4c5 e8g8 c5b7 f5f4 c2c4 f4b8 b5c6 c8c6 c4d5 e6d5 b7a5 c6c5 b3b4 b8b4 a5b3 c5c3 a1b1 b4a3 d1d2 c3c4 d2b2 f8a8 b3d2 c4c3');
INSERT INTO evaluations VALUES(32809,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(32809,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(32809,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(32809,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(32809,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(32809,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(32809,36,'cp -43','b2b3 c6c5 b1c3 c5d4 c3a4 f8b4 g5d2 b4d2 d1d2 b6c7 f1b5 b8c6 b5c6 b7c6 e1g1 c6c5 c2c3 d4d3 c3c4 d5d4 b3b4 c5b4 f3d4 c7e5 d4f5 e5f5 d2b4 g8e7 f1d1 a8d8 a4b2 e7c6 b4a4 d8c8 b2d3 e8g8 d3e1 c8c7 e1f3 f5a5 a4b3 h7h6 h2h4 f8c8 b3e3 c6e7 f3e5 e7f5 e3e4 a7a6');
INSERT INTO evaluations VALUES(32808,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(32808,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(32808,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(32808,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(32808,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(32808,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(32808,36,'cp -59','c6c5 b1c3 c5d4 c3a4 b6a5 g5d2 a5c7 f1b5 b8c6 f3d4 c7e5 d2e3 a8c8 e1g1 g8f6 d4f5 e5f5 e3a7 f8d6 a7c5 d6c5 a4c5 e8g8 c5b7 f5f4 c2c4 c6d4 b7d6 c8d8 d6b7 f6g4 g2g3 f4e4 f2f3 e4e3 g1g2 d8b8 b7c5 d4b5 c4b5 e3c5 f3g4 c5b5 a2a4 b5b4 a4a5 b4e4 d1f3 b8b3 f3e4 d5e4 g4g5 b3b2 g2g1');
INSERT INTO evaluations VALUES(34291,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(34291,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(34291,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(34291,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(34291,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(34291,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(34291,36,'cp 130','g1f3 e6d6 d2d4 c8f5 e1g1 g8f6 f3h4 f5c8 f1e1 g7g6 e2c4 f8g7 d1e2 e7e6 h4f3 b8d7 a2a4 e8g8 a4a5 b7b5 a5b6 d7b6 b2b3 d6c7 c1a3 f8d8 e2e5 c7e5 d4e5 f6d7 a3d6 a7a5 d6c7 d8e8 c4f1 c8a6 c3a4 b6a4 a1a4 a6f1 g1f1');
INSERT INTO evaluations VALUES(34290,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(34290,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(34290,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(34290,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(34290,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(34290,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(34290,36,'cp 124','e6d6 d2d4 c8f5 e1g1 g8f6 f3h4 f5c8 f1e1 g7g6 h4f3 f8g7 h2h3 e8g8 c1e3 b8d7 d1d2 d7b6 e3f4 d6d8 a2a4 a7a5 a1d1 c8f5 f3e5 f5e6 b2b3 f8e8 e2f1 b6d7 e5c4 e6d5 f4h2 e7e6 h2d6 f6h5 d6a3 d5c4 b3c4');
INSERT INTO evaluations VALUES(40194,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40194,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40194,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40194,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(40194,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(40194,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40194,36,'cp 0','d1f3 a7a6 e1c1 d8c7 a2a3 b8c6 f3g3 f6h5 g3h4 h5f6');
INSERT INTO evaluations VALUES(53204,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(53204,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(53204,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(53204,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(53204,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(53204,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(53204,36,'cp 51','d4d5 g8f6 b1c3 e7e6 f1c4 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 d5d6 e7d8 c4d5 a8a7 a1d1 b8d7 d5c6 c8b7 c6b7 a7b7 f1e1 c5c4 g2g3 d7c5 f3e5 f6d7 e5c6 c5d3 e1e2 d7c5 b2b3 b7d7 b3c4 b5c4 c3d5 c5e6 e2e4 e6f4');
INSERT INTO evaluations VALUES(53225,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(53225,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(53225,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(53225,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(53225,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(53225,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(53225,36,'cp 62','e7e6 f1c4 g8f6 b1c3 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e7e2 c4e2 e8g8 f3d2 d6c7 d2c4 b7b5 d5d6 c7d8 e2f3 a8a7 c4e5 f8e8 c1f4 c8b7 f1e1 e8e6 a1d1 b7f3 g2f3 a7b7 a2a4 h7h6 f4g3 b5a4 c3a4 f6d7 e5d3 c5c4 e1e6 f7e6 a4c5 d7c5');
INSERT INTO evaluations VALUES(54126,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(54126,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(54126,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(54126,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(54126,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54126,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(54126,36,'cp 69','b1c3 g8f6 f1c4 e6d5 e4d5 a7a6 g1f3 f8d6 d1e2 d8e7 e1g1 e8g8 e2e7 d6e7 c1f4 b7b5 f1e1 e7d8 c4f1 b8d7 d5d6 f6h5 g2g3 h5f4 g3f4 a8b8 f1g2 c5c4 f3e5 d7e5 f4e5 g7g6 a2a4 b5a4 c3a4 g8g7 e1e2 c8g4 e2c2 g4f5 c2c4 f7f6 f2f4 f6e5 f4e5');
INSERT INTO evaluations VALUES(55483,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(55483,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55483,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(55483,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55483,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(55483,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(55483,36,'cp 34','f1d3 g8f6 e1g1 f8e7 f1e1 e8g8 h2h3 e7d6 b2b3 b8c6 a2a3 f8e8 e1e8 d8e8 a1a2 c6e7 c2c4 b7b6 b1c3 a7a6 a2e2 c8e6 c1b2 c7c6 a3a4 a6a5 d1c2 e8d7 f3e5');
INSERT INTO evaluations VALUES(55482,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(55482,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(55482,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55482,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(55482,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(55482,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(55482,36,'cp 25','f8b4 c2c3 b4d6 e1g1 g8e7 f1e1 e8g8 f3h4 b8d7 b1d2 d7f6 h2h3 c8e6 d2f1 f6e4 d3e4 d5e4 e1e4 d8d7 c1d2 a8d8 e4e1 e7g6 h4g6 f7g6 d2e3 b7b6 d1c2 d7f7 a1c1 g6g5 f2f3 d8e8 e3f2 e6f5 c2b3 f7b3 a2b3 e8e1 c1e1');
INSERT INTO evaluations VALUES(10075,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(10075,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(10075,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(10075,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(10075,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(10075,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(10075,36,'cp -10','c5e7 d2d3 d7d5 e4d5 f6d5 e1g1 e8g8 a2a4 c8e6 b4b5 c6a5 c4a2 e7f6 f1e1 a7a6 b5a6 a8a6 f3e5 c7c5 d3d4 f8e8 c1d2 c5d4 c3d4 a5c6 b1c3 c6d4 c3d5 e6d5 a2d5 d8d5 d2c3 e8e5 d1d4 d5d4 c3d4 e5e1 a1e1 h7h6 e1e8 g8h7 d4f6 g7f6 h2h3 f6f5 e8d8 a6a4 d8b8 h7g7');
INSERT INTO evaluations VALUES(11512,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(11512,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(11512,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(11512,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(11512,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(11512,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(11512,36,'cp -66','f6e4 c3e4 d7d5 c4d3 f7f5 e4c3 e5e4 d3e2 e4f3 e2f3 d5d4 d1e2 e8f7 e2c4 c8e6 f3d5 d8f6 d5e6 f6e6 c4e6 f7e6 c3e2 e6d7 d2d3 a8e8 c1d2 f8d6 a2a3 e8e7 e1f1 h8e8 e2g1 e7e5 h2h4 e5c5 a1c1 c5b5 b2b4 b5d5 c1b1 a6a5 b4b5 c6e5 a3a4 d5c5 b1c1 h7h5 g2g3');
INSERT INTO evaluations VALUES(15602,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(15602,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(15602,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(15602,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(15602,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(15602,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(15602,36,'cp 46','d2d4 e5d4 f3d4 f8e7 f1c4 e8g8 e1g1 c6d4 d1d4 c7c6 a3a4 d8c7 c1f4 f6h5 f4e3 c8e6 c4e6 f7e6 e4e5 d6e5 d4c4 h5f4 a1d1 g8h8 e3f4 e5f4 c4e6 e7f6 d1d7 c7b6 e6c4 a8e8 f1d1 b6b2 c3e4 b2e5 e4f6');
INSERT INTO evaluations VALUES(15562,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(15562,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(15562,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(15562,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(15562,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(15562,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(15562,36,'cp 41','e5d4 f3d4 f8e7 f1c4 e8g8 e1g1 c6d4 d1d4 c8e6 a3a4 h7h6 h2h3 e6c4 d4c4 f8e8 f1d1 f6d7 a4a5 e7f6 c3d5 c7c6 d5e3 d7c5 f2f3 b7b5 a5b6 d8b6 c2c3 f6e5 e3g4 c5e6 c1e3 b6b2 g4e5 d6e5 c4c6');
INSERT INTO evaluations VALUES(50834,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(50834,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50834,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(50834,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50834,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(50834,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50834,36,'cp 18','c7c5 c2c3 b8c6 g1f3 c5d4 c3d4 c8g4 b1d2 e7e6 f1d3 g8f6 a1c1 f6h5 f4e5 f8b4 h2h3 g4f3 d1f3 a8c8 e1e2 h5f6 e5f6 d8f6 f3f6 g7f6 a2a3 b4d2 e2d2 h7h5 h3h4 e8d7 g2g3 f6f5 c1c3 c6e7 h1c1 c8c3 c1c3');
INSERT INTO evaluations VALUES(53286,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(53286,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(53286,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(53286,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53286,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(53286,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(53286,36,'cp 28','e4d5 d8d5 b1c3 d5d8 g1f3 b8c6 g2g3 c8g4 f1g2 e7e6 h2h3 g4f3 d1f3 g8f6 e1g1 f8e7 c3e4 e8g8 c2c3 a8c8 e4f6 e7f6 c1e3 b7b6 a1d1 f6g5 e3g5 d8g5 h3h4 g5e5 f1e1 e5f6 f3f6 g7f6 g2c6 c8c6 d3d4 c5d4 d1d4');
INSERT INTO evaluations VALUES(53233,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53233,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(53233,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53233,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(53233,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(53233,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53233,36,'cp 30','d8d5 g1f3 b8c6 b1c3 d5d8 g2g3 g8f6 f1g2 c8g4 h2h3 g4f3 d1f3 e7e6 f3e2 d8d7 c3e4 f8e7 c2c3 e8g8 c1e3 b7b6 e4f6 e7f6 h3h4 a8d8 e1g1 c6e7 a1d1 e7d5 e3d2 d7c7 f1e1 d8d7 a2a3 a7a5 g2e4');
INSERT INTO evaluations VALUES(34369,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(34369,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(34369,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(34369,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(34369,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34369,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(34369,36,'cp 26','b1c3 d5d8 g1f3 b8c6 g2g3 c8g4 f1g2 g8f6 h2h3 g4f3 d1f3 e7e6 f3e2 d8b6 c3e4 f8e7 c2c3 e8g8 h3h4 a8d8 e1g1 d8d7 e4d2 b6a6 d2c4 f6d5 a2a4 c6a5 a1a3 f8d8 f1e1 e7f6 c4a5 a6a5 g2e4 f6e7 e4f3 h7h6');
INSERT INTO evaluations VALUES(34366,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(34366,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(34366,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(34366,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(34366,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(34366,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(34366,36,'cp 36','d5d8 g1f3 b8c6 g2g3 c8g4 h2h3 g4f3 d1f3 e7e6 f1g2 g8f6 f3e2 d8b6 c3e4 f8e7 e1g1 e8g8 c2c3 f8d8 h3h4 f6e4 g2e4 d8d7 f1e1 a8d8 c1e3 b6a6 e3f4 a6b6 a2a3 e7f6 e1c1 c6e7 b2b4 e7d5 f4d2');
INSERT INTO evaluations VALUES(28805,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(28805,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(28805,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(28805,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28805,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(28805,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(28805,36,'cp 10','f6h5 f4e5 d7d6 f3g5 h5f6 e5f6 e7f6 g5f3 g7g6 g2g3 f8g7 f1g2 e8g8 c2c4 f6f5 e1g1 b8d7 f3e1 b7g2 e1g2 d7f6 b1c3 c7c6 d1b3 f8e8 a1c1 a8c8 g2f4 d8e7 f1e1 e7d7 e1d1 g7h6 f4d3 h6g7 a2a4 h7h5');
INSERT INTO evaluations VALUES(38892,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(38892,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(38892,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(38892,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(38892,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(38892,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(38892,36,'cp 222','b7b6 g1f3 g8e7 e1g1 e8g8 f1e1 b8d7 c1g5 c8b7 c4b3 h7h6 g5h4 a8c8 c3c4 g7g5 h4g3 e7g6 h2h3 d7f6 d4d5 f6h5 g3e5 g6e5 f3e5 h5f4 b3c2 d8e7 a2a4 a7a5 g2g3 f4g6 e5g4 h6h5 g4e3 g8g7 d1d3 e7f6 e4e5 f6e5');
INSERT INTO evaluations VALUES(50147,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50147,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50147,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50147,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50147,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(50147,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(50147,36,'cp 1','c2c4 f8g7 b1c3 d7d6 g1f3 e8g8 g2g3 b8d7 f1g2 d7b6 d1d3 e7e6 e1g1 f8e8 f1e1 e6d5 c4d5 c7c6 d5c6 b7c6 f3d4 c6c5 d4c6 d8d7 c6a5 d6d5 c1g5 c5c4 d3c2 d7f5 c2f5 c8f5 a1d1 f6e4 g2e4 d5e4 a5c6 g7c3 b2c3 f5g4 c6d4 a8c8');
INSERT INTO evaluations VALUES(50145,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(50145,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(50145,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(50145,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50145,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50145,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(50145,36,'cp 3','d7d6 b1c3 f8g7 e2e4 e7e6 g1f3 e8g8 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 c1g5 d8b6 g5e3 b6b4 a2a3 b4b3 e3f4 a8c8 d3d2 f8d8 a1c1 b7b5 e1e7 a6a5 c1e1 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(56158,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(56158,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(56158,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(56158,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(56158,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(56158,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(56158,36,'cp 28','d8d5 b1c3 d5d4 d1d4 e5d4 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 b5d4 g8f6 a1d1 f6e4 d2e3 e4d6 d4b5 h8e8 e3f3 b8c6 f3g3 e8e6 f2f3 c8d7 b5d6 e6g6 g3f2 g6d6 d1d6 c7d6 f1d3 g7g6 g1e2 c6b4 h1d1 a8c8 c2c3 b4d3 d1d3 d8e7 d3d1 d7e6 a2a3 c8c5 h2h4 h7h5 e2d4 a7a6');
INSERT INTO evaluations VALUES(6588,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(6588,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(6588,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(6588,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(6588,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(6588,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(6588,36,'cp 109','c6e5 d2d4 e5g6 f1e2 d7d6 h1f1 g8e7 d1e1 e8g8 f2g1 d6d5 c3d5 e7d5 e4d5 f8e8 e1f2 d8d5 c2c4 d5f5 e2f3 c7c6 c1d2 c8e6 b2b3 h7h6 a1e1 f5f6 d2c3 g6h4 f3e4 f6f2 f1f2 a8d8 f2f1 e8f8 a2a4 d8d7 e4c2 f8d8 a4a5 h4g6');
INSERT INTO evaluations VALUES(6891,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(6891,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(6891,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(6891,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(6891,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(6891,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(6891,36,'cp 125','d2d4 e5g6 f1e2 d7d6 d1d3 g8f6 h1f1 e8g8 h2h3 h7h6 f2g1 c8e6 b2b3 a8c8 a2a4 c7c5 d4d5 e6d7 d3f3 f6h7 f3g3 d8e7 e2d3 e7e5 g3e5 g6e5 c1f4 f8e8 d3e2 f7f6 h3h4 g7g5 f4g3 g5h4 g3h4 g8g7 a1e1 h7g5 h4g3 g5f7 c3b5');
INSERT INTO evaluations VALUES(6888,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(6888,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(6888,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(6888,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(6888,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(6888,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(6888,36,'cp 118','e5g6 f1e2 d7d6 d1d3 g8f6 h1f1 e8g8 f2g1 h7h6 h2h3 c8e6 d3f3 d6d5 e4e5 f6d7 e2d3 c7c5 c3d5 d8h4 d5c7 a8c8 c7e6 f7e6 f3g4 h4g4 h3g4 f8f1 g1f1 g6e7 c2c3 c5d4 c3d4 d7b6 f1f2 g8f7 d3e4 c8c4 c1e3 c4b4 a1f1 f7g8 b2b3 b6d5 f2e2');
INSERT INTO evaluations VALUES(6815,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(6815,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(6815,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(6815,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(6815,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(6815,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(6815,36,'cp 105','f1e2 d7d6 h1f1 g8e7 f2g1 e8g8 d1e1 d6d5 h2h4 d5e4 h4h5 g6h8 e1f2 e7f5 c3e4 d8d4 e2d3 f7f6 f2d4 f5d4 c1e3 d4c6 a1e1 c6e5 e3c5 f8d8 h5h6 f6f5 h6g7 g8g7 e4g3 e5d3 e1e7 g7g8 c2d3 b7b6 c5b4 h8g6 e7c7 d8d7 c7d7 c8d7 g3f5 d7f5 f1f5');
INSERT INTO evaluations VALUES(14644,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(14644,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(14644,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(14644,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(14644,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(14644,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(14644,36,'cp 36','c8g4 c4b5 f6d7 d4e5 d6e5 h2h3 g4f3 d1f3 c6d4 f3g3 c7c6 b5d3 h6h5 c3e2 h5h4 g3g4 d7f6 g4g5 f6d7 g5d8 e8d8 e2d4 e5d4 d3c4 d7e5 c4b3 a7a5 f1d1 f8c5 c2c3 d4d3 c1f4 f7f6 f4e5 f6e5 d1d3 d8c7 a1d1 a8d8 d3d8 h8d8 g1f1 g7g5 d1d8 c7d8 f1e2 b7b5');
INSERT INTO evaluations VALUES(25070,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(25070,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(25070,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(25070,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(25070,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(25070,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(25070,36,'cp 44','c4b5 f6d7 d4e5 d6e5 h2h3 g4f3 d1f3 c6d4 f3g3 c7c6 b5d3 h6h5 c3e2 h5h4 g3g4 d7f6 g4g5 f6d7 e2d4 d8g5 c1g5 e5d4 g5d2 f8e7 b2b4 c6c5 b4c5 d7c5 a2a4 e8g8 f1c1 f8c8 g1f1 c8c7 a4a5 a8c8 f1e2 e7d6 a1b1 c5d3 e2d3');
INSERT INTO evaluations VALUES(26356,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26356,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26356,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26356,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26356,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(26356,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26356,36,'cp 0','g8e7 e1g1 e7f5 a2a4 a7a5 d4c5 f8c5 b1d2 d5d4 d2b3 c6f3 d1f3 d8b6 b3c5 b6c5 f1d1 e8g8 g2g4 f5h4 f3g3 h4g6 c3d4 c5d5 h2h4 f7f6 c1e3 f6e5 d4e5 d5e5 g3e5 g6e5 g4g5 e5f3 g1f1 f3h4 d1d7 f8f7 a1d1 a8f8 d7f7 f8f7 d1d8 f7f8 d8d7 f8f7');
INSERT INTO evaluations VALUES(24587,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24587,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24587,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(24587,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(24587,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24587,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(24587,36,'cp 8','e1g1 e7f5 a2a4 h7h6 f1e1 f8e7 d4c5 e7c5 b1d2 d5d4 d2b3 c6f3 d1f3 d8b6 c3d4 f5d4 b3d4 c5d4 a4a5 b6b4 e1d1 e8g8 f3g3 b4c5 d1d4 c5d4 c1h6 g7g6 h6f8 a8f8 h2h4 f8c8 a1e1 d4d2 g3e3 c8c2 e3d2 c2d2 b2b3 d2d3 e1e3 d3d5 a5a6 b7a6 g2g3 g8g7 g1g2 d5b5 e3c3 b5e5');
INSERT INTO evaluations VALUES(49701,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(49701,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(49701,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(49701,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(49701,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(49701,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(49701,36,'cp -42','b8c6 d2d4 d7d5 d4e5 c6e5 b1d2 f8b4 a2a3 b4d6 g1f3 c8g4 e1g1 e8g8 h2h3 g4h5 b2b4 c7c6 c1b2 d8e7 g3g4 e5f3 d1f3 h5g6 c2c4 a7a5 b4a5 a8a5 f1d1 f8d8 a3a4 d6e5 b2e5 e7e5 c4d5 a5d5 d2c4 e5e6 d1d5 c6d5 c4d2');
INSERT INTO evaluations VALUES(16617,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(16617,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(16617,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(16617,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(16617,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(16617,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(16617,36,'cp -62','d2d3 d7d5 b1d2 c8f5 e3e4 f5g4 g1f3 f6e4 d3e4 d5e4 d2e4 d8d1 e1d1 c6d4 c2c3 d4f3 h2h3 g4h5 g3g4 e8c8 d1c2 f3h4 g4h5 h4g2 h1g1 g2h4 c1g5 h4f3 g5d8 f3g1 d8c7 g1f3 c7d6 f7f5 d6f8 h8f8 e4d2 f3d2 c2d2 c8d7 d2e3 d7e6 a1d1 f8c8 d1g1 e6f6 g1d1 a7a6 a2a4 b7b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(52772,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(52772,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(52772,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(52772,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(52772,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52772,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52772,36,'cp 30','g8f6 g1f3 c8g4 h2h3 g4h5 e1g1 b8d7 c2c4 e7e6 b1d2 f8d6 b2b3 e8g8 c1b2 a8c8 a1c1 h5g6 d3g6 h7g6 f3e5 d6c7 d1e2 d8e7 c1c2 a7a5 f2f4 a5a4 c4c5 a4b3 a2b3 c8a8 b3b4 a8a2 d2f3');
INSERT INTO evaluations VALUES(44950,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44950,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44950,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44950,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(44950,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(44950,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(44950,36,'cp 25','g1f3 c8g4 h2h3 g4h5 c2c4 e7e6 e1g1 f8d6 b1d2 e8g8 b2b3 b8d7 c1b2 h5g6 d3g6 h7g6 d1c2 f8e8 a2a3 d8e7 f1e1 a7a5 e3e4 d5e4 d2e4 f6e4 c2e4 d7f6 e4d3 b7b5 b2c1 b5c4 b3c4 c6c5 d4d5 e6e5 c1g5 e5e4');
INSERT INTO evaluations VALUES(55602,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55602,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55602,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55602,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55602,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55602,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55602,36,'cp 0','g8f6 e1g1 f8e7 c2c3 b8c6 d3c2 d6d5 d2d3 d5e4 d3e4 d8d1 f1d1 a7a5 b1d2 c8e6 d2f1 e8g8 f1e3 a8d8 d1d8 f8d8 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(47884,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47884,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47884,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47884,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47884,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47884,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47884,36,'cp 0','e1g1 f8e7 c2c3 b8c6 d3c2 d6d5 d2d3 d5e4 d3e4 d8d1 f1d1 a7a5 c2a4 e7d6 b1d2 c8e6 f3g5 e6g4 g5f3');
INSERT INTO evaluations VALUES(47883,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47883,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47883,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47883,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47883,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47883,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47883,36,'cp 0','f8e7 c2c3 b8c6 d3c2 d6d5 d2d3 d5e4 d3e4 d8d1 f1d1 a7a5 c2a4 e7d6 b1d2 c8e6 f3g5 e6g4 g5f3');
INSERT INTO evaluations VALUES(56899,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(56899,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(56899,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(56899,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(56899,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(56899,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(56899,36,'cp 62','f8g7 e1g1 g8e7 d2d4 e5d4 f3d4 b8c6 d4c6 e7c6 b1c3 d7d6 f1e1 c8e6 c4f1 e8g8 c1e3 c6e5 f2f4 e5c4 f1c4 e6c4 e3d4 f7f6 b2b3 c4f7 d1f3 c7c6 f4f5 f8e8 a2a4 e8e7 a1d1 d8a5 h2h3 a8e8 f3f2 d6d5 e4d5 e7e1 d1e1 e8e1 f2e1');
INSERT INTO evaluations VALUES(39050,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(39050,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(39050,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(39050,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(39050,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(39050,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(39050,36,'cp 63','e1g1 d7d6 c2c3 g8f6 d2d4 f6e4 d4e5 d6e5 d1d8 e8d8 c4f7 c8f5 c1e3 b8d7 b1d2 e4d6 f7b3 a7a5 a2a4 c7c6 f3g5 g7f6 f1e1 h7h6 g5e6 d8c8 a1d1 h8e8 e6c5 d7c5 e3c5 c8c7 c5d6 c7d6 d2e4 d6e7 e4d6 e8d8 d6b7 d8d1 b3d1');
INSERT INTO evaluations VALUES(13890,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(13890,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(13890,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(13890,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(13890,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(13890,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(13890,36,'cp 22','c1e3 f6g4 e3c1 c8e6 f2f3 g4f6 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 e1c1 e8c8 a2a4 d5b3 d1d8 c8d8 c2b3 d8c7 f1c4 f7f6 h1d1 b7b6 h2h4 h7h5 c4e6 f8d6 d1d2 h8d8 e6f7 d8h8 d2c2 c7d7 f7c4 c6e7 c4b5 d7e6');
INSERT INTO evaluations VALUES(13887,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(13887,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(13887,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13887,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(13887,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(13887,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(13887,36,'cp 28','f6g4 e3c1 g4f6 f2f3 c8e6 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5d3 c8c7 c2c3 f8e7 c1c2 d5e6 a2a4 d8d7 d3e2 e6f5 c2c1 d7d1 h1d1 h7h6 e2d3 f5d3 d1d3 h8d8 c1c2 f7f5 a4a5 e5e4 e3b6 c7c8 d3d8 e7d8 b6d8 c8d8 f3e4 f5e4 c2d2 c6e5 d2e2 e5c4');
INSERT INTO evaluations VALUES(18698,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(18698,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18698,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(18698,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(18698,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18698,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(18698,36,'cp 20','d7d6 c2c4 g8f6 b1c3 a7a6 b5d4 f8e7 f1e2 e8g8 e1g1 d6d5 c4d5 e6d5 d4c6 b7c6 e4d5 f6d5 c3d5 c6d5 c1e3 c8f5 b2b3 f5e4 e2d3 e7f6 a1c1 e4d3 d1d3 d5d4 e3d2 f8e8 c1c6 d8d5 f1c1 d5b5 c6c4 h7h6 a2a4 b5e5 g2g3 e5e2 d3e2');
INSERT INTO evaluations VALUES(20655,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20655,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20655,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(20655,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(20655,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(20655,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(20655,36,'cp 9','c2c4 g8f6 b1c3 a7a6 b5d4 f8e7 f1e2 e8g8 e1g1 d6d5 e4d5 e6d5 c4d5 f6d5 d4c6 b7c6 c3d5 c6d5 c1e3 c8f5 b2b3 f5e4 e2d3 e7f6 a1c1 e4d3 d1d3 d5d4 e3d2 d8d7 c1c4 f8e8 f1e1 d7b5 e1c1 h7h6 a2a4 b5e5');
INSERT INTO evaluations VALUES(20653,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(20653,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(20653,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(20653,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20653,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(20653,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(20653,36,'cp 9','g8f6 b1c3 a7a6 b5d4 f8e7 f1e2 e8g8 e1g1 d6d5 c4d5 e6d5 e4d5 f6d5 d4c6 b7c6 c3d5 c6d5 c1e3 c8f5 b2b3 e7f6 e3d4 f6d4 d1d4 f8e8 f1e1 e8e4 d4d2 d8f6 e2f3 e4d4 d2e3 f5e4 f3e2 h7h6 f2f3 e4f5 a1d1 d4d1 e1d1 a8c8 e3d4 f6d4 d1d4');
INSERT INTO evaluations VALUES(22420,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(22420,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(22420,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(22420,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(22420,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(22420,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(22420,36,'cp 79','c2c4 g8f6 b1c3 e7e5 d2d4 g7g6 f1e2 f8g7 d1c2 e8g8 e1g1 f8e8 c1e3 e5d4 f3d4 c7c6 a1d1 d8e7 e3g5 d7c5 f2f3 a7a5 f1e1 a5a4 e2f1 e7c7 g5e3 c8e6 h2h3 c7a5 e3f4 a8d8 f4h2 f6h5 g2g4 h5f6 g1h1 h7h5');
INSERT INTO evaluations VALUES(23288,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(23288,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(23288,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(23288,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(23288,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(23288,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(23288,36,'cp 114','g8f6 e1g1 f8c5 b1c3 f5f4 d3d4 c5d4 f3d4 c6d4 c1f4 d7d6 f4g5 d4e6 f2f4 e5f4 g5h4 c7c6 f1e1 g7g5 e4e5 d6e5 h4g5 e6g5 e1e5 c8e6 c4e6 g5e6 e5e6 e8f7 e6d6 d8e7 d1d4 h8d8 d6d8 e7d8 a1d1 d8b6 d4b6 a7b6 d1d4 f7e6 d4f4 a8d8 a2a4 e6e5 f4f3 d8d2 f3f2 d2d8');
INSERT INTO evaluations VALUES(17107,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(17107,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(17107,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(17107,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(17107,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(17107,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(17107,36,'cp 142','e1g1 f8c5 b1c3 f5f4 d3d4 c5d4 f3d4 c6d4 c1f4 d7d6 f4g5 c7c6 f2f4 d8e7 f4e5 d6e5 g5f6 g7f6 d1h5 e8d8 h5h6 h8f8 f1f2 c8d7 a1f1 d8c7 c4d3 d4e6 c3e2 a7a5 h2h3 a5a4 f2f6 f8f6 f1f6 b7b5 g1h2 a8g8 f6f2 c7b6 f2f1 b6b7 h6f6 e7g7 f6g7 g8g7 g2g4 b7c7 b2b3 a4a3 e2g3 e6f4 g3f5');
INSERT INTO evaluations VALUES(24955,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(24955,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(24955,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(24955,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(24955,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(24955,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24955,36,'cp 50','g4f3 g2f3 e7e5 d4e5 c6e5 f3g2 c7c6 b1d2 f8c5 c2c4 d5c4 d1c2 d8d4 d2f3 e5f3 g2f3 f6d5 c1d2 e8g8 a1d1 c5e7 d2g5 d5b4 c2a4 d4c5 g5e7 c5e7 a2a3 b4a6 a4c4 a6c7 h2h4 a8d8 b2b4 c7e6 g1g2 d8d1 f1d1 f8d8 d1d8 e7d8 e2e3 d8e7 h4h5 h7h6 f3g4 e6c7 a3a4 a7a6 e3e4 e7f6');
INSERT INTO evaluations VALUES(24969,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24969,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(24969,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(24969,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(24969,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(24969,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(24969,36,'cp 43','g2f3 e7e5 c2c3 d8d7 f3g2 h7h5 d4e5 c6e5 b1d2 f8c5 d2f3 e5f3 e2f3 e8g8 c1g5 d7f5 f3f4 c7c6 g5f6 g7f6 b2b4 c5b6 a2a4 a7a5 h2h4 f8d8 b4a5 b6a5 a1b1 d8d7 d1d4 d7e7 c3c4 d5c4 d4c4 g8g7 g2f3');
INSERT INTO evaluations VALUES(24968,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(24968,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(24968,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(24968,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(24968,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(24968,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24968,36,'cp 43','e7e5 c2c3 d8d7 f3g2 f8e7 d4e5 c6e5 b1d2 a8d8 d2f3 e5f3 e2f3 e8g8 c1f4 h7h6 f1e1 e7d6 f4d2 f8e8 d1c2 d7c6 a1d1 b7b5 g2f1 e8e1 d1e1 a7a6 b2b3 d6c5 g1g2 c5b6 f1d3 d8e8 e1e8 f6e8 h2h4 b6a7 d2c1 a7b6');
INSERT INTO evaluations VALUES(29246,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29246,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29246,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(29246,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29246,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(29246,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(29246,36,'cp 12','f5g6 e1g1 b8d7 d1e2 f6e4 f3d2 e4d2 c1d2 f8e7 e3e4 d5c4 d3c4 d7f6 e4e5 f6d7 d2e3 d7b6 c4d3 g6d3 e2d3 b6d5 a1c1 e8g8 c3e4 d5b4 d3b1 b4d5 a2a3 d8b6 b1c2 f8d8 h2h3 h7h6 c1d1 d8d7 d1d3 a7a5 f1d1 a5a4');
INSERT INTO evaluations VALUES(29319,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(29319,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29319,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29319,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29319,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29319,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(29319,36,'cp 5','d1e2 b8d7 e1g1 f6e4 f3d2 e4d2 c1d2 f8e7 e3e4 d5c4 d3c4 d7b6 d2e3 b6c4 e2c4 d8d7 c4b3 e8g8 f2f3 h7h6 e3f2 e7d6 f1d1 f7f5 d1e1 a8e8 a1d1 d6c7 f2g3 c7b6 g3e5 f5e4 c3e4 g6e4 e1e4 d7d5 b3a3 c6c5 d1e1');
INSERT INTO evaluations VALUES(38893,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(38893,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(38893,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(38893,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(38893,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(38893,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(38893,36,'cp 220','b2c3 g8e7 g1f3 b8d7 e1g1 e8g8 f1e1 b7b6 c1g5 c8b7 a2a4 h7h6 g5h4 a7a5 d1d2 d8e8 h4g3 e7g6 c4d3 e8d8 a1d1 d8e7 g3c7 d7c5 c7g3 c5d3 d2d3 b7c6 d1a1 e7d7 a1b1 c6a4 b1b6 a4c6 h2h4 a5a4 h4h5 g6e7 f3e5');
INSERT INTO evaluations VALUES(45354,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45354,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45354,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45354,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45354,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45354,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45354,36,'cp 0','e2e3 c5d4 e3d4 a7a6 g1f3 b8c6 f3e5 c8d7 f1e2 e7e6 e1g1 f8e7 a2a3 e8g8 d1d2 a8c8 d2e3 d7e8 a1d1 d8b6 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(45352,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45352,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45352,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45352,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45352,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45352,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45352,36,'cp 0','c5d4 e3d4 c8g4 f1e2 g4e2 d1e2 a7a6 g1f3 e7e6 e1g1 f8e7 c3d1 b8d7 c2c3 f6h5 f4e5 e8g8 d1e3 h5f6 e5f4 f6h5');
INSERT INTO evaluations VALUES(45462,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(45462,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45462,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45462,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45462,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45462,36,'cp 0','e3d4 a7a6 g1f3 b8c6 f1e2 c8f5 f3e5 c6e5 d4e5 d5d4 e5f6 d4c3 b2c3 d8d1 a1d1 f5c2 e2f3 c2d1 f3b7 a8a7 b7c6 e8d8 e1d1 e7f6 d1c2 f8c5 h1d1 d8e7 d1e1 e7d8');
INSERT INTO evaluations VALUES(45425,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45425,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45425,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45425,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45425,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45425,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45425,36,'cp 0','a7a6 g1f3 b8c6 f1e2 c8f5 f3e5 c6e5 d4e5 d5d4 e5f6 d4c3 b2c3 d8d1 a1d1 f5c2 e2f3 c2d1 f3b7 a8a7 b7c6 e8d8 e1d1 e7f6 d1c2 f8c5 h1d1 d8e7 d1e1 e7d8');
INSERT INTO evaluations VALUES(43431,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43431,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43431,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43431,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43431,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43431,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43431,36,'cp 0','g1f3 b8c6 f1e2 c8f5 f3e5 c6e5 d4e5 d5d4 e5f6 d4c3 b2c3 d8d1 a1d1 f5c2 e2f3 c2d1 f3b7 a8a7 b7c6 e8d8 e1d1 e7f6 d1c2 f8c5 h1d1 d8e7 d1e1 e7d8');
INSERT INTO evaluations VALUES(47664,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(47664,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(47664,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47664,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(47664,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(47664,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(47664,36,'cp 25','g5f6 d8f6 b1c3 f8b4 e2e3 c7c5 c4d5 e6d5 d1b3 e8g8 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 f6e4 c3c4 d5c4 f1c4 e4c5 e1e2 c8d7 d1c1 a8c8 h1d1 c5e6 c4d5 c8c1 d1c1 e6d4 e3d4 f8c8 c1c8 d7c8 e2e3 g8f8 d5f3 f8e7 d4d5');
INSERT INTO evaluations VALUES(47659,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(47659,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(47659,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47659,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(47659,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47659,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47659,36,'cp 19','f8b4 b1c3 d8f6 d1b3 c7c5 c4d5 e6d5 e2e3 e8g8 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 c8d7 f2f3 f8c8 d1b1 c8c5 b1b7 c5c3 e1d2 c3a3 b7b2 a8c8 f1e2 f6e8 h1a1 e8d6 b2b3 a3a4 a2a3 g8f8 b3b4 a7a5 b4b6 f8e7 e2d3 c8c7 g2g3 d6c4 d3c4 a4c4 h2h4');
INSERT INTO evaluations VALUES(53738,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(53738,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(53738,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(53738,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(53738,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(53738,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(53738,36,'cp -78','d8e7 d1e2 e7e2 e1e2 f7f6 h1e1 b8a6 e2f1 e8f7 d2d4 a6c7 a4b3 h7h5 h2h4 c8f5 b1c3 a7a5 a2a4 g8e7 c3e2 e7g6 c2c4 d5c4 b3c4 f5e6 c4e6 c7e6 e1d1 e6c7 d1d3 f8b4 c1d2 c7d5 d2b4 d5b4 d3c3 h8d8 g2g3 g6e7 f3d2');
INSERT INTO evaluations VALUES(6228,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(6228,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(6228,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(6228,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6228,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6228,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(6228,36,'cp 10','d7d6 c1e3 a7a6 b1d2 g8e7 e1g1 e8g8 c2c3 e7g6 d1c2 c5a7 f1e1 a7e3 f2e3 c6b8 d3d4 b8d7 b2b4 d7f6 c4d3 a6a5 h2h3 b7b6 b4b5 d8e7 e1f1 c8b7 f1f2 a8e8 a1f1 e5d4 c3d4 b7e4 d2e4');
INSERT INTO evaluations VALUES(5748,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5748,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(5748,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5748,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(5748,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(5748,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(5748,36,'cp 10','b1c3 a7a6 c1e3 g8f6 e1g1 e8g8 e3c5 d6c5 c3d5 d8d6 d5e3 b7b5 c4a2 c6d4 h2h3 a8b8 f1e1 a6a5 a3a4 c7c6 f3h4 d6c7 c2c3 d4e6 a2e6 c8e6 h4f5 f8e8 f5g3 c5c4 a4b5 c6b5 d3c4 b5c4');
INSERT INTO evaluations VALUES(5745,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(5745,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(5745,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(5745,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5745,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(5745,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(5745,36,'cp 8','g8f6 e1g1 e8g8 c3a4 c5b6 f1e1 c8e6 c4e6 f7e6 c2c3 d8e8 b2b4 c6e7 a4b6 a7b6 a3a4 e7g6 h2h3 e8d7 a1a3 d7e7 c1e3 f8f7 a4a5 f6d7 f3h2 g6f4 d1c2 b6a5 e3f4 f7f4 a3a5 a8a5 b4a5 f4f7 d3d4 e7f6 h2f3 f6g6 f3h4');
INSERT INTO evaluations VALUES(8549,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(8549,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(8549,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(8549,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(8549,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(8549,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(8549,36,'cp 74','d2d4 c8d7 d4d5 c6b8 b5c4 a7a5 h2h3 b8a6 c1e3 c7c6 d1e2 c6d5 c4d5 a6c7 d5c4 a5a4 e3g5 e8g8 g5f6 e7f6 a1d1 f6e7 c3d5 c7d5 c4d5 a8c8 c2c3 d8b6 d1d2 c8c5 f1d1 d7c6 a2a3 c6d5');
INSERT INTO evaluations VALUES(8544,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(8544,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(8544,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(8544,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(8544,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(8544,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(8544,36,'cp 73','c8d7 d4d5 c6b8 a2a4 e8g8 b2b3 a7a5 b5d7 f6d7 c1a3 b8a6 d1e2 e7g5 a1e1 g5h6 g2g3 d7c5 f3d2 b7b6 d2c4 f7f5 e4f5 f8f5 f2f3 f5f7 c3e4 c5d7 e2d3 d7f6 h2h4 f6e4 f3e4 d8e8 f1f3 a6b4 a3b4 a5b4 f3f7 e8f7');
INSERT INTO evaluations VALUES(15882,30,'mate 1',NULL);
INSERT INTO evaluations VALUES(15882,31,'mate 1',NULL);
INSERT INTO evaluations VALUES(15882,32,'mate 1',NULL);
INSERT INTO evaluations VALUES(15882,33,'mate 1',NULL);
INSERT INTO evaluations VALUES(15882,34,'mate 1',NULL);
INSERT INTO evaluations VALUES(15882,35,'mate 1',NULL);
INSERT INTO evaluations VALUES(15882,36,'mate 1','f3f7');
INSERT INTO evaluations VALUES(18907,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(18907,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(18907,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18907,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(18907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18907,36,'cp 0','g8f6 d4c6 b7c6 e4e5 f6d5 b1d2 f8e7 d2c4 e8g8 f1d3 a7a5 e1g1 c8a6 d1g4 f7f5 e5f6 f8f6 g4e4 f6f5 e4e2 f5f8 e2e4');
INSERT INTO evaluations VALUES(12479,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12479,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12479,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12479,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12479,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12479,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12479,36,'cp 0','d4c6 b7c6 e4e5 f6d5 b1d2 f8e7 d2c4 e8g8 f1d3 a7a5 e1g1 c8a6 d1h5 f7f5 e5f6 d5f6 h5e2 a5a4 c1f4 a6c4 d3c4 f6d5 f4e5 e7f6 e5d6 f6e7');
INSERT INTO evaluations VALUES(31881,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31881,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31881,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(31881,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(31881,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31881,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(31881,36,'cp 28','f1g2 c7c6 e1g1 e7e6 d2d3 g8f6 f3h4 f5g4 h2h3 g4h5 f2f4 f8e7 g3g4 h5g6 e2e3 a7a5 b1d2 e8g8 a2a4 b8d7 d1e1 e7b4 e1f2 b7b5 d2f3 d8b6 g1h1 a8e8 f2g3 d7c5 h4g6 h7g6 c2c3 c5b3 c3b4 b3a1 b4a5 b6a5');
INSERT INTO evaluations VALUES(31882,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31882,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31882,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31882,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31882,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31882,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(31882,36,'cp 15','e7e6 e1g1 c7c6 d2d3 f5g6 c2c4 d5c4 d3c4 d8d1 f1d1 g8f6 b1c3 b8d7 f3h4 g6h5 g2f3 h5f3 h4f3 h7h6 b2b3 e8c8 c1b2 f8b4 c3a4 h8e8 a2a3 b4e7 b3b4 g7g5 f3d2 b7b6 c4c5 b6c5 a4c5 d7c5 b4c5');
INSERT INTO evaluations VALUES(32333,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32333,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32333,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(32333,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(32333,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(32333,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(32333,36,'cp 31','c2c4 c7c6 e1g1 d5c4 a2a4 g8f6 b1a3 f8a3 a1a3 e8g8 b2b3 c4c3 d2c3 f5e4 b3b4 b8d7 c3c4 a7a5 b4a5 d8a5 c1b2 f8e8 d1a1 e6e5 f1d1 a5c7 b2c3 e4c2 d1c1 c2f5 f3d2 d7c5 a4a5 f5e4 d2e4 f6e4 c1d1 e4c3 a1c3');
INSERT INTO evaluations VALUES(55912,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(55912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55912,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(55912,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(55912,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(55912,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55912,36,'cp -11','d5e4 b1c3 f7f5 f2f3 b8c6 f3e4 g8f6 g1f3 f8c5 h1f1 e8g8 d1e2 c6d4 f3d4 c5d4 d2d3 c7c5 c1d2 a8b8 e1c1 b7b5 e2e1 f8e8 c3e2 d4e5 d2f4 e5f4 g3f4 c8b7 e1g3 d8c7 f1g1 a7a5 e4e5 f6g4 g2b7');
INSERT INTO evaluations VALUES(56086,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(56086,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(56086,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(56086,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(56086,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56086,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(56086,36,'cp -12','b1c3 f7f5 d2d3 g8f6 d3e4 d8d1 e1d1 f5e4 c3e4 b8c6 c2c3 e6e5 e4f6 g7f6 g2h3 c8h3 g1h3 h7h5 d1c2 f8d6 h1e1 e8f7 f2f3 b7b5 c1e3 b5b4 a1d1 b4c3 b2c3 c6a5 f3f4 a5c4 f4e5 h8e8 e3c1 e8e5 e1e5 c4e5 c1f4 a8e8 d1d4 e5f3');
INSERT INTO evaluations VALUES(56720,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(56720,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(56720,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(56720,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(56720,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(56720,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(56720,36,'cp 145','d8e7 f3g5 g8f6 c4f7 e8d8 f7b3 d8e8 g5f7 h8g8 b1c3 c8e6 f7g5 e6b3 a2b3 c7c6 f2f4 h7h6 g5f3 b8d7 e1g1 e7e6 d1e1 a7a6 c1e3 f6g4 e3d2 e8d8 c3a4 b7b5 f4f5 e6f7 h2h3 g4f6 d2a5 d8c8 a4b6 d7b6 a5b6 f8e7 b6f2 g8e8 g1h1 c8b7');
INSERT INTO evaluations VALUES(5099,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(5099,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(5099,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(5099,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(5099,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(5099,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(5099,36,'cp 26','d2d4 c5d4 c3b5 d7d6 f3d4 c6d4 b5d4 g8f6 f1e1 a7a6 c4f1 e8g8 c2c4 c8d7 a2a4 f6g4 h2h3 d8b6 a4a5 b6d4 h3g4 d4d1 e1d1 d7c6 f2f3 f8d8 a1b1 d6d5 e4d5 e6d5 c1g5 d8d7 g5e3 g7f8 e3b6 a8c8 b2b3 f8e7 g1f2 g8f8 b1c1 d5c4 f1c4 d7d1 c1d1 f8g7 d1d3');
INSERT INTO evaluations VALUES(20294,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20294,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20294,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20294,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20294,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20294,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20294,36,'cp 0','e4d5 e6d5 d3d4 c5c4 f1c4 d5c4 d4d5 d8e7 c1e3 c6e5 d1e2 e5f3 g2f3 c8f5 e1c1 f7f6 d5d6 e7d7 e2c4 f5e6 c4d3 e8f7 h1e1 g7g6 c3b5 e6f5 d3d5 f5e6');
INSERT INTO evaluations VALUES(20236,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(20236,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20236,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(20236,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(20236,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(20236,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20236,36,'cp 0','e6d5 d3d4 c5c4 f1c4 d5c4 d4d5 d8e7 c1e3 c6e5 d1e2 e5f3 g2f3 c8f5 e1c1 f7f6 d5d6 e7d7 c3e4 d7c6 d1d4 f5e6 d6d7 e6d7 h1d1 d7f5 d4c4 c6e6 c4c7 e6a6 d1d3 a6a2 c7b7 a2a1 c1d2 a1a5 d2c1');
INSERT INTO evaluations VALUES(20532,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(20532,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20532,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20532,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20532,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20532,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(20532,36,'cp 2','d3d4 c5c4 f1c4 d5c4 d4d5 d8e7 c1e3 c6e5 d1e2 e5f3 g2f3 c8f5 e1c1 f7f6 d5d6 e7d7 c3e4 d7c6 d1d4 a8c8 h1d1 f5d7 e4c5 b7b5 c5d7 c6d7 a2a4 a7a6 e2d2 e8f7 d2a5 g7g6 a5a6 g8h6 a6b5 d7b5 a4b5 h6f5 d6d7 c8d8 d4c4 f8d6 e3c5 d8d7 c5d6 f5d6 c4c6 f7e7 c2c4 h8d8 b2b3 d8a8 c1c2 a8a2 c2c3');
INSERT INTO evaluations VALUES(24804,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(24804,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(24804,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(24804,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(24804,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(24804,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(24804,36,'cp 113','c4d5 d8d5 b1c3 d5d6 a2a3 e7e6 b2b4 f8e7 f1e2 e8g8 c1b2 f6e4 e1g1 e4c3 b2c3 a7a6 f3d2 c6a7 e2f3 a7b5 c3b2 c7c6 d2b3 d6d8 b3c5 b5d6 d1e2 d8c7 e3e4 f5g6 a1c1 a6a5 f1e1 a5b4 a3b4 c7b6 d4d5 c6d5 e4d5');
INSERT INTO evaluations VALUES(30241,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(30241,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(30241,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(30241,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(30241,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(30241,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(30241,36,'cp 26','h2h3 g4f5 f1d3 f5d3 d1d3 c7c6 e1g1 f8e7 f3h4 e8g8 h4f5 f8e8 c1h6 g7g6 f5e7 e8e7 a2a4 d8d6 h6g5 b8d7 f2f4 e7e8 f4f5 f6e4 c3e4 e8e4 d3f3 a8e8 c2c3 e4e2 a4a5 f7f6 f5g6 h7g6 g5f6 d7f6 f3f6 d6g3 f6f3 g3f3 f1f3 e2b2 a5a6');
INSERT INTO evaluations VALUES(30238,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30238,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(30238,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(30238,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30238,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(30238,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(30238,36,'cp 38','g4f5 c1f4 f8d6 d1e2 f5e6 f4d6 d8d6 e2e5 d6b6 e1c1 b8d7 e5g3 e8c8 f3g5 h8e8 f1d3 h7h6 g5e6 f7e6 d3g6 e8e7 h1e1 a7a6 a2a3 d7f8 f2f3 f8g6 g3g6 b6d6 c1b1 c8b8 g6d3 f6d7 d3e3 d7b6 h3h4 c7c6');
INSERT INTO evaluations VALUES(38648,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(38648,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(38648,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(38648,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(38648,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(38648,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(38648,36,'cp 112','g8f6 b1c3 d8e7 c1g5 d6b4 g5f6 e7f6 d1d4 b4c3 b2c3 b8d7 e2e4 e8g8 f1d3 f6h6 e1g1 b7b6 f1e1 d7c5 d4e3 h6e3 e1e3 f8d8 f3d4 c8d7 f2f4 g7g6 g1f2 g8g7 e4e5 a8c8 h2h4 h7h5 a1d1 d7e6 d4e6 c5e6 g2g3 e6c5 d1e1 a7a5 a2a3');
INSERT INTO evaluations VALUES(41744,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(41744,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(41744,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(41744,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(41744,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(41744,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(41744,36,'cp 123','b1c3 d8e7 c1g5 d6b4 g5f6 e7f6 d1d4 b4c3 b2c3 b8d7 e2e4 e8g8 f1d3 f6h6 e1g1 b7b6 f1e1 d7c5 d4e3 h6e3 e1e3 c8d7 f3d4 a8d8 f2f4 g7g6 a1e1 f8e8 g2g3 a7a5 g1f2 g8g7 e4e5 h7h5 d3c4 d7g4 f2g2 c5e6 c4e6 g4e6 d4e6 e8e6 e3d3');
INSERT INTO evaluations VALUES(41742,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(41742,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(41742,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(41742,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(41742,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(41742,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(41742,36,'cp 117','d8e7 c1g5 d6b4 g5f6 e7f6 d1d4 b4c3 b2c3 b8d7 e2e4 e8g8 f1d3 f6h6 e1g1 b7b6 f1e1 d7c5 d4e3 h6e3 e1e3 c8d7 f3d4 a8d8 f2f4 g7g6 g1f2 h7h5 d3c4 g8g7 c4d5 h5h4 f2f3 f8e8 c3c4 d8c8 d4c6 d7c6 d5c6 e8d8 e4e5');
INSERT INTO evaluations VALUES(39919,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(39919,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(39919,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(39919,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(39919,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(39919,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(39919,36,'cp 15','g2g3 e8g8 f1g2 c5d4 f3d4 d7d5 c4d5 f6d5 d1b3 d8a5 c1d2 b8c6 d4c6 b7c6 e1g1 b4c3 d2c3 d5c3 b2c3 c8a6 b3b4 a5b4 c3b4 a8b8 a2a3 a6e2 f1c1 f8c8 g2c6 g7g6 c1c5 a7a6 g1g2 g8g7 a1c1 c8d8 c1c2 e2b5 c2c3 d8d2 g2f3 h7h6 c6b5 b8b5');
INSERT INTO evaluations VALUES(50789,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(50789,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(50789,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(50789,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(50789,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(50789,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(50789,36,'cp 50','f1d3 b8c6 g1f3 c8d7 a2a3 g8e7 e1g1 e7f5 d4c5 f8c5 b1d2 c5a7 f1e1 h7h6 d2b3 e8g8 c1f4 f5h4 f3h4 d8h4 f4g3 h4d8 d3c2 c6e7 b3d4 a7d4 d1d4 e7f5 d4d2 d8b6 a3a4 a8c8 c2d3 b6b3 a4a5 d7b5 d3f5 e6f5 e5e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(51898,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(51898,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51898,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(51898,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(51898,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51898,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51898,36,'cp -3','g8f6 c1e3 b8a6 e3d4 f8c5 e2e3 b7b6 f1e2 e8g8 c2c4 d7d5 c4d5 f6d5 e1g1 c8b7 b1c3 c5d4 d1d4 d5c3 d4c3 d8e7 f1d1 a8c8 c3e5 a6c5 a1c1 h7h6 f3d4 f8d8 h2h3 b7d5 a2a3 a7a5 e5g3 c5e4 g3e5 c8c5 c1c5');
INSERT INTO evaluations VALUES(44154,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44154,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(44154,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44154,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44154,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44154,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44154,36,'cp 0','c1e3 b8a6 e3d4 f8c5 e2e3 b7b6 f1e2 e8g8 e1g1 c8b7 d4f6 d8f6 d1d7 b7f3 e2f3 f6b2 b1c3 a8d8 d7a7 b2c3 a7a6 d8d2 a1c1 g7g6 a6a4 c5a3 c1b1 a3c5');
INSERT INTO evaluations VALUES(51977,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(51977,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51977,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(51977,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(51977,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(51977,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(51977,36,'cp 14','g1f3 d7d5 f1b5 c8d7 b5d7 d8d7 e4d5 e6d5 e1g1 g8f6 d2d4 f8e7 c1b2 c5d4 f3d4 b8c6 d1f3 e8g8 d4f5 f8e8 f5e7 d7e7 b2f6 e7f6 f3f6 g7f6 b1a3 a8c8 f1d1 c6b4 c2c4 e8d8 g1f1 a7a6 c4d5 d8d5 d1d5 b4d5 a3c4 b7b5 c4d6 c8c2 a2a3 g8g7 h2h3 g7g6 a1d1 d5c3');
INSERT INTO evaluations VALUES(52809,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(52809,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52809,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52809,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(52809,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(52809,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(52809,36,'cp -18','e7e5 g1f3 f8d6 d3d4 e5d4 f1d3 g8e7 f3d4 e8g8 e1g1 b8d7 e4d5 e7d5 d2e4 d7e5 e4d6 d8d6 f1e1 e5d3 d1d3 h7h6 c1d2 d5f6 d2c3 c6c5 d4b3 d6d3 c2d3 b7b6 c3f6 g7f6 d3d4 c5c4 b3d2 f8d8 d2c4 c8e6 c4e3 d8d4 a2a3 d4d2 b2b4 g8g7 e1d1 a8d8');
INSERT INTO evaluations VALUES(53770,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(53770,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(53770,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(53770,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53770,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(53770,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53770,36,'cp -5','g1f3 f8d6 d3d4 e5d4 f1d3 g8e7 f3d4 e8g8 e1g1 b8d7 e4d5 e7d5 d2e4 d7e5 e4d6 d8d6 f1e1 c8g4 f2f3 g4h5 c1g5 a8e8 g5h4 e5d3 d1d3 d6b4 h4f2 e8e1 a1e1 b4b2 d3b3 b2b3 d4b3 b7b6 c2c4 d5b4 e1e7 f8a8 b3d4 c6c5 a2a3 b4d3 d4b5 g8f8 e7c7 h7h6 b5d6 d3f2');
INSERT INTO evaluations VALUES(53769,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53769,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(53769,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53769,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(53769,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53769,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53769,36,'cp 0','f8d6 d3d4 e5d4 f1d3 g8e7 f3d4 e8g8 e1g1 b8d7 e4d5 e7d5 d2e4 d7e5 f1e1 d6c7 d3f1 h7h6 h2h3 e5g6 c2c3 d5f4 d1f3 g6h4 f3d1');
INSERT INTO evaluations VALUES(38091,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38091,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38091,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38091,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38091,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38091,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38091,36,'cp 0','d3d4 e5d4 f1d3 g8e7 f3d4 b8d7 e4d5 e7d5 e1g1 e8g8 d2e4 d7e5 f1e1 d6c7 d3f1 h7h6 h2h3 e5g6 c1d2 d5f4 c2c3 d8e7 d1c2 c8d7 e4g3 e7f6 g3e4');
INSERT INTO evaluations VALUES(1527,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(1527,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(1527,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(1527,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(1527,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(1527,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(1527,36,'cp 85','c3d4 g8e7 b1c3 e7f5 c3a4 b6d8 b2b4 f8e7 f1d3 e8g8 c1b2 f7f6 d3f5 e6f5 e1g1 f5f4 f1e1 f6e5 d4e5 c8e6 a4c3 a7a6 c3e2 d8e8 e2d4 c6d4 f3d4 e8g6 d1f3 a8c8 a1c1 g8h8 d4e6 g6e6 c1c8 f8c8 f3f4');
INSERT INTO evaluations VALUES(5463,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5463,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5463,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5463,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5463,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5463,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5463,36,'cp 0','c1e3 g8f6 h2h3 e8g8 e1g1 b7b6 f1e1 c8b7 d1d2 a7a6 c4b3 e7e6 a2a4 d8c7 a1d1 a8d8 e3h6 c6e5 f3h2 e5c6');
INSERT INTO evaluations VALUES(5464,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5464,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5464,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5464,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5464,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5464,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5464,36,'cp 0','g8f6 h2h3 e8g8 e1g1 a7a6 a2a4 b7b6 c3d5 e7e6 d5f6 g7f6 d3d4 c5d4 f3d4 c8b7 d4c6 b7c6 d1d3 d8d7 e3d4 f6d4 d3d4 b6b5 a4b5 c6b5 c4b5 d7b5 c2c4 b5b4 f1b1 a6a5 b2b3 f8b8 d4d3 g8g7 d3d4');
INSERT INTO evaluations VALUES(8248,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8248,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8248,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8248,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8248,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8248,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8248,36,'cp 0','h2h3 e8g8 e1g1 b7b6 d1d2 c8b7 e3g5 g8h8 g5h6 e7e6 f1e1 a7a6 a2a4 d8c7 a1d1 a8d8 c4b3 c6e5 f3h2 e5c6');
INSERT INTO evaluations VALUES(11441,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(11441,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(11441,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(11441,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(11441,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(11441,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(11441,36,'cp 187','d2d3 f8c5 e5f3 d8e7 b1c3 c8g4 h2h3 g4h5 d1e2 e8c8 a2a3 h8e8 g2g4 h5g6 c1e3 c5e3 f2e3 f6d7 e1c1 f7f6 f3d4 g6f7 h1f1 c6c5 d4f5 e7f8 e2f2 d7e5 f5g3 c8b8 d3d4 e5d7 f1e1 f7g6 c1b1 f8e7 g3f5');
INSERT INTO evaluations VALUES(15029,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(15029,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(15029,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(15029,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(15029,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(15029,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(15029,36,'cp 163','f1b5 c8d7 e4d5 c6d4 b5d7 d8d7 d1d4 d7g4 d4g4 f6g4 h2h3 g4e5 c1f4 f8d6 e1c1 e8c8 h1e1 h8e8 c3e4 e5d3 d1d3 d6f4 c1d1 e8e5 e4c3 f7f5 d3f3 e5e1 d1e1 f4c1 b2b3 g7g6 c3e2 c1g5 f3g3 g5e7 g3d3 d8e8 e1f1 c8d7 e2d4 e7b4 c2c3 b4d6 g2g3 a7a6 c3c4 d6b4 d4c2 a6a5 a2a3 b4d6');
INSERT INTO evaluations VALUES(15859,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15859,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15859,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15859,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15859,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15859,36,'cp 0','e2e3 f8e7 h2h4 e8g8 g1f3 c7c6 b5c3 d8a5 f3d2 a5b6 c3a4 b6a5');
INSERT INTO evaluations VALUES(15858,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15858,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15858,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15858,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15858,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15858,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15858,36,'cp 0','f8e7 h2h4 e8g8 g1f3 c7c6 b5c3 d8a5 f3d2 a5b6 c3a4 b6a5');
INSERT INTO evaluations VALUES(25709,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(25709,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(25709,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(25709,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(25709,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(25709,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(25709,36,'cp 83','b5c6 d7c6 d2d4 f7f6 f3h4 g7g6 f2f4 e5f4 c1f4 g8e7 h4f3 f8g7 d4d5 c6d7 d1d3 b7b5 a1e1 e8g8 a2a3 d8c8 d3d2 g6g5 f4e3 h7h6 c3e2 c8b7 e3d4 e7g6 e2g3 a8e8 b2b3 c7c5 d4b2 g6f4 h2h3 h6h5 f1f2 h5h4');
INSERT INTO evaluations VALUES(25701,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(25701,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(25701,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(25701,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(25701,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(25701,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(25701,36,'cp 82','d7c6 d2d4 f7f6 f1e1 g8e7 d1d3 d8c8 a2a4 e7g6 c3d5 f8e7 d3b3 e7d8 a4a5 c6d7 d4e5 d6e5 c1e3 d7e6 b3a3 e8f7 a1d1 h8e8 f3d2 f7g8 b2b4 h7h6 f2f3 e6f7 d2b3 f6f5 b3c5 f5f4 e3f2 a8b8 a3c3 b8a8 c5d3 c8d7');
INSERT INTO evaluations VALUES(25663,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(25663,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(25663,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(25663,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(25663,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(25663,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(25663,36,'cp 74','d2d4 f7f6 a2a4 g8e7 a4a5 d8c8 d1d3 e7g6 c3d5 c6d5 e4d5 f8e7 c1d2 c8d7 a1e1 e8g8 h2h4 d7g4 e1e4 g4d7 d4e5 d6e5 h4h5 g6h8 f1e1 a8e8 h5h6 g7g6 d3b3 h8f7 b3b7 f7d6 b7a6 d6e4 e1e4 d7d5 a6e2 d5e6 a5a6 e8a8');
INSERT INTO evaluations VALUES(25660,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(25660,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(25660,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(25660,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(25660,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(25660,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(25660,36,'cp 75','f7f6 f3h4 g7g6 f2f4 e5f4 c1f4 g8e7 h4f3 f8g7 d1d3 b7b5 a1e1 e8g8 d4d5 c6d7 f1f2 e7c8 b2b3 c8b6 c3e2 d8e7 e2d4 a8e8 f4g3 e8c8 f2e2 e7f7 g3f2 c8e8 h2h3 h7h6 f2h4 g8h7 h4g3 f6f5 e4f5 d7f5');
INSERT INTO evaluations VALUES(39239,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(39239,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(39239,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(39239,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(39239,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(39239,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(39239,36,'cp -21','g8f6 g2g3 e8g8 f1g2 a7a5 g1f3 d7d6 e1g1 c7c6 h2h3 f8e8 g1h2 h7h6 d3d4 e5d4 f3d4 a5a4 a1b1 b8d7 b2b4 a4b3 a2b3 c5b4 c1b2 b4c3 b2c3 f6e4 g2e4 e8e4 d4f5 d7f6 f5d6 e4e6 d6c8 d8c8 c3f6 e6f6 d1d3 c8c7 h2g2 a8d8');
INSERT INTO evaluations VALUES(43293,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(43293,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(43293,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(43293,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(43293,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(43293,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(43293,36,'cp 154','c2c4 d5b4 f1e2 e8g8 b1c3 d7d6 e5d6 c7d6 e1g1 b8d7 c1e3 e7e6 d1d2 b4c6 e3g5 c6e7 a1d1 d7f6 f1e1 d6d5 f3e5 f6e4 c3e4 d5e4 e5g4 f7f6 g5h6 g7h6 g4h6 g8g7 h6g4 c8d7 g4e3 a8b8 b2b4 a7a6 d1c1 b7b5 a2a3 b5c4 e2c4');
INSERT INTO evaluations VALUES(43292,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(43292,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(43292,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(43292,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(43292,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(43292,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(43292,36,'cp 137','d5b4 b1c3 e8g8 f1e2 d7d6 e5d6 c7d6 e1g1 c8g4 h2h3 g4f3 e2f3 b8c6 a2a3 b4a6 c1e3 a6c7 a1c1 e7e6 d4d5 e6d5 f3d5 c7d5 c3d5 g7b2 c1b1 b2g7 b1b7 a8b8 b7b5 g8h8 d1a4 c6d4 b5b1 d4e2 g1h1 d8h4 b1b8 f8b8 a4a7');
INSERT INTO evaluations VALUES(52885,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(52885,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(52885,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(52885,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(52885,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(52885,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(52885,36,'cp 41','b7b5 a2a4 e7e6 a4b5 c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8e7 f3e5 d8c7 c1a3 e8g8 e1g1 b8c6 a3e7 c6e7 b1d2 e7g6 e5g6 h7g6 d1b3 c8b7 f1c1 f8c8 c4f1 c7d7 c1c8 d7c8 a1a5 f6e4 d2e4 b7e4 b3a3 c8c2 a5a7 a8a7 a3a7 e4d5 a7e7');
INSERT INTO evaluations VALUES(51650,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(51650,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51650,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(51650,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(51650,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(51650,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(51650,36,'cp 51','a2a4 e7e6 a4b5 c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8d6 e1g1 e8g8 c1a3 c8b7 b1d2 a7a5 d1b3 a8a7 a3d6 d8d6 b3a3 d6a3 a1a3 b7f3 g2f3 b8c6 f1c1 f8a8 g1f1 c6b4 c4b3 f6d5 d2c4 g7g6 a3a4 g8g7 c1a1 b4c6 f3f4 a8b8');
INSERT INTO evaluations VALUES(51660,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51660,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51660,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(51660,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(51660,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(51660,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(51660,36,'cp 55','e7e6 a4b5 c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8d6 e1g1 e8g8 c1a3 c8b7 c4e2 a7a5 b1d2 b8c6 a3d6 d8d6 d2c4 d6c7 d1d2 f8b8 f1c1 c7d8 a1a3 h7h6 h2h3 f6e4 d2b2 e4d6 c4d6 d8d6 b2a1 b7a6 e2a6 a8a6 f3d2 d6d8');
INSERT INTO evaluations VALUES(51697,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(51697,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(51697,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(51697,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(51697,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(51697,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(51697,36,'cp 42','a4b5 c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8e7 f3e5 d8c7 c1a3 e8g8 e1g1 e7a3 b1a3 b8d7 a3b5 c7b6 c4d3 d7e5 d4e5 f6d7 b5d6 d7e5 d3h7 g8h7 d1h5 h7g8 h5e5 f7f6 e5g3 f8d8 d6c8 d8c8 f1d1 b6c7 g3g4 f6f5 g4g6 c7e7 g2g3 a7a5');
INSERT INTO evaluations VALUES(51690,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(51690,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(51690,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(51690,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(51690,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51690,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(51690,36,'cp 51','c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8d6 c1a3 e8g8 e1g1 d6a3 a1a3 d8e7 d1b3 c8b7 b1d2 f8c8 f1a1 b7f3 d2f3 b8c6 c4f1 h7h6 h2h3 c8c7 a3a4 a8b8 b3a2 b8b4 a4b4 c6b4 a2b1 b4d5 a1a5 e7b4 b1b4 d5b4 a5b5 b4d5 b5b8 g8h7 f3e5');
INSERT INTO evaluations VALUES(51716,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(51716,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(51716,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(51716,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(51716,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(51716,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(51716,36,'cp 50','b2b3 g8f6 b3c4 b5c4 f1c4 f8d6 c1a3 e8g8 b1d2 a7a5 e1g1 b8c6 a3d6 d8d6 a1a4 c8a6 c4a6 a8a6 d1a1 f8a8 h2h3 d6d5 f1c1 h7h6 c1c5 d5d8 d2b3 f6d7 c5c4 a8b8 b3c5 a6b6 f3d2 d8e7 a1c3 d7f6 a4a1 g8h8');
INSERT INTO evaluations VALUES(51715,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(51715,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(51715,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(51715,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(51715,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(51715,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(51715,36,'cp 48','g8f6 b3c4 b5c4 f1c4 f8d6 c1a3 e8g8 e1g1 d6a3 a1a3 d8e7 d1b3 c8b7 b1d2 f8c8 f1a1 b7f3 d2f3 b8c6 a3a6 g7g6 f3d2 c8c7 c4f1 c6b4 a6a4 b4d5 h2h3 a8c8 b3a3 e7a3 a4a3 c7c2 d2f3 f6e4 f1d3 c2c1 f3e1 e4d2 a3a7 d2b3 a1c1 c8c1');
INSERT INTO evaluations VALUES(11514,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(11514,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(11514,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(11514,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(11514,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(11514,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(11514,36,'cp 19','f8c5 e1g1 d7d6 c3d5 f6d5 c4d5 e8g8 c2c3 c5a7 h2h3 h7h6 a2a4 a6a5 c1e3 a7e3 f2e3 c6b8 d1c2 b8d7 f1f2 d8e7 d5a2 c7c6 a1e1 d7f6 e1f1 c8e6 a2e6 f7e6 f3h4 g8h7 d3d4 f6h5 f2f8 a8f8 f1f8 e7f8 h4f3 h5g3');
INSERT INTO evaluations VALUES(12014,30,'cp -204',NULL);
INSERT INTO evaluations VALUES(12014,31,'cp -196',NULL);
INSERT INTO evaluations VALUES(12014,32,'cp -176',NULL);
INSERT INTO evaluations VALUES(12014,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(12014,34,'cp -202',NULL);
INSERT INTO evaluations VALUES(12014,35,'cp -214',NULL);
INSERT INTO evaluations VALUES(12014,36,'cp -180','d8a5 b1d2 a5e5 f1e2 e5b8 e1g1 e7e5 b2b4 d7d5 a1b1 f8d6 b4b5 c6c5 c2c4 d5d4 e2d3 e8g8 f2f3 c8b7 g2g3 f6h5 d1c2 g7g6 g3g4 h5f4 d3e4 f4e2 g1h1 e2c3 e4b7 b8b7 b1b3 f7f5 g4f5 g6f5 b3c3 d4c3 d2b1 f8f6 b1c3 b7f7 c3d5 f6g6 c1b2 a8e8 b2c3 d6f8 d5e3 e5e4 f3e4 e8e4 e3f5 e4c4');
INSERT INTO evaluations VALUES(706,30,'cp -212',NULL);
INSERT INTO evaluations VALUES(706,31,'cp -221',NULL);
INSERT INTO evaluations VALUES(706,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(706,33,'cp -188',NULL);
INSERT INTO evaluations VALUES(706,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(706,35,'cp -184',NULL);
INSERT INTO evaluations VALUES(706,36,'cp -187','b1d2 a5e5 f1e2 e5b8 e1g1 e7e5 c2c4 f8e7 d2f3 e8g8 d1b3 f8e8 c1e3 f6g4 e3d2 e7c5 d2c3 b8c7 h2h3 g4f6 b3c2 a7a5 a1d1 h7h6 f1e1 d7d6 a2a3 a5a4 e2d3 c8d7 f3h4 c7a7 h4f5 d7f5 d3f5 c5d4 f5e4 d4c3 c2c3 f6e4 e1e4 f7f5 e4e1 e8e6');
INSERT INTO evaluations VALUES(15353,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(15353,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(15353,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(15353,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(15353,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(15353,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(15353,36,'cp -16','f1e2 f8d6 e1g1 e8g8 f1e1 f8e8 e2f1 h7h6 b1d2 c8f5 h2h3 d6f8 a2a4 a7a5 d2c4 d8f6 f3h2 f8c5 d1f3 e8e6 e1e2 a8d8 h2g4 f5g4 f3g4 f6e7 c1d2 d5b6 c4e3 e6g6 g4f5 e7d6 e2e1 c5e3 d2e3 b6d5 e3d2 g6f6 f5g4');
INSERT INTO evaluations VALUES(15352,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(15352,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(15352,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(15352,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(15352,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(15352,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(15352,36,'cp -21','f8d6 e1g1 h7h6 f1e1 e8g8 b1d2 c8f5 e2f1 f8e8 h2h3 a7a5 d2c4 a5a4 c1d2 d5b6 c4d6 d8d6 d2e3 b6d5 e3d2 e8d8 d1e2 f5h7 a2a3 d5b6 f3e5 d6e5 e2e5 c6e5 e1e5 h7d3 f1d3 d8d3 d2e3 b6c4 e5b5 c4e3 f2e3 d3e3 b5b7 e3e2 b7c7');
INSERT INTO evaluations VALUES(20907,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(20907,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(20907,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(20907,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(20907,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(20907,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(20907,36,'cp 119','c2c4 g8f6 g1f3 f8g7 b1d2 e8g8 a1c1 d5c4 f1c4 f6h5 f4g5 h7h6 g5h4 d8e8 e1g1 e7e5 d4d5 c6e7 e3e4 e8d8 c4b3 g6g5 h4g3 h5g3 h2g3 c7c6 f1e1 c8g4 d2f1 c6d5 e4d5 e7f5 f1h2 g4f3 g2f3 f5d6 h2g4 d8e7 g4e3 f8d8 a2a4 e5e4 f3e4');
INSERT INTO evaluations VALUES(26547,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(26547,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(26547,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(26547,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(26547,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26547,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(26547,36,'cp 26','e7e6 e2e3 f8d6 f4g3 g8f6 c2c4 e8g8 b1c3 c6e7 a1c1 c7c6 f1e2 d5c4 e2c4 e7d5 g3d6 d8d6 e1g1 d5c3 c1c3 f8d8 c4e2 h7h6 d1b3 d6e7 c3c1 a7a5 f3e5 f6e4 f1d1 a5a4 b3c4 e4d6 c4c3 f5e4 f2f3 e4d5 e3e4 d5b3 d1d2');
INSERT INTO evaluations VALUES(26821,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(26821,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(26821,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(26821,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(26821,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(26821,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(26821,36,'cp 34','e2e3 f8d6 f4g3 g8f6 c2c4 e8g8 b1c3 c6e7 a1c1 c7c6 f1e2 d5c4 g3d6 d8d6 e2c4 e7d5 e1g1 d5c3 c1c3 f8d8 c4e2 h7h6 d1b3 f5e4 f3e5 d6e7 c3c1 e4d5 b3c3 f6d7 e5d3 e7g5 d3f4 d7f6 f2f3 d8e8 g2g4');
INSERT INTO evaluations VALUES(27188,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(27188,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(27188,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(27188,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(27188,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(27188,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(27188,36,'cp 113','e1g1 f8e7 f1e1 a7a6 b5c6 d7c6 b1c3 f7f6 c3e2 g7g6 e2f4 d8d7 a2a4 e8c8 f4d3 e7d6 c1f4 d8e8 f4d6 e8e1 f3e1 c7d6 d3b4 g8e7 c2c3 h8e8 e1d3 c8b8 b4c6 d7c6 d1f3 c6c4 d3f4 e7f5 f3d5 c4d5 f4d5 e8e2 d5f6 e2b2 f6h7 b2c2');
INSERT INTO evaluations VALUES(27187,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(27187,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(27187,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(27187,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(27187,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(27187,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(27187,36,'cp 105','a7a6 b5c6 d7c6 f1e1 f8e7 b1c3 g8f6 d1e2 c6d7 e2e5 e8f8 c1f4 e7d6 e5e3 h7h6 h2h3 f8g8 f3e5 d7f5 e3f3 f5c2 c3d5 c7c6 d5c3 c2g6 e5g6 f7g6 f4e5 g8h7 e5f6 d8f6 f3f6 g7f6 e1e6 h8d8 e6f6 h7g7 f6f3 d8e8 a2a3 e8f8 f3e3 a8e8 e3e8 f8e8 g1f1 h6h5');
INSERT INTO evaluations VALUES(44310,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(44310,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(44310,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(44310,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(44310,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(44310,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(44310,36,'cp 125','g1f3 f8d6 d2d3 b8c6 f3h4 d6f8 h4f5 d7d6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 f5e3 a8b8 e1g1 b7b5 c4b3 f8e7 g2g3 c8h3 f1e1 e7g5 d1h5 h3e6 f2f4 g7g6 h5d1 e5f4 g3f4 g5f6 c2c3 e6d5 e3d5 e8g8 g1h1 f6g7 d1f3 g8h8 e1g1');
INSERT INTO evaluations VALUES(44309,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(44309,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(44309,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(44309,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(44309,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(44309,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(44309,36,'cp 86','f8d6 d2d3 b8c6 f3h4 d6f8 h4f5 d7d6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 f5e3 a8b8 e1g1 b7b5 c4b3 f8e7 g2g3 c8h3 f1e1 h6h5 c2c3 g7g6 a2a4 b5b4 b3c4 e8f8 c4b5 b8c8 b5c6 c8c6 c3b4 c5b4 d5b4 c6c8 b4d5 e7g5 b2b4 f8g7 b4b5 g5e3 d5e3 h3e6 h2h4 c8c3 d1d2 c3b3 g1h1 g7h7 e3d5');
INSERT INTO evaluations VALUES(51058,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(51058,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(51058,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(51058,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(51058,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(51058,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(51058,36,'cp 95','g1f3 d7d6 d2d4 e5d4 f3d4 g8f6 c1f4 f8e7 d1f3 c8g4 f3g3 d8d7 g3e3 b8c6 h2h3 g4e6 e1c1 c6d4 e3d4 e8g8 g2g4 a8d8 g4g5 f6h5 f4d2 c7c6 f1e2 g7g6 e2h5 g6h5 d1e1 f7f5 e4f5 e6f5 c3e4 d7e6 e4f6 e7f6 g5f6 e6f6');
INSERT INTO evaluations VALUES(51057,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(51057,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(51057,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(51057,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(51057,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(51057,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(51057,36,'cp 100','d7d6 d2d4 e5d4 f3d4 g8f6 c1f4 f8e7 d1f3 c8g4 f3g3 c7c5 d4e2 g4e6 e1c1 b8c6 f4d6 f6h5 g3f3 e7d6 f3h5 e8g8 f2f4 c6d4 f4f5 e6d7 e2d4 c5d4 c3d5 d6e5 c1b1 a8c8 f1d3 d7c6 f5f6 e5f6 e4e5 g7g6 d5f6 d8f6 e5f6 g6h5 b1c1 h5h4 h1g1 g8h8 g1e1');
INSERT INTO evaluations VALUES(52078,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(52078,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(52078,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(52078,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(52078,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(52078,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(52078,36,'cp -20','d4c3 b1c3 b8c6 f1e2 g8f6 e1g1 d8c7 d1a4 d7d6 c1f4 c8d7 a1c1 f8e7 f1d1 c7b8 a4a3 d6d5 f4b8 e7a3 e4d5 e6d5 b2a3 a8b8 c3d5 f6d5 d1d5 d7e6 d5b5 e8g8 e2d3 e6a2 d3e4 f7f6 e4c6 b7c6 b5c5 f8d8 c5c6 d8d7 h2h3 h7h6 c6c7 d7c7 c1c7');
INSERT INTO evaluations VALUES(52085,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(52085,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(52085,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(52085,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(52085,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(52085,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(52085,36,'cp -21','b1c3 b8c6 f1e2 g8f6 e1g1 d8c7 d1a4 d7d6 c1f4 c8d7 a1c1 f8e7 f1d1 c7b8 a4a3 d6d5 f4b8 e7a3 e4d5 e6d5 c3d5 f6d5 b2a3 a8b8 d1d5 d7e6 d5c5 e8e7 e2c4 h8d8 g2g3 e6c4 c1c4 f7f6 c5b5 d8d6 b5b3 b7b6 b3c3 b8c8 g1g2 e7d7 h2h4 g7g6 c4f4 a7a6 h4h5 g6h5 a3a4 c8e8');
INSERT INTO evaluations VALUES(55003,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(55003,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(55003,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(55003,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(55003,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(55003,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(55003,36,'cp 48','d2d4 e7e5 d4e5 d6e5 d1d8 e8d8 b1c3 g8f6 f2f4 b8c6 f1g2 f8d6 g2c6 b7c6 g1f3 f6d7 c1e3 f7f6 e1c1 d8e7 c3e4 a7a5 f3h4 g7g6 h1f1 h8e8 f4e5 d6e5 e3d4 e7f7 h4f3 e5d4 d1d4 f6f5 e4c3 f7g7 f1d1 d7f6 f3g5 e8e7 e2e4');
INSERT INTO evaluations VALUES(54968,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(54968,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(54968,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(54968,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(54968,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(54968,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(54968,36,'cp 33','e7e5 d4e5 d6e5 d1d8 e8d8 b1c3 g8f6 f2f4 b8c6 f1g2 f8d6 g2c6 b7c6 g1f3 f6d7 c1e3 d8e7 e1c1 f7f6 c3e4 a7a5 f4e5 d7e5 f3e5 d6e5 e3d4 c8f5 e4c5 e5d6 e2e4 f5e6 a2a4 h8e8 d1e1 e7f8 c5d3 f8g8 h1f1 e6c4 f1f3 c4d3 c2d3 d6e5 d4e5 e8e5 c1c2 a8d8');
INSERT INTO evaluations VALUES(1601,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(1601,31,'cp -102',NULL);
INSERT INTO evaluations VALUES(1601,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(1601,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(1601,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(1601,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(1601,36,'cp -101','d5d4 d1e2 e7e5 c1e3 d4d6 g1f3 a7a6 a1d1 d6g6 g2g3 g8f6 f1g2 f8d6 h2h3 h7h6 f3h4 g6h7 e1g1 e8g8 f2f4 e5f4 e3f4 d6f4 f1f4 c8e6 c3e4 f6d7 f4f1 a8e8 g3g4 c6e5 e2f2 g8h8 g4g5 e5g6 h4g6 f7g6');
INSERT INTO evaluations VALUES(1633,30,'cp -115',NULL);
INSERT INTO evaluations VALUES(1633,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(1633,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(1633,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(1633,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(1633,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(1633,36,'cp -94','d1e2 e7e5 c1d2 d4g4 e1c1 c8d7 f2f3 g4e6 f3f4 e8c8 g1f3 f8d6 f4e5 c6e5 e2e3 c8b8 f3e5 d6e5 f1c4 e6c4 e3e5 f7f6 e5g3 c4g4 g3g4 d7g4 d1e1 h7h5 h2h3 g4d7 c3d5 d7c6 d5e7 g8e7 e1e7 g7g5 d2c3 h8f8 h1g1 b7b6 e7h7 h5h4');
INSERT INTO evaluations VALUES(10208,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(10208,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(10208,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(10208,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(10208,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(10208,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10208,36,'cp 0','d7d6 d2d3 a7a6 c1e3 c6d4 e3d4 e5d4 c3b1 e8g8 a2a4 c5a7 b1d2 f8e8 f1e1 h7h6 e2f1 f6g4 f1e2');
INSERT INTO evaluations VALUES(9392,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(9392,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9392,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(9392,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(9392,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(9392,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(9392,36,'cp 0','d2d3 a7a6 c1e3 c6d4 a2a4 d4f3 e2f3 c5e3 f2e3 e8g8 d1d2 c8e6 h2h3 h7h6 f1f2 b7b6 a1f1 f6h7 b2b3 d8h4 d3d4 h4g5 g1h1 h7f6 h1h2 b6b5 a4b5 a6b5 c3b5 e5d4 b5d4 g5e5 h2g1 f6e4 f3e4 e5e4 f2f4 e4d5 d2c3 c7c5 d4e6 d5e6 g1h2 f7f6 f1f3');
INSERT INTO evaluations VALUES(9398,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9398,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9398,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9398,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(9398,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(9398,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(9398,36,'cp -5','a7a6 c1e3 c6d4 a2a4 d4f3 e2f3 c5e3 f2e3 c8e6 h2h3 e8g8 d1d2 h7h6 a4a5 b7b5 a5b6 c7b6 a1a3 d8c7 f1a1 c7b7 c3d5 f6d5 e4d5 e6d5 f3d5 b7d5 a3a6 a8a6 a1a6 d5b5 a6a2 f8e8 c2c3 d6d5 a2a1 b5d7 a1f1 e8e6 d2f2 b6b5 f2f5 g7g6 f5f2 g8g7 e3e4 d5e4 d3e4');
INSERT INTO evaluations VALUES(19483,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(19483,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(19483,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19483,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(19483,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(19483,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(19483,36,'cp -18','d7d5 g1f3 e7e6 c2c3 f8e7 f1e2 g8h6 e1g1 e8g8 f3e5 c6a5 c1e3 f7f6 e5g4 h6f7 e4d5 e6d5 f1e1 f6f5 g4f2 c8e6 d3d4 c5d4 e3d4 f7d6 b1d2 a5c6 d2f3 d6e4 f2d3 d8c7 d3e5 c6d4 f3d4 e7c5 e5d3 c5d4 c3d4 c7b6');
INSERT INTO evaluations VALUES(21177,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(21177,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(21177,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(21177,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(21177,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(21177,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(21177,36,'cp 50','g8f6 b1c3 c8f5 a2a3 e7e6 e2e3 f5g6 f1d3 a7a6 d3g6 h7g6 e3e4 d6d5 e4d5 f6d5 c3d5 e6d5 d1e2 f8e7 e1c1 f7f6 e2d3 e8f7 h2h4 e7d6 f4d6 d8d6 g2g3 a8e8 h1e1 d6d7 c1b1 e8e1 d1e1');
INSERT INTO evaluations VALUES(21187,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(21187,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(21187,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(21187,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(21187,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(21187,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(21187,36,'cp 82','e7e5 d2d4 g8f6 f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 d8d7 h2h3 a7a5 c1e3 f8b8 e4e5 f6d5 c3d5 c6d5 f3d5 d7e6 d5f3 d6e5 f4e5 f7f6 e5f6 e7f6 c2c3 b8e8 e3d4 f6d4 c3d4 a8d8 a1c1 e6b6 g1h1 b6d4 f3b3 g8h8 c1c7');
INSERT INTO evaluations VALUES(30830,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(30830,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(30830,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30830,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(30830,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(30830,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(30830,36,'cp 0','e7e6 g1f3 d7d5 e4d5 e6d5 f3e5 g7g6 f4f5 g8e7 f5f6 e7c6 e5c6 b8c6 d1e2 e8d7 e2g4 d7e8 g4e2');
INSERT INTO evaluations VALUES(30700,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30700,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30700,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(30700,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(30700,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(30700,36,'cp 0','g1f3 d7d5 e4d5 e6d5 d2d4 g8f6 f3e5 f8d6 g2g4 e8g8 g4g5 d6e5 d4e5 d5d4 h1g1 f6e4 c3e2 b8c6 f1g2 c6b4 a2a3 b4c2 d1c2 d4d3 c2d1 d3e2 d1e2 d8d5 e2e4 d5e4 g2e4 b7e4 g1g3 c7c5 g3c3 f8c8 c1e3 e4f5 a1d1 g7g6 h2h4 f5g4 d1d5 c8d8 d5d8 a8d8 b2b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(30699,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30699,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30699,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30699,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(30699,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(30699,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30699,36,'cp 0','d7d5 e4d5 e6d5 f3e5 g8f6 d2d4 f8b4 f1d3 e8g8 e1g1 c7c5 c3e2 c5d4 e2d4 b4c5 c1e3 b8c6 c2c3 d8e7 f1e1 c6d4 c3d4 c5b4 e1e2 b4d6 a1c1 f6e4 d1b3 g8h8 e5c6 e7e6 c6e5');
INSERT INTO evaluations VALUES(36524,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(36524,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36524,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36524,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36524,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36524,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(36524,36,'cp 11','g1f3 c7c5 h2h3 c5d4 e3d4 d6d5 f1e2 b8c6 e1g1 d5c4 e2c4 c6a5 b2b3 b7b6 c4e2 c8b7 f3e5 a8c8 a1c1 e7e6 d1d2 a5c6 e5c6 b7c6 e2a6 c8a8 f4e5 d8d7 f1d1 a8d8 a2a4 c6b7 a6b5 d7e7 d2f4 a7a6 b5e2 f8e8 e2f3 b7f3');
INSERT INTO evaluations VALUES(37886,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(37886,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(37886,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(37886,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(37886,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(37886,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(37886,36,'cp 59','d2d4 b8d7 a2a4 g8f6 e1g1 e8g8 h2h3 h7h6 f1e1 f8e8 a4a5 e5d4 f3d4 e7f8 c1f4 d7e5 c4b3 e5d7 d1f3 d7c5 b3c4 c5e6 c4e6 c8e6 a5a6 d8b6 f3d3 e6d7 d4f3 b7a6 f4d6 f8d6 d3d6 b6b2 d6d3 b2b7 e4e5 f6d5 a1b1 b7c7 c3d5');
INSERT INTO evaluations VALUES(37891,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(37891,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(37891,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(37891,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(37891,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(37891,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(37891,36,'cp 49','b8d7 a2a4 g8f6 e1g1 e8g8 f1e1 h7h6 h2h3 f8e8 a4a5 e5d4 f3d4 e7f8 c1f4 d7e5 c4b3 e5d7 d1f3 d7c5 b3a2 c5e6 a2e6 c8e6 a5a6 b7a6 a1a6 e6c8 a6c6 c8b7 c6c4 d6d5 c4c7 d5e4 f3e2 d8d4 c7b7 f6d5 c3d5 d4d5 b7b5 d5e6 b5a5 f8d6 f4d6');
INSERT INTO evaluations VALUES(5074,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(5074,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(5074,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(5074,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(5074,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(5074,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(5074,36,'cp 53','a2a4 g8f6 e1g1 e8g8 h2h3 h7h6 f1e1 f8e8 a4a5 d8c7 c4f1 d7f8 d4d5 c8d7 a5a6 c6d5 e4d5 e8b8 a6b7 b8b7 f3d2 b7b4 c3a2 b4b8 c2c4 a7a5 a2c3 d7f5 e1e3 e7d8 b2b3 f8d7 c1a3 d7c5 a3c5 c7c5 c3a4 c5b4 e3c3');
INSERT INTO evaluations VALUES(6201,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(6201,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(6201,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(6201,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(6201,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(6201,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(6201,36,'cp 76','d8f6 e1g1 d7c6 d1h5 c5b6 c1g5 f6g6 h5g6 f7g6 g5d2 g8f6 b1c3 a7a5 e4e5 f6h7 e5e6 e8e7 a1e1 h8d8 e1e2 a5a4 a2a3 h7f6 c4d3 g6g5 g1h1 d8d6 f1e1 a8a5 h1g1 d6e6 e2e6 c8e6 d3c4 f6d5 c3d5 c6d5 c4d3 e7f7 d2a5 b6a5 c2c3 c7c6 d3c2 b7b5 c2d3 g5g4 f2f4 g4f3 g2f3 a5c7 g1g2');
INSERT INTO evaluations VALUES(54947,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54947,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54947,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54947,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(54947,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(54947,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54947,36,'cp 6','c7c6 b5e2 e7e5 g1f3 g8f6 d2d3 f8e7 e1g1 e8g8 c2c4 c6c5 b1c3 b8c6 f3e1 a7a6 e1c2 c6d4 b2b3 c8d7 f2f4 a8b8 c1d2 b7b5 c2d4 c5d4 c3d5 b5c4 d5e7 d8e7 b3c4 d7c6 d1e1 e7b7 e2f3 f6d7 f4f5');
INSERT INTO evaluations VALUES(52941,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52941,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52941,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52941,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52941,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52941,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52941,36,'cp 0','b5e2 e7e5 g1f3 g8f6 d2d3 c6c5 e1g1 f8e7 c2c4 e8g8 b1c3 b8c6 f3e1 a7a6 e1c2 c6d4 a2a4 c8d7 a1b1 d8a5 b1a1 a5d8');
INSERT INTO evaluations VALUES(839,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(839,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(839,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(839,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(839,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(839,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(839,36,'cp 0','c1b2 c8d7 f1e2 h7h5 e1g1 f8e7 b4b5 c6a5 a3a4 a8c8 h2h3 a7a6 b1a3 a5c4 b2c1 a6b5 a4b5 b6a5 a3c2 a5c3 a1a7 d7b5 e2d3 e8g8 d3f5 e6f5 a7b7 c4e5 b7b5 e5f3 d1f3 c3f3 g2f3 c8c2 b5d5 f8c8 c1e3 g7g6 d5d7 c8c7 d7c7 c2c7 f1a1 c7d7 a1a5 e7d8 a5b5 d8e7');
INSERT INTO evaluations VALUES(838,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(838,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(838,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(838,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(838,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(838,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(838,36,'cp 0','c8d7 f1e2 h7h5 e1g1 f8e7 b4b5 c6a5 a3a4 g7g5 f3e1 a8c8 e1c2 a5b3 a1a2 g5g4 e2d3 b3d4 c2d4 f5d4 b1c3 b6c5 c3e2 d4e2 d1e2 c5b4 a2a1 b4f4 b2c1 f4d4 c1b2 d4b4 f1c1 c8c1 a1c1 e7c5 a4a5 h5h4 b2c3 b4f4 c3d2 f4d4 d2c3');
INSERT INTO evaluations VALUES(27363,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27363,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27363,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27363,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27363,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27363,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27363,36,'cp 0','f1e2 f8e7 e1g1 a8c8 d1d2 e8g8 e2d3 h7h6 d3f5 e6f5 b1c3 d7e6 c3a4 b6b5 a4c3');
INSERT INTO evaluations VALUES(27362,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27362,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27362,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27362,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27362,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27362,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27362,36,'cp 0','a8c8 e1g1 f8e7 d1d2 e8g8 e2d3 h7h6 g1h1 e7g5 f3g5 h6g5 g2g4 f5d4 d2g5 d4f3 g5h5 g7g6 d3g6 f7g6 h5g6 g8h8 g6h6 h8g8');
INSERT INTO evaluations VALUES(6343,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(6343,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(6343,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(6343,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(6343,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(6343,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(6343,36,'cp 77','d4f5 d8f6 b1c3 g8e7 e3b6 a7b6 f5e3 d7d6 f1b5 e8d8 a2a3 h8f8 e1g1 g7g5 d1d2 c8e6 a1d1 c6e5 f2f3 d8c8 b5e2 f8d8 c3b5 c8b8 h2h4 g5h4 f3f4 e5c6 e3d5 e6d5 e4d5 c6a7 d2c3 f6c3 b5c3 e7f5');
INSERT INTO evaluations VALUES(23484,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(23484,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(23484,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(23484,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(23484,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(23484,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(23484,36,'cp 88','e7e6 b1c3 g8f6 d1d4 e5g6 h2h4 e6e5 d4d3 h7h5 c1g5 f8c5 g2g3 d7d6 f1h3 a7a6 e1c1 c8h3 g1h3 b7b5 c1b1 d8d7 d1c1 b5b4 c3d1 d7b5 f2f3 g6e7 h3f2 c7c6 c2c4 b4c3 d1c3 b5d3 f2d3 c5d4 d5c6 e7c6');
INSERT INTO evaluations VALUES(44570,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(44570,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(44570,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(44570,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(44570,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(44570,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(44570,36,'cp -9','b8c6 c2c3 e7e5 c3c4 a8b8 b1c3 d7d6 f1e2 f8e7 e1g1 a7a6 a2a4 e8g8 c1d2 c8d7 f3e1 c6d4 e1c2 b7b5 c4b5 a6b5 a4b5 d7b5 c3b5 b8b5 d2c3 d8d7 c2a3 b5b7 a3c4 f8b8 g2g3 d7e6 e2f3 g7g6 f1e1 g8h8 f3g2 h8g7 a1a6 f6d7');
INSERT INTO evaluations VALUES(55623,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(55623,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(55623,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(55623,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(55623,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55623,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(55623,36,'cp -15','e5f4 d1f3 b8c6 f3f4 c6d4 f1d3 g8f6 f4f2 d4e6 g1f3 f8e7 e1g1 e8g8 d3e2 f6g4 f2e1 g4e5 c3d5 e5f3 e2f3 e7h4 g2g3 c7c6 d5e3 h4f6 d2d3 f8e8 f3g2 e6c7 e1f2 d6d5 e4d5 c7d5 e3d5 c6d5 c1e3 d5d4 e3f4 c8g4 f1e1 d8d7');
INSERT INTO evaluations VALUES(55691,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(55691,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55691,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55691,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55691,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55691,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(55691,36,'cp 0','d1f3 b8c6 f3f4 c6d4 f1d3 d4e6 f4f2 f8e7 g1f3 g8f6 e1g1 e8g8 d3e2 f6g4 f2e1 g4e5 c3d5 e5f3 e2f3 e7h4 g2g3 c7c6 d5e3 h4f6 f3g2 f8e8 d2d3 e6c5 a2a4 a7a5 e1d1 d6d5 d1h5 e8e5 h5d1');
INSERT INTO evaluations VALUES(56448,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(56448,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(56448,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(56448,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(56448,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(56448,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(56448,36,'cp -62','g8f6 d2d4 e5e4 f2f3 e4f3 g1f3 f8e7 e1g1 e8g8 c3c4 d5c4 f3e5 c7c6 b1a3 c8e6 a3c4 f8e8 b2b3 b8d7 e5d7 d8d7 e2e4 a8d8 c1b2 f6g4 d1d3 b7b5 c4e3 g4e3 d3e3 a7a6 a1d1 h7h6 e3f3 e7f8 d4d5');
INSERT INTO evaluations VALUES(48867,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(48867,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(48867,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(48867,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(48867,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(48867,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(48867,36,'cp -54','d2d4 e5e4 f2f3 e4f3 g1f3 f8e7 c3c4 d5c4 b1c3 e8g8 e1g1 b8c6 c1f4 f6d5 c3d5 d8d5 e2e3 c8f5 f3e5 f5e4 g2e4 d5e4 e5c4 g7g5 f4c7 a8c8 c7d6 b7b5 d6e7 e4e7 c4e5 c6e5 d4e5 c8c5 d1d6 c5e5 d6e7 e5e7 f1e1 f8e8 a2a4 b5b4 a1d1 e7e3 e1e3');
INSERT INTO evaluations VALUES(5679,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(5679,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(5679,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(5679,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(5679,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(5679,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(5679,36,'cp 101','c2c4 g8f6 b1d2 e8g8 a1c1 c8f5 a2a3 f6e4 f1e2 e4d2 d1d2 d5c4 c1c4 f5e4 d2c3 d8d7 e1g1 e4d5 c4c5 b7b6 c5d5 d7d5 e2c4 d5f5 c4f7 f8f7 c3c6 a8d8 f3g5 f7f8 g2g3 f5d5 c6e6 d5e6 g5e6 c7c5 e6d8 f8d8 d4c5 b6c5 b2b4 c5b4 a3b4 d8d5 f1c1 a7a5 e3e4 d5b5');
INSERT INTO evaluations VALUES(5675,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(5675,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(5675,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(5675,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(5675,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(5675,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(5675,36,'cp 105','g8f6 b1d2 e8g8 a1c1 c8f5 a2a3 f6e4 f1e2 e4d2 d1d2 d5c4 c1c4 a8c8 d2c3 e7e6 f4g3 c6e7 f3d2 c7c6 c3c1 g6g5 h2h4 h7h6 h4g5 h6g5 e1g1 f5h7 c4c5 e7f5 g3h2 d8e7 b2b4 f8d8');
INSERT INTO evaluations VALUES(10696,30,'cp 282',NULL);
INSERT INTO evaluations VALUES(10696,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(10696,32,'cp 297',NULL);
INSERT INTO evaluations VALUES(10696,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(10696,34,'cp 299',NULL);
INSERT INTO evaluations VALUES(10696,35,'cp 301',NULL);
INSERT INTO evaluations VALUES(10696,36,'cp 325','c1a3 d7d5 c4b5 c3a1 f1e1 c8e6 d1c2 d8d7 f3e5 c6e5 b5d7 e5d7 e1a1 e8c8 a3d6 c7c6 h2h4 d7b6 c2c5 d8d6 c5d6 h8e8 h4h5 b6c4 d6f4 c8d7 f4g5 g7g6 h5h6 c4d6 f2f3 d6f5 a1b1 b7b6 g5f4 e8g8 g2g4 f5d6 b1e1 g6g5');
INSERT INTO evaluations VALUES(10697,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(10697,31,'cp 297',NULL);
INSERT INTO evaluations VALUES(10697,32,'cp 285',NULL);
INSERT INTO evaluations VALUES(10697,33,'cp 282',NULL);
INSERT INTO evaluations VALUES(10697,34,'cp 266',NULL);
INSERT INTO evaluations VALUES(10697,35,'cp 303',NULL);
INSERT INTO evaluations VALUES(10697,36,'cp 288','d7d5 c4b5 c3a1 f1e1 c8e6 d1c2 d8d7 f3e5 c6e5 b5d7 e5d7 e1a1 e8c8 a3d6 c7c6 a1c1 d7b8 d6e5 f7f6 e5g3 b8a6 c2e2 e6d7 e2e7 h8g8 e7d6 d8e8 h2h4 c8d8 g1h2 e8e6 d6a3 d7c8 a3d3 g8h8 a2a4 d8e8 d3f5 e8d8 f5f3 h8e8 f3h5 e8h8 h5d1 c8d7 d1f3 b7b6 f3a3 d8e8 a3d3');
INSERT INTO evaluations VALUES(20808,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(20808,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(20808,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20808,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(20808,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(20808,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(20808,36,'cp 38','h7h6 g5e3 g8f6 f1d3 e7g6 e1g1 f8e7 c2c4 e8g8 b1c3 f6h7 b2b3 a7a6 d3c2 a6a5 a1c1 b7b6 c3b5 c8g4 h2h3 g4d7 a2a4 e7g5 f3g5 h6g5 f2f3 g6f4 d1d2 h7f6 b5c3 f6h5 c3e2 f4e2 d2e2 h5f4 e2d2 g8h8');
INSERT INTO evaluations VALUES(20802,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(20802,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(20802,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(20802,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(20802,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(20802,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(20802,36,'cp 30','g5e3 g8f6 f1d3 e7g6 e1g1 f8e7 c2c4 e8g8 b1c3 f6h7 a1c1 h7g5 f3e1 g5h7 g2g3 c8h3 e1g2 e7g5 e3d2 g5d2 d1d2 d8d7 f2f3 a7a6 g1h1 f7f5 e4f5 h3f5 c1e1 f5d3 d2d3 g6e7 f3f4 e7f5 f4e5 d6e5');
INSERT INTO evaluations VALUES(26512,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26512,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(26512,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(26512,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26512,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(26512,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26512,36,'cp 28','e7e6 g1e2 d8d7 e1g1 h7h5 c1e3 h5h4 h2h3 a7a6 b5d3 g8e7 e3g5 f5d3 d1d3 e7f5 f1e1 f8e7 g5e7 c6e7 a2a4 g7g5 b1d2 h8g8 d2b3 d7c6 c2c3 g5g4 e2f4 g4h3 d3h3 e8c8 b3c5 b7b6 c5d3 a6a5');
INSERT INTO evaluations VALUES(26794,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26794,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(26794,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(26794,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(26794,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(26794,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(26794,36,'cp 41','g1e2 f7f6 e1g1 d8d7 b1d2 a7a6 b5c6 d7c6 c2c3 h7h5 d2f3 f5e4 e5f6 g7f6 f3h4 e8c8 f1e1 f8d6 f2f3 e4h7 e2f4 d6f4 c1f4 g8e7 d1d2 d8g8 g1h1 e7f5 d2f2 c6d7 a2a4 g8g7 e1e2 d7f7 a1e1 h8e8 b2b4 f5h4 f2h4 h7d3 e2b2 e8h8 b4b5');
INSERT INTO evaluations VALUES(27998,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(27998,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(27998,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(27998,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(27998,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(27998,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(27998,36,'cp 129','b2c3 d7d6 a2a4 g8f6 f3d2 b7a6 d3a6 b8a6 f2f4 e8g8 e1g1 a6b8 e4e5 f6d5 d2e4 b8c6 c3c4 d5e7 d4d5 c6a5 d1d3 e6d5 e4g5 e7g6 c4d5 h7h6 g5f7 f8f7 d3g6 d6e5 g6d3 a5b7 f4e5 b7c5 d3d4 f7f1 g1f1 d8d7 c1a3');
INSERT INTO evaluations VALUES(51757,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(51757,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(51757,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(51757,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(51757,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(51757,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(51757,36,'cp 124','c2c3 d7d5 b1d2 b7b6 f1d3 a7a5 e1g1 c8a6 d3a6 a8a6 f1e1 g8f6 e4e5 f6d7 d2f1 b6b5 f1g3 c6c5 g3h5 g7g6 h5g3 d8b6 c1e3 a5a4 h2h4 b8c6 a1c1 a4a3 b2b3 c5c4 b3c4 b5c4 h4h5 g6g5');
INSERT INTO evaluations VALUES(31132,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(31132,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(31132,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(31132,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(31132,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(31132,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31132,36,'cp 29','h2h3 g4h5 b1c3 d5e4 d3e4 d8d1 c3d1 h5f3 g2f3 g7g6 d1c3 b8d7 f3f4 e7e6 h3h4 g8f6 a2a3 h7h5 e1e2 a7a5 c1e3 f6g4 a1d1 g4e3 f2e3 d7c5 e4e5 f8e7 f1g2 a5a4 g2f3 f7f6 e5f6');
INSERT INTO evaluations VALUES(31130,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(31130,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(31130,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31130,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(31130,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31130,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(31130,36,'cp 16','g4h5 b1c3 d5e4 d3e4 d8d1 c3d1 h5f3 g2f3 b8d7 a2a4 g7g6 c1e3 g8f6 h3h4 e7e6 a4a5 f6h5 d1c3 f8c5 c3e2 a7a6 e2d4 c5d6 c2c3 d6f4 h1g1 e8c8 e1c1 d7e5 f1e2 c6c5 d4b3 d8d1 e2d1 e5d7 b3d2 f4c7 d2c4 h5f4 d1b3 h8f8');
INSERT INTO evaluations VALUES(38238,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(38238,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38238,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(38238,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(38238,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(38238,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(38238,36,'cp 41','d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 f1d3 f8e8 e1g1 a7a6 c1f4 c8d7 a2a4 b8c6 d4f3 e7f8 f1e1 h7h6 a4a5 c6b4 d3c4 c7c5 d1d2 d7e6 c4f1 b4c6 f4g3 f6h5 g3h2 h5f6 c3d5 e6d5 e4d5 e8e1 f3e1 c6d4 d2d1 f6d5');
INSERT INTO evaluations VALUES(38231,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(38231,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(38231,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(38231,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38231,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(38231,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(38231,36,'cp 33','e5d4 f3d4 g8f6 b1c3 e8g8 f1d3 a7a6 e1g1 f8e8 f1e1 h7h6 c1f4 c7c5 d4f3 b8c6 d1d2 e7f8 c3d5 b7b5 a2a4 b5a4 a1a4 f6d5 e4d5 c6b4 d3f1 c8b7 c2c4 b7c8 a4a1 c8f5 e1e8 d8e8 a1e1 e8d8');
INSERT INTO evaluations VALUES(38261,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38261,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(38261,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(38261,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38261,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(38261,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(38261,36,'cp 35','f3d4 g8f6 b1c3 e8g8 f1d3 a7a6 e1g1 f8e8 a2a4 c8d7 d3c4 b8c6 d4c6 d7c6 d1f3 g8h8 f1d1 f6d7 f3h5 e8f8 c3d5 d7f6 h5f5 f6d5 c4d5 c6d5 f5d5 a8b8 d5b3 e7f6 c1e3 d8e7 f2f3 c7c6 e3f2 f6e5');
INSERT INTO evaluations VALUES(38253,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(38253,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(38253,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38253,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38253,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38253,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(38253,36,'cp 42','g8f6 b1c3 e8g8 f1d3 f8e8 e1g1 c7c6 f1e1 e7f8 c1f4 a7a5 a2a4 b8d7 d3f1 h7h6 d4f5 d7e5 f5g3 e5g6 f4e3 g6h4 d1e2 d6d5 e4d5 c6d5 c3b5 h4f5 g3f5 c8f5 e2d1 a8c8 c2c3 f5d7 d1f3 f8c5 a1d1 c5e3 e1e3 d8b6');
INSERT INTO evaluations VALUES(48175,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(48175,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(48175,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(48175,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48175,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(48175,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48175,36,'cp -23','e6d5 d2d4 f8b4 f1d3 d5c4 d3c4 e8g8 g1f3 c8g4 e1g1 b8c6 c1e3 d8d6 a1c1 a8e8 c3b5 d6d7 b5c3 a7a6 c4e2 d7d6 d1b3 b4c3 b2c3 b7b5 b3c2 c6a5 a2a4 a5c4 e2c4 b5c4 f3e5 d6e6 c1e1 g4f5 c2e2 f6d7 e5c4');
INSERT INTO evaluations VALUES(48634,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(48634,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(48634,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48634,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(48634,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48634,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(48634,36,'cp -5','d2d4 f8b4 f1d3 d5c4 d3c4 e8g8 g1e2 b8d7 c4b3 d7b6 e1g1 c7c6 c1g5 f8e8 e2f4 d8d6 f4d3 c8g4 d1c1 b6d5 b3d5 f6d5 c3d5 c6d5 g5f4 d6f8 d3b4 f8b4 c1d2 b4d2 f4d2 g4d7 f2f3 f7f6 a1c1 a8c8 g1f2 d7b5 c1c8 e8c8 f1c1 c8e8 c1e1 e8e1 f2e1');
INSERT INTO evaluations VALUES(49542,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49542,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(49542,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(49542,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(49542,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(49542,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(49542,36,'cp -13','f3e5 d7d5 c4b3 f8d6 c1f4 e8g8 e1g1 f8e8 d3d4 c7c5 f1e1 c5d4 c2c3 d4c3 b1c3 b8c6 b3d5 f6d5 d1d5 d6e5 d5d8 e8d8 f4e5 c8e6 f2f3 c6e5 e1e5 g8f8 e5b5 a8b8 a2a4 d8d4 a1d1 d4d1 c3d1 e6d7 b5a5 a7a6 d1c3 d7c6 b2b4 b8c8 b4b5 c6e8 c3e4 a6b5 a4b5 f8e7 a5a7 c8c1 g1f2');
INSERT INTO evaluations VALUES(49488,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(49488,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(49488,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(49488,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49488,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49488,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(49488,36,'cp -19','d7d5 c4b3 f8d6 c1f4 e8g8 e1g1 f8e8 d3d4 c7c5 f1e1 b8c6 e5c6 e8e1 d1e1 b7c6 f4d6 d8d6 d4c5 d6c5 b1c3 c8d7 h2h3 h7h6 e1d2 c5d6 c3e2 d6e5 c2c3 a7a5 b3c2 a8e8 a1e1 e5g5 d2g5 h6g5 b2b4 a5b4 c3b4 g8f8 a2a3');
INSERT INTO evaluations VALUES(48753,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(48753,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48753,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(48753,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(48753,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48753,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48753,36,'cp -24','c4b3 f8d6 c1f4 e8g8 e1g1 f8e8 d3d4 c7c5 f1e1 b8c6 e5c6 e8e1 d1e1 b7c6 f4d6 d8d6 d4c5 d6c5 b1c3 c5d6 a1d1 c8f5 f2f3 f5e6 c3e4 f6e4 f3e4 d5d4 b3e6 f7e6 e1e2 h7h6 b2b3 c6c5 e2d3 d6e5 d1e1 a8f8 h2h3 f8e8 d3c4');
INSERT INTO evaluations VALUES(48760,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(48760,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48760,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(48760,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(48760,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48760,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48760,36,'cp -26','f8d6 c1f4 e8g8 e1g1 f8e8 d3d4 c7c5 f1e1 b8c6 e5c6 e8e1 d1e1 b7c6 f4d6 d8d6 d4c5 d6c5 b1c3 h7h6 e1d2 c5b4 h2h3 a7a5 a2a3 b4h4 c3a4 f6e4 d2e3 c8f5 a1d1 a8e8 a4c5 e4f6 e3f3 e8e5 b3a4 f5g6 f3g3 h4g3 f2g3 g8f8');
INSERT INTO evaluations VALUES(41732,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(41732,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(41732,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(41732,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(41732,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(41732,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(41732,36,'cp -21','c1f4 e8g8 e1g1 f8e8 d3d4 c7c5 f1e1 b8c6 e5c6 e8e1 d1e1 b7c6 f4d6 d8d6 d4c5 d6c5 b1c3 c8e6 h2h3 a8e8 e1d2 h7h6 a1e1 c5a5 e1b1 a5b4 b1d1 a7a5 a2a3 b4b8 d1e1 b8c7 c3e2 e6d7 e2c3 e8d8 d2e3 d7e6 c3a4 d5d4 e3g3 c7g3 f2g3 e6b3');
INSERT INTO evaluations VALUES(55088,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(55088,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(55088,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(55088,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(55088,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(55088,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55088,36,'cp -44','g8f6 g1f3 c7c5 g2g3 e7e6 f1g2 f8e7 e1g1 e8g8 a2a4 b8c6 b1c3 d5d4 c3b5 f6d5 c1d2 a7a6 b5a3 a8b8 a3c4 b7b5 a4b5 a6b5 f3e5 c8d7 e5c6 d7c6 c4e5 c6e8 a1a6 d8c7 d1a1 f7f6 a6a7 c7b6 e5g4 c5c4');
INSERT INTO evaluations VALUES(47242,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(47242,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(47242,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(47242,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(47242,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(47242,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(47242,36,'cp -37','g2g3 e7e6 b1c3 f8e7 f1g2 d5d4 c3e4 f6e4 g2e4 c7c5 e2e3 e8g8 d1e2 d8c7 c1d2 d4e3 e2e3 b8d7 g1f3 a8b8 e1g1 b7b5 a2a4 b5b4 f3e5 d7f6 e4g2 f6d5 e3f2 c8b7 a1e1 e7f6 g3g4 d5e7 e5c4 b7g2 f2g2 f6d4 g1h1');
INSERT INTO evaluations VALUES(15579,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15579,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15579,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(15579,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15579,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15579,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15579,36,'cp 0','d6d5 c1g5 d5e4 d3e4 d8d1 a1d1 f8b4 g5d2 c8g4 a2a3 b4c3 d2c3 f6e4 c3e5 c6e5 f3e5 g4e2 e1e2 e4d6 e2f3 e8c8 h1e1 h8e8 g2g3 f7f6 e5d3 e8e1 d1e1 b7b6 d3f4 d6c4 e1e7 d8d7 e7e8 d7d8');
INSERT INTO evaluations VALUES(36505,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(36505,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36505,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(36505,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(36505,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(36505,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36505,36,'cp 0','f6e4 c4f7 f8f7 c3e4 d6d5 e4g3 b8c6 c2c3 e7e5 d4e5 c6e5 f3e5 g7e5 e1g1 e5g7 c1e3 b7b6 e3d4 c7c5 d4g7 g8g7 f1e1 d8g5 e1e3 c8b7 d1e1 a8d8 a1d1 d8d7 e3e5 g5f6 e5f5 f6d6 f5f7 d7f7 g3e2 b7a6 e2g3');
INSERT INTO evaluations VALUES(36705,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(36705,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36705,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36705,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36705,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36705,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36705,36,'cp 0','c4f7 f8f7 c3e4 d6d5 e4g3 b8c6 c2c3 e7e5 d4e5 c6e5 f3e5 g7e5 e1g1 e5g7 c1e3 b7b6 e3d4 g7f8 d4e5 d8d7 f1e1 c8a6 d1d2 f8c5 e5d4 c5d6 d4e5');
INSERT INTO evaluations VALUES(37679,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37679,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37679,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(37679,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(37679,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37679,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(37679,36,'cp 0','g8f6 f1d3 d7d5 e4d5 e6d5 b1c3 e8g8 e1g1 h7h6 c1f4 b8c6 d1d2 f8e8 f1e1 c8e6 a2a3 a7a6 h2h3 a8c8 a1d1 b7b5 e1e2 c5f8 d1e1 d8b6 f4e3 b6c7 e3f4');
INSERT INTO evaluations VALUES(39904,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39904,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39904,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39904,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39904,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39904,36,'cp 0','f1d3 d7d5 e4d5 e6d5 e1g1 e8g8 b1c3 b8c6 c1g5 h7h6 g5f6 d8f6 c3d5 f6b2 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(47383,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(47383,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(47383,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(47383,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(47383,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(47383,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(47383,36,'cp -57','b1c3 e4f3 g1f3 g7g6 c1f4 f8g7 d1d2 e8g8 e1c1 c7c5 d4d5 b8d7 d5d6 e7d6 f4d6 f8e8 d2f4 f6h5 f4d2 d8a5 d2d5 d7b6 d5c5 g7c3 c5c3 a5c3 b2c3 c8e6 d1d4 e8c8 f1d3 e6a2 h1e1 a2c4 e1e7 c4d3 d4d3 b6c4 e7b7 c4d6 d3d6 c8c3');
INSERT INTO evaluations VALUES(9777,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(9777,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(9777,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(9777,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(9777,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(9777,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(9777,36,'cp 21','c5e3 f2e3 d7d6 a2a4 e8g8 b1d2 c8e6 h2h3 c6e7 f3h4 e6c4 d2c4 d8d7 d1e2 d7e6 f1f2 c7c6 c4d2 d6d5 a1f1 a8d8 e2f3 d5e4 d2e4 f6e4 f3e4 a7a5 b2b3 b7b5 h4f3 f7f6 d3d4 e7d5 d4e5 d5c3 f3d4 c3e4 d4e6 e4f2 e6d8 f8d8');
INSERT INTO evaluations VALUES(9858,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(9858,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9858,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(9858,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(9858,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(9858,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9858,36,'cp 22','f2e3 e8g8 b1d2 d7d6 a2a4 c8e6 h2h3 e6c4 d2c4 d8e7 d1e1 d6d5 e4d5 f6d5 f3h4 g8h7 h4f5 e7d7 e1h4 a8e8 e3e4 d5f4 g1h2 f7f6 a4a5 c6d4 f5d4 d7d4 h4f2 d4f2 f1f2 f4e6 c4e3 c7c6 a1f1 f8f7 b2b4 e8d8 e3f5');
INSERT INTO evaluations VALUES(9856,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(9856,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(9856,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9856,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(9856,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(9856,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(9856,36,'cp 12','d7d6 a2a4 c8e6 b1d2 e8g8 h2h3 e6c4 d2c4 d8e7 d1e1 b7b6 e1g3 g7g6 g3h4 f6h5 h4e7 c6e7 g2g4 h5g7 c4d2 f7f5 g1f2 g7e6 c2c3 f5e4 d2e4 a7a5 f2e2 e6c5 h3h4 c5b3 a1d1 e7d5 f3d2 b3d2 e2d2 f8f1 d1f1 a8f8');
INSERT INTO evaluations VALUES(8891,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(8891,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(8891,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(8891,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(8891,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(8891,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(8891,36,'cp 9','c4b3 e8g8 c2c4 a7a6 b1c3 c8e6 h2h3 a8b8 a2a4 c6b4 a4a5 b7b6 a5b6 b8b6 b3a4 b6b8 f1f2 c7c6 d1e2 f6h5 g1h2 f7f5 f3e5 d6e5 e2h5 d8d3 f2f3 d3c4 e4f5 e6f7 h5g4 c4g4 h3g4 f8d8 h2g3 a6a5 a1c1 b4d3 c1c2 b8b4 a4c6 d3b2');
INSERT INTO evaluations VALUES(16996,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(16996,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(16996,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(16996,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(16996,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(16996,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(16996,36,'cp 146','f3e5 d7d5 e4d5 f8d6 c2c3 d6e5 c3d4 e5d4 d1e2 d8e7 b1c3 e7e2 e1e2 d4e5 f2f4 e5c3 b2c3 e8g8 e2d2 f8d8 h1e1 h7h6 e1e5 f6d7 c1a3 d7e5 f4e5 c7c6 d5d6 d8e8 d3d4 c8e6 c4d3 a8d8 c3c4 f7f6 a1e1 f6e5 d4e5 b7b6 c4c5 e6a2 d2c3 a2d5 d3f5 d5e6 f5e4 e6d7 e4g6 e8e6 g6f5');
INSERT INTO evaluations VALUES(16938,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(16938,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(16938,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(16938,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(16938,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(16938,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(16938,36,'cp 137','d7d5 e4d5 f8d6 c2c3 d6e5 c3d4 e5d4 d1e2 d8e7 b1c3 e7e2 e1e2 e8g8 c1f4 a7a6 a2a4 c7c6 d5c6 b7c6 f4e3 f8e8 e2d2 d4e3 f2e3 c8e6 c4e6 e8e6 h2h3 a8d8 a1d1 e6d6 d2c2 d8b8 e3e4 f6d7 d3d4 f7f6 d4d5 c6d5 d1d5 d6d5 e4d5 g8f7 h1f1 h7h5 f1f4 a6a5 b2b3');
INSERT INTO evaluations VALUES(28489,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(28489,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(28489,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(28489,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(28489,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(28489,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(28489,36,'cp 31','d1b3 a7a5 c1d2 e8g8 a2a3 b4c3 d2c3 b8d7 f1e2 d5c4 b3c4 f5e4 e1g1 e4d5 c4d3 d5f3 e2f3 d7e5 d3d1 e5f3 d1f3 f6d5 c3d2 c7c6 f1e1 d8f6 f3f6 d5f6 e3e4 f8d8 d2c3 f6e8 a3a4 e8d6 b2b3 h7h6 a1c1 f7f5 f2f3 f5e4');
INSERT INTO evaluations VALUES(37615,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(37615,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(37615,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(37615,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(37615,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(37615,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(37615,36,'cp 93','d2d4 c5d4 f3d4 a7a6 c1e3 b7b5 c4b3 c8b7 f2f3 g8f6 d1d2 d7d5 e4d5 b7d5 c3d5 f6d5 e3g5 d5e7 e1c1 b8c6 d4c6 d8d2 d1d2 e7c6 c2c3 e8g8 h1d1 h7h6 g5e3 h6h5 c1b1 c6e5 e3g5 a8c8 a2a3 a6a5 h2h4 a5a4 b3c2 c8b8 c2e4 f8c8 d2d8 c8d8 g5d8 e5c4');
INSERT INTO evaluations VALUES(37627,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(37627,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(37627,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(37627,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(37627,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37627,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(37627,36,'cp 84','c5d4 f3d4 a7a6 c1e3 b7b5 c4b3 c8b7 f2f3 g8f6 d1d2 d7d5 e4d5 b7d5 c3d5 f6d5 e3g5 d5e7 e1c1 b8c6 d4c6 d8d2 d1d2 e7c6 h1d1 e8g8 c2c3 h7h6 g5e3 c6e5 c1b1 a6a5 a2a3 f8c8 d2d6 h6h5 h2h4 a8b8 d6a6 b5b4 c3b4 a5b4 a3a4 e5c4 e3f4');
INSERT INTO evaluations VALUES(40216,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(40216,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(40216,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(40216,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(40216,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(40216,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(40216,36,'cp 44','c4d5 c5d4 f3d4 f6d5 g5e7 d8e7 c3d5 e6d5 e2e3 e8g8 a1c1 b8c6 f1e2 e7f6 e1g1 f8d8 c1c5 c6d4 d1d4 f6d4 e3d4 c8e6 c5c7 d8d7 f1c1 g8f8 f2f3 d7c7 c1c7 a8b8 e2b5 a7a6 b5d7 e6d7 c7d7 b7b6 d7c7 b8e8 g1f2 e8e7 c7c6 e7e6 c6c1 g7g5 c1c7 f8e8 b2b4 e8d8 c7a7');
INSERT INTO evaluations VALUES(42465,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(42465,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(42465,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(42465,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(42465,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(42465,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(42465,36,'cp 103','e4e5 f6d5 c3d5 e6d5 c2c3 b4e7 d2d4 d7d6 f1d3 d6e5 f3e5 b8d7 e1g1 e8g8 c1f4 d7e5 f4e5 c7c6 d1c2 g7g6 a1e1 e7h4 h2h3 c8e6 e1e3 a7a5 a2a4 d8d7 e5f4 f8e8 f1e1 d7d8 e3f3 h4g5 f4g5 d8g5 e1e5 g5h6');
INSERT INTO evaluations VALUES(42437,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(42437,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(42437,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(42437,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(42437,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(42437,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(42437,36,'cp 93','f6d5 c3d5 e6d5 c2c3 b4e7 d2d4 d7d6 f1d3 d6e5 f3e5 e8g8 e1g1 b8d7 e5d7 d8d7 f1e1 c7c6 d1f3 d7d8 c1f4 e7d6 h2h4 d8f6 f4d6 f6d6 e1e5 c8d7 f3g3 g7g6 h4h5 f7f6 e5e3 d6g3 e3g3 g6g5 f2f4 h7h6 f4g5 h6g5 a1e1');
INSERT INTO evaluations VALUES(44337,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(44337,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(44337,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(44337,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(44337,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(44337,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(44337,36,'cp 89','g7g6 f3d2 d7d6 e2e4 f8g7 f1e2 e8g8 e1g1 b8d7 a2a4 f8e8 b1c3 g6g5 f1e1 a7a6 d1c2 d7e5 d2f1 g5g4 f1g3 e5g6 e1f1 c8d7 g1h1 a8b8 a4a5 h7h5 f2f3 b7b5 a5b6 h5h4 g3f5 d7f5 e4f5 g6e5');
INSERT INTO evaluations VALUES(47243,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(47243,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(47243,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(47243,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(47243,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(47243,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47243,36,'cp -28','g1f3 c7c5 b2b3 g7g6 f1b5 c8d7 b5d7 d8d7 c1a3 b7b6 f3e5 d7e6 c2c4 f8g7 d1f3 b8c6 e5c6 e6c6 a3b2 e8g8 b1c3 f8d8 e1c1 c6d7 c3d5 f6d5 f3d5 d7d5 c4d5 d8d5 b2c3 a8c8 c3g7 g8g7 d2d3 c8d8 c1c2 h7h5 h1e1 e7e6 g2g3 b6b5 a2a4 b5a4');
INSERT INTO evaluations VALUES(56429,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(56429,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(56429,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(56429,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(56429,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(56429,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(56429,36,'cp -62','b8c6 d2d3 c8e6 b1d2 f7f6 a2a3 d8d7 b3b4 h7h5 h2h4 g8h6 f1g2 f8e7 g1f3 e8g8 e1g1 c6d8 f3h2 h6g4 c2c4 c7c6 a1c1 d8f7 e2e3 a7a5 h2g4 e6g4 d1b3 a5a4 b3c3 g4f5 c4d5 c6d5 c3c7 d7e6 c7b7 f5d3');
INSERT INTO evaluations VALUES(9034,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(9034,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(9034,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(9034,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(9034,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(9034,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(9034,36,'cp 73','b4d6 d1c2 e8g8 e1g1 h7h6 a1d1 d8e7 a2a3 e6e5 h2h3 f8e8 c4a2 a7a6 a2b1 g7g6 d4d5 e5e4 c3e4 e7e4 c2e4 f6e4 d5c6 b7c6 d2a5 c6c5 f3d2 e4d2 d1d2 a8b8 a5c3 c8e6 f1d1 d6f8 b1e4 f8g7 c3g7 g8g7 d1c1 b8d8 d2d8 e8d8');
INSERT INTO evaluations VALUES(13306,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(13306,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13306,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(13306,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(13306,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13306,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13306,36,'cp 0','g1e2 g7g6 c3d5 e7e5 d5f6 d8f6 d2d3 f8g7 e1g1 e8g8 c2c3 f6d8 c1e3 c8e6 d1d2 a8c8 f2f4 d8d7 a1d1 b7b6 b2b3 d6d5 f4f5 g6f5 e3h6 f5e4 h6g7 g8g7 d2g5 g7h8 g5f6 h8g8 f6g5');
INSERT INTO evaluations VALUES(13307,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13307,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13307,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13307,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13307,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13307,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13307,36,'cp 0','g7g6 c3d5 f8g7 e1g1 e8g8 c2c3 e7e5 d5e3 c8e6 f2f4 e5f4 g3f4 f6g4 d2d4 c5d4 e2d4 c6d4 e3g4 e6g4 d1g4 d4c2 a1b1 d8b6 g1h1 c2a3 b1a1 a3c2 a1b1');
INSERT INTO evaluations VALUES(13015,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13015,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13015,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13015,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13015,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13015,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13015,36,'cp 0','e1g1 f8g7 c3d5 e8g8 c2c3 e7e5 d5f6 d8f6 d2d3 f6d8 c1e3 c8e6 d1d2 a8c8 f2f4 d8d7 a1d1 f8d8 b2b3 d6d5 f4f5 g6f5 e3h6 d5e4 h6g7 g8g7 d2g5 g7h8 g5f6 h8g8 f6g5');
INSERT INTO evaluations VALUES(14626,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(14626,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(14626,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(14626,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(14626,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(14626,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(14626,36,'cp 32','c2c3 f8e7 a2a4 a7a6 b1d2 e8g8 e1g1 c6a5 c4a2 c7c5 d2c4 a5c4 a2c4 c8d7 a4a5 d7c6 f1e1 d8d7 c1e3 a8e8 b2b4 c5b4 c3b4 e8c8 a1c1 f8e8 d1d2 c6b5 e3b6 b5c6 d2c3 e7d8 b6d8 c8d8 c4a6 c6e4 d3e4 b7a6 c3c4 d7b7 f3d2');
INSERT INTO evaluations VALUES(14620,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14620,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(14620,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(14620,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(14620,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(14620,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(14620,36,'cp 23','f8e7 a2a4 a7a6 b1d2 c6a5 c4a2 c7c5 d2c4 a5c4 a2c4 c8d7 a4a5 d7c6 e1g1 e8g8 c4b3 d8d7 c1d2 a8b8 f1e1 f8e8 b3a4 e7d8 d1c2 d6d5 a4c6 d7c6 e4d5 c6d5 c3c4 d5d6 a1b1 b7b5 a5b6 b8b6 d2c3 f6d7 e1e3 f7f5 b1e1 d8f6 b2b3 g8h8');
INSERT INTO evaluations VALUES(8443,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(8443,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(8443,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(8443,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(8443,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(8443,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(8443,36,'cp 30','a2a4 a7a6 b1d2 e8g8 e1g1 c6a5 c4a2 c7c5 d2c4 a5c4 a2c4 c8d7 a4a5 d7c6 f1e1 d8d7 c1e3 a8e8 d1c2 e8d8 b2b4 c5b4 c3b4 d8c8 a1c1 c6b5 c2b1 f8e8 e3b6 e7d8 c4b5 c8c1 e1c1 d7b5 b6e3 d6d5 e3c5');
INSERT INTO evaluations VALUES(16754,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(16754,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(16754,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(16754,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(16754,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(16754,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(16754,36,'cp 224','h2h3 g4f6 b5c6 d7c6 f3e5 c8e6 f1e1 d8d4 e5f3 d4d7 d2d4 e8c8 b1d2 h8g8 c2c3 h7h6 d2f1 c8b8 f1g3 g7g5 b2b4 d7c8 f3e5 f8d6 e5d3 h6h5 e4e5 h5h4 g3f1 g5g4 h3g4 e6g4 d1c2 h4h3 g2g3 g4f3 e5d6 c7d6 c1f4 f6h5');
INSERT INTO evaluations VALUES(16753,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(16753,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(16753,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(16753,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(16753,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(16753,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(16753,36,'cp 224','g4f6 b5c6 d7c6 f3e5 c8e6 f1e1 d8d4 e5f3 d4d7 d2d4 e8c8 b1d2 h8g8 d2f1 h7h6 c2c3 g7g5 f1g3 d7e8 f3h2 e8d7 b2b4 f8e7 h2f3 d7e8 f3e5 e7d6 e5d3 d6e7 b4b5 c6c5 a2a4 h6h5 g3f5 e6f5 e4f5 g5g4 h3h4');
INSERT INTO evaluations VALUES(18778,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(18778,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(18778,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(18778,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18778,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(18778,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(18778,36,'cp 23','g1f3 d7d5 c4d5 e6d5 d2d4 g8f6 f1g2 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 a7a5 c5d4 a5a4 d1d3 f6e4 f1d1 h7h6 f2f3 e4g5 g2f1 g5e6 d4f2 a8b8 d3c3 d8f6 c3f6 g7f6');
INSERT INTO evaluations VALUES(57433,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(57433,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(57433,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(57433,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(57433,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(57433,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(57433,36,'cp 36','b8c6 f1e2 d7d5 e4d5 d8d5 b1c3 d5f7 d3d4 c8d7 d4d5 c6e7 c3b5 e8c8 b5a7 c8b8 d5d6 e7d5 d6c7 d5c7 a7b5 c7b5 e2b5 f8c5 e1g1 g8h6 b5d7 d8d7 d1e2 h6f5 a2a4 h7h5 f1e1 h5h4 h2h3 f7d5 c2c3 h8d8 e2c2 g7g6 b2b4 c5a7 c3c4 d5d3 c2d3 d7d3 a1b1 f5d4 f3d4 a7d4 c1e3 d4e3 e1e3 d3e3 f2e3');
INSERT INTO evaluations VALUES(23240,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(23240,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(23240,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(23240,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(23240,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(23240,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(23240,36,'cp 76','f1e2 f8c5 c1e3 c5e3 f2e3 d7d6 b1c3 c8e6 d1d2 g8h6 d3d4 c6e7 e1c1 c7c6 h2h3 h6f7 g2g4 d8a5 a2a3 d6d5 d4e5 f6e5 e4d5 e7d5 e2c4 e8c8 c3e4 a5b6 h1e1 c8b8 d2e2 h7h6 f3d2 b8a8 e2f3 h8f8 c4b3 f7g5 f3g3');
INSERT INTO evaluations VALUES(4852,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(4852,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(4852,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(4852,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(4852,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(4852,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(4852,36,'cp 175','b5c6 d7c6 f3e5 g8f6 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 f2f4 a8d8 d2b3 c5e3 e4e3 h7h6 b3c5 e6h3 a1d1 b7b6 c5e4 h3e6 d1d8 f8d8 e4f2 e6a2 f4f5 a2d5 e5e6 e7g5 e3g5 h6g5 f2h3');
INSERT INTO evaluations VALUES(4845,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(4845,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(4845,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(4845,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(4845,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(4845,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(4845,36,'cp 162','d7c6 f3e5 g8f6 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 e3c5 e7c5 b1d2 e8g8 e4e3 c5e3 e1e3 a8d8 d2e4 a6a5 f2f4 h7h6 a2a3 a5a4 a1f1 b7b6 g1g2 c6c5 f1e1 d8d5 g2f3 e6d7 h2h3 d5e5 e4f6 g7f6 f4e5 f6e5 f3e2 d7b5 e2d2');
INSERT INTO evaluations VALUES(4788,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(4788,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(4788,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(4788,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(4788,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(4788,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(4788,36,'cp 183','f3e5 g8f6 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 f2f4 c5e3 e4e3 a8d8 d2b3 h7h6 b3c5 e6d5 c5b7 d8b8 c3c4 b8b7 c4d5 b7b2 d5c6 e7b4 e1d1 b4c4 d1d2 b2d2 e3d2 c4c5 d2f2 c5c6 a1e1 h6h5 f2e2 h5h4 e5e6 f8e8 e2e4 c6e4 e1e4');
INSERT INTO evaluations VALUES(4778,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(4778,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(4778,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(4778,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(4778,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(4778,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(4778,36,'cp 182','g8f6 c2c3 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 e8g8 f2f4 a8d8 d2b3 c5e3 e4e3 e6g4 a2a4 d8d5 c3c4 d5d8 b3c5 b7b6 c5e4 e7b4 b2b3 b6b5 a4b5 a6b5 e4c5 b5c4 e5e6 f8e8 b3c4 b4c4 e6e7');
INSERT INTO evaluations VALUES(6222,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6222,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(6222,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(6222,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(6222,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6222,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(6222,36,'cp 30','c5e3 f2e3 d7d6 a2a3 g8f6 b1c3 c8e6 h1f1 a7a6 d1e2 e6c4 d3c4 c6b8 f3d2 b8d7 e1c1 a6a5 f1f5 g7g6 f5f2 e8g8 g2g4 c7c6 h2h4 g8g7 d1g1 d7c5 c1b1 d8e7 e2f3 c5e6 g1f1 f6d7 f3g3 d7c5');
INSERT INTO evaluations VALUES(6272,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(6272,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(6272,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(6272,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(6272,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(6272,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(6272,36,'cp 43','f2e3 d7d6 a2a3 c8e6 b1c3 g8f6 h1f1 e6c4 d3c4 d8e7 c3d5 f6d5 e4d5 c6b8 e3e4 b8d7 d1e2 g7g6 e1c1 e8c8 c1b1 d8f8 g2g3 f7f5 e4f5 f8f5 f3d4 f5f1 d1f1 h8f8 f1f8 e7f8 d4e6 f8f6 b2b3 a7a6 a3a4 d7c5 e6c5 d6c5 a4a5 h6h5');
INSERT INTO evaluations VALUES(6271,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(6271,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(6271,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(6271,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(6271,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6271,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(6271,36,'cp 27','d7d6 a2a3 c8e6 h1f1 g8f6 b1c3 e6c4 d3c4 d8e7 c3d5 f6d5 e4d5 c6b8 e3e4 b8d7 d1e2 e8c8 f3d2 g7g6 b2b4 d8f8 a3a4 f7f5 a4a5 a7a6 e1c1 c8b8 c1b2 f5e4 e2e4 e7g7 b2b3 g6g5 e4d3 f8f1 d1f1 h8f8 f1f8 g7f8 d3g6 f8f2 d2e4 f2g2 g6e8 b8a7 e8d7 g2e4 d7c7');
INSERT INTO evaluations VALUES(5756,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(5756,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(5756,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(5756,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(5756,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(5756,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(5756,36,'cp 23','a2a3 g8f6 b1c3 c8e6 h1f1 e6c4 d3c4 d8e7 c3d5 f6d5 e4d5 c6b8 e3e4 b8d7 d1e2 e8c8 f3d2 d8f8 f1f3 f7f5 e4f5 e7h4 e2f2 h4h2 e1c1 c8b8 d1g1 d7f6 g1e1 f8e8 d2e4 h2h5 b2b3 a7a6 c4c5 d6c5 e4c5 h5g5 c1b1');
INSERT INTO evaluations VALUES(16511,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(16511,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(16511,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(16511,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(16511,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(16511,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(16511,36,'cp -87','f1c4 f8c5 d2d3 h7h6 a2a3 d7d6 b2b4 c5b6 c3a4 b6d4 c2c3 d4g1 h1g1 e8g8 g1f1 a7a6 c4a2 b7b5 a4b2 d6d5 f1f2 a6a5 a3a4 a5b4 a4b5 c6e7 c3b4 c8d7 e1f1 d7b5 f1g1 d8d6 c1d2 c7c6 a2b3 e7g6 a1a8 f8a8 b2a4');
INSERT INTO evaluations VALUES(16384,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(16384,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(16384,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(16384,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(16384,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(16384,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(16384,36,'cp -93','f8c5 d2d3 h7h6 a2a3 d7d6 c3a4 c5d4 c4b5 f6d7 g1e2 d8h4 g2g3 h4h3 c2c3 d4b6 d1c2 f7f5 e4f5 h3f5 a4b6 d7b6 e1g1 f5g6 c1e3 c8f5 c2d2 a7a6 b5c6 b7c6 e3b6 c7b6 d3d4 e8g8 a1d1 f5h3 f1f2 a8d8 d4e5 d6e5 d2d8 f8d8 d1d8 g8h7');
INSERT INTO evaluations VALUES(17370,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(17370,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(17370,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(17370,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(17370,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(17370,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(17370,36,'cp -2','d2d4 c5d4 f3d4 g8f6 b1c3 d8c7 c4e2 c6d4 d1d4 f8d6 c1e3 d6e5 d4d2 e5c3 d2c3 c7c3 b2c3 f6e4 c3c4 b7b5 c4b5 a6b5 e2b5 e4d6 b5d3 c8a6 e3c5 a6d3 c2d3 d6f5 e1d2 d7d6 c5b6 e8d7 a3a4 h8c8 h1b1 f5e7 a4a5');
INSERT INTO evaluations VALUES(17378,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(17378,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(17378,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(17378,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(17378,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17378,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(17378,36,'cp -4','c5d4 f3d4 g8f6 b1c3 d8c7 c4e2 c6d4 d1d4 f8d6 c1e3 d6e5 d4d3 e5c3 b2c3 e8g8 e2f3 d7d5 e4d5 f6d5 f3d5 e6d5 e1g1 c8d7 d3d5 d7c6 d5g5 f7f6 g5g3 c7g3 h2g3 f8e8 f1d1 e8e4 d1d4 e4d4 c3d4 c6d5 a1c1 b7b5 a3a4 g8f7 a4b5 a6b5 f2f3 b5b4');
INSERT INTO evaluations VALUES(30838,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(30838,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30838,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30838,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(30838,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(30838,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(30838,36,'cp 26','g1f3 g8f6 g2g3 e7e6 f1g2 g7g6 e1g1 f8g7 c2c4 e8g8 d2d4 d7d5 c4d5 e6d5 b1c3 f8e8 e2e3 b8d7 f1e1 c7c6 e1e2 a7a5 e2c2 d8e7 a1c1 h7h5 a2a4 f6g4 h2h3 g4h6 c3b1 h6f5 c2c6 b7c6');
INSERT INTO evaluations VALUES(30837,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30837,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(30837,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(30837,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(30837,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(30837,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(30837,36,'cp 18','g8f6 g2g3 e7e6 c2c4 g7g6 d2d4 f8g7 f1g2 e8g8 e1g1 d7d5 b1d2 c7c5 a1c1 b8d7 d4c5 b6c5 e2e3 d8e7 d1c2 a8c8 f3e1 h7h5 e1d3 h5h4 f1e1 h4g3 h2g3 f8e8 e1d1 f6g4 b2g7 g8g7 d2f3 e7f6 c4d5 e6d5');
INSERT INTO evaluations VALUES(28828,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(28828,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(28828,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(28828,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(28828,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28828,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(28828,36,'cp 13','g2g3 c7c5 c2c4 g7g6 b2f6 e7f6 b1c3 b7f3 e2f3 d8e7 f1e2 f6f5 h2h4 h7h5 c3d5 e7d8 e1g1 f8g7 f3f4 b8c6 a1b1 e8g8 e2f3 a8c8 b3b4 c6d4 f1e1 f8e8 e1e8 d8e8 f3g2 d7d6 a2a4 c5b4 b1b4 g8h7 g1h2 e8d7 a4a5 b6a5 d1a4');
INSERT INTO evaluations VALUES(40327,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(40327,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(40327,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40327,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40327,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40327,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(40327,36,'cp 14','d2d4 c5d4 f3d4 b8c6 h2h3 c6d4 d1d4 e8g8 f1d1 c8d7 c4b3 d7c6 d4c4 f6d7 c1e3 d8a5 e3d4 a8c8 c4e2 d7c5 e2e3 a5b6 a1b1 g7d4 e3d4 c5d7 d4d2 g8g7 a2a3 b6c5 g1h1 d7f6 f2f3 a7a5 b3d5');
INSERT INTO evaluations VALUES(40342,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40342,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(40342,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(40342,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(40342,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40342,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(40342,36,'cp 9','c5d4 f3d4 b8c6 c4b3 e8g8 f1e1 c6d4 d1d4 c8d7 d4d3 a8c8 c1e3 d7c6 a1d1 a7a6 f2f3 d8c7 d3d2 f8d8 d2f2 f6d7 h2h3 b7b5 e3d4 g7d4 d1d4 c7b6 a2a3 a6a5 f3f4 e7e5 d4d6 b6f2 g1f2 e5f4 b3d5 c6d5');
INSERT INTO evaluations VALUES(40384,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40384,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(40384,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(40384,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40384,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40384,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40384,36,'cp 12','f3d4 b8c6 c4b3 c6d4 d1d4 e8g8 f1e1 c8d7 d4d3 a8c8 c1e3 d7c6 a1d1 a7a6 f2f3 d8c7 d3d2 f8d8 d2f2 f6d7 e3d4 b7b5 a2a3 a6a5 g1h1 c8b8 d4g7 g8g7 d1d4 d7f6 e1d1 d8c8 f2d2 b5b4 a3b4 b8b4 h2h3 c7b6 c3e2');
INSERT INTO evaluations VALUES(43174,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(43174,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(43174,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(43174,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(43174,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(43174,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(43174,36,'cp -38','c7c5 b1c3 c5d4 d1d4 d7d6 d4d2 e8g8 a1d1 a7a6 f1e2 b8d7 f4g5 h7h6 g5f6 d7f6 g1f3 b7b6 e1g1 c8b7 h2h3 a8c8 d1c1 c8c7 c3d5 f6d5 c4d5 c7c5 e3e4 b6b5 b2b3 c5c1 f1c1 d8b6 e2f1 f8c8 c1c8 b7c8 d2b4 b6c5 b4c5 d6c5 e4e5');
INSERT INTO evaluations VALUES(46298,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(46298,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(46298,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(46298,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(46298,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(46298,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(46298,36,'cp 34','e2e3 d8b6 g5f6 g7f6 a2a3 c8d7 b2b4 f8d6 f1d3 f6f5 g1f3 b6d8 e1g1 e8g8 d1b3 d8e7 f1c1 a7a6 g2g3 d7c6 g1g2 g8g7 a1b1 b8d7 a3a4 d7b6 b4b5 a6b5 d3b5 f8c8 c1c2 b6c4 b5c6 b7c6 c3e2 c6c5 d4c5 c8c5');
INSERT INTO evaluations VALUES(46295,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(46295,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(46295,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(46295,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(46295,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(46295,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(46295,36,'cp 40','d8b6 a1c1 b8c6 g5f6 g7f6 c1c2 c8d7 a2a3 h7h5 g1f3 h5h4 f1d3 f8d6 e1g1 e8f8 d1e2 h4h3 g2g3 c6e7 f3h4 f6f5 f1c1 a8c8 e2d2 f8g7 h4f3 b6d8 c3e2 c8c2 c1c2 d8b6 f3e5 d7b5 d3b5 b6b5 e5d3 b5d7 e2f4 d6f4');
INSERT INTO evaluations VALUES(49069,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(49069,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(49069,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(49069,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(49069,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(49069,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(49069,36,'cp -85','c4d3 d5e4 d3e4 f7f5 e4f3 f8d6 f3d5 b8c6 c2c3 d6e7 d5c6 b7c6 d2d4 e5d4 d1d4 d8d4 c3d4 c8a6 c1e3 e8c8 e1c1 g7g5 g1f3 a6e2 d1e1 e2f3 g2f3 e7f6 c1c2 h7h5 h2h3 c8b7 h1g1 f5f4 e3d2 h8g8 e1e2 f6d4 d2c3 d4c3 b2c3');
INSERT INTO evaluations VALUES(47258,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(47258,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(47258,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(47258,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(47258,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(47258,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(47258,36,'cp -31','c7c5 d2d4 e7e6 d4c5 b8c6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 d1e2 a8c8 f1d1 c6d4 f3d4 c5d4 c1e3 e6b3 d1c1 d8b6 e3d4 b6d4 c3b5 d4b6 c1c8 f8c8 b5c3 c8c5 d3b5');
INSERT INTO evaluations VALUES(54025,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(54025,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(54025,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(54025,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(54025,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(54025,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(54025,36,'cp 115','g1f3 g8f6 f1d3 b8c6 b1c3 f8e7 a2a3 c8d7 e1g1 e8g8 h2h3 a8c8 c1e3 a7a6 f1e1 b7b5 d1e2 f8e8 e4e5 f6d5 c3e4 d5e3 e2e3 d6d5 e4g5 e7g5 f3g5 h7h6 d3h7 g8f8 g5f3 b5b4 h3h4 b4a3 a1a3 a6a5 e3f4 c6e7 h4h5 d8b6');
INSERT INTO evaluations VALUES(7602,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(7602,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(7602,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(7602,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(7602,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(7602,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(7602,36,'cp 75','e4e5 f6d5 d1g4 e8f8 e1g1 d7d6 g4h5 d6e5 h5e5 c5d6 e5h5 h7h6 f1e1 c8e6 h5d1 d8h4 g2g3 h4g4 d1g4 e6g4 b1a3 d5b4 d3e4 f7f5 e4g2 f8f7 b2b3 h8e8 c1b2 e8e2 e1e2 g4e2');
INSERT INTO evaluations VALUES(9289,30,'cp 432',NULL);
INSERT INTO evaluations VALUES(9289,31,'cp 426',NULL);
INSERT INTO evaluations VALUES(9289,32,'cp 427',NULL);
INSERT INTO evaluations VALUES(9289,33,'cp 431',NULL);
INSERT INTO evaluations VALUES(9289,34,'cp 440',NULL);
INSERT INTO evaluations VALUES(9289,35,'cp 436',NULL);
INSERT INTO evaluations VALUES(9289,36,'cp 429','d2d4 c5b6 d4d5 a7a6 b5a4 f6e4 d5c6 e8g8 b1d2 e4c5 c6b7 c8b7 a4c2 f7f5 d2c4 d8f6 f3g5 d6d5 c4b6 c7b6 b2b4 c5e4 g5e4 d5e4 d1d7 b7c6 c2b3 g8h8 d7e6 f6e6 b3e6 f8f6 e6b3 f6g6 b3c2 a8c8 f1e1 h7h5 h2h4 e4e3 g2g3 c6e4 c2e4 e3f2 g1f2 f5e4 c1g5 g6c6 e1e4 c6c3');
INSERT INTO evaluations VALUES(9282,30,'cp 431',NULL);
INSERT INTO evaluations VALUES(9282,31,'cp 428',NULL);
INSERT INTO evaluations VALUES(9282,32,'cp 436',NULL);
INSERT INTO evaluations VALUES(9282,33,'cp 428',NULL);
INSERT INTO evaluations VALUES(9282,34,'cp 425',NULL);
INSERT INTO evaluations VALUES(9282,35,'cp 431',NULL);
INSERT INTO evaluations VALUES(9282,36,'cp 405','c5b6 d4d5 a7a6 b5a4 f6e4 d5c6 e8g8 b1d2 e4c5 c6b7 c8b7 a4c2 f7f5 d2c4 d8f6 c4b6 c7b6 f3g5 d6d5 b2b4 c5e4 g5e4 d5e4 d1d7 b7c6 c2b3 g8h8 d7e6 f5f4 e6f6 g7f6 g2g3 a6a5 b4a5 b6a5 c1a3 a5a4 b3c2 f8c8 a1d1 e4e3 f2e3');
INSERT INTO evaluations VALUES(10231,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(10231,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(10231,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(10231,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(10231,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(10231,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(10231,36,'cp -30','d3c2 d7d5 e4d5 d8d5 d1e2 c8e6 d2d3 e8c8 b1d2 a7a5 e1g1 c5a7 h2h3 h7h6 f1e1 h8e8 d2e4 f6e4 d3e4 d5c4 e2c4 e6c4 f3d2 c4e6 d2f1 d8d7 f1e3 e8d8 e1d1 b7b5 d1d7 d8d7 c2b3 c6d8 g1f1 a5a4 b3e6 d8e6');
INSERT INTO evaluations VALUES(19765,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(19765,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(19765,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(19765,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(19765,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(19765,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(19765,36,'cp 206','d1d8 c6d8 b1c3 d8e6 c1e3 g7g6 e1c1 f8g7 f1b5 e8f8 b5c4 g7c3 b2c3 h7h5 f3e5 g8f6 f2f3 f8g7 h2h4 b7b6 h1e1 a7a6 a2a4 a8a7 a4a5 b6a5 e5c6 a7c7 c6a5 h8d8 c4d5 f6e8 e4e5 g7f8 g2g4 f8g8 c3c4 e8g7 a5c6 d8d7 f3f4 c8b7');
INSERT INTO evaluations VALUES(2258,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(2258,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(2258,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(2258,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(2258,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(2258,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(2258,36,'cp 202','c6d8 c1e3 d8e6 b1c3 g7g6 e1c1 f8g7 f1b5 e8f8 b5c4 g7c3 b2c3 f8g7 f3e5 g8f6 f2f3 a8b8 a2a4 b7b6 h2h4 h7h6 e3g1 h8d8 g1h2 c8b7 d1d3 b8c8 h1d1 d8d3 c2d3 e6d8 a4a5 f6e8 c4b5 e8d6 b5d7 c8c7 h2g3 d8c6 e5c6 b7c6');
INSERT INTO evaluations VALUES(2382,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(2382,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(2382,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(2382,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(2382,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(2382,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(2382,36,'cp 184','b1c3 d8e6 c1e3 g7g6 e1c1 f8g7 f1b5 e8f8 b5c4 g7c3 b2c3 g8f6 f3e5 h7h5 f2f3 f8g7 h2h4 b7b6 h1e1 a7a6 a2a4 a8a7 a4a5 b6a5 e5c6 a7c7 c6a5 h8e8 e3f2 c7d7 a5b3 e6f4 c4f1 f4e6 b3c5 e6c5 f2c5 d7c7 c5b4 c8e6 d1d4 e8b8 e1d1 b8b7 g2g3');
INSERT INTO evaluations VALUES(24136,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(24136,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(24136,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(24136,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(24136,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(24136,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(24136,36,'cp 54','h2h3 g4h5 b1d2 c5b6 a2a4 a7a5 c4a2 b6a7 d2c4 h7h6 f1e1 f6h7 g2g4 h5g6 g1g2 h7f6 a1b1 d8d7 f3h4 g6h7 d1f3 e8c8 h4f5 h7f5 f3f5 d7f5 e4f5 h8e8 f2f4 a7c5 c1d2 e5f4 d2f4 d6d5 c4e5 c6e5 e1e5 e8e5 f4e5 c7c6 d3d4 c5d6');
INSERT INTO evaluations VALUES(24134,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(24134,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(24134,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(24134,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(24134,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(24134,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(24134,36,'cp 62','g4e6 c4e6 f7e6 b2b4 c5b6 b1d2 a7a6 a2a4 e8g8 a1b1 h7h6 d1b3 d8d7 d2c4 d7f7 c4b6 c7b6 c1e3 b6b5 b1a1 f6h5 a4b5 a6b5 f1b1 d6d5 b1e1 a8a1 e1a1 d5e4 d3e4 f7g6 b3d1 h5f6 e3c5 f8d8 d1e2 f6e4 e2b5 e4c5 b4c5 e5e4 f3d4 c6d4 c3d4');
INSERT INTO evaluations VALUES(32862,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(32862,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(32862,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(32862,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(32862,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(32862,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(32862,36,'cp -64','g4f3 g2f3 b8c6 d2d3 c6d4 c3e4 e7e6 c2c3 d4e2 d1e2 e8c8 c1e3 e5d5 d3d4 h7h6 b2b4 f7f5 e4c5 f8c5 d4c5 g8e7 a1d1 d5c6 b4b5 d8d1 e1d1 c6d5 d1c1 h8d8 h1d1 d5d1 e2d1 d8d1 c1d1 e6e5 c3c4 c7c6 f3f4 e5f4 e3f4 g7g5 f4d6 c8d7 d1d2 e7g6 d2d3 a7a6 b5b6 d7e6 d3e3 h6h5 d6c7 g6e5');
INSERT INTO evaluations VALUES(32870,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(32870,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(32870,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(32870,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(32870,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(32870,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(32870,36,'cp -55','g2f3 b8c6 d2d3 e7e6 c3e4 c6d4 c2c3 d4e2 d1e2 e5d5 d3d4 e8c8 a2a3 g8e7 c1e3 e7c6 e1c1 d5a2 e2c2 c6a5 e4d2 a5c4 c2b1 a2b1 d2b1 f8e7 b1d2 c4d2 c1d2 d8d5 d2e2 d5h5 a3a4 c8d7 b2b4 e7d6 h2h3 h8g8 d1b1');
INSERT INTO evaluations VALUES(32869,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(32869,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(32869,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(32869,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(32869,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(32869,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(32869,36,'cp -50','b8c6 d2d3 e7e6 c3e4 c6d4 c2c3 d4e2 d1e2 e5d5 d3d4 e8c8 a2a3 g8e7 c1e3 e7f5 e1c1 f8e7 c1b1 f5d6 e4d2 e7f6 f3f4 d5f5 b1a2 a7a6 h1e1 f5d5 a2b1 h7h6 b1a1 d8d7 a1b1 d5f5 b1c1 h8d8 d2f3 f5h5 f3e5 h5e2 e1e2');
INSERT INTO evaluations VALUES(27624,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(27624,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(27624,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(27624,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(27624,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(27624,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(27624,36,'cp -38','d2d3 c6d4 c3e4 e7e6 c2c3 d4e2 d1e2 e8c8 c1e3 e5d5 d3d4 g8f6 e4d2 f6h5 d2c4 f8d6 e1c1 h5f4 e3f4 d6f4 c1c2 g7g6 b2b3 c7c5 d4c5 d5c5 b3b4 c5c7 h2h4 d8d5 a2a4 h8d8 a4a5 d5d1 h1d1 d8d1 c2d1 c7d8 d1c2');
INSERT INTO evaluations VALUES(27625,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(27625,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(27625,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(27625,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(27625,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(27625,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(27625,36,'cp -43','c6d4 c3e4 e8c8 c1e3 e5a5 c2c3 d4e2 e1e2 a5a6 e4c5 a6c6 d1a4 c6a4 c5a4 e7e6 d3d4 g8f6 b2b4 f6d5 b4b5 g7g6 c3c4 d5e7 a1d1 e7f5 a4c5 f8c5 d4c5 h8e8 a2a4 e6e5 d1d8 e8d8 h1d1 d8d1 e2d1 f5e3 f2e3 c8d7 d1d2 d7e6 d2c3 f7f5 c3b4');
INSERT INTO evaluations VALUES(34242,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(34242,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(34242,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(34242,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(34242,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(34242,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(34242,36,'cp 107','c1d2 a5b6 g1f3 g8f6 e4e5 f6e4 d2c1 d7d5 e5d6 b6a5 c2c3 e4d6 c1f4 c8f5 f1e2 b8d7 e1g1 e7e6 b1d2 f8e7 a2a4 e8g8 f1e1 f5g6 h2h3 a5f5 f4h2 f5c2 f3e5 d7e5 d4e5 c2d1 e1d1 d6c8 a4a5 f8d8 d2c4 b7b5 d1d8 e7d8 a1d1 d8e7');
INSERT INTO evaluations VALUES(34243,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(34243,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(34243,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(34243,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(34243,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(34243,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(34243,36,'cp 115','a5b6 g1f3 g8f6 e4e5 f6e4 d2c1 d7d5 e5d6 b6a5 c2c3 e4d6 c1f4 c8f5 f1e2 e7e6 e1g1 f8e7 c3c4 b8d7 f1e1 f5g4 b1d2 d6f5 d2b3 a5b4 f4d2 b4d6 d4d5 e8g8 d5e6 f7e6 d2c3 e7f6 d1d6 f5d6 c3f6 g7f6 a1d1');
INSERT INTO evaluations VALUES(34234,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(34234,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(34234,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(34234,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(34234,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(34234,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(34234,36,'cp 114','g1f3 g8f6 e4e5 f6e4 d2c1 d7d5 e5d6 b6a5 c2c3 e4d6 c1f4 c8f5 f1e2 e7e6 e1g1 b8d7 f1e1 f8e7 c3c4 f5g4 b1d2 d6f5 d2b3 a5d8 d4d5 e8g8 d5c6 b7c6 h2h3 g4f3 e2f3 a8c8 e1e2 d7f6 e2d2 d8b6 d1e2 f5h4 a1d1 h4f3 e2f3 f8d8 d2d8 c8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(47811,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(47811,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(47811,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(47811,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(47811,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(47811,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(47811,36,'cp -19','b8c6 g1e2 c8e6 d2d3 d6d5 e4d5 e6d5 c4d5 f6d5 a2a3 d8d7 c1d2 g7g6 b1c3 d5c3 e2c3 e8c8 e1c1 f7f5 d2g5 f8e7 g5e7 d7e7 f3e3 h8e8 c1b1 a7a6 g2g4 c6d4 h1e1 e7f6 g4f5 g6f5 f2f4 e5e4 d3e4 f5e4 e3g3 d4f5 d1d8 f6d8');
INSERT INTO evaluations VALUES(15505,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(15505,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(15505,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(15505,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(15505,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(15505,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(15505,36,'cp -25','g1e2 c8e6 d2d3 d6d5 e4d5 e6d5 c4d5 f6d5 a2a3 d8d7 c1d2 c6d4 e2d4 e5d4 e1g1 e8c8 f1e1 f7f6 e1e4 f8c5 b2b4 c5d6 c2c4 d4c3 b1c3 f6f5 e4e2 d5c3 d2c3 h8e8 a1e1 e8e2 f3e2 b7b6 e2e3 c8b7 g1f1 d6f8 d3d4 g7g6 e3e6 d7e6 e1e6');
INSERT INTO evaluations VALUES(50343,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50343,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(50343,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(50343,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(50343,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(50343,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(50343,36,'cp -10','d7d5 g1f3 c8f5 f3h4 f5d7 e2e3 c7c5 c2c3 b8c6 h4f3 c5d4 e3d4 d8c7 h2h3 e7e6 f1d3 f8d6 e1g1 e8g8 f1e1 f8e8 d2f1 h7h6 c1d2 a8d8 a2a4 a7a5 d3c2 b7b6 a1c1 e6e5 f3e5 c6e5 d4e5 d6e5 d2e3 e5d6 e3d4 e8e1 d1e1 d8e8 f1e3');
INSERT INTO evaluations VALUES(47209,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(47209,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47209,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47209,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47209,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47209,36,'cp 0','g1f3 c8f5 f3h4 f5d7 e2e3 c7c5 c2c3 b8c6 h4f3 c5d4 e3d4 d8c7 h2h3 d7f5 f3h4 f5d7');
INSERT INTO evaluations VALUES(53602,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(53602,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(53602,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(53602,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(53602,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53602,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(53602,36,'cp 37','f1d3 b8d7 e1g1 f8d6 c2c4 g8f6 b1c3 e8g8 e3e4 d5c4 d3c4 e6e5 c1e3 d8e7 c4b3 e5d4 e3d4 f8d8 d1e2 d7c5 d4c5 d6c5 e4e5 f6d5 c3d5 c6d5 a1d1 c8e6 d1d3 d5d4 e2e4 e6b3 d3b3 b7b6 b3d3 e7e6 f3g5 e6g6 e4g6 h7g6 e5e6 f7e6 g5e6 d8d7 e6f4 a8e8');
INSERT INTO evaluations VALUES(53601,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(53601,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(53601,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(53601,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(53601,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(53601,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(53601,36,'cp 34','b8d7 b1c3 f7f5 c3e2 g8h6 c2c4 f8d6 c4c5 d6c7 b2b4 e8g8 e1g1 e6e5 d4e5 d7e5 e2d4 h6g4 c1b2 d8f6 f3e5 c7e5 h2h3 e5h2 g1h1 h2e5 a1b1 f5f4 e3f4 f6f4 d4f3 e5b2 b1b2 g4f6 h1g1 c8d7 f1e1 a8e8 a2a3 e8e1 d1e1 f8e8 e1d2 f4d2 b2d2 a7a5 b4a5');
INSERT INTO evaluations VALUES(11061,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(11061,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(11061,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(11061,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(11061,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11061,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(11061,36,'cp 46','f1c4 e7e6 d1e2 b7b5 c4b3 d8c7 a2a3 a8b8 f2f4 f8e7 b3e6 f7e6 d4e6 c7c4 e6g7 e8f7 e2c4 b5c4 g7f5 h8e8 e1c1 e7f8 g5h6 f6e4 c3e4 e8e4 h6f8 f7f8 f5d6 e4e7 d6c4 c8b7 d1d2 d7f6 c4e5 e7g7 h1g1 b8e8 g2g3 f6g4 e5d7 f8g8 h2h3 b7c8 g1d1 c8d7 d2d7 g7d7 d1d7 g4f6 d7d6 f6e4 d6a6');
INSERT INTO evaluations VALUES(11060,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11060,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(11060,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11060,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11060,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(11060,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(11060,36,'cp 44','d8b6 c4b3 e7e6 g5f6 d7f6 d1d2 g7g6 e1c1 f8g7 c1b1 e8g8 f2f4 c8d7 h2h3 a8d8 g2g4 e6e5 d4e2 e5f4 d2f4 d7e6 d1d3 b6c5 g4g5 f6h5 f4d2 c5e5 h1d1 e5h2 a2a3 e6b3 c2b3 f8e8 b3b4 g7c3 e2c3 h2d2 d3d2 h5g3');
INSERT INTO evaluations VALUES(17199,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(17199,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(17199,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(17199,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(17199,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(17199,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(17199,36,'cp 191','f3e5 f6e4 d1e2 f7f5 b1d2 e4c5 d2b3 d7d6 e5f3 c5e4 f3g5 g7g6 g5e4 f5e4 e2e4 f8g7 f1b5 c8d7 b3d4 e8g8 d4e6 d7e6 e4e6 f8f7 e1g1 e7f5 c2c3 d8f8 e6e4 f7e7 e4a4 a7a6 b5d3 g7h6 a4b3 h6c1 a1c1 b7b5 a2a4 b5a4 b3a4 a8b8 c1e1 b8b2 e1e7 f5e7 a4c4 b2d2 h2h4 e7f5 f1e1');
INSERT INTO evaluations VALUES(17192,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(17192,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(17192,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(17192,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(17192,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(17192,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(17192,36,'cp 173','f6e4 d1e2 f7f5 b1d2 d7d6 e2b5 c7c6 d5c6 b7c6 e5c6 a7a6 b5c4 e7c6 c4c6 c8d7 c6d5 e4f6 d5d4 d6d5 f1d3 f8d6 e1g1 e8g8 d2f3 d8c7 c1e3 f6g4 h2h3 g4e3 d4e3 a8e8 e3d4 d6c5 d4c3 a6a5 a1e1 e8e4 c3d2 g8h8 c2c3 c7d6 e1d1 e4f4 d3e2 d7c6 f3d4 f4e4 d4c6 d6c6 e2f3');
INSERT INTO evaluations VALUES(17205,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(17205,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(17205,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(17205,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(17205,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(17205,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(17205,36,'cp 138','b1c3 e7c6 d4e3 f8b4 c1d2 d7d6 a2a3 b4c5 e3g3 e8g8 e1c1 g8h8 f2f4 a8b8 c3d5 f7f6 d2c3 a7a5 f1b5 c6a7 b5d3 c7c6 d5e3 a7b5 d3b5 c6b5 e3d5 c8e6 b2b4 a5b4 a3b4 c5b6 h1e1 e6g8 d5b6 d8b6 c3b2');
INSERT INTO evaluations VALUES(23054,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(23054,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(23054,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(23054,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(23054,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(23054,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(23054,36,'cp 145','c7c6 c2c3 d4b5 g1f3 d7d6 c3c4 b5c7 b1c3 g8f6 h2h3 f8e7 f1d3 a7a5 e1g1 c7a6 a2a3 c8d7 d5c6 d7c6 a1b1 e8g8 b2b4 a5b4 a3b4 a6c7 f1e1 c7e6 c3d5 f8e8 b4b5 c6d7 d1c2 e7f8 e3b6 d8c8 d5f6 g7f6 b6e3 g8h8 d3f1 c8c7 b1d1');
INSERT INTO evaluations VALUES(26013,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(26013,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(26013,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(26013,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(26013,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26013,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(26013,36,'cp 9','b7c6 b5c4 g8f6 b1c3 e7e6 e1g1 f8e7 c1f4 e6e5 f4g3 e8g8 h2h3 d7e6 c4b3 d8b6 d1e2 a8d8 f1d1 e6c8 a1b1 g8h8 b3c4 b6b7 b2b3 a7a5 g1h1 d6d5 c4d3 f6d7 c3a4 f7f5 e4d5 c6d5 g3e5 d7e5 e2e5');
INSERT INTO evaluations VALUES(25832,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(25832,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(25832,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(25832,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(25832,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(25832,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(25832,36,'cp 13','b5c4 g8f6 b1c3 e7e6 e1g1 f8e7 d1e2 e8g8 b2b3 d8c7 c1b2 f8e8 f1e1 d6d5 e4e5 d5c4 e5f6 e7f6 e2c4 c6c5 c3a4 f6e7 h2h3 a8d8 a1d1 e7f8 b2e5 c7a5 e5c3 a5c7 a4b2 d7c6 c4g4 f7f5 g4g5 h7h6 g5g6 d8d1 e1d1');
INSERT INTO evaluations VALUES(25833,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(25833,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(25833,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(25833,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(25833,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(25833,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(25833,36,'cp 3','g8f6 b1c3 e7e6 e1g1 f8e7 c1f4 e6e5 f4g3 d8b6 b2b3 a8d8 h2h3 e8g8 d1e1 g8h8 a1d1 f6h5 g3h2 d7c8 g1h1 b6c7 e1e3 c7b6 e3b6 a7b6 a2a4 f7f5 e4f5 f8f5 c4d3 f5f8 c3e2 h5f6 c2c4 e5e4 d3c2 d6d5 c4d5 c6d5 e2d4');
INSERT INTO evaluations VALUES(24394,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(24394,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24394,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(24394,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(24394,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(24394,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(24394,36,'cp 16','b1c3 e7e6 e1g1 f8e7 h2h3 e8g8 c1f4 d6d5 c4d3 d7e8 f1e1 f6d7 d1g4 g8h8 e4d5 c6d5 g4g3 a8c8 c3b5 e6e5 f4e5 d7e5 g3e5 e7f6 e5e2 e8b5 d3b5 f6b2 a1b1 b2f6 b5d3 g7g6 b1b7 c8c7 b7c7 d8c7');
INSERT INTO evaluations VALUES(24393,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24393,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(24393,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(24393,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24393,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(24393,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24393,36,'cp 14','e7e6 e1g1 f8e7 h2h3 e8g8 c1f4 d6d5 c4d3 d7e8 d1e2 f6d7 f1e1 d7c5 a1d1 e7f6 b2b4 c5d3 e2d3 a7a5 b4b5 e6e5 e4d5 e5f4 b5c6 e8c6 d5c6 d8d3 d1d3 f8c8 c3d5 c8c6 d5f6 g7f6 e1e4 c6c2 a2a4 c2b2 e4f4 b2b4 d3f3 b4f4');
INSERT INTO evaluations VALUES(32643,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(32643,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32643,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(32643,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(32643,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(32643,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(32643,36,'cp 51','h2h3 g4h5 d2d3 e7e6 g2g4 h5g6 c1d2 a5c5 d1e2 b8c6 a2a3 a7a6 d2e3 c5e7 e1c1 e8c8 d3d4 f6d5 c4d5 e6d5 f3e5 c6e5 d4e5 f7f6 d1d5 f6e5 h1d1 d8d5 c3d5 e7e6 e2c4 e6c6 c4c6 b7c6 d5b4 f8b4 a3b4 h7h5 f2f3 c8b7 d1d7 h5g4');
INSERT INTO evaluations VALUES(32641,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(32641,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(32641,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(32641,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(32641,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(32641,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(32641,36,'cp 60','g4h5 d2d3 c7c6 c1d2 b8d7 d1e2 h5f3 e2f3 a5e5 c3e4 e7e6 e1c1 f6e4 d3e4 d7c5 h1e1 a8d8 c1b1 f8e7 d2c3 d8d1 e1d1 e5g5 g2g3 g5g6 f3e2 h7h5 c4d3 e8g8 c3b4 f8e8 a2a3 g6g5 b4c5 e7c5 f2f4 g5g3 e2h5 c5e3 e4e5 g3f4 h5h7 g8f8');
INSERT INTO evaluations VALUES(32636,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(32636,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(32636,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(32636,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(32636,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(32636,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(32636,36,'cp 91','d2d3 c7c6 d1e2 h5f3 e2f3 e7e6 c1d2 a5e5 c3e2 b8d7 e1c1 e5c7 d3d4 f6d5 h3h4 b7b5 c4d3 d7b6 c1b1 a7a5 c2c3 f8d6 g2g4 b6c4 d2c1 a5a4 a2a3 e8c8 g4g5 c4a5 h4h5 g7g6 f3e4 c8b8 e4h4 a5b3 h5h6 h8g8 d1e1');
INSERT INTO evaluations VALUES(38818,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(38818,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(38818,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(38818,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(38818,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(38818,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(38818,36,'cp 82','g8e7 e2e3 e8g8 f1d3 d5c4 d3c4 b7b6 e1g1 c7c5 e3e4 c8a6 c4a6 b8a6 d1b3 d8c7 d4d5 c5c4 b3a4 a6c5 a4c4 f8d8 d5e6 c5e6 c4c7 e6c7 f3d4 c7e6 c1e3 e6c5 f2f3 c5a4 f1c1 a8c8 c3c4 a4b2 d4b5 d8d3 b5a7 d3e3 a7c8 e7c8');
INSERT INTO evaluations VALUES(53166,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(53166,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(53166,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(53166,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53166,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53166,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53166,36,'cp 0','e7e6 e2e4 b8c6 e4d5 e6d5 b1c3 g8f6 c3b5 f8c5 b5c7 e8f8 c7a8 f6e4 f4e3 d5d4 e3d2 e4f2 d1h5 d8e7 e1f2 d4d3 f2g3 c5d6 g3f2');
INSERT INTO evaluations VALUES(54067,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54067,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54067,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54067,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54067,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54067,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54067,36,'cp 0','e2e4 b8c6 e4d5 e6d5 b1c3 g8f6 c3b5 f8c5 b5c7 e8f8 c7a8 f6e4 f4e3 d5d4 e3d2 e4f2 d1h5 d8e7 e1f2 d4d3 f2g3 c5d6 g3f2');
INSERT INTO evaluations VALUES(55055,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(55055,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(55055,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(55055,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(55055,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(55055,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(55055,36,'cp -31','c7c5 c2c3 c5d4 c3d4 b8c6 a2a3 c8f5 b1c3 e7e6 g1f3 f8d6 e3g5 g8e7 e2e3 h7h6 g5h4 e8g8 f1d3 f5d3 d1d3 c6a5 e1g1 a8c8 h4g3 a5c4 d3e2 e7f5 g3d6 f5d6 b2b3 c4a5 e2b2 b7b6 a1c1 d8e7 h2h3 c8c6 c3e2 f8c8 c1c6 c8c6 f1c1 c6c1 e2c1');
INSERT INTO evaluations VALUES(2786,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(2786,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(2786,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(2786,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(2786,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(2786,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(2786,36,'cp 149','d2d4 c5d6 d4e5 d6e5 f3e5 c6e5 f2f4 e5g6 e4e5 f6e8 d1f3 a8b8 c1e3 d7d6 b1d2 c8f5 e3a7 b8c8 a7f2 d8d7 d2e4 c8d8 a1e1 f5e6 f2d4 c7c5 d4f2 e6b3 a2b3 f7f5 e4g5 d6e5 f4e5 h7h6 e5e6 d7d5 e1d1 d5f3 g5f3 d8c8 d1d5 e8f6 d5f5 f6e4 f5f8 c8f8');
INSERT INTO evaluations VALUES(2779,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(2779,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(2779,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(2779,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(2779,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(2779,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(2779,36,'cp 127','c5d6 d4e5 d6e5 f3e5 c6e5 f2f4 e5g6 e4e5 f6e8 d1f3 a8b8 c1e3 d7d6 b1d2 a6a5 f3g3 d8h4 e3a7 b8a8 a7d4 h4g3 h2g3 c8e6 b3e6 f7e6 f1e1 d6e5 d4e5 g6e5 e1e5 e8d6 e5e6 b5b4 a1e1 b4c3 b2c3 f8b8 a2a4 b8b2 d2e4 b2c2 e4d6');
INSERT INTO evaluations VALUES(21164,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(21164,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(21164,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(21164,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(21164,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(21164,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(21164,36,'cp 57','a7a6 b5c6 b7c6 d2d4 g8f6 b1d2 a6a5 e1g1 e7e6 a2a4 f8e7 f1e1 c8b7 d2b3 c6c5 d1d3 e8g8 d4c5 d6c5 c1d2 d8d3 c2d3 b7a6 b3a5 a6d3 a5c6 e7d8 e4e5 f6d7 b2b4 f8e8 e1c1 c5b4 d2b4 d7b8 c6b8 a8b8 b4c5 d3a6 a4a5 h7h6 h2h3 d8g5 f3g5 h6g5 c1e1 b8b2 c5e3');
INSERT INTO evaluations VALUES(42092,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(42092,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(42092,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(42092,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(42092,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(42092,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(42092,36,'cp 40','b2c3 e8g8 c4d5 e6d5 f1d3 b7b6 g1e2 c7c5 e1g1 c8a6 d3a6 b8a6 d1d3 a6c7 d4c5 b6c5 c3c4 a8b8 f1d1 d5c4 d3c2 d8e7 c2c4 f8d8 c1d2 c7e6 d2e1 e7b7 e2c3 d8d1 a1d1 h7h6 a3a4 b7b3 c4b3 b8b3 c3b5 a7a6 b5d6 b3b2 d1c1 f6d5 h2h3 b2a2 a4a5');
INSERT INTO evaluations VALUES(51947,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51947,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(51947,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(51947,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(51947,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(51947,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(51947,36,'cp 48','g1f3 g8f6 f1g2 b8c6 b1c3 d7d5 c4d5 e6d5 d2d4 c5d4 f3d4 f8c5 d4b3 c5b6 c3a4 e8g8 a4b6 a7b6 c1e3 h7h5 e1g1 f8e8 e3g5 d8d6 g5f4 d6e7 f1e1 c8f5 b3d4 f5e4 f2f3 e4g6 e2e3 c6e5 f4e5 e7e5 d1b3 h5h4 g3h4');
INSERT INTO evaluations VALUES(10709,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(10709,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10709,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(10709,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10709,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(10709,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(10709,36,'cp 15','e8g8 d2d3 h7h6 b5a4 a7a6 a2a3 b7b5 a4b3 d6c5 c3d5 d7d6 c2c3 a8b8 d3d4 c5a7 c1e3 c8g4 d4e5 f6d5 e3a7 d5c3 b2c3 c6a7 e5d6 d8d6 d1d6 c7d6 f3d4 b8c8 f2f3 g4d7 d4e2 d7e6 b3e6 f7e6 a3a4 d6d5 e4d5 e6d5 f1d1 c8c5 h2h3 b5a4 a1a4 a7b5 a4a6 b5c3 e2c3');
INSERT INTO evaluations VALUES(4509,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(4509,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4509,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4509,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4509,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(4509,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4509,36,'cp 16','d2d3 h7h6 b5a4 f8e8 a2a3 a7a6 h2h3 d6f8 a4b3 d7d6 c1e3 c6e7 f3h4 g7g5 h4f3 e7g6 d3d4 b7b5 d4e5 d6e5 d1d8 e8d8 a3a4 b5b4 c3d5 f6e4 d5c7 a8b8 a1d1 c8b7 c7d5 f8c5 a4a5 c5e3 f2e3 b7d5 d1d5 d8d5 b3d5 e4d6 f3d2');
INSERT INTO evaluations VALUES(17405,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(17405,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(17405,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(17405,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(17405,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(17405,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(17405,36,'cp -26','d1e2 f8e7 d2d3 d7d5 c4b3 g8f6 c2c3 d8c7 e1g1 h7h6 f1e1 g7g5 b1d2 c8d7 e4d5 e6d5 f3e5 e8g8 e5d7 c7d7 d2f1 a8e8 e2c2 e7d6 c1d2 d5d4 e1e8 f8e8 a1e1 e8e1 d2e1 b7b5 c2d1 d6c7 a4b5 a6b5 c3d4 c6d4 e1c3 d7d6 g2g3 d4b3 c3f6 d6f6 d1b3');
INSERT INTO evaluations VALUES(21594,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(21594,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(21594,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(21594,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(21594,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21594,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(21594,36,'cp 40','g1f3 c8f5 c1g5 f8e7 g5e7 g8e7 e2e3 e8g8 f1d3 f5d3 d1d3 h7h6 e1g1 e7c8 f1c1 a7a5 a2a3 c8b6 d3c2 c6e7 b2b4 a5b4 a3b4 a8a1 c1a1 b6c4 b4b5 d8d6 a1c1 c7c6 b5c6 b7c6 f3e1 f8a8 e1d3 e7c8 h2h3 c8b6 c2b3');
INSERT INTO evaluations VALUES(21894,30,'cp -550',NULL);
INSERT INTO evaluations VALUES(21894,31,'cp -569',NULL);
INSERT INTO evaluations VALUES(21894,32,'cp -578',NULL);
INSERT INTO evaluations VALUES(21894,33,'cp -586',NULL);
INSERT INTO evaluations VALUES(21894,34,'cp -578',NULL);
INSERT INTO evaluations VALUES(21894,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(21894,36,'cp -582','d8g5 e1g1 g5g6 b1c3 g8f6 c3b5 c8g4 f2f3 g4h3 d1e2 e8c8 c4d5 f6d5 e4d5 c6b4 d2d3 b4d5 b5a7 c8d7 g1h1 h3f5 c1d2 f8e7 a7b5 d8a8 b5c3 d5b4 d2e3 f5e6 a2a3 b4c6 c3b5 h8e8 c2c4 f7f5 a1d1 d7c8 f1e1 e6d7 e2d2 f5f4 e3f2 g6f7 b2b3 f7f6 d3d4 e5d4 f2d4 c6d4 d2d4 f6d4');
INSERT INTO evaluations VALUES(1747,30,'cp -532',NULL);
INSERT INTO evaluations VALUES(1747,31,'cp -544',NULL);
INSERT INTO evaluations VALUES(1747,32,'cp -525',NULL);
INSERT INTO evaluations VALUES(1747,33,'cp -530',NULL);
INSERT INTO evaluations VALUES(1747,34,'cp -551',NULL);
INSERT INTO evaluations VALUES(1747,35,'cp -551',NULL);
INSERT INTO evaluations VALUES(1747,36,'cp -551','e1g1 g5g6 b1c3 g8f6 c3b5 c8g4 f2f3 g4h3 d1e2 e8c8 c4d5 c6e7 d2d3 e7d5 e4d5 f6d5 b5a7 c8d7 c1d2 f8e7 g1h1 h3f5 a7b5 d8a8 a2a4 h7h5 h2h3 d5f6 h3h4 f6d5 e2f2 f5e6 a4a5 g6f5 b5c3 d5f4 d2f4 e5f4 c3e4 a8a5 b2b4 a5a1 f1a1 f5e5 c2c3');
INSERT INTO evaluations VALUES(26170,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(26170,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(26170,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(26170,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26170,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26170,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(26170,36,'cp 10','e1g1 g8f6 d2d4 f8e7 b1c3 e8g8 c1g5 h7h6 g5h4 a8c8 d4c5 d7c5 f1e1 c5e6 a1b1 d5d4 c3e2 e7b4 c2c3 d4c3 b2c3 d8d1 e1d1 b4c3 b1b7 a7a5 e2c3 c8c3 h4f6 g7f6 h2h3 c3c2 a2a4 c2a2 d1d7 a2a1 g1h2 a1a4');
INSERT INTO evaluations VALUES(26169,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(26169,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(26169,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(26169,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26169,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26169,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(26169,36,'cp 0','g8f6 f1e1 f8e7 d1e2 a8c8 b1c3 c8c6 d2d4 c6e6 c1e3 e8g8 f3g5 e6a6 d4c5 e7c5 a1d1 c5e3 e2e3 f8e8 e3d2 e8e1 d2e1 d7f8 g5f3 h7h6 f3d4 f8e6 e1e3 e6d4 e3d4 a6b6 c3a4 b6c6 a4c3');
INSERT INTO evaluations VALUES(24530,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(24530,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24530,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(24530,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24530,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(24530,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(24530,36,'cp 7','d2d4 f8e7 d4c5 d7c5 c1e3 e8g8 b1d2 f8e8 c2c3 e7d6 d2b3 c5e4 d1d3 d8d7 f1e1 a7a5 a2a4 b7b6 b3d4 d6c5 h2h3 e4d6 e3f4 e8e1 f3e1 a8e8 e1f3 f6e4 b2b3 d7c8 a1e1 d6f5 e1e2 c5d4 f3d4 f5d4 c3d4 c8c6');
INSERT INTO evaluations VALUES(43366,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43366,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(43366,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43366,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43366,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(43366,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(43366,36,'cp 28','g2g3 d5c4 a2a4 e7e6 f1g2 c6c5 d4c5 d8d1 c3d1 f8c5 d1e3 c8d7 f3e5 d7c6 g2c6 b8c6 e5c6 b7c6 e3c4 e8e7 f2f3 h8b8 c1d2 c5b4 e1c1 b4d2 d1d2 b8b4 c4a5 b4a4 a5c6 e7f8 c1b1 a4c4 h1c1 c4c1 b1c1 g7g5 d2d8 a8d8 c6d8 f8e7 d8b7 f6d7 c1d2 f7f5 d2c3 g5g4 f3f4 e7f6 c3d4 e6e5 f4e5 d7e5 b7c5');
INSERT INTO evaluations VALUES(1438,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(1438,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(1438,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(1438,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(1438,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(1438,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(1438,36,'cp -29','c5d4 f3d4 a7a6 d4c6 a6b5 c6d4 b5b4 c3e2 g8e7 e1g1 e7g6 f2f4 f8c5 c1e3 c8d7 e3f2 e8g8 d4b3 c5f2 f1f2 f8c8 d1d4 b6d4 b3d4 a8a6 h2h3 g6e7 f2f3 f7f6 f3b3 c8c4 c2c3 f6e5 f4e5 b4c3 b2c3 e7g6 d4f3 c4c7 a1b1 a6a2 b3b7 c7b7 b1b7 a2e2');
INSERT INTO evaluations VALUES(1436,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(1436,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(1436,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(1436,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(1436,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(1436,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(1436,36,'cp -22','f3d4 a7a6 d4c6 a6b5 c6d4 b5b4 c3e2 g8e7 c1e3 e7g6 d4b3 b6c7 f2f4 b7b6 e2d4 f8e7 e1g1 e8g8 d1d2 e7c5 h2h3 c8a6 f1e1 h7h6 d4f3 f8c8 b3d4 g6e7 d2f2 e7c6 a1c1 c6d4 f3d4 a6b7 d4b5 c7d7 e3c5 b6c5 b5d6 c8c6 c2c3');
INSERT INTO evaluations VALUES(12864,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(12864,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(12864,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(12864,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12864,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(12864,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(12864,36,'cp -7','e7e6 e1g1 d7d6 d1e2 d8c7 f1d1 a7a6 c1f4 f8e7 a1c1 e8g8 c4b3 c7b8 c3a4 b7b5 c1c6 b5a4 b3a4 c8d7 c6c4 d7b5 a4b5 a6b5 c4c6 a8a2 f4d6 e7d6 c6d6 h7h6 g2g3 b8b7 d6d4 a2a4 b2b4 b7b8 e2d3 f8c8 d1d2 b8a8 f3e5 a4a2 d4d8 c8d8 d3d8 a8d8 d2d8 g8h7 e5f7 f6e4');
INSERT INTO evaluations VALUES(18215,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(18215,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(18215,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(18215,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(18215,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(18215,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(18215,36,'cp 80','f1e1 e7e5 c2c3 g8e7 d2d4 e7g6 d4c5 d6c5 d1a4 f8e7 a4c6 c8d7 c6d5 d7e6 d5d8 a8d8 c1e3 d8d3 b2b3 c5c4 e1c1 e8g8 b1d2 f8c8 f3e1 d3d7 g2g3 a6a5 d2c4 e6c4 b3c4 c8c4 e1f3 d7d3 h2h4 h7h5 a1b1 a5a4 b1b7 d3c3 c1c3 c4c3 g1f1 f7f6');
INSERT INTO evaluations VALUES(19423,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(19423,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(19423,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(19423,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(19423,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(19423,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(19423,36,'cp 170','d4d5 c6b8 e2e4 d7d6 a2a4 g7g6 b1c3 g8f6 h2h3 f8g7 f1e2 a7a6 c1f4 e8g8 d1d2 b8d7 e1g1 f6e8 d2e3 a8b8 a4a5 b7b6 a5b6 d8b6 c3a4 b6c7 f1e1 g7f6 a1d1 b8b4 b2b3 f6g7 e2c4 c8b7 f4h6');
INSERT INTO evaluations VALUES(19512,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(19512,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(19512,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(19512,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(19512,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(19512,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(19512,36,'cp 180','c6b8 e2e4 d7d6 b1c3 g7g6 h2h3 g8f6 f1e2 f8g7 c1f4 e8g8 a2a4 b7b6 e1g1 c8a6 e2b5 a6b7 f1e1 a7a6 b5f1 f6h5 f4h2 b8d7 d1d2 d8c7 a1d1 a8d8 f1c4 g8h8 c4e2 d8e8 e2f1 g7f6 g2g4 h5g7 g4g5 d7e5');
INSERT INTO evaluations VALUES(30658,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(30658,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30658,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30658,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(30658,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(30658,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(30658,36,'cp 25','e2e3 g8f6 f1d3 c7c5 e1g1 b8c6 a2a3 d7d5 c2c3 f8d6 b1d2 e8g8 f1e1 d8c7 f4d6 c7d6 e3e4 d5e4 d2e4 f6e4 d4c5 d6c5 d3e4 a8d8 d1e2 c5e7 a1d1 d8d1 e1d1 h7h6 h2h3 f8d8 d1d2 d8d2 f3d2 e7d8 d2f3 c6a5 e4b7 a5b7');
INSERT INTO evaluations VALUES(44486,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(44486,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(44486,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(44486,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(44486,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(44486,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(44486,36,'cp 118','e4e5 f6d5 c2c4 d5c7 c1e3 d7d6 f2f4 b7b5 g1f3 b5c4 f1c4 e7e6 e1g1 c7d5 c4d5 c6d5 b1c3 f8e7 f4f5 e8g8 f5f6 g7f6 e3h6 b8c6 e5f6 e7f6 h6f8 d8f8 a1c1 c8a6 f1f2 a8c8 g1h1 f8g7 d1a4 a6b7 c3b5');
INSERT INTO evaluations VALUES(44457,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(44457,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(44457,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(44457,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(44457,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(44457,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(44457,36,'cp 109','f6d5 c2c4 d5c7 b1c3 d7d6 f2f4 g7g6 g1f3 f8g7 c1e3 e8g8 f1e2 b7b5 h2h4 h7h5 f3g5 b5c4 g2g4 d6e5 f4e5 c7d5 c3d5 c6d5 g4h5 d8a5 d1d2 a5d2 e1d2 c8f5 h5g6 f7g6 b2b3 b8c6 b3c4 a8b8 e2f3 b8b2 d2c1 b2c2 c1d1 f8d8 f3d5 d8d5 c4d5 c6b4 a1b1 a7a5');
INSERT INTO evaluations VALUES(44496,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(44496,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(44496,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(44496,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(44496,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(44496,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(44496,36,'cp 89','c2c4 d5c7 b1c3 d7d6 f2f4 g7g6 g1f3 f8g7 c1e3 e8g8 f1e2 b7b5 h2h4 h7h5 f3g5 b5c4 g2g4 c7d5 c3d5 c6d5 g4h5 b8c6 h5g6 f7g6 h4h5 d6e5 d4e5 d8a5 d1d2 a5d2 e3d2 g7h6 e2f3 e7e6 h1g1 c6d4 f3d1 a8b8 b2b3 c8d7 h5g6 f8f5 b3c4 h6g5 g1g5 d5c4');
INSERT INTO evaluations VALUES(44495,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(44495,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(44495,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(44495,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(44495,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(44495,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(44495,36,'cp 101','d5c7 b1c3 d7d6 f2f4 g7g6 h2h4 h7h5 g1f3 f8g7 c1e3 e8g8 d1d2 c8f5 f1d3 f5d3 d2d3 d6e5 f4e5 b7b5 a1d1 b5c4 d3c4 c7d5 e3g5 b8d7 c3d5 c6d5 c4d5 e7e6 d5b7 d8c8 b7c8 f8c8 e1e2 a8b8 b2b3 c8c2 d1d2 c2c3 d2b2 f7f6 g5d2 c3c8 e5f6');
INSERT INTO evaluations VALUES(47211,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(47211,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(47211,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(47211,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(47211,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(47211,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(47211,36,'cp -12','c8f5 f1d3 e7e6 g1f3 f8d6 d3f5 e6f5 d1d3 d8d7 c1d2 e8g8 d2b4 b8c6 b4d6 c7d6 b1c3 f8e8 e1g1 h7h6 c3e2 a8c8 f1c1 c6e7 a3a4 g7g5 a4a5 d7e6 c2c3 a7a6 e2g3 g8h8 c1e1 h8g7 h2h3 f6e4 f3d2 e4g3 f2g3 e8h8');
INSERT INTO evaluations VALUES(29595,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(29595,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29595,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29595,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29595,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(29595,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(29595,36,'cp -8','f1d3 e7e6 d3f5 e6f5 d1d3 d8d7 g1f3 f8d6 c1d2 b8a6 e1g1 e8g8 c2c4 d5c4 d3c4 f8e8 d2c3 c7c6 b1d2 f6e4 c4b3 e4d2 f3d2 a6c7 d2c4 a8b8 c3b4 d6b4 a3b4 a7a6 c4e5 d7e7 f1e1 g7g6 h2h3 c7d5 e5d3 d5b6 d3c5');
INSERT INTO evaluations VALUES(49385,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(49385,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(49385,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(49385,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(49385,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(49385,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(49385,36,'cp 8','d7d5 d2d4 f8b4 c1d2 b4e7 d1c2 e8g8 f1g2 c7c6 e1g1 b8d7 f1d1 b7b6 d2f4 c8b7 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 d8c7 e2e3 a8c8 a2a4 a7a6 c2e4 a6a5 b2b3 d7f6 e4e5 f6d7 e5c7 c8c7 a1c1 e7b4 c4e5 f8c8 c1c2 b4e7 e5c4 b7a6 c3e4 c8d8 e4d6');
INSERT INTO evaluations VALUES(48241,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(48241,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(48241,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(48241,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(48241,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(48241,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(48241,36,'cp 9','f1g2 d5c4 d1a4 c7c6 a4c4 b7b5 c4c2 c8b7 e1g1 b8d7 a2a4 a7a6 d2d4 c6c5 c1g5 c5d4 g5f6 d7f6 a4b5 a6b5 b1d2 d4d3 c2d3 d8d3 e2d3 f8b4 f3d4 b7g2 g1g2 e8g8 d2f3 b4c5 a1a8 f8a8 f1c1 c5d4 f3d4 b5b4 c1c4 f6d5 d4c6 a8c8 c4b4 d5b4 c6e7 g8f8 e7c8');
INSERT INTO evaluations VALUES(10061,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(10061,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(10061,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10061,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10061,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10061,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10061,36,'cp 0','b1c3 e8g8 f1e2 h7h6 g5h4 f8e8 c3d5 g7g5 h4g3 d7d6 d5f6 d8f6 h2h4 g5g4 f3h2 h6h5 f2f3 f6h6 f3g4 h5g4 c2c3 c6e7 d1d2 c5e3 d2d1');
INSERT INTO evaluations VALUES(31909,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31909,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31909,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31909,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31909,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31909,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31909,36,'cp 0','b8c6 c1f4 e7e6 d1d2 g8f6 f3e5 c6e5 d4e5 f6e4 c3e4 d5e4 e1c1 d8d2 d1d2 a8d8 h2h3 g4f5 d2d8 e8d8 g2g4 f5g6 g4g5 h7h6 g5h6 g7h6 h3h4 h8g8 h1g1 g8h8');
INSERT INTO evaluations VALUES(32168,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(32168,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(32168,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(32168,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(32168,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(32168,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(32168,36,'cp -77','h7h6 d2d4 g7g5 g2g3 f8g7 g3f4 b8c6 c4b5 g8e7 d4d5 a7a6 b5e2 c6b8 c2c3 g4f3 e2f3 b8d7 e4e5 d6e5 f4g5 h6g5 b1d2 d7f6 d1b3 d8d7 b3b7 a8d8 b7a6 e8g8 a6e2 e7d5 d2e4 f6e4 e2e4 d5f4 c1f4 e5f4');
INSERT INTO evaluations VALUES(32054,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(32054,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(32054,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(32054,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(32054,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(32054,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(32054,36,'cp -44','d2d4 g7g5 g2g3 f8g7 g3f4 b8c6 c4b5 g5f4 d4d5 a7a6 d5c6 a6b5 c6b7 a8b8 b1c3 b8b7 c1f4 g8e7 d1d3 b5b4 c3d5 e8g8 f3d4 e7d5 e4d5 d8f6 f4e3 f6g6 d3g6 f7g6 f1f8 g8f8 a2a4 b4a3 b2a3 f8e8 a3a4 g4f3 c2c4 b7b4 a1e1 g7d4 e3d4 e8d7');
INSERT INTO evaluations VALUES(32055,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(32055,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(32055,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(32055,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(32055,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(32055,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(32055,36,'cp -53','g7g5 g2g3 f8g7 g3f4 b8c6 c4b5 a7a6 b5c6 b7c6 d1d3 g5f4 c1f4 g8e7 b1c3 e7g6 c3e2 e8g8 a1e1 f7f5 e4e5 g8h7 e5d6 c7d6 f4d2 a8b8 b2b3 d8f6 e2f4 g6f4 d2f4 b8e8 f4g3 f8g8 e1e8 g8e8 c2c3 e8e4 f3d2 e4e2');
INSERT INTO evaluations VALUES(32024,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(32024,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(32024,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(32024,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(32024,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(32024,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(32024,36,'cp -56','g2g3 f8g7 g3f4 b8c6 c4b5 a7a6 b5c6 b7c6 d1d3 g5f4 c1f4 g8e7 b1c3 e7g6 c3e2 e8g8 a1e1 g8h7 f4g3 f7f5 e4e5 f8g8 e5d6 c7d6 e2f4 g6f4 g3f4 g7f6 g1f2 d8d7 f1g1 g4f3 g1g8 f3e4 d3h3 f6d4 f2f1 h7g8 e1e4 d4g7 e4e2 g8h7 h3d3 d6d5 b2b3 a8g8');
INSERT INTO evaluations VALUES(36721,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(36721,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(36721,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(36721,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(36721,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(36721,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(36721,36,'cp -10','c7c5 d4c5 f6e4 a1c1 b8d7 c4d5 d8a5 d1c2 e4c3 b2c3 d7c5 f3d4 e7e5 d5e6 c8e6 d4e6 c5e6 f1e2 a8c8 e1g1 c8c3 c2b1 e6f4 e3f4 c3c1 f1c1 a5d2 c1c2 d2f4 b1b7 f4f5 b7c6 h7h5 g2g3 f8d8 e2f3 h5h4 g3h4 g7e5 c6e4 f5e4 f3e4 g8g7 e4f3 d8b8 h4h5');
INSERT INTO evaluations VALUES(36025,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(36025,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(36025,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(36025,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(36025,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(36025,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(36025,36,'cp -12','d4c5 f6e4 a1c1 b8d7 c4d5 d8a5 d1c2 e4c3 b2c3 d7c5 f3d4 e7e5 d5e6 c8e6 d4e6 c5e6 f1e2 a8c8 e1g1 c8c3 c2b1 e6f4 e3f4 f8e8 e2b5 e8d8 c1c3 g7c3 b1b3 a5b4 b5c4 b4b3 c4b3 d8d2 f1d1 d2d1 b3d1 g8f8 g1f1 f8e7 f1e2 c3d4 d1c2 f7f5 c2b3 e7d6 b3g8');
INSERT INTO evaluations VALUES(36023,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(36023,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(36023,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(36023,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(36023,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(36023,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(36023,36,'cp -5','f6e4 a1c1 b8d7 c4d5 d8a5 d1c2 e4c3 b2c3 d7c5 f3d4 e7e5 d5e6 c8e6 d4e6 c5e6 f1e2 f8c8 e1g1 c8c3 c2b1 e6f4 e3f4 a8e8 e2c4 c3c1 b1c1 b7b5 c4b3 a5c3 g2g3 g7d4 c1c3 d4c3 f1d1 e8e1 d1e1 c3e1 g1g2 e1c3 a2a4 b5a4 b3a4 c3d4 g2f3 d4c5 a4b3 g8f8 f3e2 f8e7 f2f3');
INSERT INTO evaluations VALUES(41813,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41813,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41813,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41813,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41813,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41813,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41813,36,'cp 0','b7b6 c2c3 e8g8 g1f3 c7c5 f1d3 c8b7 d1e2 f6e4 f3e5 d8c7 e2h5 b8d7 e5d7 c7d7 g3d6 d7d6 h5h4 h7h6 d2e4 d5e4 d3e4 g7g5 h4g4 f7f5 g4h5 b7e4 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8 f2f3 e4d3 e1c1 c5c4 g6h6 h8g8 h6g6');
INSERT INTO evaluations VALUES(49705,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(49705,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(49705,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(49705,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(49705,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(49705,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(49705,36,'cp -68','c2c4 c7c6 b1c3 d7d5 d2d4 f8b4 a2a3 b4c3 b2c3 e8g8 f3d1 e5e4 a3a4 f8e8 c4d5 c6d5 g1e2 b8c6 h2h3 c6a5 e2f4 d8c7 f1e2 b7b6 e1g1 a5c4 e2c4 c7c4 c1a3 c8a6 f1e1 c4c7 a4a5 a6c4 a5b6 a7b6 a3b4 a8c8 b4a3 g7g5');
INSERT INTO evaluations VALUES(14789,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(14789,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(14789,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(14789,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(14789,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(14789,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(14789,36,'cp 64','g1f3 f6e4 d1c2 e4c3 c2c3 a7a5 c1f4 d5c4 c3c4 f8d6 f4g3 h7h5 e2e3 a5a4 g3h4 f7f6 h4g3 g7g5 f1e2 h5h4 g3d6 c7d6 h2h3 e8f7 e1g1 c6e7 b2b4 c8d7 a1c1 e7d5 c4c2 a8c8 c2d2 d8b6 c1c8 h8c8 e2d3 d5c3 d2c2');
INSERT INTO evaluations VALUES(14791,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14791,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(14791,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(14791,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(14791,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(14791,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(14791,36,'cp 35','d5c4 b1d2 b7b5 a3a4 b5a4 f1c4 c8d7 c4b5 c6b4 b5d7 f6d7 d1a4 f8e7 f3e5 f7f6 e5c6 d7b6 c6d8 b6a4 a1a4 e8d8 e1e2 d8d7 d2e4 b4d5 c1d2 h8b8 d2c3 a7a5 h1a1 b8b5 e2d3 e7b4 c3b4 d5b4 d3e2 b4d5 b2b3 f6f5 e4c5 b5c5 d4c5 d5c3 e2d3 c3a4');
INSERT INTO evaluations VALUES(15565,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(15565,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(15565,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(15565,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(15565,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(15565,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(15565,36,'cp 28','b1c3 a7a6 h2h3 c8e6 f1c4 g7g6 c4e6 f7e6 d2d4 e5d4 f3d4 c6d4 d1d4 f8g7 e1c1 f6h5 d4e3 e8g8 g2g3 e6e5 c3d5 f8f7 h1f1 d8d7 g3g4 h5f4 d5f4 f7f4 f2f3 a8f8 c1b1 d7f7 d1d3 d6d5 d3d5 f4f3 f1f3 f7f3 e3f3 f8f3 a2a4 f3h3 d5d8 g8f7');
INSERT INTO evaluations VALUES(15567,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15567,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(15567,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(15567,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(15567,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(15567,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(15567,36,'cp 19','a7a6 h2h3 c8e6 f1c4 g7g6 c4e6 f7e6 d2d4 c6d4 f3d4 e5d4 d1d4 f8g7 d4e3 e8g8 e1c1 f6h5 g2g3 e6e5 c3d5 f8f7 c1b1 d8d7 g3g4 h5f4 d5f4 f7f4 f2f3 a8f8 h1f1 d7f7 d1d3 d6d5 d3d5 f4f3 f1f3 f7f3 b2c1 f3e3 c1e3 f8f3 d5d8 f3f8 d8d7 f8f7 d7f7 g8f7 b1b2 f7e7 g4g5 e7e6 b2c3');
INSERT INTO evaluations VALUES(16497,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(16497,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(16497,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(16497,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(16497,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(16497,35,'cp -121',NULL);
INSERT INTO evaluations VALUES(16497,36,'cp -102','d7d5 b1d2 f8c5 f1e2 d5e4 d3e4 f6g4 e2g4 d8h4 d1e2 h4g4 g1f3 f7f6 e1g1 c8d7 h2h3 g4g6 f3h4 g6f7 c2c3 c6e7 b3b4 g7g5 b4c5 g5h4 f2f4 h8g8 f4e5 d7h3 f1f2 e8c8 e5f6 e7c6 c3c4 h3e6 a1d1 f7d7 e2e3 h4h3 g2g3 e6c4 e3f3 d7d3 f3d3 d8d3');
INSERT INTO evaluations VALUES(15127,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(15127,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(15127,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(15127,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(15127,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(15127,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(15127,36,'cp -90','b1d2 f8c5 f1e2 d5e4 d3e4 f6g4 e2g4 d8h4 d1e2 h4g4 g1f3 f7f6 e1g1 a7a6 h2h3 g4e6 f3h4 c8d7 h4f5 e6f7 c2c3 e8c8 a2a4 d7f5 e4f5 c5a7 b3b4 c6e7 b4b5 f7d5 d2e4 d5d3 e2d3 d8d3 b5a6 b7a6 b2a3 e7f5 a3c5 a7c5 e4c5 d3c3 c5a6 c8b7 a6b4 c3b3 f1b1 f5d4 b1b3 d4b3 a1b1');
INSERT INTO evaluations VALUES(19903,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(19903,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(19903,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(19903,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(19903,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(19903,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(19903,36,'cp -39','d8d6 b1c3 g8f6 d2d4 c5d4 c3b5 d6b8 b5d4 e7e5 d4c6 b7c6 f1c4 e5e4 d1e2 f8c5 f3g5 e8g8 c4f7 g8h8 f7e6 h7h6 e6c8 b8c8 g5h3 a8b8 e1g1 c8f5 c2c3 c5d6 f2f3 b8e8 f3e4 e8e4 f1f5 e4e2 g1f1 e2c2 c1f4 f6d5 f5f8 d6f8 a1e1 g7g5 f4g3 c2b2 g3e5 f8g7 c3c4 d5c3 a2a3 g7e5');
INSERT INTO evaluations VALUES(1418,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(1418,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(1418,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(1418,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(1418,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(1418,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(1418,36,'cp -17','b1c3 g8f6 d2d4 c5d4 c3b5 d6b8 b5d4 e7e5 d4c6 b7c6 f1c4 f8d6 c1g5 e5e4 g5f6 e4f3 f6c3 f3g2 d1e2 c8e6 h1g1 d6b4 c4e6 b4c3 b2c3 e8g8 e1c1 b8f4 e2e3 f7e6 g1g2 a8d8 d1d8 f4e3 f2e3 f8d8 g2g4 g8f7 g4f4 f7e7 f4g4 g7g6 g4a4 d8d7 a4c4 d7c7 c1d2 c6c5 d2e2 e6e5 e2f3 e7f6 h2h4 h7h5 c4a4 f6f5 e3e4 f5f6 a4a6 f6g7 a2a3 c7f7 f3g3');
INSERT INTO evaluations VALUES(22360,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(22360,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(22360,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(22360,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(22360,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22360,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22360,36,'cp 0','g8f6 f3d2 c7c6 c2c4 a7a5 b1c3 d7c5 b2b3 c8g4 f2f3 g4d7 f1e2 f8e7 d2f1 h7h5 f1e3 h5h4 a1b1 f6h5 e1g1 e8g8 g1h1 h5f4 g2g3 h4g3 h2g3 f4h5 h1g2 e7h4 g3h4 h5f4 g2g3 f4h5');
INSERT INTO evaluations VALUES(15904,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(15904,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(15904,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(15904,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15904,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(15904,36,'cp 11','f3d2 c7c6 c2c4 a7a5 d1c2 a5a4 f1e2 f8e7 b1c3 d7c5 e1g1 e8g8 b2b4 a4b3 d2b3 c5a4 c3a4 a8a4 b3d2 a4a8 d2b1 f6d7 a2a4 f7f5 c1a3 f5e4 b1c3 d7f6 d5c6 b7c6 c3e4 c8f5 e4f6 f8f6 c2d1 d8e8 a3b2 a8b8 b2c3 e8g6 a4a5 f5e4 f2f3');
INSERT INTO evaluations VALUES(34980,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(34980,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(34980,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(34980,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(34980,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(34980,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(34980,36,'cp 139','g1f3 e5d4 d1d4 c8e6 f1b5 b8d7 c1g5 c7c6 b5c4 g8f6 c4e6 f7e6 e1c1 e8c8 a2a3 d8e8 h1e1 f8c5 d4d2 f6d5 g5h4 d7f6 c3e4 a5d2 d1d2 f6e4 e1e4 h8f8 h4g3 c5b6 f3g5 d5f6 e4e6 h7h6 e6e8 f8e8 g5f3 b6c7 c1d1 c7g3 h2g3 c8c7 c2c3 f6g4 d2e2 e8e2 d1e2 c6c5 e2f1');
INSERT INTO evaluations VALUES(34979,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(34979,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(34979,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(34979,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(34979,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(34979,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(34979,36,'cp 138','f8b4 c1d2 e5d4 f3d4 a5e5 d1e2 e5e2 f1e2 g8f6 e1c1 e8g8 a2a3 b4a5 h1e1 c7c6 d2f4 f8e8 f4g3 a5c3 b2c3 b8d7 e2f3 d7c5 e1e8 f6e8 c3c4 c8e6 d4b3 c5b3 c2b3 e8f6 a3a4 a8e8 a4a5 a7a6 d1e1 e8d8 c1b2 d8d3 g3b8 h7h6 e1d1 d3d1 f3d1 f6d7 b8c7');
INSERT INTO evaluations VALUES(37176,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37176,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37176,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37176,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37176,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(37176,36,'cp 9','b7b6 e2e3 c7c5 f1d3 d7d5 c4d5 e6d5 d4c5 b6c5 e1g1 c8g4 h2h3 g4h5 d3e2 b8c6 d1a4 a8c8 a2a3 b4a5 e2b5 h5f3 b5c6 f3h5 b2b4 c5b4 a3b4 a5b4 a4b4 c8c6 a1a7 h7h6 f1a1 c6c4 b4a5 h5g6 a7a8 d8a5 a1a5');
INSERT INTO evaluations VALUES(35655,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(35655,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(35655,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(35655,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35655,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(35655,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(35655,36,'cp 9','e2e3 d7d5 c4d5 e6d5 a1c1 c7c5 f1e2 b4c3 d2c3 c5c4 b2b3 b6b5 a2a4 a7a6 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 f6e4 b3c4 d5c4 e2f3 f8e8 e5g6 h7g6 f3e4 e8e4 d4d5 b8d7 d5d6 d8b6 d1d5');
INSERT INTO evaluations VALUES(35653,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(35653,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35653,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35653,36,'cp 8','d7d5 c4d5 e6d5 a1c1 b4e7 f1d3 c7c5 e1g1 c8b7 f3e5 b8c6 a2a3 e7d6 e5g4 f6g4 d1g4 g7g6 c3d5 c6d4 e3d4 b7d5 d2g5 d8c8 g4c8 f8c8 d3a6 c8c7 d4c5 c7c5 f1d1 d5e6 c1c5 d6c5 b2b4 c5f8 g5f4 h7h6 h2h4 e6b3 d1c1 a8d8 c1c7 b3e6');
INSERT INTO evaluations VALUES(40267,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(40267,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(40267,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(40267,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(40267,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(40267,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(40267,36,'cp 52','f6e4 g5h4 e8g8 f1e2 g7f6 h4f6 e4c3 b2c3 e7f6 e1g1 b7b6 a1c1 c8a6 c4d5 a6e2 d1e2 c6d5 c3c4 b8c6 c4c5 d8d7 f3e1 g8g7 e1d3 a8b8 c1c3 f8c8 f1c1 c6e7 h2h3 b6c5 d3c5 d7d6 g2g4 b8b6 g1g2 c8b8 c5d3 b6b1 e2c2 b1c1 c2c1');
INSERT INTO evaluations VALUES(5385,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5385,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5385,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5385,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(5385,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(5385,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(5385,36,'cp -2','d8a5 c2c3 g8f6 b1d2 e8g8 d1e2 c6e5 c4b3 e5g4 d2c4 a5h5 e3d2 f6e4 e2e4 d7d5 e4e2 d5c4 b3c4 b7b5 c4b3 b5b4 h2h3 b4c3 b2c3 g4f6 e2h5 f6h5 e1g1 c8d7 d2e3 f8c8 a1c1 a7a5 b3d5 a8b8 d5f3 a5a4 g2g4 e7e5 g4h5 e5d4 c3d4 d7h3 f1e1 g7f8 e1d1 h3d7 h5g6 h7g6');
INSERT INTO evaluations VALUES(6069,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(6069,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(6069,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(6069,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(6069,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(6069,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(6069,36,'cp 57','b4d2 b1d2 g8f6 d4d5 c6e7 d1b3 e8g8 e1g1 c7c6 d5c6 e7c6 f1e1 d8b6 b3b6 a7b6 a2a3 f8d8 a1c1 c8e6 c4e6 f7e6 d2c4 b6b5 c4d6 d8d6 e4e5 c6e5 e1e5 b5b4 a3b4 d6b6 b4b5 f6d5 b2b3 h7h6 h2h3 a8d8 c1c5 d5f4 e5e4 f4d3 e4d4 d8d4 c5c8 g8f7 c8c7 f7f6 f3d4');
INSERT INTO evaluations VALUES(6077,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(6077,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(6077,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(6077,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(6077,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(6077,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(6077,36,'cp 60','b1d2 g8f6 c4b5 e8g8 e1g1 c6e7 b5d3 c7c6 a1c1 f6h5 g2g3 g7g6 f1e1 d6d5 d1b3 a7a5 a2a4 h5g7 h2h4 d8d6 e4d5 c6d5 d2f1 e7c6 f1e3 c6b4 b3d1 c8d7 d3b5 a8c8 d1d2 d7b5 a4b5 d6d7 c1c5 c8c5 d4c5 d7b5');
INSERT INTO evaluations VALUES(6075,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(6075,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(6075,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(6075,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(6075,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(6075,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(6075,36,'cp 61','g8f6 c4b5 e8g8 e1g1 c6e7 f1e1 c7c6 b5d3 e7g6 a1c1 g6f4 d3b1 g7g6 h2h3 f6h5 g1h2 d6d5 e4e5 f7f6 c1c3 f6e5 e1e5 d8d7 f3g1 h5g7 d2f3 d7d6 g1e2 f4e2 d1e2 c8f5 b1f5 f8f5 c3b3 d6c7 h2g1 f5e5 e2e5 c7e5 f3e5 g7f5 b3b7 f5d4 e5d7 d4e6 g1f1');
INSERT INTO evaluations VALUES(9454,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(9454,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(9454,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(9454,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(9454,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(9454,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(9454,36,'cp 75','e1g1 e8g8 c4b5 c6e7 f1e1 c7c6 b5d3 f6h5 g2g3 g7g6 h2h4 d6d5 e4e5 c8g4 d1b3 d8b6 b3a3 f8e8 a3c3 h5g7 f3h2 g4c8 d2b3 a7a5 b3c5 h7h5 h2f3 b6c7 a1c1 b7b6 c5a4 c8g4 f3g5 a8b8 a2a3 c7d7 c3d2 g4f5 a4c3 f5d3');
INSERT INTO evaluations VALUES(9453,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(9453,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(9453,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(9453,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(9453,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(9453,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(9453,36,'cp 65','e8g8 c4b5 c6e7 f1e1 c7c6 b5d3 f6h5 g2g3 g7g6 h2h4 d6d5 e4e5 c8g4 d1c2 h5g7 f3h2 g4f5 d2f3 h7h5 f3g5 a7a5 d3f5 e7f5 h2f3 f8e8 a1d1 d8b6 d1d3 e8e7 g1g2 a8f8 e1e2 f8e8 a2a3 b6c7 d3c3 c7d7 e2e1 g7e6');
INSERT INTO evaluations VALUES(3890,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(3890,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(3890,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(3890,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(3890,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(3890,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(3890,36,'cp 62','c4b5 c6e7 f1e1 c7c6 b5d3 f6h5 g2g3 g7g6 h2h4 d6d5 e4e5 c8g4 d1c2 h5g7 f3h2 g4f5 d2f3 h7h5 f3g5 a7a5 d3f5 e7f5 h2f3 f8e8 a1d1 a5a4 d1d3 a8a6 a2a3 d8d7 g1g2 d7d8 e1b1 d8d7 b1c1 d7e7 g5h3 a6b6 c1e1 g7e6 c2d2 e7c7 f3g5 e6g5 h3g5');
INSERT INTO evaluations VALUES(10123,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(10123,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(10123,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10123,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(10123,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(10123,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(10123,36,'cp 20','g1e2 d7d6 b1c3 c6d4 e2d4 c5d4 a2a4 c7c6 a4a5 e8g8 e1g1 c8e6 c4e6 f7e6 f3h3 d8e7 c3e2 d4c5 c2c3 a7a6 b2b4 c5a7 c1e3 a7e3 h3e3 a8e8 f2f3 f6d7 a1e1 h7h6 d3d4 e5d4 e2d4 f8f7 e1d1 e8f8 h2h3 g8h8 f3f4 e6e5 d4f5 e7e6');
INSERT INTO evaluations VALUES(10122,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10122,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10122,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10122,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10122,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(10122,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10122,36,'cp 13','c6a5 b1c3 c7c6 h2h3 a5c4 d3c4 d7d6 b2b3 c8e6 c1e3 d8b6 e3g5 f6d7 e1g1 b6c7 a1d1 f7f6 g5c1 a7a6 c3a4 b7b5 a4c5 d7c5 e2g3 e8g8 f1e1 c5b7 g3f5 e6f5 e4f5 a8d8 c1e3 d8e8 a2a3 h7h6 g2g4 b5c4 b3c4');
INSERT INTO evaluations VALUES(23863,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(23863,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23863,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23863,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(23863,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(23863,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(23863,36,'cp 9','h2h3 g4h5 d2d3 c6a5 c1g5 d8c8 a2a4 a5c4 d3c4 c5d4 c3b5 h7h6 g5d2 g8f6 b5d4 e5d4 d1e2 e8g8 f1e1 f8e8 e2d3 h5g6 f3d4 e8e4 e1e4 f6e4 f2f3 e4f6 d3b3 a7a6 a1e1 a8b8 d2f4 c8d7 f4g3 f6h5 g3f2 c7c5 d4e2 d7f5 e2g3 h5g3 f2g3 f5c2');
INSERT INTO evaluations VALUES(23862,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(23862,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(23862,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(23862,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(23862,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(23862,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23862,36,'cp 17','g4h5 d2d3 g8f6 c3d5 f6d5 c4d5 e8g8 c1e3 c6e7 d5b3 c5b6 a2a4 c7c6 g2g4 h5g6 e3b6 d8b6 a4a5 b6c7 d1d2 g8h8 f3h4 d6d5 f2f4 f7f6 f4f5 g6f7 d2e3 b7b6 a5b6 a7b6 h4f3 a8a1 f1a1 d5d4 e3d2 f7b3 c2b3 h7h6 g1g2');
INSERT INTO evaluations VALUES(23856,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(23856,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23856,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(23856,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23856,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(23856,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(23856,36,'cp 23','d2d3 g8f6 c3d5 f6d5 c4d5 e8g8 c1e3 c6e7 d5b3 c5e3 f2e3 g8h8 f1f2 c7c6 g2g4 h5g6 d1e2 f7f6 a1f1 d8d7 g1h2 d6d5 g4g5 d7d6 f3h4 a7a5 a2a4 d5e4 h4g6 e7g6 g5f6 f8f6 f2f6 g7f6 d3e4 a8d8 e2f3 d6d2 f1f2');
INSERT INTO evaluations VALUES(23857,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23857,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(23857,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(23857,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(23857,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(23857,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(23857,36,'cp 12','c6a5 c1g5 d8c8 d1e2 a5c4 d3c4 f7f6 g5e3 b7b6 e3c5 b6c5 e2d3 h5f3 d3f3 g8e7 b2b3 e8g8 f3d3 a7a5 c3d5 c8d7 d5e7 d7e7 a2a4 g7g6 a1e1 g8g7 d3d2 h7h5 e1e3 f6f5 e4f5 f8f5 e3e4 e7e6 f2f4 e6f6 f1f3 f5f4 f3f4 e5f4');
INSERT INTO evaluations VALUES(39740,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(39740,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(39740,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(39740,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(39740,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(39740,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(39740,36,'cp 34','h7h5 f1e2 b7b5 e1g1 e8g8 e3f2 c8d7 a2a4 b5b4 c3d1 d8c7 d2b4 e6e5 d4b3 c7c2 b4c4 c2c4 e2c4 b8c6 d1e3 c6b4 f1d1 f8c8 b3d2 a8b8 b2b3 d7e6 c4e6 f7e6 d2c4 d6d5 c4e5 d5e4 e3c4 b4c2');
INSERT INTO evaluations VALUES(53119,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(53119,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(53119,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(53119,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(53119,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(53119,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(53119,36,'cp -24','e7e6 g1f3 a7a6 e1g1 b7b5 c4e2 f8e7 b2b4 g8f6 a2a4 b5a4 e4e5 d6e5 d1d8 e7d8 f3e5 d8c7 c1f4 c8b7 c3a4 e8g8 a4c5 f6d5 f4g3 b7c6 e5c6 b8c6 g3c7 d5c7 e2f3 c7d5 f1b1 f8c8 c5d7 c6d4 f3d5 e6d5');
INSERT INTO evaluations VALUES(54015,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(54015,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(54015,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(54015,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(54015,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(54015,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54015,36,'cp 0','g1f3 f8e7 d1e2 c8d7 e1g1 b8c6 f1d1 a7a6 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 a1c1 d8e8 a2a3 d7e6 h2h3 a8c8 e2d2 e8d8 d2d3 e6d5 d3d5 d8e8 b2b4 e8d7 e3c5 f8d8 c5b6 d8f8');
INSERT INTO evaluations VALUES(56179,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(56179,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(56179,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(56179,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(56179,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(56179,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(56179,36,'cp -78','d5d4 d2d3 c7c5 b1d2 b8c6 a2a3 g8f6 f1e2 f8e7 e1g1 e8g8 f3g5 f6e8 f2f4 h7h6 g5e4 f7f5 e4g3 a7a5 e2f3 a8a6 f3c6 a6c6 d2f3 g8h7 c1d2 e8f6 f3e5 c6a6 g3h5 a6e6 f1e1 d8c7 h5f6 e7f6 b2b4 a5b4 a3b4 f6e5 f4e5 e6e5 e1e5 c7e5 d1f3 e5f6 a1e1 c5b4');
INSERT INTO evaluations VALUES(56647,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(56647,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(56647,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(56647,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(56647,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(56647,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(56647,36,'cp -76','d2d3 c7c5 b1d2 b8c6 a2a3 g8f6 f1e2 f8e7 f3g5 e8g8 e1g1 f6e8 f2f4 h7h6 g5e4 f7f5 e4g3 a7a5 e2f3 a8a6 b2b3 g8h7 f3c6 a6c6 d2f3 c6a6 f1e1 e7d6 a1a2 a5a4 b3a4 e8f6 a2e2 c8d7 d1b3 d8c7 h2h3 f6g8 b3b2 d7a4 b2d2 g8e7');
INSERT INTO evaluations VALUES(57238,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(57238,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(57238,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(57238,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(57238,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(57238,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(57238,36,'cp -29','c2c4 d7d5 c4d5 c8h3 g2h3 d8d5 h1g1 d5e6 f1g2 c7c6 d2d3 b8d7 b1d2 a7a5 d1b3 e6b3 d2b3 f8b4 b3d2 g8e7 a2a3 b4d6 h3h4 g7g6 d2e4 d6c7 c1g5 e7d5 g2f3 f7f5 e4c3 d5f4 g5f4 e5f4 d3d4 c7d6 a1d1 d6c7 d4d5 d7e5 d5c6 b7c6');
INSERT INTO evaluations VALUES(9788,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9788,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9788,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9788,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9788,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9788,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9788,36,'cp 2','d7d6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1e2 d8e7 b1d2 e8c8 e1c1 d8f8 c1b1 a7a6 d1f1 f6d7 f3h2 c8b8 f1f8 h8f8 h2g4 d7f6 h1f1 f6g4 e2g4 f8f1 d2f1 e7f7 g4e2 c6e7 f1d2 b8a8 d2f3 e7c6 e2f2 b7b5 f2g3');
INSERT INTO evaluations VALUES(8860,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8860,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8860,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8860,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8860,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8860,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8860,36,'cp 7','c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1e2 d8e7 b1d2 e8c8 e1c1 d8f8 d1f1 a7a6 f3h2 f6d7 f1f8 h8f8 h1f1 f8f1 d2f1 e7f7 f1g3 f7g6 e2f3 g6g5 h2f1 g5h4 f3h5 h4h5 g3h5 g7g6 h5g3 d7c5 c1b1 h6h5');
INSERT INTO evaluations VALUES(42013,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(42013,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(42013,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(42013,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(42013,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(42013,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(42013,36,'cp 46','b1d2 b4d6 f4g3 e8g8 f1d3 c7c5 c2c3 d8b6 g3d6 b6d6 f3e5 d6b6 a1b1 c8d7 a2a4 d7e8 f2f4 f6d7 b2b4 c5c4 d3c2 a7a5 b4b5 f7f6 e5f3 b6d6 e1g1 d7b6 d1e1 d6a3 g2g4 a3c3 b1c1 c3a3 e1h4 e8g6 c2g6 a3e3 g1g2 h7g6');
INSERT INTO evaluations VALUES(47949,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(47949,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(47949,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(47949,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(47949,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(47949,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(47949,36,'cp 56','b1c3 d6d5 c3d5 f6d5 e4d5 d8d5 f1d3 d5d6 d1e2 c8e6 e2e4 b8c6 c1f4 d6d7 d3c4 e8c8 c4e6 f7e6 c2c3 d7d5 e4c2 f8d6 f4d6 c7d6 e1g1 c8b8 c2b3 d5b3 a2b3 h8e8 a1e1 e6e5 d4d5 c6e7 c3c4 h7h6 f3d2 e8f8 f1f8 d8f8 d2e4 b8c7 c4c5 e7d5 e4d6 c7c6');
INSERT INTO evaluations VALUES(48367,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48367,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48367,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48367,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48367,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48367,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(48367,36,'cp 5','g1f3 f8e7 b2b3 b7b6 b1d2 e8g8 c2c4 c8b7 c1b2 c7c5 d4c5 b6c5 f1d3 f6e4 a1c1 b8d7 e1g1 f7f5 c4d5 e6d5 f3e5 d7e5 b2e5 a8c8 c1c2 e7f6 e5f6 d8f6 d2f3 f6b6 f3e5 b6e6 f2f4 c8c7 f1e1 f8c8 d1a1');
INSERT INTO evaluations VALUES(49992,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(49992,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(49992,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(49992,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(49992,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(49992,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(49992,36,'cp 32','d2d4 e5d4 d1d4 b8c6 d4e3 d8e7 f1g2 e7e3 c1e3 f8b4 e3d2 c6e5 a2a3 b4d6 b2b3 c7c6 f2f4 e5f7 d2e3 e8g8 e3d4 d6e7 e2e3 g7g6 g1e2 d7d6 e1g1 f8e8 a1d1 f6d7 e3e4 f5e4 c3e4 d7c5 d4c5 d6c5 h2h3 c8f5 e2c3 a8d8 g3g4 f5e6 f1e1 d8d1 e1d1 e8d8 d1d8 f7d8');
INSERT INTO evaluations VALUES(12325,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(12325,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12325,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12325,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(12325,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(12325,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(12325,36,'cp 18','d7d6 c1f4 e6e5 f4g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 f6g5 a3c2 c6e7 a2a4 b5a4 d5e7 d8e7 a1a4 e7b7 a4b4 b7c7 f1c4 a6a5 c4b5 c8d7 b4c4 c7b7 b5d7 b7d7 e1g1 e8g8 d1d3 d7b7 b2b4 a5a4 d3d6 a8c8 c4c8 f8c8 d6d3 c8d8 d3c4');
INSERT INTO evaluations VALUES(13909,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(13909,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(13909,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13909,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13909,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(13909,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(13909,36,'cp 25','c1f4 e6e5 f4g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 f6g5 a3c2 a8b8 g2g3 a6a5 h2h4 g5h6 c2e3 h6e3 d5e3 e8g8 f1e2 c8e6 e1g1 a5a4 e2g4 c6a5 g4e6 f7e6 d1d3 d8c7 a1d1 f8d8 d1d2 a5c4 e3c4 c7c4 d3c4 b5c4');
INSERT INTO evaluations VALUES(13910,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(13910,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13910,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(13910,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(13910,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(13910,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(13910,36,'cp 26','e6e5 f4g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 f6g5 a3c2 a8b8 a2a4 b5a4 c2b4 c6b4 c3b4 e8g8 a1a4 a6a5 b4b5 c8d7 d5c3 g8h8 h2h4 g5e7 f1c4 f7f5 e4f5 d7f5 h4h5 e7f6 e1g1 e5e4 c4d5 f6c3 b2c3 b8b5 h5h6 g7h6 d5e4 f5e4 a4e4 b5e5 e4d4 d8e7');
INSERT INTO evaluations VALUES(13840,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(13840,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(13840,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(13840,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(13840,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(13840,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(13840,36,'cp 35','f4g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 a8b8 a3c2 f6g5 a2a4 b5a4 c2b4 c6b4 c3b4 e8g8 a1a4 a6a5 b4b5 c8d7 d5c3 f7f5 e4f5 g8h8 f1e2 d7f5 e1g1 f5e6 c3d5 e6d7 d1b3 d7e6 b3d3 e6f5 d3c3 f5d7 c3a5 d8a5 a4a5 g5d8 a5a4 d8b6 a4a6 b6d4 b2b4 d7b5 e2b5 b8b5');
INSERT INTO evaluations VALUES(12348,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(12348,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(12348,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(12348,33,'cp -184',NULL);
INSERT INTO evaluations VALUES(12348,34,'cp -206',NULL);
INSERT INTO evaluations VALUES(12348,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(12348,36,'cp -200','h5h4 d7d5 e4d5 e6d5 c4b5 d8e7 e1d1 e7d8 f3e5 d8c7 h1e1 f8e7 e5c6 b7c6 h4g3 c7b7 e1e7 e8e7 b5f1 h7h5 g3e3 e7d8 e3c5 f6e4 c5a5 b7b6 a5b6 a7b6 f2f3 e4d6 a2a4 h5h4 b1c3 d6f5 d2d3 d5d4 c3e4 c6c5 d1e2 f7f6 e2f2 c8d7 b2b3 d8c7 c1d2 c7c6 f1e2 c6b7 f2f1 d7e6 f1f2 h8e8 e2d1');
INSERT INTO evaluations VALUES(12349,30,'cp -196',NULL);
INSERT INTO evaluations VALUES(12349,31,'cp -181',NULL);
INSERT INTO evaluations VALUES(12349,32,'cp -189',NULL);
INSERT INTO evaluations VALUES(12349,33,'cp -186',NULL);
INSERT INTO evaluations VALUES(12349,34,'cp -191',NULL);
INSERT INTO evaluations VALUES(12349,35,'cp -199',NULL);
INSERT INTO evaluations VALUES(12349,36,'cp -203','d7d5 e4d5 e6d5 c4b5 d8e7 e1d1 e7d8 f3e5 d8c7 h1e1 f8e7 e5c6 b7c6 h4g3 c7b7 e1e7 e8e7 b5f1 h7h5 g3e3 e7d8 e3c5 f6e4 c5e3 b7b6 e3b6 a7b6 f2f3 e4d6 b1c3 h5h4 a2a4 d6f5 d2d3 f7f6 b2b4 h8e8 c1d2 d8c7 a4a5 b6a5 d2f4 f5d6 b4a5 g7g5 f4d2 d6f5 c3a4 f5e3 d2e3 e8e3 d1d2 d5d4 a4c5');
INSERT INTO evaluations VALUES(17855,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(17855,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(17855,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(17855,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(17855,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(17855,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(17855,36,'cp 98','d2d4 e5d4 f3d4 c6e5 c4b3 f8c5 d4f3 d8f6 f3e5 f6e5 b1c3 g8f6 d1f3 c5d6 g2g3 e8g8 c1f4 e5e7 f1e1 d6f4 g3f4 d7d6 e4e5 d6e5 f4e5 f6h7 c3d5 e7g5 f3g3 g5g3 f2g3 h7g5 d5c7 a8b8 e1e3 c8g4 a1f1 g4h3 f1d1 b8c8 c7d5 f8e8 c2c3 h3g4 d1f1 g4h3');
INSERT INTO evaluations VALUES(17862,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(17862,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(17862,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(17862,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(17862,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(17862,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(17862,36,'cp 103','e5d4 f3d4 c6e5 c4b3 f8c5 d4f3 d8e7 f3e5 e7e5 b1c3 g8f6 d1f3 d7d6 c1f4 e5h5 f3g3 h5g4 e4e5 g4g3 f4g3 d6e5 f1e1 e8g8 g3e5 c7c6 e5f6 g7f6 c3e4 c5e7 a1d1 g8g7 h2h3 a6a5 a2a4 b7b5 e4g3 f8e8 d1d6 b5a4 b3a4 e7d6 e1e8 d6g3 f2g3 c6c5 a4c6');
INSERT INTO evaluations VALUES(17871,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(17871,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(17871,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(17871,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(17871,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(17871,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(17871,36,'cp 122','f3d4 c6e5 c4b3 f8c5 d4f3 d8e7 f3e5 e7e5 b1c3 g8f6 g1h1 d7d6 f2f4 e5d4 d1d4 c5d4 e4e5 f6g4 c3d5 e8g8 d5c7 a8b8 e5e6 f7e6 h2h3 g4f2 h1h2 d6d5 f1e1 f2e4 c7d5 e4c5 d5e7 g8f7 e7c8 f8c8 c2c3 c5b3 a2b3 d4f2 e1e2 f2h4 g2g3 h4e7 c1e3 c8d8 h2g2 d8d3 g2f3 b8d8 b3b4');
INSERT INTO evaluations VALUES(17867,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(17867,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(17867,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(17867,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(17867,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(17867,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(17867,36,'cp 132','c6e5 c4b3 f8c5 d4f3 d8f6 f3e5 f6e5 b1c3 g8f6 g1h1 b7b6 f2f4 e5d4 d1d4 c5d4 e4e5 d4c3 b2c3 f6g8 c1a3 a6a5 f4f5 a5a4 b3d5 a8a5 c3c4 g8e7 a1d1 c7c6 a3e7 e8e7 f5f6 g7f6 e5f6 e7f8 d5e4 h6h5 e4f5 f8e8 d1e1 e8d8 e1e7 h8e8 e7f7');
INSERT INTO evaluations VALUES(26351,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(26351,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(26351,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(26351,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(26351,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(26351,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(26351,36,'cp 209','d4d5 g4f3 g2f3 c6b8 h2h4 h7h5 c1e3 g7g6 b1c3 b8d7 e3d4 h8h7 d4e3 f8g7 f1h3 g7c3 b2c3 g8f6 e1g1 e8f8 d1e2 a8c8 c3c4 c8c7 f1b1 b7b6 a2a4 d7e5 b1c1 f6d7 a4a5 d7c5 a5b6 a7b6 c1b1 f8g8 b1b6 h7g7 f3f4 e5d7 b6c6 c5e4');
INSERT INTO evaluations VALUES(26348,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(26348,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(26348,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(26348,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(26348,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(26348,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(26348,36,'cp 221','g4f3 g2f3 c6b8 h2h4 h7h5 b1c3 g7g6 c1e3 b8d7 e3d4 h8h7 d4e3 f8g7 f1h3 g7c3 b2c3 g8f6 d1e2 e8f8 e1g1 d8c7 a1c1 a8e8 g1h1 d7c5 f1g1 e7e5 d5e6 c5e6 e2d2 f8g8 c1d1 f6d7 d2d6 c7d6 d1d6 d7e5 h1g2 e5c4 d6d7 c4e3 f2e3 e6c5 d7c7');
INSERT INTO evaluations VALUES(31053,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(31053,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(31053,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(31053,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(31053,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(31053,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(31053,36,'cp 102','c6d5 d1b3 g8f6 b3b7 b8d7 g1f3 a8b8 b7a7 e7e6 e2e3 b8a8 a7b7 a8b8 b7a6 f8b4 c1d2 e8g8 a6a4 b4e7 f1b5 d7b6 a4d1 b6c4 d1c1 f5d3 b5c4 d5c4 f3e5 d8b6 e5d3 c4d3 a1b1 f8c8 e1g1 e7b4 c1d1 b6a6 a2a4 b4c3 d2c3 f6e4 h2h4 e4c3 b2c3 b8b1 d1b1 c8c3 a4a5 c3c8 f1d1 a6a5 b1d3');
INSERT INTO evaluations VALUES(2482,30,'mate 3',NULL);
INSERT INTO evaluations VALUES(2482,31,'mate 3',NULL);
INSERT INTO evaluations VALUES(2482,32,'mate 3',NULL);
INSERT INTO evaluations VALUES(2482,33,'mate 3',NULL);
INSERT INTO evaluations VALUES(2482,34,'mate 3',NULL);
INSERT INTO evaluations VALUES(2482,35,'mate 3',NULL);
INSERT INTO evaluations VALUES(2482,36,'mate 3','f3d5 d8d5 c4d5 c8e6 d5e6');
INSERT INTO evaluations VALUES(5940,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(5940,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(5940,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(5940,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5940,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(5940,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5940,36,'cp 14','g8e7 d2d4 e5d4 f3d4 e8g8 c1e3 c5d4 e3d4 f7f5 f2f3 f5e4 b5c6 e7c6 c3e4 c8f5 f1e1 d6d5 e4c5 b7b6 c5e6 f5e6 e1e6 d8d7 e6e2 c6d4 d1d4 c7c5 d4d3 a8e8 a1e1 e8e2 e1e2 d5d4 d3e4 f8d8 b2b3 b6b5 e4d3 d8c8 c2c4 b5c4 d3c4 d7f7 c4d3 c5c4 b3c4 f7c4 d3e4 c4c6 e4d4');
INSERT INTO evaluations VALUES(8721,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(8721,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8721,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(8721,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8721,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8721,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(8721,36,'cp 13','d2d4 e5d4 f3d4 e8g8 c1e3 c5d4 e3d4 f7f5 f2f3 f5e4 b5c6 e7c6 c3e4 c8f5 f1e1 d8d7 d4c3 d7f7 d1d2 h7h6 e1e2 a8e8 a1e1 f5d7 b2b3 a7a6 h2h3 c6e7 e4g3 e7d5 e2e8 f8e8 e1e8 d7e8 c3a1 d5f4 g3e2 f4e2 d2e2 e8c6 a1d4 f7f5 d4b2');
INSERT INTO evaluations VALUES(14936,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(14936,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(14936,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(14936,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(14936,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(14936,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(14936,36,'cp 203','a7a6 b5c6 b7c6 d5c6 e5e4 d1e2 d8e7 f3d4 e7e5 d4b3 f8b4 d2d4 e5b5 c1d2 b4c3 d2c3 e8g8 e2b5 a6b5 b3c5 c8e6 c5e6 f7e6 c3b4 f8d8 b4c5 a8a6 c2c3 g8f7 a2a3 a6c6 a3a4 b5a4 a1a4 e6e5 e1g1 f6d5 f1e1 d8b8 c5a3 e5d4 a4d4');
INSERT INTO evaluations VALUES(19834,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(19834,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19834,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(19834,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(19834,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19834,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19834,36,'cp 1','g8f6 b1c3 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 e1g1 e7e6 f1d1 f8e7 a1c1 e8g8 b2b3 d8a5 c3a4 f8d8 c1c4 h7h5 h2h4 a5d5 f3e2 d5f5 g2g3 d8d7 a4c3 g7g6 e2e5 c8d8 e5f5 g6f5 c4c6 d7d4 d1d4 d8d4 c6c7');
INSERT INTO evaluations VALUES(13126,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13126,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13126,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13126,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13126,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(13126,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13126,36,'cp 9','b1c3 c8g4 c1g5 d5c4 f1c4 e7e6 d4d5 e6d5 c4d5 f8e7 d5c6 b7c6 e1g1 d8d1 f1d1 h7h6 g5e3 e8g8 a1c1 a7a5 h2h3 g4f3 g2f3 f8e8 g1g2 e7b4 c3e4 f6d5 c1c6 d5e3 f2e3 f7f5 e4g3 e8e3 d1d7 a8e8 a2a3 b4f8 g3f5 e3e2 g2g3 e2b2 d7c7');
INSERT INTO evaluations VALUES(36414,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36414,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36414,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36414,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(36414,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(36414,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36414,36,'cp 21','h2h3 c7c5 d4c5 e7c5 b1c3 a7a6 a2a3 f8e8 f1e1 c8e6 b2b4 c5a7 c1e3 a7e3 e1e3 b8c6 c3e2 d8c7 d1d2 e6d7 a1e1 e8e3 d2e3 h7h6 g1f1 a8e8 e3f4 c7f4 e2f4 g8f8 f4e2 g7g6 c2c3 f6e4 e1c1 c6e5 f3e5 e8e5 e2d4 e5e8');
INSERT INTO evaluations VALUES(39263,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(39263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39263,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(39263,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(39263,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39263,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39263,36,'cp -12','f3g3 g8f6 g3e5 c5e7 e5g3 e8g8 d2d3 d7d5 b1c3 b8c6 f1e2 c6d4 e2d1 d5e4 d3e4 b7b5 g1f3 b5b4 f3d4 d8d4 c3b5 d4e4 c1e3 a8b8 d1f3 e4c2 g3c7 c2c7 b5c7 c8g4 e1e2 e7d6 h1c1 d6e5 c7a6 b8b5 c1c4 g4f3 e2f3 e5b2 a1b1 b5f5 f3e2 b2c3 a6b4 c3b4 b1b4 f6d5 e3c5 f8e8 c4e4');
INSERT INTO evaluations VALUES(44739,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(44739,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44739,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(44739,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44739,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44739,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44739,36,'cp 0','f8g7 h2h3 a7a6 a2a4 e8g8 e1g1 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(10458,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(10458,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10458,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(10458,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(10458,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10458,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10458,36,'cp 0','f2e3 d7d5 e4d5 f6d5 d1d2 d8d6 d3d4 e5d4 f3d4 d5c3 d2c3 c6d4 c3d4 c8e6 f1b5 e8f8 d4d6 c7d6 h1f1 f8e7 b5d3 g7g5 e1d2 h7h5 a2a4 a8c8 a4a5 h5h4 b2b4 c8c7 f1f2 c7c8');
INSERT INTO evaluations VALUES(29591,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(29591,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29591,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29591,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(29591,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(29591,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29591,36,'cp 27','c2c4 c7c6 g1f3 e7e6 d1b3 d8b6 c4c5 b6b3 a2b3 b8a6 e1g1 h7h6 h2h4 f6d7 c1d2 a6c7 b1c3 a7a6 f1e1 f8e7 b3b4 e8c8 e2e4 d5e4 c3e4 d7f6 e4c3 h8g8 d2e3 f6d5 f3e5 d5b4 e5f7 b4c2 f7d8 c2e1');
INSERT INTO evaluations VALUES(32079,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(32079,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(32079,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(32079,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(32079,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(32079,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(32079,36,'cp 86','h5g6 b1c3 c7c6 d1e2 b8d7 h3h4 h7h5 g4g5 b7b5 c4b3 a7a5 a2a3 d7c5 d2d4 c5b3 c2b3 b5b4 e2c4 a8c8 a3b4 g6e4 c3e4 d6d5 c4e2 d5e4 e2e4 f8b4 c1d2 g8e7 d2b4 a5b4 d4e5 e7g6 a1d1 d8e7 e1g1 e7e6 f3d4 e6g4 e4g4 h5g4 f1e1 g6h4 e5e6 f7e6 e1e6 e8f7 e6d6');
INSERT INTO evaluations VALUES(50494,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(50494,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(50494,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(50494,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(50494,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(50494,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(50494,36,'cp -57','c4b3 b7b5 d2d3 c8b7 c1f4 f8e7 g1e2 g8f6 b1d2 d7d5 c2c3 d5e4 d3e4 c5c4 b3c2 b8d7 e1g1 e8g8 e2g3 d8b6 a1d1 g7g6 f3e2 e6e5 f4e3 b6a5 a2a4 a8d8 g1h1 d7c5 f2f4 e5f4 e3f4 c5e6 f4e5 f6d7 e5d4 e6d4');
INSERT INTO evaluations VALUES(57306,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57306,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(57306,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(57306,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57306,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57306,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57306,36,'cp 0','f3d4 d7d5 c2c4 d5c4 e2e3 a7a6 f1c4 c7c5 d1a4 b8d7 d4f3 f8d6 e3e4 g8e7 e4e5 d6c7 e1g1 e8g8 f1d1 e7c6 e5e6 f7e6 f3g5 d8e7 g5e6 c7h2 g1h2 e7h4 h2g1 h4f2 g1h1 f2h4 h1g1');
INSERT INTO evaluations VALUES(57283,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57283,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57283,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57283,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57283,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57283,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57283,36,'cp 0','d7d5 c2c4 d5c4 e2e3 a7a6 f1c4 c7c5 d1a4 b8d7 d4f3 f8d6 e3e4 g8e7 e4e5 d6c7 e1g1 e8g8 f1d1 e7c6 e5e6 f7e6 f3g5 d8e7 g5e6 c7h2 g1h2 e7h4 h2g1 h4f2 g1h1 f2h4 h1g1');
INSERT INTO evaluations VALUES(7223,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(7223,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(7223,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(7223,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(7223,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(7223,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(7223,36,'cp 29','d2d4 f6d7 d4e5 d7e5 c1f4 e5g6 f4d6 d8d6 d1d6 c7d6 c3a4 e8d8 a4b6 a8b8 a2a4 h8e8 f1e1 c8e6 a4a5 d8c7 b2b4 g6f8 f3d4 f8d7 b6a4 g7g5 f2f3 h7h5 g1f2 h5h4 e1e3 b8d8 a1d1 e6c4 d4f5 d6d5 a4c5 d7c5 b4c5 d5e4');
INSERT INTO evaluations VALUES(7217,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(7217,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(7217,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(7217,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(7217,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(7217,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7217,36,'cp 21','f6d7 d4e5 d7e5 c1f4 e5g6 f4d6 c7d6 d1d2 e8g8 f3d4 f7f5 e4f5 g6h4 c3e4 d6d5 e4g3 h4f5 a1e1 d8d6 c2c3 c6c5 d4f3 c8d7 g3f5 f8f5 e1e3 d7b5 f1e1 a8f8 e3e6 d6f4 d2e3 f4e3 e6e3 f8d8 h2h3 d5d4');
INSERT INTO evaluations VALUES(9430,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9430,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9430,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9430,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(9430,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9430,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9430,36,'cp 0','a2a3 b4c3 d2c3 e8g8 f1e2 f8e8 e1g1 d6d5 f1e1 a7a5 h2h3 d8d6 e2f1 h7h6 b2b3 d5d4 c3d2 c8e6 f3h4 g8h7 d1e2 d6c5 e2d1 c5d6');
INSERT INTO evaluations VALUES(9432,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9432,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(9432,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9432,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9432,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9432,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9432,36,'cp 0','b4c3 d2c3 e8g8 f1e2 f8e8 e1g1 d6d5 f1e1 a7a5 h2h3 d8d6 e2f1 c8e6 b2b3 h7h6 e4d5 e6d5 c3b2 a8d8 f3d2 d5e6 d1f3 e6d5 f3d1');
INSERT INTO evaluations VALUES(15244,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(15244,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(15244,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(15244,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(15244,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(15244,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(15244,36,'cp 232','f3e5 d5c3 d2c3 d8d1 e1d1 c8d7 e5d7 e8d7 h1e1 f8c5 d1e2 a7a6 e1d1 c5d6 b5a4 a8e8 e2f1 d7c8 c1d2 e8e7 a4c6 b7c6 d1e1 h8d8 e1e7 d6e7 d2e3 f7f6 a1e1 e7d6 e1d1 d8e8 h2h3 a6a5 a2a4 h7h5 b2b3 c6c5 c3c4 h5h4 e3d2');
INSERT INTO evaluations VALUES(15242,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(15242,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(15242,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(15242,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(15242,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(15242,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(15242,36,'cp 221','d5c3 d2c3 d8d1 e1d1 c8d7 e5d7 e8d7 h1e1 f8d6 d1e2 a8e8 e2f1 e8e1 f1e1 g7g6 h2h3 a7a6 b5e2 f7f5 e1f1 c6d8 c1d2 d8e6 a1d1 h8e8 b2b4 d7c8 a2a4 d6f4 d2e1 f4d6 c3c4 e6f4 e2f3 d6e5 e1d2 f4e6 c2c3 e5f4 c4c5 f4d2 d1d2 c7c6');
INSERT INTO evaluations VALUES(15261,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(15261,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(15261,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(15261,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(15261,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(15261,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(15261,36,'cp 226','d2c3 d8d1 e1d1 c8d7 e5d7 e8d7 h1e1 f8d6 d1e2 a8e8 e2f1 e8e1 f1e1 a7a6 b5e2 h8e8 e1f1 f7f5 c1d2 c6e5 a1d1 e5g4 d2f4 g7g6 d1d4 g4e5 h2h3 b7b5 g2g4 c7c5 d4d1 f5g4 h3g4 d7c7 b2b3 c7c6 f4e3 c5c4 e3f4 e5f7 f4c1 c4b3 a2b3');
INSERT INTO evaluations VALUES(17895,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(17895,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(17895,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(17895,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(17895,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(17895,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(17895,36,'cp 63','d7d6 d2d4 e5d4 f3d4 c8d7 d4c6 d7c6 a2a4 g8f6 b1c3 f8e7 f2f3 f6d7 c1e3 e7f6 d1d2 e8g8 b2b4 d7e5 c4b3 b7b5 a4a5 c6b7 a1d1 d8e7 c3d5 b7d5 d2d5 f8d8 e3f2 c7c6 d5d2 c6c5 f3f4 e5c4 b3c4 b5c4 b4c5 d6c5 d2e2 f6d4');
INSERT INTO evaluations VALUES(17793,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(17793,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(17793,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17793,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(17793,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(17793,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(17793,36,'cp 76','d2d4 e5d4 f3d4 c8d7 b1c3 g8f6 a2a4 f8e7 a4a5 e8g8 c1f4 f8e8 f1e1 e7f8 d4c6 d7c6 d1f3 g7g6 a1d1 f8g7 h2h3 f6h5 f4d2 d8e7 c4b3 g7e5 b3a2 e7f6 f3f6 h5f6 f2f3 f6d7 b2b4 b7b5 c3d5 c6d5 a2d5');
INSERT INTO evaluations VALUES(26357,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(26357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26357,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(26357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26357,36,'cp 3','g1f3 g8e7 e1g1 e7f5 b1d2 c5d4 f3d4 f5d4 c3d4 d8b6 d2b3 a7a5 a2a4 f8e7 f1e1 e8g8 e1e3 f7f6 h2h3 f6e5 e3e5 c6d7 c1d2 e7d6 e5e2 f8f7 d2c3 h7h6 b3c5 d6c5 d4c5 b6c5 d1d2 c5c4 e2e3 c4f4 c3e5 f4f5 d2e1 a8c8 e3f3 f5e4 e1e4 d5e4');
INSERT INTO evaluations VALUES(28023,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(28023,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(28023,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(28023,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(28023,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(28023,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(28023,36,'cp 109','e2e4 d7d6 f1d3 b8d7 e1g1 g8f6 h2h3 e7e5 d4d5 a7a5 a2a3 b7c8 a1b1 e8g8 b2b4 g8h8 d1c2 f6g8 c3b5 c8a6 c1b2 g8f6 f1e1 f6h5 g2g3 a6b5 c4b5 g7h6 b1d1 h5g7 g1g2 a5b4 a3b4 h8g8 d1a1');
INSERT INTO evaluations VALUES(29896,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(29896,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(29896,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(29896,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(29896,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(29896,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(29896,36,'cp 19','b8c6 h2h3 g4h5 f1e1 f8e7 a2a4 d8d7 c4e2 e8g8 d2d3 h5g6 f3h4 c6d4 h4g6 d4e2 e1e2 h7g6 d3d4 e5d4 d1d4 a7a6 c1f4 f8e8 a1d1 a8d8 d4c4 d7c8 c3d5 f6d5 d1d5 e7f6 b2b3 e8e7 f4g3 d8e8 f2f3 f6e5 g3e5 e7e5 d5e5 e8e5 c4d3');
INSERT INTO evaluations VALUES(32893,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(32893,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(32893,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(32893,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(32893,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(32893,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(32893,36,'cp 100','c1d2 e7e6 f1c4 c7c6 d1e2 g8f6 c3d5 a5d8 d5f6 g7f6 e1c1 b7b5 c4d3 d8d5 d3f5 d5f5 g1h3 f8d6 g2g4 f5d5 c1b1 h7h5 h3f4 d6f4 d2f4 b8d7 h2h4 e8c8 g4g5 d5f5 f4c1 d7b6 h1g1 d8d5 d1d3 f5g6 d3f3 f6f5 f3c3 c8b7 g1d1 h8d8 c3d3 a7a5 e2f3 a5a4 c1f4 b6c4 b2b3 c4b6');
INSERT INTO evaluations VALUES(32894,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(32894,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(32894,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(32894,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(32894,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(32894,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(32894,36,'cp 87','e7e6 f1c4 c7c6 d1e2 g8f6 c3d5 a5d8 d5f6 g7f6 e1c1 b7b5 c4d3 d8d5 d3f5 d5f5 g1h3 h8g8 f2f3 h7h5 h1g1 h5h4 h3f4 b8d7 e2e1 f8d6 e1h4 f5g5 h4g5 f6g5 f4d3 d6h2 g1h1 h2d6 g2g4 d7f6 h1h6 f6d5 c2c3 a7a5 d1h1 a5a4 h1h5 d5f4 d2f4 d6f4 d3f4 g5f4');
INSERT INTO evaluations VALUES(36616,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(36616,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(36616,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(36616,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(36616,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(36616,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(36616,36,'cp -35','b8d7 f4g3 f8e8 d3e2 e7e5 e1g1 d8e7 d4e5 d7e5 f3e5 d6e5 g3h4 e7b4 d2c4 c8f5 c2c3 b4c5 a2a4 f6e4 c4d2 e4d6 f1e1 c5b6 d1b3 f5e6 b3c2 a7a5 a1c1 d6f5 h4g3 f5g3 h2g3 b6c5 e1d1 a8d8 b2b3');
INSERT INTO evaluations VALUES(4078,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(4078,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(4078,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(4078,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(4078,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(4078,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4078,36,'cp -3','e1g1 d8e8 e3e4 e7e5 d4e5 d6e5 f4e3 f6g4 e3g5 f7f6 g5h4 d7c5 b2b4 c5e6 a2a4 g4h6 d3c4 g8h8 f1e1 h6f7 h4g3 h7h5 d2b3 c8d7 c4d5 c7c6 d5c4 g7h6 d1e2 a8d8 a4a5 h8h7 e1d1 e8e7 c4e6 d7e6 b3c5 e6c8 h2h3 b7b6');
INSERT INTO evaluations VALUES(44045,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(44045,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44045,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(44045,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(44045,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44045,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44045,36,'cp 25','d7d5 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 f3d4 c8d7 d4c2 f8d8 c2b4 c6b4 b5d6 d7c6 h4g3 f6e4 d1d4 e4d6 g3d6 b4d5 d6c5 d5b6 c4b3 c6b5 f1d1 d8d4 c5d4 a8d8');
INSERT INTO evaluations VALUES(57506,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(57506,31,'cp 239',NULL);
INSERT INTO evaluations VALUES(57506,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(57506,33,'cp 216',NULL);
INSERT INTO evaluations VALUES(57506,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(57506,35,'cp 227',NULL);
INSERT INTO evaluations VALUES(57506,36,'cp 215','d4e5 f5e4 b1c3 f8b4 c1d2 d7d5 e5d6 g8f6 d6c7 d8c7 c3b5 c7e7 d2b4 e7b4 c2c3 b4c5 d1d6 c5d6 b5d6 e8e7 d6c8 h8c8 f2f3 b8c6 f3e4 f6e4 f1d3 c8e8 g1f3 e4c5 e1c1 c5d3 d1d3 a8d8 d3d8 e8d8 c1c2 d8f8 h1d1 h7h6 d1d2 e7e6 b2b4 b7b5 c2b3');
INSERT INTO evaluations VALUES(1367,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(1367,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(1367,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(1367,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(1367,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(1367,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(1367,36,'cp 90','e1g1 c6e4 d2d4 e4c6 c2c4 g8f6 b1c3 d7d6 f1e1 e7e6 d4d5 c6d7 f3g5 e6e5 f2f4 f8e7 f4e5 d6e5 g5f3 e5e4 c3e4 f6e4 e1e4 e8g8 c1f4 e7d6 d1d2 d7c7 a1e1 f7f6 e4e7 c8d7 f4d6 c7d6 d2e2 a8e8 b2b3 e8e7 e2e7 d6e7');
INSERT INTO evaluations VALUES(2738,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(2738,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(2738,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(2738,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(2738,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(2738,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(2738,36,'cp 39','b3c2 c7c5 d2d4 d8c7 b1d2 c8d7 d2f1 a5c4 b2b3 c4b6 d1e2 c5c4 c1e3 a8b8 f1d2 a6a5 b3c4 b6c4 d2c4 b5c4 e1b1 d7b5 f3d2 f6d7 e2d1 e7f6 b1c1 f8c8 a1b1 g7g6 c2a4 b5a4 d1a4 b8b1 c1b1 e5d4 c3d4 c4c3 b1c1 f6g7 d2f3');
INSERT INTO evaluations VALUES(2741,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(2741,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(2741,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(2741,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(2741,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(2741,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(2741,36,'cp 40','c7c5 d2d4 d8c7 b1d2 c8d7 d2f1 a5c4 b2b3 c4b6 f1e3 c5c4 d1e2 f8c8 b3c4 b6c4 e3c4 b5c4 a1b1 a8b8 c1e3 c7a5 b1b8 c8b8 e2c4 b8c8 c4b3 a5c3 b3c3 c8c3 c2b3 e5d4 e3d4 c3c7 e4e5 d6e5 f3e5 d7e8 d4b6 c7b7 e5c4 g8f8 b6e3 e8d7 c4b6');
INSERT INTO evaluations VALUES(2826,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(2826,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(2826,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(2826,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(2826,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(2826,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(2826,36,'cp 60','d2d4 d8c7 b1d2 c5d4 c3d4 c8d7 d2f1 f8e8 b2b3 a8c8 f1e3 a5c6 c1b2 c7b7 a2a3 e7d8 d4e5 c6e5 f3d4 d8a5 b3b4 a5b6 d4f5 b6e3 e1e3 d7f5 e4f5 b7d5 c2b3 d5d1 a1d1 e5c4 b3c4 e8e3 c4f7 g8f7 f2e3 f6e4 d1d4 e4g3');
INSERT INTO evaluations VALUES(2825,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(2825,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(2825,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(2825,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(2825,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(2825,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(2825,36,'cp 30','d8c7 b1d2 c5d4 c3d4 a5c6 a2a3 c8d7 d4d5 c6b8 e1e3 f8c8 e3c3 c7b7 c3c8 b7c8 a3a4 b5a4 c2a4 a6a5 d1c2 c8c2 a4c2 b8a6 a1a5 a8c8 a5a6 c8c2 a6a8 e7f8 a8a1 d7b5 f3e1 c2c8 f2f3 f6d7 b2b3 f7f5 e4f5 d7f6 c1a3 f6d5 d2e4');
INSERT INTO evaluations VALUES(2181,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(2181,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(2181,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(2181,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(2181,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(2181,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(2181,36,'cp 38','b1d2 c5d4 c3d4 a5c6 a2a3 c8d7 d4d5 c6b8 e1e3 f8c8 e3c3 c7b7 c3c8 b7c8 a3a4 b5a4 c2a4 a6a5 d1c2 b8a6 c2c8 a8c8 d2b3 d7a4 a1a4 e7d8 c1g5 c8c2 g5e3 f6d7 a4a3 h7h6 b3a5 d8a5 a3a5 a6c5 f3d2 c2b2 a5a7 b2c2 f2f3 c2c1 g1f2 c1c2 f2g3 d7f6 e3c5');
INSERT INTO evaluations VALUES(2182,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(2182,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(2182,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(2182,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(2182,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(2182,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(2182,36,'cp 43','c8d7 d2f1 c5d4 c3d4 a8c8 f1e3 f8e8 b2b3 a5c6 c1b2 e7f8 d4d5 c6b4 c2b1 a6a5 a2a3 b4a6 b3b4 c8a8 d1d2 g7g6 b1d3 f8h6 b2c3 e8c8 a1c1 c7b6 d2b2 a5b4 a3b4 a6c7 c3d2 a8a4 d3c2 a4a8 c1b1 h6f4 e1c1 b6b7 b1a1 a8a1 c1a1 c8a8 a1a5 a8a5 b4a5 c7a6');
INSERT INTO evaluations VALUES(5837,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(5837,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(5837,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(5837,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(5837,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(5837,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(5837,36,'cp 89','b4d2 b1d2 f7f6 f1d3 c8d7 c2c3 f6e5 d4e5 g8h6 e1g1 e8g8 b2b4 a7a6 d1c2 h6f5 a2a4 d8e7 b4b5 c6a5 f3d4 c7c5 b5c6 b7c6 d3f5 e6f5 e5e6 c6c5 e6d7 c5d4 c3d4 e7d7 c2d3 a8c8 d2f3 a5c4 f1b1 h7h6 h2h3 d7e7 a4a5');
INSERT INTO evaluations VALUES(5843,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(5843,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(5843,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(5843,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(5843,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(5843,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(5843,36,'cp 111','b1d2 f7f6 f1d3 c8d7 e1g1 f6e5 f3e5 c6e5 d4e5 d8e7 c2c4 d5c4 d2c4 d7c6 a1c1 e8c8 c4a5 c6d5 d1a4 e7d7 d3b5 c7c6 b5c4 c8b8 c4d5 c6d5 a4d7 d8d7 a5b3 g8h6 b3d4 d7e7 f2f4 h8f8 g2g3 h6f5 d4f5 f8f5 g1f2');
INSERT INTO evaluations VALUES(5841,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(5841,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(5841,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(5841,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(5841,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(5841,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(5841,36,'cp 84','f7f6 f1d3 c8d7 e1g1 f6e5 f3e5 c6e5 d4e5 d8h4 g2g3 h4e7 c2c4 g8h6 c4d5 e6d5 d1c2 d7h3 f1e1 g7g6 a1c1 c7c6 f2f4 e8g8 d2f3 a7a5 f3g5 h3f5 a2a3 a5a4 g1g2 f5d3 c2d3 h6f5 d3c3 f5g7 h2h4 g7e6 e1e2 a8d8 c1e1 e7d7 g5e6');
INSERT INTO evaluations VALUES(6140,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(6140,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(6140,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(6140,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(6140,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(6140,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(6140,36,'cp 100','f1c4 e7e6 a2a4 g8e7 a4a5 e8g8 e1g1 a7a6 f1e1 h7h6 c1e3 b7b5 a5b6 d7b6 c4d3 c8b7 d1d2 g8h7 c3e2 f7f5 e4f5 e7f5 e2f4 d8f6 d2d1 f6f7 c2c3 b6d5 f4d5 b7d5 e3d2 a6a5');
INSERT INTO evaluations VALUES(6129,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(6129,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(6129,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(6129,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(6129,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(6129,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(6129,36,'cp 88','e7e6 a2a4 g8e7 a4a5 e8g8 e1g1 a7a6 f1e1 b7b5 a5b6 d7b6 c4d3 c8b7 h2h4 e7c6 d3f1 h7h6 c1e3 g8h7 h4h5 g6g5 d4d5 c6e5 f3e5 g7e5 e3d4 b6d7 b2b4 d8f6 c3e2 e5d4 d1d4 f6d4 e2d4 e6d5 e4d5 b7d5 a1a6 d7f6');
INSERT INTO evaluations VALUES(10064,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(10064,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(10064,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(10064,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(10064,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(10064,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(10064,36,'cp -79','e8g8 d2d3 h7h6 g5f3 d7d5 e4d5 f6d5 e1g1 d5b6 c4b5 f8e8 c1e3 c5d6 f1e1 a7a6 b5c6 b7c6 b1d2 c6c5 d3d4 e5d4 c3d4 b6d5 d2c4 d5e3 e1e3 e8e3 c4e3 c5d4 d1d4 c8b7 e3f5 d6f8 a1c1 c7c5 d4d8 a8d8 f5e3 f7f6 b2b3 d8d7 c1e1 g8f7 e3c4 d7d5 h2h3');
INSERT INTO evaluations VALUES(4299,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(4299,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(4299,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(4299,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(4299,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(4299,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(4299,36,'cp -75','d2d3 d7d5 e4d5 f6d5 e1g1 h7h6 g5f3 c8f5 f1e1 f8e8 b2b4 c5f8 d1b3 f5e6 a2a4 a7a6 b1a3 d5c3 b3c3 f8b4 c3b2 b4e1 f3e1 d8d4 a3c2 d4d6 c4e6 d6e6 c1d2 b7b6 h2h3 a8d8 d2c3 d8d7 c2e3 f7f6 e3f1 a6a5 f1e3 c6e7 a1c1 e8d8 b2e2 e7g6');
INSERT INTO evaluations VALUES(4300,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(4300,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(4300,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(4300,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(4300,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(4300,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(4300,36,'cp -52','d7d5 e4d5 f6d5 e1g1 h7h6 g5f3 d5b6 c4b5 c5d6 f1e1 f8e8 b1d2 a7a6 b5c6 b7c6 d2e4 c8g4 h2h3 g4h5 d3d4 e5d4 e4d6 d8d6 g2g4 e8e1 f3e1 h5g6 d1d4 d6e6 c1e3 b6c4 b2b3 c4e3 d4e3 e6e3 f2e3 a8d8 g1f2 d8d2 f2f3 f7f5 e1g2');
INSERT INTO evaluations VALUES(3717,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(3717,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(3717,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(3717,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(3717,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(3717,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(3717,36,'cp -75','e4d5 f6d5 e1g1 h7h6 g5f3 d5b6 c4b5 f8e8 f1e1 c8d7 b1d2 c5d6 d2c4 b6c4 b5c4 d8f6 c4d5 a8d8 d5e4 b7b6 b2b4 c6e7 c1e3 f6e6 d3d4 e7f5 d4e5 d6e5 f3e5 e6e5 e3d4 e5e6 e4f5 e6f5 e1e8 d8e8 d1f3 f5f3 g2f3 e8e2');
INSERT INTO evaluations VALUES(3709,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(3709,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(3709,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(3709,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(3709,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(3709,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(3709,36,'cp -56','f6d5 e1g1 h7h6 g5f3 d5b6 c4b5 c5d6 f1e1 c8g4 b1d2 a7a6 b5c6 b7c6 h2h3 g4h5 d2e4 f8e8 d3d4 e5d4 e4d6 d8d6 g2g4 h5g6 d1d4 e8e1 f3e1 b6d5 e1g2 c6c5 d4d1 d6e6 c1f4 a8e8 f4g3 f7f5 d1f3 d5f6 a1e1 f5g4 e1e6 g4f3 e6e8 f6e8');
INSERT INTO evaluations VALUES(3735,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(3735,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(3735,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(3735,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(3735,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(3735,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(3735,36,'cp -53','e1g1 h7h6 g5f3 d5b6 c4b5 c5d6 f1e1 f8e8 c1e3 a7a6 b5c6 b7c6 b1d2 c6c5 d3d4 e5d4 c3d4 c5d4 e3d4 e8e1 d1e1 c8b7 d2e4 d6f8 a1d1 d8e7 e4c5 b7f3 e1e7 f8e7 g2f3 a8d8 d1e1 e7c5 d4c5 d8d2 b2b4 d2a2 e1e8 g8h7 e8e7 b6d5 e7f7 h7g6');
INSERT INTO evaluations VALUES(10261,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10261,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10261,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10261,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(10261,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(10261,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10261,36,'cp -14','e8g8 b2b4 c5b6 c3a4 d7d5 e4d5 c6d4 a4b6 a7b6 c2c4 c8g4 f1e2 d4e2 d1e2 b6b5 e2e3 f8e8 f3d2 b5c4 d3c4 b7b5 e1g1 b5c4 d2c4 d8d5 c4a5 c7c5 h2h3 c5b4 a3b4 g4d7 f1e1 d7c6 f2f3 e5e4 c1b2 e4f3 e3f2 e8e1 a1e1 f3g2 b2f6 g7f6 a5c6 d5c6 f2g2 c6g2 g1g2');
INSERT INTO evaluations VALUES(4395,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(4395,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(4395,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4395,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(4395,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(4395,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4395,36,'cp 0','c1g5 h7h6 g5h4 f8e8 f1e2 a7a5 d1d2 c5e7 h2h3 d7d6 h4g3 a8b8 c3b5 a5a4 e1g1 f6d7 a1d1 d7c5 f1e1 c8d7 e2f1 e7f6 b5c3 d7e6 c3b5');
INSERT INTO evaluations VALUES(10877,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(10877,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(10877,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(10877,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(10877,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10877,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10877,36,'cp -4','h7h6 g5f6 e7f6 c3d5 g7g6 g2g3 d7d6 f1g2 f6g7 c2c3 e8g8 d5e3 c6e7 d3d4 f7f5 d4e5 d6e5 e1g1 d8d1 f1d1 f5f4 g3f4 e5f4 e3d5 e7d5 e4d5 c8f5 f3d4 g7e5 g2f3 e5d6 g1f1 f8f7 a2a4 a8e8 d4f5 f7f5 d1e1 e8e1 a1e1');
INSERT INTO evaluations VALUES(10825,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10825,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10825,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(10825,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10825,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10825,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10825,36,'cp 0','g5f6 e7f6 c3d5 g7g6 g2g3 f6g7 f1g2 d7d6 c2c3 e8g8 d5e3 c8e6 e1g1 a7a5 a2a4 a8b8 d1e2 d8d7 f3h4 d7d8');
INSERT INTO evaluations VALUES(12229,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(12229,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(12229,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(12229,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(12229,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(12229,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(12229,36,'cp 47','c4d5 f6d5 e2e4 d5c3 b2c3 h7h6 g5d2 e7e5 g1f3 f8d6 f1d3 d7f8 d2e3 f8g6 e1g1 e8g8 d1c2 c8e6 f1e1 f8e8 a2a4 b7b6 d3f1 d8d7 h2h3 a8d8 a4a5 b6b5 a5a6 f7f5 d4e5 d6e5 f3e5 g6e5 f2f4 e5c4 e4e5');
INSERT INTO evaluations VALUES(15167,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(15167,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(15167,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(15167,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(15167,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(15167,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(15167,36,'cp -69','f8c5 g1f3 d8e7 e1g1 e5e4 d3e4 d5e4 f3d4 c8d7 d4c6 d7c6 e2e3 h7h5 b1d2 e8c8 d1e2 c8b8 a2a3 h5h4 b3b4 c5d6 b4b5 c6d7 b2f6 e7f6 d2e4 f6h6 e4d6 h4g3 f2g3 h6h2 g1f2 d7h3 e2f3 h3g2 f3g2 d8d6 g2h2 h8h2 f2g1 d6d2 f1f7 a7a6 b5a6 d2g2 g1f1 g2c2 f1g1 c2g2 g1f1 g2g3 f7f8 b8a7');
INSERT INTO evaluations VALUES(9095,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(9095,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(9095,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(9095,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(9095,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(9095,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(9095,36,'cp -44','g1f3 d8e7 e1g1 e5e4 d3e4 d5e4 f3d4 c8d7 e2e3 h7h5 d4c6 d7c6 b1d2 e8c8 d1e2 h5h4 b2f6 g7f6 e2g4 c8b8 d2e4 h4g3 h2g3 h8h6 e4c5 e7c5 a1d1 d8h8 e3e4 a7a5 d1d2 f6f5 g4g5 h6h5 d2d8 b8a7 g5f6 h8h6 f6d4 c5d4 d8d4 f5e4 f1d1 f7f5 d4d8 b7b5 d1d4 b5b4');
INSERT INTO evaluations VALUES(15870,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(15870,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(15870,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(15870,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(15870,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(15870,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(15870,36,'cp 81','f3d5 d8d5 c4d5 f6g5 d5c6 b7c6 d2d3 g5g4 c1e3 a7a5 a2a4 f8e7 b1d2 c8e6 b2b3 h8f8 e1e2 e8d7 a1g1 h7h6 d2e4 f8f7 e3d2 e6f5 f2f3 g7g5 f3g4 f5g4 e2e1 g4f5 g1f1 f7f8 e1e2 f5g4 e2e3 g4f5 h2h3 h6h5 d2c3 d7e6 e3e2 f5e4 d3e4');
INSERT INTO evaluations VALUES(16530,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(16530,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(16530,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(16530,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(16530,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(16530,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(16530,36,'cp -67','d7d5 d2d4 c8g4 b1d2 d5e4 d4e5 c6e5 d2e4 d8e7 c1f4 g4f3 g2f3 e8c8 a1d1 e5f3 e2f3 e7e4 f3e4 d8d1 e1d1 f6e4 d1c2 f8d6 f4e3 h8f8 f1d3 f7f5 f2f3 e4c5 d3c4 f8e8 e3d4 g7g6 h2h4 c7c6 b2b4 c5e6 d4a7 b7b5 c4f1 e6f4 a7f2 c8c7');
INSERT INTO evaluations VALUES(15141,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(15141,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(15141,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(15141,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(15141,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(15141,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(15141,36,'cp -58','d2d4 c8g4 b1d2 d5e4 d4e5 e4f3 g2f3 g4e6 e5f6 d8f6 d2e4 f6e7 c1e3 e8c8 f1g2 e7e8 e1g1 f7f5 e4g3 e8f7 f1d1 e6d5 f3f4 d5g2 d1d8 c6d8 g1g2 a7a6 a1d1 g7g6 e2d3 f8g7 g3e2 d8c6 d3d5 f7d5 d1d5 h8d8 d5d8 c6d8 g2f1 c8d7 f1e1 d8e6 e1d2 g7f8 d2d3 e6c5 d3c2 f8d6 b2b4 c5e6 a2a4 d6e7 c2b3');
INSERT INTO evaluations VALUES(17432,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(17432,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(17432,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(17432,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(17432,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(17432,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(17432,36,'cp 25','d8c7 d4c6 b7c6 d1d3 d7d5 d3g3 c7g3 h2g3 g8f6 g3g4 f6e4 c3e4 d5e4 c1e3 c8b7 e1c1 a8d8 d1d8 e8d8 h1d1 d8c7 d1d4 f8c5 d4e4 c5e3 e4e3 c6c5 e3c3 c7b6 c3b3 b6c7 e2f3 b7d5 b3a3 c7b6 g4g5 a6a5 f3d5 e6d5 a3f3 h8f8 c1d2 b6c6 f3f4 f7f6 g5f6 f8f6 f4f6 g7f6 d2e3 c6d6 e3f4 d6e6 b2b3 f6f5 a2a3 d5d4');
INSERT INTO evaluations VALUES(1265,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(1265,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(1265,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(1265,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(1265,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(1265,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(1265,36,'cp 39','d4c6 b7c6 d1d3 g8f6 d3g3 c7g3 h2g3 d7d5 g3g4 f6e4 c3e4 d5e4 c1e3 c8b7 e1c1 a8d8 d1d8 e8d8 h1d1 d8c7 d1d4 f8c5 d4e4 c5e3 e4e3 c6c5 e3c3 c7b6 c3b3 b6c7 e2f3 b7f3 b3f3 h8f8 g4g5 f7f6 g5f6 f8f6 f3f6 g7f6 c1d2 c7d6 d2d3 d6d5 c2c4 d5d6 d3c3 f6f5 b2b4 c5b4 c3b4 e6e5 g2g3 h7h5 c4c5 d6d5');
INSERT INTO evaluations VALUES(21395,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(21395,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(21395,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(21395,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(21395,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21395,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(21395,36,'cp 23','f8d6 e2e3 g8f6 b1d2 e8g8 f1d3 d8e7 f3e5 f6d7 d2f3 f7f6 e5d7 c8d7 d1c2 g7g6 f4g3 e6e5 d4e5 c6e5 f3e5 d6e5 e1g1 a8e8 c3c4 c7c6 c4d5 c6d5 a1c1 e8c8 c2d2 d7e6 f1e1 f8d8 c1c8 d8c8 g3e5 f6e5');
INSERT INTO evaluations VALUES(5919,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(5919,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(5919,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(5919,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(5919,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(5919,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(5919,36,'cp 37','e2e3 g8f6 b1d2 e8g8 f3e5 f6d7 d2f3 f7f6 e5d7 c8d7 f4g3 d7e8 a2a3 c6e7 c3c4 c7c6 f1d3 d8d7 a1c1 e8h5 e1g1 a7a5 f1e1 a5a4 c4c5 d6c7 d3c2 g8h8 c2a4 c7g3 h2g3 e6e5 d1b3 e5e4 f3d2 e7f5 b3b4');
INSERT INTO evaluations VALUES(30095,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(30095,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(30095,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(30095,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(30095,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(30095,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30095,36,'cp 0','f8d6 f3e5 f6d7 e5d7 b8d7 f4d6 c7d6 b1c3 d8b6 d1c1 e8g8 f1d3 f5d3 c2d3 f8c8 c1d2 d7f6 e1g1 a7a6 a2a4 c8c6 a4a5 b6d8 f1c1 a8c8 c3e2 c6c1 a1c1 c8c1 e2c1 d8d7 c1e2 h7h5 d2b4 d7c7 b4c3 c7d7');
INSERT INTO evaluations VALUES(28452,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(28452,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(28452,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(28452,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(28452,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(28452,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(28452,36,'cp -12','f3e5 f6d7 e5d7 b8d7 f4d6 c7d6 b1c3 d8b6 d1c1 e8g8 f1d3 f5d3 c2d3 d7f6 c1d2 a8c8 e1g1 c8c6 c3e2 f8c8 f1c1 c6c1 a1c1 c8c1 d2c1 g7g6 a2a3 g8g7 c1d2 f6g8 d2c2 g8e7 e2c3 h7h5 g1f1 e7g8 c3e2 b6b5 f1e1 h5h4');
INSERT INTO evaluations VALUES(28461,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(28461,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(28461,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(28461,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(28461,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28461,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28461,36,'cp 0','f6d7 e5d7 b8d7 f4d6 c7d6 b1c3 d8b6 d1c1 e8g8 f1d3 f5d3 c2d3 d7f6 e1g1 a8c8 c1d2 h7h6 f1c1 c8c6 c3e2 f8c8 c1c6 c8c6 b2b4 f6d7 a2a4 a7a6 a1a2 c6c7 a2a1');
INSERT INTO evaluations VALUES(50189,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50189,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(50189,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50189,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(50189,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(50189,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(50189,36,'cp 23','c7c5 d4d5 f8g7 e2e4 e7e6 c1e3 d7d6 b1c3 a7a6 a2a4 e8g8 f1e2 d8e7 e3g5 e6d5 c4d5 f8e8 d1d2 e7c7 g5e3 b8d7 g1h3 c5c4 h3f2 d7c5 e2c4 c5e4 f2e4 f6e4 c3e4 c7c4 e1f2 c4b3 e3d4 b3d5 h1d1 f7f5 d4g7 d5d2 e4d2 g8g7 d2c4');
INSERT INTO evaluations VALUES(53122,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53122,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(53122,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53122,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(53122,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53122,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53122,36,'cp 0','g8f6 g1f3 g7g6 f3g5 e7e6 b1c3 b8c6 d4d2 f8g7 e1g1 e8g8 f1d1 d8a5 d2f4 h7h6 g5f3 g6g5 f4g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(18716,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(18716,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(18716,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(18716,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(18716,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(18716,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(18716,36,'cp -81','g8e7 c3e4 e7g6 e4d6 f8d6 e5d6 d8f6 h2h4 b7b6 c2c3 g6e5 d2d3 c8a6 c1e3 e8g8 f3e5 f6e5 d3d4 c5d4 c3d4 e5d5 f1a6 d5a5 d1d2 a5a6 h4h5 a6b5 h5h6 g7g6 d4d5 e6d5 a1d1 d5d4 e3d4 c6d4 d2d4 a8e8 e1d2 e8e2 d2c1 f8c8 c1b1 b5f5 b1a1 f5e5 d4e5 e2e5 f2f3');
INSERT INTO evaluations VALUES(52621,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(52621,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(52621,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(52621,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(52621,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(52621,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(52621,36,'cp 86','d2d4 g8f6 d4c5 b8c6 f1c4 d6c5 d1e2 f8g7 e1g1 e8g8 e4e5 f6d5 f1d1 e7e6 h2h4 d8c7 d1e1 h7h6 c4d3 a7a6 a2a4 b7b6 b1a3 c8b7 h4h5 c6a5 d3c2 d5e7 a3b1 g8h8 b1d2 g6g5 e2d3 e7f5 d2e4 g7e5 f3e5');
INSERT INTO evaluations VALUES(52614,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(52614,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(52614,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(52614,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(52614,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(52614,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(52614,36,'cp 81','g8f6 d4c5 b8c6 f1c4 f6e4 c4f7 e8f7 d1d5 f7g7 d5e4 c8f5 e4e2 d6c5 c1f4 e7e5 f3e5 d8d5 e2c4 d5c4 e5c4 a8e8 c4e3 f5d3 b1d2 c6d4 c3d4 c5d4 a1c1 d4e3 f2e3 f8b4 c1c7 g7f6 c7b7 b4c5 f4h6 h8g8 b7b3 c5e3 b3d3 e3h6 e1d1 e8d8 d2e4 f6e7 d3d8 g8d8 d1c2 d8d4 h1e1 e7e6 e4g5 e6f5 g5f3 d4e4 e1e4 f5e4');
INSERT INTO evaluations VALUES(11316,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(11316,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(11316,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(11316,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(11316,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(11316,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11316,36,'cp 52','f1d3 g7g6 h2h3 f8g7 g1f3 f6d7 e3e4 e7e5 f4g5 f7f6 e4d5 c6d4 f3d4 d7c5 d4b3 c5d3 d1d3 c8f5 d3c4 f6g5 b1d2 d8d6 e1c1 b7b5 c4c5 e8c8 g2g4 f5d3 c5e3 d6d5 f2f3 d5c6 b3c5 d3e2 e3e2 c6c5 d2b3 c5b6 e2e4 c7c6 h3h4 d8d1 c1d1 g7h6 h4g5 h8d8 d1c2');
INSERT INTO evaluations VALUES(11318,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(11318,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(11318,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(11318,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(11318,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(11318,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11318,36,'cp 38','g7g6 h2h3 c8f5 d3f5 g6f5 g1e2 e7e6 b1d2 h8g8 e1g1 f8d6 g1h1 d8d7 a2a4 f6e4 f4d6 d7d6 b2b4 c6e7 d2e4 f5e4 b4b5 d6b6 d1b3 c7c6 e2f4 a6b5 a4b5 a8a1 f1a1 b6b5 b3b5 c6b5 a1a7 b7b6 f4h5 g8g6 a7b7 g6h6 h5f4 e7g6');
INSERT INTO evaluations VALUES(12412,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12412,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(12412,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(12412,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(12412,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(12412,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(12412,36,'cp 16','b1d2 f8e7 g1f3 f6h5 d4c5 h5f4 e3f4 d8c7 g2g3 e7c5 f1d3 e8g8 e1g1 b7b6 a2a3 c8b7 f1e1 h7h6 d1e2 a7a5 e2e4 g7g6 e4e2 c6e7 b2b4 c5d6 d3e4 e7d5 e2d3 d6e7 a1c1 d5f6 e4b7 c7b7 f3e5 d7d6 e5g6 f7g6 d3g6 g8h8');
INSERT INTO evaluations VALUES(12411,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(12411,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12411,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(12411,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(12411,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12411,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(12411,36,'cp 11','f8e7 g1f3 e8g8 h2h3 d7d5 f1d3 e7d6 d4c5 d6c5 e1g1 c5d6 f4d6 d8d6 c3c4 f8d8 d1b3 b7b6 c4d5 d6d5 d3c4 d5d6 c4e2 c8b7 f1d1 a8c8 d2e4 d6e7 d1d8 c8d8 e4c3 h7h6 a1c1 c6b4 b3a4 a7a6 a2a3');
INSERT INTO evaluations VALUES(22957,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(22957,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(22957,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(22957,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(22957,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(22957,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(22957,36,'cp -49','g8f6 b1c3 f8b4 f2f3 d7d5 a2a3 b4c3 d2c3 d5e4 c1d2 d8d5 f1c4 d5e5 e1c1 e8g8 g1e2 e5c3 d2c3 c8e6 c4b5 e4f3 g2f3 f6d5 h1e1 c6e7 e2d4 c7c6 d4e6 f7e6 b5c4 g8f7 c3d4 b7b6 a3a4 e7f5 d4f2 g7g6 c4d3 f8d8 d3e4 d8d6 c2c4 d5e7 d1d6 f5d6');
INSERT INTO evaluations VALUES(16718,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(16718,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(16718,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(16718,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(16718,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(16718,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(16718,36,'cp -49','b1c3 f8b4 f2f3 d7d5 a2a3 b4c3 d2c3 d5e4 c1d2 d8d5 f1c4 d5e5 e1c1 e8g8 g1e2 e5c3 d2c3 c8e6 c4b5 e4f3 g2f3 f6d5 b5c6 b7c6 c3d2 f8e8 h1e1 f7f6 b2b3 d5b6 c2c4 a8d8 e2g3 g8f7 d2f4 d8d1 e1d1 e8c8 f4e3 c8b8 g3e2 b6d7');
INSERT INTO evaluations VALUES(16700,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(16700,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(16700,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(16700,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(16700,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(16700,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(16700,36,'cp -49','f8b4 f2f3 d7d5 a2a3 b4c3 d2c3 d5e4 c1g5 e4f3 g1f3 e8g8 g5f6 d8f6 c3f6 g7f6 e1c1 c6e5 f1d3 c8e6 d3e4 c7c6 f3d4 e6g4 d1e1 a8d8 h2h3 g4h5 d4f5 h5g6 b2b3 f8e8 h1f1 g6f5 e4f5 e5g6 e1e8 d8e8');
INSERT INTO evaluations VALUES(10491,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(10491,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(10491,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(10491,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(10491,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(10491,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(10491,36,'cp -48','f2f3 d7d5 a2a3 b4c3 d2c3 d5e4 c1g5 e4f3 g1f3 e8g8 g5f6 d8f6 c3f6 g7f6 e1c1 c6e5 f1d3 c8e6 d3e4 c7c6 f3d4 a8d8 b2b3 e6g4 d1d2 g4h5 d4f5 d8d2 c1d2 h5g6 h1f1 f8e8 f1f4 e8d8 d2e3 b7b6 e3e2 c6c5 f5g3 g8g7 h2h3 d8e8 e2d2');
INSERT INTO evaluations VALUES(25377,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(25377,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(25377,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(25377,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(25377,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(25377,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(25377,36,'cp 36','e1g1 f8e7 c3d5 e8g8 d5f6 e7f6 c4e6 f7e6 b2b4 c6b4 a1b1 b4c6 b1b7 d8c8 b7b2 c8a6 f3d2 f8b8 d1g4 c6d8 d2b3 a6a4 h2h4 a7a5 h4h5 a4e8 b2b1 d8c6 c2c3 a5a4 b3d2 h7h6 b1b8 a8b8 d2f3 g8h7 g2g3 c6d8 f3h2 b8b1 g4e2 h7h8 c1e3 b1f1 h2f1 d8f7 c3c4 f7g5');
INSERT INTO evaluations VALUES(25376,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(25376,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(25376,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(25376,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(25376,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(25376,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(25376,36,'cp 29','f8e7 c3d5 e8g8 a2a4 e6d5 e4d5 c6b4 d3d4 b4d5 d4e5 d6e5 f3e5 c7c6 g2g3 f8e8 e5g4 f6e4 d1f3 e7g5 g4e3 g5e3 c1e3 d8c7 f1e1 d5e3 f3e3 e4f6 e3c5 b7b6 c5a3 e8e1 a1e1 a8e8 e1e8 f6e8 c4f1 c7d7 a3e3 g7g6 b2b3 e8f6');
INSERT INTO evaluations VALUES(28582,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(28582,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(28582,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(28582,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(28582,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(28582,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(28582,36,'cp 112','b1c3 e7e6 e2e4 c5c4 g1f3 f8b4 c1g5 h7h6 g5f6 d8f6 d1d4 f6d4 f3d4 e8g8 d4c2 b4a5 c2e3 e6d5 e4d5 f8c8 e1c1 a5c3 b2c3 d7d6 e3f5 c8d8 d1d2 g8f8 f1e2 g7g6 f5d4 a6b7 e2c4 d8c8 c4b3 c8c3 c1b2 c3c5 h1c1 b8d7 c1c5 d7c5 d4c6 a8c8 c6b4 f8e7 a2a4');
INSERT INTO evaluations VALUES(28581,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(28581,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(28581,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(28581,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(28581,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(28581,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(28581,36,'cp 115','g7g6 e2e4 a6f1 e1f1 d7d6 g2g3 f8g7 g1f3 e8g8 f1g2 b8a6 c1g5 d8d7 h1e1 f8b8 d1d2 h7h6 g5f6 e7f6 a1b1 b8e8 a2a3 f6f5 e4f5 d7f5 e1e8 a8e8 b1e1 a6c7 e1e8 c7e8 a3a4 g7c3 b2c3');
INSERT INTO evaluations VALUES(28579,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(28579,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(28579,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(28579,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(28579,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(28579,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(28579,36,'cp 121','e2e4 a6f1 e1f1 d7d6 g2g3 f8g7 f1g2 b8a6 a2a4 e8g8 g1f3 d8c8 d1e2 a6c7 h1e1 c8a6 e2c2 f8b8 c1g5 h7h6 g5f6 g7f6 f3d2 e7e6 d5e6 c7e6 c3b5 e6d4 c2d3 f6g7 a1a3 h6h5 d2c4 d4b5 a4b5 a6b5 a3a8 b8a8 d3d5 a8b8 c4d6 b5d7 b2b3');
INSERT INTO evaluations VALUES(28576,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(28576,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(28576,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(28576,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(28576,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(28576,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(28576,36,'cp 105','a6f1 e1f1 d7d6 g2g3 f8g7 g1f3 e8g8 a2a4 d8c8 f1g2 b8a6 d1e2 a6c7 c1g5 c8a6 e2c2 f8b8 h1e1 h7h6 g5f6 g7f6 h2h4 b8b4 h4h5 g6h5 a1d1 a6b7 e4e5 d6e5 c2f5 b4d4 d5d6 c7e6 d6e7 f6e7 f5e5 a8d8 e5h5 e6g5 a4a5 d4d1 e1d1 d8d1 c3d1');
INSERT INTO evaluations VALUES(28575,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(28575,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(28575,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(28575,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(28575,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(28575,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(28575,36,'cp 132','e1f1 d7d6 g2g3 f8g7 g1f3 e8g8 f1g2 d8b6 a2a4 b8a6 a4a5 b6b7 f3d2 f6d7 d2c4 a6b4 d1e2 d7e5 c4b6 a8a7 h1d1 e5d7 b6d7 b7d7 c1f4 d7b7 e4e5 d6e5 f4e3 f8c8 e2b5 b7b5 c3b5 a7b7 b5c3 b4c2 a1a2 c2e3 f2e3 e5e4 c3e4 g7b2');
INSERT INTO evaluations VALUES(28574,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(28574,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(28574,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(28574,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(28574,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(28574,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(28574,36,'cp 125','d7d6 g2g3 f8g7 g1f3 b8a6 c1g5 e8g8 f1g2 d8b6 a1b1 a6c7 h1e1 h7h6 g5f6 g7f6 a2a4 f8b8 d1d2 f6g7 h2h4 b6b7 h4h5 g6g5 f3h2 c7a6 h2g4 c5c4 g4e3 a6c5 e3c4 b7b4 d2e2 c5a4 e4e5 d6e5 c4e5 a4c3 b2c3 b4d6 b1d1 g7e5 e2e5 d6e5 e1e5');
INSERT INTO evaluations VALUES(28585,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(28585,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(28585,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(28585,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(28585,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(28585,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(28585,36,'cp 109','g2g3 f8g7 g1f3 e8g8 a2a4 d8c8 f1g2 b8a6 d1e2 a6c7 c1g5 h7h6 g5f6 g7f6 h2h4 c8a6 e2c2 a8b8 h4h5 g6g5 h1e1 b8b4 a4a5 a6b7 e4e5 d6e5 c2d3 e5e4 c3e4 c7d5 e4f6 e7f6 e1d1 f8d8 a5a6 b7c6 g2g1');
INSERT INTO evaluations VALUES(28584,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(28584,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(28584,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(28584,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(28584,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(28584,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(28584,36,'cp 99','f8g7 g1f3 b8a6 f1g2 e8g8 a2a4 d8c8 d1e2 a6c7 h1e1 c8a6 e2c2 f8b8 a1b1 e7e6 d5e6 c7e6 b2b3 b8b4 c1a3 b4b6 b3b4 e6d4 f3d4 c5d4 a4a5 a8c8 a5b6 c8c3 c2d2 c3a3 d2d4 f6d7 d4d6 a6b5 d6d5 b5d5 e4d5 d7b6 d5d6');
INSERT INTO evaluations VALUES(28194,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(28194,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(28194,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(28194,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(28194,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(28194,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(28194,36,'cp 129','f1g2 e8g8 g1f3 b8a6 a1b1 d8c8 a2a3 a6c7 h1e1 c8a6 d1c2 f8c8 c1g5 c7b5 c3b5 a6b5 f3d2 a8a4 e1c1 c8b8 d2c4 b5b3 g5f6 g7f6 f2f4 f6d4 c2b3 b8b3 e4e5 a4a8 c1c2 g8f8 e5d6 e7d6 c4d6 a8a3 b2a3 b3b1 d6c4 b1a1 c2e2 a1c1 c4d6 c1b1 f4f5 g6f5 d6f5');
INSERT INTO evaluations VALUES(28195,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(28195,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(28195,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(28195,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(28195,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(28195,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(28195,36,'cp 109','e8g8 g1f3 b8a6 a1b1 d8a5 c1f4 a6c7 e4e5 f6h5 f4g5 f7f6 e5d6 e7d6 g5d2 f6f5 a2a4 g7c3 b2c3 a5a4 d2h6 a4d1 h1d1 f8b8 b1b8 a8b8 c3c4 h5f6 f3d2 c7e8 d1a1 f6g4 h6f4 h7h6 h2h3 g6g5 h3g4 g5f4 g4f5 f4g3');
INSERT INTO evaluations VALUES(27752,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(27752,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(27752,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(27752,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(27752,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(27752,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(27752,36,'cp 111','g1f3 b8a6 c1g5 d8d7 h1e1 h7h6 g5f6 g7f6 e4e5 f6g7 d1e2 d7b7 a1d1 f8b8 d1d2 a6c7 h2h4 c7b5 c3b5 b7b5 b2b3 b5e2 e1e2 g8f8 d2c2 b8b4 e2e3 f8g8 e3c3 b4b6 c2e2 b6b4 c3c4 g7e5 f3e5');
INSERT INTO evaluations VALUES(27751,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(27751,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(27751,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(27751,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(27751,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(27751,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(27751,36,'cp 101','b8a6 a2a4 d8c8 d1e2 a6c7 c1g5 h7h6 g5f6 g7f6 h2h4 f6g7 h1e1 c8b7 a1b1 f8b8 e2c2 b7b3 c2d2 b3c4 e4e5 b8b4 e1e4 c4a6 b1e1 a6b7 e5d6 b4b2 d2f4 g7c3 d6c7 c3e1 e4e1 b7d5 e1e7 b2b6 f4h6 b6e6 e7e6 d5e6 h6f4 e6c6');
INSERT INTO evaluations VALUES(29542,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(29542,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(29542,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(29542,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(29542,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(29542,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(29542,36,'cp -7','e7e6 e2e3 f8d6 f3e5 f6d7 e5d7 b8d7 f4d6 c7d6 b1c3 e8g8 a2a4 f5g6 f1d3 a8c8 a4a5 a7a6 e1g1 d7f6 d1d2 d8e7 c3a4 c8c6 a4b6 f6d7 b6d7 e7d7 c2c3 g6d3 d2d3 f8c8 d3e2 g7g6 e3e4 d5e4 e2e4 b7b5 a5b6 c6b6');
INSERT INTO evaluations VALUES(30099,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30099,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30099,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30099,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30099,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(30099,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(30099,36,'cp 0','e2e3 f8d6 f3e5 f6d7 e5d7 b8d7 f4d6 c7d6 b1c3 f5g6 a2a4 a8c8 f1d3 e8g8 a4a5 c8c6 e1g1 a7a6 d1d2 d7f6 f1c1 d8e7 c3a4 f8c8 a4b6 c8c7 b6a8 c7c8');
INSERT INTO evaluations VALUES(38820,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(38820,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(38820,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(38820,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(38820,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(38820,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(38820,36,'cp 89','b2c3 g8e7 c4d5 e6d5 e2e3 e8g8 f1e2 f8e8 e1g1 b8d7 c3c4 d7f6 a2a4 c8g4 c1a3 f6e4 d1c2 c7c6 a1b1 b7b6 f1c1 e7g6 c4d5 c6d5 h2h3 g4f3 e2f3 g6h4 f3e4 d5e4 c2c7 d8g5 c7g3 g5g3 f2g3 h4f5');
INSERT INTO evaluations VALUES(54678,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(54678,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(54678,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(54678,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(54678,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(54678,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(54678,36,'cp 210','f1c4 e7e5 g1f3 b8c6 e1g1 h7h5 b2b4 g7g5 a2a4 a7a6 b1c3 h5h4 c3d5 g5g4 f3e1 e5d4 c1b2 f8g7 b2d4 c6e5 c4b3 g8e7 d5e3 g7h6 f2f4 g4f3 e1f3 h6e3 d4e3 c8g4 h2h3 g4f3 f1f3 e5f3 d1f3 d8d7 f3f6 h8f8 f6h4 e8c8 b4b5 a6a5 b5b6 c8b8 g1h1 b8a8 e3g5 d8e8');
INSERT INTO evaluations VALUES(20166,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(20166,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(20166,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(20166,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(20166,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(20166,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(20166,36,'cp -52','c8d7 b5c6 b7c6 e1g1 c6c5 d4d3 g8e7 c2c4 a8b8 b1c3 d5c4 d3c4 e7f5 f1d1 d8c7 c3e2 f8e7 e2g3 e8g8 g3f5 e6f5 c1g5 d7c6 g5e7 c7e7 c4c3 c6f3 c3f3 b8b2 f3f5 b2e2 d1e1 e2e1 a1e1 c5c4 f5c2 e7c5 c2c3 f8d8 c3e3 d8c8 e3c5 c8c5 f2f4 c4c3');
INSERT INTO evaluations VALUES(25918,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(25918,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(25918,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(25918,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(25918,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(25918,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(25918,36,'cp -28','b5c6 b7c6 e1g1 g8e7 c2c4 e7f5 d4f4 f8c5 b2b3 e8g8 c1b2 d8e7 b1c3 c5b6 a1d1 f8d8 c3e2 d5c4 b3c4 c6c5 e2g3 h7h6 b2c3 b6c7 f1e1 a8b8 d1c1 a7a5 f3d2 f5d4 d2e4 d7c6 g3h5 d4f5 f4g4 g8h8 c1d1 d8d1');
INSERT INTO evaluations VALUES(25906,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(25906,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(25906,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(25906,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(25906,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(25906,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(25906,36,'cp -26','b7c6 e1g1 g8e7 c2c4 e7f5 d4f4 h7h6 b2b3 d8a5 c1d2 a5a6 b1c3 f8e7 f4g4 d5c4 f1c1 a8d8 c3e4 c6c5 d2e3 d7c6 e4d6 e7d6 e5d6 f5e3 g4g7 h8f8 f2e3 c6f3 g2f3 a6d6 c1c4 d6d2 g7g3 d2e3 g1h1 e8e7 a1e1 e3g5 g3c7 d8d7 c7c5 g5c5 c4c5 f8d8 c5a5 d7d2 a5a7 e7f6 a7c7 d2a2');
INSERT INTO evaluations VALUES(25824,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(25824,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(25824,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(25824,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(25824,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(25824,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(25824,36,'cp -19','c2c4 c6c5 d4d3 d8a5 c1d2 a5a6 a2a4 a6c4 d3c4 d5c4 b1a3 d7c6 a3c4 g8h6 d2h6 g7h6 h1g1 h8g8 f3d2 g8g4 e1e2 g4h4 h2h3 h4d4 f2f3 a8b8 b2b3 f8e7 g1c1 b8d8 c1d1 c6d5 a1c1 e8f8 c4a5 f8g7 d2c4 f7f6 e5f6 g7f6 c4e3 d4d1 e3d1 f6g5');
INSERT INTO evaluations VALUES(21353,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21353,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(21353,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(21353,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(21353,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(21353,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(21353,36,'cp 16','f8d6 b5d6 c7d6 g1f3 f7f6 e2e3 g8e7 f4g3 e6e5 f1e2 e7f5 e1g1 e8g8 a1c1 c8e6 c2c4 a8c8 c4d5 e6d5 d4e5 d6e5 a2a3 d5e6 d1a4 d8b6 e2c4 e6c4 a4c4 g8h8 b2b4 c6e7 c4d3 c8d8 d3b1 h7h6 b1b3 b6a6 f1d1 d8d1 c1d1');
INSERT INTO evaluations VALUES(5915,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(5915,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(5915,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(5915,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(5915,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(5915,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5915,36,'cp 14','b5d6 c7d6 g1f3 f7f6 e2e3 g8e7 f1e2 e6e5 f4g3 e7f5 e1g1 e8g8 c2c4 d5c4 d4e5 d6e5 e2c4 g8h8 a2a3 d8e7 d1c2 c8e6 c4d3 g7g6 a1c1 a8c8 c2b1 c6a5 f3d2 f5g3 h2g3 a5b3 d2b3 e6b3 d3e4 h8g7 c1c8 f8c8 f1c1 c8c1 b1c1 b3e6 e4f3 b7b6 b2b4');
INSERT INTO evaluations VALUES(5891,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(5891,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(5891,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(5891,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(5891,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(5891,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(5891,36,'cp 0','c7d6 g1f3 f7f6 f4g3 e6e5 e2e3 g8e7 f1e2 e7f5 e1g1 c8e6 c2c4 e8g8 c4d5 e6d5 d4e5 d6e5 d1b1 e5e4 f3d2 f5g3 h2g3 d8e7 f1d1 f6f5 d2b3 d5b3 a2b3 g8h8 b1a2 a7a6 a2a3 c6b4 a3a5 b4c6');
INSERT INTO evaluations VALUES(5602,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(5602,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(5602,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5602,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(5602,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(5602,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5602,36,'cp 7','g1f3 f7f6 f4g3 g8e7 e2e3 e6e5 f1e2 e7f5 e1g1 e8g8 c2c4 d5c4 d4e5 d6e5 e2c4 g8h8 a2a3 d8e7 d1c2 c8e6 c4d3 g7g6 a1c1 a8c8 c2b1 c6a5 f3d2 f5g3 h2g3 a5b3 d2b3 e6b3 c1c8 f8c8 f1c1 c8c1 b1c1 h8g7 c1c3 b3e6 d3c4 e6c4 c3c4');
INSERT INTO evaluations VALUES(24920,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24920,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24920,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(24920,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(24920,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(24920,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(24920,36,'cp 9','e7e6 b1d2 f8d6 h2h3 g4f3 d2f3 d6f4 e3f4 d8d6 d1d2 c6b4 e1g1 b4d3 c2d3 f6g8 b2b4 g8e7 a2a4 e8g8 g2g3 a7a5 b4b5 c7c6 a1b1 b7b6 d2b2 f8c8 f1c1 a8b8 f3e5 f7f6 e5c6 e7c6 b5c6 c8c6 b2b5 b8c8 c1c6 c8c6');
INSERT INTO evaluations VALUES(25211,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(25211,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(25211,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(25211,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(25211,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(25211,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(25211,36,'cp 11','b1d2 f8d6 h2h3 g4f3 d2f3 d6f4 e3f4 d8d6 d1d2 c6b4 e1g1 b4d3 c2d3 f6g8 b2b4 g8e7 g2g3 e8g8 g1g2 b7b6 f1c1 f8c8 a1b1 h7h6 d2b2 a7a5 b4b5 c7c6 c1c2 a8b8 b1c1 c6b5 b2b5 c8c2 c1c2 b8c8 c2c8 e7c8 b5e8 d6f8');
INSERT INTO evaluations VALUES(25212,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(25212,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(25212,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(25212,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(25212,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(25212,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(25212,36,'cp 2','f8d6 c2c3 d6f4 e3f4 c6e7 h2h3 g4f5 d3f5 e7f5 d1e2 e8g8 g2g4 f5d6 f3e5 f6e4 h3h4 c7c5 d2e4 d6e4 d4c5 d8c7 e5d3 e4c5 d3c5 c7c5 e1c1 b7b5 e2e3 f8c8 e3c5 c8c5 d1d4 a8c8 a2a3 g7g6 h4h5 c5c4 c1d2 c4d4 c3d4 c8c4 d2e3');
INSERT INTO evaluations VALUES(41229,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41229,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41229,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(41229,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(41229,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41229,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(41229,36,'cp 0','c7c5 a2a4 e8g8 e1g1 a7a6 c3d5 f6d5 c4d5 d8c7 h2h3 e7e6 d5c4 b8c6 c2c3 a8b8 c1f4 b7b5 a4b5 a6b5 c4a2 c8b7 a2b3 b8a8 d1d2 c6a5 b3c2 a5c6');
INSERT INTO evaluations VALUES(41365,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(41365,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(41365,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(41365,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(41365,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(41365,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(41365,36,'cp 47','c1e3 e8g8 d1d2 a7a6 e4e5 d6e5 d4e5 d8d2 e3d2 f6d7 f2f4 b7b5 h3h4 b8c6 h4h5 f8d8 g1f3 c8b7 e1c1 d7c5 f1e2 c5e6 d2e3 d8d1 h1d1 a8d8 a2a4 d8d1 c1d1 b5a4 h5g6 h7g6 c3a4 g6g5 f3g5 e6g5 a4c5 b7c8 f4g5 a6a5 c2c3 c6e5 e3f4 c8f5 d1c1 f5g6 g2g4 f7f6 g5f6 g7f6 c5b3');
INSERT INTO evaluations VALUES(41352,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(41352,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(41352,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(41352,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(41352,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(41352,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(41352,36,'cp 60','e8g8 d1d2 a7a6 e4e5 d6e5 d4e5 d8d2 e3d2 f6d7 f2f4 b7b5 f1e2 c8b7 e2f3 b7f3 g1f3 f8d8 e1c1 b8c6 d2e3 e7e6 c3e4 d7b6 e3b6 c7b6 g2g3 h7h6 h1e1 g7f8 d1d8 a8d8 e1d1 d8d1 c1d1 f8e7 c2c3 g8f8 d1d2 b5b4 f3d4 b4c3 d2c3 c6d4 c3d4 h6h5 g3g4 h5g4 h3g4 f8g8 d4c3 g8f8');
INSERT INTO evaluations VALUES(36648,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(36648,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(36648,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(36648,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(36648,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(36648,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(36648,36,'cp 56','d1d2 a7a6 e4e5 d6e5 d4e5 d8d2 e3d2 f6d7 f2f4 b8c6 e1c1 d7b6 g1f3 f7f6 e5f6 e7f6 d1e1 c8d7 c3e4 a8d8 e4c5 d7c8 c5e6 c8e6 e1e6 g8f7 e6e1 f8e8 f1d3 g7f8 a2a3 e8e1 h1e1 c6d4 f3d4 d8d4 d2e3 d4d7 g2g3 f6f5 g3g4 b6d5 e3d2 f5g4 h3g4');
INSERT INTO evaluations VALUES(36649,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(36649,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36649,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(36649,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(36649,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(36649,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(36649,36,'cp 42','a7a6 e4e5 d6e5 d4e5 d8d2 e3d2 f6d7 f2f4 b8c6 e1c1 d7b6 g1f3 f7f6 e5f6 e7f6 d1e1 c8d7 c3e4 a8d8 e4c5 d7c8 c5e6 c8e6 e1e6 g8f7 e6e1 f8e8 f1d3 g7f8 a2a3 f8c5 g2g4 c5f2 e1e8 d8e8 g4g5 c6d4 f3d4 f2d4 h3h4 b6a4 c2c3 d4e3 d2e3 e8e3 d3c2 a4c5 h4h5 f7g7');
INSERT INTO evaluations VALUES(42756,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(42756,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(42756,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(42756,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(42756,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(42756,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(42756,36,'cp 29','c2c3 b4a5 g1f3 d7d6 c1g5 c7c6 b1d2 a5c7 a2a4 b8d7 e1g1 h7h6 g5h4 d7f8 d3d4 f8g6 h4g3 e8g8 f1e1 f8e8 d1b3 d8e7 f3h4 g6h4 g3h4 a8b8 b3c2 c8e6 c4e6 e7e6 d4e5 d6e5 h4f6 e6f6 g2g3 a7a5 a1d1 b7b5 a4b5 c6b5');
INSERT INTO evaluations VALUES(42755,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(42755,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(42755,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(42755,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(42755,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(42755,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(42755,36,'cp 28','b4a5 c1g5 c7c6 b1d2 a5c7 g1f3 d7d6 e1g1 b8d7 f1e1 h7h6 g5h4 d7f8 h4g3 f8g6 d3d4 e8g8 a2a4 f8e8 d1b3 d8e7 f3h4 g6h4 g3h4 a8b8 h2h3 g7g5 h4g3 f6h5 d2f1 h5f4 a1d1 h6h5 f1e3 e7f6 b3c2 c7b6 d4e5');
INSERT INTO evaluations VALUES(42845,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(42845,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(42845,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(42845,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(42845,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(42845,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(42845,36,'cp 28','g1f3 d7d6 e1g1 c7c6 c1g5 b8d7 f1e1 a5c7 b1d2 h7h6 g5h4 d7f8 d3d4 f8g6 h4g3 e8g8 a2a4 f8e8 d1b3 d8e7 f3h4 g6h4 g3h4 a8b8 h2h3 g7g5 h4g3 f6h5 d2f1 e7f6 d4e5 d6e5 f1e3 e8d8 b3c2 h5f4 e1d1 d8d1 a1d1 h6h5 b2b4 c8e6 c4e6 f4e6 e3c4 h5h4 g3h2 b8d8 d1d8 f6d8');
INSERT INTO evaluations VALUES(45191,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(45191,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(45191,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(45191,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45191,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45191,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(45191,36,'cp 0','g7g6 g1e2 f8g7 c3d5 f6d5 e4d5 h7h5 h2h3 c8f5 d2d4 b8d7 e1g1 c5c4 e2c3 d7f6 a2a4 e8g8 c1g5 a8b8 f1e1 a7a6 a4a5 f8e8 a1a4 b7b5 a5b6 d8b6 g5c1 b8c8 a4a2 c8b8');
INSERT INTO evaluations VALUES(44762,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44762,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(44762,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44762,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(44762,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(44762,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(44762,36,'cp 13','g1e2 f8g7 e1g1 e8g8 c3d5 e7e5 d5f6 d8f6 c2c3 f6e7 f2f4 e5f4 g3f4 b8c6 d2d3 f7f5 c1e3 c8e6 e2g3 f5e4 d3e4 g8h8 d1c2 a7a5 a2a4 a8d8 a1e1 e6c4 f1f2 e7f7 e4e5 c4b3 c2e2 b3c4 e2g4 d6e5 e3c5 e5f4 c5f8 d8f8');
INSERT INTO evaluations VALUES(51871,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(51871,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(51871,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51871,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(51871,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(51871,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(51871,36,'cp 42','d2d4 e5d4 d1d4 g8f6 g1f3 d7d5 c2c4 b8d7 c4d5 f8c5 d4c4 f6d5 e1g1 e8g8 e2e4 d5c7 b1c3 a7a5 c1g5 d8e8 a1d1 c7e6 g5c1 b7b6 f1e1 b6b5 c4e2 c8a6 e2c2 f7f6 c1e3 c5e3 e1e3 d7e5 c3e2 a8d8 e2f4 d8d1 c2d1 e6f4 g3f4');
INSERT INTO evaluations VALUES(7125,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(7125,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(7125,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(7125,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(7125,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(7125,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(7125,36,'cp 53','a4c6 b7c6 d2d4 e5d4 f3d4 c8d7 c2c4 h7h6 b1c3 e8g8 c1f4 f8e8 h2h3 c6c5 d4f5 d7f5 e4f5 d8d7 d1d3 f6h7 c3d5 e7f6 d5f6 h7f6 f4d2 e8e1 a1e1 a8e8 b2b3 e8e1 d2e1 d6d5 c4d5 f6d5 e1d2 d7d6 d3c4 d5b6 c4e2 d6d4 e2e3 b6d7 g1h2 d7f6 e3d4');
INSERT INTO evaluations VALUES(12407,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12407,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(12407,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(12407,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(12407,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(12407,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(12407,36,'cp 12','d7d5 c4d5 e6d5 f1b5 c5d4 e3d4 f8d6 e1g1 e8g8 c1g5 c8g4 b5e2 f8e8 h2h3 g4f3 e2f3 h7h6 g5h4 g7g5 h4g3 d6g3 f2g3 d8d6 d1d2 e8e7 a1d1 a8e8 d2f2 a7a6 f3e2 f6e4 c3e4 d5e4 d4d5 c6e5 f2d4 e5d7 a2a3 e4e3 d4c3 d6e5 c3e5 d7e5');
INSERT INTO evaluations VALUES(48402,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(48402,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(48402,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(48402,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(48402,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(48402,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48402,36,'cp -25','g2g3 f8e7 b2b3 c7c5 a2a4 b7b6 f1g2 c8b7 d1e2 b8c6 e1g1 f6g4 d2d3 e8g8 b1a3 a7a6 h2h3 g4h6 e3e4 d8c7 c1b2 a8d8 g1h2 f7f6 a1e1 e7d6 f3d2 c6d4 e2f2 d5e4 d3e4 e6e5 f4f5 b6b5 a4b5 a6b5 b2d4 c5d4 a3b5');
INSERT INTO evaluations VALUES(52048,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52048,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52048,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(52048,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(52048,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52048,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(52048,36,'cp 12','g8f6 b1c3 d7d5 g1f3 f8e7 c4d5 f6d5 f1d3 b8c6 e1g1 e8g8 f1e1 e7f6 d3e4 c6e7 f3e5 c8d7 d1h5 g7g6 h5f3 d7c6 e4d5 e7d5 c1h6 f6g7 h6g7 g8g7 e5c6 b7c6 c3e4 d8b6 b2b3 b6d4 a1c1 a8c8 c1c5 d4b4 e1d1 b4b6 h2h4 g7g8');
INSERT INTO evaluations VALUES(44280,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44280,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44280,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(44280,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44280,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44280,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(44280,36,'cp 24','b1c3 d7d5 g1f3 f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 h7h6 g5h4 d8d6 h4g3 d6d8 b5c6 b7c6 f3e5 d8b6 a2a4 f6e4 f1e1 a7a5 c3c4 e4g3 c4c5 b6c7 h2g3 a8b8 e1e3 b8b4 d1d3 f8e8 a1e1 c8e6 d3d2 b4a4');
INSERT INTO evaluations VALUES(44278,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(44278,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(44278,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44278,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(44278,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44278,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44278,36,'cp 17','d7d5 g1f3 f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 d8d6 g5f6 d6f6 d1b3 f6d6 f1e1 c8d7 a2a4 a8c8 h2h4 f8e8 a4a5 e8e1 f3e1 d7e6 e1d3 b7b6 a5b6 a7b6 a1a6 c6a5 b3b4 d6c7 d3e5 c7c3 b4c3 c8c3');
INSERT INTO evaluations VALUES(20822,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(20822,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(20822,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(20822,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(20822,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(20822,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20822,36,'cp 0','e7f6 g1e2 c8f5 e2g3 f5g6 h2h4 d8d7 h4h5 g6f7 c1f4 e8c8 b1c3 f8b4 d1d3 g8e7 a2a3 b4d6 f4d6 d7d6 e1c1 c8b8 d3d2 h8e8 f1e2 c6a5 c3b5 d6b6 b5c3 b6d6');
INSERT INTO evaluations VALUES(22244,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(22244,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(22244,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(22244,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22244,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22244,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22244,36,'cp 0','g1e2 c8g4 h2h3 g4h5 b1c3 d8d7 c1e3 e8c8 d1d2 g8e7 g2g4 h5f7 e1c1 h7h5 g4g5 h5h4 g5f6 g7f6 e2f4 e7f5 f1e2 f5e3 f2e3 f8h6 d1f1 h8g8 e2g4 f6f5 g4f3 g8g3 c3e2 g3g8');
INSERT INTO evaluations VALUES(37791,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(37791,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(37791,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(37791,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(37791,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(37791,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(37791,36,'cp 71','d4d5 g7c3 b2c3 d7d6 e2e4 e7e5 a2a4 a7a5 f1d3 b8d7 g1f3 h7h6 c1e3 d8e7 f3d2 e8d8 f2f3 d8c7 d1b3 b7b6 e1c1 c8a6 h2h4 e7f8 g2g4 g8e7 h4h5 g6g5 d2f1 a8b8 b3a2 f7f6 a2e2 e7c8 c1c2 f8e7 e2d2');
INSERT INTO evaluations VALUES(37838,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(37838,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(37838,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(37838,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(37838,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(37838,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(37838,36,'cp 51','g7c3 b2c3 d7d6 e2e4 e7e5 f1d3 b8d7 g1f3 h7h6 a2a4 d8e7 f3d2 a7a5 f2f3 b7b6 d2f1 c8a6 d1e2 e8c8 c1d2 d8f8 e1c1 c8c7 f1e3 g8f6 g2g3 h6h5 h2h4 f6e8 d3c2 h8h7 d2e1 f8g8 h1g1 g8h8 e2d2 e8f6 c2d3 e7e8');
INSERT INTO evaluations VALUES(37720,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(37720,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(37720,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(37720,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(37720,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(37720,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(37720,36,'cp 56','b2c3 e7e5 a2a4 d7d6 e2e4 a7a5 g2g3 b8d7 g1f3 g8f6 d1c2 h7h6 c1e3 d7b6 f1e2 c8h3 f3d2 d8e7 h1g1 e8d8 g3g4 h6h5 g4g5 f6g4 c2b3 d8c7 e3c5 d6c5 g1g3 g4f2 e1f2 h3d7 f2g1 h5h4 a1b1 h4g3');
INSERT INTO evaluations VALUES(37719,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(37719,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(37719,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(37719,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(37719,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(37719,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(37719,36,'cp 52','e7e5 e2e4 d7d6 g2g3 d8e7 g1f3 g8f6 d1c2 e8d8 a2a4 a7a5 f3d2 b8d7 d2b1 d7b6 b1a3 c8d7 a3b5 d7b5 c4b5 h7h6 h2h4 d8c7 f1h3 b6c4 c2d3 c4b6 c3c4 b6d7 b5b6 d7b6 e1g1 g6g5 a1b1 g5g4 d3b3 a8a6 h3g2 f6d7 f1e1 f7f6 c1d2 h6h5 g2f1');
INSERT INTO evaluations VALUES(40933,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(40933,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(40933,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40933,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(40933,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(40933,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(40933,36,'cp 9','c2c4 h7h6 g5h4 e8g8 b1c3 b7b6 f1e2 d5c4 e2c4 b8d7 e1g1 c8b7 d1e2 f6d5 h4e7 d5c3 e7d8 c3e2 c4e2 f8d8 f1c1 c7c5 d4c5 d7c5 b2b4 c5e4 c1c7 e4d6 f3e5 d8c8 c7d7 c8d8 d7d8 a8d8 a1c1 d8c8 c1c8 d6c8 f2f3 c8d6');
INSERT INTO evaluations VALUES(42821,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(42821,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(42821,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(42821,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42821,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(42821,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(42821,36,'cp 37','e4e5 f6h5 a2a3 b4c3 d2c3 d7d6 f1e2 c8g4 e5d6 e8g8 e1g1 c7d6 h2h3 d8b6 f1f2 g4d7 d1d4 b8c6 d4b6 a7b6 f3d4 h5g3 c1f4 c6d4 f4g3 d4e2 f2e2 d6d5 e2e7 d7c6 a1e1 f8e8 g1f2 f7f6 g3c7 e8e7 e1e7 b6b5 c7d6');
INSERT INTO evaluations VALUES(42720,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(42720,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(42720,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42720,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(42720,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(42720,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(42720,36,'cp 23','f6h5 c3d5 b8c6 f1b5 b4a5 e1g1 e8g8 f3e1 d8g5 d2d4 d7d6 b5e2 c8e6 e2h5 e6d5 c1f4 g5h4 g2g3 h4e7 e5d6 c7d6 c2c3 a8e8 h5f3 d5f3 d1f3 e7d7 e1d3 h7h6 f3d5 a5b6 a1d1 b6c7 a2a4 e8e6 d1e1 f8e8 e1e6 e8e6 f4d2 c7b6');
INSERT INTO evaluations VALUES(45062,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(45062,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(45062,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(45062,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(45062,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(45062,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(45062,36,'cp 107','e2e4 d5c3 b2c3 e7e5 g1f3 e5d4 f1c4 b7b5 c4b3 a7a5 f3d4 a5a4 b3c2 f8c5 e1g1 e8g8 c1e3 c5d4 c3d4 c8e6 f1e1 f7f5 d1d2 b8a6 f2f3 f5e4 f3e4 e6c4 a1d1 c4a2 c2a4 a2c4 a4c2 a6c7 c2b1 d8e7 g1h1 e7h4 h2h3 c4b3 d1c1 b3c4 e1d1 c7e6');
INSERT INTO evaluations VALUES(45064,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(45064,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(45064,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(45064,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(45064,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(45064,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(45064,36,'cp 98','d5c3 b2c3 e7e5 g1f3 e5d4 f1c4 b7b5 c4b3 a7a5 f3d4 a5a4 b3c2 f8c5 c1e3 e8g8 e1g1 c5d4 c3d4 c8e6 f1e1 f7f5 d1d2 f5e4 c2e4 d8d7 e3f4 e6d5 e4d3 f8f7 f4g3 b8a6 e1e3 a6c7 a1e1 c7e6 d3b1 a8d8 g3e5 d8f8 f2f3 g7g6 e5g3 e6g7 e3e5');
INSERT INTO evaluations VALUES(45080,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(45080,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(45080,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(45080,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(45080,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(45080,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(45080,36,'cp 93','b2c3 e7e5 g1f3 e5d4 f1c4 b7b5 c4b3 a7a5 f3d4 a5a4 b3c2 f8c5 e1g1 e8g8 c1e3 c5d4 c3d4 c8e6 f1e1 f7f5 d1d2 b8a6 f2f3 f5e4 f3e4 e6c4 a1c1 d8e7 c2b1 e7f7 h2h3 a6c7 g1h1 a8d8 e1d1 d8d7 c1c3 c7e6 b1d3 c4d3 c3d3 h7h6 d2c2 e6f4 e3f4 f7f4 c2c6 d7c7');
INSERT INTO evaluations VALUES(45079,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(45079,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(45079,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(45079,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(45079,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(45079,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(45079,36,'cp 95','e7e5 g1f3 e5d4 f1c4 b7b5 c4b3 a7a5 f3d4 a5a4 b3c2 f8c5 e1g1 e8g8 c1e3 c5d4 c3d4 c8e6 f1e1 f7f5 d1d2 b8a6 f2f3 f5e4 f3e4 e6c4 a1c1 d8d6 c2d3 c4d3 d2d3 h7h6 a2a3 a8e8 e3f2 f8f2 g1f2 d6h2 e1e3 a6c7 c1c6 c7e6 c6e6 e8e6 e4e5 e6e8 e3f3 h2h4 f2g1 h4e1 g1h2 e1h4 f3h3 h4f2 d3b5');
INSERT INTO evaluations VALUES(49606,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(49606,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(49606,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(49606,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(49606,34,'cp -138',NULL);
INSERT INTO evaluations VALUES(49606,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(49606,36,'cp -135','d7d5 b1d2 d5e4 d2e4 c8f5 e4g3 f5g6 h2h4 h7h5 g1f3 b8c6 f1e2 f6g4 e1g1 f8e7 c1g5 f7f6 g5d2 d8d7 d1e1 e8g8 b3b4 a8e8 a2a4 c6d8 a4a5 d8e6 d2e3 a7a6 e1d2 f6f5 e3g5 e7g5 h4g5 d7e7 b4b5 a6b5');
INSERT INTO evaluations VALUES(48816,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(48816,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(48816,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(48816,33,'cp -120',NULL);
INSERT INTO evaluations VALUES(48816,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(48816,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(48816,36,'cp -120','b1d2 d5e4 d2e4 c8f5 e4g3 f5g6 g1f3 h7h5 h2h4 f6g4 f1e2 b8c6 e1g1 f8e7 c1g5 f7f6 g5d2 d8d7 a1b1 c6d8 a2a4 d8e6 d1c1 e8g8 c1b2 a7a5 b3b4 a5b4 d2b4 e7b4 b2b4 b7b6 b4b5 d7b5 a4b5 a8a2 f1c1 f8a8 d3d4 a8a4 d4e5 g4e5');
INSERT INTO evaluations VALUES(55581,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(55581,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(55581,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(55581,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(55581,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(55581,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(55581,36,'cp 177','d4e5 b8c6 c1g5 f7f6 e5f6 g8f6 e2e3 h7h6 g5f6 d8f6 b1c3 f6f7 c3b5 e8d8 b5d4 c6d4 f3d4 f7d5 d1f3 d5f3 g2f3 g7g5 h2h4 f8g7 a1d1 g7f6 f1h3 c8h3 h1h3 d8d7 a2a4 a8e8 e1e2 g5h4 d4f5 f6b2 h3h4 h8g8 h4h5');
INSERT INTO evaluations VALUES(55781,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(55781,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(55781,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(55781,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(55781,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(55781,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(55781,36,'cp -32','e6d5 f1g2 g8f6 b2b3 a7a5 g1f3 f8e7 c1b2 b8a6 a2a3 c7c6 d1c2 a6c5 b1c3 c8g4 e1g1 e8g8 e2e3 g4h5 d2d3 h5f3 g2f3 d8b6 f3d1 d5d4 e3d4 c5e6 c3a4 b6d8 a4c5 e6d4 c2c4 e7c5 c4c5 d4e6 c5f5 f8e8 b2f6 d8f6 f5f6');
INSERT INTO evaluations VALUES(56292,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(56292,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(56292,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(56292,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(56292,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(56292,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(56292,36,'cp -17','f1g2 f8e7 g1f3 g8f6 e1g1 e8g8 d2d3 h7h6 c1f4 d5d4 e2e4 d4e3 f4e3 c7c6 b1c3 f8e8 a2a3 b8a6 f3e5 c8e6 f2f4 a6c7 e3f2 e7d6 d3d4 d8c8 f1e1 c7d5 c3a4 d5e7 a4c3 f6g4 e5g4 e6g4 d1c2 c8d8 c2b3');
INSERT INTO evaluations VALUES(56293,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(56293,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(56293,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(56293,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(56293,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(56293,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(56293,36,'cp -33','g8f6 d2d3 f8e7 g1h3 b8a6 e1g1 c7c6 b1d2 e8g8 a2a3 a6c7 e2e3 a7a5 h3f4 c7e6 f4e6 c8e6 b2b3 b7b5 d1c2 d8b6 c1b2 f6d7 a3a4 f8b8 f1c1 b5a4 a1a4 c6c5 b2c3 e7f8 h2h3 h7h6');
INSERT INTO evaluations VALUES(48750,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48750,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(48750,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(48750,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(48750,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(48750,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(48750,36,'cp -22','g1f3 f8e7 e1g1 e8g8 d2d3 h7h6 c1f4 d5d4 e2e4 d4e3 f4e3 c7c6 f1e1 b8d7 b1c3 f8e8 a2a3 d7b6 c3e4 c8f5 d1c2 b6d7 a1c1 a7a5 e3f4 a5a4 f3d4 f5g6 d4f3 d8a5 e4d6 e7d6 f4d6 e8e1 f3e1 a5f5 h2h3');
INSERT INTO evaluations VALUES(48747,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(48747,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48747,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(48747,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(48747,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48747,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(48747,36,'cp -26','f8e7 d2d3 e8g8 e1g1 h7h6 c1f4 d5d4 e2e4 d4e3 f4e3 c7c6 f1e1 f8e8 b1c3 c8f5 d3d4 f5e6 c3a4 b8d7 a2a3 e7f8 a4c5 d7c5 d4c5 d8d1 a1d1 f6d5 e3d4 a7a5 f3d2 d5c7 d2e4 c7b5 d4e3 a8d8 a3a4 d8d1 e1d1');
INSERT INTO evaluations VALUES(1985,30,'cp 385',NULL);
INSERT INTO evaluations VALUES(1985,31,'cp 420',NULL);
INSERT INTO evaluations VALUES(1985,32,'cp 408',NULL);
INSERT INTO evaluations VALUES(1985,33,'cp 401',NULL);
INSERT INTO evaluations VALUES(1985,34,'cp 424',NULL);
INSERT INTO evaluations VALUES(1985,35,'cp 444',NULL);
INSERT INTO evaluations VALUES(1985,36,'cp 437','f3d4 b4a6 f2f4 b2b6 e2e3 c7c6 a1b1 b6c7 f1a6 b7a6 d1f3 d7d5 e1g1 c8d7 h2h3 g8h6 e3e4 f8c5 d2e3 e8g8 e4d5 c5d4 e3d4 c6d5 g2g4 a8e8 f4f5 e8e5 d4e5 c7e5 c3d5 f7f6 b1d1 d7b5 f1e1 e5b8 f3f4 b8f4 d5f4 h6f7 e1e7 f8c8 f4e6 c8c2 e7a7 h7h5 e6d4 c2c3 d4b5 a6b5 d1d7 c3h3 g4h5 h3h5 d7f7 h5f5 f7g7 g8f8');
INSERT INTO evaluations VALUES(1982,30,'cp 418',NULL);
INSERT INTO evaluations VALUES(1982,31,'cp 433',NULL);
INSERT INTO evaluations VALUES(1982,32,'cp 451',NULL);
INSERT INTO evaluations VALUES(1982,33,'cp 438',NULL);
INSERT INTO evaluations VALUES(1982,34,'cp 412',NULL);
INSERT INTO evaluations VALUES(1982,35,'cp 399',NULL);
INSERT INTO evaluations VALUES(1982,36,'cp 418','b4a6 f2f4 b2b6 e2e3 c7c6 a1b1 b6c7 f1a6 b7a6 d1f3 d7d5 e1g1 c8d7 h2h3 g8h6 e3e4 f8c5 d2e3 e8g8 e4d5 c6d5 c3d5 c7d8 f1d1 a8c8 d4b3 c5e3 d5e3 c8c3 b3d4 d8c8 b1b3 g8h8 b3c3 c8c3 d1d3 c3c5 g1h2 h6g8 f3e4 g8e7 d4b3 c5c8 d3d6 d7e6 b3d4 e6a2 f4f5 f8e8 f5f6');
INSERT INTO evaluations VALUES(19779,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19779,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19779,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19779,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19779,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(19779,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19779,36,'cp 11','e2e3 c5d4 e3d4 c8f5 c2c4 e7e6 b1c3 d5c4 f1c4 g8f6 e1g1 f8e7 d4d5 f6d5 c3d5 e6d5 c4d5 e8g8 f1e1 e7f6 d5c6 b7c6 d1c1 f8e8 f4e5 f5g4 e1e3 g4f3 e5f6 d8f6 e3f3 f6e5 f3c3 a8c8 b2b4 e5f6 a2a3 h7h6 h2h3 e8e2 c3f3 f6d4 a1b1 c6c5');
INSERT INTO evaluations VALUES(29658,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(29658,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(29658,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(29658,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(29658,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(29658,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(29658,36,'cp -12','e7e6 h2h3 g4f3 d1f3 c7c5 f1b5 b8d7 e1g1 a7a6 b5d7 d8d7 f1e1 a8c8 d4c5 f8c5 e3e4 f6e4 c3e4 d5e4 f3e4 d7c6 e4g4 c6b6 g4e2 e8g8 c2c3 f8d8 c1f4 d8d7 e1d1 c8d8 d1d7 d8d7 b2b4 c5e7 a1d1 b6c6 d1d7 c6d7 e2e4 h7h6');
INSERT INTO evaluations VALUES(30173,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(30173,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(30173,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(30173,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(30173,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(30173,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(30173,36,'cp -14','h2h3 g4f3 d1f3 c7c5 f1b5 b8d7 d4c5 f8c5 e3e4 d5e4 c3e4 f6e4 b5d7 d8d7 f3e4 e8g8 e1g1 f8d8 c1f4 h7h6 e4e2 d7a4 f4e5 c5d6 b2b3 a4a5 e5d6 d8d6 e2f3 a5d5 f3d5 d6d5 f1d1 a8d8 d1d5 d8d5 g1f1 d5d2 a1c1 g8f8 f1e1 d2d8 e1e2 f8e7 c2c3 b7b6 c3c4');
INSERT INTO evaluations VALUES(30171,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(30171,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(30171,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(30171,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(30171,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(30171,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(30171,36,'cp -15','g4f3 d1f3 c7c5 f1b5 b8d7 d4c5 f8c5 e3e4 f6e4 c3e4 d5e4 b5d7 d8d7 f3e4 e8g8 e1g1 f8d8 c1f4 d7d4 e4d4 d8d4 g2g3 g8f8 f1d1 a8d8 d1d4 d8d4 g1f1 e6e5 f4e3 d4c4 e3c5 c4c5 c2c3 f8e7 f1e2 e7e6 a2a4 f7f5 a4a5 e6d6 a5a6 b7b6 e2d3 c5d5 d3e3');
INSERT INTO evaluations VALUES(43670,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(43670,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43670,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43670,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43670,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(43670,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43670,36,'cp 8','b7b5 c4b3 c7c5 e1g1 c8b7 e3e4 c5d4 f3d4 b8c6 d4c6 d8d1 f1d1 b7c6 f2f3 f8c5 g1f1 e8e7 c3e2 h8c8 e2d4 c6d7 c1f4 c5b6 a2a4 b5a4 b3a4 c8c4 a4d7 f6d7 d4b5 c4c2 d1d2 c2d2 f4d2 d7c5 b5a3 c5b3 d2b4 e7e8 a1d1 b6c7 b4c3');
INSERT INTO evaluations VALUES(43700,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43700,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43700,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(43700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43700,36,'cp 0','c4b3 c7c5 e1g1 c8b7 e3e4 c5d4 f3d4 b8c6 d4c6 d8d1 f1d1 b7c6 f2f3 f8c5 g1f1 e8e7 c1f4 h8c8 a1c1 f6d7 h2h4 e7e8 g2g4 a6a5 a2a4 b5b4 c3e2 c5b6 d1d6 b6c5');
INSERT INTO evaluations VALUES(43702,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43702,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43702,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43702,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43702,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43702,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43702,36,'cp 0','c7c5 e1g1 c8b7 e3e4 c5d4 f3d4 b8c6 d4c6 d8d1 f1d1 b7c6 f2f3 f8c5 g1f1 e8e7 c1f4 h8c8 h2h4 f6d7 a1c1 e7e8 g2g4 a6a5 a2a4 b5b4 c3e2 c5b6 d1d6 b6c5');
INSERT INTO evaluations VALUES(44959,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44959,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44959,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(44959,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44959,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44959,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(44959,36,'cp 14','f1g2 g7g6 c2c4 f8g7 c4d5 c6d5 f3e5 e8g8 b1c3 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 h7h6 h2h4 e7e6 e1g1 f8d8 f1c1 d8c8 g2f3 c6e5 f4e5 f5d3 c1d1 d3c4 e5d6 g7f8 d6f8 g8f8 g1g2 g6g5 d1d2 f8g7 h4g5 h6g5');
INSERT INTO evaluations VALUES(52499,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(52499,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(52499,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(52499,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(52499,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(52499,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(52499,36,'cp -66','d7d5 e4d5 g8f6 a2a3 d8d5 g1f3 d5a5 b1d2 e7e6 f1d3 b4a3 e1g1 f8e7 d2c4 a5c7 f3e5 e8g8 c1f4 c7d8 c4a3 a7a5 c2c4 b8c6 e5c6 b7c6 c4c5 g7g6 d1a4 c8a6 d3a6 a8a6 f1c1 a6a8 a3c4 d8d4 f4e3 d4d5 c4b6');
INSERT INTO evaluations VALUES(53350,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(53350,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(53350,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(53350,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(53350,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(53350,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(53350,36,'cp -42','e4d5 g8f6 a2a3 d8d5 g1f3 e7e6 f1d3 b8c6 e1g1 f8e7 a3b4 c6b4 d3e2 e8g8 c2c3 b4c6 c3c4 d5d8 b1c3 a7a5 c1f4 c6b4 f3e5 c8d7 e2f3 d7c6 e5c6 b7c6 g2g3 e7d6 f4e3 e6e5 d4d5 c6d5 c3d5 b4d5 f3d5 a8c8 e3g5 d6b4 d1f3 h7h6 g5f6 d8f6');
INSERT INTO evaluations VALUES(53544,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(53544,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53544,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(53544,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(53544,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(53544,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(53544,36,'cp 0','e6d5 f1d3 f8d6 g1f3 g8e7 c3e2 c8f5 d3f5 e7f5 e2g3 f5e7 e1g1 e8g8 f1e1 f8e8 f3h4 d6g3 h2g3 e7g6 h4f5 b8d7 e1e8 d8e8 f5d6 e8b8 d6f5');
INSERT INTO evaluations VALUES(53827,30,'cp 265',NULL);
INSERT INTO evaluations VALUES(53827,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(53827,32,'cp 274',NULL);
INSERT INTO evaluations VALUES(53827,33,'cp 270',NULL);
INSERT INTO evaluations VALUES(53827,34,'cp 269',NULL);
INSERT INTO evaluations VALUES(53827,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(53827,36,'cp 265','d4e5 g8e7 e5d6 e7g6 d1d4 d8d6 c1e3 b8d7 b1c3 d6f6 h2h3 f8d6 d4f6 d7f6 e1c1 d6b4 f3d2 e8g8 a2a3 b4c3 b2c3 b7b6 f2f3 c6c5 a3a4 c8e6 h1g1 f8e8 f1b5 e6d7 c3c4 d7b5 c4b5 a7a6 d2c4 a6b5 c4b6');
INSERT INTO evaluations VALUES(53804,30,'cp 265',NULL);
INSERT INTO evaluations VALUES(53804,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(53804,32,'cp 248',NULL);
INSERT INTO evaluations VALUES(53804,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(53804,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(53804,35,'cp 264',NULL);
INSERT INTO evaluations VALUES(53804,36,'cp 270','g8e7 e5d6 e7g6 d1d4 d8d6 c1e3 b8d7 b1c3 h7h5 e1c1 d6d4 f3d4 f8b4 f2f3 e8g8 c3e2 f8e8 a2a3 b4f8 e2c3 f8c5 g2g3 d7f6 d4f5 c5e3 f5e3 h5h4 f1e2 g6e5 h1e1 c8e6 g3h4 a8d8 f3f4 e5g4 e3g4 d8d1 e1d1 e6g4 e2g4 f6g4');
INSERT INTO evaluations VALUES(2716,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(2716,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(2716,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(2716,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2716,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(2716,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2716,36,'cp 7','c8g4 c1e3 e5d4 c3d4 c6a5 b3c2 a5c4 e3c1 c7c5 b2b3 c4b6 b1d2 a8c8 c1b2 g4h5 a1c1 c5d4 b2d4 b6d7 c2b1 d7e5 h2h3 f8e8 c1c8 d8c8 d1c2 c8c2 b1c2 h5g6 f3h4 e5c6 d4a1 f6d7 h4g6 h7g6 e1d1 e7f6 a1f6 d7f6 c2d3 c6d4 d1c1 d6d5');
INSERT INTO evaluations VALUES(23621,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(23621,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(23621,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(23621,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(23621,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(23621,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(23621,36,'cp 20','c1e3 e5d4 c3d4 c6a5 b3c2 a5c4 e3c1 c7c5 b2b3 c4b6 b1d2 c5d4 h2h3 g4h5 g2g4 d4d3 c2d3 h5g6 c1b2 a8c8 a1c1 c8c1 d1c1 f6d7 d3b1 f8e8 d2f1 d6d5 f1g3 d5e4 b1e4 e7f8 g3f5 d7c5 e4c2 e8e1 c1e1 c5e6 f3d4 d8d7 d4e6 d7e6 g1f1 b6d5 e1e6');
INSERT INTO evaluations VALUES(4048,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(4048,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(4048,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(4048,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(4048,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(4048,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(4048,36,'cp 71','c1e3 e7e5 d4d5 d7c5 f3d2 f6e8 h3h4 h7h5 b2b4 c5d7 d2b3 f7f5 e3g5 d7f6 f1d3 d8d7 a2a4 d7f7 f2f3 f5e4 d3e4 f6e4 c3e4 c8f5 e1g1 f5e4 f3e4 e8f6 c4c5 f7d7 b3d2 f6h7 d2f3 a7a5 a1b1 d7g4 b4a5 h7g5 f3g5');
INSERT INTO evaluations VALUES(4041,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(4041,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(4041,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(4041,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(4041,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(4041,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(4041,36,'cp 79','e7e5 d4d5 d7c5 f3d2 f6e8 h3h4 h7h5 f2f3 f7f5 b2b4 c5d7 f1d3 a7a5 a2a3 b7b6 d2b3 e8f6 e3g5 f5e4 d3e4 d8e8 b4a5 b6a5 d1c2 f6e4 c3e4 a5a4 b3d2 d7f6 e1g1 c8f5 a1b1 a8b8 b1b8 e8b8 f1b1 b8a7 g1h2 f6d7 b1b5 d7f6 c2b1 a7d4 b5b4');
INSERT INTO evaluations VALUES(3950,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(3950,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(3950,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(3950,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(3950,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(3950,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(3950,36,'cp 62','d4d5 d7c5 f3d2 f6e8 h3h4 h7h5 b2b4 c5d7 d2b3 f7f5 e3g5 g7f6 d1d2 f6g5 d2g5 d7f6 f1d3 e8g7 f2f3 c7c6 e1c1 g8h7 a2a3 c6d5 e4d5 b7b6 c1b2 c8d7 g2g3 a8c8 h1g1 a7a6 d1e1 d8e8 g5d2 e8e7 d2e3 e7d8 a3a4 a6a5 b4b5 d7e8 b3d2 f6d7 b2a1');
INSERT INTO evaluations VALUES(3928,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(3928,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(3928,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(3928,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(3928,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(3928,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(3928,36,'cp 59','d7c5 f3d2 f6e8 h3h4 h7h5 b2b4 c5d7 a2a3 f7f5 f2f3 d8e7 f1d3 a7a5 d2b3 f5e4 c3e4 a5b4 a3b4 a8a1 d1a1 d7f6 e3g5 e7f7 b3d2 f6e4 d2e4 c8f5 e1g1 f5e4 f3e4 e8f6 c4c5 f7d7 a1a7 f6h7 g5e3 h7f6 a7b7 f6g4 c5c6 d7e7 f1f8 e7f8 b7a7 f8e7');
INSERT INTO evaluations VALUES(15402,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(15402,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(15402,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(15402,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(15402,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(15402,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(15402,36,'cp -72','e1g1 f8c5 d2d3 e8g8 h2h3 f8e8 c1e3 c5f8 d3d4 e5d4 f3d4 c6d4 d1d4 d8d4 e3d4 c8f5 c3b5 e8d8 d4e3 a7a6 b5d4 f5g6 c2c3 c7c5 d4e2 f7f6 f1c1 g6h5 e2f4 h5f7 h3h4 d8d7 c1d1 a8d8 d1d7 d8d7 h4h5 g7g5 h5g6 h7g6');
INSERT INTO evaluations VALUES(15401,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(15401,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(15401,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(15401,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(15401,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(15401,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(15401,36,'cp -91','f8c5 d2d3 e8g8 h2h3 f8e8 c1e3 c5f8 d3d4 e5d4 f3d4 c6d4 d1d4 c8f5 d4d8 a8d8 a1c1 a7a6 f1d1 d8d1 c3d1 f7f6 c2c4 f5d3 c4c5 g8f7 b2b4 f8e7 d1b2 d3e4 c1d1 g7g5 f2f3 e4c6 b2c4 f7e6 c4a5 c6d5 d1d4 e8d8');
INSERT INTO evaluations VALUES(20640,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(20640,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(20640,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(20640,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(20640,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(20640,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(20640,36,'cp 22','f8e7 f1c4 c8e6 d1e2 g8f6 e1g1 a8c8 c3d5 e8g8 f1d1 e6d5 c4d5 f6d5 d1d5 c6b4 d5d2 d8c7 c2c3 b4a6 d2d1 c7c6 f3e1 a6c5 f2f3 c5a4 e1d3 a7a5 c1e3 h7h6 d1d2 a4b6 e3b6 c6b6 d3f2 e7g5 d2d5 c8c6 d5b5 b6a6 a2a4 f8c8');
INSERT INTO evaluations VALUES(22443,30,'cp -198',NULL);
INSERT INTO evaluations VALUES(22443,31,'cp -204',NULL);
INSERT INTO evaluations VALUES(22443,32,'cp -180',NULL);
INSERT INTO evaluations VALUES(22443,33,'cp -180',NULL);
INSERT INTO evaluations VALUES(22443,34,'cp -187',NULL);
INSERT INTO evaluations VALUES(22443,35,'cp -198',NULL);
INSERT INTO evaluations VALUES(22443,36,'cp -191','c6d4 f3g3 f8g7 b5a4 b7b5 a4b3 a7a5 c2c3 d4b3 a2b3 d7d5 d2d3 g8e7 g3e3 e8g8 g1f3 d5d4 e3e2 c7c5 e1g1 e7c6 b1d2 a8a6 c3c4 b5b4 f1d1 g8h8 d2f1 d8d7 d1e1 d7c7 f1g3 c6d8 h2h3 f7f6 e2d1 c7f7 d1e2 d8e6 f3h2 f8e8 h2f1');
INSERT INTO evaluations VALUES(22495,30,'cp -192',NULL);
INSERT INTO evaluations VALUES(22495,31,'cp -192',NULL);
INSERT INTO evaluations VALUES(22495,32,'cp -192',NULL);
INSERT INTO evaluations VALUES(22495,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(22495,34,'cp -183',NULL);
INSERT INTO evaluations VALUES(22495,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(22495,36,'cp -185','f3g3 f8g7 b5a4 b7b5 a4b3 a7a5 c2c3 d4b3 a2b3 g8e7 d2d3 d7d5 g3e3 e8g8 g1f3 d5d4 e3e2 c7c5 b1d2 d8d7 e1g1 e7c6 f1d1 d7c7 h2h3 a8a6 d1e1 c7e7 c3c4 b5b4 d2f1 e7c7 f1g3 g8h8 e2d1 c6d8 c1d2 h7h6 d1e2 d8e6 e2d1');
INSERT INTO evaluations VALUES(26399,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(26399,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(26399,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(26399,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26399,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26399,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(26399,36,'cp 0','b1d2 c5d4 f3d4 g4e2 d1e2 c6d4 c3d4 g8e7 e1g1 a8c8 d2f3 e7c6 c1e3 f8e7 g2g3 d8d7 h2h4 e8g8 f1c1 a7a6 f3e1 c6a5 e2g4 g8h8 e1d3 c8c1 a1c1 f8c8 c1c8 d7c8 e3g5 e7f8 g5d2 a5c6 h4h5 h7h6 g4f3 h8g8 f3g4');
INSERT INTO evaluations VALUES(33256,30,'cp 320',NULL);
INSERT INTO evaluations VALUES(33256,31,'cp 342',NULL);
INSERT INTO evaluations VALUES(33256,32,'cp 306',NULL);
INSERT INTO evaluations VALUES(33256,33,'cp 301',NULL);
INSERT INTO evaluations VALUES(33256,34,'cp 322',NULL);
INSERT INTO evaluations VALUES(33256,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(33256,36,'cp 319','c5b6 d2d4 f6d8 f1d3 g8e7 e1g1 e8g8 a2a4 c7c6 c1g5 d7d5 b1d2 h7h6 g5h4 c8e6 f1e1 f8e8 b2b4 b8d7 c2c4 d5e4 d2e4 b6c7 a4a5 d8c8 h4g3 e7f5 g3c7 c8c7 a5a6 a8d8 d1c2 b7a6 c2c3 f5e7 e4g3 e7g6 a1a6 g6f4');
INSERT INTO evaluations VALUES(50197,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50197,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(50197,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50197,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50197,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50197,36,'cp 0','f8g7 g1f3 c7c5 e2e3 e8g8 f1e2 d8b6 d1c1 b8c6 c2c3 d7d5 a2a4 c8f5 b1d2 c5d4 e3d4 c6b4 c3b4 f8c8 c1d1 f5c2 d1c1');
INSERT INTO evaluations VALUES(43215,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43215,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43215,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43215,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43215,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43215,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43215,36,'cp 0','e2e3 c7c5 g1f3 e8g8 f1e2 d8b6 d1c1 d7d5 c2c3 b8c6 e1g1 c8f5 b1d2 a8c8 d2b3 f6e4 f1d1 c5d4 e3d4 a7a5 f4e3 b6d8 a2a4 e4d6 e3f4 b7b6 e2a6 c8a8 a6e2');
INSERT INTO evaluations VALUES(360,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(360,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(360,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(360,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(360,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(360,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(360,36,'cp 126','c3d4 c5b6 b1c3 g8e7 a2a4 c7c6 a4a5 b6c7 f1d3 d7d5 e4e5 f6h4 d1c2 e8g8 e1g1 h4h5 c3b5 c7a5 a1a5 c6b5 f1a1 a7a6 a5b5 a8b8 b5c5 c8e6 h2h3 e7f5 c2d2 f5h4 d2d1 h5d1 a1d1 f8c8 f2f4 g7g6 c5c8 b8c8 e3f2 h4f5 g2g4 f5g7 g1g2 c8c6 d1b1 c6b6');
INSERT INTO evaluations VALUES(359,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(359,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(359,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(359,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(359,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(359,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(359,36,'cp 105','c5b6 b1c3 g8e7 a2a4 c7c6 a4a5 b6c7 f1d3 e8g8 e1g1 d7d5 e4e5 f6h4 d1c2 h4h5 f1e1 f8d8 c3e2 c8f5 e2f4 f5d3 f4d3 d8e8 h2h3 h5g6 c2e2 a7a6 f2f4 a8b8 e3f2 h7h5 e2f3 e7f5 g2g4 h5g4 h3g4 f5h6 f4f5 g6g4 f3g4 h6g4');
INSERT INTO evaluations VALUES(10232,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(10232,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(10232,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(10232,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(10232,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(10232,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(10232,36,'cp -32','d7d5 b1d2 e8g8 f1e2 a7a5 e1g1 c8e6 f1e1 h7h6 e4d5 d8d5 e2f1 f8e8 d2c4 e5e4 d3e4 d5d1 e1d1 f6e4 f3d4 c6d4 c3d4 e8d8 c1e3 c5e7 e3f4 e7f6 f4c7 d8d4 d1d4 f6d4 c7b6 d4b6 c4b6 a8d8 b6c4 e6c4 f1c4 d8d2 a1e1 e4d6 b2b3 d6c4 b3c4 d2d4 g2g3 d4c4');
INSERT INTO evaluations VALUES(9091,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(9091,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(9091,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(9091,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(9091,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(9091,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(9091,36,'cp -31','b1d2 a7a5 f1e2 e8g8 e1g1 c8e6 e4d5 e6d5 f1e1 h7h6 d2f1 c5d6 f1g3 f8e8 c1e3 d8d7 d1c2 d5e6 g3e4 f6d5 d3d4 f7f5 e4d6 c7d6 c3c4 d5e3 f2e3 e6f7 d4e5 d6e5 a1d1 d7e6 e3e4 c6b4 c2b1');
INSERT INTO evaluations VALUES(13240,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13240,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13240,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(13240,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(13240,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13240,36,'cp 0','c3d5 f6d5 c4d5 e7e6 d5c6 b7c6 f2f4 d6d5 c1e3 e6e5 g1e2 d5d4 e3d2 d8h4 e2g3 e5f4 d1h5 h4f6 g3e2 g7g6 e4e5 f6d8 h5f3 d8d5 f3f4 d5g2 h1g1 g2d5 f4f6 h8g8 e2f4 d5d7 f6h4 g8h8 h4f6');
INSERT INTO evaluations VALUES(17556,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(17556,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(17556,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(17556,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(17556,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(17556,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(17556,36,'cp 74','b5c6 b7c6 e4e5 g8h6 e5d6 e7d6 d2d4 h6f5 d4c5 d8e7 c3e2 d6c5 e1g1 e7d6 d1e1 f8e7 c1f4 d6e6 e2g3 f5g3 h2g3 e6e1 f1e1 c8e6 f3e5 a8c8 e1e2 e7f6 e5d3 e8g8 f4e5 f6e7 a1e1 f8e8 d3f4 e6c4 e2e3 c4a2 e5c3 g8f8 c3g7 f8g7 e3e7 g7f8');
INSERT INTO evaluations VALUES(17541,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17541,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(17541,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(17541,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(17541,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(17541,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17541,36,'cp 78','b7c6 e4e5 g8h6 e1g1 h6f5 b2b3 f5d4 h2h3 g7g6 f1e1 c8f5 d2d3 f8g7 e5d6 d8d6 c1g5 d4e6 g5d2 h7h5 a1b1 e6d4 d2g5 d4f3 d1f3 g7c3 g5e7 d6d5 e7f6 c3e1 b1e1 f5e6 f3d5 c6d5 f6h8 e8e7 h8c3 d5d4 c3d2 e7d6 g1f1 a8c8 d2a5 c8e8 f2f3');
INSERT INTO evaluations VALUES(18209,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(18209,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(18209,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(18209,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(18209,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(18209,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(18209,36,'cp 88','e4e5 g8h6 e1g1 h6f5 b2b3 f5d4 h2h3 g7g6 e5d6 e7d6 f1e1 f8e7 c3a4 e8g8 c2c3 d4e6 d2d4 c5d4 f3d4 e6d4 c3d4 c8e6 c1f4 f8e8 a1c1 a8c8 d1d2 h7h5 e1e2 e7f8 c1e1 e6d7 e2e8 d7e8 f4g5 f7f6 g5h6 f8h6 d2h6 e8f7 a4c3 d8a5');
INSERT INTO evaluations VALUES(49430,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(49430,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(49430,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(49430,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(49430,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(49430,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49430,36,'cp 9','e2e4 d7d5 e4e5 f6g8 c2c3 c7c5 c1e3 d8b6 d1d2 b8c6 f3f4 g8h6 d4c5 f8c5 e3c5 b6c5 b1a3 f7f6 e5f6 g7f6 f1d3 c8d7 a3c2 h6f7 d2e3 c5e3 c2e3 c6e7 e1c1 f7d6 d3c2 e8c8 f4f5 d6f5 e3f5 e7f5 c2f5 e6f5 g1e2 h8e8 e2d4');
INSERT INTO evaluations VALUES(49443,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(49443,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(49443,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(49443,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(49443,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(49443,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(49443,36,'cp 14','d7d5 e4e5 f6g8 f3f4 c7c5 g1f3 g8h6 c2c3 b8c6 b1a3 c5d4 c3d4 f8b4 c1d2 b4a3 b2a3 h6f5 f1d3 c6d4 f3d4 f5d4 e1g1 d4c6 d2e3 c8d7 e3c5 c6e7 a1b1 b7b6 c5d6 e8g8 b1c1 g7g6 d1d2 a8c8 c1c8 d7c8 g2g4 c8b7');
INSERT INTO evaluations VALUES(48382,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(48382,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(48382,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(48382,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(48382,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(48382,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(48382,36,'cp 27','e4e5 f6g8 f3f4 c7c5 g1f3 g8h6 c2c3 b8c6 b1a3 c5d4 c3d4 h6f5 a3c2 d8b6 f1d3 c6b4 c2b4 f8b4 e1f1 h7h5 c1d2 c8d7 d3f5 e6f5 h2h3 b4e7 d2c3 h5h4 f1g1 b6g6 c3e1 g6h6 e1d2 e8g8 d1b3 h6b6 b3b6 a7b6 g1h2 f8c8 d2e1 a8a4 e1h4 e7h4 f3h4');
INSERT INTO evaluations VALUES(48430,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(48430,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(48430,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(48430,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(48430,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48430,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48430,36,'cp 13','f6g8 f3f4 c7c5 g1f3 g8h6 c2c3 b8c6 c1e3 d8b6 d1d2 c8d7 f1d3 c5d4 e3d4 c6d4 c3d4 d7b5 d3b5 b6b5 b1c3 b5d7 c3e2 e8c8 e1g1 c8b8 h2h3 d8c8 f1c1 c8c1 a1c1 g7g6 d2c3 h6f5 f3g5 f8h6 g2g4 h6g5 f4g5 f5e7 c3f3 e7c6 g1g2');
INSERT INTO evaluations VALUES(50564,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(50564,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(50564,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(50564,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(50564,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(50564,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(50564,36,'cp 72','f8b4 b1d2 d7d5 e4e5 c8d7 a2a3 b4e7 f1d3 d7b5 d3b5 a6b5 d2b3 b8c6 e1g1 h7h5 c1e3 d8b6 b3c5 e7c5 d4c5 b6c7 d1d3 g8e7 d3b5 e8g8 b2b4 e7g6 b5e2 c6e5 a3a4 e5f3 e2f3 f8d8 f3h5 d5d4 e3d2 d8d5 h5e2 d4d3 e2e3 g6e7 h2h3 e7f5 e3e4 c7d7');
INSERT INTO evaluations VALUES(56363,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(56363,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(56363,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(56363,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(56363,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(56363,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(56363,36,'cp 104','d4e5 d5d4 c3b5 c7c5 b5a3 b8c6 e2e3 a7a6 e3d4 d8d4 d1d4 c5d4 f2f4 b7b5 g1f3 f8c5 f1d3 g8e7 a3b1 e7d5 d3e4 c6e7 a2a4 b5b4 b1d2 c8f5 e4f5 e7f5 d2b3 a8c8 e1e2 e8g8 c1d2 f5e3 g2g3 e3c2 a1c1 c5a7 e2d3 c2e3 c1c8 f8c8');
INSERT INTO evaluations VALUES(56313,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(56313,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(56313,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(56313,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(56313,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(56313,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(56313,36,'cp 76','d5d4 c3e4 d8d5 e4g3 g8e7 g1f3 b8c6 e2e4 d4e3 c1e3 c8g4 c2c3 e8c8 d1a4 g4f3 g2f3 d5e5 f3f4 e5e6 f1c4 e6f6 a1d1 a7a6 g3e4 f6f5 e4c5 d8d1 a4d1 e7g6 d1d3 f5d3 c5d3 c6d8 h2h4 f8d6 e1d2 g6e7 a2a4 c7c6 b2b4 h8e8 c4b3 d8e6 b3e6 f7e6 d2c2 e7d5 h4h5');
INSERT INTO evaluations VALUES(56531,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(56531,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(56531,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(56531,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(56531,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(56531,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(56531,36,'cp 94','c3b5 c7c5 e2e3 a7a6 b5a3 b8c6 e3d4 d8d4 d1d4 c5d4 f2f4 b7b5 f1d3 c8b7 g1f3 g8e7 a3b1 e7d5 a2a3 f8c5 g2g3 e8g8 b1d2 c5b6 d2e4 c6a5 e1f2 a5c4 b2b3 f8c8 b3c4 b5c4 e4d6 c4d3 d6c8 a8c8 c2d3 c8c2 c1d2');
INSERT INTO evaluations VALUES(21698,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21698,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(21698,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21698,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21698,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21698,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21698,36,'cp 0','d5e4 c3e4 f7f5 e4c3 e5e4 d4d5 e4f3 d5c6 d8d1 c3d1 g8e7 d1c3 e7c6 c3b5 c6b4 b5c7 e8d8 c1f4 b4c2 e1d2 c2a1 c7a8 c8e6 d2c3 f8e7 f4e5 e7f6 e5f6 g7f6 g2f3 d8e7 f1d3 h8c8 c3d2 c8a8 h1a1 a8d8 d2c3 d8c8 c3d2');
INSERT INTO evaluations VALUES(21799,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21799,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21799,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21799,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21799,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21799,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21799,36,'cp 0','c3e4 f7f5 e4c3 e5e4 d4d5 e4f3 d5c6 d8d1 c3d1 g8e7 d1c3 e7c6 c3b5 c6b4 b5c7 e8d8 c1f4 b4c2 e1d2 c2a1 c7a8 c8e6 d2c3 f8e7 f4e5 e7f6 e5f6 g7f6 g2f3 d8e7 f1d3 h8c8 c3d2 c8a8 h1a1 a8d8 d2c3 d8c8 c3d4 f5f4 a1e1 c8d8 d4c3 d8c8 c3d2 c8d8');
INSERT INTO evaluations VALUES(21796,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21796,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21796,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21796,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21796,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21796,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21796,36,'cp -9','e5d4 f1b5 c8f5 b5c6 b7c6 e4g3 d8e7 g3e2 f5e4 d1d4 e7b4 c2c3 b4d4 e2d4 g8f6 e1g1 c6c5 d4b5 e4f3 g2f3 e8d7 f1d1 d7c6 a2a4 f8d6 b5d6 c7d6 b2b4 a7a6 c1g5 h8e8 g5f6 g7f6 d1d2 d6d5 b4c5 c6c5 d2d4 a8b8 d4f4 b8b2 f4f6 e8e2 f6a6 e2f2');
INSERT INTO evaluations VALUES(21839,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21839,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21839,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(21839,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21839,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21839,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21839,36,'cp 0','f1b5 c8f5 b5c6 b7c6 e4g3 d8e7 g3e2 e8c8 e1g1 f5g4 d1d3 g4f3 d3f3 g8f6 f3c6 e7e2 c1f4 e2e7 b2b4 c8b8 c6b5 b8c8');
INSERT INTO evaluations VALUES(36870,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(36870,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(36870,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(36870,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(36870,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(36870,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(36870,36,'cp 61','d4d5 c8d7 a2a3 a7a5 e1g1 b8a6 d3c2 c7c6 d5c6 d7c6 b2b3 a6c7 a3a4 c7a6 c1e3 a6b4 a1c1 d8e7 c2b1 f8c8 c3d5 e7f8 h2h3 c6d5 c4d5 c8c1 d1c1 a8c8 c1d1 g7h6 e3b6 c8a8 f3d2 f6d7 d2c4 d7b6');
INSERT INTO evaluations VALUES(36842,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(36842,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(36842,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(36842,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(36842,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(36842,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(36842,36,'cp 37','a7a5 a2a3 f6d7 a1b1 d7c5 d3c2 f7f5 b2b4 f5e4 f3g5 c5d3 c2d3 e4d3 e1g1 a5b4 a3b4 c8f5 g5e4 b8d7 c1g5 d8e8 d1d3 h7h6 g5d2 a8a3 b1a1 a3b3 f1b1 f5e4 d3e4 e8f7 f2f3 b3b1 a1b1 f8a8 e4c2 d7f6 b1e1 g8h7 d2e3 b7b6 c2d3 f7d7 h2h3 f6h5');
INSERT INTO evaluations VALUES(35614,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(35614,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(35614,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(35614,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(35614,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(35614,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(35614,36,'cp 41','a2a3 f6d7 a1b1 f7f5 b2b4 f5f4 h2h3 h7h6 c1b2 b7b6 b1a1 b8a6 d1b3 d7f6 d3e2 a6b8 f3d2 b8d7 c3b5 g8h8 e1d1 h6h5 d1c2 c8a6 b3d3 a6b5 c4b5 a5b4 a3b4 a8a1 h1a1');
INSERT INTO evaluations VALUES(37410,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(37410,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(37410,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(37410,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(37410,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(37410,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(37410,36,'cp 36','b1c3 d7d6 g1f3 c7c5 d4d5 e7e6 e1g1 f8e8 c1g5 h7h6 g5f6 d8f6 f3e1 e6e5 e2e4 h6h5 h2h4 g7h6 d1e2 f6e7 e1c2 g8h7 g1h2 c8g4 e2e1 g4c8 c2e3 e8f8 e1e2 b8d7 g2h3 a8b8 e3g2 a7a6 a1b1 d7f6 h3c8 b8c8 f2f3 c8e8');
INSERT INTO evaluations VALUES(41989,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41989,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(41989,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41989,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(41989,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(41989,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(41989,36,'cp 24','e4e5 h7h6 g5c1 f6e4 d1g4 e8f8 g1e2 c7c5 a2a3 h6h5 g4f4 b4a5 b2b4 e4c3 e2c3 c5b4 a3b4 a5b4 c1d2 c8d7 f1e2 d8c8 c3d1 b4d2 f4d2 c8c7 e1g1 a7a6 d1e3 b7b5 h2h3 g7g6 f1b1 a8a7 c2c3 b8c6 g2g3 f8g7 h3h4 c7b6 e3c2 h8c8 e2d3 b6d8 d2f4 g7g8 c2e3 c6e7');
INSERT INTO evaluations VALUES(42121,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(42121,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(42121,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(42121,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42121,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(42121,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(42121,36,'cp 34','h7h6 g5d2 b4c3 b2c3 f6e4 d1g4 g7g6 d2c1 e4c3 f1d3 c7c5 d4c5 d8a5 c1d2 a5a4 g4f3 c3e4 d3e4 a4e4 f3e4 d5e4 a1b1 b7b6 g1e2 b8d7 c5b6 a7b6 e1g1 c8a6 f1e1 e8g8 d2h6 f8c8 e2g3 c8c2 g3e4 c2a2 h2h4 a6d3');
INSERT INTO evaluations VALUES(41657,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(41657,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(41657,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(41657,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(41657,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(41657,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(41657,36,'cp 56','g5c1 f6e4 d1g4 g7g6 g1e2 e4c3 e2c3 h6h5 g4f4 a7a6 h2h4 c8d7 h1h3 b4e7 h3f3 h8h7 f1d3 c7c5 d4c5 b8c6 c1e3 d8b8 e3d4 b8c7 c3d5 e6d5 e5e6 c7f4 e6d7 e8d7 f3f4 c6d4 f4d4 d7c6 e1c1 e7f6 d4f4 f6e5 f4a4 c6c5 d3e2 a8d8 g2g3 b7b5');
INSERT INTO evaluations VALUES(45453,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45453,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45453,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45453,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45453,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45453,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(45453,36,'cp 12','g1f3 e7e6 f1d3 f8d6 e1g1 b8c6 b1c3 e8g8 h2h3 h7h6 d1e2 a7a6 f1d1 d6c7 d3c2 f8e8 e3e4 d5e4 c3e4 c6b4 e4f6 d8f6 c2e4 e8d8 c1e3 b4d5 a1c1 c7b6 d1e1 c8d7 e4d5 e6d5 f3e5 d7e8 a2a3 a8c8');
INSERT INTO evaluations VALUES(45454,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45454,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45454,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45454,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(45454,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45454,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(45454,36,'cp 0','a7a6 c1d2 e7e6 f1d3 f8d6 e1g1 e8g8 f3e5 f6e4 d3e4 d5e4 b1c3 f7f5 c3a4 c8d7 e5d7 b8d7 d1b3 d8e7 b3b7 f8b8 b7c6 b8c8 c6b7');
INSERT INTO evaluations VALUES(47522,30,'cp 413',NULL);
INSERT INTO evaluations VALUES(47522,31,'cp 390',NULL);
INSERT INTO evaluations VALUES(47522,32,'cp 419',NULL);
INSERT INTO evaluations VALUES(47522,33,'cp 413',NULL);
INSERT INTO evaluations VALUES(47522,34,'cp 399',NULL);
INSERT INTO evaluations VALUES(47522,35,'cp 406',NULL);
INSERT INTO evaluations VALUES(47522,36,'cp 425','d1a4 b8c6 d4d5 b4d5 e4d5 d8d5 b1c3 d5d7 c1e3 e7e5 f1b5 f8d6 g1f3 e8g8 e1g1 h7h6 a1d1 d7e7 c3d5 e7d8 b5c6 b7c6 a4c6 c8d7 c6c2 d7g4 h2h3 g4f3 g2f3 f7f5 f3f4 e5e4 f2f3 d8h4 c2g2 c7c6 d5c3 d6f4 e3f4 h4f4 f3e4 f4e5 e4f5');
INSERT INTO evaluations VALUES(47551,30,'cp 412',NULL);
INSERT INTO evaluations VALUES(47551,31,'cp 425',NULL);
INSERT INTO evaluations VALUES(47551,32,'cp 412',NULL);
INSERT INTO evaluations VALUES(47551,33,'cp 409',NULL);
INSERT INTO evaluations VALUES(47551,34,'cp 417',NULL);
INSERT INTO evaluations VALUES(47551,35,'cp 413',NULL);
INSERT INTO evaluations VALUES(47551,36,'cp 435','b8c6 d4d5 b4d5 e4d5 d8d5 b1c3 d5d4 f1c4 d4e5 c1e3 e5a5 a4a5 c6a5 c4e2 e7e5 g1f3 a5c6 e1c1 f8b4 c3d5 b4d6 f3g5 e8g8 g5e4 c8f5 d5c3 d6e7 g2g4 f5e6 h2h4 c6d4 e3d4 e5d4 d1d4 c7c6 h1d1 f7f5 g4f5 e6f5 h4h5');
INSERT INTO evaluations VALUES(14460,30,'cp 432',NULL);
INSERT INTO evaluations VALUES(14460,31,'cp 425',NULL);
INSERT INTO evaluations VALUES(14460,32,'cp 437',NULL);
INSERT INTO evaluations VALUES(14460,33,'cp 413',NULL);
INSERT INTO evaluations VALUES(14460,34,'cp 441',NULL);
INSERT INTO evaluations VALUES(14460,35,'cp 421',NULL);
INSERT INTO evaluations VALUES(14460,36,'cp 436','d4d5 b4d5 e4d5 d8d5 b1c3 d5d7 g1f3 e7e5 f1b5 f8d6 c1e3 e8g8 e1g1 d7e7 a1d1 c8d7 f3g5 g8h8 c3d5 e7d8 f2f4 e5f4 e3f4 f7f6 f4d6 c7d6 g5f3 a7a6 b5c6 b7c6 d5f4 d8b6 a4d4 b6d4 d1d4 d6d5 f4d3 f8b8 d3c5 b8b2 c5d7 a8d8 d7c5 b2a2');
INSERT INTO evaluations VALUES(14409,30,'cp 413',NULL);
INSERT INTO evaluations VALUES(14409,31,'cp 439',NULL);
INSERT INTO evaluations VALUES(14409,32,'cp 429',NULL);
INSERT INTO evaluations VALUES(14409,33,'cp 436',NULL);
INSERT INTO evaluations VALUES(14409,34,'cp 438',NULL);
INSERT INTO evaluations VALUES(14409,35,'cp 411',NULL);
INSERT INTO evaluations VALUES(14409,36,'cp 428','b4d5 e4d5 d8d5 b1c3 d5d7 f1b5 e7e5 g1f3 f8d6 c1e3 e8g8 e1g1 a8b8 f1d1 a7a6 b5c6 d7c6 a4c6 b7c6 c3e4 b8b2 e4d6 c7d6 d1d6 f7f6 a2a4 c6c5 h2h3 c5c4 a1c1 b2b4 f3d2 b4a4 d2c4 a4b4 g1h2 b4c4 c1c4 a6a5 c4c7 h7h5');
INSERT INTO evaluations VALUES(57718,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(57718,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(57718,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(57718,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(57718,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(57718,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(57718,36,'cp 41','e7e6 b1c3 d7d5 c1f4 f8d6 f4d6 d8d6 e2e3 g8f6 c4d5 f6d5 a1c1 e8g8 f1d3 b8d7 g1f3 b7b6 c3d5 e6d5 e1g1 c7c5 b2b3 f8d8 d1c2 c8b7 f1d1 a8c8 c2b2 d7f8 b2a3 c8c7 c1c2 f8e6 d1c1 a7a5 a3b2 d6e7 h2h3 d8a8');
INSERT INTO evaluations VALUES(17334,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(17334,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(17334,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(17334,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(17334,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(17334,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(17334,36,'cp 56','b7c6 f1d3 d8c7 d1e2 c8b7 f2f4 d7d5 f4f5 e6e5 g2g4 d5d4 c3d1 f8b4 c2c3 b4a5 e1g1 a5b6 c3d4 b6d4 c1e3 d4e3 e2e3 e8c8 d1c3 g8f6 g4g5 f6g4 e3e2 c8b8 a1e1 c7b6 g1h1 d8d3 e2d3 g4f2 f1f2 b6f2 d3e3 f2e3 e1e3 c6c5 e3h3 b8c7 b2b3 c7c6 h1g2 c5c4 h3h5 g7g6 f5g6 f7g6');
INSERT INTO evaluations VALUES(19844,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(19844,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(19844,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(19844,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(19844,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(19844,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19844,36,'cp 23','e2e3 g8f6 f1b5 c8g4 g1f3 e7e6 b5c6 b7c6 e1g1 f8d6 f4d6 d8d6 b1d2 e8g8 a1c1 f8c8 d2b3 a8b8 h2h3 g4h5 d1e2 e6e5 d4e5 h5f3 e2f3 d6e5 c1c2 f6e4 b3d4 c6c5 f1c1 g7g6 f3d1 h7h5 d4f3 e5d6 f3d2 d6e6 d2e4');
INSERT INTO evaluations VALUES(19841,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(19841,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(19841,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19841,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(19841,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(19841,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(19841,36,'cp 19','g8f6 f1b5 c8g4 g1f3 e7e6 e1g1 f8d6 b5c6 b7c6 f4d6 d8d6 b1d2 e8g8 d1a4 g4f3 d2f3 f8c8 f1c1 c6c5 c1c2 a7a5 d4c5 c8c5 a1c1 a8c8 g2g3 h7h6 g1g2 d6b6 f3e5 c5c2 c1c2 c8c7 h2h3 g7g6 a2a3 c7c2 a4c2 f6e4 e5c6 b6c7');
INSERT INTO evaluations VALUES(29468,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(29468,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(29468,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(29468,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(29468,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(29468,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(29468,36,'cp 117','c4d5 e6d5 f2f3 b7b6 c3d5 d8d5 e2e4 d5d7 c1e3 e7e6 g1e2 f8b4 e2c3 e8g8 a2a3 b4c3 b2c3 f8d8 d1b3 c7c5 f1b5 d7c7 e1g1 a7a6 b5e2 b8c6 f1c1 c6a5 b3a2 c5c4 e2d1 e6e5 a3a4 h7h6 a1b1 a8b8 a2d2 b6b5 d1c2 f6d7 f3f4 a5b3 c2b3 c4b3 b1b3');
INSERT INTO evaluations VALUES(29464,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(29464,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(29464,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(29464,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(29464,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(29464,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(29464,36,'cp 126','e6d5 f2f3 b7b6 c3d5 d8d5 e2e4 d5d7 g1e2 e7e6 e2c3 f8b4 a2a3 b4c3 b2c3 e8g8 f1e2 c7c5 d4c5 d7c8 c5b6 a7b6 d1b3 c8c5 b3b4 f6d7 b4c5 d7c5 c1e3 c5a4 e1f2 b8d7 h1d1 d7c5 d1b1 e6e5 b1b5 h7h5 e3c5 b6c5 c3c4 h5h4');
INSERT INTO evaluations VALUES(43014,30,'cp -112',NULL);
INSERT INTO evaluations VALUES(43014,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(43014,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(43014,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(43014,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(43014,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(43014,36,'cp -98','d7d5 e5g3 d5e4 b1c3 e8g8 g1e2 b8c6 c1g5 h7h6 g5f6 e7f6 e1c1 f8e8 c1b1 c8f5 d4d5 c6e7 d5d6 c7d6 d1d6 d8b8 d6f6 b8g3 e2g3 g7f6 f1b5 e8d8 g3f5 e7f5 c3e4 f5d4 b5d3 f6f5 e4f6 g8f8 b1c1 a8c8 f6h5 c8c5 h5g3 d8e8 h2h3 e8e5 g3h5 d4e6');
INSERT INTO evaluations VALUES(49981,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(49981,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(49981,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(49981,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(49981,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(49981,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(49981,36,'cp 120','d8e7 d1e2 d7d6 e5f3 c8f5 d2d3 b8c6 b1c3 e8c8 e2e7 f8e7 f1e2 h7h6 c1e3 h8e8 e1g1 c6b4 e2d1 f6d5 a2a3 d5c3 a3b4 c3d1 f1d1 a7a6 h2h3 f5d7 d1e1 g7g5 e3d2 g5g4 h3g4 d7g4 d2h6 g4f3 g2f3 e7f6 c2c3 e8h8 h6f4 c7c5 b4c5 d6c5 a1a5');
INSERT INTO evaluations VALUES(33998,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(33998,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(33998,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(33998,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(33998,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(33998,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(33998,36,'cp 135','d1e2 d7d6 e5f3 c8f5 e2e7 f8e7 d2d3 b8c6 f1e2 e8c8 b1c3 d6d5 e2d1 h7h6 e1g1 g7g5 c3e2 d8e8 e2g3 f5g6 c1d2 f6d7 d2c3 h8g8 f3d4 c6d4 c3d4 c7c5 d4c3 e7d6 d1g4 c8c7 a1e1 d5d4 c3d2 d6f4 e1e8 g8e8 d2f4 g5f4 g3h5');
INSERT INTO evaluations VALUES(33999,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(33999,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(33999,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(33999,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(33999,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(33999,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(33999,36,'cp 119','d7d6 e5f3 b8c6 b1c3 c8f5 d2d3 e8c8 e2e7 f8e7 f1e2 h7h6 c1e3 h8e8 h2h3 a7a6 e1c1 d6d5 d3d4 c6b4 f3e1 f6e4 h1f1 g7g5 c1b1 e4c3 b2c3 b4c6 e1d3 e7d6 d1e1 c6a5 e2h5 e8e7 g2g4 f5e4 f2f3 e4h7 f3f4 h7d3 c2d3 d6f4 e3f4 e7e1 f1e1 g5f4');
INSERT INTO evaluations VALUES(33771,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(33771,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(33771,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(33771,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(33771,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(33771,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(33771,36,'cp 132','e5f3 c8f5 d2d3 b8c6 e2e7 f8e7 f1e2 h7h6 b1c3 e8c8 e1g1 g7g5 c1e3 c6b4 e2d1 b4d5 e3d2 d8e8 h2h3 f5d7 f1e1 h8f8 c3d5 f6d5 c2c3 e7f6 d1c2 d5f4 d2f4 g5f4 d3d4 f8g8 c2b3 e8e1 a1e1 g8e8 g1f1 e8e1 f1e1 c7c5 e1d2 c8c7 d2d3 b7b6 d4d5');
INSERT INTO evaluations VALUES(52573,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(52573,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(52573,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(52573,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(52573,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(52573,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(52573,36,'cp 67','g1f3 g8f6 c4d5 c6d5 c1f4 b8c6 e2e3 f6h5 f4e5 f7f6 e5g3 e7e6 f1d3 c8d7 e1g1 f8e7 e3e4 d5e4 c3e4 e8g8 f1e1 f6f5 e4c5 h5g3 h2g3 c6d4 c5d7 d8d7 d3f5 f8f5 d1d4 d7d4 f3d4 f5d5 d4e6 a8c8 e1e2 b7b6 e6f4 d5d7 g1f1 e7f6 a1e1 c8d8');
INSERT INTO evaluations VALUES(55601,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(55601,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(55601,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(55601,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(55601,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(55601,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(55601,36,'cp 45','d2d4 g8f6 b1d2 b8d7 g1f3 g7g6 f1d3 f8g7 e1g1 e8g8 f1e1 f8e8 d1c2 b7b6 a2a4 a7a6 d2c4 e5d4 c3d4 d6d5 c4d2 d5e4 d2e4 c8b7 c1g5 h7h6 g5h4 d8c8 e4f6 d7f6 f3e5 b7d5 d3c4 c7c6 h4f6 g7f6 a1d1 d5c4 c2c4 f6e5');
INSERT INTO evaluations VALUES(21513,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(21513,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(21513,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(21513,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(21513,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(21513,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21513,36,'cp 89','c3e4 g8f6 f1d3 c6b4 e4f6 g7f6 d3c4 b7b6 c2c3 b4d5 c4b5 c8d7 b5a6 d7c8 a6c8 d8c8 e1g1 f8g7 c3c4 d5e7 b2b3 c8d7 d1e2 e8g8 f1d1 a8d8 c1a3 f8e8 d1e1 d7c6 a3e7 e8e7 a1d1 e7e8 g2g3 d8d7');
INSERT INTO evaluations VALUES(38784,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(38784,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(38784,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(38784,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(38784,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(38784,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(38784,36,'cp -9','g8f6 e2e3 b7b6 f1b5 c7c6 b5d3 c8b7 e1g1 e8g8 a2a3 b4e7 e3e4 d5e4 c3e4 b8d7 d1e2 c6c5 d4c5 f6e4 d3e4 b7e4 e2e4 d7c5 e4e2 d8c7 a1d1 a8c8 c2c3 f8d8 d2e3 c7c6 f1e1 h7h6 h2h4 c6e4 d1d8 e7d8 e3c5 e4e2 e1e2 c8c5 g1f1 g8f8');
INSERT INTO evaluations VALUES(42019,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42019,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42019,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(42019,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(42019,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42019,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(42019,36,'cp -4','e2e3 e8g8 f1d3 b8d7 e1g1 b7b6 a2a3 b4e7 e3e4 d5e4 c3e4 c8b7 d1e2 c7c5 d4c5 b7e4 d3e4 f6e4 e2e4 d7c5 e4e2 d8c7 a1d1 a8c8 c2c3 f8d8 d2e3 h7h6 f1e1 c7c6 g2g3 c6b7 f3e5 d8d5 e3c5 c8c5 e5d3 c5c7 d3e5 e7f6 d1d5 b7d5');
INSERT INTO evaluations VALUES(50061,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(50061,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(50061,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(50061,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(50061,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(50061,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(50061,36,'cp 111','d2d4 a7a6 c2c3 d7d6 f1d3 g7g6 c1e3 f8g7 h2h4 h7h6 b1d2 b8d7 d1e2 b7b5 a2a4 b5a4 e1g1 e8g8 d2c4 d6d5 e4d5 e7d5 e3d2 c8b7 c4a5 a8b8 a1a4 c7c5 h4h5 g6g5 a5b7 b8b7 c3c4 d5f6 b2b4 c5b4 d2b4 f8e8 b4d6 g7f8 d6f8 g8f8 a4a6');
INSERT INTO evaluations VALUES(27141,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(27141,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(27141,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(27141,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(27141,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(27141,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(27141,36,'cp 30','d8d7 b1c3 d7g4 f3e5 g4d1 c3d1 d6e5 d1e3 g8f6 f2f3 f6d7 c1d2 d7c5 a2a4 f7f6 b2b4 c5e6 a4a5 h7h5 c2c3 f8e7 e1e2 e8f7 g2g3 a8d8 e3d5 c7c6 d5e7 f7e7 d2e3 a7a6 f3f4 e5f4 g3f4 g7g6 h1d1 d8d1 a1d1 h8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(27661,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(27661,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(27661,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(27661,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(27661,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(27661,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(27661,36,'cp 28','b1c3 d7g4 f3e5 g4d1 c3d1 d6e5 d1e3 g8f6 f2f3 f6d7 c1d2 d7c5 a2a4 c5e6 e3c4 f7f6 b2b4 h7h5 c2c3 h5h4 e1d1 g7g6 d1c2 f8h6 d2h6 h8h6 h1d1 e8e7 c4a5 a8b8 d1d5 h6h8 a1d1 h8d8 d1d2 e6f4 d5c5 d8d2 c2d2 f4e6');
INSERT INTO evaluations VALUES(28867,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(28867,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(28867,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28867,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28867,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(28867,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(28867,36,'cp 8','c2c4 d7b5 d4b5 g7g6 b1c3 f8g7 c1e3 b8c6 a1c1 e8g8 e1g1 a7a6 b5d4 c6d4 e3d4 b7b5 c3d5 a8c8 d5f6 g7f6 d4f6 e7f6 c4b5 c8c1 d1c1 a6b5 c1c6 b5b4 h2h4 g8g7 f1d1 d8a8 c6a8 f8a8 d1d4 a8a2 d4b4 a2a5 g2g4 d6d5 f2f3 d5e4 f3e4 f6f5 g4f5 g6f5 g1h2 f5e4 b4e4');
INSERT INTO evaluations VALUES(41808,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(41808,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41808,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(41808,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41808,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41808,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(41808,36,'cp 0','c7c5 d4c5 d6c5 b1d2 e8g8 g1f3 f6h5 c2c4 h5g3 h2g3 g7g6 e1g1 c5e7 a2a3 a7a5 a1c1 e7f6 b2b4 c8d7 c4d5 e6d5 d2b3 d7a4 d3c2 a4b5 c2d3');
INSERT INTO evaluations VALUES(42384,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(42384,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(42384,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42384,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42384,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(42384,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42384,36,'cp 28','e8g8 e1g1 h7h6 c4b3 b8c6 h2h3 d6c5 c2c3 d7d6 f1e1 c5b6 b1d2 c6a5 b3c2 c7c5 d2c4 b6c7 c4a5 c7a5 c1d2 f8e8 a2a4 c8d7 b2b4 c5b4 c3b4 a5c7 c2b3 a7a5 b3c4 d7e6 c4b5 e6d7 b5d7 d8d7 b4a5 c7a5 d2a5 a8a5 d1b3 d6d5');
INSERT INTO evaluations VALUES(37156,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(37156,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37156,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(37156,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(37156,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(37156,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(37156,36,'cp 55','e1g1 h7h6 c2c3 c7c5 d3d4 c5d4 c3d4 b8c6 d4d5 c6e7 d1d3 a7a6 a2a4 e7g6 c1e3 b7b6 b1c3 d6b4 d5d6 c8b7 c3d5 b7d5 c4d5 f6d5 d3d5 d8f6 f1d1 b4c5 h2h3 a8e8 e3c5 b6c5 d5c5 g6f4 c5e3 e8e6 g2g3 f6g6 f3h4 g6g5 h4f5 f4h3 g1h2 g5e3');
INSERT INTO evaluations VALUES(37155,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(37155,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(37155,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(37155,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(37155,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(37155,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(37155,36,'cp 31','h7h6 c2c3 c7c5 d3d4 c5d4 c3d4 b8c6 d4d5 c6e7 d1d3 e7g6 c1e3 a7a6 a2a4 b7b6 b1c3 f6g4 e3d2 d6c5 d5d6 c8b7 c4d5 b7c6 a1d1 d8b8 d5c6 d7c6 h2h3 g4f6 d2e3 f8d8 e3c5 b6c5 a4a5 f6e8 d6d7 e8f6 d3d6 f6d7 d6c6 b8b2 d1d7 b2c3 d7d8 a8d8 c6a6 g6f4');
INSERT INTO evaluations VALUES(44349,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(44349,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(44349,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(44349,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(44349,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(44349,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(44349,36,'cp 75','c4d5 d7d6 e2e4 g7g6 b1c3 f8g7 f3d2 b8d7 f1e2 e8g8 e1g1 f8e8 f1e1 a7a6 a2a4 g6g5 a4a5 h7h6 a1a3 a8b8 d1c2 d7e5 c3d1 c8d7 d2f1 c5c4 f1g3 d7b5 f2f3 e5d3 e2d3 c4d3 c2d2 f6d7 g3f5 d7c5 d1f2');
INSERT INTO evaluations VALUES(51502,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(51502,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(51502,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51502,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(51502,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51502,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51502,36,'cp 0','c7c5 b2b3 b8c6 c1b2 c6b4 b1a3 c8b7 f3e2 g8f6 e4e5 f6d5 g1f3 d5f4 e2e3 f4d5');
INSERT INTO evaluations VALUES(54752,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(54752,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(54752,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(54752,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(54752,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(54752,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(54752,36,'cp 178','c4d5 d8d5 b1c3 d5d8 g1f3 g8f6 e2e3 e7e6 f1c4 f8d6 e1g1 e8g8 f1e1 b8c6 e3e4 f5e4 c3e4 h7h6 c1d2 c8d7 a2a3 f6d5 b2b4 a7a6 d1e2 g8h8 e4c5 d6c5 d4c5 d8f6 a1d1 a8d8 c4a2 d7e8 e2e6 f6e6 e1e6 e8f7 e6e1 f7h5');
INSERT INTO evaluations VALUES(54745,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(54745,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(54745,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(54745,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(54745,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(54745,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(54745,36,'cp 185','d8d5 b1c3 d5d8 e2e3 g8f6 g1f3 e7e6 f1c4 f8d6 e1g1 b8c6 f1e1 e8g8 e3e4 f5e4 c3e4 h7h6 c1d2 g8h8 f3h4 f6e4 h4g6 h8g8 e1e4 f8f7 d1e1 d8f6 g6e5 d6e5 d4e5 f6f5 e1e2 c8d7 a1e1 a8f8 f2f3 b7b6 e4g4 f5h5 h2h3 c6e7 g4d4 e7d5 d4d5 e6d5 c4d5 f8e8');
INSERT INTO evaluations VALUES(34502,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(34502,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(34502,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(34502,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(34502,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(34502,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(34502,36,'cp 191','b1c3 d5d8 e2e3 e7e6 f1c4 g8f6 g1f3 b8c6 e1g1 f8d6 f1e1 e8g8 e3e4 f5e4 c3e4 h7h6 c1d2 a7a5 a1c1 c8d7 a2a3 g8h8 e4d6 c7d6 c4e6 d7e6 e1e6 d8d7 d1e2 a5a4 c1e1 a8e8 d4d5 c6e5 f3e5 d6e5 h2h3 e8e6 d5e6 d7e6');
INSERT INTO evaluations VALUES(4152,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(4152,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(4152,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(4152,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(4152,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(4152,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(4152,36,'cp -31','g5f3 f8e8 b1d2 c5f8 f1e1 d7d5 e4d5 f6d5 d2e4 d5b6 c1g5 d8d7 c4b3 c6a5 g5e3 a5b3 a2b3 b6d5 e3c5 f8c5 e4c5 d7c6 d3d4 d5f4 c5d3 f4d3 d1d3 e5e4 d4d5 c6g6 d3c4 c8h3 f3h4 g6f6 e1e4 e8e4 c4e4 h3d7');
INSERT INTO evaluations VALUES(4190,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(4190,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(4190,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(4190,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(4190,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(4190,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(4190,36,'cp -21','f8e8 b1d2 c6a5 d2b3 c5b6 b3a5 b6a5 c2c3 d7d6 b2b4 a5b6 a2a4 a7a5 d1b3 d8e7 h2h3 c8e6 c4e6 e7e6 b3e6 e8e6 f1e1 d6d5 c1d2 d5e4 d3e4 e6d6 g1f1 a8d8 d2e3 b6e3 e1e3 f6d7 b4a5 d8a8 a1b1 a8a5 b1b7 a5a4 b7c7');
INSERT INTO evaluations VALUES(6752,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(6752,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(6752,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(6752,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(6752,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(6752,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(6752,36,'cp 53','f7f5 d2d4 f5e4 f3e5 g8f6 e5g4 e8g8 e1g1 d6e7 d4d5 f6g4 d1g4 d7d6 g4e4 c6e5 b1d2 c8f5 e4e1 a7a6 b5e2 d8e8 e1d1 e7f6 d2b3 e5g6 e2h5 e8f7 b3d4 f6d4 d1d4 a8e8 f2f3 e8e5 c3c4 f5d7 h5g6 f7g6 c1f4 e5e2 f4g3');
INSERT INTO evaluations VALUES(7336,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7336,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(7336,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7336,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7336,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7336,36,'cp 0','f6g4 e1f1 e8g8 b1c3 g4f6 f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 c3d5 f6e4 f1e1 e5h2 g1h2 d8h4 h2g1 h4f2 g1h1 f2h4 h1g1');
INSERT INTO evaluations VALUES(8682,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(8682,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(8682,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(8682,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(8682,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(8682,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(8682,36,'cp 76','c7c6 e1c1 d8b6 g5f6 g7f6 d4d2 b6c5 c4b3 c8e6 c1b1 e8c8 h1e1 c8b8 f2f3 h8e8 g2g4 h7h5 h2h3 a7a6 c3e2 d6d5 e4d5 e6d5 e2d4 d5b3 a2b3 d8d5 c2c3 c5d6 d2c2 e8d8 b3b4 h5g4 f3g4 d6d7 c2h7 e7d6');
INSERT INTO evaluations VALUES(14291,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14291,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(14291,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(14291,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14291,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14291,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14291,36,'cp 38','g1f3 f6e4 f1b5 e7e6 c2c4 c8d7 e1g1 f8d6 b1c3 a7a6 b5c6 d7c6 c4c5 d6e7 f3e5 c6b5 c3b5 a6b5 c1d2 c7c6 d2e1 e8g8 a2a4 b5a4 a1a4 d8c7 e5d3 b7b6 f2f3 e4f6 c5b6 c7b6 e1h4 b6b5 a4a8 f8a8 d1c2 b5c4 f1c1 c4c2 c1c2 a8a1 g1h2 e7d6 d3e5');
INSERT INTO evaluations VALUES(14293,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(14293,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(14293,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(14293,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(14293,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(14293,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(14293,36,'cp 22','f6e4 a2a3 c8f5 c2c4 e7e6 d1b3 c6a5 b3a4 c7c6 c4c5 b7b5 a4d1 h7h5 b1d2 g7g6 d2e4 f5e4 c1d2 e4f3 g2f3 a5c4 a3a4 c4d2 d1d2 a7a6 f1d3 f8e7 f3f4 e8g8 h1g1 g8h8 e1f1 e7h4 d2c2 d8e7 b2b4 h4f6 a1a2 f6g7 f1g2 e7f6 a2a3 f8g8 a4b5');
INSERT INTO evaluations VALUES(25466,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(25466,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(25466,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(25466,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(25466,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(25466,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(25466,36,'cp 8','e1g1 f8d6 b5d3 e6e5 d4e5 c6e5 e3e4 e5f3 d1f3 d7g4 f3e3 d5d4 e3e1 d4c3 e4e5 e8g8 e1c3 d6e7 e5f6 e7f6 c3b4 d8c8 b4e4 g7g6 e4f4 f6g7 a1b1 f8e8 h2h3 g4e6 b2b3 e6d5 c1e3 g7e5 f4h4 c8e6 e3g5 e6c6 f2f3 f7f6 g5d2 a8d8 g1h1 b7b6 b1e1 d5f7 d2h6 b6b5');
INSERT INTO evaluations VALUES(25465,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(25465,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(25465,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(25465,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(25465,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(25465,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(25465,36,'cp 12','f8d6 b5d3 e6e5 d4e5 c6e5 e3e4 e5f3 d1f3 d7g4 f3e3 d5d4 e3e1 d4c3 e4e5 e8g8 e1c3 d6e7 e5f6 e7f6 c3c4 d8c8 f1e1 g4f5 d3f5 c8f5 c4c7 f8c8 c7b7 f5c2 c1e3 c2b2 b7a6 c8b8 g2g3 f6d4 e3d4 b2d4 a1d1 d4b6 a6a4 b8d8 e1e7 d8d1 a4d1 a8d8 d1a4 g7g6 e7a7 g8g7 a7d7 d8d7 a4d7 b6b1 g1g2 b1e4 g2g1');
INSERT INTO evaluations VALUES(27531,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(27531,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27531,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27531,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27531,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27531,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27531,36,'cp 0','c5d4 f1e1 g8h6 d1e2 a8c8 c1h6 g7h6 b1d2 b6b2 c3c4 c6b4 c4d5 b4d5 a1b1 b2a3 d2e4 d5c3 e4c3 d4c3 b1b7 a3c5 f3d4 c5d4 d3b5 d7b5 e2b5 e8d8 b5a5 d8e8 a5b5');
INSERT INTO evaluations VALUES(27552,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27552,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27552,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27552,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(27552,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27552,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(27552,36,'cp 0','f1e1 g8e7 b1d2 d4c3 b2c3 e7g6 d2b3 b6c7 d1e2 f8e7 a1b1 e8g8 h2h4 f7f6 d3g6 h7g6 c1f4 f6e5 f4e5 c6e5 f3e5 f8f6 b3d4 a8f8 e5d7 c7d7 d4e6 f8f7 e2b5 d7b5 b1b5 f6f2 b5b7 e7d6 e6d8 f7f8 d8e6');
INSERT INTO evaluations VALUES(56607,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(56607,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(56607,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56607,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(56607,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(56607,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(56607,36,'cp 14','f1c4 b8c6 g1f3 f8e7 d2d3 g8f6 e4g3 h7h6 e1g1 e8g8 a2a3 a7a5 h2h3 e7d6 a3a4 f8e8 f1e1 c8e6 c4b5 e6d7 c2c3 d7e6 c1e3 f6d7 b2b3 d7f6 a1c1 f6d5 e3d2 d5f6 c1b1 e6d7 b3b4 a5b4 c3b4 c6d4 f3d4');
INSERT INTO evaluations VALUES(17364,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(17364,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(17364,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(17364,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(17364,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(17364,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(17364,36,'cp -31','c4a2 b7b5 d2d3 g8f6 g1f3 f8d6 e1g1 c8b7 a3a4 d8b6 c3e2 e8g8 c2c3 c6e7 a2b1 e7g6 e2g3 b5a4 b1c2 a6a5 c2a4 d6c7 c1g5 f6g4 h2h3 h7h6 g5d2 g4e5 f3e5 g6e5 d2f4 d7d6 f4e5 d6e5 d1e2 g8h8');
INSERT INTO evaluations VALUES(42079,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(42079,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(42079,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(42079,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(42079,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(42079,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(42079,36,'cp 43','e8g8 a1c1 b8d7 f1e2 b7b6 c4d5 f6d5 e1g1 c7c5 d2e4 c5d4 d1d4 c8b7 e4d6 d8b8 d6c4 d5f6 f1d1 f8d8 b2b4 h7h6 a2a4 b7d5 a4a5 b6a5 b4a5 b8b4 d4a1 a8c8 c4e5 c8c1 d1c1 d7e5 f3e5 b4d2');
INSERT INTO evaluations VALUES(37060,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(37060,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(37060,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(37060,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(37060,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(37060,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(37060,36,'cp 39','a1c1 b8d7 f1e2 b7b6 c4d5 f6d5 e1g1 c7c5 d2e4 c5d4 d1d4 c8b7 e4d6 d8b8 d6c4 f8d8 f1d1 h7h6 h2h3 d5e7 d4d6 e7f5 d6b8 a8b8 g2g4 f5e7 f3d4 d7c5 b2b4 c5a4 c4e5 e7d5 e5c6 b7c6 d4c6 a4c3 e2f3 c3d1');
INSERT INTO evaluations VALUES(48581,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(48581,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(48581,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(48581,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(48581,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(48581,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(48581,36,'cp -2','f6d5 f1c4 d5b4 d2d3 c8e6 d1e2 b8c6 g1f3 f8e7 e1g1 d8d7 a2a3 b4d5 c4d5 e6d5 f3e5 c6e5 e2e5 d5e6 c1f4 e8g8 e5c7 f8d8 c7d7 d8d7 f1d1 e6b3 d1d2 a8d8 f4e3 d7d3 d2d3 d8d3 e3a7 d3d2 a1b1 d2c2 a7e3 e7f6 e3c5 b3c4 c5e3 f6c3');
INSERT INTO evaluations VALUES(48990,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48990,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(48990,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(48990,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48990,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(48990,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(48990,36,'cp -10','f1c4 d5b6 c4b3 b8c6 g1e2 f8d6 d2d4 e8g8 e1g1 c8f5 c1e3 c6a5 b3c2 f5g6 e2f4 g6c2 d1c2 a5c6 a1d1 f8e8 d4d5 c6e5 f4e2 e5g4 e3f4 g4f6 a2a4 a7a6 c2b3 d6f4 e2f4 d8d6 g2g3 b6d7 f1e1 d7c5 b3c4 e8e1 d1e1 a8e8');
INSERT INTO evaluations VALUES(48979,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48979,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(48979,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(48979,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48979,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(48979,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(48979,36,'cp -20','d5b6 c4b3 b8c6 g1e2 f8d6 d2d4 e8g8 e1g1 c8f5 c1e3 c6a5 b3c2 f5g6 e2f4 g6c2 d1c2 a5c6 a1d1 d8f6 d4d5 c6e5 c3b5 e5g4 b5d6 f6d6 c2e2 g4e3 e2e3 f8e8 e3f3 b6d7 g2g3 a8d8 g1g2 d7f6 a2a3 h7h6 f3b3 d6b6');
INSERT INTO evaluations VALUES(52391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52391,36,'cp 0','d2d3 d7d5 e2e4 d5e4 d3e4 d8d1 e1d1 g8f6 e4e5 f6d5 g1f3 c8g4 c2c4 d5c7 b1d2 h7h5 d2b3 c7e6 d1c2 h5h4 f3d4 e6d4 b3d4 e7e6 h2h3 g4h5 f1e2 h5g6 e2d3 g6h5');
INSERT INTO evaluations VALUES(54342,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(54342,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(54342,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(54342,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54342,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(54342,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54342,36,'cp 31','e3d4 f8b4 b1c3 g8e7 f1d3 d5c4 d3c4 e8g8 c4d3 b8d7 e1g1 h7h6 f1e1 d7f6 f3e5 e7f5 d1f3 d8d4 c1d2 c8d7 a2a3 b4a5 a1d1 d7a4 c3a4 d4a4 f3b7 a5d2 d1d2 a4f4 d2c2 f5d4 c2c4 f4d2 b7b4 d2b4 c4b4 f8d8 b4b7 d4c6 e5c6 d8d3');
INSERT INTO evaluations VALUES(11834,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(11834,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(11834,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(11834,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(11834,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(11834,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(11834,36,'cp 65','f3e5 d5e4 d2d4 f8b4 e1g1 b4c3 b2c3 e8g8 f1e1 d8d6 f2f3 e4f3 d1f3 c8e6 a2a4 e6d5 f3f1 c6c5 c1a3 f6e4 e1e4 d5e4 a3c5 d6d5 c3c4 d5e6 c5f8 a8f8 f1e2 e4g6 e2d2 f8e8 c4c5 g6h5 e5d3 e6c4 d3f4 h5g6 c2c3 a7a5 h2h3 g6e4 d2e3 h7h6 g1h2 f7f6');
INSERT INTO evaluations VALUES(11829,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(11829,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(11829,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(11829,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(11829,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(11829,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(11829,36,'cp 66','d5e4 d2d4 f8b4 e1g1 b4c3 b2c3 e8g8 f1e1 d8d6 c1g5 f8e8 f2f3 f6d7 e1e4 d7e5 e4e5 e8e5 d4e5 d6c5 d1d4 c5b6 g5e3 c8e6 a1d1 b6d4 e3d4 e6a2 d1a1 a2e6 a1a6 h7h5 h2h3 e6f5 g1f2 f5c2 g2g4 h5g4 h3g4 c2b3 f3f4 c6c5 d4c5');
INSERT INTO evaluations VALUES(28267,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(28267,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(28267,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(28267,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(28267,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(28267,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(28267,36,'cp 19','f1d3 e8g8 e1g1 d7d5 c4d5 e6d5 f3e5 b8d7 d1c2 d8e7 c3b5 a7a6 c2c7 d7e5 d4e5 f6g4 c7e7 b4e7 d3e2 g4e5 b5d4 e5c6 c1d2 c6d4 e3d4 f8c8 f1c1 c8c1 a1c1 a8c8 c1e1 g8f8 e2d3 a6a5 d2f4 e7b4 e1c1 c8c1 f4c1 b7c6 c1f4 b4e7 f2f3 e7f6');
INSERT INTO evaluations VALUES(28273,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(28273,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(28273,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(28273,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(28273,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(28273,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(28273,36,'cp 14','e8g8 e1g1 d7d5 c4d5 e6d5 f3e5 b8d7 d1c2 f8e8 c3b5 d7e5 d4e5 e8e5 c2c7 d8c7 b5c7 a8c8 c7b5 a7a6 b5d4 f6e4 f1d1 g7g6 a2a3 b4f8 c1d2 e4d2 d1d2 c8c5 h2h4 e5e8 h4h5 a6a5 a1d1 e8c8 g2g3 c5c1 g1g2 c1d1 d2d1 g6h5 d3e2');
INSERT INTO evaluations VALUES(27779,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(27779,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(27779,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(27779,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(27779,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(27779,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(27779,36,'cp 25','e1g1 d7d5 c4d5 e6d5 f3e5 b8d7 f2f4 c7c5 c1d2 f6e4 c3e4 d5e4 d3c4 b4d2 d1d2 d7e5 f4e5 b7d5 b2b3 c5d4 e3d4 a8c8 a1c1 d8d7 d2e3 f8d8 c4d5 d7d5 c1c8 d8c8 f1f4 c8c2 e3e4 d5e4 f4e4 c2d2 a2a4 g8f8 e4h4 h7h5 b3b4 d2a2 a4a5 b6a5 b4a5 a2a4 h2h3');
INSERT INTO evaluations VALUES(27778,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(27778,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(27778,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(27778,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(27778,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(27778,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(27778,36,'cp 24','d7d5 a2a3 b4c3 b2c3 b7a6 c4d5 a6d3 d1d3 d8d5 c1b2 c7c5 c3c4 d5e4 d3b3 b8c6 b2c3 f8d8 d4c5 b6c5 b3b5 a8c8 h2h3 e6e5 b5a6 e4f5 a1d1 d8d1 f1d1 f6e4 c3a5 h7h6 d1d5 f5e6 f3d2 e4d2 a5d2');
INSERT INTO evaluations VALUES(27772,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(27772,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(27772,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(27772,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(27772,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(27772,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(27772,36,'cp 15','a2a3 b4d6 b2b4 d5c4 d3c4 b8d7 c1b2 a8c8 a1c1 c7c5 d4c5 d6b8 c5c6 c8c6 d1e2 d7e5 f3e5 b8e5 f2f4 e5c3 b2c3 f6e4 c3a1 d8e7 e2b2 e4f6 c4e2 f8d8 c1c6 b7c6 f1c1 e7b7 e2f1 c6e4 b2e2 h7h6 a1d4 d8c8 c1c3 e4d5 c3c8 b7c8 h2h3 c8b7 b4b5');
INSERT INTO evaluations VALUES(28280,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(28280,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(28280,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(28280,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(28280,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(28280,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(28280,36,'cp 134','d1c2 b4c3 b2c3 c7c5 g1e2 d7d6 e1g1 b8c6 f2f4 c5d4 c3d4 e8g8 c2b1 a8c8 a2a4 b7a6 e4e5 d6e5 f4e5 f6h5 c1a3 f8e8 a3d6 d8h4 a4a5 c6a5 d3h7 g8h8 f1f7 a6c4 h7d3 h4g5 f7f3 c4d5 d3e4 a5c4 e4d5 e6d5 b1d3 c4d6 e5d6 h5f6 e2f4 h8g8');
INSERT INTO evaluations VALUES(28281,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(28281,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(28281,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(28281,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(28281,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(28281,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(28281,36,'cp 125','b4c3 b2c3 c7c5 g1e2 d7d6 e1g1 b8d7 f2f4 f6h5 a2a4 e8g8 h2h3 d8h4 a4a5 b6a5 c1e3 h5g3 f1b1 f8b8 e2g3 h4g3 c2f2 g3f2 g1f2 a5a4 a1a4 a7a5 b1a1 f7f6 f2g1 g8f7 g2g4 e6e5 f4e5 f6e5 d4d5 h7h6 a4a5 a8a5 a1a5 b8a8 a5a8 b7a8');
INSERT INTO evaluations VALUES(39117,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(39117,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(39117,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(39117,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(39117,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(39117,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(39117,36,'cp 62','g8f6 d4f5 d6b4 c1d2 b4d2 d1d2 e8g8 d2g5 g7g6 b1c3 d7d5 e1c1 c8f5 e4f5 c7c6 h2h4 g8h8 f2f3 b8d7 c3e2 d8e7 e2f4 a8e8 f5g6 f7g6 h4h5 g6h5 f4h5 f8g8 g5d2 f6h5 h1h5 e7e3 d2e3 e8e3 c1d2 e3e7 d1e1 g8e8 e1e7 e8e7 b2b4 h8g7 a2a4 d7e5 f1e2 e5g6 g2g3 e7e5 h5e5 g6e5 f3f4 e5d7 a4a5 d7f6');
INSERT INTO evaluations VALUES(42398,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(42398,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(42398,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(42398,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(42398,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(42398,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(42398,36,'cp 75','d4f5 d6b4 c2c3 b4f8 e4e5 d7d6 f5d6 f8d6 e5d6 c7d6 f1e2 b8c6 b1a3 e8g8 e1g1 f8e8 c1e3 a7a6 f1e1 h7h6 e2f1 c8e6 f2f3 a8c8 d1d2 c6e5 e3d4 f6h5 a3c2 e5c4 f1c4 c8c4 b2b3 c4c8 a1d1 d8a5 c3c4 a5d2 d1d2');
INSERT INTO evaluations VALUES(43130,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43130,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43130,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43130,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43130,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43130,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43130,36,'cp -14','e8g8 e2e3 d7d5 f1e2 h7h6 g5h4 f6e4 c3e4 d5e4 f3e5 c7c5 c2c3 g7e5 d4e5 b8c6 d1d8 f8d8 f2f4 c8e6 g2g4 c5c4 a1d1 b7b5 a2a3 d8d1 e2d1 e6d5 h4g3 e7e6 h2h4 a8b8 e1f2 g8g7');
INSERT INTO evaluations VALUES(37362,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(37362,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(37362,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37362,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37362,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37362,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37362,36,'cp 0','e2e3 d7d5 f1e2 h7h6 g5h4 c8e6 e1g1 b8d7 f3e5 c7c5 f2f4 c5d4 e3d4 d8b6 a2a4 b6b2 c3b5 f8c8 e2d3 a7a6 a1b1 b2a2 b1a1');
INSERT INTO evaluations VALUES(43447,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(43447,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(43447,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(43447,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(43447,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(43447,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(43447,36,'cp 29','c8e6 c4b3 b8d7 c1e3 f8e7 e1g1 b7b5 a2a4 b5b4 c3d5 f6d5 b3d5 e6d5 d1d5 e8g8 d5d3 d8c7 f3d2 f8c8 f1c1 h7h6 c2c3 e7g5 e3g5 h6g5 c3b4 c7c1 a1c1 c8c1 d2f1 d7f6 h2h3 a8c8 d3a6 c8c2 a6d6 c2b2 b4b5 g5g4 h3g4 f6g4 d6d8 g8h7');
INSERT INTO evaluations VALUES(44173,30,'cp -112',NULL);
INSERT INTO evaluations VALUES(44173,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(44173,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(44173,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(44173,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(44173,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(44173,36,'cp -92','h5e2 b8c6 g1f3 f6e4 e2e4 d7d5 c4d5 d8d5 d2d3 b7b6 e4d5 e6d5 b1c3 c8e6 f3g5 e8d7 c1d2 f8d6 h2h4 h7h6 g5e6 f7e6 c3e2 b6b5 h4h5 h8f8 f2f4 b5b4 e1g1 a7a5 g2g4 a5a4 g1g2 e6e5 f4f5 a4a3 b2b3 f8f7 a1d1 a8f8 d1e1');
INSERT INTO evaluations VALUES(8051,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(8051,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(8051,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8051,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(8051,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8051,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(8051,36,'cp -5','f6g4 d4c6 d7c6 d2d8 e8d8 e3g5 g7f6 g5f6 g4f6 f1c4 h8f8 a2a4 a7a5 e1c1 d8c7 f2f3 f6d7 h2h4 h7h5 d1d2 d7c5 h1d1 c8d7 b2b3 e7e5 c1b2 a8d8 c3b1 f7f6 d2d6 d7c8 d6d8 f8d8 d1d8 c7d8 b2c3 d8e7');
INSERT INTO evaluations VALUES(8067,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(8067,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8067,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8067,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(8067,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(8067,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8067,36,'cp 0','d4c6 d7c6 d2d8 e8d8 e3f4 g7e5 f4e5 g4e5 e1c1 d8c7 f2f4 e5d7 e4e5 d7b6 f1e2 h7h5 g2g3 c8f5 h1e1 a7a5 e2f3 f5g4 f3e2');
INSERT INTO evaluations VALUES(7995,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7995,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7995,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7995,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7995,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7995,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7995,36,'cp 0','d7c6 d2d8 e8d8 e3f4 g7e5 f4e5 g4e5 f2f4 e5d7 e1c1 d8c7 e4e5 d7b6 g2g3 c8f5 f1d3 a8d8 h2h4 f5d3 d1d3 d8d3 c2d3 h8d8 c1d2 h7h5 h1d1 b6d5 c3e4 d5b4 d2c3 b4d5');
INSERT INTO evaluations VALUES(15754,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(15754,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(15754,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(15754,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(15754,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(15754,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(15754,36,'cp 25','h7h6 g5f4 g7g5 f4e3 f8g7 b1c3 e8g8 d4d2 f6g4 e3d4 g7d4 d2d4 c8e6 c4f1 d8f6 d4d2 d6d5 h2h3 d5d4 h3g4 d4c3 d2c3 f6c3 b2c3 e6g4 h1h6 g8g7 h6h1 f8h8 h1g1 g4e6 f1d3 a8d8 f2f3 g7f6 a2a4 a7a5 e1e2 b7b6 g1h1 f6e5 e2e3 e6d7 h1h8 d8h8 g2g3');
INSERT INTO evaluations VALUES(22311,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(22311,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(22311,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(22311,33,'cp 204',NULL);
INSERT INTO evaluations VALUES(22311,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(22311,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(22311,36,'cp 200','h2h3 g8e7 c1e3 a7a6 f1e2 g7g6 a2a3 f6f5 d4e5 d6e5 d1d8 c6d8 e1c1 f5f4 e3c5 e7c6 c5f8 h8f8 c3d5 d8e6 e2c4 c8d7 d5b4 e6d4 f3d4 c6d4 b4d3 d4c6 d3c5 d7c8 c4d5 f8f6 f2f3 c6d4 c2c3 d4e2 c1c2 e2g3 h1g1 e8e7 d5b7 c8b7');
INSERT INTO evaluations VALUES(41438,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41438,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(41438,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(41438,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41438,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41438,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41438,36,'cp 1','e8g8 a1c1 c8e6 e2e3 d5c4 f3g5 e6d5 e3e4 h7h6 e4d5 h6g5 f4g5 f6d5 f1c4 d5b6 c4b3 d8d4 d1d4 g7d4 c3b5 d4b2 c1c7 b8c6 e1g1 e7e6 c7b7 c6a5 b7a7 a8a7 b5a7 f8a8 f1b1 b2g7 b3c2 a5c4 a7c6 a8a2 c2b3 a2a1 b1a1 g7a1 g1f1 g8g7 b3c4 b6c4 f2f3 a1f6 g5f4 e6e5 f4g3 c4e3 f1f2 e3c4 f2e2');
INSERT INTO evaluations VALUES(52052,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(52052,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52052,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(52052,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(52052,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52052,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(52052,36,'cp 15','e3d4 g8f6 b1c3 d7d5 c4d5 e6d5 f1d3 h7h6 g1f3 f8b4 h2h3 e8g8 e1g1 b4d6 c3b5 b8c6 b5d6 d8d6 f1e1 f6e4 d3e4 d5e4 e1e4 d6d5 e4e1 f8d8 c1e3 c8f5 a2a3 f5e4 a1c1 e4f3 d1f3 d5f3 g2f3 d8d7 e1d1 a8d8 g1g2 c6d4');
INSERT INTO evaluations VALUES(4246,30,'cp 289',NULL);
INSERT INTO evaluations VALUES(4246,31,'cp 302',NULL);
INSERT INTO evaluations VALUES(4246,32,'cp 272',NULL);
INSERT INTO evaluations VALUES(4246,33,'cp 291',NULL);
INSERT INTO evaluations VALUES(4246,34,'cp 289',NULL);
INSERT INTO evaluations VALUES(4246,35,'cp 293',NULL);
INSERT INTO evaluations VALUES(4246,36,'cp 283','d4d5 c6e7 e4e5 f6g4 h2h3 g4f2 f1f2 b6f2 g1f2 d7d6 c1f4 b7b5 c4b5 d6e5 f4e5 c8b7 b1c3 e7d5 c3d5 b7d5 b5f1 c7c6 e5c3 f7f6 f2g1 d8b6 d1d4 a8d8 d4f2 b6f2 g1f2 f8e8 a2a4 d5e4 f1c4 e4d5 c4e2 g8h8 a4a5 h7h6 a5a6 d5e4 a1e1 d8b8 f3d4 e4d5 d4f5 e8e4 f5e3 e4f4 f2g1 f4e4');
INSERT INTO evaluations VALUES(17984,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(17984,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(17984,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(17984,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(17984,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(17984,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(17984,36,'cp 56','a4c2 f8e7 d2d4 g8f6 e1g1 e8g8 b1d2 f8e8 a2a3 e7f8 f1e1 h7h6 h2h3 a6a5 b2b3 c6b8 d2f1 b8d7 f1g3 c7c5 c1e3 d8c7 d4d5 c5c4 b3c4 b5c4 f3d2 a8b8 c2a4 c8a6 g3f5 e8c8 a4c6 b8b2 a3a4 c8b8 a1c1 d7c5 e3c5');
INSERT INTO evaluations VALUES(17981,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(17981,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17981,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(17981,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(17981,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(17981,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(17981,36,'cp 45','f8e7 e1g1 g8f6 d2d4 e8g8 b1d2 f8e8 a2a3 h7h6 f1e1 e7f8 h2h3 a8b8 d2f1 a6a5 b2b3 d6d5 e4d5 d8d5 d4e5 d5d1 c2d1 f6d5 c1d2 b5b4 a3b4 a5b4 f3d4 c6d4 c3d4 c8e6 d1f3 d5c3 e1c1 e8d8 d2c3 b4c3 c1c3 d8d4 c3c7 b8b3');
INSERT INTO evaluations VALUES(18767,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(18767,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(18767,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(18767,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(18767,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(18767,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(18767,36,'cp -37','g8f6 f1e2 d7d5 c4d5 e6d5 e4d5 f6d5 e1g1 f8e7 d3d4 e8g8 d4c5 e7c5 c1g5 c5e7 e2c4 e7g5 f3g5 d8g5 d1d5 g5f4 d5c5 c6e5 c4e2 f8e8 b1c3 b7b6 c5e3 f4e3 f2e3 e5g4 c3d5 g4e3 d5c7 e3f1 c7e8 f1e3 e8c7 a8b8');
INSERT INTO evaluations VALUES(12397,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(12397,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(12397,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(12397,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(12397,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(12397,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(12397,36,'cp -26','f1e2 d7d5 c4d5 e6d5 e1g1 f8e7 e4d5 f6d5 d3d4 e8g8 d4c5 h7h6 b1d2 e7c5 d2e4 c5e7 d1b3 d5b4 f1d1 d8c7 a2a3 c8e6 b3a4 b4d5 c1d2 f8d8 h2h3 d8d7 e2b5 d5b6 a4c2 a8d8 b5c6 b7c6 d2a5 d7d1 a1d1 d8d1 c2d1 e6d5');
INSERT INTO evaluations VALUES(31089,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(31089,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(31089,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31089,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(31089,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(31089,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(31089,36,'cp 0','c1f4 g8f6 e2e3 e7e6 f1d3 f5g6 e1g1 f8e7 h2h3 e8g8 d1e2 d8b6 f1b1 a7a5 d3g6 h7g6 a2a4 b8d7 f4h2 a8c8 h2f4');
INSERT INTO evaluations VALUES(31076,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(31076,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(31076,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31076,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(31076,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(31076,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31076,36,'cp 0','g8f6 e2e3 e7e6 f1d3 f5g6 e1g1 f8e7 d3g6 h7g6 d1e2 d8b6 e3e4 e8g8 f1b1 c6c5 d4c5 e7c5 c3a4 b6a6 e2a6 b8a6 a4c5 a6c5 e4d5 f6d5 f4d6 f8c8 c2c4 d5b4 b1d1 b4d3 d6c5 d3c5 b2b4 c5a6 f3e5 a6b4 d1d7 f7f6 e5g6 c8c4 d7b7 b4c6 a1b1');
INSERT INTO evaluations VALUES(28978,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(28978,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(28978,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(28978,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28978,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28978,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28978,36,'cp 1','e2e3 e7e6 f1d3 f8e7 e1g1 f5g6 f1e1 f6h5 d3g6 h7g6 f4b8 a8b8 e3e4 h5f6 a2a3 b8c8 e4e5 f6d7 g2g3 d7b6 g1g2 e8g8 c3e2 b6c4 b2b4 b7b6 h2h4 c6c5 c2c3 a7a5 e2f4 c5d4 c3d4 b6b5 f4d3 d8c7 d1e2');
INSERT INTO evaluations VALUES(28976,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28976,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28976,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(28976,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28976,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(28976,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(28976,36,'cp 17','e7e6 f1d3 f5g6 e1g1 f8e7 f1e1 f6h5 d3g6 h7g6 f4b8 a8b8 e3e4 h5f6 a2a3 b8c8 e4e5 f6d7 g2g3 g6g5 g1g2 c6c5 h2h3 c5d4 c3e2 d8b6 a1b1 d7c5 e2d4 c5e4 c2c3 e7c5 d1e2 c5d4 c3d4 e8g8 b1c1 b6b3 f3d2 e4d2 e2d2 c8c1');
INSERT INTO evaluations VALUES(29253,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(29253,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29253,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(29253,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29253,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(29253,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(29253,36,'cp 15','f1d3 f8e7 e1g1 b8d7 h2h3 e8g8 d3f5 e6f5 b2b3 g7g6 c3e2 a8c8 f4h6 f8e8 c2c4 c6c5 a1c1 c5d4 d1d4 d5c4 c1c4 c8c4 d4c4 d7c5 f1c1 d8d5 c4d5 f6d5 e2f4 e8d8 f4d5 d8d5 f3d4 a7a5 g1f1 g6g5 h3h4');
INSERT INTO evaluations VALUES(29255,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(29255,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(29255,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(29255,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29255,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(29255,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(29255,36,'cp 10','b8d7 e1g1 f8e7 h2h3 e8g8 d3f5 e6f5 b2b3 a8c8 c3e2 d8a5 d1d3 g7g6 f3e5 d7e5 f4e5 f6e4 c2c4 f8e8 a1d1 e7d6 e5d6 e4d6 c4c5 d6e4 a2a3 a7a6 b3b4 a5b5 f2f3 e4g5 e2c3 b5d3 d1d3 g5e6 d3d1');
INSERT INTO evaluations VALUES(41648,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(41648,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(41648,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(41648,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(41648,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(41648,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(41648,36,'cp -12','d5c4 h4f6 d8f6 d1a4 b8c6 a2a3 b4c3 b2c3 e8g8 e2e3 f8d8 f1e2 c8d7 a4c4 a8c8 e1g1 c6a5 c4b4 b7b6 f3e5 d7e8 b4b2 c7c5 a3a4 a5c6 e5c4 f6e7 f1d1 c5d4 c3d4 e7b4 d1c1 b4b2 c4b2 c6b4 b2c4 b4d5 e2f1 d5f6 h2h4 f6d5 f1d3 c8c7 a4a5');
INSERT INTO evaluations VALUES(48678,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(48678,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(48678,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(48678,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(48678,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(48678,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48678,36,'cp -33','f8d6 b1c3 e8g8 d1c2 c7c6 f1d3 f8e8 e1g1 d8e7 f1e1 f6e4 f3e5 d6e5 d4e5 e7e5 f2f3 e4f6 c1d2 b8d7 e3e4 d7c5 c3d5 c5d3 d5f6 e5f6 c2d3 f6b2 a2a4 b2b6 d2e3 b6a5 e1c1 c8e6 d3a3 e8d8 c1c5 a5a6 g1f2 d8d7 c5c3 h7h6 a3c5 a8d8 a4a5');
INSERT INTO evaluations VALUES(41718,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(41718,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(41718,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(41718,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(41718,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(41718,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(41718,36,'cp -20','d1c2 e8g8 b1c3 f8e8 f1d3 c7c6 e1g1 d8e7 f1e1 f6e4 f3e5 d6e5 d4e5 e7e5 f2f3 e4c3 b2c3 b8d7 c1b2 e5g5 d3h7 g8h8 f3f4 g5h4 h7d3 d7c5 c3c4 c5d3 c2d3 d5c4 d3c4 c8e6 c4c2 h8g8 a1c1 a8c8 c2c3 f7f6 e3e4 h4f4 e1f1');
INSERT INTO evaluations VALUES(50106,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(50106,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(50106,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(50106,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(50106,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(50106,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(50106,36,'cp 67','g1f3 e7e6 c2c4 f8e7 b1c3 e8g8 f1g2 d7d5 e1g1 c7c6 d1c2 b8d7 f3g5 d7b8 f1d1 a7a5 b2b3 h7h6 g5f3 c8d7 c3a4 d7e8 f3e5 b8d7 e5d3 e8h5 f2f3 f8e8 c1f4 h5f7 c4c5 g7g5 f4e5 d7e5 d3e5 f6d7 e5f7 g8f7');
INSERT INTO evaluations VALUES(51481,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(51481,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(51481,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(51481,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(51481,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(51481,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(51481,36,'cp 26','e7e5 g1f3 b8c6 b1c3 g8f6 a2a3 g7g6 g2g3 d7d6 d2d3 f8g7 f1g2 c8g4 h2h3 g4f3 g2f3 c6d4 f3g2 a7a5 h3h4 e8g8 e1g1 f6d7 a1b1 d7c5 b2b4 a5b4 a3b4 c5e6 h4h5 g8h8 b4b5 g6h5 d1h5 d8f6 c3d5 f6g6 h5h4 d4e2 g1h1 e2c1');
INSERT INTO evaluations VALUES(51442,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(51442,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(51442,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(51442,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(51442,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51442,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(51442,36,'cp 36','g1f3 b8c6 b1c3 g8f6 f1e2 f8c5 a2a3 e8g8 e1g1 f8e8 d2d3 a7a5 c1e3 d7d6 d1d2 h7h6 a1e1 f6d7 e2d1 d7f8 c3d5 a5a4 d3d4 e5d4 f3d4 c6d4 e3d4 c7c6 d5b4 c8b7 d4c3 d8g5 d2g5 h6g5 b4d3 f8e6 d3c5 e6c5 d1f3 b7a6 e4e5 a6c4 f3c6 c4f1');
INSERT INTO evaluations VALUES(54814,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(54814,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(54814,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(54814,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(54814,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(54814,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(54814,36,'cp 49','g8f6 b1c3 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 b8a6 c1g5 h7h6 g5e3 d8e7 d4d5 f6g4 e3c1 f7f5 h2h3 g4f6 f3h4 g8h7 e4f5 c8f5 h4f5 g6f5 a1b1 a6c5 d1c2 f6e4 c3e4 c5e4 e2d3');
INSERT INTO evaluations VALUES(18353,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(18353,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(18353,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(18353,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(18353,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(18353,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(18353,36,'cp 199','d4d5 c6a5 b1c3 e7e5 a2a3 a7a6 f2f4 d7d6 g1f3 e5f4 b2b4 a5b7 c1f4 g8e7 f1d3 e7g6 f4g3 f8e7 e1g1 e7f6 d1d2 e8g8 a1e1 c8g4 g1h1 f8e8 c3e2 f6e5 f3e5 d6e5 d2e3 f7f6 c2c4 g4e2 d3e2 g6f4 g3f4 e5f4 f1f4 b7d6 c4c5 b6c5');
INSERT INTO evaluations VALUES(18348,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(18348,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(18348,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(18348,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(18348,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(18348,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(18348,36,'cp 195','c6a5 b1c3 e7e6 d1f3 a7a6 h2h4 g8f6 h4h5 d7d6 h5h6 g7h6 d5e6 c8e6 c1h6 f8h6 h1h6 f6g8 h6h1 d8f6 f3e3 a5c4 f1c4 e6c4 g1f3 e8c8 e4e5 f6e6 b2b3 c4b5 e1c1 g8e7 f3d4 e6e5 e3f3 e5c5 c1b2 d6d5 f3f7 h8f8 f7e6 b5d7 e6e5 f8h8 d1d2 e7c6 e5d5 c5d5 c3d5 c6d4');
INSERT INTO evaluations VALUES(30426,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(30426,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(30426,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(30426,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(30426,34,'cp 241',NULL);
INSERT INTO evaluations VALUES(30426,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(30426,36,'cp 217','a4b5 a6b7 g1f3 g8f6 f1c4 e7e6 d1e2 c7c5 e1g1 f8d6 b2b3 e8g8 f1d1 b8d7 b1d2 c5d4 f3d4 d8c7 h2h3 d6h2 g1h1 h2d6 c1b2 a8c8 h1g1 d6h2 g1f1 h2d6 a1c1 d6b4 f1g1 c7b6 c4d3 c8c1 d1c1 d7c5 d2c4 b6c7 d3b1');
INSERT INTO evaluations VALUES(30423,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(30423,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(30423,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(30423,33,'cp 235',NULL);
INSERT INTO evaluations VALUES(30423,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(30423,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(30423,36,'cp 240','a6b7 g1f3 g8f6 f1c4 e7e6 d1e2 c7c5 e1g1 f8d6 b2b3 e8g8 c1b2 b8d7 b1d2 d8b8 d4c5 d6c5 f1d1 f8d8 h2h3 c5f8 f3d4 d7e5 a1c1 b8d6 e2e1 a8c8 c4f1 a7a6 c1c8 d8c8 b5a6 b7a6 d4b5 a6b5 f1b5');
INSERT INTO evaluations VALUES(32042,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(32042,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(32042,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(32042,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(32042,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(32042,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(32042,36,'cp 67','h2h3 g4d7 d3d4 b8c6 b1c3 g8f6 c1e3 e5d4 f3d4 f8e7 a2a4 e8g8 e1g1 c6d4 d1d4 f8e8 d4d3 d7c6 c4b3 f6d7 c3d5 e7h4 f2f3 d7f6 d5c3 f6d7 a1d1 h4f6 e3f2 d8e7 c3d5 c6d5 b3d5 f6b2');
INSERT INTO evaluations VALUES(32040,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(32040,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(32040,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(32040,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32040,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(32040,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(32040,36,'cp 65','g4d7 b1c3 g8f6 c1e3 f8e7 g2g4 b8c6 a2a3 f6h7 d1e2 h7g5 e1c1 g5f3 e2f3 d7e6 c1b1 e7g5 f3e2 d8d7 c4b5 g5e3 e2e3 a7a6 b5c6 b7c6 d3d4 f7f6 f2f4 e5f4 e3f4 e8g8 h1f1 a8b8 b2b3 f8e8 b1b2 g8h8 f1f3 e6g8');
INSERT INTO evaluations VALUES(53809,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(53809,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53809,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(53809,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(53809,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(53809,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(53809,36,'cp 16','f8e7 e1g1 g8f6 a2a4 e8g8 b1c3 c8e6 c4e6 f7e6 d3d4 e5d4 f3d4 d8d7 d1e2 c6c5 d4f3 b8c6 e4e5 d6e5 f3e5 c6e5 e2e5 f6d5 e5e1 d5c3 e1c3 f8f5 c3c4 a8d8 f1e1 g8f7 c2c3 h7h6 c4e2 a7a6 h2h3 e7f6 c1e3 d7d3 e2d3 d8d3 a1d1 d3d5');
INSERT INTO evaluations VALUES(37889,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(37889,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(37889,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(37889,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(37889,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37889,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(37889,36,'cp 28','a2a4 g8f6 e1g1 e8g8 f1e1 c8g4 b1d2 b8d7 c4a2 d7c5 a4a5 c5e6 a5a6 b7b5 h2h3 g4h5 d2f1 f8e8 f1g3 h5g6 b2b4 e7f8 a2b3 d8c7 c2c4 b5c4 d3c4 a8b8 c1d2 e6d4 f3d4 e5d4 b4b5 c7b6 d2a5 b6c5 b3a4 c6b5 c4b5 f6e4 g3e4');
INSERT INTO evaluations VALUES(8082,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(8082,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8082,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8082,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8082,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(8082,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(8082,36,'cp 17','f3e5 f6d7 e5d7 c8d7 f1e2 e8g8 e1g1 d7e8 f4g3 b7b6 a2a4 a7a6 b2b3 f7f5 h2h3 h7h6 f1e1 e7d6 g3d6 d8d6 d2f3 a6a5 d1d2 a8d8 e2b5 e8h5 b5c6 h5f3 c6b5 f3h5 b5e2 h5g6 e2d3 e6e5 d4e5 d6e5 c3c4');
INSERT INTO evaluations VALUES(14915,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(14915,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(14915,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(14915,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(14915,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(14915,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(14915,36,'cp 114','f1c4 a7a6 a2a3 b7b5 c4d3 f8d6 d1c2 c8b7 e3e4 e6e5 c3b5 a6b5 d3b5 e8g8 b5c6 b7c6 c2c6 e5d4 e1g1 f8e8 f1e1 f6g4 c1g5 d8b8 b2b4 d4d3 h2h3 g4e5 f3e5 d6e5 a1a2 b8b6 c6c4 b6e6 c4e6 e8e6 e1d1 h7h6 g5e3 e5d6 b4b5 e6e4 d1d3 e4a4');
INSERT INTO evaluations VALUES(19474,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(19474,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(19474,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(19474,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19474,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(19474,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(19474,36,'cp -16','g7g6 f1c4 e7e6 d2d3 d7d6 c1e3 g8f6 g1f3 f8g7 c4a2 e8g8 h2h3 d6d5 e4d5 f6d5 c3d5 e6d5 e1g1 g7b2 e3c5 b2a1 c5f8 a1c3 f8h6 c8e6 a3a4 a7a6 a2b3 b7b5 a4b5 a6b5 h6d2 d8a5 d2c3 a5c3 f1e1 g8g7 d1d2 c3d2 f3d2 c6d4 c2c3 d4b3');
INSERT INTO evaluations VALUES(19291,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(19291,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(19291,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(19291,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19291,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19291,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19291,36,'cp -7','f1c4 f8g7 c4a2 d7d6 d2d3 e7e6 c1e3 g8f6 g1f3 e8g8 e1g1 b7b6 h2h3 h7h6 d3d4 c8b7 d1d2 g8h7 a1d1 c6e7 d4c5 d6c5 e3f4 d8d2 f3d2 a8d8 f4h2 g6g5 f1e1 e7g6 a2c4 f6h5 c4f1 h5f4 g2g3 f4h5 f1d3 h5f6 d2c4 b7a6 d3f1 f6d7 f2f4 g7d4 g1g2');
INSERT INTO evaluations VALUES(19238,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(19238,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19238,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(19238,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(19238,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19238,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(19238,36,'cp 0','f8g7 c4a2 e7e6 d2d3 d7d6 c1e3 g8f6 g1f3 d8b6 a1b1 e8g8 e1g1 b6a5 c3e2 b7b5 c2c3 c8b7 b2b4 c5b4 a3b4 a5a2 b1a1 a2b2 a1b1 b2a3 b1a1');
INSERT INTO evaluations VALUES(5302,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(5302,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(5302,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(5302,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(5302,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(5302,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(5302,36,'cp -15','c4a2 e7e6 d2d3 d7d6 c1e3 g8f6 g1f3 e8g8 e1g1 b7b6 h2h3 h7h6 d1d2 g8h7 d3d4 c6e7 d4c5 d6c5 d2e2 d8c7 e4e5 f6d5 c3b5 c8a6 a2c4 a6b5 c4b5 d5e3 e2e3 a7a6 b5d3 b6b5 c2c3 a8d8 d3c2 e7f5');
INSERT INTO evaluations VALUES(21101,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(21101,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(21101,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(21101,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(21101,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(21101,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(21101,36,'cp 78','g8f6 d2d4 c8g4 d4d5 c6a5 c4d3 f6d5 d3b5 c7c6 d1d5 c6b5 d5b5 g4d7 b5e2 e7e6 e1g1 f8e7 f1d1 f7f6 f3d4 d8c8 d4e6 d7e6 e2h5 e6f7 h5a5 e8g8 c1e3 f8d8 d1d8 e7d8 a5a4 d8c7 a1d1 c7e5 e3d4 e5b8 h2h3 c8c7 d4e3 c7h2 g1f1 b8e5 a4g4 g8h8 d1d7 e5c3');
INSERT INTO evaluations VALUES(14360,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(14360,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(14360,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(14360,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(14360,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(14360,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(14360,36,'cp 68','d2d4 c8g4 e1g1 e7e6 h2h3 g4h5 f1e1 f8b4 d4d5 c6a5 c4b5 c7c6 d5c6 a5c6 d1d8 e8d8 b5c6 b7c6 f3e5 d8c7 a2a3 b4d6 c1d2 h5g6 e5g6 h7g6 c3e4 f6e4 e1e4 h8h5 c2c4 e6e5 d2a5 c7d7 a1d1 d7e6 e4e2 a8c8 a5c3 c8c7 f2f4 c7d7 d1e1 d6c5 g1f1');
INSERT INTO evaluations VALUES(14365,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(14365,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(14365,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(14365,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(14365,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(14365,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14365,36,'cp 69','c8g4 e1g1 e7e6 h2h3 g4h5 f1e1 f8b4 d4d5 c6a5 c4b5 c7c6 d5c6 a5c6 d1d8 e8d8 b5c6 b7c6 f3e5 d8c7 a2a3 b4d6 c1d2 h5g6 e5g6 h7g6 c3e4 f6e4 e1e4 h8h5 d2c3 a8g8 a1e1 c7d7 g1f1 h5d5 h3h4 d6c7 g2g3 a7a5 f1g2 c7b6 e1e2 b6c7');
INSERT INTO evaluations VALUES(24987,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(24987,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(24987,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(24987,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(24987,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(24987,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(24987,36,'cp 59','e1g1 e7e6 h2h3 g4h5 f1e1 f8b4 d4d5 c6a5 c4b5 c7c6 d5c6 a5c6 d1d8 e8d8 b5c6 b7c6 f3e5 d8c7 a2a3 b4d6 e5c4 h5g6 e1d1 f6d5 c3d5 c6d5 c4d6 c7d6 c2c4 h8d8 c4d5 e6d5 c1f4 d6e6 f4e3 f7f6 a1c1 a8c8 c1c8 d8c8 e3a7 c8c4');
INSERT INTO evaluations VALUES(21123,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(21123,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(21123,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(21123,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(21123,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(21123,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(21123,36,'cp -82','d8d4 d1c2 d4d6 f1c4 c6d4 c2d3 e7e5 g1e2 d4e6 d3g3 c7c6 c1e3 g8f6 a1d1 d6c7 h2h3 f8e7 c4e6 c8e6 e2d4 e8g8 d4e6 f7e6 e1g1 g8h8 a2a3 a7a5 a3a4 a8d8 d1d8 f8d8 g3f3 e7b4 e3g5 c7e7 f1d1 b4c5 g5f6 d8d1 c3d1 e7f6 f3h5 g7g6 h5e2 c5d4');
INSERT INTO evaluations VALUES(1628,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(1628,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(1628,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(1628,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(1628,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(1628,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(1628,36,'cp -76','d1c2 d4d6 f1c4 c6d4 c2d3 e7e5 c1e3 c8e6 e1c1 e8c8 f2f4 e6c4 d3c4 d6c6 c4c6 d4c6 d1d8 c8d8 g1f3 e5f4 e3f4 d8c8 f3g5 g8h6 h1f1 f8e7 c3d5 e7d8 f4d2 c6e5 g5f3 e5d7 d2h6 g7h6 f3d2 h8g8 g2g3 d7e5 c1c2 a7a6 d2f3 e5d7 e4e5 g8e8 d5f4 d7e5 f1e1 f7f6');
INSERT INTO evaluations VALUES(21675,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(21675,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(21675,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(21675,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(21675,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(21675,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(21675,36,'cp 145','d4e5 g7g5 f4g3 f8g7 f1b5 g8e7 d1h5 h7h6 b1c3 c8e6 e1c1 d8d7 h5e2 a7a6 b5c6 d7c6 g1f3 c6b6 f3d4 h6h5 h2h4 g5g4 e3e4 e8c8 c1b1 d5e4 e2e4 e7d5 c3e2 g7f8 g3f4 d5f4 e2f4 f8a3 d4b3 a3e7 f4e6 d8d1 h1d1 f7e6 f2f3 e7h4 f3g4 b6c6');
INSERT INTO evaluations VALUES(21643,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(21643,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(21643,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(21643,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(21643,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(21643,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(21643,36,'cp 133','g7g5 f4g3 f8g7 f1b5 g8e7 d1h5 h7h6 b1c3 c8e6 e1c1 d8d7 h5e2 a7a6 b5c6 d7c6 g1f3 c6b6 f3d4 h6h5 h2h4 g5g4 c1b1 e8c8 f2f4 g4f3 g2f3 e7c6 d4c6 b6c6 e2h2 g7f8 g3f4 f8c5 f4g5 d8d7 c3e2 c6b6 e2d4 c8b8 h2f4 b8a8 d4b3 c5f8 b1a1 h8g8');
INSERT INTO evaluations VALUES(21708,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(21708,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(21708,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(21708,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(21708,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(21708,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(21708,36,'cp 137','c3d5 f7f5 d5c3 g8f6 f1b5 f8d6 d1e2 e8g8 b5c6 b7c6 d2d3 h7h6 e4f5 c8f5 f3d2 f6d5 d2e4 a7a5 e1g1 a5a4 f2f3 a4a3 b2b3 d6b4 c3a4 d8e7 g1h1 a8d8 e2f2 f5g6 c1d2 g6e4 d2b4 e7b4 d3e4 d5f4 a1d1 g8h8 a4c5 d8d4 d1d4 e5d4');
INSERT INTO evaluations VALUES(21557,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(21557,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(21557,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(21557,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(21557,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(21557,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(21557,36,'cp 139','f7f5 d5c3 g8f6 f1b5 f8d6 b5c6 b7c6 d1e2 e8g8 d2d3 d8e8 f3d2 d6c5 e4f5 c8f5 e1g1 f6d5 d2e4 c5d4 c3a4 e8g6 c2c3 d4b6 g1h1 d5f4 c1f4 e5f4 f2f3 a8e8 a4b6 c7b6 e2d2 c6c5 a1e1 f5e6 a2a3 h7h6 d3d4 c5d4 d2d4');
INSERT INTO evaluations VALUES(42231,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42231,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42231,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42231,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42231,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(42231,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42231,36,'cp 0','d1a4 b8c6 f1d3 e8g8 g1e2 f8e8 c1d2 b4f8 e1g1 c6b4 d3b1 b4a6 b1d3');
INSERT INTO evaluations VALUES(42462,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(42462,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(42462,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(42462,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(42462,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(42462,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(42462,36,'cp 52','d7d6 d1f3 b8c6 f1b5 e6e5 g1e2 e8g8 b5c6 b7c6 d4e5 d6e5 h2h3 c8e6 e2g3 h7h6 a2a3 b4c5 c3e2 d8e7 d2c3 f6d7 g3f5 e6f5 e4f5 e5e4 f3g3 d7f6 e1c1 f8d8 c1b1 e4e3 d1d8 a8d8 c3f6 e7f6 f2e3 f6f5');
INSERT INTO evaluations VALUES(45544,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(45544,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45544,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(45544,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(45544,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(45544,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(45544,36,'cp 9','d5c4 e2e4 g7g5 h4g3 b7b5 f3e5 c8b7 h2h4 g5g4 f1e2 b8d7 e5d7 d8d7 g3e5 d7e7 e1g1 f8g7 d1d2 e8g8 d2f4 a8d8 a1d1 g8h7 f4e3 f6h5 e2g4 e7h4 g4h5 h4h5 e5g7 h7g7 c3e2 f8g8 e2g3 h5g6 b2b3 g7h7 b3c4 b5c4 d1b1');
INSERT INTO evaluations VALUES(45548,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(45548,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(45548,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45548,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45548,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45548,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45548,36,'cp 0','e2e4 g7g5 h4g3 b7b5 f1e2 c8b7 e1g1 b8d7 f3e5 f8g7 e5d7 f6d7 g3d6 g7f8 d6g3');
INSERT INTO evaluations VALUES(45549,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45549,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45549,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(45549,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45549,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45549,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45549,36,'cp 0','g7g5 h4g3 b7b5 f3e5 c8b7 h2h4 g5g4 e5g4 f6g4 d1g4 d8d4 g4f4 f8b4 a1d1 b4c3 e1e2 d4f6 f4c7 e8g8 c7b7 b8a6 b2c3 f6c3 g3d6 c3c2 e2e1 c2c3 e1e2');
INSERT INTO evaluations VALUES(45536,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45536,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45536,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45536,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45536,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(45536,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45536,36,'cp 29','h4g3 b7b5 f3e5 c8b7 h2h4 g5g4 f1e2 b8d7 e5d7 d8d7 g3e5 d7e7 b2b3 f8g7 b3c4 e8g8 c4c5 b5b4 e5d6 e7d8 e4e5 b4c3 e5f6 d8f6 d1b3 g4g3 e1g1 a8b8 b3c3 g3f2 f1f2 f6d4 c3d4 g7d4 a1b1 b7a6 d6b8 a6e2 b8d6 f8a8 b1b4 d4e3 b4e4 e3f2 g1f2 e2h5 g2g4 h5g6');
INSERT INTO evaluations VALUES(45537,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45537,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(45537,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45537,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45537,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45537,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45537,36,'cp 0','b7b5 f3e5 c8b7 h2h4 g5g4 e5g4 f6g4 d1g4 d8d4 g4f4 f8b4 a1d1 b4c3 b2c3 d4c3 e1e2 c3b2 d1d2 c4c3 d2b2 c3b2 f4e5 b2b1q e5h8 e8e7 g3b8 b1e4 e2d2 e4b4 d2c2 b4a4 c2c1 a4a3 c1d1 a3a4 d1d2');
INSERT INTO evaluations VALUES(43953,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43953,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43953,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43953,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43953,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43953,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43953,36,'cp 25','f3e5 c8b7 h2h4 g5g4 f1e2 b8d7 e5d7 d8d7 g3e5 f8g7 e2g4 h8g8 g4f3 c6c5 d4c5 b5b4 e5f6 g7f6 d1d7 e8d7 e1c1 d7e7 c3d5 e6d5 e4d5 e7d7 c5c6 d7d6 c6b7 a8b8 h1e1 b8b7 e1e6 f7e6 d5e6 d6e6 f3b7 g8d8 d1e1 e6d6 e1e4 d6c7 b7a6 d8d4 e4e6 d4f4 f2f3 f6h4 e6h6 h4f6 h6h5 c7b6');
INSERT INTO evaluations VALUES(6424,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6424,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6424,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6424,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(6424,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(6424,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6424,36,'cp 7','g8f6 g1f3 h7h6 e1g1 e8g8 h2h3 a7a6 c1e3 d7d6 e3c5 d6c5 a2a4 b7b6 c3d5 c8e6 d5e3 e6c4 e3c4 d8e7 f1e1 a8d8 c2c3 e7e6 g1h2 f6e8 d1e2 g8h7 c4e3 e6d7 a1d1 e8d6 f3d2 f7f6 g2g3');
INSERT INTO evaluations VALUES(6761,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(6761,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(6761,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(6761,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(6761,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(6761,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(6761,36,'cp 107','c6e5 b1c3 g8f6 c3b5 e5f3 g2f3 d6e7 e4e5 f6h5 f1e2 a7a6 b5c3 d7d6 f3f4 g7g6 e2h5 g6h5 d1h5 d8d7 e5d6 e7d6 h1g1 d7e6 c1e3 c8d7 e1c1 e8c8 g1g7 d7e8 g7h7 h8h7 h5h7 e8d7 h2h4 e6g6 h7g6 f7g6 c3e4 d7f5 e4d6 d8d6 d1g1 d6d8 b2b3 d8e8 c1d2 e8h8 g1h1 b7b6 d2c3 h8h5');
INSERT INTO evaluations VALUES(6800,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(6800,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(6800,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(6800,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(6800,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(6800,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(6800,36,'cp 105','b1c3 g8f6 c3b5 e5f3 g2f3 d6e7 h1g1 e8g8 c1h6 f6e8 d1d2 d7d6 e1c1 g8h8 h6f4 a7a6 b5c3 c7c6 h2h4 d8a5 a2a3 c8e6 f4g5 f7f6 c3e2 a5d2 g5d2 e6f7 f1h3 d6d5 e2f4 e8c7 h4h5 e7d6 h3f5 f8e8 g1g4 c7e6 f4e6 f7e6 h5h6 g7g6 f5e6');
INSERT INTO evaluations VALUES(15086,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(15086,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(15086,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(15086,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(15086,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(15086,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(15086,36,'cp 26','e4d5 f6d5 e1g1 f8e7 f1e1 c8g4 d2d4 e5d4 h2h3 g4h5 c3d4 e8g8 b1c3 d5b6 c4b3 e7f6 g2g4 h5g6 a2a4 c6d4 f3d4 f6d4 a4a5 d4c3 b2c3 b6d7 d1d4 c7c6 c1f4 d7f6 d4d8 f8d8 e1e7 b7b5 f4e3 a7a6 a1c1 f6e4 e7c7 e4d2 e3d2 d8d2 c7c6');
INSERT INTO evaluations VALUES(14985,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(14985,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14985,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(14985,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(14985,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(14985,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(14985,36,'cp 11','f6d5 e1g1 f8e7 d2d4 e5d4 f3d4 e8g8 d4c6 b7c6 b1d2 f8e8 f1e1 a7a5 d1f3 a5a4 h2h3 d8d7 c4d3 e7f8 e1e8 d7e8 d2f1 e8e1 a1b1 e1e6 c1d2 e6f6 f3f6 d5f6 c3c4 f6d7 d3e4 a8a6 d2c3 d7c5 e4c2 c5e6 b1d1 a4a3');
INSERT INTO evaluations VALUES(15311,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(15311,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(15311,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15311,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(15311,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(15311,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(15311,36,'cp 20','e1g1 f8e7 f1e1 c8g4 d2d4 e5d4 h2h3 g4e6 c3d4 e8g8 c4b3 c6a5 b3c2 d5b4 b1c3 b4c2 d1c2 f8e8 c1f4 a5c6 c2d2 c6b4 f4g5 c7c6 e1e5 e6d5 g5e7 e8e7 e5e7 d8e7 f3e5 d5e6 a1e1 a8d8 c3e4 b4a6 e5g4 e6g4');
INSERT INTO evaluations VALUES(15310,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(15310,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(15310,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(15310,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(15310,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(15310,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(15310,36,'cp 22','c8g4 h2h3 g4h5 d2d4 e5d4 c3d4 d5b6 c4b3 f8e7 f1e1 e8g8 b1c3 e7f6 g2g4 h5g6 a2a4 c6d4 f3d4 f6d4 a4a5 d4c3 b2c3 b6d7 a5a6 b7a6 b3d5 a8b8 d5c6 d7f6 d1a4 d8d3 e1e3 d3c2 a4c2 g6c2 g4g5 f6e8 a1a6 b8b1 e3e1 e8d6 a6a7 c2f5 g1h2 b1b3 a7c7 b3c3 c1f4 c3h3 h2g2');
INSERT INTO evaluations VALUES(15948,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(15948,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(15948,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15948,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(15948,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(15948,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(15948,36,'cp 1','a7a6 d1c2 d5c4 a2a4 c7c5 f1c4 c5d4 e3d4 f8e7 c4b3 b7b6 e1g1 c8b7 f1d1 d8c7 c1g5 e8g8 a1c1 h7h6 g5f6 e7f6 d4d5 e6d5 c3d5 c7c2 d5f6 d7f6 c1c2 f8d8 c2d2 d8d2 f3d2 g8f8 d2f1 a8c8 d1d6 b7c6 f1e3 f6e8 d6d1 e8f6 e3f5 b6b5');
INSERT INTO evaluations VALUES(18507,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(18507,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(18507,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(18507,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(18507,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18507,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18507,36,'cp 28','f1c4 g8f6 d1e2 d8c7 c1g5 f8e7 e1c1 e8g8 d1d2 a7a5 h1d1 a8b8 g5h4 f8d8 c1b1 c8e6 c4b3 h7h6 f2f3 b8b4 b3e6 f7e6 a2a3 b4b8 c3a4 d6d5 b1a1 d8f8 a1a2 f6h7 h4e7 c7e7 b2b3 h7f6 e2a6 e7c7');
INSERT INTO evaluations VALUES(18504,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(18504,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(18504,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(18504,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(18504,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(18504,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(18504,36,'cp 18','g8f6 d1e2 d8c7 c1g5 f8e7 e1c1 e8g8 d1d2 c8e6 h1d1 f8d8 c4b3 h7h6 g5h4 a8b8 f2f3 b8b4 c1b1 c7b8 b1a1 b8c7 e2f1 c7b8 f1f2 a7a5 f2e2 g8h8 b3e6 f7e6 b2b3 d6d5 c3a4 b4b7 c2c4 e7b4 d2c2 d5e4 h4f6 g7f6 f3e4 d8d4 a1b1 b7f7 c4c5 h8h7 d1d4 e5d4');
INSERT INTO evaluations VALUES(12089,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(12089,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(12089,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12089,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12089,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(12089,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(12089,36,'cp 11','d1d3 f8e7 c1g5 h7h6 g5f6 e7f6 b2b4 e8g8 e1g1 d8b6 a1b1 b6d4 b4b5 c8e6 c4b3 a8c8 f1d1 e6g4 d1d2 d4c5 c3a4 c5d4 h2h3 d4d3 d2d3 g4e2 d3d2 c6b5 d2e2 b5a4 b3a4 c8c7 a4b3 f6e7 g2g3 g7g6 b3d5 f8c8 g1g2 c7c2');
INSERT INTO evaluations VALUES(37570,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(37570,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(37570,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(37570,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(37570,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(37570,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(37570,36,'cp 99','g1f3 g8f6 e2e4 d7d6 h2h3 e8g8 c1g5 b8d7 d1c2 e7e5 e1c1 f8e8 d4d5 d7c5 f3e1 c8d7 f2f3 h7h6 g5h4 a8b8 b2b4 c5b7 e1d3 a7a6 c1b1 c7c6 d5c6 d7c6 d3b2 a6a5 b4b5 c6d7 f1e2 d7e6 c3d5 e6d5 c4d5 b7c5');
INSERT INTO evaluations VALUES(37569,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(37569,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(37569,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(37569,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(37569,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(37569,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(37569,36,'cp 74','g8f6 e2e4 d7d6 c1g5 h7h6 g5e3 f6g4 e3c1 e7e5 h2h3 e5d4 f3d4 g4f6 c1e3 c8b7 f1d3 b8c6 d4c6 b7c6 e1g1 f6d7 d1d2 a7a5 a1e1 d7c5 d3c2 d8f6 f2f3 h6h5 e1d1 e8g8 g1h1 a5a4 e3g5 f6e6 c3d5 c6d5 c4d5 e6e5 f3f4 e5b2');
INSERT INTO evaluations VALUES(39844,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(39844,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(39844,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(39844,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(39844,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(39844,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(39844,36,'cp 84','e2e4 d7d6 c1g5 h7h6 g5e3 f6g4 e3d2 e7e5 d4d5 g4f6 h2h3 b8d7 f1d3 d7c5 d3c2 f6h5 g2g3 e8g8 d2e3 f7f5 e4f5 c8f5 f3h4 f5c2 d1c2 g6g5 h4f5 e5e4 f5g7 h5g7 e1c1 g7f5 c3e4 c5e4 c2e4 d8f6 e4d3 f5e3 f2e3 a8e8 h1e1');
INSERT INTO evaluations VALUES(39845,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(39845,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(39845,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(39845,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(39845,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(39845,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(39845,36,'cp 77','d7d6 h2h3 e8g8 c1g5 b8a6 d1d2 e7e5 e1c1 d8e8 f1e2 c8b7 d4d5 c7c6 c1b1 c6d5 c4d5 a6c5 d2e3 a7a6 f3d2 b6b5 b2b4 c5d7 d2b3 a8c8 d1c1 h7h6 b3a5 c8c3 c1c3 b7d5 g5f6 d7f6 f2f3 d5e6 c3c6 d6d5 h1c1 e8d7 c1d1');
INSERT INTO evaluations VALUES(39812,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(39812,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(39812,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(39812,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(39812,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(39812,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(39812,36,'cp 88','f1e2 e8g8 c1e3 f6g4 e3d2 b8d7 h2h3 g4f6 e1g1 e7e5 d4e5 d7e5 f3e5 d6e5 d2e3 c8e6 d1c2 c7c6 b2b4 d8c7 c4c5 f8c8 c5b6 a7b6 a2a4 g7f8 c2b2 c6c5 f1c1 f6d7 e2b5 c5b4 c3d5 c7d6 c1c8 a8c8');
INSERT INTO evaluations VALUES(40233,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(40233,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(40233,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(40233,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(40233,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(40233,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(40233,36,'cp 31','e2e3 e8g8 f1d3 c8d7 e1g1 d7c6 g5f4 b8d7 h2h3 d8b6 d1c2 f8c8 f1c1 h7h6 c2e2 b6d8 c1c2 f6e4 a1c1 e7b4 d3e4 d5e4 f3d2 f7f5 a2a3 b4e7 d2c4 d7f6 c4e5 g8h7 e5c6 c8c6 f4e5 f6d7 c3b5 d7e5 d4e5 d8d3 e2d3');
INSERT INTO evaluations VALUES(40232,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(40232,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(40232,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40232,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(40232,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(40232,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(40232,36,'cp 28','c8d7 f1d3 e8g8 f3e5 b8c6 e1g1 f6e8 g5f4 c6b4 d3e2 e8d6 e5d7 d8d7 c3a4 d6e4 f2f3 e4d2 d1d2 d7a4 a2a3 b4c6 f1c1 a8c8 b2b3 a4a5 b3b4 a5b6 a1b1 f8d8 c1c3 e6e5 f4e5 c6e5 c3c8 d8c8');
INSERT INTO evaluations VALUES(46891,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(46891,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46891,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(46891,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(46891,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46891,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(46891,36,'cp 34','e7e6 c2c4 f8d6 f4d6 c7d6 b1c3 d5c4 f1c4 e8g8 e1g1 a7a6 a2a4 b8c6 d4d5 e6d5 c3d5 c8e6 d5f6 d8f6 c4e6 f6e6 h2h3 f8c8 a1c1 c6e5 f3d4 e6d7 b2b3 c8c5 c1c2 b7b5 a4b5 a6b5 f2f4 e5g6 d1d2 a8c8 f1d1 d7b7');
INSERT INTO evaluations VALUES(48948,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(48948,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(48948,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(48948,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(48948,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(48948,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(48948,36,'cp 41','d1b3 d8e7 g1e2 e7b4 b1c3 b4b3 c4b3 b8a6 e1g1 f8e7 f1e1 e8g8 e2g3 f8e8 c1g5 g8f8 g5f4 c7c6 a2a3 e8d8 a1d1 h7h6 f4e5 e7d6 c3e4 f6e4 g3e4 d6c7 b3c4 b7b5 c4b3 c7e5 d4e5 c8f5 e4d6 a6c5 b3f7 c5d3 d6f5 d3e1');
INSERT INTO evaluations VALUES(54992,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54992,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(54992,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(54992,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54992,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(54992,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54992,36,'cp 3','e7e5 c2c4 c7c5 b1c3 b8c6 g2g3 g7g6 f1g2 g8e7 d2d3 f8g7 d1d2 e8g8 e2e3 c8e6 c3d5 d8d7 h2h4 a8e8 h4h5 h7h6 g1e2 g6g5 e2c3 f7f6 e1c1 a7a6 f2f4 b7b5 d5b6 d7d8 c3d5 e7d5 c4d5 d8b6 d5e6 f6f5 c1b1');
INSERT INTO evaluations VALUES(10264,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10264,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(10264,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10264,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(10264,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10264,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(10264,36,'cp 0','b1c3 e8g8 c3a4 c5e7 c2c4 d7d6 a4c3 c6d4 f3d4 e5d4 c3e2 c7c5 b2b3 b7b5 e2g3 a8b8 f1e2 b5c4 b3c4 d8b6 e1g1 b6b3 d1e1 b3c2 c1f4 f8e8 a1c1 c2a2 c1a1');
INSERT INTO evaluations VALUES(23180,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(23180,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(23180,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23180,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23180,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(23180,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(23180,36,'cp 0','d7c6 b1c3 g8f6 e4f5 e5e4 f3g5 c8f5 e1g1 f8e7 d1e2 d8d4 f1e1 e8g8 d2d3 e4d3 e2e7 d3d2 c1d2 d4d2 g5f3 d2d6 a1d1 d6e7 e1e7 f5c2 d1d2 c2f5 e7c7 f8f7 c7f7 g8f7 f3e5 f7e7 e5c4 f5e6 c4a5 a8b8 f2f3 e6d5 g1f2 e7d6 c3e4 d6e5 e4c3');
INSERT INTO evaluations VALUES(22339,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(22339,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(22339,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(22339,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(22339,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(22339,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(22339,36,'cp 6','b1c3 g8f6 e4f5 e5e4 f3g5 c8f5 e1g1 f8e7 d1e2 d8d4 f1e1 e8g8 d2d3 e4d3 e2e7 d3d2 c1d2 d4d2 g5f3 d2d6 a1d1 d6e7 e1e7 f5c2 d1d2 c2f5 e7c7 f8f7 c7f7 g8f7 f3e5 f7e6 e5c4 f6d5 c4a5 a8b8 f2f3 e6d6 a5c4 d6c5 c3d5 c6d5 c4e3 f5e6 e3c2 g7g5 g1f2 c5d6 c2d4 e6d7 d4e2 d7c6 a2a3 a7a5 g2g3 b8c8');
INSERT INTO evaluations VALUES(37677,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37677,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37677,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37677,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37677,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37677,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(37677,36,'cp 0','b8c6 g1f3 a7a6 f1d3 d7d6 e1g1 g8e7 c3a4 c5a7 c2c4 e8g8 a4c3 c6e5 d1e2 d8c7 f1d1 e5g4 d1f1');
INSERT INTO evaluations VALUES(5183,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5183,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5183,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(5183,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(5183,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(5183,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5183,36,'cp 0','f1d3 a7a6 d1g4 g8f6 g4g3 d7d6 g1f3 f6h5 g3h3 h5f6');
INSERT INTO evaluations VALUES(55989,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(55989,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(55989,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(55989,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(55989,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(55989,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(55989,36,'cp 17','g2g3 b8c6 f1g2 d4d3 e1g1 g8e7 c2d3 e7f5 b1c3 a8b8 b2b3 b7b6 c1b2 c8b7 c3e2 f8c5 d3d4 c6d4 f3d4 c5d4 e2d4 f5d4 g2b7 b8b7 b2d4 d8d4 d1f3 d4d5 f3d5 e6d5 a1c1 e8d7 d2d4 h8c8 f1d1 c7c6 f2f4 f7f6 g1f2 f6e5 d4e5 d7e6 g3g4 c6c5 h2h4 d5d4 f2f3');
INSERT INTO evaluations VALUES(13785,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(13785,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(13785,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(13785,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(13785,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(13785,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(13785,36,'cp -18','e2e4 d5d4 a2a4 c8g4 h2h3 g4h5 g3g4 h5g6 f3e5 c6e5 f2f4 f6d7 f4e5 d7e5 b1d2 f8d6 d2f3 f7f6 f3e5 d6e5 d1f3 d8c7 c1f4 e8g8 f3g3 a8e8 b2b3 e5f4 f1f4 g6f7 g3h2 e8e5 h3h4 h7h6 f4f2 c7d7 g2f3 d7e7 a4a5 e7c7 f3e2');
INSERT INTO evaluations VALUES(20868,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(20868,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(20868,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(20868,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(20868,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(20868,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(20868,36,'cp 26','f1b5 c8g4 d2d4 f6e5 d4e5 e7e6 b1d2 a7a6 b5c6 b7c6 h2h3 g4h5 d2f1 g8e7 f1g3 h5f3 d1f3 e7g6 e1g1 d8d7 f3e2 f8e7 f2f4 e8g8 c2c4 a6a5 c1d2 e7h4 g3h5 d5d4 a1d1 c6c5 b2b3 a5a4 b3b4 c5b4 d2b4 h4e7 b4e7 d7e7');
INSERT INTO evaluations VALUES(22692,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(22692,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(22692,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(22692,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(22692,34,'cp -140',NULL);
INSERT INTO evaluations VALUES(22692,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(22692,36,'cp -104','g8f6 h5d1 d7d5 e4d5 f6d5 g1f3 c8g4 f1b5 d5c3 b2c3 d8d5 d1e2 e8c8 b5c4 d5d7 e1g1 f8d6 a1b1 c6a5 e2e4 f7f6 f1e1 g4f5 e4d5 c7c5 c4b5 d7e7 f3h4 f5c2 b1b2 c2g6 d5f3 c8b8 d2d3 d6c7 a2a4 e7f7 h4g6 h7g6 c1e3 f6f5');
INSERT INTO evaluations VALUES(16375,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(16375,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(16375,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(16375,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(16375,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(16375,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(16375,36,'cp -103','h5d1 d7d5 e4d5 f6d5 g1f3 e5e4 c3e4 d8e7 d1e2 c8g4 e4g3 d5b4 e1d1 c6d4 e2e7 f8e7 c2c3 d4f3 h2h3 g4e6 g2f3 b4a2 d1c2 a2c1 c2c1 a7a5 h3h4 g7g6 f1d3 h7h5 c1c2 e8f8 f3f4 a5a4 f4f5 e6b3 c2c1 f8g7 d3e4 c7c6 f5g6 f7g6 g3e2 h8d8 e2d4 b3d5 e4d5 d8d5 f2f4 g7f7 c1c2');
INSERT INTO evaluations VALUES(16518,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(16518,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(16518,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(16518,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(16518,34,'cp -110',NULL);
INSERT INTO evaluations VALUES(16518,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(16518,36,'cp -107','d7d5 e4d5 f6d5 g1f3 d5c3 b2c3 f8d6 d2d3 e8g8 f1e2 f7f5 e1g1 h7h6 f1e1 d8f6 c1b2 c8d7 f3d2 a8d8 c3c4 b7b6 a2a4 d8e8 e2f3 e8e7 f3d5 g8h8 b2a3 d6a3 a1a3 f8d8 a3a1 c6d4 d2f3 c7c6 f3d4 c6d5');
INSERT INTO evaluations VALUES(46703,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(46703,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(46703,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(46703,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(46703,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(46703,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(46703,36,'cp 48','f8g7 c1e3 e8g8 d1d2 e7e5 d4e5 d6e5 e1c1 d8d2 d1d2 b7b6 f1c4 b8c6 g1f3 c8b7 h1e1 h7h6 g2g4 a8d8 a2a4 c6a5 c4d3 f6e8 d2d1 e8d6 f3d2 f7f5 g4f5 g6f5 f2f3 f8f7 c3b5 f5f4 b5d6 c7d6');
INSERT INTO evaluations VALUES(50928,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(50928,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(50928,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(50928,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(50928,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(50928,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(50928,36,'cp 74','c2c4 f8e7 b1c3 e7f6 c1e3 b8c6 d1d2 f6d4 e3d4 c6d4 d2d4 g8f6 f1e2 e8g8 e1g1 f8e8 a1d1 d8e7 f1e1 e7e5 f2f3 c8d7 a2a3 a8d8 d1d2 e5d4 d2d4 g8f8 g1f2 a6a5 g2g4 d7c6 e1c1 f6d7 g4g5 h7h6 g5h6 g7h6');
INSERT INTO evaluations VALUES(54253,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(54253,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54253,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54253,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(54253,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54253,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54253,36,'cp 0','b8c6 g2g3 g8f6 f1g2 f8e7 e1g1 a7a6 c2c3 e8g8 f1e1 b7b5 d3d4 c5d4 c3d4 f6e4 d2e4 d5e4 f3e5 c8b7 g2e4 a8c8 d1h5 g7g6 e5g6 h7g6 e4g6 f7g6 h5g6 g8h8 g6h5 h8g8');
INSERT INTO evaluations VALUES(20297,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20297,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20297,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20297,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20297,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20297,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20297,36,'cp 0','g2g3 g8f6 f1g2 f8e7 e1g1 e8g8 f1e1 d8c7 c2c3 f8d8 d1e2 h7h6 e4e5 f6d7 a2a4 b7b6 d2f1 d5d4 c3c4 c8b7 c1d2 c6b4 a1c1 b4a2 c1a1');
INSERT INTO evaluations VALUES(20299,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20299,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20299,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20299,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20299,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20299,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20299,36,'cp 0','g8f6 f1g2 f8e7 e1g1 e8g8 f1e1 a7a6 c2c3 b7b5 d3d4 c5d4 f3d4 c6d4 c3d4 d5e4 d2e4 f6e4 g2e4 a8a7 d1h5 h7h6 c1h6 g7h6 h5h6 f7f5 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(13607,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13607,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13607,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13607,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13607,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13607,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13607,36,'cp 0','f1g2 f8e7 e1g1 e8g8 f1e1 a7a6 c2c3 b7b5 d3d4 c5d4 f3d4 c6d4 c3d4 d5e4 d2e4 f6e4 g2e4 a8a7 d1h5 h7h6 c1h6 g7h6 h5h6 f7f5 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(13608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13608,36,'cp 0','f8e7 e1g1 e8g8 f1e1 a7a6 c2c3 b7b5 d3d4 c5d4 f3d4 c6d4 c3d4 f6e4 d2e4 d5e4 g2e4 a8a7 d1h5 h7h6 c1h6 g7h6 h5h6 f7f5 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(8094,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8094,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8094,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8094,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8094,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8094,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8094,36,'cp 0','e1g1 e8g8 f1e1 b7b5 e4d5 e6d5 d2b3 c8f5 c1e3 f6d7 f3d2 f5e6 d2f3');
INSERT INTO evaluations VALUES(8093,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8093,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8093,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8093,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8093,36,'cp -8','e8g8 f1e1 d8c7 e4d5 e6d5 d2f1 h7h6 c1f4 c7d8 c2c3 c8e6 d3d4 d8b6 e1e2 f8e8 d4c5 e7c5 f4e3 a8d8 e3c5 b6c5 f3d4 c6d4 d1d4 c5d4 c3d4 d8c8 f1e3 a7a5 e2c2 c8c2 e3c2 e8c8 c2e3 a5a4 a2a3 f6e4');
INSERT INTO evaluations VALUES(3099,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3099,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(3099,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3099,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3099,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3099,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(3099,36,'cp -7','f1e1 d8c7 c2c3 h7h6 d1e2 f8d8 a2a3 b7b5 d3d4 c5d4 f3d4 c6d4 c3d4 d5e4 d2e4 c8b7 e2b5 f6e4 g2e4 b7e4 e1e4 c7c2 b5e2 c2e2 e4e2 d8d4 c1e3 d4d3 a1c1 e7f6 c1c7 a7a5 c7b7 a8c8 b7a7 d3d5 g1g2 c8c4 a7a8 g8h7 e3b6 a5a4');
INSERT INTO evaluations VALUES(3100,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3100,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3100,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3100,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3100,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3100,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3100,36,'cp 0','d8c7 c2c3 f8d8 d1e2 a8b8 e4d5 e6d5 d2f1 c8e6 c1f4 e7d6 f4g5 d6e7');
INSERT INTO evaluations VALUES(2232,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2232,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(2232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2232,36,'cp 0','c2c3 f8d8 d1e2 a8b8 e4d5 f6d5 a2a4 b7b6 h2h4 h7h6 d2c4 c8b7 f3h2 e7f8 e2g4 d5f6 g4e2');
INSERT INTO evaluations VALUES(16469,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16469,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16469,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(16469,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(16469,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(16469,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(16469,36,'cp -16','f8b4 g1e2 e5d4 e3d4 e8g8 a2a3 b4c3 e2c3 d7d5 c4c5 f8e8 f1e2 f6e4 c1e3 e4c3 b2c3 c8f5 e1g1 c6a5 f1e1 d8d7 e2f1 b7b6 h2h3 h7h6 a3a4 c7c6 c5b6 a7b6 e3f4 e8e1 d1e1 a8e8 e1d1 f5g6 f4g3 a5c4 f1c4 d5c4 f2f3 e8e3');
INSERT INTO evaluations VALUES(10422,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(10422,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(10422,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(10422,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(10422,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(10422,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(10422,36,'cp -10','g1e2 e5d4 e3d4 d7d5 a2a3 b4c3 e2c3 e8g8 c4d5 f6d5 f1e2 c6e7 e1g1 f8e8 f1e1 c7c6 c1g5 c8e6 e2d3 d8d7 d1f3 h7h6 g5d2 a8d8 c3d5 d7d5 f3d5 e7d5 a1c1 d8d7 f2f3 e8d8 c1c5 d5c7 d2c3 f7f6 g1f2 g8f7');
INSERT INTO evaluations VALUES(22593,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(22593,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(22593,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(22593,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(22593,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(22593,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(22593,36,'cp 47','f8b4 e2e3 g8f6 d1c2 e8g8 c1d2 a7a5 g1f3 b7b6 f1e2 c8a6 e1g1 b4c3 d2c3 c6b4 c2c1 b4d5 c3e1 d5e7 b2b3 d7d6 e1c3 a6b7 c1c2 e7g6 a1d1 d8e7 d4d5 e6d5 c3f6 e7f6 c4d5 a8c8 h2h3 c7c5 f1e1 g6e5 f3e5 f6e5 e2f3 f7f5');
INSERT INTO evaluations VALUES(6370,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(6370,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(6370,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(6370,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(6370,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6370,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(6370,36,'cp 49','e2e3 g8f6 d1c2 e8g8 c1d2 a7a5 g1f3 b7b6 f1e2 c8a6 e1g1 b4c3 d2c3 c6b4 c2b1 b4d5 c3d2 d5e7 b1c2 a6b7 a1d1 d7d6 b2b3 e7g6 f3e1 d8e7 f2f3 f6h5 a2a3 c7c5 d2c3 a8c8 d4c5 d6c5 e3e4 h5f4 g2g3 f4e2 c2e2 b7a6');
INSERT INTO evaluations VALUES(21499,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(21499,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(21499,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(21499,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(21499,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(21499,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(21499,36,'cp 96','f1c4 g8f6 b1c3 a7a6 a2a3 b7b5 c4d3 f8d6 d1c2 c8b7 c1d2 c6a5 e1g1 b7f3 g2f3 a5c4 d3c4 b5c4 c2a4 f6d7 c3e4 e8g8 a4c4 d8h4 e4d6 c7d6 d2b4 f8c8 c4e2 h4e7 f1c1 c8c1 a1c1 a6a5 b4d2 e7g5 g1f1 g5h5 e2c4 h5f3 c4c6 f3c6 c1c6 d6d5 f1e2 a8b8 b2b4 a5b4 d2b4 d7f6 f2f3');
INSERT INTO evaluations VALUES(44325,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(44325,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(44325,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(44325,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(44325,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(44325,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(44325,36,'cp 51','b8c6 d2d3 f8e7 e2e3 d7d6 g1e2 c8e6 c3d5 d8d7 e1g1 e8g8 d3d4 e5d4 e3d4 e6f5 d5e3 f5h3 d4c5 h3g2 g1g2 d6c5 d1d7 f6d7 e3d5 a8d8 c1e3 b7b6 a1d1 d7f6 e2c3 f6d5 c4d5 c6e5 e3f4 e7f6');
INSERT INTO evaluations VALUES(49707,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(49707,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(49707,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(49707,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(49707,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(49707,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(49707,36,'cp -37','e5e4 f3h4 h7h6 d2d3 d7d5 d3e4 d5e4 d1d8 e8d8 f2f4 f8c5 b1c3 h8e8 b2b3 c7c6 c1b2 d8c7 c3a4 c5b4 c2c3 b4d6 c3c4 b8a6 e1c1 b7b6 a4c3 a6c5 b3b4 c5b7 a2a3 c8g4 h2h3 g4d7 e2e3 a8d8 g3g4 a7a5 g4g5 h6g5 f4g5 a5b4 a3b4');
INSERT INTO evaluations VALUES(50418,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(50418,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(50418,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(50418,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(50418,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(50418,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(50418,36,'cp 143','e4c5 d3d4 c5a6 g1f3 d7d6 b1c3 c7c6 c1e3 c8f5 f3h4 f5e6 f2f4 g7g6 h4f3 a6b4 a2a3 b4d5 c3d5 c6d5 f3g5 e6c8 f1d3 h7h6 g5f3 c8f5 d3f5 g6f5 e1g1 e7e6 c2c4 d6e5 f3e5 d5c4 d1a4 b8c6 e5c6 b7c6 a4c6 e8e7 c6c5 e7e8 c5c4 d8d5 c4d5 e6d5');
INSERT INTO evaluations VALUES(50400,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(50400,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(50400,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(50400,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(50400,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(50400,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(50400,36,'cp 154','d3d4 c5a6 g1f3 d7d6 b1c3 c7c6 c1e3 c8f5 f3h4 f5e6 f2f4 g7g6 h4f3 a6b4 a2a3 b4d5 c3d5 c6d5 f3g5 e6c8 f1d3 h7h6 g5f3 c8f5 d3f5 g6f5 e1g1 e7e6 c2c4 d6e5 f3e5 b8c6 c4c5 h6h5 b2b4 a7a6 a1b1 h5h4 h2h3 f7f6 e5d3 e8f7 b4b5 a6b5 b1b5 a8a7 a3a4 f8e7 d1b3');
INSERT INTO evaluations VALUES(50399,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(50399,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(50399,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(50399,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(50399,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(50399,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(50399,36,'cp 132','c5a6 g1f3 d7d6 b1c3 c7c6 c1e3 c8f5 f3h4 f5e6 f2f4 g7g6 h4f3 a6b4 a2a3 b4d5 c3d5 c6d5 f3g5 e6c8 f1d3 h7h6 g5f3 c8f5 d3f5 g6f5 e1g1 e7e6 c2c4 d6e5 f3e5 d5c4 d1a4 b8c6 e5c6 b7c6 a4c6 e8e7 c6c4 f8g7 a1c1 d8b6 c4c7 e7f6 c7b6 a7b6 c1c6 f6g6 c6b6');
INSERT INTO evaluations VALUES(51886,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(51886,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(51886,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(51886,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(51886,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51886,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(51886,36,'cp 30','f7e6 e2e4 b8c6 f1b5 d8a5 b1c3 g8f6 g1h3 f8e7 e1g1 e8g8 a2a4 d7d6 h3f4 c6e5 b5e2 c8d7 g2g4 a5b4 f2f3 b4d4 d1d4 c5d4 c3b5 g7g5 f4d3 d7b5 a4b5 h7h6 f3f4 e5f7 f4f5 e6f5 f1f5 f8c8 e2d1 f6e4 d1f3 d6d5 f5d5 e4d6 d5c5 c8c5 d3c5 a8c8');
INSERT INTO evaluations VALUES(52220,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(52220,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(52220,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(52220,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(52220,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(52220,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(52220,36,'cp 23','e2e4 b8c6 f1b5 d8a5 b1c3 g8f6 c1d2 a5c7 g1f3 f8e7 d2g5 e8g8 b5c6 c7c6 d1e2 b7b5 e2b5 c6b5 c3b5 c8b7 e1c1 b7e4 f3e5 a7a6 b5c3 e4g2 h1g1 g2c6 g5h6 f6e8 e5c6 d7c6 c3e4 c5c4 h6e3 e8f6 d1d4 f6e4 d4e4 c4c3 b2c3 g8f7 g1g4 c6c5 e3g5 a8e8 g5e7 e8e7 e4f4 f7g8 f4f8 g8f8');
INSERT INTO evaluations VALUES(52219,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52219,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52219,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52219,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(52219,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52219,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(52219,36,'cp 26','b8c6 f1b5 d8a5 b1c3 g8f6 c1d2 a5c7 g1h3 f8e7 e1g1 e8g8 f1e1 d7d6 b5c4 c6e5 c4f1 a7a6 f2f4 e5g6 a2a4 g8h8 g2g3 c8d7 f1g2 a8e8 a4a5 d6d5 h3g5 c5c4 e4e5 f6g8 h2h4 g8h6 g2h3 h6f5 h3f5 f8f5 d2e3');
INSERT INTO evaluations VALUES(1467,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(1467,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(1467,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(1467,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(1467,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(1467,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(1467,36,'cp -25','b6b3 a2b3 c5d4 f3d4 c6e5 d4b5 e8d7 a1a7 a8a7 b5a7 f8c5 f1b5 d7c7 a7c8 c7c8 e1e2 g8f6 b3b4 c5b6 h1d1 h7h5 b1d2 f6e8 d2f3 e5f3 e2f3 e8d6 b5e2 d6e4 c1e3 b6e3 f3e3 c8c7 c3c4 e4f6 d1c1 d5c4 c1c4 c7b8 h2h3 h5h4 e3d2');
INSERT INTO evaluations VALUES(1535,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(1535,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(1535,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(1535,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(1535,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(1535,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(1535,36,'cp -16','a2b3 c5d4 f3d4 c6e5 d4b5 e8d7 a1a7 a8a7 b5a7 f8c5 f1b5 d7c7 a7c8 c7c8 e1e2 g8f6 f2f4 e5g6 b5d3 c5b6 b1a3 c8d7 b3b4 g6e7 c1d2 h8a8 h1a1 e7c8 a3c2 a8a1 c2a1 h7h5 d3b5 d7e7 a1b3 c8d6 b5d3 e7d7 h2h3');
INSERT INTO evaluations VALUES(1534,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(1534,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(1534,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(1534,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(1534,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(1534,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(1534,36,'cp -23','c5d4 f3d4 c6e5 d4b5 e8d7 a1a7 a8a7 b5a7 f8c5 f1b5 d7c7 a7c8 c7c8 b1d2 g8f6 b3b4 c5b6 e1e2 f6e8 f2f4 e5g4 b5d3 c8d7 c3c4 g4f2 h1f1 f2d3 e2d3 e8f6 b2b3 h8a8 c1b2 a8a2 b2f6 g7f6 g2g3 h7h5 c4d5 e6d5 d2f3 d7e6 f1e1 e6f5');
INSERT INTO evaluations VALUES(1547,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(1547,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(1547,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(1547,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(1547,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(1547,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(1547,36,'cp -17','f3d4 c6e5 d4b5 e8d7 a1a7 a8a7 b5a7 f8c5 f1b5 d7d8 a7c8 d8c8 b1d2 g8e7 f2f4 e5g4 e1e2 e7f5 d2f3 c8c7 b5d3 g4f2 h1e1 f2d3 e2d3 c5e7 c1d2 h8a8 c3c4 a8d8 d3c2 h7h5 d2c3 e7c5 g2g3 f5e3 c2c1 d5c4 b3c4 e3c4 c3g7 c4e3 g7c3 e3d5 c3e5 c7c8');
INSERT INTO evaluations VALUES(3837,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(3837,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(3837,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(3837,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(3837,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(3837,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(3837,36,'cp -8','a7a5 d2d3 c8e6 c4e6 f7e6 c1e3 b7b6 c3a4 c5e3 f2e3 b6b5 a4c3 d8b8 d1e2 b5b4 c3d1 h7h6 a3b4 a5b4 d1f2 a8a1 f1a1 b8b6 f3d2 g8h8 d2b3 f6d7 a1f1 c6e7 b3d2 e7g6 d2c4 b6c5 f2g4 f8f1 g1f1 d7b6 c4b6 c5b6');
INSERT INTO evaluations VALUES(13593,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13593,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13593,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(13593,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(13593,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(13593,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(13593,36,'cp 7','e1g1 f8e7 b1d2 e8g8 d4c5 e7c5 b2b4 c5d6 b4b5 c6a5 c3c4 a5c4 d3c4 d5c4 d2c4 d6c5 c1b2 b7b6 a2a4 c8b7 f3e5 a7a6 a1c1 a8c8 e5d3 a6b5 d3c5 c8c5 d1d8 f8d8 c4b6 c5c1 f1c1 b5a4 b6a4 d8d2 h2h3 b7a6 a4c5 d2b2 c5a6 g8f8 g2g4');
INSERT INTO evaluations VALUES(13592,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13592,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(13592,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(13592,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(13592,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(13592,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13592,36,'cp 0','f8d6 d4c5 d6c5 b2b4 c5d6 b4b5 c6a5 b1d2 e8g8 c3c4 d5c4 d3c4 a5c4 d2c4 d6c5 c1b2 b7b6 a2a4 c8b7 d1e2 a7a6 c4e5 a6b5 a4b5 a8a1 f1a1 d8d5 a1d1 d5a2 d1a1');
INSERT INTO evaluations VALUES(8397,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8397,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8397,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8397,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8397,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(8397,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(8397,36,'cp 18','d4c5 d6c5 b1d2 e8g8 b2b4 c5d6 b4b5 c6a5 c3c4 d5c4 d3c4 a5c4 d2c4 d6c5 c1b2 b7b6 d1c2 c8b7 f1d1 d8e8 a2a4 a8c8 a1c1 f6e4 f3e5 c5e7 c2e2 f7f6 e5d7 f8f7 f2f3 e4c5 d7c5 e7c5 c4d6 c5d6 d1d6 f7d7 c1c8 e8c8 d6d7 c8d7 e3e4 d7e7 e2d3 e6e5 d3c4 e7f7');
INSERT INTO evaluations VALUES(17966,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(17966,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(17966,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(17966,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(17966,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(17966,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(17966,36,'cp 98','d2d4 d7d6 f1d3 e7e5 d4d5 c6e7 c2c4 g7g6 b1c3 f8g7 a2a4 f7f5 a4a5 g8f6 e1g1 e8g8 f1e1 f6e4 c3e4 f5e4 d3e4 e7f5 c1g5 d8d7 g5d2 f5d4 d2e3 d4f5 h2h3 f5e3 e1e3 g7h6 e3b3 a8b8');
INSERT INTO evaluations VALUES(17962,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(17962,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(17962,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(17962,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(17962,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(17962,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(17962,36,'cp 76','d7d6 f1d3 g8f6 h2h3 c6b4 e1g1 b4d3 d1d3 e7e6 c2c4 f8e7 b1c3 e8g8 f1e1 b7b6 c1e3 h7h6 a1d1 c8b7 a2a3 b6b5 d4d5 b5c4 d3c4 e6d5 e4d5 d8d7 f3d4 f8e8 b2b4 e7f8 d1c1 e8e5 d4c6 b7c6 c4c6 d7c6 d5c6 f6d5 c3d5');
INSERT INTO evaluations VALUES(17730,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17730,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(17730,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(17730,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(17730,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(17730,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(17730,36,'cp 103','b1c3 g8f6 d4d5 c6b8 c1e3 c7c6 a2a4 b8d7 a4a5 g7g6 f1c4 a8b8 f3g5 b7b5 a5b6 d7b6 c4b3 c6d5 e3b6 b8b6 c3d5 b6b3 c2b3 f8g7 d5c3 e8g8 e1g1 c8b7 f1e1 d8b6 c3d5 f6d5 e4d5 g7b2 e1e7 b2a1');
INSERT INTO evaluations VALUES(29265,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(29265,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29265,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(29265,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(29265,36,'cp 0','b8d7 f1d3 d5c4 d3c4 f8e7 e1g1 h7h6 g5h4 e8g8 a2a4 f6d5 h4e7 d8e7 a4a5 d5c3 b2c3 b7b5 a5b6 a7b6 d1b3 b6b5 c4e2 f8b8 a1a8 b8a8 c3c4 b5c4 b3b7 e7d8 e2c4 f5e4 f3d2 e4d5 f1c1 a8b8 b7a6 b8a8');
INSERT INTO evaluations VALUES(24500,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(24500,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(24500,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24500,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(24500,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(24500,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(24500,36,'cp 13','f1d3 d5c4 d3c4 f8e7 e1g1 e8g8 a2a4 h7h6 g5h4 f6d5 h4e7 d8e7 h2h3 d5b4 a4a5 f5g6 d1b3 b7b5 a5b6 a7b6 a1a8 f8a8 f1c1 c6c5 c4e2 g6d3 e2d3 b4d3 c1d1 d3b4 b3c4 d7f6 d4c5 b6c5 f3d2 e7b7 d2e4 f6e4 c3e4 b4d5 c4c5 b7b2');
INSERT INTO evaluations VALUES(34377,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(34377,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(34377,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(34377,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(34377,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(34377,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(34377,36,'cp 16','c3d4 e7e5 g1f3 e5d4 d1d4 g8f6 b1c3 d5e6 c1e3 b8c6 d4a4 f8b4 f1b5 e8g8 e1g1 b4c3 b2c3 c8d7 f1e1 f8e8 a4b3 c6a5 b3e6 e8e6 b5d3 e6e8 c3c4 d7e6 a1c1 a5c6 h2h3 h7h6 g2g4 b7b6 e1d1 e8d8 d3e2');
INSERT INTO evaluations VALUES(34373,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(34373,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(34373,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(34373,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(34373,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(34373,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(34373,36,'cp 8','e7e5 g1f3 e5d4 d1d4 g8f6 b1c3 d5e6 c1e3 b8c6 f1b5 f8e7 e1g1 e8g8 d4a4 c8d7 f1d1 a7a6 b5c4 e6f5 c4d3 f5h5 a4f4 h7h6 h2h3 a8c8 a1c1 d7e6 e3b6 f8e8 a2a3 f6d5 c3d5 e6d5 d3c4 d5f3 f4f3 h5f3 g2f3 c6e5');
INSERT INTO evaluations VALUES(35125,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(35125,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(35125,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(35125,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(35125,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(35125,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(35125,36,'cp 59','g1f3 b8c6 c2c3 d7d6 g2g3 f6g6 f3h4 g6f6 f1g2 g7g5 h4f3 g5g4 f3h4 g8e7 f2f3 g4f3 e1g1 e7g6 h4f3 f6g7 d3d4 h7h5 h2h4 c8d7 d4d5 c6d8 b1d2 f7f6 f1f2 d8f7 d1b3 e8c8 d2f1 f7h6 c1h6 g7h6 f3d2 g6e7 g1h2 f6f5');
INSERT INTO evaluations VALUES(38716,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38716,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38716,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(38716,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(38716,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(38716,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(38716,36,'cp 19','g8f6 b2b4 c5b6 g1f3 b8c6 b4b5 c6d4 f3d4 b6d4 c1b2 e8g8 e1g1 c8e6 d2d3 c7c6 b5c6 b7c6 e2e3 d4c3 b2c3 d8d7 c4c5 e6h3 c5d6 h3g2 g1g2 d7d6 a1c1 f8e8 d1c2 h7h6 f1d1 c6c5 c3a1 a8c8 h2h3 d6e6');
INSERT INTO evaluations VALUES(44078,30,'cp -110',NULL);
INSERT INTO evaluations VALUES(44078,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(44078,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(44078,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(44078,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(44078,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(44078,36,'cp -111','d7d5 c4b3 a7a5 a2a3 a5a4 b3a2 d8a5 b1d2 f8d6 h2h3 c8e6 g1e2 b8d7 e2g3 e8g8 e1g1 f8e8 f3e2 h7h6 d2f3 a5b6 f1e1 d6f8 e2d1 f8c5 d1e2 c5d6 f3h4 d6f8 h4f5 g8h7 f5e3 g7g6 e2f3 d5d4 e3g4 e6a2 g4f6 d7f6 a1a2 c6c5 a2a1 b6c6');
INSERT INTO evaluations VALUES(45805,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(45805,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(45805,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(45805,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(45805,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(45805,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(45805,36,'cp -90','c4b3 a7a5 a2a3 a5a4 b3a2 d8a5 b1d2 f8d6 f3g3 e8g8 g1e2 d5e4 e2c3 e4d3 c2d3 c8f5 e1g1 a5c7 d2c4 b8d7 d3d4 f5e6 c4d6 c7d6 d4e5 d6e5 c1f4 e5f5 a2b1 f5h5 c3e4 f6e4 b1e4 d7f6 f4e5 h5g4 e5f6 g7f6 g3g4 e6g4 f2f3 g4e6 a1d1 a8a5 g1f2');
INSERT INTO evaluations VALUES(46764,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(46764,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46764,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46764,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46764,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46764,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46764,36,'cp 0','f8g7 g1f3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(47959,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(47959,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(47959,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(47959,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(47959,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(47959,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(47959,36,'cp 53','d2d4 d6d5 e4d5 f6d5 c3d5 d8d5 f1d3 d5d6 d1e2 c8e6 e2e4 b8c6 c1f4 d6d7 d3c4 e8c8 c4e6 f7e6 c2c3 d7d5 e4e3 f8d6 e1g1 h8e8 f4d6 c7d6 b2b3 c8b8 e3d2 d5a5 b3b4 a5b6 d2g5 e6e5 g5g7 e8e7 g7f6 e5d4 f3d4 c6d4 f6d4 b6d4 c3d4 e7e4');
INSERT INTO evaluations VALUES(49813,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(49813,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49813,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49813,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(49813,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(49813,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(49813,36,'cp 25','b1c3 c7c6 c3d5 c6d5 d2d4 g7g5 h2h4 g5g4 f3g1 f4f3 g2f3 d7d6 e5d6 f8d6 f3g4 b8c6 c1g5 d6g3 e1d2 d8b6 g1e2 g3f2 c2c3 f7f6 g5f4 c8g4 d1b3 c6a5 b3b6 a7b6 h1h2 a5c4 d2c1 f2e3 f4e3 c4e3 f1h3 h8g8 b2b3 e8f7 c1d2 a8e8 a1e1 e3f5 e2f4 f5h4 h3g4 g8g4 e1e8 h4f3 d2d3 f3h2');
INSERT INTO evaluations VALUES(55349,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(55349,31,'cp -187',NULL);
INSERT INTO evaluations VALUES(55349,32,'cp -169',NULL);
INSERT INTO evaluations VALUES(55349,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(55349,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(55349,35,'cp -186',NULL);
INSERT INTO evaluations VALUES(55349,36,'cp -166','g5g4 f3g1 f4f3 g2f3 d8h4 e1d2 h4f2 d2d3 b8c6 c2c3 f7f5 b1d2 g8f6 d3c2 f5e4 f3e4 d6d5 d1e2 f2e2 c4e2 f6e4 d2e4 c8f5 e2d3 d5e4 d3c4 c6a5 c4b5 c7c6 b5e2 b7b5 c1f4 a5c4 a1f1 f8e7 e2c4 b5c4 f4e3 f5e6 g1e2 e8d7 e3h6 e4e3 h6e3 e6d5 h1g1');
INSERT INTO evaluations VALUES(55358,30,'cp -192',NULL);
INSERT INTO evaluations VALUES(55358,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(55358,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(55358,33,'cp -195',NULL);
INSERT INTO evaluations VALUES(55358,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(55358,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(55358,36,'cp -175','f3g1 f4f3 e1f1 b8c6 g2f3 d8h4 b1c3 g8f6 c1f4 g4f3 d1f3 c6d4 f3g3 h4g4 c3d5 f6d5 e4d5 h8g8 a1e1 e8d8 c2c3 b7b5 c4d3 c8f5 d3e4 f5e4 e1e4 d4f5 g3f3 f5e7 f4d2 g4f3 g1f3 e7d5 e4h4 g8h8 f3d4 d8d7 d4b5 a8b8 c3c4 f8g7 b2b3');
INSERT INTO evaluations VALUES(1176,30,'cp -548',NULL);
INSERT INTO evaluations VALUES(1176,31,'cp -572',NULL);
INSERT INTO evaluations VALUES(1176,32,'cp -571',NULL);
INSERT INTO evaluations VALUES(1176,33,'cp -550',NULL);
INSERT INTO evaluations VALUES(1176,34,'cp -563',NULL);
INSERT INTO evaluations VALUES(1176,35,'cp -571',NULL);
INSERT INTO evaluations VALUES(1176,36,'cp -563','d4e4 d1e2 e4e2 e1e2 e8f7 d2d3 f8d6 h1e1 g8f6 e2f1 c8f5 b1c3 h8e8 c1g5 d6h2 a1d1 h7h5 e1e2 e8e2 c3e2 a8e8 d1e1 h2d6 e2g3 e8e1 f1e1 f5e6 a2a4 c6c5 e1f1 e6g4 g5f6 d6g3 f6c3 g3f4 b2b3 f4d6 c3b2 g4d7 c2c4 h5h4');
INSERT INTO evaluations VALUES(1178,30,'cp -563',NULL);
INSERT INTO evaluations VALUES(1178,31,'cp -548',NULL);
INSERT INTO evaluations VALUES(1178,32,'cp -563',NULL);
INSERT INTO evaluations VALUES(1178,33,'cp -553',NULL);
INSERT INTO evaluations VALUES(1178,34,'cp -563',NULL);
INSERT INTO evaluations VALUES(1178,35,'cp -540',NULL);
INSERT INTO evaluations VALUES(1178,36,'cp -550','d1e2 e4e2 e1e2 e8f7 f2f3 g8f6 d2d3 f8d6 b1c3 h8e8 c3e4 d6e5 e2f2 c6c5 c2c3 c5c4 d3d4 f6e4 f3e4 e5d6 h1e1 f7g8 c1g5 c8e6 e1f1 c7c6 f2f3 e6f7 g5f4 d6e7 g2g4 f7g6 f3g3 g6e4 f1e1 e4d5 h2h4 g8h8 g4g5 d5f7 e1e7 e8e7 a2a4 f7d5');
INSERT INTO evaluations VALUES(1179,30,'cp -540',NULL);
INSERT INTO evaluations VALUES(1179,31,'cp -551',NULL);
INSERT INTO evaluations VALUES(1179,32,'cp -563',NULL);
INSERT INTO evaluations VALUES(1179,33,'cp -571',NULL);
INSERT INTO evaluations VALUES(1179,34,'cp -570',NULL);
INSERT INTO evaluations VALUES(1179,35,'cp -567',NULL);
INSERT INTO evaluations VALUES(1179,36,'cp -555','e4e2 e1e2 e8f7 d2d3 f8d6 b1c3 g8f6 c3e4 h8e8 f2f3 d6e5 e2f2 e5d4 f2e2 c6c5 c2c3 d4e5 e2f2 c5c4 d3d4 f6e4 f3e4 e5d6 h1e1 f7g8 c1g5 c8e6 f2g1 e6f7 e4e5 d6f8 a2a4 f7d5 e1f1 a6a5 g5f4 h7h6 f4g3 d5e6 f1f8 e8f8 a1f1 f8f1 g1f1 g8f7 g3f4 a8e8 f1f2 b7b6');
INSERT INTO evaluations VALUES(1181,30,'cp -544',NULL);
INSERT INTO evaluations VALUES(1181,31,'cp -553',NULL);
INSERT INTO evaluations VALUES(1181,32,'cp -555',NULL);
INSERT INTO evaluations VALUES(1181,33,'cp -555',NULL);
INSERT INTO evaluations VALUES(1181,34,'cp -550',NULL);
INSERT INTO evaluations VALUES(1181,35,'cp -548',NULL);
INSERT INTO evaluations VALUES(1181,36,'cp -550','e1e2 e8f7 d2d3 f8d6 f2f3 g8f6 b1d2 h8e8 d2e4 d6e5 c2c3 c6c5 e2f2 c5c4 d3d4 f6e4 f3e4 e5d6 h1e1 f7g8 c1g5 c8e6 a2a4 e8f8 f2g1 a8e8 e1f1 f8f7 f1f7 e6f7 e4e5 d6f8 a1e1 f7e6 e1f1 h7h6 g5h4 g7g5 f1f8 e8f8 h4g3 g8f7 h2h4 g5h4 g3h4 e6d5 a4a5 f7e6 h4g3 f8f7');
INSERT INTO evaluations VALUES(1180,30,'cp -537',NULL);
INSERT INTO evaluations VALUES(1180,31,'cp -555',NULL);
INSERT INTO evaluations VALUES(1180,32,'cp -568',NULL);
INSERT INTO evaluations VALUES(1180,33,'cp -571',NULL);
INSERT INTO evaluations VALUES(1180,34,'cp -563',NULL);
INSERT INTO evaluations VALUES(1180,35,'cp -544',NULL);
INSERT INTO evaluations VALUES(1180,36,'cp -563','e8f7 d2d3 f8d6 b1d2 g8f6 d2e4 h8e8 f2f3 d6e5 c2c3 c6c5 e2f2 c5c4 d3d4 f6e4 f3e4 e5d6 h1e1 f7g8 c1g5 b7b5 g2g3 c8g4 e1e3 a8b8 e4e5 d6f8 e5e6 e8e6 e3e6 g4e6 a1e1 e6d5 e1e2 b5b4 h2h4 b8b6 e2e8 d5c6 e8f8 g8f8 a2a3 b4c3 b2c3');
INSERT INTO evaluations VALUES(2166,30,'cp -555',NULL);
INSERT INTO evaluations VALUES(2166,31,'cp -546',NULL);
INSERT INTO evaluations VALUES(2166,32,'cp -555',NULL);
INSERT INTO evaluations VALUES(2166,33,'cp -555',NULL);
INSERT INTO evaluations VALUES(2166,34,'cp -555',NULL);
INSERT INTO evaluations VALUES(2166,35,'cp -559',NULL);
INSERT INTO evaluations VALUES(2166,36,'cp -551','d2d3 f8d6 b1d2 g8f6 d2e4 h8e8 f2f3 d6e5 c2c3 c6c5 e2f2 c5c4 d3d4 f6e4 f3e4 e5d6 h1e1 f7g8 c1g5 b7b5 e1e3 b5b4 e4e5 d6f8 c3b4 h7h6 g5f4 c8e6 e3g3 e8b8 f4h6 b8b4 a1f1 b4b2 f2g1 c4c3 f1f8 a8f8 g3g7 g8h8 h2h4 c3c2 g1h2 f8g8 g7g8 h8g8 h6g5 b2b1');
INSERT INTO evaluations VALUES(33297,30,'cp 657',NULL);
INSERT INTO evaluations VALUES(33297,31,'cp 659',NULL);
INSERT INTO evaluations VALUES(33297,32,'cp 657',NULL);
INSERT INTO evaluations VALUES(33297,33,'cp 654',NULL);
INSERT INTO evaluations VALUES(33297,34,'cp 655',NULL);
INSERT INTO evaluations VALUES(33297,35,'cp 648',NULL);
INSERT INTO evaluations VALUES(33297,36,'cp 655','f2g1 h4e4 d2d4 b8c6 e5f3 e4e7 b1c3 d7d6 c1d2 g8f6 d2g5 e8g8 c3d5 f6d5 g5e7 c6e7 h2h3 c8f5 d1d2 f5e4 f3g5 e4g6 g1h2 f8e8 f1e2 h7h6 g5f3 e7f5 a1e1 d5f6 e2d3 f5e7 f3h4 g6d3 d2d3 d6d5 h1f1 e7c6 e1e8 a8e8 d3g3 g8h8 g3c7 c6d4 c7b7');
INSERT INTO evaluations VALUES(33298,30,'cp 627',NULL);
INSERT INTO evaluations VALUES(33298,31,'cp 628',NULL);
INSERT INTO evaluations VALUES(33298,32,'cp 626',NULL);
INSERT INTO evaluations VALUES(33298,33,'cp 625',NULL);
INSERT INTO evaluations VALUES(33298,34,'cp 618',NULL);
INSERT INTO evaluations VALUES(33298,35,'cp 628',NULL);
INSERT INTO evaluations VALUES(33298,36,'cp 624','d7d6 e5f3 h4e4 b1c3 e4e6 d2d4 d6d5 h2h3 g8f6 c1f4 e8g8 g1h2 e6b6 f1d3 c7c6 d1d2 b8d7 h1f1 b6b2 f3e5 b2a3 a1e1 f6h5 c3e2 f7f6 e5d7 c8d7 d2c1 a3c1 f4c1 g7g6 e2g3 h5g3 h2g3 d7f5 d3f5 g6f5 g3h2');
INSERT INTO evaluations VALUES(20164,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20164,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20164,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20164,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20164,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20164,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20164,36,'cp 0','c8d7 b1c3 d8b6 b5c6 b6c6 g1e2 b7b5 a2a3 b5b4 a3b4 f8b4 e1g1 g8e7 c3a2 b4a5 a2c3');
INSERT INTO evaluations VALUES(25916,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25916,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25916,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25916,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25916,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25916,36,'cp 0','b1c3 d8b6 b5c6 b6c6 g1e2 b7b5 e2f4 b5b4 c3e2 a8c8 h2h4 g8e7 h4h5 c6c2 d1c2 c8c2 a2a3 b4b3 h1h3 d7a4 e2c3 a4d7');
INSERT INTO evaluations VALUES(50088,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50088,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(50088,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(50088,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(50088,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50088,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(50088,36,'cp 29','e7e6 a2a3 f8e7 g1f3 e8g8 e2e3 d7d6 f1d3 a7a6 h2h3 b7b5 e1g1 d8e8 a3a4 b5b4 c3e2 c8b7 d3c4 e8d7 c2c3 b7f3 g2f3 d6d5 c4b3 b8c6 a4a5 b4c3 b2c3 g8h8 c3c4 a8b8 c4d5 f6d5 b3d5 e6d5 f4e5 b8b5 d1c2 e7d6 e5d6 c7d6 f1c1 b5a5 a1a5 c6a5 e2f4');
INSERT INTO evaluations VALUES(49946,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(49946,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(49946,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(49946,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(49946,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(49946,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(49946,36,'cp 31','a2a3 f8e7 g1f3 f6e4 c3e4 f5e4 f3d2 e8g8 e2e3 d7d5 f1e2 e7d6 f4d6 c7d6 e1g1 b8c6 c2c4 d5c4 e2c4 d6d5 c4a2 c8d7 f2f3 e4f3 f1f3 f8f3 d1f3 d8e7 a1f1 a8f8 f3g4 f8f1 g1f1 e7d6 d2f3 d6c7 g4h4 a7a5 a2b1 h7h6');
INSERT INTO evaluations VALUES(54104,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(54104,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(54104,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(54104,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(54104,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(54104,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(54104,36,'cp 90','c5d4 d1d4 d8d4 f3d4 a7a6 c1f4 g8f6 e4d6 f8d6 f4d6 c8d7 d4b3 b8c6 f2f3 h7h5 e1c1 f6d5 d1d2 e8c8 h2h4 d7e8 d6c5 f7f6 g2g3 e8g6 a2a3 g6f5 f1e2 e6e5 h1d1 d5c7 c5b6 d8d2 b3d2 h8d8 g3g4 f5g6 f3f4 h5g4 e2g4 f6f5 g4h3 e5f4 d1g1');
INSERT INTO evaluations VALUES(54460,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(54460,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(54460,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(54460,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(54460,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(54460,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(54460,36,'cp 86','d1d4 d8d4 f3d4 a7a6 c1f4 c8d7 d4b3 g8f6 e4d6 f8d6 f4d6 b8c6 f2f3 f6d5 e1f2 e8c8 a1d1 f7f6 b3c5 d7e8 d6g3 e8f7 f1d3 d5e7 d3c4 e7d5 h1e1 e6e5 f2g1 h8e8 g3f2 e8e7 c4d3 f7g6 d3f1 c8c7 c2c3 d5f4 d1d8 c6d8 g2g3 f4e6 h2h4 g6f5 c5a4 e7d7 a4b6');
INSERT INTO evaluations VALUES(54459,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(54459,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(54459,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(54459,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(54459,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(54459,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(54459,36,'cp 80','d8d4 f3d4 a7a6 c1f4 c8d7 e4d6 f8d6 f4d6 g8f6 d4b3 b8c6 f2f3 f6d5 e1f2 e8c8 a1d1 f7f6 b3c5 d7e8 d6g3 e8f7 f1d3 d5e7 d3c4 e7d5 c4d5 e6d5 d1d2 h7h5 h2h4 h8e8 h1d1 c6e5 c5a4 b7b5 g3e5 b5a4 e5d4 d8d7 c2c3 d7e7 b2b3 e7c7 b3a4 c8d7 d1b1');
INSERT INTO evaluations VALUES(8084,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(8084,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8084,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8084,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8084,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8084,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8084,36,'cp 7','f6h5 d4c5 h5f4 e3f4 e7c5 e1g1 c5b6 g2g3 e8g8 b2b4 d5d4 c3d4 b6d4 f3d4 d8d4 d2f3 d4b4 a1b1 b4a3 d1e2 f8d8 f1d1 a3e7 h2h4 e6e5 f3e5 c6e5 f4e5 g7g6 d3e4 d8d1 e2d1 e7e5 e4b7 c8b7 b1b7 a8b8 b7b8 e5b8');
INSERT INTO evaluations VALUES(8920,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(8920,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(8920,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(8920,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(8920,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(8920,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(8920,36,'cp 46','h2h3 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 a7a5 d1b3 c8f5 e4e1 d8d6 c1e3 d6b4 b3a3 e7c8 c3b5 b4a3 a1a3 f5d7 e3b6 c8b6 b5c7 a8c8 d5d6 d7c6 a3d3 b6a4 c7d5 a4c5 d3d4 c6d5 d4d5 f8d8');
INSERT INTO evaluations VALUES(16255,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(16255,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(16255,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(16255,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(16255,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(16255,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(16255,36,'cp -43','d7d5 e4d5 e6d5 c4b5 f8d6 d2d4 h7h6 e1g1 e8g8 h2h3 f8e8 b5d3 c6b4 d3e2 c7c5 d4c5 d6c5 a2a3 b4c6 e2d3 f6e4 d3e4 d5e4 d1d8 c6d8 f1e1 c8f5 g2g4 d8c6 g4f5 e4f3 c1e3 c5e3 e1e3 e8d8 a1e1 d8d2 c3e4 d2c2 e4d6 c2b2 e3e8 a8e8 e1e8 g8h7 d6f7');
INSERT INTO evaluations VALUES(14766,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(14766,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(14766,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(14766,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(14766,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(14766,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(14766,36,'cp -50','e4d5 e6d5 c4b5 f8d6 d2d4 h7h6 e1g1 e8g8 h2h3 a7a6 b5c6 b7c6 f1e1 f8e8 e1e8 d8e8 c3a4 f6e4 c1e3 c8f5 a4c5 f7f6 c5b7 d6f8 b7c5 a6a5 a2a4 e8f7 f3h4 f5c8 h4f3 f8d6 d1e2 c8f5 c5b7 d6f8 b7c5 a8b8 b2b3 b8e8 c5b7 f8b4');
INSERT INTO evaluations VALUES(14747,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(14747,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(14747,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(14747,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(14747,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(14747,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(14747,36,'cp -43','e6d5 c4b5 f8d6 d2d4 h7h6 h2h3 e8g8 e1g1 a7a6 b5c6 b7c6 f1e1 f8e8 e1e8 d8e8 c1e3 f6e4 c3e4 e8e4 a1c1 c8e6 f3e1 a8e8 d1d3 e4h4 e1f3 h4f6 e3d2 e6f5 d3c3 f5e4 c1e1 e8e6 e1e3 f6f5 g1f1 e4c2 e3e6 f7e6 c3c6 c2e4');
INSERT INTO evaluations VALUES(15022,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(15022,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(15022,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(15022,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(15022,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(15022,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(15022,36,'cp -40','c4b5 f8d6 d2d4 e8g8 c1g5 f8e8 b5e2 c8e6 c3b5 d6e7 g5f4 e7b4 c2c3 b4a5 e1g1 a7a6 b5a3 c6e7 h2h3 c7c6 e2d3 e7g6 f4h2 f6e4 f1e1 f7f5 a3c2 a5c7 h2c7 d8c7 d1c1 a8d8 c2e3 e8e7 a2a4 e6d7 e3f1 d8e8 c3c4 g6f4 f3e5 f4d3 e5d3 d5c4');
INSERT INTO evaluations VALUES(18602,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(18602,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(18602,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(18602,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(18602,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(18602,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(18602,36,'cp 8','a7a6 b5d3 g8f6 f1e1 e7e6 d3f1 b7b6 c2c4 f8e7 b1c3 c8b7 d2d4 c5d4 f3d4 e8g8 b2b3 d8c7 d1d3 h7h6 c1d2 d7c5 d3f3 f8e8 a1d1 c5d7 h2h3 a8d8 f3g3 f6h5 g3e3 h5f6 e3f3');
INSERT INTO evaluations VALUES(17321,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(17321,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(17321,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(17321,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(17321,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(17321,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(17321,36,'cp 18','b5d3 g8f6 f1e1 e7e6 d3f1 f8e7 d2d4 c5d4 f3d4 d7c5 d1f3 d8c7 c2c4 b7b6 b2b4 c5d7 f3g3 e8g8 b1d2 c8b7 c1b2 h7h6 a1d1 f8e8 a2a3 a8d8 h2h3 f6h5 g3g4 h5f6 g4e2 d7e5 b4b5 a6a5 f2f4 e5g6 e2f2 e7f8');
INSERT INTO evaluations VALUES(19993,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19993,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(19993,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(19993,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(19993,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(19993,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(19993,36,'cp -29','e7e5 f1c4 a7a6 a2a4 g7g6 b1c3 f8g7 c3d5 a8b8 d1e2 g8e7 d5e7 d8e7 e1g1 e8g8 h2h3 c8e6 d2d3 e7d7 b2a3 c6d4 f3d4 c5d4 a1e1 f8e8 c4e6 e8e6 f2f4 e5f4 f1f4 g7e5 f4f1 b8e8 e2f2 b7b5 a4b5 a6b5 f2f3');
INSERT INTO evaluations VALUES(20612,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(20612,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(20612,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(20612,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(20612,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(20612,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(20612,36,'cp -13','f1c4 a7a6 a2a4 g7g6 b1c3 f8g7 c3d5 g8f6 d2d3 f6d5 c4d5 c6e7 d5c4 e8g8 h2h3 a8b8 a4a5 e7c6 b2c3 h7h6 e1g1 g8h7 c4d5 f7f5 f3d2 f5f4 f2f3 g6g5 d2c4 h6h5 c4b6 d8f6 b6c8 b8c8 g1h1 c8d8 d1e2 d8e8 e2f2');
INSERT INTO evaluations VALUES(20591,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(20591,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(20591,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(20591,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(20591,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(20591,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(20591,36,'cp -23','a7a6 a2a4 g7g6 b1c3 a8b8 d1e2 f8g7 c3d5 g8e7 d5e7 d8e7 e1g1 e8g8 c2c3 c8e6 d2d3 e7d7 b2a3 b7b5 c4e6 d7e6 b3b4 c5b4 a3b4 c6b4 c3b4 d6d5 e2a2 e6d6 a2b3 d5e4 d3e4 f8c8 f1d1 b5a4 a1a4 d6c6 f3g5 b8b7 g2g3 g7f8 d1a1');
INSERT INTO evaluations VALUES(29630,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(29630,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(29630,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(29630,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(29630,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(29630,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(29630,36,'cp 49','d1b3 g4c8 c4d5 d8d5 b3d5 f6d5 g1f3 c7c5 b1c3 e7e6 c3d5 e6d5 d4c5 f8c5 c1d2 b8c6 f1b5 c8d7 e1g1 e8g8 f1d1 a8c8 d2c3 d7e6 d1d2 f8d8 h2h3 c5b6 a1d1 a7a6 b5a4 c6a5 c3d4 b6d4 f3d4 g8f8 a4b3 a5c6 a2a3 c6d4 d2d4');
INSERT INTO evaluations VALUES(32279,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(32279,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(32279,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(32279,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32279,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(32279,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(32279,36,'cp 5','c7c6 g1f3 f8e7 f1d3 f5g6 e1g1 g8f6 b2b3 b8d7 c1b2 e8g8 d3g6 h7g6 d1e2 f8e8 e3e4 d5e4 c3e4 f6e4 e2e4 d8a5 f1d1 a5f5 e4f5 g6f5 d4d5 c6d5 c4d5 d7b6 d5e6 f7e6 g2g3 b6d5 a1c1 e8c8 h2h4 a7a5 a2a4 c8c1 d1c1 a8d8');
INSERT INTO evaluations VALUES(48193,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(48193,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(48193,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(48193,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(48193,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(48193,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(48193,36,'cp -38','e6d5 d2d4 f8e7 f1g2 e8g8 g1f3 c7c6 d1c2 e7d6 e1g1 b8d7 a2a3 f8e8 f1e1 a7a5 c1d2 h7h6 a1d1 d7b6 f3e5 a5a4 e2e4 d5e4 c3e4 c8f5 e1e2 f5h7 d1e1 d6f8 g3g4 f6e4 g2e4 h7e4 c2e4');
INSERT INTO evaluations VALUES(48731,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(48731,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48731,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(48731,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(48731,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(48731,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48731,36,'cp -20','d2d4 c7c6 d1c2 f8d6 c1f4 e8g8 g1f3 g7g6 f4d6 d8d6 f1g2 c8f5 c2d2 b8d7 e1g1 f6e4 c3e4 f5e4 a1c1 e4f3 e2f3 a7a5 h2h4 d7b6 b2b3 a5a4 g2h3 a4b3 a2b3 a8a3 c1b1 f8a8 f1e1 a3a2 d2e3 d6b4');
INSERT INTO evaluations VALUES(48660,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(48660,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(48660,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(48660,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(48660,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(48660,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(48660,36,'cp -41','c7c6 c1f4 f8b4 g1f3 e8g8 f1g2 c8f5 d1b3 d8b6 e1g1 b4e7 f1c1 b8d7 f3e5 h7h6 e5d7 f6d7 e2e4 d5e4 c3e4 f8d8 b3b6 d7b6 e4c5 f5c8 f4e3 a7a5 a2a3 b6d5 e3d2 e7f6 d2c3 d5c3 b2c3 a5a4 f2f4');
INSERT INTO evaluations VALUES(45772,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(45772,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(45772,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(45772,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(45772,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(45772,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(45772,36,'cp -35','c1f4 f8e7 d1c2 e8g8 e2e3 f8e8 a2a3 c8e6 f1d3 h7h6 g1f3 b8d7 h2h3 c6c5 g3g4 a8c8 g4g5 h6g5 f3g5 c5d4 g5e6 f7e6 e3d4 e6e5 d4e5 d7e5 f4e5 e7d6 e1c1 d6e5 h1e1 e5c3 e1e8 f6e8 b2c3 e8d6 c2b3 d8f6 b3d5 g8f8 c1b1 c8c3 d3f1');
INSERT INTO evaluations VALUES(51905,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(51905,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(51905,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(51905,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(51905,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(51905,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(51905,36,'cp -63','d7d6 g1f3 b8c6 b2b3 g8h6 c1b2 d6d5 b1c3 f8e7 f1b5 c8d7 b5c6 d7c6 d2d4 h6f5 c3e2 c6b5 e2g3 e7h4 d1d2 h4g3 h2g3 f5g3 h1h3 g3e4 d2e3 d8a5 f3d2 c5c4 a2a4 b5c6 b2c3 a5c3 e3c3 e4c3 h3c3 c4b3 c2b3 a7a5 d2f3 e8g8 e1f2 f7f6 e5f6 g7f6 a1e1 e6e5');
INSERT INTO evaluations VALUES(53199,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(53199,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(53199,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(53199,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(53199,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(53199,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(53199,36,'cp 44','d4d5 a7a6 a2a4 e7e6 e2e4 e6d5 c3d5 g8e7 f1c4 e7d5 c4d5 b8c6 g1f3 f8e7 d5c6 b7c6 d1d8 e7d8 c1f4 e8g8 e1c1 f8e8 h1e1 c8e6 c1c2 d8a5 f4d2 a5c7 f3g5 c7h2 g5e6 e8e6 d2c3 h2e5 f2f3 e5c3 c2c3');
INSERT INTO evaluations VALUES(53226,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(53226,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(53226,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(53226,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(53226,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(53226,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(53226,36,'cp 30','a7a6 a2a4 e7e6 e2e4 e6d5 e4d5 f8d6 f1c4 b8d7 g1f3 g8f6 d1e2 d8e7 e1g1 e7e2 c4e2 e8g8 f3d2 d6e5 e2f3 e5d4 d2c4 d7e5 c4e5 d4e5 c1e3 c8g4 f3g4 f6g4 e3c5 f8d8 c5b6 d8d7 f2f4 e5c3 b2c3 d7d5 c3c4 d5d7 f1d1 g4f6');
INSERT INTO evaluations VALUES(56258,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(56258,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(56258,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(56258,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(56258,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(56258,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(56258,36,'cp 55','d2d3 g8f6 d3e4 f6e4 b1d2 e4d2 c1d2 c7c6 d2c3 b8d7 e3e4 d5e4 g1e2 e4e3 f2e3 d8g5 d1d4 d7f6 e1c1 f8e7 h2h3 c8d7 d4f4 g5f4 e2f4 e8c8 e3e4 d7e6 c3d4 d8d7 g2f3 h8d8 f4e6 f7e6 d4a7 d7d1 h1d1');
INSERT INTO evaluations VALUES(56257,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(56257,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(56257,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(56257,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(56257,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(56257,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(56257,36,'cp 45','g8f6 d3e4 f6e4 b1d2 e4f6 g1e2 b8c6 e1g1 h7h5 e3e4 d5e4 d2e4 f6e4 d1d8 c6d8 g2e4 d8e6 f1e1 c7c6 e4f5 f8e7 h2h4 a7a5 c2c3 e8g8 e2f4 f8e8 f5e6 c8e6 f4e6 e7f6 c1g5 f7e6 a1d1 f6g5 h4g5 a8d8 d1d8 e8d8 e1e6 d8d2 g5g6 g8f8 e6e5 d2b2 e5h5 f8e7 h5f5 e7e6 f5a5 e6f6 g1g2 f6g6');
INSERT INTO evaluations VALUES(3330,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(3330,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(3330,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(3330,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(3330,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(3330,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(3330,36,'cp 44','e7d6 f4g3 b7b6 d1c2 h7h6 g3h4 c6e7 h4f6 g7f6 g2g4 g8h8 e1c1 c7c5 h2h4 c5c4 d3e2 b6b5 g4g5 b5b4 c3b4 d6b4 g5f6 e7f5 e3e4 c4c3 e4f5 c3d2 c1b1 d8f6 f3e5 f6f5 e2d3 f5f4 h1g1 b4d6 c2d2 f4d2 d1d2 a8b8 d2e2 c8b7 d3c2');
INSERT INTO evaluations VALUES(3758,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(3758,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(3758,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(3758,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(3758,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(3758,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(3758,36,'cp 75','h2h4 d8e7 f3e5 d6e5 d4e5 f6d7 d2f3 f7f6 e5f6 d7f6 f3e5 c8d7 e5d7 e7d7 e1g1 e6e5 f4h2 a7a6 d1c2 d7e7 d3e2 g8h8 a1d1 f8d8 c2b3 c6a5 b3b4 e7b4 c3b4 a5c6 b4b5 a6b5 e2b5 d8b8 b5c6');
INSERT INTO evaluations VALUES(16715,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(16715,31,'cp -174',NULL);
INSERT INTO evaluations VALUES(16715,32,'cp -172',NULL);
INSERT INTO evaluations VALUES(16715,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(16715,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(16715,35,'cp -194',NULL);
INSERT INTO evaluations VALUES(16715,36,'cp -174','d7d5 b1c3 f8b4 g1e2 c8e6 a2a3 b4a5 b2b4 a5b6 c3a4 d8e7 a4b6 a7b6 b4b5 c6a5 a3a4 a5c4 e4e5 f6d7 f3f4 e8g8 e2c3 e7c5 f1c4 c5c4 d1d3 d5d4 d3c4 e6c4 c3e2 c4b5 e2d4 b5a4 e1g1 c7c5 e5e6 d7f6 e6e7 f8e8 d4f5 a4c2 a1a8 e8a8 f5e3 c2b3 c1b2 a8e8 b2f6 g7f6 f1b1 b3a2 b1b6 e8e7 g1f2 a2e6');
INSERT INTO evaluations VALUES(15061,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(15061,31,'cp -186',NULL);
INSERT INTO evaluations VALUES(15061,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(15061,33,'cp -164',NULL);
INSERT INTO evaluations VALUES(15061,34,'cp -177',NULL);
INSERT INTO evaluations VALUES(15061,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(15061,36,'cp -182','b1c3 f8b4 g1e2 c8e6 a2a3 b4a5 c1g5 d8e7 b2b4 a5b6 c3d5 e6d5 e4d5 c6e5 d1d2 e8g8 e1c1 e5c4 d2d3 b6e3 g5e3 c4e3 d5d6 c7d6 d1e1 e3d5 g2g4 f8e8 f1h3 e7d7 g4g5 d7h3 g5f6 d5f6 e2c3 e8e1 h1e1 h3h2 c1b2 h7h5 d3b5 b7b6 b5c6 a8d8 c6c7 d8d7 c7c8 g8h7 c3d5 h2h4');
INSERT INTO evaluations VALUES(27213,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(27213,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(27213,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(27213,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(27213,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(27213,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(27213,36,'cp 40','a7a6 b5c6 d7c6 f1e1 g8e7 c2c4 e7g6 b1c3 f8e7 d2d4 e8g8 c3d5 d8d7 d1b3 e7d8 d4e5 d6e5 c1e3 f7f5 a1d1 d7f7 b3d3 f5e4 d3e4 d8f6 b2b3 a8e8 h3h4 h7h6 d1d2 b7b6 h4h5 g6f4 e3f4 e5f4 e4g6 e8e1 f3e1 c6d5 d2d5 c7c6 d5d6');
INSERT INTO evaluations VALUES(25714,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(25714,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(25714,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(25714,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(25714,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(25714,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(25714,36,'cp 41','b5c6 d7c6 f1e1 g8e7 c2c4 e7g6 b1c3 f8e7 d2d4 e8g8 c3d5 d8d7 d1b3 e7d8 d4e5 d6e5 c1e3 f7f6 a1d1 d7f7 b3d3 g8h8 d1d2 c6e8 e1d1 c7c6 d5b6 d8b6 e3b6 g6f4 d3d6 c6c5 b6c5 e8c6 c5e3 f4e6 d6d3 f8e8 b2b3 f7g6 f3h4 g6e4 d3e4 c6e4');
INSERT INTO evaluations VALUES(34664,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34664,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(34664,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(34664,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(34664,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(34664,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(34664,36,'cp 19','g8f6 d2d3 c8d7 c1f4 b8a6 e1c1 e8c8 a2a3 a6c5 f2f3 e7e6 h2h4 d7c6 h4h5 h7h6 c3e4 c5e4 d3e4 d8d1 c1d1 f8c5 e2c1 h8d8 c1d3 c6b5 f4d2 c5d6 d1c1 f6d7 f1e2 a7a6 g2g4 b5d3 e2d3 d7e5 d3e2 f7f6');
INSERT INTO evaluations VALUES(33723,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(33723,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(33723,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(33723,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(33723,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(33723,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(33723,36,'cp 26','d2d4 c8f5 d4d5 c7c5 c1e3 f5c2 a1c1 c2d3 e3c5 g7g6 c1d1 d3a6 c5d4 f8g7 g2g4 e8g8 g4g5 f6e8 h2h4 b8d7 d4g7 e8g7 e2g3 a6f1 e1f1 a8d8 f1g2 d7b6 a2a4 b6c4 b2b3 c4d6 c3e4 g7f5 g3f5 g6f5 e4d6 d8d6 g2f3');
INSERT INTO evaluations VALUES(35506,30,'cp -413',NULL);
INSERT INTO evaluations VALUES(35506,31,'cp -416',NULL);
INSERT INTO evaluations VALUES(35506,32,'cp -459',NULL);
INSERT INTO evaluations VALUES(35506,33,'cp -448',NULL);
INSERT INTO evaluations VALUES(35506,34,'cp -443',NULL);
INSERT INTO evaluations VALUES(35506,35,'cp -458',NULL);
INSERT INTO evaluations VALUES(35506,36,'cp -459','f2f4 d7d6 g1f3 b8c6 f4e5 d6e5 d2d4 f6e4 e1g1 e4f6 c2c3 h7h6 c1e3 e5d4 f3d4 f7g8 b1d2 d8d5 d4c6 d5c6 e3d4 h8h7 d1f3 c6f3 d2f3 f8d6 d4f6 g7f6 a1e1 c8d7 f3d4 h7g7 b2b4');
INSERT INTO evaluations VALUES(48379,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(48379,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(48379,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(48379,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(48379,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(48379,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(48379,36,'cp 104','e4e5 f6d7 c2c3 c7c5 a2a3 b7b5 d3b5 d8b6 a3a4 c8a6 d4c5 f8c5 d1e2 e8g8 g1f3 f8c8 e1g1 b6a5 b2b4 a6b5 e2b5 a5b5 a4b5 c5b6 c1d2 d7f8 b1a3 b8d7 c3c4 d5c4 f1c1 a7a6 c1c4 a6b5 c4c8 a8c8 a3b5');
INSERT INTO evaluations VALUES(48428,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(48428,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(48428,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(48428,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(48428,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(48428,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(48428,36,'cp 95','f6d7 c2c3 c7c5 g1e2 b8c6 e1g1 d8b6 d3c2 f8e7 a2a3 e8g8 f2f4 f7f6 g1h1 f8f7 b2b4 c5d4 c3d4 f6e5 d4e5 a7a5 b4b5 b6b5 b1c3 b5a6 a1b1 b7b6 c2d3 a6b7 c3d5 e6d5 e5e6 f7f8 e6d7 b7d7 b1b6 e7c5 b6b5 d7e7 e2c3');
INSERT INTO evaluations VALUES(49163,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(49163,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(49163,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(49163,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(49163,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(49163,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(49163,36,'cp 92','g1e2 c7c5 c2c3 b8c6 e1g1 d8b6 d3c2 f8e7 a2a3 e8g8 f2f4 g7g6 f1f3 a7a5 g1h1 b6c7 c1e3 b7b5 b1d2 c5d4 e2d4 c6d4 e3d4 f7f5 c2d3 d7c5 d3b5 c5e4 a3a4 c8a6 d1e2 e4d2 e2d2 f8b8 d2e2 a6b5 a4b5 e7c5 d4e3 c5e3 f3e3 c7b6');
INSERT INTO evaluations VALUES(49183,30,'cp -470',NULL);
INSERT INTO evaluations VALUES(49183,31,'cp -463',NULL);
INSERT INTO evaluations VALUES(49183,32,'cp -467',NULL);
INSERT INTO evaluations VALUES(49183,33,'cp -473',NULL);
INSERT INTO evaluations VALUES(49183,34,'cp -478',NULL);
INSERT INTO evaluations VALUES(49183,35,'cp -492',NULL);
INSERT INTO evaluations VALUES(49183,36,'cp -482','e8f7 f2f4 d7d6 g1f3 b8c6 d2d3 e5f4 c1f4 f8e7 f3g5 f7g8 g5f3 h7h6 d1e2 c8e6 b1c3 d8d7 e1g1 a8f8 d3d4 d6d5 c3d5 f6d5 e4d5 d7d5 e2e3 e7f6 a1e1 f8e8 e3d3 c6d4 g1h1 d5b5 c2c4 b5b2 f3d4 b2d4 d3d4');
INSERT INTO evaluations VALUES(52557,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(52557,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(52557,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(52557,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(52557,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(52557,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(52557,36,'cp 91','c8g4 b1d2 f6e5 d4e5 e7e6 h2h3 g4h5 f1e2 g8h6 e1g1 b8a6 c2c3 h6f7 f1e1 d8d7 d2f1 h5f3 e2f3 f8e7 f1g3 e8g8 f3g4 e7g5 c1g5 f7g5 d1d2 g5f7 g3h5 d7e7 b2b4 a6c7 a2a3 e7h4 d2e3');
INSERT INTO evaluations VALUES(30945,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(30945,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(30945,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(30945,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(30945,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(30945,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(30945,36,'cp 122','b1d2 e7e6 f1e2 b8d7 e1g1 g8h6 c2c4 f8e7 e5f6 d7f6 h2h3 g4f5 f3e5 h6f7 e5f7 e8f7 d2f3 h8e8 c4c5 f7g8 c1f4 f6e4 d1c1 a7a5 f3e5 e7g5 f4g5 d8g5 f2f3 g5c1 f1c1 e4g3 g1f2 g3e2 f2e2 g7g5 e2d2 h7h5 c1e1 e8e7 a2a3 a8f8');
INSERT INTO evaluations VALUES(56959,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(56959,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(56959,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(56959,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(56959,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(56959,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(56959,36,'cp 78','e2e4 d7d5 e4e5 c7c5 c2c3 d8c7 f1e2 b8c6 b1a3 a7a6 a3c2 c8d7 e1g1 c5d4 c3d4 g8e7 c1d2 a8c8 e2d3 c6a5 c2e3 c7b6 f1e1 d7b5 d3c2 a5c4 d2c3 c4a3 c2b3 a3c4 a1c1 c4e3 e1e3 b5c4 h2h3');
INSERT INTO evaluations VALUES(937,30,'cp -463',NULL);
INSERT INTO evaluations VALUES(937,31,'cp -492',NULL);
INSERT INTO evaluations VALUES(937,32,'cp -485',NULL);
INSERT INTO evaluations VALUES(937,33,'cp -490',NULL);
INSERT INTO evaluations VALUES(937,34,'cp -494',NULL);
INSERT INTO evaluations VALUES(937,35,'cp -505',NULL);
INSERT INTO evaluations VALUES(937,36,'cp -485','d6c6 e1g1 c6c4 f7h8 c4c2 b1c3 c2d1 f1d1 c8e6 d2d3 e8d7 f2f4 f8c5 g1h1 a8h8 f4e5 f6d5 c1g5 h8e8 a1c1 b7b6 d1f1 d5b4 c3e4 c5f8 g5e3 d4f5 e3f2 f8e7 g2g4 f5h6 h2h3 b4a2 c1a1 a7a5 d3d4 a2b4 f2e3 b4d5 e3h6 g7h6');
INSERT INTO evaluations VALUES(936,30,'cp -498',NULL);
INSERT INTO evaluations VALUES(936,31,'cp -480',NULL);
INSERT INTO evaluations VALUES(936,32,'cp -487',NULL);
INSERT INTO evaluations VALUES(936,33,'cp -480',NULL);
INSERT INTO evaluations VALUES(936,34,'cp -497',NULL);
INSERT INTO evaluations VALUES(936,35,'cp -495',NULL);
INSERT INTO evaluations VALUES(936,36,'cp -489','e1g1 c6c4 f7h8 c4c2 b1c3 c2d1 f1d1 c8e6 d2d3 e8d7 f2f4 f8c5 g1h1 a8h8 c3e4 c5b6 f4e5 f6g4 h2h3 d4c2 a1b1 g4e3 c1e3 c2e3 d1d2 e6a2 b1e1 a2d5 e4c3 d5e6 d3d4 e3d5 e1f1 a7a6 c3e4 h8d8 e4g5 h7h6 g5e6 d7e6 d2d3 d5e7 d3g3 b6d4 g3g7');
INSERT INTO evaluations VALUES(4920,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(4920,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(4920,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(4920,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(4920,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(4920,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(4920,36,'cp 77','a4b3 c5b6 a2a4 b5a4 b3a4 g8e7 e1g1 e8g8 b1a3 d7d5 e4d5 d8d5 a4b3 d5d6 a3c4 d6f6 f1e1 h7h6 b3c2 f8d8 d1e2 e7g6 c2g6 f6g6 f3e5 c6e5 c4e5 g6f6 d2d4 c7c5 d4c5 b6c5 e2f3 f6f3 e5f3 c8g4 c1e3 g4f3 e3c5 f3c6 c5e3 c6b5 b2b3');
INSERT INTO evaluations VALUES(31436,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31436,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31436,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(31436,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(31436,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(31436,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(31436,36,'cp 21','f8d6 f4d6 d8d6 e2e3 g8f6 f1d3 f5d3 d1d3 e8g8 e1g1 b8d7 h2h3 f8e8 f1d1 h7h6 b2b3 d6e7 a1c1 d7f8 f3e5 f8d7 e5d7 f6d7 c3a4 a8d8 d3c3 d5c4 b3c4 c6c5 d4d5 b7b6 c3b3 e6d5 c4d5 d7f6 a4c3 e7e5 b3c2 f6e4 c2e4');
INSERT INTO evaluations VALUES(28092,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(28092,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(28092,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(28092,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(28092,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(28092,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(28092,36,'cp 15','f4d6 d8d6 d1b3 d6e7 e2e3 g8f6 f3e5 e8g8 f1e2 h7h6 e1g1 f8c8 f2f3 c8c7 c4d5 e6d5 e3e4 f5e6 b3d1 c6c5 f3f4 c5d4 c3d5 e6d5 e4d5 e7b4 a1c1 c7c1 d1c1 b8d7 c1c7 d4d3 e5d3 b4d2 f1f2 a8e8 d3e5 d7e5 f4e5 d2d4');
INSERT INTO evaluations VALUES(36683,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(36683,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(36683,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(36683,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(36683,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(36683,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(36683,36,'cp 116','e1c1 c7c6 f4h6 b7b5 f2f3 d8a5 h2h4 b5b4 c3b1 c8a6 f1a6 b8a6 h6g7 g8g7 h4h5 g6g5 h5h6 g7h8 g1h3 f8g8 a2a3 g5g4 h3f4 a8b8 b2b3 a6c7 a3b4 a5b4 d2b4 b8b4 b1c3 g8a8 e4e5 d6e5 f4d3 b4d4 d3e5 d4d1 c3d1');
INSERT INTO evaluations VALUES(36682,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(36682,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(36682,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(36682,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(36682,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(36682,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(36682,36,'cp 88','c7c6 f4h6 b7b5 f2f3 d8a5 h2h4 b5b4 c3b1 c8a6 f1a6 b8a6 h4h5 g7h6 d2h6 a6c7 g1e2 c7e6 h5g6 f7g6 b1d2 f8f7 c1b1 a5b5 d1e1 b5g5 h6g5 e6g5 e1c1 d6d5 e4e5 f6h5 c2c3 b4c3 c1c3 a8c8 g2g3 g5e6 h1c1 h5g7 c3c6 c8c6 c1c6 h7h5 f3f4');
INSERT INTO evaluations VALUES(6589,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(6589,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(6589,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(6589,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(6589,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(6589,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(6589,36,'cp 104','e1f2 c6e5 d2d4 e5g6 f1e2 d7d6 h1f1 g8e7 f2g1 e8g8 d1e1 d6d5 e4d5 e7d5 c3d5 d8d5 e1f2 f8e8 c2c4 d5f5 e2f3 f5f6 c1d2 c7c6 a1e1 e8e1 f1e1 c8d7 d2c3 f6g5 d4d5 a8e8 e1e8 d7e8 f2a7 g6h4 g1f2 h4f3 a7b8 g5h4 f2f3 h4h5 f3f2 f7f6');
INSERT INTO evaluations VALUES(11932,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11932,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(11932,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11932,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11932,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(11932,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11932,36,'cp 0','f8b4 e1g1 d8e7 d2d4 b4c3 b2c3 d7d6 f1e1 e8g8 c1a3 f6d7 c3c4 e5d4 f3d4 c6c5 d4f5 e7f6 f5e3 a8b8 d1d3 d7e5 e3d5 f6h4 d3c3 e5g4 c3g3 h4g3 h2g3 g4e5 a3c1 c8e6 d5c7 e5c4 e1d1 b8b7 c7d5 e6d5 d1d5');
INSERT INTO evaluations VALUES(7609,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7609,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7609,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(7609,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(7609,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(7609,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7609,36,'cp 13','e1g1 b4c3 b2c3 d7d6 d2d4 d8e7 d4e5 d6e5 a2a4 e8g8 d1e2 a7a5 c1a3 c6c5 f3d2 e7e6 a3c5 c8a6 c3c4 f8b8 f1d1 f6d7 e2e3 h7h6 f2f3 e6c6 c5e7 f7f6 e3c3 b8e8 e7a3 d7b6 c3a5 a6c4 a5c3 a8a4 a3b4 a4a1 d1a1');
INSERT INTO evaluations VALUES(37832,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(37832,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(37832,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(37832,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(37832,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(37832,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(37832,36,'cp -20','a2a4 b8c6 g1f3 a7a6 d2d3 d7d6 f1e2 e7e6 e1g1 g8e7 d1e1 e8g8 c3d1 c8d7 c2c3 b7b5 d1f2 b5b4 e4e5 c6a5 e2d1 d7c6 f2e4 d6e5 f4e5 c6e4 e1e4 a5c6 c1g5 d8c7 d3d4 b4c3 b2c3 c5d4 c3d4 a8b8');
INSERT INTO evaluations VALUES(831,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(831,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(831,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(831,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(831,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(831,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(831,36,'cp 93','d4c5 f8c5 b1d2 b6d8 d2b3 c5b6 e2d3 e8g8 a2a4 h7h6 a4a5 b6c7 f1e1 f7f6 d3f5 e6f5 e5f6 f8f6 c1e3 f5f4 e3d4 f6f7 d4b6 a7b6 a5b6 a8a1 b6c7 f7c7 d1a1 c7e7 e1e7 d8e7 a1a8 e7c7 h2h3 c7b8 a8a3 c8f5 a3c5 f5e4 b3d4 c6d4 f3d4');
INSERT INTO evaluations VALUES(828,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(828,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(828,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(828,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(828,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(828,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(828,36,'cp 91','f8c5 b1d2 b6d8 b2b4 c5e7 d2b3 e8g8 f1e1 a7a5 b4b5 a5a4 b3d4 f5d4 c3d4 c6a5 e2d3 a5c4 f3d2 c4d2 c1d2 c8d7 a1c1 d8b6 d1g4 f7f5 e5f6 e7f6 d2e3 b6d6 g4h5 g7g6 h5d1 d6b4 c1b1 b4d6 h2h3 d6b6 d1g4 f8f7 h3h4 e6e5 d4e5');
INSERT INTO evaluations VALUES(475,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(475,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(475,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(475,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(475,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(475,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(475,36,'cp 124','b1d2 b6d8 b2b4 c5e7 d2b3 e8g8 f1e1 a7a5 b4b5 a5a4 b3d4 c6d4 c3d4 c8d7 e2d3 e7b4 c1d2 b4d2 d1d2 f5h4 f3h4 d8h4 g2g3 h4h6 d2h6 g7h6 a1c1 f8c8 g1g2 f7f6 e5f6 g8f7 b5b6 f7f6 c1c7 c8c7 b6c7 f6e7 d3h7 d7c6');
INSERT INTO evaluations VALUES(9422,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(9422,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(9422,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(9422,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(9422,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(9422,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(9422,36,'cp 50','c5b6 h2h3 e8g8 f1e1 f6e4 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 c8f5 e4e1 b6a5 f3h4 f5g6 d1b3 d8d6 c1g5 e7f5 b3b5 a5c3 b2c3 f5h4 g5h4 c7c6 b5b7 d6d5 e1e7 d5c4 h4g3 c4c3 a1e1 c3c5 b7c7 a8d8 c7a7 c5a7');
INSERT INTO evaluations VALUES(9423,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(9423,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(9423,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(9423,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(9423,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(9423,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(9423,36,'cp 78','c3d4 c5b6 h2h3 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 c8f5 e4e1 f8e8 a4a5 b6c5 c1f4 h7h6 d1a4 a8c8 a4c4 b7b6 f3d4 d8d7 a1d1 f5g6 b2b4 c5d4 d1d4 e7f5 e1e8 c8e8 d4d1 f5d6 c4c6 e8d8 f2f3 f7f6 g1f2 b6a5 b4a5 g8h8 c6d7 d8d7');
INSERT INTO evaluations VALUES(15457,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15457,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(15457,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(15457,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(15457,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(15457,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(15457,36,'cp 11','d6e5 d1d8 c6d8 f3e5 f8d6 e5d3 f6e4 f2f3 e4c5 c1e3 c5d3 c4d3 d8c6 b1c3 c8e6 c3e4 d6e7 e1c1 c6b4 a2a3 b4d3 d1d3 e6d7 e3d4 h8g8 d3c3 f7f5 e4g3 e8c8 d4e5 c7c6 g3h5 g7g6 h5f4 g8e8 h1e1 b7b6 f4d3 a7a5 b2b4 a5a4 c1b1 c8b7 d3b2');
INSERT INTO evaluations VALUES(15208,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(15208,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(15208,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(15208,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(15208,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(15208,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(15208,36,'cp 12','d1d8 c6d8 f3e5 f8d6 e5d3 f6e4 f2f3 e4c5 c1e3 c5d3 c4d3 d8c6 b1c3 c8e6 c3e4 d6e7 e1c1 c6b4 a2a3 b4d3 d1d3 e6d7 e3d4 h8g8 d4c5 e7c5 e4c5 e8c8 h1e1 d8e8 d3e3 e8e3 e1e3 g8e8 c1d2 e8e3 d2e3 d7e8 e3d4 b7b6 c5d3 c8d7 c2c4 f7f6 b2b3 d7e6 d3f4 e6f5');
INSERT INTO evaluations VALUES(2250,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(2250,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(2250,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(2250,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(2250,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(2250,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(2250,36,'cp 4','c6d8 f3e5 f8d6 e5d3 f6e4 f2f3 e4c5 c1e3 c5d3 c4d3 d8c6 b1c3 c8e6 c3e4 d6e7 e1c1 c6b4 a2a3 b4d3 d1d3 e6d7 e3c5 e7c5 e4c5 e8c8 h1d1 d7c6 d3e3 h8e8 d1d8 c8d8 c1d2 e8e3 d2e3 b7b6 c5d3 d8e7 c2c4 c6e8 b2b4 f7f6 e3d4 e7d6 c4c5 d6c6 c5b6 c7b6 d3f4');
INSERT INTO evaluations VALUES(2379,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2379,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(2379,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(2379,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(2379,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(2379,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(2379,36,'cp 4','f3e5 f8d6 e5f3 f6e4 e1g1 c8e6 f1e1 e6c4 b1d2 c4d5 d2e4 d6e7 f3d4 d8e6 e4c3 a8d8 c3d5 d8d5 d4e6 f7e6 c1f4 e8f7 e1e4 e7f6 c2c3 e6e5 f4e3 h8d8 a1e1 a7a6 e4c4 c7c6 g1f1 f7g6 h2h3 h7h5 b2b4 f6g5 c4e4 g5f6 a2a4 d5d1 f1e2');
INSERT INTO evaluations VALUES(2374,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2374,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(2374,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(2374,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(2374,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(2374,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2374,36,'cp 4','f8d6 e5d3 f6e4 f2f3 e4c5 c1e3 c5d3 c4d3 c8e6 b1c3 d8c6 c3e4 d6e7 e1c1 c6b4 a2a3 b4d3 d1d3 e6d7 e3c5 e7c5 e4c5 e8c8 d3d7 d8d7 c5d7 c8d7 h1d1 d7c6 d1e1 c6d6 c1d2 h7h5 e1e4 h5h4 d2e3 h8h5 e4d4 d6e6 e3f2 f7f5 d4b4 b7b6 b4b5 c7c5');
INSERT INTO evaluations VALUES(18982,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(18982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18982,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18982,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(18982,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(18982,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(18982,36,'cp -13','g7g6 b1c3 f8g7 a1b1 g8e7 a2a3 a7a5 g2g3 d7d6 f1g2 e8g8 e1g1 c8e6 c1e3 h7h6 f3d2 a8b8 d1e2 d8d7 f2f4 e5f4 g3f4 c6d4 e3d4 g7d4 g1h1 d4g7 e2e1 b7b6 e1g3 b8e8 b1e1 f7f5 e4f5 e6f5 g2e4');
INSERT INTO evaluations VALUES(19487,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(19487,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(19487,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(19487,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(19487,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(19487,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(19487,36,'cp 0','e2e4 e7e6 f1b5 d8c7 e1g1 a7a6 b5c6 c7c6 d2d3 b7b5 d1e1 g8f6 f4f5 d7d6 c1g5 f8e7 b1c3 e8g8 f5e6 c8e6 f3h4 a8e8 g5f6 e7f6 f1f6 g7f6 e1f2 c6c7 h4f5 g8h8 f5d6 c7d6 f2f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(21954,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(21954,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(21954,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(21954,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(21954,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(21954,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21954,36,'cp 0','f7f5 b1c3 g8f6 d2d4 f5e4 d4e5 e4f3 e5f6 d8f6 d1f3 f6f3 g2f3 c8e6 e1c1 e8c8 f3f4 d6d5 h1g1 a7a6 f4f5 e6f5 c3d5 c6b4 b2g7 f8g7 d5e7 c8b8 d1d8 h8d8 g1g7 f5c2 a2a3 d8d1 c1b2 d1b1 b2c3 b1b3 c3d2 b3a3 g7g8 b8a7 e7c8 a7b8');
INSERT INTO evaluations VALUES(29797,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(29797,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(29797,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(29797,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(29797,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(29797,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(29797,36,'cp 66','d1b3 d5c3 b2c3 c7c5 g1f3 b8d7 e2e3 c5d4 c3d4 e7e6 b3b7 a8b8 b7a6 f8b4 c1d2 b4d2 f3d2 e8g8 f1e2 e6e5 e1g1 e5d4 e3d4 d7b6 a6a7 b8a8 a7b7 d8d4 d2f3 d4f6 f1e1 f8b8 b7c7 b8c8 c7g3 c8e8 e2f1 e8e1 a1e1 h7h6 e1d1 b6d7 d1d6');
INSERT INTO evaluations VALUES(29795,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(29795,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(29795,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(29795,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(29795,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(29795,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(29795,36,'cp 73','d5c3 b2c3 b8d7 b3b7 c7c5 g1f3 a8b8 b7a7 d8c8 e2e4 b8a8 f1a6 a8a7 a6c8 f5e4 c8d7 a7d7 e1g1 c5d4 f3d4 f7f6 f1e1 e4d3 f2f4 e8f7 a2a4 g7g6 a4a5 f8h6 d4c6 h8a8 c1e3 d3c4 g2g4 c4b5 c6d4 b5d3 g1g2 d3c4');
INSERT INTO evaluations VALUES(35183,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(35183,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(35183,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(35183,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(35183,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(35183,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(35183,36,'cp 44','f8b4 c1d2 b4c3 d2c3 g8f6 d4e5 f6e4 d1f3 e4c3 f3c3 h4e4 g1e2 b8c6 f2f4 b7b6 e1c1 c8b7 h2h4 e8c8 h1h3 e4f5 h3e3 h8e8 g2g3 f7f6 e2d4 c6d4 c3d4 f6e5 f4e5 e8e7 b2b3 f5e6 c1b2 d8e8 f1e2 e6g6 d1e1 c8b8 e2g4 g6f7 g4h3 f7d5 d4d5 b7d5');
INSERT INTO evaluations VALUES(33301,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(33301,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(33301,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(33301,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(33301,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(33301,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(33301,36,'cp 58','c1d2 b4c3 d2c3 g8f6 d4e5 f6e4 d1f3 e4c3 f3c3 h4e4 g1e2 b8c6 f2f4 c6b4 e1f2 b4d5 c3f3 e4c4 a1d1 c4c5 e2d4 d7d6 a2a3 a7a5 e5d6 e8g8 d6c7 d5c7 f1e2 f8d8 c2c3 c8d7 h1e1 d7c6 f3h3 g7g6 h3g3 c7e6 g3e3 e6d4 d1d4 d8e8 e3d2 a8d8 e2f3 e8e1 f2e1 d8d4 d2d4 c5d4');
INSERT INTO evaluations VALUES(50668,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(50668,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(50668,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(50668,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(50668,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(50668,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(50668,36,'cp 100','d2d4 g8f6 d4c5 b8c6 b1d2 d6c5 e4e5 f6h5 d1a4 g7g6 d2b3 c8d7 a4e4 f8g7 f1e2 d8c7 b3c5 d7f5 e4a4 e8g8 e5e6 f7e6 e1g1 h5f6 c1f4 e6e5 a4b3 g8h8 f4g3 f6h5 c5e6 f5e6 b3e6 h5f4 e6c4 c7b6 f1b1 c6a5 c4e4 b6c6 e4e3 e5e4 f3g5 f4e2 e3e2');
INSERT INTO evaluations VALUES(50660,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(50660,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(50660,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(50660,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(50660,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(50660,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(50660,36,'cp 81','g8f6 d4c5 b8c6 b1d2 d6c5 e4e5 f6h5 d2c4 d8d1 e1d1 c8g4 f1e2 a8d8 d1e1 g7g6 h2h3 g4e6 c1e3 e6d5 h1g1 f7f6 g2g4 h5g7 e3c5 g7e6 c5b6 e6f4 b2b3 f4e2 e1e2 d5f3 e2f3 d8d3 f3g2 f6e5 a1c1 f8g7 g1d1 e5e4 c4d2 g7c3 d2e4 d3d1 c1d1 c3b4 e4g5 b4d6 b6e3 h7h5 g5e4 h5g4 h3g4 d6b4 g2g3 e8g8 f2f4');
INSERT INTO evaluations VALUES(52471,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52471,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52471,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52471,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52471,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52471,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(52471,36,'cp -12','g8f6 d2d4 e7e6 b1d2 d7d5 g1f3 b8c6 f1d3 b7b6 e1g1 c8b7 f1e1 f8e7 d1e2 e8g8 b2b3 a7a5 a2a4 a8c8 c1b2 d8c7 d2f1 f8d8 f1g3 e7d6 a1c1 c7e7 c1d1 h7h6 c3c4 c6b4 d3b1 d6g3 h2g3 c5d4 e3d4');
INSERT INTO evaluations VALUES(16041,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(16041,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(16041,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(16041,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(16041,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(16041,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(16041,36,'cp -74','d8h4 d1e2 c6d4 g2g3 d4e2 g3h4 e2d4 f7h8 d4c2 e1e2 e4f6 b1c3 c7c6 c4f7 e8d8 a1b1 d7d5 d2d3 c2d4 e2e3 d8c7 f2f4 e5f4 e3f2 f6g4 f2g1 g4e3 c1e3 f4e3 b1c1 c8h3 c3d5 c7d7 d5e3 f8d6 c1c4 d4f3 g1f2 f3e5 h1g1 a8h8 g1g7 h8f8 g7h7 e5c4');
INSERT INTO evaluations VALUES(1042,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(1042,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(1042,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(1042,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(1042,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(1042,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(1042,36,'cp -44','d1e2 c6d4 g2g3 d4e2 g3h4 e2d4 f7h8 d4c2 e1e2 e4f6 b1c3 c7c6 c4f7 e8d8 a1b1 d7d5 d2d3 c2d4 e2d2 d8c7 f2f4 e5f4 b2b4 f8d6 f7d5 f6d5 c3d5 c6d5 h8f7 c8h3 f7d6 c7d6 d2c3 d4e2 c3d2 f4f3 c1b2 d5d4 h1e1 e2c3 b2c3 d4c3 d2e3 h3g4 e3f2 a8c8 e1g1 h7h5 h2h3 g4h3 g1g7 h3g4 g7b7 c8e8 b1e1');
INSERT INTO evaluations VALUES(27164,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(27164,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(27164,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(27164,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(27164,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(27164,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(27164,36,'cp 27','b7c6 b1c3 c6c5 d4e2 f8e7 e1g1 g8f6 f1e1 e8g8 b2b3 f8e8 e2g3 a7a5 a2a4 e7f8 c1f4 d7c6 h2h3 g7g6 d1d2 d8b8 f2f3 f8g7 g3e2 f6d7 a1d1 c5c4 e2d4 g7d4 d2d4 c4b3 c2b3 b8b3 f4h6 f7f6 h6d2');
INSERT INTO evaluations VALUES(25811,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(25811,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(25811,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(25811,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(25811,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25811,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(25811,36,'cp 31','e1g1 c6c5 d4e2 a7a5 c1f4 g8f6 b1c3 f8e7 d1d3 e8g8 a2a4 a8b8 b2b3 b8b4 f2f3 f8e8 a1d1 h7h6 h2h3 d8a8 c3a2 b4b7 f4g3 d7e6 e2f4 c5c4 b3c4 f6d7 f4d5 d7c5 d3a3 a8a6 d5e7 e8e7 e4e5 a6c4 a2c3');
INSERT INTO evaluations VALUES(25810,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(25810,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(25810,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(25810,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(25810,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(25810,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(25810,36,'cp 31','c6c5 d4e2 a7a5 a2a4 g8f6 b1c3 f8e7 b2b3 e8g8 e2g3 f8e8 h2h3 d7c6 c1b2 f6d7 c3d5 e7h4 d1d3 c6d5 d3d5 d8e7 d5d3 h4g3 d3g3 f7f6 f1e1 e7f7 b2c3 c5c4 b3b4 a5b4 c3b4 f7g6 g1h2 g6g3 h2g3 d7c5 b4c5 d6c5 a1d1 a8a4');
INSERT INTO evaluations VALUES(40468,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(40468,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(40468,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(40468,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(40468,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(40468,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(40468,36,'cp 53','e2e3 h7h6 g5h4 e8g8 d1c2 b8d7 c4d5 e6d5 f1d3 f8e8 e1g1 b4e7 h4g3 a7a6 h2h3 c6c5 d4c5 d7c5 a1c1 c8e6 d3f5 a8c8 f5e6 c5e6 f1d1 e7b4 g3e5 b4c3 e5c3 c8c5 c2d3 f6e4 c3b4 c5c1 d1c1 e6g5 f3d4 d8b6 b4a3 g5e6 d4e6');
INSERT INTO evaluations VALUES(40467,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(40467,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(40467,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(40467,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(40467,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(40467,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(40467,36,'cp 54','e8g8 d1c2 b8d7 c4d5 e6d5 f1d3 h7h6 g5h4 f8e8 e1g1 b4e7 h4g3 a7a6 h2h3 c6c5 d4c5 d7c5 a1c1 c8e6 g3e5 a8c8 f1d1 e7d6 e5d6 d8d6 f3d4 e6d7 c2b1 b7b5 c3e2 b5b4 d3f5 c5e6 d4e6 f7e6 f5g6 e8f8 e2d4 c8c1');
INSERT INTO evaluations VALUES(36100,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(36100,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(36100,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(36100,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(36100,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(36100,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(36100,36,'cp 67','d1c2 b8d7 c4d5 e6d5 f1d3 h7h6 g5h4 b4e7 e1g1 f8e8 h4g3 a7a6 a1d1 e7b4 g3h4 b4e7 f1e1 f6e4 d3e4 e7h4 e4h7 g8h8 h7d3 d7f8 f3h4 d8h4 f2f3 h4g5 c2f2 h8g8 a2a3 c8d7 h2h3 d7f5 h3h4 g5f6 e3e4 f5h7 e4e5 f6e7 f3f4 h7d3');
INSERT INTO evaluations VALUES(53046,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(53046,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(53046,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(53046,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(53046,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(53046,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(53046,36,'cp 121','d4c5 d8a5 b1c3 a5c5 e2e4 g8f6 c1e3 c5a5 f3d2 b8c6 d2b3 a5c7 f2f4 c8d7 d1d2 e7e6 f1e2 d6d5 e4e5 f6e4 c3e4 d5e4 e1c1 c6e7 e3c5 e7d5 c5f8 h8f8 g2g3 e8c8 d2d4 d5b6 h1e1 d7c6 d4c5 f8e8 c5b4 b6d5 b4e4 d5c3 d1d8 e8d8 e4e3 c3a2 c1b1');
INSERT INTO evaluations VALUES(52993,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(52993,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(52993,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(52993,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(52993,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(52993,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(52993,36,'cp 117','d8a5 b1c3 a5c5 e2e4 g8f6 c1e3 c5a5 f3d2 b8c6 d2b3 a5c7 c3d5 c7d8 d5f6 e7f6 b3d4 f8e7 f1c4 c6d4 d1d4 d8a5 c2c3 c8e6 e1c1 e8g8 c4d5 f8c8 c1b1 f6f5 g2g3 e6d5 d4d5 a5d5 e4d5 b7b5 d1d4 g7g6 d4b4 a7a6 a2a4 b5a4 b4a4 a8b8');
INSERT INTO evaluations VALUES(2046,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(2046,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(2046,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(2046,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(2046,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(2046,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(2046,36,'cp -9','g8f6 h5d1 c6a5 b1c3 c7c6 g1e2 f6e4 c3e4 d7d5 c4b3 d5e4 d3e4 e7c7 e2g3 f8c5 d1e2 e8g8 c1e3 a5b3 a2b3 c5d4 e3d4 e5d4 e1g1 c6c5 e4e5 c8e6 f2f4 f7f5 e5f6 f8f6 e2f3 c7f7 g3e4 f6f5 e4g5 e6d5 g5f7');
INSERT INTO evaluations VALUES(1095,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(1095,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(1095,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(1095,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(1095,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1095,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(1095,36,'cp -1','h5d1 c6a5 b1c3 c7c6 g1e2 f6e4 c3e4 d7d5 c4b3 d5e4 d3e4 e7c7 e2g3 g7g6 e1g1 f8g7 d1f3 h7h6 f1d1 e8g8 c1e3 b7b6 h2h3 a5b3 a2b3 c8e6 b3b4 c7e7 e3d2 c6c5 b4c5 e7c5 d2c3 b6b5 a1a5 c5c6 d1a1 a7a6 g3f1 f7f5 f1e3 c6e4 f3e4 f5e4');
INSERT INTO evaluations VALUES(5728,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(5728,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5728,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5728,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5728,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5728,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(5728,36,'cp 0','c6a5 c4b3 g8e7 c1e3 c5e3 f2e3 e8g8 e1g1 e7g6 d3d4 c7c6 d4e5 a5b3 a2b3 g6e5 f3e5 d6e5 d1d8 f8d8 c3a4 c8e6 a4c5 b7b6 c5e6 f7e6 b3b4 a7a5 b4a5 a8a5 a1a5 b6a5 f1a1 d8d2 a1a5 d2c2 b2b4 c2e2 a5e5 e2e3 e5e6 e3b3 e6c6 b3b4 c6c8 g8f7 c8c7 f7f8 e4e5 b4e4 c7d7');
INSERT INTO evaluations VALUES(13911,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(13911,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(13911,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(13911,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(13911,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(13911,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(13911,36,'cp -48','h7h6 g5f4 a7a6 c4b3 g7g5 f4g3 f8g7 e1g1 e8g8 e4e5 f6h5 e5d6 h5g3 h2g3 d8d6 c3e4 d6e7 c2c3 b7b6 f1e1 f8d8 d1e2 a6a5 b3a4 c8d7 e4d2 c6d4 f3d4 c5d4 a4d7 d4c3 b2c3 e7d7 d3d4 b6b5 d2e4 b5b4 a1d1 d7c7 d1d3 c7c6 e2h5');
INSERT INTO evaluations VALUES(17021,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(17021,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(17021,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(17021,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(17021,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(17021,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(17021,36,'cp 32','d1d4 f8d6 c1g5 d8e7 b1c3 d6c5 d4d2 f6e4 c4f7 e7f7 c3e4 c5b6 a2a4 e8g8 a4a5 f7c4 f2f3 b6c5 e4c5 c4c5 g5e3 c5b5 d2d4 d7d6 c2c4 b5f5 c4c5 d6c5 d4c5 c8d7 c5f5 f8f5 e1f2 a7a6 e3d4 a8f8 h1e1 f8f7 d4c3 h7h6 a1d1 d7c6 e1e5 f5e5 c3e5 c6b5 g2g4');
INSERT INTO evaluations VALUES(37238,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(37238,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(37238,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(37238,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(37238,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(37238,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(37238,36,'cp 7','g1f3 d7d6 c2c3 c7c6 e1g1 d6d5 e4d5 c6d5 c4b3 e5e4 f3d4 d8e7 c1e3 b8c6 d3e4 d5e4 d4c6 b7c6 e3c5 e7c5 d1d4 c5g5 d4e3 g5e3 f2e3 a7a5 f1d1 a5a4 b3c4 c8d7 b1d2 f8e8 d2f1 d7e6 c4e6 e8e6 f1d2 f6d7 d2c4 d7c5 d1d4');
INSERT INTO evaluations VALUES(37232,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(37232,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37232,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(37232,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(37232,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(37232,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37232,36,'cp 15','d7d6 e1g1 c7c6 f1e1 d8b6 e1e2 f8e8 c4b3 b6d8 c2c3 c8e6 b3e6 e8e6 b1d2 b8d7 d2c4 b7b5 c4e3 c5b6 a2a4 b5a4 d1a4 d6d5 f3g5 e6e8 a4c6 h7h6 g5f3 d5e4 d3e4 a8c8 c6b5 f6e4 e3f5 b6f2 e2f2 e4f2');
INSERT INTO evaluations VALUES(36996,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(36996,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(36996,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36996,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(36996,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(36996,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(36996,36,'cp 7','e1g1 c7c6 c2c3 d6d5 e4d5 c6d5 c4b3 e5e4 f3d4 d8e7 c1e3 c8d7 d3e4 d5e4 b1d2 b8c6 f1e1 h7h6 f2f3 e4f3 e3f2 e7d8 d2f3 d8c7 d1d3 c5b6 b3c2 c6d4 f3d4 c7f4 e1e7 f4d6 a1e1 a8e8 e7e8 f8e8 e1e8 d7e8 d3g3 d6d8 g3f3 d8c7 c2b3 e8c6 d4c6 b6f2 f3f2 b7c6');
INSERT INTO evaluations VALUES(36995,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(36995,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(36995,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(36995,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(36995,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36995,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36995,36,'cp 7','c7c6 c2c3 d6d5 e4d5 c6d5 c4b3 e5e4 f3d4 b8c6 c1e3 d8e7 d3e4 d5e4 d4c6 b7c6 e3c5 e7c5 d1d4 c5d4 c3d4 f8d8 f1d1 d8d6 b1c3 c8e6 b3e6 d6e6 d4d5 c6d5 c3d5 f6d5 d1d5 e4e3 f2e3 e6e3 a1d1 h7h6 d1d2 a8e8 d5d7 a7a5 d7a7 e3e2 d2e2 e8e2 a7a5 e2b2 a5a7 g8h7');
INSERT INTO evaluations VALUES(43465,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(43465,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(43465,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43465,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(43465,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43465,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(43465,36,'cp 1','f8e7 f1c4 e8g8 c4b3 b8c6 d1e2 c6a5 e3g5 c8e6 e1g1 a5b3 a2b3 f6g4 g5e7 d8e7 f1d1 b7b5 c3d5 e6d5 d1d5 f8d8 d5d3 g4f6 h2h3 e7b7 a1d1 b7e4 e2e4 f6e4 f3e1 f7f6 f2f3 e4g5 d3d6 g5f7 d6d7 a6a5 d7d8 a8d8 e1d3 d8d4');
INSERT INTO evaluations VALUES(39679,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(39679,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(39679,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(39679,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(39679,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(39679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39679,36,'cp 5','f1c4 e8g8 e1g1 b8c6 c4b3 h7h6 f3h4 c6a5 h4g6 a5b3 e3b6 d8d7 g6e7 d7e7 a2b3 c8e6 f2f4 e5f4 f1f4 a8c8 d1d2 f8e8 a1d1 c8c6 b6a5 f6d7 a5b4 d7c5 b4c5 d6c5 c3d5 e6d5 e4d5 c6d6 d2f2 e7c7 c2c4 a6a5');
INSERT INTO evaluations VALUES(39668,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(39668,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(39668,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(39668,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(39668,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(39668,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(39668,36,'cp 9','e8g8 e1g1 b8c6 d1e2 c6a5 c4d3 c8e6 c3a4 b7b5 e3b6 d8c8 b6a5 b5a4 b2b3 e7d8 a5b4 c8b8 b4d2 a6a5 b3a4 d8b6 d2g5 f6d7 d3b5 h7h6 g5e3 b6e3 e2e3 f8c8 c2c4 d7b6 f3d2 b6c4 d2c4 e6c4 f1c1 c4e6 c1c8 b8c8 e3d3 c8d8 a1d1 e6a2 d3d6 d8d6');
INSERT INTO evaluations VALUES(46142,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(46142,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(46142,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(46142,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(46142,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(46142,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(46142,36,'cp 36','c4d5 c5d4 d1d4 f8e7 e2e4 e6d5 g5f6 e7f6 e4e5 f6e7 f1b5 b8c6 d4d5 e8g8 g1f3 a7a6 d5d8 f8d8 b5c6 b7c6 a1d1 c8g4 d1d8 a8d8 e1e2 d8b8 b2b3 e7b4 h1c1 b8e8 c3d1 e8e5 d1e3 g4f3 e2f3 e5e6 c1c4 a6a5 c4e4 e6f6 f3e2 g8f8 h2h3 b4c5 e4c4 c5b6 f2f3');
INSERT INTO evaluations VALUES(46125,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(46125,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(46125,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(46125,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(46125,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(46125,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46125,36,'cp 44','c5d4 d1d4 f8e7 e2e4 e6d5 f1b5 b8c6 g5f6 e7f6 e4e5 f6e7 d4d5 e8g8 g1f3 a7a6 d5d8 f8d8 b5c6 b7c6 a1d1 c8e6 f3d4 e6d7 e1g1 g8f8 h2h3 c6c5 d4c2 d7e6 c2e3 d8d4 c3d5 a8d8 d5e7 f8e7 d1c1 d4d2 c1c5 d2b2 a2a3 e7f8 c5a5 b2b5 a5b5 a6b5 f1b1 d8d3 b1b5');
INSERT INTO evaluations VALUES(49881,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(49881,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(49881,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(49881,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(49881,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(49881,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(49881,36,'cp -94','d7d5 c4d3 d5e4 d3e4 f7f5 e4f3 f8d6 d2d3 e8g8 g1e2 b8c6 e1g1 c8d7 c1d2 g8h8 c2c4 d8f6 d2c3 a8d8 d1d2 b7b6 b2b4 c6d4 e2d4 e5d4 c3b2 c7c5 b4c5 b6c5 f3d5 f5f4 a1e1 f6g6 f2f3 d7f5 d5e4 f5e4 e1e4 f8e8 f1e1 e8e4');
INSERT INTO evaluations VALUES(55025,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(55025,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(55025,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(55025,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(55025,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(55025,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(55025,36,'cp -119','g8f6 h5d1 e7e5 d2d4 b8d7 g1f3 f8d6 c2c4 e5e4 f3d2 c7c6 b1c3 e8g8 a2a4 f8e8 d1b3 d5c4 f1c4 d8e7 a4a5 a8b8 c4e2 d7f8 e1g1 c8f5 f2f4 e4f3 d2f3 b7b5 a5b6 a7b6 f3h4 f5c8 e3e4 f6e4 c3e4 e7e4 b3f7 g8h8 e2c4 c8e6 c4e6 e4d4 g1h1 e8e6');
INSERT INTO evaluations VALUES(47125,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(47125,31,'cp -102',NULL);
INSERT INTO evaluations VALUES(47125,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(47125,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(47125,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(47125,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(47125,36,'cp -121','h5d1 e7e5 d2d4 b8d7 g1f3 e5e4 f3d2 f8d6 c2c4 c7c6 a2a4 e8g8 b1c3 f8e8 c4d5 c6d5 d1b3 a7a5 c3d5 f6d5 b3d5 d7b6 d5h5 d6b4 b2b3 g7g6 h5d1 b6d5 c1a3 c8e6 f1c4 d5c3 d1c2 d8g5 e1f1 a8c8 h2h4 g5f5 a3b4 a5b4 c4e6 e8e6 d2c4 h7h5');
INSERT INTO evaluations VALUES(55036,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(55036,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(55036,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(55036,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(55036,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(55036,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(55036,36,'cp -72','d5d4 g1f3 b8c6 c1a3 g8f6 d2d3 a7a5 g2g3 g7g6 f1g2 f8g7 e1g1 e8g8 a3b2 e7e5 b1a3 f8e8 a3c2 f6d7 d1d2 d7c5 b2a3 c5a6 e2e4 d4e3 c2e3 a6b4 a1d1 g7h6 a3b2 c6d4 a2a3 b4c6 b3b4 d4f3 g2f3');
INSERT INTO evaluations VALUES(6366,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(6366,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(6366,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(6366,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(6366,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(6366,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(6366,36,'cp 183','c2c3 b4f8 f1b5 a7a6 b5a4 d7d6 c3c4 c8d7 b1c3 c6e7 a4d7 d8d7 d1b3 d7c8 h2h4 g8f6 h4h5 e7c6 d4d5 c6e5 f3e5 d6e5 h5h6 g7g6 f2f3 e6d5 c4d5 f8c5 c1d2 f6d7 e1c1 e8g8 c1b1 c5d6 d2e3 b7b5 d1c1 c8b7 h1d1');
INSERT INTO evaluations VALUES(6365,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(6365,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(6365,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(6365,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(6365,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(6365,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(6365,36,'cp 169','b4e7 e4e5 d7d6 f1d3 f7f5 d1e2 d6e5 d4e5 c8d7 d3c2 a7a6 e1g1 g8h6 c1h6 g7h6 f1d1 e8g8 e2d2 d7c8 d2h6 d8e8 h2h4 e8g6 h6g6 h7g6 b1d2 b7b6 a2a4 c8b7 b2b4 g8f7 a1c1 a8c8 c2b3 f8d8 b3c4 a6a5 b4b5');
INSERT INTO evaluations VALUES(6523,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(6523,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(6523,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(6523,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(6523,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(6523,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(6523,36,'cp 37','f1b5 d7d6 d2d4 e5d4 e3d4 d8f6 d1d2 c8d7 g1e2 c5b6 c2c4 f6h6 e1g1 h6d2 b1d2 g8f6 b5c6 b7c6 f1c1 b6a5 b2c3 a5c3 e2c3 e8c8 c1e1 d8e8 f2f3 h7h6 g1f2 g7g5 e1e8 h8e8 a1e1 e8e1 f2e1 f6h5 g2g3 h5g7 e1f2 c8d8 f3f4 g5f4 g3f4');
INSERT INTO evaluations VALUES(6396,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(6396,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(6396,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(6396,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(6396,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(6396,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(6396,36,'cp 35','d7d6 d2d4 e5d4 e3d4 d8f6 d1d2 c8d7 g1e2 c5b6 c2c4 f6h6 e1g1 h6d2 b1d2 g8f6 b5c6 b7c6 a1c1 b6a5 b2c3 a5c3 e2c3 e8c8 f1e1 h7h6 f2f3 h8e8 g1f2 g7g5 e1e8 d8e8 c1e1 e8e1 f2e1 f6h5 e1f2 h5f4 g2g3 f4d3 f2e2 d3b4 d2e4 f7f5 e4f6 b4c2 f6d7');
INSERT INTO evaluations VALUES(5952,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(5952,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(5952,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(5952,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(5952,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(5952,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(5952,36,'cp 45','d2d4 e5d4 e3d4 d8f6 c2c4 c5b6 d1d2 a7a6 b5a4 f6g6 g1e2 g8f6 e2f4 g6g5 a4c6 b7c6 e1g1 e8g8 f4d5 g5h4 d5b6 c7b6 d4d5 c6d5 g2g3 h4h5 b2f6 g7f6 d2d5 h5d5 c4d5 f8e8 b1c3 b6b5 f2f3 c8f5 g1f2 b5b4 c3e2 f5d3 f1e1 e8e5 a1d1 e5d5');
INSERT INTO evaluations VALUES(10151,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(10151,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(10151,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(10151,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(10151,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(10151,35,'cp -130',NULL);
INSERT INTO evaluations VALUES(10151,36,'cp -145','e5d4 c3a4 c5b4 c2c3 d4c3 b2c3 b4d6 e1g1 e8g8 f1e1 f8e8 d1b3 e8e7 c4d3 d8e8 b3c2 b7b5 d3b5 f6e4 b5d3 e4f6 c1g5 c6e5 d3h7 g8f8 h7e4 e5f3 g2f3 c8b7 g5f6 g7f6 g1f1 b7e4 e1e4 e7e4 c2e4 e8e4 f3e4 a8e8 f2f3 d6h2 a1b1 h2e5 f1e2 d7d6 b1b7');
INSERT INTO evaluations VALUES(10291,30,'cp -115',NULL);
INSERT INTO evaluations VALUES(10291,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(10291,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(10291,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(10291,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(10291,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(10291,36,'cp -130','c3a4 c5b4 c2c3 d4c3 b2c3 b4d6 e1g1 e8g8 f1e1 f8e8 c1g5 h7h6 g5h4 g7g5 h4g3 d6g3 h2g3 d7d6 d1c2 c6e5 f3e5 d6e5 a4b2 d8e7 a2a4 c8d7 b2d1 a8d8 f2f3 b7b6 c2f2 a7a5 c4b5 d7b5 a4b5 d8d3 c3c4 e8d8 d1e3');
INSERT INTO evaluations VALUES(29664,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(29664,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(29664,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(29664,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(29664,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(29664,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(29664,36,'cp 0','g1e2 b8c6 f2f3 g4d7 c2c4 d5c4 d3c4 c6a5 c4d3 e7e6 a2a3 c7c5 d4c5 a5c6 d1c2 c6e5 b2b4 b7b6 d3e4 a8c8 c1b2 d8c7 f3f4 e5g4 c5c6 d7c6 e4c6 c7c6 c2c6 c8c6 e2d4 c6d6 e1e2 f6d5 h1c1 g4e3 g2g3 f8e7 d4c6 d6d7 c6b8 d7d8 b8c6');
INSERT INTO evaluations VALUES(41282,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41282,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(41282,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(41282,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(41282,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(41282,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(41282,36,'cp 40','f1g2 e8g8 e1g1 c8f5 b1c3 f6e4 c3d5 c7c6 d5e3 f5e6 d4d5 e6d7 f3d4 e4f6 b2b3 d8b6 c1b2 a7a5 a1b1 b8a6 d4f3 f8e8 a2a3 a6c5 b2d4 b6c7 f3g5 h7h6 b3b4 a5b4 a3b4 c5a4 g5e4 f6e4 g2e4 a4c3 d4c3 g7c3 e4g2 c6c5 b4c5');
INSERT INTO evaluations VALUES(3374,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(3374,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(3374,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(3374,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3374,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3374,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(3374,36,'cp 12','a2a3 c8e6 c3d5 a7a5 e1g1 e6d5 e4d5 c6b8 f1e1 c7c6 d5c6 b7c6 c4a2 b8d7 c2c3 f8e8 d1b3 d6d5 f3e5 e7a3 c1f4 d7e5 f4e5 a3d6 e5d4 e8e1 a1e1 a5a4 b3b7 d8c8 b7c8 a8c8 a2b1 c8b8 d4f6 g7f6 b1c2 c6c5 c2a4 b8b2 a4c6');
INSERT INTO evaluations VALUES(3393,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(3393,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3393,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(3393,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(3393,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(3393,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(3393,36,'cp 11','c8e6 c3d5 a7a5 d5f6 e7f6 c4e6 f7e6 e1g1 d8e8 c1d2 h7h6 b2b3 f6e7 c2c3 e8h5 b3b4 f8f7 f3h2 h5d1 f1d1 d6d5 g1f1 e7d6 f1e2 b7b6 h2f3 g7g5 d1c1 c6e7 a3a4 e7g6 b4a5 a8a5 g2g3 a5a8 c3c4');
INSERT INTO evaluations VALUES(23699,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(23699,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23699,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(23699,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(23699,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(23699,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(23699,36,'cp 23','c3d5 a7a5 e1g1 e6d5 e4d5 c6b8 c4a2 b8d7 f1e1 f8e8 c2c4 a5a4 a2b1 d7c5 b1c2 g7g6 a1b1 e7f8 f3g5 f8e7 c1d2 f6d7 g5e4 c5e4 d3e4 d7c5 d2e3 d8d7 d1g4 d7g4 h3g4 f7f6 g1f1 e8b8 e3c5 d6c5 f1e2 b7b6 e1h1 e7d6 e2f3');
INSERT INTO evaluations VALUES(5269,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(5269,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(5269,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(5269,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(5269,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5269,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(5269,36,'cp 17','g8f6 f1e1 e8g8 d2d4 d7d5 e4e5 f6e4 c1e3 c5d4 c3d4 d8b6 d1e2 c8d7 a2a4 a7a6 b5c6 b6c6 f3d2 a8c8 d2e4 d5e4 b1c3 d7e6 e2c2 f7f6 e5f6 e7f6 c2e4 c6e4 c3e4 c8c2 e4c5 e6d5 e1c1 c2c1 a1c1 f8c8 a4a5 g8f7 c1c3 f6f5 f2f3 c8e8 h2h3');
INSERT INTO evaluations VALUES(7999,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(7999,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(7999,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(7999,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(7999,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(7999,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(7999,36,'cp 28','f1e1 e8g8 d2d4 d7d5 e4e5 f6e4 c1e3 c5d4 c3d4 d8b6 d1e2 c8d7 b5a4 a8c8 b1c3 e4c3 b2c3 b6d8 a4b3 c6a5 a1c1 a5c4 e3f4 a7a5 f3d2 b7b5 h2h4 h7h5 f4g5 d7f5 a2a4 f7f6 e5f6 e7f6 g5f4 f5g4 f2f3 b5a4 b3a4');
INSERT INTO evaluations VALUES(8000,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8000,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8000,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(8000,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8000,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(8000,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(8000,36,'cp 22','e8g8 d2d4 d7d5 e4e5 f6e4 c1e3 d8b6 d1e2 c5d4 c3d4 a7a6 b5c6 b7c6 h2h3 a6a5 e1c1 a5a4 b1c3 c8e6 f3e1 e4c3 b2c3 e6f5 e2d2 h7h5 f2f3 f7f6 e1d3 f5d3 d2d3 g8h7 a1b1 b6a7 b1b4 f6e5 d4e5 a7c7 e3d4 c6c5');
INSERT INTO evaluations VALUES(3042,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(3042,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(3042,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(3042,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(3042,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(3042,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(3042,36,'cp 32','d2d4 d7d5 e4e5 f6e4 c1e3 d8b6 d1e2 c5d4 c3d4 a7a6 b5c6 b7c6 b1c3 e4c3 b2c3 a6a5 e3g5 f7f6 g5f4 b6a6 e2d2 c8g4 e5f6 e7f6 h2h3 g4f5 f3h4 f5d7 f4h6 a8e8 a1c1 e8e1 c1e1 f8e8 h6g7 g8g7 g2g4 a6c8 e1e3 g6g5 e3e8 c8e8 h4f3 a5a4 g1g2');
INSERT INTO evaluations VALUES(34214,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(34214,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(34214,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(34214,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(34214,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(34214,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(34214,36,'cp 140','f8b4 f1c4 g8e7 e1g1 d7d6 c3e2 e8g8 c4b3 b4c5 c1e3 d6d5 e2g3 c5b6 f1e1 b8d7 e4d5 d7e5 g3h5 f6h4 h5f4 e5g4 d4f3 g4e3 f3h4 e3d1 a1d1 e7d5 f4d5 c6d5 b3d5 c8g4 h4f3 a8d8 d5b7 d8d1 e1d1 h7h6 b7d5 f8d8 g1f1 g4e6');
INSERT INTO evaluations VALUES(32998,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(32998,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(32998,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(32998,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(32998,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(32998,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(32998,36,'cp 151','f1c4 g8e7 e1g1 d7d6 c4b3 e8g8 c3e2 b4c5 c1e3 b8d7 c2c3 c5b6 e2g3 d7c5 f2f4 c5b3 a2b3 e7g6 d1d2 h7h6 h2h3 f8e8 d2f2 f6h4 b3b4 c8d7 g1h2 g8h8 a1e1 a8d8 f2d2 d7g4 b2b3 g4c8 d2f2 f7f6 d4f5 c8f5 e3b6 a7b6');
INSERT INTO evaluations VALUES(43826,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43826,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(43826,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(43826,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43826,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43826,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43826,36,'cp 0','d7d5 f1d3 c8b7 e1g1 f8d6 b2b3 e8g8 c4d5 e6d5 c1a3 d6a3 b1a3 c7c5 a1c1 d8e7 a3b5 b8c6 f1e1 f8d8 b5c3 a8c8 d3f5 c8a8');
INSERT INTO evaluations VALUES(46962,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46962,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46962,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46962,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(46962,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46962,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(46962,36,'cp 5','c7c6 e2e4 b8d7 g1f3 g7g6 f1e2 f8g7 e1g1 d7c5 d5c6 c5e4 c3e4 f6e4 c6b7 c8b7 e2b5 e8f8 c2c3 h7h5 c1e3 g7f6 f1e1 f8g7 a2a4 d8c7 a4a5 e4c5 b5e2 a7a6 h2h3 a8c8 e2f1 c5d7 f3g5 d6d5 d1d2 e7e5 b2b4');
INSERT INTO evaluations VALUES(45085,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45085,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45085,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(45085,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(45085,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(45085,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45085,36,'cp 21','e2e4 b8d7 g1f3 g7g6 f1e2 f8g7 e1g1 d7c5 d5c6 c5e4 c3e4 f6e4 c6b7 c8b7 e2b5 e8f8 c2c3 h7h5 c1e3 g7f6 f1e1 f8g7 a2a4 a7a6 b5f1 d6d5 f3d2 e4d2 d1d2 e7e5 e1d1 a6a5 e3c5 d8c7 d2e3 h8c8 c5b6 d5d4 b6c7 d4e3 c7b6 e3f2 b6f2 c8d8 f2b6 d8d1 a1d1');
INSERT INTO evaluations VALUES(45084,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45084,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(45084,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45084,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(45084,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45084,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45084,36,'cp 0','b8d7 g1f3 g7g6 f1e2 f8g7 e1g1 d7c5 d5c6 c5e4 c6b7 c8b7 c3e4 f6e4 e2b5 e8f8 c2c3 h7h5 a2a4 g7f6 c1e3 a7a6 b5d3 a8b8 e3a7 b8a8');
INSERT INTO evaluations VALUES(48548,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(48548,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(48548,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(48548,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48548,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48548,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48548,36,'cp 0','c7c6 b5d3 f8d6 d1e2 d6e7 c2c3 e8g8 e1g1 c8g4 c1f4 f8e8 e2d1 e7d6 f4d6 g4f3 d1f3 d8d6 b1d2 b8d7 a1e1 c6c5 d4c5 d7c5 d3c2 a7a5 f3d1 d6b6 d1c1 a5a4 e1e8 a8e8 h2h3 e8e2 c1d1 e2e8');
INSERT INTO evaluations VALUES(45721,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45721,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45721,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45721,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45721,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45721,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45721,36,'cp 0','b5d3 f8d6 d1e2 d6e7 e1g1 e8g8 c2c3 c8g4 c1f4 f8e8 e2d1 e7d6 f4d6 g4f3 d1f3 d8d6 b1d2 g7g6 a1e1 b8d7 f3d1 c6c5 d4c5 d7c5 d2b3 c5e6 e1e3 e6c5');
INSERT INTO evaluations VALUES(45777,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45777,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45777,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45777,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45777,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45777,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45777,36,'cp 0','f8d6 d1e2 d6e7 e1g1 e8g8 c2c3 c8g4 c1f4 f8e8 e2d1 e7d6 f4d6 d8d6 b1d2 g7g6 d1c2 b8d7 a1e1 g8g7 h2h3 g4f3 d2f3 d6f4 c2b3 f4c7 b3c2');
INSERT INTO evaluations VALUES(40525,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40525,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40525,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40525,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40525,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40525,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40525,36,'cp 0','d1e2 d6e7 c2c3 e8g8 e1g1 c8g4 c1f4 f8e8 e2d1 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(54711,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(54711,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(54711,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(54711,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(54711,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(54711,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(54711,36,'cp 107','f6f5 c1f4 e7e6 e2e3 f8d6 f1d3 g8f6 c3e2 b8d7 e1g1 e8g8 c2c4 d5c4 d3c4 d7b6 c4b3 c8d7 f3e5 b6d5 a1c1 a7a5 a2a3 g8h8 f1e1 a5a4 b3a2 d8e7 d1c2 d7b5 h2h3 b7b6 a2c4 e7e8 c4b5 e8b5 e5d3 b5b3 c2b3 a4b3 f4d6');
INSERT INTO evaluations VALUES(57194,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(57194,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(57194,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(57194,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(57194,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(57194,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(57194,36,'cp -86','d7d5 c1b2 g8f6 e2e3 c7c6 d2d4 e5e4 f2f3 a7a5 a2a3 f8d6 f3e4 d5e4 g1e2 e8g8 d1d2 d8e7 a3a4 b8a6 e2g3 h7h5 f1c4 h5h4 g3e2 a6c7 e1g1 c7d5 c4d5 f6d5 c3d5 c6d5 c2c4 c8e6 e2f4 e7g5 d2e2 d6f4 f1f4 h4h3 a1f1 h3g2 e2g2 g5g2 g1g2 d5c4 b3c4 e6c4');
INSERT INTO evaluations VALUES(56432,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(56432,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(56432,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(56432,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(56432,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(56432,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(56432,36,'cp -101','c1b2 g8f6 e2e3 c7c6 d2d4 e5e4 f2f3 a7a5 d1d2 h7h5 a2a3 c8f5 f3e4 f6e4 c3e4 f5e4 g1f3 f8d6 f1d3 d8e7 d3e4 e7e4 d2d3 e4e7 f3d2 e7h4 e1d1 e8g8 e3e4 a5a4 e4e5 h4g4 d3e2 d6c7 c2c4 f7f6 e2g4 h5g4 d1c2 f6e5');
INSERT INTO evaluations VALUES(2781,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2781,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(2781,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2781,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(2781,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(2781,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(2781,36,'cp 15','c2c3 d7d6 c1g5 c5b6 a2a4 a8b8 b1d2 h7h6 g5h4 b5b4 d2c4 c6a5 c4b6 b8b6 b3c2 b4b3 c2b1 c8g4 d3d4 c7c5 h2h3 g4f3 d1f3 c5d4 c3d4 a5c6 d4d5 c6d4 f3d1 g7g5 h4g3 b6b4 f2f3 f6h5 g3e1 b4b8 e1f2 b8b4 b1d3 a6a5 f2e1 b4b8 e1c3 h5f4');
INSERT INTO evaluations VALUES(2788,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2788,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2788,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2788,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(2788,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(2788,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(2788,36,'cp 0','d7d6 c1g5 h7h6 g5h4 c5b6 a2a4 g7g5 h4g3 a8b8 b1d2 f8e8 f1e1 c6e7 d3d4 e7g6 a4a5 b6a7 d2f1 c8b7 b3c2 d8c8 h2h3 e5d4 e4e5 g6e5 f3d4 c8d7 f1e3 e5g6 c2f5 d7d8 f5c2');
INSERT INTO evaluations VALUES(2759,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(2759,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(2759,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(2759,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(2759,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(2759,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2759,36,'cp 21','c1g5 h7h6 g5h4 c5b6 a2a4 a8b8 b1d2 b5b4 b3d5 c6e7 h4f6 g7f6 d3d4 f6f5 d4e5 b4c3 b2c3 d6e5 f3e5 e7d5 e4d5 d8d5 d2f3 d5d1 f1d1 b6c5 e5d7 c8d7 d1d7 b8b2 a1f1 c5b6 f3d4 f5f4 g2g3 b2a2 d4f5 g8h7 g1g2 f4g3 h2g3 a2a4');
INSERT INTO evaluations VALUES(11013,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11013,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(11013,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11013,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(11013,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(11013,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(11013,36,'cp 49','f3e5 c6e5 d2d4 c7c6 b5a4 c5d6 d4e5 d6e5 a4b3 e8g8 e1g1 d7d5 e4d5 e5c3 b2c3 c6d5 f1e1 c8e6 d1d3 f8e8 e1e2 a8c8 c1e3 d8c7 e3d2 c7d7 a2a4 e6f5 d3b5 a7a6 b5b6 e7c6 e2e8 c8e8 d2e3 h7h6 a1d1 d7e6 d1d5');
INSERT INTO evaluations VALUES(19955,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(19955,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(19955,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(19955,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19955,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(19955,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(19955,36,'cp -17','g8f6 d2d3 g7g6 e1g1 f8g7 b1c3 e8g8 h2h3 e7e6 c4a2 h7h6 c1e3 b7b6 d3d4 c8b7 d4d5 e6d5 e4d5 c6e7 e3f4 a7a6 a3a4 b6b5 a4b5 a6b5 c3b5 f6d5 f4d6 g7b2 d1d3 b7c6 a2d5 b2a1 d5c6 e7c6 d6f8 d8d3 c2d3');
INSERT INTO evaluations VALUES(13264,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(13264,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(13264,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(13264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13264,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(13264,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(13264,36,'cp 0','d2d3 e7e6 e1g1 f8e7 b1d2 e8g8 f1e1 d6d5 c4b5 c8d7 c2c3 d8c7 b5a4 d5e4 d3e4 a8d8 d2c4 c6a5 c4a5 c7a5 a4d7 f6d7 d1e2 b7b5 c1e3 a5a4 h2h3 a7a6 e3f4 h7h6 e1d1 d7f6 d1e1');
INSERT INTO evaluations VALUES(13261,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13261,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(13261,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13261,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13261,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(13261,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13261,36,'cp -7','e7e6 e1g1 f8e7 b1d2 d6d5 c4b5 c8d7 b5c6 d7c6 f3e5 d8d6 e5c6 d6c6 d1e2 b7b5 e4d5 f6d5 e2e4 b5b4 d2c4 e8g8 c1f4 e7f6 f1c1 f8c8 f4g3 h7h6 h2h3 a7a6 e4f3 d5e7 f3e4 e7f5 g3e5 f6e5 e4c6 c8c6 c4e5');
INSERT INTO evaluations VALUES(13947,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13947,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13947,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13947,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13947,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13947,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13947,36,'cp -5','e1g1 f8e7 f1e1 e8g8 b1d2 d6d5 c4b5 c8d7 c2c3 d8c7 b5a4 d5e4 d3e4 a8d8 a4c2 e6e5 d2c4 d7e6 d1e2 h7h6 a3a4 e6c4 e2c4 c6a5 c4b5 a7a6 b5e2 c5c4 c1e3 e7c5 e3c5 c7c5 g2g3 b7b6 a1d1 c5c7 d1d8 f8d8 e1d1 d8d1 c2d1');
INSERT INTO evaluations VALUES(13946,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13946,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13946,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13946,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13946,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13946,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(13946,36,'cp -19','f8e7 b1d2 e8g8 f1e1 d6d5 c4b5 c8d7 b5a4 d5e4 d3e4 d8c7 c2c3 a7a6 d2f1 f6g4 h2h3 g4e5 c1f4 f8d8 f3e5 c6e5 a4d7 e7f6 f1e3 d8d7 d1h5 g7g6 h5e2 c7c6 f4e5 f6e5 e3g4 e5g7 a1d1 a8d8 e4e5 b7b5 f2f4 h7h6 d1d7 d8d7 g4f2 c5c4 e2g4');
INSERT INTO evaluations VALUES(8172,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(8172,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(8172,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(8172,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(8172,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(8172,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(8172,36,'cp 0','f1e1 e8g8 b1d2 d6d5 c4b5 c8d7 c2c3 d8c7 b5a4 d5e4 d3e4 a8d8 a4c2 e6e5 d1e2 h7h6 a3a4 f6h5 d2c4 d7e6 g2g3 b7b6 c4e3 h5f6 f3d2 d8d7 d2c4 f8d8 f2f3 f6h7 e1d1 d7d1 e3d1 d8d7 d1e3 e7g5 c2d3 g5e3 c4e3');
INSERT INTO evaluations VALUES(30686,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30686,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(30686,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(30686,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(30686,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(30686,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30686,36,'cp -20','f8b4 e4e5 c7c5 g1f3 c5d4 f3d4 g8h6 c1e3 b4c3 b2c3 b8c6 d4b5 h6f5 e3f2 c6e7 d1d2 a7a6 b5d6 f5d6 d2d6 d8c8 e1c1 b6b5 f1d3 c8c6 h1e1 a8c8 f2d4 e7d5 d6c6 b7c6 g2g3 d5e7 d1d2 c6f3 h2h3 f3d5 e1e3 h7h5 a2a4 b5a4 d3a6');
INSERT INTO evaluations VALUES(27957,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(27957,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(27957,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(27957,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27957,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27957,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27957,36,'cp 0','e4e5 c7c5 a2a3 b4c3 b2c3 g8e7 g1f3 d8c7 f1e2 b8c6 e1g1 c5d4 c3d4 a8c8 a1b1 e7f5 f1f2 h7h5 a3a4 c6a5 e2d3 b7d5 d3a6 d5b7');
INSERT INTO evaluations VALUES(41431,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41431,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(41431,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(41431,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(41431,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41431,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41431,36,'cp 29','g5f6 g7f6 c4d5 c7c5 d4c5 b8d7 e2e3 e8g8 g1e2 d7c5 e2d4 d8a5 f1c4 a5b4 c4b3 c8f5 e1g1 f5d3 f1e1 f8d8 d1d2 a8c8 h2h3 b4b6 e1d1 a7a6 a1c1 a6a5 g2g3 h7h5 h3h4 f6g7 b3c2 d3c4 c1b1 c4d5 b2b4 g7d4 d2d4');
INSERT INTO evaluations VALUES(41421,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(41421,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41421,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(41421,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(41421,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41421,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(41421,36,'cp 21','g7f6 c4d5 c7c5 d4c5 b8d7 e2e3 e8g8 f1c4 d8a5 g1e2 d7c5 e2d4 c5e4 a1c1 c8d7 e1g1 e4c3 c1c3 a8c8 d1e2 c8c7 f1c1 f8c8 g2g3 d7h3 e2d3 a5b4 c4b3 a7a5 a2a3 b4b6 c3c7 c8c7 c1c7 b6c7 d3c3 c7b6 b3c4 b6c5 c4f1');
INSERT INTO evaluations VALUES(41167,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41167,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(41167,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(41167,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(41167,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41167,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(41167,36,'cp 15','c4d5 c7c5 d4c5 b8d7 e2e3 e8g8 f1c4 d8a5 g1e2 d7c5 a2a3 c5e4 e1g1 e4c3 e2c3 f6c3 b2c3 a5c3 d1b3 c3e5 a1c1 a8b8 a3a4 c8d7 h2h3 f8c8 d5d6 e5f6 d6e7 f6e7 e3e4 d7e6 c4e6 e7e6 b3e6 f7e6 f1d1 c8c1 d1c1 a7a5 c1b1 b8c8 b1b7 c8c4');
INSERT INTO evaluations VALUES(47626,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(47626,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47626,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(47626,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(47626,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(47626,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(47626,36,'cp 42','c8g4 d2d4 b8d7 h2h3 g4h5 d1d3 h5f3 d3f3 e5d4 f3d3 g7g6 d3d4 f8g7 d4d1 e8g8 b1c3 d7c5 c1f4 f6d7 d1d2 d8h4 a1d1 g7c3 b2c3 a8e8 c4d3 g8h7 f2f3 b7b6 f4e3 c5e6 f3f4 e8e7 g1h1 d7c5 e4e5 c5d3 c2d3 d6e5 f4e5');
INSERT INTO evaluations VALUES(51367,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(51367,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(51367,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(51367,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(51367,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(51367,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(51367,36,'cp 90','h2h4 c7c5 c2c3 d8d7 b1d2 h7h5 d2f3 c8a6 f1a6 b8a6 g1e2 g8e7 e1f1 a6b8 h1h3 b8c6 e2g3 e7g6 f3g5 e8c8 d1d3 c8b8 g3e2 d7e8 a2a3 c6a5 d3f3 a5b3 a1b1 d8d7 f1g1 b3c1 b1c1 e8e7 f3g3 d7c7 b2b4 f7f6');
INSERT INTO evaluations VALUES(51948,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51948,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(51948,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(51948,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(51948,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(51948,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(51948,36,'cp 23','b8c6 g1f3 d7d5 c4d5 e6d5 e1g1 g8f6 d2d4 c5c4 b1c3 f8e7 f3e5 e8g8 c1e3 f8e8 e3g5 c8e6 f2f4 f6g4 f4f5 c6e5 g5e7 e8e7 f5e6 g4e3 d1d2 e5g4 e6f7 g8f8 f1f4 e3g2 g1g2 g4e3 g2f2 d8d7 f4f3 e3g4 f2g1 d7e6 d2f4 g4e3 f4g5 h7h6');
INSERT INTO evaluations VALUES(18783,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(18783,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(18783,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(18783,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(18783,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(18783,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(18783,36,'cp 31','g1f3 d7d5 c4d5 e6d5 e1g1 g8f6 d2d4 c5c4 b1c3 f8e7 f3e5 e8g8 b2b3 c4b3 a2b3 c8f5 c1g5 c6b4 g5f6 e7f6 e2e3 a7a5 c3a2 b4a2 a1a2 a8c8 d1d2 b7b6 e5d3 f6e7 b3b4 a5b4 d3b4 e7b4 d2b4');
INSERT INTO evaluations VALUES(18782,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(18782,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(18782,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18782,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(18782,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(18782,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(18782,36,'cp 44','d7d5 c4d5 e6d5 d2d4 g8f6 e1g1 c5c4 b2b3 c4b3 d1b3 f8e7 c1g5 e8g8 e2e3 c6a5 b3d3 c8e6 b1c3 a5c4 f3e5 a8c8 a1c1 d8a5 g5f6 e7f6 c3d5 c4e5 d4e5 c8c1 d5f6 g7f6 f1c1 a5e5 g2b7 e6a2 d3a3 f8b8 a3a7 a2e6 b7e4 b8d8 a7b7 d8b8 b7c6 e6h3');
INSERT INTO evaluations VALUES(23448,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(23448,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(23448,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(23448,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(23448,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(23448,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(23448,36,'cp -68','g8f6 b1c3 e7e6 a2a3 f8c5 f3g3 h7h5 e4e5 f6g4 c3e4 c5e7 d2d4 c6d4 g3c3 d4c6 g1f3 f7f5 e5f6 g7f6 h2h3 d7d5 h3g4 d5e4 f3d2 e6e5 h1h5 h8h5 g4h5 c8f5 d2b3 d8d7 c1e3 e8c8 f1e2 c6d4 a1d1 d4e2 d1d7 e2c3 d7d8 c8d8 b2c3 e7a3');
INSERT INTO evaluations VALUES(31254,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(31254,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(31254,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(31254,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(31254,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(31254,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(31254,36,'cp 70','d2d4 c5d4 d1d4 b8c6 f1b5 g4d7 b5c6 d7c6 c1e3 e7e5 d4d3 f8e7 g2g4 g8f6 g4g5 f6h5 e1c1 a8c8 c1b1 e8g8 c3d5 c6d5 d3d5 d8c7 d1c1 c7c6 c2c4 b7b6 h2h4 g7g6 f3d2 c6d5 c4d5 f7f5 c1c8 f8c8 h1c1 c8c1 b1c1 f5e4 d2e4');
INSERT INTO evaluations VALUES(32473,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32473,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(32473,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32473,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(32473,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(32473,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(32473,36,'cp 5','d7b5 c3b5 c7c6 b5c3 b8d7 e1g1 g8f6 a2a4 f8e7 d2d4 e8g8 h2h3 f8e8 c1e3 e5d4 e3d4 b7b6 f1e1 d8c7 d1d3 d7e5 d3e3 e5g6 c3e2 d6d5 e4e5 e7c5 d4c5 b6c5 e3c5 f6e4 c5d4 g6e5 f3e5 e8e5 e2g3 e5e6 g3e4');
INSERT INTO evaluations VALUES(32069,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32069,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(32069,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(32069,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(32069,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(32069,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(32069,36,'cp 17','c3b5 c7c6 b5c3 f8e7 e1g1 b8d7 d2d4 g8f6 a2a4 e8g8 f1e1 f8e8 h2h3 d8a5 c1e3 h7h6 d1d3 a8d8 d4d5 c6d5 c3d5 f6d5 e4d5 d7f6 d3b5 a5b5 a4b5 f6d5 a1a7 d8d7 c2c4 d5e3 f2e3 e8c8 b2b3 e5e4 f3d4');
INSERT INTO evaluations VALUES(32068,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(32068,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(32068,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(32068,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(32068,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(32068,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(32068,36,'cp 0','c7c6 b5c3 g8f6 e1g1 f8e7 d2d4 b8d7 f1e1 e8g8 a2a4 f8e8 c1e3 d8a5 h2h3 h7h6 d4e5 d6e5 f3h4 e7f8 h4f5 d7c5 d1f3 e8e6 e1d1 a5b4 e3c5 f8c5 f3g3 g7g6 f5h6 g8g7 h6f5 g7h8 f5h6');
INSERT INTO evaluations VALUES(33430,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33430,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(33430,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(33430,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(33430,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33430,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(33430,36,'cp 18','g1f3 e8g8 d1e2 c7c5 e2e5 f6e5 d4e5 b7b6 e1c1 c8b7 f1b5 b7d5 h1e1 a7a6 b5f1 b6b5 f3d2 f8d8 d2e4 b8d7 e4c3 d7b6 c3d5 d8d5 g2g3 a8d8 d1d5 d8d5 f1g2 d5d8 f2f4 g8f8 e1d1 d8d1 c1d1 b6c4 g2b7 c4b2 d1c1 b2a4 b7a6 b5b4 c1d2 h7h6 a6c4 a4b6 c4b3');
INSERT INTO evaluations VALUES(33429,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(33429,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(33429,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33429,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(33429,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33429,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33429,36,'cp 15','c7c5 f1b5 c8d7 d1e2 c5d4 b5d7 b8d7 e1c1 e8g8 d1d4 d7b6 c1b1 f8d8 h1d1 d8d4 d1d4 h7h6 e2e4 a8d8 d4d8 f6d8 b2b3 d8d5 e4d5 b6d5 b1b2 d5f4 f3e1 g8f8 g2g3 f4h3 f2f3 f8e7 c2c4 e7d6 b2c2 e6e5 b3b4 f7f6 c2d3 g7g6 d3e3 h3g5 e1d3');
INSERT INTO evaluations VALUES(51099,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(51099,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(51099,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51099,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(51099,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(51099,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(51099,36,'cp 0','e7e6 d2d4 d7d5 e4d5 e6d5 c4b3 f8d6 g1f3 g8f6 d1e2 d6e7 e1g1 e8g8 f1e1 f8e8 c1g5 c8e6 c2c4 b8c6 b1c3 d5c4 b3c4 e6g4 h2h3 g4f3 e2f3 c6d4 f3b7 a8b8 b7a6 b8a8 a6b7 a8b8');
INSERT INTO evaluations VALUES(51024,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51024,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51024,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51024,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51024,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51024,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51024,36,'cp 0','d2d4 d7d5 e4d5 e6d5 c4b3 f8d6 g1f3 g8f6 d1e2 d6e7 e1g1 e8g8 f1e1 f8e8 c1g5 c8e6 c2c4 b8c6 b1c3 d5c4 b3c4 e6g4 h2h3 g4f3 e2f3 c6d4 f3b7 a8b8 b7a7 b8a8');
INSERT INTO evaluations VALUES(51025,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51025,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51025,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51025,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51025,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51025,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51025,36,'cp 0','d7d5 e4d5 e6d5 c4b3 f8d6 g1f3 g8e7 e1g1 c8e6 f1e1 e8g8 b1d2 b8d7 d2f1 e7g6 c2c3 h7h6 f1g3 d8f6 g3h5 f6f5 h5g3');
INSERT INTO evaluations VALUES(54917,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(54917,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(54917,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(54917,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(54917,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(54917,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(54917,36,'cp 146','c4d5 g8f6 g1f3 c8f5 d1a4 c7c6 d5c6 b8c6 e2e3 a7a6 f3e5 b7b5 e5c6 d8c8 a4d1 c8c6 a2a4 a8b8 a4b5 a6b5 b1c3 f5d7 d1f3 b5b4 f3c6 d7c6 c3e2 e7e6 f2f3 f8e7 e3e4 e8g8 c1e3 b8a8 a1a8 f8a8 e2c1 f6e8 f1d3 e8d6 e1f2 f7f5 e4e5 d6e8 c1e2 g7g5 h1c1 c6d5 d3b5 g8f8 b5c6 d5c6 c1c6');
INSERT INTO evaluations VALUES(54912,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(54912,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(54912,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(54912,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(54912,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(54912,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(54912,36,'cp 124','g8f6 g1f3 c8f5 b1c3 f6d5 d1b3 d5c3 b2c3 b8d7 b3b7 e7e6 g2g3 a8b8 b7a7 f5e4 a7a4 f8e7 f1g2 e8g8 e1g1 c7c5 f3e1 e4g2 e1g2 c5d4 c3d4 b8b4 a4a7 d7b8 e2e3 b8c6 a7a6 d8a8 a6a8 f8a8 a2a3 b4a4 g2f4 c6a5 f1d1 a5c4 h2h4 g8h8 a1b1 c4a3');
INSERT INTO evaluations VALUES(56336,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(56336,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(56336,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(56336,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(56336,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(56336,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(56336,36,'cp -52','b8d7 e2e3 g8f6 c2c4 c7c6 g1f3 f8d6 d2d3 e8g8 f1e2 f8e8 e1g1 d7f8 f1e1 h7h6 b1d2 f8g6 e2f1 b7b6 d1c2 c8e6 g2g3 a8c8 e3e4 d5d4 b2d4 e5d4 e4e5 g6e5 f3e5 d8c7 e5f3 c6c5 b4b5 e6f5 a3a4 c8d8 f1g2 g7g5 e1e8 d8e8 a1e1 e8e1 f3e1');
INSERT INTO evaluations VALUES(31908,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(31908,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(31908,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(31908,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(31908,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(31908,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(31908,36,'cp 50','f2f3 g4d7 e2e4 e7e6 e4d5 e6d5 b1c3 d7e6 h2h4 a7a6 d1d2 d8d7 h4h5 b8c6 e1c1 e8c8 g1e2 f8d6 f4e3 f7f6 e2f4 g8e7 f4e6 d7e6 d1e1 e6f7 g2g4 f6f5 c3e2 h8f8 c1b1 c8b8 a2a3 e7c8 e2c3 d6g3 e1d1 f5f4 e3g1');
INSERT INTO evaluations VALUES(37687,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37687,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37687,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37687,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(37687,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37687,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(37687,36,'cp 11','e2e4 g8f6 f1d3 d7d5 e4e5 f6d7 e1g1 b8c6 f1e1 h7h6 c2c4 d5c4 d3c4 d8b6 d1e2 c6d4 f3d4 c5d4 b1a3 d7e5 c1f4 e5g6 c4b5 e8f8 f4g3 h6h5 h2h3 e6e5 b5d3 h5h4 g3h2 b6f6 a3b5 d4b6 a2a4 c8d7 a4a5 b6c5 e1c1 a8c8 b2b4 c5b4 c1c8 d7c8 b5a7 c8d7 a5a6 b7a6 a1a6');
INSERT INTO evaluations VALUES(41264,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41264,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(41264,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(41264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41264,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(41264,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(41264,36,'cp 12','e8g8 g1f3 c7c5 h2h3 c5d4 e3d4 b8c6 f1e2 d6d5 e1g1 d5c4 e2c4 c6a5 b2b3 b7b6 c4e2 c8b7 f3e5 a8c8 a1c1 f6e4 c3e4 c8c1 f4c1 b7e4 f1e1 d8d6 e2f1 e4a8 c1f4 a5c6 e5c6 d6c6 d4d5 c6d7 d5d6 e7d6 d1d6 d7d6 f4d6 f8d8 d6f4 a8b7 e1c1 h7h5 c1c7');
INSERT INTO evaluations VALUES(44559,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(44559,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(44559,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(44559,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(44559,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(44559,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(44559,36,'cp -81','f6e4 d4d5 d7d6 f1d3 e4f6 e1g1 g7g6 c2c4 f8g7 b1c3 e7e5 h2h3 e8g8 c1g5 h7h6 g5h4 d8d7 f3d2 f6h5 d3c2 a7a6 g2g4 h5f4 g1h2 d7c7 f1g1 b8d7 d2e4 g6g5 h4g3 d7f6 e4f6 g7f6 c3e4 f6g7 g3f4 e5f4');
INSERT INTO evaluations VALUES(44654,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(44654,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(44654,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(44654,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(44654,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(44654,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(44654,36,'cp -55','b1d2 e4d2 c1d2 c5d4 f3d4 d7d6 d2c3 e7e5 f1b5 c8d7 d1e2 b8c6 e1c1 f8e7 d4c6 b7c6 b5d3 e8g8 f2f4 e5f4 e2e4 f7f5 e4d4 f8f7 d3e2 d8b6 d4f4 d6d5 h1e1 b6d8 c1b1 e7g5 f4f2 g5f6 c3f6 d8f6 c2c4 d7e6 f2d4 a8b8 e2h5 f7e7 d4f6 g7f6 c4d5 c6d5 b1a1 g8g7 h5f3');
INSERT INTO evaluations VALUES(45003,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(45003,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(45003,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(45003,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(45003,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(45003,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(45003,36,'cp 79','e7e6 g1f3 c6c5 e1g1 c5d4 e3d4 f8e7 f1e1 e8g8 a2a3 c8d7 d4d5 e6d5 c3d5 f6d5 d1d5 b8c6 d5h5 e7f6 f3g5 f6g5 c1g5 d8a5 h5h4 a5f5 g5d2 d7e6 c4f1 f8d8 d2c3 f5g4 h4g4 e6g4 f2f3 g4e6 g1f2 h7h6 a1d1 d8d1');
INSERT INTO evaluations VALUES(45699,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(45699,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(45699,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(45699,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(45699,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(45699,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(45699,36,'cp 99','g1f3 c6c5 e1g1 c5d4 e3d4 f8e7 a2a3 e8g8 f1e1 c8d7 d1e2 d7c6 f3e5 c6e8 c1e3 b8d7 e5d7 e8d7 d4d5 e6d5 c3d5 f6d5 c4d5 e7f6 e3c5 f8e8 e2c4 d7e6 d5e6 f7e6 e1e6 e8e6 c4e6 g8h8 e6b3 b7b6 c5e3 d8g8 b3b5 g8e8 b5e8 a8e8 a1b1 e8f8 b2b4 f8d8 b1c1');
INSERT INTO evaluations VALUES(49121,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(49121,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(49121,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(49121,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(49121,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(49121,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(49121,36,'cp 125','c2c4 f8d6 b1c3 e8g8 a1c1 b8c6 a2a3 c8d7 f1e2 c6e7 e1g1 d7e8 d1b3 a7a5 f1e1 a5a4 b3c2 f6e4 f3e5 g8h8 e1d1 h7h6 f2f3 e4c3 c2c3 g7g5 f4g3 c7c6 c4c5 d6c7 d1e1 e7g6 e5g6 e8g6 e2d3 d8e7 g3c7 e7c7');
INSERT INTO evaluations VALUES(49110,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(49110,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(49110,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(49110,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(49110,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(49110,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(49110,36,'cp 123','f8d6 b1c3 e8g8 a1c1 b8c6 a2a3 c8d7 f1e2 c6e7 e1g1 d7e8 f1e1 c7c6 c3a4 e7g6 f4d6 d8d6 a4c5 d6e7 d1b3 a8b8 c5d3 f6e4 c4c5 b8a8 f3e5 g6e5 d3e5 e4g5 e5d3 g5f7 b3c3 g7g5 f2f4 a7a5 b2b4 a5b4 a3b4 g5g4 c1a1');
INSERT INTO evaluations VALUES(51297,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51297,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(51297,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51297,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(51297,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51297,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(51297,36,'cp 74','c2c4 e7e6 c4d5 e6d5 b1c3 g8f6 f3e5 f8e7 c1f4 e8g8 e2e3 c7c5 d1f3 c8b7 f1d3 a7a6 e1g1 b8c6 f1d1 c5d4 e5c6 b7c6 e3d4 f8e8 a1c1 b6b5 d1e1 a8c8 h2h3 e7f8 a2a3 c6b7 e1e8 d8e8 f4e5 f6d7 c1e1 d7e5 e1e5');
INSERT INTO evaluations VALUES(51288,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(51288,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(51288,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(51288,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(51288,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(51288,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(51288,36,'cp 63','e7e6 b1c3 g8f6 c4d5 e6d5 f3e5 f8e7 c1f4 e8g8 e2e3 c8b7 f1d3 b8d7 e1g1 a7a6 a1c1 f8e8 e5d7 f6d7 d1f3 d7f6 f4e5 g7g6 f1d1 f6d7 e5f4 d7f6 h2h3 f6h5 f4e5 e7d6 e5d6 d8d6 b2b4 h5f6 a2a3');
INSERT INTO evaluations VALUES(51363,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(51363,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(51363,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(51363,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(51363,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(51363,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(51363,36,'cp 62','b1c3 g8f6 c4d5 e6d5 f3e5 c7c6 g2g3 c8b7 f1g2 b8d7 e1g1 d7e5 d4e5 f6d7 e2e4 d5d4 e5e6 f7e6 d1d4 d7c5 d4e5 d8f6 e5f6 g7f6 f1d1 a8d8 c1e3 c5d3 g2f1 d3e5 f1e2 h8g8 d1d8 e8d8 a1d1 d8c7 f2f4 e5g4 e2g4 g8g4 g1f2 e6e5 f4f5 b7a6 f2f3');
INSERT INTO evaluations VALUES(51359,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(51359,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(51359,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(51359,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(51359,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(51359,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51359,36,'cp 54','g8f6 c4d5 e6d5 f3e5 c7c6 c1f4 c8b7 e2e3 b8d7 d1f3 f8e7 f1d3 d7e5 f4e5 e8g8 e1g1 f8e8 a1d1 b6b5 h2h3 a7a5 a2a4 b5b4 c3e2 g7g6 d1c1 f6d7 e5g3 e7f8 c1c2 d8f6 f3f6 d7f6 f1c1 f6d7');
INSERT INTO evaluations VALUES(19088,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19088,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19088,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19088,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19088,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(19088,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19088,36,'cp 5','b1c3 f8b4 f1b5 g8f6 g1e2 a8b8 a2a3 b4e7 b5d3 a7a6 e1g1 b7b5 a4b3 d7d6 c1g5 e8g8 g5f6 e7f6 c3d5 f6g5 a3a4 b5b4 c2c3 b4c3 b3c3 c6e7 d5e7 d8e7 d3c4 g8h8 e2g3 g5f4 f1d1 f8d8 h2h3 c8e6 c4e6 f7e6');
INSERT INTO evaluations VALUES(20972,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(20972,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(20972,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(20972,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(20972,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(20972,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20972,36,'cp 28','d5c4 g1f3 e7e6 e2e3 c6a5 f1c4 a5c4 d1a4 c8d7 a4c4 c7c5 b1c3 c5d4 f3d4 a8c8 c4b3 d8b6 d4b5 g8f6 e1g1 d7b5 b3b5 b6b5 c3b5 f8c5 b5d6 c5d6 f4d6 c8c2 f1d1 f6e4 d6a3 f7f5 f2f3 e4f6 d1c1 c2c1 a1c1 e8d7 c1d1 d7c8 d1d6 e6e5 h2h4 h8d8 d6d8 c8d8');
INSERT INTO evaluations VALUES(21116,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(21116,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(21116,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21116,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(21116,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(21116,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21116,36,'cp 33','g1f3 e7e6 e2e3 c6a5 f1c4 a5c4 d1a4 c8d7 a4c4 c7c5 e1g1 c5d4 c4d4 g8f6 f1d1 d8b6 d4b6 a7b6 b1c3 d7c6 a2a3 f8c5 b2b4 c5e7 f3d4 e8g8 f4c7 c6d7 d4b5 f6d5 c3d5 e6d5 b5c3 e7f6 a1c1 f6c3 c1c3 f8c8');
INSERT INTO evaluations VALUES(21115,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(21115,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(21115,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21115,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(21115,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(21115,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(21115,36,'cp 30','a7a6 d4d5 e7e5 f4g3 c6e7 e2e4 g8f6 b1c3 e7g6 f3e5 g6e5 g3e5 b7b5 d1d4 f8d6 a2a4 c8d7 f1e2 e8g8 e5g3 f8e8 e4e5 b5b4 c3e4 f6e4 d4e4 c4c3 b2c3 d8e7 f2f4 d6c5 g3f2 b4c3 e1g1 c5f2 f1f2 e7c5 e2d3');
INSERT INTO evaluations VALUES(26176,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26176,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(26176,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(26176,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(26176,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(26176,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(26176,36,'cp 25','g6h7 f1d3 h7d3 d1d3 e7e6 c1d2 g8f6 e1c1 f8e7 g3e4 e8g8 h1g1 f6e4 d3e4 d7f6 e4e2 d8d5 c1b1 d5e4 e2e4 f6e4 d2e3 c6c5 c2c3 a8c8 d4d5 e6d5 d1d5 e4f6 d5e5 f8e8 e3f4 c8d8 g1e1 g8f8 b1c2 f6g4 e5e2 g4f6');
INSERT INTO evaluations VALUES(35116,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35116,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(35116,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35116,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(35116,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(35116,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35116,36,'cp 10','f8c5 c2c3 g8e7 f1e2 b8c6 b3b4 c5b6 d2d3 e8g8 a2a4 a7a6 e1g1 d7d5 c1g5 f6e6 b1d2 h7h6 g5e3 d5e4 d2e4 f7f5 e4c5 e6f7 b4b5 a6b5 a4b5 a8a1 d1a1 f5f4 b5c6 e7c6 e3f4 f7f4 a1a3 f4f7 f3d2 c6e7 d2c4 e7d5 c4b6 c7b6 c5e4 d5f4 f1e1');
INSERT INTO evaluations VALUES(33229,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(33229,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(33229,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33229,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33229,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33229,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(33229,36,'cp 0','c2c3 g8e7 f1e2 c5b6 d2d4 e5d4 c1g5 f6g6 g5e7 e8e7 e1g1 d7d6 c3d4 h8e8 b1c3 e7f8 a1c1 b8c6 g1h1 g6h6 d4d5 c6e5 c3a4 b6a5 a2a3 c8d7 b3b4 d7a4 d1a4 a5b6 f3e5 e8e5 a4c2 h6f4 e2d3 e5h5 g2g3 f4f3 h1g1 h5g5 c2d2 g5g3 h2g3 f3g3 g1h1 g3h3 h1g1');
INSERT INTO evaluations VALUES(39312,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39312,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39312,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39312,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(39312,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39312,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39312,36,'cp 0','e2e4 a7a5 f1d3 g8e7 g1e2 e8g8 e2g3 d7d6 e1g1 b8a6 a2a3 c5b6 b1c3 c7c6 d5c6 b7c6 c3a4 b6a7 c1e3 a7e3 f2e3 c8e6 a1b1 a6b8 d3e2 d8c7 e2g4 e6c4 g4e2');
INSERT INTO evaluations VALUES(39311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39311,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39311,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39311,36,'cp 0','d7d6 f1d3 g8e7 g1e2 e8g8 e1g1 a7a5 e2g3 b8a6 b1c3 c5d4 g3e2 d4c5');
INSERT INTO evaluations VALUES(38753,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38753,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38753,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38753,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38753,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38753,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(38753,36,'cp -34','f1d3 g8f6 g1f3 e8g8 e1g1 c8g4 h2h3 g4h5 c2c4 b8d7 b1c3 c7c6 f1e1 a8c8 a1b1 c5b4 e1e3 d7c5 d3c2 b4c3 b2c3 d8a5 a2a4 a5a6 g2g4 h5g6 f3d2 f6d7 c1a3 f8e8 b1a1 a6b6 a3c5 b6c5 d1e2 d7b6 c2d3');
INSERT INTO evaluations VALUES(52128,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52128,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52128,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52128,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(52128,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52128,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(52128,36,'cp 0','b8c6 g2g3 g8e7 c2c4 d7d6 b1c3 g7g6 f1g2 f8g7 a1b1 e8g8 e1g1 c8e6 a2a3 a7a5 c1d2 a8b8 f3g5 e6c8 g5f3 c8e6');
INSERT INTO evaluations VALUES(7679,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(7679,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(7679,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(7679,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7679,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7679,36,'cp -1','d1f3 e8g8 f1c4 h7h6 g5f6 e7f6 e1g1 g8h8 f3h5 d7d6 a1d1 d8c7 b2b3 h8h7 h5e2 f6e7 b3b4 a8b8 b4b5 c6b5 c3b5 c7c5 d1d5 c8g4 e2g4 c5c4 b5d6 e7d6 d5d6 c4a2 g4f5 h7g8 f5e5 a2c2 e5d5 b8b1 g2g3 a7a5 f1b1 c2b1 g1g2 a5a4 h2h4 b1b4');
INSERT INTO evaluations VALUES(8246,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(8246,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(8246,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(8246,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8246,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(8246,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(8246,36,'cp 12','c3d5 f6d7 c2c3 e7e6 d5e3 d7f6 f1e1 e8g8 h2h3 d8c7 a2a4 b7b6 c4a2 a7a6 c1d2 b6b5 a2b1 b5b4 d3d4 a6a5 d4d5 e6d5 e3d5 f6d5 e4d5 c6e5 f3e5 g7e5 d2h6 b4c3 h6f8 g8f8 e1e5 d6e5 b2c3 c7d6 b1c2 a8b8');
INSERT INTO evaluations VALUES(9356,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(9356,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(9356,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9356,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(9356,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9356,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9356,36,'cp 18','c2c3 c5b6 d2d3 e8g8 a2a4 a7a5 h2h3 h7h6 b1a3 c6e7 c4a2 c7c6 d3d4 e7g6 d4e5 d6e5 a3c4 b6c7 d1d8 f8d8 c1e3 a8a6 c4d2 b7b5 a2b3 f6d7 e1d1 d7b6 a4b5 c6b5 c3c4 b5c4 d2c4 d8d1 a1d1 c8e6 c4d2 e6b3 d2b3 b6c4 d1d7 c4e3 d7c7');
INSERT INTO evaluations VALUES(9340,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(9340,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(9340,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(9340,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9340,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(9340,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(9340,36,'cp 9','c5b6 d2d3 e8g8 h2h3 a7a5 a2a4 c8e6 b1a3 h7h6 d1b3 d8d7 c1e3 b6e3 e1e3 e6c4 a3c4 f8e8 a1e1 b7b6 c4a3 c6e7 d3d4 e7g6 a3b5 f6h5 d4e5 d6e5 e1d1 d7c8 g2g3 e8e7 g1h2 e7d7 b3c2 h5f6 e3e1 c7c6 b5d6 c8c7 d6f5');
INSERT INTO evaluations VALUES(28797,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(28797,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(28797,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(28797,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(28797,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(28797,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28797,36,'cp 12','e7e6 a2a3 d7d5 c4d5 f6d5 e2e3 f8e7 d1a4 c7c6 c3d5 d8d5 f1c4 d5d6 e1g1 b8d7 a4c2 e8g8 f1d1 d6c7 c1d2 a8c8 c4d3 h7h6 a1c1 c7b8 d2c3 c6c5 d3h7 g8h8 h7e4 c5c4 e4b7 b8b7 e3e4 d7f6 d1e1 b6b5 c1d1 f8d8 h2h3');
INSERT INTO evaluations VALUES(36970,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(36970,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36970,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36970,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(36970,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36970,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36970,36,'cp 0','b7b6 e1g1 d6g3 h2g3 b8d7 c2c4 d5c4 d3c4 c8b7 b1c3 c7c5 d1e2 a7a6 a2a4 d8e7 f1d1 c5d4 e3d4 e7d6 f3e5 h7h6 c4a2 d6e7 a1c1 f8e8 a2c4 e8c8 c4d3 c8d8 d3c4 d8e8 c4d3');
INSERT INTO evaluations VALUES(35649,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35649,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35649,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35649,36,'cp 0','c2c4 d5c4 d3c4 c8b7 e1g1 d6g3 h2g3 b8d7 b1c3 c7c5 d1e2 a7a6 a2a4 d8e7 f1d1 c5d4 e3d4 e7d6 f3e5 h7h6 a1c1 f8d8 c4a2 d6e7 f2f4 h6h5 f4f5 d7e5 d4e5 f6g4 f5e6 e7c5 g1h1 g4f2 h1h2 d8d1 c1d1 f2g4 h2h1 g4f2');
INSERT INTO evaluations VALUES(41347,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(41347,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(41347,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(41347,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(41347,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(41347,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41347,36,'cp 34','e8g8 c2c4 c8f5 b2b3 c7c6 c1b2 b8d7 b1c3 d6d5 e2e3 f6e4 a1c1 e7e6 f3h4 d7f6 c3e2 f8e8 e2f4 d8e7 f4d3 a8d8 a2a4 e4g5 h4f5 e6f5 a4a5 g5e6 a5a6 b7b6 c1c2 h7h5 d1e2 f6e4 f1a1');
INSERT INTO evaluations VALUES(46365,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(46365,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46365,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(46365,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(46365,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(46365,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46365,36,'cp 15','e3d4 b8c6 f1b5 a7a6 b5a4 b7b5 a4c2 a6a5 e1g1 f8d6 f4d6 d8d6 a2a3 e8g8 c2d3 d6b8 d1c2 c8d7 b1d2 f8c8 d2b3 b5b4 b3c5 h7h6 a3b4 a5b4 a1a8 b8a8 f1e1 b4c3 b2c3 a8a5 h2h3 c8b8 f3d2 a5c7 d2b3 e6e5 d4e5');
INSERT INTO evaluations VALUES(46336,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(46336,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(46336,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46336,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(46336,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(46336,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46336,36,'cp 22','b8c6 f1d3 f6h5 f4e3 f8d6 b1d2 h5f4 e3f4 d6f4 e1g1 c8d7 d1e2 e8g8 f3e5 c6e5 d4e5 d8b6 d2f3 f7f6 g2g3 f4e5 f3e5 f6e5 e2e5 a8c8 a1b1 h7h6 g1g2 c8e8 f2f4 d7b5 d3b5 b6b5 f1e1 b5d3 e5e2 d3e2 e1e2 g8f7 e2e5 e8d8 a2a3 b7b6 b1d1');
INSERT INTO evaluations VALUES(13660,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13660,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(13660,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(13660,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(13660,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(13660,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(13660,36,'cp 16','f1b5 a7a6 b5d3 f6h5 f4e3 f8d6 e1g1 e8g8 b1d2 h5f4 d3c2 f4g6 f1e1 c8d7 d2b3 b7b6 a2a4 h7h6 b3c1 a8c8 d1e2 e6e5 c2g6 f7g6 d4e5 c6e5 f3e5 d6e5 c1d3 e5d6 e3d4 f8e8 d3e5 d8c7 f2f4 d6e5 f4e5');
INSERT INTO evaluations VALUES(49435,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(49435,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(49435,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(49435,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(49435,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(49435,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(49435,36,'cp -40','d7d5 c1f4 c7c5 e2e3 a7a6 g1f3 b7b5 d4c5 f8c5 f1d3 c8b7 d1e2 b8c6 e3e4 c6d4 f3d4 c5d4 a1d1 d4c3 b2c3 e8g8 e1g1 d5e4 d3e4 d8c8 e4b7 c8b7 c3c4 f8c8 d1d4 c8c4 d4c4 b5c4 e2c4 a8c8 c4e2 f6e4 f2f3 e4f6 f4e5 b7d5 g1h1 c8c4 e5f6');
INSERT INTO evaluations VALUES(51794,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(51794,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(51794,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(51794,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(51794,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51794,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(51794,36,'cp 60','e2e4 d7d5 b1d2 g8f6 e4e5 f6d7 c2c3 c6c5 f4e3 b8c6 g1f3 d8b6 d1c2 c5d4 c3d4 c6b4 c2d1 b4c6 a1b1 f7f6 b2b4 f6e5 d4e5 b6d8 f1b5 f8b4 e1g1 b4d2 d1d2 e8g8 b5c6 b7c6 d2c2 d7b6 e3b6 a7b6 c2c6 a8a6 f1c1 h7h6 c6d6 d8d6 e5d6 b6b5 b1b5 a6d6 f3e5 d6d8 h2h4 c8a6 b5b3');
INSERT INTO evaluations VALUES(51806,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(51806,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(51806,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(51806,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(51806,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(51806,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(51806,36,'cp 66','d7d5 b1d2 g8f6 e4e5 f6d7 f4e3 c6c5 c2c3 d8b6 a1b1 b8c6 g1f3 c5d4 c3d4 f7f6 b2b4 f8b4 a2a3 b4d2 f3d2 b6c7 d1h5 g7g6 h5g4 e8g8 g4e6 g8g7 e6d5 f6e5 b1c1 e5d4 d5d4 c6d4 c1c7 d4e6 c7c1 d7f6 f2f3 f6d5 e3f2 f8e8 d2e4 e6f4 g2g3');
INSERT INTO evaluations VALUES(55084,30,'cp -211',NULL);
INSERT INTO evaluations VALUES(55084,31,'cp -204',NULL);
INSERT INTO evaluations VALUES(55084,32,'cp -201',NULL);
INSERT INTO evaluations VALUES(55084,33,'cp -209',NULL);
INSERT INTO evaluations VALUES(55084,34,'cp -201',NULL);
INSERT INTO evaluations VALUES(55084,35,'cp -196',NULL);
INSERT INTO evaluations VALUES(55084,36,'cp -206','d5e4 d2d3 g8f6 d1d2 c8f5 d3e4 d8d2 b1d2 f6e4 g1f3 e7e6 f1d3 e4d6 d3f5 d6f5 d2e4 h7h5 c1d2 b8c6 e1c1 f7f6 e4g3 f5d6 h1e1 e8f7 c2c3 a8e8 c1c2 c6e7 a2a4 h5h4 g3e4 e7d5 g2g3 h4g3 h2g3 c7c6');
INSERT INTO evaluations VALUES(55294,30,'cp -205',NULL);
INSERT INTO evaluations VALUES(55294,31,'cp -190',NULL);
INSERT INTO evaluations VALUES(55294,32,'cp -190',NULL);
INSERT INTO evaluations VALUES(55294,33,'cp -196',NULL);
INSERT INTO evaluations VALUES(55294,34,'cp -190',NULL);
INSERT INTO evaluations VALUES(55294,35,'cp -205',NULL);
INSERT INTO evaluations VALUES(55294,36,'cp -195','d2d3 g8f6 d1d2 c8f5 d3e4 d8d2 b1d2 f6e4 g1f3 e7e6 f1d3 e4d6 d3f5 d6f5 d2c4 b8c6 c1d2 h7h5 e1c1 f7f6 h1e1 e8f7 c2c3 a7a6 c4e3 f5d6 c1c2 a8e8 e3f1 c6e7 a2a4 e7f5 f1e3 d6e4 e3c4 e4f2 d1a1 f2g4');
INSERT INTO evaluations VALUES(9014,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(9014,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(9014,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(9014,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(9014,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(9014,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(9014,36,'cp 84','c2c4 e8g8 b1d2 h7h6 a1c1 f8e8 f3e5 c6e5 d4e5 f6e4 d2e4 d5e4 d1g4 d8e7 c4c5 f7f5 e5f6 e7f6 c1c3 f6f5 g4f4 c8d7 f1c4 f5f4 g3f4 d7c6 h1h5 a8d8 g2g4 c6d5 h5e5 c7c6 c4b3 a7a5 e1e2 g8f7 b3c2 d5a2 c2e4 a2d5 c3c1 d8d7 e4d5 e6d5');
INSERT INTO evaluations VALUES(15259,30,'cp 324',NULL);
INSERT INTO evaluations VALUES(15259,31,'cp 340',NULL);
INSERT INTO evaluations VALUES(15259,32,'cp 344',NULL);
INSERT INTO evaluations VALUES(15259,33,'cp 340',NULL);
INSERT INTO evaluations VALUES(15259,34,'cp 346',NULL);
INSERT INTO evaluations VALUES(15259,35,'cp 352',NULL);
INSERT INTO evaluations VALUES(15259,36,'cp 354','d2d3 e4f2 f3e5 f2d1 e5c6 f8e7 c6d8 e8d8 e1d1 c7c6 b5a4 c8g4 d1e1 h8e8 h2h3 g4h5 c1f4 e7c5 g1f1 e8e1 f1e1 d8d7 e1f1 a8e8 b1c3 a7a5 a1e1 e8e1 f1e1 b7b5 c3b5 c6b5 a4b5 d7d8 b5c6 d5d4 c6e4 f7f6 e4h7 h5f7 f4d2 a5a4 b2b3 g7g6 b3a4');
INSERT INTO evaluations VALUES(15257,30,'cp 376',NULL);
INSERT INTO evaluations VALUES(15257,31,'cp 354',NULL);
INSERT INTO evaluations VALUES(15257,32,'cp 359',NULL);
INSERT INTO evaluations VALUES(15257,33,'cp 344',NULL);
INSERT INTO evaluations VALUES(15257,34,'cp 360',NULL);
INSERT INTO evaluations VALUES(15257,35,'cp 364',NULL);
INSERT INTO evaluations VALUES(15257,36,'cp 363','f8b4 c2c3 e4f2 b5c6 b7c6 e1e5 b4e7 g1f2 e8g8 f2g1 a8b8 b1d2 e7d6 e5e1 c8g4 d2f1 a7a5 d3d4 f7f5 h2h3 g4h5 d1e2 f8e8 e2f2 h5f3 e1e8 d8e8 f2f3 e8e4 f3e4 d5e4 g1f2 c6c5 d4c5 d6c5 f2e2 c5d6 f1d2 c7c5 d2b3 c5c4 b3a5');
INSERT INTO evaluations VALUES(19980,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(19980,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(19980,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19980,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(19980,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19980,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(19980,36,'cp 12','g7g6 e2e3 f8g7 g1e2 e7e5 a2a3 c8f5 d2d3 g8e7 e1g1 e8g8 a1b1 a7a5 b2b3 d8d7 c3d5 e7d5 c4d5 c6b8 e3e4 f5h3 g2h3 d7h3 g1h1 b7b5 a3a4 b5a4 b3a4 b8a6 e2c3 h3d7 c3b5 a8b8 d1c2 a6c7 b5c7 d7c7 b1b8 f8b8');
INSERT INTO evaluations VALUES(19665,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19665,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(19665,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19665,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(19665,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(19665,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19665,36,'cp 27','e2e3 f8g7 g1e2 e7e5 d2d3 g8e7 a2a3 e8g8 a1b1 a7a5 e1g1 c8f5 b2b3 a8b8 c3b5 b7b6 e2c3 d8d7 d1c2 b8d8 c1b2 f5h3 g2h3 d7h3 c3d5 f7f5 d5e7 c6e7 f2f4 d8d7 d3d4 c5d4 e3d4 e5f4 f1f4 d6d5 b1e1 h3h5 f4f2 e7c6');
INSERT INTO evaluations VALUES(22755,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(22755,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(22755,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(22755,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(22755,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(22755,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(22755,36,'cp 69','d4d5 c6e7 e2e4 d7d6 a2a4 g8f6 f1c4 h7h6 g1e2 c7c6 e1g1 c6d5 c3d5 e7d5 c4d5 f6d5 d1d5 f8e7 e2c3 e8g8 c1e3 a7a5 c3b5 c8e6 d5b7 a8b8 b7a7 f7f5 e3d2 e6c4 e4f5 c4f1 a1f1 d6d5 d2a5 d8e8 b5c7 e8f7');
INSERT INTO evaluations VALUES(22659,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(22659,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(22659,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(22659,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(22659,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(22659,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(22659,36,'cp 63','c6e7 e2e4 d7d6 a2a4 g8f6 f1c4 h7h6 g1e2 c7c6 e1g1 c6d5 c3d5 e7d5 c4d5 f8e7 c1e3 e8g8 e2c3 c8e6 d1d3 f6d5 c3d5 f7f5 e4f5 e6f5 d3b3 f8f7 f2f4 f5e6 f4f5 e6d5 b3d5 d8d7 a4a5 a8c8 c2c3 a7a6 g2g4 d7b5 d5b5 a6b5 e3f2 d6d5 f2b6 e7c5 g1g2 c5b6');
INSERT INTO evaluations VALUES(28251,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(28251,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(28251,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(28251,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(28251,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(28251,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(28251,36,'cp 41','d7d5 e4d5 d8d5 e1g1 b4c3 b2c3 e8g8 c3c4 d5h5 f3e5 h5e2 d3e2 f8d8 f1d1 b8d7 e5d7 d8d7 a2a4 a7a5 f2f3 b7a6 e2d3 f6e8 a1a3 a8d8 c1e3 e8d6 a3c3 d6c8 h2h4 c8e7 d1b1 h7h6 h4h5 e7c6 c4c5 a6d3 c5b6 c6b4 b6c7 d8c8 c2d3 c8c7 c3c7 d7c7 e3d2 c7b7');
INSERT INTO evaluations VALUES(28211,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(28211,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(28211,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(28211,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(28211,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(28211,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(28211,36,'cp 50','e4d5 d8d5 e1g1 b4c3 b2c3 e8g8 c3c4 d5h5 f3e5 h5e2 d3e2 f8d8 f1d1 b8d7 e5d7 d8d7 a2a4 a7a5 f2f3 b7a6 e2d3 h7h6 h2h4 a8d8 c1e3 g8f8 a1b1 f6e8 e3f2 e8f6 d3f1 a6b7 c4c5 f6d5 c5b6 c7b6 d1d2 b7c6 c2c4 d5b4 b1a1 d8c8');
INSERT INTO evaluations VALUES(38554,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(38554,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(38554,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(38554,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(38554,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(38554,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(38554,36,'cp 91','g8f6 f1e2 e8g8 e1g1 e6e5 c1e3 f6g4 e3g5 f7f6 g5c1 b8a6 h2h3 g4h6 a1b1 h6f7 b2b4 c7c6 b4b5 a6c7 b5c6 b7c6 d1a4 f6f5 d4e5 f7e5 c4c5 d6c5 f3e5 g7e5 c1h6 f8f7 e2c4 c8e6 e4f5 g6f5 f1d1 d8e7 a4b3 e6c4 b3c4');
INSERT INTO evaluations VALUES(41609,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(41609,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(41609,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(41609,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(41609,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(41609,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(41609,36,'cp 99','f1e2 e8g8 e1g1 e6e5 c1e3 c8g4 d4d5 b8d7 f3d2 g4e2 d1e2 f6h5 g2g3 d8e8 c3b5 e8c8 g1h1 a7a6 b5c3 c8e8 d2f3 h5f6 a1d1 f6g4 e3c1 f7f6 e2c2 g4h6 h1g2 h6f7 b2b3 g7h6 c1b2 f7g5 f3h4 b7b5 b2a3 b5c4 b3c4');
INSERT INTO evaluations VALUES(41019,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(41019,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(41019,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(41019,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(41019,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(41019,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(41019,36,'cp 33','e2e4 c7c5 f1c4 c5d4 f3d4 d8a5 g5d2 a5c5 c4b5 c8d7 d4b3 c5c7 a1c1 e8g8 e1g1 d7c6 b5d3 b8d7 d1e2 a8d8 c3b5 c7b8 b5d4 d7e5 d4c6 e5c6 g2g3 d8d7 d3b1 f8d8 d2c3 c6e5 f1d1 d7d1 c1d1 d8d1 e2d1 e5c6 g1h1 b7b6');
INSERT INTO evaluations VALUES(41020,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(41020,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41020,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(41020,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(41020,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(41020,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41020,36,'cp 25','c7c5 f1c4 c5d4 f3d4 d8a5 g5d2 a5c5 c4b5 c8d7 d2e3 c5c7 e1g1 e8g8 a1c1 b8c6 d4c6 d7c6 b5c6 b7c6 d1e2 f8d8 c1c2 h7h6 f2f3 f6d7 c3a4 c7b7 b2b3 d7e5 a4c5 b7b5 e2f2 e5d3 c5d3 b5d3 c2c6 a8c8 c6c8 d8c8 e3a7 c8c2 f2e3');
INSERT INTO evaluations VALUES(47880,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(47880,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(47880,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(47880,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(47880,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(47880,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(47880,36,'cp -12','c7c5 c3c4 b8c6 b1c3 f8e7 f1e2 e8g8 e1g1 a7a6 c1d2 c8d7 a2a4 a8b8 c3d5 b7b6 h2h3 g8h8 d5e7 d8e7 f3e1 b6b5 a4b5 a6b5 b2b3 b8b6 f2f4 f8b8 f4e5 d6e5 d2g5 c6d4 e1f3 b5c4 f3d4 c5d4 b3c4 h7h6');
INSERT INTO evaluations VALUES(53014,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(53014,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(53014,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(53014,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(53014,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(53014,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(53014,36,'cp -6','g8f6 d2d3 b8c6 g1f3 e7e6 e1g1 d6d5 c4b5 c8d7 f1e1 a7a6 b5c6 d7c6 e4e5 f6d7 d3d4 f8e7 b1c3 e8g8 c3e2 c5d4 e2d4 d7c5 c1e3 a6a5 h2h3 a5a4 c2c3 d8b6 d1c2 h7h6 f3h2 b6c7 f2f4 c5e4 a1d1 f8d8 c2e2 e7c5');
INSERT INTO evaluations VALUES(45150,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(45150,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(45150,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(45150,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(45150,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(45150,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(45150,36,'cp -13','d2d3 e7e6 b1c3 d6d5 e4d5 e6d5 c4b3 f8e7 g1e2 e8g8 e1g1 c8e6 c1g5 h7h6 g5f6 e7f6 e2f4 f6e5 f4e6 f7e6 d1g4 d8d6 c3b5 d6d7 b5c3 d7f7 f2f4 e5d4 g1h1 b8c6 a1e1 a8e8 g4h3 g8h8 g2g4 f7f6 g4g5 f6f5 h3f5');
INSERT INTO evaluations VALUES(45149,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(45149,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(45149,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(45149,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(45149,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(45149,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(45149,36,'cp -19','e7e6 g1f3 f8e7 e1g1 e8g8 d1e2 a7a6 c4a2 b7b5 c1f4 b8d7 f1e1 f6h5 f4e3 d8c7 b1d2 h5f6 c2c3 c8b7 e3f4 e6e5 f4g5 h7h6 g5h4 a8e8 d2f1 g8h8 f1g3 d6d5 h4f6 d7f6 e4d5 f6d5 a2d5 b7d5 e2e5');
INSERT INTO evaluations VALUES(46007,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(46007,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(46007,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(46007,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(46007,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(46007,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(46007,36,'cp -13','g1f3 f8e7 e1g1 e8g8 d1e2 a7a6 c1f4 b7b5 c4b3 c8b7 h2h3 b8d7 b1d2 f8e8 c2c3 e7f8 a1d1 d8b6 f1e1 a8c8 b3a2 a6a5 f3h2 b7c6 f4g3 a5a4 h2g4 d6d5 e4d5 c6d5 a2b1 d5a8 d2e4 f6g4 h3g4');
INSERT INTO evaluations VALUES(46006,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(46006,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(46006,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(46006,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(46006,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(46006,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(46006,36,'cp -20','a7a6 e1g1 b7b5 c4b3 c8b7 f1e1 b8d7 h2h3 f8e7 c1f4 e8g8 b1d2 f8e8 c2c3 d8c7 a3a4 e7f8 d1e2 c7b6 f4g3 a8c8 a1a2 h7h6 e1a1 e6e5 a4b5 a6b5 c3c4 c8a8 a2a8 e8a8 a1a8 b7a8 c4b5 b6b5');
INSERT INTO evaluations VALUES(43499,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(43499,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(43499,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(43499,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(43499,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(43499,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(43499,36,'cp -6','c4b3 b7b5 e1g1 c8b7 f1e1 b8d7 h2h3 f8e7 c1f4 d7f8 b1d2 f8g6 f4h2 d8b6 c2c3 e8g8 d1e2 f8e8 d3d4 e7f8 d4c5 b6c5 e4e5 d6e5 f3e5 a8d8 d2f3 f8d6 a1d1 f6h5 a3a4 b7f3 e5f3 d6h2 g1h2 h5f4 e2e3 c5e3');
INSERT INTO evaluations VALUES(56597,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56597,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(56597,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(56597,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(56597,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56597,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(56597,36,'cp -20','d2d3 b8c6 g1f3 g8f6 e1g1 f8e7 c2c3 c8g4 c3d4 c6d4 b1d2 e8g8 h2h3 d4f3 d2f3 g4f3 d1f3 e7c5 c1g5 c5d4 a1c1 c7c6 c1c2 d8d6 g5f6 d6f6 f3f6 g7f6 g2f3 a7a5 g1g2 a5a4 f3g4 f8d8 f1b1 g8g7 g2f3 d8d6 b1c1 a4a3 b2b3 c6c5 c1b1 a8d8');
INSERT INTO evaluations VALUES(56587,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(56587,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(56587,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(56587,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(56587,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(56587,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(56587,36,'cp -10','g8f6 b1d2 b8c6 g1e2 h7h5 h2h3 g7g5 a2a3 a7a5 c2c3 d4c3 d2f3 g5g4 h3g4 c8g4 b2c3 f8c5 a1b1 d8d6 d3d4 e8c8 d1d3 g4f3 d3f3 e5d4 f3f5 f6d7 c1f4 c6e5 e2d4 c5d4 c3d4 e5d3 e1f1 d3f4 f5f4 d6d4');
INSERT INTO evaluations VALUES(48926,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48926,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48926,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48926,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48926,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48926,36,'cp 0','g1e2 b8c6 a2a3 a7a5 b1d2 f6d7 e1g1 f8e7 f2f4 f7f6 d2f3 a5a4 f3h4 g7g6 b2b4 a4b3 c2b3 e7d6 g2h3 d8e7 f4e5 c6e5 e2f4 e8g8 f4d5 e7g7 d5f4');
INSERT INTO evaluations VALUES(10699,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(10699,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(10699,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(10699,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(10699,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(10699,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(10699,36,'cp -72','d7d5 c4d3 e8g8 f1e1 c8f5 d1c2 f5g6 c1f4 f8e8 h2h3 b7b6 c2b2 c6a5 f3e5 c7c6 d3f1 e4f6 f2f3 f6d7 f4g3 d7e5 d4e5 b6b5 b2b4 a7a6 g3f2 a5c4 f1c4 d5c4 b4c5 d8d5 c5d5 c6d5 f2d4 g6d3 a2a4 h7h6 a4b5 a6b5');
INSERT INTO evaluations VALUES(9187,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(9187,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(9187,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(9187,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(9187,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(9187,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(9187,36,'cp -49','d1c2 e8g8 c4d3 f8e8 f1e1 c8f5 c1f4 f5g6 c2b2 c6a5 f3e5 c7c6 h2h3 e4f6 d3f1 f6d7 e5g6 h7g6 e1e8 d8e8 c3c4 e8e4 g2g3 d7f6 c4d5 c6d5 f4e5 b7b6 b2b5 a8d8 a1c1 a5c4 b5a6 c4d2 f1g2 d2f3 g2f3 e4f3 e5f6 f3f6 a6a7 f6d4 a7e7 d8f8 c1c6 d4a4 c6b6 d5d4 e7e4 f8d8 g1g2 a4a2');
INSERT INTO evaluations VALUES(13367,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(13367,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(13367,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(13367,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(13367,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13367,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(13367,36,'cp 25','e7e5 d4b3 c8e6 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5d3 c8c7 c2c3 f8e7 c1c2 d5e6 a2a4 f7f5 h1e1 d8d5 d3e2 d5d1 e1d1 h8d8 d1d8 c6d8 e2d3 g7g6 a4a5 d8c6 e3b6 c7c8 d3e2 e7g5 g2g3 g5d8 b6d8 c6d8');
INSERT INTO evaluations VALUES(13885,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(13885,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(13885,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13885,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(13885,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(13885,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(13885,36,'cp 22','d4b3 c8e6 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5d3 c8c7 c2c3 f8e7 c1c2 b7b5 d3e4 d5e4 f3e4 d8d1 h1d1 h8d8 d1a1 b5b4 a1f1 f7f6 h2h3 a6a5 b3c5 e7c5 e3c5 b4c3 b2c3 a5a4 f1b1 d8c8');
INSERT INTO evaluations VALUES(13888,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(13888,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13888,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13888,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(13888,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(13888,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13888,36,'cp 30','c8e6 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5d3 c8c7 c2c3 d5e6 c1c2 h7h6 h1e1 f8e7 a2a4 d8d7 d3e4 h8d8 d1d7 d8d7 b3d2 b7b5 a4b5 a6b5 e1a1 e7g5 e3g5 h6g5 a1a6 d7d6 e4c6 d6c6 a6a7 c7b6 a7e7 c6c7 e7e8 c7c8 e8c8 e6c8 d2e4 f7f6');
INSERT INTO evaluations VALUES(13565,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(13565,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13565,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(13565,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(13565,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(13565,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(13565,36,'cp 0','f8d6 d4c5 d6c5 f1e2 e8g8 e1g1 f8e8 b2b4 c5f8 b4b5 c6a5 b1d2 b7b6 c3c4 c8b7 c4d5 f6d5 f4e5 a8c8 a1c1 d8e7 d2c4 a5c4 e2c4 d5f6 d1e2 e8d8 e5b2 h7h6 f1d1 d8d1 c1d1 f6e4 f3e5 e7g5 e5f3');
INSERT INTO evaluations VALUES(8386,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(8386,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(8386,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(8386,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(8386,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(8386,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8386,36,'cp -6','d4c5 d6c5 b1d2 c5d6 f4d6 d8d6 f1b5 c8d7 e1g1 e8g8 a1c1 a7a6 b5a4 f8d8 c3c4 a8c8 a2a3 c6e5 f3e5 d6e5 a4d7 d8d7 c4d5 c8d8 d1b3 e5d5 b3d5 f6d5 d2e4 g8f8 c1c4 h7h6 f1d1 f8e7 d1c1 a6a5 e4g3 f7f5 e3e4 f5e4 g3e4');
INSERT INTO evaluations VALUES(15633,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(15633,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(15633,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(15633,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(15633,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(15633,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(15633,36,'cp 46','f3d4 f8e7 c1f4 e8g8 e1g1 c6e5 c4e2 f8e8 f1e1 e7f8 f4g3 c7c6 a2a4 a7a5 d1d2 d8b6 g1h1 e5g6 f2f3 c8d7 b2b3 d6d5 e4d5 f6d5 c3d5 c6d5 e2f1 e8e1 a1e1 f8c5 e1d1 a8c8 c2c3 c8d8 g3f2 h7h6');
INSERT INTO evaluations VALUES(24977,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(24977,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(24977,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(24977,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(24977,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(24977,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(24977,36,'cp 81','g1f3 e7e6 e1g1 f8b4 c1d2 e8g8 a2a3 b4d6 f3h4 c6e7 d1e2 e7g6 h4f5 e6f5 e2f3 c7c5 f3f5 c5d4 c3b5 d4e3 d2e3 d6e5 a1d1 d8c8 f5c8 a8c8 b2b3 a7a6 b5d6 c8c7 a3a4 h7h5 h2h3 g6h4 c4e2 e5d6 d1d6 h4f5 e3b6');
INSERT INTO evaluations VALUES(46880,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(46880,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46880,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46880,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(46880,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(46880,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(46880,36,'cp 23','g5f6 e7f6 e2e3 c8e6 b1d2 b8d7 c2c4 d5c4 f1c4 e6c4 d2c4 f8b4 e1e2 d7b6 d1b3 b4d6 h1c1 e8g8 e2f1 f8e8 c4d6 d8d6 a2a4 c7c6 a4a5 b6d5 e3e4 d5e7 g2g3 d6d7 a5a6 b7b6 c1c3 a8c8 a1c1 c6c5 d4c5 b6c5 b3b7');
INSERT INTO evaluations VALUES(46877,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(46877,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46877,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(46877,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(46877,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(46877,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46877,36,'cp 27','e7f6 e2e3 c8e6 b1d2 c7c6 c2c4 f8e7 d1b3 d8d7 f1d3 e8g8 e1g1 c6c5 f1d1 b8a6 c4d5 e6d5 d3c4 d5c4 b3c4 a8c8 d2e4 c5d4 c4d4 d7e6 e4c3 a6c5 d4f4 g7g5 f4g3 c5e4 c3e4 e6e4 f3d4 f8d8 g3f3 e4f3 g2f3 g8h7 d4f5 d8d1 a1d1');
INSERT INTO evaluations VALUES(47727,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(47727,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(47727,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47727,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(47727,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(47727,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(47727,36,'cp 16','e2e3 c8e6 b1d2 c7c6 c2c4 f8e7 f1d3 e8g8 e1g1 b8d7 d1b3 d5c4 d3c4 e6c4 d2c4 d7b6 a1c1 b6c4 c1c4 d8d7 f1c1 f8e8 a2a4 a7a5 f3e1 g7g6 h2h4 h6h5 e1d3 e7d6 d3c5 d6c5 c4c5 e8d8 b3c4 d7e7 c1c3 g8g7');
INSERT INTO evaluations VALUES(47726,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(47726,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(47726,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(47726,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(47726,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(47726,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47726,36,'cp 21','c8e6 b1d2 c7c6 f1d3 a7a5 e1g1 f8e7 c2c4 e8g8 d1c2 b8d7 a1c1 f8e8 f1d1 e7f8 d3h7 g8h8 h7f5 e6f5 c2f5 g7g6 f5c2 d5c4 d2c4 b7b5 c4d2 a8c8 d2b3 h8g7 b3c5 d8c7 a2a4 f8d6 b2b3 d6c5 d4c5 b5a4 b3a4 d7e5 f3d4 c7b7');
INSERT INTO evaluations VALUES(50499,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(50499,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(50499,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(50499,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(50499,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(50499,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(50499,36,'cp -21','g1f3 d7d5 e4d5 e6d5 c4e2 f8d6 e1g1 g8f6 d2d4 c5d4 f3d4 b8c6 b1c3 e8g8 c1e3 f8e8 d1d2 d8c7 h2h3 c8e6 a1d1 a8d8 f1e1 h7h6 e2f1 e6c8 g2g3 d6e5 f1g2 f6e4 d2d3 c6d4 e3d4 e5d4 d3d4 e4c3 e1e8 d8e8 d4c3 c7c3 b2c3');
INSERT INTO evaluations VALUES(57502,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(57502,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(57502,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(57502,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(57502,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(57502,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(57502,36,'cp 32','g8f6 g1f3 b8c6 e1g1 f8e7 b1c3 e8g8 d2d3 d7d6 a1b1 a7a5 a2a3 d8e8 b2b4 a5b4 a3b4 c6d8 e2e3 d8f7 c1b2 c7c6 b4b5 c8e6 b5c6 b7c6 c3e2 g7g5 d1c2 g5g4 f3d2 f7g5 d3d4 e5e4 d4d5 c6d5 e2f4');
INSERT INTO evaluations VALUES(49996,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(49996,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(49996,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(49996,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(49996,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(49996,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(49996,36,'cp 32','g1f3 b8c6 e1g1 f8e7 b1c3 e8g8 a1b1 a7a5 d2d3 d7d6 a2a3 d8e8 b2b4 a5b4 a3b4 c6d8 e2e3 d8f7 c1b2 c7c6 b4b5 c8e6 d1c2 d6d5 c3e2 d5c4 f3e5 f7e5 b2e5 c6b5 e5f6 e7f6 e2f4 a8d8 f4e6 e8e6 d3c4 b5c4 b1b7 g8h8 b7c7');
INSERT INTO evaluations VALUES(873,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(873,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(873,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(873,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(873,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(873,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(873,36,'cp 164','e6d7 d2d4 e7e6 f1e1 f8e7 e2b5 e8g8 f3e5 d7e8 b5c6 b7c6 c1g5 c8b7 d1d3 a8d8 d3c4 h7h6 g5h4 f6d7 h4e7 e8e7 e5d7 d8d7 c3e4 d7d5 c2c3 f8d8 e4c5 b7c8 f2f4 e7h4 g2g3 h4h5 e1e3 h5f5 a1e1 a7a5 b2b3 f5g4 c4e2 g4e2 e3e2 g7g5 f4g5');
INSERT INTO evaluations VALUES(12961,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(12961,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12961,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(12961,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(12961,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(12961,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(12961,36,'cp 21','c5d4 c3d4 e7f5 d3f5 e6f5 b1c3 c8e6 h2h4 d8b6 e1f1 f8e7 c3e2 f5f4 e2f4 e6g4 c1e3 b6b2 f1g1 e8g8 f4d5 f8d8 d5e7 c6e7 d1b1 b2b1 a1b1 d8d7 a2a4 h7h6 a4a5 d7c7 g1h2 g4e6 h1c1 a8c8 c1c7 c8c7 f3d2 e7d5');
INSERT INTO evaluations VALUES(12981,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12981,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12981,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12981,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12981,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(12981,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(12981,36,'cp 32','c3d4 e7f5 d3f5 e6f5 h2h4 c8e6 b1c3 d8b6 e1g1 h7h6 c3a4 b6b5 f1e1 f8e7 c1f4 e8g8 e1e3 f8c8 e3d3 c6b4 d3d2 b5d7 a4c3 d7d8 h4h5 d8d7 c3e2 b4c6 f4e3 a7a5 a1b1 a5a4 e2f4 a4a3 b2b3');
INSERT INTO evaluations VALUES(12975,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(12975,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12975,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12975,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12975,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(12975,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(12975,36,'cp 11','e7f5 d3f5 e6f5 b1c3 f8e7 h2h4 e8g8 c1f4 c8e6 h4h5 h7h6 e1g1 d8b6 d1d2 f8c8 f1c1 c6a5 b2b3 e7b4 d2d3 c8c3 c1c3 b4c3 d3c3 a5c6 c3d3 a8c8 d3d2 g8h7 a1c1 c6e7 c1b1 e7c6 f4e3 b6d8 b1c1 a7a6 e3f4');
INSERT INTO evaluations VALUES(13480,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(13480,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13480,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(13480,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13480,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(13480,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(13480,36,'cp 22','d3f5 e6f5 b1c3 f8e7 h2h4 e8g8 c1f4 c8e6 d1d2 d8b6 e1g1 f8c8 h4h5 h7h6 f1c1 c6a5 c3a4 b6b5 b2b3 e7a3 c1d1 a3b4 d2e3 c8c2 f3e1 b4e1 e3e1 a8c8 d1d2 b7b6 d2c2 c8c2 e1d1 b5e2 d1e2 c2e2 a4c3 e2c2 a1c1 c2c1 f4c1 a5c6 c3e2 c6b4');
INSERT INTO evaluations VALUES(13477,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(13477,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13477,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(13477,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13477,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13477,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(13477,36,'cp 21','e6f5 b1c3 f8e7 c3e2 e8g8 c1d2 a7a5 e1g1 a5a4 a2a3 g7g5 h2h4 h7h6 f3h2 f7f6 h4g5 h6g5 f2f4 g5g4 d2c3 f8f7 f1f2 f7h7 h2f1 g8h8 f1e3 c8e6 e2g3 f6e5 g3f5 e6f5 e3f5 e5e4 d1g4 d8g8');
INSERT INTO evaluations VALUES(13826,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(13826,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(13826,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13826,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(13826,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13826,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(13826,36,'cp 19','b1c3 f8e7 h2h4 e8g8 d1b3 c8e6 h4h5 h7h6 e1g1 d8d7 c3e2 a7a5 c1d2 a5a4 b3e3 a4a3 b2b3 f5f4 e2f4 e6f5 a1c1 f8e8 f4e2 f7f6 e2g3 f5g4 f3h2 f6e5 d4e5 e7h4 h2g4 d7g4 e3e2 g4e2 g3e2 e8e5');
INSERT INTO evaluations VALUES(18747,30,'cp -187',NULL);
INSERT INTO evaluations VALUES(18747,31,'cp -218',NULL);
INSERT INTO evaluations VALUES(18747,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(18747,33,'cp -179',NULL);
INSERT INTO evaluations VALUES(18747,34,'cp -209',NULL);
INSERT INTO evaluations VALUES(18747,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(18747,36,'cp -201','g8f6 e1g1 f8e7 c4b3 c6d4 f3d3 d7d5 b3a4 c8d7 a4d7 d8d7 e4d5 f6d5 c2c3 d4c6 b1a3 e8g8 d3e2 a8d8 d2d3 a7a6 a3c2 f8e8 c1g5 f7f6 g5e3 e6e5 a1d1 e7d6 e2f3 d5e7 g1h1 e7g6 f3g3 b7b5 f2f3 f6f5 f3f4');
INSERT INTO evaluations VALUES(12375,30,'cp -192',NULL);
INSERT INTO evaluations VALUES(12375,31,'cp -191',NULL);
INSERT INTO evaluations VALUES(12375,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(12375,33,'cp -166',NULL);
INSERT INTO evaluations VALUES(12375,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(12375,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(12375,36,'cp -185','e1g1 f8e7 c4b3 c6d4 f3d3 d7d5 b3a4 c8d7 a4d7 d8d7 e4d5 f6d5 c2c3 d4c6 d3e2 e8g8 b1a3 a8d8 d2d3 a7a6 a3c2 h7h6 f2f4 f8e8 c1d2 e7d6 a1e1 d6c7 e2f3 f7f5 d2e3 d5e3 e1e3 e6e5 g1h1 c5c4 d3d4 e5d4 e3e8 d8e8 c3d4 c6d4 c2d4 d7d4 f3b7');
INSERT INTO evaluations VALUES(26189,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(26189,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(26189,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(26189,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(26189,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26189,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(26189,36,'cp 8','e7e6 e1g1 g8f6 d1e2 g6d3 e2d3 f8d6 c2c4 e8g8 b2b3 f8e8 c1b2 d8c7 g3e4 f6e4 d3e4 e6e5 d4e5 d7e5 g2g3 e5f3 e4f3 d6e5 b2c3 e5c3 f3c3 a8d8 a1e1 g7g6 h2h4 h7h5 e1e8 d8e8 f1d1 c7e7 c3d4 a7a5 g1g2 e7e2 d1d2 e2g4 d4g4 h5g4 d2d7 b7b5 c4c5');
INSERT INTO evaluations VALUES(26093,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(26093,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(26093,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(26093,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(26093,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(26093,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26093,36,'cp 7','e1g1 g8f6 d1e2 g6d3 e2d3 f8d6 c2c4 e8g8 g3e4 f6e4 d3e4 f8e8 c1g5 d8a5 f1e1 h7h6 g5d2 d6b4 d2b4 a5b4 e4c2 b4b6 a1d1 a8d8 g2g3 c6c5 a2a3 c5d4 d1d4 d7f6 d4d1 d8d1 e1d1 e6e5 b2b4 e5e4 f3d4 e8d8');
INSERT INTO evaluations VALUES(37163,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(37163,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37163,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(37163,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(37163,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37163,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(37163,36,'cp 36','e2e3 c7c5 f1d3 c5d4 e3d4 h7h6 g5h4 d7d5 c4d5 e6d5 a1c1 f8e8 g1e2 b8c6 e1g1 b4e7 d3b1 f6h5 h4e7 c6e7 e2g3 h5f6 f1e1 c8d7 g3h5 f6h5 d1h5 a8c8 e1e2 c8c4 c1d1 d8b6 c3d5 e7d5 e2e8 d7e8 h5d5 b6b4');
INSERT INTO evaluations VALUES(37161,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(37161,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(37161,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(37161,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(37161,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37161,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37161,36,'cp 12','c7c5 f1d3 c5d4 e3d4 h7h6 g5h4 b8c6 g1f3 d7d5 c4d5 e6d5 e1g1 g7g5 h4g3 c8g4 g3e5 b4c3 b2c3 f8e8 f1e1 c6e5 d4e5 f6d7 d3e2 d8c7 d1d4 g4f3 e2f3 e8e5 f3d5 a8e8 e1e5 e8e5 a1f1 e5f5 d4e4 f5f6 c3c4 d7b6 h2h4 c7f4 d5b7 b6c4 h4g5 f4e4 b7e4');
INSERT INTO evaluations VALUES(37231,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37231,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(37231,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37231,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(37231,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37231,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(37231,36,'cp 0','g1f3 d7d6 c3a4 c5b4 c2c3 d6d5 e4d5 b4e7 b2b4 a7a6 c4b3 b7b5 a4c5 a6a5 e1g1 a5a4 b3c2 d8d5 c1e3 b8d7 c5d7 c8d7 c3c4 b5c4 d3c4 d5d1 a1d1 e7b4 f3e5 d7e6 e5c6 b4a3 c2d3 e6d7 c6e5 d7e6');
INSERT INTO evaluations VALUES(37228,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37228,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(37228,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37228,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37228,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37228,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37228,36,'cp 5','d7d6 c3a4 c5b4 c2c3 d6d5 e4d5 b4e7 b2b4 f6d5 e1g1 c7c6 c4b3 a7a5 b4a5 d8a5 c1d2 c8g4 h2h3 g4f3 d1f3 b8d7 a4b2 d7c5 b3d5 c6d5 f3d5 f8d8 d5e5 e7f6 e5e2 a5a6 a1b1 c5d3 b2d3 a6d3 e2d3 d8d3 d2e3 d3c3 b1b7 c3a3 f1d1 a3a2 d1d7 a2a1 g1h2');
INSERT INTO evaluations VALUES(36993,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36993,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(36993,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(36993,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36993,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36993,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(36993,36,'cp 0','c3a4 c5b4 c2c3 d6d5 e4d5 b4e7 b2b4 a7a6 c4b3 b7b5 a4c5 a6a5 e1g1 a5a4 b3c2 d8d5 c1e3 b8d7 d3d4 e5d4 c5d7 c8d7 e3d4 d7g4 f1e1 g4f3 d1f3 d5f3 g2f3 f8e8 a1d1 c7c6 a2a3 g8f8 c2d3 a8d8 d3f1 d8d7 f3f4 f6d5 c3c4 b5c4 f1c4 e8d8');
INSERT INTO evaluations VALUES(45751,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(45751,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(45751,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(45751,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(45751,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(45751,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(45751,36,'cp -13','f8d6 f1d3 e8g8 c3e2 f8e8 h2h3 h7h6 g5h4 d8a5 d1d2 a5d2 f3d2 f6e4 d3e4 d5e4 h4g3 d6f8 d2c4 b8d7 c4e3 d7f6 c2c4 b7b5 c4b5 c6b5 e2c3 c8d7 a2a4 b5a4 e1g1 a4a3 b2a3 a8c8 f1c1 f6h5 g3h2 f7f5 e3d5 g7g5 d5c7 e8d8 c3b5 d7b5 c7b5 h5f4 c1c8');
INSERT INTO evaluations VALUES(53025,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53025,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(53025,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(53025,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(53025,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53025,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53025,36,'cp 22','g7g6 e2e3 b8c6 f1g2 e7e5 g1e2 f8g7 a2a3 g8e7 b2b4 c5b4 a3b4 c6b4 e1g1 a7a5 c1a3 e7c6 c3d5 c8f5 a3b4 c6b4 d5b4 a5b4 d1b3 f5g4 a1e1 d8a5 h2h3 g4e6 e1b1 e8g8 b3b4 e6f5 b4a5 a8a5 b1b4 f5d3 f1e1 e5e4 b4b7');
INSERT INTO evaluations VALUES(1136,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(1136,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(1136,32,'cp 269',NULL);
INSERT INTO evaluations VALUES(1136,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(1136,34,'cp 266',NULL);
INSERT INTO evaluations VALUES(1136,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(1136,36,'cp 232','c1e3 e7d6 e1c1 d6d4 e3d4 b7b6 e4e5 f6g8 c3d5 e8d8 f1c4 g8e7 d5e3 d8e8 f2f4 a8b8 h1e1 b6b5 c4f1 a7a6 a2a3 e7c6 d4c3 e8d8 g2g3 a6a5 c3d2 f8c5 e3f5 b5b4 a3b4 a5b4 f5g7 b8a8 c1b1 c6d4 d2e3 a8a1 b1a1 d4c2 a1a2 c2e3 e1e3 c5e3 g7f5 e3a7 f1c4 d8e8 a2b3 c8b7');
INSERT INTO evaluations VALUES(5633,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(5633,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(5633,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(5633,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(5633,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(5633,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(5633,36,'cp 76','d1b3 g8h6 d2d4 e8g8 c1h6 g7h6 e1g1 g8h8 b1d2 c6a5 b3a4 a5c4 a4c4 e7g5 f3g5 h6g5 f1d1 d8e7 d2f1 f7f6 f1e3 e7f7 c4f7 f8f7 a1c1 c8d7 f2f3 a8e8 c3c4 f6f5 e4f5 d7f5 c4c5 f5g6 c5d6 c7d6');
INSERT INTO evaluations VALUES(18360,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(18360,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(18360,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(18360,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(18360,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(18360,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(18360,36,'cp 81','d2d4 e7e6 a2a3 d7d5 c2c4 g8f6 c4d5 e6d5 c1g5 f8e7 b1c3 c8b7 g5f6 e7f6 c3d5 f6d4 d1c2 e8g8 a1d1 g7g6 e2e4 d4g7 g1e2 c6e5 e1g1 c7c6 d5e3 d8e7 f2f4 g7h6 f1e1 a8d8 e2c3 d8d1 c3d1 e5d7 d1f2 d7c5 e4e5 h6g7 b2b4 c5e6 c2c3 f8d8 g2c6 e6d4');
INSERT INTO evaluations VALUES(29240,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29240,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29240,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(29240,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29240,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29240,36,'cp 0','h7h6 e1g1 f8e7 b2b3 e8g8 e2d3 f8e8 c1b2 b8d7 a2a4 f5d3 d1d3 a7a5 d3c2 a8c8 a1c1 e7b4 f1d1 d8e7 h2h3 c8a8 c3e2 f6e4 e2c3');
INSERT INTO evaluations VALUES(29196,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29196,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29196,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29196,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29196,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29196,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29196,36,'cp 0','e1g1 f8e7 b2b3 e8g8 e2d3 f8e8 c1b2 b8d7 a2a4 f5d3 d1d3 a7a5 d3c2 a8c8 a1c1 e7b4 f1d1 d8e7 h2h3 c8a8 c3e2 f6e4 e2c3');
INSERT INTO evaluations VALUES(29195,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29195,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29195,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29195,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29195,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29195,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29195,36,'cp 0','f8e7 b2b3 e8g8 e2d3 f8e8 c1b2 b8d7 a2a4 f5d3 d1d3 a7a5 d3c2 a8c8 a1c1 e7b4 f1d1 d8e7 h2h3 c8a8 c3e2 f6e4 e2c3');
INSERT INTO evaluations VALUES(43019,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(43019,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(43019,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(43019,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(43019,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(43019,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(43019,36,'cp -13','b8c6 e5f4 e8g8 f1e2 d7d5 d2d3 e7b4 c1d2 f8e8 b1c3 d5e4 d3e4 b4c3 d2c3 f6e4 a1d1 d8e7 e1g1 g7g5 f4c1 e4c3 b2c3 e7f6 e2d3 h7h6 f3d4 c8g4 f2f3 g4h5 d3e4 h5g6 e4g6 f6g6 f3f4 c6d4 d1d4 a8d8 f4g5 d8d4 c3d4 e8e2 c1f4 e2c2 f4f3 h6g5 f3b7 c2a2');
INSERT INTO evaluations VALUES(10871,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(10871,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(10871,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(10871,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(10871,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(10871,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(10871,36,'cp -9','e5f4 e8g8 d2d3 d7d5 f1e2 d5e4 d3e4 f8e8 b1c3 e7b4 c1d2 b4c3 d2c3 f6e4 a1d1 d8e7 e1g1 g7g5 f4c1 e4c3 b2c3 e7f6 e2c4 h7h6 f3d4 c6d4 c3d4 c8e6 c4b3 a8d8 c1d2 e8e7 c2c3 d8e8 d1e1 c7c6 c3c4 e6f5 e1e3 f5g6 c4c5 e7e3 f2e3 f6e7 f1e1 e8d8');
INSERT INTO evaluations VALUES(44199,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(44199,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(44199,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44199,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(44199,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(44199,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(44199,36,'cp 31','g1f3 d7d5 c4d5 e6d5 d2d4 b8c6 f1g2 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 f6e4 c5d4 a7a5 f2f3 e4d6 d1d2 a8b8 e2e4 d5e4 d4f2 e4f3 g2f3 d7h3 f1d1 d8f6 f3c6 b8b2 d2d4 f6d4 f2d4');
INSERT INTO evaluations VALUES(48786,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48786,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(48786,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48786,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(48786,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48786,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48786,36,'cp -14','e4d5 c7c6 b1c3 e5e4 f3e2 c6d5 c4d5 f6d5 e2b5 b8c6 b5d5 d8d5 c3d5 f8d6 g1e2 d6e5 f2f4 e4f3 g2f3 c8h3 e1f2 e8c8 d5f4 h3f5 d2d3 h7h6 a1b1 g7g5 f4g2 f5h3 b2b4 h8e8 h1e1 c6d4 e2d4 e5d4 c1e3 h3g2 e3d4 e8e1 b1e1 d8d4 f2g2');
INSERT INTO evaluations VALUES(48591,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(48591,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(48591,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(48591,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(48591,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(48591,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(48591,36,'cp -8','e5e4 f3d1 f8c5 b1c3 e8g8 d2d3 e4d3 c2d3 c7c6 d3d4 c6d5 c4b3 c5b6 g1e2 b8c6 e1g1 c8e6 c1e3 c6a5 e2f4 d8d6 c3b5 d6c6 b5c3 a8c8 a1c1 c6d6 h2h3 h7h6 f4e6 d6e6 f1e1 a5b3 d1b3 e6f5 c3e2');
INSERT INTO evaluations VALUES(48561,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(48561,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48561,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48561,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48561,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48561,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(48561,36,'cp -12','f3d1 f8c5 b1c3 e8g8 d2d4 e4d3 c2d3 b8d7 h2h3 f8e8 g1e2 d7b6 e1g1 f6d5 c3d5 b6d5 d1b3 c7c6 d3d4 c5d6 e2c3 c8e6 c3d5 c6d5 c4b5 e8e7 c1g5 f7f6 g5d2 d8b6 d2c3 a7a6 b5a4 b6b3 a4b3 e6f7 a1e1 e7c7 g2g3 g8f8 g1g2 a6a5 a2a4 b7b6 e1e3 a8c8 f1d1 h7h6 d1e1');
INSERT INTO evaluations VALUES(4281,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4281,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4281,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4281,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4281,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(4281,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4281,36,'cp 1','c5e7 e1g1 d7d6 a2a4 g8h8 b1d2 f6h5 g5e3 f7f5 e4f5 c8f5 d3d4 e5d4 f3d4 c6d4 c3d4 h5f6 d2f3 f6g4 d1b3 f5e4 f3d2 e4g6 a1e1 c7c6 c4d3 g6d3 b3d3 g4f6 d2c4 f6d5 e3d2 d8d7 b2b3 a8e8 a4a5 d5c7 h2h3');
INSERT INTO evaluations VALUES(7001,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(7001,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(7001,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(7001,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(7001,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(7001,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(7001,36,'cp 52','d1d4 g8f6 e4e5 d7d5 e5f6 d5c4 d4d8 e7d8 f6g7 h8g8 c1h6 c8e6 b1c3 d8f6 a1d1 f6g7 h6g7 g8g7 e1g1 e6h3 f1e1 e8f8 g2g3 g7g6 d1d5 g6e6 e1e4 a8e8 f2f3 e6b6 e4e8 f8e8 d5b5 b6b5 c3b5 e8d7 c2c3 c7c6 b5d4 c6c5 d4c2 h3e6 g1f2 d7d6 c2e3 d6e5 e3g2 f7f5 f2e2 e6f7 e2e3 b7b6 a2a3 h7h6');
INSERT INTO evaluations VALUES(27964,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(27964,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(27964,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(27964,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(27964,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(27964,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(27964,36,'cp 39','d1e2 g8f6 e4e5 d7d5 e5f6 d5c4 f6g7 h8g8 e2c4 b8c6 b2b3 d8f6 c1b2 e8c8 e1c1 f6g7 h1g1 b4e7 c3b5 e7f6 b2f6 g7f6 c4c3 f6e7 c3e3 e6e5 d1e1 g8e8 c1b2 f7f5 d2d3 c8b8 b5c3 e7f8 e3g5 h7h6 g5h4 f8h8 h2h3 h8g7 e1d1');
INSERT INTO evaluations VALUES(34771,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(34771,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(34771,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(34771,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(34771,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(34771,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(34771,36,'cp 99','c8e6 c3d5 f6d8 e1g1 g8f6 f1e1 c7c6 d5e3 e6c4 e3c4 b7b5 c4e3 b8d7 d2d4 f8e7 d4e5 d6e5 e3f5 e8g8 d1d2 f8e8 d2g5 g7g6 f5h6 g8g7 g5h4 f6h5 f3g5 e7g5 c1g5 f7f6 g5e3 d7f8 g2g3 d8c7 h4g4 a7a5 e1d1 a8d8 a2a4 d8d1 a1d1 b5a4 h6f5 g7g8');
INSERT INTO evaluations VALUES(36937,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(36937,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(36937,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(36937,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(36937,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(36937,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(36937,36,'cp -24','c8g4 c3b5 f8e8 c1g5 h7h6 g5h4 d6f4 b5c3 c7c6 h2h3 g4h5 g2g4 h5g6 d3g6 f7g6 d1d3 g6g5 c3e2 f4c7 h4g3 b8d7 a1e1 d7f8 g3e5 f8e6 c2c3 c7e5 f3e5 f6d7 e5g6 d8f6 f2f4 g5f4 g6f4 e6f4 f1f4 f6g5 e1f1');
INSERT INTO evaluations VALUES(27878,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(27878,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(27878,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(27878,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(27878,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(27878,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(27878,36,'cp -16','h2h3 g4h5 c3b5 f8e8 b5d6 d8d6 g2g4 h5g6 f3e5 g6d3 e5d3 b8d7 c2c3 d6b6 c1f4 c7c5 d4c5 d7c5 d3c5 b6c5 f4g5 f6e4 g5e3 c5c6 d1d4 c6b5 b2b3 h7h6 a1d1 b5a6 c3c4 d5c4 d4c4 b7b5 c4d3 a8d8 d3d8 e8d8 d1d8 g8h7');
INSERT INTO evaluations VALUES(27877,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(27877,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(27877,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(27877,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(27877,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(27877,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(27877,36,'cp -17','g4h5 c3b5 f8e8 b5d6 d8d6 g2g4 h5g6 f3e5 g6d3 e5d3 d6b6 c2c3 b8d7 c1f4 c7c5 d4c5 d7c5 d3c5 b6c5 f4g5 f6d7 g5f4 c5c6 f1e1 h7h6 d1d4 d7f8 e1e8 a8e8 d4d2 c6b5 b2b3 e8d8 f4c7 d8c8 c7f4 b5c6 a1c1 b7b6 f4e3');
INSERT INTO evaluations VALUES(27876,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(27876,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(27876,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(27876,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(27876,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(27876,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(27876,36,'cp -15','c3b5 f8e8 c1g5 c7c6 b5d6 d8d6 g5f6 d6f6 g2g4 h5g6 d3g6 h7g6 d1d3 b8d7 a1e1 d7f8 c2c3 f6f4 f3e5 g6g5 a2a4 a7a5 d3c2 c6c5 c2f5 c5d4 c3d4 a8c8 e5f3 f4f5 g4f5 f7f6 e1e8 c8e8 f1c1 e8e2 c1c5 b7b6 c5d5 e2b2 d5d8 g8f7 d4d5 b6b5 a4b5');
INSERT INTO evaluations VALUES(39558,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(39558,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(39558,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(39558,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(39558,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(39558,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(39558,36,'cp 56','d2d4 e7e6 b1c3 g8e7 g1f3 e8g8 f1d3 d7d6 h2h4 e6e5 d4e5 d6e5 h4h5 c8g4 h5g6 e7g6 d3e2 b8c6 c3d5 g4f3 e2f3 c6d4 e1f1 c7c6 d5e3 d4f3 d1f3 d8d3 f1g1 g6f4 g1h2 a8d8 e3f5 d3f3 g2f3 f4e6 h2g3 d8d3 c1e3 f8d8 b2b4');
INSERT INTO evaluations VALUES(39620,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(39620,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(39620,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(39620,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(39620,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(39620,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(39620,36,'cp -25','c7c5 c2c3 g8f6 g2g3 e8g8 f1g2 b8c6 e1g1 d7d6 a2a4 e7e5 c3c4 a8b8 b1c3 a7a6 c1g5 h7h6 g5d2 b7b6 a1b1 c6b4 f3e1 g8h7 d2c1 f6g8 e1c2 g8e7 c2b4 c5b4 c3e2 f7f5 e4f5 e7f5 c1e3 c8b7 g2b7');
INSERT INTO evaluations VALUES(37823,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(37823,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(37823,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(37823,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(37823,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(37823,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(37823,36,'cp -17','g2g3 g8f6 f1g2 e8g8 b1d2 d7d6 e1g1 b8c6 a2a4 h7h6 d2c4 c8e6 d1e2 d8d7 c1d2 a8d8 d2c3 b7b6 a1e1 d7c7 f3d2 d6d5 e4d5 e6d5 f2f4 e7e6 c3e5 c7d7 c2c3 d5g2 e2g2 d7d5 g2f3 c6e7 e5c7 d8d7 f3d5 f6d5 c7e5 f8d8');
INSERT INTO evaluations VALUES(37824,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(37824,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(37824,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(37824,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(37824,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(37824,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(37824,36,'cp -12','g8f6 f1g2 e8g8 e1g1 d7d6 c2c3 b8c6 a2a4 h7h6 h2h3 e7e5 c3c4 g8h7 b1c3 f6g8 g1h2 g8e7 f3e1 b7b6 e1c2 a8b8 c1d2 b8b7 a1b1 a7a5 d1e2 c6b4 c2e3 e7c6 c3b5 h6h5 e3d5 b4d5 c4d5 c6d4 b5d4 e5d4');
INSERT INTO evaluations VALUES(50152,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50152,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50152,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(50152,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50152,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50152,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50152,36,'cp -1','d7d5 e2e3 f8g7 g1f3 e8g8 f1e2 h7h6 g5f4 c7c5 f3e5 b8c6 e1g1 c5d4 e3d4 c8f5 f1e1 e7e6 e5c6 b7c6 c3a4 h6h5 a4c5 d8b6 c5a4 b6b7 c2c3 f6d7 d1d2 f8e8 b2b4 e6e5 d4e5 g7e5');
INSERT INTO evaluations VALUES(46746,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46746,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46746,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46746,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46746,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46746,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46746,36,'cp 0','d1d2 h7h6 g5h4 f8g7 f2f3 e8g8 e1c1 c7c5 d4c5 b8c6 e2e3 e7e6 g1e2 b7b6 c5b6 a7b6 c1b1 b6b5 e2d4 c6d4 d2d4 c8d7 d4b4 d8b6 b4d4');
INSERT INTO evaluations VALUES(56866,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(56866,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(56866,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(56866,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(56866,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(56866,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(56866,36,'cp 76','d7d5 e4d5 e6d5 b1c3 c7c6 c4d5 c6d5 g1f3 f8g7 f1b5 b8c6 e1g1 g8e7 b5c6 b7c6 h2h3 e8g8 c1f4 f8e8 f1e1 c8f5 c3a4 f5e4 f3e5 f7f6 e5d3 e7f5 d3c5 f5h4 f2f3 g6g5 f3e4 g5f4 d1h5 h4g6 e4d5 d8d5 h5d5 c6d5 g1f1 f4f3 g2f3 g6h4');
INSERT INTO evaluations VALUES(16230,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(16230,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(16230,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(16230,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(16230,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(16230,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(16230,36,'cp -36','d7c6 f1e1 e4d6 f3e5 f8e7 d2d4 e8g8 c1f4 f8e8 b1d2 e7f8 c2c3 f7f6 e5c4 c8e6 c4e3 d8d7 d1f3 e6f7 f4d6 c7d6 d2e4 e8e7 e4g3 a8e8 a2a4 e7e6 d4d5 e6e5 e1d1 d7c8 a4a5 h7h5 g3f5 c6d5 f5g3 g7g6 f3f6');
INSERT INTO evaluations VALUES(18839,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(18839,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18839,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18839,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18839,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(18839,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(18839,36,'cp 33','d2d4 d7d5 e4d5 e6d5 c2c4 c5d4 e1g1 g8f6 f3d4 f8e7 b1c3 e8g8 c1e3 c6d4 e3d4 d5c4 e2c4 c8g4 d1a4 a7a6 d4f6 e7f6 c4f7 f8f7 a4g4 d8b6 c3e4 f6b2 a1b1 f7e7 f1d1 b6c6 e4d6 b2e5 g4c4 c6c4 d6c4 a8e8 c4e5 e7e5 b1b7');
INSERT INTO evaluations VALUES(18804,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(18804,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18804,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(18804,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(18804,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(18804,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(18804,36,'cp 28','d7d5 e4d5 e6d5 c2c4 c8e6 b1c3 g8f6 c4d5 f6d5 e1g1 f8e7 d4c5 d5c3 b2c3 d8d1 f1d1 e7c5 f3g5 a8d8 g5e6 d8d1 e2d1 f7e6 c1f4 c6e7 d1a4 e8f7 a1e1 h8d8 a4b3 e7d5 f4d2 c5e7 g2g3 e7f6 g1g2 b7b6 h2h4 h7h6 g2f3 f6c3 d2c3 d5c3');
INSERT INTO evaluations VALUES(21314,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(21314,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(21314,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(21314,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(21314,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(21314,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(21314,36,'cp 122','f3d4 g8f6 d4c6 b7c6 e4e5 f6g4 e5e6 d6d5 d1g4 c8e6 g4f3 d5c4 f3c6 e6d7 c6c4 f8e7 f1d1 e8g8 b1c3 d7e6 c4c6 e7d6 c1f4 d8f6 f4d6 c7d6 c6d6 a8b8 a1b1 f8d8 d6g3 d8d1 c3d1 b8e8 g3e3 e8e7 e3d2 e6a2 b1c1 a2d5 d1e3 d5c6 c2c4 e7d7 d2c3 f6c3 c1c3 d7d4 f2f3 d4d2 b2b3');
INSERT INTO evaluations VALUES(21304,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(21304,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(21304,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(21304,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(21304,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(21304,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(21304,36,'cp 108','g8f6 d4c6 b7c6 e4e5 f6g4 e5e6 d6d5 d1g4 c8e6 g4f3 d5c4 f3c6 e6d7 c6c4 f8e7 f1d1 e8g8 b1c3 d7e6 c4c6 d8b8 c3b5 b8b6 c6b6 c7b6 b5d4 e6d7 a2a4 a7a6 c2c4 f8d8 c1e3 d7e8 b2b3 e7f6 h2h3 d8d7 d1d2 a6a5 a1d1 a8d8');
INSERT INTO evaluations VALUES(14658,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(14658,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(14658,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(14658,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(14658,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(14658,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(14658,36,'cp 110','d4c6 b7c6 e4e5 f6g4 e5e6 d6d5 d1g4 c8e6 g4f3 d5c4 f3c6 e6d7 c6c4 f8e7 b1c3 e8g8 f1d1 d7e6 c4c6 d8b8 c3b5 b8b6 c6b6 c7b6 b5d4 a8d8 c1e3 e6d5 f2f3 f8e8 g1f2 e7c5 c2c3 d5c4 b2b3 c5d4 e3d4 e8e2 f2f1 c4a6 c3c4 d8e8 d4f2 e2c2 d1d7');
INSERT INTO evaluations VALUES(21924,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(21924,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(21924,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(21924,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(21924,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(21924,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(21924,36,'cp 77','d2d4 e5d4 f3d4 c6d4 d1d4 g8f6 e2e4 f8e7 b2b3 c7c6 f1e2 e8g8 e1g1 f8e8 c1f4 d8a5 a2a3 c8g4 f2f3 g4e6 d4d2 f6d7 c3d5 a5d2 d5e7 e8e7 f4d2 d6d5 c4d5 c6d5 d2b4 e7e8 e4d5 e6d5 e2b5 d5c6 b5c6 b7c6 a1c1 a7a5 b4d2 e8e2 f1d1');
INSERT INTO evaluations VALUES(21930,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(21930,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(21930,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(21930,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(21930,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(21930,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(21930,36,'cp 73','c6d4 f3d4 e5d4 d1d4 g8f6 e2e4 f8e7 f1e2 e8g8 e1g1 f6d7 c1e3 e7f6 d4d2 f8e8 f2f3 f6e5 a1e1 d8h4 f3f4 e5c3 d2c3 b7b6 e2d3 d7c5 e3c5 b6c5 c3a5 h4d8 e1e3 c8b7 f1e1 a8b8 b2b3 d8f6 f4f5 f6d8 a5a7 b7c6 e4e5');
INSERT INTO evaluations VALUES(45200,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(45200,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(45200,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(45200,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(45200,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(45200,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(45200,36,'cp -15','d8c7 f1b5 c8d7 b5d7 b8d7 d2d3 e7e6 e1g1 f8e7 f1e1 a7a6 a2a4 b7b6 c1f4 e8g8 h2h3 h7h6 c2e2 b6b5 f4g3 c7b7 b1d2 f8c8 e4e5 d7e5 f3e5 d6e5 g3e5 c8d8 a4b5 a6b5 d3d4 a8a1 e1a1 f6d7 d2f3 c5c4 e2c2');
INSERT INTO evaluations VALUES(45412,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(45412,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(45412,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(45412,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(45412,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45412,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(45412,36,'cp -30','b8c6 f1b5 c8g4 g1e2 a7a6 b5c6 b7c6 e1g1 e7e6 c3a4 h7h6 g5h4 f8e7 h2h3 g4f5 a1c1 e8g8 c2c4 f8e8 c4c5 f6d7 h4g3 e7f6 c1c3 e6e5 d4e5 f6e5 g3e5 d7e5 a4b6 a8b8 c3e3 d8f6 e2d4 f5g6 f2f4 e5c4 b6c4 d5c4');
INSERT INTO evaluations VALUES(4953,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4953,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(4953,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4953,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(4953,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4953,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(4953,36,'cp 0','d2d4 e5d4 f3d4 g8e7 c1g5 f7f6 g5e3 d8d7 d1d2 c5b4 e1c1 c6c5 d4e2 c8a6 e2f4 e8c8 h2h4 e7c6 h4h5 h7h6 f2f3 h8e8 e3f2 d7f7 d2d5 f7d7');
INSERT INTO evaluations VALUES(4950,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(4950,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4950,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4950,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4950,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4950,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4950,36,'cp 0','e5d4 f3d4 g8e7 c1g5 f7f6 g5e3 d8d7 d1d3 c5b4 a2a3 b4c3 d3c3 c6c5 d4e2 a7a5 e2f4 e7g6 e1g1 g6f4 e3f4 e8g8 f2f3 h7h6 h2h3 d7b5 b2b4 a5b4 a3b4 a8a1 f1a1 b5b4 c3b4 c5b4 a1a7 f8f7 a7a8 f7f8');
INSERT INTO evaluations VALUES(4961,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4961,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(4961,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4961,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4961,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4961,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(4961,36,'cp 0','f3d4 g8e7 c1g5 f7f6 g5e3 d8d7 d1d3 c5b4 a2a3 b4c3 d3c3 c6c5 d4e2 a7a5 e2f4 e7g6 f2f3 g6f4 e3f4 h7h6 e1g1 d7b5 f1d1 e8g8 b2b4 a5b4 a3b4 a8a1 d1a1 b5b4 c3b4 c5b4 a1a7 f8f7 a7a8 f7f8');
INSERT INTO evaluations VALUES(4960,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(4960,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(4960,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4960,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4960,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4960,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(4960,36,'cp 0','g8e7 c1g5 f7f6 g5e3 d8d7 d1d3 c5b4 a2a3 b4c3 d3c3 c6c5 d4e2 a7a5 e2f4 e7g6 e1g1 e8g8 f2f3 g6f4 e3f4 d7b5 b2b4 a5b4 a3b4 a8a1 f1a1 b5b4 c3b4 c5b4 a1a7 g7g5 f4g3 f8f7 a7a8 f7f8');
INSERT INTO evaluations VALUES(7299,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(7299,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(7299,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(7299,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(7299,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(7299,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(7299,36,'cp 7','c3d5 d7d6 c2c3 f6d5 c4d5 d8f6 d2d3 h7h6 h2h3 c6e7 d5b3 c8d7 b3a4 b7b5 a4c2 c5b6 a2a4 e8g8 d3d4 e7g6 c1e3 f6e7 a4a5 b6a7 b2b4 d7e6 f1e1 c7c6 d4d5 e6d7 c2b3 a7e3 e1e3 c6d5 b3d5 a8b8 d1d2');
INSERT INTO evaluations VALUES(7595,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7595,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(7595,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(7595,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7595,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7595,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(7595,36,'cp 11','e1g1 d7d5 e4d5 c6d5 c4b5 c8d7 f1e1 c5e7 d1e2 a8b8 c2c4 b8b6 c1e3 c7c5 b5d7 d8d7 c4d5 f6d5 e1d1 d7c6 b1c3 d5c3 b2c3 e8g8 g2g3 f8e8 e2d2 h7h6 d2d5 c6c8 d5d7 c8a8 a1c1 b6e6 c3c4 e8d8 d7a4 d8d1 a4d1 a8e8 d1d5 e6d6 d5b7 e7f6 g1g2');
INSERT INTO evaluations VALUES(7594,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7594,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7594,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7594,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(7594,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(7594,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(7594,36,'cp 1','d7d6 b1c3 e8g8 d1d3 c5b6 c4b3 f6d7 c1e3 a7a5 d3d2 d7c5 a1d1 d8h4 g1h1 c5b3 c2b3 b6e3 d2e3 f7f5 e4f5 c8f5 h1g1 a8e8 e3d4 h4d4 d1d4 f5d7 d4d2 e8e5 f2f3 f8b8 f1d1 b8e8 h2h3 g8f8 g1f2 h7h6');
INSERT INTO evaluations VALUES(27105,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(27105,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(27105,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(27105,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(27105,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(27105,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(27105,36,'cp 178','c1f4 g8e7 e1c1 e7c6 d4e3 a7a6 h2h4 f8e7 e4e5 e8g8 h4h5 h7h6 e3e4 f7f5 e5f6 f8f6 c3d5 e6d5 e4d5 g8h8 f4e3 d8g8 f1c4 g8d5 c4d5 a8e8 f2f4 c6b4 d5b3 b4c6 h1e1 f6f8 g2g3 e7f6 c2c3 c6a5 b3c2 a5c4 e3g1 b7b5 c2g6');
INSERT INTO evaluations VALUES(27107,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(27107,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(27107,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(27107,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(27107,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(27107,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(27107,36,'cp 159','g8e7 e1c1 e7c6 d4e3 f8e7 e4e5 e8g8 h2h4 g8h8 h4h5 h7h6 f4h6 c6e5 h6f4 e5g4 e3d2 c7c6 f2f3 g4e5 c3e4 e6d5 c1b1 f7f5 e4c3 d5e6 h5h6 g7g6 f1e2 b7b5 d2d4 d8c7 f4g3 b5b4 c3a4 c6c5 a4c5 c7c5 d4c5 d6c5 g3e5 h8h7 e5f4 a8d8 d1d8 f8d8');
INSERT INTO evaluations VALUES(32307,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(32307,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(32307,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(32307,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(32307,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(32307,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(32307,36,'cp -18','g8f6 f1d3 a7a6 d3f5 e6f5 e1g1 f8d6 b2b3 c7c6 c1b2 e8g8 a2a4 f8e8 c3e2 b8d7 c2c4 f6e4 b2a3 d6b8 a1c1 g7g6 c4d5 c6d5 e2f4 b8f4 e3f4 d7f8 a3f8 g8f8 f1e1 a8c8 f3e5 f8g7 f2f3 e4f6 c1c8 d8c8 g2g3 h7h5 e5d3 e8e1 d1e1');
INSERT INTO evaluations VALUES(37393,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(37393,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(37393,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(37393,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(37393,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(37393,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37393,36,'cp -14','c7c5 b1c3 d7d5 d4c5 f6e4 a1c1 b8d7 c4d5 d8a5 d1c2 e4c3 b2c3 d7c5 f3d4 e7e5 d5e6 c8e6 d4e6 c5e6 f1e2 e6f4 e3f4 a8c8 e1g1 c8c3 c2b1 f8e8 e2c4 c3c1 f1c1 g7d4 f4f5 g6f5 g2g3 a5d2 c1c2 d4f2 g1g2 d2e1 b1b7');
INSERT INTO evaluations VALUES(43135,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(43135,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43135,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43135,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43135,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(43135,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43135,36,'cp 0','h7h6 g5f4 d7d6 h2h3 c7c5 c2c3 c5d4 e3d4 b8c6 f1c4 d8b6 c4b3 b6a6 b1a3 e8g8 b3c4 a6b6 c4b3');
INSERT INTO evaluations VALUES(52780,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52780,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52780,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(52780,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52780,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52780,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52780,36,'cp 25','c2c4 g7g6 c4d5 c6d5 g1f3 f8g7 b1c3 g8f6 f3e5 c8f5 f1g2 e8g8 e1g1 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 a1c1 d8c8 a4a7 c3c1 f1c1 c8a8 a7a8 f8a8 e2e3 g7f8 c1a1 a8a3 g2f1 f7f6 h2h4 h7h5');
INSERT INTO evaluations VALUES(55545,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(55545,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(55545,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55545,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55545,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(55545,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55545,36,'cp 3','g8f6 d2d3 b8c6 b1c3 c6a5 g1e2 f8e7 e1g1 e8g8 f2f4 e5f4 e2f4 a5c4 d3c4 c7c6 c1e3 f8e8 b2b3 d8a5 d1d3 f6d7 d3d2 d7e5 e3d4 e7h4 a1d1 c8e6 a2a4 e5g6 d4e3 g6e5 g1h1 h4g5 f4e6 g5e3 d2e3');
INSERT INTO evaluations VALUES(47814,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(47814,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(47814,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47814,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(47814,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(47814,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(47814,36,'cp 12','d2d3 b8c6 b1c3 c6a5 g1e2 f8e7 e1g1 e8g8 f2f4 e5f4 e2f4 a5c4 d3c4 c7c6 c1e3 f8e8 b2b3 d8a5 d1d3 f6d7 a1d1 d7e5 d3d2 c8e6 d2e1 e5g6 f4e6 f7e6 c3d5 e7d8 e1a5 d8a5 d5f4 g6f4 e3f4 e6e5 f4h2 a5c7 g1h1');
INSERT INTO evaluations VALUES(47805,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(47805,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(47805,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(47805,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(47805,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(47805,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47805,36,'cp 18','b8c6 b1c3 c6a5 g1e2 c7c6 e1g1 f8e7 f2f4 e5f4 e2f4 a5c4 d3c4 c8e6 b2b3 e8g8 c1e3 d8a5 d1d3 a7a6 c3e2 f6d7 e3d4 a5g5 d3c3 e7f6 a1d1 a8e8 d4f6 d7f6 e2g3 g5e5 c3d4 e6c8 d4d6 e5d6 d1d6 f6e4 g3e4');
INSERT INTO evaluations VALUES(15499,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(15499,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15499,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(15499,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(15499,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(15499,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(15499,36,'cp 15','b1c3 c6a5 g1e2 c7c6 e1g1 f8e7 f2f4 e8g8 c4b3 a5b3 a2b3 e5f4 e2f4 d6d5 d1f3 d5e4 d3e4 f6d7 c1e3 e7d6 a1d1 d8e7 c3e2 a7a5 e2g3 g7g6 f3f2 f7f6 g1h1 f8f7 e3d4 d6c7 f4d3 b7b6 d4c3 c8a6 f1e1');
INSERT INTO evaluations VALUES(15485,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(15485,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(15485,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(15485,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15485,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15485,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(15485,36,'cp 3','c6a5 c4b3 f8e7 f2f4 e5f4 c1f4 e8g8 d1e2 a5b3 a2b3 c7c6 g1f3 d6d5 e1g1 d5d4 c3d1 c8e6 b3b4 f6h5 f4e5 e7b4 g2g4 h5f6 e5d4 f6d7 e2h2 b7b5 g1h1 c6c5 d4f2 f8e8 d1e3 h7h6 e3f5 a7a5 f2e3');
INSERT INTO evaluations VALUES(3825,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(3825,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(3825,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(3825,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(3825,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3825,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(3825,36,'cp 26','b2b4 c5b6 a2a4 a7a6 d2d3 h7h6 h2h3 b6a7 c4b3 c8e6 b1d2 f8e8 b3e6 f7e6 d2b3 d8d7 c1d2 e8f8 d3d4 e5d4 c3d4 e6e5 d4e5 c6e5 f3e5 d6e5 d2e3 a7e3 e1e3 d7d1 a1d1 a8d8 d1c1 f6h5 a4a5 f8f7 g1h2 f7f2 c1c7');
INSERT INTO evaluations VALUES(29670,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29670,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(29670,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(29670,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29670,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29670,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29670,36,'cp 0','b8d7 h2h3 g4h5 c2c4 c7c6 d1b3 d8c7 c4d5 c6d5 g2g4 h5g6 g4g5 g6c2 b3c3 c7c3 b2c3 f6e4 c1a3 a8c8 a1c1 e4d2 f3d2 c2g6 c3c4 e7e6 a3f8 e8f8 f1g2 f8e7 e1e2 g6h5 e2d3 h5g6');
INSERT INTO evaluations VALUES(33247,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(33247,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(33247,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(33247,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(33247,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33247,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(33247,36,'cp 12','d7d6 c1e3 b8c6 f1e2 f6g6 e1g1 g8f6 f2f3 d6d5 d1d2 d5e4 d4c6 c5e3 d2e3 b7c6 b1d2 e8g8 d2e4 f6d5 e3d2 c8f5 e2d3 a7a5 a1e1 f8d8 d3b1 h7h6 a2a3 a8b8 f1f2 a5a4 f2e2 d5f6 e4f6 g6f6 d2f4 f5b1 f4f6 g7f6 e1b1 c6c5 e2e7 d8d2 e7c7 d2b2 b1b2 b8b2');
INSERT INTO evaluations VALUES(41342,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41342,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41342,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(41342,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41342,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41342,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(41342,36,'cp 29','e1g1 e8g8 e3e4 c7c5 c2c3 c5d4 c3d4 b8c6 h2h3 e7e5 d4d5 c6b4 c1g5 h7h6 g5f6 d8f6 d3b5 h6h5 b1d2 g7h6 a2a3 b4a6 a1c1 f6e7 c1c3 a6c7 b5d3 c7e8 a3a4 c8d7 f1e1 a8c8 c3c8 d7c8 d2c4 e8f6 d1b3 f6d7 b3a3 d7c5');
INSERT INTO evaluations VALUES(41341,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(41341,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41341,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41341,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(41341,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41341,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41341,36,'cp 25','e8g8 e3e4 c7c5 c2c3 c5d4 c3d4 b8c6 h2h3 e7e5 d4d5 c6b4 c1g5 d8d7 g5f6 g7f6 d3e2 a7a5 b1d2 b7b5 d1b3 a8b8 f1c1 b4a6 c1c6 a6b4 c6c3 b4a6 b3c2 a5a4 b2b3 a4b3 d2b3 b8a8 c2d2 c8b7 a2a3 f8c8 a1c1 f6d8 d2e3 c8c3');
INSERT INTO evaluations VALUES(36641,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36641,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36641,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36641,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(36641,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(36641,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36641,36,'cp 23','h2h3 b8d7 e3e4 e7e5 d4e5 d7e5 f3e5 d6e5 d1f3 f6h5 b1c3 c8e6 f1d1 d8e7 d3f1 c7c6 a2a4 a7a5 b2b3 f8d8 d1d8 a8d8 f3e3 d8d4 c3e2 d4d7 c1a3 e7h4 e2c3 g7f8 a3f8 g8f8 a1d1 d7d1 c3d1 h4f4 e3c5 f8g7 c5a5 f4e4 d1c3 e4c2 a5e5 h5f6');
INSERT INTO evaluations VALUES(45468,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(45468,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(45468,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(45468,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(45468,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(45468,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(45468,36,'cp -45','b8c6 d2d4 c8g4 c1e3 g4h5 h2h3 e7e6 g2g4 h5g6 f3h4 f8d6 h4g6 h7g6 b1d2 e8g8 c2c3 f8e8 e1g1 d8c7 d1f3 c6a5 g1g2 b7b5 e3g5 f6h7 g5h4 a8b8 h4g3 h7f6 g3d6 c7d6 f3g3 d6b6 f1e1 b5b4 a1c1 e8c8 h3h4 b4c3 b2c3 b6c6 e1h1 a5c4 b3c4 d5c4 d2f3');
INSERT INTO evaluations VALUES(13177,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(13177,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(13177,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(13177,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(13177,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(13177,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(13177,36,'cp -33','d2d4 c8g4 c1e3 e7e6 h2h3 g4h5 g2g4 h5g6 f3h4 f8d6 h4g6 h7g6 b1d2 e8g8 c2c3 f8e8 e1g1 b7b5 d1f3 d6c7 f3g2 c6a5 b3c2 a8b8 a1e1 a5c4 d2c4 d5c4 e3g5 b5b4 c2e4 d8d6 g5f6 g7f6 h3h4 f6f5 g4f5 e6f5 e4f5 b4c3 b2c3');
INSERT INTO evaluations VALUES(13144,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(13144,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(13144,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(13144,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(13144,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(13144,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(13144,36,'cp -34','c8g4 c1e3 e7e6 h2h3 g4h5 g2g4 h5g6 f3h4 f8d6 h4g6 h7g6 b1d2 e8g8 c2c3 f8e8 e1g1 b7b5 d1f3 d6c7 f3g2 c6a5 b3c2 c7d6 e3g5 a8c8 a2a3 d8c7 h3h4 e6e5 d4e5 d6e5 h4h5 g6h5 g4h5 f6h5 a1e1 h5f4 g5f4 e5f4 e1e8 c8e8 g2d5 f4d2 d5d2');
INSERT INTO evaluations VALUES(24633,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(24633,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(24633,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(24633,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(24633,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(24633,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(24633,36,'cp -28','c1e3 e7e6 h2h3 g4h5 g2g4 h5g6 f3h4 f8d6 h4g6 h7g6 b1d2 e8g8 c2c3 f8e8 e1g1 b7b5 d1f3 a8c8 e3g5 d6c7 f3g2 c6a5 f1e1 d8d6 h3h4 a5b3 a2b3 a7a5 g2h3 e6e5 d4e5 e8e5 g5f6 g7f6 h4h5 d6e6 d2f3 e5e4 e1e4 d5e4 f3d4');
INSERT INTO evaluations VALUES(6659,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(6659,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(6659,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(6659,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(6659,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(6659,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(6659,36,'cp 82','b4d2 b1d2 c6e7 f1d3 d7d5 e1g1 g8h6 f1e1 e8g8 e4e5 c8f5 d3f1 c7c6 d1b3 b7b6 a1c1 a8c8 b3a3 e7g6 b2b4 g6f4 g2g3 f4h3 f1h3 f5h3 a3a7 h3g4 a7a4 h6f5 a4a3 c8a8 a3b3 d8d7 a2a4 f7f6 e5f6 f8f6');
INSERT INTO evaluations VALUES(6711,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(6711,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(6711,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(6711,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(6711,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(6711,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(6711,36,'cp 83','b1d2 c6e7 f1d3 d7d5 e1g1 g8h6 f1e1 e8g8 e4e5 c8f5 d3f5 h6f5 a1c1 c7c6 d2b3 b7b6 h2h4 g7g6 c1c3 f5g7 b3c1 a8c8 c1e2 e7f5 h4h5 g7e6 d1d2 c6c5 e2f4 f5d4 f3d4 e6d4 c3g3 d4f5 g3g4 d5d4 h5g6 f7g6');
INSERT INTO evaluations VALUES(8942,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(8942,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(8942,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(8942,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(8942,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(8942,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(8942,36,'cp 91','e2e3 e8g8 c4d5 e6d5 f1d3 b4e7 d1b3 c8e6 e1g1 a8b8 h2h3 f6e4 g5e7 c6e7 b3c2 e4c3 b2c3 h7h6 c3c4 d5c4 d3c4 e6c4 c2c4 c7c6 f1c1 d8d5 c4a4 a7a6 a1b1 f8c8 f3e5 d5e6 e5d3 e7d5 a4c2 b7b5 c2e2 e6d6 e2f3 c8c7');
INSERT INTO evaluations VALUES(8941,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(8941,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(8941,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(8941,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(8941,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(8941,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(8941,36,'cp 54','h7h6 g5h4 g7g5 h4g3 f6e4 f3d2 e4c3 b2c3 b4c3 a1c1 c3b4 f1d3 d5c4 d3c4 f7f5 e1g1 e8g8 d2f3 d8e7 d1c2 b4a3 c1b1 b7b6 c4b5 c6a5 f3e5 e7g7 f2f4 c8b7 f4g5 h6g5 b5d7 g7e7 c2a4 a3d6 e5g6 e7f7 g6f8 a8f8 b1d1 f7e7 d7b5 d6g3 h2g3');
INSERT INTO evaluations VALUES(8788,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(8788,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(8788,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(8788,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(8788,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(8788,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(8788,36,'cp 99','g5h4 g7g5 h4g3 f6e4 f3d2 e4c3 b2c3 b4c3 a1c1 c3b4 f1d3 d5c4 d3c4 f7f5 e1g1 e8g8 d2f3 b4d6 f1e1 d6b4 e1e2 b4a3 c1b1 a3d6 e2c2 d8e7 d1d2 a8b8 g3d6 c7d6 d4d5 c6d8 d5e6 c8e6 f3d4 e6c4 c2c4 d8e6 b1b5 e6d4 c4d4 b8d8');
INSERT INTO evaluations VALUES(8789,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(8789,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(8789,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(8789,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(8789,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(8789,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(8789,36,'cp 87','g7g5 h4g3 f6e4 f3d2 e4c3 b2c3 b4c3 a1c1 c3b4 f1d3 d5c4 d3c4 f7f5 e1g1 e8g8 d2f3 b4a3 c1b1 a3d6 a2a4 g8h8 g3d6 d8d6 d1c2 g5g4 f3e1 b7b6 e1d3 c8b7 b1c1 c6a5 d3f4 a5c4 c2c4 f8f7 c4e6 d6e6 f4e6 a8c8 e6f4 h8g8 f1d1 a7a5 d4d5');
INSERT INTO evaluations VALUES(21012,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21012,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21012,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(21012,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21012,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(21012,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(21012,36,'cp 12','g1f3 g8f6 c2c4 e7e6 b1c3 d5c4 d3c4 c6b4 c4b3 c7c5 e2e3 c5d4 f3d4 f8e7 a2a3 b4d5 f1b5 c8d7 c3d5 f6d5 e1g1 e8g8 f1d1 d7b5 d4b5 f7f5 a3a4 a7a6 b5d4 d8d7 d4e2 a8c8 c1d2 e7f6 g2g3');
INSERT INTO evaluations VALUES(21011,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(21011,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(21011,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(21011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21011,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(21011,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(21011,36,'cp 0','c8g4 e2e3 e7e6 c1d2 g4f3 g2f3 g8f6 d3b3 a8b8 c2c4 f8e7 c4d5 e6d5 a2a3 e8g8 h2h4 d8d7 f1h3 d7d8');
INSERT INTO evaluations VALUES(24986,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(24986,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(24986,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(24986,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(24986,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(24986,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(24986,36,'cp 56','e7e6 h2h3 g4h5 e1g1 f8d6 f1e1 e8g8 g2g4 h5g6 e3e4 e6e5 d4d5 c6e7 c1g5 f6d7 c4f1 h7h6 g5e3 a7a6 a2a3 d7f6 d1d3 g6h7 f3d2 g8h8 a1c1 e7g6 d2c4 b7b5 c4d6 c7d6 e3d2 f6d7 c3e2 g6h4 d3e3 a6a5 e2g3 d7c5');
INSERT INTO evaluations VALUES(25237,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25237,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(25237,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(25237,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(25237,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(25237,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(25237,36,'cp 69','a2a3 g4h5 c4e2 f8d6 e3e4 h5f3 e2f3 e6e5 d4d5 c6d4 c1e3 d6c5 e1g1 d8d7 f3e2 d4e2 d1e2 c5e3 e2e3 e8g8 h2h3 f8d8 a1d1 c7c6 d1d2 c6d5 c3d5 f6d5 d2d5 d7e7 f1e1 h7h5 e3c3 d8e8 e1d1 a8d8 f2f3 d8d5');
INSERT INTO evaluations VALUES(26127,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(26127,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26127,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(26127,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(26127,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(26127,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(26127,36,'cp 54','b1d2 g8e7 d2b3 f5g6 e1g1 e7f5 c2c3 a8c8 a2a4 a7a6 a4a5 c6c5 b3c5 f8c5 d4c5 c8c5 c1g5 d8b8 d1a4 c5c7 a4b4 h7h6 g5d2 d7c5 a1d1 c7c8 c3c4 d5c4 d1c1 e8g8 c1c4 c5d7 f1c1 c8c4 e2c4 g6h5 c4e2 d7e5 d2f4');
INSERT INTO evaluations VALUES(39541,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(39541,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(39541,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(39541,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(39541,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(39541,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(39541,36,'cp -38','c7c5 c2c3 d7d6 f1b5 c8d7 b5d7 d8d7 e5d6 d7d6 e1g1 g8f6 b1a3 e8g8 f1e1 b8c6 d2d4 f8d8 d1b3 c5d4 a3b5 d6c5 b5d4 c6d4 f3d4 d8d7 d4f3 f6d5 f3e5 g7e5 e1e5 a8c8 e5e4 d5f6 c1e3 c5c6 e4d4 f6d5 e3h6 e7e5 d4d1 f7f6 h6e3');
INSERT INTO evaluations VALUES(37756,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(37756,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(37756,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(37756,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(37756,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(37756,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(37756,36,'cp -26','c2c3 d7d6 f1b5 c8d7 b5d7 d8d7 e5d6 d7d6 e1g1 g8f6 b1a3 e8g8 f1e1 b8c6 d2d4 f8d8 d1b3 c5d4 a3b5 d6d5 b5d4 c6d4 f3d4 e7e6 c1f4 a7a6 d4f3 d5b3 a2b3 f6d5 f4d2 h7h6 g2g3 d5e7 b3b4 e7f5 d2e3 f5e3 e1e3 g6g5 e3e2 a8c8 f3e5 g7e5 e2e5');
INSERT INTO evaluations VALUES(47277,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(47277,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(47277,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(47277,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(47277,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(47277,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(47277,36,'cp -44','b8c6 d2d4 e7e5 d4e5 c6e5 b1d2 f8d6 g1f3 c8g4 h2h3 g4h5 e1g1 e8g8 b2b3 f6e4 g3g4 e5f3 d2f3 h5g6 c1b2 c7c6 d1d4 f7f6 c2c4 d6c5 d4d1 d5c4 d1c1 g6f7 b3c4 d8e7 b2d4 a8d8 c1c2 f7g6 f3h4 c5d4 e3d4');
INSERT INTO evaluations VALUES(14343,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(14343,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(14343,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(14343,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(14343,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(14343,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(14343,36,'cp -47','d2d4 e7e5 d4e5 c6e5 b1d2 f8b4 a2a3 b4d6 g1f3 c8g4 h2h3 g4h5 e1g1 e8g8 b2b4 c7c6 c1b2 d8e7 d1c1 e5f3 d2f3 f6e4 f3d4 h5g6 c2c4 d5c4 c1c4 e4g3 f2g3 e7e3 g1h1 g6d3 d4f5 d3c4 f5e3 c4f1 a1f1 f8e8 e3f5 d6f8 b4b5 e8e2 b2d4 g7g6 b5c6 b7c6 f5e3 f8a3 g2c6');
INSERT INTO evaluations VALUES(14290,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(14290,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(14290,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(14290,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(14290,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(14290,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(14290,36,'cp -38','e7e5 d4e5 c6e5 b1d2 c7c6 g1f3 e5f3 d2f3 f8e7 e1g1 a7a5 d1e2 a5a4 c2c4 d5c4 e2c4 e8g8 c1d2 c8e6 c4c2 e6d5 f1d1 d5e4 c2c4 d8e8 f3d4 c6c5 g2e4 f6e4 d4b5 e8c6 b5c3 e4c3 d2c3 b7b5 c4g4 e7f6 c3f6 c6f6 g4e2');
INSERT INTO evaluations VALUES(52478,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(52478,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(52478,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(52478,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(52478,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(52478,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(52478,36,'cp -27','d7d5 g1f3 g8f6 c2c4 d5c4 d3d4 c8e6 b1a3 c5d4 d1d4 e6d5 f1c4 b8c6 d4d1 d5e4 e1g1 e7e6 d1d8 a8d8 c4e2 f8e7 b2b3 f6d5 c1b2 e8g8 a3c4 d5b4 f1d1 h7h6 a2a3 b4d3 b2c3 d3c5 f3d2 e4d3 e2d3 d8d3 a1c1 f8d8');
INSERT INTO evaluations VALUES(53318,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(53318,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(53318,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(53318,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(53318,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(53318,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(53318,36,'cp -17','g1f3 g8f6 d3d4 c5d4 e3d4 b8c6 c2c3 c8f5 c1f4 e7e6 f1d3 f5d3 d1d3 f6h5 f4g3 d8b6 b2b4 a8c8 b1d2 h5g3 h2g3 h7h6 e1g1 f8d6 f1c1 e8g8 a2a3 c6e7 c3c4 d5c4 c1c4 b6a6 a1c1 e7d5');
INSERT INTO evaluations VALUES(57398,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57398,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57398,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57398,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57398,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57398,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57398,36,'cp 0','a2a3 d7d5 d2d4 e5e4 c2c4 c7c6 b1c3 f6f5 e2e3 g8f6 g1h3 c8e6 d1b3 f8d6 b4b5 b8d7 c4c5 d6c7 b3a4 c6b5 c3b5 c7b8 b5c3 b8c7');
INSERT INTO evaluations VALUES(57399,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57399,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57399,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57399,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57399,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57399,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57399,36,'cp 0','d7d5 d2d4 e5e4 c2c4 c7c6 b1c3 f6f5 e2e3 g8f6 g1h3 c8e6 d1b3 f8d6 b4b5 b8d7 c4c5 d6c7 b3a4 c6b5 c3b5 c7b8 b5c3 b8c7');
INSERT INTO evaluations VALUES(56816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56816,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56816,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56816,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56816,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56816,36,'cp 0','e2e3 c8e6 d2d4 e5e4 c2c4 c7c6 b1c3 f8d6 g1h3 f6f5 b4b5 g8f6 d1b3 b8d7 c4c5 d6c7 b3a4 c6b5 c3b5 c7b8 b5c3 b8c7');
INSERT INTO evaluations VALUES(56815,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56815,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56815,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56815,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56815,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56815,36,'cp 0','c8e6 d2d4 e5e4 c2c4 c7c6 b1c3 f8d6 g1h3 f6f5 b4b5 g8f6 d1b3 b8d7 c4c5 d6c7 b3a4 c6b5 c3b5 c7b8 b5c3 b8c7');
INSERT INTO evaluations VALUES(3935,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(3935,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(3935,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(3935,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(3935,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(3935,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3935,36,'cp 15','d7c5 d1c2 a7a5 c1g5 h7h6 g5e3 b7b6 f3d2 f6h7 d2b3 c8d7 b3c5 b6c5 a1e1 a5a4 c3b5 h6h5 e2d3 g7f6 c2d2 f8e8 a2a3 g8g7 h2h3 f6h4 e1e2 d7b5 c4b5 h4g5 d2c3 d8f6 e3g5 h7g5 h3h4 g5h7 g2g3 h7f8 f2f4 f8d7 e2f2 e8f8 f4f5 g6g5');
INSERT INTO evaluations VALUES(3549,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(3549,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(3549,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(3549,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(3549,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(3549,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(3549,36,'cp 31','d1c2 a7a5 c1g5 b7b6 f3d2 h7h6 g5e3 f6h7 d2b3 h6h5 b3c5 b6c5 c3b5 g7f6 a1e1 a5a4 e2d3 c8d7 c2d2 f8e8 a2a3 d7b5 c4b5 f6g5 f2f4 e5f4 e3f4 g8g7 f1f3 g5f4 d2f4 f7f6 h2h4 d8e7 e4e5 d6e5 f4g3 h7f8 f3f2 e7d6 d3c2 a8c8 c2a4 d6d5 a4c2');
INSERT INTO evaluations VALUES(3551,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(3551,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3551,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(3551,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(3551,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(3551,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(3551,36,'cp 5','a7a5 c1g5 b7b6 f3d2 h7h6 g5e3 f6h7 d2b3 c5b3 a2b3 c8d7 g1h1 g8h8 e3d2 f7f5 e4f5 g6f5 f2f3 d8h4 f3f4 h4e7 f4e5 d6e5 a1e1 h7f6 c2c1 h8h7 e2d3 f6g4 g2g3 e7d6 c1c2 e5e4 d3e2 d6g6 d2f4 a8e8');
INSERT INTO evaluations VALUES(2581,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(2581,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(2581,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(2581,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(2581,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(2581,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(2581,36,'cp 16','c1g5 b7b6 f3d2 h7h6 g5e3 f6h7 d2b3 c5b3 a2b3 c8d7 b3b4 a5b4 a1a8 d8a8 c3b5 h7f6 f2f3 f8c8 c2b3 a8a5 e3d2 h6h5 d2b4 a5a8 b3a3 d7b5 c4b5 g7h6 f1a1 h6e3 g1f1 a8a3 b2a3 f6d7 a3a4 e3d4 a1a2 d7c5 a4a5 b6a5 b4a5 c5b3');
INSERT INTO evaluations VALUES(20028,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(20028,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(20028,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(20028,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(20028,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(20028,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(20028,36,'cp 48','c2c3 g8f6 d2d4 e7e6 d4d5 e6d5 e4d5 c6e7 c3c4 e7g6 e1g1 f8e7 b1c3 e8g8 d1c2 f6d7 f3d2 g6f4 e2d1 a7a6 d2e4 f4g6 f2f4 b7b5 c4b5 f7f5 e4g3 a6b5 c3b5 d7b6 d1e2 g8h8 f1d1 c8b7 c2b3 a8a4 g3h5 a4b4 b3c3 f8f7 c3c2 b6d5 e2c4 b4c4 c2c4');
INSERT INTO evaluations VALUES(20010,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(20010,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(20010,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(20010,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(20010,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(20010,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(20010,36,'cp 40','e7e6 d2d4 g8f6 d4d5 e6d5 e4d5 c6e7 c3c4 g7g6 b1c3 f8g7 h2h3 h7h6 e1g1 e8g8 e2d3 c8f5 f1e1 f8e8 c1f4 a7a6 g2g4 f5d3 d1d3 d8d7 a1d1 g6g5 f4h2 e7g6 d3f5 d7f5 g4f5 g6f4 h2f4 g5f4 g1g2 b7b5 c4b5 a6b5 c3b5 e8e1 d1e1 f6d5 b5d6 g7b2');
INSERT INTO evaluations VALUES(47482,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47482,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47482,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(47482,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47482,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47482,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47482,36,'cp 0','g5h4 c7c5 d4c5 b8c6 e2e3 g7g6 c2c3 f8g7 b1d2 e4c5 f1e2 a7a5 e1g1 e8g8 a2a4 b7b6 f3d4 c8d7 e2f3 c6e5 f3e2');
INSERT INTO evaluations VALUES(47489,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47489,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47489,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47489,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(47489,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(47489,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(47489,36,'cp -11','c7c5 d4c5 b8c6 e2e3 h7h6 b1d2 g7g5 h4g3 e4c5 d2b3 c5e4 c2c4 e7e6 c4d5 d8d5 d1d5 e6d5 f1d3 a7a5 b3d4 f8g7 e1g1 c8d7 a1c1 e8g8 d4c6 b7c6 f3e5 e4g3 e5d7 g3f1 d7f8 f1e3 f2e3 g8f8 c1c6 d5d4 g1f2 d4e3 f2e3 a5a4 d3b5 a8e8 e3d2 g7b2 b5a4');
INSERT INTO evaluations VALUES(52294,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(52294,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(52294,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(52294,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(52294,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(52294,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(52294,36,'cp 105','d4c5 f8g7 b1c3 g7c3 b2c3 d8a5 d1d4 g8f6 d4b4 a5c7 f1d3 b7b6 c5b6 a7b6 a1b1 b8a6 b4b2 a6c5 c3c4 c5a4 b2d4 d7d6 g1e2 e8g8 f2f3 c8e6 c1g5 f8c8 b1b4 a4c5 e1g1 a8a2 e2c3 a2a8 g5f6 e7f6 d4f6 c5d7 f6d4 c7c5 d4c5 d7c5 c3b5');
INSERT INTO evaluations VALUES(52261,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(52261,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(52261,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(52261,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(52261,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(52261,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(52261,36,'cp 113','f8g7 b1c3 g7c3 b2c3 d8a5 d1d4 g8f6 d4b4 b8c6 b4a5 c6a5 f2f3 b7b6 c5b6 a7b6 c1h6 h8g8 h6g5 c8a6 f1d3 a6d3 c2d3 a5b3 a1b1 b3c5 d3d4 c5e6 g5f6 e7f6 b1a1 f6f5 e4f5 g6f5 e1f2 e6f4 g2g3 f4d5 c3c4 d5b4 a1b1 a8a2 g1e2 b4d3 f2e3 d3b2 b1a1 a2a1 h1a1');
INSERT INTO evaluations VALUES(52977,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(52977,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(52977,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(52977,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(52977,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(52977,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(52977,36,'cp 64','d2d4 e7e5 f2f4 e5d4 d1d4 d8b6 g1f3 g8f6 c1d2 f8e7 e1c1 b8d7 h2h3 b6d4 f3d4 d7c5 d1e1 f6h5 g2g4 h5g3 h1g1 g3f1 g1f1 c8d7 f4f5 e8c8 d4f3 d7e8 e4e5 d6e5 f3e5 e7d6 e5c4 d6c7 d2e3 c5d7 e3f4 d7b6 c4b6 a7b6 f4c7 c8c7');
INSERT INTO evaluations VALUES(53382,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53382,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53382,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53382,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53382,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53382,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(53382,36,'cp -8','c8f5 d1b3 d8d7 b1c3 e7e6 c1d2 b8c6 g1f3 f8d6 a1c1 g8e7 c3a4 e8g8 f1e2 h7h6 a4c5 d6c5 c1c5 e7c8 e1g1 c8d6 f1c1 a8c8 d2e1 d6e4 c5c2 e4g5 c2c3 f5g4 f3g5 g4e2 b3c2 h6g5 c2e2 c6e7 h2h3 a7a6 e2h5 c8c3 c1c3 f7f6 h5e2');
INSERT INTO evaluations VALUES(31318,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31318,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31318,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31318,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31318,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31318,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31318,36,'cp 0','d1b3 d8d7 g1f3 b8c6 b1c3 e7e6 c1d2 f8d6 f1e2 a7a6 f3h4 f5g6 c3a4 a8b8 e1g1 d7c7 g2g3 b7b5 a4c3 g8f6 h4g6 h7g6 f1c1 e8g8 b3d1 c7b7 d2e1 b8c8 c1c2 c6b4 c2c1 b4c6');
INSERT INTO evaluations VALUES(55646,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(55646,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(55646,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(55646,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(55646,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(55646,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(55646,36,'cp 23','c2c4 f8e7 b1c3 f7f5 f1e2 g8f6 d2d4 e5e4 f3d2 c7c5 d2f1 b8c6 f1g3 c5d4 e3d4 d6d5 c4d5 f6d5 e1g1 c8e6 f2f3 e4e3 c3d5 d8d5 c1e3 e8g8 g3h5 e6f7 b2b3 d5d7 e2c4 e7d6 e3f2 a8d8 c4f7 d7f7 h5g3 c6b4 g3e2 b4d5 a1c1 f7e6 d1d2 e6g6 g1h1');
INSERT INTO evaluations VALUES(5313,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5313,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5313,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5313,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5313,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5313,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5313,36,'cp 0','c2c3 e7e6 c4e2 g8e7 d2d4 c5d4 c3d4 d8b6 b1a3 c6d4 a3c4 b6b4 c1d2 b4c5 d2e3 c5b4');
INSERT INTO evaluations VALUES(12477,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(12477,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(12477,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12477,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(12477,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(12477,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(12477,36,'cp 19','f8b4 e1g1 b4c3 b2c3 f6e4 e2d3 d7d5 d4c6 b7c6 c1a3 c6c5 f1e1 e8g8 d3e4 d5e4 a3c5 f8e8 e1e4 d8a5 c5e3 e6e5 e4c4 c8f5 c4c5 a5a4 d1f3 a4c2 a1c1 c2e4 f3e4 f5e4 f2f3 e4f5 c3c4 f7f6 c5a5 e8c8 e3a7 g8f7 h2h3 g7g5 a7b6 a8a5 b6a5 h7h5 c4c5');
INSERT INTO evaluations VALUES(7835,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7835,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7835,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7835,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7835,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(7835,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(7835,36,'cp 17','e1g1 b4c3 b2c3 f6e4 e2d3 d7d5 d4c6 b7c6 c1a3 c6c5 f1e1 e8g8 d3e4 d5e4 a3c5 f8e8 e1e4 c8b7 d1d8 e8d8 e4a4 f7f6 h2h4 d8c8 a4a5 b7e4 a1d1 e4c2 d1d6 e6e5 h4h5 h7h6 f2f3 c8c7 c5b4 e5e4 g1f2 e4f3');
INSERT INTO evaluations VALUES(25516,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(25516,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(25516,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(25516,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(25516,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(25516,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(25516,36,'cp 40','f1e1 f8e7 b5a4 e8g8 c2c3 f8e8 d3d4 h7h6 h2h3 e7f8 d4d5 c6a5 b1a3 b7b6 a4d7 d8d7 d1d3 a8c8 c3c4 a5b7 f3h4 g7g6 a3b5 c8a8 d3f3 f6h7 f3d1 a7a5 c1e3 b7c5 d1c2 f8e7 h4f3 e7f8 b5c3 f8g7 c2d2 g6g5 e3c5');
INSERT INTO evaluations VALUES(25517,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(25517,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(25517,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25517,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(25517,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25517,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25517,36,'cp 40','f8e7 c2c3 e8g8 h2h3 f8e8 b5a4 h7h6 d3d4 e7f8 d4d5 c6a5 b1a3 b7b6 c3c4 a7a6 a4c2 a5b7 a3b1 c7c6 a2a4 b6b5 a4b5 a6b5 a1a8 d8a8 c4b5 c6d5 e4d5 d7b5 b1c3 b5d7 b2b4 e8c8 d1d3 g7g6 f3d2 f8g7 c2b3 d7f5 d2e4 f6e4 c3e4 f5e4 d3e4');
INSERT INTO evaluations VALUES(26946,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(26946,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(26946,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(26946,33,'cp 235',NULL);
INSERT INTO evaluations VALUES(26946,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(26946,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(26946,36,'cp 228','f3e5 c6e5 d1h5 g8f6 h5e2 e5c4 e2c4 c7c6 c4b3 d8d7 d2d4 f8e7 e1g1 e8g8 a2a4 d6d5 e4e5 f6e8 a4a5 e8c7 c1e3 f7f6 c3e2 c6c5 c2c3 c5c4 b3c2 b7b5 a5b6 a7b6 e2f4 b6b5 e5e6 d7c6 c2f5 b5b4 f5h5 a8a1 f1a1 e7d6 c3b4 d6b4');
INSERT INTO evaluations VALUES(26933,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(26933,31,'cp 246',NULL);
INSERT INTO evaluations VALUES(26933,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(26933,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(26933,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(26933,35,'cp 235',NULL);
INSERT INTO evaluations VALUES(26933,36,'cp 235','c6e5 d1h5 e5c4 h5b5 c7c6 b5c4 g8f6 c4b3 d8d7 d2d4 f8e7 a2a4 e8g8 c1e3 d6d5 e4e5 f6e8 a4a5 e8c7 e1g1 f7f6 c3e2 g8h8 f2f4 g7g6 g1h1 c7e6 c2c4 d5c4 b3c4 e6c7 e2c3 c7d5 e3d2 b7b5 c4b3 a8c8 a1e1 b5b4 c3e4 f6f5');
INSERT INTO evaluations VALUES(37831,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(37831,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(37831,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(37831,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(37831,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(37831,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(37831,36,'cp -32','b8c6 a2a4 a7a6 d2d3 d7d6 f1e2 e7e6 e1g1 g8e7 d1e1 e8g8 c3d1 c8d7 c2c3 b7b5 d1e3 b5a4 e2d1 c6a5 d1a4 d7a4 a1a4 a5b3 a4a3 b3c1 e1c1 d6d5 e4e5 a6a5 c1c2 d8b6 f1a1 f8b8 g2g3 b6b2 c2b2 b8b2');
INSERT INTO evaluations VALUES(38842,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(38842,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(38842,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(38842,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(38842,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(38842,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(38842,36,'cp 37','d2c3 g8f6 e4d5 d8d5 g1f3 b7b6 f1d3 c8b7 e1g1 e8g8 b2b3 c7c5 d4c5 d5c5 c3b2 b8d7 f1e1 f8d8 f3e5 c5d5 f2f3 d5c5 g1h1 d7e5 e1e5 c5f2 d1e1 f2e1 a1e1 a8c8 h1g1 h7h6 a2a4 a7a6 g1f2 f6d5 e5e4 d5c3');
INSERT INTO evaluations VALUES(38840,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38840,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38840,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(38840,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38840,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(38840,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(38840,36,'cp 34','d5e4 d1g4 g8f6 g4g7 h8g8 g7h6 b7b6 g1e2 c8b7 e2f4 b8d7 f1e2 d8e7 e1c1 e8c8 g2g3 f6g4 e2g4 g8g4 d4d5 e6e5 f4g2 e7g5 h6g5 g4g5 h2h4 g5f5 d1d2 f5f3 h4h5 d8g8 h1h4 f7f5 g2e1 e4e3 e1f3 e3d2 c1d2 g8d8 a2a4 d8e8 f3g5 d7f6 h5h6 b7d5 f2f3 e8g8 c3e5 g8g5 e5f6');
INSERT INTO evaluations VALUES(38905,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38905,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38905,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(38905,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38905,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(38905,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(38905,36,'cp 39','d1g4 g8f6 g4g7 h8g8 g7h6 b7b6 g1h3 b8d7 c3d2 c8b7 e1c1 d8e7 d2g5 g8g6 h6h4 h7h6 g5d2 f6g8 h4f4 e8c8 f4e3 g6g7 h3f4 e7d6 f1e2 g8e7 e3h3 d8g8 d2e3 c8b8 c1b1 e7f5 g2g4 f5e3 f2e3 d7f6 f4h5 f6h5 h3h5 g7h7 h5h3');
INSERT INTO evaluations VALUES(38903,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38903,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38903,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(38903,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38903,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38903,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(38903,36,'cp 50','g8f6 g4g7 h8g8 g7h6 b7b6 e1c1 c8b7 g1h3 b8d7 c3d2 d8e7 d2g5 g8g6 h6h4 h7h6 g5d2 f6g8 h4e7 g8e7 h3f4 g6g8 f4h5 g8g6 h1g1 e8c8 f1e2 d7f6 h5f4 g6g8 c2c3 e7f5 g2g4 f5h4 h2h3 f6h7 f4h5 h7g5 g1g3 f7f5 d2f4 g8g6 g4f5 e6f5 d1e1');
INSERT INTO evaluations VALUES(44965,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44965,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44965,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44965,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44965,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44965,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44965,36,'cp -12','d5e4 d3e4 f6e4 c3e4 c8f5 e4g3 f5g6 g1f3 b8d7 e1g1 e7e6 c1f4 f8e7 f1e1 e8g8 d1e2 d8b6 h2h4 h7h6 h4h5 g6h7 c2c4 f8e8 c4c5 b6b5 e2b5 c6b5 a2a4 b5a4 a1a4 h7c2 a4a5 a7a6 f4d6 e7d6 c5d6 a8d8 e1c1');
INSERT INTO evaluations VALUES(45044,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45044,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45044,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45044,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45044,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45044,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45044,36,'cp 0','d3e4 f6e4 c3e4 d8d5 d1e2 c8f5 e4g3 d5g2 g3f5 g2h1 f5d6 e8d8 d6f7 d8e8 e2g4 b8d7 f7h8 d7f6 g4g5 f6e4 g5g4');
INSERT INTO evaluations VALUES(47190,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(47190,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(47190,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(47190,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(47190,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(47190,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(47190,36,'cp -40','e2e3 c7c5 g1f3 b8c6 d4c5 e7e6 c2c4 f8c5 b1c3 e8g8 a2a3 d5c4 d3c4 c5d6 f1e2 c6e5 c4d4 e5f3 e2f3 d8e7 e1g1 f8d8 d4c4 d6e5 h2h3 a8b8 f1d1 d8d1 c3d1 b7b5 c4e2 a7a6 c1d2 c8b7 f3b7 e7b7 d2c3');
INSERT INTO evaluations VALUES(49670,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49670,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49670,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49670,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49670,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(49670,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49670,36,'cp 0','d2d4 f6e4 f2f3 e4g5 d4e5 g5e6 b1c3 b8c6 f3f4 d7d6 e5d6 f8d6 c1e3 e8g8 d1d2 d6c5 e1c1 d8d2 e3d2 e6d4 c3e4 c5b6 e2g3 c8g4 d1e1 a8d8 f1c4 c6a5 c4d3 a5c6');
INSERT INTO evaluations VALUES(54276,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(54276,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(54276,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(54276,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(54276,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(54276,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(54276,36,'cp -41','e5d6 d8d6 g1f3 b8c6 e1g1 g8f6 c2c3 f8e7 d2d4 c5d4 c3d4 b7b6 c1e3 e8g8 b1c3 c8b7 a2a3 a8d8 c4a2 h7h6 d1e2 f8e8 a1d1 e7f8 a2c4 c6e7 c4b5 b7c6 f3e5 c6b5 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(53863,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(53863,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(53863,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(53863,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(53863,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(53863,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(53863,36,'cp -54','g8f6 g1f3 b8c6 b1c3 d8d6 e1g1 a7a6 f1e1 f8e7 c4d3 b7b6 c3e4 d6d7 e4f6 g7f6 d3f1 c8b7 c2c3 h8g8 g2g3 e8c8 f1g2 c8b8 d1b3 b8a7 d2d4 c5d4 c3d4 d7d5 b3d5 d8d5 c1e3 c6b4 e1c1 d5d7 a2a3 b4d5 c1c2 h7h5 e3d2 a7b8 a1e1 d7d8 e1c1');
INSERT INTO evaluations VALUES(7855,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7855,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(7855,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(7855,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7855,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7855,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7855,36,'cp 0','f1c4 e8g8 a2a3 d7d6 d1d3 c8e6 c1g5 c5d4 f3d4 e5d4 g5f6 e6c4 d3c4 d8f6 c3d5 f6h4 e1g1 h4e4 a1e1 e4g6 d5e7 c6e7 e1e7 b7b5 c4d4 g6c2 d4d5 c2b2 e7f7 f8f7 d5a8 f7f8 a8d5 f8f7 d5a8');
INSERT INTO evaluations VALUES(7853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7853,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(7853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7853,36,'cp 0','e8g8 a2a3 d7d6 c1g5 c8e6 d1d3 c5d4 f3d4 e5d4 g5f6 e6c4 d3c4 d8f6 c3d5 f6g6 e1g1 g6e4 a1e1 e4g6 d5e7 c6e7 e1e7 b7b5 c4d4 g6c2 b2b4 a7a5 b4a5 c2a4 d4d6 a4a5 e7e3 a8e8 e3e8 f8e8 d6c5 h7h6 h2h3 g7g6 g2g3 e8e1 f1e1 a5e1 g1g2 e1e4 g2g1');
INSERT INTO evaluations VALUES(3018,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(3018,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(3018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3018,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(3018,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(3018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3018,36,'cp 0','a2a3 d7d6 c1g5 c8e6 c4a2 h7h6 g5f6 d8f6 c3d5 f6g6 d1e2 g8h8 e1g1 a7a5 c2c3 a5a4 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(13315,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(13315,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(13315,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(13315,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(13315,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(13315,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(13315,36,'cp 56','d2d4 e7e6 d4d5 e6d5 e4d5 c6e7 c3c4 g7g6 b1c3 f8g7 h2h3 e8g8 e1g1 a7a6 f1e1 f8e8 c1f4 e7f5 e2d3 e8e1 d1e1 c8d7 a2a4 d8c7 c3e2 a8e8 e1d2 h7h5 a4a5 c7d8 f4g5 b7b5 a5b6 d8b6 d2c2 e8b8 g5d2');
INSERT INTO evaluations VALUES(13294,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(13294,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(13294,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(13294,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(13294,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(13294,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(13294,36,'cp 40','e7e6 d4d5 e6d5 e4d5 c6e7 c3c4 b7b5 e1g1 b5c4 e2c4 e7g6 b2b4 f8e7 b4c5 d6c5 b1c3 e8g8 f1e1 a8b8 h2h3 b8b4 d1e2 e7d6 a2a3 b4b6 e2c2 h7h6 a3a4 a7a6 c4d3 g6f4 c1f4 d6f4 f3e5 f4e5 e1e5 d8d6 a1e1 b6b4');
INSERT INTO evaluations VALUES(22804,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(22804,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(22804,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(22804,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(22804,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(22804,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(22804,36,'cp -31','d7d5 e4d5 d8d5 g1f3 g8h6 b1c3 f8b4 f1e2 h6f5 e1g1 b4c3 b2c3 e8g8 a2a4 b7b6 e3c1 c8b7 f1e1 f7f6 f3d2 a8d8 c1a3 f8f7 e2f1 d5d7 d2c4 f5h4 d1h5 h4g6 c4e3 c6a5 d3d4 f6f5 a1d1 b7c6 h5h3 c6a4 f1d3');
INSERT INTO evaluations VALUES(41033,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41033,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41033,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(41033,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41033,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41033,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41033,36,'cp 8','a7a6 e3e4 b7b5 c4d3 c8b7 e4e5 f6d5 c3e4 h7h6 e1g1 b8d7 d1e2 d5b4 d3b1 c7c5 d4c5 d7c5 f1d1 b4d5 f3d4 d8b6 e4c5 e7c5 d4b3 c5e7 b1d3 g7g6 c1d2 h6h5 d3e4 e8g8 b3a5 a8c8 a5b7 b6b7 g2g3 b7b6 d2h6 f8d8 a1c1 c8c1 d1c1 b6b8 e4b1');
INSERT INTO evaluations VALUES(14267,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14267,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14267,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14267,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14267,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14267,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14267,36,'cp 0','g1f3 c8f5 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(16183,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(16183,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(16183,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(16183,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(16183,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(16183,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(16183,36,'cp 0','c6e5 c4b3 f8c5 b1c3 e8g8 c1e3 d7d6 h2h3 f8e8 f2f4 e5c6 d1d3 c6d4 e3d4 c5d4 d3d4 c8e6 a1d1 e6b3 a2b3 c7c6 f1f3 d8c7 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(17237,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17237,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17237,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17237,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17237,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17237,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17237,36,'cp 0','e7e5 e2e4 f8b4 d2d3 d7d6 f1e2 b4c5 e1g1 e8g8 c1e3 c8g4 f3e1 g4e6 e1f3');
INSERT INTO evaluations VALUES(37473,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(37473,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(37473,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(37473,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(37473,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(37473,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(37473,36,'cp -10','d7d5 c2c4 c7c6 c4d5 c6d5 f3e5 b8d7 d3d4 f6e4 c1f4 d7e5 d4e5 c8f5 b1d2 e4c5 a1c1 c5e6 d2b3 e6f4 g3f4 e7e6 b3d4 d8b6 d4f5 g6f5 b2b3 f7f6 e5f6 g7f6 e2e3 a8d8 d1e2 f8f7 f1d1 f7g7 g1h1 d5d4 c1b1 d4e3 f2e3 d8d1 b1d1 h7h6 a2a4');
INSERT INTO evaluations VALUES(36783,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36783,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36783,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36783,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36783,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36783,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36783,36,'cp 0','c2c4 b8c6 c4d5 f6d5 d1d2 f8e8 b1c3 e7e5 d2c2 d5e7 e2e3 c6b4 c2a4 c7c5 a2a3 c8d7 a4b3 d7e6 b3a4');
INSERT INTO evaluations VALUES(40901,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(40901,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(40901,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(40901,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(40901,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(40901,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(40901,36,'cp -10','e7f6 c4d5 e6d5 e2e3 e8g8 g1f3 c8e6 d1b3 b7b6 a1d1 g7g6 f1e2 c7c6 e1g1 b8d7 b3c2 a7a6 e2d3 e6g4 c2e2 b6b5 h2h3 g4e6 e2c2 c6c5 d4c5 d7c5 f3d4 a8c8 c3e2 d8d6 c2d2 f8e8 e2f4 e6d7 d3c2 c5e4 c2e4 d5e4 b2b3');
INSERT INTO evaluations VALUES(41965,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(41965,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(41965,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(41965,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(41965,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(41965,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(41965,36,'cp -2','c4d5 e6d5 e2e3 c7c6 f1d3 g7g6 g1f3 e8g8 e1g1 a7a5 h2h3 f8e8 f1e1 c8e6 d1b3 e8e7 a1c1 b8d7 c3e2 d7f8 a2a3 e7c7 b3c2 d8d6 c1b1 d6e7 e1c1 f8d7 e2f4 a5a4 f4e6');
INSERT INTO evaluations VALUES(41962,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(41962,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(41962,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(41962,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(41962,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(41962,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(41962,36,'cp -9','e6d5 e2e3 g7g6 f1d3 e8g8 g1f3 c7c6 e1g1 d8e7 h2h3 f8d8 a1b1 c8e6 c3e2 a7a5 a2a3 b8a6 b1c1 a6c7 e2c3 c7e8 c3a4 h7h5 d1c2 e8d6 a4b6 a8a7 b6a4 f6g7 f3e5 e6c8 f2f4 c8f5 d3f5 d6f5');
INSERT INTO evaluations VALUES(42198,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(42198,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(42198,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(42198,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(42198,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42198,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(42198,36,'cp -4','e2e3 c7c6 f1d3 e8g8 g1f3 g7g6 e1g1 f6e7 e3e4 d5e4 d3e4 b8d7 d4d5 c6d5 e4d5 e7f6 d1e1 d7c5 a1d1 f8e8 d5b7 e8e1 d1d8 f6d8 b7a8 e1f1 g1f1 c8a6 f1g1 c5d3 c3d5 d3b2 f3e5 g8g7 e5c6 d8b6 d5b6 a7b6 c6e5 b2c4 a8b7 a6b7 e5c4 b7d5 c4b6 d5a2 b6a4 a2e6 a4c3 e6g4 f2f3');
INSERT INTO evaluations VALUES(42196,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(42196,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(42196,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(42196,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(42196,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(42196,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(42196,36,'cp -10','c7c6 f1d3 g7g6 g1f3 e8g8 h2h3 f8e8 e1g1 a7a5 d1c2 f6g7 a2a3 c8e6 c3e2 b8a6 c2b3 e6c8 e2c3 b7b5 c3e2 c8d7 f1c1 a6c7 e2f4 a5a4 b3d1 c7e6 f4e6 e8e6 c1c2 e6e8 g2g3 d8f6 a1c1 f6e7 d3f1 g7f8 f3e5');
INSERT INTO evaluations VALUES(43538,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(43538,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(43538,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43538,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(43538,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(43538,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(43538,36,'cp 43','c1e3 b7b5 d1d2 b5b4 c3e2 e6e5 d4b3 a6a5 a2a4 b4a3 a1a3 d6d5 a3a5 a8a5 b3a5 d5e4 d2d8 e8d8 e3d2 e4f3 g2f3 f6d5 e2c1 f7f6 c1d3 c8f5 f1e2 d8c7 c2c4 d5e7 d3b4 f5e6 e2d3 b8d7 b4a6 c7b6 c4c5 d7c5 a6c5 b6c5');
INSERT INTO evaluations VALUES(46423,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(46423,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(46423,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46423,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(46423,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(46423,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(46423,36,'cp 11','f1b5 b8c6 g1f3 c5d4 f3d4 c8d7 e1g1 f8d6 d4f3 d7g4 h2h3 g4h5 b5e2 e8g8 f3h4 h5e2 c3e2 f6e4 h4f3 a8c8 c1d2 c6e5 f3e5 d6e5 d2b4 f8e8 d1b3 d8f6 a1d1 e5b2 d1d5 b7b6 e2d4 b2d4 d5d4 a7a5 b4a3 c8c3 b3a4 f6c6');
INSERT INTO evaluations VALUES(52348,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(52348,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(52348,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52348,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(52348,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(52348,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(52348,36,'cp 63','d2d4 c5d4 f3d4 g8f6 b1c3 d7d6 c1e3 a7a6 f1c4 e7e6 a2a3 b7b5 c4b3 d8c7 f2f3 b8c6 d4c6 c7c6 h2h4 f8e7 d1d2 c6b7 e1c1 a6a5 g2g4 a5a4 b3a2 b5b4 a3b4 b7b4 d2d4 b4d4 d1d4 a4a3 c3b5 a3b2 c1b1 e8g8 g4g5 f6e8 g5h6 g7g6 h4h5 g6g5');
INSERT INTO evaluations VALUES(52347,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(52347,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(52347,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(52347,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(52347,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(52347,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(52347,36,'cp 59','c5d4 f3d4 e7e5 d4b5 g8f6 b5d6 f8d6 d1d6 d8e7 d6e7 e8e7 f2f3 d7d6 c2c4 c8e6 b2b3 h8c8 b1c3 a7a6 c1e3 b7b5 c3d5 f6d5 c4d5 e6d7 e1d2 a6a5 f1e2 b8a6 h1c1 a6c5 h2h4 b5b4 c1c4 a5a4 b3a4 a8a4 e2d1 d7b5 c4c5 d6c5 d1a4 b5a4');
INSERT INTO evaluations VALUES(52350,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(52350,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(52350,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(52350,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(52350,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(52350,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(52350,36,'cp 66','f3d4 e7e5 d4b5 g8f6 b1c3 d7d6 c3d5 f6d5 d1d5 b8c6 c1e3 c8e6 d5d2 a7a6 b5c3 c6b4 a1d1 a8c8 f1e2 f8e7 a2a3 b4c6 e1g1 e8g8 c3d5 e7g5 e3g5 d8g5 d5b6 c6d4 f2f4 d4e2 d2e2 e5f4 b6c8 f8c8 e2f2 f4f3 f2f3 c8c2 b2b4 e6c4 f1e1 a6a5 d1d6 a5b4 a3b4');
INSERT INTO evaluations VALUES(52349,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(52349,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(52349,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(52349,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(52349,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(52349,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(52349,36,'cp 67','e7e5 d4b5 g8f6 b5d6 f8d6 d1d6 d8e7 d6e7 e8e7 f2f3 d7d6 c1e3 c8e6 b1c3 a7a6 e1c1 b7b5 g2g4 b8d7 a2a3 a8c8 h2h4 c8c6 h4h5 d7b6 g4g5 f6d7 g5h6 g7h6 e3f2 h8g8 f2h4 f7f6 h4f2 c6c8 h1h2 c8c6 f2e3 e6f7 c1b1 c6c8 e3h6 g8h8 h6e3 h8h5');
INSERT INTO evaluations VALUES(54557,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(54557,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(54557,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(54557,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(54557,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(54557,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(54557,36,'cp -13','b8c6 d2d4 c5d4 f3d4 f8c5 d4b3 c5b6 b1c3 g8f6 e1g1 e8g8 c1g5 h7h6 g5h4 f8e8 e2f3 c8e6 c3a4 g7g5 h4g3 e6f5 a4b6 d8b6 g3d6 e8e6 d6c5 b6c7 h2h3 a8e8 b3d4 c6d4 c5d4 f5c2 d1d2 f6e4 f3e4 c2e4 a1c1');
INSERT INTO evaluations VALUES(20537,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(20537,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(20537,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(20537,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(20537,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(20537,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(20537,36,'cp 0','e1g1 g8f6 d2d4 c5d4 f3d4 f8d6 c1e3 e8g8 b1d2 h7h6 f1e1 c8d7 d2f1 f8e8 c2c3 a7a6 d1c2 d8c7 a1d1 f6g4 d4f3 g4f6');
INSERT INTO evaluations VALUES(20535,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(20535,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20535,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(20535,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(20535,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(20535,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(20535,36,'cp -16','g8f6 d2d4 c5d4 f3d4 f8d6 c1e3 e8g8 b1d2 f8e8 f1e1 d8c7 d2f1 a7a6 c2c3 c8d7 d1c2 e8e7 a1d1 a8e8 d4f5 d7f5 c2f5 c6e5 e3d4 e7e6 f5c2 h7h5 h2h3 c7c6 f1d2 d6c7 e2f1 g7g6 d4e5 e6e5 e1e5 c7e5 d2b3');
INSERT INTO evaluations VALUES(13772,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13772,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13772,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(13772,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(13772,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(13772,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(13772,36,'cp -14','d2d4 c5d4 f3d4 f8d6 b1c3 e8g8 f1e1 f8e8 c1e3 a7a6 e2f3 h7h6 c3d5 f6d5 f3d5 d6h2 g1h2 d8d5 h2g1 c6e5 e3f4 c8g4 f2f3 e5g6 f4g3 e8e1 g3e1 g4h5 d1d2 g6e5 e1f2 f7f6 b2b3 a8d8 a1d1 d5f7 d1e1 f7c7');
INSERT INTO evaluations VALUES(9386,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(9386,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(9386,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(9386,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(9386,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(9386,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(9386,36,'cp 26','e1g1 e8g8 e2e3 c8f5 d2d3 d8d7 d3d4 c5b6 d4e5 d6e5 d1d7 f5d7 b2b4 e5e4 f3g5 c6b4 c1a3 a7a5 g5e4 f6e4 c3e4 d7c6 c4c5 b6a7 a3b4 a5b4 f1c1 f8e8 c1c4 e8d8 a1c1 h7h6 g3g4 a7b8 e4g3 c6b5 c4b4 c7c6 g2f1 b8g3 h2g3 b5f1 c1f1');
INSERT INTO evaluations VALUES(9385,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(9385,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(9385,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(9385,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(9385,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(9385,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(9385,36,'cp 37','e8g8 e2e3 f8e8 d2d3 h7h6 b2b3 a7a5 a2a3 c5a7 h2h3 c6e7 d3d4 e7g6 d4e5 d6e5 d1d8 e8d8 c1b2 c7c6 c3a4 c8f5 f1c1 f6d7 b2c3 a7c5 f3d2 f5e6 b3b4 a5b4 a3b4 c5e7 b4b5 c6b5 c4b5 d8b8 g2f1 e5e4 f1e2 g6e5 d2e4 e5c4');
INSERT INTO evaluations VALUES(14235,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(14235,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(14235,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14235,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(14235,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(14235,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(14235,36,'cp 20','e2e3 e7e6 c2c4 f8d6 f4d6 c7d6 b1c3 c6e7 f1d3 d5c4 d3c4 a7a6 g1e2 b7b5 c4d3 c8b7 e1g1 e8g8 a2a3 a8c8 d1d2 d8b6 e2g3 h7h6 f1c1 e7d5 c3e2 d5e7 c1d1 b7d5 a1c1 d5c4 d3b1 e7g6 e3e4 c4e2 g3e2 c8c1');
INSERT INTO evaluations VALUES(19775,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(19775,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(19775,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19775,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(19775,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(19775,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(19775,36,'cp -12','g8f6 a2a3 e7e6 e2e3 f8d6 f4g3 e8g8 b1d2 d8e7 f3e5 f6d7 e5d7 c8d7 f1e2 f8d8 e1g1 d7e8 f1e1 d6g3 h2g3 e6e5 d4e5 c6e5 d1c2 e8c6 d2f3 g7g6 f3e5 e7e5 e2f3 d8d6 a1d1 b7b6 c3c4 d5c4 d1d6 e5d6 e1d1 d6e6 f3c6 e6c6 c2c4 a8e8 g1h1');
INSERT INTO evaluations VALUES(23372,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(23372,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(23372,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(23372,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(23372,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(23372,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(23372,36,'cp 68','a7a6 g1e2 h7h5 d1d3 d7d6 g2g3 h5h4 c1g5 h4g3 e2g3 g8f6 h2h4 f6h7 g5e3 e7g6 e1c1 h7f6 f1e2 g6f4 d3d2 g7g6 c1b1 c8d7 d1g1 f8h6 h4h5 d8e7 a2a3 e8c8 e2f3 c8b8 b2b3 h6g5 h5g6 f7g6');
INSERT INTO evaluations VALUES(28039,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(28039,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(28039,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(28039,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(28039,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(28039,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(28039,36,'cp 101','g8f6 e4e5 f6d5 c3d5 b7d5 c2c4 d5b7 d4d5 e7e6 h2h4 h7h6 f1e2 c7c6 d5d6 c6c5 h1h3 d8c8 d1d2 b7f3 g2f3 g7e5 f3f4 e5f6 h4h5 g6h5 f4f5 b8c6 a1d1 c8a6 f5e6 f7e6 e2h5 e8d8 b2b3 d8c8 h3f3 f6d4 f3f7 d4e3 d2e3 a6a2');
INSERT INTO evaluations VALUES(28333,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(28333,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(28333,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(28333,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(28333,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(28333,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(28333,36,'cp 137','e4e5 f6d5 c3d5 b7d5 h2h4 h7h6 f1e2 e7e6 c2c4 d5b7 d4d5 c7c6 d5d6 c6c5 h1h3 b8c6 e3f4 a7a6 e1f1 b6b5 a1c1 d8b6 d1d2 b5c4 e2c4 a8c8 f1g1 h6h5 h3g3 e8g8 c4d3 c6d4 f3d4 c5d4 c1e1 c8c5 b2b4 c5c3 f4h6 b7d5 h6g7 g8g7 d2f4 b6b4');
INSERT INTO evaluations VALUES(29179,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29179,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(29179,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(29179,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29179,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(29179,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29179,36,'cp 17','c1f4 b8c6 b1d2 e7e6 e1g1 f8d6 f4d6 d8d6 c2c3 e8g8 f1e1 g4h5 h2h3 a7a6 e2d3 a8e8 d1b3 d6c7 a2a4 e8a8 b3a3 h5g6 d3g6 h7g6 a4a5 f8e8 d2b3 f6e4 b3c5 e4d6 b2b4 e6e5 d4e5 d6c4 a3b3 c6e5 f3e5 e8e5 e1e5 c7e5');
INSERT INTO evaluations VALUES(29667,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29667,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(29667,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29667,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(29667,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(29667,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(29667,36,'cp 27','g4f3 d1f3 e7e6 c2c4 f8b4 b1d2 c7c5 c4d5 f6d5 d4c5 e8g8 f1e2 b4c5 e1g1 b8c6 d2e4 c5b6 c1d2 d8h4 e4c3 a8c8 f3e4 h4e4 c3e4 b6a5 a1c1 a5d2 e4d2 f8d8 f1d1 g7g6 d2f3 g8g7 a2a3 a7a5 c1c2 h7h6 e2f1 d5f6 d1c1');
INSERT INTO evaluations VALUES(29684,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(29684,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(29684,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29684,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(29684,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(29684,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(29684,36,'cp 26','d1f3 e7e6 c2c4 f8b4 c1d2 b4d2 b1d2 e8g8 f1d3 d8e7 e1g1 b8d7 a2a3 c7c5 c4d5 f6d5 d2b3 a8c8 a1c1 c5d4 f3e4 g7g6 e4d4 d5b6 d3e2 f8d8 d4e4 d7f6 e4e5 f6d7 e5a5 a7a6 b3d4 g8g7 f1d1 h7h5 e2f1 b6d5 d4f3 c8c1 d1c1 d7f6 c1c2');
INSERT INTO evaluations VALUES(29683,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(29683,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(29683,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(29683,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29683,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(29683,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29683,36,'cp 15','e7e6 c2c4 f8b4 b1d2 c7c5 c4d5 f6d5 d4c5 e8g8 f1d3 b8c6 e1g1 b4c5 f3e2 a8c8 f1d1 d8c7 d2f3 c6e5 d3e4 f8d8 c1d2 d5f6 e4c2 e5f3 e2f3 c7b6 d2c3 f6d5 f3e4 g7g6 c3e5 c5e3 f2e3 c8c2 e4c2 b6e3 c2f2 e3e5 d1d2 e5c7 f2a7');
INSERT INTO evaluations VALUES(42147,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(42147,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(42147,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42147,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(42147,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(42147,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(42147,36,'cp 30','e8g8 g1f3 b7b6 e1g1 c8b7 d1e2 c7c5 f1d1 c5d4 f3d4 b8d7 a2a3 b4c5 d2e1 d8e7 d4b3 c5d6 c3b5 d6e5 e1b4 d7c5 b3c5 b6c5 b4a5 a8c8 b5c3 h7h6 a1c1 e5c7 a5c7 e7c7 e3e4 c8d8 f2f3 d8d1 c3d1 f6h5 g1h1');
INSERT INTO evaluations VALUES(37078,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(37078,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(37078,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(37078,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(37078,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(37078,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(37078,36,'cp 34','g1f3 b7b6 d1e2 c8b7 e1g1 c7c5 f1d1 c5d4 f3d4 b8d7 a2a3 b4c5 d2e1 d8e7 d4b3 c5d6 c3b5 d6e5 e1b4 d7c5 b3c5 b6c5 b4a5 h7h6 a1c1 f6d5 b2b3 a8c8 b5c3 d5c3 a5c3 e5c3 c1c3 e7g5 f2f3 f8d8');
INSERT INTO evaluations VALUES(37080,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(37080,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(37080,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(37080,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(37080,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(37080,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(37080,36,'cp 36','b7b6 e1g1 c8b7 d1e2 c7c5 f1d1 c5d4 f3d4 b8d7 a2a3 b4e7 a1c1 a7a6 c4a2 b6b5 e3e4 d8b8 d2e3 d7e5 e3f4 f8d8 b2b4 e7d6 f4g3 d6c7 a3a4 e5g6 a4b5 a6b5 a2b1 c7g3 h2g3 b7a6 d4b3 g6e5 b3c5 d8d1 e2d1 b8b6');
INSERT INTO evaluations VALUES(57279,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(57279,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(57279,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(57279,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(57279,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(57279,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(57279,36,'cp 77','d1b3 b8c6 e2e3 f8b4 c1d2 b4d2 b1d2 f7f5 g2g4 d7d6 g4f5 c8f5 g1e2 g8f6 e2g3 f5g6 h2h4 h7h6 c4c5 a8b8 c5d6 c7d6 b3e6 d8e7 e6e7 c6e7 h4h5 g6h7 f1h3 e8g8 a1c1 g8h8 a2a3 e7d5 h3e6 f8e8 e6f7 e8c8 e1e2 d5b6 f7g6 d6d5 f2f3 e4f3 d2f3 b6c4 g3f5 f6e4 f3e5 c8c7 e5c4');
INSERT INTO evaluations VALUES(3613,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(3613,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(3613,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(3613,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(3613,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(3613,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(3613,36,'cp -12','c5e3 f2e3 c6a5 c4b3 a5b3 a2b3 a7a5 d1d2 c7c6 c3e2 b7b5 f1f2 c8e6 e2g3 f8e8 d3d4 c6c5 a1f1 e5d4 e3d4 d6d5 f1e1 d5e4 g3e4 f6e4 e1e4 d8d6 e4e5 c5d4 d2d4 d6a6 d4d3 a8b8 f3d4 e6d7 e5e8 b8e8 f2e2 e8e2 d4e2 a6b6 g1h2 b6c7 h2g1 a5a4 b3b4 c7b6 d3d4');
INSERT INTO evaluations VALUES(3652,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(3652,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(3652,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(3652,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(3652,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(3652,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(3652,36,'cp -10','f2e3 c6a5 c4b3 a5b3 a2b3 a7a5 b3b4 c7c6 d1e1 c8e6 a1a5 a8a5 b4a5 d8a5 c3d5 a5e1 d5f6 g7f6 f1e1 f8a8 f3h4 h6h5 g1f2 b7b5 g2g4 d6d5 e4d5 c6d5 h4f5 g8h7 e3e4 d5e4 e1e4 a8a1 e4b4 h7g6 f5e3 e6d7 e3d5 h5g4 h3g4 g6g5 d5e3 a1c1');
INSERT INTO evaluations VALUES(3649,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(3649,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(3649,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(3649,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(3649,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(3649,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(3649,36,'cp -10','c6a5 c4b3 a5b3 a2b3 a7a5 b3b4 c7c6 d1e1 c8e6 a1a5 a8a5 b4a5 d8a5 c3d5 a5e1 d5f6 g7f6 f1e1 f8a8 g1f2 h6h5 f3h4 g8g7 c2c4 a8a4 f2g3 g7f8 g3f3 a4b4 e1e2 f8e7 g2g4 h5g4 h3g4 b7b5 c4b5 b4b5 h4f5 e7d7 e2d2 b5b3 f5h6 b3b8 d2g2 c6c5 h6f5 d7c6 f5h6 d6d5 e4d5 c6d5 g2c2');
INSERT INTO evaluations VALUES(3660,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(3660,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(3660,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(3660,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(3660,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(3660,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(3660,36,'cp -5','c4b3 a5b3 a2b3 a7a5 d1e1 c7c6 b3b4 c8e6 b4a5 a8a5 a1a5 d8a5 c3d5 a5e1 d5f6 g7f6 f1e1 f8a8 g1f2 h6h5 f2g3 g8g7 e1e2 b7b5 e2f2 a8a2 b2b3 a2a3 g3h4 g7g6 f2d2 c6c5 h4g3 g6h6 f3h4 a3a1 d2f2 h6g7 h4f5 e6f5 f2f5');
INSERT INTO evaluations VALUES(3661,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(3661,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(3661,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(3661,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(3661,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(3661,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3661,36,'cp 0','a5b3 a2b3 c7c6 c3e2 a7a5 e2g3 b7b5 f1f2 c6c5 d3d4 c8e6 d1e2 a5a4 a1f1 a4b3 c2b3 a8b8 e2c2 b8c8 c2e2');
INSERT INTO evaluations VALUES(14444,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(14444,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(14444,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(14444,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(14444,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(14444,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(14444,36,'cp 74','f1b5 e7e6 b1d2 e4d2 d1d2 f8d6 c2c3 c8d7 b5d3 a7a5 e1c1 a5a4 e3e4 d5e4 d3e4 c6a5 f3e5 d7b5 h1e1 d8e7 d2e3 e8c8 e4c2 c8b8 f4g3 f7f6 e5d3 h8e8 g3d6 e7d6 e3g3 e6e5 d4e5 d6e6 d3b4 d8d1 e1d1 e6e5 c2h7 e5g3 h2g3 e8e2 h7d3 b5d3');
INSERT INTO evaluations VALUES(16158,30,'cp 277',NULL);
INSERT INTO evaluations VALUES(16158,31,'cp 265',NULL);
INSERT INTO evaluations VALUES(16158,32,'cp 277',NULL);
INSERT INTO evaluations VALUES(16158,33,'cp 300',NULL);
INSERT INTO evaluations VALUES(16158,34,'cp 301',NULL);
INSERT INTO evaluations VALUES(16158,35,'cp 298',NULL);
INSERT INTO evaluations VALUES(16158,36,'cp 294','f3e5 c6e5 d4e5 f6e4 c4f7 e8f7 d1d5 f7e8 d5e4 d7d5 e4g6 e8e7 f2f4 d8d7 g6d3 c7c6 c1e3 d7f5 d3b3 e7f7 b1c3 f7g8 e1g1 f5f7 c3e2 g7g6 a1e1 h8h7 b3c3 c8f5 e3c5 h7g7 c5f8 a8f8 h2h3 h6h5 g1h2 g8h7 e2d4 f5e4 c3c5 f7e7 c5e7 g7e7 e1d1 c6c5');
INSERT INTO evaluations VALUES(19213,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19213,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19213,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(19213,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19213,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19213,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(19213,36,'cp 16','g8f6 d1e2 d8b6 e1g1 f8g7 f1d1 d7d6 d2d4 c8g4 d4d5 a7a6 b5d3 c6e5 b1d2 b6c7 h2h3 e5f3 d2f3 g4f3 e2f3 c5c4 d3c2 e8g8 c1e3 b7b5 e3d4 a8e8 a2a4 f6d7 f3e3 g7d4 d1d4 c7b6 a1e1 d7c5 f2f4 b5a4 d4c4 b6b2 c2a4');
INSERT INTO evaluations VALUES(19579,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(19579,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(19579,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(19579,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(19579,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(19579,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(19579,36,'cp 45','a7a6 b5c4 e7e6 d2d3 b7b5 c4b3 c8b7 c1f4 f8c5 d1g4 d8f6 g4g3 f6g6 a2a4 g6g3 f4g3 d7d6 f2f4 g8f6 h2h3 e8g8 b1d2 b7c6 g3f2 a8b8 a4a5 f6d7 d2f3 e6e5 g2g4 c5b4 h3h4 b8e8 f2g3 c6b7 g4g5 d7c5 f4e5 c5b3 c2b3 d6e5 f3e5 b4d6');
INSERT INTO evaluations VALUES(17533,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17533,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(17533,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(17533,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(17533,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(17533,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17533,36,'cp 45','b5a4 e7e6 d2d3 b7b5 a4b3 c8b7 c1f4 f8c5 d1g4 d8f6 g4g3 f6g6 a2a4 g6g3 f4g3 d7d6 f2f4 g8e7 a4b5 a6b5 a1a8 b7a8 b1d2 e8d7 d2f3 f7f6 f1a1 a8b7 g1f1 e6e5 f4e5 d6e5 a1a5 d7c6 g3f2 h8a8 a5a8 b7a8 f1e2 c5b4 f2e1 b4e1');
INSERT INTO evaluations VALUES(21926,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(21926,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(21926,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(21926,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(21926,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(21926,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(21926,36,'cp 46','g1f3 c6d4 f1g2 g7g6 b2b4 f8g7 a1b1 g8e7 d2d3 e8g8 f3d2 h7h5 e2e3 d4e6 a2a4 f7f5 b4b5 a7a6 b5b6 c7c6 a4a5 h5h4 e1g1 f8f7 d2b3 g6g5 d3d4 e5e4 f2f3 e4f3 d1f3 g5g4 f3d1 d6d5 c4d5 e7d5 c3d5 c6d5 b3c5 e6g5 c1a3 h4g3');
INSERT INTO evaluations VALUES(21925,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21925,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(21925,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21925,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21925,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(21925,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(21925,36,'cp 24','c6d4 f1g2 g7g6 b2b4 f8g7 a1b1 c8g4 e1g1 g8e7 e2e3 g4f3 g2f3 d4f3 d1f3 a8b8 b4b5 e8g8 b1b3 f7f5 d2d3 c7c6 f3d1 d8d7 d3d4 b8d8 c1a3 e5d4 e3d4 f8e8 b5c6 e7c6 c3b5 d6d5 b5d6 d5c4 d6e8 d8e8');
INSERT INTO evaluations VALUES(43467,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43467,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(43467,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43467,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43467,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(43467,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43467,36,'cp 0','f8e7 c1g5 b8d7 e1g1 h7h6 g5e3 b7b5 f3h4 d7c5 e3c5 d6c5 d1d8 e7d8 h4f5 e8g8 f1d1 c8e6 f5e3 c5c4 a2a4 a8b8 d1d6 b8b6 d6d1');
INSERT INTO evaluations VALUES(39682,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39682,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39682,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39682,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39682,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39682,36,'cp 2','c1g5 b8d7 e1g1 h7h6 g5e3 b7b5 f3h4 d7c5 e3c5 d6c5 d1d8 e7d8 h4f5 e8g8 f1d1 c5c4 a2a4 b5b4 c3d5 f6d5 d1d5 c8b7 d5c5 b7e4 f5e3 d8g5 c5c4 g5e3 f2e3 f7f5 c4b4 f8d8 a1d1 d8d1 e2d1 a8d8 d1e2 d8d2 g1f1 a6a5 b4b5 e4g2 f1e1 d2c2 b5e5 c2b2');
INSERT INTO evaluations VALUES(45207,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(45207,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(45207,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(45207,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(45207,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(45207,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(45207,36,'cp -12','b8c6 a2a4 g7g6 f1g2 f8g7 e1g1 e8g8 b1d2 h7h6 d2c4 c8e6 f1e1 d8d7 c1d2 e6h3 c4a5 h3g2 g1g2 c6a5 d2a5 a7a6 a5c3 b7b5 d1d2 g8h7 a4b5 a6b5 c3f6 g7f6 c2c3 d7b7 d2e2 a8a1 e1a1 b5b4 d3d4 f6g7 d4c5 d6c5');
INSERT INTO evaluations VALUES(13329,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(13329,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13329,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(13329,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(13329,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(13329,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(13329,36,'cp -2','b1c3 g7g6 c3d5 f6d5 e4d5 c6d4 f3d4 c5d4 f1g2 f8g7 e1g1 e8g8 f1e1 f8e8 c1g5 h7h6 g5d2 c8f5 g3g4 f5d7 h2h3 a8c8 a1c1 b7b6 c2c4 d4c3 b2c3 e7e5 d5e6 e8e6 d1b3 c8c7 c3c4 e6e1 c1e1 g7d4 b3b4 d4c5 b4b3 d7e6 e1e2 d8h4');
INSERT INTO evaluations VALUES(54131,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(54131,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(54131,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54131,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(54131,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(54131,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54131,36,'cp 19','e6d5 g1f3 b8c6 f1e2 c5d4 f3d4 f8d6 e1g1 g8f6 b2b3 d8e7 c1b2 e7e5 g2g3 e8g8 b1d2 c6d4 b2d4 e5e7 d4f6 e7f6 e2g4 h7h6 g4c8 f8c8 d2f3 f6f5 f3d4 f5e4 d1e2 a7a6 e2d2 d6e5 a1d1 e5d4 d2d4 e4d4 d1d4 c8c2 d4d5 c2a2 d5d7 b7b6 f1d1 a8c8 d1d4 c8c2');
INSERT INTO evaluations VALUES(54523,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(54523,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54523,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(54523,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(54523,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54523,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54523,36,'cp 21','g1f3 g8f6 f1e2 c5d4 f3d4 f8d6 e1g1 b8c6 b2b3 e8g8 c1b2 f8e8 b1d2 c8d7 d4b5 d6e5 b2e5 c6e5 b5d4 a8c8 d2f3 e5c6 a1c1 h7h6 h2h3 d8a5 d4b5 e8d8 a2a3 c6e5 f3d4 f6e4 c1c8 d8c8 d1a1 d7b5 d4b5 e5c6 b3b4');
INSERT INTO evaluations VALUES(54524,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54524,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(54524,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(54524,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(54524,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(54524,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54524,36,'cp 19','g8f6 f1e2 c5d4 f3d4 f8c5 e1g1 e8g8 b2b3 b8c6 c1b2 c5b6 d4c6 b7c6 d1c2 d8d6 b1d2 f8e8 a1c1 c8g4 e2g4 f6g4 d2f3 a8c8 h2h3 g4f6 b2f6 d6f6 c2c3 f6c3 c1c3 f7f6 f1c1 c6c5 c3c2 c8d8 g1f1 g8f7 f3d2');
INSERT INTO evaluations VALUES(54980,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(54980,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54980,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54980,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(54980,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(54980,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54980,36,'cp 0','c7c5 g1f3 b8c6 b1c3 g8f6 f1c4 e7e6 e1g1 f8e7 d2d4 c5d4 f3d4 e8g8 c1e3 a7a6 c4b3 c6d4 d1d4 b7b5 f2f4 c8b7 a2a4 a8b8 a4b5 a6b5 e4e5 d6e5 f4e5 d8d4 e3d4 f6e4 c3e4 b7e4 a1a7 e4b7 d4e3 f8d8 a7a5 b5b4 a5a7');
INSERT INTO evaluations VALUES(53082,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(53082,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53082,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53082,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(53082,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53082,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53082,36,'cp 8','g1f3 b8c6 b1c3 g8f6 f1c4 e7e6 e1g1 f8e7 d2d4 c5d4 f3d4 c6d4 d1d4 e8g8 b2b3 a7a6 c4d3 b7b5 c1b2 d8c7 c3e2 c8b7 c2c4 b5c4 f1c1 c7d7 d4c4 d6d5 e4d5 b7d5 c4c2 f8d8 e2f4 d5b7 c2e2 e7d6 f4h5 f6h5 e2h5');
INSERT INTO evaluations VALUES(57067,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57067,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(57067,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(57067,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57067,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(57067,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(57067,36,'cp 8','d2d4 d7d5 g1f3 g8f6 c1f4 f8d6 e2e3 e8g8 f4g3 c7c5 b1d2 d8c7 d4c5 c7c5 b2b4 c5c7 c3c4 c7e7 g3d6 e7d6 d1b1 b7b6 f1e2 a7a5 c4d5 f6d5 b4a5 b6a5 f3g5 d5f6 e2f3 a8a7 e1g1 b8d7 g5e4 f6e4 d2e4 d6a3 b1c1 a3e7');
INSERT INTO evaluations VALUES(8965,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(8965,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(8965,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(8965,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(8965,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(8965,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(8965,36,'cp 47','d5c4 f1c4 f6d5 e1g1 d5f4 e3f4 e8g8 f1e1 d8d6 g2g3 b4c3 b2c3 b7b6 d1c2 c8b7 f3g5 g7g6 c4d3 c6e7 h2h4 e7f5 h4h5 a8e8 d3e4 b7d5 a1d1 d6c6 e4d5 e6d5 g5f3 e8e1 d1e1 f8e8 f3e5 c6d6 c2a4 e8e7 h5g6 h7g6 a4a7 f7f6 a7a8 g8h7 e5f3 e7e1 f3e1 d6e6 g1f1 h7g7 g3g4');
INSERT INTO evaluations VALUES(9041,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(9041,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(9041,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(9041,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(9041,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(9041,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(9041,36,'cp 54','f1c4 f6d5 e1g1 d5f4 e3f4 e8g8 f1e1 d8d6 g2g3 b4c3 b2c3 b7b6 d1c2 c8b7 f3g5 g7g6 g5e4 d6e7 c4f1 c6a5 c3c4 a8d8 a1d1 b7e4 e1e4 e7f6 c2a4 c7c6 e4e5 d8d7 c4c5 f6d8 h2h4 h7h5 f1e2 f8e8 e5g5 b6b5 a4a3 d7d5 a3c3 d8c7 e2f3 d5g5 f4g5 e8d8');
INSERT INTO evaluations VALUES(9030,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(9030,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(9030,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(9030,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(9030,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(9030,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(9030,36,'cp 53','f6d5 e1g1 d5f4 e3f4 e8g8 f1e1 d8d6 g2g3 b4c3 b2c3 b7b6 d1c2 c8b7 f3g5 g7g6 g5e4 d6e7 c4f1 c6a5 c3c4 a8d8 a1d1 b7e4 e1e4 d8d7 d4d5 e7d8 f1g2 c7c5 e4e1 e6d5 c4d5 a5b7 h2h4 b7d6 g2h3 d7c7 h4h5 d8f6 h5h6 f8e8 e1e8 d6e8 f4f5 g8f8 d1e1 g6f5');
INSERT INTO evaluations VALUES(15911,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(15911,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(15911,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(15911,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(15911,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(15911,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(15911,36,'cp 42','f8e7 a2a4 e8g8 e1g1 e5d4 f3d4 d7e5 c4a2 e5c6 f1e1 h7h6 h2h3 f8e8 c1f4 e7f8 d4c6 b7c6 e4e5 d6e5 d1d8 e8d8 f4e5 f8d6 a1d1 c8f5 a2b3 d8e8 e5h2 a7a5 f2f3 e8e1 d1e1 g8f8 e1d1 d6h2 g1h2 a8b8 g2g4 f5g6 h2g1');
INSERT INTO evaluations VALUES(9575,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(9575,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(9575,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(9575,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(9575,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(9575,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(9575,36,'cp 49','a2a4 e8g8 e1g1 h7h6 h2h3 f8e8 f1e1 c7c6 a4a5 d8c7 c4f1 e5d4 f3d4 e7f8 c1f4 a8b8 b2b4 a7a6 d1d2 b7b5 a5b6 b8b6 f4g3 b6b8 d4f5 d7e5 f5e3 e5g6 d2d4 g6e5 f2f4 f6h5 g3h2 e5g6 f4f5 g6e5 f1a6 c8a6 a1a6');
INSERT INTO evaluations VALUES(32827,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(32827,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(32827,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(32827,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(32827,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(32827,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(32827,36,'cp 91','d2d4 g8f6 h2h3 g4h5 g2g4 h5g6 f1g2 b8d7 f3h4 c7c6 f2f4 e7e6 e1g1 d7b6 a2a4 a7a5 f4f5 e6f5 h4f5 g6f5 f1f5 f8e7 g4g5 f6d5 c3d5 b6d5 g2d5 c6d5 d1f3 e8g8 f3d5 d6g3 d5g2 g3e1 f5f1 e1e6 g2b7 a8e8 c1f4 e6h3 a1e1 e7d8 b7g2 h3d7 c2c3 e8e1 f1e1 d7a4 f4d6');
INSERT INTO evaluations VALUES(38809,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(38809,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(38809,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(38809,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(38809,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(38809,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(38809,36,'cp -9','d5c4 a2a3 b4a5 c2a4 b8c6 e2e3 e6e5 f1c4 e5d4 e3d4 d8d4 g1e2 a5c3 b2c3 d4f6 c1f4 g8e7 f4c7 c8e6 e1g1 a8c8 c7g3 e6c4 a4c4 e8g8 a1d1 e7f5 e2f4 c6e7 c4b4 f6c3 b4b7 c3a3 d1a1 a3c5 b7a7 c5a7 a1a7 h7h6 a7d7 e7c6 f1e1 f5g3 f2g3');
INSERT INTO evaluations VALUES(9098,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(9098,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(9098,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(9098,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9098,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9098,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9098,36,'cp 6','f3e5 c6e5 d3d4 c5d6 d4e5 d6e5 c2c4 c7c6 c4d5 f6d5 e2e4 d5b4 d1d8 e8d8 b1d2 b4c2 a1b1 c8e6 b2b3 d8e7 f2f4 e5d4 g1h1 e6g4 d2f3 d4b6 h2h3 g4f3 g2f3 h8d8 h1g2 a7a5 b1b2 c2b4 h3h4 h7h6 h4h5 b6c5 b2e2 a5a4 e4e5 a4b3 a2b3 d8d3');
INSERT INTO evaluations VALUES(11760,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(11760,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(11760,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(11760,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(11760,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(11760,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(11760,36,'cp 83','c8b7 d2d3 h7h6 a2a4 b5b4 c3d5 e7g6 a4a5 f8c5 h2h4 g6e7 c1e3 c6d4 d5e7 d8e7 f3d4 e5d4 e3f4 e8c8 d1g4 g7g5 f4g3 f7f6 e1c1 c8b8 c1b1 c5d6 b3a4 d6e5 h4g5 f6g5 g3e5 e7e5 a4d7 e5a5 g4f5');
INSERT INTO evaluations VALUES(12153,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(12153,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(12153,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(12153,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(12153,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(12153,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(12153,36,'cp 36','f6h5 f4g5 f8e7 g5e7 d8e7 f1d3 h5f6 e1g1 e8g8 c4d5 e6d5 h2h3 f6e4 d1c2 f8e8 b2b4 d7f6 b4b5 c6b5 d3b5 c8d7 b5d7 e7d7 c3e4 f6e4 f3e5 d7e6 c2b2 b7b6 a2a4 e4d6 f1c1 e8c8 a4a5 b6b5 a5a6 f7f6 e5d3 c8c4 d3c5');
INSERT INTO evaluations VALUES(12378,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(12378,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(12378,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(12378,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(12378,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(12378,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(12378,36,'cp -109','f6e4 c3e4 d7d5 d1e2 d5c4 e2c4 b7b6 d2d3 c8b7 c1e3 f8e7 e1c1 e8g8 e4c3 e6e5 h1e1 d8d7 c4d5 a8d8 d5d7 d8d7 a3a4 f7f5 c1b1 f8e8 e3c1 e7f6 f3d2 c6d4 f2f3 f6h4 g2g3 h4f6 f3f4 g8f7 f4e5 f6e5 c3b5 e5f6 e1e8 f7e8 b5d4');
INSERT INTO evaluations VALUES(14742,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14742,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14742,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(14742,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(14742,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(14742,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(14742,36,'cp 27','g1f3 g7g6 f1b5 f8g7 f3e5 c8d7 b5c6 d7c6 e5c6 b7c6 b1c3 e8g8 e1g1 a7a5 c1d2 h7h5 h2h3 f6e4 c3e4 d5e4 d2c3 a5a4 b2b4 f7f5 a1b1 d8d5 d1e2 f8f7 b4b5 c6b5 b1b5 f5f4 e3f4 f7f4 c5c6 d5c6 b5c5 c6a6 e2a6 a8a6 c5c7 f4f7 c7c4 f7b7 f1d1 g7f8 d4d5 e6d5 d1d5 a4a3 c4e4');
INSERT INTO evaluations VALUES(14743,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(14743,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(14743,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(14743,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(14743,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(14743,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(14743,36,'cp 15','g7g6 f1b5 f8g7 f3e5 c8d7 b5c6 d7c6 e5c6 b7c6 e1g1 a7a5 b1c3 d8c8 d1c2 e8g8 b2b3 c8e8 c1b2 e6e5 d4e5 e8e5 c3e2 e5e7 e2d4 e7d7 d4f3 f8b8 b2d4 d7e7 a1b1 h7h5 f1e1 b8e8 b1d1 e7e6 d1c1 f6e4 d4g7 g8g7 c1d1 a5a4 b3a4');
INSERT INTO evaluations VALUES(18964,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(18964,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(18964,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(18964,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(18964,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(18964,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(18964,36,'cp 104','d2d3 d7d6 g1e2 g8f6 e1g1 a7a6 a2a4 c8d7 e2g3 f8e7 f2f4 e5f4 c1f4 e8g8 h2h3 a8c8 g1h1 g8h8 f4e3 c8b8 d1d2 c6e5 a4a5 e5c6 c3e2 d7e6 e2f4 e6c4 d3c4 d8d7 f4d5 b8e8 b2b3 h8g8 a1e1 f6d5 c4d5 c6e5 g3f5 e7f6');
INSERT INTO evaluations VALUES(18965,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(18965,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(18965,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(18965,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(18965,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(18965,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(18965,36,'cp 91','d7d6 g1e2 f8e7 e1g1 g8f6 f2f4 a7a6 a2a4 c8d7 e2g3 e5f4 c1f4 e8g8 d1d2 a8c8 g1h1 g8h8 h2h3 c6b4 a1e1 d7e6 f4e3 d6d5 e4d5 f6d5 e3g1 e7h4 c3e4 d8c7 g1h2 b4c6 g3f5 h4e1 f1e1 c7d7 e4c5');
INSERT INTO evaluations VALUES(20203,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(20203,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(20203,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(20203,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20203,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(20203,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(20203,36,'cp 18','b5c6 b7c6 d2d3 g8e7 h2h4 e7f5 h4h5 f8e7 g2g4 f5d4 b1d2 e8g8 f3d4 c5d4 f2f4 f7f5 g4f5 f8f5 d1g4 c6c5 d2f3 a7a5 c1d2 c8d7 e1c1 d8f8 d1g1 g8h8 g1g2 f8f7 g4h3 d7e8 h5h6 g7g6 f3g5 e7g5 g2g5 a5a4 g5f5 f7f5 h3g2 h8g8');
INSERT INTO evaluations VALUES(20147,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20147,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20147,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(20147,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20147,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(20147,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(20147,36,'cp 19','b7c6 d2d3 g8e7 h2h4 e7f5 h4h5 f8e7 c1f4 a8b8 b2b3 h7h6 b1d2 c8a6 a2a3 e8g8 e1g1 c5c4 d3c4 d5c4 d2c4 d8c7 d1e2 f8d8 f1d1 c6c5 d1d8 b8d8 a1d1 f5d4 f3d4 c5d4 e2d3 d8d5 d1b1 d5d8 b1e1 d8d5');
INSERT INTO evaluations VALUES(18481,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(18481,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(18481,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(18481,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(18481,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(18481,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(18481,36,'cp 20','b2b3 g8e7 d2d3 c8a6 e1g1 e7g6 f1e1 f8e7 b1c3 e8g8 d1e2 g8h8 c3a4 a6b5 g2g3 f7f6 a1b1 c5c4 e5f6 f8f6 c1g5 c4d3 c2d3 f6f5 h2h4 e6e5 g5e7 d8e7 f3d4 f5f8 h4h5 g6f4 e2e5 e7e5 e1e5 f4d3 e5e2 h7h6 f2f4 d3f4 g3f4 b5e2 d4e2');
INSERT INTO evaluations VALUES(26053,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(26053,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(26053,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(26053,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(26053,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(26053,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(26053,36,'cp 83','d2d4 a7a6 b5c4 b7b5 c4b3 c5d4 c3d4 c6a5 b3c2 g7g6 b1c3 f8g7 e1g1 d7c6 c3e2 g8f6 e2g3 c6b7 b2b3 a5c6 c1g5 h7h6 g5d2 a8c8 h2h3 e7e6 c2d3 e8g8 a1c1 d8b6 d2e3 b6a5 d1e2 c6e7 f1d1 c8c1 e3c1');
INSERT INTO evaluations VALUES(26046,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(26046,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(26046,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(26046,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(26046,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(26046,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(26046,36,'cp 82','a7a6 b5c6 d7c6 d4d5 c6d7 e1g1 e7e5 d5e6 d7e6 f3g5 e6c8 d1b3 d8c7 c3c4 f8e7 b1c3 g8f6 c1f4 e8g8 a1d1 f8e8 f1e1 h7h6 g5f3 c8e6 a2a4 f6h5 f4e3 e7f6 c3d5 e6d5 e4d5 f6e5 e3d2 e5d4 e1e8 a8e8 f3d4 c5d4 b3d3 h5f6 h2h3 f6e4');
INSERT INTO evaluations VALUES(42033,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42033,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42033,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(42033,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42033,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(42033,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(42033,36,'cp 0','c7c5 g1f3 c5d4 d1a4 b8c6 f3d4 c8d7 d4c6 b4c3 b2c3 d7c6 a4a3 d5c4 a1d1 d8c8 f1c4 c6g2 c4b5 g2c6 c3c4 c6b5 c4b5 c8c4 h1g1 a8c8 g1g7 h8g8 g7g8 f6g8 a3b2 f7f6 f4d6 c4e4 b2d4 e4h1 e1e2 c8c2 d1d2 c2c1 d2d1');
INSERT INTO evaluations VALUES(9765,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9765,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(9765,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9765,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9765,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9765,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9765,36,'cp 0','d7d6 e1g1 c5b6 a2a4 g7g5 b2b4 a7a5 c1e3 g5g4 h3g4 c8g4 b1d2 h8g8 e3b6 c7b6 d3d4 a5b4 d1b3 d8d7 d4e5 g4f3 d2f3 g8g2 g1g2 d7g4 g2h2 g4f4 h2g1 f4g4');
INSERT INTO evaluations VALUES(8841,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8841,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8841,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8841,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8841,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8841,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(8841,36,'cp 0','b1d2 e8g8 b2b4 c5b6 a2a4 a7a5 b4b5 c6e7 e1g1 e7g6 f1e1 f8e8 d3d4 e5d4 c3d4 d6d5 e4d5 e8e1 d1e1 f6d5 d2b3 c8e6 c1d2 d8f6 e1e4 c7c6 a1e1 e6f5 e4e2 f5e6 e2e4');
INSERT INTO evaluations VALUES(9999,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(9999,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(9999,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(9999,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(9999,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(9999,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(9999,36,'cp 115','c2c3 b4d6 b1d2 a7a6 b5a4 b7b5 a4c2 d6e7 d3d4 d7d6 e1g1 e8g8 f1e1 f8e8 h2h3 e7f8 d2f1 h7h6 f1g3 a6a5 c2d3 b5b4 d3c2 f6d7 c1d2 d7b6 b2b3 b4c3 d2c3 c6b4 c2b1 b4c6 d4d5 c6e7 b1c2 e7g6');
INSERT INTO evaluations VALUES(9998,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(9998,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(9998,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(9998,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(9998,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(9998,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(9998,36,'cp 91','b4d6 b1d2 a7a6 b5a4 b7b5 a4c2 d6e7 e1g1 e8g8 d3d4 d7d6 h2h3 f8e8 f1e1 c8d7 d2f1 a6a5 f1g3 a5a4 c2d3 a8b8 b2b3 a4b3 a2b3 e7f8 c1g5 h7h6 g5d2 b5b4 d4d5 b4c3 d2c3 c6e7 f3d2 e7g6 d2c4 b8a8 a1a8 d8a8 c3d2 e8b8 d3f1');
INSERT INTO evaluations VALUES(29767,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(29767,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(29767,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29767,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29767,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29767,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(29767,36,'cp 9','f3b7 b8d7 f1e2 g4e2 g1e2 e7e6 c1f4 f8d6 b7c6 a8b8 e1c1 d8e7 f4d6 e7d6 d4d5 d6c6 d5c6 d7e5 e2d4 b8b6 d4b5 e5c6 b5c7 e8f8 a2a3 h7h5 b2b4 b6b7 c7a6 b7d7 a6c5 d7d1 c3d1 h5h4 h2h3 f8e7 h1e1 h8h5 c2c4 h5g5 d1e3 c6d4 c1b2 g5g3 b4b5 f6d7 c5d7');
INSERT INTO evaluations VALUES(28611,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(28611,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(28611,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28611,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28611,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28611,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28611,36,'cp 0','b8d7 f1e2 g4e2 g1e2 e7e6 b7f3 f8e7 b2b3 c7c5 c1b2 c5d4 e2d4 d7e5 f3b7 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(24407,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(24407,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24407,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24407,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24407,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24407,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24407,36,'cp 10','f1e2 g4e2 g1e2 e7e6 c1f4 f8d6 b7c6 a8b8 e1c1 d8e7 f4d6 e7d6 d4d5 d6c6 d5c6 d7e5 e2d4 b8b6 d4b5 e5c6 b5c7 e8e7 h1e1 c6b4 c7d5 b4d5 c3d5 f6d5 d1d5 b6a6 d5c5 h8d8 a2a3 g7g6 e1e3 a6d6 c5c7 d8d7 c7d7 e7d7 c2c3 d6a6 c1c2 a6a4 c2d3 a7a5 c3c4 f7f5 d3c3 d7d6 e3d3 d6c6');
INSERT INTO evaluations VALUES(52188,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(52188,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(52188,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(52188,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(52188,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(52188,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(52188,36,'cp -32','b8c6 d4a4 g8f6 f1b5 f8b4 c1d2 d8e7 b5c6 b4d2 b1d2 d7c6 g1e2 e8g8 e1g1 f8d8 f1d1 a7a5 d2f1 c8g4 f2f3 e7c5 g1h1 g4e6 e2f4 e6c4 a4c2 f6d7 a1c1 c4f1 d1f1 c5c2 c1c2 d7e5 h1g1 d8d7 f1c1 a8d8 g1f1 d7d2 c2d2');
INSERT INTO evaluations VALUES(19091,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(19091,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(19091,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(19091,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(19091,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(19091,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(19091,36,'cp -27','d4a4 g8f6 f1b5 f8b4 c1d2 d8e7 b5c6 b4d2 b1d2 d7c6 g1e2 e8g8 e1g1 f8d8 f1d1 h7h6 a4c2 a7a5 f2f3 a5a4 c2c3 a4a3 b2b3 f6d5 e4d5 e7e2 d1e1 e2b5 d5c6 b7c6 a1c1 b5b6 g1h1 b6f2 d2e4 f2b2 c1c2 b2c3 e4c3 c8f5 c3e4 f5e4 f3e4 d8d3 c2c6 d3d2');
INSERT INTO evaluations VALUES(55310,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(55310,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(55310,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(55310,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(55310,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(55310,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(55310,36,'cp 96','d2d4 e7e6 f1d3 g8f6 d1e2 c7c5 d4c5 f8c5 b1c3 e8g8 c1g5 b8d7 c3e4 d8b6 e1g1 c5e7 a2a4 b6c7 g5h4 f6d5 h4g3 e6e5 d3c4 d5f4 g3f4 e5f4 c2c3 f8d8 a4a5 d7f6 f1d1');
INSERT INTO evaluations VALUES(55236,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(55236,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(55236,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(55236,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(55236,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(55236,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(55236,36,'cp 70','e7e6 f1e2 b8d7 c2c4 b7b6 b1c3 f8d6 d1a4 g8f6 c3b5 c8b7 b5d6 c7d6 c1f4 d8c7 e1g1 e8g8 a1d1 f8d8 d1c1 h7h6 h2h3 f6e4 f3d2 e6e5 d2e4 b7e4 f4e3 e5d4 e3d4 d7c5 a4a3 c5e6 d4e3 c7b7 f2f3 e4c6 f1d1 d6d5 c4d5 c6d5');
INSERT INTO evaluations VALUES(57055,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(57055,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(57055,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(57055,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(57055,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(57055,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(57055,36,'cp -41','g1f3 c7c5 f1g2 d7d5 d2d4 b8c6 c2c4 d5c4 d1a4 c5d4 f3d4 d8d4 g2c6 c8d7 c1e3 d4d6 c6d7 d6d7 a4d7 e8d7 b1d2 e6e5 d2c4 d7e6 a1c1 g8f6 g4g5 f6d5 e3d2 f8e7 h2h4 d5f4 d2f4 e5f4 e2e3 f4e3 c4e3 a8c8 e1e2 f7f6 c1g1 e6f7 e2f3 f6g5 h4g5');
INSERT INTO evaluations VALUES(57606,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(57606,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(57606,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(57606,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(57606,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(57606,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(57606,36,'cp 43','d2d4 e7e6 g1f3 g8f6 c2c4 b7b6 f1e2 f8b4 b1d2 b4e7 b2b3 c8b7 e1g1 c7c5 c1b2 e8g8 a1c1 b8c6 a2a3 d8e8 e2d3 e8h5 f1e1 c6d8 d3b1 d7d6 b3b4 a8c8 d4c5 d6c5 e3e4 f5e4 d2e4 b7e4 b1e4 f6e4 e1e4');
INSERT INTO evaluations VALUES(27911,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(27911,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(27911,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(27911,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(27911,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(27911,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(27911,36,'cp 87','d7d6 f1e1 g8e7 c1e3 b8d7 d1d2 a7a6 a2a4 d7f6 c4d3 e8g8 e3g5 c7c5 e4e5 d6e5 d4e5 f6d5 c3e4 d5b4 h2h4 b4d3 c2d3 b7e4 e1e4 a8a7 h4h5 g6h5 e4h4 a7d7 d3d4 f7f5 e5f6 g7f6 h4h5 e7f5 d2e2 f5d4 f3d4 c5d4 e2e6 g8h8');
INSERT INTO evaluations VALUES(30512,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(30512,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(30512,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(30512,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(30512,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(30512,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(30512,36,'cp 46','c2c4 c6d5 c4d5 e7e6 b1c3 g8f6 c1g5 h7h6 g5h4 e6d5 e4e5 d8e7 g1f3 g7g5 h4g3 f6e4 f1e2 e4g3 h2g3 g5g4 f3h2 d5d4 d1d4 b8c6 d4d5 e7e5 h2g4 e5d5 c3d5 e8c8 g4f6 c6d4 e2d3 d4e6 e1f1 c8b8 a1d1 f8c5 d1d2 h6h5 d3b5 b7d5');
INSERT INTO evaluations VALUES(30510,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(30510,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(30510,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(30510,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(30510,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(30510,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(30510,36,'cp 55','c6d5 c4d5 g8f6 b1c3 e7e6 c1g5 h7h6 g5h4 e6d5 e4e5 d8e7 g1f3 g7g5 h4g3 f6e4 f1e2 e4g3 h2g3 f8g7 e1g1 e8g8 f1e1 b8c6 c3d5 e7d8 e2c4 g5g4 f3d4 c6e5 d4f5 d8g5 d5e7 g8h8 c4b3 a8e8 d1d6 e5g6 d6d7 g6e7 e1e7 e8d8 d7c7 g5f5 c7b7 f5f6 e7f7 f8f7 b3f7');
INSERT INTO evaluations VALUES(30506,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(30506,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(30506,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(30506,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(30506,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(30506,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(30506,36,'cp 83','c4d5 g8f6 b1c3 e7e6 c1g5 e6d5 e4e5 d8e7 g1f3 h7h6 g5h4 g7g5 h4g3 f6e4 f1e2 e4g3 h2g3 g5g4 f3h2 d5d4 d1d4 b8c6 d4d5 e7e5 h2g4 e5d5 c3d5 e8c8 e1c1 d8e8 e2f3 c6e5 g4f6 e8e6 f3e2 f8c5 f2f4 b7d5 f6d5 e5c6 e2h5 c6e7 c1b1 c8d8 a2a3 e7d5 d1d5 h8e8 h5f7');
INSERT INTO evaluations VALUES(30504,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(30504,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(30504,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(30504,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(30504,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(30504,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(30504,36,'cp 63','e7e6 b1c3 g8f6 c1g5 h7h6 g5h4 e6d5 e4e5 d8e7 g1f3 g7g5 h4g3 f6e4 f1e2 e4g3 h2g3 g5g4 f3h2 d5d4 d1d4 b8c6 d4d5 e7e5 h2g4 e5d5 c3d5 e8c8 e1c1 c8b8 h1h5 f8g7 g4f6 c6b4 h5f5 b7d5 f6d5 d8c8 c1b1 c8c2');
INSERT INTO evaluations VALUES(34557,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(34557,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(34557,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(34557,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(34557,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(34557,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(34557,36,'cp 172','f1e2 e6d6 d2d4 c8f5 c2c4 c7c6 e1g1 e7e6 f3h4 f5b1 a1b1 g8f6 b2b4 f8e7 g2g3 e8g8 h4f3 d6c7 c1f4 e7d6 f4d2 f6e4 d2e1 h7h6 d1c2 e4f6 e1d2 d6e7 f1e1 b8d7 a2a4 b7b6 g1g2 f8e8 e2d3 a8d8 c4c5 c7c8 a4a5 b6b5');
INSERT INTO evaluations VALUES(34558,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(34558,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(34558,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(34558,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(34558,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(34558,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(34558,36,'cp 158','e6d6 d2d4 c8f5 c2c4 c7c6 e1g1 e7e6 f3h4 f5b1 a1b1 f8e7 h4f3 g8f6 b2b4 d6c7 g2g3 b8d7 d1c2 e8g8 c1f4 e7d6 f4d2 f8e8 a2a4 d6f8 f1e1 b7b6 c4c5 f6d5 e2c4 h7h6 h2h4 a8d8 c4d5 e6d5 d2f4 c7c8 g1g2 b6c5');
INSERT INTO evaluations VALUES(35797,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(35797,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(35797,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(35797,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(35797,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(35797,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(35797,36,'cp 44','a2a4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 e7f6 c2c3 f6d4 c3d4 b8d7 e4c2 f8e8 d4d5 c6d5 d1d5 d7f6 d5d8 e8d8 a4a5 c8e6 f1d1 d8c8 c2d3 e6b3 d1e1 b3c4 d3c2 c8c5 c1e3 c5b5 e3d4 a7a6 e1c1 c4e6 c1d1 a8c8 c2d3 b5d5 d4e3');
INSERT INTO evaluations VALUES(52103,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(52103,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52103,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52103,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(52103,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(52103,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(52103,36,'cp 22','g8f6 f3b3 d8e7 d2d3 b8c6 g1e2 c6a5 b3c3 a5c4 c3c4 d7d6 c4b3 c8e6 c2c4 g7g6 b3b5 e7d7 a2a4 f8e7 b1c3 a7a6 b5d7 f6d7 h2h4 h7h5 c3d5 e6d5 e4d5 f7f5 c1g5 a8b8 e1d2 b7b5 a4b5 a6b5 c4b5 b8b5 a1a8 b5b8');
INSERT INTO evaluations VALUES(44317,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(44317,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44317,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(44317,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(44317,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44317,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(44317,36,'cp 22','f3b3 d8e7 d2d3 b8c6 g1e2 c6a5 b3c3 a5c4 c3c4 d7d6 c4b3 c8e6 c2c4 g7g6 b1c3 b7b6 b3a4 e6d7 a4c2 f8g7 f2f3 d7e6 c2a4 e6d7 a4d1 d7c6 a2a3 e8g8 b2b4 f6h5 b4b5 c6b7 e1g1 a8e8 a3a4 f7f5 a4a5');
INSERT INTO evaluations VALUES(26462,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(26462,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(26462,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(26462,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(26462,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(26462,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(26462,36,'cp 97','f4g3 c6b4 b1a3 e7e6 c2c3 b4c6 d1b3 a8b8 f1e2 f8d6 e1g1 a7a6 c3c4 g8e7 c4d5 e6d5 a3b1 e8g8 b1c3 c6a5 b3a4 d6g3 h2g3 c7c6 f3h4 b7b5 a4d1 f5e6 a1c1 a5b7 e2d3 e6f7 d1c2 g7g6 c3e2 b7d6 h4f3');
INSERT INTO evaluations VALUES(44626,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(44626,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44626,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44626,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44626,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44626,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44626,36,'cp 0','d4a4 b8c6 g1f3 d8c7 b1c3 e7e6 c3b5 c7b8 f1d3 f8c5 c2c3 e6e5 c3c4 d7d6 b5c3 c8g4 a4d1 c6d4 d1a4');
INSERT INTO evaluations VALUES(44818,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(44818,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(44818,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(44818,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(44818,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(44818,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(44818,36,'cp 72','g5f6 e7f6 c4d5 c6d5 e2e3 b8c6 g1e2 f8b4 e2f4 e8g8 a1c1 c8e6 f1e2 a8c8 e1g1 c6a5 c3b5 c8c1 d1c1 a5c6 c1c2 a7a6 a2a3 b4d6 b5d6 d8d6 c2a4 f8c8 f1c1 e6d7 c1c5 c6e7 a4b4 c8c5 d4c5 d6c7 e2f3 d7c6 b4d2 c7e5 h2h3 g7g6 f4e2');
INSERT INTO evaluations VALUES(44816,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(44816,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(44816,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(44816,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(44816,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(44816,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44816,36,'cp 65','e7f6 c4d5 c6d5 e2e3 c8e6 d1b3 b8d7 b3b7 a8b8 b7a7 b8b2 f1b5 f8b4 g1e2 b2c2 e1g1 b4c3 a1c1 c2e2 b5e2 c3a5 c1c6 e8g8 f1b1 f6f5 b1b5 d8a8 c6a6 a8a7 a6a7 a5d2 b5b2 d2e1 a2a4 d7f6 b2b1 e1d2 a4a5 f6e4 e2d3 f8c8 a5a6 c8c3 a7b7 c3d3 a6a7 d3a3');
INSERT INTO evaluations VALUES(45553,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(45553,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(45553,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(45553,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(45553,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(45553,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(45553,36,'cp 64','c4d5 c6d5 e2e3 b8c6 g1e2 f8b4 e2f4 e8g8 a1c1 c8e6 f1e2 a8c8 e1g1 c6a5 c3b5 c8c1 d1c1 a5c6 c1c2 a7a6 a2a3 b4d6 b5d6 d8d6 c2a4 e6d7 f1c1 g7g6 e2f3 c6e7 a4b3 d7c6 c1c5 f8d8 a3a4 d6d7 a4a5 g8g7 b3b4 d8c8 b4b6 h6h5 h2h4');
INSERT INTO evaluations VALUES(56905,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(56905,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(56905,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(56905,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(56905,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(56905,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(56905,36,'cp 86','d2d4 f8g7 d4e5 g7e5 e2e4 d7d6 d1c2 b8c6 g1f3 e5g7 c1e3 g8f6 f1e2 e8g8 a1d1 f8e8 e1g1 b7b6 f1e1 c8b7 a2a3 a7a5 h2h3 d8c8 e3g5 h7h6 g5d2 c8d8 e2f1 f6d7 b2b3 c6e5 f3e5 d7e5 d2e3 e5d7 f2f3 g8h7 e3f2 g7e5 c3d5 b7d5 c4d5 d7c5');
INSERT INTO evaluations VALUES(15264,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(15264,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(15264,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(15264,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(15264,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(15264,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(15264,36,'cp 128','e2e3 h5f4 e3f4 d8e7 d1d2 g7g5 e1c1 g5f4 d2f4 c8e6 g1f3 e8c8 c1b1 c8b8 b5d4 c6d4 f3d4 e6c8 g2g4 c7c5 d4f3 e7c7 f1d3 f7f6 f4f6 c8g4 f6h8 g4f3 e5e6 f3h1 d1h1 c5c4 d3f5 f8c5 h8h7 c7h7 f5h7 d8e8 h2h4 e8e6 f2f4 e6e2 h7g6 c5d6 h4h5 d6f4 h5h6 e2h2 h1h2 f4h2');
INSERT INTO evaluations VALUES(15263,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(15263,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(15263,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(15263,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(15263,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(15263,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(15263,36,'cp 86','h5f4 e3f4 d8e7 c2c3 g7g5 g1f3 g5f4 d1d5 c8e6 d5d2 a7a6 b5d4 c6d4 d2d4 h8g8 e1c1 a8d8 d4a4 c7c6 f3d4 g8g5 f1c4 e6c4 a4c4 g5e5 c4b3 f8h6 d4f3 d8d1 h1d1 e5e4 d1d2 e8f8 b3d1 e7c5 d1c2 c5c4 b2b3 c4f1 c1b2 e4e8 c2h7 h6g7 g2g3 e8e2 h7d3 e2d2 d3d2 g7f6 g3f4 f1g2 f3d4 g2d5');
INSERT INTO evaluations VALUES(15269,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(15269,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(15269,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(15269,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(15269,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(15269,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(15269,36,'cp 100','e3f4 d8e7 d1d5 c8e6 d5d2 a8d8 f1d3 e6c4 b5a3 c4d5 e1f1 c6b4 h2h4 b4d3 c2d3 e7d7 h1h3 f8e7 h4h5 e8g8 h5h6 g7g6 h3g3 f7f6 g1f3 f6e5 f3e5 d7f5 a1e1 e7d6 g3g5 f5f4 d2f4 f8f4 e5g6 d5g2 g5g2 h7g6 a3c4 f4f6 c4d6 d8d6 e1e7 d6d3 f1e2');
INSERT INTO evaluations VALUES(15267,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(15267,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(15267,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(15267,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(15267,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(15267,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(15267,36,'cp 94','d8e7 d1d5 c8e6 d5d2 a8d8 f1d3 e6c4 b5a3 c4d5 e1f1 c6b4 h2h4 b4d3 c2d3 e7d7 a3c2 d5c4 c2e1 c4e6 g1f3 f8e7 e1c2 e8g8 c2e3 c7c5 h4h5 h7h6 d2c2 f7f6 a1d1 f6e5 f3e5 d7d4 e5g6 f8e8 g6e7 e8e7 h1h4 b7b6 a2a3 d4d7 d1d2 e7e8 f1g1 e6f7 b2b3');
INSERT INTO evaluations VALUES(20091,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(20091,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(20091,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20091,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20091,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(20091,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20091,36,'cp 0','g8f6 g1f3 e7e6 c1f4 a7a6 e4e5 d6e5 d1d8 e8d8 f3e5 c6e5 f4e5 c8d7 e1c1 d8e8 f2f4 f8e7 f1d3 a8d8 e5c7 d8c8 c7e5');
INSERT INTO evaluations VALUES(37245,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(37245,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37245,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(37245,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(37245,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37245,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(37245,36,'cp 1','c7c6 g1e2 d7d5 c4d5 f6d5 d2d4 e5d4 c3d5 c6d5 e3d4 c5b6 e1g1 c8g4 h2h3 g4e2 d1e2 b8c6 c1e3 c6d4 e3d4 b6d4 a1d1 d8b6 b2b3 a8d8 h3h4 g7g6 h4h5 f8e8 e2d3 d4f6 g2d5 g8g7 d3c4 d8d7 d5f3 d7c7 c4a4 e8e5 d1d7 g6h5 d7c7 b6c7 a4a7');
INSERT INTO evaluations VALUES(35664,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(35664,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(35664,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35664,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(35664,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35664,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(35664,36,'cp 12','g1e2 c5b6 d2d4 e5d4 e3d4 d7d5 c4d5 f6d5 c3d5 c6d5 e1g1 c8g4 h2h3 g4e2 d1e2 b8c6 c1e3 c6d4 e3d4 b6d4 a1d1 d8b6 d1d3 a8d8 f1d1 f8e8 e2d2 d4f6 g2d5 d8d7 h3h4 b6b2 d2b2 f6b2 d5f3 d7d3 d1d3 b2f6 d3d7 e8e1 g1h2 g7g6 d7b7 e1e7 b7b5 g8g7 a2a4 h7h5 b5a5 f6d4');
INSERT INTO evaluations VALUES(35665,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(35665,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(35665,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(35665,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(35665,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(35665,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(35665,36,'cp 11','d7d5 c4d5 f6d5 c3d5 c6d5 d2d4 e5d4 e3d4 c5b6 e1g1 b8c6 c1e3 c8g4 h2h3 g4e2 d1e2 c6d4 e3d4 b6d4 a1d1 d8f6 d1d3 d4b2 f1b1 b2d4 b1b7 a8d8 e2d2 d4c5 g2d5 d8d5 d3d5 f6f3 d5c5 f3b7 c5e5 f8b8 d2d5 b7d5 e5d5 g7g6 d5a5 b8b7 g1g2 b7c7 h3h4 g8g7 a5a4');
INSERT INTO evaluations VALUES(36110,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(36110,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36110,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(36110,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36110,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36110,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36110,36,'cp 8','c4d5 f6d5 d2d4 e5d4 e3d4 c5b6 c3d5 c6d5 e1g1 b8c6 c1e3 c8g4 h2h3 g4e2 d1e2 c6d4 e3d4 b6d4 a1d1 d8b6 d1d2 a8d8 f1d1 f8e8 e2g4 d4f6 g2d5 g7g6 h3h4 e8e5 g1g2 g8g7 d5b3 d8d2 d1d2 e5c5 g4a4 b6c6 a4c6 c5c6 d2d7');
INSERT INTO evaluations VALUES(36109,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(36109,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(36109,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(36109,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(36109,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36109,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(36109,36,'cp 9','f6d5 d2d4 e5d4 e3d4 c5b6 c3d5 c6d5 e1g1 c8g4 h2h3 g4e2 d1e2 b8c6 c1e3 c6d4 e3d4 b6d4 a1d1 d8b6 d1d3 g7g6 f1d1 a8e8 e2d2 d4f6 g2d5 b6b2 d2b2 f6b2 d5b7 e8e7 d3b3 b2f6 b7a6 g8g7 b3f3 e7c7 d1d5 f6e7 g1g2 f8e8');
INSERT INTO evaluations VALUES(41320,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41320,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41320,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41320,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41320,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41320,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41320,36,'cp 16','f1e2 c7c5 g1f3 e8g8 e1g1 d8b6 d1c1 b8c6 c2c3 c8e6 b1d2 a8c8 d4c5 b6c5 e3e4 f8d8 f1e1 c5b6 a2a4 b6c7 e2f1 h7h6 d2c4 e6c4 f1c4 c6e5 c4f1 e5f3 g2f3 e7e5 f4h6 g7h6 c1h6 c7b6 h6c1 a7a6');
INSERT INTO evaluations VALUES(43445,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(43445,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43445,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(43445,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(43445,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(43445,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(43445,36,'cp 40','b8d7 a2a4 h7h6 g5e3 d8c7 f3d2 f8e7 a4a5 d7c5 f2f3 c8e6 d2c4 e8g8 c4b6 a8b8 f1d3 c5d7 e1g1 d7b6 e3b6 c7d7 d1d2 b8c8 f1d1 c8c6 c3a4 d6d5 e4d5 f6d5 d3e4 d5b6 a4b6 d7d2 d1d2 c6c7 b6d5 e6d5 d2d5 f7f5 e4d3 e5e4 f3e4 f5e4');
INSERT INTO evaluations VALUES(44327,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(44327,31,'cp 247',NULL);
INSERT INTO evaluations VALUES(44327,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(44327,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(44327,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(44327,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(44327,36,'cp 237','f3e5 b8c6 e5c6 b7c6 e4e5 d8e7 d1e2 f6d5 g2g3 d7d6 f1g2 e7e5 e2e5 d6e5 c3d5 c6d5 g2d5 a8b8 b2b3 f7f6 h2h4 f8d6 d2d3 e8e7 c1d2 c8e6 d5e4 e7f7 e4g2 b8d8 e1g1 h7h5 f1e1 d6e7 e1e2 e6g4 e2e4 d8d4 d2e3 d4e4 g2e4 g4e6 e3d2 h8d8 e4f3 e6g4 f3g2 g4e6');
INSERT INTO evaluations VALUES(44306,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(44306,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(44306,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(44306,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(44306,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(44306,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(44306,36,'cp 219','b8c6 e5c6 b7c6 e4e5 d8e7 d1e2 f6d5 b2b3 d5c3 d2c3 a7a5 a2a4 g7g6 c1f4 c8a6 e2e3 a6f1 f4g5 e7e6 h1f1 c5c4 g5f6 h8g8 e3b6 c4b3 c2b3 f8d6 f2f4 d6e7 e1c1 e7f6 e5f6 e6f6 b6c5 d7d5 f4f5 g6f5 f1f5 f6e6 c5f2 a8d8 c1b2 g8g6');
INSERT INTO evaluations VALUES(54039,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(54039,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(54039,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(54039,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(54039,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(54039,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(54039,36,'cp 87','b8c6 e4d5 e6d5 d2d4 g8f6 e1g1 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 c8g4 b5c6 b7c6 f3e5 g4h5 e5d7 f8d8 d5f6 g7f6 d3g3 h5g6 d7c5 h6g5 g3c3 g5e5 c3e5 f6e5 c2c3 d8d2 b2b4 a8d8 a2a4 d2c2 a1c1 d8d2 c1c2 d2c2 c5d7 c2c3 d7e5 g6f5 f1e1 f7f6');
INSERT INTO evaluations VALUES(20183,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(20183,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(20183,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(20183,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(20183,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(20183,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(20183,36,'cp 118','e4d5 e6d5 e1g1 g8f6 d2d4 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6g6 e4g6 h7g6 a1d1 c8f5 c2c3 f5g4 b5e2 a8c8 d1d2 c6e7 f3d4 g4d7 e2f3 e7d5 f3d5 b7b5 f1d1 d7e8 d5f3 c5e7 d4b3 d8d2 d1d2 e7f6');
INSERT INTO evaluations VALUES(20173,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(20173,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(20173,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(20173,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(20173,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(20173,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(20173,36,'cp 109','e6d5 e1g1 g8f6 d2d4 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6g6 e4g6 h7g6 f1d1 c8g4 b5c6 b7c6 d5e3 g4f3 g2f3 f7f6 g1g2 a8b8 b2b3 g6g5 d1d8 b8d8 a1d1 d8d1 e3d1 g8f7 g2f1 f7e6 f1e2 c5e7 d1b2 e6d5 c2c3 e7d6 h2h3 d6e5');
INSERT INTO evaluations VALUES(6777,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(6777,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(6777,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(6777,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(6777,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(6777,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(6777,36,'cp 109','d2d4 g8f6 d4e5 c6e5 c3b5 e5f3 g2f3 d6b4 c1d2 b4d2 d1d2 e8g8 e1c1 d7d5 e4d5 a7a6 b5c3 c8f5 d5d6 c7d6 d2d6 d8a5 f1c4 b7b5 c4b3 f5g6 h1e1 a8d8 d6f4 h7h6 a2a3 d8d1 e1d1 f8d8 h2h4 d8d1 c3d1 a5d8 d1e3 g8h7 f4b4 d8c7 f3f4 c7b6 f4f5');
INSERT INTO evaluations VALUES(6773,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(6773,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(6773,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(6773,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(6773,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(6773,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(6773,36,'cp 80','e5d4 f3d4 g8e7 c1e3 e8g8 d1d2 c6d4 d2d4 e7c6 d4d2 f8e8 f2f3 d8h4 e3f2 h4f6 e1c1 d6b4 f2e1 d7d6 a2a3 b4a5 f1e2 c8e6 h2h4 a5c3 d2c3 f6c3 e1c3 h7h5 h1g1 a7a6 b2b3 a8d8 a3a4 e6d7 c1d2 d7e6 d2e3 d6d5 c3b2 f7f6 e4d5 e6d5 e3f2 c6b4');
INSERT INTO evaluations VALUES(6848,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(6848,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(6848,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(6848,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(6848,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6848,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(6848,36,'cp 6','c2c3 b4c6 d2d4 e5d4 c3d4 c5b6 e1g1 c6a5 c4d3 g8e7 b1c3 d7d6 c3a4 e8g8 d4d5 c8g4 a4b6 a7b6 a1b1 e7g6 h2h3 g4f3 d1f3 g6e5 f3e2 f8e8 c1d2 c7c6 a2a4 e5d7 d2e3 d8f6 f2f4 c6d5 e4d5 d7c5 b1b6 c5d3 e2d3 f6g6');
INSERT INTO evaluations VALUES(6847,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(6847,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6847,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6847,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6847,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6847,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(6847,36,'cp 5','b4c6 d2d4 e5d4 c3d4 c5b6 b1c3 d7d6 e1g1 c6a5 c4d3 g8e7 c3d5 e8g8 d5b6 a7b6 d4d5 c7c6 c1b2 b6b5 b2c3 c8g4 h2h3 g4f3 d1f3 e7g6 f3g3 d8h4 g3h4 g6h4 f1b1 h4g6 a2a4 a5c4 a4b5 c6d5 a1a8 f8a8 e4d5 c4b6');
INSERT INTO evaluations VALUES(7306,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(7306,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(7306,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(7306,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(7306,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(7306,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7306,36,'cp 12','e1g1 d7d6 b2b4 c5a7 b1d2 e8g8 h2h3 c8e6 c4b3 h7h6 f1e1 e6b3 d2b3 d8d7 a2a4 a8e8 c1e3 a7e3 e1e3 c6e7 d1c2 e7g6 d3d4 f6h5 d4e5 g6e5 b3d4 e5c4 e3e1 h5f6 d4f5 d6d5 e4d5 f6d5 a4a5 g8h8 f5d4 e8e1 a1e1 f8e8 e1e8 d7e8');
INSERT INTO evaluations VALUES(7305,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(7305,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7305,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(7305,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7305,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7305,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7305,36,'cp 0','d7d6 b1d2 e8g8 h2h3 h7h6 f1e1 c8e6 b2b4 c5a7 c4b3 f8e8 a2a4 d8d7 b3c2 e6h3 g2h3 d7h3 e1e2 e8e6 d1f1 h3h5 f1g2 f6d5 e4d5 e6g6 d5c6 g6g2 g1g2 h5g4 g2h2 g4h5 h2g3 h5g6 g3h3 g6h5');
INSERT INTO evaluations VALUES(7172,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(7172,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(7172,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7172,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7172,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7172,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(7172,36,'cp 17','c4b3 c5a7 b1d2 h7h6 f1e1 c6a5 d2f1 a5b3 a2b3 e8g8 f1g3 f8e8 h2h3 b7b5 d3d4 c8b7 d4d5 a6a5 c3c4 c7c6 f3h4 b5c4 b3c4 c6d5 e4d5 b7a6 b2b3 a7d4 a1b1 d4c5 h4f5 a6c8 c1d2 c8f5 g3f5 h6h5 d1f3');
INSERT INTO evaluations VALUES(16608,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(16608,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(16608,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(16608,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(16608,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(16608,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(16608,36,'cp -21','e3e4 d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 e8g8 e1g1 h7h6 c2c3 d8f6 f1e1 c8f5 a2a4 c6a5 f3e5 a8d8 e5g4 f6h4 h2h3 c7c5 d1f3 f5g4 f3g4 h4g4 h3g4 c5d4 c3d4 a5b3 a1b1 b3d4 g4g5 h6h5 b5d3 d4e6 a4a5');
INSERT INTO evaluations VALUES(16506,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16506,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(16506,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(16506,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(16506,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(16506,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(16506,36,'cp -11','d7d5 e4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 e5d4 e1g1 e8g8 c3d4 h7h6 c2c3 d8f6 f1e1 c8g4 h2h3 g4h5 b5d3 f8e8 c1e3 c6a5 a1b1 b7b6 b1b5 h5g6 c3c4 a8d8 d3g6 f6g6 d1a4 a5c6 f3h4 g6e6 h4f3 e6d7 a4c2 c6a5 a2a4');
INSERT INTO evaluations VALUES(29491,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(29491,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(29491,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(29491,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(29491,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(29491,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(29491,36,'cp 65','f6d5 d1b3 d5c3 b2c3 b8d7 g1f3 a8b8 c1g5 h7h6 g5h4 g7g5 h4g3 f8g7 e2e3 e8g8 f1e2 d7f6 a1c1 f6e4 g3e5 e7e6 e5g7 g8g7 f3e5 g5g4 b3b2 h6h5 e1g1 e4g5 f2f3 g4f3 e5f3 g5f3 e2f3 f5g6 b2e2 d8g5 e3e4 h5h4 g1h1 f8d8 e4e5 g5h6');
INSERT INTO evaluations VALUES(56372,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(56372,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(56372,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(56372,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(56372,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56372,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(56372,36,'cp -9','d5e4 g1e2 f8b4 c1e3 b8c6 d4e5 d8d1 a1d1 g8e7 e3d2 c8f5 e2g3 e8c8 g3f5 e7f5 c3e4 b4d2 d1d2 c6e5 e1d1 f5h4 h2h3 f7f5 e4c5 g7g5 c5e6 d8d2 d1d2 h7h6 g2g3 h8e8 e6g5 h6g5 g3h4 e5f3 d2c3 g5h4 f1d3 e8e5 a2a4 a7a5 b2b4 a5b4 c3b4 c7c6 b4c3 c8d7 h1d1 d7c7');
INSERT INTO evaluations VALUES(56565,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(56565,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(56565,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(56565,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(56565,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(56565,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56565,36,'cp -8','g1e2 b8c6 c1e3 f8b4 d4e5 d8d1 a1d1 g8e7 e3d2 c8f5 e2g3 f5g4 f1e2 g4e2 c3e2 e8c8 c2c3 b4a5 d2g5 d8d5 g5e7 c6e7 d1d5 e7d5 g3e4 h8e8 e4c5 e8e5 c5d3 e5e7 e1d1 c7c6 e2d4 g7g6 d4b3 a5c7 g2g3 b7b6 h1e1 c8d7 f2f4 e7e1 d3e1 d5f6 d1e2 c6c5 c3c4');
INSERT INTO evaluations VALUES(57181,30,'cp -209',NULL);
INSERT INTO evaluations VALUES(57181,31,'cp -219',NULL);
INSERT INTO evaluations VALUES(57181,32,'cp -202',NULL);
INSERT INTO evaluations VALUES(57181,33,'cp -203',NULL);
INSERT INTO evaluations VALUES(57181,34,'cp -230',NULL);
INSERT INTO evaluations VALUES(57181,35,'cp -179',NULL);
INSERT INTO evaluations VALUES(57181,36,'cp -187','d7d5 g4g5 b8c6 f1g2 c8e6 b1c3 h7h6 d2d3 h6g5 h4g5 h8h1 g2h1 g8e7 c1d2 d8d6 e2e3 e5e4 f2f4 e4d3 c3b5 d6d7 c2d3 e7f5 d1a4 f8c5 d3d4 c5b6 e1c1 e8c8 b5c3 d8h8 g1f3 h8h3 c1b1 f5d6 f3e5 c6e5 d4e5 d7a4 c3a4 d6f5 a4b6 a7b6 b1c1');
INSERT INTO evaluations VALUES(56418,30,'cp -203',NULL);
INSERT INTO evaluations VALUES(56418,31,'cp -181',NULL);
INSERT INTO evaluations VALUES(56418,32,'cp -188',NULL);
INSERT INTO evaluations VALUES(56418,33,'cp -180',NULL);
INSERT INTO evaluations VALUES(56418,34,'cp -199',NULL);
INSERT INTO evaluations VALUES(56418,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(56418,36,'cp -169','g4g5 b8c6 f1g2 c8e6 b1c3 h7h6 d2d3 h6g5 h4g5 h8h1 g2h1 g8e7 a2a3 e7g6 c1d2 f8e7 e2e3 e7g5 d1f3 c6e7 f3g2 g5f6 g1f3 d8d7 f3g5 e6g4 g5h7 e8c8 h7f6 g7f6 f2f3 g4h3 g2g3 h3e6 e1c1 d5d4 e3d4 d7d4 c3e2 d4h4 g3g2 h4a4 d2e3 g6h4');
INSERT INTO evaluations VALUES(56320,30,'cp -170',NULL);
INSERT INTO evaluations VALUES(56320,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(56320,32,'cp -210',NULL);
INSERT INTO evaluations VALUES(56320,33,'cp -182',NULL);
INSERT INTO evaluations VALUES(56320,34,'cp -188',NULL);
INSERT INTO evaluations VALUES(56320,35,'cp -195',NULL);
INSERT INTO evaluations VALUES(56320,36,'cp -206','b8c6 f1g2 c8e6 b1c3 d8d6 d2d3 e8c8 e2e4 d5e4 c3e4 d6d7 c1e3 f7f5 g5f6 g7f6 a2a3 f6f5 e4g5 e6d5 g2d5 d7d5 d1f3 d5d7 f3h3 g8f6 g1e2 f6g4 h1g1 h7h6 g5f3 h8g8 e1c1 f8d6 f3h2 h6h5 e2c3 f5f4 e3d2 g4f2 h3d7 c8d7 d1f1 g8g1');
INSERT INTO evaluations VALUES(4368,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(4368,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(4368,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4368,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(4368,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(4368,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(4368,36,'cp 32','d7d5 c4d5 c6b4 d2d4 e5d4 e3d4 c5e7 e1g1 b4d5 e2f4 d5f4 c1f4 e7d6 f4e3 f8e8 a1c1 h7h6 d1b3 a7a6 f1e1 a8b8 d4d5 d8e7 b3c2 e7e5 e3d2 e5h5 e1e8 f6e8 c3e4 c8f5 c1e1 f5e4 e1e4 e8f6 e4e1 b8e8 e1e8 f6e8 d2c3 e8f6 h2h3 h5g6 c2g6 f7g6');
INSERT INTO evaluations VALUES(7329,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(7329,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(7329,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(7329,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(7329,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(7329,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(7329,36,'cp 66','d2d4 c5a7 e1g1 e8g8 c1g5 h7h6 g5h4 e5d4 c3d4 d7d6 a4c6 b7c6 b1d2 c8g4 d1c2 g7g5 h4g3 d8d7 a1c1 c6c5 d4c5 a7c5 h2h3 g4f3 d2f3 f6h5 g3h2 f8e8 e4e5 h5g7 f1d1 d7e6 c2d2 e6a2 e5d6 c7d6 h2d6 c5d6 d2d6 e8e6');
INSERT INTO evaluations VALUES(7323,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(7323,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(7323,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(7323,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(7323,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(7323,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(7323,36,'cp 62','c5a7 e1g1 e8g8 c1g5 h7h6 g5h4 b7b5 a4b3 d7d6 a2a4 e5d4 a4b5 a6b5 c3d4 g7g5 h4g3 c8g4 e4e5 g4f3 g2f3 d6e5 d1c2 c6d4 c2g6 g8h8 g6h6 f6h7 g3e5 f7f6 e5d4 d8d4 b1c3 b5b4 a1e1 d4h4 h6h4 g5h4 c3d5 f8f7 f3f4 f7g7 g1h1 a7d4 d5b4 d4b2 e1e2 b2d4 f1e1 h7f8 b4c6 g7d7');
INSERT INTO evaluations VALUES(11937,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(11937,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(11937,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(11937,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(11937,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(11937,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(11937,36,'cp -76','h7h6 g5f4 f6e4 d1e2 d8e7 f2f3 e4f6 f4c7 f6d5 e2e7 f8e7 c7g3 d5e3 e1d2 e3f5 g3f2 e7g5 d2d1 f5e3 f2e3 g5e3 f1d3 e8g8 c2c3 d7d5 d1c2 a7a5 b1d2 g7g6 h1e1 e3c5 g2g3 g8g7 f3f4 h6h5 d2f3 a8a7 a2a4 a7c7 a1b1 c8h3 b2b4 a5b4 c3b4 c5d6');
INSERT INTO evaluations VALUES(11905,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(11905,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(11905,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(11905,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(11905,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(11905,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(11905,36,'cp -89','g5f4 f6e4 d1e2 d8e7 f2f3 e4d6 b1d2 e7e2 f1e2 a7a5 e1c1 a5a4 h1e1 e8d8 a2a3 d6f5 d2c4 f7f6 g2g4 f5h4 f4g3 g7g5 g3h4 g5h4 e2d3 d7d5 d3g6 c8d7 e1e2 h8g8 g6h5 f8c5 c4e3 d7e8 e3f5 e8h5 g4h5 g8g5 f5d4 g5h5 e2e6 d8d7 e6f6 a8g8 d4c6 c5d6');
INSERT INTO evaluations VALUES(17286,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17286,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17286,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(17286,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17286,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17286,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17286,36,'cp 13','e2e3 d7d5 f1b5 e7e6 d1e2 c8d7 d2d3 f8c5 b2b3 c6e7 b5d7 d8d7 f3e5 d7d8 c1b2 c7c6 b1d2 e7f5 d3d4 c5e7 e1c1 f5d6 c2c4 d6e4 c1b1 e7b4 d2f3 e4c3 b2c3 b4c3 f3g5 e8g8 e3e4 f6e4 g5e4 d5e4 e2e4 a7a5');
INSERT INTO evaluations VALUES(17285,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(17285,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(17285,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(17285,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17285,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(17285,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17285,36,'cp 7','d7d5 d2d4 c8g4 f1e2 e7e6 e1g1 g4f3 e2f3 c6e7 d1e2 c7c6 b1d2 g7g6 c2c3 h7h5 e3e4 f6e4 d2e4 d5e4 f3e4 d8d6 a2a4 f8g7 h2h3 e7f5 e4f5 g6f5 a4a5 d6d5 a5a6 b7b5 c1e3 g7f6 e3f2 e8c8 a1c1');
INSERT INTO evaluations VALUES(22632,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(22632,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(22632,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(22632,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(22632,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(22632,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(22632,36,'cp 37','d2d4 g8f6 e4e5 f6e4 g1f3 f7f5 e5f6 e4f6 b1c3 b7b6 f1d3 c8b7 f4f5 f8d6 e1g1 d8e7 c3b5 e8c8 f1e1 h8e8 f5e6 d7e6 c2c3 c8b8 d1e2 e6e5 b5d6 c7d6 c1g5 e7c7 e2c2 e5d4 c3d4 h7h6 g5h4 e8e1 a1e1 g7g5 h4g3 c6b4 c2c7 b8c7 d3f5 b4d5');
INSERT INTO evaluations VALUES(22619,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(22619,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(22619,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22619,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(22619,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(22619,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(22619,36,'cp 29','g8f6 e4e5 f6e4 g1f3 f7f5 e5f6 e4f6 b1c3 b7b6 f1d3 c8b7 f4f5 f8d6 e1g1 d8e7 f1e1 e8c8 c3b5 h8e8 c2c3 c8b8 f5e6 d7e6 d3c4 e6e5 c1f4 a7a6 b5d6 c7d6 f4g3 e7c7 c4b3 e5d4 e1e8 d8e8 f3d4 c6e5 d4e6 c7c6 d1e2 e8e7 a1e1 f6e4 g3e5 d6e5');
INSERT INTO evaluations VALUES(29744,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(29744,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(29744,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(29744,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(29744,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(29744,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(29744,36,'cp 160','d1b3 e7e6 b3b7 b8d7 b1c3 f8d6 c4e2 e8g8 e2g4 f6g4 h2h3 g4f6 g1f3 c7c5 e1g1 d8c8 b7c8 f8c8 f1d1 c5d4 f3d4 d6b4 c1d2 d7b6 c3b5 b4d2 d1d2 f6e4 d2c2 c8c2 d4c2 a7a6 b5d4 a8c8 f2f3 e4g3 b2b3 g3e2 g1f1 e2d4 c2d4');
INSERT INTO evaluations VALUES(39547,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(39547,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39547,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39547,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(39547,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(39547,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(39547,36,'cp 23','b8c6 g1f3 e7e5 a2a4 d7d6 d2d3 h7h6 c1e3 c6d4 c3b5 d4f3 d1f3 g8f6 d3d4 a7a6 b5c3 c8g4 f3g3 e5d4 e3d4 f6h5 g3e3 g7d4 e3d4 d8f6 d4f6 h5f6 c3d5 f6d5 c4d5 c7c6 d5a2 g4e6 a2e6 f7e6 e1d2 d6d5 d2e3 d5e4 a1d1 e8e7 e3e4 a8d8 c2c4 h8f8 f2f3 c6c5 d1d8 f8d8 e4e3 b7b6');
INSERT INTO evaluations VALUES(51030,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(51030,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(51030,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(51030,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(51030,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(51030,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(51030,36,'cp 17','d7d5 e4d5 e6d5 b1c3 g8f6 g1f3 f8b4 f1e2 d5c4 e2c4 e8g8 e1g1 b8c6 c1g5 c8g4 c3d5 b4e7 d5e7 d8e7 h2h3 g4f3 d1f3 e7e4 f3e4 f6e4 g5e3 e4f6 a1c1 c6e7 e3f4 c7c6 f4c7 f8e8 f1e1 e7d5 c7h2 a8d8 g1f1 e8e1 c1e1 h7h6 h2e5');
INSERT INTO evaluations VALUES(53361,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53361,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53361,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53361,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(53361,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53361,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53361,36,'cp 7','c4d5 d8a5 b2b4 a5b4 c1d2 b4c5 d1b3 g8f6 f1c4 e7e6 d5e6 c8e6 c4e6 f7e6 g1f3 d4e3 b3e6 f8e7 d2e3 c5c6 e6b3 c6d5 b3b2 f6e4 b1c3 e4c3 b2c3 e8g8 e1g1 e7f6 e3d4 f6d4 f3d4 f8d8 d4f3 b8c6 c3a3 d8d7 a1b1 a7a6 f1c1 a8d8');
INSERT INTO evaluations VALUES(57380,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(57380,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(57380,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(57380,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(57380,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(57380,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(57380,36,'cp 149','d2d4 d7d5 f1d3 b8c6 c2c3 f8d6 b2b4 d5e4 d3e4 g8e7 e1g1 a7a6 a2a3 e8g8 c3c4 f6f5 e4c2 e7g6 f1e1 d6e7 c1b2 e7f6 b2c3 g8h8 d4d5 e6d5 c4d5 c6a7 a3a4 c8d7 c2b3 a7c8 d1c2 f6c3 c2c3 d8f6 c3f6 f8f6 f3d4 c8d6 b1d2 a8d8 d2f3');
INSERT INTO evaluations VALUES(57377,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(57377,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(57377,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(57377,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(57377,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(57377,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(57377,36,'cp 176','d7d5 f1d3 b8c6 a2a3 f8d6 e4d5 e6d5 e1g1 g8e7 c2c4 d5c4 d3c4 c8g4 h2h3 g4f3 d1f3 d8d7 f1e1 c6d4 f3b7 a8d8 b7e4 e8f8 b1c3 d7f5 a1b1 g7g5 g1f1 d4c6 c4b3 f8g7 b3a4 c6e5 a4c2 f5e4 c3e4 e7d5 e4g3 h8e8 g3f5 g7f7 f5d6 d8d6 c2h7 d5f4 h7c2 e8d8 b2b4 e5c4 b1b3 c4d2 c1d2 d6d2');
INSERT INTO evaluations VALUES(56802,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(56802,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(56802,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(56802,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(56802,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(56802,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(56802,36,'cp 157','f1d3 b8c6 b1c3 f8b4 e1g1 g8e7 a2a3 b4c3 b2c3 d5e4 d3e4 b7b6 a3a4 c8b7 d1e2 d8d7 f1e1 e8c8 e4d3 e7g6 e2e6 d7e6 e1e6 c6a5 e6e3 h8e8 e3e8 d8e8 c1e3 b7f3 g2f3 g6h4 f3f4 g7g6 g1f1 h4f5 f1e2 c8d7 d3a6 a5c6 e2d3 d7d6 a1b1 c6a5 c3c4 c7c5 d4c5 b6c5');
INSERT INTO evaluations VALUES(15448,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(15448,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(15448,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(15448,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(15448,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(15448,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(15448,36,'cp 75','f3d4 c8d7 b1c3 f8e7 d4c6 d7c6 b5c6 b7c6 b2b3 a7a5 c1b2 e8g8 d1d4 f8e8 f2f4 d8c8 a1e1 c8g4 d4f2 e7f8 g1h1 g4h5 f2g3 h5g6 g3f3 e8e7 e4e5 d6e5 f3c6 a8e8 f4f5 g6h6 c3e4 f6e4 c6h6 g7h6 e1e4 f7f6');
INSERT INTO evaluations VALUES(15443,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(15443,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(15443,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(15443,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(15443,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(15443,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(15443,36,'cp 81','c8d7 b1c3 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 h2h3 d8b8 a2a4 b8b4 g1h1 f8e8 e4e5 f6d5 c3d5 c6d5 f3d5 a8d8 d5f3 b4c4 c1d2 d6e5 d2a5 e7f6 f4e5 f6e5 a5c3 e5c3 f3c3 c4c3 b2c3 g7g6 a4a5');
INSERT INTO evaluations VALUES(21083,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(21083,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(21083,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21083,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(21083,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(21083,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(21083,36,'cp 45','d2d4 g8f6 g1f3 c8f5 e1g1 e7e6 c2c4 f8e7 b1c3 e8g8 f3e5 a7a5 a2a3 f6e4 e5c6 b7c6 c3e4 f5e4 g2e4 d5e4 c1e3 f7f5 d1c2 e7f6 a1d1 d8e7 f2f3 e6e5 f3e4 e5d4 e3d4 f6d4 d1d4 e7c5 c2c3 f5e4 f1f8 a8f8 g1g2');
INSERT INTO evaluations VALUES(34429,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34429,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34429,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34429,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34429,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34429,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34429,36,'cp 0','c8d7 d4c5 f8c5 d1d2 g8e7 f1d3 e7g6 h2h4 b6c7 d2e2 b8c6 c3b5 c7b8 d3g6 f7g6 c1e3 c5e7 e3g5 e7c5');
INSERT INTO evaluations VALUES(32740,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32740,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32740,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32740,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32740,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32740,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32740,36,'cp 0','d4c5 f8c5 d1d2 g8e7 a2a3 b6c7 b2b4 c5b6 c1b2 a7a6 f1d3 e7g6 e1g1 b8c6 c3a4 c6e5 a4b6 e5f3 g2f3 c7b6 b2g7 h8g8 d2h6 b6d8 h6h7 d8g5 g1h1 g5f4 h7g8 e8e7 g8a8 f4f3 h1g1 f3g4 g1h1');
INSERT INTO evaluations VALUES(45289,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(45289,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(45289,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(45289,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(45289,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45289,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(45289,36,'cp 11','c8d7 b5d3 b8c6 c2c3 d7g4 c1f4 e7e6 h2h3 g4h5 b1d2 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 f8e8 d1b3 d6c7 e1e3 c6a5 b3a4 h5f3 e3f3 a5c6 f3e3 e6e5 d4e5 e8e5 e3e5 c6e5 d3e2 a8e8 e2f1 e5c6 d2f3 a7a6 a4b3');
INSERT INTO evaluations VALUES(28877,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(28877,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(28877,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(28877,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(28877,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28877,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(28877,36,'cp 11','b5d3 d7g4 h2h3 g4h5 c1f4 e7e6 b1d2 f8d6 f4d6 d8d6 c2c3 b8c6 e1g1 e8g8 f1e1 d6c7 d1e2 c6e7 a2a4 c7b6 e2e3 h5g6 f3e5 g6d3 e5d3 e7f5 e3f4 a7a5 g2g4 f5d6 h3h4 a8c8 g4g5 f6h5 f4h2 g7g6 d2f3 c8c4');
INSERT INTO evaluations VALUES(41259,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41259,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(41259,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(41259,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(41259,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41259,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41259,36,'cp 34','e8g8 d2d4 c7c5 d4d5 e7e6 e1g1 e6d5 c4d5 f8e8 f1e1 a7a6 a2a4 b8d7 c1f4 f6g4 h2h3 g4e5 c3e4 d8e7 a4a5 a8b8 f3e5 d7e5 f4g5 f7f6 g5f4 c8d7 d1b3 d7b5 e4c3 b5d7 e2e4 e7f7 f4e5 f6e5 a1d1 h7h6');
INSERT INTO evaluations VALUES(42352,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(42352,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(42352,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42352,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(42352,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(42352,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42352,36,'cp 17','g5h4 c5e7 h4g3 d7d6 a2a4 e8g8 h2h3 g8h8 g1f3 d8e8 b1d2 b8c6 c2c3 c6a5 c4a2 f6h5 b2b4 a5c6 e1g1 a7a6 f3h4 h5g3 f2g3 e7h4 g3h4 e8e7 d1h5 c8e6 a2e6 e7e6 h5e2 h8g8 g1h1 c6e7 h4h5');
INSERT INTO evaluations VALUES(42353,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42353,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42353,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42353,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(42353,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(42353,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42353,36,'cp 0','c5e7 h4g3 d7d6 a2a4 e8g8 h2h3 c8e6 b1d2 b8d7 g1f3 f6h5 g3h2 e7g5 e1g1 g5d2 f3d2 d8g5 d2f3 g5g6 f3h4 g6g5');
INSERT INTO evaluations VALUES(44796,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44796,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(44796,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44796,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(44796,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(44796,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(44796,36,'cp 17','c2c4 b8c6 d4e2 f8g7 b1c3 e8g8 e1g1 e7e6 h2h3 d6d5 e4d5 e6d5 c1g5 c6b4 c4d5 b4d5 d3c4 h7h6 g5f6 d5f6 d1b3 d8b6 b3b6 a7b6 e2f4 g6g5 f4g6 f8e8 f1e1 c8d7 g6e5 d7e6 c4e6 e8e6 e5c4 e6e1 a1e1 b6b5 c3b5 a8a2 b5c3 a2a7 e1d1 g7f8 d1d8 a7a1 g1h2');
INSERT INTO evaluations VALUES(57431,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(57431,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(57431,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(57431,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(57431,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(57431,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(57431,36,'cp 84','d7d5 e4d5 d8d5 d2d4 e5d4 c3d4 d5f7 b1c3 f8d6 f1d3 g8e7 e1g1 e8g8 d3c2 g8h8 h2h3 b8c6 c2b3 f7h5 c3e4 e7d5 e4d6 c7d6 f3g5 h5d1 f1d1 f6g5 b3d5 h7h6 a2a3 c6e7 d5c4 c8d7 c1d2 f8c8 a1c1 e7f5 d4d5 a7a6 c1c3 f5d4 f2f3 d7f5');
INSERT INTO evaluations VALUES(1638,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(1638,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(1638,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(1638,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(1638,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(1638,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(1638,36,'cp 34','c8g4 h2h3 g4h5 g2g4 h5g6 e2b5 e7e6 b5c6 b7c6 d2d3 f8d6 c1d2 g8e7 e1c1 f7f6 d1e1 h7h5 f3h4 e8f7 g4g5 g6f5 c3e4 c6c5 b2b3 a8d8 f2f4 e7d5 e1f1 d5e7 f1g1 h8g8 g1e1 a7a6 e1e2 g8e8 h4f5 e7f5 h1g1');
INSERT INTO evaluations VALUES(27630,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(27630,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(27630,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(27630,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(27630,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(27630,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(27630,36,'cp 31','h2h3 g4h5 g2g4 h5g6 e2b5 e7e6 b5c6 b7c6 d2d3 f8d6 c1d2 g8e7 e1c1 f7f6 f3h4 e8f7 f2f4 h7h5 g4g5 g6f5 c3e4 h8e8 d1g1 e7g6 h4g6 f5g6 b2b3 a8d8 d2e3 d6b4 g5f6 g7f6 g1g3 e8g8 h1g1 g6e4 d3e4 g8g3 g1g3');
INSERT INTO evaluations VALUES(16374,30,'mate 1',NULL);
INSERT INTO evaluations VALUES(16374,31,'mate 1',NULL);
INSERT INTO evaluations VALUES(16374,32,'mate 1',NULL);
INSERT INTO evaluations VALUES(16374,33,'mate 1',NULL);
INSERT INTO evaluations VALUES(16374,34,'mate 1',NULL);
INSERT INTO evaluations VALUES(16374,35,'mate 1',NULL);
INSERT INTO evaluations VALUES(16374,36,'mate 1','h5f7');
INSERT INTO evaluations VALUES(17085,30,'cp 546',NULL);
INSERT INTO evaluations VALUES(17085,31,'cp 557',NULL);
INSERT INTO evaluations VALUES(17085,32,'cp 550',NULL);
INSERT INTO evaluations VALUES(17085,33,'cp 544',NULL);
INSERT INTO evaluations VALUES(17085,34,'cp 548',NULL);
INSERT INTO evaluations VALUES(17085,35,'cp 546',NULL);
INSERT INTO evaluations VALUES(17085,36,'cp 565','g2g4 d7d6 g4h5 d8e7 b1c3 c7c6 c1e3 d6e5 d4a4 e7b4 a4b4 f8b4 e1c1 c8e6 h1g1 a8d8 d1e1 f7f6 g1g7 d8d7 h5h6 h8g8 a2a3 d7g7 h6g7 b4c3 b2c3 g8g7 c1b2 b7b6 f1a6 e8f7 a6d3 g7g6 b2a1');
INSERT INTO evaluations VALUES(17083,30,'cp 561',NULL);
INSERT INTO evaluations VALUES(17083,31,'cp 559',NULL);
INSERT INTO evaluations VALUES(17083,32,'cp 551',NULL);
INSERT INTO evaluations VALUES(17083,33,'cp 546',NULL);
INSERT INTO evaluations VALUES(17083,34,'cp 532',NULL);
INSERT INTO evaluations VALUES(17083,35,'cp 537',NULL);
INSERT INTO evaluations VALUES(17083,36,'cp 540','d8h4 c1e3 d7d6 f1e2 g7g6 g4h5 h4d4 e3d4 f8g7 b1c3 c8d7 e1c1 g7e5 d4e5 d6e5 e2f3 e8c8 h1e1 e5e4 f3e4 d7g4 d1d8 h8d8 e4d3 a7a6 c3e4 g4f5 e4f6 h7h6 d3f5 g6f5 e1e7 b7b5 e7f7 d8d6 f6e8 d6d7 f7f5 d7e7 e8f6 e7e2 f5f3');
INSERT INTO evaluations VALUES(25893,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(25893,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(25893,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(25893,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(25893,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(25893,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(25893,36,'cp 180','d2d4 e7e6 d4d5 c6e7 d1e2 e6d5 e4d5 d7d6 c1f4 g8f6 e1c1 g7g6 f3d4 f8g7 d4c6 b7c6 d5c6 e8g8 e2f3 d8c8 f4g5 c8f5 f3f5 e7f5 g5f6 g7f6 c3d5 f6d8 f1a6 f8e8 g2g4 f5h4 f2f4 g8f8 h1g1 a8b8 g1g3 b6b5 a6b7 e8e2 g3a3 h4g2 a3a7');
INSERT INTO evaluations VALUES(25888,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(25888,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(25888,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(25888,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(25888,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(25888,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(25888,36,'cp 175','e7e6 d4d5 c6e7 d1e2 g8f6 c1d2 a7a6 e1c1 b6b5 h2h4 c7c5 g2g3 d8b6 f3e5 d7d6 e5g4 f6g4 e2g4 e6e5 h4h5 b7c8 g4e2 c8d7 f2f4 b5b4 c3b1 e7c8 f4e5 d6e5 d2e3 c8d6 b1d2 f8e7 f1h3 d7h3 h1h3');
INSERT INTO evaluations VALUES(26528,30,'cp 320',NULL);
INSERT INTO evaluations VALUES(26528,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(26528,32,'cp 341',NULL);
INSERT INTO evaluations VALUES(26528,33,'cp 329',NULL);
INSERT INTO evaluations VALUES(26528,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(26528,35,'cp 333',NULL);
INSERT INTO evaluations VALUES(26528,36,'cp 393','c4d5 f5b1 d5c6 b1e4 c6b7 e4b7 f3e5 e7e6 d1a4 c7c6 e2e3 f8d6 f1a6 d8b6 a6b7 b6b7 a4c6 b7c6 e5c6 e8d7 c6e5 d6e5 d4e5 g8e7 e1e2 h8b8 b2b3 b8b5 f2f4 g7g5 c1d2 g5f4 e3f4 h7h5 a1c1 b5d5 c1c3 d7e8 c3c4 e7f5 h1c1 a8d8 d2b4 d5d4 c4d4 d8d4');
INSERT INTO evaluations VALUES(26515,30,'cp 356',NULL);
INSERT INTO evaluations VALUES(26515,31,'cp 368',NULL);
INSERT INTO evaluations VALUES(26515,32,'cp 382',NULL);
INSERT INTO evaluations VALUES(26515,33,'cp 366',NULL);
INSERT INTO evaluations VALUES(26515,34,'cp 348',NULL);
INSERT INTO evaluations VALUES(26515,35,'cp 373',NULL);
INSERT INTO evaluations VALUES(26515,36,'cp 371','f5b1 d5c6 b1e4 c6b7 e4b7 d1a4 c7c6 e2e3 e7e6 f3e5 f8d6 f1a6 d8b6 a6b7 b6b7 a4c6 b7c6 e5c6 e8d7 c6e5 d6e5 d4e5 g8e7 c1d2 h8b8 d2c3 e7d5 c3d4 b8c8 e1d2 f7f5 e5f6 g7f6 e3e4 e6e5 d4e5 f6e5 e4d5 c8b8 b2b3 a7a5 a1b1 b8b4 h1c1 a8b8 c1c6 a5a4 c6h6 b8c8 b1b2 b4d4 d2e2 a4a3 b2d2 c8c1 h6h7 d7d6');
INSERT INTO evaluations VALUES(27216,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(27216,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(27216,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(27216,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(27216,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27216,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27216,36,'cp 0','g7g5 b5c6 b7c6 d2d4 g5g4 f3d2 f7f6 b1c3 g8e7 d2f1 f8g7 f1g3 h7h5 d4e5 f6e5 c1g5 d8c8 h2h4 g4h3 g3h5 h8g8 h5g7 g8g7 d1h5 e7g6 f2f4 d7g4 h5h6 c8d7 f4f5 g6f8 g2h3 g7h7 h6f6 h7h3 e1e3 h3h7 e3g3 h7f7 f6h6 f7h7');
INSERT INTO evaluations VALUES(47875,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(47875,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(47875,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(47875,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47875,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47875,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(47875,36,'cp 15','f2f4 c7c5 g2g3 e5f4 c1f4 f8e7 g1e2 e8g8 f1g2 b8c6 e1g1 a8b8 d1d2 b7b5 a2a3 h7h6 d3d4 c5d4 e2d4 c6e5 b2b3 c8d7 d4f5 d7f5 e4f5 f8e8 a1d1 a7a6 c3e4 e5g4 e4f6 g4f6 f4d6 d8b6 g1h1 e7d6 d2d6');
INSERT INTO evaluations VALUES(10234,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(10234,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(10234,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(10234,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(10234,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(10234,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(10234,36,'cp -14','d7d5 e4d5 f6d5 b2b4 c5d6 e1g1 e8g8 b4b5 c6e7 c3c4 d5f6 d1c2 e7g6 b1c3 c7c6 b5c6 b7c6 a1b1 h7h6 h2h3 d8e7 c1e3 c8f5 c3e4 f5e4 d3e4 d6c5 e3d2 e7c7 f1d1 f6d7 g2g3 f8e8 d2c3 g6f8 f3d2 c5d4 c3d4 e5d4 c4c5');
INSERT INTO evaluations VALUES(29672,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(29672,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(29672,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(29672,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(29672,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(29672,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(29672,36,'cp 18','f3e5 g4f5 c2c4 c7c6 d1b3 d8b6 c4d5 f6d5 f1g2 b6b3 a2b3 d5b4 b1a3 f7f6 e5c4 b8a6 c1d2 e7e6 e2e4 f5g6 h2h4 h7h5 d2c3 e8c8 e1g1 f8e7 c4e3 c8b8 f2f4 a6c7 a1d1 c6c5 c3b4 c5b4 a3c4 g6f7 g1h2 d8d7 f4f5 h8d8 f5e6 f7e6');
INSERT INTO evaluations VALUES(41935,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(41935,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(41935,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(41935,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(41935,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(41935,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(41935,36,'cp 78','d1b3 b4e7 e2e4 d5c3 b3c3 e8g8 c1f4 b7b6 f4c7 d8d7 c7b8 a8b8 f1d3 c8b7 e1g1 f8d8 a1d1 d7a4 c3d2 b8c8 d2e2 a7a6 d3b1 a4c4 e2e3 a6a5 d1d2 e7b4 d2c2 c4b5 c2c8 d8c8 b1d3 b5a4 a2a3 b4e7 f1b1 e7f6 h2h4 c8d8');
INSERT INTO evaluations VALUES(44636,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(44636,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44636,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44636,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44636,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(44636,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(44636,36,'cp 47','e4e5 f6d5 d1d4 e7e6 f1d3 d7d6 e1g1 b8c6 d4e4 c6e5 f3e5 d6e5 e4e5 c8d7 d3e4 d8b8 e5b8 a8b8 e4d5 e6d5 c1e3 b8c8 b1c3 a7a6 f1e1 d7e6 e3d4 e8d7 a1d1 h8g8 a2a3 d7c6 d1d3 b7b5 c3e2 f8d6 c2c3 c6b7 d4e3 g7g5 d3d1 h7h5 e2d4 h5h4 d4e6 f7e6');
INSERT INTO evaluations VALUES(44605,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(44605,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(44605,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44605,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(44605,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(44605,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(44605,36,'cp 42','f6d5 d1d4 e7e6 f1d3 d7d6 e1g1 b8c6 d4e4 d5b4 f1d1 b4d3 e4d3 d6e5 d3d8 c6d8 f3e5 d8c6 e5c6 b7c6 b1d2 e6e5 d2c4 f7f6 c4d6 f8d6 d1d6 c8f5 d6c6 e8d7 c6a6 f5c2 c1e3 h8b8 b2b3 b8b7 a1c1 c2f5 h2h4 h7h5 c1c5 d7e8 c5a5 e8f7 a6a7 a8a7 a5a7 b7a7 e3a7');
INSERT INTO evaluations VALUES(44673,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(44673,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(44673,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(44673,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44673,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(44673,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(44673,36,'cp 23','d1d4 e7e6 f1d3 b8c6 d4e4 d7d6 e1g1 d5b4 f1d1 b4d3 e4d3 d6e5 d3d8 c6d8 f3e5 d8c6 e5c6 b7c6 b1d2 c8b7 d2c4 c6c5 c1f4 b7d5 c4d6 f8d6 f4d6 a8c8 a1c1 c5c4 b2b4 f7f6 a2a4 e8f7 d6c5 a7a6 d1d4 d5c6 a4a5 h8d8 d4c4 e6e5 h2h4 c6b5 c4c3 h7h5 f2f3 f7e6 c1e1 e6f5');
INSERT INTO evaluations VALUES(44670,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(44670,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44670,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(44670,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(44670,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(44670,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(44670,36,'cp 38','e7e6 f1d3 d7d6 e1g1 b8c6 d4e4 d5b4 f1d1 b4d3 e4d3 d6e5 d3d8 c6d8 f3e5 d8c6 e5c6 b7c6 b1d2 e6e5 d2c4 f7f6 c4d6 f8d6 d1d6 c8f5 d6c6 a7a5 c1e3 e8d7 c6b6 f5c2 a1c1 h8c8 b6b7 d7e6 b7g7 c2e4 c1c8 a8c8 f2f3 e4d3 g7b7 a5a4 b7b6 e6f7 b6d6 d3c4 d6d7 f7g6');
INSERT INTO evaluations VALUES(53597,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(53597,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(53597,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(53597,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(53597,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(53597,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(53597,36,'cp 73','e2e4 f8b4 e4e5 c6c5 a2a3 b4c3 b2c3 g8e7 a3a4 h7h6 h2h4 d8a5 c1d2 a5c7 h4h5 b8c6 f1e2 c6a5 e1g1 a5c4 d2c1 c8d7 f3h4 c4b6 a4a5 b6a4 d1d3 c5d4 c3d4 a4c3 c1b2 c3e2 d3e2 e8g8 b2a3 f8e8 a3c5 g8h8 c5e7 e8e7 f2f4');
INSERT INTO evaluations VALUES(10941,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(10941,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(10941,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(10941,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(10941,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(10941,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(10941,36,'cp 42','f3d4 e8g8 e1g1 d7d6 c1f4 c6e5 c4e2 f8e8 f4g3 e7f8 f1e1 e5g6 e2c4 a7a6 a2a4 b7b6 c3d5 c8b7 d1f3 f6d5 c4d5 b7d5 e4d5 f8e7 d4c6 d8d7 b2b3 e7f6 a1d1 e8e1 d1e1 a8e8 e1e8 d7e8 g1f1 a6a5 h2h3 g6e7 f3e3 h7h5 e3f3 h5h4 g3f4 e8d7 c6e7 d7e7');
INSERT INTO evaluations VALUES(21176,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(21176,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(21176,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(21176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21176,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(21176,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21176,36,'cp 0','e7e5 d4e5 d6e5 d1d8 e8d8 f4g3 h7h5 g1f3 f7f6 b1c3 c8e6 a2a3 g8e7 h2h3 d8e8 e1c1 e7f5 g3h2 a8d8 f1e2 h5h4 f3d2 e8f7 h1e1 f8d6 d2e4 d6e7 c3b5 d8c8 b5c3');
INSERT INTO evaluations VALUES(21936,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21936,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21936,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(21936,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21936,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21936,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21936,36,'cp 0','d4e5 d6e5 d1d8 e8d8 f4g3 h7h5 g1f3 f7f6 b1c3 c8e6 a2a3 g8e7 e1c1 d8e8 h2h3 a8d8 f1e2 e7f5 g3h2 a7a6 c3e4 f5d6 f3d2 d6e4 d2e4 e6d5 e2f3 d5e6');
INSERT INTO evaluations VALUES(21884,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(21884,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21884,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21884,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21884,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21884,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(21884,36,'cp 0','d6e5 d1d8 e8d8 f4g3 h7h5 g1f3 f7f6 b1c3 c8e6 a2a3 g8e7 h2h3 d8e8 e1c1 e7f5 g3h2 a8d8 f1e2 a7a6 f3d2 f8c5 g2g4 h5g4 h3g4 f5d6 h2g3 e8f7 e2f3 c5a7 h1h8 d8h8 c3d5 h8c8 d5c3');
INSERT INTO evaluations VALUES(21827,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21827,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21827,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21827,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(21827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21827,36,'cp -3','d1d8 e8d8 f4g3 h7h5 g1f3 f7f6 b1c3 c8e6 a2a3 g8e7 e1c1 d8e8 h2h3 a8d8 f1e2 e7f5 g3h2 f8c5 f3d2 e8e7 c3a4 c5b6 a4b6 a7b6 e2d3 g7g5 d2e4 f5h4 h1g1 g5g4 b2b3 g4h3 g2h3 h4f3 g1g7 e6f7 h2g3 h5h4 d3c4 d8d1 c1d1 c6d8 c4f7');
INSERT INTO evaluations VALUES(30780,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(30780,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(30780,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(30780,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(30780,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(30780,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(30780,36,'cp 40','g8f6 e4e5 f6e4 d1f3 d7d5 d2d4 e7e6 g1e2 d8d7 c4d5 e6d5 c1f4 b8c6 h2h4 e8c8 g2g3 c8b8 f1h3 d7e8 e1c1 h7h6 c1b1 g7g5 f4c1 f8e7 a2a3 g5h4 h3f5 h4g3 f2g3 e4c3 f3c3 b7a6 e2f4 e7g5 f4h5 a6b5 h1e1 c6a5 f5c2 e8c6 c3c6');
INSERT INTO evaluations VALUES(31047,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(31047,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(31047,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(31047,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(31047,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(31047,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(31047,36,'cp 132','h2h4 h7h6 f3e5 d8d6 d2d4 b8d7 e5g6 d6g6 c2c3 g6d6 f1c4 e7e6 d1f3 d6c7 g3e4 g8f6 c1f4 c7d8 e1c1 f6e4 f3e4 d7f6 e4e2 f8d6 f4d2 a7a5 c4d3 f6d5 g2g3 b7b5 e2g4 d8f6 h1e1 a5a4 a2a3 f6e7 d3c2 e8c8 g4e2 d5b6 c2d3 c8b7 h4h5 e7c7 c1b1 h8e8');
INSERT INTO evaluations VALUES(31046,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(31046,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(31046,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(31046,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(31046,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(31046,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(31046,36,'cp 109','h7h6 f3e5 d8d6 d2d4 b8d7 e5c4 d6d5 f1e2 e7e6 c1f4 f8e7 c2c3 g8f6 a2a4 c6c5 c4e3 d5c6 e2b5 c6c8 d4d5 a7a6 b5d7 c8d7 d5d6 e7d8 e3c4 d7c6 c4e5 c6d5 e5g6 d5d1 a1d1 f7g6 f4e5 e8d7 h4h5 g6g5 e1g1');
INSERT INTO evaluations VALUES(30969,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(30969,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(30969,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(30969,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(30969,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(30969,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(30969,36,'cp 114','f3e5 d8d6 e5g6 d6g6 d2d4 b8d7 c2c3 g6d6 f1c4 d6c7 d1f3 e7e6 g3h5 g8f6 c1f4 c7d8 h5f6 d7f6 e1c1 f8d6 f4d2 f6d5 d1e1 d6e7 c4d3 e7f6 f3h5 g7g6 h5f3 a7a5 c1b1 h6h5 c3c4 d5b4 d3e4 e8g8');
INSERT INTO evaluations VALUES(30961,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(30961,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(30961,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(30961,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(30961,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(30961,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(30961,36,'cp 112','d8d6 d2d4 b8d7 e5c4 d6d5 f1e2 e7e6 e1g1 g8f6 b2b3 g6e4 f1e1 f8e7 g3e4 d5e4 g2g3 e8g8 e2f3 e4f5 c1b2 a7a5 a2a4 f6d5 f3e4 f5h3 e4g2 h3f5 a1c1 f8d8 d1e2 d7f6 c4e5 d8e8 c2c3 e7d6 c1d1 a8d8 d1d3 f5h7 e1d1 h7f5 d3f3 f5h5 d1d3 d6e5 d4e5');
INSERT INTO evaluations VALUES(32876,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(32876,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(32876,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(32876,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(32876,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(32876,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(32876,36,'cp 129','g1f3 e5d6 d2d4 c7c6 e1g1 g8f6 f3h4 f5c8 f1e1 g7g6 e2c4 f8g7 d1e2 e7e6 h4f3 b8d7 a2a4 e8g8 a4a5 b7b5 a5b6 d7b6 b2b3 d6c7 e2e5 c7e5 f3e5 c6c5 c1a3 f6d7 c4b5 c5d4 e5d7 c8d7 a3f8 g7f8 b5d7 b6d7 e1d1 f8b4 c3a4 e6e5');
INSERT INTO evaluations VALUES(32874,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(32874,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(32874,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(32874,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(32874,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(32874,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(32874,36,'cp 126','e5d6 e1g1 c7c6 d2d4 e7e6 f1e1 g8f6 f3e5 b8d7 g2g4 f5g6 h2h4 e8c8 e5c4 d6c7 h4h5 d7b6 c4b6 a7b6 h5g6 f8c5 c1e3 h7g6 e2f3 c7h2 g1f1 e6e5 d4c5 d8d1 a1d1 h2h4 f1e2 f6g4 e1h1 g4h2 c3e4 b6c5 h1h2 h4h2 d1h1 h2h1 f3h1 h8h1 e4d6 c8d7 d6f7 h1a1 a2a3');
INSERT INTO evaluations VALUES(32825,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(32825,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(32825,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(32825,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(32825,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(32825,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(32825,36,'cp 136','d2d4 c7c6 e1g1 g8f6 f3h4 f5c8 f1e1 g7g6 e2c4 f8g7 h2h3 e8g8 h4f3 b7b5 c4b3 a7a5 a2a4 b5b4 c3e4 f6e4 e1e4 e7e6 c1f4 d6e7 f3e5 b8d7 e5c6 e7h4 d1f3 c8b7 f4d6 b7c6 e4h4 c6f3 d6f8 g7f6 f8h6 f6h4 g2f3 h4f6 h6e3');
INSERT INTO evaluations VALUES(48889,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(48889,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(48889,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(48889,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(48889,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(48889,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(48889,36,'cp -99','f8d6 c2c4 d5d4 e3d4 e5d4 g1f3 c7c5 d1e2 d6e7 g2g4 e8g8 f1g2 f8e8 e1g1 b8c6 b1d2 h7h6 e2d1 d8c7 f1e1 c8d7 g4g5 h6g5 f3g5 c6e5 d2e4 e5g6 b2b4 f6e4 e1e4 e7g5 c1g5 d7c6 b4c5 c6e4 g2e4 c7c5 h3h4 e8e5 d1g4 a8f8');
INSERT INTO evaluations VALUES(41738,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(41738,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(41738,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(41738,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(41738,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(41738,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(41738,36,'cp -92','c2c4 d5d4 e3d4 e5d4 g1e2 b8c6 e2g3 e8g8 f1e2 a7a5 b1d2 d6g3 f2g3 d8d6 e1g1 d6g3 d1e1 g3e3 g1h1 e3e7 e2f3 e7e1 f1e1 c8f5 d2b3 f5d3 f3c6 b7c6 b3d4 c6c5 d4b5 a8b8 b5c7 h7h6 c7d5 f6e4 d5e7 g8h7 e7c6 b8e8 c1f4 g7g5 f4c7 d3c4 c6a5 c4b5 h1g1 f7f5');
INSERT INTO evaluations VALUES(24933,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(24933,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(24933,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(24933,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(24933,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24933,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24933,36,'cp 0','g4f3 d1f3 e7e5 f1b5 f8d6 d4e5 d6e5 e1g1 e8g8 f1d1 a7a6 b5e2 e5c3 b2c3 b7b5 a2a4 c6e5 f3f4 e5g6 f4f5 c7c6 c3c4 b5c4 e2c4 f6e4 c4d3 g6h4 f5f4 h4g6');
INSERT INTO evaluations VALUES(24961,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24961,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24961,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24961,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24961,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24961,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24961,36,'cp 0','d1f3 e7e5 f1b5 f8d6 d4e5 d6e5 e1g1 e8g8 f1d1 a7a6 b5e2 e5c3 b2c3 b7b5 a2a4 c6e5 f3f4 e5g6 f4f5 c7c6 c3c4 b5c4 e2c4 f6e4 c4d3 g6h4 f5f4 h4g6');
INSERT INTO evaluations VALUES(24960,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24960,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24960,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24960,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24960,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24960,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24960,36,'cp 0','e7e5 f1b5 f8d6 d4e5 d6e5 e1g1 e8g8 f1d1 a7a6 b5e2 e5c3 b2c3 b7b5 a2a4 c6e5 f3f4 e5g6 f4f5 c7c6 c3c4 b5c4 e2c4 f6e4 c4d3 g6h4 f5f4 h4g6');
INSERT INTO evaluations VALUES(25251,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25251,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25251,36,'cp 0','f1b5 f8d6 d4e5 d6e5 e1g1 e8g8 f1d1 a7a6 b5e2 e5c3 b2c3 b7b5 a2a4 c6e5 f3f4 e5g6 f4f5 c7c6 c3c4 b5c4 e2c4 f6e4 c4d3 g6h4 f5f4 h4g6');
INSERT INTO evaluations VALUES(25782,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(25782,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(25782,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(25782,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(25782,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(25782,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(25782,36,'cp 73','d4e5 d6e5 d1d8 a8d8 c1e3 a7a5 h2h3 g4h5 a2a4 f7f6 b1d2 h5f7 e1e2 h7h5 b2b3 g8e7 f3e1 e7g6 e1d3 d8a8 d2c4 f8d6 e3d2 f7c4 b3c4 g6f8 c4c5 d6e7 d2e3 f8e6 h1b1 e8d7 h3h4 h8d8 f2f3 a8b8 g2g3 d7c8 b1b8 c8b8 a1b1 b8c8 e3f2');
INSERT INTO evaluations VALUES(25777,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(25777,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(25777,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(25777,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(25777,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(25777,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(25777,36,'cp 86','d6e5 d1d8 a8d8 c1e3 a7a5 h2h3 g4h5 b1d2 f7f6 a2a4 h5f7 e1c1 d8b8 b2b3 g8e7 f3e1 e7c8 e1d3 f8a3 c1b1 c8b6 b1a1 b6d7 g2g4 f7e6 h3h4 c6c5 g4g5 a3b4 d3b4 b8b4 a1b2 c5c4 b2a3 c4b3 c2b3 c7c5 d1c1 e8g8 g5f6 f8f6 e3c5');
INSERT INTO evaluations VALUES(25774,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(25774,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(25774,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(25774,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(25774,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(25774,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(25774,36,'cp 67','d1d8 a8d8 c1e3 a7a5 h2h3 g4h5 a2a4 f7f6 b1d2 h5f7 e1e2 f8d6 f3e1 h7h5 e1d3 g8e7 h3h4 f6f5 f2f3 f5e4 f3e4 e8g8 h1f1 e7g6 g2g3 f7e6 b2b3 e6g4 e2e1 d8a8 d2c4 f8f3 f1f3 g4f3 c4d2 f3g4');
INSERT INTO evaluations VALUES(23553,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(23553,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(23553,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(23553,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(23553,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(23553,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(23553,36,'cp 78','a8d8 c1e3 a7a5 h2h3 g4h5 b1d2 f7f6 a2a4 h5f7 e1c1 d8b8 b2b3 f8a3 c1b1 g8e7 f3e1 e7c8 e1d3 c8b6 b1a1 b6d7 g2g4 f7e6 f2f3 h7h5 g4g5 a3d6 a1b2 f6f5 h3h4 e8g8 b2c3 d6a3 d1a1 a3d6 h1f1 f5f4 e3f2 e6f7 c3b2 c6c5 b2c3');
INSERT INTO evaluations VALUES(62,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(62,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(62,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(62,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(62,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(62,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(62,36,'cp 86','c1e3 a7a5 h2h3 g4h5 b1d2 f7f6 a2a4 h5f7 e1c1 d8b8 b2b3 f8a3 c1b1 g8e7 f3e1 e7c8 e1d3 c8b6 b1a1 b6d7 g2g4 h7h5 g4g5 a3d6 a1b2 f7e6 f2f3 c6c5 b2c3 f6f5 h3h4 e8g8 h1g1 g7g6 g1f1 b8a8 e3f2 e6f7 f1g1 f7e6 d2c4 e6c4 c3c4');
INSERT INTO evaluations VALUES(31618,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(31618,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(31618,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(31618,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(31618,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(31618,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(31618,36,'cp 119','f1b5 b8c6 e1g1 f8e7 f1e1 g8f6 e4e5 d6e5 d1d8 a8d8 b5c6 b7c6 f3e5 g4d7 c1g5 e8g8 b1a3 h7h6 g5h4 g7g5 h4g3 f8e8 a3c4 f6d5 a1d1 e7f6 c4d6 f6e5 g3e5 e8e6 f2f3 d7c8 d6c8 d8c8 c2c4 d5b6 b2b3 c8e8 e5g3 f7f6 e1e6 e8e6');
INSERT INTO evaluations VALUES(39511,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(39511,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(39511,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(39511,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(39511,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(39511,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(39511,36,'cp 175','d1h5 g7g6 h5e5 g8f6 b1c3 e8g8 g1f3 c7c6 e1g1 b7b5 c4b3 b5b4 c3a4 c6c5 e5f4 d7d6 d2d4 c5d4 f1d1 c8b7 d1d4 b8c6 d4d1 d8c7 d1e1 a8e8 a2a3 a7a5 a3b4 a5b4 c1d2 e7d8 a1c1 c6a5 e4e5 d6e5 f3e5 f6h5 b3f7 f8f7 f4f7 c7f7 e5f7 g8f7 e1e8 f7e8 d2b4');
INSERT INTO evaluations VALUES(39508,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(39508,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(39508,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(39508,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(39508,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(39508,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(39508,36,'cp 167','g7g6 h5e5 g8f6 b1c3 e8g8 g1f3 c7c6 e1g1 b7b5 c4b3 b5b4 c3a4 d7d6 e5f4 c6c5 d2d4 c5d4 f3d4 c8b7 f2f3 d6d5 f4g3 d5e4 d4f5 f6h5 f5e7 d8e7 g3e1 b8d7 f3e4 b7a6 f1f2 d7e5 h2h3 a6c4 c1d2 a7a5 a4b6 c4b3 a2b3');
INSERT INTO evaluations VALUES(40805,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(40805,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(40805,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(40805,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(40805,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(40805,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(40805,36,'cp -145','f6e4 g5e7 d8e7 f1b5 c7c6 c3e4 e7e4 d1e2 e4e2 b5e2 e8g8 e1c1 a7a5 d1d2 f8d8 b2b3 g8f8 c2c4 b8a6 e2f3 a6c5 h1d1 g7g6 a2a3 f8e7 c1b2 c5e6 d1e1 e7f6 e1d1 e6d4 d2d4 f6e7 h2h4 h7h5 b2c3 c8f5 d4f4 f5e6 d1e1 d6d5 c4d5 c6d5');
INSERT INTO evaluations VALUES(46988,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(46988,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(46988,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(46988,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(46988,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(46988,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(46988,36,'cp 75','b1c3 g7g6 d2d4 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 b8a6 c1e3 d8e7 d1c2 e5d4 f3d4 c7c6 f1d1 f8e8 f2f3 d6d5 c4d5 a6b4 c2d2 b4d5 e3g5 d5c3 b2c3 e7a3 g5e3 c8d7 a1b1 b7b5 d4c2 a3e7 c2b4 a7a6 a2a4');
INSERT INTO evaluations VALUES(46985,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46985,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(46985,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(46985,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(46985,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(46985,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(46985,36,'cp 66','g7g6 d2d4 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 f6e8 a1b1 b8d7 a2a3 f7f5 b2b4 b7b6 c1d2 g8h8 c3b5 e8f6 d1c2 f5e4 f3e2 f6g8 b4a5 b6a5 e2g4 g7h6 g4d7 d8d7 d2h6 g8h6 c4c5 h6f5 c2e4 a8e8 e4c2');
INSERT INTO evaluations VALUES(46655,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(46655,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(46655,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(46655,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(46655,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(46655,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(46655,36,'cp 80','d2d4 f8g7 e2e4 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 h2h3 g4f3 e2f3 f6d7 a1b1 d7c5 b2b3 b8d7 a2a3 f7f5 b3b4 a5b4 a3b4 c5e4 f3e4 f5e4 d1c2 h7h6 c3e4 d7f6 e4g3 g8h7 c1e3 d8e7 b1a1 f6d7 a1a8 f8a8 g3e4 d7f6');
INSERT INTO evaluations VALUES(53341,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(53341,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(53341,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(53341,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(53341,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(53341,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(53341,36,'cp 50','d2d4 c5d4 d1d4 b8c6 d4f4 e7e6 f1d3 c6b4 b1c3 c8d7 e1g1 b4d3 c2d3 g8e7 h2h4 e7f5 c3e2 f8e7 h4h5 e8g8 g2g4 f5h6 f4g3 f7f6 c1h6 g7h6 a1c1 d8b6 e5f6 e7f6 d3d4 f8f7 f3e5 f6e5 g3e5 a8f8 c1c7 b6b2 f1c1');
INSERT INTO evaluations VALUES(54057,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(54057,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54057,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(54057,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(54057,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(54057,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(54057,36,'cp -4','d5d4 c3a4 c8d7 b2b4 d7a4 d1a4 b8d7 g1f3 a7a5 f3d4 a5b4 a4b4 f8c5 b4b2 g8f6 e2e3 c5a3 b2b5 a3c1 a1c1 e8g8 f1e2 d8a5 b5a5 a8a5 a2a3 a5a3 c4c5 g7g6 e1g1 f6e4 c5c6 b7c6 d4c6 d7f6 e2f3 f8a8 h2h3 a3a2 c6b4 a2a4 b4c6 g8g7 c6e5 a4a5 e5g4 a8a7 c1c2 h7h6 g1h2');
INSERT INTO evaluations VALUES(4164,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4164,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4164,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(4164,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(4164,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(4164,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4164,36,'cp 8','d2d3 a7a6 a2a4 d7d6 c1e3 c5e3 f2e3 c8e6 b2b3 c6a5 c4e6 f7e6 d3d4 a5c6 f1f2 d8e7 d1e2 e5d4 e3d4 f6h5 e2c4 a8e8 c3e2 h5f4 a1e1 d6d5 e4d5 e6d5 c4c3 f4e6 h3h4 e7f6 e1f1 e6f4 e2f4 f6f4 c3c5 f4d6 c5d6 c7d6 c2c3');
INSERT INTO evaluations VALUES(6091,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(6091,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(6091,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(6091,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(6091,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(6091,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(6091,36,'cp 58','d1d8 e8d8 f2f3 f7f6 a2a4 a7a5 b1c3 g8h6 f1c4 d8e7 c1f4 c5d6 c3e2 h6f7 f4d6 c7d6 c4f7 e7f7 e1c1 h8d8 d1d2 c8e6 e2f4 f7e7 h2h4 e6f7 h4h5 a8b8 h1h4 d8g8 f4e2 g8d8 d2d1 g7g5 h5g6 h7g6 h4h7 d8h8');
INSERT INTO evaluations VALUES(2243,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(2243,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(2243,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(2243,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(2243,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(2243,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(2243,36,'cp 52','e8d8 f1c4 f7f6 c1f4 c8d7 b1c3 d8c8 f4g3 g8h6 e1c1 b7b5 c4b3 h8e8 d1d3 d7e6 b3e6 e8e6 h1d1 e6e7 a2a4 b5b4 c3b1 f6f5 d3d8 c8b7 d8a8 b7a8 b1d2 c5b6 f2f3 f5e4 d2e4 h6f5 g3f4 a7a5 f4g5 e7e6 d1d8 a8b7 g5d2 f5e3');
INSERT INTO evaluations VALUES(2345,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(2345,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(2345,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(2345,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(2345,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(2345,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(2345,36,'cp 74','b1c3 f7f6 f1e2 g8h6 c1d2 c8d7 h1f1 b7b5 a2a4 b5b4 c3a2 a7a5 a2c1 h8e8 c1d3 c5d4 f2f3 h6f7 e1c1 f7d6 c2c3 b4c3 b2c3 d4b6 d2f4 d6f5 c1c2 f5e3 f4e3 b6e3 d3b2 d8e7 b2c4 e3b6 f3f4 a8d8 g2g3 d7e6');
INSERT INTO evaluations VALUES(18449,30,'cp 472',NULL);
INSERT INTO evaluations VALUES(18449,31,'cp 463',NULL);
INSERT INTO evaluations VALUES(18449,32,'cp 498',NULL);
INSERT INTO evaluations VALUES(18449,33,'cp 500',NULL);
INSERT INTO evaluations VALUES(18449,34,'cp 498',NULL);
INSERT INTO evaluations VALUES(18449,35,'cp 519',NULL);
INSERT INTO evaluations VALUES(18449,36,'cp 516','f3c6 c8d7 c6a6 e7e6 g1f3 f8b4 c1d2 b4d2 b1d2 g8f6 f3e5 e8g8 e5d7 d8d7 f1e2 a8c8 e1g1 f8d8 a6a5 d7e8 d2f3 f6d5 f3e5 c8c7 e2f3 d8c8 e3e4 d5b6 f3d1 h7h6 e5f3 e8d7 h2h3 c8a8 d1c2 a8c8 a5d2 c4c3 b2c3 c7c3');
INSERT INTO evaluations VALUES(18448,30,'cp 522',NULL);
INSERT INTO evaluations VALUES(18448,31,'cp 529',NULL);
INSERT INTO evaluations VALUES(18448,32,'cp 521',NULL);
INSERT INTO evaluations VALUES(18448,33,'cp 520',NULL);
INSERT INTO evaluations VALUES(18448,34,'cp 525',NULL);
INSERT INTO evaluations VALUES(18448,35,'cp 532',NULL);
INSERT INTO evaluations VALUES(18448,36,'cp 498','c8d7 c6a6 e7e6 g1f3 f8b4 c1d2 b4d2 b1d2 g8e7 f3e5 e8g8 e5d7 d8d7 f1e2 f8d8 e2f3 e7d5 e1g1 a8b8 d2e4 d5b4 a6a5 b4c6 a5c3 d7c7 e4c5 c7d6 b2b3 e6e5 f3c6 d6c6 b3c4 b5c4 c3c4 e5d4 e3d4 b8a8 c4d3 a7a5 d3b1 d8d4 c5b3 c6c3 b3d4');
INSERT INTO evaluations VALUES(25831,30,'cp 525',NULL);
INSERT INTO evaluations VALUES(25831,31,'cp 535',NULL);
INSERT INTO evaluations VALUES(25831,32,'cp 534',NULL);
INSERT INTO evaluations VALUES(25831,33,'cp 536',NULL);
INSERT INTO evaluations VALUES(25831,34,'cp 530',NULL);
INSERT INTO evaluations VALUES(25831,35,'cp 512',NULL);
INSERT INTO evaluations VALUES(25831,36,'cp 540','c6a6 e7e6 g1f3 g8f6 f3e5 f6d5 c1d2 d5c7 a6a5 f8d6 f1e2 e8g8 e2f3 a8b8 d2b4 d7e8 b4d6 d8d6 a5a7 c7d5 e1g1 d6d8 b1c3 f7f6 c3d5 e6d5 f3d5 d8d5 a7b8 f6e5 b8e5 d5e5 d4e5 b5b4 e5e6 e8g6 f2f3 c4c3 b2c3 b4b3 a1a7');
INSERT INTO evaluations VALUES(25841,30,'cp 523',NULL);
INSERT INTO evaluations VALUES(25841,31,'cp 523',NULL);
INSERT INTO evaluations VALUES(25841,32,'cp 512',NULL);
INSERT INTO evaluations VALUES(25841,33,'cp 505',NULL);
INSERT INTO evaluations VALUES(25841,34,'cp 522',NULL);
INSERT INTO evaluations VALUES(25841,35,'cp 523',NULL);
INSERT INTO evaluations VALUES(25841,36,'cp 528','e7e6 g1f3 g8f6 f3e5 f6d5 c1d2 d5c7 a6a5 f8d6 f1e2 e8g8 e2f3 a8b8 d2b4 d7e8 b4d6 d8d6 a5a7 c7d5 e1g1 d6b6 a7b6 d5b6 b1d2 b8c8 a1a6 b6a4 b2b3 a4b2 b3c4 b5c4 f1c1 c4c3 d2e4 f7f6 f3g4 f6e5 g4e6 e8f7 e6c8 e5d4 e3d4 f8c8 c1c3 b2c4 e4g5 g7g6 g5f7 g8f7');
INSERT INTO evaluations VALUES(22455,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(22455,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(22455,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(22455,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(22455,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(22455,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(22455,36,'cp 25','f8g7 d2d3 h7h6 c1e3 c6d4 e3d4 e5d4 c3e2 g8e7 e2d4 c7c5 d4b5 d7d5 e4d5 a7a6 b5c3 b7b5 c4b3 e8g8 e1g1 c8b7 c3e4 b7d5 e4c5 g7b2 a1b1 b2c3 c5e4 c3g7 c2c4 d5c6 b1c1 a8c8 c4b5 a6b5 d1d2 e7f5 f1e1 f5d4 f3d4 d8d4 h2h3 g8h8 e4g3 h8g8 d2e3 d4e3 e1e3');
INSERT INTO evaluations VALUES(6177,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(6177,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(6177,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(6177,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(6177,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(6177,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(6177,36,'cp 27','d2d3 h7h6 c1e3 c6d4 e3d4 e5d4 c3e2 g8e7 e2d4 c7c5 d4b5 d7d5 e4d5 a7a6 b5c3 b7b5 c4b3 c8b7 e1g1 e8g8 c3e4 b7d5 e4c5 g7b2 a1b1 b2c3 d3d4 a8c8 d1d3 c3d4 c5a6 d4g7 f1d1 d8a5 b3d5 e7d5 b1b5 a5a6 b5d5 a6d3');
INSERT INTO evaluations VALUES(6178,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(6178,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(6178,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6178,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(6178,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(6178,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(6178,36,'cp 30','c6a5 c1g5 f7f6 g5e3 g8e7 d1e2 a5c4 d3c4 d7d6 e1c1 b7b6 f3d2 d8d7 f2f3 f6f5 h2h4 f5f4 e3f2 h7h6 e2d3 e7c6 h1g1 d7f7 b2b3 c6b4 d3e2 b4c6 g2g4 h6h5 a2a4 h5g4 f3g4 c8e6 c3d5 e8c8 c1b2 c6d4');
INSERT INTO evaluations VALUES(47673,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(47673,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(47673,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(47673,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47673,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(47673,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(47673,36,'cp 33','b8d7 b2b3 f8b4 c1b2 b7b6 a2a3 b4c3 b2c3 c8b7 f1e2 d5c4 b3c4 c7c5 e1g1 e8g8 a1c1 d8c7 d1b3 f6e4 c3b2 d7f6 h2h3 f8d8 a3a4 a8c8 b3a3 c7d6 e2d3 c5d4 e3d4 d6f4 f1e1 e4g5 f3g5 h6g5 d3f1 f6e4 a3e3 e4d6 c4c5 f4e3');
INSERT INTO evaluations VALUES(50485,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(50485,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(50485,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(50485,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(50485,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(50485,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(50485,36,'cp -27','d7d5 e4d5 e6d5 c4e2 g8f6 d2d4 f8d6 c1g5 b8d7 e1g1 e8g8 b1d2 c5c4 a2a4 d8c7 g5h4 a8b8 h4g3 f8e8 f1e1 d7f8 g3d6 c7d6 f3e5 f8g6 e2f1 d6c7 e5g6 h7g6 e1e8 f6e8 d1f3 e8f6 f3g3 c7g3 h2g3 c8d7 d2f3 f6e4 f3e5 d7e6 a4a5');
INSERT INTO evaluations VALUES(51934,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(51934,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(51934,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(51934,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(51934,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(51934,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(51934,36,'cp -32','d7d5 c4d5 e6d5 e4d5 g8f6 g1f3 f6d5 d3d4 b8c6 f1c4 d8e7 c4e2 c5d4 e1g1 d5c7 b1a3 c7e6 a3c2 e7f6 b2b4 f8e7 b4b5 c6e5 f3e5 f6e5 f2f4 e5c7 f4f5 e6c5 d1d4 e8g8 c1e3 f8d8 d4c3 a7a6 c2d4 c8d7 a1d1 d7b5 d4b5 a6b5 e2b5 d8d1');
INSERT INTO evaluations VALUES(6750,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(6750,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(6750,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6750,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6750,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(6750,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(6750,36,'cp 25','c6d4 f3d4 e5d4 e1g1 c7c6 b5c4 d6c7 d1d4 d8h4 g2g3 h4f6 d4c3 d7d6 a2a4 a7a5 c1e3 f6c3 b1c3 c8e6 c4e2 e8c8 f1e1 g8e7 e3d4 f7f6 a1d1 e6h3 e2f1 h3g4 d1d3 h8e8 f1e2 g4e6 f2f4 e6f7 e2f3 h7h5 h2h4 d6d5 e4d5 e7d5 e1e8 d8e8');
INSERT INTO evaluations VALUES(9147,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(9147,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(9147,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(9147,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(9147,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(9147,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(9147,36,'cp 111','c8e6 g5e6 f7e6 e1g1 d8d7 f3h5 g7g6 h5d1 e8c8 b1d2 d5b6 c4b5 h8f8 d1e2 f8f5 d2f3 d8f8 c1h6 f5f3 g2f3 f8f5 c2c4 a7a6 b5c6 d7c6 a1d1 b6a4 e2e4 a4b2 d1b1 c6e4 d3e4 f5h5 h6e3 b2a4 b1b3');
INSERT INTO evaluations VALUES(24084,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(24084,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(24084,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(24084,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(24084,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(24084,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(24084,36,'cp 121','g5e6 f7e6 e1g1 h8f8 f3h5 g7g6 h5h7 d8f6 h7h3 c5f2 g1h1 f6h4 h3h4 f2h4 f1f8 e8f8 b1d2 h4g5 d2f3 g5c1 a1c1 f8e7 c1e1 e7f6 c2c3 a8d8 g2g3 d5b6 h2h4 a7a5 h1g2 b6c4 d3c4 a5a4 e1f1 f6e7 f3g5 e5e4 g5e4 c6e5 c4c5 a4a3 b2b3 d8d3 f1e1 e5f3 e1e2 e6e5 b3b4');
INSERT INTO evaluations VALUES(24083,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(24083,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(24083,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(24083,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(24083,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(24083,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(24083,36,'cp 100','f7e6 e1g1 d8d7 f3h5 g7g6 h5d1 e8c8 b1d2 h8f8 a2a4 c6a5 d1e1 a5c4 d2c4 d5b4 e1e5 c5f2 f1f2 f8f2 g1f2 b4c2 a1b1 d7d3 b2b3 c2e3 c1e3 d3b1 e5e6 c8b8 c4e5 b1d1 e6e7 d1d6 e7g7 d6d1 e3f4 d1b3 e5f3 b3b6 f4e3 b6a6 g7h7 a6a4');
INSERT INTO evaluations VALUES(24167,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(24167,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(24167,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(24167,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(24167,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(24167,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(24167,36,'cp 111','e1g1 d8d7 f3h5 g7g6 h5d1 h8f8 b1d2 e8c8 d2e4 c5e7 c1d2 c8b8 a2a4 a7a6 a4a5 d5f6 e4g5 f6d5 c2c3 e7g5 d2g5 c6e7 d1b3 d7d6 b3a3 d8e8 a3d6 c7d6 g2g3 f8f5 h2h4 b8c7 g5h6 c7d7 c4b3 e8d8 b3c2 f5h5 h6d2 d8f8 b2b4');
INSERT INTO evaluations VALUES(14294,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(14294,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(14294,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14294,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14294,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(14294,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14294,36,'cp 36','c8g4 c2c4 d5c4 e2c4 e7e6 e1g1 f8d6 h2h3 g4h5 b1d2 a7a6 c4e2 e8g8 b2b3 h5g6 f3h4 f6d5 h4g6 d5c3 d1e1 c3e2 e1e2 h7g6 c1b2 d8g5 d2c4 g5b5 f1d1 a6a5 b2c3 a5a4 a1b1 a4b3 a2b3 b7b6 e2c2 f8d8');
INSERT INTO evaluations VALUES(21044,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(21044,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(21044,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(21044,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(21044,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(21044,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(21044,36,'cp -40','e7e6 e4d5 e6d5 c1e3 f8b4 b1c3 g8e7 e3f2 e8g8 f1d3 b4a5 a2a3 a5c3 b2c3 c8f5 g1e2 f8e8 d1d2 c6a5 a1b1 b7b6 e1g1 a5c4 d3c4 d5c4 b1e1 h7h6 g2g4 f5d7 e2f4 d7c6 f2g3 d8d7 h2h4 f7f6 e1e2 e7d5 f1e1 d5f4 d2f4 e8e2 e1e2');
INSERT INTO evaluations VALUES(21444,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(21444,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(21444,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(21444,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(21444,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(21444,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(21444,36,'cp -32','e4d5 e6d5 c1e3 f8b4 b1c3 g8e7 e3f2 e8g8 f1d3 b4c3 b2c3 c8f5 d1d2 c6a5 g1e2 a5c4 d3c4 d5c4 e2f4 e7d5 f4d5 d8d5 e1g1 h7h6 a2a4 a7a5 f2g3 c7c6 g3c7 f8e8 c7b6 f7f6 f1e1 f5g6 h2h3 d5f5 e1e8 a8e8 a1e1 e8e1 d2e1 f5c2 b6a5 c2a4 e1e6 g6f7 e6c8 g8h7 c8b7 a4a5 b7f7 a5c3');
INSERT INTO evaluations VALUES(26564,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(26564,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(26564,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(26564,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(26564,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26564,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(26564,36,'cp 21','e7e6 e1g1 g8f6 d3b5 f8d6 c2c4 d5c4 b1d2 e8g8 b5c6 b7c6 d2c4 f5g4 c4e5 d6e5 d4e5 f6d7 e3e4 g4f3 d1f3 d7e5 f3g3 e5g6 g3c3 a7a5 c3c6 d8d4 c6c2 f8b8 f1d1 d4e5 a1b1 a5a4 c1e3 b8b4 f2f3 h7h6 b2b3 a4b3 a2b3 a8b8 e3c5 b4b5 b3b4 e5g5 g1h1 e6e5 b1c1 g6f4 d1d7 f4e6 d7e7');
INSERT INTO evaluations VALUES(26823,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(26823,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26823,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(26823,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26823,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(26823,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26823,36,'cp 24','d3b5 f8d6 c2c4 d5c4 b1d2 g8f6 b5c6 b7c6 d2c4 e8g8 e1g1 f5g4 f1e1 c6c5 d4c5 d6c5 d1d8 f8d8 c1d2 f6e4 d2a5 f7f6 a5c7 d8c8 c7a5 e6e5 f3d2 e4d2 a5d2 h7h5 e1c1 g4e6 b2b3 c5b6 d2b4 c8c6 c4a5 c6c1 a1c1 b6a5 b4a5 a8c8 c1c8 e6c8');
INSERT INTO evaluations VALUES(38422,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(38422,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(38422,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38422,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(38422,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38422,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(38422,36,'cp 57','g8f6 b1c3 e8g8 e1g1 e7e5 d4e5 d6e5 f3e5 f6e4 c3e4 g7e5 c2c3 d8d1 f1d1 f8e8 c1g5 g8g7 d1e1 b8c6 e2f3 h7h6 g5e3 a7a5 a1d1 a5a4 e4c5 e8e7 a2a3 g6g5 f3c6 b7c6 e3d4 e5d4 c3d4 e7e1 d1e1 c8f5 e1c1 g5g4 c1c4 a8e8 f2f3 e8e2 c5a4');
INSERT INTO evaluations VALUES(41389,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(41389,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(41389,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(41389,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(41389,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(41389,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(41389,36,'cp 46','b1c3 e8g8 e1g1 e7e5 d4e5 d6e5 f3e5 f6e4 c3e4 d8d1 f1d1 g7e5 c2c3 f8e8 c1g5 g8g7 d1e1 b8c6 e2f3 h7h6 g5e3 a7a5 a1d1 e8d8 e4c5 d8d1 e1d1 e5d6 c5e4 d6e5 g2g3 g6g5 f3g2 f7f5 f2f4 g5f4 g3f4 f5e4 f4e5');
INSERT INTO evaluations VALUES(43096,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(43096,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(43096,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(43096,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(43096,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(43096,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(43096,36,'cp 91','b2b4 e8g8 e2e3 d7d6 c1b2 e7e6 f1e2 a7a5 b4b5 b7b6 a1c1 b8d7 e1g1 c8b7 b2a3 a8b8 c3a4 f6e4 c4c5 b6c5 d4c5 d6c5 a4c5 d7c5 a3c5 d8d1 f1d1 e4c3 c1c3 g7c3 c5f8 g8f8 d1d7 c7c5 d7c7 b7d5 c7c5');
INSERT INTO evaluations VALUES(8435,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(8435,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(8435,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(8435,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(8435,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(8435,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(8435,36,'cp 32','a2a3 e8g8 c1e3 c8e6 c3d5 a7a5 e1g1 f8e8 c4a2 f6d7 c2c3 e7f8 d1b3 d8c8 b3c2 c6d8 c3c4 c7c6 d5c3 c6c5 a2b3 d8c6 b3a4 g7g6 c3d5 f8g7 a4b5 g8h7 a1b1 c8d8 b2b4 a5b4 a3b4 c6d4 f3d4');
INSERT INTO evaluations VALUES(8448,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(8448,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(8448,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(8448,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8448,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(8448,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(8448,36,'cp 35','e8g8 e1g1 a7a5 c4a2 f8e8 c1e3 c8e6 c3d5 f6d7 a1c1 e7f8 c2c3 c6e7 c3c4 c7c5 f3d2 e7c6 a2b3 f8e7 g1h2 c6d4 g2g3 d4b3 d1b3 d7f6 b3b7 a8b8 b7c7 b8b2 c1b1 b2a2 b1b7 f6d5 c7d8 e7d8');
INSERT INTO evaluations VALUES(3309,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(3309,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(3309,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3309,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(3309,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(3309,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3309,36,'cp 28','c1e3 c8e6 c3d5 a7a5 e1g1 f8e8 c4a2 f6d7 c2c3 e7f8 d1b3 d8c8 b3c2 c6b8 c3c4 c7c6 d5c3 c6c5 a2b3 b8c6 b3a4 c8d8 c3d5 f8e7 a4b5 e8f8 a1e1 g8h7 c2d1 f8g8 g1h1 g7g6 f3d2 e7g5 b5a4 a8b8 f2f4 e5f4 d5f4');
INSERT INTO evaluations VALUES(3310,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(3310,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(3310,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3310,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(3310,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(3310,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3310,36,'cp 38','c8e6 c3d5 a7a5 e1g1 a5a4 c4a2 e6d7 c2c4 c6a5 a1c1 c7c5 d5c3 d8b6 c1b1 b6d8 f3h2 f6h7 f2f4 e5f4 e3f4 h7g5 h2g4 a5c6 c3d5 d7g4 d1g4 e7f6 f4e3 f6d4 e3d4 c6d4 h3h4 g5e6 f1f2 d4c6 b1f1 c6e5 g4g3 g8h8 a2b1 e6d4');
INSERT INTO evaluations VALUES(21144,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(21144,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(21144,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21144,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(21144,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(21144,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(21144,36,'cp 20','c2c3 e7e5 f1e2 g8e7 e1g1 e7g6 c3d4 e5d4 f3e1 a7a5 f2f4 a5a4 b1d2 f8e7 b2b3 a4b3 a2a4 b7b6 d2b3 e8g8 c1d2 d8d7 g2g3 h7h6 e2f3 c6b4 a4a5 b6a5 a1a5 a8a5 b3a5 d7b5 d2b4 b5b4 a5c6 b4b6 e4e5 c8b7 c6e7 g6e7 f3b7 b6b7');
INSERT INTO evaluations VALUES(21143,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(21143,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21143,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(21143,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(21143,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(21143,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(21143,36,'cp 12','e7e5 b1d2 g8e7 h2h4 a7a5 d2c4 f7f6 f1e2 c8e6 e1g1 d8d7 c3d4 c6d4 f3d4 d7d4 c1e3 d4d7 d3d4 e5d4 e3d4 e8c8 d4e3 d7d1 a1d1 d8d1 f1d1 e7c6 f2f4 f8e7 g1f2 h8d8 d1d8 c8d8 h4h5 g7g6 h5g6 h7g6 a2a3 g6g5');
INSERT INTO evaluations VALUES(21806,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21806,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(21806,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21806,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(21806,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(21806,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21806,36,'cp 21','b1d2 g8e7 h2h4 a7a5 d2c4 f7f6 f1e2 c8e6 e1g1 d8d7 c3d4 e5d4 d1a4 e7c8 c1f4 a8a6 a1c1 f8e7 a4b5 c8b6 c4d2 e6a2 c1c6 d7c6 b5c6 b7c6 f3d4 e7d6 f4d6 c7d6 f1a1 a2f7 d4f5 d6d5 d3d4 b6c4 f5g7 e8d8 e2c4 d5c4');
INSERT INTO evaluations VALUES(33628,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(33628,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(33628,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(33628,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(33628,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(33628,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(33628,36,'cp 63','b8c6 d2d3 e4e6 c7g3 d7d5 g3e3 f8d6 e3e6 f7e6 a2a3 e8g8 h2h3 c8d7 g2g3 c6e7 b1c3 e6e5 f1g2 a8c8 d3d4 e5e4 c1d2 b7b5 e1g1 c8c4 f1c1 f8c8 f2f3 e4f3 g2f3 a7a5 d2g5 g8h8 g3g4 a5a4 f3g2 b5b4 a3b4 c4b4 c3d1 f6e4 g2e4 d5e4');
INSERT INTO evaluations VALUES(867,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(867,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(867,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(867,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(867,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(867,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(867,36,'cp 43','d2d3 e4e6 c7g3 d7d5 g3e3 f8d6 g2g3 e8g8 e3e6 f7e6 h2h3 c8d7 c1d2 c6b4 d2b4 d6b4 c2c3 b4d6 b1d2 e6e5 f1g2 d7f5 e2c1 a7a5 d2f1 d5d4 f1d2 d4c3 b2c3 e5e4 d3e4 f6e4 d2e4 f5e4 g2e4 a8e8 e1g1 e8e4 a1b1 f8f3 b1b7 f3c3 b7a7 e4a4');
INSERT INTO evaluations VALUES(866,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(866,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(866,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(866,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(866,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(866,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(866,36,'cp 43','e4e6 c7g3 d7d5 g3e3 f8d6 e3e6 f7e6 g2g3 e8g8 h2h3 c6b4 e1d1 e6e5 a2a3 b4c6 f1g2 c8e6 f2f4 d5d4 b1d2 e6d5 g2d5 f6d5 d2c4 d6b8 f4e5 c6e5 e2d4 e5c4 d3c4 d5b6 b2b3 b8g3 c1e3 g3f4 e3f4 f8f4 d4e6 f4f6 h1e1 a8e8 e6c5 e8f8 c5b7');
INSERT INTO evaluations VALUES(42109,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(42109,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(42109,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42109,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(42109,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(42109,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(42109,36,'cp 57','g3g4 d8e7 b1c3 b8c6 f1d3 g7g5 f2f4 h7h5 f4g5 f6g4 d1g4 h5g4 h1h8 e8d7 h8g8 b7b5 e1c1 c8b7 g8g7 a8h8 g5g6 c6d8 g6f7 h8f8 g7g4 f8f7 g1f3 d7c8 f3e5 f7g7 e5g6 e7d6 c3b5 d6h2 b5c3 c8b8 g4h4 h2g2 h4h8 g2g5 c1b1 b8c8 g6f4 g7g8 h8h7 c8b8 a2a3 g5g3 h7h3 g3g7 d1h1 b7c8 c3e2 e6e5 h3h7 g7g5 d4e5 g5e5');
INSERT INTO evaluations VALUES(47838,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(47838,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47838,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(47838,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(47838,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(47838,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(47838,36,'cp 52','b1c3 c7c6 e2e4 c8e6 d2d3 f8e7 g1f3 h7h6 e1g1 e8g8 b2b3 c6c5 f3e1 a7a6 e1c2 b8c6 f2f4 e6d7 f4f5 c6d4 c2e3 d7c6 h2h4 b7b5 g1h2 a6a5 c3b5 d4b5 c4b5 c6b5 a2a4 b5a6 f1f2 a8b8 e3g4 b8b6 g4f6 e7f6');
INSERT INTO evaluations VALUES(49345,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(49345,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(49345,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(49345,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(49345,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(49345,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(49345,36,'cp -27','e6d5 c4d5 f8b4 b1c3 e8g8 c1g5 h7h6 g5h4 d7d6 e2e3 f8e8 a1c1 b4c3 c1c3 c7c5 d5c6 b8c6 h4f6 d8f6 g1f3 c8e6 a2a3 c6e5 f3d4 e6d5 f2f3 a8c8 c3c8 e8c8 f1e2 f6h4 g2g3 h4h3 e2f1 h3d7 f1e2 a7a6 e1g1 b7b5 d1d2 d5c4 f1d1 c4e2 d2e2 d7h3');
INSERT INTO evaluations VALUES(49486,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(49486,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(49486,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(49486,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49486,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(49486,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(49486,36,'cp -23','c4d5 f8b4 b1c3 e8g8 c1g5 h7h6 g5h4 d7d6 e2e3 f8e8 a1c1 b4c3 b2c3 c8f5 f1b5 b8d7 b5d7 f5d7 g1e2 g7g5 h4g3 f6e4 e1g1 f7f5 f1e1 e4c5 e2d4 d8f6 f2f3 e8e7 d1d2 a8e8 g1h1 h6h5 h2h3 a7a6 c3c4 e7g7 d4b3 c5b3 a2b3 h5h4 g3h2 g5g4 h2f4 g4f3');
INSERT INTO evaluations VALUES(49479,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(49479,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(49479,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(49479,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(49479,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(49479,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(49479,36,'cp -17','f8b4 b1c3 e8g8 c1g5 h7h6 g5h4 d7d6 e2e3 f8e8 a1c1 b4c3 b2c3 c8f5 f1b5 b8d7 b5d7 f5d7 g1e2 g7g5 h4g3 f6e4 e1g1 f7f5 f1e1 e4c5 e2d4 d8f6 f2f3 e8e7 c3c4 b7b6 d1d2 a8f8 g1h1 h6h5 h2h3 f8e8 c1d1 e7g7 g3h2 f6g6 d1c1 h5h4 e1g1 g5g4 h2f4');
INSERT INTO evaluations VALUES(11632,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11632,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(11632,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11632,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(11632,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(11632,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11632,36,'cp 26','a2a4 a8b8 a4b5 a6b5 e1g1 f8e7 b1c3 e8g8 c3d5 f6d5 b3d5 c6b4 d5b3 c7c5 h2h3 c8e6 b3e6 f7e6 c1d2 b4c6 d1e2 d8d7 c2c3 h7h6 f3h2 g8h7 e2h5 c5c4 d3c4 b5c4 h2g4 d7e8 h5e8 f8e8 f1b1 b8a8 a1a8 e8a8 g4e3 c6a5 b1a1 e7d8 e3f1');
INSERT INTO evaluations VALUES(19243,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(19243,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(19243,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(19243,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(19243,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(19243,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(19243,36,'cp -22','e7e6 c4e2 b7b6 e1g1 d7d5 d2d3 g8e7 c2c3 c8a6 b1d2 f8h6 f3h4 e8g8 e4e5 d8d7 d2f3 d5d4 c3c4 a6b7 c1d2 h6g7 f3g5 e7f5 h4f5 e6f5 e2f3 h7h6 g5h3 g8h7 f3d5 a8b8 d1e2 c6e7 d5b7 b8b7 b2b3 b6b5 a1e1 b5c4');
INSERT INTO evaluations VALUES(18626,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(18626,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(18626,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(18626,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(18626,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(18626,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(18626,36,'cp 0','c4e2 b7b6 c2c3 d7d5 d2d3 g8e7 e1g1 c8a6 f1e1 f8g7 b1d2 e8g8 e4e5 f7f6 d3d4 a6e2 e1e2 f6e5 f4e5 c5d4 c3d4 e7f5 d2f1 f5h4 f3e1 h4f5');
INSERT INTO evaluations VALUES(26958,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(26958,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(26958,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(26958,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(26958,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(26958,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(26958,36,'cp -25','c6d4 c1e3 g8f6 e3d4 e5d4 c3b5 c7c5 e1g1 a7a6 b5a3 b7b5 c4b3 f8e7 h2h3 g4c8 f1e1 e8g8 a3b1 c8b7 c2c3 d4c3 b1c3 a8c8 a1c1 d8b6 a2a3 f8d8 b3a2 e7f8 d1d2 g7g6 b2b4 f8g7 d2e3 d8e8 a3a4 f6d7 b4c5 b6c5 e3c5 c8c5 a4b5');
INSERT INTO evaluations VALUES(27092,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(27092,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27092,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27092,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(27092,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(27092,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(27092,36,'cp 0','h2h3 g4f3 g2f3 g7g6 f3f4 f8h6 f4e5 h6c1 d1c1 d6e5 c3b5 g8f6 b5d4 d8d4 c2c3 d4d6 c1g5 h7h6 g5e3 c7c5 e1c1 e8f8 c1b1 a8d8 h1f1 f6h5 d3d4 c5d4 c3d4 e5d4 e3b3 d6f6 b3b7 f8g7 f2f4 h5f4 c4f7 f6f7 b7f7 g7f7 f1f4 f7e6 b1c2 d8d7 f4g4');
INSERT INTO evaluations VALUES(27530,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(27530,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(27530,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(27530,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(27530,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(27530,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(27530,36,'cp -17','c5d4 c3d4 c6b4 e1g1 b4c2 d1c2 g8e7 b1c3 e7f5 f1d1 a8c8 h2h4 h7h6 h4h5 f8e7 c1f4 e8g8 c2d2 c8c4 a1c1 f8c8 g2g4 f5h4 f3h4 e7h4 g1g2 b6d8 b2b3 c4c6 c3e2 h4e7 c1c6 c8c6 d1c1 c6c1 e2c1 d7e8 f2f3 e8b5 c1e2 d8b6 f4e3 b5e2');
INSERT INTO evaluations VALUES(27551,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(27551,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(27551,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(27551,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(27551,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(27551,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(27551,36,'cp -12','c3d4 c6b4 e1g1 b4c2 d1c2 g8e7 b1c3 a8c8 f1d1 e7f5 h2h4 h7h5 c1g5 f8b4 a2a3 b4c3 b2c3 e8g8 d1b1 b6a6 c2d2 c8c7 a3a4 c7c4 a4a5 f8c8 a1a3 d7a4 f3e1 a6a5 e1d3 c4c3 d3c5 c3c5 d4c5 a5d2 g5d2 a4c6 a3a7');
INSERT INTO evaluations VALUES(32513,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(32513,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(32513,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(32513,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(32513,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(32513,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(32513,36,'cp 70','f1d3 d7c6 g1f3 b8d7 c2c3 f8e7 d1e2 g8f6 e4f6 e7f6 c1e3 c6d5 e1g1 e8g8 d3c2 g7g6 a1d1 d5f3 e2f3 c7c6 f3e2 f6g7 g2g3 d7f6 a2a4 f8e8 f1e1 a7a5 e3c1 h7h5 c1g5 d8b6 d1b1 e8d8 g5f4 d8d5 f4e5 c6c5 d4c5');
INSERT INTO evaluations VALUES(34273,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(34273,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(34273,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(34273,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(34273,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(34273,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(34273,36,'cp 142','d2d4 d7d6 f1c4 h7h6 c1e3 g8f6 h2h3 e7c7 c4b3 b7b5 a2a3 f8e7 g2g4 f6h7 h1g1 b8d7 h3h4 a7a5 d4d5 a5a4 b3a2 c6c5 c3b5 c7a5 b5c3 a8b8 a2c4 h7f6 d1d2 b8b2 g4g5 f6h5 c3d1 a5d2 e1d2 b2b8 g5h6 g7h6 d1c3');
INSERT INTO evaluations VALUES(34272,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(34272,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(34272,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(34272,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(34272,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(34272,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(34272,36,'cp 130','d7d6 f1c4 h7h6 c1e3 e7c7 h2h3 g8f6 c4b3 b7b5 a2a3 b8d7 f3h4 a7a6 e1g1 c8b7 d4e5 d6e5 d1e2 c6c5 b3d5 b7d5 c3d5 f6d5 e4d5 g7g6 c2c4 f8d6 a1e1 e8f8 e3d2 b5b4 a3b4 c5b4 d2e3 f8g7 f2f4 e5f4 e3d4 g7h7 d4h8 h7h8 h4f3 a6a5');
INSERT INTO evaluations VALUES(34405,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(34405,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(34405,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(34405,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(34405,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(34405,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(34405,36,'cp 144','c1e3 g8f6 f1c4 b7b5 c4b3 h7h6 a2a3 e7c7 f3h4 a7a5 e1g1 f8e7 d4d5 c7b7 d1d3 a5a4 b3a2 a8a5 f2f4 f6g4 h4f5 c8f5 e4f5 g4e3 d3e3 b8d7 d5c6 b7c6 c3d5 c6c5 e3c5 d7c5 f4e5 d6e5 f5f6 e7f6 d5f6 g7f6 f1f6');
INSERT INTO evaluations VALUES(52792,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(52792,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(52792,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(52792,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(52792,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(52792,35,'cp -110',NULL);
INSERT INTO evaluations VALUES(52792,36,'cp -126','d5e4 d1h5 g7g6 h5h4 a7a5 b1c3 f8g7 h4e4 g8f6 e4e2 b7b5 a2a3 c8a6 d2d3 e8g8 g1f3 a5a4 b3a2 c6c5 e1g1 e7e6 f1e1 d8b6 c1f4 b8c6 h2h3 f8d8 a1d1 h7h6 c3e4 f6e4 e2e4 a6b7 e4e2 c6d4 f3d4 d8d4 f4e5 b6c6');
INSERT INTO evaluations VALUES(52921,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(52921,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(52921,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(52921,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(52921,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(52921,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(52921,36,'cp -125','d1h5 g7g6 h5e2 g8f6 b1c3 c8f5 f2f3 e4f3 g1f3 f8g7 d2d4 e8g8 e1g1 b7b5 h2h3 a7a5 a2a3 a5a4 b3a2 d8d6 f1e1 b8d7 e2e7 d6e7 e1e7 a8e8 e7e8 f8e8 f3g5 e8f8 c1f4 h7h6 g5f3 f5c2 a1e1 g6g5 f4h2 c2f5 e1e7 f5g6 h2d6 f8e8 g2g4 e8e7');
INSERT INTO evaluations VALUES(52866,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(52866,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(52866,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(52866,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(52866,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(52866,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(52866,36,'cp -125','g7g6 h5e2 g8f6 b1c3 c8f5 f2f3 e4f3 g1f3 f8g7 e1g1 e8g8 d2d4 b7b5 c1f4 a7a5 a2a3 b8d7 h2h3 h7h5 a1e1 e7e6 a3a4 b5b4 c3d1 c6c5 d1e3 f5e4 f4d6 c5d4 f3d4 d8b6 d6f8 a8f8 e1d1 e4b7 e2b5 b7a6 b5b6 d7b6 d4b5 a6b5 a4b5 a5a4 b3a2 f6e4 e3c4');
INSERT INTO evaluations VALUES(54240,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(54240,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(54240,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(54240,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(54240,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(54240,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(54240,36,'cp -19','e4d5 e6d5 b1c3 b8c6 f1b5 c5c4 g1e2 f8b4 b2b3 c4b3 a2b3 g8e7 e1g1 e8g8 c3a4 a7a6 b5c6 e7c6 c2c3 b4e7 c1a3 e7a3 a1a3 c6e7 a3a1 c8d7 f1e1 a6a5 a4c5 b7b6 c5d3 d7b5 e2f4 b5d3 f4d3 f8e8 d1d2 d8d7 g2g3');
INSERT INTO evaluations VALUES(57222,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(57222,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(57222,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(57222,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(57222,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(57222,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(57222,36,'cp -33','d7d5 d2d4 b8c6 c2c4 g8f6 d4e5 c6e5 c4d5 f6d5 b1c3 d5b4 e2d4 c7c5 d4f3 d8d1 e1d1 e5f3 f1b5 e8e7 g2f3 c8e6 a2a3 a8d8 d1e2 a7a6 a3b4 a6b5 b4c5 e6c4 e2e1 e7e6 c1d2 f8c5 h1g1 g7g6 c3e4 c5e7 e4g5 e6d7 d2c3 f7f6 g5e4 h8f8');
INSERT INTO evaluations VALUES(56477,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56477,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(56477,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(56477,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(56477,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56477,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(56477,36,'cp -21','d2d4 b8c6 c2c4 g8f6 c4d5 f6d5 a2a3 d5b6 b1c3 e5d4 e3d4 c8e6 c1e3 d8d7 g2g3 f8d6 f1g2 e6h3 g2h3 d7h3 d1c2 e8g8 e1c1 c6a5 c1b1 f8e8 e3c1 h3g4 h1e1 h7h6 d4d5 a5c4 b1a2 d6e5 f2f4 e5f6 c3e4 g4g6 e4f6 g6f6 e2c3');
INSERT INTO evaluations VALUES(6133,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(6133,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(6133,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(6133,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(6133,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(6133,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(6133,36,'cp 78','g8f6 f1e2 e7e5 e1g1 e8g8 c1e3 e5d4 f3d4 f8e8 d1c2 a7a6 a1d1 c7c6 h2h3 d8e7 e3g5 d7c5 e2f3 a6a5 b2b3 h7h6 g5h4 g6g5 h4g3 f6e4 c3e4 c5e4 f1e1 f7f5 f3h5 e8d8 h5g6 g7d4 d1d4 e7f6 d4e4 f5e4 g6e4 d8f8 e4h7 g8h8');
INSERT INTO evaluations VALUES(41001,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41001,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41001,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(41001,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41001,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(41001,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41001,36,'cp 4','e1g1 e8g8 b1d2 c7c5 c2c3 b8d7 e3e4 c5c4 d3c2 f6e4 d2e4 d5e4 c2e4 d7f6 e4c2 b7b5 f1e1 c8b7 f3e5 d8d5 f2f3 b5b4 d1e2 a8c8 c1d2 c8c7 a2a4 b4a3 b2a3 d5b5 a3a4 b5b2 e2d1 f6d5 d1b1 b2b1 a1b1 f8c8 a4a5 f7f6');
INSERT INTO evaluations VALUES(41000,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41000,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41000,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41000,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41000,36,'cp 0','e8g8 b1d2 c7c5 d4c5 a7a5 c2c3 a5a4 b2b4 a4b3 d2b3 b8a6 c3c4 a6c5 b3c5 e7c5 c1b2 b7b6 d1e2 c8b7 f1d1 h7h6 f3e5 d8e7 a2a4 f8d8 c4d5 b7d5 h2h3 f6e4 e2c2 e4f6');
INSERT INTO evaluations VALUES(36356,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36356,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(36356,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36356,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36356,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36356,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36356,36,'cp 0','b1d2 c7c5 d4c5 a7a5 c2c3 a5a4 b2b4 a4b3 d2b3 b8a6 c3c4 a6c5 b3c5 e7c5 c4d5 d8d5 d1e2 d5h5 c1b2 b7b6 f1d1 c8b7 a2a4 f6e4 e2c2 e4f6');
INSERT INTO evaluations VALUES(36357,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(36357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36357,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(36357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36357,36,'cp 0','c7c5 b2b3 b7b6 c1b2 c8b7 d4c5 b6c5 c2c4 d5d4 e3d4 b8c6 a2a3 c6d4 f3d4 c5d4 d2f3 b7f3 d1f3 a7a5 a1d1 e7c5 b3b4 a5b4 a3b4 c5b4 b2d4 d8e7 d1a1 a8a1 f1a1 b4c5 d4c3 c5b4 c3b2 b4c5');
INSERT INTO evaluations VALUES(730,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(730,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(730,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(730,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(730,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(730,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(730,36,'cp 95','c5d4 c3d4 f7f6 e5f6 d7f6 e1g1 f8d6 c1f4 d6f4 e2f4 e8g8 f1e1 c8d7 d1d2 a8e8 a1d1 b6b4 d2b4 c6b4 h2h4 b4a2 d1a1 f6e4 a1a2 f8f4 a2a7 d7c6 b2b4 e4d6 h4h5 c6b5 d3b5 d6b5 a7b7 b5d4 f3e5 d4c2 e1b1 c2a3 b1a1 a3c4 e5c4 d5c4 b7c7 f4g4');
INSERT INTO evaluations VALUES(741,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(741,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(741,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(741,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(741,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(741,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(741,36,'cp 75','c3d4 f7f6 e5f6 d7f6 e1g1 f8d6 c1f4 d6f4 e2f4 e8g8 f1e1 c8d7 d1d2 a8e8 a1c1 c6d4 f3d4 e6e5 f4d5 f6d5 d4f3 d5f4 f3e5 b6f6 d3c4 d7e6 d2d4 g8h8 c4e6 e8e6 g1h1 f4g6 e5g6 h7g6 d4f6 f8f6 h1g1 e6e1 c1e1 f6d6 b2b4 a7a5 a2a3 a5b4 a3b4 d6d2 g1f1 h8h7');
INSERT INTO evaluations VALUES(736,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(736,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(736,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(736,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(736,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(736,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(736,36,'cp 73','f7f6 e5f6 d7f6 e1g1 f8d6 c1f4 d6f4 e2f4 e8g8 f1e1 c8d7 d1d2 a8e8 d3c2 b6b4 d2c3 b4c3 b2c3 f6e4 c2e4 f8f4 e4d3 h7h6 a1b1 b7b6 e1e3 f4f6 d3b5 e8c8 h2h4 d7e8 b1e1 c6a5 a2a4 e8b5 a4b5 c8c7 h4h5 a5c4 e3e6 c4d6 e6e7 c7e7 e1e7 d6b5 f3e5 b5c3');
INSERT INTO evaluations VALUES(756,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(756,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(756,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(756,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(756,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(756,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(756,36,'cp 78','e5f6 d7f6 c1f4 f6h5 f4e3 f8d6 e1g1 e8g8 d1c2 h5f6 a2a3 c8d7 a1c1 a8c8 c2b1 c6a5 c1c8 f8c8 e2f4 d7b5 f1c1 c8c1 b1c1 b6c6 d3b5 c6b5 c1c8 b5e8 c8e6 e8e6 f4e6 g8f7 e6c5 a5c4 e3c1 d6c5 d4c5');
INSERT INTO evaluations VALUES(755,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(755,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(755,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(755,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(755,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(755,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(755,36,'cp 70','d7f6 e1g1 f8d6 c1f4 d6f4 e2f4 e8g8 f1e1 c8d7 g2g3 a8e8 d3c2 c6d4 d1d4 b6d4 f3d4 e6e5 f4d5 e5d4 d5e7 g8h8 a1d1 f8f7 e7f5 d7f5 c2f5 e8d8 g1g2 g7g6 f5d3 h8g7 d1c1 d8d6 e1e2 d6c6 c1c4 c6c4 d3c4 f7d7 c4d3 f6d5 a2a3 g7f6 e2e4 d5c7 f2f4');
INSERT INTO evaluations VALUES(865,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(865,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(865,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(865,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(865,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(865,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(865,36,'cp 72','e1g1 f8d6 c1f4 d6f4 e2f4 e8g8 f1e1 c8d7 d1d2 a8e8 a1c1 c6d4 f3d4 e6e5 f4d5 f6d5 d4f3 d5f4 f3e5 b6f6 d3c4 d7e6 d2d4 g8h8 c4f1 e6a2 g2g3 f4g6 e5g6 h7g6 e1e8 f8e8 d4a4 a2f7 a4d7 a7a5 d7b7 e8d8 c1c6 f6f5 b7a7 f7d5 a7d4 f5e4 d4e4 d5e4 c6e6 e4f5');
INSERT INTO evaluations VALUES(864,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(864,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(864,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(864,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(864,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(864,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(864,36,'cp 67','f8d6 c1f4 d6f4 e2f4 e8g8 f1e1 c8d7 d1d2 a8e8 a1d1 c6d4 f3e5 d7a4 b2b3 a4c6 d3h7 g8h7 d2d4 f6e4 d4b6 a7b6 f4h3 f8f5 f2f4 e8a8 a2a4 b6b5 a4b5 c6b5 e5f3 e4c3 d1d2 c3e4 f3g5 h7h8 g5e4 d5e4 e1e4 a8a1 g1f2 f5d5 d2d5 e6d5');
INSERT INTO evaluations VALUES(481,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(481,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(481,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(481,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(481,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(481,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(481,36,'cp 97','c1f4 d6f4 e2f4 e8g8 f1e1 c8d7 d1d2 b6b4 d2e3 f6g4 f3g5 g4f6 g5e6 a8e8 d3f5 b4d6 e3g3 f8f7 e1e3 f6e4 f5e4 d5e4 d4d5 c6b4 e3e4 b4d5 a1d1 d7e6 f4d5 d6g3 h2g3 f7d7 a2a3 g8f7 d5e3 d7d1 e3d1 e6d7 e4f4 f7g6 d1e3 d7c6 f4d4 e8e7 d4d3 c6b5 d3d8 b5c6 g3g4 a7a5 g1h2 a5a4 h2g3 h7h6 d8a8 g6f7 a8a7 e7e4 e3f5');
INSERT INTO evaluations VALUES(6443,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6443,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6443,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(6443,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6443,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6443,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6443,36,'cp 0','g8f6 d2d3 f6g4 c4f7 e8f7 f3g5 f7g8 d1g4 c6d4 g4d1 e5f4 c1f4 h7h6 g5f3 d8f6 c2c3 f6f4 f3d4 c5d4 c3d4 f4e3 d1e2 e3c1 e2d1');
INSERT INTO evaluations VALUES(10150,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10150,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10150,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10150,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10150,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10150,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10150,36,'cp 10','d2d3 e5f4 c1f4 d7d5 e4d5 f6d5 f4g5 c5e7 d1e2 c8e6 g5e7 d8e7 c4d5 e6d5 e2e7 e8e7 h1f1 f7f6 b1c3 d5f3 f1f3 a8d8 e1d2 e7f7 a1e1 b7b6 a2a4 h7h5 c3b5 d8d7 b2b3 h5h4 f3f2 c6e7 b5c3 h4h3 g2g3 c7c6');
INSERT INTO evaluations VALUES(10148,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(10148,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10148,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10148,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(10148,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(10148,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(10148,36,'cp 0','e5f4 c1f4 d7d6 b1c3 e8g8 d1d2 c8e6 c4e6 f7e6 e1c1 f6g4 d1e1 g4e5 f3e5 d6e5 f4g3 d8d6 h1f1 c5b4 f1f8 a8f8 e1e2 c6d4 e2f2 d4b5 f2f8 d6f8 g3e5 f8f1 d2d1 f1g2 d1h5 b5c3 h5e8 b4f8 e8e6 g8h8 e5c3 g2g1 c1d2 g1h2 d2d1 h2h5 d1d2');
INSERT INTO evaluations VALUES(10284,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10284,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10284,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10284,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10284,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10284,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10284,36,'cp 0','c1f4 e8g8 b1c3 d7d6 d1d2 c8e6 c4e6 f7e6 e1c1 f6g4 f4g5 d8e8 d1e1 h7h6 g5h4 e8h5 h4g3 c6d4 h1f1 f8f7 f3d4 f7f1 e1f1 c5d4 c3d1 b7b5 d2b4 d4e5 b4b3 a8e8 g3e5 h5e5 h2h3 g4f6 c1b1 f6d7 d1e3 e8f8 e3f5 f8f7 d3d4');
INSERT INTO evaluations VALUES(10282,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10282,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10282,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10282,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10282,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10282,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10282,36,'cp 0','e8g8 b1c3 d7d6 d3d4 c5b6 f4g5 h7h6 g5f6 d8f6 c3d5 f6g6 d1d3 c8g4 e1c1 a8e8 h2h3 g4d7 h1e1 c6a5 d5f4 g6f6 f4d5');
INSERT INTO evaluations VALUES(41948,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(41948,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(41948,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(41948,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(41948,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(41948,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(41948,36,'cp 73','d1c2 c7c6 f1d3 f7f5 e1g1 e8g8 c3e2 b4d6 b2b3 a7a5 c1b2 b8d7 f3e5 d8f6 f2f4 d6e7 d3e4 f5e4 a2a4 d7e5 f4e5 f6h4 f1f8 e7f8 e2f4 c8d7 a1f1 b7b6 c2c1 h4e7 c4c5 b6b5 b2c3 d7e8 c3e1 g7g5 f4e2 e8h5 c1c2 b5b4 f1f6 f8g7');
INSERT INTO evaluations VALUES(41949,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(41949,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(41949,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(41949,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(41949,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(41949,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(41949,36,'cp 64','c7c6 f1d3 f7f5 e1g1 e8g8 c3e2 b8d7 b2b3 a7a5 c1b2 b4d6 e2f4 d8f6 d3e2 f6h6 f3e5 d6e7 a1d1 g7g5 f4d3 g5g4 f2f3 g4g3 h2h3 d7e5 d4e5 e7g5 f3e4 g5e3 g1h1 d5e4 b2c1 e4d3 d1d3 f5f4 c1e3 f4e3');
INSERT INTO evaluations VALUES(49433,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(49433,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(49433,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49433,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49433,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49433,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49433,36,'cp -2','g1f3 c7c5 e2e3 d7d5 a2a3 a7a6 d4c5 f8c5 b2b4 c5e7 c2c4 e8g8 b1d2 d5c4 f1c4 b7b5 c4e2 c8b7 c1b2 b8d7 e1g1 a8c8 a1c1 c8c1 d1c1 d8a8 c1a1 f8d8 f1c1 d8c8 f3e1 c8c1 a1c1 e7d6 e2f3 h7h6 b2d4 b7d5 c1a1 d5f3 e1f3');
INSERT INTO evaluations VALUES(49426,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49426,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(49426,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49426,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49426,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49426,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49426,36,'cp 0','c7c5 e2e3 d7d5 f1e2 b7b6 e1g1 f8d6 c2c4 c8b7 b1c3 e8g8 d4c5 b6c5 d1b3 b7c6 b3c2 c6b7');
INSERT INTO evaluations VALUES(51036,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(51036,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(51036,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(51036,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(51036,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51036,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(51036,36,'cp 46','b7b5 f1d3 c8b7 g1h3 c7c5 d4d5 g8f6 e1g1 c5c4 d3e2 d8c7 c1g5 f8e7 g5f4 e6e5 f4e3 e8g8 f2f3 e7c5 d1d2 d7d6 c3d1 b8d7 b2b3 a6a5 b3c4 b5b4 a2a3 d7b6 e3c5 c7c5 h3f2 b4a3 d1e3 b7a6 f2d3 c5c7 a1a3');
INSERT INTO evaluations VALUES(51149,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(51149,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51149,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(51149,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(51149,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(51149,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(51149,36,'cp 69','f1d3 c7c5 d4d5 c8b7 g1h3 g8f6 e1g1 c5c4 d3e2 f8d6 c1f4 d6f4 h3f4 e8g8 d1d4 f8e8 f4h5 e6d5 e4d5 f6h5 e2h5 d7d6 d4f4 g7g6 h5f3 e8e5 a2a4 d8b6 f3g4 f7f5 g4f3 b8d7 a4b5 a6b5 a1a8 b7a8 f1a1 a8b7 h2h4 e5e8 h4h5 b5b4 c3a4 b6c7 h5h6 b7a6 c2c3 d7e5');
INSERT INTO evaluations VALUES(55314,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55314,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55314,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55314,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55314,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55314,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(55314,36,'cp 0','b8d7 e2e4 c7c6 a4c4 e7e5 f1e2 f8d6 e1g1 g8f6 d2d3 e8g8 c1e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(5348,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5348,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5348,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5348,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(5348,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5348,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5348,36,'cp 0','d7d6 g1e2 e7e5 c3d5 g8e7 c1g5 d8a5 g5d2 a5d8');
INSERT INTO evaluations VALUES(18528,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(18528,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(18528,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(18528,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(18528,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(18528,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(18528,36,'cp 123','f1d3 c6c5 c2c3 g8e7 e1g1 e7c6 a2a3 f8e7 c1e3 e8g8 b1d2 d8c7 b2b4 c5c4 d3c2 f7f6 e3f4 f6e5 f4g3 e7d6 b4b5 c6d4 c3d4 e5e4 c2e4 d5e4 g3d6 c7d6 d2e4 d6d5 e4c3 d5h5 f1e1 d7b6 f3e5 h5d1 a1d1 c8d7 e5d7 b6d7 e1e6 d7f6 e6e7');
INSERT INTO evaluations VALUES(18527,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(18527,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(18527,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(18527,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(18527,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(18527,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(18527,36,'cp 115','c6c5 c2c3 g8e7 e1g1 e7c6 a2a3 f7f6 e5f6 d7f6 c1f4 f8d6 f4d6 d8d6 d4c5 d6c5 b2b4 c5e7 f1e1 e8g8 b1d2 c8d7 b4b5 c6d8 c3c4 e7d6 f3e5 d8f7 d2f3 f7e5 f3e5 a8c8 a3a4 d7e8 a1c1 e8h5 d1d2 f6g4 e5g4 h5g4 c4d5 c8c1 e1c1');
INSERT INTO evaluations VALUES(19406,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19406,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(19406,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(19406,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(19406,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(19406,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19406,36,'cp 4','g8f6 b1c3 e7e6 d2d4 d7d5 c4d5 e6d5 f1b5 c5d4 f3d4 c8d7 e1g1 f8d6 e3e4 d5e4 c1g5 e8g8 c3e4 d6e7 e4f6 e7f6 d4c6 b7c6 g5f6 d8f6 d1d7 c6b5 d7b5 a8b8 b5a5 f6b2 a5a7 b8a8 a7e3 a8a2 a1a2 b2a2 e3f4');
INSERT INTO evaluations VALUES(12810,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(12810,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(12810,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12810,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12810,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12810,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(12810,36,'cp 9','d2d4 c5d4 e3d4 d7d5 b1c3 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 a1d1 e7e6 f3g3 h7h5 h2h4 d8c7 g3g5 g7g6 g5f6 h8g8 d1d3 f8e7 f6e5 c6c5 e5c7 c8c7 d4d5 e7f6 e1e2 f6c3 b2c3 e8d7 d5e6 d7e6 h1d1 g8c8 d3e3 e6f6 d1d6 f6g7 d6a6 c7b7 e2f3');
INSERT INTO evaluations VALUES(12815,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(12815,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12815,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(12815,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(12815,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12815,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(12815,36,'cp 12','c5d4 e3d4 d7d5 b1c3 c8g4 c1g5 f6e4 g5e3 e7e6 c4d5 e6d5 d1b3 f8b4 f3e5 d8a5 a2a3 b4c3 b2c3 c6e5 d4e5 a5c3 b3c3 e4c3 f2f3 g4d7 a1c1 c3a4 c1c7 d7c6 e3d4 e8g8 e1d2 f8e8 f1d3 a4b6 h1e1 b6c4 d2c3 e8c8 c7c8 a8c8 d3c4 d5c4 d4a7 c8a8 a7c5 a8c8 c3c4');
INSERT INTO evaluations VALUES(23424,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(23424,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(23424,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(23424,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(23424,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(23424,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(23424,36,'cp -57','e7e5 d2d3 f8c5 g1f3 d7d6 a2a3 a7a5 c1e3 g8e7 d3d4 e5d4 f3d4 e8g8 b1c3 c8e6 f1e2 c6d4 e3d4 e7c6 d4c5 d6c5 e1g1 d8h4 f2f4 a8d8 c3d5 c6d4 d1e1 h4e1 a1e1 e6d5 c4d5 f7f5 e4f5 d4e2 e1e2 f8f5 e2e7 f5d5 e7c7 d8d7 c7d7 d5d7 g2g3 b7b6 f1e1');
INSERT INTO evaluations VALUES(30289,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(30289,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(30289,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(30289,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(30289,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(30289,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(30289,36,'cp 29','b5c4 f8e7 d2d4 e5d4 f3d4 c7c6 e1g1 b7b5 c4d3 e8g8 b2b3 b5b4 c3e2 c6c5 d4f3 b8c6 h2h3 a7a5 a2a4 b4a3 c1a3 f8e8 a3b2 e7f8 e2g3 c6b4 f1e1 b4d3 c2d3 d7c6 b2c3 h7h6 d1d2 a5a4 b3a4');
INSERT INTO evaluations VALUES(30290,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(30290,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(30290,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(30290,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(30290,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(30290,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(30290,36,'cp 29','f8e7 d2d4 e5d4 f3d4 c7c6 a2a4 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 d1f3 e7f6 c1e3 d8e7 e4d3 b8a6 c2c3 a6c7 d3c4 e7e4 f3e4 e8e4 f1d1 d7g4 d1d3 a8e8 h2h3 g4h5 d3d2 e8d8 c4b3 e4e8 g2g4 h5g6 a1d1 a7a6 g4g5 e8e3');
INSERT INTO evaluations VALUES(35077,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(35077,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(35077,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(35077,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(35077,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(35077,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(35077,36,'cp 167','e2e4 f6d8 b1c3 d7d6 f1d3 b8d7 d1e2 a7a6 c1e3 g8f6 e1g1 c7c5 d4d5 f8e7 d5e6 f7e6 e4e5 d7e5 f3e5 d6e5 e3g5 e8g8 a1d1 c8d7 e2e5 d8b8 e5e2 b8c7 d3c4 g8h8 c4e6 d7e6 e2e6');
INSERT INTO evaluations VALUES(37574,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(37574,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(37574,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(37574,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(37574,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(37574,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(37574,36,'cp 73','d7d6 g1f3 g8f6 c1e3 e8g8 f1e2 e7e5 d4e5 d6e5 d1c2 c8b7 e1g1 d8e8 a1d1 c7c5 c3d5 b8c6 e3g5 e8e6 d5c7 e6c8 c7a8 c8a8 c2b1 c6b4 d1d5 b4c6 e2d3 a8b8 a2a3 b8c7 b1c2 f6d5 c4d5 c6d4 f3d4 e5d4 f2f4 c7d6 b2b3 f7f5 e4f5 d6d5 f5g6 h7g6');
INSERT INTO evaluations VALUES(37548,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(37548,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(37548,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(37548,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(37548,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(37548,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(37548,36,'cp 82','g1e2 g8f6 g2g3 b8d7 f1g2 e7e5 e1g1 e8g8 c1g5 h7h6 g5e3 f6g4 e3c1 g4f6 b2b3 f8e8 d1c2 e5d4 e2d4 c8b7 c1b2 c7c6 a1d1 d8c7 c2d2 a8d8 f1e1 a7a6 d4c2 b6b5 d2d6 c7d6 d1d6 b5c4 e4e5 f6g4 f2f4 d7c5 e1d1 d8d6 d1d6 c4b3');
INSERT INTO evaluations VALUES(39672,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(39672,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(39672,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(39672,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(39672,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(39672,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(39672,36,'cp 16','f1e2 c8e6 e1g1 b8d7 a2a4 a8c8 d1d3 e8g8 f1d1 f6g4 e3a7 c8a8 h2h3 d7f6 e2g4 f6g4 c3d5 a8a7 h3g4 e7g5 d3e2 a7a8 c2c3 a8b8 a4a5 b7b5 a5b6 e6d5 e4d5 d8b6 b3a5 f8c8 e2d3 b6b2 a5c4 b2b5 c4d6 b5d3 d1d3');
INSERT INTO evaluations VALUES(39671,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(39671,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(39671,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(39671,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(39671,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(39671,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(39671,36,'cp 6','c8e6 e1g1 e8g8 d1d3 b8d7 a2a4 a8c8 a4a5 d8c7 f1d1 f8d8 e2f3 d7c5 b3c5 d6c5 d3e2 d8d1 e2d1 h7h6 c3d5 e6d5 e4d5 c8d8 c2c4 e5e4 f3g4 c7e5 g4e2 f6d7 d1b3 e5c7 a1d1 d8e8 f2f4 e4f3 e2f3 d7e5 e3f4 e5f3 b3f3 c7a5');
INSERT INTO evaluations VALUES(28180,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(28180,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(28180,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28180,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28180,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(28180,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28180,36,'cp 22','e1g1 e8g8 d1d3 b8d7 a2a4 a8c8 a4a5 d8c7 f1e1 d7c5 b3c5 d6c5 e3c1 f8d8 d3g3 g8f8 c1e3 h7h6 h2h3 d8d4 e3d4 c5d4 c3d5 e6d5 e4d5 e7d6 g3b3 e5e4 a1a4 c7c2 b3c2 c8c2 a4d4 c2b2 d4a4 f8e7 g1f1 e7d7 e1d1 b2c2 a4c4');
INSERT INTO evaluations VALUES(28179,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(28179,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28179,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28179,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28179,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(28179,36,'cp 12','e8g8 d1d3 b8d7 a2a4 a8c8 a4a5 d8c7 f1d1 h7h6 e2f3 d7c5 b3c5 d6c5 c3a4 f8d8 d3c3 d8d1 a1d1 f6d7 h2h3 e7d8 f3g4 e6g4 h3g4 c7c6 b2b3 c6b5 c3d3 b5d3 d1d3 d7f6 a4c5 d8e7 g4g5 e7c5 g5f6 c5e3 f2e3 c8c2 d3d8 g8h7 d8d7 g7f6 d7f7 h7g6 f7b7 c2a2');
INSERT INTO evaluations VALUES(44927,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44927,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(44927,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(44927,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(44927,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44927,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44927,36,'cp 2','d8b6 b2b3 c8g4 g1e2 c6c5 b1c3 b8c6 e1g1 c5d4 c3a4 b6d8 f2f3 g4d7 e3d4 e7e6 c2c3 f6h5 f4d2 f8d6 d1e1 b7b6 g2g4 h5f6 e1h4 h7h6 a1e1 f6h7 h4h3 e8g8 a4b2 e6e5 d2h6 f7f5 h6e3 h7g5 h3g2 f5f4 d4e5 f4e3 e5d6');
INSERT INTO evaluations VALUES(11423,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(11423,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(11423,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(11423,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(11423,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(11423,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(11423,36,'cp 31','a4c6 b7c6 d2d4 e5d4 f3d4 c8d7 e1g1 f8e7 f1e1 e8g8 d1f3 f8e8 h2h3 h7h6 c1d2 e7f8 a1d1 d6d5 e4d5 c6d5 c3d5 f6d5 f3d5 d7h3 d5d8 e8d8 d2h6 h3g4 f2f3 g4f3 d4f3 g7h6 g1f1 d8d1 e1d1 g8g7 f3h4 g7f6 d1d5 f8b4 d5h5 b4d6 h5h6 f6g5');
INSERT INTO evaluations VALUES(15839,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(15839,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(15839,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(15839,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(15839,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(15839,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(15839,36,'cp -176','h7h6 g5f3 a5b3 a2b3 e5e4 f3g1 c8f5 g1e2 f8c5 e1g1 e8g8 b1c3 f8e8 d2d3 e4d3 c2d3 f6d5 d3d4 c5f8 c3d5 d8d5 e2c3 d5e6 c1d2 e6g6 d1f3 c7c6 f1e1 f5g4 f3g3 g6c2 d2h6 g4f5 e1d1 c2b3 d1d2 f5c2 h2h3 b3b2 a1a2 b2c1 g1h2 c2e4 h6f4 f8b4 c3e2 c1f1 f4e5');
INSERT INTO evaluations VALUES(44501,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44501,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44501,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44501,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44501,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44501,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44501,36,'cp 0','e7e6 c1e3 d8a5 e3d2 a5c5 e2e3 b7b6 g1f3 b8c6 f1e2 c8b7 e1g1 a8c8 d2c3 f6e4 d1d3 c6b4 d3d1');
INSERT INTO evaluations VALUES(44151,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44151,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44151,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(44151,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44151,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(44151,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44151,36,'cp 0','c1e3 d8a5 e3d2 a5c5 e2e3 f8e7 g1f3 e8g8 b1c3 b7b6 a1c1 b8c6 f1d3 c5h5 e3e4 c8b7 e1g1 d7d6 h2h3 a8d8 f1e1 c6e5 d3f1 h5g6 f3h4 g6h5');
INSERT INTO evaluations VALUES(47338,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(47338,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(47338,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(47338,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(47338,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(47338,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(47338,36,'cp 41','e7e5 d4e5 d8d1 c3d1 f6e4 d1e3 c8e6 e3d5 b8a6 a2a3 c7c6 d5f4 f8c5 c1e3 c5e3 f2e3 e6f5 g1f3 g7g5 f3d4 f5g6 f4g6 h7g6 b2b4 c6c5 f1d3 e4g3 h1g1 h8h2 d4f3 h2h5 e1f2 g5g4 f2g3 g4f3 g2f3 c5b4 a3b4');
INSERT INTO evaluations VALUES(47402,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(47402,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(47402,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47402,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47402,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47402,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47402,36,'cp 0','e4e5 f6d5 d2d4 c7c5 g1f3 c5d4 d1d4 b8c6 f1b5 c8f5 b1c3 e7e6 c3d5 e6d5 e1g1 f8e7 d4f4 f5g6 c1e3 e8g8 c2c3 f7f6 e5f6 f8f6 f4g3 g6e4 g3h3 e4f5 h3h5 f5g6 h5h3');
INSERT INTO evaluations VALUES(47308,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47308,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47308,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47308,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47308,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(47308,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47308,36,'cp 0','f6d5 d2d4 c7c5 g1f3 c5d4 d1d4 b8c6 f1b5 c8f5 b1c3 e7e6 c3d5 e6d5 e1g1 f8e7 d4f4 f5g6 c1e3 e8g8 c2c3 f7f6 e5f6 f8f6 f4g3 g6e4 g3h3 e4f5 h3h5 f5g6 h5g4');
INSERT INTO evaluations VALUES(2377,30,'cp 247',NULL);
INSERT INTO evaluations VALUES(2377,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(2377,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(2377,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(2377,34,'cp 255',NULL);
INSERT INTO evaluations VALUES(2377,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(2377,36,'cp 249','f2f3 d8e6 c1e3 f6d7 e5c4 b7b6 b1c3 h7h5 h2h4 c8b7 e1c1 e8c8 c3e2 g7g6 c2c3 f8g7 e2d4 d7c5 c1c2 a7a5 f1e2 h8e8 h1e1 c8b8 g2g3 a5a4 b2b4 a4b3 a2b3 e6d4 e3d4 c5e6 d4g7 d8d1 c2d1 e6g7 d1d2 f7f5 e2d3 f5e4');
INSERT INTO evaluations VALUES(5192,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(5192,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(5192,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(5192,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(5192,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(5192,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(5192,36,'cp 89','b1c3 d7d5 e4e5 f7f6 f1d3 f6e5 f3e5 g8f6 a2a3 b4a5 e1g1 e8g8 c1e3 c8d7 c3a4 a5c7 e5d7 d8d7 a4c5 d7f7 c5b7 e6e5 d4e5 c6e5 d3e2 e5c4 e2c4 d5c4 a1c1 c7h2 g1h2 f7b7 d1d4 a8d8 d4h4 h7h6 h4c4 g8h8 e3d4 f6e4');
INSERT INTO evaluations VALUES(5189,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(5189,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(5189,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(5189,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(5189,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(5189,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(5189,36,'cp 90','d7d6 f1d3 g8e7 c1e3 e8g8 a1c1 d6d5 e4e5 e7f5 a2a3 b4e7 e1g1 f5e3 f2e3 c8d7 e3e4 d5e4 c3e4 d7e8 g1h1 g8h8 d3b1 f7f5 e5f6 g7f6 e4c5 e7c5 c1c5 e8f7 d1d2 f8g8 c5c1 g8g7 d2h6 c6d4 f3d4 d8d4');
INSERT INTO evaluations VALUES(10470,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(10470,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(10470,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(10470,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(10470,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(10470,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(10470,36,'cp -46','h7h6 d4c6 b7c6 g5h4 e8g8 f1d3 d7d5 e1g1 b4c3 b2c3 d5e4 h4f6 d8f6 d3e4 f6c3 d1f3 c3f3 e4f3 f8d8 f3c6 a8b8 c6e4 c8e6 f2f3 c7c5 f1e1 g7g5 g2g4 d8d4 g1f2 g8g7 f2g3 a7a5 a2a3 d4a4 e1e3 a4c4 h2h4 g7f6 a3a4 b8b4 e4d3 c4c3');
INSERT INTO evaluations VALUES(9867,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(9867,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(9867,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(9867,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(9867,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(9867,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(9867,36,'cp -39','d4c6 b7c6 g5h4 e8g8 f1d3 d7d5 e1g1 b4c3 b2c3 d5e4 h4f6 d8f6 d3e4 f6c3 d1d3 c3c5 f1e1 c8e6 d3e3 c5e3 e1e3 e6d5 e4d5 c6d5 e3c3 f8c8 c3c5 c7c6 c2c4 d5c4 a1c1 a8b8 c5c4 b8b7 g2g3 b7c7 c4c5 g8f8 c1d1 c8e8 g1g2 e8e2 d1c1 e2a2 c5c6 c7c6 c1c6');
INSERT INTO evaluations VALUES(15293,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(15293,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(15293,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(15293,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(15293,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(15293,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(15293,36,'cp -55','f7f6 g5d2 c8e6 d1e2 f8c5 b1c3 d5c3 d2c3 e6c4 d3c4 a7a5 a2a3 e8g8 f3d2 c5b4 a3b4 a5b4 a1a8 d8a8 e1g1 b4c3 b2c3 a8a5 e2f3 a5a2 f3d5 g8h8 d5b5 c6d8 b5b3 a2a8 c4c5 a8a6 f1b1 d8e6 b3b5 a6a3 b5b3 a3b3 d2b3 f8a8 g1f1 f6f5 g2g3');
INSERT INTO evaluations VALUES(15867,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(15867,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(15867,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(15867,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(15867,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(15867,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(15867,36,'cp -58','g5d2 c8e6 d1e2 f8b4 c2c3 b4d6 f3d4 e6f7 d4c6 b7c6 b1a3 a7a5 e1g1 e8g8 e2f3 g8h8 c4b3 a5a4 b3c2 d5b6 c3c4 d6a3 b2a3 b6c4 d2b4 f7d5 f3h3 c4d6 f1e1 d8c8 d3d4 e5e4 h3c8 f8c8 a1d1 g7g6');
INSERT INTO evaluations VALUES(16009,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(16009,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(16009,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16009,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(16009,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16009,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(16009,36,'cp 16','h2h3 e7e5 f4h2 d8e7 g1f3 g7g6 b1c3 a7a6 f1c4 e5d4 d1d4 d7b6 e1c1 f8g7 e3e4 e8g8 e4e5 d6e5 h2e5 c7c5 d4h4 b6c4 h4c4 b7b5 c4f4 f8e8 h1e1 c8e6 c3e4 f6e4 e5g7 g8g7 e1e4 e8d8 d1d8 a8d8 f3g5 e7d6 g5e6 f7e6 f4e5 g7f7 e5d6 d8d6');
INSERT INTO evaluations VALUES(21016,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(21016,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(21016,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(21016,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(21016,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(21016,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(21016,36,'cp -28','e7e5 g1f3 e5e4 f3d2 c6e7 c2c4 c7c6 b1c3 g8f6 c4d5 c6d5 f1b5 c8d7 d1b3 d7b5 b3b5 d8d7 b5d7 e8d7 e1g1 a8c8 f2f3 e4f3 d2f3 e7g6 f3e5 d7e6 e5d3 f8d6 c3b5 c8c2 f1f2 c2f2 d3f2 d6b8 f2d3 f6e4 b5c3 g6h4 c3e4 d5e4 d3c5 e6d5 c5b7');
INSERT INTO evaluations VALUES(21691,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(21691,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(21691,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(21691,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(21691,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(21691,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(21691,36,'cp -6','g1f3 e5e4 f3d2 c6e7 c2c4 f7f5 f1e2 g8f6 e1g1 c7c6 b1c3 a7a6 c4d5 e7d5 d2c4 c8e6 c1d2 f8d6 d1b3 b7b5 c4d6 d8d6 b3c2 h7h5 c3d5 e6d5 a2a4 e8f7 c2c5 d6c5 d4c5 g7g5 d2c3 f7e6 f2f3 h8e8 f1d1 e4f3 g2f3 g5g4 f3g4 h5g4 h3g4 f5g4');
INSERT INTO evaluations VALUES(23193,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(23193,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(23193,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(23193,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(23193,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(23193,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(23193,36,'cp 165','g8f6 g2g3 d7d5 f1g2 c8e6 e1g1 d8d7 a2a3 f8e7 b2b4 e8c8 c1b2 h7h5 b1d2 h5h4 f3h4 h8h4 g3h4 e6h3 g1h1 h3g2 h1g2 d7g4 g2h1 g4h4 d2f3 h4h5 b4b5 c6a5 b2f6 e7f6 f1g1 f6a1 d1a1 d5d4 g1g7 h5b5');
INSERT INTO evaluations VALUES(17117,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(17117,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(17117,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(17117,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(17117,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(17117,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(17117,36,'cp 177','g2g3 d7d5 f1g2 c8e6 e1g1 d8d7 a2a3 e8c8 b2b4 h7h5 h2h4 e6h3 d1d3 f8e7 c1b2 h3g2 g1g2 f6e4 c2c4 e7f6 b2f6 g7f6 b1c3 e4c3 d3c3 d5d4 c3d3 c6e5 f3e5 f6e5 d3e4 d7e6 a1d1 c8b8 e2e3 d4e3 d1d8 h8d8 f2e3');
INSERT INTO evaluations VALUES(39161,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(39161,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(39161,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(39161,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(39161,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(39161,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(39161,36,'cp 63','b4d2 b1d2 d7d6 d1f3 b8c6 f3c3 c8d7 g1e2 d8f6 g2g3 h7h5 f1g2 h5h4 f2f4 e8c8 e1c1 h4g3 e4e5 f6e7 h2g3 g8h6 g2f3 h6f5 c1b1 d6d5 g3g4 f5h4 a2a3 h4f3 d2f3 b7b6 e2c1 h8h1 d1h1 c8b8');
INSERT INTO evaluations VALUES(39180,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(39180,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(39180,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(39180,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(39180,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(39180,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(39180,36,'cp 65','b1d2 g8e7 f1d3 b8c6 g1f3 d7d6 d2c4 e8g8 d1d2 f7f5 e4f5 e7f5 c2c3 d8f6 e1c1 h7h6 h2h4 b7b5 c4e3 a8b8 d3c2 b5b4 c3c4 e6e5 d4e5 c6e5 e3d5 f6f7 b2b3 c8e6 d5f4 e5f3 g2f3 b8e8 d2b4');
INSERT INTO evaluations VALUES(42293,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(42293,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(42293,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(42293,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(42293,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(42293,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(42293,36,'cp 32','e7g5 h4g5 d8g5 d1d3 h7h6 g1h3 g5e7 e1c1 b8c6 d3g3 g7g6 h3g5 a7a6 f1d3 d7f8 g3e3 c8d7 f2f4 e8c8 g5f3 c8b8 g2g4 d7e8 a2a3 h6h5 c3e2 f7f6 h1h3 f8d7 d1h1 h8g8 g4h5 g6h5 c1b1 f6e5 f4e5 d7b6 e2f4 b6c4 e3f2');
INSERT INTO evaluations VALUES(42286,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(42286,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(42286,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(42286,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(42286,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(42286,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(42286,36,'cp 25','h4g5 d8g5 g1h3 g5h4 g2g3 h4e7 d1g4 g7g6 h3g5 h7h6 f1d3 b8c6 e1c1 d7f8 g5f3 c8d7 h1h2 f7f5 g4f4 g6g5 f4e3 e7g7 f3d2 f8g6 f2f4 e8c8 d1h1 g6e7 a2a3 a7a5 d2f3 g5g4 f3e1 b7b6 b2b3 c8b7 c3e2 d7e8 c1b2 h6h5 e3f2 g7f7');
INSERT INTO evaluations VALUES(42285,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(42285,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(42285,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42285,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(42285,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(42285,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(42285,36,'cp 20','d8g5 g1h3 g5h4 g2g3 h4e7 d1g4 g7g6 h3g5 h7h6 e1c1 b8c6 f1d3 d7f8 g5f3 c8d7 h1h2 f7f5 g4f4 g6g5 f4e3 e7g7 f3d2 f8g6 f2f4 e8c8 d1h1 g6e7 a2a3 a7a5 d2f3 g5g4 f3e1 c8b8 b2b3 h6h5 c3e2 a5a4 b3b4 c6a7 h2h5 a7b5 c1b2 h8h5 h1h5 d8h8 h5h8 g7h8 d3b5');
INSERT INTO evaluations VALUES(33432,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(33432,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(33432,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(33432,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(33432,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(33432,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(33432,36,'cp 20','g1h3 g5h4 g2g3 h4e7 d1g4 g7g6 h3g5 h7h6 f1d3 b8c6 e1c1 d7f8 h1h2 c8d7 g5f3 f7f5 g4f4 g6g5 f4e3 e7g7 f3d2 e8c8 f2f4 f8g6 d1h1 g6e7 a2a3 a7a5 d2f3 g5g4 f3e1 b7b6 b2b3 c8b7 c1b1 d7e8 b1b2 h6h5 d3b5 g7f8');
INSERT INTO evaluations VALUES(33431,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(33431,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(33431,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(33431,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(33431,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(33431,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(33431,36,'cp 7','g5h4 g2g3 h4e7 d1g4 g7g6 h3g5 h7h6 e1c1 b8c6 f1d3 d7f8 g5f3 c8d7 h1h2 f7f5 g4f4 g6g5 f4e3 e7g7 f3d2 e8c8 f2f4 f8g6 d1h1 h6h5 d2b3 g5g4 h2h5 h8h5 h1h5 d8h8 h5h8 g7h8 a2a3 b7b6 d3a6 c8b8 e3g1 d7c8 a6b5 c8d7 b5c6');
INSERT INTO evaluations VALUES(43170,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43170,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(43170,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43170,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(43170,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43170,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43170,36,'cp 21','d2d4 c7c6 b1c3 d7d5 d1b3 e8g8 f1g2 d8b6 e1g1 c8f5 f1e1 b8d7 c4d5 b6b3 a2b3 c6d5 h2h3 h7h5 c1f4 a7a6 e1c1 a8c8 f3e5 e7e6 f2f3 g6g5 f4g5 d7e5 d4e5 f6d7 g3g4 f5h7 f3f4 h5g4 h3g4 d7c5 g5f6 g7f6 e5f6 c5b3');
INSERT INTO evaluations VALUES(50436,30,'cp -196',NULL);
INSERT INTO evaluations VALUES(50436,31,'cp -193',NULL);
INSERT INTO evaluations VALUES(50436,32,'cp -190',NULL);
INSERT INTO evaluations VALUES(50436,33,'cp -192',NULL);
INSERT INTO evaluations VALUES(50436,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(50436,35,'cp -217',NULL);
INSERT INTO evaluations VALUES(50436,36,'cp -193','d7d5 c4e2 c7c5 e1g1 b8c6 f1e1 e7e6 e2b5 e4d6 b5c6 b7c6 c2c3 f8e7 d2d4 d6f5 b2b3 e8g8 c1a3 c5d4 a3e7 d8e7 c3d4 c6c5 b1c3 c8b7 a1c1 a8c8 c3e2 b7a6 d4c5 a6e2 d1e2 c8c5 e2b2 c5b5 f3d4 f5d4 b2d4');
INSERT INTO evaluations VALUES(47485,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(47485,31,'cp -188',NULL);
INSERT INTO evaluations VALUES(47485,32,'cp -197',NULL);
INSERT INTO evaluations VALUES(47485,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(47485,34,'cp -200',NULL);
INSERT INTO evaluations VALUES(47485,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(47485,36,'cp -185','c4e2 c7c5 e1g1 b8c6 e2b5 c8d7 d2d3 e4f6 c2c4 e7e6 c4d5 f6d5 b1c3 a7a6 b5a4 d5f6 d3d4 c5d4 f3d4 b7b5 a4c2 f8e7 d4f3 e8g8 d1e2 d8c7 c3e4 c6b4 e4f6 e7f6 c2e4 a8c8 c1d2 b4d5 e4d5 e6d5 a1c1 c7b7 d2e3 f8d8');
INSERT INTO evaluations VALUES(52496,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(52496,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(52496,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(52496,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(52496,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(52496,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(52496,36,'cp 224','e2e4 e7e5 f1c4 f8c5 g1e2 d7d6 e1g1 g8e7 a2a4 b8d7 b1d2 e8g8 e2g3 a7a5 c4b5 e7g6 d2b3 c5b6 c1e3 b6e3 f2e3 d8g5 d1e1 g6h4 b3a5 d7c5 b2b4 c5e4 b5d3 e4g3 e1g3 g5h6 a5b3 f7f5 a4a5 e5e4 d3c4');
INSERT INTO evaluations VALUES(52495,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(52495,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(52495,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(52495,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(52495,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(52495,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(52495,36,'cp 206','e7e5 f1c4 f8c5 g1e2 d7d6 e1g1 g8e7 a2a4 b8d7 e2g3 e8g8 b1c3 e7g6 c3e2 d7f6 h2h3 h7h6 c4d3 c8d7 a4a5 b7b5 c1d2 a7a6 d1e1 c5a7 b2b3 f6h7 g1h2 d8h4 f2f3 a7c5 g3f5 h4e1 f1e1 d7f5 e4f5 g6e7 b3b4 c5a7');
INSERT INTO evaluations VALUES(4803,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(4803,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(4803,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(4803,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(4803,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(4803,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(4803,36,'cp -20','g8e7 d3d4 e5d4 d1d4 e8g8 e4e5 e7f5 d4e4 f8e8 b1c3 f7f6 c1f4 f6e5 f1e1 d8f6 f3e5 e8e7 e4c4 c8e6 c4d3 a8e8 c3e4 f6f8 e4d6 c7d6 e5f3 e6d5 e1e7 e8e7 h2h3 f8f6 a1e1 e7e1 f3e1 d5a2 b2b3 h7h6 d3e4');
INSERT INTO evaluations VALUES(13117,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(13117,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(13117,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(13117,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(13117,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(13117,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(13117,36,'cp -24','e7e5 e2e4 d5d4 a2a4 c8g4 h2h3 g4h5 g3g4 h5g6 f3e5 c6e5 f2f4 f6d7 f4e5 d7e5 b1a3 f7f6 c1f4 f8d6 f4e5 d6e5 a3c4 d8e7 c4e5 e7e5 h3h4 e8g8 d1d2 a8e8 d2f4 g6f7 a4a5 h7h6 g2h3 f7e6 b2b3 e5f4 f1f4 g7g5 f4f1');
INSERT INTO evaluations VALUES(17849,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(17849,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(17849,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(17849,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(17849,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(17849,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(17849,36,'cp 35','g8f6 d3d4 f8d6 d4e5 c6e5 c4b3 e8g8 c1f4 d8e7 f4g3 e5f3 d1f3 d6g3 h2g3 d7d6 f3f4 c8e6 e1c1 a8d8 f2f3 e6b3 a2b3 f8e8 g3g4 e7e5 c3e2 e5f4 e2f4 a6a5 c2c3 g8f8 c1c2 b7b6 d1d2 d8b8 f4e2 f6d7 e2d4');
INSERT INTO evaluations VALUES(18842,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(18842,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18842,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(18842,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18842,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18842,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18842,36,'cp 0','g8f6 d2d3 d7d5 d1e2 f8e7 a2a4 e8g8 e1g1 h7h6 c2c3 a7a6 e4e5 f6d7 c1f4 a8b8 c3c4 d5c4 d3c4 d8c7 h2h4 b7b6 f1d1 c8b7 b1c3 b8d8 c3d5 c7c8 d5c3');
INSERT INTO evaluations VALUES(12441,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12441,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12441,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12441,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12441,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(12441,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(12441,36,'cp 1','b1c3 d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 c8g4 d4e2 f8b4 h2h3 b4c3 b2c3 g4f5 c1g5 h7h6 g5f6 d8f6 e1g1 e8g8 d1d5 f5c2 d5b5 a8e8 b5b7 e8e2 b7c6 f6c6 g2c6 c2f5 f1e1 e2e1 a1e1 f5e6 e1e5 e6a2 e5a5 a2e6 a5a7 f8c8 c6e4 e6h3 e4d5 c8c3');
INSERT INTO evaluations VALUES(25792,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(25792,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(25792,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(25792,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(25792,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(25792,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(25792,36,'cp 69','f8e7 b1c3 g8f6 f2f4 a7a5 d1f3 e8g8 c1e3 d8c8 h2h3 c8b7 e3d4 c6c5 d4f2 a5a4 a1d1 f8e8 b2b3 e7f8 f1e1 a4b3 a2b3 f6d7 g1h2 c7c6 f2g3 b7b4 d1d2 b4b7 e1d1 a8a3 f4f5 d7e5 g3e5 e8e5 f3g3 e5e8 f5f6');
INSERT INTO evaluations VALUES(23790,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(23790,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(23790,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(23790,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(23790,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(23790,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(23790,36,'cp 90','b1c3 g8f6 f2f4 e8g8 d1f3 a7a5 c1e3 d8c8 h2h3 c8b7 b2b3 d6d5 e4e5 e7b4 e5f6 b4c3 f6g7 f8e8 a1d1 b7b4 e3f2 c3b2 d1e1 b2g7 a2a4 b4c3 f3d1 c3b4 d1g4 e8e4 e1e4 b4e4 f1e1 e4c2 e1e7 c2b3 g1h2 b3a4');
INSERT INTO evaluations VALUES(40906,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(40906,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(40906,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(40906,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(40906,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(40906,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40906,36,'cp 20','e7f6 g5f6 d8f6 g1f3 e8g8 d1d3 f8d8 e1c1 c7c5 d3e4 b8c6 f1d3 f6h6 e4e3 h6e3 f2e3 c5d4 d3e4 d4d3 e4c6 b7c6 d1d3 d8d3 c2d3 c8a6 d3d4 a6d3 h1g1 f7f6 b2b3 a7a5 c1b2 a5a4 f3d2 d3b5 g1e1 a4b3 a2b3 g8f7 d2e4 e6e5 d4e5 b5d3');
INSERT INTO evaluations VALUES(41972,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(41972,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(41972,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41972,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(41972,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41972,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(41972,36,'cp 21','g5f6 d8f6 g1f3 e8g8 d1d3 c7c5 e1c1 f8d8 d3e4 b8c6 f1d3 f6h6 e4e3 h6e3 f2e3 c5d4 d3e4 d4d3 e4c6 b7c6 d1d3 d8d3 c2d3 c8a6 d3d4 a6d3 h1g1 d3e4 f3d2 e4g6 d2b3 e6e5 g2g3 a8e8 c1d2 g6e4 d2c3 e5d4 c3d4 e4d5 b3c5 e8d8 d4c3 f7f5 g1a1 g8f7');
INSERT INTO evaluations VALUES(41960,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(41960,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(41960,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(41960,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(41960,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41960,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41960,36,'cp 15','d8f6 g1f3 c7c5 f1b5 c8d7 d1d3 c5d4 e1c1 b8c6 f3d4 c6d4 b5d7 e8d7 d3d4 f6d4 d1d4 d7e7 h1d1 h8d8 d4d8 a8d8 d1d8 e7d8 c1d2 d8d7 d2d3 d7d6 h2h4 d6c5 h4h5 f7f6 g2g4 a7a5 d3e4 c5c4 g4g5 h7h6 g5g6 b7b6 f2f3 c4c5 f3f4 c5c4 a2a4 c4b4');
INSERT INTO evaluations VALUES(55594,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55594,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55594,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55594,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55594,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55594,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55594,36,'cp 0','b8c6 g1f3 f7f5 b1c3 g8f6 d2d4 f5e4 d4e5 e4f3 e5f6 d8f6 d1f3 f6f3 g2f3 c8e6 e1c1 e8c8 f3f4 d6d5 f1g2 d5d4 c3b5 e6g4 d1g1 g4e6');
INSERT INTO evaluations VALUES(21955,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21955,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21955,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21955,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21955,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21955,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21955,36,'cp 0','g1f3 f7f5 b1c3 g8f6 d2d4 f5e4 d4e5 e4f3 e5f6 d8f6 d1f3 f6f3 g2f3 c8e6 e1c1 e8c8 f3f4 d6d5 f1g2 d5d4 c3b5 e6g4 d1g1 g4e6');
INSERT INTO evaluations VALUES(56924,30,'cp 326',NULL);
INSERT INTO evaluations VALUES(56924,31,'cp 320',NULL);
INSERT INTO evaluations VALUES(56924,32,'cp 337',NULL);
INSERT INTO evaluations VALUES(56924,33,'cp 318',NULL);
INSERT INTO evaluations VALUES(56924,34,'cp 334',NULL);
INSERT INTO evaluations VALUES(56924,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(56924,36,'cp 342','d2d4 e5d4 c1g5 f8b4 c2c3 b4e7 d1d4 e7f6 d4e3 d7d6 h2h4 h7h6 g5f4 b8c6 b1a3 a7a6 e1c1 f6g7 e4e5 d6d5 a3c2 g8e7 f1d3 c8g4 h1e1 d8d7 f3d4 e8c8 f2f3 g4e6 c1b1 c6d4 c2d4 e7f5 d4f5 e6f5 g2g4 f5d3 e3d3 c8b8 f4g3 h8e8');
INSERT INTO evaluations VALUES(57684,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(57684,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57684,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57684,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57684,36,'cp 0','d7d5 d2d4 g8f6 c1f4 f8g7 e2e3 e8g8 f1e2 c7c5 e1g1 b7b6 c3b5 b8c6 b5c7 a8b8 c7b5');
INSERT INTO evaluations VALUES(5266,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(5266,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5266,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5266,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5266,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5266,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(5266,36,'cp 12','e7e5 d2d3 g8e7 c1g5 h7h6 g5e7 c6e7 a2a4 e8g8 b5c4 d7d6 c3d5 c8d7 c2c3 a8c8 a4a5 g8h7 d1b3 c8b8 d5e7 d8e7 f1b1 d7c6 c4d5 a7a6 d5c6 b7c6 b3a4 f8c8 g2g3 d6d5 g1g2 e7d7 b2b4 c5b4 b1b4 b8b4');
INSERT INTO evaluations VALUES(10630,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(10630,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(10630,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(10630,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(10630,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(10630,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(10630,36,'cp -50','d7d5 f3e5 c6e5 c4b3 e8g8 f2f3 e4f6 d4c5 f8e8 b1a3 d8e7 c1g5 e7c5 d1d4 c5d4 c3d4 e5g6 g5f6 g7f6 b3d5 g6f4 d5c4 c8e6 a1c1 a8d8 f1d1 c7c6 g1f1 d8d7 g2g3 f4d5 a3c2 d5b6 c4e6 f7e6 c2a1 e8d8 a1b3 b6a4 d1d2 e6e5 c1d1 e5d4 d2d4 d7d4 d1d4 d8d4 b3d4 a4b2');
INSERT INTO evaluations VALUES(9169,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(9169,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(9169,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(9169,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(9169,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(9169,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(9169,36,'cp -44','f3e5 c6e5 c4b3 e8g8 f2f3 e4f6 d4c5 f8e8 b1a3 d8e7 c1g5 e7c5 d1d4 c5d4 c3d4 e5g6 g5f6 g7f6 b3d5 g6f4 d5c4 c8e6 a1c1 a8d8 f1e1 c7c6 c1d1 e6c4 e1e8 d8e8 a3c4 b7b5 c4a5 e8e2 g2g3 f4d5 d1c1 e2b2 a2a3 a7a6 c1c6 d5e3 c6a6 b2g2 g1h1 g2a2 h1g1 a2a3 g1f2');
INSERT INTO evaluations VALUES(20908,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(20908,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(20908,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(20908,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(20908,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(20908,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(20908,36,'cp 98','e7e6 c2c4 f8d6 f4g5 f7f6 g5h4 d5c4 f1c4 g8h6 h4g3 h6f5 b1c3 d8e7 g3d6 f5d6 c4e2 e8g8 a1c1 b7b6 e1g1 c8b7 a2a3 c6d8 c3b5 d6b5 e2b5 c7c6 b5e2 d8f7 b2b4 g8g7 d1b3 f7d6 f1d1 f8d8 f3d2 a8c8 b3b2');
INSERT INTO evaluations VALUES(22222,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(22222,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(22222,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(22222,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(22222,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(22222,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(22222,36,'cp 42','g8f6 f1b5 f8d6 c2c4 d5c4 b1c3 e8g8 b5c4 g8h8 e1g1 e6e5 f3e5 c6e5 d4e5 d6e5 d1d8 f8d8 f1e1 e5d6 c1g5 h7h6 g5h4 g7g5 h4g3 c8f5 f2f3 d6g3 h2g3 d8d7 c3a4 a8e8 e1e8 f6e8 a1e1 f5g6 g3g4 h8g7 a4c5 d7d2 c5b7 d2b2 b7d8 b2b1 e1b1 g6b1');
INSERT INTO evaluations VALUES(15862,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(15862,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(15862,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(15862,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(15862,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(15862,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(15862,36,'cp 42','f1e2 f8d6 c2c4 d5c4 e2c4 e8g8 e1g1 g8h8 b1c3 e6e5 d4e5 c6e5 f3e5 d6e5 d1d8 f8d8 f1e1 e5d4 c1e3 c8f5 e3d4 d8d4 c4b3 a8d8 f2f3 h7h6 e1e7 d4d7 a1d1 d7d1 c3d1 d8d7 e7d7 f5d7 g1f2 g7g5 f2e3 c7c5 d1c3 b7b6 c3e4 h8g7 e4d6 g7f8 d6f7 f8g7');
INSERT INTO evaluations VALUES(22721,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(22721,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(22721,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(22721,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(22721,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(22721,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(22721,36,'cp -62','c6d4 g1f3 d4f3 d1f3 g8f6 f3b3 d8e7 b1c3 c7c6 a2a4 d7d6 e1g1 c8e6 c4e6 f7e6 f2f4 e5f4 c1f4 e6e5 f4g5 e7f7 b3b4 a7a5 b4b6 a8a6 b6e3 f8e7 e3d3 f7e6 a1d1 a6b6 g5f6 e7f6 b2b3 b6b4 d3d6 e6d6 d1d6 f6e7 d6d3 e7c5 g1h1 b4d4 d3d4 c5d4 c3d1 e8e7 c2c3 d4c5 d1b2 h8d8');
INSERT INTO evaluations VALUES(23109,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(23109,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(23109,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(23109,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(23109,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(23109,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(23109,36,'cp -71','g1f3 d4f3 d1f3 g8f6 f3b3 d8e7 b1c3 c7c6 a2a4 d7d6 e1g1 c8e6 c4e6 f7e6 f2f4 e5f4 c1f4 e6e5 f4g5 e7f7 b3b4 a7a5 b4b6 a8a6 b6e3 f8e7 a1d1 f7e6 e3d3 a6b6 g5f6 e7f6 b2b3 b6b4 d3d6 e6d6 d1d6 f6e7 d6d1 e7c5 g1h1 b4d4 d1d4 c5d4 c3e2 e8e7 e2d4 e5d4 h2h3 g7g6 f1d1 h8d8 h1g1 b7b5');
INSERT INTO evaluations VALUES(34709,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(34709,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(34709,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(34709,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(34709,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(34709,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(34709,36,'cp 67','d6e5 b1c3 g8f6 f1c4 c8g4 e1g1 b8c6 h2h3 g4f3 d1f3 c6d4 f3d3 e8c8 a2a4 c7c6 c3e2 e7c7 a1b1 c8b8 c2c3 d4b3 d3f3 b3c1 e2c1 f8d6 b2b4 d8d7 c1b3 h7h5 b3a5 h8d8 f3f5 h5h4 f1e1 a7a6 f5f3 d6e7 c4f7 g7g5');
INSERT INTO evaluations VALUES(34744,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(34744,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(34744,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(34744,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(34744,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(34744,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(34744,36,'cp 66','b1d2 f7f6 f1b5 c7c6 b5c4 g8h6 d2b3 h6f7 c1e3 f7d6 c4d3 c8e6 f3d2 b7b6 d1e2 c6c5 d2c4 e7c7 c4d6 f8d6 d3c4 c7d7 e1c1 b8c6 c2c3 d7f7 b3d2 e8g8 c4e6 f7e6 e2c4 g8f7 c4e6 f7e6 d2c4 d6e7 a2a4 f8d8 c1c2 d8d1 c2d1 c6d8 f2f3');
INSERT INTO evaluations VALUES(46209,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(46209,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46209,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46209,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46209,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(46209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46209,36,'cp 0','e2e3 b8c6 a2a3 c5d4 e3d4 f6h5 f4g5 d8b6 b1c3 c8d7 c3a4 b6c7 c2c4 d5c4 f1c4 f8e7 g5e3 c6d4 d1d4 d7a4 c4e6 a4c6 e6a2 e8g8 e1g1 c6f3 g2f3 h5g3 f2g3 c7g3 g1h1 g3h3 h1g1 h3g3 g1h1');
INSERT INTO evaluations VALUES(46205,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(46205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46205,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(46205,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46205,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46205,36,'cp 0','d8b6 d1c1 c8d7 b1d2 a7a5 a2a4 f8d6 f4d6 b6d6 f1d3 c5c4 d3e2 e8g8 e1g1 d7c6 b2b3 c4b3 d2b3 f6e4 c2c3 b8d7 f3d2 e4d2 c1d2 d7b6 b3c5 b6d7');
INSERT INTO evaluations VALUES(54007,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(54007,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(54007,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54007,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(54007,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(54007,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(54007,36,'cp 30','d2d4 c5d4 f3d4 g8f6 b1c3 a7a6 f1d3 b8d7 a2a4 d7c5 c1e3 g7g6 d4b3 c5b3 c2b3 f8g7 d3e2 e8g8 e1g1 b7b6 e2f3 c8b7 d1e2 d6d5 a1d1 d8e7 e4d5 f6d5 c3d5 e6d5 e2f2 e7b4 e3b6 b4b3 a4a5 f8e8 h2h3 b3b2 f2b2 g7b2');
INSERT INTO evaluations VALUES(53997,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(53997,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(53997,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(53997,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(53997,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(53997,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(53997,36,'cp 26','c5d4 f3d4 a7a6 f1e2 g8f6 b1c3 f8e7 c1e3 e6e5 f4e5 d6e5 d4f5 d8d1 a1d1 c8f5 e4f5 e7b4 e1g1 b4c3 b2c3 b8c6 g2g4 h7h6 e3b6 a8c8 h2h4 c6b8 d1d3 e8g8 g4g5 h6g5 h4g5 f6e4 e2f3 e4c3 f3b7 c3e2 g1h1 c8c2 f1e1 c2b2 b7e4 b2b6 e1e2 f8c8 h1g2');
INSERT INTO evaluations VALUES(14736,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(14736,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(14736,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(14736,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(14736,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(14736,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(14736,36,'cp 13','f8d6 b5d6 c7d6 c2c3 f6e4 f3d2 e8g8 d2e4 d5e4 e2e3 e6e5 f4g3 e5d4 c3d4 c6e7 f1e2 c8e6 e1g1 a8c8 h2h3 d8b6 a2a4 b6b4 a4a5 c8c6 d1e1 b4e1 f1e1 e7f5 g3f4 f8c8 e2b5 c6c2 b5a4 c2b2 e1b1 b2b1 a1b1 b7b6 a5b6 a7b6');
INSERT INTO evaluations VALUES(9220,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(9220,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(9220,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9220,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(9220,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(9220,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(9220,36,'cp 19','b5d6 c7d6 e2e3 e8g8 c2c3 f8e8 h2h3 e6e5 f4h2 d8b6 d1b3 b6c7 f1e2 c8f5 e1g1 c6a5 b3d1 a7a6 a1c1 a8c8 b2b3 b7b5 d1e1 f6e4 c3c4 d5c4 d4e5 d6e5 h2e5 c7b6 e5d4 b6d8 f3d2 a5c6 d2e4 f5e4 b3c4 c6d4 e3d4');
INSERT INTO evaluations VALUES(9195,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(9195,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(9195,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(9195,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9195,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(9195,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(9195,36,'cp 6','c7d6 e2e3 e8g8 c2c3 f8e8 h2h3 e6e5 f4h2 d8b6 d1b3 b6c7 f1e2 c8f5 e1g1 c6a5 b3d1 a7a6 a1c1 a8c8 c3c4 d5c4 b2b4 a5c6 d4e5 d6e5 e2c4 c7e7 b4b5 a6b5 c4b5 e8d8 d1b3 f6e4 b5c6 b7c6 h2e5 c6c5 f1d1 d8d1 c1d1 e7e6 b3b2 f7f6 e5f4 h7h6 g1h2 g7g5 f4g3 e6a6 f3d2 e4g3 f2g3 f5d3');
INSERT INTO evaluations VALUES(8419,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(8419,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8419,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8419,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(8419,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(8419,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(8419,36,'cp 25','c2c3 e8g8 e2e3 f8e8 h2h3 e6e5 f4h2 h7h6 f1e2 c8f5 e1g1 f5g6 a1c1 a8c8 f1e1 d8c7 d4e5 d6e5 b2b4 a7a6 a2a4 g6h5 d1d2 c7e7 d2a2 e5e4 f3d4 c6d4 e2h5 f6h5 c3d4 h5f6 h2e5 f6d7 e5f4 e7b4 a2d5 c8c1 e1c1');
INSERT INTO evaluations VALUES(16128,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(16128,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(16128,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(16128,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(16128,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(16128,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(16128,36,'cp 35','f2f4 f8b4 g1e2 d7d5 e4d5 f6d5 e1g1 d5f6 a2a3 b4c5 g1h1 e8g8 c1d2 e5f4 e2f4 c5d4 d1e1 f8e8 e1h4 c6e5 c4b3 c7c5 h4g3 c8f5 c3d5 e5c6 d5f6 d8f6 c2c3 d4e5 g3f2 f6d6 f4h5 f5e6 b3c2 a8d8 a1e1 f7f5 h5g3 e5g3 f2g3 d6g3');
INSERT INTO evaluations VALUES(16153,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(16153,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(16153,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(16153,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(16153,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(16153,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(16153,36,'cp 37','f8c5 g1f3 d8e7 f4e5 c6e5 f3e5 e7e5 d1f3 d7d6 c3d5 f6d5 c4d5 e5e7 c2c3 c7c6 d5b3 c8e6 c1f4 e6b3 a2b3 c5b6 e1c1 e7e6 h2h4 a7a5 f3g3 e8c8 c1c2 h8g8 h4h5 b6c7 g3f2 d6d5 f4c7 c8c7 h1e1 d5d4 f2f4 c7b6 b3b4 a5b4 c3b4 d8a8 d1a1 a8a1 e1a1');
INSERT INTO evaluations VALUES(24315,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(24315,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(24315,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(24315,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(24315,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(24315,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(24315,36,'cp 66','b5c4 d7g4 a2a4 f8e7 h2h3 g4h5 d2d3 d8d7 c1g5 c6d8 b1d2 d8e6 g5h4 e8g8 h4f6 h5f3 d2f3 e7f6 g2g3 b7b5 c4d5 c7c6 d5a2 f8d8 h3h4 g7g6 g1g2 h7h5 a1c1 a8c8 d3d4 e5d4 c3d4 c6c5 a4b5 d7b5 a2e6 f7e6 d4d5');
INSERT INTO evaluations VALUES(54159,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(54159,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(54159,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(54159,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(54159,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(54159,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(54159,36,'cp -29','e4d5 e6d5 c4e2 b8c6 g1f3 g8f6 d2d4 c5c4 b1c3 f8d6 c1g5 c8e6 e1g1 h7h6 g5h4 e8g8 f1e1 f8e8 d1d2 d6e7 a1d1 d8b6 d2c1 f6e4 h4e7 e8e7 c3e4 d5e4 f3e5 c6e5 d4e5 b6b5 c1f4 b5b2 f4e4 b2c3 f2f4 a8e8 f4f5 e6d7 e5e6 f7e6 e2c4 b7b5');
INSERT INTO evaluations VALUES(54123,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(54123,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(54123,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(54123,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54123,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(54123,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(54123,36,'cp -29','e6d5 c4e2 g8f6 d2d4 h7h6 g1f3 c5c4 c1f4 d8b6 d1c1 c8e6 b1d2 f8d6 f4d6 b6d6 c2c3 e8g8 e1g1 b8c6 f1e1 a8e8 e2d1 e6g4 e1e8 f8e8 d2f1 c6a5 f3d2 g4d1 c1d1 a7a6 f1e3 b7b5 d1f3 a5b7 a1e1');
INSERT INTO evaluations VALUES(54503,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(54503,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(54503,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(54503,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(54503,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(54503,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(54503,36,'cp -34','c4e2 b8c6 g1f3 f8d6 d2d4 c5c4 b1c3 g8f6 c1g5 c8e6 e1g1 h7h6 g5h4 e8g8 f1e1 f8e8 d1d2 d6e7 a1d1 d8b6 d2c1 f6e4 h4e7 e8e7 c3e4 d5e4 f3e5 c6e5 d4e5 a8c8 h2h3 b6b5 f2f4 e4f3 e2f3 e6f5 g1h1 e7e5 e1e5 b5e5');
INSERT INTO evaluations VALUES(22853,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(22853,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(22853,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(22853,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(22853,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(22853,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(22853,36,'cp 22','d2d4 e5d4 e3d4 d7d5 g1f3 g8f6 f1b5 f8b4 b5c6 b7c6 e1g1 e8g8 f3e5 c8a6 f1e1 f8e8 c1g5 d8d6 g5f6 d6f6 a2a3 b4c3 b2c3 f6d6 d1h5 e8e7 a3a4 a8e8 e1e3 f7f6 e5g4 e7e3 g4e3 g7g6 h5h6 a6c8 h2h3 c8e6 a4a5 e6f7 a5a6 c6c5');
INSERT INTO evaluations VALUES(22753,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(22753,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(22753,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(22753,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(22753,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(22753,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(22753,36,'cp 16','e5d4 e3d4 d7d5 g1f3 g8f6 f1b5 f8b4 e1g1 e8g8 b5c6 b7c6 f3e5 b4c3 b2c3 c8f5 f1e1 f8e8 c1g5 a7a5 d1f3 f5e4 f3f4 h7h6 g5h4 e8e6 f4d2 e4f5 f2f3 d8d6 a2a4 a8e8 h4g3 c6c5 e5d3 e6e1 a1e1 e8e1 d2e1');
INSERT INTO evaluations VALUES(22935,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(22935,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(22935,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(22935,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(22935,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(22935,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(22935,36,'cp 11','e3d4 d7d5 g1f3 g8f6 f1b5 f8b4 e1g1 e8g8 b5c6 b7c6 f3e5 c8a6 f1e1 f8e8 c1f4 e8e6 f4g5 h7h6 g5h4 d8d6 h4f6 e6f6 a2a3 b4c3 b2c3 c6c5 e1e3 f6f4 e3f3 f4f3 d1f3 f7f6 e5g4 a8e8 g4e3 c7c6 f3h5 d6d7 h2h3 c5d4 c3d4 e8e4');
INSERT INTO evaluations VALUES(30726,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(30726,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(30726,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(30726,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(30726,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(30726,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(30726,36,'cp 113','h2h4 h7h5 c2c4 f8g7 b1c3 d7d6 g1f3 b8d7 f3d4 d7e5 f1e2 g8h6 c1g5 c7c5 d4f3 e5g4 d1c2 e8g8 e1g1 f7f6 g5d2 b7c8 a1e1 c8d7 e2d3 g4e5 f3e5 f6e5 f2f3 h6f7 f1f2 g8h7 g2g3 a7a6 b2b3 g7h6 d2h6 h7h6 c2d2 h6g7 f2g2 g7h7 d3c2');
INSERT INTO evaluations VALUES(39416,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(39416,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(39416,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(39416,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(39416,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(39416,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(39416,36,'cp -114','g8f6 g1f3 e8g8 e1g1 d7d6 d1c2 b8c6 d2b3 c6e5 a2a3 e5c4 c2c4 b4c5 b3c5 d6c5 c4c5 f6e4 c5e5 d8f6 f1e1 f6e5 b2e5 e4c5 e5c7 c5e6 c7d6 f8d8 a1d1 f7f6 d6b4 d8d7 d1d7 c8d7 e1d1 d7e8 d1d6 e6f4 h2h4 g8f7 g1h2 a8c8 b4d2 f4e6 f3d4 e6c5 f2f3 b7b6 d4f5 c5e6 f5d4 c8d8 d6d8');
INSERT INTO evaluations VALUES(42830,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(42830,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(42830,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(42830,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(42830,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(42830,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(42830,36,'cp -109','g1f3 e8g8 e1g1 d7d6 d1c2 b8c6 d2b3 c6e5 c4e2 e5f3 e2f3 c7c6 a1d1 d6d5 a2a3 b4e7 f1e1 f8e8 b2f6 e7f6 e4d5 e8e1 d1e1 c6d5 f3d5 c8d7 c2c4 a8c8 d5f7 g8h8 c4e4 d7c6 e4g4 d8d7 g4d7 c6d7 f7d5 b7b6 g2g3 h7h6 b3d2 d7f5 d2c4');
INSERT INTO evaluations VALUES(42828,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(42828,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(42828,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(42828,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(42828,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(42828,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(42828,36,'cp -93','e8g8 e1g1 d7d6 d1c2 b8c6 d2b3 c6e5 c4e2 c7c6 a1d1 e5f3 e2f3 d6d5 a2a3 b4e7 f1e1 c8e6 b3d4 e6d7 e4d5 f6d5 d4f5 d7f5 c2f5 g7g6 f5d3 e7f6 b2f6 d5f6 d3c3 f6d5 c3b3 d8c7 f3d5 c6d5 d1d5 b7b6 g2g3 f8e8 e1e8 a8e8 h2h4 h7h5 b3f3 g8g7');
INSERT INTO evaluations VALUES(39884,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(39884,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(39884,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(39884,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(39884,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(39884,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(39884,36,'cp 31','g1f3 d7d6 d1c2 b4c3 c2c3 a7a5 d2d3 e8g8 e1g1 h7h6 b2b3 c6c5 h2h3 b8c6 a2a4 f8e8 c1b2 c6d4 a1e1 c8d7 f3d2 d7c6 e2e3 c6g2 g1g2 d4c6 d2e4 f6e4 d3e4 d8d7 e1d1 a8a6 f2f3 a6b6 g3g4 b6b4 c3d3');
INSERT INTO evaluations VALUES(39883,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(39883,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(39883,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(39883,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(39883,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(39883,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(39883,36,'cp 22','d7d6 d1c2 e8g8 e1g1 b4c3 c2c3 a7a5 d2d3 h7h6 h2h3 c6c5 b2b3 b8c6 a2a4 a8b8 g1h2 b7b6 e2e3 f8e8 e3e4 f6d7 c1b2 c6d4 f3d4 c5d4 c3c2 d7c5 f2f4 c8d7 c2d1 f7f6 f4f5 g8f8 b2c1 d7c6 g2f3 b8a8 f3h5');
INSERT INTO evaluations VALUES(42933,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(42933,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(42933,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(42933,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(42933,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(42933,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(42933,36,'cp 53','g2g3 e8g8 f1g2 d7d5 e1g1 b8d7 d1c2 c7c6 f3g5 d7b8 c1d2 c8d7 g5f3 d7e8 c2b3 d8b6 c4c5 b6b3 a2b3 f6e4 c3e4 d5e4 f3e5 b8d7 e5d7 e8d7 d2f4 g8f7 b3b4 e7f6 f4e5 a7a6 f2f3 e4f3 e2f3 f8c8 f1d1 f6e5 d4e5');
INSERT INTO evaluations VALUES(47665,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(47665,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(47665,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(47665,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(47665,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(47665,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(47665,36,'cp 40','g5f6 d8f6 a2a3 g7g6 e2e4 f6e7 d1e2 c7c6 e2e3 b8d7 f1d3 h6h5 e1g1 f8h6 e3e2 d5e4 c3e4 e8g8 b2b4 a7a5 a1b1 a5b4 a3b4 e6e5 e4c5 d7c5 b4c5 c8g4 e2e5 e7d7 e5g3 g4f3 g3f3 d7d4 b1b7 d4c5 f1d1 h6g7 g2g3 a8b8 b7c7 b8d8 f3c6 c5c6');
INSERT INTO evaluations VALUES(47660,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47660,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(47660,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(47660,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(47660,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47660,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(47660,36,'cp 24','d8f6 a2a3 f6e7 e2e4 g7g6 d1e2 d5e4 e2e4 f8g7 e1c1 b8d7 f1c4 e8g8 h1e1 a8b8 e4f4 c7c6 c3e4 b7b5 c4f1 c8b7 f4d6 e7d6 e4d6 b7a8 c2c3 b8d8 d6e4 a8b7 g2g3 b7a8 f1g2 a7a5 h2h4 a5a4 g3g4 c6c5 e4c5 d7c5 d4c5');
INSERT INTO evaluations VALUES(33859,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(33859,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(33859,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(33859,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(33859,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(33859,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(33859,36,'cp 15','a2a3 g7g6 e2e4 f6e7 d1e2 d5e4 e2e4 f8g7 e1c1 b8d7 f1c4 e8g8 h1e1 a8b8 e4f4 c7c6 c4f1 b7b5 c3e4 c8b7 f4d6 e7d6 e4d6 b7a8 f3e5 d7e5 d4e5 c6c5 f2f4 c5c4 d1d2 f8d8 g2g3 g7f8 e1d1 f8e7 f1g2 a8g2 d2g2 e7d6 e5d6 b8b6 g2d2 g8f8 d2d4');
INSERT INTO evaluations VALUES(47905,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(47905,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(47905,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(47905,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(47905,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47905,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(47905,36,'cp 30','g1f3 e5f4 c1f4 f8e7 f1e2 c7c5 c2c4 b8c6 b1c3 e8g8 e1g1 f6g4 c3d5 g4e5 f3e5 d6e5 f4e3 e7d6 d1d2 c8e6 f1f2 c6d4 a1f1 f7f6 h2h3 d8d7 e2h5 a8d8 d2d1 f6f5 e4f5 d4f5 h5g4 f5e3');
INSERT INTO evaluations VALUES(47904,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47904,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(47904,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47904,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(47904,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(47904,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(47904,36,'cp 29','e5f4 c1f4 f8e7 f1e2 c7c5 c2c4 e8g8 e1g1 b8c6 b1c3 f6g4 c3d5 g4e5 f3e5 d6e5 f4e3 c8e6 e2g4 b7b6 g1h1 c6d4 e3d4 c5d4 g4e6 f7e6 d5e7 d8e7 d1h5 e7d6 h2h3 f8f4 f1f4 e5f4 e4e5 d6c5 a2a3 a8f8 b2b4 c5c7 a1f1 h7h6');
INSERT INTO evaluations VALUES(48712,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(48712,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(48712,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(48712,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(48712,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(48712,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(48712,36,'cp -90','d2d4 f8e7 f3e5 e8g8 c1e3 c7c5 d4c5 b8d7 b1c3 d7e5 f4e5 f6g4 e3d4 g4e5 f1e2 e5c6 e1g1 c8e6 d4f2 d5d4 c3e4 b7b6 c5b6 d8b6 b2b3 f8d8 e4d2 a8c8 d2c4 b6c5 e2d3 c6b4 f2e1 b4d3 c2d3 c5b5 e1d2 b5d5 d1e2 e7c5 e2f3 f7f6 f3d5 d8d5');
INSERT INTO evaluations VALUES(48650,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(48650,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(48650,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(48650,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(48650,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(48650,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(48650,36,'cp -88','f8e7 c1e3 e8g8 f3e5 f8e8 f1e2 c7c5 d4c5 d8c7 b2b4 b8c6 b1c3 c6b4 c3b5 c7b8 a1b1 b4a2 b1b2 a7a6 b2a2 a6b5 a2a8 b8a8 e1g1 e8d8 e3d4 f6e4 e2h5 f7f6 e5d3 c8f5 c2c3 a8a4 f1e1 a4d1 h5d1 d8a8 g2g4 f5g6 f4f5 g6f7 d1f3 a8a4 f3e4 d5e4 e1e4');
INSERT INTO evaluations VALUES(41083,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(41083,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(41083,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(41083,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(41083,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(41083,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(41083,36,'cp -75','f3e5 e8g8 c1e3 c7c5 d4c5 b8d7 b1c3 d7e5 f4e5 f6g4 e3d4 g4e5 f1e2 e5c6 e1g1 c8e6 d4f2 d5d4 c3e4 f7f5 e4d6 e7d6 c5d6 d8d6 c2c3 a8d8 f1e1 e6d5 f2d4 d6c7 e2f3 d5f3 d1f3 c6d4 c3d4 d8d4 f3b3 g8h8 b3a3 f8d8 a3a7 h7h6 a1c1 c7f7 a7b6');
INSERT INTO evaluations VALUES(49450,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(49450,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(49450,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49450,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(49450,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(49450,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49450,36,'cp 0','c7c5 e4e5 f6d5 c2c4 d5e7 b1c3 b8c6 c1g5 d8b6 d1b3 d7d5 b3b6 a7b6 f1e2 d5d4 c3b5 e8d7 b5d6 f7f5 d6f7 h8g8 h2h4 d7e8 f7d6 e8d7');
INSERT INTO evaluations VALUES(52714,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(52714,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(52714,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(52714,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(52714,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52714,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52714,36,'cp 21','c6d5 b1c3 g8f6 d2d4 b8c6 c1f4 c8f5 e2e3 e7e6 f1b5 f6d7 e1g1 a8c8 d1b3 f8e7 h2h3 e8g8 f1c1 d7b6 b5c6 b7c6 c3a4 f7f6 a4c5 b6c4 e3e4 d5e4 f3h4 c4a5 b3c3 f8f7 g1h1 g8h8 b2b4 e7c5');
INSERT INTO evaluations VALUES(53425,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(53425,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(53425,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53425,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(53425,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(53425,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(53425,36,'cp 28','d2d4 g8f6 b1c3 b8c6 c1f4 c8f5 d1b3 c6a5 b3a4 f5d7 a4c2 a8c8 e2e3 e7e6 f1d3 b7b5 a2a3 f8e7 e1g1 e8g8 h2h3 d7e8 f1c1 a5c4 c2e2 d8b6 c1c2 a7a6 a1c1 h7h6 g2g4 b6b7 g4g5 h6g5 f3g5 b5b4 c3d1 b4b3');
INSERT INTO evaluations VALUES(820,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(820,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(820,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(820,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(820,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(820,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(820,36,'cp 46','c1h6 g7h6 d1d2 c8d7 e1g1 f8g7 b1a3 e8g8 a3c2 c5d4 c3d4 a7a5 a1b1 f7f6 e5f6 f8f6 b2b4 a5b4 c2b4 b6a5 a2a3 a5a3 b4d3 d7e8 b1b7 f6f7 b7f7 e8f7 d2f4 f7g6 d3e5 c6e5 f3e5 a8f8 f4d2 a3d6 d2c3 g7e5 d4e5 d6b6 c3d2 h6h5 h2h4 g8h8');
INSERT INTO evaluations VALUES(815,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(815,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(815,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(815,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(815,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(815,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(815,36,'cp 37','g7h6 d1d2 a7a5 e1g1 f8g7 b1a3 e8g8 a3c2 c5d4 c3d4 c8d7 a1b1 f7f6 e5f6 f8f6 b2b4 a5b4 c2b4 b6a5 f1c1 c6b4 b1b4 g7f8 b4b7 a5d2 f3d2 a8a2 d2f3 f8a3 b7a7 d7e8 e2d3 f6f7 a7a6 a3c1 a6a2 c1f4 g1f1 g8f8 d3e2 f8e7 g2g3 f4c7 a2a7');
INSERT INTO evaluations VALUES(18264,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18264,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18264,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18264,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(18264,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(18264,36,'cp 0','d2d4 d5e4 f3e5 g8e7 c2c3 f7f6 e5c4 c8a6 b1a3 d8d5 d1a4 a6b5 a4b3 b5a6');
INSERT INTO evaluations VALUES(20039,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(20039,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(20039,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(20039,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(20039,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(20039,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20039,36,'cp 0','g8f6 d2d3 e7e6 a2a4 f8e7 f1e2 a7a6 e1g1 e8g8 d1e1 c8d7 b1d2 d8c7 f3h4 c7d8');
INSERT INTO evaluations VALUES(13340,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13340,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13340,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13340,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13340,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13340,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13340,36,'cp 0','d2d3 e7e6 f1e2 f8e7 e1g1 e8g8 a2a4 a7a6 d1e1 c8d7 b1d2 d8c7 f3h4 c7d8');
INSERT INTO evaluations VALUES(13341,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13341,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13341,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13341,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13341,36,'cp 0','e7e6 f1e2 f8e7 a2a4 a7a6 e1g1 e8g8 d1e1 c8d7 b1d2 d8c7 f3h4 c7d8');
INSERT INTO evaluations VALUES(30162,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30162,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30162,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30162,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30162,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30162,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30162,36,'cp 0','b8d7 e1g1 c7c5 b2b3 a8c8 c2c4 f8e7 f1c1 e8g8 b1c3 d7b6 c3e2 b6d7');
INSERT INTO evaluations VALUES(33226,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(33226,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(33226,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(33226,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(33226,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(33226,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(33226,36,'cp 97','e2e3 c5b6 b1c3 g8e7 b2b4 d7d6 a2a4 a7a5 b4b5 b8d7 g1e2 f6e6 d2d3 b6c5 e1g1 e8g8 d1b3 c5b4 c1a3 d7c5 b3d1 c7c6 d3d4 b4a3 a1a3 c5d7 d1b3 e6h6 c3e4 g8h8 d4e5 d7e5 e2d4 c8g4 b3c2 d6d5 e4d2 a8d8 c4d5 c6d5');
INSERT INTO evaluations VALUES(33225,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(33225,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(33225,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(33225,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(33225,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(33225,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(33225,36,'cp 87','d7d6 b1c3 b8c6 c3d5 f6d8 g1e2 g8f6 d2d4 c5b6 d5b6 a7b6 d4d5 c6e7 e2c3 e8g8 a2a4 f6d7 b2b3 f7f5 f2f4 e7g6 c1b2 d8e7 e1g1 d7f6 d1d2 h7h5 a1e1 e5e4 c3b5 h5h4 d2e2 h4g3 h2g3 f6g4 b2d4');
INSERT INTO evaluations VALUES(39264,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39264,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39264,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39264,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39264,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39264,36,'cp 0','f1g2 g8f6 c2c3 d7d6 g1e2 e8g8 d2d4 c5b6 e1g1 f8e8 c1g5 h7h6 g5f6 d8f6 b1a3 e5d4 e2d4 b8c6 d4c6 b7c6 a3c4 b6c5 d1a4 c8d7 b2b4 c5b6 a4c2 a7a5 a2a3 e8d8 f1e1 d7e6 c4b6 c7b6 b4b5 e6d7 b5c6 d7c6 c2b3 a8c8 c3c4 c6d7 b3b6 d7e6 a1d1 e6c4');
INSERT INTO evaluations VALUES(39265,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39265,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(39265,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(39265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39265,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(39265,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(39265,36,'cp -9','g8f6 b1c3 e8g8 g1e2 c7c6 e1g1 d7d5 e4d5 c6d5 d2d4 e5d4 e2d4 c8g4 d4e2 d8d7 c1e3 c5e3 f2e3 b8c6 d1d2 g4e2 c3e2 a8d8 a1d1 d7c7 e2c3 d5d4 c3b5 c7b6 b5d4 c6d4 e3d4 b6b2 d4d5 h7h6 d5d6 b2a2');
INSERT INTO evaluations VALUES(49931,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(49931,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(49931,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(49931,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(49931,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(49931,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(49931,36,'cp 36','b7b6 f1e2 c8b7 f2f3 b8c6 g1h3 g7g6 e1g1 f8g7 a2a3 e8g8 b2b4 d8e7 d1b3 d7d6 c1d2 c6d8 a1e1 a7a5 d4d5 g7h6 d2c1 c7c6 e3e4 h6c1 e1c1 f5e4 f3e4 a5b4 a3b4 e6d5 c4d5 c6d5 h3f4 e7e5 f4d5 f6d5 c3d5 b7d5 b3d5 e5d5 e4d5');
INSERT INTO evaluations VALUES(56430,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(56430,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(56430,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(56430,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(56430,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(56430,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(56430,36,'cp 112','b2e5 c7c5 e2e3 b8c6 e5b2 g8f6 g1f3 f8d6 f1e2 e8g8 e1g1 d8e7 d2d3 c8f5 b1d2 a8d8 a2a3 f8e8 f1e1 d6c7 b3b4 c5b4 a3b4 e7b4 b2f6 g7f6 g2g3 c7a5 d2b3 a5b6 f3h4 f5g6 d3d4 b4c3 h4g6 h7g6 d1d3 c3d3 e2d3 c6b4 e1c1 d8c8 d3b5 e8d8');
INSERT INTO evaluations VALUES(56300,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(56300,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(56300,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(56300,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(56300,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(56300,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(56300,36,'cp 115','c7c5 e5b2 g8f6 e2e3 b8c6 g1f3 f8d6 g2g3 c8g4 f1g2 e8g8 e1g1 f8e8 h2h3 g4h5 d2d3 d5d4 e3e4 d6c7 b1d2 f6d7 d1c1 g7g5 f3h2 f7f6 g2f3 h5f7 c1d1 g8h8 f3h5 d8e7 h5f7 e7f7 g1g2 a8d8 a2a4 e8g8 b2c1 g8g7 h2g4 f7g6 d2c4 h7h5');
INSERT INTO evaluations VALUES(57726,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(57726,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(57726,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(57726,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(57726,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(57726,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(57726,36,'cp 61','e2e4 e7e6 c2c3 g8f6 f1d3 c7c5 d3c2 b8c6 e1g1 b7b5 e4e5 f6d5 c2e4 f8e7 d2d4 c5c4 b1d2 a7a5 f1e1 a5a4 d2f1 a4a3 e4d5 a3b2 c1b2 e6d5 f1e3 e8g8 e3d5 d7d6 a2a4 d6e5 d5e7 c6e7 f3e5 a8a4 a1a4 b5a4 e5c4');
INSERT INTO evaluations VALUES(57724,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(57724,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(57724,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(57724,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(57724,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(57724,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(57724,36,'cp 59','e7e6 d2d4 d7d5 e4e5 c7c5 c2c3 b8c6 f1d3 d8c7 e1g1 c5d4 c3d4 c6b4 d3b5 c8d7 b5d7 c7d7 b1c3 g8e7 c1d2 b4c6 c3e2 e7g6 h2h4 f8e7 g2g3 e7d8 g1g2 e8g8 h4h5 g6e7 a1c1 a8c8 f1e1 e7f5 c1c2 d8b6 d2c3 b6a5 d1d2 a5c3 c2c3 c6e7 e2f4 c8c3 d2c3');
INSERT INTO evaluations VALUES(19069,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19069,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19069,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19069,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19069,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19069,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19069,36,'cp 0','g8f6 g1e2 f8c5 e2g3 e8g8 f1d3 a7a5 a2a3 d7d6 h2h3 c8e6 e1g1 a8c8 a1b1 h7h6 c1e3 c5e3 f2e3 c6b8 d1d2 b8d7 b1d1 d8b6 c3a4 b6c6 a4c3');
INSERT INTO evaluations VALUES(29754,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(29754,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(29754,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(29754,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(29754,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(29754,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(29754,36,'cp 54','f1c4 e7e6 b1c3 c7c5 h2h3 g4h5 g2g4 h5g6 h3h4 c5d4 e3d4 f6g4 c1g5 d8b6 d1e2 f8b4 d4d5 e8g8 d5e6 g6h5 e1g1 b4c3 b2c3 b6c7 e2e4 g4f6 g5f6 g7f6 e6e7 f8e8 e4d5 e8e7 f1e1 h5f3 d5f3 b8c6 f3f6 c6e5 e1e3');
INSERT INTO evaluations VALUES(31252,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(31252,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(31252,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(31252,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(31252,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(31252,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(31252,36,'cp 187','d4c5 d8a5 b1c3 a5c5 h2h3 g4d7 c1e3 c5c7 f3d4 g8f6 d1d2 e7e6 g2g4 h7h6 e1c1 a7a6 f2f4 b8c6 c1b1 b7b5 f1d3 c6d4 e3d4 b5b4 c3e2 e6e5 d4e3 a6a5 e2g3 a8c8 d3a6 c8a8 a6f1 c7c6 g4g5 h6g5 f4g5 f6e4');
INSERT INTO evaluations VALUES(31239,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(31239,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(31239,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(31239,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(31239,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(31239,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(31239,36,'cp 190','d8a5 b1c3 a5c5 h2h3 g4d7 c1e3 c5c7 f3d4 g8f6 d1d2 e7e6 e1c1 a7a6 g2g4 h7h6 f2f4 b8c6 f1e2 c6d4 d2d4 e6e5 d4d3 e5f4 e3f4 d7e6 c1b1 f8e7 h3h4 f6d7 g4g5 h6g5 h4g5 h8h1 d1h1 d7e5 d3g3 e8c8 h1h7 d8g8 c3d5 e6d5 e4d5 c8b8 a2a3');
INSERT INTO evaluations VALUES(33380,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(33380,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(33380,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(33380,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(33380,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(33380,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(33380,36,'cp -9','b8c6 a2a3 h7h6 b1c3 a7a6 c1e3 e8g8 e1g1 c8e6 h2h3 f8d8 d3d4 e7f8 f1e1 b7b5 e2d3 a8b8 e3d2 e6c4 d3c4 b5c4 a1b1 c6e7 g2g4 e7d5 c3d5 f6d5 b2b3 c4b3 b1b3 b8b3 c2b3');
INSERT INTO evaluations VALUES(37861,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(37861,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(37861,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(37861,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(37861,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(37861,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(37861,36,'cp -14','b8c6 f1d3 g8f6 e1g1 d7d5 d3b5 c8d7 e4d5 f6d5 d4c6 b7c6 b5c4 d7f5 d1a4 d8d7 b1a3 e8g8 f1e1 a7a5 c4f1 d7c7 a3c4 h7h5 c1g5 g7f6 g5f6 e7f6 a1d1 g8g7 d1d2 f8d8 h2h3 f5e6 a4c2 a5a4 g1h1 h5h4 e1d1 d8b8 c4a3 d5f4 c3c4');
INSERT INTO evaluations VALUES(43870,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(43870,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43870,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(43870,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(43870,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43870,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(43870,36,'cp 18','e7e6 g2g3 c8a6 b2b3 f8b4 c1d2 b4e7 f1g2 e8g8 e1g1 c7c6 d2c3 d7d5 b1d2 b8d7 f1e1 a6b7 e2e4 d5e4 d2e4 c6c5 f3g5 c5d4 c3d4 d8c7 e4f6 e7f6 g2b7 c7b7 g5e4 f6e5 d1f3 b7c7 a1d1 e5d4 d1d4 d7e5 f3f4 f7f6 e1d1 f8d8 g1g2 d8d4');
INSERT INTO evaluations VALUES(49043,30,'cp 391',NULL);
INSERT INTO evaluations VALUES(49043,31,'cp 400',NULL);
INSERT INTO evaluations VALUES(49043,32,'cp 394',NULL);
INSERT INTO evaluations VALUES(49043,33,'cp 386',NULL);
INSERT INTO evaluations VALUES(49043,34,'cp 375',NULL);
INSERT INTO evaluations VALUES(49043,35,'cp 400',NULL);
INSERT INTO evaluations VALUES(49043,36,'cp 405','d2d3 f7f5 d3e4 f5e4 f3d4 f8e7 c2c4 c7c5 d4b3 d5d4 e2e4 e8g8 e4d5 d8d5 c4d5 a7a6 a2a4 b8d7 f1c4 d7b6 d5d6 b6c4 d6e7 f8e8 e1g1 b7b6 a4a5 c4a5 b3a5 b6a5 a1a5 e8e7 a5c5 c8d7 c5c7 a8e8 c1g5 e7f7 c7c1 h7h6 g5h4 d7b5 f1e1 e8e2 e1e2 b5e2 h4g3');
INSERT INTO evaluations VALUES(49041,30,'cp 381',NULL);
INSERT INTO evaluations VALUES(49041,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(49041,32,'cp 401',NULL);
INSERT INTO evaluations VALUES(49041,33,'cp 398',NULL);
INSERT INTO evaluations VALUES(49041,34,'cp 390',NULL);
INSERT INTO evaluations VALUES(49041,35,'cp 397',NULL);
INSERT INTO evaluations VALUES(49041,36,'cp 386','f7f5 d3e4 f5e4 f3d4 f8e7 c2c4 c7c5 d4b3 d5d4 e2e4 e8g8 e4d5 d8d5 c4d5 a7a6 a2a4 b8d7 f1c4 d7b6 d5d6 b6c4 d6e7 f8e8 e1g1 e8e7 b3c5 c8f5 b1a3 c4a3 a1a3 a8c8 c5d3 e7e2 c1f4 c8f8 f4d6 f8c8 a3b3 c8c2 a4a5');
INSERT INTO evaluations VALUES(4034,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(4034,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(4034,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(4034,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(4034,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(4034,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(4034,36,'cp 67','e1g1 e7e5 e2e4 e5d4 f3d4 d7c5 h2h3 f8e8 f1e1 a7a6 b2b3 c8d7 c1a3 a8b8 a1c1 b7b5 e4e5 d6e5 a3c5 e5d4 e1e8 d8e8 c5d4 c7c6 d1f3 e8e7 c1d1 b8d8 f3f4 f6h5 f4c7 g7d4 d1d4 h5f6 d4d1 d8e8 c7d6 e7e5 d6e5 e8e5 c3e4 f6e4 d1d7');
INSERT INTO evaluations VALUES(11975,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(11975,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(11975,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(11975,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(11975,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(11975,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(11975,36,'cp -10','c4d3 d7d5 d1e2 f8e7 e1g1 e8g8 c2c4 c8a6 b1d2 f6d7 c4d5 a6d3 e2d3 c6d5 e4d5 d7c5 d3e2 d8d5 d2c4 d5d3 e2d3 c5d3 f1d1 f8d8 c1e3 f7f5 g1f1 e6e5 f2f3 e7f6 a1b1 a8c8 b2b3 e5e4 f3e4 f5e4 e3a7');
INSERT INTO evaluations VALUES(21278,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(21278,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(21278,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(21278,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(21278,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(21278,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(21278,36,'cp 36','g8f6 d2d4 f8e7 f1e1 c8g4 d4d5 c6b8 b1d2 c7c6 h2h3 c6d5 e4d5 g4f5 d2f1 e8g8 f1g3 f5h7 c4d3 h7d3 d1d3 b8d7 c1e3 a8c8 c3c4 d7c5 d3e2 b7b5 a1c1 b5c4 c1c4 f8e8 e1c1 d8d7 e2d2 e7f8 b2b4 c5a4 c4c8 e8c8 c1c8');
INSERT INTO evaluations VALUES(14618,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(14618,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14618,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14618,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14618,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(14618,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14618,36,'cp 46','d2d4 f8e7 d4d5 c6b8 c4d3 c8g4 h2h3 g4h5 c1e3 h5g6 d1b3 b8d7 b1d2 d8c8 c3c4 e8g8 a2a3 b7b6 b3d1 c7c6 a1c1 d7c5 e3c5 b6c5 b2b4 c5b4 a3b4 c6d5 c4d5 c8b7 d1c2 a7a5 b4b5 a5a4 c2c6 b7d7 c6d7 f6d7 c1c7');
INSERT INTO evaluations VALUES(14646,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(14646,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(14646,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14646,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(14646,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14646,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14646,36,'cp 37','f8e7 f1e1 c8g4 d4d5 c6b8 b1d2 c7c6 d5c6 b8c6 d2f1 g4e6 c4d5 e8g8 b2b3 d8d7 c3c4 e7d8 f1g3 a8c8 h2h3 d8b6 c1e3 b6e3 e1e3 c8c7 d5e6 f7e6 e3d3 f8d8 d1e2 a7a6 a1d1 d7e7 e2d2 c7d7 a2a3 g8h8 d2c3');
INSERT INTO evaluations VALUES(42031,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(42031,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(42031,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(42031,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(42031,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(42031,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(42031,36,'cp 179','e4e5 f6g8 d1g4 b4f8 g1f3 h7h5 g4g3 h5h4 g3h3 b7b6 c1g5 d8d7 g5h4 c8a6 d3a6 b8a6 a2a3 c7c5 e1g1 g8e7 h3g4 e7f5 h4g5 a6c7 a1d1 c5c4 g4f4 f8e7 g2g4 f7f6 e5f6 g7f6 g4f5 f6g5 f3g5 e6f5');
INSERT INTO evaluations VALUES(42124,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(42124,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(42124,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(42124,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(42124,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(42124,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(42124,36,'cp 182','f6g8 a2a3 b4c3 b2c3 d8d7 d1g4 f7f5 g4e2 d7f7 h2h4 c7c5 d4c5 g8e7 a3a4 h7h6 g1f3 e8g8 f3d4 b8d7 c1a3 f5f4 c3c4 e7c6 d4f3 f7h5 e1c1 d7e5 f3e5 h5e5 a3b2 e5e2 d3e2 d5d4 b2d4 c6d4 d1d4 e6e5 d4d5 e5e4 d5e5 c8d7 e5e4 d7c6 e4e6');
INSERT INTO evaluations VALUES(42694,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(42694,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42694,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42694,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(42694,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42694,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(42694,36,'cp -24','f6d5 e4d5 e5e4 f3d4 e8g8 c2c3 b4c5 d2d3 c5d4 c3d4 d8e7 f1e2 e4d3 d1d3 b7b6 c1e3 c7c6 d5c6 c8a6 d3a3 e7a3 b2a3 a6e2 e1e2 b8c6 h1c1 f7f6 a1b1 f8e8 e2d3 e8e6 e3f4 g7g5 d4d5 c6e5 d3d4 e6e8 f4e5 e8e5 c1c7 e5e2 c7d7 e2f2');
INSERT INTO evaluations VALUES(42889,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(42889,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(42889,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(42889,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(42889,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(42889,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42889,36,'cp -21','e4d5 e5e4 f3d4 e8g8 c2c3 b4c5 d2d3 c5d4 c3d4 d8e7 f1e2 e4d3 d1d3 b7b6 d3e3 e7d6 e3g3 c8b7 g3d6 c7d6 f2f3 b7d5 e1f2 b8c6 h1d1 f8e8 c1f4 e8e6 e2d3 h7h6 h2h4 c6b4 d3b5 d5c6 b5f1 a7a5 a2a3 b4d5 f4h2 a5a4');
INSERT INTO evaluations VALUES(42888,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(42888,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(42888,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(42888,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(42888,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(42888,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(42888,36,'cp -25','e5e4 f3d4 e8g8 c2c3 b4c5 d2d3 c5d4 c3d4 d8e7 f1e2 e4d3 d1d3 b7b6 d3e3 e7d6 e3g3 c8b7 g3d6 c7d6 f2f3 f8e8 e1f2 b7d5 h1d1 b8c6 c1f4 e8e6 e2d3 a7a6 a2a3 b6b5 a1c1 h7h6 h2h4 e6f6 f4e3 a8e8 d1e1 e8b8 h4h5 g8f8 e3d2 d5c4 d3c4');
INSERT INTO evaluations VALUES(53819,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(53819,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(53819,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(53819,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(53819,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(53819,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(53819,36,'cp 103','d2d4 e5d4 f3d4 g8f6 e2e4 f8e7 f1e2 e8g8 c1f4 f8e8 d1c2 e7f8 a1d1 b8d7 d4f5 d7e5 f5d6 f8d6 c4c5 f6d5 c3d5 c6d5 c5d6 d8d6 e1g1 d5d4 c2b3 d6d8 f2f3 e5c6 e2c4 d8f6 f4g3 c8e6 c4e6');
INSERT INTO evaluations VALUES(53824,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(53824,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(53824,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(53824,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(53824,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(53824,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(53824,36,'cp 100','e5d4 f3d4 f8e7 e2e4 g8f6 f1e2 e8g8 c1f4 f8e8 e1g1 e7f8 d1c2 a7a5 a1d1 b8a6 f2f3 a6c5 f4e3 c8d7 d4b3 c5b3 c2b3 d8e7 e3f2 d7e6 b3c2 g7g6 f1e1 a8d8 e2f1 f6d7 f2g3 d7c5 c2d2 a5a4 d2c2');
INSERT INTO evaluations VALUES(53840,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(53840,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(53840,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(53840,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(53840,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(53840,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(53840,36,'cp 112','f3d4 f8e7 e2e4 g8f6 c1f4 e8g8 f1e2 f8e8 e1g1 e7f8 d1c2 b8d7 a1d1 d7c5 f2f3 a7a5 g1h1 f6h5 f4e3 d8h4 e3f2 h4e7 b2b3 g7g6 f1e1 f8g7 e2f1 g7f6 g2g3 h5g7 c2d2 h7h5 h1g1 h5h4 f2e3 h4g3 h2g3');
INSERT INTO evaluations VALUES(9963,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(9963,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(9963,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(9963,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(9963,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(9963,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(9963,36,'cp 88','b2c3 b7c6 f1d3 e8g8 e1g1 f8e8 f1e1 d7d6 c1g5 d8e7 h2h3 e7e6 f2f4 f6d7 g5h4 d7c5 h4f2 c5d3 c2d3 c6c5 f4f5 e6f6 c3c4 c8d7 d1d2 a8b8 a1b1 d7c6 b1b8 e8b8 d2a5 f6d8 a5a6 d8d7 f2g3 f7f6 g1h2 b8a8 a6a5');
INSERT INTO evaluations VALUES(9962,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(9962,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(9962,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(9962,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(9962,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(9962,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(9962,36,'cp 59','b7c6 f1d3 e8g8 e1g1 f8e8 f1e1 d7d6 c1g5 d8e7 h2h3 e7e6 g5h4 f6d7 f2f4 a8b8 d1d2 f7f6 h4f2 c6c5 a2a4 c8b7 c3c4 b7a6 d2a5 b8b6 a1b1 a6c4 a5a7 c4d3 c2d3 e8b8 b1b6 c7b6 a7c7 b8a8 f4f5 e6f7 c7d6 a8a4 d6c6');
INSERT INTO evaluations VALUES(7620,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(7620,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(7620,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(7620,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7620,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(7620,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(7620,36,'cp 59','f1d3 e8g8 e1g1 f8e8 f1e1 d7d6 h2h3 c6c5 e4e5 d6e5 e1e5 c8b7 e5e8 d8e8 c3c4 e8e5 a1b1 b7c6 d1d2 a8e8 c1b2 e5e7 d2g5 h7h6 g5f5 c6e4 b2f6 e4f5 f6e7 f5d3 c2d3 e8e7 b1b7 e7d7 b7a7 f7f6 a7a3 g8f7 g1h2 d7d6 f2f3');
INSERT INTO evaluations VALUES(7621,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(7621,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(7621,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(7621,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(7621,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(7621,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(7621,36,'cp 74','e8g8 e1g1 f8e8 f1e1 d7d6 h2h3 c6c5 e4e5 d6e5 e1e5 c8b7 d1e2 d8d6 c1f4 d6c6 f2f3 e8e5 e2e5 c5c4 d3f1 f6d5 f4g3 a8e8 e5d4 c6b6 g3f2 b6d4 f2d4 d5b6 g1f2 b7c6 h3h4 c6a4 a1c1 e8c8 g2g4 h7h6 g4g5 c7c5 d4e5 h6g5 h4g5 a4d7 c1d1 d7f5 d1d6 c8e8 f3f4 f5c2');
INSERT INTO evaluations VALUES(2881,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(2881,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(2881,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(2881,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(2881,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(2881,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(2881,36,'cp 76','e1g1 f8e8 f1e1 d7d6 c1g5 d8e7 h2h3 e7e6 f2f4 f6d7 g5h4 a8b8 d1f3 d7c5 h4f2 c5d3 c2d3 b8b2 f4f5 e6e7 a1b1 b2b1 e1b1 c6c5 c3c4 a7a6 f2g3 f7f6 b1b3 c8d7 b3a3 d7c8 f3f2 e7d7 a3b3 h7h6 b3b8');
INSERT INTO evaluations VALUES(2880,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(2880,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(2880,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(2880,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(2880,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(2880,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(2880,36,'cp 78','f8e8 f1e1 d7d6 h2h3 c6c5 e4e5 d6e5 e1e5 c8b7 e5e8 d8e8 c3c4 e8e5 a1b1 b7c6 d1d2 a8e8 c1b2 f6e4 d2e3 e5g5 e3g5 e4g5 b2c3 c6e4 d3e4 g5e4 c3a5 c7c6 b1b7 e4d6 b7a7 h7h5 g1f1 e8e4 a2a4 e4c4 a5b6 c4c2 a4a5 c2a2 b6c5 a2a1 f1e2 d6e4 c5d4 a1a4 e2e3');
INSERT INTO evaluations VALUES(16605,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(16605,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(16605,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(16605,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(16605,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(16605,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(16605,36,'cp -64','d7d5 d2d3 f8c5 h2h3 c8e6 b1d2 d8e7 e2e3 e8c8 a2a3 c5b6 g1e2 d5d4 e3d4 e5d4 d2f3 h8e8 e1g1 e6b3 e2d4 c6d4 c2b3 e7e2 b2d4 b6d4 f3d4 d8d4 b3b4 e2d1 f1d1 a7a6 a1c1 e8e2 g2f1 e2a2 c1a1 a2c2 a1c1 c2c1 d1c1 c7c6');
INSERT INTO evaluations VALUES(33234,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(33234,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(33234,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(33234,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(33234,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(33234,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(33234,36,'cp 228','d2d4 e5d4 c3d4 c5b6 b1c3 d7d6 a2a4 b6a5 f1e2 f6d8 e1g1 b8c6 e2c4 g8f6 c3d5 h7h6 b2b4 f6d5 b4a5 d5e7 d1d2 b7b6 a5b6 a7b6 d4d5 c6a5 c4b5 c8d7 d2c3 e8g8 c1b2 f7f6 f3d4 a5b7 b5d7 d8d7 d4e6 f8f7 c3c7 d7c7 e6c7 a8c8 c7b5 f6f5 f2f3 f5e4 f3e4 f7f1 a1f1');
INSERT INTO evaluations VALUES(33228,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(33228,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(33228,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(33228,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(33228,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(33228,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(33228,36,'cp 225','e5d4 c3d4 c5b6 b1c3 d7d6 a2a4 b6a5 f1e2 b8c6 e1g1 f6d8 e2c4 g8f6 c3d5 f6d5 e4d5 c6e7 b2b4 a5b4 a4a5 b7b5 c4b5 e8f8 d1b3 a8b8 b5d3 c8d7 a1b1 c7c6 d5c6 d7c6 f3g5 c6d5 d3c4 d5c4 b3c4 d6d5 c4d3 h7h6 g5f3 f8g8 c1f4 b4d6 f4d6 b8b1 f1b1 d8d6 a5a6');
INSERT INTO evaluations VALUES(33252,30,'cp 210',NULL);
INSERT INTO evaluations VALUES(33252,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(33252,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(33252,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(33252,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(33252,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(33252,36,'cp 244','c3d4 c5b6 a2a4 b6a5 b1c3 d7d6 f1b5 c7c6 b5c4 c8g4 d1b3 b8d7 f3g5 g8h6 h2h3 g4h5 c1e3 e8g8 e1g1 d6d5 c4d3 f6d8 b3b7 h5g6 b7c6 d7f6 a1c1 a8b8 e3f4 b8b2 f1d1 b2b6 c6c7 d5e4 g5e4 a5c3 e4c3 g6d3 d1d3 h6f5 c7a7 f8e8');
INSERT INTO evaluations VALUES(33250,30,'cp 246',NULL);
INSERT INTO evaluations VALUES(33250,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(33250,32,'cp 230',NULL);
INSERT INTO evaluations VALUES(33250,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(33250,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(33250,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(33250,36,'cp 214','c5b6 a2a4 b6a5 b1c3 d7d6 f1e2 f6d8 e1g1 b8c6 e2c4 g8f6 c3d5 h7h6 b2b4 f6d5 b4a5 d5e7 d1e1 e8g8 d4d5 c6b8 e1c3 b8d7 c1b2 d7f6 f1e1 c7c6 d5c6 b7c6 e4e5 f6d5 c3d2 d6e5 f3e5 c8e6 a1d1 d5f6 c4e6 d8d2 e6f7 f8f7');
INSERT INTO evaluations VALUES(33152,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(33152,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(33152,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(33152,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(33152,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(33152,35,'cp 254',NULL);
INSERT INTO evaluations VALUES(33152,36,'cp 219','b1c3 d7d6 a2a4 b6a5 f1e2 b8c6 e1g1 f6d8 d4d5 c6e7 a1b1 a5c3 b2c3 g8f6 c1g5 f6d7 f3d4 e8g8 e2d3 d7c5 d3c2 f7f6 g5h4 d8e8 a4a5 c8d7 f1e1 e7g6 h4g3 g6e5 f2f4 e5c4 f4f5 e8e7 d4e6 d7e6 d5e6 c4a5 d1h5');
INSERT INTO evaluations VALUES(43131,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(43131,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(43131,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(43131,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(43131,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(43131,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(43131,36,'cp 29','e2e4 d7d5 g5f6 e7f6 e4d5 e8g8 g2g3 f8e8 g1e2 b7b5 a2a3 b5b4 a3b4 b8a6 f1g2 a6b4 d1d2 c8f5 e1g1 f5c2 c3a2 b4a2 d2c2 a2b4 c2c4 d8d6 e2c3 a7a6 c3e4 d6b6 e4c5 a8c8 c5a6 b4a6 a1a6 b6b2 a6c6 g7f8 c6f6 b2b4 c4d3 b4e7 f6a6 e7e2 d3e2 e8e2 f1d1 f8d6');
INSERT INTO evaluations VALUES(43138,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43138,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(43138,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(43138,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(43138,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(43138,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(43138,36,'cp 35','d7d5 g5f6 e7f6 e4d5 e8g8 g2g3 f8e8 g1e2 b7b5 a2a3 b5b4 a3b4 b8a6 f1g2 a6b4 d1d2 c8f5 e1g1 f5c2 c3a2 b4a2 d2c2 a2b4 c2c4 d8d6 a1a4 a7a5 f1a1 d6d7 e2c3 f6f5 a4a5 a8a5 a1a5 b4c2 a5a7 c2d4 a7c7 d7d6 c7c8 e8c8 c4c8 g7f8');
INSERT INTO evaluations VALUES(50866,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(50866,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(50866,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(50866,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(50866,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(50866,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(50866,36,'cp -41','b8c6 g1f3 e7e5 f3e5 c6e5 d4e5 d8d1 c3d1 g8e7 e2e4 e7c6 f2f4 c6d4 a1a2 d4c6 e1f2 c8e6 c1e3 g7g6 a2a1 e8c8 a1c1 c6a5 f1e2 a5b3 c1c4 b3d2 c4d4 d8d4 e3d4 d2e4 f2e3 e4c5 d1c3 f8e7 e3f3 h8d8 d4c5 e7c5 h1d1 c5d4 f3e4 f7f5 e5f6 d4f6 d1d8 c8d8 a4a5');
INSERT INTO evaluations VALUES(17729,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(17729,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(17729,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(17729,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(17729,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(17729,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(17729,36,'cp -26','g1f3 e7e5 f3e5 c6e5 d4e5 d8d1 c3d1 g8e7 f2f4 e7c6 e2e4 c6d4 a1a2 d4c6 e1f2 c8e6 c1e3 g7g6 a2a1 e8c8 a1c1 c6a5 f1e2 a5b3 c1c4 b3d2 c4d4 d8d4 e3d4 d2e4 f2e3 e4c5 d1c3 f8e7 e3f3 h8d8 d4c5 e7c5 h1d1 c5d4 g2g4 e6b3 d1d3 d4c3 d3c3 b3a4 f4f5 a4b5 e2b5');
INSERT INTO evaluations VALUES(54529,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(54529,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54529,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54529,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(54529,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54529,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(54529,36,'cp 13','g8f6 f1b5 c8d7 b5d7 b8d7 e1g1 f8e7 d4c5 d7c5 d2b3 c5e4 f1e1 e8g8 c2c3 e7d6 c1e3 f8e8 e3d4 h7h6 b3c1 d8c7 c1d3 b7b6 g2g3 d6c5 g1g2 a7a5 a1c1 c7b7 a2a3 c5d4 f3d4 b7d7 d3f4 a8d8 e1e2 e4c5 e2e8 d8e8 c1c2');
INSERT INTO evaluations VALUES(46439,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46439,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(46439,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46439,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(46439,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(46439,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(46439,36,'cp 23','f1b5 c8d7 b5d7 b8d7 e1g1 f8e7 d4c5 d7c5 d2b3 c5e4 f1e1 e8g8 c2c3 f8e8 d1c2 d8c7 c1e3 h7h6 h2h3 e7d6 a1d1 a8d8 b3d2 b7b6 f3d4 c7d7 d2f3 f6h5 c2b3 h5f4 e3f4 d6f4 g2g3 f4c7 g1g2 a7a6 d4c2');
INSERT INTO evaluations VALUES(46391,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46391,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(46391,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(46391,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(46391,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(46391,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(46391,36,'cp 20','c8d7 b5d7 b8d7 e1g1 f8e7 d4c5 d7c5 d2b3 c5e4 f1e1 e8g8 c2c3 f8e8 d1c2 a8c8 c1e3 h7h6 h2h3 a7a6 a1d1 b7b5 a2a3 e7d6 b3c1 d6b8 c1d3 d8d7 e3d4 e8e6 a3a4 b5a4 d1a1 a4a3 a1a3 c8e8 a3a5');
INSERT INTO evaluations VALUES(28859,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(28859,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(28859,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(28859,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(28859,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28859,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(28859,36,'cp 16','b5d7 b8d7 e1g1 f8e7 d4c5 d7c5 f3d4 g7g6 d2f3 e8g8 h2h3 f8e8 f1e1 a8c8 c2c3 a7a6 d4c2 c5e6 f3d4 e6c5 c1e3 c5a4 a1b1 a4c5 c2b4 e7d6 d4f3 c5e6 b4d3 f6e4 d1c2 e6g7 d3f4 d6f4 e3f4 g7f5');
INSERT INTO evaluations VALUES(57641,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(57641,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(57641,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(57641,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(57641,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(57641,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(57641,36,'cp 3','c1b2 g8f6 g1f3 f8g7 e2e3 d7d6 c2c4 e7e5 d2d3 e8g8 f1e2 e5e4 d3e4 f6e4 d1c1 g7b2 c1b2 a7a5 b4b5 d8f6 b2f6 e4f6 f3d4 b8d7 h2h4 f6e4 e2f3 d7c5 f3e4 c5e4 b1d2 e4c5 h4h5 g6g5 d4b3 c5d7 h5h6 d7b6 e1g1');
INSERT INTO evaluations VALUES(57642,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57642,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(57642,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(57642,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57642,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57642,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(57642,36,'cp 1','g8f6 g1f3 f8g7 e2e3 e8g8 f1e2 d7d6 c2c4 e7e5 d2d3 e5e4 d3e4 f6e4 d1c1 g7b2 c1b2 a7a5 b4b5 b8d7 e1g1 b7b6 f3d4 c8b7 b2c2 d8g5 e2f3 d7c5 b1c3 a8e8 c3e2 f7f5 a1d1 g5f6 e2f4 g6g5 f4d5 b7d5 c4d5');
INSERT INTO evaluations VALUES(13715,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(13715,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(13715,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(13715,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(13715,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(13715,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(13715,36,'cp -52','f8d6 d1e2 c8e6 f3g5 d8e7 e1g1 e8g8 g5e6 f7e6 b5c6 b7c6 c1g5 a8e8 a1e1 e6e5 c3d1 d6c7 g5f6 f8f6 d1e3 e7d7 c2c4 e8f8 e3g4 f6g6 f2f3 g6g5 g1h1 g5f5 g4e3 f5f7 e3g4 f8e8 h1g1 f7f5 g2g3 h7h5 g4f2 c7a5 e1c1 h5h4');
INSERT INTO evaluations VALUES(15733,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(15733,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(15733,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(15733,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(15733,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(15733,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(15733,36,'cp 115','a2a3 b4a6 f1a6 b7a6 c1e3 f8e7 d1d3 f6g4 e1c1 e8g8 f3d2 f7f5 e4f5 a6a5 f5f6 f8f6 d2e4 f6f8 h2h3 g4e3 f2e3 c8d7 d1f1 f8f1 h1f1 d8c8 b2b3 c8b7 a3a4 a8f8 f1f8 g8f8 d3c4 c7c6 g2g4 b7b6 d5c6 d7c6');
INSERT INTO evaluations VALUES(15734,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(15734,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(15734,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(15734,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(15734,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(15734,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(15734,36,'cp 109','b4a6 f1a6 b7a6 c1e3 f8e7 d1d3 e8g8 f3d2 f6g4 e1c1 f7f5 h2h3 g4e3 f2e3 e7g5 e4f5 c8f5 d3e2 h7h6 d2e4 a6a5 g2g4 f5g6 e4g5 d8g5 h3h4 g5e7 d1g1 f8f7 h4h5 g6h7 g4g5 h6g5 h5h6 a8f8 h6g7 f7g7 c1b1 h7g6 e3e4 f8f4 b2b3');
INSERT INTO evaluations VALUES(15857,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(15857,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(15857,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(15857,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(15857,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(15857,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(15857,36,'cp 125','f1a6 b7a6 d1d3 f8e7 c1e3 c8b7 f3d2 e8g8 d2c4 c7c6 d5c6 b7c6 e3g5 d6d5 g5f6 d5c4 d3c4 c6b5 f6e7 d8e7 c3b5 a6b5 c4e2 b5b4 a3b4 e7b4 c2c3 b4b7 e1g1 a7a5 a1a3 a8b8 b2b3 f8d8 f1a1 b7b5 c3c4 b5b4 h2h3 h7h6 a3a5 b4b3 a5b5 b8b5 c4b5 b3c3 a1b1');
INSERT INTO evaluations VALUES(17601,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(17601,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(17601,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(17601,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(17601,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(17601,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(17601,36,'cp 57','d8c7 d2d3 e7e5 a2a4 a6a5 c1e3 f8d6 b1d2 g8e7 f3h4 h7h6 d1f3 b7b6 d2c4 c8e6 h4f5 e7f5 e4f5 e6c4 d3c4 e8c8 f3e4 c8b7 a1d1 c7e7 d1d2 d6c7 f1d1 d8d2 d1d2 h6h5 g1f1 h8e8 d2d1 f7f6 b2b3 e8h8 d1d3 h8f8 f1e2 g7g5 f5g6 f6f5 e4f3');
INSERT INTO evaluations VALUES(29102,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(29102,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(29102,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(29102,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(29102,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(29102,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(29102,36,'cp 98','e4e5 d6d5 c4b3 c5c4 b3a4 f6d7 c3d5 d8a5 d5e3 a5a4 e3g4 e7e6 g4e3 b8c6 d1e2 d7b6 b2b3 c4b3 c2b3 a4e4 d2d3 e4g6 e1g1 f8e7 d3d4 e8g8 c1b2 f8d8 a2a3 b6d5 e3d5 d8d5 b3b4 e7d8 b4b5 c6a5 a3a4 d8b6 f1d1 a8d8 a1c1 h7h6 g2g3 a5b3');
INSERT INTO evaluations VALUES(32672,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(32672,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(32672,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(32672,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(32672,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(32672,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(32672,36,'cp 163','e6f8 h8f8 d2d3 b8c6 b1c3 e8c8 c1e3 e7e6 d1d2 d6d5 e4d5 f6d5 f2f3 b7b6 c3d5 e6d5 e1g1 d8d6 b2b3 d6g6 g1h1 c6d4 e3d4 e5d4 a1e1 g6c6 h2h3 g7g5 e1e4 c8b7 f1e1 d5c5 e1e2 h7h6 h1h2 c5d6 h2g1 d6a3 e4d4 a3a2 d2c1');
INSERT INTO evaluations VALUES(32552,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(32552,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(32552,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(32552,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(32552,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(32552,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(32552,36,'cp 166','h8f8 d2d3 b8c6 e1g1 e8c8 b1c3 h7h6 c1e3 e7d7 a1b1 d6d5 c3d5 f6d5 e4d5 d7d5 d1g4 c8b8 b2b3 g7g5 b1e1 b7b6 h2h3 d8e8 e1e2 b8b7 e2d2 a7a5 f1e1 e8d8 c2c3 d5f7 f2f3 d8d6 g4c4 f7c4 b3c4');
INSERT INTO evaluations VALUES(32963,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(32963,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(32963,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(32963,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(32963,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(32963,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(32963,36,'cp 165','d2d3 b8c6 b1c3 e8c8 e1g1 e7d7 c1e3 h7h6 a1b1 d6d5 c3d5 f6d5 e4d5 d7d5 d1g4 c8b8 b2b3 g7g5 b1e1 a7a5 e1e2 b7b6 f1e1 b8b7 h2h3 f8f6 f2f3 c6d4 e3d4 d5d4 g4d4 e5d4 e2e5 a5a4 g1f2 f6c6 e1e2 c6c3 e5e6 c3c6 e6c6 b7c6');
INSERT INTO evaluations VALUES(43779,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(43779,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(43779,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(43779,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(43779,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(43779,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(43779,36,'cp 49','a2a4 c7c6 a4b5 c6b5 b2b3 e7e6 b3c4 b5c4 f1c4 f8d6 e1g1 e8g8 c1a3 c8b7 b1d2 a7a5 d1b3 a8a7 f1c1 d6a3 b3a3 b8d7 c4e2 a5a4 d2c4 d8a8 c4d6 b7c6 c1c2 c6e4 d6e4 f6e4 c2c4 d7b6 c4c1 f8b8 e2b5 a8d5 b5c6');
INSERT INTO evaluations VALUES(43781,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(43781,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(43781,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(43781,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(43781,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(43781,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(43781,36,'cp 55','c7c6 a4b5 c6b5 b1c3 d8b6 c1d2 e7e6 b2b3 f8b4 b3c4 b5c4 f1c4 c8b7 e1g1 a7a5 f3e5 e8g8 c3a4 b6c7 a1c1 b4a3 c1c3 c7e7 c4d3 a3b4 c3c1 b8a6 f2f3 b4d2 d1d2 a6b4 d3e2 b7a6 e2a6 a8a6 c1c4 a6a8 f1c1 f6d7 e5d7 e7d7 a4b6');
INSERT INTO evaluations VALUES(43942,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(43942,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(43942,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(43942,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(43942,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(43942,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43942,36,'cp 41','a4b5 c6b5 b2b3 e7e6 b3c4 b5c4 f1c4 f8d6 e1g1 e8g8 c1a3 c8b7 b1d2 a7a5 d1b3 a8a7 c4e2 d6a3 b3a3 b8d7 d2c4 f6e4 c4a5 b7d5 e2d3 d8a8 f1c1 e4f6 f3e1 e6e5 a3c3 e5d4 e3d4 h7h6 a5c4 f8c8 c3b2 c8b8 a1a7 a8a7 b2a1 a7a1 c1a1 d5c4 d3c4 d7b6 c4e2 b6d5 e1d3 d5c3 e2f1');
INSERT INTO evaluations VALUES(43931,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(43931,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(43931,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(43931,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(43931,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(43931,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(43931,36,'cp 37','c6b5 b1c3 d8b6 c1d2 e7e6 b2b3 f8b4 b3c4 b5c4 f1c4 c8b7 e1g1 a7a5 f3e5 e8g8 c3a4 b6c7 a1c1 b4a3 c1c3 c7e7 c4d3 a3b4 c3c1 b8a6 f2f3 e7d6 g1h1 b4d2 d1d2 a6b4 d3b5 b7a6 b5a6 d6a6 a4c5 a6d6 c1a1 f8c8 f1c1 d6d8 a1a3 b4c6 e5c4 c6b4 e3e4');
INSERT INTO evaluations VALUES(43950,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43950,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(43950,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(43950,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(43950,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(43950,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(43950,36,'cp 61','b2b3 e7e6 b3c4 b5c4 f1c4 f8d6 c1a3 e8g8 e1g1 c8b7 c4e2 b8c6 d1b3 c6b4 a3b4 d8b6 b1d2 b6b4 b3d3 d6b8 f1b1 b4e7 d3b5 b7d5 e2d3 b8c7 b5a6 c7b6 b1c1 d5b7 a6a3 e7a3 a1a3 f8c8 d2c4 b7d5 f3d2 a8b8 c1a1 c8c7');
INSERT INTO evaluations VALUES(54993,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(54993,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(54993,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(54993,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(54993,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54993,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(54993,36,'cp 29','c7c5 e2e4 e7e5 f1c4 f8e7 d2d3 g8f6 c1g5 b8c6 a2a4 c6b4 g5f6 e7f6 c3d5 b4d5 c4d5 a8b8 e1g1 e8g8 c2c3 g7g6 h2h3 c8d7 b2b4 c5b4 c3b4 a7a6 d1d2 d7c6 d5c6 b7c6 f1c1 d8d7 g2g3 f8c8 g1g2 g8g7 d3d4 e5d4');
INSERT INTO evaluations VALUES(53101,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53101,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(53101,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(53101,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(53101,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(53101,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(53101,36,'cp 42','e2e4 e7e5 f1c4 b8c6 d2d3 h7h6 h2h3 f8e7 a2a4 g8f6 f3h2 e8g8 h2g4 c8e6 c3d5 a7a6 g4e3 b7b5 a4b5 a6b5 a1a8 d8a8 c4b3 a8b7 e1g1 c6d4 b3a2 b7d7 c2c3 d4c6 d1f3 g8h7 e3f5 f6d5 a2d5');
INSERT INTO evaluations VALUES(55531,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(55531,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(55531,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(55531,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(55531,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(55531,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(55531,36,'cp -32','h5d1 g7g6 d2d4 d8e7 f1e2 g8f6 g1f3 f8g7 b2b4 e8g8 e1g1 f8e8 c2c3 c8f5 c1b2 a7a5 a2a3 c7c6 c3c4 f6e4 b4a5 a8a5 a3a4 b8d7 b1c3 d7f6 d1b3 a5a8 a4a5 e5d4 c3e4 f6e4 b2d4 g7d4 f3d4 e4d2 d4f5 d2b3 f5e7 e8e7');
INSERT INTO evaluations VALUES(6480,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6480,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(6480,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(6480,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(6480,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(6480,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6480,36,'cp 9','d7d6 c3a4 a7a5 a4c5 d6c5 d2d3 b7b6 c2c3 g8e7 e1g1 e8g8 c1e3 e7g6 d3d4 e5d4 c3d4 c6d4 f3d4 c5d4 d1d4 d8d4 e3d4 g6f4 e2f3 c8a6 f1d1 f4e2 f3e2 a6e2 d1e1 e2g4 h2h3 g4e6 d4c3 a8e8 a2a3 f7f6 f2f3 e8d8 e1d1 c7c5');
INSERT INTO evaluations VALUES(6011,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6011,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(6011,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6011,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(6011,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(6011,36,'cp 9','c3a4 g8e7 c2c3 a7a5 a4c5 d6c5 d2d3 b7b6 c1e3 f7f5 e4f5 c8f5 e1g1 d8d6 d1a4 e8g8 d3d4 c5d4 c3d4 e5e4 f3e5 f5e6 e5c6 e7c6 a1c1 e6d5 a4c2 d6g6 a2a3 h7h6 h2h3 g8h8 e2g4 g6d6 g4e2');
INSERT INTO evaluations VALUES(15736,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(15736,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(15736,32,'cp 259',NULL);
INSERT INTO evaluations VALUES(15736,33,'cp 250',NULL);
INSERT INTO evaluations VALUES(15736,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(15736,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(15736,36,'cp 258','f3d4 e5d4 d1d4 f8e7 b2b3 c7c5 d5c6 b7c6 c1b2 e8g8 e1c1 d8b6 f1e2 f8e8 c1b1 c8e6 f2f4 a8d8 e2f3 b6d4 d1d4 d6d5 e4d5 c6d5 h1e1 e7c5 d4d1 c5b6 e1e2 h7h5 c3b5 a7a6 b2d4 b6d4 b5d4 e6g4 e2e8 d8e8');
INSERT INTO evaluations VALUES(15735,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(15735,31,'cp 244',NULL);
INSERT INTO evaluations VALUES(15735,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(15735,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(15735,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(15735,35,'cp 254',NULL);
INSERT INTO evaluations VALUES(15735,36,'cp 264','e5d4 d1d4 f8e7 b2b3 e8g8 c1b2 f6h5 e1c1 e7f6 d4c4 c8d7 c4e2 h5f4 e2f3 f4g6 h2h4 f6e5 c1b1 f7f5 h4h5 f5e4 f3e4 g6e7 f1d3 d7f5 e4e2 h7h6 h1e1 f5d3 d1d3 d8d7 g2g4 a8e8 c3e4 e5b2 b1b2 a7a5 f2f3 a5a4 e2d2 a4b3');
INSERT INTO evaluations VALUES(15731,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(15731,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(15731,32,'cp 265',NULL);
INSERT INTO evaluations VALUES(15731,33,'cp 270',NULL);
INSERT INTO evaluations VALUES(15731,34,'cp 265',NULL);
INSERT INTO evaluations VALUES(15731,35,'cp 278',NULL);
INSERT INTO evaluations VALUES(15731,36,'cp 268','d1d4 f8e7 b2b3 e8g8 c1b2 f6h5 g2g3 f7f5 f1e2 e7f6 d4c4 g7g6 f2f4 f5e4 e1c1 d8e7 e2h5 g6h5 c3e4 f6b2 c1b2 e7g7 c4c3 g7c3 b2c3 c8d7 d1e1 h7h6 c3d4 h5h4 c2c4 b7b6 e1e3 a8e8 h1g1 f8f7 e4c3 g8f8 e3e8 f8e8 c3b5 d7g4 g1e1 e8f8 b3b4 f8g7 e1e8 g4f5 b5c3 f5g4');
INSERT INTO evaluations VALUES(31972,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(31972,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31972,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(31972,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(31972,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(31972,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(31972,36,'cp -5','d1f3 d8c8 f1c4 e7e6 g1e2 g8f6 d4d5 f5g4 f3g3 f8d6 c1f4 e8g8 d5e6 g4e6 f4d6 c7d6 c4b3 e6b3 a2b3 c8g4 e1c1 f8d8 g3g4 f6g4 d1d4 g4f6 h1d1 f6e8 c3b5 g8f8 d4d2 b8c6 g2g3 d6d5 d2d5 d8d5 d1d5 e8f6 d5d2 a8e8 b5c3');
INSERT INTO evaluations VALUES(31971,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(31971,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(31971,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31971,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(31971,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(31971,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(31971,36,'cp -33','d8c8 g1e2 c7c6 c1f4 g8f6 e1c1 g7g6 h2h3 f8g7 c1b1 e8g8 g2g4 f5e6 f4c1 f6d5 c3e4 b8d7 e4g5 d7b6 g5e6 c8e6 h3h4 d5b4 b2b3 a7a5 a2a3 b4d5 c2c4 d5f6 g4g5 e6e4 d1d3 e4f3 d3f3 f6d7 h4h5 e7e5 h5h6 g7h8 c1b2 a5a4 d4e5 h8e5');
INSERT INTO evaluations VALUES(37773,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37773,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37773,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37773,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37773,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37773,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37773,36,'cp 0','c5d4 c2c3 d4c3 b1c3 b8c6 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4g5 d8a5 c3d5 h6f5 g5d2 a5d8 d2g5');
INSERT INTO evaluations VALUES(37851,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37851,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37851,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37851,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37851,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37851,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37851,36,'cp 0','c2c3 d4c3 b1c3 b8c6 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4g5 d8a5 c3d5 h6f5 g5d2 a5d8 d2g5');
INSERT INTO evaluations VALUES(42321,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(42321,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(42321,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(42321,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(42321,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(42321,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(42321,36,'cp 89','e2e4 d7d6 c1e3 e8g8 f1e2 e6e5 e1g1 e5d4 f3d4 b8c6 d1d2 c6d4 e3d4 f8e8 f2f3 c8d7 b2b3 d7c6 a1e1 a7a5 e2d3 f6d7 d4g7 g8g7 f3f4 d8f6 d3c2 g7g8 a2a3 f6g7 e1e3 d7f6 h2h3 f6d7 g1h1 d7c5 f1e1 c5e6');
INSERT INTO evaluations VALUES(42323,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(42323,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(42323,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(42323,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(42323,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(42323,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(42323,36,'cp 97','d7d6 h2h3 e8g8 c1e3 b8d7 d1c2 e6e5 e1c1 d8e7 d4e5 d7e5 f3e5 e7e5 f2f4 e5a5 c1b1 f8e8 f1d3 c8d7 e3d2 a5b6 h1e1 d7c6 d2e3 b6a5 e3d4 e8e7 g2g4 a8e8 g4g5 f6d7 d4g7 g8g7 c3d5 c6d5');
INSERT INTO evaluations VALUES(42505,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42505,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42505,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(42505,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(42505,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(42505,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(42505,36,'cp 12','e8g8 e2e3 d7d6 f1d3 d8e7 e1g1 e6e5 d1c2 h7h6 c4c5 c8g4 d4e5 d6e5 d3e4 c7c6 h2h3 f6e4 c2e4 g4f3 d2f3 b8d7 f1d1 a7a5 a1c1 f8d8 e4c2 e5e4 f3d4 d7f6 a2a3 d8d5 d4e2 a8d8 b2b4 a5b4 a3b4 e7e5 d1d4 h6h5 c2c3 d8e8 d4d5 f6d5 c3e5 e8e5');
INSERT INTO evaluations VALUES(37192,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(37192,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(37192,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(37192,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(37192,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37192,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(37192,36,'cp 14','d1c2 d7d6 e2e3 d8e7 f1d3 h7h6 e1g1 e6e5 c4c5 b8c6 c5d6 c7d6 d4e5 d6e5 d2e4 c8g4 e4f6 e7f6 d3e4 a8c8 c2b3 c8c7 f1d1 g4f5 e4f5 f6f5 h2h3 f8c8 b3a4 a7a6 a1c1 f5e6 f3d2 f7f5 d2b3 e5e4 c1c5 b7b6 c5c1 g8h7 b3d4 c6d4 c1c7 c8c7 a4d4 c7c2');
INSERT INTO evaluations VALUES(55556,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55556,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(55556,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(55556,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(55556,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(55556,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55556,36,'cp 26','g8f6 d2d4 c7c6 f1d3 g7g6 g1e2 b8d7 e3e4 e5d4 e2d4 f8g7 h2h3 e8g8 e1g1 f8e8 c1e3 a7a6 f1e1 d7c5 d1c2 c5d3 c2d3 c6c5 d4f3 b7b5 c4b5 a6b5 d3b5 c8a6 b5c6 d8b8 a1d1 b8b2 c3d5 a6b7 d5f6 g7f6 c6d6 b7e4 f3d2 e4f5 e3c5 b2c3 e1e8 a8e8 c5e3');
INSERT INTO evaluations VALUES(47830,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(47830,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(47830,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(47830,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(47830,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(47830,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(47830,36,'cp 37','g1f3 g7g6 d2d4 b8d7 f1e2 f8g7 e1g1 e8g8 b2b4 e5d4 e3d4 d6d5 b4b5 f8e8 f1e1 d5c4 e2c4 d7b6 c4b3 c8g4 e1e8 d8e8 c1f4 g4e6 a1c1 b6d5 f4e5 a8d8 a2a4 d5c3 c1c3 e6b3 d1b3 f6d5');
INSERT INTO evaluations VALUES(47831,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47831,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47831,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(47831,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(47831,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47831,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(47831,36,'cp 20','g7g6 d2d4 d8e7 f1e2 f8g7 b2b4 c8f5 b4b5 e8g8 e1g1 b8d7 a2a4 f6e4 c3e4 f5e4 a4a5 h7h5 c1b2 e4f3 e2f3 e5e4 f3e2 c7c5 b5c6 b7c6 a1b1 c6c5 b2a1 a8b8 d1d2 a7a6 a1c3 b8b1 f1b1 f8b8 d2c2 b8b1 c2b1 c5d4 e3d4 e4e3');
INSERT INTO evaluations VALUES(6414,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(6414,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(6414,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6414,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(6414,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(6414,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(6414,36,'cp 15','c5b6 a2a4 a7a6 f1c4 g8f6 d2d3 h7h6 e1g1 e8g8 b1d2 d7d6 c4b3 c8e6 d2c4 b6a7 a4a5 c6e7 c1e3 a7e3 c4e3 e7g6 f1e1 b7b5 a5b6 c7b6 g2g3 d8d7 f3d2 a6a5 b3e6 f7e6 e3c4 d7c7 g1g2 g6e7 b4a5 b6a5 a1b1 a5a4');
INSERT INTO evaluations VALUES(6313,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(6313,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(6313,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(6313,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(6313,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(6313,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(6313,36,'cp 28','a2a4 a7a6 f1c4 g8f6 d2d3 h7h6 e1g1 e8g8 f1e1 d7d6 b1d2 c8e6 h2h3 d8d7 c4b3 e6b3 d2b3 d6d5 e4d5 f6d5 c1d2 d5f6 a4a5 b6a7 d2e3 a7e3 f2e3 e5e4 b3c5 d7e8 f3d4 c6e5 c5e4 f6e4 d3e4 a8d8 d1e2 e8e7 b4b5 a6b5 e2b5 c7c6');
INSERT INTO evaluations VALUES(6329,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(6329,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(6329,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(6329,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(6329,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(6329,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(6329,36,'cp 15','a7a6 f1c4 g8f6 d2d3 d7d6 e1g1 e8g8 b1d2 h7h6 f1e1 c8e6 h2h3 d8d7 c4b3 e6b3 d2b3 f8e8 a4a5 b6a7 c1e3 a7e3 e1e3 b7b6 d1c2 d6d5 e3e1 d7d6 f3d2 c6e7 g2g3 e7g6 g1h2 e8f8 d2f3 d6d7 c3c4 d5e4 d3e4');
INSERT INTO evaluations VALUES(18989,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(18989,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(18989,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(18989,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(18989,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(18989,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(18989,36,'cp 35','a2a3 a7a5 d2d3 g7g6 f1g2 d7d6 d1a4 c8d7 a1b1 f7f5 c1d2 g8f6 a4d1 h7h6 e2e3 f8g7 g1e2 e8g8 e1g1 a8b8 c3d5 f6d5 g2d5 g8h7 b2b4 a5b4 a3b4 c6e7 d5g2 b7b5 d1c2 d8c7 f1c1 b5c4 c2c4 d7b5 c4c2 c7d7 e2c3 b5c6 g2c6 e7c6 b4c5 d6c5');
INSERT INTO evaluations VALUES(23872,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(23872,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(23872,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(23872,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(23872,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(23872,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(23872,36,'cp 48','c2c3 a7a5 h2h3 g4h5 b1d2 g8e7 f1e1 e8g8 d2f1 h5g6 f3h4 g8h8 c1e3 c5e3 f1e3 e7c8 d1g4 c8b6 a1d1 a5a4 a2a3 b6d7 h4f5 c6e7 f5g3 d7f6 g4f3 c7c6 d3d4 d8c7 d4e5 d6e5 c4f1');
INSERT INTO evaluations VALUES(23869,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(23869,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(23869,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(23869,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(23869,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(23869,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(23869,36,'cp 40','a7a5 h2h3 g4h5 a2a4 g8f6 b1d2 h7h6 f1e1 e8g8 d2f1 h5g6 f1g3 f6d7 c1d2 d8f6 d1e2 c5b6 c4b5 a8d8 a1d1 d7c5 d2e3 b6a7 f3h2 g8h8 h2g4 f6e6 b5c4 e6d7 b2b3 c6e7 f2f4 e5f4 e3f4 c5e6 f4e3');
INSERT INTO evaluations VALUES(38604,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(38604,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38604,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(38604,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(38604,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(38604,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38604,36,'cp 19','d8d6 c2c4 g8f6 b1c3 e8g8 e2e3 d5c4 f1c4 c7c5 e1g1 b8c6 d4c5 d6c5 c4e2 f8d8 d1c2 c8d7 a1c1 a8c8 f1d1 h7h6 h2h3 c6e5 f3e5 c5e5 e2f3 e5c7 a2a3 c7b6 b2b4 c8c7 c2b2 d8c8 c3e4 f6e4 f3e4 c7c1');
INSERT INTO evaluations VALUES(33104,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(33104,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(33104,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(33104,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(33104,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(33104,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(33104,36,'cp 12','c2c4 g8f6 e2e3 e8g8 f1e2 d5c4 b1a3 d6b4 d1d2 b4d2 f3d2 c7c5 d4c5 c4c3 b2c3 b8d7 c5c6 b7c6 e1g1 a7a5 f1d1 c8a6 e2a6 a8a6 a3c4 d7c5 a1c1 f8a8 g1f1 g8f8 f2f3 f8e7 f1e2 f6d7 d2b3 c5a4');
INSERT INTO evaluations VALUES(42687,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(42687,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(42687,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(42687,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(42687,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(42687,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(42687,36,'cp 188','e4e5 f6h5 d2d4 c5e7 d1d3 g7g6 g2g4 f4g3 c1h6 d7d5 e5d6 d8d6 e1c1 c8f5 d3c4 c7c6 d4d5 b8d7 h2g3 d7b6 c4b3 e8c8 f3d4 b6d5 d4f5 g6f5 h1h5 d6h6 h5h6 e7g5 d1d2 g5h6 c3d5 d8d5 f1d3 h8g8 c1d1 h6d2 d1d2');
INSERT INTO evaluations VALUES(54212,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54212,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(54212,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(54212,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54212,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54212,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54212,36,'cp 0','b2b3 b7b6 g1f3 g8f6 c1b2 c8b7 e1g1 f8d6 b1d2 e8g8 d4c5 b6c5 c2c4 f8e8 f1e1 b8c6 a2a3 h7h6 d1c2 f6d7 a1d1 d8c7 d1c1 d7e5 b2e5 c6e5 f3e5 d6e5 c4d5 e5h2 g1f1 h2d6 d5e6 e8e6 d2c4 d6e7 e1d1');
INSERT INTO evaluations VALUES(10309,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(10309,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(10309,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(10309,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(10309,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(10309,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(10309,36,'cp -63','c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g1f3 a7a6 e4e5 d7d6 c3d5 e8d8 e5d6 h8e8 f1e2 c5d6 a2a3 c8e6 a1d1 d8d7 d5e3 a8d8 e1g1 d7e7 g2g3 e7f8 b2b4 d6e7 e2c4 e6h3 d1d8 e8d8 f1d1 d8d1 e3d1 e7f6 c4d5 h3d7 f3d2 b7b6 d1e3 c6e5 c2c3 g7g5 d2c4 e5c4 d5c4');
INSERT INTO evaluations VALUES(21291,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(21291,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(21291,32,'cp -175',NULL);
INSERT INTO evaluations VALUES(21291,33,'cp -148',NULL);
INSERT INTO evaluations VALUES(21291,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(21291,35,'cp -164',NULL);
INSERT INTO evaluations VALUES(21291,36,'cp -188','e5f4 d2d4 g7g5 b1c3 f8g7 c3d5 c6e7 d5e7 g8e7 c2c3 c8g4 d1d3 d8d7 c1d2 g4h5 a2a4 e8c8 f1e1 d6d5 e4d5 h5g6 d3f1 e7d5 f3e5 g7e5 e1e5 d5f6 a1e1 h8e8 e5e8 d8e8 c4b5 c7c6 e1e8 f6e8 b5d3 g6d3 f1d3');
INSERT INTO evaluations VALUES(21302,30,'cp -161',NULL);
INSERT INTO evaluations VALUES(21302,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(21302,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(21302,33,'cp -151',NULL);
INSERT INTO evaluations VALUES(21302,34,'cp -161',NULL);
INSERT INTO evaluations VALUES(21302,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(21302,36,'cp -171','b2b3 g7g5 c1b2 g8f6 b1c3 f8g7 c3d5 e8g8 d5f6 g7f6 d2d4 f6g7 h2h3 c8e6 c4e6 f7e6 d1e2 d8f6 a1e1 f6g6 c2c4 e6e5 d4e5 d6e5 b3b4 a8e8 c4c5 a7a6 b4b5 a6b5 e2b5 c6d8 a2a4 g8h8');
INSERT INTO evaluations VALUES(33421,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(33421,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(33421,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33421,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(33421,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33421,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(33421,36,'cp 14','c2c4 e8g8 f1e2 c7c5 e1g1 b8c6 c4d5 f6d5 d4c5 d6c5 a2a3 c8d7 b1d2 c5e7 a1c1 f8c8 d2e4 a7a6 h2h3 d7e8 d1d3 h7h6 d3b1 c6a7 c1c8 a8c8 f1c1 c8d8 f3e5 d5b6 e2f3 a7b5 e4c5 b5d6 e5d3');
INSERT INTO evaluations VALUES(33414,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33414,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(33414,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(33414,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(33414,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(33414,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(33414,36,'cp 10','e8g8 b1c3 d5c4 f1c4 c7c5 c3b5 d6b6 d4c5 b6c5 c4e2 b8c6 e1g1 f8d8 a1c1 c5e7 d1c2 a7a6 b5c3 c8d7 a2a3 c6a7 f1d1 a8c8 c2b3 d7e8 d1d8 c8d8 c1d1 d8d1 b3d1 e7c7 f3d4 a7c6 h2h3 h7h6 d4c6 e8c6 d1d2 c7e7 b2b4');
INSERT INTO evaluations VALUES(42745,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(42745,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(42745,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(42745,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(42745,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(42745,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(42745,36,'cp 65','c2c3 b4d6 f1e2 b8c6 b1d2 d6e7 e1g1 e8g8 b2b4 d7d6 f1e1 f8e8 c1b2 e7f8 e2f1 c8d7 a2a3 c6e7 d3d4 e7g6 d4d5 c7c6 c3c4 a8c8 a3a4 c6d5 c4d5 a7a6 d2c4 b7b5 a4b5 a6b5 c4a5 d8b6 g2g3 h7h6 d1e2 c8c7 h2h3');
INSERT INTO evaluations VALUES(42744,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(42744,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(42744,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(42744,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(42744,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(42744,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(42744,36,'cp 58','b4d6 f1e2 b8c6 b1d2 d6e7 b2b4 d7d6 a2a3 e8g8 e1g1 f8e8 f1e1 e7f8 e2f1 c6e7 d3d4 e7g6 d4d5 a7a5 d2c4 a5b4 c3b4 c7c6 d5c6 b7c6 c1g5 c8e6 a3a4 e6c4 f1c4 d6d5 e4d5 f8b4 d5c6 b4e1 d1e1');
INSERT INTO evaluations VALUES(50569,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(50569,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(50569,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(50569,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(50569,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(50569,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(50569,36,'cp 0','c1f4 d7d6 f1e2 f8e7 e1g1 b8c6 d1b3 g8f6 f1d1 d8c7 b3a3 e6e5 f4g5 d6d5 a3a4 d5e4 g5f6 e7f6 c3d5 c7d8 d5e3 d8a5 a4e4 c8e6 e2c4 a5b4 a2a3 b4e7 e3d5 e6d5 c4d5 e8g8 a1c1 a8c8 h2h4 f8d8 h4h5 h7h6 g2g3');
INSERT INTO evaluations VALUES(52642,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(52642,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(52642,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(52642,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(52642,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(52642,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(52642,36,'cp 72','f8g7 b1c3 c8g4 c1e3 g8f6 f1b5 f6d7 d1d2 e8g8 b5e2 d7b6 h2h4 b8d7 b2b3 a8c8 a2a4 d7f6 e1g1 d6d5 e4e5 f6e4 c3e4 d5e4 f3g5 g4f5 g2g4 f5d7 a1c1 d7c6 h4h5 g6h5 g4h5 h7h6 a4a5 h6g5 a5b6 a7b6 e3g5');
INSERT INTO evaluations VALUES(52739,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(52739,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52739,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(52739,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(52739,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(52739,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52739,36,'cp 17','g1f3 g8f6 d2d4 c8f5 b1c3 e7e6 f3h4 f5g6 h4g6 h7g6 c1d2 b8d7 c4d5 e6d5 d1c2 f8b4 e1c1 b4c3 d2c3 e8g8 h2h4 a7a5 f1d3 b7b5 h4h5 b5b4 c3d2 g6h5 g2g4 c6c5 g4g5 c5c4 d3f5 b4b3 c2b1 f6e4 f5e4 d5e4 h1h5 g7g6');
INSERT INTO evaluations VALUES(4260,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(4260,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(4260,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4260,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(4260,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(4260,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(4260,36,'cp 13','c5b6 d4e5 f6e4 d1d5 e4c5 f1e1 d7d6 e5d6 c8e6 d5d1 d8d6 d1d6 c7d6 c1e3 c6e5 f3e5 d6e5 e3c5 b6c5 e1e5 a8c8 b1d2 f8d8 d2f3 g7g6 h2h3 a7a6 b5a4 e6d5 a1d1 d5f3 d1d8 c8d8 g2f3 c5d6 e5e1 b7b5 a4c2 b5b4 c3b4 d6b4 e1e2 a6a5');
INSERT INTO evaluations VALUES(9824,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(9824,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(9824,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(9824,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9824,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9824,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9824,36,'cp 0','d7d6 c2c3 f6d5 c4d5 e8g8 e1g1 a7a5 a2a4 c5a7 c1e3 a7e3 f2e3 c6b8 d1c2 b8d7 h2h3 c7c6 d5a2 d8e7 f1f2 d7c5 f3d2 c8e6 a2e6 c5e6 d3d4 e6g5 a1f1 e5d4 e3d4 a8e8 c2b3 e7d7 b3c2');
INSERT INTO evaluations VALUES(8878,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8878,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8878,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8878,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8878,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8878,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8878,36,'cp 0','e1g1 f6d5 c4d5 a7a5 c1e3 c6e7 e3c5 d6c5 f3e5 e7d5 e4d5 d8d5 f1e1 e8g8 d1f3 d5f3 e5f3 c8e6 a2a4 f8e8 f3d2 e6d5 e1e3 b7b6 b2b3 f7f5 a1e1 g8f7 f2f4 e8e3 e1e3 a8e8 e3e8 f7e8 d2c4 e8e7 g1f2 e7d7 c4e3 d7e6 e3d5 e6d5 c2c3 d5d6 f2f3 d6e6 h2h3 e6f7 f3f2');
INSERT INTO evaluations VALUES(8877,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8877,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8877,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8877,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(8877,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8877,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8877,36,'cp 0','f6d5 c4d5 a7a5 c1e3 c6e7 e3c5 d6c5 f3e5 e7d5 e4d5 d8d5 f1e1 e8g8 d1f3 d5f3 e5f3 c8e6 a2a4 f8e8 f3d2 e6d5 e1e3 b7b6 b2b3 f7f5 a1e1 g8f7 e3e8 a8e8 e1e8 f7e8 f2f4 g7g5 g2g3 e8e7 d2f1 e7e6 f1e3 d5c6 g1f2 e6f6 f2e2 f6e6');
INSERT INTO evaluations VALUES(8898,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8898,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8898,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8898,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8898,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8898,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8898,36,'cp 0','c4d5 a7a5 c1e3 c6e7 e3c5 d6c5 f3e5 e7d5 e4d5 d8d5 f1e1 e8g8 d1f3 d5f3 e5f3 c8e6 a2a4 f8e8 f3d2 e6d5 e1e3 b7b6 b2b3 f7f5 a1e1 g8f7 e3e8 a8e8 e1e8 f7e8 f2f4 g7g5 g2g3 e8e7 d2f1 e7e6 f1e3 d5c6 g1f2 c6e8 e3c4 e6d5 c4e3');
INSERT INTO evaluations VALUES(8896,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8896,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8896,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8896,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8896,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8896,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8896,36,'cp 4','d8f6 c1e3 c6e7 d3d4 e7d5 e4d5 e5d4 e3d4 c5d4 d1d4 f6d4 f3d4 c8d7 f1e1 e8d8 a2a4 a7a5 a1a3 c7c5 d4b5 d7b5 a4b5 h8e8 e1e8 d8e8 g1f1 e8d7 c2c4 b7b6 a3g3 g7g6 g3h3 h6h5 h3e3 a8e8 e3e8 d7e8 f1e2 e8e7 e2e3 e7f6 e3f4 g6g5 f4e4 h5h4 f2f4 g5g4');
INSERT INTO evaluations VALUES(13921,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(13921,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(13921,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(13921,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(13921,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(13921,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(13921,36,'cp -47','e1g1 a7a6 b1d2 f8e7 c4b3 c6a5 f1e1 e8g8 d2f1 b7b5 c1d2 a5b3 a2b3 a6a5 f2f4 c5c4 f1g3 c4d3 c2d3 d8b6 d2e3 b6c7 h2h3 c8d7 a1c1 c7b7 g5f3 f8c8 c1a1 b7c7 a1c1 c7d8 c1a1 a5a4 g1h2 c8c7 f3d4 a8c8 b3a4 b5a4');
INSERT INTO evaluations VALUES(13920,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(13920,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(13920,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(13920,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(13920,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(13920,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(13920,36,'cp -45','a7a6 b1d2 b7b5 c4b3 c6a5 a2a4 f8e7 a4b5 a6b5 c2c3 e8g8 b3c2 c8b7 f1e1 h7h6 g5f3 d8c7 d2f1 a5c6 a1a8 f8a8 c1f4 c6e5 f3e5 d6e5 f4g3 f6d7 c2b3 d7f8 d1g4 c5c4 d3c4 b5c4 b3c2 f8g6 h2h4 g6f4 g4f3 b7c6 f1d2');
INSERT INTO evaluations VALUES(23410,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(23410,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(23410,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(23410,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(23410,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(23410,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(23410,36,'cp 203','d4d5 c6b8 d1d4 g8f6 e4e5 f6g8 c1f4 f8g7 b1c3 d7d6 e1c1 a7a6 f1e2 g7e5 f4e5 d6e5 d4e5 g8f6 g1f3 d8d6 e5e3 e8g8 f3g5 b7b5 e2f3 c8f5 h1e1 b5b4 c3a4 b8d7 g5e6 d6h2 e6f8 a8f8 e3e7 a6a5 e7e3 f8e8 e3g5 e8e5 g5d2 e5e1 d1e1 d7e5');
INSERT INTO evaluations VALUES(23402,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(23402,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(23402,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(23402,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(23402,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(23402,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(23402,36,'cp 162','c6b8 d1d4 g8f6 e4e5 f6g8 b1c3 f8g7 c1f4 d7d6 e1c1 b8d7 e5d6 g7d4 d6c7 d8c7 f4c7 d4f2 d5d6 e7e6 f1e2 g8f6 g1h3 f2e3 c1b1 e8g8 h1f1 e3h6 g2g4 h6g7 a2a4 b7b6 e2b5 f6g4 b5c6 a8b8 h3g5 g4h6 a4a5 h6f5 a5b6 a7b6 c7b8');
INSERT INTO evaluations VALUES(27558,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(27558,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(27558,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(27558,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(27558,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(27558,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(27558,36,'cp -41','d1e2 a7a6 b1d2 d4a4 d2b3 d7b5 c1e3 b5d3 e2d3 g8e7 b3c5 a4b5 d3b5 a6b5 c5b7 e7c6 f2f4 e8d7 e3c5 d7c7 c5f8 h8f8 b7c5 a8a7 a1c1 c7b6 a2a3 c6a5 c5d3 a5c4 f1f2 a7c7 f4f5 f8c8 f5e6 f7e6 h2h4 c4e3 c1c7 c8c7 f2f8 c7c2 f8f2 c2f2 g1f2 d5d4 h4h5 b6c6');
INSERT INTO evaluations VALUES(47436,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(47436,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(47436,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(47436,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(47436,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(47436,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(47436,36,'cp -22','d5f6 f3e2 c7c5 g1f3 a7a6 c2c3 b7b5 c4b3 e7e6 e1g1 b8d7 d2d4 c8b7 a2a4 f8e7 c1f4 e8g8 f1d1 f6h5 f4c1 c5c4 b3c2 d8b6 c1g5 e7g5 f3g5 h5f6 b1d2 h7h6 g5f3 b7d5 d1e1 b6b7 d2e4 f6e4 c2e4 d5e4 e2e4 b7e4');
INSERT INTO evaluations VALUES(47330,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(47330,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(47330,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(47330,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(47330,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(47330,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(47330,36,'cp -17','f3e2 c7c5 g1f3 a7a6 c2c3 b8c6 e1g1 c8g4 h2h3 g4f3 e2f3 c6e5 f3e2 e5c4 e2c4 e7e6 a2a4 f8d6 d2d3 d8c7 b1d2 a8d8 d2e4 f6e4 c4e4 e8g8 c1e3 e6e5 e4c4 d8d7 f1d1 f8d8 c4e4 d6f8 a4a5 h7h6 g1f1 f8e7 f1e2 g8h8 g2g3');
INSERT INTO evaluations VALUES(49155,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(49155,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(49155,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(49155,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(49155,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(49155,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(49155,36,'cp 40','c7c5 c2c3 b8c6 d2f3 d8b6 a2a3 f8e7 h2h4 a7a5 b2b3 b6c7 c1e3 c7b6 f1d3 c5d4 f3d4 d7c5 g1e2 c6d4 e2d4 e8g8 d3c2 b6c7 h4h5 c8d7 h5h6 g7g6 e1g1 f7f5 e5f6 e7f6 d1f3 c5e4 c2e4 d5e4 f3e4 c7c3 d4e6');
INSERT INTO evaluations VALUES(43990,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(43990,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(43990,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(43990,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(43990,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(43990,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(43990,36,'cp 72','c2c3 b8c6 d2f3 d8b6 a2a3 c5d4 c3d4 f8e7 h2h4 c6a5 g1e2 b6b3 d1d2 a5c4 d2c3 b3c3 e2c3 d7b6 e1d1 c4a5 d1c2 h7h5 b2b3 a5c6 c1d2 c8d7 c2b2 a7a5 a3a4 e7b4 c3b5 e8g8 f1e2 b4d2 f3d2 a8b8 a1c1 c6b4 b5d6 b6c8 c1c7 c8d6 e5d6');
INSERT INTO evaluations VALUES(43988,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(43988,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(43988,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(43988,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(43988,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(43988,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(43988,36,'cp 66','c5d4 c3d4 b8c6 d2f3 d8b6 a2a3 f8e7 h2h4 c6a5 g1e2 b6b3 d1d2 a5c4 d2c3 b3c3 e2c3 d7b6 e1d1 c4a5 d1c2 h7h5 b2b3 a5c6 c1d2 a7a5 a3a4 c8d7 c2b2 e8g8 f1e2 e7b4 h1g1 f7f6 a1c1 a8c8 e2d3');
INSERT INTO evaluations VALUES(52395,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52395,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52395,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52395,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52395,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(52395,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(52395,36,'cp 12','e2e3 d7d5 d2d4 c8f5 g1f3 b8d7 f1d3 e7e6 e1g1 g8f6 c2c4 f6e4 d1c2 f8d6 a2a4 f5g6 c1a3 d6a3 a1a3 a7a5 b1c3 e4c3 c2c3 g6d3 c3d3 e8g8 a3a1 f8e8 d3c2 e6e5 f3e5 d7e5 d4e5 e8e5 f1d1 d8e7');
INSERT INTO evaluations VALUES(52393,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(52393,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(52393,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(52393,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(52393,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(52393,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52393,36,'cp 0','d7d5 g1f3 c8g4 c1b2 b8d7 f1e2 e7e6 d2d3 g8f6 b1d2 f8d6 h2h3 g4f3 d2f3 e8g8 e1g1 d8e7 c2c4 e6e5 a1c1 f8e8 f3h4 e7e6 h4f3');
INSERT INTO evaluations VALUES(52827,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52827,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52827,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52827,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(52827,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52827,36,'cp 0','g1f3 c8g4 c1b2 b8d7 f1e2 g8f6 d2d3 e7e6 b1d2 a7a5 c2c4 g4h5 e1g1 f8e7 a2a3 e8g8 d3d4 f6e4 d2e4 d5e4 f3d2 h5e2 d1e2 f7f5 f2f3 e4f3 d2f3 d7f6 f3g5 d8d7 g5f3');
INSERT INTO evaluations VALUES(53272,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53272,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53272,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53272,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(53272,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53272,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53272,36,'cp 0','c5d4 f1g2 g8f6 e1g1 h7h6 b2b3 e7e6 c1b2 f8c5 f3d4 e8g8 c2c4 e6e5 d4f3 e5e4 f3d4 f8e8 c4d5 d8d5 d4c2 d5e6 b1d2 b8d7 b3b4 c5b6 c2d4 e6d5 d4c2');
INSERT INTO evaluations VALUES(53400,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53400,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53400,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53400,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53400,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53400,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53400,36,'cp 0','f1g2 g8f6 e1g1 h7h6 b2b3 e7e6 c1b2 f8c5 f3d4 e6e5 d4f3 e5e4 f3d4 b8c6 d4c6 b7c6 c2c4 e8g8 b1c3 e4e3 f2f4 d5c4 d1d8 f8d8 c3a4 c5e7 g2c6 a8b8 b2e5 b8b4 e5c3 b4b8');
INSERT INTO evaluations VALUES(18983,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(18983,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(18983,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(18983,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(18983,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(18983,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18983,36,'cp 10','b1c3 d7d6 g1e2 a7a6 g2g3 a8b8 f1g2 b7b5 e1g1 f8e7 b2b3 g8f6 c1d2 e8g8 c3d5 b5c4 b3c4 f6d5 e4d5 c6d4 e2d4 e5d4 g2e4 f7f5 e4g2 e7g5 f2f4 g5e7 a1b1 c8d7 b1b8 d8b8 d1c2 b8c7 f1b1 f8b8 b1b8 c7b8 h2h3 b8c7 c2b2 e7f6');
INSERT INTO evaluations VALUES(28740,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28740,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(28740,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(28740,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(28740,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(28740,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28740,36,'cp 8','f6h5 f4g5 f8e7 h2h4 c7c5 g5e7 d8e7 g2g4 h5f6 g4g5 f6e4 d2e4 b7e4 f1d3 e4b7 d4c5 e7c5 a2a3 b8c6 h1h3 e8c8 d1e2 c8b8 e1c1 e6e5 f3d2 d7d5 d2b3 c5e7 f2f4 h7h6 d3b5 h6g5 f4g5 e7c7 h4h5 c6e7 h5h6');
INSERT INTO evaluations VALUES(30564,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(30564,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(30564,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(30564,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(30564,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(30564,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(30564,36,'cp 128','d4d5 g8f6 f3d4 e6e5 f1b5 f6d7 b5c6 d8c8 c6b7 c8b7 d4f5 h7h5 f5e3 d7f6 d1e2 a7a6 a2a4 b8d7 h2h4 g7g6 e3c4 f8h6 c1h6 h8h6 f2f3 h6h8 e1c1 e8f8 c1b1 d7c5 g2g4 b6b5 c4a5 b7b6 a5c6 b5a4 b1a1 f6d7');
INSERT INTO evaluations VALUES(30558,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(30558,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(30558,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(30558,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(30558,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(30558,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(30558,36,'cp 137','g8f6 f3d4 e6e5 f1b5 f6d7 b5c6 d8c8 c6b7 c8b7 d4f5 h7h5 a2a4 a7a6 f5e3 d7f6 f2f3 g7g6 e3c4 b8d7 h2h4 f8h6 c1h6 h8h6 d1d2 h6h8 b2b3 b7a7 c4e3 d7c5 c3d1 e8f8 d1b2 c5d7 a4a5 b6b5 b3b4 a8e8 c2c4');
INSERT INTO evaluations VALUES(41766,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41766,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41766,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(41766,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(41766,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(41766,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(41766,36,'cp 14','b5d6 c7d6 e2e3 b8c6 g1f3 e8g8 c2c3 h7h6 h2h3 f8e8 f1e2 e6e5 f4h2 c8f5 e1g1 a7a6 a1c1 f5g6 f1e1 d8c7 f3d2 b7b5 a2a4 c7b6 b2b4 e5d4 a4a5 b6d8 e3d4 a8c8 e2f1 c6e7 d1b3 d8d7 b3a3 f6e4 d2e4 d5e4');
INSERT INTO evaluations VALUES(41677,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(41677,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41677,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41677,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41677,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41677,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(41677,36,'cp 17','c7d6 c2c3 b8c6 g1f3 d8b6 f4d6 c6d4 d1d4 b6d6 e2e4 d6b6 d4b6 a7b6 e4d5 f6d5 f1c4 d5f6 f3e5 e8e7 e1c1 f6d7 e5f3 d7f6 h1e1 c8d7 f3d4 d7a4 d1d2 h8c8 b2b3 a4c6 f2f3 c6d5 c4d5 f6d5 c3c4 d5b4 c1b1 c8d8 e1d1 g7g5 d4e2 d8d2 d1d2 h7h6');
INSERT INTO evaluations VALUES(40582,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40582,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(40582,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40582,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40582,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(40582,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(40582,36,'cp 11','c2c3 b8c6 g1f3 e8g8 e2e3 f8e8 f1e2 h7h6 h2h3 e6e5 f4h2 c8f5 e1g1 a8c8 a1c1 f5g6 b2b4 a7a6 b4b5 a6b5 e2b5 g6h5 g2g4 h5g6 b5c6 b7c6 d4e5 d6e5 f3e5 d8a5 c3c4 a5a2 c4d5 c6d5 c1c8 e8c8 e5g6 f7g6');
INSERT INTO evaluations VALUES(55216,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(55216,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(55216,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(55216,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(55216,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(55216,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(55216,36,'cp 29','g1f3 c7c5 f1c4 a7a6 e1g1 g8f6 d2d4 e7e6 b1d2 c5d4 f3d4 f8d6 b2b3 e8g8 c4e2 e6e5 d4f3 b8c6 c1b2 d6c7 d2c4 e5e4 f3d4 c6d4 d1d4 c8e6 f1d1 d8d4 b2d4 f6d5 f2f3 e4f3 e2f3 f8d8 a2a4 a8c8');
INSERT INTO evaluations VALUES(19489,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(19489,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19489,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19489,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19489,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19489,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(19489,36,'cp -1','e7e5 c2c4 g7g6 b1c3 f8g7 g2g3 g8e7 f1g2 d7d6 g1e2 c8e6 e1g1 e8g8 c3d5 d8d7 e2c3 e6f5 d2d3 e7d5 c3d5 a8e8 b2c3 c6e7 b3b4 e7d5 g2d5 f5h3 f1e1 h3e6 a1b1 b7b6 b4c5 d6c5 e3e4 f7f5 f2f3 h7h6 d1d2 e8d8 a2a3');
INSERT INTO evaluations VALUES(19028,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19028,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(19028,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(19028,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(19028,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19028,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19028,36,'cp 6','c2c4 g7g6 b1c3 f8g7 g2g3 g8e7 f1g2 e8g8 d2d3 d7d6 g1e2 a8b8 e1g1 c8e6 c3d5 b7b5 d1d2 a7a5 a1c1 b5b4 f1d1 a5a4 d3d4 e5d4 e3d4 e6d5 g2d5 e7d5 c4d5 c6d4 b2d4 c5d4 e2d4 g7d4 d2d4 a4b3 a2b3');
INSERT INTO evaluations VALUES(22074,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(22074,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(22074,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(22074,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(22074,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(22074,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(22074,36,'cp 74','g8f6 a2a3 f8e7 f1d3 e6e5 d4d5 c6b8 c3a4 e8g8 c2c4 c8g4 a4c3 b8d7 h2h3 g4h5 c1e3 c7c6 a1c1 d7c5 e3c5 d6c5 g2g4 h5g6 d1e2 e7d6 f3d2 h7h5 a3a4 a7a6 c1d1 d8e7 a4a5 a8e8 d3c2 e7c7 c2a4 c7a5 d5c6 b7c6 a4c6');
INSERT INTO evaluations VALUES(26527,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(26527,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(26527,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(26527,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(26527,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(26527,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(26527,36,'cp 43','e7e6 a2a3 g8f6 b2b4 f6e4 c1b2 f8e7 b1c3 e8g8 c4c5 e7f6 f1e2 c6b8 c3e4 f5e4 e1g1 c7c6 a1c1 h7h6 f3d2 e4h7 b2c3 d8c7 a3a4 e6e5 d4e5 f6e5 c3e5 c7e5 d2f3 e5f6 f3d4 b8d7 e2d3 h7d3 d1d3 a7a6 d3c2 f6e7');
INSERT INTO evaluations VALUES(26802,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(26802,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(26802,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(26802,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(26802,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(26802,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(26802,36,'cp 51','a2a3 f8e7 b2b4 g8f6 c1b2 f6e4 b1d2 e4d2 d1d2 e8g8 c4c5 e7f6 f1e2 c6e7 e1g1 c7c6 a3a4 a7a6 b4b5 e7g6 b2c3 h7h6 d2b2 d8e7 a1a3 a6b5 a4b5 a8a3 b2a3 e6e5 b5c6 b7c6 d4e5 g6e5 f3e5 f6e5 c3e5 e7e5');
INSERT INTO evaluations VALUES(34066,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(34066,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(34066,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(34066,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(34066,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(34066,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(34066,36,'cp 1','d7d5 b1c3 e4c3 b2c3 b8d7 c1f4 d7e5 e2e5 e7e5 f4e5 c8f5 h2h4 f7f6 e5f4 f8a3 e1d2 e8c8 f1e2 h8e8 a1e1 e8e7 e2f3 c7c6 e1e7 a3e7 f3e2 d8e8 g2g4 f5d7 f2f3 e7d8 e2d3 h7h6 a2a4 a7a5 c3c4 d5c4 d3c4 b7b6 f4d6');
INSERT INTO evaluations VALUES(33844,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(33844,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33844,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33844,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(33844,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(33844,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(33844,36,'cp 4','b1c3 e4c3 b2c3 b8d7 c1f4 d7e5 e2e5 e7e5 f4e5 c8f5 h2h4 f7f6 e5f4 e8c8 e1d2 f8a3 f1e2 h7h5 h1e1 c7c6 a1b1 b7b6 e2a6 c8d7 c3c4 d5c4 a6c4 a3d6 f4d6 d7d6 e1e3 d8e8 g2g3 b6b5 c4d3 f5d3 d2d3 g7g5 e3f3 h8h6 c2c4 g5h4');
INSERT INTO evaluations VALUES(34330,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(34330,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(34330,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(34330,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(34330,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(34330,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(34330,36,'cp 63','g1f3 g8f6 c1d2 a5c7 g2g3 b8d7 d1e2 e7e6 f1g2 f8d6 f3g5 e8g8 h2h4 a7a5 e1c1 d6b4 a2a3 b4c3 d2c3 h7h6 g5e4 f6d5 c3d2 b7b5 d1e1 a8b8 g3g4 b5b4 a3a4 f7f5 e4g5 d7c5 d3d4 c5a4 g5e6 c8e6 e2e6 f8f7');
INSERT INTO evaluations VALUES(34329,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(34329,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(34329,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(34329,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(34329,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(34329,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(34329,36,'cp 51','g8f6 c1d2 a5c7 g2g3 e7e6 d1e2 b8d7 f1g2 f8d6 e1c1 e8g8 h2h4 a7a5 f3g5 f6d5 c3e4 d6b4 c2c4 d5e7 d3d4 b4d2 e2d2 b7b6 g3g4 c8a6 g2f1 a6b7 c4c5 b6c5 d4c5 e7d5 e4d6 h7h6 d2c2 d7f6 g5e4 f6e4');
INSERT INTO evaluations VALUES(33595,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(33595,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(33595,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(33595,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(33595,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(33595,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(33595,36,'cp 59','c1d2 g7g6 g2g3 f8g7 f1g2 e8g8 h2h3 a5c7 e1g1 e7e5 f1e1 b8d7 b2b4 f8e8 a2a4 d7f8 d1c1 f6d5 c3e4 f7f5 e4c5 b7b6 c5b3 a7a5 b4a5 b6a5 d2h6 g7f6 c1d2 f8e6 a1b1 e5e4 d3e4 d5c3 e4f5 c3b1 e1b1 g6f5 f3d4 e8d8 h6e3 f5f4 g3f4 e6d4 b3d4');
INSERT INTO evaluations VALUES(33596,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(33596,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(33596,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(33596,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(33596,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(33596,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(33596,36,'cp 29','a5b6 h2h3 b6b2 a1b1 b2a3 d3d4 a3d6 f1d3 g7g6 d1e2 f8g7 e1g1 e8g8 f1e1 f6d5 a2a4 d5c3 d2c3 b8d7 c3b4 c6c5 b4a3 f8e8 d4c5 d6c7 e2e3 d7f6 d3c4 c8d7 f3e5 e7e6 c4b5 a7a6 e5d7 f6d7');
INSERT INTO evaluations VALUES(39287,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(39287,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(39287,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(39287,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(39287,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(39287,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(39287,36,'cp 153','c4f7 e8f7 d1h5 f7f8 h5c5 d7d6 c5c4 c3b2 c1b2 d8e7 g1e2 b8c6 e1g1 g8f6 b1d2 c8d7 f2f4 a8e8 e2g3 h7h5 e4e5 d6e5 a1e1 d7e6 c4c1 f8g8 f4e5 f6d7 d2e4 d7e5 g3f5 e6f5 f1f5 h8h6 e4g5 e7d6 g1h1 h5h4 f5f4 d6g6');
INSERT INTO evaluations VALUES(38974,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(38974,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(38974,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(38974,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(38974,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(38974,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(38974,36,'cp 150','e8f7 d1d5 f7e8 d5h5 e8f8 h5c5 d7d6 c5c4 c3b2 c1b2 g8f6 g1f3 d6d5 c4b3 b8a6 e1g1 d5e4 b1c3 d8e7 a1e1 a6c5 b3c2 c5d3 e1e3 d3b2 c2b2 c8f5 f3h4 f5g6 b2b7 a8d8 h2h3 c7c5 b7b3 e7c7 f1e1 c5c4 b3a4 f8g8 h4g6 h7g6 c3e4 f6e4 e3e4 h8h5 e4c4 c7d7 a4c2 d7f5 e1e4 g8h7 c4c7 f5d5 c7a7 h5e5 e4e5 d5e5');
INSERT INTO evaluations VALUES(37489,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(37489,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(37489,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(37489,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(37489,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(37489,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(37489,36,'cp 146','d1d5 f7f8 d5c5 d7d6 c5c4 c3b2 c1b2 g8f6 g1f3 d6d5 c4b3 b8a6 e1g1 d5e4 f3e5 d8e7 b1d2 a6c5 b3e3 b7b6 a1c1 c8a6 f1e1 a8e8 e3g3 a6c8 c1c5 b6c5 d2e4 c8f5 e4f6 g7f6 e5g6 h7g6 e1e7 e8e7 g3c3 e7f7 f2f3 f8g7 g2g4 f5e6 g4g5 h8d8 g5f6 g7h7');
INSERT INTO evaluations VALUES(37490,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(37490,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(37490,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(37490,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(37490,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(37490,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(37490,36,'cp 146','f7f8 d5c5 d7d6 c5c4 c3b2 c1b2 b8c6 g1f3 d8e8 e1g1 c8e6 c4c2 e8g6 b1d2 g8f6 a1e1 a8e8 f3h4 g6h5 h4f5 f8f7 d2f3 h8f8 b2f6 e6f5 e4f5 g7f6 c2b3 f7g7 b3b7 c6e5 f3e5 d6e5 b7c7 g7h8 e1e3 h5f7 c7c2 e8c8 c2b1 c8b8 e3b3 b8b3 a2b3 f8d8 f1d1 d8d4');
INSERT INTO evaluations VALUES(44577,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(44577,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44577,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(44577,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44577,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44577,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44577,36,'cp 2','b1c3 e7e6 g1f3 f8e7 e4e5 f6d5 c3d5 e6d5 d2d4 d7d6 c1e3 d6e5 d4e5 b8c6 f1d3 d8b6 a1b1 c6b4 e1g1 d5d4 e3f2 b4d3 d1d3 b6g6 d3b5 g6c6 b5c6 b7c6 f2h4 c8f5 f1f2 a7a5 h2h3 e7h4 f3h4 f5e4 b1e1 e4d5 f4f5 d5a2');
INSERT INTO evaluations VALUES(44576,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44576,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44576,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(44576,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(44576,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(44576,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44576,36,'cp 13','e7e6 g1f3 b8c6 f1b5 d8c7 e4e5 f6g8 d2d4 c6d4 f3d4 c5d4 d1d4 a7a6 b5e2 f8c5 d4d3 d7d5 e5d6 c5d6 c1e3 g8f6 h1f1 e8g8 e1c1 f8d8 g2g4 b7b5 g4g5 f6e8 e2f3 a8b8 c1b1 d6e7 d3e2 c8b7 c3e4 d8d5 d1d5 b7d5 f4f5 e6f5');
INSERT INTO evaluations VALUES(44251,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(44251,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44251,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(44251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44251,36,'cp 1','g1f3 b8c6 f1b5 c6d4 e4e5 d4b5 c3b5 a7a6 b5c3 f6d5 c3d5 e6d5 d2d4 d7d6 c1e3 b7b6 e1g1 f8e7 d1d3 e8g8 c2c3 d6e5 f4e5 c5c4 d3d2 b6b5 f1f2 b5b4 c3b4 d8b6 a2a3 h7h6 h2h3 c8f5 f3h4 f5e4 h4f5 a8e8 f5g3 e4d3 a1e1 b6g6 g1h2');
INSERT INTO evaluations VALUES(44250,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(44250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44250,36,'cp 0','b8c6 f1b5 d8b6 d2d3 d7d5 a2a4 f8e7 e1g1 e8g8 b5c6 b6c6 f3e5 c6b6 h2h3 c5c4 g1h2 c4d3 c2d3 f8d8 d1e2 d5e4 d3e4 c8d7 c1e3 b6b4 e3d2 b4b6');
INSERT INTO evaluations VALUES(12443,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12443,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(12443,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(12443,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12443,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(12443,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(12443,36,'cp 0','f1b5 c6d4 e4e5 d4b5 c3b5 a7a6 b5c3 f6d5 c3d5 e6d5 d2d4 d7d6 c1e3 d6e5 f4e5 c5d4 d1d4 f8e7 e1g1 e8g8 d4f4 d8c7 c2c3 f7f6 e5f6 c7f4 e3f4 e7c5 g1h1 f8f6 f4e5 f6f8 a1d1 c8e6 f1e1 a8e8 e5d4 c5d6 d4e5');
INSERT INTO evaluations VALUES(12324,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12324,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12324,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12324,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(12324,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12324,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12324,36,'cp 0','c6d4 e4e5 d4b5 c3b5 a7a6 b5c3 f6d5 c3d5 e6d5 d2d4 d7d6 c1e3 b7b6 e1g1 f8e7 c2c4 d5c4 e5d6 d8d6 d4c5 d6d1 a1d1 b6c5 f3e5 e8g8 e5c6 e7f6 e3c5 f8e8 c5d4 c8g4 d1d2 f6d8 h2h3 g4f5 f1c1 d8c7 d4e5 c7b6 e5d4');
INSERT INTO evaluations VALUES(52151,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52151,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(52151,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52151,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(52151,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(52151,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(52151,36,'cp 3','g1f3 b8c6 c3d4 c5d4 f3d4 f8b4 b1c3 g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 f1d1 b4e7 e2f4 e7f6 d3e4 c8g4 f2f3 d5f4 f3g4 f4e2 g1f2 e2c1 a1c1 f6b2 c1c6 f8d8 e4d5 b2e5 f2f3 e5h2 c6c1 a8c8 d5b3 h2e5 c1c8 d8c8');
INSERT INTO evaluations VALUES(52150,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(52150,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(52150,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(52150,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(52150,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(52150,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(52150,36,'cp 5','b8c6 c3d4 c5d4 f3d4 g8f6 b1c3 f8b4 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 b4d6 e2d4 c8d7 d4f3 a7a5 f1d1 f8b8 d3c4 d7e8 b2b3 a5a4 c1b2 d6e7 b2e5 b8b7 d1d5 c6d5 c4d5 a8a7 d5b7 a7b7 e5c3');
INSERT INTO evaluations VALUES(52939,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(52939,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(52939,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52939,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52939,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52939,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52939,36,'cp 0','a4c4 b8d7 g1f3 e7e5 f3e5 d7e5 d4e5 d8a5 b1c3 a5e5 g2g3 c8e6 c4a4 f8c5 f1g2 g8f6 c1f4 e5d4 e1g1 d4a4 c3a4 c5e7 f1d1 f6d5 g2d5 e6d5 a4c3 d5e6 c3e4 e6d5');
INSERT INTO evaluations VALUES(52877,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52877,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52877,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52877,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52877,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52877,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52877,36,'cp 0','b8d7 g1f3 e7e5 d4e5 d7e5 f3e5 d8a5 b1c3 a5e5 g2g3 c8e6 c4a4 f8c5 f1g2 g8f6 c1f4 e5d4 e1g1 d4a4 c3a4 c5e7 f1d1 f6d5 g2d5 e6d5 a4c3 d5e6 c3e4 e6d5');
INSERT INTO evaluations VALUES(54654,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(54654,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(54654,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(54654,33,'cp 243',NULL);
INSERT INTO evaluations VALUES(54654,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(54654,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(54654,36,'cp 234','f1c4 g7g6 e1g1 a7a5 c1e3 b7b5 c4e2 f8h6 d1c1 h6e3 c1e3 d8b6 a2a4 b5b4 c3b1 g8e7 b1d2 e8g8 d2c4 b6c7 d4e5 d6e5 c4b6 a8a7 f1d1 b8d7 b6d7 c8d7 f3d2 d7e6 e2c4 e6c4 d2c4 a7a8 h2h4 c6c5 d1d6 a8d8 a1d1 e7c8 d6d8 f8d8 d1d8 c7d8 e3c5');
INSERT INTO evaluations VALUES(54652,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(54652,31,'cp 247',NULL);
INSERT INTO evaluations VALUES(54652,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(54652,33,'cp 249',NULL);
INSERT INTO evaluations VALUES(54652,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(54652,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(54652,36,'cp 252','g7g6 c1e3 g8h6 d1e2 d8e7 e1c1 h6g4 e3d2 f8g7 c1b1 g4h6 d2e3 h6f7 d4e5 d6e5 f3d2 c8e6 h2h4 b8d7 d2b3 e8g8 h4h5 e6c4 e2c4 f8d8 h5g6 h7g6 a2a4 d7b6 d1d8 e7d8 c4e2 b6d7 a4a5 b7b5 c3a2 d8c7 a2b4 a8e8 b4a6 c7c8 a6c5 d7c5');
INSERT INTO evaluations VALUES(5972,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5972,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5972,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5972,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5972,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5972,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5972,36,'cp 16','g8f6 d2d3 a7a5 e1g1 h7h6 b1d2 c5a7 f1e1 e8g8 d2f1 c8e6 c4b3 e6b3 d1b3 a5a4 b3c2 d6d5 a1b1 f8e8 c1e3 d5d4 c3d4 c6d4 e3d4 a7d4 f3d4 d8d4 f1e3 a8d8 c2c7 d8d7 c7c2 d4d3 c2a4 d3e4 a4e4 f6e4');
INSERT INTO evaluations VALUES(9338,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9338,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9338,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9338,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9338,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9338,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9338,36,'cp 0','d2d3 a7a5 e1g1 h7h6 c4b5 c5a7 b1d2 e8g8 f1e1 f6h7 d2f1 f7f5 e4f5 c8f5 f1g3 f5g6 c1e3 a7e3 e1e3 h7g5 b5c4 g8h8 c4d5 g5f3 d5f3 f8f4 g3e2 f4f8');
INSERT INTO evaluations VALUES(22530,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(22530,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(22530,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(22530,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(22530,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(22530,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(22530,36,'cp 21','f8b4 d1b3 b4c3 b3c3 g8f6 e4e5 f6e4 c3e3 d7d5 e5d6 c8f5 f3d2 e8g8 d2e4 f5e4 e1g1 e4g6 f1d1 c7d6 e3g3 g8h7 g3d6 g6h5 d6d8 a8d8 d1d8 f8d8 c1e3 c6d4 a1c1 d4e2 c4e2 h5e2 f2f3 e2a6 c1c7 d8d3 g1f2 f7f6 g2g4 d3d5 e3a7 d5d2 f2g3 d2b2 a7d4 b2a2 d4f6');
INSERT INTO evaluations VALUES(6278,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(6278,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(6278,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(6278,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(6278,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(6278,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(6278,36,'cp 9','d1b3 b4c3 b3c3 g8f6 e4e5 f6e4 c3e3 d7d5 e5d6 c8f5 f3d2 e8g8 d2e4 f5e4 e1g1 e4g6 f1d1 c7d6 e3g3 g8h7 d1d6 d8e7 c1f4 f8d8 a1d1 d8d6 f4d6 e7f6 g3c3 f6c3 b2c3 a8c8 h2h4 h7g8 g2g4 h6h5 f2f3 h5g4 f3g4 c6d8 c4b3 c8c3 d6b4 c3d3 d1d3 g6d3 g1f2 d8e6 f2e3');
INSERT INTO evaluations VALUES(6276,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(6276,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(6276,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6276,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(6276,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6276,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6276,36,'cp 8','b4c3 b3c3 g8f6 e4e5 f6e4 c3e3 d7d5 e5d6 c8f5 f3d2 e8g8 d2e4 f5e4 e1g1 e4g6 f1d1 c7d6 e3g3 g8h7 g3d6 c6e5 c4f1 d8d6 d1d6 f8d8 d6d8 a8d8 c1f4 f7f6 f4e3 a7a6 h2h3 d8d7 f2f4 e5c6 a2a3 c6d4 a1c1 d4c2 g1f2 c2e3 f2e3 d7e7 e3f3 g6e4 f3f2 e7d7 f2e3 e4c6 g2g4 g7g5 f1e2 h7g6 f4g5 f6g5');
INSERT INTO evaluations VALUES(29766,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29766,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(29766,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(29766,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29766,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29766,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29766,36,'cp 0','h2h3 g4f5 d1e2 b8c6 g2g4 f5g6 c1g5 d8d6 e1c1 f6d5 c3b5 d6d7 c2c4 d5b4 b2b3 a7a6 b5c3 e8c8 d4d5 f7f6 d5c6 d7c6 d1d8 c8d8 g5e3 b4d3 c1d1 d3b2 d1c1');
INSERT INTO evaluations VALUES(29763,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29763,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29763,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29763,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29763,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29763,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29763,36,'cp -6','g4f5 g2g4 f5g6 d1e2 b8c6 c1g5 c6d4 f3d4 d8d4 e2b5 f6d7 a1d1 d4e5 b5e5 d7e5 g5f4 f7f6 c3b5 a8c8 f4e5 f6e5 f1c4 g6c2 d1c1 c2a4 c4e6 a4b5 e6c8 e7e6 c8b7 e8d7 b7f3 a7a5 e1f2 f8d6 f2e3 h8b8 h1d1 d7e7 d1d2 b5a6 f3e4 h7h6 e4d3 a6b7 d3e4 b7e4 e3e4 b8b4 e4e3 e5e4 d2f2 g7g6 c1f1 a5a4');
INSERT INTO evaluations VALUES(38346,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(38346,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(38346,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(38346,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(38346,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(38346,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(38346,36,'cp -10','e7e6 d2d4 g8f6 b1d2 e8g8 c2c3 b8d7 e1g1 d6d5 e4d5 e6d5 c4d3 c7c5 f1e1 c5d4 c3d4 f8e8 g5f3 e8e1 d1e1 d7f8 d2b3 f8e6 c1e3 f6e4 f3e5 c8d7 d3e4 d5e4 a1d1 d8e7 e5d7 e7d7 b3c5 d7d5 c5e6 f7e6 e1e2');
INSERT INTO evaluations VALUES(38546,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(38546,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(38546,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(38546,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(38546,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(38546,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(38546,36,'cp -6','d2d4 g8f6 b1d2 e8g8 c2c3 b8d7 e1g1 d6d5 e4d5 e6d5 c4d3 c7c5 f1e1 c5d4 c3d4 d7b8 h2h3 b8c6 d2b3 c6b4 d3b1 f8e8 c1e3 h7h6 g5f3 f6e4 f3e5 e4d6 d1f3 b4c6 e5c6 b7c6 b1d3 c8f5 b3c5');
INSERT INTO evaluations VALUES(38547,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(38547,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(38547,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(38547,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(38547,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(38547,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(38547,36,'cp -11','g8f6 c4d3 e8g8 e1g1 e6e5 d4e5 d6e5 g5f3 d8e7 f1e1 f8d8 d1e2 h7h6 h2h3 b8c6 c2c3 f6h5 c1e3 h5f4 e3f4 e5f4 b1d2 d8e8 d3c4 c8d7 c4d5 a8d8 a1d1 a7a6 d2b3 g6g5 e2c2 h6h5 d5c6 d7c6 d1d8 e7d8 b3d4 g5g4 d4c6');
INSERT INTO evaluations VALUES(41607,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(41607,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(41607,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41607,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(41607,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41607,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(41607,36,'cp -1','b1c3 a7a6 c4d3 h7h6 g5f3 b8d7 d1e2 e8g8 e4e5 d6e5 d4e5 f6d5 h2h4 d5c3 b2c3 d7c5 e1g1 c5d3 c2d3 b7b6 d3d4 c8b7 c1a3 f8e8 h4h5 g6g5 f3h2 a6a5 f1d1 d8d5 f2f3 b7a6 e2e3 d5c4 d1d2 a6b7 h2g4 a8d8 g4f2 c4d5');
INSERT INTO evaluations VALUES(47180,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(47180,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(47180,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(47180,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(47180,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(47180,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(47180,36,'cp -88','c7c5 d4c5 b8c6 a2a3 d5d4 c3e4 f6e4 d3e4 g7g6 e4d3 c8f5 e2e4 d4e3 d3d8 a8d8 c2c3 e3f2 e1f2 e7e5 c1e3 f5e6 g1f3 f7f5 g2g3 f8g7 f1b5 h7h6 h1d1 g6g5 d1d8 e8d8 a1e1 e6d7 b5c4 d8c8 b2b4 e5e4 e3d4 g7d4 f3d4 c6e5 c4e2 h8f8 d4b5 e5g4 f2g1 d7b5 e2b5');
INSERT INTO evaluations VALUES(48325,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(48325,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(48325,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(48325,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(48325,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(48325,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(48325,36,'cp -44','c7c5 e2e3 f8e7 g1e2 e8g8 e1g1 b8c6 d4c5 e7c5 e3e4 f6e4 c3e4 d5e4 g2e4 e6e5 d1d8 f8d8 c1e3 c5e3 f2e3 c8g4 e2c3 d8d7 c3d5 g4h3 f1d1 a8d8 d5c3 h3g4 d1d7 d8d7 a1f1 h7h6 e4d5 g8f8 f1f2 c6b4 e3e4 f7f6 f2d2 g4f3 d2f2 b4d5');
INSERT INTO evaluations VALUES(46230,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(46230,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(46230,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(46230,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(46230,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(46230,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(46230,36,'cp -40','g1f3 c5d4 f3d4 f8b4 e1g1 e8g8 c1d2 b8c6 d4c6 b7c6 c3a4 b4d6 c2c4 d8e7 a1c1 a8b8 a2a3 c8a6 c4d5 c6d5 b2b4 f8c8 f1e1 h7h6 d2e3 a6b5 a4c3 b5c4 c3a4 e7c7 e3d4 e6e5 d4c5 d6c5 a4c5 a7a5 d1d2 a5b4');
INSERT INTO evaluations VALUES(46229,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(46229,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(46229,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(46229,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(46229,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(46229,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(46229,36,'cp -37','c5d4 f3d4 f8b4 e1g1 e8g8 c1d2 b8c6 d4c6 b7c6 e2e4 c8a6 f1e1 f8e8 c3a4 b4f8 e4e5 f6d7 b2b3 d8c7 d2f4 a8b8 c2c3 a6b5 a4b2 f8a3 a1b1 d7f8 d1c2 f8g6 c3c4 d5c4 b2c4 g6f4 g3f4 a3b4 e1d1 e8d8 f4f5 b5c4');
INSERT INTO evaluations VALUES(52303,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(52303,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(52303,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52303,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(52303,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52303,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(52303,36,'cp -10','f8g7 c2c3 g8f6 e4e5 f6d5 d1b3 e7e6 d3d4 b8c6 c1g5 d8a5 g5d2 a5b6 b3b6 a7b6 f1c4 e8g8 e1g1 c5d4 c3d4 d7d6 f1e1 d6e5 d4e5 c6b4 e1c1 c8d7 b1c3 d7c6 a2a3 d5c3 a3b4 c3a4 b2b3 a4b2 c4e2 a8a1 c1a1 c6f3 e2f3 g7e5 f3b7');
INSERT INTO evaluations VALUES(56463,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(56463,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(56463,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(56463,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(56463,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(56463,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(56463,36,'cp -75','f8d6 c2c4 c7c6 d3d4 e5e4 b1c3 g8f6 d1c2 a7a6 c4c5 d6c7 c1d2 b8d7 e1c1 b7b5 c5b6 d7b6 f2f3 e8g8 f3e4 d5e4 c3e4 f6e4 c2e4 f8e8 e4c2 d8d5 f1d3 d5a2 g1e2 h7h6 c2b1 a2b3 b1c2 c8e6 c2b3 e6b3 d1e1 e8d8 e3e4 a6a5');
INSERT INTO evaluations VALUES(1433,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(1433,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(1433,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(1433,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(1433,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(1433,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(1433,36,'cp -103','b7c6 c2c3 c8a6 a2a3 g8e7 b2b4 c5d4 c3d4 e7f5 b1c3 h7h5 c3a4 b6b8 c1d2 a6b5 a4c3 b5c4 c3e2 b8b5 e2g3 c4d3 g3f5 d3f5 f3g1 f5d3 g1e2 f8e7 e2f4 a7a5 b4a5 e8g8 f4d3 b5d3 d2b4 e7b4 a3b4 d3c3 e1e2 a8b8 d1d3 c3b2 e2f3 b8b4 h1c1');
INSERT INTO evaluations VALUES(1311,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(1311,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(1311,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(1311,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(1311,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(1311,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(1311,36,'cp -96','c2c3 c8a6 a2a3 g8e7 b2b4 c5d4 c3d4 e7f5 b1c3 h7h5 c3a4 b6b7 d1c2 a6c4 a4b2 c4b5 c1d2 f8e7 h2h3 h5h4 a3a4 b5a6 b2d3 a6d3 c2d3 e7b4 e1g1 b4d2 d3d2 e8g8 f1c1 a8b8 d2d3 a7a5 a1b1 b7c7 c1c5 b8b4 b1c1 f8c8 d3a6 f5d4 f3d4 b4d4 c5c6');
INSERT INTO evaluations VALUES(1308,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(1308,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(1308,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(1308,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(1308,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(1308,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(1308,36,'cp -104','c8a6 a2a3 g8e7 b2b4 c5d4 c3d4 e7f5 b1c3 h7h5 c3a4 b6b7 d1c2 a6c4 c1d2 b7a6 d2e3 a8c8 e1d2 c4e2 h1c1 f8e7 f3g1 e2c4 g1f3 e8g8 a4b2 c4b5 b2d3 a6b6 d3c5 b6c7 d2e1 a7a5 c2d2 b5c4 e3g5 a5b4 a3b4 e7c5 d4c5 c8a8 a1a8');
INSERT INTO evaluations VALUES(27500,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(27500,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(27500,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(27500,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(27500,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(27500,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(27500,36,'cp -100','a2a3 g8e7 b2b4 c5d4 c3d4 e7f5 b1c3 h7h5 c3a4 b6b7 d1c2 a6c4 a4b2 c4b5 c1d2 a8c8 a1c1 h5h4 a3a4 b5a6 b2d3 b7b6 d3c5 a6c4 d2c3 a7a5 b4a5 b6a7 c5d3 c6c5 d4c5 f8c5 e1g1 c5a3 c3b4 a3b4 d3b4 a7a5 c2d2');
INSERT INTO evaluations VALUES(22410,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(22410,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22410,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(22410,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(22410,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(22410,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(22410,36,'cp 81','e2e4 g7g6 f1e2 g8f6 b1c3 e7e5 g1f3 f8g7 e1g1 e8g8 c1e3 f8e8 d4d5 f6g4 e3g5 f7f6 g5d2 a7a5 a2a3 f6f5 f3g5 d7c5 e4f5 g6f5 e2g4 f5g4 f2f3 h7h6 g5e4 c5e4 f3e4 c8d7 d1c2 b7b6 d2e3 d8e7 c3b5 a8c8 b5a7');
INSERT INTO evaluations VALUES(22411,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(22411,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(22411,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(22411,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(22411,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(22411,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(22411,36,'cp 77','g8f6 d1c2 g7g6 b1c3 e7e5 g1f3 f8g7 f1e2 e8g8 e1g1 e5d4 f3d4 f8e8 c1e3 a7a6 a1d1 c7c6 f1e1 d8c7 d4b3 b7b5 c4b5 c6b5 c2d2 g7f8 d1c1 c7b8 b3a5 c8b7 a5b7 b8b7 e2f3 d7e5 e3g5 e8e6 e1d1 e5f3 g2f3 f8g7 c1c2 b5b4 c3d5 f6d5 d2d5 b7d5');
INSERT INTO evaluations VALUES(52285,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(52285,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(52285,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(52285,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(52285,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(52285,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(52285,36,'cp 67','e2e4 d7d6 d2d4 c5d4 f3d4 b8c6 f1e2 c6d4 d1d4 g8f6 e1g1 f8g7 b1c3 e8g8 d4d3 a7a5 c1e3 a5a4 e3d4 d8a5 b2b3 c8d7 a1b1 a4b3 a2b3 a5g5 f2f4 g5h4 d3d2 d7c6 d4f2 h4h6 e2f3 f6d7 f2e3 g7c3 d2c3 h6g7 e3d4 e7e5 f4e5 d7e5 c3d2 f7f6');
INSERT INTO evaluations VALUES(1589,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(1589,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(1589,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(1589,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(1589,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(1589,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(1589,36,'cp 160','d2d4 e6d7 d4d5 c6d8 c1f4 g8f6 f4e5 c7c6 e5f6 e7f6 d1d4 c6d5 c3d5 d8e6 d4e4 f8d6 e2c3 e8g8 f1d3 e6g5 e4e2 f8e8 d5e3 d7e7 h2h4 g5e6 e3f5 e7d8 f5d6 d8d6 e1c1 d6f4 e2d2 f4d2 d1d2 g7g6 d3e4 g8g7 e4f3 e8d8 h1d1 d8d2 d1d2 a8b8 c3b5 a7a6 b5d6 b7b5 b2b4 f6f5');
INSERT INTO evaluations VALUES(1580,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(1580,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(1580,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(1580,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(1580,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(1580,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(1580,36,'cp 120','e6d7 d4d5 c6d8 c1f4 g8f6 f4e5 c7c6 e5f6 e7f6 d1d4 c6d5 c3d5 d8e6 d4e4 f8d6 e2c3 e8g8 f1d3 e6g5 e4e2 f8e8 d5e3 e8d8 e1c1 d7e8 h1e1 d6b4 e3d5 e8e2 d3e2 b4c3 d5c3 c8d7 f2f4 g5e6 f4f5 e6f4 e2f3 d7f5 e1e7 d8d1 c3d1 a8b8 d1e3 f5e6 f3b7 g8f8 e7c7 f4e2 c1d2');
INSERT INTO evaluations VALUES(1925,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(1925,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(1925,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(1925,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(1925,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(1925,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(1925,36,'cp 39','b1c3 f7f6 a2a3 f8c5 d2d4 e5d4 f3d4 c5d4 d1d4 g8e7 e1g1 e8g8 c1e3 d7d6 f2f3 c8e6 e3f2 a7a6 f1e1 a8e8 a1d1 e7g6 c3e2 f6f5 e4f5 e6f5 c2c4 f5e6 c4c5 d6c5 d4c5 c6c5 f2c5 f8f5 c5f2 f5d5 e2c3 d5d1 e1d1 h7h6 f2g3');
INSERT INTO evaluations VALUES(1923,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(1923,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(1923,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(1923,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(1923,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(1923,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(1923,36,'cp 31','f7f6 a2a3 f8c5 d2d4 c5d4 f3d4 e5d4 d1d4 g8e7 e1g1 e8g8 f1e1 d7d6 h2h3 a7a6 c1f4 c8e6 a1d1 e7g6 f4g3 a8e8 d4d2 f6f5 e4f5 e6f5 c3d5 c6c4 d5e3 c4f7 e3f5 f7f5 d2c3 f5f7 b2b3 h7h6 c3d2 f7f5 e1e8 f8e8 f2f3 f5f7');
INSERT INTO evaluations VALUES(2047,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(2047,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(2047,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(2047,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(2047,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(2047,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(2047,36,'cp -60','g8f6 h5h4 c6d4 f3d4 e5d4 e1g1 f6e4 h4f4 c7c6 d2d3 e4c5 f4d4 a7a5 a2a3 d7d5 c4a2 c5e6 d4e5 e7d6 e5d6 f8d6 c1e3 e8g8 b1d2 f8e8 f1e1 g8f8 g2g3 a5a4 d2f3 e6c7 c2c4 c8h3 c4d5 c7d5 a2d5 c6d5 f3d4 a8c8 f2f3 h3d7 a1d1 f7f6 g1f2');
INSERT INTO evaluations VALUES(1096,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(1096,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(1096,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(1096,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(1096,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(1096,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(1096,36,'cp -86','h5h4 d7d5 e4d5 g7g5 f3g5 c6d4 d5d6 c7d6 b1a3 d6d5 c2c3 d5c4 c3d4 h8g8 e1g1 h7h6 d4e5 h6g5 e5f6 g5h4 f6e7 f8e7 a3c4 c8h3 g2g3 h3f1 g1f1 e8c8 b2b3 b7b5 c4e3 d8d7 d2d4 h4g3 h2g3 d7d4 e3f5 d4d7 c1b2 e7g5 a1e1 c8b7 b2c3 d7d3');
INSERT INTO evaluations VALUES(1097,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(1097,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(1097,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(1097,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(1097,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(1097,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(1097,36,'cp -99','d7d5 e4d5 g7g5 f3g5 c6d4 d5d6 c7d6 c4f7 e8d8 f7b3 d6d5 b1c3 h8g8 d2d3 h7h6 e1f1 a7a5 g5f3 d4b3 a2b3 g8g4 c3d5 f6d5 h4e7 d8e7 f3e5 g4g8 c1d2 e7f6 d3d4 d5b4 d2b4 f8b4 f2f3 c8f5 c2c3 b4d6 e5c4 d6c7 f1f2 b7b5 c4d2 a8e8 g2g3 f6f7 h1e1 f5c2 b3b4 e8e1 f2e1 a5b4 a1a7 g8e8 d2e4 e8e7 e1f2 b4b3');
INSERT INTO evaluations VALUES(6236,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(6236,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(6236,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(6236,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(6236,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(6236,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(6236,36,'cp 32','f3e5 c6e5 d3d4 c5d6 d4e5 d6e5 f2f4 e5c3 b2c3 d7d6 c1a3 g8f6 e4e5 c8g4 d1d4 f6d7 f1d3 d8h4 g2g3 h4h3 d4e4 e8c8 d3f1 g4f5 e4c4 f5e6 f1h3 e6c4 e1c1 d6d5 h1e1 h8e8 f4f5 f7f6 e5f6 e8e1 d1e1 d7f6 e1e6 d5d4 c3d4 d8d4 e6e7 d4d7 e7d7 f6d7 a3b2 d7f6 a2a4 c7c6 h3g2 c4e2');
INSERT INTO evaluations VALUES(16323,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(16323,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(16323,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(16323,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(16323,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(16323,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(16323,36,'cp 55','f3e5 f8e7 d2d4 e8g8 b5a4 c6a5 b1d2 d7d6 e5f3 c8d7 c2c3 f8e8 a4c2 e7f8 e1e8 d8e8 d2f1 h7h6 b2b3 d7e6 c1f4 a5c6 f1g3 c6e7 c3c4 e7g6 f4e3 c7c6 h2h3 e8d7 d1d3 a8e8 a1e1 b7b6 c2b1 d7c7 d3d2 c7d7 d2c1 d6d5 c4c5');
INSERT INTO evaluations VALUES(16301,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(16301,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(16301,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(16301,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(16301,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(16301,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(16301,36,'cp 45','f8e7 d2d4 e8g8 h2h3 a7a6 b5f1 d7d5 e5c6 b7c6 b1d2 f8e8 c2c4 c6c5 d4c5 e7c5 d2b3 c5f8 c1d2 d5c4 f1c4 c8d7 d1c2 e8e1 a1e1 d7b5 d2a5 d8d7 e1d1 d7c6 d1c1 b5c4 c2c4 c6c4 c1c4 f6d5 g1f1 g7g6 b3c1 a8e8 c1d3 f8d6 b2b3');
INSERT INTO evaluations VALUES(21344,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(21344,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(21344,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(21344,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(21344,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(21344,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(21344,36,'cp 41','g8f6 e1g1 f8d6 c2c4 c8d7 c4c5 d6e7 a2a3 e8g8 b5d3 a7a5 b1c3 b7b6 c5b6 c7b6 f1e1 d8b8 h2h3 f8d8 c1d2 h7h6 a1b1 d7e8 b1c1 e7d6 d1e2 c6e7 e1d1 b8b7 c3b5 f6e4 d2e1 d8c8 f3d2 e8b5 d3b5 c8c1 d1c1 a8c8 b5a6');
INSERT INTO evaluations VALUES(22092,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(22092,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(22092,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(22092,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(22092,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(22092,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(22092,36,'cp 90','b4a6 f1b5 c8d7 b5d7 d8d7 c2c4 g8f6 b1c3 c7c6 c1g5 f8e7 e1g1 h7h6 g5f6 e7f6 g2g3 e8g8 g1g2 f8c8 f1e1 a6c7 d1a4 c7e8 e1d1 g7g6 a1c1 d7e7 a4c2 h6h5 a3a4 f6g7 a4a5 a8b8 a5a6 b7a6 c4c5 g7h6 c5d6 e8d6');
INSERT INTO evaluations VALUES(22204,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(22204,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(22204,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(22204,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(22204,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(22204,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(22204,36,'cp 101','f1b5 c8d7 b5d7 d8d7 c2c4 c7c6 b1c3 g8f6 b2b4 f8e7 e1g1 c6d5 c4d5 e8g8 d1d3 a6c7 h2h3 f6e8 f3d2 f7f5 e4f5 d7f5 d2e4 f5g6 c1e3 e8f6 a3a4 a7a6 f1c1 c7e8 d3c4 h7h6 a1b1 f6e4 c4e4 g6f7 e4d3 e7g5 c3e4 g5e3 f2e3 f7d7 a4a5');
INSERT INTO evaluations VALUES(22200,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(22200,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(22200,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(22200,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(22200,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(22200,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22200,36,'cp 88','c8d7 b5d7 d8d7 c2c4 g8f6 b1c3 c7c6 b2b4 f8e7 h2h3 e8g8 c1e3 a6c7 a3a4 c6d5 c4d5 f8c8 d1b3 b7b5 e1g1 h7h6 f3d2 a7a6 f1c1 e7d8 a4a5 c7e8 c3e2 c8c1 a1c1 a8c8 b3d3 c8c1 e2c1 d7c8 c1b3 f6d7 d2f3 e8f6 e3d2');
INSERT INTO evaluations VALUES(22867,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22867,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(22867,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22867,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22867,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22867,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(22867,36,'cp 0','c2c4 g8f6 g1f3 d7d5 c4d5 f6d5 f1e2 f8d6 b1c3 c8e6 e3e4 d5e7 d3d4 c6d4 f3d4 e5d4 d1d4 e8g8 c1e3 e7g6 g2g3 d8e7 e1c1 f8d8 d4a4 g6e5 c3d5 e6d5 e4d5 d6c5 a4e4 d8e8 h1e1 c5b4 e1f1 b4c5');
INSERT INTO evaluations VALUES(27493,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(27493,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(27493,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(27493,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(27493,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(27493,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(27493,36,'cp -69','f7f6 h2h3 g4d7 e2f1 g8e7 b1c3 e8f7 d2d3 e7g6 a2a4 a6a5 b2b3 f8b4 c1d2 h8d8 c3e4 b4f8 e1e2 b7b6 a1e1 d7c8 d2e3 f8b4 e3d2 h7h5 d2b4 a5b4 e2e3 c6c5 h3h4 c8g4 g2g3 d8d7 f1g1 g6f8 e4d2 g4f5 d2c4 f8e6 f3d2 e6d4');
INSERT INTO evaluations VALUES(31914,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31914,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(31914,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(31914,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31914,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(31914,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(31914,36,'cp 38','b8d7 e2e4 d5e4 d2e4 e7e6 h2h3 g4f5 f1d3 g8f6 d1e2 f6e4 d3e4 f5e4 e2e4 c7c6 e1g1 f8d6 c2c4 h7h6 c1d2 e8g8 f1e1 d7f6 e4e2 a7a5 a2a3 d8c7 a1d1 a8d8 b2b4 a5b4 a3b4 d8d7 g2g3 d6e7 d2c3 f8d8');
INSERT INTO evaluations VALUES(36735,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(36735,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(36735,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(36735,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(36735,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(36735,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(36735,36,'cp -27','c7c5 h2h3 c5d4 e3d4 b8c6 e1g1 b7b6 c4d5 c6b4 f1e1 c8b7 c1g5 a8c8 a2a3 b4d3 d1d3 b7d5 c3d5 f6d5 a1c1 c8c1 e1c1 f7f6 g5d2 d8d7 d3b3 f8d8 h3h4 e7e6 b3d3 d8c8 c1c8 d7c8 g2g3 c8d7 g3g4 d5e7 d2b4 e7c6 d3e4');
INSERT INTO evaluations VALUES(36030,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(36030,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(36030,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(36030,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(36030,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(36030,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(36030,36,'cp -13','e1g1 c5d4 e3d4 b8c6 h2h3 b7b6 c4d5 c6b4 f1e1 c8b7 c1g5 h7h6 g5f6 e7f6 d3c4 b4d5 c3d5 b7d5 c4d5 d8d5 d1b3 d5b3 a2b3 f8c8 e1c1 a7a5 c1c4 c8d8 g2g3 d8d5 a1c1 d5b5 c1c3 g7f8 c4c7 a8d8 c7b7 b5b4 g1f1 g8g7 c3c6 b4b3 b7b6');
INSERT INTO evaluations VALUES(45260,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(45260,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(45260,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(45260,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45260,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(45260,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(45260,36,'cp 0','d1d4 b8c6 d4d3 g7g6 f1e2 f8g7 e1g1 e8g8 f1d1 b7b6 c1g5 h7h6 g5h4 c8b7 b1a3 d8c8 d3e3 g6g5 h4g3 c8e6 f3d2 f6g4 e3d3 g4f6');
INSERT INTO evaluations VALUES(51890,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(51890,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(51890,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(51890,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(51890,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(51890,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(51890,36,'cp -35','a7a5 b4b5 f8c5 e2e3 d8f6 c2c3 b7b6 b1d2 c8b7 g1f3 d7d6 c1b2 a5a4 a1c1 b8d7 f1d3 g8e7 d3e4 d6d5 e4d3 e8g8 e1g1 a4a3 b2a1 f6h6 d3b1 a8c8 c3c4 f8d8 d1e2 e7g6 h2h3 c5e7 c4d5 b7d5 f3d4 d7c5 f2f4 e7f6 d2c4 c5a4 c4a3 c8c1 f1c1');
INSERT INTO evaluations VALUES(50462,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(50462,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(50462,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(50462,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(50462,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(50462,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(50462,36,'cp -39','b4b5 f8c5 e2e3 d8f6 c2c3 b7b6 c1b2 c8b7 g1f3 d7d6 b1d2 a5a4 a1c1 f6g6 d2c4 e6e5 b2a3 c5a3 c4a3 g8e7 a3c4 e7c8 d1d2 e8g8 c1d1 e5e4 f3d4 d6d5 c4e5 g6d6 f2f4 e4f3 e5f3 d6c5 f1d3 c8d6 e1g1 b8d7 d4f5 d6f5 d3f5 d7f6 d2d4 a8e8');
INSERT INTO evaluations VALUES(52266,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(52266,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(52266,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(52266,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(52266,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(52266,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(52266,36,'cp 99','h5c5 b8c6 g1f3 g8f6 e4e5 d7d6 e5d6 e7d6 c5e3 c8e6 f1b5 f8g7 f3d4 f6g4 e3e4 d6d5 d4c6 d5e4 c6d8 e8d8 b1c3 f7f5 h2h3 g4e5 d2d4 e4d3 c1g5 d8c8 e1c1 d3c2 c1c2 a7a6 b5e2 h7h6 g5e3 c8c7 c3a4 a8e8 c2b1 e6d7 e3b6 c7c8 d1c1 c8b8 a4c5 d7c6 f2f4 e5f7 e2c4 f7d6 c4b3 c6e4 c5e4 d6e4 c1c7');
INSERT INTO evaluations VALUES(52262,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(52262,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(52262,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(52262,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(52262,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(52262,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(52262,36,'cp 77','b8c6 g1f3 g8f6 e4e5 d7d6 e5d6 e7d6 c5e3 f8e7 d2d4 e8g8 f1e2 c8f5 e3d2 d6d5 e1g1 c6b4 f3e1 d8c7 b1c3 b4c6 e1f3 a8d8 d2d1 f8e8 h2h3 c6b4 e2d3 f5d3 c2d3 b4c6 d1b3 c7d7 b3a4 f6h5 f3e5 d7c7 c1e3 h5g7 e5c6 c7c6 a1c1 c6a4 c3a4');
INSERT INTO evaluations VALUES(4215,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4215,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4215,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4215,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4215,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4215,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4215,36,'cp 0','d7d6 a2a4 f6d5 c4d5 d8f6 c2c3 a7a5 c1e3 c6e7 d5c4 e7g6 d3d4 e5d4 c3d4 c5b4 g1h1 c8e6 c4d3 c7c5 a1c1 e6d7 d3b1 a8d8 h2h3 f8e8 f3h2 c5d4 e3d4 f6f4 d4e3 f4f6');
INSERT INTO evaluations VALUES(3646,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3646,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3646,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3646,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3646,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3646,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3646,36,'cp 0','a2a4 f6d5 c4d5 d8f6 d5c6 b7c6 c1e3 c5b6 a4a5 b6e3 f2e3 c6c5 d1e1 f6e7 f3h4 a8b8 b2b3 e7g5 g1h1 f7f5 h4f5 c8f5 e4f5 f8f5 f1f5 g5f5 e3e4 f5d7 h1g1 d7b5 a1a4 a7a6 h2h3 b8f8 e1c3 b5d7 c3e1');
INSERT INTO evaluations VALUES(13059,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(13059,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(13059,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13059,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13059,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(13059,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(13059,36,'cp 25','e1g1 c8g4 b1d2 d8b6 a2a4 a7a6 b5c6 b6c6 h2h3 g4f3 d2f3 e7e6 c1f4 f8d6 f4d6 c6d6 a4a5 e8g8 c2c3 d6c7 f3e5 h7h6 d1e2 f6e8 g2g3 e8d6 h3h4 f8b8 a1a2 b8c8 f1e1 d6c4 e2f3 b7b5 a5b6 c4e5 e1e5 c7b6 g1g2');
INSERT INTO evaluations VALUES(13058,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(13058,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(13058,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(13058,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(13058,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(13058,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(13058,36,'cp 24','c8g4 h2h3 g4f3 d1f3 e7e6 c2c3 h7h6 c1e3 f8d6 b1d2 e8g8 f1e1 a8c8 b5f1 a7a6 g2g3 f6d7 a2a4 c6e7 h3h4 d8b6 a1b1 b6c7 b1c1 c7b6 b2b4 e7f5 d2b3 f5e3 e1e3 d7f6 b3c5 b6c7 c1e1 d6c5');
INSERT INTO evaluations VALUES(24605,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(24605,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24605,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24605,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(24605,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24605,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24605,36,'cp 19','h2h3 g4f3 d1f3 e7e6 c2c3 h7h6 c1e3 f8d6 b1d2 a7a6 b5d3 e8g8 f3e2 e6e5 d4e5 c6e5 d2f3 e5d3 e2d3 f8e8 f1e1 d8d7 e3d4 f6e4 a1d1 d6c7 d4e3 a8d8 c3c4 e4f6 c4c5 f6e4 b2b4 f7f5 e3d4 d7f7');
INSERT INTO evaluations VALUES(15094,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(15094,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(15094,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(15094,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(15094,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(15094,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(15094,36,'cp 156','e4d5 f6d5 e1g1 c8e6 c4b5 e5d4 f3d4 d8d7 b5c6 b7c6 d4e6 f7e6 b1d2 f8d6 d2e4 e6e5 f1e1 d5f6 e4f6 g7f6 d1e2 h8g8 c1e3 d7g4 f2f3 g4a4 e2d3 e8c8 d3h7 c8b7 h7e4 a4a5 g1h1 a5d5 a1d1 d5e6 e4d3 d6e7 d3b3 e6b3 a2b3');
INSERT INTO evaluations VALUES(14992,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(14992,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(14992,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(14992,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(14992,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(14992,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(14992,36,'cp 143','f6d5 e1g1 d5b6 c4b5 c8g4 b1c3 e5d4 f1e1 f8e7 d1d4 d8d4 f3d4 g4d7 d4c6 b7c6 b5a6 e8f8 c1e3 a8b8 a1d1 d7g4 d1d2 h7h5 h2h3 g4e6 c3e4 b6d5 e3d4 h8h6 e4c5 b8d8 a6c4 a7a5 d2d1 h5h4 a2a3 e7d6 c5b7 d8a8 d4c5 d6c5 b7c5 h6f6');
INSERT INTO evaluations VALUES(15331,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(15331,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(15331,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(15331,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(15331,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(15331,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(15331,36,'cp 122','f3e5 c6e5 d4e5 c8e6 e1g1 d8d7 b1d2 e8c8 d1e2 d5f4 e2e3 d7d4 c4e6 f4e6 d2f3 d4a4 b2b3 a4a6 e3e4 f8c5 f1e1 d8d7 a2a4 a6a5 c1e3 c5e3 e4e3 h8d8 b3b4 a5b6 e3e4 g7g6 a4a5 b6b5 a5a6 b7a6 e4a8 b5b8');
INSERT INTO evaluations VALUES(20316,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(20316,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(20316,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(20316,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(20316,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(20316,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(20316,36,'cp -93','c5d4 f3d4 g8e7 d4c6 e7c6 c4d5 e6d5 f1b5 d8b6 b5c6 b7c6 e1g1 f8e7 b1c3 d5d4 c3a4 b6d8 c1f4 c8e6 d1c2 a8c8 a4c5 d8d5 b2b4 a7a5 a2a3 e7c5 c2c5 d5c5 b4c5 e8g8 f4d2 a5a4 f2f3 e6b3 a1c1 f8d8 f1e1');
INSERT INTO evaluations VALUES(20407,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(20407,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(20407,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(20407,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(20407,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(20407,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(20407,36,'cp -66','c1f4 h7h6 f3d4 f8b4 b1c3 g8e7 c4d5 d8d5 d4c6 b4c3 b2c3 d5c6 d1f3 c6f3 g2f3 b7b6 f4e3 c8b7 f1b5 e8f8 e1e2 e7g6 h1g1 g6e5 f3f4 e5f3 g1g3 f3h4 a1d1 h4f5 g3g1 b7e4 d1d7 f5e3 e2e3 e4f5 b5d3 f5d3 e3d3 g7g5 d3e4 g5f4 e4f4 h8g8 g1g8 f8g8');
INSERT INTO evaluations VALUES(22010,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(22010,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22010,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(22010,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(22010,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(22010,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(22010,36,'cp -9','b1c3 g8f6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7e6 c4c7 c8d7 c7e5 e8g8 c1e3 f8c8 e5e6 d7e6 f3d4 a7a6 h2h3 e6d7 d4e2 b7b5 e2f4 d7c6 a2a3 c8e8 a1c1 a8c8 c3d5 f6d5 f4d5 c6d5 c1c8 e8c8');
INSERT INTO evaluations VALUES(26662,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(26662,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(26662,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(26662,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(26662,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(26662,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(26662,36,'cp 135','d1b3 e7e6 b3b7 c6b4 c4b5 e8e7 b1a3 a8b8 b7a7 e7f6 b5e2 g7g6 a7a4 b8a8 a4b3 f6g7 e1f1 a8b8 b3d1 h7h5 g1f3 g8f6 h2h4 f8d6 f1g1 d8e7 c1d2 b4d3 e2d3 f5d3 d2c3 d3a6 d1c2 f6d5 c3d2 c7c5 d4c5 d6c5 e3e4 h8c8 e4d5 c5f2 g1f2 c8c2 a3c2 e7c5 c2e3');
INSERT INTO evaluations VALUES(31437,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(31437,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(31437,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(31437,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(31437,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(31437,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(31437,36,'cp 55','d1b3 d8b6 c4c5 b6b3 a2b3 b8d7 b3b4 a7a6 h2h3 a8c8 g1f3 f7f6 g2g4 f5g6 h1g1 e8f7 e2e3 h7h5 f1e2 h5g4 h3g4 c8e8 e1c1 f8e7 g1h1 h8h1 d1h1 e7d8 f4d6 g8e7 f3d2 e7c8 d6g3 e6e5 d2b3 f7g8 f2f4 e5d4 e3d4 c8e7 f4f5 g6f7 e2d3');
INSERT INTO evaluations VALUES(39154,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(39154,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(39154,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(39154,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(39154,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(39154,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(39154,36,'cp 115','c2c3 b4e7 e2e4 d7d5 e4e5 c7c5 f1d3 b8c6 e1g1 a7a6 f1e1 c8d7 h2h3 c5d4 f3d4 g8h6 d4c6 d7c6 c1h6 g7h6 b1d2 d5d4 d3e4 d4c3 e4c6 b7c6 b2c3 e7g5 d2e4 d8d1 a1d1 e8g8 c3c4 f8b8 d1d6 b8b2 d6c6 b2a2 c4c5 g5e7');
INSERT INTO evaluations VALUES(39152,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(39152,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(39152,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(39152,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(39152,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(39152,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(39152,36,'cp 121','b4e7 e2e4 c7c5 f1d3 d7d5 e4e5 b8c6 e1g1 c8d7 d4c5 e7c5 b2b4 c5b6 f1e1 g8e7 b1d2 h7h6 d2b3 e8g8 a2a4 e7g6 c1a3 f8e8 d3c2 b6c7 c2g6 f7g6 b3c5 d8c8 c3c4 c7b6 c4d5 e6d5 b4b5 b6c5');
INSERT INTO evaluations VALUES(41770,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41770,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(41770,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41770,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41770,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41770,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41770,36,'cp 12','f4d6 d8d6 g1f3 e8g8 e2e3 d5c4 f1c4 c7c5 e1g1 b8c6 d4c5 d6c5 c4d3 f8d8 d1e2 c8d7 a1c1 a8c8 c3e4 c5e7 e4f6 e7f6 f1d1 h7h6 h2h3 c6e5 d3e4 d7c6 d1d8 f6d8 e2c2 e5f3 e4f3 c6d7 c2b1 c8c1 b1c1 d8c8 c1c8 d7c8 b2b4 e6e5 a2a3');
INSERT INTO evaluations VALUES(41672,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(41672,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41672,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41672,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41672,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(41672,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(41672,36,'cp 9','d8d6 g1f3 e8g8 e2e3 d5c4 f1c4 c7c5 e1g1 b8c6 d4c5 d6c5 c4d3 f8d8 a1c1 c5d6 d3e2 d6e7 d1a4 h7h6 f1d1 c8d7 a2a3 a8c8 c3b5 d7e8 d1d8 c8d8 b5c3 d8c8 h2h3 a7a6 c3e4 f6d5 b2b4 c6a7 c1c8 a7c8 a4c2 c8a7 c2b2');
INSERT INTO evaluations VALUES(33412,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(33412,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(33412,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33412,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(33412,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(33412,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(33412,36,'cp 11','e2e3 e8g8 g1f3 d5c4 f1c4 c7c5 e1g1 b8c6 d4c5 d6c5 c4d3 f8d8 a1c1 c5e7 d1e2 b7b6 a2a3 c8b7 f1d1 a8c8 d3a6 h7h6 h2h3 d8d1 c1d1 c8d8 d1d8 c6d8 a6b7 e7b7 b2b4 d8c6 e2d3 b7c8 d3c4 c8b7 c3b5 a7a6 b5d4 c6d4 c4d4');
INSERT INTO evaluations VALUES(42170,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(42170,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(42170,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(42170,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(42170,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(42170,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(42170,36,'cp 47','f1c4 e8g8 a2a3 b4c3 b2c3 b7b6 e1g1 c7c5 f1e1 c8b7 c4d3 b7e4 d3e4 f6e4 d1c2 e4f6 d4c5 b6c5 c3c4 b8c6 a1b1 d8c7 c1d2 h7h6 d2c3 a8b8 h2h3 b8b1 e1b1 f8d8 b1b5 f6d7 f3d2 d8b8 c2d3 c6e5 b5b8 d7b8 d3e4 b8c6');
INSERT INTO evaluations VALUES(46633,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(46633,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46633,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(46633,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(46633,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46633,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(46633,36,'cp 26','f8g7 e2e4 h7h6 g5f4 e8g8 d1d2 e7e5 d4e5 d6e5 f4h6 f6e4 d2d8 f8d8 h6g7 e4c3 g7f6 d8d6 f6e5 d6e6 f2f4 f7f6 b2c3 f6e5 e1c1 b8c6 f4e5 c6e5 g1f3 e5f3 g2f3 g8f7 d1d8 e6d6 d8d6 c7d6 h2h4 c8f5 h4h5 f7g7 h5g6 g7g6 h1h4 g6f6 c1d2 a8g8 d2e3 g8e8 e3f4 e8e5 h4h6 f6e7');
INSERT INTO evaluations VALUES(48969,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(48969,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48969,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(48969,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48969,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(48969,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48969,36,'cp 4','d8d5 d2d4 b8c6 c1e3 c8f5 d4e5 d5d1 e1d1 e8c8 d1c1 d8e8 f1b5 f5d7 c2c3 c6e5 b5d7 c8d7 f3e5 e8e5 h1d1 f8d6 a2a4 a7a5 d1d4 d7c8 d4c4 b7b6 c1c2 c8b7 g2g3 h8e8 a1d1 h7h5 b2b4 g7g5 c4d4 f7f6 h2h3 a5b4 c3b4 e5e4 d4e4 e8e4');
INSERT INTO evaluations VALUES(57353,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(57353,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(57353,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(57353,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(57353,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(57353,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(57353,36,'cp 55','g8f6 e2e3 f8e7 f1d3 e8g8 c2c4 b7b6 h2h3 f6e4 f3d2 e4d2 d1d2 c8b7 e1g1 d7d6 b1c3 b8d7 a1d1 e7h4 d3c2 d7f6 f4h2 g7g6 d4d5 e6e5 f2f4 e5f4 e3f4 f6h5 c2a4 d8e7 f1f3 e7g7 g1h1 h4f6 f3e3 g6g5 f4g5 f6g5');
INSERT INTO evaluations VALUES(16124,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(16124,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(16124,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(16124,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(16124,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(16124,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(16124,36,'cp -23','f6e4 c3e4 d7d5 c4d3 d5e4 d3e4 f8d6 e1g1 e8g8 d1e2 c8d7 d2d3 d8f6 f1e1 b7b6 b2b3 f8e8 c1b2 a8d8 a1d1 c6d4 f3d4 e5d4 e2f3 f6f3 e4f3 c7c5 a2a4 g7g5 g1f1 g8f8 e1e8 d8e8 b2c1 f7f5 c1d2 f8g7 f3d5 e8e7 d1e1 g7f6 e1e7 d6e7 d5c4 f5f4 c4b5 d7b5');
INSERT INTO evaluations VALUES(25391,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(25391,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(25391,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(25391,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(25391,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(25391,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(25391,36,'cp 44','h2h3 g4f3 d1f3 f6d7 b1d2 d7c5 d2c4 f8d6 c2c3 c5e6 a2a4 d8d7 e1g1 c6c5 f3h5 g7g6 h5h4 d6e7 h4g4 h7h5 g4g3 h5h4 g3g4 h8h5 g4f3 a8d8 f1d1 e8f8 a4a5 f8g7 c4e3 e7g5 e3d5 c7c6 c1g5 e6g5 f3f6 g7h7 a5a6 b7b5 d5e3 g5e6');
INSERT INTO evaluations VALUES(25389,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(25389,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(25389,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(25389,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(25389,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(25389,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25389,36,'cp 37','g4f3 d1f3 f6d7 b1d2 d7c5 d2c4 f8d6 a2a4 c5e6 c2c3 d8d7 e1g1 c6c5 f3h5 e8c8 c4e5 d6e5 h5e5 d7d3 c1e3 d3d6 e5d6 d8d6 f2f4 h8e8 f4f5 e6d8 e3c5 d6d2 f1f2 d2f2 g1f2 e8e4 c5a7 e4f4 f2e2 f4f5 a7e3 d8e6 a1f1 f5f1 e2f1 c8d7 f1e2 d7d6 e2f3');
INSERT INTO evaluations VALUES(25394,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(25394,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(25394,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(25394,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(25394,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(25394,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(25394,36,'cp 44','d1f3 f6d7 f3g3 d8e7 b1d2 f7f6 d2c4 e8c8 a2a4 h7h5 f2f4 e5f4 c1f4 d7b6 c4b6 a7b6 e1g1 e7c5 g3f2 f8e7 g2g3 e7d6 f2c5 d6c5 g1g2 c5d6 g2f3 d8e8 f3e3 c8d7 h3h4 e8e7 f1f3 h8e8 a1f1 e8a8 f4d6 c7d6 f3f5');
INSERT INTO evaluations VALUES(25393,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(25393,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(25393,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(25393,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(25393,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(25393,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(25393,36,'cp 50','f6d7 b1d2 d7c5 d2c4 f8d6 a2a4 c5e6 c2c3 d8d7 e1g1 f7f6 c1e3 c6c5 f1d1 e8g8 h3h4 a8d8 h4h5 b7b6 g2g3 d7f7 g1f1 d8d7 f3f5 d7d8 f1e2 f7d7 e3c1 d7f7 c4e3 d8d7 f5g4 f8d8 e3f5 g8h8 b2b3 d6f8 f5h4 a7a6');
INSERT INTO evaluations VALUES(40750,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(40750,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(40750,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(40750,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(40750,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(40750,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(40750,36,'cp 122','e5g3 d6d5 b1c3 b8c6 f1e2 c6d4 e2d1 d5e4 d3e4 e8g8 g1e2 f8e8 g3d3 d4e2 d1e2 e7d6 e1g1 f6g4 h2h3 d8h4 c3b5 d6h2 g1h1 e8d8 d3c3 c7c6 e2g4 c8g4 h1h2 c6b5 c3g3 h4g3 h2g3 g4e6 c1e3 a7a5 f1c1 h7h5 f2f3 a8c8 a2a4 b5b4');
INSERT INTO evaluations VALUES(40857,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(40857,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(40857,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(40857,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(40857,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(40857,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(40857,36,'cp 118','e8g8 f1e2 b8c6 c1h6 f6e8 g1f3 g8h8 h6d2 e8f6 h2h3 d6d5 e4d5 f6h5 g3h2 c6b4 b1a3 e7d6 g2g3 f8e8 f3g5 h8g8 g5e4 h5f6 h2g2 f6d5 e1g1 d6f8 c2c3 b4c6 a1e1 f8a3 b2a3 c8f5 e2d1 c6e5 g3g4 f5e4 e1e4 d5f6 d2g5 e5d3 e4e8 d8e8 g2b7');
INSERT INTO evaluations VALUES(49543,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(49543,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(49543,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(49543,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(49543,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(49543,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(49543,36,'cp -25','d2d4 e5d4 e3d4 d7d5 c4d3 f8d6 b1c3 e8g8 g1e2 f8e8 e1g1 c7c6 c1f4 c8g4 d1d2 g4h5 f1e1 h5g6 f4d6 d8d6 e2f4 b8d7 e1e8 a8e8 c3e2 d7f8 f2f3 e8e7 a1e1 f6e8 f4g6 f8g6 a2a4 d6f6 b2b3 e8d6 a4a5');
INSERT INTO evaluations VALUES(2052,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(2052,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(2052,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(2052,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(2052,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(2052,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(2052,36,'cp 59','g8f6 a2a4 d7d6 e1g1 c8e6 c4e6 e7e6 c2c4 f8e7 b1c3 a7a5 c3d5 e7d8 c1g5 e6d7 g5d2 e8g8 d2c3 c6b8 f1e1 c7c6 d5f6 d8f6 d1d2 d7c7 d3d4 f8e8 a1d1 b8d7 b2b3 e5d4 f3d4 c7b6 f2f3 h7h6');
INSERT INTO evaluations VALUES(1106,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(1106,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(1106,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(1106,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(1106,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(1106,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(1106,36,'cp 71','e1g1 d7d6 c2c3 h7h6 d3d4 f6e4 f1e1 e4g5 b1d2 g7g6 b2b4 f8g7 d1a4 e8g8 b4b5 c6b8 d4e5 b8d7 e5d6 g5f3 d2f3 e7d6 c1a3 d7c5 a4c2 c8f5 c2e2 b7b6 a1d1 d6f6 a3c5 b6c5 a2a4 a8d8 h2h3 d8d1 e1d1 c7c6 g2g4 c6b5 a4b5 f5e6 c4e6 f6e6 e2e6 f7e6');
INSERT INTO evaluations VALUES(1105,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(1105,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(1105,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(1105,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(1105,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(1105,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(1105,36,'cp 54','d7d6 c2c3 g7g6 f1e1 f8g7 h2h3 b7b6 d3d4 e8g8 b2b4 c8b7 d4d5 c6b8 c4f1 a7a5 a2a4 a5b4 c3b4 c7c6 b1c3 c6d5 e4d5 b8a6 f1a6 a8a6 c1g5 f8c8 a1c1 a6a8 d1b3 h7h6 g5f6 g7f6 f3d2 f6g5 c3e4 g5d2 e4d2 h6h5 c1c8 a8c8 d2e4');
INSERT INTO evaluations VALUES(12819,30,'cp 315',NULL);
INSERT INTO evaluations VALUES(12819,31,'cp 318',NULL);
INSERT INTO evaluations VALUES(12819,32,'cp 330',NULL);
INSERT INTO evaluations VALUES(12819,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(12819,34,'cp 328',NULL);
INSERT INTO evaluations VALUES(12819,35,'cp 354',NULL);
INSERT INTO evaluations VALUES(12819,36,'cp 343','e4e5 f6d5 c2c4 d5c7 d4d5 c6d4 f3d4 c5d4 d1d4 g7g6 e5e6 f7f6 e6d7 c8d7 c1e3 e7e5 d4d2 b7b6 b1c3 f8c5 e3c5 b6c5 f1e2 e8g8 d2e3 a8c8 e3c5 c7b5 c5e3 b5d4 e3d2 a7a5 e1g1 f6f5 g1h1 g8h8 f2f3 d8h4 g2g3 h4f6');
INSERT INTO evaluations VALUES(12768,30,'cp 354',NULL);
INSERT INTO evaluations VALUES(12768,31,'cp 348',NULL);
INSERT INTO evaluations VALUES(12768,32,'cp 344',NULL);
INSERT INTO evaluations VALUES(12768,33,'cp 348',NULL);
INSERT INTO evaluations VALUES(12768,34,'cp 341',NULL);
INSERT INTO evaluations VALUES(12768,35,'cp 339',NULL);
INSERT INTO evaluations VALUES(12768,36,'cp 337','f6d5 c2c4 d5c7 d4d5 c6d4 f3d4 c5d4 d1d4 g7g6 e5e6 f7f6 e6d7 c8d7 c1e3 e7e5 d4d2 b7b6 b1c3 a8c8 f1e2 f8b4 a1c1 e8g8 a2a3 b4d6 c3e4 c7e8 f2f3 f6f5 e4d6 e8d6 e3g5 d8e8 e1g1 f5f4 g5h6 f8f6 f1e1 d6f7');
INSERT INTO evaluations VALUES(18379,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18379,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18379,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18379,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18379,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18379,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18379,36,'cp 0','g8e7 e4e5 e7g6 b2b3 f7f6 c1b2 d7d6 e5d6 f8d6 b1a3 e8g8 a3c4 d6c7 b2a3 d8d5 d2d3 e6e5 c4e3 d5d6 e3c4');
INSERT INTO evaluations VALUES(25057,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(25057,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(25057,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(25057,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(25057,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(25057,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(25057,36,'cp 68','h2h3 g4h5 g2g4 h5g6 a2a4 d8d7 a4a5 a7a6 g1g2 f8e7 c1e3 g6h7 f1g1 c6d8 c4b3 d8e6 c3d5 e7d8 d1d2 c7c6 d5b6 d8b6 a5b6 g7g5 d2c3 d7e7 f3d2 e6f4 e3f4 g5f4 f2f3 f6d7 d2c4 e8g8 d3d4 g8g7 d4e5 d7e5 c4e5 d6e5 a1d1');
INSERT INTO evaluations VALUES(25056,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(25056,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(25056,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(25056,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(25056,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(25056,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(25056,36,'cp 61','g4e6 c4b5 f6d7 c1e3 a7a6 b5c6 b7c6 d3d4 d8f6 d4d5 c6d5 c3d5 f6d8 c2c4 f8e7 d1a4 d8b8 a4c6 e7d8 c4c5 e8g8 f1d1 d7f6 f3d2 e6c8 c5d6 c7d6 d5f6 d8f6 d2c4 c8b7 c6d6 b7e4 d6a3 b8b7 c4d6 f6e7 f2f3 e7d6 d1d6 e4c6 a1d1 a8e8 b2b4 c6b5 a3b3');
INSERT INTO evaluations VALUES(34718,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(34718,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(34718,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(34718,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(34718,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(34718,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(34718,36,'cp 65','g8f6 e1g1 g7g6 d3d4 f8g7 f1e1 e8g8 c2c3 a7a5 a2a4 c8e6 c4d3 b8a6 h2h3 f8e8 d3c2 c7c6 c1e3 e5d4 c3d4 a6b4 b1c3 b4c2 d1c2 e7d7 a1c1 d6d5 e4e5 e6f5 c2d1 f6e4 c3e4 d5e4 f3d2 f7f6 e5f6 g7f6 d2c4 f5e6');
INSERT INTO evaluations VALUES(33781,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(33781,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(33781,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(33781,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(33781,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(33781,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(33781,36,'cp 63','e1g1 g7g6 d3d4 f8g7 f1e1 e8g8 c2c3 a7a5 a2a4 c8e6 c4d3 f8e8 h2h3 c7c6 c1e3 b8a6 b1d2 e5d4 f3d4 a6c5 d3c2 e6d7 e3g5 b7b5 a4b5 c6b5 d2f3 e7f8 e4e5 f6h5 d4b3 c5e6 b3a5 e6g5 f3g5 e8e5 e1e5 g7e5 c2b3 g8g7 g5f7');
INSERT INTO evaluations VALUES(42581,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(42581,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(42581,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(42581,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(42581,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(42581,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(42581,36,'cp -38','c7c6 d1c2 b8a6 a2a3 d7d5 e4d5 c6d5 c1g5 a6c7 g1f3 e8g8 f1e2 c8f5 g5f6 d8f6 c4d5 c5b6 e1g1 a8d8 c2b3 f6d6 f3d2 c7d5 d2c4 d6e6 c4b6 e6b6 b3b6 d5b6 f1e1 f8e8 e2f3 e8e7 f3e4 f5e4 d3e4 g8f8 a1d1 e7d7 d1d7 d8d7 b2b3 f8e7 e1c1');
INSERT INTO evaluations VALUES(44128,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(44128,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(44128,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(44128,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(44128,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(44128,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(44128,36,'cp -23','f7e6 g2g3 d7d5 c4d5 d8d5 g1f3 d5d1 e1d1 b8c6 f1g2 f8e7 b1c3 c8d7 d1e1 a8d8 c1f4 f6d5 a1d1 d5c3 b2c3 e7f6 d1c1 b7b6 h2h4 e6e5 f4e3 c6e7 f3g5 h7h6 g5e4 d7e6 e3d2 e8g8 a2a4 e7f5 c1a1 f5d6 e4d6 d8d6');
INSERT INTO evaluations VALUES(44379,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(44379,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(44379,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(44379,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(44379,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(44379,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44379,36,'cp 0','c1g5 d7d5 e2e3 b8c6 g1f3 f8e7 f1e2 d5d4 e3d4 c5d4 e1g1 e6e5 e2d3 e8g8 b1d2 f6h5 g5e7 d8e7 f3d4 h5f4 d4c6 b7c6 d1c2 c8h3 d3h7 g8h8 h7e4 f4g2 g1h1 g2h4 f1g1 f8f2 c2c3 a8d8 c3h3 d8d2 h3c8 e7f8 c8f8 f2f8 a1f1 f8f4 f1f4 e5f4 e4c6 d2b2 g1g4 b2b1 g4g1');
INSERT INTO evaluations VALUES(44376,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(44376,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(44376,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(44376,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(44376,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(44376,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(44376,36,'cp 0','b8c6 e2e3 f8e7 g1f3 d7d5 f1e2 d5d4 e3d4 c5d4 e1g1 e8g8 e2d3 e6e5 b1d2 f6h5 g5e7 d8e7 f3d4 h5f4 d4c6 b7c6 d1c2 c8h3 d3h7 g8h8 h7e4 f4g2 g1h1 a8d8 c2c3 e7h4 e4g2 h3g2 h1g2 d8d2 c3d2 h4g4 g2h1 g4f3 h1g1 f3g4');
INSERT INTO evaluations VALUES(49796,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(49796,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(49796,32,'cp 238',NULL);
INSERT INTO evaluations VALUES(49796,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(49796,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(49796,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(49796,36,'cp 231','f3e5 b8c6 e5c6 d7c6 d2d4 d8h4 b1c3 f8b4 d1f3 f7f5 c1f4 c8e6 a2a3 b4c3 b2c3 e8c8 f3g3 h4g3 h2g3 f5e4 f4h6 g7h6 h1h6 d8d6 e1d2 h8f8 d2e3 e6g4 h6h4 d6g6 f1c4 h7h5 a3a4 b7b6 a1e1 c8b7 c4e2 c6c5 e2g4 h5g4 e1e2 f8e8 d4c5 g6c6 c5b6 c6c3 e3f4 a7b6 f4g4');
INSERT INTO evaluations VALUES(49788,30,'cp 239',NULL);
INSERT INTO evaluations VALUES(49788,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(49788,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(49788,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(49788,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(49788,35,'cp 246',NULL);
INSERT INTO evaluations VALUES(49788,36,'cp 244','b8c6 e5c6 d7c6 d2d4 d8e7 b1c3 c8e6 f1e2 e8c8 c1e3 e7e8 e1g1 f8d6 f1e1 c8b8 e2h5 e8f8 d1e2 f7f6 e3h6 g7h6 a1d1 h8g8 a2a3 f8g7 g2g3 e6h3 b2b4 g8f8 c3a4 f6f5 e4e5 d6e7 f2f4 h3g4 h5g4 f5g4');
INSERT INTO evaluations VALUES(18181,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(18181,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18181,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18181,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18181,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(18181,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(18181,36,'cp 46','d8c7 e4e5 f7f6 d2d4 c5d4 d1d4 c6c5 d4g4 f6e5 c3d5 c7d6 g4h5 g7g6 h5e5 d6e5 f3e5 a8a7 d5b6 a7c7 b6c8 c7c8 c1f4 f8h6 f4h6 g8h6 e1c1 h8f8 h1e1 d7d6 e5c4 e8d8 f2f3 h6f5 c2c3 c8b8 g2g3 f8f6 e1e4 h7h5 a2a3 g6g5 b2b4 c5b4 a3b4 h5h4');
INSERT INTO evaluations VALUES(31864,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(31864,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(31864,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(31864,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(31864,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31864,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(31864,36,'cp 0','e7e6 c2c4 c7c6 b1c3 d5c4 e2e4 f5g6 g1h3 f8b4 h3f4 g8f6 f4g6 h7g6 e1g1 b8d7 d1e2 b7b5 a2a4 a7a6 d4d5 e8g8 d5c6 d7e5 c1e3 b4c3 b2c3 e5c6 e3c5 f8e8 f1d1 d8c8 a1b1 c6e5 c5d4 e5c6');
INSERT INTO evaluations VALUES(32324,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32324,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32324,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32324,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(32324,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(32324,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(32324,36,'cp 12','c2c4 c7c6 b1c3 g8f6 g1f3 f8e7 d1b3 d8b6 c4c5 b6b3 a2b3 h7h6 b3b4 e8g8 e1g1 a7a6 a1a3 b8d7 c1e3 f6e4 f1d1 f8d8 c3e4 f5e4 f3d2 e4g2 g1g2 f7f5 d2b3 g7g5 h2h3 g8f7 e3d2 a8b8 f2f4 e7f6 g3g4 f5g4 h3g4 g5f4 d2f4');
INSERT INTO evaluations VALUES(44339,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(44339,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(44339,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(44339,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(44339,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(44339,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(44339,36,'cp 106','b1c3 d7d6 e2e4 g7g6 f2f4 f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5c4 d7b6 c4e2 c8g4 h2h3 g4f3 f1f3 e8g8 c1e3 b6d7 d1d2 f8e8 a1f1 f7f5 e4f5 g6f5 g1h1 d7f6 e3f2 f6e4 c3e4 f5e4 f3g3 g8h8 f2g1 e8g8 e2c4 d8f6 f4f5');
INSERT INTO evaluations VALUES(47654,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(47654,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(47654,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(47654,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(47654,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(47654,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(47654,36,'cp 124','c1f4 f8e7 d1d2 e8g8 e1c1 b8c6 f2f3 c6d4 d2d4 c8e6 g2g4 c7c6 h2h4 d8a5 g4g5 f6h5 g5h6 g7g6 f4d2 e7f6 d4f2 a5c5 f2c5 d6c5 f1h3 e6h3 h1h3 a8d8 c3e2 d8d7 f3f4 f8e8 e4e5 f6d8 d2e3 d7d1 c1d1 f7f6 e2g3 f6e5 g3h5 g6h5');
INSERT INTO evaluations VALUES(50251,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(50251,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(50251,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(50251,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(50251,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(50251,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(50251,36,'cp 6','c1b2 f8g7 g2g3 d7d6 f1g2 e7e5 e1g1 e8g8 c2c4 a7a5 d2d3 b8a6 d1c2 f8e8 a2a3 c8d7 f1e1 c7c6 b1d2 c6c5 d2b1 d7c6 b1c3 a6c7 f3d2 f6h5 e2e4 h5f6 a3a4 c7e6 d2f3 f6d7 g2h3 e6d4 c2d1 a8c8');
INSERT INTO evaluations VALUES(50628,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50628,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(50628,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50628,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50628,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(50628,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(50628,36,'cp -8','g1f3 b8c6 f1e2 g8f6 e4e5 f6g4 c1f4 d7d6 e1g1 g4e5 f3e5 d6e5 d1d8 e8d8 f1d1 c8d7 c3d5 a8c8 d5b6 c8c7 f4e3 d8e8 a1c1 d7f5 g2g4 f5g6 h2h4 h7h5 b6a8 c7d7 d1d7 e8d7 c1d1 c6d4 e3d4 e5d4 d1d4 d7c8 g4h5 g6h5 a8b6 c8c7 e2h5 c7b6 h5f7 e7e5 d4d8');
INSERT INTO evaluations VALUES(50627,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(50627,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50627,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(50627,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50627,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50627,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(50627,36,'cp -1','e7e6 c1f4 b8c6 f1e2 g8e7 a1c1 e7g6 f4e3 f8e7 h2h4 h7h5 e1g1 b7b5 a2a4 e8g8 a4b5 a6b5 e2b5 e7h4 f3h4 d8h4 f2f4 c8a6 b5a6 a8a6 d1d7 h4g4 f1f2 f8d8 d7b7 d8a8 b7d7 c6a5 c3a4 a6c6 c1c6 a5c6 d7c6 g4d1 f2f1 d1a4 c6a4');
INSERT INTO evaluations VALUES(52947,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(52947,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52947,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52947,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(52947,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(52947,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(52947,36,'cp 12','g8f6 e4e5 d6e5 f3e5 e7e6 e1g1 c6c5 c2c3 f8e7 d2d4 e8g8 d4c5 e7c5 c1f4 b8d7 d1e2 f6d5 f4g3 d7e5 e2e5 b7b6 f1d1 d8f6 e5f6 d5f6 b1d2 f8d8 c4e2 c8b7 g1f1 a7a5 d2b3 d8d1 a1d1 c5e7 b3d4 e7c5 a2a4 f6e4 g3f4 b7d5');
INSERT INTO evaluations VALUES(45089,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(45089,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(45089,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45089,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(45089,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45089,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45089,36,'cp 20','e4e5 d6e5 f3e5 e7e6 d2d4 c6c5 c2c3 c5d4 e1g1 f8d6 c3d4 b8c6 e5c6 b7c6 b1c3 e8g8 c1g5 h7h6 g5h4 d6e7 b2b3 d8d7 f1e1 f8d8 d1f3 d7d4 c3e4 f6e4 h4e7 d8e8 e7a3 e4d2 f3d3 e8d8 d3d4 d8d4 c4e2 c8b7 a1d1 d4d7 d1c1 a8d8 f2f3');
INSERT INTO evaluations VALUES(54773,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(54773,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(54773,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(54773,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(54773,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(54773,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(54773,36,'cp 81','d4e5 b8c6 f1b5 e7e6 d1h5 g7g6 h5h3 d8d7 g1f3 a7a6 b5c6 d7c6 c2c3 c6b5 f3d4 b5d7 e1g1 f8g7 f1e1 d7f7 b2b4 g8e7 c1g5 e8g8 f2f4 c8d7 b1d2 a6a5 b4b5 e7f5 a2a4 c7c5 b5c6 f5d4 c3d4 d7c6 d2f3 f7d7 g5f6 g7f6 e5f6 f8f6 f3e5 d7g7');
INSERT INTO evaluations VALUES(54770,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(54770,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(54770,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(54770,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(54770,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(54770,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(54770,36,'cp 52','b8c6 f1b5 e7e6 d1h5 g7g6 h5h3 d8d7 g1f3 a7a6 b5c6 d7c6 c2c3 c6b5 f3d4 b5d7 e1g1 f8g7 f1e1 d7f7 b2b4 g8e7 c1g5 h7h6 g5e7 f7e7 b1d2 b7b6 h3g3 g6g5 d2f3 c7c5 b4c5 b6c5 d4c2 e8g8 c2e3 c8b7 a1d1 d5d4 c3d4 b7f3 g2f3 c5d4 d1d4 e7f7 h2h4 f7f3');
INSERT INTO evaluations VALUES(20889,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(20889,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(20889,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(20889,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(20889,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(20889,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(20889,36,'cp 59','f1b5 e7e6 d1h5 g7g6 h5h3 d8d7 g1f3 a7a6 b5c6 d7c6 c2c3 c6b5 f3d4 b5d7 e1g1 f8g7 f1e1 d7f7 b2b4 g8e7 c1g5 h7h6 g5e7 f7e7 a2a4 b7b6 b1d2 e8g8 h3e3 c7c5 d4b3 c5b4 c3b4 e7g5 e3e2 a6a5 d2f3 g5f4 b3d4 a5b4 e2b2 g6g5 h2h3 c8d7');
INSERT INTO evaluations VALUES(2434,30,'cp -290',NULL);
INSERT INTO evaluations VALUES(2434,31,'cp -305',NULL);
INSERT INTO evaluations VALUES(2434,32,'cp -311',NULL);
INSERT INTO evaluations VALUES(2434,33,'cp -306',NULL);
INSERT INTO evaluations VALUES(2434,34,'cp -311',NULL);
INSERT INTO evaluations VALUES(2434,35,'cp -307',NULL);
INSERT INTO evaluations VALUES(2434,36,'cp -314','f7g8 d2d4 h7h6 g5h3 c8g4 d4e5 c6e5 h3f4 c7c6 h2h3 g4f3 g2f3 d8f6 e1f1 f6f5 f1g2 f8d6 f4d3 e5g6 h3h4 d6e7 h4h5 g6h4 g2g3 f5h5 f3f4 h4f5 g3g2 h5g6 g2f1 f5h4 h1g1 g6f5 d1g4 f5g4 g1g4 g8h7 c1e3 e7f6 e3d2 h8f8 d2c3 f6c3 b2c3');
INSERT INTO evaluations VALUES(2485,30,'cp -299',NULL);
INSERT INTO evaluations VALUES(2485,31,'cp -298',NULL);
INSERT INTO evaluations VALUES(2485,32,'cp -313',NULL);
INSERT INTO evaluations VALUES(2485,33,'cp -329',NULL);
INSERT INTO evaluations VALUES(2485,34,'cp -294',NULL);
INSERT INTO evaluations VALUES(2485,35,'cp -289',NULL);
INSERT INTO evaluations VALUES(2485,36,'cp -307','d2d4 h7h6 g5h3 c8g4 d4e5 c6e5 h3f4 c7c6 h2h3 g4f3 g2f3 d8f6 e1f1 f6f5 c1e3 a8e8 h3h4 g8h7 h4h5 f8d6 f4g6 e5g6 h5g6 f5g6 c2c4 d5c4 h1g1 g6f6 e3d4 d6e5 d1c2 g7g6 a1d1 h8g8 d4e5 e8e5 d1d7 g8g7 d7g7 h7g7 g1g4 f6f5 g4e4 e5e4 f3e4 f5h3 f1e2 h3d3 c2d3 c4d3 e2d3 h6h5 d3e2 h5h4 f2f4 b7b6');
INSERT INTO evaluations VALUES(2483,30,'cp -334',NULL);
INSERT INTO evaluations VALUES(2483,31,'cp -306',NULL);
INSERT INTO evaluations VALUES(2483,32,'cp -303',NULL);
INSERT INTO evaluations VALUES(2483,33,'cp -318',NULL);
INSERT INTO evaluations VALUES(2483,34,'cp -320',NULL);
INSERT INTO evaluations VALUES(2483,35,'cp -320',NULL);
INSERT INTO evaluations VALUES(2483,36,'cp -325','h7h6 g5h3 c8g4 d4e5 c6e5 h3f4 c7c6 h2h3 g4f3 g2f3 d8g5 c1e3 g5f5 e1f1 a8e8 h3h4 g8h7 c2c3 f8d6 d1b1 f5b1 a1b1 e5f3 f1g2 h8f8 g2f3 d6f4 f3e2 h6h5 h1h3 e8e4 b1g1 f4e3 f2e3 g7g6 b2b3 f8f4 h3g3 f4f6 g3h3 h7h6 a2a4 a7a5 e2d3 e4g4 h3h1 f6f8 g1g4 h5g4');
INSERT INTO evaluations VALUES(5788,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(5788,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(5788,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(5788,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5788,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5788,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5788,36,'cp 0','b7c6 d2d4 e5d4 f3d4 g8e7 c1g5 f7f6 g5e3 d8d7 d1d3 c5b6 e1c1 c6c5 d4e2 c8b7 h1e1 e8c8 e2f4 h8e8 e3d2 c8b8 f2f3 e7c6 f4d5 d7f7 d3c4 c6e5 c4b3 c5c4 b3a4 f6f5 d2g5 b7c6 a4a6 c6b7');
INSERT INTO evaluations VALUES(15595,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(15595,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(15595,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15595,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(15595,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(15595,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(15595,36,'cp 23','e1g1 h7h6 c2c3 f8e7 b2b4 e8g8 a2a3 c8d7 c3c4 a7a6 c1b2 f8e8 f1e1 e7f8 b1d2 f6h7 g2g3 h7g5 f3g5 d8g5 e2f3 g7g6 f3g2 d7g4 d1c2 f8g7 d2f1 h6h5 h2h3 g4e6 a1c1 g7h6 f1e3 g5d8');
INSERT INTO evaluations VALUES(15594,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(15594,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(15594,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(15594,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(15594,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(15594,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(15594,36,'cp 34','h7h6 c2c3 f8e7 b2b4 e8g8 c1b2 f8e8 b1d2 e7f8 f1e1 c6e7 d3d4 e7g6 e2f1 c7c6 a2a3 c8d7 c3c4 a7a5 d4e5 d6e5 c4c5 a5b4 a3b4 a8a1 d1a1 b7b6 c5b6 d8b6 d2c4 b6b7 c4e5 e8a8 a1b1 g6e5 f3e5 b7b4 f1c4');
INSERT INTO evaluations VALUES(20747,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(20747,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(20747,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(20747,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(20747,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(20747,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(20747,36,'cp 176','d4d5 d8c7 b1c3 g8f6 d5d3 a7a6 c1g5 f8e7 g5f6 e7f6 c3d5 c7d8 d5f6 d8f6 e1c1 e8g8 d3g3 f6e5 g3e5 d6e5 d1d6 c8e6 c4e6 f7e6 f2f3 f8f6 d6b6 a8f8 b6b7 f6g6 g2g3 f8f3 b7a7 g6g4 a7a6 g4e4 a6e6 g8f7 e6a6 f3f2 a6c6 e4a4 c6c5 a4a2 c5e5');
INSERT INTO evaluations VALUES(20760,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(20760,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(20760,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(20760,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(20760,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(20760,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(20760,36,'cp 167','d8c7 b1c3 a7a6 e1g1 g8f6 d5g5 c8e6 c3d5 c7d8 f1e1 f6g4 b2b3 d8g5 c1g5 a8b8 a1d1 g4e5 c4e2 h7h6 g5e3 g7g5 h2h4 e6d5 d1d5 f8e7 e3c5 b8c8 c5d6 e7d6 d5d6 c8c2 d6b6 c2a2 b6b7 e8g8 b7b6 g5h4 b6h6 f8d8 h6h4 a2b2 e2a6 b2b3');
INSERT INTO evaluations VALUES(41317,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41317,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(41317,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41317,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41317,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41317,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41317,36,'cp 0','g1f3 e8g8 e1g1 c7c5 b1c3 c5d4 e3d4 c8g4 h2h3 g4f3 d1f3 b8c6 c3e2 e7e5 d4e5 d6e5 f4e3 e5e4 d3e4 c6e5 f3f4 f6h5 f4g5 g7f6 g5h6 f6g7');
INSERT INTO evaluations VALUES(46899,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(46899,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(46899,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(46899,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(46899,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46899,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(46899,36,'cp 28','c1f4 c8f5 e2e3 e7e6 f1d3 f5d3 c2d3 f6h5 f4e5 c7c6 d1b3 d8b6 b3c2 f8e7 e1g1 e8g8 c3a4 b6d8 h2h3 h5f6 f1c1 b8d7 e5g3 f8e8 a2a3 a7a5 a4c5 e7c5 d4c5 e6e5 b2b4 e5e4 f3d4 e4d3 c2d3 d7e5 g3e5');
INSERT INTO evaluations VALUES(46886,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46886,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46886,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(46886,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46886,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46886,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(46886,36,'cp 18','c8f5 e2e3 e7e6 f1d3 f5d3 c2d3 f6h5 f4e5 c7c6 d1b3 d8b6 b3c2 f8e7 c3a4 b6b4 e1e2 b8d7 a2a3 b4b5 b2b4 f7f6 e5g3 b7b6 h1c1 a8c8 e2f1 b5a6 a4c3 e8g8 f3h4 h5g3 h2g3 e7d6 h4g6 f8f7 c2d1 e6e5 f1g1 d7f8 g6f8 d6f8');
INSERT INTO evaluations VALUES(48744,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(48744,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(48744,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(48744,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48744,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(48744,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48744,36,'cp -19','f8d6 d2d4 h7h6 c2c4 d5c4 b1d2 e8g8 d2c4 c8e6 f1e1 b8c6 a2a3 f8e8 c1e3 f6d5 a1c1 d8f6 c4d6 c7d6 h2h3 a7a6 e3d2 a8c8 e2d3 c6d4 d3e4 d4f3 e4f3 f6b2 d2h6 d5f6 h6g5 b2a3 c1b1 e6f5 e1e8 c8e8 b1b7');
INSERT INTO evaluations VALUES(41729,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(41729,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(41729,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41729,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41729,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(41729,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(41729,36,'cp -6','d2d4 e8g8 c1g5 c8e6 b1d2 b8d7 c2c4 h7h6 g5h4 f8e8 d1b3 a7a5 f1e1 a5a4 b3c2 c7c5 a1d1 d8c7 h4g3 d6g3 h2g3 e8c8 b2b3 c7a5 d2f1 b7b5 c4b5 a4b3 a2b3 c5c4 b3c4 d5c4 f1e3 a5b5');
INSERT INTO evaluations VALUES(2509,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(2509,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(2509,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(2509,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(2509,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(2509,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(2509,36,'cp 175','f7b3 d8e8 d2d3 d7d6 g5f3 a7a5 c2c3 e8g6 f3h4 g6g4 d1g4 f6g4 e1g1 f8e8 a2a4 h8f8 h4f3 g4f6 b1a3 h7h6 h2h3 e8d8 a3c2 f6h5 d3d4 e5d4 c2d4 c6e5 f3e5 d6e5 d4f3 c5d6 f3d2 h5f4 d2c4 f4e2 g1h2 e2c1 a1c1 c8d7 c4d6 c7d6 f2f3 d8c7');
INSERT INTO evaluations VALUES(2513,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(2513,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(2513,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(2513,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(2513,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(2513,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(2513,36,'cp 169','d8e8 d2d3 d7d6 g5f3 a7a5 c2c3 e8g6 f3h4 g6g4 d1g4 f6g4 e1g1 a5a4 b3d1 g4f6 b1a3 f8e8 a3c2 h7h6 h2h3 g7g5 h4f5 c8f5 e4f5 e8d7 d1f3 d6d5 f1e1 h8b8 g2g4 b7b5 c1d2 b5b4 c3b4 c5b4 c2b4 c6b4 e1e5');
INSERT INTO evaluations VALUES(29378,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29378,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(29378,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29378,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(29378,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(29378,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(29378,36,'cp 0','d1b3 d8c7 f3h4 f5g6 h4g6 h7g6 g2g3 b8c6 f1g2 c7d7 e1g1 c6a5 b3d1 f8b4 c3a4 e8g8 b2b3 b7b6 a4b2 a8c8 b2d3 b4d6 c1b2 a5c6 d1e2 c6b4 d3b4 d6b4 f1c1 b4d6 g2f3 d6b4');
INSERT INTO evaluations VALUES(50841,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(50841,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(50841,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50841,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50841,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(50841,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50841,36,'cp 24','g8f6 c2c4 e7e6 b2b3 c7c5 f1e2 d5c4 b3c4 b7b6 e1g1 f8d6 b1c3 e8g8 a1b1 b8d7 a2a4 c5d4 e3d4 h7h6 h2h3 c8b7 c1e3 d8c7 d1b3 a8b8 f1d1 f6e4 c3e4 b7e4 b1c1 f8c8 f3d2 e4c6 e2f3 c6f3 d2f3');
INSERT INTO evaluations VALUES(43599,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43599,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43599,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(43599,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43599,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43599,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(43599,36,'cp 16','c2c4 e7e6 b2b3 b7b6 c4d5 e6d5 c1a3 f8a3 b1a3 b8d7 a1c1 c7c5 a3b1 c8b7 f1d3 a8c8 e1g1 e8g8 d4c5 b6c5 b1c3 f8e8 c3a4 h7h6 h2h3 d8e7 f1e1 f6e4 a4c3 d7f8 d1c2 e4f6 c3a4 f8e6 d3f5 c5c4 f5e6 f7e6 b3c4 d5c4');
INSERT INTO evaluations VALUES(43585,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43585,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(43585,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43585,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(43585,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43585,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(43585,36,'cp 13','e7e6 b2b3 c7c5 f1e2 c5d4 e3d4 b7b6 e1g1 c8b7 b1c3 f8d6 c1g5 d5c4 b3c4 e8g8 a1b1 b8d7 c4c5 d6c7 c5b6 d7b6 f3e5 a8b8 e2d3 h7h6 g5h4 b6d5 c3e4 g7g5 e4f6 d5f6 h4g3 d8d5 d1f3 d5f3 g2f3 f8c8 f1c1 f6d5 e5d7 c7g3 h2g3');
INSERT INTO evaluations VALUES(5630,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5630,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5630,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(5630,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(5630,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5630,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5630,36,'cp 0','c6a5 c4b3 a5b3 a2b3 f7f5 d3d4 f5e4 c3e4 g8f6 d1e2 f6e4 e2e4 e5d4 c1g5 h7h6 e4g6 e8d7 g6f5 d7e8 f5g6');
INSERT INTO evaluations VALUES(6675,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(6675,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(6675,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(6675,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(6675,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(6675,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(6675,36,'cp 8','c1d2 g8f6 a2a3 b4d2 b1d2 e8g8 e1c1 d7d5 e4d5 f8e8 e3b3 d8d5 b3d5 f6d5 f1c4 d5f6 g1f3 c8d7 h1e1 e8e1 d1e1 a8e8 e1e8 d7e8 c1d1 g8f8 d2f1 f6g4 d1e2 h7h6 f1e3 g4e3 e2e3 f8e7 c4d3 e7d6 f3d2 c6e5 d3e4 c7c6');
INSERT INTO evaluations VALUES(6676,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(6676,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(6676,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(6676,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(6676,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6676,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(6676,36,'cp 0','g8f6 a2a3 b4d2 b1d2 e8g8 e1c1 d7d5 e4d5 f8e8 e3c5 d8d5 c5d5 f6d5 f1c4 d5f6 g1f3 c8d7 c4d3 f6g4 d2e4 c6e5 f3e5 g4e5 e4c3 e5d3 d1d3 d7c6 f2f3 e8e5 h1d1 a8e8 c3d5 c6d5 d3d5 e5e2 d1d2 e2e1 d2d1');
INSERT INTO evaluations VALUES(9318,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(9318,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(9318,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(9318,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(9318,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(9318,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(9318,36,'cp -33','e8g8 h2h3 c6a5 c4b3 a5b3 a2b3 a7a5 e1g1 c7c6 g5f3 h7h6 c1e3 c5e3 f2e3 b7b5 d1d2 c8d7 d2f2 g8h7 c3e2 d7e6 f3h4 f6d7 e2c3 b5b4 c3b1 d6d5 b1d2 d7c5 h4f5 g7g6 f5g3 d8e7 d3d4 c5d7 d2f3 f7f6 f1e1 f8e8 f2d2');
INSERT INTO evaluations VALUES(20946,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(20946,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(20946,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(20946,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(20946,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(20946,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(20946,36,'cp 112','e7e6 e1g1 f8d6 f1e1 g8e7 d2d4 a7a6 b5f1 e8g8 c3e4 h7h6 b2b3 b7b5 c1b2 a8b8 d1d2 c8b7 e4d6 c7d6 c2c4 b5c4 b3c4 e7g6 a1b1 d6d5 b2a3 f8e8 c4c5 c6a5 d2c3 a5c4 f3d2 g6e7 d2c4 d5c4 c3c4 b7c6 b1b8 d8b8 c4a6 b8f4');
INSERT INTO evaluations VALUES(21496,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(21496,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(21496,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(21496,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(21496,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(21496,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(21496,36,'cp 102','f1c4 a7a6 g1f3 b7b5 c4d3 c6b4 d3e2 g8f6 a2a4 b5a4 e1g1 f8d6 e3e4 d6e7 d1a4 c8d7 a4a5 e8g8 c1f4 d7c6 f4c7 d8c8 d4d5 e6d5 e4e5 f6e4 f3d4 c8d7 f1d1 a8c8 c3e4 d7c7 a5c7 c8c7 e4d6 f7f6 d4e6 f6e5 e6c7 e7d6');
INSERT INTO evaluations VALUES(30968,30,'cp 293',NULL);
INSERT INTO evaluations VALUES(30968,31,'cp 315',NULL);
INSERT INTO evaluations VALUES(30968,32,'cp 277',NULL);
INSERT INTO evaluations VALUES(30968,33,'cp 313',NULL);
INSERT INTO evaluations VALUES(30968,34,'cp 295',NULL);
INSERT INTO evaluations VALUES(30968,35,'cp 318',NULL);
INSERT INTO evaluations VALUES(30968,36,'cp 325','f3e5 d8d6 f1c4 b8d7 d1e2 d7e5 d4e5 d6c7 e5e6 g8f6 e1g1 f7e6 b2b3 h6h5 c4e6 h5h4 g3f5 h8h5 c1b2 g6f5 b2f6 f5e6 e2h5 e6f7 h5e5 c7e5 f6e5 a7a5 h2h3 a5a4 a1d1 f7g6 c2c3 b7b5 d1d4 g6f7 b3a4 a8a4 f1a1 a4a2 a1a2 f7a2 d4h4 e8f7 h4d4');
INSERT INTO evaluations VALUES(30960,30,'cp 321',NULL);
INSERT INTO evaluations VALUES(30960,31,'cp 325',NULL);
INSERT INTO evaluations VALUES(30960,32,'cp 318',NULL);
INSERT INTO evaluations VALUES(30960,33,'cp 311',NULL);
INSERT INTO evaluations VALUES(30960,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(30960,35,'cp 286',NULL);
INSERT INTO evaluations VALUES(30960,36,'cp 317','d8d6 f1c4 b8d7 d1e2 d7e5 d4e5 d6c7 e5e6 g8f6 e1g1 a8d8 c1d2 h6h5 h2h4 f7e6 d2c3 c7f4 c4e6 d8d6 f1d1 f6e4 d1d6 e7d6 a1e1 e4c3 e6f7 e8d8 e2e8 d8c7 f7g6 c3b5 a2a4 f4a4 g3e4 c7b6 e4g5 a4h4 g5f7 h8g8 g6h7 d6d5 h7g8 f8c5 e8d8 h4d8 f7d8 c5d4');
INSERT INTO evaluations VALUES(34144,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(34144,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(34144,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(34144,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(34144,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(34144,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(34144,36,'cp 21','g8f6 d1d3 b7b5 d3g3 d7d6 b2b4 c8b7 a2a4 b5a4 a1a4 f8e7 b4b5 e8g8 c1h6 f6e8 e1g1 e7f6 f1d1 c7d8 h6e3 a6b5 a4a8 b7a8 d4b5 b8c6 c3b1 f6h4 g3h3 h4e7 e3f4 d8b8 h3a3 e8f6 b1d2 d6d5 f4b8 e7a3 e4d5 c6b8 b5a3 a8d5');
INSERT INTO evaluations VALUES(33532,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(33532,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(33532,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(33532,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(33532,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33532,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(33532,36,'cp 13','e1g1 b8c6 c1e3 d7d6 f2f4 f8e7 d1e1 e8g8 a2a3 c6d4 e3d4 b7b5 e1g3 c8b7 g1h1 a8e8 a1d1 h7h6 e2f3 e8d8 d1d2 g8h8 f1d1 b7c6 h2h3 c7b7 d2d3 b7c7 d1f1 c6b7 d3d1 b7c6 d1d2 c6b7 d2d3 b7c6 f1d1 f8g8 f4f5');
INSERT INTO evaluations VALUES(33531,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33531,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(33531,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(33531,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(33531,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(33531,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33531,36,'cp 21','b8c6 c1e3 d7d6 d1e1 f8e7 f2f4 c6d4 e3d4 e8g8 e1g3 b7b5 g1h1 c8b7 a2a3 a8e8 a1d1 b7c6 f1e1 g8h8 e2f3 e8d8 d1d3 c6b7 b2b4 f8e8 h2h3 h7h6 e1d1 b7c6 e4e5 d6e5 f4e5 f6h7 f3c6 c7c6 g3f3 c6f3 g2f3 h7f8 c3e4');
INSERT INTO evaluations VALUES(667,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(667,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(667,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(667,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(667,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(667,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(667,36,'cp 28','c1e3 d7d6 f2f4 f8e7 d1e1 e8g8 e1g3 c6d4 e3d4 b7b5 g1h1 c8b7 a2a3 a8e8 a1e1 g8h8 e1d1 h8g8 e2f3 e8d8 f1e1 g8h8 d1d3 h7h6 d3d1 b7c6 d1d2 d8c8 h2h3 f8d8 e1d1 c6b7 g3f2 d8e8 e4e5 b7f3 f2f3 d6e5');
INSERT INTO evaluations VALUES(660,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(660,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(660,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(660,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(660,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(660,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(660,36,'cp 22','f8e7 f2f4 d7d6 d1e1 e8g8 e1g3 c6d4 e3d4 b7b5 a2a3 c8b7 a1d1 a8d8 e2f3 g8h8 g1h1 b7c6 h2h3 h7h6 f1e1 c6b7 d1d3 f8e8 d3d2 e8g8 c3d1 d6d5 e4e5 f6e4 f3e4 d5e4 d1e3 a6a5 e1d1 b5b4 a3b4 a5b4 b2b3 c7c8 h1h2 g7g6 e3c4 h6h5');
INSERT INTO evaluations VALUES(439,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(439,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(439,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(439,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(439,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(439,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(439,36,'cp 13','f2f4 d7d6 d1e1 c6d4 e3d4 e8g8 e1g3 b7b5 a2a3 c8b7 g1h1 a8d8 a1d1 h7h6 e2f3 g8h8 h2h3 b7c6 d1d3 c6b7 d3d2 f8g8 f1e1 g8f8 d2d1 f8e8 d1d3 b7c6 e4e5 d6e5 d4e5 c7b6 c3e4 c6e4 f3e4 h8g8 e1f1 d8d3 e4d3 e8d8 f4f5');
INSERT INTO evaluations VALUES(34435,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(34435,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(34435,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34435,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(34435,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34435,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34435,36,'cp 0','b1c3 c5d4 e3d4 b6a5 a2a3 b8c6 g1f3 a7a6 d4e3 a5c7 c3a4 c8d7 e3b6 c7b8 b6e3');
INSERT INTO evaluations VALUES(37437,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37437,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(37437,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37437,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37437,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(37437,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(37437,36,'cp -5','d7d6 c2c3 b8c6 f1e2 f6d7 d4d5 c6e5 f3e5 d7e5 e1g1 e7e6 d5e6 c8e6 e3e4 d6d5 f1e1 f8e8 d1c2 d8d6 e4d5 e6d5 h2h3 h7h5 a1d1 d6f6 f4g3 a8d8 e2b5 c7c6 b5f1 b7b5 a2a4 a7a6 f2f3 e5c4 d2c4 b5c4 e1e8 d8e8');
INSERT INTO evaluations VALUES(1645,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(1645,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(1645,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(1645,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(1645,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(1645,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(1645,36,'cp 198','d4d5 c6b8 g1f3 c7c6 c1d2 g8f6 d5c6 a7a6 c6b7 c8b7 f1c4 a5c7 d1e2 e7e6 f3e5 b8c6 e5c6 c7c6 c4b3 c6g2 e1c1 f8e7 h1g1 g2f3 g1g7 f3e2 c3e2 h7h5 e2f4 a8c8 d2e3 b7f3 d1g1 f6g4 e3d4 e7d6 f4e6 f7e6 b3e6');
INSERT INTO evaluations VALUES(5017,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(5017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5017,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5017,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5017,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(5017,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5017,36,'cp -6','f1c4 g8f6 d1f3 d7d6 e1g1 e8g8 c1g5 h7h6 a2a3 b4a5 g5f6 d8f6 f3f6 g7f6 b2b4 a5b6 c3a4 f6f5 a4b6 a7b6 f2f3 b6b5 c4b3 f5e4 f3e4 c8e6 f1d1 f8d8 a3a4 b5a4 a1a4 e6b3 a4a8 d8a8 c2b3 a8d8 g1f2 h6h5 f2f3 d8b8 d1d6 b8b4 d6d3 g8f8 d3c3 f8e7 g2g3');
INSERT INTO evaluations VALUES(11865,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11865,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(11865,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(11865,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11865,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11865,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11865,36,'cp -5','d2d3 f8e7 b1c3 e8g8 e1g1 c6c5 f3d2 c8e6 d2c4 c7c6 b2b3 f6d7 c4e3 f8e8 c1d2 e7g5 e3c4 d6d5 c4d6 e8f8 d2g5 d8g5 d1f3 d7b6 e4d5 c6d5 c3b5 a8d8 a1e1 d8d7 f3g3 g5g3 f2g3');
INSERT INTO evaluations VALUES(11861,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(11861,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(11861,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(11861,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(11861,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(11861,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(11861,36,'cp -17','g7g6 c2c3 f8g7 e1g1 e8g8 f1e1 a7a5 d3d4 f6d7 d4e5 d7e5 f3e5 g7e5 c1h6 f8e8 b1d2 e5h8 d2f3 c8e6 e4e5 d6e5 f3e5 d8d5 h6f4 a5a4 c3c4 d5c5 d1f3 e6c4 e5c4 e8e1 a1e1 c5c4 e1c1 c4d5 b2b4 a4b3 f3d5 c6d5 a2b3 a8a1 c1a1 h8a1 f4c7 a1c3 g1f1 d5d4 f2f3 f7f5');
INSERT INTO evaluations VALUES(12940,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(12940,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(12940,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(12940,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(12940,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12940,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(12940,36,'cp 41','c8d7 b1c3 d8b6 e1g1 e7f5 b5a4 a8c8 c3e2 f8e7 a4c2 c6b4 c2f5 e6f5 e2c3 d7e6 c1g5 e7g5 f3g5 e8g8 g5f3 h7h6 d1d2 c8c7 a1c1 f8c8 f1d1 b4c6 h2h4 b6b4 a2a3 b4a5 h4h5 a5b6 b2b4 a7a5 b4b5 c6d8 a3a4 c7c4 c3e2');
INSERT INTO evaluations VALUES(24448,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(24448,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(24448,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24448,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(24448,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(24448,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(24448,36,'cp 47','b1c3 d8b6 e1g1 e7f5 b5a4 f8e7 c3e2 a8c8 a4c2 c6b4 c2f5 e6f5 e2c3 d7e6 c1g5 e7g5 f3g5 e8g8 d1d2 c8c4 g5f3 f8c8 f1d1 h7h6 h2h4 c4c7 a1c1 b4c6 c3e2 c6b4 c1c7 c8c7 e2c3 c7c4 d1c1 b6a5 h4h5 b4a6 d2e3 a6c7 f3d2 c4c6 d2b3');
INSERT INTO evaluations VALUES(21505,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(21505,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(21505,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(21505,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(21505,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(21505,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(21505,36,'cp 181','f1c4 g8f6 b1c3 f8b4 c4d3 f6d7 a2a3 b4c3 b2c3 b7b6 e1g1 c8b7 f1e1 e8g8 c1g5 f7f6 g5h4 c6a5 f3d2 e6e5 d2c4 a5c4 d3c4 g8h8 h4g3 a7a5 a3a4 d8e7 d1e2 b7c6 f2f3 h7h6 e2a2 a8c8 a1d1 f8d8 d1d3 d7f8 f3f4 e5f4 g3f4 f8g6 f4g3');
INSERT INTO evaluations VALUES(46191,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(46191,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(46191,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(46191,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(46191,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(46191,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(46191,36,'cp -30','a7a6 d4c5 f8c5 f1d3 b8d7 e3e4 e6e5 f4g5 d8b6 e1g1 d5e4 c3e4 f6e4 d3e4 e8g8 g5h4 f8e8 d1d3 h7h6 e4d5 d7f6 h4f6 b6f6 d3c3 f6e7 f3e5 e7e5 a1e1 c5f2 g1f2 e5f5 c3f3 f5c2 f2g1 c8e6 d5e6 f7e6 f3b7 a8b8 b7f7 g8h7 b2b3 c2a2 e1e6');
INSERT INTO evaluations VALUES(50207,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(50207,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(50207,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(50207,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(50207,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(50207,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(50207,36,'cp -29','d7d5 f1e2 f8g7 g1f3 e8g8 e1g1 a7a6 f1e1 f8e8 e2f1 b8c6 f3e5 c6e5 d4e5 f6e4 c3d5 g7e5 d5f4 e5d6 d1e2 e7e5 f4d3 c7c5 f2f3 e4f6 e3e4 c5c4 d3f2 c8e6 c1e3 d8c7 e1d1 b7b5 a2a4');
INSERT INTO evaluations VALUES(46782,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(46782,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(46782,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(46782,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(46782,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(46782,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(46782,36,'cp -15','g1f3 f8g7 a2a4 e8g8 f1e2 c7c6 e1g1 b8d7 a4a5 f8e8 f3e5 d7e5 d4e5 f6e4 c3e4 d5e4 d1d4 d8d5 a1a4 g7e5 d4e4 c8f5 e4c4 d5c4 a4c4 f5e6 c4b4 e8b8 e2f3 e5d6 b4a4 b8d8 f1d1 d6e5 g2g4 d8d1 f3d1 a8d8 d1e2 b7b6 f2f4 e5f6 e3e4 b6b5 a4a1');
INSERT INTO evaluations VALUES(57715,30,'cp 242',NULL);
INSERT INTO evaluations VALUES(57715,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(57715,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(57715,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(57715,34,'cp 252',NULL);
INSERT INTO evaluations VALUES(57715,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(57715,36,'cp 238','b1c3 d7d6 h2h4 g5h4 g1f3 f8g7 f3h4 b8c6 c1e3 e7e5 d4e5 g7e5 d1d2 e5c3 d2c3 d8f6 e3d2 f6c3 d2c3 h8h7 f2f3 h6h5 e1f2 c6e5 a1e1 c8e6 h4f5 g8f6 f1b5 e6d7 a2a4 a7a6 b5d7 e8d7 h1h4 b7b5 e1h1 a8g8 f5g3 g8e8 a4b5 a6b5');
INSERT INTO evaluations VALUES(57714,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(57714,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(57714,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(57714,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(57714,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(57714,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(57714,36,'cp 210','d7d6 h2h4 g5h4 g1f3 f8g7 f3h4 b8c6 c1e3 e7e5 d4e5 g7e5 d1d2 e5c3 d2c3 d8f6 e3d2 h8h7 h4f3 f6c3 d2c3 f7f6 f1c4 h7g7 f3h4 c6e5 c4b3 a7a5 a2a4 g8e7 f2f3 h6h5 e1f2 b7b6 a1e1 g7h7 h4f5 c8f5 e4f5 h5h4 b3e6');
INSERT INTO evaluations VALUES(23283,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(23283,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(23283,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(23283,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(23283,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(23283,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(23283,36,'cp 140','g8f6 d2d4 e5d4 f3d4 f8b4 d4c6 b7c6 a2a3 b4c3 b2c3 e8g8 f1d3 d8e7 c1e3 d7d6 d1f3 f6d5 g2g4 e7e5 e1g1 e5c3 a1b1 c8d7 b1b3 c3a5 e3g5 a8e8 c2c4 d5b6 f1b1 e8e1 g1g2 a5e5 g5f4 e1b1 b3b1 e5d4 h2h3 b6c4 b1b7');
INSERT INTO evaluations VALUES(39977,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(39977,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(39977,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(39977,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(39977,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(39977,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(39977,36,'cp 47','e2e4 c6c5 d2d4 c5d4 d1d4 b8c6 d4d2 e8g8 g1e2 d7d6 b2b3 c8g4 c1b2 d8a5 f2f3 g4d7 a1d1 a7a6 e1g1 f8b8 c3d5 a5d2 d1d2 f6d5 c4d5 c6e5 f3f4 e5g4 b2g7 g8g7 f1c1 b8c8 d2c2 c8c2 c1c2 a8c8 c2c8 d7c8 g2f3 h7h6 f3g4 c8g4 g1f2 g4e2 f2e2 f7f5 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(39974,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(39974,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(39974,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(39974,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(39974,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(39974,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(39974,36,'cp 37','c6c5 d2d4 c5d4 d1d4 b8c6 d4d2 e8g8 g1e2 a8b8 b2b3 a7a6 a2a4 d7d6 e1g1 c8g4 a1b1 g4e2 d2e2 f6d7 c1d2 e7e6 f1c1 d7c5 d2e3 g7d4 b3b4 d4e3 e2e3 c5d7 c1d1 d8c7 f2f4 f8c8 c3e2 a6a5');
INSERT INTO evaluations VALUES(41997,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(41997,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(41997,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(41997,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(41997,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(41997,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(41997,36,'cp 52','e8g8 e2e3 c7c5 c4d5 e6d5 f1d3 c5c4 d3c2 b8d7 e1g1 d8a5 f3e5 a5c3 e5d7 f6d7 d1b1 d7f6 a2a4 f8e8 g5f6 g7f6 c2h7 g8g7 h7f5 b7b6 h2h3 c3d2 f5c8 a8c8 b1f5 c8d8 h3h4 e8e4 h4h5 e4e6 f1d1 d2e2 d1e1 e2d3 f5d3 c4d3 e1d1 d8h8 d1d3 h8h5');
INSERT INTO evaluations VALUES(37028,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(37028,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(37028,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(37028,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(37028,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(37028,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(37028,36,'cp 52','e2e3 c7c5 f1e2 b8d7 c4d5 e6d5 e1g1 d8a5 c3c4 h7h6 g5h4 d5c4 e2c4 b7b5 c4e2 f6e4 d4c5 e4c3 d1c2 c8b7 f3d4 d7c5 h4e7 b7e4 c2d2 c3e2 d2e2 e4d3 e2g4 d3f1 a1f1 a5b6 d4f5 b6g6 g4f3 g6b6 h2h4 c5e6 e7f8 a8f8 f5e7 g8h8 f1d1 b6a5 f3b7 a5a4 d1d7 a4h4 b7b5 h4e4');
INSERT INTO evaluations VALUES(37027,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(37027,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(37027,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(37027,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(37027,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(37027,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(37027,36,'cp 50','c7c5 f1e2 d5c4 e1g1 b8d7 e2c4 d8a5 a2a4 b7b6 f3e5 c5d4 e5d7 c8d7 g5f6 g7f6 c3d4 d7c6 d1g4 g8h8 g4f4 a5g5 f4g5 f6g5 f1c1 c6b7 c4b5 f8c8 b5d7 c8c1 a1c1 g5g4 c1c7 b7d5 g1f1 h8g8 h2h3 h7h5 d7c6 d5c6 c7c6 a8b8 h3g4 h5g4 c6c7 b6b5');
INSERT INTO evaluations VALUES(42876,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(42876,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(42876,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(42876,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(42876,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(42876,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(42876,36,'cp 125','e2e3 d7d6 d2d4 c5b6 h2h3 e5d4 e3d4 g4f6 d1e2 d8e7 e2e7 e8e7 g1e2 b8c6 c1g5 h8e8 e1c1 e7f8 g5f6 g7f6 d1d2 b6a5 e2f4 f6f5 h1d1 c8d7 c1c2 c6b4 c2b3 b4c6 f4d5 f8g7 b3c2 d7e6 a2a3 a5c3 c2c3 e6d5 g2d5 a8b8 d5c6 b7c6 b2b4 d6d5 c4c5 b8b5 c3b3');
INSERT INTO evaluations VALUES(42875,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(42875,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(42875,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(42875,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(42875,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(42875,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(42875,36,'cp 150','d7d6 d2d4 c5b6 g1f3 b8c6 b2b4 e8g8 h2h3 e5d4 e3d4 g4f6 c4c5 d6c5 d4c5 d8d1 e1d1 f8d8 c1d2 c6b4 c5b6 a7b6 f3e5 c7c6 d1c1 c8e6 a2a3 b6b5 d2e3 b4d5 c3d5 f6d5 h1e1 d5e3 e1e3 g7g6 e3d3 f7f6 d3d8 a8d8 e5g4 g8f7 g4e3 d8d3 e3c2');
INSERT INTO evaluations VALUES(41926,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(41926,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(41926,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(41926,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(41926,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(41926,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(41926,36,'cp 157','d2d4 c5b6 g1f3 e8g8 h2h3 e5d4 f3d4 b6d4 d1d4 g4e5 b2b3 b8c6 d4d2 c6b4 e1f1 e5d3 f1g1 a7a5 g1h2 d3c5 c1a3 c8f5 e3e4 f5g6 h1e1 c7c6 a1d1 f7f5 e4f5 g6f5 d2d6 d8d6 d1d6 f5g6 e1e2 a8d8 d6d8 f8d8 g3g4 h7h5 f2f4 d8d3 a3b2 h5h4 f4f5');
INSERT INTO evaluations VALUES(44487,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(44487,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(44487,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(44487,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(44487,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(44487,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(44487,36,'cp 46','d7d5 e4e5 f6e4 d2d4 e4c3 b2c3 c6c5 d4c5 e7e6 d1d4 d8a5 c1d2 b8d7 c3c4 a5c5 c4d5 c5d5 d4d5 e6d5 c2c4 d7b6 c4d5 b6d5 f1c4 d5b6 c4b3 f8a3 e1g1 e8g8 d2e3 c8g4 h2h3 g4f5 f3d4 f5d3 f1d1 a3b2 e5e6 b2a1 d1d3');
INSERT INTO evaluations VALUES(7862,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(7862,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(7862,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(7862,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(7862,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(7862,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(7862,36,'cp -18','d7d5 e1g1 d5d4 c3e2 e8g8 h2h3 a7a5 a2a3 b4d6 c2c3 a5a4 b3d2 c8e6 d2f3 d8b6 c3d4 e5d4 c1f4 d6f4 e2f4 e6b3 d1d2 f6d7 f4d5 b6a5 d5b4 d7c5 a1c1 h7h6 d3b1 f8d8 b4c6 b7c6 d2a5 a8a5 f3e5 d8c8 f2f4 f7f6 e5c4 b3c4 c1c4 d4d3');
INSERT INTO evaluations VALUES(8374,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(8374,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(8374,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(8374,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(8374,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(8374,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(8374,36,'cp -20','e1g1 d5d4 c3e2 e8g8 h2h3 f8e8 c2c3 b4f8 c3d4 e5d4 d3b1 f6e4 e2d4 c8d7 b1e4 e8e4 d4c6 d7c6 d1d8 a8d8 c1e3 b7b6 f1d1 d8d1 a1d1 e4e7 b3d4 e7d7 d1c1 c6d5 d4b5 d5e4 c1c8 e4b7 c8b8 b7c6 b5d4 c6d5 b2b3 f7f6 h3h4 g8f7 f2f3 f8c5 g1f2 c5d6 b8h8');
INSERT INTO evaluations VALUES(29673,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(29673,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(29673,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(29673,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(29673,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29673,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29673,36,'cp 22','e7e6 c2c4 c7c6 d1b3 d8b6 h2h3 g4f3 g2f3 b8d7 e1g1 f8e7 b1c3 e8g8 f1d1 a7a5 e2e4 f6e4 c3e4 b6b3 a2b3 d5e4 f3e4 e7b4 c1e3 f8d8 g3g4 d7f6 e4f3 h7h6 h3h4 f6e8 g1g2 d8d7 g4g5 a8d8 d4d5 c6d5 e3b6 e8c7 c4d5 e6d5 g5h6');
INSERT INTO evaluations VALUES(30182,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(30182,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(30182,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30182,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(30182,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(30182,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(30182,36,'cp 25','c2c4 c7c6 d1b3 d8b6 h2h3 g4f3 g2f3 b8d7 e1g1 f8e7 b1c3 e8g8 f1d1 a7a5 e2e4 b6b3 a2b3 f6e4 c3e4 d5e4 f3e4 f8d8 c1f4 d7f6 e4f3 d8d7 h3h4 g8f8 g3g4 h7h6 f4e3 e7b4 g4g5 h6g5 h4g5 f6e8 d4d5 c6d5 c4d5 e6d5 g1g2 e8c7 a1c1 c7e6');
INSERT INTO evaluations VALUES(31062,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(31062,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(31062,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(31062,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(31062,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(31062,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(31062,36,'cp 61','d1b3 d8b6 c4d5 b6b3 a2b3 f5b1 d5c6 b8c6 a1b1 c6d4 e2e3 d4c6 f1b5 f7f6 b1c1 a8c8 e3e4 e7e5 b5c6 c8c6 c1c6 b7c6 f4e3 f8b4 e1e2 a7a5 g1f3 g8e7 h1d1 e7g6 g2g3 e8e7 f3d2 h8d8 d1c1 b4d2 e3d2 d8b8 c1c6 b8b3');
INSERT INTO evaluations VALUES(45201,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45201,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(45201,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(45201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45201,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45201,36,'cp 0','c8g4 h2h3 g4h5 d3e2 e7e6 d2d3 b8c6 c1f4 f8e7 b1d2 h7h6 e1g1 e8g8 a2a4 a7a6 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(48947,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(48947,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(48947,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48947,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(48947,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(48947,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(48947,36,'cp 25','f8d6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 b1c3 e7d6 c3b5 b8d7 b5d6 c7d6 c1f4 d7b6 b2b3 f8d8 f4g5 c8e6 f3d2 a8c8 f1c1 e6c4 d2c4 b6c4 c1c4 b7b5 c4c1 a7a5 f2f3 a5a4 g5f6 g7f6 d4d5 c8c1 a1c1 a4b3 a2b3 d8a8 c1c7 a8a1 g1f2 g8g7');
INSERT INTO evaluations VALUES(41741,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(41741,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(41741,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(41741,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(41741,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(41741,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41741,36,'cp 24','d1e2 d8e7 e2e7 d6e7 b1c3 e8g8 e1g1 e7d6 c1g5 b8d7 f1e1 h7h6 g5h4 d7b6 c4b3 a7a5 a2a4 f8e8 h4f6 e8e1 a1e1 g7f6 c3e4 d6e7 e4g5 h6g5 e1e7 c8e6 b3e6 f7e6 e7c7 b6a4 c7b7 a8c8 h2h3 c8c1 g1h2 c1b1 b7e7 b1b2 e7e6 g8f7 e6a6 b2f2 a6a5');
INSERT INTO evaluations VALUES(49658,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(49658,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(49658,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(49658,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(49658,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(49658,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(49658,36,'cp -102','b8c6 b1c3 c6d4 f3d1 d7d5 d2d3 c7c6 g1f3 f8d6 f1e2 e8g8 c1g5 d8b6 f3d4 e5d4 g5f6 g7f6 c3b1 d5e4 d3e4 d6f4 b1d2 b6b2 e1g1 b2c3 d2f3 c8e6 a1b1 b7b5 a2a4 a7a6 a4b5 c6b5 d1d4 f8c8 e2d3 c3d4 f3d4 f4e5 d4e6 f7e6 g2g3 e5c3 b1b3 b5b4');
INSERT INTO evaluations VALUES(16575,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(16575,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(16575,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(16575,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(16575,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(16575,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(16575,36,'cp -84','b1c3 c6d4 f3d1 d7d5 d2d3 c7c6 g1f3 f8d6 f1e2 e8g8 e1g1 d6c7 f1e1 d8d6 a2a4 c8d7 e2f1 a8e8 g2g3 d4f3 d1f3 d5d4 c3b1 d7e6 b1d2 a7a5 f3e2 f6d7 d2f3 h7h6 f1g2 g8h7 f3h4 d6e7 h4f5 e7f6 c1d2 e6f5');
INSERT INTO evaluations VALUES(52460,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(52460,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(52460,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(52460,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(52460,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(52460,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(52460,36,'cp -61','c2c4 d7d5 c4d5 d8d5 g1f3 c8g4 b1c3 d5e6 f1g2 b8c6 d2d3 e6d7 c1e3 e7e6 d1a4 g8f6 h1g1 a8c8 f3d2 h7h5 a1c1 f6d5 h2h3 g4f5 c3d5 e6d5 e3c5 f5h3 g2h3 d7h3 d2f3 h3e6 e1d1 f8c5 c1c5 e8g8 a4h4 f8e8 c5c2 c6e7 f3d4 e6b6 e2e3 c8c2 d1c2 e7c6 d4c6 b6c6 c2b1');
INSERT INTO evaluations VALUES(54817,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(54817,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(54817,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(54817,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(54817,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54817,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54817,36,'cp 14','f8g7 g1f3 b8d7 h2h3 e7e5 d4e5 d6e5 f4h2 d8e7 b1c3 g8f6 c3b5 e8g8 b5c7 a8b8 f1e2 b7b5 a2a4 b5a4 a1a4 b8b2 e1g1 c8b7 a4a7 e7c5 d1a1 b7f3 e2f3 b2c2 a1a6 f8b8 c7d5 c2c1 a7c7 b8b1 c7c5 c1f1 a6f1 b1f1 g1f1 d7c5 d5f6 g7f6 f3d5 c5d3 f1e2 d3c5 e2d2 e5e4 h2d6');
INSERT INTO evaluations VALUES(56530,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(56530,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(56530,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(56530,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(56530,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(56530,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(56530,36,'cp 129','d1d8 e8d8 b1c3 b8c6 g1f3 c8e6 f3g5 c6e5 f2f4 e5c6 f4f5 h7h6 f5e6 h6g5 e6f7 g8f6 c1g5 c6e5 e2e3 c7c6 e1c1 d8e7 g5f4 e5d3 f1d3 c4d3 d1d3 e7f7 h1f1 f8b4 c3e2 b4a5 g2g4 f7g6 g4g5 f6d5 f4e5 a8e8 e2f4 d5f4 e3f4 h8h2');
INSERT INTO evaluations VALUES(35404,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(35404,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(35404,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(35404,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(35404,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(35404,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(35404,36,'cp 110','e8d8 g1f3 b8c6 b1c3 c8e6 f3g5 c6e5 f2f4 e5c6 e2e4 d8e8 c1d2 a8d8 g5e6 f7e6 f1c4 e6e5 e1c1 g8f6 h1e1 f8b4 g2g3 b4c3 b2c3 d8d6 c1c2 e8d7 d2c1 a7a6 c1a3 d6d1 e1d1 d7c8 f4f5 c8b8 d1e1 h8d8');
INSERT INTO evaluations VALUES(35450,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(35450,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(35450,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(35450,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(35450,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(35450,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(35450,36,'cp 139','b1c3 b8c6 g1f3 h7h6 e2e4 c8e6 c1e3 g8e7 c3b5 a7a6 e1c1 d8c8 b5d4 b7b5 d4e6 f7e6 g2g3 e7g6 f1h3 g6e5 f3e5 c6e5 c1b1 c8b7 f2f4 e5d3 h3e6 f8b4 e4e5 a8d8 b1c2 b4c5 e3c5 d3c5 d1d8 h8d8 e6f5');
INSERT INTO evaluations VALUES(12146,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(12146,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(12146,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(12146,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(12146,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12146,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(12146,36,'cp 38','c4d5 e6d5 f1d3 f8e7 g1f3 e8g8 e1g1 f6e4 g5f4 d7f6 d1c2 e7d6 f4d6 e4d6 f3e5 g7g6 f1e1 c8f5 f2f3 f8e8 a1d1 f5d3 c2d3 f6d7 e5d7 d8d7 e3e4 d6c4 d3c2 d7e7 c2f2 a8d8 h2h3 d5e4');
INSERT INTO evaluations VALUES(12143,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(12143,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12143,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12143,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(12143,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(12143,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(12143,36,'cp 37','e6d5 f1d3 f8e7 g1f3 d7f8 d1c2 f8e6 g5h4 g7g6 e1g1 e8g8 a1b1 f6e8 h4e7 d8e7 b1e1 e8f6 f3e5 e6g7 f2f3 c8e6 c2f2 f6d7 e5d7 e7d7 e1d1 a8d8 f1e1 f8e8 d3b1 d7e7 e3e4 d8d7 e4e5 e6f5 b1f5 g7f5 g2g4 f5g7 g1h1 e7g5 a2a3');
INSERT INTO evaluations VALUES(20713,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(20713,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(20713,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20713,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(20713,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(20713,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(20713,36,'cp 42','b1c3 g8f6 c1e3 f8e7 d1e2 e6e5 d4f5 c8f5 e4f5 c6d4 e3d4 e5d4 c3e4 d8a5 e4d2 e8g8 e1g1 e7d8 e2d3 d6d5 c4b3 f8e8 g2g3 d8b6 g1g2 h7h6 a1d1 a7a6 a2a3 a8d8 f1e1 d8c8 e1e8 c8e8 b3a2 a5c5 d2f3 e8e4 b2b4 c5c6 h2h4 g8f8 a2b3 a6a5 d1e1 a5a4');
INSERT INTO evaluations VALUES(27535,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(27535,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(27535,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(27535,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(27535,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(27535,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(27535,36,'cp 30','b2b4 c5d4 c3d4 a8c8 c1b2 g8h6 b1c3 c6a5 c3a4 b6c6 a4c5 a5c4 b2c1 f8c5 d4c5 h6f5 f1d3 e8g8 e1g1 b7b6 c5b6 c6b6 h2h4 h7h6 f1e1 d7c6 d3f5 e6f5 d1d4 c6d7 c1f4 f8e8 d4b6 a7b6 e1d1 d7a4 d1d3 a4c2 d3d5 c2e4 d5d1 e4f3 g2f3 c4e5 f4e5 e8e5 a1c1 c8c1 d1c1 e5e2 a3a4');
INSERT INTO evaluations VALUES(27519,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(27519,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(27519,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(27519,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(27519,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(27519,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(27519,36,'cp 37','c5d4 c3d4 a8c8 c1b2 g8h6 b1c3 c6a5 c3a4 b6c6 a4c5 a5c4 b2c1 f8c5 d4c5 h6f5 f1d3 b7b6 c5b6 c6b6 e1g1 e8g8 f1e1 h7h6 h2h4 d7c6 d3f5 e6f5 c1f4 d5d4 d1d4 f8d8 d4b6 a7b6 h4h5 c6f3 g2f3 d8d3 e1c1 d3f3 f4e3 f3h3 a3a4 f5f4 e3f4 h3h5 a4a5 h5f5');
INSERT INTO evaluations VALUES(27539,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(27539,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(27539,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(27539,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(27539,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(27539,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(27539,36,'cp 44','c3d4 a8c8 c1b2 c6a5 b1d2 g8e7 a1c1 c8c1 b2c1 a5c6 d2b3 c6d8 f1d3 d7b5 e1g1 a7a6 c1e3 e7f5 h2h4 f8e7 e3g5 e7g5 h4g5 d8c6 d1d2 b5d3 d2d3 e8g8 f1c1 f8c8 g2g3 c6a7 c1c8 a7c8 d3c3 c8a7 c3c5 b6c5 b3c5 a7c6 c5b7 f5d4 f3d4');
INSERT INTO evaluations VALUES(27538,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(27538,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(27538,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(27538,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(27538,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(27538,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(27538,36,'cp 37','a8c8 c1b2 c6a5 b1d2 g8e7 a1c1 c8c1 d1c1 a5c6 f1d3 c6d8 c1c5 e7c8 c5c2 a7a5 b4a5 b6a5 e1g1 h7h6 f1b1 a5a4 c2c1 f8e7 h2h4 c8b6 b2c3 a4c6 d2f1 b6a4 b1b3 a4c3 b3c3 c6b6 f1g3 d8c6 g3h5 e8d8 h5g7');
INSERT INTO evaluations VALUES(46,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(46,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(46,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(46,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46,36,'cp 32','c1b2 g8e7 b1c3 c6a5 c3a4 b6c6 a4c5 a5c4 b2c3 b7b6 c5d7 c6d7 f1d3 g7g6 e1g1 f8g7 d1e2 e8g8 f1c1 c8c7 c3e1 f8c8 c1c2 a7a5 g2g3 h7h5 c2a2 a5a4 h2h3 g7h6 a2c2 b6b5 a1a2 g8h8 c2c3 e7f5 h3h4');
INSERT INTO evaluations VALUES(30785,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30785,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(30785,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(30785,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(30785,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(30785,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(30785,36,'cp 42','d2d4 g8f6 g2g3 e7e6 f1g2 f8b4 c1d2 b4d2 d1d2 d7d5 e1g1 e8g8 c4d5 e6d5 b1c3 b8d7 f1c1 c7c6 c1c2 f8e8 e2e3 f6e4 d2d1 d7f6 b2b3 f6g4 c3e4 d5e4 f3d2 g4f6 a2a4 d8e7 d2c4 c6c5 a4a5 b6b5 a5a6 b7c8 c4e5');
INSERT INTO evaluations VALUES(38336,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38336,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38336,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(38336,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38336,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(38336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38336,36,'cp 1','g8f6 f1d3 c7c6 c2c4 c6d5 e4d5 e8g8 e1g1 e7e5 b1c3 b8a6 d3e2 a6c5 b2b4 c5e4 c3e4 f6e4 c1b2 h7h5 a2a3 a7a5 e2d3 e4f6 f1e1 a5b4 a3b4 a8a1 b2a1 c8g4 h2h3 g4f3 d1f3 d8d7 a1c3 f8a8 e1a1 a8a1 c3a1 d7a4 a1c3 e5e4 d3e4 f6e4 c3g7 g8g7');
INSERT INTO evaluations VALUES(41193,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41193,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41193,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(41193,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41193,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41193,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41193,36,'cp 0','f1d3 c7c6 c2c4 c6d5 e4d5 e8g8 e1g1 e7e5 b1c3 b8a6 d3e2 a6c5 b2b4 c5e4 c3e4 f6e4 c1b2 h7h5 a2a3 a7a5 e2d3 e4f6 f1e1 f8e8 f3g5 f6g4 g5f3');
INSERT INTO evaluations VALUES(45239,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45239,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45239,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45239,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45239,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(45239,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(45239,36,'cp -14','c8d7 c1f4 b8c6 e1g1 a8c8 a1c1 h7h6 e4e5 c6e5 f3e5 d6e5 d1b3 e7e6 f4e5 d8b6 c3b5 f6g4 b5d6 f8d6 b3b6 a7b6 e5d6 d7c6 a2a4 g4f6 b2b3 f6d5 c4d5 e6d5 d6f4 e8g8 f1d1 f7f6 f4e3 b6b5 a4b5 c6b5 c1c8 f8c8 d1d5 c8a8 d5d1 b5c6 f2f3');
INSERT INTO evaluations VALUES(45753,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(45753,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(45753,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(45753,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(45753,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(45753,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(45753,36,'cp 21','d1e2 c8e6 f2f4 g7g6 g2g4 h7h6 g5h4 g6g5 f4g5 f6g4 g1f3 h6g5 h4g5 f8e7 e2d2 b8d7 h2h4 e7g5 f3g5 d8f6 f1g2 h8h6 e1c1 f6f2 d1f1 f2d2 c1d2 e8e7 c3e2 d7f6 e2f4 e6d7 f1e1 e7f8 a2a4 a8e8 h4h5 f8g7 e1e8 f6e8 h1g1 e8d6 a4a5 d6c4 d2c1 c4a5 g2f3');
INSERT INTO evaluations VALUES(53537,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(53537,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(53537,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(53537,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(53537,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(53537,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(53537,36,'cp -29','g1f3 b7b6 c5b6 a7b6 e2e3 g8f6 f1d3 b8d7 e1g1 f8d6 d1c2 c8b7 b2b3 e8g8 c1b2 d8e7 a2a4 c6c5 f1c1 f8c8 c2e2 e6e5 f3e5 d7e5 d4e5 d6e5 b2e5 e7e5 b1d2 e5e7 d2f3 c8c7 h2h3 h7h6 c1d1 c7c6 d3f5 b7c8 f5d3 c8e6 a1c1');
INSERT INTO evaluations VALUES(53536,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(53536,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(53536,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(53536,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(53536,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(53536,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(53536,36,'cp -35','b7b6 c5b6 a7b6 e2e3 b8d7 d1c2 c8b7 f1d3 g8f6 e1g1 f8d6 b2b3 d8e7 c1b2 e8g8 f1c1 f8c8 a2a4 c6c5 c2d1 e6e5 f3e5 d7e5 d4e5 d6e5 b2e5 e7e5 b1d2 g7g6 d1c2 e5e7 h2h3 h7h5 c2b2 f6d7 c1d1 d7e5 d3e2 c8d8 d2f3 e5c6 a1c1 c6b4 e2b5 a8c8');
INSERT INTO evaluations VALUES(4393,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4393,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4393,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4393,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4393,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(4393,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(4393,36,'cp 2','f1c4 d7d6 d2d3 c6d4 e1g1 c7c6 c4a2 a7a5 c1e3 d4f3 d1f3 c5e3 f3e3 c8e6 a2e6 f7e6 f2f4 e5f4 f1f4 d8c7 c3e2 f6d7 a1f1 a8e8 f4f8 e8f8 f1f8 d7f8 e3a7 d6d5 a7e3 c7e5 e2d4 a5a4 g1f2 d5e4 e3e4 e5e4 d3e4 g8f7 d4f3 f7f6 f2e3 e6e5 b2b3');
INSERT INTO evaluations VALUES(10245,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(10245,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(10245,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(10245,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(10245,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(10245,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(10245,36,'cp -11','c2c4 d7d6 f1e2 h7h6 a2a3 a7a5 c1e3 e8g8 e1g1 f6h7 b1c3 f7f5 e4f5 c8f5 d3d4 c6d4 f3d4 e5d4 e3d4 c7c6 f2f4 d8b6 d4c5 b6c5 g1h2 a8e8 d1d2 h7f6 a1d1 e8e6 d2d4 f8e8 d4c5 d6c5 e2d3 f5d3 d1d3 e6e3 f1d1');
INSERT INTO evaluations VALUES(15289,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(15289,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(15289,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(15289,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(15289,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(15289,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(15289,36,'cp 195','e1g1 c6e5 f1e1 d5e3 g5f7 e5f7 f2e3 f8e7 e3d4 e8g8 e1f1 b7b5 c4b5 c7c5 c1e3 e7f6 b1d2 c5d4 e3f4 c8f5 b5d3 f5d3 c2d3 f6e5 d2c4 e5f4 f1f4 d8d5 d1f3 d5f3 f4f3 f8e8 f3f5 a8d8 f5a5 f7d6 c4d6 d8d6 a5a7 d6c6 b2b4 c6c3 a2a4 c3d3 b4b5 d3c3');
INSERT INTO evaluations VALUES(15288,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(15288,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(15288,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(15288,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(15288,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(15288,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(15288,36,'cp 222','c6e5 f1e1 d5e3 g5f7 e5f7 f2e3 f8e7 e3d4 e8g8 e1f1 b7b5 c4b5 c7c5 c1f4 d8d4 d1d4 c5d4 b5d3 a7a5 b1d2 c8a6 d2c4 a5a4 a1e1 e7c5 f4g3 g7g6 e1e6 a6b5 c4d2 b5d3 c2d3 f7d8 f1f8 c5f8 e6b6 a4a3 b2b3 a8c8 d2c4 d8c6 g3d6 f8d6 c4d6 c8c7');
INSERT INTO evaluations VALUES(16557,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(16557,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16557,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(16557,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(16557,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(16557,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(16557,36,'cp -21','d7d5 e4d5 f6d5 b1c3 f8b4 c1d2 d5c3 d2c3 b4c3 b2c3 d8e7 g1e2 e8g8 e1g1 c8e6 f1e1 e7d6 d3d4 a8d8 d4e5 d6c5 d1b1 b7b6 e2f4 e6f5 e1e3 g7g5 f4h5 c6e5 b1b4 c5b4 c3b4 e5g4 e3e7 f5g6 h2h3 g6h5 h3g4 h5g4 e7c7 d8d2 c7a7 d2c2 a7a8 f8a8 g2a8');
INSERT INTO evaluations VALUES(15161,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(15161,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(15161,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(15161,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(15161,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15161,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(15161,36,'cp -8','e4d5 f6d5 b1c3 f8b4 c1d2 d5c3 d2c3 b4c3 b2c3 d8e7 g1e2 e8g8 e1g1 c8g4 d1b1 a8b8 f1e1 e7d6 a2a4 b7b6 e2c1 a7a6 c1b3 g4e6 b1b2 e6d5 g2d5 d6d5 b2a3 a6a5 b3d2 f8d8 a3b3 d5b3 d2b3');
INSERT INTO evaluations VALUES(19001,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(19001,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(19001,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(19001,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19001,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19001,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19001,36,'cp 2','b1c3 g8f6 f1b5 c6d4 b5c4 d7d6 d2d3 c8e6 g1e2 g7g6 b2c1 f8g7 c1g5 h7h6 g5f6 g7f6 c3d5 f6g7 a2a4 e8g8 e1g1 d4c6 d5e3 a7a6 c2c3 a8b8 d1d2 b7b5 c4d5 c6e7 d5e6 f7e6 b3b4 c5b4 c3b4');
INSERT INTO evaluations VALUES(31282,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(31282,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(31282,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(31282,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(31282,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(31282,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(31282,36,'cp 42','d4c5 b8c6 c2c3 g4f3 d1f3 e7e6 b2b4 d8c7 f1e2 g8e7 e1g1 g7g6 f1e1 f8g7 c1f4 e8g8 b1a3 c6e5 f3e3 e7f5 e3d2 c7d8 e2f1 e5c6 a3c2 a7a6 a2a4 d8f6 a1a3 h7h6 c2e3 f5e3 d2e3 d5d4 c3d4 c6d4 e3e4 d4c2 e4c2 f6f4 a3b3');
INSERT INTO evaluations VALUES(33747,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(33747,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(33747,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(33747,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(33747,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(33747,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(33747,36,'cp 0','c8g4 f1e2 g4e2 d1e2 b8c6 e1c1 e8c8 g1f3 e7e6 c1b1 a5b6 d2e3 b6a5 h1e1 h7h6 h2h3 f8b4 e3d2 a5b6 c3e4 b4a3 d2c1 a3b4');
INSERT INTO evaluations VALUES(38045,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38045,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(38045,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(38045,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(38045,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(38045,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(38045,36,'cp 33','g1f3 g8f6 f1d3 e8g8 e1g1 b7b6 c4d5 c6d5 c1d2 c8b7 a2a3 b4e7 d1e2 b8c6 f1c1 c6a5 a1b1 a8c8 c3b5 a5c6 d2e1 a7a6 b5c3 b6b5 c3a2 f6e4 c1c2 e7d6 b1c1 d8e7 a2c3 f7f5 c3a2 f5f4 a2b4 c6b4 e1b4 c8c2 e2c2 f4e3');
INSERT INTO evaluations VALUES(38046,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38046,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38046,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(38046,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38046,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(38046,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(38046,36,'cp 39','g8f6 d1c2 e8g8 a2a3 b4c3 c2c3 b7b6 f1d3 d5c4 d3c4 c8b7 e1g1 b8d7 f1d1 d8e7 b2b3 f8d8 c1b2 c6c5 a1c1 b7f3 g2f3 c5d4 c3d4 d7f8 d4c3 f8g6 d1d8 a8d8 c4e2 e6e5 c1d1 d8d1 e2d1 e7e6 b3b4 e6f5 c3c7');
INSERT INTO evaluations VALUES(38184,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(38184,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(38184,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(38184,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(38184,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(38184,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(38184,36,'cp 140','d8e7 d5e4 g8f6 e4e7 e8e7 c1f4 b8c6 e1c1 c8e6 f2f3 c6b4 d1e1 b4d5 c3d5 f6d5 g1h3 h8d8 f4g5 f7f6 g5d2 e7f7 f1c4 g7g5 e1e2 d5c7 c4e6 c7e6 c2c3 c5d6 h1e1 e6g7 h3f2 a8c8 f2e4 d6b8 d2e3 h7h5 h2h3 b7b5 a2a3 f7g6 c1c2 g7e6 g2g3 a7a6 e3b6 d8d7');
INSERT INTO evaluations VALUES(33024,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(33024,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(33024,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(33024,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(33024,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(33024,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(33024,36,'cp 133','d5e4 g8f6 e4e7 e8e7 c1f4 b8c6 e1c1 c8e6 f2f3 c6b4 d1e1 b4d5 c3d5 f6d5 f4g5 f7f6 f1c4 e7f7 g1h3 a8d8 g5d2 g7g5 h1f1 h8e8 h3f2 d5e3 d2e3 c5e3 e1e3 e6c4 f1e1 e8e3 e1e3 c4b5 b2b3 b5f1 g2g4 f1g2 f2e4 b7b5 e4g3 b5b4 g3f5 a7a5 e3e7 f7g8 f5h6 g8f8 e7h7 g2f3');
INSERT INTO evaluations VALUES(47766,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(47766,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(47766,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(47766,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(47766,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(47766,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(47766,36,'cp 21','f8e7 c1e3 c7c6 c2c4 b7b5 b1c3 b5b4 c3a4 b8a6 a2a3 e8g8 d5c6 b4a3 a1a3 a6b4 f1d3 b4c6 g1e2 c8e6 e2c3 c6d4 e1g1 a8b8 g1h1 d6d5 e4d5 e7a3 b2a3 e6d7 a4c5 d8e7 c5a6 b8b6 c4c5 b6a6 d3a6 e7c5 c3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(40744,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(40744,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(40744,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(40744,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(40744,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(40744,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(40744,36,'cp 31','c2c4 a7a5 c1e3 f6d7 b1c3 b8a6 f1d3 e7g5 e3f2 d8f6 g1e2 f6h6 a2a3 g5e3 e1g1 d7f6 e2g3 c8d7 b2b4 a5b4 a3b4 e8g8 a1b1 c7c5 d5c6 b7c6 d1e2 e3d4 c4c5 d4c3 d3a6 d6c5 f2c5 c3d4 c5d4 e5d4 a6c4 d7e6 f1d1 h6e3 e2e3 d4e3 c4e6');
INSERT INTO evaluations VALUES(51044,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(51044,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51044,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51044,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51044,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51044,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(51044,36,'cp 9','b5a4 g8f6 b1c3 f8c5 g1f3 b7b5 a4b3 d7d6 e1g1 c8e6 d2d3 e8g8 c3e2 c5b6 c2c3 h7h6 e2g3 b8d7 b3e6 f7e6 h2h3 d8e8 f1e1 a8d8 c1e3 b6e3 e1e3 c7c5 a2a4 b5b4 c3b4 c5b4 d3d4 a6a5 b2b3 e5d4 d1d4 d7c5 e4e5 f6d5 e5d6');
INSERT INTO evaluations VALUES(51060,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51060,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51060,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51060,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51060,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51060,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51060,36,'cp 0','g8f6 d2d3 f8c5 g1f3 e8g8 f3e5 d7d5 e1g1 d5e4 d3e4 f6e4 d1d8 f8d8 a4b3 e4f2 e5f7 f2d1 g1h1 d1f2');
INSERT INTO evaluations VALUES(52089,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52089,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52089,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52089,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52089,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52089,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52089,36,'cp 0','f8c5 b1c3 g8f6 d1f3 d7d6 f1d3 b8c6 g1e2 h7h6 e1g1 e8g8 h2h3 a7a6 f3g3 g8h8 g3h4 h8g8');
INSERT INTO evaluations VALUES(37706,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37706,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37706,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37706,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37706,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37706,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37706,36,'cp 0','b1c3 g8f6 d1f3 d7d6 f1d3 b8c6 g1e2 h7h6 h2h3 a7a6 a2a3 b7b5 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(37704,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37704,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37704,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37704,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37704,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37704,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37704,36,'cp 0','g8f6 d1f3 b8c6 c1g5 c6d4 f3d3 e8g8 e1c1 h7h6 g5f6 d8f6 c3d5 f6f2 c2c3 d4e6 g2g3 c5d6 d1d2 f2e1 d2d1');
INSERT INTO evaluations VALUES(39956,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39956,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39956,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39956,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39956,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39956,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39956,36,'cp 0','d1f3 b8c6 c1g5 c6d4 f3d3 e8g8 e1c1 h7h6 g5f6 d8f6 c3d5 f6f2 c2c3 d4e6 g2g3 c5d6 d1d2 f2e1 d2d1');
INSERT INTO evaluations VALUES(55614,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(55614,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55614,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55614,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(55614,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55614,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(55614,36,'cp 15','b1c3 b8c6 f1b5 g8e7 g1e2 a7a6 b5c6 e7c6 d2d4 e5d4 e2d4 f8e7 e1g1 c6d4 d1d4 e7f6 d4d3 c8e6 c1e3 b7b5 f1e1 e8g8 e3d4 b5b4 c3e2 a6a5 b2b3 f6h4 g2g3 h4f6 e2f4 a5a4 d4f6 d8f6 d3d2 a4b3 c2b3');
INSERT INTO evaluations VALUES(55596,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(55596,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55596,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(55596,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55596,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(55596,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(55596,36,'cp 4','b8c6 f1b5 g8e7 g1e2 a7a6 b5c6 e7c6 e1g1 f8e7 d2d4 c6d4 e2d4 e5d4 d1d4 e7f6 d4d3 c8e6 a2a4 e8g8 c1e3 d8d7 b2b3 a8e8 a1d1 d7c6 c3d5 e6d5 e4d5 c6d7 e3f2 h7h5 c2c4 e8e7 h2h3 f8e8 f1e1 e7e1 d1e1 e8e1 f2e1 g7g6 g1f1 g8f8 e1f2 f6b2 f1e2 b2g7 f2e3');
INSERT INTO evaluations VALUES(5101,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(5101,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(5101,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(5101,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(5101,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(5101,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(5101,36,'cp -28','g8e7 c1f4 d7d5 c4b5 e8g8 b5c6 e7c6 e4e5 h7h6 h2h3 c8d7 f1e1 a8c8 f4h2 g8h7 a2a4 a7a6 h2g3 c6e7 d3d4 c5d4 c3e2 e7f5 g3f4 d8b6 g2g4 f5e7 d1d4 b6d4 e2d4 g6g5 f4h2 e7c6 a4a5 c6d4 f3d4 c8c4 c2c3 d7b5');
INSERT INTO evaluations VALUES(7961,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(7961,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(7961,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(7961,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(7961,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(7961,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(7961,36,'cp -21','c1f4 d7d5 c4b5 e8g8 b5c6 e7c6 e4e5 h7h6 f1e1 c8d7 h2h3 a8c8 d1d2 g8h7 c3e2 f7f6 e5f6 d8f6 f4d6 f8f7 e2c3 f6d8 f3e5 f7f5 e5c6 c8c6 d6g3 d8g5 d2g5 f5g5 c3e2 c6b6 a1b1 g5f5 b2b3 e6e5 f2f3 d5d4 a2a4 h7g8 b1c1 d7e6');
INSERT INTO evaluations VALUES(8946,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(8946,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(8946,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(8946,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(8946,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(8946,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(8946,36,'cp 29','f1d3 e8g8 a2a3 b4c3 b2c3 f6e4 e1g1 e4c3 d1e1 c3e4 c2c4 b7b6 c4d5 e6d5 a1c1 c8b7 f3d2 e4d6 e1d1 h7h6 d1h5 c6e7 d2f3 b7c8 c1c2 c8e6 f1c1 a8c8 h2h3 c7c6 f3e5 c6c5 d4c5 c8c5 c2c5 b6c5 c1c5 d8b6 c5c2 d6c4 h3h4 f8c8 a3a4 b6b4 c2c1 b4a4');
INSERT INTO evaluations VALUES(8948,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(8948,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(8948,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(8948,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(8948,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(8948,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(8948,36,'cp 36','e8g8 e1g1 b4d6 f3e5 f6d7 e5d7 d8d7 f4d6 c7d6 a1c1 c6b4 d1d2 d7d8 d3e2 b4c6 f1d1 c6e7 e3e4 d5e4 c3e4 d6d5 e4g5 h7h6 g5f3 c8d7 f3e5 f7f6 e5f3 d7c6 e2d3 a7a6');
INSERT INTO evaluations VALUES(15629,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15629,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(15629,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(15629,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(15629,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(15629,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(15629,36,'cp 51','b1c3 f8e7 e1g1 e8g8 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 f8e8 h2h3 d7c6 e2f3 c6c4 d4d2 a8d8 b2b3 c4b4 g2g4 h7h6 f3g2 b7b5 a4b5 a6b5 f1e1 e6d7 f4e3 d7c6 e1e2 f6d7 c3d5 b4d2 e3d2 c6d5 e4d5 e7f6 e2e8 d8e8');
INSERT INTO evaluations VALUES(20314,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(20314,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(20314,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(20314,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(20314,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(20314,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(20314,36,'cp -41','e5d6 g8f6 b1c3 d8d6 e1g1 a7a6 f1e1 f8e7 c4d3 b7b6 c3e4 d6d7 e4f6 g7f6 d3f1 c8b7 c2c3 h8g8 g2g3 h7h5 f1g2 a8d8 d2d4 c5d4 c3d4 h5h4 f3h4 c6d4 c1e3 b7g2 h4g2 e7b4 d1d4 d7d4 e3d4 d8d4 e1d1 d4d1 a1d1 e8e7 a2a3 b4d6 h2h4 f6f5 g2e3 d6c5 b2b4');
INSERT INTO evaluations VALUES(20563,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(20563,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(20563,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(20563,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(20563,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(20563,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(20563,36,'cp -47','g8f6 d2d4 d8d6 c1e3 f8e7 b1c3 c5d4 f3d4 e8g8 e1g1 f8d8 d4c6 d6c6 d1e2 b7b6 c4b5 c6c7 f1d1 e7d6 h2h3 c8b7 e2c4 c7e7 b5a6 f6d5 e3d4 d5f4 a6b7 e7b7 c4f1 f4g6 d1d2 d6f4 d2d3 g6h4 a1d1 h7h6 d4e3 d8d3 d1d3 f4e3 d3e3 a8d8 e3e4 h4f5 f1e2');
INSERT INTO evaluations VALUES(29515,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29515,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(29515,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(29515,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(29515,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(29515,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29515,36,'cp 0','e7e6 d1b3 b8c6 b3b7 c6b4 a1c1 a8b8 b7a7 f8d6 f4d6 c7d6 a7a4 e8e7 a2a3 b4c2 c1c2 b8a8 a4b3 a8b8');
INSERT INTO evaluations VALUES(43146,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(43146,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(43146,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(43146,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(43146,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(43146,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(43146,36,'cp 91','e4e5 f6e4 e1g1 e8g8 c2c3 d7d5 c4e2 c7c6 d2d3 e4c5 f1e1 c5e6 d3d4 c6c5 b2b4 c5c4 a2a4 b8c6 b1a3 e6c7 c1f4 a7a6 f4g3 f7f6 b4b5 a6b5 a3b5 c8f5 d1d2 c7b5 a4b5 a8a1');
INSERT INTO evaluations VALUES(46238,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46238,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46238,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46238,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46238,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46238,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46238,36,'cp 0','c5d4 c3d4 f8e7 e1g1 e8g8 b1c3 d8b6 c3a4 b6a6 a4c5 a6b6');
INSERT INTO evaluations VALUES(55247,30,'cp -239',NULL);
INSERT INTO evaluations VALUES(55247,31,'cp -201',NULL);
INSERT INTO evaluations VALUES(55247,32,'cp -222',NULL);
INSERT INTO evaluations VALUES(55247,33,'cp -222',NULL);
INSERT INTO evaluations VALUES(55247,34,'cp -224',NULL);
INSERT INTO evaluations VALUES(55247,35,'cp -201',NULL);
INSERT INTO evaluations VALUES(55247,36,'cp -199','e7e5 g1e2 f7f5 b1c3 b8c6 d4e5 d8d1 c3d1 c8d7 c1f4 e8c8 d1e3 d8e8 e3c4 g8e7 e1c1 e7g6 f4e3 a7a6 h2h3 g6e5 e2f4 b7b5 c4e5 c6e5 f4h5 c8b7 f1e2 h8g8 d1d4 d7e6 a2a4 e8a8 h1d1 e6c4 e2c4 e5c4');
INSERT INTO evaluations VALUES(13001,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(13001,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(13001,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(13001,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13001,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13001,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13001,36,'cp -3','e1g1 f8g7 h2h3 e8g8 c1e3 a7a6 a2a4 b7b6 f1e1 c8b7 d1d2 e7e6 c4b3 d8c7 d3d4 c5d4 f3d4 c6a5 e3h6 g7h6 d2h6 d6d5 e4d5 a5b3 c2b3 f6d5 c3d5 b7d5 a1c1 c7b7 h6g5 a8c8 b3b4 b7d7 b4b5 a6b5');
INSERT INTO evaluations VALUES(13000,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13000,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13000,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13000,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13000,36,'cp 0','f8g7 c3d5 e8g8 d5f6 g7f6 a2a3 c8g4 h2h3 g4f3 d1f3 b7b5 c4d5 d8d7 c2c3 f6g7 c1e3 e7e6 d5c6 d7c6 a1d1 f7f5 f3g3 a7a5 e4f5 f8f5 d3d4 c5d4 e3d4 e6e5 d4e3 d6d5 g3g4 h7h5 g4e2 a8e8 f2f3 g8h7 e2d2 e5e4 f3e4');
INSERT INTO evaluations VALUES(18060,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(18060,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(18060,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(18060,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(18060,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(18060,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(18060,36,'cp 41','e5d4 d1d4 d8f6 d4a4 f8c5 b1c3 g8e7 e1g1 a8b8 f1e1 e8g8 c1e3 c5e3 e1e3 d7d5 a1b1 f6d6 b2b3 c8g4 b1d1 f7f5 e4d5 c6d5 d1e1 g4f3 e3f3 c7c5 f3d3 g8h8 a4a3 b8e8 c3a4 e7g6 e1e8 f8e8 a3c5');
INSERT INTO evaluations VALUES(18066,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(18066,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(18066,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(18066,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(18066,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(18066,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(18066,36,'cp 43','d1d4 d8f6 d4a4 f8c5 b1c3 g8e7 c1g5 f6e6 f3d4 c5d4 a4d4 f7f6 g5d2 d7d6 e1c1 c6c5 d4d3 c8b7 h1e1 e8c8 c1b1 h8e8 d2c1 e7c6 b2b3 e6e5 h2h4 e5d4 d3f3 d4e5 h4h5 h7h6 c1a3');
INSERT INTO evaluations VALUES(18065,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(18065,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(18065,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(18065,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(18065,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(18065,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18065,36,'cp 53','d8f6 d4a4 f8c5 e1g1 g8e7 b1c3 a8b8 f1e1 e7g6 e4e5 f6f5 a4e4 f5e4 c3e4 c5b6 h2h4 h7h5 b2b3 e8g8 c1e3 b6a5 c2c3 d7d5 e5d6 c7d6 f3d4 c8b7 d4f5 d6d5 b3b4 d5e4 b4a5 b7c8 f5g3 g6h4 g3e4 c8e6 e3b6 f8e8 e4c5 e6c4 c5d7 b8c8 e1e8 c8e8 a2a4');
INSERT INTO evaluations VALUES(21335,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21335,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21335,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21335,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21335,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(21335,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(21335,36,'cp 0','b7c6 d2d4 d5e4 f3e5 g8e7 e1g1 f7f6 e5c4 e7g6 f1e1 f6f5 f2f3 f8e7 f3e4 e8g8 e4f5 c8f5 c1e3 d8d5 d1e2 e7h4 b1c3 d5e6 e1f1 f5g4 f1f8 a8f8 e2d3 g4f5 d3e2');
INSERT INTO evaluations VALUES(22694,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(22694,31,'cp -130',NULL);
INSERT INTO evaluations VALUES(22694,32,'cp -152',NULL);
INSERT INTO evaluations VALUES(22694,33,'cp -144',NULL);
INSERT INTO evaluations VALUES(22694,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(22694,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(22694,36,'cp -142','d7d5 h5e2 g8f6 g1f3 f8c5 c1g5 c8e6 b1c3 c5b4 a2a3 b4c3 b2c3 h7h6 g5f6 d8f6 e2e3 e8c8 f1e2 d5e4 d3e4 d8d6 e1g1 h8d8 e2b5 c8b8 a1b1 b7b6 c3c4 c6d4 f3d4 d6d4 c4c5 f6f4 c5b6 a7b6 e3f4 e5f4 b5d3 e6c4 d3c4');
INSERT INTO evaluations VALUES(24596,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24596,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(24596,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24596,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24596,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24596,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24596,36,'cp 0','c1g5 e7e6 d4d3 f8e7 e1c1 d8a5 g5h4 e8g8 f3d4 f8e8 c1b1 a8c8 f2f3 a7a6 d4b3 a5e5 h4g3 e5h5 g3d6 e8d8 e4e5 f6e8 c3e4 e7d6 e5d6 c6d5 e4c3 d5c6');
INSERT INTO evaluations VALUES(24594,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24594,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24594,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24594,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24594,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24594,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24594,36,'cp 0','e7e6 g5f6 d8f6 d4e3 h7h5 h2h4 f8e7 e1c1 f6h6 f3g5 e6e5 f2f4 e5f4 e3f4 f7f6 g5h3 h6f4 h3f4 e8f7 h1f1 a8g8 f4h3 f7e6 h3f4');
INSERT INTO evaluations VALUES(24690,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24690,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24690,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24690,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24690,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24690,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24690,36,'cp 0','e1c1 f8e7 h1e1 e8g8 e4e5 d6e5 f3e5 d8c7 d4c4 h7h6 g5f4 e7d6 f4g3 f8c8 d1d2 c6e8 c4e2 d6b4 e5d3 b4d6');
INSERT INTO evaluations VALUES(24689,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24689,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24689,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24689,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24689,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24689,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24689,36,'cp 0','f8e7 h1e1 e8g8 e4e5 d6e5 f3e5 d8c7 d4c4 h7h6 g5f4 e7d6 f4g3 f8c8 d1d2 c6e8 c4e2 d6b4 e5d3 b4d6');
INSERT INTO evaluations VALUES(24027,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24027,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24027,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24027,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24027,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24027,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24027,36,'cp 0','h1e1 e8g8 e4e5 d6e5 f3e5 d8c7 d4c4 h7h6 g5f4 e7d6 f4g3 f8c8 d1d2 c6e8 c4e2 d6b4 e5d3 b4d6');
INSERT INTO evaluations VALUES(25363,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(25363,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(25363,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(25363,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(25363,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(25363,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(25363,36,'cp 153','f7e6 g5e6 d8d7 c3b5 a8c8 e6f8 h8f8 b5c3 c6d4 e1g1 e8f7 d2d3 f7g8 a2a4 d4e6 a4a5 f8f7 f2f3 c7c5 c3e2 d6d5 e4d5 f6d5 e2g3 c5c4 d3c4 c8c4 c2c3 e6c5 f1e1 d5f4 c1f4 e5f4 g3e4 c5e4 e1e4 c4e4 f3e4 d7b5 d1d5 b5b2');
INSERT INTO evaluations VALUES(25427,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(25427,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(25427,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(25427,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(25427,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(25427,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(25427,36,'cp 148','g5e6 d8d7 c3b5 a8c8 e6f8 h8f8 b5c3 c6d4 f2f3 e8f7 c3e2 c7c5 c2c3 d4e2 d1e2 f7g8 e1g1 b7b6 d2d3 h7h6 c1d2 a7a5 a2a4 f8f7 d2e3 d6d5 e3f2 f6h5 e4d5 h5f4 e2e4 d7d5 e4d5 f4d5 a1d1 c8d8 f2h4 d8d7 f1e1 d5f4 e1e5 f4d3');
INSERT INTO evaluations VALUES(25424,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(25424,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(25424,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(25424,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(25424,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(25424,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(25424,36,'cp 151','d8d7 c3b5 a8c8 e6f8 h8f8 b5c3 e8f7 e1g1 f7g8 d2d3 c6d4 f2f3 c7c5 c3e2 b7b6 c2c3 d4e2 d1e2 h7h6 c1e3 d6d5 c3c4 d5d4 e3d2 d7e6 g2g3 e6h3 a2a4 c8e8 a4a5 f8f7 b2b4 c5b4 d2b4 f6d7 a5b6 a7b6 a1a6 e8c8 a6a7');
INSERT INTO evaluations VALUES(27472,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(27472,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(27472,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(27472,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(27472,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(27472,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(27472,36,'cp 163','c3b5 a8c8 e6f8 h8f8 b5c3 e8f7 e1g1 f7g8 d2d3 c6d4 c3e2 c7c5 c2c3 d4e2 d1e2 b7b5 a2a3 a7a5 h2h3 a5a4 c1e3 h7h6 a1d1 d7e6 f2f3 f6h5 g1h2 f8f7 e2d2 e6g6 d2f2 g6e6 d1d2 c8f8 d3d4 c5d4 c3d4 d6d5 e4d5 e6d5');
INSERT INTO evaluations VALUES(30577,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(30577,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(30577,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(30577,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(30577,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(30577,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(30577,36,'cp 141','e4d5 e6d5 f3e5 g8f6 d1f3 f8e7 c1f4 e8g8 e1c1 f6e4 c3e4 d5e4 f3g3 e7h4 g3e3 d8e7 g2g3 h4f6 e5g4 b8d7 f1e2 c7c6 f2f3 f8e8 f3e4 e7e4 e3e4 e8e4 e2f3 e4e8 h1e1 f6e7 g4e3 e7f8 e3f5 e8e1 d1e1 g7g6 f5d6');
INSERT INTO evaluations VALUES(30570,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(30570,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(30570,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(30570,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(30570,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(30570,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(30570,36,'cp 123','e6d5 f3e5 g8f6 d1f3 f8e7 c1f4 e8g8 e1c1 f6e4 c3e4 d5e4 f3g3 e7h4 g3e3 d8e7 f1c4 b8d7 h1g1 d7e5 d4e5 e7e8 e3e2 b7c6 c1b1 a7a5 g2g3 h4e7 c4d5 c6d5 d1d5 e7c5 g1d1 e8c6 a2a3 c6g6 g3g4 a8e8 h2h4 g6e6 d5d7 e6c6 h4h5');
INSERT INTO evaluations VALUES(30585,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(30585,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(30585,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(30585,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(30585,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(30585,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(30585,36,'cp 149','f3e5 g8f6 d1f3 f8e7 c1f4 e8g8 e1c1 c7c5 d4c5 e7c5 f1c4 b8c6 e5c6 b7c6 c3d5 c6d5 c4d5 f6d5 d1d5 d8f6 h1d1 f8e8 f4g3 f6f3 g2f3 a7a5 c2c3 f7f6 d1d2 a5a4 c1c2 g8f7 f3f4 g7g6 f4f5 e8e1 d5d7 e1e7 d2d5 a8e8 d7e7 e8e7 b2b4 a4b3 a2b3 g6f5 b3b4');
INSERT INTO evaluations VALUES(42449,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42449,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42449,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42449,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42449,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42449,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42449,36,'cp 4','d1d3 c7c5 d4c5 b8a6 e2e3 a6c5 d3d4 d8e7 e1c1 d7d5 c3b5 c5a6 f4g5 c8d7 g1f3 h7h6 g5f6 g7f6 e3e4 b4c5 d4a4 c5f2 e4d5 e6d5 c1b1 e8f8 a4f4 f2e3 f4d6 e7d6 b5d6 a6c5 d1d5 d7c6 d5d1 c5e4 d6e4 c6e4');
INSERT INTO evaluations VALUES(54161,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54161,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54161,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54161,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(54161,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54161,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(54161,36,'cp 9','e4d5 e6d5 c4b5 b8c6 g1f3 g8f6 e1g1 f8e7 d4c5 e7c5 c1g5 e8g8 b1c3 c8e6 d1d3 a8c8 a1e1 h7h6 g5h4 e6g4 f3e5 c6e5 e1e5 g4e6 c3d5 e6d5 h4f6 d8f6 e5d5 f6b2 b5c4 c8c7 c2c3 g8h8 g1h1 c5e7 h2h3 b2b6 f1d1 b6f2');
INSERT INTO evaluations VALUES(54124,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54124,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54124,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54124,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(54124,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(54124,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(54124,36,'cp 10','e6d5 c4b5 b8c6 g1f3 f8d6 b1c3 g8e7 d4c5 d6c5 c1e3 c5e3 f2e3 e8g8 d1d2 a7a6 b5d3 b7b5 e1g1 c8g4 f3d4 c6e5 a2a4 b5b4 c3e2 g4e2 d2e2 e5d3 c2d3 a6a5 a1c1 a8c8 e2f2 c8c1 f1c1 d8b6 b2b3 h7h6 h2h3');
INSERT INTO evaluations VALUES(54504,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54504,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54504,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54504,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(54504,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54504,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(54504,36,'cp 0','c4b5 b8c6 g1f3 f8d6 d4c5 d6c5 e1g1 g8e7 b1d2 c8g4 d2b3 c5d6 f1e1 e8g8 a2a4 a7a6 b5e2 h7h6 h2h3 g4h5 c2c3 d6c7 c1e3 f8e8 e3c5 e7g6 g2g4 d8d7 g4h5 d7h3 h5g6 h3g4 g1h1 g4h3');
INSERT INTO evaluations VALUES(54464,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54464,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54464,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(54464,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(54464,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(54464,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54464,36,'cp 7','b8c6 g1f3 f8d6 b1c3 g8e7 d4c5 d6c5 c1e3 c5e3 f2e3 e8g8 d1d2 a7a6 b5d3 c8e6 e1c1 d8c7 c1b1 a8d8 a2a3 h7h6 c3e2 c6e5 f3e5 c7e5 e2d4 e7f5 d4f5 e6f5 d3f5 e5f5 h1f1 f5e6 f1f4 f8e8 d2f2 d8d6 d1d3 e6d7 f2d2 g8h8');
INSERT INTO evaluations VALUES(20466,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20466,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20466,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20466,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20466,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20466,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(20466,36,'cp 0','g1f3 f8d6 b1c3 g8e7 d4c5 d6c5 c1e3 c5e3 f2e3 e8g8 d1d2 a7a6 b5c6 b7c6 e1c1 c8f5 h1e1 d8a5 a2a3 f8d8 d2f2 a5b6 c3a4 b6a5');
INSERT INTO evaluations VALUES(9673,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(9673,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(9673,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(9673,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(9673,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(9673,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(9673,36,'cp 15','h4g3 d7d6 c2c3 e8g8 a2a4 a7a5 e1g1 c8e6 b1d2 c5a7 f1e1 d8e7 c4b3 f6d7 h2h3 e7f6 d2f1 a8e8 a1b1 h6h5 f1e3 a7e3 b3e6 f6e6 e1e3 h5h4 g3h2 e6g6 b2b4 a5b4 c3b4 c6d4 a4a5 d4e6');
INSERT INTO evaluations VALUES(9674,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(9674,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(9674,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(9674,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(9674,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(9674,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9674,36,'cp 36','d7d6 c2c3 a7a6 a2a4 c5a7 b1d2 e8g8 e1g1 g8g7 f1e1 g5g4 f3h4 f6h5 d2f1 d8f6 f1e3 a7e3 e1e3 c6e7 d3d4 h5g3 h2g3 e7g6 h4g6 f7g6 d1d2 c8e6 d4e5 d6e5 c4e6 f6e6 d2d5 e6d6 d5d6 c7d6 a1d1 f8f6 e3d3 a8d8 a4a5 h6h5 d3d5 g6g5');
INSERT INTO evaluations VALUES(8764,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(8764,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8764,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(8764,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(8764,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8764,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(8764,36,'cp 14','c2c3 a7a6 b1d2 c5a7 e1g1 d8e7 g1h1 e8g8 a2a4 f6h5 h2h3 e7f6 f3d4 c6d4 d1h5 d4e6 d3d4 g8g7 c4e6 f6e6 h5e2 c8d7 e2d3 a8e8 f2f3 f7f5 e4f5 e6d5 f5f6 f8f6 g3f2 d7f5 d2e4 f6f8 c3c4 d5f7 b2b4 f5g6 d4d5 a7f2 f1f2 g6e4 d3e4 g7h8');
INSERT INTO evaluations VALUES(9751,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(9751,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(9751,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(9751,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(9751,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(9751,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(9751,36,'cp 12','d2d3 a7a6 c1e3 d7d6 d3d4 e5d4 f3d4 c5d4 e3d4 e8g8 e1g1 c8e6 c4e2 b7b5 d4e3 b5b4 c3d5 f6e4 a2a3 a8b8 d5b4 c6b4 a3b4 b8b4 e2a6 e4c5 b2b3 c5a6 a1a6 b4e4 d1d2 f8e8 a6a7 c7c5 f1a1 d8b6 a7a8 e8a8 a1a8 g8h7 d2c3 e6f5');
INSERT INTO evaluations VALUES(13659,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(13659,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13659,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13659,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(13659,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13659,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(13659,36,'cp 17','f6h5 f4e3 f8d6 f1d3 h5f4 e3f4 d6f4 g2g3 f4d6 e1g1 c8d7 a2a3 e8g8 f1e1 g8h8 a1c1 a8c8 b2b4 c6e7 d1e2 d7e8 c3c4 d5c4 c1c4 c8c7 c4c7 d8c7 d2c4 e7d5 e1c1 g7g6 e2b2 f7f6 c1e1 c7d7 c4d6 d7d6');
INSERT INTO evaluations VALUES(26914,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(26914,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(26914,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(26914,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(26914,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(26914,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(26914,36,'cp 92','d2d4 g8f6 e1g1 g4d7 f1e1 f8e7 h2h3 h7h6 b1d2 e8g8 b5d3 f6h7 d4d5 c6b8 b2b4 a7a5 d2c4 a5b4 c3b4 f7f5 a2a4 d8e8 a4a5 f5e4 d3e4 d7f5 e4f5 f8f5 c4d2 b8a6 d2e4 f5f8 a1b1 e8f7 b4b5 a6b8 b1a1 h7f6 e4f6 e7f6 a1a3');
INSERT INTO evaluations VALUES(26905,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(26905,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(26905,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(26905,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(26905,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(26905,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(26905,36,'cp 75','g8f6 e1g1 g4d7 f1e1 f8e7 h2h3 e8g8 b5d3 h7h6 d4d5 c6b8 c1e3 a7a5 c3c4 b8a6 a2a3 f6h7 c4c5 h7g5 b1d2 g5f3 d2f3 f8e8 c5d6 e7d6 a1c1 b7b6 f3d2 a6c5 d3b1 d8f6 d2c4 f6g6 c4d6 c7d6 d1f3 a8c8 b2b4 a5b4 a3b4 c5a4 e3d2');
INSERT INTO evaluations VALUES(27503,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(27503,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(27503,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(27503,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(27503,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(27503,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(27503,36,'cp -75','b6d4 d1e2 a7a6 b1d2 d4a4 d2b3 d7b5 c1e3 g8e7 b3c5 b5d3 e2d3 a4b5 d3b5 a6b5 c5b7 e7c6 f2f4 f8e7 e3c5 c6a5 f1c1 a5b7 c5e7 e8d7 e7b4 h8c8 c1c8 a8c8 b2b3 b7d8 a2a4 d8c6 b4c5 b5a4 a1a4 c6e5 a4a7 d7c6 c5d4 e5d7 d4g7 c8b8 g7e5 d7e5 f4e5 d5d4 a7f7 c6d5 f7h7 b8b3 g1f2 d5e4 h7d7 b3b2 f2g3');
INSERT INTO evaluations VALUES(36504,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(36504,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(36504,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36504,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36504,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(36504,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36504,36,'cp 0','f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 e1g1 d7f6 e4d3 f6d5 e3g5 c7c5 d4c5 g7b2 g5h6 b2g7 h6g7 g8g7 f1e1 d5f6 d1e2 c8g4 a1b1 a8b8 e2e7 g4f3 e7d8 f8d8 g2f3 f6d7 c5c6 b7c6 b1b8 d8b8 e1e7 d7c5 e7a7 c5d3 c2d3 b8b2 g1g2 b2d2 a7c7 d2d3 c7c6 d3a3 c6b6 a3a8 b6d6 h7h6 d6d4 a8a2 g2g3 a2b2');
INSERT INTO evaluations VALUES(36704,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(36704,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36704,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36704,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(36704,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36704,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(36704,36,'cp 0','c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 e1g1 d7f6 e4d3 f6d5 e3c1 d8d6 f1e1 c7c5 d4c5 d6c5 d3f1 c8f5 c2c3 f8d8 d1b3 a7a5 a2a4 f5g4 f3e5 g4e6 e5f3');
INSERT INTO evaluations VALUES(45759,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45759,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45759,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45759,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(45759,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45759,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45759,36,'cp 0','b1c3 f8b4 f1e2 d5c4 e2c4 e8g8 e1g1 c8g4 c1g5 g4f3 d1f3 d8d4 g5f6 d4f6 f3f6 g7f6 f1d1 b4c3 b2c3 a7a5 d1d6 g8g7 a1e1 b8a6 e1e4 f8e8 e4e8 a8e8 c4a6 b7a6 g1f1 e8e4 d6c6 e4a4 c3c4 a4a2 c6a6 a2c2 a6a5 c2c4 f1e2 c4c2 e2e3 c2b2 a5d5 b2c2');
INSERT INTO evaluations VALUES(45757,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45757,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45757,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(45757,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45757,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45757,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45757,36,'cp 0','f8b4 f1d3 e8g8 e1g1 d5c4 d3c4 c8g4 c1g5 g4f3 d1f3 d8d4 g5f6 d4f6 f3f6 g7f6 f1d1 b4c3 b2c3 a7a5 d1d6 g8g7 a1e1 b8a6 e1e4 f8e8 e4e8 a8e8 c4a6 b7a6 g1f1 e8e4 d6c6 e4a4 c6a6 a4a2 c3c4 a2c2 a6a5 c2c4 f1e2 g7g6 h2h3 c4b4 e2f3 b4b2 a5a6 b2d2 a6a4 d2c2 a4a1 c2c3 f3f4');
INSERT INTO evaluations VALUES(40501,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40501,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40501,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40501,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40501,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(40501,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40501,36,'cp 0','f1e2 d5c4 e2c4 e8g8 e1g1 c8g4 c1g5 g4f3 d1f3 d8d4 g5f6 d4f6 f3f6 g7f6 f1d1 f8e8 g1f1 b4c3 b2c3 a7a5 d1d6 g8g7 a1e1 e8e1 f1e1 b8a6 d6d7 b7b5 c4f7 a6c5 d7c7 c5a6 c7b7 a6c5');
INSERT INTO evaluations VALUES(50736,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50736,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(50736,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(50736,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50736,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(50736,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50736,36,'cp 37','d2d4 c5d4 f3d4 f8e7 c4d3 g8f6 c2c4 b7b6 b1c3 c8b7 c1e3 b8d7 a1c1 d7e5 h2h3 e8g8 d1e2 e5d3 e2d3 a8c8 b2b3 f6d7 d3d2 d7c5 f2f3 e7f6 d4e2 d8e7 f1d1 f8d8 e2g3 f6e5 f3f4 e5c3 c1c3 c5e4 g3e4 b7e4');
INSERT INTO evaluations VALUES(50745,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50745,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(50745,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(50745,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(50745,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(50745,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(50745,36,'cp 27','c5d4 f3d4 f8e7 c4d3 b7b6 c2c4 c8b7 c1e3 g8f6 b1c3 b8d7 a1c1 e8g8 h2h3 f8e8 d1e2 e7f8 d3b1 a8c8 d4f3 c8c7 f1d1 d8a8 e3f4 e8c8 f4d6 c7c4 d6f4 f8b4 f4d2 b6b5 e4e5 b4c3 d2c3 b7f3 g2f3');
INSERT INTO evaluations VALUES(8150,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(8150,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(8150,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(8150,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(8150,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(8150,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(8150,36,'cp -25','a2a4 h7h6 g5h4 c8d7 h2h3 e8g8 h4g3 d8b6 e1g1 a7a6 c4b3 c6a5 f3d2 d7c6 f1e1 b6d8 b3a2 b7b5 d1e2 d8d7 a4b5 a6b5 c3d1 d6d5 e4d5 e6d5 d1e3 f8e8 d2f3 e7d6 g3d6 d7d6 c2c3 a8d8 d3d4 a5c4 d4c5 d6c5');
INSERT INTO evaluations VALUES(13195,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(13195,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(13195,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(13195,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(13195,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(13195,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(13195,36,'cp 129','f1b5 d8b6 b5a4 c8g4 e1g1 e7e6 d2d4 b6a6 b1c3 f8b4 d1e2 a6a5 a4c6 b7c6 c3e4 a5d5 e4g3 e8g8 f1d1 g4f3 g2f3 d5d7 c2c4 d7c7 c1g5 b4e7 e2e3 a7a5 g5f4 c7b7 b2b3 a5a4 a1b1 a4b3 a2b3 f8d8 g1g2 f6d7 d4d5 c6d5 c4d5 d7f8 d5d6 e7f6 g3e4 f8g6 e4f6 g7f6');
INSERT INTO evaluations VALUES(13066,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(13066,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(13066,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(13066,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(13066,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(13066,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(13066,36,'cp 136','d8b6 b5a4 c8g4 e1g1 e7e6 d2d4 b6a6 b1c3 f8b4 d1e2 a6a5 a4c6 b7c6 c3e4 a5d5 e4g3 g4f3 g2f3 e8g8 f1d1 d5a5 c2c4 a5c7 c1g5 b4e7 e2e3 f8d8 b2b3 d8d7 d1d2 a8d8 g5f4 e7d6 f4d6 c7d6 a1d1 h7h6 g1g2 d6e7');
INSERT INTO evaluations VALUES(39944,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(39944,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39944,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(39944,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(39944,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(39944,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(39944,36,'cp 28','d7d6 e1g1 e8g8 c1d2 b8c6 d4c6 b7c6 c3a4 b4d2 d1d2 e6e5 f1e1 f6h5 d3f1 h5f4 a1d1 d8g5 g2g3 f8d8 f1g2 f4h3 g2h3 g5d2 d1d2 c8h3 f2f3 g8f8 g3g4 h7h5 g4h5 f8e7 c2c4 h3e6 b2b3 d8h8 c4c5 a8d8 c5d6 d8d6 d2d6 e7d6 e1d1 d6e7 d1c1 e6d7 a4b2 h8h5 b2c4 f7f6');
INSERT INTO evaluations VALUES(41328,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(41328,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41328,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(41328,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(41328,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(41328,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41328,36,'cp 32','b1c3 e8g8 e2e4 b8c6 f4e3 a7a6 d4d5 c6b8 a2a4 e7e6 f1c4 e6d5 c4d5 f6d5 e4d5 b8d7 e1g1 d7e5 f3e5 g7e5 e3d4 f8e8 f1e1 c8f5 d4e5 e8e5 e1e5 d6e5 d1e2 d8e7 g2g4 f5d7 e2c4 c7c6 a1e1 a8e8 e1e3 h7h5 g4h5 e7g5 g1f1 c6d5 c3d5');
INSERT INTO evaluations VALUES(53630,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(53630,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(53630,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(53630,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(53630,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(53630,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(53630,36,'cp 34','c6c5 c4d5 c5d4 f1b5 c8d7 b5d7 b8d7 d5e6 f7e6 d1d4 d8c7 b1c3 f8c5 d4e4 c7e5 g1f3 e5e4 c3e4 c5b4 c1d2 b4d2 e4d2 g8f6 e1g1 e8g8 f1e1 d7c5 f3e5 f6d5 d2f3 d5b4 e1e2 a8d8 g2g3 d8d5 a1c1 b7b6 a2a3 b4d3 c1c3 d3e5 f3e5 a7a5 g1g2 f8d8 e2e1 d5d2');
INSERT INTO evaluations VALUES(54280,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(54280,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(54280,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(54280,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(54280,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(54280,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54280,36,'cp 32','c4d5 c5d4 f1b5 c8d7 b5d7 b8d7 d5e6 f7e6 d1d4 d8c7 b1c3 f8c5 d4e4 c7e5 g1f3 e5e4 c3e4 c5b4 c1d2 b4d2 e4d2 g8f6 e1g1 e8g8 a1c1 f8d8 f1e1 a8c8 d2b3 e6e5 g1f1 c8c1 e1c1 e5e4 f3d4 d7e5 d4e6 d8e8');
INSERT INTO evaluations VALUES(53996,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(53996,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(53996,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(53996,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(53996,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(53996,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(53996,36,'cp 201','d4c5 d6c5 d1d8 e8d8 c1f4 b8d7 h2h4 g8e7 a2a4 e7c6 h4h5 f7f6 b1c3 a7a6 e1c1 d8e8 f4e3 b7b6 f3d2 f8d6 f2f4 c6d4 d2f3 c8b7 h1h3 e8e7 f3d4 c5d4 d1d4 d6c5 d4d1 c5e3 h3e3 g7g5 f4f5 d7e5 f5e6 e7e6');
INSERT INTO evaluations VALUES(53959,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(53959,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(53959,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(53959,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(53959,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(53959,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(53959,36,'cp 180','d8a5 b1c3 a5c5 c1e3 c5c7 c3b5 c7d8 e3g5 f7f6 g5f4 e6e5 f4e3 a7a6 b5c3 c8e6 f1e2 g8e7 d1d2 e6f7 e1g1 b8d7 f1d1 a8c8 a2a3 d7b6 d2d6 d8d6 d1d6 e7c6 d6d2 b6c4 e2c4 f7c4 c3a4 c8d8 d2d8 c6d8 a4b6 c4e6 c2c4 d8c6 b2b4 f8e7 f3d2 e8f7 f2f3 h8d8 b6d5');
INSERT INTO evaluations VALUES(56143,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(56143,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(56143,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(56143,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56143,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(56143,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56143,36,'cp -20','g8f6 d2d3 f8b4 d3e4 f6e4 d1d4 d8e7 g1e2 e8g8 c1d2 e4d2 d4d2 b4c5 d2d3 b8d7 e1c1 d7f6 h2h3 e7e3 d3e3 c5e3 c1b1 h7h5 d5d6 c7d6 g2g3 e3c5 f1g2 f8e8 a2a3 a7a5 e2c1 c8f5 c1b3 e8e3 g3g4 h5g4 h3g4 f5g4 b3c5 e3c3 b2c3 g4d1 h1d1');
INSERT INTO evaluations VALUES(48576,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48576,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48576,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(48576,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48576,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48576,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48576,36,'cp -10','f1b5 c7c6 d5c6 b8c6 d2d4 f8b4 g1e2 e8g8 e1g1 c6e7 a2a3 b4c3 e2c3 a7a6 b5c4 e7f5 f1e1 d8d4 d1d4 f5d4 c4b3 c8e6 c1e3 d4b3 c2b3 e6b3 e3d4 f6d5 e1e4 f7f5 e4e6 a8d8 c3d5 b3d5 e6e7 f8f7 e7f7 d5f7 d4c3 f7d5 g1f2 d8e8 a1e1 e8e1 c3e1');
INSERT INTO evaluations VALUES(26210,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26210,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26210,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26210,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26210,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26210,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26210,36,'cp 0','f3d4 g7g6 e1g1 f8g7 c2c4 d8b6 c1e3 b6b2 b1d2 a8c8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(26208,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26208,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26208,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26208,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26208,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26208,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26208,36,'cp 0','g7g6 e1g1 f8g7 c2c4 d8b6 c1e3 b6b2 b1d2 a8c8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(29098,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(29098,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(29098,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(29098,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(29098,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(29098,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(29098,36,'cp 134','h2h3 g4f3 d1f3 d6d5 e4e5 f6d7 f1d3 e7e6 c3e2 c6c5 c2c3 c5d4 c3d4 f8b4 e1f1 d8e7 f3h5 b8c6 c1g5 g7g6 h5h4 e7f8 g5h6 f8e7 h4f4 f7f6 e5f6 d7f6 a1c1 b4d6 f4e3 e8f7 g2g4 h8e8 f2f4 f7g8 f1g2 a8c8 h1f1 e7f7 h6g5 c6b4 d3b1 c8c1 e2c1 e8f8 a2a3 b4c6 c1d3 c6a5 b1a2 a5c4 e3e2 f8c8 d3e5');
INSERT INTO evaluations VALUES(29096,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(29096,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(29096,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(29096,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(29096,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(29096,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(29096,36,'cp 150','g4f3 d1f3 d6d5 e4e5 f6d7 c1g5 d8b6 e1c1 e7e6 c1b1 h7h6 g5e3 a7a5 a2a4 b8a6 h3h4 f8b4 e3c1 e8c8 c3a2 f7f6 e5f6 d7f6 f3b3 f6e4 c2c3 b4c5 b3c2 c5e7 f1d3 e4d6 d1e1 a6c7 c1f4 h8f8 f4g3 d6c4 g3c7 c8c7 d3c4 d5c4 e1e6 e7f6');
INSERT INTO evaluations VALUES(30691,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(30691,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30691,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30691,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30691,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30691,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30691,36,'cp 0','g8f6 e4e5 f6d5 c3e4 f7f5 e5f6 d5f6 c2c3 f6e4 d3e4 b7e4 f3g5 e4b7 d1c2 g7g6 g5h7 h8h7 c2g6 h7f7 c1g5 f8e7 h2h4 b8c6 g6g8 f7f8 g8g6');
INSERT INTO evaluations VALUES(28768,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28768,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28768,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28768,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28768,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(28768,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28768,36,'cp 10','e4e5 f6d5 c3e4 f7f5 e5f6 d5f6 c2c3 f6e4 d3e4 b7e4 f3g5 e4b7 f1d3 d8f6 g5h7 f6e5 d1e2 e5e2 e1e2 f8e7 h2h4 b7a6 h7g5 a6d3 e2d3 b8a6 c1e3 a6c5 d3c2 e7g5 e3g5 d7d6 g2g4 e8g8 f2f4 c5e4 h1f1 e4g5 h4g5 a8e8 a1e1 b6b5 e1e4');
INSERT INTO evaluations VALUES(36657,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(36657,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(36657,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(36657,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(36657,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(36657,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(36657,36,'cp 46','c7c5 d4c5 d6c5 d2d8 f8d8 e3c5 b8c6 g1e2 b7b6 c5e3 c8a6 e2f4 a6f1 h1f1 f6e8 c3d1 c6b4 f1f2 g7e5 a2a3 b4c6 a1c1 e8d6 b2b3 g6g5 f4e2 h7h6 d1c3 a8c8 c1d1 f7f6 g2g3 b6b5 f3f4 e5c3 e2c3 b5b4 a3b4 c6b4 e3d4 g5f4 g3f4 d6e4 c3e4 b4c2 f2c2 c8c2 d4f6 d8d1 e1d1 c2h2');
INSERT INTO evaluations VALUES(43078,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(43078,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(43078,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(43078,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(43078,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(43078,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(43078,36,'cp 65','b1c3 e8g8 f1d3 d7d6 h2h4 e6e5 d4e5 d6e5 h4h5 c8g4 h5g6 e7g6 d3e2 d8d1 c3d1 b8c6 d1e3 g4d7 c1d2 a8d8 a1d1 c6d4 d2c3 d7a4 b2b3 d4e2 e1e2 g6f4 e2e1 a4c6 e3d5 c6d5 c4d5 f7f5 g2g3 f5e4 g3f4 e4f3 f4e5');
INSERT INTO evaluations VALUES(48666,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(48666,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(48666,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(48666,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(48666,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(48666,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(48666,36,'cp -13','f8d6 f1d3 d8e7 d3e2 c8g4 e1g1 e8g8 f1e1 f8e8 c1e3 c7c6 b1d2 e7c7 d1c2 g4h5 e2d3 b8d7 e3g5 a7a6 g5h4 h5g6 d3g6 h7g6 h4g3 e8e1 a1e1 d6g3 h2g3 a8e8 e1e8 f6e8 c3c4 d7f8 c4d5 c6d5 c2b3 f8e6 d2f1 e8f6');
INSERT INTO evaluations VALUES(41706,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(41706,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(41706,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(41706,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(41706,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(41706,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(41706,36,'cp -2','f1d3 d8e7 d3e2 e8g8 e1g1 c8g4 f1e1 f8e8 c1e3 c7c6 b1d2 e7c7 d1c2 b8d7 e2d3 g4h5 e3g5 a7a6 a2a4 h5g6 g2g3 g6d3 c2d3 h7h6 g5f6 d7f6 g1g2 g7g6 h2h3 a8d8 b2b4 a6a5 e1e8 d8e8 b4a5 c7a5 c3c4 d6b4 c4c5');
INSERT INTO evaluations VALUES(41707,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(41707,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(41707,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(41707,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(41707,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(41707,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(41707,36,'cp -13','d8e7 d3e2 e8g8 e1g1 c8g4 f1e1 f8e8 c1e3 c7c6 b1d2 e7c7 d1c2 b8d7 e2d3 a7a5 a2a4 g4h5 e3g5 h7h6 g5f6 d7f6 h2h3 h5f3 d2f3 g7g6 g2g3 g8g7 e1e8 a8e8 a1e1 e8e1 f3e1 c7e7 e1g2 e7e6 h3h4 f6e4 d3e4 e6e4');
INSERT INTO evaluations VALUES(50616,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(50616,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(50616,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(50616,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(50616,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(50616,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(50616,36,'cp -22','e7e6 d2d3 b8c6 g1f3 d7d5 c4b3 g8f6 d1e2 h7h6 e1g1 f8e7 c2c3 d8c7 e4e5 f6h7 b3c2 c8d7 b1d2 g7g5 d2b3 g5g4 f3e1 c6e5 d3d4 c5d4 c3d4 e5c6 e2g4 e8c8 g4e2 e7d6 g2g3 h7g5 e1d3 c8b8 c1e3 f7f5');
INSERT INTO evaluations VALUES(50542,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(50542,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(50542,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(50542,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(50542,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(50542,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(50542,36,'cp -29','d2d3 b8c6 c4b3 g8f6 g1f3 d7d5 b1c3 d5e4 c3e4 f8e7 c2c3 e8g8 e1g1 b7b6 e4f6 e7f6 c1f4 c8b7 d1e2 c6e7 f1d1 e7f5 d3d4 c5d4 f3d4 f5d4 c3d4 a8c8 h2h3 h7h6 e2g4 b7d5 b3d5 d8d5 f4h6 c8c2 d1d2 c2d2');
INSERT INTO evaluations VALUES(50540,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(50540,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(50540,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(50540,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(50540,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(50540,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(50540,36,'cp -24','b8c6 b1c3 g8f6 g1f3 d7d5 c4b3 d5e4 c3e4 b7b6 e1g1 f8e7 c2c3 c8b7 e4f6 e7f6 c1f4 e8g8 d1e2 c6e7 a1d1 e7f5 f1e1 f8e8 h2h3 d8d7 f3e5 d7e7 e5g4 f5h4 f2f3 f6g5 f4g5 e7g5 d3d4 a8d8 d4c5 g5c5 g1h1 d8d1');
INSERT INTO evaluations VALUES(17403,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(17403,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(17403,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(17403,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(17403,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(17403,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(17403,36,'cp -12','c4b3 g8f6 g1f3 d7d5 b1c3 d5e4 c3e4 f8e7 e1g1 e8g8 c2c3 b7b6 d1e2 c8b7 e4f6 e7f6 c1f4 c6e7 d3d4 b7f3 e2f3 c5d4 a1d1 b6b5 a4b5 d8b6 c3d4 b6b5 f4d6 a6a5 d4d5 a5a4 b3c2 e6d5 d6e7 f6e7 d1d5 b5b4 f1d1 a8c8 f3e4');
INSERT INTO evaluations VALUES(55347,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(55347,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(55347,32,'cp -149',NULL);
INSERT INTO evaluations VALUES(55347,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(55347,34,'cp -189',NULL);
INSERT INTO evaluations VALUES(55347,35,'cp -190',NULL);
INSERT INTO evaluations VALUES(55347,36,'cp -194','h7h6 c2c3 b8c6 d2d4 f8g7 b1d2 g8f6 a2a4 e8g8 c4d3 c6e7 b2b4 e7g6 h2h3 c7c6 a4a5 f8e8 f1e1 c8e6 c1b2 a8c8 d2f1 f6h7 e4e5 h6h5 e5d6 d8d6 d1c2 g5g4 h3g4 h5g4 f3e5 h7f8 d3g6 f7g6 e5g6 f8g6');
INSERT INTO evaluations VALUES(55393,30,'cp -160',NULL);
INSERT INTO evaluations VALUES(55393,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(55393,32,'cp -176',NULL);
INSERT INTO evaluations VALUES(55393,33,'cp -177',NULL);
INSERT INTO evaluations VALUES(55393,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(55393,35,'cp -158',NULL);
INSERT INTO evaluations VALUES(55393,36,'cp -160','d2d4 b8c6 b1c3 f8g7 c3d5 g8e7 d5e7 c6e7 c2c3 c8g4 d1d3 d8d7 c1d2 h8f8 a2a4 e8c8 f1e1 c8b8 b2b4 d6d5 e4d5 e7d5 a4a5 g4f5 d3f1 f8e8 a5a6 b7b6 e1e8 d8e8 a1e1 e8e1 f1e1 f5e6 f3e5 g7e5 e1e5 f7f6');
INSERT INTO evaluations VALUES(55394,30,'cp -187',NULL);
INSERT INTO evaluations VALUES(55394,31,'cp -174',NULL);
INSERT INTO evaluations VALUES(55394,32,'cp -180',NULL);
INSERT INTO evaluations VALUES(55394,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(55394,34,'cp -195',NULL);
INSERT INTO evaluations VALUES(55394,35,'cp -201',NULL);
INSERT INTO evaluations VALUES(55394,36,'cp -195','f8g7 b1c3 b8c6 c3d5 c6e7 d5e7 g8e7 c2c3 c8g4 d1d3 d8d7 c1d2 g4h5 a1e1 c7c6 a2a4 e8g8 d3e2 a8e8 e2f2 d6d5 e4d5 c6d5 c4b5 e7c6 h2h4 f7f6 b5d3 e8e1 d2e1 c6e7 d3b5 d7d8 b5d3 g5g4 f3d2 h5g6');
INSERT INTO evaluations VALUES(38514,30,'cp -185',NULL);
INSERT INTO evaluations VALUES(38514,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(38514,32,'cp -180',NULL);
INSERT INTO evaluations VALUES(38514,33,'cp -184',NULL);
INSERT INTO evaluations VALUES(38514,34,'cp -179',NULL);
INSERT INTO evaluations VALUES(38514,35,'cp -191',NULL);
INSERT INTO evaluations VALUES(38514,36,'cp -182','b1c3 b8c6 c3d5 c6e7 d5e7 g8e7 c2c3 c8g4 d1c2 d8d7 c1d2 g4e6 c4e6 f7e6 a2a4 e7g6 d2e1 e6e5 c2b3 g5g4 f3h4 g6h4 e1h4 d7f7 b3b7 e8g8 d4e5 g7e5 b2b4 g8h7 h4g3 h6h5 a1d1 h7h8 b4b5 f7g6 g3h4 f4f3 b7c7');
INSERT INTO evaluations VALUES(38512,30,'cp -196',NULL);
INSERT INTO evaluations VALUES(38512,31,'cp -183',NULL);
INSERT INTO evaluations VALUES(38512,32,'cp -198',NULL);
INSERT INTO evaluations VALUES(38512,33,'cp -175',NULL);
INSERT INTO evaluations VALUES(38512,34,'cp -190',NULL);
INSERT INTO evaluations VALUES(38512,35,'cp -185',NULL);
INSERT INTO evaluations VALUES(38512,36,'cp -189','b8c6 c3d5 c6e7 d5e7 g8e7 c2c3 c8g4 d1c2 d8d7 c1d2 g4e6 c4b3 e7g6 b3e6 d7e6 c2a4 c7c6 a4b3 e8g8 h2h4 g5h4 a1e1 h4h3 b3e6 f7e6 g2h3 e6e5 g1f2 g8f7 f2e2 g7f6 d4e5 d6e5 e1d1 a8d8 c3c4 f8g8 d2c3 d8d1 f1d1 f7e6 d1g1 f6e7 b2b4 b7b6 a2a4 h6h5');
INSERT INTO evaluations VALUES(5709,30,'cp -205',NULL);
INSERT INTO evaluations VALUES(5709,31,'cp -190',NULL);
INSERT INTO evaluations VALUES(5709,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(5709,33,'cp -198',NULL);
INSERT INTO evaluations VALUES(5709,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(5709,35,'cp -194',NULL);
INSERT INTO evaluations VALUES(5709,36,'cp -177','c3d5 c6e7 d5e7 g8e7 c2c3 c8g4 d1c2 d8d7 c1d2 g4e6 c4d3 c7c5 a1e1 c5c4 d3e2 d6d5 e4d5 e6d5 b2b3 b7b5 b3c4 b5c4 h2h4 g5g4 f3e5 g7e5 d4e5 f4f3 e2d1 d7e6 g2g3 a8d8 d2e3 e7f5 c2f2 e8g8 e3f4 e6g6 d1c2 f8e8 f2c5 d5e6 c5f2');
INSERT INTO evaluations VALUES(55683,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(55683,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(55683,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(55683,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(55683,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(55683,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(55683,36,'cp -15','g8f6 d1d4 b8d7 f1b5 f8e7 e1g1 a7a6 b5d7 d8d7 c3c4 b7b5 b1c3 e8g8 d4d3 b5c4 d3c4 c7c5 c1g5 d7b7 f1d1 c8e6 c4e2 a8d8 a1b1 f8e8 e2c2 h7h6 g5f6 e7f6 c3d5 e6d5 d1d5 e8e6 b2b3 d8e8 f3d2 f6d4 b1d1 b7b5');
INSERT INTO evaluations VALUES(47954,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(47954,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(47954,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(47954,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(47954,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(47954,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(47954,36,'cp -20','d1d4 b8d7 f1b5 f8e7 e1g1 e8g8 f1e1 f8e8 b5c4 d7c5 c1f4 c8e6 c4f1 a7a5 h2h3 e7f8 f4g3 h7h6 e4e5 d6e5 g3e5 c5d7 e5g3 c7c6 b1d2 d7c5 d4d8 e8d8 g3c7 d8c8 c7e5 f6d7 e5h2 e6d5 f3d4 a5a4 b2b4 c5e6 d4e6 d5e6');
INSERT INTO evaluations VALUES(57042,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(57042,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(57042,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(57042,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(57042,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(57042,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(57042,36,'cp -48','d7d5 e4d5 e6d5 h3f4 g8f6 d2d4 f8d6 f1d3 e8g8 e1g1 b8c6 c2c3 c8g4 f2f3 g4d7 b1a3 f8e8 g2g3 a7a6 a3c2 h7h6 f1e1 e8e1 d1e1 c6e7 e1e2 d7f5 d3f5 e7f5 f4d3 d8d7 c1f4 d6f4 d3f4 f5d6 f4d3 a8e8 e2f2 b7b6 a1e1 a6a5');
INSERT INTO evaluations VALUES(55913,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(55913,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(55913,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(55913,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55913,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(55913,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(55913,36,'cp -45','e4d5 e6d5 h3f4 g8f6 d2d4 f8d6 f1d3 b8c6 b1c3 c6b4 d1e2 d6e7 c1e3 e8g8 e1c1 c7c6 f2f3 f8e8 c1b1 e7d6 e2d2 d8c7 d1e1 h7h5 a2a3 b4d3 f4d3 c8f5 h2h3 a7a5 e3g5 f6h7 g5h4 b7b5 e1e8 a8e8 h1e1 e8e1 d3e1 f5g6 e1d3');
INSERT INTO evaluations VALUES(2960,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(2960,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(2960,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(2960,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(2960,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(2960,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(2960,36,'cp 36','c4d5 e6d5 e1g1 f6e8 g5e7 d8e7 f1e1 d7f6 f3e5 c8e6 f2f3 f6d7 e5d7 e6d7 d1d2 e7f6 a1c1 e8d6 d2f2 f8e8 b2b4 a7a6 a2a4 e8e7 h2h3 d6f5 b4b5 a6b5 a4b5 a8a3 d3f5 f6f5 b5c6 b7c6');
INSERT INTO evaluations VALUES(9663,30,'cp -1033',NULL);
INSERT INTO evaluations VALUES(9663,31,'cp -1032',NULL);
INSERT INTO evaluations VALUES(9663,32,'cp -1052',NULL);
INSERT INTO evaluations VALUES(9663,33,'cp -1086',NULL);
INSERT INTO evaluations VALUES(9663,34,'cp -1078',NULL);
INSERT INTO evaluations VALUES(9663,35,'cp -1040',NULL);
INSERT INTO evaluations VALUES(9663,36,'cp -1098','d8h4 g2g3 e4g3 h2g3 h4h1 c4f1 e8g8 d1e2 c6d4 e2g2 h1h5 f1e2 h5f7 b1c3 d4c2 e1d1 c2a1 b2b3 f7g6 d2d3 d7d6 c3d5 f8f7 g3g4 h7h5 d5e3 h5g4 g2g3 c8e6 c1b2 f7f3 g3g1 a8f8 b2a1 g4g3 a1c3 e6g4 e3g4 g6g4 c3e1 g3g2 d1c2 e5e4 c2c1 e4d3 e2f3');
INSERT INTO evaluations VALUES(535,30,'cp -1082',NULL);
INSERT INTO evaluations VALUES(535,31,'cp -1094',NULL);
INSERT INTO evaluations VALUES(535,32,'cp -1098',NULL);
INSERT INTO evaluations VALUES(535,33,'cp -1086',NULL);
INSERT INTO evaluations VALUES(535,34,'cp -1105',NULL);
INSERT INTO evaluations VALUES(535,35,'cp -1101',NULL);
INSERT INTO evaluations VALUES(535,36,'cp -1117','g2g3 e4g3 h2g3 h4h1 c4f1 e8g8 d1e2 c6d4 e2g2 h1h5 b1a3 f8f7 d2d3 d7d5 c1e3 f7f1 g2f1 d4f3 e1d1 f3h2 d1d2 h2f1 a1f1 d5d4 e3d4 e5d4 d2c1 h5e2 f1h1 e2f3 h1e1 f3g3 e1h1 g3g2 h1e1 g2g5 c1b1 b7b5 e1e8 g8f7 e8e1 a7a5 e1d1 c8g4 d1e1 h7h5 e1c1 g4f5 c1e1 a5a4 e1c1');
INSERT INTO evaluations VALUES(533,30,'cp -1119',NULL);
INSERT INTO evaluations VALUES(533,31,'cp -1110',NULL);
INSERT INTO evaluations VALUES(533,32,'cp -1135',NULL);
INSERT INTO evaluations VALUES(533,33,'cp -1132',NULL);
INSERT INTO evaluations VALUES(533,34,'cp -1132',NULL);
INSERT INTO evaluations VALUES(533,35,'cp -1117',NULL);
INSERT INTO evaluations VALUES(533,36,'cp -1137','e4g3 h2g3 h4h1 c4f1 e8g8 d1e2 c6d4 e2g2 h1h5 f7h6 h5h6 b1a3 d7d5 d2d3 h6h5 c2c4 d4f3 e1d1 c8g4 f1e2 h5g6 c1e3 d5d4 e3d2 e5e4 d1c1 f3d2 c1d2 g4e2 d2e2 e4d3 e2d2 a8e8 a1h1 g6g4 d2d3 c7c6 d3d2 d4d3 g2h2 f8f3 h2h7 g8f7 h7h5 g4h5 h1h5 e8e2 d2c3 d3d2 c3c2');
INSERT INTO evaluations VALUES(539,30,'cp -1127',NULL);
INSERT INTO evaluations VALUES(539,31,'cp -1140',NULL);
INSERT INTO evaluations VALUES(539,32,'cp -1132',NULL);
INSERT INTO evaluations VALUES(539,33,'cp -1117',NULL);
INSERT INTO evaluations VALUES(539,34,'cp -1128',NULL);
INSERT INTO evaluations VALUES(539,35,'cp -1128',NULL);
INSERT INTO evaluations VALUES(539,36,'cp -1117','h2g3 h4h1 c4f1 e8g8 d1e2 c6d4 e2g2 h1h5 d2d3 h5f7 b1c3 d4c2 e1d2 c2a1 f1e2 d7d5 b2b3 d5d4 c1b2 d4c3 b2c3 a1b3 a2b3 f7b3 g2e4 b3a2 d2c1 f8e8 e2h5 e8e7 h5d1 a2a3 c1d2 a3c5 d1c2 g7g6 c2b3 g8g7 g3g4 c5f2 d2c1 f2f6 c3d2 a7a5 g4g5 f6c6 e4c6 b7c6 b3a4 c8f5 a4c6 a8d8 c6e4 f5e4 d3e4 a5a4');
INSERT INTO evaluations VALUES(537,30,'cp -1154',NULL);
INSERT INTO evaluations VALUES(537,31,'cp -1148',NULL);
INSERT INTO evaluations VALUES(537,32,'cp -1140',NULL);
INSERT INTO evaluations VALUES(537,33,'cp -1155',NULL);
INSERT INTO evaluations VALUES(537,34,'cp -1132',NULL);
INSERT INTO evaluations VALUES(537,35,'cp -1153',NULL);
INSERT INTO evaluations VALUES(537,36,'cp -1155','h4h1 c4f1 e8g8 d1e2 c6d4 e2g2 h1h5 f1e2 h5f7 d2d3 d4c2 e1d2 c2a1 b1c3 d7d5 b2b3 d5d4 c1b2 d4c3 b2c3 a1b3 a2b3 f7b3 g2e4 b3a2 d2c1 f8e8 e2h5 e8e7 h5d1 a2a3 c1d2 a3c5 d1c2 g7g6 c2b3 g8g7 g3g4 c5f2 d2c1 f2f6 c1b2 e7e8 b3d1 a7a5 d1a4 c7c6 d3d4 e5d4 e4e8');
INSERT INTO evaluations VALUES(540,30,'cp -1162',NULL);
INSERT INTO evaluations VALUES(540,31,'cp -1151',NULL);
INSERT INTO evaluations VALUES(540,32,'cp -1162',NULL);
INSERT INTO evaluations VALUES(540,33,'cp -1163',NULL);
INSERT INTO evaluations VALUES(540,34,'cp -1162',NULL);
INSERT INTO evaluations VALUES(540,35,'cp -1147',NULL);
INSERT INTO evaluations VALUES(540,36,'cp -1147','c4f1 e8g8 d1e2 c6d4 e2g2 h1h5 f1e2 h5f7 d2d3 d4c2 e1d2 c2a1 b1c3 d7d5 b2b3 d5d4 c1b2 d4c3 b2c3 a1b3 a2b3 f7b3 g2e4 b3a2 d2c1 f8e8 e2h5 e8e7 h5d1 a2a3 c1d2 a3c5 d1c2 g7g6 c2b3 g8g7 g3g4 c5f2 d2c1 f2f6 c1b2 e7e8 b3d1 a7a5 e4e1 a5a4 c3d2 f6c6 e1e3 a4a3 b2a1 c6h1 e3e2 c7c6 d2e3');
INSERT INTO evaluations VALUES(16379,30,'cp -590',NULL);
INSERT INTO evaluations VALUES(16379,31,'cp -617',NULL);
INSERT INTO evaluations VALUES(16379,32,'cp -596',NULL);
INSERT INTO evaluations VALUES(16379,33,'cp -615',NULL);
INSERT INTO evaluations VALUES(16379,34,'cp -601',NULL);
INSERT INTO evaluations VALUES(16379,35,'cp -632',NULL);
INSERT INTO evaluations VALUES(16379,36,'cp -612','c6d4 f3d1 d7d5 c4e2 d5e4 b1c3 b7b6 h2h4 c8b7 h1h3 f6d5 d2d3 d5b4 e1f1 e4d3 c2d3 d8d7 a2a3 b4c6 c1e3 d4e2 d1e2 e8c8 h4h5 c8b8 a1d1 h7h6 h3g3 f7f5 g4f5 c6e7 g1f3 e7f5 f3e5 f5g3 f2g3 d7f5 e3f4 f8d6 c3b5 d6e5 e2e5 f5h3 f1e2');
INSERT INTO evaluations VALUES(16986,30,'cp -622',NULL);
INSERT INTO evaluations VALUES(16986,31,'cp -623',NULL);
INSERT INTO evaluations VALUES(16986,32,'cp -605',NULL);
INSERT INTO evaluations VALUES(16986,33,'cp -606',NULL);
INSERT INTO evaluations VALUES(16986,34,'cp -621',NULL);
INSERT INTO evaluations VALUES(16986,35,'cp -612',NULL);
INSERT INTO evaluations VALUES(16986,36,'cp -611','f3d1 d7d5 c4e2 d5e4 b1c3 b7b6 d2d3 c8b7 g4g5 e4d3 e2d3 b7h1 g5f6 d8f6 c1e3 f8b4 d1h5 g7g6 h5g4 h7h5 g4h3 b4c3 b2c3 d4f3 e1e2 f3h4 h3g3 h4f5 g3g5 f6e6 d3f5 g6f5 f2f3 f5f4 e3d2 e8f8 e2f2 h8g8 g5h5 a8d8 a1d1 g8g6 h5h8 f8e7');
INSERT INTO evaluations VALUES(19224,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19224,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19224,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19224,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19224,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19224,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19224,36,'cp 0','f1b5 f8g7 b5c6 b7c6 e1g1 f7f6 e5e6 d7e6 d2d3 g8h6 f3d2 e6e5 b1c3 h6f5 b2b3 e8g8 c3a4 c5c4 d2c4 c8e6 a4c5 e6c4 b3c4 d8d6 c5e4 d6c7 e4c5');
INSERT INTO evaluations VALUES(19209,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19209,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(19209,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(19209,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19209,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(19209,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(19209,36,'cp -19','f8g7 b5c6 b7c6 e1g1 f7f6 f1e1 f6e5 f3e5 d7d6 e5f3 a8b8 h2h3 d8c7 d2d3 g8f6 b1c3 e8g8 c3e4 f6d5 c2c4 d5f4 a1b1 e7e5 b2b3 f4e6 c1e3 h7h6 d1d2 f8f3 g2f3 c7d7 b1d1 d6d5 e4c5 e6c5 e3c5 d7h3 c4d5 c6d5 d3d4');
INSERT INTO evaluations VALUES(22667,30,'cp -351',NULL);
INSERT INTO evaluations VALUES(22667,31,'cp -362',NULL);
INSERT INTO evaluations VALUES(22667,32,'cp -364',NULL);
INSERT INTO evaluations VALUES(22667,33,'cp -352',NULL);
INSERT INTO evaluations VALUES(22667,34,'cp -371',NULL);
INSERT INTO evaluations VALUES(22667,35,'cp -353',NULL);
INSERT INTO evaluations VALUES(22667,36,'cp -370','c6e5 b1c3 e5g6 g2g3 g8f6 d2d4 d7d5 e4e5 f6e4 c3e4 d5e4 c2c3 c7c5 c1e3 c5d4 e3d4 d8d5 d1b3 d5b3 a2b3 g6e7 f1g2 c8f5 a1a4 a7a6 e1g1 e8c8 d4e3 e7d5 e3a7 f8e7 g2e4 f5e6 f2f4 f7f5 e5f6 d5f6 e4g2 e6b3 a4a5 d8d2 f1f2 d2d1 f2f1 d1f1 g1f1 h8e8 a7d4 e7d6 a5g5 g7g6');
INSERT INTO evaluations VALUES(23083,30,'cp -336',NULL);
INSERT INTO evaluations VALUES(23083,31,'cp -362',NULL);
INSERT INTO evaluations VALUES(23083,32,'cp -350',NULL);
INSERT INTO evaluations VALUES(23083,33,'cp -365',NULL);
INSERT INTO evaluations VALUES(23083,34,'cp -347',NULL);
INSERT INTO evaluations VALUES(23083,35,'cp -349',NULL);
INSERT INTO evaluations VALUES(23083,36,'cp -343','b1c3 e5g6 g2g3 g8f6 d2d4 d7d5 e4e5 f6e4 c3e4 d5e4 c2c3 c7c5 c1e3 c5d4 e3d4 d8d5 d1b3 d5b3 a2b3 g6e7 f1g2 c8f5 a1a4 a7a6 e1g1 e8c8 d4e3 e7d5 g2e4 f5e6 a4c4 c8b8 e3g5 f8e7 e4d5 e7g5 d5e6 f7e6 h2h4 g5h6 f2f4 d8d3 g1g2 d3d2 f1f2 h8d8 h4h5');
INSERT INTO evaluations VALUES(26303,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(26303,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(26303,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(26303,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(26303,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(26303,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(26303,36,'cp 9','c2c4 g8f6 e1g1 e7e6 c4c5 f8e7 a2a3 a7a5 c1e3 e8g8 b1c3 f6e4 a1b1 h7h6 h2h3 g4h5 f3d2 h5e2 d1e2 e4d2 e2d2 b7b6 b2b4 a5b4 a3b4 b6c5 b4c5 d8d7 f1c1 f8b8 e3f4 b8b1 c1b1 e7f6 f4e3 a8a3 c3b5 a3a8 b5d6 a8b8 b1b8 c6b8 g2g3 b8c6 d2d3 f6e7');
INSERT INTO evaluations VALUES(26655,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(26655,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(26655,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26655,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(26655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26655,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26655,36,'cp 0','e7e6 b2b4 g8e7 c2c3 f7f6 f2f4 e7g6 f1d3 f5d3 d1d3 a7a5 b4b5 c6e7 g1e2 f6e5 f4e5 e7f5 e1g1 f8e7 e2g3 g6h4 g3e2 h4g6');
INSERT INTO evaluations VALUES(26854,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(26854,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(26854,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(26854,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(26854,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(26854,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26854,36,'cp 27','g1f3 a7a6 f1e2 f7f6 e1g1 f6e5 f3e5 g8e7 f2f4 c6e5 f4e5 c7c5 c2c3 d8b6 d4c5 b6c5 g1h1 c5c6 b1d2 e7g6 d2b3 f8c5 b3c5 c6c5 d1a4 b7b5 a4d4 c5d4 c3d4 e8g8 h1g1 h7h6 c1e3 g6e7 g2g4 f5e4 f1f8 a8f8');
INSERT INTO evaluations VALUES(26853,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(26853,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(26853,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(26853,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(26853,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(26853,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(26853,36,'cp 3','d8d7 b2b4 g8e7 c1e3 a7a5 b4b5 c6a7 b1c3 h7h6 f3h4 f5h7 f1d3 h7d3 d1d3 a7c8 e1g1 c8b6 a3a4 b6c4 c3e2 c4b2 d3c3 b2c4 f1b1 e7c8 h4f3 c8b6 e2g3 f8e7 g3h5 e8g8 c3d3 c4e3 d3e3 b6c4 e3f4 f7f5 f4g3 g7g5 c2c3 c7c6 g3h3 c6b5 b1b5 g8h8');
INSERT INTO evaluations VALUES(48389,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(48389,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(48389,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(48389,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(48389,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48389,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48389,36,'cp 0','e4e5 f6d7 f2f4 f8e7 a2a4 c7c5 g1f3 b8c6 g2g3 e8g8 f1g2 a7a6 e1g1 a8b8 c3e2 d8c7 d3d4 b7b5 a4b5 b8b5 g1h1 c5d4 f3d4 c6d4 e2d4 b5b8 c2c3 a6a5 d1a4 d7b6 a4d1');
INSERT INTO evaluations VALUES(49432,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(49432,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(49432,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(49432,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(49432,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(49432,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(49432,36,'cp 10','c7c5 g1f3 c5d4 f3d4 d7d5 e1g1 e6e5 d4b3 c8e6 e2e4 b8c6 e4d5 e6d5 g2d5 d8d5 b1c3 d5e6 c1g5 a8d8 d1f3 f8e7 g5f6 e6f6 f3f6 e7f6 f1d1 f6e7 d1d8 e8d8 a1d1 d8c8 c3d5 h8d8 g1f1 d8d7 c2c3 h7h6 a2a4 f7f5 f1e2 b7b6');
INSERT INTO evaluations VALUES(53663,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(53663,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(53663,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(53663,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(53663,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(53663,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(53663,36,'cp 154','f1d3 f8e7 c1f4 g8f6 e1g1 b8d7 f1e1 e8g8 d1e2 a7a6 a2a4 e7b4 h2h3 d7b6 f4d2 a6a5 c3e4 b4d2 e2d2 b6d5 f3e5 d5b4 e4f6 d8f6 d3c4 f6e7 e1e3 e7c7 c2c3 b4d5 e3g3 f7f6 e5d3 g8h8 a1e1 b7b6 h3h4 c8d7 h4h5 g7g5 h5h6 d5f4');
INSERT INTO evaluations VALUES(53664,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(53664,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(53664,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(53664,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(53664,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(53664,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(53664,36,'cp 150','f8e7 d1e2 g8f6 c1f4 b8a6 a2a3 a6c7 e1g1 c7d5 c3d5 c6d5 c2c3 e8g8 f3e5 f6d7 e2c2 h7h6 a1e1 d7e5 d4e5 e7c5 c2e2 d8h4 f4g3 h4e7 d3b1 e7g5 g1h1 g7g6 f2f4 g5e7 b2b4 c5b6 e2g4 g8g7');
INSERT INTO evaluations VALUES(53868,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(53868,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(53868,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(53868,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(53868,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(53868,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(53868,36,'cp 35','a7a6 b5c3 g8f6 c1g5 f8e7 g5f6 e7f6 c3d5 e8g8 f1c4 b7b5 d5f6 d8f6 c4d5 a8a7 e1g1 a7c7 b1c3 b8d7 d1e2 d7c5 a2a3 f8d8 f1d1 f6e7 d5a2 c8b7 c3d5 b7d5 a2d5 c5e6 d5e6 e7e6 d1d2 e6c4 d2d3');
INSERT INTO evaluations VALUES(50712,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(50712,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(50712,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(50712,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(50712,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(50712,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(50712,36,'cp 43','b5c3 g8f6 c1g5 f8e7 g5f6 e7f6 c3d5 e8g8 a2a4 b8c6 f1c4 f6g5 e1g1 a8b8 d1d3 g8h8 b2b4 c8e6 b4b5 a6b5 a4b5 c6e7 b1c3 e7d5 c4d5 d8c7 g2g3 b8c8 a1a3 g7g6 f1d1 g5e7 d3e2 h8g8 d5e6 f7e6 g1g2 g8g7');
INSERT INTO evaluations VALUES(50724,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50724,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50724,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50724,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(50724,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(50724,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(50724,36,'cp 33','g8f6 c1g5 f8e7 g5f6 e7f6 c3d5 e8g8 a2a4 b8c6 b1a3 a8b8 f1e2 c8e6 e2c4 f6g5 e1g1 g8h8 c2c3 g7g6 d1e2 e6d7 b2b4 c6e7 a4a5 f7f5 d5e7 d8e7 c4d5 f5e4 d5e4 d7e6 c3c4 f8f6 a1d1 b8f8 a3c2 e6f5 e4f5 g6f5');
INSERT INTO evaluations VALUES(19024,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(19024,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(19024,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(19024,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(19024,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(19024,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(19024,36,'cp 82','f4e5 g8e7 g1f3 e7g6 f1c4 f8e7 e1g1 e8g8 d2d3 c6e5 f3e5 g6e5 c1f4 d7d6 f4e5 d6e5 d1e2 c8e6 a2a4 e7g5 g2g3 d8d7 c4e6 f7e6 g1g2 g5e7 b2b3 h7h6 f1f8 a8f8 a1f1 f8f1 g2f1 a7a6 c3d1 e7f6 h2h4 b7b5 a4b5 a6b5 d1f2');
INSERT INTO evaluations VALUES(18960,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(18960,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(18960,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(18960,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(18960,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(18960,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(18960,36,'cp 86','g8e7 g1f3 e7g6 f1c4 f8e7 d2d3 e8g8 e1g1 g6e5 f3e5 c6e5 c1f4 d7d6 f4e5 d6e5 a2a4 c8e6 d1e2 e7g5 f1f2 d8d4 c3b5 d4d7 a1f1 a7a6 b5c3 d7d6 g1h1 a8d8 a4a5 d8d7 e2h5 g5d8 f2f3 d8a5 c4e6 d6e6 c3d5 a5d8 g2g3');
INSERT INTO evaluations VALUES(31310,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(31310,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(31310,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31310,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(31310,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31310,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31310,36,'cp 18','e7e6 d1b3 d8b6 b3b6 a7b6 c3b5 f8b4 c1d2 b4d2 f3d2 e8e7 a2a3 g8f6 a1c1 b8c6 f2f3 g7g5 g2g4 f5g6 h2h4 g5h4 h1h4 f6e8 e1f2 e8d6 e2e3 h7h5 g4g5 d6b5 f1b5 c6a7 b5e2 a8c8 h4h1 c8c1 h1c1 h8c8 c1c8 a7c8 e3e4 c8d6 f2e3 b6b5 e3f4 d6c4 d2f1 c4b2 e2b5 b2c4 a3a4 h5h4 e4d5 e6d5');
INSERT INTO evaluations VALUES(31694,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(31694,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31694,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(31694,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31694,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(31694,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(31694,36,'cp 22','d1b3 d8b6 b3b6 a7b6 c3b5 f8b4 c1d2 b4d2 f3d2 e8e7 a2a3 g8f6 a1c1 b8c6 e2e3 g7g5 h2h3 h7h5 h3h4 g5g4 f1e2 h8c8 e1g1 c6a5 b5c7 a8b8 f1e1 f6e4 d2e4 f5e4 f2f3 e4f5 g1f2 a5c4 e2c4 c8c7 c4b3 c7c8 b3a4 f5d3 e1d1 d3c4 c1c3 g4f3');
INSERT INTO evaluations VALUES(31778,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(31778,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(31778,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(31778,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(31778,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(31778,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(31778,36,'cp 96','c7c6 d1b3 d8b6 c4c5 b6b3 a2b3 g7g6 c1f4 e6f5 g1f3 b8d7 e2e3 f5g4 f3e5 d7e5 f4e5 f7f6 e5g3 e7e5 b3b4 f8g7 b4b5 g8e7 f1d3 g4d7 e1g1 h7h5 h2h4 e8g8 b2b4 e7f5 b5c6 b7c6 d3f5 g6f5 f2f4 f8e8 a1a3 e5d4 e3d4 e8b8 f1b1 g8f7 b4b5 c6b5');
INSERT INTO evaluations VALUES(31015,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(31015,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(31015,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(31015,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(31015,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(31015,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(31015,36,'cp 108','d1b3 d8b6 c4c5 b6b3 a2b3 g7g6 c1f4 e6f5 g1f3 b8d7 e2e3 f5g4 f3e5 d7e5 f4e5 f7f6 e5g3 e7e5 b3b4 f8g7 b4b5 g8e7 f1d3 g4d7 e1g1 h7h5 h2h4 e8g8 b2b4 e7f5 d3f5 g6f5 b5c6 b7c6 f2f4 f8e8 a1a3 e5d4 e3d4 e8e3 g3f2 e3d3 f1a1 a7a6 c3a4 d3a3 a1a3 d7e6 a3a1 a8a7 a4b6');
INSERT INTO evaluations VALUES(45681,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(45681,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(45681,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(45681,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(45681,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(45681,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(45681,36,'cp 137','e4e5 f6d7 f3g5 f8e7 h2h4 h7h6 g5e6 f7e6 d1h5 e8f8 h1h3 d8e8 h3f3 d7f6 h5g4 b8d7 c1d2 f8g8 g4e6 g8f8 e5f6 d7f6 e6e5 f8f7 e1c1 c8g4 f1d3 e8d7 d3f5 g4f5 e5f5 d7f5 f3f5 f7g6 f5e5 a8e8 h4h5 g6f7 d1h1 e7d6 e5f5 e8e6 g2g4 b7b5 f2f3 a7a5 d2f4 d6f4 f5f4 h8e8 a2a4 b5b4');
INSERT INTO evaluations VALUES(45689,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(45689,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(45689,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(45689,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(45689,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(45689,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(45689,36,'cp 115','f6d7 f3g5 f8e7 h2h4 h7h6 g5e6 f7e6 d1h5 e8f8 h1h3 d8e8 h3f3 d7f6 h5g4 b8d7 c1d2 f8g8 g4e6 e8f7 e6f7 g8f7 e1c1 d7f8 c3e2 c8g4 e5f6 e7f6 f3b3 b7b6 b3c3 g4d7 h4h5 f8e6 d2e3 h8e8 c3d3 f7g8 f2f3 d7c8 d3a3 b6b5 a3c3 c8d7 e2f4 a7a5 f4g6 b5b4 c3d3 e6f8 g6f8 e8f8');
INSERT INTO evaluations VALUES(43977,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(43977,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(43977,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(43977,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(43977,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(43977,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(43977,36,'cp 112','f3g5 f8e7 h2h4 h7h6 g5e6 f7e6 d1h5 e8f8 h1h3 d8e8 h3f3 d7f6 h5g4 b8d7 c1d2 f8g8 g4e6 e8f7 e6f7 g8f7 e1c1 d7f8 c3e2 c8g4 e5f6 e7f6 f3b3 b7b6 b3c3 g4d7 h4h5 f8e6 d2e3 h8e8 f2f3 f7g8 c3d3 a7a5 e2f4 b6b5 c2c3 b5b4 f4g6 b4c3 d3c3 c6c5 d4c5 f6c3 b2c3 e6f8 e3d4 f8g6 h5g6');
INSERT INTO evaluations VALUES(638,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(638,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(638,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(638,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(638,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(638,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(638,36,'cp 230','e5f3 d4b6 d2d3 c8e6 e1g1 h7h6 a1b1 c6c5 b2b3 e8c8 f3e5 g7g5 c1b2 f8e7 c3a4 b6d6 e5f7 e6f7 e4e5 d6c6 e5f6 e7f6 b2f6 c6f6 a4c5 f6c6 d3d4 f7g6 d1g4 c8b8 b1b2 h6h5 g4e6 c6d6 e6d6 c7d6 c5e6 d8e8 e6g5 e8e2 f2f4 g6c2');
INSERT INTO evaluations VALUES(642,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(642,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(642,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(642,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(642,34,'cp 223',NULL);
INSERT INTO evaluations VALUES(642,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(642,36,'cp 210','d4b6 d2d3 c8e6 e1g1 e8c8 a1b1 c6c5 f3g5 f8d6 d1e1 h8e8 g5e6 e8e6 b2b3 d8g8 c1g5 b6c6 f2f4 f6d7 f4f5 e6e8 c3d5 b7b6 a2a4 c8b7 a4a5 b6b5 c2c4 f7f6 g5e3 g7g6 g2g3 g6f5 f1f5 d7e5');
INSERT INTO evaluations VALUES(1861,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(1861,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(1861,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(1861,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(1861,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(1861,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(1861,36,'cp 110','d7d6 b1c3 g8e7 c3b5 e8d8 d2d4 c8g4 c4e2 g4f3 e2f3 a7a6 d4e5 c6e5 b5d4 c7c5 d4f5 d8c7 c1f4 a8d8 f4e5 d6e5 d1e1 c7b8 f5e3 e7g6 e3d5 f6g5 a1d1 h7h5 e1c3 f8d6 c3b3 g6f4 d5f4 g5f4 g2g3 f4h6 h2h4 h6g6 d1d5 d6c7 f1d1 d8d5 b3d5 h8d8 d5c5 d8d1 f3d1 g6e4');
INSERT INTO evaluations VALUES(1746,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(1746,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(1746,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(1746,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(1746,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(1746,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(1746,36,'cp 135','b1c3 g8e7 c3b5 e8d8 d2d4 c8g4 c4e2 g4f3 e2f3 a7a6 d4d5 a6b5 d5c6 e7c6 c2c3 b5b4 c1e3 b4c3 d1b3 c3b2 b3b2 d6d5 e4d5 c6a5 f3e2 f8d6 b2c2 d8e8 e3d2 e8f8 d2c3 f6g5 g2g3 b7b6 a2a4 f8g8 h2h4 g5g6 e2d3 g6h5 f1e1 a5b7 a4a5 b6a5 e1b1 b7c5 a1a5');
INSERT INTO evaluations VALUES(5309,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(5309,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(5309,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(5309,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(5309,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(5309,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(5309,36,'cp -46','e7e6 e1g1 g8e7 c4b3 e8g8 d2d3 d7d5 e4e5 e7f5 c3e2 b7b5 c2c3 d5d4 e2g3 c8b7 g3f5 g6f5 a2a4 d4c3 a4b5 c3b2 c1b2 c6b4 b3c4 d8b6 h2h3 a7a6 b5a6 b7a6 g1h2 f8d8 a1a3 g7h6 b2c1 a6b7 a3a8 d8a8 d1e2 a8a1');
INSERT INTO evaluations VALUES(5117,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(5117,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(5117,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(5117,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(5117,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(5117,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(5117,36,'cp -32','d2d3 g8e7 c4b3 e8g8 e1g1 d7d5 e4e5 e7f5 c3e2 b7b5 c2c3 d5d4 e2g3 c8b7 g3f5 g6f5 c3c4 b5c4 b3c4 g8h8 d1e2 g7h6 f3g5 c6a5 b2b3 d8d7 e2h5 h6g5 f4g5 f8g8 f1f2 a5c4 b3c4 g8g6 c1d2 b7c6 a1f1 a8b8 d2f4 b8g8 h5h4 a7a6');
INSERT INTO evaluations VALUES(5118,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(5118,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(5118,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(5118,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(5118,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(5118,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(5118,36,'cp -25','g8e7 c4b3 d7d5 e1g1 e8g8 e4e5 e7f5 c3e2 b7b5 c2c3 d5d4 e2g3 c8b7 g3f5 g6f5 c3c4 b5c4 b3c4 g8h8 f3g5 g7h6 d1e2 a7a5 c1d2 h6g5 f4g5 a5a4 h2h4 f8g8 f1f2 c6e7 a1f1 b7d5 c4d5 d8d5 h4h5 d5a2 g5g6 f7g6 d2g5 e7d5 g5f6 d5f6 e5f6 a2d5');
INSERT INTO evaluations VALUES(7968,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(7968,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(7968,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(7968,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(7968,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(7968,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(7968,36,'cp -34','e1g1 d7d5 c4b3 e8g8 e4e5 e7f5 c3e2 b7b5 c2c3 d5d4 e2g3 c8b7 g3f5 g6f5 c3c4 b5c4 b3c4 a7a5 a2a4 c6b4 c1d2 b4d5 a1c1 g8h8 f3g5 g7h6 d1h5 h6g5 h5g5 d8g5 f4g5 d5e3 d2e3 d4e3 c1e1 f8g8 e1e3 g8g5 e3g3 g5g3 h2g3 h8g7 b2b3 a8d8 f1c1 b7d5 c4d5 d8d5');
INSERT INTO evaluations VALUES(7967,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(7967,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(7967,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(7967,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(7967,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(7967,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(7967,36,'cp -36','d7d5 c4b3 e8g8 e4e5 e7f5 c3e2 b7b5 c2c3 d5d4 e2g3 c8b7 g3f5 g6f5 c3c4 c6a5 c1d2 a5b3 a2b3 a7a5 h2h3 b7c6 c4b5 c6d5 d1e1 d8d7 a1a5 a8a5 d2a5 d7b5 b3b4 f8a8 e1d2 g7f8 a5c7 a8a2 c7d6 b5b4 d2b4 c5b4 d6f8 g8f8');
INSERT INTO evaluations VALUES(17010,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(17010,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(17010,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(17010,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(17010,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(17010,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(17010,36,'cp 119','f3d4 e5d4 d1d4 d7d6 c1f4 f8e7 f2f3 c8e6 e1c1 e8g8 c1b1 f6d7 d4d2 e7f6 h2h4 f6e5 f4g5 f7f6 g5e3 c7c6 f3f4 e5c3 d2c3 d8e7 c3a3 e6f7 d1d6 a8d8 h4h5 d7b6 d6d8 e7a3 b2a3 f8d8 f1d3 h7h6 g2g4 b6a4 b1c1 f7a2 e3a7 a2f7');
INSERT INTO evaluations VALUES(17008,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(17008,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(17008,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(17008,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(17008,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(17008,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(17008,36,'cp 127','e5d4 d1d4 d7d6 c1f4 f8e7 e1c1 e8g8 c1b1 c8e6 f2f3 f6d7 h2h4 e7f6 d4d2 d7b6 b2b3 f6e5 g2g4 d8e7 f4g5 f7f6 g5e3 d6d5 c3d5 b6d5 e4d5 a8d8 f1c4 c7c6 d5d6 e7f7 c4d3 f8e8 e3a7 d8d6 a7b6 d6d7 h1e1');
INSERT INTO evaluations VALUES(17043,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(17043,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(17043,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(17043,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(17043,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(17043,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(17043,36,'cp 113','d1d4 d7d6 c1f4 f8e7 e1c1 e8g8 c1b1 c8e6 f2f3 f6d7 d4d2 e7f6 h2h4 f6e5 f4g5 f7f6 g5e3 c7c6 f3f4 e5c3 d2c3 d8e7 f1e2 f6f5 c3a3 f5e4 a3d6 f8e8 d6e7 e8e7 g2g4 d7b6 h1e1 b6d5 e3c5 e7e8 f4f5 e6f7 b1c1 a8d8 c5f2 a7a5 a2a4 d5f4 d1d8 f4e2 e1e2 e8d8 b2b3 g7g6 f2b6 d8d7 e2e4 g6f5');
INSERT INTO evaluations VALUES(26325,30,'cp 232',NULL);
INSERT INTO evaluations VALUES(26325,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(26325,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(26325,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(26325,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(26325,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(26325,36,'cp 220','c4f7 e8f7 f3g5 f7e8 d1g4 g8f6 g4e2 d8d7 g5f3 g7g6 d2d3 f8g7 e4e5 c6e5 f3e5 d6e5 e2e5 e8f7 e5c5 a8c8 c5e3 b7b5 e1g1 b5b4 c3e4 f6e4 d3e4 d7e6 e3f3 f7e8 a2a3 c8c2 a3b4 h8f8 f3d3 e6c4 d3c4 c2c4 b4b5 c4e4 a1a7 g7d4 a7a8 e8f7 a8a4');
INSERT INTO evaluations VALUES(32442,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(32442,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(32442,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(32442,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32442,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32442,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32442,36,'cp 0','c7c5 a2a3 b8c6 b2b4 b7b6 e1g1 g8f6 d2d3 f8e7 c1d2 e8g8 c3e2 a7a6 e2g3 d8d7 h2h3 f8b8 b4c5 b6c5 a1b1 h7h6 d2e3 a6a5 c2c3 b8b1 d1b1 a5a4 b1a2 c6a5 f1b1 e7d8 f3d2 d8c7 d3d4 e5d4 c3d4 c5d4 e3d4');
INSERT INTO evaluations VALUES(41993,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(41993,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(41993,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(41993,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(41993,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(41993,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(41993,36,'cp -48','d5c4 g1f3 b7b5 a2a4 c7c6 g2g3 c8b7 f1g2 b8d7 e1g1 d8b6 f3e5 a7a6 g5e3 b6c7 a4b5 c6b5 c3b5 a6b5 a1a8 b7a8 g2a8 e8g8 e5d7 c7d7 a8f3 f6d5 f1a1 f8c8 g1g2 c4c3 a1a2 b4f8 b2c3 c8c3 c2e4 d5e3 f2e3 b5b4 a2c2 d7c7 c2c3 c7c3');
INSERT INTO evaluations VALUES(46192,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46192,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(46192,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(46192,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(46192,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(46192,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(46192,36,'cp -11','e2e3 a7a6 d4c5 f8c5 f1d3 b8d7 e3e4 c5b4 e4e5 d5d4 f3d4 f6d5 d1d2 d8c7 f4g3 d5c3 a2a3 b4a5 d4b3 d7e5 b3a5 e5d3 c2d3 c7a5 d2c3 a5c3 b2c3 f7f6 e1e2 c8d7 a3a4 e8f7 c3c4 b7b5 a4b5 a6b5 h1b1 b5c4 d3c4 h8c8 a1a8 c8a8 b1b7 f7e7 e2d3 e6e5 d3c3 a8a4 b7b3 d7c6 f2f3');
INSERT INTO evaluations VALUES(54009,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(54009,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(54009,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(54009,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(54009,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(54009,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(54009,36,'cp 78','e6d5 c4d5 g7g6 e2e4 f8g7 g1f3 a7a6 c1f4 g8f6 d1a4 f6d7 f1e2 e8g8 a4b3 d8e7 a2a4 d7e5 f4e5 g7e5 f3e5 e7e5 e1g1 g6g5 f2f3 b8d7 g1h1 e5g7 g2g3 g7h6 b3c2 a8b8 f3f4 g5f4 g3f4 g8h8 a4a5');
INSERT INTO evaluations VALUES(53938,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(53938,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(53938,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(53938,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(53938,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(53938,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(53938,36,'cp 85','c4d5 g7g6 g1f3 g8f6 h2h3 f8g7 e2e4 e8g8 c1e3 f8e8 f3d2 f6e4 d2e4 f7f5 f1b5 e8e4 c3e4 f5e4 e1g1 b8d7 b2b4 g7a1 d1a1 c5b4 b5d7 c8d7 a1d4 a7a5 e3h6 d8e7 d4b6 a8c8 b6b7 g6g5 h6g5 e7g5 b7d7 c8c1 d7d6 c1f1 g1f1 g5c1 f1e2 c1c2 e2e3 c2a2');
INSERT INTO evaluations VALUES(53918,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(53918,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(53918,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(53918,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(53918,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(53918,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(53918,36,'cp 88','g7g6 g1f3 g8f6 f3d2 f8g7 e2e4 e8g8 f1e2 b8a6 e1g1 a6c7 a2a4 b7b6 f2f3 d8e7 d2c4 c8a6 c1g5 a6c4 e2c4 a7a6 d1d2 e7d7 d2e2 h7h6 g5e3 f8b8 g1h1 d7e7 c4d3 b6b5 a4b5 c7b5 d3b5 a6b5 c3b5 f6d7 h2h3 b8e8');
INSERT INTO evaluations VALUES(55700,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55700,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(55700,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(55700,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(55700,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(55700,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55700,36,'cp 20','g8f6 b1c3 e6d5 e4d5 f8e7 g1e2 b8d7 h2h3 c7c6 e2f4 e8g8 a2a4 a7a5 g2g4 d7b6 f1g2 b6d5 c3d5 f6d5 f4d5 c6d5 g2d5 c8e6 e1g1 e7f6 c1e3 f6b2 a1b1 b2c3 b1b7 d8f6 d1f3 f6f3 d5f3 c3b4 c2c4 e6c4 f1d1 f8e8');
INSERT INTO evaluations VALUES(1458,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(1458,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(1458,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(1458,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(1458,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(1458,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(1458,36,'cp -54','g1f3 g8h6 f1d3 c5d4 c3d4 c8d7 d3c2 c6b4 b1a3 b6a6 c2b1 a8c8 e1f2 f8e7 h1e1 h6f5 d1e2 a6e2 f2e2 f5h4 f3h4 e7h4 e1f1 h4e7 c1d2 a7a5 f4f5 e6f5 b1f5 d7f5 f1f5 b4c2 a3c2 c8c2 a1c1 c2b2 e2d3 e8g8 a2a4 e7b4 f5f2 b4d2 f2d2 b2b4');
INSERT INTO evaluations VALUES(13249,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13249,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(13249,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(13249,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(13249,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(13249,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13249,36,'cp -1','e7e6 b1d2 a7a6 a2a4 a8b8 e1g1 f8e7 c2c3 b7b5 c4b3 e8g8 f1e1 b5b4 d2f1 d8b6 b3a2 a6a5 c1f4 c8a6 f1e3 f6d7 f4g3 d7e5 f3e5 c6e5 g3e5 d6e5 a2c4 e7g5 c4a6 g5e3 a6c4 e3g5 g2g3 g7g6 h3h4 g5e7 d1g4 b4c3 b2c3');
INSERT INTO evaluations VALUES(13936,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13936,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(13936,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(13936,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(13936,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(13936,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(13936,36,'cp 0','b1d2 a7a6 c2c3 b7b5 c4b3 c8b7 e1g1 d8c7 f1e1 f8e7 d2f1 e8g8 b3c2 a6a5 a2a4 b5b4 c1f4 c6e5 f1e3 b4c3 b2c3 c5c4 e3c4 e5c4 d3c4 e6e5 f4g5 h7h6 g5f6 e7f6 a1b1 b7a6 f3d2 a8b8 c2b3 f6g5 d2f3 g5f6');
INSERT INTO evaluations VALUES(17232,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(17232,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(17232,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(17232,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(17232,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(17232,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(17232,36,'cp -152','f6e4 d2d4 d7d5 c4b5 c8g4 h2h3 g4f3 d1f3 e7e6 e1g1 f8e7 b1d2 e4d6 b5c6 b7c6 d2b3 d6e4 b3d2 e4d2 c1d2 e8g8 c2c4 a7a5 a1c1 e7d6 f1e1 d8h4 f3g4 h4g4 h3g4 h7h5 g4g5 a5a4 c4c5 d6e7');
INSERT INTO evaluations VALUES(17308,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(17308,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(17308,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(17308,33,'cp -138',NULL);
INSERT INTO evaluations VALUES(17308,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(17308,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(17308,36,'cp -113','d2d4 d7d5 c4b5 e7e6 e1g1 f8d6 f1e1 e8g8 c2c4 c8d7 b1c3 a7a6 b5c6 d7c6 c4c5 d6e7 f3e5 c6e8 c1f4 f7f6 e5d3 e8g6 c3e2 d8d7 h2h4 e6e5 f4h2 e5d4 a1c1 f6f5 e2d4 e7h4 d4f3 g6h5 f3e5 h5d1 e5d7 f8d8 c5c6 b7c6 e1d1 d8d7 c1c6');
INSERT INTO evaluations VALUES(19206,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19206,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(19206,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(19206,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(19206,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(19206,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(19206,36,'cp 0','d8b6 d4c6 b7c6 b5c4 f8g7 b1c3 d7d6 e1g1 g8f6 h2h3 e8g8 f1e1 f6d7 c3a4 b6a5 c4b3 c8a6 c1d2 a5c7 d2c3 g7h6 c3d2 h6g7');
INSERT INTO evaluations VALUES(27175,30,'cp -184',NULL);
INSERT INTO evaluations VALUES(27175,31,'cp -187',NULL);
INSERT INTO evaluations VALUES(27175,32,'cp -181',NULL);
INSERT INTO evaluations VALUES(27175,33,'cp -182',NULL);
INSERT INTO evaluations VALUES(27175,34,'cp -181',NULL);
INSERT INTO evaluations VALUES(27175,35,'cp -182',NULL);
INSERT INTO evaluations VALUES(27175,36,'cp -189','c6e5 f3e5 d7b5 b1c3 b5a6 d1h5 g7g6 h5f3 d8f6 c1f4 f8h6 g2g3 h6f4 g3f4 g8e7 e1c1 a8d8 h1e1 f6h4 e1g1 b7b6 g1g3 e7f5 g3h3 h4f6 c1b1 e8g8 a2a4 g8g7 c3b5 a6b5 a4b5 f5d6 e5g4 f6e7 g4h6 d6b5 f3g3 b5d6');
INSERT INTO evaluations VALUES(31416,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(31416,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(31416,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(31416,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(31416,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(31416,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(31416,36,'cp 46','h2h4 c6c5 d3b5 b8d7 h4h5 g6f5 f3h4 c5d4 b5d7 d8d7 h4f5 e6f5 d1d4 g8e7 b1c3 e7c6 d4d5 d7d5 c3d5 e8c8 c2c4 c6e5 b2b3 f8c5 c1g5 f7f6 g5e3 c5e3 d5e3 f5f4 e3d5 b7b5 h5h6 g7g5 d5f6 b5c4 b3c4 e5d7 f6d7 d8d7 e1g1 h8e8 a1e1 e8e1');
INSERT INTO evaluations VALUES(32503,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32503,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(32503,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(32503,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(32503,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(32503,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(32503,36,'cp 21','c2c3 c7c5 g1f3 b8c6 f1e2 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 a7a6 c1e3 f8e7 e2d3 f5e3 f2e3 e8g8 a1c1 f7f6 e5f6 g7f6 e3e4 d5e4 c3e4 f6f5 e4c5 g8h8 c5b7 d8c7 b7c5 c6d4 c5d7 c7d7 g1h1 e7f6 d3c4 f8d8 f3d4 d7d4 d1d4 f6d4 c4e6 d4b2');
INSERT INTO evaluations VALUES(34095,30,'cp 1005',NULL);
INSERT INTO evaluations VALUES(34095,31,'cp 1002',NULL);
INSERT INTO evaluations VALUES(34095,32,'cp 1005',NULL);
INSERT INTO evaluations VALUES(34095,33,'cp 1001',NULL);
INSERT INTO evaluations VALUES(34095,34,'cp 1009',NULL);
INSERT INTO evaluations VALUES(34095,35,'cp 1001',NULL);
INSERT INTO evaluations VALUES(34095,36,'cp 1001','f1e2 e4g2 e2f3 g2g5 h8g8 g5e5 g1e2 e5g7 g8g7 f8g7 b1c3 c7c6 c3e4 e8e7 d2d4 d7d5 e4c5 b8d7 c5d7 c8d7 c1g5 e7d6 g5f4 d6e7 a2a4 a7a5 e1c1 h7h5 b2b3 e7d8 f4g3 a8a7 h2h3 d8c8 c1b2 b7b6 f3g2 d7f5 e2f4 a7e7 c2c3 g7f6 h3h4');
INSERT INTO evaluations VALUES(34096,30,'cp 986',NULL);
INSERT INTO evaluations VALUES(34096,31,'cp 986',NULL);
INSERT INTO evaluations VALUES(34096,32,'cp 978',NULL);
INSERT INTO evaluations VALUES(34096,33,'cp 978',NULL);
INSERT INTO evaluations VALUES(34096,34,'cp 986',NULL);
INSERT INTO evaluations VALUES(34096,35,'cp 986',NULL);
INSERT INTO evaluations VALUES(34096,36,'cp 980','e4g2 e2f3 g2g5 h8g8 g5e5 g1e2 e5g7 g8g7 f8g7 b1c3 c7c6 c3e4 e8e7 d2d4 d7d5 e4c5 b8d7 c5d7 c8d7 c1g5 e7d6 a2a4 a7a5 g5f4 d6e7 e1c1 h7h5 f4g3 g7h6 c1b1 d7f5 b2b3 a8a7 b1b2 e7d8 g3f4 h6f8 f4e3 b7b6 h2h3 d8c8 e2f4');
INSERT INTO evaluations VALUES(34098,30,'cp 1010',NULL);
INSERT INTO evaluations VALUES(34098,31,'cp 1009',NULL);
INSERT INTO evaluations VALUES(34098,32,'cp 1009',NULL);
INSERT INTO evaluations VALUES(34098,33,'cp 1009',NULL);
INSERT INTO evaluations VALUES(34098,34,'cp 1004',NULL);
INSERT INTO evaluations VALUES(34098,35,'cp 1001',NULL);
INSERT INTO evaluations VALUES(34098,36,'cp 1009','e2f3 g2g5 h8g8 g5e5 g1e2 e5g7 g8g7 f8g7 d2d4 b8c6 c2c3 c6e7 b1a3 d7d6 e1f1 a7a6 e2f4 a8a7 a3c2 c7c6 d4d5 c6c5 a2a4 c8d7 a4a5 e8f8 c2e3 h7h5 f1g2 h5h4 f3d1 g6g5 f4h5 g7e5 e3g4 d7g4 d1g4 e7d5 c1g5');
INSERT INTO evaluations VALUES(49297,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(49297,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(49297,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49297,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49297,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49297,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49297,36,'cp 0','c4b3 b8c6 e1g1 d6f5 f1e1 f8e7 f3e5 c6e5 e1e5 d7d6 e5e1 e8g8 c2c3 d6d5 c1f4 c7c6 b1d2 f5h4 d3d4 f8e8 f4g3 h4f5 g3f4');
INSERT INTO evaluations VALUES(49299,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49299,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49299,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49299,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49299,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49299,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49299,36,'cp 0','a7a5 a2a4 b8c6 e1g1 d6f5 f1e1 f8e7 f3e5 e8g8 c2c3 c6e5 e1e5 d7d6 e5e1 c7c6 c1f4 e7g5 f4g5 d8g5 d1d2 g5f6 d2f4 d6d5 b1d2 f6d6 e1e5 f5e7 f4g3 e7f5');
INSERT INTO evaluations VALUES(49365,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(49365,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(49365,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(49365,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(49365,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(49365,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(49365,36,'cp -21','c7c5 b4c5 f8c5 e2e3 e8g8 c2c4 b8c6 g1f3 d7d5 d2d4 c5e7 b1d2 c6a5 b2c3 a5c4 d2c4 d5c4 f1c4 d8c7 d1b3 b7b5 c4e2 a7a6 e1g1 c8b7 c3b4 b7d5 b3b2 e7b4 a3b4 c7b7 f1c1 f8c8 f3e1 c8c1 a1c1 f6d7 e2f3 d5f3 e1f3 a8c8 h2h3 c8c1 b2c1 d7b6 f3d2 b7e7');
INSERT INTO evaluations VALUES(56212,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(56212,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(56212,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(56212,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56212,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(56212,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(56212,36,'cp 0','f8d6 c2c4 c7c6 c4d5 c6d5 b1c3 g8f6 f1d3 b8c6 g1f3 e8g8 e1g1 h7h6 f1e1 c8e6 a2a3 f8e8 c1d2 a8c8 a1c1 e6d7 e1e8 d7e8 d2e3 a7a6 f3h4 c6e7 h4f5 e7f5 d3f5 c8a8 d1f3 e8c6 e3f4 d6f4 f3f4 d8b6 f4d2');
INSERT INTO evaluations VALUES(38629,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38629,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38629,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38629,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38629,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38629,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(38629,36,'cp 17','c2c4 c7c6 c4d5 c6d5 b1c3 g8e7 f1d3 b8c6 g1e2 e8g8 e1g1 c8e6 c1f4 d6f4 e2f4 d8b6 f4e2 b6b2 a1b1 b2a3 b1b3 a3a5 b3b5 a5a3 d1c2 h7h6 b5b3 a3d6 b3b7 f8b8 f1b1 b8b7 b1b7 c6b4 d3h7 g8h8 c2b1 a8b8 b7b8 d6b8 h7d3 b8b6 c3a4 b6a5 a4c5 b4d3');
INSERT INTO evaluations VALUES(19234,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19234,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19234,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19234,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19234,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19234,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19234,36,'cp 7','b1c3 d7d6 c1g5 a7a6 a2a4 f8g7 f2f4 a8b8 g1e2 h7h6 g5h4 g8f6 f4f5 g6g5 h4g3 e7e6 f5e6 f7e6 e4e5 f6g4 e5d6 e8g8 d1d2 c6d4 e2d4 c5d4 c3d1 g8h7 h2h3 g4e5 d1f2 d8d6 f2e4 d6c7 c2c3 e5c4 g3c7 c4d2 e1d2');
INSERT INTO evaluations VALUES(19229,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19229,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19229,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19229,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19229,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19229,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19229,36,'cp 7','d7d6 c1g5 a7a6 a2a4 h7h6 g5h4 g6g5 h4g3 f8g7 h2h4 g5g4 g1e2 e7e6 d1d2 g8f6 d3d4 f6h5 d4d5 c6a5 c4a2 e6e5 c3d1 b7b5 d1e3 b5b4 f2f3 g4f3 g2f3 a8b8 h1g1 b4b3 a2b1 b3c2 b1c2 b8b2 g3f2 b2b4 e1c1 b4b8');
INSERT INTO evaluations VALUES(19649,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(19649,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19649,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(19649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19649,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(19649,36,'cp 15','c1g5 a7a6 a2a4 g8f6 f2f4 f8g7 g1f3 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 b7b5 a4b5 c6d4 f3d1 a6b5 c3b5 a8a1 d1a1 d4c2 a1a7 h7h6 g5h4 d6d5 e4d5 f6d5 a7c5 c2e3 f1f3 e3f5 h4f2 e7e6 c4d5 d8d5 c5d5 e6d5 b5c3 f8b8 g2g4 f5d4 f3e3 d4e6 c3d5 b8b2 g1f1 g7d4');
INSERT INTO evaluations VALUES(19518,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19518,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19518,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19518,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(19518,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19518,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19518,36,'cp 0','a3b4 c6b4 d2d4 e7e6 c2c3 b4c6 b1d2 f8e7 f1d3 d7d6 f2f4 d6d5 e4e5 g8h6 d2f3 c6a5 d1a4 a5c6');
INSERT INTO evaluations VALUES(19515,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19515,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19515,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19515,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19515,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19515,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19515,36,'cp 0','c6b4 d2d4 e7e6 c2c3 b4c6 b1d2 d7d6 f1d3 f8e7 f2f4 d6d5 e4e5 g8h6 d2f3 c6a5 d1a4 a5c6');
INSERT INTO evaluations VALUES(19612,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(19612,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(19612,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(19612,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(19612,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19612,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19612,36,'cp 0','d2d4 e7e6 c2c3 b4c6 b1d2 d7d6 f1d3 f8e7 f2f4 d6d5 e4e5 g8h6 d2f3 c6a5 d1a4 a5c6');
INSERT INTO evaluations VALUES(19610,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19610,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19610,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19610,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(19610,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19610,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(19610,36,'cp 0','e7e6 g1f3 d7d5 e4d5 e6d5 f1b5 b4c6 f3e5 d8c7 d1e2 c8e6 b5c6 b7c6 a1a6 c7b7 e1g1 b7b1 e5f7 e8f7 a6c6 a8e8 e2h5 g7g6 h5f3 e6f5 f3d5 e8e6 c6e6 f5e6 d5f3 e6f5 f3d5');
INSERT INTO evaluations VALUES(52832,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52832,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(52832,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52832,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52832,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(52832,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52832,36,'cp 0','g1f3 c8f5 d2d3 h7h6 b1d2 g8f6 g2g3 e7e6 f1g2 f8e7 e2e4 f5h7 e1g1 e8g8 d1e2 a7a5 a2a3 f8e8 f1e1 d8c7 h2h4 e7f8 b2e5 c7e7 e5b2');
INSERT INTO evaluations VALUES(52830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52830,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52830,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52830,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52830,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52830,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52830,36,'cp 0','c8f5 d2d3 h7h6 b1d2 g8f6 g2g3 e7e6 f1g2 f8e7 e2e4 f5h7 e1g1 e8g8 d1e2 a7a5 a2a3 f8e8 f1e1 e7f8 h2h4 d8c7 b2e5 c7d8');
INSERT INTO evaluations VALUES(54902,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(54902,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54902,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54902,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54902,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54902,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(54902,36,'cp 3','g8f6 e2e3 f8g7 c2c3 e8g8 g1f3 c7c5 b1d2 c5d4 e3d4 b8c6 f1d3 f6h5 e1g1 d8b6 d2b3 c8g4 d3e2 h5g3 h2g3 a7a5 a2a4 a8d8 f1e1 g4f3 g2f3 e7e5 d4e5 c6e5 g1g2 e5c4 d1c2 f8e8 e2c4 d5c4');
INSERT INTO evaluations VALUES(46862,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46862,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(46862,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(46862,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46862,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(46862,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(46862,36,'cp 15','e2e3 f8g7 b1d2 e8g8 f1d3 c7c5 c2c3 b8c6 g1e2 f6h5 e1g1 d8b6 a1c1 c5d4 c3d4 h5g3 e2g3 e7e5 d4e5 c6e5 d3f5 d5d4 e3d4 b6d4 d2e4 d4b2 e4d6 c8e6 g3h5 g7h8 f5e6 f7e6 c1b1 b2a2 f2f4 g5f4 h5f4 b7b6');
INSERT INTO evaluations VALUES(56361,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(56361,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(56361,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(56361,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(56361,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(56361,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(56361,36,'cp -26','e5e4 f2f3 g8f6 f3e4 f6e4 c3e4 d5e4 g1e2 f8d6 c2c4 c7c6 d1c2 e8g8 c1d2 b8d7 e1c1 f7f5 g2g3 d7f6 c4c5 d6c7 e2f4 c7f4 f1c4 f6d5 g3f4 c8e6 a2a4 g8h8 b2b3 f8f7 d1g1 a8b8 h2h4 h7h6 g1g6 f7f6 g6g2 b7b5 a4b5 c6b5 c4d5 d8d5 c1b2 b5b4 h1g1 f6f7');
INSERT INTO evaluations VALUES(56912,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(56912,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(56912,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(56912,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(56912,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(56912,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(56912,36,'cp 164','d2d4 d7d6 g1f3 b8d7 f1c4 h7h6 h2h3 g8f6 c1e3 d8e7 d1d2 f8g7 e1g1 c7c6 a1d1 g6g5 f1e1 e8g8 d4d5 c6c5 c4e2 f6h7 f3h2 d7f6 h2f1 a7a6 a2a4 g5g4 h3g4 f6g4 e2g4 c8g4 f2f3 g4d7 f1g3 b7b5 g3f5 d7f5 e4f5 e7d7 c3e4 d7f5 e4d6');
INSERT INTO evaluations VALUES(1926,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(1926,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(1926,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(1926,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(1926,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(1926,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(1926,36,'cp 46','f8b4 c1d2 g8f6 d1e2 e8g8 f2f3 b4c5 e2d3 d7d6 d2g5 f6d7 c3d5 d7e5 d5e7 g8h8 e7c6 e5d3 f1d3 b7c6 g2g4 c8e6 b2b3 c5d4 e1c1 c6c5 a2a4 a7a5 d3b5 h7h6 g5f4 f8b8 h1e1 g7g5 f4d2 h8g7 h2h4 g7f6 e1h1 d4e5 h4g5 h6g5 h1h7');
INSERT INTO evaluations VALUES(350,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(350,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(350,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(350,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(350,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(350,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(350,36,'cp 58','c1d2 g8f6 c3d5 b4d2 d1d2 f6e4 d2d4 e8d8 e1c1 e4f6 f1c4 f6d5 c4d5 c6h6 c1b1 c7c6 d5f7 h8f8 f7c4 f8f4 d4d3 d8c7 c4g8 d7d6 h1e1 h6f6 g8e6 d6d5 e6c8 a8c8 f2f3 f4b4 d3a3 b4b5 e1e7 c7b8 d1e1 a7a5 e7e5 c8d8 a3e7 a5a4 e7f6 g7f6 e5e7 a4a3 b2b3');
INSERT INTO evaluations VALUES(351,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(351,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(351,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(351,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(351,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(351,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(351,36,'cp 49','g8f6 f1b5 c6e6 e1g1 e8g8 b5d3 b4c3 d2c3 f6e4 d3e4 e6e4 f1e1 e4c4 d1f3 d7d6 e1e7 c4g4 f3g4 c8g4 e7c7 b7b6 a1e1 f8c8 c7c8 a8c8 e1e7 a7a5 e7b7 a5a4 a2a3 g4f5 b7b6 f5c2 b6d6 c2b3 f2f3 h7h5 g1f2 c8c5 d6d7 f7f6 d7a7 c5d5');
INSERT INTO evaluations VALUES(33214,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(33214,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(33214,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(33214,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(33214,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(33214,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(33214,36,'cp 35','b1c3 g8e7 d2d3 d7d6 c1e3 c5e3 f2e3 f6g6 g1e2 e8g8 e1g1 c8g4 f3g3 b8d7 g3h4 d7f6 h2h3 g4e6 c4b3 c7c6 f1f2 b7b5 e2g3 h7h6 a1f1 g6g5 h4g5 h6g5 f2f6 g7f6 f1f6 d6d5 e4d5 e7d5 c3d5 c6d5 g3f5 e6f5 f6f5');
INSERT INTO evaluations VALUES(33205,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(33205,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(33205,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(33205,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(33205,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(33205,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(33205,36,'cp 30','g8e7 d2d3 d7d6 c1e3 c5e3 f2e3 c8e6 f3f6 g7f6 g1f3 h8g8 f3h4 c7c6 e1g1 g8g4 g2g3 b8d7 c4b3 e8c8 d3d4 d8g8 d4d5 e6d5 e4d5 g4h4 d5c6 b7c6 e3e4 h4f4 b3f7 f4f1 a1f1 g8f8 f7e6 c8c7 b2b4 d7b6 b4b5 d6d5 b5c6 c7c6 e4d5 e7d5 c3e4 h7h6 f1d1');
INSERT INTO evaluations VALUES(33751,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(33751,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(33751,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(33751,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(33751,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(33751,35,'cp -121',NULL);
INSERT INTO evaluations VALUES(33751,36,'cp -106','d5e4 d1e2 e4g6 g1f3 c8g4 h2h3 g4f3 e2f3 b8c6 d2d3 e7e5 a2a3 a8d8 f3g4 c6d4 g4g6 h7g6 a1b1 d4b3 f1e2 f7f5 c1e3 e8f7 e2d1 b3d4 b2b4 f8e7 e3d4 d8d4 e1e2 f7f6 c4c5 c7c6 d1b3 g6g5 g2g3 g7g6 b3c4 d4d7 b1e1 h8d8 e2f1');
INSERT INTO evaluations VALUES(33764,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(33764,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(33764,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(33764,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(33764,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(33764,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(33764,36,'cp -100','d1e2 e4g6 g1f3 b8c6 f3e5 c6e5 e2e5 f7f6 e5g3 e7e5 d2d3 c8d7 c1d2 e8c8 d2c3 d7f5 a1d1 f8c5 f1e2 c7c6 g3g6 h7g6 h2h4 c5d4 g2g4 f5e6 f2f3 d4e3 c3d2 e3d2 d1d2 c6c5 b2b3 c8c7 e1f2 e6d7 h4h5 d7c6 h1g1 g6h5 g4h5');
INSERT INTO evaluations VALUES(38972,30,'cp -264',NULL);
INSERT INTO evaluations VALUES(38972,31,'cp -286',NULL);
INSERT INTO evaluations VALUES(38972,32,'cp -297',NULL);
INSERT INTO evaluations VALUES(38972,33,'cp -303',NULL);
INSERT INTO evaluations VALUES(38972,34,'cp -311',NULL);
INSERT INTO evaluations VALUES(38972,35,'cp -294',NULL);
INSERT INTO evaluations VALUES(38972,36,'cp -309','e8f7 d1h5 f7f8 h5e5 d7d6 e5g3 g8f6 d2d3 b8c6 g1e2 f8f7 c1e3 c5e3 g3e3 h8e8 a2a3 d6d5 b1d2 d8e7 f2f3 f7g8 e1g1 c8f5 a1e1 f5g6 b2b4 e7d7 e3f2 d5e4 f3e4 a7a5 b4b5 c6e5 h2h3 d7b5 e2f4 b5c6 d3d4 e5c4 f4g6 c4d2 f2d2 h7g6 e4e5 f6d5 c2c4 c6c4');
INSERT INTO evaluations VALUES(37486,30,'cp -290',NULL);
INSERT INTO evaluations VALUES(37486,31,'cp -294',NULL);
INSERT INTO evaluations VALUES(37486,32,'cp -305',NULL);
INSERT INTO evaluations VALUES(37486,33,'cp -307',NULL);
INSERT INTO evaluations VALUES(37486,34,'cp -290',NULL);
INSERT INTO evaluations VALUES(37486,35,'cp -317',NULL);
INSERT INTO evaluations VALUES(37486,36,'cp -299','d1h5 f7f8 h5e5 d7d6 e5g3 g8f6 b1c3 b8c6 g1e2 f8f7 d2d3 h8f8 c1e3 c5e3 g3e3 f7g8 f2f4 d6d5 e3c5 d8d7 h2h3 b7b6 c5f2 d5d4 c3b1 c6b4 e2g3 c7c5 e1g1 c8b7 b1a3 b6b5 a3b1 b4d5 e4d5 f6d5 b1d2 d5f4 f2f4 f8f4 f1f4 a8f8 f4f8 g8f8 d2e4');
INSERT INTO evaluations VALUES(37485,30,'cp -300',NULL);
INSERT INTO evaluations VALUES(37485,31,'cp -295',NULL);
INSERT INTO evaluations VALUES(37485,32,'cp -290',NULL);
INSERT INTO evaluations VALUES(37485,33,'cp -298',NULL);
INSERT INTO evaluations VALUES(37485,34,'cp -318',NULL);
INSERT INTO evaluations VALUES(37485,35,'cp -302',NULL);
INSERT INTO evaluations VALUES(37485,36,'cp -316','f7f8 h5e5 d7d6 e5g3 b8c6 g1e2 g8f6 b1c3 f8f7 d2d3 h8f8 c1e3 c5e3 g3e3 d6d5 e3c5 c6e7 f2f3 f7g8 e1c1 b7b6 c5e3 e7c6 e3g1 a7a6 g2g4 d5d4 c3b1 c8e6 h2h4 d8d6 h4h5 e6a2 f3f4 f6d7 e4e5 d6c5 h5h6 g7g6 b1d2');
INSERT INTO evaluations VALUES(35557,30,'cp -308',NULL);
INSERT INTO evaluations VALUES(35557,31,'cp -311',NULL);
INSERT INTO evaluations VALUES(35557,32,'cp -306',NULL);
INSERT INTO evaluations VALUES(35557,33,'cp -304',NULL);
INSERT INTO evaluations VALUES(35557,34,'cp -301',NULL);
INSERT INTO evaluations VALUES(35557,35,'cp -295',NULL);
INSERT INTO evaluations VALUES(35557,36,'cp -302','h5e5 d7d6 e5g3 g8f6 d2d3 b8c6 c1e3 c5e3 g3e3 f8f7 g1e2 h8e8 f2f3 d6d5 e3f2 c8e6 h2h3 c6b4 b1a3 f7g8 e1g1 e6f7 a1d1 d8e7 c2c3 d5e4 f3e4 b4c6 a3c2 g8h8 c2e3 f7g6 e2f4 c6e5 f4g6 e5g6 e3f5 e7d7');
INSERT INTO evaluations VALUES(35556,30,'cp -313',NULL);
INSERT INTO evaluations VALUES(35556,31,'cp -313',NULL);
INSERT INTO evaluations VALUES(35556,32,'cp -305',NULL);
INSERT INTO evaluations VALUES(35556,33,'cp -300',NULL);
INSERT INTO evaluations VALUES(35556,34,'cp -313',NULL);
INSERT INTO evaluations VALUES(35556,35,'cp -305',NULL);
INSERT INTO evaluations VALUES(35556,36,'cp -301','d7d6 e5f4 g8f6 g1e2 b8c6 d2d3 f8f7 c1e3 c5e3 f4e3 h8e8 e1g1 f7g8 f2f3 d6d5 e3f2 c8e6 b1c3 d8e7 d3d4 d5e4 f3e4 e6c4 e4e5 f6g4 f2f4 g4h6 b2b3 c4e2 c3e2 c6d8 f4e4 c7c6 e2f4 d8e6 c2c3 e7d7 a1e1 e8f8 h2h3 a8e8 f4e6 f8f1 e1f1 d7e6 g2g4');
INSERT INTO evaluations VALUES(35555,30,'cp -301',NULL);
INSERT INTO evaluations VALUES(35555,31,'cp -300',NULL);
INSERT INTO evaluations VALUES(35555,32,'cp -303',NULL);
INSERT INTO evaluations VALUES(35555,33,'cp -301',NULL);
INSERT INTO evaluations VALUES(35555,34,'cp -292',NULL);
INSERT INTO evaluations VALUES(35555,35,'cp -308',NULL);
INSERT INTO evaluations VALUES(35555,36,'cp -300','e5g3 g8f6 d2d3 b8c6 c1e3 c5e3 g3e3 f8f7 g1e2 h8e8 f2f3 d6d5 e3f2 f7g8 b1c3 c8e6 e1g1 d5e4 f3e4 g8h8 h2h3 e6g8 a1e1 d8d6 c3b5 d6e5 b5d4 c6d4 e2d4 c7c5 d4f3 e5c7 e4e5 f6d7 e1e4 d7f8 f3g5 a8d8 g5f7 g8f7');
INSERT INTO evaluations VALUES(42965,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42965,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(42965,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42965,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42965,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42965,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(42965,36,'cp 0','h7h6 g5f4 d7d5 e2e3 e8g8 a1c1 c7c5 d4c5 e7c5 g1f3 b8c6 c4d5 e6d5 f1e2 d5d4 e3d4 c5d4 f3d4 c6d4 f4e3 d4e2 d1d8 f8d8 e1e2 c8e6 f2f3 f6d5 c3d5 d8d5 a2a3 d5b5 e3d4 b5d5');
INSERT INTO evaluations VALUES(53156,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(53156,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(53156,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(53156,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(53156,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(53156,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(53156,36,'cp -25','c8d7 b5d7 b8d7 f2f4 e7e6 g1f3 g8e7 e1g1 e7f5 d2d3 d7b8 b1a3 h7h5 c2c3 f8e7 a3c2 d5d4 c3c4 b8c6 d1e2 a7a6 b2b3 d8d7 a1b1 a8d8 a2a4 e8g8 g2g3 g7g6 c1a3 f8e8 e2e4 f5e3 c2e3 d4e3 f1d1 c6d4 f3d4 c5d4 a3e7 e8e7 a4a5 d7c7');
INSERT INTO evaluations VALUES(30941,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(30941,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(30941,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(30941,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(30941,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(30941,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(30941,36,'cp -40','b5d7 b8d7 f2f4 e7e6 g1f3 g8h6 e1g1 h6f5 d2d3 d8b6 b2b3 c5c4 d3d4 a8c8 g2g4 f5e7 b3c4 h7h5 h2h3 b6c6 b1d2 h5g4 h3g4 f7f5 e5f6 g7f6 f1e1 f6f5 g4g5 d5c4 d1e2 e8f7 a1b1 e7g6 e2c4 c6c4 d2c4 b7b6 d4d5 e6d5 f3e5 g6e5 c4e5 d7e5 e1e5 c8c2 c1e3 h8h3 e5f5 f7g8 e3d4 f8c5 d4c5 h3g3 g1f1 b6c5');
INSERT INTO evaluations VALUES(30867,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(30867,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(30867,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(30867,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(30867,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(30867,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(30867,36,'cp -26','b8d7 f2f4 g8h6 g1f3 e7e6 e1g1 h6f5 d2d3 d8b6 c2c3 f8e7 a2a4 h7h5 b1a3 d7b8 d1e2 b8c6 c1e3 f5e3 e2e3 a8d8 e3e2 g7g6 f1b1 c6a5 a3c2 a5b3 a1a2 b6a6 g2g3 e8g8 h2h3 b7b5 d3d4 c5d4');
INSERT INTO evaluations VALUES(53393,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(53393,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(53393,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53393,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(53393,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(53393,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(53393,36,'cp 8','e3d4 b8c6 g1f3 c8f5 c2c4 e7e6 c4d5 e6d5 b1c3 f8d6 f4e5 g8f6 f1b5 f6e4 e1g1 e4c3 b2c3 e8g8 b5d3 f5g6 d3g6 h7g6 d1b3 d6e5 d4e5 d8d7 f1e1 a8d8 a1d1 f8e8 a2a4 c6a5 b3b5 d7b5 a4b5 a5b3 d1b1');
INSERT INTO evaluations VALUES(53994,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(53994,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(53994,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(53994,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(53994,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(53994,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(53994,36,'cp 87','b8d7 f1d3 f8e7 e1g1 g8f6 f1e1 e8g8 d1e2 e6e5 h2h3 c5d4 c3d4 e5d4 f3d4 d6d5 e4e5 f6e4 d3e4 d5e4 e2e4 d7e5 b1c3 e5g6 e1d1 e7h4 d4e6 h4f2 g1h2 c8e6 d1d8 f8d8 c1e3 f2e3 e4e3 d8e8 c3b5 e6d7 e3b3 d7b5 b3b5 e8e7 a1d1 g6f8 a2a4 a7a6 b5b6 a8e8 a4a5');
INSERT INTO evaluations VALUES(54689,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(54689,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(54689,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(54689,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(54689,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(54689,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(54689,36,'cp 57','d1d3 b8c6 g1f3 g8f6 b1c3 d8d7 a2a3 g7g6 e2e3 f8g7 e1c1 e8g8 f4g5 e7e6 g5f6 g7f6 h2h4 h7h6 d4d5 c6d8 d5e6 d8e6 c3d5 d7f7 d5f6 f7f6 f3d4 e6c5 d3d2 a7a6 f1c4 g8h7 f2f3 c8d7 d4e2 c5a4 d2b4 a4c5 c4a2 a8e8 e2f4 e8e3');
INSERT INTO evaluations VALUES(723,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(723,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(723,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(723,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(723,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(723,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(723,36,'cp 62','c3a4 b6c7 a4c5 c6d4 f3d4 d7c5 a2a3 c8d7 c2c4 d7a4 d1c1 c7d7 c4d5 d7d5 b2b4 c5b3 c1c2 f8e7 f1c4 e7h4 e1d1 d5c4 c2c4 b3a5 c4c2 a4c2 d1c2 a5c4 c2d3 c4e3 d3e3 e8c8 d4b5 c8b8 b5d6 d8d7 h1d1 h8d8 e3e2 f7f6 d1d3 f6e5 f4e5 h4e7 a1d1 e7d6 d3d6 d7d6');
INSERT INTO evaluations VALUES(733,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(733,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(733,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(733,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(733,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(733,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(733,36,'cp 53','b6c7 a4c5 c6d4 f3d4 d7c5 d4b5 c7c6 a1c1 a7a6 b5d4 c6b6 d4b3 b6b4 c2c3 b4e4 e1f2 b7b6 f1e2 c8b7 b3d4 g7g5 g2g3 g5f4 e3f4 e4g6 f2g2 h7h5 h2h4 f8h6 g2h2 h6f4 g3f4 e8c8 h1g1 g6h6 d1d2 c8b8 d2e3 c5e4 c1f1 d8c8 g1g2 h8g8 f1g1 g8g2 g1g2 f7f6');
INSERT INTO evaluations VALUES(2137,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(2137,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(2137,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(2137,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(2137,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(2137,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(2137,36,'cp 43','c2c4 e7e6 b1c3 g8f6 f1e2 c6d4 d1d4 f8c5 d4d3 e8g8 c1e3 b7b6 e3c5 c7c5 e1g1 f8d8 f1d1 c8b7 d1d2 d7d6 a1d1 d8d7 e2f1 a8d8 d3f3 c5c6 d1e1 c6c5 h2h3 b7c6 a2a3 a7a5 e1d1 g8f8 d2d4 h7h6 f3e2 f8g8 e2e3 g8f8 f2f3');
INSERT INTO evaluations VALUES(30060,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(30060,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30060,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30060,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(30060,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30060,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30060,36,'cp 0','f8b4 g1f3 e8g8 f1e2 f6e4 d1b3 b8c6 c4d5 e6d5 e1g1 b4c3 b2c3 c6a5 b3b4 c7c6 f1d1 f5g4 h2h3 g4h5 g2g4 h5g6 f3d2 e4d2 d1d2 f8e8 b4d6 d8h4 g1g2 e8e6 d6c7 b7b6 f4g3 h4f6 g3e5 f6g5 e5f4');
INSERT INTO evaluations VALUES(28484,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28484,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28484,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28484,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28484,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28484,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28484,36,'cp 0','g1f3 e8g8 f1e2 f6e4 d1b3 b8c6 c4d5 e6d5 e1g1 b4c3 b2c3 c6a5 b3b4 c7c6 f1d1 f5g4 h2h3 g4h5 g2g4 h5g6 f3d2 e4d2 d1d2 f8e8 b4d6 d8h4 g1g2 e8e6 d6c7 b7b6 f2f3 e6e8 c7d6 e8e6');
INSERT INTO evaluations VALUES(28485,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28485,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28485,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28485,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28485,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28485,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28485,36,'cp 0','e8g8 f1e2 f6e4 d1b3 b8c6 c4d5 e6d5 e1g1 b4c3 b2c3 c6a5 b3b4 c7c6 f1d1 f5g4 h2h3 g4h5 g2g4 h5g6 f3d2 e4d2 d1d2 f8e8 b4d6 d8h4 g1g2 e8e6 d6c7 b7b6 d2b2 g6e4 f2f3 e4g6 a2a4 h7h5 f4g3 h4f6 g3e5 f6h4 e5g3');
INSERT INTO evaluations VALUES(52110,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52110,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52110,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52110,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52110,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(52110,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52110,36,'cp 16','b8c6 g2g3 g7g6 f1g2 f8g7 e1g1 d7d6 a1b1 g8e7 d2d3 e8g8 a2a3 a7a5 f3e1 c8e6 c1g5 f7f6 g5d2 f6f5 d2g5 d8d7 c3d5 a8b8 b2b4 a5b4 d5e7 c6e7 a3b4 e7c6 b4b5 c6d8 g5d8 b8d8 d1c2 h7h6 e1f3 g8h8 f1d1 d7f7');
INSERT INTO evaluations VALUES(16364,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(16364,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(16364,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(16364,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(16364,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(16364,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(16364,36,'cp -53','f8b4 g1f3 h7h6 g5f6 d8f6 a2a3 b4c3 b2c3 e8g8 f1e2 b7b6 f3d2 c8b7 e2f3 c6e7 c3c4 f6d6 e1g1 f7f5 f1e1 b7c6 a3a4 a7a5 e4f5 f8f5 d2e4 d6g6 f3g4 f5f8 g4h5 g6f5 h5g4 f5f7 d1f3 f7f3 g4f3 a8d8 h2h4 f8f4 e4d2 d7d6');
INSERT INTO evaluations VALUES(16534,30,'cp -328',NULL);
INSERT INTO evaluations VALUES(16534,31,'cp -332',NULL);
INSERT INTO evaluations VALUES(16534,32,'cp -331',NULL);
INSERT INTO evaluations VALUES(16534,33,'cp -354',NULL);
INSERT INTO evaluations VALUES(16534,34,'cp -343',NULL);
INSERT INTO evaluations VALUES(16534,35,'cp -344',NULL);
INSERT INTO evaluations VALUES(16534,36,'cp -341','d7d5 d2d3 c8g4 f3e3 d5e4 b1d2 e4d3 e3d3 d8e7 f1e2 g4d7 d3c2 e8c8 d2c4 e5e4 c1g5 h7h6 g5f6 e7f6 c2e4 f8c5 e4f3 f6e7 e1c1 h8e8 h2h4 g7g6 b2b4 c5b6 f3f4 c8b8 c4b6 a7b6 e2b5 g6g5 h4g5 h6g5 f4g3 g5g4 h1h5 c6b4 b5d7 d8d7 d1d7 e7d7 c3b4 e8e1 c1c2');
INSERT INTO evaluations VALUES(15144,30,'cp -334',NULL);
INSERT INTO evaluations VALUES(15144,31,'cp -331',NULL);
INSERT INTO evaluations VALUES(15144,32,'cp -340',NULL);
INSERT INTO evaluations VALUES(15144,33,'cp -329',NULL);
INSERT INTO evaluations VALUES(15144,34,'cp -336',NULL);
INSERT INTO evaluations VALUES(15144,35,'cp -320',NULL);
INSERT INTO evaluations VALUES(15144,36,'cp -338','d2d3 c8g4 f3g3 d5e4 c1g5 d8d7 b1d2 e8c8 f2f3 e4f3 g1f3 f8e7 g3f2 d7d5 f1e2 g4f5 g5f6 e7f6 d2e4 f5e4 d3e4 d5e4 e1g1 c8b8 e2d1 f6e7 d1c2 e4c4 g1h1 a7a5 f3d2 c4a6 c2e4 g7g6 d2f3 f7f6 f2c2 a5a4 b2b4 f6f5 e4c6 a6c6 f3e5');
INSERT INTO evaluations VALUES(19316,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19316,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(19316,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19316,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(19316,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(19316,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(19316,36,'cp -13','f8g7 f1d3 g8f6 e1g1 e8g8 d4c6 b7c6 e4e5 f6d5 f1e1 a7a5 b1d2 d7d6 e5d6 d8d6 d2c4 d6c7 c1g5 a5a4 g5h4 e7e5 d1d2 c8e6 h4g3 d5f4 d3f1 e6d5 c4e5 g7e5 e1e5 c7e5 g3f4 e5f5 c3c4 d5e6 f4h6 f8d8 d2e3 f7f6 h2h3 g8f7 a1e1 c6c5 g2g4');
INSERT INTO evaluations VALUES(22015,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(22015,31,'cp -181',NULL);
INSERT INTO evaluations VALUES(22015,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(22015,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(22015,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(22015,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(22015,36,'cp -169','g7g5 d2d4 h7h6 c2c3 f8g7 d1b3 d8d7 c4d3 g8f6 c1d2 e8g8 b3c2 f8e8 a2a4 c6e7 h2h3 d7d8 b2b4 e7g6 d2e1 a7a5 b4b5 c7c5 b5c6 b7c6 e1f2 a8b8 f1e1 g5g4 h3g4 f6g4 b1d2 g8h8 a1b1 b8b1 e1b1 g4f2 g1f2');
INSERT INTO evaluations VALUES(21249,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(21249,31,'cp -166',NULL);
INSERT INTO evaluations VALUES(21249,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(21249,33,'cp -165',NULL);
INSERT INTO evaluations VALUES(21249,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(21249,35,'cp -181',NULL);
INSERT INTO evaluations VALUES(21249,36,'cp -188','c2c3 h7h6 d2d4 f8g7 d1b3 d8d7 c4d3 g8f6 f1e1 e8g8 b1d2 c6e7 e4e5 f6h5 b3d1 g5g4 h2h3 g4h3 f3g5 d7g4 d1g4 c8g4 g5h3 d6e5 h3f2 g4d7 d4e5 a8d8 d3e4 d7c6 e4f3 c6f3 d2f3 e7g6 e5e6 f7e6 e1e6 g8f7 e6e1 h5g3 c1d2 f8e8 e1e8');
INSERT INTO evaluations VALUES(22049,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(22049,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(22049,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(22049,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(22049,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(22049,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(22049,36,'cp 19','g8f6 b1c3 f8e7 f1d3 e8g8 h2h3 f8e8 e1g1 e7f8 f1e1 c6e5 f3e5 d6e5 d1f3 c7c6 c1g5 c8e6 e1d1 f8e7 d3c4 d8c8 c4e6 c8e6 b2b3 a8d8 c3a4 b7b6 a4b2 d8d4 c2c3 d4d8 d1d8 e8d8 a1d1 h7h6 d1d8 e7d8 g5f6 d8f6 b2c4 f6e7 c4e3');
INSERT INTO evaluations VALUES(15659,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(15659,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15659,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(15659,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(15659,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(15659,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(15659,36,'cp 30','b1c3 f8e7 f1d3 e8g8 h2h3 f8e8 e1g1 e7f8 f1e1 c6e5 f3e5 d6e5 d1f3 c8e6 c1g5 c7c6 e1d1 f8e7 d3c4 d8c8 c4e6 c8e6 f3f5 e6f5 e4f5 e5e4 a2a4 h7h6 g5h4 g8f8 g1f1 a7a5 g2g4 a8d8 f1e2 f6d5 h4e7 e8e7 c3d5');
INSERT INTO evaluations VALUES(15653,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15653,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(15653,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(15653,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(15653,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(15653,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(15653,36,'cp 22','f8e7 f1d3 e8g8 h2h3 f8e8 e1g1 e7f8 f1e1 c6e5 f3e5 d6e5 d1f3 c7c6 c1g5 h7h6 g5e3 d8e7 e1d1 f6d7 d3c4 d7f6 c3e2 c8e6 c4e6 e7e6 e2g3 a7a6 a2a4 b7b5 f3e2 a8d8 d1d8 e8d8 c2c3 d8b8 f2f3 f6d7 g3f5');
INSERT INTO evaluations VALUES(8659,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(8659,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(8659,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(8659,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(8659,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(8659,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(8659,36,'cp 23','f1d3 e8g8 h2h3 f8e8 e1g1 e7f8 f1e1 c6e5 f3e5 d6e5 d1f3 c8e6 c1g5 h7h6 g5f6 d8f6 f3f6 g7f6 c3d5 e6d5 e4d5 e8d8 a1d1 d8d6 g1f1 a8d8 d3e4 c7c6 c2c4 c6d5 e4d5 b7b5 b2b3 b5c4 b3c4 d8d7 d1b1 d6b6 a2a4 f8b4 e1e4 d7d8 f2f4 b4d6 b1b6 a7b6 f4e5 f6e5');
INSERT INTO evaluations VALUES(35301,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(35301,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(35301,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(35301,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(35301,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(35301,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(35301,36,'cp 20','d7d5 d2d3 g8f6 c1f4 b8c6 f1e2 d5d4 c3d5 f6d5 e4d5 c6b4 d5d6 c7d6 f4g5 e7c7 f3d4 f8e7 g5e7 c7e7 c2c3 b4c6 d4f3 e8g8 e1g1 d6d5 d3d4 c8f5 d1d2 f5e4 a1e1 h7h6 d2f4 f7f5 h2h4 e7d8 e2d1 c6a5 f3e5 a5c4 e5c4 d5c4 d1a4 d8b6 f1f2');
INSERT INTO evaluations VALUES(44703,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(44703,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(44703,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(44703,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(44703,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44703,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(44703,36,'cp 18','d5c4 a2a4 f6d5 e2e4 d5c3 b2c3 b7b5 d1d2 a7a6 f1e2 b8d7 h2h4 h7h6 g5f4 f8g7 e1g1 h6h5 f1b1 e8g8 a4b5 c6b5 b1b5 a6b5 a1a8 d8b6 f4e3 d7f6 f3g5 b6b7 a8a5 f6e4 g5e4 b7e4 a5b5 c8e6 e2f3 e4d3 d2d3 c4d3 f3e4 f8c8 e3d2 e6c4 b5b4 c4a6 b4b6 a6c4 b6c6 c8c6 e4c6 g8f8 d2e3');
INSERT INTO evaluations VALUES(54274,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(54274,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(54274,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(54274,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(54274,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(54274,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(54274,36,'cp -40','c5b4 a2a3 b4a3 c2c3 c8d7 b1a3 b8c6 d2d4 a7a6 f1d3 g8e7 e1g1 h7h6 a3c2 a8c8 c1d2 c6a5 d1b1 d8c7 c2e3 a5c4 d2c1 d7b5 f3d2 c7c6 d2c4 b5c4 d3c4 d5c4 c1a3 e7d5 e3d5 c6d5 a3f8 h8f8');
INSERT INTO evaluations VALUES(54403,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(54403,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(54403,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(54403,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(54403,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(54403,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(54403,36,'cp -25','a2a3 b4a3 c2c3 h7h6 d2d4 c8d7 b1a3 b8c6 f1d3 a8c8 e1g1 a7a6 f3d2 g8e7 d2b3 e7f5 f1e1 f8e7 a3c2 c6a7 b3a5 b7b6 a5b7 d8c7 d3a6 e8g8 b7d6 f5d6 e5d6 e7d6 a6c8 a7c8 h2h4 c7c3 c1a3 c3c7 d1d3 d6a3 a1a3 c8e7');
INSERT INTO evaluations VALUES(54404,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(54404,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(54404,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(54404,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(54404,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(54404,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(54404,36,'cp -25','b4a3 c2c3 a3a2 a1a2 b8c6 d2d4 a8b8 f1b5 g8e7 b5d3 a7a6 b1d2 b7b5 d2b3 b5b4 e1g1 b4c3 b3c5 c6b4 a2a3 e7c6 a3c3 b4d3 d1d3 c6b4 d3e3 f8e7 c1d2 e8g8 f1a1 a6a5 h2h4 c8a6 h4h5 h7h6 c5b3 a6c4 b3a5 b8a8');
INSERT INTO evaluations VALUES(54439,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(54439,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(54439,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(54439,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(54439,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(54439,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(54439,36,'cp -33','c2c3 c8d7 b1a3 b8c6 d2d4 a8c8 f1d3 a7a6 e1g1 c6a5 a3b1 d8b6 d3c2 g8e7 f3g5 h7h6 g5h3 g7g5 f2f4 g5g4 d1g4 c8c3 b1c3 b6d4 f1f2 d4c3 a1b1 h8g8 g4h5 b7b5 f4f5 e7f5 c1b2 c3b4 c2f5 e6f5 e5e6 d7e6 h3f4 f8c5 f4e6');
INSERT INTO evaluations VALUES(7117,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(7117,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(7117,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(7117,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(7117,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(7117,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(7117,36,'cp 31','b7b5 a4c2 e8g8 h2h3 d6d5 e4d5 d8d5 d2d4 e5d4 c3d4 c8f5 b1c3 f5c2 d1c2 d5d7 a2a4 a8b8 a4b5 a6b5 c1f4 b5b4 c3e2 b4b3 c2c4 b8b4 c4d3 b4b6 a1c1 f8d8 e1d1 h7h6 f3e5 c6e5 f4e5 f6d5 d3f3');
INSERT INTO evaluations VALUES(7400,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(7400,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(7400,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(7400,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(7400,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(7400,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(7400,36,'cp 51','a4c2 e8g8 h2h3 d6d5 d2d4 e5d4 e4e5 f6e4 c3d4 c6b4 c1e3 c8f5 c2b3 d8d7 a2a3 b4c6 b1d2 c6a5 b3c2 e4d2 e3d2 a5c4 d2c1 a6a5 b2b3 c4b6 c1g5 e7g5 f3g5 h7h6 c2f5 d7f5 g5f3 a5a4 a1c1 a4b3 d1b3 f5d7 e5e6 f7e6');
INSERT INTO evaluations VALUES(7399,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(7399,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(7399,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(7399,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(7399,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(7399,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(7399,36,'cp 44','c8g4 h2h3 g4h5 a2a4 e8g8 d2d3 d6d5 g2g4 d5e4 g4h5 e4f3 d1f3 d8d5 c2d1 a8d8 a4b5 a6b5 h5h6 d8d6 b1d2 d5d3 f3d3 d6d3 d1e2 d3d5 e2f3 d5d6 f3c6 d6c6 e1e5 e7c5 d2b3 c5b6 c1g5 g7h6 g5h6 f8d8 h6g5 g8g7 e5b5');
INSERT INTO evaluations VALUES(39592,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39592,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39592,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39592,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39592,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(39592,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39592,36,'cp 0','g8f6 g2g3 e8g8 f1g2 d7d5 e1g1 b8d7 a2a4 a7a5 f1e1 c7c6 b1d2 c6c5 d1b3 e7e6 e2e4 c5d4 c3d4 f6e4 d2e4 d5e4 e1e4 b7b6 c1g5 g7f6 e4e1 c8b7 g5h6 f6g7 h6g5');
INSERT INTO evaluations VALUES(40869,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(40869,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(40869,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(40869,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(40869,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(40869,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(40869,36,'cp 38','f1d3 d6d5 e1g1 c8f5 f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 f5d3 d1d3 d5d7 a1b1 b7b6 d4d5 a8e8 c1f4 c6a5 c3c4 a5b7 f3e5 d7a4 e5c6 e7d6 f4d6 b7d6 d3b3 e8e1 b1e1 a4b3 a2b3 a7a5 f2f3 a5a4 b3a4 f8a8 g1f2 a8a4');
INSERT INTO evaluations VALUES(40868,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(40868,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40868,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(40868,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(40868,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(40868,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(40868,36,'cp 42','d6d5 e1g1 c8f5 f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 a8e8 c1f4 f5d3 d1d3 d5d7 e1e3 h7h6 d4d5 c6b8 a1e1 e7d6 f4e5 b8a6 c3c4 a6c5 d3c3 d6e5 f3e5 d7d6 g2g3 e8e7 c3d4 f8e8 g1g2 a7a5 h2h4 c5d7 e5d7 d6d7 e3e7 e8e7 e1e4 e7e4 d4e4 b7b6 h4h5 a5a4 a2a3');
INSERT INTO evaluations VALUES(41119,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(41119,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(41119,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(41119,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(41119,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(41119,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(41119,36,'cp 47','e1g1 c8f5 f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 f5d3 d1d3 d5d7 c1f4 a8e8 a1d1 h7h6 d4d5 c6b8 h2h3 b8a6 f3e5 d7d8 d3g3 e7g5 f4g5 h6g5 c3c4 a6c5 g3a3 d8d6 e5f3 c5e4 a3e3 d6f6 c4c5 a7a6 e3a3 f6c3 a3a4 e4c5 a4e8 f8e8');
INSERT INTO evaluations VALUES(41118,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(41118,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(41118,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41118,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41118,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(41118,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41118,36,'cp 37','c8f5 f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 d5d7 c1f4 e7d6 f4d6 c7d6 d4d5 c6a5 h2h4 a8c8 f3d4 f5d3 d1d3 a5c4 h4h5 c4e5 d3f5 f8d8 h5h6 d7f5 d4f5 c8c3 f5g7 g8f8 a1b1 b7b6 e1e4 e5d3 b1d1 d3e5 g7f5 c3c5 a2a4 a7a6 e4e1 b6b5 f2f4');
INSERT INTO evaluations VALUES(28409,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(28409,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(28409,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(28409,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28409,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28409,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(28409,36,'cp 40','f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 f5d3 d1d3 a8e8 c1f4 d5d7 e1e3 e7f6 d3b5 b7b6 a1e1 e8e3 f2e3 f6e7 e3e4 e7d6 f4g5 f7f6 g5h4 c6b8 b5d3 d7f7 h4f2 b8c6 e1e2 c6e7 f2g3 d6g3 h2g3');
INSERT INTO evaluations VALUES(28410,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(28410,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(28410,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(28410,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(28410,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(28410,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(28410,36,'cp 41','e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 a8e8 c1f4 f5d3 d1d3 d5d7 a1d1 e7f6 f4g5 d7d6 g5f6 d6f6 h2h4 h7h6 g2g3 b7b6 g1g2 f6d6 d3b5 e8e7 e1e7 c6e7 c3c4 d6c6 b5b4 f8e8 d4d5 c6d7 d1e1 e7f5 f3e5 d7c8 b4c3 f5d6 c4c5 b6c5');
INSERT INTO evaluations VALUES(27829,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(27829,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(27829,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(27829,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(27829,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(27829,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(27829,36,'cp 47','c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 a8e8 c1f4 f5d3 d1d3 d5d7 e1e3 h7h6 d4d5 c6b8 a1e1 e7d6 f4e5 d7d8 c3c4 b8d7 e5d4 d7c5 d3a3 b7b6 d4e5 d8d7 e5d6 e8e3 a3e3 c7d6 f3d4 d7g4 h2h3 g4h4 d4f5 h4f6 f5e7 g8h7 e7c6 a7a6');
INSERT INTO evaluations VALUES(42382,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(42382,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(42382,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(42382,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(42382,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(42382,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(42382,36,'cp -26','e8g8 b1c3 d6c5 g5f3 d7d6 d2d3 a7a5 a2a4 c8e6 c4e6 f7e6 e1g1 b8d7 c3e2 d8e8 e2g3 e8g6 d1e2 h7h6 c1d2 f6h5 g3h5 g6h5 d2e3 f8f7 f3d4 h5e2 d4e2 c5e3 f2e3 g8f8 e2c3 c7c6 f1f7 f8f7 h2h3 f7e7 b2b3');
INSERT INTO evaluations VALUES(37151,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(37151,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(37151,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(37151,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(37151,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(37151,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(37151,36,'cp -17','b1c3 d6c5 g5f3 d7d6 d2d3 c7c6 a2a4 f8e8 c4b3 b8d7 e1g1 c5b6 h2h3 d7c5 b3a2 a7a5 c1e3 h7h6 f1e1 c8d7 a2c4 d7e6 c4e6 c5e6 d1d2 d8c7 e3b6 c7b6 c3e2 d6d5 e4d5 f6d5 f3e5 b6b2 d3d4 d5b4');
INSERT INTO evaluations VALUES(37149,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(37149,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(37149,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(37149,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(37149,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(37149,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(37149,36,'cp -32','d6c5 g5f3 d7d6 d2d3 c7c6 a2a4 f8e8 e1g1 h7h6 h2h3 b8d7 c1e3 c5b4 f3h2 d7f8 d1f3 d6d5 e4d5 b4c3 b2c3 c6d5 c4b3 c8e6 h2g4 f6g4 h3g4 a8c8 f1e1 f8g6 a4a5 d5d4 c3d4 e6b3 c2b3 e5d4 e3f4 d8f6 f4g3 e8e1 a1e1 f6f3 g2f3');
INSERT INTO evaluations VALUES(44481,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(44481,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(44481,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(44481,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(44481,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(44481,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(44481,36,'cp 110','e2e4 g7g6 f1e2 f8g7 f2f4 d7d6 g1f3 e8g8 e1g1 b8d7 e4e5 f6d5 c3d5 c6d5 c2c3 d8b6 d1b3 b6b3 a2b3 a7a5 e2d3 d6e5 f4e5 d7b6 c1g5 f8e8 f1e1 c8f5 d3f5 g6f5 g1f2 e7e6 g5f6 g7f8 h2h3 a5a4 b3a4 b6a4');
INSERT INTO evaluations VALUES(56885,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(56885,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(56885,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(56885,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(56885,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(56885,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(56885,36,'cp 70','f4g3 h7h5 h2h4 g5g4 e2e4 d7d6 g3f4 b8c6 b1c3 e6e5 d4e5 d6e5 d1d8 e8d8 e1c1 d8e8 f4e3 c8e6 f1b5 a7a6 b5c6 b7c6 g1e2 g8f6 b2b3 f6d7 c3a4 a6a5 e2g3 f8d6 g3f5 e6f5 e4f5 a8b8 c1b1 h8g8 f2f3 g4f3 g2f3 g8g2 e3g5');
INSERT INTO evaluations VALUES(56886,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(56886,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(56886,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(56886,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(56886,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(56886,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(56886,36,'cp 54','h7h5 h2h4 g5g4 e2e4 d7d6 g3f4 b8c6 b1c3 e6e5 d4e5 d6e5 d1d8 e8d8 e1c1 d8e8 c3b5 c8d7 f4e3 a8c8 f1c4 d7e6 c4d5 g8e7 g1e2 a7a6 b5c3 c8d8 e2g3 f8h6 e3h6 h8h6 h1e1 d8d7 d5c6 e7c6 d1d7 e8d7 e1d1 d7c8 c3d5 b7b6 b2b3 a6a5 d5e3 h6h8 c1b2 c8b7 a2a3 c6e7 e3c4');
INSERT INTO evaluations VALUES(277,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(277,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(277,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(277,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(277,34,'cp 234',NULL);
INSERT INTO evaluations VALUES(277,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(277,36,'cp 241','c3d5 f6d8 e1g1 c5e7 d2d4 d7d6 f3g5 g8h6 c2c3 e5d4 c3d4 e8g8 f2f4 e7f6 d5f6 d8f6 d4d5 f6d4 d1d4 c6d4 f1d1 d4c2 a1b1 h6g4 c4b3 c2b4 c1d2 b4a6 d1e1 a6c5 b3c2 g4f6 b2b4 c5d7 b1c1 h7h6 g5f3 a7a5 c2b3 a5a4 b3c2 c7c6 d5c6 b7c6');
INSERT INTO evaluations VALUES(300,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(300,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(300,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(300,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(300,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(300,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(300,36,'cp 196','f6d8 e1g1 d7d6 c2c3 c5b6 d2d4 e5d4 f3d4 h7h6 d5b6 a7b6 f2f4 c8d7 e4e5 d6d5 d4c6 b7c6 c4d3 g7g6 f4f5 g6f5 d1h5 g8e7 c1g5 h8g8 g5f6 g8g6 d3f5 d7f5 f1f5 d8d7 f6e7 d7e7 a1f1 e8c8 f5f7 e7g5 h5g5 g6g5 f1e1 c6c5 f7f6 d5d4 e5e6 d4c3 b2c3');
INSERT INTO evaluations VALUES(10611,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(10611,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(10611,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(10611,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(10611,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(10611,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(10611,36,'cp 142','c3d5 c5e7 d2d4 d7d6 c2c3 g4f6 h2h3 a7a5 f1e1 e8g8 d5e7 d8e7 c1g5 h7h6 g5h4 g8h7 f3h2 c8d7 h2f1 g7g5 h4g3 a8e8 a2a3 h6h5 f2f3 h5h4 g3f2 f6h5 f1e3 e7f6 c4b5 h5f4 g1h2 e8d8 d4d5 c6e7 b5d7 d8d7 b2b3');
INSERT INTO evaluations VALUES(11639,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(11639,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(11639,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(11639,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(11639,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(11639,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(11639,36,'cp 67','f3g5 d6d5 c3d5 c6d4 d5c3 d4b3 a2b3 h7h6 g5f3 b5b4 c3e2 f6e4 e1g1 f8d6 d2d4 e5d4 d1d4 e4f6 c1f4 a6a5 f4d6 c7d6 d4e3 d8e7 e3e7 e8e7 e2d4 c8b7 d4f5 e7d7 f3d4 g7g6 f5e3 h8c8 e3c4 a8a6 a1a4 c8c5 f1a1 c5d5 c2c3 b4c3 b2c3 f6e4 d4e2 e4c5 a4a5 c5b3');
INSERT INTO evaluations VALUES(18002,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(18002,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(18002,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(18002,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(18002,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(18002,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(18002,36,'cp 151','e5f3 g7g6 d2d4 f8g7 b1c3 h7h6 c1e3 a8b8 b2b3 g8e7 h2h4 e8g8 d1d2 g8h7 e1c1 c8g4 d1g1 a6a5 f3h2 g4d7 f2f4 f7f5 e4e5 a5a4 h2f3 a4b3 a2b3 d8e8 d2d3 d6e5 f3e5 e7d5 h4h5 g6h5 c3d5 c6d5 e5d7');
INSERT INTO evaluations VALUES(18021,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(18021,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(18021,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(18021,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(18021,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(18021,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(18021,36,'cp 169','g7g6 d2d4 f8g7 b1c3 h7h6 c1e3 g8f6 d1d2 d8e7 e1c1 f6e4 c3e4 e7e4 d1e1 e8f8 e3h6 g7h6 e1e4 h6d2 c1d2 c8e6 e4e3 a6a5 a2a4 e6d5 h2h3 a8b8 b2b3 b8b4 d2c3 h8h5 h3h4 d5f3 e3f3 h5d5 f3d3 c6c5 d4c5 d5c5 c3d2');
INSERT INTO evaluations VALUES(21376,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(21376,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(21376,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21376,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21376,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(21376,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(21376,36,'cp 26','d5c4 e2e3 c6a5 f1c4 a5c4 d1a4 c8d7 a4c4 c7c5 b1c3 c5d4 f3d4 a8c8 c4b3 d8b6 d4b5 f8c5 e1g1 g8e7 b5d6 c5d6 b3b6 a7b6 f4d6 f7f6 f1d1 d7c6 f2f3 h7h5 e3e4 e8f7 d1d2 h5h4 a1d1 h8d8 d6b4 d8d2 d1d2 e7g6 g1f2 g6f4 g2g3 f4g6');
INSERT INTO evaluations VALUES(21495,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(21495,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(21495,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(21495,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(21495,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(21495,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(21495,36,'cp 37','e2e3 c6a5 b1d2 g8f6 d2c4 a5c4 f1c4 f8b4 f3d2 e8g8 e1g1 b7b6 d1f3 a8b8 f3e2 c8b7 a1c1 b4d6 f4g3 b8c8 f1d1 d6g3 h2g3 h7h6 b2b3 d8d7 d2f3 d7e7 c4a6 c7c5 a6b7 e7b7 d4c5 c8c5 c1c5 b6c5 e2c2 b7c7 f3d2 f8d8 d1c1 c7b7');
INSERT INTO evaluations VALUES(25006,30,'cp 389',NULL);
INSERT INTO evaluations VALUES(25006,31,'cp 366',NULL);
INSERT INTO evaluations VALUES(25006,32,'cp 402',NULL);
INSERT INTO evaluations VALUES(25006,33,'cp 408',NULL);
INSERT INTO evaluations VALUES(25006,34,'cp 377',NULL);
INSERT INTO evaluations VALUES(25006,35,'cp 413',NULL);
INSERT INTO evaluations VALUES(25006,36,'cp 463','d4d5 c6b8 e4e5 f6d7 d1b3 d7b6 a2a4 c7c6 f3g5 e7e6 h2h3 g4f5 g2g4 f5c2 b3c2 c6d5 f1b5 b8c6 a4a5 b6d7 e1g1 a7a6 b5c6 b7c6 c3a4 c6c5 a4b6 f8e7 b6a8 d8a8 c1f4 a8b7 a1b1 e7g5 f4g5 h7h6 g5f4 e8g8 f4g3 b7c6 c2e2 c5c4 f2f4 f8b8 f4f5');
INSERT INTO evaluations VALUES(24795,30,'cp 451',NULL);
INSERT INTO evaluations VALUES(24795,31,'cp 448',NULL);
INSERT INTO evaluations VALUES(24795,32,'cp 459',NULL);
INSERT INTO evaluations VALUES(24795,33,'cp 436',NULL);
INSERT INTO evaluations VALUES(24795,34,'cp 435',NULL);
INSERT INTO evaluations VALUES(24795,35,'cp 444',NULL);
INSERT INTO evaluations VALUES(24795,36,'cp 440','c6b8 e4e5 f6d7 d1b3 d7b6 a2a4 e7e6 a4a5 g4f3 a5b6 f3d5 c3d5 c7b6 f1b5 b8c6 d5b4 d8c7 e1g1 f8e7 b4c6 b7c6 b3f3 a8c8 b5e2 e8g8 f3g3 g8h8 c1e3 a7a5 a1c1 b6b5 e2b5 c6b5 c1c7 c8c7 e3b6 c7d7 b6a5 f8c8 g3f3 h8g8 a5b6 h7h6 b6e3 d7d5 g2g3 c8c4 f1d1 d5d1 f3d1 e7c5 d1d8 g8h7 d8d3 h7g8 b2b3');
INSERT INTO evaluations VALUES(26201,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26201,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26201,36,'cp 0','c5d4 d1d4 g8f6 e1g1 e7e6 c2c4 d8b6 b1c3 b6d4 f3d4 a7a6 b2b3 h7h6 f2f4 d7c5 f1e1 c5d3 e1d1 d3c5');
INSERT INTO evaluations VALUES(29606,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(29606,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(29606,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(29606,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(29606,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(29606,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(29606,36,'cp 35','c2c4 e7e6 e1g1 c7c6 d1b3 d8b6 d2d3 b8d7 c4d5 e6d5 b1c3 b6b3 a2b3 f8e7 c1e3 a7a6 f3d4 f5g6 d4c2 f6g4 e3d4 d7c5 b3b4 c5b3 d4g7 b3a1 f1a1 h8g8 g7d4 e8c8 d4b6 d8e8 f2f3 g4f6 g2h3 f6d7 g1f2 f7f5 b6d4 c8c7 f3f4');
INSERT INTO evaluations VALUES(32274,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(32274,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(32274,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(32274,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(32274,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(32274,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(32274,36,'cp 11','g8f6 g1f3 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 b8c6 e1g1 e8g8 b5c6 b7c6 a1c1 a8c8 d1b3 c6c5 d4c5 d6c5 e3e4 c5c4 b3b4 g6h5 f1d1 d8c7 b4a5 h5g6 a5c3 f8d8 b2b3 f7f6 b3c4');
INSERT INTO evaluations VALUES(35539,30,'mate 11',NULL);
INSERT INTO evaluations VALUES(35539,31,'mate 11',NULL);
INSERT INTO evaluations VALUES(35539,32,'mate 11',NULL);
INSERT INTO evaluations VALUES(35539,33,'mate 11',NULL);
INSERT INTO evaluations VALUES(35539,34,'mate 11',NULL);
INSERT INTO evaluations VALUES(35539,35,'mate 11',NULL);
INSERT INTO evaluations VALUES(35539,36,'mate 11','e5f5 g6h6 h2h4 d8e7 d2d4 g7g5 c1g5 h6g7 g5e7 f8e7 f5f7 g7h6 g2g4 d7d6 g4g5 e7g5 h4g5 h6g5 h1h5 g5g4 c4e2');
INSERT INTO evaluations VALUES(35540,30,'mate 10',NULL);
INSERT INTO evaluations VALUES(35540,31,'mate 10',NULL);
INSERT INTO evaluations VALUES(35540,32,'mate 10',NULL);
INSERT INTO evaluations VALUES(35540,33,'mate 10',NULL);
INSERT INTO evaluations VALUES(35540,34,'mate 10',NULL);
INSERT INTO evaluations VALUES(35540,35,'mate 10',NULL);
INSERT INTO evaluations VALUES(35540,36,'mate 10','g6h6 h2h4 d8e7 d2d4 g7g5 c1g5 h6g7 g5e7 f8e7 f5f7 g7h6 g2g4 d7d6 g4g5 e7g5 h4g5 h6g5 h1h5 g5g4 c4e2');
INSERT INTO evaluations VALUES(35542,30,'mate 10',NULL);
INSERT INTO evaluations VALUES(35542,31,'mate 10',NULL);
INSERT INTO evaluations VALUES(35542,32,'mate 10',NULL);
INSERT INTO evaluations VALUES(35542,33,'mate 10',NULL);
INSERT INTO evaluations VALUES(35542,34,'mate 10',NULL);
INSERT INTO evaluations VALUES(35542,35,'mate 10',NULL);
INSERT INTO evaluations VALUES(35542,36,'mate 10','h2h4 d8e7 d2d4 g7g5 c1g5 h6g7 g5e7 f8e7 f5f7 g7h6 g2g4 d7d6 g4g5 e7g5 h4g5 h6g5 h1h5 g5g4 c4e2');
INSERT INTO evaluations VALUES(46692,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(46692,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(46692,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(46692,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(46692,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(46692,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(46692,36,'cp 54','e2e4 f8g7 c1e3 e8g8 d1d2 a7a6 e3h6 b7b5 f1d3 b8c6 c3d5 g7h6 d2h6 f6g4 h6d2 e7e6 d5e3 g4e3 f2e3 c8b7 h2h4 c6b8 e4e5 d6e5 d4e5 b8d7 e3e4 f7f5 e5f6 d8f6 e1c1 d7c5 h4h5 b7e4 d3e4 c5e4 d2e1 f6f4 c1b1 a8d8 h5g6 d8d1 e1d1 h7g6');
INSERT INTO evaluations VALUES(56748,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(56748,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(56748,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(56748,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(56748,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(56748,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(56748,36,'cp 21','b8c6 e4f5 g8e7 g2g4 h7h5 g4h5 e7f5 h1g1 f8e7 c3d5 c8e6 f1h3 h8f8 d5e7 d8e7 c2c3 f5d4 f3d4 e5d4 h3e6 e7e6 d1e2 e6e2 e1e2 e8f7 e2f1 f8h8 h5h6 g7h6 c1d2 d4c3 d2c3 h8g8 g1g3 g8g3 h2g3 a8e8 f1g2 e8e2 a1h1 c6e5 h1h6 e5d3 h6h7 f7e6');
INSERT INTO evaluations VALUES(22327,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(22327,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(22327,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(22327,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(22327,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(22327,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(22327,36,'cp 31','e4f5 g8e7 g2g4 h7h5 g4h5 e7f5 h1g1 f8e7 c3d5 c8e6 f1h3 h8f8 d5e7 d8e7 c2c3 f5d4 f3d4 e5d4 h3e6 e7e6 d1e2 e6e2 e1e2 e8f7 e2f1 f8h8 g1g5 h8h7 c3d4 c6d4 c1e3 d4e6 g5f5 f7e7 e3g5 e6g5 f5g5 a8h8 f1e2 e7d7 d3d4 d7c6 a1c1 c6b6 c1c4 a7a5 h2h4 h7h5 g5g7 h5h4 c4c7');
INSERT INTO evaluations VALUES(4590,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(4590,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(4590,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(4590,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(4590,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(4590,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(4590,36,'cp 39','d2d4 c6d4 f3d4 e5d4 e4e5 f6e8 d1d4 d7d6 b1c3 c7c6 e5d6 e8d6 c4b3 e7f6 d4f4 c8e6 c1e3 e6b3 a2b3 f8e8 g2g3 d6b5 c3b5 c6b5 e1d1 d8c8 c2c3 h7h6 e3d4 e8e6 d4f6 e6f6 f4e4 a7a6 b3b4 c8c7 g1g2 f6e6 e4f3');
INSERT INTO evaluations VALUES(4599,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(4599,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(4599,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(4599,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(4599,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(4599,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(4599,36,'cp 32','e5d4 e4e5 f6e8 f3d4 c6d4 d1d4 d7d6 b1c3 c7c6 e5d6 e8d6 c4b3 e7f6 d4f4 c8e6 c1e3 e6b3 a2b3 d6b5 c3b5 c6b5 e1d1 d8e7 c2c3 f8d8 g2g3 a7a6 b3b4 d8d7 d1d7 e7d7 f4f3 a8d8 g1g2 d7d5 a1a5');
INSERT INTO evaluations VALUES(4609,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(4609,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(4609,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(4609,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(4609,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(4609,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(4609,36,'cp 38','e4e5 f6e8 f3d4 c6d4 d1d4 c7c6 b1c3 d7d5 e5d6 e8d6 c4b3 e7f6 d4f4 c8e6 c1e3 e6b3 a2b3 d6b5 c3b5 c6b5 e1d1 d8e7 c2c3 a7a6 b3b4 f8d8 g2g3 d8d7 e3b6 a8e8 g1g2 d7d1 a1d1 e7e6 d1d6 e6c8 f4f3 e8e6 d6e6 f7e6');
INSERT INTO evaluations VALUES(6715,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6715,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(6715,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6715,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6715,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(6715,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(6715,36,'cp 51','e3d4 c6d4 d1d4 d8f6 e4e5 f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 a2a3 c7c6 c2c4 d7d6 e1c1 d8e7 d1e1 e7d8 f1d3 a8a5 e1e4 c8f5 e4d4 d8d7 h1d1 f5e6 d4f4 d7e7 d1e1 a5e5 f4e4 h7h5 h2h4 e7d7 c1d2');
INSERT INTO evaluations VALUES(6706,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(6706,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(6706,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(6706,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(6706,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(6706,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(6706,36,'cp 33','c6d4 d1d4 d8f6 e4e5 f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 f2f3 d7d5 a2a3 d8e7 e1f2 f6f5 a1e1 e7f6 f3f4 c8d7 e1e5 c7c6 f1d3 a8e8 h1e1 e8e5 e1e5 d7e6 f2e3 h8g8 g2g3 g8a8 g3g4 f5g4 e5h5 a8h8 h5h6 f6e7 h6h7 h8h7 d3h7');
INSERT INTO evaluations VALUES(6905,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(6905,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(6905,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(6905,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(6905,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(6905,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(6905,36,'cp 41','d1d4 d8f6 e4e5 f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 a2a3 h8e8 e1d2 d7d6 f2f3 h7h5 a1e1 e8h8 d2e3 c8f5 c2c3 c7c6 e3f4 f5c2 h2h4 a8a4 f4e3 d6d5 f1e2 d8d7 g2g4 a4a8 g4g5 f6g5 h4g5 c2g6 f3f4 a8e8 e3d2 e8e4 e2d3 e4e1 h1e1');
INSERT INTO evaluations VALUES(14675,30,'cp 252',NULL);
INSERT INTO evaluations VALUES(14675,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(14675,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(14675,33,'cp 245',NULL);
INSERT INTO evaluations VALUES(14675,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(14675,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(14675,36,'cp 257','e4e5 c7c5 d4d3 d6e5 c4f7 e8e7 d3b3 d8b6 f7g6 b6b3 a2b3 c8d7 f1e1 e5e4 b1d2 d7e8 g6f5 e8d7 f5e4 f6e4 e1e4 e7f7 d2c4 d7e6 c1d2 e6c4 e4f4 f7g6 f4c4 f8e7 d2c3 h8e8 g1f1 e7f8 a1d1 e8d8 d1d8 a8d8 f1e2 a7a6 h2h4 d8e8 e2f3 h6h5 g2g3 b7b5 c4f4 e8d8');
INSERT INTO evaluations VALUES(14668,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(14668,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(14668,32,'cp 264',NULL);
INSERT INTO evaluations VALUES(14668,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(14668,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(14668,35,'cp 256',NULL);
INSERT INTO evaluations VALUES(14668,36,'cp 249','c7c5 d4d3 d6e5 c4f7 e8e7 d3b3 d8b6 f7g6 b6b3 a2b3 c8d7 b1d2 d7e8 g6d3 f6d5 c2c3 e8c6 d2c4 d5f4 c1f4 e5f4 c4e5 e7f6 e5c6 b7c6 a1a6 f8d6 a6c6 h8d8 d3c4 a8b8 c6a6 d8d7 f1d1 b8b6 a6a5 d7c7 g2g3 h6h5 g1g2 g7g6 d1a1 g6g5 a5a7 c7a7 a1a7 f4g3 h2g3 g5g4 f2f4 g4f3 g2f3 f6f5 a7h7 f5g5 h7d7 g5f6 c4d3 d6e7');
INSERT INTO evaluations VALUES(18690,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(18690,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(18690,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(18690,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(18690,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(18690,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(18690,36,'cp 13','b5c6 b7c6 d2d3 d7d5 c1d2 g8f6 g1f3 f8e7 b2b3 f6d7 d1e2 e8g8 e1g1 a7a5 h2h3 c8a6 f1e1 f8e8 a1d1 d8c7 e4e5 a6b5 d2g5 e7g5 f3g5 h7h6 a2a4 b5a6 g5f3 d7f8 c3b1 f8g6 e2e3 c5c4 d3c4 d5c4 b1c3 c4b3 c2b3');
INSERT INTO evaluations VALUES(18675,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(18675,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(18675,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(18675,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18675,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(18675,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(18675,36,'cp 16','b7c6 d2d3 d7d5 g1f3 g8e7 h2h4 h7h5 c3e2 c8a6 b2b3 e7g6 e2f4 g6f4 c1f4 d5e4 d3e4 c5c4 d1d2 f7f6 e4e5 d8d2 f4d2 f8e7 f3d4 e8d7 d4f3 d7e8 d2c3 e8f7 f3d2 c4b3 a2b3 a6b5 h1h3 h8d8 d2c4 d8d7 h3d3');
INSERT INTO evaluations VALUES(18377,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(18377,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(18377,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(18377,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(18377,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(18377,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(18377,36,'cp 21','d2d3 d7d5 g1f3 g8e7 h2h4 h7h5 c3e2 c8a6 b2b3 d5e4 d3e4 d8d1 e1d1 a8d8 d1e1 e7g6 c1e3 c5c4 e3a7 f8e7 a7d4 e8g8 d4c3 c6c5 f3e5 g6e5 c3e5 d8a8 f2f3 c4b3 a2b3 c5c4 b3b4 e7b4 e1f2 f8d8 e5d4 b4e7 g2g3 a6b5 h1b1 b5a4 b1b7 e7d6 a1a2 e6e5 d4e3 f7f6 e2c3 a4e8 b7a7 a8a7 a2a7 d6b4');
INSERT INTO evaluations VALUES(20953,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(20953,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(20953,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(20953,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(20953,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(20953,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(20953,36,'cp 98','c6b8 e2e4 e7e5 g1f3 b8d7 f1c4 f8d6 e1g1 g8e7 a2a4 a7a5 c1e3 h7h6 f3d2 e8g8 a1c1 d7f6 c4e2 d6b4 e2f3 c7c6 d5c6 e7c6 d2c4 c8e6 c3d5 a8a6 e3b6 a6b6 c4b6 c6d4 d1d3 f6d5 b6d5 d8h4 c1c7');
INSERT INTO evaluations VALUES(30129,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(30129,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30129,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(30129,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30129,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(30129,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(30129,36,'cp 9','f8e7 c2c4 c7c6 b1c3 h7h6 d1b3 d8b6 c4c5 b6a6 f3e5 b7b6 e2e4 d5e4 f1e1 b6c5 c3e4 f5e4 g2e4 e8g8 e4d3 a6b6 e5g6 f8e8 g6e7 e8e7 d4c5 b6c5 c1e3 c5d6 e3f4 d6d8 a1c1 d8b6 b3c2 f6d5 f4d6 e7d7 d6a3 b8a6 d3a6 b6a6 c2c6 a6b7 c6b7 d7b7 c1c4');
INSERT INTO evaluations VALUES(31829,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(31829,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(31829,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(31829,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(31829,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31829,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31829,36,'cp 0','e7e6 e2e3 g8f6 d1b3 d8c8 g1f3 c7c5 c3c4 a7a6 b1d2 b8d7 a1c1 b7b5 c4d5 f6d5 f4g3 d5f6 f1e2 c5c4 d2c4 b5c4 c1c4 c8d8 g3c7 d8c8 c7e5 c8d8');
INSERT INTO evaluations VALUES(32292,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32292,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32292,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32292,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32292,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32292,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32292,36,'cp 0','e2e3 g8f6 d1b3 d8c8 g1f3 c7c5 c3c4 a7a6 b1d2 b8d7 a1c1 b7b5 c4d5 f6d5 f4g3 d5f6 a2a4 c5c4 f1c4 b5c4 c1c4 c8d8 g3c7 d8c8 c7e5 c8d8');
INSERT INTO evaluations VALUES(32044,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(32044,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(32044,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(32044,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(32044,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(32044,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(32044,36,'cp 92','d2d4 b8d7 d1d3 g8f6 d3b3 g4h5 b1d2 a8b8 a2a4 f8e7 d4e5 d7e5 f3e5 d6e5 b3g3 e8g8 f2f3 f6d7 f1d1 g8h8 d2f1 h5g6 f1e3 e7d6 c4f1 d8e7 e3c4 d6c5 c1e3 f8d8 e3c5 e7c5 g1h1 c5e7 d1d5 c7c6 d5a5 a7a6 c4e5 d7e5 g3e5 e7b4 a5c5 f7f6 e5c3 b4c3 c5c3 d8d2');
INSERT INTO evaluations VALUES(35512,30,'cp -449',NULL);
INSERT INTO evaluations VALUES(35512,31,'cp -450',NULL);
INSERT INTO evaluations VALUES(35512,32,'cp -440',NULL);
INSERT INTO evaluations VALUES(35512,33,'cp -455',NULL);
INSERT INTO evaluations VALUES(35512,34,'cp -451',NULL);
INSERT INTO evaluations VALUES(35512,35,'cp -439',NULL);
INSERT INTO evaluations VALUES(35512,36,'cp -451','f3e5 f7g8 e1g1 b8c6 d2d3 e4f6 f2f4 d8e8 b1c3 d7d5 c1d2 f8c5 g1h1 c6e7 d1f3 h7h5 d2e3 c5d6 a1e1 c7c6 e3f2 e7f5 e5c4 d6e7 c4e3 h8h6 e3f5 c8f5 h2h3 e8d7 f2h4 a8e8 h4g5 h6h7 e1e5 e7d6 f1e1 d6e5 f4e5 f6g4 d3d4 g8h8');
INSERT INTO evaluations VALUES(35509,30,'cp -481',NULL);
INSERT INTO evaluations VALUES(35509,31,'cp -478',NULL);
INSERT INTO evaluations VALUES(35509,32,'cp -471',NULL);
INSERT INTO evaluations VALUES(35509,33,'cp -478',NULL);
INSERT INTO evaluations VALUES(35509,34,'cp -474',NULL);
INSERT INTO evaluations VALUES(35509,35,'cp -471',NULL);
INSERT INTO evaluations VALUES(35509,36,'cp -472','f7g8 d2d3 e4f6 e1g1 b8c6 d3d4 d7d6 e5c6 b7c6 f1e1 c8f5 b1d2 d8d7 c2c4 h7h6 b2b4 a8e8 e1e8 d7e8 a2a4 g8h7 d2f3 g7g5 c1b2 f8g7 d4d5 c6d5 c4d5 h8f8 a1c1 e8f7 b2f6 f7f6 c1c7 h7h8 c7a7 f6b2 b4b5');
INSERT INTO evaluations VALUES(35518,30,'cp -453',NULL);
INSERT INTO evaluations VALUES(35518,31,'cp -471',NULL);
INSERT INTO evaluations VALUES(35518,32,'cp -450',NULL);
INSERT INTO evaluations VALUES(35518,33,'cp -440',NULL);
INSERT INTO evaluations VALUES(35518,34,'cp -451',NULL);
INSERT INTO evaluations VALUES(35518,35,'cp -440',NULL);
INSERT INTO evaluations VALUES(35518,36,'cp -461','d2d3 e4f6 e1g1 d7d6 e5f3 b8c6 c2c4 h7h6 d3d4 c8g4 h2h3 g4h5 d1b3 h5f3 b3f3 c6d4 f3d3 d4c6 b1c3 c6e5 d3c2 d8e8 c1e3 e8f7 e3d4 a8e8 a1e1 f7c4 d4e5 d6e5 b2b3 c4f7 c3e4 f6d5 f2f3 c7c6 c2e2 d5f4');
INSERT INTO evaluations VALUES(37316,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(37316,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(37316,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(37316,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(37316,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(37316,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(37316,36,'cp -12','b1c3 f8e8 e5f4 e7b4 g1e2 b4c3 e2c3 d7d5 c4b5 c7c6 b5e2 f6e4 d2d3 e4c3 b2c3 c6c5 c1e3 d8a5 e1g1 a5c3 e2f3 c8e6 a1b1 b7b6 b1b3 c3f6 f4f6 g7f6 c2c4 e8d8 c4d5 e6d5 f3d5 d8d5 d3d4 c5d4 e3h6 d4d3 f1d1 g8h8 h2h3 d3d2 d1d2 d5d2 h6d2 b8d7 d2c3 a8e8');
INSERT INTO evaluations VALUES(39981,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(39981,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(39981,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(39981,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(39981,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(39981,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(39981,36,'cp 92','e2e4 e8g8 f1e2 d7d6 e1g1 b8a6 c1e3 e7e5 d4d5 f6g4 e3g5 f7f6 g5h4 c6c5 f3e1 g4h6 g2g4 h6f7 e1g2 f7g5 g1h1 g7h6 f2f4 e5f4 g2f4 d8e7 h4g3 c8d7 d1b3 a8e8 a1e1 b7b6 f4g2 g5f7 b3d1 f7e5 b2b3');
INSERT INTO evaluations VALUES(39987,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(39987,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(39987,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(39987,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(39987,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(39987,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(39987,36,'cp 87','e8g8 f1e2 d7d6 e1g1 e7e5 d4d5 a7a5 c1g5 b8a6 f3d2 a6c5 d2b3 c5b3 a2b3 h7h6 g5e3 c6c5 g1h1 b7b6 d1d2 g8h7 f2f3 c8d7 g2g3 f6g8 f1f2 d8e7 a1g1 a8d8 h2h4 f7f5 e2d3 h7h8 e4f5 d7f5 d3c2 f5c2');
INSERT INTO evaluations VALUES(35695,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(35695,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(35695,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(35695,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(35695,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(35695,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(35695,36,'cp 74','f1e2 d7d6 e1g1 e7e5 d4d5 b8a6 f3d2 f6e8 d2b3 c6d5 c4d5 f7f5 a2a4 c8d7 c1e3 f5f4 e3d2 a6c5 b3c5 d6c5 f2f3 e8d6 d2e1 g7f6 c3b5 d7b5 e2b5 a7a6 b5e2 b7b6 g1h1 h7h5 d1c2 d8e7 a1d1 g6g5');
INSERT INTO evaluations VALUES(42998,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(42998,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(42998,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(42998,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(42998,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(42998,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(42998,36,'cp 25','c2c4 c7c5 d4c5 e7c5 g1f3 c5b4 b1d2 e8g8 f1e2 b7b6 e1g1 c8b7 a1c1 f6h5 f4g5 b4e7 g5e7 d8e7 d2b1 h5f6 b1c3 f8d8 d1d4 b8a6 f1d1 a6c5 d4h4 c5e4 c3e4 b7e4 a2a3 e4b7 d1d2');
INSERT INTO evaluations VALUES(45937,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(45937,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(45937,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(45937,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(45937,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(45937,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(45937,36,'cp 137','h7h5 h2h4 g7g6 g2g3 b8a6 f2f3 a6c7 g1h3 c8d7 c1g5 f8h6 g5h6 h8h6 h3f2 h6h8 f1e2 a7a6 a2a4 e8f8 a4a5 c7e8 d1d2 a8b8 f3f4 d8e7 f2d3 e5f4 d2f4 f8g7 e1g1 d7h3 f1e1 h8f8 d3f2 h3g4 f2g4 h5g4 e2g4 f6g4 f4g4 e8f6 g4g5');
INSERT INTO evaluations VALUES(50678,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50678,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(50678,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(50678,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50678,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(50678,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(50678,36,'cp 17','g8f6 e1g1 g7g6 b1c3 f8g7 a2a4 e8g8 h2h3 b8c6 c1e3 c8d7 f1e1 a8c8 c4f1 c6e5 d1d2 e5c4 f1c4 c8c4 a4a5 e7e5 d4e2 f6e4 d2d3 d6d5 e3b6 d8c8 c3d5 d7e6 a1d1 e6d5 d3d5');
INSERT INTO evaluations VALUES(43399,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(43399,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(43399,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43399,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43399,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43399,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43399,36,'cp 20','e1g1 g7g6 f1e1 f8g7 b1c3 e8g8 c4b3 b8c6 d4c6 b7c6 c1f4 f6d7 d1d2 d7c5 a1d1 a8b8 h2h3 c8e6 f2f3 c5b3 c2b3 c6c5 f4e3 d8c8 e3h6 g7d4 g1h1 f8e8 c3e2 d4g7 h6g7 g8g7 e2c3 f7f6 d2f2 a6a5');
INSERT INTO evaluations VALUES(53467,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(53467,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(53467,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(53467,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(53467,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(53467,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(53467,36,'cp 37','d2d4 g8f6 g1f3 e7e6 f1c4 f8e7 e1g1 e8g8 f1e1 a7a6 d4d5 e6d5 c4d5 b8c6 d5c6 b7c6 c1g5 h7h6 g5h4 c8e6 f3d4 e6d7 d1a4 f6d5 h4e7 d5e7 d4b3 e7d5 c3d5 c6d5 a4d4 a8c8 h2h3 f8e8 d4d5 d7e6 d5d8');
INSERT INTO evaluations VALUES(53462,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(53462,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53462,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(53462,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(53462,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(53462,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(53462,36,'cp 24','g8f6 g1f3 e7e6 f1c4 f8e7 e1g1 e8g8 c1f4 b8c6 a2a3 a7a6 f1e1 b7b5 d4d5 e6d5 c4d5 c8b7 d5a2 d8d1 a1d1 a8e8 b2b4 h7h6 c3d5 f6d5 a2d5 e7f6 d5f7 g8f7 d1d7 f7g6 d7b7 c6d4 g1f1 d4f3 e1e8 f8e8 g2f3');
INSERT INTO evaluations VALUES(45494,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(45494,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45494,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45494,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(45494,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45494,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45494,36,'cp 28','g1f3 e7e6 f1c4 f8e7 e1g1 e8g8 a2a3 c8d7 c1e3 d8c7 d1e2 d7c6 f1d1 b8d7 d4d5 e6d5 c3d5 c6d5 c4d5 f6d5 d1d5 d7f6 d5d2 b7b6 a1d1 a8d8 f3d4 d8d5 d4b5 d5d2 d1d2 c7b7 b5c3 e7c5 e2f3 b7f3 g2f3 c5e3 f2e3');
INSERT INTO evaluations VALUES(45493,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(45493,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(45493,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(45493,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(45493,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(45493,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(45493,36,'cp 30','e7e6 f1c4 f8e7 e1g1 e8g8 a2a3 c8d7 c1e3 d8c7 d1e2 d7c6 f3e5 b8d7 e3f4 d7e5 d4e5 f6d5 c3d5 c6d5 c4d5 e6d5 f1d1 a8d8 d1d3 c7b6 b2b4 d8d7 a1d1 b6e6 g2g3 b7b6 e2e3 f8c8 h2h4 c8c4 d3d4 c4d4 e3d4 f7f6 d4e3 d7d8 e5f6 e6e3 f4e3');
INSERT INTO evaluations VALUES(46380,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(46380,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46380,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(46380,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(46380,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(46380,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(46380,36,'cp 33','f1c4 f8e7 e1g1 e8g8 f1e1 b8c6 a2a3 b7b6 c1f4 c8b7 c4a2 h7h6 d1d2 e7d6 f4h6 c6d4 d2d4 b7f3 d4e3 f3b7 e3g5 f6e8 g5g4 d6c5 a1d1 d8f6 h6f4 a8d8 g4g3 d8d1 e1d1 e6e5 f4e3 c5e3 g3e3 e8d6 a2d5 f8d8 e3e2');
INSERT INTO evaluations VALUES(46373,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(46373,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46373,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46373,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(46373,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(46373,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(46373,36,'cp 37','f8e7 e1g1 b8c6 f1e1 e8g8 a2a3 b7b6 c4a2 c8b7 c1f4 f8e8 d1d3 e7d6 f4g5 h7h6 g5h4 g7g5 h4g3 d6g3 h2g3 g5g4 f3h2 d8d4 d3e2 d4c5 h2g4 f6g4 e2g4 c5g5 g4h3 a8d8 c3e4 g5g6 h3h4 g8g7 a1c1 c6d4 c1c7 b7e4 h4e4 g6e4 e1e4');
INSERT INTO evaluations VALUES(40242,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(40242,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40242,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(40242,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(40242,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(40242,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(40242,36,'cp 36','e1g1 e8g8 f1e1 b8c6 c1f4 b7b6 a2a3 c8b7 c4a2 f8e8 d1d2 e7f8 d2d3 f8d6 f4g5 h7h6 g5h4 g7g5 h4g3 d6g3 h2g3 g5g4 f3h2 d8d4 d3e2 d4c5 h2g4 f6g4 e2g4 c5g5 g4g5 h6g5 c3e4 e8d8 a1d1 d8d4 e4g5 a8d8 d1c1 d4d2 a2b1 d8d7 b1e4 c6d4 e4b7 d7b7');
INSERT INTO evaluations VALUES(54002,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(54002,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(54002,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(54002,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(54002,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(54002,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(54002,36,'cp 108','d2d4 g8f6 f1d3 f8e7 e1g1 c8d7 d1e2 c5d4 c3d4 b8c6 b1c3 d8b6 c1e3 d6d5 f3e5 d5e4 c3e4 b6d8 e4f6 e7f6 d3e4 e8g8 f1d1 g7g6 a1c1 a8c8 b2b3 d7e8 h2h3 f6g7 c1c5 c8c7 e4f3 c6e7 e3f4');
INSERT INTO evaluations VALUES(54609,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(54609,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(54609,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(54609,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(54609,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(54609,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(54609,36,'cp -22','e5e4 f3d4 f8c5 d4e2 d8h4 g2g3 h4e7 b1c3 g8f6 c3d5 f6d5 b3d5 e8g8 d2d4 e4d3 d1d3 b8a6 d3e4 e7e4 d5e4 f8e8 e4d3 a6b4 c1e3 c5e3 f2e3 e8e3 e1d2 e3e7 e2d4 c8g4 a1e1 e7d7 c2c3 b4d3 d2d3 h7h5 h1f1 a8c8 d3d2 f7f6 f1f4 b7b6 e1e3 g8f7 h2h4 b6b5 a2a3 a7a6');
INSERT INTO evaluations VALUES(54592,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(54592,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(54592,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(54592,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(54592,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(54592,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(54592,36,'cp -22','f3d4 d8g5 b3a4 c8d7 a4d7 b8d7 e1g1 g5e5 c2c3 d7c5 d4f3 e5c7 d2d4 c5e6 f3e5 e8c8 f2f4 f7f6 f4f5 f6e5 f5e6 c7b6 g1h1 b6e6 d4e5 g8h6 d1d4 c8b8 c1e3 b7b6 e3h6 e6h6 b1d2 f8c5 d4a4 h6e6 d2b3 d8c8 a1d1 h8d8 b3d4 c5d4 c3d4 b8b7 a4b3 d8e8 h1g1 e8e7 d1e1 c8c4');
INSERT INTO evaluations VALUES(54582,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(54582,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(54582,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(54582,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(54582,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(54582,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(54582,36,'cp -35','d8g5 b3a4 c8d7 a4d7 b8d7 e1g1 g5e5 c2c3 d7c5 d4f3 e5c7 d2d4 c5e6 f3e5 e8c8 f2f4 f7f6 f4f5 f6e5 f5e6 c7b6 d1g4 c8b8 c1e3 g8f6 g4g3 b6e6 g3e5 e6e5 d4e5 f6d7 e3d4 d8e8 b1a3 f8a3 b2a3 d7e5 f1f5 e5d3 a1d1 e8e7 f5d5 h8e8 a3a4');
INSERT INTO evaluations VALUES(54899,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(54899,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54899,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54899,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(54899,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54899,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(54899,36,'cp 11','h4g3 g8f6 e2e3 f8g7 b1d2 e8g8 f1d3 c7c5 c2c3 b8c6 g1e2 f6h5 e1g1 c5d4 e3d4 d8b6 d1b3 h5g3 e2g3 b6b3 d2b3 e7e5 d4e5 g7e5 f1e1 e5g7 d3f5 d5d4 c3d4 c6d4 b3d4 g7d4 e1e2 f8d8 f5c8 a8c8 g3f5 g8f8 f5h6');
INSERT INTO evaluations VALUES(57251,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(57251,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(57251,32,'cp -136',NULL);
INSERT INTO evaluations VALUES(57251,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(57251,34,'cp -133',NULL);
INSERT INTO evaluations VALUES(57251,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(57251,36,'cp -134','d7d5 c2c3 g8f6 d2d4 b8c6 d4e5 c6e5 c1f4 e5g4 g1h3 f8d6 g2g3 h7h6 f1g2 e8g8 e1g1 g4e5 b1d2 c8g4 f4e5 d6e5 d2f3 e5d6 e2e3 d8d7 h3f4 c7c6 d1c2 f8e8 c3c4 g4f3 g2f3 d5d4 c4c5 d6f4 e3f4 d4d3 c2c4 a8d8 a1d1');
INSERT INTO evaluations VALUES(56502,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(56502,31,'cp -139',NULL);
INSERT INTO evaluations VALUES(56502,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(56502,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(56502,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(56502,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(56502,36,'cp -136','c2c3 b8c6 g2g3 g8f6 d2d4 f8d6 d4e5 c6e5 g1f3 e8g8 f3e5 d6e5 c1g5 c8g4 f1g2 h7h6 g5f6 d8f6 e1g1 a8d8 d1c2 f8e8 b1d2 c7c6 f1e1 e5c7 e2e3 e8e7 d2f1 a7a5 c2b3 f6g6 b3b7 d8b8 b7a7 b8b2 a7d4 c7b6 d4f4');
INSERT INTO evaluations VALUES(5001,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(5001,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(5001,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(5001,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(5001,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(5001,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(5001,36,'cp 40','e1g1 g8f6 d1d6 d8e7 c1f4 e7d6 f4d6 g7f8 d6f8 h8f8 e4e5 f6d5 c3e4 e8e7 f1d1 a7a5 g2g3 f7f5 e4c5 f8d8 c5a4 d7d6 e5d6 d8d6 d1d2 d5f6 d2d6 e7d6 a1d1 d6e7 d1e1 f6e4 c4d3 e4f6 b2b3 f6d7 f2f4 c8a6 d3a6 a8a6 g1f2 a6a8 f2e3 d7f6 e3d4 a8d8 d4c5');
INSERT INTO evaluations VALUES(5000,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(5000,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(5000,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(5000,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(5000,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(5000,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(5000,36,'cp 33','g8f6 d1d6 d8e7 c1f4 e7d6 f4d6 g7f8 d6f8 e8f8 e4e5 f6d5 c3e4 a7a5 f1d1 f8e7 b2b3 a5a4 g2g3 f7f6 c4d5 e6d5 e5f6 e7f7 e4c3 a4a3 b3b4 c8a6 b4b5 a6b5 c3b5 c6b5 d1d5 a8c8 a1c1 c8c3 d5b5 h8c8 c1d1 c3c2 d1d7 f7f6 d7h7 c2a2 b5b6 f6e5 b6g6 a2d2');
INSERT INTO evaluations VALUES(10718,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(10718,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(10718,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(10718,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(10718,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(10718,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(10718,36,'cp 64','a7a6 b5a4 b7b5 a4b3 c8b7 d2d4 e8g8 c1g5 f8e8 c2c3 d6e7 d4e5 f6e4 b3f7 g8f7 d1d5 f7f8 d5e4 e7g5 e4f5 g5f6 b1d2 g7g6 f5f6 d8f6 e5f6 h7h6 g2g4 g6g5 h2h4 g5h4 f3h4 f8f7 h4f5 f7f6 f2f4 c6e7 e1e7 e8e7 f5e7 f6e7 g1f2 a6a5 a1e1 e7f6 d2b3');
INSERT INTO evaluations VALUES(7355,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(7355,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(7355,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(7355,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(7355,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(7355,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(7355,36,'cp 76','b5a4 b7b5 a4b3 c8b7 d2d4 e8g8 c1g5 f8e8 c2c3 d6e7 a2a4 f6e4 g5e7 c6e7 f3e5 e4d6 b1d2 e7d5 d1g4 d5f6 g4h3 e8e7 d2f1 b7d5 b3c2 d5e4 c2e4 d6e4 h3d3 c7c6 h2h3 e7e8 d3f3 e4d6 b2b3 d8c7 f1g3 a6a5 c3c4 b5c4 b3c4 c6c5 a1c1 a8b8 c1d1 c5d4');
INSERT INTO evaluations VALUES(7360,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(7360,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(7360,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(7360,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(7360,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(7360,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(7360,36,'cp 75','e8g8 d2d4 b7b5 a4b3 c8b7 c1g5 f8e8 c2c3 d6e7 a2a4 f6e4 g5e7 c6e7 f3e5 e4d6 b1d2 e7d5 d1f3 d5f6 f3h3 e8e7 d2f1 b7d5 b3c2 d5e4 c2e4 d6e4 h3d3 c7c6 h2h3 e7e8 d3f3 e4d6 b2b3 b5a4 a1a4 d6e4 e1e4 f6e4 f3f7 g8h8 f7f5 e4g5 f5d7 d8d7 e5d7 e8e1');
INSERT INTO evaluations VALUES(2694,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(2694,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(2694,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(2694,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(2694,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(2694,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(2694,36,'cp 72','d2d4 b7b5 a4b3 c8b7 c1g5 f8e8 c2c3 d6e7 a2a4 f6e4 g5e7 c6e7 f3e5 e4d6 b1d2 e7d5 d1f3 d5f6 f3h3 e8e7 d2f1 b7d5 b3c2 d5e4 c2e4 d6e4 h3d3 c7c6 h2h3 e7e8 d3f3 e4d6 b2b3 d8c7 f1e3 b5a4 a1a4 c6c5 e1a1 c5d4 c3d4 c7b6 e3d5 f6d5 f3d5 a8d8');
INSERT INTO evaluations VALUES(11180,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(11180,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(11180,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(11180,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(11180,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11180,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(11180,36,'cp 32','e1g1 g7g6 d4c6 b7c6 c1g5 d8c7 d1d2 f8g7 a1d1 e8g8 b2b3 f8e8 e2c4 c7a5 g5f6 e7f6 c3a4 a5d2 d1d2 g7f8 a4b6 a8b8 b6c8 b8c8 d2e2 d6d5 c4a6 c8a8 a6b7 a8a2 b7c6 e8e4 e2e4 d5e4 c6e4 f6f5 e4d3');
INSERT INTO evaluations VALUES(11179,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(11179,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(11179,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11179,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(11179,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(11179,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(11179,36,'cp 36','g7g6 d4c6 b7c6 c1f4 d8c7 d1d2 f8g7 a1d1 e8g8 b2b3 f8e8 g1h1 c8b7 f2f3 f6d7 c3a4 c6c5 f4h6 g7h6 d2h6 e7e6 h6f4 d7e5 a4b2 a8d8 h1g1 g8g7 f4g3 f7f6 b2d3 e5d3 e2d3 a6a5 f1e1 b7c6 a2a4 c6d7 h2h3 d7c6');
INSERT INTO evaluations VALUES(14041,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(14041,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(14041,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(14041,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14041,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(14041,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(14041,36,'cp 41','a7a6 b5d3 c6b4 d3e2 c7c5 d4c5 b4c6 b1c3 e7g6 c3a4 f8c5 a4c5 d8a5 c2c3 a5c5 c1e3 c5a5 e3d4 c8d7 h2h4 c6d4 d1d4 g6e7 d4d2 a5c7 h4h5 h7h6 e1g1 f7f6 d2e3 e8c8 e2d3 e7c6 f1e1 c6e5 f3e5 f6e5 e3e5 c7e5 e1e5 c8c7 f2f4 h8f8 g2g3 c7d6');
INSERT INTO evaluations VALUES(11281,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(11281,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(11281,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(11281,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11281,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11281,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11281,36,'cp 42','b5d3 c6b4 d3e2 c7c5 d4c5 b4c6 b1c3 e7g6 c3a4 f8c5 a4c5 d8a5 c2c3 a5c5 c1e3 c5a5 e3d4 c8d7 h2h4 c6d4 d1d4 g6e7 h4h5 d7b5 e2d1 e7f5 d4d2 d5d4 c3d4 a5d2 e1d2 e8c8 a1c1 b5c6 d1c2 f5d4 f3d4 d8d4 d2e3 h8d8 c2h7 c8b8');
INSERT INTO evaluations VALUES(18078,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(18078,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(18078,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(18078,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(18078,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(18078,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(18078,36,'cp 144','c8b7 c2c3 d8f6 c3d4 f8b4 b1c3 f6d4 e1g1 g8e7 c3e2 d4f6 c1e3 b4a5 a1c1 e8g8 e2g3 a8e8 d3d4 a5b6 d1d2 g8h8 d4d5 b6e3 f2e3 f6d6 d2d4 f7f6 c1c5 a6a5 f1c1 c7c6 d5c6 d6d4 e3d4 d7c6 a2a3 e8d8 c1d1 a5a4 b3a2 g7g6 g3e2 h8g7');
INSERT INTO evaluations VALUES(19989,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(19989,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(19989,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(19989,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(19989,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(19989,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(19989,36,'cp 131','d4d5 c6b8 f1b5 c8d7 b5d3 g7g6 c3c4 f8g7 b1c3 g8f6 g1e2 b8a6 e1g1 e8g8 h2h3 e7e6 c1g5 h7h6 g5f4 e6e5 f4d2 f6h5 g2g4 a6b4 d3b1 h5f6 g1g2 h6h5 f2f3 b4a6 d1c1 f6h7 f1h1 h5h4 h1g1 b7b5 c4b5 a6c7 a2a4 a8b8');
INSERT INTO evaluations VALUES(20049,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(20049,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(20049,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(20049,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(20049,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(20049,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(20049,36,'cp 129','c6b8 f1b5 c8d7 b5d3 g7g6 c3c4 f8g7 b1c3 g8f6 h2h3 e8g8 g1f3 b8a6 a2a3 e7e6 e1g1 e6e5 c1g5 h7h6 g5e3 f6h7 f3h2 f7f5 f2f4 e5f4 e3f4 d8b6 f4d2 f5f4 d1c2 a6b4 a3b4 c5b4 g1h1 b4c3 d2c3 g7c3 c2c3');
INSERT INTO evaluations VALUES(21038,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(21038,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21038,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(21038,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(21038,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(21038,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(21038,36,'cp 39','g8f6 e2e3 c8g4 b1d2 e7e5 h2h3 g4f3 d2f3 e5d4 e3d4 f8d6 f1d3 d8e7 d3e2 e8g8 e1g1 a8e8 f1e1 e7d7 c2c4 d5c4 e2c4 e8e1 d1e1 h7h6 c1d2 c6e7 f3e5 d7f5 c4d3 f5h5 e5c4 e7f5 c4d6 f5d6 d2f4 f6d5 f4h2 f8e8 h2e5 h5g5');
INSERT INTO evaluations VALUES(14312,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14312,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(14312,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(14312,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(14312,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14312,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(14312,36,'cp 45','e2e3 c8g4 b1d2 e7e5 h2h3 g4f3 d2f3 e5d4 e3d4 f8d6 f1d3 d8e7 d3e2 e8g8 e1g1 a8e8 e2d3 c6d8 c2c4 c7c6 d1c2 d5c4 d3c4 d8e6 c1d2 e6f4 a1e1 e7c7 f3e5 f4d5 c2f5 g7g6 f5g5 f6h5 g5h4 d6e5 d4e5 e8e5 e1e5 c7e5 d2h6');
INSERT INTO evaluations VALUES(14310,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(14310,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(14310,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14310,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(14310,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14310,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(14310,36,'cp 47','c8g4 b1d2 e7e5 h2h3 g4f3 d2f3 e5d4 e3d4 f8d6 c1e3 e8g8 f1d3 c6e7 e1g1 c7c6 f1e1 e7g6 c2c4 d5c4 d3c4 d8c7 e3g5 f6d5 h3h4 f8e8 h4h5 e8e1 d1e1 g6f4 g5f4 d5f4 f3g5 f4d5 e1e4 d5f6 e4h4 a8f8 h5h6 c7d7 a1e1 d7g4 h4g4 f6g4');
INSERT INTO evaluations VALUES(24957,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(24957,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(24957,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(24957,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(24957,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(24957,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(24957,36,'cp 49','b1d2 e7e5 h2h3 g4f3 d2f3 e5d4 e3d4 f8d6 c1e3 e8g8 f1d3 c6e7 e1g1 c7c6 f1e1 e7g6 c2c4 d5c4 d3c4 d8c7 d1c2 a8e8 e3g5 f6d5 g5d2 c7d7 a1d1 d6f4 d2c3 e8e1 d1e1 f4d6 c3d2 d6f4 d2a5 h7h6 c4d3 b7b6 a5c3 g6e7 f3e5 f4e5 d4e5 d5c3 c2c3 e7g6 d3g6 f7g6');
INSERT INTO evaluations VALUES(24419,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(24419,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(24419,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(24419,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(24419,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(24419,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(24419,36,'cp 102','b5a4 f8e7 d2d3 e8g8 e1g1 c6c5 f3f5 a8c8 b1c3 c5c4 c3e4 g7g6 f5h3 h7h5 e4f6 e7f6 g5e4 c4d3 c2d3 b7c6 e4f6 d8f6 a4d1 c6b5 f1e1 a5c6 h3g3 f6f5 c1h6 f8d8 a1c1 b5d3 h2h3 f7f6 h6d2 c6d4 c1c8 d8c8 d2c3 d4c2 d1c2 d3c2');
INSERT INTO evaluations VALUES(24418,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(24418,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(24418,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(24418,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(24418,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(24418,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(24418,36,'cp 85','f8e7 e1g1 e8g8 d2d3 c6c5 f3f5 c5c4 b1c3 c4d3 c2d3 a8c8 f5e5 a5c6 e5f4 c6b4 f4g3 d8d3 f1d1 d3g3 h2g3 f8d8 d1d8 c8d8 a4b3 f6d5 c1d2 h7h6 g5f3 d5b6 a1e1 e7f6 e1d1 b4d3 d2e3 d3b2 d1d8 f6d8 f3e5 d8f6');
INSERT INTO evaluations VALUES(23998,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(23998,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(23998,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(23998,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(23998,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(23998,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(23998,36,'cp 93','d2d3 e8g8 e1g1 c6c5 f3f5 c5c4 b1c3 a8c8 c3e4 g7g6 f5e5 c4d3 c2d3 a5c6 e5b5 c8b8 g5f3 c6d4 f3d4 f6e4 d4c6 b7c6 b5c6 e4c5 a4c2 b8b6 c6f3 c5e6 a1b1 e6d4 f3d1 e7g5 c1e3 f8e8 c2a4 e8e5 e3d4 d8d4 a4b3 g8g7 g2g3 h7h5 d1c2 g5f6 c2c4 e5e7 g1g2 h5h4 g3h4');
INSERT INTO evaluations VALUES(41441,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(41441,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(41441,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(41441,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(41441,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(41441,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(41441,36,'cp -13','e8g8 c4d5 f6d5 f4e5 g7e5 f3e5 b8d7 e5d7 c8d7 f1c4 d5b6 c4b3 e7e5 e1g1 e5d4 d1d4 d7b5 f1c1 d8d4 e3d4 b5c6 b1c3 f8d8 d4d5 c6d5 c3d5 b6d5 b3d5 d8d5 c1c7 b7b6 g2g3 d5d2 a1b1 a8d8 g1f1 g8g7 f1g2 a7a5 c7b7 d2c2 b7b6');
INSERT INTO evaluations VALUES(36724,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36724,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(36724,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36724,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36724,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36724,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(36724,36,'cp -13','c4d5 f6d5 f4e5 g7e5 f3e5 b8d7 e5d7 c8d7 f1c4 d5b6 c4b3 e7e5 e1g1 e5d4 d1d4 d7b5 f1c1 d8d4 e3d4 b5c6 b1c3 f8d8 d4d5 b6d5 c3d5 c6d5 b3d5 d8d5 c1c7 b7b6 h2h3 a7a5 c7b7 d5d2 b7b6 a5a4 b2b4 a4a3 g2g3 d2b2 b4b5');
INSERT INTO evaluations VALUES(48196,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(48196,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(48196,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(48196,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(48196,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(48196,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(48196,36,'cp 142','f1b5 c7c6 d5c6 b7c6 b5e2 f8e7 e1g1 d8c7 b1a3 e7a3 b2a3 c6c5 c1b2 e8g8 b2e5 c7a5 e5f6 g7f6 f1e1 b8c6 e2d3 c6e7 f3h4 e7g6 d1f3 a8b8 f3f6 a5d8 f6d8 f8d8 h4f3 b8b6 g2g3 c8b7 d3e4 b7d5 c2c3 g8g7 e4d5 d8d5 e1b1 d5d8 b1b6 a7b6');
INSERT INTO evaluations VALUES(54624,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(54624,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(54624,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(54624,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(54624,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(54624,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(54624,36,'cp 137','c1f4 g8f6 e2e3 f8e7 f1d3 e8g8 h2h3 f6e4 e1g1 b8d7 a1c1 g8h8 c3e2 a7a5 f4h2 e7d6 h2d6 e4d6 e2f4 d8e7 h3h4 h8g8 d1c2 d7f6 f1e1 d6e4 f3g5 e4g5 h4g5 f6e4 g5g6 e7h4 c4d5 e6d5 f2f3 e4g3 g6h7 h4h7');
INSERT INTO evaluations VALUES(54620,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(54620,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(54620,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(54620,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(54620,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(54620,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(54620,36,'cp 117','d5c4 e2e4 b7b5 e4f5 g8f6 g2g3 f8b4 f1g2 f6d5 e1g1 d5f4 g3f4 e8g8 f5e6 b4c3 b2c3 c8e6 f3g5 e6d5 f1e1 d5g2 d1h5 h7h6 g5e6 d8e8 h5g4 e8d7 g4g2 b8a6 e6f8 a8f8 g2f3 a6c7 e1e5 a7a5 a2a3 b5b4 a3b4 a5b4 c3b4 d7d4 a1e1 f8f4 e5e4 f4f3 e4d4 c4c3 d4d6 f3f6 d6f6 g7f6 e1c1 c7d5 f2f3');
INSERT INTO evaluations VALUES(13995,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(13995,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(13995,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(13995,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(13995,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(13995,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(13995,36,'cp 54','d1d2 f8e7 f2f3 e8g8 g2g4 d6d5 g4g5 f6h5 f3f4 c6d4 d2d4 b7b6 e4d5 e7c5 d4d2 c5e3 d2e3 d8c7 e1c1 c7f4 e3f4 h5f4 f1c4 e6d5 c3d5 c8e6 d5e7 g8h8 c4e6 f4e6 h2h4 g7g6 e7c6 h8g7 d1d7 h7h6 h1e1 e6c5 g5h6 g7h6 d7e7 a7a5 a2a4 h6g7 b2b3 a8c8 c6a7 c8a8 a7b5');
INSERT INTO evaluations VALUES(15037,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15037,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(15037,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(15037,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(15037,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(15037,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(15037,36,'cp -6','f8d6 c3b5 e8g8 b5d6 d8d6 f1e2 c8g4 h2h3 g4h5 e1g1 a7a6 c1d2 f6e4 a1c1 f7f5 f3e1 h5e2 d1e2 f5f4 e1f3 a8e8 f1e1 f4e3 d2e3 e8e7 e2c2 d6g6 g1h2 g6f7 h2g1 f8e8 c2b3 e4d6 e3d2 e7e1 c1e1 e8e1 d2e1 h7h6');
INSERT INTO evaluations VALUES(27464,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(27464,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(27464,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(27464,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(27464,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(27464,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(27464,36,'cp 76','f1d1 f8e7 b1c3 e7d8 a2a4 d7e6 f2f3 e8g8 a4a5 f8e8 c1e3 e6c8 d4d3 d8e7 c3e2 f6d7 h2h3 a7a6 d3c3 e7f8 e3f2 c7c6 e2f4 c8c7 c3d2 a8d8 c2c4 d7c5 d2c3 d6d5 c4d5 c7f4 f2c5 f8c5 c3c5 c6d5 d1d5 d8d5');
INSERT INTO evaluations VALUES(41784,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41784,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41784,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(41784,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(41784,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41784,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(41784,36,'cp 0','e2e3 e8g8 f1d3 d6f4 e3f4 b7b6 c3e2 c7c5 c2c3 c8a6 d3a6 b8a6 e1g1 a6b8 d1d3 b8c6 f4f5 f8e8 f5e6 e8e6 e2g3 d8d7 a1e1 a8e8 e1e6 e8e6 d4c5 b6c5 d3b5 d7d6 g3f5 d6f4 f5g3');
INSERT INTO evaluations VALUES(41783,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(41783,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(41783,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(41783,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(41783,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41783,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41783,36,'cp 0','e8g8 f1d3 d6f4 e3f4 b7b6 c3e2 c7c5 c2c3 c8a6 d3a6 b8a6 e1g1 a6c7 d1c2 a7a5 f1e1 c5c4 f3e5 b6b5 a2a3 f8e8 f4f5 b5b4 a3b4 a5b4 a1a8 d8a8 f5e6 c7e6 h2h3 a8b7 e2g3 g7g6 e5g4 f6g4 h3g4 b4c3');
INSERT INTO evaluations VALUES(42010,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(42010,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(42010,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42010,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42010,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(42010,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(42010,36,'cp 29','b4d6 b1d2 e8g8 f1d3 d6f4 e3f4 c7c5 d4c5 a7a5 d1e2 d8c7 f3e5 a5a4 a1c1 b8d7 e1g1 d7c5 d3b1 c8d7 f1e1 f8d8 g2g4 d7e8 g4g5 f6d7 e2e3 d7e5 f4e5 a8c8 h2h4 a4a3 b2b4 c5a4 f2f4');
INSERT INTO evaluations VALUES(41787,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(41787,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(41787,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(41787,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(41787,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41787,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(41787,36,'cp 36','b1d2 e8g8 f1d3 d6f4 e3f4 b7b6 f3e5 c7c5 d1e2 a7a5 h2h4 c8a6 d3a6 b8a6 a2a3 d8c7 g2g4 f8c8 e1g1 f6e8 a1e1 e8d6 h4h5 c8e8 e5f3 h7h6 f3e5 a6b8 g4g5 b8c6 d2f3 c6e5 f3e5 c5d4 g5h6 g7h6 c3d4 g8h8 e1c1 c7b7 g1h1 d6f5 f1e1');
INSERT INTO evaluations VALUES(42463,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(42463,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(42463,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(42463,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(42463,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(42463,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(42463,36,'cp 72','c7c5 d4c5 b8c6 g1f3 d8c7 e1g1 b4c5 d1e2 f6g4 c3b5 c7b8 h2h3 g4e5 f3e5 c6e5 c1e3 c5e3 e2e3 a7a6 b5d4 e8g8 a1d1 b7b5 f2f4 e5d3 d1d3 b8c7 f1d1 c8b7 d4b3 a8c8 c2c3 d7d5 e4e5 f7f6 b3d4 c8e8 e3g3 f6e5 f4e5');
INSERT INTO evaluations VALUES(43631,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(43631,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(43631,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(43631,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(43631,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(43631,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(43631,36,'cp 65','f3g5 d6d5 e4d5 h7h6 g5e4 f6e4 c3e4 d8h4 c4d3 c8g4 d3e2 g4e2 d1e2 b8d7 d2d3 f7f5 e4g3 g7g6 c2c4 e8c8 c1d2 f8c5 d2c3 h6h5 e1g1 h4e7 d3d4 e5d4 e2e7 c5e7 c3d4 h8e8 a1d1 b7b5 b2b3 b5c4 b3c4 e7b4 d4e3 d7b6 e3b6 c7b6');
INSERT INTO evaluations VALUES(43628,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(43628,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(43628,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(43628,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(43628,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(43628,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(43628,36,'cp 81','d6d5 e4d5 h7h6 g5e4 f6e4 c3e4 d8h4 d1e2 c8g4 e2e3 b8d7 e3g3 g7g5 g3h4 g5h4 f2f3 g4h5 c4e2 f7f5 g2g4 h5g4 f3g4 f5e4 d2d3 d7c5 d3e4 c5e4 e2d3 e4d6 c1e3 f8e7 c2c4 b7b6 b2b4 e5e4 d3e2 e7g5 e1d2 a6a5 c4c5 d6f7 d5d6 e8c8 h1f1 f7e5 e3g5 h6g5 f1f5 c7d6 c5d6 d8d6 d2e3 e5d3 b4a5 d3f4 a1b1 c8c7 a5b6 d6b6 b1b6');
INSERT INTO evaluations VALUES(44385,30,'cp 334',NULL);
INSERT INTO evaluations VALUES(44385,31,'cp 312',NULL);
INSERT INTO evaluations VALUES(44385,32,'cp 318',NULL);
INSERT INTO evaluations VALUES(44385,33,'cp 320',NULL);
INSERT INTO evaluations VALUES(44385,34,'cp 345',NULL);
INSERT INTO evaluations VALUES(44385,35,'cp 366',NULL);
INSERT INTO evaluations VALUES(44385,36,'cp 332','d2d4 d7d5 c4b3 d5e4 c3e4 c8f5 d1e2 e5d4 e1g1 f5e4 e2e4 b8a6 e4g4 d8d7 b3e6 d7d6 f1e1 h7h5 g4h4 a6c5 c1f4 d6d8 e6h3 e8f7 a1d1 d8d5 c2c4 d5c4 d1d4 c4c2 b2b4 c5d3 h3e6 f7e8 d4d3 c2d3 h4g3 h5h4 g3g4 g7g6 h2h3 f8g7 f4e3 h8h5 e3c5 a8d8 e6b3 h5c5 b4c5');
INSERT INTO evaluations VALUES(44388,30,'cp 355',NULL);
INSERT INTO evaluations VALUES(44388,31,'cp 351',NULL);
INSERT INTO evaluations VALUES(44388,32,'cp 353',NULL);
INSERT INTO evaluations VALUES(44388,33,'cp 359',NULL);
INSERT INTO evaluations VALUES(44388,34,'cp 355',NULL);
INSERT INTO evaluations VALUES(44388,35,'cp 387',NULL);
INSERT INTO evaluations VALUES(44388,36,'cp 386','d7d5 c4b3 d5e4 c3e4 c8f5 d1e2 e5d4 e1g1 b8a6 f1d1 f5e4 f3d4 d8d6 e2e4 a6c5 e4f3 c5b3 d4b5 d6e5 b5d6 e5d6 d1d6 b3a1 f3e2 a8d8 d6d8 e8d8 e2d1 d8c8 c1d2 a1c2 d1c2 e7d5 c2e4 c8d7 e4g4 d7c7 g4e6 f8b4 d2b4 d5b4 e6e7 c7b6 e7b4 b6a6 b4a3 a6b6 a3e7 b6a6 b2b4 b7b6 e7g7');
INSERT INTO evaluations VALUES(53548,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(53548,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(53548,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(53548,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(53548,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(53548,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(53548,36,'cp 43','e6d5 e4e5 c6c5 g1f3 c5d4 f3d4 b8c6 f1b5 f8b4 b1c3 g8e7 c1g5 c8d7 e1g1 c6d4 g5e7 d8e7 b5d7 e7d7 d1d4 b4c3 b2c3 a8c8 d4a7 e8g8 a1b1 c8c7 h2h3 f8c8 f2f4 h7h6 f4f5 b7b5 a7b6 c7c3 b6b5 d7b5 b1b5 c3c2 f1f2');
INSERT INTO evaluations VALUES(53729,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(53729,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53729,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(53729,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(53729,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(53729,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(53729,36,'cp 44','e4e5 c6c5 g1f3 c5d4 f3d4 f8b4 b1c3 b8c6 f1b5 g8e7 c1g5 c8d7 e1g1 c6d4 g5e7 d8e7 d1d4 b4c3 b5d7 e7d7 b2c3 a8c8 d4a7 e8g8 a1b1 c8c3 b1b7 d7c6 b7b8 g7g6 h2h4 c3c4 h4h5 g6h5 f1d1 h5h4 b8f8 g8f8 a7b8 f8g7 b8d8 c6g6 d8d5 h4h3');
INSERT INTO evaluations VALUES(53742,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(53742,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53742,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(53742,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(53742,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(53742,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(53742,36,'cp 32','c6c5 g1f3 c5d4 f3d4 f8b4 b1c3 g8e7 f1b5 b8c6 c1g5 c8d7 e1g1 c6d4 g5e7 d8e7 d1d4 b4c3 b5d7 e7d7 b2c3 a8c8 d4a7 e8g8 a1b1 c8c3 b1b7 d7c6 h2h3 h7h6 a2a4 c3a3 b7b4 f8a8 b4b8 a8b8 a7b8 g8h7 b8d6 c6c3 d6d5 a3a4 f1d1 c3c4 d5d7');
INSERT INTO evaluations VALUES(846,30,'cp -126',NULL);
INSERT INTO evaluations VALUES(846,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(846,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(846,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(846,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(846,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(846,36,'cp -94','b1c3 f5d4 f3d4 b6d4 c3b5 d4d1 f1d1 e8d7 c1f4 a7a6 b5c3 d7e8 a1c1 c8d7 c3a4 a8b8 a4c5 f8e7 f4g3 g7g5 e2h5 c6d8 f2f3 b8c8 g3f2 d7b5 b2b3 b7b6 c5a4 c8c1 d1c1 b5a4 b3a4 e8d7 f2b6 d8c6 h5f7 c6e5 c1c7 d7d6 c7a7 e5f7 b6a5 h8b8 a5c7 d6c6 c7b8 e7c5 g1f1 c5a7');
INSERT INTO evaluations VALUES(843,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(843,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(843,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(843,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(843,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(843,35,'cp -130',NULL);
INSERT INTO evaluations VALUES(843,36,'cp -131','f5d4 f3d4 b6d4 c3b5 d4d1 f1d1 e8d7 c1f4 a7a6 b5c3 b7b5 a1c1 c8b7 a2a4 b5b4 c3e4 f8e7 e4c5 e7c5 c1c5 h8c8 d1c1 c6d4 e2d1 a6a5 c5c8 a8c8 c1c8 b7c8 f4e3 d4c6 e3b6 c8a6 d1c2 h7h6 f2f4 g7g5 g2g3 g5f4 g3f4 a6e2 g1f2 e2c4 b2b3 c4a6');
INSERT INTO evaluations VALUES(836,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(836,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(836,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(836,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(836,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(836,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(836,36,'cp -98','f3d4 b6d4 c3b5 d4d1 f1d1 e8d7 c1f4 a7a6 b5c3 b7b5 a2a4 b5b4 c3e4 c8b7 a1c1 f8e7 e4c5 e7c5 c1c5 h8c8 d1c1 c6d4 e2d1 c8c5 c1c5 a6a5 h2h4 d4c6 d1g4 d7c7 f4e3 b7a6 f2f4 a6c4 g4d1 b4b3 e3d2 c7b7 h4h5 b7b6 d2e3 b6c7 g1f2 c7d7 g2g4 c6b4 d1e2 c4e2 f2e2 a8c8 c5c8 d7c8');
INSERT INTO evaluations VALUES(834,30,'cp -110',NULL);
INSERT INTO evaluations VALUES(834,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(834,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(834,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(834,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(834,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(834,36,'cp -103','b6d4 c3b5 d4d1 f1d1 a8b8 c1e3 a7a6 b5d6 f8d6 e5d6 e8d7 a1c1 d7d6 e3c5 d6c7 f2f4 c8d7 f4f5 b8c8 f5e6 d7e6 a2a4 h8g8 c5a7 a6a5 e2b5 g8d8 a7e3 c7b8 e3b6 d8d7 g1f2 f7f6 b5c6 b7c6 b6a5 b8b7 d1e1 e6f5 a5b4 b7a6');
INSERT INTO evaluations VALUES(858,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(858,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(858,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(858,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(858,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(858,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(858,36,'cp -98','c3b5 d4d1 f1d1 a8b8 c1e3 a7a6 b5d6 f8d6 e5d6 e8d7 a1c1 d7d6 e3c5 d6c7 f2f4 b7b6 c5f2 c7d6 d1d3 c8d7 d3b3 b6b5 b3c3 b8c8 f2c5 d6c7 a2a4 b5a4 e2d1 c6a5 c5f8 a5c4 f8g7 h8g8 g7d4 d7b5 c3h3 h7h6 h3h6 c7d7 h6h7 c4d6 c1c3 c8c3');
INSERT INTO evaluations VALUES(4822,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(4822,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(4822,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(4822,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(4822,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(4822,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(4822,36,'cp 32','c1e3 d7d6 d3d4 c6d4 f3d4 e5d4 e3d4 c5d4 d1d4 d8f6 d4e3 g8e7 e1c1 c8e6 c4e2 e7c6 f2f4 e8c8 a2a3 c8b8 g2g4 f6h4 e3g3 h4f6 h2h3 g7g5 f4f5 e6d7 g3e3 h8e8 c1b1 f6g7 c3d5 c6e7 e2c4 d7c6 d5b4 a6a5 b4c6 e7c6 e3b3 e8e4 c4f7');
INSERT INTO evaluations VALUES(4823,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(4823,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(4823,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(4823,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(4823,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(4823,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(4823,36,'cp 26','d7d6 d3d4 c5d4 f3d4 c6d4 e3d4 e5d4 d1d4 d8f6 d4e3 g8e7 e1c1 c8e6 c4e2 e7c6 f2f4 e8c8 h2h4 c8b8 g2g4 d8e8 c1b1 g7g6 e3d2 f6d8 c3d5 e6d5 e4d5 c6e7 h4h5 e7c8 e2d3 d8f6 c2c3 g6h5 h1h5 c8b6 a2a3');
INSERT INTO evaluations VALUES(8597,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(8597,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8597,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(8597,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(8597,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(8597,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(8597,36,'cp 25','c2c3 e8g8 a2a4 c8e6 c4e6 f7e6 b2b4 a7a6 e1g1 d6d5 b1d2 d8e8 c1b2 e7d6 b4b5 c6e7 c3c4 e7g6 g2g3 e8f7 d1e2 h7h6 f3h2 c7c6 b5c6 b7c6 d2f3 f6d7 b2c3 a8b8 a1b1 d7c5 b1b8 f8b8 e4d5 e6d5 c3e5 b8e8 d3d4 c5a4 f1a1 a4b6 a1a6 b6c4 a6c6 c4e5');
INSERT INTO evaluations VALUES(8591,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(8591,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(8591,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(8591,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(8591,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8591,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(8591,36,'cp 16','e8g8 a2a4 c8e6 c4e6 f7e6 b2b4 d6d5 b1d2 d5e4 d2e4 d8d5 e1g1 f6e4 d3e4 d5e4 b4b5 c6a5 d1d7 f8f7 d7e6 e4c4 e6e5 e7f6 e5h5 g7g6 h5g4 c4g4 h3g4 f6c3 a1a3 c3g7 a3d3 c7c5 b5c6 a5c6 f1d1 b7b6 f3g5 f7e7 c1a3');
INSERT INTO evaluations VALUES(3382,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(3382,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(3382,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(3382,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(3382,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(3382,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(3382,36,'cp 23','a2a4 c8e6 c4e6 f7e6 b2b4 d6d5 b1d2 d5e4 d2e4 d8d5 e1g1 f8d8 e4f6 e7f6 f3g5 d5d3 d1h5 f6g5 c1g5 d8d5 b4b5 c6a5 c3c4 d5d4 g5e3 d4c4 h5e5 d3f5 e5g3 f5g6 g3g6 h7g6 f1d1 b7b6 d1d7 a8f8 a1d1 f8f7 d7d8 f7f8 d8d4 f8f5 d4c4 a5c4 d1c1 c4e3');
INSERT INTO evaluations VALUES(9470,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9470,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9470,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9470,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9470,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9470,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9470,36,'cp 0','f6e4 d1e2 d8d5 e5d3 e8g8 f1e1 c8f5 d3c5 e4c5 b1c3 d5d7 d2d3 f8e8 c1e3 f5g6 e2f3 c5e6 c3e2 e8e7 e2f4 a8e8 f4e6 e7e6 h2h3 b7b6 a2a4 a7a5 e3d2 f7f6 e1e6 e8e6 a1e1 e6e1 d2e1 c6c5 f3f4 d7d4 f4c7 d4b2 c7d8 g8f7 d8c7 f7g8');
INSERT INTO evaluations VALUES(9506,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9506,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9506,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9506,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9506,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(9506,36,'cp 6','d1e2 d8d5 e5d3 e8g8 f1e1 c8f5 d3c5 e4c5 d2d3 c5e6 b1c3 d5d7 c1e3 a8e8 e2f3 f5g6 c3e2 e8e7 a2a4 a7a5 h2h4 h7h5 e2f4 f8e8 f4e6 e7e6 f3f4 c6c5 e3d2 b7b6 e1e6 e8e6 a1e1 e6e1 d2e1 d7e6 e1c3 e6a2 f2f3 f7f6 f4c7 a2a4 c7b6 a4c2 b6a5');
INSERT INTO evaluations VALUES(10439,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(10439,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(10439,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(10439,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(10439,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(10439,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(10439,36,'cp -31','f6e4 c4d5 e4d6 e1g1 e5e4 f3d4 c6d4 c3d4 c7c6 d5b3 e8g8 d2d3 d8f6 f1e1 e4d3 d1d3 d6f5 c2c3 d7d5 b3c2 f6h4 g2g3 h4h3 c1f4 c8d7 f2f3 g7g6 d3d2 f8e8 c2d3 f5g7 a3a4 e8e1 a1e1 a8e8 e1e8 g7e8 d2e3 h3e6 e3e6 d7e6 a4a5 f7f6');
INSERT INTO evaluations VALUES(10686,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(10686,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(10686,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(10686,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(10686,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(10686,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(10686,36,'cp -19','c4d5 e8g8 e1g1 e4f6 d5c6 d7c6 f3e5 f6d5 f1e1 f7f6 e5f3 f8e8 c3c4 d5e7 h2h3 e7g6 c1b2 e8e1 d1e1 c8d7 e1e3 d8e7 c4c5 b7b6 e3e7 g6e7 d2d3 e7g6 a1e1 g8f7 f3d2 g6f4 d2e4 f4e6 f2f3 a8b8 g1f2 b6c5 b2c3 c5c4');
INSERT INTO evaluations VALUES(14182,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(14182,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14182,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14182,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14182,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(14182,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(14182,36,'cp 39','f6d5 g1f3 d5c3 b2c3 e7e5 d4d5 e5e4 d5c6 d8d1 e1d1 e4f3 e2f3 b7c6 f1c4 f8d6 h1e1 e8f8 h2h4 h7h5 c1e3 c8f5 d1d2 a7a5 c4d3 f5d3 d2d3 f8e7 a2a4 e7d7 a1b1 h8e8 b1b7 e8d8 d3c4 d8e8 e1b1 d7c8 c4d3 e8d8 d3e2 g7g6 b7b3 c8d7 b3b2 c6c5 g2g3 d7c6');
INSERT INTO evaluations VALUES(17770,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(17770,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(17770,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(17770,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(17770,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(17770,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(17770,36,'cp 147','c2c3 f7f6 f1d3 f6e5 d4e5 g8h6 d1e2 h6f7 h2h4 g7g6 b2b4 b7b6 c1f4 c8b7 b1d2 f8h6 f4h6 f7h6 a2a4 d8e7 a4a5 e8g8 e1g1 d5d4 d2e4 d4c3 e4f6 f8f6 e5f6 e7f6 a5b6 c7b6 d3a6 c3c2 e2c4 a8a6 a1a6 b7a6 c4a6 c6b4 a6b6');
INSERT INTO evaluations VALUES(17768,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(17768,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(17768,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(17768,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(17768,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(17768,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(17768,36,'cp 107','g8h6 f1d3 f7f6 d1e2 f6e5 d4e5 c8d7 b1d2 f8c5 d2b3 c5b6 c1g5 c6e7 e1c1 h6f7 g5e3 b6e3 f2e3 e7g6 d3g6 h7g6 e3e4 f7g5 e4d5 g5f3 g2f3 d8g5 d1d2 e6d5 e5e6 d7b5 e2g2 g5g2 d2g2 h8h6 h1g1 b5d3 b3c5 d3f5 c5b7 f5e6 g2g6 h6g6 g1g6 e8f7 g6g2');
INSERT INTO evaluations VALUES(18179,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(18179,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18179,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18179,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(18179,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(18179,36,'cp 20','e7e6 b1c3 c5d4 d1d4 d7d5 e1g1 g8f6 f1d1 d8c7 e4d5 c6d5 c1f4 c7a7 c3a4 a7d4 d1d4 c8d7 a4b6 a8a7 b6d7 f6d7 c2c4 f8c5 d4d1 d5c4 a1c1 d7b6 f3e5 a7a8 e5c4 b6d5 f4d2 e8e7 a2a3 h8c8 g2g3 a8b8 b2b4 c5d4 c4a5 c8c1 d1c1 e7d7 g1g2 e6e5');
INSERT INTO evaluations VALUES(18141,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(18141,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(18141,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(18141,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(18141,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(18141,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(18141,36,'cp 12','e1g1 c5d4 d1d4 d7d5 b1c3 g8f6 f1d1 d8c7 e4d5 c6d5 c1f4 c7a7 c3a4 a7d4 f3d4 c8d7 a4b6 a8a7 c2c4 a7b7 b6d7 e8d7 f4e5 d5c4 b2b3 c4b3 a2b3 f6d5 a1a6 f8c5 e5g7 h8c8 g7e5 c5d4 e5d4 b7b3 d1a1 b3b7 g2g3 d7e7 g1g2 c8c4 a6a7 b7a7 a1a7 e7d6');
INSERT INTO evaluations VALUES(18140,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(18140,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18140,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(18140,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18140,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(18140,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(18140,36,'cp 13','c5d4 d1d4 d7d5 b1c3 g8f6 f1d1 d8c7 e4d5 c6d5 c1f4 c7c5 c3a4 c5d4 f3d4 c8d7 a4b6 a8a7 c2c4 d5c4 a1c1 f6d5 b6d7 a7d7 c1c4 f8e7 f4e5 e8g8 d4c6 f8e8 c6e7 d7e7 g2g3 e7b7 b2b3 a6a5 e5c3 d5c3 c4c3 g7g6 d1c1');
INSERT INTO evaluations VALUES(21396,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(21396,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21396,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21396,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(21396,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21396,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(21396,36,'cp 23','g1f3 f8d6 e2e3 g8f6 b1d2 e8g8 f4g3 b7b6 f1d3 c6e7 e3e4 d5e4 d2e4 f6e4 d3e4 a8b8 e1g1 c8b7 e4b7 b8b7 d1e2 d8d7 f1e1 b7b8 a1d1 h7h6 h2h4 b8d8 f3e5 d6e5 g3e5 e7f5 h4h5 a7a5 c3c4 f8e8');
INSERT INTO evaluations VALUES(25373,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(25373,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(25373,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(25373,33,'cp 271',NULL);
INSERT INTO evaluations VALUES(25373,34,'cp 263',NULL);
INSERT INTO evaluations VALUES(25373,35,'cp 256',NULL);
INSERT INTO evaluations VALUES(25373,36,'cp 258','g5e6 f7e6 c4e6 d8e7 e6h3 g7g6 e1g1 c6d8 a2a4 d8f7 a4a5 f8g7 c1e3 g7h6 e3h6 f7h6 b1d2 h6f7 g2g3 e8g8 d1e2 g8h8 c2c3 a8e8 h3g2 c7c6 d3d4 h8g7 d4e5 d6e5 a1e1 h7h5 d2f3 b7b5 a5b6 a7b6 e1a1 b6b5');
INSERT INTO evaluations VALUES(25369,30,'cp 269',NULL);
INSERT INTO evaluations VALUES(25369,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(25369,32,'cp 266',NULL);
INSERT INTO evaluations VALUES(25369,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(25369,34,'cp 261',NULL);
INSERT INTO evaluations VALUES(25369,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(25369,36,'cp 253','f7e6 c4e6 d8e7 e6h3 g7g6 b1c3 f8g7 e1g1 e8g8 c1e3 g8h8 d1d2 c6d4 a1e1 c7c6 e3d4 e5d4 c3e2 f6d7 f2f4 b7b5 g1h1 d7c5 g2g3 c5a4 c2c3 d4c3 e2c3 a4c3 b2c3 a8e8 d3d4 a7a5 h3g2 e7d7 a2a3 a5a4 d2d3 g6g5');
INSERT INTO evaluations VALUES(25431,30,'cp 269',NULL);
INSERT INTO evaluations VALUES(25431,31,'cp 269',NULL);
INSERT INTO evaluations VALUES(25431,32,'cp 269',NULL);
INSERT INTO evaluations VALUES(25431,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(25431,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(25431,35,'cp 281',NULL);
INSERT INTO evaluations VALUES(25431,36,'cp 264','c4e6 d8e7 e6h3 g7g6 b1c3 f8g7 e1g1 e8g8 c1e3 a8e8 g2g3 g8h8 a2a4 b7b6 g1h1 e7f7 d1d2 c6d4 f2f4 c7c5 f4e5 d6e5 c3d5 f7b7 d2g2 f6d5 e4d5 b7e7 c2c3 d4f5 e3d2 f5d6 h3e6 f8f1 a1f1 e8f8 h1g1 f8f1 g1f1 e5e4 d3e4');
INSERT INTO evaluations VALUES(25423,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(25423,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(25423,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(25423,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(25423,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(25423,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(25423,36,'cp 241','d8e7 e6h3 g7g6 c2c3 c6d8 e1g1 d8f7 a2a4 f8h6 c1h6 f7h6 b1d2 h6f7 g2g3 a7a5 d1e2 e8g8 h3g2 f7g5 a1e1 a8e8 f2f4 e5f4 f1f4 g8g7 d3d4 h7h5 h2h4 g5e6 f4f3 e6d8 e2b5 f6g4 b5a5 c7c6 f3f8 e8f8 d2f3 d8f7 g2h3');
INSERT INTO evaluations VALUES(45190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45190,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(45190,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(45190,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(45190,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(45190,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(45190,36,'cp 0','f1g2 g7g6 g1e2 b8c6 c3d5 f8g7 e1g1 e8g8 c2c3 e7e5 d5e3 f6g4 e3d5');
INSERT INTO evaluations VALUES(48042,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(48042,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(48042,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(48042,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(48042,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(48042,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(48042,36,'cp 72','d6d5 g2g3 f8e7 f1g2 b8d7 d1d3 e8g8 e1g1 b7b6 c4d5 e6d5 c1f4 c8b7 a1d1 f6h5 e2e3 c7c6 f1e1 h5f4 e3f4 a7a5 f3e5 b6b5 h2h4 d7e5 d4e5 f8e8 h4h5 e7f8 g2h3 b7c8 f4f5 d8g5 d3f3 a8a7 a2a3 a7c7 e5e6 f7e6 f5e6');
INSERT INTO evaluations VALUES(50401,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(50401,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(50401,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(50401,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(50401,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(50401,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(50401,36,'cp 63','f4g3 d5f6 g1f3 e7e6 e2e3 d7d5 c2c4 f6e4 f1e2 e4g3 h2g3 g7g6 b1c3 f8g7 e1g1 e8g8 d1d2 d5c4 e2c4 e6e5 a1d1 e5d4 e3d4 b8c6 f1e1 c8g4 d4d5 g4f3 g2f3 c6d4 g1g2 d8f6 d2d3 f6b6 c3e4 d4f5 f3f4 f5d6 c4b3 a7a5 e4d6 c7d6 e1e7');
INSERT INTO evaluations VALUES(50402,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(50402,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(50402,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(50402,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(50402,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(50402,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(50402,36,'cp 37','d5f6 g1f3 e7e6 e2e3 d7d5 c2c4 f6e4 b1c3 e4g3 h2g3 g7g6 f1e2 f8g7 b2b4 e8g8 d1b3 d5c4 e2c4 a7a5 b4b5 c7c5 b5c6 b8c6 e1g1 a5a4 b3a3 g7f6 a1b1 f6e7 a3c1 c8d7 f1d1 c6a5 c4d3 a8c8 c1d2 d7c6 f3e5 a4a3 d1c1 c8c7 c3b5 c6b5');
INSERT INTO evaluations VALUES(55186,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(55186,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(55186,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(55186,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(55186,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(55186,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(55186,36,'cp 29','g8f6 g1f3 a7a6 e1g1 e7e6 d2d4 c7c5 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 b2b3 b8d7 c4e2 b7b6 c1b2 c8b7 a2a3 h8c8 b3b4 c5d6 d2c4 d6c7 a3a4 f6d5 b4b5 d7c5 f3d2 f7f6 b2d4 a6b5 a4b5 a8a1 d4c5 b6c5 f1a1 c8a8');
INSERT INTO evaluations VALUES(47328,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(47328,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(47328,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(47328,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(47328,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47328,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47328,36,'cp 35','g1f3 a7a6 e1g1 e7e6 d2d4 c7c5 b2b3 c5d4 f3d4 f8d6 c1b2 e8g8 b1d2 e6e5 d4f3 b8c6 d1b1 d8e7 f1d1 c8g4 d2e4 f6e4 b1e4 g4h5 a1c1 a8d8 h2h3 h5g6 e4h4 e5e4 h4e7 d6e7 f3e5 c6e5 b2e5 d8d1 c1d1 f8d8 d1d8 e7d8 e5b2');
INSERT INTO evaluations VALUES(47329,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(47329,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(47329,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(47329,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(47329,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(47329,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(47329,36,'cp 25','e7e6 e1g1 a7a6 d2d4 c7c5 b2b3 c5d4 f3d4 f8d6 c4e2 e6e5 d4f3 b8c6 b1d2 e8g8 c1b2 d6c7 d1c2 d8e7 a1c1 f8d8 a2a3 h7h6 f1d1 c8e6 b3b4 e6d5 h2h3 a8c8 e2c4 e7e6 e3e4 d5c4 c2c4 d8d7 c4e6');
INSERT INTO evaluations VALUES(48464,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(48464,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(48464,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(48464,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(48464,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(48464,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(48464,36,'cp 29','e1g1 a7a6 d2d4 c7c5 b2b3 c5d4 f3d4 f8d6 c1b2 e8g8 b1d2 e6e5 d4f3 d8e7 d1b1 b7b5 c4d3 b8d7 d2e4 c8b7 f3g5 f6e4 d3e4 b7e4 g5e4 d7f6 e4d6 e7d6 b1f5 f8e8 f1d1 d6e6 f5e6 e8e6 h2h3 a8c8');
INSERT INTO evaluations VALUES(57688,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(57688,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(57688,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(57688,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57688,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(57688,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(57688,36,'cp 10','c2c4 c7c5 g2g3 b8c6 f1g2 e7e6 b1c3 f8g7 c1e3 d7d6 d1d2 c6d4 a1b1 d4f5 e3g5 f7f6 g5f4 g6g5 f4e3 g8e7 b2b4 f5e3 d2e3 c5b4 b1b4 f6f5 b4b3 e7g6 h2h3 h7h6 g1f3 e8g8 e1g1 g8h8 e3d2');
INSERT INTO evaluations VALUES(57647,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57647,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(57647,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(57647,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(57647,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(57647,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57647,36,'cp 3','c7c5 b1c3 b8c6 g2g3 f8g7 f1g2 e7e6 c1e3 d7d6 d1d2 c6d4 a1d1 d4f5 e3g5 f7f6 g5e3 g8e7 g1f3 e6e5 d1b1 a7a5 a2a3 c8e6 e1g1 e8g8 b2b3 e7c6 f3e1 f5e3 f2e3 g8h8 e1c2 g7h6 c3d5');
INSERT INTO evaluations VALUES(6601,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(6601,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(6601,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(6601,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(6601,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(6601,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(6601,36,'cp 49','b4d2 d1d2 e5d4 f3d4 c6d4 d2d4 d8f6 e4e5 f6b6 d4b6 a7b6 b1c3 f7f6 c3d5 e8d8 e5f6 g8f6 d5f6 g7f6 a2a3 h8e8 e1d2 d7d6 f2f3 h7h5 a1e1 e8e1 d2e1 d8e7 h2h4 c8f5 c2c3 a8a5 f1e2 a5d5 g2g4 f5g6 f3f4 e7f8 e2f3 d5b5 b2b4 h5g4 a3a4 b5h5 f3g4 h5d5 h1h3 g6e4 h4h5 d5d3 h3d3');
INSERT INTO evaluations VALUES(6642,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(6642,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(6642,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(6642,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(6642,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(6642,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(6642,36,'cp 29','d1d2 e5d4 f3d4 c6d4 d2d4 d8f6 d4b4 f6b6 b4b6 a7b6 f2f3 d7d6 c2c4 f7f5 e4f5 c8f5 b1c3 e8c8 a1d1 g8f6 f1d3 f5d3 d1d3 d8e8 e1f2 h8f8 h2h4 f6g4 f2g3 g4e3 b2b3 e8e5 f3f4 e5e7 c3d5 e3d5 d3d5 h7h6');
INSERT INTO evaluations VALUES(6641,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(6641,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(6641,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(6641,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6641,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6641,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(6641,36,'cp 36','e5d4 f3d4 c6d4 d2d4 d8f6 e4e5 f6b6 d4c3 g8e7 f1c4 e8g8 e1g1 d7d6 e5d6 b6d6 f1e1 c8d7 b1d2 e7g6 a1d1 a8e8 d2f3 e8e1 c3e1 d6c6 d1d4 f8e8 e1b4 b7b6 h2h4 h7h6 b4b3 d7e6 b3d3 e6c4 d4c4 c6e6 c4c7 e6a2 d3d7 e8e7 d7c8 g8h7 c7e7 g6e7');
INSERT INTO evaluations VALUES(10124,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(10124,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(10124,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(10124,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10124,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(10124,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(10124,36,'cp 0','g1f3 d7d6 c2c3 a7a5 e1g1 h7h6 b1d2 e8g8 f1e1 c8e6 c4b5 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 b5a4 e7g6 d3d4 c7c6 a4c2 a5a4 b2b4 b7b5 d1d2 a7c7 e3f5 a8d8 f5g3 g6f4 e1e3 f4g6');
INSERT INTO evaluations VALUES(10393,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10393,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10393,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10393,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10393,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(10393,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(10393,36,'cp 0','b4d2 b1d2 e8g8 e1g1 d7d6 c4b3 c6a5 d3d4 e5d4 f3d4 f8e8 f1e1 a7a6 d1f3 c7c5 d4e2 c8g4 f3g3 g4e2 e1e2 d6d5 e4e5 c5c4 e5f6 d8f6 e2e8 a8e8 g3c3 f6c3 b2c3 c4b3 c2b3 g8f8 g1f1 e8c8 a1c1 a5c6 d2f3 h7h6 c1b1 c6e7 b1c1');
INSERT INTO evaluations VALUES(10435,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10435,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10435,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10435,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10435,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(10435,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(10435,36,'cp -1','b1d2 d7d6 c4b3 e8g8 e1g1 c6a5 d3d4 e5d4 f3d4 f8e8 b3a4 c8d7 a4d7 d8d7 f1e1 a5c6 c2c3 a8d8 d1c2 d6d5 d4c6 d7c6 e4e5 f6d7 d2f3 c6g6 c2d2 g6b6 a1d1 c7c6 e1e3 h7h6 h2h3 d7c5 d2c2 c5e4 e3e2 a7a6 c3c4 a6a5');
INSERT INTO evaluations VALUES(15806,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15806,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15806,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(15806,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(15806,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15806,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15806,36,'cp 0','d3c4 h7h6 g5f3 f8d6 c1e3 e8g8 b1c3 f8e8 a2a3 c8f5 e1g1 d8d7 f3h4 f5h7 d1e2 b7b6 h2h3 a7a6 a3a4 d6b4 c3a2 b4d6');
INSERT INTO evaluations VALUES(15803,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15803,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15803,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15803,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15803,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15803,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15803,36,'cp 0','h7h6 g5f3 f8d6 c1e3 e8g8 b1c3 f8e8 a2a3 c8f5 e1g1 d8d7 f3h4 f5h7 h4f3 h7g6 f3h4');
INSERT INTO evaluations VALUES(14681,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14681,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14681,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14681,36,'cp 18','g5f3 f8d6 c1e3 e8g8 b1c3 f8e8 a2a3 c7c6 d5c6 b7c6 c4c5 d6c7 d1d8 e8d8 e1g1 c8f5 f1c1 a7a5 h2h3 a8b8 a1b1 f6d5 c3d5 c6d5 b2b4 d5d4 e3d2 d8d5 f3h4 f5e6 c2c4 d4c3 d2c3 a5b4 a3b4 e5e4 g2g4 d5d3 b4b5');
INSERT INTO evaluations VALUES(14682,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(14682,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14682,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(14682,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14682,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14682,36,'cp 5','f8d6 b1c3 e8g8 c1e3 f8e8 a2a3 c8f5 e1g1 f5g4 c4c5 e5e4 c5d6 d8d6 f1e1 e4f3 h2h3 g4h5 g2g4 h5g6 d1f3 g6c2 e3d4 f6h7 f3g3 d6g3 f2g3 c2b3 e1e8 a8e8 d4a7 h7f6 d5d6 c7d6 a1c1 d6d5 a7d4 f6e4 c1e1 f7f6 e1e3 e4d6 e3e8 d6e8 g1f2 e8d6 h3h4 d6e4 f2f3 e4c3 d4c3 g8f7 a3a4 b3a4');
INSERT INTO evaluations VALUES(21188,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(21188,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(21188,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(21188,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(21188,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(21188,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(21188,36,'cp 87','d2d4 e7e5 g1e2 e5d4 e2d4 g8f6 f1b5 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 d8d7 h2h3 a7a5 c1e3 f8b8 b2b3 d6d5 e4e5 e7b4 e3d4 b4c3 d4c3 f6e4 a1d1 d7f5 a2a4 b8e8 c3b2 f7f6 e5f6 e4f6 f3g3');
INSERT INTO evaluations VALUES(21184,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(21184,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(21184,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(21184,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(21184,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(21184,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(21184,36,'cp 87','e7e5 g1e2 e5d4 e2d4 f8e7 d4c6 b7c6 f1d3 g8f6 e1g1 e8g8 f2f4 a7a5 g1h1 f6d7 c1e3 a8b8 b2b3 d6d5 e4d5 e7f6 d5c6 f6c3 c6d7 d8d7 a1c1 f8e8 e3d2 c3b4 d2b4 a5b4 f4f5 d7c6 d1f3 c6f3 f1f3 c8b7 f3g3 g8h8 f5f6 g7f6');
INSERT INTO evaluations VALUES(21945,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(21945,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21945,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(21945,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(21945,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(21945,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(21945,36,'cp 85','g1e2 e5d4 e2d4 g8f6 f1b5 c8d7 e1g1 c6d4 b5d7 d8d7 d1d4 f8e7 f1d1 d7c6 b2b3 c6b6 d4b6 a7b6 a2a4 e8c8 c1e3 f6d7 c3d5 e7f6 a1b1 h8e8 f2f3 f6e5 c2c4 d7c5 g2g4 h7h6 h2h3 g7g5 e3f2 c5e6 g1g2 e6f4 d5f4 e5f4');
INSERT INTO evaluations VALUES(21692,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(21692,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(21692,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(21692,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(21692,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(21692,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(21692,36,'cp 141','f3e5 c6e5 d4e5 c8e6 b2b3 f8b4 c1d2 a7a5 a2a3 b4c5 d2c3 g8e7 f1d3 e7c6 e1g1 d5d4 e3d4 c5d4 d3b5 d4b6 b1d2 e8g8 b5d3 d8h4 d1e2 c6d4 c3d4 h4d4 d2c4 b6c5 g1h1 d4f4 g2g3 f4d4 a1d1 d4c3 d3e4 a5a4 e4b7 a8d8');
INSERT INTO evaluations VALUES(31837,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(31837,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(31837,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(31837,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(31837,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(31837,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(31837,36,'cp -72','g8f6 e2e3 e7e6 f1d3 f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 e8g8 e1g1 f6e4 c2c4 b8d7 c1a3 e7a3 b1a3 d7f6 a3b1 f8e8 b1c3 c7c6 h2h3 e4c3 d3c3 c8e6 f1e1 f6e4 c3b2 a7a5 a1c1 a5a4 f3e5 f7f6 e5d3 a4b3 b2b3');
INSERT INTO evaluations VALUES(46240,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46240,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46240,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(46240,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(46240,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(46240,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(46240,36,'cp -8','b7b6 d1e2 f8e7 g1f3 e8g8 b2b3 c8b7 e1g1 b8d7 c1b2 a8c8 f1d1 d8c7 h2h3 f8d8 a1c1 c7b8 d3b1 b8a8 b1c2 e7f8 f3e5 c8c7 a2a4 d8c8 d2f3 c5d4 e3d4 d7e5 f3e5 f6e4 c2d3 e4c3 b2c3 c7c3');
INSERT INTO evaluations VALUES(47183,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(47183,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(47183,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(47183,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(47183,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(47183,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(47183,36,'cp -110','c7c5 c1f4 e7e6 e2e3 f8e7 b1d2 b8c6 f4g3 h7h5 g1h3 c8d7 f1e2 h5h4 g3e5 d8b6 d1b3 c6e5 d4e5 f6g8 c3c4 d7c6 e1c1 g8h6 c4d5 e6d5 h3f4 e8c8 g2g4 c5c4 b3b6 a7b6 e3e4 d5e4 d2e4 b6b5 e4d6 e7d6 e5d6');
INSERT INTO evaluations VALUES(52976,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(52976,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(52976,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(52976,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(52976,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(52976,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(52976,36,'cp 52','e7e5 d2d4 b8d7 f1c4 f8e7 a2a4 g8f6 e1g1 e8g8 f1e1 e5d4 f3d4 d6d5 e4d5 d7b6 c4b3 b6d5 c3d5 f6d5 b3d5 c6d5 c1e3 e7f6 c2c3 f8e8 h2h3 c8d7 a4a5 h7h6 d4c2 d7b5 e3d4 e8e1 c2e1 a7a6 e1c2 f6g5 d1g4 d8d7 g4d7 b5d7');
INSERT INTO evaluations VALUES(5323,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(5323,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5323,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(5323,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5323,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(5323,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5323,36,'cp 0','e7e6 g1e2 g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 c3d5 e6d5 e1g1 e8g8 c1e3 c8g4 f1e1 f8e8 h2h3 g4f5 e2c3 f5e6 c3e2');
INSERT INTO evaluations VALUES(5127,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5127,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5127,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5127,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5127,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5127,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(5127,36,'cp 4','g1e2 g8e7 e1g1 e8g8 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 f1e1 f8e8 c1d2 c8d7 c3d5 e6d5 d2c3 d7g4 h2h3 g4e2 e1e2 e8e2 d1e2 g7d4 c3d4 c6d4 e2e5 d4c6 e5d5 d8b6 b2b3 a8d8 d5c4 d8d2 c4f4 b6d8 g2c6 b7c6 f4e5 d2c2 e5e3 c6c5 a1e1 c2a2');
INSERT INTO evaluations VALUES(5128,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(5128,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5128,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5128,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5128,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5128,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(5128,36,'cp 2','g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 c3d5 e6d5 c1e3 e8g8 e1g1 c8g4 d1d2 g4e2 d2e2 g7d4 e3d4 c6d4 e2d2 d8f6 a1d1 d4e6 g2d5 a8d8 d2e3 b7b6 d5e6 f6e6 e3e6 f7e6 h2h4 g8f7 d1c1 f7e7 c1c7 d8d7 c7d7 e7d7 b2b3 f8c8 f1e1 c8c2');
INSERT INTO evaluations VALUES(7974,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(7974,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7974,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(7974,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7974,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7974,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(7974,36,'cp 5','d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 e1g1 e8g8 c3d5 e6d5 c1e3 c8g4 d1d2 g4e2 d2e2 g7d4 e3d4 c6d4 e2d2 d4f5 a1c1 d8b6 d2d5 f5d4 c1c4 f8d8 d5b7 b6b7 g2b7 a8b8 c4c7 d4e6 c7e7 a7a5 f1c1 d8d2 g1g2 d2b2 b7d5');
INSERT INTO evaluations VALUES(9475,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(9475,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(9475,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(9475,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(9475,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(9475,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(9475,36,'cp -50','c8g4 h2h3 g4h5 d2d3 d8e7 d1e2 h7h6 c1e3 c5d6 g2g4 h5g6 c3b1 f6d7 b1d2 h6h5 d2c4 h5g4 h3g4 f7f6 g1g2 e7e6 f3d2 e8c8 f1h1 d6e7 h1h8 d8h8 a1h1 h8h1 g2h1 b7b6 e2f3 g6f7 d2f1 c8b7 a2a4 a7a5');
INSERT INTO evaluations VALUES(24160,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(24160,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(24160,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(24160,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(24160,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(24160,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(24160,36,'cp -42','d2d3 f6d7 a2a3 f7f6 c1e3 d8e7 c3a4 c5d6 c2c4 c6c5 a4c3 c7c6 h2h3 g4f3 d1f3 d7f8 c3e2 f8e6 f1b1 e8g8 h3h4 g7g6 b2b4 c5b4 a3b4 d6b4 e3a7 b4c5 a7c5 e6c5 h4h5 g8g7 f3e3 f8d8 d3d4 e5d4 e2d4 c5e4 b1e1 a8a1');
INSERT INTO evaluations VALUES(16832,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(16832,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16832,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(16832,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(16832,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(16832,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16832,36,'cp -38','e4f6 f1e1 f8e7 f3e5 c6e5 e1e5 d7d5 c4b3 e8g8 h2h3 c7c6 e5e1 e7d6 c2c4 d5d4 b1d2 a7a5 b3c2 h7h6 d2e4 d6c7 d1f3 f6e4 d3e4 c7e5 f3d3 d8h4 e1f1 g7g5 f2f4 g5f4 d3f3 c8e6 c2d3 a5a4 c1f4 h4f4 f3f4 e5f4 f1f4');
INSERT INTO evaluations VALUES(16421,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(16421,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(16421,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(16421,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(16421,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(16421,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(16421,36,'cp -48','f1e1 f8e7 f3e5 c6e5 e1e5 d7d5 c4b3 e8g8 h2h3 e7d6 e5e1 h7h6 c1d2 c7c6 d2c3 c8e6 b1d2 f8e8 d2f3 a7a5 a2a4 c6c5 c3e5 d6e5 f3e5 d8c7 d1e2 e8e7 f2f4 a8d8 e2f2 d8e8 a1d1 c5c4 b3a2 c4d3 c2d3 e6f5 d1c1 c7d6 f2d4');
INSERT INTO evaluations VALUES(22608,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22608,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(22608,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(22608,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(22608,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(22608,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(22608,36,'cp 65','g1f3 d7d5 c2c4 g8f6 b1c3 f8e7 a2a3 e8g8 b2b4 d5c4 f1c4 e7d6 e1g1 e6e5 h2h3 e5d4 e3d4 h7h6 f1e1 c6e7 f3e5 c8e6 c4e6 f7e6 c1f4 e7d5 c3d5 f6d5 f4g3 a7a5 b4b5 d8g5 d1b3 g5f5 b3b2 d5f6 a1c1 f6d5 g3h2');
INSERT INTO evaluations VALUES(22609,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(22609,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(22609,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(22609,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(22609,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(22609,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(22609,36,'cp 66','d7d5 a2a3 g8f6 b2b4 c6b8 c2c4 a7a5 b4b5 c7c5 c1b2 b8d7 b1d2 c5d4 e3d4 d5c4 d2c4 d7b6 c4b6 d8b6 f1d3 f8d6 e1g1 e8g8 a3a4 d6b4 f3e5 c8d7 d1f3 f8d8 h2h4 d7e8 g2g4 a8c8 g4g5 f6d5 a1c1 b4c3 b2c3');
INSERT INTO evaluations VALUES(25493,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25493,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(25493,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25493,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25493,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25493,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25493,36,'cp 0','f8e7 a2a4 c6b4 e1g1 c7c6 b5c4 e8g8 c4b3 h7h6 g5d2 b4a6 d2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(24188,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(24188,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(24188,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24188,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(24188,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(24188,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24188,36,'cp 12','a2a4 a7a6 b5c4 d7e6 g5f6 e7f6 c3d5 c6e7 d5e7 d8e7 c4e6 e7e6 e1g1 g7g6 c2c3 e8g8 d1c2 a8d8 f1e1 f6g7 e1e3 e6e7 e3e2 c7c6 a1e1 e7c7 g2g3 d6d5 g1g2 f8e8 b2b4 e8e7 e1b1 b7b5 h2h3');
INSERT INTO evaluations VALUES(38451,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(38451,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(38451,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(38451,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(38451,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(38451,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(38451,36,'cp 80','c4d5 g8f6 d1a4 c7c6 d5c6 b8c6 b1c3 c8f5 e2e3 a7a6 a4b3 b7b5 f1e2 a8c8 e1g1 e8g8 h2h3 b5b4 c3a4 a6a5 a2a3 e7e5 d4e5 c6e5 f3d4 f6e4 d4f5 g6f5 a3b4 a5b4 b3b4 c8c2 b4e1 e5d3 e1d1 d3b4 e2f3 d8h4 f3e4 f5e4 a4c3 g7c3 b2c3 f8d8 d1g4 h4g4 h3g4 b4d3 c1a3 c2c3');
INSERT INTO evaluations VALUES(38445,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(38445,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(38445,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(38445,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(38445,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(38445,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(38445,36,'cp 80','g8f6 d1a4 c7c6 d5c6 b8c6 b1c3 c8f5 e2e3 e8g8 f1e2 a7a6 e1g1 b7b5 a4b3 a8c8 h2h3 f8e8 f1d1 c6a5 b3a3 f5c2 d1d2 c2f5 f3e5 e7e6 b2b4 a5c4 e5c4 b5c4 e2f3 g6g5 a3a6 g5g4 h3g4 f6g4 f3g4 f5g4 c1b2 d8h4 c3e2 g4e2 d2e2');
INSERT INTO evaluations VALUES(49740,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(49740,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(49740,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(49740,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(49740,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(49740,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(49740,36,'cp 68','f3g5 c7c6 g5e4 f6e4 c3e4 d7d5 c4d5 c6d5 e4g3 b8c6 d2d4 h7h5 e2e3 f8d6 a2a3 h5h4 g3e2 d8f6 e2c3 c8f5 d1f3 f6g6 c1d2 e8c8 f1e2 c8b8 a1c1 c6a5 c3d5 a5b3 c1c3 f5e4 d5f4 g6h7 f3g4 b3d2 e1d2 g7g5 g4g5 h4h3 f4h3 h8g8 g5b5 e4g2 h3g5');
INSERT INTO evaluations VALUES(49721,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(49721,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(49721,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(49721,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(49721,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(49721,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(49721,36,'cp 62','c7c6 g5e4 f6e4 c3e4 d7d5 c4d5 c6d5 e4g3 b8c6 e2e3 h7h5 d2d4 f8d6 a2a3 h5h4 g3e2 c8f5 c1d2 d8f6 e2c3 f6g6 d1f3 e8c8 h2h3 c6a5 a1d1 f5c2 f3d5 d6c7 d5f3 c8b8 f3g4 g6h7 c3b5 c7b6 d2a5 b6a5 b2b4 c2d1 g4d1 a5b6 f1d3 g7g6 e1g1 f7f5 a3a4 h8e8 a4a5 b6c7 d1f3 h7e7 b5c7');
INSERT INTO evaluations VALUES(52117,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52117,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(52117,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(52117,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(52117,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(52117,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(52117,36,'cp 33','b8c6 a2a3 g7g6 b1c3 a7a5 d2d3 d7d6 c3d5 h7h6 c1d2 g8f6 e2e3 f8g7 g1e2 e8g8 d5f6 d8f6 d1b3 c8f5 e2c3 a8b8 c3d5 f6d8 e1g1 g8h7 f1d1 b7b6 b3c2 c6e7 b2b4 f5g4 d1b1 e5e4 g2e4 c5b4 d5e7 g7a1');
INSERT INTO evaluations VALUES(18995,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(18995,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(18995,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(18995,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(18995,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(18995,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(18995,36,'cp 42','a2a3 a7a5 b1c3 d7d6 d2d3 g7g6 c3d5 h7h6 c1d2 g8f6 e2e3 f8g7 g1e2 e8g8 d5f6 d8f6 e2c3 c8e6 e1g1 f6d8 a1b1 g8h7 d1a4 f7f5 a4c2 h6h5 c2d1 d8d7 c3d5 a8b8 b2b4 a5b4 a3b4 e6d5 c4d5 c6b4 d2b4');
INSERT INTO evaluations VALUES(53547,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53547,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53547,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(53547,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(53547,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(53547,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(53547,36,'cp 14','e6d5 f1d3 f8d6 c1g5 g8e7 c2c4 h7h6 g5h4 d5c4 d3c4 d8c7 e1g1 e8g8 h4e7 c7e7 f1e1 e7c7 d1d3 c8g4 b1d2 b8d7 h2h3 g4f3 d2f3 d7f6 e1e2 a8e8 a1e1 e8e2 e1e2 c7d7 a2a3 f8e8 e2e8 f6e8 d3g6 d7e7 g2g3 e8c7');
INSERT INTO evaluations VALUES(7313,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7313,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7313,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7313,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7313,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7313,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(7313,36,'cp 0','d2d3 d7d6 e1g1 e8g8 c1e3 c5e3 f2e3 c8e6 c3d5 e6d5 e4d5 c6e7 e3e4 c7c6 d5c6 b7c6 b2b4 a6a5 d1e1 a5b4 e1b4 f6g4 a3a4 g4e3 f1f2 e3g4');
INSERT INTO evaluations VALUES(7314,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7314,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7314,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7314,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7314,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7314,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7314,36,'cp 0','d7d6 e1g1 e8g8 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d3d4 d8e8 d1d3 h7h6 h2h3 e5d4 e3d4 e6e5 d4e5 c6e5 f3e5 e8e5 d3c4 g8h8 c4c7 e5d4 g1h2 d4e5');
INSERT INTO evaluations VALUES(14425,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(14425,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(14425,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(14425,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(14425,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(14425,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(14425,36,'cp 122','c2c4 d5f6 b1c3 e7e6 a2a3 a7a6 f1e2 g7g6 c1f4 f8g7 d1d2 e8g8 a1d1 b7b5 c4b5 a6b5 c3b5 f6d5 f4g3 c6e7 e1g1 e7f5 g3e5 g7h6 d2c2 c8b7 e2d3 d8d7 b5c3 d5c3 c2c3 f5e7 c3c7 d7c7 e5c7 f8c8 c7g3 b7d5 f1e1 e7f5');
INSERT INTO evaluations VALUES(14416,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(14416,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(14416,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(14416,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(14416,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(14416,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(14416,36,'cp 114','d5f6 b1c3 e7e6 a2a3 a7a6 f1e2 g7g6 c1f4 f8g7 d1d2 e8g8 a1d1 b7b5 c4b5 a6b5 e2b5 c8b7 e1g1 c6a5 d2e2 f6d5 f4g3 a5b3 f3e5 d5e7 e2c4 b3a5 c4a2 e7d5 c3e4 d5e7 e4c5 b7d5 b2b3 a5c6 b5c6 e7c6 a2a1 d8e7 e5c6 d5c6 a3a4');
INSERT INTO evaluations VALUES(14370,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(14370,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(14370,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(14370,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(14370,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(14370,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(14370,36,'cp 120','b1c3 e7e6 a2a3 a7a6 f1e2 g7g6 c1f4 f8g7 d1d2 e8g8 a1d1 b7b5 c4b5 a6b5 e2b5 c8b7 b5e2 f6d5 f4g5 f7f6 g5e3 c6e7 e1g1 e7f5 c3d5 b7d5 d1c1 d8d7 f1e1 c7c6 h2h3 f8d8 c1c5 g7f8 c5c3 f8d6 d2c1 a8a7 e3d2 d8a8 e2c4 g8g7 b2b4 h7h5 c3b3 d7b7 e1e6 d5e6 c4e6');
INSERT INTO evaluations VALUES(16336,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(16336,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(16336,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(16336,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(16336,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(16336,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(16336,36,'cp 79','c6b8 d1f3 d7d6 h2h3 h7h5 c1e3 a7a6 a2a4 b8d7 a4a5 f6g4 e3d2 g4f6 f1d3 c7c6 g1e2 c6d5 e4d5 g7g6 d2g5 f8e7 e1g1 e8g8 f3g3 f6d5 g5h6 h5h4 g3f3 d5f6 h6f8 d7f8 d3c4 g8g7 f1d1 a8b8 c3d5');
INSERT INTO evaluations VALUES(16724,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(16724,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(16724,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(16724,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(16724,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(16724,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(16724,36,'cp -66','f8b4 f1d3 d7d5 g1e2 d5e4 d3b5 c8d7 c1f4 a7a6 b5c6 b4c3 e2c3 d7c6 a4a5 a8c8 a5e5 d8e7 e1c1 e7e5 f4e5 e8f8 e5f6 g7f6 c3d5 h8g8 g2g3 f6f5 d1d4 g8g6 h1d1 g6h6 h2h4 c6d5 d4d5 h6f6 d5e5 c8e8 d1d5 e8e5 d5e5 f8g7 c1d2 g7g6 e5c5 f6b6 c5c7 b6b2 d2e3 b2a2 c7b7 a2a3 e3d2 a6a5 b7a7 f7f6 a7a6');
INSERT INTO evaluations VALUES(10533,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(10533,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(10533,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(10533,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(10533,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(10533,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(10533,36,'cp -20','c1d2 e8g8 e1c1 d7d6 a2a3 b4a5 g1f3 c8d7 f1d3 a7a6 a4b3 b7b5 c3d5 a5d2 d1d2 f8e8 h1e1 d7e6 d2d1 h7h6 h2h3 a8b8 d3f1 e6d5 e4d5 e8e1 f3e1 c6e5 e1d3 f6e4 d3e5 d8g5 b3e3 g5e3 f2e3 d6e5 f1e2 g8f8 e2f3');
INSERT INTO evaluations VALUES(21237,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(21237,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(21237,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(21237,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(21237,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(21237,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21237,36,'cp 89','e2e3 g8f6 g1f3 e7e6 c2c4 f8b4 b1d2 c7c5 a2a3 b4d2 d1d2 f6e4 d2d1 d8a5 f3d2 e4d2 d1d2 a5d2 e1d2 c5d4 e3d4 d5c4 f1c4 e8g8 h1c1 a6b8 d4d5 e6d5 c4d5 b8c6 c1c6 b7c6 d5c6 c8d7 c6a8 f8a8 a1c1 a7a5 c1c7 d7e6 b2b4 a5b4 a3b4 a8a2 d2c3 a2f2 f4d6 f2f1 b4b5');
INSERT INTO evaluations VALUES(21235,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(21235,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(21235,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(21235,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(21235,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(21235,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21235,36,'cp 75','g8f6 g1f3 e7e6 c2c4 f8b4 b1d2 c7c5 a2a3 b4d2 d1d2 f6e4 d2c2 d8a5 f3d2 e4d2 c2d2 a5d2 e1d2 c5d4 e3d4 d5c4 f1c4 e8g8 b2b4 a6b8 b4b5 b8d7 h1c1 d7f6 a3a4 f8d8 f4e5 c8d7 a4a5 a8c8 c4d3 f6d5 e5g3 d5b4 d3e2 d7e8 c1c8 d8c8 a1c1 c8c1');
INSERT INTO evaluations VALUES(23074,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(23074,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(23074,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(23074,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(23074,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(23074,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(23074,36,'cp 84','e5g6 f4g3 g8f6 f1d3 d7d6 h2h3 f8e7 e1g1 e8g8 a2a4 c7c5 d4f5 c8f5 e4f5 g6e5 g3e5 d6e5 d1e2 d8c7 b1d2 a8d8 a4a5 f6d5 g2g3 d5f6 d2e4 f6e4 d3e4 c5c4 c2c3 e7c5 h3h4 b7b5 a5b6 c5b6 a1a6');
INSERT INTO evaluations VALUES(22995,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(22995,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(22995,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(22995,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(22995,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(22995,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(22995,36,'cp 112','f4g3 g8f6 f1d3 d7d6 e1g1 f8e7 a2a4 e8g8 h2h3 c7c5 d4f5 c8f5 e4f5 g6e5 g3e5 d6e5 d1e2 d8d4 b1c3 a8d8 g2g3 a7a6 a4a5 d4d7 e2e5 c5c4 d3c4 e7d6 e5e2 d7f5 g1g2 f8e8 c4d3 f5c5 e2f3 d6c7 f3b7 c7a5 b7f3 a5c3');
INSERT INTO evaluations VALUES(29587,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29587,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29587,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(29587,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29587,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29587,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29587,36,'cp 4','c2c4 e7e6 b1c3 f8b4 c4d5 f6d5 c1d2 d5c3 d2c3 b4c3 b2c3 c7c5 e2e3 d8a5 d1d2 e8g8 f1e2 b8c6 e1g1 f8d8 d2b2 c5d4 c3d4 a5b4 b2b4 c6b4 f1c1 a8c8 f3d2 f5c2 a2a4 c8c7 d2c4 c2d3 e2d3 b4d3 c1c3 d8c8 c3d3 c7c4');
INSERT INTO evaluations VALUES(38131,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38131,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(38131,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38131,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(38131,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38131,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38131,36,'cp 0','a2a3 g8f6 c2c4 e8g8 c4d5 e6d5 f1e2 b8c6 b2b4 c5b6 c1b2 c8f5 b4b5 c6a5 b1d2 f6e4 e1g1 e4d2 d1d2 a8c8 a1c1 a5c4 d2c3 c4b2 c3b2 d8e7 a3a4 f5e4 f3d2 e4f5');
INSERT INTO evaluations VALUES(38133,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38133,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38133,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38133,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38133,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38133,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38133,36,'cp 0','g8f6 c2c4 e8g8 c4d5 e6d5 f1e2 b8c6 b2b4 c5b6 c1b2 c8f5 e1g1 a8c8 h2h3 f8e8 e2b5 a7a6 b5c6 b7c6 b1d2 h7h6 a1c1 a6a5 d2b3 f6e4 b3a5 b6a5 b4a5 d8a5 f3d4 f5d7 d4b3 a5a6 b3c5 e4c5 c1c5 c8d8');
INSERT INTO evaluations VALUES(40619,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40619,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40619,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40619,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40619,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40619,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40619,36,'cp 0','c2c4 e8g8 c4d5 e6d5 f1e2 b8c6 b2b4 c5b6 c1b2 c8f5 b4b5 c6a5 b1d2 f6e4 e1g1 e4d2 d1d2 a8c8 a1c1 a5c4 d2c3 c4b2 c3b2 d8e7 a3a4 f5e4 f3d2 e4f5');
INSERT INTO evaluations VALUES(47406,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(47406,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(47406,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(47406,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(47406,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(47406,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(47406,36,'cp -11','c3e4 b8c6 g1f3 e7e5 c2c3 a7a5 e4f6 d8f6 f1e2 f8c5 e1g1 e8g8 c1e3 c5e3 f2e3 f6h6 d1d2 e5e4 d3e4 f8d8 d2e1 h6e3 e1f2 e3f2 g1f2 d8e8 e2c4 c8e6 c4b5 a5a4 f1d1 a8a5 f3d4 e6d7 b2b3 a4b3 a2b3 a5a1 d1a1 e8e4 a1a8 e4e8');
INSERT INTO evaluations VALUES(47392,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(47392,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(47392,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(47392,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(47392,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(47392,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(47392,36,'cp -22','b8c6 c2c3 e7e5 g1f3 a7a5 e4f6 d8f6 f1e2 f8c5 e1g1 e8g8 f3d2 f6e7 d2e4 c5a7 c1e3 a7e3 f2e3 f7f5 e4f2 e7d6 d3d4 c8e6 d1a4 c6e7 d4e5 d6e5 e3e4 e7g6 e4f5 e5e2 f5e6 e2b2 a1b1 b2c3 a4e4 a8e8 b1c1 c3f6 f2d3 f6e6 e4e6 e8e6 c1c7 e6e2 f1f8 g6f8 c7b7 e2a2');
INSERT INTO evaluations VALUES(56198,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(56198,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56198,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56198,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(56198,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(56198,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(56198,36,'cp 11','f8d6 f4d6 d8d6 b1c3 c8f5 g1f3 g8f6 f3h4 f5g6 e2e3 b8d7 d1b3 e8g8 f1e2 a7a5 a2a4 c7c6 e1g1 d6b4 b3a3 d7b6 h4g6 h7g6 c3b1 b6c4 a3b4 a5b4 b2b3 c4d6 a4a5 f8c8 b1d2 g6g5 a5a6 b7a6 a1a6 a8a6');
INSERT INTO evaluations VALUES(5774,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(5774,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(5774,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(5774,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(5774,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(5774,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(5774,36,'cp 80','d4e5 d8e7 e5d6 e7e4 c4e2 e4g6 e1g1 g6d6 d1d6 c7d6 b1d2 d6d5 f1e1 g8f6 e2f1 e8d8 e1d1 d8e7 d2b3 h8e8 a2a4 a7a6 h2h3 e7f8 a4a5 b6a7 b3d4 h7h6 c1e3 c8e6 f3e1 e6d7 d4c2 a8d8 e3a7 c6a7 c2d4 f6e4 f2f3 e4d6 e1c2 d6c4 f1c4');
INSERT INTO evaluations VALUES(5769,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(5769,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(5769,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(5769,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(5769,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(5769,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(5769,36,'cp 80','d8e7 e5d6 e7e4 c4e2 e4g6 e1g1 g6d6 d1d6 c7d6 b1d2 d6d5 f1e1 g8f6 e2f1 e8d8 e1d1 d8e7 d2b3 h8e8 a2a4 a7a6 h2h3 e7f8 a4a5 b6a7 b3d4 h7h6 c1e3 c8e6 g2g3 e6d7 d4c2 a7e3 c2e3 d7e6 f3d4 c6d4 d1d4 a8d8 e3c2 e6f5');
INSERT INTO evaluations VALUES(22756,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(22756,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(22756,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(22756,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(22756,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(22756,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(22756,36,'cp -31','d7d5 g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 a2a3 a7a6 c4d5 f6d5 f1c4 d8d7 c4d5 e6d5 c3d5 d7d5 d1b3 e8c8 b3d5 d8d5 h2h4 h7h5 e1e2 h8h6 g2g3 d5b5 b2b4 f8d6 d2b1 c6e7 b1c3 b5b6 c1d2 c8d7 h1c1 b6c6 c1c2');
INSERT INTO evaluations VALUES(25950,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(25950,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(25950,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(25950,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(25950,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(25950,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(25950,36,'cp 0','d8a5 d4c5 a5c7 b2b4 c6e5 c1f4 e5d3 f1d3 c7f4 b1d2 g8e7 c3c4 g7g6 e1g1 f8g7 f1e1 e8g8 a1c1 g7b2 c1b1 b2g7 b1c1');
INSERT INTO evaluations VALUES(26296,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(26296,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(26296,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(26296,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(26296,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(26296,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(26296,36,'cp -30','e7e6 h2h3 g4h5 c1f4 f8d6 f4d6 d8d6 c3b5 d6b8 c2c3 g8f6 a2a4 b8d8 b5a3 e8g8 e1g1 d8b6 a1b1 h7h6 f1e1 a8c8 f3h4 h5e2 e1e2 a7a6 h4f3 c8c7 f3e1 c6a5 e1d3 f6e4 a3c2 b6b3 b1a1 a5c4 c2e3 b3d1 e3d1 e4d6 d1e3');
INSERT INTO evaluations VALUES(26422,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(26422,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(26422,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(26422,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(26422,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(26422,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(26422,36,'cp -18','h2h3 g4h5 c1f4 f8d6 f4d6 d8d6 c3b5 d6b8 c2c3 g8f6 a2a4 b8d8 b5a3 e8g8 e1g1 d8b6 a1b1 a7a6 a3c2 h7h6 c2e3 h5g6 e2d3 f6e4 d1e2 f7f6 f3h4 g6h7 h4f3 f6f5 f3d2 g8h8 d2f3 a8d8 f1e1 b6c7 b1c1 h7g6 f3h4');
INSERT INTO evaluations VALUES(26420,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(26420,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(26420,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(26420,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(26420,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(26420,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(26420,36,'cp -17','g4h5 c1f4 f8d6 f4d6 d8d6 e1g1 g8f6 c3b5 d6d8 c2c3 e8g8 f3e5 h5g6 e5g6 h7g6 b5a3 d8b6 a1b1 b6c7 a3c2 f8e8 a2a3 f6e4 c2e3 f7f5 e2d3 a8d8 d1c2 g8h7 b1d1 a7a6 f1e1 e6e5 d3e4 d5e4 d4d5');
INSERT INTO evaluations VALUES(26417,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(26417,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(26417,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(26417,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(26417,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(26417,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(26417,36,'cp -23','c1f4 f8d6 f4d6 d8d6 c3b5 d6d8 e1g1 g8f6 c2c3 h5f3 e2f3 d8b6 d1e2 e8g8 b5a3 a8c8 a1b1 b6a5 b1a1 g7g6 a3c2 g8g7 e2d2 b7b5 f3e2 f6e4 d2f4 a5b6 f1d1 a7a5 a1b1 b5b4 c3c4 b4b3 a2b3');
INSERT INTO evaluations VALUES(37980,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(37980,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(37980,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(37980,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(37980,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(37980,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(37980,36,'cp 58','g7c3 b2c3 e7e5 g1f3 d8e7 g2g3 g8f6 f3d2 b8d7 f1g2 h7h5 h2h4 f6h7 a2a4 d7f6 d1e2 e8g8 e1g1 b7b6 d2f3 c8g4 e2c2 g8g7 f3h2 f8e8 f1e1 e7c7 h2f1 g4d7 f2f3 f6g8 c2e2 c7d8 f1e3 g7h8 g1h2 g8h6');
INSERT INTO evaluations VALUES(41465,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41465,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(41465,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41465,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(41465,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(41465,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(41465,36,'cp 16','b1d2 e8g8 f1e2 b8d7 h2h3 c7c5 g1f3 b7b6 a2a4 a7a6 e2d3 c8b7 e1g1 c5c4 d3c2 b6b5 d1e2 b7c6 f1e1 d8b6 e3e4 d5e4 d2e4 b6b7 f3e5 d7e5 d4e5 f6e4 c2e4 a8d8 a4b5 a6b5 e2f3 c6e4 f3e4 b7e4 e1e4');
INSERT INTO evaluations VALUES(45654,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(45654,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(45654,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(45654,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(45654,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(45654,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(45654,36,'cp -44','d5c4 e2e4 b7b5 b2b3 c4b3 d1b3 a7a5 f1d3 f8e7 e1g1 c8a6 e4e5 a5a4 b3b1 f6d5 c1d2 d5c3 d2c3 h7h6 f1c1 e8g8 b1c2 a6b7 c2e2 c6c5 d3b5 b8d7 h2h3 d7b6 b5d3 d8d5 c1b1 b6d7 d3b5 b7c6 b5c6 d5c6');
INSERT INTO evaluations VALUES(49983,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(49983,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(49983,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(49983,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(49983,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(49983,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(49983,36,'cp 176','f1c4 d8e7 d2d4 b8c6 e1g1 f5e4 e5g4 f6g4 d1g4 c6d4 b1c3 c7c6 g4h5 g7g6 h5d1 f8g7 c3e4 d7d5 c1g5 e7e5 f2f4 e5e4 f1e1 e4e6 f4f5 d4f5 e1e6 c8e6 d1e1 e8f7 c4d3 h8e8 g2g4 f5d6 e1b4 d6e4 d3e4 d5e4 a1e1 b7b6 e1e4 e6d5 e4f4 f7g8 g5f6 g7f8 b4d2 f8c5 f6d4 c5d6 f4f2 a8d8 g4g5 b6b5');
INSERT INTO evaluations VALUES(49982,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(49982,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(49982,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(49982,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(49982,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(49982,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(49982,36,'cp 200','d8e7 d2d4 b8c6 e1g1 f5e4 e5g4 d7d5 g4f6 e7f6 c4d5 c8f5 c1e3 e8c8 b1c3 c8b8 d1h5 g7g6 h5g5 f6g5 e3g5 f8e7 g5e3 e7b4 d5c6 b4c3 b2c3 b7c6 a2a4 b8c8 a4a5 a7a6 h2h3 c8d7 c3c4 d8b8 d4d5 h8e8 f1d1 b8b2 d1d2 b2b8 d5c6 d7c6 d2d5');
INSERT INTO evaluations VALUES(50107,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(50107,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(50107,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(50107,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(50107,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(50107,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(50107,36,'cp 56','d7d5 c2c4 e7e6 g1h3 f8d6 b1c3 c7c6 d1c2 e8g8 e1g1 a7a5 b2b3 f8e8 f1d1 b8d7 c1f4 d6b4 a2a3 b4f8 h3g5 f6e4 g5f3 g7g5 f4c1 h7h6 c1b2 d8e7 a1c1 e4c3 c2c3 b7b6 h2h4 c8b7 c4d5 c6d5 c3d3 a8c8 c1c8');
INSERT INTO evaluations VALUES(46603,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(46603,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(46603,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(46603,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(46603,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(46603,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(46603,36,'cp 55','c2c4 e7e6 g1h3 f8d6 b1c3 c7c6 d1c2 e8g8 e1g1 a7a5 b2b3 f8e8 f1d1 b8d7 c1f4 d6f8 h3g5 h7h6 g5f3 g7g5 f4c1 f6e4 c1b2 e4c3 b2c3 b7b5 g3g4 d7f6 f3e5 f6g4 e5g4 f5g4 c2g6 f8g7 c4b5 c6b5 a1c1 b5b4 c3b2 c8d7 a2a3 b4a3');
INSERT INTO evaluations VALUES(50568,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(50568,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(50568,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(50568,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(50568,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(50568,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(50568,36,'cp 76','c3d4 f8b4 b1c3 d7d5 e4e5 c8d7 f1d3 d7b5 d3c2 g8e7 a2a3 b4c3 b2c3 a6a5 c1d2 b8d7 f3g5 h7h6 g5h3 d8c7 h3f4 d7b6 a3a4 b5a6 f4d3 b6c4 d2c1 b7b6 e1g1 e8c8 a1b1 e7f5 d1f3 f5h4 f3e2 h4f5 f1e1 g7g5 h2h3 d8g8 d3b2 c8b7');
INSERT INTO evaluations VALUES(52417,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(52417,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(52417,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(52417,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(52417,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(52417,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(52417,36,'cp -48','d7d5 c4b5 c8d7 b5d7 d8d7 g1f3 b8c6 d2d4 g8f6 d1e2 e7e6 e1g1 c5d4 e3d4 f8d6 b2b3 e8g8 c2c4 f6e4 c1b2 f7f5 b1c3 d7f7 c3b5 d6b8 f3e5 c6e5 d4e5 a7a6 b5c3 f7c7 c3e4 d5e4 b2d4 b8a7 d4a7 a8a7 f2f3 a7a8 f3e4 c7e5');
INSERT INTO evaluations VALUES(53239,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(53239,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(53239,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(53239,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(53239,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(53239,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(53239,36,'cp -37','c4b5 c8d7 b5d7 d8d7 g1f3 b8c6 d2d4 e7e6 d4c5 g8f6 c2c4 d5c4 d1e2 f8c5 e2c4 c5b6 e1g1 a8c8 c4a4 e8g8 f1d1 c6d4 a4d7 d4f3 g2f3 f6d7 c1d2 f8d8 d2c3 d7f6 d1c1 f6d5 g1f1 g8f8 f1e2 d5c3 b1c3 c8c4 c1d1 g7g5 d1d8 b6d8');
INSERT INTO evaluations VALUES(52753,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(52753,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(52753,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(52753,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(52753,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(52753,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(52753,36,'cp -67','d5e4 b1c3 e7e5 c1e3 e5d4 d1d4 d8d4 e3d4 g8f6 g1e2 c8f5 e1c1 b8d7 e2g3 f5g6 g3e4 f6e4 c3e4 g6e4 d1e1 e8c8 e1e4 d7c5 d4c5 f8c5 f2f4 g7g6 g2g3 h7h5 f4f5 g6g5 f1e2 g5g4 h1e1 a7a5 c1c2 c8c7 e2d3 c7b6 a2a3 a5a4 e1e2 d8d6 e4e8 h8e8 e2e8 c5g1');
INSERT INTO evaluations VALUES(52876,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(52876,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(52876,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(52876,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(52876,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(52876,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(52876,36,'cp -63','b1c3 g8f6 g1e2 c8f5 e2g3 e7e6 f1e2 f5g6 c1e3 b8d7 e1g1 f8d6 d1c2 d8c7 h2h3 e8g8 g3e4 f6e4 c3e4 c6c5 a1d1 c5d4 e3d4 a8d8 e2f3 d6h2 g1h1 h2e5 c2e2 e5d4 d1d4 d7e5 f1d1 d8d4 d1d4 e5f3 g2f3 e6e5 d4d1 f8c8 h1g1 c7c4 e2c4 c8c4');
INSERT INTO evaluations VALUES(52875,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(52875,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(52875,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(52875,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(52875,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(52875,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(52875,36,'cp -59','g8f6 g1e2 c8f5 e2g3 e7e6 f1e2 f5g6 c1e3 b8d7 e1g1 f8d6 d1c2 e8g8 g3e4 f6e4 c3e4 d6h2 g1h2 d8h4 h2g1 g6e4 c2d2 h7h6 e3f4 h4f6 d2e3 f6g6 f2f3 e4f5 f1d1 f8e8 d1d2 a7a5 a1e1 d7f6 f4e5');
INSERT INTO evaluations VALUES(45014,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(45014,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(45014,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(45014,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(45014,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(45014,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(45014,36,'cp -68','g1e2 c8f5 e2g3 e7e6 f1e2 f5g6 e1g1 d8c7 c1e3 f8d6 c4c5 d6f4 d1d2 f4e3 d2e3 f6d5 c3d5 c6d5 f2f3 e4f3 e2f3 e8g8 b2b4 b8d7 b4b5 f8e8 a1c1 d7f6 a2a4 a8d8 c5c6 b7c6 c1c6 c7b8 e3e5 b8e5 d4e5 f6d7 c6c7 d7e5 c7a7');
INSERT INTO evaluations VALUES(55253,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(55253,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(55253,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(55253,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(55253,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55253,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(55253,36,'cp -22','c3e4 e7e5 e2e3 f7f5 e4g3 d4e3 d2e3 d8d1 e1d1 b8c6 f1b5 f8d6 b2b3 g8e7 c1b2 c8d7 b5c4 g7g6 a2a3 h8f8 h2h4 h7h6 d1e1 e5e4 f3d4 c6e5 c4e6 a7a6 a1d1 e5g4 e6d7 e8d7 d4e2 a8e8 e2f4 b7b6 c2c4 g6g5 h4g5 h6g5 f4h5 d7c8 e1e2 e8d8');
INSERT INTO evaluations VALUES(55258,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(55258,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(55258,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(55258,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55258,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(55258,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(55258,36,'cp -32','e7e5 e2e3 f7f5 e4g3 d4e3 d2e3 d8d1 e1d1 f8d6 b2b3 b8c6 f1b5 g8e7 c1b2 c8d7 b5c4 g7g6 a2a3 h7h6 b3b4 h8f8 f3d2 e5e4 c4b3 c6e5 d2c4 e5c4 b3c4 e7c6 b2g7 a7a5 b4b5 c6e5 c4e2 h6h5 g7f8 e8f8 d1e1 h5h4');
INSERT INTO evaluations VALUES(55424,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(55424,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(55424,32,'cp -180',NULL);
INSERT INTO evaluations VALUES(55424,33,'cp -184',NULL);
INSERT INTO evaluations VALUES(55424,34,'cp -191',NULL);
INSERT INTO evaluations VALUES(55424,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(55424,36,'cp -183','g7g5 c2c3 b8c6 d2d4 f8g7 b1d2 g8f6 a2a4 e8g8 c4d3 c6e7 b2b4 c8e6 h2h3 e7g6 a4a5 g5g4 h3g4 f6g4 d1e2 c7c6 c1b2 f8e8 f1e1 g8h8 c3c4 g6h4 f3h4 d8h4 d2f3 h4g3 d4d5 c6d5 e4d5 e6d7 b2g7 h8g7 e2d2 g7f8');
INSERT INTO evaluations VALUES(55747,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(55747,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(55747,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(55747,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(55747,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(55747,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(55747,36,'cp 53','d2d4 d6d5 b1c3 f8b4 e4e5 b7b6 f1b5 c7c6 b5d3 c8a6 g1f3 g8e7 d3a6 b8a6 e1g1 h7h5 d1e2 a6c7 c3d1 e7f5 c1e3 b4e7 e3f2 c6c5 c2c3 d8d7 a2a4 g7g6 h2h3 c5c4 a4a5 b6b5 g2g4 f5g7 g1g2 d7c6 f1g1 e8c8 f3g5 e7g5 f4g5 h5g4');
INSERT INTO evaluations VALUES(57166,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(57166,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(57166,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(57166,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(57166,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(57166,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(57166,36,'cp -32','b8c6 g1f3 g8e7 g2g3 d7d5 e4d5 d8d5 b1c3 d5a5 f1g2 c8g4 e2c4 g4e6 c4b5 a5b5 c3b5 a8c8 d2d3 a7a6 b5c3 c6b4 e1d1 f7f6 a2a3 b4c6 c1e3 h7h5 h2h3 e8f7 f3d2 c8d8 b2b4 e7f5 g2c6 b7c6 d2e4 f5e3 f2e3 f8e7 e4c5 e7c5 b4c5 e6d5 e3e4');
INSERT INTO evaluations VALUES(22824,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(22824,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(22824,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(22824,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(22824,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(22824,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(22824,36,'cp -22','g1f3 g8f6 g2g3 f8c5 f1g2 e8g8 e1g1 f8e8 d2d3 d7d5 e4d5 c6d4 f3d4 e5d4 e2d2 f6d5 d2g5 c8e6 g5d8 a8d8 c1d2 c5f8 a2a4 h7h6 f1e1 g7g5 b1a3 c7c6 a3c4 f8b4 g2d5 c6d5 d2b4 d5c4 d3c4 e6c4 e1e8 d8e8 b2b3');
INSERT INTO evaluations VALUES(22809,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(22809,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(22809,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(22809,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(22809,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(22809,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22809,36,'cp -32','g8e7 g2g3 d7d5 e4d5 d8d5 b1c3 d5c5 d2d3 c6d4 f3d4 e5d4 c3e4 c5c6 e2h5 c6c2 f1e2 e7g6 e1g1 f8e7 c1g5 f7f6 g5f4 e8g8 a1c1 c2b2 c1c7 c8e6 e2f3 a8d8 c7b7 b2a3 f1e1 a3a2 h5b5 e6d5 b7d7 d8d7 b5d7 d5f7 f4c7 e7b4 e1c1 a2a3 c1d1 f8e8 d7d4 g6e5');
INSERT INTO evaluations VALUES(4219,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4219,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4219,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4219,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(4219,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4219,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(4219,36,'cp 9','e1g1 d7d6 a2a3 b4c3 d2c3 c8e6 c4e6 f7e6 h2h3 d8d7 b2b4 f8f7 c3d2 b7b6 d1e2 a7a5 b4b5 c6e7 d3d4 e5d4 e4e5 f6d5 f3d4 e7g6 e2e4 g6e5 f2f4 e5c4 d2c1 a8f8 f4f5 c4e5 d4e6 d7b5 e6f8 d5c3 e4e3 c3e2 g1h2 f7f8');
INSERT INTO evaluations VALUES(4218,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4218,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4218,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4218,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4218,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4218,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4218,36,'cp 0','d7d6 a2a3 b4c3 d2c3 c8e6 c4e6 f7e6 h2h3 d8d7 d1e2 d7e8 b2b4 e8h5 e2d1 f8f7 a3a4 a8f8 b4b5 c6e7 f3h2 h5h4 h2f3');
INSERT INTO evaluations VALUES(3648,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(3648,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(3648,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3648,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3648,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3648,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(3648,36,'cp 4','a2a3 b4c3 d2c3 c8e6 c4e6 f7e6 h2h3 d8e8 b2b4 f6h5 c3d2 f8f6 a3a4 h5f4 b4b5 c6e7 f3h2 a7a6 g2g3 f6g6 h3h4 a6b5 g1h1 f4h3 a4b5 a8a1 d1a1 e8b5 d2e3 b5e8 a1a5 e8c6');
INSERT INTO evaluations VALUES(14966,30,'cp -504',NULL);
INSERT INTO evaluations VALUES(14966,31,'cp -542',NULL);
INSERT INTO evaluations VALUES(14966,32,'cp -543',NULL);
INSERT INTO evaluations VALUES(14966,33,'cp -537',NULL);
INSERT INTO evaluations VALUES(14966,34,'cp -548',NULL);
INSERT INTO evaluations VALUES(14966,35,'cp -555',NULL);
INSERT INTO evaluations VALUES(14966,36,'cp -536','f6d5 d2d4 d5f6 d4d5 h7h6 g5f3 f6e4 e1g1 f8c5 c2c4 c6e7 f3e5 e8g8 d1f3 c8f5 b1d2 e4f2 f1f2 d8d6 d2f1 d6e5 c1e3 f5g4 f3f4 e5f4 f2f4 c5e3 f1e3 g4d7 e3c2 e7g6 f4f1 a8e8 a1e1 e8e1 f1e1 f7f6 b2b3 g8f7 h2h3 g6e5 e1d1 d7f5 c2e3 f5e4 d5d6 c7d6');
INSERT INTO evaluations VALUES(15304,30,'cp -559',NULL);
INSERT INTO evaluations VALUES(15304,31,'cp -547',NULL);
INSERT INTO evaluations VALUES(15304,32,'cp -548',NULL);
INSERT INTO evaluations VALUES(15304,33,'cp -536',NULL);
INSERT INTO evaluations VALUES(15304,34,'cp -511',NULL);
INSERT INTO evaluations VALUES(15304,35,'cp -528',NULL);
INSERT INTO evaluations VALUES(15304,36,'cp -529','d2d4 d5f6 d4d5 h7h6 g5f3 f6e4 e1g1 f8c5 c2c4 c6e7 f3e5 e8g8 d1f3 c8f5 b1c3 e4c3 f3c3 e7g6 e5g6 f5g6 b2b4 c5e7 c1b2 e7f6 c3b3 f6b2 b3b2 c7c6 d5c6 b7c6 a1d1 d8h4 h2h3 f8e8 b2c3 a8d8 d1e1 h4g5 e1e8 d8e8 f1e1 e8e1 c3e1 g8h7 a2a3 g5f5 f2f3 f5b1 e1b1 g6b1 g1f2 b1d3');
INSERT INTO evaluations VALUES(19072,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(19072,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(19072,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(19072,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19072,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19072,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(19072,36,'cp -30','c3g3 f7f5 f1b5 g8f6 b1c3 f8b4 g1e2 f5f4 g3d3 e8g8 a2a3 b4a5 f2f3 a7a6 b5c4 g8h8 c1d2 b7b5 c4b3 d7d6 e1g1 a5b6 g1h1 c6d4 e2d4 b6d4 c3d5 d4b2 a1a2 b2d4 d5f4 a6a5 f4e6 c8e6 b3e6 d8e7 e6b3 a5a4');
INSERT INTO evaluations VALUES(53401,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53401,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(53401,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53401,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(53401,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53401,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(53401,36,'cp 35','e4d5 d8d5 c3d4 e7e5 g1f3 e5d4 d1d4 d5d4 f3d4 f8b4 b1c3 g8f6 d4b5 b4a5 c1g5 e8g8 e1c1 c8e6 g5f6 g7f6 b5d6 b8d7 c3d5 a8b8 f1c4 f8d8 f2f4 b7b5 d6b5 g8f8 h1f1 d7b6 d5b6 a5b6 c4e6 f7e6');
INSERT INTO evaluations VALUES(53346,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(53346,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(53346,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53346,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(53346,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53346,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(53346,36,'cp 12','d8d5 c3d4 e7e5 g1f3 e5d4 d1d4 g8f6 b1c3 d5e6 c1e3 b8c6 d4a4 f8b4 f1b5 b4c3 b2c3 e8g8 e1g1 c8d7 f1e1 f8e8 a4b3 c6a5 b3e6 e8e6 b5d3 a8c8 a1c1 a5c6 h2h3 h7h6 d3e2 e6e7 g1h2 d7f5 f3d4 c6d4 c3d4 c8c1 e1c1 f6d5');
INSERT INTO evaluations VALUES(6435,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(6435,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(6435,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6435,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6435,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6435,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(6435,36,'cp 0','g8f6 d2d3 d7d6 c2c3 a7a5 e1g1 h7h6 c4b5 c5a7 b1d2 e8g8 f1e1 f6h7 d2f1 f7f5 e4f5 c8f5 f1g3 f5g6 c1e3 a7e3 e1e3 h7g5 b5c4 g8h8 c4d5 d8f6 d1d2 g6f7 d5f7 f8f7 f3g5 f6g5 g3e4 g5g6 e3g3 g6f5 g3f3 f5g6');
INSERT INTO evaluations VALUES(10127,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10127,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10127,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10127,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10127,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(10127,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10127,36,'cp 6','d2d3 h7h6 c2c3 a7a5 b1d2 d7d6 e1g1 c8e6 f1e1 e8g8 c4b5 d8b8 d2f1 b8a7 c1e3 c5e3 f1e3 c6e7 a2a4 e7g6 b5c4 e6c4 e3c4 a7c5 d1c2 f8e8 b2b3 c5c6 b3b4 b7b6 b4b5 c6c5 c4e3 d6d5 e4d5 f6d5 e3d5 c5d5 a1d1');
INSERT INTO evaluations VALUES(6459,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(6459,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(6459,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(6459,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(6459,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(6459,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(6459,36,'cp 62','g1f3 d7d6 a2a3 a7a6 b2b4 c5a7 c1b2 g8e7 f1e2 e8g8 e1g1 e7g6 d1c2 c8g4 a1d1 f7f5 h2h3 g4h5 b4b5 e5e4 f3e1 h5e2 c3e2 c6e5 b5a6 b7a6 b2e5 g6e5 d2d4 e5g6 g2g3 d8e7 e1g2 a7b8 h3h4 c7c6 c4c5');
INSERT INTO evaluations VALUES(6927,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(6927,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(6927,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(6927,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(6927,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(6927,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(6927,36,'cp 212','c2c3 d7d6 d2d4 c5b6 a2a4 a7a6 b2b4 c8g4 h2h3 g4f3 d1f3 e5d4 f3h5 g7g6 h5d1 g8e7 c3d4 f6f5 b1c3 b6d4 c1b2 h8f8 b4b5 a6b5 a4b5 a8a1 b2a1 d4c3 a1c3 c6e5 c3e5 d6e5 d1a1 e7c8 e4f5 d8d4 a1c1 c8d6 c4b3 f8f5 c1c7 e5e4 c7h7 e4e3 f1e1 f5f2 h7g6 e8d8 g6g5 d8d7 g5e3 d4e3 e1e3 f2b2 b3e6 d7d8 e6g4 d8c7');
INSERT INTO evaluations VALUES(6923,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(6923,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(6923,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(6923,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(6923,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(6923,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(6923,36,'cp 240','d7d6 d2d4 c5b6 a2a4 a7a6 b2b4 c8g4 h2h3 g4f3 d1f3 e5d4 f3h5 g7g6 h5d1 d4d3 c4d3 c6e5 d3e2 g8e7 d1b3 e5f7 e2c4 e8g8 c1h6 d6d5 c4d5 e7d5 h6f8 d5f4 f8c5 b6c5 b4c5 g8g7 f1d1 d8c8 c3c4 f7g5 b1c3 f4h3 g1f1 h3f4 c3d5');
INSERT INTO evaluations VALUES(10358,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(10358,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(10358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10358,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10358,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10358,36,'cp 0','c3d5 b4c5 f3e5 c5f2 e1f2 f6e4 f2g1 c6e5 d2d4 c7c6 d4e5 c6d5 c1e3 d8a5 d1d4 e8g8 e2f3 f8e8 a1d1 a5c7 d4d5 c7e5 d5e5 e8e5 e3f4 e5e7 h2h4 b7b6 f4d6 e7e8 d6f4');
INSERT INTO evaluations VALUES(12920,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12920,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12920,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(12920,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12920,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(12920,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12920,36,'cp 0','d7d6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5c6 d7c6 f1e1 f8e7 d4d5 b6d5 c3d5 c6d5 d1d5 e6d5 c1f4 a8d8 a1c1 d8d7 c1c8 d7d8 c8c7 d8d7');
INSERT INTO evaluations VALUES(13411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13411,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13411,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13411,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13411,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13411,36,'cp 0','f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 b5c6 d7c6 e1g1 g7g6 b1c3 f8g7 e5c6 b7c6 c1e3 e8g8 b2b3 c6c5 d4c5 g7c3 c5b6 c3a1 d1a1 a7b6 e3h6 f7f6 h6f8 g8f8 f1d1 d8c7 a1b2 a8c8 b2d4 c7c5 d4d7 c5c6 d7d4 c6c5');
INSERT INTO evaluations VALUES(13410,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13410,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13410,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13410,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13410,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13410,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13410,36,'cp 0','d5b6 c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5c6 d7c6 f1e1 f8e7 d4d5 b6d5 c3d5 c6d5 d1d5 e6d5 c1f4 a8d8 a1c1 d8d7 c1c8 d7d8');
INSERT INTO evaluations VALUES(13048,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13048,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13048,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13048,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13048,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13048,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13048,36,'cp 0','c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5c6 d7c6 c1e3 b6d5 a1c1 f8e7 d1b3 a8d8 c3d5 c6d5 b3a4 d5d7 a4a5 d7d5');
INSERT INTO evaluations VALUES(19482,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(19482,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(19482,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(19482,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(19482,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(19482,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(19482,36,'cp -56','d7d5 e4d5 d8d5 b1a3 g7g6 g1f3 f8g7 f1c4 d5d6 d1e2 g8h6 d2d4 c5d4 a3b5 d6d8 b5d4 c6d4 f3d4 e8g8 e1g1 d8b6 c1e3 h6g4 a1d1 g4e3 e2e3 e7e6 d1d2 b6c5 c4e2 c8d7 f1d1 e6e5 d4c2 c5e3 c2e3 d7a4 b2b3 e5f4 e3d5 a4c6 d5e7 g8h8 e7c6 b7c6 e2f3');
INSERT INTO evaluations VALUES(21109,30,'cp -197',NULL);
INSERT INTO evaluations VALUES(21109,31,'cp -213',NULL);
INSERT INTO evaluations VALUES(21109,32,'cp -213',NULL);
INSERT INTO evaluations VALUES(21109,33,'cp -220',NULL);
INSERT INTO evaluations VALUES(21109,34,'cp -219',NULL);
INSERT INTO evaluations VALUES(21109,35,'cp -200',NULL);
INSERT INTO evaluations VALUES(21109,36,'cp -212','d8d4 e2e4 g8f6 b1c3 c8e6 c4e2 e8c8 g1f3 d4b4 a2a3 b4a5 a1b1 e6g4 e2b5 a5b5 f1b5 g4f3 g2f3 c6d4 b5e2 a7a6 b2b4 g7g6 f3f4 f8g7 e2d1 f6d7 c1e3 d7b6 b4b5 a6b5 c3b5 d4b5 b1b5 b6c4 e1e2 e7e6 b5b3 c4e3');
INSERT INTO evaluations VALUES(1627,30,'cp -204',NULL);
INSERT INTO evaluations VALUES(1627,31,'cp -188',NULL);
INSERT INTO evaluations VALUES(1627,32,'cp -194',NULL);
INSERT INTO evaluations VALUES(1627,33,'cp -197',NULL);
INSERT INTO evaluations VALUES(1627,34,'cp -190',NULL);
INSERT INTO evaluations VALUES(1627,35,'cp -204',NULL);
INSERT INTO evaluations VALUES(1627,36,'cp -210','e2e4 g8f6 b1c3 c8e6 c4e2 e8c8 g1f3 d4b4 a2a3 b4a5 a1b1 e6g4 e2b5 a5b5 f1b5 g4f3 g2f3 c6d4 b5e2 e7e5 c3b5 d4c2 e1f1 a7a6 b5c3 c2d4 b2b4 c7c6 f1g2 b7b5 e2d1 f6h5 c3e2 f8d6 h1e1 d4e2 e1e2 d6b4 d1b3 b4c5 b3f7 h5f4 c1f4');
INSERT INTO evaluations VALUES(27184,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(27184,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(27184,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(27184,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(27184,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(27184,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(27184,36,'cp 113','a2a3 a7a6 f1d3 c6a7 c2c3 c7c5 d3c2 c5d4 c3d4 a7b5 e1g1 d8b6 h2h4 a8c8 a3a4 b5a7 b1c3 a7c6 c3e2 c6a5 e2f4 g8e7 f1e1 e7c6 e1e3 a5c4 e3b3 c6b4 c2b1 c4a5 b3c3 c8c3 b2c3 b4c6 h4h5 h7h6');
INSERT INTO evaluations VALUES(31052,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(31052,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(31052,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31052,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(31052,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(31052,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(31052,36,'cp 0','g1f3 g8f6 b1c3 b7b6 b2b4 g7g6 e2e3 f8g7 f1e2 e8g8 c1b2 a7a5 a2a3 a5b4 a3b4 b8d7 f3h4 f5e6 e1g1 b6c5 b4c5 a8a1 b2a1 d8c7 h2h3 f6e4 d1c2 d7f6 h4f3 e6f5 e2d3 f8a8 f1b1 e4c3 a1c3 f5d3 c2d3 f6e4 c3e1');
INSERT INTO evaluations VALUES(39932,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39932,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(39932,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39932,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39932,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39932,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39932,36,'cp 0','e2e3 b7b6 f1d3 c8b7 f2f3 b8c6 g1e2 e8g8 e3e4 f6e8 e1g1 b7a6 d1a4 c6a5 d4c5 d8c8 c1e3 d7d6 c5b6 a7b6 f1b1 a5c4 d3c4 a6c4 a4d1 b6b5 e2d4 c8a6 d4c2 a6c6 c2d4');
INSERT INTO evaluations VALUES(39929,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39929,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(39929,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39929,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39929,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(39929,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(39929,36,'cp -8','b7b6 f1d3 c8b7 f2f3 b8c6 g1e2 e8g8 e3e4 f6e8 e1g1 b7a6 d1a4 c6a5 d4c5 d8c8 c1e3 d7d6 c5b6 a7b6 f1b1 a5c4 d3c4 a6c4 a4c2 b6b5 e2d4 c8d7 a3a4 b5a4 b1b4 d6d5 e4d5 d7d5 b4a4 e8f6 a4a8 f8a8 a1a8 d5a8 c2d2 e6e5 d4c2 h7h6 h2h3 c4e6 c2b4 a8b7 g1h2');
INSERT INTO evaluations VALUES(12644,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(12644,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(12644,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(12644,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(12644,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(12644,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(12644,36,'cp 148','f8e7 e5g3 e8g8 c1h6 f6h5 g3g4 d7d6 g4h5 g7h6 h5h6 e7f6 f1d3 f6g7 h6h5 f8e8 e1g1 e8e6 d3c4 e6g6 h5f3 c8g4 f3f7 g8h8 c3e2 b7b5 c4b5 a8b8 b5d3 g4e6 f7a7 b8a8 a7b7 c5c4 e2f4 c4d3 f4g6 h7g6 c2d3 a8c8 b7a7 g7b2 a1b1 d8f6 a7e3 b2d4');
INSERT INTO evaluations VALUES(7956,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(7956,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(7956,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(7956,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(7956,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(7956,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(7956,36,'cp 175','e5g3 e8g8 c1h6 f6h5 g3g4 d7d6 g4h5 g7h6 h5h6 e7f6 f1d3 f6g7 h6h5 c5c4 d3c4 d8c7 c4d3 g7c3 b2c3 c7c3 e1e2 c8e6 h1e1 e6c4 e2f1 c3c2 d3c4 c2c4 f1g1 c4e6 a1d1 a8c8 d1d5 h7h6 h2h3 b7b6 h5f3 c8c5 f3g3 e6g6 g3g6 f7g6 d5d6 c5c2 f2f3 g8g7 e4e5 f8f7 e5e6 f7e7 d6d5 c2c8 d5d7 e7d7 e6d7');
INSERT INTO evaluations VALUES(14773,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(14773,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(14773,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(14773,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(14773,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(14773,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(14773,36,'cp 55','e2e3 f8d6 f4g3 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 e1g1 c8f5 a1c1 f5d3 d1d3 a7a5 c1c2 c7c6 f1c1 d8d7 g3d6 d7d6 h2h4 f8c8 f3g5 d6g6 d3d1 h7h6 g5h3 g6d6 h3f4 a5a4 f4d3 a8a5 d1g4 c6c5 d4c5 b6c5');
INSERT INTO evaluations VALUES(15462,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(15462,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(15462,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(15462,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(15462,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15462,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(15462,36,'cp -9','c6a5 c4b3 a5b3 a2b3 f8e7 b1c3 h7h6 g5h4 e8g8 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f1e1 f8e8 h4g3 c7c6 d3d4 d6d5 e4d5 e5d4 f3d4 f6d5 c3d5 d8d5 d4f5 e7f6 e1e8 a8e8 d1d5 c6d5 a1a7 e8e2 c2c3 e2b2 a7b7 f6c3 b7d7 g6f5 g4f5 d5d4 g3e5 b2b3 e5d4');
INSERT INTO evaluations VALUES(16074,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(16074,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(16074,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(16074,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(16074,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(16074,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16074,36,'cp -23','c6d4 f3g3 d7d6 b1a3 f8g7 h2h4 c8e6 g1e2 e6c4 d3c4 d8e7 h4h5 e8c8 c1g5 f7f6 g5d2 e7f7 e2d4 e5d4 e1c1 f6f5 e4f5 g6f5 g3d3 a7a6 f2f3 h8e8 b2b4 f5f4 b4b5 g7e5 g2g4 f4g3 d2h6 g3g2 h1e1 f7h5');
INSERT INTO evaluations VALUES(16090,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(16090,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(16090,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(16090,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(16090,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(16090,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(16090,36,'cp -14','f3g3 d7d6 b1a3 f8g7 h2h4 c8e6 g1e2 e6c4 d3c4 d4e6 a3b5 f7f5 b5c3 d8d7 c1d2 f5f4 g3h3 e6c5 h3d7 e8d7 a2a4 a7a5 f2f3 h6f7 c3d5 a8a6 d5c3 h7h5 e1c1 a6a8 g2g3 f4g3 d1g1 a8g8 g1g3 g7f6 c3d5 f6d8 d2a5 c5a4 a5b4 d7e8 b4a3 a4c5');
INSERT INTO evaluations VALUES(20909,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(20909,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(20909,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(20909,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(20909,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(20909,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(20909,36,'cp 106','c2c4 g8f6 e2e3 f8g7 b1d2 d5c4 f1c4 e8g8 a1c1 f6h5 f4e5 h5f6 e5g3 f6h5 g3h4 g7f6 h4f6 e7f6 d1c2 f6f5 c4b3 h5f6 a2a3 d8e7 h2h3 c8e6 b3e6 e7e6 d2b3 f6e4 b3c5 e4c5 c2c5 f8d8 c5c3 e6d5 e1g1');
INSERT INTO evaluations VALUES(42857,30,'cp 405',NULL);
INSERT INTO evaluations VALUES(42857,31,'cp 435',NULL);
INSERT INTO evaluations VALUES(42857,32,'cp 402',NULL);
INSERT INTO evaluations VALUES(42857,33,'cp 386',NULL);
INSERT INTO evaluations VALUES(42857,34,'cp 385',NULL);
INSERT INTO evaluations VALUES(42857,35,'cp 433',NULL);
INSERT INTO evaluations VALUES(42857,36,'cp 413','c4f7 e8f8 d2d4 d7d6 d4c5 d8e7 f7d5 e4f6 b1c3 b8c6 c5d6 e7e5 c1e3 e5d6 d5b3 c6a5 d1d6 c7d6 e1c1 a5b3 a2b3 c8d7 d1d6 d7c6 f2f3 f8f7 h1d1 h8e8 d6d3 e8e5 e3f4 e5f5 d3d4 f6h5 f4e3 h5f6 d1d3 a8e8 b3b4 h7h6 h2h3 a7a6 b2b3 c6f3 g2f3 f5f3');
INSERT INTO evaluations VALUES(42302,30,'cp 418',NULL);
INSERT INTO evaluations VALUES(42302,31,'cp 399',NULL);
INSERT INTO evaluations VALUES(42302,32,'cp 391',NULL);
INSERT INTO evaluations VALUES(42302,33,'cp 394',NULL);
INSERT INTO evaluations VALUES(42302,34,'cp 412',NULL);
INSERT INTO evaluations VALUES(42302,35,'cp 390',NULL);
INSERT INTO evaluations VALUES(42302,36,'cp 397','e8f8 d2d4 d7d6 d4c5 d8e7 f7d5 e4f6 b1c3 d6e5 c1g5 b8d7 d5c4 c7c6 c3e4 h7h6 g5f6 d7f6 e4d6 c8e6 d1e2 b7b6 e1c1 e6d5 h1e1 b6c5 e2e5 e7e5 e1e5 d5g2 d6f7 h8h7 e5c5 g2d5 f7e5 g7g5 e5c6 d5c4 c5c4 a8e8 f2f3 f6d7 c4d4 d7f6 d4d8 h7c7 d8e8 f8e8');
INSERT INTO evaluations VALUES(45262,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(45262,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(45262,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(45262,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(45262,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45262,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(45262,36,'cp 0','e4e5 f6g4 e5d6 e7d6 h2h3 g4e5 f1e2 f8e7 e1g1 b8c6 c3d5 e8g8 c1e3 c8e6 b2b4 a7a6 a1c1 a8c8 d5b6 c8c7 b6d5');
INSERT INTO evaluations VALUES(46750,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(46750,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(46750,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(46750,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46750,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46750,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(46750,36,'cp 12','f8g7 e2e3 e8g8 c4d5 f6d5 c3d5 d8d5 f4c7 b8a6 f1a6 d5g2 d1f3 g2f3 g1f3 b7a6 a1c1 f7f6 e1g1 c8b7 f3d2 e7e5 c7d6 f8d8 d4e5 f6e5 d6b4 a8c8 b4a5 c8c1 f1c1 d8c8 c1c8 b7c8 a5c3 g8f7 d2e4 f7e6 e4g5 e6f5 g5h7 g7h6 a2a3 c8e6 c3b4 e6b3 h7f8 h6f8 b4f8 f5e6 f8h6 e6f5');
INSERT INTO evaluations VALUES(41439,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(41439,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(41439,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41439,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41439,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41439,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41439,36,'cp 22','e2e3 e8g8 c4d5 f6d5 c3d5 d8d5 f4c7 b8a6 f1a6 d5g2 d1f3 g2f3 g1f3 b7a6 a1c1 f7f6 h1g1 c8b7 e1e2 a8c8 g1d1 e7e6 f3d2 e6e5 d4e5 f6e5 c7d6 f8d8 d6e7 d8e8 e7b4 e8d8 c1c8 d8c8 d2c4 b7c6 c4d6 c8f8 e3e4 g7h6 b4c5 a6a5 c5e3');
INSERT INTO evaluations VALUES(41437,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(41437,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(41437,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41437,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(41437,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41437,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41437,36,'cp 14','e8g8 c4d5 f6d5 c3d5 d8d5 f4c7 b8a6 f1a6 d5g2 d1f3 g2f3 g1f3 b7a6 a1c1 f7f6 h1g1 c8b7 e1e2 a8c8 g1d1 e7e6 f3d2 e6e5 c7d6 f8d8 d4e5 f6e5 d6e7 d8e8 e7b4 b7d5 b2b3 e5e4 d2c4 d5e6 c4d6 e6g4 e2f1 e8d8 d6c8 d8d1 c1d1 g4d1 c8a7 g7e5 h2h3 g8f7 a7c6 f7e6 c6e5');
INSERT INTO evaluations VALUES(36720,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36720,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36720,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(36720,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36720,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36720,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36720,36,'cp 23','c4d5 f6d5 c3d5 d8d5 f4c7 b8a6 f1a6 d5g2 d1f3 g2f3 g1f3 b7a6 a1c1 f7f6 h1g1 c8b7 e1e2 a8c8 g1d1 e7e6 h2h4 f8f7 c7a5 b7c6 f3d2 f7d7 c1c3 e6e5 d1c1 c6b5 e2e1 c8c3 b2c3 b5c6 d2b3 g7f8 c3c4 f8a3 c1c3 c6a4 c4c5 a3b2 c3c2 b2a3');
INSERT INTO evaluations VALUES(48501,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(48501,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(48501,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(48501,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(48501,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(48501,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(48501,36,'cp -32','c7c5 g1f3 b7b5 c3b5 d8a5 b5c3 f6e4 f1e2 e4c3 b2c3 a5c3 c1d2 c3b3 d1c1 b3a4 e1g1 f8e7 d4c5 e8g8 c1c4 a4c4 e2c4 e7c5 f1c1 a7a6 c4e6 c5f2 g1f2 c8e6 f3d4 b8d7 d4e6 f7e6 f2g1 f8c8 a1b1 c8c1 b1c1 d7f6 c1c6 e6e5 d2a5 g8f7 c6c7 f7g6 c7e7 e5e4 a5c3');
INSERT INTO evaluations VALUES(46123,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(46123,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(46123,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46123,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46123,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46123,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46123,36,'cp 0','g1f3 b7b5 c3b5 d8a5 b5c3 f6e4 c1d2 e4d2 f3d2 c5d4 d2c4 a5d8 d1f3 b8c6 c3b5 c8b7 f3g3 e6e5 a1c1 g7g6 c4e5 d8e7 c1c6 b7c6 b5d4 c6d7 f1c4 f8g7 c4f7 e8f8 f7e6 a8d8 g3f4 f8e8 e6f7 e8f8');
INSERT INTO evaluations VALUES(1189,30,'cp -342',NULL);
INSERT INTO evaluations VALUES(1189,31,'cp -340',NULL);
INSERT INTO evaluations VALUES(1189,32,'cp -332',NULL);
INSERT INTO evaluations VALUES(1189,33,'cp -312',NULL);
INSERT INTO evaluations VALUES(1189,34,'cp -329',NULL);
INSERT INTO evaluations VALUES(1189,35,'cp -358',NULL);
INSERT INTO evaluations VALUES(1189,36,'cp -355','g7g6 h5g5 d4e4 g5e3 e4g2 e5f3 g8e7 h1g1 g2h3 d2d3 c8e6 g1g3 h3h5 e3g5 h5g5 c1g5 f8g7 b1c3 h7h6 g5f4 e8c8 e1c1 h8e8 d1e1 e7f5 g3g1 c6c5 f3d2 b7b6 f4e5 f7f6 e5g3 f5d4 c3e2 d4e2 e1e2 e6d7 e2e3 f6f5 d2f3 e8e3 f2e3 d8e8 f3h4 g6g5');
INSERT INTO evaluations VALUES(1182,30,'cp -342',NULL);
INSERT INTO evaluations VALUES(1182,31,'cp -325',NULL);
INSERT INTO evaluations VALUES(1182,32,'cp -331',NULL);
INSERT INTO evaluations VALUES(1182,33,'cp -336',NULL);
INSERT INTO evaluations VALUES(1182,34,'cp -356',NULL);
INSERT INTO evaluations VALUES(1182,35,'cp -351',NULL);
INSERT INTO evaluations VALUES(1182,36,'cp -355','h5g5 d4e4 g5e3 e4g2 e5f3 g8e7 h1g1 g2h3 d2d3 c8e6 g1g3 h3h5 e3g5 h5g5 c1g5 f8g7 b1c3 h7h6 g5e3 e7f5 g3g1 e8c8 e1c1 b7b6 a2a4 a6a5 e3f4 d8e8 h2h4 e6d7 d1e1 e8e1 g1e1 h8e8 e1e4 e8e6 f4e5 f7f6 e5f4 e6e7 b2b3 e7e8 c1d2 e8h8 f4e3 g6g5 f3d4 f5d4 e3d4 c6c5 h4g5 c5d4');
INSERT INTO evaluations VALUES(6994,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(6994,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(6994,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(6994,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(6994,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(6994,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(6994,36,'cp 75','b1c3 d7d6 d4c6 b7c6 f1d3 g8f6 e1g1 e8g8 f2f4 d6d5 e4e5 f6g4 c3a4 f7f6 h2h3 g4h6 c1e3 c8f5 a4c5 f6e5 f4e5 d8e8 d1d2 f5d3 f1f8 e7f8 c5d3 h6f7 a1e1 f7d8 e3c5 d8e6 c5f8 e8f8 b2b4 a7a5 b4a5 f8a3 d3f4 e6f4 d2f4');
INSERT INTO evaluations VALUES(6993,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(6993,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(6993,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(6993,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(6993,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(6993,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(6993,36,'cp 85','d7d6 f1b5 c8d7 d4c6 b7c6 b5d3 e7f6 e1g1 g8e7 f2f4 e8g8 c1e3 f6c3 b2c3 c6c5 f4f5 d8e8 f5f6 e7d5 f6g7 d5e3 g7f8q e8f8 d1h5 e3f1 a1f1 d7e6 e4e5 f8g7 e5d6 c7d6 a2a3 g8h8 d3e4 a8e8 h5h4 e8g8 g1h1 d6d5');
INSERT INTO evaluations VALUES(5662,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(5662,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(5662,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(5662,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(5662,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(5662,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(5662,36,'cp 81','f1b5 c8d7 d4c6 b7c6 b5d3 e7f6 e1g1 g8e7 f2f4 e8g8 c1e3 f6c3 b2c3 c6c5 f4f5 f7f6 e4e5 d6e5 e3c5 f8e8 d3e4 c7c6 c3c4 d8a5 c5e3 a8d8 d1e1 a5e1 a1e1 g8f7 h2h3 g7g6 f5g6 h7g6 g2g4 f7g7 e1b1 d7e6 b1b7 e6c4 e4c6 c4f1');
INSERT INTO evaluations VALUES(11509,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11509,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11509,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11509,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11509,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11509,36,'cp 0','d7d5 e4d5 c6a5 b2b3 h7h6 g5f3 e5e4 d1e2 d8e7 f3d4 b7b5 d5d6 c7d6 c4d5 f6d5 c3d5 e7e5 d4b5 a6b5 e2b5 e8d8 b5b6 d8d7 b6b5');
INSERT INTO evaluations VALUES(11389,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11389,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11389,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11389,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11389,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11389,36,'cp 0','e4d5 c6a5 b2b3 h7h6 g5f3 e5e4 d1e2 d8e7 f3d4 b7b5 d5d6 c7d6 c4d5 f6d5 c3d5 e7e5 d4b5 a6b5 e2b5 e8d8 b5b6 d8d7 b6b5');
INSERT INTO evaluations VALUES(11386,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11386,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11386,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11386,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11386,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11386,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11386,36,'cp 0','c6a5 b2b3 h7h6 g5f3 e5e4 d1e2 d8e7 f3d4 b7b5 d5d6 c7d6 c4d5 f6d5 c3d5 e7e5 d4b5 a6b5 e2b5 e8d8 b5b6 d8d7 b6b5');
INSERT INTO evaluations VALUES(16003,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(16003,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(16003,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(16003,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(16003,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(16003,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(16003,36,'cp 85','g7g6 e2e4 f8g7 d1c2 e7e5 f1e2 e8g8 e1g1 e5d4 f3d4 f8e8 c1e3 a7a6 a1d1 c7c6 f1e1 d8c7 d4b3 a6a5 a2a4 b7b6 e3f4 d7e5 c2d2 c8a6 d2d6 c7d6 d1d6 e5c4 e2c4 a6c4 b3d2 f6d5 d2c4 d5f4 c4b6 a8d8 e1d1 d8d6 d1d6 g7c3 b2c3 e8e6 d6e6 f4e6 b6c4 e6c5');
INSERT INTO evaluations VALUES(15990,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(15990,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(15990,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(15990,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(15990,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(15990,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(15990,36,'cp 91','e2e4 e7e5 f1e2 f8g7 e1g1 e8g8 c1e3 d8e7 d1c2 e5d4 f3d4 f8e8 f2f3 c7c6 e3f2 a7a6 f2h4 e7d8 a1d1 d8c7 d4b3 a6a5 g1h1 a5a4 b3c1 a4a3 b2b4 d7e5 f3f4 e5g4 c2d3 d6d5 c4d5 c6d5 e4e5 c8f5 d3f3 f6e4');
INSERT INTO evaluations VALUES(19716,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(19716,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19716,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19716,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19716,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(19716,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(19716,36,'cp 36','c2c4 e7e6 c4c5 e6e5 g1f3 e5d4 f3d4 c8d7 b5c6 b7c6 b2b4 f8e7 e1g1 e7f6 c1b2 g8e7 b1d2 a7a5 a2a3 e8g8 d2b3 a5b4 a3b4 a8a1 d1a1 d8a8 d4c2 a8a1 b2a1 f6g5 f2f4 g5h6 a1c3 f7f6 c2d4 e7f5 h2h3 f5d4');
INSERT INTO evaluations VALUES(25030,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(25030,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(25030,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(25030,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(25030,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(25030,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(25030,36,'cp 105','c1e3 e7e5 d4d5 c6e7 d1d3 g4d7 f1e2 f6g4 e3d2 a7a6 a2a3 a8c8 h2h3 g4f6 e1c1 b7b5 c1b1 c7c6 d5c6 e7c6 d2g5 d7e6 g5f6 d8f6 h3h4 f8e7 c3d5 e6d5 d3d5 e8g8 g2g4 f6e6 d5d3 e6g4 f3e5 c6e5 e2g4 e5d3 c2d3 c8d8 d3d4 e7f6 f2f4 h7h6');
INSERT INTO evaluations VALUES(35104,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(35104,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(35104,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(35104,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35104,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35104,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35104,36,'cp 0','b8c6 c2c3 d7d6 d2d3 h7h5 b1d2 g8e7 b2b4 c8g4 f3f6 g7f6 g1e2 f8h6 d2f1 h6c1 a1c1 g4e6 f1e3 e8c8 b4b5 c6a5 c4d5 f6f5 d5e6 f7e6 e4f5 e7f5 e1g1 h5h4 c3c4 b7b6 e3f5 e6f5 f2f4 h8e8 g1f2 e5e4');
INSERT INTO evaluations VALUES(39742,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(39742,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(39742,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(39742,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(39742,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(39742,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(39742,36,'cp 26','g1h1 e8g8 f2f4 d8c7 c1e3 b7b5 e4e5 d6e5 f4e5 c7e5 d1f3 e5c7 c3e4 f6d5 e4g5 e7g5 e3g5 c7c5 d4b3 c5d6 a2a4 b5b4 d3e4 f7f5 e4d5 d6d5 f3d5 e6d5 g5e7 f8e8 e7b4 b8c6 f1e1 c8e6 b4c3 e6f7 h1g1 a8c8 e1e8 c8e8 g1f1 f5f4 b3d4 c6d4 c3d4 f7g6 c2c3');
INSERT INTO evaluations VALUES(40074,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(40074,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40074,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(40074,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(40074,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(40074,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(40074,36,'cp 23','e8g8 a2a3 b8d7 d1c2 f6h5 g2g3 h5f4 g3f4 d5c4 f1c4 c6c5 d4c5 d7c5 a1d1 d8c7 c3b5 c7b6 f3e5 a7a6 b5d4 e7f6 d4f3 f6e5 f3e5 c5d7 c2e4 b6c7 c4e2 d7e5 e4e5 c7e5 f4e5 b7b5 d1d6 c8b7 h1g1 a8c8 f2f4 g7g6');
INSERT INTO evaluations VALUES(35748,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(35748,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(35748,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(35748,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(35748,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(35748,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(35748,36,'cp 42','f1d3 d5c4 d3c4 b8d7 e1g1 b7b5 c4d3 b5b4 c3e4 c6c5 d1c2 c8b7 d4c5 f6e4 d3e4 b7e4 c2e4 e7c5 f1d1 d8e7 a1c1 a8c8 f4e5 f8d8 h2h3 h7h6 b2b3 d7e5 f3e5 d8d1 c1d1 c5d6 e5c6 e7d7 d1c1 c8c7 g2g3 d6f8 g1g2 d7d2 c1c2 d2d6 c2c4 g7g6 e4f3 d6d7');
INSERT INTO evaluations VALUES(35749,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(35749,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(35749,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(35749,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(35749,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(35749,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(35749,36,'cp 54','d5c4 d3c4 a7a6 c4e2 b7b5 f3e5 c8b7 e2f3 d8b6 a2a4 a6a5 e1g1 b8d7 f4g5 e7d6 e5d7 f6d7 c3e4 d6b4 a1c1 e6e5 d1c2 h7h6 e4c5 d7c5 d4c5 b6c7 g5h4 b7a6 f1d1 f8c8 f3e2 b5a4 e2a6 a8a6');
INSERT INTO evaluations VALUES(40215,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40215,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40215,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40215,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(40215,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(40215,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(40215,36,'cp 4','c5d4 f3d4 d5c4 f1c4 e8g8 e1g1 c8d7 a1c1 h7h6 g5h4 b8c6 d4f3 a8c8 c4e2 d8b6 c3a4 b6c7 a4c5 f8d8 c5d7 c7d7 d1a4 c6e5 a4b3 e5f3 e2f3 b7b6 c1d1 e7d6 h4g3 d7c7 g3d6 d8d6 d1d6 c7d6 f1d1 d6c5');
INSERT INTO evaluations VALUES(40230,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(40230,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(40230,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(40230,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40230,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(40230,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40230,36,'cp 16','f3d4 d5c4 f1c4 c8d7 d4f3 e8g8 e1g1 d7c6 d1b3 d8a5 g5h4 b8d7 f3d4 d7c5 d4c6 c5b3 c6a5 b3a5 c4e2 f8d8 f1c1 a8c8 c1c2 a5c6 a1d1 d8d1 e2d1 f6d5 h4e7 d5c3 c2c3 c6e7 c3c8 e7c8 g1f1 g8f8 f1e2 c8d6 e2d3 f8e7 d1b3 f7f5 f2f3');
INSERT INTO evaluations VALUES(40226,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(40226,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40226,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(40226,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(40226,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40226,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40226,36,'cp 20','d5c4 f1c4 c8d7 e1g1 e8g8 d4f3 h7h6 g5f4 d7c6 f3e5 d8d1 f1d1 f6h5 c4e2 h5f4 e3f4 g7g5 g2g3 g5f4 g3f4 f8c8 a1c1 b8a6 e5c6 c8c6 e2f3 c6c7 c3b5 c7c1 d1c1 a8d8 a2a3 f7f5 f3b7 a6c5 b7f3 c5d3');
INSERT INTO evaluations VALUES(41066,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(41066,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(41066,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(41066,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(41066,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(41066,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(41066,36,'cp -56','e8g8 e1c1 f8e8 e2d2 h7h6 g5h4 c8f5 f1d3 f5d3 h4f6 e7f6 d2d3 c7c6 g1f3 d8d6 h1e1 b8d7 g2g3 g7g6 e1e8 a8e8 d1e1 e8e1 f3e1 d7b6 e1f3 b6c4 c3d1 d6e6 d1e3 c4d6 c1d2 a7a5 c2c3 a5a4 d2e1 g8g7 e1f1 b7b5 f3d2 f6g5 f1g1 g5e3 d3e3 e6e3 f2e3 g6g5 g1f1');
INSERT INTO evaluations VALUES(36390,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(36390,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(36390,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(36390,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(36390,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(36390,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(36390,36,'cp -39','e1c1 c7c6 d1e1 c8e6 g1h3 f8e8 h3f4 e6d7 f4h5 b8a6 g5h4 f6e4 h4e7 d8e7 f2f3 e7g5 e2e3 g5e3 e1e3 e4c3 e3e8 a8e8 b2c3 e8e1 c1b2 a6c7 h5g3 g7g6 f1d3 e1h1 g3h1 g8f8 h1g3 f8e7 c3c4 d5c4 d3c4 b7b5 c4b3 d7e6 g3e4 a7a5 b2c3 a5a4 b3e6 e7e6 e4g5 e6e7 g5h7 c7d5 c3d2 f7f6 a2a3');
INSERT INTO evaluations VALUES(36389,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(36389,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(36389,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(36389,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(36389,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(36389,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(36389,36,'cp -60','f8e8 e2d2 h7h6 g5h4 b8d7 f1d3 c7c5 d3b5 c5d4 h4f6 e7f6 c3d5 a7a6 b5d7 c8d7 g1f3 d7f5 d5f6 d8f6 f3d4 a8d8 d2f4 d8d5 d4f5 d5f5 f4d4 f6d4 d1d4 f5f2 d4d2 e8e2 d2e2 f2e2 g2g4 g7g5 h2h3 g8g7 b2b4 e2f2 a2a4 f2f4 c2c3 f4f2 b4b5 a6b5');
INSERT INTO evaluations VALUES(57792,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(57792,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(57792,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(57792,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(57792,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(57792,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(57792,36,'cp -51','e2d2 h7h6 g5h4 b8d7 f1d3 c7c5 d3b5 c5d4 d2d4 a7a6 b5a4 b7b5 a4b3 d7b6 g1f3 c8e6 h1e1 b6c4 d4d3 d8a5 f3d4 e7a3 b3c4 d5c4 d3g3 f6h5 d4c6 h5g3 c6a5 g3f5 h4g3 a3f8 a2a3 f5g3 h2g3 a8c8 f2f3 e6f5 e1e8 c8e8 a5b7 f5g6 g3g4 e8e6 b7d8 e6d6 d1d6 f8d6');
INSERT INTO evaluations VALUES(43276,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(43276,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(43276,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(43276,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43276,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(43276,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(43276,36,'cp -13','c7c5 c2c4 e8g8 f1e2 d7d5 c4d5 d8d5 b1c3 d5f5 d1c1 b7b6 d2d4 c8b7 e1g1 b8d7 f1d1 a8d8 f3h4 f5e6 h4f3 c5d4 f3d4 e6e5 e2f3 e5b8 c1c2 a7a6 f3b7 b8b7 c2e2 d8c8 e2f3 b7f3 d4f3 d7c5 a1c1 f8d8 d1d8 c8d8 b2a3 d8c8 c3e2 f6e4');
INSERT INTO evaluations VALUES(45376,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45376,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(45376,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(45376,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45376,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45376,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45376,36,'cp -8','e7e6 b1d2 b8d7 f1d3 b7b6 d1e2 c8b7 e1g1 d8c7 e3e4 c5d4 c3d4 d5e4 d2e4 f8e7 c1g5 e8g8 a1c1 c7b8 f1e1 h7h6 g5h4 b8d8 e4c3 f6h5 h4e7 d8e7 e2e3 a8c8 e1d1 d7f6 f3e5');
INSERT INTO evaluations VALUES(47668,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(47668,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(47668,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(47668,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(47668,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(47668,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47668,36,'cp 31','f8d6 f4d6 d8d6 e2e3 e8g8 a1c1 a7a6 d1d2 b7b6 c4d5 e6d5 f1d3 c7c5 e1g1 c8g4 f3e5 b8c6 h2h3 g4e6 e5c6 d6c6 d4c5 b6c5 f1d1 c5c4 d3c2 a8b8 b2b3 c6b6 c3e2 f8c8 e2f4 c8c5 b3c4 d5c4 c1b1');
INSERT INTO evaluations VALUES(41623,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41623,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(41623,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(41623,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(41623,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(41623,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(41623,36,'cp 22','f4d6 d8d6 e2e3 e8g8 a1c1 b7b6 f1e2 c8b7 c4d5 f6d5 c3d5 b7d5 d1c2 f8c8 e1g1 a7a5 f1d1 a5a4 e2c4 b8d7 c4d5 d6d5 c2c6 a8a5 h2h3 a5b5 d1d2 b5b4 a2a3 b4b5 c1c2 d5c6 c2c6 b5b3');
INSERT INTO evaluations VALUES(47827,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(47827,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(47827,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47827,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(47827,36,'cp 0','b1c3 c7c5 d2d3 b8c6 a2a3 f8e7 f1e2 e8g8 e1g1 c8d7 c1d2 h7h6 d2e3 a7a5 f3d2 c6d4 a3a4 d7e6 c3d5 e6d7');
INSERT INTO evaluations VALUES(48351,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(48351,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(48351,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(48351,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(48351,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(48351,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(48351,36,'cp 0','b7b6 c2c4 c8b7 f1d3 f8d6 e1g1 b8d7 c4d5 e6d5 b2b3 e8g8 c1a3 c7c5 b1c3 a8c8 a1c1 f8e8 a3b2 d7f8 d3b5 e8e7 b5d3 e7e8');
INSERT INTO evaluations VALUES(49584,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(49584,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(49584,32,'cp 267',NULL);
INSERT INTO evaluations VALUES(49584,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(49584,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(49584,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(49584,36,'cp 244','d4e5 f6g4 e2e4 g4e5 f2f4 e5c6 c1d2 g7g6 d1f3 c6d4 f3d3 f8g7 e1c1 e8g8 d3g3 b8c6 g1f3 d7d5 c3d5 f8e8 d5c3 d8e7 h2h4 e7c5 g3g5 d4e6 g5c5 e6c5 e4e5 c8g4 f1e2 h7h5 d2e3 c5e6 c3d5 c6e7 e2c4 e7d5 c4d5 g7h6 g2g3 a8b8 d5e4');
INSERT INTO evaluations VALUES(49509,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(49509,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(49509,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(49509,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(49509,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(49509,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(49509,36,'cp 242','f6g4 e2e4 g4e5 f2f4 e5c6 c1d2 g7g6 d1f3 c6d4 f3d3 f8g7 e1c1 e8g8 g1e2 d7d5 c3d5 c7c6 d5e3 d4e2 f1e2 d8d3 e2d3 c8e6 e4e5 b8d7 h1e1 f7f6 d3c4 e6c4 e3c4 f6e5 f4e5 f8e8 d2f4 d7f8 c4d6 e8e7 f4g3 f8e6 d6e4 a8d8 h2h3 h7h5 e4d6');
INSERT INTO evaluations VALUES(49869,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(49869,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(49869,32,'cp 252',NULL);
INSERT INTO evaluations VALUES(49869,33,'cp 260',NULL);
INSERT INTO evaluations VALUES(49869,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(49869,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(49869,36,'cp 242','e2e4 g4e5 f2f4 e5c6 c1d2 g7g6 d1f3 c6d4 f3d3 f8g7 e1c1 e8g8 d3g3 b8c6 g1f3 d7d5 c3d5 f8e8 d5c3 d8e7 h2h4 e7c5 g3g5 d4e6 g5c5 e6c5 e4e5 c8g4 f1c4 c5e6 c4d5 a8d8 d2e3 h7h5 h1e1 c6e7 d5b7 d8d1 c3d1');
INSERT INTO evaluations VALUES(51771,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(51771,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(51771,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(51771,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(51771,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51771,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(51771,36,'cp 32','d2d4 d7d6 g1f3 d6e5 f3e5 b8d7 b1c3 d7e5 d4e5 d8d1 c3d1 c8d7 h2h4 g8e7 h4h5 c6c5 c1d2 e7c6 f2f4 e8c8 d1e3 f7f6 e1c1 f6e5 e3c4 b7b5 c4e5 c6e5 f4e5 d7c6 a2a4 a7a6 h1h3 f8e7 h3g3 h8g8 a4b5 a6b5 f1e2 d8d4 d1f1 d4e4 f1f7 e4e2');
INSERT INTO evaluations VALUES(54748,30,'cp 378',NULL);
INSERT INTO evaluations VALUES(54748,31,'cp 382',NULL);
INSERT INTO evaluations VALUES(54748,32,'cp 380',NULL);
INSERT INTO evaluations VALUES(54748,33,'cp 397',NULL);
INSERT INTO evaluations VALUES(54748,34,'cp 381',NULL);
INSERT INTO evaluations VALUES(54748,35,'cp 370',NULL);
INSERT INTO evaluations VALUES(54748,36,'cp 388','d1h5 e8d7 g1f3 d8e8 f3e5 d7d8 e5f7 d8d7 f5f6 g8f6 h5h3 e7e6 f7h8 f8e7 d2d4 e8h8 f1d3 d7d8 h3g3 f6h5 g3f3 g7g6 c2c3 h8f6 b1d2 c7c5 d4c5 e7c5 e1g1 b8c6 f3d1 d8c7 d2b3 c5b6 a2a4 a7a5 c1e3 e6e5 g1h1 c8e6 b3c5 e6g8 f1e1');
INSERT INTO evaluations VALUES(54746,30,'cp 380',NULL);
INSERT INTO evaluations VALUES(54746,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(54746,32,'cp 398',NULL);
INSERT INTO evaluations VALUES(54746,33,'cp 389',NULL);
INSERT INTO evaluations VALUES(54746,34,'cp 371',NULL);
INSERT INTO evaluations VALUES(54746,35,'cp 392',NULL);
INSERT INTO evaluations VALUES(54746,36,'cp 353','e8d7 g1f3 d8e8 f3e5 d7d8 e5f7 d8d7 f5f6 g8f6 h5h3 e7e6 f7h8 b8c6 b1c3 d7d8 h3g3 c8d7 f1e2 a7a6 d2d3 d8c8 e1g1 b7b6 f1e1 c6d4 c1f4 d4f5 g3g5 c8b7 e2g4 h7h6 g5g6 e8g6 h8g6 f6g4 h2h3 g4f6 c3e2 a8e8 d3d4 f8d6 g2g4 f5e7 g6e5 d7b5 a2a4 b5e2 e1e2 e7c6 e5c6');
INSERT INTO evaluations VALUES(20577,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(20577,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(20577,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(20577,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(20577,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(20577,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(20577,36,'cp 104','d2d4 e5d4 c3d4 c5d4 f3d4 c8d7 e1g1 f8e7 b1c3 g8f6 h2h3 e8g8 c1e3 a7a6 b5d3 c6e5 d3e2 f8e8 f2f4 e5c6 e2f3 a8c8 g1h1 b7b5 a2a3 c6d4 e3d4 d7e6 f1f2 e7f8 f2d2 e6c4 e4e5 d6e5 d4e5 d8a5 a1c1 g8h8 c3d5 f6d5');
INSERT INTO evaluations VALUES(20571,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(20571,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(20571,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(20571,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(20571,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(20571,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(20571,36,'cp 87','e5d4 c3d4 c5d4 f3d4 c8d7 b1c3 g8f6 e1g1 f8e7 f1e1 e8g8 b5f1 a7a6 c1e3 a8c8 d1d2 f6g4 d4c6 d7c6 e3d4 f8e8 a1d1 b7b5 a2a3 h7h6 f2f3 g4e5 d4f2 e7f8 h2h3 d8a5 f2e3 e5d7 g1h1 d7e5 c3d5 a5d2');
INSERT INTO evaluations VALUES(27924,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(27924,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(27924,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(27924,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(27924,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(27924,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(27924,36,'cp 139','c1e3 d7d6 d1d2 g8e7 e1c1 b8d7 h1e1 a7a6 d4d5 e6e5 h2h4 h7h6 h4h5 g6g5 g2g3 d7f6 f3h2 f6h5 d2e2 h5f6 f2f4 g5f4 g3f4 e5f4 e3d4 c7c5 d5c6 e7c6 h2f3 e8f8 d4f2 b6b5 e4e5 f6g4 f2h4 d8e8 c1b1 e8e6 e5d6 e6e2 d3e2 b5b4');
INSERT INTO evaluations VALUES(27925,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(27925,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(27925,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(27925,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(27925,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(27925,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(27925,36,'cp 114','d7d6 d1d2 g8e7 e1c1 b8d7 h1e1 a7a6 h2h4 h7h6 d4d5 e6e5 g2g3 d7c5 h4h5 c5d3 d2d3 g6h5 f3h4 d8d7 e1h1 g7f6 h4f5 e7f5 e4f5 h5h4 c3e4 d7e7 e4f6 e7f6 h1h4 e8c8 d1h1 h6h5 h4h5 h8h5 h1h5 d8h8 h5h8 f6h8');
INSERT INTO evaluations VALUES(27904,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(27904,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(27904,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(27904,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(27904,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(27904,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(27904,36,'cp 110','d1d2 b8d7 e1c1 g8e7 h1e1 h7h6 d4d5 e6e5 h2h4 a7a6 h4h5 g6g5 g2g3 d7f6 f3h2 f6h5 d2e2 h5f6 f2f4 g5f4 g3f4 e5f4 e3d4 c7c5 d5c6 e7c6 h2f3 c6d4 f3d4 f6d7 e4e5 g7e5 d3e4 d8c7 e4b7 c7b7 d4f5 e8f8 f5d6 b7c6 d6c4 a8e8 d1d7 c6d7 c4e5');
INSERT INTO evaluations VALUES(31336,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(31336,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(31336,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(31336,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(31336,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(31336,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(31336,36,'cp 137','g2g4 h5g6 f3e5 a7a6 f1g2 e7e6 c2c4 f8b4 b1c3 b8c6 c4d5 e6d5 e1g1 g8e7 e5g6 h7g6 a2a3 b4d6 c1g5 d8d7 a1c1 e8g8 c3a4 a8d8 f1e1 f7f6 g5e3 g8h7 a4b6 d7c7 d1b3 d6f4 b6d5 e7d5 g2d5 f4e3 e1e3 c7d7 d5c6 b7c6 g1g2 d7d4 c1c6');
INSERT INTO evaluations VALUES(31333,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(31333,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(31333,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(31333,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(31333,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(31333,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(31333,36,'cp 151','h5g6 f3e5 a7a6 f1g2 e7e6 c2c4 f8d6 c4d5 e6d5 b1c3 g8e7 c1g5 f7f6 e5g6 h7g6 g5d2 b8c6 c3d5 e7d5 g2d5 d8d7 d1f3 d6e7 d2e3 a8d8 d5c6 b7c6 a1c1 d7d5 f3d5 c6d5 e1e2 e8d7 h3h4 e7d6 e2d3 d8e8 h4h5 g6h5 h1h5 h8h5 g4h5 e8e6 c1h1 e6e7 f2f3 d7e8');
INSERT INTO evaluations VALUES(31265,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(31265,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(31265,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(31265,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(31265,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(31265,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(31265,36,'cp 140','f3e5 b8c6 h3h4 d8c7 c1f4 c6e5 f4e5 c7c2 d1c2 g6c2 b1c3 a8c8 e1d2 f7f6 e5g3 c2g6 a1c1 e8d8 f1e2 e7e5 c3d5 c8c6 d2e3 g6f7 e2f3 h7h5 g4g5 f6g5 d4e5 g8h6 h4g5 h5h4 g5h6 h4g3 c1d1 c6h6 h1h6 f8c5 e3f4 h8h6 f2g3 f7e6 d5c3 d8e7 f3b7');
INSERT INTO evaluations VALUES(39208,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(39208,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(39208,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(39208,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39208,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(39208,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39208,36,'cp -12','c5d4 g1f3 b8c6 f3d4 c6d4 c1e3 d4c6 b1c3 g8f6 g2g4 h7h6 f2f3 d7d6 d1d2 c8e6 c3d5 h6h5 g4g5 f6d7 g5g6 c6d4 e3d4 e5d4 g6f7 e6f7 d2d4 d8h4 d4f2 h4f2 e1f2 e8g8 a1d1 d7e5 c4e2 f7d5 d1d5 f8f4 f2e3 a8f8 h2h4 g7g6 h1h3 f8f7 d5d1 b7b6 d1g1 a7a5 b2b3 f7f6 g1d1 g8f7 d1g1');
INSERT INTO evaluations VALUES(39373,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(39373,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(39373,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(39373,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(39373,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(39373,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(39373,36,'cp -21','g1f3 b8c6 f3d4 c6d4 c1e3 d4c6 b1c3 g8f6 f2f3 d7d6 c4e2 c8e6 d1d2 a7a6 e1c1 d8d7 h1e1 c6e7 e3g5 e8c8 g5f6 g7f6 f3f4 e5f4 d2f4 e7g6 f4f6 c8b8 e1f1 h7h5 a2a3 h5h4 c1b1 d8e8 f6d4 g6e5 f1f2 d7d8 d1f1 h8g8');
INSERT INTO evaluations VALUES(39372,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(39372,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(39372,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(39372,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(39372,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(39372,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(39372,36,'cp -13','b8c6 f3d4 c6d4 c1e3 d4c6 b1c3 g8f6 g2g4 h7h6 f2f3 d7d6 d1d2 c8e6 c3d5 h6h5 g4g5 f6d7 g5g6 c6d4 g6f7 e6f7 e3d4 e5d4 d2d4 d8h4 d4f2 h4f2 e1f2 e8g8 a1d1 d7e5 c4e2 f7d5 d1d5 f8f4 f2e3 a8f8 h2h4 b7b6 h1h3 a7a5 a2a3 f8f7 b2b3 g7g6 b3b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(6638,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(6638,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(6638,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(6638,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(6638,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(6638,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(6638,36,'cp -11','f3d4 c6d4 c1e3 d4c6 b1c3 g8f6 f2f3 d7d6 c4e2 c8e6 d1d2 d8d7 e1c1 e8c8 c3d5 e6d5 e4d5 c6e7 e2c4 c7c5 d5c6 e7c6 e3f2 d6d5 c4b3 h7h6 h1e1 h8e8 a2a3 d5d4 c1b1 f6d5 c2c3 d7f5 b3c2 f5f4 c3d4 c6d4 f2d4 f4d2 d1d2 e5d4 e1e8 d8e8');
INSERT INTO evaluations VALUES(6636,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(6636,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(6636,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(6636,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(6636,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(6636,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(6636,36,'cp -17','c6d4 c1e3 d4c6 b1c3 g8f6 c4e2 d7d6 f2f3 c8e6 d1d2 d8d7 e1c1 e8c8 c3d5 e6d5 e4d5 c6e7 e2c4 c7c6 d5c6 e7c6 e3f2 d6d5 c4b3 h7h6 h1e1 h8e8 a2a3 d7c7 c1b1 a7a6 f2h4 g7g5 h4f2 c6a5 h2h4 a5c4 d2b4 e8e6 h4g5 h6g5 e1h1 b7b5 b3c4 b5c4');
INSERT INTO evaluations VALUES(6894,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(6894,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(6894,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(6894,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(6894,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(6894,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(6894,36,'cp -7','c1e3 d4c6 b1c3 g8f6 c4e2 d7d6 f2f3 c8e6 d1d2 d8d7 e1c1 a7a6 h1e1 h7h6 c1b1 c6e7 b2b3 e8c8 b1b2 c8b8 e3f2 e7c6 c3d5 e6d5 e4d5 c6d4 f2d4 e5d4 d2d4 d8e8 d4b4 b8a7 e2d3 e8e1 d1e1 a7a8 b4c3 h8d8 b2c1 a8b8 c1b2 d8e8 e1e8 d7e8');
INSERT INTO evaluations VALUES(49693,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(49693,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(49693,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49693,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(49693,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49693,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(49693,36,'cp -9','d2d4 e5d4 e3d4 d7d5 f1d3 f8d6 g1e2 e8g8 h2h3 f8e8 e1g1 b8d7 c1f4 d7f8 f4g3 f6e4 d3e4 e8e4 b1d2 e4e7 d2f3 f8g6 d1c2 f7f6 f1e1 c8e6 a2a4 e6f7 g3d6 d8d6 e2g3 a8e8 g3f5 e7e1 a1e1 d6b6 g1h1 e8e1 f3e1');
INSERT INTO evaluations VALUES(49585,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(49585,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(49585,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(49585,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(49585,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(49585,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(49585,36,'cp -1','e5d4 e3d4 d7d5 f1d3 f8d6 g1e2 e8g8 h2h3 f8e8 e1g1 b8d7 f1e1 d7f8 b1d2 c7c6 d2f3 f6e4 c1f4 c8f5 f4d6 d8d6 f3e5 f7f6 e5g4 f5g6 a2a4 e8e7 g4e3 a8e8 d3e4 g6e4 e2g3 f8g6 d1g4 g6f4 g3e4 e7e4');
INSERT INTO evaluations VALUES(15198,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(15198,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(15198,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(15198,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(15198,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(15198,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(15198,36,'cp -55','b1d2 a7a5 g1f3 c8f5 c2c4 d5d4 e3d4 e5d4 f1e2 f8e7 e1g1 h7h6 f3e1 e8g8 f2f4 f8e8 g2g4 f5h7 e2f3 e7d6 f4f5 e8e3 f3c6 b7c6 d2f3 e3e8 e1g2 c6c5 c1f4 d6f4 g2f4 d8d6 d1d2 e8e3 d2h2 g7g5 f5g6 h7g6 a1e1 a8e8 e1e3 e8e3 f4g6 f7g6 h2d6');
INSERT INTO evaluations VALUES(22227,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(22227,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(22227,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(22227,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(22227,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(22227,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(22227,36,'cp 130','c2c4 f8b4 b1c3 g8e7 a1c1 e8g8 a2a3 b4c3 c1c3 c6a5 c4c5 c8d7 h2h4 a5c4 d1c1 e6e5 d4e5 c4e5 f3e5 f6e5 f4e5 e7f5 c1d1 c7c6 f1d3 d8e7 d3f5 f8f5 e5d6 e7e6 e1g1 a8e8 d1b3 d7c8 d6g3 e6f6 b3c2 f6g6 c3b3 h7h6 f1c1');
INSERT INTO evaluations VALUES(30801,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(30801,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(30801,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(30801,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(30801,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(30801,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(30801,36,'cp 75','e2e4 e7e6 f1d3 g8f6 h2h4 a7a6 e4e5 f6d5 c3d5 b7d5 d1g4 c7c5 c2c4 d5b7 d4d5 e6d5 c1g5 d8c8 h4h5 d5c4 d3f5 b8c6 g1f3 c6d4 f3d4 c5d4 g4d4 f8b4 e1f1 c8c6 d4g4 g7g6 h5g6 f7g6 a1d1 c6g2 g4g2 b7g2 f1g2 g6f5 g5f6 h8g8 g2f3 a8c8 h1h7 c4c3 b2c3 c8c3 f3f4 c3c2');
INSERT INTO evaluations VALUES(37162,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37162,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37162,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(37162,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37162,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37162,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37162,36,'cp 0','c7c5 e2e3 c5d4 d1d4 b8c6 d4d3 h7h6 g5h4 d7d5 a1c1 b4e7 a2a3 d5c4 d3c4 e6e5 c1d1 d8c7 c4a4 a7a6 f1d3 c8e6 e1g1 f8d8 h4f6 e7f6 d3e4 c6e7 a4b4 e7c6');
INSERT INTO evaluations VALUES(35668,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(35668,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(35668,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(35668,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(35668,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(35668,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35668,36,'cp 0','e2e3 c5d4 d1d4 b8c6 d4d3 h7h6 g5h4 d7d5 a1c1 g7g5 h4g3 f6e4 d3c2 h6h5 h2h4 e4g3 f2g3 g5g4 f3g5 f7f5 c4d5 d8d5 a2a3 d5e5 a3b4 e5e3 c3e2 c6b4 c2c7 b4d3 e1d1 d3f2 d1e1 f2d3');
INSERT INTO evaluations VALUES(49783,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49783,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49783,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49783,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49783,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49783,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(49783,36,'cp 0','f1e2 f8c5 c1f4 b8c6 b1c3 d7d5 e4e5 f6d7 g1f3 c6d4 c3d5 e8g8 b2b4 d4f3 e2f3 c5d4 c2c3 d4e5 e1g1 e5f4 d5f4 c7c6 d3d4 a7a5 a2a3 a5b4 a3b4 a8a1 d1a1 d7f6 a1a5 d8e8 a5e5 c8g4 e5e8 f8e8 f3g4 f6g4 d4d5 c6d5 f4d5 e8e2 f1d1 g8f8 h2h3 g4e3 d5e3 e2e3 d1d7 e3c3 d7b7 c3d3');
INSERT INTO evaluations VALUES(53857,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(53857,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(53857,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(53857,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(53857,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(53857,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(53857,36,'cp 100','a2a3 b8d7 g1f3 a7a6 g2g4 b7b5 g4g5 d6d5 h2h4 b5b4 a3b4 f8b4 c1d2 g8e7 f1g2 a6a5 e1g1 e8g8 f1e1 c8b7 h4h5 d8c7 h5h6 e7g6 h6g7 g8g7 e4e5 g7h8 f3h2 g6h4 h2g4 h4f5 c3e2 f8g8 e2g3 f5g3 f2g3 b4d2 d1d2');
INSERT INTO evaluations VALUES(53892,30,'cp 323',NULL);
INSERT INTO evaluations VALUES(53892,31,'cp 303',NULL);
INSERT INTO evaluations VALUES(53892,32,'cp 318',NULL);
INSERT INTO evaluations VALUES(53892,33,'cp 299',NULL);
INSERT INTO evaluations VALUES(53892,34,'cp 323',NULL);
INSERT INTO evaluations VALUES(53892,35,'cp 301',NULL);
INSERT INTO evaluations VALUES(53892,36,'cp 294','d4e5 b8c6 f1b5 f7f6 e5f6 g8f6 e1g1 f8e7 e4e5 d6e5 f3e5 d8d1 f1d1 c8d7 e5d7 e8c8 c1f4 d8d7 d1d7 c8d7 b1c3 h8d8 g1f1 e7d6 a1d1 d7c7 d1d6 d8d6 b5d3 c7d7 f4d6 d7d6 c3e4 f6e4 d3e4 h7h6 f1e2 b7b5 e4c6 d6c6 e2e3 c6d5 g2g4 g7g5 f2f4 a7a5 f4f5 d5e5 b2b3 b5b4');
INSERT INTO evaluations VALUES(53871,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(53871,31,'cp 314',NULL);
INSERT INTO evaluations VALUES(53871,32,'cp 319',NULL);
INSERT INTO evaluations VALUES(53871,33,'cp 305',NULL);
INSERT INTO evaluations VALUES(53871,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(53871,35,'cp 328',NULL);
INSERT INTO evaluations VALUES(53871,36,'cp 291','b8c6 f1b5 f7f6 e5f6 g8f6 e1g1 f8e7 e4e5 d6e5 f3e5 e8g8 e5c6 b7c6 d1d8 f8d8 b5c6 a8b8 b1c3 c8f5 f1e1 g8f7 e1e2 f5g4 f2f3 g4e6 a1b1 e7d6 c6b5 f6d5 c3d5 b8b5 d5c3 b5b6 c1g5 d8e8 b2b3 c5c4 c3e4 d6f8 g5f4 h7h6 f4e3 b6a6');
INSERT INTO evaluations VALUES(6317,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(6317,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(6317,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6317,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6317,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6317,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6317,36,'cp 0','g8f6 d2d3 e8g8 e1g1 d7d6 a2a4 a7a5 b4b5 c6e7 b1d2 e7g6 c4b3 f8e8 d2c4 b6a7 f3g5 e8f8 g5f3');
INSERT INTO evaluations VALUES(9251,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(9251,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(9251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9251,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(9251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9251,36,'cp 0','e6e5 f1b5 e5e4 g1e2 a7a6 b5c6 b7c6 c3c4 e8g8 d1a4 c8g4 g3h4 a8b8 b2b3 g4d7 c4c5 d6e7 a4a6 b8a8 a6b7 a8b8');
INSERT INTO evaluations VALUES(10949,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(10949,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(10949,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(10949,33,'cp -150',NULL);
INSERT INTO evaluations VALUES(10949,34,'cp -143',NULL);
INSERT INTO evaluations VALUES(10949,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(10949,36,'cp -153','d7d5 e4e5 f6g4 g1f3 c6d4 b1a3 e8g8 f1e2 d4f3 e2f3 e7h4 g3f4 g4e5 e1g1 h4f6 f3h5 e5g6 f4f3 f8e8 c2c3 d5d4 h5g6 h7g6 c1d2 c8f5 f1e1 d8d7 c3c4 c7c5 f3g3 a7a6 h2h3 g6g5 a1d1 g5g4 h3g4 f5g4 f2f3 g4f5 d2g5 f6g5 g3g5 f7f6 g5g3 e8e1 g3e1');
INSERT INTO evaluations VALUES(18048,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(18048,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(18048,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(18048,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(18048,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(18048,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(18048,36,'cp 89','c4d5 g8e7 e1g1 d7d6 a2a4 b5b4 d5a2 h7h6 d3d4 e7g6 c2c3 a8b8 b1d2 b4c3 b2c3 f8e7 d2c4 e5d4 f3d4 c6d4 c3d4 e8g8 f2f4 c8b7 d4d5 b7c8 c1a3 f8e8 a1b1 c8d7 b1b8 d8b8 d1c2 b8a7 g1h1 e7f6 h2h3 d7c8 a2b1 c8d7 e4e5');
INSERT INTO evaluations VALUES(18043,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(18043,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(18043,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(18043,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(18043,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(18043,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(18043,36,'cp 97','g8e7 e1g1 e7d5 e4d5 c6e7 f3e5 d7d6 e5c6 e7c6 d5c6 f8e7 b1c3 c8e6 f1e1 e8g8 d1f3 a8b8 c3e2 e7f6 d3d4 d6d5 c2c3 b8b6 e2f4 b6c6 f4d3 f6e7 h2h3 e6c8 c1f4 f8e8 e1e5 c6e6 e5d5 e7d6 a2a4 b5a4 d5a5 c8d7 f4d6 c7d6 a5a6 d7b5 a6a7');
INSERT INTO evaluations VALUES(18732,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(18732,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(18732,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(18732,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(18732,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(18732,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(18732,36,'cp -96','g8f6 c4b3 c6d4 f3d3 b7b5 c3b5 d4b5 d3b5 f6e4 g1f3 d8c7 b5e2 c8b7 e1g1 f8d6 d2d3 e4f6 h2h3 e8g8 f1e1 a8b8 c2c3 b7a8 d3d4 c5d4 f3d4 f8c8 b3c2 a7a5 a2a4 h7h6 d4b5 c7c6 f2f3 d6g3 e1f1 f6d5 g1h1 a8b7 f1d1 d5f4 c1f4 g3f4');
INSERT INTO evaluations VALUES(12354,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(12354,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(12354,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(12354,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(12354,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(12354,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(12354,36,'cp -74','c4b3 c6d4 f3f4 b7b5 g1f3 d4f3 g2f3 a7a6 d2d3 d7d5 a2a3 c8b7 c1d2 f8e7 h2h4 e8g8 h4h5 d8b8 f4g3 b8g3 f2g3 h7h6 b3a2 d5e4 d3e4 f6d7 e1f2 f8c8 h1d1 e7f6 d2e3 b7c6 a1b1 a6a5 c3e2 c8d8 c2c3 f6e7 e3f4 a5a4 f4d6 e7f8 b1c1');
INSERT INTO evaluations VALUES(19713,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(19713,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(19713,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(19713,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(19713,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(19713,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(19713,36,'cp 0','g8f6 g1f3 c8g4 e1g1 e7e6 h2h3 g4f3 d1f3 d8b6 f3g3 a8c8 f1d1 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(13057,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13057,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13057,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13057,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13057,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13057,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13057,36,'cp 0','g1f3 c8g4 h2h3 g4f3 d1f3 e7e6 e1g1 d8b6 f3g3 a8c8 f1d1 f6h5 g3d3 h5f6');
INSERT INTO evaluations VALUES(20246,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(20246,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(20246,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(20246,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(20246,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(20246,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(20246,36,'cp -37','e4d5 e6d5 c4e2 g8f6 d2d4 c5c4 e1g1 f8d6 c1g5 h7h6 g5f6 d8f6 b1c3 c8e6 e2c4 d5c4 d4d5 e8c8 d1e2 e6g4 d5c6 g4f3 e2f3 f6f3 g2f3 b7c6 f1e1 h8e8 a1d1 e8e1 d1e1 d6f4 c3a4 f7f5 e1e7 d8d7 e7e8 c8c7 e8f8 f4c1 f8f5 d7d2 f5f7 c7d6 f7g7 d2c2 g7g6 d6d5');
INSERT INTO evaluations VALUES(20224,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(20224,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(20224,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(20224,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(20224,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(20224,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(20224,36,'cp -27','e6d5 c4e2 g8f6 e1g1 f8d6 d2d4 c5d4 f3d4 e8g8 b1c3 f8e8 c1e3 d6c7 f1e1 h7h6 e2f3 a7a6 d4c6 b7c6 c3a4 c8f5 c2c4 f6e4 c4d5 c6d5 a4c3 e4c3 b2c3 f5e4 f3e4 d5e4 g2g3 d8f6 d1d5 f6c3 d5c4 c3c4 a1c1 c7e5 c1c4');
INSERT INTO evaluations VALUES(20507,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(20507,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(20507,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(20507,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(20507,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(20507,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(20507,36,'cp -27','c4e2 g8f6 e1g1 f8d6 d2d4 c5c4 c1g5 h7h6 g5h4 g7g5 h4g3 d6g3 h2g3 d8b6 b1c3 g5g4 f3e5 b6d4 d1d4 c6d4 a1d1 d4e2 c3e2 e8g8 e2f4 c8e6 f2f3 f8e8 f4e6 e8e6 e5g4 f6g4 f3g4 a8d8 f1f4 d8d6 b2b3 c4b3 c2b3 g8g7 g1f2 g7g6 b3b4 b7b6 a3a4 e6e8 f4d4');
INSERT INTO evaluations VALUES(21813,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(21813,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(21813,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(21813,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(21813,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(21813,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(21813,36,'cp -26','h2h3 c8e6 d2d3 f8b4 c2c3 d4c3 e2c3 g8e7 f1e2 e8g8 a2a3 b4c5 b2b4 c5d4 c1d2 e7g6 e1g1 g6h4 b4b5 h4f3 e2f3 c6e7 a1c1 a7a6 f3g4 e6g4 d1g4 a6b5 c3b5 d4b6 d2b4 f8e8 c1c3 e7c6 b4c5 b6a5 c3b3 a5d2 c5e3 d2e3 f2e3 e8f8 g1h1 g8h8 f1b1 f7f6');
INSERT INTO evaluations VALUES(25688,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(25688,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(25688,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(25688,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(25688,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(25688,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(25688,36,'cp 56','h2h3 f8b4 b1d2 g4f3 d1f3 g8e7 a2a3 b4d6 h3h4 d8d7 d2c4 h7h5 c1e3 c6c5 f3g3 e7g6 a3a4 b7b6 e1e2 a8d8 b2b3 d7e6 g3h3 e6h3 h1h3 f7f6 g2g3 e8f7 h3h1 g6e7 f2f4 e7c6 c2c3 g7g6 h1f1 d8g8 f4e5 d6e5 c4e5 c6e5 e3f4');
INSERT INTO evaluations VALUES(25686,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(25686,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(25686,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(25686,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25686,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(25686,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(25686,36,'cp 50','g4f3 d1f3 f8b4 b1d2 g8e7 f3g3 e7g6 a2a3 b4d6 d2c4 d8d7 c1e3 c6c5 a3a4 b7b6 g3g4 h7h5 g4d7 e8d7 h3h4 g6f8 e1e2 f8e6 c2c3 f7f6 g2g3 g7g6 a1f1 a8f8 c4d2 f8e8 f1a1 e8a8 h1g1 a8e8 g1f1 e8f8');
INSERT INTO evaluations VALUES(25694,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(25694,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(25694,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(25694,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(25694,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(25694,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(25694,36,'cp 59','d1f3 f8b4 b1d2 g8e7 f3g3 e7g6 a2a3 b4d6 d2c4 d8d7 c1e3 c6c5 a3a4 b7b6 g3g4 h7h5 g4d7 e8d7 h3h4 g6f8 e1e2 f8e6 c2c3 g7g6 g2g3 f7f6 h1f1 g6g5 h4g5 f6g5 f1h1 h5h4 e2f3 a8f8 f3g4 e6d8 g3h4 g5h4');
INSERT INTO evaluations VALUES(25693,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(25693,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(25693,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(25693,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(25693,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(25693,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(25693,36,'cp 55','f8d6 f3g3 g7g6 b1d2 g8f6 d2c4 f6h5 g3f3 h5g7 c1h6 g7e6 e1c1 d8e7 g2g3 e8c8 f3g4 b7b5 c4a5 e7d7 f2f4 e6d4 g4d7 d8d7 f4e5 d6e5 g3g4 f7f6 c2c3 d4e2 c1d2 e2g3 h1e1 c6c5 a5c6 h8e8 d2c2');
INSERT INTO evaluations VALUES(39618,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(39618,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(39618,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(39618,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(39618,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(39618,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(39618,36,'cp 53','d2d4 d7d5 e4e5 f7f6 g1f3 f6e5 f3e5 b8d7 e5d7 c8d7 b1d2 g8h6 d2f3 e8g8 f1e2 c7c6 e1g1 h6f7 c1f4 d8b6 h2h3 a8e8 f4e3 d7f5 d1d2 b6c7 f1e1 e7e5 d4e5 f7e5 a1d1 e5f3 e2f3 f5e4 f3e4 e8e4 f2f3 e4c4 e3f2 b7b6 d2e2');
INSERT INTO evaluations VALUES(43245,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(43245,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43245,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43245,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43245,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(43245,36,'cp 0','d7d5 e1g1 e8g8 c2c4 d5c4 e2c4 c7c5 d4c5 b8d7 c5c6 d7b6 c4e2 b7c6 b1c3 f6d5 c3e4 a7a5 a2a3 d8c7 e4c5 c7d6 c5e4');
INSERT INTO evaluations VALUES(41505,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41505,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41505,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41505,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41505,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41505,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41505,36,'cp 0','e1g1 e8g8 c2c4 d5c4 e2c4 c7c5 d4c5 b8d7 c5c6 b7c6 b1c3 d7b6 c4e2 f6d5 c3e4 a7a5 a2a3 d8c7 e4c5 c7d6 c5e4');
INSERT INTO evaluations VALUES(41504,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(41504,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(41504,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(41504,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41504,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41504,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41504,36,'cp 0','e8g8 b2b4 a7a5 b4b5 c7c5 c1b2 c5c4 b1c3 b8d7 f3e5 d7b6 f2f4 f6e8 a2a4 f7f6 e5f3 e8d6 b2a3 c8e6 a3c5 b6d7 c5a3');
INSERT INTO evaluations VALUES(36779,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36779,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36779,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36779,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36779,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36779,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36779,36,'cp 0','b2b4 a7a5 b4b5 c7c5 c1b2 b8d7 c2c4 d5c4 b1d2 c4c3 b2c3 f6d5 c3b2 c5d4 f3d4 d7b6 a2a4 e7e5 d4b3 d5b4 d2e4 b6d5 e2c4 d5b6');
INSERT INTO evaluations VALUES(44049,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(44049,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44049,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(44049,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(44049,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(44049,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(44049,36,'cp 29','c1f4 d7d5 e2e3 f8d6 f4d6 d8d6 b1c3 e8g8 a1c1 f8d8 h2h3 h7h6 c4d5 c6d5 c3b5 d6b4 d1d2 b8c6 d2b4 c6b4 a2a3 b4c6 f1d3 c8d7 c1c2 f6e8 e1g1 a7a6 b5c3 d8c8 f1c1 e8d6 c3a4 d7e8 a4c5 c8c7');
INSERT INTO evaluations VALUES(46115,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46115,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46115,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(46115,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46115,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46115,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(46115,36,'cp 9','a7a6 e1g1 b7b5 c4e2 b8d7 d4d5 e6d5 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 c3b5 f8e7 b2b4 c5b4 d1a4 e8g8 c1d2 d7c5 a4b4 c5e4 b4b2 e4d2 f3d2 a8c6 b5d4 c6d7 f1b1 d8c7 g2g3 h7h6 e2d3 e7d6 d2f3 f8b8 b2b8 c7b8 b1b8 d6b8 g1g2 g7g6 d3b5 d7b5 d4b5');
INSERT INTO evaluations VALUES(43558,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(43558,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43558,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(43558,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(43558,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(43558,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43558,36,'cp 9','e1g1 b7b5 c4e2 b8d7 d4d5 e6d5 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 c3b5 f8e7 b2b4 c5b4 d1a4 e8g8 c1d2 d7c5 a4b4 c5e4 b4d4 e4d2 f3d2 f6e4 d2e4 d5e4 d4a7 e7f6 e2c4 g7g6 g2g3 d8e7 a7e7 f6e7 f1a1 f8c8 b5d4 a8b7 a1a7 c8c4 a7b7');
INSERT INTO evaluations VALUES(43557,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(43557,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(43557,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(43557,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43557,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(43557,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(43557,36,'cp 11','b7b5 c4e2 b8d7 d4d5 e6d5 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 c3b5 f8e7 b2b4 c5b4 d1a4 e8g8 c1d2 d7c5 a4b4 f6e4 d2c1 d8b6 c1a3 f8b8 f3d4 e7f8 b4b2 a8c6 f1b1 c6b5 b2b5 b6b5 b1b5 b8b5 d4b5 c5e6 a3b2 f8e7 f2f3 e4f6 g1f1 f6d7 f3f4 e7c5 b5d4');
INSERT INTO evaluations VALUES(56892,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(56892,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(56892,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(56892,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(56892,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(56892,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(56892,36,'cp 192','f8g7 d2d4 g8f6 f1d3 d7d5 e4d5 f6d5 e1g1 e8g8 e5f3 f8e8 c1g5 g7f6 g5f6 d8f6 d3c4 d5b6 c4b3 b8c6 c2c3 c8f5 b1d2 g8g7 f1e1 e8e1 d1e1 c6a5 d2e4 f5e4 e1e4 c7c6 e4d3 a8e8 g2g3 a7a6 g1g2 b6d7 a1e1 e8e1 f3e1 a5b3 a2b3 f6f5');
INSERT INTO evaluations VALUES(39047,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(39047,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(39047,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(39047,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(39047,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(39047,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(39047,36,'cp 207','d2d4 g8f6 b1c3 d7d6 e5f3 e8g8 f1d3 c7c5 d4d5 f8e8 e1g1 c5c4 d3c4 f6e4 c3e4 e8e4 c4d3 e4e8 c1g5 d8c7 f1e1 c8d7 c2c3 e8e1 d1e1 b8a6 d3a6 b7a6 e1d2 c7c4 g5h6 g7f6 a1e1 a8e8 e1e8 d7e8 d2e3 e8b5 h2h4 c4c5 e3c5 d6c5 h6e3 b5c4 d5d6 f6d8 e3c5 c4a2 c5a7');
INSERT INTO evaluations VALUES(4797,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(4797,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(4797,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(4797,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(4797,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(4797,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(4797,36,'cp -37','c6c5 d2d3 c8g4 a2a4 g4h5 h2h3 g8e7 a4a5 e7c6 c3d5 d8d7 c1e3 e8c8 f3e5 h5d1 e5d7 d1c2 d7c5 d6c5 e3c5 c2d3 f1e1 h8e8 f2f3 f7f5 d5b4 c6b4 c5b4 f5e4 f3e4 d3e4 a1d1 d8d1 e1d1 g7g6 d1f1 e4d5 f1d1 d5b3 d1e1 e8e1 b4e1 c8d7 g1f2 d7c6 g2g4 c6b5 e1c3 b5a4 h3h4 b3e6 f2g3 b7b5');
INSERT INTO evaluations VALUES(6606,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6606,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(6606,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(6606,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(6606,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(6606,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(6606,36,'cp -8','g1e2 g8e7 a2a3 b4c3 e2c3 d7d5 e4d5 e7d5 f1e2 e8g8 e1g1 c6d4 f1e1 d5c3 b2c3 d4e2 d1e2 c8d7 f2f4 e5f4 c1f4 d8f6 e2d2 c7c5 e1e5 b7b6 e5d5 d7a4 d5d6 f6f5 a1f1 a8e8 f4g3 f5h5 f1f3 f7f6 d2f2 h5f7 f3f4 a4d7 c3c4 e8e6');
INSERT INTO evaluations VALUES(7371,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(7371,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(7371,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(7371,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(7371,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(7371,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(7371,36,'cp 13','b7b5 a4b3 d7d6 a2a4 c8d7 h2h3 e8g8 f1e1 c6a5 b3a2 b5a4 b1c3 c7c5 f3d2 h7h6 d2f1 a8b8 f1e3 g8h7 e1f1 d8e8 c3d5 a5c6 d5e7 e8e7 a2c4 c6a5 c4d5 f6d5 e3d5 e7d8 f2f4 f7f5 f4e5 d6e5 e4f5');
INSERT INTO evaluations VALUES(7499,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(7499,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(7499,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(7499,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(7499,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(7499,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(7499,36,'cp 22','a4b3 d7d6 a2a4 c8d7 c2c3 e8g8 b3c2 f8e8 b1d2 h7h6 f1e1 e7f8 d2f1 a8b8 f1g3 b5b4 d3d4 c6a5 h2h3 b4c3 b2c3 c7c6 g3f1 d8c7 f1d2 b8b6 d1e2 d7e6 c1a3 a5b3 a1d1 b3d2 f3d2 a6a5');
INSERT INTO evaluations VALUES(11470,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(11470,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(11470,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(11470,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(11470,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(11470,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(11470,36,'cp 9','f8c5 c1e3 d7d6 c3d5 c5e3 f2e3 e8g8 d1e1 f6d5 c4d5 c8e6 d5e6 f7e6 e1g3 d8e7 f3h4 f8f1 a1f1 a8f8 a2a3 f8f1 g1f1 c6b8 g3e1 b8d7 h4f3 e7e8 e1a5 e8c8 f1e1 c7c5 b2b3 b7b6 a5c3 d7f6 e1d2 b6b5 d2c1 f6g4 c3e1 c5c4 d3c4 b5c4');
INSERT INTO evaluations VALUES(17379,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17379,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(17379,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17379,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(17379,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(17379,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(17379,36,'cp 64','d2d4 c5d4 f3d4 d8c7 a2a3 g8f6 c1e3 f8d6 g2g3 e8g8 f2f4 d6e7 e4e5 f6e8 d4c6 b7c6 d1c2 d7d6 f1d3 h7h6 e5d6 e8d6 e1c1 d6f5 e3f2 a8b8 d1d2 c6c5 d3f5 e6f5 c3d5 c7b7 d5e7 b7e7 d2d5 e7b7 f2c5 c8e6 c5f8 e6d5 c4d5');
INSERT INTO evaluations VALUES(17383,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(17383,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(17383,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17383,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(17383,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(17383,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(17383,36,'cp 62','c5d4 f3d4 d8c7 c1e3 g8f6 a1c1 f8b4 f1d3 c6e5 f2f3 b4c5 d3e2 e8g8 e1g1 d7d6 a2a3 c8d7 f3f4 e5g6 b2b4 c5a7 d1d2 a8c8 d4b3 c7b8 g1h1 a7e3 d2e3 b7b6 b3d4 e6e5 f4e5 d6e5 d4f5 d7f5 f1f5 g6f4');
INSERT INTO evaluations VALUES(17429,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17429,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(17429,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(17429,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(17429,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(17429,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(17429,36,'cp 70','f3d4 d8c7 c1e3 g8f6 a1c1 f8b4 f1d3 c6e5 f2f3 b4c5 d3e2 e8g8 e1g1 d7d6 g1h1 h7h6 a2a3 c8d7 b2b4 c5a7 f3f4 e5g6 a3a4 c7d8 c4c5 d6c5 b4c5 a7c5 d4e6 d7e6 e3c5 d8d1 c1d1 f8c8 c5d4 g6e7 e2f3 e7c6 d4f6 g7f6 c3d5');
INSERT INTO evaluations VALUES(17426,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(17426,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(17426,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(17426,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(17426,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(17426,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17426,36,'cp 62','d8c7 c1e3 g8f6 a1c1 f8b4 f1d3 c6e5 f2f3 b4c5 d3e2 d7d6 e1g1 e8g8 a2a3 c8d7 f3f4 e5g6 b2b4 c5a7 d1d2 a8d8 d4c2 a7e3 d2e3 d8c8 a3a4 e6e5 f4e5 d6e5 f1f6 g7f6 c3d5 c7d8 e3h6 g6f4 d5f6 d8f6 h6f6 f4e2 g1h1 e2c1');
INSERT INTO evaluations VALUES(1258,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(1258,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(1258,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(1258,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(1258,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(1258,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(1258,36,'cp 71','c1e3 g8f6 a1c1 f8b4 f1d3 c6e5 f2f3 b4c5 d3e2 e8g8 e1g1 d7d6 g1h1 c7e7 d1d2 c8d7 c1d1 a6a5 e3g1 f8c8 a2a4 e5g6 b2b3 c8d8 e2d3 a8c8 d4e2 d7c6 f3f4 c5g1 h1g1 h7h6 d3c2 d6d5 e4e5 f6e4');
INSERT INTO evaluations VALUES(1257,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(1257,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(1257,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(1257,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(1257,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(1257,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(1257,36,'cp 70','g8f6 a1c1 f8b4 f1d3 c6e5 f2f3 b4c5 d3e2 e8g8 e1g1 d7d6 g1h1 h7h6 a2a3 c8d7 f3f4 e5g6 b2b4 c5a7 d1e1 a8c8 e1g3 f8e8 c1d1 g8h8 a3a4 d7c6 e2d3 a7d4 e3d4 e6e5 f4e5 d6e5 d4e3 f6h5 g3f3 h5f4 d3b1');
INSERT INTO evaluations VALUES(658,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(658,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(658,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(658,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(658,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(658,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(658,36,'cp 76','a1c1 f8b4 f1d3 c6e5 f2f3 b4c5 d3e2 e8g8 e1g1 d7d6 g1h1 h7h6 a2a3 c8d7 b2b4 c5a7 f3f4 e5g6 d1e1 d7c6 e2d3 f8e8 c1d1 c6d7 e3g1 a8c8 d4e6 d7e6 g1a7 b7b6 f4f5 c7a7 f5g6 f7g6 e1g3 g6g5 c3d5 e6d5 e4d5 e8e5 d1e1 e5e1 g3e1 c8e8 e1f2 f6e4');
INSERT INTO evaluations VALUES(25094,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(25094,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(25094,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(25094,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(25094,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25094,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(25094,36,'cp 33','c1f4 f8e7 e1c1 e8g8 h1e1 f8e8 d4c4 d8c8 f3g5 e8f8 f2f3 h7h6 g5h3 f8e8 c1b1 c8d7 e4e5 f6h7 e5d6 c7d6 c4d3 h7f6 f4g3 e7f8 e1e8 a8e8 a2a3 d7e6 h3f4 e6d7 f4e2 d6d5 e2d4 a7a5');
INSERT INTO evaluations VALUES(34301,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(34301,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(34301,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(34301,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(34301,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(34301,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(34301,36,'cp 181','b1c3 d5a5 e2e4 g8f6 f1d3 g7g6 g1f3 c8g4 e1g1 f8g7 h2h3 g4f3 d1f3 e8g8 f3e2 e7e6 c1e3 b8d7 a1d1 e6e5 d4e5 d7e5 d3c2 b7b5 f2f4 e5c4 c2b3 c4e3 e2e3 b5b4 c3a4 a8d8 e4e5 d8d1 f1d1 f6d5 e3f2');
INSERT INTO evaluations VALUES(34298,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(34298,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(34298,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(34298,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(34298,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(34298,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(34298,36,'cp 174','d5a5 e2e4 g8f6 f1d3 g7g6 g1f3 c8g4 e1g1 f8g7 h2h3 g4f3 d1f3 e8g8 f1d1 e7e5 d4e5 a5e5 c1f4 e5e7 d3c4 f6e8 h3h4 a7a5 e4e5 g7e5 f3e2 e8d6 f4g5 e7e8 g5h6 b8d7 h6f8 g8f8 c4b3 d7c5 h4h5 c5b3 a2b3');
INSERT INTO evaluations VALUES(34322,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(34322,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(34322,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(34322,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(34322,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(34322,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(34322,36,'cp 198','e2e4 g8f6 f1d3 g7g6 g1f3 c8g4 e1g1 f8g7 h2h3 g4f3 d1f3 e7e6 f1d1 e8g8 a2a3 b8d7 c1e3 a5d8 d3b1 d8e7 b1a2 h7h6 f3e2 a8d8 g2g3 g8h7 a1c1 d7b6 h3h4 d8d7 d1d3 f8d8 c1d1 h7g8 g1g2 g7f8 b2b4 f8g7 e3c1 e6e5 d4e5 d7d3 d1d3 d8d3 e2d3');
INSERT INTO evaluations VALUES(39094,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(39094,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(39094,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(39094,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(39094,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(39094,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39094,36,'cp 10','d6c5 f1d3 d7d6 b1c3 b8d7 c3a4 g8e7 g1e2 e8g8 e1g1 f7f5 e4f5 e7f5 d3f5 f8f5 a4c5 d7c5 c1e3 b7b6 d1d2 c5e4 d2b4 e4c5 e2g3 f5f8 b4d2 a7a5 b2b3 c8f5 g3f5 f8f5 a1e1 d8h4 f2f3 h4b4 e1e2 b4d2 e3d2 h7h6 f1e1 a8f8 h2h3');
INSERT INTO evaluations VALUES(42566,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(42566,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(42566,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(42566,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(42566,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(42566,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(42566,36,'cp -125','e8g8 e1g1 b8c6 c2c3 d7d5 c4d5 c8h3 g2h3 f6d5 e4d5 c6e7 c3c4 c7c6 b1c3 c6d5 c3d5 e7d5 c4d5 d8d6 d2d3 a8d8 f1d1 d6d5 f3d5 d8d5 c1e3 c5e3 f2e3 f8d8 a1c1 g7g6 c1c7 d8d7 c7d7 d5d7 d3d4 g8f8 g1f2 f8e7 h3h4 d7d5 d1d3 e5d4 e3d4 d5h5 f2g3 e7e6 d3a3 a7a5');
INSERT INTO evaluations VALUES(37240,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(37240,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(37240,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(37240,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(37240,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(37240,35,'cp -128',NULL);
INSERT INTO evaluations VALUES(37240,36,'cp -116','e1g1 b8c6 c2c3 d7d5 c4d5 c8h3 g2h3 f6d5 e4d5 c6e7 c3c4 c7c6 b1c3 c6d5 c3d5 e7d5 c4d5 d8d6 d2d3 a8d8 c1e3 c5e3 f2e3 d6d5 f3d5 d8d5 f1d1 g7g6 a1c1 f8d8 c1c7 d8d7 c7d7 d5d7 d3d4 g8f8 g1f2 f8e7 d1d3 e5d4 e3d4 e7d6 d3b3 b7b6 h3h4 d7e7');
INSERT INTO evaluations VALUES(49699,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(49699,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49699,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(49699,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(49699,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(49699,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(49699,36,'cp 0','g1f3 b8c6 c2c4 d7d5 c4d5 f6d5 a2a3 d8f6 c1d2 c8e6 b1c3 f6g6 f3h4 g6h6 h4f3');
INSERT INTO evaluations VALUES(52115,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52115,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(52115,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52115,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52115,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(52115,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(52115,36,'cp 12','b1c3 g8f6 g1f3 b8c6 f1e2 d7d5 d2d4 c5d4 e3d4 e5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 e2c4 f6d5 d1e3 f8b4 c1d2 d5e3 f2e3 b4d2 e1d2 e8e7 d2c3 c8e6 c4e6 e7e6 a1f1 h8d8 f1f4 d8d5 f4e4 e6d7 e4g4 d5c5 g4c4 c5c4 c3c4 a8b8 b2b4 h7h5 h1b1 b8b5');
INSERT INTO evaluations VALUES(52109,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(52109,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(52109,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(52109,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52109,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(52109,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52109,36,'cp 16','g8f6 f1e2 d7d5 c4d5 f6d5 g1f3 b8c6 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e2c4 b7b6 c4d5 c8b7 e3e4 c6a5 b2b3 f8e7 c1e3 b7d5 d1d5 e8f7 g1f1 h8d8 f1e2 a8c8 d5d8 c8d8 g2g4 h7h5 g4g5 a5b7 g5f6 g7f6 f3h4 b7d6 f2f3 c5c4');
INSERT INTO evaluations VALUES(44323,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44323,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44323,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(44323,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(44323,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(44323,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(44323,36,'cp 12','g1f3 b8c6 f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e2c4 b7b6 c4b5 c8d7 e3e4 e8c8 c1e3 c6a5 b5d7 d8d7 b2b3 f8e7 g1f1 h8d8 d1d7 c8d7 a1d1 d7e6 d1d8 e7d8 f3d2 g7g6 f1e2 f6f5 f2f3 f5f4 e3f2 d8e7 e2d3 b6b5');
INSERT INTO evaluations VALUES(10413,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(10413,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10413,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10413,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10413,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10413,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(10413,36,'cp 0','e8g8 c3d5 e5e4 f3h4 d7d6 d5b4 c6b4 a2a3 b4c6 d2d3 d6d5 c4d5 d8d5 c1d2 d5b5 d2c3 e4d3 d1d3 b5d3 e2d3 c6e7 c3f6 g7f6 e1d2 f8d8 a1c1 c7c6 c1c5 c8e6 g2e4 d8d7 h1e1 a8d8 e1e3 e7d5 e4d5 e6d5 h4f5 d5e6 f5h4');
INSERT INTO evaluations VALUES(4443,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4443,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4443,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4443,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(4443,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4443,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4443,36,'cp 11','c3d5 e5e4 f3h4 d7d6 d5b4 c6b4 a2a3 b4c6 d2d3 d6d5 e1g1 e4d3 e2d3 f8e8 h4f3 d5c4 d3c4 d8d1 f1d1 c6a5 f3d4 a5c4 b2b3 c4e5 c1f4 c7c5 d4b5 c8g4 d1d2 e8e7 f4e5 e7e5 b5d6 a8d8 d6f7 d8d2 f7e5 g4e6 g2b7 e6b3 a1c1 d2c2 c1c2 b3c2 e5c6');
INSERT INTO evaluations VALUES(13552,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13552,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13552,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13552,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13552,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13552,36,'cp 0','d5c4 d3c4 a7a6 e1g1 b7b5 c4d3 c5d4 e3d4 c6b4 d3e2 f8e7 a2a3 b4c6 c1e3 e8g8 f3e5 c8b7 e2f3 d8c7 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(13501,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13501,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13501,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13501,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13501,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13501,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13501,36,'cp 0','d3c4 a7a6 e1g1 b7b5 c4d3 c5d4 e3d4 c6b4 d3e2 f8e7 a2a3 b4c6 c1e3 c8b7 f3e5 e8g8 e2f3 d8c7 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(13497,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13497,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13497,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13497,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13497,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13497,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13497,36,'cp 0','a7a6 e1g1 b7b5 c4d3 c5d4 e3d4 c6b4 d3e2 f8e7 a2a3 b4c6 c1e3 c8b7 f3e5 e8g8 e2f3 d8c7 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(26403,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26403,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(26403,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26403,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(26403,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(26403,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(26403,36,'cp 20','e1g1 c5d4 c3d4 g8e7 b1c3 e7f5 a2a3 f8e7 c1d2 e8g8 c3e2 f7f6 e5f6 e7f6 d2c3 d8d7 a1e1 a8e8 h2h3 f5d6 e2c1 d6e4 e1e2 d7f7 c3b4 f6e7 b4e7 e8e7 f3d2 e6e5 d2e4 d5e4 e2e4 c6d4 d3e3 h7h6 e4e5 e7e5 e3e5');
INSERT INTO evaluations VALUES(26404,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(26404,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(26404,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(26404,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(26404,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(26404,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(26404,36,'cp 20','c5d4 c3d4 g8e7 b1c3 e7f5 c3e2 c6b4 d3b3 a7a5 a2a4 f8e7 c1d2 e8g8 g2g4 f5h4 f3h4 e7h4 d2b4 a5b4 f2f4 a8c8 b3d3 d8b6 b2b3 c8c6 f1c1 f8c8 c1c6 c8c6 a1f1 b6a6 d3a6 b7a6 f1c1 c6c1 e2c1 h4e7 g1g2 a6a5');
INSERT INTO evaluations VALUES(26435,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26435,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26435,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(26435,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(26435,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(26435,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(26435,36,'cp 33','c3d4 g8e7 b1c3 e7f5 c3e2 f8e7 g2g4 f5h4 f3h4 e7h4 f2f4 d8b6 c1e3 e8g8 f4f5 a8c8 a2a3 c6a5 b2b4 a5c4 e3f2 h4e7 a1a2 a7a5 b4a5 c4a5 h2h4 b6a6 d3a6 b7a6 e2f4 g7g5 f4h5 g5h4 f2e3 g8h8 e3h6 f8g8 a2g2 a5c4 f5f6 e7f8 h6g7 f8g7 f6g7 g8g7 h5g7 h8g7 a3a4');
INSERT INTO evaluations VALUES(26433,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26433,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26433,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26433,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26433,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(26433,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26433,36,'cp 31','g8e7 b1c3 e7f5 c3e2 f8e7 g2g4 f5h4 f3h4 e7h4 f2f4 d8b6 c1e3 e8g8 f4f5 a8c8 a2a3 c6a5 b2b4 a5c4 e3f2 h4e7 a1a2 a7a5 b4a5 c4a5 h2h4 c8c4 g4g5 e6f5 d3f5 b6b3 a2d2 a5c6 d2d3 b3c2 f5f3 b7b5 e2c3 b5b4 c3d5 b4a3 d5e7 c6e7 d3a3 c2g6 a3a8 f8a8 f3a8 c4c8 a8a4');
INSERT INTO evaluations VALUES(24588,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24588,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24588,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(24588,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(24588,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24588,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24588,36,'cp 8','b1c3 e7f5 a2a3 f8e7 c1d2 e8g8 c3e2 f7f6 e5f6 e7f6 d2c3 d8d7 h2h3 f6d8 a1e1 d8c7 e2g3 c7g3 f2g3 f5g3 f3g5 f8f1 e1f1 g3e4 g5e4 d5e4 d3e4 c6e7 f1e1 e7d5 e4e6 d7e6 e1e6 a8c8 e6d6 d5c3 b2c3 c8c3 d6d7 c3a3 d7b7 a3a2 b7c7 a7a5 d4d5');
INSERT INTO evaluations VALUES(30826,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(30826,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(30826,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(30826,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(30826,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(30826,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(30826,36,'cp 37','d2d4 e7e6 f1d3 d8f6 f3g3 f6d4 g1f3 d4d6 c1f4 d6e7 f4c7 b8a6 b1c3 a6c7 g3c7 b7c6 f3d4 e7c5 c3b5 a7a6 d4b3 c5e7 b5d4 e7d8 c7d8 e8d8 d4c6 d7c6 e4e5 d8c7 g2g3 a6a5 e1c1 a5a4 b3d4 g8e7 d3e4 a8d8 f2f4 e7d5 d4f3 f8c5 f3g5');
INSERT INTO evaluations VALUES(38426,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38426,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(38426,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(38426,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38426,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(38426,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(38426,36,'cp 30','g8f6 b1c3 e8g8 d1d2 c8g4 e1c1 b8c6 f1e2 g4f3 e2f3 e7e5 d4e5 d6e5 f4e3 d8c8 c1b1 f8d8 c3d5 f6d5 e4d5 c6d4 f3e4 f7f5 e4d3 c7c5 c2c3 c5c4 c3d4 c4d3 d4e5 f5f4 e3d4 c8c4 d2d3 c4d3 d1d3 d8d5 h1d1 g7e5 d4e5 d5e5 d3d7 e5e2 d7b7');
INSERT INTO evaluations VALUES(41395,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(41395,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(41395,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(41395,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41395,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(41395,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(41395,36,'cp 27','b1c3 e8g8 d1d2 b8c6 e1c1 c8g4 f4e3 e7e5 d4e5 d6e5 d2d8 a8d8 d1d8 f8d8 f3d2 g4e6 f2f3 a7a6 f1c4 e6c4 d2c4 b7b5 c4d2 c6d4 d2b3 d4b3 a2b3 g7f8 e3g5 f8e7 h2h4 g8g7 g5f6 e7f6 c3d5 d8d7 c2c4 f6e7 d5e7 d7e7 c1c2 h7h5 h1a1 e7e6 c4c5 e6c6 b3b4');
INSERT INTO evaluations VALUES(39711,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(39711,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(39711,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(39711,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(39711,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(39711,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(39711,36,'cp -46','f6e4 g5e7 e4c3 e7d8 c3d1 d8c7 d1b2 c7d6 b2d3 c2d3 c8d7 a2a4 b8c6 d4b3 c6e7 f1c1 e7f5 d6b4 d7c6 b3a5 a8b8 a5c6 b7c6 c1c6 e8d7 c6c4 h8c8 c4c8 d7c8 b4c5 b8b3 g2g4 f5h4 g1f1 b3d3 a1c1 c8d7 c1b1 h4f3 f1g2 f3g5 c5e3');
INSERT INTO evaluations VALUES(48186,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(48186,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(48186,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(48186,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(48186,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(48186,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(48186,36,'cp 140','f1b5 c8d7 b5d7 d8d7 d5e6 d7e6 g1e2 f8d6 e1g1 e8g8 b1d2 c7c5 d2f3 f8e8 f1e1 b8c6 c1e3 f6g4 d4d5 g4e3 d5e6 e3d1 e6f7 g8f7 a1d1 a8d8 e2c3 e8e1 f3e1 f7e6 g1f1 d6e7 d1d8 e7d8 c3d1 c6b4 a2a3 b4c6 f1e2 c5c4 e1f3 b7b5 c2c3 h7h6 d1e3 d8b6 h2h3 g7g5 e3g4');
INSERT INTO evaluations VALUES(51328,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(51328,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(51328,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(51328,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(51328,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(51328,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51328,36,'cp 0','a7a6 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 b1c3 e7e6 c3b5 a8e4 f1c4 c7c6 b5c3 e4f3 d1f3 d8d4 f3f4 d4f4 c1f4 b8d7 e1g1 f8b4 f1a1 e8e7 a1a6 b4d6 f4d6 e7d6 c3e4 d6e7 a6a8 d7f8 a8a7 f8d7');
INSERT INTO evaluations VALUES(51355,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(51355,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(51355,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(51355,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(51355,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(51355,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(51355,36,'cp 112','c2c4 g8f6 b1c3 c8b7 f1d3 b8d7 e1g1 g7g6 c1g5 h7h6 g5e3 f8g7 h2h3 e8g8 d1d2 g8h7 a2a3 e6e5 d4d5 a7a5 f1e1 f6h5 d2d1 h7g8 g2g3 b7c8 d3c2 d7c5 g1g2 d8d7 e1h1 d7e7 b2b4 a5b4 a3b4 a8a1');
INSERT INTO evaluations VALUES(51350,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(51350,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(51350,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(51350,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(51350,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(51350,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(51350,36,'cp 108','c8b7 f1d3 b8d7 b1c3 g7g6 e1g1 g8f6 h2h3 f8g7 c1g5 h7h6 g5e3 e8g8 d1d2 g8h7 a2a3 f6h5 g2g4 h5f6 a1e1 e6e5 d4d5 a7a5 b2b4 d8e7 g1h2 a5b4 a3b4 c7c5 b4b5 a8a3 e1a1 f8a8 a1a3 a8a3 d2b2 a3a8 f1e1 b7c8 e1a1');
INSERT INTO evaluations VALUES(57392,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(57392,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(57392,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(57392,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(57392,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(57392,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(57392,36,'cp 64','b8c6 f1b5 f8c5 b5c6 d7c6 b1c3 h7h5 e5f6 c5g1 f6f7 e8f8 d1f3 c8g4 f3f4 g1d4 h2h3 g8h6 h3g4 g7g5 f4f3 h5g4 f3g3 d8d6 g3d6 c7d6 d2d3 h6f7 h1h8 d4h8 c1e3 f8g7 e1d2 g7g6 a1h1 a7a6 c3e2 a8e8 e3d4 h8d4 e2d4 d6d5 e4d5 c6d5 c2c3 f7d6 h1f1 e8h8');
INSERT INTO evaluations VALUES(4970,30,'cp 336',NULL);
INSERT INTO evaluations VALUES(4970,31,'cp 323',NULL);
INSERT INTO evaluations VALUES(4970,32,'cp 341',NULL);
INSERT INTO evaluations VALUES(4970,33,'cp 343',NULL);
INSERT INTO evaluations VALUES(4970,34,'cp 320',NULL);
INSERT INTO evaluations VALUES(4970,35,'cp 339',NULL);
INSERT INTO evaluations VALUES(4970,36,'cp 322','c4f7 e8f7 d1h5 f7f8 h5c5 d8e7 c5c4 e7f7 c4a4 g8f6 b1c3 f8g8 c1e3 c8b7 a4b4 b7c8 b4c5 c8b7 e1c1 d7d6 c5b4 b7c8 b4a4 c8b7 e4e5 d6e5 c3e4 f6e4 a4e4 a8e8 b2b3 b7c8 c1b2 c6c5 e3c5 c8f5 e4d5 f7d5 d1d5 f5e4 d5d7 e4g2 h1g1');
INSERT INTO evaluations VALUES(15816,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(15816,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(15816,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(15816,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(15816,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(15816,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(15816,36,'cp 110','d2d3 f6d5 g5f3 f7f6 e1g1 c7c6 d3d4 e5d4 f1e1 f8e7 d1e2 e8f7 c3d4 e7d6 b1c3 h8e8 e2e8 d8e8 e1e8 f7e8 c3d5 c6d5 c4d5 f5e7 d5e4 a8b8 f3d2 d6c7 g2g3 c8f5 e4g2 b8d8 d2b3 b7b6 c1e3 e8f8 b3d2 e7d5 a1c1 f5g6 d2c4 g6f7');
INSERT INTO evaluations VALUES(19358,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(19358,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(19358,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(19358,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(19358,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(19358,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(19358,36,'cp 21','e7e6 c1e3 g8f6 c2c3 f8e7 g2g3 b7b6 c5b6 f6g4 d1c1 a7b6 f1g2 e8g8 e1g1 c8a6 h2h3 g4e3 c1e3 d7d5 a2a4 d8c7 b1a3 c6a5 f3d2 h7h6 a3b5 a6b5 a4b5 e7g5 f2f4 g5f6 e3d3 a5b7 e2e4 b7c5 d3e3 d5e4 d2e4 c5e4 g2e4 a8a1 f1a1 f6e7');
INSERT INTO evaluations VALUES(18702,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(18702,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18702,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(18702,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(18702,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(18702,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(18702,36,'cp 29','c1e3 g8f6 c2c3 f8e7 g2g3 f6g4 e3d4 b7b6 c5b6 e6e5 f1h3 e5d4 h3g4 d8b6 c3d4 h7h5 g4h3 b6b2 b1d2 c6d4 e1g1 d4f3 d2f3 a8b8 d1d3 b2a3 d3d5 a3c5 f1d1 c5d5 d1d5 b8b2 a1c1 e8g8 h3d7 c8a6 e2e3 b2a2 d5h5 g7g6 h5d5 a6b7 d7c6 f8c8 f3e5 a2b2 d5d7 e7f6 e5g4 c8c6 c1c6 b7c6');
INSERT INTO evaluations VALUES(19941,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(19941,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(19941,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(19941,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(19941,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(19941,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(19941,36,'cp -13','e7e6 g1f3 g8f6 d3d4 c5d4 f3d4 f8e7 e1g1 e8g8 h2h3 c6d4 d1d4 b7b6 c4d3 c8b7 c1g5 a7a6 g5h4 d8c7 a1d1 h7h6 f1e1 f8d8 c3e2 b6b5 f2f3 e6e5 d4c3 c7b8 e2g3 d6d5 g3f5 e7f8 g1h1 d8e8 c3d2 d5e4 f3e4');
INSERT INTO evaluations VALUES(20668,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(20668,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(20668,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(20668,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(20668,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20668,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20668,36,'cp -5','g1f3 f8e7 d3d4 c5d4 f3d4 g8f6 e1g1 e8g8 a2a4 d6d5 e4d5 e6d5 c4e2 c8e6 e2f3 a8c8 c1e3 a7a6 c3e2 c6e5 c2c3 e7c5 e2c1 e6g4 f3g4 f6g4 e3f4 f8e8 h2h3 g4f6 f4e5 e8e5 c1d3 c5d4 c3d4 e5e7 f1e1 e7e1 d1e1 d8b6 e1e5 h7h6');
INSERT INTO evaluations VALUES(36498,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36498,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36498,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36498,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36498,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36498,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36498,36,'cp 0','h2h3 c7c5 e1g1 a7a6 a2a4 b7b6 c1g5 h7h6 g5h4 e7e6 f3h2 b8c6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(36496,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36496,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36496,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36496,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36496,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36496,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36496,36,'cp 0','c7c5 e1g1 a7a6 a2a4 b7b6 c1g5 h7h6 g5h4 e7e6 f3h2 b8c6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(35930,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35930,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35930,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35930,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35930,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35930,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35930,36,'cp 0','e1g1 a7a6 a2a4 b7b6 c1g5 h7h6 g5h4 e7e6 f3h2 b8c6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(35929,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35929,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35929,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35929,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35929,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35929,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35929,36,'cp 0','a7a6 a2a4 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(45694,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45694,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45694,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(45694,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(45694,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(45694,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(45694,36,'cp 29','e2e4 b7b5 e4e5 h7h6 g5h4 g7g5 f3g5 h6g5 h4g5 b8d7 e5f6 c8b7 h2h4 d7f6 d1f3 f8e7 f1e2 d8d4 a1d1 d4e5 g5f4 e5c5 h4h5 a8d8 d1d8 e8d8 h5h6 b5b4 c3e4 c5f5 e4f6 e7f6 e2c4 f5b1 f3d1 b1d1 e1d1 f6b2 f4e3 c6c5 d1c2 b2d4 e3d4 c5d4 f2f3 d8e7 c2b3 a7a5 b3a4 h8a8 h6h7 b7c6 c4b5');
INSERT INTO evaluations VALUES(45695,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(45695,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(45695,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(45695,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45695,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(45695,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(45695,36,'cp 46','b7b5 e4e5 h7h6 g5h4 g7g5 f3g5 h6g5 h4g5 b8d7 e5f6 c8b7 g2g3 d8b6 f1g2 e8c8 e1g1 c6c5 d4d5 b5b4 a1b1 b6a6 d5e6 b7g2 e6e7 g2f1 g1f1 f8e7 f6e7 d8g8 c3e4 c4c3 f1g1 g8g5 e4g5 a6g6 b2c3 g6g5 d1e2 g5e5 e2a6 c8b8 a6b5 b8c7 c3b4 e5e7 b5a5 c7b7 b4c5 b7a8 c5c6 d7e5 a5d5 e7e6 d5e6 f7e6 c6c7');
INSERT INTO evaluations VALUES(43965,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(43965,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(43965,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(43965,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(43965,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(43965,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(43965,36,'cp 47','e4e5 h7h6 g5h4 g7g5 f3g5 h6g5 h4g5 b8d7 e5f6 c8b7 g2g3 c6c5 d4d5 d7f6 f1g2 f8h6 h2h4 e6d5 d1f3 h6g5 h4g5 h8h1 g2h1 f6e4 c3e4 d8e7 f3g4 d5e4 g5g6 e7e6 g6f7 e6f7 e1c1 a8d8 h1e4 d8d1 c1d1 b7e4 g4e4 e8f8 d1e2 b5b4 e4e5 f8g8 e5c5 f7e6 c5e3');
INSERT INTO evaluations VALUES(43963,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43963,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(43963,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(43963,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(43963,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(43963,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(43963,36,'cp 40','h7h6 g5h4 g7g5 f3g5 h6g5 h4g5 b8d7 e5f6 c8b7 g2g3 c6c5 d4d5 d8b6 f1g2 e8c8 e1g1 b5b4 c3a4 b6a6 a2a3 b7d5 g2d5 d7e5 a4c5 f8c5 a3b4 c5f2 g1f2 a6b6 f2g2 d8d5 d1e2 b6c6 g2g1 e5d3 a1a5 d3b4 a5d5 c6d5 g5e3 b4d3 e3a7 d5h5 e2h5 h8h5 g1g2 h5d5 a7e3');
INSERT INTO evaluations VALUES(43956,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(43956,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(43956,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(43956,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(43956,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(43956,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(43956,36,'cp 59','g5h4 g7g5 f3g5 h6g5 h4g5 b8d7 e5f6 c8b7 g2g3 c6c5 d4d5 f8h6 g5h6 h8h6 d1d2 d8f6 c3e4 f6f3 e4d6 e8e7 d6b7 h6h5 d5d6 e7e8 h1g1 c4c3 b2c3 h5e5 f1e2 e5e2 d2e2 f3c3 e1f1 c3a1 f1g2 a1e5 e2e5 d7e5 b7c5 a8c8 c5e4 f7f5 e4f6 e8f7 f6h7 f7g6 h7g5 g6g5 f2f4 g5f6 f4e5 f6e5 g1d1 c8d8 d6d7 e5f6 h2h4 b5b4 g2f3 f6e7');
INSERT INTO evaluations VALUES(43957,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(43957,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(43957,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(43957,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(43957,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(43957,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43957,36,'cp 37','g7g5 f3g5 h6g5 h4g5 b8d7 e5f6 c8b7 g2g3 d8b6 f1g2 e8c8 e1g1 c6c5 d4d5 b5b4 a1b1 b6a6 d5e6 b7g2 e6e7 g2f1 g1f1 f8e7 f6e7 d8g8 c3e4 c4c3 f1g1 g8g5 e4g5 a6g6 b2c3 g6g5 d1e2 g5e5 e2a6 c8b8 a6b5 b8c7 c3b4 e5e4 b5a5 c7b7 b1d1 e4e7 a5b5 b7c7 b5a4 a7a6 a4a6 h8h5 h2h4 e7e6 a6e6 f7e6 b4c5 h5c5 g1g2 c5c2');
INSERT INTO evaluations VALUES(43952,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43952,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(43952,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(43952,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(43952,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(43952,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(43952,36,'cp 62','f3g5 h6g5 h4g5 b8d7 e5f6 c8b7 g2g3 d8b6 f1g2 e8c8 e1g1 c6c5 d4d5 b5b4 a1b1 b6a6 d5e6 b7g2 e6e7 g2f1 g1f1 f8e7 f6e7 d8g8 c3e4 g8g5 e4g5 h8e8 d1d5 e8e7 g5e4 e7e5 d5d6 e5e6 d6d7 c8d7 e4c5 d7e8 c5a6 e6a6 b1a1 e8f8 f1e2 a6e6 e2d2 e6f6 a1f1 f8g8 a2a3 b4a3 b2a3 f6d6 d2c3 d6d3 c3b4 a7a5 b4a5 d3a3 a5b5');
INSERT INTO evaluations VALUES(43951,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43951,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43951,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43951,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(43951,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(43951,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(43951,36,'cp 37','h6g5 h4g5 b8d7 e5f6 c8b7 g2g3 c6c5 d4d5 d7f6 f1g2 f8h6 g5f6 d8f6 e1g1 a8d8 d1e2 e8g8 a1d1 b5b4 c3e4 f6f5 d5d6 b7d5 f2f4 h6g7 g3g4 f5h7 e4g5 h7g6 g2e4 d5e4 g5e4 g7b2 g1h1 b2d4 f4f5 e6f5 g4f5 g6h6 e2c4 f8e8 d1d4 c5d4 c4d4');
INSERT INTO evaluations VALUES(43955,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(43955,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(43955,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(43955,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(43955,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(43955,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(43955,36,'cp 71','h4g5 b8d7 e5f6 c8b7 g2g3 d8b6 f1g2 e8c8 e1g1 d7e5 d4e5 d8d1 a1d1 b6c7 g5f4 f8h6 f4h6 h8h6 f2f4 c7b6 g1h1 c6c5 c3e4 b7e4 g2e4 h6h8 f4f5 e6f5 e4f5 c8b8 d1d7 b6c6 h1g1 c4c3 b2c3 c5c4 f1f2 c6c5 e5e6 f7e6 f5e6 h8h2 d7d4 h2f2 g1f2 b5b4 f6f7 b4c3 f2f3');
INSERT INTO evaluations VALUES(43954,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(43954,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(43954,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(43954,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(43954,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(43954,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(43954,36,'cp 61','b8d7 e5f6 c8b7 g2g3 c6c5 d4d5 f8h6 g5h6 h8h6 d1d2 h6f6 e1c1 d7b6 f1g2 b7d5 f2f4 b5b4 c3d5 e6d5 g2d5 d8d5 d2d5 b6d5 d1d5 a8c8 h1e1 f6e6 e1e6 f7e6 d5e5 e8e7 e5h5 a7a5 a2a4 e7f7 h5g5 c8h8 h2h4 h8g8 g5g8 f7g8 c1d2 g8h7 g3g4 h7g8 h4h5 g8h8 d2d1 h8g7 d1c1 g7h7 c1c2 h7g8 g4g5 g8g7 c2d1 g7h8 d1d2');
INSERT INTO evaluations VALUES(11956,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(11956,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(11956,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(11956,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(11956,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(11956,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(11956,36,'cp 53','e5f6 c8b7 g2g3 c6c5 d4d5 f8h6 g5h6 h8h6 d1d2 h6f6 e1c1 d7b6 f1g2 b7d5 f2f4 b5b4 c3d5 e6d5 g2d5 d8d5 d2d5 b6d5 d1d5 a8c8 h1e1 f6e6 e1e6 f7e6 d5e5 e8e7 e5h5 a7a5 a2a4 e7f7 h2h3 f7e7 h5g5 e7f7 h3h4 c8g8 g5g8 f7g8 c1d2 g8h7 g3g4 h7g7 h4h5 g7h8 g4g5 h8h7 d2c2 h7g7 c2c1 g7g8 c1b1 g8h8 h5h6');
INSERT INTO evaluations VALUES(54714,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(54714,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(54714,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(54714,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(54714,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(54714,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(54714,36,'cp 111','c2c4 e7e6 f1d3 f6f5 g1h3 c7c6 e1g1 g8h6 b1c3 f8e7 f2f3 e8g8 d1c2 b8d7 g1h1 d7f6 c1d2 h6f7 a1e1 b7b6 h3f4 g8h8 c4d5 c6d5 e3e4 d5e4 f3e4 e6e5 d4e5 f7e5 e4f5 e5d3 c2d3 d8d3 f4d3 c8a6 e1e7 a6d3 f1f3 a8d8 d2e1 d3a6 f3f2');
INSERT INTO evaluations VALUES(54700,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(54700,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(54700,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(54700,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(54700,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(54700,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(54700,36,'cp 90','e7e6 b1c3 f6f5 g1h3 c7c6 f2f3 g8f6 f1d3 f8e7 e1g1 e8g8 a2a3 b7b6 b2b4 b8a6 c4d5 f6d5 c3d5 c6d5 c1d2 a6c7 d1e2 c8d7 d2e1 a7a6 e1g3 d7b5 a1c1 e7d6 f1e1 h7h6 h3f4');
INSERT INTO evaluations VALUES(256,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(256,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(256,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(256,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(256,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(256,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(256,36,'cp 171','c3d5 f6d8 e1g1 g8e7 c2c3 e8g8 d2d4 e5d4 c3d4 c5d6 d5e3 d6f4 d4d5 c6b8 f1e1 d7d6 h2h3 c7c6 b5a4 c6d5 e3d5 f4c1 a1c1 b8c6 a4c6 e7c6 d1b3 a8b8 c1c6 b7c6 b3b8 c6d5 e4d5 d8f6 b8b3 c8f5 b3a3 f5g6 a3a7 f6b2 f3d4 f8d8');
INSERT INTO evaluations VALUES(263,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(263,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(263,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(263,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(263,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(263,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(263,36,'cp 187','f6d8 e1g1 g8e7 c2c3 e8g8 d2d4 e5d4 c3d4 c5d6 c1g5 f7f6 b5c4 g8h8 g5d2 a8b8 d5e7 d8e7 f1e1 b7b6 a2a3 e7e8 d2c3 a7a5 c4f1 e8h5 g2g3 b6b5 b2b3 b5b4 a3b4 c6b4 f1c4 c8b7 d4d5 b7a6 a1a5 a6c4 b3c4 b4c2 d1c2');
INSERT INTO evaluations VALUES(570,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(570,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(570,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(570,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(570,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(570,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(570,36,'cp 0','d7d5 e1g1 e8g8 d4c6 f6c6 e3c5 c6c5 b1d2 c8e6 d1a4 c5b6 e4d5 e7d5 d2c4 b6c5 c4d2 c5b6');
INSERT INTO evaluations VALUES(484,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(484,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(484,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(484,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(484,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(484,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(484,36,'cp 0','e1g1 e8g8 d4c6 f6c6 e3c5 c6c5 b1d2 c8e6 d1a4 c5b6 e4d5 e7d5 d2c4 b6c5 c4a5 c5b6');
INSERT INTO evaluations VALUES(8986,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8986,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8986,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8986,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(8986,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8986,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8986,36,'cp 0','b4c3 b2c3 e8g8 a3a4 c6a5 c1a3 f8e8 f3e5 f6d7 e5d7 c8d7 c4d5 e6d5 f1d3 h7h6 d1h5 d8g5 h5f3 d7g4 f3g3 g4d7');
INSERT INTO evaluations VALUES(9000,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9000,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9000,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(9000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9000,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9000,36,'cp 0','b2c3 e8g8 c4d5 e6d5 c3c4 c6a5 c4d5 d8d5 a1b1 c7c5 d4c5 d5d1 e1d1 f6e4 d1e1 e4c5 f3d4 f8d8 f2f3 c5e6 d4f5 e6c5');
INSERT INTO evaluations VALUES(8997,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8997,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8997,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8997,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8997,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8997,36,'cp 0','e8g8 c4d5 e6d5 c3c4 c6a5 c4d5 d8d5 a1b1 c7c5 d4c5 d5d1 e1d1 f6e4 d1e1 e4c5 f3d4 f8d8 f2f3 c5e6 d4f5 e6c5');
INSERT INTO evaluations VALUES(3684,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3684,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3684,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3684,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3684,36,'cp 0','c4d5 e6d5 c3c4 c6a5 c4d5 d8d5 a1b1 c7c5 d4c5 d5d1 e1d1 f6e4 d1e1 e4c5 f3d4 f8d8 f2f3 c5e6 d4f5 e6c5');
INSERT INTO evaluations VALUES(3670,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3670,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3670,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3670,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3670,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3670,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3670,36,'cp 0','e6d5 c3c4 c6a5 c4d5 d8d5 a1b1 c7c5 d4c5 d5d1 e1d1 f6e4 d1e1 e4c5 f3d4 f8d8 f2f3 c5e6 d4f5 e6c5');
INSERT INTO evaluations VALUES(13165,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(13165,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13165,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13165,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(13165,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(13165,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(13165,36,'cp 21','h2h3 c6b4 e1g1 b4d3 d1d3 e7e6 c1g5 f8e7 f1e1 e8g8 f3e5 c8d7 b1d2 h7h6 g5h4 f6e8 h4e7 d8e7 b2b3 e8d6 c2c4 d5c4 b3c4 f8d8 d4d5 a8c8 a1c1 b7b6 e5d7 e7d7 d5e6 f7e6 d3e2 d8e8 d2f3 d7c7 f3e5 d6f7 e5f7');
INSERT INTO evaluations VALUES(13163,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13163,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13163,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13163,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(13163,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(13163,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(13163,36,'cp 21','c6b4 e1g1 b4d3 d1d3 e7e6 c1g5 f8e7 f3e5 e8g8 f1e1 c8d7 b1d2 h7h6 g5h4 f6e8 h4e7 d8e7 b2b3 e8f6 c2c4 d5c4 b3c4 f8d8 a2a4 d7c6 a4a5 a8c8 d2f3 e7c7 e1c1 c6e4 d3a3 e4f3 e5f3 b7b6 a1b1 c7d6 b1b3 d6a3 b3a3 f6e4 a5b6 a7b6');
INSERT INTO evaluations VALUES(19112,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(19112,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(19112,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(19112,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(19112,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(19112,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(19112,36,'cp 141','b1c3 d8f6 d4d3 f8b4 c1d2 g8e7 e1c1 e8g8 a2a3 b4d6 h2h4 f8e8 g2g3 d6c7 d3e2 a7a5 d2g5 f6g6 e2d2 h7h6 g5f4 d7d6 f4d6 c7d6 d2d6 c8g4 d1d3 g6d6 d3d6 g4f3 h1g1 e7g6 f1g2 f3g2 g1g2 b7b5 h4h5 g6e5 f2f4 b5b4 c3b1 e5c4');
INSERT INTO evaluations VALUES(19111,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(19111,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(19111,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(19111,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(19111,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(19111,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(19111,36,'cp 125','d8f6 d4d3 f8b4 c1d2 g8e7 a2a3 b4c5 d2e3 f6d6 d3e2 c5e3 e2e3 e8g8 f1d3 e7g6 e1c1 f8e8 h2h4 d6f4 e3f4 g6f4 d3f1 f4e6 g2g3 e6c5 f1g2 a7a5 h1e1 b7b5 e4e5 a8b8 c3e4 c5e4 g2e4 b5b4 f2f4 b4a3 b2a3 g7g6 d1d6 g8f8 g3g4');
INSERT INTO evaluations VALUES(20074,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(20074,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(20074,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20074,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20074,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20074,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20074,36,'cp 0','g8f6 f1d3 g7g6 e1g1 f8g7 h2h3 e8g8 d1e2 d8c7 f1d1 c8d7 d4f3 a7a6 b1a3 f8d8 a3c4 b7b5 c4e3 d7e6 f3g5 e6d7');
INSERT INTO evaluations VALUES(13371,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13371,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13371,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13371,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13371,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13371,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13371,36,'cp 0','f1d3 g7g6 e1g1 f8g7 h2h3 e8g8 d1e2 c8d7 f1d1 d8c7 d4f3 f8d8 b1a3 a8c8 a3b5 c7b8 b5a3');
INSERT INTO evaluations VALUES(13372,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13372,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13372,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13372,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13372,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13372,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13372,36,'cp 0','g7g6 h2h3 f8g7 e1g1 e8g8 d1e2 d8c7 f1d1 c8d7 b1a3 f8d8 a3b5 c6d4 b5d4 d7a4 d1e1 a7a6 c3c4 f6d7 c1e3 c7b6 d4b5 d7c5 b5d4');
INSERT INTO evaluations VALUES(23387,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(23387,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(23387,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(23387,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(23387,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(23387,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(23387,36,'cp 38','d7d6 e5d6 e7e5 d2d4 f8d6 d4e5 c6e5 f3e5 d6e5 f1d3 g8f6 e1g1 e8g8 b1d2 f6e4 d2e4 f5e4 d3c4 g8h8 d1d8 f8d8 c1g5 e5f6 g5f6 g7f6 a1e1 c8f5 f2f3 e4f3 f1f3 f5c2 f3f6 d8d7 f6e6 a8d8 b2b4 c2f5 e6e5 f5g6 e5e7 g6f5 c4f7 f5g6 f7g6 h7g6');
INSERT INTO evaluations VALUES(44065,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(44065,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(44065,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(44065,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(44065,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(44065,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(44065,36,'cp 109','d8e7 d2d4 d7d6 e5f3 f6e4 f1e2 e4c3 b2c3 e7d8 e1g1 f8e7 f1e1 e8g8 e2d3 c8e6 a1b1 b7b5 e1e6 f7e6 d1e2 d8d7 e2e4 f8f5 g2g4 d6d5 e4e3 f5f8 f3e5 d7e8 e3h3 g7g6 c1h6 b8d7 h6f8 d7f8 f2f4 e7d6 b1e1 d6e5 e1e5 f8d7 e5g5');
INSERT INTO evaluations VALUES(18808,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(18808,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(18808,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(18808,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(18808,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(18808,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(18808,36,'cp 0','a7a6 g2g3 d7d5 e4d5 e6d5 d1e2 c8e6 f1g2 d5d4 e1g1 f8e7 f3e5 c6e5 e2e5 a8b8 c2c3 b7b6 c3d4 e7f6 e5e2 g8e7 e2a6 e8g8 b1a3 e7f5 a3c2 f6d4 c2d4 f5d4 a6d3 e6f5 d3c4 f5e6');
INSERT INTO evaluations VALUES(17386,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(17386,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(17386,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17386,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17386,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(17386,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(17386,36,'cp 13','g2g3 d7d5 e4d5 e6d5 d1e2 c8e6 d2d3 h7h6 f1h3 d8e7 h3e6 e7e6 e2e6 f7e6 b1d2 g8e7 e1g1 e7f5 a1e1 e8d7 f3e5 c6e5 b2e5 h8g8 f2f4 f8e7 d2f3 a8f8 g3g4 f5h4 f3h4 e7h4 e1e2 f8f7 c2c4 g8f8 c4d5 e6d5 f1c1 f8c8 g1g2 h4f6 g2g3 c8c6 f4f5');
INSERT INTO evaluations VALUES(21915,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21915,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21915,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(21915,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(21915,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(21915,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(21915,36,'cp 11','c8g4 d2d3 g7g6 f1e2 g8f6 c1g5 g4f3 e2f3 h7h5 c3e2 f8e7 g5e3 a7a5 g2g3 f6g4 e3d2 g4f6 h2h4 f6d7 d2e3 d7c5 f3g2 c6b4 e2c1 e8g8 e1g1 e7f6 a1b1 d8e7 d1d2 f6g7 g2h3 b4c6 c1e2 a5a4 e3g5 g7f6');
INSERT INTO evaluations VALUES(26990,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(26990,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(26990,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(26990,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(26990,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(26990,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(26990,36,'cp 14','d2d3 g7g6 f1e2 g8f6 c1g5 g4f3 e2f3 c6d4 g2g3 h7h5 h2h4 c7c6 f3g2 f8h6 g5h6 h8h6 b2b4 e8f8 a1b1 h6h8 e1g1 f8g7 c3e2 d8b6 g2h3 h8e8 a2a4 a7a6 a4a5 b6a7 e2d4 a7d4 d1e2 a8b8 f1c1 b7b6 e2d2 b6a5 b4a5 b8b1 c1b1');
INSERT INTO evaluations VALUES(22517,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(22517,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22517,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22517,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(22517,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(22517,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(22517,36,'cp 16','g8f6 d2d3 f8c5 b1c3 d7d6 c3a4 c5b6 a4b6 a7b6 a2a4 d8e7 e1g1 c8e6 c4b5 e8g8 b2b3 c6a7 b5c4 a7c6 c4e6 f7e6 c2c3 e7e8 f1e1 e8g6 f3h2 b6b5 b3b4 d6d5 a4a5 f8f7 c1e3 d5e4 d3e4 g6e4 e3d2');
INSERT INTO evaluations VALUES(16143,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16143,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(16143,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(16143,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(16143,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(16143,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(16143,36,'cp 10','d2d3 f8c5 b1c3 a7a6 c1e3 d7d6 d3d4 e5d4 f3d4 c5d4 e3d4 e8g8 e1g1 c6d4 d1d4 b7b5 c4d5 a8b8 a2a3 f8e8 f1e1 c7c5 d4d3 b5b4 a3b4 c5b4 c3e2 f6d5 e4d5 b8b5 d3c4 e8e5 e2d4 b5d5 e1e5 d6e5 d4c6 d8d6 c6b4 d5d1 a1d1');
INSERT INTO evaluations VALUES(29220,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(29220,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(29220,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29220,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(29220,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(29220,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(29220,36,'cp 0','e6d5 f3h4 f5c8 f1d3 f8d6 h4f5 c8f5 d3f5 e8g8 e1g1 f8e8 g1h1 b7b5 a2a3 b8d7 f2f3 g7g6 f5d3 a7a5 d1c2 a8c8 e3e4 d5e4 c3e4 d6f8 c2f2 f6d5 c1g5 f8e7 g5h6 e7f8');
INSERT INTO evaluations VALUES(29327,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29327,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29327,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29327,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(29327,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(29327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29327,36,'cp 0','f3h4 f5e6 f1d3 f8d6 e1g1 e8g8 h4f5 e6f5 d3f5 f8e8 d1c2 b8d7 f2f3 c6c5 g1h1 a7a6 d4c5 d7c5 f1d1 g7g6 f5h3 c5e6 g2g3 a8c8 c2e2 c8c5 e2d3 d6e5 c1d2 c5c6 a1c1 e6c5 d3c2 c5e6');
INSERT INTO evaluations VALUES(34788,30,'cp 242',NULL);
INSERT INTO evaluations VALUES(34788,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(34788,32,'cp 238',NULL);
INSERT INTO evaluations VALUES(34788,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(34788,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(34788,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(34788,36,'cp 240','b1c3 f6d8 c1e3 g8f6 h2h3 c7c6 f2f4 d8b6 a1b1 c6c5 d4e2 b6a5 e3d2 b8c6 g2g4 f8e7 f1g2 h7h6 e1g1 a5b6 d2c1 b6c7 e2g3 c8d7 c3b5 c7b6 c2c4 e8c8 g3f5');
INSERT INTO evaluations VALUES(34787,30,'cp 248',NULL);
INSERT INTO evaluations VALUES(34787,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(34787,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(34787,33,'cp 260',NULL);
INSERT INTO evaluations VALUES(34787,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(34787,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(34787,36,'cp 255','f6d8 c1e3 f8e7 f2f4 g8f6 d1f3 e8g8 h2h3 c7c6 e1c1 d6d5 e4e5 f6e8 g2g4 c6c5 d4f5 d5d4 e3f2 b8c6 f5e7 d8e7 c3d5 e7d8 f1d3 f7f5 d5e3 g8h8 e3f5 c8e6 d3e4 d8a5 a2a3 a5c7 h1e1 g7g6 f5d6 e8d6 e5d6 c7d6 e4c6 b7c6');
INSERT INTO evaluations VALUES(35974,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(35974,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(35974,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(35974,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(35974,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(35974,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(35974,36,'cp 35','c8g4 e1g1 e7e6 d3e2 e6d5 e4d5 b8d7 c1d2 d7b6 d1b3 g4f5 a1e1 f8e8 h2h3 f6e4 c3e4 f5e4 f3g5 h7h6 g5e4 e8e4 g2g3 d8f8 e2d3 e4e1 f1e1 a8e8 e1e8 f8e8 b3a3 e8e5 d2e3 a7a5 a3a5 e5b2 g1g2 g7d4 e3d4 b2d4 a5b6 d4d3');
INSERT INTO evaluations VALUES(27797,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(27797,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27797,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(27797,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(27797,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(27797,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(27797,36,'cp 64','e1g1 e7e6 h2h3 g4f3 d1f3 e6d5 c4d5 a7a6 c1f4 f8e8 a2a4 d8c7 a4a5 b8d7 f3e2 f6h5 f4h2 a8c8 d3c2 c5c4 g1h1 d7e5 e2d2 b7b6 a5b6 c7b6 a1b1 b6b4 f2f3 a6a5 f1e1 e8e7 f3f4 e5d3 c2d3 c4d3 d2d3 g7c3');
INSERT INTO evaluations VALUES(37184,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37184,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(37184,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(37184,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37184,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37184,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37184,36,'cp -4','b4c3 d2c3 f6e4 a1c1 d7d6 g1f3 b8d7 d1c2 e4c3 c2c3 b7b6 g2g3 a7a5 f1g2 c8b7 e1g1 a5a4 c1d1 f7f5 f3e1 b7g2 e1g2 d8e7 c3c2 f8e8 e2e4 f5e4 c2e4 e6e5 g2e3 e5d4 e4d5 e7f7 d5f7 g8f7 d1d4 d7c5 d4f4 f7g8 f1d1 e8f8 e3f5');
INSERT INTO evaluations VALUES(37201,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(37201,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(37201,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(37201,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(37201,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(37201,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(37201,36,'cp -2','d2c3 f6e4 a1c1 d7d6 g1f3 b8d7 d1c2 e4c3 c2c3 b7b6 g2g3 c8b7 f1g2 a7a5 e1g1 a5a4 f1d1 d8e7 f3e1 b7g2 g1g2 e6e5 d4e5 d7e5 e1d3 e5d3 e2d3 f8e8 d3d4 h7h6 d1e1 e7d7 c3d3 d6d5 c4c5 e8e1 c1e1');
INSERT INTO evaluations VALUES(37194,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(37194,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37194,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(37194,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37194,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37194,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37194,36,'cp 0','f6e4 d1c2 e4c3 c2c3 b7b6 e2e3 c8b7 g1f3 d7d6 f1e2 b8d7 e1g1 c7c5 f1d1 d8e7 d4c5 d6c5 d1d3 d7f6 a1d1 f6e4 c3c2 e4f6');
INSERT INTO evaluations VALUES(37160,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37160,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37160,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37160,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37160,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(37160,36,'cp 0','d1c2 e4c3 c2c3 b7b6 e2e3 c8b7 g1f3 d7d6 f1e2 b8d7 e1g1 c7c5 f1d1 d8e7 d4c5 d6c5 d1d3 d7f6 a1d1 f6e4 c3c2 e4f6');
INSERT INTO evaluations VALUES(43512,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43512,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(43512,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(43512,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43512,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43512,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43512,36,'cp 5','f8e7 f2f3 b7b5 g2g4 e8g8 h2h4 c8b7 a2a3 b8c6 e1c1 d8c7 g5e3 f8b8 c1b1 c6d4 e3d4 e6e5 d4f2 d6d5 g4g5 d5d4 c3a2 f6h5 c2c3 d4c3 d2c3 c7c3 a2c3 b5b4 c3a4 b4a3 a4b6 b7c6 b1a1 a3b2 a1b1 e7a3 b6a8 b8a8');
INSERT INTO evaluations VALUES(39708,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39708,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(39708,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39708,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39708,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39708,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39708,36,'cp 0','f2f3 b7b5 a2a3 c8b7 g5e3 b8c6 e1c1 e8g8 h2h4 c6d4 e3d4 d6d5 g2g4 b5b4 a3b4 e7b4 d2g5 a8c8 e4d5 b4c3 b2c3 h7h6 d4f6 h6g5 f6d8 f8d8 c3c4 e6d5 h4g5 d5c4 d1d8 c8d8 f1c4 b7f3 h1f1 d8d4 c4b3 f3d5 b3d5 d4d5');
INSERT INTO evaluations VALUES(39701,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39701,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39701,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39701,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39701,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39701,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39701,36,'cp 0','b7b5 e1c1 h7h6 g5e3 b5b4 c3a4 d8a5 b2b3 c8d7 d4f5 e6f5 e3b6 a5e5 b6d4 e5a5');
INSERT INTO evaluations VALUES(44489,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(44489,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(44489,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(44489,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(44489,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(44489,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(44489,36,'cp 108','e4e5 f6d5 c2c4 d5c7 d2d4 d7d6 b1c3 b8d7 e5d6 e7d6 f1d3 f8e7 e1g1 e8g8 c1f4 d7f6 h2h3 f8e8 f1e1 h7h6 f4g3 f6h5 g3h2 e7f8 e1e8 d8e8 c4c5 e8d8 c3e4 d6c5 d4c5 c7d5 d3c2 c8f5 e4d6 f8d6 c5d6 f5c2 d1c2');
INSERT INTO evaluations VALUES(54072,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(54072,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(54072,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54072,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(54072,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(54072,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54072,36,'cp -17','f8c5 a2a3 g8f6 f1e2 e8g8 b2b4 c5d6 c1b2 a7a5 b4b5 b8d7 c2c4 a5a4 b1d2 d8e7 d1c2 d7c5 e1g1 c8d7 b2e5 d6e5 f3e5 f8d8 f1d1 d5c4 e5d7 d8d7 c2c4 a8c8 a1c1 g7g6 e2f3 f6d5 h2h3 c8c7');
INSERT INTO evaluations VALUES(55245,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(55245,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(55245,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(55245,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(55245,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(55245,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(55245,36,'cp -119','g8f6 g1e2 c8f5 h2h3 b8c6 g2g4 f5g6 b1c3 e7e6 d1d2 f8b4 a2a3 b4a5 e1c1 d8e7 e2f4 e8c8 f1c4 a7a6 c1b1 e7d6 c4b3 f6d5 b3d5 e6d5 b1a1 a5c3 d2c3 f7f6 h3h4 g6f7 f4e2 h7h6 g4g5 h6g5 h4g5 c8b8 h1h8 d8h8 g5f6');
INSERT INTO evaluations VALUES(47381,30,'cp -116',NULL);
INSERT INTO evaluations VALUES(47381,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(47381,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(47381,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(47381,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(47381,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(47381,36,'cp -96','g1e2 c8f5 h2h3 b8c6 d1d2 e7e6 a2a3 h7h6 g2g4 f5g6 f1g2 f8d6 c2c4 e8g8 e1g1 f8e8 b2b4 e6e5 d4d5 c6d4 e2d4 e5d4 e3d4 h6h5 d2g5 h5g4 h3g4 d6e7 b1c3 c7c6 g5f4 e7d6 f4g5 d8c7 d5c6 d6h2 g1h1 b7c6 g5h4');
INSERT INTO evaluations VALUES(4039,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(4039,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(4039,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(4039,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(4039,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(4039,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(4039,36,'cp 60','b1c3 e7e5 e2e4 e5d4 f3d4 d7c5 h2h3 f8e8 f1e1 c8d7 b2b3 a7a5 a1b1 h7h5 c1g5 d8c8 h3h4 c7c6 f2f3 c8c7 d1d2 a5a4 b3b4 c5e6 g5e3 e6d4 e3d4 d7e6 g2f1 f6d7 f3f4 g7d4 d2d4 c7b6 d4b6 d7b6 c4c5 d6c5 b4c5 b6d7 b1b7 d7c5 b7c7 c5d7');
INSERT INTO evaluations VALUES(4318,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(4318,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(4318,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(4318,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(4318,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4318,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(4318,36,'cp -10','c6d4 f3d4 c5d4 d2d3 d7d6 h2h3 c8e6 c4b3 c7c6 c3e2 d4b6 e2g3 a7a5 b3e6 f7e6 c1e3 b6c7 g3h5 h7h6 h5f6 d8f6 e1f1 f8f7 d1g4 d6d5 e3c5 a8d8 a2a4 c7d6 c5d6 d8d6 a1e1 d6d8 g4e2 f6h4 e2d2');
INSERT INTO evaluations VALUES(20994,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(20994,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(20994,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(20994,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(20994,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(20994,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(20994,36,'cp 15','c8f5 a2a3 e7e6 c2c4 c6b8 b1c3 c7c6 g1f3 b8d7 e2e3 f8e7 f1d3 g8f6 d3f5 e6f5 e1g1 e8g8 c4c5 f6e4 b3b4 f8e8 c3e2 g7g6 e2f4 a7a5 d1c2 e7f6 f4d3 d8c7 f3e5 f6g5 b2c3 a5b4 a3b4 a8a1 f1a1');
INSERT INTO evaluations VALUES(26553,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(26553,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26553,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26553,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(26553,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(26553,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26553,36,'cp 19','a2a3 c6b8 c2c4 c7c6 g1f3 e7e6 b1c3 b8d7 e2e3 g8f6 f1d3 f5d3 d1d3 d8b6 b3b4 d5c4 d3c4 f8e7 c3a4 b6b5 c4b5 c6b5 a4c5 b7b6 c5d7 f6d7 e1e2 e8g8 a1c1 f8c8 c1c8 a8c8 h1c1 c8c1 b2c1 a7a6 e3e4 h7h6');
INSERT INTO evaluations VALUES(37922,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(37922,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(37922,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(37922,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(37922,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(37922,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(37922,36,'cp 66','c4d5 c6d5 c1f4 g8f6 e2e3 e8g8 h2h3 b8c6 f1e2 c8f5 e1g1 f6e4 a1c1 a8c8 d1b3 d8d7 b3a3 e4c3 c1c3 f7f6 f4g3 a7a6 f1d1 f8d8 d1c1 e7e5 a3b3 e5e4 f3d2 g7f8 a2a3 d8e8 b3b6 f5e6 d2b3 e6f7');
INSERT INTO evaluations VALUES(37916,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(37916,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(37916,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(37916,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(37916,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(37916,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(37916,36,'cp 65','c6d5 c1f4 g8f6 h2h3 e8g8 e2e3 d8b6 d1b3 c8d7 b3b6 a7b6 f1e2 f8c8 f3e5 d7c6 a1c1 b8d7 e5d7 f6d7 e1g1 d7f6 f2f3 e7e6 a2a3 g7f8 c3b5 f6e8 g1f2 f8e7 g2g4 g6g5 f4h2 h7h6 e2d3 c6b5 d3b5 e8d6 b5d7 c8c1 f1c1 d6c4 c1c2 g8g7 d7b5 e7d6');
INSERT INTO evaluations VALUES(38013,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(38013,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(38013,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(38013,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(38013,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(38013,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(38013,36,'cp 62','c1f4 g8f6 h2h3 d8b6 d1b3 c8d7 b3b6 a7b6 e2e3 e8g8 f1e2 f8c8 f3e5 d7c6 a1c1 b8d7 e5d7 f6d7 e1g1 d7f6 a2a3 e7e6 g2g4 g7f8 g4g5 f6e4 h3h4 e4d6 f4d6 f8d6 f2f4 h7h5 g1f2 d6e7 c3b5 g8g7 f1h1 a8a4');
INSERT INTO evaluations VALUES(46190,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(46190,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(46190,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(46190,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(46190,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(46190,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(46190,36,'cp 52','c3b5 b8a6 c2c3 f8e7 a2a4 c8d7 g1f3 e8g8 f1d3 d7b5 a4b5 a6b8 h2h4 b8d7 b2b3 d8b6 e1g1 f8c8 c3c4 a7a6 b5a6 a8a6 a1a6 b7a6 f3e5 c8d8 e5d7 d8d7 f4e5 a6a5 d1c2 h7h6 d3e2 d7d8 d4c5 e7c5 f1d1 d8c8 e2f3 c5e7 e5d4 b6c7');
INSERT INTO evaluations VALUES(46087,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(46087,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(46087,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(46087,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(46087,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(46087,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(46087,36,'cp 44','b8a6 a2a4 f8e7 c2c3 c8d7 g1f3 e8g8 f1d3 d7b5 a4b5 a6b8 h2h4 b8d7 b2b3 d8b6 e1g1 e7d6 f4d6 b6d6 a1a4 f8e8 g2g3 g7g6 c3c4 d5c4 b3c4 a7a6 b5a6 a8a6 a4a6 d6a6 d1b3 a6c6 d3e2 e8a8 b3b2 f6e4 f1b1 c5d4 e3d4');
INSERT INTO evaluations VALUES(14027,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(14027,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(14027,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14027,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14027,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(14027,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(14027,36,'cp 57','c2c3 c8d7 a2a4 f8e7 g1f3 e8g8 f1d3 c5c4 d3b1 d7b5 a4b5 a6b8 h2h4 b8d7 g2g4 a7a6 b5a6 f6g4 b1h7 g8h7 f3g5 h7g8 d1g4 a8a6 a1a6 b7a6 e1f1 d7f6 g4f3 d8b6 f3e2 a6a5 h4h5 a5a4 h5h6 a4a3 h6g7 g8g7 h1g1 a3b2 g5e6 g7h8');
INSERT INTO evaluations VALUES(47961,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(47961,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47961,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47961,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47961,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47961,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47961,36,'cp 0','d6d5 g2g3 f8c5 f1g2 e8g8 e1g1 f8e8 a2a3 h7h6 b2b4 c5f8 a1b1 c7c5 f1e1 e8e1 d1e1 b8c6 b4c5 f8c5 c3a4 c5f8 a4c3');
INSERT INTO evaluations VALUES(54686,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(54686,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(54686,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(54686,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(54686,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(54686,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(54686,36,'cp 77','b1c3 g8f6 g1f3 g7g6 b2b4 a7a5 b4b5 b8d7 e2e3 e7e5 f1e2 e5e4 f3d2 d7b6 e1g1 c8e6 d1c2 f8e7 a2a4 e8g8 c1a3 d6d5 c4d5 e6d5 a3e7 d8e7 c3d5 b6d5 d2c4 f5f4 c2d2 f6d7 e3f4 d5f4');
INSERT INTO evaluations VALUES(54684,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(54684,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(54684,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(54684,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(54684,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(54684,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(54684,36,'cp 50','g8f6 g1f3 g7g6 g2g3 f8g7 f1g2 e8g8 e1g1 c7c6 a1b1 a7a5 a2a3 f6e4 f3g5 d6d5 c1e3 e4g5 e3g5 d5c4 d4d5 h7h6 g5e3 g8h7 c3a4 a8a6 a4c5 a6a8 b1c1 c6d5 g2d5 b8c6 d5c6 d8d1 f1d1 b7c6 c5d7 c8d7 d1d7 g7b2 d7e7 h7g8 c1c4 b2a3 e7c7');
INSERT INTO evaluations VALUES(46574,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(46574,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(46574,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(46574,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(46574,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(46574,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(46574,36,'cp 50','g2g3 e7e5 d4e5 d6e5 d1d8 e8d8 g1f3 e5e4 f3d4 b8c6 d4c6 b7c6 f2f3 e4f3 e2f3 f8d6 f1d3 h8e8 e1f2 d6c5 f2g2 c5d6 c1d2 d8e7 h1e1 e7f7 e1e8 f6e8 d2e3 c8e6 c3e2 c6c5 e2f4 g7g6 f4e6 f7e6 a1e1 e6f7 b2b4 d6e7 e3c5 e7c5 b4c5 e8g7 g2f2');
INSERT INTO evaluations VALUES(57368,30,'cp 221',NULL);
INSERT INTO evaluations VALUES(57368,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(57368,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(57368,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(57368,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(57368,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(57368,36,'cp 200','e4f5 e6f5 d2d4 g8f6 f1d3 d8e7 d3e2 e7d8 e1g1 f8e7 c2c4 d7d6 e2d3 c7c5 f1e1 e8g8 b1c3 b8c6 d3c2 c6b4 h2h3 f8e8 c2b1 e7f8 e1e8 d8e8 c3b5 e8f7 c1f4 f7c4 b5d6 c4d5 d6c8 a8c8 a2a3 b4c6 b1a2 c5c4 d1c2 c6d4 f3d4 d5d4 c2f5');
INSERT INTO evaluations VALUES(57339,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(57339,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(57339,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(57339,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(57339,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(57339,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(57339,36,'cp 209','e6f5 d2d4 g8f6 f1d3 d7d6 e1g1 f8e7 f1e1 e8g8 c2c4 f6e4 b1d2 e4g5 b2b4 a7a6 c1b2 g5f3 d2f3 e7f6 d1d2 b8c6 h2h4 c6e7 h4h5 h7h6 d4d5 f6b2 d2b2 f5f4 b2e2 e7f5 e2e4 d8f6 e4f4 c8d7 a1b1 a8e8 g2g4 e8e1 b1e1 f5h4 f4f6 h4f3 f6f3 f8f3');
INSERT INTO evaluations VALUES(57525,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(57525,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(57525,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(57525,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(57525,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(57525,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(57525,36,'cp 205','d2d4 g8f6 f1d3 f6e4 e1g1 f8e7 f1e1 e8g8 c2c4 d7d6 b1d2 e4g5 b2b4 a7a6 c1b2 g5f3 d2f3 e7f6 b4b5 a6b5 c4b5 c7c6 a2a4 c6b5 d1b3 g8h8 d3b5 b8a6 b2a3 a6c7 b5c4 f8e8 e1e8 c7e8 a1e1 e8c7 a3b4 h7h6 a4a5 c7a6 b4c3 a6c7');
INSERT INTO evaluations VALUES(1823,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(1823,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(1823,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(1823,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(1823,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(1823,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(1823,36,'cp -62','f1e2 d7c6 e1g1 c8e6 b1d2 e4e5 d2f3 e5d5 f3d4 e8c8 d4e6 d5e6 e2d3 c8b8 f1e1 e6f6 d1g4 f8c5 g4f4 h7h6 g2g3 c5d4 a1b1 a7a6 g1g2 g8e7 f4f6 d4f6 h2h4 e7d5 c1d2 b7b5 a2a4 b8b7 h4h5 h8e8 e1e8 d8e8 a4b5 c6b5');
INSERT INTO evaluations VALUES(7688,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7688,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7688,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7688,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(7688,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(7688,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(7688,36,'cp 9','e4e5 f6g8 e3d4 d8a5 f2f4 a8b8 d1d2 b8b2 c3e4 a5d2 e1d2 b2b4 c2c3 b4b2 d2c1 b2b8 d4a7 b8a8 a7d4 f7f6 h2h3 f6e5 f4e5 g8h6 g2g4 a8a5 e4c5 e8g8 a2a4 h6f7 c5b3 a5a8 f1c4 d7d5 e5d6 e7d6 a4a5 c6c5 d4g7 g8g7 c4d5 a8a7 c1c2 h7h5 b3d2 h5g4 h3g4 c8g4 a5a6 g4f5 c2b3');
INSERT INTO evaluations VALUES(7683,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(7683,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7683,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7683,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7683,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7683,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(7683,36,'cp 0','f6g8 e3d4 d8a5 f2f4 a8b8 d1d2 b8b2 c3e4 a5d2 e1d2 b2b4 c2c3 b4b2 d2c1 b2b8 d4a7 b8a8 a7d4 f7f6 f1c4 f6e5 f4e5 g8h6 h1e1 h6f5 a2a4 f5d4 c3d4 h8f8 a4a5 e8d8 c1c2 d8c7 c2c3 d7d5 e5d6 e7d6 e4g5 c7d8 g5f7 d8d7 f7g5');
INSERT INTO evaluations VALUES(5031,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5031,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5031,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5031,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(5031,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5031,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5031,36,'cp 0','e3d4 d8a5 f2f4 a8b8 d1d2 b8b2 c3d5 a5d2 e1d2 b2b8 d5c7 e8d8 c7a6 c8a6 f1a6 b8b4 h1b1 b4d4 d2e3 c6c5 c2c3 d4a4 b1b8 d8c7 b8c8 c7b6 a6d3 b6b7 c8c5 b7b6 c5c8 b6b7');
INSERT INTO evaluations VALUES(13546,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13546,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13546,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(13546,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13546,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13546,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13546,36,'cp 0','b1c3 c5d4 f3d4 f8b4 d4c6 b7c6 d1a4 a7a5 a4c6 c8d7 c6c7 d8c7 f4c7 f6e4 a2a3 b4c3 b2c3 d5c4 f2f3 e4c3 c7e5 c3a4 e5g7 h8g8 g7f6 g8g6 f6d4 a8b8 f1c4 g6g2 c4f1 g2c2 f1d3 c2g2');
INSERT INTO evaluations VALUES(13544,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(13544,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13544,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(13544,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(13544,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13544,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13544,36,'cp 0','c5d4 f3d4 f8b4 d4c6 b7c6 d1a4 b4c3 b2c3 e8g8 f1d3 d5c4 a4c4 a7a5 c4d4 c6c5 d4d8 f8d8 e1e2 c8b7 f4g5 a8b8 e3e4 b7c6 g5f6 b8b2 e2e3 g7f6 h1b1 d8b8 b1b2 b8b2 a1b1 b2a2 b1b8 g8g7 b8c8 c6a4 d3c4 a2a1 f2f3 a1e1 e3d3 e1d1 d3e3');
INSERT INTO evaluations VALUES(13636,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(13636,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13636,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(13636,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(13636,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(13636,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13636,36,'cp 0','f3d4 f8b4 a1c1 e8g8 d4c6 b7c6 f1d3 c8a6 a2a3 b4a5 d1e2 d5c4 d3c4 a6c4 e2c4 a5c3 c4c3 f6d5 c3d2 d5f4 e3f4 a8b8 d2d8 f8d8 c1c2 d8d3 e1e2 d3b3 h1b1 a7a5 c2c6 b3b2 b1b2 b8b2 e2e1 b2b1 e1e2');
INSERT INTO evaluations VALUES(36385,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36385,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36385,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36385,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(36385,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36385,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(36385,36,'cp 3','e2e3 h7h6 g5h4 c7c6 d1c2 f6e4 h4e7 d8e7 f1d3 c8f5 f3e5 f8e8 e1g1 b8d7 f2f3 e4c3 d3f5 d7e5 d4e5 c3b5 f3f4 f7f6 a2a4 b5c7 f5h7 g8h8 h7g6 e8f8 c2c3 a7a5 e3e4 d5e4 g6e4 c7d5 e4d5 c6d5 a1e1 f6e5 c3e5 e7e5 e1e5 d5d4 e5e7 h8g8 e7b7');
INSERT INTO evaluations VALUES(36384,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(36384,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(36384,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(36384,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36384,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36384,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36384,36,'cp 12','h7h6 g5h4 c7c6 d1c2 f6e4 h4e7 d8e7 f1d3 c8f5 f3e5 f7f6 e5f3 f5h7 e1g1 a7a5 a2a3 b8d7 b2b4 g8h8 f1c1 b7b5 c3e2 f8c8 f3d2 d7b6 d3e4 d5e4 b4a5 a8a5 c2c5 e7c5 c1c5 a5a8 h2h3 b6d5 e2c3 d5c3 c5c3');
INSERT INTO evaluations VALUES(36204,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(36204,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(36204,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(36204,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(36204,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36204,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(36204,36,'cp 24','g5h4 c7c6 f1d3 f6e4 h4e7 d8e7 d1c2 f8e8 e1g1 b8d7 c3e2 e4g5 f3d2 d7f6 a1c1 c8d7 c2c5 e7d8 e2g3 a7a5 a2a3 g7g6 c5c2 d8e7 h2h4 g5e6 d2f3 e6g7 h4h5 f6h5 g3h5 g7h5 f3e5');
INSERT INTO evaluations VALUES(36207,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(36207,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(36207,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(36207,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36207,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36207,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(36207,36,'cp 14','c7c6 f1d3 f6e4 h4e7 d8e7 d1c2 f8e8 e1g1 b8d7 c3e2 e4g5 f3d2 d7f6 a1c1 c8d7 c2c5 e7d8 e2f4 a7a5 f1e1 g5e6 f4e6 e8e6 a2a3 e6e7 d2f3 d7e6 f3e5 f6d7 e5d7 e6d7 h2h3 e7e6 e1e2 d8f6 e3e4 d5e4');
INSERT INTO evaluations VALUES(35729,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(35729,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(35729,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35729,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(35729,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(35729,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35729,36,'cp 24','d1c2 f8e8 f1d3 f6e4 h4e7 d8e7 e1g1 b8d7 a2a3 d7f6 f3e5 f6g4 d3e4 d5e4 e5g4 c8g4 h2h3 g4h5 a1c1 a8d8 c3e2 h5e2 c2e2 d8d5 b2b4 a7a5 c1c5 a5b4 a3b4 e7e6 f1b1 e8c8 e2g4 e6g4 h3g4 c8a8 g1h2 a8a3 b4b5 d5c5 d4c5 a3a2 h2g3 c6b5 b1b5');
INSERT INTO evaluations VALUES(44187,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(44187,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(44187,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(44187,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(44187,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(44187,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(44187,36,'cp -67','g1e2 d7d5 c4b5 c8d7 b5d7 d8d7 e4e5 f6g8 d2d4 c5d4 e2d4 g8e7 f3e2 b8c6 d4c6 b7c6 b1d2 e7f5 c2c3 a7a5 d2f3 d7b7 e1g1 f8c5 a1b1 h7h6 f1d1 c5b6 f3d4 b6d4 c3d4 e8g8 b2b3 b7b4 e2d2 c6c5 d4c5 b4c5');
INSERT INTO evaluations VALUES(51957,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(51957,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(51957,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(51957,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(51957,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(51957,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(51957,36,'cp -22','c5d4 d1d4 b8c6 d4d3 d7d5 b1c3 d8a5 e2e4 d5e4 d3e4 g8f6 e4c4 c8d7 e1c1 a8c8 c4b5 a5b4 f4e3 b4b5 f1b5 f8b4 d1d3 a7a6 b5c6 d7c6 f3e5 c6d5 h1d1 b4c3 d3c3 c8c3 b2c3 e8g8 f2f3 d5a2 c3c4 b7b5 c4b5 a6b5 e5d7 f6d7');
INSERT INTO evaluations VALUES(52072,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(52072,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(52072,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(52072,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(52072,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(52072,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52072,36,'cp -1','d1d4 b8c6 d4d3 d7d5 b1c3 d8a5 e2e4 d5e4 d3e4 g8f6 e4c4 c8d7 f1e2 f8e7 e1g1 a8c8 c4b5 a5b4 b5b4 c6b4 a1c1 e8g8 f3e5 d7e8 f1e1 f6d5 c3d5 b4d5 f4d2 f7f6 e5d3 e8b5 d3f4 b5e2 f4d5 e6d5 e1e2');
INSERT INTO evaluations VALUES(16632,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(16632,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(16632,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(16632,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(16632,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(16632,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(16632,36,'cp -120','f8e7 g5f6 e7f6 b1c3 f6c3 b2c3 b7b6 f1d3 d8e7 g1e2 c8b7 e1g1 e8c8 f1e1 h8e8 a1b1 d7d5 e4d5 d8d5 e2g3 d5e5 d1g4 c8b8 e1e5 e7e5 g4e2 g7g6 e2e5 c6e5 f2f3 e8d8 b1d1 e5d3 c2d3 b7a6 d3d4 a6c4 a2a3 c7c5 g1f2 b8c7 g3e2 d8e8 e2f4');
INSERT INTO evaluations VALUES(10931,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(10931,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(10931,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(10931,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(10931,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(10931,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(10931,36,'cp -106','g5f6 e7f6 b1c3 f6c3 b2c3 d8e7 f1d3 b7b6 g1e2 c8b7 e1g1 e8c8 f1e1 h8e8 d1d2 c6e5 e2d4 g7g5 a1b1 g5g4 d4f5 e7a3 c3c4 c8b8 b1b5 b7c6 b5b1 a3c5 d2e3 d7d6 b1d1 c6d7 f5g7 e8f8 e3c5 d6c5 g7f5 b8b7 f5g7 f8g8 g7f5');
INSERT INTO evaluations VALUES(10857,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(10857,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(10857,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(10857,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(10857,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(10857,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(10857,36,'cp -119','e7f6 b1c3 f6c3 b2c3 d8e7 f1d3 b7b6 g1e2 c8b7 c3c4 c6a5 e1g1 e7c5 d1d2 e8c8 d2c3 h8e8 c3g7 b7e4 e2g3 e4g6 d3g6 f7g6 g7h7 e8e6 a1e1 c5c4 e1e6 c4e6 h2h4 a5c6 h4h5 g6h5 h7h5 d8g8 h5f5 e6c4 f5d3 c4d3 c2d3');
INSERT INTO evaluations VALUES(9992,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(9992,31,'cp -123',NULL);
INSERT INTO evaluations VALUES(9992,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(9992,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(9992,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(9992,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(9992,36,'cp -118','b1c3 f6c3 b2c3 d8e7 f1d3 b7b6 g1f3 c8b7 a2a4 e8c8 e1g1 a7a5 d1d2 h8e8 f3d4 e7f6 d4b5 d7d5 e4d5 d8d5 a1e1 d5e5 f2f4 e5e7 e1e7 f6e7 d2f2 c8b8 f1d1 g7g6 h2h3 h7h5 d3f1 h5h4 f1d3 e7f6 c3c4 f6e7 d3f1 f7f5 c4c5 e7c5 f2c5');
INSERT INTO evaluations VALUES(34695,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(34695,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(34695,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(34695,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(34695,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(34695,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(34695,36,'cp 18','c7c6 b5e2 c8f5 g1f3 b8d7 f3h4 f5g6 e1g1 e7e6 d2d3 g8f6 g2g3 g6h5 e2h5 a5h5 d1h5 f6h5 f2f4 f8c5 g1g2 e8g8 h4f3 h5f6 f1e1 f6d5 c3d5 e6d5 c1e3 f8e8 e3c5 d7c5 b2b4 c5a4 e1e8 a8e8 a1e1 e8e1 f3e1 a4c3 a2a3');
INSERT INTO evaluations VALUES(34334,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(34334,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(34334,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(34334,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(34334,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(34334,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(34334,36,'cp 25','b5e2 c8f5 g1f3 b8d7 f3h4 f5g6 e1g1 e7e6 d2d3 g8f6 g2g3 g6h5 e2h5 a5h5 d1h5 f6h5 f2f4 f8c5 g1g2 e8g8 f1e1 a7a5 a2a4 h5f6 c1d2 f8d8 e1e2 d7b6 h4f3 h7h6 c3e4 c5e7 d2e3 b6d5 e3d4 f6e4 d3e4 d5b4 c2c3');
INSERT INTO evaluations VALUES(39593,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(39593,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(39593,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(39593,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(39593,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(39593,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(39593,36,'cp 50','e2e4 d7d5 e4e5 g8h6 g1f3 e8g8 h2h3 f7f6 e5f6 e7f6 f1e2 f8e8 e1g1 c7c6 f1e1 h6f7 c1f4 b8d7 b1d2 d7f8 a2a4 f8e6 f4h2 g7f8 a4a5 f8d6 d2b3 c8d7 h2d6 f7d6 e2d3 b7b6 d1d2 g8g7 h3h4 c6c5 a5b6 a7b6 a1a8');
INSERT INTO evaluations VALUES(39625,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(39625,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(39625,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(39625,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(39625,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(39625,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(39625,36,'cp 0','f1c4 e7e6 g1e2 c7c5 d2d3 b8c6 b1c3 g8e7 c4b3 d7d6 h2h4 h7h5 c1g5 a7a6 f3f4 g7e5 f4f3');
INSERT INTO evaluations VALUES(43343,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(43343,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(43343,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(43343,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(43343,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(43343,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(43343,36,'cp 76','d1d2 f8g7 f2f3 h7h5 h2h3 b8c6 f1e2 c8d7 e1c1 h5h4 h1e1 b7b5 c1b1 e8g8 c3d5 f6d5 e4d5 c6d4 e3d4 e7e5 d5e6 d7e6 a2a3 f8e8 d4g7 g8g7 e2d3 a8c8 e1e4 c8c5 d2f2 e8h8 e4f4 c5d5 f2a7 d8g5 f4e4 g5f6 a7a6 f6g5 e4e6 f7e6 a6b7 g7h6');
INSERT INTO evaluations VALUES(43344,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(43344,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(43344,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(43344,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(43344,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(43344,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(43344,36,'cp 70','f8g7 f2f3 h7h5 h2h3 b8c6 e1c1 c8d7 d4c6 d7c6 d2f2 f6d7 e3d4 g7d4 f2d4 h8f8 f1c4 d8b6 d4d2 e8c8 c4d5 c8b8 h1e1 d7e5 d5b3 b6c5 f3f4 e5d7 a2a3 c5a5 e1e3 d8c8 b3d5 d7b6 d5c6 c8c6 c3d5 a5d2 d1d2');
INSERT INTO evaluations VALUES(39760,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(39760,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(39760,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(39760,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(39760,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(39760,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(39760,36,'cp 62','f2f3 h7h5 f1e2 c8d7 h2h3 b8c6 e1c1 h5h4 h1e1 b7b5 c3d5 a8c8 d4c6 d7c6 d5f6 g7f6 e3d4 f6d4 d2d4 f7f6 f3f4 d8c7 d1d2 c6d7 e4e5 f6e5 f4e5 d6e5 d4e3 d7f5 e2d3 e8g8 d3f5 f8f5 e3h6 g8f7 h6h4 f7g7 h4e4 c8d8 d2d8 c7d8');
INSERT INTO evaluations VALUES(39763,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(39763,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(39763,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(39763,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(39763,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(39763,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(39763,36,'cp 80','b8c6 e1c1 c6d4 e3d4 c8e6 c3d5 e6d5 e4d5 h7h5 c1b1 d8d7 f1d3 e8g8 h2h3 f6d5 d4g7 g8g7 d3g6 f7g6 d2d5 a8c8 h1e1 c8c7 e1e4 f8f6 a2a3 d7f5 d5d2 e7e5 d2b4 c7d7 b4c4 f5e6 c4e2 e6f7 e2d2 f6f4 d2e3 f4e4');
INSERT INTO evaluations VALUES(7102,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(7102,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(7102,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(7102,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(7102,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(7102,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(7102,36,'cp 107','e1c1 c6d4 e3d4 c8e6 c3d5 e6d5 e4d5 h7h5 c1b1 e8f8 f1d3 d8d7 h1e1 a8c8 e1e2 f6d5 d3g6 d5c3 b2c3 f7g6 d4g7 f8g7 d1e1 c8c7 e2e7 d7e7 e1e7 c7e7 d2d6 e7f7 g2g4 h5g4 f3g4 h8c8 h2h4 c8c6 d6d4 g7h7 b1b2 f7e7 d4d3 b7b5 d3d5 c6f6 d5c5 e7e6 c5c7 h7g8 c7c8 g8f7');
INSERT INTO evaluations VALUES(7101,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(7101,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(7101,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(7101,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(7101,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(7101,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(7101,36,'cp 114','h7h5 c1b1 c8d7 c3d5 a8c8 d4c6 c8c6 f1d3 f6d5 e4d5 c6c8 h1e1 e8g8 h2h3 e7e6 d5e6 d7e6 d3e4 d8f6 e3d4 f6d4 d2d4 g7d4 d1d4 d6d5 e4d5 c8d8 c2c4 e6d5 c4d5 d8d7 b1c2 g8g7 c2b3 f8d8 e1d1 g7f6 b3b4 b7b6 a2a4 f6e5 f3f4 e5d6 d4c4 d6e7 c4c6 d7d5 d1d5 d8d5');
INSERT INTO evaluations VALUES(49550,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(49550,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(49550,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(49550,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(49550,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(49550,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(49550,36,'cp -47','f6e4 d1e2 e4f6 f3e5 f8e7 d2d4 e8g8 e2d1 e7b4 b1c3 c7c5 e5f3 d7d5 f1e2 d5c4 e1g1 b8c6 d4c5 c8e6 f3g5 e6f5 e2c4 h7h6 g5f3 b4c5 c1f4 a8c8 d1d8 f8d8 a1d1 a7a6 d1d8 c6d8 c4b3 f5e6 b3e6 d8e6 f4e3 c5e3 f2e3');
INSERT INTO evaluations VALUES(49898,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(49898,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(49898,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(49898,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(49898,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(49898,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(49898,36,'cp -56','d1e2 e4f6 f3e5 f8e7 d2d4 e8g8 e2d1 e7b4 b1c3 c7c5 e5f3 c5d4 f3d4 b8c6 f1e2 f6e4 d1d3 d7d5 d4c6 b7c6 e1g1 c8f5 c3e4 f5e4 d3g3 d8f6 c4d5 c6d5 c1f4 f8e8 a2a3 b4c5 b2b4 c5d4 a1d1 h7h6 e2f3 a8c8 f3e4 e8e4 f4e3 d4e3');
INSERT INTO evaluations VALUES(50536,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50536,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50536,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(50536,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(50536,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50536,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50536,36,'cp 0','g1f3 b7b5 d2d4 c5d4 f3d4 c8b7 f1d3 b8c6 d4b3 d8h4 g2g3 h4d8 c1e3 g8f6 a2a3 c6a5 b3a5 d8a5 h1f1 f8e7 e3d2 a5c7 e4e5 f6d5 d1e2 e8g8 c3d5 b7d5 f4f5 e6f5 f1f5 d5e6 f5f4 a6a5 e1c1 b5b4 a3b4 a5b4 d2b4 a8a1 c1d2 a1d1 d2d1');
INSERT INTO evaluations VALUES(50535,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(50535,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50535,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50535,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50535,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(50535,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50535,36,'cp 0','b7b5 d2d4 c5d4 f3d4 c8b7 f1d3 b8c6 d4b3 d8h4 g2g3 h4d8 c1e3 g8f6 a2a3 a8c8 d1e2 c6a5 e3d4 f8e7 e1g1 a5c6 d4e3 c6a5');
INSERT INTO evaluations VALUES(51224,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51224,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51224,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51224,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51224,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51224,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(51224,36,'cp 13','d2d4 c5d4 f3d4 c8b7 f1d3 b8c6 d4b3 d8h4 g2g3 h4d8 a2a3 g8f6 c1e3 c6a5 b3a5 d8a5 h1f1 f8e7 e3d2 a5b6 d1e2 e8g8 f4f5 e6f5 f1f5 g7g6 f5f1 d7d5 e4d5 f6d5 c3d5 b7d5 e2e5 b6e6 e5e6 d5e6 e1c1 a8c8 d2b4 f8e8 b4e7 e8e7');
INSERT INTO evaluations VALUES(53399,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(53399,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53399,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(53399,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53399,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(53399,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53399,36,'cp 9','e3d4 b8c6 c2c3 c8g4 c1f4 e7e6 f1d3 f8d6 f4d6 d8d6 e1g1 g8e7 b1d2 a8c8 f1e1 h7h6 d2f1 e8g8 h2h4 d6f4 d1d2 f4d2 f3d2 g4f5 d3e2 a7a5 f1e3 c8c7 g2g3 f5h7 a2a4 f8c8 d2b3');
INSERT INTO evaluations VALUES(6502,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(6502,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(6502,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(6502,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6502,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6502,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(6502,36,'cp -23','d7d6 f1e2 g8f6 e1g1 e8g8 h2h3 f8e8 c2c4 c8d7 b1c3 a7a6 a2a3 c6d4 b2b4 d4f3 e2f3 c5e3 f2e3 c7c6 d1d2 b7b5 f1b1 d7e6 f3e2 d6d5 e4d5 c6d5 c4b5 d5d4 c3e4 a6b5 e4f6 d8f6 e3d4 e6d5 d4e5 f6g6 e2f1 e8e5 d2f2 a8f8 b1c1 g6d6');
INSERT INTO evaluations VALUES(6022,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(6022,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6022,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6022,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6022,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6022,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6022,36,'cp -9','f1e2 g8f6 e1g1 e8g8 h2h3 a7a5 c2c4 c8d7 d1d2 f8e8 b1c3 h7h6 e2d1 f6h7 a1b1 h7f8 d3d4 e5d4 f3d4 c6d4 e3d4 d7c6 f1e1 c5d4 d2d4 d8g5 d4e3 f8e6 e3g5 h6g5 b2b3 e6c5 d1c2 g7g6 g1h2 g8g7 b1d1 a8d8 f2f3 c5e6');
INSERT INTO evaluations VALUES(10318,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10318,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10318,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10318,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10318,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10318,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(10318,36,'cp 0','f6e4 d1e2 e4f6 d2d4 c6d4 e2e5 d4e6 c1g5 c5e7 d5e7 d8e7 e1c1 d7d6 e5e3 e6g5 e3e7 e8e7 f3g5 c8e6 h1e1 a8e8 g5e6 f7e6 c4e6 e7d8 e6b3 h8f8 f2f3 f6d7 e1e8 f8e8 d1d5 e8e1 d5d1');
INSERT INTO evaluations VALUES(10649,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10649,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10649,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10649,36,'cp 0','d1e2 e4f6 d2d4 c6d4 e2e5 d4e6 c1g5 c5e7 d5e7 d8e7 e1c1 d7d6 e5e3 e6g5 e3e7 e8e7 f3g5 c8e6 h1e1 a8e8 g5e6 f7e6 c4e6 e7d8 g2g3 e8e7 e6h3 h8e8 e1e7 e8e7 c2c4 e7e2 d1d2 e2e4 b2b3 e4e1 d2d1 e1e2 d1d2');
INSERT INTO evaluations VALUES(13485,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(13485,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(13485,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(13485,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(13485,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(13485,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(13485,36,'cp -11','d8b6 d1e2 c8d7 h2h3 a7a6 b5a4 f8e7 e1g1 e8g8 a4c2 f6e4 b1d2 f7f5 a1b1 d7e8 f3e5 c6e5 f4e5 e8b5 c2d3 b5d3 e2d3 c5c4 d3e2 e4d2 e2d2 b6c6 b1a1 b7b5 a2a4 b5b4 c3b4 c6b7 d2c2 f8c8 a4a5 e7b4 c2e2 b4e7 f1b1 b7b5 e2c2 b5b4 a1a4 b4b7');
INSERT INTO evaluations VALUES(26239,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(26239,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(26239,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(26239,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(26239,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(26239,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26239,36,'cp 34','f3e5 a8c8 d1e2 g8f6 c1f4 f5d7 e5c6 d7c6 b5d3 e7e6 c2c3 f8d6 f4d6 d8d6 b1d2 c6a4 e1g1 a7a6 d2f3 d6b6 f3e5 e8g8 h2h4 f6e4 f1e1 a4b5 d3b5 b6b5 e2b5 a6b5 e5d3 e4d6 g2g4 h7h6 a2a3 c8c6 g1g2 f8c8 e1e2');
INSERT INTO evaluations VALUES(26250,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(26250,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(26250,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(26250,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(26250,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(26250,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(26250,36,'cp 43','a8c8 d1e2 f5d7 e5c6 d7c6 b5d3 e7e6 b1d2 f8d6 e1g1 g8e7 f1e1 e8g8 d2f3 c6d7 f3e5 d6e5 d4e5 e7g6 b2b3 a7a6 c1a3 f8e8 a3d6 d8b6 g2g3 f7f6 c2c4 f6e5 d6e5 g6e5 e2e5 d5c4 b3c4');
INSERT INTO evaluations VALUES(31894,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(31894,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(31894,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(31894,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(31894,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(31894,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(31894,36,'cp 70','g4f3 g2f3 d5c4 b1c3 e7e6 e2e3 g8f6 f1c4 b8d7 e3e4 a7a6 c4e2 c7c5 d4d5 e6d5 c3d5 f8d6 f3f4 f6e4 e2f3 e8g8 f3e4 f8e8 d5c3 d7f6 f2f3 f6e4 f3e4 d8h4 e1f1 d6f4 d1f3 f4e5 h1g1 h4h2 c3e2 b7b5 c1g5 e5b2 a1b1');
INSERT INTO evaluations VALUES(31902,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(31902,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(31902,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(31902,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(31902,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(31902,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(31902,36,'cp 85','g2f3 d5c4 b1c3 e7e6 e2e4 f8b4 f1c4 b8c6 a2a3 b4a5 c4b5 d8d6 d4d5 a5c3 b2c3 e8c8 c3c4 c6e5 c1e3 a7a6 c4c5 d6e7 d5d6 c7d6 b5e2 d6c5 d1b3 e7c7 a1c1 e5d7 h1g1 g7g6 g1g5 e6e5 b3f7 g8f6 e1f1 c7c6 c1b1 c8b8 f7b3 d8c8 e2c4 c8c7 a3a4 h8d8');
INSERT INTO evaluations VALUES(41666,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(41666,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(41666,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(41666,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(41666,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(41666,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(41666,36,'cp 179','f1d3 b8c6 e1g1 b4e7 a2a3 e8g8 c1e3 g8h8 f1e1 f7f6 d1c2 c8d7 a1d1 d8c8 b2b4 d7e8 b4b5 c6d8 a3a4 e8h5 d3e2 e7b4 c2b3 b4a5 d1c1 d8f7 f3d2 h5e2 c3e2 c8d7 e2f4 f8e8 f4d3 e6e5 d4e5 f7e5 d3c5 d7g4 f2f3');
INSERT INTO evaluations VALUES(41667,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(41667,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(41667,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(41667,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(41667,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(41667,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(41667,36,'cp 208','b8c6 e1g1 e8g8 e4e5 b4e7 c3e4 g8h8 e4g5 g7g6 d3b1 h8g7 d1d3 c6b4 d3e4 c8d7 a2a3 b4d5 h2h4 d8e8 h4h5 h7h6 g5h3 f8h8 h5g6 f7g6 f3h4 e7h4 e4h4 d5e7 h4g4 b6d5 c1d2 d7c6 b1e4 e7f5 f1e1 e8f7 a1c1');
INSERT INTO evaluations VALUES(1968,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(1968,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(1968,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(1968,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(1968,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(1968,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(1968,36,'cp 37','f6d4 e3d4 b7b6 h2h4 h7h5 b1c3 c8b7 f2f3 e8c8 f1c4 f7f6 a1d1 g8e7 e1f2 e7c6 d4e3 f8b4 c3b5 a7a6 b5d4 c6e5 c4b3 a6a5 d4e2 d8f8 e3d4 b4c5 d4c5 b6c5 e2g3 g7g6 g3f1 d7d6 f1e3 c8d8 e3c4 e5c4 b3c4 f6f5 e4f5');
INSERT INTO evaluations VALUES(2009,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(2009,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(2009,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(2009,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(2009,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(2009,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(2009,36,'cp 46','e3d4 b7b6 f2f3 c8b7 b1c3 e8c8 f1c4 f7f6 h2h4 h7h5 a1d1 g8e7 e1f2 e7c6 d4e3 f8b4 c3b5 a7a6 b5d4 c6e5 c4e2 b4e7 d4f5 e7f8 a2a3 a6a5 e3d4 e5c6 d4c3 f8c5 f2g3 g7g6 f5d4 c5d6 g3h3 d6e5 d4c6 b7c6 c3e5 f6e5 e2a6 c8b8 a6c4 d7d6 c4f7 c6e8');
INSERT INTO evaluations VALUES(2007,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(2007,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(2007,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(2007,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(2007,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(2007,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(2007,36,'cp 44','b7b6 f2f3 c8b7 b1c3 e8c8 f1c4 f7f6 h2h4 h7h5 a1d1 g8e7 e1f2 e7c6 d4e3 f8b4 c3d5 b4d6 d5f4 c6e5 c4d5 d8f8 c2c3 g7g6 a2a4 b7a6 e3d4 c7c6 d5b3 e5g4 f3g4 d6f4 g4h5 g6h5 h1h3 c6c5 h3f3 c5d4 f3f4 d4c3 b2c3 a6b7 b3c4');
INSERT INTO evaluations VALUES(8835,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(8835,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(8835,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8835,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(8835,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(8835,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(8835,36,'cp -13','c5e3 f2e3 c6a5 c4b3 e8g8 h2h3 c8d7 d1e1 c7c6 f3h4 a5b3 a2b3 a7a5 f1f3 b7b5 e1g3 f6h7 a1f1 d7e6 h4f5 e6f5 f3f5 a8a7 g1h2 a7d7 h2h1 d8e7 c3e2 d7b7 g3f2 g7g6 f5f3 f8a8 e2g3 h6h5 h1h2 b5b4 g3e2');
INSERT INTO evaluations VALUES(8889,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(8889,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(8889,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(8889,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(8889,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(8889,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(8889,36,'cp -18','f2e3 c6a5 c4b3 e8g8 d1d2 c7c6 h2h3 a5b3 a2b3 a7a5 c3e2 b7b5 e2g3 c6c5 f1f2 c8e6 d3d4 a5a4 b3a4 a8a4 a1f1 a4a2 b2b3 e6b3 d4c5 a2c2 d2b4 c2f2 f1f2 b3e6 b4b5 d6c5 b5c5 f6d7 c5c3 d8b8 f2c2 g8h7 g1h2 f8e8 c3a5 b8b3 c2c3');
INSERT INTO evaluations VALUES(8886,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(8886,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(8886,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8886,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(8886,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(8886,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(8886,36,'cp -14','c6a5 c4b3 e8g8 h2h3 a5b3 a2b3 a7a5 b3b4 c7c6 d1e1 c8e6 b4a5 a8a5 a1a5 d8a5 c3d5 a5e1 d5f6 g7f6 f1e1 f8a8 g1f2 h6h5 f3h4 g8h7 c2c4 h7h6 e1d1 a8a2 d1d2 a2a4 f2f3 a4b4 g2g4 b7b5 c4b5 c6b5 d2g2 d6d5 h4f5 e6f5 g4f5 d5e4 d3e4');
INSERT INTO evaluations VALUES(21753,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(21753,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(21753,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(21753,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(21753,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(21753,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(21753,36,'cp -20','g1f3 g8f6 e1g1 c8e6 b1d2 f8e7 c2c3 h7h6 b2b4 e5e4 f3e1 e8g8 d3e4 d5e4 b4b5 c6e5 e1c2 a7a6 b5a6 a8a6 d2e4 d8c8 c2b4 a6a5 d1c2 f6e4 g2e4 f8d8 a2a4 e7f8 b4d3 e5c4 d3f4 e6f5 e4f5 c8f5 c2f5 a5f5 e2e4 f5c5 f4d5 b7b6');
INSERT INTO evaluations VALUES(26509,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(26509,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(26509,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(26509,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(26509,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(26509,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(26509,36,'cp 45','d8d7 b1d2 a7a6 b5e2 g8f6 c2c4 e7e6 e1g1 f8e7 b2b3 h7h6 c1b2 e8g8 a2a3 a6a5 e2d3 g4f5 d3f5 e6f5 f3e5 d7e6 e5d3 e7d6 c4c5 d6e7 d3f4 e6c8 d1c2 c8d7 f4d3 b7b5 d2f3 f6e4 f3e5 c6e5 d3e5');
INSERT INTO evaluations VALUES(27678,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(27678,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(27678,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(27678,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(27678,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(27678,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(27678,36,'cp 51','b1d2 a7a6 b5e2 e7e6 c2c4 g8f6 a2a3 f8d6 c4c5 d6e7 b2b4 f6e4 c1b2 f7f5 d2b3 e7f6 e1g1 g7g5 f3e1 g4e2 d1e2 g5g4 f2f3 g4f3 g2f3 d7g7 g1h1 e4g5 e1d3 g5h3 e2g2 g7g2 h1g2 h3g5 a3a4 c6e7 g2h1 h8g8 b2c3 e7g6 c3e1');
INSERT INTO evaluations VALUES(29814,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(29814,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(29814,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(29814,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(29814,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(29814,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(29814,36,'cp 69','e7e6 e1g1 b8c6 c2c4 d5b6 b1c3 f8e7 d4d5 e6d5 c4d5 g4f3 e2f3 c6e5 f3e4 e8g8 d1c2 h7h6 b2b3 b6c8 g2g3 c8d6 e4g2 f8e8 c1b2 e7f6 a1d1 a7a5 f1e1 e5d7 e1e8 d8e8 b2c1 e8e7 c3e4 a5a4 b3a4 d6e4 g2e4 d7c5 e4h7 g8h8');
INSERT INTO evaluations VALUES(30048,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(30048,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(30048,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(30048,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(30048,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(30048,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(30048,36,'cp 92','e1g1 f8d6 h2h3 g4f5 c2c4 d5f4 c1f4 d6f4 e2d3 f5d3 d1d3 c7c6 b1c3 b8d7 d4d5 e8g8 d5e6 f7e6 a1d1 d7e5 f3e5 f4e5 d3e2 d8f6 c3e4 f6e7 f1e1 a8d8 d1d8 f8d8 g2g3 e5d4 b2b3 b7b6 h3h4 h7h6 g1g2 c6c5 f2f3 a7a5 e1d1 d8f8 h4h5');
INSERT INTO evaluations VALUES(30047,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(30047,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(30047,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(30047,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(30047,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(30047,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(30047,36,'cp 71','b8c6 c2c4 d5b6 b1c3 f8e7 d4d5 e6d5 c4d5 g4f3 e2f3 c6e5 f3e4 e8g8 d1c2 h7h6 g2g3 f8e8 b2b3 e7f6 c1b2 b6c8 a1d1 c8d6 e4g2 a7a5 f1e1 e5g6 h2h4 e8e1 d1e1 g6e5 c3e4 d6e4 c2e4 e5d7 b2a3 f6e5 d5d6 e5d6 a3d6 c7d6 e4b7 a8b8 b7c6 d7f6 c6c4');
INSERT INTO evaluations VALUES(31905,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(31905,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31905,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(31905,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(31905,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31905,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(31905,36,'cp 25','g2f3 e7e6 e2e4 g8f6 c2c4 d5e4 b1c3 f8d6 f4g5 b8c6 f3e4 e6e5 d4d5 c6d4 d1d3 h7h6 g5d2 f6h5 e1c1 d8h4 c1b1 e8g8 d3h3 h4h3 f1h3 c7c5 c3b5 d4b5 c4b5 h5f4 h3g4 g7g6 h2h4 g8g7 d1g1 b7b6 h4h5 g6g5 a2a4 f4d3 f2f3 d3f4 b1c2 f4e2 g1g2 e2f4 d2f4 e5f4 g4f5 g7f6 h1d1 a8d8');
INSERT INTO evaluations VALUES(31903,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31903,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(31903,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31903,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31903,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31903,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(31903,36,'cp 17','e7e6 e2e4 g8f6 d1d3 c7c5 d3b5 d8d7 d4c5 d7b5 f1b5 b8c6 b1c3 f8c5 e4d5 f6d5 c3d5 e6d5 e1c1 e8c8 f4g3 h8e8 h1e1 e8e1 d1e1 c5d6 f3f4 c8c7 e1d1 a7a6 b5a4 d6c5 c2c3 d5d4 f4f5 c7b6 f5f6 g7f6 c1c2 d4c3 d1d8 c6d8 c2c3');
INSERT INTO evaluations VALUES(55384,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(55384,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(55384,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(55384,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(55384,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(55384,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(55384,36,'cp 108','c7c5 c2c3 b8c6 f1d3 f8g7 e1g1 g8e7 b1a3 e8g8 a3c2 c5d4 c3d4 f7f6 e5f6 f8f6 b2b4 f6f3 d1f3 e6e5 f3d1 e5e4 d3e2 e7f5 c1b2 d8b6 a2a4 a7a5 b4b5 c6d4 e2g4 d4c2 d1c2 g7b2 c2b2 d5d4 a1d1 d4d3 d1c1 b6d6 f1e1');
INSERT INTO evaluations VALUES(53475,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(53475,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(53475,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(53475,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(53475,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(53475,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(53475,36,'cp 113','c2c3 b8c6 f1d3 f8g7 e1g1 g8e7 b1a3 e8g8 a3c2 f7f6 e5f6 f8f6 f1e1 c5d4 c2d4 d8d6 d1e2 c6d4 c3d4 e7c6 c1e3 c8d7 d3b5 a7a6 b5c6 d7c6 a1c1 f6f5 e2d2 a8f8 e3g5 f5f3 g2f3 e6e5 d4e5 g7e5 f3f4 e5g7 c1d1 d5d4');
INSERT INTO evaluations VALUES(11764,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(11764,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(11764,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(11764,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(11764,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(11764,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(11764,36,'cp 88','a4b3 c8b7 d2d3 h7h6 c1e3 d7d6 a2a3 g7g6 d3d4 f8g7 d4e5 d6e5 e3c5 e8g8 d1e2 f8e8 a1d1 c6d4 f3d4 e5d4 e1g1 e7f5 e2g4 e8e5 g4g6 g8h8 g6f7 e5c5 e4f5 d8e8 c3e2 e8f7 b3f7 c5f5 f7e6 f5f6 e6h3 c7c5 b2b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(13662,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13662,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13662,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13662,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13662,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(13662,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(13662,36,'cp 20','f6h5 f4e3 f8d6 e1g1 h5f4 e3f4 d6f4 b1d2 e8g8 g2g3 f4d6 a2a3 c8d7 d1e2 a7a5 f3e5 c6e5 d4e5 d6c5 d2f3 a5a4 a1d1 d7c6 f3d4 d8b6 d1d2 c5d4 c3d4 f7f6 e5f6 f8f6 f1e1 a8f8 e2e3 c6b5 d3b5 b6b5');
INSERT INTO evaluations VALUES(25746,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(25746,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(25746,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(25746,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(25746,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(25746,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(25746,36,'cp 200','f8d6 b1c3 g8e7 h2h3 g4h5 g2g4 h5g6 d1d4 f7f6 c1e3 e8g8 e1c1 a8b8 e4e5 b8b4 d4b4 d6b4 d1d8 f8d8 e5f6 g7f6 e3d2 b4c3 d2c3 e7d5 h1e1 g8f7 c3a5 d8e8 e1e8 f7e8 a5d2 c6c5 f3h4 g6f7 c2c4 d5b4 d2b4 c5b4 b2b3 e8d7 c1d2 d7c6 d2e3 c6c5 e3e4 a6a5 h4f5 f7g6');
INSERT INTO evaluations VALUES(23789,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(23789,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(23789,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(23789,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(23789,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(23789,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(23789,36,'cp 212','h2h3 g4h5 c1e3 g8e7 g2g4 h5g6 d1d4 f7f6 b1c3 a8b8 e1c1 e8g8 e4e5 b8b4 d4a7 f6e5 f3e5 d8b8 e5g6 e7g6 a7b8 f8b8 b2b3 a6a5 a2a3 b4b7 a3a4 b7b4 h1e1 c6c5 c3b5 c5c4 b5d4 c4b3 c2b3 b8b6 d4b5 b6c6 c1b2 d6e5 e3d4 e5d4 d1d4 b4d4 b5d4 c6d6 b2c3 g6f4');
INSERT INTO evaluations VALUES(32306,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(32306,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(32306,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(32306,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(32306,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(32306,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(32306,36,'cp -12','g1f3 g8f6 f1d3 a7a6 d3f5 e6f5 c3e2 f8d6 e1g1 c7c6 b2b3 d8e7 c1b2 e8g8 c2c4 g7g6 e2c1 f8e8 c1d3 a6a5 d1c2 f6e4 a2a3 b8d7 a3a4 h7h5 f3e5 e4g5 a1e1 g5e6 d3f4 a8c8 f4d3 c8d8 h2h3 d7e5 d4e5 d6c5 d3c5 e7c5');
INSERT INTO evaluations VALUES(47759,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(47759,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(47759,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(47759,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(47759,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(47759,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47759,36,'cp 33','e5e4 g2g3 f8e7 f1g2 c8f5 g1h3 e8g8 e1g1 f8e8 f2f3 e4f3 e2f3 h7h6 c1e3 f6d7 h3f2 e7f6 e3d4 b8a6 c3b5 d7e5 a1c1 d8d7 b2b3 c7c6 d5c6 b7c6 f3f4 c6b5 f4e5 d6e5 d4e3 a8d8 f2e4 f5e4 d1d7 d8d7 g2e4 b5c4');
INSERT INTO evaluations VALUES(49410,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(49410,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(49410,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(49410,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(49410,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(49410,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(49410,36,'cp -28','e2e3 f8e7 c2c3 d7d5 g1f3 c7c5 f1d3 e8g8 e1g1 c5c4 d3c2 b7b5 f3e5 a7a5 b1d2 c8b7 d1e2 b8d7 e3e4 d5e4 d2e4 f6e4 c2e4 b7e4 e2e4 b5b4 e4e2 d7e5 f4e5 a5a4 e2g4 g8h8 a2a3 b4a3 b2a3 d8d5 a1b1 a8c8 g4f3 d5f3 f1f3 h8g8 g2g4 f7f6 e5f6 f8f6');
INSERT INTO evaluations VALUES(49408,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(49408,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(49408,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(49408,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(49408,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(49408,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(49408,36,'cp -27','c7c5 c2c3 f8e7 f1d3 d7d5 g1f3 b7b6 d1e2 f6e4 b2b3 c5d4 c3d4 e8g8 e1g1 a7a5 c1b2 c8a6 d3a6 b8a6 b1c3 e4c3 b2c3 a6c7 f1c1 c7e8 c1c2 a5a4 a1c1 e7a3 c3b2 a4b3 a2b3 e8d6 c2c7 a3b2 e2b2');
INSERT INTO evaluations VALUES(44226,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(44226,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(44226,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(44226,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(44226,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(44226,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(44226,36,'cp -24','c2c3 f8e7 g1f3 d7d5 f1d3 e8g8 e1g1 b7b6 d1e2 f6e4 b2b3 b8d7 c1b2 c8b7 b1d2 a7a5 f3e5 e4d2 e2d2 a5a4 a1d1 d7e5 f4e5 a4b3 a2b3 e7g5 d2e2 d8e7 h2h3 a8a2 f1f3 c5c4 d3b1 c4b3 b1a2 b3a2 d1a1 f8a8 e2c2 f7f5 e5f6');
INSERT INTO evaluations VALUES(53924,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(53924,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(53924,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(53924,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(53924,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(53924,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(53924,36,'cp 96','h7h5 b1c3 g7g6 g2g3 h5h4 f1h3 c8h3 g1h3 f8h6 g3g4 h6c1 d1c1 g8f6 f2f3 b8d7 e1g1 a7a6 a2a4 e8f8 g1h1 f8g7 c1d2 d8b6 d2c2 b6d8 f1g1 d7f8 g1g2 a8b8 b2b3 f8h7 c2f2 d8e7 a1e1 b7b6 g2g1 e7d7');
INSERT INTO evaluations VALUES(55568,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(55568,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(55568,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(55568,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55568,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(55568,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(55568,36,'cp 20','e5e4 f3g5 f7f5 g5h3 g8f6 b1c3 f8e7 c1g5 c8e6 e2e3 e6f7 h3f4 b8d7 h2h3 c7c6 g2g4 f5g4 h3g4 d8a5 g5f6 d7f6 d1b3 e8c8 g4g5 a5g5 f1h3 c8b8 e1c1 g5h4 d4d5 c6d5 h3e6 h4f2 c3d5 f6d5 e6d5 f7d5 f4d5');
INSERT INTO evaluations VALUES(56550,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56550,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(56550,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(56550,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56550,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56550,36,'cp 13','g8f6 f1c4 e5d4 e3d4 f8d6 g1f3 e8g8 e1g1 b8c6 c1g5 h7h6 g5h4 c8g4 f1e1 a7a6 c4d5 a8b8 d5c6 b7c6 b2b3 d6b4 e1e3 c6c5 c3e4 g7g5 e4f6 d8f6 h4g3 c5d4 g3e5 g4f3 d1d4 f6b6 d4b6 b8b6 e3f3 c7c5 a1d1 f8e8 e5c3 b4c3');
INSERT INTO evaluations VALUES(48915,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(48915,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48915,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(48915,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48915,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(48915,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(48915,36,'cp 14','f1c4 e5d4 e3d4 f8d6 g1f3 e8g8 e1g1 b8c6 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 a2a4 a7a6 a4a5 g4h5 e1e3 d6f4 h4f6 d8f6 c3d5 f6d6 d5f4 d6f4 c4d5 a8d8 d5c6 b7c6 a1c1 c6c5 c1c5 h5f3 g2f3 f4d4 d1d4 d8d4 c5c7 d4d5 e3e7 d5a5');
INSERT INTO evaluations VALUES(48910,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(48910,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(48910,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(48910,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(48910,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(48910,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(48910,36,'cp 0','e5d4 e3d4 f8d6 g1f3 e8g8 e1g1 b8c6 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 a1c1 a8b8 c3e4 d6e7 h4f6 e7f6 e4f6 d8f6 c4d5 c6e7 h2h3 e7d5 h3g4 c7c6 e1e5 h8g8 d1c2 f8e8 c1e1 e8e5 e1e5 f6d6 g2g3 d5f6 g4g5 h6g5 f3g5');
INSERT INTO evaluations VALUES(5380,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(5380,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(5380,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(5380,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(5380,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(5380,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(5380,36,'cp -24','d7d6 f3g5 c6e5 d1b3 e5c4 b3c4 g8h6 c1e3 c8d7 h2h3 g7c3 c4c3 f7f6 g5f3 h6f7 a2a4 e8g8 a4a5 d7c6 f3d4 d8d7 f2f3 f8c8 c3b4 e7e5 d4e2 d6d5 e4d5 d7d5 e2c3 d5e6 a1d1 b7b6 d1c1 f7d6 a5b6 a7b6 b4b6 d6f5');
INSERT INTO evaluations VALUES(5520,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(5520,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(5520,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(5520,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(5520,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(5520,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(5520,36,'cp -30','f3g5 c6e5 d1b3 e5c4 b3c4 g8h6 c1e3 c8d7 h2h3 g7c3 c4c3 f7f6 g5f3 h6f7 c3b3 b7b6 f1c1 e8g8 a2a4 e7e5 f3d2 f8e8 b3b4 a8b8 a4a5 b6a5 b4a5 d8a5 a1a5 e8c8 c1b1 d7b5 e3a7 b8b7 f2f3 b5d3 b1a1 b7b2 a7e3 g8g7');
INSERT INTO evaluations VALUES(17764,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(17764,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(17764,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(17764,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(17764,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(17764,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(17764,36,'cp 75','c2c4 g8f6 b1c3 h7h6 f4g3 f8d6 g3h4 d6e7 d1c2 c6b4 c2b3 d5c4 f1c4 b4d5 c3d5 f6d5 h4e7 d5e7 e1g1 e8g8 c4d3 b7b6 d3e4 a8b8 a1c1 c8b7 e4b7 b8b7 f1d1 d8d6 h2h3 f8d8 e3e4 c7c6 b3c3 d6c7');
INSERT INTO evaluations VALUES(19212,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(19212,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19212,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(19212,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(19212,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(19212,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(19212,36,'cp 22','b5c6 b7c6 d2d3 f8g7 f2f4 g7c3 b2c3 d7d5 g1f3 d5e4 d3e4 d8d1 e1d1 g8f6 h1e1 c8g4 h2h3 g4f3 g2f3 f6d7 d1e2 f7f5 e2f2 f5e4 e1e4 d7f6 e4e2 c5c4 c1a3 f6d5 a1e1 e8g8 a3e7 f8f4 e7a3 d5c3 e2e6 c3d5 e6c6 c4c3');
INSERT INTO evaluations VALUES(19203,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(19203,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(19203,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(19203,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(19203,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(19203,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(19203,36,'cp 24','b7c6 d2d3 f8g7 f2f4 g7c3 b2c3 d7d5 g1f3 d5e4 d3e4 d8d1 e1d1 g8f6 h1e1 c8g4 h2h3 g4f3 g2f3 f6d7 d1e2 f7f5 e4f5 e8f7 c3c4 g6f5 c1a3 h8d8 e2f1 e7e6 a1b1 d7b6 a3c5 a8b8 b1b4 d8d2 e1e2 d2d1 f1f2 d1h1 f2g3 b8g8 g3h4');
INSERT INTO evaluations VALUES(18409,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(18409,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(18409,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(18409,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(18409,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(18409,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(18409,36,'cp 28','d2d3 f8g7 f2f4 d7d5 g1f3 d5e4 d3e4 g7c3 b2c3 d8d1 e1d1 g8f6 h1e1 c8g4 h2h3 g4f3 g2f3 e8c8 d1e2 c8c7 a2a4 c5c4 c1e3 f6d7 a1d1 a7a5 e1g1 d8b8 g1g5 b8a8 h3h4 f7f6 g5g1 a8b8 d1d2 b8b2 h4h5 b2a2 h5g6 h7g6 g1g6 h8h7 f4f5 d7b6 e3f4 c7b7');
INSERT INTO evaluations VALUES(30524,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(30524,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(30524,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(30524,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(30524,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(30524,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(30524,36,'cp 78','c2c4 g7g6 g2g3 b8d7 b1c3 g8f6 g1e2 f8g7 f1g2 a7a5 c1g5 d7c5 f2f3 e8g8 e1g1 h7h6 g5e3 e7e5 d1d2 h6h5 a1e1 b7c8 b2b3 h5h4 g3g4 f6h7 g1h1 g7f6 h2h3 f8e8 f3f4 e5f4 e3f4 d8e7 c3b5 c5e4');
INSERT INTO evaluations VALUES(30521,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(30521,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(30521,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(30521,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(30521,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(30521,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(30521,36,'cp 76','g7g6 b1c3 f8g7 g1e2 g8f6 g2g3 b8d7 f1g2 a7a5 h2h3 d7c5 d1c2 e8g8 c1e3 e7e6 e1g1 e6e5 b2b3 h7h5 a2a3 f6h7 f2f4 g7h6 a1e1 h5h4 g3g4 c5d7 g1h1 f8e8 c2d2 e5f4 e2f4 d7e5 c3b5 b7a6 b5d4');
INSERT INTO evaluations VALUES(30729,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(30729,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(30729,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(30729,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(30729,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(30729,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(30729,36,'cp 112','d7d6 d2d4 f8g7 d1e2 a7a6 c4b3 e7e6 c1f4 b8d7 e1c1 b6b5 c1b1 g8e7 a2a3 h7h6 h2h4 a8b8 h4h5 g6g5 f3g5 h6g5 f4g5 g7h6 g5h6 h8h6 g2g4 c7c5 g4g5 h6h8 f2f4 c5c4 b3a2 d8b6 d4d5 e6d5 e4d5 e8d8 h5h6 e7f5 e2h5 f5g3 h5f7 g3h1 d1h1');
INSERT INTO evaluations VALUES(39222,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(39222,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(39222,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(39222,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(39222,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(39222,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(39222,36,'cp 75','b1c3 c5b4 g1e2 g8f6 a2a3 b4c3 e2c3 d7d5 c4d5 f6d5 f1c4 d5b6 c4a2 d8e7 e1g1 c8e6 a2e6 e7e6 d2d4 e8g8 d4d5 e6g6 e3e4 c7c6 c1e3 c6d5 e4d5 b8d7 d1b3 a8c8 a3a4 e5e4 a4a5 b6c4 e3a7 d7e5 g1h1 e5d3 b3b7');
INSERT INTO evaluations VALUES(42445,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42445,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42445,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42445,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42445,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42445,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42445,36,'cp 0','b1d2 h7h6 g5f6 d8f6 a2a3 b4d2 d1d2 c7c5 d4c5 b8a6 g2g3 a6c5 a1d1 c5e4 d2e3 f6b2 e3e4 b2c3 d1d2 c3a1 d2d1');
INSERT INTO evaluations VALUES(43510,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(43510,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43510,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(43510,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(43510,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(43510,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(43510,36,'cp 0','f8e7 a2a4 e8g8 e1g1 b8c6 g5e3 c8d7 d4b3 a8c8 a4a5 c6e5 b3d2 d7c6 f2f4 e5g6 g2g3 d6d5 e3b6 d8e8 e4d5 f6d5 c3d5 c6d5 c2c4 d5c6 d2b3 e7d8 b6c5 d8e7');
INSERT INTO evaluations VALUES(39706,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39706,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39706,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39706,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39706,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39706,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39706,36,'cp 0','a2a4 c8d7 e1g1 e8g8 g5e3 b8c6 d4b3 c6a5 b3a5 d8a5 d1d2 f8e8 f1d1 d7c6 e3d4 h7h6 f2f3 a5c7 g1h1 a8c8 e2f1 f6h5 d4e3 h5f6');
INSERT INTO evaluations VALUES(45176,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(45176,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(45176,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(45176,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(45176,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(45176,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(45176,36,'cp -47','f6e4 f1d3 d6d5 d1e2 c5d4 d3e4 d5e4 e2e4 d4d3 e1g1 f7f6 f1d1 e7e5 d1d3 d8b6 b2b3 b8a6 d3d1 f8c5 e4e2 e8g8 c3c4 b6c6 b1c3 a6c7 c1e3 c5b4 c3d5 c7d5 c4d5 c6d6 e3d2 b4d2 e2d2 c8d7 a1c1 e5e4 f3e1 f8e8 h2h3 a8d8 a2a4 a7a6');
INSERT INTO evaluations VALUES(45270,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(45270,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(45270,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(45270,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(45270,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(45270,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(45270,36,'cp -25','f1d3 d6d5 d1e2 c5d4 d3e4 d5e4 e2e4 d4d3 e1g1 f7f6 f1d1 e7e5 d1d3 d8b6 b2b3 b8a6 e4e2 c8e6 d3d1 f8c5 c3c4 e8g8 b1c3 a8d8 c1d2 c5e7 d2e3 b6a5 e3d2 a5c7 c3b5 c7b8 d2e3 b7b6 b5c3 b8c7 c3d5 e6d5 d1d5 d8d5 c4d5 a6b4');
INSERT INTO evaluations VALUES(45932,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(45932,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(45932,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(45932,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(45932,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(45932,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(45932,36,'cp 81','e2e4 g7g6 f1e2 f8g7 f3d2 e8g8 e1g1 f8e8 b1c3 b8d7 a2a4 d7e5 d2b1 a7a6 h2h3 c5c4 c1e3 c8d7 a4a5 a8c8 b1d2 d7b5 b2b3 c4b3 c3b5 a6b5 d1b3 e5c4 e2c4 b5c4 d2c4 f6e4 a1c1 h7h5 f1e1 e4c5 b3a3 c5e4 c4b6 c8c1 e1c1');
INSERT INTO evaluations VALUES(54277,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(54277,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(54277,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(54277,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(54277,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(54277,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(54277,36,'cp -44','b8c6 c4d5 e6d5 f1b5 g8e7 d2d4 a7a6 b5a4 b7b5 a4c2 c8g4 d4c5 c6e5 b1d2 e7c6 h2h3 g4f3 d2f3 e5f3 d1f3 f8c5 c1e3 c5e3 f3e3 d8e7 e1g1 e8g8 f1d1 e7e3 f2e3 a8d8 c2b3 d5d4 a1c1 d8d6 a2a4 d4d3 a4b5 a6b5 c1c5 f8e8 d1f1 c6d8 c5b5 d8e6 b3d1 g7g6');
INSERT INTO evaluations VALUES(20315,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(20315,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(20315,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(20315,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(20315,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(20315,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(20315,36,'cp -48','c4d5 e6d5 f1b5 g8e7 d2d4 a7a6 b5a4 b7b5 a4c2 c8g4 d4c5 e7g6 b1d2 f8c5 h2h3 g4e6 e1g1 g6e5 f3e5 c6e5 d2b3 c5b6 d1h5 e5c4 c1g5 d8d6 f1e1 h7h6 a1d1 a8c8 b3d4 b6d4 d1d4 c4b2 c2b3 b2c4 g5f4 d6b6 d4d1 e8g8 f4h6 g7h6 h5h6');
INSERT INTO evaluations VALUES(20239,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(20239,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(20239,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(20239,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(20239,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(20239,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(20239,36,'cp -57','e6d5 f1b5 g8e7 d2d4 a7a6 b5a4 b7b5 a4c2 c8g4 d4c5 c6e5 b1d2 e7c6 h2h3 e5f3 d2f3 g4f3 d1f3 f8c5 c1e3 c5e3 f3e3 d8e7 e1g1 e8g8 f1d1 e7e3 f2e3 a8e8 a1c1 e8e3 d1d5 c6b4 d5d4 e3e1 c1e1 b4c2 e1d1 c2d4 d1d4 f8a8 g1f2 h7h6 f2f3 g7g6 d4d6 a6a5 d6b6');
INSERT INTO evaluations VALUES(20542,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(20542,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(20542,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(20542,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(20542,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(20542,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(20542,36,'cp -61','f1b5 g8e7 d2d4 a7a6 b5a4 c8g4 d4c5 g4f3 g2f3 e7g6 b1c3 f8c5 f3f4 d8h4 d1f3 a8d8 f3g3 b7b5 g3h4 g6h4 a4d1 e8g8 e1g1 g8h8 c1e3 d5d4 c3e4 c5b4 e3c1 d4d3 c1e3 c6d4 a2a3 b4e7 a3a4 f7f6 a4b5 a6b5 a1a7 h4f5 e4g3 f6e5 g3f5 d4f5 f4e5');
INSERT INTO evaluations VALUES(787,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(787,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(787,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(787,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(787,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(787,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(787,36,'cp 33','b1a3 e7f5 a3c2 c8d7 e1g1 a7a5 g1h1 a8c8 a2a4 h7h5 c1d2 b6b3 d2c3 c6b4 c2b4 b3d1 f1d1 f8b4 c3b4 a5b4 d1c1 e8g8 a4a5 g7g5 e2d3 g5g4 d3f5 e6f5 f3d2 f5f4 d2b3 c8c4 b3c5 d7c6 a5a6 b7a6 c5a6 h5h4 c1c4 d5c4');
INSERT INTO evaluations VALUES(788,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(788,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(788,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(788,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(788,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(788,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(788,36,'cp 25','c8d7 e1g1 e7f5 a3c2 a8c8 e2d3 c6b4 c2b4 f8b4 g1h1 b4f8 a2a4 a7a5 h2h3 f8b4 c1d2 h7h6 a1b1 f5e7 d1b3 e8g8 f1g1 e7f5 d2b4 a5b4 b3d1 b4b3 b1a1 b6b4 g2g4 f5e7 g4g5 h6h5 d1d2 b4d2 f3d2 b7b6 d2b3 c8a8 a4a5 b6a5');
INSERT INTO evaluations VALUES(10578,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(10578,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(10578,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(10578,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(10578,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(10578,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(10578,36,'cp 126','d1e2 c5f2 g1h1 d7d5 d2d3 e4g3 h2g3 f2g3 c1g5 f7f6 g5h4 g3h4 f3h4 e8g8 b5c6 b7c6 b1d2 f6f5 g2g3 d8d6 a1e1 e5e4 e2e3 c8d7 h4g2 a7a5 d2b3 e4d3 e3d3 c6c5 d3e3 d7b5 f1f3 d5d4 e3e6 d6e6 e1e6 b5c4 e6e7 c4d5 f3f2 d4c3');
INSERT INTO evaluations VALUES(27446,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(27446,31,'cp -160',NULL);
INSERT INTO evaluations VALUES(27446,32,'cp -175',NULL);
INSERT INTO evaluations VALUES(27446,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(27446,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(27446,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(27446,36,'cp -163','h2h3 g4h5 c1d2 e8c8 d2c3 c6b4 g2g4 h5g6 d1e2 b4c6 a2a3 h7h5 f3h4 g6h7 h4f5 c6d4 c3d4 e5d4 e2f3 d8e8 e1f1 e8e5 e4d2 h7f5 f3d5 e5d5 g4f5 d5f5 b2b4 c8d7 a1e1 b7b6 e1e4 c7c5 b4c5 f8c5 a3a4 h8e8 e4e8 d7e8 d2e4 f7f6 f1g2 f5f4');
INSERT INTO evaluations VALUES(27444,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(27444,31,'cp -185',NULL);
INSERT INTO evaluations VALUES(27444,32,'cp -186',NULL);
INSERT INTO evaluations VALUES(27444,33,'cp -154',NULL);
INSERT INTO evaluations VALUES(27444,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(27444,35,'cp -159',NULL);
INSERT INTO evaluations VALUES(27444,36,'cp -181','g4h5 c1d2 e8c8 d2c3 c6b4 g2g4 h5g6 d1e2 b4c6 a2a3 h7h5 f3h4 g6h7 h4f5 c6d4 c3d4 e5d4 e2f3 d8e8 e1f1 e8e5 e4d2 h7f5 f3d5 e5d5 g4f5 d5f5 a1e1 c8d7 e1e4 c7c5 f1e2 b7b6 c2c3 d4c3 b2c3 g7g6 d2c4 f8d6 a3a4 h8d8 h1e1 d6c7 c4e3 f5f4 e4f4');
INSERT INTO evaluations VALUES(27443,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(27443,31,'cp -166',NULL);
INSERT INTO evaluations VALUES(27443,32,'cp -170',NULL);
INSERT INTO evaluations VALUES(27443,33,'cp -192',NULL);
INSERT INTO evaluations VALUES(27443,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(27443,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(27443,36,'cp -185','c1d2 e8c8 d2c3 c6b4 g2g4 h5g6 d1e2 b4c6 a2a3 f8e7 h3h4 h7h5 f3d2 f7f6 e4g3 g6f7 g4h5 f7h5 e2e4 h5f7 e4d5 f7d5 f2f3 g7g6 h4h5 g6h5 g3f5 d8g8 f5e7 c6e7 e1f2 h5h4 a1g1 g8g1 h1g1 d5e6 d2e4 e7d5 d3d4 d5c3');
INSERT INTO evaluations VALUES(32323,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32323,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32323,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32323,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32323,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32323,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(32323,36,'cp 0','h7h6 c2c4 c7c6 e1g1 b8d7 f3d2 f8e7 e2e4 d5e4 b1c3 g8f6 d2e4 e8g8 e4f6 d7f6 d1b3 d8b6 d4d5 b6b3 a2b3 e6d5 c4d5 f6d5 c3d5 c6d5 g2d5 f5e6 d5b7 a8b8 b7e4 e7c5 c1d2 b8b3 d2c3 f8c8 f1c1 c5b6 e4c2 b3b5');
INSERT INTO evaluations VALUES(45522,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(45522,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(45522,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(45522,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(45522,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(45522,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(45522,36,'cp 58','a7a6 h2h3 b7b5 c1f4 f8g7 f1d3 e8g8 e1g1 d8e7 d1d2 b8d7 f1e1 f8e8 a1c1 d7e5 f3e5 d6e5 f4e3 c5c4 d3c2 e7d6 f2f3 c8d7 b2b3 c4b3 c2b3 e8c8 c3e2 b5b4 c1c8 a8c8 e1c1 a6a5 c1c8 d7c8 e2c1 g7f8 c1d3 f6d7 d3b2 c8a6');
INSERT INTO evaluations VALUES(54762,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(54762,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(54762,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(54762,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(54762,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(54762,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(54762,36,'cp 70','c1f4 a7a6 e2e3 g8f6 f1d3 e7e6 g1f3 c7c5 d4c5 f8c5 e1g1 b8c6 c3e2 d8e7 e2d4 c8d7 c2c4 c5a7 a1c1 e8g8 a2a3 a8c8 f1e1 h7h6 b2b4 c6d4 e3d4 d5c4 d3c4 f6e4 d1e2 d7e8 e2e3 a7b6 f3d2 c8d8 f4e5 e4d2 e3d2 e8c6 c4b3');
INSERT INTO evaluations VALUES(54753,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(54753,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(54753,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(54753,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(54753,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(54753,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(54753,36,'cp 54','e7e6 e2e3 g8f6 g1f3 a7a6 f1d3 c7c5 d4c5 f8c5 e1g1 b8c6 c3e2 d8e7 e2d4 e8g8 c2c4 c8d7 a1c1 c5a7 a2a3 a8c8 f1e1 d5c4 d3c4 f6e4 d1b3 c6d4 f3d4 a7d4 e3d4 b7b5 c4f1 c8c1 e1c1 f8c8 b3e3 e4f6 h2h3 f6d5 e3e5 c8c1 f4c1');
INSERT INTO evaluations VALUES(56763,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(56763,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(56763,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(56763,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(56763,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(56763,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(56763,36,'cp 60','e2e3 g8f6 f1d3 a7a6 g1f3 c7c5 d4c5 f8c5 e1g1 b8c6 c3e2 d8e7 e2d4 e8g8 c2c4 c8d7 a2a3 c5a7 f1e1 c6d4 e3d4 d5c4 d3c4 f6e4 a1c1 a8c8 a3a4 d7e8 d1b3 e8f7 f4e5 e7d7 a4a5 h7h6 h2h3 f8d8 b3e3 f7h5 g2g4 h5g6 g4f5 g6f5 c4d3');
INSERT INTO evaluations VALUES(55934,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55934,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(55934,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55934,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55934,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(55934,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(55934,36,'cp 22','c1b2 g8f6 e2e3 f8d6 c2c4 e8g8 f1e2 b7b6 b1c3 c8b7 c4d5 e6d5 c3b5 d6e7 e1g1 c7c5 d2d4 b8d7 e2d3 a7a6 b5c3 e7d6 c3e2 f8e8 a1c1 d8e7 f1e1 a8c8 e2f4 h7h6 f4e2 g7g6 d1c2 f6e4 e2g3 h6h5 d4c5 b6c5');
INSERT INTO evaluations VALUES(57512,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(57512,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(57512,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(57512,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(57512,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(57512,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(57512,36,'cp 111','e4f5 d8f6 d1h5 g7g6 f5g6 h7g6 h5f3 b8c6 f3f6 g8f6 g1f3 d7d5 c1g5 f6h7 g5d2 c8e6 b1c3 e8c8 f1e2 f8c5 h1f1 c6d4 e1c1 d4e2 c3e2 e6g4 f3g1 d8f8 d2e3 d5d4 e3h6 f8f7 h6d2 h8f8 f2f3 g4e6 a2a3 h7f6 e2g3 c5d6 d2g5 f6h5 g3e4 h5g7');
INSERT INTO evaluations VALUES(57486,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(57486,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(57486,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(57486,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(57486,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(57486,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(57486,36,'cp 94','d8f6 d1h5 g7g6 f5g6 h7g6 h5f3 f6g7 b1c3 b8c6 g2g3 c6d4 f3d1 c7c6 f1h3 g8f6 c3e4 d7d5 e4f6 g7f6 h3c8 a8c8 c2c3 d4e6 d1f3 f8g7 f3f6 g7f6 h2h4 e8f7 e1e2 g6g5 g1f3 g5h4 f3h4 c8e8 c1e3 d5d4 e3d2 e6f4 g3f4 e5f4 e2f3 h8h4 h1h4 f6h4 c3d4');
INSERT INTO evaluations VALUES(4186,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(4186,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4186,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(4186,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(4186,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(4186,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(4186,36,'cp 3','d7d6 b1d2 a7a6 c4b3 c8e6 d2f1 e6b3 d1b3 f8e8 c1e3 c5e3 f1e3 d8d7 a2a4 b7b6 h2h3 a6a5 b3b5 c6e7 b5d7 f6d7 d3d4 d7f6 e3g4 f6g4 h3g4 e7c6 a1c1 a8c8 c1c2 c8d8 c2d2 e5d4 f3d4 c6d4 c3d4 d6d5');
INSERT INTO evaluations VALUES(3621,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(3621,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(3621,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(3621,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(3621,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(3621,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(3621,36,'cp 25','b2b4 c5b6 b1d2 c6e7 a2a4 a7a5 b4b5 e7g6 d3d4 f8e8 h2h3 e5d4 c3d4 d6d5 e4d5 e8e1 d1e1 f6d5 d2b3 c8e6 b3c5 b6c5 d4c5 d8f6 a1a3 f6e7 c4d3 g6f4 c1f4 d5f4 e1e5 f4d5 f3d4 a8d8 d4e6 e7e6 e5e6 f7e6 d3c4 g8f7 a3f3 d5f6 g2g3 d8d1 g1g2 d1b1 c5c6 b7c6');
INSERT INTO evaluations VALUES(8683,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(8683,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(8683,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(8683,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(8683,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(8683,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(8683,36,'cp 86','b1c3 c7c6 e1c1 d8b6 g5f6 g7f6 d4d2 b6c5 c4e2 c8e6 f2f4 e8c8 f4f5 e6c4 e2c4 c5c4 h1f1 d8d7 f1f3 e7d8 f3d3 d8a5 d2e3 c8b8 b2b3 c4a6 e3d4 a5c3 d4c3 a6a2 c3f6 h8d8 f6c3 a7a5 e4e5 d6d5 c3d4 d8e8 d1e1 d7e7 e5e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(12048,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(12048,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(12048,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(12048,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(12048,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(12048,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(12048,36,'cp 102','g5e4 f6e4 d3e4 f8e7 d2d3 e8g8 e1g1 f7f5 e4f3 c8a6 f1e1 d8c7 b1d2 a8d8 d2b3 e5e4 c1d2 a5b3 a2b3 c7b7 f3h5 e7f6 d1b1 e4d3 c2c4 b7c8 d2a5 d8d6 e1d1 c6c5 d1d3 d6d3 b1d3 f6b2 a1a2 b2f6 h5f3 a6b7 f3d5 g8h8 a5c3 b7d5 c4d5 f8e8 g2g3 f6c3 d3c3 c8d7 a2d2 d7d6');
INSERT INTO evaluations VALUES(12051,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(12051,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(12051,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(12051,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(12051,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(12051,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(12051,36,'cp 104','f6e4 d3e4 f8e7 d2d3 e8g8 e1g1 f7f5 e4f3 c8a6 f1e1 d8c7 b1d2 a8d8 d2b3 e5e4 c1d2 a5b3 a2b3 c7b7 f3h5 e7f6 d1b1 e4d3 c2c4 b7c8 d2a5 d8d7 e1d1 c6c5 d1d3 d7d3 b1d3 f6b2 a1a2 b2f6 h5f3 a6b7 f3d5 b7d5 c4d5 f6d4 d5d6 c8c6 a5c7 f8e8');
INSERT INTO evaluations VALUES(12063,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(12063,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(12063,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(12063,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(12063,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(12063,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(12063,36,'cp 90','d3e4 f8d6 g2g3 c8h3 h1g1 h3e6 d2d3 e8g8 g3g4 d6e7 b1c3 a8b8 e4f5 a5b7 f5e6 f7e6 g4g5 h6g5 d1g4 d8b6 c3d1 f8f5 c1e3 e7c5 a2a3 b7d6 b2b4 c5e3 d1e3 f5f7 c2c3 c6c5 g4e6 b8e8 e6d5 e5e4 d3d4 c5d4 d5d4 b6d4 c3d4 d6b5 g1g5 b5d4 e1c1');
INSERT INTO evaluations VALUES(12062,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(12062,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(12062,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(12062,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(12062,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(12062,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(12062,36,'cp 86','f8d6 d2d3 e8g8 g2g4 a8b8 b1c3 d6e7 d1f3 e7g5 c1g5 d8g5 h2h4 g5g4 e1c1 g4f3 e4f3 c6c5 d1e1 f7f6 f3d5 g8h8 h1f1 c8b7 d5b7 a5b7 h4h5 b8e8 e1e4 b7d8 e4a4 f8f7 c3e4 f7c7 a4a6 d8e6 f1h1 e6g5 a6a5 g5e4 d3e4 e8d8 h1d1 d8d4 d1d4 e5d4 c1d2 h8g8');
INSERT INTO evaluations VALUES(18711,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(18711,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(18711,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(18711,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(18711,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(18711,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(18711,36,'cp -112','c5d4 c2c3 d8a5 c1d2 c6e5 c3d4 e5f3 d1f3 a5d8 f3g3 d7d5 b1c3 a7a6 f1d3 g8f6 a1c1 c8d7 e1g1 d7c6 c3e2 g7g6 g3h3 f6h5 c1c2 f8e7 g2g4 h5g7 d2f4 e7d6 f4e5 e8g8 h3g3 g7e8 e2f4 a8c8 h2h4 e8f6 f1c1 d6e5 d4e5 f6d7 h4h5 g6g5');
INSERT INTO evaluations VALUES(18872,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(18872,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(18872,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(18872,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(18872,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(18872,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(18872,36,'cp -104','c2c3 d8a5 c1d2 c6e5 c3d4 e5f3 d1f3 a5d8 b1c3 d7d5 f3g3 a7a6 d2f4 c8d7 f1d3 g8f6 f4c7 d8e7 e1g1 d7c6 a1e1 g7g6 h2h4 e7d7 d3e2 f8g7 c7d6 f6e4 c3e4 d5e4 h4h5 f7f6 h5g6 h7g6 g3g6 d7f7 g6g3 f6f5 e1c1 g7d4 f1d1');
INSERT INTO evaluations VALUES(19602,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(19602,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(19602,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(19602,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(19602,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(19602,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(19602,36,'cp 101','f3d4 c5d4 d1d4 e7e6 b1c3 a7a6 f1e2 d8c7 c1d2 c7c5 d4d3 c5c7 d3g3 c7g3 h2g3 b7b5 a2a4 b5b4 c3a2 a6a5 c2c3 b4c3 a2c3 f8b4 f2f3 e8e7 e1c1 c8a6 e2a6 a8a6 c1b1 g8f6 e4e5 f6d5 c3d5 e6d5 d2b4 a5b4');
INSERT INTO evaluations VALUES(19600,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(19600,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(19600,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(19600,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(19600,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(19600,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(19600,36,'cp 87','c5d4 d1d4 e7e6 b1c3 a7a6 f1e2 d8c7 a2a4 g8e7 c1e3 e7c6 d4b6 c7b6 e3b6 f8b4 e1c1 d7d6 d1d3 c8d7 f2f4 g7g5 f4g5 c6e5 d3d4 b4c5 b6c5 d6c5 d4d1 e8e7 h2h4 b7b5 a4b5 a6b5 e2b5 d7b5 c3b5 a8a1 c1d2 a1d1 h1d1 h8b8 b5a7 b8b2 a7c8 e7f8 d2c3 b2b8');
INSERT INTO evaluations VALUES(19617,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(19617,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(19617,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(19617,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(19617,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(19617,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(19617,36,'cp 91','d1d4 e7e6 c1f4 a7a6 f1e2 d7d6 f4e3 g8f6 c2c4 f8e7 b1c3 e8g8 e1g1 b7b6 b2b4 c8b7 a1c1 f6d7 f1d1 h7h6 f2f4 b7c6 g1h1 a8c8 a2a3 f8e8 d4d3 e7f8 h2h3 d7f6 e3g1 c6b7 d3e3 d8c7 e3b6 f6e4 b6c7');
INSERT INTO evaluations VALUES(32126,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(32126,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32126,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(32126,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(32126,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(32126,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(32126,36,'cp 42','e5d4 f3b3 g8f6 f1c4 d6d5 c4d5 f6d5 e4d5 f8c5 e1g1 e8g8 c2c4 b7b6 c1f4 c7c6 b1d2 c5d6 f4d6 d8d6 d2e4 d6g6 e4g3 b8d7 d5c6 g6c6 b3d3 f8e8 d3d4 d7e5 b2b3 a8d8 d4c3 e5d3 g3h5 c6g6 a1d1 d3f2 d1d8 e8d8 h5f6 g6f6 c3f6 g7f6 g1f2');
INSERT INTO evaluations VALUES(32197,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(32197,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(32197,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(32197,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(32197,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(32197,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(32197,36,'cp 53','f3b3 g8f6 f1c4 d6d5 c4d5 f6d5 e4d5 f8e7 c2c3 b8d7 b3b7 d7b6 b7c6 d8d7 c3d4 e8c8 c1e3 d7d5 c6d5 b6d5 b1c3 e7f6 e1g1 h8e8 c3d5 d8d5 f1d1 e8d8 d1d3 f6d4 a1d1 c7c5 g1f1 c8c7 e3d4 c5d4 d1c1 c7d6 c1c4 d6e7 b2b3 e7f8 f1e2 d8e8 e2d2 d5f5 f2f3 e8e6 g2g4');
INSERT INTO evaluations VALUES(40747,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(40747,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(40747,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(40747,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(40747,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(40747,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(40747,36,'cp 69','e5g3 e8g8 b1c3 f6e4 c3e4 d6d5 c4d5 d8d5 d2d3 c8f5 e4c3 f8e8 c1e3 d5d7 g1e2 c7c5 e1c1 b8c6 f2f3 c6d4 e2f4 e7d6 g3f2 b7b5 c3e4 d6f4 e3f4 d7d5 e4c3 d5c6 f4e3 a8d8 h1e1 f5e6 c1b1 b5b4 c3e4 a7a5 e3d4 c5d4');
INSERT INTO evaluations VALUES(40818,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(40818,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(40818,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(40818,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(40818,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(40818,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(40818,36,'cp 75','e8g8 b1c3 f6e4 c3e4 d6d5 c4d5 d8d5 d2d3 c8f5 e4c3 f8e8 c1e3 d5d7 g1e2 c7c5 e1c1 b8c6 f2f3 c6d4 e2f4 e7d6 g3f2 b7b5 h1e1 b5b4 c3e4 e8c8 e4d6 d7d6 c1b1 d4b5 f2g3 d6d7 g3h4 a7a5 g2g4 f5e6 f4e6 d7e6 e3f2 e6d5 e1e7 d5c6 d1e1');
INSERT INTO evaluations VALUES(42496,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(42496,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(42496,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(42496,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(42496,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(42496,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(42496,36,'cp 24','d7d6 f3d2 e6e5 e2e4 b8c6 c1b2 c6e7 g2g3 b7b6 f1g2 h7h5 h2h4 e8g8 e1g1 c8e6 f1e1 d8c8 b2c1 e6g4 f2f3 g4h3 d2f1 h3g2 g1g2 f6h7 f1e3 c8e6 e3d5 a8d8 d1c2 c7c6 d5e7 e6e7 c1e3 e7c7 a2a4 f8e8 g2h2 d6d5 c4d5 c6d5 e4d5');
INSERT INTO evaluations VALUES(42599,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(42599,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(42599,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42599,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(42599,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(42599,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(42599,36,'cp 22','g1f3 d7d6 a2a3 e8g8 e2e3 b8d7 b2b4 c5b6 c3a4 a7a5 f1e2 b6a7 e1g1 c7c6 c1d2 a7b8 d1c2 f8e8 b4a5 b8c7 d2b4 d8e7 c4c5 d6c5 a4c5 c7a5 c5d7 e7d7 b4a5 a8a5 f3d2 d7e7 d2c4 a5c5 a3a4 e5e4 d3d4 c5g5');
INSERT INTO evaluations VALUES(42597,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(42597,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(42597,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(42597,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(42597,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42597,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(42597,36,'cp 8','d7d6 a2a3 e8g8 e2e3 b8d7 b2b4 c5b6 c3a4 a7a5 f1e2 b6a7 e1g1 c7c6 c1d2 a7b8 d1c2 b8c7 a1b1 a5b4 a3b4 d6d5 c4d5 c6d5 f1c1 c7d6 b1a1 d8e7 c2b2 h7h6 d2e1 f8e8 h2h3 d7f8 a4c5 a8a1 b2a1');
INSERT INTO evaluations VALUES(43336,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43336,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43336,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(43336,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(43336,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43336,36,'cp 0','e7e5 d4b3 f8b4 c1d2 e8g8 f1d3 d7d6 e1g1 c8e6 a2a3 b4c3 d2c3 b8c6 f1e1 d8c7 c3d2 f6d7 d2e3 b7b5 d1d2 d7b6 b3a5 b6a4 a5c6 c7c6 a1b1 e6a2 b1a1');
INSERT INTO evaluations VALUES(43329,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43329,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43329,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43329,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43329,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43329,36,'cp 1','d4b3 f8b4 c1d2 d7d6 f1d3 e8g8 e1g1 c8g4 d3e2 g4e2 d1e2 b8d7 f1d1 d8c7 a2a3 b4c3 d2c3 a8c8 f2f3 d7b6 a3a4 f6h5 g2g3 f7f5 e4f5 h5f6 c3a5 c7c4 d1d2 c4e2 d2e2 b6c4 a1d1 c4b2 d1d6 b2c4 d6d1 c4b2 d1c1 f8e8 a5b4 b2a4 b3a5 b7b5 g1h1');
INSERT INTO evaluations VALUES(43330,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43330,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43330,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43330,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43330,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43330,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43330,36,'cp 0','f8b4 c1d2 d7d6 f1d3 e8g8 e1g1 b8c6 a2a3 b4c3 d2c3 c8e6 f1e1 d8c7 c3d2 f6d7 d2e3 b7b5 d1d2 d7b6 b3a5 b6a4 a5c6 c7c6 a1b1 e6a2 b1a1');
INSERT INTO evaluations VALUES(39656,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39656,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39656,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39656,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39656,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39656,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39656,36,'cp 0','c1d2 d7d6 f1d3 e8g8 e1g1 b8c6 a2a3 b4c3 d2c3 c8e6 f1e1 d8c7 c3d2 f6d7 d2e3 b7b5 d1d2 d7b6 b3a5 b6a4 a5c6 c7c6 a1b1 e6a2 b1a1');
INSERT INTO evaluations VALUES(2787,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2787,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2787,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2787,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2787,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2787,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2787,36,'cp 0','d7d6 d2d4 c5b6 a2a4 c8g4 c1g5 e5d4 a4b5 a6b5 a1a8 d8a8 g5f6 g7f6 b3d5 a8e8 d5c6 g4f3 d1f3 e8c6 f3f6 f8e8 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(9666,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9666,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9666,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(9666,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9666,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9666,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(9666,36,'cp -3','d2d3 e8g8 c1g5 c5e7 b1d2 c6a5 g1e2 b7b5 c4b5 a8b8 a2a4 c7c6 b2b4 c6b5 b4a5 d7d5 a4b5 b8b5 e1g1 f8e8 g5f6 e7f6 f1d1 f6g5 d2f1 d5e4 d3e4 d8c7 f1e3 g5e3 f3e3 c8e6 a5a6 b5c5 d1c1 e8d8 c1d1 d8d6 d1d6 c7d6');
INSERT INTO evaluations VALUES(9667,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9667,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9667,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9667,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9667,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9667,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9667,36,'cp 0','e8g8 c1g5 c5e7 b1d2 c6a5 g1e2 d7d6 h2h3 a5c4 d2c4 c8e6 c4e3 c7c6 d3d4 e5d4 e2d4 e6d7 g5h6 f8e8 e1g1 e7f8 h6g5 f8e7');
INSERT INTO evaluations VALUES(21365,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(21365,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21365,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21365,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21365,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(21365,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(21365,36,'cp 26','e6d5 b1c3 c8f5 a2a3 g8f6 c1g5 f8e7 e2e3 f6e4 g5e7 c6e7 f1d3 c7c6 d3e4 f5e4 c3e4 d5e4 f3d2 f7f5 a1c1 d8d5 c1c5 d5f7 d1b3 f7b3 d2b3 e7d5 e1e2 e8c8 b3d2 g7g6 h2h4 h8e8 g2g3 d5f6 h1c1 e8e6 d2c4 b7b6 c5e5 e6e5 d4e5');
INSERT INTO evaluations VALUES(21737,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(21737,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(21737,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(21737,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(21737,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(21737,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(21737,36,'cp -61','d2d3 c8e6 c1b2 d8d7 a2a3 f7f6 b1d2 h7h5 h2h4 g8h6 g1f3 f8e7 b3b4 e8g8 c2c4 d5d4 d2e4 a7a5 b4b5 c6d8 b2c1 h6f7 a3a4 c7c5 f3h2 e6h3 g2h3 d7h3 e1d2 h3d7 h2f3 f6f5 e4g5 e7g5 h4g5');
INSERT INTO evaluations VALUES(23360,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(23360,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(23360,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(23360,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(23360,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(23360,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(23360,36,'cp 113','d7d6 h1g1 h7h5 f1e2 g7g6 f3h4 f8g7 g2g3 a7a6 d1d3 b7b5 a2a3 a8b8 f2f3 g8h6 c1d2 c8d7 c3a2 b8a8 d3b3 d8c8 a2b4 h6f7 e1c1 f7d8 c1b1 a6a5 b4a2 b5b4 a3b4 a5b4 a2b4 a8b8 f3f4 d8f7 f4e5 f7e5');
INSERT INTO evaluations VALUES(20799,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(20799,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(20799,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(20799,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(20799,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(20799,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(20799,36,'cp 117','f3h4 a7a6 f1d3 d8d7 h2h3 g7g6 a2a4 f8h6 g2g3 h6g7 g3g4 g8h6 c1e3 h6f7 d1e2 e8g8 a4a5 d7d8 f2f3 c8d7 h4g2 f6f5 h3h4 f5g4 f3g4 d8c8 g4g5 d7g4 e2d2 c8d7 d3c4 h7h5 c4e2 c7c6 e1c1 g4e2 d2e2 c6d5');
INSERT INTO evaluations VALUES(36864,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(36864,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(36864,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(36864,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(36864,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(36864,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(36864,36,'cp 93','d4d5 c7c6 d1d2 b8a6 g2g4 c8d7 g1h3 c6d5 c4d5 h7h5 g4g5 f6e8 h3f2 a6c5 a1c1 e8c7 f2d3 b7b6 f1e2 a7a5 e1g1 c5d3 e2d3 b6b5 b2b3 d7h3 f1f2 h3d7 c3d1 c7e8 h2h4 a8b8 d3f1 f7f5 e4f5 d7f5 d1c3 b5b4 c3e4');
INSERT INTO evaluations VALUES(36837,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(36837,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(36837,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(36837,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(36837,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(36837,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(36837,36,'cp 84','c7c6 d1d2 c6d5 c4d5 b8a6 g2g4 h7h5 g4g5 f6e8 g1h3 c8d7 h3f2 e8c7 a1b1 d8e8 f1e2 a6c5 e1g1 a7a5 b2b3 b7b5 f1c1 e8d8 a2a3 a8b8 b3b4 a5b4 a3b4 c5a4 f2d1 a4b6 d1b2 d8e7 b1a1 b8b7 c3d1 f8c8 a1a5 g7f8 d1c3 b7b8 h2h4');
INSERT INTO evaluations VALUES(36059,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(36059,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(36059,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(36059,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(36059,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(36059,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(36059,36,'cp 95','d1d2 c6d5 c4d5 c8d7 g2g4 b8a6 g1h3 h7h5 g4g5 f6e8 h3f2 a6c5 b2b4 c5a4 c3a4 d7a4 b4b5 a7a6 b5b6 a8c8 d2b4 a4d7 f1e2 f7f6 h2h4 f6g5 h4g5 f8f4 h1g1 f4f7 a2a4 g7f8 b4d2 e8g7 a4a5 d8e8 d2b2 f8e7 f2d3 e8f8 a1c1 c8c1 e3c1');
INSERT INTO evaluations VALUES(55385,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(55385,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(55385,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(55385,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(55385,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(55385,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(55385,36,'cp 93','g1f3 c7c5 c2c3 b8c6 f1d3 f8g7 e1g1 g8e7 b1a3 e8g8 a3c2 f7f6 e5f6 f8f6 f1e1 c5d4 c2d4 d8d6 d1e2 c6d4 c3d4 e7c6 c1e3 c8d7 d3b5 a8f8 b5c6 d7c6 a1c1 f6f5 e3d2 f5f3 g2f3 g7d4 e2e6 d6e6 e1e6');
INSERT INTO evaluations VALUES(771,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(771,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(771,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(771,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(771,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(771,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(771,36,'cp -35','c5d4 b2b4 a7a6 a2a3 d4c3 b1c3 e7g6 f1e1 b6d8 e2d3 f8e7 c3a4 b7b5 a4c5 e8g8 d1c2 e7c5 b4c5 d8c7 c1b2 c8b7 h2h4 h7h6 d3g6 f7g6 c2g6 c7f7 g6g3 f7f4 a1d1 f4g3 f2g3 a6a5 f3d4 c6d4 b2d4 b7c6 h4h5 g8h7 d1c1');
INSERT INTO evaluations VALUES(792,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(792,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(792,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(792,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(792,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(792,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(792,36,'cp -44','b2b4 a7a6 a2a3 d4c3 b1c3 e7g6 f1e1 f8e7 c1b2 b6d8 e2d3 e8g8 a1c1 e7g5 c1c2 c8d7 c3e2 g5e7 d1d2 g6h4 f3h4 e7h4 d2e3 f7f5 g2g3 h4g5 f2f4 g5e7 e1c1 f8f7 g1g2 g8h8 b2d4 d8f8 d4b6 f8e8');
INSERT INTO evaluations VALUES(1680,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(1680,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(1680,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(1680,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(1680,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(1680,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(1680,36,'cp 142','f1d3 g8f6 e1g1 d5d8 f1e1 c6b4 d3c4 f8d6 a2a3 b4d5 c4f1 d5e7 c2c4 e7g6 b2b4 e8g8 d1c2 d6e7 c1b2 c7c6 b1d2 a7a5 b4a5 d8a5 a3a4 c6c5 d2b3 a5c7 b3c5 b7b6 c5d3 c8b7 f3e5 f6d7 e5g4 e7d6 b2a3 d6a3 a1a3 f8d8 h2h3 g6h4');
INSERT INTO evaluations VALUES(1933,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(1933,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(1933,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(1933,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(1933,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(1933,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(1933,36,'cp 20','g8f6 b1c3 f8b4 c1d2 d7d6 a2a3 b4c5 f2f3 c8e6 c3b5 c5b6 c2c3 e8c8 e1c1 d6d5 e4e5 f6d7 d1e1 d7c5 b5d4 e6d7 h2h4 c5e6 d4e6 d7e6 h4h5 b6f2 e1e2 f2g3 h5h6 g7g6 d2g5 d8e8 g5f6 g3f4 c1b1 h8f8 b2b4 c7c6 f6g7');
INSERT INTO evaluations VALUES(1060,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(1060,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(1060,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(1060,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(1060,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(1060,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(1060,36,'cp 28','b1c3 f8b4 c1d2 d7d6 a2a3 b4c5 f2f3 a7a6 e1c1 c8e6 c3e2 d6d5 e4e5 f6d7 e2f4 d7e5 d1e1 e5c6 f4d5 e8c8 d5f4 e6f5 f1c4 h8f8 g2g4 f5d7 f4d3 c5a7 h2h4 c6d4 d3e5 d7b5 c4f7 d4f3 f7e6 c8b8 e5f3 f8f3 e6f5 g7g6 d2g5 d8e8');
INSERT INTO evaluations VALUES(1059,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(1059,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(1059,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(1059,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(1059,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(1059,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(1059,36,'cp 18','f8b4 c1d2 d7d6 f2f3 a7a6 a2a3 b4c5 c3e2 d6d5 e4d5 f6d5 e2f4 d5f4 d2f4 c5b6 a3a4 c8d7 a4a5 b6a7 f1c4 e8c8 e1c1 h8e8 d1d2 e8e7 h2h4 d8e8 h1d1 d7b5 c4b5 a6b5 c2c3 f7f6 b2b4 h7h5 c1b2 e7e2 b2b3 e2d2 d1d2 c7c6');
INSERT INTO evaluations VALUES(557,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(557,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(557,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(557,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(557,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(557,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(557,36,'cp 21','c1d2 d7d6 f2f3 a7a6 c3d5 f6d5 e4d5 b4c5 f1d3 e8g8 e1c1 c8d7 b2b3 f8e8 d1e1 h7h5 a2a4 g7g6 h1f1 c5d4 h2h3 e8e1 f1e1 a8e8 e1e8 d7e8 g2g4 e8d7 a4a5 h5g4 h3g4 g8f8 c1d1 d4e5 d1e2 f8e7 g4g5 c7c6 d3e4 c6d5 e4d5');
INSERT INTO evaluations VALUES(5275,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(5275,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(5275,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(5275,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(5275,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(5275,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(5275,36,'cp -18','c6d4 e1g1 a7a6 b5e2 e7e6 e4e5 f7f5 d2d3 d7d6 a2a4 g8e7 e5d6 d8d6 f3d4 g7d4 g1h1 b7b6 c3b1 e8g8 b1d2 c8b7 e2f3 b7d5 d2c4 d6c7 c2c3 d4f6 f3d5 e7d5 d1e2 f8e8 f1e1 a8d8 h2h3 b6b5 a4b5 a6b5 c4e5 b5b4');
INSERT INTO evaluations VALUES(5421,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(5421,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(5421,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(5421,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(5421,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(5421,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(5421,36,'cp -26','e1g1 a7a6 b5e2 b7b5 d2d3 c8b7 f3d4 c5d4 c3b1 e7e6 a2a4 g8e7 a4b5 a6b5 a1a8 b7a8 b1d2 d8c7 d2b3 e8g8 e2f3 e6e5 f4e5 g7e5 g2g3 d7d5 c1h6 f8e8 d1e2 d5e4 d3e4 c7b6 e2d3 a8c6 f3g2 e7c8 f1a1');
INSERT INTO evaluations VALUES(5420,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(5420,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(5420,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(5420,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(5420,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(5420,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(5420,36,'cp -24','a7a6 b5e2 b7b5 d2d3 c8b7 f3d4 c5d4 c3b1 e7e6 f4f5 d8b6 a2a4 e6f5 e4f5 g8f6 e2f3 e8g8 c1g5 b7f3 d1f3 a8c8 a4b5 a6b5 f5g6 f7g6 b1d2 h7h6 g5f6 f8f6 f3d5 b6e6 d5e6 f6e6 f1f2 h6h5 d2f3 e6c6 a1a7 c6c2 a7d7 c2f2 g1f2 c8c2 f2f1 c2b2 d7d8 g8f7');
INSERT INTO evaluations VALUES(8436,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(8436,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(8436,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(8436,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(8436,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(8436,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(8436,36,'cp 20','a2a3 e8g8 h2h3 a7a5 d3d4 e5d4 f3d4 f8e8 c1e3 c6d4 e3d4 c7c6 a3a4 c8e6 c4d3 c6c5 d4e3 d6d5 c3d5 f6d5 e4d5 d8d5 d3b5 d5d1 f1d1 e8d8 g2g4 g7g5 g1g2 e7f6 c2c3 b7b6 f2f4 g5f4 e3f4 e6b3 d1d6 d8d6 f4d6');
INSERT INTO evaluations VALUES(8449,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(8449,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(8449,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(8449,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(8449,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8449,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(8449,36,'cp 21','e8g8 h2h3 a7a5 d3d4 c6d4 f3d4 e5d4 d1d4 c8e6 c1e3 c7c6 c4e6 f7e6 e4e5 d6e5 d4e5 d8d6 e5d6 e7d6 f1e1 d6c7 e3c5 f8e8 a1d1 f6d5 c3e4 a8d8 c2c4 b7b6 c5d4 e6e5 d4e3 d5e3 d1d8 e8d8 e1e3 g8f7 g1f1 a5a4 e3e1');
INSERT INTO evaluations VALUES(12395,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(12395,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(12395,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(12395,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(12395,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(12395,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(12395,36,'cp -34','d7d5 e4d5 e6d5 c4d5 f6d5 d3d4 c5d4 f3d4 f8b4 d4c6 b7c6 c1d2 d5c3 b2c3 b4c5 f1e2 e8g8 e1g1 c8f5 e2f3 d8f6 f1e1 a8d8 d1c1 f8e8 d2e3 c5e3 e1e3 g7g6 e3e8 d8e8 c1a3 h7h5 a3c5 e8e6 a1d1 f5g4 f3g4 h5g4 d1f1 e6e4 c5a7 f6c3 a2a4 g8g7 f2f3');
INSERT INTO evaluations VALUES(13538,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(13538,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(13538,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(13538,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(13538,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(13538,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(13538,36,'cp -34','e4d5 e6d5 c4d5 f6d5 d3d4 d8a5 d1b3 c5d4 b3d5 d4c3 d5a5 c6a5 f1b5 c8d7 b5d7 e8d7 e1g1 a8e8 f1d1 d7c8 b2c3 f8c5 c1f4 f7f6 d1d5 b7b6 a1d1 e8d8 g1f1 d8d5 d1d5 h8d8 d5d8 c8d8 f4e3 c5e3 f2e3 g7g6 f1e2 d8d7 f3d4 a5c4 e2d3');
INSERT INTO evaluations VALUES(31167,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(31167,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(31167,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31167,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31167,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(31167,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31167,36,'cp 0','e7e6 e2g3 f5g6 h2h4 h7h5 f1d3 g6d3 c2d3 d8b6 c1e3 b6b2 e1g1 g8e7 b1d2 b2a3 d2f3 e7g6 e3c1 a3a5 c1g5 b8d7 a2a4 a8c8 a1b1 b7b6 f3e1 a5c3 g3e2 c3a5');
INSERT INTO evaluations VALUES(31543,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31543,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31543,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31543,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31543,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31543,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31543,36,'cp 0','h2h4 h7h5 e2g3 f5g6 f1d3 g6d3 c2d3 d8b6 c1e3 b6b2 e1g1 g8e7 b1d2 b2a3 d2f3 e7g6 e3c1 a3a5 c1g5 b8d7 a1b1 b7b6 f1e1 a8c8 f3h2 a5c3 g3e2 c3a5');
INSERT INTO evaluations VALUES(41435,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41435,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41435,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41435,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41435,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41435,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41435,36,'cp 0','e2e3 e8g8 f1e2 h7h6 g5h4 c8e6 e1g1 b8d7 f3e5 d7e5 d4e5 f6e4 f2f4 e4c3 b2c3 d8d7 c3c4 c7c6 a1b1 a8c8 c4d5 c6d5 e2b5 d7c7 b5d3 c7d7');
INSERT INTO evaluations VALUES(41434,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41434,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41434,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41434,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41434,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41434,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41434,36,'cp 0','e8g8 f1e2 h7h6 g5h4 c8e6 e1g1 b8d7 a2a4 c7c5 d4c5 d7c5 a4a5 f8e8 e2b5 e8f8');
INSERT INTO evaluations VALUES(36717,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36717,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36717,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36717,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36717,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36717,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36717,36,'cp 0','f1e2 h7h6 g5h4 c8e6 e1g1 b8d7 a2a4 c7c5 d4c5 d7c5 a4a5 f8e8 e2b5 e8f8');
INSERT INTO evaluations VALUES(43831,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43831,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(43831,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(43831,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(43831,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(43831,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43831,36,'cp 28','b1c3 c8b7 d4d5 f8d6 g1f3 e8g8 c1g5 h7h6 g5f6 d8f6 g2g3 b8a6 f1g2 a6c5 e1g1 a7a5 d1c2 a5a4 f1e1 e6e5 e2e3 f6g6 e3e4 a8a5 a1d1 b7a6 f3d2 d6e7 h2h4 a6b7 g1h1 d7d6 g2h3 a5a8');
INSERT INTO evaluations VALUES(43829,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(43829,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43829,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(43829,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43829,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43829,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43829,36,'cp 37','c8b7 c1f4 d7d5 c4d5 f6d5 c3d5 d8d5 e2e3 c7c5 a1c1 f8e7 g1f3 e8g8 f1c4 d5d8 e1g1 e7f6 b2b4 c5d4 e3d4 b8c6 d4d5 e6d5 d1d5 d8d5 c4d5 c6d8 d5b7 d8b7 g2g3 b7d8 f1d1 f8e8 g1g2 f6b2 c1b1 b2f6 d1d6 d8e6 f4e5 e8d8 d6d8 a8d8 e5f6 g7f6 f3h4 d8d3 b1c1 d3a3 c1c8 e6f8 h4f5');
INSERT INTO evaluations VALUES(28724,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(28724,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(28724,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(28724,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28724,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(28724,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(28724,36,'cp 22','d4d5 f8d6 g1f3 e8g8 g2g3 c7c6 c1g5 c6d5 c4d5 b8a6 f1g2 e6d5 c3d5 b7d5 g5f6 d8f6 d1d5 f6b2 e1g1 d6e7 f3d4 a6c5 e2e3 a8c8 a1b1 b2c3 h2h4 f8d8 f1d1 g7g6 d4b5 c3a5 d1c1 h7h5 c1c4 g8g7 a3a4 a7a6 b5d6 e7d6 d5d6 c5e6 c4c8');
INSERT INTO evaluations VALUES(48758,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(48758,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(48758,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(48758,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(48758,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(48758,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(48758,36,'cp -10','d5e4 f1e2 e4d3 e5d3 f8e7 e1g1 b8c6 f1e1 e8g8 b1c3 h7h6 h2h3 f8e8 c1f4 c6d4 d3e5 d4e2 d1e2 c7c6 e2f3 c8e6 a2a4 e7b4 a1d1 f6d5 f4d2 d8f6 c3d5 f6f3 e5f3 b4d2 d5e7 e8e7 f3d2 b7b6');
INSERT INTO evaluations VALUES(50425,30,'cp 558',NULL);
INSERT INTO evaluations VALUES(50425,31,'cp 545',NULL);
INSERT INTO evaluations VALUES(50425,32,'cp 567',NULL);
INSERT INTO evaluations VALUES(50425,33,'cp 548',NULL);
INSERT INTO evaluations VALUES(50425,34,'cp 570',NULL);
INSERT INTO evaluations VALUES(50425,35,'cp 571',NULL);
INSERT INTO evaluations VALUES(50425,36,'cp 559','d1g4 d7d6 g4e4 g7g6 d2d4 f8g7 g1f3 e8g8 b1c3 c7c5 d4c5 b8c6 f1b5 d6e5 b5c6 b7c6 e1g1 c8f5 e4e2 d8c7 h2h3 a7a5 f1e1 a5a4 f3e5 g7e5 e2e5 c7e5 e1e5 f7f6 e5e7 f8e8 e7c7 a8c8 c7c8 e8c8 c3a4 f5c2 c1h6 c2a4 b2b3 c8a8 b3a4');
INSERT INTO evaluations VALUES(50424,30,'cp 543',NULL);
INSERT INTO evaluations VALUES(50424,31,'cp 527',NULL);
INSERT INTO evaluations VALUES(50424,32,'cp 540',NULL);
INSERT INTO evaluations VALUES(50424,33,'cp 532',NULL);
INSERT INTO evaluations VALUES(50424,34,'cp 542',NULL);
INSERT INTO evaluations VALUES(50424,35,'cp 559',NULL);
INSERT INTO evaluations VALUES(50424,36,'cp 537','d7d6 g4e4 g7g6 g1f3 f8g7 d2d4 e8g8 b1c3 c7c5 d4c5 b8c6 f1b5 d6e5 b5c6 c8f5 e4e2 b7c6 f3e5 d8c8 e1g1 c8e6 f1e1 e6e5 e2e5 g7e5 e1e5 f7f6 e5e1 f8e8 f2f3 a8d8 c3e4 a7a5 e4g3 e7e5 g3f5 g6f5 c1h6 g8f7 a1d1 f5f4 d1d6 f7g6 d6c6 g6h6 c6f6 h6g7 f6a6');
INSERT INTO evaluations VALUES(47072,30,'cp 567',NULL);
INSERT INTO evaluations VALUES(47072,31,'cp 567',NULL);
INSERT INTO evaluations VALUES(47072,32,'cp 565',NULL);
INSERT INTO evaluations VALUES(47072,33,'cp 575',NULL);
INSERT INTO evaluations VALUES(47072,34,'cp 586',NULL);
INSERT INTO evaluations VALUES(47072,35,'cp 575',NULL);
INSERT INTO evaluations VALUES(47072,36,'cp 559','g4e4 g7g6 d2d4 c8f5 e4b7 b8d7 f1b5 f8g7 b5d7 f5d7 g1f3 e8g8 b1c3 a8b8 b7e4 c7c5 d4c5 d6e5 f3e5 d7f5 e4e3 d8c7 e5d3 f8d8 e1g1 b8c8 f1e1 f5d3 c2d3 c7c5 e3e7 c5e7 e1e7 d8d3 c1e3 c8d8 a1e1 d3d7 e7d7 d8d7');
INSERT INTO evaluations VALUES(50966,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(50966,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(50966,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(50966,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(50966,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(50966,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(50966,36,'cp 62','g1f3 f8b4 e4d5 e6d5 f1d3 g8f6 e1g1 e8g8 c1g5 b4e7 f1e1 b8c6 h2h3 h7h6 g5h4 c6b4 d3f1 b4c6 a2a3 f6h5 h4e7 c6e7 d1d2 h5f6 e1e2 e7f5 a1e1 f5d6 f3e5 f8e8 g2g4 c8d7 f1g2 c7c6 f2f4 d8c7 a3a4');
INSERT INTO evaluations VALUES(51373,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(51373,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51373,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(51373,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(51373,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(51373,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(51373,36,'cp 49','a2a4 c7c6 a4b5 c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8e7 e1g1 e8g8 c1a3 c8b7 b1d2 a7a5 d1b1 a8a7 f1c1 b8d7 c4b5 h7h6 f3e1 e7a3 a1a3 e6e5 e1f3 d8e7 b1a1 e5e4 b5d7 e4f3 d7c6 f3g2 a3a5 a7a5 a1a5 f6g4 c6b7 e7b7 a5c3');
INSERT INTO evaluations VALUES(51377,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(51377,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(51377,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(51377,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(51377,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(51377,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(51377,36,'cp 38','c7c6 a4b5 c6b5 b2b3 g8f6 b3c4 b5c4 f1c4 f8d6 c1a3 e8g8 e1g1 c8b7 b1d2 a7a5 d1b3 a8a7 f1c1 d6a3 a1a3 b8d7 c4f1 d8e7 c1a1 f8a8 b3d1 b7f3 d2f3 g7g6 a1a2 f6d5 h2h3 d5b4 a2a1 d7f6');
INSERT INTO evaluations VALUES(53831,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(53831,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(53831,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53831,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(53831,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(53831,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(53831,36,'cp 26','f8e7 d3d4 b8d7 f1d3 g8f6 e1g1 e8g8 f1e1 f8e8 a2a4 e5d4 f3d4 d7c5 h2h3 a7a5 b2b3 e7f8 c1f4 d6d5 e4e5 f6d7 d3f5 g7g6 e5e6 d7f6 e6f7 g8f7 e1e8 d8e8 f5c8 e8c8 d1d2 c8d7 d4f3 f7g8 a1e1 f8g7 f4g5 c5e6 g5h6 g7h6 d2h6 d7f7 c3e2 f6e4');
INSERT INTO evaluations VALUES(55210,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(55210,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(55210,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55210,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(55210,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(55210,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(55210,36,'cp 22','c7c5 f1c4 b8c6 c4b5 c8d7 b1c3 e7e6 e1g1 d8b6 f1e1 g8f6 d2d3 f8e7 b5c4 e8g8 e4e5 f6e8 h2h4 a8d8 d1e2 e8c7 a2a3 h7h6 b2b4 c5b4 c1e3 e7c5 c3a4 c6d4 e3d4 d7a4 d4c5 b6c5 a3b4 c5c6 e1c1 b7b5 c4a2 c6b6');
INSERT INTO evaluations VALUES(56014,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(56014,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(56014,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(56014,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(56014,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56014,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56014,36,'cp -12','c7c6 a2a4 g8f6 e2e3 b7b5 a4b5 c6b5 c3b5 f8b4 b5c3 f6d5 d1c2 b4c3 b2c3 d5f4 e3f4 c8b7 g1f3 b7f3 g2f3 e8g8 f1c4 b8c6 f4f5 e6f5 e1g1 a8c8 c2f5 c6d4 c3d4 c8c4 a1a7 d8d4 a7e7 h7h6 f1e1 d4d2 e7e2 d2g5 f5g5 h6g5 g1g2 g7g6 e2e4 c4c8 e4e7 c8b8 e7a7');
INSERT INTO evaluations VALUES(6780,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(6780,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(6780,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(6780,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(6780,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(6780,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(6780,36,'cp 199','f4e5 c6e5 d2d4 e5f3 d1f3 d8h4 g2g3 h4f6 f3f6 g7f6 f1e2 c7c6 b1c3 h7h5 c1e3 d6f8 e1d2 d7d6 a1f1 c8h3 f1f4 f6f5 e2f1 h3f1 h1f1 g8e7 e4f5 f8h6 f4f3 d6d5 a2a3 e8d7 f5f6 e7g6 e3h6 h8h6 c3d1 h5h4 g3g4 b7b5 d1e3');
INSERT INTO evaluations VALUES(6762,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(6762,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(6762,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(6762,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(6762,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(6762,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(6762,36,'cp 175','c6e5 d2d4 e5f3 d1f3 d8h4 g2g3 h4f6 f3f6 g7f6 b1c3 c7c6 c1e3 h7h5 f1e2 d6c7 e1d2 f6f5 e4f5 d7d5 e2d3 c8d7 a1f1 e8c8 e3g5 d8e8 c3d1 c7d8 g5d8 c8d8 f1f4 g8f6 h1e1 d8c7 e1e8 h8e8 d1f2 c7d6 d3e2 e8g8 a2a4 g8g5 e2d3 g5g8 c2c4 b7b6 b2b4 c6c5 b4c5 b6c5 d4c5 d6c5 c4d5 c5d5');
INSERT INTO evaluations VALUES(6801,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(6801,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(6801,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(6801,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(6801,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(6801,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(6801,36,'cp 186','d2d4 e5f3 d1f3 d8h4 g2g3 h4f6 f3f6 g7f6 f1e2 c7c6 c1e3 h7h5 b1c3 d6c7 e1c1 b7b5 d1f1 d7d6 c3d1 c8h3 f1f2 h3g4 e2d3 a7a6 h2h3 g4e6 d1c3 d6d5 e4d5 c6d5 c3e2 e8d7 e3d2 a8e8 a2a4 b5a4 d3a6 e8b8 a6d3 d7d8 d2f4 h5h4 g3g4 c7f4 e2f4');
INSERT INTO evaluations VALUES(7593,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7593,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(7593,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7593,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(7593,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(7593,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(7593,36,'cp 0','d7d6 e1g1 e8g8 d1d3 c5b6 c1f4 f8e8 h2h3 c8e6 c4a6 a8b8 b2b3 f6h5 f4h2 d8f6 a1d1 h5f4 d3g3 f4h5');
INSERT INTO evaluations VALUES(7558,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7558,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(7558,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7558,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7558,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7558,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7558,36,'cp 16','e1g1 e8g8 c1g5 h7h6 g5h4 d8e7 d1d3 e7e5 a1e1 e5d4 h4f6 d4f6 c3a4 c5d4 c2c3 d4b6 a2a3 f8e8 d3c2 f6f4 a4b6 a7b6 e1e3 c8a6 c4a6 a8a6 g2g3 f4e5 c3c4 e5e6 f1e1 c6c5 g1g2 e6d7 e4e5 d7c6 f2f3 d6e5 e3e5');
INSERT INTO evaluations VALUES(18198,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(18198,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(18198,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(18198,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(18198,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(18198,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(18198,36,'cp 89','b1c3 e7e6 e4e5 g8e7 b2b3 e7g6 c3a4 f7f6 d1e2 d8c7 f1e1 f8e7 c1a3 f6e5 f3e5 c7e5 e2e5 g6e5 e1e5 e7f6 a3b2 c5c4 e5e1 a8b8 f2f4 f6b2 a4b2 e8e7 a1b1 c4b3 a2b3 c6c5 b2d3 h8f8 b1a1 e7d6 g2g3 b8b5 a1a4 c8b7 g1f2 a6a5 e1h1');
INSERT INTO evaluations VALUES(18255,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(18255,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(18255,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(18255,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(18255,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(18255,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(18255,36,'cp 117','c2c3 a7a5 f1b5 g8e7 d1e2 c8d7 e1g1 c6a7 b5d3 c7c5 b2b3 c5d4 c3d4 e7f5 b1a3 f8b4 a3c2 b4c3 c1g5 f5e7 a1c1 h7h6 g5e3 b6b5 f3h4 b5b4 f2f4 d8b6 g2g4 h6h5 f4f5 h5g4 f5e6 d7e6 h4g2 e8c8 g2f4 c8b7 f4e6 b6e6 e3g5 h8h5 h2h4');
INSERT INTO evaluations VALUES(37360,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(37360,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(37360,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(37360,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(37360,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(37360,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(37360,36,'cp 31','b1c3 h7h6 g5h4 d7d6 e2e3 c7c5 d4d5 g6g5 h4g3 f6h5 d1c2 a7a6 f1e2 f7f5 e1g1 d8e8 h2h3 b7b5 c4b5 g7c3 b2c3 a6b5 g3h2 h5f6 a2a3 f6e4 e2d3 e8f7 d3b5 f7d5 c3c4 d5e6 f3d2 c8b7 a3a4 e4d2 c2d2 b8c6');
INSERT INTO evaluations VALUES(41274,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(41274,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41274,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(41274,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(41274,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(41274,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41274,36,'cp 27','e8g8 f1g2 c7c5 e1g1 b8c6 d4d5 c6a5 f3d2 e7e6 a1b1 e6d5 c4d5 c8d7 b2b3 f8e8 e2e3 b7b5 c1b2 b5b4 c3e2 a5b7 a2a3 d7b5 f1e1 a7a5 a3a4 b5a6 b1c1 a8c8 h2h4 c5c4 d2c4 a6c4 c1c4 c8c4 b3c4 d8c7 e2d4');
INSERT INTO evaluations VALUES(43637,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(43637,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(43637,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(43637,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(43637,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(43637,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(43637,36,'cp 94','c1f4 f8e7 d1f3 c8g4 f3g3 d8d7 g3e3 b8c6 f2f3 g4e6 e1c1 e8g8 g2g4 c6d4 e3d4 a8d8 d4f2 c7c6 f4e3 b7b5 c1b1 d7b7 g4g5 f6h5 f3f4 f7f5 f1h3 e6d7 h3f5 d7f5 e4f5 f8f5 c3e4 d8f8 f2f3 g7g6 e4d6 e7d6 d1d6 h5f4 e3f4 f5f4 f3c6 b7c6 d6c6');
INSERT INTO evaluations VALUES(54389,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(54389,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(54389,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54389,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(54389,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(54389,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54389,36,'cp 28','e3d4 g8f6 c2c3 b8c6 f1d3 f6h5 f4e3 f8d6 e1g1 e8g8 f1e1 h5f4 e3f4 d6f4 g2g3 f4d6 b1d2 g7g6 a1c1 c8d7 b2b4 a7a6 a2a4 d8c7 b4b5 c6e7 c3c4 d5c4 d3c4 a8c8 d2b3 a6b5 c4e6 c7d8 c1c8 d8c8 e6d7');
INSERT INTO evaluations VALUES(54364,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(54364,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(54364,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54364,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(54364,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(54364,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(54364,36,'cp 38','g8f6 f1d3 d8b6 b1c3 c8d7 e1g1 f8e7 b2b3 e8g8 c3e2 d7c6 c2c3 a7a5 a2a4 b8d7 d1c2 f8c8 h2h3 f6e4 f3e5 d7e5 f4e5 h7h6 f2f3 e4f6 g1h1 c6e8 g2g4 f6d7 e5g3 e7d6 f3f4 b6c7 c3c4');
INSERT INTO evaluations VALUES(46344,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46344,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(46344,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(46344,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(46344,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(46344,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(46344,36,'cp 48','f1d3 d8b6 b1c3 c8d7 e1g1 a7a6 f1e1 f8e7 c3e2 d7b5 e2c1 e8g8 c2c3 b5c4 a1b1 c4d3 c1d3 a6a5 f3e5 b8c6 e5g4 f6e4 a2a4 e7d8 d1e2 a8c8 f2f3 e4f6 g2g3 f6g4 f3g4 d8c7 g1g2 c7f4 d3f4 b6c7 g4g5 c6e7 f4d3');
INSERT INTO evaluations VALUES(46340,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46340,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46340,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(46340,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46340,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46340,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46340,36,'cp 40','d8b6 b1c3 c8d7 e1g1 f8e7 b2b3 e8g8 c3e2 b8c6 a2a3 c6a5 f3e5 a8c8 a3a4 d7e8 e2g3 f6d7 e5f3 e7d6 f4d6 b6d6 f1e1 a5c6 c2c3 f7f6 c3c4 e8f7 a4a5 f8d8 c4c5 d6e7 d3f1 f7g6 d1d2 e6e5 b3b4');
INSERT INTO evaluations VALUES(57515,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(57515,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(57515,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(57515,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(57515,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(57515,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(57515,36,'cp 79','e4f5 d8e7 g1f3 d7d6 f4e5 d6e5 d2d4 e5d4 e1f2 e7f7 f1b5 c8d7 d1d4 f7f5 h1e1 e8d8 c1g5 g8f6 e1e5 f5c2 b1d2 b8c6 b5c6 c2c6 d2e4 d8c8 e4f6 g7f6 g5f6 c6f6 e5e8 d7e8 d4f6 f8c5 f2f1 e8b5 f1e1 c5b4 e1f2 b4c5 f2g3 c5d6 g3h4 h8f8 f6g5 b5d7 f3e5 d7e6 g5g7 a7a5 a1e1 a8a6 g7h7 a6b6 e5g6 b6b4 h4h5 e6f7 h5h6 f7g6 h7g6 b4b2 g6e6 c8b8');
INSERT INTO evaluations VALUES(57484,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(57484,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(57484,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(57484,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(57484,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(57484,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(57484,36,'cp 84','d8e7 f1e2 d7d6 f4e5 d6e5 e2h5 e8d8 d2d3 c8f5 h5f3 d8c8 b1d2 e7h4 g2g3 h4e7 f3g2 b8c6 g1f3 g8f6 e1g1 e7d7 g1h1 f8c5 f3g5 f5g4 d1e1 a7a5 d2c4 a8a6 c1e3 h7h6 e3c5 h6g5 e1e3 g4e6 c4e5 c6e5 e3e5 f6g4 e5g5');
INSERT INTO evaluations VALUES(35095,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(35095,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(35095,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(35095,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(35095,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(35095,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(35095,36,'cp -19','b8c6 b1c3 f6d8 c1g5 g8f6 g5f6 d8f6 c3d5 f6d8 g1f3 d7d6 h5g5 d8g5 f3g5 f8e7 h2h4 e7d8 c2c3 c6e7 d5e3 e8g8 g2g3 h7h6 g5f3 f7f5 f1g2 f5f4 g3f4 e5f4 e3c2 e7c6 h4h5 c6e5 d3d4 e5f3 g2f3 c7c6');
INSERT INTO evaluations VALUES(35862,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(35862,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(35862,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(35862,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(35862,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(35862,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(35862,36,'cp 63','c4d5 f6d5 f1e2 b8d7 e1g1 d7f6 a1c1 h7h6 g5h4 d5c3 b2c3 c6c5 h2h3 c8f5 a2a4 a8c8 d1b3 d8b6 b3a3 c5d4 c3d4 b6d6 a3b2 f6e4 c1c8 f8c8 b2b7 g6g5 g2g4 d6d7 b7d7 f5d7 e2d3 e4d6');
INSERT INTO evaluations VALUES(35857,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(35857,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(35857,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(35857,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(35857,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(35857,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(35857,36,'cp 69','f6d5 f1e2 h7h6 g5h4 d5c3 b2c3 c6c5 e1g1 c5d4 c3d4 b8c6 a1b1 d8d5 a2a4 b7b6 e2b5 c8b7 f1e1 d5e4 h4g3 f8d8 b5d3 e4e6 h2h4 e6g4 d1b3 a8c8 e1c1 c6a5 c1c8 d8c8 b3a3 g4d7');
INSERT INTO evaluations VALUES(50594,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(50594,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(50594,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(50594,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(50594,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(50594,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(50594,36,'cp -26','e7e6 c4a2 d7d5 e4d5 e6d5 b1c3 g8f6 d2d4 c5c4 c1g5 c8e6 d1f3 f8e7 g1e2 b8c6 h2h3 h7h6 g5f4 e6f5 e1c1 b7b5 f4e5 f5e6 e2f4 e8g8 h1e1 f8e8 e5f6 e7f6 c3d5 c6d4 f3e4 e6d5 e4e8 d8e8 e1e8 a8e8 f4d5 d4e2 c1b1');
INSERT INTO evaluations VALUES(53315,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53315,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53315,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(53315,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53315,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53315,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(53315,36,'cp 16','e2e4 e7e6 e4d5 e6d5 d3d4 b8c6 f1b5 f8d6 b1c3 g8f6 c1g5 e8g8 e1g1 c8e6 f1e1 h7h6 g5h4 a7a6 b5c6 b7c6 c3a4 c5d4 d1d4 f8e8 a4c5 d8e7 b2b4 a6a5 a2a3 a5b4 h4f6 e7f6 d4f6 g7f6 a3b4 d6c5 b4c5 a8a1');
INSERT INTO evaluations VALUES(53285,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53285,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(53285,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53285,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53285,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53285,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53285,36,'cp 0','e7e6 e4d5 e6d5 d3d4 a7a6 f1e2 g8f6 e1g1 b8c6 c2c4 f8e7 d4c5 e7c5 c4d5 d8d5 b1c3 d5d1 f1d1 e8g8 c1g5 c8e6 g5f6 g7f6 c3d5 g8g7 a1c1 c5a7 b2b3 f8d8 d5f4 d8d1 c1d1 a8d8 d1d8 c6d8 g1f1 f6f5 e2c4 g7f6 f4d5 f6g7');
INSERT INTO evaluations VALUES(15047,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(15047,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(15047,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(15047,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(15047,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(15047,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(15047,36,'cp 40','e1g1 f8e7 c2c3 e8g8 f1e1 f8e8 c1f4 f6h5 f4d2 e7d6 e1e8 d8e8 b1a3 h5f4 d2f4 d6f4 d1b3 e8d8 a1e1 g7g6 g2g3 f4h6 a3c2 c6a5 b3a4 c7c6 c2e3 c8e6 a4d1 b7b5 f3e5 d8b6 d1f3 a5c4 d3c4 b5c4 e3g4 h6g7 g4f6 g7f6 f3f6 a8e8');
INSERT INTO evaluations VALUES(15046,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(15046,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(15046,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(15046,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(15046,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(15046,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(15046,36,'cp 28','f8e7 c2c3 e8g8 f1e1 h7h6 f3e5 e7d6 b1d2 c6e7 d2f1 c7c5 d4c5 d6c5 c1f4 c5d6 h2h3 e7c6 f4g3 c6e5 g3e5 d6e5 e1e5 c8e6 f1g3 d8b6 d1d2 a8d8 a1d1 a7a5 b2b3 f8e8 d1e1 e6d7 e5e8 f6e8 g3e2 e8f6 e2d4 d8e8 e1e8 d7e8');
INSERT INTO evaluations VALUES(8519,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(8519,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(8519,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(8519,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8519,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(8519,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(8519,36,'cp 29','c2c3 e8g8 f1e1 h7h6 b1d2 e7d6 f3e5 c6e7 d2f1 c7c5 d4c5 d6c5 c1f4 c5d6 f4g3 e7c6 e5f3 d6g3 h2g3 c8g4 d1b3 d8c7 f3d4 f8e8 f1e3 g4e6 d4e6 f7e6 d3e2 a7a6 a1d1 a8c8 b3a4 c8d8 e2f3');
INSERT INTO evaluations VALUES(17806,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(17806,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(17806,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(17806,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(17806,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(17806,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(17806,36,'cp -13','b7b5 a4b3 c6a5 d2d4 e5d4 d1d4 c8b7 c2c4 g8f6 b1c3 f8e7 e1g1 e8g8 c1e3 f8e8 f1e1 h7h6 f3d2 e7f8 f2f3 b5c4 d2c4 a5b3 a2b3 d6d5 c4a5 a8b8 e4d5 b7d5 c3d5 d8d5 d4d5 f6d5 e3d2 e8e1 d2e1 b8b5 g1f1 f7f6 a5c4 b5b3 a1a6 f8b4 e1b4 b3b4');
INSERT INTO evaluations VALUES(17986,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(17986,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(17986,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(17986,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(17986,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(17986,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17986,36,'cp -10','a4b3 c6a5 d2d4 e5d4 d1d4 c8b7 c2c4 g8f6 b1c3 f8e7 e1g1 e8g8 c1e3 f8e8 f1d1 h7h6 f3d2 e7f8 c3d5 f6d7 d4c3 c7c6 d5f4 c6c5 c4b5 a6b5 b3d5 b7d5 f4d5 d7b6 c3d3 b6d5 e4d5 d8f6 a2a4 b5b4 d1e1 f8e7 d2c4 a5c4');
INSERT INTO evaluations VALUES(17976,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(17976,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(17976,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(17976,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(17976,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(17976,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(17976,36,'cp -9','c6a5 d2d4 e5d4 d1d4 c8b7 c2c4 g8f6 b1c3 f8e7 d4d3 e8g8 e1g1 f8e8 f1e1 h7h6 c1f4 f6d7 c4b5 d7c5 d3c2 a5b3 a2b3 a6b5 b3b4 c5e6 f4g3 a8a1 e1a1 d8d7 c2d3 e8a8 a1a8 b7a8 f3d4 e6d4 d3d4 a8c6 d4a7 e7f8 c3d5 c6d5 e4d5 f8e7');
INSERT INTO evaluations VALUES(20928,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(20928,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(20928,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(20928,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(20928,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(20928,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(20928,36,'cp 97','c2c4 g8f6 b1c3 d5c4 f1c4 e7e6 a2a3 f8d6 f4g3 d6g3 h2g3 e6e5 d4d5 c6b8 g1f3 b8d7 e3e4 e8g8 e1g1 d8e7 f1e1 d7b6 d1b3 f8d8 c4b5 c8g4 a1c1 a8c8 c3a4 g4d7 a4b6 a7b6 b5d3 c7c6 b3b6 c6d5 c1c8 d7c8 e4d5 e5e4 d3c2');
INSERT INTO evaluations VALUES(23075,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(23075,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(23075,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(23075,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(23075,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(23075,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(23075,36,'cp 143','e5g6 b1c3 f8b4 d4f5 d8f6 c1d2 g8e7 c3b5 f6c6 f5g7 e8f8 c2c3 b4a5 g7h5 c6e4 d1e2 d7d5 e1c1 c8f5 e2e4 f5e4 c3c4 a5d2 d1d2 g6h4 h5f6 c7c6 f6e4 d5e4 b5d6 e4e3 d2d3 e7f5 d6f5 h4f5 g2g3 f8g7 f1h3 g7f6 h1e1 a8e8 c1c2 h7h5 d3d7 e3e2 h3f5 f6f5 d7b7 f7f6');
INSERT INTO evaluations VALUES(22997,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(22997,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(22997,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(22997,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(22997,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(22997,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(22997,36,'cp 187','b1c3 f8b4 d4f5 d8f6 f1d3 d7d6 c1d2 c8f5 e4f5 g6e7 a2a3 b4a5 c3b5 a5d2 d1d2 e8c8 d2a5 e7d5 e1c1 a7a6 g2g4 c8b8 g4g5 f6e7 h1e1 e7d7 b5c3 d5b6 c3d5 b6d5 a5d5 h7h6 h2h4 h6g5 h4g5 f7f6 d5b3 d6d5 d3a6 c7c6 e1e6 g8e7 b3b7 d7b7 a6b7');
INSERT INTO evaluations VALUES(22996,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(22996,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(22996,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(22996,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(22996,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(22996,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(22996,36,'cp 161','f8b4 d4f5 d8f6 f1d3 d7d6 c1d2 c8f5 e4f5 g6e7 a2a3 b4a5 c3b5 a5d2 d1d2 e8c8 d2a5 e7d5 b5a7 c8b8 a5d5 d8e8 e1d2 b8a7 a1e1 g8e7 d5a5 a7b8 e1e4 e7c6 a5c3 f6c3 d2c3 e8e4 d3e4 c6e7 g2g4 c7c6 h1d1 f7f6 d1d6 b8c7 d6d1 h8e8 d1e1 e7d5 c3d4 d5f4');
INSERT INTO evaluations VALUES(23298,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(23298,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(23298,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(23298,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(23298,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(23298,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(23298,36,'cp 214','f3e5 f5e4 e5c6 b7c6 d1h5 e8e7 c1g5 g8f6 b1d2 d8e8 h5h4 e7d8 e1c1 f8e7 d2e4 f6e4 g5e7 e8e7 h4e7 d8e7 d1e1 d7d5 f2f3 c8f5 f1d3 e7d6 f3e4 f5e4 d3e4 d5e4 e1e4 h8e8 h1e1 e8e4 e1e4 a8f8 c1d2 f8f2 e4e2 f2f5 d2d3 c6c5 d4c5 f5c5 e2e4 c5f5 g2g3 f5f2 e4e2 f2f5 d3d4 h7h5 c2c3 f5g5 b2b3 g5g4 d4d3 a7a5 e2e4');
INSERT INTO evaluations VALUES(38672,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38672,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38672,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(38672,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(38672,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38672,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(38672,36,'cp 30','c2c4 b7b6 a2a3 g8f6 b2b4 c7c6 b1d2 b8d7 c4c5 d6e7 c1b2 a7a5 f1e2 e8g8 e1g1 c8a6 e2a6 a8a6 b2c3 d8a8 d1c2 a8b7 f1b1 a5b4 a3b4 f8a8 a1a6 b7a6 b1a1 a6b7 c2b2 h7h6 h2h3 a8a6 a1a6 b7a6 b2c2 e7f8 d2b3 f8e7 c5b6 d7b6 b3c5 a6b5 c2a2');
INSERT INTO evaluations VALUES(38656,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(38656,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(38656,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(38656,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(38656,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38656,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38656,36,'cp 15','b7b6 a2a3 g8f6 b2b4 c7c6 b1d2 a7a5 c4c5 d6e7 c1b2 e8g8 f1e2 c8a6 e1g1 b8d7 e2a6 a8a6 d1c2 d8c7 f1b1 f8a8 f3e1 f6g4 d2f3 g4f6 e1d3 f6e4 c5b6 a6b6 d3c5 a5b4 a3b4 a8a1 b1a1 e7c5 b4c5 b6b8 b2c3 c7b7 f3d2 d7f6 d2e4 d5e4');
INSERT INTO evaluations VALUES(42000,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42000,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42000,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42000,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42000,36,'cp 0','d5c4 e2e4 c7c5 g5f6 d8f6 f1c4 c5d4 c3d4 e8g8 g1f3 b8c6 d1d3 f8d8 a1d1 c6d4 f3d4 e6e5 c4d5 e5d4 d3d4 f6d4 d1d4 c8e6 e1e2 e6d5 d4d5 d8d5 e4d5 g8f8 h1b1 b7b6 b1c1 a8d8 c1c7 d8d5 c7a7 d5b5 a7a8 f8e7 h2h4 b5b2 e2e3 e7f6 a8a7 b2b3 e3f4 b3b2');
INSERT INTO evaluations VALUES(43025,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(43025,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(43025,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(43025,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(43025,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(43025,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(43025,36,'cp -25','c2c4 h7h6 g5e3 c7c6 g1f3 f6g4 b1c3 g4e3 f2e3 d7d6 f1e2 c8g4 e1g1 e8g8 d1b3 d8c7 d3d4 b8d7 a1d1 g4h5 b3c2 a8e8 b2b4 d7f6 h2h3 h5f3 e2f3 d6d5 e4d5 e7b4 d5c6 b7c6 g1h1 b4c3 c2c3 e5e4');
INSERT INTO evaluations VALUES(52276,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(52276,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52276,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(52276,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(52276,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52276,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52276,36,'cp -17','b8c6 d2d3 f8g7 c1e3 d7d6 g2g3 a7a6 f1g2 b7b5 e4e5 c8b7 e5d6 e7d6 g1f3 b5b4 c3d5 g8e7 e1g1 e7d5 c4d5 c6e7 d3d4 e8g8 d4c5 e7d5 e3d4 d6c5 d4c5 f8e8 c5d4 a8c8 a2a3 b4a3 a1a3 d5f6 d4e3 f6g4 e3g5 d8d1 f1d1');
INSERT INTO evaluations VALUES(54004,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(54004,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(54004,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(54004,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(54004,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(54004,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(54004,36,'cp 36','g8f6 d2d4 c5d4 f3d4 a7a6 f1e2 f8e7 c1e3 e6e5 f4e5 d6e5 d4f5 c8f5 d1d8 e7d8 e4f5 d8a5 e1c1 a5c3 b2c3 b8d7 h1e1 h7h6 h2h3 f6e4 e2f3 e4c3 d1d3 c3a2 c1d2 a2b4 d3b3 b4c6 b3b7 e8c8 b7b1 h8e8 d2c1 c8c7 f3e2 c6d4 e2a6 d4f5');
INSERT INTO evaluations VALUES(46025,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46025,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(46025,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(46025,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(46025,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46025,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(46025,36,'cp 49','d2d4 c5d4 f3d4 a7a6 d1f3 e6e5 f4e5 d6e5 d4f5 d8c7 f5e3 c8e6 f1d3 b8c6 e1g1 a8d8 e3f5 e6f5 f3f5 c6b4 g1h1 h7h6 f5h3 b4d3 c2d3 c7d6 h3g3 f6h5 g3f2 d6e6 c3d5 f8d6 c1e3 e8g8 a1d1 d8d7 d1c1 e6g6 f2f5 g6f5 f1f5');
INSERT INTO evaluations VALUES(55575,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(55575,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(55575,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(55575,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(55575,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(55575,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(55575,36,'cp 54','g1f3 b8d7 h2h3 c7c6 e2e4 f8e7 a2a4 g8f6 f1c4 e8g8 e1g1 f8e8 f1e1 h7h6 a4a5 e5d4 f3d4 d7c5 c1f4 e7f8 e4e5 d6e5 f4e5 c5d7 e5f4 e8e1 d1e1 d7b6 c4b3 b6d5 c3d5 f6d5 f4e5 c8d7 a5a6 b7a6 b3d5 c6d5 a1a6 d8g5 d4f3 g5h5');
INSERT INTO evaluations VALUES(4356,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4356,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4356,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4356,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4356,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4356,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4356,36,'cp 0','d7d6 h2h3 h7h6 b1c3 a7a6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 f1f2 d8e7 d1e2 c6b8 f3h4 g8h7 h4f3 h7g8');
INSERT INTO evaluations VALUES(12895,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(12895,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(12895,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(12895,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(12895,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(12895,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(12895,36,'cp -23','d7d5 e4d5 d8d5 g1f3 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 e7e6 c1e3 a7a6 e1e2 a8c8 b1d2 c6b4 c2c3 b4d3 e2d3 f8e7 f3f4 e7c5 d3e2 e8e7 a2a4 c5e3 f2e3 c8c5 h3h4 g7g6 d2b3 c5c4 b3d2 c4c8 a4a5 c8c5 d2b3');
INSERT INTO evaluations VALUES(16742,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(16742,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(16742,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(16742,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(16742,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(16742,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(16742,36,'cp 95','g1e2 d7d6 d2d3 c8e6 c1h6 g7h6 c4e6 f7e6 f3h3 d8f6 f2f4 h6h5 f4f5 e6f5 b1c3 f6g5 c3d5 e8d8 e4f5 f8h6 e1g1 c6e7 d5c3 g5g4 h3f3 g4f3 f1f3 d8d7 f5f6 e7g6 a1f1 a8f8 c3e4 f8f7 c2c4 h8g8 g1h1 g6h4 f3g3 g8g4 e2c3 g4f4 f1f4');
INSERT INTO evaluations VALUES(51978,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(51978,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(51978,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(51978,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(51978,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(51978,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(51978,36,'cp 0','d7d5 e4d5 e6d5 f1b5 b8c6 g1f3 d8e7 b5e2 e7c7 d2d4 c8e6 e1g1 g8f6 b1c3 a7a6 c3a4 c5d4 f3d4 f8d6 d4c6 c7c6 c2c4 d5c4 b2f6 g7f6 e2c4 e6c4 b3c4 e8c8 d1b3 d6c7 a1b1 h8e8 g2g3 d8d4 a4b6 c7b6 b3b6 c6b6 b1b6 e8e6 f1b1 d4d1 g1g2 d1b1 b6b1 c8c7 g2f3 e6e5 b1b3');
INSERT INTO evaluations VALUES(54244,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(54244,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(54244,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54244,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(54244,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54244,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(54244,36,'cp 18','e4d5 e6d5 f1b5 b8c6 g1f3 d8e7 b5e2 g8f6 e1g1 e7c7 d2d4 c5d4 f3d4 f8e7 e2f3 e8g8 b1c3 f8d8 f1e1 e7c5 d4c6 b7c6 c3a4 c5d6 b2f6 g7f6 g2g3 c8f5 d1d2 f5g6 g1g2 h7h5 a1d1 g8g7 h2h4 a7a5 a4b2 d6b4 c2c3 b4d6 b2d3 d8e8 e1e2 a5a4 b3b4 a4a3 e2e8 a8e8 d3f4 d6f4');
INSERT INTO evaluations VALUES(54139,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(54139,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(54139,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54139,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54139,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54139,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(54139,36,'cp 21','e6d5 f1b5 b8c6 g1f3 d8e7 b5e2 g8f6 e1g1 e7d8 d2d4 f8e7 c2c4 e8g8 b1c3 c5d4 c3d5 f6d5 c4d5 d8d5 f3d4 c6d4 d1d4 d5d4 b2d4 a7a5 e2c4 c8e6 c4e6 f7e6 a2a4 a8d8 a1d1 f8f5 d4b6 d8d5 g2g3 e7f6 d1d5 f5d5 f1c1 d5d3 c1c7 d3b3');
INSERT INTO evaluations VALUES(54543,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(54543,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54543,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(54543,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(54543,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(54543,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(54543,36,'cp 8','f1b5 b8c6 g1f3 d8e7 b5e2 e7d8 d2d4 g8f6 e1g1 f8e7 c2c4 e8g8 b1c3 c6d4 f3d4 c5d4 c3d5 f6d5 c4d5 d8d5 d1d4 d5d4 b2d4 a7a5 e2c4 c8e6 c4e6 f7e6 g2g3 a5a4 a1e1 a4b3 a2b3 g8f7 e1e4 a8d8 f1c1 d8d7 c1e1 f8e8 e4g4 e7f6 d4f6 g7f6 g4h4');
INSERT INTO evaluations VALUES(14609,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(14609,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14609,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14609,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14609,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14609,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14609,36,'cp 0','g7g5 d2d4 g5g4 h3g4 c8g4 d4e5 c6e5 c4e2 g4f3 e2f3 h6h5 c1g5 h8g8 g5h4 c7c6 f3h5 d8d7 h4f6 d7h3 g2g3 g8g3 f2g3 h3g3 g1h1 g3h3 h1g1');
INSERT INTO evaluations VALUES(21244,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(21244,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(21244,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(21244,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(21244,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(21244,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(21244,36,'cp 148','h2h4 c8d7 f1d3 d8e7 e1g1 h7h6 c2c3 g6g5 b2b4 g5g4 f3h2 e7h4 h2g4 f7f5 g4h2 h6h5 b1d2 h4e7 d2b3 e7g7 b4b5 c6d8 a2a4 d8f7 g1h1 g8h6 d1e2 e8c8 a4a5 f8e7 c3c4 h6g4 c1f4 d8g8 f2f3 e7g5 b5b6 a7a6 f3g4 h5g4 c4d5 g5f4 f1f4 g7h6 g2g3 e6d5');
INSERT INTO evaluations VALUES(23223,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(23223,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(23223,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(23223,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(23223,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(23223,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(23223,36,'cp 119','f8c5 b1c3 d7d6 c3a4 c5b6 a4b6 a7b6 d2d4 c8g4 d4d5 c6d4 c1e3 b6b5 c4f1 g8e7 e3d4 e5d4 f1b5 c7c6 b5e2 d8a5 d1d2 c6d5 e4d5 e7d5 f3d4 g4e2 e1e2 a5d2 e2d2 d5f4 g2g3 f4g6 b2b3 e8c8 d2c3 h8e8 a2a4 g6e7 h1e1 e7d5 c3b2 e8e5 a4a5 d5e3 b3b4 c8b8');
INSERT INTO evaluations VALUES(6928,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(6928,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(6928,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(6928,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(6928,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(6928,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(6928,36,'cp 101','b1c3 d7d6 c3a4 d8e7 d2d3 c8e6 c4b5 e6d7 a4c5 d6c5 e1g1 e8c8 d1e1 c6d4 f3d4 e5d4 b5d7 d8d7 f4f5 h7h5 c1f4 g8h6 e1g3 d7d8 a1e1 d8e8 c2c3 e8d8 c3c4 h6g4 e4e5 f6e5 f4e5 g4e3 e5g7 h8g8 f5f6 e7f7 e1e3 d4e3 g3e3');
INSERT INTO evaluations VALUES(43210,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(43210,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43210,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43210,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43210,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43210,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43210,36,'cp 0','e8g8 c2c3 d7d6 g1f3 b8d7 h2h3 d8e8 f1e2 e7e5 d4e5 d7e5 e1g1 e5f3 e2f3 f6d7 a2a4 d7c5 a4a5 e8b5 d1c2 b5d3 c2d1');
INSERT INTO evaluations VALUES(37433,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37433,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37433,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(37433,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37433,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37433,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37433,36,'cp 0','g1f3 d7d6 c2c3 d8e8 h2h3 b8d7 f1e2 e7e5 d4e5 d6e5 f4g3 b7b6 e1g1 e8e7 d1c2 c8b7 a2a4 a7a5 e2b5 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(47848,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(47848,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(47848,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(47848,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(47848,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(47848,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(47848,36,'cp -18','e5d4 d1d4 f8e7 f1e2 e8g8 e1g1 f8e8 c1f4 b8d7 f1e1 d7c5 e2c4 c8e6 c4f1 e6g4 b1d2 c5e6 d4e3 f6h5 f4g3 a7a5 e4e5 h5g3 h2g3 d6d5 f3d4 e6d4 e3d4 d8d7 f1e2 g4e6 a2a4 c7c6 d2f3 g7g6 d4e3 d7c7');
INSERT INTO evaluations VALUES(48665,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48665,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(48665,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(48665,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(48665,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48665,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(48665,36,'cp 0','b8c6 c1e3 c6b4 g1f3 f8d6 e1g1 b4d3 d1d3 h7h6 c3b5 e8g8 b5d6 c7d6 f1e1 d8b6 a2a4 c8d7 a4a5 b6c7 f3d2 a8c8 c2c3 f8e8 f2f3 a7a6 d2f1 d7b5 d3f5 b5d7 f5f4 d7b5');
INSERT INTO evaluations VALUES(51476,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51476,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51476,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(51476,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(51476,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(51476,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(51476,36,'cp 16','c7c5 c2c3 c8b7 d2d4 e7e6 g1f3 b8c6 f1d3 c5d4 e1g1 g8e7 c3d4 c6b4 d3e2 e7g6 g2g3 f7f6 b1c3 f8e7 a2a3 b4d5 c3d5 b7d5 h2h4 e8g8 h4h5 g6h8 f3h2 f6e5 d4e5 d8c7 e2f3 c7e5 f3d5 e5d5 d1d5 e6d5 f1e1');
INSERT INTO evaluations VALUES(52889,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(52889,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(52889,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(52889,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(52889,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(52889,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(52889,36,'cp 138','f1c4 g8f6 b1c3 b7b5 c4d3 d8d4 g1f3 d4d8 e4e5 b5b4 e5f6 b4c3 b2c3 c8g4 e1g1 g7f6 f1e1 d8d6 e1e4 h8g8 c1f4 e7e5 f3e5 g4d1 e5g4 e8d8 f4d6 d1f3 g2f3 f8d6 e4d4 d8c7 g1f1 d6c5 d4f4 b8d7 g4f6 g8g7 f6g4 d7b6 f4f6 a8f8 a1e1');
INSERT INTO evaluations VALUES(52873,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(52873,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(52873,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(52873,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(52873,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(52873,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(52873,36,'cp 110','g8f6 b1c3 b7b5 c4d3 d8d4 g1f3 d4d8 e4e5 b5b4 e5f6 b4c3 b2c3 c8g4 h2h3 g4h5 e1g1 g7f6 f1e1 d8d6 a1b1 b8d7 d3e4 d6d1 e1d1 e8c8 e4c6 d7b6 d1d8 c8d8 f3d4 d8c7 b1b5 e7e5 c6e4 a7a6 b5a5 b6c4 a5a6 c4d6 e4f3 h5f3 d4f3 c7b7 a6a4 f8e7 c1e3 d6f5 a4a7 b7c6 g2g4 f5e3');
INSERT INTO evaluations VALUES(45010,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(45010,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(45010,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(45010,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(45010,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(45010,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(45010,36,'cp 147','b1c3 b7b5 c4d3 d8d4 g1f3 d4d8 e4e5 b5b4 e5f6 b4c3 b2c3 e7f6 e1g1 f8c5 d1c2 c8e6 d3h7 b8d7 f1e1 d8c7 h7f5 e8c8 f5e6 f7e6 e1e6 d8e8 e6e8 h8e8 c1d2 d7e5 f3d4 c5b6 a1d1 e5c4 d2c1 c7e5 h2h3 c8b7 c2b3 e5d5 d4f3 d5c5 d1d4 c4e5 f3e5 e8e5 c1f4 e5e2 f4g3');
INSERT INTO evaluations VALUES(45009,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(45009,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(45009,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(45009,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(45009,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(45009,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(45009,36,'cp 135','b7b5 c4d3 d8d4 g1f3 d4d8 e4e5 b5b4 e5f6 b4c3 b2c3 e7f6 e1g1 f8e7 f1e1 e8g8 f3d4 e7d6 d1h5 g7g6 h5h4 f8e8 c1h6 d6f8 e1e8 d8e8 h6f8 e8f8 h4f6 b8d7 f6c6 a8b8 a1e1 b8b6 c6c7 g8g7 d3c2 f8c5 c7g3 d7f6 g3e5 c5e5 e1e5 c8e6 c2b3 e6b3');
INSERT INTO evaluations VALUES(43936,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(43936,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(43936,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(43936,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(43936,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(43936,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(43936,36,'cp 150','c4d3 d8d4 g1f3 d4d8 e4e5 b5b4 e5f6 b4c3 b2c3 e7f6 e1g1 f8c5 d1c2 c8e6 f1e1 b8d7 d3h7 e8f8 c1f4 d7e5 f3e5 f6e5 e1e5 c5d6 a1d1 d6e5 d1d8 a8d8 f4e5 h8h7 h2h3 h7h5 c2e2 h5h6 c3c4 f8g8 g1h2 d8e8 e5f4 h6f6 e2e5 f6f5 e5e3 e8f8');
INSERT INTO evaluations VALUES(53160,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53160,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53160,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(53160,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(53160,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(53160,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(53160,36,'cp 52','b2b4 a7a5 b4b5 e7e5 g1f3 b8d7 e2e3 f8c5 e3d4 e5d4 f1e2 g8f6 e1g1 e8g8 f3d4 f6e4 d4b3 c5b6 c1e3 b6e3 f2e3 a5a4 b3d2 e4f6 b1c3 d8b6 c4c5 b6c5 d2c4 d7e5 d1d6 c5d6 c4d6 c8g4 d6b7 f8c8 e2g4 c8c3 g4f3 e5f3 g2f3 f6d5 e3e4 d5f4 g1h1 f4d3 a1d1 d3f4 b7d6');
INSERT INTO evaluations VALUES(57326,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(57326,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(57326,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(57326,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(57326,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(57326,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57326,36,'cp 28','e2e3 g8f6 b1c3 d7d5 d2d3 f8d6 d3e4 d5e4 g1e2 e8g8 e2g3 d6e5 d1d8 f8d8 a1d1 d8e8 e1d2 c8g4 f1e2 g4e2 g3e2 c7c6 h2h3 b8d7 d2c1 a8d8 g2g4 h7h6 a2a3 d7c5 a3a4 g7g6 d1d8 e8d8 h1d1 f6d7');
INSERT INTO evaluations VALUES(57325,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(57325,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(57325,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(57325,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(57325,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(57325,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(57325,36,'cp 8','g8f6 d2d3 d7d5 b1c3 f8d6 d3e4 d5e4 g1e2 e8g8 d1d2 d8e7 e1c1 b8c6 h2h3 f8d8 c1b1 d6e5 d2c1 d8d1 c1d1 c8e6 e2f4 a8d8 d1c1 h7h5 f1e2 e5f4 e3f4 c6d4 c1e3 c7c5 g2g4 h5g4 h3g4 e6g4 e2g4 f6g4 e3e4 e7e4 c3e4 b7b6');
INSERT INTO evaluations VALUES(17054,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(17054,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(17054,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(17054,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(17054,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(17054,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(17054,36,'cp 92','f3e5 a5c4 e5c4 d7d5 e4d5 f6d5 d1f3 c8e6 e1g1 f8e7 c1d2 e8g8 b1c3 d5c3 d2c3 b7b5 f3g3 f7f6 f1e1 e6f7 c4e3 e7d6 g3f3 b5b4 c3d2 d8d7 e3f5 a8e8 f5d6 d7d6 f3f5 c7c5 b2b3 a7a5 h2h3 d6c6 a2a4 h7h6 f5g4 g8h8 g1f1');
INSERT INTO evaluations VALUES(17047,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(17047,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(17047,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(17047,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(17047,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(17047,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(17047,36,'cp 90','a5c4 e5c4 d7d5 e4d5 f8e7 b1c3 f6d5 c1d2 c8e6 e1g1 e8g8 d1f3 c7c6 c4e3 d5e3 d2e3 f8e8 f1e1 d8d7 h2h3 e7f8 a2a3 b7b6 a3a4 f8b4 e3d2 a8d8 e1e4 c6c5 a1e1 f7f6 e1e3 e6f7 e4e8 d8e8 c3e4');
INSERT INTO evaluations VALUES(16940,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(16940,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(16940,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(16940,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(16940,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(16940,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(16940,36,'cp 95','e5c4 d7d5 e4d5 f6d5 e1g1 f8e7 c1d2 c8e6 b1c3 e8g8 d1f3 c7c5 c4e3 d5e3 d2e3 b7b6 f1e1 e7f6 h2h3 f8e8 f3g3 f6h4 g3f4 e6d7 e3d2 d8f6 c3e4 f6f4 d2f4 e8e6 g1f1 a8e8 e4c3 e6e1 a1e1 e8e1 f1e1 d7c6');
INSERT INTO evaluations VALUES(17032,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(17032,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(17032,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(17032,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(17032,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(17032,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(17032,36,'cp 96','d7d5 e4d5 f8e7 b1c3 f6d5 c1d2 e8g8 d1f3 c8e6 e1g1 c7c5 a2a3 f8e8 h2h3 d5c3 d2c3 b7b5 c4e3 a7a5 f1d1 b5b4 c3e5 f7f6 e5g3 d8d4 b2b3 e7f8 a3a4 e6f7 d1e1 a8a6 g1f1 h7h5 a1d1 h5h4 g3h2 d4d7 e3c4 a6e6 e1e6');
INSERT INTO evaluations VALUES(22649,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(22649,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(22649,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(22649,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(22649,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(22649,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(22649,36,'cp -69','e5e4 f3d4 d8g5 b5f1 d7d5 d4c6 b7c6 d2d4 g8f6 c2c4 d5c4 b1d2 f8d6 d2c4 e8g8 h2h4 g5f5 f1e2 f5g6 e1f1 h7h5 c1d2 c8g4 b2b4 f8e8 a1c1 e8e6 f1g1 a8e8 d2e1 f6d5 e2g4 h5g4 g2g3 g6f5 d1c2 d6b4 e1b4 d5b4');
INSERT INTO evaluations VALUES(22652,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(22652,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(22652,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(22652,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(22652,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(22652,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(22652,36,'cp -76','f3d4 d8g5 b5f1 d7d5 d4c6 b7c6 d2d4 c8g4 d1d2 a7a5 b2b3 f7f5 c2c4 g8f6 c1a3 f8a3 b1a3 e8g8 c4d5 c6d5 h2h3 g5h4 g2g3 h4h6 h1g1 g4f3 a1c1 f8c8 a3b5 f6e8 b5c3 c7c6 c3a4 c8c7 h3h4 e8f6 f1e2 f6g4 e2f3 e4f3 e1d1 g8h8');
INSERT INTO evaluations VALUES(29875,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(29875,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(29875,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(29875,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(29875,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(29875,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(29875,36,'cp 95','f2f3 g4c8 c1e3 f8e7 h2h4 a7a6 d1d2 h7h5 g1h3 c7c6 h3g5 b8d7 a2a4 e8g8 f1c4 d7b6 c4d3 c6d5 e4d5 d8c7 a4a5 b6c4 d3c4 c7c4 b2b3 c4c7 e3b6 c7b8 c3a4 c8f5 c2c4 f5g6 e1g1 f8c8 b6e3 f6d7 f3f4 e5f4 f1f4 d7e5');
INSERT INTO evaluations VALUES(29873,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(29873,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(29873,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(29873,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(29873,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(29873,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(29873,36,'cp 109','g4c8 c1e3 f8e7 h2h4 f6d7 e3f2 h7h5 g1e2 c7c6 d1d2 d7f6 e2g3 g7g6 f1e2 b8d7 a2a4 a7a6 a4a5 c6d5 e4d5 e8g8 c3a4 a8b8 a4b6 d7b6 f2b6 d8e8 c2c4 c8d7 b2b4 e7d8 b6e3 b8c8 e1g1 e8e7 e2d3 f8e8 g3e4 f6e4');
INSERT INTO evaluations VALUES(32037,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(32037,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(32037,32,'cp 278',NULL);
INSERT INTO evaluations VALUES(32037,33,'cp 263',NULL);
INSERT INTO evaluations VALUES(32037,34,'cp 289',NULL);
INSERT INTO evaluations VALUES(32037,35,'cp 272',NULL);
INSERT INTO evaluations VALUES(32037,36,'cp 264','f3e5 d6e5 d1g4 b8c6 g4g3 h6h5 c4b5 d8d6 b5c6 b7c6 b2b3 e8c8 g3e3 c8b7 c1b2 g8f6 e1c1 d6c5 e3f3 c5a5 c3a4 f8b4 c2c3 b4d6 c1b1 f6d7 b2c1 d7b6 a4b2 d6a3 d2d4 f7f6 h2h3 e5d4 d1d4 a3b2 c1b2 a5c5 h1d1 d8d4 c3d4 c5b4');
INSERT INTO evaluations VALUES(32027,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(32027,31,'cp 286',NULL);
INSERT INTO evaluations VALUES(32027,32,'cp 290',NULL);
INSERT INTO evaluations VALUES(32027,33,'cp 266',NULL);
INSERT INTO evaluations VALUES(32027,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(32027,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(32027,36,'cp 267','d6e5 d1g4 b8c6 d2d3 h6h5 g4f5 d8f6 f5f6 g8f6 a2a4 h5h4 h2h3 a7a6 c1e3 a8d8 h1f1 c6d4 e1d1 f6h5 c3d5 c7c6 d5b6 h5f4 c2c3 f8c5 e3f4 e5f4 a4a5 d4b5 c4b5 a6b5 d1e2 f7f5 e4e5 f4f3 g2f3 e8f7 d3d4 c5b6 a5b6 f5f4');
INSERT INTO evaluations VALUES(32482,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(32482,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(32482,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(32482,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(32482,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(32482,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(32482,36,'cp 63','g8f6 e1g1 b8c6 b1c3 f8e7 f1e1 c6d4 d1d4 e8g8 c1f4 a7a6 a2a4 d7c6 a4a5 f6h5 f4c1 h5f6 f2f3 f6d7 c1e3 e7h4 e1f1 f8e8 d4d2 h4f6 a1d1 h7h6 b2b3 d7e5 c4d5 e5g6 g2g3 f6c3 d2c3 c6d5 d1d5 d8e7 d5d1 a8d8 g1g2 g6e5 e3d4 f7f6');
INSERT INTO evaluations VALUES(30294,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(30294,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(30294,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(30294,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(30294,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(30294,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(30294,36,'cp 78','b1c3 f8e7 e1g1 b8c6 a2a4 e8g8 c1f4 f8e8 d4c6 b7c6 f1e1 e7f8 h2h3 d7e6 c4f1 a7a5 b2b3 f6d7 f4e3 g7g6 d1d2 d8b8 c3e2 d7c5 e3c5 d6c5 e2f4 f8h6 g2g3 c5c4 b3c4 b8b2 a1d1 b2a2 d2c3 a2a4 f4e6 e8e6 f1g2');
INSERT INTO evaluations VALUES(30293,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(30293,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(30293,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(30293,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(30293,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(30293,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(30293,36,'cp 59','f8e7 a2a4 e8g8 c1f4 f8e8 e1g1 b8c6 d4c6 b7c6 h2h3 e7f8 f1e1 d7e6 c4f1 a7a5 b2b3 f6d7 d1d2 d8b8 a1d1 d7c5 f2f3 b8b4 f4e3 h7h6 e3d4 f7f6 c3e2 b4d2 d1d2 d6d5 e4e5 f6e5 d4e5 f8d6');
INSERT INTO evaluations VALUES(28506,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(28506,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(28506,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(28506,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(28506,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(28506,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(28506,36,'cp 61','a2a4 b8c6 c1f4 e8g8 e1g1 f8e8 d4c6 b7c6 b2b3 e7f8 f1e1 d7e6 c4f1 a7a5 h2h3 f6d7 f4e3 d8f6 d1d2 f6g6 f2f4 f7f5 f1d3 d7f6 e3d4 d6d5 e4d5 f6d5 c3d5 e6d5 e1e5 e8e5 f4e5 a8d8 a1f1 f8b4 c2c3 d5e6 d2c2');
INSERT INTO evaluations VALUES(37765,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37765,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37765,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37765,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37765,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37765,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37765,36,'cp 0','e7e6 d3d4 c5d4 f3d4 d7d5 c4b5 c8d7 e4d5 e6d5 b5d7 d8d7 e1g1 g8e7 c1e3 b8c6 b1c3 e8g8 d1d2 a8c8 a1d1 c6e5 b2b3 e5g4 e3g5 f8e8 h2h3 g4f6 f1e1 a7a6 d2f4 f6h5 f4d2');
INSERT INTO evaluations VALUES(37620,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37620,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(37620,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(37620,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37620,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(37620,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37620,36,'cp -3','d3d4 c5d4 f3d4 d7d5 c4b5 c8d7 e4d5 e6d5 e1g1 g8e7 b5d7 d8d7 c1e3 e8g8 b1d2 f8e8 h2h3 e7c8 c2c3 b8c6 d4c6 b7c6 d1c2 a7a5 d2f3 c8d6 e3d4 a5a4 f1e1 d7b7 a1d1 e8e1 d1e1 d6e4 d4g7 g8g7 c3c4 b7b4 c4d5 c6d5');
INSERT INTO evaluations VALUES(38702,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(38702,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38702,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38702,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(38702,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(38702,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(38702,36,'cp 1','g8f6 e1g1 e8g8 d2d3 c7c6 c2c3 b8d7 c4b3 c5b6 b1d2 f8e8 d3d4 d7f8 b3c2 b6c7 f1e1 h7h6 d2f1 d6d5 f3e5 f6e4 c1f4 c8f5 f4h2 e4d6 c2f5 d6f5 d1c2 d8f6 f1e3 f5h4 g1h1 f8e6 a2a4 a8d8 e3g4 f6g5 g4e3');
INSERT INTO evaluations VALUES(41870,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(41870,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41870,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(41870,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41870,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41870,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(41870,36,'cp 6','e1g1 e8g8 d2d3 c7c6 c2c3 d6d5 e4d5 c6d5 c4b3 e5e4 f3d4 b8c6 d3e4 d5e4 c1e3 d8e7 d4c6 b7c6 e3c5 e7c5 d1d4 c5g5 d4e3 g5e3 f2e3 a7a5 f1f2 c8a6 c3c4 f6d7 b1c3 d7e5 c3e4 a6c4 e4c5 c4b3 c5b3 f8d8 f2c2 a5a4 b3c5 a4a3');
INSERT INTO evaluations VALUES(40017,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(40017,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(40017,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(40017,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(40017,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(40017,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(40017,36,'cp 89','e2e4 e8g8 f1d3 d7d6 h2h3 e7e6 g1f3 e6d5 e4d5 f8e8 c1e3 a7a6 e1g1 b8d7 d1d2 b7b5 c4b5 d7b6 b5a6 c8a6 d3a6 a8a6 b2b3 d8a8 a1d1 f6e4 c3e4 e8e4 d2d3 e4e8 f1e1 e8c8 d3b5 c8b8 b5c6 g7f8 c6a8 b6a8 e1e2 a8c7 g2g4 h7h6 g1g2 a6a3 g4g5 h6g5 e3g5 f8g7');
INSERT INTO evaluations VALUES(40015,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(40015,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(40015,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(40015,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(40015,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(40015,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(40015,36,'cp 82','d7d6 h2h3 e8g8 f1d3 e7e6 g1f3 e6d5 e4d5 f8e8 c1e3 c8f5 e1g1 d8d7 e3f4 f5d3 d1d3 b8a6 a2a3 a6c7 f4h2 a7a6 a3a4 a8b8 f3d2 b7b6 f1d1 e8e7 a1a3 c7e8 d2f3 d7c7 a3a1 f6d7 d1e1 e7e1 a1e1 e8f6 h2f4 f6h5 f4d2');
INSERT INTO evaluations VALUES(41864,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(41864,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(41864,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(41864,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(41864,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41864,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41864,36,'cp 2','e8g8 e1g1 c7c6 c2c3 d6d5 e4d5 c6d5 c4b3 e5e4 f3d4 d8e7 c1e3 c8d7 d3e4 d5e4 d1e2 b8c6 f1d1 h7h6 d4c6 d7c6 e3c5 e7c5 b1d2 a7a5 a2a4 c5b6 e2e3 f6d7 b3c2 b6e3 f2e3 d7c5 d2c4 g7g6 c4b6 a8d8 b6c4');
INSERT INTO evaluations VALUES(42484,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(42484,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(42484,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(42484,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42484,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42484,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(42484,36,'cp 25','b1d2 e8g8 g1f3 d7d5 a2a3 b4e7 b2b3 b7b6 c1b2 c7c5 c4d5 e6d5 a1c1 c8b7 f1d3 b8d7 e1g1 e7d6 f3e5 f8e8 f2f4 c5d4 e3d4 d7f8 d1f3 a7a5 f3h3 f8g6 c1e1 g6f8 d3b5 e8e7 f4f5 e7c7 f1f3 b7a6 b5a6 a8a6');
INSERT INTO evaluations VALUES(45020,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(45020,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(45020,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(45020,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(45020,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(45020,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(45020,36,'cp 93','f1c4 e7e6 g1f3 c6c5 e1g1 c5d4 e3d4 f8e7 a2a3 e8g8 f1e1 d8c7 c4a2 b8c6 c1e3 h7h6 a1c1 c7a5 d1e2 c8d7 e1d1 f8e8 h2h3 a8c8 e3f4 a5a6 e2a6 b7a6 f3e5 e7d6 f4g3 d6e5 g3e5 c6e5');
INSERT INTO evaluations VALUES(49657,30,'cp -275',NULL);
INSERT INTO evaluations VALUES(49657,31,'cp -279',NULL);
INSERT INTO evaluations VALUES(49657,32,'cp -263',NULL);
INSERT INTO evaluations VALUES(49657,33,'cp -288',NULL);
INSERT INTO evaluations VALUES(49657,34,'cp -296',NULL);
INSERT INTO evaluations VALUES(49657,35,'cp -279',NULL);
INSERT INTO evaluations VALUES(49657,36,'cp -274','b8c6 f1d3 g7g5 f3e3 g5g4 h3g1 d7d5 d3b5 f6e4 d2d3 f8c5 e3h6 d8f6 h6f6 e4f6 g1e2 h8g8 c2c3 c8d7 b1d2 e8c8 e1g1 h7h5 d3d4 c5d6 b5c6 d7c6 d4e5 d6e5 d2b3 c6b5 f1e1 d8e8 c1e3 h5h4 a1d1 h4h3 e3d4 f6e4 d4e5 e8e5 e2f4 b5c4');
INSERT INTO evaluations VALUES(16574,30,'cp -266',NULL);
INSERT INTO evaluations VALUES(16574,31,'cp -265',NULL);
INSERT INTO evaluations VALUES(16574,32,'cp -265',NULL);
INSERT INTO evaluations VALUES(16574,33,'cp -252',NULL);
INSERT INTO evaluations VALUES(16574,34,'cp -261',NULL);
INSERT INTO evaluations VALUES(16574,35,'cp -273',NULL);
INSERT INTO evaluations VALUES(16574,36,'cp -276','d2d3 d7d5 c2c3 c8g4 f3e3 d5e4 b1d2 e4d3 e3d3 f8c5 d3d8 a8d8 f2f3 g4h3 g2h3 f6d5 d2b3 c5b6 c1d2 a7a6 e1c1 d5e3 d2e3 b6e3 c1c2 d8d1 c2d1 f7f5 f1d3 g7g6 d1c2 e8f7 h1e1 e3f2 e1f1 f2b6 a2a4 h8d8 f1e1 f7f6 b3d2 h7h5 d2c4 h5h4');
INSERT INTO evaluations VALUES(55340,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(55340,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(55340,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(55340,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(55340,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(55340,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(55340,36,'cp 96','g8f6 c1e3 f8g7 d1d2 b8c6 f2f3 e8g8 e1c1 d6d5 d4c6 b7c6 f1c4 c8e6 c4b3 d8e7 d2f2 f8d8 e3c5 e7e8 e4d5 f6d5 c3e4 e6f5 h1e1 e8e5 c5d4 e5f4 c1b1 f5e4 e1e4 f4h2 d1e1 d5f6 e4e7 f6d5 d4e5 g7e5 e7e5 h2f4 g2g3 f4f6 a2a3');
INSERT INTO evaluations VALUES(47581,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(47581,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(47581,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(47581,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(47581,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(47581,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(47581,36,'cp 90','c1e3 f8g7 d1d2 e8g8 f2f3 b8c6 e1c1 d6d5 d4c6 b7c6 e3h6 c8e6 e4d5 e6d5 h6g7 g8g7 d2f4 d8e7 f1d3 a7a5 a2a3 f8d8 h1e1 e7d6 f4d6 c7d6 c3d5 f6d5 d3b5 c6b5 d1d5 b5b4 a3b4 a5b4 c1d2 a8a2 e1b1 h7h5 d5d4 d8e8 d4b4 a2a5 b4e4 a5d5 d2c3 d5c5 e4c4 c5c4 c3c4 e8e2 c4c3 e2g2');
INSERT INTO evaluations VALUES(3067,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(3067,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(3067,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(3067,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(3067,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(3067,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(3067,36,'cp 72','f1e2 d7d6 e1g1 a7a6 d1d2 c8d7 f2f3 c6d4 e3d4 a8c8 d4e3 d7e6 b2b3 d8a5 a1c1 f8e8 g1h1 f6d7 e3d4 g7d4 d2d4 e8f8 f3f4 d7f6 a2a4 a5c5 d4d2 e6d7 h2h3 d7c6 e2f3 c5b6 c1b1 b6b4 f1d1 b7b5 a4b5');
INSERT INTO evaluations VALUES(3066,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(3066,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(3066,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(3066,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(3066,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(3066,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(3066,36,'cp 62','d7d6 e1g1 c6d4 e3d4 c8d7 d1c2 d8c7 f1d1 a7a5 b2b3 d7c6 d4e3 f6d7 a1c1 d7c5 c3d5 c7d8 e3g5 f8e8 e2f3 h7h6 g5e3 h6h5 e3c5 d6c5 d5f6 g7f6 d1d8 a8d8 c2e2 f6e5 c1d1 d8d1 e2d1 e8a8 d1e1 a5a4 b3a4 c6a4');
INSERT INTO evaluations VALUES(4146,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(4146,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(4146,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(4146,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(4146,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(4146,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(4146,36,'cp -26','g5f3 d7d6 d2d3 c6a5 c4b3 a5b3 a2b3 a7a5 c1e3 c5e3 f2e3 c7c6 d1e1 b7b5 h2h3 c8e6 e1g3 a8a7 f3h4 g8h7 g3f2 a7a8 h4f3 d8e7 f3h4 e7c7 h4f5 e6f5 f2f5 h7g8 f5f3 a8e8 c3e2 e8e6 c2c4 f6d7 c4b5 c6b5');
INSERT INTO evaluations VALUES(4165,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(4165,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(4165,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(4165,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(4165,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(4165,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(4165,36,'cp -27','f8e8 f1e1 d7d6 h2h3 a7a6 a2a4 c8e6 d2d3 d8d7 c1e3 e6c4 d3c4 c5b4 e3d2 b4c3 d2c3 c6d8 b2b3 d8e6 f3d2 d7c6 a4a5 e8f8 f2f3 f6h5 d2f1 f7f5 e4f5 f8f5 d1d5 h5f4 d5c6 b7c6 c3d2 c6c5 f1g3 f5f7 d2f4 e6f4');
INSERT INTO evaluations VALUES(15416,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(15416,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(15416,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(15416,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(15416,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(15416,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(15416,36,'cp -22','a7a6 b5c6 b7c6 b1c3 h7h6 g5e3 c6c5 e1g1 c8e6 h2h3 f8e7 f3d2 g7g5 d2c4 h8g8 e3d2 h6h5 d1e2 d8d7 a1e1 h5h4 f2f3 f6h5 e2d1 e7d8 b2b3 f7f6 c3d5 e6d5 e4d5 h5g3 f1f2 f6f5');
INSERT INTO evaluations VALUES(16540,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(16540,31,'cp -138',NULL);
INSERT INTO evaluations VALUES(16540,32,'cp -162',NULL);
INSERT INTO evaluations VALUES(16540,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(16540,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(16540,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(16540,36,'cp -140','d7d5 f3e2 f8c5 g1f3 c8e6 f3g5 e6g4 g5f3 d8d6 b1d2 g4h5 e4d5 d6d5 d2e4 e8c8 e2c2 h5g6 e4c5 d5c5 c1e3 c5d5 c2a4 g6d3 e1c1 d5e4 a4e4 d3e4 d1d8 h8d8 f3g5 e4g6 f1b5 f6d5 b5c6 b7c6 h1e1 h7h6 g5f3 f7f6 f3d2 d5e3');
INSERT INTO evaluations VALUES(15151,30,'cp -154',NULL);
INSERT INTO evaluations VALUES(15151,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(15151,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(15151,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(15151,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(15151,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(15151,36,'cp -129','h2h3 c8e6 b1d2 h8g8 f3e2 d8d7 g1f3 h7h6 b2b4 e8c8 c1b2 c8b8 g2g3 g7g5 f1g2 d5e4 d3e4 g5g4 h3g4 f6g4 e1g1 f7f6 f1d1 d7f7 f3h4 h6h5 d2f1 f8d6 f2f3 e6c4 e2e1 g4h6 b2c1 f7g7 c1h6 g7h6 h4f5');
INSERT INTO evaluations VALUES(15150,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(15150,31,'cp -147',NULL);
INSERT INTO evaluations VALUES(15150,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(15150,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(15150,34,'cp -147',NULL);
INSERT INTO evaluations VALUES(15150,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(15150,36,'cp -147','c8e6 b1d2 a7a5 f3d1 f8c5 f1e2 d8e7 g1f3 e8c8 d1c2 c5b6 d2f1 c8b8 c1e3 f6h5 d3d4 e5d4 f3d4 c6d4 c3d4 g7g6 g2g4 h5g7 e4e5 h7h5 h1g1 h5g4 h3g4 f7f6 e5f6 e7f6 e1c1 e6d7 c1b1 g7e6 g4g5 f6g7 c2c3 g7f7 f1g3 e6f4 e2f1 f4h3 f1h3 h8h3 b1a1 f7f3 g1e1');
INSERT INTO evaluations VALUES(20702,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(20702,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(20702,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(20702,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(20702,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(20702,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(20702,36,'cp 47','d2d4 c5d4 f3d4 f8e7 c1e3 g8f6 d1d2 f6g4 e3g1 c6d4 d2d4 e8g8 e1c1 e6e5 f4e5 g4e5 f1e2 f8e8 c1b1 a7a6 d4b6 d8b6 g1b6 c8g4 h1e1 a8c8 e2g4 e5g4 h2h3 g4f6 a2a4 h7h6 b6d4 f6d7 d1d3 e7h4 e1e2 h4d8 c3d5 d7c5 d4c5 c8c5 c2c3 e8e5 b1c2');
INSERT INTO evaluations VALUES(22370,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(22370,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(22370,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(22370,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(22370,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(22370,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(22370,36,'cp 52','g8f6 a2a4 f8e7 f1c4 e8g8 e1g1 h7h6 f1e1 f8e8 h2h3 c7c6 a4a5 d8c7 c4f1 d7f8 d4d5 c8d7 c1e3 f8g6 f1c4 e7f8 f3d2 g6e7 d1f3 c6d5 e4d5 e7f5 c4d3 f5e3 e1e3 a8b8 g2g4 g8h8 d2c4 b7b5 a5b6 a7b6');
INSERT INTO evaluations VALUES(23800,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(23800,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23800,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(23800,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(23800,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(23800,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(23800,36,'cp 26','d1d2 g8f6 f2f3 e8g8 e1c1 a7a6 b5e2 a8c8 h2h4 f6h5 c1b1 h5g3 h1e1 c6e5 e2d3 b7b5 a2a3 f7f5 e4f5 e5d3 d2d3 g3f5 g2g4 f5e3 e1e3 c8c4 e3e4 g8h8 c3d5 e7e5 d4e2 c4c8 e2c3 d7e6 d3e3 a6a5 h4h5 g6h5 g4h5');
INSERT INTO evaluations VALUES(23801,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(23801,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(23801,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(23801,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(23801,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(23801,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(23801,36,'cp 37','g8f6 f2f3 c6d4 b5d7 d8d7 e3d4 e8g8 e1c1 a8c8 d2e2 c8c6 c1b1 f8c8 e4e5 f6e8 e5d6 g7d4 d1d4 c6d6 d4d6 e8d6 h1d1 g8f8 c3d5 d7e6 e2e6 f7e6 d5c3 c8c5 d1e1 f8f7 c3e4 c5d5 b1c1 d6e4 e1e4 a7a6 a2a4 b7b5 a4b5 a6b5 c2c3 e6e5');
INSERT INTO evaluations VALUES(24003,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(24003,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(24003,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(24003,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(24003,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24003,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(24003,36,'cp 29','f2f3 c6d4 e3d4 d7b5 c3b5 e8g8 e1c1 a8c8 c1b1 a7a6 b5c3 c8c4 h1e1 b7b5 b2b3 c4d4 d2d4 f6e4 d4e4 g7c3 e4e7 c3e1 e7d8 f8d8 d1e1 g8f8 b1b2 h7h5 e1e4 f7f5 e4d4 f8e7 a2a4 d8c8 h2h4 e7e6 c2c3 c8c5 b2c2 c5e5 c2d2 e6e7 g2g3 a6a5 a4b5');
INSERT INTO evaluations VALUES(30085,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(30085,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(30085,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(30085,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(30085,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(30085,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(30085,36,'cp -13','f8d6 g1f3 d6f4 e3f4 e8g8 f1e2 b8d7 e1g1 c7c5 d4c5 d7c5 d2b3 c5a4 d1c1 h7h6 b3d4 f5h7 b2b3 a4c5 f3e5 a8c8 c1e3 c5e4 f1c1 a7a6 a2a4 f6d7 f2f3 e4d6 a4a5 d6f5 d4f5 h7f5 b3b4 c8c7 b4b5 d7e5 f4e5 a6b5 e2b5');
INSERT INTO evaluations VALUES(28450,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(28450,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(28450,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(28450,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(28450,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(28450,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(28450,36,'cp -11','g1f3 d6f4 e3f4 e8g8 f1e2 b8d7 e1g1 h7h6 a2a4 c7c6 f3e5 c6c5 d4c5 d7c5 a4a5 d8c7 d2f3 f6e4 f3d4 e4d6 d4f5 d6f5 e2d3 c5d3 d1d3 f8c8 f1d1 b7b6 g2g4 f5e7 d3a6 b6a5 a6a5 c7a5 a1a5 a8b8 d1a1 b8b2');
INSERT INTO evaluations VALUES(30161,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30161,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30161,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30161,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(30161,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30161,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30161,36,'cp 0','d3b5 b8d7 b5b7 a8b8 b7a7 b8b2 g1f3 f8d6 f4d6 c7d6 e1g1 e8g8 a7a3 b2c2 a3d6 f6e4 d6b4 d8a8 b1a3 c2e2 b4b5 e2f2 f1f2 e4f2 a3b1 f2e4 b5d7 a8b8 b1d2 b8b2 d2b3 b2f2 g1h1 f2e3 a1f1 e4f2 h1g1 f2d3 g1h1');
INSERT INTO evaluations VALUES(40140,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(40140,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(40140,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(40140,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(40140,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(40140,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(40140,36,'cp 34','a2a3 e8g8 c4a2 f8e8 e1g1 b8d7 f1e1 e7f8 c1g5 h7h6 g5h4 d7c5 f2f3 a7a5 h2h3 c5e6 d1d2 e6d4 d2d4 c8e6 a2e6 f7e6 g1h1 e6e5 d4c4 g8h8 a1d1 d8e7 h4f6 e7f6 d1d3 f6g5 e1d1 a8d8 c4a4');
INSERT INTO evaluations VALUES(45136,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(45136,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(45136,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(45136,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(45136,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(45136,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(45136,36,'cp -7','e7e6 g1f3 f8e7 a2a4 b8c6 e1g1 e8g8 c4a2 c8d7 f1e1 a7a6 c2c3 b7b5 c1f4 d8c7 b1d2 f8d8 f4g3 h7h6 a2b3 d7e8 d1e2 c7b6 b3c2 a8b8 a4b5 a6b5 e4e5 d6e5 f3e5 e7d6 d2f3 b6c7 e5c6 e8c6 f3e5 c6b7');
INSERT INTO evaluations VALUES(45998,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(45998,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(45998,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(45998,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45998,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45998,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45998,36,'cp 0','g1f3 f8e7 a2a4 b8c6 e1g1 e8g8 c4a2 a7a6 c1f4 c8d7 f4g3 b7b5 b1d2 d8c7 f1e1 f8d8 c2c3 d7e8 a2b3 h7h6 b3c2 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(45993,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(45993,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(45993,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(45993,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(45993,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(45993,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(45993,36,'cp -14','f8e7 e1g1 e8g8 c1f4 d6d5 e4d5 e6d5 c4b3 b8c6 f1e1 h7h6 b1d2 e7d6 f4d6 d8d6 b3a4 c8f5 a4c6 d6c6 d3d4 c6b6 d4c5 b6b2 d2b3 b2c2 d1c2 f5c2 e1e7 c2b3 a2b3 f8c8 e7b7 c8c5 a1a7 a8a7 b7a7 c5c1 g1h2 c1c2 h2g3 f6h5 g3g4');
INSERT INTO evaluations VALUES(47929,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47929,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47929,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47929,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47929,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47929,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(47929,36,'cp 0','b1c3 g7g6 f3g5 d6d5 e4d5 c8g4 g5f3 d8e7 d1e2 e7e2 c3e2 b8d7 d2d3 g4f3 g2f3 d7b6 e2f4 e8c8 c1d2 f8d6 e1c1 d6f4 d2f4 f6d5 f4e5 b6c4 d3c4 f7f6 e5d4 d5b6 b2b3 b6d7 h2h4 d8e8 a2a4 b7b6 a4a5 h8f8 a5b6 a7b6 h4h5 g6g5');
INSERT INTO evaluations VALUES(47927,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47927,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47927,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47927,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47927,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47927,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47927,36,'cp 0','g7g6 f3g5 d6d5 e4d5 c8g4 g5f3 d8e7 d1e2 e7e2 c3e2 b8d7 d2d4 e8c8 c1f4 d7b6 f4e5 f6d7 c4b3 g4f3 g2f3 d7e5 d4e5 b6d5 a1d1 c7c6 b3d5 c6d5 f3f4 f8g7 e2d4 h8e8 e1f2 f7f6 e5e6 f6f5 h1e1 g7f6 e1e2 e8e7 f2f3 f6d4 d1d4 d8d6 b2b3 d6e6 e2e6 e7e6 d4d5 e6e8');
INSERT INTO evaluations VALUES(49418,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49418,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(49418,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49418,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49418,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49418,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49418,36,'cp 2','d7d5 c1f4 f8d6 f4g3 c7c5 e2e3 e8g8 b1d2 b7b6 f1d3 c8b7 d1e2 d8c7 g3d6 c7d6 e3e4 d5e4 d2e4 f6e4 d3e4 b7e4 e2e4 b8d7 e1g1 a8d8 a1d1 h7h6 e4e2 d7f6 d4c5 d6c5 c3c4 c5c7 h2h3 f6d7 b2b3 e6e5 d1d5 f8e8 f1e1 e5e4');
INSERT INTO evaluations VALUES(5179,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(5179,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5179,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5179,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(5179,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(5179,36,'cp 14','d8b6 c3a4 b6a5 c2c3 c5d4 e3d4 c6d4 b2b4 d4f3 g2f3 a5e5 d1d4 e5d4 c3d4 d7d6 e1d2 c8d7 b4b5 g8f6 a1c1 e8e7 f1d3 a8c8 f3f4 f6g4 d2e2 g4f6 f2f3 h7h6 h1g1 g7g6 a4c3 g6g5 e2e3 g5f4 e3f4 h8g8 g1g8 c8g8 f4e3 g8g2');
INSERT INTO evaluations VALUES(110,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(110,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(110,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(110,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(110,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(110,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(110,36,'cp 14','c3a4 b6a5 c2c3 c5d4 e3d4 c6d4 b2b4 d4c2 d1c2 a5e5 f1d3 g8f6 c2e2 d7d5 e4d5 e5e2 d3e2 f6d5 e2f3 c8d7 f3d5 e6d5 a4c5 a7a5 e1d2 b7b6 h1e1 e8d8 c5d3 h8e8 e1e8 d7e8 a2a3 d8e7 d3f4 e7d6 f4e2 f7f6 g2g3 e8g6 e2d4 a5b4 c3b4 d6e5 d2e3 g6e4');
INSERT INTO evaluations VALUES(111,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(111,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(111,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(111,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(111,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(111,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(111,36,'cp 0','b6a5 c2c3 c5d4 e3d4 c6d4 b2b4 d4c2 d1c2 a5e5 f1d3 g8f6 c2e2 d7d5 e4d5 e5e2 d3e2 f6d5 e2f3 c8d7 f3d5 e6d5 a4c5 a7a5 e1d2 b7b6 h1e1 e8d8 c5d7 d8d7 a2a4 h8c8 a1b1 c8c6 h2h3 d7d6 e1e3 g7g6 b4a5 a8a5 b1b4 a5c5 e3d3 d6e6 d3e3 e6d6');
INSERT INTO evaluations VALUES(112,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(112,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(112,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(112,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(112,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(112,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(112,36,'cp 3','c2c3 c5d4 e3d4 c6d4 b2b4 d4c2 d1c2 a5e5 f1d3 g8f6 c2e2 d7d5 e4d5 e5e2 d3e2 f6d5 e2f3 c8d7 f3d5 e6d5 a4c5 a7a5 e1d2 b7b6 h1e1 e8d8 c5d7 d8d7 a2a4 h8c8 a1b1 c8c4 b4a5 a8a5 b1b6 c4f4 f2f3 a5a4 b6b7 d7c6 e1e7 a4a2 d2d3 a2g2 e7f7 g2h2 f7f4 c6b7 f4f7 b7b6');
INSERT INTO evaluations VALUES(9915,30,'cp 286',NULL);
INSERT INTO evaluations VALUES(9915,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(9915,32,'cp 245',NULL);
INSERT INTO evaluations VALUES(9915,33,'cp 250',NULL);
INSERT INTO evaluations VALUES(9915,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(9915,35,'cp 256',NULL);
INSERT INTO evaluations VALUES(9915,36,'cp 249','f6e4 d1e2 b6a5 e1f1 d7d5 e5d6 c8f5 c1g5 f7f6 g5f4 e8f8 c4d3 d8e8 b1c3 e4c3 b2c3 e8e2 f1e2 f5d3 e2d3 c7d6 f4d6 f8f7 a1b1 b7b6 f3d2 a8d8 d2c4 d8d7 d6f4 h8d8 h1c1 h7h5 d3e2 h5h4 f4e3 h4h3 g2g3 d8e8 a2a4 f7g8 e2f3 d7c7 d4d5 c6e5 c4e5 e8e5 c3c4 c7d7 f3g4');
INSERT INTO evaluations VALUES(9945,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(9945,31,'cp 263',NULL);
INSERT INTO evaluations VALUES(9945,32,'cp 278',NULL);
INSERT INTO evaluations VALUES(9945,33,'cp 274',NULL);
INSERT INTO evaluations VALUES(9945,34,'cp 258',NULL);
INSERT INTO evaluations VALUES(9945,35,'cp 285',NULL);
INSERT INTO evaluations VALUES(9945,36,'cp 276','d1e2 b6a5 e1f1 d7d5 e5d6 c8f5 c1g5 f7f6 g5f4 e8f8 c4d3 d8e8 b1c3 e4c3 b2c3 e8e2 f1e2 a8e8 e2d2 f5d3 d2d3 g7g5 f4g3 c7d6 a1b1 b7b6 h2h4 g5g4 g3d6 f8f7 f3d2 e8d8 d2c4 d8d7 d6f4 h8d8 h1c1 a5b4 d3e2 d8e8 e2f1 b4f8 c4e3 c6e7 e3g4 e7d5 f4d2 d5c7 g4e3 f8h6 c1c2 h6e3 d2e3 c7d5');
INSERT INTO evaluations VALUES(12251,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(12251,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(12251,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12251,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(12251,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(12251,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(12251,36,'cp 12','f6e4 d3e4 d7f6 e4d3 c8g4 c2c3 e7e6 h2h3 g4h5 d1e2 d8d5 g2g4 h5g6 e1g1 g6d3 e2d3 d5d8 c3c4 d8c7 f1e1 f8d6 c1d2 e8g8 b2b3 h7h6 a1d1 d6e7 f3e5 f6d7 d2f4 d7e5 f4e5 e7d6 g1g2 a8d8 e5g3 d6g3 d3g3 c7g3 g2g3 d8d7 e1e5 f8d8');
INSERT INTO evaluations VALUES(12255,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12255,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12255,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(12255,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12255,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(12255,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(12255,36,'cp 23','d3e4 d7f6 e4d3 c8g4 d3e2 e7e6 e1g1 f8d6 f3e5 g4f5 f1e1 e8g8 e5c4 d6e7 c2c3 f6d5 e2f3 d8c7 c4e3 d5e3 c1e3 h7h6 g2g3 a8d8 d1e2 d8d7 a2a4 a7a5 a1d1 e7d6 e2c4 f8d8 g1g2 d6e7 h2h4 e7f6 c4e2 b7b6 e3f4');
INSERT INTO evaluations VALUES(12254,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(12254,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(12254,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(12254,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12254,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(12254,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(12254,36,'cp 15','d7f6 e4d3 c8g4 d3e2 e7e6 e1g1 f8d6 f3e5 g4f5 f1e1 d8c7 e2f3 f6d7 e5d7 c7d7 d1e2 e8g8 c1e3 a7a5 a1d1 h7h6 c2c3 f8e8 g2g3 d7c7 g1g2 a8d8 e2c4 f5c2 d1d2 c2f5 a2a4 c7e7 h2h4 e7f6 c4e2 f6e7 g2g1 f5h7');
INSERT INTO evaluations VALUES(13419,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(13419,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(13419,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(13419,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13419,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13419,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(13419,36,'cp 14','e4d3 c8g4 d3e2 e7e6 e1g1 f8d6 h2h3 g4f5 e2d3 e8g8 f1e1 f5d3 d1d3 h7h6 c1d2 d8c7 c2c4 a7a5 b2b3 f8d8 d3e2 d6b4 e1d1 b4d2 d1d2 c6c5 a1c1 c7b6 c1d1 a5a4 d2d3 c5d4 f3d4 d8d7 d4b5 d7d3 d1d3 a4b3');
INSERT INTO evaluations VALUES(13417,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(13417,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(13417,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(13417,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(13417,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(13417,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(13417,36,'cp 9','c8g4 c2c3 e7e6 h2h3 g4h5 d1e2 f8d6 c1d2 d8a5 e1g1 a8d8 c3c4 a5c7 f1e1 e8g8 a1d1 f8e8 d2e3 d6b4 e1f1 c6c5 g2g4 h5g6 d4c5 f6d7 d3g6 h7g6 g1g2 e6e5 d1d5 b4c5 f1d1 c5e3 e2e3 e5e4');
INSERT INTO evaluations VALUES(24682,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24682,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(24682,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(24682,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(24682,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(24682,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24682,36,'cp 14','d3e2 e7e6 e1g1 f8d6 f3e5 g4f5 f1e1 e8g8 e5c4 d6e7 e2f3 f6d5 c2c3 d8c7 a2a4 a8d8 d1b3 c6c5 c4e3 d5e3 c1e3 b7b6 b3c4 d8d7 a4a5 h7h6 a5b6 a7b6 c4b5 e7d6 f3c6 d6h2 g1h1 c5d4 e3d4 d7d4 c3d4 h2f4');
INSERT INTO evaluations VALUES(18033,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(18033,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(18033,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(18033,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(18033,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(18033,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(18033,36,'cp 93','c8b7 c2c3 d7d6 d2d4 g8f6 f1e1 f8e7 b1d2 e8g8 d2f1 f8e8 f1g3 e7f8 b3c2 d6d5 d4e5 f6e4 g3e4 d5e4 c2e4 d8d1 e1d1 a8b8 c1f4 g7g5 f4g3 g5g4 e4c6 b7c6 f3d4 c6d7 e5e6 d7e6 g3c7 b8c8 c7g3 e6d7 g3f4 h6h5 d1d2 f8e7 d4c2 d7e6 c2e3 g8g7');
INSERT INTO evaluations VALUES(34348,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(34348,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(34348,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(34348,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(34348,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(34348,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(34348,36,'cp 207','f1e2 g7g6 d2d4 f8g7 d4c5 g8f6 c1e3 e8g8 h2h3 b8c6 e1g1 f8d8 d1c1 f6e4 c3e4 e6e4 c2c3 c8f5 f1e1 e4c2 a2a4 c2c1 a1c1 a7a5 e2b5 f5e6 c1d1 h7h6 g1f1 g6g5 d1d8 a8d8 f3d2 c6e5 f1e2 f7f5 g2g3 e6d5');
INSERT INTO evaluations VALUES(34349,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(34349,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(34349,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(34349,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(34349,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(34349,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(34349,36,'cp 230','g7g6 e1g1 g8f6 d2d4 f8g7 d4c5 e8g8 h2h3 e6f5 c1e3 b8c6 d1c1 c6b4 f3d4 f5e5 f1d1 f6d5 c3d5 b4d5 e3g5 d5f6 c5c6 b7c6 g5f4 e5c5 f4e3 f6d5 d4e6 c5b4 c2c3 g7c3 e6f8 b4b2 a1b1 b2c1 b1c1 c3g7 c1c6 c8b7 d1d5 b7c6 d5c5 c6b7 f8d7');
INSERT INTO evaluations VALUES(36401,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36401,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36401,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(36401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36401,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36401,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36401,36,'cp 0','f1d3 f8e8 e1g1 e7d6 f1e1 h7h6 c3b5 d6f8 b5c3');
INSERT INTO evaluations VALUES(36410,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36410,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36410,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36410,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(36410,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36410,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36410,36,'cp -12','f8e8 e1g1 e7d6 f1e1 c7c6 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 c1d2 b8d7 d1b3 e8e7 e4c2 d7f6 f3e5 c8e6 b3d3 d8b6 e5c4 b6c7 c4d6 c7d6 d2c3 e7d7 e1e5 d6e5 d4e5 d7d3 c2d3 f6d5 c3e1 d5f4 d3f1 a8d8 f2f3 e6d5 a1d1 f4e6 a2a3');
INSERT INTO evaluations VALUES(42034,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(42034,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(42034,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42034,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42034,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(42034,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(42034,36,'cp -5','d5c4 a2a3 b4c3 b2c3 b7b5 a3a4 c7c6 d1b1 a7a6 e2e4 e8g8 f1e2 b8d7 e1g1 c8b7 f3d2 d8c8 f4d6 f8d8 f2f3 e6e5 d4e5 d7e5 d6e5 d8d2 b1e1 c8d8 e5f4 d2d7 e1f2 d8e7 f2b6 a8d8 f4e3 d8e8 e3d4 e7d8 f1d1 f6d5 b6d8 d7d8 e2c4 b5c4 e4d5 c6d5');
INSERT INTO evaluations VALUES(42167,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(42167,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(42167,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(42167,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42167,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42167,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42167,36,'cp -5','a2a3 b4c3 b2c3 c7c6 a3a4 b7b5 d1b1 a7a6 e2e4 b8d7 f4d6 d7b6 d6c5 c8b7 c5b6 d8b6 f1c4 c6c5 a4b5 e8g8 b5a6 b6b1 a1b1 b7a6 c4a6 a8a6 f3d2 a6a3 e1e2 a3c3 d4c5 c3c5 b1c1 c5e5 c1c4 g7g5 h1c1 e5b5 c4c7 b5b4 h2h3 b4b5 c1c2 g5g4 f2f3 g4h3 g2h3');
INSERT INTO evaluations VALUES(50607,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(50607,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(50607,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(50607,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(50607,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(50607,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(50607,36,'cp 38','g1f3 d7d5 e4d5 d8d5 d2d4 g8f6 c1e3 c5d4 c3d4 d5d8 b1c3 e7e6 f1d3 b8c6 a2a3 f8e7 e1g1 b7b5 a3a4 b5b4 c3e4 c8b7 a1c1 e8g8 d1e2 c6a5 e4c5 e7c5 d4c5 b7f3 g2f3 f6d5 c5c6 a8c8 f1d1 d8f6 d3a6 c8c6 c1c6 d5e3');
INSERT INTO evaluations VALUES(53851,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(53851,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(53851,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(53851,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(53851,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(53851,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(53851,36,'cp 146','e2e4 g8f6 g1f3 f8e7 f1d3 b8d7 h2h3 d8c7 e1g1 a7a6 f1e1 e6e5 c1e3 b7b5 c4b5 a6b5 d1c2 c7b8 a2a4 b5a4 c3a4 c8a6 d3a6 a8a6 d4e5 d7e5 f3e5 d6e5 a4c5 a6a1 e1a1 e8g8 c5d3 f6e4 c2c6 e4f6 a1a7 e5e4 d3f4');
INSERT INTO evaluations VALUES(3483,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(3483,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(3483,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(3483,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(3483,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(3483,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(3483,36,'cp 45','e7e5 d4d5 c6b8 e2e4 a7a5 a2a3 b8d7 c1g5 h7h6 g5e3 f6g4 e3d2 d7c5 b2b4 a5b4 a3b4 a8a1 d1a1 c5b3 a1b2 b3d2 b2d2 g4f6 h2h3 c8d7 f1d1 g8h7 d2b2 d8c8 g1h2 c8e8 d1a1 c7c6 f3d2 e8d8 c4c5 d8e7 c5d6 e7d6 d2c4 d6e7 a1a7 c6d5 c3d5 f6d5');
INSERT INTO evaluations VALUES(3540,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(3540,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(3540,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(3540,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(3540,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(3540,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(3540,36,'cp 56','d4d5 c6b8 e2e4 b8d7 c1g5 a7a5 a2a3 h7h6 g5e3 f6g4 e3d2 d7c5 b2b4 a5b4 a3b4 a8a1 d1a1 c5b3 a1a2 b3d2 a2d2 g4f6 f1a1 c7c6 d5c6 b7c6 a1d1 c8e6 f3e5 d8c7 e5c6 c7c6 e4e5 c6c4 e5f6 g7f6 c3d5 f6g7 d5e7 g8h7 g2d5 e6d5 e7d5 f8b8 d2d3 c4d3 d1d3 b8b5 g1g2 h6h5');
INSERT INTO evaluations VALUES(3531,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(3531,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(3531,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(3531,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(3531,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(3531,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(3531,36,'cp 48','c6b8 e2e4 b8d7 c1g5 h7h6 g5e3 f6g4 e3d2 a7a5 b2b3 d7c5 d1c2 g4f6 f3h4 c8d7 d2e3 b7b6 c3b5 a8c8 a1d1 f6g4 e3d2 g4f6 f1e1 f6h5 a2a3 h5f4 b3b4 f4g2 g1g2 a5b4 a3b4 c5a6 d1a1 c8a8 c2b3 d8e8');
INSERT INTO evaluations VALUES(7493,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(7493,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(7493,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(7493,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(7493,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(7493,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(7493,36,'cp -35','f6e4 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 f1e1 d7d5 d2d3 e4c5 f3e5 c6e5 e1e5 e8g8 e5e1 e7d6 b1a3 a8c6 d3d4 c5b3 d1b3 f8e8 a3c2 e8e1 c2e1 d8h4 g2g3 h4e4 c1e3 f7f6 b3c2 g7g5 c2e4 d5e4 h2h4 h7h6 h4g5 f6g5 e1c2 c6d5 c2b4 d5b3 b4a6 b3c4 a6c5 c4d5 c5d7 g8f7 d7e5 f7f6');
INSERT INTO evaluations VALUES(10073,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(10073,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(10073,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(10073,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(10073,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(10073,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(10073,36,'cp -27','c5b6 f1c4 c6d4 e1g1 d7d6 f3d4 b6d4 a1b1 e8g8 d1e1 c8e6 c4e6 f7e6 d2d3 f6g4 c3d1 d8h4 h2h3 g4f6 g1h2 b7b5 f2f3 h4e1 f1e1 a7a5 c2c3 d4b6 c1e3 b6e3 d1e3 a5b4 a3b4 f6h5 d3d4 e5d4 c3d4 h5f4 e1d1');
INSERT INTO evaluations VALUES(9918,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(9918,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(9918,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(9918,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(9918,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(9918,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(9918,36,'cp -9','f1c4 a7a5 a1b1 a5b4 a3b4 d7d6 h2h3 e8g8 e1g1 h7h6 d2d3 c6e7 c1e3 b6e3 f2e3 c7c6 c4b3 e7g6 d1d2 c8e6 b3e6 f7e6 d2e1 d8e7 b4b5 c6c5 b1a1 a8a1 e1a1 d6d5 a1e1 g8h7 e4d5 e6d5 e3e4');
INSERT INTO evaluations VALUES(12292,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(12292,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(12292,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(12292,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(12292,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(12292,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(12292,36,'cp -26','b7c6 d1e2 d7d5 c4d3 f8e7 b1c3 e8g8 e1g1 f6d7 c1f4 d7c5 a1d1 f8e8 f1e1 d8b6 b2b3 d5d4 c3b1 c5d3 e2d3 c8a6 d3g3 f7f6 e4e5 f6f5 h2h4 c6c5 b1d2 g8h8 c2c3 d4c3 g3c3 e7h4 d2c4 a6c4 c3c4 h4e7 f4e3 b6c6 e3d2 a7a5 d2e3');
INSERT INTO evaluations VALUES(12311,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(12311,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(12311,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(12311,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12311,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(12311,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12311,36,'cp 35','a2a3 d7d6 f1d3 g7g6 c1g5 d8c7 b1c3 d5c3 b2c3 f8g7 e5d6 c7d6 e1g1 e8g8 f1e1 b7b6 h2h4 h7h6 g5c1 c6e7 a3a4 c8b7 h4h5 g6g5 a4a5 f8d8 a5a6 b7d5 c1a3 d6c7 f3e5 e7f5 a3b4 a8c8 d1e2 d5a8 a1d1 f5h4 d3e4 a8e4 e2e4');
INSERT INTO evaluations VALUES(17820,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(17820,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(17820,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(17820,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(17820,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(17820,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(17820,36,'cp -69','c6c5 d4e2 d8d1 e1d1 c8e6 b1c3 e8c8 d1e1 f7f5 e4e5 g8e7 c1g5 h7h6 g5e7 f8e7 h2h4 g7g5 a1d1 d8d1 e1d1 h8d8 d1e1 f5f4 h4g5 e7g5 f2f3 b7b6 b2b3 d8e8 c3e4 e6d7 e4g5 h6g5 h1h7 c5c4 h7g7 e8e5 e1d2 d7b5 e2d4 c4b3 d4b5 e5b5 a2b3 b5c5');
INSERT INTO evaluations VALUES(17825,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(17825,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(17825,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(17825,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(17825,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(17825,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(17825,36,'cp -71','d4e2 d8d1 e1d1 c8e6 b1c3 e8c8 d1e1 f7f5 e4e5 g8e7 c1g5 e6f7 h2h4 d8d7 g5e7 f8e7 a2a3 b7b5 a1d1 b5b4 d1d7 c8d7 c3d1 b4a3 b2a3 c5c4 h1h3 g7g6 d1e3 e7a3 h4h5 g6h5 h3f3 h8e8 f3f5 f7g6 f5f6 e8e6 e3c4 e6f6 e5f6 a3b4 c2c3 b4c5 c4e5 d7e6');
INSERT INTO evaluations VALUES(18846,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(18846,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(18846,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(18846,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(18846,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(18846,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18846,36,'cp 9','d8b6 f1c4 d7d6 d2d3 f8e7 e1g1 g8f6 b2b4 c5b4 a3b4 c6b4 c1e3 b6d8 a1a7 a8a7 e3a7 b4c6 d1a1 e8g8 f1b1 c6a7 a1a7 d6d5 e4d5 e6d5 c4b3 d8c7 c3b5 c7c6 f3d4 c6c5 b3a4 g7g6 a7c5 e7c5 d4b3 c5b6 b5d4 g8g7 a4b5');
INSERT INTO evaluations VALUES(23726,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(23726,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(23726,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(23726,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(23726,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(23726,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(23726,36,'cp 104','c3d5 c6d4 d1d2 d4f3 g2f3 g4f3 h2h3 c7c6 d5f6 g7f6 g5h4 f3h5 g1h2 h5g6 b2b4 c5b6 c4b3 d6d5 d2h6 b6d4 a1c1 d4c3 f1g1 c3b4 e4d5 g8h8 d5c6 b7c6 c1b1 a8b8 b3f7 f8f7 g1g6 b4f8 b1b8 d8b8 h6h5 b8d6 c2c3 e5e4 h4g3 d6e6 d3e4 e6e4 g6g4');
INSERT INTO evaluations VALUES(23731,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(23731,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(23731,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(23731,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(23731,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(23731,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(23731,36,'cp 105','c6d4 d1d2 d4f3 g2f3 g4f3 h2h3 f3h5 b2b4 c5b6 g1h2 h5g6 d5f6 g7f6 g5h4 c7c6 d2h6 d6d5 e4d5 d8d6 f1g1 c6c5 b4b5 b6c7 a1f1 f6f5 h4g5 g8h8 f2f4 e5e4 d3e4 f5e4 g1g3 f8g8 g5h4 c7d8 h4d8 a8d8 f4f5 g6h5');
INSERT INTO evaluations VALUES(28802,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(28802,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(28802,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(28802,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(28802,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(28802,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(28802,36,'cp 35','g2g3 e7e6 f1g2 f8b4 c1d2 b4d2 d1d2 d7d5 c4d5 e6d5 b1c3 e8g8 e1g1 f8e8 b2b4 b8d7 d2b2 c7c6 f1c1 a7a5 b4a5 a8a5 a2a4 d8a8 f3d2 h7h6 d2b3 a5a6 e2e3 c6c5 c3b5 e8c8 d4c5 d7c5 b3c5 b6c5 b2a3 f6d7 a4a5 c5c4');
INSERT INTO evaluations VALUES(38810,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(38810,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38810,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38810,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38810,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38810,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(38810,36,'cp 16','b4e7 g1f3 g8f6 b1c3 e8g8 c4d5 e6d5 f1d3 f8e8 e1g1 e7d6 d1c2 b8a6 a2a3 c7c6 f1e1 h7h6 d3a6 b7a6 c3e2 a8b8 d2b4 d6b4 a3b4 b8b4 f3e5 d8b6 e1b1 c6c5 d4c5 b6c7 e5d3 b4c4 c2d1 c8f5 b2b3 c4c5 d3c5 f5b1 d1b1 c7c5 e2d4');
INSERT INTO evaluations VALUES(46805,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(46805,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(46805,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(46805,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(46805,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(46805,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(46805,36,'cp 89','e4e5 f6e4 c3e2 f7f6 f2f3 e4g5 c1g5 f6g5 h2h4 g5h4 f3f4 c7c5 g1f3 c8g4 f3h4 f8g7 d1d2 d8b6 c2c3 b8c6 d4c5 b6c5 e2d4 c6d4 c3d4 c5b6 a2a4 e8g8 a4a5 b6c7 f1d3 a8c8 e1f2 g6g5 h4f3 f8f4 h1h7 g4f5 h7h5 f5d3');
INSERT INTO evaluations VALUES(46813,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(46813,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(46813,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(46813,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(46813,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(46813,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(46813,36,'cp 64','f6e4 c3e2 f7f6 f2f3 e4g5 c1g5 f6g5 h2h4 g5h4 f3f4 c7c5 g1f3 b8c6 f3h4 f8g7 d4c5 d8a5 c2c3 a5c5 d1d2 c8g4 e2d4 c6d4 c3d4 c5b6 a2a4 e8g8 a4a5 b6e6 f1d3 b7b6 a5a6 b6b5 g2g3 a8c8 e1g1 g7h6 f1f2 e6b6 d2e3 e7e6 h4f3 g8h8 g1h2');
INSERT INTO evaluations VALUES(46851,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(46851,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(46851,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(46851,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(46851,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(46851,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(46851,36,'cp 71','c3e2 f7f6 c1e3 f8g7 f2f3 e4g5 d1d2 g5f7 e5f6 e7f6 h2h4 e8g8 e1c1 f8e8 g2g4 b8d7 e2g3 d7b6 h4h5 g6g5 h5h6 g7f8 d1e1 a7a5 g1h3 a5a4 a2a3 f7d6 d2f2 d6c4 g3h5 d8d6 f1d3 c4e3 e1e3 c8d7 e3e8 a8e8');
INSERT INTO evaluations VALUES(49842,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(49842,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(49842,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(49842,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(49842,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(49842,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(49842,36,'cp 142','a2a3 d7d6 d1c2 e4c5 e5d6 f8d6 g1f3 d8f6 b1c3 e8g8 e2e4 c8g4 c1g5 f6g6 g5e3 b8c6 e1c1 f7f5 e4e5 g4f3 g2f3 c6e5 d1d5 b7b6 b2b4 c5d7 c4c5 b6c5 b4c5 d7f6 c2a2 g8h8 c5d6 c7d6 f1a6 f5f4 e3d4 f8d8 a2c2 f6d5 c2g6 h7g6 c3d5 e5f3');
INSERT INTO evaluations VALUES(49843,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(49843,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(49843,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(49843,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(49843,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(49843,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(49843,36,'cp 155','d7d6 d1c2 e4c5 e5d6 f8d6 b1c3 e8g8 g1f3 d8f6 e2e4 b8c6 c1e3 c8g4 c3d5 f6g6 h2h3 g4f3 g2f3 f7f5 e4f5 g6f5 c2f5 f8f5 e1c1 c5e6 c1b1 f5f3 b2b4 f3f8 c4c5 d6e5 h3h4 g8h8 h4h5 a8d8 h5h6 g7h6 h1h6 e6d4 b4b5 d8d5 b5c6 b7c6 f1c4 d4f5');
INSERT INTO evaluations VALUES(51124,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(51124,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(51124,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(51124,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(51124,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(51124,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(51124,36,'cp 167','d1f3 a8a7 a4b5 c8b7 f3h3 g8f6 g1f3 a7a8 f1c4 e7e6 f3e5 d8c8 e1g1 f8e7 b5a6 b8a6 b1d2 a6b4 c4b5 c7c6 a1a8 c8a8 b5c4 b7a6 f1d1 a8b7 c4a6 b7a6 h3f3 e8g8 h2h3 c6c5 d4c5 e7c5 d2c4 a6a8 f3a8 f8a8 g1f1 c5e7');
INSERT INTO evaluations VALUES(6339,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(6339,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(6339,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(6339,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(6339,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(6339,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(6339,36,'cp 120','c1f4 d7d6 c2c3 g8e7 b1d2 e8g8 e1g1 e7g6 f4g5 d8e8 g1h1 c6e7 c4b3 c8e6 a2a4 c7c6 a4a5 b6d8 d4d5 c6d5 f3d4 g8h8 g5e7 d8e7 b3d5 e6d5 e4d5 g6e5 d1e2 e8d7 a5a6 a8e8 b2b4 h7h6 a6b7 d7b7 d4f5 h8g8 d2e4 b7d5');
INSERT INTO evaluations VALUES(6337,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(6337,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(6337,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(6337,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(6337,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(6337,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(6337,36,'cp 137','g8f6 b1c3 e8g8 d4d5 f8e8 d5c6 f6e4 e1f1 e4c3 b2c3 d8f6 d1d2 d7c6 f4g5 f6f5 c4d3 f5c5 g5h4 b6a5 h4f2 c5c3 d2c3 a5c3 a1b1 a7a6 h2h4 h7h6 h4h5 b7b5 h1h4 c8e6 a2a4 e6a2 b1c1 c3b2 c1e1 e8e1 f2e1 a2d5 h4f4 b2a3 e1b4 a3b4 f4b4 c6c5 b4b1 b5b4 f1f2');
INSERT INTO evaluations VALUES(9941,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(9941,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(9941,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(9941,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(9941,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(9941,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(9941,36,'cp 138','b1c3 e8g8 d4d5 f8e8 d5c6 f6e4 e1f1 e4c3 b2c3 d8f6 d1d2 d7c6 f4g5 f6f5 c4d3 f5c5 a1e1 c8f5 g5h4 b6a5 h4f2 f5d3 c2d3 c5c3 d2c3 a5c3 e1e3 c3b4 d3d4 b7b6 g2g4 f7f6 f1g2 e8e3 f2e3 a8e8 g2f2 g8f7 h1c1 f7e6 c1c4 b4d6 c4c6 e6d5 c6c3 a7a6 f2e2 b6b5');
INSERT INTO evaluations VALUES(9940,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(9940,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(9940,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(9940,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(9940,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(9940,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(9940,36,'cp 151','e8g8 d4d5 f8e8 d5c6 f6e4 e1f1 e4c3 b2c3 d8f6 d1d2 d7c6 f4g5 f6f5 c4d3 f5c5 a1e1 c8f5 g5h4 b6a5 h4f2 f5d3 c2d3 c5c3 d2c3 a5c3 e1e3 c3a5 d3d4 a5b6 g2g4 a8d8 e3e8 d8e8 g4g5 e8e4 h1g1 g8f8 h2h4 f7f6 h4h5 f8f7 g1g3 e4f4 f2e3 f4e4');
INSERT INTO evaluations VALUES(6727,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(6727,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(6727,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(6727,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(6727,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(6727,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(6727,36,'cp -72','g8f6 d2d4 f6e4 d4e5 e8g8 c4d5 e4c5 c1g5 c6e7 b1d2 h7h6 g5h4 d8e8 d2c4 a5b6 c4e3 d7d6 e5d6 c7d6 f1e1 c8e6 d5e6 f7e6 e3c4 e7f5 c4d6 f5d6 d1d6 e8a4 d6d1 f8f4 h4g3 a8d8 d1b1 f4e4 h2h3 a4c4 b1c2 b6a5 e1e4 c5e4 g3e5 c4d3 c2d3 d8d3 a1e1 e4c3 e5c3 a5c3');
INSERT INTO evaluations VALUES(10539,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(10539,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(10539,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(10539,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(10539,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(10539,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(10539,36,'cp -71','d2d4 f6e4 d4e5 e8g8 c4d5 e4c5 c1g5 c6e7 b1d2 h7h6 g5h4 d8e8 d2c4 a5b6 c4e3 d7d6 e5d6 c7d6 f1e1 c8e6 d5e6 f7e6 e3c4 e7f5 c4d6 f5d6 d1d6 e8a4 h4g3 a8d8 d6e5 d8d3 f3d4 f8f6 g3h4 f6f7 h4g3 a4c4 d4e6 d3g3 h2g3 c5d3 e5b8 g8h7 e6f8 f7f8');
INSERT INTO evaluations VALUES(10543,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(10543,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(10543,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(10543,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(10543,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(10543,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(10543,36,'cp -64','f6e4 d4e5 e8g8 c4d5 e4c5 c1g5 c6e7 b1d2 h7h6 g5h4 d8e8 d2c4 a5b6 c4e3 d7d6 e5d6 c7d6 f1e1 c8e6 d5e6 f7e6 e3c4 e7c8 c4d6 c8d6 d1d6 e8a4 d6d1 f8f4 h4g3 a8d8 d1b1 f4e4 h2h3 a4c4 b1c2 d8d3 g3e5 d3d5 e1e4 c5e4 a1f1 d5d3 a2a4');
INSERT INTO evaluations VALUES(9269,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(9269,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(9269,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(9269,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(9269,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(9269,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(9269,36,'cp 43','e8g8 e3e4 e6e5 d4d5 c6e7 h2h3 a7a6 a2a4 c7c6 c1e3 c6d5 c4d5 e7d5 c3d5 f6d5 d1d5 c8e6 d5b7 f7f6 f1c1 d8e7 b7e7 d6e7 c1c6 e6f7 f3h4 g7g6 c6c7 e7d6 c7c2 f8c8 a1c1 c8c2 c1c2 f7e8 b2b3 a6a5 h4f3 e8f7 c2c6 d6b4 c6f6 f7b3 f3e5');
INSERT INTO evaluations VALUES(3770,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(3770,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(3770,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(3770,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(3770,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(3770,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(3770,36,'cp 45','e3e4 e6e5 d4d5 c6e7 h2h3 h7h6 c1e3 f6h7 a1c1 g8h8 f3d2 a7a6 c4b3 f7f5 e4f5 e7f5 d2e4 c8d7 d1d2 h7f6 e4c5 f5e3 d2e3 d6c5 e3c5 d7f5 c3e2 f5d3 f1e1 a8c8 c5e3 d3e2 e1e2 f6d5 b3d5 d8d5 c1c5 d5d1 g1h2 d1d6 e3e5 d6e5 c5e5 c8d8');
INSERT INTO evaluations VALUES(3773,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(3773,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(3773,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(3773,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(3773,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(3773,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(3773,36,'cp 44','e6e5 d4d5 c6e7 h2h3 h7h6 c1e3 f6h7 f1e1 g8h8 c4b3 f7f5 e4f5 e7f5 b3c2 f5h4 f3h4 d8h4 c3e4 c8f5 e4d6 c7d6 c2f5 f8f5 a1c1 f5f7 d1d3 a8f8 c1c4 h4d8 a2a4 d8a5 e1d1 h7f6 f2f4 e5e4 d3d2 a5d2 d1d2 a7a6 d2d1');
INSERT INTO evaluations VALUES(14189,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14189,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14189,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14189,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(14189,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(14189,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(14189,36,'cp 35','d5c4 g1f3 h7h6 g5f6 e7f6 e2e3 f8d6 f1c4 c6e7 h2h3 c7c6 e1g1 e8g8 d1e2 f6f5 f1e1 c8d7 a1d1 e7g6 g2g3 f8e8 h3h4 g6f8 e2d3 d8c7 d4d5 a8d8 d5c6 d7c6 c3b5 c6b5 c4b5 d6b4 d3e2 b4e1 b5e8 d8e8 e2e1 f8e6 e1b4 a7a6');
INSERT INTO evaluations VALUES(14354,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(14354,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(14354,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14354,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14354,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(14354,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(14354,36,'cp 47','g1f3 h7h6 g5f6 e7f6 e2e3 f8d6 f1c4 c6e7 d1e2 e8g8 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 a1d1 g6h7 a2a3 c7c6 d4d5 e7d5 c4d5 c6d5 d1d5 d8b6 f1d1 f8d8 e2c4 d6f8 d5d7 d8d7 d1d7 b6b2 g1g2 h7g6 f3h4 b2c2 h4g6 c2g6 d7b7');
INSERT INTO evaluations VALUES(14353,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14353,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(14353,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(14353,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(14353,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(14353,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14353,36,'cp 39','h7h6 g5f6 e7f6 e2e3 f8d6 f1c4 e8g8 e1g1 c6e7 d1e2 c7c6 h2h3 f6f5 f1e1 e7g6 g2g3 f8e8 h3h4 g6f8 e2c2 d8f6 a1d1 c8d7 d4d5 c6c5 c3b5 f8g6 a2a4 d7b5 c4b5 g6e5 f3d2 e8e7 g1g2 a7a6 b5d3 g7g6 h4h5 a8d8');
INSERT INTO evaluations VALUES(14147,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14147,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(14147,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14147,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(14147,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14147,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(14147,36,'cp 45','g5f6 e7f6 e2e3 f8d6 f1c4 e8g8 e1g1 c6e7 d1e2 c7c6 h2h3 f6f5 f1d1 e7g6 d4d5 d8f6 d5c6 b7c6 c4b3 a8b8 a1c1 f8d8 d1d2 d8e8 e2d3 d6b4 d3d4 f5f4 d4f6 g7f6 e3f4 g6f4 c1d1 a7a5 f3d4 c8d7 a2a3 b4d6');
INSERT INTO evaluations VALUES(29696,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(29696,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(29696,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(29696,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(29696,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(29696,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(29696,36,'cp 24','c2c4 c7c6 f3e5 g4f5 c4d5 c6d5 d1a4 b8c6 b1c3 f5d7 e5d7 d8d7 e1g1 e7e6 d2d4 f8e7 c1d2 e8g8 f1c1 a8c8 e2e3 g7g6 h2h4 f8d8 a1b1 h7h5 d2e1 g8g7 a2a3 b7b6 c3b5 e7d6 b5d6 d7d6 e1d2 d6d7 a4b3');
INSERT INTO evaluations VALUES(43491,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(43491,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(43491,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(43491,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(43491,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(43491,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(43491,36,'cp -9','a2a4 f8e7 e1g1 e8g8 f1e1 b8c6 h2h3 d8c7 d3d4 c5d4 f3d4 c8d7 d4c6 d7c6 c1f4 f8d8 d1e2 h7h6 c4d3 c7a5 f4d2 a5e5 a4a5 d6d5 e4d5 e5e2 d3e2 f6d5 c3d5 d8d5 d2c3 a8c8 a1d1 d5d1 e1d1 e7g5 b2b3');
INSERT INTO evaluations VALUES(43503,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(43503,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(43503,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(43503,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43503,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(43503,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(43503,36,'cp -23','b8c6 e1g1 f8e7 h2h3 e8g8 f1e1 h7h6 c1f4 b7b6 c4b3 c8d7 c3e2 d8c7 e2g3 b6b5 c2c3 c6e5 d3d4 c5d4 c3d4 e5g6 f4d2 b5a4 b3a4 d7a4 a1a4 f8b8 d2c3 d6d5 e4e5 f6d7 g3e2 d7b6 a4a2 b6c4 d1c2 c7b7');
INSERT INTO evaluations VALUES(11256,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(11256,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(11256,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(11256,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(11256,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(11256,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(11256,36,'cp -7','e1g1 f8e7 h2h3 e8g8 f1e1 c8d7 d3d4 c5d4 f3d4 c6d4 d1d4 d7c6 c4d3 a8c8 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d4d5 c6d5 b2b3 f8d8 e1d1 d8d7 c2c4 d5c6 d3e2 c8d8 e3b6 d7d1 a1d1 d8d1 e2d1 e7a3 d1e2 f7f5 a4a5 g8f8 f2f4 f8e7');
INSERT INTO evaluations VALUES(11255,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(11255,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(11255,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(11255,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(11255,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(11255,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(11255,36,'cp -15','f8e7 f1e1 e8g8 h2h3 c8d7 d3d4 c5d4 f3d4 a8c8 c4f1 h7h6 d4c6 d7c6 a4a5 d6d5 e4d5 f6d5 c3d5 c6d5 c1f4 e7g5 f4g5 h6g5 f1d3 g7g6 c2c3 d8f6 d1e2 f8d8 a1d1 d5c6 d3c2');
INSERT INTO evaluations VALUES(7079,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(7079,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(7079,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(7079,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(7079,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(7079,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(7079,36,'cp -12','h2h3 e8g8 f1e1 c8d7 c1f4 d8c7 d1e2 h7h6 c3d1 a8b8 d1e3 b7b5 a4b5 a6b5 c4b3 b8a8 c2c3 c6e5 f3d2 a8a1 e1a1 c7b7 f4g3 f8a8 a1a8 b7a8 f2f4 a8a1 g1h2 e5c6 e4e5 d6e5 f4e5 f6h7 d2e4 h7f8 e3g4 f8g6 e2f2 b5b4 b3c4');
INSERT INTO evaluations VALUES(7077,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(7077,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(7077,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(7077,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(7077,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(7077,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(7077,36,'cp -26','e8g8 c1f4 b7b6 f1e1 c8b7 f4h2 d8c7 c3b1 c6a5 c4a2 b6b5 c2c3 c5c4 a4b5 a6b5 b1a3 c7b6 h2f4 b7c6 f3d4 b6b7 d4c6 b7c6 f4g3 d6d5 e4d5 c6d5 d3c4 d5d1 a1d1 e7a3 b2a3 a5c4 a2c4 b5c4 d1d4');
INSERT INTO evaluations VALUES(2534,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(2534,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(2534,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(2534,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(2534,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(2534,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(2534,36,'cp -16','f1e1 h7h6 c1f4 d8c7 c3b1 f6d7 b1d2 d7e5 f3e5 d6e5 f4e3 a8b8 c2c3 c8d7 d1e2 f8d8 e1f1 d7e8 g1h1 b7b5 a4b5 a6b5 c4a2 b5b4 a2b3 g8h8 f2f4 e5f4 e3f4 e7d6 f4e3 b4c3 b2c3 h8g8 d3d4 c5d4 c3d4 b8a8 e4e5');
INSERT INTO evaluations VALUES(56764,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(56764,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(56764,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(56764,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(56764,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(56764,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(56764,36,'cp 114','c2c4 f8d6 g1f3 g8f6 b1c3 e8g8 a1c1 b8c6 a2a3 c6e7 c4c5 d6f4 e3f4 c8d7 f1d3 d7e8 f3e5 f6d7 e5d7 e8d7 e1g1 e7c6 d3c2 b7b5 d1d2 d8f6 c1d1 a7a6 f1e1 g8h8 b2b4 f8b8 c3e2 a6a5 h2h3 a5b4 a3b4 a8a3');
INSERT INTO evaluations VALUES(16515,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(16515,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(16515,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(16515,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(16515,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16515,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(16515,36,'cp 0','f1g2 f8c5 g1e2 e8g8 d2d3 a7a5 e1g1 f8e8 h2h3 c6d4 c3a4 c5a7 e2d4 a7d4 c2c3 d4a7 c3c4 c7c6 a4c3 d7d6 g1h2 a7d4 a2a4 h7h6 c3e2 d4a7 e2c3');
INSERT INTO evaluations VALUES(25159,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25159,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25159,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25159,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25159,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25159,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25159,36,'cp 0','f6e4 h2h4 f8d6 d1c2 d6f4 e3f4 d8f6 c2e2 h7h5 e1g1 f6g6 d3b5 g6g4 e2e3 f7f6 f1e1 e4d6 b5c6 b7c6 b1d2 e8g8 b2b3 a7a5 a1c1 f8b8 a2a3 d6b5 c1a1 b5d6');
INSERT INTO evaluations VALUES(36552,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(36552,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(36552,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36552,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36552,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36552,36,'cp 1','b1c3 f8e8 d3e2 e7e6 e1g1 c7c5 a2a3 b7b6 b2b4 c8b7 c1b2 b8c6 d4c5 d6c5 b4c5 b6c5 d1c2 d8e7 h2h3 a8b8 a1b1 b7a8 f1d1 h7h6 c3a4 f6d7 b2g7 b8b1 d1b1 g8g7 a4c3 e8c8 b1d1 g7g8 c3e4 f7f5 e4d6 c8d8 c2a4 d7f6');
INSERT INTO evaluations VALUES(1346,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(1346,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(1346,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(1346,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(1346,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(1346,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(1346,36,'cp 70','e7e6 b5c6 b7c6 d2d3 g8e7 b2b3 b6c7 d1e2 e6e5 g2g3 h7h5 f3d2 d7d5 c1a3 c7a5 a3b2 a5c7 e1c1 c8e6 c3a4 e7g6 d1f1 d5d4 h2h4 a8b8 d2c4 e6c4 d3c4 f8d6 c1b1 g6f8 f2f4 f8e6 f4f5 e6f8 b2c1 f8d7 b1a1 f7f6 a4b2');
INSERT INTO evaluations VALUES(1321,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(1321,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(1321,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(1321,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(1321,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(1321,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(1321,36,'cp 67','b5c6 b7c6 d2d3 g8e7 b2b3 b6c7 d1e2 e6e5 g2g3 h7h5 c1a3 d7d6 e1c1 f7f6 f3d2 c8h3 f2f4 e8c8 d1g1 c8b8 c1b1 e7c8 c3d1 h3e6 a3b2 b8a8 h2h3 f8e7 f4e5 f6e5 g3g4 h5g4 h3g4 e7g5 d2c4 c8b6');
INSERT INTO evaluations VALUES(2324,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(2324,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(2324,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2324,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(2324,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(2324,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(2324,36,'cp 14','c8e6 c1g5 f8e7 e1c1 d8e8 g5e7 e8e7 c3d5 e6d5 d1d5 f7f6 d5d3 g8h6 d3g3 g7g6 g3b3 a8b8 f1c4 h6f7 c4f7 e7f7 h1d1 f7e7 a2a4 h7h5 a4a5 c7c6 b3c3 b8c8 c3g3 g6g5 g3d3 h8d8 d3d8 c8d8 d1d8 e7d8 c1d2 d8c7 d2d3 b7b6 a5b6 a7b6 d3c4 c7d6 c4b4 d6c7 g2g3 c7d6');
INSERT INTO evaluations VALUES(10541,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(10541,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(10541,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(10541,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(10541,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(10541,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(10541,36,'cp 71','c1g5 a5b6 b1d2 d7d6 e1g1 e8g8 a2a4 a7a6 g5h4 h7h6 b2b4 g7g5 h4g3 g5g4 f3h4 f6h5 c4b3 d8g5 d2c4 b6a7 b4b5 c6e7 b5b6 a7b6 c4b6 c7b6 d3d4 h5f4 f2f3 c8e6 f3g4 e6g4 h4f3 g5g6 g3f4 e5f4 f1e1 b6b5 a4b5 a6b5 h2h3');
INSERT INTO evaluations VALUES(11538,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(11538,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(11538,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(11538,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(11538,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(11538,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(11538,36,'cp -26','f6e4 d1e2 e4c5 f3e5 d8e7 a4c6 d7c6 d2d4 c5d7 c1f4 d7e5 e2e5 e7e5 f4e5 e8d8 f2f3 f7f6 e5f4 c8e6 h2h4 h7h5 e1f2 a6a5 b1d2 b7b6 d2e4 d8c8 a2a4 c8b7 h1e1 e6f7 e1e3 a8d8 e3e2 d8d7 a1e1 f7b3 e1a1 b3d5 e2e3 d5f7');
INSERT INTO evaluations VALUES(11564,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(11564,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(11564,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(11564,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(11564,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(11564,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(11564,36,'cp -20','d1e2 e4c5 f3e5 d8e7 a4c6 d7c6 d2d4 c5d7 c1f4 d7e5 e2e5 e7e5 f4e5 e8d8 f2f3 f7f6 e5g3 c8e6 b1d2 h7h5 h2h4 b7b6 d2e4 e6f7 e1f2 a6a5 h1e1 d8c8 g3f4 c8d7 a2a4 a8d8 e1e2 d7c8 g2g3 d8d7 e2e3 c8b7 e3e2 f8e7 e2e1 h8e8 e1e3 f7b3');
INSERT INTO evaluations VALUES(15096,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(15096,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(15096,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(15096,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(15096,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(15096,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(15096,36,'cp 14','c4d5 f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 c8d7 e1g1 e8g8 b5d3 f8e8 e3e4 d7g4 d4d5 c6a5 h2h3 g4d7 c1e3 b7b6 f3d2 a5b7 d2b3 a7a5 a2a4 d8e7 d1e2 f7f5 d3a6 a8b8 e4f5 e5e4 b3d4 e7e5 g2g3');
INSERT INTO evaluations VALUES(14998,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(14998,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(14998,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(14998,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(14998,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(14998,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14998,36,'cp 13','f6d5 f1b5 d5c3 b2c3 f8d6 d2d4 c8d7 e1g1 e8g8 b5c6 d7c6 d4e5 d6c5 f3d4 d8e7 d4c6 b7c6 f2f4 f7f6 e5f6 f8f6 g1h1 a8e8 h2h3 h7h6 d1b3 g8h8 a1b1 c5e3 f1e1 f6e6 c1e3 e6e3 e1e3 e7e3 b3f7 e3e7 f7e7 e8e7 b1b8 h8h7 b8b7 a7a5 h1h2 e7e3 f4f5 e3c3 b7c7 h7g8');
INSERT INTO evaluations VALUES(15341,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(15341,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(15341,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(15341,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(15341,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(15341,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(15341,36,'cp 12','f1b5 d5c3 b2c3 f8d6 d2d4 c8d7 e1g1 e8g8 b5d3 f8e8 e3e4 d7g4 c1e3 e5d4 c3d4 c6b4 d3b1 g4f3 d1f3 c7c5 e4e5 c5d4 e5d6 d4e3 f2e3 d8d6 f3f7 g8h8 b1h7 h8h7 f1f4 d6e7 f7e7 e8e7 f4b4 a7a5 b4h4 h7g8 a1b1 a8c8 e3e4 c8c2 a2a4 c2a2 e4e5');
INSERT INTO evaluations VALUES(15243,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(15243,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15243,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(15243,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(15243,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(15243,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(15243,36,'cp 21','d5c3 b2c3 f8d6 d2d4 c8d7 e1g1 e8g8 b5d3 f7f5 e3e4 d8f6 f1e1 f5e4 d3e4 e5d4 c3d4 a8e8 c1g5 f6f7 g5h4 g8h8 h4g3 d7f5 e4f5 e8e1 d1e1 f7f5 h2h3 a7a5 e1c3 f5d5 a1b1 h7h6 g3d6 c7d6 c3b3 d5b3 b1b3 f8f7 b3b5 h8g8 a2a3 g7g5');
INSERT INTO evaluations VALUES(21780,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(21780,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(21780,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(21780,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(21780,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(21780,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(21780,36,'cp 65','g8e7 f1g2 e7g6 e1g1 g6e5 f3e5 c6e5 d1a4 c8d7 a4b3 f8c5 c1f4 e5c6 b3b7 a8b8 b7c7 d8c7 f4c7 b8b2 c7f4 c5b4 f1d1 b2e2 b1d2 e8g8 g2f3 d4d3 d2b3 c6e5 f4e5 e2e5 d1d3 d7h3 f3g2 h3g4 d3e3 e5e3 f2e3 f8e8 c4c5 e8e3 c5c6');
INSERT INTO evaluations VALUES(14058,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(14058,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(14058,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(14058,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(14058,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(14058,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(14058,36,'cp 70','f1g2 e7g6 e1g1 g6e5 f3e5 c6e5 d1a4 c8d7 a4b3 f8c5 b3b7 a8b8 b7e4 d8e7 c1f4 e5g6 e4e7 c5e7 f4c1 e8g8 b1d2 f8e8 f1d1 c7c5 b2b3 a7a5 g2d5 a5a4 d2e4 e7f8 f2f3 e8c8 c1g5 d7c6 b3a4 c6d5 c4d5 f7f5 e4d2 c5c4 a1c1');
INSERT INTO evaluations VALUES(14059,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(14059,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(14059,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14059,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(14059,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(14059,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(14059,36,'cp 82','e7g6 e1g1 g6e5 f3e5 c6e5 d1a4 c8d7 a4b3 f8c5 b3b7 a8b8 b7d5 d8e7 c1f4 f7f6 f4e5 e7e5 b2b3 d7g4 d5e5 f6e5 g2f3 g4h3 f1d1 e8e7 f3e4 a7a5 b1d2 c5b4 e4d3 g7g6 d2e4 h8d8 f2f3 h3f5 g1f2 b4a3 f2g2 f5e4 d3e4 a3b4 a1b1 b4c3 a2a3 c7c5');
INSERT INTO evaluations VALUES(15240,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(15240,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(15240,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(15240,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(15240,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(15240,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(15240,36,'cp 62','e1g1 g6e5 f3e5 c6e5 d1a4 c8d7 a4b3 f8c5 b3b7 a8b8 b7e4 d8e7 c1f4 e5g6 e4e7 c5e7 f4c1 e8g8 b1d2 f8e8 d2b3 e7f6 e2e3 d4e3 c1e3 b8b4 e3a7 f6b2 a1b1 b2a3 b3c5 b4b1 f1b1 d7f5 b1a1 e8d8 c5b7 d8a8 b7d6 a8a7 d6f5 a7a4 g2d5 a3c5 g1g2 h7h5 h2h4 g6e5 a1e1');
INSERT INTO evaluations VALUES(26615,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26615,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(26615,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(26615,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26615,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(26615,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(26615,36,'cp 19','g4f3 d1f3 g8f6 f1b5 d8d7 e1g1 e7e6 c2c3 f8d6 b1d2 a7a6 b5a4 e8g8 e3e4 e6e5 d4e5 c6e5 f3d1 b7b5 a4c2 d5e4 d2e4 f6e4 c2e4 a8d8 a2a4 h7h6 d1c2 f8e8 f1d1 d7e6 e4d5 e6e7 c1e3 e5c4 d5c4 b5c4 d1d4 e7e5 e3f4 e5e2 c2e2 e8e2 f4d6 c7d6');
INSERT INTO evaluations VALUES(26629,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(26629,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(26629,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(26629,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26629,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(26629,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26629,36,'cp 27','d1f3 g8f6 f1b5 d8d7 e1g1 e7e6 b1d2 a7a6 b5a4 f8d6 c2c3 e8g8 e3e4 e6e5 d4e5 c6e5 a4d7 e5f3 d2f3 d5e4 d7a4 e4f3 g2g3 f6e4 a4d1 f8e8 d1f3 d6c5 g1g2 c7c6 a2a4 a8d8 b2b4 c5b6 c3c4 e4d2 c1d2 d8d2 c4c5 b6a7 f1d1 e8d8');
INSERT INTO evaluations VALUES(26628,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(26628,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26628,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26628,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(26628,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(26628,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(26628,36,'cp 21','g8f6 f1b5 d8d7 b5a4 a7a6 e1g1 e7e6 c2c3 f8d6 b1d2 e8g8 e3e4 e6e5 d4e5 c6e5 f3d1 b7b5 a4c2 f6e4 d2e4 d5e4 c2e4 a8e8 a2a4 h7h6 e4c2 d7c6 d1h5 e5g6 a4b5 a6b5 c1e3 e8e5 h5g4 h6h5 g4f3 c6f3 g2f3 e5e8 c2e4 d6f4 e3c5 f4d6 c5d6 c7d6 e4g6');
INSERT INTO evaluations VALUES(24967,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(24967,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(24967,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24967,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24967,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24967,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24967,36,'cp 22','f1b5 d8d7 b5a4 a7a6 e1g1 e7e6 b1d2 f8d6 c2c4 b7b5 c4b5 c6b4 d2c4 a6b5 c4d6 d7d6 a4b3 e8g8 f3e2 f6e4 c1d2 e4d2 e2d2 b4a6 f1c1 a8b8 b3d1 c7c5 d4c5 a6c5 d2c2 c5e4 d1f3 e4f6 a2a3 g7g6 c2c7 d6c7 c1c7 f8c8 a1c1 c8c7 c1c7 g8g7 g1f1 f6e8 c7c5 e8d6 f3e2 b8d8');
INSERT INTO evaluations VALUES(34306,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(34306,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(34306,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(34306,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(34306,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(34306,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(34306,36,'cp 115','g1f3 e5c7 d2d4 c8f5 e1g1 b8d7 d4d5 e7e5 d5e6 f5e6 f3g5 e8c8 e2g4 e6g4 d1g4 g8h6 g4h3 c7e5 c1d2 e5d6 a1d1 d6g6 f1e1 f8d6 g5e4 d6e5 d2g5 f7f6 g5h6 g6h6 e4d6 e5d6 h3h6 g7h6 d1d6 h8e8 g1f1 e8e1 f1e1 d8e8 c3e2 e8e7 e1f1 c8c7 d6d4 d7e5 e2f4');
INSERT INTO evaluations VALUES(34305,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(34305,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(34305,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(34305,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(34305,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(34305,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(34305,36,'cp 115','e5a5 e1g1 c8f5 d2d4 g8f6 f3h4 f5e6 f1e1 b8d7 a2a4 g7g6 h4f3 f8g7 f3g5 e6d5 c3d5 c6d5 e2d3 e8g8 c2c3 f8e8 g5f3 a7a6 g2g3 a8c8 g1g2 a5c7 c1f4 c7b6 e1e2 b6c6 a4a5 f6h5 f4d2 e7e5 d4e5 d7e5 f3e5 e8e5 e2e5 g7e5 d1f3 h5f6');
INSERT INTO evaluations VALUES(34328,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(34328,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(34328,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(34328,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(34328,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(34328,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(34328,36,'cp 112','d2d4 c8f5 e1g1 g8f6 f3h4 f5e6 f1e1 b8d7 a2a4 g7g6 h4f3 f8g7 f3g5 e6d5 c3d5 c6d5 c1f4 e8g8 c2c3 a7a6 h2h3 a5b6 d1d2 d7b8 h3h4 b8c6 b2b4 h7h6 a4a5 b6d8 g5f3 f6e4 d2e3 e7e6 g2g3');
INSERT INTO evaluations VALUES(41269,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41269,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(41269,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(41269,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(41269,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(41269,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(41269,36,'cp 14','b1c3 c8f5 e2e3 f6e4 c3e4 f5e4 f1d3 e4f3 d1f3 c7c6 e1c1 b8d7 h2h4 e7e5 f4h2 d8h4 h2e5 d7e5 h1h4 e5f3 g2f3 h7h5 f3f4 g7f6 h4h3 e8e7 d1h1 h5h4 c1d2 a7a5 d2e2 a5a4 h1b1 a8a5 d3e4 c6c5 e2d3 c5d4 e4b7 h8b8');
INSERT INTO evaluations VALUES(46322,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(46322,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(46322,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(46322,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(46322,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(46322,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(46322,36,'cp 64','b2b3 c4b3 a2b3 a7a5 f1b5 c8d7 b5d3 f8e7 h2h3 e8g8 e1g1 d7c6 f3e5 c6e8 c3c4 b8c6 b1c3 c6b4 d3b1 b7b6 d1e2 f6d7 e5f3 d7f6 f1d1 e8c6 e3e4 d5e4 c3e4 f6e4 b1e4 c6e4 e2e4 a8c8 d1d2 h7h6 f3e5 e7d6 a1d1');
INSERT INTO evaluations VALUES(46406,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(46406,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(46406,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(46406,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(46406,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(46406,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(46406,36,'cp 63','f8e7 f1b5 b8c6 d4c5 e8g8 e1g1 e7c5 g5f6 d8f6 c3d5 f6d6 d5c3 c6e5 b5e2 d6c7 f3e5 c7e5 e2f3 c8e6 f1e1 e5f4 c3d5 f4f5 c2c4 a8d8 d1b3 b7b6 a1d1 e6d5 c4d5 f8e8 e1e8 d8e8 b3a4 f5e5 g2g3 c5d6 a4a7 e5b2 a7d7 b2e5 g1g2 g7g6');
INSERT INTO evaluations VALUES(40247,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(40247,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(40247,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(40247,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(40247,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(40247,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(40247,36,'cp 88','f1b5 b8c6 e1g1 e8g8 d4c5 e7c5 g5f6 d8f6 c3d5 f6d6 d5c3 c6e5 b5e2 d6c7 c3d5 c7d6 c2c4 c8e6 d1b3 e6d5 c4d5 f8e8 a1c1 a8d8 f3e5 e8e5 e2f3 b7b6 g2g3 a7a5 f1d1 g7g6 g1g2 d6e7 h2h4 h7h5 a2a3 d8d7 b3c3 e5f5 c1c2 e7f6');
INSERT INTO evaluations VALUES(53105,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53105,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53105,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(53105,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(53105,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(53105,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53105,36,'cp 23','e7e5 c2c4 b8c6 b1c3 g7g6 a2a3 f8g7 e1g1 g8e7 b2b4 e8g8 a1b1 f7f5 f3e1 a8b8 d2d3 b7b6 e1c2 g8h8 c1g5 h7h6 g5e7 c6e7 c2e3 c8d7 e3d5 e7d5 c3d5 h6h5 a3a4 h5h4 b4b5 d7e6 d1d2 e6f7 f2f4 e5f4 d5f4 d8e7');
INSERT INTO evaluations VALUES(56759,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(56759,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(56759,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(56759,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(56759,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(56759,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(56759,36,'cp 124','g1f3 g8f6 c1f4 f8d6 e2e3 e8g8 a1c1 b8c6 a2a3 c8d7 f1e2 c6e7 e1g1 d7e8 d1b3 a7a5 f3e5 a5a4 b3c2 f6e4 f1d1 h7h6 h2h3 e4c3 c2c3 d5c4 c3c4 e7d5 e2f3 c7c6 c4c2 d8e7 d1e1 d5f4 e3f4 g8h8 g2g3 a8d8');
INSERT INTO evaluations VALUES(56758,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(56758,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(56758,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(56758,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(56758,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(56758,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(56758,36,'cp 123','g8f6 c1f4 f8d6 e2e3 e8g8 a1c1 a7a6 f4d6 d8d6 f1d3 c7c6 e1g1 f6e4 c3e2 b8d7 e2f4 a6a5 a2a3 d6e7 h2h4 d7f6 d1b3 a5a4 b3b4 e7b4 a3b4 a4a3 b2a3 a8a3 c1a1 a3b3 f1b1 b3b1 a1b1 d5c4 d3c4 e4c3 b1b3');
INSERT INTO evaluations VALUES(49113,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(49113,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(49113,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(49113,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(49113,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(49113,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(49113,36,'cp 120','c1f4 f8d6 e2e3 e8g8 a1c1 b8c6 a2a3 c8d7 f1e2 d7e8 f3g5 e8d7 e1g1 c6e7 g5f3 d7e8 d1b3 a8b8 f1e1 c7c6 f4d6 d8d6 e2d3 e7g6 b3b4 d6b4 a3b4 d5c4 d3c4 e8f7 f3g5 f6d5 b4b5 g6e7 c4d5 e7d5 b5c6 b7c6 c3d5 e6d5 c1c6');
INSERT INTO evaluations VALUES(56796,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(56796,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(56796,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(56796,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(56796,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(56796,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(56796,36,'cp 94','c2c4 f8b4 f4d2 c7c5 e2e3 g8e7 d2b4 c5b4 f1d3 d5c4 d3c4 b8c6 e1g1 e8g8 b1d2 c6a5 c4d3 c8d7 d2e4 b7b6 e4g3 a8c8 h2h4 d7c6 h4h5 h7h6 e3e4 f6f5 e4f5 e6f5 a1c1 c6f3 d1f3 c8c1 f1c1 d8d4');
INSERT INTO evaluations VALUES(8429,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(8429,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(8429,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(8429,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(8429,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(8429,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8429,36,'cp 54','e8g8 a1c1 f6e4 h4e7 d8e7 c4d5 e4c3 c1c3 e6d5 f1d3 e7d6 h2h3 c6e7 e1g1 c7c6 d1b1 a7a5 a2a3 c8e6 f1c1 a5a4 f3e5 g7g6 b2b4 a4b3 c3b3 b7b5 b3c3 e6d7 b1b4 d6b4 a3b4 a8a7 h3h4 g8g7 e5d7 a7d7 h4h5 f8a8 h5g6');
INSERT INTO evaluations VALUES(3296,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(3296,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(3296,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(3296,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(3296,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(3296,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(3296,36,'cp 47','a1c1 f6e4 h4e7 c6e7 f1e2 b7b6 e1g1 e4c3 c1c3 d5c4 c3c4 a7a5 a2a3 c8b7 h2h3 e7d5 c4c2 d8d6 d1d2 d5f6 f1c1 f6e4 d2e1 c7c5 d4c5 e4c5 f3d4 a8c8 b2b4 a5b4 a3b4 c5e4 c2c8 f8c8 c1c8 b7c8');
INSERT INTO evaluations VALUES(17788,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(17788,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(17788,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(17788,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(17788,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(17788,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(17788,36,'cp 70','d2d4 c8g4 h2h3 g4f3 d1f3 d8f6 f3f6 g8f6 d4d5 c6d4 c4d3 f6d7 c1e3 g7g6 a2a4 h7h5 a4a5 f8h6 b2b4 c7c5 b4c5 d7c5 e3d4 e5d4 c3a4 e8c8 e1g1 h8e8 a4c5 d6c5 a1a4 e8e7 g2g3 h5h4 a4c4 h6d2 c4c5 e7c7 c5c7 c8c7 f1a1 d2c3 a1a3 f7f5 g1g2 c3b4 a3a4 f5e4 d3e4 h4g3');
INSERT INTO evaluations VALUES(17794,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(17794,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(17794,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(17794,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(17794,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(17794,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(17794,36,'cp 67','c8g4 h2h3 g4f3 d1f3 d8f6 f3f6 g8f6 d4d5 c6d4 c4d3 g7g6 c1e3 f6d7 a2a4 h7h5 a4a5 f8h6 b2b4 c7c5 b4c5 d7c5 e3d4 e5d4 c3a4 e8c8 e1g1 h8e8 a4c5 d6c5 a1a4 h6d2 f1b1 e8e7 g2g3 d2b4 g1g2 h5h4 b1b4 c5b4 a4b4 h4g3 f2g3 e7c7 b4d4 c7c5');
INSERT INTO evaluations VALUES(20031,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(20031,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(20031,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(20031,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(20031,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(20031,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(20031,36,'cp 38','g1e2 g8f6 e1g1 c8d7 h2h3 h7h6 a2a4 e7e6 c2c4 e6e5 d2d3 f8e7 b1c3 a8b8 b2b3 a7a6 f2f4 e8g8 g1h2 f8e8 d1d2 c6a5 d2c2 a5c6 c2d1 c6b4 c1b2 e8f8 d1d2 d8c7 a4a5 b7b5 a5b6 b8b6 a1e1 f8b8 b2a1');
INSERT INTO evaluations VALUES(20032,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(20032,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(20032,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20032,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(20032,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(20032,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(20032,36,'cp 45','g8f6 e1g1 c8d7 h2h3 h7h6 d2d4 c5d4 e2d4 g7g6 c1e3 f8g7 d1d2 a8c8 f1d1 a7a6 b1c3 c6e5 b2b3 d8a5 c3e2 a5d2 d1d2 b7b5 a2a4 b5b4 c2c4 b4c3 d2c2 e5d3 c2c3 c8c3 e2c3 e8g8 a4a5 d3c5 g3g4 c5e6 d4e6 d7e6');
INSERT INTO evaluations VALUES(13335,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(13335,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(13335,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(13335,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(13335,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(13335,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(13335,36,'cp 54','e1g1 c8d7 h2h3 e7e6 d2d4 c5d4 e2d4 f8e7 c2c4 a7a6 d4c6 d7c6 b1c3 e8g8 d1e2 f6d7 c1e3 b7b6 a1c1 d8b8 e3f4 f8c8 f1d1 d7e5 a2a4 b8b7 b2b3 a8b8 h3h4 h7h6 f4e3 e7f8 h4h5 f8e7 f2f4 e5d7 b3b4');
INSERT INTO evaluations VALUES(21310,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(21310,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(21310,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(21310,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(21310,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(21310,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(21310,36,'cp 49','d4d3 g8f6 b1c3 f8e7 c1f4 e8g8 e1c1 c8e6 h1e1 f6g4 f4g3 g4e5 g3e5 c6e5 f3e5 d6e5 c4e6 f7e6 f2f3 a7a6 d3d7 d8d7 d1d7 f8f7 d7d3 c7c5 e1d1 a8f8 d3d7 b7b5 b2b3 c5c4 c1b2 c4b3 b2b3 e7c5 d7f7 g8f7');
INSERT INTO evaluations VALUES(55132,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55132,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(55132,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(55132,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(55132,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(55132,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(55132,36,'cp -15','e7e5 d2d4 e5e4 c2c4 c7c6 b1c3 g8f6 c1d2 f8e7 g1e2 b8a6 c4d5 c6d5 d1b3 a6c7 e2f4 h7h5 c3b5 c7b5 b3b5 e8f8 d2b4 g7g5 b4e7 f8e7 b5b4 d8d6 b4d6 e7d6 f4e2 h5h4 e2c3 f6e8 a1c1 a7a6 c3a4 a8b8 f1e2 f7f5 f2f3 e8f6 a4c5 f5f4 e1g1 e4f3');
INSERT INTO evaluations VALUES(6399,30,'cp -370',NULL);
INSERT INTO evaluations VALUES(6399,31,'cp -377',NULL);
INSERT INTO evaluations VALUES(6399,32,'cp -360',NULL);
INSERT INTO evaluations VALUES(6399,33,'cp -351',NULL);
INSERT INTO evaluations VALUES(6399,34,'cp -367',NULL);
INSERT INTO evaluations VALUES(6399,35,'cp -379',NULL);
INSERT INTO evaluations VALUES(6399,36,'cp -374','c6e5 c4e2 e5c6 c2c3 c5b6 d2d4 d7d5 e4d5 d8d5 e1g1 c8e6 f1e1 d5d7 b1d2 g8f6 d2f3 e8g8 a2a4 a7a5 e2b5 e6g4 h2h3 g4h5 d1d3 a8e8 c1g5 h5g6 d3f1 d7f5 e1e8 f8e8 g5h4 g6h5 f3e5 e8e5 b5c6 e5e7 c6b7 h5g6 b7f3 e7e8 h4g3 h7h6');
INSERT INTO evaluations VALUES(6840,30,'cp -372',NULL);
INSERT INTO evaluations VALUES(6840,31,'cp -373',NULL);
INSERT INTO evaluations VALUES(6840,32,'cp -368',NULL);
INSERT INTO evaluations VALUES(6840,33,'cp -367',NULL);
INSERT INTO evaluations VALUES(6840,34,'cp -384',NULL);
INSERT INTO evaluations VALUES(6840,35,'cp -363',NULL);
INSERT INTO evaluations VALUES(6840,36,'cp -373','c4e2 c5b6 d2d4 e5c6 c2c3 d7d5 e4d5 d8d5 e1g1 c8e6 b1d2 g8f6 f1e1 e8g8 d2f3 d5d7 a2a4 f8e8 b2b4 a7a5 b4b5 c6e7 c3c4 e7g6 c1e3 a8d8 c4c5 b6a7 d1c2 e6d5 a1d1 d5e4 c2c1 d7f5 f3g5 e4d5 e2d3 f5g4 f2f3');
INSERT INTO evaluations VALUES(16954,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(16954,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(16954,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(16954,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(16954,34,'cp 252',NULL);
INSERT INTO evaluations VALUES(16954,35,'cp 278',NULL);
INSERT INTO evaluations VALUES(16954,36,'cp 263','f2f4 e5g6 e4e5 f6g8 d1e2 g8h6 c1d2 f8c5 d4b3 c5b6 e1c1 e8g8 g2g4 d7d6 h2h3 c7c6 e5d6 d8d6 f4f5 h6f5 g4f5 c8f5 e2h5 d6f6 c1b1 g6e5 f1e2 f5g6 h5g5 f6g5 d2g5 f7f6 g5d2 a8d8 a2a4 f8e8 d2f4 e5f7 a4a5 b6f2 d1d8 f7d8');
INSERT INTO evaluations VALUES(19322,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(19322,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(19322,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(19322,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(19322,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(19322,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(19322,36,'cp 203','d4d5 c6b8 d1d4 g8f6 e4e5 f6g8 d4a4 f8g7 g1f3 e8f8 f1e2 d7d6 e5d6 d8d6 b1c3 b8a6 e1g1 g8f6 c3b5 c8d7 b5d6 d7a4 d6b7 a6b4 d5d6 e7d6 b7d6 a8d8 d6b7 d8e8 e2c4 b4c2 a1b1 f6e4 b2b3 a4c6 b7a5 c6a8 c4d3 c2d4 c1a3 f8g8 f1e1 d4f3 g2f3 g7c3 f3e4');
INSERT INTO evaluations VALUES(19301,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(19301,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(19301,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(19301,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(19301,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(19301,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(19301,36,'cp 177','c6b8 d1d4 g8f6 e4e5 f6g8 g1f3 f8g7 d4a4 e8f8 b1c3 d7d6 e5d6 d8d6 f1b5 b8a6 e1g1 d6b4 a4c2 c8g4 a2a3 b4d6 b5e2 g8f6 f1d1 g4f5 e2d3 f5g4 c2e2 h7h5 h2h3 g4f3 e2f3 g7h6 d3a6 b7a6 c3e4 f6e4 f3e4 f8g7 c1h6 g7h6');
INSERT INTO evaluations VALUES(27161,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(27161,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(27161,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(27161,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(27161,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(27161,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(27161,36,'cp 53','b7c6 d2d4 f7f6 e1g1 g7g6 c1e3 g8e7 h2h3 f8g7 d1d2 h7h6 c3e2 g6g5 e2g3 e8g8 a1e1 d7e6 e1d1 d8c8 f1e1 a7a5 h3h4 e5d4 e3d4 e6g4 d4c3 a5a4 d2d3 e7g6 g3f5 g6f4 d3c4 c8e6 c4e6 f4e6 f5e7 g8f7');
INSERT INTO evaluations VALUES(29022,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29022,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29022,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(29022,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29022,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(29022,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(29022,36,'cp 0','e7e6 d1b3 d8c7 h2h3 g4f5 f3h4 f5e4 b1c3 b8d7 h4f3 e4f5 f3h4');
INSERT INTO evaluations VALUES(29280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29280,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(29280,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(29280,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29280,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(29280,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29280,36,'cp 6','h2h3 g4h5 d1b3 d8b6 c1d2 b8d7 b1c3 f8e7 c4d5 f6d5 e1g1 e8g8 f1c1 h5g6 f3e1 a7a5 c3d5 e6d5 b3b6 d7b6 a2a4 f8e8 b2b3 b6d7 e2d3 f7f6 d3g6 h7g6 e1d3 g8f8 g1f1 e7d6 f1e2 g6g5 b3b4 d6b4 d2b4 a5b4');
INSERT INTO evaluations VALUES(33740,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(33740,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(33740,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(33740,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(33740,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(33740,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(33740,36,'cp 36','c7c6 d2d4 c8f5 e1g1 e7e6 c1f4 b8d7 c3b1 f8e7 b1d2 e8g8 a2a4 f6h5 d2c4 a5d8 f4c1 h5f6 c1g5 d7b6 c4e5 a7a5 c2c3 h7h6 g5h4 f6e4 h4e7 d8e7 e2d3 e4d6 d1c2 e7c7 h2h3 b6d7 f1e1 f5d3 e5d3 c6c5 d3c5 d7c5');
INSERT INTO evaluations VALUES(33594,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(33594,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(33594,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(33594,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(33594,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(33594,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(33594,36,'cp 39','e1g1 c8f5 f3h4 f5g6 g2g3 b8d7 d2d3 e7e6 h4g6 h7g6 h2h4 f8c5 e2f3 e8g8 d1e2 a5c7 g1g2 d7e5 c1f4 c5d6 c3e4 f6e4 f3e4 f7f5 e4f3 a8e8 f4e5 d6e5 c2c3 e5f6 d3d4 e6e5 d4e5 e8e5 e2d3 e5e8 f1e1 e8d8 d3c2 c7f7 a2a4');
INSERT INTO evaluations VALUES(35324,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(35324,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(35324,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(35324,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(35324,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(35324,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(35324,36,'cp 100','f1d3 g8f6 h2h3 e7e6 d1e2 f8d6 g1f3 b8d7 e1c1 b7b5 c3e4 f6e4 e2e4 d7f6 e4h4 f6d5 f3g5 h7h6 g5e4 d6f4 e4c5 a7a5 a2a3 e8g8 h1e1 c7d8 h4d8 f4d2 d1d2 f8d8 g2g3 g8f8 a3a4 a8b8 c2c3 b5a4 c5a4 c8d7');
INSERT INTO evaluations VALUES(36219,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36219,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36219,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36219,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36219,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36219,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36219,36,'cp 0','c7c6 a2a4 c8e6 c4e6 f7e6 d3d4 e5d4 f3d4 d8d7 d4f3 e6e5 d1e2 b8a6 c1e3 a6c7 a1d1 d7e6 f3g5 e6d7');
INSERT INTO evaluations VALUES(35787,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35787,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35787,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35787,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35787,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35787,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35787,36,'cp 0','a2a4 c8e6 c4e6 f7e6 d3d4 e5d4 f3d4 d8d7 d4f3 e6e5 d1e2 b8a6 c1e3 a6c7 a1d1 d7e6 f3g5 e6d7');
INSERT INTO evaluations VALUES(46956,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(46956,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(46956,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(46956,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(46956,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(46956,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(46956,36,'cp 45','c7d6 c2c4 d5f6 b1c3 g7g6 g2g3 d6d5 c4d5 f6d5 f1g2 d5c3 b2c3 f8g7 g1f3 c8e6 e1g1 e8g8 f1e1 e6d5 d1d3 f8e8 c1f4 b8d7 a1c1 a8c8 h2h4 d7b6 h4h5 d5c4 d3b1 c4d5 h5h6 g7f6 a2a4 d8d7 b1b5 e8d8 f3d2 d5g2 g1g2');
INSERT INTO evaluations VALUES(45277,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(45277,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(45277,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(45277,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45277,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(45277,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(45277,36,'cp 39','c2c4 d5f6 g1f3 g7g6 b1c3 f8g7 f1e2 d6d5 e1g1 e8g8 f1e1 b8c6 h2h3 d5c4 e2c4 e7e6 c1g5 h7h6 g5f4 a7a6 a2a4 b7b6 d4d5 c6a5 c4a2 e6d5 c3d5 f6d5 a2d5 c8b7 d5b7 d8d1 e1d1 a5b7 f4c7 b6b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(48046,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(48046,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(48046,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(48046,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(48046,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(48046,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(48046,36,'cp 63','g2g3 f8e7 f1g2 d6d5 e1g1 b8d7 d1c2 e8g8 f1d1 a7a5 b1c3 c7c6 b2b3 b7b5 a2a3 a5a4 c4b5 c6b5 c3b5 d8a5 b3a4 a5a4 c2a4 a8a4 b5c3 a4c4 c1d2 c8a6 e2e3 f6e4 c3e4 d5e4 f3e1 f7f5 g2f1 c4a4 f1a6 a4a6 d2b4 e7b4 a3b4 a6a1 d1a1 f8b8 a1b1 g8f7 b4b5');
INSERT INTO evaluations VALUES(48085,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(48085,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(48085,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(48085,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(48085,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(48085,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(48085,36,'cp 41','d5c3 b2c3 c7c5 e2e4 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 d7f6 c4d3 h7h6 d2e3 b7b6 e4e5 f6d5 e3e4 g7g6 h2h4 c8b7 h4h5 d8e7 e4g4 d5b4 d3b1 b7f3 g4f3 a8d8 b1e4 d8d4 a1d1 d4d1 f3d1 f8d8 d1f3 e7g5 h5g6');
INSERT INTO evaluations VALUES(52519,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52519,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52519,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52519,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52519,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52519,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52519,36,'cp 0','b8c6 d4a4 d7d6 e2e4 g7g6 f1e2 g8f6 e1g1 c8d7 b1c3 f8g7 a4b3 e8g8 f1d1 c6a5 b3b4 a5c6');
INSERT INTO evaluations VALUES(19551,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19551,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19551,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19551,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19551,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19551,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19551,36,'cp 0','d4a4 d7d6 e2e4 g8f6 f1e2 c8d7 e1g1 e7e6 c2c3 f8e7 a4c2 d8c7 c3c4 a8c8 b1c3 c6a5 f3d2 e8g8 c2d1 a7a6 a1b1 a5c6 d2f3 c6a5');
INSERT INTO evaluations VALUES(14285,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(14285,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(14285,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(14285,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(14285,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(14285,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(14285,36,'cp -73','e7e5 g1e2 f8d6 c3c4 e5e4 d3c2 c6b4 c4d5 e8g8 a2a3 b4d5 b1d2 f8e8 d2c4 c7c6 c4d6 d8d6 h2h3 c8f5 c1d2 h7h5 a1c1 h5h4 e2c3 d5c3 d2c3 d6d7 e1g1 e8e6 g1h2 f6d5 c2b3 e6g6 b3d5 c6d5 d1h5 g6h6 h5e2 a7a5 h2h1 h6g6');
INSERT INTO evaluations VALUES(15114,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(15114,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(15114,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(15114,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(15114,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(15114,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(15114,36,'cp -62','g1e2 e5e4 d3c2 f8d6 c3c4 c6b4 c4d5 e8g8 a2a3 b4d5 b1d2 f8e8 d2c4 c7c6 c4d6 d8d6 c1d2 c8f5 a1c1 h7h5 h2h3 d6d7 e2c3 d5c3 b2c3 d7d6 c3c4 d6a3 c4c5 b7b5 c1a1 a3b2 e1g1 f5e6 d1c1 b2c1 f1c1 e6d5 a1a6 e8e7 c2d1 g8h7 c1a1 e7c7 d1c2 h5h4 d2b4 f6h5');
INSERT INTO evaluations VALUES(34458,30,'cp 371',NULL);
INSERT INTO evaluations VALUES(34458,31,'cp 349',NULL);
INSERT INTO evaluations VALUES(34458,32,'cp 378',NULL);
INSERT INTO evaluations VALUES(34458,33,'cp 348',NULL);
INSERT INTO evaluations VALUES(34458,34,'cp 369',NULL);
INSERT INTO evaluations VALUES(34458,35,'cp 371',NULL);
INSERT INTO evaluations VALUES(34458,36,'cp 394','c1g5 f6g6 d4e5 c8g4 d1d2 b8d7 e1c1 f7f6 e5f6 g8f6 f1e2 g6f7 c1b1 f8e7 f3d4 g4e2 d2e2 e8g8 d4f5 a8e8 f2f3 f7e6 e2d2 d7e5 d2d4 a7a6 c3e2 e5f7 h2h4 e7d8 d4b4 e6d7 f5d6 f7d6 d1d6 d7f7 h1d1 d8c7');
INSERT INTO evaluations VALUES(34452,30,'cp 394',NULL);
INSERT INTO evaluations VALUES(34452,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(34452,32,'cp 392',NULL);
INSERT INTO evaluations VALUES(34452,33,'cp 394',NULL);
INSERT INTO evaluations VALUES(34452,34,'cp 370',NULL);
INSERT INTO evaluations VALUES(34452,35,'cp 382',NULL);
INSERT INTO evaluations VALUES(34452,36,'cp 373','f6g6 d4e5 c8g4 d1d2 b8d7 e1c1 f7f6 e5f6 g8f6 f1e2 g4f3 g2f3 f8e7 h1g1 g6f7 d2e3 h7h6 g5f4 d7e5 f4g3 e8g8 f3f4 e5c4 e2c4 f7c4 f4f5 f6g4 e3f3 g4e5 f3g2 f8f7 f2f4 e5d7 g3f2 e7f6 d1d6 a8e8 g2g6 e8e4 d6d7 f7d7 g6f6 e4f4 f6e6 d7f7 f2e3 f4g4 e6c4 g4c4 e3h6');
INSERT INTO evaluations VALUES(34462,30,'cp 385',NULL);
INSERT INTO evaluations VALUES(34462,31,'cp 384',NULL);
INSERT INTO evaluations VALUES(34462,32,'cp 381',NULL);
INSERT INTO evaluations VALUES(34462,33,'cp 397',NULL);
INSERT INTO evaluations VALUES(34462,34,'cp 403',NULL);
INSERT INTO evaluations VALUES(34462,35,'cp 378',NULL);
INSERT INTO evaluations VALUES(34462,36,'cp 406','d4e5 c8g4 d1d2 b8d7 e1c1 f7f6 e5f6 g8f6 f1e2 g6f7 c1b1 f8e7 f3d4 g4e2 d2e2 e8g8 d4f5 f7e6 f2f3 f8e8 f5e7 e6e7 e2d2 d7e5 g5f6 e7f6 d2d6 f6f7 b2b3 a7a5 d6c5 f7f4 a2a4 g8h8 c5f2 b7b5 d1d6 b5a4 c3a4 a8d8 d6d8 e8d8 f2g3 f4g3 h2g3 d8d2');
INSERT INTO evaluations VALUES(36674,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(36674,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(36674,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(36674,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(36674,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(36674,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(36674,36,'cp 58','e1g1 b8d7 a2a4 e7e5 b1d2 b7b6 f1e1 a7a5 d3b5 c8b7 d4d5 d8c8 b2b3 c7c6 d5c6 b7c6 c1a3 c6b5 a4b5 f8d8 c3c4 d7c5 a3c5 c8c5 d2f1 c5c7 f1e3 a8c8 e3d5 c7b7 d5f6 g7f6 h2h3 c8c7');
INSERT INTO evaluations VALUES(36673,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(36673,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(36673,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(36673,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36673,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(36673,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(36673,36,'cp 52','b8d7 c1e3 e7e5 b1d2 b7b6 a2a4 a7a5 h2h3 c8b7 d1c2 f8e8 a1d1 d8c8 d4e5 d7e5 f3e5 d6e5 d3b5 e8d8 c2b3 b7a6 c3c4 c8e6 b5a6 a8a6 d2b1 a6a8 b1c3 d8d1 f1d1 g7f8 b3b5 c7c6 b5b6 e6c4 d1d8 a8d8 b6d8 f6e4');
INSERT INTO evaluations VALUES(4098,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(4098,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(4098,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(4098,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(4098,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(4098,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(4098,36,'cp 59','c1e3 e7e5 b1d2 f8e8 a2a4 c7c6 f1e1 d6d5 d4e5 f6g4 e5e6 e8e6 e3g5 g7f6 g5f6 g4f6 f3d4 e6e7 e4d5 f6d5 d3c4 d7f6 d1b3 e7e1 a1e1 d8c7 c4d5 c6d5 d2f3 c8d7 e1e7 b7b6 f3e5 a8e8 b3a3 c7c5 e5d7 c5e7 a3e7 e8e7');
INSERT INTO evaluations VALUES(36756,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36756,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36756,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36756,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36756,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36756,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36756,36,'cp 24','b1d2 c7c5 c2c3 b7b6 a2a4 b8c6 f1b5 c8b7 e1g1 a7a6 b5c6 b7c6 f3e5 c6e8 b2b3 a8c8 d1e2 a6a5 f1e1 f6d7 e5d3 d7f6 e1c1 e7e6 d4c5 b6c5 f4e5 d8e7 c3c4 e8c6 e5c3 h7h6 c3a5 f6e4 d2e4 d5e4');
INSERT INTO evaluations VALUES(37265,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(37265,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(37265,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(37265,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(37265,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(37265,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(37265,36,'cp 33','f3e5 d7d6 e5f3 f6e4 e1g1 e4f6 a2a3 b4c3 d2c3 f8e8 h2h3 h7h6 f1e1 e8e1 f3e1 b8c6 c4f1 b7b6 e1d3 c8b7 c1f4 c6e7 d1e2 e7f5 a1e1 d8d7 e2d2 f6e4 d2d1 e4f6 f4h2 c7c5 d1d2 f6e4 d2f4 a8e8');
INSERT INTO evaluations VALUES(37260,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(37260,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(37260,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(37260,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(37260,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(37260,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(37260,36,'cp 19','d7d6 e5f3 f6e4 e1g1 e4f6 a2a3 b4c3 d2c3 h7h6 h2h3 c8f5 f3d4 f5g6 c4d3 d8d7 c3c4 b8c6 d4c6 d7c6 b2b4 f8e8 c1b2 g6d3 c2d3 e8e6 f1e1 e6e1 d1e1 a8e8 e1d2 d6d5 b4b5 c6d6 d2c3 d5c4 d3c4 f6h5 a1e1 e8e1 c3e1');
INSERT INTO evaluations VALUES(41090,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(41090,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(41090,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(41090,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(41090,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(41090,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(41090,36,'cp -48','e8g8 e2d1 e7b4 f1e2 b4c3 b2c3 f6e4 c1b2 b8d7 c3c4 d7b6 c4d5 b6a4 c2c4 a4b2 d1b3 b2c4 e2c4 c8g4 f3e5 g4h5 b3h3 h5g6 e1g1 e4d2 f1c1 d2c4 e5g6 h7g6 c1c4 d8d5 c4c7 d5d4 a1c1 b7b6 h3f3 a7a5 g2g3 a8d8 h2h4');
INSERT INTO evaluations VALUES(36405,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(36405,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(36405,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(36405,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(36405,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(36405,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(36405,36,'cp -73','e2d1 e7b4 f1e2 b4c3 b2c3 f6e4 c1b2 b8d7 f3d2 e4d2 d1d2 d7b6 e1g1 c8f5 a1e1 d8d7 b2c1 f8e8 e2d3 b6a4 c3c4 f5d3 c2d3 d5c4 d3c4 a4b6 d2d3 d7c6 c4c5 b6d5 c1d2 b7b5 c5b6 a7b6 e1e8 a8e8 a2a3 h7h6 h2h3 c6a4 f1c1 e8e6 g1h2 a4d7 c1e1 e6f6 d3e2 c7c6 a3a4');
INSERT INTO evaluations VALUES(43225,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(43225,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43225,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(43225,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43225,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(43225,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(43225,36,'cp 0','e8g8 e2e3 b7b6 g2g3 c8b7 f1g2 d7d5 e1g1 b8d7 c2c4 c7c5 d1e2 d5c4 b3c4 e7e6 f1d1 c5d4 e3d4 d8c7 b1d2 a8c8 a1c1 f8e8 d2f1 b7a6 f3d2 a6b7');
INSERT INTO evaluations VALUES(37446,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(37446,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37446,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(37446,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(37446,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37446,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(37446,36,'cp 5','e2e3 b7b6 g2g3 d7d5 f1g2 b8d7 c2c4 c7c5 e1g1 c8b7 d1e2 d5c4 b3c4 e7e6 b1c3 a7a6 f1d1 c5d4 e3d4 d8c7 f3d2 b7g2 g1g2 c7c6 c3e4 a8c8 a1c1 g7h6 f2f3 f6e4 e2e4 f7f5 e4c6 c8c6 d4d5 e6d5 c4d5 c6c1 b2c1 d7c5');
INSERT INTO evaluations VALUES(37445,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(37445,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(37445,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37445,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(37445,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37445,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37445,36,'cp 0','b7b6 g2g3 c7c5 f1g2 d7d5 e1g1 b8d7 c2c4 c8b7 d1e2 d5c4 b3c4 e7e6 f1d1 c5d4 e3d4 a8c8 b1d2 d8c7 a1c1 f8e8 d1e1 b7c6 e2d1 c7b7 a2a4 e8d8 d1b3 g7h6 c1d1 h6g7 d1c1');
INSERT INTO evaluations VALUES(44422,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(44422,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(44422,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(44422,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(44422,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(44422,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(44422,36,'cp 89','d4d5 e7e6 b1c3 e6d5 c4d5 f8g7 d5d6 b7b6 e2e4 c8b7 e4e5 f6g4 c1g5 b7f3 d1d2 f7f6 e5f6 g7f6 g2f3 f6g5 d2d5 b8c6 f3g4 d8f6 f1g2 e8c8 c3e4 f6b2 e1g1 g5f4 d5c4 b2d4 c4a6 c8b8 a1b1 c6b4 b1b4 d4b4 e4g3 f4g3 f2g3');
INSERT INTO evaluations VALUES(44429,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(44429,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(44429,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(44429,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(44429,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(44429,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(44429,36,'cp 81','f8g7 b1c3 e7e6 e2e4 e8g8 h2h3 e6d5 e4d5 f8e8 f1e2 d7d6 e1g1 c8f5 f1e1 h7h6 c1f4 f6e4 c3e4 e8e4 d1d2 g6g5 f4h2 b7b5 c4b5 e4b4 e2f1 d8f8 f3g5 b4d4 d2c1 h6g5 c1g5 f5e4 f2f3 e4d5 e1e7 f8d8 a1e1 b8d7');
INSERT INTO evaluations VALUES(40018,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(40018,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(40018,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(40018,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(40018,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(40018,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(40018,36,'cp 86','b1c3 e7e6 e2e4 e8g8 h2h3 d7d6 f1d3 e6d5 e4d5 f8e8 c1e3 c8f5 e1g1 d8d7 e3f4 f5d3 d1d3 b8a6 a2a3 a6c7 a1d1 f6h5 f4h2 f7f5 b2b4 b7b6 f1e1 h5f6 h2f4 e8e1 d1e1 a8e8 e1e8 c7e8 b4c5 b6c5 a3a4 h7h6 a4a5 a7a6 f4e3 g6g5 e3d2');
INSERT INTO evaluations VALUES(40016,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(40016,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(40016,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(40016,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(40016,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(40016,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(40016,36,'cp 85','e8g8 h2h3 e7e6 e2e4 e6d5 e4d5 d7d6 f1d3 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1c2 e4f6 a1e1 d8d7 c2d2 b8a6 e3g5 e8e1 f1e1 a8e8 e1e8 f6e8 b2b3 h7h5 a2a3 g7f6 g5f4 d7d8 g2g3 a6b8 d2e2 b7b6 g1h1 d8e7 h1g2 e7e2');
INSERT INTO evaluations VALUES(35712,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(35712,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(35712,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(35712,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(35712,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(35712,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(35712,36,'cp 95','e2e4 e7e6 h2h3 e6d5 e4d5 d7d6 f1d3 f8e8 c1e3 c8f5 e1g1 f5d3 d1d3 a7a6 e3f4 d8c7 a1e1 b8d7 g2g4 e8e1 f1e1 h7h6 b2b3 d7f8 g1g2 a8e8 e1e8 f6e8 c3e4 c7d7 e4g3 b7b5 f4d2 e8f6 d2e3 h6h5 g4g5 f6e8 g3e4 b5b4');
INSERT INTO evaluations VALUES(35710,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(35710,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(35710,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(35710,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(35710,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(35710,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(35710,36,'cp 79','e7e6 h2h3 f8e8 f1e2 e6d5 e4d5 d7d6 e1g1 c8f5 f1e1 b8a6 g2g4 f5d7 c1g5 h7h6 g5h4 h6h5 g4h5 d7h3 d1d2 d8d7 h5g6 f7g6 f3g5 h3f5 a2a3 a6c7 d2f4 e8e5 g5f3 e5e7 h4g3 c7e8 f3g5 f6h7 g5e6 g7e5 f4e3 f5e6 d5e6 d7e6 g3e5 d6e5 e2f3');
INSERT INTO evaluations VALUES(49669,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49669,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(49669,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(49669,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(49669,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(49669,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(49669,36,'cp -13','g1f3 b8c6 b1c3 d7d5 e4d5 f6d5 e2b5 d5c3 b2c3 f8d6 e1g1 e8g8 d2d4 e5d4 c3d4 h7h6 c2c3 d8f6 h2h3 c8f5 a2a4 c6a5 f1e1 c7c6 b5d3 f8e8 c1e3 d6c7 c3c4 a8d8 d3f5 f6f5 c4c5 a5c4 d1b3 c4e3 e1e3 e8e3 f2e3');
INSERT INTO evaluations VALUES(49650,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49650,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(49650,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(49650,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(49650,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(49650,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(49650,36,'cp -18','b8c6 b1c3 d7d5 e4d5 f6d5 e2b5 d5c3 b2c3 f8d6 d2d4 e5d4 c3d4 h7h6 e1g1 e8g8 c1b2 d8f6 a1b1 c8g4 b5e2 c6b4 c2c4 f8e8 a2a3 g4f5 b1c1 b4c6 d1d2 a8d8 d4d5 c6e5 b2e5 d6e5 f3e5 e8e5 e2f3 f5e4 f3e4 e5e4 f2f3');
INSERT INTO evaluations VALUES(50553,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50553,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50553,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50553,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50553,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(50553,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(50553,36,'cp 0','f3d4 d8c7 c4d3 g8f6 e1g1 f8c5 d4b3 c5a7 c2c4 b8c6 d1e2 e8g8 b1c3 c6e5 c1f4 d7d6 h2h3 c8d7 f1d1 a8c8 a1c1 h7h6 f4g3 f8e8 g3h4 c7d8 h4g3');
INSERT INTO evaluations VALUES(57357,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(57357,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(57357,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(57357,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(57357,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(57357,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(57357,36,'cp 110','e2e4 f5e4 c3e4 b8c6 g1f3 d7d5 e4g3 f8d6 c2c4 d6b4 c1d2 b4d2 d1d2 g8e7 f1d3 e8g8 a1d1 e7g6 d3g6 h7g6 e1g1 d5c4 f1e1 b7b5 h2h4 a7a5 a2a3 c6e7 f3g5 e7d5 g3e4 f8f5 e4c5 d8d6 g2g3 c8d7 e1e2 d5f6 d1e1 d7c6 c5e6 d6d5 f2f3 f5f3 g5f3');
INSERT INTO evaluations VALUES(57364,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(57364,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(57364,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(57364,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(57364,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(57364,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(57364,36,'cp 135','f5e4 c3e4 b8c6 g1f3 d7d5 e4g3 f8d6 c2c4 d6b4 c1d2 b4d2 d1d2 g8e7 f1d3 e8g8 a1d1 c8d7 e1g1 g8h8 d2e3 c6b4 c4c5 b4d3 e3d3 e7g6 f1e1 g6f4 d3c3 d7e8 e1e3 e8g6 d1e1 b7b6 f3e5 d8g5 b2b4 h7h6 e3f3 g6h7 b4b5 b6c5 c3c5');
INSERT INTO evaluations VALUES(57370,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(57370,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(57370,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(57370,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(57370,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(57370,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(57370,36,'cp 183','c3e4 g8f6 f1d3 b8c6 e4g5 d7d5 g1f3 f8d6 e1g1 e8g8 f1e1 d8e8 f3e5 c6d4 g5h7 f6h7 d3g6 e8d8 d1d4 d8f6 e1e2 d6e5 g6h7 g8h7 e2e5 h7g8 c1g5 c7c5 d4e3 d5d4 e3g3 f6g6 e5c5 c8d7 g3d6 a8c8 c5c8 d7c8 g5f4 g6c2 d6d4 f8f7 f4e5 c8d7 d4a7 d7c6 a7e3');
INSERT INTO evaluations VALUES(4180,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(4180,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(4180,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(4180,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(4180,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4180,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4180,36,'cp 12','d7d6 b1d2 a7a6 f1e1 c8e6 a2a4 c5a7 c4b3 f6d7 d2f1 e6b3 d1b3 d7c5 b3c2 d6d5 c1e3 d5e4 d3e4 d8d3 c2d2 d3d2 f3d2 c5d3 e1d1 d3b2 d1b1 b2d3 b1b7 f8c8 e3a7 a8a7 b7a7 c6a7 f1e3 d3f4 d2c4 f7f6 g1f1 c8d8 a4a5');
INSERT INTO evaluations VALUES(22281,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(22281,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(22281,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(22281,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(22281,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(22281,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(22281,36,'cp 208','f3e5 c6e5 f4e5 g8f6 c2c4 c8e6 c4d5 f6d5 b1d2 d8g5 e5g3 e8c8 f1c4 e6g4 d1c1 c8b8 e1g1 d5f6 c4e2 f8d6 d2f3 g5h5 f1d1 d6g3 h2g3 h5f7 a2a3 a7a6 d1d8 h8d8 c1c3 f7e7 a1d1 d8c8 c3b3 g4e6 b3c2 e6g4 e2f1 h7h5 b2b4 c8d8 d1c1 f6e4 c2b2');
INSERT INTO evaluations VALUES(22276,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(22276,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(22276,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(22276,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(22276,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(22276,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(22276,36,'cp 200','g8f6 f1b5 c8d7 b5c6 d7c6 e5c6 b7c6 e1g1 f8d6 b1d2 e8g8 c2c4 d6f4 e3f4 d8d6 g2g3 a8e8 d2f3 d5c4 a1c1 d6d5 f3d2 c4c3 c1c3 d5a2 c3c6 a2b2 d2f3 b2a2 f3e5 a7a5 c6c7 e8c8 c7a7 c8a8 a7e7 a8d8 d1c1 d8c8 c1e3 f8e8 e7a7');
INSERT INTO evaluations VALUES(26967,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(26967,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(26967,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(26967,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(26967,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(26967,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(26967,36,'cp 26','g4h5 c2c3 g8f6 d2d3 d8d7 b2b4 h8g8 c1g5 h7h6 g5f6 g7f6 g1h2 f8e7 b1d2 c6d8 g2g3 d8e6 d1b3 e8c8 a2a4 f6f5 a4a5 e6g5 f3g5 e7g5 a1a2 d6d5 e4d5 g5d2 a2d2 f5f4 b4b5 h5f3 d5d6 d7d6 c4f7 g8f8 b3e6 c8b8 a5a6 b7b6 e6d6 d8d6');
INSERT INTO evaluations VALUES(26952,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(26952,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(26952,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(26952,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(26952,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(26952,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(26952,36,'cp 65','c4b5 g8e7 d2d3 f7f6 f1e1 d6d5 b1d2 a7a6 b5a4 d5d4 a1b1 d8d6 c2c4 d6d7 b2b4 e7g6 a4c6 b7c6 f3e5 h5d1 e5d7 d1c2 d7f8 h8f8 b1b2 c2d3 d2b3 e8c8 b3a5 g6e5 b2d2 g7g5 e1d1 d3c4 c1b2 c4a2 b2d4 a2f7 d4e5 d8d2 d1d2 f6e5');
INSERT INTO evaluations VALUES(28728,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(28728,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(28728,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(28728,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(28728,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(28728,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(28728,36,'cp 134','f1d3 d7d6 g1e2 g7g6 h2h3 f8g7 c1g5 b8d7 e1g1 c7c5 d4d5 h7h6 g5h4 e6e5 f2f3 a7a6 a2a3 b7c8 a1b1 g7f8 h4f2 h6h5 b2b4 f8e7 f2e3 f6h7 d1d2 e7g5 e3g5 h7g5 d3c2 e8g8 c2a4 g5h7 a4c6 a8b8 d2h6 g8h8 f1f2 f7f5 e4f5 g6f5');
INSERT INTO evaluations VALUES(28729,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(28729,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(28729,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(28729,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(28729,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(28729,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(28729,36,'cp 110','d7d6 g1e2 g7g6 e1g1 f8g7 h2h3 e8g8 c1g5 b8d7 f2f4 h7h6 g5h4 e6e5 f4e5 d6e5 d4d5 c7c6 a1b1 a7a6 g1h1 d8c7 a2a3 a8c8 d5c6 b7c6 b2b4 f6h5 c3d5 c6d5 c4d5 c7b7 h4e7 f8e8 d5d6 h5f4 d3c2 b6b5 e2f4 e5f4 f1f4');
INSERT INTO evaluations VALUES(28645,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(28645,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(28645,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(28645,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(28645,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(28645,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(28645,36,'cp 135','g1e2 g7g6 e1g1 f8g7 h2h3 b8d7 c1g5 e8g8 f2f4 e6e5 f4e5 d6e5 d4d5 h7h6 g5h4 c7c6 a2a3 d8c7 a1c1 a8c8 b2b4 c7d6 g1h1 a7a5 c1b1 a5b4 a3b4 f6h5 h4f2 h5f4 c4c5 b6c5 f2c5 d7c5 b4c5 d6d7 d5d6');
INSERT INTO evaluations VALUES(29538,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29538,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29538,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29538,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29538,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29538,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29538,36,'cp 0','e7e6 f1d3 f5d3 c2d3 c7c5 d1b3 d8d7 d4c5 f8c5 g1f3 b8c6 e1g1 e8g8 f3e5 c6e5 f4e5 f8c8 e5f6 g7f6 d3d4 c5e7 b1c3 c8c4 f1c1 a8c8 b3d1 a7a5 d1g4 g8h8 g4h5 h8g7 h5g4');
INSERT INTO evaluations VALUES(30093,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30093,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30093,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30093,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30093,36,'cp 0','f1d3 f5d3 c2d3 c7c5 d1b3 d8d7 d4c5 f8c5 g1f3 b8c6 e1g1 e8g8 f3e5 c6e5 f4e5 f8c8 e5f6 g7f6 d3d4 c5e7 b1c3 c8c4 f1c1 a8c8 b3d1 a7a5 d1g4 g8h8 g4h5 h8g7 h5g4');
INSERT INTO evaluations VALUES(30039,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(30039,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(30039,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(30039,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(30039,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(30039,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(30039,36,'cp 131','g1f3 c7c5 b1c3 b8c6 f1b5 d8b6 e1g1 e8c8 b5c6 b6c6 c1e3 f8d6 d1e2 e6g4 f1d1 a7a6 d4c5 d6c5 e2c4 c5e3 c4c6 b7c6 f2e3 d8d1 a1d1 h8e8 d1e1 g4h5 e3e4 c8c7 g1f2 h5g6 e4e5 f6d7 e1e2 d7c5 g2g3 c5d7 c3a4 a6a5 b2b3 g6h5 h2h3 h5g6 a2a3');
INSERT INTO evaluations VALUES(32104,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(32104,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(32104,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(32104,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(32104,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(32104,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(32104,36,'cp 187','d1f3 d8d7 f3b3 c7c5 a2a4 g8f6 a4a5 a7a6 b1c3 b8c6 c3d5 c6d4 d5f6 g7f6 b3d3 e8c8 c2c3 d4e6 d3f3 e6g5 f3f6 f8e7 f6f5 d7f5 e4f5 h8g8 g2g3 g5f3 g1g2 f3h4 g2h3 d6d5 c4e2 h4f5 e2g4 g8g4 h3g4 d8g8 g4f3 c5c4 f3e2 h7h5 d2d3 c4d3 e2d3 f5d6 d3e2');
INSERT INTO evaluations VALUES(32102,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(32102,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(32102,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(32102,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(32102,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(32102,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(32102,36,'cp 186','d8d7 f3b3 c7c6 d2d4 e5d4 c2c3 d4d3 f1e1 b8a6 c4a6 b7a6 b3c4 g8f6 c4a6 f8e7 a6d3 e8g8 c1f4 d7b7 b2b3 a7a5 b1d2 f8d8 d3f3 a5a4 b3b4 h7h6 a1d1 d6d5 e4d5 d8d5 f4g3 a8d8 d2c4 d5d1 e1d1 d8d1 f3d1');
INSERT INTO evaluations VALUES(43227,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(43227,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(43227,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43227,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43227,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43227,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(43227,36,'cp 16','e3e4 d7d6 c1e3 e8g8 g1f3 f6g4 e3g5 c7c5 d4d5 c5c4 a1b1 b7b5 c3b5 d8b6 b5d4 e7e5 d5e6 f7e6 g5h4 b8c6 d4c6 b6c6 f3d2 g4e5 f1e2 d6d5 e1g1 d5e4 b2b3 c8a6 b3c4 e5c4 e2c4 a6c4 d2c4 c6c4 b1b7 c4d5 b7d7 d5d1 f1d1 a8c8 d7a7 c8c2 a7a6 g7f6 h4g3 c2b2 a6e6');
INSERT INTO evaluations VALUES(48236,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48236,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48236,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(48236,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48236,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48236,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48236,36,'cp 0','f8d6 b2b3 e8g8 c1b2 b8d7 f1e2 a7a6 e1g1 b7b6 c4d5 e6d5 f3d4 d7e5 f2f4 e5g6 e2f3 g6e7 c3e2 d8d7 d4c2 d7e6 c2d4');
INSERT INTO evaluations VALUES(35040,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(35040,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(35040,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(35040,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(35040,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(35040,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(35040,36,'cp 137','g1f3 f8b4 b1d2 h4e7 e2e4 d7d5 f1d3 d5e4 d3e4 g8f6 e4c2 b7b6 e1g1 c8b7 a2a3 b4d6 f1e1 e8g8 b2b4 b8d7 c1b2 a7a5 b4b5 a8c8 h2h3 c7c5 d4d5 c8e8 d2f1 e6e5 f1e3 g7g6 f3g5 f6h5');
INSERT INTO evaluations VALUES(35039,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(35039,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(35039,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(35039,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(35039,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(35039,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(35039,36,'cp 154','h4d8 e2e4 d7d5 e4d5 e6d5 c4d5 b8d7 f1c4 g8f6 b1c3 d7b6 d1e2 f8e7 e1g1 b6c4 e2c4 a7a6 f1e1 e8g8 c1f4 e7d6 f3e5 f6d7 c4b3 d7b6 a2a4 a6a5 e5d3 d6f4 d3f4 d8f6 g2g3 f6d4 c3b5 d4b4 b3b4 a5b4 b5c7 a8a4 a1a4 b6a4');
INSERT INTO evaluations VALUES(37206,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(37206,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(37206,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(37206,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(37206,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(37206,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(37206,36,'cp -25','d7d5 c4b3 d5e4 e1g1 c5d4 e5c4 c8g4 d1d2 e4d3 d2d3 b8c6 b1c3 d4c3 d3d8 f8d8 b2c3 g4e6 c4b2 f6e4 b2d1 d8d7 f2f3 a8d8 c1f4 e4c5 d1b2 e6b3 a2b3 c5e6 f4e3 f7f6 g1f2 g8f7 f1d1 d7d1 a1d1 d8d1 b2d1 f6f5 g2g3 g7g5 d1b2 b7b6 e3d2 a7a6 b2d3 a6a5 d2e3 h7h5');
INSERT INTO evaluations VALUES(37097,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(37097,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(37097,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(37097,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(37097,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(37097,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(37097,36,'cp -19','c4b3 d5e4 e1g1 c5d4 e5c4 c8g4 d1d2 e4d3 d2d3 a7a5 a2a4 b8a6 b1c3 a6c5 d3g3 d4c3 b2c3 f8e8 f2f3 c5b3 c2b3 g4e6 c1g5 d8e7 a1e1 e7c5 g5e3 c5c6 e3d4 e6c4 b3c4 e8e1 f1e1 a8e8 c4c5 e8e1 g3e1 g8f8 e1e5 f6e8 e5e4 c6e4 f3e4 e8f6 d4f6 g7f6 c3c4 f8g8');
INSERT INTO evaluations VALUES(37405,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37405,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(37405,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(37405,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37405,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37405,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37405,36,'cp -6','d7d5 e1g1 c7c5 d4c5 d5c4 d3c4 b8d7 c5c6 b7c6 d1c2 d7b6 f1d1 d8c7 c4d3 f6d5 c1d2 c6c5 d2a5 c8e6 d3e2 a8c8 b1c3 d5b4 c2e4 h7h6 a5b4 c5b4 e4b4 g7c3 b4c3 c7c3 b2c3 c8c3 a2a4 e6c4 e2c4 c3c4 d1d4 c4d4 f3d4');
INSERT INTO evaluations VALUES(36736,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(36736,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36736,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(36736,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36736,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36736,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36736,36,'cp 0','e1g1 c7c5 d4c5 d5c4 d3c4 b8d7 c5c6 d7b6 c6b7 c8b7 c4e2 d8c7 b1c3 f8d8 d1b3 f6e4 f3d4 e4c5 b3a3 a7a5 d4b5 c7e5 b5d4');
INSERT INTO evaluations VALUES(40813,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(40813,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(40813,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(40813,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(40813,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(40813,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(40813,36,'cp -7','f1e2 h7h6 g5h4 e8g8 d3d4 f8e8 e1g1 f6e4 h4e7 e8e7 c2c4 c8f5 e2d3 d6d5 c4d5 d8d5 d3e4 f5e4 b1c3 d5d6 f1e1 e4f3 d1f3 b8c6 e1e7 d6e7 d4d5 a8e8 h2h3 c6d4 f3d3 d4e2 c3e2 e7e2 d3e2 e8e2 a1c1 e2e7 g1f1 f7f6 a2a3 g8f7 f2f3 e7d7');
INSERT INTO evaluations VALUES(40811,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(40811,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(40811,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(40811,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(40811,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(40811,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(40811,36,'cp -12','e8g8 d3d4 h7h6 g5h4 c8f5 e2d3 f5d3 d1d3 b8d7 e1g1 c7c6 f1e1 f8e8 c2c4 d6d5 b1d2 d5c4 d2c4 d7b6 c4e5 f6d7 h4e7 e8e7 e5g4 e7e1 a1e1 b6d5 g2g3 d8b6 a2a3 d7f6 g4f6 d5f6 b2b4 b6c7 f3e5 a8d8 e5f3 a7a5 d3b3 d8e8 e1e8 f6e8 f3e5 a5b4 a3b4');
INSERT INTO evaluations VALUES(44866,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44866,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44866,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44866,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(44866,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44866,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44866,36,'cp 0','c6d5 g1f3 b8c6 b1c3 e7e6 f1d3 f8d6 e1g1 h7h6 e3e4 d5e4 c3e4 e8g8 f1e1 d6b4 c1d2 b4d2 d1d2 f6d5 e4c3 d5f6 c3e4');
INSERT INTO evaluations VALUES(3222,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(3222,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3222,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3222,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(3222,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3222,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(3222,36,'cp 0','h2h3 f6e4 c3e4 d6d5 d4c6 b7c6 c4d3 d5e4 d3e4 c8a6 d1d8 f8d8 f1d1 g7b2 a1b1 b2e5 e4c6 d8d1 b1d1 a8c8 c6a4 e5d6 a4b3 c8c7 d1d5 a6e2 d5a5 a7a6 c2c4 e2c4 b3c4 c7c4 a5a6 d6f4 e3f4 c4f4 g2g3 f4d4 a6a7 e7e6 g1g2 d4d3 a7a8 g8g7 g2f1 d3d2 a8a7 d2b2 f1g2 b2d2');
INSERT INTO evaluations VALUES(3220,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3220,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3220,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3220,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(3220,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(3220,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3220,36,'cp 0','f6e4 c3e4 d6d5 d4c6 b7c6 c4d3 d5e4 d3e4 c8a6 d1d8 f8d8 f1d1 g7b2 a1b1 b2e5 e4c6 d8d1 b1d1 a8c8 c6a4 e5d6 a4b3 c8c7 d1d5 a6e2 d5a5 a7a6 c2c4 e2c4 b3c4 c7c4 a5a6 d6f4 e3f4 c4f4 g2g3 f4b4 a6a7 e7e6 g1g2 g8g7 h3h4 h7h5 g2f3 g7f6 f3e3 b4b2 a2a4 b2b3 e3e2 b3b2');
INSERT INTO evaluations VALUES(5363,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(5363,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(5363,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(5363,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(5363,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(5363,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(5363,36,'cp 0','e1g1 d7d6 c2c3 e7e5 d2d3 g8e7 a2a3 e8g8 b2b4 a7a6 c1e3 b7b6 b1d2 h7h6 f1e1 c8g4 a1b1 g4e6 d2f1 a8c8 d1a4 e6d7 a4d1');
INSERT INTO evaluations VALUES(5362,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5362,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5362,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5362,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5362,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5362,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5362,36,'cp 0','d7d6 c2c3 e7e5 d2d3 g8e7 a2a3 a7a6 b2b4 e8g8 c1e3 b7b6 b1d2 c8e6 f3g5 e6c8');
INSERT INTO evaluations VALUES(5513,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5513,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5513,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5513,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5513,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5513,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5513,36,'cp 0','c2c3 e7e5 d2d3 g8e7 a2a3 e8g8 b2b4 a7a6 c1e3 b7b6 b1d2 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(5507,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5507,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5507,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5507,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5507,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5507,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5507,36,'cp 0','e7e5 d2d3 g8e7 a2a3 e8g8 b2b4 a7a6 c1e3 b7b6 b1d2 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(7051,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(7051,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(7051,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(7051,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(7051,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(7051,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(7051,36,'cp 89','d8e7 d6e7 c6e7 c1g5 b7b5 e1c1 c8b7 f2f3 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e7f5 d1e1 f5d6 e1e3 h6h5 g2g3 h5h4 g3g4 h8g8 f1h3 e8d8 h1e1 a6a5 b2b3 g8g5 e3c3 a8c8 c3c8 d6c8 c2c4 c8b6 c4b5 b6d5 c1b2');
INSERT INTO evaluations VALUES(429,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(429,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(429,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(429,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(429,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(429,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(429,36,'cp 105','d6e7 c6e7 c1g5 b7b5 e1c1 c8b7 f2f3 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e7f5 d1e1 f5d6 g2g3 a8c8 e1e3 e8e7 f1g2 c8c4 b2b3 c4c5 h1d1 h8c8 d1d2 d6f5 e3e2 e7d6 f3f4 f5d4 e2f2 c5c3 c1b2 c8c5 a2a3 a6a5 f4e5 f6e5 g2e4 c3e3');
INSERT INTO evaluations VALUES(423,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(423,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(423,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(423,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(423,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(423,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(423,36,'cp 109','c6e7 c1g5 b7b5 e1c1 c8b7 f2f3 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e7f5 d1e1 f5d6 e1e3 a8c8 f1d3 e8d7 h1e1 h8g8 g2g3 g8e8 f3f4 e5f4 e3e8 c8e8 e1e8 d6e8 g3f4 d7d6 c2c4 d6c5 c4b5 a6b5 c1c2 e8d6 d3f1 c5d5 c2c3 d5c5 a2a3 f6f5 f1g2 d6e4 c3d3 e4f2 d3e2');
INSERT INTO evaluations VALUES(426,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(426,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(426,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(426,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(426,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(426,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(426,36,'cp 103','c1g5 b7b5 e1c1 c8b7 f2f3 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e8d7 f3f4 e7f5 g2g3 f5d6 f4e5 f6e5 f1h3 d7e7 d1e1 f7f6 h1f1 a8a7 f1f4 h6h5 f4h4 a7c7 h3g4 c7c5 h4h5 h8h5 g4h5 c5d5 e1d1 d5c5 h5f3 e5e4');
INSERT INTO evaluations VALUES(425,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(425,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(425,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(425,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(425,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(425,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(425,36,'cp 100','b7b5 f2f3 c8b7 e1c1 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e8d7 f3f4 e7f5 d1e1 a8e8 g2g3 f5d6 f1h3 d7c7 e1e3 e5f4 e3c3 d6c4 h1f1 f4g3 c3g3 h8g8 g3d3 e8e2 b2b3 c4e5 d5d6 c7c6 d3d4 e2h2 h3f5 h2e2 d6d7 c6c7 f1d1 g8d8 a2a4 b5a4 d4a4 e2f2 f5h3 f6f5 d1d5');
INSERT INTO evaluations VALUES(8487,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(8487,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(8487,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(8487,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(8487,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(8487,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(8487,36,'cp 77','b1d2 e7d6 f1d3 e8g8 f3e5 f6d7 d2f3 f7f6 e5c6 b7c6 d1c2 h7h6 f4g3 c6c5 e1g1 c5c4 d3e2 d7b6 f3h4 a7a5 h4g6 f8e8 e2h5 a5a4 b2b4 c4b3 a2b3 c8a6 f1b1 a4b3 c2b3 a6d3 b1e1 d3c4 b3c2 d8d7 g3d6 a8a1 e1a1 c7d6 h2h3 e8b8 a1b1 f6f5 g2g4 f5g4 g6f4 d7e7 h5g4');
INSERT INTO evaluations VALUES(8486,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(8486,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(8486,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(8486,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(8486,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(8486,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(8486,36,'cp 62','f6h5 g2g3 c6b8 f1d3 h5f4 g3f4 c7c5 d4c5 e7c5 d1c2 b8c6 h2h4 h7h6 h1g1 d8f6 d3b5 c8d7 b5c6 d7c6 f3e5 c5d6 d2f3 d6e5 f3e5 f6h4 e1c1 h4f6 c2d3 h6h5 d3d4 h5h4 g1g2 h4h3 g2h2 e8f8 d1g1 a7a6 g1g3 f6f5 e5c6 b7c6 d4g7 f8e7 g7g5 f5g5 f4g5 a8g8');
INSERT INTO evaluations VALUES(17026,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(17026,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(17026,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(17026,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(17026,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(17026,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(17026,36,'cp 132','e1g1 c7c6 f1e1 d7d6 b1d2 f8e7 d2f3 c8g4 e4e5 g4f3 d1f3 d6e5 e1e5 e8g8 c1g5 d8d7 h2h4 h7h6 g5d2 b7b5 c4b3 c6c5 e5f5 a8c8 c2c4 d4c3 d2c3 d7c6 f3g3 g8h8 a2a4 c6d7 g3f3 b5b4 c3e5 c8e8 g2g3 e7d6 f5f6 d6e5 f6f7 f8f7 b3f7');
INSERT INTO evaluations VALUES(17025,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(17025,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(17025,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(17025,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(17025,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(17025,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(17025,36,'cp 131','c7c6 b1d2 d7d5 e4d5 f6d5 f1e1 f8e7 d1h5 e8g8 d2f3 e7b4 c1d2 b4d2 f3d2 d5f4 h5e5 f4g6 e5c5 d8b6 b2b4 b6c5 b4c5 b7b6 a2a4 b6c5 d2b3 c8f5 b3c5 a8e8 c5b3 e8e1 a1e1 f8d8 g2g3 g8f8 h2h4 h7h5 a4a5 g6e7 e1e5 f5g4');
INSERT INTO evaluations VALUES(21261,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(21261,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(21261,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(21261,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(21261,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(21261,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(21261,36,'cp 163','c2c3 b7b6 b2b4 c6b8 f1d3 c7c5 e1g1 g8e7 c1f4 a7a6 b1d2 c5b4 c3b4 c8d7 a2a3 d7b5 d3b5 a6b5 h2h4 b8c6 d1e2 c6a7 h4h5 e7f5 f1c1 a8c8 g2g3 f8e7 g3g4 f5h4 f3h4 e7h4 c1c8 d8c8 d2f3 h4e7 a1c1 c8a6 c1c3 e8g8 g1g2 f8c8');
INSERT INTO evaluations VALUES(21639,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(21639,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(21639,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21639,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21639,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21639,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(21639,36,'cp 0','f7f6 f1b5 f6e5 d1h5 g7g6 h5e5 d8e7 b1c3 c8d7 b5c6 d7c6 g1f3 f8g7 e5e7 g8e7 c1f4 e8c8 f4e5 g7e5 f3e5 e7f5 e1c1 h8f8 g2g3 g6g5 h2h4 g5h4 g3g4 f5d6 h1h4 f8f2 h4h7 f2g2 d1d2 g2g1 d2d1');
INSERT INTO evaluations VALUES(22249,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(22249,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(22249,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(22249,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22249,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22249,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22249,36,'cp 0','f1b5 f6e5 d1h5 g7g6 h5e5 d8e7 b1c3 c8d7 b5c6 d7c6 c1f4 f8g7 e5e7 g8e7 g1f3 e7f5 a1d1 e8c8 e1g1 h8f8 f4e5 g7h6 f1e1 f5d6 e5d6 d8d6 h2h4 h6g7 e1e7 d6d7 e7e2 d7d6');
INSERT INTO evaluations VALUES(21904,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(21904,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(21904,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21904,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(21904,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(21904,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(21904,36,'cp 38','c2c3 g8f6 g1f3 f8e7 a2a4 h7h6 e1g1 e8g8 f1e1 a7a6 b1d2 c6a5 c4a2 c7c5 d2c4 a5c6 c4e3 c8e6 e3d5 a8b8 c1e3 d8d7 a4a5 e7d8 d5f6 d8f6 a2e6 f7e6 f3d2 f6d8 d2c4 d8c7 d1g4 d7f7 e1d1 g8h7 h2h3 b8d8 c4b6');
INSERT INTO evaluations VALUES(26898,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(26898,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(26898,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(26898,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(26898,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(26898,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(26898,36,'cp 63','a7a6 b5e2 c6b8 f3d2 g4e2 d1e2 f8e7 d2f1 e7g5 c1e3 g5e3 f1e3 g8e7 h2h4 e8g8 g2g4 b8d7 h4h5 b7b5 b1c3 h7h6 e1c1 d7f6 f2f3 f6h7 e2d2 h7g5 h1f1 d8b8 f3f4 e5f4 f1f4 f8e8 c3e2 b5b4 e2g3 a6a5 b2b3 a5a4');
INSERT INTO evaluations VALUES(25666,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(25666,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(25666,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(25666,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(25666,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(25666,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(25666,36,'cp 73','b5e2 c6b8 a2a4 f8e7 a4a5 g8f6 f3d2 g4e2 d1e2 c7c6 c2c4 b7b5 a5b6 d8b6 b1c3 e8g8 e1g1 a6a5 b2b3 f8c8 a1a3 b8a6 d5c6 c8c6 c1b2 c6c8 f1a1 a6b4 d2f3 f6d7 b2c1 d7c5 c1e3 b6c6 e3c5 c6c5 a3a4 b4c6 h2h3 c6d4 f3d4 e5d4');
INSERT INTO evaluations VALUES(34858,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(34858,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(34858,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(34858,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(34858,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(34858,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(34858,36,'cp 132','d2d4 g8f6 c1d2 c8d7 f1c4 f8b4 e1g1 e8g8 f3e5 b4c3 d2c3 a5b6 a2a4 b6d6 b2b3 b8c6 f1e1 c6e5 e1e5 d7c6 c3b2 f8e8 c4d3 f6d5 c2c4 d5f4 d3f1 f4g6 e5e3 a8d8 d1e1 g6h4 b3b4 b7b6 e3e5 c6b7 b4b5 h4f5 c4c5 d6e7');
INSERT INTO evaluations VALUES(44340,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(44340,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(44340,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(44340,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(44340,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(44340,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(44340,36,'cp 156','c2c4 d7d6 b1c3 b8d7 g1f3 f8e7 h2h3 d7f8 f1d3 f8g6 e1g1 e8g8 f1e1 f6h5 a2a4 h5f4 d3c2 c8d7 a1a3 f4h5 a4a5 f8e8 a5a6 b7b6 c1d2 h5f6 c2d3 d8c8 a3a1 h7h6 d1c2 e7f8 e1e8 c8e8 a1e1 g6e5 c2d1');
INSERT INTO evaluations VALUES(44334,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(44334,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(44334,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(44334,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(44334,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(44334,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(44334,36,'cp 131','d7d6 b1c3 g7g6 d1e2 f8e7 c1h6 f6g4 h6g5 e8f8 e2e7 d8e7 g5e7 f8e7 h2h3 g4f6 e1d2 b8a6 g2g4 a8b8 b2b3 c8d7 a1e1 e7d8 g4g5 f6h5 c3e4 h8e8 e1e3 d8c7 g1e2 f7f5 g5f6 e8f8 e3f3 d7f5 e2g3 h5g3 e4g3 f8f6 f1d3 b8f8 d3f5 c7d8 h1e1 a6c7 h3h4 g6f5 a2a4 f5f4');
INSERT INTO evaluations VALUES(45925,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(45925,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(45925,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(45925,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(45925,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(45925,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(45925,36,'cp 149','b1c3 g7g6 d1e2 f8e7 c1h6 f6g4 h6g5 e8f8 e2e7 d8e7 g5e7 f8e7 h2h3 g4f6 e1d2 b8a6 g2g4 a8b8 b2b3 c8d7 g1e2 a6c7 a1e1 b7b5 g4g5 f6g8 e2g3 e7f8 h3h4 h7h6 f2f4 h6g5 f4g5 g8e7 h4h5 g6h5 h1h5 h8h5 g3h5 b5c4 h5f6 d7f5 f1c4 e7g6 c3e4');
INSERT INTO evaluations VALUES(45923,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(45923,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(45923,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(45923,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(45923,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(45923,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(45923,36,'cp 137','d8e7 f1e2 g7g6 g1f3 f8g7 c1f4 c8d7 e1g1 e8g8 f1e1 f8e8 e2f1 e7f8 d1b3 b7b6 b3a3 h7h6 h2h3 g6g5 f4h2 e8e1 a1e1 f6e8 h3h4 g5g4 f3d2 f7f5 c3e2 f8f6 h2f4 f6h4 e2g3 g7d4 d2b3 d4e5 f4e5 d6e5 e1e5 f5f4 g3f5 h4f6 f5e7 g8g7 e5e4 e8d6 e4e1 g4g3 f2f3');
INSERT INTO evaluations VALUES(46701,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(46701,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(46701,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(46701,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(46701,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(46701,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(46701,36,'cp 53','a7a6 c1e3 f8g7 d1d2 b7b5 e1c1 b8d7 h2h4 c8b7 g1h3 c7c5 g2g4 c5d4 e3d4 d7e5 f1e2 e5c6 d4e3 h7h5 g4h5 f6h5 c3d5 h5g3 h1g1 d8a5 g1g3 a5a2 d2d3 a2b2 c1d2 c6e5 d3b3 b2b3 c2b3 b7d5 e4d5 h8h4 h3g5 g7f6 e3b6');
INSERT INTO evaluations VALUES(51204,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(51204,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(51204,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(51204,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(51204,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(51204,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(51204,36,'cp -8','c4b3 d7d6 d2d3 g8f6 e1g1 b8d7 c1f4 c8b7 f1e1 f8e7 c2c3 e8g8 b1d2 f8e8 h2h3 h7h6 f4g3 e7f8 f3h2 g7g6 h2f1 a8b8 g3h2 b7c6 f1e3 d8b6 b3c2 b6b7 h2g3 b8d8 f2f4 d6d5 e4e5 f6h5 g3h2');
INSERT INTO evaluations VALUES(51872,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(51872,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(51872,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(51872,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(51872,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(51872,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(51872,36,'cp -37','g8f6 b1c3 f8c5 g1f3 d7d6 a4b3 e8g8 h2h3 a7a5 e1g1 b7b5 a2a3 b8d7 d2d3 h7h6 b3a2 f8e8 c1e3 c5b6 f3h4 d7f8 d1f3 f8h7 e3b6 d8b6 c3e2 b5b4 e2g3 h7g5 f3e3 b6e3 f2e3 c8e6 a2e6 g5e6 a3b4 a5b4');
INSERT INTO evaluations VALUES(44118,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(44118,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(44118,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(44118,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(44118,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(44118,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(44118,36,'cp -25','b1c3 a7a5 a2a3 f8c5 g1f3 d7d6 h2h3 b8d7 d2d3 e8g8 e1g1 h7h6 a4b3 c5b6 b3a2 a5a4 f3h4 d6d5 h4f3 d5e4 d3e4 f8e8 f3h4 d7f8 d1f3 f8e6 c1e3 b6e3 f3e3 e6f4 f1e1 d8c7 h4f3 c8e6 a2e6 e8e6 a1d1 c7e7');
INSERT INTO evaluations VALUES(54312,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(54312,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(54312,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(54312,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(54312,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54312,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(54312,36,'cp 35','d2d4 c8d7 g1f3 b8c6 a2a3 g8e7 f1d3 c5d4 c3d4 d8b6 d3c2 g7g5 h2h3 e7f5 c2f5 e6f5 c1g5 b6b2 b1d2 b2b5 a3a4 b5d3 d1e2 d3e2 e1e2 h8g8 g2g3 c6d4 f3d4 g8g5 h1b1 g5g6 b1b7 g6b6 b7c7 b6b4 d2f3 b4a4 a1a4 d7a4 d4f5 a4d7 f5e3');
INSERT INTO evaluations VALUES(55139,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(55139,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(55139,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(55139,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(55139,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55139,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55139,36,'cp 0','c7c5 e2e3 g8f6 d2d4 b8c6 c2c4 e7e6 c4d5 e6d5 f1b5 c5d4 e3d4 f8d6 b1c3 e8g8 e1g1 h7h6 f1e1 f8e8 e1e8 d8e8 b5d3 c8e6 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(56356,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(56356,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(56356,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(56356,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(56356,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(56356,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(56356,36,'cp -25','e5e4 f3d2 c7c6 b1c3 g8f6 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 e8g8 e1g1 b4c3 b2c3 b8d7 d1b3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 f8c8 f3d2 f5g6 c1a3 d7b6 a1c1 b6c4 d2c4 c8e8 c4e5 a8c8 c1c8 e8c8 e5g6 h7g6 f1c1 c8c1 a3c1 f6e4 c1a3');
INSERT INTO evaluations VALUES(56186,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56186,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(56186,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(56186,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(56186,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(56186,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(56186,36,'cp -14','f3d2 c7c6 b1c3 g8f6 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 b8d7 e1g1 b4c3 b2c3 e8g8 d1b3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 a8c8 f3d2 d7b6 c4f7 f8f7 f1f5 f7c7 c1a3 b6d5 e3e4 c7c2 d2f1 d5c3 f1e3 c3e2 g1f1 e2d4 e3c2 d4f5 e4f5 c8c2 a1c1 c2c1 a3c1 f6e4 c1a3 e4c3 a3c5');
INSERT INTO evaluations VALUES(791,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(791,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(791,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(791,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(791,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(791,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(791,36,'cp 82','b1c3 e7f5 c3a4 b6d8 b2b4 f8e7 c1b2 f7f6 f1d3 f6e5 d4e5 e8g8 e1g1 c8d7 a1c1 a8c8 d3b1 d7e8 g2g4 f5h6 h2h3 e8g6 b1g6 h7g6 d1d3 d8d7 a4c5 e7c5 c1c5 b7b6 c5c1 c6e7 f3d4 c8c1 b2c1 h6f7 f1e1 f8c8 f2f4 e7c6 d4c6');
INSERT INTO evaluations VALUES(789,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(789,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(789,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(789,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(789,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(789,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(789,36,'cp 93','e7f5 c3a4 b6d8 b2b4 f8e7 f1d3 b7b6 e1g1 e8g8 c1b2 c8d7 a4c3 f5d4 f3d4 c6d4 c3d5 d4f5 d3f5 e6d5 f5d3 a7a5 d1d2 d8e8 f1e1 d7e6 b2d4 e8a4 e1b1 f8b8 h2h3 a4e8 f2f4 e8d7 g1h2 g7g6 b1e1 a5b4 a3b4 a8a4 a1a4');
INSERT INTO evaluations VALUES(849,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(849,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(849,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(849,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(849,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(849,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(849,36,'cp 97','c3a4 b6d8 b2b4 f5h4 f3h4 d8h4 c1e3 f7f6 e5f6 h4f6 f1b5 f8d6 b5c6 b7c6 d1c1 c8d7 a4c5 f6f7 e1g1 e8g8 c1c3 a7a5 c5b7 d6c7 b7a5 c6c5 a5b3 c5c4 b3c5 d7a4 f2f4 c7d6 e3d2 f7g6 f1f3 f8f6 a1f1 g6f5 f3g3 a4b5 g3h3 g8h8');
INSERT INTO evaluations VALUES(850,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(850,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(850,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(850,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(850,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(850,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(850,36,'cp 78','b6d8 b2b4 c8d7 c1b2 f5h4 f3h4 d8h4 f1d3 b7b5 a4c5 f8c5 b4c5 a7a6 e1g1 e8g8 f2f4 f7f5 d1d2 f8b8 b2c3 h7h6 d2e3 d7e8 h2h3 e8g6 f1b1 h4e7 b1b2 b8b7 a1b1 a8b8 e3d2 e7c7 d2f2 c7e7 a3a4 b5a4 b2b7 b8b7');
INSERT INTO evaluations VALUES(12047,30,'cp -488',NULL);
INSERT INTO evaluations VALUES(12047,31,'cp -501',NULL);
INSERT INTO evaluations VALUES(12047,32,'cp -502',NULL);
INSERT INTO evaluations VALUES(12047,33,'cp -496',NULL);
INSERT INTO evaluations VALUES(12047,34,'cp -525',NULL);
INSERT INTO evaluations VALUES(12047,35,'cp -530',NULL);
INSERT INTO evaluations VALUES(12047,36,'cp -521','d8d4 e1g1 f8d6 e5c4 d4c4 b1c3 e8g8 d2d3 e4d3 d1d3 c4e6 c1d2 f6g4 h2h3 g4e5 d3e4 e5c4 e4e6 c8e6 c3e4 c4d2 e4d2 a8b8 b2b3 f8d8 f1d1 d6b4 d2f3 b4c3 a1c1 e6f5 f3e1 c6c5 e1d3 c5c4 d3c5 d8d1 c1d1 g8h8 c5a6 b8c8 b3b4 a5c6 g2g4 f5g6 d1d6 c6d4 d6g6 f7g6 a2a3');
INSERT INTO evaluations VALUES(716,30,'cp -506',NULL);
INSERT INTO evaluations VALUES(716,31,'cp -500',NULL);
INSERT INTO evaluations VALUES(716,32,'cp -496',NULL);
INSERT INTO evaluations VALUES(716,33,'cp -509',NULL);
INSERT INTO evaluations VALUES(716,34,'cp -503',NULL);
INSERT INTO evaluations VALUES(716,35,'cp -501',NULL);
INSERT INTO evaluations VALUES(716,36,'cp -509','e1g1 f8d6 e5c4 d4c4 b1c3 e8g8 d2d3 e4d3 d1d3 c4e6 c1d2 f6g4 h2h3 g4e5 d3e4 e5c4 e4e6 c8e6 c3e4 c4d2 e4d2 a8b8 b2b3 f8d8 f1d1 d6e7 d2f1 e7f8 d1e1 c6c5 e1e6 f7e6 c2c3 f8e7 a1e1 g8f7 e1e5 d8d1 b3b4 d1a1 b4a5 c5c4');
INSERT INTO evaluations VALUES(21058,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(21058,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21058,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(21058,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(21058,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21058,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(21058,36,'cp 0','g1f3 a7a6 a2a3 c8f5 c2c4 e7e6 d2d4 g8f6 b1c3 h7h6 c4d5 e6d5 f1d3 f5d3 d1d3 f8d6 c1d2 e8g8 f3e5 c6e7 e1g1 c7c6 a1c1 f6d7 c3a4 d7f6 a4c3');
INSERT INTO evaluations VALUES(32245,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(32245,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(32245,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(32245,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(32245,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(32245,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(32245,36,'cp 20','f1d3 g8f6 e1g1 f8d6 b1c3 e8g8 h2h3 a7a6 c3e2 f8e8 e2f4 e6d7 c2c4 d5c4 d3c4 e8e7 f4d5 f6d5 c4d5 d7c6 d5c6 b8c6 c1g5 f7f6 d1b3 g8h8 g5d2 b7b6 b3c4 d8d7 a1c1 c6d8 f1e1 e7e1 c1e1 d8c6 a2a3 c6e7 d2b4 d6b4 c4b4 a8e8 b4b3 d7d6 g1h1');
INSERT INTO evaluations VALUES(32243,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(32243,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(32243,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(32243,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(32243,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(32243,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(32243,36,'cp 24','g8f6 e1g1 f8d6 h2h3 e8g8 b1c3 a7a6 f1e1 c7c5 d4c5 d6c5 a2a3 f8e8 c1f4 b8c6 b2b4 c5f8 c3e2 h7h6 a1c1 a8c8 f4e3 f8d6 e2f4 e6d7 c2c4 d6f4 e3f4 d5c4 e1e8 d8e8 d3c4 d7e6 f4e3 c8d8 d1b3 e6c4 b3c4');
INSERT INTO evaluations VALUES(30025,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(30025,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(30025,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(30025,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(30025,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(30025,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(30025,36,'cp 21','e1g1 f8d6 h2h3 e8g8 b1c3 a7a6 f1e1 c7c5 d4c5 d6c5 a2a3 f8e8 b2b4 c5a7 c1e3 a7e3 e1e3 b8c6 c3e2 d8c7 d1d2 e6d7 a1e1 e8e3 d2e3 a8e8 e3c5 c7b8 e1d1 h7h6 e2g3 e8c8 d1e1 b8c7 c5e3 c8e8 e3d2 e8e1 d2e1');
INSERT INTO evaluations VALUES(30024,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(30024,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(30024,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(30024,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(30024,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(30024,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30024,36,'cp 31','f8d6 b1c3 e8g8 h2h3 a7a6 f1e1 c7c5 d4c5 d6c5 c3e2 f8e8 e2d4 b8c6 c1e3 d8b6 e1e2 e6d7 c2c3 h7h6 d1c2 c6d4 e3d4 e8e2 d3e2 a8e8 d4c5 b6c5 f3d4 f6e4 a1e1 e4d6 e1d1 d6e4 e2d3 c5b6 d1e1 e4d6 e1e8 d6e8 c2d2 e8d6');
INSERT INTO evaluations VALUES(40797,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40797,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40797,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40797,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(40797,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(40797,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40797,36,'cp 25','d3d4 e5d4 f3d4 e8g8 f1d3 a7a6 e1g1 c7c5 d4f3 b7b5 f1e1 f8e8 c3d5 f6d5 e4d5 c8g4 a2a4 c5c4 d3e2 g4f3 e2f3 b8d7 c1f4 e7f8 e1e8 d8e8 h2h3 d7c5 a4b5 a6b5 a1a8 e8a8 f4e3 a8d8 g2g3 g7g6 d1d2 d8f6');
INSERT INTO evaluations VALUES(48050,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(48050,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(48050,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(48050,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(48050,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(48050,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(48050,36,'cp 37','f8e7 h2h3 c7c5 d4c5 d8a5 f3d2 a5c5 c2c4 e8g8 b1c3 f8d8 a1c1 a7a5 f1e2 c8d7 e1g1 d7c6 d1c2 b8a6 a2a3 c5b6 c1b1 d8c8 e3e4 a6c5 b2b3 b6d8 f1d1 d8c7 f4e3 h7h6 b1c1 f6d7 c2b1 d7f6 b3b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(40888,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(40888,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40888,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40888,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40888,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40888,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(40888,36,'cp 40','h2h3 c7c5 d4c5 d8a5 f3d2 a5c5 c2c4 e8g8 b1c3 f8d8 a1c1 a7a5 f1e2 c8d7 e1g1 d7c6 d1c2 b8d7 d2b3 c5b6 f1d1 a5a4 b3d4 d7c5 f4g5 h7h6 g5h4 g7g5 d4c6 b7c6 h4g3 a8c8 e2f3 b6a6 c2e2 a6a5 g3h2 e7f8 c1b1');
INSERT INTO evaluations VALUES(51695,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(51695,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(51695,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(51695,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(51695,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(51695,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(51695,36,'cp 0','f1e2 g8f6 e1g1 c8b7 a2a4 a7a6 b2b3 b5b4 e4e5 f6d5 c3d5 c6d5 b3c4 d5c4 e2c4 f8e7 c1e3 b8c6 a1c1 c6a5 c4d3 a8c8 d1e2 c8c1 f1c1 e8g8 f3d2 d8a8 c1c7 e7d8 c7c1');
INSERT INTO evaluations VALUES(22860,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(22860,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(22860,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(22860,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(22860,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(22860,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(22860,36,'cp -17','d2d4 d7d5 g1f3 e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 f1e2 d8d7 a2a3 d5c4 d2c4 h7h5 b2b4 d7f7 b4b5 e6c4 b5c6 b7b6 e1g1 h5h4 f2f3 f7e6 c1b2 h4h3 e2c4 e6c4 f3e4 f5e4 f1f6 g7f6 d1g4 c4c6 c3e4 h8h6 d4d5 c6d7 e4f6 h6f6 g4d7 e8d7 b2f6 f8c5 g2h3 c5e3 g1g2');
INSERT INTO evaluations VALUES(23177,30,'cp 683',NULL);
INSERT INTO evaluations VALUES(23177,31,'cp 682',NULL);
INSERT INTO evaluations VALUES(23177,32,'cp 690',NULL);
INSERT INTO evaluations VALUES(23177,33,'cp 694',NULL);
INSERT INTO evaluations VALUES(23177,34,'cp 701',NULL);
INSERT INTO evaluations VALUES(23177,35,'cp 690',NULL);
INSERT INTO evaluations VALUES(23177,36,'cp 692','f3h4 h6h5 h4g6 d7d5 c4d5 h8h6 c1h6 g8h6 d4e5 c6b4 b1c3 f6e5 d1h5 d8f6 d5b3 c8g4 h5e5 f6e5 g6e5 a8d8 a2a3 b4a6 f2f3 g4c8 e1e2 a6c5 a1d1 f8d6 e5c4 c5b3 d1d6 c7d6 c2b3 d6d5 c3d5 c8e6 d5c7 e8d7 c7e6 d7e6 f3f4 h6f7 h1e1');
INSERT INTO evaluations VALUES(29945,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(29945,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(29945,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(29945,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(29945,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(29945,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(29945,36,'cp 32','d3d4 b8c6 f1b5 f6d7 d4d5 c6d4 b5e2 d4e2 d1e2 f8e7 h2h3 g4f3 e2f3 e7g5 c1g5 d8g5 h3h4 g5g6 a2a4 h7h5 h1h3 g6g4 a4a5 a7a6 f3g3 e8e7 g3h2 b7b5 f2f3 g4g6 e1c1 b5b4 c3a2 a8b8 h3g3 g6h6 c1b1 h6f4 a2c1 g7g6 c1d3 f4e3 h2g1 e3g1 d1g1 f7f5');
INSERT INTO evaluations VALUES(32254,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(32254,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(32254,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(32254,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(32254,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(32254,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(32254,36,'cp 135','b8c6 g1f3 c6d4 f1e2 g8e7 f3d4 d8d4 e1g1 e7f5 d2d3 d4d7 c3e4 f8e7 c1g5 h7h6 g5e7 d7e7 d1d2 e8g8 a1e1 b7b6 e2g4 e7d7 g4f5 e6f5 e4g3 f5g6 b2b3 f8e8 a2a4 a7a5 d2c3 c7c5 e1e8 a8e8 h2h3 e8e6 f1e1');
INSERT INTO evaluations VALUES(34766,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(34766,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(34766,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(34766,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(34766,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(34766,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(34766,36,'cp 205','f6e6 b1c3 g8f6 d1d2 c7c6 e1c1 f8e7 h2h4 h7h6 d4d5 e6d7 g5f6 e7f6 d5c6 b8c6 f1b5 e8g8 d2d6 d7e6 d6e6 c8e6 b5c6 b7c6 c3a4 f8d8 d1d8 a8d8 b2b3 e6g4 a4b2 g4f3 g2f3 f6e7 b2d3 f7f6 h4h5 a7a5 a2a4 g8f7 c1d2 g7g6 d2e2 g6h5 h1h5 f7g6 h5h3 d8d4 c2c3 d4d7 h3h1 h6h5');
INSERT INTO evaluations VALUES(34800,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(34800,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(34800,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(34800,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(34800,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(34800,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(34800,36,'cp 193','b1c3 g8f6 d1d2 f8e7 e1c1 c7c6 h2h4 h7h6 d4d5 e6d7 g5f6 e7f6 d5c6 b8c6 f1b5 e8g8 d2d6 d7e6 b5c6 b7c6 d6e6 c8e6 c3a4 f8d8 d1d8 a8d8 b2b3 e6g4 a4b2 h6h5 a2a4 g4f3 g2f3 f6e7 b2d3 f7f6 c1d2 e7c5 d2e2 c5d4 h1g1 a7a5 f3f4 e5f4 d3f4 g8f7 f4h5');
INSERT INTO evaluations VALUES(46551,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(46551,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(46551,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(46551,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(46551,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(46551,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(46551,36,'cp 134','c2c4 c7c5 f1e2 c5d4 f3d4 e7c6 c1e3 f8e7 e1g1 e8g8 b1c3 b8d7 f2f4 a7a6 d1d2 c6d4 e3d4 b7b6 a1d1 c8b7 f4f5 e6e5 d4f2 d7f6 d2e3 a8c8 a2a4 f6d7 e3d3 d7c5 d3c2 d8c7 e2f3 h7h6 h2h3 e7g5 h3h4 g5f6 g2g4 c5d7 b2b3 d7c5 b3b4 c5d7');
INSERT INTO evaluations VALUES(46547,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(46547,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(46547,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(46547,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(46547,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(46547,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(46547,36,'cp 112','c7c5 f1e2 c5d4 f3d4 e7c6 e1g1 f8e7 c1e3 e8g8 b1c3 b8d7 f2f4 a7a6 a1c1 f8e8 g1h1 d8c7 h2h3 c6d4 d1d4 b7b6 c1d1 a8b8 a2a4 e8f8 d4d2 h7h6 f4f5 c8b7 e3f4 d7e5 b2b3 b8c8 f4g3 b7c6');
INSERT INTO evaluations VALUES(47068,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(47068,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(47068,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(47068,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(47068,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47068,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(47068,36,'cp -13','d6e5 f4e5 c8f5 g1f3 e7e6 b1c3 f8b4 c1d2 b4c3 b2c3 c7c5 f1b5 b8c6 c3c4 d5e7 e1g1 h7h6 d2c3 c5d4 f3d4 e8g8 b5c6 e7c6 d4f5 e6f5 f1f5 d8b6 f5f2 a8d8 d1f3 c6d4 f3e3 b6c5 a1f1 c5c4 c3d4 c4d4 e3d4 d8d4 e5e6 f7f6 f1b1 f8e8 b1b7 e8e6 b7a7 e6e1 f2f1');
INSERT INTO evaluations VALUES(47477,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(47477,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47477,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(47477,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(47477,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47477,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(47477,36,'cp 0','f4e5 c8f5 g1f3 e7e6 f1d3 f5d3 d1d3 c7c5 c2c4 d5b4 d3b3 d8b6 a2a3 b4c6 b3b6 a7b6 d4d5 e6d5 b1c3 d5c4 c3d5 a8a7 e1g1 f8e7 c1g5 e7g5 f3g5 c6e5 a1e1 f7f6 g5e6 e8d7 e6g7 d7c6 f1f6 c6d5 g7f5 h8d8 f5e7 d5d4 e7f5');
INSERT INTO evaluations VALUES(47475,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47475,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47475,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47475,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47475,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47475,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47475,36,'cp -18','c8f5 g1f3 e7e6 f1b5 c7c6 b5d3 f5d3 d1d3 c6c5 c1g5 d8d7 b1c3 h7h6 g5h4 b8c6 c3d5 d7d5 c2c4 d5d7 d4d5 g7g5 h4g3 e6d5 c4d5 c6b4 d3e2 b4d5 e1g1 e8c8 a1c1 f8e7 f3d2 d5f4 g3f4 g5f4 d2c4 d7d3 e2g4 c8b8 f1f4 e7g5 c1c3 d3d7 g4d7 d8d7 f4f1 h8d8');
INSERT INTO evaluations VALUES(47192,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(47192,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(47192,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(47192,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(47192,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(47192,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(47192,36,'cp -32','c8g4 g1f3 b8d7 f1g2 e7e6 c2c4 c7c6 d1b3 d8b6 e1g1 f8e7 b1c3 e8g8 c4d5 e6d5 b3c2 g7g6 f3h4 f6h5 h2h3 g4e6 h4f3 f8e8 f1d1 h5g7 b2b3 e6f5 c2e2 h7h5 c1b2 a8d8 c3a4 b6c7 a1c1 a7a5 f3e5 d7e5');
INSERT INTO evaluations VALUES(49185,30,'cp -466',NULL);
INSERT INTO evaluations VALUES(49185,31,'cp -438',NULL);
INSERT INTO evaluations VALUES(49185,32,'cp -447',NULL);
INSERT INTO evaluations VALUES(49185,33,'cp -462',NULL);
INSERT INTO evaluations VALUES(49185,34,'cp -475',NULL);
INSERT INTO evaluations VALUES(49185,35,'cp -459',NULL);
INSERT INTO evaluations VALUES(49185,36,'cp -455','e8f7 f3e5 f7g8 e1g1 b8c6 d2d3 e4f6 d3d4 d7d6 e5c6 b7c6 f1e1 h7h6 d1f3 d6d5 h2h3 g7g5 b1d2 h8h7 f3d3 g5g4 h3h4 c8d7 g2g3 h7e7 e1e5 f8g7 d2b3 f6e4 c1f4 g7e5 d4e5 d8e8 d3e3 e8g6 b3c5 e4c5 e3c5 e7e8 c5e3 g8h7 a1c1');
INSERT INTO evaluations VALUES(54617,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(54617,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(54617,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(54617,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(54617,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(54617,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(54617,36,'cp -14','d2d4 e5e4 g1h3 b8c6 b1c3 h7h6 e1g1 g8f6 f2f3 f8b4 f3e4 d5e4 d4d5 b4c3 d5c6 c3a5 d1d8 a5d8 h3f2 b7c6 f2e4 d8b6 g1h1 f6e4 g2e4 c8d7 c1f4 e8c8 a1c1 g7g5 f4d6 h8e8 e4f3 g5g4 f3d5 e8e2 f1f7 b6e3 c1c6 d7c6 f7c7 c8b8 c7c6 b8a8');
INSERT INTO evaluations VALUES(54605,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(54605,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(54605,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(54605,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(54605,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54605,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(54605,36,'cp -18','e5e4 g1h3 f8b4 c1d2 b4d2 d1d2 g8f6 b1c3 e8g8 e1g1 f8e8 h3f4 b8c6 e2e3 c8f5 f2f3 a8c8 d2f2 c6b4 f3e4 f5e4 f4h5 c8c6 g2e4 d5e4 h5f6 c6f6 f2d2 b4d5 c3d5 f6f1 a1f1 d8d5 b2b3 h7h6 f1c1 d5d7 c1c5');
INSERT INTO evaluations VALUES(54588,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(54588,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(54588,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(54588,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(54588,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54588,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(54588,36,'cp -17','g1h3 b8c6 e1g1 h7h6 b1c3 g8f6 f2f3 f8b4 f3e4 d5e4 d4d5 b4c3 d5c6 d8b6 h3f2 c3d4 e2e3 d4e5 f2e4 e8g8 c6b7 c8b7 e4f6 e5f6 d1b3 b6b3 a2b3 f8b8 a1a3 b7g2 g1g2 f6e7 a3a4 b8b3 f1f2 e7c5 f2c2 c5e3 c1e3 b3e3 b2b4 g7g5 c2a2 e3b3 h2h4 g8g7 h4g5 h6g5 a4a5 g7g6 a5a6 g6h5 a6a7 a8e8 a7f7 e8e3 g2f1');
INSERT INTO evaluations VALUES(7372,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(7372,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(7372,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(7372,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(7372,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(7372,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(7372,36,'cp 148','a4c6 b7c6 f3e5 c8b7 e1g1 e8g8 e5c4 a8b8 c4e3 d7d5 b1d2 e7d6 e3f5 f8e8 f5d6 c7d6 f2f3 d5e4 d2e4 f6e4 f3e4 b7c8 b2b3 d6d5 d1f3 f7f6 c1f4 b8b7 e4d5 d8d5 f3d5 c6d5 a1e1 e8e1 f1e1');
INSERT INTO evaluations VALUES(7362,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(7362,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(7362,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(7362,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(7362,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(7362,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(7362,36,'cp 139','b7c6 f3e5 a6a5 e1g1 e8g8 b1c3 c8b7 f1e1 a5a4 d3d4 d7d6 e5d3 c6c5 d4c5 d6c5 c1e3 f6e4 c3e4 b7e4 e3c5 e7c5 e1e4 c5d4 a2a3 d8d5 d1e2 a8d8 a1d1 d5c6 h2h4 f8e8 d1e1 c6e4 e2e4 e8e4 e1e4 c7c5 b2b3 f7f5 e4e1 d4f6 b3a4 c5c4 d3e5 f6h4 e5c4');
INSERT INTO evaluations VALUES(14244,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(14244,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(14244,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(14244,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(14244,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(14244,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(14244,36,'cp -28','c6b4 g1f3 b4d3 c2d3 c8f5 b1c3 e7e6 e1g1 f8e7 d1e2 f5g4 g1h1 e8g8 h2h3 g4f3 e2f3 c7c5 d4c5 e7c5 e3e4 a8c8 f1d1 d8b6 c3a4 d5e4 d3e4 b6b4 a4c5 b4c5 e4e5 f6d5 f4f5 c5c2 f5f6');
INSERT INTO evaluations VALUES(17277,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(17277,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(17277,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(17277,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(17277,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(17277,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(17277,36,'cp -18','e7e5 c2c4 f8c5 f1e2 d7d6 h2h3 e8g8 e1g1 c8e6 b1c3 a7a6 c1e3 c6d4 e3d4 e5d4 c3d5 a6a5 d5f6 d8f6 f3d2 a5a4 e2g4 a4a3 a1b1 a3b2 a2a4 c5a3 d1b3 e6g4 h3g4 f6f4 b3a3 f4d2 b1b2 a8a4 a3b3 d2c3 b3b7 c3d3 b7c7 a4c4 c7d6');
INSERT INTO evaluations VALUES(21068,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(21068,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(21068,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(21068,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(21068,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(21068,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(21068,36,'cp -60','e7e5 d2d3 c8e6 b1d2 d8d7 a2a3 f7f6 f1g2 h7h5 h2h4 g8h6 g1f3 f8e7 b3b4 c6d8 e1g1 c7c6 c2c3 d8f7 e2e4 e8g8 e4d5 c6d5 c3c4 a8c8 c4d5 e6d5 f1e1 f8d8 d3d4 e5d4 d2e4 h6g4 f3d4');
INSERT INTO evaluations VALUES(35294,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35294,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(35294,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(35294,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(35294,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(35294,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(35294,36,'cp 75','b1c3 g8f6 g1e2 d7d5 c3d5 f6d5 c4d5 c7c6 e1g1 c6d5 e4d5 e7d6 f1e1 f8e7 b2g7 h8g8 g7c3 c8g4 d1d4 e8f8 e2g3 b8c6 d4e4 c6b4 h2h3 g4d7 a1d1 a8e8 e4h7 d6g6 h7g6 g8g6 e1e7 f8e7 c3b4 e7d8 a2a3 e8g8 g1h2 b7b6 b4c3 d8e7 c3e5 f7f6 e5f4 g8c8 d1e1 e7f8 f2f3');
INSERT INTO evaluations VALUES(35290,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(35290,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(35290,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(35290,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(35290,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(35290,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(35290,36,'cp 62','g8f6 g1e2 d7d5 c3d5 f6d5 c4d5 c7c6 e1g1 c6d5 e4d5 e7b4 b2c3 b4d6 f1e1 f8e7 c3g7 h8g8 g7c3 c8g4 d1d4 e8f8 e2g3 b8c6 d4e4 c6b4 h2h3 g4d7 a1d1 a8e8 e4h7 d6g6 h7g6 g8g6 e1e7 f8e7 c3b4 e7d8 g1h2 e8g8 a2a3 d7a4 d1g1 b7b6 g3f5 a4b3 f5e3 d8d7 g2g4 g8c8');
INSERT INTO evaluations VALUES(43143,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(43143,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(43143,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(43143,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(43143,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(43143,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(43143,36,'cp -43','f6e4 c4f7 e8f7 c3e4 d7d5 e4g3 c7c5 c2c3 h8f8 e1g1 b8c6 d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 c1e3 f7g8 d1d2 c8a6 f1e1 e5e4 a1d1 d8c7 b2b3 a6d3 e3c5 f8f7 f2f3 a8e8 c5d4 c6c5 d4g7 g8g7 d2g5 c7e5 g5e5 e8e5 f3e4');
INSERT INTO evaluations VALUES(51949,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(51949,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51949,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(51949,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(51949,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(51949,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51949,36,'cp 0','c5d4 d1d4 g8f6 d4d3 b8c6 g1f3 d7d5 c4d5 c6b4 d3d1 f6d5 c3d5 d8d5 a2a3 d5d1 e1d1 b4c6 e2e3 f7f6 c1d2 e8f7 f1c4 c8d7 a1c1 a8c8 d1e2 f8e7 d2c3 c8c7 c1c2 h8c8 h1c1 e7d6 c1d1 d6e7');
INSERT INTO evaluations VALUES(52051,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52051,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52051,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52051,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52051,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52051,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52051,36,'cp 0','d1d4 b8c6 d4d3 g8f6 g1f3 d7d5 c4d5 e6d5 c1e3 c8g4 a1d1 f8b4 a2a3 b4c3 d3c3 e8g8 e3g5 a8c8 g5f6 g7f6 c3d2 g8h8 f3d4 d8b6 h2h3 c6d4 d2d4 b6d4 d1d4 g4e6 g2g4 c8c2 d4d2 c2c1 d2d1 c1c2');
INSERT INTO evaluations VALUES(52049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52049,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52049,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52049,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52049,36,'cp -13','g8f6 d4d3 b8c6 g1f3 d7d5 e2e3 f8e7 f1e2 e8g8 e1g1 d5c4 d3c4 c8d7 c1d2 a8c8 a1c1 d8c7 c3e4 c7b6 e4f6 e7f6 d2c3 c6a5 c4b4 f6c3 b2c3 b6c7 c3c4 b7b6 f1d1 f8d8 b4e7 a5c6 e7a3 c6a5');
INSERT INTO evaluations VALUES(52008,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52008,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(52008,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(52008,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52008,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(52008,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52008,36,'cp 23','g1f3 d7d6 d2d4 c5d4 f3d4 g8f6 f1d3 b8c6 d4c6 b7c6 e1g1 f8e7 c2c4 d6d5 c4d5 c6d5 b1c3 d5e4 c3e4 e8g8 e4f6 e7f6 d1f3 a8b8 f1d1 d8e7 c1f4 e6e5 f4d2 e5e4 d3e4 b8b2 a1b1 b2b1 d1b1 f8d8 d2a5 d8e8 b1e1 c8e6 e4f5 f6d4 f3c6 e8b8 f5e6 f7e6 a5c3 d4c3 c6c3');
INSERT INTO evaluations VALUES(52007,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52007,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(52007,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(52007,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52007,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(52007,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52007,36,'cp 22','b8c6 d2d4 c5d4 f3d4 f8c5 d4b3 c5e7 c2c4 e7f6 f2f4 d7d5 e4d5 e6d5 c4d5 c6e7 b1c3 f6c3 b2c3 d8d5 d1d5 e7d5 c3c4 d5f6 a3a4 c8f5 b3d4 g8e7 d4f5 e7f5 f1d3 f5d4 e1g1 e8c8 c1b2 h8e8 f1e1');
INSERT INTO evaluations VALUES(18848,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(18848,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(18848,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(18848,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(18848,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(18848,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(18848,36,'cp 10','d2d4 c5d4 f3d4 f8c5 d4b3 c5b6 b1c3 g8e7 c1g5 e8g8 d1d2 h7h6 g5h4 d7d5 e4d5 e6d5 e1c1 c8e6 c1b1 d5d4 c3a4 f8e8 f1d3 d8d5 h4e7 e8e7 a4b6 a7b6 f2f3 e7d7 d2e2 c6a5 b3c1 a8d8 d1e1 a5c4 d3c4 d5c4 e2c4');
INSERT INTO evaluations VALUES(52138,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(52138,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(52138,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(52138,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(52138,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(52138,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(52138,36,'cp 36','g1f3 e5f4 f1c4 b8c6 d2d3 g8f6 b1c3 f8d6 e1g1 e8g8 c1d2 d6b8 d1c1 f6h5 c3d5 c6e5 d5f4 e5c4 d3c4 b8f4 d2f4 d8e7 c1d2 h5f4 d2f4 b7b6 e4e5 f7f6 f4e4 c8a6 a1d1 a8e8 b2b3 e7e6 e4d5 f8f7 f1e1 f7e7 e5f6 e6d5 c4d5');
INSERT INTO evaluations VALUES(52137,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52137,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(52137,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(52137,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(52137,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52137,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52137,36,'cp 38','e5f4 f1c4 b8c6 d2d3 g8f6 b1c3 f8d6 e1g1 e8g8 c1d2 d6b8 d1c1 d7d5 e4d5 c6b4 d2f4 b4d5 f4b8 a8b8 c3d5 f6d5 a2a4 d8d6 c1e1 c8f5 f3g5 f5g6 g5e4 g6e4 e1e4 d5f6 e4f4 d6f4 f1f4 f8e8 d3d4 c5d4 f4d4 e8e7 a1d1 b8e8 b2b3');
INSERT INTO evaluations VALUES(52157,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52157,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(52157,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52157,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52157,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52157,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(52157,36,'cp 39','f1c4 b8c6 d2d3 g8f6 b1c3 f8d6 c1d2 e8g8 e1g1 a7a6 a2a4 b7b6 d1e1 d8c7 a4a5 b6a5 e1h4 c6e7 e4e5 d6e5 f3e5 c7e5 h4f4 e5f4 d2f4 d7d5 f4d6 f8e8 c4b3 c5c4 d3c4 d5c4 b3c4 c8e6 c4e6 f7e6 a1a5 e7d5 h2h3 d5e3 f1f3 e3c2 f3f1');
INSERT INTO evaluations VALUES(52143,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(52143,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(52143,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(52143,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52143,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(52143,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52143,36,'cp 18','b8c6 d2d3 g8f6 b1c3 f8d6 e1g1 e8g8 c1d2 d6b8 d1c1 d7d5 e4d5 c6e7 d2f4 e7d5 f4b8 a8b8 c3d5 f6d5 a2a4 h7h6 c1d2 c8e6 a1e1 d8d6 d2f2 b7b6 f3e5 b8e8 c4b5 e8d8 f2h4 d5e7 b5c4 e6c4 e5c4 e7g6 c4d6 g6h4 d6c4 h4g6 b2b3 d8d7');
INSERT INTO evaluations VALUES(19031,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(19031,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19031,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(19031,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(19031,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(19031,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(19031,36,'cp 31','d2d4 c5d4 c1f4 g8f6 e1g1 d7d5 e4d5 f6d5 f3g5 c8e6 g5e6 f7e6 d1h5 g7g6 h5h3 f8e7 f4g3 d8d7 b1d2 e8c8 a1e1 e6e5 g3e5 d7h3 g2h3 h8e8 e5g3 d4d3 c2d3 e7c5 g1h1 e8e1 f1e1 c5b4 e1e2 c6d4 e2e5 d5b6 d2e4 b6c4 d3c4 d4f5 g3f2 d8d1 h1g2 d1c1 g2f3 c1c2 f2a7 c2b2 e5e8 c8d7 e4f6 d7d6 e8e2 b2e2 f3e2 d6e6 f6d5 b4e7');
INSERT INTO evaluations VALUES(52442,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(52442,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(52442,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(52442,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(52442,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(52442,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(52442,36,'cp -19','g7g6 g1f3 f8g7 e1g1 g8f6 f1e1 e8g8 c2c3 d7d5 e4e5 f6e8 h2h3 e8c7 d3f1 d5d4 b2b4 d4c3 b4c5 c3d2 b1d2 c8e6 a1b1 e6a2 b1b2 a2d5 d2c4 d8c8 c4e3 f8d8 e3d5 c7d5 d1b3 b8c6');
INSERT INTO evaluations VALUES(53353,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(53353,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(53353,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(53353,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(53353,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(53353,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(53353,36,'cp -36','d5d4 g2g3 g8f6 f1g2 b8c6 d2d3 f6d7 b2b4 c6b4 e1g1 e7e5 f1e1 f7f6 f3d4 d7c5 d4b3 b4d3 e1f1 d3c1 d1d8 e8d8 f1c1 c5b3 a2b3 f8b4 b1c3 b4c3 c1c3 d8c7 b3b4 c8d7 c3a3 a7a6 b4b5 c7b6 b5a6 a8a6 c4c5 b6c5 a3a6 b7a6 a1a6');
INSERT INTO evaluations VALUES(53454,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(53454,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(53454,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(53454,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(53454,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(53454,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(53454,36,'cp -5','g2g3 g8f6 f1g2 b8c6 d2d3 f6d7 b1d2 e7e5 a2a3 a7a5 d2e4 f8e7 f3h4 d7f6 e4f6 e7f6 g2e4 g7g6 h4g2 f6g7 e1g1 c8h3 c1d2 e8g8 b2b4 d8d7 d1e2 c6e7 f1e1 h3g4 e4f3 g4f3 e2f3 a5b4 a3b4 a8a1 e1a1 b7b5 c4b5 d7b5 a1a8 f8c8 a8c8 e7c8');
INSERT INTO evaluations VALUES(3300,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3300,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3300,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(3300,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(3300,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3300,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3300,36,'cp 25','a2a3 c8e6 f1e1 a7a5 c1d2 f8e8 c3d5 a5a4 b2b4 a4b3 c2b3 f6d7 a3a4 c6d4 d2e3 e7f6 b3b4 c7c6 d5f6 d8f6 e3d4 e5d4 c4e6 f7e6 e4e5 d7e5 f3e5 d6e5 d1g4 a8d8 g4g3 d8d5 e1e4 e8f8 a1e1 f6e7 e4g4 e7f7 e1e5 d5e5 g3e5 f7f2 g1h2');
INSERT INTO evaluations VALUES(3308,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(3308,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(3308,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(3308,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3308,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(3308,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(3308,36,'cp 25','c8e6 f1e1 a7a5 c1d2 f8e8 c3d5 a5a4 b2b4 a4b3 c2b3 e7f8 b3b4 c6b8 d5c3 c7c6 c4e6 e8e6 d3d4 e6e8 d4d5 c6d5 e4d5 b8d7 d2e3 a8c8 d1b3 d7b6 f3d2 c8c3 b3c3 b6d5 c3d3 d5e3 e1e3 d6d5 d3b5 e5e4 a1d1 d5d4 d2f1 e8e7 b5c4 d4d3 c4c3 f6h5');
INSERT INTO evaluations VALUES(5285,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(5285,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(5285,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(5285,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(5285,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(5285,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(5285,36,'cp -109','e7e6 d2d3 g8e7 c4b3 h7h6 g5f3 e8g8 e1g1 b7b6 f1e1 c8b7 h2h3 g8h7 c1f4 d7d6 c3b5 e6e5 f4d2 f7f5 b5c3 f5e4 d3e4 c6d4 f3d4 e5d4 c3d5 b6b5 c2c4 e7d5 e4d5 d8h4 f2f4 b7a6 c4b5 a6b5 d1g4 h4g4 h3g4 c5c4 a2a4');
INSERT INTO evaluations VALUES(5093,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(5093,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(5093,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(5093,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(5093,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(5093,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(5093,36,'cp -99','d2d3 h7h6 g5f3 g8e7 e1g1 e8g8 f1e1 b7b6 c4b3 c8b7 c3b5 d8b8 b5c3 g8h7 c3e2 c6e5 f3e5 b8e5 c2c3 e5c7 c1f4 e6e5 f4g3 a8d8 c3c4 f7f5 f2f3 f5f4 g3f2 d7d6 g1h1 e7c6 e2c3 g6g5 b3a4 c7f7 c3b5');
INSERT INTO evaluations VALUES(5094,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(5094,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(5094,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(5094,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(5094,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(5094,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(5094,36,'cp -96','g8e7 e1g1 e8g8 h2h3 b7b6 f1e1 h7h6 g5f3 g8h7 c4b3 c8b7 c1f4 d7d6 f4e3 a7a6 a2a3 d6d5 e4d5 c6d4 b3a2 a8a7 e3f4 d4f3 d1f3 e6d5 d3d4 g7d4 a1d1 b7c6 f4e5 d4e5 e1e5 d8d6 d1e1 d5d4 c3d5 e7d5 a2d5 c6d5 e5d5 d6c6 b2b4 a7d7 d5d7 c6d7 b4c5');
INSERT INTO evaluations VALUES(5614,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(5614,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(5614,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(5614,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(5614,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(5614,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(5614,36,'cp 136','c2c3 f7f6 f1d3 g8h6 e5f6 e7f6 e1g1 e8g8 d1e2 d8d6 b1a3 a7a6 c1h6 g7h6 a1e1 d6e7 a3c2 e7g7 a2a3 a6a5 g2g3 c8d7 c2e3 a8e8 d3b5 g8h8 e2d1 f6e7 e3g2 c6e5 f3e5 d7b5 e5d3 e7d6 f2f4 b5d3 d1d3 g7f7 e1e3 c7c5 f1e1 c5d4 c3d4 d6c7 d3d2');
INSERT INTO evaluations VALUES(11679,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(11679,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(11679,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(11679,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(11679,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(11679,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(11679,36,'cp 138','d2d4 d7d6 d4e5 d6e5 d1d8 e8d8 b3f7 f8d6 f1e1 h8f8 f7b3 c6a5 b1c3 a5b3 c2b3 c8b7 a2a3 a6a5 c3b5 f6e4 c1e3 b7c6 a3a4 d8d7 a1d1 d7e6 d1c1 c6b5 a4b5 f8b8 c1c4 e4g5 e3g5 h6g5 f3g5 e6d7 c4g4 b8b5 h2h4 b5b3 g5e4 b3b2 g4g7 d7e6 g7g6 e6d5 e4d6 c7d6 e1d1 d5e4');
INSERT INTO evaluations VALUES(11876,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(11876,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(11876,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(11876,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(11876,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(11876,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(11876,36,'cp 38','f3d4 c6c5 d4c6 d8d7 c6a5 f8e7 e1g1 c7c6 c1f4 e8g8 e4e5 d6e5 d1d7 c8d7 f4e5 e7d8 a5b7 d7f5 a1c1 f8e8 b7d8 a8d8 e5f6 g7f6 f2f3 a7a5 f1f2 g8g7 c3a4 e8e5 c2c4 h7h5 b2b3 h5h4 f2b2 h4h3 g2h3 f5h3 g1f2 h3e6 b2e2 e5g5');
INSERT INTO evaluations VALUES(11873,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(11873,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(11873,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11873,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(11873,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(11873,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11873,36,'cp 37','c6c5 d4c6 d8d7 c6a5 f8e7 e1g1 e8g8 c1f4 c7c6 e4e5 d6e5 f4e5 e7d8 d1d7 c8d7 a5b7 d7f5 f1e1 f5c2 e5d6 f8e8 e1e8 f6e8 b7d8 a8d8 d6c5 a7a6 f2f3 c2f5 c5e3 e8d6 a1d1 f7f6 d1d4 d6f7 d4a4 f5d3 a4a5 f7e5 b2b3 h7h5 a5c5 d3g6 c3a4 d8d5 c5c1');
INSERT INTO evaluations VALUES(11882,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11882,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(11882,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(11882,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11882,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(11882,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(11882,36,'cp 28','d4c6 d8d7 c6a5 f8e7 e1g1 e8g8 f1e1 c7c6 c1f4 d7c7 a5c4 f8d8 d1e2 c8g4 f2f3 g4e6 e4e5 d6e5 f4e5 c7b7 a1b1 e6c4 e2c4 f6d5 b2b3 e7f8 c3e4 b7b5 c4f1 b5f1 g1f1 f7f6 e5g3 f6f5 e4g5 d5c3 b1a1 d8d2 g5e6 d2c2 e6f8 a8f8 g3d6 f8c8 d6c5 c3a2 e1e7 c2c5');
INSERT INTO evaluations VALUES(14161,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(14161,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14161,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(14161,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(14161,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(14161,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(14161,36,'cp 0','e7f6 e2e3 a7a6 f1d3 c8e6 g1f3 f8d6 e3e4 e6g4 d3e2 d5e4 c3e4 e8g8 c2c3 f8e8 e4d6 d8d6 e1g1 g4h5 f1e1 e8e7 d1d2 a8e8 h2h3 g7g5 b2b4 g8g7 a2a4 h5g6 e2f1 e8d8 a1d1 h7h5 a4a5 g6e4 f3h2 e4g6');
INSERT INTO evaluations VALUES(15236,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15236,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15236,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15236,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15236,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15236,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15236,36,'cp 0','e2e3 h7h5 f1d3 c8e6 g1e2 f8d6 a2a3 a7a6 e3e4 d5e4 d3e4 f6f5 e4f3 e6c4 b2b3 c4e2 c3e2 g7g6 b3b4 e8g8 e1g1 f8e8 d1d2 h5h4 c2c3 d8f6 g2g3 g8g7 e2f4 h4g3 h2g3 d6f4 d2f4 e8e7 a1e1 a8e8 e1e7 e8e7');
INSERT INTO evaluations VALUES(37575,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(37575,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(37575,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(37575,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(37575,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(37575,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(37575,36,'cp 82','d7d6 b1c3 g8f6 c1g5 h7h6 g5e3 e7e5 d4e5 f6g4 d1d5 c7c6 d5d6 g4e3 f2e3 d8d6 e5d6 b8d7 e1c1 c8b7 h2h4 g7c3 b2c3 c6c5 e4e5 e8c8 f1d3 b7f3 g2f3 d7e5 d3e2 e5d7 h4h5 d8e8 f3f4 e8e3 e2d3 c8b8 c1c2 e3e6 f4f5 e6d6 f5g6 f7g6 h5g6 b8c7');
INSERT INTO evaluations VALUES(37549,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(37549,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(37549,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(37549,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(37549,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(37549,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(37549,36,'cp 77','c1e3 g8f6 b1c3 e8g8 f1e2 e7e5 d4e5 d6e5 d1c2 d8e8 e1g1 c7c5 c3d5 c8b7 e3g5 e8e6 f1d1 b8c6 d5c7 e6c8 c7a8 c8a8 g5f6 g7f6 e2d3 a8d8 c2e2 d8e7 d1d2 f8d8 a1d1 c6d4 f3d4 d8d4 d3c2');
INSERT INTO evaluations VALUES(39069,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(39069,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(39069,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(39069,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(39069,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(39069,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(39069,36,'cp -58','h2h4 h7h6 d2d4 b8c6 b1c3 d7d6 d1d3 g5g4 f3g1 f4f3 g2f3 g7d4 c1f4 d4c3 d3c3 d8f6 f4d2 h8h7 e1c1 c8e6 c4d3 f6c3 d2c3 c6e5 d3e2 e8c8 f3f4 e5d7 f4f5 e6a2 b2b3 g8f6 c3f6 d7f6 e2d3 d8e8 c1b2 a2b3 c2b3 f6e4 d1e1 d6d5 d3e4 h7h8 g1e2 e8e4 e2g3 e4e1 h1e1 c8d7 g3h5');
INSERT INTO evaluations VALUES(52518,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(52518,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52518,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(52518,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(52518,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52518,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(52518,36,'cp -30','b8c6 d4d1 e7e5 c2c4 g8f6 b1c3 f8b4 c1d2 e8g8 g1e2 d7d5 c3d5 f6d5 c4d5 d8d5 d2b4 c6b4 e2c3 d5a5 a2a3 f8d8 d1b3 b4d5 f1c4 c8e6 e1g1 d5c3 b2c3 e6c4 b3c4 a8c8 c4b4 a5d5 h2h3 b7b6 b4a4 d5d7 a4a6 c8c3 f1d1 d7d1 a1d1 d8d1 g1h2 d1d7');
INSERT INTO evaluations VALUES(19550,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(19550,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(19550,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(19550,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(19550,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(19550,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(19550,36,'cp -25','d4d1 d7d5 c2c4 d5d4 e3d4 d8d4 d1d4 c6d4 f1d3 e7e5 b1c3 c8e6 c1e3 g8f6 g1f3 d4f3 g2f3 f8b4 a1c1 e8c8 d3e2 c8b8 f3f4 f6g4 e2g4 e6g4 e1g1 b4d6 c3b5 a7a6 b5d6 d8d6 f1e1 d6g6 g1f1 e5e4 c1c3 g4f5 a2a3 f5h3 f1e2 g6g2 b2b3 g2h2');
INSERT INTO evaluations VALUES(19573,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(19573,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(19573,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(19573,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(19573,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(19573,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(19573,36,'cp -32','d7d5 c2c4 d5d4 e3d4 d8d4 d1d4 c6d4 f1d3 e7e5 b1c3 c8e6 c1e3 g8f6 g1e2 e8c8 a1d1 f6g4 e3g5 f7f6 g5d2 f8c5 e1g1 f6f5 h2h3 d4e2 d3e2 g4f6 d2g5 d8d1 f1d1 h7h6 g5f6 g7f6 c3d5 c5d4 b2b3 e6d5 c4d5 h8d8 e2c4');
INSERT INTO evaluations VALUES(54536,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(54536,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(54536,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(54536,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(54536,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(54536,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(54536,36,'cp -40','f1b5 b8c6 b1c3 g8f6 b5c6 b7c6 d2d3 d8c7 e1g1 f8e7 f3e5 e8g8 d1f3 c8a6 c1d2 f6d7 a1e1 a8e8 b2b3 e7f6 e5d7 c7d7 c3a4 f6d4 g1h1 c5c4 c2c3 d4b6 d3c4 d5c4 e1e8 f8e8 f1e1 e8c8 d2e3 c6c5 e3f2 h7h6 h1g1 a6b7 f3g3 d7f5 a4b6 a7b6 b3c4 b7e4 h2h3 c8e8');
INSERT INTO evaluations VALUES(54466,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(54466,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(54466,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(54466,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(54466,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(54466,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(54466,36,'cp -38','b8c6 b1c3 g8f6 b5c6 b7c6 e1g1 f8e7 d2d3 d8c7 f3e5 e8g8 d1f3 c8a6 f1e1 f8e8 c1d2 f6d7 a1d1 e7f6 e5d7 c7d7 b2b3 c5c4 d3c4 f6d4 g1h1 d5c4 d2e3 c6c5 f3d5 d7d5 c3d5 a8d8 d5c7 d4e3 c7a6 d8d1 e1d1 e3f4 b3c4 f7f5 h1g1 f4e3 g1f1 e8e4 a6c7 e4c4');
INSERT INTO evaluations VALUES(20470,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(20470,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(20470,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(20470,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(20470,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(20470,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(20470,36,'cp -25','b1c3 g8f6 b5c6 b7c6 e1g1 f8e7 d2d3 e8g8 f3e5 d8c7 d1f3 c8a6 f1e1 f8e8 c1d2 f6d7 a1d1 e7f6 e5d7 c7d7 b2b3 e8e1 d1e1 f6h4 e1f1 h4e7 f1d1 a8e8 h2h3 e7f8 f4f5 h7h6 d1e1 e8e1 d2e1 c5c4 d3c4 d5c4 c3e4 d7d4 e1f2');
INSERT INTO evaluations VALUES(20468,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(20468,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(20468,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(20468,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(20468,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(20468,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(20468,36,'cp -19','g8f6 b5c6 b7c6 e1g1 f8e7 d2d3 d8c7 f3e5 e8g8 d1f3 f8e8 b2b3 e7d6 h2h3 f6d7 e5d7 c8d7 c1d2 f7f6 f4f5 d6e5 a1d1 a8d8 g1h1 e8e7 c3a4 c7d6 d2c3 e5d4 c3d4 c5d4 f3f4 e7e5 d1e1 d8e8 e1e5 f6e5 f4g4');
INSERT INTO evaluations VALUES(13711,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(13711,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(13711,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(13711,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(13711,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(13711,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(13711,36,'cp -17','b5c6 b7c6 e1g1 f8e7 d2d3 e8g8 f3e5 d8c7 d1f3 c8a6 f1e1 f8e8 c1d2 f6d7 a1d1 e7f8 e5d7 c7d7 f4f5 e8e1 d1e1 a8e8 e1e8 d7e8 f3f4 e8d8 g1h1 d8f6 g2g4 f8d6 f4f2 f6e5 f2e2 f7f6 h2h3 c5c4 e2e5 d6e5');
INSERT INTO evaluations VALUES(2075,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(2075,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(2075,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(2075,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(2075,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(2075,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(2075,36,'cp 55','e7e5 f2f4 e5e6 e2e4 b7b6 f1d3 c8b7 d1e2 g8f6 c1d2 f8b4 e4e5 f6d5 c3b5 b4d2 e2d2 e8c8 e1c1 d5e7 a2a3 c8b8 h1e1 e7c6 c1b1 a7a6 b5c3 d7d6 e5d6 e6d6 d2f2 d6h6 g2g4 h8e8 g4g5 h6h3 e1e8 d8e8 d3c4 f7f6 g5f6 g7f6 c4d5 e8d8');
INSERT INTO evaluations VALUES(1995,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(1995,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(1995,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(1995,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(1995,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(1995,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(1995,36,'cp 76','d1d3 c7c6 e2e4 d7d5 c1e3 d5e4 d3e4 e5e4 c3e4 c8f5 f1d3 f5e4 d3e4 e8c8 e4f3 c8c7 e1g1 f8d6 a1d1 g8f6 g2g3 d6e5 e3c1 f6d5 f1e1 f7f6 c2c3 b7b5 g1g2 d5b6 h2h4 d8d1 f3d1 h8d8 d1c2 h7h6 h4h5');
INSERT INTO evaluations VALUES(18803,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18803,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18803,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18803,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18803,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18803,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18803,36,'cp 0','c5d4 f3d4 g8f6 d1f3 e6e5 d4b3 f8b4 c2c3 b4e7 f1c4 d7d6 b1d2 e8g8 h2h3 c8e6 c4d5 e6d7 d5c4');
INSERT INTO evaluations VALUES(18923,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18923,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18923,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18923,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18923,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18923,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18923,36,'cp 0','f3d4 g8f6 d1f3 e6e5 d4b3 f8b4 c2c3 b4e7 f1c4 d7d6 b1d2 e8g8 h2h3 c8e6 c4d5 e6c8');
INSERT INTO evaluations VALUES(18928,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(18928,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(18928,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(18928,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(18928,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(18928,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(18928,36,'cp -4','g8f6 f1d3 a7a6 g1f3 d7d6 a2a4 f8e7 e1g1 e8g8 h2h3 b7b6 c1e3 c6b4 f3d4 e6e5 d4e2 d6d5 c3d5 f6d5 e4d5 b4d5 e2c3 d5e3 f2e3 f7f5 d3c4 g8h8 d1d8 e7d8 f1d1 g7g6 g1f2 a8b8 b2b3 f5f4 e3e4 d8e7 c3d5 e7c5 f2e2 b6b5 a4b5 a6b5 c4d3 b5b4');
INSERT INTO evaluations VALUES(19807,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19807,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(19807,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(19807,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19807,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19807,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(19807,36,'cp 9','e2e3 e7e6 d2d4 g8f6 f1d3 b7b6 a2a3 c5d4 e3d4 f8d6 d1e2 c8b7 b1d2 a8c8 f3e5 e8g8 e1g1 c8c7 a1e1 c6e7 g2g3 e7f5 g3g4 f5h4 f2f4 b6b5 f1f2 a7a5 e1f1 c7c8 f4f5 e6f5 g4f5 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(27591,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(27591,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(27591,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(27591,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(27591,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(27591,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(27591,36,'cp 56','b1c3 g4f3 g2f3 d5d4 d1d4 c6d4 c3b5 e8c8 b5d4 d8d4 c1e3 d4d6 e3a7 g7g6 a1c1 d6c6 c1c6 b7c6 a7d4 g8f6 f1c4 f8g7 c4f7 f6h5 d4g7 h5g7 f7c4 g7f5 e1e2 f5d4 e2e3 d4f5 e3f4 h8f8 f4g5 c8d7 h1d1 f5d6 g5g4 f8f5 f3f4 d7c7 c4d3 f5h5 h2h4 h5d5 d3c2 d5c5 c2b3 c5b5 d1e1 d6f5 e1e5');
INSERT INTO evaluations VALUES(27589,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(27589,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(27589,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(27589,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(27589,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(27589,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(27589,36,'cp 61','g4f3 g2f3 d5d4 d1d4 c6d4 c3b5 e8c8 b5d4 d8d4 c1e3 d4d6 e3a7 g7g6 a1c1 d6c6 c1c6 b7c6 a7d4 g8f6 f1c4 f8g7 c4f7 f6h5 d4g7 h5g7 f7c4 g7f5 e1d2 f5h4 h1c1 c8d7 d2e2 h8b8 c1c2 b8f8 c2c3 h4f5 c4d3 f5d4 e2e3 d4f3 d3e4 f3e5 f2f3 d7d6 a2a4');
INSERT INTO evaluations VALUES(35825,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(35825,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(35825,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(35825,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(35825,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(35825,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(35825,36,'cp 27','d4c5 d5c4 f1c4 d8d1 a1d1 e7c5 e1g1 a7a6 d1c1 h7h6 g5f6 g7f6 f1d1 c5e7 f3d4 c8d7 c4e2 f8d8 e2f3 a8a7 c3e2 g8f8 c1c7 b8c6 d4c6 d7c6 d1d8 e7d8 c7c8 c6f3 g2f3 f8e7 g1f1 e7d7 c8c1 f6f5 c1d1 d7e7 b2b3 d8c7 d1c1');
INSERT INTO evaluations VALUES(35819,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(35819,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(35819,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(35819,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(35819,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(35819,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(35819,36,'cp 25','d5c4 f1c4 d8d1 a1d1 e7c5 e1g1 c5b4 d1c1 c8d7 c4e2 b8c6 f3d2 f8d8 a2a3 b4e7 f1d1 d7e8 d2c4 f6d5 c3d5 e6d5 g5e7 c6e7 c4e5 a8c8 e5f3 c8c1 d1c1 d8c8 c1c8 e7c8 g2g3 g8f8 g1f1 c8d6 e2d3 h7h6 f1e2 f8e7 f3d4');
INSERT INTO evaluations VALUES(40278,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(40278,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40278,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(40278,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(40278,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(40278,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(40278,36,'cp 38','e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 e7e6 f1d1 b8d7 c1d2 d7b6 f3e2 d8e7 b2b3 e6e5 d4e5 e7e5 c4d5 f6d5 c3d5 b6d5 a1c1 f8d8 d2e1 e5e7 e1a5 d8d7 a2a4 b7b6 a5e1 d5b4 e1b4 e7b4 c1c6 b4b3');
INSERT INTO evaluations VALUES(35877,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(35877,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(35877,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(35877,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(35877,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(35877,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(35877,36,'cp 35','e1g1 c8g4 h2h3 g4f3 d1f3 e7e6 f1d1 b8d7 a2a4 a7a5 c1d2 f8e8 d2e1 d7b8 f3e2 b8a6 a1c1 a6b4 d3b1 d8e7 b2b3 e6e5 d4e5 e7e5 c4d5 f6d5 c3d5 b4d5 e2f3 e5e7 c1c4 h7h5 c4e4 e7c7 g2g3 e8e4 b1e4 c7e7');
INSERT INTO evaluations VALUES(35876,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(35876,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(35876,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35876,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(35876,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(35876,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(35876,36,'cp 43','c8g4 h2h3 g4f3 d1f3 e7e6 f1d1 b8d7 b2b3 a7a6 f3e2 a8c8 c1b2 f8e8 a1c1 b7b5 e2c2 d8a5 c2d2 a5d8 d3f1 d8e7 d2c2 e8d8 c3e2 c6c5 c4b5 a6b5 d4c5 c8c5 b2a3 c5c2 a3e7 d8c8 c1c2 c8c2 d1c1 c2a2 c1c8 g7f8');
INSERT INTO evaluations VALUES(27795,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(27795,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(27795,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(27795,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(27795,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(27795,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(27795,36,'cp 30','h2h3 g4f3 d1f3 e7e6 b2b3 a7a6 c1b2 b8d7 f1d1 d8a5 f3e2 c6c5 d4c5 d5c4 d3c4 a8c8 c3a4 d7c5 b2c3 a5c7 a4c5 c7c5 a1c1 c5e7 a2a4 f8d8 e2f3 d8d1 c1d1 b7b5 a4b5 a6b5 c3f6 e7f6 f3f6 g7f6 c4b5 c8d8 d1c1 h7h5 b5c4 g8g7 b3b4 d8d2 b4b5 d2b2');
INSERT INTO evaluations VALUES(43912,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(43912,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43912,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43912,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(43912,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43912,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(43912,36,'cp 32','b5b4 g2g3 g7g6 b2b3 a7a5 c1b2 f8g7 f1g2 a8a6 a2a3 e8g8 b1d2 e7e6 a3b4 a5b4 a1a6 b8a6 d5e6 d7e6 d1a1 a6b8 g1e2 d8e7 d2e4 b8d7 e1g1 f8e8 e2f4 f6e4 b2g7 c8b7 g7h6 e6e5 f4d5 b7d5 c4d5 e4c3 e3e4 c3b5 a1a6');
INSERT INTO evaluations VALUES(44267,30,'cp -171',NULL);
INSERT INTO evaluations VALUES(44267,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(44267,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(44267,33,'cp -180',NULL);
INSERT INTO evaluations VALUES(44267,34,'cp -168',NULL);
INSERT INTO evaluations VALUES(44267,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(44267,36,'cp -171','d8a5 b1c3 a5e5 f1e2 a7a6 a2a3 e5c7 g2g4 b7b5 g4g5 c8b7 h1f1 f6e4 c3e4 b7e4 c1e3 b5b4 f2f3 e4b7 a3b4 f8b4 c2c3 b4d6 f3f4 e8g8 h2h4 b8c6 d4c6 b7c6 h4h5 a6a5 d1d2 f8b8 h5h6 d6f8 e3d4 g7g6 c3c4 a5a4 d4e5 c7b6 e5b8 a8b8 d2c3 b6b2 c3b2 b8b2');
INSERT INTO evaluations VALUES(33555,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(33555,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(33555,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(33555,33,'cp -176',NULL);
INSERT INTO evaluations VALUES(33555,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(33555,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(33555,36,'cp -171','b1c3 a5e5 f1e2 a7a6 e1g1 d7d5 f1e1 e5d6 a2a4 f8e7 a4a5 d6d8 d4b3 c8d7 c3a4 d7a4 a1a4 e8g8 c1e3 b8d7 c2c4 a8c8 e2f1 d7e5 e3b6 d8d7 b3d2 e5c6 c4c5 e6e5 b2b4 d7e6 b4b5 a6b5 f1b5 g8h8 d1e2 f6d7 d2b3 e6g6 f2f3 d7b6 c5b6 e7b4 e1d1');
INSERT INTO evaluations VALUES(33553,30,'cp -183',NULL);
INSERT INTO evaluations VALUES(33553,31,'cp -157',NULL);
INSERT INTO evaluations VALUES(33553,32,'cp -172',NULL);
INSERT INTO evaluations VALUES(33553,33,'cp -178',NULL);
INSERT INTO evaluations VALUES(33553,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(33553,35,'cp -169',NULL);
INSERT INTO evaluations VALUES(33553,36,'cp -177','a5e5 f1e2 a7a6 a2a3 e5c7 g2g4 b7b5 g4g5 c8b7 h1f1 f6d5 c3d5 b7d5 e2f3 c7b7 f3d5 b7d5 c1e3 f8e7 h2h4 b8c6 d4c6 d5c6 f2f4 a8c8 c2c3 a6a5 d1d4 e8g8 h4h5 b5b4 h5h6 f7f6 a3b4 a5b4 a1a7 b4c3 h6g7 f8f7 b2c3 f6g5 f4g5 f7f1 e1f1 c6f3 f1g1 e7c5');
INSERT INTO evaluations VALUES(33551,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(33551,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(33551,32,'cp -174',NULL);
INSERT INTO evaluations VALUES(33551,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(33551,34,'cp -163',NULL);
INSERT INTO evaluations VALUES(33551,35,'cp -168',NULL);
INSERT INTO evaluations VALUES(33551,36,'cp -175','f1e2 a7a6 a2a3 e5c7 g2g4 b7b5 g4g5 c8b7 h1f1 f6d5 c3d5 b7d5 e2f3 c7e5 c1e3 b8c6 f3d5 e5d5 d4c6 d5c6 d1d3 b5b4 a3b4 f8b4 c2c3 b4e7 f2f4 a6a5 f4f5 a5a4 d3d4 f7f6 f5e6 c6e6 g5f6 e7f6 d4c5 e8f7 e1c1 h8c8 f1f6 e6f6 d1d7 f7g8 c5d5 g8h8 e3d4 f6f4 c1b1');
INSERT INTO evaluations VALUES(46184,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46184,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(46184,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(46184,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(46184,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(46184,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(46184,36,'cp -10','d5c4 f1c4 a7a6 g1f3 b7b5 c4e2 c8b7 e1g1 b8d7 d4c5 f8c5 b2b4 c5e7 a3a4 e7b4 c1b2 b5a4 d1a4 d8e7 c3a2 b4d6 b2f6 g7f6 f1b1 a8b8 e2a6 b7f3 b1b8 d6b8 g2f3 e8g8 a2c3 b8e5 a1c1 e5c3 c1c3 d7e5 f3f4 e5f3 g1g2 f3h4 g2f1 g8h8 a4c6 f8d8 c6c7 e7c7 c3c7');
INSERT INTO evaluations VALUES(54211,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(54211,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54211,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(54211,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(54211,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54211,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(54211,36,'cp -15','g8f6 g1f3 b7b6 b1d2 f8e7 e1g1 c8b7 h2h3 b8d7 b2b3 e8g8 d1e2 e7d6 c1b2 a8c8 f1d1 c5d4 c3d4 f6e4 d1c1 c8c1 a1c1 d8b8 d3b5 d7f6 d2e4 f6e4 f3d2 e4f6 a2a4 f8c8 c1c8 b8c8 d2f3 f6e4 f3e5 f7f6 e5d3');
INSERT INTO evaluations VALUES(46241,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(46241,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46241,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46241,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46241,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46241,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46241,36,'cp 0','b1d2 b7b6 g1f3 c8b7 d1e2 b8c6 e1g1 f8e7 b2b3 e8g8 c1b2 a7a5 d3b5 c6a7 b5d3');
INSERT INTO evaluations VALUES(54471,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(54471,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(54471,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(54471,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(54471,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(54471,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(54471,36,'cp -37','b8c6 b1c3 g8f6 g1e2 f8e7 b5c6 b7c6 e1g1 e8g8 c3a4 f8e8 b2b3 f6d7 c1a3 c8a6 e2g3 a6b5 a4c3 e7f6 d1d2 g7g6 f1e1 d8a5 a3b2 f6g7 e1e8 a8e8 a1e1 e8e1 d2e1 d7f8 g1f1 a5d8 e1c1 h7h5 g3e2 f8e6 c3b5 c6b5 b2g7 e6g7 a2a4 b5b4 e2f4 h5h4 h2h3');
INSERT INTO evaluations VALUES(20480,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(20480,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(20480,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(20480,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(20480,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(20480,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(20480,36,'cp -28','b1c3 g8f6 g1e2 f8e7 e1g1 e8g8 b5c6 b7c6 c3a4 f6d7 c1e3 d8a5 b2b3 d7b6 e3d2 a5a6 a4b6 a7b6 h2h3 c5c4 d3c4 d5c4 f1e1 a6b7 e2g3 f8d8 d1e2 c8e6 g3h5 e7f8 d2g5 d8d5 g5f4 c4c3 g2g4 a8d8 e2f3 b7d7 f3c3');
INSERT INTO evaluations VALUES(54885,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(54885,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(54885,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(54885,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(54885,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(54885,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(54885,36,'cp 112','e2e4 f8g7 f1c4 b8c6 d4d5 c6e5 c4b3 c7c6 f2f4 e5d7 d5c6 b7c6 g1f3 d8b6 e4e5 g8h6 c3a4 b6a6 d1c2 e8g8 c1e3 h6f5 e3f2 d7b6 a4c5 a6a5 c2d2 a5d2 f3d2 b6d5 g2g3 g6g5 d2c4 g5f4 g3f4 d5f4 e1c1 f4e6 h1e1 e6c5');
INSERT INTO evaluations VALUES(54888,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(54888,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(54888,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(54888,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(54888,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(54888,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(54888,36,'cp 136','f8g7 f1c4 b8c6 d4d5 c6e5 c4b3 c7c6 c1e3 c6d5 d1d5 c8e6 d5b7 a8b8 b7a7 e6b3 a2b3 g8f6 a1d1 d8c8 a7a4 f6d7 f2f4 e5g4 e3d4 e8g8 h2h3 b8a8 a4c4 e7e5 c3d5 c8c4 b3c4 e5d4 h3g4 f8b8 b2b4 d7b6 e4e5 b6d5 c4d5 b8b4 g1e2 d4d3 d1d3 b4b1 d3d1 b1d1 e1d1 a8d8 d1d2 d8d5 d2e3 g7f8 g4g5 f8c5 e3e4 d5d2');
INSERT INTO evaluations VALUES(38500,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(38500,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(38500,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(38500,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(38500,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(38500,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(38500,36,'cp 107','f1c4 b8c6 d4d5 c6e5 c4b3 c7c6 c1e3 c6d5 d1d5 c8e6 d5b7 a8b8 b7a7 e6b3 a2b3 g8f6 a7a4 d8d7 a4d7 f6d7 a1a3 e8g8 g1f3 e5f3 g2f3 f8c8 e1d2 e7e6 d2c2 d7e5 f3f4 e5c6 h1d1 c6b4 c2d2 b4c6 e3c5 c8d8 d2e3 d8d1 c3d1 b8d8 d1c3 c6d4 c5d4');
INSERT INTO evaluations VALUES(55427,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(55427,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(55427,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(55427,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(55427,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(55427,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(55427,36,'cp -62','g7g5 h2h4 f8g7 d1d3 b8c6 b1c3 g5g4 f3g1 f4f3 g2f3 g7d4 c1f4 d4e5 f4e3 e5g3 e1d2 c6e5 d3f1 e5c4 f1c4 g3e5 a1e1 g8e7 d2c1 c8e6 c4d3 e7c6 f3f4 e5g7 g1e2 d8d7 d3b5 a7a6 b5h5 e8c8 f4f5 e6c4 h5g4 g7f6 c1b1 c4e2 c3e2 h8g8 g4f4 d8e8 h1g1 h6h5 e3c1 c8b8 g1g8');
INSERT INTO evaluations VALUES(55395,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(55395,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(55395,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(55395,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(55395,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(55395,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(55395,36,'cp -61','h2h4 f8g7 d1d3 b8c6 b1c3 g5g4 f3g1 f4f3 g2f3 g7d4 c1f4 d4e5 f4e3 e5g3 e1d2 c6e5 d3f1 e5c4 f1c4 g3e5 f3g4 g8f6 g4g5 c8e6 c4d3 e5c3 b2c3 f6g4 e3d4 g4e5 d4e5 d6e5 g1f3 d8d6 a2a4 e6g4 f3h2 e8c8 d3d6 d8d6 d2e3');
INSERT INTO evaluations VALUES(1582,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(1582,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(1582,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(1582,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(1582,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(1582,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(1582,36,'cp 168','f1e2 e6d7 d2d4 e7e6 e1g1 g8f6 f1e1 f8e7 e2b5 e8g8 f3e5 d7e8 b5c6 b7c6 c1g5 c8b7 d1d3 h7h6 g5f6 e7f6 c3a4 f6e7 a4c5 e7c5 d4c5 a8d8 d3e4 f7f6 e5g6 f8f7 e4e6 e8e6 e1e6 f7d7 a1e1 d7d2 e6e7 d2c2 e7c7 g8h7 g6e7 c2b2 e7f5');
INSERT INTO evaluations VALUES(10039,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(10039,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(10039,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(10039,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(10039,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(10039,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(10039,36,'cp 0','c5b4 e1f1 d7d5 e4e5 f6e4 a2a3 b4e7 d1c2 e8g8 c1e3 f7f6 b1c3 f6e5 b5c6 e4c3 c2c3 b7c6 f3e5 e7d6 h2h4 d8f6 c3c6 g8h8 f1g1 d6e5 c6a8 e5f4 a8d5 f4e3 f2e3 f6f2 g1h2 f2h4 h2g1 h4f2 g1h2 f2h4');
INSERT INTO evaluations VALUES(10013,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10013,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10013,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10013,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10013,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10013,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10013,36,'cp 0','e1f1 d7d5 e4e5 f6e4 a2a3 b4a5 d1a4 a5b6 b5c6 b7c6 a4c6 c8d7 c6d5 e4c5 f1g1 d7e6 d5d8 a8d8 c1g5 c5b3 g5d8 e8d8 a1a2 b3c1 a2a1');
INSERT INTO evaluations VALUES(12365,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(12365,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(12365,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(12365,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(12365,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(12365,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(12365,36,'cp -9','c4b5 d8c7 g1f3 f8e7 e1g1 e8g8 c1g5 d7d6 b1d2 a7a6 b5c6 c7c6 a2a4 b7b6 f1e1 h7h6 g5h4 c8b7 d2c4 a8e8 h2h3 c6c7 h4g3 f6d7 d1e2 b6b5 a4b5 a6b5 c4a5 e8a8 a5b7 c7b7 e4e5 a8a1 e1a1 d7e5 f3e5 d6e5 e2e5');
INSERT INTO evaluations VALUES(14509,30,'cp -537',NULL);
INSERT INTO evaluations VALUES(14509,31,'cp -524',NULL);
INSERT INTO evaluations VALUES(14509,32,'cp -525',NULL);
INSERT INTO evaluations VALUES(14509,33,'cp -532',NULL);
INSERT INTO evaluations VALUES(14509,34,'cp -525',NULL);
INSERT INTO evaluations VALUES(14509,35,'cp -537',NULL);
INSERT INTO evaluations VALUES(14509,36,'cp -536','e8f7 e4d5 b7b5 d5c6 b5c4 e1g1 e5e4 b1c3 c8g4 d1e1 d8e8 b2b3 e8c6 h2h3 g4h5 b3c4 c6c4 d2d3 e4d3 c2d3 c4d3 e1d2 d3d2 c1d2 a7a6 d2g5 f8a3 g5f6 g7f6 c3d5 a8d8 d5e3 h5g6 g2g4 h8e8 f1e1 d8d7 a1d1 d7d1 e1d1');
INSERT INTO evaluations VALUES(2425,30,'cp -512',NULL);
INSERT INTO evaluations VALUES(2425,31,'cp -521',NULL);
INSERT INTO evaluations VALUES(2425,32,'cp -517',NULL);
INSERT INTO evaluations VALUES(2425,33,'cp -530',NULL);
INSERT INTO evaluations VALUES(2425,34,'cp -536',NULL);
INSERT INTO evaluations VALUES(2425,35,'cp -525',NULL);
INSERT INTO evaluations VALUES(2425,36,'cp -528','e4d5 b7b5 d5c6 b5c4 e1g1 e5e4 b1c3 c8g4 d1e1 d8e8 h2h3 g4h5 e1e3 e8c6 d2d3 e4d3 c2d3 f8d6 e3d4 c4d3 d4d3 a8b8 c1g5 h8d8 d3c2 h5g6 c2c1 h7h6 g5f6 g7f6 a2a4 d8g8 f1d1 g6e4 c3e4 c6e4 g2g4 f6f5 d1d6 c7d6 c1c7 e4e7 c7c2 f5g4 c2h7 g8g7 h7f5 f7g8');
INSERT INTO evaluations VALUES(2424,30,'cp -550',NULL);
INSERT INTO evaluations VALUES(2424,31,'cp -544',NULL);
INSERT INTO evaluations VALUES(2424,32,'cp -544',NULL);
INSERT INTO evaluations VALUES(2424,33,'cp -550',NULL);
INSERT INTO evaluations VALUES(2424,34,'cp -509',NULL);
INSERT INTO evaluations VALUES(2424,35,'cp -526',NULL);
INSERT INTO evaluations VALUES(2424,36,'cp -542','b7b5 d5c6 b5c4 e1g1 e5e4 b1c3 c8g4 d1e1 d8e8 h2h3 g4h5 b2b3 e8c6 c3d1 f8c5 a1b1 a8e8 b3c4 h5f3 b1b5 f6h5 b5c5 c6c5 e1e3 c5f5 d2d3 f3d1 f1d1 e4d3 e3d3 f5d3 c2d3 e8e2 c1e3 h8b8 a2a3 h5f6 e3c5 a7a5 c5d4 b8e8 a3a4 f6d7');
INSERT INTO evaluations VALUES(16532,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(16532,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(16532,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(16532,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(16532,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(16532,35,'cp -121',NULL);
INSERT INTO evaluations VALUES(16532,36,'cp -128','f1b5 d7d5 d2d4 e5d4 g1e2 f8d6 e4d5 f6d5 c3d4 d8h4 g2g3 h4h3 e1f2 c8e6 b1c3 e8g8 c3d5 e6d5 e2c3 d5e6 c3e4 d6e7 b5c6 b7c6 c1f4 a7a5 a1c1 a8b8 d1e2 e6d5 e4c3 f8e8 c3d5 c6d5 c1c7 e7f6 f4e5 f6e5 d4e5 d5d4');
INSERT INTO evaluations VALUES(16740,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(16740,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(16740,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(16740,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(16740,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(16740,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(16740,36,'cp 140','d7d6 d2d4 e5d4 f3d4 h6g4 b1c3 f8e7 c4b5 c8d7 b5c6 b7c6 d4c6 d7c6 d1g4 e7f6 c3d5 e8g8 a1b1 f6e5 f1d1 c6d7 g4h5 c7c6 d5e3 d8a5 a2a3 a5a4 h5f3 f7f5 e4f5 d7f5 b2b3 a4e4 f3e4 f5e4 c1b2 e5f4 b2d4 f4e3 d4e3');
INSERT INTO evaluations VALUES(15672,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(15672,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(15672,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(15672,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(15672,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(15672,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(15672,36,'cp 133','d2d4 e5d4 f3d4 h6g4 b1c3 f8e7 c4b5 c8d7 b5c6 b7c6 d4c6 d7c6 d1g4 e8g8 c3d5 f7f5 d5e7 d8e7 e4f5 e7e4 g4e4 c6e4 f2f3 e4f5 f1f2 f8e8 c1f4 a7a5 a2a4 g8f7 a1d1 h7h5 h2h3 f7e6 c2c4 a8b8 c4c5 b8b4 f4h2 e6f7 c5d6 c7d6 h2d6 b4a4 f2d2 f5e6');
INSERT INTO evaluations VALUES(22681,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(22681,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(22681,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(22681,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(22681,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(22681,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(22681,36,'cp 10','c6e5 c2c4 b7b6 g1f3 d8f6 f1e2 c8b7 c1d2 e5f3 e2f3 b7f3 d1f3 f6f3 g2f3 g8f6 h2h4 g7g6 d2c3 f8g7 c3d4 e8c8 b1c3 f6h5 a1d1 a7a5 e1e2 g7d4 d1d4 h5g7 a2a3 h8e8 f3f4 c7c6 d4d1 f7f5 b2b3 c8c7 e2f3');
INSERT INTO evaluations VALUES(23089,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(23089,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(23089,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(23089,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(23089,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(23089,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(23089,36,'cp 14','c2c4 b7b6 g1f3 d8f6 f1e2 c8b7 c1d2 e5f3 e2f3 b7f3 d1f3 f6f3 g2f3 g8f6 d2c3 f8e7 b1d2 e8c8 d2e4 f6e4 f3e4 h8g8 e1e2 d8e8 h2h4 a7a5 f2f4 h7h5 e2f3 g7g6 a1d1 e7c5 c3d4 c5e7 b2b3 f7f5 e4e5 d7d6 d4b2 g8f8 f3e2 d6e5 b2e5 e7c5');
INSERT INTO evaluations VALUES(28557,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28557,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(28557,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(28557,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28557,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(28557,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(28557,36,'cp 26','g1f3 e7e6 f3e5 f8d6 f1d3 f5d3 c2d3 e8g8 e1g1 d8e7 f4g3 b8d7 e3e4 f8d8 e5d7 f6d7 e4e5 d6b4 c3e2 b4a5 f2f4 a5b6 g1h1 f7f5 g3f2 e7f7 a1c1 d7f8 h2h3 a6a5 c1c2 f8g6 a2a3 d8f8 g2g3 g6e7 f1g1 f7h5');
INSERT INTO evaluations VALUES(28558,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(28558,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(28558,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(28558,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(28558,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(28558,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(28558,36,'cp 0','e7e6 f3e5 f6d7 g2g4 f5g6 e5g6 h7g6 e3e4 d7f6 f1g2 f8b4 e4d5 f6d5 f4d2 c7c6 c3e4 b4d2 d1d2 d8c7 c2c4 d5f4 g2f3 b8d7 e1c1 d7f6 e4f6 g7f6 h2h4 e8f8 h4h5 f8g7 h5g6 h8h1 d1h1 f7g6 c1b1 a8h8 h1h8 g7h8 f3e4 h8g7 e4c2 c7d6 d2e3 g6g5 e3e4 b7b5 e4h7 g7f8 h7h6 f8f7 h6h7');
INSERT INTO evaluations VALUES(28570,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(28570,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28570,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28570,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(28570,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(28570,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(28570,36,'cp 13','f3e5 f6d7 g2g4 f5g6 e5g6 h7g6 e3e4 d7f6 f1g2 f8b4 e4d5 f6d5 f4d2 c7c6 c3e2 b4d2 d1d2 b8d7 c2c4 d5e7 h2h3 d8c7 e1g1 d7f6 f1d1 e8g8 b2b3 a8d8 d2e3 d8d7 e2c3 f8d8 d1d2 a6a5 c3a4 e6e5 a1d1 e5d4 d2d4 d7d4 d1d4 d8d4 e3d4 f6d7');
INSERT INTO evaluations VALUES(41408,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41408,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(41408,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(41408,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41408,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(41408,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(41408,36,'cp 16','d3d4 e8g8 c1e3 b8d7 a2a4 b7b6 f1d3 e7e5 d4e5 d6e5 f3d2 c7c6 d1e2 d8c7 d2b3 f8e8 f2f3 d7f8 e1g1 f6h5 e2f2 h5f4 a4a5 f4d3 c2d3 b6b5 b3c5 f8e6 c3e2 e8d8 f1d1 e6c5 e3c5');
INSERT INTO evaluations VALUES(54400,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(54400,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(54400,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(54400,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(54400,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54400,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54400,36,'cp 29','d2d4 b8c6 c1f4 f8d6 f4d6 d8d6 f1d3 c8d7 e1g1 g8f6 c3e2 e8g8 c2c3 f8e8 d1d2 e6e5 d4e5 c6e5 f3e5 e8e5 a1e1 a8e8 e2d4 a7a5 e1e5 e8e5 f1e1 e5e1 d2e1 d6b6 e1d2 h7h6 h2h3 f6e4 d3e4 d5e4 b2b4 d7b5 d4b5');
INSERT INTO evaluations VALUES(1815,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(1815,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(1815,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(1815,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(1815,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(1815,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(1815,36,'cp 195','e4e5 f6f5 b1c3 g8e7 f2f4 e7c6 d4f2 b7b6 c4d3 f5e6 c3b5 e8d8 c1e3 f7f6 e1g1 c8b7 a1e1 f8e7 f2g3 f6e5 f4e5 d8c8 g1h1 a7a6 b5c3 e7b4 e3g1 c6e7 g3f4 c7c5 f4f7 b4c3 b2c3 c8c7 f7e6 d7e6 f1f7 a8e8 e1d1 c7b8 f7g7');
INSERT INTO evaluations VALUES(1813,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(1813,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(1813,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(1813,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(1813,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(1813,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(1813,36,'cp 201','f6f5 b1c3 g8e7 f2f4 e7c6 d4f2 b7b6 c3b5 f5e4 f2e2 e4e2 e1e2 e8d8 c4f7 a7a6 b5c3 d7d6 f7d5 c8b7 e5d6 f8d6 c1e3 d8d7 a1d1 a8e8 c3e4 d6e7 h2h4 d7c8 h4h5 h8f8 g2g3 e7f6 e4f6 f8f6 c2c3 c8b8 e2f2 c6e5 d5b7 e5g4 f2f3 e8e3 f3g4 b8b7 h1e1');
INSERT INTO evaluations VALUES(12672,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(12672,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(12672,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(12672,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(12672,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(12672,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(12672,36,'cp 17','d2d4 c5d4 f3d4 c6d4 d1d4 a7a6 b5e2 e7c6 d4e3 d8c7 e3g3 c7g3 h2g3 f8c5 a2a3 b7b5 g3g4 c8b7 c1d2 d7d6 f2f3 h7h6 e1c1 e8e7 c1b1 c6e5 a3a4 b5b4 c3a2 a6a5 a2c1 b7c6 b2b3 f7f6 c1d3 e5d3 e2d3 e6e5 d3b5 c6b5');
INSERT INTO evaluations VALUES(12669,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(12669,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(12669,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(12669,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12669,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(12669,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(12669,36,'cp 21','c5d4 f3d4 a7a6 b5e2 c6d4 d1d4 e7c6 d4e3 d8c7 e3g3 c7g3 h2g3 f8c5 c1f4 b7b5 g3g4 e8e7 g4g5 c8b7 a2a3 d7d6 e1c1 c6d4 e2f1 e6e5 f4d2 a6a5 c3d5 b7d5 e4d5 h7h6 f1d3 h6g5 d2g5 f7f6 g5e3 a5a4 c1b1 f6f5 h1h8 a8h8 c2c3 d4b3 e3c5 b3c5 d3f5');
INSERT INTO evaluations VALUES(20758,30,'cp 321',NULL);
INSERT INTO evaluations VALUES(20758,31,'cp 340',NULL);
INSERT INTO evaluations VALUES(20758,32,'cp 330',NULL);
INSERT INTO evaluations VALUES(20758,33,'cp 344',NULL);
INSERT INTO evaluations VALUES(20758,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(20758,35,'cp 318',NULL);
INSERT INTO evaluations VALUES(20758,36,'cp 328','d5c6 b7c6 b1c3 d8b6 c1e3 b6d4 e3d4 h7h5 e1c1 h5h4 f1e2 h4h3 g2g3 f7f6 b2b4 c8d7 b4b5 f8e7 b5c6 d7c6 h1e1 g8h6 f2f3 h6f7 f3f4 c6d7 e2c4 e8g8 e1e3 f8e8 c4d5 a8c8 g3g4 g7g5 f4f5 g8g7 e3h3 e8h8 h3h8 c8h8');
INSERT INTO evaluations VALUES(20156,30,'cp 295',NULL);
INSERT INTO evaluations VALUES(20156,31,'cp 313',NULL);
INSERT INTO evaluations VALUES(20156,32,'cp 321',NULL);
INSERT INTO evaluations VALUES(20156,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(20156,34,'cp 330',NULL);
INSERT INTO evaluations VALUES(20156,35,'cp 348',NULL);
INSERT INTO evaluations VALUES(20156,36,'cp 352','b7c6 b1c3 d8b6 c1e3 g8f6 e1c1 b6d4 e3d4 c8g4 f2f3 g4e6 f3f4 a7a5 f1e2 f8e7 f4f5 e6c8 g2g4 e8g8 g4g5 f6d7 h1g1 g7g6 f5f6 e7d8 h2h4 d7e5 d4e5 d6e5 d1d6 d8b6 g1h1 h7h5 g5h6 c8e6 d6c6 b6e3 c1b1 a8c8 c6a6 c8a8 a6a8 f8a8 h1d1 e3h6');
INSERT INTO evaluations VALUES(18508,30,'cp 337',NULL);
INSERT INTO evaluations VALUES(18508,31,'cp 325',NULL);
INSERT INTO evaluations VALUES(18508,32,'cp 329',NULL);
INSERT INTO evaluations VALUES(18508,33,'cp 325',NULL);
INSERT INTO evaluations VALUES(18508,34,'cp 338',NULL);
INSERT INTO evaluations VALUES(18508,35,'cp 322',NULL);
INSERT INTO evaluations VALUES(18508,36,'cp 344','b1c3 d8b6 c1e3 b6d4 e3d4 h7h5 e1c1 h5h4 f1e2 h4h3 g2g3 c8d7 h1e1 g8h6 f2f3 f7f6 b2b4 f8e7 b4b5 e7d8 b5c6 d7c6 e2c4 a8c8 c4e6 c6d7 e6d7 e8d7 e1e3 d7e6 c3d5 c8c4 d4a7 c4a4 a7d4 a4a2');
INSERT INTO evaluations VALUES(30548,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(30548,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(30548,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(30548,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(30548,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(30548,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(30548,36,'cp 170','c3d5 e7e6 d5c3 g8f6 f2f3 b8d7 c1e3 c7c5 d4c5 f8c5 e3c5 d7c5 g1h3 d8d1 a1d1 e8e7 h3f2 a8c8 f1e2 a7a6 f2d3 b6b5 a2a3 f6d7 d3c5 c8c5 e1f2 h8c8 d1d4 d7b6 h1d1 b7c6 g2g3 f7f6 f3f4 g7g6 f2e3 g6g5 h2h4 g5f4 g3f4 c6b7 d4d6 b7c6');
INSERT INTO evaluations VALUES(31148,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(31148,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(31148,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(31148,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(31148,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(31148,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(31148,36,'cp -19','g4f3 d1f3 e7e6 d2d4 c6c5 f1b5 b8c6 c2c3 d8b6 f3d3 a7a6 b5c6 b6c6 e1g1 g8e7 d4c5 c6c5 b1d2 a8c8 d2f3 c5c7 f1e1 c7c4 d3d1 e7f5 f3d4 f5d4 c3d4 c4c2 d1g4 h7h5 g4f3 f8e7 c1e3 c2f5 f3f5 e6f5 e1c1 e8d7 a1b1 h5h4 b2b3 c8c1 b1c1');
INSERT INTO evaluations VALUES(31152,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(31152,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(31152,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(31152,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(31152,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(31152,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(31152,36,'cp -10','d1f3 e7e6 d2d4 c6c5 f1b5 b8c6 c2c3 d8b6 f3d3 a7a6 b5c6 b6c6 c1e3 c5d4 e3d4 g8e7 b1d2 e7f5 d2b3 a6a5 a2a4 f8e7 e1g1 e8g8 g1h2 c6c4 d3c4 d5c4 b3d2 f5d4 c3d4 f8d8 d2c4 d8d4 b2b3 e7b4 f1d1 d4d1 a1d1');
INSERT INTO evaluations VALUES(31151,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(31151,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(31151,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31151,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(31151,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(31151,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(31151,36,'cp -27','e7e6 d2d4 c6c5 c2c3 b8c6 f1b5 c5d4 c3d4 g8e7 e1g1 e7f5 f3d3 f8e7 b1d2 e8g8 d2f3 c6b4 d3d1 a8c8 b5a4 b4c6 a2a3 a7a6 g2g4 f5h4 f3h4 e7h4 c1e3 c6a5 a4c2 f7f6 e5f6 h4f6 b2b3 g7g6 a1c1 f6g7 g1g2 a5c6 c2b1 d8b6 d1d3');
INSERT INTO evaluations VALUES(31504,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(31504,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(31504,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31504,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(31504,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(31504,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(31504,36,'cp -10','d2d4 c6c5 f1b5 b8c6 c2c3 c5d4 c3d4 g8e7 e1g1 e7f5 f3d3 f8e7 b1d2 e8g8 d2f3 c6b4 d3d1 a8c8 b5a4 b4c6 a2a3 a7a6 d1d3 d8b6 f1d1 c8c7 b2b4 f8c8 c1f4 c6a7 a1c1 a7b5 a4b5 b6b5 d3b5 a6b5 g2g4 f5h4 f3h4 c7c1 d1c1 c8c1 f4c1 e7h4 c1e3 g8f8 f2f4 g7g6 e3f2');
INSERT INTO evaluations VALUES(36369,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(36369,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(36369,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(36369,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(36369,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36369,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(36369,36,'cp 27','c7c5 d4c5 d8d1 f1d1 e7c5 b2b3 b7b6 c1b2 c8b7 c3b5 b8c6 f3e5 c6e5 b2e5 f6d5 a2a3 a7a6 c4d5 e6d5 b5c3 f8e8 e5d4 a8c8 c3e2 g7g6 a1b1 c5a3 d4b6 a3c5 b6c5 c8c5 e2d4 c5c3 b3b4 e8c8 h2h3 g8g7 b1b3');
INSERT INTO evaluations VALUES(37224,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(37224,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(37224,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(37224,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(37224,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(37224,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(37224,36,'cp 0','d7d6 d2d4 e5d4 f3d4 f8e8 c1g5 b8d7 d4b3 c5b6 a2a4 a7a5 g1h1 h7h6 g5h4 d7e5 h4f6 d8f6 c3d5 f6h4 d5b6 e5g4 h2h3 c7b6 h1g1 g4f6 f2f3 c8h3 d1d6 h3g2 g1g2 f6e4 c4f7 g8h8 d6h2 h4g5 g2h3 g5f5 h3g2');
INSERT INTO evaluations VALUES(53865,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(53865,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(53865,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(53865,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(53865,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(53865,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(53865,36,'cp 51','b8d7 d4f5 a7a6 b5a4 b7b5 a4b3 d7c5 b1c3 g8f6 c1g5 c8f5 e4f5 f8e7 e1g1 e8g8 g5f6 e7f6 c3d5 e5e4 c2c3 f8e8 d1g4 c5d3 g4e2 a8c8 g2g3 c8c5 a1d1 a6a5 a2a3 a5a4 b3a2 d8a8 d5f6 g7f6 e2g4 g8h8 a2f7 d3e5 g4h5 e8e7 f7a2 d6d5 h5h6 e7f7 g1h1');
INSERT INTO evaluations VALUES(10428,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10428,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10428,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(10428,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10428,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(10428,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10428,36,'cp 0','d7d6 a2a4 a7a5 d1e2 h7h6 b2b3 c8e6 f3d2 e8g8 c1a3 c6e7 f2f3 e7g6 g2g3 g8h8 c4d3 f6d7 a1e1 e6h3 f1f2 b7b6 d2f1 d8g5 e2d2 g6f4 f1e3 f4d3 c2d3 f7f5 e4f5 h3f5 e3d5 g5d8 d3d4 f5h3 d4e5 d7e5 e1e5 d6e5 a3f8 d8f8');
INSERT INTO evaluations VALUES(14785,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(14785,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(14785,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(14785,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(14785,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(14785,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(14785,36,'cp 32','d5c4 d1a4 f8b4 c1d2 f6d5 d2b4 d5b4 e1g1 a8b8 b1a3 e8g8 a4b5 b7b6 b5c4 c8a6 a3b5 d8d5 c4d5 b4d5 a2a4 a6b7 f3e5 c6e5 d4e5 a7a6 b5a3 f8c8 a1c1 d5e7 g2b7 b8b7 f1d1 g7g5 d1d7 e7g6 a3c4 b7b8 c4e3 g6e5 d7c7 c8c7 c1c7 a6a5 f2f4 g5f4 g3f4 e5g6');
INSERT INTO evaluations VALUES(14917,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(14917,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(14917,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14917,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(14917,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(14917,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14917,36,'cp 32','d1a4 f8b4 c1d2 f6d5 d2b4 d5b4 e1g1 a8b8 b1a3 e8g8 a4b5 b7b6 b5c4 c8a6 a3b5 d8d5 c4d5 b4d5 a2a4 a6b7 f1c1 a7a6 b5c3 d5b4 c3a2 b4a2 a1a2 f8d8 e2e3 c6b4 a2a1 c7c5 d4c5 b4d3 c5c6 d3c1 a1c1 b8c8 f3d4 e6e5 d4b3 b7c6 g2c6 d8d3 c6f3 c8c1 b3c1 d3d2');
INSERT INTO evaluations VALUES(16217,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(16217,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(16217,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(16217,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(16217,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(16217,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(16217,36,'cp 182','d7d5 e4d5 f8e7 c4b3 e8g8 c2c4 c8f5 b1c3 f6d7 c1f4 e7b4 a1e1 d8h4 b3d1 a8c8 a2a3 b4d6 f4e5 h4d4 e5d4 f8e8 d1a4 e8e1 f1e1 d7c5 b2b4 c5a4 c3a4 a7a5 c4c5 d6f8 f2f3 a5b4 a3b4 c8d8 a4c3 b7b6 c5b6 c7b6 d4b6');
INSERT INTO evaluations VALUES(17375,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(17375,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(17375,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17375,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17375,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17375,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17375,36,'cp 0','c5d4 c3d4 d7d5 e4d5 e6d5 c4b3 f8b4 b1c3 c8g4 e1g1 g8e7 h2h3 g4f3 d1f3 c6d4 f3d1 b4c3 b2c3 d4b3 d1b3 d8c7 a1b1 b7b5 a2a4 e8g8 a4b5 a6b5 c1a3 f8d8 a3e7 c7e7 f1d1 e7c7 b3b5 c7c3 d1d5 g7g6 d5d8 a8d8 h3h4 d8d2 b5b8 g8g7 b8f4 c3d4 f4d4 d2d4 h4h5 d4c4 h5g6 h7g6');
INSERT INTO evaluations VALUES(17424,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17424,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17424,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17424,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17424,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(17424,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17424,36,'cp 0','c3d4 d7d5 e4d5 e6d5 c4b3 f8b4 b1c3 c8g4 e1g1 g8e7 h2h3 g4f3 d1f3 c6d4 f3d1 d4b3 d1b3 b4c3 b2c3 d8c7 a1b1 b7b5 a2a4 e8g8 a4b5 a6b5 c1a3 f8d8 a3e7 c7e7 b3b5 e7c7 f1d1 c7c3 d1d5 g7g6 g2g3 d8d5 b5d5 a8c8 d5d7 h7h6 b1d1 c3c6 d7e7 g8g7 d1d8 c8d8 e7d8 c6e4 d8d2 g7f6 d2h6 e4e1 g1g2');
INSERT INTO evaluations VALUES(17421,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17421,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17421,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17421,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17421,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17421,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17421,36,'cp 0','d7d5 e4d5 e6d5 c4b3 f8b4 b1c3 c8g4 e1g1 g8e7 h2h3 g4f3 d1f3 c6d4 f3d1 d4b3 d1b3 b4c3 b2c3 d8c7 a1b1 b7b5 a2a4 e8g8 a4b5 a6b5 c1a3 f8d8 a3e7 c7e7 b3b5 e7c7 f1d1 c7c3 d1d5 a8c8 g2g3 d8d5 b5d5 g7g6 g1g2 c3f6 b1b7 c8c1 b7d7 c1c2 d5d4 f6d4 d7d4 c2b2 d4d7 g8g7');
INSERT INTO evaluations VALUES(23251,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(23251,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(23251,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(23251,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(23251,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(23251,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(23251,36,'cp 212','g8e7 b1c3 c6d4 d1d4 e7c6 d4e3 f8b4 e1g1 d7d6 c3d5 b4c5 e3c3 a7a5 c1h6 g7h6 d5f6 e8f8 c3f3 c6e5 f6d7 f8g7 d7e5 d8f6 f3g3 f6g5 e5d3 c5a7 a1e1 h8e8 a2a4 c8d7 b2b3 a8c8 e1e2 g7h8 e4e5 g5g3 h2g3 d7f5 f1e1 f5d3 c4d3 d6e5');
INSERT INTO evaluations VALUES(26126,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26126,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26126,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(26126,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(26126,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(26126,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(26126,36,'cp 35','a7a5 c1e3 f5g6 b1c3 g8e7 f3h4 e7f5 h4f5 g6f5 g2g4 f5g6 f2f4 f7f5 e2d3 f8e7 g1h1 a5a4 d1e2 d7f8 g4f5 g6f5 f1g1 f5d3 e2d3 g7g6 b2b3 a4b3 c2b3 e7a3 c3e2 d8d7 g1g2 a3b2 a1g1 a8a2');
INSERT INTO evaluations VALUES(33374,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33374,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33374,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33374,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33374,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(33374,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33374,36,'cp 0','h7h6 g5d2 e8g8 c2c4 c7c5 b1c3 b8c6 d2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(33370,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33370,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33370,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33370,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33370,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33370,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33370,36,'cp 0','g5d2 e8g8 c2c4 c7c5 b1c3 b8c6 d2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(33944,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(33944,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(33944,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(33944,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(33944,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(33944,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(33944,36,'cp -70','g1f3 b8c6 b1c3 f8b4 c1d2 b4c3 d2c3 f7f6 a2a3 c8g4 h2h3 g4h5 d1e2 c6d8 e2e4 d5e4 d3e4 h5f3 g2f3 d8e6 e1c1 a8d8 d1d8 e8d8 h3h4 d8e7 c3d2 h7h5 d2e3 g7g5 c1d2 e7d7 b2b4 b7b6 b4b5 d7e7 a3a4 h8d8 d2c3 e7f7 a4a5 b6a5 e3a7 g5h4 h1h4');
INSERT INTO evaluations VALUES(33940,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(33940,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(33940,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(33940,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(33940,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(33940,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(33940,36,'cp -81','b8c6 b1c3 f8b4 c1d2 b4c3 d2c3 f7f6 a2a3 e8g8 d1e2 c8e6 e1g1 e6g4 f1e1 a8d8 e2e4 g4f3 e4f3 d5f3 g2f3 g8f7 e1e4 g7g5 h2h4 h7h6 g1f1 f8e8 a3a4 c6e7 h4g5 h6g5 d3d4 e5d4 e4d4 e7f5 d4d8 e8d8 a4a5 f5h4 a5a6 b7b6 f1e2 d8e8 e2f1 c7c6');
INSERT INTO evaluations VALUES(38916,30,'cp 301',NULL);
INSERT INTO evaluations VALUES(38916,31,'cp 317',NULL);
INSERT INTO evaluations VALUES(38916,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(38916,33,'cp 289',NULL);
INSERT INTO evaluations VALUES(38916,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(38916,35,'cp 325',NULL);
INSERT INTO evaluations VALUES(38916,36,'cp 300','b2b4 a5b6 d1g4 g7g6 h2h4 a7a5 b4b5 h7h5 g4g3 c7c5 f1d3 b8d7 c1g5 d8c7 g1e2 c5c4 d3c2 a5a4 b1d2 d7f8 a1b1 f8h7 g5f4 g8e7 g3e3 c8d7 f4h6 h8g8 d2f3 a8a5 f3g5 h7g5 h6g5 a5b5 e1g1 d7c6 b1b5 c6b5 f1b1 b5c6 e3f4 e7f5 c2f5 g6f5 g5f6 g8g6 g2g3 g6f6');
INSERT INTO evaluations VALUES(38914,30,'cp 316',NULL);
INSERT INTO evaluations VALUES(38914,31,'cp 315',NULL);
INSERT INTO evaluations VALUES(38914,32,'cp 314',NULL);
INSERT INTO evaluations VALUES(38914,33,'cp 321',NULL);
INSERT INTO evaluations VALUES(38914,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(38914,35,'cp 312',NULL);
INSERT INTO evaluations VALUES(38914,36,'cp 302','a5b6 d1g4 g7g6 h2h4 h7h5 g4g3 a7a5 b4b5 c7c5 f1d3 b8d7 g1f3 d7f8 c1g5 d8c7 a2a4 g8e7 e1g1 c5c4 d3c2 h8g8 g5f4 c7d8 b1d2 e7f5 c2f5 e6f5 f4h6 c8e6 f3g5 d8e7 h6f8 g8f8 g5h7 e8c8 h7f8 d8f8 d2f3 c8b8 f1d1 b8a7 f3g5 f8h8 g3f3 h8c8 g2g3 e7f8');
INSERT INTO evaluations VALUES(38599,30,'cp 308',NULL);
INSERT INTO evaluations VALUES(38599,31,'cp 316',NULL);
INSERT INTO evaluations VALUES(38599,32,'cp 317',NULL);
INSERT INTO evaluations VALUES(38599,33,'cp 298',NULL);
INSERT INTO evaluations VALUES(38599,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(38599,35,'cp 321',NULL);
INSERT INTO evaluations VALUES(38599,36,'cp 301','d1g4 g7g6 h2h4 h7h5 g4g3 a7a5 b4b5 c7c5 g1f3 b8d7 f1d3 d7f8 c1g5 d8c7 d4c5 b6c5 b1d2 b7b6 d2b3 c5e7 e1g1 c8b7 b3d4 g8h6 g5e7 c7e7 d4c6 h6f5 d3f5 b7c6 b5c6 e6f5 f1b1 f8e6 f3d4 f5f4 g3h3 e8g8 b1b6 e7c7 b6b7 c7e5');
INSERT INTO evaluations VALUES(41534,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(41534,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(41534,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(41534,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(41534,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(41534,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(41534,36,'cp 78','c7c5 d4d5 e8g8 e1g1 c8g4 e3e4 a7a6 a2a4 f6e8 c4e2 g4f3 e2f3 b8d7 f3g4 e8d6 d1e2 d7e5 g4h3 c5c4 f2f4 e5d3 e4e5 d6f5 h3f5 g6f5 c1e3 e7e6 d5e6 f7e6 a4a5 a8c8 f1d1 d8e7 c3a4 e7b4 a4b6 c8c6 a1b1 b4b3 g1h1 f8d8 d1d2 b3b5 b1d1 b5a5 b6c4 c6c4 d2d3 d8d3 e2d3');
INSERT INTO evaluations VALUES(43474,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(43474,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43474,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43474,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(43474,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(43474,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(43474,36,'cp 32','d4f3 f8e7 c1g5 c8e6 g5f6 e7f6 c3d5 b8c6 f1c4 a8c8 d5e3 e8g8 c4e6 f7e6 c2c3 d8d7 e1g1 c8d8 d1e2 d6d5 e3g4 d5d4 f1d1 d7c7 a1c1 h7h6 h2h3 c7e7 d1d3 g8h7 g2g3 b7b5 a4b5 a6b5 g1g2');
INSERT INTO evaluations VALUES(43476,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(43476,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43476,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43476,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(43476,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(43476,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43476,36,'cp 15','f8e7 c1g5 b8c6 g5f6 e7f6 f1c4 c8e6 c3d5 a8c8 d5e3 e8g8 e1g1 e6c4 e3c4 c6d4 b2b3 d4c2 d1c2 b7b5 a4b5 a6b5 c2d3 b5c4 b3c4 d8b6 g2g3 b6c6 a1c1 c8a8 f1d1 h7h6 h2h4 a8a4 f3d2 f8c8 d2f1 a4c4 c1c4 c6c4 f1e3 c4d3 d1d3 c8c6 d3d1 f6d8');
INSERT INTO evaluations VALUES(39687,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(39687,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(39687,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(39687,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(39687,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(39687,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(39687,36,'cp 8','c1g5 c8e6 g5f6 e7f6 c3d5 b8d7 f1c4 a8c8 d1e2 d7b6 d5b6 d8b6 c4e6 f7e6 c2c3 e8g8 e1g1 a6a5 f1d1 f6e7 e2c2 b6c6 h2h3 e7d8 c2b3 f8f6 b3b5 d8b6 b5c6 c8c6 d1d2 h7h6 a1d1 b6c7 d2e2 f6f8 d1a1 c7b6 f3d2 g8f7 g1f1');
INSERT INTO evaluations VALUES(53647,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(53647,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(53647,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(53647,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(53647,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(53647,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(53647,36,'cp 83','g1f3 c6c5 e1g1 g8f6 d1e2 b8c6 f1d1 c5d4 b1c3 f8e7 f3d4 d8c7 d4c6 b7c6 b2b3 c8b7 c1b2 e8g8 a1c1 c6c5 c3a4 f8d8 c4d3 a8c8 a4c5 e7c5 b3b4 c7c6 f2f3 c5e3 e2e3 c6b6 e3b6 a7b6 c1c8 d8c8 b4b5 f6d5 a2a4');
INSERT INTO evaluations VALUES(53648,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(53648,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(53648,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(53648,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(53648,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(53648,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(53648,36,'cp 87','g8f6 e1g1 a7a6 b1c3 c6c5 d1e2 b7b5 c4d3 c5d4 e3d4 f8e7 a2a4 b5a4 d4d5 e8g8 f1e1 e7b4 a1a4 a6a5 c1g5 b8d7 d5e6 f7e6 e1d1 d8e7 d3b5 g8h8 g5h4 d7b6 a4a1 c8b7 f3d4 e6e5 d4c6 b7c6 b5c6 a8c8 c6f3 e7e6 h4f6');
INSERT INTO evaluations VALUES(45705,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(45705,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(45705,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(45705,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(45705,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(45705,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(45705,36,'cp 82','e1g1 c6c5 b1c3 a7a6 c4b3 b8c6 d1e2 c5d4 f1d1 f8e7 e3d4 e8g8 a2a3 f6d5 h2h3 c6a5 b3a2 e7f6 f3e5 a5c6 a2b3 c6e7 c3e4 e7f5 b3c2 c8d7 e4f6 d5f6 c2f5 e6f5 c1g5 d7c6 e5c6 b7c6 a1c1 f8e8 e2f3');
INSERT INTO evaluations VALUES(45703,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(45703,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(45703,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(45703,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(45703,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(45703,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(45703,36,'cp 75','a7a6 b1c3 c6c5 c4b3 b8c6 d1e2 c5d4 f1d1 f8e7 e3d4 e8g8 f3e5 f6d5 e2f3 d5c3 e5c6 c3e2 f3e2 b7c6 h2h3 c8b7 c1e3 a6a5 a1c1 a5a4 b3c4 f8e8 a2a3 g7g6 e2f3 e7f8 e3f4 f8g7 f4e5 d8g5 e5g7 g8g7 f3c3 g5f6 c4d3 a8d8 d3e4');
INSERT INTO evaluations VALUES(1475,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(1475,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(1475,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(1475,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(1475,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(1475,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(1475,36,'cp -48','c5d4 c3d4 c6d4 f1d3 d4f3 d2f3 f8b4 f3d2 g8e7 e1g1 e7c6 d2f3 b4e7 a2a3 e8g8 d3c2 f7f5 e5f6 e7f6 f3g5 f6g5 c1g5 b6b2 d1d3 g7g6 a1d1 c8d7 d1b1 b2e5 g5h6 f8f7 b1b7 c6a5 b7b1 a5c4 b1b7 e5h5 h6d2 h5f5 d2c3 f5d3 c2d3 e6e5');
INSERT INTO evaluations VALUES(1522,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(1522,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(1522,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(1522,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(1522,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(1522,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(1522,36,'cp -41','c3d4 c6d4 f1d3 d4c6 e1g1 g8e7 d2b3 e7g6 c1e3 b6d8 e3g5 f8e7 g5e7 d8e7 d3g6 h7g6 a1c1 c8d7 d1d2 e8g8 c1c3 f8c8 f1c1 c6b4 a2a3 c8c3 c1c3 b4c6 d2e3 a8c8 h2h3 b7b6 b3d4 c6d4 c3c8 d7c8 f3d4 e7d8');
INSERT INTO evaluations VALUES(1509,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(1509,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(1509,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(1509,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(1509,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(1509,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(1509,36,'cp -37','c6d4 f1d3 d4c6 e1g1 g8e7 d2b3 e7g6 c1e3 b6d8 d3g6 h7g6 a1c1 c8d7 d1d2 f8b4 d2e2 b4e7 e3c5 e8g8 e2e3 b7b6 c5e7 d8e7 a2a3 f8c8 c1c3 e7e8 h2h3 c6e7 b3d4 c8c3 e3c3 a8c8 c3d2 e7c6 f1c1 c6d4 f3d4 c8c4 c1c4');
INSERT INTO evaluations VALUES(1429,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(1429,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(1429,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(1429,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(1429,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(1429,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(1429,36,'cp -36','f1d3 d4c6 e1g1 g8e7 d2b3 e7g6 d3g6 h7g6 c1e3 b6d8 a1c1 c8d7 d1d2 f8b4 d2e2 b4e7 e3c5 e8g8 e2e3 b7b6 c5e7 d8e7 c1c3 a8c8 f1c1 f7f6 b3d4 c6e5 f3e5 c8c3 b2c3 f6e5 e3e5 e7f6 d4f3 f6e5 f3e5 d7e8 c1e1 f8f4 e5f3 f4e4 e1e4 d5e4 f3g5');
INSERT INTO evaluations VALUES(5826,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(5826,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(5826,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(5826,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(5826,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(5826,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(5826,36,'cp 80','g8f6 c2c4 e8g8 b1d2 d8e7 a1c1 f8d8 a2a3 h7h6 c4d5 e6d5 d1c2 a7a5 f1d3 d8d6 e1g1 c8g4 f1e1 a8e8 d2b3 b7b6 b3d2 g7g6 c2c3 f6e4 d2e4 d5e4 d3b5 g4d7 f3d2 h6h5 b2b4 a5b4 a3b4 h5h4');
INSERT INTO evaluations VALUES(11729,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(11729,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(11729,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(11729,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(11729,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(11729,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(11729,36,'cp 221','e1g1 f8d6 d2d4 e8g8 f2f3 f8e8 e5d3 d6f8 c1g5 d7d6 g5h4 c6c5 d4c5 d6c5 b1c3 h7h6 d3f2 c8e6 f3f4 d8d1 f1d1 f6g4 f4f5 g4f2 g1f2 e6c8 f2f3 c8b7 c3d5 f8d6 d1d2 a6a5 h4g3 a5a4 g3d6 c7d6 a1d1 b7c6 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(11728,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(11728,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(11728,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(11728,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(11728,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(11728,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(11728,36,'cp 213','f8d6 d2d4 e8g8 f2f3 f8e8 e5d3 d6f8 c1g5 d7d6 g5h4 c6c5 d4c5 d6c5 b1c3 h7h6 d3f2 f8e7 f1e1 a8b8 d1d8 e8d8 b2b3 c5c4 e1d1 c8e6 g1f1 g7g5 h4g3 f6e8 d1d8 b8d8 a1d1 c4b3 a2b3 d8d1 f2d1 f7f5 d1e3 f5e4 c3e4');
INSERT INTO evaluations VALUES(12998,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(12998,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(12998,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(12998,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12998,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12998,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(12998,36,'cp 15','c3d5 f8g7 d5f6 g7f6 c2c3 c8g4 h2h3 g4f3 d1f3 a8b8 d2d3 e8g8 c4d5 e7e6 d5c6 b7c6 c1h6 f6g7 h6g7 g8g7 a1b1 d8a5 a2a3 b8b7 f3e2 d6d5 f2f4 c5c4 f4f5 e6f5 e4f5 c4d3 e2d3 f7f6 d3d4 f8f7 b2b4 a5b5 d4d2 g6g5 b1d1 b5c4');
INSERT INTO evaluations VALUES(26872,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26872,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(26872,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26872,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26872,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26872,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26872,36,'cp 0','f1b5 g7g5 h2h4 g8e7 h4g5 e7g6 d1e2 a7a6 b5c6 b7c6 b1c3 h7h6 g5h6 h8h6 h1h6 f8h6 c1d2 c6c5 d4c5 g4f3 g2f3 d8h4 e2f2 h4h1 f2f1 h1h4');
INSERT INTO evaluations VALUES(28085,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(28085,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(28085,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(28085,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(28085,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(28085,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28085,36,'cp 23','f1d3 g8e7 e1g1 f5d3 c2d3 b8d7 c3e2 b4a5 a1b1 e8g8 b2b4 a5b6 a2a4 a7a5 b4b5 h7h6 c1a3 c6c5 d4c5 b6c5 a3c5 d7c5 e2d4 e7g6 g2g3 a8c8 b1c1 c5d7 d1e2 d7c5 e2d2 c5a4 c1c8 d8c8 d2a5 a4c3');
INSERT INTO evaluations VALUES(34412,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(34412,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(34412,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(34412,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(34412,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(34412,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(34412,36,'cp 109','c1e3 c8g4 d3d4 g4f3 d1f3 f6f3 g2f3 b8d7 d4d5 f8e7 h2h4 g8f6 a2a4 e8g8 a4a5 f8c8 a5a6 b7b5 d5c6 b5c4 c6d7 f6d7 a1a4 e7d8 e1d2 g8f8 c3d5 c8c6 h4h5 d8b6 h1a1 b6e3 f2e3 a8b8 d2c3 d7b6 a4b4 g7g5 h5g6 f7g6');
INSERT INTO evaluations VALUES(34413,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(34413,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(34413,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(34413,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(34413,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(34413,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(34413,36,'cp 111','c8g4 d3d4 g4f3 d1f3 f6f3 g2f3 e5d4 e3d4 g8f6 e1c1 b8d7 f3f4 b7b5 c4e2 b5b4 c3a4 e8c8 e2f3 c6c5 d4e3 a7a5 c2c3 g7g5 c3b4 a5b4 a2a3 b4a3 b2a3 g5f4 e3f4 d7e5 f4e5 d6e5 a4b6 c8c7 d1d8 c7d8 b6c4 d8c7 c1c2 f8e7 c4e5 f6d7 e5c4 e7f6 f3e2 h8e8 f2f3 f6d4');
INSERT INTO evaluations VALUES(35632,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(35632,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(35632,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(35632,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(35632,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(35632,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35632,36,'cp 26','c4d5 f6d5 g5e7 d8e7 c3d5 e6d5 a1c1 c8b7 d1a4 c7c5 b2b3 f8c8 f1e2 a7a5 e1g1 b8d7 e2b5 d7f8 f1d1 h7h6 d4c5 b6c5 a4g4 f8e6 g4g3 e6c7 b5d3 a5a4 f3h4 c7e6 h4f5 e7f6 f5d6 c8c7 d3f5 a4b3 a2b3');
INSERT INTO evaluations VALUES(35630,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(35630,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(35630,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(35630,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(35630,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35630,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(35630,36,'cp 22','f6d5 c3d5 e6d5 g5e7 d8e7 a1c1 c8b7 f1d3 c7c5 d1c2 g7g6 d4c5 f8c8 e1g1 b6c5 f1e1 c5c4 d3e2 b8d7 b2b3 c8c7 f3d4 a8c8 e1d1 a7a6 e2g4 h7h5 g4d7 e7d7 c2b2 c4c3 b2e2 a6a5 c1c2 a5a4 d1c1 a4a3 e2b5 d7b5');
INSERT INTO evaluations VALUES(35629,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(35629,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(35629,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(35629,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(35629,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(35629,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(35629,36,'cp 25','c3d5 e6d5 g5e7 d8e7 a1c1 c8b7 f1d3 c7c5 d1c2 g7g6 d4c5 f8c8 e1g1 b6c5 f1e1 c5c4 d3e2 b8d7 b2b3 c8c7 f3d4 a8c8 e1d1 b7a6 c2d2 d7e5 d2a5 e7d6 b3b4 h7h5 a2a3 e5g4 d4f3 g4f6 b4b5 a6b7 a5a7 c4c3');
INSERT INTO evaluations VALUES(40002,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(40002,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40002,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40002,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(40002,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40002,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40002,36,'cp 24','e8g8 d2d4 d7d5 c4d5 f6d5 e1g1 c5d4 f3d4 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 c8f5 c6a8 g7c3 c1g5 c3a1 f1a1 d8a8 g5e7 a8e8 e7c5 a7a5 g1f1 e8b8 f1e1 b8b1 a1b1 f5b1 a2a3 f7f6 e1d2 a5a4 d2c3 b1a2 c5b6 f6f5 c3b4');
INSERT INTO evaluations VALUES(35703,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(35703,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(35703,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(35703,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35703,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35703,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(35703,36,'cp 23','d2d4 d7d5 d4c5 d8a5 e1g1 d5c4 d1d4 b8c6 d4c4 c8e6 c4h4 a5c5 f3g5 e6f5 h2h3 c5a5 e2e4 f5d7 g5f3 d7e6 c1d2 a5h5 h4h5 f6h5 d2e3 f8d8 a1c1 h7h6 f1d1 d8d1 c1d1 g7c3 b2c3 h5f6 f3d4 e6a2 d4c6 b7c6 d1a1 a2b3 a1a7 a8a7 e3a7');
INSERT INTO evaluations VALUES(41226,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(41226,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(41226,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(41226,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41226,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41226,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(41226,36,'cp 9','c7c5 d2d4 c5d4 f3d4 b8c6 f1e1 e8g8 c4b3 c6d4 d1d4 a7a6 a2a4 f6d7 d4b4 d7c5 c1e3 c5b3 b4b3 c8e6 c3d5 e6d5 e4d5 b7b5 a4b5 a6b5 b3b5 a8b8 b5d3 g7b2 a1a7 b8a8 a7b7 b2f6 e1b1 d8a5 c2c4 f8c8 b7b4 a5a3 d3e4 a3a5 e3d4 f6d4 e4d4');
INSERT INTO evaluations VALUES(43444,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(43444,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(43444,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43444,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(43444,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43444,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(43444,36,'cp 43','c8e6 f2f4 e5f4 d1d2 f8e7 e1c1 b8d7 g5f4 d7e5 b3d4 e8g8 d4e6 f7e6 g2g3 d8a5 f1h3 g8f7 c1b1 a8c8 f4e5 d6e5 h1f1 c8c6 d2e2 f7g8 d1d3 g8h8 g3g4 h7h6 g4g5 h6g5 f1g1 b7b5 g1g5 b5b4 c3d1 f8c8');
INSERT INTO evaluations VALUES(28540,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(28540,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(28540,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28540,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(28540,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(28540,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(28540,36,'cp 44','f2f4 e5f4 d1d2 f8e7 e1c1 b8c6 g5f4 e8g8 f4d6 e6g4 d6e7 d8e7 d1e1 a8d8 d2f2 h7h6 c1b1 c6e5 f1e2 g4e2 e1e2 e5c4 e4e5 f6d5 f2f3 e7d7 h1e1 f8e8 g2g3 d7c6 b1a1 d5b4 f3c6 b4c6 e5e6 f7e6 e2e6 e8e6 e1e6 c4d2 b3d2 d8d2');
INSERT INTO evaluations VALUES(51421,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(51421,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(51421,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(51421,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(51421,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(51421,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(51421,36,'cp 93','d7d5 g1f3 d5e4 d3e4 c7c6 f3e5 g8f6 e4f3 c8b7 c1f4 f8e7 e1g1 e8g8 f1e1 d8c8 c2c4 b8d7 b1c3 f8e8 d1d2 c6c5 d4d5 e6d5 c4d5 d7e5 f4e5 c8f5 d5d6 b7f3 d6e7 f3c6 e5g3 h7h6 a1d1 f5g6 a2a4 f6e4 c3e4 c6e4 a4a5 e4c6 a5b6');
INSERT INTO evaluations VALUES(118,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(118,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(118,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(118,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(118,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(118,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(118,36,'cp 192','c1f4 g7g5 f4c7 b7b6 b1c3 c8b7 f2f3 h7h5 h2h4 g5h4 a1d1 b7c6 f1c4 g8f6 e1f2 h8g8 c3b5 c6b5 c4b5 a7a6 b5e2 g8g6 e4e5 f6h7 e2d3 g6h6 d3e4 a8c8 c7d6 e7d6 d1d6 h6d6 e5d6 h7f6 e4b7 c8b8 b7a6 b8a8 a6c4 e8f8 h1h4');
INSERT INTO evaluations VALUES(19497,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(19497,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(19497,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(19497,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(19497,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(19497,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(19497,36,'cp 30','e2e4 e7e6 d2d4 c5d4 f3d4 a7a6 f1e2 d8c7 e1g1 g8f6 c1e3 f8e7 f2f4 d7d6 d1e1 e8g8 e1g3 d6d5 e4e5 c6d4 e3d4 e7c5 d4c5 c7c5 g3f2 f6d7 e2d3 c5f2 g1f2 f7f6 e5f6 d7f6 a1e1 c8d7 f2g1 a8e8 c3e2 e6e5 f4e5');
INSERT INTO evaluations VALUES(36729,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36729,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(36729,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(36729,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36729,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(36729,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(36729,36,'cp -12','c7c5 d4c5 d5c4 e1g1 b8a6 c5c6 d8c7 e2c4 c7c6 d1e2 c8e6 c4a6 c6a6 e2a6 b7a6 f3d4 f6d5 d4e6 f7e6 c3e4 a8c8 e4g5 c8c6 f1d1 c6c2 f2f4 g7b2 c1b2 c2b2 g5e6 d5e3 d1d3 e3c2 a1d1 f8b8 d3d8 b8d8 d1d8 g8f7 e6g5 f7g7 d8d7 b2a2 d7e7 g7f8 e7a7 h7h6 g5f3');
INSERT INTO evaluations VALUES(36029,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36029,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36029,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36029,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36029,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36029,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36029,36,'cp 0','d4c5 d5c4 e1g1 b8a6 c5c6 d8b6 e2c4 b6c6 d1e2 c8e6 c4a6 c6a6 e2a6 b7a6 f3d4 f6d5 d4e6 f7e6 c3e4 a8b8 e4c5 g7b2 c1b2 b8b2 c5e6 f8c8 f1d1 d5c3 d1d7 c3e2 g1f1 c8c2 d7e7 e2g3 f2g3 c2f2 f1e1 f2e2 e1f1');
INSERT INTO evaluations VALUES(41466,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41466,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(41466,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(41466,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(41466,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41466,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41466,36,'cp 19','c7c5 b1d2 e8g8 f1e2 b8d7 h2h3 b7b6 a2a4 a7a6 e1g1 c8b7 d1b3 b7c6 b3c2 f8e8 f1c1 c5c4 b2b3 b6b5 b3c4 b5c4 e3e4 f6e4 d2e4 d5e4 f3d2 e7e5 f4e3 e5d4 c3d4 d7b6 a4a5 b6d5 e2c4 d5e3 f2e3 c6d5 c4d5 d8d5 g1h1 d5f5');
INSERT INTO evaluations VALUES(42357,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(42357,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42357,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(42357,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42357,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42357,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42357,36,'cp 10','g1f3 b8c6 e1g1 e8g8 h2h3 a7a6 c1e3 d7d6 e3c5 d6c5 a2a4 f8e8 c3e2 f6h5 g1h2 h5f4 f3e5 c6e5 e2f4 d8d6 g2g3 e5c4 d3c4 e8e4 d1d6 c7d6 b2b3 c8f5 a1d1 g7g5 f4d5 e4e2 d5e3 f5c2 e3c2 e2c2 d1d6 g8g7 h2g2');
INSERT INTO evaluations VALUES(42355,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(42355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42355,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(42355,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(42355,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42355,36,'cp 0','b8c6 h2h3 a7a6 c1e3 d7d6 d3d4 e5d4 f3d4 c5d4 e3d4 e8g8 e1g1 c8e6 c4b3 c6d4 d1d4 f6h5 d4e3 f8e8 a1d1 d8h4 g2g4 e6b3 a2b3 h5f6 g1g2 a8d8 e3f4 f6h5 f4e3');
INSERT INTO evaluations VALUES(47526,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47526,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47526,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47526,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47526,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47526,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47526,36,'cp 0','e7e6 g1f3 f8d6 d2d4 e8g8 e1g1 c7c5 d4d5 e6d5 d1d5 c8e6 d5b7 e6c4 b7a8 c4d5 a8a7 d5f3 g2f3 d8h4 f1e1 h4h2 g1f1 h2h3 f1e2 h3e6 e2f1');
INSERT INTO evaluations VALUES(48135,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(48135,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(48135,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48135,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(48135,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(48135,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(48135,36,'cp 13','d2d4 c7c5 g1f3 f8d6 d4c5 d6c5 d1d8 e8d8 f3g5 d8e7 g5e4 b8d7 c4b5 b7b6 b5c6 a8b8 e4c5 b6c5 c1a3 c8b7 c6d7 e7d7 a3c5 b8c8 c5a7 c8c3 a2a4 c3c2 e1g1 h8c8 f1b1 b7a6 a7e3 e6e5 a4a5 c2c6 h2h4 d7e6 h4h5 f7f6 b1b4 c6c4 b4b6 c4c6 a1b1 e6d5 h5h6 g7h6 e3h6 d5e6 b6c6 c8c6');
INSERT INTO evaluations VALUES(51151,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(51151,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(51151,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(51151,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(51151,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(51151,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(51151,36,'cp 121','f1d3 c7c5 c2c3 g8f6 e1g1 c5c4 d3c2 f8e7 b2b3 c4b3 a2b3 d7d6 c3c4 c8b7 d4d5 e8g8 b1c3 e6d5 c3d5 f6d5 c4d5 e7f6 f3d4 b8d7 c1e3 a8c8 a1c1 g7g6 d1d2 f8e8 f1d1 d8b6 f2f3 d7e5 h2h3 f6g7');
INSERT INTO evaluations VALUES(51150,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(51150,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(51150,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(51150,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(51150,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(51150,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(51150,36,'cp 78','c7c5 c2c3 c8b7 e1g1 g8f6 f1e1 d7d6 a2a4 c5c4 d3c2 b8d7 b1a3 f8e7 a1b1 d8a5 d4d5 e8g8 d5e6 f7e6 f3d4 d7c5 f2f3 f6d7 c1e3 d7e5 a4b5 a6b5 a3b5 c5d3 e1f1 b7a6 b2b4 a5b6 b5a3 d3f4 b1a1 a6b7 e3f4 f8f4 g2g3');
INSERT INTO evaluations VALUES(52506,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(52506,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(52506,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(52506,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(52506,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(52506,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(52506,36,'cp -83','b8c6 g1f3 g8f6 d1e2 e7e6 c2c3 f6e4 e2e4 d7d5 c4d5 d8d5 e4d5 e6d5 f3d4 c6d4 c3d4 c8e6 c1e3 e8d7 e1d2 a8c8 b1c3 c8c4 h2h4 f8d6 h4h5 h8c8 a1c1 h7h6 a2a3 a7a6 f2f3 b7b5 c3e2 a6a5 e2f4 b5b4 a3b4 d6b4 d2d1 a5a4 c1c4 d5c4 f4e6 d7e6');
INSERT INTO evaluations VALUES(19527,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(19527,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(19527,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(19527,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(19527,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(19527,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(19527,36,'cp -78','g1f3 g8f6 d1e2 e7e6 c2c3 f6e4 e2e4 d7d5 c4d5 d8d5 e4d5 e6d5 f3d4 c6d4 c3d4 c8e6 c1e3 a8c8 b1c3 f8e7 e1d2 h7h5 h1c1 e8d7 c3e2 g7g5 c1c8 h8c8 a1c1 c8c4 b2b3 c4c1 e2c1 f7f6 g2g3 e6f5 c1d3 b7b6 h2h4 d7e6 d2e2 f5g6 a2a4 g6d3 e2d3 e6f5 d3c3 e7d6 h4g5');
INSERT INTO evaluations VALUES(55171,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(55171,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(55171,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(55171,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(55171,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(55171,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(55171,36,'cp -38','b8c6 g1f3 c8g4 h2h3 g4f3 d1f3 e7e6 f1a6 a8b8 a6b7 c6e5 f3e4 c7c5 e4e5 b8b7 e1g1 g8f6 d2d3 f8d6 e5e2 e8g8 c3d4 c5d4 b1d2 h7h6 d2c4 d8b8 a2a3 b7b5 c1d2 d6h2 g1h1 h2f4 b2b4 f4d2 e2d2 f8c8 a1c1 b5f5 d2e2 f6d5 e2e4 f5g5');
INSERT INTO evaluations VALUES(5932,30,'cp 283',NULL);
INSERT INTO evaluations VALUES(5932,31,'cp 296',NULL);
INSERT INTO evaluations VALUES(5932,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(5932,33,'cp 303',NULL);
INSERT INTO evaluations VALUES(5932,34,'cp 313',NULL);
INSERT INTO evaluations VALUES(5932,35,'cp 309',NULL);
INSERT INTO evaluations VALUES(5932,36,'cp 351','e1f1 a7a6 b5c6 b7c6 d1a4 a6a5 a2a3 c8b7 a4c2 g8f6 a3b4 a5b4 a1a8 d8a8 e4e5 f6d7 e5d6 e8g8 d6c7 c6c5 d4c5 f8c8 c1e3 c8c7 h2h3 d7c5 f1g1 b7f3 g2f3 a8d8 c2e2 d8d3 e2d3 c5d3 g1h2 d3b2 b1d2 b2d3 d2b3 c7c2 h1d1 c2f2 e3f2');
INSERT INTO evaluations VALUES(9358,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9358,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9358,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(9358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9358,36,'cp -3','c8g4 h2h3 g4h5 g2g4 h5g6 d2d3 d8d7 g1g2 e8c8 f3h4 a7a6 f2f4 e5f4 c1f4 c5d4 c4b3 c8b8 b3a4 h8e8 d1f3 d4e5 f4e5 d6e5 f3g3 b7b5 a4b3 c6d4 a1d1 d7d6 g2h2 d4b3 c2b3 f6d7 b3b4 f7f6 d1d2 d7f8 h4f5 g6f5 g4f5 g7g6 g3f2 g6f5');
INSERT INTO evaluations VALUES(24137,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(24137,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(24137,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24137,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24137,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24137,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24137,36,'cp -2','h2h3 g4h5 g2g4 h5g6 d2d3 d8d7 g1g2 e8c8 f3h4 c8b8 f2f4 e5f4 c1f4 a7a6 c4b3 h8e8 g2h2 c5d4 d1d2 d4e5 f4e5 d6e5 d2g2 c6d4 b3a2 d7e7 f1f2 f6d7 g2g3 f7f6 c3d5 e7c5 d5e3 d7f8 a1f1 d4e6 e3d5 c5d6 b2b4 c7c6 d5e3 d6c7 a2e6 f8e6 h4g2 e6g5');
INSERT INTO evaluations VALUES(10193,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(10193,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(10193,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(10193,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(10193,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(10193,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(10193,36,'cp 75','e5d4 e3d4 c5b4 d4d5 b4c3 b2c3 c6a5 f1d3 d8e7 d3e2 e8g8 e1g1 d7d6 f1e1 h7h6 e2d3 e7d8 c1e3 c8g4 h2h3 g4h5 g2g4 h5g6 g4g5 h6g5 e3g5 f8e8 e1e8 d8e8 g5f6 g7f6 d1d2 b7b6 f3d4 e8d7 d3g6 f7g6 d2h6 d7h7');
INSERT INTO evaluations VALUES(10299,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(10299,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(10299,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(10299,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(10299,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(10299,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(10299,36,'cp 70','e3d4 c5b4 d4d5 b4c3 b2c3 c6a5 f1d3 d8e7 d3e2 e8g8 e1g1 d7d6 f1e1 c8g4 c1g5 a8e8 f3d4 g4e2 e1e2 e7d7 g5f6 g7f6 d1d3 e8e2 d4e2 d7g4 c4c5 d6c5 e2g3 a5c4 h2h3 g4f4 g3h5 f4g5 h5f6 g5f6 d3c4 b7b6 a1e1 f8d8 c4g4 f6g6 g4f3 g8f8 c3c4 d8e8 e1e8 f8e8 g2g4 h7h6 g1g2 e8f8');
INSERT INTO evaluations VALUES(10298,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(10298,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(10298,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(10298,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(10298,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(10298,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(10298,36,'cp 75','c5b4 d4d5 d8e7 f1e2 c6e5 e1g1 e8g8 f3e5 e7e5 e2d3 d7d6 f1e1 c8g4 f2f3 b4c5 g1h1 e5h5 c1f4 g4d7 g2g4 h5h3 d3f1 h3h4 d1d2 h4f2 f1e2 f2d4 e2d3 d4f2 d2f2 c5f2 e1e2 f2d4 a1c1 f8e8 h1g2 g8f8 h2h3 h7h6 f4g3 e8e2 c3e2 d4e3 c1c2 a7a5 a2a3 a8e8');
INSERT INTO evaluations VALUES(10511,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(10511,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(10511,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(10511,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(10511,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(10511,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(10511,36,'cp 73','d4d5 d8e7 f1e2 c6e5 e1g1 e8g8 f3e5 e7e5 e2d3 d7d6 f1e1 c8g4 f2f3 b4c5 g1h1 e5h5 c1f4 g4d7 g2g4 h5h3 d3f1 h3h4 d1d2 h4f2 f1e2 f2d4 e2d3 d4f2 d2f2 c5f2 e1e2 f2d4 a1c1 h7h6 h1g2 f8e8 h2h4 a7a5 f4d2 a5a4 c1e1 e8e2 d3e2 g8f8 e2f1 a8e8 e1e8 f6e8 h4h5 d4c3 d2c3 b7b6');
INSERT INTO evaluations VALUES(16065,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(16065,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(16065,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(16065,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(16065,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(16065,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(16065,36,'cp -85','b1c3 f8g7 d2d3 e8g8 a2a3 d7d6 c4a2 d8e8 h2h4 h7h6 g1e2 g8h7 c1d2 c8g4 f3g3 c6d4 e2d4 e5d4 c3d5 f6d5 a2d5 d4e3 f2e3 g4e6 d5e6 g7b2 a1b1 b2e5 g3g4 f7e6 b1b7 e8c6 h4h5 g6h5 g4e4 c6e4 d3e4 a8c8 h1h5 e5g3 e1d1 f8f2 d1c1 f2g2 d2c3');
INSERT INTO evaluations VALUES(16064,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(16064,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(16064,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(16064,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(16064,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(16064,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(16064,36,'cp -70','f8g7 d2d3 e8g8 a2a3 d7d6 c4a2 d8e8 h2h4 h7h6 c1d2 a7a5 g1e2 g8h7 e3e4 c6d4 e2d4 e5d4 c3d5 f6d5 a2d5 f7f5 e1g1 f5f4 g2g3 f4g3 f3g3 e8e5 g3e5 d6e5 f2f4 c7c6 d5c4 c8h3 f1f2 e5f4 f2f4 f8f4 d2f4 b7b5');
INSERT INTO evaluations VALUES(16316,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(16316,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(16316,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16316,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(16316,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(16316,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(16316,36,'cp -23','f8c5 e1g1 e8g8 b5c6 b7c6 f3e5 d8e7 e5f3 e7e4 e2e4 f6e4 d2d3 e4f6 c1g5 c5e7 f1e1 e7d8 c2c4 d7d6 b1c3 h7h6 g5d2 c8f5 d3d4 f8e8 b2b3 a7a5 e1e8 f6e8 a1e1 f5d7 e1e2 g7g5 h2h4 g5g4 f3e1 d8h4 d2h6 e8g7');
INSERT INTO evaluations VALUES(10035,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(10035,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(10035,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(10035,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(10035,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(10035,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10035,36,'cp -17','e1g1 e8g8 b5c6 b7c6 f3e5 d8e8 e5f3 e8e4 e2e4 f6e4 d2d3 e4f6 c1g5 c5e7 f1e1 e7d8 c2c4 d7d6 b1c3 h7h6 g5d2 a7a5 b2b3 c8g4 d3d4 f6d7 a1d1 g4h5 d4d5 c6d5 c3d5 a5a4 d2b4 a4b3 a2b3 d7f6 b4c3 f6d5 d1d5 h5f3');
INSERT INTO evaluations VALUES(24644,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(24644,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(24644,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(24644,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24644,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24644,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(24644,36,'cp 23','c1f4 e7e6 h2h3 g4h5 b1d2 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 d6c7 d1e2 c6e7 a2a4 h5g6 f3e5 g6d3 e2d3 f8c8 e5g4 f6e8 d2f3 e8d6 f3e5 a7a5 d3e2 d6e4 e5d3 c7d6 g4e3 c8c7 h3h4 h7h6');
INSERT INTO evaluations VALUES(24631,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24631,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(24631,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24631,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(24631,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24631,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24631,36,'cp 25','e7e6 h2h3 g4h5 b1d2 f8d6 f4d6 d8d6 e1g1 e8g8 f1e1 d6c7 d1b1 a8e8 f3e5 c6e5 e1e5 h5g6 d3g6 h7g6 b1d3 a7a5 e5e2 b7b5 a2a3 e8a8 d2f3 c7b6 h3h4 f8c8 g2g3 b5b4 c3b4 a5b4 a3a4 b6a6 d3a6 a8a6 f3e5 f6e4');
INSERT INTO evaluations VALUES(24723,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24723,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(24723,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24723,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(24723,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(24723,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24723,36,'cp 14','h2h3 g4f3 d1f3 f8d6 f4d6 d8d6 e1g1 a8d8 b1d2 e8g8 f1e1 f8e8 d3b5 e6e5 d4e5 c6e5 f3e3 e5c6 e3d3 a7a6 b5c6 d6c6 d2b3 e8e1 a1e1 d8e8 e1e2 e8e2 d3e2 c6d6 e2e3 h7h6 b3c5 d6b6 b2b4 f6e4 a2a4 e4c5 b4c5 b6f6');
INSERT INTO evaluations VALUES(26813,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26813,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(26813,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(26813,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(26813,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(26813,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(26813,36,'cp 12','g8f6 d3f5 e6f5 e1g1 f8d6 f4g5 h7h6 g5f6 d8f6 c2c4 d5c4 b1d2 e8g8 d2c4 c6e7 d1b3 a8b8 h2h3 e7d5 f3e5 d6e5 c4e5 f8d8 a1c1 c7c6 f1d1 f6e6 a2a4 e6e7 a4a5 a7a6 e5d3 d8e8 b3c2 g7g6 c2c5 e7e4 c5a3 e4e6');
INSERT INTO evaluations VALUES(37775,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(37775,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37775,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37775,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37775,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37775,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37775,36,'cp 0','g2g3 a7a6 a2a4 d7d6 f1g2 e7e6 g1e2 b8c6 e1g1 g8e7 a1b1 e8g8 d2d3 c8d7 c1e3 d8a5 d3d4 c5d4 e2d4 c6e5 d4b3 a5b4 c4c5 e5c4 e3g5 e7c6 c3a2 b4a4 a2c3 a4b4 c3a2 b4a4');
INSERT INTO evaluations VALUES(38246,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38246,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(38246,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38246,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(38246,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(38246,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38246,36,'cp 13','g8f6 c1f4 b8c6 d4d3 c8e6 b1d2 e6c4 d3c4 e8g8 e1g1 f8e8 f1e1 e7f8 h2h3 h7h6 a1d1 f6d7 f4e3 a7a6 a2a4 d7f6 e3f4 f6d7 c2c3 d8f6 f4g3 f6e6 c4d3 c6e5 d3c2 e5f3 d2f3 d7c5 e4e5 d6e5 g3e5 e6c6');
INSERT INTO evaluations VALUES(50220,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50220,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(50220,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(50220,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50220,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50220,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(50220,36,'cp 27','c7c5 d4c5 d8a5 b1c3 f8g7 c1d2 a5c5 e2e4 b8c6 f1e2 d7d5 d2e3 c5a5 e4e5 f6e4 d1d5 e4c3 d5a5 c6a5 b2c3 e8g8 e1c1 c8e6 e3c5 e6a2 c5e7 f8e8 d1d7 a8c8 h1d1 a2e6 d7d8 c8d8 d1d8 e8d8 e7d8 a5c4 d8f6 g7h6 c1b1 e6d5 h3h4 b7b5 e2c4 d5c4 f3d4 a7a5 f2f3 a5a4 b1b2');
INSERT INTO evaluations VALUES(50687,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(50687,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(50687,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(50687,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(50687,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(50687,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(50687,36,'cp 121','g8f6 b1c3 g7g6 h2h3 f8g7 f1d3 e8g8 e1g1 e7e6 c1g5 h7h6 g5f4 b8c6 d1d2 g6g5 f4e3 e6e5 d4d5 c6e7 f3h2 c8d7 a2a4 f6h5 a1a3 f7f5 a4a5 f5f4 e3b6 d8e8 b6c7 f4f3 h2f3 f8f3 g2f3 h5f4 c7d6 d7h3 c3e2 h3f1');
INSERT INTO evaluations VALUES(43420,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(43420,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(43420,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(43420,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(43420,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(43420,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(43420,36,'cp 119','b1c3 g7g6 h2h3 f8g7 f1d3 e8g8 e1g1 e7e6 f1e1 b7b5 e4e5 d6e5 d4e5 f6d5 d3e4 c8b7 c1g5 d8d7 e4d5 b7d5 d1e2 d7b7 c3d5 e6d5 e2e3 b8d7 g5h6 b7b6 e3d2 f8e8 h6g7 g8g7 d2d5 a8d8 d5e4 d7f8 a1c1 d8c8 e4f4 h7h6 c1d1 c8d8 d1d8 e8d8 a2a3 f8e6');
INSERT INTO evaluations VALUES(43418,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(43418,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(43418,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(43418,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(43418,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(43418,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(43418,36,'cp 111','g7g6 h2h3 f8g7 f1d3 e8g8 e1g1 e7e6 c1g5 h7h6 g5f4 b8c6 d1d2 g6g5 f4e3 e6e5 d4d5 c6e7 f3h2 c8d7 a2a4 f6h5 a1a3 f7f5 e4f5 d7f5 a3b3 d8d7 a4a5 h5f4 e3f4 f5d3 d2d3 f8f4 f1e1 f4d4 d3b1 a8f8 h2g4 d4f4 c3e2 f4f7 e2g3 e7d5');
INSERT INTO evaluations VALUES(55418,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(55418,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55418,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55418,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(55418,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(55418,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(55418,36,'cp 30','d3d4 g8f6 d4e5 f6e4 f1c4 b8c6 c4f7 e8f7 d1d5 c8e6 d5e4 d6e5 f3e5 c6e5 e4e5 f8d6 e5h5 f7g8 e1g1 d8f6 b1c3 e6f5 c3b5 f5c2 b5d6 f6d6 c1e3 g8h7 f1c1 c2e4 f2f3 e4d3 c1c3 h8e8 e3f2 d3a6 h5c5 d6c5 f2c5 e8e2 b2b4 a8e8');
INSERT INTO evaluations VALUES(4109,30,'cp -289',NULL);
INSERT INTO evaluations VALUES(4109,31,'cp -275',NULL);
INSERT INTO evaluations VALUES(4109,32,'cp -281',NULL);
INSERT INTO evaluations VALUES(4109,33,'cp -278',NULL);
INSERT INTO evaluations VALUES(4109,34,'cp -285',NULL);
INSERT INTO evaluations VALUES(4109,35,'cp -280',NULL);
INSERT INTO evaluations VALUES(4109,36,'cp -288','f8f7 c4f7 g8f7 b1c3 h7h6 c3a4 c5e7 c2c4 e7b4 a4c3 d7d6 h2h3 f7g8 e1g1 b4c5 c3d5 c8e6 g1h1 e6f7 c1e3 a7a5 h1h2 f6d7 a2a3 c5d4 a1b1 a5a4 d1g4 g8h8 g4f5 f7g8 f5h5 g8h7 h5f7 a8b8 h2h1 d7c5 f7c7 d8c7 d5c7');
INSERT INTO evaluations VALUES(4647,30,'cp -278',NULL);
INSERT INTO evaluations VALUES(4647,31,'cp -271',NULL);
INSERT INTO evaluations VALUES(4647,32,'cp -280',NULL);
INSERT INTO evaluations VALUES(4647,33,'cp -285',NULL);
INSERT INTO evaluations VALUES(4647,34,'cp -281',NULL);
INSERT INTO evaluations VALUES(4647,35,'cp -287',NULL);
INSERT INTO evaluations VALUES(4647,36,'cp -262','c4f7 g8f7 b1c3 h7h6 c1e3 d7d6 d1d2 f7g8 e1c1 c8e6 e3c5 d6c5 f2f4 e5f4 d2f4 a7a5 f4h4 d8d6 c3b5 d6d7 h1f1 a8f8 b5c3 c6d4 e4e5 f6h7 f1f8 h7f8 d1f1 e6f5 h4f2 f8g6 f2e3 g8h7 f1e1 a5a4 c3e2 d4e6 e1f1 a4a3 b2a3');
INSERT INTO evaluations VALUES(4638,30,'cp -278',NULL);
INSERT INTO evaluations VALUES(4638,31,'cp -287',NULL);
INSERT INTO evaluations VALUES(4638,32,'cp -273',NULL);
INSERT INTO evaluations VALUES(4638,33,'cp -268',NULL);
INSERT INTO evaluations VALUES(4638,34,'cp -287',NULL);
INSERT INTO evaluations VALUES(4638,35,'cp -264',NULL);
INSERT INTO evaluations VALUES(4638,36,'cp -272','g8f7 b1c3 f7g8 e1g1 d7d6 c3a4 c8g4 d1d2 g4e6 a4c5 d6c5 f2f3 f6h5 d2g5 d8g5 c1g5 b7b6 g5e3 a7a5 f1f2 a5a4 f2d2 h7h6 g1f2 a8d8 g2g3 h5f6 f2e2 g8f7 c2c3 f6h5 e3f2 g7g5 f2e3 f7g7 b2b3 a4b3 a2b3 e6b3 d2b2 b3e6');
INSERT INTO evaluations VALUES(11429,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11429,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(11429,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(11429,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(11429,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(11429,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(11429,36,'cp -10','b7b5 a4b3 c6a5 c1d2 a5b3 a2b3 f8e7 d2a5 e8g8 c2c4 d8d7 b1c3 f8e8 f1e1 c7c6 d3d4 e5d4 d1d4 d7b7 e4e5 d6e5 f3e5 e7f8 e1e2 f6d7 e5c6 e8e2 c6d8 f8c5 d8b7 c5d4 c3e2 d4b2 a1d1 b2f6 b7d6 b5c4 d6c4 a8b8 e2f4 d7c5 b3b4 c5b3 c4b6 b3a5 b4a5');
INSERT INTO evaluations VALUES(11652,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11652,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11652,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11652,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(11652,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(11652,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(11652,36,'cp 0','a4b3 c6a5 c1d2 a5b3 a2b3 f8e7 f1e1 e8g8 d2a5 d8d7 b1c3 f8e8 d3d4 e5d4 d1d4 c7c6 h2h3 e7f8 d4d3 d7b7 a1d1 c8e6 f3d4 e6d7 d4f3 d7e6');
INSERT INTO evaluations VALUES(11631,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11631,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11631,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11631,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(11631,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11631,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(11631,36,'cp -13','c6a5 c1d2 a5b3 a2b3 f8e7 f1e1 e8g8 b1c3 c8b7 c3e2 f8e8 c2c4 e7f8 h2h3 f6d7 d2a5 d7b8 e2g3 g7g6 d1d2 b8d7 b3b4 f8g7 b2b3 h7h5 a1c1 a8c8 c4c5 d6c5 b4c5 d8f6 b3b4 d7f8 g3f1 f8e6 c5c6 b7a8 d2e3 e6f4');
INSERT INTO evaluations VALUES(19276,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19276,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19276,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19276,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19276,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19276,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19276,36,'cp 0','g2g3 f8g7 g1e2 e7e6 f1g2 d7d6 c1e3 c6d4 e3d2 g8e7 e2d4 c5d4 c3e2 d6d5 e4d5 e7d5 c2c4 d4c3 b2c3 e8g8 e1g1 d5b6 d2e3 b6d5');
INSERT INTO evaluations VALUES(24577,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(24577,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(24577,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(24577,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(24577,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(24577,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(24577,36,'cp -31','g7g6 d1e2 f8g7 e1g1 a7a6 a2a4 e8g8 f1d1 d8c7 d4f3 f8c8 h2h3 d7b6 e4e5 d6e5 f3e5 e7e6 e5d3 f6d5 c3d5 b6d5 c2c3 a6a5 h3h4 h7h6 d1e1 c7c4 d3e5 c4e2 e1e2 c8c7 c1d2 a8c8 a1d1 d5b6 b2b3 b6d7 d2f4 d7e5 f4e5 g7e5');
INSERT INTO evaluations VALUES(24546,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(24546,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(24546,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(24546,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(24546,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(24546,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(24546,36,'cp -20','e1g1 f8g7 d1e2 e8g8 f1d1 d8c7 a2a4 a7a6 d4f3 f8c8 h2h3 b7b6 c1g5 c7b7 d1e1 c8c5 a1d1 b6b5 a4b5 a6b5 g5e3 c5c8 e3d4 b5b4 c3d5 f6d5 e4d5 g7f6 c2c4 b4c3 b2c3 d7c5 c3c4 f6d4 f3d4');
INSERT INTO evaluations VALUES(24545,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(24545,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(24545,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(24545,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(24545,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(24545,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(24545,36,'cp -19','f8g7 a2a4 e8g8 d1e2 d8c7 f1d1 f8c8 h2h3 a7a6 d4f3 d7b6 c1g5 e7e6 f3d2 f6e8 a1a3 b6c4 d2c4 c7c4 e2d3 g7c3 b2c3 c4c5 g5c1 a8b8 d3e3 e6e5 f2f3 g8g7 e3c5 c8c5 a3b3 g7f6 c1d2 f6e6 b3b6 c5c6');
INSERT INTO evaluations VALUES(24024,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(24024,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(24024,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(24024,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(24024,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(24024,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(24024,36,'cp -12','d1e2 e8g8 f1d1 d8c7 a2a4 f8c8 h2h3 a7a6 d4f3 d7b6 c1g5 e7e6 f3d2 f6e8 a1a3 a8b8 d1a1 b6c4 d2c4 c7c4 e2d3 b7b5 a4b5 a6b5 d3c4 b5c4 a1a2 e8c7 a3a4 g7c3 b2c3 d6d5 e4e5 h7h5 f2f3 c7b5 g5d2 g8g7 a4a5 d5d4 c3d4 b5d4 a5a7');
INSERT INTO evaluations VALUES(24578,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(24578,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(24578,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(24578,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24578,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24578,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24578,36,'cp 0','d1e2 d8b6 d4b3 a7a6 b1c3 a8c8 c1e3 b6c6 b3d4 c6c4 e2c4 c8c4 f2f3 h7h5 e1c1 e7e6 h1e1 b7b5 d4b3 h5h4 a2a4 b5a4 b3a5 c4b4 a5c6 b4c4');
INSERT INTO evaluations VALUES(34387,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(34387,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(34387,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(34387,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(34387,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(34387,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(34387,36,'cp 155','c1d2 a5c5 c4d5 c5d5 b1c3 d5d8 e2e4 a7a6 d2g5 h7h6 d1d8 e8d8 g5f4 e7e6 g1f3 d8e8 f3e5 b8d7 e5c4 f8c5 f1e2 g8e7 c4d6 e8f8 e1c1 e7g6 f4g3 f8e7 d1d2 g6e5 h1d1 g7g5 c1b1 b7b5 a2a4 b5a4 c3a4 c5b4 d2d4 a6a5 d6c4 f7f6 a4b6 d7b6');
INSERT INTO evaluations VALUES(38284,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38284,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38284,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38284,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(38284,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(38284,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(38284,36,'cp 29','g8f6 c1f4 c7c5 d4c5 e7c5 e2e3 e8g8 c4d5 f6d5 c3d5 e6d5 a2a3 b8c6 f1d3 c5b6 e1g1 d8f6 b2b4 c8f5 b4b5 c6a5 f4e5 f6e6 d3f5 e6f5 h2h3 f5e4 e5d4 h7h6 a1c1 f8c8 d4b6 a7b6 c1c8 a8c8 f3d4 a5c4 a3a4 c8c5 d1b3');
INSERT INTO evaluations VALUES(39359,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39359,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(39359,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39359,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(39359,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(39359,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39359,36,'cp 0','b4c3 d2c3 d7d6 e2e4 g8f6 f1d3 a7a5 c4c5 b8a6 c5d6 d8d6 e1g1 a6c5 d3c2 d6d1 f1d1 c8g4 c1e3 g4f3 g2f3 c5e6 a2a4 e8e7 b2b4 h8d8 d1d8 e7d8 c2b3 d8e7 g1f1 f6d7 b3d5 c7c6 d5e6 e7e6 f1e2 d7f6 e2d3 f6d7');
INSERT INTO evaluations VALUES(39380,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39380,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39380,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39380,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39380,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(39380,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(39380,36,'cp 8','d2c3 d7d6 e2e4 g8f6 f1d3 a7a5 c4c5 e8g8 c5d6 c7d6 d3c2 b8a6 c1e3 b7b5 e1g1 b5b4 c3b4 a6b4 c2b1 c8b7 a2a3 b4a6 f3d2 a8c8 f2f3 d6d5 e4d5 f6d5 e3f2 d5f4 d2e4 b7e4 d1d8 f8d8 b1e4 f4e2 g1h1 a6c5 e4f5 c5e6 b2b4 a5b4 a3b4 e2c3 a1a7');
INSERT INTO evaluations VALUES(44880,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(44880,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(44880,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44880,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(44880,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(44880,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(44880,36,'cp -17','f6e4 g5f4 d8a5 a2a3 e4c3 b2c3 b8d7 d1d2 a5a4 e2e3 e7e6 f1e2 f8e7 e1g1 e8g8 d2d3 c6c5 c3c4 d5c4 d3c4 a4c4 e2c4 d7b6 c4d3 b6d5 d4c5 d5f4 e3f4 e7c5 a3a4 a7a5 f1d1 g7g6 f3e5 a8a7 d3b5 b7b6 d1d3 h7h6 a1d1 a7c7 g2g3 c8b7');
INSERT INTO evaluations VALUES(49079,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(49079,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(49079,32,'cp -154',NULL);
INSERT INTO evaluations VALUES(49079,33,'cp -162',NULL);
INSERT INTO evaluations VALUES(49079,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(49079,35,'cp -146',NULL);
INSERT INTO evaluations VALUES(49079,36,'cp -147','b8c6 b1c3 e4c3 b2c3 d8d6 e2b5 c8e6 b5b7 a8b8 b7a6 e5e4 f3d4 c6d4 b3a4 e8d8 a6d6 f8d6 c3d4 h7h5 h2h4 f7f6 e1g1 d8e7 f1e1 e7f7 a4b3 a7a5 e1e3 c7c5 d4c5 d6c5 e3e1 h8c8 a2a4 f6f5 c1a3 c5a3 a1a3 f7f6 f2f3 b8b4 c2c3 b4b6 f3e4 d5e4 b3e6 f6e6');
INSERT INTO evaluations VALUES(15389,30,'cp -155',NULL);
INSERT INTO evaluations VALUES(15389,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(15389,32,'cp -159',NULL);
INSERT INTO evaluations VALUES(15389,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(15389,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(15389,35,'cp -157',NULL);
INSERT INTO evaluations VALUES(15389,36,'cp -152','b1c3 e4c3 b2c3 d8d6 e2b5 c8e6 b5b7 a8b8 b7a6 e5e4 f3d4 c6d4 b3a4 e8d8 a6d6 f8d6 c3d4 h7h5 h2h4 d8e7 e1g1 f7f6 c2c3 g7g5 h4g5 h5h4 f1e1 h4h3 g2g3 f6g5 d2d3 e4d3 c1g5 e7f7 a1d1 e6f5 a4c6 b8b2 c6d5 f7g6 g5e3 b2c2 e3d2 f5g4 g1h1 g4d1 d5e4 g6g7');
INSERT INTO evaluations VALUES(52828,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(52828,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52828,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52828,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52828,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52828,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52828,36,'cp 0','c8f5 g2g4 f5g6 h2h4 h7h5 g4g5 b8d7 g1f3 e7e6 d2d3 d8c7 c2c4 g8e7 b1d2 c6c5 c4d5 e6d5 a1c1 e7c6 a2a3 e8c8 d3d4 c8b8 d4c5 f8c5 f1g2 g6d3 g2f1');
INSERT INTO evaluations VALUES(31096,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31096,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31096,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31096,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31096,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31096,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31096,36,'cp 0','g2g4 f5g6 h2h4 h7h5 g4g5 b8d7 g1f3 e7e6 d2d3 g8e7 b1d2 c6c5 a2a3 d8c7 c2c4 e8c8 c4d5 e6d5 d3d4 e7c6 a1c1 c8b8 d4c5 f8c5 f1g2 g6d3 g2f1');
INSERT INTO evaluations VALUES(54005,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(54005,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(54005,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(54005,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(54005,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(54005,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(54005,36,'cp 32','g1f3 g8f6 d2d4 c5d4 f3d4 a7a6 f1e2 e6e5 d4f5 c8f5 e4f5 b8c6 e1g1 h7h6 e2f3 f8e7 c3d5 e8g8 c2c3 e5f4 c1f4 f8e8 f1f2 f6d5 f3d5 e7f6 d1h5 e8e7 g1f1 c6e5 f2e2 d8c7 a1d1 a8e8 e2e3');
INSERT INTO evaluations VALUES(56391,30,'cp -370',NULL);
INSERT INTO evaluations VALUES(56391,31,'cp -364',NULL);
INSERT INTO evaluations VALUES(56391,32,'cp -382',NULL);
INSERT INTO evaluations VALUES(56391,33,'cp -396',NULL);
INSERT INTO evaluations VALUES(56391,34,'cp -387',NULL);
INSERT INTO evaluations VALUES(56391,35,'cp -378',NULL);
INSERT INTO evaluations VALUES(56391,36,'cp -395','d5e4 d3e4 f7f5 e4d3 e5e4 d3b5 c7c6 b5c4 e4f3 d1f3 f8d6 e1g1 d8h4 f3e2 e8d8 g2g3 h4d4 c2c3 d4g4 d2d4 g8f6 e2g4 f5g4 c4d3 c8e6 c3c4 b8d7 f1e1 h8e8 b1c3 e6f7 c1d2 d6b4 c4c5 a8c8 a1d1 h7h6 d2f4 e8e1 d1e1 f6d5 f4d6 d7f6 d3f5 c8a8 h2h3 b4c3 b2c3');
INSERT INTO evaluations VALUES(56619,30,'cp -385',NULL);
INSERT INTO evaluations VALUES(56619,31,'cp -384',NULL);
INSERT INTO evaluations VALUES(56619,32,'cp -371',NULL);
INSERT INTO evaluations VALUES(56619,33,'cp -371',NULL);
INSERT INTO evaluations VALUES(56619,34,'cp -379',NULL);
INSERT INTO evaluations VALUES(56619,35,'cp -384',NULL);
INSERT INTO evaluations VALUES(56619,36,'cp -370','d3e4 f7f5 e4d3 e5e4 d3b5 c7c6 b5c4 e4f3 d1f3 f8d6 d2d3 b8d7 e1g1 d8h4 g2g3 d7e5 f3g2 h4f6 f1e1 g8e7 c4b3 c8d7 f2f4 e5g4 c1d2 d6c5 g1f1 f6h6 h2h3 g4f6 d2a5 f6d5 b1d2 d7e6 d2f3 e8g8 c2c4 d5e3 e1e3 c5e3 a1e1 e3b6 a5b6 a7b6 f1g1 g8h8');
INSERT INTO evaluations VALUES(56603,30,'cp -391',NULL);
INSERT INTO evaluations VALUES(56603,31,'cp -392',NULL);
INSERT INTO evaluations VALUES(56603,32,'cp -376',NULL);
INSERT INTO evaluations VALUES(56603,33,'cp -400',NULL);
INSERT INTO evaluations VALUES(56603,34,'cp -396',NULL);
INSERT INTO evaluations VALUES(56603,35,'cp -390',NULL);
INSERT INTO evaluations VALUES(56603,36,'cp -389','f7f5 e4d3 e5e4 d3b5 c7c6 b5c4 e4f3 e1g1 f3g2 f1e1 e8d7 d2d4 d7c7 b1a3 g8f6 d1f3 a7a6 c4b3 f6e4 f3g2 b7b5 c1f4 c7b7 a1d1 d8h4 f4g3 h4f6 c2c4 f8b4 e1e3 e4g3 h2g3 c8d7 c4b5 a6b5 a3b5 h8e8 e3e8 d7e8 b5c3 b7a7 d4d5 b4c3 b2c3');
INSERT INTO evaluations VALUES(4012,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(4012,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(4012,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(4012,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(4012,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(4012,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(4012,36,'cp 37','g5h4 h7h6 f1d3 b7b6 e1g1 a7a6 d4d5 d7c5 d3c2 c8g4 h2h3 g4f3 d1f3 f6d7 e3e4 b6b5 c4b5 a6b5 a2a3 c7c6 f3e2 c6d5 e4d5 d7f6 b2b4 c5a4 c2a4 b5a4 a1c1 d8d7 f1e1 f8e8 e2d3 e8c8 e1d1 d7c7 b4b5 f6d7');
INSERT INTO evaluations VALUES(6626,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(6626,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(6626,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6626,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(6626,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(6626,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(6626,36,'cp 25','c3d5 e5e4 f3d4 c6d4 d5b4 c7c6 e2e3 d8e7 e3d4 e7b4 d1g4 g8e7 g4g7 h8g8 g7e5 d7d5 b2b3 c8e6 c1b2 e7g6 e5f6 d5c4 b2c3 b4d6 f1c4 g6f4 f2f3 f4g2 e1d1 e8d7 c4e6 d6e6 f6e6 d7e6 f3e4 g8g4 d4d5 c6d5 e4d5 e6d5 h1f1 g4f4 d1c2 a8c8 h2h3 b7b6 f1g1 g2h4 g1g7 h4f3 g7h7');
INSERT INTO evaluations VALUES(19063,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(19063,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(19063,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(19063,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(19063,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(19063,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(19063,36,'cp 77','d4e5 d7d6 f1b5 d6e5 d1d8 e8d8 g1f3 f8d6 b1c3 g8f6 c1g5 c8e6 e1g1 d8e7 c3d5 e6d5 e4d5 c6b4 f3h4 h8d8 h4f5 e7f8 f1d1 b4c2 g5f6 g7f6 a1c1 c2d4 f5d4 e5d4 g2g3 a8c8 b5d3 c8c1 d1c1 f8g7 f2f4 d6b4 g1f2 d8d5 c1c7');
INSERT INTO evaluations VALUES(20105,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20105,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20105,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20105,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20105,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20105,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20105,36,'cp 0','g1f3 g7g6 f1e2 f8g7 e1g1 c8d7 a4b3 g8f6 b1c3 e8g8 f1d1 d8c7 h2h3 a8c8 c1e3 a7a6 a2a4 d7e6 b3a3 c6a5 a3b4 a5c4 e3d4 c4a5');
INSERT INTO evaluations VALUES(20104,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20104,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20104,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20104,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20104,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20104,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(20104,36,'cp -25','g8f6 f1e2 e7e6 e1g1 d8c7 c1f4 c8d7 a4c4 a8c8 b1c3 a7a6 a2a4 e6e5 f4g5 d7e6 c4d3 f8e7 a1d1 c6b4 d3d2 e8g8 g5f6 e7f6 d2d6 b4c2 d6c7 c8c7 c3d5 c7d7 d5b6 d7d1 f1d1 f8d8 d1c1 c2d4 f3d4 e5d4 e2d3 f6e5 c1d1 g7g6 f2f3 f7f5 e4f5');
INSERT INTO evaluations VALUES(13404,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13404,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13404,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13404,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13404,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13404,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13404,36,'cp 0','f1e2 g7g6 e1g1 c8d7 b1c3 f8g7 a4b3 e8g8 f1d1 a8c8 h2h3 d8c7 a2a4 a7a6 c1e3 d7e6 b3a3 c6a5 f3d4 e6d7 d4b3 d7e6');
INSERT INTO evaluations VALUES(30030,30,'cp -196',NULL);
INSERT INTO evaluations VALUES(30030,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(30030,32,'cp -143',NULL);
INSERT INTO evaluations VALUES(30030,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(30030,34,'cp -133',NULL);
INSERT INTO evaluations VALUES(30030,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(30030,36,'cp -128','b8c6 g1f3 f6e4 b1c3 d8f6 c1b2 e8c8 d1c1 e4c5 f1e2 c6b4 e1g1 c5d3 c1b1 e6f5 c3d1 d3b2 b1b2 f6b2 d1b2 g7g6 a2a3 b4c6 b2a4 f8g7 a1e1 c6d4 f3d4 d8d4 d2d3 a7a5 e1d1 h8d8 a4c3 c7c6 c3e4 g7f8 a3a4 f8e7 h2h3 c8c7 f1e1 e7b4 e1f1 h7h5 g2g4');
INSERT INTO evaluations VALUES(25117,30,'cp -112',NULL);
INSERT INTO evaluations VALUES(25117,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(25117,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(25117,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(25117,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(25117,35,'cp -158',NULL);
INSERT INTO evaluations VALUES(25117,36,'cp -148','g1f3 f6e4 b1c3 d8f6 c1b2 e8c8 d1c1 e4c5 f1e2 c6b4 e1g1 c5d3 c1b1 e6f5 c3d1 d3b2 b1b2 f6b2 d1b2 g7g6 a2a3 b4c6 b2a4 f8g7 a1a2 h8e8 a4c3 c6a5 b3b4 a5b3 c4c5 b3d2 f3d2 g7c3 d2c4 f5e6 a2c2 c3g7 c4a5 e6f5 c2a2 c8b8 e2c4 e8e7 a2e2 e7e2 c4e2 d8d2 f1e1');
INSERT INTO evaluations VALUES(38130,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38130,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38130,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(38130,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(38130,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(38130,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(38130,36,'cp 22','e2e4 g8f6 f1b5 c8d7 b5d7 d8d7 e4d5 f6d5 c3d5 e6d5 e1g1 e8g8 d1d3 b8c6 f1e1 f8e8 c1d2 h7h6 a2a3 a7a6 b2b4 c5d6 d2e3 d7c7 d3d5 c6e5 f3e5 d6e5 a1d1 c7c2 e3c5 e5c3 e1e8 a8e8 g2g3 c3f6 c5b6 c2c3 d5b7 c3a3');
INSERT INTO evaluations VALUES(38120,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(38120,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(38120,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(38120,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(38120,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(38120,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38120,36,'cp 16','g8f6 f1b5 c8d7 b5d7 d8d7 e4d5 f6d5 c3d5 e6d5 e1g1 e8g8 d1d3 b8c6 a2a3 f8e8 c1d2 h7h6 f1e1 a8d8 h2h3 e8e1 a1e1 a7a6 b2b4 c5d6 g1f1 c6e7 c2c3 d7c7 g2g3 c7c4 d3c4 d5c4 d2f4 e7d5 f4d6 d8d6 e1e8 g8h7 f3e5 d5c3 e5c4 d6d4 c4e5');
INSERT INTO evaluations VALUES(40609,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40609,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40609,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40609,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(40609,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40609,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40609,36,'cp 24','f1b5 c8d7 b5d7 d8d7 e4d5 f6d5 c3d5 e6d5 e1g1 e8g8 d1d3 b8c6 f1e1 c6b4 d3b3 f8e8 c1d2 b4c6 b3d3 h7h6 a2a3 a7a6 b2b4 e8e1 a1e1 c5d6 d2c3 a8d8 g2g3 d7g4 g1g2 g4c4 e1d1 c4d3 d1d3 d6e7 d3d2 f7f6 c3d4 c6d4 f3d4 g8f7');
INSERT INTO evaluations VALUES(47170,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(47170,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(47170,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(47170,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(47170,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(47170,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(47170,36,'cp -25','c7c5 d4c5 e7e6 e2e3 f8c5 a2a3 e8g8 c2c4 d5c4 d1d8 f8d8 f1c4 b7b6 b1d2 c8b7 g1f3 b8d7 e1g1 c5e7 a1c1 a8c8 c4e2 a7a6 f1d1 f6d5 b3b4 a6a5 b4a5 b6a5 d2b3 a5a4 b3a5 b7a8 a5c4 d5b6 f3e5 a8d5 e5d7');
INSERT INTO evaluations VALUES(51652,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(51652,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(51652,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(51652,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(51652,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(51652,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(51652,36,'cp -25','e7e6 a2a4 f8b4 f1e2 c8b7 e1g1 a7a6 c3a2 b4e7 e4e5 c6c5 c1e3 g8h6 d4c5 e8g8 a4b5 a6b5 f3d4 d8c8 a2c3 a8a1 d1a1 b8d7 c3b5 e7c5 b5d6 c5d6 e5d6 e6e5 d4c2 h6f5 a1a5 f5d6 a5b4 c8c7 f1c1 d6f5');
INSERT INTO evaluations VALUES(56925,30,'cp -187',NULL);
INSERT INTO evaluations VALUES(56925,31,'cp -200',NULL);
INSERT INTO evaluations VALUES(56925,32,'cp -199',NULL);
INSERT INTO evaluations VALUES(56925,33,'cp -183',NULL);
INSERT INTO evaluations VALUES(56925,34,'cp -183',NULL);
INSERT INTO evaluations VALUES(56925,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(56925,36,'cp -189','h7h6 g5f7 e8f7 d2d4 g8f6 c1f4 d7d5 b1c3 f7g7 c3d5 f8d6 f4d6 d8d6 d5f6 d6g3 e1e2 g7f6 d1e1 g3e1 a1e1 f6g7 e2d2 h8d8 c2c3 c8e6 f1e2 b8d7 h1f1 d7f6 e2d3 c7c6 a2a4 a7a5 b2b3 d8d7 e4e5 f6d5 d3f5 e6f5 f1f5 a8f8 f5f8 g7f8 g2g3 d7f7 e5e6 f7e7 c3c4');
INSERT INTO evaluations VALUES(56941,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(56941,31,'cp -193',NULL);
INSERT INTO evaluations VALUES(56941,32,'cp -184',NULL);
INSERT INTO evaluations VALUES(56941,33,'cp -185',NULL);
INSERT INTO evaluations VALUES(56941,34,'cp -181',NULL);
INSERT INTO evaluations VALUES(56941,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(56941,36,'cp -181','g5f7 e8f7 d2d4 d7d5 c1f4 g8f6 b1c3 f7g7 c3d5 f8d6 f4d6 d8d6 d5f6 d6g3 e1e2 g7f6 d1e1 g3e1 a1e1 f6g7 e2d2 h8d8 c2c3 c7c5 d2e3 b8c6 d4d5 c6e5 b2b4 b7b6 b4c5 b6c5 e1b1 c8d7 f1b5 d8b8 b5d7 e5d7 h1e1 g7f6 e3f4 d7e5 b1b8 a8b8 e1f1 b8b2 f4g3 f6e7 f1f5 e5d7 f5h5');
INSERT INTO evaluations VALUES(56853,30,'cp -184',NULL);
INSERT INTO evaluations VALUES(56853,31,'cp -196',NULL);
INSERT INTO evaluations VALUES(56853,32,'cp -194',NULL);
INSERT INTO evaluations VALUES(56853,33,'cp -202',NULL);
INSERT INTO evaluations VALUES(56853,34,'cp -169',NULL);
INSERT INTO evaluations VALUES(56853,35,'cp -169',NULL);
INSERT INTO evaluations VALUES(56853,36,'cp -188','e8f7 d2d4 d7d5 c1f4 g8f6 b1c3 f7g7 c3d5 f8d6 f4d6 d8d6 d5f6 d6g3 e1e2 g7f6 d1d3 g3d3 e2d3 f6g7 d3e3 h8e8 f1d3 c7c5 d4d5 b8d7 d3b5 e8e5 e3f4 e5e7 h1e1 d7e5 a1d1 e5g6 f4g3 e7e5 a2a3 a7a6 b5d3 h6h5 b2b4 c5b4 a3b4 c8d7 e1f1 a6a5 b4a5 a8a5');
INSERT INTO evaluations VALUES(35534,30,'cp -229',NULL);
INSERT INTO evaluations VALUES(35534,31,'cp -207',NULL);
INSERT INTO evaluations VALUES(35534,32,'cp -210',NULL);
INSERT INTO evaluations VALUES(35534,33,'cp -231',NULL);
INSERT INTO evaluations VALUES(35534,34,'cp -243',NULL);
INSERT INTO evaluations VALUES(35534,35,'cp -236',NULL);
INSERT INTO evaluations VALUES(35534,36,'cp -260','b1c3 g8f6 e4e5 d7d5 e5f6 d8f6 d2d4 f8d6 f1e2 f4f3 g2f3 g4f3 e2f3 h8e8 e1f2 c7c6 c1h6 f6h6 f3h5 f7e7 h5e8 e7e8 d1e2 e8d8 a1g1 b8d7 e2e3 d6f4 e3f3 d7f6 h4h5 f4g5 g1g2 d8c7 f3e2 c8g4 g2g4 a8e8 e2f3 f6g4 f3g4 g5f6 h1f1 h6d2 f2g3 f6d4');
INSERT INTO evaluations VALUES(57577,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(57577,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(57577,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(57577,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(57577,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(57577,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(57577,36,'cp 33','e7e5 g1f3 d7d6 b1c3 f8e7 f1g2 g8f6 e1g1 e8g8 b2b4 g8h8 d2d3 a7a6 a1b1 c7c6 a2a4 c8e6 b4b5 a6b5 a4b5 d8c7 b5c6 b7c6 g2h3 b8a6 e2e4 f5e4 h3e6 e4f3 d1f3 a8d8 f3e3 d6d5 e6h3 d5d4');
INSERT INTO evaluations VALUES(2358,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(2358,31,'cp 277',NULL);
INSERT INTO evaluations VALUES(2358,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(2358,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(2358,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(2358,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(2358,36,'cp 248','e5c4 d6b4 c2c3 b4e7 b1a3 c8d7 f2f3 a7a6 a3c2 d8c6 c1f4 e8c8 c2d4 g7g5 d4c6 d7c6 f4e3 e7f6 f1e2 h7h5 e1f2 c6d7 h2h3 f6e7 c4e5 d7e6 e2c4 e6c4 e5c4 g8f6 f2e2 f6d7 a1d1 b7b6 e3d4 h8e8 c4e3 e7c5 d4c5 d7c5 d1d8 e8d8');
INSERT INTO evaluations VALUES(2360,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(2360,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(2360,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(2360,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(2360,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(2360,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(2360,36,'cp 251','d6b4 c1d2 b4d2 b1d2 c8e6 e1c1 g8f6 c4e5 f6d7 e5d7 e6d7 d2b3 d8e6 f1c4 e8e7 c4e6 d7e6 b3c5 b7b6 c5e6 e7e6 d1d2 h8d8 d2d8 a8d8 h1d1 d8a8 f2f3 h7h5 d1d5 g7g6 c1d2 c7c6 d5d3 h5h4 d2e3 a8h8 b2b4 h8h5');
INSERT INTO evaluations VALUES(6662,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(6662,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(6662,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(6662,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(6662,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(6662,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(6662,36,'cp 30','b1c3 d7d5 e4e5 g8e7 g1f3 b4a5 f1e2 a5b6 c1e3 e8g8 e1g1 h7h6 c3a4 e7f5 a4b6 a7b6 e3f4 d8e7 g2g4 f5h4 f3h4 e7h4 f4g3 h4e7 f2f4 f7f5 e5f6 e7f6 e2f3 c6d4 f3g2 c7c5 f4f5 c8d7 g2d5 g8h8 d5b7 a8e8 f1f2 e8e3 d1d2 f6e7');
INSERT INTO evaluations VALUES(6658,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(6658,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(6658,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(6658,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(6658,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(6658,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(6658,36,'cp 36','d7d5 e4e5 g8e7 f1e2 b4a5 g1f3 a5b6 e1g1 c8g4 c1e3 e7f5 d1d3 e8g8 g1h1 c6b4 d3d2 b4c6 a1e1 h7h6 h2h3 g4f3 e2f3 c6d4 e3d4 f5d4 f3d5 c7c6 d5c4 d4e6 d2c1 d8d4 e1e4 d4d7 f2f4 e6d4 c3a4 d4f5 a4b6 a7b6 h1g1');
INSERT INTO evaluations VALUES(25437,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(25437,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(25437,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(25437,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(25437,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(25437,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(25437,36,'cp 25','e1g1 d8d7 a2a3 h7h6 b2b4 g7g5 c1e3 f8g7 c3e2 c6e7 a3a4 e7g6 c2c4 a7a6 b4b5 e8g8 h2h3 g5g4 h3g4 f6g4 f3h2 g4e3 f2e3 d7e7 f1f8 a8f8 d1e1 e7g5 e1g3 g5h5 e2c3 g7f6 h2g4 f6g5 a1f1 f8f1 g1f1 h5h1 f1e2 h1g1 g3f3 a6b5 c4b5');
INSERT INTO evaluations VALUES(25436,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(25436,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(25436,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(25436,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(25436,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25436,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(25436,36,'cp 17','f8e7 c3e2 e8g8 e2g3 d6d5 c2c3 a7a5 d1e2 e7d6 d3d4 d5e4 f3e5 c6e5 d4e5 d6e5 g3e4 f6e4 e2e4 d8d5 e4e2 a8d8 f1e1 e5f6 g2g3 a5a4 a2a3 d5c6 c1f4 e6e5 f4e3 g8h8 a1d1 h7h6 h2h4 c6e4 d1d2 d8d2 e3d2 e4e2 e1e2 f8d8 g1g2 d8d3');
INSERT INTO evaluations VALUES(31146,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(31146,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(31146,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31146,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(31146,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(31146,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(31146,36,'cp 0','e4g6 e5e6 f7e6 h2h4 h7h5 f1d3 g6d3 d1d3 d8d6 g1e2 b8d7 d3g6 e8d8 c1g5 h5g4 e2f4 g8f6 g6f7 e6e5 f4e6 d8c8 g5f6 d7f6 b1d2 c8d7 e6c5 d7c7 c5e6');
INSERT INTO evaluations VALUES(31029,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31029,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31029,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31029,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31029,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31029,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31029,36,'cp 0','e5e6 f7e6 h2h4 h7h5 f1d3 g6d3 d1d3 d8d6 g1e2 b8d7 d3g6 e8d8 c1g5 h5g4 e2f4 g8f6 g6f7 e6e5 f4e6 d8c8 g5f6 d7f6 b1d2 c8d7 e6c5 d7c7 c5e6');
INSERT INTO evaluations VALUES(52475,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(52475,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(52475,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52475,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52475,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52475,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52475,36,'cp 0','d7d5 e2e4 e7e6 e4d5 e6d5 d3d4 b8c6 f1b5 g8f6 c1g5 f8e7 d4c5 e7c5 g5f6 d8f6 d1d5 e8g8 e1g1 f6b2 d5c5 b2a1 b5d3 a1a2 d3h7 g8h7 c5f8 f7f6 f8e8 a2c2 e8h5 h7g8 h5e8');
INSERT INTO evaluations VALUES(55931,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(55931,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55931,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55931,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(55931,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(55931,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(55931,36,'cp 11','g1f3 g8f6 c1b2 f8d6 d2d4 b7b6 f1d3 c8b7 e1g1 e8g8 c2c4 b8d7 b1d2 c7c5 c4d5 f6d5 d2e4 d6e7 d4c5 d7c5 e4c5 e7c5 d1e2 a8c8 d3a6 d8e7 a2a3 c5d6 a6b7 e7b7 h2h3 d6e7 f1c1 c8c1 a1c1 f8c8 c1c4 b7a6 c4c8 a6c8 e2d3');
INSERT INTO evaluations VALUES(237,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(237,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(237,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(237,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(237,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(237,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(237,36,'cp 0','d7c6 b1c3 c8e6 c1e3 c5a3 e3c1');
INSERT INTO evaluations VALUES(203,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(203,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(203,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(203,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(203,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(203,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(203,36,'cp 0','b1c3 c8e6 c1e3 c5a3 e3c1');
INSERT INTO evaluations VALUES(199,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(199,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(199,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(199,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(199,36,'cp 0','c8e6 c1e3 c5a3 e3c1');
INSERT INTO evaluations VALUES(27315,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27315,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27315,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27315,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27315,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27315,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27315,36,'cp 0','c1e3 c5a3 e3c1 a3c5');
INSERT INTO evaluations VALUES(4036,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(4036,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4036,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4036,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(4036,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4036,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(4036,36,'cp 19','e3e4 e7e5 d4d5 d7c5 d3c2 a7a5 e1g1 d8e7 c1g5 c8d7 a1b1 e7e8 b2b3 f6h5 a2a3 f7f5 b3b4 c5e4 c3e4 f5e4 c2e4 a5b4 a3b4 h7h6 g5e3 h5f6 f3d2 f6e4 d2e4 d7f5 d1d3 g6g5 f2f3 e8g6 c4c5 g5g4 c5d6 c7d6');
INSERT INTO evaluations VALUES(10292,30,'cp 383',NULL);
INSERT INTO evaluations VALUES(10292,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(10292,32,'cp 375',NULL);
INSERT INTO evaluations VALUES(10292,33,'cp 392',NULL);
INSERT INTO evaluations VALUES(10292,34,'cp 349',NULL);
INSERT INTO evaluations VALUES(10292,35,'cp 392',NULL);
INSERT INTO evaluations VALUES(10292,36,'cp 369','c3d4 c5b6 e4e5 f6g4 c1g5 f7f6 e5f6 g7f6 f1e1 e8f8 g5f4 d7d5 c4b3 c6e7 h2h3 e7g6 f4c1 h7h5 b3c2 f6f5 b1c3 f8g7 c3e2 g4f6 e2f4 f6e4 f4g6 g7g6 a1b1 c7c6 c1f4 b6c7 f3e5 c7e5 f4e5 e4c3 d1f3 c3b1 e5h8 d8h8 f3g3 g6f7 g3c7 f7g6 c2b1 h5h4 b1c2 h8d4 e1e7');
INSERT INTO evaluations VALUES(10288,30,'cp 372',NULL);
INSERT INTO evaluations VALUES(10288,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(10288,32,'cp 368',NULL);
INSERT INTO evaluations VALUES(10288,33,'cp 377',NULL);
INSERT INTO evaluations VALUES(10288,34,'cp 377',NULL);
INSERT INTO evaluations VALUES(10288,35,'cp 370',NULL);
INSERT INTO evaluations VALUES(10288,36,'cp 364','c5b6 e4e5 f6g4 c1g5 f7f6 e5f6 g7f6 f1e1 e8f8 g5f4 d7d5 c4b3 b6a5 h2h3 a5e1 d1e1 c6e7 h3g4 c8g4 b1c3 c7c6 f3h4 e7g6 f2f3 g6f4 f3g4 h8g8 e1g3 f4e6 h4f5 g8g5 c3e2 d8c7 g3h4 e6f4 e2g3 a8e8 h4h6 f8g8 h6f6 g5g6 f5h6 g6h6 f6h6 e8e6 h6h4 f4g6 h4h2 g6f4 a1f1 f4e2 g3e2 c7h2 g1h2 e6e2 f1d1 e2e4 d1d2 b7b5 b3d1 h7h6 g4g5 h6g5 d1f3');
INSERT INTO evaluations VALUES(9937,30,'cp 394',NULL);
INSERT INTO evaluations VALUES(9937,31,'cp 411',NULL);
INSERT INTO evaluations VALUES(9937,32,'cp 391',NULL);
INSERT INTO evaluations VALUES(9937,33,'cp 408',NULL);
INSERT INTO evaluations VALUES(9937,34,'cp 381',NULL);
INSERT INTO evaluations VALUES(9937,35,'cp 369',NULL);
INSERT INTO evaluations VALUES(9937,36,'cp 406','e4e5 f6g4 c1g5 f7f6 g5c1 c6d4 f3d4 g4e5 c4b3 d7d5 b1c3 c7c6 c1a3 e8f7 c3e4 h8e8 e4d6 f7g8 d6e8 d8e8 g1h1 e8g6 d1d2 g6d3 d2d3 e5d3 a1d1 d3e5 h2h3 c8d7 f1e1 e5g6 b3c2 g6e5 a3c1 b6a5 e1e3 c6c5 d4f3 e5f3 e3f3 d7c6');
INSERT INTO evaluations VALUES(9914,30,'cp 413',NULL);
INSERT INTO evaluations VALUES(9914,31,'cp 404',NULL);
INSERT INTO evaluations VALUES(9914,32,'cp 402',NULL);
INSERT INTO evaluations VALUES(9914,33,'cp 405',NULL);
INSERT INTO evaluations VALUES(9914,34,'cp 391',NULL);
INSERT INTO evaluations VALUES(9914,35,'cp 386',NULL);
INSERT INTO evaluations VALUES(9914,36,'cp 425','f6g4 c1g5 f7f6 g5c1 c6d4 f3d4 g4e5 c4b3 d7d5 b1c3 e8g8 c3d5 g8h8 d5b6 a7b6 c1e3 c7c5 d4e6 c8e6 b3e6 f6f5 d1d8 a8d8 g2g3 d8d6 e6b3 c5c4 b3c2 b6b5 a1b1 e5f3 g1g2 f3d2 e3c5 d6d8 c5f8 d8f8 b1b5 d2f1 g2f1 f5f4 b5b7 f4g3 h2g3');
INSERT INTO evaluations VALUES(10466,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(10466,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(10466,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(10466,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(10466,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(10466,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(10466,36,'cp -6','f1b5 e8g8 d2d3 c6a5 c3c4 c7c6 b5a4 d7d5 e4d5 e5e4 d3e4 a5c4 a4b3 c6d5 b3c4 d5c4 d1d8 f8d8 c1g5 b7b5 e4e5 h7h6 g5h4 d8e8 h4f6 g7f6 e1c1 c8b7 e5f6 b7f3 g2f3 e8e6 d1d5 e6f6 d5b5 f6f3 c1b2 f3f2 b5b7 g8f8 h1d1 a8e8 d1d4 f2h2 d4d7 e8e2 d7f7 f8g8 f7g7 g8h8 g7h7 h8g8');
INSERT INTO evaluations VALUES(10009,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(10009,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(10009,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(10009,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(10009,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(10009,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(10009,36,'cp -23','e8g8 b5c6 b7c6 d2d3 d7d5 d1e2 d5e4 d3e4 a7a5 e1g1 d8e7 a3a4 f8d8 f1e1 f6d7 f3d2 d7b6 d2b3 c8e6 f2f3 h7h6 c1e3 b6c4 b3c5 a8b8 a1b1 b8b6 e2f2 d8b8 b1c1 f7f6 c5d3 c4e3 f2e3 b8d8 d3c5 b6b2 e3f2');
INSERT INTO evaluations VALUES(12154,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(12154,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(12154,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(12154,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(12154,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(12154,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(12154,36,'cp 55','e2e3 f8e7 f1d3 d5c4 d3c4 b7b5 c4d3 c8b7 e1g1 a7a6 a2a4 e8g8 f3e5 d7e5 f4e5 c6c5 a4b5 c5d4 e5d4 a6b5 d3b5 d8d6 f2f4 a8a1 d1a1 f8a8 a1c1 d6b4 c1c2 e7c5 f1d1 h7h6 h2h3 c5d4 d1d4 b4e7 b5f1 e6e5 f4e5 e7e5');
INSERT INTO evaluations VALUES(16563,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16563,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(16563,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16563,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(16563,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(16563,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(16563,36,'cp 0','c2c4 f8b4 b1c3 d7d6 f1e2 e8g8 a2a3 b4c3 b2c3 c6e7 h2h4 c7c6 g2g4 d8a5 g4g5 f6d7 h4h5 f7f5 g5f6 f8f6 h5h6 a5c3 e3d2 c3b2 a1b1 b2a3 d2b4 a3a2 b1a1 a2b2 a1b1 b2a2');
INSERT INTO evaluations VALUES(19737,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19737,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19737,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(19737,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19737,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(19737,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19737,36,'cp 13','c8d7 g1f3 e7e6 e1g1 g8f6 c2c4 d5c4 b3c4 f8d6 b1c3 e8g8 d1e2 a7a6 b5a4 d8c7 h2h3 a8d8 d2d3 b7b6 a1d1 c6e7 b2a1 d7a4 c3a4 c7c6 a1f6 c6a4 f6a1 b6b5 e2b2 f7f6 c4b5 a4b5 d3d4 b5b2 a1b2');
INSERT INTO evaluations VALUES(26020,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(26020,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26020,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(26020,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26020,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26020,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(26020,36,'cp 15','g1f3 g8f6 e1g1 e7e6 c2c4 d5c4 b3c4 f8d6 d1e2 a7a6 b5a4 e8g8 b1c3 d8c7 h2h3 f8d8 a1d1 c6e7 d1b1 e7g6 b2a1 a8b8 a4d7 d8d7 c3a4 b7b6 d2d4 c7c6 d4c5 b6c5 b1b8 d6b8 e2c2 d7d8 a1f6 g7f6 a4c3 g6e5 f3e5 b8e5 f2f4 e5c3 c2c3 f6f5 c3f6');
INSERT INTO evaluations VALUES(26021,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(26021,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(26021,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(26021,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(26021,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(26021,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26021,36,'cp 4','e7e6 c2c4 d5c4 b3c4 g8f6 b1c3 f8d6 e1g1 a7a6 b5a4 e8g8 d1e2 d8c7 d2d3 f8d8 h2h3 a8b8 a1b1 h7h6 b2a1 c6e5 f3e5 d6e5 a4d7 c7d7 f1d1 b7b6 a2a4 f6e8 a1b2 e8d6 e2c2 d6f5 c2b3 d7e7 c3e4 e5c7 f2f4');
INSERT INTO evaluations VALUES(20497,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(20497,31,'cp -168',NULL);
INSERT INTO evaluations VALUES(20497,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(20497,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(20497,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(20497,35,'cp -166',NULL);
INSERT INTO evaluations VALUES(20497,36,'cp -178','c4b5 f8d6 e1g1 g8e7 b1c3 e8g8 d2d4 c5d4 f3d4 d6c5 b5c6 b7c6 c1e3 c5b6 c3a4 c6c5 d4f3 d5d4 e3f2 e7d5 c2c3 d4c3 a4c3 d5f4 d1a4 f4d3 a1d1 c8f5 f3e5 f8e8 f2g3 e8e5 g3e5 c5c4 g1h1 f5d7 d1d3 c4d3 a4e4 f7f6 e5f4 d7e8 c3d5 e8f7 d5e7 g8h8 f1d1 d8f8 f4d6 a8d8');
INSERT INTO evaluations VALUES(20469,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(20469,31,'cp -155',NULL);
INSERT INTO evaluations VALUES(20469,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(20469,33,'cp -168',NULL);
INSERT INTO evaluations VALUES(20469,34,'cp -164',NULL);
INSERT INTO evaluations VALUES(20469,35,'cp -158',NULL);
INSERT INTO evaluations VALUES(20469,36,'cp -156','f8d6 e1g1 g8e7 d2d4 d8b6 b5d3 c5d4 g1h1 e8g8 c2c3 c8f5 c3d4 c6d4 c1e3 d6c5 e3d4 c5d4 b1c3 d4c3 b2c3 b6f6 d3f5 e7f5 f1e1 f8d8 d1b3 b7b6 a1d1 f5e7 e1e5 g7g6 f3d4 a8c8 d1e1 e7c6 d4c6 f6c6 f4f5 g6f5 e5f5 c6c3 b3c3 c8c3');
INSERT INTO evaluations VALUES(22968,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(22968,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(22968,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(22968,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(22968,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(22968,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(22968,36,'cp -70','f8e7 c2c4 g8f6 e4c2 d7d5 g1f3 e8g8 a2a3 f8e8 f1e2 f6e4 b1c3 e4c3 b2c3 c8g4 c4d5 d8d5 e1g1 a8d8 f1d1 d5f5 c2f5 d8d1 e2d1 g4f5 f3d4 f5d7 a3a4 c6a5 d4b3 a5b3 d1b3 b7b6 g1f1 h7h6 a4a5 b6b5 c1d2 e7d6');
INSERT INTO evaluations VALUES(6997,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(6997,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(6997,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(6997,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(6997,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(6997,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(6997,36,'cp -59','c2c4 g8f6 e4c2 d7d5 g1f3 e8g8 a2a3 f8e8 f1e2 f6e4 b1c3 e4c3 b2c3 c8g4 c4d5 d8d5 e1g1 g4f5 c2a2 d5c5 c1b2 f5e4 f3d2 e4g6 c3c4 a8d8 d2f3 g6e4 f1d1 c5h5 d1d8 e8d8 a2b3 h5g6 a1d1 b7b6 b3a4 d8d1 a4d1 h7h6 h2h3 e7f8 f3h4 g6d6 d1d6 f8d6 h4f3');
INSERT INTO evaluations VALUES(40989,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(40989,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(40989,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(40989,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(40989,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(40989,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40989,36,'cp 9','e8g8 c2c3 c7c5 f3e5 c5c4 f1e2 b7b5 a2a4 b5b4 e2f3 c8a6 e1g1 d8c8 a4a5 a6b7 f1e1 b8d7 e3e4 d7e5 f4e5 f8d8 d1e2 b7c6 e4d5 c6d5 e5f6 e7f6 f3d5 e6d5 d2f1 c8c6 g2g3 a8b8 e2f3 g7g6 f1e3 b4c3 b2c3 b8b3 e3g4 g8g7 g1g2 h7h5');
INSERT INTO evaluations VALUES(36353,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36353,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36353,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36353,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36353,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36353,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36353,36,'cp 24','f3e5 c7c5 c2c3 c5c4 f1e2 b7b5 a2a4 b5b4 e2f3 c8a6 e1g1 d8c8 a4a5 b8c6 e3e4 c6e5 f4e5 f8d8 f1e1 a6b5 e5f6 e7f6 e4d5 e6d5 d2f1 b5c6 g2g3 g7g6 f1e3 f6g7 f3g2 a8b8 h2h4 b4c3 b2c3 b8b3 d1d2');
INSERT INTO evaluations VALUES(43644,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(43644,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(43644,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43644,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43644,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43644,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(43644,36,'cp 3','g1f3 b8d7 d1c2 d5c4 a2a4 c7c5 f1c4 d8c7 c2e2 f8d6 h2h3 e8g8 e1g1 c5d4 e3d4 b7b6 d4d5 e6d5 c3d5 f6d5 c4d5 c8b7 d5b7 c7b7 f1d1 d6c5 e2d3 d7f6 c1e3 c5e3 d3e3 f8e8 e3b3 a8d8 d1d8 e8d8 a1d1 d8d1 b3d1 h7h6 d1d8 g8h7 d8d3 h7g8 d3c4 b6b5 c4d4 b7c6 f3e5 c6e6 a4b5');
INSERT INTO evaluations VALUES(45629,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(45629,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(45629,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(45629,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45629,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(45629,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(45629,36,'cp 28','e2e3 f8d6 f4d6 d8d6 b1c3 e8g8 a1c1 d6e7 d1c2 h7h6 f1d3 d5c4 d3c4 c6c5 d4c5 e7c5 c4e2 f8d8 e1g1 b8c6 a2a3 c5e7 f1d1 c8d7 b2b4 d7e8 d1d8 a8d8 h2h3 d8c8 c2b2 c6b8 e2f1 b8d7 e3e4 d7b6 e4e5 f6d7 c3e4 c8c1');
INSERT INTO evaluations VALUES(45836,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(45836,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(45836,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(45836,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(45836,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(45836,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(45836,36,'cp 60','g1f3 d5c3 b2c3 b8d7 e1g1 f8d6 d2d4 e8g8 d1c2 f8e8 e2e4 d8c7 a2a4 b7b5 c1e3 b5a4 d4e5 d7e5 f3e5 e8e5 c2a4 e5b5 f1d1 c8e6 e4e5 d6e5 c3c4 b5b7 a1c1 a8b8 g2c6 b7b4 a4a6 e5b2 e3a7 b2c1 a7b8 b4b8');
INSERT INTO evaluations VALUES(45835,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(45835,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(45835,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(45835,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(45835,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(45835,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(45835,36,'cp 67','b8d7 e1g1 f8d6 d2d4 d5c3 b2c3 e8g8 d1c2 f8e8 e2e4 b7b5 c1e3 d8c7 d4e5 d7e5 f3e5 d6e5 f2f4 e5d6 a2a4 d6f8 e4e5 c8b7 c2e4 e8b8 e5e6 c6c5 e4e5 c7e5 e6f7 g8f7 f4e5 f7g8 a4b5 b7g2 g1g2 b8b5 a1a6 a8e8 e5e6 b5b2 g2f3 b2h2 f1d1 h2c2');
INSERT INTO evaluations VALUES(47760,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(47760,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(47760,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(47760,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(47760,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(47760,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(47760,36,'cp 65','b1c3 a7a5 e2e4 b8a6 f1d3 g7g6 d3c2 f8g7 c1g5 h7h6 g5e3 e8g8 h2h3 f6h5 a2a3 h5f4 e1f1 d8e8 g2g3 f4h5 g1f3 f7f5 e4f5 g6f5 h1g1 e5e4 f3d4 a6c5 c3e2 e8f7 e2f4 h5f4 g3f4 g8h8');
INSERT INTO evaluations VALUES(1344,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(1344,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(1344,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(1344,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(1344,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(1344,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(1344,36,'cp 70','b6c6 e1g1 d7d6 d2d4 c5d4 f3d4 c6e4 b1c3 e4g4 d1d3 c8d7 f2f4 g8f6 h2h3 g4h5 f4f5 a8c8 c1d2 e7e5 f5e6 f7e6 a1e1 h5c5 b2b4 c5c4 d4e6 c4d3 e6f4 f8e7 c2d3 e8d8 g2g4 h7h6 b4b5 h8e8 a2a4 b7b6 g1g2 c8c5 a4a5 d7b5 c3b5 c5b5');
INSERT INTO evaluations VALUES(19637,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19637,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(19637,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(19637,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(19637,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(19637,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19637,36,'cp 19','c8d7 c1e3 g8f6 f2f3 a7a6 b5c6 b7c6 d4e2 a6a5 b2b3 f8g7 d1d2 e8g8 e3h6 g7h6 d2h6 c6c5 h2h4 a5a4 h4h5 a4b3 c2b3 d8a5 h5g6 f7g6 e1g1 f8c8 h6e3 c5c4 e3d4 a5a7 b3b4 a7d4 e2d4 g6g5 d4c2 g5g4 b4b5 g4f3 g2f3 g8h8');
INSERT INTO evaluations VALUES(32231,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32231,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32231,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32231,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(32231,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32231,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32231,36,'cp 0','d1e2 e6c4 e2c4 b8c6 b1c3 g7g5 h2h4 g5g4 f3d4 c6e5 c4b5 c7c6 b5b7 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(41683,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(41683,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(41683,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(41683,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(41683,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(41683,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(41683,36,'cp 0','h7h6 g5h4 d8e7 d3e2 c8e6 e1g1 b8d7 e2d3 e8c8 b1c3 a7a6 d1e2 g7g5 h4g3 d6g3 h2g3 e7b4 e2d2 h8e8 a2a3 b4b6 b2b4 c8b8 c3a4 b6c6 a4c5 f6e4 d2e3 e4c5 b4c5 e6g4 e3d2 g4f3 g2f3 c6f6 d2b4 f6f3 c5c6 d7b6 c6b7');
INSERT INTO evaluations VALUES(41618,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(41618,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41618,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(41618,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(41618,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41618,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41618,36,'cp 4','g5h4 d8e7 d3e2 c8e6 e1g1 b8d7 e2d3 e8c8 b1c3 g7g5 h4g3 c7c6 d1d2 h8e8 f1e1 f6h5 g3d6 e7d6 e1e3 e6g4 a1e1 h5g7 h2h3 g4f3 e3f3 e8e1 d2e1 d8e8 e1d2 d6e7 g2g3 d7f6 f3e3 e7b4 c3e4 b4d2 e4d2 c8d7 g1g2 b7b6 d2f1 g7e6 d3f5 a7a5 c2c3 h6h5');
INSERT INTO evaluations VALUES(44402,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(44402,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(44402,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(44402,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(44402,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(44402,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(44402,36,'cp 83','e2e4 d7d6 g1f3 f8g7 f1e2 e8g8 e1g1 b8a6 c1e3 e7e5 d4d5 f6g4 e3g5 f7f6 g5h4 c6c5 f3e1 g4h6 g2g4 h6f7 e1g2 g7h6 g1h1 c8d7 f2f3 f7g5 f3f4 e5f4 g2f4 d8e7 e2d3 a6b4 d3b1 b4a6 f4g2 a8e8 a2a3 h6g7 b1c2 g5f7 h4f2 a6c7');
INSERT INTO evaluations VALUES(44404,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(44404,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(44404,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(44404,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(44404,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(44404,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(44404,36,'cp 66','d7d6 g1f3 f8g7 f1e2 e8g8 e1g1 e7e5 d4d5 b8a6 c1e3 f6g4 e3g5 f7f6 g5h4 c6c5 f3e1 g4h6 f2f3 h6f7 g1h1 a6c7 e1d3 g7h6 h4e1 b7b6 a2a4 a7a5 e1d2 h6d2 d1d2 c7e8 a1e1 e8g7 g2g4 f6f5 g4f5 g6f5 f3f4 f5e4 c3e4');
INSERT INTO evaluations VALUES(44723,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(44723,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(44723,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(44723,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(44723,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(44723,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(44723,36,'cp 75','f1e2 f8g7 g1f3 e8g8 e1g1 b8a6 c1e3 e7e5 d4d5 f6g4 e3g5 f7f6 g5h4 c6c5 g1h1 g4h6 f3e1 c8d7 a2a3 b7b6 f2f3 h6f7 g2g4 g7h6 e1g2 a8b8 h4f2 a6c7 b2b4 f7g5 f2e3 g5f7 e3h6 f7h6 d1d2 h6f7 b4b5 c7e8 g2e3 f7g5 a1d1');
INSERT INTO evaluations VALUES(47045,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(47045,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(47045,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(47045,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(47045,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(47045,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(47045,36,'cp -16','d2d4 g7g6 f1c4 f8g7 f3e2 e8g8 g1f3 c8g4 c2c3 b8d7 h2h3 g4f3 e2f3 e7e5 c1e3 d6d5 e4d5 e5d4 c3d4 d7b6 c4b3 f6d5 b1c3 d5e3 f2e3 d8h4 f3f2 h4e7 e1g1 a8e8 e3e4 c7c5 d4c5 b6d7 a1d1 g7c3 b2c3 d7c5 b3d5 c5e4 f2d4 e4d6 d4a7 d6f5');
INSERT INTO evaluations VALUES(50940,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(50940,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(50940,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(50940,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(50940,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(50940,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(50940,36,'cp 85','g1f3 b8c6 f1d3 g8f6 c1e3 f8e7 e1g1 e8g8 a2a3 b7b5 b2b4 f6d7 d1e2 c8b7 f1e1 e7f6 e2d2 d7b6 c3e2 c6b8 a1d1 b8d7 e2g3 f6e7 h2h3 f8e8 g3h5 d7f6 h5f6 e7f6 c2c3 a8c8 d2c2 h7h6 f3d2 e6e5 d4d5 c7c5 b4c5 d6c5');
INSERT INTO evaluations VALUES(6581,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6581,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6581,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6581,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6581,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6581,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(6581,36,'cp 0','g8f6 f1c4 d7d6 d2d3 c6a5 b2b4 f6d5 b4c5 d5f4 c1f4 e5f4 c5d6 a5c4 d3c4 c7d6 d1d2 d8f6 d2d4 f6d4 f3d4 c8d7 e1d2 e8e7 h1b1 a8b8 d2d3 h8c8 a2a4 c8c5 a4a5 c5h5 b1h1 b8c8 a1b1 c8b8');
INSERT INTO evaluations VALUES(10322,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10322,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(10322,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(10322,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(10322,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(10322,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(10322,36,'cp 0','f1c4 f6e4 d1e2 e4f6 d2d4 c5d4 c1g5 d7d6 f3d2 c8g4 d5f6 g7f6 e2g4 f6g5 d2e4 c6a5 c4d5 c7c6 d5f7 e8f7 c2c3 d4b6 a1d1 d6d5 e4g5 f7e8 g4h5 e8d7 h5g4');
INSERT INTO evaluations VALUES(7368,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(7368,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(7368,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(7368,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(7368,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(7368,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(7368,36,'cp -44','f6e4 d1e2 e4f6 f3e5 c6e5 e2e5 e8g8 d2d4 d7d6 e5e2 b7b5 a4c2 f8e8 b1d2 e7f8 e2d1 c8b7 d2f3 h7h6 a2a4 d8d7 c1e3 b7d5 h2h3 g7g6 c2b3 d7c6 a4b5 a6b5 a1a8 e8a8 b3d5 c6d5 d1c2');
INSERT INTO evaluations VALUES(7376,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(7376,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(7376,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(7376,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(7376,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(7376,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(7376,36,'cp -25','d1e2 e4c5 a4c6 d7c6 f3e5 e8g8 d2d4 c5e6 b1d2 c6c5 d4c5 e6c5 f1d1 d8e8 d2e4 a6a5 e4c5 e7c5 d1e1 a8a6 c1f4 a6e6 e2d2 f7f6 e5c4 e8b5 c4e3 c7c6 a2a4 b5b3 f4c7 e6e4 c7a5 e4a4 a5c7 a4a1 e1a1');
INSERT INTO evaluations VALUES(14342,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(14342,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(14342,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(14342,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(14342,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(14342,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(14342,36,'cp -28','e7e5 e1g1 c8e6 f3g5 e6g4 c2c4 d5d4 c4c5 a8b8 d1b3 g4h5 b3c4 f8e7 b2b4 f6d7 g5f3 f7f6 c1d2 h5f7 c4c2 a7a5 a2a3 e8g8 e2e3 d4e3 f2e3 e5e4 d3e4 b8a8 f3d4 a5b4 d4c6 b7c6 d2b4 f8e8 b1d2');
INSERT INTO evaluations VALUES(18666,30,'cp 247',NULL);
INSERT INTO evaluations VALUES(18666,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(18666,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(18666,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(18666,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(18666,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(18666,36,'cp 226','d1b3 d8f6 c3b5 f8b4 c1d2 b4d2 f3d2 e8d8 e1g1 a7a6 b5c3 c6d4 b3a3 g8e7 a1c1 a8b8 d2b3 b7b6 b3d4 e5d4 c3d5 e7d5 c4d5 c8b7 d5b7 b8b7 a3a6 b7b8 a6d3 b8c8 c1c8 d8c8 f1d1 c8b8 d3d4 f6d4 d1d4 h8c8 g1f1 c8c1 f1e2 b8c7 d4d1 c1c2 d1d2 c2c1 h2h4 b6b5 e2f3 c1e1 g2g3 d7d6 a2a3 c7d7 d2d5 d7c6 d5f5 e1b1 f5f7 b1b2 f7g7 b2b3 f3f4 b3a3');
INSERT INTO evaluations VALUES(26554,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26554,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26554,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26554,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26554,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26554,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26554,36,'cp 0','f1d3 e7e6 g1f3 g8f6 e1g1 h7h5 a2a3 f6e4 f3d2 f8d6 d3e4 d5e4 d1e1 d8f6 f2f3 e4f3 d2f3 f5c2 f3e5 c2f5 e5c6 b7c6 e3e4 f6d4 c1e3 d4e5 e3f4 e5d4');
INSERT INTO evaluations VALUES(28014,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(28014,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(28014,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(28014,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(28014,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(28014,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(28014,36,'cp 117','d2d4 d7d6 h2h4 h7h6 c1e3 b8d7 d1e2 e7e6 e1c1 a7a6 c4b3 d8e7 c1b1 g8f6 h1e1 e8g8 h4h5 f6e4 c3e4 b7e4 h5g6 e4g6 e2d2 h6h5 e3g5 g7f6 b3d5 f6g5 f3g5 a8e8 d5c6 g8g7 c6e4 f8h8 e4g6 g7g6 g5h3 g6g7 h3f4');
INSERT INTO evaluations VALUES(33799,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(33799,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(33799,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33799,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(33799,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(33799,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33799,36,'cp -8','f6d5 e3g5 f7f6 g5d2 b8c6 b1a3 c8f5 a3b5 e8c8 b5d4 c6d4 f3d4 f5d7 e2e7 f8e7 e1c1 f6f5 g2g3 e7f6 d4e2 d7c6 d1e1 d8e8 h1g1 g7g5 c2c4 d5b6 h2h4 g5h4 f1h3 c6d7 g3h4 f6h4 e2g3 e8e1 d2e1 h4g3 f2g3');
INSERT INTO evaluations VALUES(42216,30,'cp 475',NULL);
INSERT INTO evaluations VALUES(42216,31,'cp 480',NULL);
INSERT INTO evaluations VALUES(42216,32,'cp 508',NULL);
INSERT INTO evaluations VALUES(42216,33,'cp 509',NULL);
INSERT INTO evaluations VALUES(42216,34,'cp 501',NULL);
INSERT INTO evaluations VALUES(42216,35,'cp 526',NULL);
INSERT INTO evaluations VALUES(42216,36,'cp 513','d1a4 b8c6 d5c6 b4c3 b2c3 e8g8 c1g5 h7h6 a1d1 d8e7 g5f6 e7f6 f1c4 f6g5 g1e2 g5g2 h1g1 g2h2 c4d5 b7b6 c3c4 h2h3 a4c2 h3a3 c2c3 a3c3 e2c3 a7a6 d1c1 c8e6 e1e2 g7g5 e2e3 g8g7 c4c5 b6c5 g1d1 a8d8 d5e6 d8d1 c1d1');
INSERT INTO evaluations VALUES(42218,30,'cp 512',NULL);
INSERT INTO evaluations VALUES(42218,31,'cp 519',NULL);
INSERT INTO evaluations VALUES(42218,32,'cp 488',NULL);
INSERT INTO evaluations VALUES(42218,33,'cp 517',NULL);
INSERT INTO evaluations VALUES(42218,34,'cp 528',NULL);
INSERT INTO evaluations VALUES(42218,35,'cp 496',NULL);
INSERT INTO evaluations VALUES(42218,36,'cp 497','b8c6 d5c6 b4c3 b2c3 e8g8 c1g5 a7a6 c6b7 c8b7 a4c4 a6a5 a1d1 d8e7 g1f3 a5a4 g5f6 g7f6 f1e2 f8b8 c4d3 e7e6 d3c2 e6g4 e1g1 g4e4 f3e1 e4c2 e1c2 f6f5 f1e1 f5f4 f2f3 g8g7 c2a3 e5e4 f3e4 b7e4 a3c4 e4c2 d1d2');
INSERT INTO evaluations VALUES(55772,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(55772,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(55772,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(55772,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(55772,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(55772,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(55772,36,'cp -32','e6d5 g1f3 g8f6 c1b2 f8e7 d2d4 e8g8 f1e2 c8f5 c2c4 f8e8 b1c3 c7c6 e1g1 b8d7 e2d3 f5g6 f1e1 d7f8 f3e5 g6d3 d1d3 f8g6 a1d1 e7d6 c4d5 c6d5 d3b5 g6e5 d4e5 a7a6 b5b7 d6e5 c3a4 e5h2 g1h2 e8e1 d1e1');
INSERT INTO evaluations VALUES(56265,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(56265,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(56265,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(56265,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(56265,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(56265,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(56265,36,'cp -19','d2d4 f8d6 f1d3 g8f6 g1e2 e8g8 b1c3 b8c6 a2a3 c8g4 f2f3 g4h5 e1g1 h5g6 c1g5 d6e7 g5h4 f8e8 h4f2 d8d7 e2f4 c6d8 f1e1 c7c6 c3a4 e7d6 e1e8 f6e8 f4g6 h7g6 a4c5 d7c7 g2g3 b7b6 c5a6 c7d7 d1d2 e8f6 g1g2 d8e6 c2c3');
INSERT INTO evaluations VALUES(17548,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(17548,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(17548,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17548,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17548,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(17548,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(17548,36,'cp 31','b5c6 b7c6 c2c4 e7e6 c1f4 g8e7 b1d2 f7f6 e1g1 e7f5 f1e1 f8e7 a1c1 e8g8 h2h4 c6c5 c4d5 d8d5 d2c4 d5b7 c4a5 b7b6 a5b3 f5d4 f3d4 c5d4 f4c7 b6b7 b3d4 e7b4 d4e6 b4e1 e6f8 e1f2 g1f2 g8f8 f2g3 c8e6');
INSERT INTO evaluations VALUES(17539,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(17539,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(17539,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(17539,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(17539,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(17539,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(17539,36,'cp 12','b7c6 c1f4 c8g4 b1d2 e7e6 h2h3 g4f3 d2f3 f8d6 f4e5 g8f6 e1g1 d6e5 f3e5 f6d7 e5d3 e8g8 d1g4 d8b6 a1c1 c6c5 c2c4 d5c4 d4c5 d7c5 d3c5 b6c5 c1c4 c5e5 g4d4 e5a5 c4a4 f8d8 a4a5 d8d4');
INSERT INTO evaluations VALUES(18202,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(18202,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18202,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(18202,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(18202,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(18202,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(18202,36,'cp 44','c2c4 e7e6 c4c5 f7f6 e1g1 g7g5 f3e1 h7h5 b1c3 f8g7 c3a4 g8e7 a4b6 a8a7 b2b4 g5g4 e1d3 e6e5 d4e5 f6e5 c1b2 e7g6 f2f4 e8g8 d3e5 g6e5 b2e5 c8f5 a2a4 f5e4 d1d4 d8h4 b6c4 g4g3 h2h3 f8f4 f1f4 h4f4 e5g7 f4f2 d4f2 g3f2 g1f2 a7g7 c4a5 g7g2 f2e3 g2g6 e3d4 h5h4 a1e1 e4g2');
INSERT INTO evaluations VALUES(19503,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(19503,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19503,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(19503,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(19503,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19503,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(19503,36,'cp 12','d2d4 c5d4 e3d4 d7d5 c2c3 c8g4 c1f4 e7e6 h2h3 g4f5 d1b3 d8d7 b1d2 g8e7 f4h2 h7h5 f1e2 e7g6 e1g1 f8d6 h2d6 d7d6 f1e1 e8g8 a2a4 g6f4 b3d1 d6c7 e2f1 a7a5 e1e3 f4g6 a1c1 b7b6 d1e2');
INSERT INTO evaluations VALUES(19422,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19422,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19422,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(19422,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(19422,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19422,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19422,36,'cp 5','c5d4 e3d4 d7d5 c2c4 g8f6 b1c3 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 a1d1 e7e6 f3g3 d8d6 d1d3 d6g3 h2g3 h7h5 e1e2 e8d7 b2b3 f8d6 c3e4 a7a5 d3c3 a5a4 e4d6 d7d6 b3a4 c8b8 h1c1 b8b2 c3c2');
INSERT INTO evaluations VALUES(19553,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(19553,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19553,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(19553,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19553,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19553,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19553,36,'cp 9','e3d4 d7d5 c2c3 c8g4 h2h3 g4h5 c1f4 e7e6 f1b5 f8d6 f4d6 d8d6 e1g1 a7a6 b5d3 g8f6 f1e1 e8g8 b1d2 a8c8 d1e2 d6f4 g2g3 f4c7 e2e3 c8e8 g1g2 f6d7 f3g5 h5g6 d3g6 h7g6 h3h4 e6e5 d4e5 c7e5');
INSERT INTO evaluations VALUES(19542,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(19542,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19542,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(19542,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(19542,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19542,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(19542,36,'cp 18','d7d5 c2c3 c8g4 c1f4 e7e6 f1d3 f8d6 f4d6 d8d6 e1g1 g8f6 f1e1 e8g8 b1d2 g4f5 f3e5 c6e7 d2f1 d6b6 d1e2 f5d3 e5d3 a7a5 f1e3 a8c8 h2h3 f6e4 a2a4 c8c7 h3h4 e4f6 h4h5 h7h6 e2f3 f6e4 f3h3 f7f5');
INSERT INTO evaluations VALUES(20274,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(20274,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(20274,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20274,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(20274,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(20274,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(20274,36,'cp 8','f1d3 b7b6 e3e4 c5d4 f3d4 c8b7 b1d2 g8e7 d2f3 d5e4 d3e4 d8d7 e1g1 a8d8 d1e2 c6d4 f3d4 b7e4 e2e4 d7d5 e4e2 e7g6 e2b5 d8d7 b5d5 d7d5 d4c6 d5d7 c1e3 d7c7 c6d4 f8c5 d4b5 c7d7 e3c5');
INSERT INTO evaluations VALUES(20278,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(20278,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(20278,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20278,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20278,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20278,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(20278,36,'cp 2','g8f6 e1g1 f8d6 b1d2 e8g8 d4c5 d6c5 b2b4 c5d6 c1b2 a7a6 a2a3 c6e5 f3e5 d6e5 d2f3 e5c7 c3c4 d5c4 d3c4 b7b5 d1d8 f8d8 c4e2 c8b7 f1c1 a8c8 a3a4 c7d6 a4b5 c8c1 a1c1 a6b5 f3d4 d6b4 e2b5 b7d5 d4c6 d5c6 b5c6 d8d2 b2f6 g7f6');
INSERT INTO evaluations VALUES(22453,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(22453,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(22453,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(22453,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(22453,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(22453,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(22453,36,'cp -93','h5g5 d8g5 f3g5 g8f6 b1c3 c6d4 f1d3 d7d6 c3b5 d4b5 d3b5 c7c6 b5d3 d6d5 e4d5 f6d5 d3c4 h7h6 g5e4 c8e6 e4c3 f8c5 c4b3 a7a5 a2a3 f7f5 d2d3 e8f7 h2h4 b7b5 g2g3 a5a4 b3d5 e6d5 c3d5 c6d5 e1e2 a8c8 c2c3 c5e7 c1e3');
INSERT INTO evaluations VALUES(22447,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(22447,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(22447,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(22447,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(22447,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(22447,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(22447,36,'cp -94','d8g5 f3g5 g8f6 b1c3 c6d4 f1d3 d7d6 c3b5 d4b5 d3b5 c7c6 b5d3 d6d5 e4d5 f6d5 d3c4 h7h6 g5e4 c8e6 e4c3 f8c5 c4b3 a7a5 a2a3 d5c3 b2c3 e6b3 c2b3 e5e4 f2f3 e4f3 g2f3 a5a4 b3b4 c5e7 e1e2 f7f5 d2d4 e8f7 c1e3 g6g5 h2h4 f5f4 e3f2');
INSERT INTO evaluations VALUES(23864,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(23864,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(23864,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(23864,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(23864,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(23864,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(23864,36,'cp 27','g4e6 c4e6 f7e6 c3a4 c5b6 e1g1 g8f6 a4b6 a7b6 a2a4 e8g8 c2c3 h7h6 b2b4 d8e8 c1e3 e8g6 f1e1 g8h7 g1h2 f8f7 a1a2 g6h5 h2g1 h5g6 d1e2 a8f8 f3h4 g6h5 e2h5 f6h5 e1a1 h5f4 a2d2 c6e7 e3f4 f7f4 h4f3 e7g6 b4b5');
INSERT INTO evaluations VALUES(25571,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(25571,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(25571,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(25571,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(25571,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(25571,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(25571,36,'cp 12','g1f3 e7e6 e1g1 f8d6 b5d3 e6e5 d4e5 c6e5 e3e4 e5f3 d1f3 d7g4 f3e3 d5d4 e3e1 d4c3 e4e5 e8g8 e1c3 d6e7 e5f6 e7f6 c3c4 d8c8 f1e1 g4f5 d3f5 c8f5 c4c7 f8c8 c7b7 f5c2 a2a4 a7a5 b7f3 h7h6 g2g3 a8b8 c1f4 b8b3 f4e3 f6b2 a1d1 b3c3');
INSERT INTO evaluations VALUES(25572,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(25572,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(25572,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(25572,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(25572,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(25572,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(25572,36,'cp 13','e7e6 e1g1 f8d6 b5d3 e6e5 d4e5 c6e5 e3e4 e5f3 d1f3 d7g4 f3e3 d5d4 e3e1 d4c3 e4e5 e8g8 e1c3 d6e7 e5f6 e7f6 c3c4 d8c8 c2c3 f8e8 h2h3 g4e6 c4e4 g7g6 e4f4 c8d8 f1d1 c7c6 c1e3 d8e7 d3c4 e6c4 f4c4 a7a6 a2a4 a8d8 e3b6 d8d1 a1d1 e7e2 c4e2 e8e2');
INSERT INTO evaluations VALUES(41782,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(41782,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(41782,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(41782,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(41782,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41782,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41782,36,'cp 17','d1f3 b8c6 e1c1 a7a6 g2g4 d6f4 f3f4 d8d6 f4d6 c7d6 f2f3 b7b5 a2a3 c6e7 g1h3 e7g6 f1e2 h7h5 g4g5 f6g8 e2d3 g8e7 d3g6 e7g6 h3f4 g6h4 h1f1 c8b7 c1d2 e8e7 f1f2 h4f5 f4d3 h5h4 a3a4 b5a4 c3a4 b7c6 d1a1 c6d7');
INSERT INTO evaluations VALUES(143,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(143,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(143,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(143,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(143,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(143,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(143,36,'cp 106','b1c3 g8e7 e1g1 e7f5 g1h1 e8g8 a2a3 b4e7 e2d3 b6d8 d3c2 f7f6 g2g4 f5d4 f3d4 f6e5 d4c6 b7c6 f2f3 c8a6 c2d3 a6d3 d1d3 a8b8 d3e2 d8c7 a1b1 c6c5 a3a4 b8b3 c1d2 a7a5 g4g5 c5c4 c3b5 c7c6 e2e5 b3f3 f1f3 f8f3');
INSERT INTO evaluations VALUES(138,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(138,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(138,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(138,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(138,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(138,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(138,36,'cp 99','g8e7 e1g1 e7f5 g1h1 e8g8 a2a3 b4e7 e2d3 b6d8 d3c2 f7f6 g2g4 f5d4 f3d4 f6e5 d4e2 b7b6 f2f4 c8a6 b2b4 a6c4 c2d3 b6b5 c3b5 c4d3 d1d3 d5d4 d3c4 a8c8 c4e6 g8h8 c1b2 f8f6 e6b3 a7a6 g4g5 f6g6 b5d4 c6d4 b2d4 e5d4 a1d1 g6d6 b3d3 d8d7 f1f3 h8g8 h1g1 c8f8 f3h3 g7g6 e2d4');
INSERT INTO evaluations VALUES(471,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(471,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(471,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(471,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(471,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(471,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(471,36,'cp 131','e1g1 e7f5 g1h1 e8g8 a2a3 b4e7 e2d3 b6d8 d3c2 f7f6 g2g4 f5d4 f3d4 f6e5 d4e2 b7b6 f2f4 c8a6 b2b4 a6c4 c2d3 b6b5 c3b5 c4d3 d1d3 d5d4 d3e4 a8c8 c1b2 a7a6 a1c1 d8d5 e4d5 e6d5 f4e5 d4d3 b5d4 f8f1 c1f1 d3e2 d4e2 e7g5 f1d1 d5d4 e5e6 c8e8 b2d4');
INSERT INTO evaluations VALUES(470,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(470,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(470,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(470,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(470,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(470,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(470,36,'cp 131','e7f5 g1h1 b6d8 e2d3 c8d7 c1d2 b4c3 b2c3 c6a5 a1b1 d8c7 f1e1 h7h6 h1g1 b7b6 d1e2 a7a6 h2h4 a5c4 d2c1 f5e7 f3h2 c4a5 e2g4 c7c3 e1e3 a5c4 e3f3 e7c6 g4g7 e8c8 d3c4 c3c4 c1h6 c4d4');
INSERT INTO evaluations VALUES(477,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(477,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(477,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(477,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(477,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(477,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(477,36,'cp 138','g1h1 e8g8 a2a3 b4e7 e2d3 b6d8 d3c2 f7f6 g2g4 f5d4 f3d4 f6e5 d4e2 b7b6 f2f4 c8a6 b2b4 a6c4 c2d3 b6b5 c3b5 c4d3 d1d3 d5d4 d3e4 a8c8 c1b2 a7a6 b5d4 c6d4 b2d4 e5d4 e2d4 d8d5 a1e1 c8c4 e4d5 e6d5 d4f5 e7d8 e1e6 f8f6 e6e5 g7g6 f5e3 c4e4 e3d5 f6f4 e5e4 f4e4 d5f6 d8f6 f1f6 g8g7 f6a6 e4g4 a6a7 g7h6');
INSERT INTO evaluations VALUES(4325,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(4325,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(4325,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(4325,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4325,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4325,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4325,36,'cp 0','c5e3 f2e3 d7d6 a2a4 c8e6 c4e6 f7e6 e1g1 h7h6 c3e2 d8e8 d1d2 a7a5 f1f2 f6g4 f2f1 g4f6');
INSERT INTO evaluations VALUES(4446,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4446,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4446,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4446,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4446,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4446,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4446,36,'cp 0','f2e3 d7d6 a2a4 c8e6 c4e6 f7e6 e1g1 h7h6 c3e2 c6e7 h2h3 d8d7 d1d2 c7c5 e2g3 f8f7 f1f2 a8f8 b2b3 a7a6 d2e2 b7b5 a4b5 a6b5 a1f1 g8h7 f3h2 d7e8 e2d2 e7c6 f2f3 e8d7 d2e2 b5b4 f1f2 d7a7 f2f1');
INSERT INTO evaluations VALUES(4445,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4445,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4445,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4445,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4445,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4445,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4445,36,'cp 0','d7d6 a2a4 c8e6 c4e6 f7e6 e1g1 h7h6 c3e2 a7a5 h2h3 d8e8 d1d2 e8g6 f1f2 f8f7 a1f1 a8f8 f3h4 g6h5 h4f3');
INSERT INTO evaluations VALUES(3883,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3883,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3883,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3883,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3883,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3883,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3883,36,'cp 0','a2a4 c8e6 c4e6 f7e6 e1g1 d8e8 d1d2 e8g6 f1f2 f8f7 h2h3 a8f8 a1f1 b7b6 f3h4 g6h6 h4f3 h6g6');
INSERT INTO evaluations VALUES(9254,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(9254,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(9254,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(9254,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(9254,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(9254,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(9254,36,'cp -29','d6g3 h2g3 e6e5 d3b5 e5e4 g1h3 e8g8 b5e2 c8h3 g2h3 c6e7 c3c4 c7c6 b1c3 d8d7 c4d5 e7d5 a1c1 g7g6 h3h4 h7h5 c3d5 f6d5 a2a3 f8d8 e2f1 d7e6 f1h3 e6e7 c1c5 d8d6 d1c2 a8d8 h3g2 f7f5 e1g1 a7a6 f1e1 d8d7 c2b3 g8h8 g2h3');
INSERT INTO evaluations VALUES(9010,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(9010,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(9010,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(9010,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(9010,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(9010,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(9010,36,'cp -18','h2g3 e6e5 d3b5 e5e4 b5e2 c6e7 c3c4 e8g8 b1c3 c7c6 g1h3 c8h3 g2h3 d8d7 c4d5 e7d5 a1c1 g7g6 h3h4 h7h5 c3d5 f6d5 a2a3 g8g7 d1c2 d7e7 e1g1 f8d8 e2c4 d8d6 c4d5 d6d5 c2e2 d5f5 g1g2 a8d8 b2b4');
INSERT INTO evaluations VALUES(9009,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(9009,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(9009,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(9009,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(9009,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(9009,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(9009,36,'cp -22','e6e5 d3b5 e5e4 g1h3 e8g8 b5e2 c6e7 c3c4 c7c6 b1c3 c8h3 g2h3 a8c8 d1a4 d5c4 a4c4 e7d5 a1c1 h7h5 c3d5 f6d5 h3h4 g7g6 c4b3 d8d7 e2f1 c8c7 f1g2 f8e8 c1c5 a7a6 a2a3 b7b6 g2h3 d7e7 c5c2 g8g7 h3g2');
INSERT INTO evaluations VALUES(14113,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(14113,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(14113,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(14113,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(14113,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(14113,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(14113,36,'cp 103','c2c4 f8g7 b1d2 c8e6 h2h3 e8g8 a1c1 f6e4 f1d3 e4d2 d1d2 d5c4 d3c4 e6c4 c1c4 d8d5 d2c3 a8e8 e1g1 e7e5 f3e5 c6e5 f4e5 g7e5 d4e5 e8e5 b2b4 e5g5 e3e4 d5e5 c3e3 e5e7 f1c1 c7c6 e3a7 f8d8 a2a4 g5e5 a4a5 d8d2 a7e3 d2b2 f2f4');
INSERT INTO evaluations VALUES(14104,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(14104,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(14104,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(14104,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(14104,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(14104,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(14104,36,'cp 112','f8g7 b1d2 e8g8 a1c1 c8f5 a2a3 f6e4 f1e2 e4d2 d1d2 d5c4 c1c4 a8c8 h2h4 f5e4 b2b4 f8e8 h4h5 e7e5 d4e5 d8d2 e1d2 c8d8 d2e1 e4f3 g2f3 d8d7 e5e6 e8e6 b4b5 c6e5 c4a4 a7a6 b5a6 b7a6 e1f1 e5d3 f4g3 e6c6 f1g2 d3c5 a4g4 d7d2 e2c4 c5d3 g4e4');
INSERT INTO evaluations VALUES(20530,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(20530,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(20530,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(20530,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(20530,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(20530,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(20530,36,'cp -39','d8e7 d1e2 g8f6 d2d4 e7e2 e1e2 c5c4 b3c2 f8d6 c1g5 f6h5 h1e1 e8g8 e2f1 f7f6 g5c1 b7b5 f3h4 g7g6 g2g3 g8f7 h4g2 f8d8 g2e3 c8h3 f1g1 h3e6 b2b3 a8b8 b3c4 b5c4 c1a3 c6e7 a3d6 d8d6 b1a3 d6b6 e1b1 f6f5 b1b6 b8b6');
INSERT INTO evaluations VALUES(22060,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(22060,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(22060,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(22060,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(22060,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(22060,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(22060,36,'cp -98','d6e5 d4e5 d8d1 e1d1 g8e7 c1f4 e7d5 f4g3 c8d7 b1d2 c6e7 c2c4 d5b6 d1c2 e8c8 a1d1 d7c6 f3g5 c6e8 f2f4 e7c6 c2c1 f8e7 f1e2 h7h6 g5e4 c6d4 h1e1 d4e2 e1e2 e8c6 b2b3 b6d7 d2f3 h8g8 e4c3 b7b6 f3d4 c6b7 c1c2 g7g5');
INSERT INTO evaluations VALUES(21481,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(21481,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(21481,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(21481,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(21481,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(21481,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(21481,36,'cp -98','d4e5 d8d1 e1d1 g8e7 c1f4 c8d7 b1d2 e7d5 f4g3 c6e7 c2c4 d5b6 d1c2 e8c8 a1d1 d7c6 f1d3 h7h6 h2h4 e7f5 d3f5 e6f5 h4h5 f8b4 g3h4 d8e8 h1e1 c6f3 g2f3 b4d2 d1d2 b6c4 d2e2 e8e6 b2b3 c4b6 e1g1 g7g5 h5g6 e6g6 g1g6 f7g6 h4f6 h8f8');
INSERT INTO evaluations VALUES(21480,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(21480,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(21480,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(21480,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(21480,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(21480,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(21480,36,'cp -71','d8d1 e1d1 g8e7 c1f4 c8d7 b1d2 e7g6 f4g3 h7h5 h2h4 e8c8 f1d3 c6e7 c2c3 d7a4 d3c2 a4e8 c2e4 f7f5 e4c2 f5f4 g3h2 e7f5 f3g5 f5e3 d1c1 e3c2 c1c2 g6e5 h2f4 e8g6 d2e4 e5c4 a1e1 e6e5 f4c1 b7b6 b2b3 c4d6 f2f3 d8d7 e1e2 f8e7 h1e1 d6e4 f3e4');
INSERT INTO evaluations VALUES(1682,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(1682,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(1682,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(1682,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(1682,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(1682,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(1682,36,'cp -66','e1d1 g8e7 c1f4 c8d7 b1d2 e7g6 f4g3 h7h5 h2h4 e8c8 f1d3 c6e7 c2c3 d7a4 d3c2 a4e8 c2e4 e7f5 d1c2 f5g3 f2g3 g6e7 f3g5 g7g6 a1f1 e7f5 f1f3 f8h6 h1e1 e8b5 d2b3 h6g5 h4g5 b5a4 f3f4 a4b3 c2b3 d8d7 g3g4 h5g4 f4g4 c7c6 e4f3 d7d2 g4a4 a7a6 g2g4 f5g3');
INSERT INTO evaluations VALUES(1681,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(1681,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(1681,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(1681,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(1681,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(1681,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(1681,36,'cp -93','g8e7 c1f4 c8d7 b1d2 e7g6 f4g3 h7h5 h2h4 e8c8 f1d3 c6e7 c2c3 d7a4 d3c2 a4e8 c2e4 e7f5 d1c2 f5g3 f2g3 g6e7 f3g5 g7g6 a1f1 e7f5 f1f3 f8h6 d2c4 b7b6 b2b3 h6g5 h4g5 e8b5 a2a4 b5a6 h1e1 a6b7 e4b7 c8b7 b3b4 d8d7 e1d1 d7d1 c2d1 h8d8 d1c2 d8d5 b4b5 a7a5 f3f4');
INSERT INTO evaluations VALUES(26267,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26267,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26267,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26267,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26267,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26267,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26267,36,'cp 0','g4f3 e2f3 e7e6 e1g1 g8e7 b1a3 c5d4 c3d4 e7f5 a3c2 h7h5 f3e2 a8c8 e2b5 g7g6 c2e3 f5e7 e3c2');
INSERT INTO evaluations VALUES(40873,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(40873,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40873,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(40873,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(40873,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(40873,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40873,36,'cp 28','e8g8 d1d2 b8d7 e1c1 c7c6 h2h4 d6d5 c1b1 d7f6 f1d3 f8e8 f3g5 a7a5 f2f3 c6c5 c3c4 b7b5 c4d5 c5c4 d3e4 e7d6 g2g4 h7h6 d2f2 a8a6 g5f7 g8f7 g4g5 f6e4 f3e4 f7g8 g5h6 d8f6 f2d2 e8e4 h1g1 d6f8 h6g7 f8g7 e3d4 f6d4 d2d4 e4d4 d1d4 a6d6 g1g5 g8h7 d4d1 b5b4');
INSERT INTO evaluations VALUES(36279,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(36279,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(36279,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(36279,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36279,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36279,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(36279,36,'cp 18','d1d2 b8d7 e1c1 c7c6 c1b1 d6d5 h2h4 d7f6 f1d3 f8e8 f3g5 c6c5 c3c4 d5d4 e3f4 h7h6 d1e1 c8d7 g5e4 d7c6 e4g3 e7f8 f2f3 d8d7 g3f5 e8e1 h1e1 a8e8 e1e8 d7e8 d2e2 e8e2 d3e2 c6d7 e2d3 d7c6 a2a3 f6d7 b2b3 a7a6');
INSERT INTO evaluations VALUES(36280,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(36280,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36280,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(36280,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(36280,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36280,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36280,36,'cp 9','b8d7 e1c1 c7c6 c3c4 d7b6 e3g5 e7g5 d2g5 c8e6 g5f4 d6d5 f1d3 h7h6 h1e1 d8b8 f4b8 a8b8 c4d5 e6d5 d3e4 d5e4 e1e4 f8e8 e4e1 g8f8 f3e5 b8d8 d1d8 e8d8 b2b3 b6d5 g2g3 f7f6 e5d3 b7b6 a2a4 f8f7 c1d2 d5b4');
INSERT INTO evaluations VALUES(42074,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(42074,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(42074,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(42074,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(42074,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(42074,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(42074,36,'cp 52','e8g8 f1e2 d5c4 c1a3 f8e8 e2c4 b7b6 e1g1 c8b7 f1e1 c7c5 c4d3 b8c6 d1e2 a8c8 a1d1 d8c7 e3e4 c6e7 f3e5 e8d8 e2b2 h7h6 f2f3 e7c6 e5c6 b7c6 d1c1 c6b7 g1h1 d8d7 d3f1 b7c6 b2f2 c6a4 d4c5 b6c5');
INSERT INTO evaluations VALUES(37058,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(37058,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(37058,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(37058,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(37058,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(37058,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(37058,36,'cp 42','f1e2 d5c4 c1a3 f8e8 e2c4 b7b6 e1g1 c8b7 c4e2 b8d7 d1b3 c7c5 d4c5 f6e4 c5b6 a7b6 a1d1 b7c6 b3b2 e6e5 a3b4 c6a4 d1b1 a4c6 c3c4 d8f6 f3d2 e4c5 b4c5 b6c5 f1d1 f6g6 f2f3 e8b8 d2b3 h7h6 b2d2 c6a4 e3e4 g6f6 h2h3 a4b3');
INSERT INTO evaluations VALUES(45021,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45021,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(45021,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(45021,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(45021,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45021,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(45021,36,'cp -16','b7b5 a2a4 b5b4 c3a2 e7e6 f1c4 c8b7 e1g1 f8e7 d1e2 b8d7 f1d1 e8g8 e3e4 a7a5 c1f4 d8b6 c4d3 b7a6 a2c1 c6c5 c1b3 a6d3 e2d3 f8c8 d1c1 c5d4 d3d4 e7c5 b3c5 c8c5 f4e3 c5c1 a1c1 b6d4 f3d4 f6e4');
INSERT INTO evaluations VALUES(43938,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(43938,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(43938,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(43938,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(43938,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43938,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(43938,36,'cp 0','a2a4 b5b4 c3b1 c8a6 b1d2 c4c3 b2c3 b4c3 d2b1 a6f1 e1f1 b8d7 g2g3 e7e6 c1a3 f8a3 a1a3 c6c5 a3c3 e8g8 f1g2 d8a5 d4c5 d7c5 d1c2 c5d7 h1c1 f6d5 c3a3 d7b6 c2c5 a5c5 c1c5 b6d7 c5c4 d7b6');
INSERT INTO evaluations VALUES(43941,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(43941,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(43941,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(43941,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(43941,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(43941,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43941,36,'cp 0','b5b4 c3b1 c8a6 b1d2 c4c3 b2c3 b4c3 d2b1 a6f1 e1f1 b8d7 g2g3 e7e6 c1a3 f8a3 a1a3 c6c5 a3c3 e8g8 f1g2 a8c8 d4c5 c8c5 c3c5 d7c5 d1d8 f8d8 b1d2 g7g5 h1c1 c5a4 c1a1 a4c5 a1a7 g5g4 f3d4 g8g7 d2c4 d8d5 c4b6 d5d8');
INSERT INTO evaluations VALUES(43944,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(43944,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(43944,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43944,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43944,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43944,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43944,36,'cp -1','c3b1 c8a6 b1d2 c4c3 b2c3 b4c3 d2b1 a6f1 e1f1 b8d7 g2g3 e7e6 c1a3 f8a3 a1a3 c6c5 a3c3 d8a5 f1g2 e8g8 d4c5 d7c5 d1c2 c5d7 b1d2 h7h6 c3c7 d7b6 d2e4 f6d5 c7c5 a5a4 c2a4 b6a4 c5a5 a4b6 h1a1 f7f5 e4d6 f8d8 d6b5 b6c4 a5a6');
INSERT INTO evaluations VALUES(53946,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(53946,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(53946,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(53946,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(53946,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(53946,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(53946,36,'cp 96','c3d4 e5d4 f3d4 b8c6 f1c4 g8f6 e1g1 f8e7 b1c3 e8g8 f1e1 c6d4 d1d4 c8e6 c4e6 f7e6 e4e5 f6e8 c1e3 d6e5 d4e5 d8c8 a1c1 f8f5 e5g3 c8c6 g3g4 e8f6 g4e2 f6d5 c3e4 c6d7 e1d1 a8f8 e4g3 f5f6 e3g5 f6g6 g5e7 d7e7');
INSERT INTO evaluations VALUES(53944,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(53944,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(53944,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(53944,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(53944,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(53944,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(53944,36,'cp 103','e5d4 f3d4 b8c6 b1c3 g8f6 f1c4 f8e7 e1g1 e8g8 f1e1 a7a6 c1f4 c8g4 d4c6 b7c6 d1d3 f8e8 c4b3 g4h5 a1c1 h5g6 d3c4 d8b6 e1d1 f6e4 c3e4 d6d5 c4a4 d5e4 f4e3 c6c5 d1d7 b6b5 c1d1 b5a4');
INSERT INTO evaluations VALUES(53953,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(53953,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(53953,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(53953,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(53953,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(53953,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(53953,36,'cp 99','f3d4 b8c6 b1c3 g8f6 f1c4 f8e7 e1g1 e8g8 f1e1 a7a6 h2h3 c6d4 d1d4 b7b5 c4f1 f8e8 c1f4 h7h6 e1d1 c8e6 f4d6 e7d6 d4d6 d8d6 d1d6 b5b4 c3e2 f6e4 d6d4 e6f5 f2f3 e4f6 e2g3 f5e6 d4b4 a6a5 b4a4 e8d8 b2b3 d8d2 f1b5 e6d7 b5d7 f6d7 a1c1');
INSERT INTO evaluations VALUES(53952,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(53952,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(53952,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(53952,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(53952,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(53952,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(53952,36,'cp 94','b8c6 f1c4 g8f6 e1g1 f8e7 b1c3 e8g8 f1e1 a7a6 c1f4 c8g4 d1d2 c6d4 d2d4 a8c8 c4f1 f8e8 h2h3 g4e6 a1d1 c8c6 f1e2 h7h6 d4e3 f6d7 d1d4 d7b6 e1d1 b6c4 e2c4 c6c4 f4d6 c4d4 e3d4');
INSERT INTO evaluations VALUES(20644,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(20644,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(20644,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(20644,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(20644,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(20644,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(20644,36,'cp 104','f1c4 g8f6 e1g1 f8e7 b1c3 e8g8 f1e1 a7a6 h2h3 c6d4 d1d4 b7b5 c4f1 f8e8 c1f4 h7h6 a2a4 b5a4 d4a4 e7f8 a1d1 c8e6 a4a3 d6d5 a3a4 d5d4 a4d4 d8d4 d1d4 f8c5 d4d2 c5b4 f2f3 a6a5 g1h2 e8c8 e1c1 b4c3 c1c3 c8c3 b2c3');
INSERT INTO evaluations VALUES(55836,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(55836,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(55836,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(55836,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55836,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55836,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55836,36,'cp 2','g1f3 b7b6 c1b2 c8b7 e2e3 g8f6 b1d2 f8e7 f1d3 c7c5 e1g1 b8d7 d1e2 e8g8 f3e5 c5d4 e3d4 d7e5 d4e5 f6d7 e2h5 g7g6 h5e2 d7c5 d2f3 c5d3 e2d3 a7a5 f1e1 b7a6 d3e3 a8c8 f3d4 e7c5 c2c3 d8c7 a1d1 c7e7');
INSERT INTO evaluations VALUES(55835,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(55835,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(55835,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55835,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(55835,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(55835,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(55835,36,'cp 0','b7b6 e2e3 f8b4 c2c3 b4d6 f1d3 g8f6 c3c4 e8g8 e1g1 c8b7 b1c3 b8d7 c4d5 e6d5 c3b5 d6e7 c1a3 c7c5 a1c1 f8e8 a3b2 a7a6 b5c3 e7d6 c3e2 a8c8 e2g3 g7g6 f1e1 d6f8 d1e2 b6b5 d4c5 d7c5 d3b1 f6e4');
INSERT INTO evaluations VALUES(56735,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(56735,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(56735,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(56735,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(56735,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(56735,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(56735,36,'cp 92','f5e4 f3e5 d6e5 d1h5 e8d7 h5f5 d7c6 f5e4 c6d7 e4f5 d7c6 f5f3 c6d7 d2d4 d8f6 f3g3 f6g6 d4e5 g6g3 f2g3 b8c6 c1f4 d7e8 c3d5 f8b4 c2c3 b4a5 b2b4 a5b6 a2a4 a7a5 b4b5 c6d8 e1c1 c8g4 d1d2 g8e7 d5b6 c7b6 h2h3 a8c8 c4a2 g4f5 g3g4 f5e6 d2d8 c8d8 a2e6 h8f8 f4e3 e7d5 e6d5 d8d5 c1b2 d5e5 e3b6 e5e6 b6a5 f8f2 b2b3 f2g2 h1d1 b7b6 a5b4 e6e2');
INSERT INTO evaluations VALUES(4474,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(4474,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(4474,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(4474,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(4474,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(4474,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(4474,36,'cp 62','c1g5 h7h6 g5f6 b4c3 b2c3 d8f6 e1g1 d7d6 f1e1 c6e7 f3d2 c8d7 a2a4 c7c5 e4e5 f6g6 e5d6 d7c6 g2g3 g6d6 d2e4 c6e4 e1e4 c5d4 e4d4 d6c5 d1e2 e7c6 d4d5 c5e7 e2e7 c6e7 d5d7 e7c6 d7c7 a8d8 c7b7 c6a5 b7c7 a5c4 c7c4');
INSERT INTO evaluations VALUES(10336,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(10336,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(10336,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(10336,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(10336,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(10336,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(10336,36,'cp -58','h7h6 g5f6 d8f6 a2a3 b4c3 b2c3 e8g8 f1e2 b7b6 a3a4 a7a5 f3d2 c6e7 c3c4 c8b7 e1g1 a8d8 d2b1 f6d6 e2f3 f7f5 b1c3 f5e4 c3e4 d6g6 f3h5 g6f5 h5g4 f5h7 d1e2 d7d6 g2g3 g8h8 g4h5 b7e4 e2e4 h7e4 d3e4 e7g8 f1e1 h8h7 h5g4 g8f6');
INSERT INTO evaluations VALUES(9829,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(9829,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(9829,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(9829,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(9829,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(9829,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(9829,36,'cp -53','g5f6 d8f6 f1e2 c6e7 e1g1 e8g8 a2a3 b4c3 b2c3 d7d6 f3d2 b7b6 e2f3 c8b7 g2g3 f6e6 c3c4 b7c6 d1e2 a8e8 c2c3 g8h8 f3g2 e6d7 d3d4 f7f6 f1e1 f6f5 d4d5 c6a4 e4f5 d7f5 g2e4 f5f7 g1g2 a4d7 f2f3');
INSERT INTO evaluations VALUES(15528,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(15528,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15528,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(15528,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(15528,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(15528,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(15528,36,'cp 19','g7g6 g2g3 f8g7 f1g2 c8g4 h2h3 g4f3 d1f3 e8g8 c1g5 c6d4 f3d1 d4e6 g5e3 c7c6 e1g1 a7a5 c3e2 a5a4 a1b1 f6d7 f2f4 g7h6 d1d2 f7f5 e4f5 f8f5 g2e4 f5f8 f4f5 h6e3 d2e3 d8g5 e3g5 e6g5 f5g6 h7g6 e4g6 g5h3 g1g2 h3g5 f1f5 f8f5 g6f5 d7c5 d3d4 c5e4 d4e5 d6e5');
INSERT INTO evaluations VALUES(22457,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(22457,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(22457,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(22457,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(22457,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(22457,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(22457,36,'cp 55','f8g7 f3g5 g8h6 a2a4 f7f6 g5f3 h6f7 a4a5 d7d6 b1c3 e8g8 h2h4 g8h8 c1e3 h7h5 a5a6 b7a6 c4a6 c8a6 a1a6 c6b4 a6a4 a7a5 e1g1 f6f5 e4f5 g6f5 d3d4 e5e4 f3g5 f7g5 h4g5 d8e8 e3f4 h8h7 d1d2 c7c5 d4c5 d6c5');
INSERT INTO evaluations VALUES(6180,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(6180,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(6180,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(6180,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(6180,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(6180,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(6180,36,'cp 60','f3g5 g8h6 a2a4 f7f6 g5f3 h6f7 b1c3 d7d6 c1e3 e8g8 d1d2 c8g4 h2h4 d8d7 f3h2 g4e6 c4e6 d7e6 c3d5 a8c8 h4h5 f6f5 h5g6 h7g6 e3g5 f5f4 h2f3 f7g5 f3g5 e6g4 g5h3 g4d7 e1c1 c6d4 h3g5 c7c6 d5c3 d7d8 g5f3 d4e6 c1b1');
INSERT INTO evaluations VALUES(6175,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(6175,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(6175,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(6175,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(6175,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6175,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(6175,36,'cp 64','g8h6 a2a4 f7f6 g5f3 h6f7 b1c3 d7d6 c1e3 e8g8 a4a5 g8h8 h2h4 h7h5 d1d2 a7a6 c3d5 a8b8 e1g1 h8h7 c4b3 c8d7 c2c3 c6e7 d3d4 e7d5 b3d5 d8e7 d4e5 f7e5 f3d4 c7c6 d5a2 f8e8 e3f4 c6c5 d4e2 b7b5 f2f3');
INSERT INTO evaluations VALUES(9699,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(9699,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(9699,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(9699,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(9699,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(9699,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(9699,36,'cp 77','a2a4 f7f6 g5f3 h6f7 b1c3 d7d6 a4a5 e8g8 c1e3 g8h8 h2h4 h7h5 d1d2 a7a6 e1g1 h8h7 c3d5 a8b8 b2b4 c6e7 d3d4 c8g4 f3h2 g4d7 c4b3 e7d5 b3d5 c7c6 d5b3 d8e7 d4e5 e7e5 b3f7 f8f7 a1d1 e5e4 d2d6 b8d8 d6g3 e4b4');
INSERT INTO evaluations VALUES(23230,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(23230,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(23230,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(23230,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(23230,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(23230,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(23230,36,'cp 117','d7d5 f1d3 c8e6 e1g1 d8d7 b2b4 a7a6 e4d5 e6d5 b4b5 a6b5 d3b5 f8d6 d4e5 d6e5 f3e5 f6e5 f1e1 g8e7 e1e5 d7d6 e5e1 e8g8 c3c4 d5f7 d1d6 c7d6 c1b2 e7f5 e1d1 f8c8 a2a4 c6e5 b1d2 e5d3 a1b1 d3b2 b1b2 f5d4 d2e4 d4b5 c4b5 a8a4');
INSERT INTO evaluations VALUES(24569,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(24569,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24569,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(24569,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24569,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24569,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(24569,36,'cp 8','e1g1 e7e6 d2d3 f8e7 d1e2 a7a6 a2a4 e8g8 c1f4 d7b8 e4e5 d6e5 f3e5 e7d6 a4a5 d8c7 f4g3 b8c6 e5c6 c7c6 g3d6 c6d6 f1e1 a8c8 a1a4 f6d5 c3d5 e6d5 g2g3 d6d7 a4a1 f8e8 e2d2 h7h6 e1e8 c8e8 a1e1 e8e1 d2e1 c5c4 d3c4 d5c4 e1e4 d7c7');
INSERT INTO evaluations VALUES(24568,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24568,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(24568,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(24568,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24568,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24568,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(24568,36,'cp 12','e7e6 d2d3 f8e7 d1e2 a7a6 a2a4 e8g8 c1f4 d7b8 e4e5 d6e5 f3e5 e7d6 a4a5 d8c7 f1e1 b8c6 e5c6 d6f4 c6e5 a8d8 g2g3 f4e5 e2e5 d8d6 c3e4 f6e4 e1e4 f7f6 e5c3 f8c8 e4c4 d6c6 a1e1 g8f7 c4e4 c7d7 h2h4 d7d5 g1f1 c8d8 c3c4 d8d6 c4d5');
INSERT INTO evaluations VALUES(24516,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(24516,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(24516,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24516,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(24516,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(24516,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(24516,36,'cp 12','d2d3 f8e7 d1e2 a7a6 a2a4 d8c7 c1f4 d7b8 e4e5 d6e5 f4e5 c7d8 d3d4 c5d4 f3d4 e8g8 f1d1 d8c8 h2h3 b8c6 e5g3 f8d8 d4f3 d8d1 a1d1 c8e8 c3e4 f6e4 e2e4 a8d8 d1d3 d8d3 e4d3 c6b4 d3e4 b4c6 c2c3 e8d7 e4e2');
INSERT INTO evaluations VALUES(46745,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(46745,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(46745,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(46745,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(46745,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(46745,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(46745,36,'cp -23','f6e4 g5f4 f8g7 c3e4 d5e4 f3e5 b7b5 f2f3 c8b7 f3e4 b7e4 e2e3 a7a6 d1g4 b8d7 f1d3 e4d3 e5d3 e8g8 e1g1 c7c5 c2c3 c5c4 d3b4 e7e5 f4g5 h7h5 g5d8 h5g4 d8e7 f8b8 d4d5 e5e4 a1d1 f7f5 h2h3 g4h3 b4c6 h3g2 g1g2 b8c8 e7d6');
INSERT INTO evaluations VALUES(47105,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47105,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47105,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47105,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47105,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47105,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47105,36,'cp 0','f6d5 d3d4 c7c5 d4c5 e7e6 c2c4 d5f6 d1d8 e8d8 b1c3 f8c5 a2a3 a7a5 c1f4 c8d7 g1f3 d8e7 e1c1 b8c6 h2h3 h8d8 f4c7 d8c8 c7h2 c8d8');
INSERT INTO evaluations VALUES(47460,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47460,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47460,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47460,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47460,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47460,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47460,36,'cp 0','d3d4 c7c5 d4c5 e7e6 c2c4 d5f6 d1d8 e8d8 a2a3 a7a5 g1f3 f8c5 b1c3 d8e7 c1f4 c8d7 e1c1 b8c6 h2h3 h8d8 f4c7 d8c8 c7h2 c8d8');
INSERT INTO evaluations VALUES(47493,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(47493,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(47493,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(47493,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(47493,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(47493,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(47493,36,'cp -65','c3e4 d5e4 c1e3 e7e6 d1d2 b7b6 e1c1 c8b7 g1e2 b8c6 e2c3 f8b4 a2a3 b4c3 d2c3 e8g8 e3f4 d8d7 h2h4 a8c8 f1b5 a7a6 b5c6 b7c6 b2b3 f7f6 c1b2 c6d5 c3e3 b6b5 h1g1 a6a5 g2g4 a5a4 b3b4 d7c6 g4g5 f6f5');
INSERT INTO evaluations VALUES(47488,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(47488,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(47488,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(47488,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(47488,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(47488,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(47488,36,'cp -79','d5e4 c1e3 e7e6 g1e2 b8d7 e2c3 f7f5 d1h5 g7g6 h5e2 d7f6 f2f3 e4f3 e2f3 f6d5 c3d5 d8d5 f3d5 e6d5 e3f4 c7c6 h2h4 e8f7 c2c3 c8d7 g2g3 f8e7 e1d2 a7a5 a2a4 e7f6 a1e1 a8c8 f1g2 c8e8 g2f3 e8e6 e1e6 d7e6');
INSERT INTO evaluations VALUES(47539,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(47539,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(47539,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(47539,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(47539,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(47539,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(47539,36,'cp -55','c1e3 e7e6 f2f3 c8d7 d1e2 d7c6 e1c1 b8d7 g2g4 d7b6 f1g2 e4f3 g2f3 d8d6 c1b1 e8c8 c2c4 d6d7 b2b3 f8a3 h2h4 h7h6 h4h5 h8f8 g4g5 h6g5 e3g5 f7f6 g5e3 d8e8 e2g2 a7a5 g1e2 c6f3 g2f3 a5a4 h1f1 e6e5 d4e5 f6e5 d1d7 f8f3 f1f3 c8d7');
INSERT INTO evaluations VALUES(55091,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(55091,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(55091,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(55091,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(55091,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(55091,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(55091,36,'cp -68','h7h5 f1g2 b8c6 d2d4 h5h4 c2c3 g8f6 g1h3 f6e4 h3f2 f7f5 b1d2 h4g3 d2e4 d5e4 h2g3 h8h1 g2h1 e7e6 e2e3 d8f6 d1h5 f6f7 h5f7 e8f7 c1d2 b7b6 h1g2 c8a6 g2f1 a6f1 e1f1 f8d6 g3g4 a8h8 f1g2 c6a5 b2b3 c7c5 a1h1 h8h1');
INSERT INTO evaluations VALUES(55463,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(55463,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(55463,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(55463,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(55463,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(55463,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(55463,36,'cp 37','c2c4 g8f6 b1c3 f8d6 f4d6 d8d6 a1c1 e8g8 c4d5 f6d5 a2a3 d5c3 c1c3 b8d7 g1f3 f8d8 d1c2 c7c6 f1e2 e6e5 e1g1 e5d4 f3d4 d7f6 h2h3 d6e7 f1c1 c8e6 c3d3 d8d7 d4e6 e7e6 d3d7 e6d7 c1d1 d7e7');
INSERT INTO evaluations VALUES(998,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(998,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(998,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(998,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(998,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(998,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(998,36,'cp -62','d2d3 c8f5 f1e2 e8c8 e1g1 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 c5e3 f2e3 f5f6 c2c3 h8e8 a1d1 d8d1 f1d1 f6e7 d1e1 e7e4 f3g5 e4g6 g5f3 a7a6 e2c4 f7f6 h2h3 h7h6 e3e4 c8b8 c4d5 g6g3 e1d1 g3f4 d5f7 e8e7 f7f8 b8a7 d1e1');
INSERT INTO evaluations VALUES(997,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(997,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(997,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(997,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(997,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(997,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(997,36,'cp -75','c8f5 f1e2 e8c8 e1g1 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 h8e8 a1d1 d8d1 f1d1 c5e3 f2e3 c6e5 e3e4 e5f3 g2f3 e8e6 e2e3 f5h3 e3e2 h3h5 e2e3 b7b6 e3d3 e6g6 g1h1 g6d6 d3e2 d6d1 e2d1 h5g5 b2b3 g5c5 h1g2 c5a3 d1d5 a3e7 d5a8 c8d7 a8d5 d7e8');
INSERT INTO evaluations VALUES(4357,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(4357,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(4357,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4357,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(4357,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4357,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4357,36,'cp 10','b1c3 d7d6 c3a4 c5b6 e1g1 h7h6 f1e1 c8e6 c4e6 f7e6 a4b6 a7b6 c2c3 d8e8 c1e3 e8g6 a3a4 f8f7 b2b4 a8f8 f3h4 g6g4 d1g4 f6g4 h4f3 g4e3 f2e3 g7g5 e1f1 g5g4 f3d2 h6h5 f1f7 f8f7 a1f1 c6e7 d2c4 e7g6 f1f7 g8f7');
INSERT INTO evaluations VALUES(11914,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(11914,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(11914,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(11914,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(11914,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(11914,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(11914,36,'cp 196','f8c5 d4e5 d8e7 g5f6 g7f6 e5c7 e7e5 c7e5 f6e5 c3d5 e8f8 h2h4 a7a5 a1d1 d7d6 a2a3 a5a4 h4h5 h8g8 d5f6 g8g7 g2g3 c8e6 h1h4 h7h6 c2c3 e6b3 d1d2 f8e7 f6d5 e7f8 f1e2 g7g5 g3g4 b3d5 d2d5 f8e7 e1f1 e7f6 h4h1 f6e7 f2f3 c5e3 f1g2');
INSERT INTO evaluations VALUES(7575,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(7575,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(7575,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(7575,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(7575,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(7575,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(7575,36,'cp 200','d4e5 d8e7 g5f6 g7f6 e5c7 e7e5 c7e5 f6e5 c3d5 e8f8 h2h4 a7a5 a1d1 c8b7 a2a3 b7c6 h1h3 d7d6 h3c3 c6d5 d1d5 h8g8 g2g3 g8g4 f1g2 b6b5 c3b3 b5b4 a3a4 h7h6 e1e2 f8g7 e2f3 g4g6 g2h3 g6f6 h3f5 c5d4 d5b5');
INSERT INTO evaluations VALUES(7574,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(7574,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(7574,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(7574,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(7574,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(7574,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(7574,36,'cp 194','d8e7 g5f6 g7f6 e5c7 e7e5 c7e5 f6e5 c3d5 e8f8 a1d1 a7a5 h2h4 d7d6 a2a3 a5a4 g2g3 f7f5 e4f5 c8f5 f1d3 e5e4 d3c4 f8g7 h4h5 h8f8 h1h4 g7h6 d5b6 f5e6 h4e4 e6c4 b6c4 c5f2 e1e2 f2g3 e4e6 h6h5 d1d5 h5h4 e6h6 h4g4 c4e3 g4f4 d5d4 f4g5 h6h7 g5g6 h7h3');
INSERT INTO evaluations VALUES(15421,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(15421,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(15421,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(15421,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(15421,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(15421,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(15421,36,'cp 100','d4e5 f6e4 d1d5 e4c5 c1e3 a7a6 b5c6 b7c6 d5c6 c8d7 c6d5 d7e6 d5d4 d6e5 d4e5 f7f6 e5c3 d8d5 b1d2 f8d6 e1g1 e8g8 f1d1 c5a4 c3b3 a4c5 b3d5 e6d5 d2b3 c5b3 a2b3 d5e4 f3d4 f8e8 c2c3 c7c5 d4c6 d6f8 c6a5 a8d8 h2h3 d8d1 a1d1');
INSERT INTO evaluations VALUES(15415,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(15415,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(15415,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(15415,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(15415,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(15415,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(15415,36,'cp 110','f6e4 d1d5 e4c5 c1e3 a7a6 b5c6 b7c6 d5c6 c8d7 c6d5 d7e6 d5d4 d6e5 d4e5 f7f6 e5d4 d8d4 f3d4 e6d5 e1g1 d5f7 d4c6 c5e6 f1d1 e6d8 c6a5 a8b8 a5b3 f8b4 c2c3 b4e7 d1e1 a6a5 e3a7 b8b7 a7c5 d8e6 c5a3 a5a4 b3d4 b7b6 b2b3 c7c5 d4f5 b6b7 f5e7 b7e7 f2f3 e8g8');
INSERT INTO evaluations VALUES(16555,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(16555,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(16555,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(16555,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(16555,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(16555,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(16555,36,'cp -82','c2c4 c6d4 f3d1 f8c5 f1e2 c7c6 g1f3 d7d5 f3d4 c5d4 b1d2 e8g8 e1g1 d5e4 d3e4 c6c5 d1c2 b7b6 d2f3 c8b7 e2d3 h7h6 f3h4 b7c8 h4f3 f8e8 a1b1 c8b7 f3d4 e5d4 f2f3 d8d6 a2a3 f6d7 b2b4 a7a5 c1d2 d7e5 b4c5 d6c5');
INSERT INTO evaluations VALUES(20020,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(20020,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(20020,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(20020,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(20020,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(20020,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(20020,36,'cp -14','g8f6 c2c3 g7g6 f1g2 e7e5 e1g1 f8g7 c3c4 e8g8 b1c3 a8b8 a1b1 a7a6 a2a4 c8g4 h2h3 g4d7 c1g5 h7h6 g5d2 c6b4 b2b3 b7b6 d2e3 g8h7 g1h2 b6b5 a4b5 a6b5 b1a1 f6e8 c3d5 b4c6 a1a6 b5c4 b3c4 e8c7 a6b6 b8b6');
INSERT INTO evaluations VALUES(21953,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(21953,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(21953,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(21953,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(21953,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(21953,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(21953,36,'cp 24','d2d4 e5d4 f3d4 c6d4 d1d4 d8f6 c1b2 f6d4 b2d4 g8e7 d4b2 c8d7 h2h4 h7h5 b1c3 e8c8 a1d1 e7g6 f2f3 f8e7 g2g3 g6e5 e1f2 e7f6 f1e2 h8e8 c3d5 e5d3 e2d3 f6b2 c2c3 c8b8 d1d2 b2a3 h1e1 c7c6 d5e3');
INSERT INTO evaluations VALUES(39519,30,'cp -154',NULL);
INSERT INTO evaluations VALUES(39519,31,'cp -184',NULL);
INSERT INTO evaluations VALUES(39519,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(39519,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(39519,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(39519,35,'cp -161',NULL);
INSERT INTO evaluations VALUES(39519,36,'cp -181','g8f6 b1c3 e8g8 f1e2 f6h5 e1g1 h5g3 f1e1 d7d6 d2d4 e7h4 c1f4 g3h5 f3h4 h5f4 h4f3 c8g4 d1d2 d8f6 a1d1 b8d7 f3h2 f4e2 c3e2 g4e2 d2e2 f8e8 h2f3 f6e6 e2b5 b7b6 e4e5 h7h6 c2c4 e6f5 e5d6 f5b5 c4b5 c7d6 e1e8 a8e8');
INSERT INTO evaluations VALUES(43049,30,'cp -165',NULL);
INSERT INTO evaluations VALUES(43049,31,'cp -160',NULL);
INSERT INTO evaluations VALUES(43049,32,'cp -157',NULL);
INSERT INTO evaluations VALUES(43049,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(43049,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(43049,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(43049,36,'cp -165','b1c3 e8g8 f1e2 f6h5 d2d4 h5g3 h1h2 e7b4 e2d3 d7d5 e4e5 f7f6 c1f4 g3e4 f4d2 b4c3 d2c3 c7c5 d3e4 d5e4 f3g1 d8d5 d4c5 d5c5 d1d4 c5d4 c3d4 f6e5 d4e5 b8c6 e5c3 b7b5 a2a3 e4e3 g1f3 c8g4 h4h5 h7h6 h2h4 g4f3 g2f3 f8f3');
INSERT INTO evaluations VALUES(47667,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(47667,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(47667,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(47667,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(47667,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(47667,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(47667,36,'cp 65','g5f6 d8f6 a2a3 c7c6 g1f3 b8d7 f1d3 f6d8 d1e2 f8e7 h2h4 a7a6 e1c1 d8a5 c3a2 a5c7 g2g4 d5e4 d3e4 c7f4 c1b1 d7f6 a2c3 f6g4 c3a4 c8d7 e4d3 e8c8 d3a6 c8b8 a6d3 f4c7 h1e1 h6h5 a4c5 e7c5 d4c5');
INSERT INTO evaluations VALUES(47661,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(47661,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(47661,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(47661,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(47661,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(47661,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(47661,36,'cp 53','d8f6 a2a3 c7c6 g1f3 b8d7 h2h4 f6d8 f1d3 f8e7 d1e2 a7a5 e1f1 d7f6 e4e5 f6g4 g2g3 h6h5 f1g2 a5a4 c3d1 c6c5 c2c3 d8b6 d1e3 c8d7 h1b1 g4e3 e2e3 d7b5 d3b5 b6b5 f3g5 b5c4 b1c1 e8g8 a1b1 e7g5 h4g5 c5d4');
INSERT INTO evaluations VALUES(33860,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(33860,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(33860,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(33860,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(33860,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(33860,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(33860,36,'cp 45','a2a3 c7c6 g1f3 b8d7 h2h4 f6d8 f1d3 f8e7 d1e2 a7a5 e1f1 d7f6 e4e5 f6g4 g2g3 h6h5 f1g2 c8d7 c3d1 c6c5 c2c3 d8b6 a1b1 d7a4 b1c1 a4d7 h1e1 a5a4 c1b1 e8c8 e2d2 c8b8 d2f4 h8f8 d1e3 g4e3 f4e3 d8c8 d3e2 c5d4 f3d4 g7g6 e1c1 d7c6 e3d2 e7c5');
INSERT INTO evaluations VALUES(49779,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(49779,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49779,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(49779,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(49779,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(49779,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(49779,36,'cp 0','b8c6 f1e2 f8c5 b1c3 d7d5 e4e5 f6d7 g1f3 c6d4 c3d5 e8g8 b2b4 d4f3 e2f3 c5d4 c2c3 d4e5 e1g1 c7c6 d3d4 e5f4 d5f4 a7a5 d4d5 a5b4 d5c6 b7c6 c3b4 c8a6 f1e1 d8b6 g1h1 d7f6 a2a3 f8d8 d1b3 b6d4 b3e3 a6b5 h1g1 d4b2 e1b1 b2d4');
INSERT INTO evaluations VALUES(16720,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16720,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(16720,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16720,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16720,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(16720,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(16720,36,'cp -3','f1e2 f8c5 b1c3 d7d5 e4e5 f6d7 g1f3 c6d4 c3d5 e8g8 b2b4 d4f3 e2f3 c5d4 c2c3 d4e5 e1g1 c7c6 d3d4 e5f4 d5f4 a7a5 d4d5 a5b4 c3b4 d7e5 a2a3 c8d7 f3e4 f8e8 d5c6 d7c6 e4c6 d8b6 g1h1 b6c6 f4d5 e5c4 d1f3 c4e5');
INSERT INTO evaluations VALUES(54856,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54856,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54856,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(54856,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54856,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54856,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(54856,36,'cp 7','g8f6 c1f4 f8g7 e2e3 e8g8 f1e2 c7c5 e1g1 c5d4 e3d4 b8c6 f3e5 c8f5 f1e1 f6d7 c3d5 c6e5 d4e5 e7e6 d5e3 g7e5 f4e5 d7e5 e3f5 e6f5 d1d8 f8d8 a1d1 a8c8 d1d8 c8d8 e1d1 d8d1 e2d1 g8f8 g1f1 f8e7 f1e2 e7d6 h2h3 b7b5 c2c3 e5c4 b2b3');
INSERT INTO evaluations VALUES(56464,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(56464,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(56464,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(56464,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(56464,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(56464,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(56464,36,'cp -48','f8d6 g1f3 g8f6 c2c4 c7c6 a2a3 e8g8 b2b4 f8e8 c1b2 b8d7 b1d2 a7a5 c4c5 d6c7 e1g1 d7f8 f1e1 f8g6 e3e4 c8g4 e2f1 g6h4 h2h3 h4f3 d2f3 g4f3 d1f3 d5d4 g2g3 d8e7 e1b1 f6d7 g1h1 d7f8 f3d1 g7g6');
INSERT INTO evaluations VALUES(56617,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(56617,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(56617,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56617,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(56617,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(56617,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56617,36,'cp 0','c3e4 f7f5 e4g3 e5e4 f3e5 d8d6 d2d4 b8c6 e5c6 d6c6 c2c4 g8f6 f1e2 f8d6 g3h5 e8g8 d1b3 f6h5 c4c5 g8h8 c5d6 h5f6 d6c7 c6c7 c1d2 f8d8 e1g1 c7d6 d2b4 d6d7 b4d2');
INSERT INTO evaluations VALUES(7211,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(7211,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(7211,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(7211,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(7211,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(7211,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(7211,36,'cp 163','e5c4 d6c5 d2d3 f6g4 c4e3 c5e3 c1e3 g4e3 f2e3 d8h4 e1d2 c8e6 d1f3 e8c8 a1f1 h7h5 h2h3 h4e7 f1f2 h5h4 d2c1 h8h6 f3e2 f7f6 f2f4 b7b6 e2f2 d8h8 c1b1 c8b8 b2b3 b8b7 d3d4 h6h5 h1d1 h5g5 b1b2 h8h6 d1g1');
INSERT INTO evaluations VALUES(12335,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(12335,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(12335,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(12335,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(12335,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(12335,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(12335,36,'cp -68','h7h6 g5d2 a7a6 d1e2 f8e7 c4b3 d7d5 c2c4 d5c4 d3c4 d8c7 b1c3 c6d4 e2e3 e8g8 f3d4 c5d4 e3d4 f8d8 d4e3 b7b5 e3e2 c8b7 f2f3 d8d2 e2d2 b5c4 b3a4 a8d8 d2e2 f6h5 g2g3 e7c5 a1d1 c5d4 d1d4 d8d4 e1g1 h5f6 a4c2 d4d8 e2e3');
INSERT INTO evaluations VALUES(12270,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(12270,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(12270,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(12270,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(12270,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(12270,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(12270,36,'cp -60','g5d2 a7a6 c4b3 d7d5 d1e2 f8e7 e1g1 b7b5 c2c3 c5c4 e4d5 e6d5 d3c4 d5c4 b3c2 e8g8 a2a4 c8b7 f1e1 e7c5 d2e3 f8e8 e2d2 c6e5 f3e5 e8e5 e3c5 d8d5 c2e4 e5e4 e1e4 d5e4 f2f3 e4f5 c5f2 f6d5 a4b5 a6b5 a1a8 b7a8 d2d1 d5f4 f2g3 f4d3');
INSERT INTO evaluations VALUES(12799,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(12799,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(12799,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(12799,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(12799,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(12799,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(12799,36,'cp -13','b1c3 e7e6 c1f4 a7a6 f1e2 d7d6 g1f3 f8e7 a2a3 f6d7 f4e3 d7e5 e1g1 e5f3 e2f3 e8g8 a1b1 c6d4 b2b4 c8d7 f3g4 b7b5 c3e2 d4c6 f2f4 c5b4 a3b4 a6a5 b4a5 b5c4 a5a6 a8a6 d3c4 a6a3 e3c1 a3a5 g1h1 d8c7 e2c3 f8b8 c1e3 b8b1');
INSERT INTO evaluations VALUES(12797,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(12797,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(12797,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(12797,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(12797,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(12797,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(12797,36,'cp -24','e7e6 c1f4 a7a6 f1e2 d7d6 g1f3 f8e7 e1g1 e8g8 a2a3 f6d7 a1b1 c6d4 b2b4 b7b6 d1d2 c8b7 f4g3 d4e2 d2e2 d8c7 f1c1 d7e5 b4c5 b6c5 f3e5 d6e5 b1b2 b7c6 c1b1 a8b8 h2h3 h7h6 b2b8 f8b8 b1b8 c7b8 c3d1 e7d6');
INSERT INTO evaluations VALUES(20379,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(20379,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(20379,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(20379,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(20379,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(20379,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(20379,36,'cp 190','f1b5 c8d7 e4d5 e6d5 c3d5 g8e7 b5c6 b7c6 d5e3 e7g6 c1d2 f8d6 d1f3 e8g8 e1c1 c6c5 d4e2 d8c8 d2c3 d7c6 f3h5 d6c7 e2g3 c7f4 c1b1 f4e3 f2e3 c6d7 h2h3 c8c6 g3f5 d7f5 h5f5 a8e8 f5d5 c6d5 d1d5 f7f6 d5c5');
INSERT INTO evaluations VALUES(20157,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(20157,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(20157,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(20157,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(20157,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(20157,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(20157,36,'cp 218','c8d7 e4d5 e6d5 c3d5 g8e7 b5c6 b7c6 d5e3 e7g6 c1d2 f8d6 d1f3 e8g8 e1c1 c6c5 d4e2 d8c8 d2c3 d7c6 f3h5 d6c7 e2g3 c7f4 c1b1 f4e3 f2e3 c6d7 h2h3 c8c6 e3e4 f7f6 h1e1 g6e7 h5f3 a8e8 b2b3 e7g6 b1b2 g6e5 c3e5 e8e5 f3d3 d7e6');
INSERT INTO evaluations VALUES(25908,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(25908,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(25908,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(25908,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(25908,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(25908,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(25908,36,'cp 221','e4d5 e6d5 c3d5 g8e7 b5c6 d7c6 d5f4 g7g6 d4c6 e7c6 d1e2 d8e7 e2e7 f8e7 c1d2 e8c8 e1c1 h8e8 h1e1 e7f6 c2c3 c6e5 c1c2 f6h4 g2g3 h4f6 h2h3 e5f3 e1e8 d8e8 d2e3 f6g5 e3a7 g5f4 g3f4 e8e2 c2b3 f3d2 b3a4 d2e4');
INSERT INTO evaluations VALUES(25464,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(25464,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(25464,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(25464,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25464,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(25464,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(25464,36,'cp 41','b1d2 f8d6 c2c3 d8e7 e1g1 f6e4 d2e4 d5e4 f3d2 a7a6 f4d6 c7d6 b5e2 d6d5 f2f3 e4f3 e2f3 f7f5 c3c4 d5c4 d2c4 e8g8 d1b3 d7e8 a1d1 a8d8 d4d5 e6d5 c4b6 e8f7 f3d5 g8h8 d5f7 e7f7 d1d8 c6d8 e3e4 f7b3 a2b3 d8c6 b6d7 f8e8 e4f5');
INSERT INTO evaluations VALUES(28048,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(28048,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(28048,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(28048,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(28048,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(28048,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(28048,36,'cp 134','b1c3 d7d6 h2h4 h7h5 d1e2 c7c5 d4d5 b8d7 e1g1 a7a6 a2a4 g8f6 c1f4 f6g4 e2d2 d7e5 f3e5 g4e5 d3e2 b7c8 a1e1 b6b5 a4b5 a6b5 c3b5 e8g8 c2c4 f7f5 e4f5 c8f5 f4g5 d8d7 f2f4 e5f7 e2d3 f7g5 h4g5');
INSERT INTO evaluations VALUES(29657,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(29657,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(29657,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(29657,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(29657,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(29657,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(29657,36,'cp 45','f2f3 g4c8 e3e4 c7c5 e4e5 f6d7 f3f4 b8c6 g1f3 c5d4 c3b5 e7e6 b5d4 c6d4 f3d4 f8e7 c2c3 d7c5 f1e2 e8g8 e1g1 c8d7 c1e3 a8c8 g1h1 a7a6 a2a4 d8c7 a4a5 g7g6 e2d3 c5d3 d1d3 c7c4 d3c2');
INSERT INTO evaluations VALUES(29659,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(29659,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29659,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(29659,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(29659,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(29659,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29659,36,'cp 35','g4c8 e3e4 c7c5 e4e5 f6d7 f3f4 e7e6 g1f3 b8c6 c1e3 a7a6 d1d2 d8a5 f1e2 b7b5 e1g1 b5b4 c3d1 f8e7 c2c3 e8g8 a2a3 b4a3 a1a3 a5c7 d4c5 d7c5 b2b4 c5e4 d2e1 c8b7 a3a1 f7f6 f3d4 c6d4 e3d4 f6e5 f4e5 a6a5 f1f8 e7f8');
INSERT INTO evaluations VALUES(30929,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(30929,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(30929,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(30929,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(30929,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(30929,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(30929,36,'cp 74','c2c4 g8f6 f2f3 g7g6 c1e3 f8g7 b1c3 b8c6 d1d2 e8g8 f1e2 c6d4 e3d4 d8a5 e1g1 a7a6 a2a3 f8c8 b2b4 a5d8 a1c1 d7e6 c3a4 a8b8 d4a7 b8a8 d2e3 f6d7 a7b6 d8f8 b6d4 a8b8 d4g7 f8g7 f3f4 d7f6');
INSERT INTO evaluations VALUES(51530,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(51530,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(51530,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(51530,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(51530,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(51530,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(51530,36,'cp 128','c4b5 a7a6 e2e3 e7e5 g1f3 e5e4 f3d4 a6b5 f1b5 g8f6 b1c3 f8c5 d4c2 e8g8 d2d4 c5d6 b5e2 c7c6 a2a4 b8a6 c1d2 a8b8 a1b1 d8c7 b2b3 d6e7 h2h3 d7d5 c3a2 h7h5 b3b4 e7d6 e1g1 f8e8 b4b5 c6b5 b1b5');
INSERT INTO evaluations VALUES(51527,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(51527,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(51527,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(51527,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(51527,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(51527,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(51527,36,'cp 112','a7a6 e2e4 c8b7 b1c3 g8f6 e4e5 f6d5 g1f3 e7e6 d2d4 f8e7 c1d2 e8g8 f1c4 d5c3 d2c3 a6b5 c4b5 b8a6 e1g1 a6b4 a2a4 b4d5 c3d2 b7a6 b5a6 a8a6 d1c2 d8a8 f1b1 f8b8 b2b3 e7b4 f3g5 b4d2 c2d2 h7h6 g5e4 d7d6 e5d6 c7d6 e4c3 d5c3');
INSERT INTO evaluations VALUES(51161,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(51161,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(51161,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(51161,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(51161,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(51161,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(51161,36,'cp 122','e2e3 e7e5 b1c3 g8f6 g1f3 e5e4 f3d4 a6b5 f1b5 f8c5 d4c2 e8g8 d2d4 c5d6 a2a4 c7c6 b5e2 b8a6 e1g1 f8e8 f2f4 a6b4 c2b4 d6b4 c3a2 f6d5 a2b4 d5b4 c1d2 b4d3 d2c3 d7d5 d1d2 c8a6 a4a5 a6b5 h2h3 h7h6 b2b4 d8d6');
INSERT INTO evaluations VALUES(55860,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55860,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(55860,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55860,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55860,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55860,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(55860,36,'cp 0','c7c5 c2c4 b8c6 e2e3 c5d4 e3d4 f8b4 b1c3 g8f6 c4d5 f6d5 c1d2 d5e7 d2e3 e7d5');
INSERT INTO evaluations VALUES(54217,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54217,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54217,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54217,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54217,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54217,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54217,36,'cp 0','c2c4 b8c6 e2e3 c5d4 e3d4 f8b4 b1c3 g8f6 c4d5 f6d5 c1d2 d5e7 d2e3 e7d5');
INSERT INTO evaluations VALUES(56185,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(56185,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(56185,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(56185,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(56185,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(56185,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(56185,36,'cp 81','c4d5 f7f5 b1c3 g8f6 f2f3 f8d6 g1h3 b8d7 f1c4 d7b6 c4b3 d8e7 a2a4 a7a5 e1g1 c8d7 f3e4 f5e4 h3f4 e8c8 c1d2 c8b8 d1e2 d8e8 d2e1 g7g5 f4e6 b6d5 c3d5 f6d5 b3d5 d7e6 d5e6 e7e6 e1a5 e8f8 a5d2 g5g4 f1f8 h8f8 d2e1 e6b3 e1g3 d6g3 h2g3 b3d3 e2d3');
INSERT INTO evaluations VALUES(56135,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(56135,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(56135,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(56135,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(56135,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(56135,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(56135,36,'cp 67','f7f5 b1c3 g8f6 g1e2 b8d7 e2g3 d7b6 f2f3 b6d5 f3e4 d5c3 b2c3 f5e4 f1e2 f8d6 d1b3 d8e7 e1g1 h7h5 a2a4 c7c6 c1a3 d6a3 b3a3 h5h4 g3f5 e7a3 a1a3 c8f5 f1f5 e8f7 a4a5 b7b5 f5c5 h8c8 a3a1 f6d5 e2g4 c8c7 a1f1 f7e7 g4f5 d5e3 f1e1 e3f5 e1e4 e7d8 c5f5');
INSERT INTO evaluations VALUES(2804,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(2804,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(2804,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(2804,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(2804,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(2804,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2804,36,'cp 26','a2a4 c8b7 f1e1 d7d6 c1d2 b5b4 c2c3 a8b8 c3b4 b7c8 b4b5 a6b5 a4b5 b8b5 h2h3 b5b8 b1a3 c8d7 a3c4 h7h6 b3a4 c6b4 d2c3 c7c5 a4d7 d8d7 a1a5 b4c6 a5a6 d7b7 d1a4 b8a8 e1a1 a8a6 a4a6 f8b8 a6a4 f6h5 g2g3 h5f6');
INSERT INTO evaluations VALUES(9008,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(9008,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(9008,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(9008,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(9008,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(9008,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(9008,36,'cp 47','h2g3 h7h6 f1b5 c8d7 b1d2 e8g8 e1g1 d8e7 f1e1 c6b8 b5d3 c7c5 e3e4 d5e4 d2e4 c5d4 f3d4 b8c6 e4f6 e7f6 d4f3 f8d8 d1e2 d7e8 d3c2 d8d6 a1d1 d6d1 e1d1 a8d8 d1e1 b7b6 a2a3 c6e7 e2c4 e7c6 c4a6 g8f8 a6e2 d8d6 e2e3 f6d8 g3g4 g7g5 c2e4');
INSERT INTO evaluations VALUES(12362,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(12362,31,'cp -102',NULL);
INSERT INTO evaluations VALUES(12362,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(12362,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(12362,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(12362,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(12362,36,'cp -91','c4b5 d7d5 e4e5 f6d7 f3e2 d8a5 b5c6 b7c6 d2d3 c8a6 g1f3 a5b5 c3c4 d5c4 b1c3 b5b4 d3d4 c5d4 f3d4 d7c5 e1g1 b4b7 e2f3 c5d3 f3c6 b7c6 d4c6 e8d7 c6d4 d3e5 f1d1 e5d3 b2b3 a8c8 b3c4 a6c4 d4b5 f8c5 c1e3 c5e3 f2e3');
INSERT INTO evaluations VALUES(26771,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(26771,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(26771,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(26771,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(26771,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(26771,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(26771,36,'cp 122','d2d4 e5d4 f3d4 f8e7 d1d3 g8f6 f2f3 f6d7 c1e3 e7f6 e1c1 d7c5 d3c4 f6d4 e3d4 c5e6 d4e3 d8h4 c3d5 c6d5 d1d5 e8g8 c1b1 b7b6 a2a4 f8e8 b2b3 h4e7 e3c1 f7f6 c1b2 e7f7 h1d1 e8e7 c4e2');
INSERT INTO evaluations VALUES(32221,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(32221,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(32221,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(32221,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(32221,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(32221,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(32221,36,'cp 23','e6b3 a2b3 c7c5 b1c3 b8c6 d2d3 f8e7 e1g1 g8f6 c1g5 e8g8 g5f6 e7f6 c3d5 g7g6 d1d2 f6g7 a1a2 a7a5 c2c3 c6e7 d5e7 d8e7 f1a1 b7b6 b3b4 c5b4 c3b4 e7b7 b4a5 a8a5 a2a5 b6a5 a1a5 f8b8 h2h4 b7b2 g2g3 h7h5 d2b2 b8b2 a5a8 g7f8');
INSERT INTO evaluations VALUES(32146,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(32146,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(32146,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(32146,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(32146,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(32146,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(32146,36,'cp 33','a2b3 c7c5 b1c3 f8e7 d2d3 b8c6 e1g1 g8f6 c1g5 e8g8 g5f6 e7f6 c3d5 g7g6 d1d2 f6g7 a1a2 a7a5 c2c3 c6e7 d5e7 d8e7 f1a1 b7b6 b3b4 c5b4 c3b4 e7b7 b4a5 a8a5 a2a5 b6a5 a1a5 f8b8 g2g3 b7b2 g1g2 g7f8 h2h4 h7h5 d2e3 b8b7 a5a6 g8g7 f3d2');
INSERT INTO evaluations VALUES(32145,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(32145,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(32145,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32145,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(32145,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(32145,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(32145,36,'cp 26','c7c5 d2d3 b8c6 b1c3 f8e7 e1g1 g8f6 c1g5 f6g8 g5d2 g8f6 f3h4 f6e4 d3e4 e7h4 c3d5 e8g8 d1g4 c6d4 c2c3 d4e6 g2g3 h4e7 h2h4 g8h8 a1d1 b7b5 d2e3 a7a5 f2f4 e5f4 e3f4 e6f4 f1f4 b5b4 d1f1 d8e8 d5c7');
INSERT INTO evaluations VALUES(32512,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(32512,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(32512,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(32512,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(32512,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(32512,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(32512,36,'cp 68','d7c6 f1d3 b8d7 c2c3 h7h6 e1g1 g8f6 e4d2 b7b6 d2c4 f8e7 c4e5 c6b7 d1e2 e8g8 f1d1 a8c8 c1f4 f6h5 f4g3 e7d6 d3b5 c7c6 b5d3 h5g3 h2g3 d8e7 e5d7 e7d7 b2b4 d7c7 a2a4 f8d8 a4a5');
INSERT INTO evaluations VALUES(32061,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(32061,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(32061,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(32061,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(32061,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(32061,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(32061,36,'cp 66','f1d3 b8d7 c2c3 g8f6 e4f6 d7f6 e1g1 c6f3 d1f3 d8d5 f3d1 f8d6 c1e3 c7c6 c3c4 d5h5 d1h5 f6h5 a1b1 h5f6 b2b4 f6g4 g2g3 a8d8 c4c5 d6e7 f1d1 g4e3 f2e3 a7a6 a2a4 h7h5 h2h4 g7g6 b4b5 a6b5 a4b5 e8g8 d3e4');
INSERT INTO evaluations VALUES(32060,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(32060,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(32060,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(32060,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(32060,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(32060,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(32060,36,'cp 74','b8d7 c2c3 g8f6 e4f6 d7f6 e1g1 c6f3 d1f3 d8d5 f3d1 a7a5 c1e3 c7c6 g2g3 f8d6 c3c4 d5h5 d1h5 f6h5 d3e2 h5f6 a1d1 e8g8 g3g4 f8d8 g1g2 g7g6 h2h4 d6e7 e2f3 h7h5 g4g5 f6e8 d4d5 e8d6 d5c6 b7c6 f3c6 a8b8 c4c5 d6f5 d1d8 b8d8');
INSERT INTO evaluations VALUES(27296,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(27296,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(27296,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(27296,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(27296,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(27296,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27296,36,'cp 76','c2c3 g8f6 e4f6 d7f6 e1g1 c6f3 d1f3 d8d5 f3d1 f8d6 c1e3 c7c6 c3c4 d5a5 a2a3 e8g8 g2g3 f8d8 b2b4 a5c7 a1b1 a7a5 b4a5 a8a5 d4d5 d6c5 e3c5 a5c5 d5e6 f7e6 d1b3 c5a5 d3c2 d8a8 b3b7 c7b7 b1b7 a5a3 b7c7 c6c5 f1d1');
INSERT INTO evaluations VALUES(48810,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(48810,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(48810,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(48810,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(48810,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(48810,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(48810,36,'cp 69','e4d5 e5d4 f1b5 c7c6 d5c6 b7c6 b5e2 f8b4 b1d2 b4e7 e1g1 e8g8 f3e5 c8e6 d2c4 c6c5 c2c3 d4c3 e2f3 e6d5 b2c3 d5f3 d1f3 d8d5 f1d1 d5f3 g2f3 h7h6 c1f4 f8c8 a1b1 b8c6 e5c6 c8c6 b1b7 c6e6 a2a4 e7f8 a4a5 e6e7 b7b5 a8c8 f4e3 e7e6 g1g2 f6h5');
INSERT INTO evaluations VALUES(49782,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(49782,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(49782,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(49782,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(49782,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(49782,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(49782,36,'cp -69','b8c6 c1f4 d7d5 b1d2 f8c5 c2c3 e8g8 d1c2 h7h6 e1c1 f6h5 g2g3 c8g4 h2h3 g4f3 d2f3 h5f4 g3f4 c5e3 c1b1 d5d4 f4f5 f8e8 h3h4 b7b5 h1h3 b5b4 c3b4 c6b4 c2g2 a8b8 f3d2 d8f6 d2c4 c7c5 b2b3 b4c6 c4e3 d4e3 h3e3 c6d4 e3g3 g8h8 d1c1 d4b3 a2b3');
INSERT INTO evaluations VALUES(16722,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(16722,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(16722,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(16722,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(16722,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(16722,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(16722,36,'cp -62','c1f4 d7d5 b1d2 f8c5 c2c3 e8g8 d1c2 h7h6 e1c1 f6h5 g2g3 d5e4 d3e4 d8e7 d1e1 c5f2 e1e2 f2b6 e2e1 c8e6 c1b1 a8d8 f1c4 h5f6 c4e6 e7e6 h2h3 f6d7 g3g4 d7c5 f4g3 f8e8 h1f1 a7a5 e4e5 a5a4 a2a3 d8d7 d2e4 c5d3 e1e2');
INSERT INTO evaluations VALUES(16719,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(16719,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(16719,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(16719,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(16719,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(16719,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(16719,36,'cp -70','d7d5 b1d2 f8c5 c2c3 e8g8 d1c2 h7h6 e1c1 f6h5 f4e5 c6e5 f3e5 c5d6 d3d4 c7c5 e5f3 c5d4 f3d4 h5f6 e4d5 f6d5 d2c4 d6f4 c1b1 d8c7 g2g3 f4g5 h2h4 g5e7 f1d3 c8g4 d3e2 g4e2 c2e2 f8e8 d4f5 a8d8 e2e5 c7e5 c4e5 e7f8 e5g4 h6h5 g4e3 d5e3 d1d8 e8d8 f5e3 d8d2');
INSERT INTO evaluations VALUES(51301,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(51301,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(51301,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(51301,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(51301,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(51301,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(51301,36,'cp 170','e4d5 g8f6 f1b5 c8d7 b5e2 f6d5 c2c4 d5f6 g1f3 b8c6 f3e5 c6e5 d4e5 f6g8 e2f3 e7e6 f3a8 d8a8 f2f3 g8e7 b1c3 e7f5 d1d3 f8e7 c1e3 c7c5 e1c1 d7c6 d3e2 e8g8 h1e1 f5h4 e3f4 a8b7 c3e4 c6e4 e2e4 b7e4 e1e4 h4g2');
INSERT INTO evaluations VALUES(51285,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(51285,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(51285,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(51285,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(51285,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(51285,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(51285,36,'cp 159','g8f6 f1b5 c8d7 b5e2 f6d5 c2c4 d5f6 g1f3 e7e6 f3e5 b8c6 e2f3 c6e5 d4e5 f6g8 f3a8 d8a8 e1g1 g8e7 d1d3 e7g6 b1c3 f8e7 c3b5 a8b7 f1d1 d7c6 d3g3 e8g8 c1g5 e7g5 g3g5 h7h6 g5g3 b7a6 b5c7 a6c4 a1c1 c4e2 g3e3 e2e3');
INSERT INTO evaluations VALUES(53487,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(53487,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53487,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53487,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(53487,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53487,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53487,36,'cp 14','b7b5 c4b3 g8f6 e1g1 f8e7 c3e2 e8g8 a2a4 b5b4 c1d2 c6c5 f3h4 d6d5 e2g3 b8c6 h4f5 d5e4 d3e4 c6d4 b3c4 d4f5 e4f5 c8b7 d1e2 b7d5 c4d5 d8d5 f1d1 d5c6 e2e5 e7d6 e5e2 a7a5 b2b3 d6g3 h2g3 a8e8 d2e3');
INSERT INTO evaluations VALUES(51681,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51681,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(51681,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51681,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(51681,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(51681,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(51681,36,'cp 19','c4b3 g8f6 e1g1 f8e7 c3e2 e8g8 e2g3 f8e8 d3d4 b8d7 c2c3 e7f8 a2a4 d8c7 a4b5 c6b5 b3c2 c8b7 f1e1 b5b4 c3b4 d6d5 c1d2 d5e4 f3e5 d7e5 d4e5 c7e5 d2c3 e5f4 d1d2 f4d2 c3d2 a7a6 h2h3 h6h5 c2a4 e8d8 d2c3 a8c8 a4b3');
INSERT INTO evaluations VALUES(51684,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51684,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(51684,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51684,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(51684,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51684,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(51684,36,'cp 8','g8f6 e1g1 f8e7 a2a4 b5b4 c3e2 e8g8 e2g3 f8e8 f1e1 e7f8 d3d4 b8d7 c2c3 b4c3 b2c3 d8c7 d1c2 c8a6 c1e3 c6c5 c2a2 c5c4 b3c2 g7g6 h2h3 a8b8 a1b1 b8b1 e1b1 e8b8 b1e1 f8g7 a2a1 e5d4 c3d4 c4c3 a4a5 b8b2 a1a4 a6b5');
INSERT INTO evaluations VALUES(53917,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(53917,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(53917,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(53917,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(53917,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(53917,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(53917,36,'cp 87','g7g6 g1f3 f8g7 b1c3 a7a6 c1f4 g8f6 d1a4 f6d7 a4c2 d8e7 a2a4 e8g8 f1e2 d7e5 f3e5 g7e5 f4e5 e7e5 e1g1 f7f5 f2f4 e5d4 g1h1 f5e4 a1d1 d4g7 c3e4 c8f5 g2g4 f5e4 c2e4 g8h8 e4e6 b8d7 e6d6 g7b2 e2f3 a8d8 d6e7 f8e8 e7h4 b7b5 a4b5 a6b5 d5d6 b2f6 h4f6 d7f6 g4g5 f6e4');
INSERT INTO evaluations VALUES(55562,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(55562,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55562,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(55562,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(55562,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(55562,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(55562,36,'cp 23','b1c3 g8f6 d2d4 g7g6 f1e2 b8d7 g1f3 f8g7 e1g1 e8g8 d1c2 f8e8 d4e5 d7e5 f3e5 d6e5 b2b3 d8e7 c3d5 e7d8 f1d1 c7c6 d5f6 d8f6 e3e4 f6e7 c4c5 c8e6 d1d6 e8d8 c1a3 g7f8 a1d1 e7h4 g2g3');
INSERT INTO evaluations VALUES(57395,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(57395,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(57395,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(57395,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(57395,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(57395,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(57395,36,'cp 163','b8c6 e2e4 f6e5 f1c4 g8f6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8e7 f1e1 e8g8 e1e5 d4f5 g5e4 e7f6 e4f6 d8f6 e5e4 f5d6 e4f4 c8f5 b1c3 f6g6 c1e3 h7h5 f2f3 a8e8 d1d2 f5c2 a1e1 f8f4 e3f4 c2f5 e1e8 d6e8 d2e3 a7a6 e3a7 e8d6 a7c5 f5h3 f4g3');
INSERT INTO evaluations VALUES(7143,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(7143,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(7143,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(7143,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(7143,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(7143,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7143,36,'cp 13','c3d5 e8g8 c2c3 c5a7 d5f6 d8f6 c4b3 c6e7 c1e3 a7e3 f2e3 e7g6 a2a4 g6h4 a4a5 h4f3 d1f3 f6e7 f3g3 c8e6 b3e6 f7e6 b2b4 f8f1 a1f1 a8f8 f1f8 e7f8 g3g6 f8f7 g6f7 g8f7 g1f2 f7e7 f2f3 g7g6 c3c4 e7d7 b4b5 c7c6 b5b6 c6c5');
INSERT INTO evaluations VALUES(14396,30,'cp 244',NULL);
INSERT INTO evaluations VALUES(14396,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(14396,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(14396,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(14396,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(14396,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(14396,36,'cp 226','b1c3 e7e6 a2a3 a7a6 f1d3 f6d7 c1e3 e6e5 d4d5 c6a7 e1g1 f8d6 f3d2 d6c5 d1e2 e8g8 h2h3 c5e3 e2e3 h7h6 a3a4 d8g5 e3g5 h6g5 f1c1 f8e8 b2b4 c7c6 d2f3 f7f6 f3d2 e8d8 d2c4 a8b8 d3e2 d7f8 c4b6 c6d5 e4d5 c8f5 a4a5');
INSERT INTO evaluations VALUES(14625,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14625,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14625,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14625,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14625,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(14625,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(14625,36,'cp -10','g7g5 f3h2 c8e6 b1c3 d6d5 e4d5 f6d5 d1f3 c6d4 f3e4 f8g7 h2f3 d4f3 e4f3 c7c6 c3d5 c6d5 c4b5 e8f8 c2c4 g7f6 c1d2 f8g7 d2c3 a7a6 b5a4 a8c8 c4d5 d8d5 f3d5 e6d5 a4d7 c8c5 d7f5 a6a5 a2a4 h8d8 f5e4 b7b6 a1c1 d5e4 d3e4 h6h5 f1e1 g5g4 h3g4');
INSERT INTO evaluations VALUES(19330,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19330,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19330,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(19330,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19330,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19330,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(19330,36,'cp 10','g1f3 f8g7 f1c4 g8f6 e1g1 d7d5 f1d1 d5d4 e3e2 e8g8 e4e5 f6g4 c1f4 d8c7 b1a3 g4e5 f3e5 g7e5 a3b5 c7b8 f4e5 b8e5 b5d4 e5e2 c4e2 c6d4 d1d4 c8e6 a2a4 a8c8 c2c3 c8c5 f2f3 f8c8 g1f2 h7h5 g2g3 g8g7 b2b4 c5c7 a4a5 a7a6');
INSERT INTO evaluations VALUES(26045,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(26045,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26045,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26045,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26045,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(26045,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(26045,36,'cp 19','c5d4 f3d4 g7g6 c1e3 g8f6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 d1d2 f8g7 c3a4 d8c7 c2c4 e8g8 e1g1 f6d7 a1d1 c7b7 c4c5 d7c5 a4c5 d6c5 e3c5 b7b2 d2b2 g7b2 c5e7 f8c8 e7c5 c8d8 f3f4 a6a5 a2a4 a8b8 e4e5 d8d1 f1d1');
INSERT INTO evaluations VALUES(26034,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26034,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26034,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(26034,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(26034,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26034,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(26034,36,'cp 15','f3d4 g7g6 c1e3 g8f6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 d1d2 f8g7 c3a4 e8g8 e1g1 d8c7 a1e1 d6d5 e3b6 c7b7 b2b3 f8e8 b6d4 e7e5 d4f2 g7f8 f2c5 f8c5 a4c5 b7a7 d2f2 g8g7 c5d3 a7f2 d3f2 a8d8 f2d3 d5e4 f3e4');
INSERT INTO evaluations VALUES(43249,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(43249,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(43249,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43249,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(43249,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(43249,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(43249,36,'cp 41','c2c4 e8g8 b1c3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 f1e2 c5d4 c3d4 b8c6 c1e3 d8a5 e3d2 a5a3 d4d5 c6e5 e1g1 c8d7 a1b1 a3a2 b1b7 f8b8 d1b1 a2b1 f1b1 e5f3 g2f3 g7e5 f3f4 d7c8 b7b5 e5d6 e4e5 d6c7 d2b4 c8d7 b4e7 d7b5 e2b5');
INSERT INTO evaluations VALUES(43297,30,'cp 512',NULL);
INSERT INTO evaluations VALUES(43297,31,'cp 497',NULL);
INSERT INTO evaluations VALUES(43297,32,'cp 482',NULL);
INSERT INTO evaluations VALUES(43297,33,'cp 480',NULL);
INSERT INTO evaluations VALUES(43297,34,'cp 486',NULL);
INSERT INTO evaluations VALUES(43297,35,'cp 496',NULL);
INSERT INTO evaluations VALUES(43297,36,'cp 500','g2g4 d7d6 g4h5 d6e5 h5h6 g7f6 d4e5 d8d1 c3d1 f6e5 g1f3 e5d6 d1c3 b8c6 c1e3 c8g4 f1e2 g4h5 e1c1 f7f6 h2h3 h8f8 h1e1 e8c8 c3e4 e7e5 e3d2 d6b4 c2c3 f6f5 c3b4 f5e4 f3g5 h5e2 e1e2 f8f5 d2e3 c6b4 a2a3 d8d1 c1d1 b4d5 g5h7 d5e3 e2e3');
INSERT INTO evaluations VALUES(43294,30,'cp 482',NULL);
INSERT INTO evaluations VALUES(43294,31,'cp 471',NULL);
INSERT INTO evaluations VALUES(43294,32,'cp 495',NULL);
INSERT INTO evaluations VALUES(43294,33,'cp 485',NULL);
INSERT INTO evaluations VALUES(43294,34,'cp 486',NULL);
INSERT INTO evaluations VALUES(43294,35,'cp 465',NULL);
INSERT INTO evaluations VALUES(43294,36,'cp 494','d7d6 g4h5 d6e5 h5h6 g7f6 d4e5 d8d1 c3d1 f6e5 g1f3 e5d6 c1e3 c8g4 f1e2 f7f6 d1c3 b8c6 c3e4 e8c8 e1c1 h8f8 h2h3 g4f5 e4d6 c7d6 f3d4 c6d4 e3d4 e7e5 d4a7 c8c7 a7e3 b7b6 e2f3 f5e6 a2a3 g6g5 h3h4 g5g4 f3g2 d6d5 c2c3 d5d4 c3d4');
INSERT INTO evaluations VALUES(46349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46349,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(46349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46349,36,'cp 0','f8d6 f1d3 b8c6 e1g1 e8g8 h2h3 h7h6 c1d2 c8d7 a1c1 a8c8 a2a3 f8e8 f1e1 e6e5 c3b5 d6b8 f3e5 c6e5 c1c8 d8c8 d4e5 b8e5 b5d4 f6e4 d3e4 d5e4 d2c3 d7h3 g2h3 c8h3 f2f4 e5d4 d1d4 h3g3 g1f1 g3f3 f1g1');
INSERT INTO evaluations VALUES(48460,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(48460,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(48460,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(48460,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(48460,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(48460,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(48460,36,'cp 17','a7a6 d2d4 b7b5 c4d3 c8b7 c1g5 b8d7 d1e2 f8e7 g5f6 e7f6 c3e4 e8g8 c2c3 b7d5 b2b3 g7g6 a1d1 f6g7 f1e1 d8e7 c3c4 d5b7 d1c1 b5c4 c1c4 d7b6 c4c5 b6d5 f3e5 d5f4 e2f3 g7e5 d4e5');
INSERT INTO evaluations VALUES(49357,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49357,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(49357,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(49357,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49357,36,'cp 0','c7c5 b1d2 c5d4 e3d4 f8e7 g1f3 b8c6 c2c3 e8g8 f1d3 f6d5 g5e7 c6e7 e1g1 d5f4 d3e4 a8b8 g2g3 f4d5 e4d3 d7d6 f1e1 b7b5 a2a4 a7a6 a4b5 a6b5 d1e2 c8d7 a1a6 d7c6 e1a1 h7h6 d2e4 f8e8 a6a5 d8b6 a5a6');
INSERT INTO evaluations VALUES(53661,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(53661,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53661,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53661,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53661,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(53661,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(53661,36,'cp 0','b7b5 a2a4 b5b4 c3b1 c8a6 c1g5 d8a5 b1d2 c4c3 d2b3 a5b6 f1a6 b6a6 b2c3 b4c3 d1c2 g8f6 g5f6 g7f6 c2c3 b8d7 f3d2 h8g8 a1c1 c6c5 g2g3 a8b8 c3c4 a6c4 c1c4 c5d4 b3d4 g8g5 e1e2 g5c5 h1c1 c5c4 c1c4 d7c5 d4c6 b8b2');
INSERT INTO evaluations VALUES(10183,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(10183,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(10183,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(10183,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(10183,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(10183,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(10183,36,'cp 13','e8g8 a2a3 a7a5 e2e3 c5a7 b2b3 f8e8 c1b2 d7d6 g1f3 c6e7 d3d4 e5e4 f3d2 d6d5 c4d5 e7d5 c3e4 f6e4 g2e4 d5f6 e4g2 a7d4 b2d4 d8d4 e1g1 d4c5 a1c1 c5a3 c1c7 a3d6 d1c2 e8d8 d2e4 d6e5 c2c3 e5c3 e4c3 a8b8');
INSERT INTO evaluations VALUES(4373,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4373,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4373,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(4373,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4373,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4373,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(4373,36,'cp 13','a2a3 a7a5 e2e3 c5a7 d1c2 d7d6 g1f3 f8e8 e1g1 h7h6 c1d2 c8e6 a1b1 d6d5 c4d5 f6d5 f1c1 a7b6 c3a4 d8d6 a4c5 b6c5 c2c5 a5a4 c5d6 c7d6 d2c3 b7b5 f3d2 e8b8 d2e4 b5b4 a3b4 c6b4 c3b4 d5b4 e4d6 b4d3 g2a8 d3c1 b1c1 b8a8');
INSERT INTO evaluations VALUES(11088,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(11088,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(11088,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11088,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(11088,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(11088,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(11088,36,'cp 17','d4c6 d8c7 e4e5 c7c6 e5f6 c6c4 d1f3 d7d5 c1g5 c8d7 e1c1 g7g6 d1d3 d7c6 h1d1 f8c5 g5e3 e8c8 e3d4 c5d4 d3d4 c4c5 f3e3 d8d7 e3e5 c5a5 f2f4 h7h5 g2g3 a5c7 a2a4 c8b8 a4a5 h5h4 e5c7 d7c7 g3g4 h8c8');
INSERT INTO evaluations VALUES(15635,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(15635,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(15635,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(15635,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(15635,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15635,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(15635,36,'cp 49','f3d4 f8e7 b1c3 e8g8 a2a4 c6d4 d1d4 c8e6 f1d1 f6g4 c3d5 e7h4 g2g3 c7c6 d5f4 e6c4 d4c4 g4e5 c4e2 h4e7 b2b3 f8e8 g1g2 e5g6 c2c4 d8d7 e2c2 a8d8 h2h3 g6f4 c1f4 a7a6 a1a2 d7e6 c2d3 d8d7 d3f3 e7f6');
INSERT INTO evaluations VALUES(15628,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15628,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(15628,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(15628,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(15628,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(15628,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(15628,36,'cp 46','f8e7 b1c3 e8g8 a2a4 c8g4 f2f3 c6d4 d1d4 g4e6 f1d1 d8d7 c1e3 f8d8 e3f2 e6c4 d4c4 c7c6 d1d2 d7e6 c4d3 d8e8 a1d1 a7a6 a4a5 a8d8 d1e1 e6e5 f2b6 d8d7 d3d4 e7d8 b6d8 e8d8 d4e5 d6e5 d2d7 d8d7 e1d1 d7d1 c3d1 g8f8 g1f2 h7h6 d1e3');
INSERT INTO evaluations VALUES(24771,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(24771,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(24771,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(24771,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(24771,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(24771,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(24771,36,'cp 104','f1e2 e6g4 h2h3 g4f3 e2f3 f6d7 d1e2 d7b6 e1g1 e7e6 c3e4 f8e7 e4d2 e6e5 d4e5 d8d3 e2d3 c4d3 b2b3 e8c8 f3c6 b7c6 c1b2 c8b7 f1d1 a7a5 d2e4 d8d5 f2f4 h8d8 g1f2 d5b5 b2c3 a5a4 a1b1 g7g5 f4g5 a4a3 h3h4 e7b4 c3b4 b5b4 f2f3');
INSERT INTO evaluations VALUES(42500,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(42500,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(42500,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(42500,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(42500,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(42500,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(42500,36,'cp 31','b2c3 d7d6 f3d2 b8c6 e2e4 e6e5 c1b2 c6e7 g2g3 e8g8 f1g2 e7g6 e1g1 b7b6 a2a4 c8d7 h2h3 c7c5 f1e1 h7h6 d1e2 d8c8 g1h2 c8c7 d2f1 h6h5 h3h4 f6g4 h2g1 a8e8 f2f3 g4h6 f1e3 f7f5 e3d5 c7c6');
INSERT INTO evaluations VALUES(44735,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44735,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44735,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(44735,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(44735,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44735,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44735,36,'cp 10','f8g7 f1e1 e8g8 d2d4 c5d4 f3d4 b8c6 c4b3 c6d4 d1d4 a7a6 a2a4 f6d7 d4b4 a6a5 b4a3 d7c5 c1e3 b7b6 a1d1 g7c3 b2c3 d8c7 e3d4 c8e6 a3a2 a8b8 d4c5 c7c5 b3e6 f7e6 a2e6 f8f7 e1e2 c5c3 e6b3 c3b3 c2b3 b8c8 h2h3');
INSERT INTO evaluations VALUES(51321,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(51321,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(51321,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(51321,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(51321,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(51321,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(51321,36,'cp 131','c3b5 c8a6 b5c3 e7e6 g1f3 g8f6 e2e3 b8d7 f3e5 c7c5 e5c4 f8e7 c4d6 e7d6 f1a6 c5d4 e3d4 d8c7 h2h3 e8g8 e1g1 a8d8 d1e2 d7b8 a6d3 b8c6 c1e3 c7b8 f1d1 d6h2 g1h1 h2c7 c3e4 f6d5 e4c5 c6e7 a1c1 c7b6 a2a4 d5e3 f2e3 b8g3 e2f3 g3f3 g2f3');
INSERT INTO evaluations VALUES(51314,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(51314,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(51314,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(51314,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(51314,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(51314,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(51314,36,'cp 115','c8a6 b5c3 e7e6 e2e3 b8d7 g1f3 g8f6 d1a4 a6b7 f3e5 c7c5 e5c4 c5d4 e3d4 f8e7 c4a5 d8c7 a5b7 c7b7 a4b5 b7c7 f1e2 e8g8 e1g1 d7b6 e2f3 a8d8 f1d1 f6d5 b5a5 d5c3 b2c3 d8c8 c1d2 b6c4 a5c7 c8c7 d2f4');
INSERT INTO evaluations VALUES(52001,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52001,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52001,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(52001,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(52001,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(52001,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(52001,36,'cp 26','f1g2 d7d5 e4d5 e6d5 g1e2 b8c6 e1g1 g8f6 d2d4 c5d4 e2d4 c8g4 d1d3 f8e7 b1c3 e8g8 h2h3 g4d7 c1f4 d8c8 f1e1 c6d4 d3d4 e7c5 d4d3 d5d4 c3e4 f6e4 g2e4 h7h6 d3f3 d7c6 g1g2 c6e4 e1e4 c8c6 a1e1 c5b6 e4e7 c6c2 f3b7 a8d8 b7e4 c2e4 e1e4');
INSERT INTO evaluations VALUES(52002,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52002,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52002,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(52002,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(52002,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(52002,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(52002,36,'cp 25','d7d5 e4d5 e6d5 g1e2 g8f6 d2d4 c5d4 e1g1 f8e7 e2d4 e8g8 b1c3 b8c6 f1e1 c8g4 d1d3 f8e8 h2h3 g4h5 c1g5 h7h6 g5f6 e7f6 e1e8 d8e8 d4f5 c6b4 d3d2 h5g6 a2a3 b4c6 d2d5 a8d8 d5b5 c6d4 b5e8 d8e8 f5d6 e8e6 d6e4 f6e5 a1d1 d4c2');
INSERT INTO evaluations VALUES(54258,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(54258,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54258,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54258,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(54258,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54258,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(54258,36,'cp 32','e4d5 e6d5 g1e2 g8f6 d2d4 c5d4 e1g1 f8e7 e2d4 e8g8 h2h3 b8c6 c1e3 e7b4 b1d2 f8e8 c2c3 b4a5 d2f3 h7h6 f1e1 c8d7 e3f4 e8e1 d1e1 a5b6 a1d1 d8c8 e1f1 c6d4 f3d4 d7a4 d1d3 c8d7 b2b3 a4c6 d4c6 b7c6 c3c4 a8c8 b3b4 c8e8 c4c5 b6c7 f4e3');
INSERT INTO evaluations VALUES(54149,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(54149,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(54149,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(54149,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(54149,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(54149,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54149,36,'cp 29','e6d5 g1e2 g8f6 e1g1 f8e7 d2d4 c5d4 h2h3 e8g8 e2d4 b8c6 c1e3 f8e8 b1c3 e7b4 c3e2 b4d6 c2c3 h7h6 a2a4 c8d7 e2f4 d6e5 d4b3 d7f5 f4d5 f6d5 d1d5 d8d5 g2d5 f5h3 d5g2 h3g2 g1g2 b7b6 f1d1 a8d8 d1d2 d8d2 b3d2 e5f6 d2c4');
INSERT INTO evaluations VALUES(54561,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(54561,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(54561,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(54561,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(54561,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(54561,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(54561,36,'cp 37','g1e2 g8f6 e1g1 f8e7 d2d4 c5d4 h2h3 b8c6 e2d4 e8g8 c1e3 e7b4 b1d2 f8e8 c2c3 b4a5 d2f3 h7h6 f1e1 c8d7 d1d3 d8c8 g1h2 a5c7 a1c1 c6e7 b2b3 e7f5 d4f5 d7f5 d3d2 c8d8 c1d1 f5g6 d2b2 c7b6 e3d4 b6d4 f3d4 a8c8 e1e8 d8e8');
INSERT INTO evaluations VALUES(52396,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(52396,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(52396,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52396,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52396,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52396,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52396,36,'cp 1','d7d5 g1f3 c8g4 d2d3 e7e6 b1d2 g8f6 a2a3 f8d6 h2h3 g4f3 d2f3 e8g8 g2g3 b8d7 f1g2 e6e5 e1g1 a7a5 e2e3 f8e8 c2c4 d8e7 d1c2 h7h6 f1d1 a8c8 b2c3 b7b6 c3b2 b6b5 a3a4 d6b4 a4b5 c6b5 c2e2 d7c5 c4b5 c5b3');
INSERT INTO evaluations VALUES(54069,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(54069,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54069,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(54069,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(54069,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54069,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(54069,36,'cp 10','g8f6 e2e4 f8c5 f1b5 b8c6 c1g5 c5e7 e4d5 e6d5 e1g1 e8g8 g5f6 e7f6 d1d5 c8e6 d5e4 d8a5 a1d1 a8e8 b5c6 b7c6 e4c6 e8c8 c6d6 f6c3 b2c3 a5c3 f3d4 f8d8 d6f4 e6a2 h2h3 h7h6 d1d3 c3b2 d4f5 d8d3 c2d3 g8h8 f5e3');
INSERT INTO evaluations VALUES(46099,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(46099,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(46099,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46099,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(46099,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(46099,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46099,36,'cp 24','e2e4 f8c5 f1b5 c8d7 b5d7 d8d7 e4d5 f6d5 c3d5 e6d5 e1g1 e8g8 f1e1 b8c6 d1d3 f8e8 c1d2 h7h6 a2a3 a8d8 e1e8 d8e8 g1f1 e8e4 d2c3 d7e6 a1d1 d5d4 d1e1 e4e1 c3e1 c6e5 f3e5 e6e5 h2h3 b7b6 a3a4 a7a5 f2f3 e5g5 b2b3 g5c1 f1e2');
INSERT INTO evaluations VALUES(9154,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(9154,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(9154,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(9154,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(9154,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9154,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9154,36,'cp 23','e8g8 a2a4 a7a5 f1e1 c8f5 b1d2 d5f6 d1e2 c5d6 d2e4 h7h6 f3d2 f6d7 e2h5 f5h7 e1e3 d7f6 e4f6 d8f6 d2e4 f6f5 h5f5 h7f5 e4d6 c7d6 e3g3 g8h7 c1e3 f5e6 c4e6 f7e6 a1e1 a8d8 e3b6 d8d7 g3g4 d6d5 b2b4 a5b4 c3b4 e5e4 d3e4 c6b4 e4d5 b4d5 e1e6');
INSERT INTO evaluations VALUES(9316,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(9316,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(9316,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(9316,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(9316,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(9316,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(9316,36,'cp -52','c8g4 c4e2 g4d7 d2d3 a7a6 g5f3 h7h6 c1e3 c6d4 e3d4 e5d4 c3b1 e8g8 b1d2 f8e8 f1e1 c5b6 a2a4 c7c5 e2f1 b6c7 g2g3 b7b5 a4b5 a6b5 f1g2 a8a1 d1a1 f6g4 h2h3 g4e5 f3e5 d6e5 a1a7 c7d6 e1a1 c5c4 d3c4');
INSERT INTO evaluations VALUES(10515,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(10515,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(10515,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(10515,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(10515,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(10515,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(10515,36,'cp 65','f6e4 d4d5 c6e7 d1d4 e4f6 c1g5 d7d6 g5f6 g7f6 d4f6 e8g8 f6d4 c7c5 d5c6 e7c6 d4f4 d8e7 e1d2 c8e6 h1e1 g8h8 c4b3 a8d8 a1d1 f8g8 g2g3 g8g6 b3c2 g6f6 f4h4 h7h6 h4e4 e6f5 e4e7 c6e7 e1e7 f5c2 d2c2 f6f3 e7b7 d8c8 d1d6 c8c3 c2d2 h8g7 b7a7');
INSERT INTO evaluations VALUES(10700,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(10700,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(10700,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(10700,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(10700,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(10700,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(10700,36,'cp 48','d4d5 c6e7 d1d4 e4f6 c1g5 d7d6 g5f6 g7f6 d4f6 e8g8 f6h6 e7g6 e1d2 f8e8 a1e1 e8e1 h1e1 d8f8 h6f8 g8f8 f3d4 c8d7 c4d3 g6f4 d3e4 a8e8 g2g3 f4g6 f2f4 h7h6 c3c4 g6e7 e4d3 e7g8 d3f5 e8e1 d2e1 g8f6 e1f2 h6h5 f5d7 f6d7 d4b5 a7a6 b5c7');
INSERT INTO evaluations VALUES(10516,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(10516,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(10516,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(10516,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(10516,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(10516,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(10516,36,'cp 51','b2c3 f6e4 d4d5 c6e7 d1d4 e4f6 c1g5 d7d6 g5f6 g7f6 d4f6 e8g8 h2h4 e7g6 f6d8 f8d8 h4h5 g6e5 f3e5 d6e5 e1c1 g8g7 f2f3 g7f6 h1e1 d8d6 c4b3 b7b5 a2a4 a7a6 a4a5 c8b7 g2g4 a8d8 g4g5 f6g5 e1e5 g5f4 e5e7 b7d5 b3d5 d6d5 d1d5 d8d5 e7f7 f4g5 f7h7');
INSERT INTO evaluations VALUES(17979,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(17979,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(17979,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(17979,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(17979,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(17979,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(17979,36,'cp 36','f8e7 a2a4 b5b4 c3d5 g8f6 d2d4 c8b7 e1g1 e8g8 d4e5 c6e5 f3e5 d6e5 d5f6 e7f6 d1d8 f8d8 f2f3 a6a5 c1e3 h7h6 f1d1 g8f8 b3c4 f6e7 g1f1 d8d1 a1d1 a8d8 d1d8 e7d8 e3f2 b7c8 f2e1 c7c6 c2c3 c6c5 f1e2 f7f6 e2d3 d8b6 d3c2 f8e7 c2b3');
INSERT INTO evaluations VALUES(31873,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(31873,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(31873,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(31873,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(31873,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(31873,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(31873,36,'cp -37','e7e6 b1c3 g8f6 d2d3 f5g6 f3e5 b8d7 e5g6 h7g6 c1d2 f8b4 g2g4 c7c5 g4g5 f6g8 a2a3 b4a5 f1g2 g8e7 e3e4 d5d4 c3e2 e7c6 c2c3 d8b6 c3d4 a5d2 d1d2 e8c8 e1g1 c8b8 f1b1 c5d4 b2b4 e6e5 a3a4 e5f4 e2f4 d7e5');
INSERT INTO evaluations VALUES(32328,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(32328,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(32328,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(32328,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(32328,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(32328,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(32328,36,'cp -36','b1c3 g8f6 d2d3 c7c5 f3e5 h7h6 g2g4 f5h7 f1g2 f6d7 e5d7 d8h4 e1f1 b8d7 d1e2 f8d6 e2f2 h4e7 h2h4 a8c8 e3e4 c5c4 g4g5 c4d3 c2d3 d6c5 d3d4 c5d6 e4d5 d7b6 g5h6 g7h6 h1h3 e8g8 d5e6 f7e6 c3e4 g8h8 e4d6 e7d6 b2b3');
INSERT INTO evaluations VALUES(32951,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(32951,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(32951,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(32951,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(32951,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(32951,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(32951,36,'cp 39','d2d4 e5d4 f3d4 g7g6 c2c4 f8g7 c1e3 g8e7 b1c3 b8c6 e1g1 e8g8 d4c6 d7c6 d1d3 f7f5 e3d4 g7d4 d3d4 c6c5 d4d3 a8e8 a1d1 e7c6 e4f5 c5f5 d3d2 f5f4 b2b3 f4d2 d1d2 g8g7 f2f3 g6g5 d2d5 e8e5 f1d1 e5d5 d1d5');
INSERT INTO evaluations VALUES(35170,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35170,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(35170,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(35170,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35170,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35170,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35170,36,'cp 0','h4e4 c4e2 e4f5 b1c3 g8f6 e1g1 d7d5 d2d4 e5e4 f3h4 f5d7 f2f3 e4f3 f1f3 d7d8 c1g5 f8e7 d1d3 e8g8 a1f1 b8c6 h4f5 c8f5 f3f5 d8d6 c3b5 d6d8');
INSERT INTO evaluations VALUES(35176,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35176,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35176,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35176,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35176,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35176,36,'cp 0','c4e2 e4f5 b1c3 g8f6 e1g1 d7d5 d2d4 e5e4 f3h4 f5d7 f2f3 e4f3 f1f3 d7d8 c1g5 f8e7 d1d3 b8c6 a1f1 e8g8 h4f5 c8f5 f3f5 d8d6 c3b5 d6d8');
INSERT INTO evaluations VALUES(35209,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35209,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35209,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35209,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35209,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35209,36,'cp 0','e4f5 b1c3 g8f6 e1g1 d7d5 d2d4 e5e4 f3h4 f5d7 f2f3 e4f3 f1f3 d7d8 c1g5 f8e7 d1d3 b8c6 a1f1 e8g8 h4f5 c8f5 f3f5 f6d7 g5f4 d7f6');
INSERT INTO evaluations VALUES(36114,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(36114,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(36114,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(36114,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(36114,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(36114,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(36114,36,'cp -19','c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 f8e8 f2f4 d8b6 g1h1 g6e4 d3e4 f6e4 c3e4 d5e4 d1e2 c6c5 e2c4 b6c7 d4c5 d6e5 f4e5 b8c6 c4e4 c6e5 f1f5 c7c5 a1e1 c5c6 e4c6 e5c6 g5d2 e8e1 d2e1 a8e8 e1g3 f7f6 h1g2 b7b6 f5f2 g8f7 c2c3 e8e4 g2f3 e4c4');
INSERT INTO evaluations VALUES(27804,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(27804,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(27804,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(27804,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(27804,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(27804,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(27804,36,'cp -5','h2h3 g4h5 g2g4 h5g6 f3e5 f8e8 f2f4 d8b6 g1h1 g6e4 d3e4 f6e4 c3e4 d5e4 d1e2 b8a6 c2c3 a6c7 g5h4 e4e3 a1e1 b6a5 f1f3 b7b5 f3e3 a5a2 e2g2 c7d5 e3e4 a7a5 e5c6 e8e4 g2e4 b5b4 c6e5 d5c7 e4c6 a2d5 c6d5 c7d5 c3b4 d6b4 e1d1 d5f4 d4d5 f7f6 e5c4 a8c8');
INSERT INTO evaluations VALUES(27803,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(27803,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(27803,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27803,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(27803,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(27803,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(27803,36,'cp -13','g4h5 g2g4 h5g6 f3e5 f8e8 f2f4 d8b6 g1h1 g6e4 d3e4 f6e4 c3e4 d5e4 d1e2 b8a6 g5h4 a6c7 a1e1 e4e3 c2c3 b6a5 f1f3 a5a2 e2c4 a2c4 e5c4 d6f8 f3e3 e8e3 c4e3 a7a5 f4f5 a5a4 f5f6 b7b5 h4g3 c7d5 f6g7 g8g7 e3d5 c6d5');
INSERT INTO evaluations VALUES(27807,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(27807,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(27807,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27807,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27807,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(27807,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(27807,36,'cp -10','g2g4 h5g6 f3e5 f8e8 f2f4 d8b6 g1h1 g6e4 c3e4 f6e4 d3e4 d5e4 d1e2 b8a6 c2c3 c6c5 e2c4 b6c7 c4d5 c5d4 c3d4 a6b4 d5e4 f7f6 a2a3 c7c6 e4c6 b4c6 e5c6 b7c6 g5h4 e8e4 a1c1 e4d4 c1c6 d6f4 h4e1 a8d8 h1g2 f4g5 e1b4 d8d7 b4c3');
INSERT INTO evaluations VALUES(27806,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(27806,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(27806,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27806,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(27806,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(27806,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(27806,36,'cp -10','h5g6 f3e5 f8e8 f2f4 d8b6 g1h1 g6e4 c3e4 f6e4 d3e4 d5e4 d1e2 b8a6 c2c3 c6c5 e2c4 b6c7 c4d5 c5d4 c3d4 a6b4 d5e4 f7f6 a2a3 c7c6 e4c6 b4c6 e5c6 b7c6 g5h4 e8e2 f1f2 e2e4 a1c1 d6f4 c1c6 e4d4 h1g2 a7a5 f2c2 f4e5 h4g3 e5g3 g2g3 d4d3 g3g2 a8d8 h3h4 d8d4 c6c4 g8f7 h4h5 d4c4');
INSERT INTO evaluations VALUES(27809,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(27809,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(27809,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(27809,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(27809,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(27809,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(27809,36,'cp -10','f3e5 f8e8 f2f4 d8b6 g1h1 g6e4 d3e4 f6e4 c3e4 d5e4 d1e2 b8a6 c2c3 c6c5 e2c4 b6c7 c4d5 c5d4 c3d4 a6b4 d5e4 f7f6 a2a3 c7c6 e4c6 b4c6 e5c6 b7c6 g5h4 e8e2 a1c1 e2b2 c1c6 d6a3 c6a6 a3b4 h4f2 b2d2 h1g2 b4c3 g2f3 c3d4 f3g3 h7h6 f2d4 d2d4 f1f3 d4c4 f3e3 a8d8');
INSERT INTO evaluations VALUES(45162,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45162,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45162,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45162,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45162,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(45162,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45162,36,'cp 0','b1c3 g7g6 g2g3 f8g7 g1e2 b8c6 f1g2 a8b8 a2a4 e8g8 e1g1 a7a6 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(45160,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45160,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45160,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45160,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45160,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45160,36,'cp 0','g7g6 g2g3 f8g7 f1g2 b8c6 g1e2 a8b8 e1g1 e8g8 a2a4 a7a6 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(44756,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44756,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44756,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44756,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44756,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44756,36,'cp 0','g1e2 f8g7 g2g3 b8c6 f1g2 a8b8 a2a4 e8g8 e1g1 a7a6 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(46514,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(46514,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(46514,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(46514,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(46514,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(46514,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(46514,36,'cp 46','f1d3 f8d6 e1g1 e8g8 c3e2 d8e7 b2b3 c8d7 f3e5 d7e8 c1b2 b8d7 e2f4 a7a5 d3e2 g7g5 f4d3 g8h8 a2a4 f8g8 d1c1 h7h5 c4d5 e6d5 b2a3 d6a3 c1a3 e7a3 a1a3 h5h4 a3a2 d7e5 d4e5 f6e4 f2f3 e4c3 a2c2 c3e2 c2e2 e8h5');
INSERT INTO evaluations VALUES(46517,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(46517,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(46517,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(46517,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(46517,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(46517,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(46517,36,'cp 48','f8d6 e1g1 e8g8 c3e2 d8e7 b2b3 c8d7 c1b2 d7e8 e2f4 b8d7 d1c2 f6h5 f3e5 a8d8 f4h3 g7g5 c4d5 c6d5 f2f3 d7e5 d4e5 d6c5 b2d4 c5d4 e3d4 h5g7 c2d2 h7h6 h3f2 e8g6 a1c1 d8c8 g2g4 c8c1 f1c1');
INSERT INTO evaluations VALUES(40665,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(40665,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(40665,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(40665,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(40665,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(40665,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(40665,36,'cp 44','e1g1 e8g8 c3e2 a7a5 b2b3 d8e7 c1b2 c8d7 e2f4 d7e8 d1c2 b8d7 f3g5 e8f7 f2f3 a8e8 g2g3 h7h6 g5f7 e7f7 a1e1 g7g5 f4h3 g5g4 f3g4 f6g4 e3e4 d5e4 d3e4 f7h5 e4g2 d7f6 e1e2 f8f7 b2c3 f6e4 g2e4 f5e4 f1f7 g8f7');
INSERT INTO evaluations VALUES(40664,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40664,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(40664,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(40664,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(40664,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(40664,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(40664,36,'cp 33','e8g8 c3e2 d8e7 b2b3 c8d7 e2f4 d7e8 c1b2 b8d7 d3e2 e8h5 f4h5 f6h5 f3e5 h5f6 a1c1 g8h8 d1c2 a8e8 e2d3 e8d8 c2e2 d6b4 c4d5 d7e5 d4e5 f6d5 g2g3 b4c5 d3c4 c5a3 f1d1 a3b2 e2b2 a7a5 b2e2 d8e8 g1g2 g7g5 c4d3');
INSERT INTO evaluations VALUES(36168,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(36168,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(36168,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(36168,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(36168,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(36168,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(36168,36,'cp 51','c3e2 d8e7 b2b3 c8d7 c1b2 d7e8 e2f4 b8d7 d3e2 e8h5 f3e5 h5e2 d1e2 a7a5 f4d3 f6e4 f2f3 e4g5 a2a4 g5f7 a1c1 a8c8 e5d7 e7d7 b2c3 c8a8 e2d2 d7c7 h2h3 b7b6 c4c5 b6c5 d4c5 d6e7 f1d1 e7f6 c3f6 g7f6');
INSERT INTO evaluations VALUES(56199,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(56199,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56199,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56199,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(56199,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56199,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56199,36,'cp 0','g8f6 b1c3 f8b4 f1d3 e8g8 g1e2 c7c6 e1g1 f6h5 f4e3 b4d6 d1d2 d8c7 e2g3 h5f4 g3f5 c8f5 d3f5 f4e6 d2d3 h7h6 g2g3 b8d7 a1e1 a8e8 f5h7 g8h8 h7f5 h8g8');
INSERT INTO evaluations VALUES(48649,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48649,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(48649,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(48649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48649,36,'cp 0','b1c3 f8b4 f1d3 e8g8 g1e2 c7c6 a2a3 b4a5 e1g1 f8e8 e2g3 b8d7 d1f3 a5b6 g3f5 d7f8 f5d6 e8e7 f4e5 f6d7 e5f4');
INSERT INTO evaluations VALUES(1640,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(1640,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(1640,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(1640,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(1640,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(1640,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(1640,36,'cp 48','c3d5 e8d8 g1f3 e7e5 b2b3 g8e7 d5e3 e7f5 c1b2 f5e3 f2e3 f7f6 e1c1 f8d6 h1f1 a7a5 a2a4 c8e6 f3h4 c6e7 g2g4 d8d7 f1f2 g7g5 h4g2 h8f8 e3e4 e7c6 d1f1 c6d4 b2d4 e5d4 f2f6 f8f6 f1f6 a8f8 f6f8 d6f8');
INSERT INTO evaluations VALUES(6943,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(6943,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(6943,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(6943,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(6943,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(6943,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(6943,36,'cp 90','d8e7 c5e3 d7d5 b1c3 d5e4 f3g5 g8f6 e1g1 h7h6 g5e4 f6e4 c3e4 c8f5 f1e1 f5e4 c1d2 a8d8 d2c3 d8d7 f2f3 e8d8 e3e4 e7e4 e1e4 h8e8 a1e1 e8e4 e1e4 d7d1 g1f2 g7g6 g2g4 d8d7 e4f4 d7e8 c3g7 h6h5 g4h5 g6h5 f4f8 e8d7 f8h8 d1d5 f3f4 c6e7 g7c3 d7e6 h8h6 e6f7 h6h7 f7e6 f2e3 e7f5 e3f3 d5d3 f3f2 d3h3 f2g2 h3h4 h7c7');
INSERT INTO evaluations VALUES(12743,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12743,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(12743,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(12743,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12743,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12743,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12743,36,'cp 0','g1f3 f8g7 f1c4 e8g8 e1g1 c6a5 c4d3 d7d5 e4e5 f6d7 c1f4 a5c6 f1e1 d7c5 d3b5 a7a6 b5c6 b7c6 f4e3 c5d7 e3f4');
INSERT INTO evaluations VALUES(12742,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12742,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12742,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12742,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12742,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12742,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12742,36,'cp 0','f8g7 f1c4 e8g8 e1g1 c6a5 c4d3 d7d5 e4e5 f6d7 f1e1 a5c6 c1f4 d7c5 d3b5 a7a6 b5c6 b7c6 f4e3 c5d7 e3f4');
INSERT INTO evaluations VALUES(8065,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8065,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8065,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8065,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8065,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8065,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8065,36,'cp 0','f1c4 e8g8 e1g1 c6a5 c4d3 d7d5 e4e5 f6d7 c1f4 a5c6 f1e1 d7c5 d3b5 a7a6 b5c6 b7c6 f4e3 c5d7 e3f4');
INSERT INTO evaluations VALUES(8033,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8033,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8033,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8033,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8033,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(8033,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(8033,36,'cp 0','e8g8 e1g1 c6a5 c4d3 d7d5 e4e5 f6d7 f1e1 a5c6 c1f4 d5d4 c3b5 a7a6 b5d4 c6d4 f3d4 d7e5 f4e5 g7e5 e1e5 d8d4 e5e7 d4b2 a2a4 c8e6 a1b1 b2a3 e7b7 a3a4 d1e2 a6a5 e2e5 f8d8 b7b8 a8b8 b1b8 d8b8 e5b8 g8g7 b8e5 g7g8');
INSERT INTO evaluations VALUES(19811,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(19811,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(19811,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(19811,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(19811,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(19811,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(19811,36,'cp -39','e7e5 e1g1 f8e7 e2e4 d5d4 b1d2 c8e6 d2c4 d8c7 c1d2 h7h5 h2h4 f7f6 a2a4 g8h6 c2c3 h6f7 a4a5 f7d6 c4d6 c7d6 c3d4 c5d4 f3e1 d6c5 f2f4 c5b5 g2f3 e6f7 f4e5 c6e5 d2f4 a7a6 f3g2 f7e6 a1c1 g7g5 f4e5 f6e5');
INSERT INTO evaluations VALUES(20546,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(20546,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(20546,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(20546,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(20546,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(20546,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(20546,36,'cp -17','e1g1 f8e7 e2e4 d5d4 b1a3 c8e6 a3c4 d8c7 c1d2 h7h5 c2c3 e6c4 d3c4 h5h4 c3d4 c6d4 f3d4 c5d4 d1a4 e8f8 a1c1 g8f6 c4c5 c7c6 a4c6 b7c6 c1c4 h4h3 g2f3 g7g6 b2b4 a7a5 b4a5 f6d7 f1c1 f8g7 f3g4 d7f6 g4f3');
INSERT INTO evaluations VALUES(27466,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(27466,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(27466,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(27466,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(27466,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(27466,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(27466,36,'cp -41','f6d5 c3d5 d7d5 d1f3 d5d7 f3h3 f7f6 h3d7 e8d7 g5e4 f6f5 e4g3 g7g6 d2d3 a5c6 c2c3 a8e8 g3f1 h7h6 h2h4 f8d6 g2g3 c6e7 a2a4 e7d5 a4a5 d5f6 a1a4 b7b5 a5b6 a7b6 e1e2 d7e6 c1e3 e8a8 a4a8 h8a8 e3h6 a8a2 h6c1 d6a3 f2f4 a3b2 c1b2 a2b2 f1d2 e5f4');
INSERT INTO evaluations VALUES(27471,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(27471,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(27471,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(27471,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(27471,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(27471,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(27471,36,'cp -41','c3d5 d7d5 d1f3 d5d7 f3h3 f7f6 h3d7 e8d7 g5e4 f6f5 e4g3 g7g6 d2d3 a5c6 c2c3 a8e8 e1d1 h7h5 h2h4 f8c5 c1e3 c5e3 f2e3 c6e7 d1d2 d7e6 a1e1 e8g8 g3e2 e7d5 c3c4 d5f6 b2b4 g8d8 h1f1 d8d7 d3d4 h8d8 a2a4 f6e4 d2c2 e5d4 e3d4 d7d4 e2d4 d8d4 a4a5');
INSERT INTO evaluations VALUES(27470,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(27470,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(27470,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(27470,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(27470,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(27470,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(27470,36,'cp -46','d7d5 d1f3 d5d7 f3h3 f7f6 h3d7 e8d7 g5e4 f6f5 e4g3 g7g6 d2d3 f8d6 c1h6 a5c6 a2a4 h8e8 e1d1 c6d4 g3e2 d4e6 h2h4 b7b6 c2c3 e6c5 d1c2 e5e4 d3e4 e8e4 e2d4 e4g4 g2g3 a8e8 a1d1 f5f4 h6f4 d6f4 g3f4 g4f4 f2f3 c5a4');
INSERT INTO evaluations VALUES(27455,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(27455,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(27455,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(27455,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(27455,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(27455,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(27455,36,'cp -44','d1f3 d5d7 f3h3 f7f6 h3d7 e8d7 g5e4 f6f5 e4g3 g7g6 d2d3 f8d6 e1d1 h7h5 h2h3 a5c6 c2c3 b7b5 g3f1 a7a5 a2a4 b5b4 f1d2 g6g5 d2c4 g5g4 c1d2 a8f8 d1c2 h5h4 a1f1 g4g3 f2f4 e5e4 d3e4 f5e4 h1g1 d6c5 d2e3 c5e3 c4e3 c6e7 f4f5');
INSERT INTO evaluations VALUES(35257,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(35257,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(35257,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(35257,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(35257,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(35257,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(35257,36,'cp 154','g1f3 d7d5 b1c3 d5e4 c3e4 g8f6 e4f6 g7f6 f1d3 c7c5 c2c3 b8c6 d4c5 e7c5 c1e3 c5h5 d3e2 h5g6 d1a4 f8h6 f3h4 g6g7 f2f4 e8g8 e1g1 e6e5 a4e4 e5f4 e3f4 c8d7 f4h6 g7h6 h4f5 d7f5 e4f5 c6e5');
INSERT INTO evaluations VALUES(35256,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(35256,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(35256,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(35256,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(35256,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(35256,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(35256,36,'cp 161','d7d5 b1c3 d5e4 c3e4 g8f6 f1d3 f6e4 d3e4 b8d7 c2c3 d7f6 e4c2 c7c5 e1g1 e7c7 f3e5 f8d6 f1e1 c8d7 c1f4 e8c8 d1e2 d7e8 d4c5 c7c5 a1d1 e8b5 e2f3 d6e5 f4e5 d8d1 c2d1 b5c6 f3f4 h8d8 e5f6 g7f6 f4f6 c5d6 d1f3 d6d2 e1b1 c6f3');
INSERT INTO evaluations VALUES(37567,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(37567,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(37567,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(37567,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(37567,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(37567,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(37567,36,'cp 100','d7d6 b1c3 e7e6 d1e2 a7a6 h2h4 h7h6 a2a3 b8d7 c1e3 c7c5 c4a2 d8c7 e1c1 c8b7 d4c5 d6c5 e2d2 g8f6 e3f4 e6e5 f4g3 b6b5 c3d5 f6d5 a2d5 b7d5 d2d5 a8a7 d5d6 c7d6 d1d6 e8g8 h1d1 d7f6 d1e1 f6g4 h4h5 g6g5 f3h2 g4h2 g3h2 c5c4 e1d1');
INSERT INTO evaluations VALUES(48168,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(48168,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(48168,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(48168,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(48168,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(48168,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(48168,36,'cp -54','b7b6 c5b6 a7b6 e2e3 f8b4 c1d2 b4d6 a2a4 f6e4 f1e2 e4d2 d1d2 e8g8 e1g1 b8c6 b1a3 c6b4 b2b3 c7c6 a3c2 b4c2 d2c2 c8b7 e2d3 g7g6 e3e4 d6b4 e4e5 c6c5 d3b5 d8e7 c2e2 a8c8');
INSERT INTO evaluations VALUES(43790,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(43790,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(43790,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(43790,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(43790,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(43790,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(43790,36,'cp -53','c5b6 a7b6 e2e3 f8b4 c1d2 b4d6 a2a4 e8g8 f1e2 c7c5 e1g1 b8c6 b1c3 d6b8 c3b5 f6e4 d2e1 d8e7 f3d2 e4f6 f2f4 c5d4 e3d4 c8a6 d2b1 f6e4 b1c3 e7d7 a1c1 h7h6 e2d3 f7f5 e1h4 b8d6 g1h1 c6e7 b5d6 a6d3 d1d3 d7d6 h4e7 d6e7');
INSERT INTO evaluations VALUES(52770,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(52770,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(52770,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(52770,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52770,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(52770,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52770,36,'cp 13','c8f5 c1f4 e7e6 d1b3 b7b6 e2e3 g8f6 f1e2 a7a5 a2a4 b8d7 h2h3 f6e4 b1d2 g7g5 f4h2 h7h5 d2e4 f5e4 h2g3 h8g8 g3h2 h5h4 b3d1 f7f5 e2d3 e4f3 d1f3 d7f6 g2g4 h4g3 f3g3 c6c5 f2f3 c5c4 d3c2 e8f7 g3c7 f7g6');
INSERT INTO evaluations VALUES(55271,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(55271,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(55271,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(55271,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(55271,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(55271,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(55271,36,'cp -43','g8f6 e4e5 f6d5 c2c4 d5b6 d2d3 b8c6 f2f4 f7f6 g1f3 f6e5 f3e5 c6e5 f4e5 g7g6 f1e2 b6d7 e5e6 d7e5 e1g1 f8g7 c1f4 e8g8 e2g4 d8d6 f4e5 d6e5 f1e1 e5d6 b1d2 g7h6 g1h1 h6e3 d2e4 d6b6 d1b3 b6b3 a2b3 a7a5');
INSERT INTO evaluations VALUES(56586,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(56586,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(56586,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(56586,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(56586,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(56586,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(56586,36,'cp -75','f3f4 e5f4 c1f4 g8e7 f4g3 e7g6 c2c3 c7c5 c3d4 c5d4 g1f3 h7h5 f1e2 b8c6 h2h3 h5h4 g3h2 f8d6 e1g1 e8g8 d1d2 d6h2 f3h2 c8e6 b1a3 d8b6 h2f3 f7f6 a3c2 f8d8 c2e1 a7a5 e2d1 a5a4 d2f2 c6e5 f3h4 g6h4 f2h4');
INSERT INTO evaluations VALUES(8003,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8003,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8003,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8003,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8003,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8003,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8003,36,'cp 0','c1e3 d7d6 h2h3 e8g8 e1g1 b7b6 d1d2 c8b7 f1e1 e7e6 e3g5 d8d7 a2a4 a7a6 c4b3 d7c7 c3e2 f6d7 a1d1 b6b5 g5h6 c6e5 f3e5 d6e5 a4b5 a6b5 e2c3 b7c6 c3e2');
INSERT INTO evaluations VALUES(16794,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16794,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(16794,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16794,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(16794,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16794,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16794,36,'cp 0','d1e2 e4c5 e1g1 c5e6 c2c3 d7d5 e5d6 f8d6 c1g5 d6e7 g5e7 d8e7 f1d1 e6f4 e2e7 e8e7 f3d4 c6d4 c3d4 c7c6 g2g3 c8e6 c4e6 f4e6 b1c3 e6c7 d4d5 c7d5 c3d5 c6d5 d1d5 h8d8 d5b5 d8d7 a1c1 e7f6 g1g2 a8e8 g2f3 b7b6 c1c2 g7g6 a2a4 e8e5 c2c6 e5e6 c6c2 e6e5');
INSERT INTO evaluations VALUES(20217,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(20217,31,'cp -175',NULL);
INSERT INTO evaluations VALUES(20217,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(20217,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(20217,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(20217,35,'cp -155',NULL);
INSERT INTO evaluations VALUES(20217,36,'cp -165','e6d5 c4b5 f8d6 e1g1 g8e7 d2d4 d8b6 b5c6 b7c6 c1e3 e7f5 d4c5 f5e3 c5b6 e3d1 f1e1 c8e6 e1d1 a7b6 f3e5 a8c8 d1e1 e8g8 g2g3 f7f6 e5f3 e6d7 b1d2 c8a8 a2a3 g8f7 c2c3 h7h5 a3a4 c6c5 b2b3 d6e7 g1f2 f8e8');
INSERT INTO evaluations VALUES(20604,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(20604,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(20604,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(20604,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(20604,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(20604,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(20604,36,'cp -3','g7g6 g2g3 g8e7 f1g2 f8g7 e1g1 e8g8 a1b1 c8e6 c1e3 h7h6 f3d2 a7a5 g1h1 g8h7 c3d5 a8b8 f2f4 f7f5 a2a4 e6d7 b2b3 b7b6 d2f3 f8f7 b1c1 f5e4 d3e4 c6d4 e3d4 c5d4 d1d3 e7d5 c4d5 d7g4');
INSERT INTO evaluations VALUES(29563,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(29563,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(29563,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(29563,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(29563,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(29563,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(29563,36,'cp -21','e7e6 c1g5 b8d7 e2e3 f8e7 f1d3 f5g6 e1g1 e8g8 d3g6 h7g6 c3b1 c7c5 c2c3 d8b6 d1b3 f8c8 b1d2 e7d6 f1c1 a7a5 b3b6 d7b6 a2a4 f6d7 h3h4 f7f6 g5f4 d6f4 e3f4 c5d4 c3d4 c8c1 a1c1 a8c8 c1c8 b6c8 d2b3 b7b6 b3c1 g8f7 c1d3 c8d6 b2b3');
INSERT INTO evaluations VALUES(30116,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(30116,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(30116,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(30116,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(30116,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(30116,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(30116,36,'cp -12','c1g5 f8e7 e2e3 b8d7 f1d3 f5g6 e1g1 e8g8 d3g6 h7g6 c3b1 a7a6 c2c3 c7c5 b1d2 d8c7 g5f4 c7b6 d1c2 b6c6 a2a4 f8c8 f1e1 b7b5 a4b5 a6b5 d4c5 d7c5 f4e5 f6d7 a1a8 c6a8 e5d4 e7f6 e1d1 e6e5 d4c5');
INSERT INTO evaluations VALUES(37236,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37236,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37236,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37236,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37236,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37236,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37236,36,'cp 0','g1e2 c7c6 c1g5 c5e7 b1c3 d7d6 a2a4 c8g4 f3g3 g4e2 c3e2 d6d5 e4d5 c6d5 c4a2 b8d7 g5h6 d8a5 c2c3 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(40446,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(40446,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(40446,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(40446,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(40446,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(40446,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(40446,36,'cp 72','f3d2 e8g8 f1e2 b8a6 e1g1 a6c7 a2a4 b7b6 f2f3 d8e7 d2c4 c8a6 c1g5 h7h6 g5e3 a6c4 e2c4 a7a6 d1d2 b6b5 a4b5 a6b5 c4b5 a8a1 f1a1 c7b5 c3b5 f8b8 a1a7 f6d7 d2e2 e7d8 g1h1 b8b6 b2b3 d8b8 a7a5 b8d8 e3f4 d7e5 h2h3 g6g5 f4g3 e5g6 e2f1 d8b8 b5d6 b6d6');
INSERT INTO evaluations VALUES(42999,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(42999,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42999,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42999,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(42999,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42999,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(42999,36,'cp 20','e8g8 f1e2 c7c5 d4c5 d8a5 c2c3 a5c5 e1g1 d7d6 c3c4 e6e5 f4g3 c8e6 f3d2 d6d5 g3e5 d5c4 e5f6 e7f6 d2e4 c5e7 e4f6 e7f6 b1c3 b8c6 d1c2 f6e5 f1d1 a8c8 c2a4 a7a6 e2f3 h7h6 a2a3 f8d8 a1c1 g7g5 d1d8 c8d8 c1d1 d8d1 a4d1 g8g7');
INSERT INTO evaluations VALUES(37309,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(37309,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37309,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(37309,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(37309,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(37309,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(37309,36,'cp 20','f1e2 d7d5 e1g1 b7b6 c2c4 c8b7 b1c3 d5c4 e2c4 a7a6 d1e2 b6b5 c4b3 c7c5 d4c5 b8d7 f1d1 d8e8 c5c6 b7c6 a2a3 a8d8 b3a2 c6f3 e2f3 d7c5 d1d8 e8d8 a1d1 d8a8 f3a8 f8a8 f4e5 g8f8 e5d4 a8c8 b2b4 c5d7 f2f3');
INSERT INTO evaluations VALUES(55617,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(55617,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(55617,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(55617,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(55617,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(55617,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(55617,36,'cp 36','f1g2 f8e7 g1e2 c7c5 e1g1 g8f6 d2d3 b8c6 b1c3 e8g8 a2a4 a7a6 f2f4 c8d7 h2h3 h7h6 g1h2 f8e8 f4f5 b7b6 g3g4 c6b4 e2g3 f6h7 c1d2 e7h4 c3d5 h4g3 h2g3 b4d5 e4d5 b6b5 a4b5 a6b5 a1a8 d8a8 g2e4');
INSERT INTO evaluations VALUES(55634,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55634,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55634,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55634,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(55634,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(55634,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55634,36,'cp -12','b8c6 g1f3 e5e4 f3d4 c6d4 b2d4 d6d5 f1g2 h7h5 d2d3 f7f5 h2h4 g8f6 d1d2 f8e7 c2c4 d5c4 d3c4 c7c5 d4e3 c8e6 b1c3 f6g4 d2d8 a8d8 e3f4 e7f6 a1c1 d8d7 c3b5 a7a6 b5c7 e8f7 f2f3 e4f3 g2f3 g4e5 c7e6 e5f3 e1f2 f7e6 e2f3 d7d3 f2g2');
INSERT INTO evaluations VALUES(21997,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(21997,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(21997,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(21997,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21997,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(21997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21997,36,'cp 0','g1f3 e5e4 f3d4 c6d4 b2d4 d6d5 f1g2 h7h5 d2d3 f7f5 h2h4 g8f6 e1g1 c7c5 d4e3 f8d6 d3e4 f5e4 c2c4 d6e5 b1d2 e5a1 d1a1 e8g8 e3c5 f8e8 c5d4 c8g4 c4d5 d8d5 d4f6 g7f6 a1f6 d5d2 f6g6 g8h8 g6f6');
INSERT INTO evaluations VALUES(10884,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(10884,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(10884,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(10884,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(10884,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(10884,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(10884,36,'cp -31','e8g8 e1g1 c6a5 c4b3 d7d5 b1c3 a5b3 a2b3 c7c6 g5f3 d5e4 c3e4 f6e4 d3e4 d8c7 h2h3 f8e8 c1d2 b7b6 d2c3 f7f6 f3h4 a7a5 h4f5 e7f8 d1f3 c8e6 f1d1 a8d8 c3d2 d8d7 d2e3 d7d1 a1d1 e8d8 d1a1 c6c5 h3h4 c7d7 f5g3 e6f7 g3f5');
INSERT INTO evaluations VALUES(4572,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(4572,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(4572,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(4572,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(4572,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(4572,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(4572,36,'cp -36','e1g1 c6a5 c4b3 d7d5 b1c3 c7c6 g5f3 d5e4 c3e4 a5b3 a2b3 f6e4 d3e4 d8c7 h2h3 a7a5 d1e2 f8e8 f1d1 b7b6 c1d2 e7f8 d2c3 f7f6 f3h4 c8e6 h4f5 a8d8 d1d2 d8d2 c3d2 e8a8 d2c3 a8d8 c3d2 d8d7 f5e3 g8h8 e3c4');
INSERT INTO evaluations VALUES(4571,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(4571,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(4571,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(4571,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(4571,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(4571,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(4571,36,'cp -34','c6a5 c4b3 d7d5 b1c3 a5b3 a2b3 c7c6 g5f3 d5e4 c3e4 f6e4 d3e4 d8c7 h2h3 f8e8 d1e2 e7f8 f1d1 b7b6 c1d2 a7a5 f3h4 c8e6 h4f5 e8d8 d2c3 g8h8 d1d8 a8d8 e2e1 d8a8 e1e2 h7h6 a1d1 h8h7 c3d2 a8d8 d2c3 d8d1 e2d1 f8c5 d1e2 g7g6 f5e3 f7f6 e3g4');
INSERT INTO evaluations VALUES(14395,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(14395,31,'cp -175',NULL);
INSERT INTO evaluations VALUES(14395,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(14395,33,'cp -181',NULL);
INSERT INTO evaluations VALUES(14395,34,'cp -183',NULL);
INSERT INTO evaluations VALUES(14395,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(14395,36,'cp -168','d8d4 d1d4 c6d4 f1d3 e7e5 g1f3 d4f3 g2f3 f8c5 c3b5 e8d8 a2a4 c7c6 b5a3 f6h5 a3c4 f7f6 c2c3 d8c7 b2b4 c5e7 d3f1 h8d8 c1e3 h5f4 h1g1 g7g5 b4b5 c8e6 b5b6 a7b6 e3b6 c7c8 b6d8 c8d8 e1d2 d8c7 c4b2 b7b5 a4a5 e7c5 g1h1 h7h5 h2h4');
INSERT INTO evaluations VALUES(16927,30,'cp 276',NULL);
INSERT INTO evaluations VALUES(16927,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(16927,32,'cp 259',NULL);
INSERT INTO evaluations VALUES(16927,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(16927,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(16927,35,'cp 245',NULL);
INSERT INTO evaluations VALUES(16927,36,'cp 250','c2c3 b4a6 f1a6 b7a6 f3e5 d8e7 d1d4 c7c6 d5c6 d7c6 b1d2 e7c7 d2b3 c8h3 g2h3 a8d8 d4c4 c7e5 c4c6 d8d7 c1e3 e5e4 c6e4 f6e4 e3d4 f8e7 e1c1 e8g8 h1e1 f7f5 d4e3 d7d1 c1d1 e7h4 e1e2 e4g5 e3g5 h4g5 e2e6 f8f6 e6f6 g5f6 b3d4 f6e5 d4f5 e5h2 d1e2 h2f4 e2f3 f4e5');
INSERT INTO evaluations VALUES(19164,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(19164,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(19164,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(19164,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(19164,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(19164,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(19164,36,'cp 181','a7a6 b1c3 d7d6 c1f4 c8e6 c4d3 g8f6 e1c1 b7b5 f1e2 b5b4 c3a4 f8e7 f4d6 c5c4 d3d4 b4b3 a2b3 c4b3 a4b6 e7d6 d4d6 d8d6 d1d6 e8e7 d6c6 a8b8 h1d1 f6e4 f2f3 b3c2 c1c2 e4f6 c6c7 e7f8 b6a4 e6f5 c2c1');
INSERT INTO evaluations VALUES(20070,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(20070,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(20070,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(20070,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(20070,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(20070,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(20070,36,'cp 66','f3d4 g8f6 b1c3 c6d4 d1d4 g7g6 f1e2 f8g7 c1e3 e8g8 d4d3 a7a6 e1g1 c8d7 a2a4 d7e6 a1b1 a8c8 b2b3 f6d7 d3d2 d7c5 e3d4 g7d4 d2d4 a6a5 d4e3 d8b6 h2h4 c5d7 e3g3 g8h8 h4h5 g6h5 c3d5 e6d5 e4d5 b6b4 e2h5');
INSERT INTO evaluations VALUES(38929,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38929,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38929,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(38929,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(38929,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(38929,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(38929,36,'cp 0','g8f6 f1d3 e8g8 e1g1 c8g4 c1g5 c7c6 h2h3 g4h5 g2g4 h5g6 f3e5 b8d7 e5d7 d8d7 g5f6 g7f6 d1f3 a8e8 c3e2 b4d6 a1e1 g6d3 f3d3 e8e4 e2c3 e4g4 h3g4 d7g4 g1h1 g4h4 h1g2 h4g4');
INSERT INTO evaluations VALUES(41304,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41304,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41304,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41304,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(41304,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41304,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(41304,36,'cp -2','e8g8 h2h3 b7b6 a2a4 c8b7 f1e2 b8d7 e1g1 a7a6 d1d3 e7e6 e3e4 d8e7 d3e3 e6e5 d4e5 d6e5 f4g5 e7b4 f3d2 d7c5 a4a5 c5e6 g5f6 g7f6 c3d5 b7d5 e4d5 e6f4 e2f3 f6g5 e3e5 b4d2 a1d1 a8e8 e5e8 f4h3 g2h3 f8e8 d1d2 g5d2 a5b6 c7b6');
INSERT INTO evaluations VALUES(36594,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(36594,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36594,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(36594,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36594,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36594,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36594,36,'cp 0','f1e2 b7b6 a2a4 c8b7 e1g1 a7a6 h2h3 e7e6 f4h2 d8e7 f1e1 f8e8 e2f1 d6d5 c3b1 c7c5 b1d2 e8d8 c2c3 b8d7 h2c7 d8c8 c7g3 b7c6 g3h4 e7d6 h4g3');
INSERT INTO evaluations VALUES(45181,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(45181,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(45181,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(45181,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(45181,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(45181,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(45181,36,'cp -2','b1c3 b8c6 f1b5 e7e5 d2d3 f8e7 h2h3 e8g8 e1g1 a7a6 b5c6 b7c6 f3d2 d8c7 d1e2 c8e6 d2c4 d6d5 e4d5 c6d5 c4e5 a8e8 f2f4 d5d4 c3a4 e7d6 e2f2 f6d7 e5c4 e6c4 b3c4 c7c6 a4b2 f7f5 f2f3 c6f3 f1f3');
INSERT INTO evaluations VALUES(45184,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(45184,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(45184,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(45184,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(45184,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(45184,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(45184,36,'cp -16','b8c6 f1b5 c8d7 e1g1 g7g6 f1e1 e7e5 d2d3 f8g7 a2a4 a7a6 b5c6 b7c6 a4a5 e8g8 c1d2 a8b8 c3a4 d7e6 a4b6 d8c7 b3b4 c5b4 d2b4 f8e8 h2h3 h7h6 a1b1 c6c5 b4d2 f6d7 b6d7 e6d7 c2c4 b8b1 d1b1 e8b8 b1c2 c7b7 d2c3 d7c6 f3d2');
INSERT INTO evaluations VALUES(47807,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(47807,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(47807,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(47807,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(47807,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(47807,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(47807,36,'cp -101','f6e4 d1e2 e4g5 b1c3 g5f3 e2f3 d8d7 d2d4 b8c6 d4e5 c6e5 f3e2 f8e7 c4b3 e8g8 c1e3 c7c6 e1c1 f8e8 f2f4 e5g6 e2d2 d7c7 g2g4 b7b5 h1e1 d6d5 f4f5 g6e5 d2f2 a7a5 e3b6 c7b8 b6d4 a5a4 e1e5 a4b3 e5e7 e8e7 c2b3 b5b4 c3a4');
INSERT INTO evaluations VALUES(48004,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(48004,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(48004,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(48004,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(48004,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(48004,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(48004,36,'cp -90','d1e2 e4g5 b1c3 g5f3 e2f3 d8d7 d2d4 b8c6 d4e5 c6e5 f3e2 f8e7 c4b3 a7a5 a2a4 b7b6 c1f4 e8g8 f4e5 d6e5 a1d1 e7d6 e1g1 c8b7 b3d5 b7d5 d1d5 c7c6 d5d3 d7e6 f1d1 d6e7 d3e3 f7f6 g2g4 e7c5 e3g3 f6f5 g4f5 f8f5 c3e4 a8f8 e4c5 b6c5 d1d2 f5f4');
INSERT INTO evaluations VALUES(51583,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(51583,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(51583,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(51583,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(51583,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(51583,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(51583,36,'cp 49','c8b7 e1g1 e7e6 f1e1 d8c7 b1c3 b8c6 d2d4 c6d4 f3d4 c5d4 d1d4 c7d6 d4d3 d6d3 c4d3 f8c5 a2a3 g8f6 c3a4 c5d4 c2c3 d4e5 f2f4 e5c7 c3c4 d7d6 b2b4 e8g8 a4c3 f8c8 c1e3 d6d5 c4d5 e6d5 e4d5 f6d5 c3d5 b7d5 d3a6');
INSERT INTO evaluations VALUES(30496,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(30496,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(30496,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(30496,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(30496,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(30496,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(30496,36,'cp 57','e1g1 e7e6 f1e1 d8c7 b1c3 b8c6 c4f1 g8e7 d2d4 c6d4 f3d4 c5d4 c3b5 c7b8 d1d4 e7c6 d4d3 f7f6 c1e3 c6e5 d3c3 b7c6 f2f3 a7a6 b5d4 c6b7 a1d1 f8d6 c3d2 e5f7 h2h3 e8g8 a2a4 d6h2 g1h1 h2g3 e3f2 f8c8 f2g3 b8g3 h1g1');
INSERT INTO evaluations VALUES(54183,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54183,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54183,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(54183,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54183,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54183,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54183,36,'cp 5','g1f3 d5c4 e2e3 g8f6 f1c4 a7a6 c4e2 b7b5 e1g1 b8d7 b2b3 c5d4 f3d4 c8b7 c1b2 f8e7 b1d2 e8g8 e2f3 d8b6 b3b4 a8c8 d4b3 b7f3 d2f3 c8c7 a1c1 f8c8 c1c7 c8c7 d1a1 e7f8 b3a5 h7h6 f1c1 f6d5 c1c7 b6c7');
INSERT INTO evaluations VALUES(57477,30,'cp 833',NULL);
INSERT INTO evaluations VALUES(57477,31,'cp 839',NULL);
INSERT INTO evaluations VALUES(57477,32,'cp 832',NULL);
INSERT INTO evaluations VALUES(57477,33,'cp 842',NULL);
INSERT INTO evaluations VALUES(57477,34,'cp 840',NULL);
INSERT INTO evaluations VALUES(57477,35,'cp 840',NULL);
INSERT INTO evaluations VALUES(57477,36,'cp 840','d1h5 g7g6 h5e5 d8e7 e5h8 g8f6 d2d3 e7g7 h8g7 f8g7 b1c3 d7d6 c3b5 e8d8 g1f3 c8e6 f1e2 b8c6 e1g1 a7a6 b5c3 b7b5 c1e3 d8d7 f3g5 e6g8 a2a3 h7h6 g5h3 c6e5 h3f4 g6g5 f4d5 g8d5 e4d5 g5g4 h2h3 g4g3 g1h1 a8f8 a3a4');
INSERT INTO evaluations VALUES(57468,30,'cp 817',NULL);
INSERT INTO evaluations VALUES(57468,31,'cp 816',NULL);
INSERT INTO evaluations VALUES(57468,32,'cp 828',NULL);
INSERT INTO evaluations VALUES(57468,33,'cp 825',NULL);
INSERT INTO evaluations VALUES(57468,34,'cp 827',NULL);
INSERT INTO evaluations VALUES(57468,35,'cp 827',NULL);
INSERT INTO evaluations VALUES(57468,36,'cp 832','g7g6 h5e5 d8e7 e5h8 e7h4 e1d1 h4f6 h8f6 g8f6 b1c3 f6g4 d1e2 f8g7 g1f3 d7d6 h2h3 g4f6 d2d4 c8e6 e2f2 b8d7 c1e3 a7a6 f1d3 e8c8 g2g4 e6f7 f3g5 f7g8 f2g3 c7c5 g5f3 h7h6 e4e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(57543,30,'cp 840',NULL);
INSERT INTO evaluations VALUES(57543,31,'cp 840',NULL);
INSERT INTO evaluations VALUES(57543,32,'cp 840',NULL);
INSERT INTO evaluations VALUES(57543,33,'cp 840',NULL);
INSERT INTO evaluations VALUES(57543,34,'cp 843',NULL);
INSERT INTO evaluations VALUES(57543,35,'cp 848',NULL);
INSERT INTO evaluations VALUES(57543,36,'cp 844','h5e5 d8e7 e5h8 g8f6 d2d3 e7g7 h8g7 f8g7 b1c3 b8c6 c3b5 e8d8 g1f3 d7d6 f1e2 c8e6 e1g1 d8d7 c2c3 a8e8 b5d4 e6g8 c1g5 h7h6 g5h4 e8f8 d4c6 b7c6 f3d2 g8e6 d3d4 f6h7 f1f8 h7f8 b2b4 c6c5 b4c5 d6c5');
INSERT INTO evaluations VALUES(57540,30,'cp 828',NULL);
INSERT INTO evaluations VALUES(57540,31,'cp 828',NULL);
INSERT INTO evaluations VALUES(57540,32,'cp 833',NULL);
INSERT INTO evaluations VALUES(57540,33,'cp 827',NULL);
INSERT INTO evaluations VALUES(57540,34,'cp 838',NULL);
INSERT INTO evaluations VALUES(57540,35,'cp 832',NULL);
INSERT INTO evaluations VALUES(57540,36,'cp 832','d8e7 e5h8 e7h4 e1d1 h4f6 h8f6 g8f6 b1c3 c7c6 d2d4 d7d6 h2h3 c8e6 f1d3 b8d7 g1f3 e8c8 c1e3 d8e8 d1d2 c6c5 a1e1 c5d4 e3d4 a7a6 g2g4 f8e7 e1f1 e6g8 a2a3 d7c5 b2b4 c5d3 c2d3 e8f8');
INSERT INTO evaluations VALUES(35244,30,'cp 840',NULL);
INSERT INTO evaluations VALUES(35244,31,'cp 840',NULL);
INSERT INTO evaluations VALUES(35244,32,'cp 843',NULL);
INSERT INTO evaluations VALUES(35244,33,'cp 840',NULL);
INSERT INTO evaluations VALUES(35244,34,'cp 848',NULL);
INSERT INTO evaluations VALUES(35244,35,'cp 850',NULL);
INSERT INTO evaluations VALUES(35244,36,'cp 843','e5h8 e7h4 e1d1 h4f6 h8f6 g8f6 b1c3 b8c6 g1f3 d7d6 d2d4 f8g7 h2h3 c8d7 c1g5 e8c8 e4e5 d6e5 d4e5 d7f5 d1c1 f6d7 g5d8 c8d8 g2g4 f5e6 f1g2 d7e5 f3e5 c6e5 g2b7 c7c6 b7a8 g7h6 c1b1 e6d5 c3d5 c6d5 a8d5 h6d2');
INSERT INTO evaluations VALUES(34108,30,'cp 832',NULL);
INSERT INTO evaluations VALUES(34108,31,'cp 840',NULL);
INSERT INTO evaluations VALUES(34108,32,'cp 832',NULL);
INSERT INTO evaluations VALUES(34108,33,'cp 835',NULL);
INSERT INTO evaluations VALUES(34108,34,'cp 835',NULL);
INSERT INTO evaluations VALUES(34108,35,'cp 832',NULL);
INSERT INTO evaluations VALUES(34108,36,'cp 836','g8f6 d2d3 b8c6 b1c3 e7g7 h8g7 f8g7 c3b5 e8d8 g1f3 d7d6 f1e2 c8e6 e1g1 a7a6 b5d4 c6d4 f3d4 e6g8 c2c3 d8e7 c1e3 a8f8 d4b3 g8e6 b3d2 h7h6 a2a4 g6g5 d3d4 e6f7 a4a5 f7g6 e4e5 d6e5 d4e5 f6d7 f1f8');
INSERT INTO evaluations VALUES(1997,30,'cp -435',NULL);
INSERT INTO evaluations VALUES(1997,31,'cp -448',NULL);
INSERT INTO evaluations VALUES(1997,32,'cp -469',NULL);
INSERT INTO evaluations VALUES(1997,33,'cp -442',NULL);
INSERT INTO evaluations VALUES(1997,34,'cp -469',NULL);
INSERT INTO evaluations VALUES(1997,35,'cp -468',NULL);
INSERT INTO evaluations VALUES(1997,36,'cp -467','d7d5 g4e3 d5c4 d2d3 c4d3 d1d3 d4e6 b1c3 c8d7 c1d2 e8c8 e1c1 g8f6 h2h4 g5c5 e3d5 c5f2 h4h5 f6d5 e4d5 e6c5 d3c4 f2g2 h1g1 g2h3 c1b1 h3h5 d2e3 b7b6 e3c5 f8c5 d5d6 c7d6 b2b4 d7e6 c4a6 c8b8 b4c5 h5c5 g1g3 c5b4 b1c1');
INSERT INTO evaluations VALUES(1767,30,'cp -463',NULL);
INSERT INTO evaluations VALUES(1767,31,'cp -466',NULL);
INSERT INTO evaluations VALUES(1767,32,'cp -447',NULL);
INSERT INTO evaluations VALUES(1767,33,'cp -438',NULL);
INSERT INTO evaluations VALUES(1767,34,'cp -463',NULL);
INSERT INTO evaluations VALUES(1767,35,'cp -448',NULL);
INSERT INTO evaluations VALUES(1767,36,'cp -456','g4e3 d5c4 d2d3 c4d3 d1d3 d4e6 h2h4 g5h5 b1c3 c8d7 g2g4 h5c5 e3f5 e8c8 c1e3 c5a5 e1c1 g8f6 f2f3 g7g6 f5d4 f8c5 d4b3 c5e3 d3e3 a5b6 e3b6 a7b6 a2a4 d7c6 b3d2 d8d4 a4a5 b6a5 d2b3 d4d1 h1d1 b7b6 c3e2 c6e4 f3e4 f6g4 b3d4 g4e3 d1d3 e3c4 d4e6 f7e6 e2d4');
INSERT INTO evaluations VALUES(11448,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(11448,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(11448,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(11448,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(11448,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(11448,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(11448,36,'cp -18','f8d6 d1e2 d8e7 c1d2 c8d7 h2h3 e8c8 e1c1 h8e8 f3h4 g7g6 g2g3 f6g8 c1b1 f7f5 h4f3 g8f6 d2g5 e7f8 f3d2 d6b4 a2a3 b4c3 b2c3 c6c5 e2e3 e8e6 c3c4 h7h6 g5f6 e6f6 e4f5 g6f5 d2b3 d8e8 e3c5 f8c5 b3c5 d7c6');
INSERT INTO evaluations VALUES(7226,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(7226,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(7226,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(7226,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(7226,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(7226,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(7226,36,'cp -10','d1e2 d8e7 c1d2 c8d7 h2h3 e8c8 e1c1 h8e8 f3h4 g7g6 g2g3 f6g8 h4f3 c8b8 g3g4 e7f8 c1b1 g8e7 g4g5 f7f5 g5f6 e7g8 f6f7 f8f7 d2g5 g8f6 f3d2 d7e6 d1f1 f7g7 e2e3 b7b6 f2f4 e5f4 g5f4 d6c5 e3e2 f6h5 d2b3 h5f4 f1f4 c5d4');
INSERT INTO evaluations VALUES(25109,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(25109,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(25109,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(25109,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(25109,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(25109,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(25109,36,'cp 105','f1e1 f8c5 c4b5 e8g8 b5c6 b7c6 e1e5 c5d6 e5e4 e6f5 e4d4 d8e7 b1d2 a8e8 h2h3 d5b6 d2e4 d6e5 f3e5 e7e5 d4b4 a7a5 c1f4 e5e7 b4b6 c7b6 f4d6 e7d8 d6f8 g8f8 e4c3 f5g6 b2b3 b6b5 d1f3 d8d7 a2a3 f7f6');
INSERT INTO evaluations VALUES(25110,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(25110,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(25110,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(25110,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(25110,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(25110,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(25110,36,'cp 73','f8c5 c4b5 e8g8 b5c6 b7c6 e1e5 c5d6 e5e4 e6f5 e4e1 f5g4 h2h3 g4h5 b1d2 d8d7 d2e4 a8e8 c1d2 d6f4 c2c4 h5f3 d1f3 f4d2 e4d2 d5b4 e1e8 f8e8 a1d1 b4a2 d3d4 d7d4 f3c6 e8f8 c6c7 a2b4 c7g3 d4b2 d2f1 b4a6 d1d7 a6c5 d7a7 b2b8 g3b8 f8b8 f1e3 b8b1 g1h2 b1b2 h2g3');
INSERT INTO evaluations VALUES(29660,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(29660,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(29660,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(29660,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29660,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29660,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(29660,36,'cp -5','g4f3 g2f3 c7c5 e2e4 e7e6 e4d5 f6d5 c3d5 d8d5 d4c5 d5d1 e1d1 f8c5 d1e2 b8c6 c2c3 e8c8 h1g1 g7g6 f1g2 d8d7 c1e3 c5e3 f2e3 h8d8 a1d1 d7d1 g1d1 d8d1 e2d1 c6e5 d1e2 g6g5 f3f4 g5f4 e3f4 e5g6 e2e3 b7b6 g2e4 f7f5 e4d3 c8d7 e3f3 d7e7 f3g3 e7f6 h3h4');
INSERT INTO evaluations VALUES(29676,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29676,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29676,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29676,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29676,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29676,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29676,36,'cp 0','g2f3 c7c5 e2e4 e7e6 e4d5 f6d5 c3d5 d8d5 d4c5 d5d1 e1d1 f8c5 d1e2 b8c6 c2c3 e8c8 h1g1 h8g8 f3f4 c6e7 f1g2 e7g6 g2e4 f7f5 e4c2 c5d6 e2f3 g6h4 f3e2');
INSERT INTO evaluations VALUES(29674,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29674,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29674,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29674,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29674,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29674,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(29674,36,'cp 2','e7e6 e2e4 f8b4 d1d3 c7c5 d4c5 b8c6 a2a3 b4c5 e4d5 f6d5 c3d5 e6d5 c1e3 d8a5 e3d2 a5b6 e1c1 e8c8 d2c3 f7f6 d3f5 c8b8 d1d5 d8d5 f5d5 c5f2 f1d3 h8d8 d5e4 h7h6 e4h7 b6c7 c1b1 a7a6 d3e4 f2d4 c3d4 c6d4 h1d1 d4e6 d1d8 e6d8 h7g8 b8a7 h3h4 d8c6');
INSERT INTO evaluations VALUES(31032,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(31032,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(31032,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(31032,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(31032,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(31032,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(31032,36,'cp -25','h7h5 g4h5 g6h5 f1e2 h5e2 d1e2 e7e6 g1f3 g8e7 h4h5 e7f5 b1d2 d8b6 c2c3 c6c5 d4c5 f8c5 d2b3 b8c6 c1g5 a8c8 b3c5 b6c5 e1f1 a7a6 a1c1 c6e7 b2b3 c5b5 e2b5 a6b5 f1g2 e7c6 a2a4 b5a4 b3a4 c6a5 f3d4 f5d4 c3d4 c8c1 g5c1 a5c4 c1g5');
INSERT INTO evaluations VALUES(30953,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(30953,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(30953,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(30953,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(30953,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(30953,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(30953,36,'cp -22','e5e6 h5g4 d1g4 g8h6 e6f7 g6f7 g4g5 b8d7 g1f3 e7e6 g5d8 a8d8 c1g5 f8e7 b1d2 h6f5 f1h3 f5h4 g5h4 e7h4 h3g2 g7g5 f3g5 e8e7 g5f7 e7f7 e1e2 d7f8 d2f3 f8g6 f3h4 h8h4 h1h4 g6h4 g2f3 e6e5 d4e5 h4g6 e5e6 f7e6 c2c3 d8e8 a1g1');
INSERT INTO evaluations VALUES(36262,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36262,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(36262,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(36262,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(36262,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(36262,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36262,36,'cp 39','f8e8 f1e1 a7a6 a2a4 b8c6 c1f4 e7f8 d4c6 b7c6 e4e5 d6e5 d1d8 e8d8 f4e5 c8f5 e5c7 d8d2 a1c1 a6a5 e2d1 d2d7 c7f4 d7b7 c1b1 a8d8 h2h3 f5g6 f4e3 h7h6 g2g3 f8b4 d1f3 g6c2 f3c6');
INSERT INTO evaluations VALUES(38856,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(38856,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(38856,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(38856,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(38856,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(38856,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(38856,36,'cp 71','g8f6 f1d3 b8c6 a2a3 b4e7 d2e3 a7a6 d1e2 b7b5 e1g1 c8b7 f1e1 e8g8 a1d1 f8e8 h2h3 c6b8 e3g5 b8d7 f3e5 e8f8 g5f4 d7b6 f4g3 a8c8 e5g4 f6g4 h3g4 e7d6 d3e4 c8b8 e4b7 b8b7 c3e4 d6g3 e4g3 d8d6');
INSERT INTO evaluations VALUES(42139,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(42139,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(42139,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(42139,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(42139,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(42139,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(42139,36,'cp 89','f1d3 b8c6 c3e4 b4e7 e4f6 e7f6 c2c3 e6e5 d4e5 d8d3 e5f6 g7f6 d2f4 d3d1 a1d1 c8e6 e1g1 e8g8 b2b3 a8c8 f1e1 a7a5 f3d2 f8d8 d2e4 g8g7 d1d8 c6d8 c3c4 a5a4 f2f3 g7g6 e1d1 d8c6 g1f2 h7h5 e4c3 a4b3 a2b3 b7b6 f4e3');
INSERT INTO evaluations VALUES(39733,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(39733,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(39733,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39733,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39733,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(39733,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(39733,36,'cp 0','g2g4 e8g8 c1e3 f6d7 h1g1 d7c5 g4g5 b7b5 h2h4 c8b7 h4h5 b8c6 d4c6 b7c6 d1g4 c5b3 a2b3 b5b4 c3e2 a6a5 e2d4 c6d7 f2f4 d8c7 g1g2 c7b7 f4f5 e6f5 d4f5 f8e8 f5h6 g7h6 g4f4 h6g5 g2g5 e7g5 f4g5 g8f8 g5h6 f8g8');
INSERT INTO evaluations VALUES(41362,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(41362,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(41362,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(41362,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(41362,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(41362,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(41362,36,'cp 57','c1e3 c7c6 a2a4 e8g8 d1d2 e7e5 e1c1 d8a5 c1b1 e5d4 e3d4 b8d7 g1e2 d7e5 h2h4 b7b5 a4b5 e5c4 d2g5 h7h6 g5f4 c4e5 e2c1 f6h5 c1b3 a5c3 f4e5 d6e5 d4c3 h5g3 h1g1 c6b5 b3c5 a8b8 f1d3 b5b4 c5a6');
INSERT INTO evaluations VALUES(41778,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(41778,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(41778,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(41778,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(41778,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(41778,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41778,36,'cp 23','c4c5 d6e7 b2b4 b7b6 c1b2 a7a5 a2a3 c7c6 b1d2 e8g8 f1e2 c8a6 e1g1 b8d7 e2a6 a8a6 d1c2 d8c7 f3e1 f8a8 e1d3 a5b4 a3b4 a6a1 b2a1 h7h6 f1b1 e7d8 a1b2 c7b7 d2b3 b6c5 b4c5 f6g4 h2h3 g4f6 b2c3 d8c7 c3e1');
INSERT INTO evaluations VALUES(41767,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41767,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41767,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(41767,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41767,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41767,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41767,36,'cp 16','d6e7 b2b4 b7b6 c1b2 a7a5 a2a3 e8g8 f1e2 c7c6 b1d2 b8d7 e1g1 c8a6 b2c3 d8b8 e2a6 a8a6 d1b3 b8c7 f1c1 a5b4 a3b4 f8a8 b3b2 h7h6 h2h3 c7b7 d2b3 f6e4 c5b6 a6a1 c1a1 a8a1 b2a1 e4c3 a1c3 b7b6 b3c5 d7c5 b4c5 b6b1 c3e1 b1e1 f3e1 e6e5 e1c2 e5d4 e3d4 e7f6 g1f1');
INSERT INTO evaluations VALUES(45878,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(45878,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(45878,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(45878,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(45878,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(45878,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(45878,36,'cp -18','a7a5 a2a3 b8d7 d2d4 f8e7 e1g1 b7b5 a4b3 e8g8 f1e1 h7h6 c1e3 f8e8 f3h4 e5d4 e3d4 a8b8 h4g6 c6c5 g6e7 d8e7 d4f6 e7f6 d1d2 c5c4 b3a2 f6g6 f2f3 d7e5 g1h1 c8e6 f3f4 b5b4 a3b4 a5b4 c3d5 b4b3 f4e5 b3a2 e5d6 e6d5 d2d5');
INSERT INTO evaluations VALUES(55074,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(55074,31,'cp -147',NULL);
INSERT INTO evaluations VALUES(55074,32,'cp -152',NULL);
INSERT INTO evaluations VALUES(55074,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(55074,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(55074,35,'cp -143',NULL);
INSERT INTO evaluations VALUES(55074,36,'cp -160','d5e4 d2d4 e7e5 f1c4 b8c6 g1e2 g8f6 c4b5 e5d4 c3d4 f8d6 b1c3 e8g8 c1g5 h7h6 g5f6 d8f6 c3e4 d6b4 e4c3 f8d8 b5c6 f6c6 e1g1 c8f5 d1b3 b4d6 f1d1 c6b6 c3d5 b6b3 a2b3 a7a6 d5e3 f5e6');
INSERT INTO evaluations VALUES(55285,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(55285,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(55285,32,'cp -156',NULL);
INSERT INTO evaluations VALUES(55285,33,'cp -150',NULL);
INSERT INTO evaluations VALUES(55285,34,'cp -145',NULL);
INSERT INTO evaluations VALUES(55285,35,'cp -144',NULL);
INSERT INTO evaluations VALUES(55285,36,'cp -130','d2d4 e7e5 f1b5 c7c6 b5a4 e5d4 d1d4 d8d4 c3d4 b8a6 b1c3 a6b4 c3e4 c8f5 e4c3 e8c8 g1f3 d8e8 e1f1 g8f6 a2a3 b4d5 c3d5 f6d5 c1d2 f8d6 a4b3 f7f6 h2h4 h7h5 g2g3 d6c7 f1g2 f5e4 h1e1 g7g5 h4g5 f6g5 e1e4');
INSERT INTO evaluations VALUES(11454,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11454,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11454,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11454,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11454,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(11454,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11454,36,'cp 0','d2d3 f8d6 b1d2 e8g8 e1g1 f8e8 d2c4 b7b5 c4e3 h7h6 a2a4 f6h5 e3g4 h5f6');
INSERT INTO evaluations VALUES(11450,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11450,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11450,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11450,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11450,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11450,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11450,36,'cp 0','f8d6 b1d2 e8g8 e1g1 f8e8 d2c4 b7b5 c4e3 h7h6 a2a4 f6h5 e3g4 h5f6');
INSERT INTO evaluations VALUES(7230,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7230,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7230,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7230,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7230,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7230,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7230,36,'cp 0','b1d2 e8g8 e1g1 f8e8 d2c4 h7h6 a2a4 b7b5 c4e3 f6h5 e3g4 h5f6');
INSERT INTO evaluations VALUES(34769,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(34769,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(34769,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(34769,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(34769,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(34769,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(34769,36,'cp 24','b1c3 g7g6 h5d1 c8e6 c3d5 f6d8 d2d4 c7c6 d5e3 e6c4 e3c4 b7b5 c4e3 g8f6 d4e5 f6e4 g1f3 f8g7 e5d6 d8d6 e1g1 b8a6 a2a4 b5b4 d1e2 a6c5 e3c4 d6e6 c1e3 e8g8 f3d4 e6d7 f1d1 d7c7 e2f3 f7f5 f3f4 f8f7 f4c7 f7c7 d4b3 c5b3 c2b3');
INSERT INTO evaluations VALUES(35768,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(35768,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(35768,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(35768,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(35768,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(35768,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(35768,36,'cp 37','f1d3 f8e8 d1c2 h7h6 g5f4 c8e6 e1g1 b8d7 h2h3 a8c8 a1c1 a7a6 a2a3 b7b5 c3e2 c6c5 d4c5 d7c5 b2b4 c5a4 c2b1 a4b6 c1c8 d8c8 f1c1 b6c4 f4e5 f6e4 d3e4 d5e4 b1e4 e7f8 e2d4 e6d7 e4f4 c8a8 d4f5 a8b7 f5d6 c4d6');
INSERT INTO evaluations VALUES(38057,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(38057,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(38057,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(38057,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(38057,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(38057,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(38057,36,'cp 148','a2a3 b4f8 f1d3 b7b6 c3e2 c8a6 d3a6 b8a6 c2c3 d8d7 g1f3 c6c5 b2b4 a6c7 e1g1 g8e7 d4c5 b6c5 e2g3 h7h5 h2h4 e7g6 c1e3 c5c4 a3a4 f8e7 b4b5 g6h4 f3h4 e7h4 g3h5 g7g6 h5f4 h4e7 d1f3 e8c8 e3a7 g6g5');
INSERT INTO evaluations VALUES(38728,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(38728,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(38728,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(38728,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(38728,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(38728,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(38728,36,'cp 77','d2d4 e5d4 c3d4 c5b6 b1c3 g8e7 h2h3 e8g8 f1c4 c7c6 e1g1 d6d5 c4b3 h7h6 c1f4 c8e6 f1e1 f8e8 f4g3 b8a6 e4d5 c6d5 b3a4 e6d7 f3e5 d7a4 c3a4 e7c6 d1f3 c6e5 e1e5 e8e5 g3e5 b6c7 a1e1 c7a5 a4c3 a5c3 b2c3 a6c7');
INSERT INTO evaluations VALUES(38722,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(38722,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(38722,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(38722,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(38722,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(38722,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(38722,36,'cp 86','e5d4 c3d4 c5b6 h2h3 g8e7 b1c3 e8g8 f1c4 c7c6 c1g5 d6d5 e4d5 h7h6 g5e7 d8e7 d1e2 e7e2 e1e2 f8e8 e2d2 g8f8 h1e1 e8d8 a2a3 b6c7 a1d1 b8d7 d5c6 b7c6 d2c2 d7b6 c4d3 c8e6 f3e5 d8d4 e5c6 d4f4 f2f3 b6d5 c3d5 e6d5 c6e7 d5b7 e7f5 a7a5 d3e4');
INSERT INTO evaluations VALUES(38749,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(38749,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(38749,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(38749,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(38749,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(38749,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(38749,36,'cp 76','c3d4 c5b6 b1c3 g8e7 f1c4 e8g8 h2h3 c7c6 c1g5 d6d5 e4d5 h7h6 g5e7 d8e7 d1e2 e7e2 e1e2 f8e8 e2d2 g8f8 h1e1 e8d8 a1d1 b6c7 d5c6 b8c6 d2e2 c8f5 g2g4 f5c2 d1d2 c2g6 d4d5 c6e5 f3e5 c7e5 e2f3 e5f6 c3e4 g6e4 e1e4 a8c8 f3g2 d8e8 e4e8 c8e8');
INSERT INTO evaluations VALUES(38747,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(38747,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(38747,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(38747,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(38747,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(38747,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(38747,36,'cp 83','c5b6 h2h3 g8e7 b1c3 e8g8 f1c4 c7c6 e1g1 d6d5 c4b3 h7h6 c1f4 b8a6 e4d5 e7d5 f4g3 c8e6 f1e1 f8e8 a1c1 b6c7 c3d5 c7g3 f2g3 c6d5 c1c3 a6b8 c3e3 d8d6 g3g4 b8c6 d1d3 e6d7 f3e5 c6e5 d4e5 d6e7 b3d5 a8d8 d3e4 d7c6 d5c6 b7c6');
INSERT INTO evaluations VALUES(38596,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(38596,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(38596,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(38596,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(38596,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(38596,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(38596,36,'cp 81','b1c3 g8e7 h2h3 e8g8 f1c4 c7c6 e1g1 d6d5 c4b3 h7h6 c1f4 c8e6 f1e1 f8e8 f4g3 b8a6 e4d5 e7d5 a1c1 b6c7 c3d5 c7g3 f2g3 c6d5 f3e5 a6b8 b3a4 b8d7 d1d2 e8f8 e5g6 f7g6 e1e6 d7f6 c1e1 g8h7 a4c2 d8c7 g1h2 a8e8 e6e5 b7b6 c2b1 c7d6 a2a4 e8e5 d4e5');
INSERT INTO evaluations VALUES(43027,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43027,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43027,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(43027,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(43027,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43027,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(43027,36,'cp 28','g1f3 b8c6 e1g1 e8g8 d2d3 d7d6 a2a3 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 f2f4 d6d5 c4a2 e5f4 c1f4 d8b6 f4e3 e7c5 g1h1 d4f5 f1f5 c5e3 a1b1 a8e8 e4d5 e3d4 c3e4 c6d5 a2d5 g7g6 e4f6 d4f6 f5f1 f6b2 d1f3 g8g7 c2c3 f7f5 d5b7 e8e7');
INSERT INTO evaluations VALUES(43026,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(43026,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43026,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43026,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(43026,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(43026,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(43026,36,'cp 27','b8c6 e1g1 e8g8 d2d3 d7d6 a2a4 c8e6 c3d5 e6d5 e4d5 c6b8 f1e1 c7c6 d5c6 b7c6 c4a2 a7a5 c2c3 b8d7 h2h3 a8b8 a1b1 f8e8 b2b4 d8c7 c1d2 e7f8 b4a5 c7a5 b1b8 e8b8 a2b3 d7c5 b3c2 c5e6 d3d4 e5d4 c3d4 a5d5');
INSERT INTO evaluations VALUES(43069,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(43069,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(43069,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(43069,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(43069,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(43069,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(43069,36,'cp 74','b1c3 d7d5 c4b3 e8g8 e1g1 d5e4 c3e4 h7h6 f1e1 b7b6 c2c3 b8d7 d1e2 c8b7 c1f4 e7d5 f4g3 d5f6 e4f6 d7f6 g3h4 a7a5 a2a4 b7d5 b3d5 d8d5 h4g3 a8c8 c3c4 d5b7 g3e5 f8e8 h2h3 c7c5 d4c5 b6c5 a1d1 c8d8');
INSERT INTO evaluations VALUES(45156,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(45156,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(45156,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(45156,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(45156,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(45156,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(45156,36,'cp -97','f6e4 d4c5 e4c5 b1c3 e7e6 c1f4 d6d5 c3b5 b8a6 c4e2 c5e4 e1g1 f8c5 c2c4 e8g8 a2a3 c8d7 e2d3 c5e7 c4d5 e6d5 a1c1 d7b5 d3b5 a6c5 f4e3 a7a6 b5e2 g7g6 g2g3 c5e6 f3d4 e7g5 d4e6 f7e6 d1b3 g5e3 b3e3 d8d6 e2f3');
INSERT INTO evaluations VALUES(45370,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(45370,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(45370,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(45370,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(45370,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(45370,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(45370,36,'cp -21','e7e6 d4c5 f8c5 f1d3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a6 h2h3 h7h6 c1d2 c8e6 e2g3 f8e8 b2b4 c5a7 d1e2 a8c8 a1c1 d8d7 a3a4 a7b8 c2c3 d4c3 d2c3 e8d8 f1d1 c6d4 c3d4 c8c1 d1c1 e5d4 b4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(50615,30,'cp -420',NULL);
INSERT INTO evaluations VALUES(50615,31,'cp -404',NULL);
INSERT INTO evaluations VALUES(50615,32,'cp -415',NULL);
INSERT INTO evaluations VALUES(50615,33,'cp -435',NULL);
INSERT INTO evaluations VALUES(50615,34,'cp -411',NULL);
INSERT INTO evaluations VALUES(50615,35,'cp -430',NULL);
INSERT INTO evaluations VALUES(50615,36,'cp -440','b7b5 a4b3 c5c4 b3c4 b5c4 b1c3 e7e6 g1f3 c8b7 e1g1 g8f6 d1e2 d7d5 d2d3 c4d3 c2d3 f8e7 e4e5 f6d7 d3d4 b8c6 c1f4 d7b6 a1c1 h7h6 f4e3 a8c8 b2b3 d8d7 c3a4 b6a4 b3a4 c6a5 c1c8 b7c8');
INSERT INTO evaluations VALUES(51248,30,'cp -428',NULL);
INSERT INTO evaluations VALUES(51248,31,'cp -433',NULL);
INSERT INTO evaluations VALUES(51248,32,'cp -417',NULL);
INSERT INTO evaluations VALUES(51248,33,'cp -424',NULL);
INSERT INTO evaluations VALUES(51248,34,'cp -426',NULL);
INSERT INTO evaluations VALUES(51248,35,'cp -425',NULL);
INSERT INTO evaluations VALUES(51248,36,'cp -397','a4b3 c5c4 b3c4 b5c4 b1c3 e7e6 g1f3 c8b7 d1e2 g8f6 e2c4 d7d5 c4b3 b8c6 e4e5 f6d7 c3e2 d8c8 d2d4 c6b4 e1g1 c8c2 b3e3 c2c4 b2b3 c4d3 c1a3 d3e3 f2e3 a6a5 e2f4 b7a6 a3b4 f8b4 f1c1 b4e7 c1c3 g7g5 f4h5 g5g4 f3e1 e8g8');
INSERT INTO evaluations VALUES(51242,30,'cp -425',NULL);
INSERT INTO evaluations VALUES(51242,31,'cp -432',NULL);
INSERT INTO evaluations VALUES(51242,32,'cp -437',NULL);
INSERT INTO evaluations VALUES(51242,33,'cp -437',NULL);
INSERT INTO evaluations VALUES(51242,34,'cp -413',NULL);
INSERT INTO evaluations VALUES(51242,35,'cp -440',NULL);
INSERT INTO evaluations VALUES(51242,36,'cp -447','c5c4 g1f3 c4b3 a2b3 c8b7 d2d3 e7e6 e1g1 d7d5 d1e2 g8f6 e4e5 f6d7 d3d4 f8e7 f1e1 b5b4 c2c3 b8c6 c3c4 d7f8 c1e3 f8g6 g2g3 d8d7 b1d2 h7h5 h2h4 a6a5 c4c5 f7f6 e5f6 g7f6 e3h6 e6e5 d4e5 f6e5 h6g7');
INSERT INTO evaluations VALUES(51233,30,'cp -402',NULL);
INSERT INTO evaluations VALUES(51233,31,'cp -401',NULL);
INSERT INTO evaluations VALUES(51233,32,'cp -419',NULL);
INSERT INTO evaluations VALUES(51233,33,'cp -425',NULL);
INSERT INTO evaluations VALUES(51233,34,'cp -413',NULL);
INSERT INTO evaluations VALUES(51233,35,'cp -418',NULL);
INSERT INTO evaluations VALUES(51233,36,'cp -418','b3c4 b5c4 b1c3 e7e6 g1f3 c8b7 e1g1 g8f6 f1e1 f8e7 b2b3 c4b3 a2b3 e8g8 e4e5 f6e8 d2d4 d7d6 c3a4 d6e5 f3e5 b8c6 e5f3 c6b4 e1e3 b7f3 e3f3 b4c6 f3d3 a8c8 c2c3 e8f6 a4c5 e7c5 d4c5');
INSERT INTO evaluations VALUES(51232,30,'cp -448',NULL);
INSERT INTO evaluations VALUES(51232,31,'cp -436',NULL);
INSERT INTO evaluations VALUES(51232,32,'cp -419',NULL);
INSERT INTO evaluations VALUES(51232,33,'cp -421',NULL);
INSERT INTO evaluations VALUES(51232,34,'cp -436',NULL);
INSERT INTO evaluations VALUES(51232,35,'cp -450',NULL);
INSERT INTO evaluations VALUES(51232,36,'cp -455','b5c4 b1c3 e7e6 g1f3 c8b7 e1g1 g8f6 f1e1 f8e7 a1b1 b8c6 e4e5 f6g4 b2b3 c4b3 b1b3 c6a5 b3b1 a8c8 c3e4 e8g8 h2h3 g4h6 d2d3 h6f5 c1d2 b7a8 c2c3 a5c6 d3d4 c8b8 d1e2 b8b1 e1b1 g8h8 d4d5 e6d5 e4d6 f5d6 e5d6 e7d6');
INSERT INTO evaluations VALUES(51251,30,'cp -423',NULL);
INSERT INTO evaluations VALUES(51251,31,'cp -436',NULL);
INSERT INTO evaluations VALUES(51251,32,'cp -436',NULL);
INSERT INTO evaluations VALUES(51251,33,'cp -423',NULL);
INSERT INTO evaluations VALUES(51251,34,'cp -429',NULL);
INSERT INTO evaluations VALUES(51251,35,'cp -442',NULL);
INSERT INTO evaluations VALUES(51251,36,'cp -425','g1f3 d7d5 e4d5 d8d5 b2b3 g8f6 b1a3 c4b3 a2b3 d5e4 d1e2 e4e2 e1e2 b8c6 d2d4 e7e6 h1d1 f8e7 f3e5 c8b7 e5c6 b7c6 f2f3 e8g8 c1d2 a6a5 a3c4 a5a4 c4b6 a8d8 b3a4 d8d4 a4a5 f6d5 c2c4 d5b6 a5b6 d4c4 a1c1 c6b5 c1c4 b5c4 e2f2 c4d5 b6b7');
INSERT INTO evaluations VALUES(14152,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(14152,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(14152,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(14152,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(14152,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(14152,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(14152,36,'cp 53','c1e3 c8f5 b1c3 e7e6 g1f3 f8e7 f1e2 e8g8 e1g1 f7f6 e5f6 e7f6 d1d2 d8e7 g1h1 a8d8 a1d1 h7h6 d2c1 g8h8 e3g1 f5g4 c4c5 b6d5 f3h4 f6h4 c3d5 e6d5 e2g4 e7g5 c1g5 h4g5 f1f8 d8f8 g2g3 g5f6 h2h4 g7g6 g4e6 f8d8 g3g4 b7b6 c5b6 c7b6 h4h5');
INSERT INTO evaluations VALUES(14150,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(14150,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(14150,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(14150,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14150,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(14150,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(14150,36,'cp 54','c8f5 b1c3 e7e6 g1f3 f8e7 f1e2 e8g8 e1g1 f7f6 e5f6 e7f6 d1d2 d8e7 g1h1 g8h8 a1d1 a8d8 d2c1 d8d7 e3g1 f8d8 c4c5 b6d5 e2c4 e7f7 a2a3 h7h6 h2h3 f7g6 d1e1 c6a5 c4a2 f5d3 c3e2 a5c6 g1h2 d3e4 b2b4 e4f3 f1f3 c6d4 e2d4');
INSERT INTO evaluations VALUES(24759,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(24759,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(24759,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(24759,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(24759,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(24759,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(24759,36,'cp 70','b1c3 e7e6 g1f3 f8e7 f1e2 f7f6 e1g1 e8g8 e5f6 e7f6 d1d2 d8e7 a1d1 a8d8 g1h1 h7h6 d2c1 g8h8 e3g1 d8d7 h2h3 f8d8 c4c5 b6d5 a2a3 e7f7 d1e1 f7g6 e2b5 a7a6 b5c6 b7c6 g1h2 d7f7 f1f2 d8f8 h2e5 h8g8 h1h2 f5d3 e5f6 g6f6 c3e4 f6f4 c1f4 f7f4');
INSERT INTO evaluations VALUES(24758,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(24758,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(24758,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(24758,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(24758,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(24758,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(24758,36,'cp 74','e7e6 g1f3 f8e7 f1e2 e8g8 e1g1 f7f6 e5f6 e7f6 d1d2 d8e7 g1h1 a8d8 a1d1 f5g6 d2c1 h7h6 h2h3 g8h8 b2b3 g6h7 e3g1 d8d7 f1e1 e7d8 c1e3 c6b4 e1f1 b4c2 e3c1 d8e7 d1d2 c7c5 d4c5 d7d2 c1d2 f8d8 d2f4 f6c3 c5b6 e6e5 f4g3 a7b6');
INSERT INTO evaluations VALUES(25076,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(25076,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(25076,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(25076,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(25076,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(25076,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(25076,36,'cp 83','f1e2 f8e7 g1f3 f7f6 e1g1 e8g8 e5f6 e7f6 d1d2 d8e7 g1h1 a8d8 a1d1 h7h6 h2h3 c6b4 a2a3 b4c2 e3g1 c7c5 d4d5 f6c3 d5d6 d8d6 d2d6 e7d6 d1d6 c3b2 g2g4 f5h7 d6e6 b2a3 g1h2 a3b2 h2d6 f8c8 h3h4 c2a3 f3e5 h7e4 h1h2 b2e5 d6e5 a3c4 e5a1');
INSERT INTO evaluations VALUES(20343,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(20343,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(20343,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(20343,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(20343,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(20343,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(20343,36,'cp -15','f7f6 g1f3 f6e5 f3e5 c6e5 d4e5 d8c7 f1b5 c8d7 b5d7 c7d7 c1e3 b7b6 a3a4 a8c8 e1g1 g8e7 b1d2 e7f5 d1h5 g7g6 h5e2 f8e7 g2g4 f5e3 f2e3 c8d8 a4a5 b6b5 e2f3 d7c7 f3f7 e8d7 d2b3 d8f8 f7g7 h8g8 g7h7 c7e5 f1f8 g8f8 b3c5 d7d6 c5b7 d6c6 h7e7 e5e3 g1h1 e3f3 h1g1 f3g4 g1h1 g4f3 h1g1 f3e3 g1h1');
INSERT INTO evaluations VALUES(31712,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31712,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31712,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31712,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31712,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31712,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31712,36,'cp 0','h2h3 g4f3 d1f3 b8c6 c1e3 g8f6 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(31710,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31710,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31710,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31710,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31710,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31710,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31710,36,'cp 0','g4f3 d1f3 b8c6 c1e3 g8f6 f1d3 f8d6 g2g4 d8a5 a2a3 a7a6 g4g5 f6d7 h3h4 d6b4 a3b4 a5a1 e1d2 a1b2 h1b1 b2a3 b1b3 a3a1 b3b1 a1a3');
INSERT INTO evaluations VALUES(42475,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(42475,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(42475,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(42475,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42475,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(42475,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42475,36,'cp -13','c7c5 g1f3 b8c6 e2e3 c5d4 f3d4 b4c5 d4f3 b7b6 f1e2 e8g8 e1g1 c8b7 b3d1 a8c8 a2a3 c5e7 b2b4 a7a5 b4b5 c6b8 c1b2 d7d6 c3a4 b8d7 f3d2 d7c5 a4c5 c8c5 e2f3 b7f3 d1f3 c5c7 a1d1 d8a8 b2f6 a8f3 g2f3 g7f6 d2e4 c7c4');
INSERT INTO evaluations VALUES(39912,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(39912,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(39912,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(39912,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(39912,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(39912,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(39912,36,'cp -12','g1f3 b8c6 e2e3 c5d4 f3d4 b4c5 d4f3 b7b6 f1e2 c8b7 e1g1 e8g8 b3d1 d7d5 c4d5 e6d5 a2a3 d5d4 e3d4 c6d4 f3d4 d8d4 e2f3 d4d1 f1d1 b7f3 g2f3 f8d8 c1e3 c5e7 b2b4 g8f8 g1g2 a8c8 c3b5 a7a6 b5d4 f6e8 d4f5 d8d1 a1d1');
INSERT INTO evaluations VALUES(43407,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(43407,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(43407,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43407,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43407,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43407,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43407,36,'cp 11','e7e5 d4e2 f8e7 f1g2 b7b5 c3d5 b8d7 e2c3 a8b8 e1g1 e8g8 c1e3 h7h6 b2b3 b5b4 d5e7 d8e7 c3d5 f6d5 d1d5 d7c5 a1d1 f8d8 e3c5 b8b5 c5d6 b5d5 d6e7 d5d1 e7d8 d1d8 f1c1 d8d2 g2f1 a6a5 f1c4 c8g4 c2c3 d2a2 c3b4 a5b4 c4d5');
INSERT INTO evaluations VALUES(43451,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43451,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43451,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(43451,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43451,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43451,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(43451,36,'cp 0','d4e2 f8e7 a2a4 b8c6 f1g2 c6b4 b2b3 c8e6 e1g1 a8c8 c1a3 d8b6 d1d2 h7h6 a3b2 b6c6 b2a3 c6b6');
INSERT INTO evaluations VALUES(52717,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(52717,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(52717,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(52717,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(52717,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(52717,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(52717,36,'cp -28','c6d5 d2d4 c8f5 f1g2 b8c6 g1f3 e7e6 e1g1 h7h6 b1c3 g8f6 c1f4 f8d6 f3e5 a8c8 a1c1 e8g8 d1d2 d6e7 h2h3 d8b6 c3a4 b6b5 e5c6 c8c6 c1c6 b7c6 a4c3 b5a6 f1c1 f8c8 a2a3 c6c5 d4c5 c8c5 f4e3 c5c8 e3d4 f6d7 e2e4 d5e4 c3e4');
INSERT INTO evaluations VALUES(53431,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(53431,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(53431,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(53431,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(53431,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(53431,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(53431,36,'cp -23','g1f3 b8c6 d2d4 c8f5 f1g2 e7e6 e1g1 g8f6 b1c3 h7h6 c1f4 f8d6 f3e5 a8c8 a1c1 e8g8 d1d2 f8e8 f1d1 d6e7 h2h4 d8b6 g2f3 e8d8 e2e3 f5h7 f3e2 f6e4 c3e4 h7e4 e5c6 c8c6 c1c6 b6c6 d2a5');
INSERT INTO evaluations VALUES(53428,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(53428,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(53428,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(53428,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(53428,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(53428,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(53428,36,'cp -27','b8c6 d2d4 c8f5 b1c3 e7e6 f1g2 h7h6 e1g1 g8f6 c1f4 f8d6 f3e5 a8c8 a1c1 e8g8 d1d2 d6e7 h2h3 f8e8 c3a4 c6e5 f4e5 b7b6 c1c8 d8c8 a4c3 c8b7 f1c1 e8c8 e5f6 e7f6 e2e3 f6e7 a2a3 e7d6 g2f1 a7a6 c3a4 c8c1 d2c1 g7g5 a4c3');
INSERT INTO evaluations VALUES(5771,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5771,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(5771,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(5771,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(5771,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(5771,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(5771,36,'cp 32','a2a4 a7a6 e1g1 g8f6 d2d3 h7h6 b1d2 e8g8 h2h3 c8e6 c4e6 f7e6 d2c4 b6a7 c1e3 a7e3 f2e3 c6e7 a4a5 e7g6 c4d2 f8f7 c3c4 d8d7 c4c5 a8d8 d2c4 f6h5 g1h2 d8f8 a1c1 d7e7 c4d2 d6d5 c5c6 b7c6 c1c6 d5d4 e3d4 e5d4 d2c4');
INSERT INTO evaluations VALUES(5775,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(5775,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(5775,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(5775,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(5775,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(5775,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(5775,36,'cp 23','a7a6 d2d3 g8f6 e1g1 e8g8 b1d2 h7h6 c4b3 c8e6 f1e1 e6b3 d1b3 d8d7 d2f1 d7e6 b3c2 f6g4 c1e3 b6e3 f1e3 g4e3 e1e3 f7f5 a1e1 f5f4 e3e2 g7g5 d3d4 g5g4 d4d5 e6f6 d5c6 b7c6 c2d3 g4f3 d3f3 g8h8 g1h1 f6e6 e2d2');
INSERT INTO evaluations VALUES(4815,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4815,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(4815,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(4815,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(4815,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(4815,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(4815,36,'cp 25','d2d3 g8f6 e1g1 e8g8 h2h3 h7h6 b1d2 c8e6 c4e6 f7e6 d2c4 b6a7 c1e3 a7e3 f2e3 c6e7 c4d2 e7g6 d1b3 d8e7 c3c4 f6d7 a4a5 a8b8 a1c1 c7c5 b4b5 d7f6 f3h2 b8a8 c1a1 f8f7 f1f2 f7f8 d2f1 f8f7 b5a6 b7a6 b3b6 f7f8 f1g3 g6h4 a1f1 h4g6');
INSERT INTO evaluations VALUES(7459,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7459,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7459,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7459,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(7459,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7459,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7459,36,'cp 0','d7d6 e1g1 h7h6 a2a4 c8d7 d3d4 c5b6 c1e3 f6g4 g1h1 e8g8 h2h3 g4e3 f2e3 c6e7 d4e5 b6e3 e5d6 c7d6 d1d6 b5a4 b3a4 d7a4 d6d8 f8d8 a1a4 a8b8 b2b4 f7f5 e4e5 f5f4 h1h2 e7f5 a4a6 f5g3 f1e1 e3f2 e1c1 f2e3');
INSERT INTO evaluations VALUES(7421,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(7421,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(7421,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(7421,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(7421,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(7421,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(7421,36,'cp 10','c1g5 h7h6 g5h4 c8e6 b1d2 e6b3 d2b3 c5b6 a2a4 g7g5 h4g3 c6e7 b3d2 f6h5 e1g1 e7g6 d3d4 g5g4 f3e1 h5g3 h2g3 e5d4 a4b5 a6b5 a1a8 d8a8 d1g4 d4c3 b2c3 a8c6 e4e5 d6e5 e1f3 e8f8 d2e4 f8e7 f1b1 c6e6 g4h5 e6c6 f3d2 c6d5 h5g4 d5e6 g4e6 e7e6 b1b5');
INSERT INTO evaluations VALUES(13591,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(13591,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(13591,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13591,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13591,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13591,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13591,36,'cp 0','b7b6 e1g1 c8b7 f1e1 f8e7 d4c5 b6c5 e3e4 d8c7 e4d5 e6d5 d2f1 e8g8 c1g5 h7h6 g5h4 a8d8 h4g3 e7d6 g3h4');
INSERT INTO evaluations VALUES(15309,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15309,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15309,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15309,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15309,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15309,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(15309,36,'cp 6','c8e6 f3g5 d5c3 g5e6 c3d1 e6d8 a8d8 e1d1 c6a5 b2b3 f8c5 h1e1 f7f6 d1e2 a5c4 b3c4 e8d7 a1b1 d7c6 a2a4 b7b6 c1e3 c5d4 h2h4 d4c3 e1h1 h8e8 b1b3 c3d4 a4a5 b6a5 b3a3 d4b6 c4c5 b6c5 e3c5 c6c5 a3a5 c5b6 h1a1 e5e4 e2d2 e4d3 c2d3 b6c6 a5a7');
INSERT INTO evaluations VALUES(25105,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(25105,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(25105,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(25105,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(25105,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(25105,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(25105,36,'cp 4','f3g5 d5c3 g5e6 c3d1 e6d8 a8d8 e1d1 c6a5 b2b3 f8c5 d1e2 a5c4 b3c4 c5d4 a1b1 e8d7 c1e3 d7c8 h1e1 d4c3 e1f1 d8d6 b1b3 c3d4 a2a4 d6a6 f1b1 h8e8 e2f3 b7b6 b3a3 h7h5 f3e4 c7c6 a4a5 a6a5 a3a5 b6a5 e3d4 e5d4 e4d4 e8e2 b1a1 e2f2 a1a5 f2g2 a5a7 g2h2');
INSERT INTO evaluations VALUES(18001,30,'cp 343',NULL);
INSERT INTO evaluations VALUES(18001,31,'cp 346',NULL);
INSERT INTO evaluations VALUES(18001,32,'cp 327',NULL);
INSERT INTO evaluations VALUES(18001,33,'cp 346',NULL);
INSERT INTO evaluations VALUES(18001,34,'cp 348',NULL);
INSERT INTO evaluations VALUES(18001,35,'cp 339',NULL);
INSERT INTO evaluations VALUES(18001,36,'cp 329','d1d8 e8d8 f3e5 d8e8 b1c3 g8e7 h2h4 f7f6 e5d3 e7g6 h4h5 g6e5 d3e5 f6e5 b2b3 c8e6 e1e2 e8d7 f2f3 c6c5 h5h6 g7g6 h1d1 f8d6 c1g5 h8f8 c3a4 d7c6 c2c4 a6a5 a1c1 f8f7 a4c3 a8f8 d1d3 f8a8 g5d2 f7f8 d2e3 f8h8 c1d1 h8b8 g2g4 b8f8');
INSERT INTO evaluations VALUES(2257,30,'cp 321',NULL);
INSERT INTO evaluations VALUES(2257,31,'cp 327',NULL);
INSERT INTO evaluations VALUES(2257,32,'cp 324',NULL);
INSERT INTO evaluations VALUES(2257,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(2257,34,'cp 317',NULL);
INSERT INTO evaluations VALUES(2257,35,'cp 332',NULL);
INSERT INTO evaluations VALUES(2257,36,'cp 333','e8d8 f3e5 d8e8 b1c3 g8e7 h2h4 f7f6 e5d3 e7g6 h4h5 g6e5 d3e5 f6e5 b2b3 c6c5 c3d5 f8d6 f2f3 c8e6 e1e2 e8d7 d5e3 g7g6 c1b2 g6g5 g2g4 d7c6 b2c3 h8f8 h1d1 a6a5 a2a4 h7h6 e3f5 e6f5 g4f5 a8b8 c3d2 f8g8 a1b1 g5g4 d2h6 g4f3 e2f3');
INSERT INTO evaluations VALUES(2303,30,'cp 340',NULL);
INSERT INTO evaluations VALUES(2303,31,'cp 328',NULL);
INSERT INTO evaluations VALUES(2303,32,'cp 337',NULL);
INSERT INTO evaluations VALUES(2303,33,'cp 337',NULL);
INSERT INTO evaluations VALUES(2303,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(2303,35,'cp 333',NULL);
INSERT INTO evaluations VALUES(2303,36,'cp 332','f3e5 d8e8 b1c3 g8e7 h2h4 f7f6 e5d3 e7g6 h4h5 g6e5 d3e5 f6e5 b2b3 c6c5 h5h6 g7g6 c3d5 f8d6 c1g5 c8e6 a1d1 a6a5 d5f6 e8f7 d1d3 e6c8 d3f3 f7e6 f6d5 a5a4 f3f6 e6d7 e1e2 a4b3 a2b3 a8a2 e2d3 c8a6 c2c4 a6b7 h1d1 b7d5 c4d5 h8f8 f6f8 d6f8 g5e3 a2b2');
INSERT INTO evaluations VALUES(2302,30,'cp 330',NULL);
INSERT INTO evaluations VALUES(2302,31,'cp 325',NULL);
INSERT INTO evaluations VALUES(2302,32,'cp 348',NULL);
INSERT INTO evaluations VALUES(2302,33,'cp 343',NULL);
INSERT INTO evaluations VALUES(2302,34,'cp 315',NULL);
INSERT INTO evaluations VALUES(2302,35,'cp 321',NULL);
INSERT INTO evaluations VALUES(2302,36,'cp 332','d8e8 b1c3 g8e7 h2h4 f7f6 e5d3 e7g6 h4h5 g6e5 d3e5 f6e5 b2b3 c6c5 c3d5 f8d6 f2f3 c8e6 e1e2 e8d7 d5e3 g7g5 h1d1 a6a5 a2a4 h8b8 g2g4 d7c6 c1d2 b8e8 d1h1 e8b8 a1b1 b8b6 d2c3 h7h6 e3f5 c5c4 f5h6 c4b3 c2b3 b6b3 b1b3 e6b3 h6f5');
INSERT INTO evaluations VALUES(19368,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(19368,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(19368,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19368,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(19368,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(19368,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(19368,36,'cp 0','h5c5 e7e5 c5e3 f7f5 c2c3 f5e4 d2d3 e4d3 g1f3 d8e7 f1d3 g8f6 d3b5 e7c5 e3c5 f8c5 e1g1 d7d6 b1d2 e8g8 d2c4 e5e4 b2b4 c5b4 c3b4 e4f3 c4d6 c8e6 f1e1 e6d5 b5c6 b7c6 d6f5 f3g2 f5e7 g8f7 e7f5');
INSERT INTO evaluations VALUES(19362,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19362,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19362,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(19362,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(19362,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(19362,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(19362,36,'cp -12','e7e5 c5e3 f7f5 c2c3 f5e4 d2d3 e4d3 g1f3 d8e7 f1d3 d7d5 c3c4 g8f6 c4d5 f6d5 e3g5 e7g5 c1g5 c6b4 d3c4 e5e4 f3d4 b4d3 e1d2 h7h6 c4d5 h6g5 b1c3 d3f2 h1e1 f8c5 c3e4 c5d4 e4f2 e8d8 f2e4 d4b2 a1b1 b2e5 e4c5 e5f4 d2c3 a8b8 c5b7 c8b7 b1b7');
INSERT INTO evaluations VALUES(23382,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(23382,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(23382,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(23382,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(23382,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(23382,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(23382,36,'cp 188','e2e4 e7e6 d4d5 c6e5 f3e5 f6e5 d1h5 g7g6 h5e5 g8f6 c1g5 f8g7 b1c3 d7d6 e5g3 e8g8 e1c1 d8e8 f1c4 f6h5 g3e3 e6e5 f2f3 e8f7 g5h6 g7h6 e3h6 a7a6 c4e2 f7f4 h6f4 h5f4 h1g1 c8d7 c1d2 f8f7 h2h4 b7b5 a2a3 a8f8 e2d3 h7h5 d2e3 g8h7 c3e2');
INSERT INTO evaluations VALUES(23383,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(23383,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(23383,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(23383,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(23383,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(23383,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(23383,36,'cp 200','e7e6 d4d5 c6e5 f3e5 f6e5 d1h5 g7g6 h5e5 g8f6 c1g5 f8g7 b1c3 d7d6 e5g3 e8g8 e1c1 d8e8 f1c4 f6h5 g3e3 e6e5 f2f3 e8f7 e3d2 a7a6 c4f1 h5f4 c1b1 c8d7 g5h4 b7b5 c3e2 f8b8 e2f4 e5f4 h4f2 a6a5 f2d4 g7d4 d2d4 f7g7 b1c1 a5a4 h2h4 a4a3 d4g7 g8g7');
INSERT INTO evaluations VALUES(23192,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(23192,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(23192,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(23192,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(23192,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(23192,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(23192,36,'cp 207','d4d5 c6e5 f3e5 f6e5 d1h5 g7g6 h5e5 g8f6 c1g5 f8g7 b1c3 d7d6 e5g3 e8g8 f1c4 d8e8 e1c1 f6h5 g3d3 e6e5 f2f3 a7a6 d3d2 h5f4 c4f1 c8d7 g5h4 b7b5 c3e2 a6a5 h4f2 b5b4 f2e3 d7b5 e2f4 e5f4 e3d4 b5f1 h1f1 e8e7 f1e1 g7e5 c1b1 e5d4 d2d4 f8e8 d4d3 e7g5 e1e2 g5f6 d3c4 e8e7');
INSERT INTO evaluations VALUES(31294,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(31294,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(31294,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(31294,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(31294,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(31294,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(31294,36,'cp -43','f1b5 b8d7 g1f3 e7e6 d2d4 h7h5 e1g1 c5c4 c2c3 g8h6 b1d2 a7a6 b5a4 b7b5 a4c2 f8e7 f1e1 a6a5 d2f1 b5b4 f1e3 f5c2 d1c2 g7g6 h2h3 h6f5 e3f5 g6f5 c2a4 e8f8 g1h1 d8e8 a4c2 d7b6 a2a3 b4b3 c2f2 h5h4 c1d2');
INSERT INTO evaluations VALUES(32262,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(32262,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(32262,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(32262,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(32262,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(32262,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(32262,36,'cp 129','g8f6 d1e2 f8b4 c1g5 h7h6 g5h4 g6e4 h4f6 e4f3 e2f3 g7f6 f3b7 b4c3 b2c3 b8d7 a1d1 d7b6 c4b3 a7a5 e1g1 a5a4 b3c2 e8e7 b7f3 d8d5 c2e4 d5a2 e4a8 h8a8 c3c4 a8d8 h2h3 a2b3 d1d3 b3b4 f1c1 b6c4 f3f4 e6e5 f4e4 a4a3 d4e5 d8d3 e5f6 e7f6 e4d3 a3a2 d3g3');
INSERT INTO evaluations VALUES(34662,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(34662,31,'cp 284',NULL);
INSERT INTO evaluations VALUES(34662,32,'cp 293',NULL);
INSERT INTO evaluations VALUES(34662,33,'cp 309',NULL);
INSERT INTO evaluations VALUES(34662,34,'cp 272',NULL);
INSERT INTO evaluations VALUES(34662,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(34662,36,'cp 293','h1g1 g2h3 e2c4 e7e6 d2d4 c7c6 d1e2 g8e7 g1g3 h3h5 c3e4 e7d5 g3g5 h5h3 c1d2 b8d7 g5g3 h3h5 e1c1 h7h6 c4d3 d5f6 e4f6 g7f6 c2c4 d7b6 e2e4 c8d7 e4f4 f8e7 f4c7 a8c8 c7b7 e7d6 d1g1 c6c5 b7a7 d6g3 g1g3 d7c6 d4c5');
INSERT INTO evaluations VALUES(34661,30,'cp 321',NULL);
INSERT INTO evaluations VALUES(34661,31,'cp 310',NULL);
INSERT INTO evaluations VALUES(34661,32,'cp 298',NULL);
INSERT INTO evaluations VALUES(34661,33,'cp 296',NULL);
INSERT INTO evaluations VALUES(34661,34,'cp 323',NULL);
INSERT INTO evaluations VALUES(34661,35,'cp 307',NULL);
INSERT INTO evaluations VALUES(34661,36,'cp 312','g2h3 e2c4 e7e6 d2d4 b8c6 g1g3 h3f5 c3b5 f8d6 c4d3 f5h5 g3g5 h5h6 g5g4 g7g5 c1g5 h6f8 d1e2 g8e7 f3e5 c8d7 g5f6 d6e5 d4e5 e8c8 e1c1 a7a6 b5d4 c8b8 c1b1 f8e8 d4c6 d7c6 f6h8 e8h8 g4f4 h8g7 e2h5 e7g6 d3g6 d8d1 h5d1 g7e5');
INSERT INTO evaluations VALUES(34660,30,'cp 293',NULL);
INSERT INTO evaluations VALUES(34660,31,'cp 324',NULL);
INSERT INTO evaluations VALUES(34660,32,'cp 317',NULL);
INSERT INTO evaluations VALUES(34660,33,'cp 310',NULL);
INSERT INTO evaluations VALUES(34660,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(34660,35,'cp 337',NULL);
INSERT INTO evaluations VALUES(34660,36,'cp 315','e2c4 e7e6 d2d4 c8d7 g1g3 h3f5 d4d5 g8f6 d1e2 b8a6 d5e6 f7e6 g3g5 f5h3 c1f4 f8d6 f4d6 c7d6 c3b5 e8g8 b5d6 f6d5 f3e5 h3h2 e5d7 h2d6 d7f8 a8f8 e2e5 d6e5 g5e5 a6c7 c4d5 e6d5 e1f1 f8f7 a1e1 g8f8 e1e3 g7g6 e3e1 b7b6 e5e3 h7h5 e3c3 f8g7 c3c6 h5h4 e1e3 d5d4');
INSERT INTO evaluations VALUES(47248,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(47248,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47248,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47248,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47248,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47248,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47248,36,'cp 2','g1f3 c8f5 d2d4 e7e6 f1d3 f5d3 d1d3 c7c5 e1g1 c5d4 f3d4 b8d7 c2c4 d5c4 d3c4 a8c8 c4e2 d8a5 b1d2 d7b6 c1b2 f8e7 a2a3 e8g8 h2h3 a5h5 e2h5 f6h5 f1c1 h5f6 g1f1 f6d5 d2c4 f8d8 c4b6 d5b6');
INSERT INTO evaluations VALUES(47250,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(47250,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(47250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47250,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(47250,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(47250,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(47250,36,'cp 0','c8f5 d2d4 e7e6 f1d3 f5d3 d1d3 c7c5 c1b2 b8d7 e1g1 f8e7 c2c4 e8g8 c4d5 f6d5 f1d1 c5d4 f3d4 d7c5 d3e2 d8b6 b1d2 f8d8 a2a4 d8d7 d4b5 b6a6 b2d4 a8d8 d4c5 e7c5 d2c4 d5f6 h2h3 c5e7 a1c1');
INSERT INTO evaluations VALUES(53570,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(53570,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(53570,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(53570,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(53570,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(53570,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(53570,36,'cp 29','e2e3 g8f6 b1c3 f8d6 f4d6 d8d6 g1f3 e8g8 a1c1 f8d8 d1b3 b8d7 c4d5 e6d5 f1d3 d6e7 b3c2 d8e8 e1g1 f6e4 f1e1 h7h6 c3e2 a7a5 a2a3 e4d6 e2c3 d7b6 b2b3 c8g4 f3e5 g4h5 g1h1 b6d7 f2f4 d7f6 h2h3');
INSERT INTO evaluations VALUES(53568,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53568,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(53568,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(53568,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(53568,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(53568,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(53568,36,'cp 26','g8f6 b1c3 f8d6 f4d6 d8d6 c4d5 c6d5 f1d3 e8g8 f2f4 b7b6 d1e2 c8b7 g1f3 f6e4 f3g5 e4f6 e1g1 b8c6 a2a3 a8c8 g5f3 c6e7 f3e5 d6d8 e2f3 e7f5 f1c1 f5d6 c3b5 d6b5 d3b5 c8c7 c1c7 d8c7 g1h1');
INSERT INTO evaluations VALUES(53582,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(53582,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(53582,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(53582,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(53582,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(53582,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(53582,36,'cp -35','d5c4 a3a4 c6c5 g1f3 b8c6 e2e3 c5d4 e3d4 c6a5 f3e5 f8b4 f1c4 a5c4 e5c4 g8e7 e1g1 e8g8 d1b3 a7a5 c1g5 h7h6 g5e7 d8e7 f1d1 f8d8 a1c1 c8d7 c4b6 a8a6 b6d7 d8d7 c3a2 b4d6 g2g3 e7d8 d4d5 d6e5 d5e6 a6e6 a2c3 e5c3 b2c3');
INSERT INTO evaluations VALUES(53660,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(53660,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(53660,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(53660,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(53660,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(53660,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(53660,36,'cp -18','a3a4 c6c5 e2e3 c5d4 e3d4 b8c6 g1f3 c6a5 f3e5 c8d7 f1c4 a5c4 e5c4 g8f6 c4e5 d7c6 e1g1 f8d6 c1g5 d6e7 a4a5 e8g8 g5f4 a8c8 e5c6 c8c6 d1b3 d8c8 d4d5 f6d5 c3d5 e6d5 f1d1 f8d8 d1d5 d8d5 b3d5 h7h6 f4e3');
INSERT INTO evaluations VALUES(1947,30,'cp 255',NULL);
INSERT INTO evaluations VALUES(1947,31,'cp 239',NULL);
INSERT INTO evaluations VALUES(1947,32,'cp 247',NULL);
INSERT INTO evaluations VALUES(1947,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(1947,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(1947,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(1947,36,'cp 262','f3d4 e5d4 f1d3 f8c5 e1g1 d7d6 b1d2 f6e7 d2b3 c5b6 b3d4 g8f6 d3b5 e8f8 f2f3 h7h5 c2c3 h5h4 c1f4 f8g8 d1d2 c8d7 b5d7 e7d7 g1h1 b6d4 d2d4 h4h3 g2g4 a8e8 c3c4 a7a5 f4g3 c7c6 d4d2 a5a4 f1e1 c6d5 c4d5');
INSERT INTO evaluations VALUES(1946,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(1946,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(1946,32,'cp 260',NULL);
INSERT INTO evaluations VALUES(1946,33,'cp 262',NULL);
INSERT INTO evaluations VALUES(1946,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(1946,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(1946,36,'cp 251','e5d4 f1d3 f8c5 e1g1 d7d6 b1d2 f6e7 d2b3 c5b6 b3d4 g8f6 d3b5 e8f8 f2f3 h7h5 c2c3 h5h4 c1f4 a7a6 b5a4 c8d7 a4d7 e7d7 g1h1 f8g8 d1d2 b6d4 d2d4 a8e8 a1e1 h4h3 g2g4 f6h7 b2b3 h7f8 c3c4 f8g6 f4g3 g6e5 g3e5 e8e5 d4d3 e5e8 a2a4 b7b5 a4b5 a6b5 c4b5');
INSERT INTO evaluations VALUES(1945,30,'cp 248',NULL);
INSERT INTO evaluations VALUES(1945,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(1945,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(1945,33,'cp 247',NULL);
INSERT INTO evaluations VALUES(1945,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(1945,35,'cp 275',NULL);
INSERT INTO evaluations VALUES(1945,36,'cp 269','f1d3 f8c5 e1g1 d7d6 b1d2 f6e7 d2b3 c5b6 b3d4 g8f6 c2c3 e8g8 c1f4 c8d7 d1d2 h7h6 f1e1 f6g4 a2a4 a7a5 d3f1 e7f6 f2f3 g4e5 f4g3 h6h5 g3f2 c7c6 d5c6 b7c6 g1h1 h5h4 h2h3 e5g6');
INSERT INTO evaluations VALUES(8954,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(8954,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(8954,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(8954,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(8954,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(8954,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(8954,36,'cp 95','c2c4 f6e4 a2a3 b4d6 f4g3 e8g8 d2e4 d5e4 f3d2 f7f5 c4c5 d6e7 b2b4 e7f6 b4b5 c6e7 f1c4 e7d5 a1c1 a7a6 b5a6 b7a6 d1b3 g8h8 e1g1 a6a5 a3a4 c8a6 c1b1 a6c4 d2c4 a8b8 b3b8 d8b8 b1b8 f8b8 c4a5 b8b2 a5c6 b2d2');
INSERT INTO evaluations VALUES(8967,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(8967,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(8967,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(8967,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(8967,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(8967,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(8967,36,'cp 72','f6e4 a2a3 b4d6 f4g3 e8g8 f1d3 f7f5 b2b4 a7a6 d1b3 e4d2 f3d2 d5c4 d2c4 f5f4 e3f4 d6f4 a1d1 d8f6 e1g1 f4g3 h2g3 c6d4 b3b2 e6e5 d1e1 f8e8 f2f4 c8f5 d3f5 f6f5 c4e5 a8d8 e1d1 f5c2 b2c2 d4c2 d1c1');
INSERT INTO evaluations VALUES(10398,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(10398,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(10398,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(10398,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(10398,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(10398,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(10398,36,'cp -43','f6e4 e1g1 b4c3 b2c3 d7d5 c4b5 e8g8 b5c6 b7c6 f3e5 f7f6 e5d3 e4c3 d1d2 c3b5 d3c5 c8f5 a2a4 b5d6 d2c3 f8e8 c1f4 e8e2 a1c1 d8e7 f1e1 a8e8 e1e2 e7e2 c5d3 f5d3 c2d3 d6f5 h2h3 e8e6 f4c7 g7g5 c3a5 g5g4 h3g4 f5d4 a5a7 e2g4 g1h1 g4h4 c7h2 h4f2');
INSERT INTO evaluations VALUES(16760,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(16760,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(16760,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(16760,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(16760,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(16760,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(16760,36,'cp 67','f1e1 f7f5 f3d4 c6d4 d1d4 c7c6 f2f3 c6b5 f3e4 d8b6 e4f5 e8f7 c1e3 b6d4 e3d4 b7b6 b1c3 b5b4 c3e4 h7h6 e4d2 d7d5 g2g4 h6h5 d2f3 h5g4 f3e5 f7g8 e5g6 c8f5 g6h8 g8h8 e1f1 f5e4 f1f7 a8c8 a1c1 a7a5 d4b6 h8g8 f7d7 a5a4 b6d4 e4c2 d7d5 a4a3 b2a3 b4a3');
INSERT INTO evaluations VALUES(16761,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(16761,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(16761,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(16761,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(16761,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(16761,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(16761,36,'cp 62','f7f5 f3d4 c6d4 d1d4 c7c6 b1c3 d8f6 d4f6 g7f6 b5d3 e8f7 c3e4 f5e4 e1e4 d7d5 e4h4 f8d6 a2a4 h7h5 a4a5 c8d7 c1e3 a7a6 c2c3 a8e8 h2h3 d6g3 h4b4 d7c8 e3b6 g3d6 b4h4 e8e5 b6d4 e5g5 d4e3 c8f5 e3g5 f6g5 h4h5 h8h5 d3f5 h5h4');
INSERT INTO evaluations VALUES(17125,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(17125,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17125,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(17125,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(17125,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(17125,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(17125,36,'cp 55','f3d4 c6d4 d1d4 c7c6 b1c3 d8f6 d4f6 g7f6 b5d3 e8f7 c3e4 f5e4 e1e4 d7d5 e4h4 h7h5 c1e3 f8e7 h2h3 c8d7 a1d1 a7a5 c2c4 d7e6 c4d5 e6d5 a2a3 a8e8 g2g3 d5e6 d1e1 e8d8 d3e2 d8d5 e3f4 d5f5 a3a4 e6d5 g1f1 e7b4 e1d1');
INSERT INTO evaluations VALUES(19695,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(19695,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(19695,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(19695,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(19695,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(19695,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(19695,36,'cp 151','d4d5 c6e5 f3e5 d6e5 f1e2 g8f6 b1c3 f8g7 d1a4 c8d7 a4b4 b7b6 a2a4 e8g8 e1g1 a7a5 b4b3 a8b8 c1e3 h7h5 a1c1 g8h7 h2h3 f6e8 c3b5 e8d6 b5d6 e7d6 e2b5 d7b5 a4b5 g7h6 e3h6 h7h6 b3e3 d8g5 c1c6 g5e3 f2e3 h6g7 c6d6');
INSERT INTO evaluations VALUES(20691,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(20691,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(20691,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20691,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(20691,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(20691,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(20691,36,'cp 28','d2d4 e6e5 d4c5 d6c5 d1d8 e8d8 c1e3 f7f6 e1c1 c6d4 f3d2 g8e7 f2f4 e7c6 f4f5 h7h5 h2h3 c8d7 f1d3 c6b4 d3b1 h5h4 e3f2 b4c6 b1d3 d7e8 d3e2 f8d6 d2f3 e8f7 b2b3 a7a5 c1b2 d8e7 f3h4 a5a4 h4g6 f7g6 f5g6 a4b3 a2b3 c6b4');
INSERT INTO evaluations VALUES(20693,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(20693,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20693,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(20693,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(20693,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(20693,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20693,36,'cp 39','e6e5 d4e5 d6e5 d1d8 e8d8 c1e3 f7f6 e1c1 c6d4 f3d2 g8e7 f2f4 e7c6 h2h3 e5f4 e3f4 c8e6 d2f3 d8e8 c3d5 a8d8 f4e3 e8f7 f3d4 c5d4 e3d4 c6d4 d1d4 f8c5 d4d3 h8e8 d5c7 e8f8 f1e2 d8d3 e2d3 c5d6 c7d5 h7h5 c1c2 a7a5 d3e2 g7g6 h1d1 h5h4 d1f1 d6e5');
INSERT INTO evaluations VALUES(31569,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31569,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31569,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31569,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31569,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31569,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31569,36,'cp 0','c1f4 b8d7 f1d3 d8b6 c2c3 f5e4 f3e4 e6e5 f4e3 e5d4 e3d4 c6c5 d4e3 b6b2 g1e2 e8c8 a1b1 d7e5 e3d4 e5d3 d1d3 b2a2 b1b7 c8b7 d3b5 b7c7 d4e5 f8d6 b5c5 c7b7 c5b5 b7c7');
INSERT INTO evaluations VALUES(37693,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37693,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37693,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37693,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37693,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37693,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37693,36,'cp 0','d1d4 b4f8 d4e5 b8c6 e5g3 d7d6 f1c4 g8f6 e1g1 c8d7 f1d1 a7a6 c4e2 a8c8 c1e3 f6h5 g3h3 h5f6');
INSERT INTO evaluations VALUES(37691,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37691,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37691,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37691,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37691,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37691,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37691,36,'cp 0','b4f8 d4e5 b8c6 e5g3 d7d6 c1e3 c8d7 a1d1 g8f6 f1b5 f6h5 g3h3 h5f6');
INSERT INTO evaluations VALUES(37796,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(37796,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(37796,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(37796,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(37796,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37796,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(37796,36,'cp 73','d4d5 e7e6 g1f3 d7d6 b1c3 e6d5 c4d5 g8f6 f1e2 e8g8 e1g1 f8e8 f3d2 b8d7 a2a4 g6g5 f1e1 h7h6 d1c2 d7e5 d2f1 a7a6 c1d2 a8b8 h2h3 c8d7 a4a5 b7b5 a5b6 d8b6 c3d1 b6b3 c2b3 b8b3 a1a6');
INSERT INTO evaluations VALUES(37837,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(37837,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(37837,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(37837,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(37837,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(37837,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(37837,36,'cp 71','e7e6 b1c3 d7d6 g1f3 e6d5 c4d5 g8f6 f3d2 e8g8 f1e2 f8e8 e1g1 b8d7 a2a4 d7e5 d2b1 h7h6 f2f4 e5d7 e2f3 c5c4 c1e3 b7b6 b1a3 d7c5 e3c5 b6c5 d1c2 g6g5 a1e1 g5f4 a3c4 c8a6 c3b5 a6b5 a4b5 a8b8');
INSERT INTO evaluations VALUES(53104,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(53104,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53104,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(53104,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(53104,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(53104,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53104,36,'cp 33','c2c4 e7e5 f1g2 g7g6 b1c3 f8g7 e1g1 b8c6 a2a3 a7a5 d2d3 g8e7 f3e1 c8e6 c1g5 f7f6 g5d2 e8g8 c3d5 a5a4 e2e3 b7b5 d5e7 c6e7 g2a8 d8a8 c4b5 f8b8 e1g2 b8b5 d2c3 e6b3 d1e2 d6d5 e3e4 d5d4 c3d2 e7c6 f2f4');
INSERT INTO evaluations VALUES(53030,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53030,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(53030,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53030,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(53030,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(53030,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(53030,36,'cp 34','e7e5 f1g2 g7g6 e1g1 b8c6 a2a3 f8g7 b1c3 a7a5 d2d3 g8e7 c1g5 e8g8 f3e1 f7f6 g5d2 f6f5 a1b1 c8e6 b2b4 a5b4 a3b4 c6b4 e1c2 e7c6 c2b4 c6b4 g2b7 a8b8 b7g2 d8d7 d1c1 b4c6 b1b8 f8b8 d2h6 g7h6 c1h6 c6d4');
INSERT INTO evaluations VALUES(56410,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(56410,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(56410,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(56410,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(56410,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(56410,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(56410,36,'cp -63','d5e4 b1c3 b8c6 d3e4 d8d1 c3d1 f8d6 d1c3 e5f4 c3b5 g8f6 f1d3 e8g8 g1e2 c8g4 b5d6 c7d6 c1f4 d6d5 e4d5 f6d5 e1g1 c6b4 e2c3 d5c3 b2c3 b4d3 c2d3 f8d8 d3d4 a8c8 a1e1 g4e6 e1e3 e6d5 f1e1 f7f6 h2h4 h7h6 a2a3 c8c4 e3e8 d8e8 e1e8 g8h7 e8d8 d5c6 h4h5 c4c3 d4d5');
INSERT INTO evaluations VALUES(56623,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(56623,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(56623,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(56623,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(56623,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(56623,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(56623,36,'cp -67','b1c3 b8c6 d3e4 d8d1 c3d1 f8d6 d1c3 e5f4 c3b5 g8f6 f1d3 e8g8 g1e2 c8g4 b5d6 c7d6 c1f4 f8e8 e2g3 d6d5 e1g1 f6e4 d3e4 d5e4 a1e1 c6b4 f1f2 b4a2 e1e4 b7b5 e4e8 a8e8 f2d2 g4e6 c2c3 f7f6 d2d4 a7a5 g3e4 b5b4 c3b4 a2b4 e4c5 e6f7');
INSERT INTO evaluations VALUES(732,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(732,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(732,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(732,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(732,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(732,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(732,36,'cp 90','d2f3 c5d4 c3d4 f7f6 e5f6 d7f6 e1g1 f8d6 c1f4 d6f4 e2f4 c8d7 f1e1 e8g8 d1d2 a8e8 d3c2 b6b4 d2c3 b4c3 b2c3 f6e4 c2e4 f8f4 e4d3 h7h6 a1b1 b7b6 e1e3 e8c8 b1e1 f4f6 d3b5 g8f8 h2h4 f8e7 b5c6 d7c6 f3e5 c6b5');
INSERT INTO evaluations VALUES(3594,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(3594,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(3594,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(3594,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(3594,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(3594,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3594,36,'cp 32','d1c1 c5b4 e1g1 b4c3 b2c3 d8e7 f1e1 c6a5 c4b3 b7b6 d3d4 g8h7 h2h3 c8b7 a2a4 a8e8 b3a2 b7c8 c1d1 c8e6 a2b3 a5b3 c2b3 g7g5 h4g3 g5g4 h3g4 e6g4 d1d3 e7d7 d4e5 d6e5 a1d1 d7d3 d1d3 f6d7 b3b4 f8g8 a4a5 g4f3');
INSERT INTO evaluations VALUES(10474,30,'cp -387',NULL);
INSERT INTO evaluations VALUES(10474,31,'cp -401',NULL);
INSERT INTO evaluations VALUES(10474,32,'cp -400',NULL);
INSERT INTO evaluations VALUES(10474,33,'cp -409',NULL);
INSERT INTO evaluations VALUES(10474,34,'cp -421',NULL);
INSERT INTO evaluations VALUES(10474,35,'cp -389',NULL);
INSERT INTO evaluations VALUES(10474,36,'cp -421','c6d4 e4e5 f6g8 d1g4 b4c3 d2c3 d4e6 f1c4 h7h5 g4e4 d8g5 h1f1 g5f4 e4e2 b7b5 c4b5 c8b7 g2g3 f4f3 e1c1 f3e2 b5e2 g8e7 f2f4 g7g6 e2c4 e8c8 d1d3 h5h4 g3g4 d8g8 f4f5 g6f5 g4f5 e6c5 e5e6 c5d3 c2d3');
INSERT INTO evaluations VALUES(10791,30,'cp -409',NULL);
INSERT INTO evaluations VALUES(10791,31,'cp -424',NULL);
INSERT INTO evaluations VALUES(10791,32,'cp -421',NULL);
INSERT INTO evaluations VALUES(10791,33,'cp -421',NULL);
INSERT INTO evaluations VALUES(10791,34,'cp -410',NULL);
INSERT INTO evaluations VALUES(10791,35,'cp -425',NULL);
INSERT INTO evaluations VALUES(10791,36,'cp -398','e4e5 f6g8 d1g4 b4c3 d2c3 d4e6 f1c4 h7h5 g4e2 d8g5 c3d2 g5g6 e1g1 g8e7 c4d3 g6g4 e2e1 e6c5 d3e2 g4h4 a2a4 b7b6 b2b4 c5e6 f2f4 h4e1 d2e1 g7g6 a4a5 c8b7 e1h4 e7f5 h4f6 e8g8 e2b5 f5e3 f1f2 e3g4 f2f1 g4f6');
INSERT INTO evaluations VALUES(10835,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(10835,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(10835,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10835,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(10835,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(10835,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(10835,36,'cp 42','a2a4 e8g8 f1e1 d7d6 c2c3 a7a6 b1d2 a6a5 h2h3 c8e6 c4e6 f7e6 d2c4 f6d7 c1d2 d7c5 d2e3 b7b6 d3d4 e5d4 f3d4 c6e5 c4e5 d6e5 d4f3 e7d6 d1e2 d8e8 e3c5 d6c5 e1d1 a8d8 d1d8');
INSERT INTO evaluations VALUES(10859,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(10859,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(10859,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(10859,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(10859,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(10859,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(10859,36,'cp 65','e1g1 e7d6 f1e1 a7a6 b5c4 d6c5 c3d5 h7h6 c2c3 c5a7 d2d4 d7d6 h2h3 e8g8 c1e3 f8e8 d1c2 f6d5 c4d5 e5d4 c3d4 c6e7 d5b3 d6d5 e4e5 c8f5 c2d1 f5e4 f3h4 d8d7 b3c2 c7c6 a1c1 e7f5 c2e4 d5e4 h4f5 d7f5 d1b3');
INSERT INTO evaluations VALUES(10858,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(10858,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(10858,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(10858,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(10858,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(10858,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(10858,36,'cp 59','e7d6 f1e1 a7a6 b5a4 e8g8 d2d4 e5d4 c3d5 f6d5 e4d5 b7b5 a4b3 c6a5 d1d4 c8b7 c1g5 f7f6 g5d2 a5b3 a2b3 f8f7 a1d1 f7e7 c2c4 e7e1 d1e1 d8f8 d4e4 c7c6 e4f5 c6d5 f5d7 a8b8 d7e6 g8h8 c4b5 a6b5 e6d7 f8d8 d7b5 d5d4 b5c4 d4d3 e1d1 d6c7');
INSERT INTO evaluations VALUES(11123,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(11123,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(11123,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(11123,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(11123,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(11123,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(11123,36,'cp 18','e7e5 d4e2 f6g4 d1d2 g4e3 d2e3 d8a5 e1c1 f8c5 e3g3 e8g8 c1b1 d7d6 f2f4 e5f4 e2f4 c5d4 f1c4 d4e5 h1f1 c8d7 c3d5 a8e8 d5e3 a5b6 c4b3 c6d4 e3c4 b6c6 c4e5 d6e5 f4d5 c6g6 g3g6 h7g6 b3c4 b7b5 c4d3 a6a5 c2c3 d4e6 g2g4 e6c5 g4g5 g8h8 d3c2 d7g4 d1d2 g4e6 d5c7');
INSERT INTO evaluations VALUES(11106,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11106,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(11106,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(11106,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(11106,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(11106,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(11106,36,'cp 11','d4c6 b7c6 f2f4 e5f4 e3f4 f8b4 d1d4 d8a5 f1d3 e8g8 e1c1 f8e8 e4e5 b4c3 d4c3 a5c3 b2c3 f6d5 f4d2 e8e5 c3c4 d5f6 d2c3 e5e6 d3f5 e6e2 h1e1 e2e1 d1e1 d7d5 f5c8 a8c8 c3f6 g7f6 e1e7 d5c4 c1b2 c8d8 e7a7 d8d2 a2a4 d2g2 a7a6 c4c3 b2c3');
INSERT INTO evaluations VALUES(11583,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(11583,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(11583,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(11583,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(11583,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(11583,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(11583,36,'cp 117','b5c6 e7c6 d4d5 c6a7 f3e5 d8e7 e5f3 e7e4 c1e3 f8b4 c2c3 b4c5 e1g1 c5e3 f1e1 e8g8 e1e3 e4f4 b1d2 d7d6 a2a4 c8d7 e3e7 f8d8 d1b3 b7b6 h2h3 h7h6 c3c4 a7c8 e7e4 f4f6 a1e1 b6b5 a4a5 c8a7 e4e7 b5c4 b3b7 d7f5 b7a6 f6b2 a6c4 a7b5 c4f4 b2f6 f4a4 b5c3');
INSERT INTO evaluations VALUES(11579,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(11579,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(11579,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(11579,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(11579,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(11579,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(11579,36,'cp 111','e7c6 d4d5 c6a7 f3e5 d8e7 e5f3 e7e4 c1e3 f8b4 c2c3 b4c5 e1g1 c5e3 f1e1 e8g8 e1e3 e4f4 b1d2 d7d6 c3c4 c8d7 d1b3 b7b6 a1e1 b6b5 d2e4 c7c5 a2a4 a7c8 c4b5 a6b5 a4b5 a8b8 e4c3 c8b6 h2h3 h7h6 f3d2 f4d4 d2e4 d4b4 b3c2 b6c4 e3g3 f7f5 b2b3 c4a3');
INSERT INTO evaluations VALUES(11556,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(11556,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(11556,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(11556,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(11556,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(11556,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(11556,36,'cp 123','d4d5 c6a7 f3e5 d8e7 e5f3 e7e4 c1e3 f8b4 c2c3 b4c5 e1g1 c5e3 f1e1 e8g8 e1e3 e4f4 b1d2 d7d6 a2a4 c8d7 d1b3 a8b8 c3c4 b7b6 b3c3 a6a5 a1e1 b8e8 d2e4 a7c8 b2b3 c8e7 g2g3 f4h6 e4f6 h6f6 c3f6 g7f6 e3e7 e8e7 e1e7 f8d8 f3d4 g8f8 e7e3 d8e8 e3e8 f8e8 g1f1 d7g4 f2f3 g4h5 f1e2 h5g6');
INSERT INTO evaluations VALUES(11553,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(11553,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(11553,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(11553,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(11553,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(11553,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(11553,36,'cp 87','c6a7 f3e5 d8e7 e5f3 e7e4 c1e3 f8b4 c2c3 b4c5 e1g1 c5e3 f1e1 e8g8 e1e3 e4f5 c3c4 d7d6 b1c3 c8d7 d1d4 b7b6 a2a4 a8e8 a4a5 a7c8 a5b6 e8e3 f2e3 c8b6 e3e4 f5f4 c3e2 f4h6 h2h3 f8e8 a1a6 f7f5 f3d2 f5e4 a6a7 e4e3 d2f1 h6g5 f1g3 g5d8 g3h5 d8g5');
INSERT INTO evaluations VALUES(11827,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11827,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(11827,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(11827,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(11827,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(11827,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(11827,36,'cp 13','c1f4 f6e4 c3e4 d5e4 d1d8 e8d8 e1c1 f8d6 f4d6 c7d6 d1d6 d8c7 d6d4 f7f5 f2f3 h8e8 f3e4 e8e4 d4e4 f5e4 f1e2 c8e6 h1f1 c7d6 c1d2 a7a5 f1f4 e6a2 f4e4 a2d5 e4g4 g7g6 e2d3 d6c5 h2h4 a8f8 d2e3 d5e6 g4f4');
INSERT INTO evaluations VALUES(14181,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(14181,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(14181,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(14181,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14181,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(14181,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(14181,36,'cp 32','c6e5 g1f3 e5f3 d1f3 e7e6 f1c4 e6d5 e4d5 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 c8g4 f3d3 f6h5 g3d6 d8d6 a1c1 g4d7 d3d4 d6b6 f1d1 h5f6 h2h3 b6d4 d1d4 f8e8 c1d1 e8e5 d5d6 c7c6 f2f3 a8e8 g2g4 g8f8');
INSERT INTO evaluations VALUES(14495,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(14495,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(14495,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(14495,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14495,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(14495,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(14495,36,'cp 40','g1f3 e5f3 d1f3 e7e6 f1c4 e6d5 e4d5 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 c8g4 f3d3 f6h5 g3d6 d8d6 a1c1 g4d7 d3d4 d6b6 d4b6 a7b6 f2f3 h5f4 f1d1 c7c6 d5c6 d7c6 a2a3 a8d8 g1f2 g8g7 d1d8 f8d8 c1d1 d8d1 c3d1 f4e6 d1e3 e6d4 f2g3 g7g6');
INSERT INTO evaluations VALUES(39650,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(39650,31,'cp 261',NULL);
INSERT INTO evaluations VALUES(39650,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(39650,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(39650,34,'cp 263',NULL);
INSERT INTO evaluations VALUES(39650,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(39650,36,'cp 270','b1c3 h7h6 c1e3 d7d6 h2h4 g5g4 g1e2 g8f6 e2g3 b8c6 d1d2 c8d7 e1c1 e7e6 f2f3 g4f3 g2f3 d8e7 d4d5 c6e5 f1e2 e6d5 e4d5 a7a6 h1e1 e8d8 f3f4 e5g6 h4h5 g6h4 e3f2 h4g2 e1g1 g2h4');
INSERT INTO evaluations VALUES(42330,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(42330,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(42330,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(42330,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(42330,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(42330,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(42330,36,'cp 46','d2d4 e5d4 d1d4 b8c6 d4e3 c6e7 e3d2 e8g8 e2e4 a7a6 g1e2 a8b8 a2a4 d7d6 e1g1 f6d7 a1b1 e7c6 b2b3 f8e8 c1a3 b7b6 c3d5 d7c5 d2c2 c5e6 a3b2 c6d4 e2d4 e6d4 b2d4 g7d4 b1d1 c7c5 d5c3 d8e7 d1d2 c8b7 f1d1 b8d8 d2d4 c5d4 d1d4 g6g5');
INSERT INTO evaluations VALUES(47548,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(47548,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(47548,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(47548,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47548,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(47548,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(47548,36,'cp 18','b2c3 g7g6 h2h4 f8g7 h4h5 c8g4 h5h6 g7f6 a1b1 b7b6 d2d4 e8g8 f1e2 c7c5 e1g1 b8c6 d4c5 f6c3 c5b6 a7b6 b1b3 c3f6 a2a3 d8c7 c1e3 f8b8 d1b1 c6e5 f1d1 e5c4 e3d4 b8e8 b3b4 c4a3 b1b2 a3c2 d4f6 c2b4 f6c3');
INSERT INTO evaluations VALUES(47544,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47544,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(47544,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(47544,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(47544,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(47544,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(47544,36,'cp 21','g7g6 h2h4 h7h5 f1e2 f8g7 d2d4 b8c6 e1g1 e8g8 c1f4 c8g4 f1e1 e7e6 f3g5 g4e2 d1e2 d8d7 a1d1 a8d8 e2e4 d7d5 f4c7 d8d7 e4d5 d7d5 g5e4 f8c8 c7f4 b7b6 e4d6 c8d8 d6c4 e6e5 f4g5 d8d7 c4e3 d5a5 d4e5 d7d1');
INSERT INTO evaluations VALUES(49496,30,'cp 812',NULL);
INSERT INTO evaluations VALUES(49496,31,'cp 797',NULL);
INSERT INTO evaluations VALUES(49496,32,'cp 794',NULL);
INSERT INTO evaluations VALUES(49496,33,'cp 801',NULL);
INSERT INTO evaluations VALUES(49496,34,'cp 797',NULL);
INSERT INTO evaluations VALUES(49496,35,'cp 801',NULL);
INSERT INTO evaluations VALUES(49496,36,'cp 801','e5c6 d8e7 c6e7 f8e7 g2g4 b8c6 g4g5 f6d5 c2c3 e8g8 h2h4 d5b6 d2d4 f8e8 c1e3 d7d6 b1d2 b6d5 f1g2 c8e6 e1g1 c6d8 a1e1 d5e3 f2e3 e6d7 d2e4 e8f8 e2d3 d8e6 c3c4 c7c6 d3b3 b7b5 c4b5 c6b5 h4h5 e7g5 e4d6');
INSERT INTO evaluations VALUES(49287,30,'cp 775',NULL);
INSERT INTO evaluations VALUES(49287,31,'cp 775',NULL);
INSERT INTO evaluations VALUES(49287,32,'cp 774',NULL);
INSERT INTO evaluations VALUES(49287,33,'cp 778',NULL);
INSERT INTO evaluations VALUES(49287,34,'cp 766',NULL);
INSERT INTO evaluations VALUES(49287,35,'cp 776',NULL);
INSERT INTO evaluations VALUES(49287,36,'cp 771','d8e7 c6e7 f8e7 g2g4 b8c6 g4g5 f6d5 c2c3 e8g8 e2f3 d5b6 d2d4 d7d6 f1d3 c8e6 e1g1 e6c4 d3c4 b6c4 h2h4 a8e8 f3g2 c4b6 b1d2 f7f6 f2f4 f6g5 h4g5 g7g6 a2a4 f8f5 d2e4 c6a5 e4g3 a5b3 g3f5 b3a1 f5e7 e8e7');
INSERT INTO evaluations VALUES(34009,30,'cp 786',NULL);
INSERT INTO evaluations VALUES(34009,31,'cp 778',NULL);
INSERT INTO evaluations VALUES(34009,32,'cp 797',NULL);
INSERT INTO evaluations VALUES(34009,33,'cp 794',NULL);
INSERT INTO evaluations VALUES(34009,34,'cp 796',NULL);
INSERT INTO evaluations VALUES(34009,35,'cp 796',NULL);
INSERT INTO evaluations VALUES(34009,36,'cp 790','c6e7 f8e7 g2g4 b8c6 g4g5 f6d5 c2c3 e8g8 e2f3 d5b6 d2d4 d7d6 f1d3 c8e6 e1g1 e6c4 d3c4 b6c4 h2h4 a8e8 f3g2 c4b6 b1a3 d6d5 a3c2 e7d6 h4h5 f7f5 g2h3 g7g6 c2e1 b6c4 e1g2 e8e4 b2b3 c4b6 a2a4 a7a5 h5g6 h7g6 h3h6 c6e7 f2f4');
INSERT INTO evaluations VALUES(33993,30,'cp 767',NULL);
INSERT INTO evaluations VALUES(33993,31,'cp 771',NULL);
INSERT INTO evaluations VALUES(33993,32,'cp 778',NULL);
INSERT INTO evaluations VALUES(33993,33,'cp 775',NULL);
INSERT INTO evaluations VALUES(33993,34,'cp 766',NULL);
INSERT INTO evaluations VALUES(33993,35,'cp 767',NULL);
INSERT INTO evaluations VALUES(33993,36,'cp 767','f8e7 g2g4 b8c6 g4g5 f6d5 c2c3 e8g8 h2h4 d5b6 d2d4 d7d5 e2f3 e7d6 h4h5 c6e7 f1d3 g7g6 e1f1 c8e6 a2a4 a8e8 b1d2 b6d7 b2b4 c7c6 b4b5 e7f5 a4a5 c6b5 a5a6 b7b6 d3b5 e8e7 c1b2 f5g7 c3c4 d5c4 d2c4 e6c4 b5c4 g7e6 b2c3 e6g5');
INSERT INTO evaluations VALUES(33476,30,'cp 798',NULL);
INSERT INTO evaluations VALUES(33476,31,'cp 789',NULL);
INSERT INTO evaluations VALUES(33476,32,'cp 789',NULL);
INSERT INTO evaluations VALUES(33476,33,'cp 794',NULL);
INSERT INTO evaluations VALUES(33476,34,'cp 785',NULL);
INSERT INTO evaluations VALUES(33476,35,'cp 794',NULL);
INSERT INTO evaluations VALUES(33476,36,'cp 790','b1c3 e8g8 e2d1 b7b6 c3b5 e7d8 d2d3 c8b7 f1e2 b7g2 h1g1 g2b7 c1h6 g7g6 h6f8 g8f8 d1d2 b8c6 e2f3 a7a6 b5c3 a8b8 d2f4 f8g7 h2h4 c6a5 f3e2 g7g8 e1c1 f6d5 c3d5 b7d5 c2c4 d5e6 h4h5 a5c6 h5g6');
INSERT INTO evaluations VALUES(56024,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(56024,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(56024,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(56024,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(56024,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(56024,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(56024,36,'cp -55','a7a6 e2e3 b7b5 a3a4 b5b4 c3b1 c7c5 g1f3 b8c6 f1c4 g8f6 e1g1 c8b7 b1d2 c5d4 d2b3 f8d6 e3d4 c6a5 b3a5 d8a5 d1e2 a8c8 f3e5 e8g8 f1e1 c8c7 c4b3 h7h6 c1d2 a5b6 d2e3 b7d5 e5c4 b6b8 g2g3 c7c8 c4d6 b8d6 b3d5 f6d5 e1c1 d5e3 e2e3 c8c1 a1c1 f8d8');
INSERT INTO evaluations VALUES(50984,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(50984,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(50984,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(50984,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(50984,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(50984,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(50984,36,'cp -32','g1f3 b7b5 a3a4 b5b4 c3b1 c7c5 e2e3 c8b7 f1c4 b8c6 e1g1 g8f6 b1d2 c5d4 d2b3 c6a5 b3a5 d8a5 e3d4 f8d6 d1e2 e8g8 f3e5 a8c8 c1g5 c8c7 c4b3 f6e4 g5e3 f8c8 f2f3 e4f6 e5c4 a5d5 a1c1 h7h6 f1d1 a6a5 c4d6 d5d6 c1c5 b7d5');
INSERT INTO evaluations VALUES(4275,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(4275,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(4275,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4275,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(4275,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(4275,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(4275,36,'cp 21','c2c3 d7d6 h2h3 h7h6 d2d3 c8d7 b1d2 a7a6 b5a4 b7b5 a4c2 c5a7 d2f1 c6e7 d3d4 e7g6 f1g3 c7c5 b2b4 c5b4 c3b4 f8e8 c1e3 e5d4 e3d4 a7b6 c2b3 g6e5 f3e5 d6e5 d4b6 d8b6 g3h5 f6h5 d1d7 b6f6 a1d1 e8d8 d7f5 h5f4 f5f6 g7f6');
INSERT INTO evaluations VALUES(13269,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(13269,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(13269,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13269,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13269,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(13269,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13269,36,'cp 0','b1c3 e7e6 d2d3 f8e7 e1g1 a7a6 a2a4 e8g8 c4b3 d6d5 e4e5 f6d7 d1e1 f7f5 g1h1 c6a5 b3a2 a5c6 a2b3');
INSERT INTO evaluations VALUES(16010,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(16010,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16010,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(16010,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(16010,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16010,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16010,36,'cp -12','f6h5 f4g5 h7h6 g5h4 g7g5 h4g3 f8g7 f1d3 e7e5 d4e5 d7c5 b1c3 h5g3 h2g3 d6e5 d3c4 c8f5 d1e2 c7c6 e3e4 f5g4 e1g1 d8e7 e2e3 e8g8 a2a4 f8d8 f1d1 c5e6 d1d8 e7d8 c4e6 g4e6 c3d1 d8e7 e3e2 g5g4 f3e1 h6h5 d1e3');
INSERT INTO evaluations VALUES(22852,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(22852,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(22852,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(22852,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(22852,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(22852,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(22852,36,'cp -80','d7d5 e2e3 g8f6 f1b5 f8d6 d2d4 e5d4 d1d4 e8g8 d4d2 c6e5 f2f4 e5g6 g1f3 c7c6 b5d3 f8e8 e1g1 d8e7 a1e1 d6b4 d2c1 c8g4 a2a3 b4d6 f3d4 g6h4 h2h3 g4d7 c1d2 d6a3 b2a3 e7a3 d2f2 h4g6 e3e4 d5e4 c3e4 f6e4 d3e4 a3c5 e4d3 e8e1 f1e1 a8e8');
INSERT INTO evaluations VALUES(21741,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(21741,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(21741,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(21741,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(21741,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(21741,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(21741,36,'cp -77','e2e3 g8f6 f1b5 f8d6 d2d4 e5d4 d1d4 e8g8 b5c6 b7c6 g1f3 f8e8 d4h4 a7a5 e1g1 d8d7 f1d1 c8b7 c3e2 f6e4 f3g5 d7f5 g5e4 e8e4 e2g3 f5f2 g1f2 e4h4 b2g7 b7c8 g7f6 h4h2 d1h1 h2h1 a1h1 a5a4 f2f3 a4b3 a2b3 a8a2 c2c4 a2c2 c4d5 c6d5');
INSERT INTO evaluations VALUES(21740,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(21740,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(21740,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(21740,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(21740,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(21740,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(21740,36,'cp -96','g8f6 f1b5 f8d6 d2d4 e5d4 d1d4 e8g8 b5c6 b7c6 g1f3 f8e8 d4h4 a8b8 e1c1 a7a5 h2h3 c8f5 f3d4 d8d7 d4f5 d7f5 h1e1 b8b4 f2f4 f6e4 a2a3 b4b8 c3e4 e8e4 a3a4 b8e8 h4g3 f7f6 c1b1 d6b4 e1e2 f5e6');
INSERT INTO evaluations VALUES(15178,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(15178,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(15178,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(15178,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(15178,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(15178,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(15178,36,'cp -86','f1b5 f8d6 d2d4 e5d4 d1d4 e8g8 d4d2 d8e7 g1f3 a7a6 b5d3 d5d4 c3d1 f6e4 d3e4 e7e4 e1g1 d6b4 d2d3 e4d3 c2d3 d4e3 d1e3 f8d8 f1d1 f7f6 a1c1 a6a5 d3d4 b4d6 d4d5 c6e7 a2a4 e7f5 e3f5 c8f5 f3d2 f5g6 d2c4 d6f8');
INSERT INTO evaluations VALUES(33802,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(33802,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(33802,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(33802,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(33802,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(33802,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(33802,36,'cp 71','b1c3 e7e2 f1e2 f8e7 c1d2 c5d7 f3d4 d7f6 e2f3 b8d7 g2g3 e8g8 e1c1 e7d8 f3g2 c7c6 d4f5 d6d5 h1e1 d8b6 f5e7 g8h8 e1e2 d7c5 e7c8 f8c8 g2h3 c8e8 d1e1 e8e2 e1e2 h8g8 d3d4');
INSERT INTO evaluations VALUES(33801,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(33801,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(33801,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(33801,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(33801,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(33801,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(33801,36,'cp 48','e7e2 f1e2 f8e7 c1d2 e8g8 e1c1 a7a6 h2h4 c5e6 d3d4 f8e8 g2g4 b7b5 h1e1 b8d7 d4d5 e6f8 g4g5 d7b6 a2a3 c8b7 d2e3 f8d7 h4h5 g7g6 h5h6 e7f8 e2f1 e8e7 b2b3 a8e8 f1h3 f7f5 h3g2 d7e5 f3e5 d6e5 a3a4 b5b4 a4a5 b4c3 a5b6');
INSERT INTO evaluations VALUES(54405,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(54405,31,'cp -155',NULL);
INSERT INTO evaluations VALUES(54405,32,'cp -152',NULL);
INSERT INTO evaluations VALUES(54405,33,'cp -147',NULL);
INSERT INTO evaluations VALUES(54405,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(54405,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(54405,36,'cp -121','b8c6 d4f4 f8c5 g1f3 g8e7 c4d5 e6d5 f1d3 c6b4 d3b5 c8d7 e5e6 e8g8 e6d7 a7a6 b5e2 b4c2 e1f1 d8d7 f4d2 c2a1 b2b4 c5b6 c1b2 a6a5 b2a1 a5b4 d2b4 a8a2 a1g7 d7c6 g7d4 b6d4 f3d4 c6c1 b4e1 c1b2 b1c3 a2a1 e2d1 a1d1 e1d1 b2c3');
INSERT INTO evaluations VALUES(20406,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(20406,31,'cp -147',NULL);
INSERT INTO evaluations VALUES(20406,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(20406,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(20406,34,'cp -119',NULL);
INSERT INTO evaluations VALUES(20406,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(20406,36,'cp -137','d4f4 f8c5 g1f3 g8e7 b1c3 e7g6 f4g3 c6b4 a1b1 b4c2 e1d1 c2d4 c4d5 e8g8 f1d3 e6d5 f3d4 c5d4 h1e1 c8e6 c3b5 d4b6 c1g5 b6f2 g5d8 f2g3 h2g3 a8d8 b1c1 d5d4 c1c7 a7a6 d3g6 f7g6 b5d6 e6d5 d6b7 d8b8 b7d6 d5g2 e1e2 d4d3 e2g2');
INSERT INTO evaluations VALUES(20403,30,'cp -176',NULL);
INSERT INTO evaluations VALUES(20403,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(20403,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(20403,33,'cp -156',NULL);
INSERT INTO evaluations VALUES(20403,34,'cp -157',NULL);
INSERT INTO evaluations VALUES(20403,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(20403,36,'cp -150','f8c5 g1f3 g8e7 b1c3 e7g6 f4g3 c6b4 a1b1 b4c2 e1d1 c2d4 c4d5 d4f3 g2f3 e8g8 f1d3 c5d4 c1g5 d8b6 h2h4 e6d5 c3d5 b6a5 d3g6 a5d5 g6e4 d4f2 e4d5 f2g3 d1e2 g3e5 g5e3 f8e8 d5e4 f7f5 e4d5 c8e6 d5b7 a8b8 b7c6 e8c8 h1c1 b8b4 e3a7 b4h4 a7f2 h4h2 b2b3 h7h5');
INSERT INTO evaluations VALUES(57316,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(57316,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(57316,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(57316,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(57316,34,'cp -113',NULL);
INSERT INTO evaluations VALUES(57316,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(57316,36,'cp -112','c3b2 c1b2 g8f6 b1c3 d7d6 f1c4 f8e7 e4e5 d6e5 d1b3 e8g8 f3e5 d8e8 e1g1 b8a6 c3b5 c7c6 b5d4 e7d6 f2f4 b7b5 c4e2 a6c5 b3c2 f6d5 e2f3 d5e3 c2f2 e3f1 f3c6 e8d8 c6a8 d6e5 d4c6 c5d3 c6d8 d3f2 b2e5 f8d8 a1f1 f2d3 f1d1 c8g4 d1d2 g4f5 a8f3 d8c8 e5d6 a7a5');
INSERT INTO evaluations VALUES(57319,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(57319,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(57319,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(57319,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(57319,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(57319,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(57319,36,'cp -88','c1b2 g8f6 b1c3 d7d6 f1c4 f8e7 e4e5 d6e5 d1b3 e8g8 f3e5 d8e8 e1g1 b8a6 c3b5 c7c6 b5d4 e7d6 f2f4 b7b5 c4e2 a6c5 b3c2 f6d5 e2f3 d5e3 c2f2 e3f1 f3c6 e8d8 c6a8 d6e5 d4c6 c5d3 c6d8 d3f2 b2e5 f8d8 a1f1 f2d3 f1d1 c8f5 a8f3 f7f6 e5c7 d8d7 c7b8 a7a6 f3c6 d7d8 b8a7 g8f7 c6f3 f5e6');
INSERT INTO evaluations VALUES(57318,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(57318,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(57318,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(57318,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(57318,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(57318,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(57318,36,'cp -95','g8f6 b1c3 d7d6 f1c4 f8e7 e4e5 d6e5 d1b3 e8g8 f3e5 d8e8 e1g1 b8a6 c3b5 c7c6 b5d4 e7d6 f2f4 a6c5 b3c2 b7b5 c4e2 f6d5 e2f3 d5e3 c2f2 e3f1 f3c6 e8d8 c6a8 d6e5 d4c6 c5d3 b2e5 d3f2 c6d8 f8d8 a1f1 f2d3 f1d1 c8g4 d1d2 g4f5 a8f3 d8e8 f3g4 f5g4 d2d3');
INSERT INTO evaluations VALUES(17134,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(17134,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(17134,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(17134,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(17134,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(17134,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(17134,36,'cp 232','a2a3 d7d6 h2h3 c6a5 c4a2 c7c5 e1g1 a5c6 d2d3 b7b6 f3d2 c6d4 d2c4 c8e6 f2f4 d8d7 f4f5 e6f7 c4e3 f7a2 a1a2 e8c8 b2b4 c8b8 c3d5 e7d5 e3d5 d4b5 c2c4 b5c7 d5c7 d7c7 b4b5 h7h5 a3a4 d8d7 a4a5 b6a5 c1e3 f8e7 d1a4 b8a8 a4a5');
INSERT INTO evaluations VALUES(21076,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(21076,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(21076,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(21076,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(21076,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(21076,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(21076,36,'cp -21','e7e5 f1g2 g8f6 e1g1 c8e6 c2c3 h7h6 b2b4 f8d6 b4b5 c6e7 c3c4 c7c6 b5c6 b7c6 d1a4 e8g8 b1d2 e6d7 c1a3 c6c5 a4a6 d8b6 a6b6 a7b6 c4d5 a8a3 d2c4 f8a8 c4d6 f6d5 f3e5 d7e6 g2d5 e7d5 e2e4 d5b4 d6c4 a3a2 a1a2 a8a2 c4b6');
INSERT INTO evaluations VALUES(22627,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(22627,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(22627,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(22627,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(22627,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(22627,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(22627,36,'cp 26','d3d4 d8h4 b1c3 f8b4 d1d3 g8f6 e4e5 h4e4 c1e3 e4d3 f1d3 f6d5 g1e2 d7d6 e5d6 c7d6 e1c1 d5e3 f2e3 h7h5 c3b5 e8e7 a2a3 b4a5 c2c4 a7a6 b5c3 a5c3 e2c3 c8d7 c1d2 h5h4 d1b1 g7g5 b2b4 f7f5 d3e2 a8f8 a3a4 f8g8 a4a5 g5g4 h1f1 g8g6');
INSERT INTO evaluations VALUES(30781,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(30781,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(30781,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(30781,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(30781,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(30781,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(30781,36,'cp 31','b1c3 e7e5 g1f3 b8c6 a2a3 f7f5 e4f5 c6d4 f1e2 d8f6 d2d3 d4e2 d1e2 f6f5 e2e5 f5e5 f3e5 d7d6 e5f3 g7g6 d3d4 e8d7 d4d5 f8g7 c1e3 b7a6 f3d2 g7c3 b2c3 g8f6 e1c1 h8e8 c1c2 c7c6 d5c6 d7c6 h1e1 f6e4 c2b3 e4d2 d1d2 e8e4 d2d4 a8e8 e1d1');
INSERT INTO evaluations VALUES(31049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31049,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31049,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31049,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31049,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(31049,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(31049,36,'cp -1','e7e5 c5b7 d8b6 b7c5 f8c5 d4c5 b6c5 c2c3 g8f6 f1d3 f5d3 d1d3 b8d7 g1e2 c5d5 d3d5 f6d5 c1d2 d7c5 e1g1 c5d3 c3c4 d5b4 f1d1 e8c8 a1b1 c6c5 a2a3 b4c6 d2c3 d8d7 e2g3 h8d8 b2b4 c5b4 a3b4 d3f4 d1d7 d8d7 b4b5 c6d4 b1d1 d7d8');
INSERT INTO evaluations VALUES(36656,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(36656,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(36656,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(36656,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(36656,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(36656,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(36656,36,'cp 66','g1e2 b8d7 d1d2 d7b6 b2b3 e7e5 d4e5 d6e5 d2d8 f8d8 e2c1 c7c6 a2a4 c8e6 c1d3 f6d5 e4d5 c6d5 c3b5 e5e4 e3d4 e4d3 f1d3 b6d7 d4g7 g8g7 e1d2 a7a5 b5d4 d7e5 a1e1 g7f6 g2g4 e5d3 d2d3 e6d7 h2h4 h7h6 h1g1 d8e8 g4g5 f6g7 d3d2 h6h5 e1e8 a8e8 c2c3 e8c8');
INSERT INTO evaluations VALUES(37239,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(37239,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(37239,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(37239,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37239,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(37239,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(37239,36,'cp 7','f3e5 d7d6 e5f3 f6e4 d2d3 e4f6 c1g5 h7h6 g5h4 d6d5 d3d4 c5b6 h4f6 d8f6 c4d5 f8d8 d1e2 b8c6 d5c6 b7c6 b1d2 b6d4 d2e4 f6f4 f3d4 d8d4 f2f3 c8f5 a1d1 a8d8 d1d4 d8d4 e2f2 d4d5 f2a7 f4e5 a7a8 g8h7 a8c6 f5e4 f3e4 e5e4');
INSERT INTO evaluations VALUES(37207,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(37207,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37207,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(37207,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(37207,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(37207,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(37207,36,'cp 9','d7d6 d2d4 c5b6 e5f3 f6e4 b1d2 d6d5 d2e4 d5c4 a2a4 a7a5 f1e1 c8e6 c2c3 f8e8 e4g3 e6d5 e1e8 d8e8 c1f4 b8d7 g3f5 e8e6 f5e3 c7c6 d1d2 a8e8 e3d5 e6d5 a1e1 e8e1 d2e1 d7f6 f4b8 f6e4 f3d2 e4f6 h2h3');
INSERT INTO evaluations VALUES(40969,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40969,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40969,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40969,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40969,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40969,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40969,36,'cp 5','e8g8 c1f4 b7b6 e2e3 c7c5 d4c5 b6c5 f1d3 c8b7 c4d5 f6d5 c3d5 e6d5 d1c2 h7h6 e1g1 b8c6 a1d1 d8b6 e3e4 d5e4 d3e4 a8c8 d1d7 e7f6 f1d1 c6d4 d7b7 b6b7 e4b7 d4c2 b7c8 f8c8 f4e5 c2d4 e5f6 d4f3 g2f3 g7f6 d1c1 g8g7 c1c4 c8b8 c4c5 b8b2 c5c7 b2b3 a3a4 b3f3 c7a7 g7g6 a4a5 f3a3 a5a6');
INSERT INTO evaluations VALUES(36339,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(36339,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36339,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36339,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36339,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(36339,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(36339,36,'cp 16','c1f4 b7b6 e2e3 c7c5 f1e2 d5c4 e2c4 c8b7 e1g1 c5d4 f3d4 a7a6 c4e2 b8c6 e2f3 a8c8 a1c1 b7a8 d4c6 a8c6 e3e4 b6b5 d1e2 d8b6 f1d1 c8d8 e4e5 f6d5 f3d5 c6d5 c3d5 d8d5 d1d5 e6d5 b2b4 b6e6');
INSERT INTO evaluations VALUES(36316,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36316,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(36316,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36316,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36316,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36316,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(36316,36,'cp 12','b7b6 e2e3 c7c5 d4c5 b6c5 f1d3 c8b7 c4d5 f6d5 c3d5 e6d5 e1g1 b8d7 d1c2 d7f6 f4e5 d8b6 a1b1 a7a5 e5f6 b6f6 d3h7 g8h8 h7d3 b7c8 c2e2 c5c4 d3c2 c8g4 f1e1 f8d8 h2h3 g4f3 e2f3 f6f3 g2f3 g7g6 f3f4 a8b8 e1d1 h8g8');
INSERT INTO evaluations VALUES(44571,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44571,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(44571,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44571,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44571,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44571,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44571,36,'cp 5','f2f4 b8c6 g1f3 e7e6 f3e5 c6a5 b1c3 d7d6 e5f3 a5c6 g2g3 g7g6 f1g2 f8g7 e1g1 e8g8 d1e2 c6d4 f3d4 c5d4 c3b1 b7b5 a2a4 b5b4 b1d2 c8b7 d2c4 d8c7 e4e5 b7g2 e2g2 d6e5 f4e5 f6d5 c1g5 g7e5');
INSERT INTO evaluations VALUES(57661,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(57661,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(57661,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57661,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(57661,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(57661,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(57661,36,'cp 11','d7d5 f1g2 c7c6 c2c4 g8f6 g1f3 f8g7 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 c1f4 d8b6 d1b3 b8c6 b3b6 a7b6 e2e3 e7e6 e1g1 f8c8 f1c1 c6e5 f4e5 f6e4 g2e4 f5e4 c3e4 d5e4 e5c7 b6b5 b2b3 b5b4 h2h4 h7h5 c7d6 c8c1 a1c1 a8a2 d6b4');
INSERT INTO evaluations VALUES(54865,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(54865,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54865,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54865,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(54865,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(54865,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(54865,36,'cp 16','g1f3 f8g7 f1g2 c7c6 c2c4 g8f6 e1g1 e8g8 c4d5 c6d5 b1c3 b8c6 f3e5 c8f5 e5c6 b7c6 c1f4 f6d7 a1c1 e7e5 d4e5 d7e5 b2b3 h7h5 d1d2 h5h4 f4g5 f7f6 g5h4 g6g5 f2f4 g5h4 f4e5 f5e6 c3a4 f6e5 c1c6 f8f1 g1f1 d8f8 f1e1 g7h6 e2e3');
INSERT INTO evaluations VALUES(54864,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54864,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54864,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54864,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54864,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(54864,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(54864,36,'cp 20','c7c6 c2c4 g8f6 f1g2 f8g7 e1g1 e8g8 c4d5 c6d5 b1c3 b8c6 f3e5 c8f5 e5c6 b7c6 c1f4 f6h5 f4e3 h5f6 d1d2 f6g4 e3f4 e7e5 d4e5 g4e5 b2b3 h7h5 a1c1 f8e8 c3a4 d8e7 f4g5 e7a3 f1d1 a8c8 a4c5 a7a5 g5f4 a3b4 h2h3 e5d7 e2e4 d7c5 e4f5 c5e4');
INSERT INTO evaluations VALUES(1897,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1897,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1897,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1897,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(1897,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(1897,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1897,36,'cp 16','b4f4 c3d5 f4e4 d5c7 e8d8 c7a8 c6e5 d1d2 b7b6 f3e5 e4e5 e1c1 f7f5 g2g4 g8f6 f1g2 f6g4 e2e3 f8c5 c1b1 h8e8 a2a3 g4f6 d2d3 f6e4 d3c4 e5e6 c4a4 a7a5 h1f1 c8b7 g2e4 b7e4 b2b4 e4c6 a4b3 e6b3 c2b3 a5b4 a3b4 c5e3 f2e3 c6a8 b1b2 g7g6 d1d3 e8e6 f1d1 a8c6 d3d6');
INSERT INTO evaluations VALUES(1907,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(1907,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(1907,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(1907,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(1907,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(1907,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(1907,36,'cp 12','c3d5 f4e4 d5c7 e8d8 c7a8 c6e5 d1d2 b7b6 f3e5 e4e5 e1c1 f7f5 g2g4 g8f6 f1g2 f6g4 e2e3 f8c5 c1b1 h8e8 a2a3 g4f6 d2d3 f6e4 d3c4 e8e7 g2e4 e5e4 c4g8 e7e8 g8g7 e4a8 g7h7 a8e4 h2h4 c5e7 h4h5 e7f6 h7f7 e4e5 f7b3 c8b7 h1h3 b7g2 h3g3 g2e4 h5h6 d8c7');
INSERT INTO evaluations VALUES(1886,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1886,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1886,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1886,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(1886,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(1886,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1886,36,'cp 0','f4e4 d5c7 e8d8 c7a8 c6e5 d1d2 b7b6 f3e5 e4e5 e1c1 f7f5 g2g4 g8f6 f1g2 f6g4 e2e3 f8c5 c1b1 h8e8 d2e2 g4f6 e2c4 e5b8 h1g1 g7g6 b2b4 c5f8 c4c6 f6e4 g2e4 e8e4 g1g6 h7g6 c6f6 d8e8 f6g6 e8d8 g6f6 d8e8');
INSERT INTO evaluations VALUES(9459,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(9459,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(9459,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(9459,33,'cp -150',NULL);
INSERT INTO evaluations VALUES(9459,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(9459,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(9459,36,'cp -136','c6d4 f3d1 b7b5 c2c3 b5c4 c3d4 c5d4 g5f3 c8g4 d1a4 d8d7 a4d7 e8d7 f3d4 e5d4 d3c4 g4e6 e1g1 c7c5 c1h6 g7h6 b1d2 f7f5 f1e1 a8b8 b2b3 f5e4 e1e4 e6f5 e4f4 h8f8 f2f3 f5g6 f4h4 h6h5 d2f1 a7a5 f1g3 a5a4 g3h5 a4b3 a2b3 b8b3 h5f4 b3b1 a1b1 g6b1 f4d5 d7c6 h4h5 f8a8');
INSERT INTO evaluations VALUES(21273,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(21273,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(21273,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21273,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(21273,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21273,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(21273,36,'cp 22','c6a5 c4b5 c7c6 b5e2 c6c5 b2b4 a5c6 b4c5 d6c5 e2b5 f8d6 b5c6 b7c6 d2d3 g8e7 f3d2 e7g6 e1g1 c8e6 d1f3 d6c7 c3e2 e8g8 c1a3 d8e7 d2b3 c7d6 e2g3 c5c4 a3d6 e7d6 d3c4 e6c4 f1d1 d6a3 f3e3 g6f4 g3f5 c4e6 b3c5 a3e3 f5e3');
INSERT INTO evaluations VALUES(30590,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(30590,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(30590,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(30590,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(30590,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(30590,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(30590,36,'cp 153','c1g5 f7f6 g5d2 d7d5 f1b5 b8c6 e4e5 f8g7 d2f4 g8e7 e5f6 g7f6 d1e2 d8d7 e1c1 a7a6 b5c6 e7c6 h1e1 e8c8 e2e6 d7e6 e1e6 f6g7 e6e3 h8f8 g2g3 c8b8 h2h4 b7c8 f3g5 g7d4 e3e6 c8e6 g5e6 d4e5 f4e5 c6e5 e6d8 f8d8 c3d5 e5g4 d5f4 d8d1 c1d1 g4f2 d1e2');
INSERT INTO evaluations VALUES(45027,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(45027,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(45027,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(45027,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(45027,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(45027,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(45027,36,'cp 103','f1c4 b7b5 c4d3 d8d4 g1f3 d4d8 e4e5 b5b4 e5f6 b4c3 b2c3 g7f6 f3d4 b8d7 d3e4 d7e5 e1g1 f6f5 d4f5 c8f5 e4f5 d8d1 f1d1 f8g7 a1b1 e8g8 f5e4 e5c4 c1g5 g7f6 g5f6 e7f6 a2a4 f8d8 e4c6 a8b8 b1b8 d8b8 g2g3 c4e5 c6g2');
INSERT INTO evaluations VALUES(55197,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(55197,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(55197,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(55197,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(55197,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(55197,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(55197,36,'cp 41','g1f3 c7c5 d2d4 c5d4 d1d4 c8d7 f1e2 b8c6 d4c3 d8c7 e1g1 g8f6 c3c2 e7e6 b1c3 c6e7 b2b4 e7g6 a1b1 f8e7 f1d1 e8g8 f3d4 a8c8 g2g3 b7b6 c1e3 f8d8 a2a4 h7h6 a4a5 b6a5 d4b5 c7e5 e3d4 e5g5 b5a7 a5b4 a7c8 d8c8 c3e4 f6e4 c2e4 d7c6');
INSERT INTO evaluations VALUES(2380,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(2380,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(2380,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(2380,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(2380,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(2380,35,'cp 236',NULL);
INSERT INTO evaluations VALUES(2380,36,'cp 235','f3e5 f8b4 f2f3 b4c3 b2c3 f6d7 e5d3 b7b6 d3f4 d7c5 c1e3 d8e6 e1c1 e6f4 e3f4 c8a6 f1a6 c5a6 d1d5 e8g8 h1d1 a8c8 h2h4 a6c5 f4e3 c5b7 h4h5 h7h6 d5d7 f8e8 e3f4 b7c5 d7c7 c5e6 c7c8 e8c8 f4e3 c8c3 d1d3 c3c4 d3d7 c4a4');
INSERT INTO evaluations VALUES(2876,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(2876,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(2876,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2876,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2876,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(2876,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(2876,36,'cp 7','f1d3 d7d5 e1g1 b4c3 d2c3 d5e4 c3f6 d8f6 d3e4 f6b2 e4c6 a8b8 a1b1 b2b1 d1b1 b8b1 f1b1 f8d8 c6f3 g8f8 b1b8 d8e8 g2g4 c8a6 b8b4 a6e2 g1g2 e2f3 g2f3 e8e5 b4c4 e5a5 a2a4 c7c6 h2h3 f8e7 c4c6 a5a4 f3e3 a4a3 e3f4 a3h3 c6c7 e7f6');
INSERT INTO evaluations VALUES(2879,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2879,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(2879,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(2879,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(2879,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(2879,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(2879,36,'cp 4','d7d5 e1g1 b4c3 d2c3 d5e4 c3f6 d8f6 d3e4 f6b2 e4c6 a8b8 a1b1 b2b1 d1b1 b8b1 f1b1 f8d8 c6f3 g8f8 b1b8 d8e8 h2h3 a7a6 b8a8 c8f5 a8a6 f5c2 a2a4 c2e4 f3e4 e8e4 g2g3 f8e7 g1g2 e4e2 a6a7 e7d6 a4a5 e2a2 a5a6 a2a4 a7a8 d6c6 a6a7');
INSERT INTO evaluations VALUES(2854,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(2854,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2854,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(2854,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2854,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(2854,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(2854,36,'cp 1','e1g1 b4c3 d2c3 d5e4 c3f6 d8f6 d3e4 f6b2 e4c6 a8b8 a1b1 b2b1 d1b1 b8b1 f1b1 f8d8 c6f3 g8f8 b1b8 d8e8 h2h3 a7a6 b8a8 c8f5 a8a6 f5c2 a2a4 c2e4 f3e4 e8e4 g2g3 e4e1 g1g2 e1a1 g2f3 f8e7 a4a5 e7d7 a6a7 g7g6 a7a8 d7d6 a5a6');
INSERT INTO evaluations VALUES(5737,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(5737,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(5737,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(5737,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(5737,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5737,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(5737,36,'cp 34','b2b4 c5b6 a2a4 a7a6 d1b3 d8f6 c1e3 g8e7 e1g1 e8g8 b1d2 e7g6 a4a5 b6a7 b3b2 c8e6 c4b3 e6b3 b2b3 f6e6 b3e6 f7e6 g2g3 f8f7 g1g2 a7e3 f2e3 g6f8 d3d4 f8d7 d4d5 c6d8 f3h4 f7f1 a1f1 d7f6');
INSERT INTO evaluations VALUES(5724,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(5724,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(5724,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(5724,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(5724,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5724,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5724,36,'cp 33','c5b6 a2a4 a7a6 d1b3 d8f6 c1e3 g8e7 b1d2 e8g8 e1g1 e7g6 g1h1 b6a7 b3b2 a7e3 f2e3 f6e7 d3d4 a6a5 b4b5 c6b8 h2h3 b8d7 c4d5 d7f6 c3c4 e5d4 e3d4 g6f4 b2c2 f6d5 e4d5 e7e2 f3h4 e2d3 c2d3 f4d3 f1f3 d3b4 d2e4');
INSERT INTO evaluations VALUES(5719,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(5719,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(5719,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(5719,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(5719,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(5719,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(5719,36,'cp 40','a2a4 a7a6 d1b3 d8f6 c1e3 g8e7 e1g1 e8g8 b1d2 e7g6 a4a5 b6a7 b3c2 f6e7 f1e1 e7f6 c4b3 c8g4 g1h1 g4e6 b3a4 g6f4 e1f1 f4g6 a1e1 f6e7 d3d4 e5d4 f3d4 c6d4 e3d4 c7c5 b4c5 d6c5 d4e3 e7c7 f2f4 c7a5');
INSERT INTO evaluations VALUES(11535,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(11535,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(11535,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(11535,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(11535,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(11535,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(11535,36,'cp 55','d2d4 d7d6 f1e2 c8d7 e1g1 e5d4 f3d4 g7g6 c1g5 h7h6 g5e3 f8g7 d1d2 c6d4 e3d4 e8g8 a1d1 f8e8 f2f3 b7b5 f1e1 d7e6 e2f1 d8e7 h2h3 a8d8 d4e3 g8h7 c3d5 e6d5 e4d5 e7d7 c2c4 f6h5 d2f2 f7f5 c4b5 a6b5 b2b4');
INSERT INTO evaluations VALUES(11532,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(11532,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(11532,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(11532,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(11532,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(11532,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(11532,36,'cp 59','d7d6 f1e2 c8d7 e1g1 e5d4 f3d4 g7g6 c1g5 h7h6 g5e3 f8g7 f2f3 e8g8 d1d2 c6d4 e3d4 b7b5 a1d1 f8e8 f1e1 d7e6 e2f1 g8h7 h2h3 c7c6 d2f2 d6d5 e4d5 e6d5 e1e8 f6e8 a3a4 e8c7 d4g7 h7g7 c3d5 c7d5');
INSERT INTO evaluations VALUES(17395,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(17395,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(17395,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(17395,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(17395,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(17395,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(17395,36,'cp 12','d2d4 c5d4 f3d4 c6d4 d1d4 b7b5 c1e3 c8b7 e1c1 a8c8 c1b1 g8f6 e4e5 f6d5 c3d5 b7d5 d4d2 f8e7 f1d3 e8g8 e3d4 e7c5 d4c3 d8b6 c3a5 b6b7 f4f5 e6f5 d3f5 d5e6 d2d3 e6f5 d3f5 b7g2 f5d7 b5b4 e5e6 f7e6 d7e6 g8h8 h1e1 g2h2 e6a6 c8a8');
INSERT INTO evaluations VALUES(20062,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(20062,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20062,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20062,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20062,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20062,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(20062,36,'cp 0','d4e3 e7e6 g1f3 g8f6 e1g1 f8e7 e3e2 e8g8 f1d1 d8c7 c1f4 f8d8 b1c3 a7a6 e4e5 d6e5 d1d8 c7d8 f3e5 d8e8 e2f3 c6d4 f3d1 d4c6');
INSERT INTO evaluations VALUES(20058,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(20058,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(20058,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20058,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(20058,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(20058,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(20058,36,'cp 0','e7e6 g1f3 g8f6 e1g1 f8e7 f1d1 e8g8 c4e2 d6d5 b1c3 f8e8 a2a3 d8c7 c3b5 c7b8 e4d5 f6d5 e3e4 d5f6 e4h4 e6e5 e2d3 h7h6 c1h6 g7h6 h4h6 e5e4 h6g5 g8f8 g5h6');
INSERT INTO evaluations VALUES(27242,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(27242,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(27242,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(27242,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(27242,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(27242,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(27242,36,'cp 77','b1c3 g8f6 e1g1 f8e7 c1f4 e8g8 f1e1 a7a6 a2a4 c6d4 d1d4 d7c6 a4a5 f6d7 c3d5 c6d5 c4d5 a8b8 d4b4 e7g5 f4g3 g5h4 g3h4 d8h4 b4c3 c7c6 d5c4 h4e7 b2b4 e7e5 c3e5 d6e5 f2f3 f8d8 e1d1 g8f8');
INSERT INTO evaluations VALUES(27240,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(27240,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(27240,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(27240,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(27240,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(27240,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(27240,36,'cp 66','g8f6 c1e3 f8e7 d4c6 d7c6 f2f3 a7a6 a2a4 f6d7 e1g1 e8g8 b2b4 f8e8 c3d5 d7f6 d5e7 d8e7 e3f2 b7b5 c4b3 b5a4 b3a4 c6a4 a1a4 e7e6 f1e1 f6d7 d1d2 e6c4 e1a1 d7e5 h2h3 h7h6 c2c3 e5d3 a4a5 d3f2 g1f2');
INSERT INTO evaluations VALUES(25542,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(25542,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(25542,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(25542,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(25542,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(25542,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(25542,36,'cp 71','c1e3 c6d4 d1d4 f8e7 e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 f8e8 e3g5 d7c6 c4d5 c6d5 c3d5 d6e7 d5e7 d8e7 g5f6 e7f6 d4f6 g7f6 h1e1 b7b6 a2a4 e8e1 d1e1 g8f8 g2g3 a8d8 e1d1 d8d1 c1d1 f8e7 d1e2 e7d6 e2e3 d6e5 e3f3 c7c6 f3e3 b6b5 a4b5 c6b5 e3d3 f6f5 h2h4 a7a5 c2c4');
INSERT INTO evaluations VALUES(34331,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(34331,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(34331,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(34331,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(34331,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(34331,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(34331,36,'cp 34','g8f6 g1f3 a5b6 a2a4 a7a5 h2h3 b8a6 g2g3 g7g6 f1g2 f8g7 e1g1 e8g8 a1b1 f8e8 f1e1 c8d7 f3e5 d7e6 d2g5 a8d8 d1d2 b6c7 g5f4 c7c8 g1h2 a6b4 e1e2 f6h5 f4h6 c8c7 h6g7 h5g7 b1e1 f7f6 e5g4 c7d7 g4e3 g8h8');
INSERT INTO evaluations VALUES(33597,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(33597,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(33597,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(33597,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(33597,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(33597,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(33597,36,'cp 44','g1f3 a5b6 a2a4 a7a5 h2h3 b8a6 g2g3 b6b2 a1b1 b2a3 f1g2 a3d6 e1g1 e7e6 d3d4 f8e7 d1e2 e8g8 c3e4 f6e4 e2e4 c6c5 f1d1 c5d4 f3d4 a6c5 e4e3 c8d7 d4b5 d6b8 d2c3 d7b5 b1b5 b8c8 c3a5 c5a4 b5b7 a8a5 b7e7');
INSERT INTO evaluations VALUES(40156,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(40156,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(40156,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(40156,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(40156,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(40156,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(40156,36,'cp 119','f1d3 e8g8 g1e2 b8d7 e1g1 a7a6 a2a4 a8b8 e2g3 d7e5 d3e2 e5g6 f2f4 f6e8 c1e3 c8d7 e2d3 b7b5 a4b5 a6b5 b2b4 e7f6 a1c1 c5c4 d3b1 b8a8 g3h5 a8a3 d1d2 f6c3 c1c3 a3c3 d2c3 d8h4 h5g3 e8f6 e3d2 f6h5 g3h5 h4h5 c3g3 h5g4 g3g4 d7g4 h2h3 g4d7 d2c3 f8e8');
INSERT INTO evaluations VALUES(40157,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(40157,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(40157,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(40157,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(40157,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(40157,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(40157,36,'cp 126','e8g8 g1e2 a7a6 a2a4 b8d7 e1g1 a8b8 e2g3 d7e5 d3e2 e5g6 h2h3 f6e8 e2d3 e7f6 g1h2 g6e5 d3e2 f6h4 f2f4 e5g6 f1f3 e8f6 d1c2 f8e8 c1e3 c8d7 a4a5 h7h5 h2h1 b7b5 a5b6 d8b6 a1a6 b6b2 c2b2 b8b2 a6d6 h4g3');
INSERT INTO evaluations VALUES(40271,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(40271,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(40271,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(40271,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(40271,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(40271,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(40271,36,'cp 45','e2e3 e8g8 f1e2 d5c4 e2c4 f6d5 c4d5 c6d5 d1b3 b8c6 b3d5 c6b4 d5d8 f8d8 e1d2 c8g4 a2a3 b4c6 d2e2 e7e5 f4e5 c6e5 d4e5 g7e5 a1d1 d8d1 h1d1 g4f3 g2f3 e5c3 b2c3 a8c8 e2d3 c8c5 d1b1 b7b6 b1b4 c5a5 a3a4 a5f5 d3d4 f5f3');
INSERT INTO evaluations VALUES(46918,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46918,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46918,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46918,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(46918,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46918,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46918,36,'cp 0','c4b3 d6e5 d1h5 e7e6 d4e5 a7a5 a2a4 b8a6 g1h3 a6c5 e1g1 c8d7 c1e3 c5b3 c2b3 g7g6 h5e2 b6d5 e3g5 f8e7 g5h6 e7f8');
INSERT INTO evaluations VALUES(46927,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(46927,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(46927,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(46927,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46927,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46927,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46927,36,'cp 0','a7a5 a2a4 d6e5 d1h5 e7e6 d4e5 b8a6 b1c3 a6c5 h5h3 b6d5 g1e2 d5c3 e2c3 d8d4 h3g3 c5e4 c3e4 d4e4 c1e3 e4b4 e1e2 b7b6 h1d1 c8a6 e2f3 a6b7 f3e2');
INSERT INTO evaluations VALUES(55004,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(55004,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(55004,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(55004,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(55004,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(55004,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(55004,36,'cp 28','c7c5 e2e4 g8f6 g1e2 e7e5 b1c3 f8e7 h2h3 b8c6 e1g1 e8g8 a2a4 a8b8 d2d3 a7a6 f2f4 b7b6 f4f5 c6b4 g3g4 h7h6 c3d5 b4d5 e4d5 f6h7 e2g3 e7h4 g3e4 h7f6 e4f6 d8f6 g1h2 c8d7 g2e4 b8a8 c1e3 b6b5 a4b5 a6b5 c2c3 a8a1');
INSERT INTO evaluations VALUES(53106,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(53106,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(53106,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(53106,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(53106,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(53106,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(53106,36,'cp 32','e2e4 g8f6 g1e2 e7e5 e1g1 f8e7 d2d3 b8c6 f2f4 a7a6 b1c3 e8g8 h2h3 c6d4 g3g4 d4e2 d1e2 e5f4 c1f4 f6d7 c3d5 d7e5 d5e7 d8e7 a2a4 c8d7 a4a5 a8e8 e2d2 d7c6 f4g3 f7f6 b2b4 e5f7 d2c3 g8h8 a1e1 h7h6 g1h2 c6b5');
INSERT INTO evaluations VALUES(55808,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55808,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(55808,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55808,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55808,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55808,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55808,36,'cp 0','d5c4 d1a4 b8d7 f1g2 a7a6 a4c4 b7b5 c4h4 f8e7 h4e4 a8b8 f3d4 c8b7 d4c6 g8f6 e4c2 b7c6 c2c6 e8g8 e1g1 d7e5 c6c3 e5d7');
INSERT INTO evaluations VALUES(2838,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(2838,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(2838,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(2838,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(2838,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(2838,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(2838,36,'cp 26','e5e1 d8h4 g2g3 h4h3 e1e4 g7g5 d1f1 h3h5 b1d2 c8h3 f1e2 h5g6 d2f3 f7f6 b3c2 h3f5 c1d2 f5e4 c2e4 g6f7 e2d1 a8e8 e4f5 d5e7 f5c2 g8h8 a2a4 e7d5 b2b3 d5e7 h2h3 e7d5 c2f5 d5e7 f5d3 f7h5 g1g2 g5g4 h3g4 h5g4 f3h2 g4d1 a1d1 f6f5 h2f3 b5a4 b3a4');
INSERT INTO evaluations VALUES(2837,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2837,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(2837,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(2837,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(2837,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(2837,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(2837,36,'cp 18','d8h4 g2g3 h4h3 e1e4 g7g5 d1f1 h3h5 b1d2 c8h3 f1e2 h5g6 d2f3 f7f6 b3c2 h3f5 c1d2 g8h8 a2a4 f5e4 c2e4 f8e8 e4g6 e8e2 g6f5 d5b6 h2h4 g5g4 f5g4 b5a4 g1f1 a8e8 h4h5 b6c4 d2c1 a4a3 b2b3 c4a5 f3e5 f6e5 f1e2 e5d4 e2d3 a5b3 c1a3 b3a1 a3d6');
INSERT INTO evaluations VALUES(2185,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(2185,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(2185,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(2185,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(2185,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(2185,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(2185,36,'cp 10','g2g3 h4h3 e1e4 g7g5 d1f1 h3h5 b1d2 c8h3 f1d1 h5g6 a2a4 b5b4 c3b4 d5b4 d1f3 a8e8 g3g4 d6f4 f3h3 e8e4 d2e4 g6e4 c1f4 e4f4 a1e1 f4d4 e1d1 d4b2 g1g2 b2f6 h3f3 f6f3 g2f3 f8b8 d1d7 b4d5 b3d5 c6d5 d7d5 h7h6 d5d6 b8b3 f3e2 b3a3 d6a6 g8g7 a4a5');
INSERT INTO evaluations VALUES(2184,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(2184,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2184,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(2184,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(2184,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(2184,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(2184,36,'cp 24','h4h3 e1e4 g7g5 d1f1 h3h5 b1d2 c8h3 f1e2 h5g6 d2f3 f7f6 c1d2 h3f5 b3c2 f5e4 c2e4 g6f7 e2d1 a8e8 e4f5 d5e7 f5c2 g8g7 a2a4 e7d5 c2f5 d5e7 f5b1 e7d5 a4b5 a6b5 b1f5 e8a8 a1c1 h7h5 c3c4 b5c4 c1c4 d5e7');
INSERT INTO evaluations VALUES(2188,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(2188,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(2188,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(2188,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(2188,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(2188,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(2188,36,'cp 18','e1e4 g7g5 d1f1 h3h5 b1d2 c8h3 f1d1 h5g6 a2a4 b5b4 c3b4 d5b4 d1f3 a8e8 g3g4 d6f4 f3h3 e8e4 d2e4 g6e4 c1f4 e4f4 a1e1 f4d4 h3f3 d4b2 e1d1 h7h6 h2h3 g8g7 g1g2 b2f6 d1d7 f6f3 g2f3 c6c5 f3e4 b4c6 e4d5 c6a5 b3c2 c5c4 d5d4 a5c6 d4c5 c6e5 d7a7 c4c3');
INSERT INTO evaluations VALUES(6838,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(6838,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(6838,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(6838,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(6838,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(6838,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(6838,36,'cp 192','g5f7 c5b4 c2c3 d4c3 b2c3 e5f7 c3b4 g8f6 e1g1 d7d6 c1b2 c7c6 b1d2 e8g8 a2a4 g8h8 a1a3 d6d5 e4d5 c6d5 c4a2 c8f5 a3e3 a8c8 d1a1 f7g5 a2b3 d8d7 f1e1 h7h6 d2f3 g5f3 e3f3 a7a6 h2h3 f5e4 b2f6 h8h7 f3f4 g7f6 f4f6 f8f6 a1f6 d7g7 f6g7 h7g7');
INSERT INTO evaluations VALUES(6157,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(6157,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(6157,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(6157,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(6157,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(6157,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(6157,36,'cp 202','c5b4 c2c3 d4c3 b2c3 e5f7 c3b4 g8f6 e1g1 d7d6 c1b2 e8g8 b1d2 c7c6 a2a4 d8e7 f1e1 c8e6 c4f1 a8d8 a1a3 d8e8 d1e2 e7c7 a4a5 f7e5 f2f4 e5g6 g2g3 b7b5 f1g2 d6d5 b2f6 g7f6 a3d3 d5e4 d2e4 e6d5 e2c2 g8h8');
INSERT INTO evaluations VALUES(15019,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(15019,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(15019,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15019,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(15019,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15019,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15019,36,'cp 0','f8e7 f1b5 e8g8 b5c6 b7c6 e1g1 h7h6 g5h4 c8g4 h2h3 g4f3 d1f3 f6h7 h4e7 d8e7 f3d3 f8e8 d3a6 e7d6 a1e1 h7g5 e1e8 a8e8 h3h4 g5f3 g2f3 d6g6 g1h2 g6h5 c3e2 e8e2 a6e2 h5h4 h2g1 h4g5 g1h1 g5h4');
INSERT INTO evaluations VALUES(8516,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(8516,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(8516,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(8516,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(8516,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(8516,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(8516,36,'cp 12','f1b5 e8g8 b5c6 b7c6 e1g1 h7h6 g5h4 c8g4 h2h3 g4f3 d1f3 f6h7 h4e7 d8e7 f3d3 f8e8 c3a4 h7g5 a4c5 g5e4 c5a6 c6c5 d4c5 e4c5 a6c5 e7c5 b2b3 c5a5 f1d1 c7c6 d3d2 a5d2 d1d2 a7a5 a1c1 a5a4 c2c4 d5c4 c1c4 a4b3 a2b3 e8e6 d2d7 a8a1 g1h2');
INSERT INTO evaluations VALUES(8509,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(8509,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(8509,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(8509,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(8509,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(8509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8509,36,'cp 8','e8g8 b5c6 b7c6 e1g1 h7h6 g5f6 e7f6 d1d3 d8d6 f1e1 c6c5 d4c5 d6c5 c3d5 f6b2 d5e7 g8h8 a1b1 b2f6 b1b5 c5c3 e7c8 c3d3 c2d3 a8c8 g2g3 f8d8 e1e3 d8d6 b5a5 a7a6 a5c5 c8d8 d3d4 f6d4 f3d4 d6d4 c5c7 d4d7 c7c4 d7d6 g1f1 h8g8 c4a4 d6c6');
INSERT INTO evaluations VALUES(16329,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(16329,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(16329,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(16329,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(16329,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(16329,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(16329,36,'cp -19','c6b4 f1c4 b4d5 e4d5 f8b4 c1d2 d8e7 d1e2 e8g8 e2e7 b4e7 e1g1 e7c5 b2b4 c5b6 a2a4 a7a5 b4a5 b6a5 f1b1 a5d2 f3d2 d7d6 a4a5 f8e8 d2f3 e8e4 b1d1 c8g4 d1b1 a8a7 a5a6 b7b6 f3d2 e4e5 d2b3 g4d7 b3d4 f6d5 d4b5 d7b5 b1b5 d5f6 b5e5');
INSERT INTO evaluations VALUES(20599,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(20599,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(20599,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(20599,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(20599,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(20599,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(20599,36,'cp 82','f3g5 g8h6 d1h5 g7g6 h5d1 f8e7 g5f3 h6g4 e1g1 g4f6 c1h6 c6a5 b1c3 f6g4 c4b5 c8d7 h6d2 d7b5 c3b5 a5c6 h2h3 g4f6 c2c4 e8g8 b5c3 a7a6 a2a4 d8d7 f3h2 c6d4 a4a5 d4e6 h2f3 g8g7 c3d5 f6d5 c4d5 e6d4');
INSERT INTO evaluations VALUES(20590,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(20590,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(20590,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(20590,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(20590,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(20590,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(20590,36,'cp 78','g8h6 d1h5 g7g6 h5d1 f8e7 g5f3 h6g4 e1g1 g4f6 c1h6 c6a5 c4b5 a5c6 a2a3 f6g4 h6d2 e8g8 b5c4 f8e8 b1c3 c8e6 c3d5 g4f6 d5e7 d8e7 c4e6 e7e6 f3g5 e6e7 f2f4 e5f4 f1f4 c6e5 d1f1 f6h5 f4f2 h7h6 g5f3 e5f3 f2f3');
INSERT INTO evaluations VALUES(13898,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(13898,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(13898,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(13898,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(13898,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(13898,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(13898,36,'cp 78','d1h5 g7g6 h5d1 f8e7 g5f3 h6g4 e1g1 g4f6 c1h6 c6a5 c4b5 a5c6 b1c3 e7f8 d1d2 f8h6 d2h6 f6g8 h6e3 g8f6 h2h3 c8d7 b5c6 d7c6 a2a3 a7a5 e3h6 f6h5 a3a4 d8f6 f3d2 f6f4 h6f4 h5f4 d2c4 e8e7');
INSERT INTO evaluations VALUES(34753,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(34753,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(34753,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(34753,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(34753,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(34753,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(34753,36,'cp 109','b1c3 c8e6 c3d5 f6d8 d3d4 b8d7 e1g1 g8e7 f1e1 e7g6 c4d3 e6g4 d3e2 d7f6 d5e3 g4c8 b2b3 f8e7 c1b2 e8g8 e2d3 f6d7 d1d2 e5d4 f3d4 e7f6 c2c4 d7c5 d3c2 c5e6 a1d1 e6d4 b2d4 f6d4');
INSERT INTO evaluations VALUES(34750,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(34750,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(34750,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(34750,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(34750,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(34750,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(34750,36,'cp 119','c8e6 c3d5 f6d8 e1g1 g8f6 f1e1 b8d7 d3d4 e6g4 d5e3 g4f3 d1f3 e5d4 e3f5 d7e5 f3b3 g7g6 f5d4 f6d7 c4f1 d7c5 b3g3 f8g7 b2b4 c5d7 c2c3 e8g8 f2f4 e5c6 c1e3 f8e8 g3f3 d8e7 e3f2 c6d4 c3d4 d7f6 a2a4 g8h8');
INSERT INTO evaluations VALUES(38651,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(38651,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(38651,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(38651,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(38651,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(38651,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(38651,36,'cp 87','g1f3 g8f6 g2g3 e8g8 f1g2 c8g4 e1g1 h7h6 d2d3 c7c5 c1e3 f8e8 d1d2 b8c6 b1c3 a8c8 g1h1 d6f8 e3g1 c6d4 a1e1 d4f3 e2f3 e8e1 f1e1 g4e6 a2a3 a7a6 c3e2 d8b6 c2c4 c8d8 g1f2 e6d7 e2c3 d7c6');
INSERT INTO evaluations VALUES(38649,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(38649,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(38649,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(38649,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(38649,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(38649,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(38649,36,'cp 57','c8g4 e2e3 g8f6 f1e2 c7c5 b1c3 e8g8 e1g1 b8c6 d2d3 g4e6 g1h1 h7h6 e3e4 f6g4 h2h3 g4e5 c1e3 d8e7 c3b5 d6b8 d1d2 e5f3 e2f3 b8e5 a1d1 a8d8 e3f4 f7f6 b5c3 c6d4 f3h5 e7c7 c3e2 d4e2 h5e2 e6a2 f4e5 c7e5 b2b3 b7b5');
INSERT INTO evaluations VALUES(28140,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(28140,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(28140,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(28140,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(28140,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(28140,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(28140,36,'cp 70','e2e3 g8f6 f1e2 c7c5 b1c3 b8c6 e1g1 h7h6 g1h1 e8g8 d2d3 g4e6 e3e4 f6g4 h2h3 g4e5 c1e3 d8e7 d1d2 a8d8 a1e1 e5f3 e2f3 c6d4 c3e2 d4f3 f1f3 f7f5 e4f5 e6d5 f3f2 e7e5 e3f4 e5f5 h1g1 f5g6 e1f1 d6e7 d2e3 f8e8 e3g3 g6g3 f4g3 d5a2 e2c3 a2e6');
INSERT INTO evaluations VALUES(43073,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(43073,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(43073,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(43073,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(43073,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(43073,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(43073,36,'cp 113','h2h4 h7h6 c1f4 d7d6 d1d2 b8c6 e1c1 e7f5 e2e3 h6h5 f1e2 e8g8 d4d5 e6d5 c4d5 c6e5 e3e4 e5f3 e2f3 f5d4 f4g5 d8e8 g5h6 g7e5 h6f8 g8f8 d2h6 e5g7 h6e3 e8e5 d1d2 b7b5 f3e2 b5b4 f2f4 e5f6 c3b5 d4b5 e2b5 c8g4 b5e2');
INSERT INTO evaluations VALUES(44538,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(44538,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(44538,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(44538,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(44538,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(44538,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(44538,36,'cp -15','c5d4 d1d4 b8c6 d4d3 g7g6 e2e4 f8g7 g1f3 d7d6 f1e2 c8g4 c1e3 g4f3 g2f3 f6d7 a1d1 d7c5 d3d2 d8a5 d1c1 e8g8 e1g1 c5e6 f3f4 e6d4 g1h1 g8h8 b2b3 f7f5 e2d3 f5e4 d3e4 a5h5 f2f3 d4f5 e3f2 c6d4 f2d4 f5d4 e4b7 a8b8 b7d5 h5h4');
INSERT INTO evaluations VALUES(44609,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(44609,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(44609,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(44609,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(44609,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(44609,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(44609,36,'cp -2','d1d4 b8c6 d4d3 g7g6 e2e4 d7d6 g1f3 f8g7 f1e2 c8g4 e1g1 g4f3 d3f3 c6d4 f3d3 f6d7 a1b1 d4e2 c3e2 d8a5 d3c2 a8c8 a2a4 a7a6 c1d2 a5c7 f1c1 e8g8 a4a5 f8e8 b2b3 e7e6 h2h3 d6d5 e4d5 e6d5 e2f4 d7f6 c2d3 d5c4 b3c4 f6e4');
INSERT INTO evaluations VALUES(44608,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(44608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44608,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(44608,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(44608,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(44608,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(44608,36,'cp -13','b8c6 d4d3 e7e6 g1f3 d7d5 e2e3 f8e7 a2a3 e8g8 c4d5 f6d5 c3d5 e6d5 c1d2 e7f6 d2c3 c8g4 f1e2 g4f3 e2f3 f6c3 d3c3 d5d4 c3c5 d4e3 f2e3 d8h4 g2g3 h4h6 e1c1 a8e8 f3c6 b7c6 d1d6 h6e3 c5e3 e8e3 d6c6');
INSERT INTO evaluations VALUES(12873,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(12873,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(12873,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(12873,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(12873,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(12873,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(12873,36,'cp -9','d4d3 e7e6 e2e3 f8e7 g1f3 d7d5 f1e2 d5c4 d3c4 e8g8 e1g1 a7a6 f1d1 d8c7 a2a4 c8d7 c1d2 a8c8 c3e4 f6e4 c4e4 e7f6 d2c3 f6c3 b2c3 h7h6 e2d3 f7f5 e4c4 c6e5 f3e5 c7e5 c4d4 e5d4 c3d4 c8c3 a4a5 f8c8 e3e4 f5e4 d3e4 d7c6 f2f3 c6d5 e4d5 e6d5');
INSERT INTO evaluations VALUES(50036,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(50036,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(50036,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(50036,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(50036,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(50036,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(50036,36,'cp 183','c7c6 d2d4 d7d5 e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 b1c3 d8d6 c1e3 d6c6 e3d4 e5c4 b2b3 c4d6 f1e1 e8f7 a1c1 d6e4 d1d3 e4c3 d4c3 c6d7 c1d1 a8c8 g2g3 e7c6 d3d5 d7d5 d1d5 c8c7 e1d1 f8e7 c3d2 h8d8 d5d8 c6d8 c2c4 d8c6');
INSERT INTO evaluations VALUES(44387,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(44387,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(44387,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(44387,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(44387,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(44387,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(44387,36,'cp 189','d2d4 d7d5 e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 b1c3 d8d7 f3e5 f6e5 f1e1 e8c8 e1e5 d5d4 c3e2 e7c6 e5e4 f8c5 e2f4 d8e8 e4e8 h8e8 f4d3 c5d6 c1g5 g7g6 g5f4 d6f4 d3f4 g6g5 f4d3 g5g4 d1d2 d7f5 g1f1 h7h5 b2b3 a7a5 a1d1 e8e3 d1e1 e3e1 d3e1 f5c5 h2h3 g4h3 g2h3 c5d5 d2d3 c6b4');
INSERT INTO evaluations VALUES(50497,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(50497,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(50497,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(50497,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(50497,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(50497,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(50497,36,'cp -33','d7d5 e4d5 e6d5 c4a2 b8c6 g1e2 g8f6 b1c3 c8e6 e2f4 f8e7 f4e6 f7e6 e1g1 e8g8 c3e2 e7d6 e2f4 d8d7 c2c3 d7f7 g2g3 a8e8 f4g2 g8h8 c1g5 b7b5 f2f4 a6a5 d1f3 c5c4 g5f6 f7f6 g1h1 c6e7 d3c4 b5c4 a1e1 f6h6 a2b1');
INSERT INTO evaluations VALUES(50777,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(50777,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(50777,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(50777,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(50777,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(50777,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(50777,36,'cp -17','e4d5 e6d5 c4a2 g8f6 g1e2 b8c6 b1c3 c8e6 e2f4 f8e7 f4e6 f7e6 e1g1 e8g8 c3e2 g8h8 e2f4 d8d6 c2c3 e6e5 f4h5 a8d8 h5f6 e7f6 c1e3 d5d4 c3d4 c6d4 a1c1 b7b6 d1g4 a6a5 h2h3 g7g6 c1d1 a5a4 e3d4 d6d4 g4d4');
INSERT INTO evaluations VALUES(50770,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(50770,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(50770,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(50770,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(50770,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(50770,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(50770,36,'cp -27','e6d5 c4a2 g8f6 b1c3 c8e6 g1e2 b8c6 e2f4 f8e7 e1g1 e8g8 f4e6 f7e6 c3e2 e7d6 e2f4 d8d7 c2c3 d7f7 g2g3 b7b5 f4g2 a6a5 c1g5 b5b4 d1e2 a5a4 a1e1 a8e8 f2f4 h7h6 g5h4 b4c3 b2c3 g8h8 h4f6 f7f6');
INSERT INTO evaluations VALUES(50801,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(50801,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(50801,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(50801,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(50801,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(50801,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(50801,36,'cp -12','c4a2 b8c6 g1e2 g8f6 b1c3 f8e7 c1g5 c8e6 e2f4 d8d6 g5f6 d6f4 f6e7 c6e7 d1d2 f4e5 d2e3 e5d6 e1g1 e8g8 c3e2 a8e8 e3d2 e7g6 a1e1 b7b5 b2b4 e8e7 e2g3 f8e8 a2b3 h7h6 b4c5 d6c5');
INSERT INTO evaluations VALUES(56726,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(56726,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(56726,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(56726,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(56726,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(56726,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(56726,36,'cp 111','c7c6 f1c4 g8f6 f3g5 d8e7 g5f7 h8g8 f7g5 g8h8 c4f7 e8d8 f7b3 d8c7 g5f7 h8g8 e1g1 c8e6 f7g5 e6b3 c2b3 h7h6 g5f3 b8d7 b3b4 a8c8 a2a4 c7b8 b4b5 e7e6 c1e3 f8e7 f1e1 f6g4 b5c6 g4e3 e1e3 c8c6 c3d5');
INSERT INTO evaluations VALUES(54655,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(54655,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(54655,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(54655,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(54655,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(54655,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(54655,36,'cp 138','f1c4 d8e7 e1g1 g8f6 b2b4 c8e6 c4b3 a7a6 d1d3 b8d7 c1d2 g7g6 f1d1 e6b3 c2b3 e7e6 f3g5 e6g8 d3h3 h7h6 g5e6 e8e7 e6f8 a8f8 h3d3 g6g5 c3e2 g8e6 e2g3 d6d5 e4d5 c6d5 a1c1 e5e4 d3c2 e7f7 b4b5');
INSERT INTO evaluations VALUES(6411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6411,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(6411,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(6411,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(6411,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(6411,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(6411,36,'cp 0','c5e7 g5e3 g8f6 c2c4 e7b4 b1c3 f6g4 d1c2 d7d6 f1e2 e8g8 e1g1 b4c5 c3d5 g4e3 f2e3 a7a5 c2d2 c8g4 h2h3 g4f3 f1f3 c5b4 d2c2 b4c5');
INSERT INTO evaluations VALUES(13581,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(13581,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(13581,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(13581,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13581,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(13581,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(13581,36,'cp -9','f8d6 f3e5 d8c7 b1d2 b7b6 e1g1 c8b7 d1e2 e8g8 g2g4 c6e7 e2g2 a7a5 f1f2 c5d4 e3d4 b6b5 g4g5 f6e4 d2e4 d5e4 d3e4 b7e4 g2e4 b5b4 c1e3 b4c3 a1c1 e7d5 b2c3 d6e5 d4e5 d5e3 e4e3 c7c4 c1d1 a8c8 d1d4 c4c5 g1g2 c5c3 e3c3 c8c3 f2b2 c3c1 b2b7 c1c2 g2h3 c2a2');
INSERT INTO evaluations VALUES(8391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8391,36,'cp 0','f3e5 d8c7 e1g1 b7b6 b1d2 c8b7 d1e2 a7a5 g2g4 f6d7 d2f3 d7f6');
INSERT INTO evaluations VALUES(14333,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(14333,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14333,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14333,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(14333,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14333,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14333,36,'cp 43','g1f3 c8g4 d2d4 e7e6 f1e2 f8b4 c2c3 b4d6 c3c4 d6b4 b1c3 f6e4 d1c2 b4c3 b2c3 e8g8 e1g1 d8e7 c3b2 a7a5 a2a3 f7f5 a1e1 h7h6 h2h3 g4h5 f3d2 h5e2 e1e2 e4d2 e2d2 g8h8 f2f3 f8e8 f1f2 a5a4 b3a4');
INSERT INTO evaluations VALUES(39200,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(39200,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(39200,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(39200,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(39200,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(39200,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(39200,36,'cp -18','d7d5 c4d5 g8f6 d2d4 e5d4 c3d4 c5b4 c1d2 b4d2 d1d2 c7c6 d5b3 f6e4 d2f4 e8g8 g1f3 e4f6 e1g1 b8a6 f1e1 a6c7 b1c3 c8e6 b3c2 e6g4 f3g5 c7e6 g5e6 g4e6 a1d1 d8d7 e1e5 a8d8 h2h3 f8e8 a2a3 h7h6 f4f3 d7c7 c3e4 f6e4 f3e4 g7g6 d1e1 g8g7');
INSERT INTO evaluations VALUES(39293,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(39293,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(39293,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(39293,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(39293,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(39293,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(39293,36,'cp 113','c4f7 e8f7 d1h5 g7g6 h5c5 g8f6 g1e2 b8c6 e2d4 f6e4 c5d5 f7g7 e1g1 e4f6 d5c4 h8e8 d4c6 d7c6 f2f3 c8f5 b1c3 f5c2 c4h4 h7h5 h2h3 d8d7 c1f4 a8d8 h4g3 d7d4 g1h2 h5h4 g3h4 f6h5 f4g5 d4h4 g5h4 d8d4 h4g5 h5f4 g5f4 d4f4 a1e1 e8e1 f1e1');
INSERT INTO evaluations VALUES(38973,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(38973,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(38973,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(38973,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(38973,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(38973,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(38973,36,'cp 96','e8f7 d1h5 g7g6 h5c5 g8f6 g1f3 b8c6 f3d4 f6e4 c5d5 f7g7 e1g1 e4f6 d5c4 h8e8 b1c3 c6e5 c4b3 e5f7 c1f4 c7c6 a1e1 d8b6 e1e8 b6b3 a2b3 f6e8 f2f3 d7d6 c3e4 c6c5 d4e2 c8f5 e2g3 f5e4 g3e4 b7b6 f1d1 a8d8 e4c3 d8d7 f4g3 e8f6 d1a1 f6h5 g3f2 h5f4');
INSERT INTO evaluations VALUES(37488,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(37488,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(37488,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(37488,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(37488,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(37488,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(37488,36,'cp 128','d1h5 g7g6 h5c5 g8f6 g1e2 b8c6 e2d4 f6e4 c5d5 f7g7 e1g1 e4f6 d5c4 h8e8 b1c3 c6e5 c4a4 e5f7 b2b4 b7b6 c1b2 g7g8 a4b3 c8b7 f2f3 a7a5 b4a5 a8a5 c3e4 f6e4 f3e4 d8e7 a1e1 e8f8 f1f2 a5e5 d4f3 e5e4 e1e4 b7e4');
INSERT INTO evaluations VALUES(37487,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(37487,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(37487,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(37487,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(37487,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(37487,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(37487,36,'cp 122','g7g6 h5c5 g8f6 g1e2 b8c6 e2d4 f6e4 c5d5 f7g7 e1g1 e4f6 d5c4 h8e8 b1c3 c6e5 c4b3 e5f7 c1f4 c7c6 a1e1 d8b6 e1e8 b6b3 a2b3 f6e8 h2h4 e8f6 f1e1 h7h6 f4g3 d7d6 c3e4 f6e4 e1e4 g7f8 d4e6 f8g8 e6d8 c8f5 e4e8 g8h7 e8e7 a8d8 e7f7 h7g8 f7b7 f5c2 b3b4 c2f5 b7a7 d8d7 a7d7 f5d7 g3d6 g8f7 f2f3');
INSERT INTO evaluations VALUES(37484,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(37484,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(37484,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(37484,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(37484,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(37484,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(37484,36,'cp 100','h5c5 g8f6 g1e2 b8c6 e2d4 f6e4 c5d5 f7g7 e1g1 e4f6 d5c4 h8e8 b1c3 c6e5 c4b3 e5f7 c1f4 c7c6 a1e1 d8b6 e1e8 b6b3 a2b3 f6e8 h2h4 h7h6 f2f3 e8f6 c3a4 g6g5 d4f5 g7g6 f5e7 g6g7 f4e3 d7d6 h4g5 h6g5 e7c8 a8c8 e3a7 c8a8 a7d4 g7g6 g1f2 a8e8 d4f6 g6f6');
INSERT INTO evaluations VALUES(37481,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(37481,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(37481,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(37481,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(37481,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(37481,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(37481,36,'cp 119','g8f6 g1f3 b8c6 f3d4 f6e4 c5d5 f7g7 e1g1 e4f6 d5c4 h8e8 b1c3 c6e5 c4b3 e5f7 c1f4 c7c6 a1e1 d8b6 e1e8 f6e8 b3a4 d7d5 f4e3 b6d8 f1e1 g7g8 d4f3 e8g7 c3e2 d8d6 e3f4 d6c5 h2h3 c8d7 a4b3 c5b6 b3c3 a8e8');
INSERT INTO evaluations VALUES(46556,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(46556,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(46556,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(46556,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(46556,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(46556,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(46556,36,'cp 150','g1f3 b8c6 f1d3 c6b4 d3e2 b4c6 a2a3 h7h6 e4e5 b7b6 e1g1 e7f5 e2b5 c8d7 f1e1 a7a6 b5f1 a6a5 c3e2 a5a4 e2f4 f8e7 a1b1 c6a5 f1d3 g7g6 b2b4 a4b3 c2b3 c7c5 d3f5 g6f5 f4h5 c5c4 b3c4 d5c4');
INSERT INTO evaluations VALUES(46555,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(46555,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(46555,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(46555,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(46555,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(46555,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(46555,36,'cp 119','b8c6 a2a3 h7h6 e4e5 b7b6 h2h4 c6b8 c3e2 c7c5 h4h5 b8c6 c2c3 c8a6 e2f4 a6f1 e1f1 c6a5 h1h3 c5d4 c3d4 d8d7 f4e2 e8c8 b2b3 c8b8 c1d2 e7c6 h3g3 d8c8 d2c3 h8g8 a1c1 f8e7 c3d2 d7b7 e2f4 b8a8 d2a5 b6a5');
INSERT INTO evaluations VALUES(14043,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(14043,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(14043,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(14043,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(14043,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(14043,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(14043,36,'cp 114','f1d3 h7h6 e4e5 b7b6 a2a3 c6b8 c3e2 c8a6 d3a6 b8a6 e1g1 c7c5 c2c3 a6c7 e2f4 c5c4 f4h5 e7c6 c1e3 d8d7 a3a4 c6a5 f3d2 e8c8 h2h3 d7e8 f1e1 g7g6 h5f6 e8c6 f6g4 h6h5 g4f6 f8e7 d2f3');
INSERT INTO evaluations VALUES(49665,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(49665,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(49665,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(49665,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(49665,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(49665,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(49665,36,'cp -63','b1c3 f8c5 f1c4 c7c6 d1f3 b7b5 c4b3 a7a5 a2a3 b8a6 e1g1 d7d6 d2d3 e8g8 f3g3 a5a4 b3a2 f6h5 g3g5 d8g5 h3g5 h7h6 g5f3 h5f6 c1e3 c5e3 f2e3 a6c5 h2h3 f8d8 f3h4 d8e8 h4f5 c8f5 f1f5 a8b8 f5f1 b8b7 a1b1 b7c7');
INSERT INTO evaluations VALUES(49617,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(49617,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(49617,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(49617,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(49617,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(49617,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(49617,36,'cp -71','b8c6 h3g5 f8c5 f1c4 e8g8 d2d3 c6a5 c4b3 a5b3 a2b3 h7h6 g5f3 d7d5 e4d5 f6d5 e1g1 f8e8 c1d2 d5b4 c3e4 c5f8 d2b4 f8b4 c2c3 b4f8 b3b4 c8f5 f1e1 f5h7 e4g3 d8d5 h2h3 h7g6 g3e4 a7a5 b4a5 a8a5 a1a5');
INSERT INTO evaluations VALUES(55072,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(55072,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(55072,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(55072,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(55072,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(55072,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(55072,36,'cp 178','e4d5 d8d5 c2c4 d5d8 g1f3 g8f6 b1c3 c7c6 c1e3 c8f5 f3h4 f5c8 h2h3 e7e6 h4f3 f8b4 d1b3 a7a5 f1d3 e8g8 e1g1 b8d7 a1d1 b7b5 c4b5 c6b5 f3e5 c8b7 d3b5 d7b6 f1e1 a8c8 a2a3 b4e7 b5d3 e7d6 e3g5');
INSERT INTO evaluations VALUES(55016,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(55016,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(55016,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(55016,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(55016,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(55016,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(55016,36,'cp 164','d8d5 c2c4 d5d8 g1f3 g8f6 b1c3 e7e6 f1d3 g7g6 c1e3 f8g7 d1e2 e8g8 e1c1 c7c5 d4c5 b8d7 h2h4 d8a5 h4h5 f6h5 h1h5 g6h5 d3c2 d7f6 f3e5 f8d8 d1h1 a5c7 e3g5 c8d7 f2f4 d7e8 g5f6 g7f6 e2h5 g8f8 h5h7 c7c5');
INSERT INTO evaluations VALUES(34590,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(34590,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(34590,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(34590,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(34590,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(34590,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(34590,36,'cp 169','c2c4 d5d8 g1f3 g8f6 b1c3 e7e6 f1d3 g7g6 d1e2 f8g7 c1f4 e8g8 a1d1 b8d7 e1g1 b7b6 c3b5 c7c6 b5d6 f6h5 f4g5 d7f6 d6c8 d8c8 d3c2 f6d7 f1e1 h5f6 h2h4 b6b5 c4c5 a7a5 g5f4 f6d5 f4d6 f8d8 a2a4 d7f6 a4b5 c6b5 c2d3 f6e8 d6e5 e8c7 e5g7 g8g7 d3b5 c7b5');
INSERT INTO evaluations VALUES(2680,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(2680,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(2680,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(2680,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(2680,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(2680,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(2680,36,'cp 62','d2d4 c5a7 c1g5 e5d4 e4e5 h7h6 g5f6 g7f6 c3d4 f8e8 b1d2 f6e5 a4c6 d7c6 d4e5 c8e6 f1e1 f7f5 d1a4 g8h7 d2f1 e8g8 a1d1 d8e7 f1e3 e7f7 a4c2 a7e3 e1e3 f7g6 g2g3 f5f4 c2g6 g8g6 e3c3 g6g7 d1d4 f4g3 h2g3 a6a5 f3d2 a8g8');
INSERT INTO evaluations VALUES(2676,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(2676,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(2676,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(2676,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(2676,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(2676,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2676,36,'cp 25','c5a7 f1e1 b7b5 a4c2 d7d6 h2h3 c8b7 c1e3 e5d4 c3d4 c6b4 d4d5 a7e3 e1e3 c7c6 d5c6 b4c6 b1c3 c6e5 c2b3 a8c8 a2a3 f8e8 d1e2 d8b6 f3e5 d6e5 e3d3 c8d8 a1d1 h7h6 e2d2 d8d3 d2d3 e8c8 d3d6 b6d6 d1d6 f6e4 c3e4 b7e4 d6a6 c8d8 a6b6 d8d2 b6b8 g8h7 b3f7 d2b2 b8d8');
INSERT INTO evaluations VALUES(4535,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(4535,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(4535,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(4535,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(4535,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(4535,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(4535,36,'cp 90','f3e5 d7d6 c3a4 c5b6 a4b6 a7b6 e5f7 f8f7 c4f7 g8f7 c2c4 f7g8 c1e3 d4e6 f2f4 c7c5 b2b4 c8d7 a2a4 d7e8 b4b5 e6d4 h2h3 e8h5 g2g4 h5f7 e3d4 c5d4 a1a2 f6e4 d3e4 f7c4 a2d2 c4f1 g1f1 d8f6 f4f5 f6e5 d2d4 e5g3 d4d3 g3h2 d1f3 a8a4');
INSERT INTO evaluations VALUES(4527,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(4527,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(4527,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(4527,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(4527,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(4527,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(4527,36,'cp 106','d7d6 c3a4 c5b6 e5f7 f8f7 a4b6 a7b6 c4f7 g8f7 c2c4 d8e7 c1e3 d4c6 f2f3 f7g8 d1d2 h7h6 a2a4 c6a5 d2c3 a5c6 b2b4 c6e5 a4a5 b6a5 a1a5 a8a5 b4a5 c7c5 f1b1 c8e6 c3d2 e5c6 b1b5 e6f7 h2h3 f6h5 g1h2 h5f6 e3f2');
INSERT INTO evaluations VALUES(3916,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(3916,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(3916,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(3916,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(3916,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(3916,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(3916,36,'cp 104','c3a4 c5b6 e5f7 f8f7 c4f7 g8f7 h2h3 c8d7 a4b6 a7b6 f2f4 f7g8 c1e3 c7c5 c2c4 d4c6 e3d2 d7e8 d2c3 f6d7 a2a4 c6d4 c3d4 c5d4 b2b4 h7h6 d1f3 d7f8 e4e5 e8c6 f3f2 d6e5 f4e5 d8g5 b4b5 c6d7 f2f7 g8h7 e5e6 g5e3 g1h1 d7e6 f7b7 a8e8');
INSERT INTO evaluations VALUES(11404,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(11404,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(11404,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(11404,33,'cp 247',NULL);
INSERT INTO evaluations VALUES(11404,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(11404,35,'cp 279',NULL);
INSERT INTO evaluations VALUES(11404,36,'cp 286','b5c6 b7c6 d4e5 f6g4 c1f4 c8e6 e5d6 f8d6 d1d2 d6f4 d2f4 d8b8 a1b1 e8g8 e1g1 g4f6 f1d1 f8e8 h2h3 h7h6 f3d4 b8b6 b2b3 c6c5 d4e6 b6e6 f2f3 e6e7 f4e3 a8d8 b1c1 c5c4 b3c4 e7b4 c3d5 b4c4 d5f6 g7f6 e3h6');
INSERT INTO evaluations VALUES(11356,30,'cp 267',NULL);
INSERT INTO evaluations VALUES(11356,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(11356,32,'cp 269',NULL);
INSERT INTO evaluations VALUES(11356,33,'cp 281',NULL);
INSERT INTO evaluations VALUES(11356,34,'cp 273',NULL);
INSERT INTO evaluations VALUES(11356,35,'cp 269',NULL);
INSERT INTO evaluations VALUES(11356,36,'cp 271','b7c6 d4e5 f6g4 c1f4 c8e6 e5d6 f8d6 d1d2 e8g8 h2h3 d6f4 d2f4 d8b8 a1b1 g4f6 e1g1 h7h6 f1d1 f8e8 d1d3 c6c5 b2b3 a6a5 b1d1 c5c4 b3c4 e6c4 d3e3 f6h7 e4e5 c4e6 f3d4 h7f8 e3g3 f8g6 f4e3 b8b6 f2f4 a8d8 g1h2 c7c5 d4e6');
INSERT INTO evaluations VALUES(11663,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(11663,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(11663,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(11663,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(11663,34,'cp 278',NULL);
INSERT INTO evaluations VALUES(11663,35,'cp 286',NULL);
INSERT INTO evaluations VALUES(11663,36,'cp 280','d4e5 f6g4 c1f4 c8e6 e5d6 f8d6 d1d2 e8g8 h2h3 d6f4 d2f4 d8b8 a1b1 g4f6 e1g1 h7h6 f1d1 f8e8 d1d3 c6c5 b2b3 b8b4 f3e5 a6a5 d3e3 b4b6 g1h2 e8d8 f4g3 d8d2 e5c4 e6c4 b3c4 b6e6 c3d5 c7c6 d5c7 f6h5 c7e6 h5g3 h2g3 f7e6');
INSERT INTO evaluations VALUES(11659,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(11659,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(11659,32,'cp 257',NULL);
INSERT INTO evaluations VALUES(11659,33,'cp 283',NULL);
INSERT INTO evaluations VALUES(11659,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(11659,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(11659,36,'cp 269','f6g4 c1f4 c8e6 e5d6 f8d6 d1d2 e8g8 h2h3 d6f4 d2f4 d8b8 a1b1 g4f6 e1g1 f8e8 f1e1 h7h6 e1e3 a6a5 b2b3 a5a4 b3a4 b8a7 g1h2 a8d8 f4g3 a7a5 a2a3 a5a7 g3f4 f6h5 f4h4 h5f6 b1e1 a7c5 e4e5 f6d5 c3e4 c5e7 h4e7 e8e7 e3d3 d8a8 f3d4 a8a4 d4c6');
INSERT INTO evaluations VALUES(16407,30,'cp -324',NULL);
INSERT INTO evaluations VALUES(16407,31,'cp -329',NULL);
INSERT INTO evaluations VALUES(16407,32,'cp -325',NULL);
INSERT INTO evaluations VALUES(16407,33,'cp -335',NULL);
INSERT INTO evaluations VALUES(16407,34,'cp -342',NULL);
INSERT INTO evaluations VALUES(16407,35,'cp -318',NULL);
INSERT INTO evaluations VALUES(16407,36,'cp -312','c6d4 f3d1 f6e4 g1f3 d7d5 c4e2 d4f3 e2f3 e4g5 e1g1 f8c5 b1c3 c7c6 f1e1 c5d6 d2d4 g5f3 d1f3 e5e4 c1f4 e8g8 f3e3 d6f4 e3f4 d8b6 c3e2 b6b2 c2c3 b2a3 f4g3 f8e8 e2f4 a3d6 a2a4 a7a5 a1b1 h7h6 f2f3 c8f5 f4h5 d6g3 h5g3');
INSERT INTO evaluations VALUES(17004,30,'cp -312',NULL);
INSERT INTO evaluations VALUES(17004,31,'cp -315',NULL);
INSERT INTO evaluations VALUES(17004,32,'cp -316',NULL);
INSERT INTO evaluations VALUES(17004,33,'cp -331',NULL);
INSERT INTO evaluations VALUES(17004,34,'cp -328',NULL);
INSERT INTO evaluations VALUES(17004,35,'cp -310',NULL);
INSERT INTO evaluations VALUES(17004,36,'cp -297','f3d1 f6e4 d2d3 e4c5 c2c3 d4f5 g1e2 d7d5 c4b3 c5b3 a2b3 f7f6 e2g3 f8e7 d1h5 g7g6 h5d1 f5g3 f2g3 h7h5 d3d4 c8g4 d1d3 g4f5 d3d1 e5d4 d1d4 c7c5 d4d2 d8d7 e1g1 a8d8 d2f2 a7a6 c1f4 e8g8 b1d2 d5d4 a1e1 g6g5');
INSERT INTO evaluations VALUES(17525,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(17525,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(17525,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(17525,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(17525,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(17525,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(17525,36,'cp 132','c3d4 d7d6 d4d5 c6e5 f3d4 e7e6 d5e6 f7e6 b1c3 g8f6 f2f4 e5c6 d4c6 b7c6 f1d3 e6e5 e1g1 c8e6 d1e2 f8e7 d3c4 d8d7 f4e5 d6e5 c1e3 e8g8 a1d1 d7c8 c3a4 g8h8 a4b6 e6g4 b6c8 g4e2 c4e2 a8c8 d1c1 a6a5 f1f5 e7d6 c1c4 h7h6 b2b3 c8b8 e3c5 d6c5 c4c5 f8d8 c5e5');
INSERT INTO evaluations VALUES(17521,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(17521,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(17521,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(17521,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(17521,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(17521,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(17521,36,'cp 118','d7d6 d4d5 c6e5 f3d4 e7e6 d5e6 f7e6 b1c3 f8e7 d1b3 e5c6 d4c6 b7c6 f1c4 d8d7 e1g1 g8f6 e4e5 d6e5 c3a4 d7c7 c1e3 f6d5 e3c5 e8g8 c5e7 a8b8 b3a3 c7e7 a3e7 d5e7 a1e1 g8f7 e1e5 e7d5 b2b3 f7e7 e5h5 d5f6 h5c5 c8d7 c5g5 g7g6');
INSERT INTO evaluations VALUES(17592,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(17592,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(17592,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(17592,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(17592,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(17592,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(17592,36,'cp 135','d4d5 c6e5 f3d4 e7e6 d5e6 f7e6 b1c3 f8e7 d1b3 e5c6 d4c6 b7c6 f1c4 d8d7 e1g1 g8f6 e4e5 d6e5 c3a4 d7c7 c1e3 f6d5 e3c5 e8g8 a1e1 e7c5 a4c5 a6a5 b3c2 g8h8 b2b3 c7d6 c5d3 d5b4 c2e2 b4d3 c4d3 a5a4 e1d1 a4b3 a2b3 d6b4 d3c4 b4c3 f1e1 c8b7 e2e5');
INSERT INTO evaluations VALUES(23327,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(23327,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(23327,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(23327,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(23327,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(23327,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(23327,36,'cp 200','g8e7 b1c3 e7c6 d4e3 f8d6 c1d2 b7b6 e1c1 c8b7 f2f4 c6a5 c4d5 d8e7 e4e5 f6e5 d5b7 a5b7 c3d5 e7f7 f4e5 d6c5 e3b3 e8g8 h1f1 f7e6 d5c7 e6b3 a2b3 a8c8 c7d5 f8f1 d1f1 a7a5 f1d1 c8e8 d2c3 b7d8 b3b4 a5b4 c3b4 e8e5 b4c5 b6c5 d5b6');
INSERT INTO evaluations VALUES(17146,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(17146,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(17146,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(17146,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(17146,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(17146,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(17146,36,'cp 192','b1c3 e7c6 d4e3 f8d6 f2f4 b7b6 c1d2 c8b7 e1c1 c6a5 c4d5 d8e7 e4e5 f6e5 d5b7 a5b7 c3d5 e7f7 f4e5 d6c5 e3b3 e8g8 h1f1 f7e6 d5c7 e6b3 a2b3 f8f1 d1f1 a8f8 f1d1 a7a5 d2c3 f8f7 e5e6 d7e6 c7e6 c5e7 c3d4 e7d6 d1e1 f7e7 d4b6 d6h2 c2c3 h2d6 c1c2 g8f7 e6g5 f7g6 e1e7 d6e7 g5e6 g6f7');
INSERT INTO evaluations VALUES(48422,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(48422,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(48422,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(48422,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(48422,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(48422,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(48422,36,'cp -52','c7c5 e2e4 f6e4 c3e4 d5e4 c2c3 a7a6 c1e3 c5d4 e3d4 b8d7 d1h5 d7e5 d4e5 h7h6 f2f4 e4f3 h5f3 d8d5 f3d5 e6d5 f1e2 c8e6 e2f3 f7f6 e5d4 f8d6 e1g1 e8c8 f1e1 d8e8 a2a4 h6h5 a1d1 d6c7 d4f2 e6g4 d1d5 e8e1 f2e1 g4f3 g2f3 g7g5 e1g3 h8d8');
INSERT INTO evaluations VALUES(14526,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(14526,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(14526,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(14526,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(14526,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(14526,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(14526,36,'cp -74','c6a5 c4b5 c7c6 b5a4 d6d5 d2d3 f8g7 c1g5 d5e4 d3e4 h7h6 g5h4 e8g8 b1d2 b7b5 a4c2 c8e6 g1e2 d8e7 b2b3 a5b7 e1g1 f8d8 f1d1 a7a5 d2f1 b7c5 f1e3 a5a4 b3b4 c5b7 e2g3 b7d6 g3f1 d6c4 e3c4 d8d1 f3d1 e6c4 f1e3 e7e6 e3c4');
INSERT INTO evaluations VALUES(16095,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(16095,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(16095,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(16095,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(16095,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(16095,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(16095,36,'cp 42','b7c6 b1c3 f8d6 d1d3 e8g8 c1d2 d6c5 e1g1 d7d6 a1e1 f8e8 b2b3 c8e6 c4a6 e6c8 a6c8 d8c8 c3a4 c5b6 d2c3 c8e6 d3g3 f6h5 g3f3 h5f6 e4e5 f6d5 c3d2 b6d4 c2c4 d5b6 e5d6 e6d6 c4c5 d4c5 e1e8 a8e8 d2f4 d6d5 a4c5');
INSERT INTO evaluations VALUES(11907,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11907,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(11907,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(11907,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(11907,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(11907,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11907,36,'cp 48','b1c3 f8d6 d1d3 e8g8 c1d2 d6c5 e1g1 d7d6 a1e1 f8e8 h2h3 c8e6 c4a6 e6c8 a6c8 d8c8 e1e2 c8e6 b2b3 e6e5 f1e1 e5d4 c3a4 d4d3 c2d3 c5d4 d2c3 c6c5 c3d4 c5d4 e1c1 c7c5 a4b2 f6d7 b2c4 d6d5 c4d6 e8e6 d6b5 a8d8 b5c7');
INSERT INTO evaluations VALUES(11906,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(11906,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(11906,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(11906,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(11906,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(11906,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(11906,36,'cp 50','f8d6 d1d3 e8g8 c1d2 d6c5 e1g1 d7d6 a1e1 f8e8 h2h3 c8e6 c4a6 e6c8 a6c8 d8c8 e1e2 c8e6 f1e1 f6d7 c3a4 c5b6 b2b3 d7e5 d3g3 e6g6 g3g6 f7g6 d2c3 g6g5 a4b2 a7a5 a2a4 e8e6 c3e5 e6e5 g2g3 a8e8 g1g2 e5e7 f2f4 g5f4 g3f4');
INSERT INTO evaluations VALUES(21285,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(21285,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21285,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(21285,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21285,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21285,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(21285,36,'cp 17','g7g5 d2d4 g5g4 h3g4 c8g4 c4b5 g8e7 d4e5 a7a6 b5e2 d6e5 b1c3 h8g8 f1e1 d8d1 e1d1 g4f3 e2f3 c6d4 f3h5 d4c2 a1b1 c7c6 c3e2 e7g6 c1d2 a8d8 d2a5 d8d1 b1d1 f8e7 a5b6 c6c5 b6c5 e7c5 d1c1 g6f4 e2f4 e5f4 c1c2 g8g5 h5f3 e8e7 g1f1 h6h5 e4e5 g5e5 f3b7 a6a5 c2c4 h5h4 c4f4 c5b4 f4e4 e5e4 b7e4 b4d6 b2b3 d6b4');
INSERT INTO evaluations VALUES(27873,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(27873,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(27873,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(27873,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(27873,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(27873,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(27873,36,'cp 14','h2h3 g4h5 b1d2 h7h6 g5h4 f8e8 c2c3 b8d7 d1c2 h5f3 d2f3 c7c6 c3c4 d5c4 d3c4 d8c7 a1e1 f6d5 c2d3 d7b6 c4b3 d5f4 d3f5 g7g6 f5c2 b6d5 h4g3 a8d8 a2a3 g8g7 c2d2 f7f6 d2c2 c7d7 b3c4 e8e1 f1e1 d8e8 e1e8 d7e8 g3f4 d6f4');
INSERT INTO evaluations VALUES(27872,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27872,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27872,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27872,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27872,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27872,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27872,36,'cp 0','g4h5 b1d2 h7h6 g5h4 f8e8 c2c3 b8d7 d1c2 h5f3 d2f3 c7c6 c3c4 d5c4 d3c4 d8c7 c2b3 e8e7 a1e1 a8e8 e1e7 e8e7 f1e1 e7e1 f3e1 f6d5 c4d5 c6d5 b3d5 g7g5 h4g3 d6g3 f2g3 c7g3 e1c2 g3d3 d5b7 d3c2 b7d7 c2c1 g1h2 c1f4 h2g1');
INSERT INTO evaluations VALUES(27875,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27875,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27875,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(27875,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27875,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(27875,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(27875,36,'cp 10','b1d2 h7h6 g5h4 f8e8 c2c3 b8d7 d1c2 h5f3 d2f3 c7c6 c3c4 d5c4 d3c4 d8c7 a1e1 f6d5 a2a3 a8c8 c2d3 g7g5 h4g3 d6g3 f2g3 c7g3 d3f5 e8e1 f1e1 c8c7 e1e7 g3f4 e7f7 f4f5 f7f5 d7b6 c4b3 g8g7 g1f2 c7e7 h3h4 g5h4 f3h4 d5e3 f5a5 b6c4 a5a7 c4b2 f2e2 e3c4 e2f2');
INSERT INTO evaluations VALUES(32316,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(32316,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(32316,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(32316,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32316,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(32316,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(32316,36,'cp 17','g8f6 d3f5 e6f5 d1d3 d8c8 b2b3 b8a6 c2c4 g7g6 c1a3 a6b4 d3d2 c7c6 f3e5 f6e4 d2e2 f7f6 e5d3 b4d3 e2d3 f8a3 b1a3 e8g8 f2f3 e4d6 e1g1 f8e8 a1c1 c8e6 f1e1 a8c8 c4c5 d6f7 b3b4 a7a6 a3c2 e6d7 a2a4 h7h5 c1b1 g8g7 b4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(40648,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(40648,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(40648,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40648,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40648,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40648,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40648,36,'cp 0','b8c6 f1e2 a7a6 e1g1 e8g8 b2b3 c5a7 c1b2 c8f5 a1c1 f8e8 h2h3 d5d4 e3d4 c6d4 f3d4 a7d4 e2f3 f5d3 f1e1 d4f2 g1f2 d8d4 f2g3 d4d6 g3f2');
INSERT INTO evaluations VALUES(8370,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8370,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8370,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8370,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8370,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8370,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8370,36,'cp 0','f1e2 e8g8 e1g1 a7a6 b2b3 c5a7 c1b2 c8f5 h2h3 f5g6 e2d3 g6h5 d3e2');
INSERT INTO evaluations VALUES(8369,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8369,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8369,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8369,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8369,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8369,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8369,36,'cp 0','a7a6 e1g1 e8g8 b2b3 c5a7 c1b2 c8f5 a1c1 f8e8 h2h3 d5d4 f3d4 c6d4 e3d4 a7d4 e2f3 f5d3 f1e1 d4f2 g1f2 d8b6 f2g3 b6c7 g3f2');
INSERT INTO evaluations VALUES(43204,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43204,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43204,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(43204,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43204,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43204,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43204,36,'cp 0','d7d6 h2h3 e8g8 e2e3 b7b6 f1e2 b8d7 f4h2 c8b7 e1g1 f8e8 a2a4 e7e5 a4a5 a7a6 a5b6 c7b6 b1d2 b6b5 d1b3 d8b6 f1c1 g7f8 f3g5 e8e7 g5f3');
INSERT INTO evaluations VALUES(41315,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(41315,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41315,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(41315,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(41315,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41315,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41315,36,'cp 0','h2h3 e8g8 b1d2 b8d7 e2e3 b7b6 f1e2 c8b7 e1g1 d8e8 d1c2 e7e5 d4e5 d6e5 f4g3 e8e7 a2a4 a7a5 a1d1 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(41313,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41313,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41313,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41313,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41313,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41313,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41313,36,'cp 0','e8g8 b1d2 b8d7 e2e3 b7b6 f1e2 c8b7 e1g1 d8e8 d1c2 e7e5 d4e5 d6e5 f4g3 e8e7 a2a4 a7a5 a1d1 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(36607,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36607,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36607,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36607,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36607,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36607,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36607,36,'cp 0','b1d2 b8d7 e2e3 b7b6 f1e2 c8b7 e1g1 d8e8 d1c2 e7e5 d4e5 d6e5 f4g3 e8e7 a2a4 a7a5 a1d1 f6h5 g3h2 h5f6');
INSERT INTO evaluations VALUES(45380,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(45380,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(45380,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(45380,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(45380,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(45380,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(45380,36,'cp -27','c8g4 d3e2 g4e2 d1e2 e7e6 g1f3 h7h6 e1g1 b8c6 f1d1 c5d4 c3d4 f8d6 b1c3 e8g8 f3e1 d8b6 e1d3 c6b4 d3e1 f6e4 c1d2 b4c6 e1d3 e4c3 d2c3 c6e7 d1c1 f8c8 c3e1 a7a5 g2g3 c8c4 c1c2 a5a4 a1c1 a8c8 e2d1 b6b5 e1c3 e7f5 c3e1');
INSERT INTO evaluations VALUES(55649,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(55649,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(55649,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(55649,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(55649,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55649,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55649,36,'cp 27','g7g6 c2c4 c7c5 b1c3 b8c6 e1g1 f8g7 a2a3 a7a5 d2d3 g8e7 c1g5 f7f6 g5d2 e8g8 a1b1 c8e6 f3e1 f6f5 b2b4 a5b4 a3b4 c6b4 e1c2 e7c6 c2b4 c6b4 g2b7 a8b8 b7g2 d8d7 d1b3 h7h6 c3d5 e6d5 g2d5 b4d5 b3b8 f8b8 b1b8 g8h7 c4d5 e5e4 d3e4 f5e4 f1b1 d7g4');
INSERT INTO evaluations VALUES(57416,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(57416,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(57416,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(57416,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(57416,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(57416,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(57416,36,'cp 126','f1d3 g7g6 b1c3 d7d6 h2h4 b8c6 h4h5 g6g5 h5h6 d8d7 d4d5 c6e7 d3b5 c7c6 b5c4 d7g4 c4e2 g4d7 f2f4 e7d5 c3d5 c6d5 d1d5 g8e7 d5b3 d7e6 f4g5 e6b3 a2b3 f6g5 g1f3 g5g4 f3g5 e7c6 g5e4 f8e7');
INSERT INTO evaluations VALUES(4017,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(4017,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(4017,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(4017,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(4017,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(4017,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(4017,36,'cp 50','h2h3 e7e5 d4d5 a7a5 d1c2 d7c5 f3d2 c8d7 f1e2 b7b6 a2a3 c7c6 e1g1 a5a4 a1b1 c6d5 c4d5 d8e7 f1c1 f8b8 b2b4 a4b3 d2b3 h7h6 g5e3 a8a3 b3c5 d6c5 c3b5 d7b5 e2b5 g8h7 b5c6 e7d6 g2g3 f6e8 c6e8');
INSERT INTO evaluations VALUES(14083,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14083,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(14083,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(14083,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(14083,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(14083,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(14083,36,'cp 53','a7a6 d1d3 h7h6 g5f6 g7f6 f3h4 h6h5 g2g3 c8d7 h2h3 f8h6 f1e2 h6g5 h4g2 f6f5 f2f4 g5f6 e1c1 e5f4 g3f4 d7b5 c3b5 a6b5 a2a3 e8f8 e4e5 d6e5 d5d6 c7d6 d3d6 d8d6 d1d6 a8a6 f4e5 f6e5 d6d8 f8g7 d8d7 a6e6 e2b5 g7f6 d7b7 h8g8 g2h4 g8g3 b5d7');
INSERT INTO evaluations VALUES(20229,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(20229,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(20229,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(20229,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(20229,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20229,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20229,36,'cp 12','d8d5 f1d3 c8d7 c1e3 g8f6 e1g1 a8d8 b1a3 a7a6 a3c4 c5d4 c3d4 f8d6 d1b3 d8b8 a1c1 e8g8 c4d6 d5d6 e3g5 d6b4 b3c2 b8c8 a2a3 b4d6 c2c5 d6c5 d4c5 f6d5 g5d2 f8d8 h2h3 f7f6 f1e1 g8f7 b2b4');
INSERT INTO evaluations VALUES(26951,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(26951,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(26951,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(26951,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(26951,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(26951,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(26951,36,'cp 116','a2a3 h5f3 d1f3 d8d7 c2c3 g7g6 e1g1 f8g7 f1e1 g8h6 a3a4 e8g8 b1d2 g8h8 f3d1 a7a6 d2f3 f7f5 b2b4 h6f7 a1b1 c6e7 c1d2 a8d8 d1b3 d6d5 e4d5 e7d5 b4b5 a6a5 d3d4 e5e4 f3g5 d7e7 c4d5 f7g5 d2g5 e7g5');
INSERT INTO evaluations VALUES(36510,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(36510,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(36510,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(36510,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(36510,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(36510,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36510,36,'cp -9','f6e4 c3e4 d6d5 d1e2 d5c4 e1c1 c7c6 e2c4 c8g4 c4e2 d8a5 a2a3 a5f5 f4g5 b8d7 h2h3 g4f3 e2f3 f5f3 g2f3 f8e8 c2c3 a8d8 h3h4 h7h5 h1e1 d7f8 e4c5 b7b6 c5d3 f7f6 g5e3 f8e6 d3b4 d8d6 a3a4 g8f7 c1c2 a7a5 b4d3');
INSERT INTO evaluations VALUES(48453,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(48453,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(48453,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(48453,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(48453,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(48453,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(48453,36,'cp 64','h7h6 g5e3 f8b4 f2f3 e8g8 a2a3 b4e7 a1c1 f6d7 g1e2 b7b6 e2f4 c8b7 d4d5 e6d5 f4d5 d7f6 d5e7 d8e7 f1b5 c7c6 b5c4 b8d7 e1g1 d7e5 c4e2 c6c5 d1a4 b7c6 a4b3 a8d8 f1d1 f8e8 a3a4 e5g6 d1d8 e8d8 c1d1 d8d1 b3d1');
INSERT INTO evaluations VALUES(51412,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(51412,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(51412,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(51412,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(51412,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51412,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(51412,36,'cp 12','c8b7 e2e3 g8f6 f1e2 d7d5 e1g1 f8d6 f4d6 c7d6 c2c4 d5c4 e2c4 e8g8 b1c3 a7a6 c4d3 b8c6 h2h3 b6b5 f1e1 c6b4 d3b1 b7f3 d1f3 d6d5 e1c1 a8c8 a2a3 b4c6 b1d3 c6a5 c1c2 d8e7 a1c1 c8c7 c3a2 c7c2 c1c2 a5c4 f3e2');
INSERT INTO evaluations VALUES(54272,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(54272,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(54272,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(54272,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(54272,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(54272,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54272,36,'cp 11','g8f6 c2c4 d5d4 e2e3 b8c6 e3d4 c5d4 d2d3 f8d6 b1a3 e8g8 a3c2 e6e5 b2b4 f8e8 f1e1 a7a6 a2a4 c8f5 c4c5 d6f8 b4b5 a6b5 a4b5 a8a1 c2a1 c6b4 f3e5 f8c5 a1b3 d8b6 c1a3 b6b5 b3c5 b5c5 d1b3 e8e5 e1e5 c5e5');
INSERT INTO evaluations VALUES(57482,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(57482,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(57482,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(57482,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(57482,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(57482,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(57482,36,'cp 25','g8f6 b1c3 b8c6 c3d5 c6a5 d5f6 d8f6 c4d5 c7c6 d1h5 g7g6 f5g6 h7g6 h5f3 f6g7 d5b3 d7d5 d2d3 c8f5 g1e2 e8c8 e2g3 f8d6 g3f5 g6f5 f3f5 c8b8 h2h4 e5e4 c1f4 e4d3 f4d6 d8d6 f5d3 g7g2 e1c1 h8h4 h1h4 g2g5 d3d2 a5b3 a2b3 g5h4 f2f4 d6d8 d1f1 d8f8 d2e3 h4g4 f4f5 f8f5 e3e8 b8c7');
INSERT INTO evaluations VALUES(49985,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(49985,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(49985,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(49985,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(49985,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(49985,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(49985,36,'cp 24','b1c3 b8c6 c3d5 c6a5 d5f6 d8f6 c4d5 c7c6 d1h5 g7g6 f5g6 h7g6 h5f3 f6g7 d5b3 d7d5 d2d3 c8f5 g1e2 e8c8 e2g3 f8d6 g3f5 g6f5 f3f5 c8b8 h2h4 e5e4 c1f4 e4d3 f4d6 d8d6 f5d3 g7g2 e1c1 h8h4 h1h4 g2g5 f2f4 g5h4 d1f1 a5b3 a2b3 d6e6 c1b1 e6e1 b1a2 e1f1 d3f1 h4f6 f4f5 b8c7 f1f4 c7d7');
INSERT INTO evaluations VALUES(1283,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(1283,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(1283,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(1283,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(1283,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(1283,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(1283,36,'cp 65','f8c5 c2c3 g8e7 d2d4 e5d4 c1g5 f6d6 e4e5 d6g6 f1e1 d7d5 e5d6 c5d6 b3d5 c8b7 g5e7 d6e7 c3d4 g6d6 b1c3 b5b4 d5c6 b7c6 c3e4 c6e4 e1e4 e8g8 d1e2 e7f6 a1c1 a6a5 g2g3 a5a4 g1g2 d6d5 c1c4 f8d8 b2b3 a4b3 a2b3 d5b7 e2c2 d8d7 c4c6 a8d8 c2c4 h7h6 f3e5');
INSERT INTO evaluations VALUES(95,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(95,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(95,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(95,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(95,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(95,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(95,36,'cp 48','c2c3 g8e7 d2d4 e5d4 c1g5 f6d6 e4e5 d6g6 b3c2 g6h5 f1e1 e8g8 b1d2 d7d5 e5d6 c5d6 g5e7 c6e7 d2e4 e7f5 e4d6 f5d6 e1e5 h5g4 h2h3 g4f4 d1d4 f4d4 f3d4 c8d7 e5c5 a8c8 a2a4 d6b7 c5d5 f8d8 a4b5 d7b5 d5e5 b7d6 e5c5 g8f8 d4c6 d8e8 c2h7');
INSERT INTO evaluations VALUES(96,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(96,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(96,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(96,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(96,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(96,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(96,36,'cp 55','g8e7 d2d4 e5d4 c1g5 f6g6 g5e7 c5e7 c3d4 e8g8 b1c3 d7d6 c3d5 e7d8 a1c1 c6a5 b3c2 g6h6 b2b3 c8e6 d5e3 c7c5 d4c5 d6c5 d1d6 c5c4 b3b4 a5b7 d6c6 e6c8 c6h6 g7h6 f1d1 d8g5 e4e5 b7d8 d1d6 d8e6 f3g5 h6g5 c2e4 a8a7 a2a3 e6f4 c1d1 c8e6 g2g3');
INSERT INTO evaluations VALUES(3989,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3989,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3989,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3989,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3989,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3989,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3989,36,'cp 0','d5c4 d3c4 c7c5 e1g1 a7a6 a2a4 c5d4 f3d4 d8b6 c3e4 h7h6 g5h4 f8e8 a4a5 b6b4 a1a4 b4b2 h4f6 d7f6 a4a2 b2b4 a2a4');
INSERT INTO evaluations VALUES(3998,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3998,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3998,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3998,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3998,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3998,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3998,36,'cp 0','d3c4 c7c5 e1g1 a7a6 a2a4 c5d4 f3d4 d8b6 c3e4 h7h6 g5h4 f8e8 a4a5 b6b4 a1a4 b4b2 h4f6 d7f6 a4a2 b2b4 a2a4');
INSERT INTO evaluations VALUES(3997,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3997,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3997,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3997,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3997,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3997,36,'cp 0','a7a6 a2a4 c7c5 e1g1 c5d4 f3d4 d8b6 c3e4 h7h6 g5h4 f8e8 a4a5 b6b4 a1a4 b4b2 a4a2 b2b4 a2a4');
INSERT INTO evaluations VALUES(7260,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7260,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7260,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7260,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7260,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7260,36,'cp 0','d7d6 e1g1 e8g8 c3e2 c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 c6e7 e2g3 d8e8 c2c3 f8f7 f1f2 e7g6 d1c2 f6h7 a1f1 b7b6 f3h2 f7f2 f1f2 e8d7 d3d4 h7g5 h2g4 a8f8 b2b4 d7e7 f2f8 g8f8 d4e5 d6e5 c2e2');
INSERT INTO evaluations VALUES(10052,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(10052,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(10052,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(10052,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10052,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10052,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10052,36,'cp 0','h7h6 g5h4 d7d6 c3a4 a7a6 a4c5 d6c5 c4d5 c6d4 c2c3 d4e6 g1f3 e6f4 e1g1 d8d6 h4g3 f6h5 d3d4 e5d4 c3d4 c5d4 f3d4 h5g3 h2g3 f4d5 e4d5 e8g8 d1b3 f8d8 f1d1 c8g4 d4f3 c7c6 d5c6 d6c6 d1d8 a8d8 f3e5 c6e6 e5g4 e6g4 b3b7 d8d1 a1d1 g4d1 g1h2 d1h5 h2g1');
INSERT INTO evaluations VALUES(9733,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(9733,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(9733,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(9733,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(9733,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(9733,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9733,36,'cp 0','g5h4 d7d6 c3a4 a7a6 a4c5 d6c5 c4d5 c6b4 d5b3 d8d6 h4f6 d6f6 d1f3 f6e7 c2c3 b4c6 g1e2 c8e6 b3e6 e7e6 h2h4 a8d8 f3h3 e6f6 h3e3 f6e6');
INSERT INTO evaluations VALUES(9742,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9742,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9742,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(9742,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(9742,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9742,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9742,36,'cp 0','d7d6 c3a4 a7a6 a4c5 d6c5 c4d5 c6b4 d5b3 d8d6 h4f6 d6f6 d1f3 f6e7 a2a3 b4c6 g1e2 c8e6 b3e6 e7e6 h2h4 e8g8 f3h3 e6h3 h1h3 a8d8 b2b3 d8d6 e1d2 c6d4 a1f1 d6g6 g2g3 d4f3 d2e3 f3d4');
INSERT INTO evaluations VALUES(14126,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(14126,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14126,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(14126,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(14126,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(14126,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14126,36,'cp 6','e7f6 e2e3 c8e6 a2a3 a7a6 f1e2 f8d6 e3e4 d5e4 c3e4 e8g8 e1g1 f8e8 c2c3 c6e7 f1e1 c7c6 e4d6 d8d6 f3d2 d6f4 e2d3 f6f5 d3c4 e6c4 d2c4 e7g6 d1d2 f4c7 e1e8 a8e8 a1e1 e8d8 c4e5 g6e5 e1e5');
INSERT INTO evaluations VALUES(14576,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(14576,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14576,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(14576,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14576,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14576,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(14576,36,'cp 14','e2e3 f8b4 f1d3 c8e6 e1g1 e8g8 c3e2 b4d6 a2a3 c6e7 b2b3 c7c6 c2c4 d6c7 c4c5 g7g6 b3b4 d8d7 a3a4 h6h5 b4b5 g8g7 e2f4 e6g4 h2h3 c7f4 e3f4 g4f3 d1f3 f8e8 a1e1 a8d8 f3e3 e7f5 d3f5 d7f5 e3e8');
INSERT INTO evaluations VALUES(14575,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14575,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(14575,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(14575,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14575,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(14575,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(14575,36,'cp -8','c8e6 a2a3 a7a6 f1e2 f8d6 e3e4 d5e4 c3e4 e8g8 e1g1 e6d5 e4d6 d8d6 c2c3 c6a5 f3d2 b7b5 f1e1 d6c6 e2f3 f8e8 f3d5 c6d5 d1c2 a5c4 d2c4 b5c4 h2h3 a6a5 c2d2 e8e1 a1e1 c7c5 e1e7 c5d4 d2d4 d5d4 c3d4 a8d8 e7a7 d8d4 a7a5');
INSERT INTO evaluations VALUES(17365,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(17365,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(17365,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(17365,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(17365,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(17365,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(17365,36,'cp -39','g8f6 c4a2 b7b5 g1f3 f8d6 e1g1 c8b7 a3a4 d8b6 h2h3 e8g8 c1e3 f8d8 c3e2 b6c7 c2c3 d6e7 f1e1 d7d5 e3f4 e6e5 a4b5 a6b5 f3e5 c6e5 d3d4 e7d6 d4e5 d6e5 f4e5 c7e5 e4d5 f6d5 a2d5 a8a1 d1a1 e5d5 f2f3');
INSERT INTO evaluations VALUES(11075,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(11075,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(11075,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(11075,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(11075,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(11075,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(11075,36,'cp -33','g1f3 b7b5 c4a2 f8d6 a3a4 d8b6 c3e2 c6e5 f3e5 d6e5 c2c3 c8b7 d3d4 c5d4 c3d4 e5c7 e4e5 f6d5 a2d5 b7d5 e2f4 b6b7 e1g1 e8g8 f4d5 b7d5 c1e3 f7f6 a1c1 c7b6 h2h3 b5a4 d1a4 f6e5 d4e5 b6e3 f2e3 d5e5 a4d7 e5e3 g1h1');
INSERT INTO evaluations VALUES(11074,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(11074,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(11074,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(11074,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(11074,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(11074,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(11074,36,'cp -41','b7b5 c4b3 f8d6 e1g1 c8b7 a3a4 b5b4 c3e2 e8g8 c2c3 h7h6 b3c2 b4c3 b2c3 d8c7 e2g3 a6a5 g3e2 f8d8 h2h3 d6f8 c1f4 e6e5 f4h2 d7d5 e4d5 f6d5 f1e1 b7a6 a1b1 a8b8 d1d2 b8b1 e1b1 f8d6 h2g3 d5f4 e2f4 e5f4');
INSERT INTO evaluations VALUES(19194,30,'cp 505',NULL);
INSERT INTO evaluations VALUES(19194,31,'cp 525',NULL);
INSERT INTO evaluations VALUES(19194,32,'cp 513',NULL);
INSERT INTO evaluations VALUES(19194,33,'cp 523',NULL);
INSERT INTO evaluations VALUES(19194,34,'cp 521',NULL);
INSERT INTO evaluations VALUES(19194,35,'cp 528',NULL);
INSERT INTO evaluations VALUES(19194,36,'cp 540','d5d6 e7f5 d1e2 f8e7 d6e7 g8e7 g2g4 f5d4 f3d4 c5d4 c1g5 f7f6 g5d2 e8g8 b1a3 e7c6 e1c1 f8e8 e2f3 d7d5 h1g1 a7a6 g4g5 e8f8 g5f6 f8f6 f3g2 d8f8 f1d3 c8f5 f2f4 f5d3 c2d3 g8h8 a3c2 a8c8 c1b1 f6g6 g2d5 g6g1');
INSERT INTO evaluations VALUES(19190,30,'cp 496',NULL);
INSERT INTO evaluations VALUES(19190,31,'cp 490',NULL);
INSERT INTO evaluations VALUES(19190,32,'cp 525',NULL);
INSERT INTO evaluations VALUES(19190,33,'cp 505',NULL);
INSERT INTO evaluations VALUES(19190,34,'cp 503',NULL);
INSERT INTO evaluations VALUES(19190,35,'cp 521',NULL);
INSERT INTO evaluations VALUES(19190,36,'cp 505','e7f5 d1e2 f8e7 d6e7 g8e7 g2g4 f5d4 f3d4 c5d4 b1a3 e8g8 c1g5 f7f6 g5d2 d7d5 e1c1 e7c6 f2f4 g8h8 e2g2 a7a6 c1b1 b7b5 b1a1 b5b4 a3b1 a6a5 h2h4 d8d7 d1e1 c8a6 f1a6 a8a6 g2e2 a6a8 h4h5 a5a4 h5h6 g7g5 f4g5 f6g5 e2e6 d7e6 e1e6 a8c8 d2g5 c6a5 g5f6 h8g8 f6d4 c8c2');
INSERT INTO evaluations VALUES(21947,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(21947,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(21947,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(21947,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(21947,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(21947,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(21947,36,'cp 61','g1f3 g8f6 f1d3 g7g6 h2h3 f8g7 e1g1 e8g8 f1e1 a7a5 a2a4 f8e8 d3c2 h7h6 c1d2 e5d4 c3d4 c6b4 b1c3 b4c2 d1c2 c7c6 a1d1 c8e6 e4e5 d6e5 d4e5 f6d7 c2c1 g6g5 c3e4 e6b3 e4d6 b3d1 c1d1 e8e5 f3e5 g7e5 d2c3 e5c3 e1e8 d8e8 d6e8 a8e8 b2c3 d7f6');
INSERT INTO evaluations VALUES(23385,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(23385,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(23385,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(23385,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(23385,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(23385,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(23385,36,'cp 191','d2d4 e7e6 d4d5 c6e5 f3e5 f6e5 d1h5 g7g6 h5e5 g8f6 c1g5 f8g7 b1c3 d7d6 e5g3 e8g8 f1c4 d8e8 e1c1 f6h5 g3e3 e6e5 f2f3 c8d7 g2g3 b7b5 c4e2 b5b4 c3b1 a7a5 b1d2 a5a4 c1b1 e8f7 d2c4 f8b8 h1f1 h5f6 g3g4 d7b5 c4d2 b5e2 e3e2 h7h6 g5f6 g7f6 d2c4');
INSERT INTO evaluations VALUES(30453,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(30453,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(30453,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(30453,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(30453,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(30453,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(30453,36,'cp 201','f1a6 b8a6 g1f3 g7g6 c2c4 f8g7 b1c3 c7c5 d4d5 a6c7 e4e5 g8h6 c1h6 g7h6 d1e2 d7d6 h2h4 e8g8 a1d1 h6g7 h4h5 d6e5 f3e5 g7e5 e2e5 d8d7 e5h2 g6g5 f2f4 c7e8 f4g5 d7g4 h2h4 g4h4 h1h4 e8d6 b2b3 h7h6 c3e4 h6g5 e4g5 a8e8 h5h6 f7f6 g5f3 g8h8 a2a3');
INSERT INTO evaluations VALUES(30450,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(30450,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(30450,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(30450,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(30450,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(30450,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(30450,36,'cp 192','b8a6 g1f3 g7g6 c2c4 c7c5 d4d5 f8g7 b1c3 a6c7 e4e5 d7d6 d1e2 g8h6 e5d6 d8d6 c3e4 d6d8 h2h4 e8g8 h4h5 h6g4 c1f4 b6b5 a1d1 b5c4 h5h6 g7h8 e2c4 c7e8 e1g1 e8d6 e4d6 e7d6 c4a6 d8b6 a6b6 a7b6 f4d6 f8e8');
INSERT INTO evaluations VALUES(25816,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(25816,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(25816,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(25816,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(25816,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(25816,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(25816,36,'cp 187','d1e2 d8c8 g1f3 g7g6 e1g1 f8g7 e4e5 c7c6 c2c4 a6c7 c1f4 c7e6 f4d2 e6c7 b1c3 g8h6 d4d5 e8g8 a1d1 c6d5 c4d5 d7d6 d2f4 f8e8 h2h3 h6f5 g2g4 d6e5 f4e5 f5d6 e5g7 g8g7 f3d4 c8d7 f1e1 a8c8');
INSERT INTO evaluations VALUES(39345,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(39345,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(39345,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(39345,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(39345,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(39345,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(39345,36,'cp 24','b4c3 d2c3 b8c6 g1e2 g8f6 e2g3 d7d5 e4d5 d8d5 d1d5 f6d5 c3c4 d5e7 c1e3 f7f6 e1c1 h7h5 h2h4 e8f7 f1e2 g7g6 d1d2 c8e6 h1d1 b7b6 c4c5 a8d8 d2d8 h8d8 c5b6 d8d1 c1d1 c7b6 e3d2 c6d4 e2d3 d4f5 g3f5 e6f5 d3e2 f5e6');
INSERT INTO evaluations VALUES(39396,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(39396,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39396,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(39396,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(39396,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(39396,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(39396,36,'cp 33','d2c3 b8c6 g1e2 g8f6 e2g3 d7d5 e4d5 d8d5 d1d5 f6d5 c3c4 d5e7 c1e3 h7h5 h2h4 f7f6 e1c1 c8e6 f1e2 g7g6 d1d2 e8f7 h1d1 b7b6 c4c5 h8d8 d2d8 a8d8 d1d8 c6d8 c5b6 c7b6 f2f3 d8b7 e3f2 b7d6 b2b3 e7d5 c1d2 d5f4 e2f1 g6g5 h4g5 f6g5');
INSERT INTO evaluations VALUES(39394,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(39394,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39394,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(39394,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(39394,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(39394,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(39394,36,'cp 23','b8c6 g1e2 g8f6 e2g3 d7d5 e4d5 d8d5 d1d5 f6d5 h2h4 f7f5 c1d2 f5f4 g3e4 c8f5 f1d3 e8c8 g2g3 h8f8 e1c1 f4g3 f2g3 d5f6 e4d6 c7d6 d3f5 c8c7 d1f1 d6d5 h4h5 h7h6 d2e1 a7a6 g3g4 c6e7 e1g3 e7f5 f1f5 f6g4 g3e5 g4e5 f5e5');
INSERT INTO evaluations VALUES(50698,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(50698,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50698,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(50698,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(50698,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(50698,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50698,36,'cp 0','b8c6 c3e2 c8f5 c2c3 e7e6 e2g3 f5g6 c1f4 g8f6 f1d3 f8e7 e1g1 e8g8 f1e1 a8c8 d3g6 h7g6 d1e2 c6a5 g3f1 f6h5 f4e5 a5c4 f1e3 c4d6 e5d6 e7d6 g2g3 h5f6 a2a3 d8c7 g1g2 f8e8 a1d1 f6e4 f3d2');
INSERT INTO evaluations VALUES(17552,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(17552,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17552,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17552,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17552,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(17552,36,'cp 0','c3e2 c8f5 e2g3 f5g6 c2c3 e7e6 c1f4 f8e7 f1d3 g8f6 e1g1 e8g8 a2a4 a8c8 f1e1 g6d3 d1d3 e7d6 f4e5 d6e5 d4e5 f6d7 g3h5 c6e7 f3d4 e7c6');
INSERT INTO evaluations VALUES(55214,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55214,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(55214,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(55214,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(55214,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(55214,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(55214,36,'cp 21','e2e3 a7a6 f1c4 g8f6 d2d4 e7e6 a2a4 c7c5 g1f3 b8c6 e1g1 f8e7 d1e2 c5d4 f1d1 e6e5 e3d4 e5d4 f3d4 c6d4 e2e5 d8d6 e5d6 e7d6 d1d4 d6e5 d4d1 e8g8 h2h3 c8d7 c3d5 f6d5 c4d5 a8b8 d5f3 d7c6 f3c6 b7c6');
INSERT INTO evaluations VALUES(3096,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3096,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(3096,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(3096,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(3096,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3096,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3096,36,'cp 21','h2h4 e7d6 f3e5 d8c7 d2f3 c8d7 d1e2 a8c8 e1g1 c6e7 a2a4 h7h6 a4a5 f6e4 e5d7 c7d7 f3e5 d6e5 f4e5 e7g6 e5g3 e4g3 f2g3 f7f5 e2h5 g6e7 g3g4 d7d6 g4f5 e6f5 d4c5 d6c5 h5f3 b7b5 h4h5 a7a6 f3g3 c8d8 d3c2 b5b4');
INSERT INTO evaluations VALUES(7261,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7261,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(7261,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(7261,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7261,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7261,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7261,36,'cp 0','d7d6 c1e3 c5e3 f2e3 c8e6 c4d5 e6d5 e4d5 c6e7 f3d2 d8d7 e3e4 e8g8 d2c4 f6h7 c4e3 f7f5 e4f5 e7f5 d1g4 h6h5 g4h3 h7g5 h3h5 f5e3 f1f8 a8f8 h5g5 e3c2 a1f1 f8f1 g1f1 d7f7 f1g1 c2d4 g5g4 g8h7 g4e4 f7g6 g1f2 c7c6 d5c6 b7c6 e4g6 h7g6 c3e4 d6d5');
INSERT INTO evaluations VALUES(17947,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17947,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(17947,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(17947,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(17947,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(17947,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(17947,36,'cp 78','b5c4 f8c5 e1g1 g8e7 c3d5 e7d5 c4d5 d7d6 c2c3 c6e7 d5c4 c5a7 d2d4 b7b5 c4d3 e8g8 a2a4 b5a4 d1a4 g8h8 h2h3 c8b7 a4c2 e5d4 f3d4 f8e8 f1e1 e7g6 c1e3 d6d5 d3a6 a7d4 a6b7 a8a1 e1a1 d4e3 f2e3');
INSERT INTO evaluations VALUES(22425,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(22425,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(22425,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(22425,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(22425,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(22425,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(22425,36,'cp 32','c2c4 g8f6 b1c3 e7e6 g1f3 d5c4 e2e3 f8d6 f4d6 c7d6 f1c4 d7b6 c4d3 b6d5 e1g1 e8g8 c3d5 f6d5 d1d2 d5f6 f1c1 c8d7 e3e4 h7h6 h2h3 a8c8 c1c8 d8c8 a1c1 c8d8 c1c3 d7c6 d2e3 d8b6 c3b3 b6a5 a2a3 f8c8 b3c3');
INSERT INTO evaluations VALUES(33102,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(33102,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(33102,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(33102,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(33102,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(33102,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(33102,36,'cp 30','d1g4 g7g6 b1c3 g8f6 g4h4 b8d7 g1f3 a7a6 f1d3 e6e5 f3e5 d7e5 d4e5 d6e5 e1c1 c8d7 e3e4 e8c8 e4d5 f6d5 h1e1 e5f4 h4f4 d5f4 d3f1 d7c6 g2g3 f4e6 f1c4 d8d1 c3d1 h8e8 d1c3 b7b5 c4d5 c6d5 c3d5 e6g7 e1e8 g7e8 d5b4 a6a5 b4d3 c8d7 d3e5 d7e6 e5c6');
INSERT INTO evaluations VALUES(44865,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44865,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44865,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44865,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44865,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44865,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44865,36,'cp 0','c6d5 g1f3 c8g4 h2h3 g4f3 d1f3 b8c6 c1e3 e7e6 f1d3 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(50200,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(50200,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(50200,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(50200,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(50200,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(50200,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(50200,36,'cp -35','d7d5 g1f3 c7c5 f1d3 d8b6 e1g1 c5d4 e3d4 b8c6 c2c3 c8f5 d3f5 g6f5 d1e2 e7e6 c1e3 a7a5 f3e5 h7h5 b1d2 b6a6 e2e1 a5a4 e3f2 h8g8 e1d1 f8e7 f2h4 c6e5 f4e5 f6g4 h4e7 g4e3');
INSERT INTO evaluations VALUES(46779,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(46779,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(46779,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(46779,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(46779,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(46779,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(46779,36,'cp -28','g1f3 c7c5 f1d3 b8c6 e1g1 c5d4 e3d4 d8b6 c2c3 c8f5 f3e5 e7e6 d1e2 f5d3 e2d3 f8g7 a2a4 e8g8 d3e2 c6e7 c1e3 e7f5 e3f2 f5d6 f2h4 b6c7 b1d2 f6e4 a4a5 e4d2 e2d2 d6e4 d2e2 f7f6 e5d3 a8c8 g2g4');
INSERT INTO evaluations VALUES(46780,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(46780,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(46780,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(46780,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(46780,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(46780,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(46780,36,'cp -48','c7c5 f1d3 c5d4 e3d4 b8c6 e1g1 d8b6 c2c3 c8f5 f3e5 e7e6 d1e2 f5d3 e2d3 f8d6 b2b3 e8g8 c1e3 b6c7 b1d2 a8c8 d3e2 b7b6 a1c1 d6a3 c1d1 c6e7 c3c4 a3b4 g2g4 b4d2 e3d2 f6e4 d1c1 e4d2 e2d2 d5c4 b3c4');
INSERT INTO evaluations VALUES(50856,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(50856,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(50856,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(50856,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(50856,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(50856,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(50856,36,'cp 32','g1f3 e7e6 f1c4 c7c5 e1g1 g8f6 d4c5 f8c5 d1d8 e8d8 c4e2 d8e7 b1d2 c8d7 d2b3 c5d6 b3a5 d7c6 c1d2 f6e4 a1c1 e4d2 f3d2 h8d8 c1c2 d6b4 a5c6 b8c6 d2c4 a8b8 e2f3 b7b5 f3c6 b5c4 c2c4 d8d3 c4c2 b8c8 f1c1');
INSERT INTO evaluations VALUES(51115,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(51115,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(51115,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(51115,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(51115,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(51115,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(51115,36,'cp 130','c8b7 b5a6 b8a6 c1d2 e7e6 g1f3 a6b4 a1a8 d8a8 f1c4 g8f6 b1c3 f8d6 c3b5 b4d5 e1g1 e8g8 d1a1 b7c6 f1c1 a8a1 c1a1 f8b8 a1a5 f6e4 d2e1 f7f6 g1f1 g8f7 b5d6 e4d6 b2b3 g7g5 h2h3 d5e7 f3d2 d6c4 b3c4 b8b2 a5a7');
INSERT INTO evaluations VALUES(30347,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(30347,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(30347,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(30347,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(30347,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(30347,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(30347,36,'cp 144','b5a6 b8a6 g1f3 g8f6 f1c4 e7e6 d1b3 a6b8 a1a8 b7a8 c4b5 b8d7 f3e5 f8d6 e5c6 d8c8 b1d2 e8g8 d2c4 a8c6 b5c6 c8a6 b3b5 a6b5 c6b5 f8b8 b5a4 d7b6 c4b6 b8b6 e1e2 e6e5 d4e5 d6e5 h1d1 g7g5 b2b3 f6e4 d1d7 g5g4 d7e7 e4c3 e2d3 c3a4 b3a4 e5h2');
INSERT INTO evaluations VALUES(53666,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(53666,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(53666,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(53666,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(53666,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(53666,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(53666,36,'cp 120','g1f3 b8d7 g2g3 f8e7 f1g2 g8f6 e4f6 e7f6 e1g1 e8g8 c1f4 d7b6 f1e1 b6d5 f4e5 f6e5 f3e5 c8d7 d1d2 d7e8 c2c4 d5e7 c4c5 e7f5 e5f3 d8c7 d2c3 a8d8 b2b4 f5e7 a2a4 h7h6 a4a5 e7d5 c3b3 b7b6 f3e5');
INSERT INTO evaluations VALUES(55097,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(55097,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(55097,32,'cp -194',NULL);
INSERT INTO evaluations VALUES(55097,33,'cp -164',NULL);
INSERT INTO evaluations VALUES(55097,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(55097,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(55097,36,'cp -161','h7h5 g4h5 e7e5 d2d3 b8c6 g1f3 h8h5 b1c3 g8f6 a2a3 f8e7 e2e3 c8e6 c3e2 e5e4 f3d4 c6d4 e3d4 h5h8 c1f4 c7c6 d1d2 e4d3 d2d3 f6e4 f1g2 g7g5 f4g3 e6f5 e1c1 e4g3 d3g3 d8d6 g3d6 e7d6 g2f3 h8h4 f3g4 f5g4 h3g4');
INSERT INTO evaluations VALUES(13589,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13589,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13589,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13589,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13589,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13589,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13589,36,'cp 0','e1g1 c5d4 c3d4 f8e7 b1c3 e8g8 f3e5 d8b6 a2a3 f8d8 b2b4 a7a5 c3a4 b6b5 a4c3');
INSERT INTO evaluations VALUES(13588,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(13588,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13588,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13588,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13588,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13588,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13588,36,'cp -10','c5d4 c3d4 f8e7 b1c3 e8g8 f3e5 h7h6 c1f4 c8d7 e5d7 d8d7 a1c1 f8c8 h2h4 e7b4 d1d3 c6a5 f4e5 f6g4 e5f4 a5c4 b2b3 c4d6 c1c2 g4f6 f1c1 b4a3 c1d1 d7c6 f4d6 c6d6 e2e3 d6b6 e3e4 f6e4 c3e4 d5e4 d3e4 c8c2');
INSERT INTO evaluations VALUES(21313,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(21313,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(21313,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(21313,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(21313,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(21313,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(21313,36,'cp 155','c3d4 g8f6 d1b3 d8d7 c4d3 d6d5 e4e5 f6e4 b1c3 c6b4 d3b5 c7c6 b5e2 f8e7 e1g1 e8g8 a2a3 e4c3 b2c3 b4a6 e2d3 d7e6 d3b1 f7f5 e5f6 e6f6 a3a4 e7d6 c1a3 d6a3 a1a3 f6d6 f3e5 a6c5 b3b4 c5e4 b1e4 d6b4 e4d5 c6d5 c3b4 c8e6 f1e1 f8c8 a4a5 e6f5 f2f3');
INSERT INTO evaluations VALUES(21306,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(21306,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(21306,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(21306,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(21306,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(21306,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(21306,36,'cp 154','g8f6 d1b3 d8d7 c4d3 d6d5 e4e5 f6e4 b1c3 c6b4 d3b5 c7c6 b5e2 f8e7 e1g1 a7a5 a2a3 b4a6 e2d3 e4c3 b2c3 a6c7 h2h3 a5a4 b3b2 d7d8 f3e1 e8g8 f2f4 f7f5 e1c2 d8e8 c2e3 e8h5 b2b6 e7d8 c1d2 c7b5 b6c5 b5c7 c5b4 h5f7 b4b1 d8e7 g2g3 g8h8');
INSERT INTO evaluations VALUES(14659,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(14659,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(14659,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(14659,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(14659,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(14659,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(14659,36,'cp 164','d1b3 d8d7 c4d3 d6d5 e4e5 f6e4 b1c3 c6b4 d3b5 c7c6 b5e2 f8e7 e1g1 e8g8 a2a3 b4a6 e2d3 d7e6 c3e2 a6c7 c1e3 c6c5 e2f4 e6d7 d3b1 c5c4 b3c2 d7d8 f3d2 c8f5 f4d5 e4g3 c2c4 g3f1 b1f5 c7d5 a1f1 g7g6 f5b1 d5e3 f2e3 a8c8 c4d3 d8b6');
INSERT INTO evaluations VALUES(22020,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(22020,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(22020,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(22020,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22020,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(22020,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(22020,36,'cp -15','g8f6 c3d4 f6e4 d1b3 d8d7 c4d3 e4f6 e1g1 f8e7 f1e1 e8g8 b1c3 h7h6 c1d2 e7d8 h2h3 a7a5 a2a3 c6e7 d4d5 a5a4 b3c4 c7c5 d5c6 b7c6 d3c2 c8a6 c4a4 a6b7 a4f4 c6c5 b2b4 e7g6 f4f5 d7f5 c2f5 d8b6 g1f1 b7f3 g2f3 d6d5 f5g6 f7g6');
INSERT INTO evaluations VALUES(15634,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(15634,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(15634,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(15634,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(15634,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15634,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15634,36,'cp 0','c3d4 f6e4 d1b3 d8d7 c4d3 e4f6 e1g1 f8e7 f1e1 e8g8 b1c3 h7h6 c1d2 e7d8 h2h3 a7a5 a2a3 a5a4 b3d1 f8e8 e1e8 d7e8 d3b5 e8e6 b5a4 c6a5 a4c2 a5c4 d4d5 f6d5 f3d4 e6e5 d4f3');
INSERT INTO evaluations VALUES(15630,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15630,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(15630,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15630,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15630,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15630,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(15630,36,'cp 0','f6e4 d1b3 d8d7 c4d3 e4f6 e1g1 f8e7 f1e1 e8g8 b1c3 d7d8 h2h3 h7h6 c1e3 d6d5 e3f4 e7d6 f4e5 c8e6 b3b7 c6b4 d3b1 a7a6 a2a3 a8b8 b7a7 b8a8');
INSERT INTO evaluations VALUES(25681,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(25681,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(25681,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(25681,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(25681,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(25681,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(25681,36,'cp 53','g2g4 h5g6 f3e5 f8d6 e5g6 h7g6 d1f3 f7f5 d2d3 d8d7 b1d2 f5g4 f3g4 h8h3 g4d7 e8d7 d2c4 h3h5 c4d6 c7d6 g1g2 g8f6 f1h1 d6d5 f2f3 d5e4 f3e4 h5c5 c2c3 c5h5 c1f4 a8f8 h1f1 f6g4 a1e1 g6g5 f4g3 f8f1 g2f1 h5h1 f1e2 h1e1 g3e1 d7e7 d3d4 e7e6 e2f3 g4f6');
INSERT INTO evaluations VALUES(25677,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(25677,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(25677,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(25677,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(25677,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(25677,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(25677,36,'cp 40','h5g6 f3e5 f8d6 e5g6 h7g6 d1f3 f7f5 e4e5 d6c5 f1d1 d8h4 d2d4 e8c8 c2c3 c5e7 d1d3 c8b8 f3g3 c6c5 b1a3 h4g3 f2g3 h8h3 g4f5 g6f5 d4d5 g7g5 g1g2 h3h8 c1e3 g8h6 a1e1 b8c8 a3c4 h6g4 e3g1 f5f4 g3f4 g5f4 e1e4');
INSERT INTO evaluations VALUES(25676,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(25676,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25676,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(25676,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(25676,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(25676,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(25676,36,'cp 62','f3e5 f8d6 e5g6 h7g6 d1f3 d8d7 d2d3 f7f5 b1d2 f5g4 f3g4 h8h3 g4d7 e8d7 d2c4 h3h5 c4d6 c7d6 g1g2 g8f6 f1h1 d6d5 f2f3 d5e4 f3e4 f6g4 c1f4 a8f8 g2f3 g4h2 f3g3 f8h8 g3g2 h2g4 a2a4 c6c5 b2b3 d7e6 a1f1 h5h1 f1h1 h8h1 g2h1 e6f6 h1g2 g6g5');
INSERT INTO evaluations VALUES(25675,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(25675,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(25675,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(25675,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(25675,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(25675,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(25675,36,'cp 46','f8d6 e5g6 h7g6 d1f3 d8d7 d2d3 f7f5 b1d2 f5g4 f3g4 h8h3 g4d7 e8d7 d2c4 h3h4 c4d6 c7d6 g1g2 g8f6 f1h1 h4h1 g2h1 a8h8 h1g2 h8h5 c1d2 g6g5 a2a4 g5g4 d2c3 d7e6 a1e1 e6f7 c3f6 g7f6 g2g3 h5c5 e1e2 c5g5 d3d4 f7e6 c2c3 g5a5 b2b3 b7b5 g3g4 b5a4');
INSERT INTO evaluations VALUES(32523,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(32523,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(32523,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(32523,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(32523,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(32523,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(32523,36,'cp 131','g1f3 e7e6 c2c4 c7c5 d4d5 f8e7 b1c3 e7f6 d1b3 d8b6 b3c2 g8e7 f1d3 e7g6 e1g1 e8g8 c1d2 a7a5 a1b1 f6d4 a2a3 e6e5 b2b3 b8a6 c3e2 b6d8 f3d4 c5d4 b3b4 a5b4 a3b4 d8e7 g1h1 b7b6 b1e1 f8c8 f2f4 d7g4 h2h3 g4e2 e1e2 e7f6');
INSERT INTO evaluations VALUES(32522,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(32522,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(32522,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(32522,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(32522,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(32522,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(32522,36,'cp 109','b8c6 c2c4 e7e5 d4d5 c6b8 b1c3 a7a5 f1d3 b8a6 a2a3 g8f6 e1g1 g7g6 a1b1 f8g7 b2b4 e8g8 c4c5 a5b4 a3b4 d6c5 d3a6 c5b4 a6b7 b4c3 b7a8 d8a8 f3e5 d7a4 d1e2 f8e8 f2f4 f6d7 e2f2 c3c2 b1b2 e8d8 e5d7 d8d7 b2c2 a4c2 f2c2');
INSERT INTO evaluations VALUES(39114,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(39114,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(39114,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(39114,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(39114,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(39114,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(39114,36,'cp 147','g1f3 e5f4 d2d4 d6e7 c1f4 g8f6 b1c3 e8g8 e4e5 f6h5 f4e3 d7d6 f1d3 b8c6 d1e2 c6b4 d3c4 c7c6 a2a3 d6d5 a3b4 d5c4 f3d2 g7g6 e1g1 c8e6 d2e4 f7f5 e5f6 e7f6 g2g4 h5g7 h2h3 f6e7 f1f8 d8f8 e3g5');
INSERT INTO evaluations VALUES(39113,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(39113,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(39113,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(39113,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(39113,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(39113,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(39113,36,'cp 121','e5f4 d2d4 c7c6 e4e5 d6c7 c1f4 d7d6 f1c4 d6e5 f3e5 c7e5 f4e5 g8f6 e1g1 b8d7 e5d6 d7b6 d1e1 c8e6 c4e6 d8d6 e6b3 e8d7 e1h4 a8e8 b1d2 d7c7 c2c3 h7h6 b3f7 e8e7 f7g6 f6d5 f1f7 h8e8 a1e1 d6g6 f7e7 d5e7 e1e7 e8e7 h4e7 b6d7 e7e4 g6g5 e4e2 a7a5 h2h3 a5a4');
INSERT INTO evaluations VALUES(42081,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(42081,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(42081,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(42081,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(42081,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(42081,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(42081,36,'cp 55','b1d2 d8e7 f1e2 e8g8 a1c1 b8d7 e1g1 c7c5 d4c5 d7c5 b2b4 c5e4 c4c5 e4d2 d1d2 c8d7 d2b2 f8c8 h2h3 d7e8 f3d4 e8d7 a2a3 h7h6 d4f3 d7e8 f3e5 a7a6 f1e1 c8c7 e2d3 f6d7 e3e4 d5e4 e1e4 d7e5 b2e5 a8d8');
INSERT INTO evaluations VALUES(49414,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(49414,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(49414,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(49414,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(49414,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(49414,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(49414,36,'cp -29','d7d5 g1f3 c7c5 d4c5 f8c5 f1d3 b8c6 e1g1 e8g8 e3e4 d5d4 c3e2 e6e5 h2h3 h7h6 a2a3 a7a5 c2c3 d4c3 e2c3 c8e6 d1e2 a8c8 f1d1 c6d4 f3d4 c5d4 c1e3 d4c3 b2c3 c8c3 a1c1 d8c7 e2d2 f8c8 d2b2 f6h5 e3d2 c3c1 d1c1 c7d6 c1c8 e6c8 d3e2 h5f4 d2a5 f4e2 b2e2 d6a3');
INSERT INTO evaluations VALUES(50975,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(50975,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(50975,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(50975,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(50975,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(50975,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(50975,36,'cp 21','g1f3 g8f6 e1g1 c7c5 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 b2b3 b8d7 c4e2 b7b6 f3e1 a6a5 c1b2 h8d8 a1c1 e7f8 d2c4 f6d5 e1d3 c5e7 c4e5 d7e5 d3e5 d5b4 b2d4 f7f6 d4b6 f6e5 b6d8 e7d8');
INSERT INTO evaluations VALUES(50976,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50976,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(50976,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50976,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(50976,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(50976,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50976,36,'cp 29','c7c5 e1g1 g8f6 b2b3 b8d7 c1b2 f8e7 d4c5 e7c5 c4d3 e8g8 b1d2 b7b5 a2a4 b5a4 a1a4 a6a5 d1a1 c5e7 d3e2 c8b7 a4a2 e7b4 f1c1 h7h6 b2d4 e6e5 d4e5 d7e5 a1e5 f8e8 e5a1 a8c8 c1c4 d8e7 a2c2 c8c4 e2c4 e8c8');
INSERT INTO evaluations VALUES(55311,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(55311,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(55311,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(55311,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(55311,34,'cp -113',NULL);
INSERT INTO evaluations VALUES(55311,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(55311,36,'cp -102','e7e5 c2c3 f8d6 g1f3 g8f6 c3d4 e5d4 f3d4 f6e4 a4c3 d6c5 c3e4 c5d4 f1c4 e8g8 d2d3 b8c6 e1g1 c6e5 f1e1 b7b6 d1b3 e5c4 d3c4 f8e8 c1g5 d8d7 b3b5 e8e5 b5d7 c8d7 g5f4 e5e6 e4c3 e6e1 a1e1 c7c5 f4d2 d7e6 b2b3 a8d8 c3e2 d4e5 e2c3 e5f6 c3e4 f6d4');
INSERT INTO evaluations VALUES(56637,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(56637,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(56637,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(56637,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(56637,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(56637,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(56637,36,'cp -90','c2c3 b8c6 b2b4 a7a6 b4b5 a6b5 f1b5 g8f6 g1f3 c8d7 e1g1 f8d6 f1e1 e8g8 a4b2 d4c3 d2c3 c6a5 a2a4 h7h6 d1c2 d8e7 h2h3 f8d8 c1e3 d7e6 f3d2 f6d7 b2c4 a5c4 b5c4 d6c5 c4e6 e7e6 d2b3 c5a7 c3c4 a7e3 e1e3');
INSERT INTO evaluations VALUES(4179,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(4179,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(4179,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(4179,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(4179,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(4179,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(4179,36,'cp 128','d2d4 e5d4 c3d4 c5b4 b1c3 d7d5 e4d5 c6a5 c4f1 c7c6 d5c6 a5c6 a2a3 b4d6 g2g3 c8g4 c1e3 d8d7 f1g2 f8d8 d1a4 a8c8 a1c1 g4e6 e1d1 d6c7 f3e5 c6e5 d4e5 d7a4 c3a4 c7e5 c1c8 d8c8 g2b7 c8c2 b2b4 c2a2 a4c5 a2a3 d1d8 g8h7 c5e6');
INSERT INTO evaluations VALUES(4203,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(4203,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(4203,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(4203,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(4203,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(4203,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(4203,36,'cp 140','e5d4 c3d4 c5b4 b1c3 d7d5 e4d5 c6a5 c4f1 c7c6 d5c6 a5c6 a2a3 b4d6 d1d3 d6c7 h2h3 c8e6 c1e3 a8c8 a1d1 e6b3 d1c1 c7b8 d3b5 c6a5 f3d2 a7a6 b5d3 b3c4 d3f5 c4e6 f5f3 f6d5 c3d5 d8d5 f3d5 e6d5');
INSERT INTO evaluations VALUES(4213,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(4213,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(4213,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(4213,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(4213,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(4213,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(4213,36,'cp 152','c3d4 c5b4 b1c3 d7d5 e4d5 c6a5 c4f1 c7c6 d5c6 a5c6 a2a3 b4d6 g2g3 c8g4 c1e3 c6e7 f1g2 e7d5 e3d2 g4e6 f3e5 a8c8 d1a4 d6e5 d4e5 d5c3 d2c3 f6d5 c3a5 b7b6 a5d2 d8e8 a4e8 f8e8 a1d1 c8c2 d2h6 g7h6 g2d5 c2b2 d5e6 e8e6 f2f4 b2a2');
INSERT INTO evaluations VALUES(4211,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(4211,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(4211,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(4211,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(4211,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(4211,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(4211,36,'cp 145','c5b4 b1c3 d7d5 e4d5 c6a5 c4f1 c7c6 d5c6 a5c6 a2a3 b4d6 g2g3 d6c7 d4d5 c6e7 d1b3 c8g4 f3e5 c7e5 e1e5 d8d7 b3b5 d7b5 f1b5 e7f5 c1e3 f8d8 a1e1 f5e3 e1e3 g8f8 f2f3 g4d7 b5d7 f6d7 e5e4 a8c8 e4d4 d7b6 d5d6 c8c6 e3d3 b6c4 c3e4');
INSERT INTO evaluations VALUES(4229,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(4229,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(4229,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(4229,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(4229,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(4229,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(4229,36,'cp 144','b1c3 d7d5 e4d5 c6a5 c4f1 c7c6 d5c6 a5c6 a2a3 b4d6 h2h3 c8e6 c1d2 a7a6 a1c1 d6c7 d2e3 f8e8 c3a4 e6d5 a4c5 a8b8 g2g3 a6a5 f1g2 c6e7 c5d3 e7f5 e3f4 e8e1 d1e1 c7f4 d3f4 d5c6 e1d1 d8d7 f3e5');
INSERT INTO evaluations VALUES(4228,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(4228,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(4228,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(4228,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(4228,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(4228,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(4228,36,'cp 142','d7d5 e4d5 c6a5 c4f1 c7c6 d5c6 a5c6 a2a3 b4d6 g2g3 c8g4 c1e3 d8d7 d4d5 c6e7 e3d4 f6d5 c3d5 e7d5 d4g7 g8g7 d1d5 a8d8 f1g2 d6c7 d5b7 c7b6 b7d7 g4d7 a1c1 f8e8 e1e8 d8e8 g2f1 d7f5 f3d2 e8d8 d2c4 b6d4 c1e1 f5e6 b2b4');
INSERT INTO evaluations VALUES(10585,30,'cp 548',NULL);
INSERT INTO evaluations VALUES(10585,31,'cp 553',NULL);
INSERT INTO evaluations VALUES(10585,32,'cp 544',NULL);
INSERT INTO evaluations VALUES(10585,33,'cp 546',NULL);
INSERT INTO evaluations VALUES(10585,34,'cp 554',NULL);
INSERT INTO evaluations VALUES(10585,35,'cp 563',NULL);
INSERT INTO evaluations VALUES(10585,36,'cp 553','d1d5 c5f2 e1e2 e8g8 d5e4 d7d5 c4d5 f2b6 h1d1 d8e8 c1e3 c6e7 d5b3 c8f5 e4f4 a7a5 b1c3 b6e3 e2e3 f5d7 e3f2 a5a4 d1d7 a4b3 d7c7 b3a2 f4e4 e7g6 c7b7 e8e6 c3d5 e6a6 d5c7 a6a4 e4a4 a8a4 c7d5 f8c8 d5c3 a4a5 h2h4 g6f8 a1a2 a5a2 c3a2 c8c2 f2g3');
INSERT INTO evaluations VALUES(10650,30,'cp 544',NULL);
INSERT INTO evaluations VALUES(10650,31,'cp 528',NULL);
INSERT INTO evaluations VALUES(10650,32,'cp 561',NULL);
INSERT INTO evaluations VALUES(10650,33,'cp 541',NULL);
INSERT INTO evaluations VALUES(10650,34,'cp 544',NULL);
INSERT INTO evaluations VALUES(10650,35,'cp 536',NULL);
INSERT INTO evaluations VALUES(10650,36,'cp 536','c5f2 e1e2 e8g8 d5e4 d7d5 c4d5 f2b6 h1d1 d8e8 c1e3 c6e7 d5b3 c8f5 e4f4 h7h6 b1c3 e8c6 e3b6 c6b6 f4e3 b6e3 e2e3 a7a6 e5e6 f7e6 d1d7 a8e8 a1e1 b7b5 e3d2 e7g6 d7c7 f5g4 d2c1 g4f3 g2f3 g6h4 e1e6 e8e6 b3e6 g8h7 a2a4 b5b4 c3d5 h4f3 d5b4 a6a5 b4d3 f3h2');
INSERT INTO evaluations VALUES(10656,30,'cp 544',NULL);
INSERT INTO evaluations VALUES(10656,31,'cp 560',NULL);
INSERT INTO evaluations VALUES(10656,32,'cp 557',NULL);
INSERT INTO evaluations VALUES(10656,33,'cp 559',NULL);
INSERT INTO evaluations VALUES(10656,34,'cp 565',NULL);
INSERT INTO evaluations VALUES(10656,35,'cp 561',NULL);
INSERT INTO evaluations VALUES(10656,36,'cp 567','e1e2 e8g8 d5e4 d7d5 c4d5 f2b6 h1d1 d8e8 c1e3 b6e3 e2e3 c6e7 d5b3 c8f5 e4f4 e8b5 e3f2 b5b6 f4e3 b6e3 f2e3 a7a6 b1d2 a8d8 e3f2 b7b5 d2f1 c7c5 c2c4 f5g4 d1d6 d8b8 f3d2 g4e6 f1e3 e7f5 e3f5 e6f5 d2f3 g7g6 a1d1');
INSERT INTO evaluations VALUES(16588,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16588,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(16588,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16588,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16588,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16588,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16588,36,'cp 0','c2c4 f8c5 b2b4 c5d4 a1a2 d7d6 g1f3 d4b6 f1e2 c8g4 b1c3 g4f3 e2f3 c6d4 g2g3 c7c6 f3g2 d4e6 h2h4 h7h6 g2h3 a7a5 b4b5 e8g8 e1g1 b6d4 c3e2 d4c5 e2c3');
INSERT INTO evaluations VALUES(16439,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16439,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16439,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16439,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16439,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16439,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16439,36,'cp -8','f8c5 b2b4 c5d4 a1a2 d7d6 g1f3 c8g4 f1e2 d4b6 b1c3 g4f3 e2f3 c6d4 c3a4 c7c6 a4b6 d8b6 c1d2 d6d5 e4d5 c6d5 e1g1 e8g8 d2g5 a7a5 b4a5 a8a5 g5f6 d4f3 d1f3 b6f6 f1b1 f6c6 a2b2 d5c4 f3c6 b7c6 d3c4 a5a3 b2b6 e5e4 b6c6 a3c3 c6c7 f7f5 h2h3 f8d8 g1h2');
INSERT INTO evaluations VALUES(18616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18616,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18616,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18616,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18616,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18616,36,'cp 0','f8g7 c3b5 d7d6 c1f4 e6e5 f4g5 g8e7 b5c3 h7h6 g5e7 c6e7 c3d5 c8d7 a2a4 e7d5 c4d5 d7c6 d5c6 b7c6 e1g1 e8g8 d1e2 d8d7 b2b3 a8e8 a4a5 a7a6 g2g3 d6d5 g1g2 f7f5 a1a4 f8f7 f3h4 g8h7 h4f3');
INSERT INTO evaluations VALUES(34716,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(34716,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(34716,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(34716,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(34716,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(34716,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(34716,36,'cp 82','c8e6 c4e6 f7e6 d2d4 e5d4 d1d4 b8c6 d4c4 e8c8 e1g1 g8f6 c4a4 e7e8 f1d1 d6d5 c1g5 d5e4 d1d8 c6d8 a4e8 f6e8 c3e4 d8c6 a1e1 h7h6 g5h4 g7g5 h4g3 h8g8 g1f1 f8g7 c2c3 b7b6 h2h4 g5h4 g3h4 g8f8 h4g3 c6e7 e4d2 c8d7 g3h4');
INSERT INTO evaluations VALUES(32910,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(32910,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(32910,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(32910,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(32910,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(32910,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(32910,36,'cp 76','d1e2 b8c6 e1g1 e7d7 c3d5 g8f6 c2c3 c6e7 d2d4 f6d5 e4d5 e7d5 d4e5 e8c8 c4b3 d6e5 f1d1 e5e4 e2e4 d5f6 d1d7 f6e4 d7d8 c8d8 b3e6 f7e6 c1e3 f8c5 a1e1 c5e3 e1e3 e4c5 b2b4 c5a4 g2g3 h8e8 h2h4 d8d7 f3g5 a4b6');
INSERT INTO evaluations VALUES(35197,30,'cp 280',NULL);
INSERT INTO evaluations VALUES(35197,31,'cp 287',NULL);
INSERT INTO evaluations VALUES(35197,32,'cp 274',NULL);
INSERT INTO evaluations VALUES(35197,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(35197,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(35197,35,'cp 300',NULL);
INSERT INTO evaluations VALUES(35197,36,'cp 260','g1f3 h4e7 c3d5 e7d8 f3e5 d7d6 e5f3 c7c6 d5c3 d6d5 e4e5 f8b4 c3e2 f7f6 c2c3 b4a5 d2d4 c8g4 h2h3 g4f3 g2f3 f6e5 d4e5 b8d7 f3f4 g8h6 e2d4 e8g8 h3h4 a5b6 c1e3 d7c5 d1h5 c5e4 d4e6 d8d7 e6f8 a8f8 e1c1 b6e3 f2e3');
INSERT INTO evaluations VALUES(35195,30,'cp 248',NULL);
INSERT INTO evaluations VALUES(35195,31,'cp 286',NULL);
INSERT INTO evaluations VALUES(35195,32,'cp 272',NULL);
INSERT INTO evaluations VALUES(35195,33,'cp 264',NULL);
INSERT INTO evaluations VALUES(35195,34,'cp 270',NULL);
INSERT INTO evaluations VALUES(35195,35,'cp 277',NULL);
INSERT INTO evaluations VALUES(35195,36,'cp 281','h4e7 d2d4 d7d6 c3d5 e7d7 d4e5 c7c6 a2a4 g8e7 d5e7 f8e7 f1c4 e8g8 e1g1 d6e5 f3e5 d7e8 f1e1 b8d7 e5d3 d7c5 c1e3 c5e4 d1f3 e4d6 e3c5 c8e6 c4e6 f7e6 f3g4 d6f5 e1e6 e8f7 c5e7 f7e6 e7f8 a8f8 h2h3 e6d6 a1d1 h7h6 g4c4 g8h7 d1e1 h7h8');
INSERT INTO evaluations VALUES(44231,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(44231,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(44231,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(44231,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(44231,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(44231,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(44231,36,'cp -8','b7b6 b1d2 c8b7 f1d3 d7d5 e1g1 f8e7 d1e2 e8g8 b2b3 b8d7 c1b2 a8c8 f1d1 e7d6 c3c4 c5d4 f3d4 d7e5 d3c2 b7a6 a1c1 d8e7 d4b5 a6b5 c4b5 c8c7 h2h3 f8c8 c2b1 e5g6 c1c7 c8c7 d2f3 e6e5 g1h1 h7h6');
INSERT INTO evaluations VALUES(45615,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(45615,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(45615,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(45615,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(45615,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(45615,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(45615,36,'cp -43','d5e4 g1e2 f8e7 g5f6 e7f6 c3e4 f6e7 g2g3 e8g8 f1h3 d8a5 e4c3 b8d7 e1g1 e6e5 d4d5 d7f6 h3g2 a5c7 d1d3 c8g4 h2h3 g4d7 a1d1 c6d5 c4d5 e7c5 g3g4 h7h6 d5d6 c7b6 c3d5 f6d5 d3d5 d7c6 d5e5 f8e8');
INSERT INTO evaluations VALUES(45692,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(45692,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(45692,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(45692,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(45692,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(45692,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(45692,36,'cp -13','c3e4 f8b4 e4c3 c6c5 d4c5 b4c3 b2c3 d8a5 g1f3 f6e4 g5e3 a5c3 f3d2 e4d2 e3d2 c3e5 d2e3 c8d7 f1e2 e5c3 e3d2 c3f6 e1g1 d7c6 d1b3 e8g8 b3g3 b8d7 d2b4 a7a5 b4a3 f6e5 g3e5 d7e5 f1d1 f7f6 d1d6 g8f7 f2f4 e5g6 e2h5 f8d8');
INSERT INTO evaluations VALUES(45691,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(45691,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(45691,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(45691,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45691,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(45691,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(45691,36,'cp -37','f8b4 e4c3 c6c5 d4c5 d8d1 a1d1 b4c3 b2c3 b8d7 g5e3 f6g4 e3g5 h7h6 g5c1 d7c5 f1e2 g4f6 g1f3 b7b6 f3e5 c8d7 c1a3 d7a4 d1d4 a8c8 f2f3 f6d7 e5d7 a4d7 e1f2 d7a4 d4d2 e8e7 h1b1 h8d8 d2d8 e7d8');
INSERT INTO evaluations VALUES(47228,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(47228,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(47228,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47228,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47228,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47228,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(47228,36,'cp 0','e4e5 f6d7 f2f4 e7e6 c2c3 c7c5 d3c2 d5d4 g1f3 f8e7 e1g1 e8g8 b1a3 b8c6 c2e4 d8c7 a3b5 c7b6 b5a3');
INSERT INTO evaluations VALUES(47294,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47294,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47294,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47294,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47294,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47294,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47294,36,'cp 0','f6d7 f2f4 e7e6 g1f3 c7c5 c2c3 f8e7 d3c2 d5d4 b1a3 b8c6 c2e4 e8g8 e1g1 d8c7 a3b5 c7b6 b5a3');
INSERT INTO evaluations VALUES(52166,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(52166,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(52166,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(52166,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(52166,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(52166,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(52166,36,'cp 201','g1e2 d7d6 a2a4 g8f6 e2c3 f8e7 b1a3 h7h6 a3c4 e8g8 f1e2 f6e8 e1g1 e7g5 c1g5 d8g5 d1c1 g5e7 a4a5 b8a6 f2f4 e5f4 c1f4 a6b4 e2d1 c8d7 f4g3 e7g5 g3f2 f7f5 e4e5 d6e5');
INSERT INTO evaluations VALUES(55267,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(55267,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(55267,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(55267,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(55267,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(55267,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(55267,36,'cp 125','c2c3 c7c5 c3d4 c5d4 f1b5 c8d7 f3d4 e7e5 b5d7 d8d7 d4f3 b8c6 e1g1 a8d8 d1b3 f8c5 d2d3 g8e7 b1c3 e8g8 c1g5 h7h6 g5e7 d7e7 c3d5 e7d6 a1c1 b7b6 b3a4 g8h8 a2a3 c6d4 f3d4 c5d4 b2b4 f7f5 c1c7 a7a5 c7c6 d6d7 b4b5 h8h7 a4b3 f5e4 d3e4');
INSERT INTO evaluations VALUES(55735,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(55735,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(55735,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(55735,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(55735,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(55735,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(55735,36,'cp 93','g8f6 g1f3 c7c5 c2c3 b8d7 e1g1 f8e7 f1e1 e8g8 a2a4 f8e8 b1a3 c5d4 c3d4 h7h6 h2h3 a7a6 a3b1 e6e5 d4d5 d7c5 d3c2 a6a5 b1c3 b7b6 c1e3 c8a6 e3c5 b6c5 c3b5 a8c8 f3d2 c5c4 b5a7 c8c7 a7c6');
INSERT INTO evaluations VALUES(48058,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(48058,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(48058,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(48058,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(48058,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(48058,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(48058,36,'cp 92','g1f3 c7c5 c2c3 b8d7 e1g1 f8e7 f1e1 e8g8 h2h3 d8c7 a2a4 b7b6 b1a3 a7a6 d4d5 e6d5 e4d5 f8e8 a3c4 e7f8 e1e8 f6e8 c1f4 g7g6 c4e3 c8b7 d1d2 e8f6 c3c4 a8e8 e3g4 f8g7 a1d1 f6g4 h3g4 d7f6 g4g5 f6e4 d3e4 e8e4');
INSERT INTO evaluations VALUES(56228,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(56228,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(56228,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(56228,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(56228,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(56228,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(56228,36,'cp -21','c3c4 c7c6 c1d2 g8f6 b1c3 f8e7 c4d5 c6d5 g1e2 b8c6 e2f4 e8g8 a2a3 c8e6 f4e6 f7e6 f1e2 e7d6 e1g1 g8h8 d2e1 c6e7 c3a4 e7f5 e1b4 d6b4 a3b4 d8e7 b4b5 e7f7 a1c1 h7h5 b5b6 a7b6 a4b6 a8d8 b6a4 g7g5');
INSERT INTO evaluations VALUES(56184,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(56184,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(56184,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(56184,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(56184,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(56184,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(56184,36,'cp -36','c7c6 b1c3 g8f6 c1d2 f8e7 g1e2 e8g8 c4d5 c6d5 e2f4 b8c6 a2a3 c8e6 f1e2 e7d6 f4e6 f7e6 e1g1 a8c8 d2e1 c6e7 d1b3 d8d7 b3a4 d7a4 c3a4 g8f7 e1b4 d6b4 a3b4 c8c2 e2d1 c2c4 b4b5 c4b4 a4c5 f8b8 a1a7');
INSERT INTO evaluations VALUES(517,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(517,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(517,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(517,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(517,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(517,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(517,36,'cp 198','e5d3 c5d6 f2f3 f6d7 d3f2 a7a5 d2d3 e8g8 c3e2 d4a7 e2g3 a5a4 a2a3 d7b6 e1g1 f7f5 g1h1 f5f4 g3e2 a7a5 c1d2 a5h5 d1c1 h5h4 c2c4 c6c5 d2c3 h4h6 c1d2 c8e6 a1e1 a8e8');
INSERT INTO evaluations VALUES(521,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(521,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(521,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(521,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(521,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(521,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(521,36,'cp 181','c5d6 f2f3 f6d7 d3f2 a7a5 a2a4 e8g8 c3e2 d4a7 b2b3 b7b5 e1g1 f7f5 e4f5 b5a4 a1a4 d7b6 a4a1 c8f5 e2g3 f5g6 g3e4 b6d5 c1b2 a8e8 g1h1 a7b6 f1e1 d5f4 d2d3 f4d5 e4d6 c7d6');
INSERT INTO evaluations VALUES(16075,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(16075,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(16075,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(16075,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(16075,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(16075,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(16075,36,'cp 1','g1e2 c6a5 d2d3 a5c4 d3c4 d7d6 h2h4 f8g7 b1c3 c7c6 f3d3 c8e6 c1d2 d8d7 f2f3 f7f5 e1c1 h6f7 b2b3 h7h5 d2e3 d7e7 c1b1 f5f4 e3f2 e8g8 d1g1 g8h7 a2a4 g7f6 g1d1 e7c7 g2g3 f4g3 f2g3 f6e7 b1a1');
INSERT INTO evaluations VALUES(33759,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(33759,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(33759,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(33759,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(33759,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(33759,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(33759,36,'cp -12','c8e6 c2c3 b8c6 d2d4 e8c8 c1e3 a7a6 f3d5 e6d5 g1f3 d5f3 g2f3 e7e5 f1h3 c8b8 d4e5 c6e5 e1e2 f8d6 a1d1 e5g6 d1d4 h8e8 h1d1 e8e5 h3g2 d8e8 e2f1 e5h5 h2h3 f7f5 f3f4 h5h4 d4d5 g6f4 d5f5 f4g2 f1g2 b8c8 f5f7 e8e7 f7e7 d6e7 d1e1 e7f6 e1d1 h4h5');
INSERT INTO evaluations VALUES(44963,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(44963,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(44963,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44963,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(44963,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(44963,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(44963,36,'cp -22','d5e4 c3e4 b8d7 e4f6 d7f6 c2c3 e7e6 f1d3 f8e7 g1e2 c6c5 c1e3 c5d4 e2d4 c8d7 e3f2 e8g8 e1g1 d8c7 f2g3 c7b6 d1b3 f8d8 b3b6 a7b6 g3f2 d7e8 f1e1 h7h6 d3c2 e8a4 c2a4 a8a4 d4b5 e7c5');
INSERT INTO evaluations VALUES(45043,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(45043,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(45043,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(45043,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(45043,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(45043,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(45043,36,'cp -12','c3e4 b8d7 e4f6 d7f6 c2c3 e7e6 f1d3 f8e7 g1e2 c6c5 c1e3 c5d4 e2d4 c8d7 e1g1 e8g8 g1h1 d8c7 a2a4 a7a6 e3f2 a8d8 a4a5 h7h6 d1c2 e7c5 c2e2 c5d6 f2g1 f6d5 e2f2 d5f4 d3c2 f8e8 f1e1 f4d5 d4e2 e6e5');
INSERT INTO evaluations VALUES(47218,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(47218,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(47218,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(47218,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(47218,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(47218,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(47218,36,'cp -94','d5e4 b1c3 e4f3 g1f3 c7c6 f1d3 c8g4 c1e3 e7e6 d1e2 f8b4 h2h3 g4h5 a2a3 b4c3 b2c3 e8g8 e1g1 c6c5 e2f2 c5d4 c3d4 h5g6 d3g6 h7g6 c2c4 b8d7 a1c1 f6e4 f2e1 g6g5 e1b4 d7b6 f3e5 f7f6 e5g6');
INSERT INTO evaluations VALUES(15655,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(15655,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(15655,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15655,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(15655,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(15655,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(15655,36,'cp 49','g1e2 g7g6 h2h3 f8g7 g2g4 e8g8 f1g2 a8b8 a2a4 a7a6 f2f4 f8e8 e1g1 f6d7 g4g5 b7b5 a4b5 a6b5 c3d5 c8b7 h3h4 c6e7 d5e3 e7c8 e2g3 d7c5 e3d5 b7d5 d1d5 c8b6 d5d1 b6c4 a1b1 c5a4 c2c3 d8d7 e4e5 b5b4 c3b4 b8b4 b2b3 a4c5 h4h5 d6e5 d1d7 c5d7');
INSERT INTO evaluations VALUES(15693,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(15693,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(15693,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(15693,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(15693,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(15693,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(15693,36,'cp 106','f6d5 d2d4 c6c5 f1e2 c8f5 c2c3 f8e7 e1g1 c7c6 e2c4 e8g8 d4c5 e7c5 b1d2 f8e8 f1e1 d8h4 d1f3 f5g4 e1e4 d5f6 f3f4 f6d5 f4g3 h4g3 h2g3 g4f5 e4e1 a8d8 d2b3 c5b6 a2a4 a7a5 c1g5 d8d7 g5d2 h7h6 a1d1 f5e6 b3d4 e6g4 f2f3 g4e6 b2b3');
INSERT INTO evaluations VALUES(17618,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(17618,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(17618,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(17618,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(17618,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(17618,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(17618,36,'cp 16','b5c6 b7c6 c2c3 c5d4 c3d4 c6c5 e1g1 c8d7 b1c3 c5d4 c3e2 g8e7 e2d4 e7g6 g2g3 d8b8 h2h4 f8b4 d1e2 b4c5 f1e1 h7h6 h4h5 g6e7 c1e3 e8g8 a1c1 f8c8 g3g4 c5d4 e3d4 b8b4 a2a3 b4b3 d4c5 e7c6 e2e3 b3a4 f3d4 c6d4 c5d4 d7b5');
INSERT INTO evaluations VALUES(17609,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(17609,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(17609,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17609,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(17609,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(17609,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(17609,36,'cp 11','b7c6 c2c3 c5d4 c3d4 c8d7 e1g1 c6c5 b1c3 c5d4 c3e2 g8e7 e2d4 e7g6 f1e1 h7h6 g2g3 a8c8 h2h4 f8c5 c1e3 e8g8 a1c1 c5d4 d1d4 c8c1 e1c1 d8b8 h4h5 g6e7 d4d2 f8c8 e3c5 e7c6 c5d6 b8b5 b2b3 d5d4 f3d4 c6d4');
INSERT INTO evaluations VALUES(18236,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(18236,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18236,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(18236,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18236,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(18236,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(18236,36,'cp 25','c2c3 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 c6c5 c1g5 d8c7 d4c5 h7h6 g5d2 f8c5 c3a4 c5e7 g2g4 f5h4 f3h4 e7h4 a1c1 c7b7 d2e3 e8g8 f2f4 h4e7 f4f5 d5d4 d1d4 f8d8 d4c4 e6f5 g4f5 e7g5 e3g5 h6g5 c4c7 d8d2 c7b7');
INSERT INTO evaluations VALUES(18233,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(18233,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(18233,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(18233,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(18233,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(18233,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18233,36,'cp 25','c5d4 c3d4 c6c5 e1g1 c8d7 b1c3 c5d4 c3e2 g8e7 e2d4 e7g6 g2g3 d8b8 h2h4 f8b4 a2a3 b4a5 d1e2 a5b6 f1d1 h7h5 c1e3 g6e5 f3e5 b8e5 d4f3 e5d6 e3b6 d6b6 d1d5 d7b5 e2e5 e8g8 d5c5 b5c6 f3d4 c6b7 b2b4 f8d8 a1c1 b6d6 e5d6 d8d6 c5c7 a8d8 c7b7 d6d4');
INSERT INTO evaluations VALUES(18240,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(18240,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(18240,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(18240,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(18240,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(18240,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(18240,36,'cp 18','c3d4 g8e7 b1c3 e7f5 e1g1 c6c5 c1g5 d8c7 d4c5 h7h6 g5f4 c7c5 c3e2 f8e7 e2d4 f5d4 f3d4 c8d7 a1c1 c5b4 f4e3 b4b8 d1h5 e8g8 c1c3 f7f6 e3f4 f8f7 d4c6 d7c6 c3c6 b8b4 f4e3 b4b2 c6e6 f6e5 e6e5 b2a2');
INSERT INTO evaluations VALUES(18238,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(18238,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(18238,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(18238,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18238,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(18238,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(18238,36,'cp 18','g8e7 e1g1 e7f5 d1c2 f8e7 b2b3 c8d7 c1a3 e7a3 b1a3 d8e7 a3b1 e8g8 f1c1 f8b8 c2c3 a6a5 b1d2 a5a4 a1b1 a4b3 d2b3 e7a3 g2g4 f5e7 b3c5 a3c3 b1b8 a8b8 c1c3 d7e8 c3b3 b8b3 a2b3 h7h6');
INSERT INTO evaluations VALUES(37217,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(37217,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(37217,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(37217,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(37217,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(37217,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(37217,36,'cp 19','g1f3 c7c6 e1g1 d7d6 c2c3 h7h6 g5h4 c5b6 a2a4 c8g4 b1d2 b8d7 h2h3 g4h5 f1e1 f8e8 d1b3 g7g5 h4g3 d7c5 b3c2 a7a5 b2b4 a5b4 c3b4 c5e6 c4e6 e8e6 a4a5 b6a7 c2c3 h5g6 a1d1 e6e8 g3h2 g5g4 h3g4 f6g4 h2g3 d6d5 f3e5');
INSERT INTO evaluations VALUES(38598,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(38598,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(38598,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(38598,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(38598,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(38598,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(38598,36,'cp 148','b1c3 g8f6 d1d2 b8c6 e1c1 c8g4 d4d5 g4f3 g2f3 c6e5 f1b5 f6d7 b5e2 d8e7 f4g5 f7f6 g5e3 e8c8 c3b5 a7a6 b5d4 g7g5 h2h4 d8g8 h4h5 h7h6 c1b1 g8e8 d4f5 e7d8 e3b6 d7b6 f5d4 c8b8 d2a5 f6f5 d4f5 h8f8 f5d4 d8e7 d4e6 e5f3 e6f8 e7f8 e2f3');
INSERT INTO evaluations VALUES(38597,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(38597,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(38597,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(38597,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(38597,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(38597,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(38597,36,'cp 171','b8c6 f1b5 a7a6 b5a4 c8d7 d1d3 g8e7 e1c1 c6b4 d3c4 b4c6 h2h4 b6a7 c4e2 b7b5 a4b3 c6a5 b3d5 c7c6 d5b3 e8g8 h4h5 h7h6 f4d6 d7g4 e4e5 d8d7 h1e1 f8e8 c1b1 a5b3 a2b3 a6a5 e2e4 e7d5 c3d5 c6d5 e4d5');
INSERT INTO evaluations VALUES(40836,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40836,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(40836,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(40836,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(40836,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(40836,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(40836,36,'cp 44','e8g8 d1d2 b8c6 d4c6 b7c6 e1c1 f6g4 e3f4 e7f6 f4g3 a8b8 h2h3 g4e5 f2f4 e5d7 e4e5 d6e5 f4e5 f6g5 g3f4 g5f4 d2f4 d8e7 f1c4 e7e5 h1f1 e5f4 f1f4 d7e5 f4e4 e5c4 e4c4 f8e8 b2b3 b8b6 c1b2 g8f8 c3a4 b6b5 c4c6 c8b7 c6c7 b7g2 d1d7 b5f5 h3h4 h7h5 c7a7 g7g5 h4g5 h5h4 a4b6 f5f2 b6c4');
INSERT INTO evaluations VALUES(36256,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36256,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(36256,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(36256,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(36256,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(36256,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36256,36,'cp 26','d1d2 b8c6 d4c6 b7c6 e1c1 f6g4 e3f4 e7f6 f4g3 a8b8 d1e1 f8e8 f2f4 g7g5 h2h3 g5f4 g3f4 g4e5 f1e2 c8e6 e1f1 g8h8 g2g4 e8g8 h1g1 f6g5 b2b3 a7a5 f4g5 d8g5 d2g5 g8g5 f1f2 h8g7 g1g3 e5g6 c1d2 b8e8 e2d3');
INSERT INTO evaluations VALUES(49576,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(49576,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(49576,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(49576,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(49576,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(49576,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(49576,36,'cp -18','d7d5 d2d4 e5d4 e3d4 c7c6 c4d5 f6d5 f1d3 f8d6 g1f3 d8e7 d3e2 e8g8 e1g1 c8e6 f1e1 e7c7 b1c3 b8d7 c3e4 f8e8 e4d6 c7d6 c1d2 d7f8 e2f1 e6g4 d1b3 g4f3 b3f3 f8e6 d2c3 a8d8 e1e5 d5f6 a1d1 h7h6 f3g3 f6d5 h2h3 e6f4 g3g4 d5c3 e5e8 d8e8 b2c3');
INSERT INTO evaluations VALUES(48799,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(48799,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(48799,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48799,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(48799,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(48799,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(48799,36,'cp -12','d2d4 e5d4 e3d4 c7c6 c4d5 f6d5 g1f3 f8d6 f1c4 d8e7 c4e2 e8g8 e1g1 b8d7 f1e1 d7f6 e2d3 e7c7 b1c3 c8d7 c3d5 f6d5 d1c2 h7h6 h2h3 d7e6 c1d2 a8e8 f3e5 d5f6 e1e2 e6d5 a1e1 e8e7 e5g4 f6g4 e2e7 d6e7');
INSERT INTO evaluations VALUES(54486,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(54486,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(54486,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(54486,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(54486,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(54486,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(54486,36,'cp 85','b2b4 a7a5 c2c3 g7g6 g1f3 f8g7 a2a4 g8e7 f1e2 a5b4 c3b4 b8c6 c1b2 c6b4 b2e5 g7e5 f3e5 e8g8 e1g1 d8c7 e5f3 c7c5 d1d2 e7c6 f1c1 c5e7 b1c3 c8e6 h2h3 c6a5 d2b2 b4c6 c1d1 e7b4 b2c2 a5b3 a1b1 d5d4 e3d4 c6d4 f3d4 b3d4 b1b4 d4c2 b4b7');
INSERT INTO evaluations VALUES(54483,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(54483,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(54483,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(54483,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(54483,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(54483,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(54483,36,'cp 104','b8c6 a2a3 a7a5 b4b5 c6b8 c1b2 g8f6 b2e5 f8c5 e5d4 b8d7 g1f3 e8g8 f1d3 b7b6 e1g1 c8b7 b1d2 a8c8 d3f5 c5d4 f3d4 g7g6 f5d7 d8d7 d1b1 d7d6 b1b2 c8c7 a3a4 f8c8 h2h3 f6d7 f1d1 d6f6 a1c1 d7c5 b2a3 f6e7');
INSERT INTO evaluations VALUES(56439,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(56439,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(56439,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(56439,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(56439,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(56439,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(56439,36,'cp -61','d5d4 c3b1 c7c5 d2d3 f8d6 g1f3 b8c6 c2c3 g8f6 e1g1 e8g8 c3d4 e5d4 e2e3 d4e3 c1e3 f6d5 e3g5 d6e7 g5e7 d5e7 b1c3 e7f5 f1e1 a8b8 d1d2 c8e6 c3e2 e6d5 d2f4 d5f3 g2f3 c6d4 f3g2 b7b6 e2c3 d8d7 g1h1 b8e8 a1d1 h7h6 e1e8');
INSERT INTO evaluations VALUES(56932,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(56932,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(56932,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(56932,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(56932,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(56932,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(56932,36,'cp -138','g8f6 f3g5 h8g8 g5f3 d7d5 e4d5 f6d5 f1b5 c7c6 d1e2 d8e7 e2e7 f8e7 b5f1 b8d7 d2d3 d7f6 b1c3 c8e6 c3d5 f6d5 e1f2 e7d6 c1d2 e8c8 a1e1 d6c7 a2a3 h7h5 f3d4 c7b6 c2c3 d5e7 d2f4 e7f5 f4e3 f5d4 c3d4 b6c7 f2e2 c7a5 b2b4 a5c7 e2d2 c7g3 e1e2 h5h4 h1g1');
INSERT INTO evaluations VALUES(57429,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(57429,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(57429,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(57429,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(57429,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(57429,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(57429,36,'cp 189','g1e2 b8c6 d2d4 d7d6 c1e3 f6f5 d4e5 c6e5 e2g3 f5e4 g3e4 g8f6 e4f6 d8f6 f1e2 c7c6 c3e4 f6e6 e4g5 e6f5 f2f4 h7h6 f4e5 h6g5 e5d6 c8e6 d1d4 e8c8 h1f1 f5d5 d4d5 e6d5 e2g4 c8b8 d6d7 f8e7 e3d4 h8h4 h2h3 d5g2 d4e5 b8a8 f1f2 g2h3 g4h3 h4h3');
INSERT INTO evaluations VALUES(15809,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(15809,31,'cp -165',NULL);
INSERT INTO evaluations VALUES(15809,32,'cp -141',NULL);
INSERT INTO evaluations VALUES(15809,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(15809,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(15809,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(15809,36,'cp -170','a2b3 h7h6 g5f3 e5e4 f3g1 c8f5 g1e2 f8c5 e1g1 e8g8 c2c4 c7c6 b1c3 c6d5 c4d5 c5b6 d2d4 e4d3 e2g3 f5g6 d1f3 b6d4 f1e1 d8d7 c1e3 d4c3 b2c3 f6d5 e3d2 a7a5 c3c4 d5b4 a1a5 b4c2 e1d1 c2d4 f3d5 d7d5 c4d5 d4b3');
INSERT INTO evaluations VALUES(15808,30,'cp -176',NULL);
INSERT INTO evaluations VALUES(15808,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(15808,32,'cp -184',NULL);
INSERT INTO evaluations VALUES(15808,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(15808,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(15808,35,'cp -186',NULL);
INSERT INTO evaluations VALUES(15808,36,'cp -179','h7h6 g5f3 e5e4 f3g1 c8f5 g1e2 f8c5 e1g1 e8g8 b1c3 f8e8 d2d4 e4d3 c2d3 f6d5 d3d4 c5f8 c3d5 d8d5 e2c3 d5d7 c1e3 a7a6 d4d5 a8d8 h2h3 f8d6 a1a4 d6e5 f1e1 b7b5 a4a6 b5b4 c3a4 f5h3 a4c5 d7c8 a6a4 e5d6 a4a5 h3f5 a5b5 c7c6 b5b4 c6d5');
INSERT INTO evaluations VALUES(16915,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(16915,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(16915,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(16915,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(16915,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(16915,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(16915,36,'cp 85','c3b5 c7c6 b5c3 d7d5 e4d5 f6d5 e1g1 f8d6 c3d5 c6d5 d1h5 e8g8 e5f3 d8c7 c2c3 c7c4 b2b3 c4e4 h2h3 e4f5 h5f5 c8f5 a2a4 f5e4 f3d4 f8e8 c1a3 d6f4 f1d1 f4e5 a1c1 f7f6 d4c2 a8d8 a3c5 e4c2 c1c2 d5d4 d1c1 b7b6 c5d4 e5d4 c3d4');
INSERT INTO evaluations VALUES(16911,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(16911,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(16911,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(16911,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(16911,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(16911,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(16911,36,'cp 81','c7c6 b5c3 d7d5 e4d5 f6d5 e1g1 f8d6 c3d5 c6d5 d1h5 e8g8 e5f3 d8c7 c2c3 c7c4 b2b3 c4e4 f1e1 c8g4 h5g5 e4g6 g5g6 h7g6 f3d4 f8e8 c1b2 d6f4 h2h3 g4d7 d2d3 f7f6 c3c4 e8e1 a1e1 d5c4 d3c4 a8d8 g1f1 g8f7 b2c3 f4e5');
INSERT INTO evaluations VALUES(18343,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(18343,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(18343,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(18343,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(18343,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(18343,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(18343,36,'cp 43','d2d4 d7d6 c1e3 a7a5 b1c3 g8e7 h2h4 d8d7 d1d2 h7h5 d4e5 d6e5 d2d7 c8d7 c3a4 e7g6 g2g3 f8b4 f3d2 d7g4 c2c3 b4d6 f2f3 g4e6 e1c1 g6f8 h1g1 f8d7 b2b3 e8f7 g1h1 c6c5 c1c2 h8b8 a4b2 d7b6');
INSERT INTO evaluations VALUES(18342,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(18342,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(18342,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(18342,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(18342,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(18342,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(18342,36,'cp 48','d7d6 b1c3 a7a5 c1e3 g8e7 h2h4 d8d7 d1d2 h7h5 e1c1 a5a4 d4e5 d6e5 d2e2 d7e6 e3c5 e6f7 e2e3 e7g6 f3e1 f7c4 c5f8 g6f8 g2g3 f8e6 f2f4 c4c5 e3c5 e6c5 e1d3 c5d3 d1d3 c8g4 c3d1 e8g8 f4e5 f6e5 h1e1 a8d8 d3d8 f8d8 d1f2 d8d4 c2c3 d4d8 c1c2 c6c5 f2d3');
INSERT INTO evaluations VALUES(18293,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(18293,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(18293,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(18293,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(18293,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(18293,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(18293,36,'cp 58','b1c3 g8e7 c1e3 a7a5 h2h4 d8d7 d1d2 h7h5 e1c1 a5a4 d4e5 d6e5 d2e2 d7e6 e3c5 e7g6 c5f8 g6f8 f3e1 e6f7 g2g3 f8e6 e2c4 c6c5 e1d3 a8a5 c3d5 e8g8 h1e1 c8a6 c4c3 a5b5 f2f4 c7c6 f4e5 b5b8 d5f4 e6f4 d3f4 f7a2 e5f6 f8f6');
INSERT INTO evaluations VALUES(19977,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(19977,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(19977,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(19977,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(19977,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(19977,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(19977,36,'cp 73','e2e3 g7g6 d2d4 f8g7 f1e2 e7e6 d4d5 e6d5 c4d5 c6e5 f3e5 g7e5 e1g1 g8h6 e3e4 d8h4 g2g3 h4h3 c1h6 h3h6 c3b5 e8g8 f2f4 a7a6 f4e5 a6b5 e2b5 h6e3 f1f2 e3e4 e5d6 f8d8 d1d2 d8d6 a1e1 e4d4 d2d4 c5d4 e1e8 g8g7 b5c4 b7b5 c4b5 c8b7 a2a4 b7d5 e8a8 d5a8 a4a5 d6d5');
INSERT INTO evaluations VALUES(19976,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(19976,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(19976,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(19976,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(19976,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(19976,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(19976,36,'cp 64','g7g6 d2d4 f8g7 f1e2 e7e6 d4d5 e6d5 c4d5 c6e5 f3e5 g7e5 e1g1 g8h6 e3e4 d8h4 g2g3 h4h3 e2b5 e8e7 f2f3 a7a6 b5d3 f7f6 c3e2 g6g5 f1f2 h6f7 c1e3 b7b5 d1d2 h8g8 a1f1 h3h6 b2b4 c5c4 d3c2 c8h3 f1c1 h3d7 e2d4 e5d4 e3d4 h6g7 d4c3 f7e5 c1f1 h7h5');
INSERT INTO evaluations VALUES(25685,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(25685,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(25685,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(25685,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(25685,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25685,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25685,36,'cp 43','h2h3 g4f3 d1f3 g8f6 d2d3 f6d7 c1d2 f7f6 f3h5 g7g6 h5g4 f8d6 h3h4 d8e7 h4h5 e8c8 e1c1 h8g8 h5g6 h7g6 h1h6 e7f7 d1h1 b7b6 c1b1 g8g7 g2g3 c8b7 a2a3 d7c5 f2f4 f6f5 g4f3 f5e4 d3e4 c5e6 c3e2 d6c5 d2c3 c5d4 e2d4');
INSERT INTO evaluations VALUES(25683,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(25683,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(25683,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25683,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(25683,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(25683,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(25683,36,'cp 35','g4f3 d1f3 g8f6 d2d3 f6d7 c1d2 f7f6 f3g4 d8e7 f2f4 e8c8 e1c1 e5f4 d2f4 h7h5 g4f3 e7e6 a2a3 c6c5 h1f1 f8d6 c3d5 d7e5 f3g3 d8d7 c1b1 c8b8 d5e3 h5h4 g3f2 e5g6 f4d6 e6d6 f1e1 h8e8 e3f5');
INSERT INTO evaluations VALUES(25692,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(25692,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(25692,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(25692,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(25692,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(25692,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25692,36,'cp 48','d1f3 g8f6 d2d3 f6d7 c1d2 f7f6 f3g4 d8e7 g4h5 g7g6 h5g4 e8c8 e1c1 e7f7 f2f4 f8d6 a2a3 e5f4 d2f4 d6f4 g4f4 c8b8 h1f1 f7e7 f4h4 d8e8 f1f2 b7b5 d3d4 h8f8 d1f1 f8f7 f2e2 d7b6 h4f2 e7d6 c1b1');
INSERT INTO evaluations VALUES(25691,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(25691,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(25691,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25691,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(25691,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(25691,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(25691,36,'cp 29','g8f6 d2d3 f6d7 c1d2 f7f6 f3g4 d8e7 g4h5 g7g6 h5g4 e8c8 e1c1 e7f7 f2f4 f8d6 c1b1 f6f5 g4h4 e5f4 d2f4 d6f4 h4f4 h8f8 h1f1 f7g7 d1e1 f5e4 f4e4 d7c5 e4c4 f8f1 e1f1 g7e7 b2b4 c5e6 c3e4 d8e8 a2a3 c8b8');
INSERT INTO evaluations VALUES(30031,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(30031,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(30031,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(30031,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(30031,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(30031,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(30031,36,'cp -9','d8e7 f1e2 b8c6 d2d4 e6c4 c1e3 e7e6 g1f3 c4e2 d1e2 e8c8 a2a3 f6d5 e1g1 f7f6 f1e1 c8b8 h2h3 e6d7 f3d2 d5e3 f2e3 f6f5 a1d1 f8d6 d2c4 h8e8 e2c2 g7g6 c4d6 d7d6 d4d5 c6e5 e3e4 d6c5 c2f2 c5f2 g1f2');
INSERT INTO evaluations VALUES(32681,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(32681,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(32681,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(32681,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(32681,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(32681,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(32681,36,'cp -10','d2d4 e6c4 f1e2 b8c6 c1e3 e7e6 g1f3 c4e2 d1e2 e8c8 a1c1 f8b4 e1g1 b4a5 f3g5 e6e7 e2c4 a5b6 g5f3 e7e6 c4e6 f7e6 c1d1 h8f8 f1e1 h7h6 c3a4 b6a5 e3d2 a5d2 d1d2 f8e8 a4c5 e6e5 d4e5 d8d2 f3d2 c6e5 f2f4 e5d7 c5e6 g7g6 f4f5 g6g5');
INSERT INTO evaluations VALUES(36832,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(36832,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(36832,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(36832,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(36832,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(36832,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(36832,36,'cp 89','f1e2 c7c5 e1g1 b8c6 c1e3 b7b6 a1c1 c8g4 d4d5 c6a5 f1e1 d8d7 c3c4 e7e5 a2a4 a5b7 d1c2 g4f3 e2f3 a8b8 f3e2 f7f5 e4f5 g6f5 f2f3 b8d8 e3g5 d8c8 e2d3 b7d6 a4a5 c8b8 g5h4 e5e4');
INSERT INTO evaluations VALUES(36831,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(36831,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(36831,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(36831,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(36831,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(36831,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(36831,36,'cp 73','c7c5 e1g1 b7b6 c1f4 c8b7 d1d3 e7e6 d3e3 c5d4 c3d4 b8d7 f1d1 d7f6 f3d2 f6e8 f4e5 a8c8 a2a4 e8d6 a4a5 b6a5 e3g3 d6e8 h2h4 g7h6 h4h5 h6d2 d1d2 b7e4 e2a6 c8c6 a6b5 c6c8 h5g6 h7g6 b5a6 c8c2 d2c2 e4c2 a1c1 c2f5');
INSERT INTO evaluations VALUES(36055,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(36055,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(36055,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(36055,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(36055,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(36055,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(36055,36,'cp 85','e1g1 b7b6 c1f4 e7e6 d1c1 c5d4 c3d4 c8b7 c1e3 b8d7 f1d1 d7f6 f3d2 f6e8 f4e5 a8c8 a2a4 e8d6 e3g3 d6e8 e5g7 g8g7 g3e5 f7f6 e5e6 c8c7 d2f1 c7e7 e6b3 b7e4 f1e3 h7h5 d4d5 e8d6 e2f1 e7b7 f1b5 b7c7 b5c6 d8e7 a4a5 b6a5 a1a5 h5h4 b3a3');
INSERT INTO evaluations VALUES(36054,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(36054,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(36054,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(36054,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(36054,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(36054,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(36054,36,'cp 80','b7b6 c1f4 c8b7 d1d3 c5d4 c3d4 b8d7 e4e5 a8c8 h2h4 f8e8 e5e6 f7e6 h4h5 e8f8 d3e3 g6h5 f3g5 b7d5 a1c1 d7f6 c1c8 d8c8 f1c1 c8e8 g5e6 d5e6 e3e6 e8f7 e2c4 f7e6 c4e6 g8h8 f4g5 f6e4 g5e3 f8d8 g1f1 g7f6 f2f3 f6g5 f1e2 g5e3 e2e3 e4d6');
INSERT INTO evaluations VALUES(40200,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(40200,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(40200,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40200,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40200,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40200,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40200,36,'cp 1','c4b3 e8g8 c1e3 b8d7 d1e2 d7c5 f2f3 c5b3 a2b3 e6e5 d4b5 a7a6 b5a3 d6d5 e4d5 e7a3 b2a3 f6d5 c3d5 d8d5 c2c4 d5c6 e1g1 c8e6 a3a4 f8d8 a1d1 c6c7 h2h3 h7h6 d1c1 f7f6 f3f4 e6f7 f4e5');
INSERT INTO evaluations VALUES(43519,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43519,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43519,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43519,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43519,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43519,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(43519,36,'cp 0','b7b5 c4b3 f8e7 f2f3 e8g8 a2a3 c8b7 d1e2 b8c6 e1c1 c6a5 g2g4 a5b3 c2b3 a8c8 c1b1 d6d5 g4g5 f6h5 e4e5 b5b4 a3b4 e7b4 d1c1 g7g6 c3a4 d8a5 f3f4 h5g7 h2h4 b4e7 e3d2 e7b4');
INSERT INTO evaluations VALUES(43748,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43748,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43748,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43748,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43748,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43748,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43748,36,'cp 0','c4b3 f8e7 f2f3 e8g8 a2a3 c8b7 d1e2 b8c6 e1c1 c6a5 g2g4 a5b3 c2b3 a8c8 c1b1 d6d5 g4g5 f6h5 e4e5 b5b4 a3b4 e7b4 d1c1 g7g6 c3a4 d8a5 f3f4 h5g7 h2h4 b4e7 e3d2 e7b4');
INSERT INTO evaluations VALUES(44204,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44204,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(44204,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(44204,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(44204,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(44204,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(44204,36,'cp 28','g1f3 d7d5 c4d5 e6d5 d2d4 b8c6 b1c3 c5d4 f3d4 f8c5 d4b3 c5b6 e1g1 d5d4 c3a4 e8g8 c1g5 h7h6 a4b6 a7b6 g5f6 d8f6 g2c6 b7c6 d1d4 f8d8 d4f6 g7f6 f1d1 d8d1 a1d1 c8e6 b3c1 e6a2 b2b3 a2b1 f2f3 a8a5 g1f2 a5c5 c1d3 b1d3');
INSERT INTO evaluations VALUES(47964,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47964,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(47964,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47964,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47964,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47964,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47964,36,'cp 0','h7h6 e1g1 d6d5 b1c3 c8f5 f3d4 f5c8');
INSERT INTO evaluations VALUES(51825,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(51825,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(51825,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(51825,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(51825,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(51825,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(51825,36,'cp -5','c2c4 g8f6 a2a3 f8c5 g1f3 d7d6 d5c6 b8c6 e2e3 a7a5 b1c3 c8f5 f1e2 e8g8 e1g1 f6e4 c3a4 d8e7 f3d2 e4f6 d2b1 e5e4 b1c3 c6e5 b2b3 a8d8 c1b2 e5d3 d1d2 f8e8 a1d1 f6g4 h2h3 g4e5 c3d5 e7g5 a4c3 d3b2 d2b2 e5f3 g1h1');
INSERT INTO evaluations VALUES(51823,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(51823,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(51823,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(51823,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(51823,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(51823,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(51823,36,'cp -23','g8f6 a2a3 f8c5 e2e3 a7a5 g1e2 d7d6 b1c3 b8a6 e2g3 h7h5 h2h4 g7g6 g3e4 f6e4 c3e4 e8g8 f1e2 d8e7 e4c3 f7f5 e2f3 c8d7 c1d2 a6c7 d1b3 c6d5 c3d5 c7d5 f3d5 g8h7 e1c1 d7e6 d5e6 e7e6 b3b7 h7h6 b7c6 e6c4 c1b1');
INSERT INTO evaluations VALUES(44063,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(44063,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(44063,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(44063,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(44063,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(44063,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(44063,36,'cp -19','a2a3 f8c5 e2e3 a7a5 b1c3 d7d6 f1d3 b7b5 g1e2 b5c4 d3c4 e8g8 e1g1 c6d5 c3d5 b8d7 e2c3 c8b7 d1d3 f6d5 c4d5 b7a6 d5c4 a6c4 d3c4 a8c8 c4d3 d7f6 f1d1 c8b8 c1d2 d6d5 a1c1 d5d4 e3d4 d8d4 d2e3 d4d3 d1d3 c5e3');
INSERT INTO evaluations VALUES(312,30,'cp -926',NULL);
INSERT INTO evaluations VALUES(312,31,'cp -932',NULL);
INSERT INTO evaluations VALUES(312,32,'cp -925',NULL);
INSERT INTO evaluations VALUES(312,33,'cp -926',NULL);
INSERT INTO evaluations VALUES(312,34,'cp -932',NULL);
INSERT INTO evaluations VALUES(312,35,'cp -928',NULL);
INSERT INTO evaluations VALUES(312,36,'cp -925','g5g2 h1f1 c6d4 b1d2 d7d5 c4e2 g8f6 c2c3 d4e2 d1e2 c8h3 f2f3 g2e2 e1e2 h3f1 e2f1 e8d7 b2b4 c5e7 a2a4 a7a5 b4b5 d5e4 d2e4 f6e4 f3e4 g7g5 f1e2 h7h5 c1e3 g5g4 a1f1 d7e6 f1g1 c7c6 b5c6 b7c6 e2d3 a8d8 d3e2 c6c5');
INSERT INTO evaluations VALUES(315,30,'cp -925',NULL);
INSERT INTO evaluations VALUES(315,31,'cp -934',NULL);
INSERT INTO evaluations VALUES(315,32,'cp -934',NULL);
INSERT INTO evaluations VALUES(315,33,'cp -934',NULL);
INSERT INTO evaluations VALUES(315,34,'cp -932',NULL);
INSERT INTO evaluations VALUES(315,35,'cp -932',NULL);
INSERT INTO evaluations VALUES(315,36,'cp -925','h1f1 c6d4 c1e3 d4f3 e1e2 c5e3 e2e3 f3h2 f1g1 g2h3 e3d2 h2f3 d2c1 h3h6 b1d2 f3g1 d1g1 g8f6 g1e1 e8g8 a2a4 f6h5 a1a3 h5f4 a4a5 h6c6 c4d3 d7d6 d2f1 a7a6 f2f3 g7g6 f1e3 c6d7 e3c4 d7e7 c4e3 h7h5 a3b3 g8g7 d3e2 f7f6');
INSERT INTO evaluations VALUES(316,30,'cp -948',NULL);
INSERT INTO evaluations VALUES(316,31,'cp -948',NULL);
INSERT INTO evaluations VALUES(316,32,'cp -950',NULL);
INSERT INTO evaluations VALUES(316,33,'cp -960',NULL);
INSERT INTO evaluations VALUES(316,34,'cp -963',NULL);
INSERT INTO evaluations VALUES(316,35,'cp -971',NULL);
INSERT INTO evaluations VALUES(316,36,'cp -980','c6d4 c1e3 d4f3 e1e2 d7d6 b1d2 c5e3 e2e3 f3h2 d1e1 g2g5 e3d3 h2f1 e1f1 g8f6 d3c3 c7c6 c4d3 f6h5 b2b3 h5f4 c3b2 e8g8 a1d1 g5f6 b2b1 h7h5 f2f3 c8e6 f1e1 a8d8 e1f2 a7a6 a2a4 f4d3 c2d3 f6e7 b1b2 g7g6 f2b6');
INSERT INTO evaluations VALUES(2032,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(2032,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(2032,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(2032,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(2032,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(2032,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(2032,36,'cp 97','f7f6 e5f6 g8f6 d5b3 d7d5 c1f4 c8d7 e2e3 e8c8 b1d2 a7a6 c2c3 h7h6 f4g3 g7g5 f3d4 g5g4 g3h4 d8e8 f1d3 c6e5 d3c2 c8b8 e1c1 e7f7 h1f1 f8g7 c1b1 h8f8 e3e4 f6e4 d2e4 d5e4 c2e4 f7b3 a2b3 g7f6 h4g3 h6h5 e4f5 d7c8 f5c8');
INSERT INTO evaluations VALUES(2013,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(2013,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(2013,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(2013,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(2013,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(2013,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(2013,36,'cp 103','e5f6 g8f6 d5b3 d7d5 c1f4 c8d7 b1d2 e8c8 e2e3 a7a6 c2c3 h7h6 f4g3 g7g5 f3d4 d8e8 e1c1 h6h5 h2h4 g5g4 f1d3 e7f7 c1b1 c6e5 d3e2 f8d6 b1a1 f7e7 d1c1 e5c4 e2c4 d6g3 c4a6 b7a6 f2g3 e7e3 c1f1 e3d2 f1f6 d2g2');
INSERT INTO evaluations VALUES(5098,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(5098,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(5098,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(5098,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(5098,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(5098,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(5098,36,'cp -82','e1g1 g8e7 f2f4 d7d5 c4b5 e8g8 c2c3 c5c4 g1h1 d8b6 b1a3 a7a6 b5c6 b7c6 d3d4 h7h6 g5f3 d5e4 f3d2 c6c5 d2c4 b6c7 d4c5 c7c5 c1e3 c5c6 d1e1 e7f5 c4e5 c6c7 a1d1 f8d8 d1d8 c7d8 a3c4 a6a5 c4b6 a8a6 b6a4 a6d6 a4c5 f5e3 e1e3');
INSERT INTO evaluations VALUES(5097,30,'cp -110',NULL);
INSERT INTO evaluations VALUES(5097,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(5097,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(5097,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(5097,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(5097,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(5097,36,'cp -80','g8e7 c4b3 e8g8 g5f3 b7b6 f1e1 d7d5 b1a3 c8a6 e4d5 e7d5 c2c3 h7h6 a3c4 d8c7 d1e2 a8d8 c4e5 a6b7 a2a4 b7a8 b3c4 g8h7 c1d2 d5f6 c4b5 c6e5 f3e5 f6h5 e5c4 d8d5 a4a5 b6a5 c4a5 f8d8 e2f3 h5f6');
INSERT INTO evaluations VALUES(10359,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(10359,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(10359,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(10359,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(10359,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10359,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10359,36,'cp 16','f1g2 d7d6 c3d5 b4c5 c2c3 f6d5 e4d5 c6e7 d2d4 e5d4 f3d4 e8g8 e1g1 a7a5 f1e1 h7h6 d4b3 c5b6 c1e3 a5a4 b3d2 b6e3 e1e3 c8d7 d1c2 d7f5 g2e4 f5e4 c2e4 e7g6 a1e1 d8d7 e4d4 f8e8 d2f3 e8e3 e1e3 a8e8 c3c4 e8e3 d4e3 d7e7 e3e7 g6e7');
INSERT INTO evaluations VALUES(10360,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(10360,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(10360,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(10360,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10360,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(10360,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(10360,36,'cp 17','e8g8 e1g1 d7d6 d2d3 a7a5 c1e3 b4c5 c3d5 c5e3 d5e3 f6g4 d1e2 g4e3 e2e3 a5a4 c2c3 h7h6 d3d4 c8g4 f3h4 d8f6 h2h3 g4h5 d4d5 c6b8 b2b4 a4b3 a2b3 b8d7 a1a7 a8a7 e3a7 h5e2 f1e1 e2a6 h4f5 f6g5 f5e3 g7g6 b3b4 g8g7 e3c2 a6d3');
INSERT INTO evaluations VALUES(21195,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(21195,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(21195,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(21195,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21195,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21195,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(21195,36,'cp 23','e2e4 e7e5 d2d3 g7g6 f1e2 f8g7 f4e5 c6e5 b1c3 g8e7 c1f4 c7c5 d1d2 e7c6 f4g5 f7f6 g5e3 e5f3 e2f3 e8g8 h2h4 h7h5 c3d5 g8h7 e1c1 c8e6 d5f4 d8e7 c1b1 c6e5 d3d4 e5c4 d2e2 c4e3 e2e3 e6f7 g2g4 c5d4');
INSERT INTO evaluations VALUES(21192,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(21192,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(21192,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(21192,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(21192,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21192,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(21192,36,'cp 18','e7e5 d2d4 e5d4 f3d4 d6d5 b1c3 c6d4 d1d4 d5e4 d4e4 f8e7 c1d2 g8f6 e4e3 e8g8 e1c1 f8e8 h2h3 c8e6 e3f3 e7c5 f1d3 e6d5 c3d5 d8d5 f3d5 f6d5 d1e1 c5e3 d2e3 d5e3 g2g3 c7c6 c1d2 e3d5 a2a4 g7g6 a4a5 g8f8 e1e8 a8e8 h1d1 e8d8 d2e2 f8g7 d3e4 d8e8 e2f3 d5f6');
INSERT INTO evaluations VALUES(21785,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21785,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21785,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21785,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21785,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21785,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21785,36,'cp 9','g8f6 d1e2 c8g4 c4b5 d8d6 b1a3 g4f3 g2f3 f6d7 c2c3 d6f6 a3c2 e8c8 d2d3 a7a6 b5c6 f6c6 c1e3 c6b5 e1c1 d7c5 c3c4 b5a4 c1b1 c5e6 e2f2 f8e7 f3f4 e5f4 e3f4 f7f5 e4f5 e6f4 f2f4 e7f6 f4e4 d8d7 h1e1 h8d8 e4g4');
INSERT INTO evaluations VALUES(15209,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(15209,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15209,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(15209,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15209,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(15209,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(15209,36,'cp 0','d1e2 c8g4 c4b5 d8d6 b1a3 e8c8 a3c4 d6e6 b5c6 e6c6 d2d3 f8b4 c2c3 b7b5 c4e5 b4c3 c1d2 c3d2 f3d2 c6e6 e5g4 e6g4 d2b3 h8e8 g2g3 h7h6 a1c1 g4d7 c1c3 f6d5 c3c5 d5f6');
INSERT INTO evaluations VALUES(22084,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(22084,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(22084,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(22084,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(22084,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(22084,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(22084,36,'cp 51','c7c6 b5e2 g8f6 c2c4 f8e7 e1g1 e8g8 b1c3 f6d7 c1e3 d7c5 a1d1 d8e8 f2f3 f7f5 b2b4 f5f4 e3f2 c5e6 d4d2 f8f6 g1h1 f6h6 f2g1 e8h5 c4c5 d6c5 b4b5 e6d4 e2c4 c8e6 c4e6 h6e6 b5c6 e6c6 g1d4 c5d4 c3d5 h5f7 d2d4 e7c5 d4a4 a8e8 d1b1');
INSERT INTO evaluations VALUES(20741,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(20741,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20741,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(20741,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(20741,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(20741,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(20741,36,'cp 50','b5e2 g8f6 b1c3 d8b6 d4d3 f8e7 e1g1 e8g8 c1e3 b6d8 a2a4 f6g4 e3d4 c6c5 d4e3 g4e3 d3e3 e7f6 a1d1 f6d4 e3g3 a7a6 e2c4 g8h8 c3e2 d4b2 g3d6 d8d6 d1d6 f7f5 c2c3 f5e4 d6d2 b7b5 a4b5 a6b5 c4b5 a8a2');
INSERT INTO evaluations VALUES(22757,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(22757,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(22757,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(22757,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(22757,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(22757,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(22757,36,'cp -30','e5e4 d4d5 e4f3 d5c6 f3g2 c6d7 d8d7 f1g2 d7d1 e1d1 a7a5 a2a4 g8f6 b1d2 f8c5 d1e1 e8g8 d2b3 c5d6 c1g5 f6g4 b3d4 h7h6 g5h4 a8a6 a1d1 a6b6 d4b5 d6e5 h4d8 e5h2 h1h2 g4h2 d8c7 b6g6 c7g3 h2g4 g2e4 f7f5');
INSERT INTO evaluations VALUES(27219,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(27219,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(27219,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(27219,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(27219,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(27219,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(27219,36,'cp 66','d2d3 e5f4 c1f4 a7a6 b5c6 b7c6 e1g1 g8e7 f4d2 e7g6 d2c3 f7f6 b1d2 f8e7 f3d4 e8g8 d1h5 f8e8 g1h1 g6e5 d4f5 d7f5 h5f5 d8d7 a1e1 d7f5 f1f5 g7g6 f5f1 c6c5 c3e5 f6e5 d2c4 a6a5 a2a4 e7g5 g2g3 a8b8 h1g2');
INSERT INTO evaluations VALUES(30926,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30926,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(30926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30926,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(30926,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(30926,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30926,36,'cp -7','g8f6 d2d3 b8c6 c1f4 e7e6 b1d2 f8e7 e1g1 e8g8 c2c3 c6a5 a2a4 f6h5 f4e3 a5c4 d2c4 d7e8 f1e1 d8c7 e3d2 a8d8 d1e2 h5f6 d2f4 f6h5');
INSERT INTO evaluations VALUES(28872,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28872,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(28872,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(28872,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28872,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28872,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28872,36,'cp 0','d2d3 b8c6 e1g1 e7e6 f1e1 h7h6 a2a4 f8e7 c4b3 d8c7 c2c3 d6d5 e4e5 f6h7 c1e3 h7g5 b1d2 g5f3 d2f3 a7a6 a1c1 g7g5 d3d4 c5d4 c3d4 e8c8 d1e2 c8b8 b3c2 g5g4 f3d2 h6h5 d2b3 e7b4 c2d3 b4e1 e2e1 b8a8');
INSERT INTO evaluations VALUES(37447,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(37447,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37447,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37447,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37447,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37447,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37447,36,'cp 4','e3e4 d7d5 e4e5 f6e4 f1d3 e4c3 b2c3 c7c5 e1g1 c8g4 h2h3 g4f3 d1f3 b8c6 c1b2 c5d4 c3d4 c6b4 c2c3 b4d3 f3d3 d8d7 f2f4 f7f5 e5f6 f8f6 a1e1 a8c8 b2c1 f6e6 g2g4 e6e1 f1e1 g8f7 d3f3 e7e6 g1g2 c8c4');
INSERT INTO evaluations VALUES(38116,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(38116,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(38116,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38116,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(38116,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38116,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38116,36,'cp 28','c4d5 g8f6 d5e6 d8d1 c3d1 c8e6 a2a3 b8c6 g1f3 c6a5 f3d2 c5b6 e2e3 e8g8 b2b4 a5b3 d2b3 e6b3 c1b2 a7a5 f1e2 a5b4 a3b4 a8a1 b2a1 f6d5 d1c3 d5b4 e1g1 b6c5 f1b1 b3e6 h2h3 f8d8 c3e4 b7b6 g2g4 b4d5 a1b2 c5e7');
INSERT INTO evaluations VALUES(38109,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38109,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(38109,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(38109,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(38109,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(38109,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38109,36,'cp 15','g8f6 d5e6 d8d1 c3d1 c8e6 a2a3 b8c6 g1f3 c6a5 f3d2 c5b6 b2b4 a5b3 d2b3 e6b3 c1b2 e8g8 e2e3 a7a5 f1e2 a5b4 a3b4 f6d5 e1g1 a8a1 b2a1 d5b4 d1c3 b6c5 c3b5 f8d8 b5d4 b3d5 f1d1 d5e4 f2f3 e4g6 g1f2 d8c8 a1b2 h7h6 h2h3 c5b6 d1c1 b4c2');
INSERT INTO evaluations VALUES(16711,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(16711,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(16711,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(16711,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(16711,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(16711,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(16711,36,'cp -31','f8b4 c2c3 b4a5 f1d3 e8g8 e1g1 a5b6 e3f4 d7d5 e4d5 f6d5 f4a4 f8e8 c1g5 d8d6 b1a3 h7h6 a3c4 d6d7 g5d2 d5c3 d2c3 d7d3 f3e5 d3d5 c4b6 c7b6 a1d1 d5c5 f1e1 c8e6 a2a3 f7f6 e5c6 b7c6 c3b4 c5c4 d1c1 c4b5 a4b5 c6b5 c1c7 a7a5 b4e7 a8c8 c7c8 e6c8 f2f3 c8d7');
INSERT INTO evaluations VALUES(17912,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(17912,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(17912,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(17912,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(17912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17912,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(17912,36,'cp 4','c2c4 d7d5 c4d5 d8d5 b1c3 d5d6 g1f3 g8f6 d1b1 c8e6 f1e2 e8c8 e1g1 h7h5 f1d1 d6e7 h2h3 h8h6 d2d4 e5d4 e3d4 f6d5 e2c4 d5c3 b2c3 e6d5 f3e5 d5c4 b3c4 c6e5 d4e5 d8d1 b1d1 e7c5 d1d5 c5d5 c4d5 f8a3 f2f4 f7f6 e5e6 h5h4');
INSERT INTO evaluations VALUES(18809,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(18809,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(18809,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(18809,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(18809,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18809,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(18809,36,'cp 6','g1f3 a7a6 g2g3 d7d5 e4d5 e6d5 d1e2 d8e7 e2e7 g8e7 b1c3 c8g4 f1g2 e8c8 f3g5 g4h5 c3a4 c6d4 e1g1 h7h6 b2d4 c5d4 g5h3 c8c7 h3f4 h5g4 b3b4 g7g5 f4d3 b7b6 a4b2 e7c6 a2a3 g4e2 f1e1 e2g4 f2f4 f8e7 d3e5 c6e5 e1e5 g4e6 b2d3 e7f6');
INSERT INTO evaluations VALUES(19797,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19797,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19797,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(19797,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(19797,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(19797,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(19797,36,'cp 11','e4d5 d8d5 b1c3 d5d8 f1b5 c8d7 b5c6 d7c6 d1e2 g8h6 f3e5 d8c8 b2b3 h6f5 d2d3 e7e6 c1b2 f8e7 e1c1 e8g8 h1e1 b7b5 e5c6 c8c6 e2e4 c6d7 g2g4 f5d4 c1b1 a8d8 c3e2 f7f5 e4g2 a7a5 g4f5 d4f5 e2g3 f5d4 c2c3 d4f5 c3c4 f5d4');
INSERT INTO evaluations VALUES(19768,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(19768,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19768,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19768,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19768,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(19768,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(19768,36,'cp 0','d8d5 b1c3 d5d8 b2b3 a7a6 c1b2 g8f6 f3e5 d8c7 e5c6 c7c6 h1g1 h7h5 d1f3 c6f3 g2f3 c8f5 e1c1 e7e6 d1e1 h8h7 c3d1 f6d5 d1e3 e8c8 e3f5 e6f5 f1c4 g7g6 a2a4 h5h4 c1d1 h4h3 a4a5 d5f4 b2f6 f4g2 e1e2 g2f4');
INSERT INTO evaluations VALUES(1400,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(1400,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(1400,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1400,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1400,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1400,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(1400,36,'cp 0','b1c3 d5d8 f1b5 c8d7 f3e5 c6e5 f4e5 d7b5 c3b5 e7e6 e1g1 d8d7 a2a4 a7a6 b5a3 d7d4 g1h1 d4e5 a3c4 e5c7 d2d3 g8f6 c1f4 c7c6 c4a5 c6b6 a5c4');
INSERT INTO evaluations VALUES(1399,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1399,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1399,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1399,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1399,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1399,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1399,36,'cp 0','d5d8 b2b3 a7a6 c1b2 g8f6 f3e5 d8c7 e5c6 c7c6 h1g1 h7h5 d1f3 c6f3 g2f3 c8f5 e1c1 e7e6 d1e1 h8h7 c3d1 f6d5 d1e3 e8c8 b2e5 h5h4 f1c4 d5b6 c4f1');
INSERT INTO evaluations VALUES(19766,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19766,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19766,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19766,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19766,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(19766,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(19766,36,'cp 23','f1b5 c8d7 b2b3 e7e6 c1b2 c6d4 b5d7 d8d7 f3e5 d7c7 c3e2 d4f5 e2g3 f5d4 b2d4 c5d4 d1e2 g8f6 e2b5 f6d7 e5d7 c7d7 b5d7 e8d7 e1e2 f8d6 g3e4 a8c8 e4d6 d7d6 e2d3 c8c5 a1c1 h8c8 h1e1 f7f6 e1e4 e6e5 f4e5 f6e5 e4h4 h7h6 c2c4 d4c3 c1c3 d6e7 h4a4 a7a6 g2g4 c5c3 d2c3 e7f6 h2h4 c8d8 d3e3');
INSERT INTO evaluations VALUES(26562,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(26562,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(26562,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(26562,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(26562,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(26562,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26562,36,'cp 26','e2e3 a7a6 c3c4 e7e6 a2a3 g8f6 b2b4 f8e7 f1e2 d5c4 e2c4 e8g8 e1g1 a6a5 b4b5 c6b8 d1e2 b8d7 b1d2 c7c6 b5c6 b7c6 h2h3 f5g6 a3a4 e7b4 c1b2 a8b8 b2a3 b4a3 a1a3 c6c5 a3c3 c5d4 f3d4 d7e5');
INSERT INTO evaluations VALUES(26560,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(26560,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26560,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26560,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(26560,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(26560,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(26560,36,'cp 22','a7a6 a2a3 e7e6 c3c4 g8f6 b2b4 f8e7 c1b2 e8g8 c4c5 b7b6 b1d2 f6e4 d2e4 f5e4 f1e2 b6c5 b4c5 e7f6 e1g1 d8e7 e2d3 e6e5 d4e5 c6e5 d3e4 d5e4 f3e5 f6e5 b2e5 e7e5 d1d7 f8d8 d7c6 e5e7 a1b1 h7h6 h2h3 a8c8');
INSERT INTO evaluations VALUES(32418,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(32418,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(32418,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(32418,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(32418,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(32418,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(32418,36,'cp 157','d1d8 e8d8 b1d2 g8f6 d2c4 b8d7 h1g1 c7c6 a2a4 g7g6 f1h3 a7a5 c1g5 h7h6 g5h4 d8c7 e1c1 a8e8 d1d3 g6g5 h4g3 h6h5 h3d7 f6d7 g1d1 d7c5 g3e5 e8e5 c4e5 c5d3 d1d3 f7f6 e5g6 h8g8 g6f8 g8f8 c1d2 f8g8 d2e2 h5h4 b2b3 b7b6 h2h3 g8b8');
INSERT INTO evaluations VALUES(27721,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(27721,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(27721,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(27721,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(27721,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(27721,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(27721,36,'cp 133','e8d8 b1c3 g8f6 f3f4 b8d7 f4e5 d7e5 c1e3 c7c6 f2f4 e5g6 h1g1 f8b4 e1c1 d8c7 f1c4 b4c3 b2c3 h8e8 e4e5 f6d5 c4d5 c6d5 d1d5 b7b6 d5d1 a8d8 f4f5 g6e5 g1g7 d8d1 c1d1 h7h5 e3f4 c7c6 f4e5 e8e5 g7f7 e5a5 f5f6 a5f5 d1d2 a7a5 d2e3 c6d5 f7h7 f5f6 h7h5 d5c4');
INSERT INTO evaluations VALUES(27736,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(27736,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(27736,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(27736,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(27736,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(27736,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(27736,36,'cp 164','b1c3 b8d7 f3f4 g8f6 f4e5 d7e5 c1e3 c7c6 f2f4 e5g6 h1g1 f8b4 e1c1 d8c7 f1c4 b4c3 b2c3 a8e8 e4e5 f6d5 c4d5 c6d5 d1d5 b7b6 c3c4 e8e7 g1d1 f7f6 e5f6 g7f6 e3f2 c7b7 f2g3 h8c8 d1d4 e7c7 f4f5 g6e5 g3e5 f6e5 d5e5 c7c4 d4c4 c8c4 f5f6 c4f4 e5e7 b7c6 e7a7 f4f6 a7h7 f6f2 h2h4');
INSERT INTO evaluations VALUES(27735,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(27735,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(27735,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(27735,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(27735,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(27735,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(27735,36,'cp 143','f8d6 h1g1 g7g6 c1g5 d8e8 c3b5 b8c6 e1c1 h7h6 g5e3 e8e7 b5d6 c7d6 f3f4 b7b6 f4e5 d6e5 f1b5 a8c8 b5c6 c8c6 d1d5 g8f6 d5e5 c6e6 e5e6 e7e6 f2f3 f6d7 f3f4 h8c8 g1d1 d7f6 d1d4 f6g4 e3g1 g4f6 h2h4 f6g4 c2c3 f7f5 a2a4 f5e4 d4e4 e6f7 e4d4 g4f6 c1d2 f7e6 g1e3 f6h5 d2c2');
INSERT INTO evaluations VALUES(34303,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(34303,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(34303,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(34303,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(34303,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(34303,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(34303,36,'cp 76','d5a5 d2d4 g8f6 f1c4 c8f5 c1d2 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 d7b6 c4b3 a7a5 a2a3 f8e7 c3e4 b6d5 c2c4 f6e4 e2e4 d5f6 e4f3 a8d8 d2f4 e7d6 f4e3 d6e7 g2g4 b7b5 c1b1 b5c4 b3c4 e8g8 h4h5 g6h5 g4h5');
INSERT INTO evaluations VALUES(49404,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(49404,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(49404,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(49404,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(49404,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(49404,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(49404,36,'cp -33','c7c5 c2c3 c5d4 c3d4 b8c6 g1f3 d8b6 f4c1 d7d5 e2e3 f6e4 a2a3 f8d6 f1d3 e8g8 b2b4 c6e7 b1d2 f7f5 e1g1 c8d7 f3e5 d7e8 d2f3 a7a6 c1b2 a8c8 d1e2 e8a4 f1c1 c8c1 a1c1 h7h6 b2c3 f8c8 e2b2 a4e8 c3e1 c8c1 b2c1 e8b5 c1b2 g7g5');
INSERT INTO evaluations VALUES(44225,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(44225,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(44225,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(44225,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(44225,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(44225,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(44225,36,'cp -34','c2c3 c5d4 c3d4 b8c6 g1f3 d8b6 f4c1 d7d5 e2e3 f6e4 a2a3 f8d6 f1d3 e8g8 b2b4 c8d7 b1d2 f7f5 e1g1 c6e7 f3e5 d7e8 d2f3 a7a6 d1e2 b6d8 e2b2 e8h5 b4b5 a6b5 b2b5 e7c6 a3a4 d8e7 a4a5 f8c8 c1b2 h5f3 e5f3 c6b4 b2a3 e4c3');
INSERT INTO evaluations VALUES(53896,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53896,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53896,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(53896,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(53896,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(53896,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53896,36,'cp -12','b8c6 c3d5 g8f6 c2c4 f8e7 f1e2 e8g8 e1g1 a7a6 h2h3 b7b5 b2b3 a8b8 c1d2 h7h6 a2a3 b5c4 b3c4 f6d5 c4d5 c6d4 f3d4 c5d4 d1c2 f7f5 f1b1 b8b1 a1b1 e7h4 d2e1 f5e4 d3e4 d8f6 e2d3 f6g5 c2c7');
INSERT INTO evaluations VALUES(54203,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(54203,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(54203,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(54203,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(54203,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(54203,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(54203,36,'cp -23','d4c5 f8c5 f1d3 b8c6 e3e4 d5d4 c3e2 g8f6 a2a3 e6e5 g1f3 a7a5 e1g1 e8g8 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 d1e2 a8c8 f1d1 d8e7 c1e3 c5e3 e2e3 e6b3 d1d2 f8d8 d3f1 a5a4 f1b5 d8d2 f3d2 c6d4 d2b3');
INSERT INTO evaluations VALUES(54068,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(54068,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(54068,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(54068,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(54068,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(54068,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(54068,36,'cp -25','b8c6 g1f3 f8c5 f1d3 e6e5 e3e4 d5d4 c3e2 g8f6 e1g1 e8g8 a2a3 a7a5 c2c3 c8g4 c3d4 c6d4 e2d4 c5d4 h2h3 g4f3 d1f3 d8b6 a1b1 a5a4 c1g5 f8c8 g5f6 b6f6 f3f6 g7f6 f1d1 g8g7 g2g3 a8a5 g1g2 a5c5 b2b3 b7b5 b3a4');
INSERT INTO evaluations VALUES(56252,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(56252,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(56252,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(56252,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(56252,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(56252,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(56252,36,'cp 29','c2c4 c7c6 c4d5 c6d5 d2d3 g8f6 b1d2 b8c6 a2a3 c8e6 d3e4 f6e4 b3b4 f8d6 g1f3 e8g8 a1c1 d8e7 f1e2 a7a6 e1g1 h7h6 d2b3 a8c8 h2h3 f8d8 b3d4 e6d7 d4c6 c8c6 d1d4 f7f6 c1c6');
INSERT INTO evaluations VALUES(1849,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(1849,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(1849,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(1849,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(1849,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(1849,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(1849,36,'cp 46','f6g6 d4e5 c6e5 g5e3 g6e4 f1e2 c7c6 e1g1 d7d6 b1d2 e4f5 f3e5 d6e5 e2d3 f5g4 f2f3 g4d7 d2c4 d7c7 f3f4 e5e4 d3e4 c8e6 e4d3 g8f6 d1f3 e8c8 g1h1 e6d5 f3e2 d5c4 d3c4 c8b8 e2f3 f8d6 g2g3 a7a6 c4d3 h8e8 e3g1');
INSERT INTO evaluations VALUES(1884,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(1884,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(1884,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(1884,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(1884,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(1884,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(1884,36,'cp 36','d4e5 c6e5 g5e3 g6e4 f1e2 c7c6 e1g1 d7d6 b1d2 e4f5 e3d4 g8f6 f3e5 d6e5 e2d3 f5g4 d4e5 g4d1 f1d1 f8e7 d2c4 e8g8 c4d6 f8d8 d3c4 e7d6 e5d6 c8f5 c2c3 d8d7 d6c5 a7a5 f2f3 f5e6 c4e6 f7e6 d1d7 f6d7 c5e3 d7f6 a1d1');
INSERT INTO evaluations VALUES(10922,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(10922,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10922,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(10922,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10922,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(10922,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(10922,36,'cp 7','d3d4 e5d4 f3d4 e7b4 d4c6 b7c6 f1d3 d7d5 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 h7h6 c1f4 a7a5 c3a4 c8e6 a1d1 f8e8 b2b3 f6d7 c2c3 b4f8 f4g3 d8f6 f3f6 d7f6 f1e1 f6d7 e1e2 g7g5 g3h2 f7f6 f2f4 d7c5 a4c5 f8c5 g1f1 e6f7 f4g5 e8e2 d3e2 h6g5');
INSERT INTO evaluations VALUES(14784,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(14784,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(14784,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(14784,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(14784,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(14784,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(14784,36,'cp 53','f1g2 d5c4 d1a4 f8b4 c1d2 f6d5 d2b4 d5b4 e1g1 a8b8 b1a3 e8g8 a4b5 b7b6 b5c4 c8a6 a3b5 d8d5 c4d5 b4d5 a2a4 a6b7 f1c1 a7a6 b5c3 d5b4 c3a2 b4a2 a1a2 f8d8 e2e3 c6b4 a2a1 c7c5 d4c5 b7f3 g2f3 b6c5 c1c5 d8d2 a1d1 d2d1 f3d1 b4d3 c5c7 d3b2 d1c2');
INSERT INTO evaluations VALUES(16416,30,'cp -142',NULL);
INSERT INTO evaluations VALUES(16416,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(16416,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(16416,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(16416,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(16416,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(16416,36,'cp -138','f6e4 g1f3 e4c3 d2c3 d8e7 d1e2 e5f4 c1f4 e7e2 c4e2 d7d6 e1c1 f8e7 f3d4 c6d4 c3d4 c8e6 d4d5 e6d7 d1e1 e8f8 e2d3 h7h5 h2h4 g7g6 f4d2 a8e8 g2g3 f8g7 h1f1 f7f5 d2f4 e7f6 c1d2 a7a5 c2c3 g7f7 e1e8 h8e8 f1f2 a5a4 f2f1 f7g7 b2b3');
INSERT INTO evaluations VALUES(24557,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(24557,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(24557,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(24557,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(24557,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(24557,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(24557,36,'cp 50','c4d5 f6d5 f3h4 d5c3 b2c3 f5e4 f2f3 e7e6 g2g3 e4g6 e1f2 a7a6 h4g6 h7g6 f2g2 b7b5 e3e4 f8e7 h2h4 d7b6 c1e3 c6c5 a1c1 a8c8 f1e2 d8d7 d1b3 c5d4 c3d4 e8g8 d4d5 b6a4 d5e6 d7e6 b3e6 f7e6 f3f4 e7a3 c1c8 f8c8 e2g4 g8f7 h1d1 c8c2 d1d2 c2c3');
INSERT INTO evaluations VALUES(24553,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(24553,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(24553,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(24553,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(24553,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(24553,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(24553,36,'cp 48','f6d5 f1d3 f5d3 d1d3 e7e6 e3e4 d5c3 b2c3 f8d6 a2a4 h7h6 c1e3 d8c7 e1g1 e8g8 g2g3 a8c8 f1d1 f8d8 c3c4 e6e5 d4d5 c6d5 c4d5 d7c5 d3c2 c5d7 c2e2 c7c2 e2c2 c8c2 f3h4 d7c5 f2f3 d6f8 h4f5 b7b6 d1c1 c2c1 a1c1 c5a4 c1c7 a7a5');
INSERT INTO evaluations VALUES(42682,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(42682,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(42682,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(42682,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(42682,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(42682,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(42682,36,'cp 42','c5b4 f1d3 d7d5 e4e5 f6e4 e1g1 e4c3 b2c3 b4c3 a1b1 b8c6 c1f4 c3d4 f3d4 c6d4 d1h5 d4e6 f4e3 d8e7 f1f2 g7g6 h5f3 e8g8 f3d5 f8e8 d5f3 e6d8 e3f4 c8d7 b1f1 d8c6 f1e1 a8d8 f3g3 d7e6 h2h3 e6a2 f4g5 e7b4 f2f1 d8d5 g5f6 b4c5 g1h1 a2c4 e5e6 e8e6');
INSERT INTO evaluations VALUES(42818,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(42818,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(42818,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(42818,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(42818,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42818,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(42818,36,'cp 48','f1d3 d7d5 e4e5 f6e4 e1g1 e4c3 b2c3 b4c3 a1b1 b8c6 c1f4 c3d4 f3d4 c6d4 f4e3 d4e6 d1f3 e8g8 f3g3 d8e7 f1f6 g7g6 e3h6 e6g7 h6g7 g8g7 h2h4 c8d7 h4h5 a8e8 b1e1 g7h8 g3f4 d7e6 f4h6 e6f5 d3f5 g6f5 f6f5 f8g8 e1f1 g8g7 h6f4 e8g8 f1f2 e7c5 g1h1 g7g4 f4h6 c5e7 f5f7');
INSERT INTO evaluations VALUES(55800,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55800,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55800,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55800,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(55800,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(55800,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(55800,36,'cp 7','g8f6 g1f3 f8e7 b2b3 e8g8 c1b2 c7c5 c4d5 f6d5 f1e2 b8c6 e1g1 e7f6 d1c2 c6b4 c2c1 d5c3 d2c3 b4d3 e2d3 d8d3 c3c4 f6b2 c1b2 f8d8 b2c1 f7f6 f1d1 d3d1 c1d1 d8d1 a1d1 g8f7 d1d8 f7e7 d8h8 h7h6 f3d2 b7b6 g1f1 c8b7');
INSERT INTO evaluations VALUES(48235,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48235,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48235,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(48235,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48235,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48235,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48235,36,'cp 0','g1f3 b7b6 c4d5 e6d5 b2b3 f8d6 c3b5 d6e7 c1b2 a7a6 b5c3 c8b7 c3e2 e8g8 g2g3 b8d7 f1g2 c7c5 e1g1 e7d6 f3h4 f8e8 d2d4 a8c8 a1c1 d6f8 d4c5 b6c5 e2c3 d7e5 c3a4 d8e7 g2h3 c8c6 h3g2');
INSERT INTO evaluations VALUES(5474,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(5474,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(5474,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(5474,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5474,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(5474,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5474,36,'cp 0','e1g1 e7e6 d2d4 c5d4 c3b5 d6d5 e4d5 e6d5 c4d5 g8e7 d5c6 b7c6 b5d4 e8g8 c2c3 c8b7 f3e5 c6c5 d4b3 d8c7 c1e3 e7d5 e3c5 d5f4 c5d6 c7b6 d6c5');
INSERT INTO evaluations VALUES(5473,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5473,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5473,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(5473,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5473,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5473,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(5473,36,'cp 0','e7e6 d2d4 c5d4 c3b5 g8e7 c4b3 c6a5 b5d4 a5b3 a2b3 e6e5 d4e2 e8g8 e2c3 c8e6 d1d3 e5f4 c1f4 d6d5 f4g5 d5e4 d3d8 f8d8 c3e4 g8f8 e4f6 h7h6 f6h7 f8e8 h7f6');
INSERT INTO evaluations VALUES(5556,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5556,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(5556,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(5556,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5556,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(5556,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5556,36,'cp -19','d2d4 c5d4 c3b5 g8e7 c4b3 c6a5 b5d4 e8g8 c1e3 b7b6 c2c3 c8a6 f1f2 e6e5 d4e2 d6d5 f3e5 d5e4 e2g3 d8d1 a1d1 a5b3 a2b3 f8d8 f2d2 d8d2 d1d2 f7f6 e5c4 f6f5 d2d6 g8f7 g3e2 f7e8 e2d4 g7d4 d6d4 a6c4 b3c4 a8d8 g2g3 e7c6 d4d5 e8e7 d5b5 d8d1 g1f2 c6a5 b2b3');
INSERT INTO evaluations VALUES(7016,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(7016,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(7016,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(7016,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(7016,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(7016,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(7016,36,'cp 26','d7d6 f1b5 d6e5 f3e5 g7e5 b5c6 b7c6 d4e5 d8d1 e1d1 h7h5 h2h4 g8h6 c1h6 h8h6 d1d2 c8f5 d2c3 e8c8 b1d2 d8d5 h1e1 h6h8 b2b3 h8d8 e1e2 c6c5 a1e1 c8b7 g2g3 b7c6 d2f3 e7e6 f3d2 d8d7 f2f3 d5d4 d2c4 d4d1 c4b2 d1d5 e2f2 c6b5 e1e3 d7d8 b2d3');
INSERT INTO evaluations VALUES(5689,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(5689,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(5689,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(5689,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(5689,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(5689,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(5689,36,'cp 21','f1b5 d6e5 f3e5 g7e5 b5c6 b7c6 d4e5 d8d1 e1d1 h7h5 h2h4 g8h6 c1h6 h8h6 d1d2 c8f5 d2c3 e8c8 b1d2 d8d5 d2c4 h6h8 h1d1 h8d8 d1d2 d5d2 c4d2 d8d5 f2f4 c6c5 g2g3 c8d7 a1e1 d7c6 d2f3 f7f6 e5f6 e7f6 f3d2 d5d6 a2a3 c6d5');
INSERT INTO evaluations VALUES(13561,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(13561,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(13561,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(13561,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(13561,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(13561,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(13561,36,'cp 44','d4c5 f8c5 b2b4 c5e7 b1d2 e8g8 c1b2 a7a6 a1c1 d5c4 f1c4 b7b5 c4e2 c8b7 e1g1 a8c8 d2b3 d8b6 b3c5 f8d8 e2d3 e7c5 c1c5 h7h6 d1b1 f6d7 c5c1 c6e7 h2h3 b7f3 c1c8 d8c8 g2f3 d7f6 b1a1 e7d5 f1c1 c8c1 a1c1 b6c7 c1c7 d5c7');
INSERT INTO evaluations VALUES(13493,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(13493,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13493,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(13493,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13493,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13493,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(13493,36,'cp 36','f8c5 b2b4 c5e7 c1b2 e8g8 b1d2 a7a6 a1c1 d5c4 f1c4 b7b5 c4e2 c8b7 e1g1 a8c8 d2b3 f6e4 e2d3 d8d5 d3c2 d5d1 f1d1 c6b8 b3a5 b7d5 c2d3 e4d6 f3d2 b8c6 e3e4 c6a5 e4d5 a5c4 d2c4 d6c4 d3c4 c8c4 c1c4 b5c4 d5e6 f7e6');
INSERT INTO evaluations VALUES(8356,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(8356,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(8356,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(8356,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(8356,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(8356,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(8356,36,'cp 24','b2b4 c5e7 c1b2 e8g8 b1d2 a7a6 a1c1 d5c4 f1c4 b7b5 c4e2 c8b7 e1g1 a8c8 d2b3 f6e4 f3d2 e4d2 b3d2 c6a7 b2d4 a7c6 d4a1 c6b8 d2b3 d8d1 f1d1 b7d5 b3c5 e7c5 b4c5 b8c6 a1c3 f8d8 f2f3 e6e5 g1f2 d5e6 g2g3 f7f6 d1d8 c6d8');
INSERT INTO evaluations VALUES(8354,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(8354,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(8354,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(8354,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8354,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8354,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(8354,36,'cp 22','c5e7 c1b2 e8g8 b1d2 a7a6 c4d5 f6d5 a1c1 c8d7 f1c4 e7f6 d1b3 f6b2 b3b2 d8f6 b2b3 d5b6 c4e2 c6e7 e1g1 d7a4 b3b1 a4c6 e2d3 g7g6 f3d4 c6d5 d4b3 b6a4 d2e4 d5e4 d3e4 e7d5 e4d5 e6d5 b3d4 a8c8');
INSERT INTO evaluations VALUES(8098,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(8098,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(8098,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(8098,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(8098,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(8098,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(8098,36,'cp 25','c1b2 e8g8 b1d2 a7a6 c4d5 d8d5 f1e2 b7b5 e1g1 f8d8 d2b3 c8b7 a1c1 a8c8 d1d5 f6d5 f1d1 d5b6 d1d8 c6d8 c1c8 b7c8 f3d2 c8b7 f2f4 b6d7 b2d4 f7f6 b3a5 b7a8 e2g4 g8f7 g1f2 e6e5 g2g3 e5d4 g4d7 d4e3 f2e3 a8d5 e3d4');
INSERT INTO evaluations VALUES(23499,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(23499,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(23499,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(23499,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(23499,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(23499,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(23499,36,'cp 136','g1f3 c7c5 b1c3 d7d6 d2d4 c5d4 d1d4 a6b8 c1f4 b8c6 d4d2 e7e5 f4g5 g8f6 g5f6 d8f6 c3d5 f6d8 f1c4 c8e6 e1g1 h7h6 f1d1 a8c8 a2a4 f8e7 d5e7 e8e7 c4e6 f7e6 c2c3 d8c7 d2e2 h8d8 h2h4 e7f8 h4h5 a7a6');
INSERT INTO evaluations VALUES(23498,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(23498,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(23498,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(23498,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(23498,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(23498,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(23498,36,'cp 102','e7e6 b1c3 d7d5 f1a6 b7a6 e4d5 e6d5 e1g1 g8e7 b2b3 e7c6 f1e1 f8e7 c1a3 e8g8 a3e7 c6e7 d2d4 d8d6 d1d2 f8e8 c3a4 c8d7 a4c5 e7g6 h2h3 d7f5 a1c1 f5e4 f3g5 e4f5 e1e3 e8e3 d2e3 f5d7 c5d7 d6d7 c1e1 h7h6 g5f3 d7d6 e3d3 g6f4 d3d2');
INSERT INTO evaluations VALUES(23881,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(23881,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(23881,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(23881,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(23881,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(23881,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(23881,36,'cp 216','e1g1 b4a5 c1g5 d8g5 c4f7 e8e7 f3g5 g4d1 a1d1 h7h6 f7g8 h8g8 g5h3 e7d7 h3f4 g8f8 c3d5 a5b6 f4e2 g7g5 f2f3 c6d8 d5e3 c7c6 g2g3 d8e6 g1g2 h6h5 f3f4 g5f4 g3f4 a8e8 e4e5 b6d4 e2d4 e6f4 g2g3 e8e5 f1f4 e5e3 g3f2 f8f4 f2e3 f4h4');
INSERT INTO evaluations VALUES(23880,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(23880,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(23880,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(23880,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(23880,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(23880,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(23880,36,'cp 195','b4a5 c1g5 d8g5 f3g5 g4d1 c4f7 e8e7 a1d1 h7h6 f7g8 h8g8 g5h3 e7d7 h3f4 g8f8 f4d5 g7g5 g2g3 c6e7 d5e7 a5c3 e7g6 f8g8 b2c3 g8g6 f2f4 g5f4 f1f4 b7b5 d1b1 a7a6 g1f2 c7c5 f2e3 d7c6 b1f1 c6b6 e3d3 a8e8 g3g4 g6g5');
INSERT INTO evaluations VALUES(25283,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(25283,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(25283,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(25283,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(25283,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(25283,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(25283,36,'cp 77','d2d4 f6d7 c2c3 f8e7 h2h3 g4h5 c1e3 a7a6 b5a4 e5d4 c3d4 e8g8 b1c3 d7b6 a4b3 c6a5 b3c2 a5c4 e3c1 d6d5 c3d5 b6d5 e4d5 c4b6 c2b3 a6a5 a2a3 f8e8 e1e5 h5g6 c1f4 e7d6 b3a2 d8d7 a1c1 g6e4 f3g5 d6e5 d4e5 e4d5');
INSERT INTO evaluations VALUES(31218,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(31218,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(31218,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(31218,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(31218,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(31218,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(31218,36,'cp 57','h2h3 g4h5 e4g3 h5f3 d1f3 g8f6 f1c4 e7e6 c2c3 b8d7 d2d4 f8d6 c4d3 e8g8 e1g1 f8e8 f1e1 h7h6 a2a4 a7a5 d3c2 d8c7 g3e4 f6e4 f3e4 d7f6 e4f3 f6d5 g2g3 a8d8 c1d2 c6c5 d4c5 d6c5 a1d1 b7b6');
INSERT INTO evaluations VALUES(31217,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(31217,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(31217,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(31217,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(31217,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(31217,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(31217,36,'cp 55','g4h5 e4g3 h5f3 d1f3 e7e6 c2c3 g8f6 f1c4 f8e7 e1g1 e8g8 d2d4 a7a5 f1e1 b8d7 c4d3 f8e8 c1e3 d8b6 b2b3 c6c5 g3e4 f6d5 e4c5 d7c5 d4c5 b6c7 e3d4 e7c5 d4e5 c5d6 e5d6 c7d6 d3b5 e8b8 a1d1 d6c7 c3c4 d5f6 d1d3 e6e5 e1d1 e5e4 f3g3 c7g3 d3g3');
INSERT INTO evaluations VALUES(38096,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(38096,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(38096,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(38096,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(38096,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(38096,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(38096,36,'cp 125','e2e4 d6b4 f1d3 d5e4 d3e4 g8f6 e4c2 c6c5 a2a3 b4c3 b2c3 c5d4 c3d4 b7b5 c1g5 b8d7 c4c5 d8a5 g5d2 a5c7 d1e2 a7a6 a3a4 c8b7 e1g1 e8g8 a4b5 a6b5 e2b5 b7c6 b5e2 f8b8 f1c1 h7h6 h2h3 f6d5 e2d3 d5f6 c1e1 c6e4 d3c4 a8a1 e1a1 e4f3 g2f3 f6d5');
INSERT INTO evaluations VALUES(38098,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(38098,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(38098,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(38098,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(38098,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(38098,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(38098,36,'cp 119','d6b4 f1d3 d5e4 d3e4 g8f6 e4c2 c6c5 a2a3 b4c3 b2c3 c5d4 c3d4 b7b5 c1g5 b8d7 c4c5 d8a5 g5d2 a5c7 e1g1 c8b7 a3a4 a7a6 d1e2 e8g8 a4b5 a6b5 e2b5 b7c6 b5e2 h7h6 h2h3 f8e8 a1a8 e8a8 f1b1 a8a2 c2b3 c6f3 g2f3 a2a8 e2b5 a8b8 b5d3 d7f8 b3c4 b8b1 d3b1 f8g6');
INSERT INTO evaluations VALUES(39638,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(39638,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(39638,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(39638,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(39638,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(39638,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(39638,36,'cp 56','d2d4 d7d5 e2e4 d5e4 c3e4 g8f6 e4f6 e7f6 f1c4 e8g8 e1g1 c8f5 h2h3 c7c6 c2c3 b8d7 f1e1 f8e8 e1e8 d8e8 a2a4 f5e6 c4e6 e8e6 a4a5 a8e8 c1f4 e6f5 f4e3 f5b5 b2b3 f6f5 c3c4 b5a6 d1d2 d7f6 a1e1 f6e4');
INSERT INTO evaluations VALUES(45114,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(45114,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(45114,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(45114,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45114,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(45114,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45114,36,'cp 34','f6e4 c5d6 b8c6 d6e7 d8d1 e1d1 f8e7 c1e3 e8g8 a2a3 c8g4 d1c1 f8d8 f1d3 g4f3 g2f3 e4c5 d3c4 c6a5 c4a2 a8c8 b1d2 c5d3 c1b1 d3f2 h1f1 f2h3 b2b4 a5c6 d2e4 h7h6 b1b2 h3g5 a1d1 b7b6 d1d8 c8d8 e4g5 e7g5 e3g5 h6g5 f1g1 d8d2 g1g5 c6d4 g5d5 d2c2 b2b1 c2d2 d5d8 g8h7 a2f7 h7h6 d8d7 g7g6');
INSERT INTO evaluations VALUES(45265,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(45265,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(45265,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(45265,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45265,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45265,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(45265,36,'cp 26','c5d6 b8c6 d6e7 d8d1 e1d1 f8e7 c1e3 e8g8 d1c1 f8d8 a2a3 c8g4 f1d3 g4f3 g2f3 e4c5 d3c4 c6a5 c4a2 a8c8 b1c3 e7f6 e3c5 c8c5 c3e4 c5f5 e4f6 f5f6 h1d1 f6d6 d1d6 d8d6 b2b4 a5c6 c1b2 g8f8 a1f1 c6e5 f3f4 e5f3 a2b3 d6d2 b2c1 h7h5 f1d1 d2d1 c1d1 f3h2 b3d5 b7b6 d1e1 g7g6 e1e2');
INSERT INTO evaluations VALUES(45109,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(45109,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(45109,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45109,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(45109,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(45109,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(45109,36,'cp 24','b8c6 d6e7 d8d1 e1d1 f8e7 c1e3 e8g8 d1c1 f8d8 a2a3 c8g4 f1d3 g4f3 g2f3 e4c5 d3e4 c6e5 b1c3 c5e4 f3e4 e5g4 a3a4 e7d6 h2h3 g4e3 f2e3 d8e8 h1d1 d6b4 d1d4 b4c5 d4d7 c5e3 c1b1 a8b8 b1a2 e3g5 a1d1 g7g6 b2b4 b8d8 a2b3 d8d7 d1d7 e8e7 d7e7 g5e7 c3d5 e7d6 b3c4');
INSERT INTO evaluations VALUES(13208,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(13208,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(13208,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(13208,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(13208,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(13208,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(13208,36,'cp 39','d6e7 d8d1 e1d1 f8e7 c1e3 e8g8 d1c1 c6b4 c2c3 b4d5 e3d4 e4c5 g2g3 f8d8 f1c4 c8f5 h1e1 a8c8 a2a4 a7a6 a4a5 c5d3 c4d3 f5d3 c1d2 d3f5 d2e2 f7f6 e2f1 g8f7 b1d2 f5g4 f1g2 g4f5 d2e4 g7g5 a1d1 g5g4 f3d2 f5g6');
INSERT INTO evaluations VALUES(47671,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(47671,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(47671,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(47671,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(47671,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(47671,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(47671,36,'cp -16','d5c4 b1c3 f8b4 h4f6 d8f6 d1a4 b8c6 a2a3 b4c3 b2c3 c8d7 a4c4 e8g8 e2e3 a8c8 f1d3 c6a5 c4b4 b7b6 f3e5 f8d8 e5d7 d8d7 b4a4 c7c6 a1d1 g7g6 h2h4 h6h5 g2g3 c8d8 e1g1 c6c5 d3a6 f6e7 d1b1 g8g7 a6b5 d7c7 f1d1 e7f6 b5a6 g6g5 h4g5 f6g5');
INSERT INTO evaluations VALUES(50591,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(50591,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(50591,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(50591,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(50591,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(50591,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(50591,36,'cp -51','e7e6 c4b3 b7b5 d2d3 c8b7 b1d2 b8c6 g1e2 g8f6 e1g1 h7h5 c2c3 d8c7 f3g3 c7g3 h2g3 f8e7 f1d1 d7d6 d2f3 e8g8 c1g5 c6a5 b3c2 f8c8 a1c1 g7g6 d3d4 a5c4 b2b3 c4a3 d4c5 c8c5 g5e3 a3c2 e3c5 f6e4 c1c2 e4c5');
INSERT INTO evaluations VALUES(51473,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(51473,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(51473,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(51473,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(51473,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(51473,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(51473,36,'cp 132','b1c3 c8b7 d2d4 f8g7 h2h4 h7h6 c1f4 g8f6 e4e5 f6d5 c3d5 b7d5 c2c4 d5b7 f1d3 e7e6 d4d5 c7c6 d5d6 c6c5 d1e2 a7a6 e1f1 b7f3 e2f3 b8c6 f3e3 b6b5 c4b5 a6b5 e3c5 a8a4 f4e3 c6e5 d3b5 a4a8 a2a4 e8g8 e3h6 g7h6 c5e5 h6g7 e5e2');
INSERT INTO evaluations VALUES(51471,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(51471,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(51471,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(51471,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(51471,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(51471,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(51471,36,'cp 137','f8g7 d2d4 d7d6 e4e5 d6e5 f3e5 c7c6 c1f4 g8f6 d1f3 c8e6 h2h4 h7h5 f1d3 e8g8 f3e2 e6g4 e2e3 d8c8 a2a3 g4f5 d3f5 c8f5 e1c1 f8d8 c1b1 f6g4 e5g4 f5g4 e3e7 b8d7 e7e3 d8e8 e3g3 g4g3 f4g3 a8d8 f2f3');
INSERT INTO evaluations VALUES(37592,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(37592,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(37592,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(37592,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(37592,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(37592,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(37592,36,'cp 137','d2d4 a7a6 h2h4 h7h6 e4e5 c8b7 c1f4 e7e6 f1d3 b8c6 e1g1 g8e7 d3e4 c6a5 e4b7 a5b7 c3e4 d7d5 e4g3 b7a5 b2b3 d8d7 f1e1 g7f8 d1e2 d7b5 e2d2 e8c8 c2c3 c8b7 e1b1 d8c8 f3e1 b5d7 f4e3 e7c6 g3e2 b7a7 g2g4 h8g8 e3h6');
INSERT INTO evaluations VALUES(52680,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(52680,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(52680,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(52680,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(52680,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(52680,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(52680,36,'cp 119','c2c4 b8d7 b1c3 e7e5 h2h3 f8e7 d1c2 g8f6 f1e2 e8g8 c1e3 e5d4 f3d4 f8e8 e1g1 e7f8 b2b4 d8c7 f1d1 a7a6 a2a4 a6a5 b4b5 d7c5 e2f3 c8d7 a1c1 a8d8 c1b1 d7c8 g2g3 f6d7 f3g2 d7e5 c3e2 c7e7 f2f4');
INSERT INTO evaluations VALUES(52676,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(52676,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(52676,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(52676,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(52676,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(52676,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(52676,36,'cp 112','c8g4 d1b3 d8b6 d4d5 g7g6 c1e3 b6b3 a2b3 g4f3 g2f3 g8f6 b1c3 f8g7 e1d2 e8g8 h1g1 b8d7 d2c2 c6d5 e4d5 f8c8 f3f4 a7a6 f4f5 g6g5 f1e2 b7b5 a1a5 g5g4 g1a1 b5b4 c3a4 h6h5 a4b6 d7b6 e3b6 f6d7 b6e3');
INSERT INTO evaluations VALUES(56584,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(56584,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(56584,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(56584,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(56584,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(56584,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(56584,36,'cp 235','f3e5 g8e7 f2f4 a7a5 g2g3 a5a4 f1g2 g7g6 e5f3 f8g7 b1d2 e8g8 e1g1 b8a6 a2a3 a6c5 d1e2 b7b5 f3h4 a8b8 e2f2 b5b4 a3b4 b8b4 h4f3 e7c6 e4e5 c8e6 f3g5 c6e5 g5e6 f7e6 f2e1 e5d7 d2c4 b4b8 c1d2 g8h8 h2h3 f8f7 g1h2');
INSERT INTO evaluations VALUES(56514,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(56514,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(56514,32,'cp 215',NULL);
INSERT INTO evaluations VALUES(56514,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(56514,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(56514,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(56514,36,'cp 224','g8e7 f2f4 f7f6 e5f3 c7c5 a2a4 b8c6 b1a3 c8e6 f1e2 g7g6 c1d2 f8g7 e1g1 e8g8 d1e1 b7b6 e1g3 g8h8 a1e1 d8d7 b2b3 c6b4 h2h4 e6g4 f3h2 g4e2 e1e2 f6f5 g3f3 f5e4 d3e4 d4d3 c2d3 d7d3 f3d3 b4d3');
INSERT INTO evaluations VALUES(57706,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(57706,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(57706,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(57706,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(57706,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(57706,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(57706,36,'cp 72','g5g4 f3e5 h7h5 d2d4 d7d6 e5c4 a7a6 a2a4 e7e6 b1c3 b8c6 h2h3 g4h3 h1h3 h5h4 c1e3 e6e5 g2g4 e5d4 e3d4 c8g4 d1g4 c6d4 e1c1 g8f6 g4g2 h8g8 g2h1 f6g4 d1d4 g4f2 h1f3 f2h3 f1h3 g8g3 f3f5 f8g7 e4e5 d8g5 f5g5 g3g5 c3d5 d6e5');
INSERT INTO evaluations VALUES(57709,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(57709,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(57709,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(57709,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(57709,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(57709,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(57709,36,'cp 79','f3e5 h7h5 d2d4 d7d6 e5c4 a7a6 c1g5 f8g7 c2c3 b8d7 b1d2 b7b5 c4e3 c7c5 e3f5 g7f6 g5e3 d7b6 d4c5 d6c5 d2b3 c8f5 e4f5 b6a4 b3c5 a4c5 e3c5 d8c7 c5e3 f6e5 f1d3 a8d8 a2a4 b5b4 c3b4 e5b2 a1b1 b2c3 e1f1 g8f6 d1c2');
INSERT INTO evaluations VALUES(3609,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(3609,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(3609,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(3609,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(3609,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(3609,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(3609,36,'cp -35','c6a5 c4b3 c7c6 h2h3 a5b3 a2b3 a7a5 c1e3 c5e3 f2e3 b7b5 d1e1 c8e6 g1h2 f8e8 e1g3 g8h7 g3h4 f6d7 h4g3 d8e7 g3f2 d7f6 f2h4 h7g8 h4g3 e8f8 h2g1 e7d8 g3f2 d8b6 f3h4 d6d5 f2g3 f6h5 g3e5');
INSERT INTO evaluations VALUES(5874,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(5874,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(5874,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(5874,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(5874,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(5874,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(5874,36,'cp 26','f1d3 c8g4 e1g1 g8f6 f1e1 b4e7 d3b5 e8g8 b5c6 b7c6 h2h3 g4f3 d1f3 f8e8 c1f4 a7a5 b2b3 h7h6 a1d1 e7d6 e1e8 f6e8 f4e5 d8g5 d1e1 d6b4 f3d3 e8d6 e5d6 b4d6 c3a4 h6h5 e1e3 d6b4 c2c4 d5c4 b3c4 h5h4 a4c3 g5f4 c4c5 f4f6');
INSERT INTO evaluations VALUES(16260,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(16260,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(16260,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(16260,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(16260,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(16260,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(16260,36,'cp 99','e4e5 d7d5 c4e2 f6d7 d2d4 f8e7 e1g1 e8g8 f1e1 c6a5 c2c3 c7c5 c1e3 c5d4 c3d4 f7f6 e5f6 d7f6 h2h3 a5c6 e2d3 c6b4 b1c3 b4d3 d1d3 e7d6 a2a4 c8d7 f3e5 d7e8 e3g5 h7h6 g5h4 a7a6 f2f4 d8c7 a4a5 a8c8 f4f5');
INSERT INTO evaluations VALUES(26975,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(26975,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(26975,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(26975,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(26975,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(26975,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(26975,36,'cp 91','d1f3 d8d7 c1e3 g7g6 b1d2 f7f5 e4f5 d7f5 e1g1 g8e7 c2c3 f8g7 b2b4 f5f3 d2f3 e8d7 a2a4 c6d8 b4b5 d8e6 a1e1 a7a6 c4e6 d7e6 d3d4 e5e4 f3g5 e6d7 g5e4 a6b5 a4b5 h8e8 e3g5 e7d5 c3c4 d5b6 e4d2 c7c6 c4c5 e8e1 f1e1 d6c5');
INSERT INTO evaluations VALUES(26974,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(26974,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(26974,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(26974,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(26974,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(26974,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(26974,36,'cp 84','d8d7 c1e3 g7g6 e1g1 f8g7 a2a4 g8f6 c2c3 e8g8 a4a5 a7a6 b1d2 g8h8 f3d1 f6g8 f2f4 g7h6 d1e2 h6f4 e3f4 e5f4 f1f4 a8e8 d3d4 d6d5 c4d3 f7f5 e4e5 c6d8 f4f2 d8e6 a1f1 g8h6 b2b4 c7c6 d2b3 d7c7 g2g4 e8e7 b3c5 e6c5 b4c5 c7a5');
INSERT INTO evaluations VALUES(41404,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(41404,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(41404,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(41404,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(41404,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(41404,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(41404,36,'cp 79','e4e5 f6d7 h2h4 d7b6 a2a4 d6e5 a4a5 b6d5 d4e5 c8f5 f1c4 e7e6 d1e2 b8c6 b1c3 a7a6 e1g1 d5c3 b2c3 f5g4 c1e3 e8g8 e2e1 b7b6 a5b6 c7b6 a1a6 d8c7 a6a8 f8a8 c4d3 c6e7 f3d4 c7d8 d4b5 a8a2 e3d4 e7f5 d4f2 d8a8 e1e4 a8e4 d3e4');
INSERT INTO evaluations VALUES(41200,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(41200,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(41200,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(41200,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(41200,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(41200,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(41200,36,'cp 57','f6d5 c2c4 d5b6 h2h3 e8g8 b1c3 c7c5 d4d5 e7e6 f1e2 e6d5 c4d5 d6e5 f4e5 b8d7 c1f4 f8e8 e5e6 f7e6 d5d6 b6d5 c3d5 e6d5 d1d5 g8h8 d5d2 d8f6 e1g1 e8e2 d2e2 f6f4 e2e7 g7f8 e7e8 d7b6 f3e5 c8d7 e8a8 f4d4 g1h1 b6a8 f1f8 h8g7 f8f7 g7g8 f7d7 d4e5 d7d8 g8g7 d6d7 e5d5');
INSERT INTO evaluations VALUES(43034,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(43034,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(43034,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(43034,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(43034,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(43034,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(43034,36,'cp -8','b1c3 c7c6 a2a4 e8g8 g1e2 d7d6 d2d4 c8g4 f3e3 e5d4 e3d4 f8e8 e1g1 b8d7 h2h3 d6d5 e4d5 e7c5 d4d1 g4h5 c1g5 d7b6 c4d3 b6d5 c3d5 c6d5 d3b5 e8e4 b5d3 e4e6 g2g4 h5g6 e2f4 e6e8 f4g6 h7g6 c2c3');
INSERT INTO evaluations VALUES(2849,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(2849,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(2849,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(2849,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(2849,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(2849,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(2849,36,'cp 25','h2h3 a7a5 b2b3 c8e6 d1e2 d6d5 c4d3 d5e4 c3e4 e6f5 c1b2 f8e8 b2f6 e7f6 a1d1 f5e4 d3e4 d8e7 d1e1 e7c5 e2f3 a8d8 g2g3 e8e5 e4d3 d8d6 e1e4 g7g6 e4c4 c5b6 f1d1 h7h5 h3h4 g8g7 g1g2 e5e6 c4f4 b6c5 d3c4 e6e5 d1d3 e5e1 g2h2 d6d3 c4d3');
INSERT INTO evaluations VALUES(15049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15049,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(15049,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(15049,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15049,36,'cp -13','b1c3 f8d6 c3b5 c8g4 b5d6 d8d6 f1e2 e8g8 e1g1 f8e8 c1d2 f6e4 a1c1 c6e7 d2e1 c7c6 f3e5 g4e2 d1e2 d6e6 f2f3 e4d6 e1d2 f7f6 e5d3 e7f5 e2f2 d6c4 d3c5 e6f7 f1e1 c4b2 c1b1 b7b6 b1b2 b6c5 d4c5');
INSERT INTO evaluations VALUES(33939,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(33939,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(33939,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(33939,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(33939,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(33939,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(33939,36,'cp -70','b8c6 f1e2 c8f5 a2a3 e8c8 e1g1 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 c5e3 f2e3 f5f6 c2c3 h8e8 a1d1 d8d1 f1d1 f6e6 g1f2 g7g6 h2h3 a7a6 c3c4 e6f6 b2b4 c8b8 f2g1 a6a5 f3d4 c6d4 d1d4 f6e5 b4a5 e5a5');
INSERT INTO evaluations VALUES(38686,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(38686,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(38686,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(38686,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(38686,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(38686,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(38686,36,'cp -27','g8f6 b1c3 e8g8 d2d4 b8d7 f1e2 a7a6 c1g5 h7h6 g5h4 b7b6 f3d2 c8b7 e2f3 d8c8 e1g1 c7c5 c3e4 f6e4 d2e4 d6b8 h4g3 b8g3 h2g3 c5d4 d1d4 c8c7 a1d1 f8d8 d4d6 a8c8 d1d2 b7e4 d6c7 c8c7 f3e4 b6b5 f1d1');
INSERT INTO evaluations VALUES(41840,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(41840,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(41840,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(41840,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(41840,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(41840,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(41840,36,'cp -32','f1e2 c7c5 d2d4 e8g8 b1c3 h7h6 d4c5 d6c5 d1d8 f8d8 e1g1 c8d7 f1d1 b8c6 a2a3 a8c8 c1f4 c5b6 g1f1 a7a6 d1d3 d7e8 a1d1 d8d3 d1d3 b6c5 d3d1 c5e7 f3e5 g7g5 f4g3 c6e5 g3e5 e8c6 a3a4 g8f8 a4a5 f6d7');
INSERT INTO evaluations VALUES(48529,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(48529,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(48529,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(48529,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(48529,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(48529,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(48529,36,'cp -99','f3e5 f8b4 f1c4 b8d7 e1g1 d7e5 d4e5 d8d1 f1d1 b4c3 b2c3 f6d7 c4e2 d7e5 c1f4 e5g6 f4c7 c8d7 a2a4 a8c8 c7d6 d7c6 e2b5 g6e7 d1b1 a7a6 d6e7 e8e7 b5c6 c8c6 b1b7 e7f6 a1e1 c6c4 a4a5 h8c8 h2h4 c4a4 b7b4 c8c4 e1b1 e4e3 f2e3 c4h4 b4b7 f6g6 b1f1 f7f6 f1d1 a4a5 d1d7');
INSERT INTO evaluations VALUES(48435,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(48435,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(48435,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(48435,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(48435,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(48435,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(48435,36,'cp -103','f8b4 f1c4 c7c5 c1e3 e8g8 e1g1 b4c3 b2c3 f6d5 d4c5 f7f6 e5g4 b8c6 f2f3 d8e7 c4d5 e6d5 e3d4 c8g4 f3g4 a8c8 h2h3 f8d8 a1b1 h7h6 a2a4 c6e5 f1f5 e5c4 h3h4 e4e3 d1e2 e7d7 b1f1 d8f8 g4g5 f6g5 f5f8 c8f8 f1f8 g8f8 d4e3 c4e3');
INSERT INTO evaluations VALUES(49528,30,'cp -313',NULL);
INSERT INTO evaluations VALUES(49528,31,'cp -312',NULL);
INSERT INTO evaluations VALUES(49528,32,'cp -297',NULL);
INSERT INTO evaluations VALUES(49528,33,'cp -320',NULL);
INSERT INTO evaluations VALUES(49528,34,'cp -316',NULL);
INSERT INTO evaluations VALUES(49528,35,'cp -327',NULL);
INSERT INTO evaluations VALUES(49528,36,'cp -323','d7d5 e4d5 c8g4 f2f3 g4h5 b1c3 f8e7 g1h3 c7c6 d2d4 d8d7 h3f2 e5d4 d1d4 h5f3 e1g1 c6d5 d4d3 d5c4 d3f3 b8c6 c1e3 e8g8 a1d1 c6e5 f3g2 d7c6 f2e4 f6e4 g2e4 c6e4 c3e4 f8d8 d1d8 e7d8 f1d1 d8c7 g1g2 f7f6 e3f4 a8e8 f4e5 c7e5 d1d7 b7b6 a2a4 f6f5 e4d6 e5d6 d7d6');
INSERT INTO evaluations VALUES(52468,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(52468,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(52468,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(52468,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(52468,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(52468,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(52468,36,'cp -24','d7d5 d2d4 e7e6 g1f3 b8c6 b2b3 g8f6 f1d3 f8d6 e1g1 e8g8 c3b5 d6e7 c2c4 d5c4 b3c4 b7b6 c1b2 c8b7 a1c1 a8c8 d1e2 a7a6 b5a3 h7h6 a3c2 d8c7 f1d1 f8d8 h2h3 c6a5 f3e5 e7d6 c2e1 c5d4 e3d4 d6e5 d4e5 f6d7 e1f3');
INSERT INTO evaluations VALUES(53308,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(53308,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(53308,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(53308,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(53308,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(53308,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(53308,36,'cp -12','g1f3 e7e6 d2d4 g8f6 d4c5 f8c5 f1d3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 c1e3 c5e3 f2e3 d8d6 a1c1 f8d8 d3b1 d6d1 f1d1 d8d1 c1d1 g8f8 b1c2 a8c8 c2a4 e6d7 a4c6');
INSERT INTO evaluations VALUES(53309,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(53309,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(53309,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(53309,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(53309,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(53309,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(53309,36,'cp -30','e7e6 d2d4 g8f6 d4c5 f8c5 f1d3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 h2h3 h7h6 a2a3 a7a5 c2c3 d4c3 e2c3 c8e6 c1e3 c5e3 f2e3 d8d6 a1c1 f8d8 d3b1 d6d1 f1d1 d8d1 c1d1 g8f8 b1c2 f6e8 c2a4 f7f6 c3d5 a8d8 d1c1 e6d5 e4d5 c6e7 e3e4');
INSERT INTO evaluations VALUES(54886,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(54886,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(54886,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(54886,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(54886,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(54886,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(54886,36,'cp 80','g1f3 c7c5 f1c4 c5d4 d1b3 e7e6 e1g1 a7a6 f3d4 g8f6 f1d1 d8c7 b1c3 f8g7 c3a4 b8d7 c4e6 e8g8 e6d7 f6d7 a4c3 d7c5 b3b4 f8d8 e3e4 c8g4 f2f3 a6a5 b4c4 d8d4 d1d4 c5e6 e4e5 c7b6 c1e3 e6d4 e3d4 g4e6 d4b6 e6c4 a1e1 c4e6 g1f2 a5a4');
INSERT INTO evaluations VALUES(13850,30,'cp 298',NULL);
INSERT INTO evaluations VALUES(13850,31,'cp 311',NULL);
INSERT INTO evaluations VALUES(13850,32,'cp 318',NULL);
INSERT INTO evaluations VALUES(13850,33,'cp 311',NULL);
INSERT INTO evaluations VALUES(13850,34,'cp 328',NULL);
INSERT INTO evaluations VALUES(13850,35,'cp 346',NULL);
INSERT INTO evaluations VALUES(13850,36,'cp 348','f3g5 d6d5 e4d5 c6d4 e1g1 h7h6 g5e4 f6e4 d3e4 f8d6 f2f4 h6h5 f4f5 c8d7 a2a4 d8f6 d1d3 g7g6 c3b5 d4b5 c4b5 e8c8 c1e3 h8g8 b5d7 d8d7 a4a5 c8b8 c2c4 d7c7 a1c1 g6g5 b2b4 g5g4 b4b5 c7d7 c1b1 b8c8 d3d2 h5h4 b1b3 g4g3');
INSERT INTO evaluations VALUES(13845,30,'cp 340',NULL);
INSERT INTO evaluations VALUES(13845,31,'cp 359',NULL);
INSERT INTO evaluations VALUES(13845,32,'cp 351',NULL);
INSERT INTO evaluations VALUES(13845,33,'cp 339',NULL);
INSERT INTO evaluations VALUES(13845,34,'cp 351',NULL);
INSERT INTO evaluations VALUES(13845,35,'cp 357',NULL);
INSERT INTO evaluations VALUES(13845,36,'cp 356','d6d5 e4d5 c6d4 e1g1 h7h6 g5e4 f6e4 d3e4 f8d6 f2f4 h6h5 f4f5 a7a6 a2a4 c8d7 a4a5 f7f6 c1d2 d8c7 c3a4 d7a4 a1a4 d6e7 d2e3 d4b5 c4e2 h5h4 h2h3 b5d6 e2h5 e8f8 d1e2 a8c8 f1a1 d6b5 h5g4 c5c4 g1h1 b5d6 g4f3 c7d7 c2c3');
INSERT INTO evaluations VALUES(14830,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(14830,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(14830,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(14830,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(14830,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(14830,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14830,36,'cp 29','c2c4 f8e7 b1c3 e8g8 a2a3 a7a6 f1d3 d5c4 d3c4 b7b6 c4d3 c8b7 e1g1 c6b8 d1e2 c7c5 d4c5 b6c5 g1h1 b8c6 c1d2 d8b6 a1c1 f8d8 f1d1 a8c8 d2e1 d8d7 h1g1 c8d8 e1h4 f6d5 c3d5 e6d5');
INSERT INTO evaluations VALUES(16104,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(16104,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(16104,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(16104,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(16104,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(16104,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(16104,36,'cp 162','b5c6 b7c6 f3e5 f8e7 e1g1 e8g8 f1e1 a7a5 b1c3 a8a6 e5f3 a5a4 d3d4 f8e8 d1d3 a6a8 a2a3 c8a6 d3d1 a6b7 h2h3 c6c5 e4e5 f6h7 d4d5 d7d6 c1f4 h7f8 f4g3 f8g6 c3b5 b7a6 c2c4 a6b5 c4b5 g6e5 f3e5 d6e5 e1e5');
INSERT INTO evaluations VALUES(16094,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(16094,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(16094,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(16094,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(16094,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(16094,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(16094,36,'cp 144','b7c6 f3e5 f8b4 c2c3 b4e7 e1g1 e8g8 f1e1 a7a5 d3d4 c6c5 d4d5 f8e8 e5c4 d7d6 c1f4 c8g4 f2f3 g4d7 f4g3 a5a4 b1d2 f6h5 f3f4 h5g3 h2g3 c7c6 c4e3 e7f8 d1c2 d8c8 a1d1 c8a6 c3c4 a8b8 b2b3 c6d5 c4d5 a4b3 a2b3');
INSERT INTO evaluations VALUES(22055,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(22055,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(22055,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(22055,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(22055,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(22055,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(22055,36,'cp 1','g1f3 g8f6 b1c3 a7a6 f1c4 f8e7 e4e5 d6e5 f3e5 e8g8 e5c6 b7c6 e1g1 e7d6 a4c6 a8b8 c6f3 f6g4 c1f4 g4h2 f4h2 d6h2 g1h2 d8h4 h2g1 h4c4 b2b3 c4c5 c3d5 c5c2 f3f4 c8e6 d5c7 b8b6 f1c1 c2b2 f4g3 b2d2 c1d1 d2c2 d1e1 c2b2 g3e3 b6c6 a1c1 c6c1 e1c1');
INSERT INTO evaluations VALUES(22054,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(22054,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22054,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22054,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22054,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22054,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22054,36,'cp 0','g8f6 b1c3 c8d7 f1e2 a7a6 e1g1 f8e7 a4c4 b7b5 c4d3 e8g8 c1f4 f8e8 c3d5 e7f8 f4g5 f8e7');
INSERT INTO evaluations VALUES(23421,30,'cp -110',NULL);
INSERT INTO evaluations VALUES(23421,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(23421,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(23421,33,'cp -119',NULL);
INSERT INTO evaluations VALUES(23421,34,'cp -120',NULL);
INSERT INTO evaluations VALUES(23421,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(23421,36,'cp -118','g8f6 h5e2 e7e5 g1f3 f8c5 c2c3 e8g8 d2d3 d7d5 c1g5 d5d4 e2d1 h7h6 g5h4 a7a5 f1e2 a5a4 e1g1 d8e7 b1d2 f8d8 h2h3 g7g5 h4g3 f6h5 g3e5 c6e5 f3e5 h5f4 d2f3 a4a3 b2b4 c5b6 c3d4 b6d4 f3d4 d8d4 e5c4');
INSERT INTO evaluations VALUES(17229,30,'cp -110',NULL);
INSERT INTO evaluations VALUES(17229,31,'cp -130',NULL);
INSERT INTO evaluations VALUES(17229,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(17229,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(17229,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(17229,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(17229,36,'cp -115','h5e2 e7e5 g1f3 f8c5 c2c3 e8g8 d2d3 d7d5 c1g5 d5d4 e2d1 h7h6 g5h4 a7a5 f1e2 a5a4 e1g1 d8e7 d1c1 c8e6 c3d4 c6d4 f3d4 c5d4 b1a3 f8c8 a3c2 c7c5 c2d4 c5d4 c1d1 g7g5 h4g3 f6d7 h2h3 f7f6 e2g4 d7c5 a1c1');
INSERT INTO evaluations VALUES(29876,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(29876,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(29876,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(29876,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(29876,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(29876,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(29876,36,'cp 65','f1d3 f8e7 h2h3 g4h5 c2c4 b8a6 b1c3 a6c5 d3c2 c7c6 c1e3 c6d5 c4d5 a8c8 e1g1 e8g8 d1e2 h7h6 a2a4 f6h7 g2g4 h5g6 g1h2 h7g5 f3d2 g5h7 f2f3 h6h5 a4a5 e7g5 e3g5 h7g5 e2e3 g5h7 c3b5 a7a6 b5a7 c8c7 b2b4 c5d7 a1c1 h5g4 f3g4 d8g5 e3g5 h7g5');
INSERT INTO evaluations VALUES(32884,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(32884,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(32884,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(32884,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(32884,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(32884,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(32884,36,'cp 34','f2f3 g4d7 c1f4 b8c6 f4c7 e7e5 f1b5 f7f6 b1d2 a8c8 b5c6 d7c6 c7a5 g8e7 g1e2 h7h5 a2a4 e7g6 h2h4 f8e7 e2g3 c6d7 d2f1 g6f4 f1e3 e7c5 a5d2 c8d8 g3f1 f6f5 g2g3 f5e4 g3f4 e5f4 e3d5 d7g4 f3g4 d8d5 d1c1 e8g8 b2b4 c5b6');
INSERT INTO evaluations VALUES(32882,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(32882,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(32882,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(32882,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(32882,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(32882,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(32882,36,'cp 36','g4d7 c1f4 b8c6 f4c7 e7e5 f1b5 f7f6 b1d2 a8c8 b5c6 d7c6 c7a5 b7b6 a5c3 g8e7 h2h4 h7h5 a2a4 e7g6 g1e2 e8f7 e2c1 f8e7 c1d3 h8d8 g2g3 g6f8 d1e2 f8e6 h1c1 e6c5 d3c5 e7c5 d2b3 c5e7 c1d1 a7a5 d1d8 e7d8 b3d2 c6b7 d2f1 b7a6 e2f2');
INSERT INTO evaluations VALUES(37734,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(37734,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(37734,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(37734,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(37734,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(37734,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(37734,36,'cp 96','b1c3 d7d6 g1f3 g8f6 f1e2 e8g8 e1g1 e7e5 d4d5 a7a5 c1g5 h7h6 g5e3 f6g4 e3c1 g4f6 f3e1 b8a6 c1e3 a6c5 f2f3 f6h5 d1d2 h5f4 a1d1 c8d7 d5c6 d7c6 f1f2 f7f5 e4f5 g6f5 e3f4 e5f4 e1d3 g7d4 d3f4 d4f2 g1f2 d8h4 f2g1');
INSERT INTO evaluations VALUES(39555,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(39555,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39555,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39555,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(39555,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39555,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39555,36,'cp -8','c7c5 d2d3 b8c6 g2g3 a7a6 a2a4 d7d6 f1g2 a8b8 c1e3 g8f6 g1e2 e8g8 e1g1 c8g4 f2f3 g4d7 a1b1 f6e8 f3f4 c6d4 e3f2 b7b5 a4b5 a6b5 e2d4 c5d4 c3b5 d7b5 c4b5 e8c7 g2h3 c7b5 g1h1 e7e6 b2b4');
INSERT INTO evaluations VALUES(48324,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(48324,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(48324,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(48324,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(48324,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(48324,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(48324,36,'cp -42','f1g2 c7c5 e2e3 f8e7 g1e2 e8g8 e1g1 b8c6 d4c5 e7c5 e3e4 f6e4 c3e4 d5e4 g2e4 e6e5 d1d8 f8d8 c1e3 c5e3 f2e3 c8g4 e2c3 d8d7 c3a4 a8c8 c2c4 c6e7 b2b3 f7f6 a4c3 c8d8 c3d5 g4e6 d5c3 b7b6 a1d1 g8f7 d1d7 d8d7 f1d1 d7d1 c3d1');
INSERT INTO evaluations VALUES(52812,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52812,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(52812,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(52812,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(52812,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(52812,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52812,36,'cp 19','d2d4 e7e6 b1c3 f8b4 c1d2 b8d7 a2a3 b4a5 f1d3 g8e7 f3f4 a5c3 d2c3 d5e4 d3e4 e8g8 g1f3 e7d5 c3d2 d8c7 e1g1 f7f5 e4d5 c6d5 d1e2 d7f6 f1c1 c8d7 a3a4 a7a5 b2b3 f6e4 e2e3 b7b5 a4b5 d7b5 f3e5 f8c8 c2c4 d5c4');
INSERT INTO evaluations VALUES(54475,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(54475,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54475,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54475,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54475,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54475,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54475,36,'cp 0','d5d4 c3e4 g8f6 e4f6 d8f6 b2b4 b8c6 c1b2 c6b4 a1c1 f6a6 b2d4 b4a2 c1a1 b7b6 d1d2 c8e6 e2e3 a6a4 a1b1 e6b3 d2d3 a4b4 d3d2');
INSERT INTO evaluations VALUES(54473,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54473,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54473,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54473,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54473,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54473,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54473,36,'cp 0','c3e4 g8f6 e4f6 d8f6 b2b4 b8c6 c1b2 c6b4 a1c1 f6a6 b2d4 b4a2 c1a1 b7b6 d1d2 c8e6 e2e3 a6a4 a1b1 e6b3 d2d3 a4b4 d3d2');
INSERT INTO evaluations VALUES(54488,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(54488,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(54488,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(54488,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(54488,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(54488,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(54488,36,'cp 180','e2e3 b8c6 e3d4 c5d4 f1d3 g8e7 e1g1 e7g6 d3e4 f8e7 e4c6 b7c6 f3d4 g6e5 f1e1 f7f6 d1h5 g7g6 h5h4 d8d7 b1c3 e8g8 e1d1 d7g4 h4h6 g4d7 h6e3 e5c4 e3e4 c4e5 c1f4 c8b7 f4e5 f6e5 e4e5 e7f6 e5e6 d7f7 c3e4 f6e7 a1c1 a8d8 e6f7 g8f7 e4c5 e7c5 c1c5 f7g8');
INSERT INTO evaluations VALUES(54487,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(54487,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(54487,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(54487,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(54487,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(54487,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(54487,36,'cp 179','b8c6 e3d4 c5d4 f1d3 g8e7 e1g1 e7g6 c1g5 d8c7 d3g6 h7g6 f3d4 f8e7 g5e7 c7e5 d4f3 e5e7 f1e1 c8e6 b1c3 e8g8 b2b3 a8d8 d1e2 f8e8 h2h3 e7f6 c3e4 f6e7 e4g3 c6d4 f3d4 d8d4 a1d1 d4d8 e2f3 d8d1 e1d1 e7a3 f3b7 a3a2 b7c7');
INSERT INTO evaluations VALUES(20493,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(20493,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(20493,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(20493,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(20493,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(20493,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(20493,36,'cp 196','f1d3 g8e7 e3d4 c5d4 e1g1 e7g6 c1g5 d8b6 b1d2 c8e6 d2e4 h7h6 g5d2 c6e5 f3e5 g6e5 d2f4 e5d3 d1d3 f7f5 e4g3 f8c5 b2b4 c5b4 g3f5 e8g8 f5h6 g8h8 f4e5 e6d7 a1b1 b6h6 b1b4 d7c6 e5d4 a8d8 d3e3 h6e3 d4e3');
INSERT INTO evaluations VALUES(56469,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(56469,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(56469,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(56469,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(56469,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(56469,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(56469,36,'cp -51','f8d6 e3e4 g8f6 e4d5 f6d5 g2g3 e8g8 f1g2 c7c6 b1c3 c8g4 d1d2 b8d7 e1g1 h7h6 d3d4 d5c3 d2c3 f8e8 h2h3 g4f3 g2f3 d8f6 d4d5 e5e4 d5c6 f6c3 b2c3 b7c6 f3e2 d7b6 a1b1 d6c5 g1g2 e4e3 f2e3 c5e3 e2d3 e3c1 f1c1 b6a4 b1b3 a4c5');
INSERT INTO evaluations VALUES(5414,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(5414,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(5414,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5414,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5414,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(5414,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(5414,36,'cp -14','g7c3 b2c3 g8f6 f1d3 d7d5 e4d5 d8d5 h2h3 c6e5 d3e2 d5d1 e2d1 e5c4 e1g1 e8g8 f1e1 f8e8 f3d4 c8d7 d1f3 a8b8 a2a4 a7a5 d4b3 b7b6 b3d2 c4d6 c3c4 d7f5 g2g4 f5e6 c1a3 e8c8 a1d1 h7h5 g4g5 f6d7 d2e4 d6e4 e1e4');
INSERT INTO evaluations VALUES(10829,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(10829,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(10829,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(10829,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(10829,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(10829,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(10829,36,'cp 25','e1g1 d7d6 a2a3 e8g8 c1e3 f8e8 h2h3 e7f8 f3h2 c8e6 c4e6 f7e6 h2g4 f8e7 b2b4 a7a6 c3b1 f6h7 b1d2 e7g5 d2f3 g5e3 f2e3 e8f8 d1e2 h7g5 g4h2 d8e7 c2c4 g5f3 f1f3 f8f3 h2f3 a8f8 e2c2 a6a5 a1b1 a5b4 a3b4 g8h8');
INSERT INTO evaluations VALUES(10828,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(10828,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(10828,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(10828,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(10828,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(10828,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(10828,36,'cp 22','e8g8 c1e3 d7d6 a2a3 c8e6 c3d5 f8e8 h2h3 a7a5 c4a2 f6d7 a1c1 e7f8 c2c3 c6e7 c3c4 c7c5 f3d2 e7c6 a2b3 f8e7 b3a4 e8f8 d1e2 e7g5 d2f3 g5e7 a4b5 a5a4 c1e1 f7f5 e4f5 e6f5 e2d1 d7f6 d5c3 c6d4 b5a4 d4f3 d1f3 f5d3');
INSERT INTO evaluations VALUES(4550,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(4550,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(4550,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(4550,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(4550,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(4550,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(4550,36,'cp 31','c1e3 d7d6 a2a3 c8e6 c3d5 f8e8 h2h3 a7a5 c4a2 f6d7 a1c1 e7f8 c2c3 c6e7 c3c4 c7c5 f3d2 e7c6 a2b3 g7g6 d2b1 f8g7 b1c3 e6d5 c3d5 d7b6 d5c3 g8h7 c1b1 c6d4 b3c2 e8f8 b2b4 a5b4 a3b4 d8c7 b4c5 d6c5 c3d5 b6d5 c4d5');
INSERT INTO evaluations VALUES(4551,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(4551,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(4551,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(4551,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(4551,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(4551,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(4551,36,'cp 33','d7d6 a2a3 f8e8 h2h3 e7f8 f3h2 c8e6 c4e6 f7e6 h2g4 f8e7 b2b4 a7a6 c3b1 f6h7 b1d2 e7g5 c2c3 g5e3 f2e3 d8e7 d1e2 e8f8 d3d4 f8f1 a1f1 a8f8 f1f8 h7f8 d2f3 e5d4 e3d4 e7f7 c3c4 f7f4 d4d5 c6d8 e2c2 d8f7 c4c5');
INSERT INTO evaluations VALUES(3303,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(3303,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(3303,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(3303,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(3303,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(3303,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(3303,36,'cp 35','a2a3 f8e8 h2h3 e7f8 f3h2 c8e6 c4e6 f7e6 h2g4 f8e7 g4f6 e7f6 d1g4 g8h7 b2b4 a7a6 c3e2 f6g5 e3g5 d8g5 g4g5 h6g5 c2c3 a6a5 f1d1 h7g8 e2g3 g8f7 g3f1 a5b4 a3b4 b7b5 f1e3 f7e7 g2g3 e7d7 g1g2 a8a1 d1a1 e8h8 f2f3 h8f8');
INSERT INTO evaluations VALUES(11016,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(11016,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(11016,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(11016,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(11016,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(11016,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(11016,36,'cp 85','e1g1 d7d5 f3e5 e8g8 b5c6 b7c6 d2d4 c5d6 f1e1 d6e5 d4e5 e7g6 e4d5 c6d5 b1d2 f8e8 d2b3 e8e5 e1e5 g6e5 d1d4 e5c6 d4a4 c6e5 b3c5 d8e7 c1e3 c8d7 a4d4 a8e8 c5d7 e7d7 d4a7 e5c4 e3f4 c4b2 a7c7');
INSERT INTO evaluations VALUES(12779,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(12779,31,'cp -180',NULL);
INSERT INTO evaluations VALUES(12779,32,'cp -200',NULL);
INSERT INTO evaluations VALUES(12779,33,'cp -187',NULL);
INSERT INTO evaluations VALUES(12779,34,'cp -198',NULL);
INSERT INTO evaluations VALUES(12779,35,'cp -206',NULL);
INSERT INTO evaluations VALUES(12779,36,'cp -214','c6e5 f3e2 e5c4 e2c4 d7d5 e4d5 d8d5 c4d5 f6d5 b1a3 c8f5 g1f3 f5d3 f3e5 d3a6 d2d3 f7f6 e5c4 b7b5 c4a5 b5b4 a3c4 b4c3 b2c3 d5c3 c1d2 c3b5 d2e3 a8c8 e1g1 b5c3 f1e1 e8f7 e3d2 c3b5 a1b1 h7h5 b1b2 b5d6 e1b1 d6c4 d3c4 e7e5 f2f4 e5f4 d2f4 g7g5');
INSERT INTO evaluations VALUES(12928,30,'cp -202',NULL);
INSERT INTO evaluations VALUES(12928,31,'cp -209',NULL);
INSERT INTO evaluations VALUES(12928,32,'cp -205',NULL);
INSERT INTO evaluations VALUES(12928,33,'cp -214',NULL);
INSERT INTO evaluations VALUES(12928,34,'cp -204',NULL);
INSERT INTO evaluations VALUES(12928,35,'cp -214',NULL);
INSERT INTO evaluations VALUES(12928,36,'cp -203','f3e2 e5c4 e2c4 d7d5 c4c5 d5e4 f2f3 d8d3 b1a3 c8d7 b2b3 a8c8 c5d4 d3d4 c3d4 d7c6 a3c4 e7e6 e1f2 b7b5 c4e3 f8e7 c1b2 e8g8 a1c1 f6d5 e3d5 c6d5 g1e2 e4f3 g2f3 e7h4 f2e3 h4g5 e2f4 f8d8 c1c8 d8c8 h1c1 c8c1 b2c1 b5b4');
INSERT INTO evaluations VALUES(12929,30,'cp -200',NULL);
INSERT INTO evaluations VALUES(12929,31,'cp -195',NULL);
INSERT INTO evaluations VALUES(12929,32,'cp -199',NULL);
INSERT INTO evaluations VALUES(12929,33,'cp -203',NULL);
INSERT INTO evaluations VALUES(12929,34,'cp -194',NULL);
INSERT INTO evaluations VALUES(12929,35,'cp -211',NULL);
INSERT INTO evaluations VALUES(12929,36,'cp -196','e5c4 e2c4 d7d5 e4d5 d8d5 c4d5 f6d5 g1e2 c8f5 d2d4 c5d4 c3c4 d5b4 e2d4 f5g6 e1g1 e7e5 a2a3 e5d4 a3b4 f8b4 c1d2 b4e7 b1a3 a7a6 b2b4 e8c8 c4c5 h8e8 d2f4 g6d3 f1c1 g7g5 f4e5 g5g4 c1e1 e7g5 a1d1 f7f6 d1d3 e8e5 e1e5 f6e5 a3c4');
INSERT INTO evaluations VALUES(14205,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(14205,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14205,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(14205,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(14205,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(14205,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(14205,36,'cp 16','g1f3 d5c4 e2e3 e7e5 f4e5 f8b4 f3d2 b7b5 e5f6 d8f6 a2a3 b4d2 b1d2 e8g8 f1e2 c8f5 e1g1 a8c8 a1c1 f8d8 b2b3 c6d4 e3d4 f6d4 d2c4 d4f4 d1e1 b5c4 e2c4 f5d3 g2g3 f4f5 c4d3 d8d3 c1c7 d3b3 c7a7 h7h6 a3a4 b3f3 e1b1 c8c2 b1b5 f5b5 a4b5 g8h7 a7b7');
INSERT INTO evaluations VALUES(16471,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(16471,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(16471,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(16471,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(16471,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(16471,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(16471,36,'cp 116','d4d5 c6b8 g1f3 e5e4 f3d4 f8b4 d1b3 a7a5 a2a3 b4c5 c1e3 d7d6 h2h3 e8g8 g2g4 f8e8 f1g2 b8d7 g4g5 f6h5 c3e4 c5d4 e3d4 d7e5 e2e3 c8f5 e1c1 e5c4 b3c4 f5e4 g2e4 e8e4 c4c2 e4h4 f2f4 a5a4 c2c3');
INSERT INTO evaluations VALUES(16335,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(16335,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(16335,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(16335,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(16335,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(16335,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(16335,36,'cp 85','c6b8 g1f3 e5e4 f3d4 f8b4 d1b3 a7a5 a2a3 b4c3 b3c3 d7d6 b2b3 b8d7 e2e3 d7c5 h2h3 e8g8 b3b4 a5b4 a3b4 a8a1 c3a1 c5d7 f1e2 h7h5 e1g1 f6h7 g1h2 h7g5 c1b2 d7e5 a1e1 f8e8 e1d1 c7c6 d5c6 b7c6 e2h5 e5d3');
INSERT INTO evaluations VALUES(39535,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(39535,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(39535,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(39535,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39535,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39535,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39535,36,'cp 0','g8f6 e4e5 f6g8 g1f3 d7d6 f1b5 c7c6 d5c6 b7c6 b5a4 d8a5 b1c3 d6e5 e1g1 g8h6 f1e1 h6f5 d1e2 c8a6 e2e4 f5d6 e4h4 d6f5');
INSERT INTO evaluations VALUES(40605,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40605,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(40605,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40605,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40605,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40605,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40605,36,'cp 0','c4d5 e6d5 e2e3 e8g8 f1e2 b8c6 e1g1 a7a6 b2b3 c5a7 c1b2 c8f5 a1c1 f8e8 e2d3 f5g4 d3e2');
INSERT INTO evaluations VALUES(40596,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40596,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40596,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40596,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40596,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40596,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40596,36,'cp 0','e6d5 e2e3 b8c6 f1e2 e8g8 e1g1 a7a6 b2b3 c5a7 c1b2 f8e8 a1c1 c8f5 h2h3 d5d4 f3d4 c6d4 e3d4 a7d4 e2f3 f5d3 f1e1 d4f2 g1f2 d8d4 f2g3 d4d6 g3f2');
INSERT INTO evaluations VALUES(40649,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40649,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40649,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40649,36,'cp 0','e2e3 b8c6 f1e2 a7a6 e1g1 e8g8 b2b3 c5a7 c1b2 c8f5 e2d3 f5g4 d3e2');
INSERT INTO evaluations VALUES(51962,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51962,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(51962,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51962,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(51962,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51962,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51962,36,'cp -2','g8f6 b1d2 d7d5 g1f3 b8d7 f1d3 b7b6 d1e2 c8b7 e1g1 f8e7 e3e4 d5e4 d2e4 e8g8 e4f6 e7f6 f1d1 c5d4 f3d4 d7e5 d3e4 b7e4 e2e4 e5g6 c1e3 d8c7 a2a4 a8d8 e4c6 c7b8 c6f3 d8c8 d4b5 f8d8 d1d8 c8d8 a1d1 d8d1 f3d1 h7h6');
INSERT INTO evaluations VALUES(44230,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44230,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44230,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44230,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44230,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44230,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(44230,36,'cp 0','b1d2 d7d5 g1f3 b8d7 f1d3 c5c4 d3c2 b7b5 e3e4 d5e4 d2e4 c8b7 d1e2 f8e7 e4f6 d7f6 f3e5 e8g8 e1g1 d8c7 a2a4 a7a6 c1g5 f6d5 g5e7 c7e7 c2e4 f8c8 e4f3 f7f6 e5g4 d5f4 e2e3 f4d5 e3e2');
INSERT INTO evaluations VALUES(53569,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(53569,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(53569,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(53569,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(53569,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(53569,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(53569,36,'cp -25','d5c4 b1d2 b7b5 b2b3 c4b3 d1b3 a7a5 e2e4 g8f6 f1e2 f8e7 e1g1 e8g8 a1c1 c8b7 f1d1 a5a4 b3e3 b8d7 f3e5 a8c8 h2h4 d8b6 e3g3 g8h8 e5d7 f6d7 f4d6 e7d6 g3d6 d7f6 e4e5 f6d5 d2e4 h7h6 d6c5 b5b4 h4h5 b4b3 c5b6 d5b6 e4c5');
INSERT INTO evaluations VALUES(54845,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(54845,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(54845,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54845,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(54845,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54845,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(54845,36,'cp 37','g8f6 c4d5 f6d5 e2e4 d5c3 b2c3 f8g7 c1e3 c7c5 g1f3 d8a5 d1d2 e8g8 a1c1 c5d4 c3d4 a5d2 f3d2 c8d7 c1c7 f8c8 c7c8 d7c8 f1b5 b8c6 d2b3 c8d7 e1d2 a8d8 d4d5 f7f5 b3c5 d7c8 f2f3 f5e4 f3e4 c6d4 b5c4 b7b6 c5d3');
INSERT INTO evaluations VALUES(57114,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(57114,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(57114,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(57114,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(57114,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(57114,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(57114,36,'cp -97','d7d5 d2d3 g8f6 g1f3 f8d6 g2g3 e8g8 f1g2 a7a5 b4b5 c8g4 e1g1 f8e8 a2a4 b8d7 b1d2 d8e7 h2h3 g4e6 d1c2 h7h6 f1e1 c7c6 b5c6 b7c6 e2e4 d5e4 d2e4 f6e4 e1e4 a8b8 e4e3 f7f5 a1b1 b8b1 c2b1 e7f7 c3c4 f5f4 g3f4 e5f4 e3e1 e6f5 b1b7');
INSERT INTO evaluations VALUES(56331,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(56331,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(56331,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(56331,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(56331,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(56331,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(56331,36,'cp -86','d2d3 b8d7 g1f3 a7a5 b4b5 g8f6 g2g3 f8c5 f1g2 e8g8 e1g1 f8e8 d1c2 c7c6 a2a4 e5e4 d3e4 f6e4 b1d2 d7f6 d2e4 f6e4 c1a3 c5a3 a1a3 e4d6 b5c6 b7c6 f3d4 c8d7 a3a1 h7h6 f1d1 a8c8 e2e3 d7g4 d4f3 g4f5');
INSERT INTO evaluations VALUES(57638,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(57638,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(57638,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57638,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(57638,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(57638,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(57638,36,'cp 3','g8f6 c2c4 f8g7 b1c3 e7e6 g1f3 e6d5 c4d5 e8g8 e2e4 d7d6 f1d3 c7c6 e1g1 c6d5 e4d5 c8g4 d3e2 b8d7 h2h3 g4f3 e2f3 d7e5 c1g5 h7h6 g5e3 e5c4 e3d4 f8e8 b2b3 c4e5 a1c1 a7a6 f3e2 a8c8 a2a4 f6d7');
INSERT INTO evaluations VALUES(5270,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(5270,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5270,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(5270,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(5270,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5270,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(5270,36,'cp 62','d2d4 d8a5 b5c6 b7c6 e1g1 c5d4 c3d4 g8f6 c1d2 a5b6 b1c3 d7d6 e4e5 d6e5 d4e5 f6d5 c3d5 c6d5 d2c3 e8g8 d1d5 c8b7 d5b3 f8c8 f1e1 e7e6 f3d2 b7d5 b3b6 a7b6 a2a4 g7f8 h2h4 f8g7 f2f3 d5c6 d2c4 g7f8 c4d6 c8d8 e1c1 a8a4 a1a4 c6a4');
INSERT INTO evaluations VALUES(5265,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(5265,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5265,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(5265,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(5265,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(5265,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(5265,36,'cp 39','d8b6 a2a4 c5d4 e1g1 a7a6 b5c6 b7c6 a4a5 b6b7 c3d4 g8f6 b1c3 e8g8 e4e5 f6d5 c3d5 c6d5 c1f4 f8d8 f1e1 d7d6 e5d6 e7d6 f4g5 f7f6 g5d2 c8f5 d2c3 a8c8 h2h3 d8e8 d1d2 b7b5 f3h2 h7h5 d2f4 g7f8 h2f1 g8h7 e1e8 b5e8 a1e1');
INSERT INTO evaluations VALUES(8743,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(8743,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(8743,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(8743,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(8743,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(8743,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(8743,36,'cp 101','d4d5 c6b8 f1e2 e7e6 d1c2 e8g8 e1g1 b8d7 d5e6 f7e6 c1e3 b7b6 f3g5 d8e7 f2f4 c8b7 e2f3 h7h6 e4e5 d6e5 f3b7 a8b8 c2g6 b8b7 g5h3 e7f7 g6g3 f7h5 a1d1 h5g4 f1f3 e5f4 e3f4 g4g3 f3g3 g8h7 d1e1 f8e8 h3f2 d7c5 b2b4');
INSERT INTO evaluations VALUES(8736,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(8736,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(8736,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(8736,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(8736,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(8736,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(8736,36,'cp 96','c6b8 h2h3 e8g8 f1d3 a7a5 e1g1 b8a6 f1e1 f6d7 c1g5 a6c5 d3c2 g7c3 b2c3 f7f6 g5e3 d7e5 f3e5 f6e5 e3c5 d6c5 d1d2 d8d6 c2a4 g8h8 a4b5 c8d7 b5d7 d6d7 d2h6 b7b6 e1e3 a5a4 h3h4 f8f7 a1b1');
INSERT INTO evaluations VALUES(10883,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(10883,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(10883,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(10883,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(10883,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(10883,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(10883,36,'cp -59','e8g8 d2d3 c6a5 c4b3 a5b3 a2b3 c7c6 g5f3 d8c7 e1g1 h7h6 f1e1 f8e8 c1d2 e7f8 c3e2 d7d5 e4d5 f6d5 e2g3 c8d7 c2c4 d5f4 d2c3 f7f6 d3d4 e5d4 e1e8 a8e8 f3d4 a7a6 d1d2 c6c5 d4f5 d7f5 g3f5 f4e2 g1h1 c7f4 d2f4');
INSERT INTO evaluations VALUES(4569,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(4569,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(4569,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(4569,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(4569,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(4569,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(4569,36,'cp -38','d2d3 d7d6 e1g1 h7h6 g5f3 c6a5 c4b3 c7c6 c3e2 d8c7 e2g3 a5b3 a2b3 d6d5 d1e2 d5e4 d3e4 f8e8 f3h4 g7g6 h2h3 e7f8 c1d2 a7a5 f1d1 b7b5 h4f3 c8e6 f3h2 g8h7 h2f3 f6d7 h3h4 a5a4 b3a4 e6c4 e2e3 a8a4 a1a4 b5a4');
INSERT INTO evaluations VALUES(4570,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(4570,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(4570,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(4570,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(4570,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(4570,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(4570,36,'cp -51','d7d6 e1g1 c6a5 c4b3 h7h6 g5f3 f8e8 h2h3 c7c6 c3e2 e7f8 c1d2 b7b5 e2g3 d8b6 d1e1 a5b3 a2b3 c6c5 d2a5 b6c6 e1e2 c8b7 c2c4 g7g6 f1e1 b5b4 f3h2 h6h5 g3f1 f8h6 f1e3 c6d7 h2f3 f6h7 e2d1 e8f8');
INSERT INTO evaluations VALUES(3369,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(3369,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(3369,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(3369,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(3369,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(3369,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(3369,36,'cp -36','e1g1 c6a5 c4b3 a5b3 a2b3 h7h6 g5f3 f8e8 d3d4 e5d4 d1d4 b7b6 c1f4 c8b7 f1e1 c7c5 d4d3 d6d5 e4d5 f6d5 c3d5 d8d5 d3d5 b7d5 f4c7 d5e6 h2h3 b6b5 e1e5 e7f8 g2g4 a7a6 a1a5 e6d7 e5e8 d7e8 c2c4');
INSERT INTO evaluations VALUES(3368,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(3368,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(3368,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(3368,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(3368,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(3368,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(3368,36,'cp -41','h7h6 g5f3 c6a5 c4b3 f8e8 f1e1 c7c6 c3e2 e7f8 e2g3 a5b3 a2b3 a7a5 d3d4 e5d4 d1d4 b7b5 h2h3 c6c5 d4d3 d8b6 c2c4 c8e6 c1d2 f6d7 d2c3 b5b4 c3d2 d7e5 f3e5 d6e5 d2e3 b6c7 d3e2 g8h7 e1d1 g7g6 e2c2 a8a6 g3f1 f8g7');
INSERT INTO evaluations VALUES(23077,30,'cp -360',NULL);
INSERT INTO evaluations VALUES(23077,31,'cp -357',NULL);
INSERT INTO evaluations VALUES(23077,32,'cp -381',NULL);
INSERT INTO evaluations VALUES(23077,33,'cp -398',NULL);
INSERT INTO evaluations VALUES(23077,34,'cp -386',NULL);
INSERT INTO evaluations VALUES(23077,35,'cp -390',NULL);
INSERT INTO evaluations VALUES(23077,36,'cp -372','d8h4 e1d2 g8f6 c2c3 h4f4 d2e1 f4e4 e1d2 g7g6 d1e2 e5g4 d2e1 f8d6 b1d2 e4e2 g1e2 e8g8 d2f3 f8e8 h2h3 f6e4 h1g1 g4h2 c1f4 e4g3 f4d6 h2f1 d6g3 f1g3 e1f2 g3e4 f2e1 b7b6 e1d1 c8b7 e2f4 f7f6 g1e1 g6g5 f4d3 e4d6 b2b3 e8e1 f3e1 g8f7 c3c4 a8e8 d4d5 c7c6 d5c6 b7c6 e1f3 h7h6 d3f2 c6b7 d1d2 f7g6');
INSERT INTO evaluations VALUES(28243,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(28243,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(28243,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28243,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(28243,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28243,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28243,36,'cp 8','h7h6 g5f6 d8f6 e1g1 b4c3 b2c3 c7c5 f1e1 e8g8 e1e3 c5d4 c3d4 d7d6 h2h3 b8c6 c2c3 e6e5 d4d5 c6a5 f3d2 f8c8 d3b5 c8c7 a2a4 b7c8 d2b3 a5b3 d1b3 c8d7 b5d7 c7d7 a4a5 a8b8 c3c4 d7c7 e3c3 b8c8 a5b6 a7b6 b3b6 c7c4 c3c4');
INSERT INTO evaluations VALUES(28217,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(28217,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28217,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28217,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28217,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28217,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28217,36,'cp 0','g5f6 d8f6 e1g1 b4c3 b2c3 c7c5 d1e2 e8g8 f3e5 c5d4 e5g4 f6h4 c3d4 b8c6 g2g3 h4g5 f2f4 g5a5 g4h6 g7h6 e2g4 g8h8 g4h3 h8g7 h3g4');
INSERT INTO evaluations VALUES(28216,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28216,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28216,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28216,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28216,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28216,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28216,36,'cp 0','d8f6 e1g1 b4c3 b2c3 c7c5 f1e1 e8g8 e1e3 c5d4 c3d4 d7d6 c2c3 f8c8 d1d2 b8c6 h2h3 c8c7 d2b2 a8c8 a1e1 e6e5 d3f1 c6a5 a2a4 f6e7 b2a3 e7f6');
INSERT INTO evaluations VALUES(32546,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32546,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32546,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32546,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32546,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32546,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32546,36,'cp 22','e1g1 b4c3 b2c3 c7c5 f1e1 e8g8 e1e3 c5d4 c3d4 d7d6 c2c3 e6e5 h2h3 b8c6 d3f1 a8c8 d1a4 c6a5 f3d2 b7c6 a4b4 c6d7 e3f3 f6e7 f3g3 g8h8 d2f3 a5c6 b4a3 f8d8 a1e1 c8c7 a3b3 c6a5 b3b2');
INSERT INTO evaluations VALUES(32545,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(32545,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(32545,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(32545,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32545,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32545,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32545,36,'cp 11','b4c3 b2c3 c7c5 f1e1 e8g8 e1e3 d7d6 e4e5 d6e5 d4e5 b7f3 d1e1 f6g5 e3f3 b8d7 f3e3 c5c4 d3c4 a8c8 e3e4 g5e7 a2a4 f8d8 c4b5 d7f8 a4a5 c8c5 c3c4 e7c7 a5b6 a7b6 e1e3 f8g6 a1e1 g6e7 h2h4 e7f5 e3f4 f5e7 e4d4 d8d4 f4d4');
INSERT INTO evaluations VALUES(38576,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(38576,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(38576,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(38576,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(38576,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(38576,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(38576,36,'cp 148','h2h4 c7c5 c2c3 b8c6 h4h5 c8d7 b1a3 a7a6 a3c2 g6h5 f1e2 g8e7 h1h5 e7g6 d4c5 f7f6 e5f6 d8f6 g2g3 h7h6 c1e3 e8c8 f3d4 g6e5 d4c6 e5c6 d1d2 d7e8 h5h4 e6e5 e1c1 e8g6 c2b4 d5d4 c3d4 e5d4 e3f4 h6h5 b4c6 f6c6 e2d3 c6c5 c1b1 g6d3 d2d3');
INSERT INTO evaluations VALUES(53441,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(53441,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(53441,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(53441,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(53441,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(53441,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(53441,36,'cp 33','b8c6 b1c3 c8f5 g1e2 e7e6 e2f4 d8b6 f1b5 f8b4 b5c6 b6c6 c1d2 g8e7 g2g4 f5g6 h2h4 h7h6 c3e2 c6b5 d2b4 b5b4 d1d2 b4d2 e1d2 h6h5 f2f3 a8c8 a1c1 c8c1 e2c1 h5g4 f3g4 g6e4 h1h3 g7g5 f4h5 g5h4 h3h4 e8d8 c1e2 d8c7 e2f4 a7a6');
INSERT INTO evaluations VALUES(19887,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(19887,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(19887,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(19887,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(19887,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(19887,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(19887,36,'cp 43','b1c3 c8f5 g1e2 e7e6 e2g3 f5g6 h2h4 h7h5 c1e3 d8b6 d1d2 g8h6 f1e2 a8c8 a1c1 f8e7 c3a4 b6b4 a4c5 b4d2 e1d2 c8c7 e3h6 g7h6 d2e3 c6e5 c5e6 c7c1 e6g7 e8f8 h1c1 e5c4 e2c4 f8g7 c4d5 h8e8 e3f3 e8d8 d5b7 d8d4 b7e4 g6e4 g3e4 e7h4 c1c5');
INSERT INTO evaluations VALUES(3853,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(3853,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(3853,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(3853,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(3853,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(3853,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(3853,36,'cp 58','e1g1 c5b6 b2b3 c8f5 d2d4 f8e8 d4d5 c6e7 e3e4 f5d7 h2h3 e7g6 a1b1 a7a6 d1d3 h7h6 c3d1 h6h5 c1g5 g6f8 e2c3 f8h7 g5d2 b6d4 c3e2 d4c5 b3b4 c5b6 e2c3 a6a5 b4b5 b6c5 c3a4 c5d4 d2e3 d4e3 d1e3');
INSERT INTO evaluations VALUES(3852,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(3852,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(3852,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(3852,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(3852,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(3852,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(3852,36,'cp 46','c5b6 b2b3 f8e8 h2h3 h7h6 g1h2 c6e7 d2d4 c7c6 d4e5 d6e5 d1c2 b6c7 f1d1 c8f5 c3e4 f5e4 g2e4 d8c8 e4g2 h6h5 e2c3 c8e6 c1b2 a8d8 d1d8 e8d8 a1d1 d8d1 c2d1 e6f5 h2g1 e7g6 d1b1 f5e6 b1c2 h5h4');
INSERT INTO evaluations VALUES(20891,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(20891,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(20891,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(20891,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(20891,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(20891,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(20891,36,'cp 45','f3e5 c6e5 d4e5 c8f5 b1d2 d8d7 d2b3 e7e6 h2h4 h7h5 b3d4 f5e4 f2f3 e4g6 f1d3 g6d3 d1d3 g8e7 c1g5 c7c5 g5e7 d7e7 d3g6 e8d7 d4e2 e7e8 g6g3 a8d8 c2c4 d7c8 e1c1 d5d4 e2f4 g7g6 c1b1 f8h6 f4d3 c8b8 f3f4 d8c8 a2a3 a7a6 g3f3');
INSERT INTO evaluations VALUES(20886,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(20886,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(20886,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(20886,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(20886,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(20886,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(20886,36,'cp 56','c6e5 d4e5 c8f5 b1d2 e7e6 h2h4 h7h5 f1e2 c7c6 c2c3 g7g6 d2f3 d8c7 e1g1 g8h6 c1g5 h6g4 e2d3 f8e7 g5e7 c7e7 d1d2 e8c8 a1e1 g4h6 f3g5 h8f8 a2a4 f5d3 d2d3 h6f5 g2g3 c6c5 g1g2 c8b8 b2b4 c5b4 c3b4');
INSERT INTO evaluations VALUES(20901,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(20901,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(20901,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(20901,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(20901,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(20901,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(20901,36,'cp 44','d4e5 c8f5 b1c3 c7c6 c3e2 d8c7 c1f4 e7e6 e2d4 c7f7 d4f5 f7f5 d1d2 f8c5 f1d3 f5h5 d2e2 h5e2 e1e2 a7a5 f4d2 g8e7 f2f4 h7h5 a2a4 h8f8 h1f1 e7f5 c2c3 g7g6 g2g3 e8d7 b2b3 b7b6 a1d1 d7e8 b3b4 a5b4 c3b4 c5e7 a4a5 b6a5 b4a5 c6c5 d1b1 c5c4');
INSERT INTO evaluations VALUES(20900,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(20900,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(20900,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(20900,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(20900,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(20900,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(20900,36,'cp 47','c8f5 b1d2 d8d7 d2b3 e7e6 c1e3 g8e7 c2c3 d7a4 b3d4 a4d1 a1d1 e8d7 c3c4 f5g6 c4d5 e7d5 f1c4 d5e3 f2e3 f8b4 e1f2 d7e7 f2g3 g6f7 h1f1 h8f8 f1f7 f8f7 d4e6 c7c6 e6c7 a8d8 d1d8 e7d8 c4f7 d8c7 g3f4 b4d2 f7e6 b7b6 b2b3 b6b5 e6g4 h7h6 f4f5 d2e3');
INSERT INTO evaluations VALUES(21231,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(21231,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(21231,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(21231,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(21231,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(21231,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(21231,36,'cp -65','g1f3 d4c6 c1d2 g8f6 f1d3 g7g6 d1e2 f8g7 e1c1 e8g8 d2c3 d8d5 e4f6 g7f6 d3c4 d5h5 e2e3 f6c3 e3c3 h5a5 c3a5 c6a5 c4e2 c8e6 f3g5 e6f5 g2g4 f5c8 h2h3 e7e5 h1e1 g8g7 e2f3 h7h6 g5e4 f7f5 g4f5 c8f5 e4g3 a8e8 g3f5 f8f5 d1d7 f5f7 e1d1 c7c6 d7f7 g7f7');
INSERT INTO evaluations VALUES(31803,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(31803,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(31803,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31803,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31803,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(31803,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(31803,36,'cp -33','c7c6 a4c2 b7b5 a2a4 d8b6 g1f3 g8f6 f1e2 g7g6 f3g5 e6d5 e3e4 h7h6 e4d5 h6g5 d5c6 b8c6 a4b5 c6d4 c2c4 h8h4 c1e3 e7e6 g2g3 h4e4 e1g1 f6d5 e3d4 e4d4 c4b3 a8c8 b1c3 d4b4 b3a3 d5f6 e2f3 f8c5 c3a4 b6b5 a4c5 c8c5 f1c1 e8f8 a3a7 c5c1 a1c1 b4b2');
INSERT INTO evaluations VALUES(31025,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(31025,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(31025,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(31025,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(31025,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(31025,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(31025,36,'cp -17','a4c2 b7b5 a2a4 g8f6 g1f3 d8b6 f1e2 g7g6 f3g5 e6d5 e3e4 h7h6 e4d5 h6g5 d5c6 b8c6 a4b5 c6d4 c2c4 f8g7 c1e3 f6e4 a1a6 b6c5 b1a3 c5c4 e2c4 e4c5 a6a5 d4b3 c4b3 c5b3 a5a7 a8a7 e3a7 h8h4 e1g1 g7b2 f1b1 b2a3 b1b3 a3d6 b3h3 h4h3 g2h3');
INSERT INTO evaluations VALUES(36465,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(36465,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(36465,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(36465,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(36465,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(36465,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(36465,36,'cp 47','g1f3 c7c5 d4d5 h7h6 g5h4 g6g5 h4g3 f6h5 d1c2 f7f5 f1e2 a7a6 e1g1 d8e8 a2a3 b8d7 e2d3 e8f7 a1e1 b7b5 c4b5 d7b6 a3a4 a6b5 a4b5 g8h8 e1a1 a8b8 a1a7 b8b7 f1a1 b7a7 a1a7 f5f4 e3f4 h5f4');
INSERT INTO evaluations VALUES(41820,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(41820,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(41820,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(41820,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(41820,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41820,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(41820,36,'cp 30','e1g1 e8g8 c2c4 b7b6 b1c3 c8b7 c4d5 e6d5 d1b3 f8e8 f1d1 b8d7 c1f4 d6f4 g3f4 c7c5 e2e3 a8c8 d4c5 d7c5 b3a3 a7a5 a1c1 c5e4 a3b3 e4c3 c1c3 d8d6 d1c1 b7a6 f3e5 c8c5 c1d1 e8d8 c3c1 c5c1 d1c1');
INSERT INTO evaluations VALUES(41819,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(41819,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(41819,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(41819,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(41819,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(41819,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(41819,36,'cp 32','b7b6 c2c4 e8g8 c4d5 e6d5 b1c3 c8b7 d1b3 f8e8 f1d1 b8d7 c1f4 d6f4 g3f4 c7c5 e2e3 h7h6 d4c5 d7c5 b3a3 f6e4 a1c1 e4c3 a3c3 a8c8 b2b4 c5e6 c3d2 d8e7 f3d4 e6d4 d2d4 e8d8 a2a4 e7h4 c1c8 b7c8');
INSERT INTO evaluations VALUES(49580,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(49580,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(49580,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(49580,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(49580,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(49580,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(49580,36,'cp -32','e5e4 f3d2 c7c6 b1c3 d7d5 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 e8g8 e1g1 b8d7 d1b3 b4c3 b2c3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 a8c8 c4b3 f5d3 f1e1 f8e8 h2h3 d3c2 c1d2 c2b3 a2b3 a7a6 e1c1 h7h6 c1c8 e8c8 g1f2');
INSERT INTO evaluations VALUES(49749,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(49749,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(49749,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(49749,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(49749,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(49749,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(49749,36,'cp -30','f3d2 c7c6 b1c3 d7d5 f2f3 e4f3 d2f3 f8b4 e2e3 c8f5 f1e2 e8g8 e1g1 b8d7 d1b3 b4c3 b2c3 d8c7 c4d5 c6d5 c3c4 d5c4 b3c4 c7c4 e2c4 f8c8 f3d2 f5c2 c4e2 f6e4 d2e4 c2e4 e2f3 d7f6 f1d1 c8c2 d4d5 a8d8');
INSERT INTO evaluations VALUES(50858,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(50858,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(50858,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(50858,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(50858,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(50858,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(50858,36,'cp -27','b7b5 g1f3 e7e6 a2a4 b5b4 c3e2 b8c6 e2g3 c6a5 c1g5 f8e7 g5e7 g8e7 a1c1 c7c5 f1c4 c5d4 d1d4 d8d4 f3d4 c8d7 c4d3 d7a4 c1a1 a5b3 a1a4 b3d4 a4b4 e7c6 b4b6 e8d7 b6a6 a8a6 d3a6 h8b8 e1g1 b8b2 f2f4 d7e7 a6d3 c6b4 f1d1 b4d3');
INSERT INTO evaluations VALUES(53427,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53427,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53427,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53427,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53427,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53427,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53427,36,'cp 0','c8f5 f3h4 f5c8 h4f3');
INSERT INTO evaluations VALUES(31332,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31332,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31332,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31332,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31332,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31332,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31332,36,'cp 0','f3h4 f5c8 h4f3 c8f5');
INSERT INTO evaluations VALUES(56533,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(56533,31,'cp 287',NULL);
INSERT INTO evaluations VALUES(56533,32,'cp 311',NULL);
INSERT INTO evaluations VALUES(56533,33,'cp 306',NULL);
INSERT INTO evaluations VALUES(56533,34,'cp 319',NULL);
INSERT INTO evaluations VALUES(56533,35,'cp 298',NULL);
INSERT INTO evaluations VALUES(56533,36,'cp 323','f2f4 b8c6 g1f3 g8h6 f1b5 c8d7 c2c3 a7a6 b5c4 d7g4 b1d2 d4c3 b2c3 f8c5 d2b3 d8d1 e1d1 e8c8 d1c2 c5a7 h2h3 g4e6 c4e6 f7e6 f3g5 d8e8 a1b1 c6d8 a2a4 h6f7 g5f7 d8f7 b3a5 h8g8 h3h4 f7d8 h4h5 a7f2 h1h3 e8f8 a5c4 g7g5 h5g6 g8g6');
INSERT INTO evaluations VALUES(11838,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11838,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(11838,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(11838,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11838,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(11838,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11838,36,'cp 39','e4d5 d8d5 e1g1 c8g4 b1c3 g4f3 c3d5 f3d1 d5f6 g7f6 f1d1 e8d7 d1e1 f8d6 c1d2 f6f5 d2c3 f7f6 e1e2 a8b8 b2b3 h8g8 e2e3 h7h5 e3h3 g8h8 a1e1 h5h4 f2f4 d6c5 g1h1 c5d4 c3d4 e5d4 g2g3 h4g3 h3g3 b8g8 h1g2 g8g3 h2g3 c6c5 e1h1 h8h1 g2h1 d7c6 a2a4 c6d6 h1g2 c7c6 g2f2 a7a5 f2e1');
INSERT INTO evaluations VALUES(11826,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11826,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(11826,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11826,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(11826,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(11826,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(11826,36,'cp 41','d8d5 e1g1 c8g4 b1c3 g4f3 c3d5 f3d1 d5f6 g7f6 f1d1 e8d7 d1e1 f8d6 c1d2 f6f5 d2c3 f7f6 e1e2 a8g8 f2f3 c6c5 f3f4 g8g6 f4e5 f6e5 a1f1 h8f8 f1f2 d7e6 f2f4 h7h5 f4h4 f8h8 h4a4 e6d5 e2f2 h8g8 b2b3 d5c6 f2e2 c6d5 a4a7 h5h4 a7a4');
INSERT INTO evaluations VALUES(14269,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(14269,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(14269,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(14269,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(14269,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14269,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(14269,36,'cp 0','c1g5 e7e6 g1f3 f8b4 d1d3 h7h6 g5f6 d8f6 e1c1 b4a5 c1b1 c6b4 d3b5 b4c6');
INSERT INTO evaluations VALUES(20280,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(20280,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(20280,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(20280,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(20280,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(20280,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(20280,36,'cp 42','e1g1 c5d4 f3d4 g8f6 c2c4 f8c5 d4b3 c5b6 b1c3 e8g8 c4d5 f6d5 c3d5 e6d5 d1d5 d8d5 g2d5 c8g4 a2a4 f8d8 d5c6 b7c6 a4a5 b6c7 f1e1 a8b8 b3c5 b8b5 c1e3 b5b2 f2f3 g4f5 g1f2 d8d5 e1c1 h7h5 c1c4 b2b5 a5a6 c7b6 c5b7 b6e3 f2e3 b5b2 c4c6 f5d7 c6c7');
INSERT INTO evaluations VALUES(20279,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(20279,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(20279,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(20279,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(20279,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(20279,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(20279,36,'cp 40','c5d4 f3d4 f8c5 d4b3 c5b6 c2c4 g8f6 c4d5 f6d5 a2a4 e8g8 c1d2 a7a6 b1a3 e6e5 a1c1 c8e6 a3c4 a8c8 c4b6 d5b6 b3c5 d8e7 c5e6 e7e6 g2c6 c8c6 c1c6 e6c6 a4a5 b6d7 d2e3 c6b5 d1d6 d7f6 d6b6 b5b6 e3b6');
INSERT INTO evaluations VALUES(25323,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(25323,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(25323,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(25323,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(25323,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(25323,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(25323,36,'cp 94','c2c3 c6a5 c4b5 c7c6 b5a4 f8e7 a4d1 b7b6 e1g1 e8g8 b1d2 a5b7 d3d4 d8c7 f3e2 b7d8 d2f3 f8e8 f1e1 a7a5 a2a4 d8e6 d1c2 a8d8 c1d2 g7g6 g2g3 e7f8 g1g2 e5d4 c3d4 d6d5 e4e5 f6e4 c2e4');
INSERT INTO evaluations VALUES(25322,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(25322,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(25322,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(25322,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(25322,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(25322,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(25322,36,'cp 103','c6a5 c4b5 c7c6 b5a4 f8e7 a4c2 b7b5 b1d2 e8g8 d2f1 d6d5 f1e3 g7g6 e1g1 d5d4 c3d4 e5d4 e3g4 f6g4 f3g4 c6c5 e4e5 c5c4 c1h6 f8e8 d3c4 d4d3 a1d1 b5c4 c2a4 d8d5 a4e8 a8e8 h6d2 a5c6 f1e1 e7b4 d2b4 c6b4 g4e4 e8b8 e4d5 b4d5 e1e4 c4c3 b2c3 d5c3');
INSERT INTO evaluations VALUES(33378,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33378,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33378,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33378,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33378,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33378,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33378,36,'cp 0','c8g4 f1e2 b8c6 h2h3 g4e6 c1e3 h7h6 e1c1 e8c8 a2a3 a7a6 d3d4 d6d5 e2d3 e7d6 d1e1 g7g5 g2g4 d8e8 h1f1 e6d7 f3d2 c6d8 d2f3');
INSERT INTO evaluations VALUES(32550,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(32550,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32550,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32550,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(32550,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(32550,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(32550,36,'cp 14','f1e2 b8c6 c1e3 a7a6 h2h3 g4e6 a2a3 h7h6 f3d4 c6d4 e3d4 c7c5 d4e3 f6d5 e2f3 d5e3 f2e3 e8c8 c3d5 e7g5 d5f4 h8e8 h3h4 g5f4 e3f4 c8c7 e1f2 b7b5 b2b4 e6d7 a1e1 e8e1 h1e1 d8e8 e1e8 d7e8 h4h5 c5b4 a3b4');
INSERT INTO evaluations VALUES(32549,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(32549,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(32549,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(32549,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(32549,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(32549,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(32549,36,'cp 0','b8c6 a2a3 g4d7 e1g1 e8g8 f1e1 h7h6 c1d2 f8e8 g2g3 c6d8 e2f1 d8e6 f1g2 a7a5 c3d1 f6d5 d1c3');
INSERT INTO evaluations VALUES(40075,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(40075,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(40075,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(40075,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(40075,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(40075,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(40075,36,'cp 40','e2e3 b8d7 f1d3 d5c4 d3c4 b7b5 c4d3 e8g8 e1g1 c8b7 a1c1 a7a6 a2a4 a8c8 d1e2 d8b6 f1d1 b5b4 c3b1 c6c5 b1d2 c5d4 d2c4 b6c5 e3d4 c5h5 f4g3 b7c6 a4a5 c6f3 e2f3 h5f3 g2f3 f6d5 d3e4 f8d8');
INSERT INTO evaluations VALUES(41613,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(41613,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(41613,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(41613,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(41613,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(41613,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41613,36,'cp 26','f1e2 e8g8 e1g1 d8e7 a1b1 a7a5 b2b3 b7b6 c3a4 b8d7 c1b2 c8b7 b1c1 f8c8 f1e1 f6e4 f3d2 e4d2 d1d2 c7c6 e2f1 c8d8 b2c3 a8b8 h2h3 b8c8 f1e2 c8a8 d2c2 d8e8 e2d3 a8d8 c1d1 c6c5 d4c5 g7c3 a4c3 b6c5');
INSERT INTO evaluations VALUES(43983,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(43983,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(43983,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(43983,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(43983,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(43983,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(43983,36,'cp -43','d8b6 d4c5 f8c5 d1d2 b8c6 e1c1 b6a5 g5e3 c5b4 a2a3 b4c3 d2c3 a5c3 b2c3 c6e5 f3e5 d7e5 f2f4 e5c4 f1c4 d5c4 d1d4 c8d7 d4c4 d7c6 h1g1 e8g8 g2g4 f7f6 g4g5 g8f7 h2h4 f8e8 h4h5 e6e5 g5f6 g7f6 h5h6 e8g8 g1g8 a8g8 f4e5 f6e5 e3a7');
INSERT INTO evaluations VALUES(33544,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(33544,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(33544,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(33544,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(33544,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(33544,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(33544,36,'cp -62','d4c5 f8c5 d1d2 b8c6 e1c1 b6a5 g5e3 e8g8 e3c5 a5c5 d1e1 a8b8 c3b5 f7f6 e5f6 d7f6 f1d3 c6b4 b5c3 b4d3 d2d3 c5f2 e1e2 f2c5 d3d4 c5d6 h1e1 b7b5 c3e4 f6e4 e2e4 b8b7 c1b1 b5b4 f3e5 a7a5 e4g4 c8d7 g4g3 d7b5 g3f3 f8c8');
INSERT INTO evaluations VALUES(46568,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(46568,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46568,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(46568,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(46568,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46568,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(46568,36,'cp 19','d2d4 c7c5 c2c3 b8c6 f1d3 c5d4 c3d4 e7f5 d3f5 e6f5 b1c3 f8e7 h2h4 e8g8 h4h5 h7h6 d1b3 c8e6 e1g1 d8d7 c1f4 f8c8 f1c1 f7f6 c3e2 e6f7 e2g3 f6e5 f4e5 e7g5 c1c3 f5f4 g3e2 f7h5 e2f4 g5f4 e5f4 h5f3 c3f3');
INSERT INTO evaluations VALUES(48625,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(48625,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48625,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(48625,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(48625,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(48625,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(48625,36,'cp -6','h7h6 g5f6 d8f6 e2e3 c7c6 f1d3 f8d6 e1g1 e8g8 b1c3 c8e6 d1b3 f6e7 f1e1 e7c7 a1c1 a7a6 e3e4 d5e4 d4d5 d6h2 g1h1 e6c8 d3e4 h2d6 d5c6 b8c6 e4b1 c8g4 c3d5 c7d8 b3d3 f7f5 c1d1 d6c7 d5c7 d8c7 d3c3 g4f3 c3f3');
INSERT INTO evaluations VALUES(56154,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(56154,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(56154,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(56154,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(56154,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(56154,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(56154,36,'cp -77','g8f6 e2e3 c7c6 d5c6 d8d1 e1d1 b8c6 f1c4 f8b4 c1d2 e8g8 a2a3 b4e7 g1f3 c8f5 c3d5 f6d5 c4d5 e5e4 d5c6 b7c6 f3d4 f5g4 f2f3 e4f3 g2f3 g4d7 b2b4 c6c5 b4c5 e7c5 d2b4 c5d4 e3d4 f8b8 d1d2 a7a5 b4c3 d7f5 h1c1 h7h6 d2e3 b8b3 a3a4 a8c8 e3f4 b3c3 c1c3 c8c3 f4f5 c3f3 f5e4 f3f2 e4e5 f2e2 e5d6 e2h2');
INSERT INTO evaluations VALUES(48597,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(48597,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(48597,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(48597,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(48597,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(48597,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(48597,36,'cp -53','e2e3 c7c6 d5c6 d8d1 e1d1 b8c6 f1c4 f8b4 c1d2 e8g8 a2a3 b4e7 c3d5 f6d5 c4d5 f8d8 e3e4 e7c5 d1e1 c6e7 d2g5 g8f8 g5e7 c5e7 g1f3 f7f6 e1e2 c8d7 h1c1 a8c8 f3d2 b7b6 d2c4 d7b5 b2b3 e7c5 a3a4 b5a6 c1c2 c5d4 a1a2 f8e7');
INSERT INTO evaluations VALUES(6849,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(6849,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(6849,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(6849,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(6849,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(6849,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(6849,36,'cp 97','d2d3 d4f3 d1f3 g8f6 g2g4 d7d5 c4d5 f6d5 c3d5 c8e6 c1e3 c5d6 g4g5 c7c6 d5c3 e8g8 f3g3 d8a5 e3d2 b7b5 h2h4 b5b4 c3e2 a5c5 e1d1 f7f5 g5f6 f8f6 d2e3 c5a5 h1g1 f6g6 g3h2 g6g1 e2g1 e6g4 f2f3 g4e6 g1e2 c6c5 b2b3 a8f8');
INSERT INTO evaluations VALUES(6850,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(6850,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(6850,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(6850,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(6850,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(6850,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(6850,36,'cp 76','d4f3 d1f3 g8f6 g2g4 d7d5 c4d5 f6d5 c3d5 c8e6 c1e3 c5d6 h2h4 e8g8 e3d2 d8d7 h4h5 f7f6 d5e3 a7a5 a2a4 b7b6 b2b3 h7h6 e1g1 a8d8 g1h2 g8h7 f1g1 d7c6 f3e2 d6c5 e3f5 c5d4 f5d4 d8d4 d2e3 d4d8 f2f3 c6d7 a1d1 d7e7');
INSERT INTO evaluations VALUES(12416,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(12416,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(12416,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12416,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(12416,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(12416,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(12416,36,'cp 23','b1d2 d7d5 c2c3 f8d6 f4g3 e8g8 f1d3 h7h6 e1g1 d6g3 h2g3 b7b6 d1e2 c8b7 f1e1 d8e7 d4c5 e7c5 a1d1 a8d8 a2a3 c5e7 e3e4 d5e4 d2e4 f6g4 e4d2 g4f6 d3b1 c6b8 d2e4 b7e4 b1e4 d8d1 e1d1 f8d8 f3d4 a7a5 e2e3 e7c7 e4f3 b8d7');
INSERT INTO evaluations VALUES(22718,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(22718,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(22718,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(22718,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(22718,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(22718,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(22718,36,'cp -86','g8f6 b1c3 c6a5 b2b3 c7c6 f3g3 a5c4 b3c4 d7d6 a1b1 f8e7 g1e2 h7h5 h2h4 h8h6 f2f3 a7a6 a2a4 h6g6 g3h2 e8f8 d2d3 f8g8 e3e4 b7b5 c4b5 a6b5 a4b5 c6b5 c1g5 d8c7 b1b5 c8d7 b5b1 d6d5 e4d5 e7d6 h2g3 e5e4 g3f2 e4f3 g2f3 a8e8');
INSERT INTO evaluations VALUES(16423,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(16423,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(16423,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(16423,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(16423,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(16423,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(16423,36,'cp -65','b1c3 c6a5 b2b3 c7c6 f3g3 a5c4 b3c4 d7d6 e3e4 f8e7 g1e2 e8g8 d2d3 d6d5 c4d5 c6d5 e1g1 f8e8 c1g5 d5e4 g5f6 e7f6 d3e4 c8e6 f1d1 d8c8 a1b1 b7b6 g3f3 c8c5 c3d5 e6d5 e4d5 c5c2 e2g3 c2a2 d5d6 a2e6 g3f5');
INSERT INTO evaluations VALUES(28997,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(28997,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(28997,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(28997,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(28997,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28997,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28997,36,'cp 8','c2c4 h7h6 b1c3 e7e6 d1b3 d8b6 c4c5 b6b3 a2b3 f8e7 b3b4 e8g8 e1g1 a7a6 f3d2 f5c2 a1a2 b8d7 b2b3 c2f5 c1b2 a8d8 f1a1 g7g5 f2f3 g5g4 e2e4 g4f3 g2f3 d5e4 d2e4 f6d5 e4d6 d5b4');
INSERT INTO evaluations VALUES(28971,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28971,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28971,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28971,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28971,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28971,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28971,36,'cp 0','h7h6 d1b3 d8b6 b1c3 e7e6 c4c5 b6b3 a2b3 f8e7 e1g1 a7a6 b3b4 e8g8 f3d2 f5c2 h2h4 b8d7 d2f3 c2f5 f3d2');
INSERT INTO evaluations VALUES(31057,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31057,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(31057,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31057,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31057,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31057,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31057,36,'cp 0','c6d5 h2h3 g4f3 d1f3 e7e6 d2d4 b8c6 c1e3 a8c8 f1d3 g8f6 a2a3 c6a5 c3e2 a5c4 e1g1 f8d6 e3c1 e8g8 b2b3 c4a5 c1g5 h7h6 g5h4 d6e7 f1e1 f8e8 e1d1 a5c6 h4g3 e7d6 g3h4');
INSERT INTO evaluations VALUES(32993,30,'cp 279',NULL);
INSERT INTO evaluations VALUES(32993,31,'cp 274',NULL);
INSERT INTO evaluations VALUES(32993,32,'cp 271',NULL);
INSERT INTO evaluations VALUES(32993,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(32993,34,'cp 282',NULL);
INSERT INTO evaluations VALUES(32993,35,'cp 286',NULL);
INSERT INTO evaluations VALUES(32993,36,'cp 275','c1e3 c5b6 d1d2 g8e7 e1c1 e8g8 e3g5 f6d4 d2d4 b6d4 d1d4 f8e8 f1e2 e7g6 g5e3 b7b6 h1d1 g6f8 e2g4 b8a6 d4d2 a6c5 e3c5 b6c5 d2d6 a8b8 f2f4 f8e6 g2g3 b8b4 e4e5 b4c4 d1e1 g8f8 g4e6 e8e6 d6e6 f7e6 c3e4 c4a4 e4c5');
INSERT INTO evaluations VALUES(32992,30,'cp 280',NULL);
INSERT INTO evaluations VALUES(32992,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(32992,32,'cp 296',NULL);
INSERT INTO evaluations VALUES(32992,33,'cp 289',NULL);
INSERT INTO evaluations VALUES(32992,34,'cp 273',NULL);
INSERT INTO evaluations VALUES(32992,35,'cp 283',NULL);
INSERT INTO evaluations VALUES(32992,36,'cp 276','c5b6 d1d2 g8e7 e1c1 e8g8 e3g5 f6d4 d2d4 b6d4 d1d4 f8e8 f1e2 e7g6 g5e3 b8a6 h2h4 b7b5 h4h5 g6f8 h5h6 a6c5 f2f3 g7g6 d4d2 c5b7 e3f4 f7f6 a2a4 a7a6 f4h2 g8f7 h1d1 f8e6 e2f1 b7c5 b2b4 c5a4 c3a4 b5a4 f1c4 a6a5 b4b5 c6b5 c4b5 e8d8 e4e5 f6e5');
INSERT INTO evaluations VALUES(39087,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(39087,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(39087,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(39087,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(39087,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(39087,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(39087,36,'cp 9','g8f6 b1c3 e8g8 c4b3 b8c6 c1f4 d7d6 h2h3 d8e8 e1g1 c8e6 f1e1 c6e5 f4e3 e5g6 e3f2 e8e7 f3d2 c7c6 d1f3 a8e8 d2c4 e6c4 b3c4 c6c5 d4c5 b6c5 c3d5 f6d5 c4d5 b7b6 a1d1 g6e5 f3g3 c5f2 g3f2 e5d7 c2c3 d7f6 f2d4 f6d5 d4d5 e8d8 e1f1 d8e8 f1f5 e7e4 d5e4 e8e4');
INSERT INTO evaluations VALUES(43324,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(43324,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(43324,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(43324,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(43324,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(43324,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(43324,36,'cp 156','e4e5 f6g8 c1f4 g8e7 c3e4 e7d5 f4g3 b7b5 c2c3 c8b7 f1d3 b8c6 d4f3 f8e7 d1e2 a8c8 e1g1 c6a5 f1e1 a5c4 a1d1 b7a8 h2h4 e8g8 h4h5 f7f6 e5f6 d5f6 b2b3 c4b6 f3g5 f6d5 g5h7 g8h7 e2g4 h7g8 h5h6 f8f7 g3e5 g7g5');
INSERT INTO evaluations VALUES(50844,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(50844,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(50844,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(50844,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(50844,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(50844,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(50844,36,'cp 124','e4d5 d8d5 g1f3 d5d8 c2c4 g8f6 b1c3 e7e6 f1d3 c7c5 d4d5 e6d5 c4d5 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 b8d7 c3e4 d6b8 f1e1 d8b6 a1c1 f6e4 d3e4 d7f6 e4b1 c8g4 d1d3 g4h5 h4f6 h5g6 d3c4 b6f6 b1g6 f6g6 c4c5 b8d6 c5d4 f8c8 g2g3 c8c1 e1c1 a8e8');
INSERT INTO evaluations VALUES(50828,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(50828,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(50828,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(50828,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(50828,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(50828,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(50828,36,'cp 108','d8d5 g1f3 d5d8 c2c4 e7e6 b1c3 g8f6 f1d3 g7g6 c1g5 f8g7 e1g1 h7h6 g5e3 e8g8 f3e5 b8c6 e5c6 b7c6 h2h3 f6d7 f1e1 a8b8 d1d2 e6e5 e3h6 e5d4 c3e4 b8b2 d2b2 g7h6 b2d4 h6g7 d4e3 g7a1 e1a1 f8e8 a1d1 g8g7 d3c2');
INSERT INTO evaluations VALUES(34124,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(34124,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(34124,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(34124,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(34124,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(34124,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(34124,36,'cp 142','c2c4 d5d8 g1f3 g8f6 b1c3 e7e6 f1d3 g7g6 c1g5 f8g7 e1g1 h7h6 g5e3 e8g8 f3e5 b8c6 f1e1 d8d6 c4c5 d6d8 d1d2 g8h7 a1d1 f6d5 c3d5 e6d5 f2f4 c6e5 d4e5 c8g4 d1c1 c7c6 h2h3 g4d7 e3f2 f8e8 c1d1 d8e7 d2e3 a8d8 b2b4');
INSERT INTO evaluations VALUES(53540,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(53540,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(53540,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(53540,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(53540,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(53540,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(53540,36,'cp -30','e6d5 c4e2 f8d6 d2d4 g8f6 c2c4 d5c4 e2c4 e8g8 e1g1 c8f5 c1g5 b8d7 c4d3 f5d3 d1d3 d7c5 d3f5 c5e4 g5f6 e4f6 b1c3 g7g6 f5c2 f8e8 g2g3 d8b6 a1e1 d6f8 e1e5 f6d7 e5e8 a8e8 f1d1 d7f6 h2h3 e8d8 g1g2 g8g7 b2b3 d8d7 c3e4 f8e7');
INSERT INTO evaluations VALUES(53707,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(53707,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(53707,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(53707,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(53707,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(53707,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(53707,36,'cp -25','c4e2 f8d6 d2d4 g8f6 e1g1 e8g8 c2c4 d5c4 e2c4 c8f5 b1c3 b8d7 c1g5 d8c7 g5h4 f8e8 h4g3 a8d8 f1e1 e8e1 d1e1 d8e8 e1d2 d7b6 c4b3 f5e6 b3e6 e8e6 a1e1 e6e1 d2e1 d6g3 h2g3 g8f8 e1d2 g7g6 b2b3 c7e7');
INSERT INTO evaluations VALUES(3827,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(3827,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(3827,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(3827,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3827,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(3827,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3827,36,'cp 12','e1g1 h7h6 b2b4 c5b6 b1d2 c6e7 f1e1 e7g6 a2a4 a7a5 b4b5 f8e8 d3d4 e5d4 c3d4 d6d5 e4d5 e8e1 d1e1 f6d5 d2b3 c8e6 b3c5 b6c5 d4c5 d8e7 c4d5 e6d5 e1e7 g6e7 f3d4 a8d8 c1e3 f7f6 f2f3 g8f7 e3f2 c7c6 f2g3 c6b5 a4b5 d5c4 g3d6 e7c8 d4f5 c4b5 a1a5 b5c6 a5a8 c8d6');
INSERT INTO evaluations VALUES(9413,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9413,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(9413,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9413,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9413,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9413,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9413,36,'cp 1','f1c4 c8e6 d2d3 a7a6 c1e3 h7h6 d1d2 e8g8 e3c5 d6c5 c4e6 f7e6 c3e2 f6h5 c2c3 a6a5 e1g1 d8e7 d2e3 a8d8 g2g3 a5a4 g1g2 h5f6 a1d1 b7b6 d3d4 f6d7 d4e5 c6e5 e2f4 e7f7 f1e1 e5f3 e3f3 d7e5 f3e2');
INSERT INTO evaluations VALUES(13349,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(13349,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(13349,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(13349,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(13349,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(13349,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(13349,36,'cp -24','a7a6 c1f4 e7e6 d1e2 f6h5 f4e3 f8e7 f1d1 c8d7 f3d4 g7g6 d4c6 d7c6 e3h6 e7g5 h6g5 d8g5 d1d6 e8e7 a1d1 h8d8 d6d8 a8d8 d1d8 e7d8 e2d1 d8e7 d1d4 h5f6 f2f3 f6d7 c4e2 g5f4 b2b3 g6g5 c3d1 h7h5 d4b4 e7e8 g2g3 f4c1 g1f2 c1c2 d1e3 c2a2 e3c4 a2c2 c4d6 e8f8 d6b7 f8g8');
INSERT INTO evaluations VALUES(11165,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(11165,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(11165,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(11165,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(11165,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(11165,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(11165,36,'cp -8','c1f4 e7e6 d1e2 f6h5 f4e3 f8e7 f1d1 c8d7 f3d4 g7g6 d4c6 d7c6 e3h6 e7g5 h6g5 d8g5 d1d6 e8e7 a1d1 h8d8 d6d8 a8d8 d1d8 e7d8 e2d1 d8e7 d1d4 g5c1 c4f1 h5f6 f2f3 f6d7 c3d1 f7f6 g1f2 c1c2 f1e2 c6b5 d1c3 b5e2 c3e2 b7b5 a2a3 c2c7 h2h3 c7b6 d4b6 d7b6 b2b3 e7d6 f2e3 d6c5');
INSERT INTO evaluations VALUES(15439,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15439,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(15439,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(15439,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(15439,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(15439,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(15439,36,'cp 29','g1e2 f8e7 e1g1 e8g8 h2h3 c8e6 b5c6 b7c6 c2c4 a8b8 d2d3 f6d7 d1c2 f7f6 b1d2 d6d5 f2f4 f8e8 d3d4 e5f4 e2f4 e6f7 a1e1 e7f8 e3e4 d5e4 d2e4 a7a5 b2c3 a5a4 e4g3 e8e1 f1e1 d7b6 f4e6 a4b3 a2b3 d8d7');
INSERT INTO evaluations VALUES(18182,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(18182,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(18182,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(18182,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(18182,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(18182,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(18182,36,'cp 86','f2f4 c5c4 b2b3 c4b3 a2b3 e7e6 g1f3 c6c5 e1g1 c8b7 f4f5 f8e7 f5e6 d7e6 d2d3 g8f6 d1e1 e8g8 e1g3 f6d7 c3e2 f7f6 c1b2 f8f7 f3d2 d7b6 g3h3 d8d7 b2c3 b7c8 h3g3 e6e5 d2c4 b6c4 d3c4');
INSERT INTO evaluations VALUES(18187,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(18187,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(18187,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(18187,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(18187,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(18187,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(18187,36,'cp 69','c5c4 b2b3 c4b3 a2b3 c8b7 g1f3 e7e6 e1g1 g8f6 e4e5 f6d5 c3d5 c6d5 c2c3 f8c5 d2d4 c5e7 c1a3 e8g8 d1d3 h7h6 a3e7 d8e7 g2g4 b7c6 f1f2 a6a5 d3c2 f8c8 f4f5 f7f6 e5f6 e7f6 f3e5 e6f5 f2f5 f6e6 c2f2');
INSERT INTO evaluations VALUES(24879,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(24879,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24879,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24879,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(24879,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24879,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(24879,36,'cp 22','e2e3 f6e4 g5f4 e4c3 b2c3 e7e6 f1d3 f5d3 c2d3 f8d6 f4g3 e8g8 e1g1 c6a5 d1a4 b7b6 c3c4 d5c4 d3c4 c7c5 f1d1 d6g3 h2g3 c5d4 e3d4 a8c8 a1c1 c8c7 c4c5 d8d7 a4c2 f8d8 f3e5 d7d5 c5c6 d5d6 f2f4 f7f6');
INSERT INTO evaluations VALUES(24878,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(24878,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24878,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(24878,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(24878,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(24878,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24878,36,'cp 20','f6e4 g5h4 d8d7 c3e4 d5e4 f3d2 e7e5 c2c3 f8d6 a2a3 e8g8 h4g3 a8c8 f1b5 e5d4 c3d4 a7a6 b5e2 c6e7 a1c1 e7d5 d2b3 d7e7 b3c5 d6g3 h2g3 c7c6 g3g4 f5g6 d1b3 c8c7 c5a4 f7f6 c1c5 g6f7 a4b6 d5b6 b3b6 f7d5 e1g1');
INSERT INTO evaluations VALUES(28117,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(28117,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(28117,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(28117,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(28117,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(28117,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(28117,36,'cp 60','b8c6 c1e3 g8f6 d4d5 c6b8 a2a4 b8d7 e1g1 e8g8 f1e1 a7a6 h2h3 g4f3 e2f3 c7c6 a4a5 d8c7 c3a4 a8e8 d5c6 b7c6 b2b4 e8b8 c2c3 c6c5 a1c1 c7d8 c1b1 d8c7 d1b3 b8b5 f3e2 f6e4 b1d1 f8b8 b3c2 d7f6 e2b5 a6b5 a4b6');
INSERT INTO evaluations VALUES(23822,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(23822,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(23822,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(23822,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(23822,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(23822,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(23822,36,'cp 67','c1e3 g8f6 d4d5 c6b8 a2a4 b8d7 h2h3 g4f3 e2f3 e8g8 a4a5 c7c6 e1g1 d8c7 f1e1 a7a6 c3a4 a8e8 d5c6 b7c6 b2b4 e8b8 c2c3 c6c5 a1c1 c7d8 c1b1 d8c7 d1d2 b8b5 e1d1 f8c8 c3c4 b5b4 b1b4 c5b4 a4b6 b4b3 b6c8 c7c8');
INSERT INTO evaluations VALUES(29286,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29286,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29286,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(29286,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29286,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(29286,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29286,36,'cp 13','f8b4 e1g1 e8g8 c1d2 b4a5 a1b1 b8d7 h2h3 h7h6 f1d1 a8c8 b2b4 a5c7 a2a4 d5c4 d3c4 f8e8 e3e4 e6e5 d4e5 d7e5 f3e5 c7e5 d2h6 d8e7 h6d2 e5c3 d2c3 e7e4 c4e4 f6e4 c3e1 e4f6 g1f1 a7a6 b4b5 c6b5 a4b5 c8c2 b5a6 b7a6');
INSERT INTO evaluations VALUES(28359,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(28359,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(28359,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28359,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(28359,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(28359,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(28359,36,'cp 15','e1g1 e8g8 f1d1 f8e8 h2h3 b8d7 c1d2 b4a5 a1b1 h7h6 b2b4 a5c7 b4b5 d7b6 c4c5 b6c4 b5c6 b7c6 e3e4 c4d2 d1d2 c7a5 d2b2 a5c3 d3c3 f6e4 c3a3 f7f6 b2b7 e6e5 b7a7 a8a7 a3a7 d8a8 a7a8 e8a8 d4e5 f6e5 f3e5 e4c5 e5c6 a8a2 c6e7 g8h7 e7d5');
INSERT INTO evaluations VALUES(41139,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41139,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(41139,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(41139,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41139,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41139,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41139,36,'cp 22','c3b5 d6b4 c2c3 b4a5 f1d3 c7c6 b5a3 e8g8 e1g1 a5c7 a3c2 h7h6 f3e5 c6c5 f1e1 c5c4 d3e2 b8c6 e5c6 b7c6 b2b3 c4b3 a2b3 a7a5 c1a3 f8e8 e2d3 c8e6 a3c5 e7f5 d1f3 d8g5 c2e3 f5e3 e1e3');
INSERT INTO evaluations VALUES(42948,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42948,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42948,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(42948,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(42948,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(42948,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42948,36,'cp 28','b7b6 g2g3 c8b7 f1g2 e8g8 e1g1 f6e4 g5e7 d8e7 a1c1 d7d6 c3e4 b7e4 c4c5 d6c5 d4c5 b8c6 c5b6 a7b6 d1b3 h7h6 a2a3 e7f6 b3c3 a8c8 f1d1 f8d8 c3f6 g7f6 d1d8 c8d8 f3e1 c6d4 g2f1 c7c6 e2e3 d4f3 e1f3 e4f3');
INSERT INTO evaluations VALUES(52259,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(52259,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(52259,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(52259,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(52259,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(52259,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(52259,36,'cp 87','d2d4 f8g7 f1d3 d7d6 h2h3 b8d7 c1e3 g8f6 e1g1 e8g8 b1d2 a7a5 a2a4 d8c7 d1e2 f8e8 c2c3 b7b6 f1e1 e7e5 d4e5 d7e5 f3e5 d6e5 b2b4 f6h5 e2f3 h5f4 d3f1 a5b4 d2c4 b6b5 c4b6 a8b8 b6c8 e8c8 c3b4');
INSERT INTO evaluations VALUES(53445,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(53445,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(53445,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(53445,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(53445,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(53445,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(53445,36,'cp 57','c3d4 b8c6 b1c3 c8f5 h2h4 h7h5 c1g5 d8d7 f1b5 a8c8 g1e2 a7a6 b5d3 f5d3 d1d3 e7e6 a1c1 f8e7 d3d2 f7f6 e5f6 g8f6 e2f4 e8g8 f4g6 f8f7 e1g1 e7d6 g5f4 c6e7 f4d6 e7g6 d6g3 g6e7 f2f3 e7f5 c3e2 f7f8 g3e5 c8c4 e2f4 c4c1 d2c1');
INSERT INTO evaluations VALUES(10155,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(10155,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(10155,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(10155,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(10155,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(10155,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(10155,36,'cp -8','d7d6 d2d3 e8g8 f1e2 c6d4 f3d4 c5d4 c1e3 d4e3 f2e3 c7c6 d1d2 a7a6 a2a4 c8e6 e1g1 d6d5 e4d5 c6d5 c4d5 f6d5 e2f3 a8c8 f1c1 h7h6 a4a5 f7f5 c3d5 c8c1 a1c1 e6d5 f3d5 d8d5 c1c7 f5f4 e3f4 e5f4 d2c3');
INSERT INTO evaluations VALUES(9374,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(9374,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(9374,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(9374,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9374,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(9374,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9374,36,'cp -1','d2d3 e8g8 f1e2 a7a6 e1g1 c8d7 a2a3 c6d4 b2b4 c5a7 a3a4 c7c6 a4a5 f8e8 c1e3 f6h5 c4c5 d7e6 f3d4 e5d4 e3d4 d6c5 d4c5 a7c5 b4c5 h5f4 g1h2 d8g5 e2f3 g5c5 c3e2 f4e2 d1e2 c5d6 e4e5 d6d4 f1c1 d4b4 e2c2');
INSERT INTO evaluations VALUES(9377,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(9377,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9377,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(9377,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(9377,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(9377,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(9377,36,'cp -22','e8g8 f1e2 c6d4 f3d4 c5d4 c1e3 d4e3 f2e3 c7c6 d1d2 a7a6 a2a4 c8e6 e1g1 d6d5 e4d5 c6d5 c4d5 f6d5 e2f3 a8c8 f3d5 e6d5 f1f5 f7f6 d3d4 d5f7 d4d5 d8d7 f5f2 d7d6 a4a5 f7g6 a1d1 c8c4 d2e2 f8c8 e3e4');
INSERT INTO evaluations VALUES(3850,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(3850,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(3850,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(3850,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3850,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3850,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3850,36,'cp 0','f1e2 a7a6 e1g1 a8b8 a1b1 h7h6 a2a3 c6d4 b2b4 c5a7 f1e1 c7c6 e2f1 c8e6 f3d4 a7d4 c3e2 d4a7 e2g3 b7b5 c1d2 f8e8 b1c1 a7d4 g3e2 d4a7');
INSERT INTO evaluations VALUES(3849,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3849,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3849,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(3849,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(3849,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(3849,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(3849,36,'cp -5','a7a6 e1g1 a8b8 a1b1 h7h6 a2a3 c6d4 b2b4 c5a7 c1e3 d4f3 e2f3 a7e3 f2e3 b7b5 a3a4 c7c6 a4b5 a6b5 c4b5 c6b5 d1d2 c8e6 b1a1 d8d7 a1a5 f8c8 f1a1 b8b7 d3d4 b7c7 d4e5 f6e8 a5a3 d6e5 d2d7 e6d7 c3d5 c7c1 a1c1 c8c1 g1f2 c1c2 f2g3');
INSERT INTO evaluations VALUES(14756,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(14756,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(14756,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14756,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(14756,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(14756,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(14756,36,'cp 62','c4d5 e6d5 g1f3 f8b4 d1b3 h7h6 g5f6 d8f6 e2e3 f6d6 f1b5 b4c3 b3c3 c8d7 e1g1 e8g8 f1c1 f8c8 c3c5 d6e6 b2b4 c6e7 b5d3 b7b6 c5c2 a7a5 a2a3 a5b4 a3b4 a8a1 c1a1 e6c6 c2d1 c8a8 a1a8 c6a8 f3e5 d7e8 d1g4 c7c6');
INSERT INTO evaluations VALUES(15601,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(15601,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(15601,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15601,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(15601,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(15601,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(15601,36,'cp 15','c6d4 f1c4 c7c6 d2d3 b7b5 c4a2 d4f3 d1f3 f8e7 c3e2 e8g8 c2c3 c8e6 e2g3 d8d7 a2e6 f7e6 e1g1 a7a5 f3e2 a5a4 f2f4 e5f4 c1f4 e6e5 f4e3 d6d5 f1f5 g7g6 f5e5 e7d6 e5g5 a8e8 a1f1 d5e4 d3e4');
INSERT INTO evaluations VALUES(32239,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(32239,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(32239,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(32239,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(32239,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32239,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(32239,36,'cp 9','g8f6 h2h3 a7a6 a2a3 f8d6 f1d3 b8d7 e1g1 e8g8 f1e1 f8e8 c3e2 d7f8 c1f4 f8g6 f4d6 d8d6 e2g3 e6d7 d1d2 d7b5 d3g6 e8e1 a1e1 h7g6 c2c3 a8e8 e1e5 e8f8 h3h4 f8e8 d2f4 b5d7 e5e8 f6e8 f3e5 d6f6 f4f6 g7f6');
INSERT INTO evaluations VALUES(30023,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(30023,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(30023,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(30023,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(30023,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(30023,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(30023,36,'cp 21','f1d3 c7c5 d3b5 b8c6 e1g1 f8e7 d4c5 e7c5 c1g5 e8g8 d1d3 c5e7 a1d1 h7h6 g5f6 e7f6 c3d5 f6b2 d3b3 b2d4 b5c6 b7c6 f3d4 c6d5 d4e6 f7e6 c2c4 d5c4 b3c4 d8b6 h2h3 a8d8 d1d8 f8d8 c4e4 b6a6 e4c2 a6d3');
INSERT INTO evaluations VALUES(37046,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37046,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37046,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37046,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37046,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37046,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37046,36,'cp 3','a2a3 b4c3 c2c3 d5c4 c3c4 b7b6 c1g5 c8a6 c4a4 h7h6 g5h4 c7c5 d4c5 b6c5 a1d1 d8b6 h4f6 g7f6 e2e3 a6f1 e1f1 b8c6 d1b1 a8d8 f1e2 d8d5 h1c1 a7a5 e2f1 f8d8 h2h3 c6e5 f3e5 d5e5 b2b4 c5b4 a3b4 g8g7 f1g1 e5g5 h3h4 g5f5 b4a5 b6a5 a4a5 f5a5');
INSERT INTO evaluations VALUES(44290,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44290,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(44290,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44290,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44290,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44290,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(44290,36,'cp 25','b8c6 d4c6 d7c6 e1g1 e6e5 b1d2 c8e6 d1f3 f8e7 a2a4 e8g8 h2h3 h7h6 d2c4 d8c7 b2b3 f6d7 c1d2 f8d8 f1d1 d7f8 d3f1 f8g6 a4a5 e7c5 b3b4 c5f8 f3g3 d8d2 d1d2 f8b4 d2d1');
INSERT INTO evaluations VALUES(12481,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(12481,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(12481,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12481,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(12481,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(12481,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(12481,36,'cp 33','d4c6 d7c6 b1d2 e6e5 e1g1 c8e6 d1f3 f8e7 d2c4 f6d7 d3e2 e8g8 a2a4 d8c7 c4e3 f8e8 f1d1 a7a5 b2b3 e7c5 e3f5 e6f5 f3f5 d7f8 c1b2 g7g6 f5f6 e8e6 f6h4 a8e8 g1h1 h7h5 f2f3 c7e7 h4e7 e6e7 e2d3 h5h4 d3c4 h4h3 g2g3 e7d7');
INSERT INTO evaluations VALUES(12294,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(12294,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12294,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(12294,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(12294,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(12294,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(12294,36,'cp 22','d7c6 e1g1 e6e5 b1d2 c8e6 d1e2 f8c5 d3c4 d8e7 c4e6 e7e6 d2b3 c5e7 f1d1 e8g8 a2a4 a8d8 c1g5 h7h6 g5f6 e7f6 g2g3 f6e7 b3d2 d8d7 d2c4 f8d8 d1d7 d8d7 g1g2 g7g6 h2h4 h6h5 a4a5 d7d4 c4d2 g8g7');
INSERT INTO evaluations VALUES(52369,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52369,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(52369,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(52369,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(52369,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(52369,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52369,36,'cp 12','d7d5 c1f4 g8f6 b1d2 c8f5 g1f3 e7e6 e2e3 h7h6 d1b3 d8c8 f1e2 b8d7 f3e5 d7e5 f4e5 f8e7 e1g1 e8g8 a2a4 a7a5 h2h3 f6d7 e5h2 b7b6 f2f3 c8b7 e3e4 f5h7 g1h1 a8c8 f1c1 c6c5 e2b5 c5c4 b3d1');
INSERT INTO evaluations VALUES(53706,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(53706,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(53706,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(53706,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(53706,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(53706,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53706,36,'cp -18','c6d5 d2d4 f8b4 b1c3 e5e4 f2f3 g8f6 c1g5 e8g8 f3e4 d5e4 g1h3 b8d7 e1g1 h7h6 g5h4 b4c3 b2c3 d8b6 e2e3 b6c6 g3g4 c6c3 h3f2 c3e3 f1e1 e3a3 f2e4 f6e4 g2e4 a3a6 d1f3 d7f6 h4f6 a6f6 f3f6 g7f6 e4f3 f6f5 g4f5 c8f5');
INSERT INTO evaluations VALUES(56859,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(56859,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(56859,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(56859,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(56859,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(56859,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(56859,36,'cp 11','d7d5 e4d5 e6d5 c4e2 g8f6 e1g1 f8g7 d2d4 e8g8 f1e1 c7c6 a2a4 a7a5 e2d3 f8e8 e1e8 f6e8 c2c3 e8d6 c1f4 b8a6 b1d2 a6c7 d1b3 c7e6 f4g3 g7h6 h2h3 h6f4 g3h4 e6g5 h4g5 f4g5 f3g5 d8g5 d2f3 g5f6 b3a3 g8g7 a1e1 c8f5 d3f5 d6f5');
INSERT INTO evaluations VALUES(4106,30,'cp -262',NULL);
INSERT INTO evaluations VALUES(4106,31,'cp -274',NULL);
INSERT INTO evaluations VALUES(4106,32,'cp -269',NULL);
INSERT INTO evaluations VALUES(4106,33,'cp -298',NULL);
INSERT INTO evaluations VALUES(4106,34,'cp -257',NULL);
INSERT INTO evaluations VALUES(4106,35,'cp -275',NULL);
INSERT INTO evaluations VALUES(4106,36,'cp -252','f8f7 g5f7 g8f7 b1c3 f7g8 e1g1 d7d6 c3a4 c8g4 d1d2 c5d4 c2c3 d4b6 h2h3 g4e6 a4b6 a7b6 d2e2 d8e8 f2f4 e5f4 c1f4 c6e5 f4e5 d6e5 g1h2 e6f7 f1f2 f6h5 d3d4 h5f4 e2d2 a8d8 a1f1 h7h6');
INSERT INTO evaluations VALUES(4644,30,'cp -257',NULL);
INSERT INTO evaluations VALUES(4644,31,'cp -254',NULL);
INSERT INTO evaluations VALUES(4644,32,'cp -271',NULL);
INSERT INTO evaluations VALUES(4644,33,'cp -252',NULL);
INSERT INTO evaluations VALUES(4644,34,'cp -280',NULL);
INSERT INTO evaluations VALUES(4644,35,'cp -270',NULL);
INSERT INTO evaluations VALUES(4644,36,'cp -267','g5f7 g8f7 b1c3 f7g8 e1g1 h7h6 g1h1 d7d6 c3a4 f6g4 a4c5 d8h4 h2h3 d6c5 c1e3 g4e3 f2e3 c5c4 b2b3 c4d3 c2d3 c8e6 a1c1 a8d8 b3b4 a7a6 h1h2 d8d7 a2a3 c6d8 d1c2 c7c6 f1f2 h4e7 c1f1 e6f7 c2e2 f7g6 e2a2 g8h7 a2c2 d8f7 c2c3 f7d6');
INSERT INTO evaluations VALUES(4641,30,'cp -287',NULL);
INSERT INTO evaluations VALUES(4641,31,'cp -289',NULL);
INSERT INTO evaluations VALUES(4641,32,'cp -270',NULL);
INSERT INTO evaluations VALUES(4641,33,'cp -283',NULL);
INSERT INTO evaluations VALUES(4641,34,'cp -256',NULL);
INSERT INTO evaluations VALUES(4641,35,'cp -280',NULL);
INSERT INTO evaluations VALUES(4641,36,'cp -286','g8f7 b1c3 d7d5 c1g5 h7h6 g5f6 d8f6 e1g1 f6g6 g1h1 d5e4 c3e4 c5b6 d1d2 c8e6 f2f4 f7g8 a1e1 a8e8 b2b4 a7a6 e4g3 e5f4 f1f4 a6a5 b4b5 c6d8 d3d4 c7c6 b5c6 b7c6 f4e4 b6c7 d2e2 c7g3 h2g3 g6g3');
INSERT INTO evaluations VALUES(5917,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(5917,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(5917,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(5917,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(5917,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(5917,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5917,36,'cp 35','g1f3 g8f6 f4g3 b7b6 f1d3 c6e7 b1d2 c8b7 b2b4 e7f5 g3e5 e8g8 e1g1 f6e4 d1c2 e4d2 f3d2 g7g6 a2a4 a7a5 d3f5 d6e5 d4e5 g6f5 b4b5 c7c5 b5c6 b7a6 f1b1 a8c8 d2f3 c8c6 f3d4 c6c5 c2b3 d8c7 b3b6 a6d3 b1b2 c7b6 b2b6 c5c3');
INSERT INTO evaluations VALUES(10405,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(10405,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(10405,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(10405,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10405,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(10405,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(10405,36,'cp 36','a2a3 b4c3 b2c3 e8g8 g1e2 c6e7 f2f3 c7c6 c4c5 b7b6 d3d4 e5d4 c3d4 b6c5 d4c5 d8a5 c1d2 a5c5 d2b4 c5e5 d1d4 e5d4 e2d4 f8e8 b4d6 a7a5 e1f2 c8a6 f1a6 a8a6 h1c1 e7c8 d6g3 c8e7 a1b1 d7d5 e4e5 f6d7 b1b7 d7f8');
INSERT INTO evaluations VALUES(12487,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(12487,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(12487,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12487,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(12487,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(12487,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(12487,36,'cp 12','f8e7 f1e2 e8g8 c1e3 d7d6 d1d3 a7a6 f2f4 e6e5 d4c6 b7c6 f4e5 d6e5 d3d8 f8d8 e2d3 a6a5 c3a4 c8e6 e3b6 d8c8 b6c5 e7c5 a4c5 g8f8 e1c1 f8e7 c5e6 f7e6 h1e1 c8d8 e1e3 d8d4 c2c3 d4d8 d3e2 d8d1 e2d1 c6c5 b2b3 a8b8 h2h4 h7h5 e3g3 f6e4');
INSERT INTO evaluations VALUES(7934,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7934,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7934,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7934,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7934,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(7934,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(7934,36,'cp 26','f1e2 d7d6 c1e3 e8g8 d1d3 a7a6 f2f4 e6e5 d4c6 b7c6 f4e5 d6e5 d3d8 f8d8 e2d3 a6a5 c3a4 c8a6 e1c1 a6d3 c2d3 a8b8 h1e1 b8b3 e3b6 d8b8 b6c7 b8b7 c7e5 f6g4 e5d4 g4h2 e4e5 h2g4 e1e2 b3b5 a4c3 b5b3 c3e4 b7b8 d4c3 b3b5 e4d6');
INSERT INTO evaluations VALUES(20626,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(20626,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(20626,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(20626,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(20626,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(20626,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(20626,36,'cp 87','b1c3 g8f6 f1b5 c8e6 e1g1 f8e7 e4e5 d6e5 f3e5 d8b6 d1f3 a8c8 b2b3 e8g8 e5c6 b7c6 b5d3 f8e8 c1d2 b6d8 f1e1 c5c4 b3c4 e6c4 d3f5 c8a8 a1d1 d8c7 d2f4 c7a5 e1e5 c4d5 c3d5 c6d5 f5d3 a5b6 c2c4 e7d6 e5e8 a8e8 c4d5 d6f4 f3f4 f6d5 f4f5');
INSERT INTO evaluations VALUES(20625,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(20625,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(20625,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(20625,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(20625,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(20625,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(20625,36,'cp 66','f8e7 f1b5 c8e6 e1g1 g8f6 e4e5 d6e5 f3e5 d8b6 d1f3 a8c8 b2b3 e8g8 e5c6 b7c6 b5d3 f8e8 c1d2 e6g4 f3f4 g4h5 f1e1 h5g6 c3a4 b6b7 d3g6 h7g6 c2c4 b7d7 d2c3 e7f8 f4f3 e8e1 a1e1 c8e8 e1e8 d7e8 g2g3 e8e6 c3d2 f6e4 d2e3 g6g5 h2h3 f7f5 g1g2 f8e7 a4b2');
INSERT INTO evaluations VALUES(38593,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(38593,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(38593,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(38593,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(38593,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(38593,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(38593,36,'cp 140','g8f6 d1d3 b8c6 b1c3 c6a5 c4b5 c7c6 b5a4 b6c7 b2b4 b7b5 c3b5 c6b5 a4b5 c8d7 b4a5 d7b5 d3b5 d8d7 b5d7 e8d7 e4e5 d6e5 f3e5 c7e5 f4e5 f6e4 a1b1 f7f6 b1b7 d7c8 b7c7 c8d8 e1g1 a7a6 c7g7 f6e5 d4e5 a8b8 h2h4 d8c8 g7a7 h8e8 f1f7 b8b1 g1h2 c8b8 f7h7 b1b5 a7a6 b5e5 h7a7 b8c8 a7a8 c8c7 a8e8 e5e8');
INSERT INTO evaluations VALUES(41664,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(41664,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(41664,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(41664,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(41664,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(41664,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(41664,36,'cp 143','d1d3 b8c6 b1c3 c6a5 c4b5 c7c6 b5a4 b6c7 b2b4 b7b5 c3b5 c6b5 a4b5 c8d7 b4a5 d7b5 d3b5 d8d7 b5d7 e8d7 e4e5 d6e5 f3e5 c7e5 f4e5 f6e4 a1b1 f7f6 b1b7 d7c8 b7c7 c8d8 e1g1 a7a6 c7g7 f6e5 d4e5 a8b8 h2h4 h7h5 g7a7 h8e8 f1f7 e8e5 f7d7 d8e8 d7h7 e8d8 a7a6 e5e7 h7h5 b8b2 h5h8 d8c7 a6a7 b2b7 a7b7 c7b7 g2g4 e7f7 g4g5');
INSERT INTO evaluations VALUES(38887,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(38887,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(38887,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(38887,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(38887,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(38887,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(38887,36,'cp 151','g8e7 g1f3 e8g8 e1g1 b7b6 f1e1 c7c5 e3e4 h7h6 d4d5 e6d5 e4d5 e7g6 c4f1 c8g4 h2h3 g4f3 d1f3 b8d7 f3g3 f8e8 e1e8 d8e8 c3c4 e8e4 c1b2 a8e8 a2a4 e4c2 g3c3 c2c3 b2c3 g6h4 f1d3 h4g6 a4a5 d7f6 c3b2 g6e5 d3f1 e5g6 a1a3 g8f8 f2f3 f6d7 a5b6 a7b6 g2g3');
INSERT INTO evaluations VALUES(39518,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(39518,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(39518,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(39518,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(39518,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(39518,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(39518,36,'cp -122','e7h4 e1d2 d7d6 f1b5 b8c6 d1e2 a7a6 b5c6 b7c6 d2d1 d8f6 e4e5 d6e5 f3e5 g8h6 e5c6 e8f8 c6e5 c8e6 b1c3 a8d8 c1d2 d8d4 d1c1 h6f5 e5f3 e6c4 e2e5 f6e5 f3e5 g7g5 e5f3 d4d8 b2b3 c4d5 c3d5 d8d5 f3h4 f5h4');
INSERT INTO evaluations VALUES(39435,30,'cp -115',NULL);
INSERT INTO evaluations VALUES(39435,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(39435,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(39435,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(39435,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(39435,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(39435,36,'cp -62','e1d2 d7d6 f1d3 c8g4 c2c3 g8e7 d2c2 e7g6 d1f1 c7c5 f3h4 d8h4 c1f4 c5d4 f4d6 b8c6 b1a3 e8c8 a3b5 d4c3 b2c3 a7a6 d6g3 h4e7 b5c7 e7a3 c7d5 g6e5 a1b1 g4e6 b1b3 a3a2 b3b2 a2a4 c2b1 e6d5 e4d5 d8d5 d3f5 c8b8 f5c2 a4a5 f1f7 d5b5 b2b5');
INSERT INTO evaluations VALUES(47020,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(47020,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(47020,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(47020,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(47020,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(47020,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(47020,36,'cp 31','g7g6 f1g2 f8g7 c2c4 e8g8 e1g1 b8c6 b1c3 e7e5 d4d5 c6b8 b2b3 a7a5 e2e4 b8a6 a2a3 a6c5 d1c2 a5a4 b3b4 c5b3 a1b1 c8d7 c1b2 f6g4 h2h3 g4h6 b1e1 f7f5 e4f5 g6f5 c3e2 b7b6 f3d2 f5f4 d2b3 a4b3 c2b3');
INSERT INTO evaluations VALUES(46697,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(46697,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(46697,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(46697,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(46697,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(46697,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(46697,36,'cp 42','c2c4 f8g7 f1g2 e8g8 e1g1 c8f5 c1e3 d6d5 b1d2 c7c6 a1c1 d8b6 b2b3 a7a5 h2h3 a5a4 b3a4 b6a5 c4d5 f6d5 e3g5 h7h6 g5h4 f5e6 d2b3 a5a4 e2e4 d5b6 b3c5 a4d1 c1d1 e6a2 c5b7 f8e8 b7c5 a2c4 f1e1');
INSERT INTO evaluations VALUES(46670,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(46670,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46670,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(46670,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(46670,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(46670,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(46670,36,'cp 44','f8g7 f1g2 e8g8 e1g1 c8f5 d1b3 b8c6 b3b7 a7a5 b7b3 c6b4 f3e1 a8b8 b3a3 f6d7 a3a5 g7d4 b1c3 d4g7 c1d2 d7c5 a5a3 c7c6 b2b3 d8c8 a1c1 f5h3 c3d1 b4a6 d2c3 g7h6 f2f4 h3g2 g1g2 c5e4 e1d3 e4c3 d1c3 h6g7');
INSERT INTO evaluations VALUES(47567,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(47567,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(47567,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(47567,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(47567,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(47567,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(47567,36,'cp 223','f3g5 d6d5 e4d5 f8g7 d2d4 e5d4 d1e2 d8e7 e2e7 e8e7 c3b5 h7h6 g5f3 b8a6 b2b3 h8e8 e1g1 e7f8 c1a3 f8g8 b5d4 c8g4 a1e1 g4f3 d4f3 a6b8 f3e5 b8d7 e5d7 f6d7 e1e8 a8e8 c4b5 e8d8 a3e7 d8e8 e7b4 e8d8 c2c4 d7f6 f1e1 a7a6 b4e7 d8c8');
INSERT INTO evaluations VALUES(47563,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(47563,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(47563,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(47563,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(47563,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(47563,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(47563,36,'cp 201','d6d5 e4d5 f8g7 d2d4 e5d4 d1e2 d8e7 e2e7 e8e7 c3b5 h7h6 g5f3 b8a6 b2b3 h8e8 e1g1 e7f8 b5d4 c8g4 c1a3 f8g8 a1e1 g4f3 d4f3 a6b8 f3e5 b8d7 e5d7 f6d7 e1e8 a8e8 c4b5 e8d8 b5d7 d8d7 c2c4 c7c6 f1d1 c6d5 d1d5 d7d5 c4d5 f7f5 g1f1 g8f7 f1e2 g7e5 h2h3 e5d4 f2f3 h6h5 e2d3 d4b6');
INSERT INTO evaluations VALUES(48049,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(48049,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(48049,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(48049,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(48049,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(48049,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(48049,36,'cp 51','c2c4 f8e7 h2h3 e8g8 g1f3 b7b6 b1c3 d6d5 f1e2 c8b7 e1g1 c7c5 c4d5 f6d5 c3d5 e6d5 b2b3 b8c6 a1c1 f8e8 e2b5 a7a6 b5e2 a8c8 d1d2 c5d4 f3d4 c6d4 e3d4 c8c1 f1c1 e7d6 f4d6');
INSERT INTO evaluations VALUES(50272,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(50272,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(50272,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(50272,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(50272,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(50272,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(50272,36,'cp 158','c2c4 d5b6 c4c5 b6d5 f1c4 e7e6 c4d5 e6d5 b1c3 c7c6 g1f3 b8a6 a1b1 a6c7 e1g1 b7b6 b2b4 c7e6 h2h4 b6c5 b4c5 f8g7 c1g5 e6g5 h4g5 c8a6 f1e1 h7h6 g5h6 g7h6 g2g3 a8b8 b1b8 d8b8 e5e6 d7e6 d1a4 a6b7 c3d5 b8d8 a4a7');
INSERT INTO evaluations VALUES(50270,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(50270,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(50270,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(50270,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(50270,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(50270,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(50270,36,'cp 150','d5b6 c4c5 b6d5 f1c4 e7e6 c4d5 e6d5 b1c3 c7c6 a1b1 b7b6 b2b4 b8a6 g1f3 a6c7 e1g1 c7e6 f1e1 f8g7 h2h4 f7f6 c3e2 f6e5 d4e5 e8g8 e2d4 e6d4 d1d4 c8a6 c1g5 d8e8 g5f6 g7f6 e5f6 e8f7 e1e7 f7f6 e7d7 f6d4 f3d4 f8f7 d7f7 g8f7 d4c6 a6c4 c6e5 f7e6 e5c4 d5c4');
INSERT INTO evaluations VALUES(50140,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(50140,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(50140,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(50140,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(50140,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(50140,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(50140,36,'cp 141','c4c5 b6d5 f1c4 e7e6 c4d5 e6d5 b1c3 c7c6 b2b4 b8a6 a1b1 a6c7 g1f3 c7e6 e1g1 b7b6 h2h4 b6c5 d4c5 f8e7 c1h6 e7h4 a2a3 h4e7 f1e1 c8a6 b4b5 a6c8 f3h2 e7g5 h6g5 d8g5 h2g4 e8f8 g4f6 f8g7 b1b4 h7h5 d1f3 g5f5 f3f5 g6f5 b4h4');
INSERT INTO evaluations VALUES(55486,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(55486,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(55486,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(55486,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(55486,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(55486,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(55486,36,'cp 63','f8d6 g1f3 g8f6 g2g3 h7h5 d2d4 b8c6 c1g5 d6e7 b1c3 c8g4 d1d2 g4f3 e2f3 d8d4 d2d4 c6d4 e1c1 a8d8 f1g2 e8g8 h1e1 e7b4 e1e5 c7c6 f3f4 b4c3 b2c3 d4b5 d1d8 f8d8 g2f3 d8d7 c1b2 d7d2 g5f6 g7f6 e5e2 d2e2 f3e2 b5d6 e2h5 f6f5');
INSERT INTO evaluations VALUES(9336,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(9336,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9336,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(9336,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(9336,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(9336,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(9336,36,'cp -18','c5e3 f2e3 c6a5 c4b3 a5b3 a2b3 a7a5 e1g1 e8g8 d1d2 c8d7 c3e2 a5a4 b3a4 d7a4 e2g3 a4d7 a1a8 d8a8 b2b3 a8d8 h2h3 h7h6 d2b4 d8c8 g1h2 b7b5 d3d4 c7c5 d4c5 d6c5 b4c3 f8e8 f3e5 c8c7 e5d7 f6d7');
INSERT INTO evaluations VALUES(9442,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(9442,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(9442,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(9442,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(9442,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(9442,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(9442,36,'cp -12','f2e3 c6a5 c4b3 e8g8 e1g1 a5b3 a2b3 c7c6 h2h3 a7a5 d1d2 h7h6 b3b4 c8e6 b4a5 a8a5 a1a5 d8a5 c3d5 a5d2 d5f6 g7f6 f3d2 f8a8 g1f2 g8g7 g2g4 a8a2 f1b1 h6h5 d2f3 g7h6 f2g3 b7b5 f3h4 d6d5 e4d5 c6d5 h4f3 a2a4 d3d4 h5g4 h3g4');
INSERT INTO evaluations VALUES(9441,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(9441,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(9441,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(9441,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(9441,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(9441,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(9441,36,'cp -16','c6a5 c4b3 c7c6 h2h3 a5b3 a2b3 e8g8 e1g1 a7a5 d1d2 h7h6 b3b4 c8e6 b4a5 a8a5 a1a5 d8a5 c3d5 a5d2 d5f6 g7f6 f3d2 f8a8 g1f2 g8g7 d2f3 h6h5 f2g3 b7b5 f1f2 a8a1 g3h4 g7g6 h4g3 a1a2 b2b3 d6d5 f3h4 g6g7 h4f5 e6f5 e4f5 g7h6 g3h4 a2a1');
INSERT INTO evaluations VALUES(9817,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(9817,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(9817,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(9817,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(9817,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(9817,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(9817,36,'cp 151','d4c6 b7c6 e4e5 f6h7 c3e4 d8e7 e1g1 e8g8 c1f4 d7d5 e4c5 d5c4 d1d4 e7h4 c5e4 c8e6 f4e3 h4g4 f2f3 g4g6 f1f2 f7f5 e4c5 e6d5 h2h3 f8e8 g1h2 h7g5 a1d1 g5e6 c5e6 g6e6 f3f4 a7a5 d4c3 e6e7 e3d4 a5a4 c3g3 a4a3 b2a3 e7a3');
INSERT INTO evaluations VALUES(9721,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(9721,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(9721,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(9721,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(9721,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(9721,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(9721,36,'cp 149','b7c6 e4e5 f6h7 c3e4 c5b4 c2c3 d7d5 c3b4 d5c4 d1d8 e8d8 c1d2 h7f8 e1c1 d8e7 d2c3 c8e6 f2f4 a8d8 c3d4 e6f5 d4c5 e7e8 d1d8 e8d8 h1d1 d8c8 e4g3 f5d3 c5a7 g7g6 a7e3 f8e6 b2b3 h8d8 b3c4 d3c4 d1d8 c8d8');
INSERT INTO evaluations VALUES(7568,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(7568,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(7568,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(7568,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(7568,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(7568,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(7568,36,'cp 139','e4e5 f6h7 c3e4 c5b4 c2c3 d7d5 e5d6 b4d6 d1f3 e8g8 e1g1 d6e5 c1f4 e5f4 f3f4 c8e6 c4e6 f7e6 f4e3 h7f6 a1d1 f6d5 e3c5 d8e8 d1d2 a7a5 f1d1 e6e5 h2h3 a5a4 c3c4 d5f4 g1h2 e8g6 f2f3 f8f5 a2a3 a8f8 c5a7 f4e6 a7a4 e6d4');
INSERT INTO evaluations VALUES(14101,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14101,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14101,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14101,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14101,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14101,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14101,36,'cp 0','f1b5 c8g4 g1e2 e7e6 f2f3 g4f5 g2g4 f6g4 f3g4 d8h4 e1d2 f8b4 c2c3 h4h6 d2e1 h6h4');
INSERT INTO evaluations VALUES(14098,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14098,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14098,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14098,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14098,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14098,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14098,36,'cp 0','c8g4 g1e2 e7e6 f2f3 g4f5 g2g4 f6g4 f3g4 d8h4 e1d2 f8b4 c2c3 h4h6 d2e1 h6h4');
INSERT INTO evaluations VALUES(16288,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(16288,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(16288,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(16288,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(16288,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(16288,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(16288,36,'cp 76','e2e4 f8b4 f1d3 d7d5 e1g1 b4c3 e4d5 e6d5 d1e2 d8e7 e2e7 c6e7 b2c3 e8g8 c1g5 e7f5 g5f6 g7f6 f1b1 c7c6 a2a4 b7b6 a4a5 b6b5 b1e1 f5d6 f3h4 c8d7 f2f3 a8e8 e1e8 f8e8 g1f2 e8c8 a1e1 d7e6 g2g4 c6c5 h4g2 c5d4');
INSERT INTO evaluations VALUES(17367,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(17367,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(17367,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(17367,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(17367,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(17367,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(17367,36,'cp -22','d7d5 e4d5 e6d5 c4a2 d8e7 d1e2 g8f6 b1c3 c8e6 c1f4 h7h6 h2h3 g7g5 f4h2 d5d4 a2e6 d4c3 e6a2 e7e2 e1e2 c3b2 a1b1 f8g7 b1b2 b7b5 b2b1 e8g8 h1e1 f8d8 e2f1 f6d5 a2d5 d8d5 c2c4 d5d3 c4b5 a6b5 b1b5 c5c4 b5c5 c6a5 f3e5 a5b3 c5c4');
INSERT INTO evaluations VALUES(17621,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(17621,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(17621,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(17621,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(17621,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(17621,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(17621,36,'cp -37','e4d5 e6d5 c4a2 d8e7 d1e2 g8f6 b1c3 c8e6 c1f4 g7g6 h2h3 d5d4 c3e4 f6e4 e2e4 e6a2 e4e7 f8e7 a1a2 e8d7 a2a1 f7f6 e1g1 a6a5 a1e1 h8e8 g2g4 b7b5 f3d2 a5a4 d2e4 h7h5 g4g5 f6g5 f4g5 e7g5 e4g5 e8e1 f1e1');
INSERT INTO evaluations VALUES(17620,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(17620,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(17620,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(17620,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(17620,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(17620,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(17620,36,'cp -36','e6d5 c4a2 d8e7 d1e2 g8f6 b1c3 c8e6 c1f4 g7g6 e1g1 d5d4 c3e4 f6e4 e2e4 e6a2 e4e7 f8e7 a1a2 f7f6 h2h4 h7h6 a2a1 e8d7 h4h5 g6g5 f4h2 a8e8 g2g4 h8f8 f1e1 f6f5 f3e5 c6e5 e1e5 e7d6 e5e8 f8e8 h2d6 d7d6 g4f5 e8e5 g1g2 a6a5 c2c3 e5f5 c3d4 c5d4');
INSERT INTO evaluations VALUES(17642,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(17642,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(17642,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(17642,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(17642,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(17642,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(17642,36,'cp -39','c4a2 d8e7 d1e2 g8f6 b1c3 c8e6 c1f4 g7g6 e1g1 d5d4 c3e4 f6e4 e2e4 e6a2 e4e7 f8e7 a1a2 e8d7 a2a1 f7f6 h2h4 h8e8 f1b1 b7b5 f3d2 f6f5 g2g3 e7d6 f4g5 d7e6 g1g2 c6e5 g5f4 e6d5 a3a4 c5c4 f4e5 d6e5 a4b5 a6b5 a1a8');
INSERT INTO evaluations VALUES(19472,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(19472,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(19472,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(19472,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(19472,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(19472,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(19472,36,'cp 36','d2d4 c5d4 e2d4 g8f6 b1c3 e7e5 d4b5 d7d6 c1g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 e8g8 a3c2 f6g5 a2a4 b5a4 a1a4 a6a5 f1c4 c6e7 c2e3 c8d7 a4a2 e7d5 e3d5 a5a4 e1g1 a8b8 d1e2 g7g6 f1a1 g8h8 g2g3 g5h6 a2a4 d7a4 a1a4 f7f5 e4f5');
INSERT INTO evaluations VALUES(22589,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(22589,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(22589,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(22589,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(22589,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(22589,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(22589,36,'cp 62','d2d4 f8b4 e2e3 b4c3 b2c3 b7b6 g1f3 c8a6 d1a4 c6a5 f1d3 c7c5 e1g1 g8e7 c1a3 d8c7 f3d2 e8g8 d2b3 a5b3 a2b3 a6b7 b3b4 d7d6 b4c5 d6c5 f1b1 e7g6 a4c2 f7f5 f2f3 a8d8 d3f1 d8d7 d4c5 b6c5 b1b5 b7c6 b5c5');
INSERT INTO evaluations VALUES(44343,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(44343,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(44343,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(44343,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(44343,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(44343,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(44343,36,'cp 146','b1c3 d7d6 e2e4 a7a6 h2h3 h7h5 g1f3 b8d7 f1e2 g7g6 a2a4 f8h6 c1h6 h8h6 d1d2 h6h8 e1g1 f6h7 f3e1 d8g5 d2g5 h7g5 h3h4 g5h7 a4a5 g6g5 e1f3 g5g4 f3e1 h8g8 e1d3 e8e7 b2b4 c5b4 d3b4 d7c5 b4d3');
INSERT INTO evaluations VALUES(44342,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(44342,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(44342,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(44342,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(44342,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(44342,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(44342,36,'cp 131','d7d6 e2e4 h7h5 h2h4 g7g6 g2g3 f8h6 c1h6 h8h6 d1d2 h6h8 g1f3 b8d7 f3g5 d7f8 f2f4 f6g4 a2a3 f7f6 g5f3 c8d7 f1h3 d8b6 e1g1 e8c8 f4e5 d6e5 h3g4 d7g4 f3h2 g4h3 f1f2 f8d7 a1e1 c8b8 h2f1 h3f1 e1f1');
INSERT INTO evaluations VALUES(45940,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(45940,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(45940,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(45940,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(45940,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(45940,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(45940,36,'cp 135','e2e4 h7h5 h2h4 g7g6 g2g3 b8a6 g1h3 d8e7 d1b3 a6c7 f2f3 b7b6 h3f2 c8d7 a2a4 f8h6 c1h6 h8h6 f1h3 h6h8 e1g1 e8f8 b3d1 f8g7 d1e2 a7a6 h3d7 e7d7 b2b3 h8f8 g1g2 f8b8 f2h3 c7e8 e2d2 f6h7 f3f4');
INSERT INTO evaluations VALUES(54357,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(54357,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(54357,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(54357,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(54357,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(54357,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(54357,36,'cp 133','f1b5 c8d7 e4d5 e6d5 e1g1 f8c5 f1e1 g8e7 d4f5 e8g8 f5e7 c5e7 d1d5 e7f6 c2c3 d7c6 d5d8 f8d8 a2a4 c6b5 a4b5 b8d7 g1f1 d7e5 b1a3 a7a6 b5a6 a8a6 c1g5 e5f3 g5f6 f3d2 f1g1 g7f6 h2h3 a6b6 b2b4 g8f8 e1e2 b6a6 a3c2 a6a1 c2a1 f6f5 a1c2 f5f4');
INSERT INTO evaluations VALUES(8970,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(8970,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(8970,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(8970,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(8970,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(8970,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(8970,36,'cp 64','a2a3 b4c3 d2c3 e8g8 g1f3 b7b6 b2b3 f6e4 a1c1 e4c3 c1c3 d5c4 f1c4 d8d6 d1c1 c8b7 e1g1 a8c8 f1d1 c6e7 c4e2 a7a5 c1b2 h7h6 d1c1 f8d8 h2h3 e7d5 c3c2 d5f6 f3d2 c7c5 d4c5 c8c5 c2c5 b6c5 d2c4 d6c7 b2c3');
INSERT INTO evaluations VALUES(8985,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(8985,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(8985,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(8985,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(8985,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(8985,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(8985,36,'cp 55','b4d6 c4c5 d6e7 f2f4 a7a5 g1f3 e8g8 d1c2 c8d7 a1b1 b7b6 b2b4 a5b4 a3b4 b6c5 b4c5 a8b8 b1b8 d8b8 f1e2 c6b4 c2b1 b8a8 e1g1 f8b8 b1b2 h7h6 f1a1 a8b7 h2h3 d7e8 g1h2 b4c6 b2b7 b8b7 f3e5 b7b2 d2e1 c6e5 d4e5');
INSERT INTO evaluations VALUES(16639,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(16639,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(16639,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(16639,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(16639,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(16639,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(16639,36,'cp -9','b1c3 d7d6 d1b3 d8d7 e1g1 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 a1c1 a8c8 e5c4 f8e8 c4a5 b7a8 c3a4 e7f8 a5c6 a7a6 f2f3 f6h5 g3f2 h5f4 c6d4 e8e5 f2g3 a8b7 g3f4 g5f4');
INSERT INTO evaluations VALUES(17902,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(17902,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17902,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(17902,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(17902,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(17902,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(17902,36,'cp 54','e5d4 c3d4 d7d5 e4e5 f7f6 b1c3 c8f5 f1d3 f5d3 d1d3 f6e5 f3e5 f8d6 e1g1 d6e5 d4e5 g8e7 c1g5 d5d4 c3e2 d8d7 f2f4 e8c8 a1d1 h7h6 g5h4 h8e8 a2a3 d7d5 h2h3 d5b5 h4e7 b5d3 d1d3 e8e7 e2g3 c8d7 g3e4 g7g5');
INSERT INTO evaluations VALUES(17921,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(17921,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(17921,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17921,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(17921,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(17921,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(17921,36,'cp 54','c3d4 d7d5 e4e5 c8f5 f1d3 d8d7 d3f5 d7f5 e1g1 e8c8 b1c3 f7f6 f3h4 f5d7 f2f4 g8e7 h4f3 e7f5 a2a3 f6e5 f4e5 f8e7 b2b4 h8f8 b4b5 a6b5 c3b5 f5h4 f3h4 f8f1 d1f1 e7h4 c1e3 d7g4 f1f7 h4e7 h2h3 g4e4 a1e1 e4g6 e1f1 g6f7');
INSERT INTO evaluations VALUES(17919,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(17919,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17919,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(17919,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(17919,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(17919,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(17919,36,'cp 30','d7d5 e4e5 c8f5 f1e2 c6b4 b1a3 g8e7 e1g1 e7c6 c1e3 f8e7 f3e1 f7f6 e5f6 e7f6 a3b1 c6a5 a2a3 b4c6 b1c3 e8g8 a1c1 g8h8 b2b4 a5c4 c3d5 c4e3 d5e3 f5e4 e1f3 c6d4 f3d4 f6d4 c1c4 d4e3 d1d8 a8d8 f2e3 f8f1 e2f1 e4c6 c4f4 h8g8 f1c4 c6d5');
INSERT INTO evaluations VALUES(20107,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(20107,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(20107,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(20107,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(20107,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(20107,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(20107,36,'cp 118','c8d7 b5d7 d8d7 b1c3 h7h6 b2b3 e7e6 c1b2 g8f6 f2f4 a7a6 e1c1 f8e7 c1b1 d7c7 d4d3 a8c8 d3e2 c7a5 d1d3 a5h5 e2d2 b7b5 a2a3 h5c5 g2g4 f6g4 d3g3 h6h5 h2h3 e7h4 g3f3 g4f6 d2g2 e8e7 h1d1 h8g8 e4e5 d6e5');
INSERT INTO evaluations VALUES(26074,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(26074,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(26074,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(26074,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(26074,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(26074,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(26074,36,'cp 99','b5d7 d8d7 b1c3 d7c6 c1g5 h7h6 g5h4 e7e5 d4d3 f8e7 h4e7 g8e7 e1c1 a8d8 d1d2 a7a6 h1d1 e8g8 d3e2 f7f5 f2f3 f5e4 e2e4 c6c7 c3d5 e7d5 d2d5 f8f6 a2a3 b7b5 e4e2 f6e6 e2e3 e6g6 e3d2 c7f7 c1b1 d8d7 b2b3 f7f4 d2f4');
INSERT INTO evaluations VALUES(25897,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(25897,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(25897,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(25897,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(25897,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(25897,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(25897,36,'cp 86','d8d7 b1c3 d7c6 c1g5 h7h6 g5h4 e7e5 d4b4 a7a5 b4b5 f8e7 h4e7 c6b5 c3b5 e8e7 e1c1 a8a6 d1d3 g8f6 f2f3 h8c8 h1d1 c8c6 b2b3 f6d7 c1b2 h6h5 d3d5 g7g6 c2c4 d7c5 g2g3 c5e6 b5c3 e6c5 h2h4 f7f6 d5d2 a6a8 c3b5 a8d8 d2d5 c5e6 b5c3 e6c5 d1d2 c6b6 g3g4 h5g4');
INSERT INTO evaluations VALUES(27581,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(27581,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(27581,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(27581,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(27581,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(27581,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(27581,36,'cp 82','b1c3 d7c6 c1g5 h7h6 g5h4 e7e5 d4b4 a7a5 b4b5 f8e7 h4e7 c6b5 c3b5 e8e7 e1c1 a8a6 f2f3 g8f6 d1d2 h8c8 b2b3 a5a4 c1b2 f6e8 h1d1 g7g6 g2g3 a4b3 a2b3 f7f5 e4f5 g6f5 f3f4 c8a8 b3b4 a6a4 b2b3 a4a1 d2d5 a1d1');
INSERT INTO evaluations VALUES(27580,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(27580,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(27580,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(27580,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(27580,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(27580,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(27580,36,'cp 97','d7c6 c1g5 h7h6 g5h4 e7e5 d4b4 a7a5 b4b5 f8e7 h4e7 c6b5 c3b5 e8e7 e1c1 a8a6 f2f3 g8f6 d1d2 h8c8 h1d1 c8c6 b2b3 h6h5 c1b2 f6d7 b5c3 g7g6 a2a4 e7e6 g2g3 d7c5 c3d5 f7f5 e4f5 g6f5 d5c3 a6b6 c3b5 b6a6 h2h3 a6b6');
INSERT INTO evaluations VALUES(21470,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(21470,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(21470,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(21470,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(21470,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(21470,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(21470,36,'cp 70','d2d4 b7b6 g1f3 c6e7 c1e3 h7h5 f3h4 c7c5 c2c3 e7c6 e3f2 g7g6 h2h3 g8h6 h4f3 h6f5 f1d3 f8e7 e1g1 c8b7 f1e1 d8c8 a2a3 b7a6 d3c2 c8d7 b2b4 c5d4 c3d4 a6c4 b1c3 b6b5 c2f5 g6f5 f2h4 a7a5 h4e7 d7e7');
INSERT INTO evaluations VALUES(21652,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(21652,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(21652,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(21652,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(21652,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(21652,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(21652,36,'cp 100','c8g4 f1e2 g8e7 b1c3 d8d7 h2h3 g4f3 e2f3 c6e5 f3d5 e8c8 d5f3 d7e8 d1e2 e5f3 e2f3 e7c6 e1g1 g7g6 b2b3 e8e5 c1b2 f8d6 g2g3 e5e6 c3e4 d6e5 b2e5 c6e5 f3f6 e6d5 e4g5 e5d7 f6f4 f7f6 g5f3 g6g5 e3e4 g5f4 e4d5 f4g3 f2g3 d7c5 c2c4 h8g8 g1g2 c5e4 g3g4 h7h5');
INSERT INTO evaluations VALUES(26874,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(26874,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(26874,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(26874,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(26874,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(26874,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(26874,36,'cp 119','f1e2 g8e7 b2b3 g4f3 g2f3 c6e5 c1b2 e5c6 d1d3 d8d7 b1c3 e8c8 e1c1 d7h3 a2a3 a7a6 f3f4 c8b8 c3a4 d5d4 e3d4 h3d3 e2d3 c6d4 d3a6 e7c6 a6d3 f8d6 f4f5 h8e8 b3b4 d6f4 c1b1 c6e5 a4c5 e5d3 c5d3 e8e4 d3f4 e4f4');
INSERT INTO evaluations VALUES(22525,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(22525,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(22525,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22525,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22525,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(22525,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(22525,36,'cp -5','g8f6 c3d5 f8c5 c2c3 a7a5 f3e5 f6d5 e4d5 c6e5 d3d4 d8e7 f1e2 c5d6 d4e5 d6e5 e1g1 e8g8 f1e1 d7d6 c1e3 a5a4 e3d4 e5d4 d1d4 f8e8 d4d2 e7d8 e2f3 c8d7 e1e8 d7e8 a1e1 a4a3 b2b3 d8h4 e1e4 h4f6 e4c4');
INSERT INTO evaluations VALUES(25615,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(25615,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(25615,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(25615,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(25615,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(25615,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(25615,36,'cp 60','d4c6 b7c6 b5c4 g7g6 e1g1 g8f6 d1e2 f8g7 h2h3 d8c8 e4e5 d6e5 e2e5 e8g8 f1e1 f6d5 c3d5 c6d5 e5d5 d7c6 d5d3 c8b7 d3b3 b7c7 b3g3 c7g3 f2g3 e7e6 c2c3 f8b8 a2a4 h7h5 a4a5 c6b5 c4b5 b8b5 e1d1 a8c8 a1a4 g8f8 d1d7 f8e8');
INSERT INTO evaluations VALUES(25614,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(25614,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(25614,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(25614,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(25614,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(25614,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(25614,36,'cp 85','b7c6 b5c4 g7g6 d1f3 g8f6 c1g5 f8g7 g5f6 g7f6 e4e5 d6e5 e1c1 e8g8 c3e4 f6g7 e4c5 a8a7 f3e3 d8c8 c4a6 a7a6 c5a6 c8a6 d1d7 a6a2 h1d1 a2a1 c1d2 a1a4 h2h3 c6c5 d2e1 a4c2 d7e7 e5e4 e3e4 c2b2 e1f1 g7d4 e4f3 b2c2 e7d7 c2a4 d7c7 a4b5 f1g1 b5b6 c7b7');
INSERT INTO evaluations VALUES(26199,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26199,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(26199,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26199,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(26199,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(26199,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(26199,36,'cp 41','d1d3 g8f6 e1g1 e7e6 c1f4 f8e7 f1d1 e8g8 c2c4 f8e8 a2a3 d7f8 h2h4 f8g6 f4g5 h7h6 h4h5 g6f8 g5f6 e7f6 g3e4 f6e7 g2g3 f8d7 g1g2 d8a5 c4c5 d7f6 e4f6 e7f6 d3e4 a5d8 b2b4 d8d5 e4d5 e6d5 a3a4');
INSERT INTO evaluations VALUES(26198,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26198,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(26198,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(26198,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(26198,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(26198,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(26198,36,'cp 37','e7e6 c1f4 g8f6 e1g1 f8e7 c2c4 e8g8 h2h3 f8e8 f1d1 d8b6 a2a4 a7a5 d3c2 a8d8 g3f1 c6c5 d4d5 d7f8 f1e3 f8g6 f4g3 e7d6 g3d6 b6d6 d5e6 d6e6 d1d8 e8d8 c2b3 e6c6 a1d1 d8d1 b3d1 c6b6 b2b3 h7h6 e3d5 b6d6');
INSERT INTO evaluations VALUES(26131,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26131,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(26131,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(26131,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(26131,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(26131,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26131,36,'cp 25','c1f4 g8f6 e1g1 f8e7 f1d1 e8g8 c2c4 f8e8 a1b1 a7a5 b2b3 d7f8 g3e4 f8g6 e4f6 e7f6 f4g3 g6e7 a2a4 e7f5 d3e4 f5g3 h2g3 h7h6 g3g4 d8c7 g2g3 b7b6 d1d3 a8d8 b1d1 c6c5 d4c5 c7c5 d3d8 e8d8 d1d8 f6d8');
INSERT INTO evaluations VALUES(30147,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(30147,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(30147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30147,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(30147,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(30147,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(30147,36,'cp 0','e2e3 f8e7 f1d3 f5g6 c3e2 c7c5 e1g1 b8d7 d4c5 d7c5 d3b5 f6d7 g5e7 d8e7 e2f4 e8g8 f4g6 h7g6 c2c4 d5c4 b5c4 a8c8 a1c1 f8d8 d1c2 d7b6 c2b1 c5a4 c4b3 a4c5');
INSERT INTO evaluations VALUES(30146,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(30146,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(30146,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(30146,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(30146,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(30146,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(30146,36,'cp -13','f8e7 f1d3 f5g6 e1g1 b8d7 g5f6 g7f6 c3e2 c7c5 c2c4 d5c4 d3c4 e8g8 e2f4 c5d4 f3d4 d7e5 c4e2 d8b6 f4g6 h7g6 d1b3 e7c5 b3b6 c5b6 d4f3 f8d8 a2a4 e5c6 f1d1 a8c8 g2g3 d8d1 a1d1 c6b4 d1d7 c8c7 d7c7 b6c7 g1g2 f6f5 f3d4');
INSERT INTO evaluations VALUES(31115,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(31115,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(31115,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(31115,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(31115,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(31115,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(31115,36,'cp 79','f3e5 g4f5 g2g4 f5e6 e2e3 b8d7 h2h4 d7e5 d4e5 d8b6 b1d2 b6b2 a1b1 b2a2 b1b7 e6c8 b7b1 e7e6 f1d3 a7a5 c2c3 g8e7 h4h5 a5a4 e1g1 a2a3 d1c2 a3c5 c3c4 c5a7 f1c1 a4a3 d3e2 c8d7 d2b3 h7h6 b3d4 c6c5 d4b5 d7b5 b1b5');
INSERT INTO evaluations VALUES(34765,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(34765,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(34765,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(34765,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(34765,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(34765,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(34765,36,'cp 62','d6e5 b1c3 b8c6 c1e3 f8b4 d1d3 f6d6 e1c1 d6d3 d1d3 b4d6 d3d1 a7a6 f3d2 c8e6 d2c4 d6e7 a2a4 h7h5 c3d5 e8c8 d5e7 g8e7 d1d8 h8d8 c2c3 f7f5 b2b4 f5e4 b4b5 a6b5 a4b5 c6b8 e3g5 d8e8 c4e5 b8d7 e5d7 c8d7 f1e2');
INSERT INTO evaluations VALUES(34884,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(34884,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(34884,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(34884,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34884,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(34884,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(34884,36,'cp 85','b1c3 b8c6 c1e3 f8b4 d1d3 f6d6 e1c1 d6d3 d1d3 b4d6 c3b5 e8e7 d3d1 g8f6 f3d2 c8e6 f2f3 a7a6 b5d6 c7d6 d2c4 e6c4 f1c4 h8c8 c4e2 b7b5 d1d2 c6d8 b2b3 d8e6 c1b2 e6f4 e3f4 e5f4 a2a3 a8b8 h1d1 c8c6 d1a1 f6d7 a3a4 b5b4 g2g3');
INSERT INTO evaluations VALUES(34883,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(34883,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(34883,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(34883,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(34883,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(34883,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(34883,36,'cp 74','b8c6 c1e3 f8d6 c3d5 f6d8 a2a3 g8f6 e3g5 d6e7 g5f6 e7f6 f1c4 e8g8 d1e2 c8e6 c2c3 c6b8 d5e3 e6c4 e2c4 c7c6 e1g1 b8d7 f1d1 d8c8 g2g3 f8d8 b2b4 c8c7 a3a4 d7b6 d1d8 a8d8 c4e2 b6d7 a4a5 f6e7 a5a6');
INSERT INTO evaluations VALUES(37573,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(37573,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(37573,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(37573,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(37573,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(37573,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(37573,36,'cp 86','e2e4 d7d6 b1c3 g8f6 f1e2 e8g8 e1g1 e7e5 c1e3 e5d4 f3d4 c8b7 f2f3 f6d7 d1d2 d7c5 d4c2 c5e6 a1d1 b8d7 e3h6 d8f6 h6g7 f6g7 a2a3 a7a5 b2b4 a5b4 a3b4 e6f4 c2e3 f4e2 c3e2 d7e5 f3f4');
INSERT INTO evaluations VALUES(39042,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(39042,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(39042,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(39042,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(39042,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(39042,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(39042,36,'cp 98','g1f3 b7b6 c1e3 c8b7 b1c3 d7d5 e4e5 h7h5 d1d2 g8e7 e3f2 e7f5 f1d3 g7f8 e1c1 c7c5 d3f5 g6f5 f2h4 d8d7 h4f6 h8g8 h2h3 b8c6 g2g4 c5d4 c3b5 d4d3 d1g1 f5g4 h3g4 h5g4 d2d3 c6b4 d3h7 d7b5 h7g8 g4f3');
INSERT INTO evaluations VALUES(39041,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(39041,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(39041,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(39041,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(39041,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(39041,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(39041,36,'cp 94','b7b6 c1e3 g8e7 f1d3 d7d5 e4e5 d8d7 e1g1 c8a6 d3a6 b8a6 d1e2 a6b8 c2c3 b8c6 e3f2 h7h5 b1a3 g7h6 g2g3 a7a5 h2h3 e7f5 g1g2 f5e7 a3c2 a5a4 b2b4 c6a7 c2a3 d7c6 e2d3 c6d7 b4b5 c7c6 c3c4 c6b5 c4b5');
INSERT INTO evaluations VALUES(41835,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(41835,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(41835,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(41835,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(41835,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(41835,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(41835,36,'cp 86','b1c3 c7c5 d4c5 d6c5 d1d8 e8d8 f3e5 d8e7 e5d3 c5b6 b2b3 h8d8 a2a4 c8d7 c1a3 e7e8 e1e2 a7a6 g2g4 d7c6 h1c1 h7h6 f2f3 f6d5 c3e4 b6e3 c4d5 e6d5 e4f6 g7f6 e2e3 b8d7 a3b2 e8f8 b2d4 a8c8 d3f4 d7e5 f4h5 f8e7 d4c5 e7e6 h5g7 e6d7 g7f5 d8e8 f5h6');
INSERT INTO evaluations VALUES(41833,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(41833,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(41833,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(41833,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(41833,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(41833,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(41833,36,'cp 61','c7c5 d4c5 d6c5 d1d8 e8d8 f3e5 d8e7 e5d3 c5b6 b2b3 h8d8 c1a3 e7e8 e1e2 b8c6 g2g4 b6a5 b3b4 a5c7 g4g5 f6d5 c3e4 c8d7 a3b2 e8f8 a2a3 b7b5 c4b3 a7a5 e4c5 d7c8 h2h4 a5b4 a3b4 a8a1 h1a1 c7d6 b3d5 e6d5 b2c3 d8e8');
INSERT INTO evaluations VALUES(3592,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(3592,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(3592,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3592,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(3592,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3592,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3592,36,'cp 0','c6d4 c3a4 b7b5 f3d4 b5a4 d4c6 d8e8 d1f3 e8c6 h4f6 c6b6 f6e7 b6b2 e7f8 c8e6 f8g7 b2a1 e1e2 a1b2 g7h6 b2c2 e2e1 c2b1 e1d2 b1b2 d2e1');
INSERT INTO evaluations VALUES(9399,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(9399,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(9399,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9399,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9399,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(9399,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(9399,36,'cp 19','c3a4 a7a5 a4c5 d6c5 e1g1 e8g8 c2c3 a5a4 c1e3 d8e7 d1c2 h7h6 h2h3 f8d8 a1d1 c8e6 c2b1 b7b5 f1e1 f6h7 d3d4 c5d4 c3d4 e5d4 f3d4 c6d4 e3d4 c7c5 d4e3 c5c4 a2a3 h7f8 d1d8 a8d8 e1d1 e6d7 b1c2 f8g6 f2f3 d8e8 c2d2 d7c6 e2f1 g6e5');
INSERT INTO evaluations VALUES(18807,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(18807,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18807,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(18807,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(18807,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(18807,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(18807,36,'cp 9','c1b2 a7a6 d1e2 d7d6 d2d4 c5d4 f3d4 c6d4 b2d4 g8f6 e2e3 f8e7 f1e2 e8g8 e1g1 b7b5 f1d1 c8b7 b1d2 f6d7 e3g3 e6e5 d4e3 d8c7 h2h4 g8h8 c2c4 b5b4 h4h5 d7f6 a2a3 b4a3 a1a3 a6a5 h5h6 g7h6 g3h4 f8g8 e3h6 g8g6');
INSERT INTO evaluations VALUES(18934,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(18934,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(18934,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(18934,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(18934,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(18934,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(18934,36,'cp 12','b1c3 g8f6 g1f3 f8e7 e4e5 f6d5 c3d5 e6d5 c1d2 e8g8 e1c1 d7d6 e5d6 d8d6 e3f4 d6f4 d2f4 e7c5 f4g3 f8d8 a2a3 c5b6 f1d3 c8g4 h2h3 g4h5 h1e1 f7f6 d1d2 h5f7 b2b4 a7a6 f3h4 g7g6 h4f3 d8e8 d2e2 e8e2 e1e2 a8d8');
INSERT INTO evaluations VALUES(25273,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(25273,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(25273,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(25273,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(25273,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(25273,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(25273,36,'cp 89','h2h3 g4f3 d1f3 f8e7 c3e2 e8g8 c2c3 d6d5 d2d3 d8d6 e2g3 g7g6 b5a4 f8d8 f1d1 d5e4 d3e4 d6e6 c1g5 d8d1 a4d1 c6d8 d1b3 e6c6 h3h4 h7h5 b3d5 c6b6 b2b4 g8g7 a2a4 c7c6 a4a5 b6a6 g5f6 e7f6');
INSERT INTO evaluations VALUES(25272,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(25272,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(25272,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(25272,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(25272,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(25272,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(25272,36,'cp 76','g4f3 d1f3 f8e7 d2d4 f6d7 b5c6 b7c6 d4e5 d7e5 f3e2 e8g8 f2f4 e5d7 b2b3 e7f6 c1b2 f6d4 g1h2 f8e8 a1d1 c6c5 c3a4 d4b2 a4b2 d8e7 d1e1 a7a5 e2f3 h7h6 f3d3 d7e5 d3d5 e5d7 e4e5 d6e5 f4e5 d7f8 d5e4 f8e6 c2c3 e8d8 b2c4 a5a4');
INSERT INTO evaluations VALUES(28031,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(28031,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(28031,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(28031,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(28031,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(28031,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(28031,36,'cp 90','b1c3 d7d6 f1e2 g8f6 d1c2 e8g8 e1g1 e7e5 d4e5 d6e5 c1g5 b8d7 a1d1 d8c8 c3d5 f8e8 f1e1 a7a5 f3d2 b7a6 h2h3 c7c6 g5f6 g7f6 d5f6 d7f6 c4c5 a6e2 e1e2 b6b5 d2f3 c8c7 d1d6 g8g7 e2d2 a5a4 d2d1 a8a7 d6f6');
INSERT INTO evaluations VALUES(30600,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(30600,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(30600,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(30600,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(30600,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(30600,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(30600,36,'cp 127','f1d3 g7g6 c1e3 f8g7 e1g1 g8e7 d1d2 d7d5 e4e5 b7a6 c3b5 e7f5 d2c3 c7c6 d3f5 a6b5 f5d3 b5d3 c2d3 g7f8 b2b4 a7a5 b4a5 a8a5 a2a4 f8e7 f1b1 g6g5 h2h3 a5a6 c3b2 b8d7 b2d2 c6c5 d4c5 b6c5 d3d4 e8g8 d4c5');
INSERT INTO evaluations VALUES(30601,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(30601,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(30601,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(30601,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(30601,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(30601,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(30601,36,'cp 118','g7g6 c1e3 g8f6 e1g1 d7d6 d1d2 f8g7 a1d1 b8d7 f1e1 a7a6 h2h3 b6b5 a2a3 g6g5 e4e5 f6d5 c3d5 b7d5 d2e2 e8g8 b2b3 d8e7 e5d6 c7d6 c2c4 b5c4 b3c4 d5c6 d4d5 e6d5 c4d5 c6b5 d3b5 a6b5 e2b5 f8b8 b5c4 e7f6 e3d4 f6g6');
INSERT INTO evaluations VALUES(31952,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(31952,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(31952,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(31952,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(31952,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(31952,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(31952,36,'cp 32','h2h3 g4h5 e5e6 d8d5 e6f7 h5f7 f1e2 b8c6 e1g1 e7e5 f3g5 f7g6 d2d3 g8f6 f2f4 e5f4 c1f4 e8c8 e2f3 d5d7 b1d2 d8e8 d2c4 h7h6 f3c6 d7c6 g5f3 f6d5 f3e5 c6e6 d1f3 d5f4 f3f4 f8c5 f4g3 g6h5 a1e1 h8f8 f1f8 c5f8 e5f3 e6d5 e1e8 h5e8 g3g4 d5d7');
INSERT INTO evaluations VALUES(31951,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(31951,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(31951,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(31951,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(31951,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(31951,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31951,36,'cp 38','g4h5 f1e2 e7e6 f3d4 d8d4 e2h5 d4e5 d1e2 b8c6 c2c3 g8f6 h5f3 e8c8 d2d4 e5e2 e1e2 f8d6 h1e1 h8e8 e2f1 e6e5 f3c6 b7c6 d4e5 e8e5 e1e5 d6e5 b1d2 f6d5 d2f3 e5d6 c1d2 f7f6 a1e1 c6c5 g2g3 h7h5 e1e4 g7g5 e4e6 c8b7 f1g2 g5g4 h3g4 h5g4');
INSERT INTO evaluations VALUES(32767,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(32767,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(32767,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32767,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32767,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32767,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(32767,36,'cp 19','d1d4 b8c6 d4d3 g7g6 b1c3 f8g7 e1g1 g8f6 c3d5 e8g8 c1g5 f6d5 e4d5 c6e5 f3e5 g7e5 a1e1 f8c8 e1e2 e5b2 e2e7 d7f5 d3f5 g6f5 e7b7 c8c2 a2a4 g8g7 b7d7 b2e5 g5e7 c2a2 e7d6 a8e8 d6e5 e8e5 d7a7 e5d5 f1e1 d5c5');
INSERT INTO evaluations VALUES(32766,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32766,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(32766,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(32766,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(32766,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(32766,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(32766,36,'cp 15','g8f6 d4d3 g7g6 e1g1 f8g7 c2c4 d7c6 b1c3 b8d7 c1e3 f6g4 e3d4 d7e5 d4e5 g4e5 f3e5 g7e5 a1c1 a8c8 b2b3 a7a6 f1d1 e7e6 c3e2 c6c5 g2g3 e8g8 g1g2 f8d8 e2g1 e5g7 g1f3 c5a5 c1c2 b7b5 c4b5 c8c2 d3c2 a5b5 f3d4 g7d4 d1d4');
INSERT INTO evaluations VALUES(33984,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(33984,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(33984,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(33984,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(33984,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(33984,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(33984,36,'cp 88','e5f3 f8e7 d2d3 e4d6 e1g1 d6c4 d3c4 e8g8 b1c3 f6g6 c3d5 e7d8 f3e5 g6f5 f1e1 d7d6 g2g4 f5e6 e5c6 e6d7 c6e7 d8e7 e1e7 d7g4 d1g4 c8g4 e7c7 b8c6 c7b7 g4f3 b2b3 f7f6 c1a3 f8f7 b7f7 g8f7 a3d6 a8d8 d6g3 f3d5 c4d5 d8d5 c2c4 d5d2 g1g2 d2c2');
INSERT INTO evaluations VALUES(33986,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(33986,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(33986,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(33986,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(33986,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(33986,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(33986,36,'cp 75','f8e7 d2d3 e4d6 c4b3 b7b6 d1e2 c8b7 b1c3 b7f3 g2f3 f6h4 c3d5 e8d8 c1e3 h8e8 e1c1 b8c6 c2c3 e7g5 d1g1 g5h6 e2d2 c6e5 g1g7 e5f3 e3h6 f3d2 h6g5 h4g5 g7g5 d2b3 a2b3 a7a5 h1g1 c7c6 d5b6 a8b8 b6c4 d6c4 b3c4 e8e2 g5g8 d8c7 g8b8 c7b8 g1g7 e2f2 g7h7 b8c7 h2h4 f2h2 h4h5 c7d6 h5h6 f7f5 c1d1 f5f4 h7h8');
INSERT INTO evaluations VALUES(43489,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43489,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43489,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43489,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43489,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(43489,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(43489,36,'cp 3','d2d4 c5d4 f3d4 b7b5 c4d3 c8b7 d1e2 b8d7 b2b4 d7e5 a2a4 b5a4 a1a4 e5d3 c2d3 f8e7 c1d2 e8g8 a4a1 f6d7 d4b3 a8c8 h2h3 f8e8 f1b1 d6d5 b3c5 d7c5 b4c5 b7a8 d3d4 d5e4 d2e3 e6e5 b1d1 e5d4 d1d4 d8c7 d4c4 a8b7 e2g4');
INSERT INTO evaluations VALUES(43500,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43500,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43500,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43500,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43500,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43500,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43500,36,'cp 0','b7b5 c4d3 c5d4 f3d4 c8b7 d1e2 b8d7 b2b4 d7e5 a2a4 b5a4 a1a4 e5d3 c2d3 f8e7 a4a1 e8g8 c1d2 d8d7 c3a4 d7d8');
INSERT INTO evaluations VALUES(5320,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5320,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(5320,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5320,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5320,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(5320,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(5320,36,'cp -6','g2g3 d7d6 g1e2 a8b8 f1g2 a7a6 a2a4 c8g4 f2f3 g4d7 a1b1 c6d4 c1e3 b7b5 a4b5 a6b5 c4b5 d4b5 c3b5 d7b5 e1g1 d8a5 f3f4 g8f6 e4e5 f6d7 e5d6 e7d6 b2b4 a5a7 b4c5 e8g8 e2d4 a7c5 d4b5 c5e3 g1h1 d7c5');
INSERT INTO evaluations VALUES(11582,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(11582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11582,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(11582,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11582,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11582,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11582,36,'cp 0','e7c6 b1c3 d7d6 c3d5 f8e7 d2d4 e8g8 c2c3 f8e8 d4e5 d6e5 d1d3 e7d6 f1d1 c8e6 h2h3 c6e7 f3g5 e6d5 e4d5 e7g6 d3f5 f7f6 g5h7 g6e7 h7f6 g7f6 f5g4 g8f7 g4h5 f7g8');
INSERT INTO evaluations VALUES(11562,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11562,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11562,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11562,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11562,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11562,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(11562,36,'cp 6','b1c3 f8e7 d2d4 d7d6 c3d5 e8g8 c2c3 f8e8 d4e5 d6e5 d1d3 d8d6 c1e3 e7d8 f3d2 d6d7 a1d1 c6e7 f2f3 f7f6 c3c4 c7c6 d5e7 d8e7 d3c2 d7e6 d2b3 e6f7 e3b6 e7b4 a2a3 b4f8 b3a5 c8e6 b2b3 g8h8 c4c5 h7h6 h2h3');
INSERT INTO evaluations VALUES(11557,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11557,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11557,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11557,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11557,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11557,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11557,36,'cp 0','d7d6 c3d5 f8e7 d2d4 e8g8 c2c3 f8e8 d4e5 d6e5 d1d3 d8d6 f1d1 e7d8 c1g5 f7f6 g5e3 d6e6 e3c5 f6f5 f3g5 e6g6 d3c4 g8h8 d5c7 d8c7 g5f7 h8g8 f7h6 g8h8');
INSERT INTO evaluations VALUES(16342,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(16342,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(16342,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(16342,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(16342,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(16342,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(16342,36,'cp 94','d8e7 b1c3 c6e5 d4e5 e7e5 f2f4 e5e6 d1d4 d7d5 e4e5 c7c5 f1b5 c8d7 b5d7 e6d7 d4d3 d5d4 e5f6 d4c3 d3c3 e8c8 c1e3 g7f6 e3c5 d7c6 c5e3 c6c3 b2c3 d8d5 e1e2 h8g8 g2g3 b7b6 e2f3 f6f5 h1d1 d5a5 d1d3 f8e7 e3d4 g8e8 d3e3 c8d7 e3e5 a5e5 d4e5 d7c6 a2a4 e7c5');
INSERT INTO evaluations VALUES(1083,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(1083,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(1083,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(1083,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(1083,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(1083,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(1083,36,'cp 101','b1c3 c6e5 d4e5 e7e5 f2f4 e5e6 d1d4 d7d5 e4e5 c7c5 f1b5 c8d7 b5d7 e6d7 d4d3 d5d4 e5f6 d4c3 d3c3 e8c8 c1e3 g7f6 e3c5 d7c6 c5e3 c6c3 b2c3 d8d6 g2g3 d6c6 e3d4 f8c5 e1c1 c5d4 c3d4 c8d7 h1e1 h8c8 e1e2 c6a6 c1b2 a6b6 b2a1 b6c6 d1d2 d7d6 a1b2 c6c3');
INSERT INTO evaluations VALUES(20639,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(20639,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(20639,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20639,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(20639,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20639,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(20639,36,'cp 16','g8f6 c1e3 f6g4 d1f3 g4e3 f3e3 f8e7 f1c4 e8g8 e1g1 c8e6 c3d5 e7g5 e3e2 a8c8 c2c3 d8d7 h2h3 c6e7 d5e7 g5e7 c4e6 f7e6 b3c1 d7c6 c1d3 e7d8 a1d1 f8f7 b2b3 d8b6 c3c4 b6d4 g1h1 c8f8 f2f3');
INSERT INTO evaluations VALUES(22355,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(22355,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(22355,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(22355,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(22355,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(22355,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(22355,36,'cp 138','d4e5 d8e7 e5d6 e7e4 c4e2 f8d6 b1c3 e4g6 e1g1 g8e7 e2d3 g6h5 c3e4 d6e5 f1e1 e8g8 h2h3 e7g6 e4g3 e5g3 f2g3 h5a5 a2a3 d7e5 f3e5 g6e5 b2b4 a5b6 c1e3 b6c6 d3f1 c8f5 d1d4 f8e8 e3f2 c6f6 d4c3 e5c6 c3f6 g7f6 e1e8 a8e8 b4b5');
INSERT INTO evaluations VALUES(22353,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(22353,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(22353,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(22353,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(22353,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(22353,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(22353,36,'cp 140','d8e7 e5d6 e7e4 c4e2 f8d6 b1c3 e4g6 e1g1 g8e7 e2d3 g6h5 c3e4 d7e5 d3b5 e7c6 f3e5 h5e5 e4d6 e5d6 d1d6 c7d6 f1d1 c8f5 c2c3 e8c8 c1e3 f5e6 b2b4 a7a6 b5f1 c6e5 e3d4 f7f6 d4b6 d8d7 d1d4 d6d5 a2a4 h8e8 f2f3 e5c4 b6c5 b7b6 f1c4 d5c4 c5b6');
INSERT INTO evaluations VALUES(23348,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(23348,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(23348,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(23348,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(23348,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(23348,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(23348,36,'cp 56','g1e2 f8g7 a2a3 f6f5 b1c3 d7d6 d2d3 d8d7 h2h4 g8e7 c4a2 h7h6 c1d2 h8f8 f3g3 f5f4 g3h2 b7b6 e1c1 c8b7 f2f3 e8c8 g2g3 f4g3 h2g3 f8f6 c1b1 d8f8 d1f1 c8b8 g3g2 b7c8 a2b3 d7e8 g2g3 e8d7 b3c4 d7e8');
INSERT INTO evaluations VALUES(29436,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(29436,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(29436,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(29436,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(29436,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(29436,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(29436,36,'cp 71','g1f3 b8a6 g2g3 g7g6 f1g2 c7c6 e1g1 f8g7 f3e1 h7h5 e2e4 f5e6 d4d5 e6g4 d1c2 c6d5 c4d5 e8g8 e1d3 g4d7 c2e2 d8c8 c1g5 a6c5 a1c1 c5d3 e2d3 a7a6 a2a4 c8c5 g5e3 c5a5 h2h3 a8c8 b2b3');
INSERT INTO evaluations VALUES(29435,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(29435,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(29435,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(29435,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(29435,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(29435,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(29435,36,'cp 72','b8a6 f3h4 f5c8 g2g3 g7g6 h4f3 f8g7 f1g2 e8g8 e1g1 c8f5 f3e1 c7c5 d4d5 f5d7 e1f3 a6c7 e2e4 b7b5 e4e5 f6e8 c4b5 c7b5 c3b5 d7b5 f1e1 e8c7 e5d6 e7d6 c1g5 g7f6 h2h4 f8e8 d1d2 f6g5 h4g5 b5d7 b2b3 e8e1 a1e1');
INSERT INTO evaluations VALUES(50347,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(50347,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(50347,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(50347,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(50347,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(50347,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(50347,36,'cp -35','e7e5 b1c3 f8c5 g1f3 d7d6 e1g1 a7a5 d3c4 a5a4 d2d3 c7c6 a2a3 e8g8 h2h3 c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 b8d7 d1d2 h7h6 d3d4 d8c7 d4e5 d7e5 f3e5 d6e5 a1d1 f8d8 d2e2 b7b5 d1d8 a8d8 f1d1 d8d1 e2d1 g8f7 d1e2');
INSERT INTO evaluations VALUES(52232,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(52232,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(52232,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(52232,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(52232,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(52232,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(52232,36,'cp 212','b1c3 g7g6 h2h4 g8h6 f1c4 b8a6 c4b3 d7d6 g1e2 e7e6 c1e3 c8d7 f2f3 d6d5 d1d2 h6f7 e1c1 d8a5 g2g4 e8c8 c1b1 c8b8 g4g5 f8e7 e2c1 b8a8 c1d3 f6f5 h4h5 g6h5 h1h5 d7e8 d1h1');
INSERT INTO evaluations VALUES(52230,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(52230,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(52230,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(52230,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(52230,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(52230,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(52230,36,'cp 200','g7g6 c1e3 d7d5 d1d2 d8a5 a2a3 d5e4 b2b4 a5d8 c3e4 g8h6 g1f3 h6f7 f1c4 b8a6 c4a6 b7a6 d2e2 f8g7 e1g1 d8d5 f3d2 e8g8 f1e1 f6f5 e4c5 g6g5 f2f3 f5f4 e3f2 e7e5 d4e5 f7e5 h2h3 c8f5 d2e4 d5c4 e4g5 c4e2 e1e2 e5g6');
INSERT INTO evaluations VALUES(52413,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(52413,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(52413,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(52413,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(52413,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(52413,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(52413,36,'cp -65','d7d5 f1b5 b8c6 f2f4 g7g6 h5e2 g8h6 g1f3 f8g7 b5c6 b7c6 d2d3 e8g8 e1g1 a7a5 a2a4 d8b6 f1e1 h6f5 b1d2 f5d6 a1a2 c5c4 d3c4 d6c4 d2c4 d5c4 e2f2 c4c3 e3e4 b6f2 g1f2 c8e6 b2b3 f8c8 c1a3 c6c5 a2a1 c8c7 h2h3 a8b8 a1b1 h7h5 f3g5 g7d4 f2g3 e6d7 g5f3 d4g7');
INSERT INTO evaluations VALUES(682,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(682,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(682,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(682,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(682,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(682,36,'cp 0','b5c6 d5c6 d1f3 c6c2 f3a8 c2c1 e1e2 c1c4 d2d3 c4g4 g5f3 f8d6 a8c6 c8d7 c6a8');
INSERT INTO evaluations VALUES(680,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(680,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(680,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(680,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(680,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(680,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(680,36,'cp 0','d5c6 e1g1 h7h6 d1f3 c6b7 f3b7 c8b7 g5f3 e5e4 f3e5 f8c5 d2d3 c5d4 e5c4 e4d3 c2c3 d4c5 f1e1 f6e4 c1e3 e8c8 b1d2 e4d2 c4d2 c5d6 e3a7 f7f5 d2c4 d6f4 a7d4 d3d2 e1d1 h8e8 c4e3 f4g5 e3f5 e8e2 f5g3 e2e6 g3f1 e6e2');
INSERT INTO evaluations VALUES(686,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(686,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(686,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(686,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(686,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(686,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(686,36,'cp 0','e1g1 h7h6 d1f3 c6b7 f3b7 c8b7 g5f3 e5e4 f3e5 f8c5 d2d3 c5d4 e5c4 e4d3 c2d3 e8c8 b1a3 b7a6 a3c2 h8e8 c1e3 d4e3 c2e3 f6d5 e3d5 d8d5 c4e3 d5d3 f1c1 f7f5 g2g3 a6b7 c1c5 g7g6 a1c1 e8e7 h2h4 d3d2 e3c4 d2d5 c4e3');
INSERT INTO evaluations VALUES(685,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(685,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(685,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(685,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(685,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(685,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(685,36,'cp 0','h7h6 d1f3 c6b7 f3b7 c8b7 g5f3 e5e4 f3e5 f8c5 d2d3 c5d4 e5c4 e4d3 c2d3 e8c8 b1a3 b7a6 a3c2 d4b6 c1e3 d8d3 c4b6 a7b6 f1c1 c8b7 c2b4 d3d7 a2a4 h8e8 b4a6 b7a6 a4a5 a6b7 g1f1 f6d5 e3d4 f7f6 a5b6 c7b6 c1c4 d5c7 d4e3 c7d5');
INSERT INTO evaluations VALUES(749,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(749,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(749,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(749,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(749,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(749,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(749,36,'cp 24','f3d2 f8b4 d1c2 d5c4 g5f6 d7f6 d2c4 a5c7 f1e2 b4e7 e1g1 e8g8 a1c1 c6c5 c3b5 c7b8 d4c5 e7c5 c4d2 c5b6 b5c7 b8c7 c2c7 b6c7 c1c7 f6d5 c7c1 c8d7 d2b3 f8c8 e2f3 g8f8 f1d1 c8c1 d1c1 f8e7 f3d5 e6d5 f2f3 a8c8 c1c8 d7c8 b3d4 c8d7');
INSERT INTO evaluations VALUES(748,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(748,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(748,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(748,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(748,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(748,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(748,36,'cp 19','d5c4 g5f6 d7f6 d2c4 a5c7 a1c1 f8d6 c4d6 c7d6 f1c4 e6e5 d4d5 d6b4 d1b3 b4b3 c4b3 c8d7 d5c6 d7c6 e1e2 e8e7 c1c2 h8d8 f2f3 a8c8 h1c1 c6d7 e3e4 b7b5 a2a3 a7a5 e2e3 c8c6 c3e2 d8c8 c2c6 c8c6 c1c6 d7c6');
INSERT INTO evaluations VALUES(4026,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(4026,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4026,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4026,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(4026,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4026,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(4026,36,'cp 23','f1e2 e7e5 d4e5 d6e5 f3e5 f6h5 e2h5 d7e5 h5e2 c8e6 f4e5 g7e5 e1g1 d8e7 d1c2 c7c6 f1d1 a8d8 h2h3 e6f5 c2b3 f5e6 b3a4 a7a6 a1c1 d8d1 c1d1 f8d8 a4c2 d8d1 c3d1 e5c7 c2d2 e7h4 f2f4 h4e7 d1f2 a6a5 b2b3 h7h5');
INSERT INTO evaluations VALUES(4025,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(4025,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4025,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4025,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(4025,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(4025,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4025,36,'cp 15','e7e5 d4e5 d6e5 f3e5 f6h5 e2h5 d7e5 h5e2 c8e6 d1c2 e5c4 e1g1 d8d2 c2d2 c4d2 f1d1 d2c4 e2c4 e6c4 f4c7 f8c8 c7d6 c4e6 h2h3 g7c3 b2c3 c8c3 d6e5 c3c5 e5d4 c5a5 a2a3 a8c8 d1d2 c8d8 f2f3 a5a4 d2b2 b7b6 g1f2 d8e8 d4f6');
INSERT INTO evaluations VALUES(14296,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(14296,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14296,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14296,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14296,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(14296,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(14296,36,'cp 57','e7e6 a2a3 f6e4 c2c4 a7a5 c4d5 e6d5 d2e4 d5e4 f3d2 f7f5 f1c4 c8d7 d2b1 f8d6 b1c3 d8f6 c1d2 a5a4 a1c1 c6a5 c4a2 a5b3 a2b3 a4b3 d1b3 f6f7 b3f7 e8f7 f2f3 e4f3 g2f3 c7c5 d4c5 d6c5 c3d5 h8c8 e1f2 c5a7 d2c3 d7c6 d5b4');
INSERT INTO evaluations VALUES(14858,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14858,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(14858,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(14858,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14858,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(14858,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(14858,36,'cp 61','a2a3 f6e4 c2c4 a7a5 d1c2 f7f5 f1e2 a5a4 e1g1 c8d7 b2b3 a4b3 d2b3 f8d6 c4c5 d6e7 f3e1 e7f6 c1b2 e8g8 e1d3 c6e7 f2f3 e4g5 b2c3 d7a4 f1b1 g5f7 c2d2 f8e8 b3c1 b7b6 e2d1 a4d1 d2d1 b6c5 d3c5 e6e5 c1d3 e5e4');
INSERT INTO evaluations VALUES(19251,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19251,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(19251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19251,36,'cp -10','g2g3 a7a6 a2a4 f8g7 f1g2 d7d6 b1c3 e7e6 g1e2 g8e7 e1g1 e8g8 a1b1 c8d7 h2h3 d8b6 c1f4 c6d4 g1h2 d4e2 c3e2 f8b8 b2b3 b6c7 b3b4 c5b4 b1b4 e7c6 b4b1 c6d4 e2d4 g7d4 b1c1 d7c6 f4e3 d4e3 f2e3 b7b5 c4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(36573,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(36573,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(36573,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(36573,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(36573,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36573,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36573,36,'cp -21','c7c5 d4c5 d6c5 d1d8 f8d8 e4e5 f6d7 f2f4 b8c6 g1e2 f7f6 c1e3 f6e5 e1c1 e5f4 g3f4 e7e5 h1e1 d7b6 g2c6 d8d1 e1d1 b7c6 d1d8 g8f7 c3e4 c8g4 e4g5 f7e7 d8a8 b6a8 c1d2 h7h6 g5e4 a8b6 b2b3 b6d7 f4e5 g7e5 e4c5 d7c5 e3c5 e7f7 e2d4 a7a6');
INSERT INTO evaluations VALUES(48089,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(48089,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(48089,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(48089,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(48089,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(48089,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(48089,36,'cp 34','g1f3 c7c5 e3e4 d5f6 b1c3 c5d4 f3d4 f8c5 c1e3 b8d7 f1e2 e8g8 e1g1 e6e5 d4f5 c5e3 f5e3 d7c5 d1d8 f8d8 f1d1 c8e6 f2f3 d8d4 a1c1 a8d8 g1f2 a7a6 b2b4 c5d7 c3d5 g8f8 c1c7 b7b5 c7b7 e6d5 e3d5 f6d5 d1d4 e5d4');
INSERT INTO evaluations VALUES(48090,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(48090,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(48090,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(48090,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(48090,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(48090,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(48090,36,'cp 25','c7c5 e3e4 d5f6 b1c3 c5d4 d1d4 c8d7 c1f4 b8c6 d4d3 f8b4 a1c1 e8g8 a2a3 b4c3 c1c3 f8e8 f4g5 e6e5 g5f6 g7f6 f1e2 d7e6 d3e3 d8b6 e3b6 a7b6 e2b5 a8c8 e1g1 e8d8 h2h3 c6d4 f3d4 e5d4 c3c8 d8c8');
INSERT INTO evaluations VALUES(46068,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(46068,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46068,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46068,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(46068,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46068,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(46068,36,'cp 42','e3e4 d5f6 b1c3 c5d4 f3d4 f8e7 f1c4 e6e5 d4f3 d8d1 e1d1 b8c6 d1e2 e8g8 h1d1 c8g4 c1e3 e7b4 c3d5 f6e4 d5b4 c6b4 h2h3 g4f3 e2f3 e4f6 a1c1 f8c8 g2g4 h7h6 a2a3 b4c6 c4b5 a8b8 b2b4 c6d4 e3d4 c8c1 d1c1 e5d4 c1c7 f6d5 c7d7');
INSERT INTO evaluations VALUES(49441,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(49441,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(49441,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(49441,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(49441,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(49441,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(49441,36,'cp 61','c7c5 d4c5 d8a5 c1d2 a5c5 b1c3 d7d5 e4d5 e6d5 g1f3 f8e7 d2e3 c5a5 e1g1 e8g8 h2h3 b8c6 f1e1 a7a6 a2a3 f8d8 d1d2 c8e6 e3g5 a5b6 b2b4 b6c7 a1d1 d8e8 c3e2 a8d8 d2e3 h7h6 g5h4 e6d7 e2d4 c6d4');
INSERT INTO evaluations VALUES(57675,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(57675,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(57675,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(57675,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(57675,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(57675,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(57675,36,'cp -24','d7d5 d2d3 g8f6 g2g3 f8g7 f1g2 b8c6 e1g1 e8g8 c2c3 f8e8 f3e5 c6e5 f4e5 f6g4 d3d4 f7f6 d1b3 f6e5 b3d5 d8d5 g2d5 g8h8 d4e5 g4e5 b1d2 c7c6 d5g2 a7a5 a2a4 c8g4 f1f2 g4e6 d2f3 e5c4 f3d4 e6g4 e2e4 c6c5 d4f3 h8g8');
INSERT INTO evaluations VALUES(54872,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(54872,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(54872,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(54872,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(54872,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54872,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(54872,36,'cp -26','g2g3 g8f6 f1g2 f8g7 e1g1 e8g8 f3e5 c7c6 d2d4 c8e6 a2a4 d8c8 c2c3 e6h3 c1e3 h3g2 g1g2 f6e8 b1d2 e8d6 e3f2 b8d7 d1c2 d7f6 c3c4 c8e6 c4c5 d6e4 d2f3 f8e8 a4a5 e4f2 f1f2 f6e4 f2f1');
INSERT INTO evaluations VALUES(887,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(887,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(887,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(887,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(887,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(887,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(887,36,'cp 106','c1d2 a7a6 f1c4 c6d4 f3d4 a5c5 c4f7 e8f7 d4f3 e7e5 d1e2 f8d6 f3g5 f7g8 g5e4 f6e4 e2e4 a8b8 h2h4 c8d7 h4h5 c5c6 e4d3 h7h6 e1c1 b8f8 f2f3 d7f5 d3e2 b7b5 a2a3 f5h7 c1b1 f8f6 d2c1 g8f7 c3d5 f6e6 g2g4 h8f8 g4g5');
INSERT INTO evaluations VALUES(888,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(888,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(888,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(888,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(888,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(888,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(888,36,'cp 93','a7a6 f1c4 c6d4 f3d4 a5c5 c4f7 e8f7 d4f3 e7e5 d1e2 f8d6 f3g5 f7g8 e1c1 c5c6 g5e4 f6e4 e2e4 c8d7 h1e1 h7h5 e4d3 a8f8 f2f3 h5h4 c3e4 c6b5 d3c3 h8h5 h2h3 d7e8 a2a3 b5c6 c3a5 e8g6 d2b4 d6b4 a5b4 f8f4 b4e7 g6e4 e1e4 f4e4 f3e4 g8h7 d1d7 c6g6 c1b1 b7b5 d7c7 g6g2');
INSERT INTO evaluations VALUES(9558,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(9558,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(9558,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(9558,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(9558,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(9558,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(9558,36,'cp 25','e2e3 c7c5 c4d5 e6d5 d1c2 d8c7 g5f6 d7f6 f1b5 e8f8 d4c5 c7c5 f3d4 b4c3 b2c3 g7g6 e1g1 a7a6 b5d3 c8d7 c2b2 f8g7 a2a4 h8c8 f1c1 c8c7 c3c4 a8c8 d4e2 d5c4 c1c4 c5a5 c4b4 a5d5 b4d4 d5a5');
INSERT INTO evaluations VALUES(9557,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(9557,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9557,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(9557,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9557,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(9557,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(9557,36,'cp 20','c7c5 c4d5 e6d5 d1c2 d8c7 g5h4 e8g8 f1e2 f6e4 e1g1 b4c3 b2c3 d7b6 c3c4 c5d4 f3d4 b6c4 f1c1 f8e8 e2c4 d5c4 c2c4 c7c4 c1c4 b7b6 c4c7 e4c5 d4b5 f7f6 h4g3 c8d7 b5a7 e8d8 a1c1 c5d3 c1c3 d3c5 g3d6 c5e6 a7c6 e6c7 c6d8 a8d8 c3c7');
INSERT INTO evaluations VALUES(11476,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(11476,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(11476,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(11476,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(11476,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11476,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(11476,36,'cp 27','e1g1 d7d6 b1d2 f8e7 c4b3 e8g8 f1e1 c6a5 b3c2 c7c5 d3d4 e5d4 c3d4 c5d4 f3d4 f8e8 d2f1 c8g4 d1d3 a8c8 h2h3 g4h5 f1g3 h5g6 d4f5 a5c4 b2b3 c4e5 d3d1 g6f5 g3f5 e7f8 c1b2 g7g6 f5e3 f8g7 a2a4 e5c6');
INSERT INTO evaluations VALUES(11475,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(11475,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(11475,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11475,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(11475,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11475,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(11475,36,'cp 22','d7d6 b1d2 f8e7 c4b3 e8g8 f1e1 c6a5 b3c2 c7c5 d3d4 c5d4 c3d4 e5d4 f3d4 f8e8 d2f1 e7f8 f1g3 c8d7 b2b3 a5c6 d4c6 d7c6 c1b2 a8c8 d1f3 d6d5 b2f6 d8f6 f3f6 g7f6 g3h5 f8g7 e4d5 c6d5 c2f5 d5e6 h5g3 b7b5 h2h3 c8d8 a1d1 d8d1 e1d1 e6f5 g3f5 e8e2 g2g4 g7f8 d1d7');
INSERT INTO evaluations VALUES(11345,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(11345,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(11345,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11345,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(11345,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(11345,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(11345,36,'cp 37','b1d2 f8e7 a2a4 a6a5 f1e1 e8g8 b2b3 f6d7 c1a3 d7b6 d3d4 e5d4 f3d4 b6c4 d2c4 c6d4 c3d4 f8e8 d1d3 e7f6 a3b2 c8d7 d3f3 b7b5 a4b5 d7b5 e4e5 f6e7 c4a5 d6e5 d4e5 e7b4 b2c3 a8a5 a1a5 b4a5 c3a5');
INSERT INTO evaluations VALUES(14787,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(14787,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(14787,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(14787,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(14787,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(14787,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(14787,36,'cp 44','f8d6 c4c5 d6e7 g1f3 a7a5 f1d3 e8g8 e1g1 b7b6 c5b6 c7b6 d1e2 e7d6 c1d2 c8b7 f1d1 c6a7 d2e1 f6e4 d1c1 a8c8 c1c2 e4c3 e1c3 d8e7 a1c1 b7c6 c3d2 c6b7 d2e1 b7c6 f3e5 c6b7 c2c8 f8c8 c1c8 a7c8 e2c2 h7h6 c2a4 c8a7');
INSERT INTO evaluations VALUES(9228,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(9228,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(9228,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(9228,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(9228,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(9228,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(9228,36,'cp 58','c4c5 d6e7 g1f3 a7a5 a1b1 e8g8 b2b4 a5b4 a3b4 f6e4 d1c2 e4c3 c2c3 e6e5 d4e5 f7f6 e5e6 c8e6 f1d3 d8d7 e1g1 e6f5 f1d1 f5d3 c3d3 a8a4 c1d2 c6e5 d3b3 e5f3 g2f3 f8a8 e3e4 a4a3 b3d5 d7d5 e4d5 a3f3 d2e3 f3f5 b1b3 g8f7 g1g2 a8a4 d5d6 c7d6');
INSERT INTO evaluations VALUES(22650,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(22650,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(22650,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(22650,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(22650,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(22650,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(22650,36,'cp -67','f8c5 c2c3 d4c3 b1c3 g8e7 e1g1 d7d6 c1g5 f7f6 g5d2 e8g8 c3a4 c8g4 h2h3 g4f3 d1f3 g8h8 a4c5 d6c5 f3e3 c6d4 b5d3 e7c6 d2c3 b7b6 f2f4 d8e7 a1e1 a8d8 e3f2 a7a5 f2h4 b6b5 d3b1 a5a4 e4e5 f6f5 h4e7 c6e7 g1h2 b5b4');
INSERT INTO evaluations VALUES(6397,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(6397,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(6397,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(6397,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(6397,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(6397,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(6397,36,'cp -53','c2c3 d4c3 b1c3 g8e7 e1g1 d7d6 c1g5 f7f6 g5d2 e8g8 c3a4 c5d4 a1c1 d4e5 f3e5 f6e5 b5c4 g8h8 a4c3 e7g6 c3e2 c8d7 d1b3 a7a5 a2a4 c6b4 d2b4 a5b4 b3b4 d7a4 b4b7 a8b8 b7a6 d8d7 e2c3 b8a8 a6b7 f8b8 b7d5 g6f4 d5d2 a4b3 c4b3 b8b3 c3d5');
INSERT INTO evaluations VALUES(36452,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36452,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36452,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36452,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(36452,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(36452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36452,36,'cp 5','e7e6 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 c1f4 a8c8 c3e4 f6e4 e1e4 g7b2 a1b1 b2e5 f4e5 d6e5 d5d6 d8f6 d6d7 c8c6 b1b7 c6d6 d3e2 f8d8 g2g3 d6d7 b7d7 d8d7 e4e5 g8g7');
INSERT INTO evaluations VALUES(38533,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(38533,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(38533,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(38533,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(38533,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(38533,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(38533,36,'cp 92','c1e3 g8f6 d1d2 e8g8 f2f3 b8c6 e1c1 d6d5 d4c6 b7c6 e3h6 c8e6 e4d5 e6d5 h6g7 g8g7 d2f4 d8e7 f1d3 a8d8 h1e1 d5e6 f4e3 e7b4 a2a3 b4b6 e3e5 e6d5 c3a4 b6f2 e5g3 f2g3 h2g3 f6d7 d3e4 f8e8 a4c3 d5e4');
INSERT INTO evaluations VALUES(38532,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(38532,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(38532,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(38532,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(38532,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(38532,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(38532,36,'cp 82','g8f6 d1d2 b8c6 f2f3 e8g8 e1c1 d6d5 d4c6 b7c6 e3h6 c8e6 e4d5 e6d5 h6g7 g8g7 d2f4 d8e7 f1d3 a7a5 a2a3 f8e8 h1e1 f6h5 f4d4 e7f6 d4f6 h5f6 b2b3 a8d8 e1e8 d8e8 c3a4 e8e5 c1d2 f6d7 c2c4 d5e6 d3e4 c6c5 a4c3 d7f6 d1e1 g6g5 e4d5 e5e1 d2e1 f6d5 c4d5 e6f5 g2g4');
INSERT INTO evaluations VALUES(41603,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(41603,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(41603,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(41603,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(41603,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(41603,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(41603,36,'cp 119','d1d2 b8c6 f2f3 c6d4 e3d4 e8g8 e1c1 c8e6 g2g4 f8e8 g4g5 f6h5 d4g7 h5g7 h2h4 c7c6 a2a3 d8a5 d2d6 a8d8 d6h2 d8d1 c3d1 b7b5 h2d2 a5c7 d1f2 a7a5 d2e3 e8d8 f1e2 g7h5 c1b1 c7f4 e3c3 d8d2 h1d1 d2d1 f2d1 f4h4 c3c6 h4g5');
INSERT INTO evaluations VALUES(40844,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(40844,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(40844,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(40844,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(40844,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(40844,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(40844,36,'cp 55','e1g1 e8g8 c2c4 b8c6 d4c6 b7c6 b1c3 f6d7 c1e3 e7f6 f2f4 f6c3 b2c3 c6c5 d1c2 f8e8 e3f2 c8b7 a1e1 h7h6 f2g3 f7f6 c2f2 a8b8 d3b1 b7a6 f2e2 b8b1 e1b1 d6d5 e2c2 d5e4 f1d1 a6c4 c2d2 c4d3 b1b7 d7b6 b7a7');
INSERT INTO evaluations VALUES(40843,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(40843,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(40843,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(40843,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(40843,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(40843,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(40843,36,'cp 59','e8g8 c2c4 b8c6 d4c6 b7c6 b1c3 f6d7 c1e3 e7f6 d1d2 d7e5 d3e2 c6c5 f2f4 e5g4 e4e5 d6e5 d2d8 f8d8 e3c5 e5f4 c3d5 f6g5 d5c7 a8b8 b2b3 g4e3 c5a7 b8b7 a7e3 f4e3 c7d5 c8e6 a1d1 b7d7 e2f3 g8f8 a2a4 d8b8 c4c5 e6d5 d1d5 d7d5 f3d5');
INSERT INTO evaluations VALUES(36268,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(36268,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(36268,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(36268,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(36268,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(36268,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(36268,36,'cp 63','c2c4 b8c6 d4c6 b7c6 b1c3 f6d7 c1e3 e7f6 d1d2 d7e5 d3e2 e5g4 e3d4 f6d4 d2d4 c6c5 d4d2 g4e5 f2f4 e5c6 a1e1 c6d4 e2d3 f8e8 d2f2 a8b8 b2b3 c8d7 e4e5 a7a5 f2g3 d7f5 d3f5 d4f5 g3g4 d6e5 g4f5 d8d4 g1h1 d4c3');
INSERT INTO evaluations VALUES(44834,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(44834,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(44834,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(44834,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(44834,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(44834,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(44834,36,'cp 49','g7g6 h2h3 f8g7 f1e2 c8f5 e1g1 e8g8 b1c3 b8c6 f1e1 a8c8 a2a3 f8e8 g2g4 f5d7 c1g5 h7h6 g5h4 g6g5 h4g3 e7e5 c3e4 f7f5 e4d6 f5f4 d6e8 d7e8 g3f4 g5f4 d4e5 c6e5 f3e5 g7e5 c4c5 d8d1 a1d1 c8c5 e2f3 e8c6 f3c6 b7c6');
INSERT INTO evaluations VALUES(44701,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(44701,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(44701,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(44701,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(44701,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(44701,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(44701,36,'cp 76','a2a4 a7a5 c1e3 f8g7 f1e2 c8g4 b1a3 b8c6 h2h3 g4f3 e2f3 a8c8 e1g1 e8g8 a1c1 d6d5 c4c5 b6a8 h3h4 a8c7 a3b5 c7e6 g2g3 h7h5 f1e1 d8d7 f3g2 c8d8 d1d2 g8h7 c1d1 h7g8 b2b3 f7f5 g2f3 f5f4 e3f4 e6f4');
INSERT INTO evaluations VALUES(5780,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(5780,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(5780,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(5780,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(5780,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5780,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(5780,36,'cp 33','g8f6 h2h3 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 d1b3 c8f5 e4e2 e7g6 c1e3 d8d6 c3b5 d6d8 a1d1 f5d7 e3b6 a7b6 b5c3 h7h6 b3c4 a8c8 c4d4 f8e8 e2e8 d7e8 f3e5 g6e5 d4e5 d8d6 e5d6 c7d6 d1e1');
INSERT INTO evaluations VALUES(5798,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(5798,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(5798,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(5798,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(5798,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(5798,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(5798,36,'cp 80','b1d2 b4d6 f4g3 f7f5 f1d3 g8f6 c2c4 e8g8 a2a3 f6e4 a1c1 g8h8 b2b4 a7a6 d2b3 d8e7 f3e5 d6e5 g3e5 c6e5 d4e5 a6a5 b3a5 c7c5 d3e4 c5b4 c4d5 e6d5 e4d3 a8a5 a3b4 e7b4 d1d2 b4d2 e1d2 g7g6 d2e2 c8d7 h1d1 d7c6');
INSERT INTO evaluations VALUES(10079,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(10079,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(10079,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(10079,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(10079,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(10079,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(10079,36,'cp -92','c5b4 c2c3 b4e7 d2d4 e5d4 e4e5 f6e4 c3d4 d7d5 e5d6 e4d6 c4d5 e8g8 b1c3 e7f6 c1f4 d6f5 d5c6 b7c6 c3e4 f6e7 e4g3 f8e8 a1c1 d8d5 d1a4 c8e6 a4c6 f5g3 h2g3 d5a2 f4c7 e7b4 f1d1 e6b3 d4d5 b3d1 c1d1 f7f6');
INSERT INTO evaluations VALUES(10394,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(10394,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(10394,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(10394,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(10394,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(10394,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(10394,36,'cp -117','c2c3 b4e7 d2d4 e5d4 e4e5 f6e4 c3d4 d7d5 e5d6 e4d6 c4d5 e8g8 b1c3 d6f5 c1b2 e7f6 d5c6 b7c6 c3e4 f6e7 e4c5 f7f6 f1e1 e7c5 d1b3 d8d5 d4c5 d5b3 a2b3 g8f7 g2g4 f5e7 e1e4 e7d5 f3d4 c8d7 a1a6 f8b8 f2f3 d5b4 a6a5 a7a6 b2a3 b4d5 g1f2');
INSERT INTO evaluations VALUES(10389,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(10389,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(10389,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(10389,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(10389,34,'cp -119',NULL);
INSERT INTO evaluations VALUES(10389,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(10389,36,'cp -108','b4e7 d2d4 e5d4 e4e5 f6e4 c3d4 d7d5 e5d6 e4d6 c4d5 e8g8 b1c3 d6f5 c1b2 e7f6 d5c6 b7c6 c3e4 f6e7 e4c5 d8d5 f1e1 e7d6 d1a4 a8b8 b2c3 f5e7 c5e4 e7g6 f3e5 c8e6 h2h3 g6f4 a4c6 f8d8 c6d5 e6d5 c3d2 d6e5 d4e5 f4g2 g1g2 d5e4 e1e4 d8d2');
INSERT INTO evaluations VALUES(10891,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(10891,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(10891,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(10891,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(10891,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(10891,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(10891,36,'cp -87','d2d4 e5d4 e4e5 f6e4 c3d4 d7d5 e5d6 e4d6 c4d5 e8g8 b1c3 d6f5 a1b1 a7a5 a2a3 e7f6 c1e3 a5a4 f1e1 h7h6 h2h3 f8e8 d1d3 f5d6 g2g4 c8e6 e3f4 e6d5 c3d5 c6a5 d5f6 d8f6 f4e5 f6d8 e5d6 e8e1 b1e1 d8d6 d3b5 d6a3 e1e8 a8e8 b5e8 a3f8 e8a4 a5c6 d4d5 c6e7 a4c4 c7c6');
INSERT INTO evaluations VALUES(10914,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(10914,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(10914,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(10914,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(10914,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(10914,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(10914,36,'cp -106','d7d6 b1d2 e8g8 c4d3 f8e8 h2h3 h7h6 f1e1 e7f8 c1b2 c8d7 a2a4 b7b6 d2f1 c6a5 f3d2 a7a6 f1e3 b6b5 a4b5 a6b5 c3c4 e5d4 b2d4 b5c4 d3c2 d7c6 d4c3 a5b3 a1a8 c6a8 e4e5 d6e5 d2c4 b3d4 c2a4 a8c6 c4e5 c6a4 d1a4 e8e5');
INSERT INTO evaluations VALUES(16309,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(16309,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16309,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16309,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(16309,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(16309,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(16309,36,'cp 12','b5c6 d7c6 d2d3 f8d6 g1e2 d6c5 h2h3 c8e6 e2g3 d8d6 e1g1 e8c8 d1f3 f6d7 c3a4 f7f6 a4c5 d6c5 f3e2 c5e7 f2f4 b7b6 c1d2 a7a5 a2a3 e5f4 f1f4 c8b7 b2b4 a5a4 g3f5 e6f5 f4f5 g7g6 f5f1 d7e5 d2c3');
INSERT INTO evaluations VALUES(16226,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16226,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16226,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16226,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16226,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16226,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16226,36,'cp 0','d7c6 d2d3 f8d6 g1e2 d6c5 c3a4 c5d6');
INSERT INTO evaluations VALUES(17781,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(17781,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(17781,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(17781,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(17781,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(17781,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17781,36,'cp 54','b5c6 b7c6 d2d4 e5d4 d1d4 c6c5 d4d3 g8f6 b1c3 f8e7 e1g1 e8g8 c1f4 a6a5 f1e1 c8a6 d3d2 f8e8 a1d1 a6b7 e4e5 d6e5 e1e5 b7f3 g2f3 d8d2 d1d2 a8b8 b2b3 e7d6 e5e8 f6e8 c3d5 d6f4 d5f4 g8f8 f4d5 b8a8 d5e3 e8d6 d2d5');
INSERT INTO evaluations VALUES(17751,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(17751,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(17751,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(17751,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(17751,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(17751,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(17751,36,'cp 37','b7c6 d2d4 f7f6 c1e3 g8e7 b1c3 e7g6 d1e2 c8e6 d4e5 d6e5 e1g1 f8d6 b2b3 e8g8 a1d1 d8e7 f3e1 g6f4 e3f4 e5f4 e1f3 d6b4 d1d3 f8d8 f1d1 e6f7 c3a4 d8d3 e2d3 f7g6 f3d4 g6e4 d4c6 e4d3 c6e7 g8f7 d1d3 f7e7 a4c3 b4d6 f2f3 e7f7');
INSERT INTO evaluations VALUES(18016,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(18016,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(18016,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(18016,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(18016,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(18016,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(18016,36,'cp 53','d2d4 f7f6 b1c3 g8e7 d1e2 c8e6 d4e5 d6e5 c1e3 e7g6 e1g1 f8d6 f3e1 d8e7 e1d3 e8g8 f2f3 h7h6 b2b3 g6f4 e2f2 f4d3 c2d3 f6f5 e4f5 f8f5 f2c2 g8h8 c3e4 e6d5 e3c5 e7e6 a1e1 e6g6 g1h2 a6a5 e1e2 f5h5 f1e1 h5f5 e2e3');
INSERT INTO evaluations VALUES(32502,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(32502,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(32502,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(32502,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(32502,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(32502,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(32502,36,'cp 16','c7c5 c2c3 b8c6 f1e2 g8e7 e1g1 c5d4 c3d4 e7f5 b1c3 a7a6 c1f4 d8b6 c3a4 b6a7 f4e3 f8e7 a1c1 c6a5 e2d3 d7b5 d3f5 b5f1 d1f1 e6f5 b2b3 a5c6 f1d3 f5f4 e3f4 e8g8 h2h4 a8c8 d3f5 c6b4');
INSERT INTO evaluations VALUES(34569,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(34569,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(34569,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(34569,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(34569,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(34569,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(34569,36,'cp 146','g1f3 g6b6 d2d4 g8f6 c1d2 g7g6 a2a3 f8g7 e1c1 c8e6 f3g5 e6f5 d2e3 h7h6 g5f3 b8d7 d4d5 b6a5 f3h4 e8c8 h4f5 g6f5 e3d4 h8e8 g2g3 a7a6 f1h3 e7e5 d5e6 f7e6 h1e1 c8b8 e2c4 d7b6 c4b4 a5b4 a3b4 b6c4 d4f6 g7f6 d1d8 e8d8 e1e6 f6d4 h3f5 d4f2 c1b1 f2d4 e6h6');
INSERT INTO evaluations VALUES(34565,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(34565,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(34565,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(34565,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(34565,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(34565,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(34565,36,'cp 124','g6b6 a2a4 g8f6 a4a5 b6e6 d2d4 e6e2 f1e2 e7e6 f3e5 c8d7 e5d7 e8d7 e1g1 a7a6 f1d1 b8c6 d4d5 e6d5 c3d5 f6d5 d1d5 d7e8 e2f3 a8d8 c2c3 d8d5 f3d5 f8d6 g2g3 c6d8 c1g5 f7f6 g5e3 e8d7 d5g2 h8e8 g2h3 d7e7 b2b4 e7f7 g1f1 e8e5 c3c4 d8c6 a1d1');
INSERT INTO evaluations VALUES(35060,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(35060,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(35060,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(35060,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(35060,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(35060,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(35060,36,'cp 38','h4d8 d2d3 d7d5 d1e2 f8e7 g1f3 c7c5 f1g2 b8c6 e1g1 g8f6 c2c4 e8g8 b1c3 a8b8 e4e5 f6e8 f4f5 d5c4 d3c4 c6d4 e2f2 d4f5 f1d1 c8d7 g3g4 f5d4 f3d4 c5d4 f2d4 d7c6 d4d8 e7d8 c1e3 d8b6 e3b6 a7b6 g4g5 f7f6 e5f6');
INSERT INTO evaluations VALUES(57050,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(57050,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(57050,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(57050,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(57050,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(57050,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(57050,36,'cp 44','d2d3 c7c5 f1g2 d7d5 d1e2 f8e7 g1f3 b8c6 e1g1 g8h6 b1a3 e8g8 c2c3 f7f6 a3c2 d5d4 c3d4 c5d4 a2a4 a7a6 g2h3 g8h8 c1d2 a8b8 b2b4 b7b5 a4b5 a6b5 e4e5 h6f5 e5f6 e7f6 h3g2 d8c7 f3g5 c6e7 c2a3 e7d5 g5e6 c8e6 e2e6');
INSERT INTO evaluations VALUES(44242,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(44242,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(44242,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(44242,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(44242,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(44242,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(44242,36,'cp 70','d4d5 d7d6 f1c4 e6d5 c3d5 f8e7 g1e2 f6d5 c4d5 e8g8 e2f4 b8a6 e1g1 a6c7 c2c3 f8e8 a2a4 c7d5 f4d5 c8e6 f2f3 d8d7 c1f4 e7f8 f4g3 f7f5 e4f5 e6f5 c3c4 d7f7 d1d2 f5e6 a1d1 a8d8 f1e1 e6d5 d2d5');
INSERT INTO evaluations VALUES(44324,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(44324,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(44324,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(44324,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(44324,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(44324,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(44324,36,'cp 52','f1g2 d7d6 d2d3 f8e7 c1g5 f6g8 g5d2 b8c6 a2a3 g8f6 g1f3 h7h6 e1g1 e8g8 a1b1 a7a5 f3e1 c8e6 c3d5 a8b8 e1c2 e6d5 c4d5 c6d4 b2b4 a5b4 a3b4 b7b5 e2e4 d4c2 d1c2 d8d7 f1c1 f8c8 b4c5 c8c5');
INSERT INTO evaluations VALUES(49451,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(49451,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(49451,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(49451,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(49451,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(49451,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(49451,36,'cp 28','e4e5 f6d5 d3d4 d7d6 g1f3 d6e5 d4e5 b8c6 a2a3 h7h6 h2h4 b7b6 b2b4 c8b7 c1b2 f8e7 g2g3 a7a6 b1d2 b6b5 c2c4 b5c4 d2c4 d5b6 c4b6 c7b6 f1g2 d8d1 a1d1 a8b8 e1g1 e8g8 d1d7 f8d8 f1d1');
INSERT INTO evaluations VALUES(51079,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(51079,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(51079,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(51079,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(51079,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(51079,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(51079,36,'cp 114','d2d4 e7e6 f1d3 d7d6 e1g1 b8c6 c2c3 g7g5 b1d2 f8g7 d2c4 g8e7 d3c2 e7g6 c4e3 e6e5 h2h3 d8f6 a2a4 e5d4 e3d5 f6d8 f3d4 c6d4 c3d4 e8g8 a1a3 c8e6 d5e3 d8f6 d4d5 e6d7 e3g4 d7g4 d1g4 g6e5 g4e2 f6g6 a3b3 b7b5 a4b5');
INSERT INTO evaluations VALUES(5812,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(5812,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(5812,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(5812,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(5812,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(5812,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(5812,36,'cp 53','g8f6 c1d2 e8g8 a2a3 b4d6 c4c5 d6e7 f1b5 f6e4 e1g1 c8d7 b5d3 f7f5 b2b4 e7f6 c3e2 a7a6 d2e1 d8e7 a1c1 b7b5 h2h3 d7e8 g1h1 e8h5 a3a4 a6a5 a4b5 c6b4 d3b1 g7g5 d1b3 g8h8 e1b4');
INSERT INTO evaluations VALUES(8434,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(8434,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8434,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(8434,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(8434,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8434,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(8434,36,'cp 0','c6a5 c4b3 c7c6 c3e2 e8g8 e2g3 a5b3 a2b3 f8e8 c1d2 e7f8 f1e1 c6c5 f3h2 b7b5 d2a5 d8e7 g3f5 e7b7 d1f3 c8f5 e4f5 b7f3 h2f3 f6d5 g2g4 a7a6 h3h4 d5e7 a5b6 e7d5');
INSERT INTO evaluations VALUES(10417,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(10417,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(10417,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(10417,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(10417,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10417,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(10417,36,'cp 11','e2e4 b4c5 f1e2 d7d6 e1g1 c8g4 f3e1 g4e2 c3e2 e8g8 e1c2 a7a5 a1b1 f6d7 b2b3 f7f5 d3d4 e5d4 c2d4 c6d4 e2d4 c5d4 d1d4 d8f6 d4f6 f8f6 c1b2 f6f7 e4f5 f7f5 f1e1 f5f7 e1e3 a8f8 e3g3 d7f6 b1f1 a5a4');
INSERT INTO evaluations VALUES(20024,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(20024,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(20024,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20024,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(20024,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(20024,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(20024,36,'cp -13','c2c4 g7g6 b1c3 a7a6 f1e2 f8g7 g1f3 c8g4 e1g1 e7e6 f3d2 g4e2 c3e2 g8e7 d2f3 e8g8 d3d4 d6d5 e4d5 e6d5 a1c1 d8a5 c4d5 e7d5 c1c5 a5a2 d1d2 c6e7 e3h6 b7b6 h6g7 g8g7 c5c1 a2a5 d2a5 b6a5 c1c2 f8b8 f1b1 e7f5 h2h3 a5a4');
INSERT INTO evaluations VALUES(43687,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(43687,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(43687,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(43687,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(43687,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(43687,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(43687,36,'cp 60','g1f3 b8c6 d2d4 f8d6 c1g5 e5d4 f3d4 d6e5 d4c6 b7c6 f2f4 e5c3 b2c3 h7h6 g5h4 d7d5 e4e5 g7g5 f4g5 f6e4 d1h5 c8f5 c4d3 f5g6 h5f3 h6g5 h4g3 d8e7 e1g1 e7c5 g3f2 c5e7 d3e4 g6e4 f3e3 e7e5 f2g3 e5e7 e3d4 e8g8 f1f6 c6c5 d4f2 g8g7 a1f1 a8e8 a2a4 e4h7 a4a5 e7e3 f2e3 e8e3 f6a6');
INSERT INTO evaluations VALUES(43686,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(43686,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(43686,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(43686,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(43686,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(43686,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(43686,36,'cp 50','b8c6 d2d4 f8d6 c1g5 e5d4 f3d4 d6e5 d4c6 b7c6 f2f4 e5c3 b2c3 h7h6 g5h4 d7d5 e4e5 g7g5 f4g5 f6e4 d1h5 c8f5 g5h6 f5g6 h5g4 d8d7 c4e2 d7g4 e2g4 h8h6 g2g3 e4c3 e1g1 g6h5 g4h3 c3e2 g1f2 e2c3 a1e1 a8b8 f2g1 c3e4 c2c4 h6g6 c4d5 c6d5 h3g2 e4c3 f1f6 g6f6 h4f6 c3e4 g2e4 d5e4 e1e4 b8b1 g1g2');
INSERT INTO evaluations VALUES(44628,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(44628,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44628,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(44628,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(44628,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44628,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(44628,36,'cp 11','e3d4 d8b6 b1a3 f6d5 f4c1 e7e6 a3c4 b6c7 c4e3 d5f6 f1d3 d7d5 c2c3 b8c6 d1e2 f8e7 g1f3 e8g8 e3c2 h7h6 f3e5 e7d6 f2f4 c6e7 e1g1 c8d7 e5d7 c7d7 c2e3 f6e4 a2a4 a7a6 g1h1');
INSERT INTO evaluations VALUES(44618,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44618,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44618,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44618,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(44618,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44618,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(44618,36,'cp 8','d8b6 b1a3 f6d5 f4c1 e7e6 a3c4 b6c7 c4e3 d5f6 f1d3 b8c6 g1f3 d7d5 e1g1 f8d6 f1e1 e8g8 g2g3 f8e8 c2c4 d5c4 e3c4 h7h6 c4d6 c7d6 c1f4 d6e7 a1c1 f6d5 f4d2 c8d7 d1e2 e8d8 a2a3 a8c8 e2e4 d5f6 e4h4 a7a6 d2c3');
INSERT INTO evaluations VALUES(47147,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47147,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(47147,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(47147,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(47147,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(47147,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(47147,36,'cp 0','d5c4 g1f3 b8c6 b1c3 e7e5 f3e5 d8d4 e5c6 d4d1 c3d1 b7c6 f1g2 f8b4 c1d2 b4d2 e1d2 f6d5 b2b3 c4b3 a2b3 e8e7 d1b2 h8d8 b2d3 c8f5 h1c1 d5b4 g2c6 b4d3 e2d3 d8d3 d2e1 a8c8 a1a7 d3b3 c6b7 c8e8 b7c6');
INSERT INTO evaluations VALUES(55489,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(55489,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(55489,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(55489,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(55489,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(55489,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(55489,36,'cp -77','c7c6 b5c4 g7g6 h5e2 g8f6 g1f3 b7b5 c4b3 f8g7 d2d3 b8d7 c1e3 e8g8 a2a4 b5b4 b1d2 d6d5 h2h4 h7h6 e4d5 c6d5 g2g4 f6g4 b3d5 g4e3 e2e3 a8b8 d5b3 d7f6 e3a7 b8b7 a7e3 f6g4 e3c5 b7c7 c5b4 f8e8');
INSERT INTO evaluations VALUES(53796,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(53796,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(53796,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(53796,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(53796,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(53796,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(53796,36,'cp -64','b5c4 g7g6 h5e2 g8f6 g1f3 b7b5 c4b3 f8g7 d2d3 e8g8 c1e3 f8e8 c2c3 a7a5 a2a4 b5b4 e1g1 d6d5 e2c2 h7h6 b1d2 c8b7 f1e1 b8d7 h2h3 d8e7 a1c1 c6c5 e4d5 f6d5 d2e4 b4c3 b2c3 g8h7 c1b1 d5e3 e1e3 f7f5 b3c4 b7c6 e4d2');
INSERT INTO evaluations VALUES(53805,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(53805,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(53805,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(53805,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(53805,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(53805,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(53805,36,'cp -78','g7g6 h5e2 f8g7 g1f3 g8e7 e1g1 e8g8 d2d3 d6d5 c4b3 f8e8 b1c3 b8a6 a2a3 h7h6 c1d2 a6c5 b3a2 c8e6 e4d5 e7d5 c3e4 d5f4 d2f4 e5f4 e4c5 e6a2 e2d2 a2d5 d2f4 d8b6 c5a4 b6b5 a4c3 b5b2 c3d5 c6d5');
INSERT INTO evaluations VALUES(951,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(951,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(951,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(951,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(951,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(951,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(951,36,'cp 98','c1g5 f6g6 f1d3 g6h5 h2h3 f7f6 g5f4 f8d6 f4d6 c7d6 e1g1 e8g8 f1e1 c6e7 d3c4 e7d5 d1d2 c8d7 c4f1 b7b5 b2b3 h5h6 d2h6 g7h6 c2c4 b5c4 f1c4 d5c7 a1c1 a8c8 a2a3 a7a5 a3a4 c8b8 f3h4');
INSERT INTO evaluations VALUES(950,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(950,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(950,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(950,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(950,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(950,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(950,36,'cp 89','f6g6 f1d3 g6h5 h2h3 f7f6 g5f4 f8d6 f4d6 c7d6 e1g1 e6e5 a1c1 e8g8 f1e1 c8e6 c2c4 e6f7 d3e4 f7g6 e4g6 h5g6 d4d5 c6b4 a2a3 b4a6 b2b4 a6c7 f3d2 b7b6 d1b3 g6f7 a3a4 a8c8 b3a3 f6f5 a4a5 b6b5 a3e3 b5c4 e3a7');
INSERT INTO evaluations VALUES(4799,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(4799,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(4799,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(4799,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(4799,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(4799,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(4799,36,'cp -21','g8e7 c1e3 e8g8 e1g1 c8e6 a2a4 b7b6 f3g5 e6c8 g5f3 c6c5 f3d2 d8d7 d2c4 c8b7 f2f4 e5f4 c4d6 c7d6 e3f4 d6d5 e4d5 e7d5 c3d5 b7d5 d1d2 h7h6 a4a5 b6b5 a1e1 f8e8 f4g3 e8e1 f1e1 a8d8 d2f2 c5c4 d3c4');
INSERT INTO evaluations VALUES(12422,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(12422,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(12422,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(12422,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(12422,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(12422,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(12422,36,'cp 124','d4d5 c6a5 c1g5 h7h6 g5f6 d8f6 f1d3 e6e5 b1d2 f8e7 e1g1 b7b6 d1e2 e8g8 d3a6 c8a6 e2a6 f6d6 a6e2 a7a6 b2b3 b6b5 g2g3 d6c7 g1g2 c5c4 b3c4 a5c4 d2c4 b5c4 a1b1 f8c8 a2a4 a8b8 f3d2 b8b1 f1b1 c7a5 d2c4 a5c3 c4e3 a6a5 b1b7 c8c7 b7b8 e7f8');
INSERT INTO evaluations VALUES(15124,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(15124,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(15124,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(15124,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(15124,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(15124,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(15124,36,'cp -28','d5e4 f3g5 c8e6 b3e6 f7e6 g5e4 f6e4 d3e4 d8h4 e1g1 h4e4 c2c3 e4g6 b1d2 e8c8 d1a4 a7a6 f1e1 f8e7 d2e4 d8d3 c1e3 h8d8 a4c4 c6a5 c4a4 d3d5 a1c1 g6e8 a4e8 d8e8 b2b3 h7h6 g1f1 d5d8 f1e2 b7b6 f2f3 e8f8 c1d1 d8d5 c3c4 d5d1');
INSERT INTO evaluations VALUES(15228,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(15228,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(15228,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(15228,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(15228,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(15228,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(15228,36,'cp -26','f3g5 c8e6 b3e6 f7e6 g5e4 f6e4 d3e4 d8h4 e1g1 h4e4 c2c3 e4f5 c1e3 a7a6 d1a4 f8e7 b1d2 e8c8 a1e1 h7h5 a4c4 d8d5 d2e4 g7g5 f2f3 g5g4 f3g4 f5g4 h2h3 g4g6 g1h1 h8g8 e1e2 c6d4 c3d4 g6e4 e3g5 e4h7 g5e7 h7e7 e2e5 e7g7 c4e2 c8b8 f1f3 d5e5 d4e5 g8d8');
INSERT INTO evaluations VALUES(30014,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(30014,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(30014,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(30014,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(30014,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(30014,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(30014,36,'cp -81','d2d4 f7f6 e5d3 d5c4 f3e4 d8d5 e4d5 e6d5 d3f4 d5f7 b1c3 b8c6 c1e3 e8c8 e1c1 f8b4 h2h4 h8e8 h4h5 c6e7 g2g4 h7h6 h1g1 a7a6 a2a3 b4c3 b2c3 e7d5 f4d5 f7d5 d1e1 d8d6 g1g3 d6e6 c1d1 b7b6 e3d2 e6e1 d2e1 a6a5 d1d2 e8e4 f2f3 e4e7');
INSERT INTO evaluations VALUES(31637,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(31637,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(31637,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(31637,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(31637,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(31637,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(31637,36,'cp 56','f1b5 b8d7 c2c4 c5d4 d1d4 a7a6 b5d7 d8d7 c4d5 d7d5 d4d5 e6d5 f3d4 f5e6 b1c3 g8e7 e1e2 e7c6 h1d1 f8e7 d4b3 a8d8 f4c7 d8d7 c7b6 e8g8 a1c1 f8c8 b3c5 e7c5 b6c5 c6e5 c5b6 c8c6 c3a4 f7f6 c1c6 b7c6 d1c1 d7b7 f2f3 e5c4 b6d4 g8f7');
INSERT INTO evaluations VALUES(32772,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(32772,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(32772,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(32772,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(32772,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(32772,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(32772,36,'cp -27','d2d4 e7e6 g1e2 f8d6 c1f4 b7b5 b1d2 g8f6 e1g1 e8g8 a2a3 b8c6 c2c3 a7a5 d2b3 h7h6 f1e1 f8c8 f4d6 d7d6 h2h3 f6d7 e2c1 b5b4 a3b4 a5b4 c1d3 a8a1 d1a1 d7b6 a1d1 b4c3 b2c3 c8a8 d3c5 d6c7 d1g4 b6c4 c5d3 a8a3 d3c5 a3a2');
INSERT INTO evaluations VALUES(32768,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(32768,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(32768,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(32768,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(32768,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(32768,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(32768,36,'cp -27','e7e6 g1e2 b8c6 c2c3 g8e7 e1g1 e7g6 e2f4 f8d6 f4h5 e8c8 b1d2 c8b8 a2a4 d7c7 d2f3 f7f6 b2b4 h8e8 h2h3 b8a8 c1a3 e6e5 b4b5 c6a5 a3d6 d8d6 f3d2 e5e4 d2b3 a5c4 b3d2 f6f5 d2c4 d5c4 d1c2 a8b8 f1e1 c7f7 a4a5 e8c8 a5a6');
INSERT INTO evaluations VALUES(37319,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(37319,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(37319,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(37319,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37319,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(37319,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(37319,36,'cp 0','e5f4 d7d5 g1f3 b8c6 f1e2 e7b4 c1d2 f8e8 b1c3 d5e4 d3e4 b4c3 d2c3 f6e4 a1d1 d8e7 e1g1 g7g5 f4c1 e4c3 b2c3 e7f6 e2d3 c8g4 f3g5 g4d1 g5h7 f6g7 c1f4 d1e2 h7f6 g8f8 f6h7');
INSERT INTO evaluations VALUES(43716,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(43716,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43716,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(43716,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(43716,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(43716,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(43716,36,'cp 62','a6b5 c3b5 e7e6 d5e6 f7e6 e2e3 d7d5 f1e2 f6e4 g1f3 d8a5 c1d2 e4d2 d1d2 f8e7 a2a4 e7f6 d2a5 a8a5 a1c1 a5a4 c1c5 e8e7 e1g1 c8d7 b5d4 b8a6 c5c2 h8c8 b2b3 a4a3 c2c8 d7c8 f1c1 c8d7');
INSERT INTO evaluations VALUES(43713,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(43713,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(43713,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(43713,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(43713,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(43713,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(43713,36,'cp 63','c3b5 e7e6 d5e6 d8a5 b5c3 f7e6 g1f3 f6e4 c1d2 e4d2 d1d2 f8e7 e2e4 e8g8 f1b5 c8a6 b5e2 b8c6 e1g1 a8d8 e2a6 a5a6 a1e1 d7d6 e1d1 h7h6 d2c2 e7f6 a2a3 c6e5 f3e5 f6e5 a3a4 a6c4 c2d3 c4d3 d1d3 d8b8 c3b5');
INSERT INTO evaluations VALUES(43712,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(43712,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(43712,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(43712,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(43712,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(43712,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(43712,36,'cp 64','e7e6 d5e6 d8a5 b5c3 f7e6 g2g3 c8b7 g1f3 f6e4 a2a4 e4c3 b2c3 a5c3 c1d2 c3c4 f1g2 f8e7 e1g1 e8g8 a4a5 b7d5 d2f4 c4a6 f1e1 b8c6 e2e4 d5c4 d1d7 f8f7 f3e5 c6e5 f4e5 e7f6 d7d6 a6d6 e5d6 f6d4 e4e5 f7f2 g2a8 f2a2 g1h1 d4a1 d6c5');
INSERT INTO evaluations VALUES(46833,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(46833,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46833,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46833,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46833,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46833,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(46833,36,'cp 0','e2e3 c7c5 d4c5 e7e6 g1f3 f8c5 f1e2 e8g8 c2c4 d5c4 d1c2 b7b5 a2a4 b5a4 a1a4 b8d7 a4c4 c8b7 c2h7 g8h7 c4h4 h7g6 h4g4 g6h7');
INSERT INTO evaluations VALUES(46832,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46832,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46832,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46832,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(46832,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46832,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46832,36,'cp 0','c7c5 d4c5 e7e6 g1f3 f8c5 f1e2 e8g8 c2c4 d5c4 d1c2 b7b5 a2a4 b5a4 a1a4 d8e7 a4c4 b8d7 c2h7 g8h7 c4h4 h7g6 h4g4 g6h6 g4h4');
INSERT INTO evaluations VALUES(47996,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(47996,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(47996,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(47996,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(47996,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(47996,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(47996,36,'cp 171','c1g5 d8d7 e5d6 f8d6 e2e3 f7f6 g5h4 b8c6 b1c3 d7f7 f1e2 c8e6 f3d4 c6d4 e3d4 f7h5 h4g3 d6g3 f2g3 h5f5 e2g4 f5g4 d1g4 e6g4 e1g1 e8c8 f1f4 g4e6 d4d5 e6f7 g3g4 d8d7 a1f1 h7h6 f1d1 h6h5 g4h5 f7h5 d1e1 c7c6 h2h3 c6d5 c3d5');
INSERT INTO evaluations VALUES(48530,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48530,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(48530,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(48530,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(48530,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(48530,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(48530,36,'cp 15','f3e4 f8b4 a2a3 b4c3 b2c3 f6e4 d1g4 e4c3 g4g7 d8h4 g2g3 h4e4 e1d2 h8f8 d2c3 e4h1 f1b5 b8d7 c1h6 c7c6 b5d3 b7b5 g7h7 h1h2 g1f3 h2g3 f3g5 d7f6 h7g7 f6d7 a1f1 c8a6 g5e6 g3g7 e6g7 e8d8 g7e6 f7e6 h6f8 b5b4 a3b4 a6d3 c3d3 d8c7 f8c5 a7a5 f1a1 d7c5 b4c5 c7d7 a1a4 a8a7 c2c3 d7e7 c3c4 a7a8 d3c2');
INSERT INTO evaluations VALUES(48523,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(48523,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(48523,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(48523,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48523,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48523,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48523,36,'cp 31','f8b4 a2a3 b4c3 b2c3 f6e4 d1g4 e4f6 g4g7 h8g8 g7h6 g8g6 h6e3 c8d7 g1f3 d7c6 g2g3 d8d5 f1e2 f6g4 e3f4 e6e5 c3c4 d5d4 f4d4 e5d4 e1g1 g4e3 c1e3 d4e3 a1e1 g6f6 e2d3 h7h6 e1e3 e8f8 f3e5 f6f1 g1f1 b8d7 e5c6 b7c6 f1e2 a8b8 e3e4 b8b6 e2d2 c6c5 e4h4');
INSERT INTO evaluations VALUES(42144,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(42144,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(42144,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(42144,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(42144,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(42144,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(42144,36,'cp 75','a2a3 b4c3 b2c3 f6e4 d1g4 e4f6 g4g7 h8g8 g7h6 g8g6 h6e3 c8d7 g1f3 d7c6 g2g3 d8d5 f1e2 f6g4 e3f4 b8d7 e1g1 e8c8 c3c4 d5h5 f3h4 d8g8 f4f7 h5h4 g3h4 g4e3 g1f2 e3c2 f7g6 g8g6 c1g5 c2a1 f1a1 h7h6 g5f4 d7f6 f2e3 f6g4 e3d2 g4f6 a1f1 g6g2 f4h6 f6e4 d2e3 g2h2');
INSERT INTO evaluations VALUES(52963,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(52963,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52963,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52963,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52963,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(52963,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52963,36,'cp 5','g8f6 b1d2 c8f5 g1f3 h7h6 c2c3 b8d7 f1e2 e7e6 h2h3 a7a5 d1b3 d8c7 a2a4 c7b6 b3b6 d7b6 d2b3 f8e7 f4h2 f6d7 f3d2 e6e5 e3e4 f5g6 f2f3 e5d4 b3d4 e8g8 d2c4 b6c4 e2c4 d7b6 c4b3 f8e8 e1c1 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(45097,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(45097,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(45097,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45097,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(45097,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45097,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(45097,36,'cp 30','c2c3 b8d7 h2h3 e7e5 f4h2 e5e4 c3c4 d6d5 b1d2 f8b4 a2a3 b4d2 d1d2 e8g8 a3a4 a7a5 b2b3 d7b6 c4c5 b6d7 g1e2 b7b6 c5b6 d8b6 e2c3 c8a6 f1a6 a8a6 a1b1 b6b7 e1g1 h7h5 h2d6 f8a8 f1c1 a6b6 d2c2 h5h4');
INSERT INTO evaluations VALUES(55116,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(55116,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(55116,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(55116,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(55116,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(55116,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(55116,36,'cp -55','e7e5 e2e4 d5d4 c3e2 g8e7 f2f4 b8c6 g2g3 e7g6 f1g2 e5f4 e2f4 f8b4 c1d2 a7a5 a2a3 b4d2 d1d2 e8g8 f4g6 h7g6 h2h3 d8d6 d2f2 c6e5 g1f3 e5f3 g2f3 c8e6 h3h4 a8a6 e1g1 a6b6 b2b3 a5a4 b3a4 f8a8 h4h5 a8a4 h5g6');
INSERT INTO evaluations VALUES(56718,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(56718,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(56718,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(56718,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(56718,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(56718,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(56718,36,'cp 165','f5e4 f3g5 d6d5 e5e6 g8f6 g5f7 d8e7 f7h8 c8e6 b1c3 b8c6 f1b5 e7d6 e1g1 e8c8 c3e2 c6e5 c1f4 a7a6 e2d4 a6b5 d4e6 d6e6 f4e5 f8e7 d1d2 d8h8 d2a5 e6e5 a5a8 c8d7 a8h8 e7d6 g2g3 d7c6 f1e1 e5b2 h8g7 b2c3 g7f7 c3e5 a1b1 f6g4');
INSERT INTO evaluations VALUES(56717,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(56717,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(56717,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(56717,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(56717,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(56717,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(56717,36,'cp 151','f3g5 d6d5 e5e6 g8f6 g5f7 d8e7 f7h8 c8e6 b1c3 b8c6 f1b5 e8c8 b5c6 b7c6 e1g1 d5d4 c3e2 e7d7 c1g5 f8c5 e2g3 d7d5 g5f6 g7f6 c2c4 d4c3 d1c2 d5d3 c2c3 d3c3 b2c3 f6f5 h8f7 e6f7 g3f5 f7c4 f1d1 d8f8 f5e3 c4d3 d1d2 h7h5 a1d1 h5h4 h2h3 c5d6 d1e1 d6a3');
INSERT INTO evaluations VALUES(56716,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(56716,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(56716,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(56716,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(56716,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(56716,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(56716,36,'cp 139','d6d5 e5e6 g8f6 g5f7 d8e7 f7h8 c8e6 b1c3 b8c6 f1b5 e8c8 b5c6 b7c6 c3e2 d5d4 e1g1 e7d7 c1g5 f8c5 c2c3 d4d3 e2g3 d7d5 d1d2 d8h8 a1e1 h7h5 g5f6 e4e3 f2e3 g7f6 f1f4 c5b6 c3c4 d5e5 f4e4 e5d6 e4h4 a7a5 g3e4 d6b4 d2b4 a5b4 c4c5 b6a7 e1d1 h8d8 b2b3 f6f5 e4g5');
INSERT INTO evaluations VALUES(56752,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(56752,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(56752,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(56752,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(56752,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(56752,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(56752,36,'cp 96','e4f5 c8f5 d2d4 e5e4 g2g4 f5g6 f3g5 d6d5 c2c4 f8b4 b1c3 d8d7 c4d5 g8f6 f1c4 f6d5 e1g1 b4c3 b2c3 b8c6 g5e4 g6e4 f1e1 e8c8 e1e4 d5c3 d1b3 c3d5 a1b1 c6a5 b3b2 d7c6 c4f1 d5c3 e4e5 c3b1 b2b1 b7b6 f1a6 a5b7 e5e7 h8e8 a6b5 c6b5 b1b5 e8e7 c1g5 e7e1 g1g2 d8d4 g5e3 e1e3 f2e3');
INSERT INTO evaluations VALUES(56733,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(56733,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(56733,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(56733,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(56733,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(56733,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(56733,36,'cp 83','c8f5 d2d4 e5e4 g2g4 f5d7 f3d2 d6d5 c2c4 c7c6 b1c3 g8e7 d1e2 d8c7 d2b3 b8a6 c4d5 c6d5 c1g5 h7h6 g5e7 f8e7 c3d5 c7d6 d5e7 d6e7 f1g2 e8g8 e1g1 d7c6 e2c4 e7f7 c4f7 f8f7 f1e1 f7f4 b3d2 a8d8 g2e4 c6e4 e1e4 f4e4 d2e4 d8d4');
INSERT INTO evaluations VALUES(32440,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(32440,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(32440,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(32440,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(32440,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(32440,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(32440,36,'cp 104','d2d4 e5e4 g2g4 f5d7 f3d2 h7h5 b1c3 e4e3 d2e4 d6d5 c3d5 d7c6 d5c3 g8f6 f1g2 f6e4 g2e4 c6e4 c3e4 e3f2 e1f2 f8e7 f2g2 d8d5 d1f3 b8d7 e4g5 d5f3 g5f3 h5g4 h3g4 h8h1 g2h1 e8c8 c1e3 c7c5 d4c5 e7c5 a1e1 c5e3 e1e3 d7f6');
INSERT INTO evaluations VALUES(7994,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7994,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(7994,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(7994,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7994,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(7994,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7994,36,'cp 0','b7c6 e4e5 f6g8 f2f4 g8h6 d1d2 e8g8 e1c1 d7d6 e5d6 h6f5 e3f2 c8e6 g2g4 f5d6 f4f5 g6f5 f2d4 g7d4 d2d4 d8b6 d4e5 d6e4 d1e1 a8b8 c3d1 e6d5 g4f5 f7f6 e5f4 e4c3 e1e7 c3a2 c1b1 a2c3 b1c1');
INSERT INTO evaluations VALUES(9412,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9412,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9412,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9412,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9412,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9412,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9412,36,'cp 0','e1g1 a7a5 c2c3 e8g8 f1e1 f8e8 a2a4 h7h6 b1a3 f6g4 e1f1 g4f6');
INSERT INTO evaluations VALUES(9411,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(9411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9411,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(9411,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(9411,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(9411,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9411,36,'cp -8','a7a5 c2c3 e8g8 f1e1 f8e8 b1a3 h7h6 a3c2 a5a4 c1e3 f6g4 h2h3 g4e3 c2e3 c8d7 e3d5 d7e6 b2b4 a4b3 a2b3 e6d5 a1a8 d8a8 e4d5 c6e7 d3d4 e5d4 f3d4 e7g6 e1e8 a8e8 b3b4 c5b6 d1c2 e8e5 g2f1');
INSERT INTO evaluations VALUES(10091,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(10091,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10091,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(10091,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10091,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10091,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(10091,36,'cp 0','g1e2 e8g8 e1g1 d7d6 d2d3 h7h6 f3g3 g8h7 c1e3 g7g5 e3c5 d6c5 c3d5 f6h5 g3f3 h5f4 c2c3 c6a5 e2g3 b7b5 c4b3 a5b3 a2b3 c8e6 a1a6 g5g4 f3e3 e6d5 e4d5 d8d5 e3e4 d5e4 g3e4 f4d3 e4f6 h7g7 f6h5 g7h7');
INSERT INTO evaluations VALUES(10092,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(10092,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10092,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10092,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10092,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10092,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10092,36,'cp 0','e8g8 e1g1 d7d6 d2d3 h7h6 h2h3 c8e6 c1e3 a7a6 c3d5 e6d5 e4d5 e5e4 f3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(13537,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(13537,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13537,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(13537,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13537,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(13537,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(13537,36,'cp 2','e4d5 e6d5 c4b5 c8g4 c1g5 f8e7 d4c5 e8g8 e1g1 h7h6 g5f6 e7f6 c3d5 f6b2 a1b1 b2a3 h2h3 g4f3 d1f3 a3c5 b5a4 d8h4 d5c3 c5b6 a4c6 b7c6 f3c6 f8d8 c6f3 a8c8 c3d5 c8c2 d5b6 a7b6 b1b6 h4c4 b6b7 c2a2 b7e7 a2c2 f3f5 c2e2 f5a5 d8c8 a5a7 e2b2 a7a5');
INSERT INTO evaluations VALUES(13515,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13515,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13515,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(13515,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(13515,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13515,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13515,36,'cp 0','e6d5 c4b5 c8g4 c1g5 f8e7 d4c5 e8g8 e1g1 h7h6 g5f6 e7f6 c3d5 f6b2 a1b1 b2a3 h2h3 g4f3 d1f3 a3c5 b5a4 d8h4 d5c3 c5b6 a4c6 b7c6 f3c6 a8c8 c6f3 f8d8 c3d5 c8c2 d5b6 a7b6 b1b6 h4c4 b6b7 c2a2 f1b1 a2a6 b7b8 a6d6 b1b7 c4d5 b8d8 d6d8 f3d5 d8d5 b7b8 g8h7 f2f3 h7g6 g1f2 d5c5 b8b4 c5a5 b4b7');
INSERT INTO evaluations VALUES(13743,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13743,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13743,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13743,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13743,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13743,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13743,36,'cp 0','c4b5 c8g4 c1g5 f8e7 d4c5 e8g8 e1g1 h7h6 g5f6 e7f6 c3d5 f6b2 a1b1 b2a3 h2h3 g4f3 d1f3 a3c5 b5a4 d8h4 d5c3 c5b6 a4c6 b7c6 f3c6 a8c8 c6f3 f8d8 c3d5 c8c2 d5b6 a7b6 b1b6 h4c4 b6b7 c2a2 f1e1 a2a6 f3e2 c4e2 e1e2 a6f6 g1h2 g7g6 b7b5 g8g7 b5b7');
INSERT INTO evaluations VALUES(13702,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13702,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13702,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13702,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13702,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13702,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13702,36,'cp 0','c8g4 c1g5 f8e7 d4c5 e8g8 e1g1 h7h6 g5f6 e7f6 c3d5 f6b2 a1b1 b2a3 h2h3 g4f3 d1f3 a3c5 b5a4 d8h4 d5c3 c6d4 f3b7 a8c8 a4b3 c5b6 c3d5 b6c5');
INSERT INTO evaluations VALUES(18190,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(18190,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18190,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18190,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(18190,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18190,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(18190,36,'cp 3','d1d4 e7e6 e1g1 d7d5 e4d5 c6d5 c2c4 c8b7 b1c3 d5c4 d4c4 g8f6 f3e5 d8a5 f1d1 f8e7 c4b3 a5b4 b3b4 e7b4 a2a3 b4a5 e5c4 a5c7 c4d6 c7d6 d1d6 f6d5 c3d5 b7d5 c1f4 a6a5 a1c1 e8g8 h2h3 f7f6 c1c7 f8d8 d6d7 d8d7 c7d7 a5a4 g1h2 a8c8 d7a7 c8c2');
INSERT INTO evaluations VALUES(18189,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(18189,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18189,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(18189,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18189,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(18189,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18189,36,'cp 0','e7e6 e1g1 d7d5 b1c3 g8f6 f1d1 c8b7 a1b1 f6e4 c3e4 d5e4 d4d8 a8d8 d1d8 e8d8 f3g5 d8e8 b2b4 e6e5 g5e4 b7c8 c1e3 f8e7 e3c5 c8f5 f2f3 f5e4 f3e4 e8d7 c5e7 d7e7 g1f2 a6a5 b4b5 c6b5 b1b5 e7e6 b5b6 e6e7');
INSERT INTO evaluations VALUES(18657,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(18657,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(18657,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(18657,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(18657,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(18657,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(18657,36,'cp 45','d7d6 h2h3 a7a6 e1g1 g8f6 c3d5 f8e7 c1e3 f6d5 c4d5 e8g8 c2c3 c8e6 d5e6 f7e6 d3d4 e5d4 c3d4 c5d4 f3d4 c6d4 e3d4 e7f6 d4f6 d8f6 d1d6 a8d8 d6b4 b7b5 a1e1 d8d4 b4c3 b5b4 c3e3 f8d8 e1e2 e6e5 b2b3 f6f4 e3f4 e5f4 f2f3 d4d2 e2d2');
INSERT INTO evaluations VALUES(20140,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(20140,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20140,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(20140,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(20140,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(20140,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(20140,36,'cp 47','f3g1 g8e7 g1e2 g7g6 c1e3 f8g7 d1d2 c8e6 e1g1 d8d7 f2f4 e8c8 a1e1 e6c4 d3c4 f7f5 f4e5 c6e5 b2b3 e5g4 g1h1 g4e3 d2e3 h8f8 e3d3 g7c3 e2c3 f5f4 e4e5 d6e5 d3d7 d8d7 e1e5 e7c6 e5e6 d7d2 c3e4 d2c2 e6g6 c2e2');
INSERT INTO evaluations VALUES(25939,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25939,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25939,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25939,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25939,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(25939,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(25939,36,'cp 12','b5c6 d7c6 d2d4 g8e7 e1g1 e7g6 g2g3 b7b6 h2h4 g6e7 b1a3 d8d7 a3c2 h7h6 f1e1 a7a5 c2e3 c6b7 e3g2 b7a6 g2f4 e7f5 g3g4 f5e7 c1e3 e7c6 a1c1 f8e7 f4h5 e8c8 d4c5 b6c5 b2b4 a5b4 c3b4 c5c4 f3d4 c6d4');
INSERT INTO evaluations VALUES(25903,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(25903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25903,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(25903,36,'cp 2','d7c6 d2d4 g8e7 e1g1 e7f5 a2a4 a7a5 d4c5 f8c5 b1d2 d5d4 d2b3 c6f3 d1f3 d8b6 b3c5 b6c5 f1d1 e8g8 g2g4 f5h4 f3g3 h4g6 c3d4 c5c2 g3d3 c2c6 d4d5 e6d5 d3d5 c6c2 c1e3 c2b2 h2h3 a8e8 d5b5 b2e5 b5e5 e8e5 a1b1 f8b8 e3a7 b8c8 d1d7 e5e4 b1b7 e4a4');
INSERT INTO evaluations VALUES(26361,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26361,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26361,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26361,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26361,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(26361,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(26361,36,'cp 0','d2d4 g8e7 e1g1 e7f5 a2a4 b7b6 f1e1 f8e7 b1d2 a7a5 d2f1 e8g8 f1e3 f5e3 e1e3 a8c8 c1d2 c6e8 h2h3 g8h8 e3e1 e8c6 d2e3 c6d7 e3f4 d7c6');
INSERT INTO evaluations VALUES(34930,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(34930,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(34930,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(34930,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(34930,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(34930,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(34930,36,'cp 23','e5d4 d1d4 d5e6 c1e3 b8c6 f1b5 c8d7 d4a4 g8f6 e1g1 f8d6 f1e1 e8g8 e3f4 e6f5 f4d6 c7d6 b1c3 a8e8 a1d1 h7h6 h2h3 a7a6 b5d3 f5c5 a4f4 e8e1 d1e1 d6d5 a2a3 b7b5 e1d1 f8e8 d3f1 g8f8 f3d4 c6d4 f4d4 c5d4 d1d4 d7f5 f1d3');
INSERT INTO evaluations VALUES(34906,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(34906,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(34906,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(34906,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(34906,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(34906,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(34906,36,'cp 33','d1d4 d5e6 c1e3 b8c6 f1b5 c8d7 d4a4 g8f6 e1g1 f8d6 f1e1 e8g8 e3g5 e6f5 b1c3 h7h6 g5f6 f5f6 c3e4 f6f5 e4d6 c7d6 c2c3 a8e8 a1d1 f5c5 a4c4 e8e1 f3e1 d7e6 c4a4 f8b8 b5e2 b7b5 a4a6 e6d5 e2f3 d5f3 e1f3');
INSERT INTO evaluations VALUES(41999,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(41999,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(41999,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(41999,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(41999,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(41999,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(41999,36,'cp 54','b2c3 e8g8 e2e3 c7c5 c4d5 e6d5 d4c5 c8e6 d1d4 b8d7 f1b5 d8c7 e1g1 h7h6 b5d7 f6d7 g5e7 f8c8 e7d6 c7d8 f3e5 d7e5 d6e5 f7f6 e5g3 b7b6 c5b6 d8b6 f1b1 b6c6 b1b3 c6d7 h2h3 c8c4 d4d2 a8c8 a1b1 d7c6 b1c1 e6d7 b3b1 c4a4 d2d1 d7e6 g3f4 a4a3');
INSERT INTO evaluations VALUES(45092,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(45092,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(45092,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(45092,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(45092,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(45092,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(45092,36,'cp 71','d6d5 e2e3 g7g6 h2h3 f8g7 f1d3 e8g8 e1g1 c8f5 d3f5 g6f5 b2b3 e7e6 c1b2 b8d7 a1c1 f6e4 c3e2 f8e8 c1c2 d7f8 e2f4 f8g6 f4h5 g8h8 c4d5 c6d5 f3e5 g7e5 d4e5 d8g5 f2f4 g5h4 b2d4 e8c8 c2c8 a8c8 d4a7 g6e7 a7b6 e7c6 a2a3 c6e7');
INSERT INTO evaluations VALUES(46067,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46067,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(46067,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(46067,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(46067,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(46067,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(46067,36,'cp 26','e2e4 d5c3 b2c3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7c5 f1d1 e6d5 e4d5 d8f6 d1e1 f6d6 f3e5 c8d7 e1e3 d7a4 a2a3 a8c8 d2d4 c5d7 c4a6 c8c5 e5d3 a4b5 a6b5 c5b5 d3b4 d7f6 a1e1 f8c8 g2g3 b5c5 g1g2');
INSERT INTO evaluations VALUES(46071,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(46071,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(46071,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(46071,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(46071,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(46071,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(46071,36,'cp 38','d5c3 b2c3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7f6 d2f4 e6d5 a1d1 c8d7 e4d5 b6b5 c4b3 a7a5 d5d6 a5a4 b3c2 d8b6 d1d4 a8c8 c2f5 d7f5 f4f5 c8c4 f5f4 h7h6 d4c4 b5c4');
INSERT INTO evaluations VALUES(46064,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(46064,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46064,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(46064,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(46064,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(46064,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(46064,36,'cp 42','b2c3 c5d4 c3d4 f8b4 c1d2 b4d2 d1d2 e8g8 f1c4 b8d7 e1g1 b7b6 d4d5 d7c5 f1e1 e6d5 e4d5 d8d6 f3e5 c8d7 d2d4 d7a4 e1e3 c5d7 f2f4 a8c8 c4a6 d6c5 e3d3 c8d8 e5d7 d8d7 d4c5 b6c5 a6c4 d7b7 c4b3 a4b3 a2b3 f7f6 d5d6 b7d7 g1f2 f8d8 a1a6 g8f7');
INSERT INTO evaluations VALUES(51439,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(51439,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(51439,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(51439,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(51439,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(51439,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(51439,36,'cp 159','d7d6 d2d4 e5d4 d1d4 d8d7 c1f4 b8c6 d4e3 c8b7 b1c3 e8c8 e1c1 f7f6 h2h4 g8e7 h4h5 c8b8 f4g3 g7g5 f3d4 c6d4 e3d4 e7c6 d4f6 f8g7 f6g5 g7c3 b2c3 c6a5 c4d5 d7a4 d5b7 a5b7 g5d5 a4a3 c1b1 b7a5 d1d4 h8e8 h1e1');
INSERT INTO evaluations VALUES(51344,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(51344,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(51344,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(51344,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(51344,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(51344,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(51344,36,'cp 165','e1g1 f8e7 d2d4 e5d4 c2c3 g8f6 c3d4 c7c6 b1c3 d6d5 e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 f3e5 d8c8 d1a4 e8g8 e5d7 c8d7 a4d7 f6d7 c3d5 e7d8 d5c3 d8e7 a2a3 f8d8 c1e3 d7f6 g2g3 f6d5 a1c1 d5c3 c1c3 e7f6 f1d1 h7h6 c3c7 d8c8 c7c1 f6g5 a3a4 g5e3 f2e3');
INSERT INTO evaluations VALUES(55580,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(55580,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(55580,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(55580,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(55580,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(55580,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(55580,36,'cp 9','e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 c7c5 f1e2 b8c6 d2b3 b7b6 f2f3 f8e7 e1g1 e8g8 f3f4 c6b8 a1b1 b8a6 a2a3 a6c7 d4d5 c8d7 b3d2 h7h5 d1c2 d8e8 b2b4 f6g4 d2b3 e7h4 h2h3 g4f6 a3a4 e8g6 b4c5 d6c5');
INSERT INTO evaluations VALUES(9523,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9523,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9523,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9523,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9523,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9523,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9523,36,'cp 0','d2d4 f6e4 f1e1 f7f5 f3e5 e8g8 f2f3 d6e5 d4e5 e4c5 c1e3 c5e6 d1e2 b7b6 c2c4 e6d4 e2f2 d4e6');
INSERT INTO evaluations VALUES(9516,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9516,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9516,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9516,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9516,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9516,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9516,36,'cp 0','f6e4 f1e1 f7f5 f3e5 e8g8 f2f3 d6e5 d4e5 e4c5 c1e3 c5e6 d1e2 b7b6 c2c4 e6d4 e2f2 d4e6 f2c2 e6d4');
INSERT INTO evaluations VALUES(10616,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(10616,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(10616,32,'cp 282',NULL);
INSERT INTO evaluations VALUES(10616,33,'cp 265',NULL);
INSERT INTO evaluations VALUES(10616,34,'cp 266',NULL);
INSERT INTO evaluations VALUES(10616,35,'cp 274',NULL);
INSERT INTO evaluations VALUES(10616,36,'cp 286','d3d4 c5d6 c4f7 e8f7 f3g5 f7f8 d1g4 h7h6 g4f5 d8f6 g5h7 h8h7 f5h7 e5d4 e1g1 b7b6 f2f4 f6e6 e4e5 c8a6 f1f3 a8e8 b2b3 d4c3 c1e3 d6b4 e3f2 a6e2 f3g3 e2g4 a2a3 c3c2 h7c2 b4e7 b1c3 g4f5 c2b2 e7h4 g3f3 h4f2 f3f2 c6d4 b3b4 c7c5 a1d1 f8g8 h2h3 f5h3');
INSERT INTO evaluations VALUES(10629,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(10629,31,'cp 285',NULL);
INSERT INTO evaluations VALUES(10629,32,'cp 261',NULL);
INSERT INTO evaluations VALUES(10629,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(10629,34,'cp 247',NULL);
INSERT INTO evaluations VALUES(10629,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(10629,36,'cp 259','c5d6 c4f7 e8f7 f3g5 f7f8 d1g4 f8g8 g4d1 d8e7 d1b3 g8f8 d4d5 c6a5 b3c2 h7h6 g5f3 b7b6 e1g1 f8f7 b1d2 h8e8 b2b4 a5b7 d2c4 f7g8 a2a3 a7a5 c1e3 e7f6 f3d2 f6g6 c4d6 b7d6 f2f3 d6f7 f1f2 d7d6 a1e1 c8d7 d2f1 e8f8 c3c4 f7g5 h2h4');
INSERT INTO evaluations VALUES(11317,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11317,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(11317,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11317,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(11317,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(11317,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(11317,36,'cp 22','c8f5 f1e2 e7e6 b1d2 f8d6 f4g3 e8g8 e1g1 c6e7 c3c4 a6a5 d1a4 c7c6 f1c1 f6e4 d2e4 f5e4 f3d2 e4g6 g3d6 d8d6 c4c5 d6c7 b2b4 a5b4 a4b4 e6e5 a2a4 e5d4 b4d4 h7h6 d2b3 e7f5 d4b2 c7e7 a4a5');
INSERT INTO evaluations VALUES(24267,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(24267,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(24267,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(24267,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24267,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(24267,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(24267,36,'cp 37','b1d2 e7e6 f1e2 f8d6 f4g3 e8g8 e1g1 a6a5 c3c4 c6e7 a2a3 b7b6 f3h4 f5e4 a1c1 h7h6 d1a4 c7c5 h4f3 e4f5 a4b5 a8c8 g3d6 d8d6 c4d5 e6d5 d4c5 c8c5 c1c5 d6c5 d2b3 c5b5 e2b5 f8c8 b3d4 f5d7 b5d7');
INSERT INTO evaluations VALUES(24266,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24266,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24266,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24266,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24266,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(24266,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(24266,36,'cp 25','e7e6 f1e2 f8d6 f4g3 e8g8 e1g1 a6a5 a2a3 c6e7 c3c4 b7b6 f3h4 f5e4 c4d5 d6g3 h2g3 e6d5 h4f3 c7c5 a1c1 e4f3 e2f3 c5d4 e3d4 d8d6 d2b1 b6b5 b1c3 b5b4 c3b5 d6b6 a3a4 a8c8 f1e1 e7f5 c1c8 f8c8 g3g4 f5h4 f3e2 h7h6 e2d3 g7g6 g2g3');
INSERT INTO evaluations VALUES(24289,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(24289,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(24289,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24289,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(24289,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(24289,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(24289,36,'cp 36','f1e2 f8d6 f4g3 e8g8 e1g1 a6a5 c3c4 c6e7 a2a3 b7b6 f3h4 f5e4 a1c1 h7h6 d1a4 c7c6 c4d5 c6d5 h4f3 e4f3 g2f3 e7f5 c1c6 f6h5 f3f4 h5f6 e2d3 f5e7 c6c2 a8c8 f1c1 c8c2 c1c2 e7f5 d2f3 f6e4');
INSERT INTO evaluations VALUES(24290,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(24290,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(24290,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(24290,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24290,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(24290,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(24290,36,'cp 23','f8d6 f4g3 c6e7 c3c4 e8g8 e1g1 b7b6 g3e5 c7c6 a1c1 a6a5 d1a4 h7h6 f1d1 d8b8 h2h3 f6d7 e5d6 b8d6 a2a3 d6b8 c1c3 a8a7 e2f1 f8d8 c3c1 f5h7 a4b3 d8c8 c4d5 c6d5 c1c8 b8c8 d1a1');
INSERT INTO evaluations VALUES(21758,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(21758,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(21758,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(21758,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(21758,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(21758,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(21758,36,'cp -25','d3d4 e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 c8e6 a2a3 a7a6 c4d5 f6d5 f1c4 d8d7 c3d5 e6d5 c4d5 d7d5 d1b3 d5b3 d2b3 a6a5 a3a4 b7b6 e1e2 e8c8 c1d2 f8d6 d2c3 c6e7 b3d2 e7d5 h2h4 c8b7 g2g3 d6e7 a1c1 e7f6 d2c4 c7c6 c3d2 d8d7');
INSERT INTO evaluations VALUES(22577,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(22577,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(22577,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(22577,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(22577,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(22577,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(22577,36,'cp -31','g8f6 c4b3 d7d5 e4d5 e6d5 d2d4 f8b4 g1e2 c6a5 e1g1 c7c6 e2g3 e8g8 c3e2 a5b3 a2b3 f8e8 c2c3 b4f8 b3b4 b7b6 e2f4 a7a5 f4d3 c8a6 b4a5 b6a5 f1e1 a6d3 e1e8 f6e8 d1d3 d8b6 d3c2 b6b5 c1f4 a5a4 h2h3');
INSERT INTO evaluations VALUES(28548,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(28548,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(28548,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(28548,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(28548,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(28548,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(28548,36,'cp 12','e1g1 f8e7 f2f4 e8g8 g1h1 h7h6 c1e3 e5f4 e3f4 b8c6 d1e1 f6d7 a1d1 c6e5 e1g3 e7h4 g3e3 d8e7 b3d4 a8c8 d4f5 e6f5 e4f5 f8e8 e3d4 c8c6 f4e5 d7e5 c3d5 e7d8 c2c3 h4g5 d4a7 b7b5 a7f2 c6c8 f5f6 g5f6 d5f6 d8f6 f2f6 g7f6 f1f6');
INSERT INTO evaluations VALUES(28547,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(28547,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(28547,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28547,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(28547,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(28547,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(28547,36,'cp 9','f8e7 d1d3 b8c6 a2a3 e8g8 c1e3 a8c8 h2h3 h7h6 a1c1 d6d5 e4d5 f6d5 c3d5 d8d5 d3d5 e6d5 f1d1 d5e6 b3c5 e7c5 e3c5 f8d8 d1d8 c8d8 c1d1 d8d1 e2d1 f7f5 c5b6 g8f7 f2f3 f7f6 d1e2 f5f4 g1f2 e6f5 c2c3 c6e7 e2f1 f6e6 f1c4 e7d5 b6d8 b7b5');
INSERT INTO evaluations VALUES(28186,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(28186,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(28186,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28186,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28186,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(28186,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(28186,36,'cp 14','c1e3 e8g8 d1d3 b8d7 a2a4 a8c8 f1d1 d8c7 a4a5 h7h6 e2f3 d7c5 b3c5 d6c5 c3a4 c7a5 a4c5 a5b4 c5e6 f7e6 a1b1 f8d8 d3e2 d8d1 e2d1 e7c5 c2c3 b4b6 e3c5 c8c5 h2h3 c5c8 d1d2 c8d8 d2c2 b6c7 f3e2 d8d6 b1e1 c7d8 e2f1 g8f7 b2b4 d6d2 c2b3 b7b5');
INSERT INTO evaluations VALUES(29530,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(29530,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(29530,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(29530,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(29530,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(29530,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(29530,36,'cp -23','e7e6 g2g4 f5g6 h2h4 h7h5 g4g5 f6d7 e2e4 d7b6 f1d3 b8c6 g1e2 f8d6 f4d6 d8d6 d1d2 d5e4 d3e4 e8g8 e1c1 b6c4 c3b5 c4d2 b5d6 d2e4 d6e4 c6b4 e2c3 b7b6 h1e1 f8d8 a2a3 b4d5 c3d5 d8d5 e4c3 d5f5 e1e3 a8d8 c3e2 b6b5 b2b3 c7c6');
INSERT INTO evaluations VALUES(30082,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(30082,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(30082,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(30082,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(30082,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(30082,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(30082,36,'cp -34','g2g4 f5g6 h2h4 h7h5 g4g5 f6d7 e2e4 d5e4 c3e4 b8c6 g1e2 d7b6 c2c3 a7a6 f4g3 b6d5 e2f4 d5f4 g3f4 d8d5 f1g2 e8c8 d1b3 e6e5 b3d5 d8d5 d4e5 c6e5 f4e5 d5e5 e1f2 g6e4 f3e4 c7c6 a1f1 f7f6 g5f6 g7f6 f1d1 c8c7 f2f3 h8g8 g2h3 f8d6 f3e3 e5a5');
INSERT INTO evaluations VALUES(30106,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(30106,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(30106,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(30106,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(30106,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(30106,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(30106,36,'cp -37','f5g6 h2h4 h7h5 g4g5 f6d7 e2e4 d5e4 c3e4 b8c6 g1e2 d7b6 c2c3 a7a6 f4g3 b6d5 d1d2 f8d6 f1h3 e8g8 e1g1 b7b5 a1c1 c6e7 b2b3 d6g3 e4g3 c7c5 c3c4 d5b6 f1d1 c5d4 d2d4 b5c4 b3c4 d8c7 d4d6 c7a7 c4c5 b6d5 e2f4');
INSERT INTO evaluations VALUES(30219,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(30219,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(30219,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(30219,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(30219,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(30219,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(30219,36,'cp -11','h2h4 h7h5 g4g5 f6d7 e2e4 d5e4 c3e4 b8c6 f1b5 c6b4 g1e2 a7a6 b5d3 c7c5 c2c3 b4d5 e4d6 f8d6 f4d6 g6d3 d1d3 d8b6 d4c5 d7c5 d6c5 b6c5 e1c1 e8c8 d3d4 c5d4 d1d4 d5e7 d4e4 d8d7 e2d4 h8d8 h1e1 e7g6 c1c2 e6e5 d4b3 d7d3');
INSERT INTO evaluations VALUES(30220,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(30220,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(30220,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(30220,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(30220,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(30220,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(30220,36,'cp -15','h7h5 g4g5 f6d7 e2e4 d5e4 c3e4 b8c6 c2c3 e6e5 d4e5 d7e5 d1d8 a8d8 f1e2 g6e4 f3e4 f8d6 a1d1 e5d3 e2d3 d6f4 g1f3 e8g8 e1g1 f8e8 d3c2 g7g6 d1d8 e8d8 g1g2 c6a5 b2b3 a5c6 f3d4 f4e5 d4c6 b7c6 f1d1 d8d1 c2d1 e5g7 g2h3 g7c3 d1f3 f7f6');
INSERT INTO evaluations VALUES(30032,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(30032,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(30032,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(30032,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(30032,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(30032,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(30032,36,'cp -17','f8c5 g1f3 b8c6 f1e2 d8e7 e1g1 e8c8 c1d2 e6f5 d2f4 e7d7 b1c3 f5d3 d1d3 d7d3 e2d3 d8d3 a1d1 h8d8 d1d3 d8d3 f1e1 d3d8 a2a3 c6d4 f3g5 d8d7 g1f1 h7h6 g5e4 f6e4 c3e4 c5e7 e1d1 d4e6 d1d7 c8d7 f4d2 f7f5 e4c3 e7g5 f1e2 d7d6 b2b4');
INSERT INTO evaluations VALUES(28471,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(28471,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(28471,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(28471,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(28471,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(28471,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(28471,36,'cp -19','g1f3 b8c6 f1e2 d8e7 c1d2 e8c8 e1g1 e6f5 d2f4 h7h6 b1c3 g7g5 f4g3 h8e8 f1e1 e7d7 a1c1 f5d3 d1d3 d7d3 e2d3 d8d3 c3d5 f6e4 d5c7 c5f2 g3f2 c8c7 f2g3 c7d7 f3e5 c6e5 e1e4 f7f6 c1f1 e8e6 g3e5 e6e5 e4e5 f6e5 f1f7 d7c6 b2b3 d3d1 g1f2 d1d2 f2f3 d2a2 f7f6 c6c5 f6h6');
INSERT INTO evaluations VALUES(35061,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(35061,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(35061,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(35061,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(35061,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(35061,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(35061,36,'cp 40','g2g3 h4d8 d1e2 f8e7 g1f3 d7d5 d2d3 c7c5 f1g2 b8c6 e1g1 g8h6 c2c3 e8g8 b1a3 f7f6 a3c2 d5d4 a2a4 e6e5 c3d4 e5d4 c1d2 h6f7 f1b1 f8e8 b2b4 a7a5 b4b5 c6b4 c2a3 c8g4 h2h3 g4e6 g1h2 g8h8 f4f5 e6d7 a3c4 f7d6 b1d1 d6c4');
INSERT INTO evaluations VALUES(37754,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(37754,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(37754,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(37754,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(37754,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(37754,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37754,36,'cp -4','d4c5 g7e5 g1f3 e5g7 b1c3 d8a5 f1d3 g8f6 e1g1 b8c6 c3b5 a7a6 c1d2 a5d8 b5d4 c6d4 f3d4 d7d6 c5d6 d8d6 d4f3 e8g8 f1e1 f6d5 c2c3 c8e6 d3f1 f8d8 f3g5 e6f5 d1b3 b7b5 a2a4 b5a4 b3a4 h7h6 g2g4 d5b6 a4b3 h6g5 g4f5 d6d2');
INSERT INTO evaluations VALUES(48641,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(48641,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(48641,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(48641,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(48641,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(48641,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(48641,36,'cp 142','c1g5 d5e4 c3e4 f8b4 c2c3 b4e7 e4f6 e7f6 f1b5 b8d7 d1e2 d8e7 g5f6 e7e2 d4e2 g7f6 e1c1 c7c6 b5c4 d7e5 c4b3 c8g4 f2f3 g4e6 h1e1 e6b3 a2b3 a8d8 e2d4 e8g8 f3f4 e5g6 g2g3 d8d5 c3c4 d5c5 d4f3 b7b5 d1d4 a7a5 c1c2 f8b8 c2c3 b5c4 b3c4 g8g7');
INSERT INTO evaluations VALUES(50344,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(50344,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50344,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(50344,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50344,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50344,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50344,36,'cp 0','d7d5 g1f3 e7e6 e2e3 b8d7 c2c4 f8e7 b2b3 e8g8 b1d2 c7c5 c4d5 f6d5 c1b2 b7b6 f1e2 c8b7 e1g1 a8c8 a1c1 e7f6 d2e4 f6e7');
INSERT INTO evaluations VALUES(53731,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(53731,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(53731,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(53731,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(53731,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(53731,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(53731,36,'cp -16','d2d4 a7a5 c2c3 f8d6 g1f3 g8f6 d1e2 d8e7 e2e7 e8e7 e1g1 h8e8 f1e1 e7f8 e1e8 f6e8 a2a4 f7f6 b1a3 c8f5 h2h4 b8d7 a3c2 d6c7 c2e3 f5g6 f3e1 e8d6 g2g3 f8f7 b3d1 h7h5 e1g2 d6e4 d1c2 a8e8 g2f4 c7f4 g3f4 e4d6 f4f5');
INSERT INTO evaluations VALUES(284,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(284,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(284,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(284,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(284,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(284,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(284,36,'cp 177','b1c3 g8e7 c3d5 e7d5 e4d5 e5e4 d1e2 e8g8 e2e4 c6d4 f3d4 f6d4 d2d3 d7d6 e4e7 d4e5 f1e1 e5f5 e7h4 f7f6 a2a4 c8d7 c1e3 c5e3 f2e3 g7g5 h4d4 f5e5 d4e5 f6e5 c4b5 c7c6 b5c4 g8g7 d5c6 b7c6 e1f1 h7h5 b2b4 a8b8 f1f8 g7f8');
INSERT INTO evaluations VALUES(276,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(276,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(276,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(276,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(276,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(276,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(276,36,'cp 202','g8e7 c3d5 e7d5 e4d5 e5e4 d1e2 e8g8 e2e4 c6d4 f3d4 f6d4 d2d3 d7d6 e4e7 d4e5 f1e1 e5f5 e7h4 f7f6 h4f4 f5h5 d3d4 c5b6 a2a4 a7a6 c1e3 g7g5 f4g3 c8f5 a4a5 b6a7 c4b3 a8e8 c2c4 h5g6 c4c5 h7h5 e3d2 e8e1 a1e1 f8e8 g3c3 e8e1 d2e1');
INSERT INTO evaluations VALUES(7592,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7592,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(7592,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7592,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(7592,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7592,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(7592,36,'cp 4','h7h6 g5h4 e8g8 f1d3 a8b8 e1g1 d7d5 f1e1 d5e4 c3e4 f6e4 e1e4 d8d5 b2b3 c8f5 e4c4 f5d3 d1d3 c5b6 h4g3 f8e8 d3d5 c6d5 c4c6 e8e6 c6e6 f7e6 g1f1 g8f7 a1e1 b8c8 f1e2 b6d4 e2d3 d4f6 c2c4 c7c5 d3c2 a7a5 h2h3 h6h5');
INSERT INTO evaluations VALUES(7566,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7566,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7566,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7566,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(7566,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7566,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(7566,36,'cp 11','g5h4 e8g8 f1d3 d7d5 e1g1 a8b8 e4d5 c6d5 c3d5 g7g5 d5f6 d8f6 h4g3 b8b2 d1e2 f6c6 f1e1 c8b7 e2f1 c5b4 e1e4 b7a6 f1e2 a6d3 e2d3 b4d6 e4c4 c6b5 g3d6 c7d6 h2h4 d6d5 c4g4 b5d3 c2d3 f7f5 g4a4 f8f7 h4h5 g8g7 a4a5 d5d4');
INSERT INTO evaluations VALUES(7570,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7570,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(7570,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(7570,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7570,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(7570,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(7570,36,'cp 8','d7d5 f1d3 e8g8 e1g1 a8b8 f1e1 d5e4 c3e4 f6e4 e1e4 d8d5 b2b3 c8f5 e4e1 c5b4 d3f5 d5f5 e1e3 f8e8 d1d3 f5d3 c2d3 f7f5 h4g3 b4d6 e3e8 b8e8 g3d6 c7d6 a1c1 g8f7 g1f1 c6c5 c1c4 f7f6 g2g3 g7g5 b3b4 c5b4 c4b4 e8e7 b4a4 g5g4 a4a5 h6h5 a2a4 e7c7');
INSERT INTO evaluations VALUES(23018,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(23018,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(23018,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(23018,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(23018,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(23018,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(23018,36,'cp 13','c7c6 b5c4 g8f6 e1g1 d7d5 e4d5 f6d5 d2d3 c8e6 b1d2 f8e7 d2b3 b7b5 c4d5 e6d5 f3g4 e8g8 b3d4 e7f6 c1e3 h7h5 g4h5 d5g2 g1g2 f6d4 a1b1 f8e8 f1e1 d4f6 h5f3 d8d7 e3c5 a7a5 g2f1 a5a4 b2b3 f6c3 e1e8 a8e8 b3a4 b5a4');
INSERT INTO evaluations VALUES(19099,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(19099,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(19099,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19099,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19099,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19099,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(19099,36,'cp 17','b5c4 g8f6 e1g1 d7d5 e4d5 f6d5 f3e4 c8e6 d2d3 f8c5 b1d2 e8g8 d2b3 d8d6 b3c5 d6c5 e4e5 f8e8 c1d2 f7f6 e5g3 d5b6 c4e6 e8e6 a1e1 e6e8 h2h3 g8h8 c2c4 d4c3 b2c3 c5d5 e1e8 a8e8 f1e1 e8e1 d2e1 d5e5 g3e5 f6e5 c3c4 b6d7 g1f1 h8g8 f1e2 b7b6 e2e3 g8f7 f2f4 f7e6 e1b4 e5f4 e3f4 d7e5 f4e4');
INSERT INTO evaluations VALUES(19108,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(19108,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(19108,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19108,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(19108,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(19108,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(19108,36,'cp 19','g8f6 e1g1 d7d5 e4d5 f6d5 d2d3 c8e6 b1d2 f8e7 d2b3 b7b5 c4d5 e6d5 f3g4 e8g8 b3d4 e7f6 c1e3 h7h5 g4h5 d5g2 g1g2 f6d4 a1b1 f8e8 f1e1 a7a5 e3d4 d8d4 h5f3 d4c5 c2c3 a8d8 d3d4 c5g5 g2h1 g5g6 b1c1 g6h6 e1e3 e8e3');
INSERT INTO evaluations VALUES(34205,30,'cp 220',NULL);
INSERT INTO evaluations VALUES(34205,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(34205,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(34205,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(34205,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(34205,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(34205,36,'cp 239','d2d4 d7d6 c1g5 f6g6 d4e5 g6e4 c4e2 b8d7 b1c3 e4g6 d1d2 h7h6 g5f4 d6e5 f4e5 d7e5 f3e5 g6d6 e1c1 d6d2 d1d2 c8e6 e2c4 g8f6 h1e1 f8b4 e5f7 e8f7 e1e6 f7f8 d2e2 a8e8 e6e8 f6e8 c3e4 e8d6 c2c3 d6e4 c3b4 e4d6 c4b3 h6h5 e2e6 d6f5 e6e5 f5d6 e5a5 h8h6 b3c2 h6f6 f2f3 f6f4 a5h5 f4b4 c2d3 f8e7 h5h8 e7d7 h8g8');
INSERT INTO evaluations VALUES(45642,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45642,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45642,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45642,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45642,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45642,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(45642,36,'cp 0','d5c4 d3c4 b7b5 c4d3 c8b7 g1f3 a7a6 d1e2 b8d7 e3e4 b5b4 e4e5 b4c3 e5f6 c6c5 b2c3 c5d4 a1b1 b7f3 e2f3 d7e5 f3e2 e5d3 e2d3 d8f6 d3d4 f6d4 c3d4 f8d6 e1e2 a8c8 c1d2 e8d7 b1b7 c8c7 h1b1 h8c8 e2d3 c7b7 b1b7 c8c7 b7b6 c7c6 b6b3 c6c8');
INSERT INTO evaluations VALUES(54042,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(54042,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(54042,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(54042,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(54042,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(54042,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54042,36,'cp 0','c8d7 b5d7 d8d7 e4d5 d7d5 b1c3 d5d7 d2d4 c5d4 d1d4 d7d4 f3d4 a7a6 c1e3 g8f6 e1c1 f8b4 e3d2 b8d7 h1e1 b4c5 d2e3 e8g8 d4b3 c5e3 e1e3 f8c8 a2a4 c8c7 e3e2 a8c8 a4a5 h7h6 g2g3 d7c5 b3d2 c5d7');
INSERT INTO evaluations VALUES(30885,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(30885,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(30885,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30885,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30885,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(30885,36,'cp 5','b5d7 d8d7 e4d5 d7d5 b1c3 d5d7 f3e5 d7c7 d2d4 c5d4 d1d4 b8c6 e5c6 c7c6 c1g5 f7f6 g5e3 g8e7 e1c1 e7f5 d4d3 f8e7 f2f3 e8f7 e3f2 h8d8 d3e4 c6e4 c3e4 f5d6 e4d6 d8d6 c2c3 h7h5 c1c2 e6e5 a2a4 f6f5 h1e1 d6d1 e1d1 f7e6 b2b4 g7g5 h2h3 a8c8 c2b3');
INSERT INTO evaluations VALUES(30856,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30856,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30856,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30856,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30856,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30856,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30856,36,'cp 0','d8d7 e4d5 d7d5 b1c3 d5d6 c3e4 d6d5');
INSERT INTO evaluations VALUES(54840,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(54840,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(54840,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(54840,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(54840,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(54840,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(54840,36,'cp 60','d2d4 f8g7 b1c3 g8f6 g1f3 c7c5 d4c5 d8a5 f1d3 a5c5 d1e2 e8g8 c1e3 c5a5 e1g1 c8g4 h2h3 g4f3 e2f3 b8c6 g1h1 f6d7 c3d5 e7e6 d5c3 g7c3 b2c3 a5c3 f4f5 c6e5 f3g3 e5d3 c2d3 e6f5 e4f5 f8e8 e3g5 f7f6 g3d6 a8d8 g5h6');
INSERT INTO evaluations VALUES(56047,30,'cp -209',NULL);
INSERT INTO evaluations VALUES(56047,31,'cp -207',NULL);
INSERT INTO evaluations VALUES(56047,32,'cp -198',NULL);
INSERT INTO evaluations VALUES(56047,33,'cp -194',NULL);
INSERT INTO evaluations VALUES(56047,34,'cp -198',NULL);
INSERT INTO evaluations VALUES(56047,35,'cp -196',NULL);
INSERT INTO evaluations VALUES(56047,36,'cp -211','d8h4 e1d2 g8f6 g1f3 h4e4 b1c3 e4g4 d1e1 f8d6 d2d1 e8g8 c3b5 b8c6 c1g5 d6f4 h2h3 g4g3 e1g3 f4g3 b5c3 f6h5 f1b5 h7h6 g5d2 h5f6 b5c6 b7c6 d1c1 c8a6 b2b3 g3d6 h1d1 c6c5 d4c5 d6c5 c1b2');
INSERT INTO evaluations VALUES(34844,30,'cp -199',NULL);
INSERT INTO evaluations VALUES(34844,31,'cp -204',NULL);
INSERT INTO evaluations VALUES(34844,32,'cp -210',NULL);
INSERT INTO evaluations VALUES(34844,33,'cp -208',NULL);
INSERT INTO evaluations VALUES(34844,34,'cp -190',NULL);
INSERT INTO evaluations VALUES(34844,35,'cp -193',NULL);
INSERT INTO evaluations VALUES(34844,36,'cp -206','e1d2 h4e4 g1f3 g8f6 b1c3 e4f4 d2e1 f4d6 c3b5 d6d8 c1f4 f6d5 f4g3 c7c6 b5c3 f8b4 d1d2 d8a5 a2a3 d5c3 a1c1 b4e7 b2c3 e8g8 f1d3 c6c5 e1f2 b8d7 h1e1 d7f6 f2g1 c5d4 c3d4 a5d2 f3d2 f8d8 d2e4 f6e4 d3e4 e7a3');
INSERT INTO evaluations VALUES(56829,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(56829,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(56829,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(56829,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(56829,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(56829,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(56829,36,'cp 28','d2d4 e5e4 c2c4 c7c6 b1c3 f6f5 c4d5 c6d5 g1e2 b8c6 e2f4 g8f6 a1c1 a7a6 f1e2 f8b4 a2a3 b4a5 b3b4 a5c7 f4h5 f6h5 e2h5 g7g6 h5e2 c6e7 d1b3 c8e6 a3a4 e8g8 e1g1 c7d6 b2a3 g6g5 h2h3 a8c8 b4b5 d6a3 b3a3 f5f4 b5a6 b7a6');
INSERT INTO evaluations VALUES(2469,30,'cp -464',NULL);
INSERT INTO evaluations VALUES(2469,31,'cp -491',NULL);
INSERT INTO evaluations VALUES(2469,32,'cp -486',NULL);
INSERT INTO evaluations VALUES(2469,33,'cp -494',NULL);
INSERT INTO evaluations VALUES(2469,34,'cp -475',NULL);
INSERT INTO evaluations VALUES(2469,35,'cp -487',NULL);
INSERT INTO evaluations VALUES(2469,36,'cp -498','b1c3 d7d5 e4d5 f6d5 e1g1 d5c3 b2c3 e5e4 f3e1 d8d5 c3c4 d5e5 a1b1 f8d6 f2f4 e5f5 d2d4 c6d4 c1e3 d4c6 b1b5 f5f6 d1d5 f6e6 d5h5 g7g6 h5e2 f7e8 f4f5 g6f5 b5f5 h8g8 e3c5 c6e7 e2h5 e6g6 f5g5 g6h5 g5h5 c8g4 c5d6 g4h5');
INSERT INTO evaluations VALUES(20852,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(20852,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(20852,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(20852,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(20852,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(20852,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(20852,36,'cp 165','e2e4 d5e4 c3e4 e7e5 f1b5 d8d5 d1e2 c8f5 e4g3 f5g4 c2c4 g4f3 c4d5 f3e2 e1e2 a7a6 d5c6 a6b5 c6b7 a8b8 h1d1 e5d4 d1d4 g8e7 c1e3 c7c6 d4d6 b8b7 g3e4 b7c7 a1c1 h7h5 g2g3 e8f7 h2h4 c7c8 d6d7 f7g6 e4d6 c8a8 a2a3 a8a6 e3d4 h8g8 e2d3 a6a8 d4e3 a8a6 d7d8 g6h7 e3d4 e7g6 d6f5 g6e5 d4e5 f6e5 d3e4 a6a4 e4d3');
INSERT INTO evaluations VALUES(20860,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(20860,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(20860,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(20860,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(20860,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(20860,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(20860,36,'cp 166','e7e6 a2a3 g8e7 f1d3 d5e4 c3e4 b7b6 e1g1 d8d7 f1e1 c8b7 d3c4 e7d5 b2b4 e8c8 c4b3 e6e5 c2c4 c6d4 c4d5 d4b3 d1b3 d7d5 b3a4 d5c4 e4d2 c4a6 a4a6 b7a6 d2f1 f8e7 c1b2 d8d7 a1c1 c8b7 b2c3 c7c5 b4c5 e7c5 c3b4 c5b4 a3b4 h8c8 c1c8 b7c8 f1e3 a6b7 e1a1 c8b8 g1f1 a7a6 a1c1 d7c7');
INSERT INTO evaluations VALUES(23086,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(23086,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(23086,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(23086,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(23086,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(23086,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(23086,36,'cp 55','e5c6 b1c3 f8b4 g1e2 g8f6 a2a3 b4a5 b2b4 a5b6 c1b2 e8g8 d1d3 f8e8 e1c1 d7d5 c3d5 f6e4 e2g3 b6d4 g3e4 d8d5 b2d4 d5e4 d3e4 e8e4 d4f2 e4f4 f2g3 f4e4 d1d2 e4e8 f1e2 c8f5 h1f1 g7g6 e2f3 f5e4 g3c7 e4f3 f1f3 f7f5 d2d7 e8e7 d7e7 c6e7');
INSERT INTO evaluations VALUES(22960,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22960,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(22960,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(22960,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(22960,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(22960,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(22960,36,'cp 73','b1c3 f8b4 g1e2 g8f6 a2a3 b4a5 b2b4 a5b6 c1b2 e8g8 d1d3 f8e8 e1c1 d7d5 c3d5 f6e4 d3f3 d8h4 g2g3 h4g4 f3g2 g4g6 e2c3 c8f5 f1d3 e4d6 d5b6 c7b6 c3d5 a8c8 d3f5 d6f5 d1d3 c6e7 d5c3 g6c6 g2c6 e7c6 d3d7 c6d4 h1d1 e8d8 d7b7 f5e3 d1d2 e3c4');
INSERT INTO evaluations VALUES(23322,30,'cp 655',NULL);
INSERT INTO evaluations VALUES(23322,31,'cp 663',NULL);
INSERT INTO evaluations VALUES(23322,32,'cp 663',NULL);
INSERT INTO evaluations VALUES(23322,33,'cp 671',NULL);
INSERT INTO evaluations VALUES(23322,34,'cp 649',NULL);
INSERT INTO evaluations VALUES(23322,35,'cp 660',NULL);
INSERT INTO evaluations VALUES(23322,36,'cp 650','f3e5 g8e7 e5f3 d7d5 b1c3 d5e4 d1d8 e8d8 c3e4 c8f5 e4c3 d8d7 f3d4 a8e8 c1e3 d7c8 f1c4 f5g4 c4f7 e8d8 f2f3 g4d7 e1c1 e7f5 d4f5 d7f5 d1d8 c8d8 h1d1 f8d6 c3b5 d8e7 b5d6 c7d6 f7d5 h8c8 c2c3 f5e6 d5b7 c8c7 b7e4 g7g6 e3f4 c7d7 e4c2 d6d5 d1d4 d7d8 f4g3');
INSERT INTO evaluations VALUES(23321,30,'cp 636',NULL);
INSERT INTO evaluations VALUES(23321,31,'cp 644',NULL);
INSERT INTO evaluations VALUES(23321,32,'cp 648',NULL);
INSERT INTO evaluations VALUES(23321,33,'cp 637',NULL);
INSERT INTO evaluations VALUES(23321,34,'cp 637',NULL);
INSERT INTO evaluations VALUES(23321,35,'cp 637',NULL);
INSERT INTO evaluations VALUES(23321,36,'cp 641','g8e7 e5f3 d7d5 b1c3 d5e4 d1d8 e8d8 c3e4 c8f5 e4c3 d8d7 f3d4 a8e8 c1e3 d7c8 f1c4 f5d7 c4f7 e8d8 e1c1 e7f5 d4f5 d7f5 d1d8 c8d8 h1d1 d8c8 c3b5 f8e7 b5a7 c8b8 a7b5 g7g5 e3a7 b8c8 d1e1 e7d8 a7c5 f5d7 a2a4 b7b6 c5b4 c8b8 c2c4 g5g4 b4c3 h8f8 f7e6 d7e6 e1e6 f6f5 c1c2 f5f4 c2d3');
INSERT INTO evaluations VALUES(33196,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(33196,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(33196,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(33196,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(33196,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(33196,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(33196,36,'cp -24','f6b6 d3d4 e5d4 b1d2 a7a5 e4e5 b8c6 f1d3 d7d5 e5d6 c7d6 e1g1 h7h6 f1e1 c8e6 g5h4 g8e7 d2e4 e7d5 f3d2 e8g8 d2c4 b6c7 h4g3 c5b4 e1e2 c6e5 e4d6 e5d3 d1d3 c7c6 a2a4 b4d6 c4d6 f8d8 d6b5 d5b4 b5d4 b4d3 d4c6');
INSERT INTO evaluations VALUES(33155,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(33155,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(33155,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(33155,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(33155,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(33155,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(33155,36,'cp 0','d3d4 e5d4 b1d2 a7a5 f1d3 d7d6 e1g1 h7h6 g5h4 b8c6 e4e5 c8e6 e5d6 c7d6 d2e4 g8e7 f1e1 e8g8 e4f6 g8h8 f6h7 f8e8 h7f6 e8f8');
INSERT INTO evaluations VALUES(40677,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(40677,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(40677,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(40677,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(40677,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(40677,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(40677,36,'cp 65','d2d4 b8d7 a2a4 e8g8 e1g1 f8e8 f1e1 c7c6 a4a5 e5d4 f3d4 e7f8 c1f4 a7a6 f4g3 b7b5 a5b6 d8b6 d4b3 d7e5 c3a4 b6c7 c4f1 a6a5 c2c4 c8a6 b3d2 g7g5 c4c5 a6f1 d2f1 f6d7 c5d6 c7d6 f1e3 d6d1 e1d1 d7c5 a4c5 f8c5');
INSERT INTO evaluations VALUES(40693,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(40693,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(40693,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(40693,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(40693,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(40693,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(40693,36,'cp 62','b8d7 e1g1 c7c6 a2a4 e8g8 a4a5 d8c7 f1e1 f8e8 c4f1 d7f8 d4d5 c8d7 f3d2 c6d5 e4d5 e8b8 d2c4 b7b5 a5b6 a7b6 a1a8 b8a8 c1e3 a8b8 c4a3 f8g6 a3b5 c7d8 f1d3 b8a8 d1e2 g6h4 f2f4 d7b5 d3b5 h4f5 f4e5 d6e5 e3f2 f5d4 e2e5 d4b5 c3b5 d8d5 f2b6 d5b7 b6f2 a8e8');
INSERT INTO evaluations VALUES(40900,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(40900,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(40900,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(40900,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(40900,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(40900,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(40900,36,'cp 91','d1g4 g7g6 g1f3 c7c5 f1b5 c8d7 a1b1 d7b5 b1b5 d8c7 e1g1 h7h5 g4h3 c7c6 a2a4 c6e4 f1e1 e4g4 b5b7 g4h3 g2h3 c5d4 c1f4 b8c6 f3d4 c6a5 b7c7 e7d8 d4e6 d8c7 e6c7 e8d7 e1d1 d7e7 d1d5 a5b7 d5b5 b7d6 b5e5 e7f6 c7a8 h8a8 e5c5 d6e4 c5c6 f6f5 f4e3 a7a6 g1g2 a8d8 c3c4 a6a5 c6c7 f5e6 c4c5 d8d5');
INSERT INTO evaluations VALUES(42042,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42042,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(42042,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(42042,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(42042,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(42042,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42042,36,'cp 0','c7c5 d4c5 b4c5 g1f3 e8g8 a2a3 d5c4 f1c4 b8d7 e1g1 a7a6 c4d3 b7b5 b2b4 c5e7 c1b2 c8b7 a1c1 a8c8 c2b1 b7f3 g2f3 d7e5 d3e2 e5c4 f1d1 d8e8 e2d3 c4e5 d3e2');
INSERT INTO evaluations VALUES(40622,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(40622,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40622,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40622,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40622,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(40622,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40622,36,'cp -6','d4c5 b4c5 g1f3 e8g8 a2a3 d5c4 f1c4 b8d7 e1g1 a7a6 c3e4 f6e4 c2e4 d7f6 e4c2 c5d6 e3e4 b7b5 c4d3 e6e5 h2h3 c8d7 f1d1 d8e7 c2e2 h7h6 c1d2 a8c8 f3h4 f8e8 e2f3 d6c5 d2c3 c8c6 h4f5 d7f5 e4f5');
INSERT INTO evaluations VALUES(47290,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(47290,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(47290,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(47290,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(47290,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(47290,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(47290,36,'cp -27','c7c5 c2c4 c5d4 d1d4 b8c6 e5c6 b7c6 c4d5 c6d5 c1f4 e7e6 e2e3 d8a5 b1c3 c8a6 f1a6 a5a6 a1d1 a8c8 d4d3 a6d3 d1d3 f8b4 e1e2 e8e7 h1c1 f6e4 f4e5 f7f6 e5d4 e7f7 a2a3 b4c3 d4c3 e4d6 f2f3 g7g5 e2f1 d6c4 f1e2 h8e8');
INSERT INTO evaluations VALUES(52269,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52269,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(52269,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(52269,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(52269,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(52269,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52269,36,'cp 0','f8g7 b1c3 d7d6 c1g5 b8c6 f2f4 c8e6 c4e6 f7e6 g1f3 g8h6 c3e2 d8b6 a1b1 b6a5 c2c3 a5a2 e1g1 e8g8 g1h1 a7a5 b2b3 b7b5 g5h6 g7h6 b1a1 a2b2 a1b1 b2a3 b1a1');
INSERT INTO evaluations VALUES(37766,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37766,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37766,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37766,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37766,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37766,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37766,36,'cp 0','b1c3 d7d6 c1g5 b8c6 f2f4 c8e6 c4e6 f7e6 g1f3 g8h6 c3e2 d8b6 a1b1 b6a5 c2c3 a5a2 e1g1 e8g8 g1h1 a7a5 b2b3 a5a4 b3a4 h6g4 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(37761,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37761,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37761,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37761,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37761,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37761,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37761,36,'cp 0','b8c6 c1g5 d7d6 f2f4 c8e6 c4e6 f7e6 g1f3 g8h6 c3e2 d8b6 a1b1 b6a5 c2c3 a5a2 e1g1 e8g8 g1h1 a7a5 b2b3 h6g4 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(5292,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5292,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5292,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5292,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5292,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5292,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(5292,36,'cp 28','c1g5 d7d6 f2f4 c8e6 c4e6 f7e6 g1f3 g8h6 c3e2 d8b6 a1b1 h6g4 h2h3 g4e3 d1d2 e3g2 e1f2 h7h6 f2g2 h6g5 f4g5 e6e5 f3h2 e7e6 h2g4 e8c8 h1f1 b6c7 c2c3 c8b8 f1f3 d6d5 e2g3 c5c4 b1f1 c4d3 d2d3 d5d4 d3c4 d4c3 b2c3');
INSERT INTO evaluations VALUES(54649,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(54649,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(54649,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(54649,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(54649,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(54649,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(54649,36,'cp 167','b1c3 g8e7 d5c6 b7c6 f1c4 c8a6 c4b3 d8c7 f3d2 b8d7 d2c4 d6d5 e4d5 e7d5 c3d5 c6d5 d1d5 e8c8 d5e6 c8b7 c1d2 d7c5 c4a5 c7a5 d2a5 c5e6 b3e6 d8d6 e6b3 d6d7 a1d1 d7d1 e1d1 f8c5 b3d5 b7b8 f2f3 c5b4 a5b4 h8d8 d1c1 d8d5 b2b3 h7h5 c2c4 d5d7 h2h4 a6b7 h1e1 d7d4 b4c3');
INSERT INTO evaluations VALUES(54648,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(54648,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(54648,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(54648,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(54648,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(54648,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(54648,36,'cp 130','g8e7 f1c4 b7b5 c4b3 b5b4 c3e2 a7a5 c2c4 c8d7 h2h4 b8a6 c1e3 h7h5 e1g1 g7g6 a2a3 c6c5 a3b4 a6b4 b3a4 d7a4 d1a4 d8d7 a4b3 f8g7 a1a3 e8g8 f1a1 d7d8 e2c3 g8h8 c3b5 h8h7 g2g3 e7g8 b3d1 g7h6 e3d2 h6d2 d1d2 g8h6');
INSERT INTO evaluations VALUES(5453,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5453,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(5453,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(5453,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(5453,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(5453,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(5453,36,'cp 27','c8d7 d1d2 a7a6 d4c6 b7c6 b5c4 g8f6 h2h3 d8a5 e1g1 e8g8 a1d1 f6h5 e3g5 f8e8 g2g4 h5f6 c4b3 d7e6 g5f6 g7f6 c3d5 a5d2 d5f6 e7f6 d1d2 e6b3 a2b3 e8e4 d2d6 e4e2 f1c1 a8e8 d6c6 e2d2 c6f6 e8e2 f6f3 d2c2 c1c2 e2c2');
INSERT INTO evaluations VALUES(6926,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(6926,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(6926,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(6926,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(6926,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(6926,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(6926,36,'cp 134','f3h4 g7g6 c2c3 d7d6 e1g1 c5b6 b2b4 d8e7 h4f3 a7a5 b4b5 c6b8 c4d5 e7g7 c3c4 c7c6 d5g8 g7g8 b1c3 b8d7 a1b1 g8e6 d1a4 c6c5 a4d1 b6d8 a2a4 b7b6 c3d5 d7f8 g1h1 c8b7 f3d2 b7d5 c4d5');
INSERT INTO evaluations VALUES(7589,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(7589,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(7589,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(7589,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(7589,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(7589,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(7589,36,'cp 89','f3e5 d8e7 e5f3 d7d5 e4e5 c8g4 d3d4 c5b6 e1g1 f6d7 c2c3 e8g8 h2h3 g4h5 c1g5 e7e6 b1d2 c6c5 b2b4 c5c4 g5h4 h7h6 f1e1 c7c6 h4g3 a7a5 a2a3 h5g6 d1c1 g6d3 d2f1 b6c7 f3h4 a8a6 f2f4 a5b4 a3b4 f8a8 a1a6 a8a6 f1e3');
INSERT INTO evaluations VALUES(7581,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(7581,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(7581,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(7581,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(7581,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(7581,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(7581,36,'cp 79','d8e7 e5f3 d7d5 e4e5 c8g4 d3d4 c5b6 e1g1 f6d7 c2c3 e8g8 h2h3 g4h5 c1g5 e7e6 b1d2 c6c5 b2b4 c5c4 a2a3 c7c6 f1e1 a7a5 g5e3 b6d8 e3f4 a8a6 f4g3 h7h6 d1c1 d8c7 f3h4 g7g5 h4f3 f8a8 c1b2 d7f8 h3h4 g5g4 f3h2 a5b4 a3b4 f7f5 a1a6 a8a6');
INSERT INTO evaluations VALUES(15323,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(15323,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(15323,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(15323,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(15323,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(15323,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(15323,36,'cp 114','f8e7 f3e5 c6e5 e1e5 c8e6 c4d5 e6d5 b1c3 d5e6 d1f3 e7f6 e5e1 e8g8 d2d3 f8e8 c1f4 f6c3 b2c3 d8f6 f4c7 f6f3 g2f3 e6d7 e1e3 d7a4 a1c1 a8c8 c7e5 f7f6 e5d4 e8d8 e3e7 a4c6 d4a7 c6f3 c3c4 c8a8 a7b6 d8d6 b6c5 d6c6 c5e3 b7b6 h2h4 h7h6 e7b7 a8a2 b7b6 c6b6');
INSERT INTO evaluations VALUES(16115,30,'cp 759',NULL);
INSERT INTO evaluations VALUES(16115,31,'cp 763',NULL);
INSERT INTO evaluations VALUES(16115,32,'cp 759',NULL);
INSERT INTO evaluations VALUES(16115,33,'cp 751',NULL);
INSERT INTO evaluations VALUES(16115,34,'cp 774',NULL);
INSERT INTO evaluations VALUES(16115,35,'cp 771',NULL);
INSERT INTO evaluations VALUES(16115,36,'cp 778','g5f7 c6a5 c4d5 f6d5 f7d8 d5e7 b2b4 a5c4 d2d3 c4d6 f2f4 e5f4 c1f4 e8d8 f4d6 c7d6 e1g1 b7b6 b1c3 c8b7 d1g4 b7c6 f1f7 d8c7 g4g3 h6h5 a2a4 h5h4 g3g5 h4h3 a4a5 h3g2 a5b6 a7b6 a1a8 c6a8 c3b5 c7d8 g1g2 d8c8 f7e7');
INSERT INTO evaluations VALUES(16040,30,'cp 759',NULL);
INSERT INTO evaluations VALUES(16040,31,'cp 778',NULL);
INSERT INTO evaluations VALUES(16040,32,'cp 763',NULL);
INSERT INTO evaluations VALUES(16040,33,'cp 774',NULL);
INSERT INTO evaluations VALUES(16040,34,'cp 777',NULL);
INSERT INTO evaluations VALUES(16040,35,'cp 782',NULL);
INSERT INTO evaluations VALUES(16040,36,'cp 778','c6a5 c4d5 f6d5 f7d8 d5e7 b2b4 a5c4 d2d3 c4d6 f2f4 e5f4 c1f4 e8d8 f4d6 c7d6 e1g1 e7c6 b1c3 b7b6 b4b5 f8e7 d1g4 h8g8 b5c6 d7c6 g4g6 c8d7 g6f7 g8e8 f7g7 e7g5 f1f7 e8e7 f7e7 g5e7 g7h6 d7e8 h6f4 e8d7 a2a4 d8c7 a4a5');
INSERT INTO evaluations VALUES(17458,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(17458,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(17458,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(17458,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(17458,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(17458,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(17458,36,'cp 95','d4f5 d7d6 f1c4 c8e6 c4b3 g7g6 f5e3 c6d4 e3d5 f8g7 e1g1 g8f6 c1g5 h7h6 d5f6 g7f6 g5f6 d8f6 c3d5 f6d8 d5e3 d4b3 a2b3 e8g8 d1d3 a8c8 f1d1 c8c6 c2c4 h6h5 g1h1 b7b5 f2f3 d8b6 d3c3 b5c4 b3c4 f8c8');
INSERT INTO evaluations VALUES(17463,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17463,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(17463,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17463,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(17463,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(17463,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(17463,36,'cp 75','d7d6 f1c4 c8e6 c4b3 h7h6 e1g1 g8f6 f5g3 a8c8 g3h5 e6b3 a2b3 f6h5 d1h5 f8e7 c1e3 c6b4 h5e2 c8c6 f1d1 b7b5 e2d2 e7g5 e3g5 d8g5 c3b5 g5d2 d1d2 b4c2 b5d6 e8e7 d6f5 e7f6 a1c1 h8c8 f2f3 g7g6 f5h6 c2d4 c1c6 c8c6 h6g4 f6e6 g4e3 f7f5 e4f5 g6f5 h2h4 c6c8 d2d1 d4b3');
INSERT INTO evaluations VALUES(20154,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(20154,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(20154,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(20154,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(20154,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(20154,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(20154,36,'cp 15','b7c6 f1c4 g8f6 d1e2 d8c7 c1g5 f8e7 e1g1 e8g8 a1d1 h7h6 g5h4 c8g4 f2f3 g4c8 h4f2 f8d8 d1d2 f6d7 f1d1 d7b6 c4b3 a7a5 c3a4 b6a4 b3a4 e7g5 f2e3 g5e3 e2e3 c8e6 a4b3 a5a4 b3e6 f7e6 c2c4 c6c5 d2d3 a8b8 e3d2 b8b6 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(22229,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(22229,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(22229,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22229,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22229,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22229,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22229,36,'cp -9','g7g5 f4g3 g5g4 f3g1 g8h6 f1b5 f8g7 e5f6 d8f6 g3e5 f6e7 b5c6 b7c6 e5g7 e7g7 b1c3 e8g8 g1e2 c8a6 d1d2 a8b8 b2b3 b8e8 f2f4 c6c5 e1c1 h6f5 c1b1 a6e2 c3e2 c5d4 e2d4 g7d4 d2d4 f5d4 d1d4 c7c5 d4a4 h7h5 a4a7');
INSERT INTO evaluations VALUES(22349,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(22349,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(22349,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(22349,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(22349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22349,36,'cp -5','f4g3 g5g4 f3g1 g8h6 f1b5 f8g7 e5f6 d8f6 g3e5 f6e7 e5g7 e7g7 b5c6 b7c6 g1e2 a8b8 d1d2 c6c5 b1c3 e8g8 f2f4 c5d4 e2d4 c7c5 d4b3 h6f5 e1c1 d5d4 c3e4 f5e3 e4c5 e3d1 h1d1 e6e5 f4e5 g7e5 c5d3 e5g7 d3c5 c8f5 d2d4 g7d4 b3d4');
INSERT INTO evaluations VALUES(22350,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(22350,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(22350,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(22350,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(22350,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(22350,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(22350,36,'cp -26','g5g4 f3g1 g8h6 f1b5 f8g7 e5f6 d8f6 g3e5 f6g6 e5g7 g6g7 b5c6 b7c6 g1e2 a8b8 b1c3 c6c5 b2b3 c5d4 d1d4 g7d4 e2d4 c8d7 d4e2 e8e7 a1d1 h6f7 e1g1 h7h5 f1e1 c7c6 c3a4 e7d6 e2g3 f7g5 a4c3 b8e8 g3e4 g5e4 c3e4 d6c7 e4c5 e6e5');
INSERT INTO evaluations VALUES(23049,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(23049,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(23049,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(23049,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(23049,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(23049,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(23049,36,'cp 205','g2g4 d8f6 e2e3 f6g6 f1d3 e5e4 d3f1 d4f3 g1f3 e4f3 h1g1 g8f6 b1c3 f8b4 c1d2 e8g8 d1f3 d7d6 f1e2 b4c3 d2c3 f6e4 c3d4 c7c5 d5c6 b7c6 b2b3 f8e8 e1c1 d6d5 h2h4 g6d6 e2d3 d6b4 c4d5 c6d5 c1b2 a7a5 a2a3 b4e7 g4g5 c8a6 d3e4 d5e4 f3f6 e7f8 f6f4 f8e7 h4h5 a6d3 d1c1 a8c8 c1c8 e8c8');
INSERT INTO evaluations VALUES(23068,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(23068,31,'cp 261',NULL);
INSERT INTO evaluations VALUES(23068,32,'cp 261',NULL);
INSERT INTO evaluations VALUES(23068,33,'cp 249',NULL);
INSERT INTO evaluations VALUES(23068,34,'cp 266',NULL);
INSERT INTO evaluations VALUES(23068,35,'cp 254',NULL);
INSERT INTO evaluations VALUES(23068,36,'cp 266','b1c3 c7c6 c3e4 d8b6 e4d6 f8d6 d4d6 c6c5 f1d3 b6c6 f2f3 c5c4 d3e4 c6d6 e5d6 g8f6 e4f5 b7b5 e1f2 e8g8 c1g5 a8b8 h1e1 b8b6 g5f6 g7f6 a1d1 g8g7 d1d4 b6a6 a2a3 c4c3 b2b3 a6a5 b3b4 a5a3 d4g4 g7h8 g4h4 a3a6 e1d1 a6c6 h4h7 h8g8 d1d4');
INSERT INTO evaluations VALUES(23067,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(23067,31,'cp 288',NULL);
INSERT INTO evaluations VALUES(23067,32,'cp 287',NULL);
INSERT INTO evaluations VALUES(23067,33,'cp 284',NULL);
INSERT INTO evaluations VALUES(23067,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(23067,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(23067,36,'cp 288','c7c6 c3e4 d8b6 e4d6 f8d6 d4d6 c6c5 f1d3 b6d6 e5d6 b7b6 e1g1 e8f8 a2a4 a7a5 f1e1 c8a6 d3f5 a8d8 b2b4 c5b4 c1e3 g8f6 e3b6 d8c8 b6d4 a6b7 e1e5 c8a8 a1e1 h7h5 e5e7 b7c6 d4f6 g7f6 f5d7 c6d7 e7d7 f8g7 e1e7 h8d8 e7f7 g7g6 g1f1 a8c8 d7d8 c8d8 f7a7 d8d6 a7a5 d6d2 a5c5');
INSERT INTO evaluations VALUES(26956,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(26956,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(26956,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(26956,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(26956,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(26956,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(26956,36,'cp 14','g8f6 h2h3 g4h5 f1e1 f8e7 a2a4 a7a6 a4a5 d8d7 c4e2 h7h6 d2d3 h5g6 e2f1 e8g8 g2g3 a8e8 f1g2 e7d8 a1a4 f6h7 d3d4 h7g5 c1g5 h6g5 d4d5 c6b8 g1h2 f7f5 e4f5 d7f5 f3d2 f5c2 d1c2 g6c2 a4b4 b8d7 d2e4 b7b5 a5b6 d7b6 e1a1');
INSERT INTO evaluations VALUES(29729,30,'cp -236',NULL);
INSERT INTO evaluations VALUES(29729,31,'cp -210',NULL);
INSERT INTO evaluations VALUES(29729,32,'cp -217',NULL);
INSERT INTO evaluations VALUES(29729,33,'cp -202',NULL);
INSERT INTO evaluations VALUES(29729,34,'cp -205',NULL);
INSERT INTO evaluations VALUES(29729,35,'cp -209',NULL);
INSERT INTO evaluations VALUES(29729,36,'cp -223','b8c6 e2c4 e7e6 g5e4 f6e4 c3e4 c6d4 c4a4 c7c6 c2c3 f5e4 d2d3 e4f5 a4d4 d8d4 c3d4 e8c8 c1e3 f8e7 a1c1 d8d5 f1e2 e7f6 c1c5 d5c5 d4c5 f6b2 e1d2 h7h5 g2g3 h8d8 h2h3 b2f6 g3g4 f5e4 h1g1 h5g4 h3g4 f6e5 a2a4 e4d5 g4g5 a7a6 f2f4 e5c7');
INSERT INTO evaluations VALUES(24983,30,'cp -208',NULL);
INSERT INTO evaluations VALUES(24983,31,'cp -206',NULL);
INSERT INTO evaluations VALUES(24983,32,'cp -219',NULL);
INSERT INTO evaluations VALUES(24983,33,'cp -225',NULL);
INSERT INTO evaluations VALUES(24983,34,'cp -223',NULL);
INSERT INTO evaluations VALUES(24983,35,'cp -217',NULL);
INSERT INTO evaluations VALUES(24983,36,'cp -211','e2c4 e7e6 g5e4 f6e4 c3e4 c6d4 c4a4 c7c6 c2c3 f5e4 d2d3 e4f5 a4d4 d8d4 c3d4 e8c8 c1e3 f8b4 e1d1 b4e7 a1c1 e7f6 c1c4 d8d6 c4a4 a7a6 f2f4 h8d8 d1d2 f6d4 a4d4 d6d4 e3d4 d8d4 g2g3 d4b4 b2b3 b4b5 f1e2 e6e5 f4e5 b5a5 a2a4 a5e5 h1f1 f5e6 b3b4 a6a5 f1b1 a5b4 b1b4 e5a5 e2f3');
INSERT INTO evaluations VALUES(31516,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(31516,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(31516,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31516,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(31516,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31516,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31516,36,'cp 0','f5g6 g1e2 c6c5 h2h4 h7h5 e2f4 g6h7 f4h5 b8c6 d4c5 f8c5 c1g5 d8b6 d1d2 b6b2 a1b1 b2c2 d2c2 h7c2 b1c1 d5d4 c3d5 e6d5 c1c2 c5b4 e1d1 c6e5 h5g7 e8f8 g7f5 d4d3 c2b2 b4c5 b2b7 e5g4 d1d2 g4f2 f1g2 f2h1 g2d5 f7f6 b7f7 f8e8 f5g7 e8d8 g7e6 d8e8');
INSERT INTO evaluations VALUES(31411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31411,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31411,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31411,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31411,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31411,36,'cp 0','g1e2 c6c5 h2h4 h7h5 e2f4 g6h7 f4h5 b8c6 d4c5 f8c5 c1g5 d8b6 d1d2 b6b2 a1b1 b2c2 d2c2 h7c2 b1c1 d5d4 c3d5 e6d5 c1c2 c5b4 e1d1 c6e5 h5g7 e8f8 g7f5 d4d3 c2b2 b4c5 b2b7 e5g4 d1d2 g4f2 f1g2 f2h1 g2d5 f7f6 b7f7 f8e8 g5e3 c5b4 d2d3 a8c8 f5g7 e8d8 e3b6 a7b6 g7e6 d8e8 e6g7');
INSERT INTO evaluations VALUES(31410,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31410,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31410,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31410,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31410,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31410,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31410,36,'cp 0','c6c5 h2h4 h7h5 e2f4 g6h7 f4h5 b8c6 d4c5 f8c5 c1g5 d8b6 d1d2 b6b2 a1b1 b2c2 d2c2 h7c2 b1c1 d5d4 c3d5 e6d5 c1c2 c5b4 e1d1 c6e5 h5g7 e8f8 g7f5 d4d3 c2b2 b4c5 b2b7 e5g4 d1d2 g4f2 f1g2 f2h1 g2d5 f7f6 b7f7 f8e8 f5g7 e8d8 g7e6 d8e8');
INSERT INTO evaluations VALUES(31625,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31625,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31625,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31625,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31625,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31625,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31625,36,'cp 0','h2h4 h7h5 e2f4 g6h7 f4h5 b8c6 d4c5 f8c5 c1g5 d8b6 d1d2 b6b2 a1b1 b2c2 d2c2 h7c2 b1c1 d5d4 c3d5 e6d5 c1c2 c5b4 e1d1 c6e5 h5g7 e8f8 g7f5 d4d3 c2b2 b4c5 b2b7 e5g4 d1d2 g4f2 f1g2 f2h1 g2d5 f7f6 b7f7 f8e8 f5g7 e8d8 g7e6 d8e8');
INSERT INTO evaluations VALUES(47397,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(47397,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(47397,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(47397,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(47397,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(47397,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(47397,36,'cp -42','e7e5 g1f3 f8c5 f1c4 e8g8 d2d3 b8c6 c1g5 h7h6 g5h4 c8e6 c4b3 a7a5 d1d2 c6d4 b3e6 d4f3 g2f3 f7e6 e1c1 f8f7 h4f2 c5f2 d2f2 f6h5 h1g1 g8h7 d1f1 d8d4 c3e2 d4f2 f1f2 b7b5 e2g3 h5f4 c1d2 c7c5 g3e2');
INSERT INTO evaluations VALUES(48934,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(48934,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(48934,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(48934,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(48934,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(48934,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(48934,36,'cp -48','f1c4 f8c5 d2d3 e8g8 g1f3 b8c6 c1g5 h7h6 g5h4 c8g4 c3d5 f6d5 h4d8 d5e3 d8e7 e3g2 e1d2 c6e7 h1g1 c5g1 d1g1 g4f3 a1f1 g2h4 g1g3 g7g5 f1f3 h4f3 g3f3 g8g7 f3g4 e7g6 g4d7 a8e8 d7c7 e8e7 c7c5 b7b6 c5f2 g6f4 a2a4 f7f5 f2e1 f5e4 e1e4 h6h5 d2c3 g5g4 c3b3 f8f6 b3a2');
INSERT INTO evaluations VALUES(48209,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48209,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(48209,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(48209,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(48209,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(48209,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(48209,36,'cp -16','h7h6 g5f6 d8f6 b1d2 c7c5 e2e4 c5d4 f3d4 f8c5 e4d5 e6d5 d2b3 c5b6 d1e2 e8f8 e2f3 f6f3 d4f3 b8c6 a1d1 c8e6 f1b5 a7a6 b5e2 g7g6 f3d4 f8g7 d4e6 f7e6 h2h4 h6h5 e1g1 a8d8 g2g3 h8e8 e2f3 c6e5 f3g2');
INSERT INTO evaluations VALUES(49232,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49232,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(49232,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(49232,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(49232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49232,36,'cp 0','g5f6 d8f6 e2e4 g7g5 c2c3 d7d6 b1a3 h6h5 e4e5 d6e5 d4e5 f6e7 d1a4 c8d7 a4b3 b8c6 b3b7 a8b8 b7c7 b8c8 c7b7 c8b8');
INSERT INTO evaluations VALUES(49228,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(49228,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(49228,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49228,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49228,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49228,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49228,36,'cp -19','d8f6 c2c3 d7d5 b1d2 c7c5 e2e4 c5d4 f3d4 f8c5 e4d5 e6d5 d2b3 c5b6 d1e2 e8f8 e2f3 f6f3 d4f3 b8c6 a1d1 c8e6 f1e2 g7g6 e1g1 f8g7 b3d4 e6g4 d1d2 h8e8 f1d1 e8e7 g1f1 a8e8 h2h3 g4d7 d4c2 a7a6 a2a3');
INSERT INTO evaluations VALUES(49647,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(49647,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(49647,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(49647,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(49647,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(49647,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(49647,36,'cp -82','d7d5 g1f3 f8d6 b1d2 d5d4 e3g5 c8e6 c2c3 c7c5 d2c4 e6c4 d3c4 h7h6 g5f6 d8f6 g2g3 b8c6 c3d4 c5d4 f1e2 e8g8 e1g1 a7a5 f3e1 c6b8 e1d3 b8d7 f2f4 e5f4 c4c5 d6e7 g3f4 d7c5 d3c5 e7c5 e2d3 g7g6 g1h1');
INSERT INTO evaluations VALUES(48842,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(48842,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(48842,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(48842,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(48842,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(48842,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(48842,36,'cp -67','g1f3 f8d6 e4d5 f6d5 e3d2 e8g8 b1c3 d5c3 d2c3 c7c5 g2g3 b8c6 f1g2 c8g4 a2a4 a8c8 a4a5 c6d4 e1g1 d8c7 f1e1 h7h6 c3d2 f8d8 c2c4 d4c6 d2c3 c7d7 d1a4 f7f6 f3h4 c8c7 g2d5 g4e6 d5e4 f6f5 e4d5 e6f7 d5f7 d7f7 h4f3 f7f6 a4b3');
INSERT INTO evaluations VALUES(56218,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56218,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56218,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56218,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56218,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56218,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56218,36,'cp -5','g8f6 b1c3 c8f5 c1f4 f8d6 f4d6 d8d6 e2e3 e8g8 f1d3 f5d3 d1d3 b8d7 d3c2 c7c6 g1f3 f8e8 e1g1 d7b6 f3e5 e8e7 f1c1 a8e8 h2h3 h7h6 a3a4 a7a5 a1b1 b6c8 e5d3 d6d8 b2b4 a5b4 d3b4 c8d6 b4d3');
INSERT INTO evaluations VALUES(10287,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(10287,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(10287,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10287,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(10287,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10287,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(10287,36,'cp 9','d4c6 b7c6 b1c3 e8g8 e1g1 d7d6 c1g5 h7h6 g5h4 d8e7 d1d3 e7e5 h4g3 e5d4 c3a4 f6e4 a4c5 d4c5 d3e4 d6d5 e4f4 c5c4 f4c4 d5c4 g3c7 c8f5 a1c1 a7a5 f2f3 f7f6 g1f2 f8f7 c7b6 a5a4 a2a3 f7b7 b6d4 a8e8 f1g1 g8f7 h2h3 f5g6 d4c3');
INSERT INTO evaluations VALUES(9957,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(9957,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(9957,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(9957,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(9957,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(9957,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9957,36,'cp 5','b7c6 b1c3 e8g8 e1g1 d7d6 d1d3 c5b6 c1g5 h7h6 g5h4 g7g5 h4g3 f6h5 c3a4 d8f6 c2c3 f8e8 a4b6 a7b6 a2a4 c8e6 f1e1 h5g3 h2g3 e6c4 d3c4 c6c5 c4b5 g8g7 b5c6 a8c8 g1f1 e8e5 c6d7 f6e6 d7e6 f7e6 f2f3 c8h8 f1f2 h6h5 e1h1');
INSERT INTO evaluations VALUES(19171,30,'cp 477',NULL);
INSERT INTO evaluations VALUES(19171,31,'cp 458',NULL);
INSERT INTO evaluations VALUES(19171,32,'cp 485',NULL);
INSERT INTO evaluations VALUES(19171,33,'cp 463',NULL);
INSERT INTO evaluations VALUES(19171,34,'cp 482',NULL);
INSERT INTO evaluations VALUES(19171,35,'cp 472',NULL);
INSERT INTO evaluations VALUES(19171,36,'cp 466','a2a3 b4d5 d1d5 g8f6 d5d1 f8e7 b1c3 e8g8 f1c4 b7b5 c4b5 d7d5 e1g1 a8b8 b5d3 h7h6 b2b4 c5b4 a3b4 e7b4 c3e2 c8g4 c1e3 f8e8 e2d4 b8c8 h2h3 g4d7 a1a7 b4c5 a7a6 f6e4 d1a1 c8b8 c2c4 d5c4 d3c4');
INSERT INTO evaluations VALUES(19172,30,'cp 475',NULL);
INSERT INTO evaluations VALUES(19172,31,'cp 459',NULL);
INSERT INTO evaluations VALUES(19172,32,'cp 481',NULL);
INSERT INTO evaluations VALUES(19172,33,'cp 463',NULL);
INSERT INTO evaluations VALUES(19172,34,'cp 487',NULL);
INSERT INTO evaluations VALUES(19172,35,'cp 486',NULL);
INSERT INTO evaluations VALUES(19172,36,'cp 494','b4d5 d1d5 g8f6 d5d1 f8e7 b1c3 e8g8 f1c4 b7b5 c4b5 a8b8 e1g1 d7d5 b5d3 h7h6 f1e1 b8b6 h2h3 b6e6 e1e6 c8e6 c1f4 d5d4 c3b1 c5c4 d3f1 d8b6 b1d2 f8c8 f4e5 c4c3 e5f6 e7f6 b2c3 d4c3 d2e4 f6e7 f1d3 f7f5 e4g3 e7f6 d3f5');
INSERT INTO evaluations VALUES(20834,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(20834,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(20834,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(20834,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20834,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(20834,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(20834,36,'cp 22','g7g5 f4g3 h7h5 h2h3 c8f5 e2e3 h5h4 g3h2 e7e6 c3b5 h8h7 a2a3 a7a6 b5c3 f8d6 f1d3 h7h8 b2b4 c6e7 h2d6 c7d6 e3e4 d5e4 c3e4 d6d5 e4c5 b7b6 c5a6 e7g6 b4b5 f5d3 c2d3 g8e7 a6b4 e8f7 e1g1 f7g7 a3a4 g6f4');
INSERT INTO evaluations VALUES(29753,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(29753,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(29753,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(29753,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(29753,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(29753,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(29753,36,'cp 104','f3e5 g4h5 f2f3 f6d7 e5c4 e7e6 e2e4 f8e7 c1e3 a7a6 a2a4 b8c6 f1e2 f7f6 e1g1 h5f7 f3f4 e8g8 g1h1 g8h8 e3g1 c6b4 d1d2 f7g8 e4e5 b7b6 e2f3 a8c8 b2b3 f6f5 a1d1 c7c5 c4d6 e7d6 e5d6');
INSERT INTO evaluations VALUES(32677,30,'cp 278',NULL);
INSERT INTO evaluations VALUES(32677,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(32677,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(32677,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(32677,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(32677,35,'cp 281',NULL);
INSERT INTO evaluations VALUES(32677,36,'cp 312','e1g1 e7d7 h2h3 g4h5 f3h4 h5e2 d1e2 d6d5 f2f3 f8e7 f3e4 e7h4 e4d5 d7e7 e2c4 b8d7 b1c3 e8g8 c4c7 d7b6 c1f4 e7c7 f4c7 f8c8 c7b6 a7b6 a1d1 c8c4 d1d3 h4f6 f1f4 f6g5 f4e4 f7f5 e4e2 g5f6 c3b5 c4b4 d3b3 f6d4 b5d4 b4d4 b3b6 a8d8 b6b7 d4d5 c2c3 d5d1 g1f2 f5f4 a2a3 h7h6');
INSERT INTO evaluations VALUES(32676,30,'cp 305',NULL);
INSERT INTO evaluations VALUES(32676,31,'cp 280',NULL);
INSERT INTO evaluations VALUES(32676,32,'cp 280',NULL);
INSERT INTO evaluations VALUES(32676,33,'cp 315',NULL);
INSERT INTO evaluations VALUES(32676,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(32676,35,'cp 286',NULL);
INSERT INTO evaluations VALUES(32676,36,'cp 319','d6d5 f1e1 b8d7 f3d2 g4e2 e1e2 f7f5 f2f3 e7f6 d2b3 e8c8 f3e4 d5e4 d1e1 c8b8 b1c3 g7g5 g2g4 f5f4 c3e4 f6g7 e1f1 d7f6 e4f6 g7f6 f1f3 f8d6 e2g2 a7a5 c2c3 a5a4 b3a5 c7c6 c1d2 f6f7 c3c4 d6c7 g2e2 h8e8 e2e8 d8e8 a1f1 c7a5 d2a5 f7c4');
INSERT INTO evaluations VALUES(32780,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(32780,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(32780,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(32780,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(32780,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(32780,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(32780,36,'cp 42','b1d2 g7g6 f1e2 g8f6 f3e5 g4e2 d1e2 f6d7 d2c4 b7b5 e5d7 b8d7 c4a5 a8c8 c1e3 f8g7 c2c3 d7b6 h1c1 e8g8 c1c2 f7f5 e4f5 f8f5 g2g3 g8f7 a1d1 f7e8 d1b1 f5h5 h2h4 g7h6 e3h6 h5h6 b2b3 h6h5 c3c4 b5c4 a5c4');
INSERT INTO evaluations VALUES(33491,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(33491,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(33491,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(33491,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(33491,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(33491,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(33491,36,'cp 193','d2d4 c5b4 f2f3 c7c5 c1f4 e8g8 d4c5 b4c3 b2c3 f6h5 e5d3 h5f4 d3f4 e7c5 d1d4 c5a5 f1c4 b8c6 d4e3 b7b6 c4b3 c8b7 e1c1 c6e5 e3d4 a8c8 c1b2 a5c5 d1d2 f8e8 h1d1 c5d4 d2d4 b7c6 h2h4 a7a5 a2a4 g8f8 h4h5 h7h6 c3c4 f8e7 d4d6 e7f8 c2c3 c8c7 b2c1 f7f6 f4g6 f8f7 g6e5 e8e5');
INSERT INTO evaluations VALUES(33490,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(33490,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(33490,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(33490,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(33490,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(33490,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(33490,36,'cp 201','c5b4 f2f3 c7c5 c1f4 e8g8 d4c5 f6h5 e5d3 b4c3 b2c3 h5f4 d3f4 e7c5 d1d4 c5a5 f1c4 b7b6 c4b3 b8c6 d4e3 c8b7 a1d1 c6e5 f4d5 g8h8 e1g1 a8c8 d5b4 h8g8 e3d4 f8d8 f3f4 e5g4 f1e1 a5c5 d4c5 c8c5 c3c4 g8f8 e4e5 g4h6 d1d3 f8e7 e1d1 h6f5 g1f2 h7h6 b4d5 b7d5');
INSERT INTO evaluations VALUES(33495,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(33495,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(33495,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(33495,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(33495,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(33495,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(33495,36,'cp 202','f2f3 c7c5 c1f4 e8g8 d4c5 f6h5 e5d3 b4c3 b2c3 h5f4 d3f4 e7c5 d1d4 c5a5 f1c4 b8c6 d4e3 b7b6 c4b3 c8b7 e1c1 c6e5 e3d4 a8c8 c1b2 a5c5 d1d2 c5d4 c3d4 e5c4 b3c4 c8c4 a2a3 d7d6 h1e1 b6b5 e1e3 a7a5 e3b3 b7c6 d4d5 c6d7 b3c3 c4c3 b2c3 f7f5 f4d3 f5e4');
INSERT INTO evaluations VALUES(37421,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(37421,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(37421,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(37421,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(37421,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(37421,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(37421,36,'cp 37','e2e4 d7d6 d1d2 b8c6 e1c1 c8g4 f4e3 e7e5 d4e5 d6e5 d2d8 f8d8 d1d8 a8d8 f3d2 g4e6 f2f3 a7a6 f1c4 e6c4 d2c4 b7b5 c4d2 c6d4 d2b3 d4b3 a2b3 g7f8 c3d1 f6d7 d1f2 f8e7 c2c3 f7f5 f2d3 a6a5 c1c2 g8f7 c3c4 f5f4 e3d2 b5c4 b3c4 a5a4 h1a1 d8a8 g2g3 f4g3 h2g3 h7h5 d2c3');
INSERT INTO evaluations VALUES(37465,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(37465,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(37465,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(37465,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(37465,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37465,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(37465,36,'cp 44','d7d6 d1d2 c8g4 e1c1 b8c6 f1e2 g4f3 e2f3 e7e5 d4e5 d6e5 f4e3 d8c8 c1b1 f8e8 c3d5 f6d5 e4d5 c6d4 f3e4 d4f5 c2c3 a7a5 f2f3 f5d6 e3c5 a8b8 c5d6 c7d6 g2g4 b7b5 d1c1 c8d7 h2h4 b5b4 c3c4 f7f5 g4f5 g6f5 e4c2 e5e4 f3e4 f5e4');
INSERT INTO evaluations VALUES(39371,30,'cp 517',NULL);
INSERT INTO evaluations VALUES(39371,31,'cp 532',NULL);
INSERT INTO evaluations VALUES(39371,32,'cp 525',NULL);
INSERT INTO evaluations VALUES(39371,33,'cp 532',NULL);
INSERT INTO evaluations VALUES(39371,34,'cp 546',NULL);
INSERT INTO evaluations VALUES(39371,35,'cp 555',NULL);
INSERT INTO evaluations VALUES(39371,36,'cp 544','e1f2 d7d6 h1f1 c8e6 d2d3 e6c4 d3c4 d8d7 f2g1 b8c6 b1c3 g8e7 f3h4 h8f8 c3d5 e8c8 c1e3 c8b8 b2b3 e7d5 c4d5 c6e7 c2c4 g7g6 h4f3 f7f5 f3g5 f5f4 e3f2 h7h6 g5e6 g6g5 c4c5 g5g4 e6d8 f8d8 c5d6 c7d6');
INSERT INTO evaluations VALUES(39370,30,'cp 523',NULL);
INSERT INTO evaluations VALUES(39370,31,'cp 553',NULL);
INSERT INTO evaluations VALUES(39370,32,'cp 549',NULL);
INSERT INTO evaluations VALUES(39370,33,'cp 553',NULL);
INSERT INTO evaluations VALUES(39370,34,'cp 517',NULL);
INSERT INTO evaluations VALUES(39370,35,'cp 525',NULL);
INSERT INTO evaluations VALUES(39370,36,'cp 514','d7d6 h1f1 g8f6 d2d4 e5d4 f2g1 c8e6 c4e6 f7e6 f3d4 d8e7 d1d3 e8g8 d3h3 b8c6 d4c6 b7c6 c1g5 h7h6 g5f6 f8f6 b1d2 a8f8 h3c3 f6f1 a1f1 f8f1 d2f1 e7h4 c3e3 h4h5 h2h3 h5d1 e3a7 d1c2 a7c7 g8h7 c7d6 c2b2 d6e6 b2d4 g1h2');
INSERT INTO evaluations VALUES(44057,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(44057,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(44057,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(44057,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(44057,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(44057,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(44057,36,'cp 162','e4e5 f6e4 f1d3 d7d5 e1g1 c6c5 f1e1 c8d7 c2c3 h7h6 b1d2 e4d2 c1d2 a7a5 d2e3 c5d4 f3d4 b8c6 d1h5 c6d4 c3d4 f8e7 a1c1 e7g5 c1c3 g7g6 h5h3 g5e3 h3e3 d7c6 e1c1 d8g5 f2f4 g5e7 a2a3 h6h5 e3h3 e8g8');
INSERT INTO evaluations VALUES(44034,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(44034,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(44034,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(44034,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(44034,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(44034,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(44034,36,'cp 155','f6e4 f1d3 d7d5 e1g1 c6c5 f1e1 b8c6 d3e4 d5e4 e1e4 c5d4 f3d4 f8c5 c1e3 c6d4 b1c3 b7b6 e3d4 c8b7 d1g4 e8g8 e4f4 a7a6 d4c5 b6c5 f4c4 d8d2 g4e2 a8d8 a1e1 d2g5 f2f4 g5e7 c3e4 b7d5 c4c3 d5e4 e2e4');
INSERT INTO evaluations VALUES(45131,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45131,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45131,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45131,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45131,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45131,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45131,36,'cp 0','e7e6 c4d3 d6d5 e4e5 c5c4 d3c2 f6d7 d2d3 c4d3 d1d3 b8c6 d3e2 d8c7 c2a4 d7c5 a4c2');
INSERT INTO evaluations VALUES(49366,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(49366,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(49366,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(49366,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(49366,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(49366,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(49366,36,'cp -47','d7d5 e4d5 e6d5 c4b3 d5d4 c3e2 f8c5 d2d3 a7a5 a2a4 d8e7 g1f3 c8g4 e1g1 g4f3 g2f3 b8c6 f1e1 c6e5 g1g2 e8c8 e2g3 c5b4 c1d2 b4d2 d1d2 e7c5 d2f4 e5g6 f4f5 c5f5 g3f5 d8d7 g2f1 h8g8 e1e2 c8d8 e2e1 b7b6 b3c4 c7c6 b2b4 a5b4 a4a5 b6a5 a1a5 d8c7 a5a7 c7b6');
INSERT INTO evaluations VALUES(48226,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(48226,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(48226,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(48226,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(48226,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(48226,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(48226,36,'cp -50','e4d5 e6d5 c4b3 d5d4 c3e2 f8c5 g1f3 c8g4 d2d3 g4f3 g2f3 a7a5 a2a3 a5a4 b3a2 e8g8 e2g3 c5d6 f3f4 d8d7 d1f3 f8e8 e1d1 c7c5 h1g1 d6f8 f4f5 a8a6 c1g5 b7b5 g5f6 a6f6 f3d5 c5c4 d3c4 b5b4 c4c5 b4a3 b2a3 d7d5 a2d5');
INSERT INTO evaluations VALUES(48172,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(48172,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(48172,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(48172,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(48172,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(48172,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48172,36,'cp -42','e6d5 c4b3 d5d4 c3e2 f8c5 d2d3 a7a5 a2a4 d8e7 g1f3 c8g4 e1g1 g4f3 g2f3 b8c6 f1e1 c6e5 e2d4 c5d4 c1f4 e7c5 d1e2 e8g8 f4e5 a8e8 f3f4 d4b2 a1b1 b2c3 e2e3 c5c6 e1e2 f6g4 e3g3 g4e5 f4e5 c6c5 e5e6 f7e6 b3e6 g8h8 b1b7 e8e6 e2e6 c5d5 g3c7 d5e6 c7c3 e6g4 g1f1');
INSERT INTO evaluations VALUES(48631,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(48631,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(48631,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(48631,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(48631,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(48631,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(48631,36,'cp -47','c4b3 d5d4 c3e2 f8c5 d2d3 a7a5 g1f3 a5a4 b3c4 c8g4 f3g5 e8g8 e1g1 h7h6 g5e4 f6e4 d3e4 b8c6 h2h3 c6e5 c4d3 e5d3 c2d3 g4d7 d1c2 c5b6 e2g3 d8f6 c1d2 f6c6 c2d1 c6b5 d1c2 c7c5 f2f4 f8e8 a1c1 b5c6 f4f5');
INSERT INTO evaluations VALUES(48719,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(48719,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(48719,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(48719,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(48719,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(48719,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(48719,36,'cp -71','d5d4 c3e2 f8c5 d2d3 a7a5 a2a4 d8e7 g1f3 c8g4 f3d2 b8c6 h2h3 g4h5 f2f3 e8c8 e1g1 h5g6 d2e4 f6d5 b3d5 d8d5 f1e1 e7d7 e2f4 d5e5 f4g6 h7g6 f3f4 e5d5 c1d2 c5e7 d1f3 c6b4 d2b4 e7b4 e1e2 h8e8 g1h1');
INSERT INTO evaluations VALUES(50202,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(50202,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(50202,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(50202,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(50202,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(50202,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(50202,36,'cp -19','c2c4 e7e6 e2e3 d7d5 g1f3 c7c5 c4d5 e6d5 f1b5 b8c6 e1g1 c5d4 e3d4 f8g7 b5c6 b7c6 b1c3 e8g8 f3e5 d8b6 c3a4 b6b5 a4c5 f6e4 a2a4 b5b6 c5e4 d5e4 c1e3 f7f6 e5c4 b6d8 a1c1 c8e6 a4a5 e6d5 b2b3 d8d7 h2h3');
INSERT INTO evaluations VALUES(50912,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(50912,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50912,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(50912,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(50912,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(50912,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(50912,36,'cp 34','f8e7 e1g1 g8f6 a2a4 e8g8 b1c3 c8e6 f3g5 e6c4 d3c4 h7h6 g5f3 a6a5 d1e2 b8a6 b2b3 c7c6 f3h4 f6e4 c3e4 e7h4 c1a3 h4e7 a1d1 a6b4 c2c3 f7f5 c3b4 a5b4 a3b4 f5e4 e2e4 f8f4 e4g6 f4f6 g6d3 d8e8 d3e2 f6g6 b4d2 g6f6 f2f4 e5f4 d2f4 e8d7');
INSERT INTO evaluations VALUES(52888,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52888,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52888,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52888,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(52888,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(52888,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52888,36,'cp 14','b7b5 b2b3 c4b3 a2a4 b5a4 a1a4 e7e6 b1d2 g8f6 f1d3 a7a5 e1g1 f8e7 d2b3 e8g8 c1d2 c8a6 d2a5 d8d7 d3a6 a8a6 d1c2 d7b7 b3c5 e7c5 d4c5 b8d7 f1b1 b7a8 b1a1 a8a7 a5b4 a6a4 a1a4 a7b7 g2g3 h7h6 g1g2 f8a8 a4a8 b7a8 c2c4 a8b7 b4c3');
INSERT INTO evaluations VALUES(51654,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(51654,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(51654,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(51654,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(51654,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(51654,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51654,36,'cp 19','b2b3 c4b3 a2a4 b5b4 f1d3 e7e6 b1d2 f8e7 e1g1 g8f6 d2b3 e8g8 c1f4 a7a5 f3e5 f6d7 e5c4 c8a6 b3a5 d7b6 c4b6 d8b6 a5b3 f8d8 a4a5 b6b7 d3a6 b7a6 d1c2 b8d7 f1c1 e6e5 d4e5 c6c5 b3d2 a6e6 d2c4 d8b8 f2f3 e7f8 g1h1 b4b3 c2c3');
INSERT INTO evaluations VALUES(54017,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54017,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54017,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54017,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54017,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54017,36,'cp 0','g8f6 e1g1 b8c6 d4e3 f8e7 f1d1 e8g8 c4f1 d8c7 e3e2 b7b5 c1f4 f6h5 f4c1');
INSERT INTO evaluations VALUES(54677,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(54677,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(54677,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(54677,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(54677,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(54677,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(54677,36,'cp 198','g7g6 f1c4 b8c6 b1c3 f8g7 a2a4 e7e6 d4d5 e6d5 d1d5 d8e7 e1g1 g8h6 a4a5 a7a6 c4b3 h6f7 c1e3 c8d7 d5d2 c6b4 c3d5 b4d5 e4d5 e8g8 f3d4 a8e8 a1e1 f6f5 h2h3 f7d8 e3g5 e7f7 c2c3 e8e1 f1e1 f8e8 g5d8 e8d8');
INSERT INTO evaluations VALUES(57148,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(57148,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(57148,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(57148,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(57148,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(57148,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(57148,36,'cp -107','b8c6 g4g3 f7f5 f1b5 f5f4 g3d3 d8f6 b5c6 d7c6 b2b3 f8d6 c1b2 c8e6 g1f3 e8c8 h2h4 g8e7 d3e2 h7h6 h4h5 c6c5 b1a3 e7c6 e1c1 e6g4 a3c4 h8e8 c1b1 c6d4 b2d4 c5d4 c4d6 d8d6 e2b5 f6e6 d2d3 d6c6 d1g1 e6f6 a2a4');
INSERT INTO evaluations VALUES(2821,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(2821,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2821,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2821,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(2821,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(2821,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(2821,36,'cp 24','e7d6 e5e1 d8h4 g2g3 h4h3 e1e4 g7g5 d1f1 h3h5 b1d2 c8h3 f1e2 h5g6 d2f3 f7f6 b3c2 h3f5 c1d2 g8h8 a2a4 f5e4 c2e4 f8e8 e4g6 e8e2 g6d3 e2e8 h2h4 g5g4 f3e1 h7h5 e1g2 h8g7 b2b3 g7f7 g1f1 e8b8 d3c2 b8e8 c3c4 d5e7 c2d3 b5c4 d3c4 f7g7 g2e3 f6f5');
INSERT INTO evaluations VALUES(6851,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(6851,31,'cp 287',NULL);
INSERT INTO evaluations VALUES(6851,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(6851,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(6851,34,'cp 279',NULL);
INSERT INTO evaluations VALUES(6851,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(6851,36,'cp 278','f3e5 d4e6 e5f3 d7d6 e1g1 g8f6 c2c3 e8g8 d3d4 c5b6 f1e1 f8e8 e4e5 d6e5 f3e5 e8e7 b1a3 c7c6 d1f3 f6d5 a3c2 b6c7 c1d2 e6f8 e5d3 f8g6 e1e7 g6e7 a1e1 c8f5 g2g3 h7h6 c4b3 a7a5 a2a4 d8d6 d3e5 d6f6 c3c4 c7e5 d4e5');
INSERT INTO evaluations VALUES(6836,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(6836,31,'cp 268',NULL);
INSERT INTO evaluations VALUES(6836,32,'cp 274',NULL);
INSERT INTO evaluations VALUES(6836,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(6836,34,'cp 276',NULL);
INSERT INTO evaluations VALUES(6836,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(6836,36,'cp 303','d4e6 f2f4 d8h4 g2g3 h4h3 f4f5 h3g2 d1f3 g2f3 e5f3 e6d4 f3d4 c5d4 c2c3 d4b6 a2a4 a7a5 h1f1 c7c6 c1f4 g8f6 f4d6 b6a7 b1d2 a7b8 d6b8 a8b8 e1c1 d7d5 c4b3 e8g8 f1f4 c8d7 d1e1 d5e4 d2e4 f6d5 b3d5 c6d5');
INSERT INTO evaluations VALUES(7257,30,'cp 247',NULL);
INSERT INTO evaluations VALUES(7257,31,'cp 259',NULL);
INSERT INTO evaluations VALUES(7257,32,'cp 245',NULL);
INSERT INTO evaluations VALUES(7257,33,'cp 246',NULL);
INSERT INTO evaluations VALUES(7257,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(7257,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(7257,36,'cp 238','d3d4 c5a7 f3e5 e8g8 e5c6 d7c6 b1d2 c6c5 e4e5 f6h7 d2b3 c5d4 c3d4 b7b5 c4d3 f7f5 d3c2 c8e6 f2f4 d8e8 c1e3 a8d8 f1f2 a6a5 f2d2 a5a4 b3c5 a7c5 d4c5 g7g5 d2d8 e8d8 b2b3 g5f4 d1d8 f8d8 e3f4 h7f8 b3a4 b5a4 c2a4');
INSERT INTO evaluations VALUES(7265,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(7265,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(7265,32,'cp 239',NULL);
INSERT INTO evaluations VALUES(7265,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(7265,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(7265,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(7265,36,'cp 236','c5a7 f3e5 e8g8 e5c6 d7c6 b1d2 c6c5 e4e5 f6h7 d2b3 c5d4 c3d4 b7b5 c4d3 f7f5 d1c2 a7d4 f1d1 c8e6 d3f1 e6b3 c2b3 g8h8 c1e3 c7c5 e3d4 c5d4 a2a4 b5a4 a1a4 a8b8 b3a3 d8c7 f2f4 g7g5 a4d4 g5f4 d4d6 c7g7 e5e6 h7f6 d6a6 b8e8 a3f3 g7c7 a6d6 f6e4');
INSERT INTO evaluations VALUES(11407,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(11407,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(11407,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(11407,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(11407,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(11407,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(11407,36,'cp 48','b5c6 b7c6 d2d4 f6d7 c1e3 a6a5 c3a4 f8e7 d4e5 d6e5 d1d3 e7d6 f1d1 e8g8 e3g5 d8e8 f3h4 d7c5 a4c5 d6c5 h4f5 c8f5 e4f5 e5e4 d3g3 f7f6 g5e3 c5d6 e3f4 a8d8 d1e1 e8e7 a1d1 f8e8 g3g4 d6f4 g4f4 d8d1 e1d1');
INSERT INTO evaluations VALUES(11358,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(11358,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(11358,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11358,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(11358,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(11358,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(11358,36,'cp 54','b7c6 d2d4 f6d7 c1e3 a6a5 c3a4 f8e7 d4e5 d6e5 d1d2 e8g8 d2c3 e7d6 f1d1 a8b8 b2b3 c8a6 c3d2 a6b5 a4b2 b5a6 d1e1 d6b4 c2c3 b4d6 a1d1 c6c5 e3g5 d8c8 f3h4 c5c4 h4f5 f7f6 d2d5 f8f7 g5e3 c4b3 a2b3 d6f8 b2c4');
INSERT INTO evaluations VALUES(11666,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(11666,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(11666,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11666,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(11666,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(11666,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(11666,36,'cp 37','d2d4 f6d7 c1e3 a6a5 c3a4 f8e7 d4e5 d6e5 d1d2 e8g8 d2c3 e7d6 f1d1 a8b8 b2b3 c8a6 c3d2 b8a8 c2c4 c6c5 d2d5 d7f6 d5c6 d8d7 f3e5 d7c6 e5c6 f6e4 d1d6 e4d6 c6e7 g8h8 e7d5 f8c8 a1d1 d6f5 a4c5 f5e3 d5e3 c8e8 d1d7 f7f5 c5d3 a5a4');
INSERT INTO evaluations VALUES(11662,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11662,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(11662,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(11662,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(11662,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(11662,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(11662,36,'cp 37','f6d7 c1e3 a6a5 c3a4 f8e7 d4e5 d6e5 d1d3 e7d6 f1d1 e8g8 d3c3 a8b8 c3a5 f7f5 c2c4 d6b4 a5a7 d8f6 e3g5 f6g6 e4f5 g6f5 a7e3 e5e4 f3d4 f5g6 a2a3 b4d6 d4c6 d6h2 g1h2 g6c6 a4c3 b8b2 e3e4 c6e4 c3e4 d7e5 d1d8 e5c4 d8f8 g8f8');
INSERT INTO evaluations VALUES(11511,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11511,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(11511,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(11511,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(11511,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(11511,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(11511,36,'cp 12','f8c5 c3d5 d7d6 c2c3 h7h6 d2d4 c5a7 d4e5 c6e5 f3e5 d6e5 c1e3 f6d5 d1h5 e8g8 a1d1 d8f6 e4d5 a7e3 f2e3 f6e7 h2h3 c8d7 c4b3 a6a5 b3c2 e5e4 d5d6 c7d6 h5d5 f8e8 d5d6 d7c6 d6e7 e8e7 d1d2');
INSERT INTO evaluations VALUES(11819,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(11819,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(11819,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(11819,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11819,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11819,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11819,36,'cp 0','d8d5 b5c6 d5c6 d1f3 c6c2 f3a8 c2c1 e1e2 c1c4 d2d3 c4g4 g5f3 f8d6 a8c6 c8d7 c6a8');
INSERT INTO evaluations VALUES(21928,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(21928,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(21928,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(21928,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(21928,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(21928,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21928,36,'cp 28','b1c3 g7g6 a1b1 a7a5 e2e3 f8g7 g1e2 g8e7 e1g1 h7h5 h2h3 c8e6 d2d3 d8d7 c3d5 e8g8 g1h2 e6d5 c4d5 c6b4 e2c3 c7c6 d5c6 b7c6 h3h4 d6d5 a2a3 b4a6 c3a4 a8b8 c1d2 b8b5 d1c2 f8b8 f1c1 g8h7 b2b3 d7d6 a4c3 b5b6 e3e4 a6c7 e4d5');
INSERT INTO evaluations VALUES(24919,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(24919,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(24919,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24919,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24919,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(24919,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24919,36,'cp 23','g1e2 e7e6 f2f3 g4h5 b1c3 h5g6 g2g4 a7a6 h2h4 f8d6 a2a3 g6d3 d1d3 d6f4 e2f4 d8d6 e1c1 e8c8 f4e2 c8b8 c1b1 h7h6 e3e4 d5e4 c3e4 d6e7 h4h5 f6e4 f3e4 d8d7 d3e3 c6a5 h1e1 h8d8 e3c3 a5c6 d1d3 e7g5 e1g1 g5h4 g1g3 h4e7');
INSERT INTO evaluations VALUES(34566,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(34566,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(34566,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34566,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(34566,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(34566,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(34566,36,'cp 80','c8f5 d2d4 b8c6 g1f3 e8c8 f3h4 d8d4 h4g6 d4d1 c3d1 h7g6 d1e3 f5d7 c2c3 a7a5 e1g1 e7e5 h2h4 g8e7 f1d1 e7f5 b2b3 f8c5 d1d5 c5a7 e2f3 f7f6 e3f5 g6f5 c1e3 a7e3 f2e3 g7g5 h4g5 e5e4 f3e2 f6g5 a1d1 d7e6');
INSERT INTO evaluations VALUES(32841,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(32841,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(32841,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(32841,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(32841,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(32841,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(32841,36,'cp 93','d2d4 b8c6 g1f3 e7e6 d4d5 e6d5 f3h4 g6e6 h4f5 e6f5 e2g4 f5g6 c3d5 f8d6 e1g1 h7h5 f1e1 g8e7 g4h3 a8d8 c2c3 e8g8 d5e7 c6e7 d1f3 h5h4 h3f1 f8e8 f1d3 g6f6 f3f6 g7f6 d3c4 h4g3 h2g3 a7a5 g1g2 g8g7 e1h1 e7g6 c1e3 b7b6 a1d1 g6e5 c4e2');
INSERT INTO evaluations VALUES(38705,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38705,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38705,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38705,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38705,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(38705,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38705,36,'cp 0','g8f6 b1c3 e8g8 d2d3 a7a5 f4e5 d6e5 c1g5 h7h6 g5h4 b8c6 d1d2 c8e6 c3d5 e6d5 e4d5 c6a7 d3d4 e5d4 e1g1 d8d6 h4g3 d6d8');
INSERT INTO evaluations VALUES(41875,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41875,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(41875,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(41875,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41875,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41875,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41875,36,'cp 0','d2d3 b8d7 b1c3 e8g8 f4e5 d7e5 f3e5 d6e5 d1f3 h7h6 c1e3 c5d4 e1g1 c8e6 c4e6 f7e6 c3d1 d4e3 f3e3 f6g4 f1f8 d8f8 e3e2 f8c5 g1h1 a8f8 g2g3 g4f2 h1g2 f2d1 a1d1 c5b4 c2c3 b4d6 a2a3 a7a5 h2h4 d6d7 d1f1 f8f1 g2f1 a5a4 f1g2 d7c6 g2f3 g8h7 f3e3 c6b6 e3f3');
INSERT INTO evaluations VALUES(41874,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41874,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41874,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41874,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41874,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41874,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41874,36,'cp 0','b8d7 b1c3 c7c6 f4e5 d7e5 d3d4 e5f3 g2f3 c5b4 h1g1 d6d5 c4b3 d5e4 g1g7 f6d5 c1g5 d5c3 b2c3 b4c3 e1e2 d8d4 d1d4 c3d4 g7f7 d4a1 f7e7 e8f8 e7f7 f8e8 f7e7');
INSERT INTO evaluations VALUES(39148,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(39148,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(39148,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(39148,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(39148,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(39148,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(39148,36,'cp 86','c2c3 b4d6 f4d6 c7d6 e2e4 g8e7 b1a3 d6d5 e4e5 d7d6 f2f4 d8b6 f1d3 c8d7 d1e2 d6e5 f4e5 b8c6 g1f3 e7g6 e2d2 e8c8 e1c1 f7f6 h1e1 d8f8 a3c2 c8b8 h2h4 f6e5 d3g6 h7g6 d4e5 d7e8 c1b1 g6g5 d2g5 e8h5 g5g3 f8c8 e1f1 h8f8 b1a1 h5f3');
INSERT INTO evaluations VALUES(39147,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(39147,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(39147,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(39147,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(39147,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(39147,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(39147,36,'cp 63','b4d6 f4d6 c7d6 e2e4 g8e7 f1d3 d6d5 e4e5 d7d6 f2f4 d8b6 d1b3 c8d7 g1f3 d6e5 f4e5 b8c6 e1g1 f7f6 f1e1 e7g6 e5f6 g7f6 d3g6 h7g6 b3d5 e8c8 d5b3 g6g5 b3b6 a7b6 h2h3 d8g8 b1d2 g5g4 h3g4 g8g4');
INSERT INTO evaluations VALUES(40573,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(40573,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40573,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40573,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(40573,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40573,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(40573,36,'cp 0','e8g8 b1d2 d8c7 d4c5 c7c5 f1d3 b8d7 e1g1 d6g3 h2g3 b7b6 e3e4 d5e4 d2e4 c5c7 d1e2 c8b7 a1d1 a8d8 f1e1 h7h6 e4f6 d7f6 f3e5 d8d6 e2e3 f8d8 e3f4 f6h5 f4g4 h5f6');
INSERT INTO evaluations VALUES(41267,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(41267,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41267,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(41267,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41267,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41267,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(41267,36,'cp 23','b1c3 e8g8 g1f3 c7c5 h2h3 c5d4 e3d4 d6d5 f1e2 d5c4 e2c4 b8c6 d4d5 c6a5 c4e2 b7b5 e2b5 c8b7 d5d6 f6h5 f4h2 e7d6 e1g1 a8c8 d1d6 b7f3 d6d8 f8d8 g2f3 g7c3 b2c3 c8c3 a1d1 a5c6 f1e1 h5f6 b5c6 d8d1 e1d1 c3c6 d1d8 g8g7 h2e5 g7h6');
INSERT INTO evaluations VALUES(43814,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(43814,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(43814,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(43814,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(43814,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43814,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43814,36,'cp -3','h7h6 g5h4 c8b7 b1c3 f8e7 e2e3 e8g8 f1e2 f6e4 h4e7 d8e7 c3e4 b7e4 e1g1 c7c5 a1c1 f8d8 e2d3 e4d3 d1d3 b8c6 f1d1 c5d4 f3d4 c6d4 d3d4 d7d5 h2h4 a8c8 c4d5 c8c1 d1c1 d8d5 d4f4 d5c5 c1c5 e7c5 g1h2 c5c1 b2b3 c1d2 f4b8 g8h7 b8a7 d2f2');
INSERT INTO evaluations VALUES(46297,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46297,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(46297,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46297,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(46297,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(46297,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(46297,36,'cp 27','b8c6 f1d3 h7h6 g5f4 a7a6 e1g1 f8d6 c3e2 c6d4 f3d4 e6e5 f4g3 e5d4 e2d4 e8g8 g3d6 d8d6 c2c3 f8e8 d1c2 f6g4 d4f3 c8d7 f1e1 d6f4 a1d1 d7a4 c2c1 f4d6 e1e8 a8e8 d1e1 g4e5 e1e5 e8e5 f3e5 d6e5 g2g3 e5d6 a2a3 d6b6 c1d2 a4d7');
INSERT INTO evaluations VALUES(13625,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(13625,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13625,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13625,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13625,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13625,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(13625,36,'cp 35','f1d3 h7h6 g5f4 a7a6 e1g1 f8d6 c3e2 c6d4 f3d4 e6e5 f4g3 e5d4 g3d6 d8d6 e2d4 e8g8 d1d2 c8d7 f2f3 d6b6 c2c3 f8e8 f1e1 e8e1 a1e1 a8e8 e1e2 g8f8 g2g3 e8e2 d2e2 a6a5 d3b5 a5a4 a2a3 d7b5 e2b5 b6b5 d4b5 f8e7 b5d4 g7g6 g1f2');
INSERT INTO evaluations VALUES(2041,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(2041,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(2041,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(2041,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(2041,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(2041,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(2041,36,'cp 39','c6e5 b1c3 c7c6 e3e4 f7f5 d1d4 e5f7 c1e3 f5e4 c3e4 d7d5 e4g5 f7g5 f3g5 h7h6 g5f3 e7b4 c2c3 b4d4 e3d4 g8f6 e1c1 f8d6 d1e1 e8d8 f3e5 h8f8 e5g6 f8e8 f1d3 c6c5 d4e5 d6e5 g6e5 d8c7 f2f3 c8d7 e1e2 e8e7 e5d7 c7d7 d3b5 d7d6 e2e7 d6e7 h1e1 e7d6 c3c4');
INSERT INTO evaluations VALUES(2084,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(2084,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(2084,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(2084,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(2084,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(2084,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(2084,36,'cp 48','e3e4 f7f5 b1c3 c7c6 f3e5 e7e5 f2f4 e5e7 d1e2 f5e4 c3e4 e7e6 f4f5 e6e5 g2g4 d7d5 e4g3 f8d6 c1d2 g8f6 e1c1 e5e2 f1e2 e8g8 d1e1 h7h6 e1f1 f8e8 e2d1 c8d7 h1g1 d6g3 g1g3 f6e4 g3g2 e4d2 c1d2 g8f7 b2b3');
INSERT INTO evaluations VALUES(18788,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(18788,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(18788,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(18788,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(18788,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(18788,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(18788,36,'cp 159','d4d5 c6e7 b1c3 d7d6 d5e6 f7e6 c1f4 e6e5 f4g5 g8f6 e2e3 c8e6 f3d2 d8b6 f1e2 e8c8 a2a4 b6a5 e1g1 h7h6 g5f6 g7f6 b2b4 c5b4 d2b3 a5b6 a4a5 b6c6 c3a2 h8g8 e2f3 d6d5 a2b4 c6d6 c4d5 e7d5 b4d5 e6d5 g1h1 d5f3 d1f3');
INSERT INTO evaluations VALUES(18863,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(18863,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(18863,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(18863,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(18863,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(18863,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(18863,36,'cp 132','c6e7 b1c3 d7d6 d5e6 c8e6 f3g5 g8f6 g5e6 f7e6 g2g3 d6d5 f1g2 d5d4 c3e4 f6e4 g2e4 d8d7 e1g1 e7f5 a1b1 g7g6 e4c2 a7a6 c1d2 f8e7 b2b4 e8g8 c2e4 a8b8 b4c5 e7c5 d2f4 f5d6 f4d6 d7d6 b1b7 b8b7 e4b7 f8b8');
INSERT INTO evaluations VALUES(20643,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(20643,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(20643,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(20643,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(20643,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(20643,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(20643,36,'cp 54','f5e3 g8f6 b1c3 c8e6 a2a4 a7a6 f1d3 c6b4 e1g1 a8c8 d3e2 f8e7 c3d5 b4d5 e4d5 e6d7 a4a5 e8g8 a1a3 c8c7 c2c4 f6e8 e3c2 e7g5 f2f4 g5f4 c1f4 e5f4 f1f4 f7f5 f4f1 e8f6 b2b4 f8e8 a3e3 e8e3 c2e3 g7g6 e2d3 d8e7 f1e1');
INSERT INTO evaluations VALUES(21349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21349,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21349,36,'cp 0','d5d4 b5c6 b7c6 d2d3 c6c5 b1d2 d8d5 d2e4 f7f5 e5f6 g8f6 c2c4 d4c3 e4c3 d5f5 e1g1 f8d6 f3h4 f5e5 h4f3');
INSERT INTO evaluations VALUES(24784,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(24784,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(24784,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(24784,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(24784,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(24784,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(24784,36,'cp -10','a8c8 f3e5 c6b4 b5a3 c7c5 c2c3 b4c6 e2e3 c5c4 g2g4 f5g6 h2h4 d8b6 b2b4 c6e5 d4e5 b6a6 a3b1 f6d7 d1d5 e7e6 d5f3 g6d3 a2a4 f8e7 f1g2 c8c7 b1d2 e8g8 h4h5 h7h6 f3g3 f7f6 e5f6 e7f6 f4d6 f6c3 d6f8 c3a1 g3c7');
INSERT INTO evaluations VALUES(56,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(56,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(56,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(56,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(56,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(56,36,'cp -3','e2e3 e7e6 f3e5 c6b4 b5a3 f6d7 e5d7 d8d7 c2c3 b4c6 a3b1 f8d6 f4d6 c7d6 f1e2 e8g8 e1g1 e6e5 b1d2 c6e7 f2f4 d7c7 d2f3 f7f6 a2a4 a7a6 f3e1 g7g6 g2g4 f5e4 e1f3 e5f4 e3f4 f6f5 f3g5 g8h8 a4a5');
INSERT INTO evaluations VALUES(37576,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(37576,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(37576,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(37576,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(37576,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(37576,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(37576,36,'cp 103','h2h4 h7h5 g1f3 g8f6 b1c3 d7d6 f1e2 e8g8 f3g5 b8c6 d4d5 c6e5 c1f4 a7a6 e1g1 c8d7 b2b3 c7c6 d1d2 c6d5 e4d5 d7f5 a1c1 f6d7 c3e4 f5e4 g5e4 d7c5 e4c3 d8d7 f4g5 a6a5 c1e1 a8e8 f2f3 g7f6 c3b5 f6g7');
INSERT INTO evaluations VALUES(38486,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(38486,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(38486,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(38486,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(38486,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(38486,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(38486,36,'cp -112','d8d1 e1d1 b8c6 f1b5 c8d7 c1f4 a7a6 b5c6 d7c6 b1d2 e7e6 c2c3 h7h6 h2h4 g8e7 h1e1 e8c8 d1c2 d8d7 d2e4 h8d8 e4c5 d7d5 c5e4 e7f5 g2g3 c6b5 e4f6 b5d3 c2c1 d5c5 a2a4 a6a5 f4d2 d3c4 d2f4 c4a6 f3d2 c5c6 d2e4 a6d3 b2b3 f5d4 c1b2 d4c2');
INSERT INTO evaluations VALUES(33067,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(33067,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(33067,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(33067,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(33067,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(33067,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(33067,36,'cp -117','e1d1 b8c6 f1b5 c8d7 c1f4 e8c8 b1d2 f7f6 h1e1 g6g5 f4e3 g8h6 e5e6 d7e8 h2h3 a7a6 b5d3 c6b4 d1e2 e8g6 d2e4 d8d5 a1d1 g5g4 h3g4 h6g4 e4c3 d5d6 d3g6 h7g6 d1d6 e7d6 e1d1 g4e3 f2e3 f6f5 e3e4 g7c3 b2c3 b4a2');
INSERT INTO evaluations VALUES(33066,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(33066,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(33066,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(33066,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(33066,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(33066,35,'cp -110',NULL);
INSERT INTO evaluations VALUES(33066,36,'cp -124','b8c6 f1b5 c8d7 h1e1 a7a6 b5c6 d7c6 c1f4 h7h6 h2h4 c6f3 g2f3 e8c8 b1d2 e7e6 c2c3 g8e7 f4g3 e7f5 d1c2 h6h5 f3f4 g7f8 d2f3 f8e7 b2b3 c7c5 a2a4 c8c7 a4a5 d8d7 a1d1 d7d1 e1d1 h8d8 d1e1 c7c6 c3c4 d8d7 e1g1 e7d8 g1a1');
INSERT INTO evaluations VALUES(163,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(163,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(163,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(163,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(163,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(163,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(163,36,'cp -125','f1b5 c8d7 h1e1 a7a6 b5c6 d7c6 c1f4 h7h6 h2h4 c6f3 g2f3 e7e6 b1d2 g8e7 c2c3 e8c8 f4g3 e7f5 d1c2 h6h5 f3f4 g7f8 b2b3 f8e7 d2f3 c7c5 a2a4 c8c7 c3c4 c7c6 e1d1 d8d1 a1d1 b7b5 a4b5 a6b5 d1a1 c6b6 c2c3 h8b8 c3d3 b8f8 c4b5 b6b5 f3g5 e7g5 f4g5');
INSERT INTO evaluations VALUES(40614,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(40614,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(40614,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(40614,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(40614,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(40614,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(40614,36,'cp -22','f1d3 b8c6 e1g1 e8g8 e3e4 d5d4 c3e2 e6e5 a2a3 a7a5 c2c3 d4c3 e2c3 h7h6 d3c4 c8g4 h2h3 g4h5 d1d8 a8d8 f3h4 g8h7 g2g4 h5g6 h4g6 h7g6 g1g2 d8d7 f2f4 c5d4 g4g5 h6g5 f4g5 d4c3 g5f6');
INSERT INTO evaluations VALUES(40615,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(40615,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(40615,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(40615,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(40615,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(40615,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(40615,36,'cp -40','b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 h2h3 h7h6 a2a3 a7a5 c1d2 f8e8 c2c3 d4c3 d2c3 d8b6 e2g3 c8e6 d1e2 c6d4 f3d4 c5d4 c3d4 b6d4 f1d1 e8d8 a1c1 a8c8 c1c8 d8c8 g3f5 e6f5 e4f5 e5e4 d3b5');
INSERT INTO evaluations VALUES(52281,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(52281,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52281,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52281,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52281,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52281,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52281,36,'cp 14','f8g7 g2g3 b8c6 f1g2 d7d6 e1g1 c8g4 h2h3 g4f3 g2f3 e7e6 f3g2 g8e7 d2d3 a8c8 a1b1 e8g8 a2a3 b7b6 d1a4 a7a5 f1d1 d6d5 c4d5 e7d5 c3d5 e6d5 e2e3 h7h6 d3d4 c5d4 e3d4 b6b5 a4b5 c6d4 b5d5 d4e2 g1h2 d8d5 g2d5 e2c1 d1c1');
INSERT INTO evaluations VALUES(56538,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(56538,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(56538,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(56538,33,'cp -119',NULL);
INSERT INTO evaluations VALUES(56538,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(56538,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(56538,36,'cp -100','h2h4 g8h6 f1c4 f8e7 b1c3 c8g4 c4f7 e8d7 f2f3 e7g5 h4g5 e4f3 g2f3 h6f7 f3g4 b8c6 d2d3 f7g5 h1h5 g5e6 c1e3 d8e7 d1f3 a8f8 f3h1 e7f7 c3e4 d7c8 e1c1 f7g6 d1g1 c8b8 c1b1 h7h6 b2b3 c6d4 e3d4 e5d4 b1b2');
INSERT INTO evaluations VALUES(3496,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(3496,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(3496,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(3496,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(3496,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(3496,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(3496,36,'cp 83','e7e5 d4d5 c6e7 f3e1 a7a5 c1e3 f6d7 f2f3 b7b6 e1d3 f7f5 a2a3 g8h8 b2b4 c8a6 d3f2 e7g8 d1c1 g8f6 e3d2 a6c8 c3b5 c8a6 c1c2 f6g8 b4a5 b6a5 a3a4 g7h6 d2a5 a6b5 a4b5 h6e3 a5d2 a8a1 f1a1 e3d4 d2c3 d4e3 e4f5 g6f5');
INSERT INTO evaluations VALUES(6227,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(6227,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(6227,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(6227,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(6227,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(6227,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(6227,36,'cp 0','g8f6 d2d3 e8g8 e1g1 a7a6 c1e3 d7d6 c3d5 c5e3 f2e3 c8e6 d5f6 d8f6 f3d4 f6g5 d4e6 g5e3 g1h1 f7e6 c4e6 g8h8 e6f7 c6d8 f7c4 d8c6');
INSERT INTO evaluations VALUES(6987,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(6987,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(6987,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(6987,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(6987,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(6987,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(6987,36,'cp 40','d1b3 e8f8 c4f7 c6a5 b3a4 f8f7 a4a5 c7c6 a5f5 g8f6 e4e5 d7d6 e5e6 f7f8 e1g1 d8e8 f1e1 e8g6 f5h3 d4c3 b1c3 g6g4 h3g4 f6g4 c1b2 f8e8 h2h3 g4f6 f3d4 f6h5 d4f5 h8f8 g2g4 f8f5 g4f5 g7g6 c3e2 g6f5 e2d4 e7f6 a1b1 h5f4 d4f5 f6b2 b1b2');
INSERT INTO evaluations VALUES(6985,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(6985,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6985,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(6985,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(6985,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(6985,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(6985,36,'cp 10','e8f8 c4f7 c6a5 b3a4 f8f7 a4a5 c7c6 a5f5 g8f6 e4e5 d7d6 e5e6 f7f8 e1g1 d8e8 f1e1 e8g6 f5h3 d4c3 b1c3 g6g4 h3g4 f6g4 c1f4 g4e5 f3e5 d6e5 e1e5 e7f6 e5e3 f6d4 e3f3 d4c3 f4e5 f8e7 f3c3 c8e6 e5g7 h8g8 g7e5 e6d5 f2f3 b7b5 g1f2 b5b4 c3e3 a7a5 g2g4 c6c5 a1d1 d5a2 e5g3 a2e6 f3f4 g8g4');
INSERT INTO evaluations VALUES(7389,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(7389,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(7389,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(7389,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(7389,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(7389,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(7389,36,'cp -31','c6a5 b1c3 e8g8 d2d3 c8b7 c3e2 f8e8 e2g3 e7f8 c2c4 c7c5 c1d2 g7g6 g3f1 f8g7 f1e3 a5b3 d1b3 b7c6 a2a3 f6d7 e3d5 h7h6 b3c2 d7f8 b2b4 f8e6 b4c5 d6c5 a1d1 g8h7 d2c3 f7f6 f3d2 c6b7 d2b3 d8d6');
INSERT INTO evaluations VALUES(7403,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(7403,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(7403,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(7403,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(7403,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(7403,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(7403,36,'cp -20','b1c3 e8g8 d2d3 f8e8 c1g5 c7c6 c3e2 h7h6 g5d2 e7f8 e2g3 c8d7 c2c3 a5b3 a2b3 c6c5 d3d4 c5d4 c3d4 a6a5 a1c1 a5a4 b3a4 b5a4 d2c3 d8b8 d1d3 d7b5 d3d2 b5a6 g3f5 a6c8 d4e5 d6e5 c3e5 e8e5 c1c8 b8c8 f3e5');
INSERT INTO evaluations VALUES(13185,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(13185,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(13185,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(13185,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(13185,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(13185,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(13185,36,'cp -9','e7e5 f1e2 c8f5 d2d3 f8b4 a2a3 b4c3 b2c3 e8g8 g1f3 e5e4 d3e4 f6e4 d1d8 f8d8 f3d4 f5d7 e2f3 e4c3 c1e3 c6e5 f3b7 a8b8 b7d5 d8c8 e1g1 e5c4 d5c4 c8c4 f1c1 h7h6 h2h3 b8c8 f2f3 g8h7 c1c2 c3d5 c2c4 c8c4 e3f2 d5f4 a1e1 a7a5 d4e2 f4e2 e1e2');
INSERT INTO evaluations VALUES(13810,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13810,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13810,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(13810,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(13810,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13810,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13810,36,'cp 0','d2d3 f8b4 g1e2 c8e6 a2a3 b4e7 e2g1 e8g8 g1f3 d8c7 f1e2 a8d8 c1d2 c6d4 f3d4 e5d4 c3b5 c7b6 a3a4 a7a6 b5a3 b6b2 a3c2 e6c4 a1b1 b2a2 b1a1 a2b2');
INSERT INTO evaluations VALUES(13811,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13811,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13811,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13811,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13811,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13811,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(13811,36,'cp 0','f8b4 g1e2 e8g8 a2a3 b4c3 e2c3 c8f5 c1g5 h7h6 g5f6 d8f6 f1e2 a8d8 c3d5 f6d6 e1g1 f5d7 e2f3 c6d4 a1c1 f7f5 g2g3 d4f3 d1f3 d7c6 f1e1 b7b6 b2b4 c6d5 c4d5 d6d5 f3d5 d8d5 a3a4 f8f7 c1c8 f7f8');
INSERT INTO evaluations VALUES(18590,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(18590,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(18590,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(18590,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(18590,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(18590,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(18590,36,'cp 45','e7e5 f1c4 f8e7 a2a4 g8f6 h2h3 e8g8 e1g1 f8e8 a4a5 d8c7 f3g5 e8f8 c1e3 e5d4 e3d4 h7h6 g5f3 f8e8 f3d2 c7b8 b2b4 d6d5 e4d5 e7b4 d5c6 b7c6 f1e1 e8e1 d1e1 b8d6 d2f3 c6c5 d4f6 d7f6 c3e4 b4e1 e4d6 e1c3 a1b1 c8e6 c4e6');
INSERT INTO evaluations VALUES(19979,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(19979,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(19979,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19979,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(19979,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(19979,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(19979,36,'cp 25','g1f3 g7g6 d2d4 c5d4 f3d4 f8g7 d4c6 b7c6 f1g2 c8d7 e1g1 g8f6 b2b3 d8c8 c1b2 e8g8 d1d2 d7h3 c3a4 h3g2 g1g2 c8e6 a1d1 f6d7 h2h4 g7b2 a4b2 a7a5 f1e1 f8b8 h4h5 g6h5 b2d3 e6g6 e2e4');
INSERT INTO evaluations VALUES(19978,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(19978,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(19978,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(19978,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(19978,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(19978,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(19978,36,'cp 20','g7g6 d2d4 c5d4 f3d4 f8g7 d4c6 b7c6 f1g2 c8d7 c1f4 g8f6 d1d2 e8g8 e1g1 d8c8 f1d1 d7h3 b2b3 h3g2 g1g2 c8e6 a1c1 f6d7 f2f3 f8b8 c3e4 b8e8 e4g5 e6f6 h2h4 a7a5 g5e4 f6e6 f4h6 g7h6 d2h6 d7f6 e4c3 e6e5 e2e4 f6d7');
INSERT INTO evaluations VALUES(25569,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(25569,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(25569,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(25569,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(25569,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(25569,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(25569,36,'cp 50','b1d2 e7e6 g1f3 f8d6 c2c3 d8e7 e1g1 f6e4 d2e4 d5e4 f3d2 f7f5 f2f3 e6e5 f4g3 e5d4 g3d6 c7d6 c3d4 e4f3 d1f3 e8g8 b5c4 g8h8 a2a3 f5f4 e3f4 e7f6 d4d5 c6a5 b2b3 f6d4 f3f2 f8f4 f2d4 f4d4 f1f7 d4d2 f7d7 a5c4 b3c4 d2d4 d7d6 d4c4');
INSERT INTO evaluations VALUES(29455,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(29455,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(29455,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(29455,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(29455,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(29455,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(29455,36,'cp 112','d2d4 d6d5 e4d5 f6d5 c3d5 d8d5 f1e2 e7e6 e1g1 d5d6 a2a4 f8e7 a4a5 e8g8 h2h3 g4h5 c2c3 b8d7 c1e3 c7c6 d1d2 d6c7 f1c1 a8d8 b2b4 h5f3 e2f3 d7f6 c3c4 c7d6 c4c5 d6c7 e3f4');
INSERT INTO evaluations VALUES(29448,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(29448,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(29448,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(29448,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(29448,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(29448,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(29448,36,'cp 109','e7e6 h2h3 g4h5 d1e2 h5g6 h3h4 h7h5 c1g5 f8e7 e4e5 f6g4 d4d5 d6e5 e1c1 b8d7 d5e6 f7e6 e2c4 e7f6 c4e6 d8e7 e6e7 f6e7 g5e7 e8e7 d1d2 a8f8 f1c4 c7c6 h1d1 d7b6 c4b3 f8f6 a2a4 a7a5 d2e2 b6d7 c3b1 h8f8 b1d2 b7b5 d1e1 f6f4 a4b5 c6b5 b3d5');
INSERT INTO evaluations VALUES(38706,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(38706,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(38706,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(38706,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(38706,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(38706,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38706,36,'cp 0','e5d4 f3d4 d8f6 d4f3 c8e6 c4e6 f6e6 e1g1 g8f6 b1c3 e8g8 f1e1 b8d7 c1e3 c5b4 e3d2 a7a6 f3d4 e6c4 d4f5 f8e8 d1f3 d7e5 f3g3 e5g6 g3f3');
INSERT INTO evaluations VALUES(44444,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(44444,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(44444,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(44444,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(44444,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(44444,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(44444,36,'cp 74','b1c3 b8c6 d4c6 d7c6 d1d8 e8d8 f1c4 d8e8 e4e5 f6d7 e5e6 f7e6 c4e6 d7c5 e6c8 a8c8 c1e3 c5e6 e1c1 f8g7 c3e4 b7b6 c2c3 c8d8 d1e1 d8d5 g2g3 e8f7 c1c2 h8d8 e1e2 c6c5 e2d2 g7f6 h1e1 d5e5 f2f3 d8d2 e3d2 c5c4 d2e3 e6g7');
INSERT INTO evaluations VALUES(44443,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(44443,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(44443,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(44443,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(44443,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(44443,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(44443,36,'cp 73','b8c6 d4c6 d7c6 d1d8 e8d8 f1c4 d8e8 e4e5 f6d7 e5e6 f7e6 c4e6 f8g7 c1e3 d7f8 e6c8 a8c8 e3a7 f8e6 a7e3 g7c3 b2c3 h8f8 g2g3 f8f5 e1e2 f5a5 h1b1 b7b5 b1b4 c8a8 c3c4 a5a2 a1a2 a8a2 c4b5 c6b5 e2d3 a2a5 e3b6 a5a3 d3d2 e6g5 b4b5 g5e4 d2c1 e4d6 b5c5 a3a1 c1b2');
INSERT INTO evaluations VALUES(46277,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(46277,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(46277,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(46277,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(46277,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46277,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46277,36,'cp 5','c2c4 b8c6 c4d5 e6d5 b1c3 d5d4 c3a4 c8d7 a1c1 d8a5 f1e2 f8e7 e1g1 e8g8 e3d4 c5d4 f3d4 c6d4 b2d4 a8d8 e2f3 d7b5 f1e1 d8d4 e1e7 b5a4 b3a4 d4a4 e7b7 a4a2 d2d4 a5d2 d1d2 a2d2 b7a7 d2d4 f3c6 d4d3 g2g3 d3d2 c6f3 d2d3');
INSERT INTO evaluations VALUES(46159,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(46159,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(46159,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46159,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(46159,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(46159,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(46159,36,'cp 13','b8c6 c4d5 e6d5 f1e2 a7a6 d2d4 c5d4 f3d4 f8b4 b2c3 b4d6 d4c6 b7c6 d1c2 e8g8 b1d2 d8e7 e1g1 a6a5 e2d3 h7h6 c3b2 c8a6 d3a6 a8a6 a1c1 a5a4 b3a4 d6a3 d2b3 a3b2 c2b2 a6a4 c1c6 f8a8 c6c2 e7a3 b2a3 a4a3 f1c1 a3a2 c2a2');
INSERT INTO evaluations VALUES(13539,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(13539,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(13539,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13539,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(13539,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13539,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(13539,36,'cp 16','c4d5 e6d5 f1e2 a7a6 d2d4 c5d4 f3d4 f8b4 b2c3 b4d6 e1g1 d8e7 b1d2 e8g8 h2h3 c6d4 c3d4 c8f5 d2f3 f5e4 d4b2 h7h6 e2d3 a8d8 d1e2 f8e8 f1d1 f6d7 a1c1 d7c5 d3b1 c5e6 b1e4 d5e4 f3d2 f7f5 d2c4 d6c7 b2a3 e7g5 c4d6 c7d6 d1d6 d8d6');
INSERT INTO evaluations VALUES(51411,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(51411,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(51411,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51411,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(51411,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(51411,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(51411,36,'cp 20','c8b7 h2h4 g8f6 g1f3 f8e7 f1d3 c7c5 c2c3 d7d6 e1g1 b8d7 b1d2 e8g8 d1e2 a8c8 a2a4 h7h6 d2c4 d8c7 f3d2 f8e8 f1e1 e7f8 f4g3 c5d4 e3d4 e6e5 c4e3 f6d5 e3d5 b7d5 d4e5 d7e5 d3e4 e5c4 d2c4 e8e4 c4e3 d5b3 e2b5');
INSERT INTO evaluations VALUES(53377,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(53377,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(53377,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(53377,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(53377,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(53377,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(53377,36,'cp -75','b8c6 c2c3 c8f5 f1e2 e7e6 g1f3 d8c7 f3e5 f8d6 e1g1 g8e7 b1d2 f7f6 e2h5 g7g6 h5e2 e8g8 e5c6 b7c6 d2b3 a8e8 b3c5 e6e5 f4e5 f6e5 c1h6 e5d4 c3d4 d6h2 g1h1 f8f7 d1d2 h2g3 f1f3 f5c8 d2e3 f7f3 e2f3 c8f5 h6g5 c7d6 a1f1 g8h8 b2b4');
INSERT INTO evaluations VALUES(19847,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(19847,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(19847,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(19847,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(19847,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(19847,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(19847,36,'cp -70','f1e2 c8f5 g1f3 e7e6 e1g1 f8d6 c2c3 h7h6 c1e3 d8b6 d1b3 b6c7 f3e5 g8e7 b1d2 e8g8 a1e1 f7f6 e5d3 c6a5 b3a4 b7b6 g2g3 a7a6 a4d1 a8c8 a2a3 c7d7 g3g4 f5h7 d3b4 d6b4 a3b4 a5c4 d2c4 d5c4 b2b3 e7d5 e3d2 c4b3 d1b3');
INSERT INTO evaluations VALUES(11994,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11994,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(11994,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11994,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(11994,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(11994,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(11994,36,'cp -12','f8b4 f2f4 e8g8 e1g1 b4d6 g1h1 e5f4 c1f4 d6f4 f1f4 d7d6 d1d2 c8g4 a1f1 f8e8 e4e5 e8e5 c3e4 f6e4 d3e4 d6d5 e4h7 g8h7 f4g4 d8e7 c2c4 a8e8 c4d5 c6d5 g4f4 f7f6 h2h3 h7g8 a2a3 e7e6 f1d1 e8e7 f4d4 e5e2 d2c3 e2e3 c3a5 e3e1 d1e1 e6e1 a5e1 e7e1 h1h2');
INSERT INTO evaluations VALUES(7659,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(7659,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(7659,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(7659,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(7659,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7659,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(7659,36,'cp -5','e1g1 e8g8 f2f4 b4d6 d1e2 h7h6 g1h1 f8e8 f4f5 d6e7 c3d1 d7d5 d1f2 a7a5 c1e3 a5a4 a1b1 d8a5 b2b4 a4b3 a2b3 c8f5 e4f5 e5e4 e3d4 e4d3 e2d3 a5a3 b1e1 e7c5 e1e8 a8e8 d4f6 c5f2 f1f2 g7f6');
INSERT INTO evaluations VALUES(13281,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(13281,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(13281,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(13281,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(13281,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(13281,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(13281,36,'cp -13','b1c3 c8g4 f1e2 g7g6 c1e3 f8g7 e1g1 e8g8 f3e1 g4e2 d1e2 c6d4 e2d1 a7a6 a2a4 f6d7 f2f4 d7b8 e1f3 f7f5 e3d4 c5d4 c3e2 f5e4 f3g5 d8d7 g5e4 b8c6 d1d2 a6a5 e2g3 g7h6 a1e1 e7e5 e4g5 a8e8 g3e4 e5f4 f1f4');
INSERT INTO evaluations VALUES(16844,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(16844,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(16844,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(16844,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(16844,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(16844,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(16844,36,'cp -102','d7d5 c4b5 e5d4 f3d4 c8d7 c2c4 f8c5 d4b3 c6e7 b1c3 c5f2 f1f2 e4f2 g1f2 c7c6 b5a4 e8g8 f2g1 e7f5 d1f1 d5c4 b3c5 b7b6 c5d7 d8d7 a4c6 d7c6 f1f5 a8d8 c1f4 f8e8 a2a4 h7h6 a1f1 e8e6 h2h4 d8d3 h4h5 a7a6 g1h2 e6f6 f5e4 c6e4 c3e4 f6c6 e4g3 c6e6 f1c1 b6b5 a4b5 a6b5 g3f5');
INSERT INTO evaluations VALUES(15382,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(15382,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(15382,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(15382,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(15382,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(15382,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(15382,36,'cp -101','c4b5 e5d4 f3d4 c8d7 c2c4 f8c5 d4b3 c5f2 f1f2 e4f2 g1f2 c6e7 b1c3 c7c6 b5a4 e8g8 f2g1 e7f5 d1f1 d5c4 b3c5 f8e8 f1f2 b7b6 c5d7 d8d7 c1g5 h7h6 a1d1 d7e6 a4c2 h6g5 c2f5 e6h6 f2d4 h6h4 d4h4 g5h4 f5d7 e8d8 d7c6 d8d1 c3d1 a8d8 d1e3 d8d2 e3c4 d2c2');
INSERT INTO evaluations VALUES(15255,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(15255,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(15255,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(15255,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(15255,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(15255,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(15255,36,'cp -94','e5d4 f3d4 c8d7 c2c4 f8c5 d4b3 c6e7 b1c3 c5f2 f1f2 e4f2 g1f2 c7c6 b5a4 e8g8 f2g1 e7f5 d1f1 d5c4 b3c5 b7b6 c5d7 d8d7 a4c6 d7c6 f1f5 a8d8 a2a4 f8e8 c1g5 d8d3 f5b5 c6c8 a1f1 e8e6 h2h3 a7a6 b5f5 c8d7 c3e4 f7f6 g5f6 g7f6 f5g4 g8h8');
INSERT INTO evaluations VALUES(21363,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21363,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21363,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(21363,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(21363,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(21363,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(21363,36,'cp 13','e6d5 g1f3 g8f6 f1b5 f8b4 e1g1 e8g8 b5c6 b7c6 f3e5 c8a6 f1e1 f8e8 c1g5 d8d6 d1f3 e8e6 e1e3 c6c5 g5f6 e6f6 f3d5 a8d8 d5d6 f6d6 d4c5 b4c5 e3e1 d6e6 e5d3 a6d3 c2d3 d8d3 e1e6 f7e6 a1d1 d3d1 c3d1 g8f7 g1f1 h7h5 f1e2 c5b6 e2f3 f7f6 d1e3');
INSERT INTO evaluations VALUES(30579,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(30579,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(30579,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(30579,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(30579,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(30579,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(30579,36,'cp 65','f8b4 f1b5 c7c6 b5e2 b4c3 b2c3 g8e7 e1g1 b7a6 e2a6 b8a6 d1d3 a6b8 c1a3 b8d7 c3c4 d5c4 d3c4 e8g8 a3e7 d8e7 c4c6 f8c8 c6e4 b6b5 f1b1 a8b8 h2h3 h7h6 b1b3 a7a5 a2a3 d7f8 e4d3 e7c7 a1c1 f8g6 g2g3 g6e7 f3d2 b5b4 d2e4 e7f5 a3b4 b8b4 b3b4');
INSERT INTO evaluations VALUES(34737,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(34737,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(34737,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(34737,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(34737,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(34737,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(34737,36,'cp 18','b1c3 e7e2 c3e2 h7h5 h2h4 c7c6 c1f4 b8d7 e2g3 g7g6 f1d3 f8g7 c2c3 d7f8 f4d2 a7a5 e1c1 g8e7 g1h3 a5a4 d1e1 f8e6 h3g5 e8g8 a2a3 c8d7 g5e6 d7e6 d2g5 a8e8 e1e2 e7f5 d3f5 e6f5 e2e8 f8e8 g3f5');
INSERT INTO evaluations VALUES(49472,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(49472,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(49472,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(49472,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(49472,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(49472,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(49472,36,'cp -58','c7c6 b5a4 d7d5 e4d5 f6d5 a4b3 a7a5 g1f3 f8d6 c2c4 d5f6 e1g1 e8g8 h2h3 c8f5 b1c3 b8d7 b3c2 f5g6 d3d4 e5d4 c2g6 h7g6 d1d4 d6c5 d4d3 a5a4 c1d2 f8e8 d3c2 a4a3 b2b3 c5b4 a1d1 d7c5 c3a4 d8d3 c2d3 c5d3 d2b4 d3b4');
INSERT INTO evaluations VALUES(44060,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(44060,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(44060,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(44060,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(44060,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(44060,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(44060,36,'cp -50','b5a4 d7d5 e4d5 f6d5 g1f3 f8d6 e1g1 e8g8 h2h3 c8f5 f1e1 b8d7 b1d2 d6c7 d2e4 a7a5 c2c4 d5f6 a4c2 f8e8 f3h4 f6e4 h4f5 e4d6 f5d6 c7d6 c1d2 d6c5 d2c3 d8b6 d1e2 c5d4 a1d1 a5a4 a2a3 f7f5 e2d2 d7c5');
INSERT INTO evaluations VALUES(44116,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(44116,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(44116,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(44116,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(44116,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(44116,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(44116,36,'cp -54','d7d5 e4d5 f6d5 a4b3 a7a5 g1f3 f8d6 c2c4 d5f6 e1g1 e8g8 h2h3 c8f5 b1c3 f5g6 d3d4 b8d7 b3c2 e5d4 c2g6 h7g6 d1d4 d6c5 d4d3 a5a4 f1e1 a4a3 b2a3 d8a5 d3c2 c5a3 c1a3 a5a3 c3e4 f8e8 e4f6 d7f6');
INSERT INTO evaluations VALUES(51324,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(51324,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(51324,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(51324,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(51324,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(51324,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(51324,36,'cp -77','c7c6 b2b3 e7e5 d4e5 g8h6 c1d2 c8g4 f3f4 b8a6 h2h3 g4e6 b1c3 d8a5 g1f3 b5b4 c3a4 c4c3 d2c1 a6c5 a4c5 f8c5 f1c4 e6c4 f4c4 a5b5 c4b5 c6b5 e3e4 h6g8 e1e2 g8e7 h1d1 a7a6 c1g5 a8c8 f3d4 e8g8 g5e3 e7c6 d4c6 c8c6 e3c5 c6c5 f2f4 g8h8');
INSERT INTO evaluations VALUES(51651,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(51651,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(51651,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(51651,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(51651,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(51651,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(51651,36,'cp -92','b2b3 g8f6 b3c4 e7e5 f3d1 e5d4 d1d4 b5b4 d4d8 e8d8 g1f3 a7a5 f3g5 d8e8 a2a3 b8a6 g5f3 a5a4 f1e2 a6c5 c1b2 c5b3 a1a2 c6c5 e2d1 c8f5 b1d2 b3d2 f3d2 b4b3 a2a1 a8a6 d1f3 f5e6 e1e2 f6d7 h1b1 f7f6 f3b7 a6b6');
INSERT INTO evaluations VALUES(11173,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(11173,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(11173,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(11173,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(11173,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11173,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11173,36,'cp 28','e7e5 d4b3 c8e6 e1g1 a8c8 d1d2 f8e7 e2f3 c6a5 b3a5 d8a5 a2a3 b7b5 f1c1 h7h6 h2h3 a5c7 a3a4 b5b4 c3d5 e6d5 e4d5 c7a5 d2d3 e8g8 f3e2 f6d5 d3a6 a5a6 e2a6 c8c7 e3d2 e5e4 c2c4 b4c3 d2c3 d5c3 c1c3 c7c3 b2c3 e7f6 a4a5 f6c3');
INSERT INTO evaluations VALUES(11227,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(11227,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(11227,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(11227,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(11227,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(11227,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(11227,36,'cp 54','d4b3 c8e6 e1g1 a8c8 d1d2 f8e7 e2f3 c6a5 b3a5 d8a5 a2a4 f6g4 e3g5 g4f6 h2h3 h7h6 g5e3 e8g8 c3d5 a5d8 d5e7 d8e7 b2b3 f6h7 d2d3 h7g5 f3d1 f7f5 e3g5 h6g5 e4f5 e6f5 d3d2 f5e6 c2c4 f8f4 d1h5 c8d8 f1e1 g5g4 h5g4 e6g4 h3g4 f4g4');
INSERT INTO evaluations VALUES(11230,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(11230,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11230,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11230,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(11230,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(11230,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(11230,36,'cp 28','c8e6 e1g1 a8c8 d1d2 f8e7 f1d1 e8g8 f2f3 f8e8 e2f1 c6b8 a2a4 b8d7 d2f2 d8c7 a4a5 c7c6 f2d2 h7h6 b3c1 c6c7 e3f2 e8d8 d2e1 d7c5 c3d5 f6d5 e4d5 e6f5 b2b4 c5d7 c2c4 f5c2 d1d2 c2g6 c1b3 e7g5 f2e3 d8e8 e3g5');
INSERT INTO evaluations VALUES(12228,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(12228,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12228,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(12228,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(12228,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12228,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12228,36,'cp 0','d5c4 a2a4 d8a5 g5d2 e7e5 e2e3 e5d4 e3d4 f8b4 f1c4 e8g8 e1g1 d7b6 c4b3 c8f5 f1e1 a8e8 f3e5 f5e6 h2h3 e6b3 d1b3 e8e6 e5d3 b4c3 d2c3 e6e1 c3e1 a5d5 b3b4 d5f5 a1d1 f5h5 d1c1 h5f5');
INSERT INTO evaluations VALUES(16613,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(16613,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(16613,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(16613,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(16613,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(16613,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(16613,36,'cp 14','c2c4 d7d5 c4d5 f6d5 b1c3 f8e7 c3d5 d8d5 f1g2 e8g8 e1g1 d5d6 a2a3 a7a5 d1c2 a5a4 c1e3 c8e6 a1c1 d6d7 f3d2 f8b8 c2c3 e6d5 g2d5 d7d5 c3c4 b8d8 c1c3 f7f5 f1b1 d5c4 c3c4 d8d5 b1c1 d5b5 c1c2 b5d5 e3c5 e7c5 c4c5');
INSERT INTO evaluations VALUES(16458,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(16458,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(16458,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16458,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(16458,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(16458,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16458,36,'cp 6','d7d5 c4d5 f6d5 b1c3 f8e7 c3d5 d8d5 f1g2 e8g8 e1g1 d5d6 d1c2 c8e6 a2a3 a7a5 c1e3 a5a4 a1c1 d6d7 f3d2 f8b8 c2c3 h7h6 f1d1 e6d5 g2d5 d7d5 c3c4 b8d8 d2e4 d8d7 e4c5 e7c5 e3c5 d5c4 c1c4 f7f6 d1c1 a8a5 c5e3 a5b5');
INSERT INTO evaluations VALUES(17462,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(17462,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(17462,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17462,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(17462,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(17462,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(17462,36,'cp 60','f1c4 h7h6 d1d5 d8e7 d5d3 g8f6 b1c3 d7d6 e1g1 e7c7 a2a3 c8e6 c3d5 e6d5 c4d5 f6d5 d3d5 f8e7 c1e3 e8g8 c2c4 a6a5 b2b3 c6b8 f1d1 b8d7 a3a4 c7c6 a1c1 f8e8 g2g3 d7f6 d5c6 b7c6 c4c5 d6c5 f3e5 f6e4 e5c6 e7f6 d1d7 e8e6 d7c7 e4c3 c6e7 e6e7 c7e7 f6e7 c1c3');
INSERT INTO evaluations VALUES(17456,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(17456,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(17456,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(17456,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(17456,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(17456,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(17456,36,'cp 52','h7h6 b1c3 g8f6 d1d3 d7d6 a2a3 b7b5 c4a2 c8e6 c3d5 e6d5 a2d5 f6d5 e4d5 c6e7 e1g1 f7f5 f3d2 g7g5 a3a4 b5a4 a1a4 f8g7 a4a6 e8g8 d3a3 a8a6 a3a6 g8h7 d2c4 e7d5 f1d1 d8c8 a6c8 f8c8 c4d6 c8c2 d6f5');
INSERT INTO evaluations VALUES(19738,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19738,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19738,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(19738,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19738,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19738,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19738,36,'cp 6','d8a5 b5c6 b7c6 d4c5 a5c5 g1f3 c5b6 b2b4 e7e6 e1g1 a7a5 c1e3 b6c7 d1d2 g8e7 e3c5 e7g6 c5f8 g6f8 d2e3 f8g6 b1d2 e8g8 a2a3 f7f6 f1e1 f6e5 f3e5 c7e5 e3e5 g6e5 e1e5 a5b4 c3b4 g8f7 e5e3 f7e7 d2f3 e7d6 a1e1 f8f4 g2g3 f4c4 f3e5 c4c2 e5f7 d6e7 f7e5');
INSERT INTO evaluations VALUES(20369,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(20369,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(20369,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(20369,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(20369,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20369,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(20369,36,'cp 17','e4d5 e6d5 c4b5 c8d7 e1g1 g8f6 b1c3 f8e7 b5e2 e8g8 e2f3 d7e6 c1e3 a7a6 c3e2 e6g4 f1e1 g4f3 d4f3 f8e8 c2c3 h7h6 d1d3 e7d6 a1d1 a8c8 e2g3 d8d7 e3b6 c6e5 f3e5 d6e5 b6d4 e5d4');
INSERT INTO evaluations VALUES(20365,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20365,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(20365,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(20365,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(20365,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(20365,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(20365,36,'cp 6','e6d5 c4b5 c8d7 e1g1 g8f6 c2c4 f8e7 b1c3 c6d4 d1d4 d5c4 d4c4 e8g8 f1d1 d7b5 c4b5 d8c7 g2g3 a7a6 b5e2 f8e8 e2f3 a8c8 c1e3 c7c6 f3c6 c8c6 d1d3 g8f8 a1d1 e7c5 e3f4 c5b4 g1g2 b4c3 d3c3 c6c3 b2c3 e8c8 f4e5 f6e4 d1d7 e4c3 d7b7 c3a2 e5b2');
INSERT INTO evaluations VALUES(20549,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(20549,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(20549,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(20549,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(20549,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(20549,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(20549,36,'cp 13','c4b5 c8d7 e1g1 f8e7 c2c4 g8f6 b1c3 c6d4 d1d4 d5c4 d4c4 e8g8 f1d1 d7b5 c4b5 d8c7 g2g3 a7a6 b5e2 f8e8 e2f3 a8c8 g1g2 c7c6 f3c6 c8c6 c1e3 g8f8 h2h3 e8d8 a1c1 f6e8 g2f3 d8d1 c1d1 e7f6 e3d4 f6d4 d1d4 f8e7 c3d5 e7e6 d5e3 e8f6');
INSERT INTO evaluations VALUES(20450,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(20450,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(20450,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(20450,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(20450,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(20450,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(20450,36,'cp 10','c8d7 e1g1 g8f6 c2c4 f8e7 b1c3 c6d4 d1d4 d5c4 d4c4 e8g8 f1d1 d7b5 c4b5 d8c7 g2g3 a7a6 b5e2 f8e8 e2f3 a8c8 c1e3 c7c6 f3c6 c8c6 h2h3 g8f8 d1d3 e7c5 e3c5 c6c5 a1d1 c5c6 g1g2 f6e4 c3e4 e8e4 g2f3 e4b4 b2b3 f8e7');
INSERT INTO evaluations VALUES(25961,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(25961,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(25961,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(25961,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(25961,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(25961,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(25961,36,'cp 13','e1g1 g8f6 c2c4 f8e7 b1c3 c6d4 d1d4 d5c4 d4c4 e8g8 f1d1 d7b5 c4b5 d8c7 g2g3 a7a6 b5e2 f8e8 e2f3 a8c8 c1e3 c7c6 f3c6 c8c6 g1g2 h7h6 e3d4 g8f8 h2h3 e8d8 d4e3 d8d1 a1d1 f8e8 d1d3 e7c5 e3c5 c6c5 d3d4 g7g5 d4d6 e8e7 d6b6');
INSERT INTO evaluations VALUES(26661,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(26661,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(26661,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(26661,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(26661,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(26661,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(26661,36,'cp 80','g8f6 g1f3 e7e6 e1g1 f8d6 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 e8g8 e3e4 f7f6 a1b1 a8b8 c1e3 d8d7 c3c4 b7b6 h2h3 b8c8 b1c1 f8d8 e1d1 d7e8 d3c2 d6a3 c1b1 a3d6 d4d5 e6d5 e4d5 c6e7 f3d4 c7c6 d4e6');
INSERT INTO evaluations VALUES(31850,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31850,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31850,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31850,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(31850,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31850,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(31850,36,'cp 0','g8f6 d1b3 d8c8 c3c4 d5c4 b3c4 e7e6 b1d2 h7h6 g2g3 b8d7 f1g2 f8e7 e1g1 e8g8 b2b3 d7b6 c4c3 b6d5 c3c4');
INSERT INTO evaluations VALUES(34355,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(34355,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(34355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34355,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(34355,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(34355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34355,36,'cp 0','d5d4 d1d4 c5d4 c3b5 b8a6 g1f3 g8f6 g2g3 d4d3 e2d3 g7g6 f1g2 f8g7 f3e5 e8g8 e1g1 f8d8 d3d4 f6d5 c1d2 a6b4 f1c1 g7e5 d4e5 b4c6 d2g5 c8e6 b5c3 d5b4 c1d1 g8f8 g5h6 f8g8');
INSERT INTO evaluations VALUES(34354,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34354,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34354,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34354,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34354,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34354,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34354,36,'cp -5','d1d4 c5d4 c3b5 e8d8 g1f3 b8c6 f3d4 c6d4 b5d4 e7e5 d4f3 f8b4 c1d2 b4d2 e1d2 f7f6 e2e4 c8e6 a1c1 e6a2 d2e3 a8c8 c1a1 a2e6 a1a7 d8c7 f1a6 c8b8 a6e2 g8e7 h1c1 e7c6 a7a1 b8a8 e2b5 a8a1 c1a1 h8c8 b5c6 c7c6 h2h4 c6c7 a1d1 e6f7 d1b1 c7b6 f3d2 c8c2');
INSERT INTO evaluations VALUES(34353,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(34353,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(34353,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(34353,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(34353,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34353,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(34353,36,'cp 0','c5d4 c3b5 e8d8 g1f3 b8c6 f3d4 c6d4 b5d4 e7e5 d4b3 c8e6 e2e4 f8b4 c1d2 b4d2 e1d2 e6b3 a2b3 g8e7 a1c1 f7f6 d2e3 e7c6 c1c3 d8c7 f1b5 h8d8 b5c6 b7c6 h1c1 d8d6 c1a1 c7b6 a1a4 a8d8 a4b4 b6c7 b4a4');
INSERT INTO evaluations VALUES(34384,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34384,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(34384,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34384,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34384,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34384,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(34384,36,'cp -10','c3b5 b8a6 g1f3 g8f6 g2g3 d4d3 e2d3 g7g6 f1g2 f8g7 f3e5 e8g8 e1g1 f8d8 d3d4 a6b4 c1d2 f6d5 a2a3 b4c6 f1e1 c8e6 e5c6 b7c6 g2d5 e6d5 b5c7 a8b8 c7d5 d8d5 e1e7 g7d4 d2c3 d4c3 b2c3 b8b2 e7a7 d5d2 a1f1 b2a2 g1g2');
INSERT INTO evaluations VALUES(34352,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34352,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34352,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34352,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34352,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(34352,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(34352,36,'cp -8','b8a6 g1f3 g8f6 b5d4 f6e4 e2e3 f7f6 f3d2 e4d6 f2f4 e7e5 f4e5 f6e5 d4b5 c8d7 b5d6 f8d6 f1a6 b7a6 d2c4 d6c7 b2b3 e8e7 c1a3 e7e6 e1f2 h8d8 h1d1 d7c6 a1c1 a8c8 d1d8 c7d8 c4d2 d8b6 d2c4 c6d5 a3b2 g7g5 b2a3 c8c6 a3b2 d5c4 b3c4 c6c8');
INSERT INTO evaluations VALUES(1420,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(1420,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(1420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1420,36,'cp -4','g1f3 g8f6 b5d4 f6e4 f3d2 e4d6 f2f4 f7f6 e2e4 e7e5 f4e5 f6e5 d4b5 c8d7 b5d6 f8d6 f1a6 b7a6 d2c4 d6c5 c1e3 a8c8 e1g1 d7b5 e3c5 b5c4 f1c1 c8c5 b2b3 e8e7 c1c4 c5c4 b3c4 h8b8 a1d1 b8b2 d1d5 b2a2 d5e5 e7d6 e5g5 g7g6 h2h4 a2e2 g5a5 e2e4 a5a6 d6c5 h4h5 g6h5 a6a7 e4c4 g1h2 c5b4 a7b7 b4a3 b7a7 a3b2');
INSERT INTO evaluations VALUES(50033,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(50033,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(50033,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(50033,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(50033,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(50033,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(50033,36,'cp 196','b8c6 a2a4 d7d6 h2h3 c6a5 c4a2 c7c5 d2d3 b7b6 f3h4 d8d7 f2f4 c8b7 d1h5 e8d8 e1g1 g7g6 h5d1 f8g7 f4f5 d6d5 e4d5 e7f5 h4f5 g6f5 d1h5 f5f4 c1d2 d7e8 h5f3 f6f5 a1e1 e8g6 c3b5');
INSERT INTO evaluations VALUES(51449,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(51449,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(51449,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(51449,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(51449,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(51449,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(51449,36,'cp 141','f2f4 e5f4 g1f3 f8e7 f1c4 g8f6 e1g1 e8g8 d2d4 c8b7 e4e5 f6h5 c3d5 b7d5 c4d5 c7c6 d5b3 g7g6 f3e1 a7a5 a2a4 c6c5 e1d3 c5d4 c1f4 h5f4 f1f4 e7g5 f4f7 f8f7 d1f3 b8c6 a1f1 g8h8 b3f7 d8e7 f7c4 a8d8 f3d5');
INSERT INTO evaluations VALUES(53909,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(53909,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(53909,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(53909,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(53909,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(53909,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(53909,36,'cp 161','d4d3 b8c6 b1c3 c8e6 c1f4 g8f6 e1c1 d8a5 a2a3 f6g4 f4g3 e8c8 f3g5 h7h5 h2h3 g4e5 d3e3 h5h4 g3h2 f8e7 g5e6 f7e6 f2f4 e5d7 f1c4 c6d4 c1b1 d7b6 c4a2 e7f6 e4e5 d6e5 f4e5 f6e7 h1f1 c8b8 f1f7 d4f5 d1d8 h8d8 e3e2 f5d4 e2g4');
INSERT INTO evaluations VALUES(54248,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(54248,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(54248,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(54248,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(54248,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(54248,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(54248,36,'cp -58','d5d4 c3b1 b8c6 a2a4 e6e5 b1a3 f8d6 g2g3 h7h6 f3d2 a7a6 f1g2 b7b5 e1g1 g8f6 d2b3 a8b8 a4b5 a6b5 c2c4 d4c3 b2c3 e8g8 c1e3 d8b6 a3c2 f6g4 h2h3 c8e6 b3d2 g4e3 c2e3 c6e7 c3c4 b5b4 e3d5 e7d5 e4d5 e6f5 d2b3 f5g6 d1e2 f7f5');
INSERT INTO evaluations VALUES(54107,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(54107,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(54107,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(54107,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(54107,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(54107,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(54107,36,'cp -40','c3b1 b8c6 a2a4 e6e5 b1a3 f8d6 g2g3 h7h6 c1d2 d6c7 f1g2 a7a6 e1g1 g8e7 a3c4 c8e6 d1e2 b7b5 c4a3 a8b8 f3h4 e8g8 a4b5 a6b5 f2f4 e5f4 g3f4 e7g6 h4g6 f7g6 e4e5 d8d7 e2e4 c6e7 c2c4 b5c4 a3c4 e6c4 d3c4 b8b2');
INSERT INTO evaluations VALUES(6486,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(6486,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(6486,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6486,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(6486,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6486,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(6486,36,'cp 9','g1f3 d7d6 c3a4 c5b6 f1e2 f7f5 e1g1 g8f6 a4b6 a7b6 e4f5 c8f5 d3d4 e8g8 d4e5 d6e5 e2c4 g8h8 d1d8 a8d8 b2b3 c6d4 c1a3 f8e8 f3g5 f5g6 a3b2 d4c2 a1d1 c2d4 f2f4 c7c5 f4e5 e8e5 g5e6 e5e6 c4e6 d4e6 d1d8 e6d8 f1f6 g7f6 b2f6 h8g8 f6d8 b6b5 d8b6 c5c4 b3c4 b5c4');
INSERT INTO evaluations VALUES(9485,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(9485,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(9485,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(9485,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(9485,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(9485,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(9485,36,'cp -71','f6g4 h1f1 e8g8 c3e2 b7b5 d2d4 b5c4 d4c5 d8h4 c1f4 g4h2 f4h2 f8d8 d1c1 h4h2 c1f4 h2h5 f4c7 c8e6 c7g3 h5c5 a1d1 c5a5 g3c3 d8d1 e1d1 a5a2 d1d2 a8d8 d2e3 a2a6 e2f4 h7h6 g2g3 a6b6 e3e2 d8d4 f2f3 g8h7 f1a1 a7a6 a1h1 d4d8 f4e6 f7e6 c3c4 b6b2');
INSERT INTO evaluations VALUES(9509,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(9509,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(9509,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(9509,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(9509,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(9509,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(9509,36,'cp -52','h1f1 e8g8 c3e2 b7b5 d2d4 b5c4 d4c5 d8h4 c1f4 g4h2 f4h2 f8d8 d1c1 h4h2 c1f4 h2h5 f4c7 c8e6 c7g3 h5c5 a1d1 c5a5 g3c3 a5a2 d1d8 a8d8 e2f4 d8b8 e1d2 b8b2 f4e6 f7e6 f2f4 a7a5 f4f5 e6f5 e4f5 a2b3 f5f6 b3c3 d2c3 b2b5 f6g7 b5g5 f1f8 g8g7 f8c8 g5g2 c8c7 g7h6 c7c6 h6g5 c6c5 g5h4 c5a5 h7h5 c3c4 h4g4 c4d4 g2c2 d4e3');
INSERT INTO evaluations VALUES(10101,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(10101,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(10101,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(10101,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(10101,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(10101,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(10101,36,'cp 153','d2d4 c5e7 d4e5 f6g4 d1d5 e8g8 h2h3 g4h6 c1f4 g8h8 b1d2 h6g8 d5d3 d7d6 e5d6 c7d6 f1e1 c8e6 c4d5 g8f6 a1b1 d8c7 f3g5 c7d7 d2f3 a8d8 f4h2 c6e5 h2e5 d6e5 c3c4 d7c7 d3b3 e6d5 e4d5 h7h6 b3b7 d8d7 b7c7 d7c7 g5f7 f8f7 f3e5 h8g8 e5f7 g8f7');
INSERT INTO evaluations VALUES(10152,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(10152,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(10152,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(10152,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(10152,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(10152,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(10152,36,'cp 133','c5e7 d4e5 f6g4 d1d5 e8g8 h2h3 g4h6 c1f4 g8h8 b1d2 h6g8 f3d4 f7f6 d4c6 b7c6 d5a5 f6e5 f4e5 d7d6 e5g3 g8f6 a1d1 f6d7 c4b3 d7c5 b3c2 c5e6 f2f4 e7h4 g3h2 a8b8 a5a7 b8b6 d2c4 b6a6 a7e3 e6c5 h2g3 h4g3 e3g3 a6a2 c2b1 a2a1 c4e3 a1b1 d1b1 c5e4');
INSERT INTO evaluations VALUES(14184,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(14184,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(14184,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(14184,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(14184,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(14184,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(14184,36,'cp 53','f6d5 g1f3 e7e5 f3e5 c6e5 d4e5 c8e6 a2a3 d8g5 d1d4 a8d8 b1d2 d5b4 d4c3 b4d5 c3d3 d5b4 d3b1 b4c6 d2f3 g5h5 c1d2 c6e5 f3d4 f8e7 d4e6 f7e6 d2c3 e7f6 f1e2 h5g6 e1g1 g6b1 a1b1 e5d3 c3f6 g7f6 b1d1 d3c5 d1d8 e8d8 b2b4 c5d7 f1e1 c7c5 e1c1 c5b4 a3b4');
INSERT INTO evaluations VALUES(14422,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(14422,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(14422,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(14422,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(14422,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(14422,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(14422,36,'cp 67','f1b5 e7e5 d4e5 d5b4 a2a3 d8d1 e1d1 b4d5 g1f3 c8d7 b5c4 d5b6 c4b3 g7g6 c1d2 f8g7 d2c3 d7g4 h2h3 g4f3 g2f3 g7e5 d1e2 f7f5 f3f4 e5c3 b1c3 c6a5 b3a2 c7c6 h3h4 a5c4 h4h5 e8c8 a2c4 b6c4 b2b3 c4d6 a1b1 d6e4 c3e4 f5e4 h5g6 h7g6');
INSERT INTO evaluations VALUES(19654,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(19654,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(19654,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(19654,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(19654,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(19654,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(19654,36,'cp 76','c2c3 g8f6 d2d4 f8g7 d4c5 d6c5 d1e2 e8g8 e4e5 f6d5 f1d1 e7e6 d1e1 a7a6 a2a4 h7h6 c4d3 c8d7 d3c2 d8c7 h2h4 h6h5 c1g5 b7b5 b1d2 c5c4 a1d1 a8c8 d2f1 b5a4 f1e3 a4a3 e3d5 e6d5 b2a3 f8e8 d1d5 d7e6 d5c5 g7f8 e2e3 f8c5 e3c5');
INSERT INTO evaluations VALUES(19650,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(19650,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(19650,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(19650,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(19650,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(19650,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(19650,36,'cp 71','f8g7 d2d4 g8f6 d4c5 d6c5 d1e2 e8g8 e4e5 f6d5 f1d1 e7e6 h2h4 d8c7 d1e1 a7a6 c4d3 b7b5 h4h5 c5c4 d3c2 c8b7 h5h6 g7h8 a2a4 b5b4 b1d2 c6e5 f3e5 d5f4 e2c4 h8e5 c4c7 e5c7 d2e4 b4c3 b2c3 b7e4 c2e4 a8d8 c1e3 f4d5 e4d5 d8d5 g2g3 f8d8 c3c4 d5d3 g1g2');
INSERT INTO evaluations VALUES(5466,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(5466,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(5466,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(5466,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(5466,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(5466,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(5466,36,'cp 75','d2d4 g8f6 d4c5 d6c5 d1e2 e8g8 e4e5 f6d5 f1d1 e7e6 c4d3 d8c7 d1e1 a7a6 a2a4 c8d7 h2h4 h7h6 d3c2 b7b5 h4h5 b5b4 b1d2 b4c3 b2c3 d5c3 e2c4 c3d5 h5g6 f7g6 c2g6 d7e8 g6e4 c6b4 g2g3 g8h8 c1a3 a8c8 a1c1 e8h5 a3b4 c5b4');
INSERT INTO evaluations VALUES(5479,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(5479,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(5479,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(5479,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(5479,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(5479,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(5479,36,'cp 69','g8f6 d4c5 d6c5 d1d8 c6d8 f1e1 e8g8 a2a4 c8e6 b1a3 e6c4 a3c4 b7b6 g1f1 a8c8 c1f4 d8e6 f4g5 e6g5 f3g5 h7h6 g5f3 f8d8 a1d1 e7e6 e4e5 f6d5 g2g3 g6g5 f3d2 g8h7 h2h3 g7f8 d2e4 f8g7 f1e2');
INSERT INTO evaluations VALUES(32534,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(32534,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(32534,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(32534,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(32534,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(32534,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(32534,36,'cp 123','e2e4 b7b5 a2a4 b5b4 c3d5 e7e6 d5e3 c4c3 f1d3 g8f6 b2c3 b8c6 e4e5 f6d5 e3d5 e6d5 g1e2 b4c3 e1g1 c6b4 e2c3 a8b8 f2f4 g7g6 d3b1 f7f5 c1a3 d7e6 g1h1 c7c6 c3e2 a7a5 b1c2 f8e7 h2h3 e8g8 f1f3 d8d7 d1d2 f8c8 c2d1 d7a7 a1c1 a7a6 f3b3');
INSERT INTO evaluations VALUES(32536,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(32536,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(32536,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(32536,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(32536,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(32536,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(32536,36,'cp 81','b7b5 a2a4 b5b4 c3d5 e7e6 d5e3 d7c6 d4d5 e6d5 e4d5 c6b7 f1c4 f8c5 e3f5 d8f6 c1e3 c5e3 f5e3 g8e7 g1f3 e8g8 e1g1 b8d7 a4a5 a8d8 f1e1 e7g6 a5a6 b7a8 d1d4 c7c5 d4f6 d7f6 e1d1 g6f4 a1a5 f4d5 a5c5 d5b6 d1d8 f8d8 c4b3 a8f3 g2f3 g8f8 c5c7 d8d7 c7b7 g7g6 e3g4 d7b7');
INSERT INTO evaluations VALUES(30435,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(30435,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(30435,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(30435,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(30435,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(30435,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(30435,36,'cp 96','a2a4 b5b4 c3d5 e7e6 d5e3 d7c6 d4d5 e6d5 e4d5 c6b7 f1c4 f8c5 d1c2 b8d7 a4a5 g8f6 g1f3 e8g8 a5a6 b7c8 e1g1 f8e8 f1e1 c5d6 b2b3 d7b6 c1b2 b6c4 e3c4 e8e1 a1e1 d6f8 d5d6 c7d6 f3g5 g7g6 c2e2 c8f5 e2f3 a8b8 c4e3 h7h6 g5h3 d8e7 e1d1 f6e4 e3f5 g6f5 h3f4 f8g7 b2g7');
INSERT INTO evaluations VALUES(30439,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(30439,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(30439,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(30439,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(30439,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(30439,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(30439,36,'cp 108','b5b4 c3d5 e7e6 d5e3 d7c6 d4d5 e6d5 e4d5 c6b7 f1c4 f8c5 e3f5 d8f6 d1e2 g8e7 f5g3 b4b3 g1f3 e8g8 e1g1 b7d5 g3e4 d5e4 e2e4 b8c6 c1g5 f6g6 e4g6 h7g6 f1e1 c5d6 a1c1 a8b8 c4b5 f7f6 g5e3 c6b4 e1d1 a7a6 b5c4 g8h7 f3d4 f8e8 c4f7 e8d8 f7b3 b4c6 c1c3 c6d4 e3d4 g6g5 g1f1 h7g6 d4c5 d6c5 d1d8 b8d8 c3c5');
INSERT INTO evaluations VALUES(48847,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48847,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48847,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48847,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48847,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48847,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48847,36,'cp 0','e4d5 f6d5 d1h5 b8c6 f1b5 d8d6 g1f3 d5f4 h5e5 d6e5 f3e5 f4g2 e1f1 c8h3 e5c6 g2h4 f1e2 h3g2 c6e5 c7c6 b5c4 g2h1 e5f7 h8g8 f7g5 g8h8 g5f7');
INSERT INTO evaluations VALUES(48619,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48619,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48619,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48619,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48619,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48619,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48619,36,'cp 0','f6d5 d1h5 b8c6 f1b5 d8d6 g1f3 d5f4 h5e5 d6e5 f3e5 f4g2 e1f1 c8h3 e5c6 g2h4 f1e2 h3g2 c6e5 c7c6 b5c4 g2h1 e5f7 h1f3 e2f1 h8g8 f7e5 g8h8 e5f7');
INSERT INTO evaluations VALUES(49029,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49029,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49029,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49029,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49029,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49029,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49029,36,'cp 0','d1h5 b8c6 f1b5 d8d6 g1f3 d5f4 h5e5 d6e5 f3e5 f4g2 e1f1 c8h3 e5c6 g2h4 f1e2 h3g2 c6e5 c7c6 b5c4 g2h1 e5f7 h1f3 e2f1 h8g8 f7e5 g8h8 e5f7');
INSERT INTO evaluations VALUES(52552,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(52552,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(52552,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(52552,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(52552,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(52552,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(52552,36,'cp 203','e2e4 c8g4 h2h3 g4h5 f1d3 e7e6 d1e2 g8e7 c1d2 h5f7 h3h4 b8d7 e1g1 d7b6 a2a4 a7a5 f1e1 d8d7 d2f4 f7h5 f4g3 h5g4 a1d1 e8d8 e2d2 e7g6 f3g5 f8b4 f2f3 h7h6 g5f7 d7f7 f3g4 b6c4 d3c4 d5c4 h4h5 g6e7');
INSERT INTO evaluations VALUES(52554,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(52554,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(52554,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(52554,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(52554,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(52554,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(52554,36,'cp 200','c8g4 f1d3 e7e6 h2h3 g4h5 d1e2 f8b4 e4d5 c6d5 e2e6 g8e7 a2a3 h5f7 e6g4 b4c3 b2c3 e8g8 e1g1 d8c8 f1e1 b8c6 a3a4 f7g6 f3h4 g6d3 c2d3 f8e8 a1b1 c8g4 h3g4 g7g5 h4f3 b7b6 g1f1 a8c8 e1e3 c6a5 c1d2 c8c7 f3e1 e8d8 f1e2 d8c8 e1c2 c8d8');
INSERT INTO evaluations VALUES(55017,30,'cp -168',NULL);
INSERT INTO evaluations VALUES(55017,31,'cp -165',NULL);
INSERT INTO evaluations VALUES(55017,32,'cp -182',NULL);
INSERT INTO evaluations VALUES(55017,33,'cp -179',NULL);
INSERT INTO evaluations VALUES(55017,34,'cp -177',NULL);
INSERT INTO evaluations VALUES(55017,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(55017,36,'cp -178','e7e6 b1c3 g8f6 c1g5 h7h6 g5f6 d8f6 f1c4 f8b4 g1e2 e4e3 e1g1 e3f2 g1h1 e8g8 c3e4 f6e5 e2g3 f8d8 e4f2 e6d5 f2d3 e5d4 c4d5 d4d5 d3b4 d5d6 c2c3 d6d1 a1d1 d8e8 b4d5 b8a6 b2b4 c8e6 g3f5 g8h7 f5e7');
INSERT INTO evaluations VALUES(57656,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(57656,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(57656,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(57656,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(57656,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(57656,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(57656,36,'cp -18','f8g7 e2e3 c7c5 c2c3 g8f6 d4c5 f6e4 c1b2 e4c5 b3b4 c5e4 b1d2 e4d2 d1d2 e8g8 g1f3 d7d5 c3c4 d5c4 b2g7 d8d2 e1d2 g8g7 f1c4 a7a5 b4a5 f8d8 d2e2 a8a5 h1b1 a5c5 c4d3 b8d7 a2a4 d7f6 b1c1 c5d5 d3c4 d5a5 c4b5 c8d7 f3d4 f6e4 e2e1 d7b5 d4b5 d8d2');
INSERT INTO evaluations VALUES(39586,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(39586,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(39586,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(39586,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(39586,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(39586,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(39586,36,'cp -8','c1a3 d7d6 g1f3 c7c5 e2e3 c5d4 f3d4 d8a5 d1d2 a5d2 b1d2 b8c6 c2c3 g8f6 f1e2 c8d7 e1g1 d7g4 e2g4 f6g4 a3b2 e8g8 c3c4 f8c8 h2h3 g4f6 f1d1 a7a6 d4e2 b7b5 g2g4 f6d7 b2g7 g8g7 g1f1 b5b4 a2a3 b4a3 a1a3');
INSERT INTO evaluations VALUES(581,30,'cp 417',NULL);
INSERT INTO evaluations VALUES(581,31,'cp 426',NULL);
INSERT INTO evaluations VALUES(581,32,'cp 432',NULL);
INSERT INTO evaluations VALUES(581,33,'cp 430',NULL);
INSERT INTO evaluations VALUES(581,34,'cp 429',NULL);
INSERT INTO evaluations VALUES(581,35,'cp 451',NULL);
INSERT INTO evaluations VALUES(581,36,'cp 428','f7h8 c5f2 e1f2 e7c5 d2d4 c5c4 d4d5 c4d4 d1d4 c6d4 b1a3 f6e4 f2e1 c7c6 h1f1 c6d5 c1e3 d4e6 h8f7 d7d6 a1d1 e6c7 f7g5 e4f6 c2c4 b7b6 c4d5 c8a6 f1f3 e8e7 g5h3 h7h6 h3f2 c7d5 e3d2 a8c8 a3b1 c8c2 f3a3 a6b5 d1c1 c2b2 a3a7 e7e6 b1a3 g7g5 a3b5 b2b5 a7a8');
INSERT INTO evaluations VALUES(414,30,'cp 428',NULL);
INSERT INTO evaluations VALUES(414,31,'cp 434',NULL);
INSERT INTO evaluations VALUES(414,32,'cp 428',NULL);
INSERT INTO evaluations VALUES(414,33,'cp 419',NULL);
INSERT INTO evaluations VALUES(414,34,'cp 431',NULL);
INSERT INTO evaluations VALUES(414,35,'cp 434',NULL);
INSERT INTO evaluations VALUES(414,36,'cp 416','c5f2 e1f2 e7c5 d2d4 c5c4 d4d5 c4d4 d1d4 c6d4 b1a3 f6e4 f2e1 c7c6 h1f1 c6d5 c1e3 d4e6 h8f7 d7d6 a1d1 e6c7 f7g5 e4f6 c2c4 b7b6 c4d5 c8a6 f1f3 h7h6 g5e6 c7e6 d5e6 e8e7 e3d2 e7e6 a3b1 a8c8 f3a3 a6c4 b1c3 d6d5 a3a7 d5d4 a7g7 d4c3 d2c3');
INSERT INTO evaluations VALUES(2417,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(2417,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(2417,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(2417,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(2417,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(2417,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(2417,36,'cp 193','c7c6 d2d4 e6d6 c1g5 h7h6 g5h4 d6c7 c4d5 c6d5 c3b5 c7b6 a2a4 a7a5 f1e1 g7g5 h4g3 e7c6 g3e5 h8h7 b2b3 c6e5 e1e5 g5g4 f3d3 h7f7 c2c4 d8f6 c4c5 b6c6 b5d6 f8d6 d3b5 c6c7 c5d6 c7b8 e5e8 f6d6 e8c8 b8c8 b5e8 c8c7 e8f7 c7b6 f7f5 a8g8 g2g3 b6a7 a1c1 g8g5 f5h7 d6f6 h7c7');
INSERT INTO evaluations VALUES(2394,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(2394,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(2394,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(2394,33,'cp 216',NULL);
INSERT INTO evaluations VALUES(2394,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(2394,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(2394,36,'cp 232','d2d4 e6d7 c1g5 d7c7 c4d5 h7h6 g5h4 c6d5 c3b5 c7b6 a2a4 a7a5 f1e1 g7g5 h4g3 e7c6 g3e5 h8h7 b2b3 c6e5 e1e5 g5g4 f3d3 h7f7 c2c4 d8f6 c4c5 b6c6 b5d6 f8d6 d3b5 c6c7 c5d6 c7b8 e5e8 f6d6 e8c8 b8c8 b5e8 c8c7 e8f7 c7b6 f7f5 a8e8 f5g4 e8e4 g4g7 b6a7 g2g3');
INSERT INTO evaluations VALUES(2396,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(2396,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(2396,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(2396,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(2396,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(2396,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(2396,36,'cp 235','e6d7 c1g5 d7c7 c4d5 h7h6 g5h4 c6d5 c3b5 c7b6 a2a4 a7a5 f1e1 d8d7 h4e7 f8e7 e1e5 h8f8 f3b3 f8f6 e5d5 d7c6 d5e5 e7b4 b3d3 f6f8 d3e3 c6f6 a1e1 c8d7 e5e6 d7e6 e3e6 b4d6 e6d6 f6d6 b5d6 b6c7 d6e4 a8d8 e1d1 d8b8 c2c4 b7b6 f2f3 g7g5 d4d5 b8e8 d1e1 f8f4 g1f2');
INSERT INTO evaluations VALUES(12812,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12812,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(12812,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(12812,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12812,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(12812,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(12812,36,'cp 4','d7d5 c4d5 f6d5 d2d4 c5d4 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 g2c6 b7c6 c1e3 e7e5 b1d2 h7h5 f2f3 c8e6 h2h4 f7f5 a2a3 a8b8 b2b4 b8a8 d1e1 a7a5 e1f2 e8f7 b4a5 a8a5 h1c1 f8a3 c1c6 h8a8 c6c7 f7g6 d2c4 e6c4 c7c4 a3e7 a1c1 a8a6 c4c7 e7f6');
INSERT INTO evaluations VALUES(18453,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(18453,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(18453,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(18453,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(18453,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(18453,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(18453,36,'cp -62','g1f3 e7e6 e1g1 f8d6 d2d4 g8e7 c2c4 e8g8 b1c3 d5c4 c3e4 d6c7 d1e2 d8d5 e4c3 d5h5 f1e1 a7a5 c3e4 e7d5 e2c4 c8a6 c4a4 d5b4 e4c5 a6b5 a4d1 f8d8 h2h3 h5g6 c1d2 b4d3 c5d3 b5d3 f3e5 c7e5 e1e5 d3e4');
INSERT INTO evaluations VALUES(18451,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(18451,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(18451,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(18451,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(18451,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(18451,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(18451,36,'cp -61','e7e6 e1g1 f8d6 d2d4 g8e7 c2c4 e8g8 b1c3 d5c4 c3e4 d6c7 d1e2 d8d5 e4c3 d5h5 f1e1 a7a5 c3e4 c8a6 e4c5 a5a4 c5a6 a8a6 e2c4 h5b5 c4b5 c6b5 c1d2 c7d6 g1f1 f8c8 a1c1 c8c1 e1c1 f7f6 f3e1 g8f7 g2g3 g7g5 e1d3 a6c6');
INSERT INTO evaluations VALUES(19000,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(19000,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(19000,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(19000,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(19000,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19000,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(19000,36,'cp -20','d7d6 f1c4 a7a6 a2a4 g7g6 b1c3 f8g7 c3d5 g8f6 d5f6 d8f6 d2d3 e8g8 a4a5 c8e6 e1g1 f6e7 f3d2 f7f5 c2c3 a8d8 c4d5 g8h8 d2c4 f8f7 d5e6 e7e6 c4b6 f5f4 f2f3 d8f8 g1h1 h7h5 b2c1 g6g5');
INSERT INTO evaluations VALUES(26103,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(26103,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26103,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26103,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26103,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26103,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(26103,36,'cp 0','c6c5 e1g1 c5d4 c3b5 a7a6 b5d4 a8c8 c2c3 f8c5 h2h4 h7h6 d3e2 g8e7 h4h5 e8g8 c1f4 c5b6 a1d1 d7c5 f3d2 e7c6 g2g3 d8d7 g1g2 c5a4 d2f3 c6d4 f3d4 b6d4 d1d4 a4b6 e2g4 g8h7 f1d1 d7b5 d4b4 b5c6');
INSERT INTO evaluations VALUES(26143,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26143,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26143,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26143,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26143,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26143,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26143,36,'cp 0','e1g1 c5d4 c3b5 a7a6 b5d4 a8c8 h2h4 f8c5 c2c3 h7h6 h4h5 g8e7 c1f4 c5b6 d3e2 d7c5 f3d2 c5d7');
INSERT INTO evaluations VALUES(29717,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(29717,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(29717,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(29717,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(29717,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(29717,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(29717,36,'cp 99','f3e5 f5e6 h2h4 c7c6 c1f4 b8d7 h4h5 d7e5 f4e5 d8a5 f1d3 f6d7 e5f4 d7f6 d1e2 f6d5 f4d2 d5c3 b2c3 e6f5 c3c4 f5d3 e2d3 a5c7 a1b1 e7e6 e1g1 a8d8 h5h6 g7h6 c2c3 b7b6 f1e1 f8g7 d3h3 e8g8 d2h6 g7h6 h3h6 c6c5 b1b3 c5d4 c3d4 d8d4');
INSERT INTO evaluations VALUES(30799,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(30799,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(30799,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(30799,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(30799,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(30799,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(30799,36,'cp 0','g8f6 c1g5 e7e6 e2e4 h7h6 g5f6 d8f6 d1d2 f8b4 f1d3 g7g5 a2a3 b4c3 b2c3 d7d6 d2e2 b8d7 f3d2 h6h5 a3a4 a7a5 e1c1 e8c8 c1b2 c8b8 d3a6 b7a6 e2a6 f6f2 h1f1 f2g2 a6b5 f7f6 d2c4 b8a7 b5c6 a7b8');
INSERT INTO evaluations VALUES(28809,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(28809,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(28809,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28809,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(28809,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28809,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28809,36,'cp -14','c1g5 e7e6 e2e4 h7h6 g5f6 d8f6 d1d2 f8b4 f1d3 b8c6 a2a3 b4c3 d2c3 g7g5 a3a4 a7a5 d4d5 f6c3 b2c3 c6e7 c3c4 e7g6 e1d2 e8e7 a1b1 d7d6 h2h4 g5h4 f3h4 a8g8 h4g6 g8g6 g2g3 b7c8 e4e5 g6g4 e5d6 e7d6 d5e6 f7e6 d3e2 g4g8 b1a1 d6e7 h1h4 c8d7 d2e3 g8g5 e2d3 h6h5 e3d2');
INSERT INTO evaluations VALUES(37774,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(37774,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(37774,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(37774,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(37774,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(37774,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(37774,36,'cp -31','b8c6 d2d3 d7d6 f1e2 c8g4 f3g1 g4d7 g1f3 g8f6 h2h3 e8g8 c1e3 e7e5 a2a3 a7a5 e1g1 h7h6 e3d2 f6e8 c3d5 e8c7 a1b1 c7e6 b2b4 c5b4 a3b4 a5b4 d2b4 c6b4 b1b4 d7c6 d1c2 h6h5 d5b6 a8a3 c2b2');
INSERT INTO evaluations VALUES(43191,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(43191,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(43191,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43191,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(43191,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43191,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(43191,36,'cp 36','e2e4 c7c5 d4c5 d8a5 b1c3 a5c5 c1f4 b8c6 d1d2 c5a5 g1e2 e8g8 f4e3 a7a6 a2a3 d7d6 e2f4 c8d7 f1e2 b7b5 e1g1 b5c4 e2c4 f8c8 f4d3 c6e5 d3e5 a5e5 c4e2 d7e6 a1c1 e5a5 f3f4 a5c7 c3d5 c7b7 c1c8 a8c8 e2a6 b7a6 d5e7 g8h8 e7c8');
INSERT INTO evaluations VALUES(43196,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43196,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(43196,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(43196,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(43196,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43196,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(43196,36,'cp 20','c7c5 d4c5 d8a5 b1c3 a5c5 c1f4 b8c6 d1d2 c5a5 g1e2 e8g8 f4e3 a7a6 a2a3 d7d6 e2f4 c8d7 f1e2 b7b5 e1g1 b5c4 e2c4 f8c8 f1c1 c6e5 c4f1 e5c4 f1c4 c8c4 b2b3 c4c6 b3b4 a5e5 a3a4 g7h6 f4d5 f6d5 c3d5 h6e3 d2e3 a8c8');
INSERT INTO evaluations VALUES(51101,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(51101,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51101,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(51101,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(51101,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(51101,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(51101,36,'cp 36','b1c3 e7e5 g1f3 b8c6 e2e3 g8f6 d2d4 e5d4 e3d4 d7d5 c1g5 f8e7 g5f6 e7f6 c4d5 c6e7 f1c4 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 f6d4 a1d1 d4c3 f3c3 e7f5 f1e1 f5d6 d1c1 f8e8 e1e8 d6e8 g2g3 d8d6 g1g2 e8f6 a2a3 a8e8 c4a2 f6d5 c3b3 c7c6 b3b7');
INSERT INTO evaluations VALUES(51100,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(51100,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(51100,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(51100,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(51100,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(51100,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(51100,36,'cp 50','e7e5 g1f3 b8c6 g2g3 g8f6 f1g2 f8b4 c3d5 d7d6 d5b4 c6b4 a2a3 b4c6 d2d3 c8e6 e1g1 d8d7 b2b4 e6h3 e2e4 h3g2 g1g2 e8g8 c1b2 c6e7 f3h4 b7b5 c4b5 a6b5 f2f4 d7e6 f4f5 e6d7 g3g4 h7h6 a1c1 g8h8 h2h3 e7c6 g2h2 c6d4');
INSERT INTO evaluations VALUES(51055,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(51055,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(51055,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(51055,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(51055,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(51055,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(51055,36,'cp 59','g1f3 b8c6 g2g3 g8f6 f1g2 f8b4 c3d5 d7d6 d5b4 c6b4 a2a3 b4c6 d2d3 c8e6 b2b4 d8d7 e1g1 e6h3 e2e4 h3g2 g1g2 e8g8 c1b2 f8e8 f1e1 h7h6 a1c1 e8e6 f3h4 c6e7 d3d4 e5d4 d1d4 a8e8 f2f3 b7b5 c1d1 e7c6 d4c3 b5c4 c3c4 c6e5 c4c2');
INSERT INTO evaluations VALUES(51053,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(51053,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(51053,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(51053,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(51053,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(51053,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(51053,36,'cp 47','b8c6 g2g3 g8f6 f1g2 f8b4 c3d5 d7d6 d5b4 c6b4 a2a3 b4c6 d2d3 c8e6 b2b4 d8d7 e1g1 e6h3 e2e4 e8g8 c1b2 h3g2 g1g2 c6e7 d3d4 e5d4 d1d4 e7g6 d4e3 d7e6 f1e1 f8e8 b2f6 e6f6 a1d1 g6e5 f3e5 e8e5 f2f4 e5e7 c4c5 a6a5');
INSERT INTO evaluations VALUES(17900,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(17900,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(17900,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(17900,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(17900,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(17900,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(17900,36,'cp 42','g2g3 g8f6 f1g2 f8b4 c3d5 d7d6 d5b4 c6b4 a2a3 b4c6 e1g1 c8e6 d2d3 d8d7 b2b4 e6h3 e2e4 h3g2 g1g2 e8g8 c1b2 c6e7 d3d4 e5d4 d1d4 e7g6 d4e3 d7e6 f1e1 f8e8 b2f6 e6f6 a1d1 a6a5 b4b5 g6e5 f3e5 e8e5 f2f4 e5e7 d1d5 g7g6 a3a4 f6b2 d5d2 b2f6 e3d3 a8e8 h2h3');
INSERT INTO evaluations VALUES(51582,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(51582,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(51582,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51582,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(51582,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(51582,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51582,36,'cp 27','b4c5 b6c5 g1f3 e7e6 b1c3 g8e7 f1e2 e7c6 e1g1 f8e7 d2d4 c5d4 f3d4 e8g8 a1b1 d7d6 c1e3 h7h6 d4c6 b8c6 d1d2 d8a5 f1d1 c8a6 c3b5 a5d2 d1d2 a6b5 b1b5 a8c8 f2f4 c8c7 c2c4 a7a6 b5b6 c6a5 c4c5 d6c5 b6a6 a5c6 a6b6 c6d4 g1f2 d4e2 f2e2 f8c8');
INSERT INTO evaluations VALUES(51577,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(51577,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(51577,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51577,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(51577,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(51577,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(51577,36,'cp 24','b6c5 g1f3 e7e6 b1c3 d7d6 f1e2 d8a5 e1g1 a7a6 a1b1 f8e7 b1b3 b8d7 d2d4 c5d4 d1d4 g8f6 c1a3 e8g8 f3d2 d7e5 f1b1 f8d8 f2f4 e5c6 d4b6 a5b6 b3b6 c6d4 e2d3 d4b5 c3b5 a6b5 a3b4 a8a2 b6b5 c8a6 b5a5 a2a5 b4a5 d8a8 d3a6 a8a6 b1b8 e7f8 a5b4 f6d7');
INSERT INTO evaluations VALUES(51620,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(51620,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(51620,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(51620,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(51620,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(51620,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(51620,36,'cp 25','g1f3 e7e6 b1c3 b8c6 f1b5 c6d4 e1g1 a8b8 b5c4 d8a5 a2a4 a7a6 c1a3 a5c7 f1e1 f8e7 e4e5 f7f5 c3b1 c8b7 f3d4 c5d4 c4f1 e7a3 b1a3 g8e7 d1e2 e8g8 e2c4 c7c4 f1c4 h7h6 a1b1 e7g6 b1b3 b7e4 c4a6 g6e5 a6e2');
INSERT INTO evaluations VALUES(52186,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(52186,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(52186,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(52186,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(52186,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(52186,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(52186,36,'cp -11','g8f6 e2e4 f6e4 d1d5 e4f6 d5e5 f8e7 f1d3 e8g8 e1g1 b8c6 e5f4 d7d5 b1c3 c8g4 f4g3 g4e6 c3b5 f6e4 g3c7 c6b4 c1e3 a7a6 c7d8 a8d8 b5d4 d8c8 f1c1 b4d3 c2d3 e4c5 c1c3 e7d6 d4e6 c5e6 f3d4 c8c3 b2c3 f8c8 d4e2 d6c5 e3c5 e6c5');
INSERT INTO evaluations VALUES(53300,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(53300,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(53300,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(53300,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53300,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53300,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53300,36,'cp 13','g1f3 b8c6 f1b5 e7e6 c2c4 g8f6 e1g1 f8d6 c4d5 e6d5 c1b2 e8g8 d2d4 c5d4 f3d4 c8d7 b1d2 d8e7 b5c6 b7c6 d1c2 c6c5 d4f5 d7f5 c2f5 e7e6 f5d3 a8d8 b2f6 e6f6 f1d1 f6e6 a1c1 d6e7 h2h3 h7h6 d2f3 f8e8');
INSERT INTO evaluations VALUES(30,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(30,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(30,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(30,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(30,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(30,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(30,36,'cp -57','h2h3 g4d7 f3g5 d7e8 e2f1 h7h6 g5f3 g8e7 b2b3 c6c5 d2d3 f7f6 c1b2 e7g6 b1d2 h6h5 h3h4 e8d7 g2g3 b7b6 e1e2 g6e7 a1e1 e7f5 a2a3 d7c6 d2e4 f8e7 f3d2 h8e8 e4c3 e7f8 e2e8 c6e8 c3e2 a6a5 e2f4 a5a4 f4e6');
INSERT INTO evaluations VALUES(1853,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(1853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1853,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(1853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1853,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(1853,36,'cp -2','c6d4 c4b3 a7a5 b1c3 f8b4 a2a3 b4c3 b2c3 d4b3 c2b3 d7d6 g1e2 g7g6 h5g5 f6g5 c1g5 f7f5 f2f3 h7h6 g5e3 g6g5 c3c4 g8e7 e1d2 c8e6 b3b4 f5f4 e3f2 e7g6 g2g3 a5b4 a3b4 e8d7 g3f4 e5f4 b4b5 b7b6 e2d4 g6e5 a1a8 h8a8');
INSERT INTO evaluations VALUES(3766,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3766,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3766,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3766,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3766,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(3766,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3766,36,'cp 21','d8e8 f3e5 d6e5 d4e5 f6d7 f2f4 f7f5 e5f6 d7f6 e1g1 e6e5 d3e2 e5f4 e3f4 c8f5 b1d2 f5d3 f1e1 e8e3 g3f2 d3e2 f2e3 e2d1 a1d1 f8e8 g2g3 a8d8 e3f2 e8e1 d1e1 d5d4 f2d4 c6d4 c3d4 d8d4 e1e2 g8f8 g1g2 d4d5 d2f3 c7c6 f3e5 c6c5');
INSERT INTO evaluations VALUES(4278,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(4278,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(4278,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(4278,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(4278,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(4278,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(4278,36,'cp 128','c3d4 c5e7 d4d5 c6b8 b1c3 d7d6 f3d4 c8d7 b5d7 b8d7 d4f5 f8e8 f2f3 e7f8 g1h1 d7e5 c1g5 h7h6 g5h4 a7a6 a2a4 c7c6 f5e3 e5g6 h4f6 d8f6 f3f4 f6d8 g2g3 g6e7 d1d3 g7g6 a4a5 c6d5 c3d5 e7d5 e3d5 f8g7 f1c1');
INSERT INTO evaluations VALUES(4277,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(4277,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(4277,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(4277,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(4277,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(4277,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(4277,36,'cp 125','c5e7 d4d5 c6b8 b1c3 d7d6 f3d4 c8d7 b5d7 b8d7 d4f5 f8e8 f2f3 e7f8 g1h1 d7e5 c1g5 h7h6 g5h4 e5g6 h4f6 d8f6 f3f4 a7a6 g2g3 b7b5 d1d3 f6d8 f5d4 d8d7 a2a4 b5b4 c3d1 g6e7 a1c1 d7a4 c1c7 a4a5 c7d7 e7d5 d4c6 a5b5 d3b5 a6b5');
INSERT INTO evaluations VALUES(4289,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4289,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4289,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4289,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(4289,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(4289,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4289,36,'cp 0','g5h4 h7h6 f1e2 f8e8 c3d5 g7g5 d5f6 d8f6 h4g3 d7d6 c2c3 c6e7 h2h4 g5g4 f3h2 h6h5 f2f3 f6h6 f3g4 h5g4 d1d2 c5e3 d2d1');
INSERT INTO evaluations VALUES(5261,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5261,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5261,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(5261,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(5261,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(5261,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5261,36,'cp 0','c1e3 g8f6 b1c3 e8g8 e1g1 f6g4 d1g4 c6d4 b5d3 d7d5 g4h4 d5e4 d3e4 c8f5 c3d5 f5e4 d5e7 g8h8 h4e4 f8e8 e3d4 d8d4 e4d4 g7d4 e7d5 a8d8 d5e3 d4e3 f2e3 e8e3 f1f7 e3e2 a1f1 d8d2 f1f2 e2e1 f2f1');
INSERT INTO evaluations VALUES(5260,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5260,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5260,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5260,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5260,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5260,36,'cp 0','g8f6 b1c3 e8g8 e1g1 f6g4 d1g4 c6d4 b5d3 d7d5 g4h4 d5e4 d3e4 c8f5 c3d5 f5e4 d5e7 g8h8 h4e4 f8e8 e3d4 d8d4 e4d4 g7d4 e7d5 a8d8 d5e3 d4e3 f2e3 e8e3 f1f7 e3e2 a1f1 e2c2 f7b7 h8g8 f1e1 g8f8 e1f1');
INSERT INTO evaluations VALUES(7997,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7997,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7997,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7997,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7997,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7997,36,'cp 0','b1c3 e8g8 e1g1 f6g4 d1g4 c6d4 b5d3 d7d5 g4h4 d5e4 d3e4 c8f5 c3d5 f5e4 d5e7 g8h8 h4e4 f8e8 e3d4 d8d4 e4d4 g7d4 e7d5 a8d8 d5e3 d4e3 f2e3 e8e3 f1f7 e3e2 a1f1 e2c2 f7b7 h8g8 f1e1 g8f8 e1f1');
INSERT INTO evaluations VALUES(7996,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7996,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7996,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7996,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7996,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7996,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7996,36,'cp 0','e8g8 e1g1 f6g4 d1g4 c6d4 b5d3 d7d5 g4h4 d5e4 d3e4 c8f5 c3d5 f5e4 d5e7 g8h8 h4e4 f8e8 e3d4 d8d4 e4d4 g7d4 e7d5 a8d8 d5e3 d4e3 f2e3 e8e3 f1f7 e3e2 a1f1 e2c2 f7b7 h8g8 f1e1 g8f8 e1f1');
INSERT INTO evaluations VALUES(3041,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3041,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3041,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3041,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3041,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3041,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3041,36,'cp -12','e1g1 f6g4 d1g4 c6d4 b5d3 d7d5 g4h4 d5e4 d3e4 c8f5 c3d5 f8e8 e3d4 g7d4 c2c3 d4g7 e4f5 d8d5 f5c2 g7f6 h4g4 e7e6 f1d1 d5a5 c2b3 a5c7 g2g3 a8b8 d1d3 b8d8 d3d8 e8d8 a1d1 g8g7 d1d8 c7d8 g4b4 d8d2 b4b7');
INSERT INTO evaluations VALUES(3040,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3040,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3040,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3040,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3040,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3040,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3040,36,'cp -14','f6g4 d1g4 c6d4 b5d3 d7d5 g4h4 d5e4 d3e4 c8f5 c3d5 f8e8 e3d4 g7d4 a1d1 f5e4 h4e4 d4b2 d5f6 b2f6 d1d8 e8d8 e4b7 a8c8 h2h4 c8c2 h4h5 c2a2 h5h6 a2d2 b7a7 f6g5 f1b1 e7e6 g2g3 d2d7 b1b7 d7b7 a7b7 g5h6 b7e7 d8d2 e7e8 h6f8 e8b8 d2d5 b8a8 d5f5 a8b8 h7h5');
INSERT INTO evaluations VALUES(10126,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(10126,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(10126,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(10126,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(10126,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(10126,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(10126,36,'cp -47','f6e4 d2d3 e4f6 f1e1 e8g8 f3e5 c6e5 e1e5 d7d5 c4b3 c5d6 e5e1 c7c6 c1e3 c8f5 d1f3 f5g6 e3f4 a7a5 a2a4 d8c7 f4d6 c7d6 b1d2 f8e8 e1e3 g6h5 g2g4 e8e3 f3e3 a8e8 e3d4 h5g6 d2f3 c6c5 d4c3 h7h5 g4g5 d6f4 g5f6 f4f3 a1e1 e8e1 c3e1 f3f6');
INSERT INTO evaluations VALUES(10628,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(10628,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(10628,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(10628,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(10628,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(10628,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(10628,36,'cp -42','d2d3 e4f6 f1e1 e8g8 f3e5 c6e5 e1e5 d7d5 c4b3 c7c6 e5e1 a7a5 a2a4 c5d6 b1d2 c8f5 d2f3 d8d7 f3e5 d7c7 e5f3 f8e8 c1d2 f6d7 e1e8 a8e8 d3d4 h7h6 c2c4 d7f6 a1c1 c7d8 c4d5 f6d5 b3c4 d8b6 d1b3 b6b3 c4b3');
INSERT INTO evaluations VALUES(21871,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21871,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21871,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21871,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(21871,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21871,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21871,36,'cp 0','g8f6 d2d4 f6e4 f1b5 e4f6 c2c4 c8d7 e1g1 f8e7 f1e1 f6d5 c4d5 c6d4 f3d4 e5d4 d1g4 d7b5 g4g7 e8d7 g7g4 d7e8 g4g7');
INSERT INTO evaluations VALUES(15452,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15452,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15452,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15452,36,'cp -4','d2d4 f6e4 f1b5 e4f6 c1g5 f8e7 b5c6 b7c6 d5e7 d8e7 d4e5 d6e5 e1g1 h7h6 g5h4 e8g8 f1e1 e5e4 d1d4 f8e8 h4f6 g7f6 f3d2 c8f5 f2f3 a8d8 d4c3 e7e5 c3e5 e8e5 d2e4 f5e4 e1e4 e5e4 f3e4 d8d2 a1f1 d2c2 f1f6 c2b2 f6h6 b2c2 h6f6 c2a2 f6c6 a7a5 c6c7');
INSERT INTO evaluations VALUES(24458,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24458,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(24458,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24458,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24458,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(24458,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(24458,36,'cp 10','d4c6 d7c6 b5c6 b7c6 d1f3 a8c8 b2b3 f8g7 c1b2 e8g8 a1d1 f6d7 c3a4 g7b2 a4b2 d8a5 b2c4 a5c5 f1e1 d7e5 c4e5 c5e5 h2h3 a7a5 a2a4 c8b8 f3e3 c6c5 d1d5 e5b2 e1e2 c5c4 b3c4 b2a1 g1h2 a1a4 e4e5 a4c4 d5a5 b8b7 f2f4 c4c6 e5d6');
INSERT INTO evaluations VALUES(25933,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(25933,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(25933,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(25933,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(25933,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(25933,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(25933,36,'cp -95','c6e5 b5d7 e5d7 c2c4 g8f6 c4d5 f6d5 b1c3 d7b6 a2a4 d5c3 b2c3 b6d5 f1e1 f8e7 c3c4 d5b4 d4d5 e6d5 d1e2 e8g8 e2e7 b4c2 a1b1 c2e1 e7e1 d5c4 c1b2 d8d3 f3e5 d3c2 e1c1 c2a4 e5c4 f7f6 c4d6 b7b6 c1d2 a4b3 d2e1 b3c2 b2c3 a8d8 b1c1 c2a2 c1a1 a2b3 a1b1');
INSERT INTO evaluations VALUES(25992,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(25992,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(25992,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(25992,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(25992,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(25992,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(25992,36,'cp -99','b5d7 e5d7 c2c4 g8f6 c4d5 f6d5 b1c3 d5c3 b2c3 c5d4 d1d4 d7c5 d4b4 d8b6 b4g4 h7h5 g4g3 c5e4 g3e5 e4f6 c1g5 f8d6 e5e2 b6a6 e2a6 b7a6 g5e3 a8c8 f1c1 f6e4 c3c4 c8c7 c4c5 d6c5 c1c4 f7f5 e3c5 c7c5 c4c5');
INSERT INTO evaluations VALUES(30905,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(30905,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(30905,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(30905,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(30905,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(30905,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(30905,36,'cp 152','b5d7 d8d7 e1g1 b8c6 f1e1 f8e7 d4c5 g8f6 b1c3 d5d4 c3e2 e8g8 e2d4 a8d8 c1e3 e7c5 c2c3 f8e8 d1b3 c6d4 e3d4 c5d4 f3d4 g7g6 e1e8 d8e8 a1d1 e8d8 h2h3 a7a6 d1e1 b7b5 a2a3 d8e8 e1d1 f6h5 d4f3 d7c6 b3b4 c6c7 a3a4 b5a4 b4a4 e8e2 d1d2 h5f4 d2e2');
INSERT INTO evaluations VALUES(30859,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(30859,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(30859,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(30859,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(30859,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(30859,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(30859,36,'cp 150','d8d7 e1g1 b8c6 f1e1 f8e7 d4c5 g8f6 b1c3 d5d4 c3b5 e8g8 b5d4 a8d8 c2c3 e7c5 c1e3 f8e8 d4c6 d7c6 e3d4 a7a6 e1e8 d8e8 h2h3 g7g6 d4c5 c6c5 d1d4 c5f5 d4b6 e8e7 a1d1 g8g7 b6b4 e7e2 d1d8 f6d7 b4d6 f5c5 d6d4 c5d4 f3d4 e2e1 g1h2 d7c5 b2b4');
INSERT INTO evaluations VALUES(32614,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(32614,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(32614,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(32614,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(32614,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(32614,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(32614,36,'cp 40','b8c6 e1g1 e8c8 c1e3 d5f5 c2c4 e7e5 d4d5 e5e4 f3d4 c6d4 e3d4 h7h5 b1c3 c7c5 d4e3 f5e5 g2g3 g4e2 d1e2 h5h4 e3f4 e5f5 f2f3 e4f3 e2f3 h4g3 f4g3 f5f3 f1f3 f8d6 a1f1 d6b8 b2b3 b7b6 f3e3 b8g3 e3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(27402,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(27402,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(27402,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(27402,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(27402,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(27402,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(27402,36,'cp 31','e1g1 e8c8 c2c3 e7e5 c3c4 d5d7 d4d5 g4f3 e2f3 c6d4 b1c3 f8b4 c1d2 d7f5 f3e2 h8e8 a1c1 d4e2 d1e2 f5h5 e2d3 b4c5 h2h3 h5g6 d3g6 h7g6 d2e3 f6d7 c3e4 c5e7 e4g5 e7g5 e3g5 f7f6 g5e3 a7a5 f1d1 g6g5 b2b3 b7b6');
INSERT INTO evaluations VALUES(27401,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(27401,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(27401,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(27401,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(27401,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(27401,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(27401,36,'cp 53','e8c8 c2c3 e7e5 c3c4 d5d7 d4d5 g4f3 e2f3 c6d4 b1c3 f8b4 c1d2 d7f5 f3e2 h8e8 a2a3 d4e2 d1e2 b4c3 d2c3 f5e4 a1e1 e4e2 e1e2 b7b5 b2b3 f6d7 f1c1 b5c4 b3c4 d7b6 c3a5 f7f6 e2d2 c7c6 d2b2 c6d5 c4d5 c8b7 a5b6 a7b6 c1c6 d8d5 c6b6 b7c8 h2h3 c8d7 b6b7 d7e6 b7g7');
INSERT INTO evaluations VALUES(44157,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(44157,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(44157,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44157,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(44157,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(44157,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(44157,36,'cp -22','d7d5 c4e2 f6d7 c2c3 d5d4 e1g1 f8e7 f1e1 e8g8 e2d3 b8c6 d3b5 d8c7 b5c6 c7c6 c3d4 b7b6 d2d3 f8d8 b1c3 c8b7 d4c5 b6c5 c3e4 h7h6 c1f4 c6a6 d1e2 d7f8 e4d6 f8g6 f4g3 b7d5 d6c4 a6c6 b2b3 c6b7 c4d6');
INSERT INTO evaluations VALUES(46284,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46284,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(46284,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(46284,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(46284,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(46284,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(46284,36,'cp -20','c4e2 f6d7 c2c3 d5d4 e1g1 f8e7 f1e1 e8g8 h2h3 a7a6 d2d3 d4c3 b1c3 b8c6 c1f4 b7b5 a2a4 c8b7 a4b5 a6b5 c3b5 d8b6 b5d6 a8a1 d1a1 d7e5 f3e5 c6e5 d6b7 e5g6 f4e3 b6b7 e2f3 b7b5 a1a7 b5b2 f3e4 f7f5 e4f3 b2b5 e1a1 b5d3 e3c5 e7c5 a7c5');
INSERT INTO evaluations VALUES(45948,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(45948,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(45948,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(45948,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(45948,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(45948,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(45948,36,'cp 61','d4f3 g7g6 e1g1 b8c6 c2c4 f8g7 b1c3 e8g8 c1g5 h7h6 g5e3 f6g4 e3c1 g4f6 a1b1 f6d7 d3c2 d7c5 b2b4 c5e6 c3d5 b7b6 a2a4 c8b7 c1d2 e6d4 c2d3 a8c8 b1c1 f7f5 f3d4 c6d4 e4f5 g6f5 d2e3');
INSERT INTO evaluations VALUES(56414,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(56414,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(56414,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(56414,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(56414,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(56414,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(56414,36,'cp -116','b1c3 b8c6 f1g2 d5d4 c3e4 h7h5 g4g5 g8e7 d2d3 a7a5 a2a3 e7g6 h3h4 c8g4 c2c3 a5a4 b2b4 a4b3 d1b3 a8a7 g2f3 g4e6 c3c4 d8d7 b3b5 f8b4 e1f1 e6g4 b5d5 d7d5 c4d5 c6e7 a1b1 b4d6 e4d6 c7d6 f3e4 f7f5 e4g2 e8g8 f1e1');
INSERT INTO evaluations VALUES(57141,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(57141,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(57141,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(57141,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(57141,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(57141,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(57141,36,'cp -22','e5d4 g1f3 d7d5 f3d4 c7c5 d4b3 b8c6 f1g2 c5c4 b3d4 f8c5 c2c3 g8f6 c1e3 c5b6 b2b3 c4b3 a2b3 e8g8 e1g1 c8g4 b1a3 f6e4 a3b5 f8e8 d4c2 h7h5 g2e4 e8e4 d1d2 h5h4 f1d1 h4h3 d2d5 g4e2 e3b6 d8b6 d5e4 e2d1');
INSERT INTO evaluations VALUES(57307,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(57307,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(57307,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(57307,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(57307,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(57307,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(57307,36,'cp -9','g1f3 d7d5 f1g2 c7c5 c2c3 d4c3 b1c3 d5d4 d1a4 c8d7 a4b3 b8c6 f3g5 g8h6 c3d5 f8e7 d5e7 d8e7 g5e4 h6f5 b3b7 a8b8 b7a6 b8b6 a6a4 e8g8 e1g1 d7e6 b2b3 f8c8 c1g5 f7f6 g5d2 c6e5 a1c1 g8h8 a4a3 c5c4 a3e7 f5e7 b3c4 e5c4 d2f4 e7g6 f1d1 g6f4 g3f4');
INSERT INTO evaluations VALUES(822,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(822,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(822,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(822,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(822,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(822,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(822,36,'cp 40','f1d3 c5d4 c3d4 c8d7 d3c2 h6f5 c2f5 e6f5 b1c3 d7e6 b2b4 f8e7 d1d3 e8g8 h2h4 h7h6 e1g1 f8c8 c1f4 c6d8 c3a4 b6c6 a4c5 b7b6 c5e6 f7e6 f1c1 c6d7 b4b5 c8c1 a1c1 a8c8 a3a4 d8f7 f4e3 g7g5 c1c8 d7c8 h4g5 h6g5 f3e1 g5g4');
INSERT INTO evaluations VALUES(7026,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(7026,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(7026,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(7026,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(7026,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(7026,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(7026,36,'cp 93','d4d5 c6b8 e2e4 d7d6 f1d3 g8f6 h2h3 e8g8 g1e2 e7e6 e1g1 e6d5 c4d5 c7c6 c1g5 b8d7 d3c2 h7h6 g5h4 c6d5 c3d5 g6g5 h4g3 f6d5 d1d5 d7e5 e2d4 d8b6 a1d1 a7a5 b2b3 b6c5 f2f4 g5f4 g3f4 a5a4 b3a4 c5d5 e4d5 c8d7 g1h1 f8e8 d4b5');
INSERT INTO evaluations VALUES(10834,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(10834,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(10834,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(10834,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(10834,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(10834,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(10834,36,'cp 34','e1g1 d7d6 a2a4 e8g8 f1e1 a7a6 b1d2 c6a5 c4a2 c7c5 d2c4 a5c4 a2c4 b7b6 d3d4 e5d4 c3d4 d6d5 e4d5 c8b7 f3h4 f8e8 h4f5 e7f8 e1e8 d8e8 d4c5 a8d8 f5d6 f8d6 c5d6 e8e5 d6d7 d8d7 d1e2 e5e2 c4e2 f6d5 c1d2 d7c7 e2d3 c7d7 d2e1 a6a5 d3f1 d5b4 a1c1 b4d3 c1d1');
INSERT INTO evaluations VALUES(10833,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(10833,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(10833,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(10833,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(10833,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(10833,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(10833,36,'cp 25','e8g8 a2a4 a7a6 b1d2 e7c5 b2b4 c5a7 h2h3 d7d6 f1e1 c8e6 c4b3 f6h5 d2f1 d8d7 f1e3 e6b3 d1b3 h5f4 b3c2 c6e7 d3d4 e5d4 f3d4 f4g6 e3f5 a7d4 f5d4 c7c5 b4c5 d6c5 d4f3 a8d8 c1a3 d7c7 a1d1 d8d1');
INSERT INTO evaluations VALUES(4554,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(4554,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(4554,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(4554,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(4554,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(4554,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(4554,36,'cp 26','f1e1 d7d6 a2a4 a7a6 h2h3 c8e6 c4e6 f7e6 b1d2 d8e8 d1b3 f6h5 d2f1 e8g6 f1h2 c6a5 b3d1 c7c5 b2b4 c5b4 c3b4 a5c6 c1d2 h5f4 d2f4 f8f4 b4b5 c6d4 f3d4 e5d4 e1e2 f4f7 d1b3 a8f8 b5a6 b7a6');
INSERT INTO evaluations VALUES(4555,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(4555,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4555,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(4555,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(4555,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(4555,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(4555,36,'cp 36','d7d6 a2a4 c8e6 c4e6 f7e6 b1a3 a7a6 a3c4 d8e8 c1d2 f6d7 b2b4 e8f7 b4b5 c6d8 d1e2 f7h5 h2h3 a6b5 a4b5 d8f7 c4a5 f8b8 d2e3 f7g5 e3g5 e7g5 a5c4 b8f8 a1a8 f8a8 d3d4 e5d4 c3d4 h5g6 e2d3 a8f8 f3g5 g6g5 c4e3 f8f7');
INSERT INTO evaluations VALUES(3307,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(3307,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(3307,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(3307,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(3307,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(3307,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(3307,36,'cp 38','a2a4 c8e6 c4e6 f7e6 b2b4 d8e8 c1e3 e8f7 b1d2 f6g4 a4a5 a7a6 d1b3 e7f6 h2h3 g4e3 f2e3 c6e7 e1f1 e7g6 d3d4 f7e7 b3c4 a8e8 d4d5 g8h7 d5e6 c7c6 c4b3 e7e6 b3e6 e8e6 a1d1 f6e7 f1f2');
INSERT INTO evaluations VALUES(12836,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(12836,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12836,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(12836,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(12836,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(12836,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(12836,36,'cp 0','f2f4 e7e6 g1f3 f8e7 f1e2 e8g8 a2a4 a8b8 e1g1 a7a6 b1a3 b7b5 a4b5 a6b5 a3c2 b5b4 c2e3 d7d6 g1h1 b4c3 b2c3 c6a5 f3d2 a5c6');
INSERT INTO evaluations VALUES(16676,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(16676,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(16676,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(16676,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(16676,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(16676,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16676,36,'cp -39','d7d5 b1c3 d5e4 d4c6 d8d1 e1d1 b7c6 c3e4 f6e4 f3e4 f8d6 f1d3 f7f6 c1e3 a7a5 g2g3 c8g4 d1d2 e8c8 a1f1 h7h5 e3f4 d6b4 c2c3 b4d6 d2c2 a5a4 h2h4 c8b7 f1b1 h8e8 b2b4 g4f3 h1e1 g7g5 h4g5 f6g5 f4g5 d6g3 g5d8 g3e1 b1e1');
INSERT INTO evaluations VALUES(19439,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19439,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19439,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(19439,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(19439,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(19439,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(19439,36,'cp 28','g8f6 d2d4 c5d4 e2d4 e7e5 d4b5 d7d6 c1g5 a7a6 b5a3 b7b5 c3d5 f8e7 g5f6 e7f6 c2c3 a8b8 a3c2 f6g5 a2a4 b5a4 c2b4 c8d7 f1a6 c6e7 d5e7 g5e7 a6c4 e8g8 d1e2 h7h6 e1g1 g7g6 f1d1 g8g7');
INSERT INTO evaluations VALUES(12826,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(12826,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(12826,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(12826,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(12826,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(12826,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(12826,36,'cp 18','d2d4 c5d4 e2d4 e7e5 d4b5 d7d6 c3d5 f6d5 e4d5 c6b8 a2a4 f8e7 f1e2 e8g8 e1g1 b8d7 c1d2 a7a6 b5a3 a6a5 a3c4 b7b6 f2f4 e5f4 d2f4 d7c5 c2c3 c8a6 b2b4 a6c4 e2c4 c5e4 d1c2 f7f5 c4d3 e7f6 d3e4 f5e4 b4a5 a8a5 c2e4 d8d7 f1f3');
INSERT INTO evaluations VALUES(27065,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(27065,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(27065,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(27065,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(27065,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(27065,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(27065,36,'cp 53','c4e6 f7e6 c2c3 g8f6 f1e1 f8e7 d2d4 d8d7 d4e5 c6e5 f3e5 d6e5 d1d7 f6d7 c1e3 a7a5 a2a4 b7b6 b1d2 g7g5 e1b1 d7c5 d2c4 c5b3 a1a2 g5g4 c4e5 h8g8 b1e1 e7c5 e3c5 b3c5 b2b4 a5b4 c3b4 a8a4 a2a4 c5a4 e1c1 c7c5 b4c5 a4c5');
INSERT INTO evaluations VALUES(27049,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(27049,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(27049,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(27049,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(27049,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(27049,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(27049,36,'cp 50','f7e6 c2c3 g8f6 d1b3 d8c8 f1e1 f8e7 d2d4 e8g8 d4e5 c6e5 f3e5 d6e5 h2h3 c8d7 b1d2 a7a5 a2a4 e7c5 d2f3 d7c6 b3c2 c5d6 c1g5 h7h6 g5h4 f6d7 h4g3 b7b5 c2b3 b5a4 b3e6 g8h7 e6d5 c6d5 e4d5 a8b8 a1a4 b8b2 a4a5 b2c2 a5a7 c2c3 f3e5 d6e5 g3e5');
INSERT INTO evaluations VALUES(34523,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(34523,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(34523,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(34523,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(34523,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(34523,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(34523,36,'cp 199','e2e4 a7a6 g1f3 g8f6 d1c2 b7b5 d2d4 b5b4 f3e5 c6b7 c3d5 f6d5 e4d5 f7f6 e5d3 c8f5 c2b3 e7e6 d3f4 e6d5 f4d5 b8c6 b3f3 f5e6 f3e4 e8d7 d5e3 e6f7 e4g4 d7d8 f1e2 b4b3 a2a3 h7h5 g4e4 a8b8 e1g1 f8d6 f1d1 c6e7 e4b7 b8b7 e2a6 b7b6 a6c4 f7c4 e3c4');
INSERT INTO evaluations VALUES(34519,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(34519,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(34519,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(34519,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(34519,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(34519,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(34519,36,'cp 211','a7a6 g1f3 g8f6 d1c2 b7b5 d2d4 b5b4 f3e5 c6b7 c3d5 f6d5 e4d5 b4b3 c2b3 f7f6 e5c4 b7d5 c1f4 e7e6 f1e2 d5b7 e1g1 b7b3 a2b3 b8c6 f1d1 a8a7 f4e3 c6b4 d4d5 a7a8 d5d6 f8d6 c4d6 c7d6 d1d6 e8g8');
INSERT INTO evaluations VALUES(37196,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(37196,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(37196,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(37196,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(37196,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37196,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(37196,36,'cp -11','c7c5 d1c2 b8c6 d4c5 d8a5 e2e3 a5c5 a3a4 f8d8 c1a3 c5h5 f1e2 b7b6 e1g1 c8b7 f1d1 a8c8 h2h3 h5g6 c2g6 h7g6 f3d4 c6d4 c3d4 b7a6 d4d5 d8e8 d5e6 e8e6 a3b2 a6c4 e2c4 c8c4 b2f6 g7f6 d1d7');
INSERT INTO evaluations VALUES(35680,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(35680,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(35680,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(35680,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(35680,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(35680,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(35680,36,'cp -4','d1c2 b8c6 d4c5 d8a5 e2e3 a5c5 a3a4 f8d8 c1a3 c5h5 f1e2 b7b6 e1g1 c8b7 f1d1 a8c8 a4a5 c6a5 a3e7 b7e4 c2a2 d8e8 e7f6 g7f6 d1d7 e4f3 e2f3 h5c5 d7a7 a5c4 f3e2 e8d8 h2h3 c4e5 c3c4 e5c4 e2c4 c5c4 a2b2 c4c3 b2b6 c8b8 b6c7 c3c7 a7c7');
INSERT INTO evaluations VALUES(37979,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(37979,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37979,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37979,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37979,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37979,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37979,36,'cp 0','f1b5 c8d7 e1g1 d7b5 c3b5 b8c6 d2d3 d8d7 c2c4 g8h6 h2h3 f7f5 a1b1 e8g8 c1e3 e7e6 a2a3 a8b8 b5c3 c6d4 d1d2 b7b6 b2b4 h6f7 g1h1 b8e8 b1e1 f7d8 e3d4 c5d4 c3e2 f5e4 f3d4 e6e5 d4c2 e4d3 d2d3 d8f7');
INSERT INTO evaluations VALUES(37951,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37951,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37951,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37951,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37951,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37951,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37951,36,'cp 0','c8d7 e1g1 d7b5 c3b5 b8c6 d2d3 d8d7 c2c4 g8h6 h2h3 f7f5 a1b1 e8g8 c1e3 e7e6 a2a4 b7b6 d1d2 a8d8 b1d1 d8e8 g1h1 a7a6 b5c3 c6d4 d1b1 d4b3 d2c2 b3d4');
INSERT INTO evaluations VALUES(28066,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28066,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28066,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28066,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(28066,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28066,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28066,36,'cp 0','e1g1 d7b5 c3b5 b8c6 d2d3 d8d7 c2c4 g8h6 h2h3 f7f5 a1b1 e8g8 c1e3 e7e6 b5c3 a8e8 d1d2 b7b6 b1e1 c6d4 d2f2 d4c6');
INSERT INTO evaluations VALUES(38388,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(38388,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(38388,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(38388,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(38388,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38388,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(38388,36,'cp 30','e2e4 a7a6 g1f3 c8g4 f1e2 b8c6 f4e3 e7e5 d4e5 d6e5 d1d8 a8d8 c3d5 d8d7 e1c1 g8f6 d5f6 g7f6 d1d7 g4d7 c2c3 f6e7 e2c4 f7f6 c4d5 c6d8 b2b4 b7b5 a2a3 d7e6 c1c2 e8f7 d5e6 d8e6 a3a4 c7c6 h1a1 h8d8 a4b5 a6b5');
INSERT INTO evaluations VALUES(38425,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(38425,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(38425,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(38425,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(38425,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38425,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(38425,36,'cp 39','b8c6 f1b5 c8d7 g1f3 a7a6 b5c6 d7c6 d4d5 c6d7 d1d2 g8f6 f4h6 g7h6 d2h6 e7e5 d5e6 d7e6 e1c1 f6g4 h6d2 d8f6 h2h3 g4e5 f3e5 f6e5 d2e3 e8c8 e3a7 c7c6 f2f3 e5c5 a7c5 d6c5 d1d8 h8d8 c3a4 c5c4 h1d1 d8e8 d1d4 e8g8 h3h4 g6g5 h4g5 g8g5 g2g4');
INSERT INTO evaluations VALUES(38551,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(38551,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(38551,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(38551,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(38551,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(38551,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(38551,36,'cp 75','d6d5 e4d5 e6d5 f1e1 g8e7 c4d3 e8g8 h2h3 f8e8 c1f4 b8d7 a2a4 d7f8 c2c3 f8e6 f4e5 g7h6 d3f1 a7a5 c3c4 c7c6 c4d5 e7d5 b1c3 e6g5 f3g5 h6g5 c3e4 g5e7 f1c4 c8e6 d1b3 d8b6 b3b6 d5b6 c4e6 f7e6 e4c5 e7c5 d4c5 b6d5 a1d1 a8d8');
INSERT INTO evaluations VALUES(40742,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40742,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(40742,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(40742,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40742,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(40742,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40742,36,'cp 31','h2h3 c7c6 g2g4 e8g8 g1e2 b7b5 a2a3 c6d5 e4d5 a7a6 f1g2 b8d7 e1g1 a8b8 c1e3 h7h6 e2g3 d7b6 b2b3 c8d7 d1d3 b8c8 c3e2 f8e8 f2f4 e7f8 a1d1 a6a5 g4g5 e5f4 e2f4 e8e3 d3e3 h6g5 f4h5 f6h5 g3h5 c8c2');
INSERT INTO evaluations VALUES(41988,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(41988,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(41988,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(41988,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(41988,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(41988,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(41988,36,'cp -13','e2e3 c7c5 f1e2 b4c3 b2c3 d8a5 e1g1 a5c3 g5f6 g7f6 d1c1 b8c6 d4c5 c8d7 a1b1 c6a5 f3d2 e8e7 d2b3 a5b3 c2b3 c3c1 f1c1 h8c8 g2g3 b7b6 c5b6 a7b6 a2a4 e7d6 g1f1 c8c5 f1e1 c5a5 h2h4 f6f5 c1c2 b6b5 a4b5 d7b5 b3b4 a5a1 b1a1 a8a1 e1d2 b5e2');
INSERT INTO evaluations VALUES(45863,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(45863,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(45863,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45863,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(45863,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(45863,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45863,36,'cp 21','e1g1 f8e7 a2a4 c8g4 h2h3 g4h5 g2g4 h5g6 b1c3 e8g8 f3h4 d6d5 c4a2 b8a6 h4g6 h7g6 e4d5 c6d5 g4g5 f6h5 c3d5 e7g5 c1g5 d8g5 d1g4 g5d8 a1e1 h5f6 d5f6 d8f6 a4a5 a8d8 g4e4 d8e8 e1e3 f6d6 e4b7 a6b4 a2b3 b4c6 b3c4 e8e7 b7a6 d6d7 f1e1');
INSERT INTO evaluations VALUES(47454,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(47454,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(47454,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(47454,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(47454,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(47454,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(47454,36,'cp -14','d5f6 c3e4 b8c6 f1b5 c8d7 g1e2 e7e6 d2d4 f6e4 f3e4 a7a6 b5a4 f8d6 c1f4 e8g8 e1g1 c6e7 a4d7 d8d7 c2c3 h7h6 f4d6 c7d6 f1e1 a6a5 e2f4 d7b5 e4c2 f8c8 c2d2 b5c6 a2a3 b7b5 h2h3 c8c7 a1d1 a8c8 h3h4 c6c4');
INSERT INTO evaluations VALUES(47417,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(47417,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(47417,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(47417,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(47417,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(47417,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(47417,36,'cp -13','c3e4 b8c6 f1b5 c8d7 g1e2 e7e5 d2d4 f6e4 f3e4 d8e7 b5c6 d7c6 d4d5 c6d7 c1d2 e8c8 e1c1 e7f6 f2f3 f6a6 e2c3 f8d6 a2a3 f7f5 e4e1 h7h6 d2e3 h8e8 e1d2 c8b8 h2h3 a6c4 h1e1 b7b6 c1b1 c4h4 d2e2 h4g3');
INSERT INTO evaluations VALUES(48412,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(48412,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(48412,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(48412,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48412,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48412,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(48412,36,'cp -6','c7c5 g1f3 f8e7 e1g1 e8g8 d2d3 b8c6 f3e5 c6e5 b2e5 c8d7 e2e4 d7c6 b1d2 f6d7 e5b2 e7f6 b2f6 d8f6 a1c1 f6c3 e4d5 c6d5 g2d5 e6d5 d2f3 f8e8 f1e1 h7h6 d1d2 c3f6 e1e8 a8e8 c1e1 e8e1 f3e1 f6a1 a2a4 b7b6 d2e2 d7e5 h2h3 a1c3');
INSERT INTO evaluations VALUES(46278,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(46278,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(46278,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(46278,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(46278,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(46278,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46278,36,'cp -11','g1f3 f8e7 e1g1 e8g8 d2d3 b8c6 f3e5 c6e5 b2e5 c8d7 a2a4 d7c6 d1e1 f6d7 e5b2 e7f6 b2f6 d8f6 b1d2 a8d8 f2f4 f6d4 e2e3 d4c3 a1c1 a7a6 d2f3 c3f6 a4a5 f6e7 e1f2 h7h6 c1e1 e6e5 f4e5 d7e5 f3e5 e7e5');
INSERT INTO evaluations VALUES(51040,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51040,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(51040,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(51040,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(51040,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(51040,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(51040,36,'cp 73','d2d4 d7d5 f1d3 c7c5 d4c5 f8c5 e4d5 e6d5 c3d5 c5f2 e1f2 d8d5 d1f3 g8f6 f3d5 f6d5 g1f3 e8g8 a2a3 h7h6 h1d1 f8e8 c2c4 d5f6 b2b4 b8c6 c1b2 f6e4 f2g1 c8f5 d3f1 f7f6 c4c5 a8d8 d1e1 g8f7 f1c4 f7f8 a1d1 d8d1 e1d1 f5e6 c4d3');
INSERT INTO evaluations VALUES(54103,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(54103,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(54103,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(54103,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(54103,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(54103,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(54103,36,'cp 184','c1e3 c5d4 f3d4 a7a6 d1f3 f8b4 e1c1 b8d7 d4f5 e6f5 c3d5 g8f6 d5b4 d8e7 b4d3 e7e6 c1b1 e8g8 d3f4 e6e7 e3d4 d7e5 f3g3 f8e8 f1e2 e5g6 d4f6 e7f6 f4d5 f6c6 e2f3 f5f4 d5f4 c6a4 f4d3 c8f5 f3b7 a8d8 g3c7 d8d3 d1d3 f5d3 c2d3 a4d4 b7a6');
INSERT INTO evaluations VALUES(54101,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(54101,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(54101,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(54101,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(54101,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(54101,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(54101,36,'cp 176','c5d4 f3d4 a7a6 d1f3 f8b4 e1c1 b8d7 d4f5 e6f5 c3d5 g8f6 d5b4 d8e7 b4d3 e7e6 b2b3 e8g8 d3f4 e6e7 e3d4 f8e8 f1d3 d7e5 f3g3 e8d8 d4b2 b7b5 c1b1 e5d3 d1d3 d8d3 g3d3 c8e6 h1e1 a8e8 d3g3 g8h8 f4e6 f7e6 f2f3');
INSERT INTO evaluations VALUES(55518,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(55518,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(55518,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(55518,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(55518,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(55518,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(55518,36,'cp 19','d6e5 d1d8 e8d8 b1c3 g8h6 e3e4 c7c6 a2a4 a7a5 f1c4 f7f6 c1h6 g7h6 e1c1 d8c7 g1e2 b8a6 e2g3 a6c5 f2f3 f8e7 c1b1 h8f8 h1f1 c8d7 d1d2 a8d8 f1d1 b7b6 g3f5 d7f5 e4f5 d8d2 d1d2 h6h5 h2h4 f8d8 d2d8 c7d8 c4g8 h7h6 g8f7');
INSERT INTO evaluations VALUES(56633,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(56633,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(56633,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(56633,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(56633,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(56633,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(56633,36,'cp 20','d1d8 e8d8 b1c3 g8h6 e3e4 b8d7 a2a4 a7a5 f1c4 f7f6 g1e2 c7c6 c1h6 g7h6 e1c1 d8c7 e2g3 d7c5 h2h4 f8e7 f2f3 h8f8 d1d2 f6f5 e4f5 c8f5 d2e2 a8e8 c1b1 e7d6 g3f5 f8f5 c4d3 c5d3 c2d3 e8d8 b1c2 d6b4 e2e4 h6h5');
INSERT INTO evaluations VALUES(35410,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(35410,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(35410,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35410,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(35410,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(35410,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(35410,36,'cp 14','e8d8 b1c3 g8h6 e3e4 c7c6 f1c4 f7f6 a2a4 a7a5 g1e2 d8c7 c1h6 g7h6 e1c1 b8d7 e2g3 d7c5 f2f3 f8e7 h2h4 h8f8 d1d2 f6f5 e4f5 c8f5 d2e2 f5d7 g3e4 c5e4 e2e4 a8e8 c4d3 e7b4 c3d1 d7f5 e4e2 f5d3 c2d3');
INSERT INTO evaluations VALUES(35463,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(35463,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(35463,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(35463,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(35463,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(35463,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(35463,36,'cp 17','b1c3 g8h6 f1c4 f7f6 e3e4 c7c6 a2a4 a7a5 g1e2 d8c7 e2g3 b8a6 c1h6 g7h6 e1c1 a6c5 f2f3 f8e7 c1b1 h8f8 d1d2 c8d7 h1d1 a8d8 g3f5 d7f5 e4f5 d8d2 d1d2 h6h5 h2h4 e7d6 c3e4 c5e4 f3e4 d6b4 c2c3 b4c5 b1c2 f8d8 d2d8 c7d8');
INSERT INTO evaluations VALUES(55746,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(55746,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(55746,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(55746,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(55746,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(55746,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(55746,36,'cp 13','d6d5 b1c3 d5d4 c3b1 f8d6 d2d3 c7c5 g2g3 e6e5 f1g2 b8c6 e1g1 f7f6 f4e5 f6e5 f3h4 g8f6 h4f5 e8g8 b1d2 g8h8 b2b3 d6c7 a2a4 c8e6 d1e2 d8e8 d2c4 e8d7 c1g5 a8e8 g2h3 f6g8 g3g4 c6b4 f5h4 b7b6 f1f8 e8f8 g1h1 b4c6');
INSERT INTO evaluations VALUES(776,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(776,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(776,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(776,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(776,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(776,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(776,36,'cp 49','d4c5 b6c5 b2b4 c5b6 c3c4 e7g6 c1e3 b6d8 b1c3 d5c4 d1d8 e8d8 f1c4 c8d7 c4e2 c6e5 a1d1 a8c8 c3b5 a7a6 e3b6 d8e7 b5d6 c8c6 f3e5 g6e5 b6e3 b7b6 f2f4 e5g4 e2g4 c6d6 d1d6 e7d6 e3b6 f7f5 g4e2 d7b5 e2b5 a6b5 e1f2 d6c6 b6d4 f8d6');
INSERT INTO evaluations VALUES(4496,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(4496,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(4496,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(4496,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(4496,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(4496,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(4496,36,'cp 106','c3d5 c5e7 h2h3 g4f6 f1e1 d7d6 d5e7 d8e7 d3d4 e5d4 f3d4 c6d4 d1d4 e7e5 d4e5 d6e5 a2a4 a7a5 f2f3 f8e8 b2b3 c8e6 c4b5 e8d8 c1b2 c7c6 b5e2 f6d7 e1d1 f7f6 d1d6 d7f8 a1d1 d8d6 d1d6 g8f7 b2a3 h7h6');
INSERT INTO evaluations VALUES(6280,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(6280,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(6280,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(6280,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(6280,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(6280,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(6280,36,'cp 122','b4d2 d1d2 d7d6 b1c3 g8f6 e1c1 e8g8 c4d3 c6e7 c1b1 c7c5 h1e1 c8g4 d4c5 d6c5 h2h3 g4e6 d2e3 d8c7 g2g4 a8d8 g4g5 h6g5 f3g5 c5c4 d3c2 d8d1 e1d1 e7c6 e3c5 c7b6 c5b6 a7b6 c3b5 f8d8 d1d8 c6d8 b5d6 g7g6 g5e6 d8e6 d6c4 f6d7');
INSERT INTO evaluations VALUES(6287,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(6287,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(6287,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(6287,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(6287,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(6287,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(6287,36,'cp 141','d1d2 d7d6 b1c3 g8f6 e1c1 e8g8 c4d3 c6e7 c1b1 c7c5 h1e1 d8b6 d3c2 c8g4 h2h3 g4f3 g2f3 c5d4 d2d4 b6d4 d1d4 e7c6 d4d6 a8d8 d6d1 d8d1 e1d1 f8d8 f3f4 f6h5 c3e2 d8d1 c2d1 f7f5 e4f5 c6e7 d1c2 g8f8 c2e4 b7b6 b1c1 f8f7 e4f3 h5f6 e2g3 e7d5');
INSERT INTO evaluations VALUES(9774,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(9774,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9774,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(9774,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(9774,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9774,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9774,36,'cp 8','a7a6 c2c3 d7d6 b1d2 c5a7 d2f1 c6a5 c4b3 a5b3 a2b3 e8g8 f1g3 f8e8 h2h3 a6a5 d3d4 e5d4 c3d4 a7b6 c1d2 c7c6 b3b4 c8e6 b4a5 b6a5 d1c2 d8b6 d2c3 a5c3 b2c3 a8a1 e1a1 b6b3 c2d2 b3b5 e4e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(12727,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(12727,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(12727,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(12727,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(12727,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(12727,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(12727,36,'cp -41','f6e4 c3e3 d7d5 b1d2 e4d2 c1d2 f8g7 d2c3 e8g8 h2h3 c8f5 f1d3 f5d3 c2d3 d5d4 f3d4 c6e5 e1g1 d8b6 f1e1 f8d8 d4f5 g6f5 e3b6 a7b6 c3e5 e7e6 d3d4 g7e5 d4e5 d8d2 a2a4 a8d8 b2b3 g8g7 e1c1 b6b5 a4b5 d8d5 b5b6 d2b2 a1a7');
INSERT INTO evaluations VALUES(12749,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(12749,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(12749,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(12749,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(12749,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(12749,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(12749,36,'cp -41','c3e3 d7d5 b1d2 e4d2 c1d2 f8g7 d2c3 e8g8 h2h3 c8f5 f1d3 f5d3 c2d3 d5d4 f3d4 c6e5 e1g1 d8b6 f1e1 f8d8 d4f5 b6e3 e1e3 g6f5 c3e5 g7e5 e3e5 e7e6 a1d1 b7b6 g1f1 a8c8 f1e2 g8g7 e5b5 d8d4 e2e3 d4a4 a2a3 f5f4 e3e2 a4d4');
INSERT INTO evaluations VALUES(13148,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13148,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13148,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13148,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13148,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13148,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13148,36,'cp 0','e7e6 f1d3 f8d6 e1g1 e8g8 h2h3 b7b6 a2a3 c8b7 d1e2 a8c8 c1d2 d8e7 f1d1 d6b8 a1c1 h7h6 c3a2 f8d8 a2c3');
INSERT INTO evaluations VALUES(13669,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(13669,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13669,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13669,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13669,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(13669,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13669,36,'cp 0','f1d3 f8d6 h2h3 h7h6 e1g1 e8g8 d1e2 c6b4 d3b1 b7b6 c1d2 c8a6 c3b5 d8d7 a2a4 b4c6 f1d1 d6b8 d2e1 f8c8 f3d2 c6b4 d2f3');
INSERT INTO evaluations VALUES(15308,30,'cp 564',NULL);
INSERT INTO evaluations VALUES(15308,31,'cp 550',NULL);
INSERT INTO evaluations VALUES(15308,32,'cp 563',NULL);
INSERT INTO evaluations VALUES(15308,33,'cp 569',NULL);
INSERT INTO evaluations VALUES(15308,34,'cp 590',NULL);
INSERT INTO evaluations VALUES(15308,35,'cp 571',NULL);
INSERT INTO evaluations VALUES(15308,36,'cp 594','c4d5 f8d6 d2d4 e8g8 d5c6 e5d4 d1d4 b7c6 c1e3 c6c5 d4d3 h7h6 e1c1 c8e6 h1e1 f8e8 c1b1 d8f6 c3e4 f6f5 e4d6 c7d6 d3f5 e6f5 d1d6 c5c4 f3d4 f5h7 d4b5 a8d8 d6d8 e8d8 b5a7 h7e4 f2f3 e4g6 a7c6 d8e8 h2h3');
INSERT INTO evaluations VALUES(15276,30,'cp 587',NULL);
INSERT INTO evaluations VALUES(15276,31,'cp 578',NULL);
INSERT INTO evaluations VALUES(15276,32,'cp 553',NULL);
INSERT INTO evaluations VALUES(15276,33,'cp 582',NULL);
INSERT INTO evaluations VALUES(15276,34,'cp 589',NULL);
INSERT INTO evaluations VALUES(15276,35,'cp 578',NULL);
INSERT INTO evaluations VALUES(15276,36,'cp 571','f8d6 d5c6 b7c6 d2d4 e5d4 d1d4 e8g8 c1e3 c6c5 d4d3 d8f6 e1c1 c8b7 f3g5 f6g6 g5e4 d6e5 e4c5 g6d3 d1d3 b7g2 h1g1 g2c6 e3d4 e5f4 c1b1 g7g6 c3e4 f7f5 e4f6 f8f6 d4f6 f4h2 g1d1 a8e8 d3c3 g8f7 f6g5 h2e5 c3a3 e8a8 c5d3 e5d6 a3c3 c6b5 b2b3');
INSERT INTO evaluations VALUES(29665,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(29665,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(29665,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(29665,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(29665,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(29665,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(29665,36,'cp -22','b8d7 d3e2 e7e6 c2c4 c7c6 d1b3 d8c7 c1d2 a7a5 b1c3 d7b6 c4d5 e6d5 a1c1 a5a4 b3c2 g4h5 e2d3 h5g6 f3e5 g6d3 e5d3 f8e7 h2h3 e8g8 e1g1 f8e8 f1d1 e7f8 b2b3 a4b3 a2b3 f8a3 c1a1 c7e7 d1b1 a3d6 a1a8 e8a8');
INSERT INTO evaluations VALUES(29872,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(29872,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(29872,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(29872,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(29872,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(29872,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(29872,36,'cp 44','c7c6 h2h3 g4h5 d5c6 b7c6 g2g4 h5g6 d1e2 f8e7 h3h4 h7h5 g4g5 f6d7 c1e3 d7b6 f1h3 e8g8 a1d1 d8c7 e1g1 b8d7 f3e1 d6d5 e4d5 c6d5 c3d5 b6d5 d1d5 a8d8 e1g2 d7b6 e3b6 c7b6 c2c4 d8d5 c4d5 f8b8 b2b3 e7d6 f1d1 e5e4');
INSERT INTO evaluations VALUES(29341,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(29341,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(29341,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(29341,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(29341,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(29341,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(29341,36,'cp 56','h2h3 g4h5 d5c6 b7c6 g2g4 h5g6 d1e2 f8e7 h3h4 h7h5 g4g5 f6d7 c1e3 d7b6 f1h3 e8g8 e1g1 d8c7 f1d1 b8d7 f3h2 c7b7 a2a4 a7a5 h2f1 f8d8 h3g2 d7f8 f1g3 f8e6 g3h5 b6c4 a1b1 c4e3 f2e3 b7b6 h5g3 d6d5 h4h5 e6f4 e2f2 f4h5 e4d5 h5g3 f2g3 c6d5 c3d5');
INSERT INTO evaluations VALUES(30487,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(30487,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(30487,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(30487,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(30487,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(30487,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(30487,36,'cp 205','d4d5 a7a6 c1f4 b6b5 d1e2 c5c4 e1c1 f8b4 f4d2 d8c7 h2h4 e6e5 h4h5 d7d6 f3h4 g8e7 f2f4 b8d7 a2a3 b4c5 g2g4 a8b8 h4f5 e7f5 e4f5 b5b4 a3b4 c5b4 f4e5 d7e5 h5h6 g7g6 f5g6 h7g6 c3e4 b7c8 g4g5 b4d2 e2d2 e8g8');
INSERT INTO evaluations VALUES(30492,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(30492,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(30492,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(30492,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(30492,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(30492,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(30492,36,'cp 207','a7a6 d1e2 b6b5 c1f4 c5c4 e1c1 f8b4 f4d2 g8f6 g2g4 e8g8 g4g5 f6h5 f3e5 b4c3 b2c3 d7d6 e5g4 g7g6 g4h6 g8h8 f1h3 d8a5 c1b1 e6d5 h3g4 d5d4 g4h5 d4c3 d2c1 g6h5 a2a3 b5b4 e2c4 a5b5 h6f7 h8g7 c4b5 a6b5 f7d6 b7c8 h1e1 b8c6 d6b5');
INSERT INTO evaluations VALUES(30782,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(30782,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(30782,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30782,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(30782,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(30782,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30782,36,'cp 25','e7e6 e2e4 g8f6 d2d3 c7c5 g2g3 f8e7 f1g2 d7d6 e1g1 e8g8 b2b3 a7a6 c1b2 f8e8 f1e1 e6e5 c3d5 b8c6 e1f1 b7c8 f3d2 c6d4 f2f4 f6d5 c4d5 e7f6 b2d4 c5d4 f4e5 f6e5 d1f3 f7f6 d2c4 c8d7 a2a4 b6b5 c4e5');
INSERT INTO evaluations VALUES(30636,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(30636,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(30636,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30636,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30636,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(30636,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(30636,36,'cp 29','e2e4 g8f6 d2d3 d7d6 g2g3 c7c5 f1g2 f8e7 b2b3 a7a6 e1g1 e8g8 c1b2 f8e8 f1e1 e6e5 c3d5 b8c6 g2h3 b6b5 f3d2 e7f8 f2f4 f6d5 c4d5 c6d4 d2f3 b7c8 h3c8 a8c8 a1c1 e5f4 b2d4 c5d4 c1c8');
INSERT INTO evaluations VALUES(30634,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(30634,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(30634,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30634,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(30634,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(30634,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30634,36,'cp 25','g8f6 d2d3 c7c5 g2g3 d7d6 f1g2 f8e7 b2b3 e8g8 e1g1 f8e8 f3e1 a7a6 e1c2 b8c6 d3d4 c5d4 c2d4 d8b8 d4c6 b7c6 a2a4 b8b7 d1e2 a8c8 f1d1 f6d7 c1f4 e7f6 a1c1 d7c5 b3b4 c5b3 c1b1 f6c3 b1b3 c3f6 d1d6 c6a4');
INSERT INTO evaluations VALUES(39032,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(39032,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(39032,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(39032,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(39032,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(39032,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(39032,36,'cp 75','d7d6 a2a4 g8f6 f1d3 b7b6 a4a5 a7a6 a5b6 c7b6 c1f4 c8b7 d1e2 e8g8 b1d2 b6b5 e1g1 b8c6 f4e3 c6b8 h2h3 b8d7 f1c1 a8c8 f3h2 e6e5 d4d5 d7c5 c3c4 b5c4 c1c4 c5d3 e2d3 d8d7 c4c8 f8c8 d2c4');
INSERT INTO evaluations VALUES(38560,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(38560,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(38560,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(38560,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(38560,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(38560,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(38560,36,'cp 86','a2a4 b8d7 a4a5 g8f6 f1d3 e8g8 e1g1 a7a6 c1g5 h7h6 g5d2 e6e5 f1e1 f6h7 d3c2 f7f5 d4e5 d6e5 c3c4 f5f4 h2h3 h7g5 d2b4 c7c5 b4a3 g5f3 d1f3 d8a5 b1c3 f8f7 e1d1 d7f8 c3d5 f8e6 b2b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(39436,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(39436,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(39436,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(39436,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(39436,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(39436,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(39436,36,'cp 144','g8f6 d4e5 d8e7 e5e7 e8e7 b1c3 d7d6 c1g5 b8d7 f1e2 h7h6 g5e3 b7b6 f2f3 c8b7 a2a4 a7a6 e1f2 d7e5 h2h4 a8e8 b2b4 e7d8 g2g4 d8c8 h4h5 e8f8 f2g3 f8e8 a1b1 e8e7 b1d1 e7e8 d1a1 e8e7 e3d4 h8e8 a4a5 b6b5 c3d1');
INSERT INTO evaluations VALUES(42842,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(42842,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(42842,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(42842,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(42842,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(42842,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(42842,36,'cp 155','b1c3 b8c6 d4d3 e8g8 c1g5 h7h6 g5h4 c6e5 d3d4 e5c6 d4a4 c6e5 c3d5 g7g5 d5f6 d8f6 h4g3 d7d5 e1c1 d5e4 a4e4 e5g6 f1d3 c8d7 h2h4 f8e8 e4b4 g6h4 g3h4 g5h4 b4h4 f6h4 h1h4 g8g7 d3e4 d7c8 d1d3 g7f8 d3g3 a8b8 f2f3 f7f5');
INSERT INTO evaluations VALUES(42841,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(42841,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(42841,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(42841,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(42841,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(42841,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(42841,36,'cp 147','b8c6 d4d3 d7d6 c1g5 h7h6 g5h4 c6e5 d3d4 c8d7 e1c1 d7c6 f1b5 e8g8 f2f4 e5g6 h4f6 d8f6 d4f6 g7f6 h1f1 g8g7 d1d4 f8h8 d4c4 g6e7 b5c6 b7c6 f4f5 g7f8 f1f4 f8e8 c3e2 e8d7 f4h4 a8g8 g2g3 g8g5 c4a4 d6d5 a4a7 d5e4 h4e4');
INSERT INTO evaluations VALUES(10527,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(10527,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(10527,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(10527,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(10527,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(10527,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(10527,36,'cp 161','d4d3 d7d6 c1d2 d8e7 e1c1 c8d7 f2f3 e8c8 h2h4 c8b8 d2e1 e7e5 g2g4 a7a6 d3d2 h7h6 h4h5 e5g5 e1h4 g5d2 c1d2 d7e6 d2c1 b8c8 f1e2 d8e8 h4f6 g7f6 b2b3 c6e5 a2a3 e5d7 f3f4');
INSERT INTO evaluations VALUES(51961,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(51961,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(51961,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(51961,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(51961,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(51961,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(51961,36,'cp 72','d4d5 g8f6 e2e4 d7d6 f1c4 f8e7 g1e2 e6d5 c4d5 e8g8 e1g1 b8c6 e2f4 f8e8 d5c4 a7a6 a2a4 e7f8 f2f3 c8d7 c1e3 a8b8 d1d2 c6e5 c4e2 b7b5 a4b5 a6b5 f4d5 f6d5 c3d5 d7e6 e3f2 e5c6 f1e1 e6d5 d2d5 c6b4 d5d2');
INSERT INTO evaluations VALUES(1760,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(1760,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(1760,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(1760,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(1760,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(1760,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(1760,36,'cp 125','d4d3 f8e7 e1g1 f6g6 b1c3 c7c6 c1f4 c8e6 a1d1 a8d8 c4e6 f7e6 f4e3 g8f6 e3a7 e8g8 a2a3 d8e8 a7d4 d6d5 e4d5 g6d3 d1d3 e6d5 g2g3 f6d7 f1e1 e7c5 e1e8 f8e8 d4c5 d7c5 d3d2 g8f7 g1g2 b7b5 d2e2 c5e6 a3a4 b5a4 c3a4');
INSERT INTO evaluations VALUES(9435,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(9435,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9435,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9435,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9435,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9435,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(9435,36,'cp 2','c4b3 c6d4 h2h3 e8g8 e1g1 d4b3 a2b3 f8e8 c1d2 a7a6 c3e2 b4d2 d1d2 f6h5 b3b4 c8d7 e2g3 h5g3 f2g3 d7e6 g3g4 d6d5 e4d5 e6d5 a1e1 d8d6 f3h4 g7g6 h4f3 d5c6 d2f2 f7f6 f3h4 g8g7 b2b3 a6a5 b4a5 a8a5');
INSERT INTO evaluations VALUES(10648,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(10648,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(10648,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(10648,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(10648,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(10648,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(10648,36,'cp -34','c3d4 d7d5 d4c5 d5c4 d1e2 d8e7 e2c4 c8f5 b1c3 e8c8 f1e1 h8e8 c1e3 e7f6 c3b5 a7a6 b5d4 f5e6 c4c2 e6g4 h2h3 g4f3 d4f3 f6g6 c2c4 g6e6 c4e6 e8e6 a1d1 d8d1 e1d1 e6e7 g1f1 e7d7 f1e2 d7d1 e2d1 c8d7 f3e1 d7e6 e1d3 e6d5 f2f3');
INSERT INTO evaluations VALUES(10646,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(10646,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(10646,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(10646,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(10646,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(10646,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(10646,36,'cp -29','d7d5 d4c5 d5c4 d1e2 d8e7 e2c4 c8f5 b1c3 e8c8 f1e1 h8e8 c1e3 e4c3 c4c3 e7f6 c3f6 g7f6 e1d1 f5g4 g1f1 g4f3 g2f3 e8e5 d1d8 c6d8 a1c1 e5h5 f1g2 d8c6 f3f4 c6e7 b2b4 a7a6 a2a4 c7c6 b4b5 c6b5 a4b5 a6b5 c1b1 b5b4 b1b4 e7d5 b4d4 c8c7 d4c4');
INSERT INTO evaluations VALUES(11536,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(11536,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(11536,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(11536,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(11536,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(11536,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(11536,36,'cp -11','e5d4 e1g1 f8e7 e4e5 f6e4 f3d4 e8g8 d4f5 d7d5 e5d6 c8f5 d6e7 c6e7 a4b3 d8d1 f1d1 a8d8 d1e1 e4c5 b1c3 c5b3 c2b3 e7g6 c1e3 f7f6 e1d1 f8e8 h2h3 g8f7 d1d8 e8d8 a1c1 g6e7 f2f3 d8d7 e3f2 d7d2 c3e4 d2b2 c1c7 b2b1 g1h2 f5e4 f3e4');
INSERT INTO evaluations VALUES(11548,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(11548,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(11548,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(11548,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(11548,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(11548,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(11548,36,'cp -12','e1g1 f8e7 e4e5 f6e4 f3d4 e8g8 d4f5 d7d5 e5d6 c8f5 d6e7 c6e7 a4b3 d8d1 f1d1 a8d8 d1e1 e4c5 b1c3 c5b3 c2b3 e7g6 c1e3 d8d7 a1d1 f8d8 d1d7 d8d7 e1d1 d7d1 c3d1 f7f6 f2f3 g8f7 d1c3 f7e6 c3e2 b7b6 e2d4 e6e5 d4f5 e5f5 g1f2 g6e7 f2e2 e7d5 e3d2');
INSERT INTO evaluations VALUES(11547,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(11547,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(11547,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(11547,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(11547,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(11547,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(11547,36,'cp -15','f8e7 e4e5 f6e4 f3d4 e8g8 d4f5 d7d5 e5d6 c8f5 d6e7 c6e7 a4b3 d8d1 f1d1 a8d8 d1e1 e4c5 b1c3 c5b3 c2b3 e7g6 c1e3 f7f6 e1d1 g8f7 f2f3 f8e8 d1d8 e8d8 a1d1 d8d1 c3d1 g6e5 h2h4 f7e6 d1c3 h7h5 c3e2 e5d3 e2f4 d3f4 e3f4 f5b1 a2a3 e6d7 g1f2');
INSERT INTO evaluations VALUES(7375,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(7375,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(7375,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(7375,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(7375,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(7375,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(7375,36,'cp -12','e4e5 f6e4 f3d4 e8g8 d4f5 d7d5 e5d6 c8f5 d6e7 c6e7 a4b3 d8d1 f1d1 a8d8 d1e1 e4c5 b1c3 e7c6 c1e3 c5b3 c2b3 f7f6 e3f4 f8f7 a1d1 c6d4 h2h4 f7d7 f2f3 c7c5 g1f2 b7b5 d1d2 d4e6 d2d7 d8d7 f4e3 g8f7 g2g3 e6d4 e3d4 c5d4');
INSERT INTO evaluations VALUES(21345,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(21345,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21345,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(21345,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21345,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(21345,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(21345,36,'cp 11','d5e4 c3e4 d8d5 d1d3 a7a6 c2c4 f8b4 c1d2 b4d2 e1d2 d5d7 b5c6 d7c6 g1f3 g8f6 h1e1 f6e4 e1e4 c6d6 a1e1 f7f6 e4h4 c8d7 h4h7 d6b4 d2e3 h8h7 d3g6 e8f8 g6h7 a8e8 h7d3 b4b2 d3b3 b2b3 a2b3 e8c8 e1a1 c7c5 d4c5 c8c5 h2h4 a6a5 f3d2 c5e5 e3d3 d7e8 d2f3 e5f5 f3d4 f5f2 a1a5 f2g2 d4e6 f8e7');
INSERT INTO evaluations VALUES(26667,30,'cp 337',NULL);
INSERT INTO evaluations VALUES(26667,31,'cp 352',NULL);
INSERT INTO evaluations VALUES(26667,32,'cp 358',NULL);
INSERT INTO evaluations VALUES(26667,33,'cp 362',NULL);
INSERT INTO evaluations VALUES(26667,34,'cp 375',NULL);
INSERT INTO evaluations VALUES(26667,35,'cp 382',NULL);
INSERT INTO evaluations VALUES(26667,36,'cp 378','d4d5 c6a5 d1a4 c7c6 b2b4 b7b5 a4a5 d8a5 b4a5 b5b4 e2e4 b4c3 e4f5 c6d5 f1e2 e7e5 f5e6 f7e6 f3g5 f8d6 g5e6 e8d7 e6d4 g8f6 c1e3 h8b8 e1g1 d6c5 f1c1 b8b2 e2d1 b2d2 d1a4 d7c8 d4b5 c5e3 f2e3 f6e4 b5c3 e4c3 c1c3 a8b8 c3c2 b8b2 c2d2 b2d2 a1d1 d2a2 a4c6 c4c3 c6d5 a2a5 d5e4 a5e5 e4d3 e5e3 d3h7 c3c2');
INSERT INTO evaluations VALUES(32007,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(32007,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(32007,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(32007,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(32007,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(32007,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(32007,36,'cp 78','g4f3 d1f3 c7c6 d2d4 e7e6 c1e3 g8f6 f1d3 d8a5 e1g1 b8d7 c3e2 f8d6 c2c4 e8g8 a2a3 f8e8 b2b4 a5d8 a1d1 h7h6 b4b5 c6b5 c4c5 d6c7 f3b7 a7a6 a3a4 b5a4 b7b4 f6d5 b4a4 d7f6 e3d2 d8e7 e2g3 c7f4 g3e4 f4d2 d1d2 e8b8 f1c1 e7c7 e4f6 d5f6 g1h1 a6a5');
INSERT INTO evaluations VALUES(32012,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(32012,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(32012,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(32012,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(32012,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(32012,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(32012,36,'cp 83','d1f3 c7c6 d2d4 e7e6 c1e3 g8f6 f1d3 d8a5 e1g1 b8d7 a2a3 f8d6 a1d1 e8g8 f1e1 a8d8 e3d2 a5b6 d2c1 f8e8 c3e4 f6e4 f3e4 d7f6 e4f3 a7a5 c2c3 d8d7 g2g3 c6c5 d4c5 d6c5 b2b4 c5e7 c1e3 b6c7 e3f4');
INSERT INTO evaluations VALUES(32011,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(32011,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(32011,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(32011,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(32011,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(32011,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(32011,36,'cp 81','c7c6 d2d4 g8f6 c1f4 e7e6 e1c1 f8b4 c3e4 f6e4 f3e4 d8d5 e4e5 d5e5 f4e5 b4f8 c2c4 b8d7 e5f4 e8c8 f1e2 d7f6 c1c2 f8d6 f4e3 a7a5 e2f3 h7h5 g2g4 h5g4 h3g4 d8g8 g4g5 f6d7 h1h8 g8h8 g5g6 f7f6 d4d5 e6d5 c4d5 c8c7 d5c6');
INSERT INTO evaluations VALUES(31226,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(31226,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(31226,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(31226,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(31226,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(31226,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(31226,36,'cp 93','d2d4 g8f6 c1e3 e7e6 a2a3 f6d5 f1d3 f8e7 c3d5 c6d5 c2c3 b8c6 e1g1 e8g8 g2g3 g7g6 a1e1 e7f6 h3h4 h7h5 e3h6 f6g7 h6f4 f8e8 d3c2 d8d7 e1e2 e6e5 d4e5 c6e5 f4e5 e8e5 e2e5 g7e5 c2b3 a8d8 f1d1 d5d4 f3e4');
INSERT INTO evaluations VALUES(40408,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(40408,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(40408,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(40408,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(40408,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(40408,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(40408,36,'cp 109','f1b5 c8d7 e4e5 d6e5 d4e5 f6g4 c1f4 d8b6 d1e2 d7b5 c3b5 b8c6 e1g1 e8g8 h2h3 g4h6 f1d1 a8d8 g2g4 a7a6 b5c3 f7f5 e5f6 e7f6 c3d5 b6b5 e2b5 a6b5 a1c1 h6f7 c1c5 b5b4 c5b5 d8a8 b5b7 a8a2 b7b6 f7d8 d1e1 c6e5 f3e5 f6e5 f4g5 a2b2 g5e7 f8f7 e7d8 f7f2 b6b4');
INSERT INTO evaluations VALUES(41883,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(41883,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41883,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(41883,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41883,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41883,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41883,36,'cp 27','e8g8 g1e2 f8e8 e1g1 c7c6 h2h3 h7h6 b2b3 d6d5 c4d5 c6d5 d2d4 c5d6 d4e5 d6e5 c1b2 b8c6 d1d2 c8f5 f1d1 f5e4 c3e4 f6e4 g2e4 d5e4 b2e5 c6e5 d2d8 a8d8 g3g4 d8d1 a1d1 e8c8 e2g3 c8c2 d1d8 g8h7 g3e4 c2a2 d8e8 e5f3 g1g2 f3g5 e4d6 a2b2 d6b7 b2b3');
INSERT INTO evaluations VALUES(36999,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36999,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36999,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(36999,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(36999,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(36999,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36999,36,'cp 33','g1e2 f8e8 e1g1 c7c6 d2d4 c5b4 d1c2 c8e6 c3e4 f6e4 g2e4 h7h6 a2a3 b4a5 b2b4 a5c7 c1b2 b8d7 a1c1 e6h3 e4g2 h3g2 g1g2 d8e7 f1d1 a7a5 b4b5 a8c8 a3a4 c6b5 c4b5 d6d5 c2b3 d7b6 d4e5 c7e5 b2e5 e7e5 g2g1 c8c4 c1c4 d5c4 b3c2 e8c8');
INSERT INTO evaluations VALUES(37000,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(37000,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(37000,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(37000,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(37000,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(37000,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(37000,36,'cp 39','f8e8 e1g1 c7c6 d2d4 c5b4 d1c2 d8e7 a2a3 b4c3 e2c3 h7h6 d4e5 d6e5 e3e4 b8a6 h2h3 c8e6 b2b3 b7b6 c1b2 a6c5 b3b4 c5d7 c3d5 c6d5 c4d5 a8c8 c2d3 e6d5 e4d5 e5e4 d3d2 e4e3 f2e3 e7e3 d2f2 e3f2 f1f2');
INSERT INTO evaluations VALUES(47715,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47715,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(47715,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(47715,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(47715,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(47715,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47715,36,'cp 12','c1f4 f8d6 f4e5 c8e6 f1d3 b8c6 e1g1 e8g8 e5d6 d8d6 h2h3 f8e8 f1e1 a7a6 c3e2 f6e4 c2c3 e6f5 d3e4 f5e4 e2g3 e4g6 d1d2 c6a5 b2b3 a5c6 e1e3 e8e3 d2e3 c6e7 a1e1 a8e8 e3d2 f7f6');
INSERT INTO evaluations VALUES(57109,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(57109,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(57109,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(57109,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(57109,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(57109,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(57109,36,'cp -125','d7d5 e2e3 b8c6 d2d4 f8d6 d4e5 c6e5 g1f3 g8f6 f3e5 d6e5 b1d2 h7h6 d2f3 c8g4 f1e2 e5d6 c1d2 g4h5 c2c4 d5c4 d2c3 h5g4 f3h4 g4e6 d1c2 d8d7 c3f6 g7f6 e2c4 e8c8 e1c1 e6g4 f2f3 g4e6 c4e6 f7e6 h4g6 h8g8');
INSERT INTO evaluations VALUES(56323,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(56323,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(56323,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(56323,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(56323,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(56323,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(56323,36,'cp -115','e2e3 b8c6 d2d4 f8d6 d4e5 c6e5 g1f3 g8f6 f3e5 d6e5 b1d2 d8e7 d2f3 e5b2 c1b2 e7b4 f3d2 b4b2 a1b1 b2e5 h5h6 g7g6 f1e2 e8g8 c2c4 d5c4 d2c4 e5c3 c4d2 f8d8 b1c1 c3a5 h1h4 c7c6 d1c2 c8e6 h4a4 a5g5 c2b2 e6d5 a4f4 f6e8 b2b7 g5g2');
INSERT INTO evaluations VALUES(1288,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(1288,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(1288,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(1288,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(1288,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(1288,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(1288,36,'cp 91','a4b3 f8c5 c2c3 g8e7 d2d4 e5d4 c1g5 f6d6 e4e5 d6g6 f1e1 e7f5 b3c2 h7h6 g5f4 g6e6 b1d2 d4c3 d2e4 c3b2 a1b1 c5b4 e4f6 g7f6 c2b3 b4e1 b3e6 f7e6 d1e1 f6e5 f3e5 c8b7 b1b2 h8g8 f2f3 c6e5 e1e5 a8c8 b2e2 b7f3 e5f5 f3e2 f5h7 g8f8 h7g6 e8e7 g6h6 f8f4 h6f4 d7d6 a2a3');
INSERT INTO evaluations VALUES(4075,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(4075,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(4075,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(4075,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(4075,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(4075,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(4075,36,'cp -31','e7e5 f4g5 h7h6 g5h4 d8e7 d3c2 d6d5 d4e5 d7e5 f3e5 e7e5 e1g1 b7b6 a2a4 c8b7 f1e1 e5e6 c2b3 e6c6 d1c2 a8e8 f2f3 e8e7 a1d1 f8e8 d2f1 e7d7 h2h3 c6c5 c2d2 a7a5 b3c2 b7a6 f1g3 c5d6 b2b3 c7c5');
INSERT INTO evaluations VALUES(9012,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(9012,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(9012,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(9012,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(9012,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(9012,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(9012,36,'cp 39','d8e7 c2c4 d5c4 d3c4 e8g8 f3e5 f8d8 e5c6 b7c6 d1b3 c6c5 b3a3 a8b8 d4c5 c8b7 e1g1 f6e4 c4e2 h7h6 b1c3 e7c5 a3c5 e4c5 b2b4 c5d3 f1b1 c7c5 b4b5 g8f8 e2d3 d8d3 c3a4 d3d5 b1c1 c5c4 b5b6 d5b5 b6a7 b8a8 c1c4 a8a7');
INSERT INTO evaluations VALUES(12824,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12824,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12824,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(12824,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(12824,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12824,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(12824,36,'cp 7','f1g2 e7e6 g1f3 d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 c8g4 d4e2 f8b4 h2h3 b4c3 b2c3 g4f5 c1g5 h7h6 g5f6 d8f6 d1d5 e8g8 e1g1 f5c2 d5b5 a8e8 b5b7 e8e2 b7c6 f6c6 g2c6 c2f5 g3g4 f5e6 c6b5 e2e5 c3c4 f8c8 a1c1 e5c5 g1g2 e6c4 b5c4 c5c4 c1c4 c8c4');
INSERT INTO evaluations VALUES(13115,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(13115,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(13115,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13115,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(13115,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(13115,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(13115,36,'cp 31','f1b5 c8d7 e1g1 e7e6 d2d4 c5d4 e3d4 d8c7 f1e1 a7a6 b5f1 b7b5 b1d2 f8e7 c2c3 e8g8 f1d3 f8c8 d1e2 h7h6 g2g3 e7f8 b3b4 c6e7 d2b3 e7f5 b3c5 f5d6 f3e5 d7e8 b2c1 f6e4 a2a3 e4c3 e2g4 f7f5 g4h3 c3e4');
INSERT INTO evaluations VALUES(13071,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(13071,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(13071,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13071,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(13071,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13071,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13071,36,'cp 22','c8d7 e1g1 e7e6 d2d4 c5d4 e3d4 d8c7 f1e1 a7a6 b5f1 b7b5 b1d2 a8b8 a2a3 f8d6 b3b4 a6a5 c2c3 a5a4 f1d3 e8g8 d1e2 c6e7 f3e5 d7e8 d2f3 e7g6 e2e3 g6e7 h2h3 e7c8 b2c1 c8b6 c1d2 b6c4 e3e2 c4d2 e2d2 g8h8 f3g5 h8g8');
INSERT INTO evaluations VALUES(24466,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(24466,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(24466,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24466,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24466,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24466,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(24466,36,'cp 29','e1g1 e7e6 d2d4 c5d4 e3d4 d8c7 f1e1 a7a6 b5f1 b7b5 b1d2 a8b8 a2a3 f8d6 b3b4 a6a5 c2c3 a5a4 f1d3 e8g8 d1e2 d7e8 f3e5 c6e7 a1c1 e7c8 c1c2 c8b6 d3b5 e8b5 e2b5 b6c4 b5a4 c4b2 c2b2 c7c3 b2c2 c3d4 e5c6 d6h2 g1h2 f6g4 h2g3 d4f2 g3g4 h7h5 g4h3 f2e1');
INSERT INTO evaluations VALUES(15283,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(15283,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(15283,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(15283,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(15283,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(15283,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(15283,36,'cp -61','d8d5 d2d3 b7b6 c1e3 c8b7 d3d4 e5d4 f3d4 f8e7 d4c6 d5c6 d1g4 a8d8 e3d4 c6g6 g4g6 h7g6 d4e3 f7f5 f1b5 e8f7 e1g1 e7f6 c2c3 b7d5 a3a4 c7c6 b5e2 f6e5 h2h3 f5f4 e3d4 e5d4 c3d4 h8e8 e2d3 f4f3 g2g3 d5e6 d3e4 e6h3 e4c6 e8e2 f1c1 e2b2 c6f3 d8d4');
INSERT INTO evaluations VALUES(18622,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(18622,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(18622,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(18622,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(18622,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(18622,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(18622,36,'cp 112','d2d4 c5d4 f3d4 g8f6 f1e1 d7d6 b1a3 f8g7 d4b5 e8g8 b5d6 a7a6 c4f1 d8c7 d6c8 a8c8 c2c3 f8d8 d1e2 c6e5 c1f4 f6h5 f4e5 g7e5 g2g3 h5f6 a3c2 f6d7 a1d1 e5g7 f2f4 g7f8 e4e5 f8c5 g1h1 b7b5 h2h4 d7b6 h1h2 d8d1 e1d1 c8d8 h4h5 d8d1 e2d1 b6d5 h5h6 d5e3 c2e3 c5e3');
INSERT INTO evaluations VALUES(18629,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(18629,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(18629,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(18629,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(18629,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(18629,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(18629,36,'cp 113','c5d4 f3d4 g8f6 b1c3 f8g7 c1f4 c6d4 d1d4 e8g8 e4e5 f6h5 f4d2 d7d5 c4e2 f7f6 f2f4 g7h8 g2g4 h5g7 c3b5 c8d7 d2b4 d7b5 e2b5 f8f7 e5f6 f7f6 b5d3 f6f7 d4e3 a7a5 b4a3 d8h4 e3g3 h4g3 h2g3 g7e8 f1e1 e8c7 g1g2 a5a4');
INSERT INTO evaluations VALUES(18643,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(18643,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(18643,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(18643,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(18643,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(18643,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(18643,36,'cp 103','f3d4 g8f6 f1e1 d7d6 b1a3 f8g7 d4b5 e8g8 b5d6 a7a6 c4f1 d8c7 d6c8 a8c8 c2c3 f8d8 d1e2 c6e5 c1f4 f6h5 f4e3 g7f8 a3c2 e5d3 e1b1 d3c5 e3g5 d8d7 b1d1 d7d1 a1d1 c7e5 g5e3 e5e4 g2g3 h5f6 d1d4 e4c6 f1g2 c6e8 e3f4 f6d7 d4d1 e6e5 f4e3 e5e4 e3d4 f8g7 b2b4 g7d4 d1d4');
INSERT INTO evaluations VALUES(18638,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(18638,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(18638,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(18638,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(18638,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(18638,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(18638,36,'cp 85','g8f6 f1e1 d7d6 b1a3 f8g7 d4b5 e8g8 b5d6 a7a6 c4f1 d8c7 d6c8 f8d8 d1e2 a8c8 c2c3 c6e5 c1f4 f6h5 f4e3 g7f8 a1d1 d8d1 e1d1 c7a5 h2h3 b7b5 e3g5 h7h6 g5e3 h5f6 e3d4 f6d7 f2f4 e5c6 d4f2 d7f6 f2h4 f8c5 g1h1 f6h5 f4f5 a5c7 a3c2 h5g3 h4g3 c7g3 f5e6 f7e6 c2d4 c6e5 d4e6');
INSERT INTO evaluations VALUES(33207,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(33207,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(33207,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(33207,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(33207,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(33207,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(33207,36,'cp 153','g1f3 g8e7 e1g1 h7h6 d2d4 e5d4 c3b5 b8a6 e4e5 f6b6 f3d4 d7d6 e5d6 c7d6 c2c3 e8g8 a2a4 d6d5 c4d3 c8d7 c1e3 b6f6 f1e1 d7b5 a4b5 a6c7 h2h4 c5b6 d3c2 c7e6 d1d3 g7g6 d4e6 f7e6 e3d4 b6d4 d3d4 a7a6 a1a4 f6d4 a4d4 a6b5 e1e6 a8a1 d4d1 a1d1 c2d1');
INSERT INTO evaluations VALUES(37733,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(37733,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(37733,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(37733,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(37733,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(37733,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(37733,36,'cp 81','d7d5 e4e5 g8h6 h2h3 e8g8 b1c3 f7f6 c1f4 c8e6 d1b3 b8d7 c4d5 c6d5 f1b5 f6e5 b5d7 d8d7 f4h6 g7h6 f3e5 d7d8 e1g1 d8b6 b3b6 a7b6 f1e1 h6d2 e1e2 d2c3 b2c3 e6f5 e5f3 f8e8 a1e1 f5e4 f3d2 a8a3 d2e4');
INSERT INTO evaluations VALUES(37924,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(37924,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(37924,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(37924,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(37924,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(37924,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(37924,36,'cp 82','e4e5 g8h6 h2h3 e8g8 b1c3 f7f6 c4d5 c6d5 f1b5 f6e5 d4e5 e7e6 e1g1 d8c7 f1e1 c8d7 c1g5 d7b5 c3b5 c7d7 b5d4 h6f5 d4b3 b7b6 g2g4 f5e7 b3d4 e7c6 g5f6 c6d4 f3d4 b8a6 f6g7 g8g7 e1e3 a6c5 a1c1 f8f4 f2f3 a8f8 e3c3 f8f7 c1c2');
INSERT INTO evaluations VALUES(46979,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(46979,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(46979,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(46979,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(46979,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(46979,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(46979,36,'cp -48','b8c6 b1c3 c6e5 f3e2 e5c4 e2c4 e7e6 c4e2 f8e7 g1f3 a7a6 a2a4 b7b6 d2d4 c8b7 c1f4 e8g8 e1g1 d8d7 f1e1 b6b5 a4b5 a6b5 e2b5 d7b5 c3b5 f6e4 f3d2 g7g5 f4e3 f8c8 c2c4 e4d2 e3d2 b7a6 h2h4 a6b5 c4b5 a8a1 e1a1 g5h4');
INSERT INTO evaluations VALUES(50587,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(50587,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(50587,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(50587,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(50587,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(50587,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(50587,36,'cp -32','e7e6 a2a4 g8f6 d2d3 d7d5 c4b3 d5e4 c3e4 f8e7 g1f3 b8c6 e4f6 e7f6 c2c3 b7b6 e1g1 c8b7 c1f4 e8g8 d1e2 c6e7 a1d1 e7f5 f1e1 b6b5 f3e5 d8e7 a4b5 a6b5 d3d4 c5d4 e2b5 b7a6 b5c6 f8c8 c6b6 a6b7 c3d4 c8d8 e5f7 g8f7');
INSERT INTO evaluations VALUES(4310,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(4310,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(4310,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(4310,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4310,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4310,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4310,36,'cp 0','c5e7 c4b3 d7d5 b1d2 a7a5 b4b5 c6a7 a2a4 c7c6 b5b6 d8b6 f3e5 b6c7 e5f3 d5e4 d2e4 c8f5 e1g1 a8d8 d1e2 c6c5 f1e1 a7c6 c1g5 f6d5 g5e7 c7e7 e2d2 f5g4 e4g3 e7f6 g3e4');
INSERT INTO evaluations VALUES(11325,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(11325,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(11325,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(11325,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(11325,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11325,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11325,36,'cp 12','f3e5 c8f5 f1e2 h7h5 e5c6 b7c6 e2d3 e7e6 d1e2 a6a5 e1g1 a5a4 c1d2 f8d6 h2h3 e8g8 e3e4 f6e4 c3e4 f5e4 d3e4 d5e4 e2e4 c6c5 d4c5 d6c5 e4e2 d8h4 d2c3 c5d4 c3d4 h4d4 b2b3 f8d8 e2h5 d4b2 h5e2 d8d5 a1d1 b2a2 d1d5 e6d5');
INSERT INTO evaluations VALUES(13278,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13278,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13278,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13278,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13278,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13278,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13278,36,'cp 0','g2g3 g7g6 f1g2 f8g7 g1e2 e8g8 e1g1 a7a6 h2h3 f6d7 a2a4 a8b8 a1b1 e7e6 c1f4 d8e7 f4e3 c6d4 d1d2 d4b3 d2d1');
INSERT INTO evaluations VALUES(14069,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(14069,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(14069,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(14069,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(14069,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(14069,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(14069,36,'cp 97','d2d4 a7a6 b5e2 e5d4 f3d4 c6d4 d1d4 e7c6 d4d3 c8e6 c3d5 f8e7 f2f4 d8d7 c1e3 f7f5 e2f3 e8c8 e1g1 d8e8 b2b4 f5e4 f3e4 e7f6 a1b1 e6d5 e4d5 c6e7 d5f3 d7f5 d3f5 e7f5 e3f2 c8b8 f1d1');
INSERT INTO evaluations VALUES(14064,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(14064,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(14064,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(14064,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(14064,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(14064,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(14064,36,'cp 76','a7a6 b5e2 d6d5 e4d5 e7d5 d4e5 d5c3 d1d8 c6d8 b2c3 d8c6 c1f4 f8c5 a1d1 c8e6 e1g1 a8d8 d1d8 c6d8 f1a1 e8g8 f3d4 f8e8 g1f1 e6d7 a1d1 c5b6 c3c4 d8e6 d4e6 d7e6 d1b1 e6c8 f2f3 h7h6 e2d3 g7g5 f4g3 b6d4 d3e4');
INSERT INTO evaluations VALUES(16618,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(16618,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(16618,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(16618,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(16618,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(16618,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(16618,36,'cp -16','d7d5 d2d3 h7h6 c2c4 d5c4 d3c4 f8d6 b1c3 e8g8 b2b3 e5e4 f3d4 c6d4 d1d4 f8e8 c1b2 d6e5 d4d8 e8d8 a1d1 c8f5 d1d8 a8d8 f1d1 d8d1 c3d1 e5b2 d1b2 g8f8 f2f3 e4f3 e2f3 f5b1 a2a3 b1c2 b3b4 c7c5 g2f1 f8e7 b4c5 f6d7 b2d3');
INSERT INTO evaluations VALUES(31547,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(31547,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31547,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31547,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31547,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31547,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(31547,36,'cp 5','c6c5 f1e2 b8c6 c2c3 c5d4 c3d4 g8e7 e1g1 a8c8 c1e3 f5g4 b1d2 e7f5 h2h3 f5e3 f2e3 g4h5 e2d3 f8e7 a1c1 e8g8 g2g4 h5g6 d3g6 h7g6 g1g2 a7a6 d2b3 c6a7 a3a4 d8b6 f1f2 c8c1 b3c1 f8c8 c1d3 c8c4');
INSERT INTO evaluations VALUES(31678,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31678,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31678,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(31678,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(31678,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(31678,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(31678,36,'cp 40','c1e3 c5d4 f3d4 g8e7 b1c3 a7a6 g2g4 f5g6 f2f4 h7h5 f4f5 e6f5 g4f5 e7f5 d4f5 g6f5 f1g2 d5d4 d1d4 d8d4 e3d4 b8c6 d4b6 f8e7 e1g1 f5e6 c3d5 e7d8 d5c7 d8c7 b6c7 h5h4 h2h3 h8h5 c2c3 e8f8 g1h2 f8e8 c7d6 a8d8 a1d1');
INSERT INTO evaluations VALUES(41037,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(41037,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(41037,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(41037,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(41037,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(41037,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41037,36,'cp 35','e1g1 c7c5 d4c5 d8d1 f1d1 e7c5 c4e2 e8g8 f3e5 b8d7 e5d3 b7b6 d3c5 d7c5 b2b4 c8a6 b4b5 a6b7 b1c3 f8c8 f2f3 a7a6 c1d2 f6d5 c3d5 e6d5 d2b4 a6b5 e2b5 b7a6 b5a6 c5a6 b4e1 a6c7 a2a4 f7f6 e1g3 c7e6');
INSERT INTO evaluations VALUES(41036,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41036,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41036,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(41036,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41036,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41036,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41036,36,'cp 28','c7c5 d4c5 d8d1 f1d1 e7c5 b1c3 e8g8 b2b3 b8c6 c1b2 b7b6 f3d2 c5e7 c4e2 f8d8 e2f3 c8b7 d2c4 a8b8 g1f1 d8d1 a1d1 c6b4 f3b7 b8b7 c4d6 b7d7 d6c8 d7d1 c3d1 e7f8 d1c3 b4c6 c3b5 a7a6 b5d4 c6d4 b2d4 f6d7 a2a4');
INSERT INTO evaluations VALUES(43642,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(43642,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(43642,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(43642,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43642,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(43642,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43642,36,'cp 34','c4c5 f6h5 f4d2 b8c6 g1f3 h5f6 a2a3 f6e4 e2e3 f8e7 f1e2 e8g8 e1g1 g8h8 a1b1 f7f5 d2e1 g7g5 h2h3 e7f6 f3h2 c8d7 b2b4 c6e7 a3a4 c7c6 b4b5 a6b5 a4b5 e6e5 b5c6 d7c6 d4e5 f6e5 c3b5 e4c5 h2f3 e5f6 b5d4 c5e4');
INSERT INTO evaluations VALUES(49599,30,'cp -273',NULL);
INSERT INTO evaluations VALUES(49599,31,'cp -264',NULL);
INSERT INTO evaluations VALUES(49599,32,'cp -267',NULL);
INSERT INTO evaluations VALUES(49599,33,'cp -268',NULL);
INSERT INTO evaluations VALUES(49599,34,'cp -267',NULL);
INSERT INTO evaluations VALUES(49599,35,'cp -271',NULL);
INSERT INTO evaluations VALUES(49599,36,'cp -297','d7d5 e4d5 c8g4 f3g2 c7c6 f1e2 g4f5 g1f3 f6d5 g2g3 b8d7 d2d4 e5d4 f3d4 f5g6 b1c3 d7f6 g3e5 d8e7 e5e7 f8e7 c3d5 f6d5 c2c3 e8c8 c1d2 h7h5 e1c1 e7d6 e2c4 d5f4 d2e3 h5h4 b2b4 f4g2 e3d2 h4h3 d1g1 g2h4');
INSERT INTO evaluations VALUES(48812,30,'cp -292',NULL);
INSERT INTO evaluations VALUES(48812,31,'cp -271',NULL);
INSERT INTO evaluations VALUES(48812,32,'cp -292',NULL);
INSERT INTO evaluations VALUES(48812,33,'cp -285',NULL);
INSERT INTO evaluations VALUES(48812,34,'cp -269',NULL);
INSERT INTO evaluations VALUES(48812,35,'cp -285',NULL);
INSERT INTO evaluations VALUES(48812,36,'cp -272','e4d5 c8g4 f3g2 c7c6 d5d6 b8d7 d2d3 f8d6 f1e2 g4e6 b1c3 d8b6 g1f3 h7h6 f3d2 e8c8 d2c4 e6c4 d3c4 e5e4 e1g1 d8e8 f1d1 b6c7 a1b1 d6f4 c1f4 c7f4 g2g3 f4g3 h2g3 h6h5 b2b4 h5h4 g3g4 c8c7 g4g5 f6h5 e2g4 d7e5 g4h5 h8h5');
INSERT INTO evaluations VALUES(52252,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(52252,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(52252,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(52252,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(52252,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(52252,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(52252,36,'cp 93','e2e4 f8g7 b1c3 d7d6 g1f3 g8f6 f1e2 e8g8 e1g1 e7e5 d4d5 b8a6 h2h3 c8d7 c1g5 h7h6 g5e3 a6c5 f3d2 a7a5 d1c2 f6e8 a2a3 c6d5 c4d5 a8c8 a3a4 h6h5 c2d1 f7f5 e4f5 g6f5 e2h5 e8f6 h5e2 f5f4 e3c5 c8c5');
INSERT INTO evaluations VALUES(52253,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(52253,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(52253,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(52253,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(52253,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(52253,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(52253,36,'cp 96','d7d5 e4e5 f8g7 b1c3 f7f6 g1f3 f6e5 f3e5 g8f6 f1e2 b8d7 e5f3 d5c4 e2c4 d7b6 c4b3 c8g4 e1g1 h8f8 f1e1 d8d7 a2a4 f6d5 c3e4 g4f3 g2f3 d5c7 a4a5 b6d5 e4c5 d7f5 c5b7 f5f3 d1f3 f8f3 b3d1 f3f8 d1a4 g7d4 a4c6 e8f7 e1e4 c7e6 e4d4');
INSERT INTO evaluations VALUES(53344,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53344,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(53344,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(53344,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53344,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(53344,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53344,36,'cp 2','c2c3 c8f5 d2d4 e7e6 f1b5 b8d7 g1e2 c5d4 c3d4 f8b4 b1c3 g8e7 e1g1 a7a6 b5d3 e8g8 e2g3 f5d3 d1d3 f7f6 a2a3 b4c3 e5f6 d7f6 b2c3 d8d7 c1g5 b7b5 a1e1 a8e8 d3d2 e7g6 g5f6 f8f6 f2f4 g6h4');
INSERT INTO evaluations VALUES(2031,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(2031,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(2031,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(2031,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(2031,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(2031,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(2031,36,'cp 120','c6b8 h2h3 d7d6 b1c3 b8d7 c1g5 g8f6 f3d2 e7d8 g5e3 f8e7 a2a4 h7h6 a4a5 e8g8 d2c4 a7a6 f1d3 f6h7 d1d2 e7g5 e3g5 h7g5 h3h4 g5h7 g2g4 d7c5 d3e2 b7b5 a5b6 c7b6 b2b4 c5b7 g4g5 b6b5 c4e3 h6g5 h4g5 d8g5');
INSERT INTO evaluations VALUES(10141,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(10141,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(10141,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(10141,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10141,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10141,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10141,36,'cp 0','d7d5 e4d5 f6d5 g1f3 c8e6 e1g1 f7f6 b1d2 c5b6 d2e4 e8g8 b2b4 a7a6 f1e1 d8e8 c1d2 e8g6 c4b3 g8h8 c2c4 d5f4 d2f4 e5f4 c4c5 e6h3 f3h4 g6h6 h4f3');
INSERT INTO evaluations VALUES(10982,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(10982,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(10982,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(10982,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(10982,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(10982,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(10982,36,'cp 100','d7d6 e1c1 e8g8 f1e2 c7c6 e4e5 d6e5 d4e5 c8d7 c3e4 f8e8 g5f6 e7f6 e4f6 d8f6 e5f6 g7f6 d1d7 e8e2 h1f1 b7b6 c1d1 e2e8 f1e1 a7a5 b2b3 e8e1 d1e1 a5a4 e1d2 a4b3 a2b3 a8e8 d7b7 b6b5 b7c7 e8e6 g2g3 f6f5 h2h4 g8g7 h4h5 f5f4 g3f4 e6h6 d2e3 h6h5 c7c6 b5b4');
INSERT INTO evaluations VALUES(25687,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25687,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25687,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25687,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25687,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25687,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25687,36,'cp 0','f7f6 h2h3 g4h5 c1e3 c6c5 c2c3 f8d6 b1d2 g8e7 d1b3 h5f7 b3b7 a8b8 b7a6 b8a8 a6b7 a8b8');
INSERT INTO evaluations VALUES(32139,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(32139,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(32139,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(32139,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(32139,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(32139,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(32139,36,'cp 31','c7c5 e1g1 f8e7 c2c4 g4d7 b1c3 b8c6 f3e1 g8f6 f2f4 a7a6 g1h1 a8b8 f4e5 d6e5 c3d5 b7b5 c1e3 f6d5 c4d5 c6d4 e2g4 e8g8 g4d7 d8d7 a1c1 a6a5 e1f3 f7f6 b2b3 g8h8 f3h4 g7g6 h4f3 e7d6 h2h3');
INSERT INTO evaluations VALUES(37429,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(37429,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(37429,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(37429,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(37429,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(37429,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(37429,36,'cp 0','g1f3 c7c5 c2c3 d7d5 b1d2 b8c6 h2h3 f6d7 e1g1 e7e5 d4e5 d7e5 f3e5 c6e5 d3e2 f8e8 d2f3 c8f5 a1c1 f5e4 f3g5 e4f5');
INSERT INTO evaluations VALUES(39674,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(39674,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(39674,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(39674,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(39674,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(39674,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(39674,36,'cp 13','c8e6 d1d2 e8g8 e1c1 b8d7 g2g4 b7b5 g4g5 f6h5 c1b1 d7b6 c3d5 b6d5 e4d5 e6f5 b3a5 f7f6 g5f6 e7f6 a5c6 d8d7 a2a3 d7f7 f1d3 f5d3 d2d3 g8h8 d3b3 a8e8 h1f1 f6d8 b1a2 f7d7 c2c4 b5c4 b3c4 d7b7 c4b4 b7b4');
INSERT INTO evaluations VALUES(43217,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(43217,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(43217,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(43217,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(43217,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(43217,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(43217,36,'cp -56','c7c5 c2c3 e8g8 g1f3 d7d6 e1g1 c5d4 e3d4 d8c7 b1d2 c8f5 d3f5 g6f5 d1e2 b8c6 d2c4 e7e6 c1d2 c6e7 a2a4 e7d5 f3e1 f8c8 c4e3 d5e3 d2e3 a7a5 e1f3 c7c6 f1c1 c8c7 g1h1 b7b6 c3c4 c6b7 b2b3 a8c8 d4d5 e6d5 e3d4');
INSERT INTO evaluations VALUES(46147,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(46147,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(46147,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(46147,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(46147,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(46147,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(46147,36,'cp 29','e2e4 d5e4 c3e4 c5d4 f1b5 c8d7 d1d4 f8e7 e1c1 f6e4 g5e7 d8e7 b5d7 b8d7 d4e4 d7c5 e4e3 e8g8 d1d4 b7b6 h1d1 a8d8 f3e5 d8d4 d1d4 f8c8 c1b1 f7f6 e5d3 e6e5 d3c5 c8c5 d4d2 g7g6 b2b3 g8g7 a2a4 h7h5 h2h3 e7c7 b1b2 c7c6');
INSERT INTO evaluations VALUES(46151,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(46151,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(46151,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(46151,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(46151,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(46151,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(46151,36,'cp 26','d5e4 c3e4 c5d4 f1b5 c8d7 d1d4 f8e7 e1c1 f6e4 b5d7 b8d7 g5e7 d8e7 d4e4 d7b6 d1d4 e8g8 h1d1 a8c8 c1b1 h7h6 f3e5 f8d8 d4d8 c8d8 d1d8 e7d8 e5d3 b6d5 g2g3 d8c7 a2a4 b7b6 e4e5 c7e5 d3e5 d5e7 c2c3 g8f8 b1c2 f8e8 c2d3 f7f6 e5c4 e8d7 a4a5 b6a5 c4a5');
INSERT INTO evaluations VALUES(48067,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(48067,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(48067,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(48067,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(48067,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(48067,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(48067,36,'cp 141','c8e6 b1c3 f8d6 d2d4 e8g8 f1e2 c7c6 c1g5 b8d7 c3e4 d6e7 e4g3 d8b6 b2b3 h7h6 g5e3 c6c5 e1g1 f6d5 d4c5 b6c7 e3d4 e7c5 d4c5 d7c5 f1e1 a8d8 d1d2 e6g4 f3d4 g4e2 d2e2 d5f4 e2c4 c5e6 c4c7 e6c7');
INSERT INTO evaluations VALUES(48423,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(48423,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(48423,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(48423,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(48423,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48423,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48423,36,'cp -25','c7c5 b1d2 f8d6 c2c3 e8g8 f2f4 b7b6 f1d3 c8b7 e1g1 b8c6 a2a4 c6e7 a4a5 b6a5 d2b3 c5d4 e3d4 a5a4 a1a4 f6e4 d3e4 d5e4 b3c5 d6c5 d4c5 f7f6 d1d8 f8d8 e5c4 d8d3');
INSERT INTO evaluations VALUES(51223,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(51223,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(51223,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(51223,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(51223,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(51223,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(51223,36,'cp -12','c8b7 d2d3 d7d6 e1g1 g8f6 c2c3 b8d7 f1e1 g7g6 c1f4 f6h5 f4g5 d8c7 a3a4 f8g7 b1a3 h7h6 g5e3 b5a4 d1a4 e8g8 a4c2 b7c6 d3d4 d7f6 e3d2 c5d4 c3d4 f8c8 e4e5 d6e5 f3e5 f6d5 e5c6 c7c6 c2c6 c8c6 a3c4 g7d4 d2h6 d4g7 c4a5');
INSERT INTO evaluations VALUES(30341,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(30341,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(30341,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(30341,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(30341,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(30341,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(30341,36,'cp -12','d2d3 d7d6 e1g1 g8f6 b1c3 h7h6 f1e1 b8d7 d3d4 c5d4 f3d4 d7e5 f2f4 e5c4 a3a4 f8e7 a4b5 a6b5 d4b5 e8g8 d1e2 d6d5 e4e5 f6d7 g1h1 e7b4 e1d1 d8h4 a1b1 f8c8 a2c4 c8c4 c1e3 b7a6 b1a1 b4c3 b2c3 d7c5 b5d6');
INSERT INTO evaluations VALUES(30339,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(30339,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(30339,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30339,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(30339,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(30339,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(30339,36,'cp -21','d7d6 e1g1 g8f6 h2h3 b8d7 f1e1 f8e7 b1d2 e8g8 c2c3 a6a5 a3a4 b5a4 d1a4 d7b6 a4d1 d8c7 a2b1 a5a4 d2c4 b6c4 d3c4 b7c6 e4e5 c6f3 d1f3 d6e5 f3e2 e5e4 b1c2 e7d6 a1a4 a8a4 c2a4 f8a8 a4c2 a8a1 c2e4 f6e4 e2e4');
INSERT INTO evaluations VALUES(30346,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(30346,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(30346,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(30346,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(30346,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(30346,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(30346,36,'cp -7','e1g1 g8f6 h2h3 b8d7 f1e1 f8e7 c2c3 d8c7 b1d2 e8g8 a2b1 f8e8 b1c2 a8c8 d2f1 e7f8 c2b1 c5c4 d3d4 e6e5 d4d5 c8a8 b1c2 a6a5 f1g3 b7c8 c1g5 h7h6 g5e3 d7c5 f3h2 g7g6 f2f4 e5f4 e3f4 c8d7 h2f3 c7b6');
INSERT INTO evaluations VALUES(51539,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(51539,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(51539,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(51539,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(51539,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(51539,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(51539,36,'cp 73','c8b7 f1b5 b7e4 e1g1 c7c6 b5a4 d8a5 b1c3 e4g6 a2a3 g8f6 b2b4 a5b6 f3e5 e7e6 d2d4 f8e7 f1e1 e8g8 c1f4 f8c8 a4b3 b6d8 c3a4 d7d5 h2h4 a7a5 c2c3 a5b4 a3b4 f6h5 f4e3 h5f6 a4c5 a8a1 d1a1 b8d7 e3f4 d7e5 f4e5');
INSERT INTO evaluations VALUES(30850,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(30850,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(30850,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(30850,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(30850,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(30850,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(30850,36,'cp 77','d2d4 b7e4 f1b5 c7c6 b5a4 d8a5 b1c3 g8f6 c1d2 a5a6 b2b3 e7e6 c3e4 f6e4 c2c4 f8e7 e1g1 e4d2 d1d2 a6b7 c4c5 e8g8 f3e5 f8d8 e5c4 d7d6 c5d6 e7d6 a1c1 b7c7 c4d6 d8d6 c1c4 c7d8 f1c1 a7a5 g2g3 g7g6 d2e3 d6d5 c4c6 b8c6 a4c6 a8c8 c6d5 c8c1 e3c1 d8d5 c1c8 g8g7 c8c3 g7g8');
INSERT INTO evaluations VALUES(56731,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(56731,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(56731,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(56731,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(56731,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(56731,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(56731,36,'cp 127','c7c5 f1c4 c8f5 e1g1 b8c6 f1e1 f8e7 c3d5 g8f6 c2c3 f6d5 c4d5 d8d7 d2d4 e8c8 h2h3 e7f6 f3g5 c8b8 g5f7 c5d4 f7h8 d8h8 d5c6 b7c6 c3d4 e5d4 c1f4 g7g5 f4g3 h7h5 d1f3 g5g4 f3a3 g4h3 a3d6 d7d6 g3d6 b8a8 d6e5 f6e5 e1e5 f5c2 e5c5 c2e4 f2f3 e4d5');
INSERT INTO evaluations VALUES(56795,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(56795,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(56795,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(56795,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(56795,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(56795,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(56795,36,'cp 89','c7c5 c2c4 c5d4 e3d4 f8d6 f4d6 d8d6 c4c5 d6c7 f1d3 g8e7 e1g1 b8c6 b1c3 e8g8 f1e1 b7b6 c5b6 a7b6 c3b5 c7b8 d1e2 c8d7 a2a3 e7g6 g2g3 g8h8 h2h4 e6e5 h4h5 g6f4 g3f4 e5e4 d3c2 b8f4 f3h2 f4h4 g1h1 f6f5 f2f3 h4h5 f3e4 h5e2 e1e2 d5e4');
INSERT INTO evaluations VALUES(1085,30,'cp 261',NULL);
INSERT INTO evaluations VALUES(1085,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(1085,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(1085,33,'cp 262',NULL);
INSERT INTO evaluations VALUES(1085,34,'cp 265',NULL);
INSERT INTO evaluations VALUES(1085,35,'cp 227',NULL);
INSERT INTO evaluations VALUES(1085,36,'cp 234','e1g1 f6g4 f1e1 e7c5 b2b3 b7b5 c1a3 b5b4 a3b2 f8e7 h2h3 g4h6 a2a3 c8b7 a3b4 c6b4 f3d4 e8g8 c2c3 b4c6 b1d2 a7a5 d4f3 d7d5 e5d6 c7d6 b2a3 c5b6 c4d5 b6c7 d1e2 a8e8 d2c4 g8h8 c4e3 h6g8 e3f5 c7d7 f5e7 e8e7 e2d2 g8f6');
INSERT INTO evaluations VALUES(1084,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(1084,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(1084,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(1084,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(1084,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(1084,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(1084,36,'cp 216','f6g4 f1e1 e7c5 b2b3 b7b5 c4d3 f8e7 d1e2 c8b7 h2h3 g4h6 b1d2 c6b4 d2e4 c5b6 c1h6 b4d3 e2d3 b6h6 f3d4 e8g8 d4f5 h6e6 e4d6 f8b8 f5e7 e6e7 d6b5 e7g5 f2f3 b7a6 c2c4 a6b5 c4b5 b8e8 d3d7 a8d8 d7g4 e8e5 g4g5 e5g5 a1d1 d8d1 e1d1 g8f8 a2a4 f8e7 b3b4 g5g6 g1h2 g6d6 d1d6');
INSERT INTO evaluations VALUES(1116,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(1116,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(1116,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(1116,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(1116,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(1116,35,'cp 242',NULL);
INSERT INTO evaluations VALUES(1116,36,'cp 266','f1e1 e7c5 b2b3 b7b5 c1a3 b5b4 a3b2 d4d3 d1d2 d3c2 d2c2 c8a6 h2h3 g4f2 c2f2 c5f2 g1f2 a6c4 b3c4 h7h5 b1d2 h8h6 f2f1 a7a5 a2a4 b4a3 b2a3 c6b4 f3d4 a5a4 e1e2 b4d3 d2e4 f8a3 a1a3 d3e5 e4c3 f7f6 d4f5 h6g6 c3a4 e8d8');
INSERT INTO evaluations VALUES(6347,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(6347,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(6347,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(6347,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(6347,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6347,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(6347,36,'cp 24','g8f6 d1e2 e8g8 c1g5 c6d4 e2d2 d4b3 a2b3 h7h6 g5e3 b6e3 d2e3 f8e8 e1c1 c7c6 e3d4 d8c7 f2f3 a7a5 d4d6 c7d6 d1d6 g8f8 f1e2 b7b5 h1d1 e8e7 h2h4 f6e8 d6d4 e8c7 d4d2 c7e6 g2g3 c8b7 d2d6 e7e8 h4h5');
INSERT INTO evaluations VALUES(9943,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(9943,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(9943,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(9943,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(9943,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(9943,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(9943,36,'cp 40','d1e2 e8g8 c1g5 c6d4 e2d2 d4b3 a2b3 h7h6 g5e3 b6e3 d2e3 f8e8 e1c1 c7c6 e3d4 d8a5 c1b1 d7d5 e4d5 f6d5 c3d5 a5d5 d4d5 c6d5 b1c1 e8e5 f2f4 e5e4 g2g3 c8g4 f1d3 e4e3 c1d2 a8e8 d1a1 a7a6 h1e1 e3e1 a1e1 e8e1 d2e1 g8f8 b3b4 f8e7 c2c3 e7d6 e1d2 d6e7 d2e3 e7d6 e3d4 b7b6 b4b5 a6b5 b2b3');
INSERT INTO evaluations VALUES(6796,30,'cp 344',NULL);
INSERT INTO evaluations VALUES(6796,31,'cp 352',NULL);
INSERT INTO evaluations VALUES(6796,32,'cp 324',NULL);
INSERT INTO evaluations VALUES(6796,33,'cp 336',NULL);
INSERT INTO evaluations VALUES(6796,34,'cp 345',NULL);
INSERT INTO evaluations VALUES(6796,35,'cp 382',NULL);
INSERT INTO evaluations VALUES(6796,36,'cp 344','f3d2 c7c6 c2c3 d4b5 a2a4 b5c7 d2c4 d8f6 f1e2 c6d5 e4d5 f6e7 c4d6 e7d6 c3c4 g8f6 e1g1 e8g8 b1c3 c7e8 a4a5 d6e7 c1e3 d7d6 b2b4 c8f5 d1d2 e8c7 f2f3 f5g6 c4c5 f8d8 a1d1 h7h6 c5d6 e7d6 e3c5 d6d7 d5d6 a8c8 c5a7 c7e8 a7b6 d7d6 b6d8 d6d2 d1d2 c8c3');
INSERT INTO evaluations VALUES(16905,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(16905,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(16905,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(16905,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(16905,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(16905,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(16905,36,'cp 135','c7c6 e4e5 c6b5 c1g5 f8e7 e5f6 e7f6 f1e1 e8f8 g5f6 d8f6 d1h5 b7b6 h5b5 c8b7 b5d7 b7c6 d7c7 g7g6 f2f3 f8g7 c7e5 h7h5 g1f2 h5h4 b1a3 g6g5 a1c1 a8c8 c2c4 d4c3 b2c3 h8e8 e5f6 g7f6 e1e8 c6e8 c3c4 e8d7 a3b1');
INSERT INTO evaluations VALUES(12620,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(12620,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(12620,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(12620,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(12620,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(12620,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(12620,36,'cp 140','e4e5 c6b5 c1g5 f8e7 e5f6 e7f6 f1e1 e8f8 g5f6 d8f6 d1h5 b7b6 h5b5 c8b7 b5d7 b7c6 d7c7 g7g6 f2f3 f8g7 c7e5 h7h5 b1d2 a8c8 d2c4 h8e8 e5f6 g7f6 e1e8 c8e8 g1f2 e8c8 c4d6 c8d8 d6e4 c6e4 d3e4 d8c8 a1c1 f6e5 f2e2 h5h4 e2d3 f7f6 a2a3 g6g5 a3a4 c8c6');
INSERT INTO evaluations VALUES(18184,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(18184,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(18184,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(18184,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(18184,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(18184,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18184,36,'cp 47','d7d6 e4e5 f7f6 e5d6 e7d6 e1g1 e8f7 c2c4 g8e7 b1c3 e7f5 d3d4 d6d5 d4c5 f8c5 f1e1 a8b8 c4d5 c6d5 b2b3 c5b4 c1d2 c8b7 c3e2 d5d4 f3d4 d8d5 d4f3 b4d6 e2f4 d6f4 d2f4 d5d1 a1d1 b8d8 f3d2 h8e8 d2c4 e8e1 d1e1 b7c6 f2f3 c6b5 e1c1 b5c4 b3c4 g7g5 f4c7 d8d2');
INSERT INTO evaluations VALUES(18214,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(18214,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(18214,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(18214,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(18214,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(18214,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(18214,36,'cp 48','e4e5 f7f6 e5d6 e7d6 e1g1 e8f7 c2c4 g8e7 f1e1 e7f5 d3d4 d6d5 d4c5 f8c5 b1c3 a8b8 c4d5 c6d5 b2b3 c5b4 d1d3 b4c3 d3c3 c8b7 c3d3 d8d7 c1b2 d5d4 f3d4 f5d4 d3d4 d7d4 b2d4 b8c8 d4e3 h7h5 b3b4 h8e8 e1d1 c8c4 e3c5 b7c6 f2f3');
INSERT INTO evaluations VALUES(22874,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(22874,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(22874,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(22874,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(22874,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(22874,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(22874,36,'cp -69','c2c4 g8f6 b1c3 g7g6 d2d3 f8g7 c1d2 e8g8 h2h4 a7a5 a2a3 d7d6 f3d1 f8e8 d1c2 h7h6 f1e2 c6b8 g1f3 c7c6 e1g1 b8a6 a1d1 a6c7 f1e1 c8g4 c3e4 c7e6 e4g3 h6h5 d2c3 b7b5 c4b5');
INSERT INTO evaluations VALUES(32865,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(32865,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(32865,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(32865,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(32865,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(32865,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(32865,36,'cp 61','g4e2 c3e2 b8c6 g1f3 e5f5 e1g1 e8c8 b2b4 e7e6 c2c3 c6e5 f3e5 f5e5 d2d4 e5b5 f1e1 b5c4 c1d2 f8d6 e2c1 g8f6 d1f3 h8g8 c1b3 c7c6 a1b1 d6c7 a2a4 g8e8 g2g3 e6e5 b3c5 c4d5 f3d5');
INSERT INTO evaluations VALUES(32872,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(32872,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(32872,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(32872,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(32872,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(32872,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(32872,36,'cp 66','c3e2 e5f5 g1f3 b8c6 e1g1 e8c8 f1e1 g8f6 d2d4 e7e5 c2c3 e5d4 e2d4 c6d4 f3d4 f5d5 d1c2 f8d6 d4f5 g7g6 f5d6 d8d6 c3c4 d5d3 c2a4 a7a6 c1f4 d6c6 f4e5 d3f5 a1d1 h8e8 g2g4 f5f3 e1e3 f3e3 f2e3 e8e5 a4a3 f6d7 b2b4');
INSERT INTO evaluations VALUES(36600,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(36600,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(36600,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(36600,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36600,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36600,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36600,36,'cp 0','g1f3 b8c6 f1e2 f6d7 d4d5 c6e5 f3e5 d7e5 e1g1 e7e6 d5e6 c8e6 e3e4 d6d5 f1e1 f8e8 d1c2 d5e4 d2e4 d8d5 e4g5 e6g4 e2g4 e5g4 g5f3 e8e1 a1e1 d5a2 f4c7 g4f6 c7e5 a7a5 c2d2 a2c4 d2d6 c4d5 d6d5 f6d5 g1f1');
INSERT INTO evaluations VALUES(44477,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44477,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44477,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(44477,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(44477,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(44477,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44477,36,'cp 3','d8b6 b2b3 g7g6 h2h3 f8g7 g1f3 d7d6 f1d3 c6c5 e1g1 e8g8 c2c3 b8c6 f1e1 c8e6 d3f1 a8d8 f4h2 f6e4 d1c2 e6d5 b1d2 e4d2 f3d2 c5d4 e3d4 e7e5 d4e5 d6e5 a1d1 b6c7 a2a4 d5e6 f1c4 c7e7 c4e6 e7e6 g1h1 f7f5 f2f3 d8d7 d2c4 f8d8 d1d7 d8d7 c2e2 f5f4 a4a5');
INSERT INTO evaluations VALUES(50964,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(50964,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(50964,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50964,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50964,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50964,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(50964,36,'cp 25','f1d3 c7c5 b2b3 g8f6 b1d2 c5d4 e3d4 f8d6 c1b2 e8g8 e1g1 b8c6 a2a3 b7b6 d1e2 c8b7 f3e5 c6e7 c2c3 e7f5 a3a4 d8c7 g2g3 d6e5 d4e5 f6e4 d2e4 d5e4 d3e4 c7e5 f1e1 b7e4 c3c4 f5d4 b2d4 e5d4 e2e4 d4e4 e1e4 a8c8 e4e1 c8c7');
INSERT INTO evaluations VALUES(50963,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(50963,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(50963,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(50963,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(50963,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50963,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(50963,36,'cp 9','c7c5 b2b3 g8f6 e1g1 b8c6 c1b2 c5d4 e3d4 f8d6 b1d2 e8g8 c2c4 b7b6 f3e5 c6b4 d3b1 d8c7 f1e1 c8b7 a2a3 b4c6 b1d3 f8d8 d2f3 d5c4 b3c4 d6e5 f3e5 c6e5 e1e5 c7c6 d4d5 e6d5 c4d5 d8d5 e5d5 c6d5');
INSERT INTO evaluations VALUES(10497,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10497,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10497,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10497,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10497,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10497,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10497,36,'cp 0','e4e5 d7d5 e5f6 d5c4 d1d8 c6d8 f6g7 h8g8 c1h6 b4c3 b2c3 d8e6 e1c1 f7f6 h1e1 e8f7 e1e4 b7b5 d1d5 e6g7 h6g7 g8g7 e4f4 g7g2 d5c5 c8d7 c5c7 f7e8 c7b7 g2f2 f4e4 e8f8 b7d7 f2f3 d7h7 f8g8 e4e7 f3c3 c1b1 c3d3 e7g7 g8f8 g7f7 f8g8');
INSERT INTO evaluations VALUES(10330,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10330,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10330,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10330,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10330,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10330,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10330,36,'cp 0','d7d5 e5f6 d5c4 d1d8 c6d8 f6g7 h8g8 c1h6 b4c3 b2c3 f7f6 e1c1 d8e6 h1e1 e8f7 e1e4 b7b5 d1d5 e6g7 h6g7 g8g7 e4f4 g7g2 d5c5 c8d7 c5c7 f7e7 f4d4 a8d8 d4d5 g2f2 f3d4 e7e8 d5h5 e8f8 h5h7 d8e8 h7h8 f8e7 h8h7');
INSERT INTO evaluations VALUES(10598,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(10598,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(10598,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(10598,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(10598,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(10598,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(10598,36,'cp 49','e8g8 f1e2 c6e5 f3e5 g4e5 f2f4 e5g6 e1g1 f8e8 g1h1 b7b6 f4f5 g6e5 f5f6 c8b7 f6g7 d8g5 e3e4 g5g7 c3d5 b7d5 c4d5 c5d6 g2g3 e5g6 e2d3 g6e5 d3c2 g8h8 d1h5 e8g8 f1f5 f7f6 c1d2 e5g4 a1f1 d6e5 d5d6 c7d6');
INSERT INTO evaluations VALUES(4490,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(4490,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(4490,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(4490,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(4490,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(4490,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(4490,36,'cp 58','a2a3 f8e8 b2b4 c5f8 f1e2 c6e5 f3e5 g4e5 f2f4 e5g6 e1g1 c7c6 e3e4 a7a5 b4b5 c6b5 c1e3 b7b6 d1d5 a8b8 d5b5 c8b7 e2d3 d8c8 g1h1 f8c5 e3c5 b7c6 b5b2 b6c5 b2d2 c8d8 d3c2 d8h4 h1g1 b8b2 a1b1 b2b1');
INSERT INTO evaluations VALUES(15741,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(15741,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(15741,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(15741,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(15741,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(15741,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(15741,36,'cp -69','d8e7 e1g1 e7e5 d4d2 e5e7 f1e1 c8e6 b1c3 c7c6 c4e6 f7e6 d2d3 e7d7 c1g5 f8e7 a1d1 d6d5 d3e2 h8f8 e2e6 d7e6 e1e6 e8d7 e6e2 b7b5 g5e3 a7a5 e3d4 a5a4 b2b3 e7d6 g2g3 h7h5 h2h4 g7g6 d1e1 a8e8 g1g2 e8e2 c3e2 f6g4 f2f3');
INSERT INTO evaluations VALUES(18914,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(18914,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(18914,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(18914,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(18914,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(18914,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(18914,36,'cp 45','d7d5 e4e5 g8e7 g1f3 e7f5 f1b5 c8d7 e1g1 d8b6 b5a4 a8c8 c3e2 b6a6 a4c2 c6b4 c2f5 d7b5 f1e1 e6f5 e2c3 b5d7 a2a3 d7e6 e1f1 b4d3 d1e2 a6c4 f1d1 d3c1 d1c1 c4e2 c3e2 c8c1 a1c1 e8d7 e2f4 f8e7 f4h5 g7g5 h5f6 e7f6 e5f6 g5g4 f3e5 d7d6 f2f3 h7h5 g1f2 f5f4');
INSERT INTO evaluations VALUES(20393,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(20393,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(20393,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20393,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(20393,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(20393,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(20393,36,'cp 56','e4e5 g8e7 g1f3 e7f5 f1b5 c8d7 e1g1 d8b6 b5a4 f8e7 c3e2 h7h6 a1b1 b6a6 e2c3 a6b6 g2g4 f5h4 f3h4 e7h4 c1e3 h4e7 b2b4 a8c8 a2a3 c6b8 d1d2 d7a4 c3a4 b6c7 f1c1 c7d8 a4c5 b7b6 c5b7 c8c1 b1c1 d8d7 b7d6 e7d6 e5d6 e8g8 c1c7 d7d6 e3f4 d6d8 h2h3 a7a6');
INSERT INTO evaluations VALUES(27279,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(27279,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(27279,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(27279,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(27279,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(27279,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(27279,36,'cp -13','d7c6 g1f3 f7f6 a2a4 e7e6 b2b3 h7h5 c1a3 f8a3 b1a3 d8d7 d1e2 g8h6 c2c4 g7g5 c4d5 e6d5 e1g1 h5h4 f3e1 h4h3 g2g3 e8g8 a3b5 f8f7 b5c3 a7a5 e1d3 b7b6 f2f4 g5g4 f4f5 c6b7 f1f4 b7a6 e2d2 a6d3 d2d3 c7c6 a1c1');
INSERT INTO evaluations VALUES(26484,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(26484,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(26484,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(26484,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(26484,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(26484,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(26484,36,'cp -13','g1f3 e7e6 f3e5 c6b5 c2c4 b5c4 e5c4 d5c4 d1a4 c7c6 a4c4 d8d5 c4f1 f8b4 b1c3 g8f6 c1d2 b4c3 d2c3 f6e4 f2f3 e4c3 b2c3 e8g8 f1d3 c6c5 e1g1 f8d8 d3b5 a8c8 a2a4 b7b6 b5a6 c5d4 c3d4');
INSERT INTO evaluations VALUES(26485,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(26485,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(26485,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(26485,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(26485,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(26485,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(26485,36,'cp -26','e7e6 f3e5 c6b5 c2c4 b5c4 d1a4 c7c6 e5c4 d5c4 a4c4 d8d5 c4f1 e6e5 b1c3 f8b4 c1d2 b4c3 b2c3 e5d4 c3d4 g8e7 f2f3 e8g8 f1f2 f8d8 e1g1 c6c5 d2c3 e7f5 f1d1 d5c4 c3b2 a8c8 a1c1 c4a4 f2e2 c5d4 c1c8 d8c8 b2d4 f5d4');
INSERT INTO evaluations VALUES(26759,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(26759,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(26759,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(26759,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(26759,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(26759,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(26759,36,'cp -13','f3e5 c6b5 c2c4 b5c4 d1a4 c7c6 e5c4 d5c4 a4c4 d8d5 c4d5 e6d5 c1d2 f8d6 f2f3 g8e7 b1c3 f7f6 a2a4 e8f7 e1e2 h7h5 g2g3 g7g5 a4a5 b7b5 a5b6 a7b6 e2d3 f7e6 e3e4 h5h4 a1a8 h8a8 g3h4 g5h4 c3e2 c6c5 d2c3 c5c4 d3c2 a8a2');
INSERT INTO evaluations VALUES(28223,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(28223,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(28223,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(28223,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(28223,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(28223,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(28223,36,'cp 53','c1d2 d5c3 b2c3 b4e7 e1g1 d7d6 d2f4 b8d7 f1e1 h7h6 a2a4 a7a5 d1e2 d6e5 f3e5 d7e5 f4e5 e8g8 e2g4 e7f6 h2h4 f6e5 e1e5 d8e7 a1e1 a8d8 f2f4 e7f6 h4h5 b7d5 f4f5 e6f5 d3f5 f6c6 e1e3 g8h8 e3g3 f8g8');
INSERT INTO evaluations VALUES(28224,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(28224,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(28224,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(28224,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(28224,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(28224,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(28224,36,'cp 62','d5c3 b2c3 b4e7 e1g1 d7d6 d2f4 b8d7 f1e1 e8g8 a2a4 d6e5 f3e5 d7e5 f4e5 g7g6 a4a5 c7c5 d4c5 e7c5 d1g4 c5d6 a5a6 b7c6 h2h4 d6e5 e1e5 d8f6 a1e1 a8d8 c3c4 h7h6 h4h5 g6g5 c4c5 b6b5 g4e2 f6f4 g2g3 f4a4 c2c3 a4a6');
INSERT INTO evaluations VALUES(28230,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(28230,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(28230,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(28230,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(28230,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(28230,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(28230,36,'cp 62','b2c3 b4e7 e1g1 d7d6 d2f4 b8d7 f1e1 e8g8 a2a4 a7a5 e1e3 d6e5 f3e5 d7e5 f4e5 g7g6 d1g4 a8c8 a1d1 e7f6 h2h4 h7h5 g4f4 g8g7 d1b1 f6e5 f4e5 d8f6 e5g3 b7d5 e3e5 f8h8 b1e1 c8d8 c3c4 d5c6 c2c3 c6a4 c4c5');
INSERT INTO evaluations VALUES(28229,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(28229,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(28229,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(28229,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(28229,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(28229,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(28229,36,'cp 35','b4e7 e1g1 d7d6 f1e1 b8d7 d2f4 e8g8 a2a4 d6e5 f3e5 d7e5 f4e5 g7g6 a4a5 c7c5 d1g4 c5d4 e5d4 e7f6 h2h4 f6d4 c3d4 h7h5 g4f4 d8d5 f2f3 b6a5 f4f6 a8c8 c2c4 d5d6 a1a5 c8d8 d3g6 d6d4 f6d4 d8d4 g6h5 d4c4 h5g4 f7f5 g4h3 b7d5 h4h5 c4c2 a5a7');
INSERT INTO evaluations VALUES(29579,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29579,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29579,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(29579,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29579,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(29579,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29579,36,'cp 6','c2c4 e7e6 g1f3 b8d7 b1c3 f8b4 c1d2 e8g8 f1e2 b4e7 e1g1 c7c6 f3h4 f5e4 c4d5 e6d5 d1b3 d7b6 f2f3 e4g6 d2e1 f8e8 h4g6 h7g6 e1f2 e7d6 e2d3 f6h5 c3e2 d8g5 e3e4 h5g3 e2g3 d6g3 e4e5 g5h4 b3c2 a8d8 a1d1');
INSERT INTO evaluations VALUES(45559,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45559,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(45559,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(45559,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45559,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45559,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(45559,36,'cp 41','a2a4 f8e7 d3d4 b8d7 e1g1 e8g8 c1e3 e5d4 e3d4 f8e8 a4a5 d8c7 h2h3 d7e5 c4a2 e5f3 d1f3 c8e6 f1d1 e6a2 a1a2 f6d7 f3g3 e7f8 f2f4 b7b5 a5b6 a7b6 a2a8 e8a8 c3e2 b6b5 g3f3 b5b4 e2g3 c6c5 d4f2 c7c6 g3f5');
INSERT INTO evaluations VALUES(50265,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50265,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(50265,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(50265,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(50265,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50265,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(50265,36,'cp 21','d2d4 f8g7 c2c4 e8g8 f1e2 c7c5 d4d5 d7d6 b1c3 e7e6 e1g1 e6d5 c4d5 c8g4 h2h3 g4f3 e2f3 a7a6 a2a4 b8d7 f3e2 d8e7 d1c2 f8e8 c1d2 a8b8 c2b3 e8c8 f1c1 f6e4 c3e4 e7e4 d2e1 d7e5 b3d1 c5c4 e2f1 c8c5');
INSERT INTO evaluations VALUES(56871,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(56871,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(56871,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(56871,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(56871,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(56871,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(56871,36,'cp 105','b1c3 d7d6 e2e4 f8g7 f1d3 a7a6 e4e5 h7h6 e1g1 g8e7 f1e1 b8c6 h2h3 e8g8 c3e4 e7f5 c2c3 d6e5 d4e5 c6e5 f3e5 g7e5 g2g4 f5d6 c1h6 d6e4 d3e4 e5g7 d1d8 f8d8 h6g5 f7f6 g5e3 f6f5 e3g5 d8e8 e4g2 e6e5 a2a3 f5g4 h3g4 c8g4 g2b7');
INSERT INTO evaluations VALUES(4792,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(4792,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(4792,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(4792,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(4792,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(4792,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(4792,36,'cp 155','e5c4 g8e7 e1g1 e8g8 d2d3 f7f5 c4d6 c7d6 b1c3 d6d5 c1g5 d5e4 d3e4 d8c7 f2f3 f5e4 c3e4 e7f5 d1e2 c7b6 f1f2 h7h6 e2c4 g8h8 g5c1 f8d8 c1f4 b6b2 a1e1 b2d4 c4d4 f5d4 c2c3 d4e6 f4d6 b7b5 a2a3 c6c5');
INSERT INTO evaluations VALUES(6150,30,'cp -393',NULL);
INSERT INTO evaluations VALUES(6150,31,'cp -393',NULL);
INSERT INTO evaluations VALUES(6150,32,'cp -398',NULL);
INSERT INTO evaluations VALUES(6150,33,'cp -409',NULL);
INSERT INTO evaluations VALUES(6150,34,'cp -397',NULL);
INSERT INTO evaluations VALUES(6150,35,'cp -413',NULL);
INSERT INTO evaluations VALUES(6150,36,'cp -412','e8f7 e1g1 g8f6 c2c3 c5b6 d2d4 d7d6 d1b3 f7f8 a2a4 e5d4 c3d4 d8e7 d4d5 c6a5 b3d1 e7e4 b1c3 e4b4 c1e3 c8d7 e3b6 c7b6 c3b5 d7b5 a4b5 b4b5 f1e1 b5d5 d1a4 d5c4 a4a3 a5c6 a3d6 f8g8 d6g3 h7h6 f3d2 c4f7 h2h3 g8h7');
INSERT INTO evaluations VALUES(4658,30,'cp -409',NULL);
INSERT INTO evaluations VALUES(4658,31,'cp -386',NULL);
INSERT INTO evaluations VALUES(4658,32,'cp -409',NULL);
INSERT INTO evaluations VALUES(4658,33,'cp -405',NULL);
INSERT INTO evaluations VALUES(4658,34,'cp -390',NULL);
INSERT INTO evaluations VALUES(4658,35,'cp -382',NULL);
INSERT INTO evaluations VALUES(4658,36,'cp -413','e1g1 g8f6 c2c3 c5b6 d2d4 d7d6 d1b3 f7f8 a2a4 e5d4 c3d4 c6d4 f3d4 b6d4 h2h3 d8e7 f1e1 c8e6 b3d1 d4c5 b1d2 f8g8 d2f3 h7h6 c1d2 a7a6 a1c1 c7c6 b2b4 c5b6 b4b5 c6b5 a4b5 a6b5 e4e5');
INSERT INTO evaluations VALUES(18741,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(18741,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(18741,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(18741,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(18741,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(18741,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(18741,36,'cp -94','d7d5 c4b3 d5e4 f3e4 g8f6 e4h4 f8e7 g1f3 a7a6 e1g1 b7b5 b1c3 e8g8 f1e1 c8b7 h4g3 e7d6 g3h3 c6d4 f3d4 c5d4 c3e2 a8c8 h3h4 c8c5 c1g5 d6e7 e2g3 h7h6 g5f6 e7f6 h4h3 b5b4 g3e4 c5f5 e4g3 f5a5 a2a4 b4a3');
INSERT INTO evaluations VALUES(19170,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(19170,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(19170,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(19170,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(19170,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(19170,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(19170,36,'cp 139','f8d6 b1c3 d6e5 c1d2 g8f6 f2f4 e5d4 e4e5 f6g4 c3b5 e8g8 b5d4 c5d4 a4d4 d7d6 e5e6 f7e6 f1e2 e6e5 f4e5 d6e5 e2c4 g8h8 d4d8 f8d8 e1g1 c8d7 d2g5 d8f8 h2h3 g4f6 a1e1 a8e8 b2b3 d7e6 c4b5 e6d7 b5d3 e5e4 d3e2 f6d5 c2c4 d5c3 f1f8 e8f8 g5e3 c3e2 e1e2 h7h6 e3a7');
INSERT INTO evaluations VALUES(21943,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(21943,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(21943,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21943,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(21943,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21943,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(21943,36,'cp 43','d4d5 c6e7 c2c4 f7f5 b1c3 g8f6 a2a4 e7g6 a4a5 f8e7 b2b4 a7a6 f1e2 e8g8 e1g1 d8e8 d1c2 f6e4 c1b2 c8d7 c4c5 e4c3 c2c3 e7f6 c3b3 d7b5 e2b5 a6b5 f3d2 e8f7 f2f4 g8h8 h2h3 f8e8 c5c6 b7c6 d5c6');
INSERT INTO evaluations VALUES(29977,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(29977,31,'cp 241',NULL);
INSERT INTO evaluations VALUES(29977,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(29977,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(29977,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(29977,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(29977,36,'cp 222','h2h3 g4f3 d1f3 d8e7 b1c3 d6d5 f3f5 e7e6 f5e6 f7e6 e4e5 f6g8 d2d4 c7c5 c3e2 e8f7 c2c3 b8c6 c1e3 g8e7 e1c1 h7h5 h3h4 e7f5 g2g3 g7g6 f1h3 f8e7 e2f4 b7b6 h3f1 c5d4 c3d4 g6g5 h4g5 e7g5 f4d5 a8d8 e3g5 d8d5 f1g2 d5d4 g2c6 d4c4 c1b1 c4c6');
INSERT INTO evaluations VALUES(29976,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(29976,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(29976,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(29976,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(29976,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(29976,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(29976,36,'cp 210','g4f3 d1f3 d8e7 b1c3 d6d5 f3f5 e7e6 f5e6 f7e6 d2d3 b8c6 f1e2 e8c8 e1g1 f8c5 e4d5 e6d5 e2d1 h7h5 c3e2 d8e8 a2a3 a7a5 c1d2 h5h4 a1b1 c5a7 b2b4 a7b6 d2c3 a5b4 a3b4 f6h5 b4b5 c6e5 e2d4 h5f4 d4f5 g7g5 c3d2');
INSERT INTO evaluations VALUES(33712,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(33712,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(33712,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(33712,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(33712,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(33712,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(33712,36,'cp -20','b8c6 f2f3 c8d7 c1e3 e8c8 d1e1 f6e8 g1e2 e7e5 a1d1 e8d6 e2g3 g7g6 e3g5 f8e7 g5e7 c6e7 b2b3 h7h5 h2h4 d7e6 c3b5 d6b5 f1b5 c7c6 b5e2 d8d1 e1d1 c8c7 d1d2 h8d8 d2c3 e7c8 a2a4 a7a5 g3f1 c8d6 c3b2 b7b6 f1e3');
INSERT INTO evaluations VALUES(882,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(882,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(882,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(882,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(882,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(882,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(882,36,'cp -12','f2f3 c8d7 c1e3 e8c8 d1e1 f6e8 g1e2 e7e6 e2d4 c6e5 a1d1 e8d6 e3f4 d6c4 f4c1 c4b6 b2b3 f8b4 d4e2 a7a5 a2a3 b4c5 h2h4 a5a4 h4h5 d7c6 b3b4 d8d1 c3d1 c5e7 d1e3 c6d7 e2c3 h8d8 e1f2 h7h6 f1e2 e7f6');
INSERT INTO evaluations VALUES(34506,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(34506,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(34506,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(34506,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(34506,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(34506,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(34506,36,'cp 135','g1f3 f8g7 d2d4 e6d6 e1g1 g8f6 c3b5 d6d8 c2c4 c7c6 b5c3 c8g4 f1e1 e8g8 h2h3 g4f3 e2f3 f8e8 c1f4 b8d7 a1b1 a8c8 b2b4 d7f8 f4e5 f8d7 d1b3 d7e5 d4e5 f6d7 e5e6 f7e6 c4c5 d7f8 c3e4 d8d4 e4g5 d4h4 g5e6 f8e6 e1e6');
INSERT INTO evaluations VALUES(34505,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(34505,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(34505,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(34505,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(34505,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(34505,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(34505,36,'cp 149','e6d6 d2d4 f8g7 e1g1 g8f6 c3b5 d6d8 c2c4 c7c6 b5c3 c8g4 f1e1 b8d7 c1f4 e8g8 d1d2 c6c5 d4d5 f6e8 f3g5 g4e2 d2e2 e8d6 g5e4 d8b6 a1d1 d6e4 c3e4 a8e8 b2b3 e7e5 f4d2 f7f5 e4g5 e5e4 g5e6 e8e6 d5e6 b6e6 g1h1 f8d8');
INSERT INTO evaluations VALUES(39054,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(39054,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39054,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(39054,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(39054,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(39054,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(39054,36,'cp 12','g8e7 d2d3 d7d6 b2b4 e8g8 a1b1 c7c6 e2e4 b8d7 g1e2 d7f6 b4b5 d6d5 b5c6 b7c6 e4d5 c6d5 e1g1 c8e6 d3d4 e5d4 e2d4 e6g4 d1d2 d5c4 h2h3 g4f5 d4f5 e7f5 d2d8 a8d8 b1b4 f6d7 b4c4 d7b6 c4c7 f5d4 c1e3 d8c8 c7c8 f8c8 c3d5 c8c2 d5b6');
INSERT INTO evaluations VALUES(42593,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(42593,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(42593,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(42593,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(42593,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(42593,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(42593,36,'cp 200','f3e5 e8g8 d2d4 c5b4 c1d2 d7d6 e5d3 c7c5 d4c5 d6c5 d3f4 c8f5 e2e3 b8c6 f1e2 f6e4 c3e4 f5e4 f2f3 e4g6 f4d5 d8h4 g2g3 h4h3 d5f4 h3f5 a2a3 b4d2 d1d2 a8d8 d2c3 f8e8 e3e4 f5e5 c3e5 c6e5 f4d5 f7f5');
INSERT INTO evaluations VALUES(42528,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(42528,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(42528,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(42528,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(42528,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(42528,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(42528,36,'cp 198','e8g8 d2d4 c5b4 c1d2 f8e8 e5f3 c7c5 d4c5 b4c5 a2a3 d7d5 c4d5 f6d5 c3d5 d8d5 e2e3 c8e6 a1c1 b8c6 f1e2 a8d8 d2c3 d5d1 e2d1 c5f8 e1g1 g7g6 f1e1 f8g7 g1f1 e6d5 h2h3 g7c3 c1c3 c6a5 f3d2');
INSERT INTO evaluations VALUES(50448,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(50448,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(50448,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(50448,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(50448,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(50448,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(50448,36,'cp -129','e4f6 c2c4 g7g6 d3d4 f8g7 b1c3 e8g8 f1e2 d7d5 e1g1 d5c4 e2c4 c8g4 d4d5 c7c6 h2h3 g4f3 d1f3 b7b5 c4b3 b5b4 c3e4 f6d5 f1d1 e7e6 c1g5 d8b6 e4f6 d5f6 g5f6 c6c5 f6g7 g8g7 b3e6 b8c6 e6d5 a8c8 a1c1 c6e5 f3e3 f8e8 b2b3 e5d7 e3f4 d7f6 d5c4 c8d8 d1d8 e8d8');
INSERT INTO evaluations VALUES(50375,30,'cp -131',NULL);
INSERT INTO evaluations VALUES(50375,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(50375,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(50375,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(50375,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(50375,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(50375,36,'cp -107','c2c4 g7g6 b1c3 f8g7 d3d4 e8g8 f1e2 d7d5 e1g1 d5c4 e2c4 c8g4 h2h3 g4f3 d1f3 c7c6 d4d5 b7b5 c4b3 b5b4 c3e4 f6d5 f1d1 e7e6 c1g5 d8b6 e4f6 g7f6 g5f6 b8d7 f6g5 a7a5 a2a4 b4a3 b2a3 a5a4 b3c2 d7e5 f3g3 f7f6 g5d2 c6c5 a1b1 b6c7 d1e1 c5c4 c2e4');
INSERT INTO evaluations VALUES(53645,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(53645,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(53645,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(53645,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(53645,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(53645,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(53645,36,'cp 59','b7b5 c4d3 a7a6 g1f3 c6c5 f3e5 g8f6 a2a4 b5b4 c3e4 b8d7 e4f6 d7f6 a4a5 c8b7 d1a4 f6d7 c1d2 a8c8 e5d7 d8d7 a4d7 e8d7 d4c5 f8c5 e1e2 f7f5 f2f3 c5d6 e3e4 f5e4 f3e4 h7h6 h2h4 d7e7 h1f1 c8c6 e4e5 d6e5 d2b4 e5d6 d3g6 d6b4 f1f7 e7d6 a1d1 d6e5 f7b7 b4a5 b7g7');
INSERT INTO evaluations VALUES(51692,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(51692,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(51692,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(51692,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(51692,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(51692,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(51692,36,'cp 73','c4d3 a7a6 g1f3 c6c5 f3e5 g8f6 a2a4 b5b4 c3e4 b8d7 e4f6 d7f6 a4a5 c8b7 d1a4 f6d7 c1d2 a8c8 d4c5 f8c5 a4d7 d8d7 e5d7 e8d7 e1e2 f7f5 f2f3 c5d6 e3e4 f5e4 f3e4 h7h6 h2h4 h8f8 h1d1 d6c5 d2e1 c5d6 a1a4 b7c6 a4b4 d6b4');
INSERT INTO evaluations VALUES(1188,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1188,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1188,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1188,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1188,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1188,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1188,36,'cp 0','e5f3 g5g2 h1g1 g2h3 g1g3 h3h5 g3g5 h5h6 d2d4 g8e7 g5g3 h6h5 g3g5 h5h6');
INSERT INTO evaluations VALUES(1200,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1200,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1200,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1200,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1200,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1200,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1200,36,'cp 0','g5g2 h1g1 g2h3 g1g3 h3h5 g3g5 h5h3');
INSERT INTO evaluations VALUES(1208,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1208,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1208,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1208,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1208,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1208,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1208,36,'cp 0','h1g1 g2h3 g1g3 h3h5 g3g5 h5h3');
INSERT INTO evaluations VALUES(1207,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1207,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1207,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1207,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1207,36,'cp 0','g2h3 g1g3 h3h5 g3g5 h5h3');
INSERT INTO evaluations VALUES(1206,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1206,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1206,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1206,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1206,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1206,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1206,36,'cp 0','g1g3 h3h5 g3g5 h5h3');
INSERT INTO evaluations VALUES(3193,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3193,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3193,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3193,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(3193,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(3193,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(3193,36,'cp 5','a1b1 a8b8 a2a3 a7a6 b2b4 c5b4 a3b4 b7b5 c4b5 a6b5 c1e3 f6g4 e3d2 c6e5 f3e5 g4e5 d1c1 c8b7 g2b7 b8b7 d2h6 g7h6 c1h6 d8d7 f1c1 d7e6 h2h3 f8c8 g3g4 b7c7 c3b5 c7c1 b1c1 c8c1 h6c1 e6b3 b5c3 b3b4 c3e4 b4b6 g1g2 g8g7 c1a1 f7f6 a1a8 e5c6 g4g5 f6g5');
INSERT INTO evaluations VALUES(3192,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3192,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(3192,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(3192,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(3192,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(3192,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(3192,36,'cp 7','a8b8 a2a3 a7a6 b2b4 c5b4 a3b4 b7b5 c4b5 a6b5 c1e3 f6g4 e3d2 g4e5 f3e5 c6e5 d1b3 c8e6 c3d5 d8d7 f1c1 f8c8 c1c8 b8c8 b1c1 c8c1 d2c1 h7h6 c1b2 e6d5 b3d5 e5f3 g2f3 g7b2 d5a8 g8g7 h2h4 h6h5 e2e3 b2f6 a8c6 d7c6 f3c6 f6c3 c6b5 c3b4 d3d4 e7e6');
INSERT INTO evaluations VALUES(6639,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(6639,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(6639,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(6639,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(6639,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(6639,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(6639,36,'cp 53','d7d6 f1b5 c8d7 d4e5 d6e5 d1d3 d8e7 a2a4 a8d8 e1g1 a7a6 b5c4 d7e6 c4d5 e6d5 e4d5 e5e4 d3d2 e7f6 d2e3 d8d5 e3e4 g8e7 c1a3 f6e6 e4f4 e6f6 f4f6 g7f6 f1b1 c6a5 b1e1 d5d7 a1b1 a5c4 e1e7 d7e7 a3e7 e8e7 b1b7 e7d7');
INSERT INTO evaluations VALUES(6059,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(6059,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(6059,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(6059,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(6059,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(6059,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(6059,36,'cp 84','f1b5 c8d7 d4e5 d6e5 d1d3 g8e7 c1a3 d8c8 a3e7 e8e7 b5c6 d7c6 f3e5 c8e6 e5c6 e6c6 e1c1 h8d8 d3e3 e7f8 d1d8 a8d8 h1d1 d8d1 c1d1 g7g6 e3d4 c6a6 d1d2 a6f1 g2g3 b7b6 e4e5 f8e7 a2a4 e7e8 h2h4 a7a5 e5e6 f7e6 d4h8 e8d7 h8h7 d7d6 h7g6 f1f2 d2c1');
INSERT INTO evaluations VALUES(9678,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(9678,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(9678,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(9678,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(9678,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(9678,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(9678,36,'cp -106','c6d4 f3d1 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 e8g8 g1f3 h7h6 g5f6 d8f6 e1g1 f6c6 f1e1 d7d6 b1d2 c6b7 h2h3 c8d7 d1c2 d7e6 a1a3 a8a6 e1a1 f8a8 d2f1 c7c6 f1e3 g8h7 c2d2 d6d5 b3b4 d5e4 d3e4 a5a4 h3h4');
INSERT INTO evaluations VALUES(16442,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16442,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16442,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(16442,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16442,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16442,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(16442,36,'cp 10','g1f3 e5e4 f3d4 f8c5 d4f5 e8g8 e2e3 d7d5 c4d5 c6b4 f5g3 f6d5 a2a3 b4d3 f1d3 e4d3 b1c3 f8e8 d1f3 c8e6 a1c1 c5f8 g3h5 d5c3 c1c3 c7c6 f3g3 e6f5 h2h4 d8d6 g3d6 f8d6 h5g3 d6g3 f2g3 f5e4 e1g1 f7f6 f1f4 b7b6 g3g4 c6c5 g4g5 f6g5 h4g5 a8d8 c3c4 e4d5');
INSERT INTO evaluations VALUES(19709,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(19709,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(19709,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(19709,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(19709,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(19709,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(19709,36,'cp -75','b7c6 g1f3 e7e6 e1g1 f8d6 d2d4 g8e7 c2c4 e8g8 b1c3 d5c4 c3e4 a7a5 f1e1 e7f5 d1c2 c8a6 e4d6 d8d6 c1d2 d6c7 d2c3 f8d8 e1e5 d8d5 a1e1 h7h6 c2e2 f5d6 h2h4 a6c8 f3d2 d6f5 e5e4 c8a6 e2g4 c7e7 d2f3 a8d8 f3e5 a6b5 h4h5 c6c5 d4c5 e7c5');
INSERT INTO evaluations VALUES(21051,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21051,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21051,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21051,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21051,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21051,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(21051,36,'cp 5','e4d5 d8d5 g1f3 e7e5 b1c3 d5d8 c1e3 c8e6 d3d4 e5d4 f3d4 c6d4 d1d4 d8d4 e3d4 e8c8 e1c1 g8f6 g2g3 f8e7 h2h3 h7h5 f1g2 a7a6 h1e1 d8d7 d4e3 d7d1 e1d1 h8e8 a2a3 g7g6 e3g5 c7c6 d1e1 e7d8 b2b4 a6a5 b4b5');
INSERT INTO evaluations VALUES(43541,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(43541,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(43541,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43541,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(43541,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(43541,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(43541,36,'cp 31','g2g4 h7h6 a2a3 b7b5 d1e2 e6e5 d4f5 b8c6 g4g5 g7g6 f5h6 c6d4 e2d1 f8g7 c1e3 f6h5 c3d5 c8e6 h3h4 e6d5 e4d5 g7h6 e3d4 e5d4 g5h6 d8f6 d1e2 e8f8 e1c1 h8h6 e2d2 f8g7 d2d4 a8e8 d4d2 h6h8 f1d3 f6f4 c1b1 f4d2');
INSERT INTO evaluations VALUES(43531,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43531,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(43531,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(43531,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(43531,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43531,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(43531,36,'cp 29','h7h6 a2a3 b7b5 d1e2 e6e5 d4b3 b8c6 c1e3 a8b8 f2f4 b5b4 a3b4 c6b4 e1c1 c8e6 e3a7 b8a8 f4e5 f6d7 a7f2 d6e5 c3d5 f8d6 c2c3 b4c6 h1g1 c6a5 b3d2 a8b8 h3h4 d6c5 g4g5 h6g5 g1g5 g7g6 f2c5 d7c5');
INSERT INTO evaluations VALUES(51913,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(51913,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(51913,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(51913,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(51913,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(51913,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(51913,36,'cp -56','g7g6 h5g5 d8g5 f3g5 g8f6 d2d3 h7h6 g5f3 b8c6 c2c3 b7b6 b1a3 g6g5 h2h3 c8b7 c1e3 e8c8 f1e2 b7a6 a3c2 d7d5 e4e5 f6d7 d3d4 a6e2 e1e2 c5d4 c3d4 c8b7 e3d2 f7f6 e5f6 d7f6 f3e5 c6e5 d4e5 f6e4 c2d4 e4d2 e2d2');
INSERT INTO evaluations VALUES(7838,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(7838,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(7838,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(7838,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(7838,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(7838,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(7838,36,'cp 105','f1d3 d7d5 e4e5 f6e4 e1g1 e4c3 b2c3 b4c3 a1b1 c3b4 f3g5 b4e7 d1h5 g7g6 h5g4 c6b4 d3b5 b4c6 c1d2 a7a5 b1b3 h7h6 g5f3 c8d7 f1b1 a5a4 b3b2 e7a3 b2c2 c6a5 b5d3 a5c4 b1b7 c4d2 c2d2 a8b8 b7b8 d8b8 d2d1 a3e7 d1b1 b8a7 h2h4 e8f8 d3g6 h8g8 h4h5 f7g6 h5g6 g8g7 g4f4 f8g8 f4h6');
INSERT INTO evaluations VALUES(7839,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(7839,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(7839,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(7839,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(7839,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(7839,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(7839,36,'cp 115','h7h6 e1g1 d7d6 d1b3 b4a5 d3c2 e8g8 e4e5 f6d5 c3d5 e6d5 h2h3 c8e6 e5d6 d8d6 b3d3 f7f5 c1d2 a8e8 f1e1 a5c7 d2c3 d6d7 e1e2 d7f7 a1e1 e6c8 b2b4 a7a6 b4b5 e8e2 d3e2 a6b5 c2b3 c7d6 e2b5 c8e6 a2a4 f8c8 a4a5');
INSERT INTO evaluations VALUES(7849,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7849,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(7849,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(7849,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(7849,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7849,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(7849,36,'cp 34','d4c6 b7c6 f1d3 e6e5 e1g1 e8g8 c1e3 b4c3 b2c3 d7d6 h2h3 h7h6 f2f4 d8e7 d1f3 f8d8 f1f2 c8a6 a1f1 f6d7 f4f5 f7f6 f1d1 d7c5 e3c5 d6c5 h3h4 e7f7 f2d2 g8h8 f3e3 f7e7 d3f1 d8d2 d1d2');
INSERT INTO evaluations VALUES(7817,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(7817,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(7817,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(7817,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(7817,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(7817,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7817,36,'cp 6','b7c6 f1d3 e6e5 e1g1 e8g8 c1e3 d7d6 h2h3 b4c3 b2c3 c8e6 f2f4 e5f4 e3f4 d8b6 f1f2 b6c5 d3c2 c5c4 f4d6 f8d8 e4e5 f6e8 c2b3 c4c3 f2f3 e6b3 a2b3 c3b2 f3d3 e8d6 e5d6 d8d7 a1c1 a8c8 d3c3 h7h6 c3c6 c8c6 c1c6 b2e5 g1h1');
INSERT INTO evaluations VALUES(7650,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(7650,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(7650,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(7650,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(7650,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(7650,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7650,36,'cp 10','f1d3 e6e5 c1e3 e8g8 e1g1 b4c3 b2c3 d7d6 h2h3 c8e6 d1d2 d6d5 c4d5 c6d5 e4d5 f6d5 d3e4 a8c8 e4d5 d8d5 d2d5 e6d5 e3a7 c8a8 a7c5 f8c8 c5d6 f7f6 a2a4 c8c3 a4a5 a8a6 d6b4 c3c2 f1c1 c2c1 a1c1 h7h5 f2f3 a6c6 c1c5 c6c5 b4c5 d5c4 c5d6 h5h4');
INSERT INTO evaluations VALUES(7651,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(7651,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(7651,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(7651,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(7651,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(7651,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(7651,36,'cp 6','e6e5 e1g1 e8g8 h2h3 b4c3 b2c3 d7d6 c1e3 c8e6 d1d2 d6d5 c4d5 c6d5 e4d5 f6d5 d3e4 a8c8 e4d5 d8d5 d2d5 e6d5 e3a7 c8a8 a7c5 f8c8 c5d6 c8c3 f1b1 f7f6 a2a4 a8a6 d6b4 c3c2 a4a5 g8f7 b1c1 c2c1 a1c1 g7g5 c1c7 f7g6 b4c3 h7h5 f2f3 g6f5 c7c5 d5e6 c3d2 g5g4 f3g4 h5g4 h3g4 f5g4');
INSERT INTO evaluations VALUES(19465,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(19465,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(19465,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(19465,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(19465,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(19465,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(19465,36,'cp -41','f3f4 d7d5 e4d5 d8d5 b1c3 d5d6 g1f3 g8f6 f1b5 g7g6 b5c6 d6c6 e1g1 f8g7 f3e5 c6c7 d2d3 e8g8 d1e2 b7b6 a2a4 a7a6 c1d2 f8d8 a1e1 e7e6 h2h3 c8b7 b2b3 a8c8 e2f2 f6d5 c3d5 b7d5 e1e2 b6b5 d2c3 d5b7 c3b2');
INSERT INTO evaluations VALUES(21015,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(21015,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(21015,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(21015,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(21015,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(21015,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(21015,36,'cp -64','e7e5 d4e5 c6e5 f1g2 g8f6 b1d2 f8b4 a2a3 b4d6 h2h3 e8g8 b2b3 c8f5 c1b2 d8e7 g1f3 h7h6 e1g1 a8d8 b3b4 f8e8 a1c1 a7a5 f3e5 d6e5 b2e5 e7e5 d2b3 a5b4 a3b4 e5e7 d1d4 f5e4 f1d1 c7c6 b4b5 e4g2');
INSERT INTO evaluations VALUES(21689,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(21689,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(21689,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(21689,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(21689,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(21689,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(21689,36,'cp -45','d4e5 c6e5 f1g2 c7c6 b1d2 g8f6 g1f3 e5f3 d2f3 f8b4 c1d2 b4e7 e1g1 e8g8 d1e2 c8f5 d2c3 d8b6 b2b3 b6c5 f3d4 f5g6 c3b2 a8d8 a2a4 f8e8 f1d1 a7a5 h2h4 e7f8 g2f3 h7h5 b2a3 c5c3 a3f8 e8f8 g1g2 d8d7 a1c1 f8d8');
INSERT INTO evaluations VALUES(21356,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(21356,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(21356,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(21356,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(21356,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(21356,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(21356,36,'cp -53','e6e5 d4e5 f8c5 g2g3 c8g4 f1g2 g8e7 b1c3 d8d7 e1g1 c5b6 c1g5 e8g8 g5e7 c6e7 d1b3 a8d8 f1d1 d5d4 c3a4 e7c6 a4b6 a7b6 d1d2 d7e7 b3a4 c6e5 d2d4 b6b5 a4d1 e5f3 g2f3 g4f3 e2f3 b7b6 d1d3 d8d4 d3d4 c7c5 d4f4 b5b4 a1c1');
INSERT INTO evaluations VALUES(31270,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(31270,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(31270,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31270,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31270,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(31270,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(31270,36,'cp 27','b8c6 c2c4 g8f6 e1g1 d5c4 b1c3 e7e6 h2h3 g4f3 d1f3 a8c8 f1d1 a7a6 b5c6 b7c6 f3e2 f8d6 e2c4 a6a5 c1g5 d6e7 a1c1 e8g8 c4e2 h7h6 g5e3 f6d5 c3a4 e7d6 c1c4 d8f6 e3d2 d6c7 d1c1 f8d8 c4c6 f6d4 c6c4 d4f6 b2b3');
INSERT INTO evaluations VALUES(39001,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(39001,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(39001,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(39001,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(39001,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(39001,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(39001,36,'cp 36','d7d5 e4d5 e6d5 c4b3 g8e7 d2d4 e8g8 c2c3 c7c6 f1e1 b8d7 a2a4 a7a5 c1f4 d7f6 h2h3 e7f5 b1d2 f5d6 f3e5 f8e8 d2f1 f6h5 f4h2 g7h6 b3c2 h5g7 g2g4 h6g5 d1f3 c8e6 f1e3 d6e4 f3d1 h7h5 e3g2 h5g4 h3g4');
INSERT INTO evaluations VALUES(44894,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44894,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(44894,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(44894,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44894,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(44894,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44894,36,'cp 15','d2d4 g7g6 c4d5 c6d5 g1f3 f8g7 b1c3 b8c6 f3e5 c8d7 c1f4 e7e6 d1d2 c6e5 f4e5 a8c8 e1g1 e8g8 f1c1 d8b6 e2e3 f8d8 c1c2 f6e8 h2h4 e8d6 e5g7 g8g7 h4h5 f7f5 g2f3 d6c4 d2e2 g6g5 a1c1 c4d6 b2b3 c8c6');
INSERT INTO evaluations VALUES(44917,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44917,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44917,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44917,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44917,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44917,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(44917,36,'cp 12','g7g6 c4d5 c6d5 g1f3 f8g7 f3e5 e8g8 b1c3 c8f5 c1f4 d8b6 d1b3 b6b3 a2b3 b8c6 e5c6 b7c6 a1a6 f8c8 b3b4 f6h5 f4e3 a8b8 b4b5 e7e5 d4e5 g7e5 e3a7 e5c3 b2c3 b8b5 e1g1 b5b2 f1a1 b2e2 a7e3 e2c2 h2h3 f5e6 g3g4 h5f6 e3d4 f6e4');
INSERT INTO evaluations VALUES(45241,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45241,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45241,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45241,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(45241,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(45241,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45241,36,'cp -25','f6e4 d1h5 e7e6 c4b5 b8d7 d4e6 e4f6 b5d7 c8d7 e6d8 f6h5 d8b7 d7c6 b7a5 c6g2 h1g1 g2e4 c2c4 g7g6 b1c3 e4f3 c3b5 a8c8 c1e3 a7a6 b5d4 f3a8 b2b4 f8g7 e1c1 g7e5 c1c2 e5h2 g1e1 e8g8 a2a4 h5f6 b4b5 a6b5 a4b5 f6g4 a5c6 g4e3 e1e3 g8g7 c2d3 h7h5 d1e1 a8c6 d4c6');
INSERT INTO evaluations VALUES(45283,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45283,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(45283,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(45283,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(45283,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(45283,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(45283,36,'cp 0','d1h5 e7e6 c4b5 b8d7 d4e6 e4f6 e6d8 f6h5 b5d7 c8d7 d8b7 d7c6 b7a5 c6g2 h1g1 g2e4 c1e3 e4c2 b1c3 f8e7 a1c1 c2g6 c3d5 e7d8 a5b7 h5f6 d5f6 d8f6 b7d6 e8d7 d6c4 d7e6 g1g3 h8c8 e3f4 f6d4 f4e3');
INSERT INTO evaluations VALUES(48127,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(48127,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(48127,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(48127,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(48127,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(48127,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(48127,36,'cp 54','f1b5 c7c6 b5d3 c6c5 d3f5 e6f5 c1e3 c5d4 e3d4 f8b4 b1c3 b8c6 e1g1 b4c3 d4c3 e8g8 d1d2 f8e8 c3d4 g7g5 a1d1 h7h6 d4c3 c8e6 f3d4 c6d4 c3d4 f5f4 f2f3 g8h7 d2d3 h7h8 f1f2 e8g8 g2g4 g8g6 h2h4 g5h4 g1h2 d8g8 f2g2');
INSERT INTO evaluations VALUES(48838,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(48838,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(48838,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(48838,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(48838,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(48838,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(48838,36,'cp -79','d5e4 d3e4 d8d1 e1d1 h7h5 f2f3 h5h4 g3g4 a7a5 a2a4 c8e6 g1h3 b8c6 c2c3 e6b3 d1e2 f6d7 c1e3 f8c5 b1d2 b3e6 g2f1 f7f6 h3f2 g7g5 e3c5 d7c5 e2e3 e8f7 f1b5 h8d8 h1d1 a8b8 e3e2 d8d6 f2d3 c5d3');
INSERT INTO evaluations VALUES(48944,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(48944,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(48944,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(48944,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(48944,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(48944,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(48944,36,'cp -66','d3e4 d8d1 e1d1 f8c5 f2f3 h7h5 g2h3 b8d7 c1g5 f6h7 g5d2 h5h4 d1e2 h7f6 b1c3 c7c6 d2e3 e8e7 c3d1 c5d4 e2f2 d7c5 f2g2 c8h3 g1h3 a8d8 h1e1 h4g3 h2g3 g7g6 h3f2 b7b5 e3d4 d8d4 d1e3 a7a5 a1d1 f6d7 c2c3');
INSERT INTO evaluations VALUES(52819,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(52819,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(52819,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(52819,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52819,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(52819,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(52819,36,'cp -17','d5e4 c3e4 g8f6 d1e2 f6e4 e2e4 b8d7 g1f3 d7f6 e4e2 c8g4 h2h3 g4f3 e2f3 e7e6 c2c3 f8d6 d2d4 d8b6 f1d3 e8g8 e1g1 a8d8 g1h1 c6c5 d4c5 d6c5 f4f5 e6e5 b2b4 c5e7 c1g5 b6c7 g5f6 e7f6 d3e4 b7b6 a1d1 d8d1 f1d1');
INSERT INTO evaluations VALUES(52930,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(52930,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52930,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52930,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52930,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52930,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(52930,36,'cp -9','c3e4 g8f6 d1e2 f6e4 e2e4 b8d7 g1f3 d7f6 e4e2 c8g4 h2h3 g4f3 e2f3 e7e6 c2c3 f8e7 d2d4 d8b6 f1d3 e8g8 e1g1 c6c5 d4c5 e7c5 g1h1 a7a5 f4f5 e6e5 c1g5 b6b2 g5f6 g7f6 d3e4 a8d8 e4b7 b2d2 a1d1 d2e3 b7d5 e3f3 g2f3 f8e8 a2a4 c5b6 c3c4');
INSERT INTO evaluations VALUES(52917,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(52917,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(52917,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(52917,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(52917,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(52917,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52917,36,'cp -21','g8f6 d1e2 f6e4 e2e4 b8d7 g1f3 d7f6 e4e2 g7g6 g2g3 f8g7 f1g2 c8g4 c2c3 e8g8 e1g1 b7b5 d2d4 b5b4 c3c4 f6e8 c1e3 e8d6 a1d1 d6f5 f1e1 a7a5 e3f2 g4f3 g2f3 f5d4 f2d4 g7d4 g1g2 d8b6 e2e7 b4b3 a2b3 b6b3 f3c6 b3b2 g2h1');
INSERT INTO evaluations VALUES(55756,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(55756,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(55756,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(55756,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(55756,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(55756,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(55756,36,'cp -48','c7c6 b5e2 d5d4 c3b1 f8d6 d2d3 e6e5 a2a4 g8f6 b1a3 a7a6 a3c4 d6c7 c2c3 c6c5 c3d4 c5d4 c1d2 b8c6 g1f3 c8e6 d1c2 e8g8 e1g1 a6a5 f3g5 e6d7 c2b3 b7b6 g5f3 h7h6 f1c1 d7e6 b3a3 a8c8 b2b4 e6c4 c1c4 c7d6 a3b3 c6b4 d2b4 a5b4 g2g3 c8c4');
INSERT INTO evaluations VALUES(53530,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(53530,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(53530,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(53530,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(53530,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(53530,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(53530,36,'cp -27','b5e2 d5d4 c3b1 f8d6 d2d3 g8e7 a2a4 e6e5 b1a3 e8g8 g1f3 f7f6 e1g1 c8e6 f3d2 d8d7 h2h3 b7b6 a3c4 d6c7 e2g4 c6c5 b2b3 a7a6 g4e6 d7e6 g1h1 b8c6 d1g4 e6f7 g4e2 b6b5 a4b5 a6b5 a1a8 f8a8');
INSERT INTO evaluations VALUES(1450,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(1450,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(1450,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(1450,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(1450,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(1450,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(1450,36,'cp 24','d1c1 g8f6 h2h3 c8d7 a2a4 a7a5 f1d3 f8e7 e1g1 e8g8 d3c2 f8d8 b1d2 d7e8 c1b1 h7h6 f1e1 a8c8 f4h2 f6d7 g1h1 d7f8 d2b3 f8d7 g2g4 d7f6 b3d2 e7d6 h2d6 d8d6 e1g1 c5d4 e3d4 f6d7 g1e1 d7f8 d2f1 d6d8 f1g3 b6c7 g3h5 f7f6 f3h4');
INSERT INTO evaluations VALUES(4432,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(4432,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(4432,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(4432,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(4432,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(4432,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(4432,36,'cp -33','b4c3 d2c3 f8e8 f1e2 d7d5 e4d5 f6d5 c3d2 d8f6 e1g1 h7h6 b2b4 c8d7 c2c4 d5f4 d2f4 f6f4 f1e1 a7a6 b4b5 a6b5 c4b5 c6d4 f3d4 f4d4 e2f3 d7b5 f3b7 a8a7 b7e4 c7c6 d1f3 d4d6 h2h3 a7a3 a1c1 e8d8 f3e3 a3b3 c1c5 d6f6 e3f3 f6f3 e4f3 b3d3 f3c6 b5c6');
INSERT INTO evaluations VALUES(4464,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(4464,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(4464,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(4464,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(4464,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(4464,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(4464,36,'cp -49','d2c3 f8e8 f1e2 d7d5 e4d5 f6d5 c3d2 d8f6 e1g1 c8f5 b2b4 h7h6 c2c4 d5f4 d2f4 e5f4 d3d4 a8d8 d4d5 g7g5 h2h3 g5g4 h3g4 f5g4 f1e1 g4f3 e2f3 e8e1 d1e1 c6e5 a1c1 e5f3 g2f3 f6g6 g1f1 g6h5 e1e7 d8d5 c4d5 h5h1 f1e2 h1c1 e7e8 g8g7 e8e5 g7g6 e5e4 g6f6');
INSERT INTO evaluations VALUES(4461,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(4461,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(4461,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(4461,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(4461,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(4461,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(4461,36,'cp -34','f8e8 f1e2 d7d5 e4d5 f6d5 c3d2 d8f6 e1g1 h7h6 b2b4 c8d7 c2c4 d5f4 d2f4 f6f4 f1e1 a8d8 b4b5 c6e7 d1c1 f4c1 a1c1 e7g6 d3d4 e5d4 f3d4 e8e4 d4b3 b7b6 c1d1 g6f4 g1f1 f4e2 e1e2 e4c4 e2d2 c4c3 d2d7 d8d7 d1d7 c3b3 d7c7 b3a3 c7c8 g8h7 c8c7 h7g6 g2g3 a3a1 f1g2');
INSERT INTO evaluations VALUES(22510,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(22510,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(22510,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(22510,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(22510,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(22510,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(22510,36,'cp -25','f8c5 d2d3 g8f6 c1e3 c5b6 e3b6 a7b6 e1g1 e8g8 b1c3 d7d6 h2h3 c6d4 b3b4 c7c6 f3d4 e5d4 c3e2 d6d5 e4d5 f6d5 b4b5 d8d6 f1e1 c8d7 a2a4 f8e8 b5c6 b7c6 e2g3 e8e1 d1e1 a8a5 g3e4 d6b4 e1b4 d5b4 c4b3 d7e6 e4d2 c6c5 b3e6 f7e6 d2c4');
INSERT INTO evaluations VALUES(6207,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(6207,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(6207,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(6207,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(6207,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(6207,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(6207,36,'cp -16','e1g1 g8f6 d2d3 e8g8 c1e3 c5b6 e3b6 a7b6 b1c3 d7d6 h2h3 c6d4 b3b4 c7c6 f3d4 e5d4 c3e2 d6d5 c4b3 d5e4 d3e4 f6e4 d1d4 f8e8 d4d8 e8d8 f1d1 c8e6 f2f3 e4f6 c2c4 g8f8 e2d4 c6c5 d4e6 f7e6 b4c5 b6c5 d1e1 f8f7 e1e5 f6d7 e5e1 f7f6 a2a4 d7e5');
INSERT INTO evaluations VALUES(6206,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(6206,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(6206,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(6206,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(6206,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(6206,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(6206,36,'cp -25','g8f6 d2d3 e8g8 b1d2 a7a6 a2a3 b7b5 b3b4 b5c4 b4c5 c4d3 c2d3 d7d5 c5d6 c7d6 c1b2 f6h5 a1c1 c8b7 f3e5 c6e5 d1h5 e5d3 c1b1 b7c6 b2d4 c6b5 h5g4 d3e5 d4e5 d6e5 f1d1 b5d7 g4f3 d8c7 f3g3 f8e8');
INSERT INTO evaluations VALUES(27557,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(27557,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(27557,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(27557,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(27557,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(27557,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(27557,36,'cp -94','a7a6 d1e2 a8c8 g1h1 d4h4 f2f4 g8h6 c1e3 f8c5 e3c5 c8c5 h2h3 e8g8 h1h2 f8c8 a1d1 f7f5 g2g3 h4d8 e2f2 d8b6 d1d2 c5c6 c3e2 b6f2 f1f2 c6c7 e2d4 h6f7 d2c2 f7d8 c2c7 c8c7 g3g4 d8c6 d4c6 c7c6 g4f5 e6f5');
INSERT INTO evaluations VALUES(27484,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(27484,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(27484,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(27484,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(27484,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(27484,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(27484,36,'cp -97','d1e2 g8e7 g1h1 d7c6 f2f4 d4b6 f4f5 e7f5 d3f5 d5d4 f5e4 d4c3 e2f3 e8c8 c1e3 c6e4 f3e4 b6b4 e4b4 f8b4 b2c3 b4a3 f1f7 d8d7 f7f3 c8c7 h1g1 c7c6 g1f2 b7b5 e3d4 h8d8 f3f4 a6a5 f2e2 a3c5 a1c1 c6b6 c1d1 d7d5 g2g3 b6c6 d1c1 h7h6 c1d1 b5b4 c3b4');
INSERT INTO evaluations VALUES(31896,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(31896,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(31896,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(31896,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(31896,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(31896,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(31896,36,'cp 210','c4d5 d8d5 b1c3 d5a5 h2h3 g4h5 g2g4 h5g6 f1g2 b8c6 c1d2 e8c8 d4d5 c6d4 a1c1 e7e5 e2e3 d4b5 d1b3 b5d6 g1f3 f7f6 e3e4 g6e8 e1g1 a5b6 b3a3 c8b8 d2e3 b6a6 a3a6 b7a6 f3d2 g8e7 d2b3 d6c4 e3c5 c4d6 f2f4 e7g6 f4f5 g6f4 f1f4 e5f4 c3e2');
INSERT INTO evaluations VALUES(36495,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36495,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(36495,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(36495,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36495,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36495,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(36495,36,'cp 20','d2d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 c2c3 c7c5 e4c2 c5d4 f3d4 f8e8 f1e1 a7a6 c1g5 h7h6 g5e3 d7f6 d4f3 c8f5 c2f5 g6f5 d1b3 b7b5 a2a4 d8d5 b3c2 e7e6 a4b5 a6b5 e3d4 f6e4 d4g7 g8g7 c2e2 a8a1 e1a1');
INSERT INTO evaluations VALUES(38243,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38243,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(38243,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38243,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38243,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38243,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38243,36,'cp 0','b1c3 e7h4 e1f1 h4e7 d2d4 g8f6 c1f4 e8g8 f1f2 c7c5 h1e1 c5d4 f3d4 b8c6 f2g1 d8b6 f4e3 b6b2 c3d5 f6d5 c4d5 b2c3 d4c6 b7c6 e3d4 c8g4 d4c3 g4d1 a1d1 c6d5 e4d5 e7g5 c3d2 g5d2 d1d2 f8e8 e1e8 a8e8 g1f2 g8f8 c2c4 e8e4 d2c2 f8e7 f2f3 f7f5 c4c5 d6c5 c2c5 e4a4 c5c7 e7d6 c7g7');
INSERT INTO evaluations VALUES(52472,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52472,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52472,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52472,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52472,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52472,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(52472,36,'cp 4','g8f6 d2d4 e7e6 c1f4 f8e7 h2h3 d7d5 e2e3 e8g8 f1e2 b7b6 b1d2 e7d6 e1g1 d6f4 e3f4 c8a6 e2a6 b8a6 f1e1 a6b8 f3e5 d8c7 d2f3 b8c6 h3h4 a8c8 h4h5 h7h6 g2g3 c7b7 g1g2 f6e4 e5c6 b7c6');
INSERT INTO evaluations VALUES(9202,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(9202,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(9202,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(9202,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(9202,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(9202,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9202,36,'cp 0','c2c4 d5c4 d3c4 h7h6 b1c3 e8g8 b2b3 f8e8 c1b2 e5e4 f3d4 c6d4 d1d4 d6e5 d4d8 e8d8 f1d1 c8e6 d1d8 a8d8 a1d1 d8d1 c3d1 e5b2 d1b2 g8f8 f2f3 e4f3 e2f3 b7b6 g1f2 f8e7 b2a4 e7d6 f2e3 d6e5 g2f1 e6d7 a4c3 g7g5 f1e2 d7c6 e2d3 f6h5 c3e2 h5g7');
INSERT INTO evaluations VALUES(16019,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(16019,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(16019,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(16019,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(16019,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(16019,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(16019,36,'cp 61','g1f3 e7e5 e1g1 e5d4 f3d4 g7g6 c1f4 f8g7 d1d2 e8g8 c2c4 d7e5 b1c3 e5d3 d2d3 f6h5 f4e3 f7f5 e4f5 c8f5 d3d2 d8d7 a1e1 a8e8 e3g5 c7c6 f2f3 h5f6 g5h4 f5e6 d4e6 e8e6 h4f2 f8e8 e1e6 d7e6 b2b3 e6d7');
INSERT INTO evaluations VALUES(16018,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(16018,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(16018,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(16018,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(16018,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(16018,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(16018,36,'cp 64','e7e5 e1g1 e5d4 f3d4 g7g6 c2c4 f8g7 b1c3 e8g8 c1f4 d7c5 d1d2 c5d3 d2d3 f6h5 f4e3 f7f5 e4f5 c8f5 d4f5 f8f5 a1d1 h5f4 e3f4 f5f4 d3d5 g8h8 g2g3 c7c6 d5e6 f4f6 e6e3 d8f8 f1e1 f6f5 f2f4 d6d5 c4d5 c6d5 c3d5 f8f7 e3f3 a8f8 f3g2');
INSERT INTO evaluations VALUES(15922,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(15922,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(15922,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(15922,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(15922,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(15922,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(15922,36,'cp 53','e1g1 e5d4 f3d4 g7g6 h2h3 f8g7 c2c4 e8g8 c1e3 f8e8 b1c3 a7a6 f1e1 c7c6 b2b4 c6c5 b4c5 d7c5 f2f3 c8e6 d3f1 f6h5 a1c1 g7e5 d1d2 h5g3 d4e6 f7e6 f1d3 a8c8 d3b1 d8c7 e1d1 e8d8 d2f2');
INSERT INTO evaluations VALUES(16910,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(16910,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(16910,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(16910,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(16910,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(16910,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(16910,36,'cp -94','c7c6 b5d3 d7d5 e4e5 f6d7 e2d4 d7e5 d1e2 d8f6 e1g1 f8e7 f1e1 e5d3 e2d3 e8g8 d3f3 c8d7 d2d3 f8e8 f3f6 e7f6 c1e3 c6c5 d4e2 d5d4 e3d2 a8c8 e2g3 f6e7 g1f1 f7f6 a2a4 g8f7 b2b3 g7g6 h2h4 f6f5 h4h5 e7d6 d2g5 e8e1 a1e1 b7b6');
INSERT INTO evaluations VALUES(12621,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(12621,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(12621,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(12621,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(12621,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(12621,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(12621,36,'cp -81','b5d3 d7d5 e4e5 f6d7 e2d4 d7e5 d1e2 d8f6 e1g1 f8e7 f1e1 e5d3 e2d3 e8g8 d3f3 c8d7 d2d3 f8e8 f3f6 e7f6 e1e8 a8e8 c1e3 c6c5 d4f3 d5d4 e3f4 e8e2 a1c1 d7c6 f4b8 c6f3 g2f3 f6h4 b8g3 h4e7 f3f4 e2d2 g1f1 f7f6 f4f5 g8f7 a2a4');
INSERT INTO evaluations VALUES(22796,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(22796,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(22796,32,'cp -154',NULL);
INSERT INTO evaluations VALUES(22796,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(22796,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(22796,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(22796,36,'cp -130','d7d5 b1c3 d5d4 c3e2 f7f5 g2g3 c8e6 g1h3 g8f6 c1d2 f8d6 h3f2 d8d7 f1g2 e8c8 e1g1 f5e4 d3e4 h8g8 e2c1 g7g5 c1d3 d7e7 a2a3 h7h5 b2b4 a7a6 a1b1 g5g4 f3f4 h5h4 a3a4 e5f4 d2f4 d6f4 d3f4 e6c4 f2d3 h4h3');
INSERT INTO evaluations VALUES(24999,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(24999,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(24999,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(24999,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(24999,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(24999,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(24999,36,'cp 77','f1c4 e7e6 e1g1 f8d6 h2h3 g4h5 f1e1 e8g8 g2g4 h5g6 e3e4 e6e5 d4d5 c6e7 c1g5 f6d7 c4f1 h7h6 g5e3 g6h7 a1c1 a7a6 f3d2 e7g6 c3e2 g6f4 e2f4 e5f4 e3d4 f8e8 f1g2 h7g6 d2c4 d6b4 d4c3 b4c3 c1c3 d8f6 c4d2');
INSERT INTO evaluations VALUES(25707,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(25707,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(25707,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(25707,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(25707,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(25707,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(25707,36,'cp 104','b5c4 d8e7 e1g1 g7g6 f1e1 f8g7 a2a4 g8f6 a4a5 e8g8 h2h3 h7h6 b1d2 a8e8 d4d5 c6d8 b2b4 f6h5 c4f1 f7f5 e4f5 g6f5 f3e5 g7e5 d1h5 e7g5 h5d1 g5g7 a1a3 f5f4 c3c4 d8f7 g1h1 f7g5 h3h4 g5h7 d2f3 e5b2 e1e8 f8e8 c1b2');
INSERT INTO evaluations VALUES(25790,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(25790,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(25790,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(25790,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(25790,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(25790,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(25790,36,'cp 46','h2h3 g4f3 d1f3 f8e7 d2d3 e7g5 b1c3 g5c1 a1c1 g8e7 f3e3 e8g8 d3d4 f7f6 c3e2 e7g6 c1d1 d8b8 e3b3 f8f7 b3c3 b8b5 d1d2 a7a5 a2a3 a5a4 f1e1 b5a6 d2d3 a6b5 e1b1 b5a6 b2b3 a8e8 b3a4');
INSERT INTO evaluations VALUES(25789,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25789,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(25789,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(25789,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(25789,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(25789,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(25789,36,'cp 54','g4f3 d1f3 f8e7 b1a3 c6c5 d2d3 e7f6 a3c4 g8e7 c4a5 d8d7 c1d2 a8b8 f1b1 e8g8 f3g4 d7e8 a2a3 e7c6 a5c6 e8c6 h3h4 c5c4 d2e3 a7a5 h4h5 a5a4 h5h6 g7g6 d3c4 c6c4 b2b4 a4b3 b1b3 b8b3 c2b3 c4b3 a3a4');
INSERT INTO evaluations VALUES(36011,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(36011,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(36011,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(36011,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(36011,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(36011,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(36011,36,'cp 62','f2f3 b8c6 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8b6 c3a4 b6a5 b2b3 a5c7 g2g4 f8d8 f1c4 c8e6 a4c5 d5f4 c5e6 f4e6 d4c3 e6f4 h2h4 d8d1 h1d1 a8d8 d1d8 c7d8 c3e3 d8c7 e3c5 e7e6 a2a4 h7h5 a4a5 f4g2 g4h5 g2h4 h5g6 h4g6 c1b1 g8g7 c4d3 g6f4 d3e4 f4d5 e4d5 e6d5');
INSERT INTO evaluations VALUES(43612,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(43612,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(43612,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(43612,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43612,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(43612,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(43612,36,'cp 37','d3d4 e5d4 f3d4 f8e7 e1g1 c7c6 f1e1 e8g8 c1f4 f8e8 c4f1 c6c5 d4e2 b8c6 d1d2 b7b5 e2g3 e7f8 a2a4 b5b4 c3d5 f6d5 e4d5 c6e7 c2c3 a8b8 h2h3 e7g6 e1e8 d8e8 f4e3 c8d7 f1a6 b4c3 b2c3 d7a4 a6f1 a4d7');
INSERT INTO evaluations VALUES(45467,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(45467,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(45467,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(45467,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(45467,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(45467,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(45467,36,'cp -43','b8c6 h2h3 c8f5 g1f3 h7h6 e1g1 e7e6 b1c3 f8c5 c3e2 e8g8 c1f4 b7b5 c2c3 a7a5 a2a3 d8b6 b3c2 c5e7 f4e3 b6b7 e2d4 c6d4 e3d4 f6d7 f3e5 d7e5 d4e5 f7f6 e5g3 e6e5 f1e1 b5b4 a3b4 a5b4 d1e2 e7d6');
INSERT INTO evaluations VALUES(13176,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(13176,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(13176,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(13176,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(13176,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(13176,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(13176,36,'cp -52','h2h3 h7h6 g1e2 e7e6 e1g1 f8d6 c1f4 d6f4 e2f4 e8g8 b1d2 e6e5 f4h5 f6h5 d1h5 c8e6 c2c3 f7f5 f2f4 e5e4 d3e4 f5e4 a1e1 e6f7 h5e2 d8f6 f1f2 a8d8 d2f1 g8h8 e2e3 f7g6 f2d2 d5d4 c3d4 c6d4');
INSERT INTO evaluations VALUES(52843,30,'cp -195',NULL);
INSERT INTO evaluations VALUES(52843,31,'cp -203',NULL);
INSERT INTO evaluations VALUES(52843,32,'cp -199',NULL);
INSERT INTO evaluations VALUES(52843,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(52843,34,'cp -211',NULL);
INSERT INTO evaluations VALUES(52843,35,'cp -197',NULL);
INSERT INTO evaluations VALUES(52843,36,'cp -198','d5e4 f2f3 g8f6 b1c3 e4f3 g1f3 c8f5 d2d4 e7e6 c1f4 f8d6 d1d2 e8g8 a4b3 d8c7 f4d6 c7d6 e1c1 b8d7 f3h4 f5e4 c3e4 f6e4 d2e2 d6f4 c1b1 f4h4 g2g3 e4g3 h2g3 h4g5 d1g1 h7h6 c2c3 a8d8 b3c2 d7f6 g3g4 c6c5 d4c5 g5c5');
INSERT INTO evaluations VALUES(52934,30,'cp -198',NULL);
INSERT INTO evaluations VALUES(52934,31,'cp -193',NULL);
INSERT INTO evaluations VALUES(52934,32,'cp -209',NULL);
INSERT INTO evaluations VALUES(52934,33,'cp -210',NULL);
INSERT INTO evaluations VALUES(52934,34,'cp -200',NULL);
INSERT INTO evaluations VALUES(52934,35,'cp -193',NULL);
INSERT INTO evaluations VALUES(52934,36,'cp -194','f2f3 c8f5 b1c3 e4f3 g1f3 e7e6 d2d4 g8f6 c1f4 f8d6 d1d2 e8g8 a4b3 a7a5 a2a4 d6b4 d2e2 f6d5 f4d2 b8d7 e1g1 d5f6 d2f4 f5g6 h2h3 d8b6 g1h1 f6d5 c3d5 c6d5 c2c4 b4d6 f4d6 b6d6 a1c1 d5c4 b3c4 d7f6 e2e3 f6d5 c4d5 e6d5');
INSERT INTO evaluations VALUES(56729,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(56729,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(56729,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(56729,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(56729,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(56729,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(56729,36,'cp 135','c8f5 e1g1 f8e7 b1c3 c7c5 c3d5 b8c6 f1e1 g8f6 c2c3 d8d7 d2d4 f6d5 c4d5 e8c8 h2h3 e7f6 f3g5 c8b8 g5f7 e5d4 f7h8 d8h8 d1b3 a7a6 c1f4 b8a7 d5e6 f5e6 e1e6 d6d5 a1e1 c5c4 b3a3 h8c8 c3d4 f6d4 a3d6 c8d8 d6d7 d8d7 f4d2 d4b6');
INSERT INTO evaluations VALUES(32434,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(32434,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(32434,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(32434,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(32434,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(32434,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(32434,36,'cp 154','e1g1 f8e7 b1c3 c7c5 f1e1 g8f6 c3d5 b8c6 c2c3 f6d5 c4d5 d8d7 d2d4 e8c8 h2h3 e5d4 d5c6 b7c6 c3d4 e7f6 d4c5 h8e8 c5d6 e8e1 f3e1 d8e8 e1f3 f5e4 f3g5 e4d3 d1d3 e8e1 g1h2 f6g5 c1g5 e1a1 g5f4 a1a2 d3h7');
INSERT INTO evaluations VALUES(5597,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(5597,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(5597,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(5597,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(5597,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(5597,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(5597,36,'cp -26','b1c3 d6e7 c4b5 c8d7 d2d3 c6d4 b5d7 d8d7 e1g1 g8f6 c1f4 b7b6 f4e5 d4c6 e5g3 e8g8 f3e5 c6e5 g3e5 d7c6 d1e2 a8d8 a2a4 f8e8 f1e1 f6d7 e5g3 d7f8 c3b5 d8d7 e2e4 c6e4 d3e4 e8d8 g3c7 d8c8 c7e5 f8g6 e5c3');
INSERT INTO evaluations VALUES(6504,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(6504,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(6504,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6504,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6504,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(6504,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6504,36,'cp 6','g8f6 f3e5 f6e4 e5f3 d7d5 d2d3 e4d6 d3d4 c5b6 c2c3 c6e7 a2a4 c7c6 c1f4 e7g6 f4g3 d6e4 b1d2 e8g8 d2e4 d5e4 f3e5 g6e5 g3e5 b6c7 e5c7 d8c7 f1e1 a7a5 d1d2 c8f5 d2e3 a8d8 a1b1 f8e8 e3g3 c7g3 h2g3 f5e6');
INSERT INTO evaluations VALUES(10254,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(10254,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(10254,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(10254,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(10254,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(10254,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(10254,36,'cp 10','f3e5 f6e4 e5f3 d7d5 d2d3 e4d6 d3d4 c5b6 c1f4 c6e7 a2a4 a7a5 c2c3 c7c6 e2d3 e8g8 b1d2 c8f5 d1c2 f5d3 c2d3 e7g6 f4d6 d8d6 g2g3 b6c7 a1e1 a8e8 d2b3 d6d7 e1e8 f8e8 f1e1 e8e1 f3e1 d7e8 d3d1 g6f8 e1d3 b7b6 b3d2 h7h6 d2f1 f8e6');
INSERT INTO evaluations VALUES(6536,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(6536,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(6536,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(6536,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(6536,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(6536,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(6536,36,'cp -4','d7d6 e1g1 g8e7 c3a4 e8g8 a4c5 d6c5 c1f4 c8g4 d1d8 a8d8 f4c7 d8d7 f3e5 d7c7 e5g4 a7a6 f1d1 b7b5 c4f1 f7f5 e4f5 e7f5 a1c1 c5c4 b2b3 h7h5 g4e3 f5e3 f2e3 f8c8 b3c4 c6e5 f1e2 g7g6 c4c5 c7c5 d1d8 c8d8 c1c5');
INSERT INTO evaluations VALUES(6036,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(6036,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(6036,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(6036,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(6036,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(6036,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(6036,36,'cp -10','e1g1 g8f6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 b2b4 c5b6 a2a4 c6e5 f3e5 d6e5 d1h5 b6d4 d5b6 e8g8 b6a8 d4a1 f1a1 d8d4 a1c1 c8e6 c4d5 f8a8 h2h3 e6d5 e4d5 c7c6 d5c6 b7c6 b4b5 c6b5 a4b5 d4b2 c1c7 a8f8 c7a7 b2c1 g1h2 c1f4 h2g1 f8c8 a7a1 c8c2 a1f1 c2b2 h5d1 f4c4 d1c1 c4c1 f1c1 b2b5 f2f3');
INSERT INTO evaluations VALUES(6035,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(6035,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(6035,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(6035,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(6035,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(6035,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(6035,36,'cp -16','g8e7 c3a4 e8g8 a4c5 d6c5 c1f4 c8d7 a1c1 b7b6 h2h3 d7e6 c4e6 f7e6 d1d8 a8d8 f4c7 d8d7 c7g3 c6d4 f3d4 d7d4 f2f3 e7c6 f1d1 f8d8 d1d4 d8d4 g3e1 c6b4 e1b4 d4b4 b2b3 b4d4 c1c2 g7g5 g1f2 g8f7 f2e3 e6e5 c2c1 h7h5 g2g3 a7a5 h3h4 g5h4 g3h4 a5a4 c1c4 a4b3');
INSERT INTO evaluations VALUES(9309,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(9309,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(9309,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(9309,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(9309,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(9309,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(9309,36,'cp -30','h7h6 g5d2 a7a6 b1c3 e8g8 e1g1 c6a5 c4b3 f8e8 d2e3 c5b4 e3d2 a5b3 a2b3 b7b6 c3e2 b4d2 d1d2 d6d5 e2g3 c8b7 f1e1 c7c5 e4d5 d8d5 b3b4 c5b4 d2b4 d5c6 b4c4 c6c4 d3c4 b7f3 g2f3 e8e6 g1f1 g8h7 b2b3 a6a5 h3h4 a8d8 e1d1 d8d1 a1d1 g7g6 g3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(8811,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(8811,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(8811,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(8811,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(8811,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(8811,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(8811,36,'cp -12','g5d2 a7a6 b1c3 c6a5 c4b3 e8g8 e1g1 f8e8 d2e3 c5e3 f2e3 b7b5 c3e2 c8d7 d1e1 c7c6 f3h4 a5b3 a2b3 a6a5 h4f5 d7f5 f1f5 e8e6 e1h4 d8b6 h4g3 a5a4 e2c3 a4a3 b3b4 a3a2 c3a2 d6d5 a1f1');
INSERT INTO evaluations VALUES(17847,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(17847,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(17847,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(17847,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(17847,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(17847,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(17847,36,'cp 15','f8c5 c3d5 d7d6 a2a4 g8f6 d2d3 f6d5 c4d5 e8g8 c1e3 c5e3 f2e3 c6e7 d5b3 c8e6 b3e6 f7e6 d1d2 d8d7 b2b4 e7g6 a4a5 d7e7 h2h3 f8f7 f1f2 a8f8 a1f1 g6h4 f3h4 f7f2 f1f2 e7h4 f2f8 g8f8 c2c4 h4g3 d2f2 g3f2 g1f2 f8f7 f2f1 f7e7 h3h4 h6h5 f1e2 e7f7');
INSERT INTO evaluations VALUES(4820,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4820,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(4820,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(4820,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4820,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(4820,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4820,36,'cp 12','c3d5 g8f6 d2d3 d7d6 c1e3 c5e3 f2e3 e8g8 d5f6 d8f6 d1e1 c8e6 f3d4 f6e7 d4c6 b7c6 c4e6 e7e6 b2b3 d6d5 e1c3 f8d8 h2h3 d8d7 a2a4 e6d6 a1d1 f7f6 g1h2 g8h8 c3a5 h8g8 a5e1 a8d8 e1g3 d5e4 d3e4 d6e6 d1d7 d8d7 f1f2');
INSERT INTO evaluations VALUES(29251,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(29251,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(29251,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(29251,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(29251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29251,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(29251,36,'cp -14','b8d7 f1d3 f5g6 d3e2 d5c4 e2c4 d8c7 f3h4 f8e7 e1g1 e8g8 c4e2 f8d8 h4g6 h7g6 d1c2 c6c5 f1d1 c5d4 d1d4 a8c8 d4c4 c7b8 c1d2 c8c4 e2c4 a7a6 c4f1 d8c8 a1c1 c8c7 c2b1 b7b5 d2e1 c7c8 c3e4 c8c1 e4f6 g7f6 b1c1');
INSERT INTO evaluations VALUES(24498,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(24498,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(24498,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(24498,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(24498,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(24498,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24498,36,'cp 0','f3h4 f5g4 d1b3 a8b8 h2h3 g4h5 c4d5 f6d5 h4f3 f8d6 c1d2 h5g6 f1e2 e8g8 e1g1 f8e8 f1d1 d6c7 a1c1 h7h6 d2e1 d8e7 e2d3 g6h5 d3e2 h5g6');
INSERT INTO evaluations VALUES(29967,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(29967,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(29967,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(29967,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(29967,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(29967,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(29967,36,'cp 167','b8c6 d4c6 g4e2 d1e2 b7c6 e2a6 d8d7 a6b7 a8c8 e1g1 a7a5 b7a7 f8e7 a7a5 e8g8 a5f5 d7f5 e4f5 d6d5 f1d1 c8a8 a2a4 e7d6 g1f1 d6e5 c3e2 f8b8 a1a2 c6c5 b2b3 d5d4 a4a5 a8a6 a2a4 f6d5 c1d2 d4d3 c2d3 b8b3 d3d4 c5d4 e2d4 b3b2 d4f3');
INSERT INTO evaluations VALUES(41109,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(41109,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(41109,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(41109,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(41109,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(41109,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(41109,36,'cp -49','d2d4 e8g8 e2d1 c7c6 f1e2 c8f5 e1g1 b8d7 f3h4 f5e6 e2d3 f6e4 h4f3 e4c3 b2c3 c6c5 d4c5 d7c5 f1e1 a8c8 f3d4 e7f6 c1d2 f8e8 a1b1 e6d7 d1f3 e8e1 d2e1 b7b6 e1d2 c5e4 d3e4 d5e4 f3e4 d7c6');
INSERT INTO evaluations VALUES(42726,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(42726,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(42726,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(42726,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(42726,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(42726,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(42726,36,'cp -90','f6e4 d1d3 d7d5 d4e5 e8g8 a2a3 b4c3 b2c3 c8f5 d3e3 f7f6 c3c4 f6e5 c4d5 d8d5 e3b3 f5e6 b3d5 e6d5 c1b2 b8c6 f1e2 a8e8 e1g1 e4d6 f1e1 d5f3 e2f3 e5e4 f3g4 c6e5 g4e2 e5c4 b2c3 a7a5 a3a4 b7b6 a1d1 e4e3 f2f3');
INSERT INTO evaluations VALUES(48150,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(48150,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(48150,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48150,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(48150,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(48150,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(48150,36,'cp -16','g7f6 g1f3 f6f5 e2e3 c7c5 g2g3 b8d7 f1b5 a7a6 b5d7 c8d7 c3e2 d8b6 a2a4 f8d6 c2c3 c5c4 a1b1 b6c7 e1g1 a6a5 e2f4 h7h5 h2h4 d7c6 b1a1 e8c8 g1h2 c8b8 d1e2 d8g8 f1b1 b8a7 b2b3 c4b3 b1b3');
INSERT INTO evaluations VALUES(50924,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(50924,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(50924,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(50924,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(50924,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(50924,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(50924,36,'cp 68','d2d4 e5d4 f3d4 g7g6 b1c3 f8g7 f1e2 g8f6 c1g5 h7h6 g5e3 d8e7 d1d2 f6e4 c3e4 e7e4 e1c1 e4h4 g2g3 h4f6 c2c3 e8g8 e3h6 g7h6 d2h6 b8c6 d4c6 b7c6 h3h4 f8e8 d1d2 f6f2 h4h5 f2e3 h5g6 e3h6 g6f7 g8f7 h1h6 f7g7 h6h5 e8h8 h5h8 g7h8');
INSERT INTO evaluations VALUES(53589,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(53589,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53589,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(53589,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(53589,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(53589,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(53589,36,'cp 34','e2e4 g8f6 f1d3 d5e4 c3e4 d8d4 g1f3 d4d5 e4f6 g7f6 e1g1 b8d7 c2c4 d5h5 d3e2 h8g8 f3d4 h5c5 f4g3 e6e5 d4f5 d7b6 d1c2 b6c4 e2c4 c8f5 c2f5 c5c4 f5f6 f8g7 f6g5 h7h6 g5e3 e8c8 e3a7 c4a6 a7a6 b7a6 a1c1 c8b7 c1c2 d8d3 g3h4 f7f6 f2f4 e5e4 f4f5 e4e3 c2e2');
INSERT INTO evaluations VALUES(53617,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(53617,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53617,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(53617,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(53617,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(53617,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(53617,36,'cp 20','g8f6 e4e5 f6d7 g1f3 c6c5 d4c5 d7c5 f1e2 f8e7 e1g1 e8g8 a2a3 a7a5 c3b5 c8d7 a3a4 d7b5 e2b5 b8c6 c2c3 d8b6 g2g3 c6a7 f3d4 a7b5 d4b5 a8d8 f1e1 c5e4 f4e3 e7c5 e3d4 c5d4 d1d4 b6d4 b5d4 d8c8');
INSERT INTO evaluations VALUES(56876,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(56876,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(56876,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(56876,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(56876,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(56876,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(56876,36,'cp 99','f8g7 d2d4 d7d6 c1f4 a7a6 d1d2 h7h6 h2h3 g8e7 f1d3 b8c6 f4e3 e6e5 d3e2 f7f5 e4f5 e7f5 d4e5 d6e5 d2d8 c6d8 e1c1 c8d7 c3d5 d8e6 g2g4 f5e3 f2e3 g6g5 e2c4 a8d8 c4d3 d7c6 d3e4 c6d5 e4d5 e6c5 b2b4 e5e4 b4c5 e4f3 d5b7');
INSERT INTO evaluations VALUES(648,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(648,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(648,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(648,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(648,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(648,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(648,36,'cp -50','h1e1 f8e7 b1c3 f6d5 c3d5 c6d5 b2b3 f7f6 a2a4 e8f7 c1a3 e7a3 a1a3 h8e8 e2f1 e8e1 f3e1 a6a5 c2c3 c8f5 b3b4 f7e7 d2d4 e7d6 f2f3 f5d7 b4b5 c7c6 b5c6 d7c6 f1e2 b7b5 a4b5 c6b5 e2d2 h7h5 g2g3 a5a4 e1c2 h5h4 g3h4');
INSERT INTO evaluations VALUES(647,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(647,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(647,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(647,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(647,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(647,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(647,36,'cp -32','f6d5 d2d3 f8e7 a2a3 e8g8 e2f1 f8e8 c1d2 c8g4 f3e5 g4f5 b1c3 f7f6 e5c4 f5d7 h2h4 b7b6 c4e3 d5c3 d2c3 g8f7 f2f3 h7h5 e3c4 e7f8 c4d2 e8e1 a1e1 c6c5 d2e4 d7a4 b2b3 a4d7 e1a1 a8d8 a3a4');
INSERT INTO evaluations VALUES(14632,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(14632,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(14632,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14632,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(14632,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(14632,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(14632,36,'cp 50','d2d4 c8g4 c4b5 f6d7 c2c3 f8e7 h2h3 g4h5 c1e3 e7g5 b5c6 b7c6 b1d2 g5e3 e1e3 e8g8 d1a4 e5d4 c3d4 c6c5 a4a5 f8e8 a1c1 h5g6 b2b3 c7c6 a5c3 c5d4 c3d4 d8c7 d2c4 e8e6 f3h4 c6c5 d4d3 a8e8 h4g6 e6g6 c1d1 d7e5 c4e5 e8e5 f2f4 e5e7 d3c3 e7e8 e3e1 a7a5');
INSERT INTO evaluations VALUES(14649,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14649,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(14649,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(14649,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(14649,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(14649,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14649,36,'cp 44','c8g4 c4b5 f6d7 c2c3 f8e7 h2h3 g4h5 c1e3 e7g5 b5c6 b7c6 b1d2 g5e3 e1e3 e8g8 d1a4 e5d4 c3d4 c6c5 a4a5 f8e8 a1e1 h5f3 d2f3 c5d4 f3d4 c7c5 a5d8 a8d8 d4b5 d7b6 e1d1 b6c4 e3e2 d8b8 b5c3 c4b2 d1d6 e8c8 d6d7 a7a5 d7a7 c5c4 a7a5 b8b7 g1h2');
INSERT INTO evaluations VALUES(30852,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(30852,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(30852,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(30852,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(30852,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(30852,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(30852,36,'cp 105','d8d7 d4e2 b8c6 b1c3 g8e7 e1g1 a8d8 c3d5 e7d5 d1d5 f8e7 c2c4 f7f5 e4f5 c6b4 d5d1 d7f5 a2a3 b4d3 e2g3 f5g6 d1f3 d6d5 c4d5 h8f8 f3d1 d8d5 c1e3 d5d7 d1b3 g6c6 a1d1 d3f4 e3f4 e5f4 g3h5 f8f7');
INSERT INTO evaluations VALUES(32749,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(32749,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(32749,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(32749,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(32749,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(32749,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(32749,36,'cp 90','d4e2 b8c6 b1c3 g8e7 e1g1 a8d8 c3d5 e7d5 d1d5 c6b4 d5b3 d6d5 a2a3 b4a6 b3g3 f7f6 e4d5 d7d5 c1e3 f8c5 e2c3 d5f7 e3c5 a6c5 f2f4 e8g8 f4e5 f7c7 e5f6 c7g3 h2g3 f8f6 f1f6 g7f6 a1f1 g8f7 g3g4 f7e6 f1e1 e6f7 g1f2 d8d2 e1e2 d2d4 c3b5 d4g4 b5d6 f7g6 b2b4 c5d7 e2d2 d7e5 d6b7 e5c4 d2d3 c4e5 d3c3 g4d4');
INSERT INTO evaluations VALUES(39231,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(39231,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(39231,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(39231,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(39231,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(39231,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(39231,36,'cp -79','c2c3 d7d5 d2d4 e5d4 c3d4 c5b6 e4d5 g8e7 g1f3 e8g8 a2a4 c7c6 d5d6 e7d5 f1e2 f8e8 e1g1 d8d6 f1e1 c8g4 b1d2 b8d7 d2c4 d6g6 c4b6 a7b6 c1d2 d7f8 a1c1 a8d8 f3e5 g4e2 d1e2 g6f5 e2c4 f8e6 h2h3 h7h6 c1d1 b6b5 a4b5 c6b5 c4d3 f5d3 e5d3 e6d4 e1e8 d8e8');
INSERT INTO evaluations VALUES(43218,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(43218,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(43218,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(43218,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(43218,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(43218,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(43218,36,'cp -28','g1f3 d7d5 f1d3 c7c5 e1g1 e8g8 c2c3 d8c7 d1e2 c8f5 d3f5 g6f5 c1d2 e7e6 d2e1 c7b6 a2a4 f6e4 e1h4 b8d7 f1c1 c5c4 c1c2 g7f6 h4e1 f6e7 b1d2 e4d6 f3e5 b6c7 g1h1 g8g7 b2b3 c4b3 d2b3');
INSERT INTO evaluations VALUES(43219,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(43219,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(43219,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(43219,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(43219,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(43219,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(43219,36,'cp -35','d7d5 f1d3 c7c5 e1g1 e8g8 c2c3 d8c7 b1d2 c8f5 d3f5 g6f5 a2a4 c5c4 f3e5 b8c6 b2b3 c4b3 c1a3 f8c8 d1b3 c6a5 b3c2 e7e6 a1c1 c7d8 c2d3 c8c7 a3b4 a5c6 b4a3 g7f8 c3c4 f8a3 d3a3 a8c8');
INSERT INTO evaluations VALUES(41483,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(41483,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(41483,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(41483,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(41483,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(41483,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(41483,36,'cp -23','f1d3 c7c5 e1g1 e8g8 c2c3 c8f5 d3f5 g6f5 d4c5 e7e6 f3d4 d8c7 b2b4 b7b6 c5b6 a7b6 c1b2 b8c6 d4b5 c7d7 a2a4 c6a7 b5d4 a7c8 d4b5 f6e4 b1d2 c8d6 b5d6 e4d6 f1f3 f8d8 f3g3 d6c4 d2c4 d5c4 d1h5 a8a4 g3g7 g8g7 h5g5 g7f8 a1a4');
INSERT INTO evaluations VALUES(41482,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(41482,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(41482,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(41482,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(41482,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(41482,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(41482,36,'cp -20','e8g8 e1g1 c7c5 c2c3 d8c7 d1e2 c8f5 d3f5 g6f5 b1d2 e7e6 a2a4 c5c4 g1h1 b8c6 f1g1 c6e7 a1b1 f8b8 f3e5 b7b5 a4b5 b8b5 g2g4 f6g4 e5g4 f5g4 e2g4 e7f5 b1a1 g8h8 d2f3 a8g8 f3e5 c7b7 g4h3 b5b6');
INSERT INTO evaluations VALUES(36762,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(36762,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(36762,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(36762,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(36762,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(36762,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(36762,36,'cp -21','b1d2 c7c5 c2c3 d8c7 e1g1 c8f5 d3f5 g6f5 a2a4 e7e6 f3e5 f8c8 d1e2 b8c6 f1f3 g8h8 e2e1 c8g8 e1h4 a8f8 f3g3 h7h6 b2b3 f8c8 g1h1 b7b6 g3h3 h8h7 d2f3 f6e4 f3g5 e4g5 h4g5 g7f8 g5f6 c6e5');
INSERT INTO evaluations VALUES(36763,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(36763,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(36763,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(36763,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(36763,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(36763,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(36763,36,'cp -29','c7c5 c2c3 d8c7 e1g1 c8f5 d3f5 g6f5 f3e5 b8c6 b2b3 c5d4 c3d4 f8c8 a2a4 c7d8 c1a3 c8c7 b3b4 a8c8 h2h3 e7e6 d1e1 g7f8 b4b5 c6e7 a4a5 e7g6 a5a6 f8a3 a1a3');
INSERT INTO evaluations VALUES(51507,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(51507,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(51507,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(51507,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(51507,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(51507,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(51507,36,'cp -19','g8f6 b2b3 c8b7 e2e3 g7g6 c1b2 f8g7 g2g3 c7c5 f1g2 e8g8 c2c4 d7d5 e1g1 b8c6 d1e2 f6e8 f3e5 d5c4 b3c4 c6e5 f4e5 b7g2 g1g2 f7f6 d2d4 d8d7 d4d5 e7e6 e3e4 f6e5 f1f8 g8f8 b1d2 e8d6 a1d1 a8d8 d2f3');
INSERT INTO evaluations VALUES(54284,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(54284,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(54284,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(54284,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(54284,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(54284,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(54284,36,'cp -145','b8c6 f1b5 c5d4 g4d4 c8d7 b5c6 b7c6 g1f3 c6c5 d4d3 g8e7 e1g1 e7g6 f1e1 d7c6 c1g5 d8b6 h2h4 h7h6 g5c1 f8e7 h4h5 g6f8 c2c4 d5d4 b1d2 f8d7 d2e4 e8g8 c1f4 a7a5 f3d2 a5a4 a1d1 b6c7 d3g3 g8h8');
INSERT INTO evaluations VALUES(20324,30,'cp -126',NULL);
INSERT INTO evaluations VALUES(20324,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(20324,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(20324,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(20324,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(20324,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(20324,36,'cp -115','f1b5 c5d4 g4d4 c8d7 b5c6 b7c6 g1f3 c6c5 d4d3 g8e7 e1g1 e7g6 c2c4 d5d4 b2b4 c5b4 b1d2 f8e7 d2e4 e8g8 c1b2 d7c6 b2d4 d8c7 d3e3 a7a5 a1c1 f8b8 f1d1 h7h6 e4d6 a5a4 d6b5 c7a5 b5d6 b4b3 a2b3 b8b3');
INSERT INTO evaluations VALUES(55213,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(55213,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(55213,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(55213,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(55213,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(55213,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(55213,36,'cp 4','g8f6 e2e3 a7a6 f1c4 b7b5 c4b3 e7e6 e1g1 c8b7 d2d4 c7c5 e3e4 c5d4 f3d4 b8c6 d4c6 d8d1 f1d1 b7c6 f2f3 f8c5 g1f1 e8e7 c3e2 h8c8 e2d4 c6d7 e4e5 f6d5 c1g5 e7f8 a1c1 b5b4 b3d5 e6d5 d4b3 d7b5 f1e1 c5b6 d1d5 f8e8 g5d2 c8c1 d2c1');
INSERT INTO evaluations VALUES(47342,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(47342,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(47342,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(47342,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(47342,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47342,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(47342,36,'cp 8','e2e3 a7a6 f1c4 b7b5 c4b3 e7e6 e1g1 c7c5 d2d4 c8b7 e3e4 c5d4 f3d4 b8c6 d4c6 d8d1 f1d1 b7c6 f2f3 f8c5 g1f1 e8e7 e4e5 f6d7 c1g5 e7e8 a1c1 h7h6 g5d2 c5b6 c3e2 a8c8 d2b4 a6a5 d1d6 c6b7 c1c8 b7c8 b4a3 b5b4 b3a4 e8e7');
INSERT INTO evaluations VALUES(10551,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(10551,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(10551,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(10551,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(10551,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10551,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10551,36,'cp 10','g5f3 e8g8 b1c3 h6g4 e1g1 d7d6 d2d3 c6d4 h2h3 g4f6 c1e3 d4f3 d1f3 c5e3 f3e3 c8e6 c4e6 f7e6 c3e2 f6h5 c2c3 d8f6 d3d4 g7g5 f2f3 a7a5 a1d1 h5f4 e2f4 g5f4 e3e2 g8h8 d4e5 f6e5 e2c4 f8f7 c4d4 e5g7 d4g7 f7g7');
INSERT INTO evaluations VALUES(18958,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(18958,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(18958,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(18958,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(18958,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(18958,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(18958,36,'cp 155','c1e3 g8f6 f1c4 d7d6 b1c3 c8e6 c4e6 f7e6 c5d6 f8d6 a2a3 e8g8 g1f3 a7a6 d1d3 d8e7 a1d1 a8d8 d3e2 d6c5 e1g1 c5d4 d1d3 h7h6 h2h3 e7c7 e2d2 d4c3 d2c3 f6e4 c3c4 d8d3 c2d3 e4f6 c4e6 c7f7 e6c4 f7c4 d3c4 f8d8 f1c1 d8d3 f3d2 c6d4 g1f1 d4b3 d2b3 d3b3');
INSERT INTO evaluations VALUES(18956,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(18956,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(18956,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(18956,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(18956,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(18956,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(18956,36,'cp 134','g8f6 f1c4 d7d6 b1c3 c8e6 c4e6 f7e6 c5d6 f8d6 a2a3 e8g8 g1f3 d8c7 d1e2 a7a6 a1d1 f6g4 e3g5 h7h6 g5h4 g4f6 h4g3 a8c8 c3a2 c6a5 e1g1 a5c4 c2c3 f6h5 a2b4 a6a5 b4d3 c8d8 d1c1 h5g3 h2g3 c7b6 g1h1 b6b3 f1d1 a5a4');
INSERT INTO evaluations VALUES(23486,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(23486,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(23486,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(23486,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(23486,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(23486,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(23486,36,'cp 60','e5g6 b1c3 e7e6 d5e6 d7e6 d1d8 e8d8 c1d2 g8f6 e2e4 f8b4 g2g3 b7b6 e1c1 c8b7 f1g2 d8e7 d1e1 h8d8 a2a3 b4c5 g2f3 e6e5 f4f5 g6f8 g3g4 f8d7 h2h4 f6e8 g4g5 e8d6 h4h5 d6c4 c3d5 e7f8 d2b4 c5b4 d5b4 a7a5 b4d3 d7c5 d3c5');
INSERT INTO evaluations VALUES(23477,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(23477,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(23477,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(23477,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(23477,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(23477,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(23477,36,'cp 86','e2e4 e7e5 d5e6 d7e6 d1d8 e8d8 g1f3 f8b4 c2c3 b4c5 b1d2 b7b6 d2b3 c5d6 f1c4 g8h6 e1g1 c7c5 g2g3 f7f6 a2a4 c8d7 c1e3 h6g4 e3d2 g4h6 h2h3 h6f7 d2e3 d8e7 a4a5 d7c6 b3d2 h8d8 c4e2 d6c7 a5b6 a7b6 g1f2');
INSERT INTO evaluations VALUES(25202,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(25202,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(25202,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(25202,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(25202,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(25202,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(25202,36,'cp 45','e1g1 f8d6 f4g5 h7h6 g5f6 d8f6 c2c4 d5c4 d3c4 f6g6 b1d2 e8g8 c4b3 a8b8 d2c4 f8c8 a1c1 g6h5 c1c3 c6a5 c4a5 h5a5 f1c1 a5d5 b3d5 e6d5 f3e1 d6b4 c3c2 b4e1 c1e1 c7c6 e1c1 a7a6');
INSERT INTO evaluations VALUES(37178,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37178,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(37178,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37178,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(37178,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37178,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37178,36,'cp 0','d7d5 a2a3 b4c3 c2c3 d5c4 c3c4 b7b6 c1g5 c8a6 c4a4 h7h6 g5h4 c7c5 d4c5 b6c5 a1d1 d8b6 h4f6 g7f6 a4g4 g8h7 g4h4 h7g7 h4g4');
INSERT INTO evaluations VALUES(37211,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(37211,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(37211,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(37211,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(37211,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(37211,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(37211,36,'cp 162','d2d4 c5b4 c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g2f3 b8c6 e5c6 b7c6 e1c1 b4d6 b2b3 g7g6 f1c4 a8b8 c1b2 g8g7 h2h4 h6h5 c3e2 d6e7 e2f4 c8b7 c2c3 b8d8 b3b4 a7a5 h1g1 a5b4 c3b4 e7b4 f4h5 g7h8');
INSERT INTO evaluations VALUES(37210,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(37210,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(37210,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(37210,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(37210,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(37210,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(37210,36,'cp 129','c5b4 c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g2f3 b8c6 e5c6 b7c6 e1c1 a8b8 h2h4 b4e7 b2b3 d7d6 f1c4 c8b7 f3f4 d6d5 e4d5 f8d8 d5c6 b7c6 d4d5 c6d7 c3e4 d7g4 d1d3 g4f5 d3e3 e7a3 c1b1 g8f8 e4g3 f5g4 f2f3 g4c8 h4h5 a7a5 h1e1 a3d6 e3e4 a5a4 b1b2');
INSERT INTO evaluations VALUES(38715,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38715,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(38715,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(38715,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(38715,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(38715,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(38715,36,'cp 45','f1g2 g8f6 b2b4 c5b6 g1f3 e8g8 e1g1 c7c6 d2d3 h7h6 a2a4 a7a5 b4b5 b6c5 d1b3 d8c7 c1a3 c5a3 b3a3 c8e6 c4c5 d6c5 a3c5 b8d7 c5a3 f6d5 f1c1 d5b4 c3e4 f8e8 f3d2 e6d5 b5c6 b7c6 d2c4 d5e4 g2e4 d7c5 e4g2');
INSERT INTO evaluations VALUES(43825,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43825,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(43825,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43825,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(43825,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(43825,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(43825,36,'cp 7','f1d3 d7d5 g1f3 c8b7 e1g1 f8d6 b2b3 e8g8 b1d2 c7c5 c4d5 f6d5 d4c5 d6c5 d2c4 b8d7 c1b2 d8e7 d1e2 a7a5 a2a3 b6b5 c4e5 d7e5 f3e5 b5b4 a3b4 a5b4 a1a8 f8a8 e2c2 h7h6 f1c1 c5b6 e5c6 e7e8 d3b5 b7c6');
INSERT INTO evaluations VALUES(45397,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45397,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45397,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45397,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(45397,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(45397,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(45397,36,'cp 0','e7e6 c2c4 b8c6 c4d5 e6d5 f1b5 c8g4 d2d4 a8c8 b1c3 c5d4 d1d4 f8e7 f3e5 g4e6 e1g1 e8g8 b5c6 b7c6 d4a7 e7d6 c3a4 c6c5 a7b6 d8e7 a1d1 c8c7 e5c6 e7d7 c6b8 d7e7');
INSERT INTO evaluations VALUES(47035,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47035,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47035,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(47035,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47035,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47035,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47035,36,'cp 0','c7c5 g2g3 e7e6 f1g2 b8c6 f2f4 a8b8 a2a4 g7g6 g1h3 f8g7 e1g1 e8g8 h3f2 a7a6 c3e2 b7b5 a4b5 a6b5 g3g4 h7h6 c1e3 b5b4 d1d2 c8b7 c2c3 c6a5 d2d1 a5c6');
INSERT INTO evaluations VALUES(47335,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(47335,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(47335,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(47335,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(47335,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(47335,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(47335,36,'cp -107','b8c6 d4d5 c6a5 d1e2 a5c4 e2c4 a7a6 c1g5 b7b5 c4d4 c8b7 e1c1 d8d6 g1e2 e8c8 a2a3 h7h6 g5f6 d6f6 d4e3 f6g5 e2g3 g5e3 f2e3 e7e6 d5e6 f7e6 d1d8 c8d8 g3e4 f8e7 h1d1 d8e8 d1f1 h6h5 b2b4 h8f8 f1f8 e8f8 c1d2 f8f7 d2d3');
INSERT INTO evaluations VALUES(14369,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(14369,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(14369,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(14369,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(14369,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(14369,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(14369,36,'cp -97','d4d5 c6a5 d1e2 a5c4 e2c4 a7a6 c1g5 b7b5 c4d4 c8b7 e1c1 d8d6 f2f3 h7h6 g5f6 d6f6 f3e4 f6d4 d1d4 e7e6 g1f3 f8c5 d4d3 e6d5 c3d5 e8c8 h1d1 h8e8 a2a3 c7c6 d5c3 c8c7 d3d8 e8d8 d1d8 c7d8 c1d2 d8e7 d2d3 c5b6 a3a4 e7d6 b2b4 c6c5 b4c5 d6c5 a4b5 a6b5 c3e2');
INSERT INTO evaluations VALUES(48491,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(48491,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48491,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(48491,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48491,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48491,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48491,36,'cp 9','a7a6 f3e5 f8b4 b1c3 f6d5 e1g1 e8g8 c3e4 b7b5 b2b3 c4b3 a2b3 c8b7 e5d3 b8d7 d3b4 d5b4 c1a3 a6a5 d1d2 c7c5 f1d1 c5d4 a3b4 a5b4 d2d4 a8a1 d1a1 b7e4 g2e4 d7f6 d4b4 f6e4 b4e4 h7h6 b3b4 d8d2 a1a5 d2d7 e4e5 d7d1 g1g2 f8c8 e5b5 c8c1 a5a8 g8h7 b5d3 d1d3 e2d3 c1b1');
INSERT INTO evaluations VALUES(50484,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(50484,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(50484,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(50484,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(50484,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(50484,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(50484,36,'cp -29','b7b5 c4b3 b8c6 g1f3 c6a5 e1g1 c8b7 c3e2 f8d6 c2c4 d6e7 e2c3 b5b4 c3e2 a5b3 a2b3 d7d6 f1e1 g8f6 e2g3 e8g8 d3d4 c5d4 f3d4 f8e8 h2h3 f6d7 c1d2 a6a5 d1g4 d7e5 g4e2 e7h4 d4b5 h4g3 f2g3 e5d7 d2f4 e6e5 f4e3 d7c5 e3c5 d6c5 e2e3 d8b6');
INSERT INTO evaluations VALUES(51196,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(51196,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(51196,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(51196,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(51196,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(51196,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(51196,36,'cp -31','c4b3 b8c6 g1f3 c6a5 e1g1 c8b7 f1e1 f8e7 h2h3 d7d6 c3e2 g8f6 e2g3 e8g8 c2c3 a5b3 a2b3 f8e8 c1d2 a6a5 d3d4 f6d7 d2e3 c5d4 f3d4 b5b4 c3c4 d7c5 d1g4 e7f8 a1d1 d8f6 d4b5 c5b3 b5d6 f8d6 d1d6 f6g6');
INSERT INTO evaluations VALUES(51213,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(51213,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(51213,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(51213,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(51213,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(51213,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(51213,36,'cp -29','b8c6 g1f3 c6a5 e1g1 c8b7 f1e1 d7d6 c3e2 f8e7 e2g3 g8f6 c1d2 e8g8 c2c3 a5b3 a2b3 d8c7 d1e2 f6d7 d3d4 f8e8 h2h3 a8c8 e1c1 c7b8 e2e3 c5d4 c3d4 c8c1 a1c1 h7h6 f3e1 e8c8 c1c8 b7c8 g3h5 b5b4 h5g3 e7f8');
INSERT INTO evaluations VALUES(53866,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(53866,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(53866,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(53866,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53866,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53866,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(53866,36,'cp 9','b8c6 d2d3 a7a6 b5c4 b7b5 c4b3 c8e6 e1g1 g8f6 c3d5 f8e7 c2c3 e6d5 e4d5 c6a5 b3c2 e8g8 d1e2 d8d7 f1e1 f6d5 d3d4 c5d4 c3d4 a5c6 d4e5 d6e5 f3e5 c6e5 e2e5 e7d6 c2f5 a8e8 f5d7 e8e5 e1e5 d6e5');
INSERT INTO evaluations VALUES(56714,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(56714,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(56714,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(56714,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(56714,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(56714,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(56714,36,'cp 81','e4f5 c8f5 g1e2 f5g6 e2g3 f8e7 f1d3 g6d3 d1d3 g8f6 b1c3 b8d7 c1e3 f6g4 e1g1 e8g8 g3e4 a7a6 h2h3 g4e3 f2e3 f8f1 a1f1 e7g5 c3d1 g5h4 b2b3 d7f6 e4f6 h4f6 g2g3 a6a5 e3e4 d8d7 g1g2 a5a4 h3h4 a8f8');
INSERT INTO evaluations VALUES(57630,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(57630,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(57630,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(57630,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(57630,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(57630,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(57630,36,'cp 91','h2h4 f8g7 h4h5 c7c5 d4d5 d7d6 g1h3 b8d7 h3f4 d7e5 b1c3 c8d7 e2e4 d8c8 e4f5 d7f5 f1e2 g8h6 h5g6 h7g6 f2f3 e8f7 c1e3 g6g5 f4h5 g7f6 g2g4 f5g6 h5g3 f7g7 d1b3 h6f7 e1c1 h8h1 d1h1 c8d7 e3d2 a8h8 h1h8 g7h8 c3e4 h8g8 d2c3');
INSERT INTO evaluations VALUES(10924,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10924,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10924,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(10924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10924,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(10924,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10924,36,'cp 0','b2b4 e8g8 b1d2 d7d5 c1b2 c8g4 d1c2 f8e8 f1e2 a7a5 a2a3 a5b4 a3b4 a8a1 b2a1 b7b5 e1g1 g4d7 f1b1 h7h6 e2f1 e7f8 a1b2 d5d4 c3d4 c6b4 c2c1 b4a2 c1c2');
INSERT INTO evaluations VALUES(12720,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(12720,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(12720,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(12720,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(12720,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(12720,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(12720,36,'cp 36','d7d6 d2d4 c5d4 f3d4 c8d7 d1d2 c6d4 d2d4 d7c6 e1g1 f8g7 d4d3 e8g8 c1d2 f6d7 b2b3 d7e5 d3c2 a8c8 f1d1 d8d7 a1c1 a7a6 c3d5 c6d5 g2d5 b7b5 c2e4 b5c4 d5c4 e5c4 c1c4 c8c4 e4c4 d7b7');
INSERT INTO evaluations VALUES(13014,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(13014,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(13014,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(13014,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(13014,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(13014,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(13014,36,'cp 41','d2d4 c5d4 f3d4 c8d7 d4c2 f8g7 a1b1 e8g8 e1g1 d8c8 d1d2 f8d8 b2b3 d7h3 c3d5 h3g2 g1g2 e7e6 d5f6 g7f6 c2e3 c6e5 c1b2 c8c5 f1d1 a8c8 b1c1 h7h5 c1c2 f6g7 h2h3 d8d7 f2f4 e5c6 b2g7 g8g7 c2c1 c8d8');
INSERT INTO evaluations VALUES(18687,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(18687,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(18687,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(18687,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(18687,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(18687,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(18687,36,'cp -62','d8c7 b5c6 c7c6 d1e2 g8e7 b1c3 e7f5 d2d3 d7d5 a2a4 c8d7 e1g1 h7h6 e2d1 c6c7 f1e1 f8e7 d3d4 e8g8 c3e2 b7b5 a4a5 a8c8 c2c3 b5b4 d4c5 b4c3 b2c3 e7c5 e2d4 f5d4 c3d4 c5b4 c1d2 c7c4 d2b4 c4b4');
INSERT INTO evaluations VALUES(21305,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(21305,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(21305,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(21305,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(21305,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(21305,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(21305,36,'cp 94','e1g1 g8f6 d4c6 b7c6 e4e5 f6g4 e5e6 d6d5 d1g4 c8e6 g4f3 d5c4 f3c6 e6d7 c6c4 f8e7 b1c3 e8g8 f1d1 a8b8 c3d5 e7d6 b2b3 c7c6 d5e3 d8c7 h2h3 d6e5 a1b1 d7e6 c4e2 f8e8 c1b2 e5h2 g1h1 e6f5 e2f3 f5g6 d1d2 b8c8 f3g4 e8e4 g4d7 c7d7 d2d7 h2f4 b1d1');
INSERT INTO evaluations VALUES(21418,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(21418,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21418,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21418,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(21418,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(21418,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(21418,36,'cp 30','c1f4 f8d6 e2e3 g8f6 b1d2 e8g8 f4g3 b7b6 f1d3 c8b7 e1g1 c6e7 b2b4 e7f5 g3e5 f6e4 d1c2 e4d2 f3d2 g7g6 c2b2 d8e7 a2a4 a7a5 f1b1 f7f6 e5d6 f5d6 b4b5 e6e5 c3c4 d5c4 d2c4 d6c4 d3c4 g8g7 d4e5');
INSERT INTO evaluations VALUES(21973,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21973,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21973,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21973,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21973,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21973,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21973,36,'cp 0','f7f5 e1g1 g8f6 e4f5 f8e7 d3d4 e5e4 f3g5 c8f5 f2f3 e4f3 g5f3 e8g8 d4d5 c6b4 f3d4 f5d7 c2c4 f6e4 a2a3 f8f1 e2f1 e7h4 c1e3 b4a6 b1d2 e4f2 e3f2 h4f2 g1f2 d8h4 f2e3 h4h6 e3f2');
INSERT INTO evaluations VALUES(30818,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30818,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(30818,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(30818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30818,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30818,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30818,36,'cp 22','e7e6 c1f4 f8b4 d1g4 b4c3 b2c3 g7g6 g1f3 b8c6 f1e2 d8e7 g4g3 d7d6 f3d4 e6e5 f4g5 f7f6 d4c6 b7c6 g5e3 e8c8 f2f4 f6f5 e1g1 d8f8 a2a4 a7a5 e4f5 g6f5 f4e5 d6e5 g3f2 c6b7 d3d4 e5d4 e3d4 g8f6 f2f5 c8b8');
INSERT INTO evaluations VALUES(30692,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(30692,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30692,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(30692,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(30692,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(30692,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30692,36,'cp 23','c1f4 a7a6 g1f3 b6b5 g2g3 d7d6 f1g2 g8f6 e1g1 b8d7 f3d2 c7c5 d1e2 d7e5 f4e5 d6e5 a2a4 b5b4 c3d1 f8e7 d2c4 d8c7 a4a5 a8d8 d1e3 b7c6 f2f4 e5f4 g3f4 e8g8 f4f5 e6e5 a1e1 c6b5 e3d5 f6d5 e4d5 e7f6 e2e3');
INSERT INTO evaluations VALUES(34912,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(34912,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(34912,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(34912,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(34912,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(34912,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(34912,36,'cp 163','b1c3 d5c6 e2e3 c8d7 f3d4 c6h6 c3e4 b8c6 d4f3 e8c8 d2d4 g8f6 e4f6 f8b4 c1d2 h6f6 f1g2 b4d2 f3d2 c6d4 d2e4 f6e7 d1d4 d7c6 d4g7 e7b4 g7c3 b4c3 b2c3 f7f5 h2h4 c6e4 g2e4 f5e4 h4h5 d8d5 e1f1 d5c5 h1h4 h8e8 f1g2 a7a5 a2a4 c5c3 h4f4 c3c5 g3g4 e8g8 f4e4 c5h5');
INSERT INTO evaluations VALUES(34909,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(34909,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(34909,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(34909,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(34909,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(34909,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(34909,36,'cp 173','d5c6 e2e3 c8d7 f3e5 c6e6 d1b3 e6b3 a2b3 g8f6 f1g2 b8c6 e5d7 f6d7 g2e4 d7c5 e4c2 c6b4 e1d1 c5d3 c2d3 b4d3 d1e2 e8c8 a1a7 c8b8 a7a4 c7c6 a4d4 d8d4 e3d4 d3b4 d2d3 f8d6 c1e3 h8e8 h2h4 f7f5 e2d2 b8c7 h1a1 g7g6 h4h5 b4d5 c3d5 c6d5 a1h1 c7c6 h5g6 h7g6 h1h6');
INSERT INTO evaluations VALUES(34762,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(34762,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(34762,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(34762,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(34762,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(34762,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(34762,36,'cp 151','e2e3 c8d7 f3e5 c6e6 d1b3 e6b3 a2b3 g8f6 f1g2 b8c6 e5d7 f6d7 g2e4 d7c5 e4c2 c6b4 e1d1 c5d3 c2d3 b4d3 d1e2 e8c8 a1a7 c8b8 a7a4 c7c6 a4d4 d8d4 e3d4 d3b4 d2d3 f8d6 c1e3 h8e8 h1c1 d6e7 d4d5 b4d5 c3d5 c6d5 e2f3 e7d6 e3c5 e8d8 h2h4 h7h5 c5e3 f7f6 e3f4 d6f4');
INSERT INTO evaluations VALUES(47511,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(47511,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(47511,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(47511,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(47511,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(47511,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(47511,36,'cp 34','d5e4 d2d4 e4d3 f1d3 b8c6 d1e2 c6d4 e2e3 d8d5 g1f3 d4f3 g2f3 c8d7 c1d2 e7e6 c2c4 d5c5 b2b4 c5e3 f2e3 d7c6 e1e2 e8c8 h1f1 c6a4 d2c3 f8e7 f1g1 g7g6 a1b1 d8d7 c3d4 h8d8 d3e4 h7h6 g1g4 c7c5 b4c5 d7c7 h2h4 d8d7 h4h5 g6h5 g4g7 e7c5');
INSERT INTO evaluations VALUES(47519,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(47519,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(47519,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(47519,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(47519,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(47519,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(47519,36,'cp 47','d2d4 e4d3 f1d3 b8c6 g1f3 c6b4 d3b5 c8d7 b5d7 d8d7 c1d2 e7e6 a2a3 b4d5 c2c4 d5e7 e1g1 e7f5 d1e2 f8e7 a1d1 d7c6 b2b4 c6a4 g2g4 f5h4 f3h4 e7h4 d2e3 h4e7 b4b5 e7a3 h2h3 a3e7 d1a1 a4b3 f1d1 e8g8 d1d7');
INSERT INTO evaluations VALUES(48380,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(48380,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(48380,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(48380,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(48380,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(48380,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(48380,36,'cp 24','d5e4 d2e4 f6e4 d3e4 b8d7 c2c3 d7f6 e4c2 c7c5 c1e3 f8e7 d4c5 d8c7 d1e2 e7c5 e2b5 f6d7 e3c5 c7c5 b5c5 d7c5 g1f3 a7a5 e1c1 e8e7 h1e1 c8d7 h2h4 d7c6 f3e5 h7h6 f2f4 h8d8 g2g3 d8d1 c2d1 c6e4 d1c2 f7f5 c2e4 c5e4');
INSERT INTO evaluations VALUES(48532,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(48532,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(48532,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(48532,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(48532,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(48532,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(48532,36,'cp 49','d2e4 f6e4 d3e4 b8d7 c2c3 c7c5 e4c2 d7f6 c1e3 d8c7 g1f3 f8e7 d1e2 e8g8 f3e5 b7b6 e1c1 f6d5 h1e1 c8b7 e2d3 f7f5 e3d2 c5d4 d3d4 a8c8 c2b3 f8d8 d4c4 c7c4 e5c4 d5c7 d2f4 b7d5 c4e3 d5b3 a2b3 g8f7 d1d8 e7d8');
INSERT INTO evaluations VALUES(48415,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48415,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48415,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48415,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(48415,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48415,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48415,36,'cp 0','f8e7 e1g1 e8g8 c2c4 c7c5 c4d5 e6d5 d3d4 c5c4 f3e5 c8e6 b1c3 b8c6 e5c6 b7c6 b2b3 d8a5 c1d2 a5a6 b3c4 a6c4 d1a4 c4a4 c3a4 f6e4 g2e4 d5e4 f1c1 f8b8 c1c6 e7f6 c6d6 f6e7');
INSERT INTO evaluations VALUES(12058,30,'cp -299',NULL);
INSERT INTO evaluations VALUES(12058,31,'cp -341',NULL);
INSERT INTO evaluations VALUES(12058,32,'cp -294',NULL);
INSERT INTO evaluations VALUES(12058,33,'cp -312',NULL);
INSERT INTO evaluations VALUES(12058,34,'cp -334',NULL);
INSERT INTO evaluations VALUES(12058,35,'cp -348',NULL);
INSERT INTO evaluations VALUES(12058,36,'cp -341','c8e6 f3e5 d8d4 a4c6 a5c6 e2b5 f8c5 b5c6 e8e7 c6b7 e7d6 e5f7 e6f7 e1g1 f7d5 b7a6 d6c7 b1a3 h8b8 a3b5 b8b5 a6b5 d5c4 b5a5 c5b6 a5c3 d4c3 d2c3 c4f1 g1f1 a8f8 c1f4 c7c6 f1e2 f6g4 f4g3 g4f2 a2a4 f2g4 a4a5 b6f2 a1f1 f2c5 f1f8 c5f8 h2h3 g4f6 g3e5 f8d6 e5d6 c6d6 e2e3 d6d5 e3f4 g7g5 f4e3');
INSERT INTO evaluations VALUES(24402,30,'cp -357',NULL);
INSERT INTO evaluations VALUES(24402,31,'cp -341',NULL);
INSERT INTO evaluations VALUES(24402,32,'cp -365',NULL);
INSERT INTO evaluations VALUES(24402,33,'cp -350',NULL);
INSERT INTO evaluations VALUES(24402,34,'cp -356',NULL);
INSERT INTO evaluations VALUES(24402,35,'cp -353',NULL);
INSERT INTO evaluations VALUES(24402,36,'cp -347','f3e5 d8d4 a4c6 a5c6 e2b5 f8c5 b5c6 e8e7 c6b7 e7d6 e5f7 e6f7 e1g1 f7d5 b7a6 d6c7 b1a3 h8b8 a3b5 b8b5 a6b5 d5c4 b5a5 c5b6 a5c3 d4c3 d2c3 c4f1 g1f1 a8f8 c1f4 c7c6 f1e2 f6g4 f4g3 b6c7 a1f1 g4h2 g3h2 c7h2 e2e3 h2g3 c3c4 h6h5 b2b4 g7g5 e3e2 g3d6 a2a3 h5h4 f1d1 g5g4 d1d5 g4g3 b4b5 c6c7 f2g3 d6g3 d5h5 c7b7 h5h7 b7a8');
INSERT INTO evaluations VALUES(17442,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(17442,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(17442,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(17442,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(17442,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(17442,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(17442,36,'cp 175','d4d5 f8b4 b1c3 c6e7 d1d4 b4c3 d4c3 g8f6 d5d6 e7g6 h2h4 h7h5 f1d3 b7b5 f3e5 e8g8 e5g6 f7g6 c1e3 f6g4 e3d4 c8b7 f2f3 a8c8 c3d2 g4h6 a2a4 b5a4 a1a4 g8h7 d4f2 h6f7 e1g1 f7e5 d3e2 d8f6 a4b4 b7c6 d2g5 f6g5');
INSERT INTO evaluations VALUES(23196,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(23196,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(23196,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(23196,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(23196,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(23196,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(23196,36,'cp -15','d7d5 d2d3 a7a6 b5c6 b7c6 e1g1 g8e7 b1d2 e7g6 d3d4 f8d6 c3c4 e8g8 c4d5 c6d5 e4d5 a8b8 d2e4 h7h6 f1e1 c8f5 h2h3 d6b4 c1d2 d8d5 e4g3 e5e4 g3f5 d5f5 d1c2 f8e8 c2b3 g8h7 d2b4 e4f3 e1e8 b8e8 b3f3 f5f3 g2f3 g6h4 b4c3 h7g6 f3f4 g6f5 a1d1 f5f4');
INSERT INTO evaluations VALUES(27851,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(27851,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(27851,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(27851,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(27851,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(27851,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(27851,36,'cp 60','e7e5 d4d5 a7a5 a2a3 b8a6 c1e3 b7b6 f3d2 g4e2 d1e2 d8d7 f1b1 f6g4 b2b4 g4e3 e2e3 d7g4 h2h3 g4h5 e3f3 h5f3 d2f3 g7h6 g1f1 f7f5 e4f5 g6f5 c3b5 f8f7 g2g3 g8h8 a1a2 f7g7 a2e2 a8f8 e2c2 h6g5 f3g5 g7g5 f1g2 g5g7 b5a7 a5b4 a3b4');
INSERT INTO evaluations VALUES(39320,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(39320,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(39320,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(39320,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(39320,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(39320,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(39320,36,'cp 217','c2c3 b4c5 c1f4 d7d5 e5d6 g8f6 e2e3 c5d6 f4d6 c7d6 b1d2 e8g8 f1d3 b8d7 d2e4 f6e4 d3e4 d8b6 d1c2 g7g6 e4d5 d7f6 d5f3 c8f5 c2b3 b6c7 g1e2 f5d3 a1d1 d3c4 b3b4 a8d8 b2b3 c4e2 f3e2');
INSERT INTO evaluations VALUES(39319,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(39319,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(39319,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(39319,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(39319,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(39319,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(39319,36,'cp 205','b4c5 c1f4 d7d5 e5d6 c5d6 f4d6 c7d6 e2e3 g8f6 b1d2 e8g8 f1d3 b8c6 g1f3 c8g4 d2e4 a7a5 e4f6 d8f6 a2a4 g4h5 d3e2 a8d8 e1g1 d6d5 f3d4 c6d4 e2h5 d4e6 h5e2 g7g6 e2f3 e6c5 f3d5');
INSERT INTO evaluations VALUES(39191,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(39191,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(39191,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(39191,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(39191,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(39191,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(39191,36,'cp 211','c1f4 d7d5 e5d6 c5d6 f4d6 c7d6 e2e3 b8c6 b1d2 c8f5 d2b3 f5e4 g1f3 g8f6 b3d2 e4f5 f3h4 f5d7 f1d3 e8g8 h4f3 d6d5 h2h3 f8e8 e1g1 f6e4 d3c2 h7h6 d2b3 d7f5 b3d4 f5h7 d4e2 a7a6 e2f4');
INSERT INTO evaluations VALUES(52698,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(52698,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(52698,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(52698,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(52698,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(52698,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(52698,36,'cp -39','b5a3 e7e5 a3c4 f7f6 d2d3 f8b4 c4d2 a7a5 f1e2 c8e6 g1h3 g8h6 a2a3 b4d6 d2f1 h6f7 e2g4 e6g4 d1g4 g7g6 f2f4 b8d7 f1d2 d8e7 e1g1 a5a4 d2f3 e8c8 f4e5 f7e5 f3e5 d6e5 c1f4 c8b8 g4g3 d8f8 g1h1 g6g5 f4e5 e7e5 g3f2 c6c5 h3g1 h7h5 g1f3');
INSERT INTO evaluations VALUES(52910,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(52910,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(52910,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(52910,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(52910,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(52910,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(52910,36,'cp -76','e7e5 a3c4 f7f6 d2d3 b7b5 c4d2 a7a5 c2c3 a5a4 c3d4 d8d4 g1f3 d4a7 f1e2 f8c5 e1g1 c8e6 b2b3 g8e7 b3a4 a7a4 c1b2 a4d1 f1d1 a8a2 a1a2 e6a2 d3d4 e5d4 d1a1 a2f7 f3d4 e8g8 d2b3 c5b6 b2a3 f8e8 a3c5 b8d7 c5b6 d7b6 b3c5 e8a8 a1a8 b6a8 g1f1 a8b6');
INSERT INTO evaluations VALUES(53789,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(53789,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(53789,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(53789,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(53789,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(53789,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(53789,36,'cp -46','a3c4 f7f6 d2d3 f8b4 c4d2 b4e7 g2g3 g8h6 f1h3 c8h3 d1h5 h6f7 h5h3 a7a5 a2a4 b8a6 f2f4 b7b5 g1f3 a6b4 e1d1 e8g8 f3h4 f8e8 f4e5 f7e5 h1f1 e7f8 d2f3 b4d3 c2d3 e5d3 h3f5 g8h8 a4b5 c6b5 f5b5 d3c5 e4e5 a5a4 e5f6 d8f6');
INSERT INTO evaluations VALUES(2552,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(2552,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(2552,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(2552,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(2552,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(2552,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(2552,36,'cp 167','g2g4 c8e6 d4e6 f7e6 g4g5 f6d7 f1h3 d7c5 f3f4 g8h8 e3c5 d6c5 d2e3 c6d4 e4e5 d8c7 c1b1 a8d8 d1d3 d4b5 c3e2 d8d3 e3d3 f8d8 d3e4 c7b6 e4e3 b6c6 h1g1 b5c7 h3g2 c7d5 e3g3 c6b5 g2f3 c5c4 g1d1 c4c3 b2b3 b5c6 h2h4 c6c5 h4h5 g6h5 f3h5');
INSERT INTO evaluations VALUES(2549,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(2549,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(2549,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(2549,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(2549,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(2549,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(2549,36,'cp 159','c8e6 d4e6 f7e6 g4g5 f6d7 f3f4 d7c5 f1h3 g8h8 e3c5 d6c5 d2e3 d8c7 e4e5 c6d4 c1b1 a8d8 d1d3 d4b5 c3e2 d8d3 e3d3 f8d8 d3f3 c7d7 a2a4 b5d4 e2d4 d7d4 h1f1 b7b5 a4b5 a6b5 h3e6 d8f8 h2h4 c5c4 f3c6 g7e5 c6b5 e5g7 e6c4');
INSERT INTO evaluations VALUES(4770,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4770,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(4770,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(4770,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(4770,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(4770,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(4770,36,'cp 31','c2c3 c5a7 d2d4 b7b5 a4b3 g8f6 c1e3 e8g8 b1d2 c8g4 h2h3 g4h5 a1c1 f8e8 f1e1 h5f3 d2f3 c6a5 e3g5 h7h6 g5h4 a5b3 a2b3 d8e7 d1d3 a7b6 b3b4 e5d4 f3d4 e7e5 h4f6 e5f6 d3f3 f6f3 g2f3 c7c5 d4c6 c5b4 c6b4 a6a5');
INSERT INTO evaluations VALUES(27692,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(27692,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(27692,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(27692,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(27692,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(27692,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(27692,36,'cp 210','f7e6 f3g4 f6f5 g4f5 e6f5 d2d3 c6b4 e1d1 e5e4 a2a3 b4d5 d3e4 f5e4 c2c4 d5f6 d1c2 e4f5 b1c3 f8c5 c1d2 h8e8 a1e1 c7c6 h2h3 g7g5 b2b4 c5b6 h1f1 h7h5 f2f3 f5g6 c3e4 f6e4 f3e4 b6c7 f1f5 c7d8 c2d3 a7a5 b4a5 d8f6 e1f1 a8d8 d3c2 d8d6 c4c5 d6e6 d2e3');
INSERT INTO evaluations VALUES(27698,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(27698,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(27698,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(27698,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(27698,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(27698,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(27698,36,'cp 200','f3h3 f6f5 h3f5 e6f5 d2d3 e5e4 d3e4 f5e4 c1e3 a8d8 b1d2 e4f5 e1c1 h7h5 h1e1 f8e7 c2c3 h8e8 c1c2 g7g5 h2h3 a7a6 g2g3 e7f6 f2f3 f5g6 e1e2 f6e5 e3f2 h5h4 g3h4 e5f4 e2e8 d8e8 d2e4 g5h4 d1g1 g6f7 f2h4 c6e5 h4g3 e5f3');
INSERT INTO evaluations VALUES(29568,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(29568,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(29568,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(29568,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(29568,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(29568,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(29568,36,'cp -48','e7e6 g1f3 h7h6 e1g1 f8e7 c3b1 c7c6 b2b3 e8g8 c2c4 b8d7 b1d2 a7a5 a2a3 d8b8 f1e1 f5g6 c1b2 f8d8 e2e3 b7b6 a1c1 b8b7 d1e2 f6e4 d2e4 g6e4 c4d5 c6d5 g2f1 a8c8 e1d1 e7d6 f3e1 c8c1 d1c1 d7f6 e1d3 e4f5');
INSERT INTO evaluations VALUES(30119,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(30119,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(30119,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(30119,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(30119,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(30119,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(30119,36,'cp -55','g1f3 h7h6 e1g1 f8e7 c3b1 c7c6 c2c4 b8d7 b2b3 a7a5 b1c3 e8g8 c1b2 d8b8 f3d2 b7b5 e2e4 d5e4 d1e2 b5c4 d2c4 f6d5 g2e4 d5c3 b2c3 f5e4 e2e4 b8b5 f1c1 a5a4 g1g2 d7f6 e4f3 f6d5 c3d2 e7f6 c4e5 f6e5 d4e5 a4b3 a2b3 f8b8 a1a8');
INSERT INTO evaluations VALUES(30118,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(30118,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(30118,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(30118,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(30118,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(30118,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(30118,36,'cp -65','h7h6 e1g1 f8e7 b2b3 e8g8 c1b2 b8d7 f3e5 f5h7 f1e1 d7e5 d4e5 f6d7 e2e4 c7c6 e4d5 c6d5 c3b5 d8b6 c2c4 d7c5 a2a4 d5c4 b3c4 a8d8 b2d4 b6a5 d1f3 d8d7 d4c3 a5a6 e1d1 d7d1 a1d1 a6a4 b5d6 e7d6 e5d6 a4c4 d1d4');
INSERT INTO evaluations VALUES(31489,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(31489,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(31489,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31489,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(31489,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31489,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31489,36,'cp 0','c6c5 g1e2 b8c6 e1g1 d8b6 b1d2 c5d4 c3d4 c6b4 d3c3 b6b5 e2g3 g8e7 a2a3 b4c6 d2f3 a8c8 c3e3 h7h5 h2h4 e7g6 g3e2 g6e7');
INSERT INTO evaluations VALUES(31652,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(31652,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(31652,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31652,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(31652,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(31652,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31652,36,'cp 17','g1e2 b8c6 e1g1 d8b6 b1d2 c5d4 c3d4 c6b4 d3c3 b6b5 e2g3 b5d3 a2a3 d3c3 b2c3 b4c6 d2b3 a8c8 a3a4 f7f6 f1e1 f6e5 d4e5 g8e7 a4a5 c6d8 g3e2 e7f5 c1d2 c8c4 g2g3 d8c6 e1b1 h7h5 a5a6 b7b5 e2d4 f5d4 b3d4 c6d4');
INSERT INTO evaluations VALUES(31650,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31650,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31650,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(31650,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31650,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31650,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31650,36,'cp 0','b8c6 e1g1 d8b6 b1d2 c5d4 c3d4 c6b4 d3c3 b6b5 e2g3 h7h5 h2h4 b5c6 d2b3 c6c3 b2c3 b4c2 a1b1 b7b6 b3d2 f8e7 g3e2 e7h4 b1b2 c2a3 b2b4 a3c2');
INSERT INTO evaluations VALUES(26405,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(26405,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(26405,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(26405,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(26405,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26405,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(26405,36,'cp 6','e1g1 d8b6 a2a3 a7a5 a3a4 c5d4 c3d4 f8b4 b1c3 g8e7 c1e3 e8g8 e2f4 a8c8 a1c1 b6d8 c3b5 e7f5 h2h3 d8d7 f4h5 b4e7 e3d2 c6b4 d3e2 c8c1 f1c1 f8c8 c1c3 c8c3 d2c3 b7b6 g2g3 b4a6 h5f4 a6c7 h3h4 h7h6 e2d3 c7b5 d3b5 d7b5 a4b5 g8h8 h4h5 h8g8');
INSERT INTO evaluations VALUES(31659,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(31659,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(31659,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(31659,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(31659,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(31659,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(31659,36,'cp -26','b8c6 c1e3 d8b6 d1d2 h7h5 d4c5 f8c5 e3c5 b6c5 d2f2 c5f2 e1f2 g8e7 f3h4 f5h7 b1d2 g7g5 f4g5 c6e5 d2f3 e5c6 f1b5 a7a6 b5c6 e7c6 h1e1 h8g8 g2g3 h7e4 f3d2 g8g5 d2e4 d5e4 e1e4 g5b5 e4e2 c6e5 a1d1 e5g4 f2g1 a8c8 d1d2 b5c5 h4f3 e8e7 d2d4 g4f6 d4f4');
INSERT INTO evaluations VALUES(26408,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(26408,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(26408,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(26408,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(26408,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(26408,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(26408,36,'cp -19','c1e3 d8b6 d1d2 f5e4 d2f2 c5d4 f3d4 f8c5 f1b5 c5d4 b5c6 b6c6 e3d4 e4b1 a1b1 g8e7 e1g1 b7b6 f4f5 e6f5 b1a1 e8g8 a2a4 c6e6 b2b4 e7c6 f2f5 a8c8 f5f4 c6d4 c3d4 f7f6 f4e3 h7h6 b4b5 c8c4 a4a5 f8c8 a5b6 a7b6');
INSERT INTO evaluations VALUES(38626,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38626,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(38626,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38626,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38626,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38626,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38626,36,'cp 2','g8f6 f1d3 e8g8 e1g1 h7h6 b1c3 b8c6 c3e2 c6b4 c1f4 d6f4 e2f4 f8e8 f3e5 f6d7 e5g4 b4d3 f4d3 d7b6 f1e1 e8e1 d1e1 c8f5 d3c5 f5g4 h3g4 d8f6 c2c3 f6c6 e1d2 b6c4 d2c2 c4d6 c5d3 a8e8 a1e1 e8e1 d3e1 d6e4 a2a4 e4g5 e1d3');
INSERT INTO evaluations VALUES(39464,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(39464,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(39464,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(39464,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(39464,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(39464,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(39464,36,'cp 259','d2d4 e5d4 f3d4 d7d6 f1c4 b8c6 d4e6 c8e6 c4e6 g8e7 e1g1 e7g6 c3d5 a7a5 c2c3 c6e5 b2b4 c5a7 b4a5 c7c6 d5b6 a7b6 a5b6 d8b6 a2a4 e5f7 c1a3 b6c7 f2f4 c6c5 h2h4 e8g8 h4h5 g6e7 h5h6 g7h6 d1h5 a8a4 e4e5 d6e5 f4e5 c7e5 h5e5 f6e5 a3c5 a4a1 f1a1 e7c6 c5f8 g8f8');
INSERT INTO evaluations VALUES(39463,30,'cp 249',NULL);
INSERT INTO evaluations VALUES(39463,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(39463,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(39463,33,'cp 242',NULL);
INSERT INTO evaluations VALUES(39463,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(39463,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(39463,36,'cp 250','e5d4 f3d4 d7d6 f1c4 g8e7 d4b3 c5b4 d1h5 g7g6 h5h6 d6d5 c4e2 d5e4 h6g7 b4c3 b2c3 h8g8 g7f6 b8c6 e1g1 e7f5 f6d8 c6d8 f2f3 e4e3 g2g4 f5d6 c1e3 d8e6 c3c4 c8d7 a1e1 e8c8 e2d3 b7b6 c4c5 d6b7 g1g2 d7c6 g2g3 e6c5 b3c5 b7c5 e3c5 b6c5 e1e5');
INSERT INTO evaluations VALUES(44881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44881,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(44881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44881,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(44881,36,'cp -2','e2e3 b8d7 g5f4 e7e6 a2a3 f8e7 f1d3 d8b6 g1f3 f6h5 f4g5 e7g5 f3g5 h5f6 g5f3 e8g8 e1g1 f8e8 d1d2 e6e5 f3e5 d7e5 d4e5 e8e5 f1e1 c8d7 b2b3 a8e8 e3e4 b6d4 e4d5 f6d5 c3d5 c6d5 h2h3 g7g6 a1b1 e5e1 b1e1 e8e1 d2e1 d7c6 b3b4 g8g7');
INSERT INTO evaluations VALUES(46729,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(46729,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(46729,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(46729,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(46729,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(46729,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(46729,36,'cp 44','d2d4 f8g7 g1f3 c7c5 d4c5 d8a5 f1d3 a5c5 d1e2 e8g8 c1e3 c5a5 e1g1 c8g4 e2f2 g4f3 f2f3 b8c6 a1b1 a7a6 a2a3 e7e6 g1h1 a5c7 c3e2 b7b5 c2c3 a8e8 f3h3 c6a5 f4f5 e6f5 e4f5 a5c4 e3g5 c4e5 d3c2');
INSERT INTO evaluations VALUES(56751,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(56751,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(56751,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(56751,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(56751,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(56751,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(56751,36,'cp 37','b1c3 b8c6 e4f5 g8e7 g2g4 h7h5 g4h5 e7f5 h1g1 f8e7 c3d5 c8e6 f1h3 h8f8 d5e7 d8e7 c2c3 f5h4 f3h4 e7h4 g1g3 h4f6 c1e3 e6h3 g3h3 e8c8 d1g4 c8b8 e1c1 f6f7 c1b1 d6d5 g4g6 d5d4 g6f7 f8f7 c3d4 e5d4 e3c1 d8h8 h5h6 g7h6 h3h6 h8h6 c1h6 f7f2 h2h4');
INSERT INTO evaluations VALUES(56784,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(56784,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(56784,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(56784,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(56784,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(56784,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(56784,36,'cp 158','e4g3 c7c5 g1f3 b8c6 f1c4 b7b5 c4d3 g8f6 e1g1 c6d4 f1e1 f8e7 f3d4 c5d4 d3b5 e8f7 b5c4 g7g6 g3e2 e7b4 c1d2 b4d6 d2f4 d6f4 e2f4 h8e8 f4e2 d8c7 d1d4 f7g7 e2c3 c8b7 c3b5 c7c6 c4f1 a8d8 d4c3 c6c3');
INSERT INTO evaluations VALUES(57640,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57640,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(57640,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(57640,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(57640,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57640,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57640,36,'cp 0','f8g7 g1f3 g8f6 b1d2 e8g8 e2e4 d7d5 e4d5 f6d5 c2c3 c7c5 d4c5 h7h6 g5h4 d5f4 d1a4 d8c7 e1c1 c8d7 a4c4 d7e6 c4a4');
INSERT INTO evaluations VALUES(39545,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39545,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39545,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39545,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39545,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39545,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39545,36,'cp 0','g1f3 c7c5 e2e3 d8b6 b1a3 d7d5 c2c3 b6b2 a3b5 b8a6 f3d2 c5d4 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(39544,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(39544,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(39544,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(39544,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39544,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39544,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39544,36,'cp 0','g8f6 e2e3 e8g8 c2c4 c7c5 b1c3 c5d4 d1d4 h7h6 g5h4 d7d5 a1d1 e7e6 c4d5 g6g5 d5e6 d8d4 e3d4 g5h4 e6f7 g8f7 f1c4 f7e7 f3h4 f8e8 e1g1 e7d8 c3b5 g7f8 c4f7 e8e7 f7c4');
INSERT INTO evaluations VALUES(3435,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(3435,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3435,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3435,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(3435,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(3435,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(3435,36,'cp 26','c7c6 f1d1 d8a5 g5d2 a5e5 d4e5 d6e5 a2a4 e7c5 g1f1 a7a5 f1e2 f8e8 f2f3 g8f8 d2e1 g7g6 e2f1 f6h5 c3e2 f7f6 e2c1 f8e7 c1d3 c5d4 c2c3 d4a7 b2b3 h5g7 d3b2 c8e6 c4e6 g7e6 b2c4 a7c5 h2h4 b7b6 e1f2');
INSERT INTO evaluations VALUES(10412,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10412,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10412,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(10412,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10412,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10412,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(10412,36,'cp 15','f1g2 e8g8 c3d5 e5e4 f3h4 d7d6 d5b4 c6b4 a2a3 b4c6 d2d3 f8e8 e1g1 h7h6 d3e4 f6e4 d1c2 e4c5 e2e3 a7a5 b2b3 g7g5 h4f3 d8f6 c1b2 c8f5 c2d2 f6g6 b3b4 c5a4 b2d4 f5e4 a1d1 c6e5 f3e5 d6e5 g2e4 g6e4 f2f3 e4g6 d4b2 a4b2 d2b2 g8h7 b2c3');
INSERT INTO evaluations VALUES(14346,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(14346,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(14346,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(14346,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(14346,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(14346,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(14346,36,'cp -23','e7e6 c1f4 f8b4 a2a3 b4c3 b2c3 c6e5 f4e5 e8g8 f2f3 f6d7 e5g3 c7c5 d1d2 d8b6 a1d1 b6c6 e2e3 b7b5 f1d3 c5c4 d3e2 c6a6 d1a1 c8b7 e1g1 f8e8 f1e1 b7c6 e2d1 a6b7 a1b1 f7f6 g3d6 a7a6 d2f2 a6a5 f2g3 e6e5');
INSERT INTO evaluations VALUES(14885,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(14885,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(14885,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(14885,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(14885,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(14885,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(14885,36,'cp -18','a2a3 f8d6 e5c6 b7c6 e2e4 d5e4 d1e2 d6e7 e2e3 a8b8 c3e4 e8g8 f1c4 f6e4 e3e4 c8b7 e4g4 h7h5 g4h5 d8d4 c4d3 g7g6 h5e2 c6c5 e1g1 b7d5 c2c3 d4h4 g2g3 h4f6 f2f3 a7a5 d3e4 e7d6 f1f2 f6e5 c1f4 e5g7 f4d6 c7d6 a1d1 d5e4 f3e4');
INSERT INTO evaluations VALUES(14790,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(14790,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(14790,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(14790,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(14790,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(14790,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(14790,36,'cp 64','g1f3 f8e7 b1c3 e8g8 b2b4 c6b8 c4c5 c7c6 f1e2 a7a5 c1d2 b8d7 e1g1 d8c7 b4b5 f6e4 a1c1 e6e5 d2e1 e4c3 c1c3 e5e4 f3d2 c6b5 e2b5 b7b6 c5c6 d7f6 f2f3 e4f3 d1f3 e7d6 e1g3 d6g3 f3g3 c7g3 h2g3');
INSERT INTO evaluations VALUES(16057,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(16057,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(16057,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(16057,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(16057,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(16057,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(16057,36,'cp 173','f3g5 d7d5 e4d5 f6d5 b1c3 f8b4 e1g1 b4c3 b2c3 e8g8 d1f3 c8e6 g5e4 d5f4 c4e6 f4e6 e4f6 g8g7 a1b1 a8b8 f1e1 h7h5 h2h4 e6f4 f6e4 f4e6 f3e3 f8h8 e3g3 b7b6 d3d4 d8d5 g3f3 f7f5 e4g5 d5f3 g5f3 b8d8 f3e5 c6e5 e1e5 h8e8 f2f3 g7f7 a2a4 e6g7 c1g5 d8d7 b1e1 e8e5 e1e5');
INSERT INTO evaluations VALUES(16049,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(16049,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(16049,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(16049,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(16049,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(16049,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(16049,36,'cp 180','d7d5 e4d5 f6d5 b1c3 c8e6 c3e4 f8e7 g5e6 f7e6 c2c3 d8d7 e1g1 c6a5 c1h6 e8c8 h6g7 h8e8 g7e5 a5c4 d3c4 d5b6 b2b3 d7d1 f1d1 b6d7 e5g7 h7h5 g1f1 g6g5 d1e1 g5g4 a1d1 e8g8 g7d4 a7a5 c4c5 h5h4 a2a4 h4h3 g2g3 g8f8 b3b4 a5b4 c3b4 d7b8 e4c3');
INSERT INTO evaluations VALUES(14515,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(14515,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(14515,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(14515,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(14515,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(14515,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(14515,36,'cp 211','e4d5 c6a5 d1e2 a5c4 d3c4 d8e7 b1c3 f8g7 c1e3 e8g8 e1c1 f6g4 e3d2 f7f5 f2f3 g4f6 h1g1 h7h6 g5h3 e5e4 g1e1 c8d7 f3e4 f5e4 h3f2 a8e8 h2h3 a7a6 g2g4 b7b5 c3e4 f6e4 f2e4 b5c4 e2c4 d7b5 c4c5 e7c5 e4c5 e8e1 d1e1 f8f1 c5e6');
INSERT INTO evaluations VALUES(17238,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17238,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17238,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(17238,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(17238,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(17238,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17238,36,'cp 0','e7e6 d2d4 d7d5 c4d5 e6d5 c1g5 c8e6 f1g2 h7h6 g5f6 d8f6 c3d5 e6d5 g2d5 c6d4 d5b7 a8b8 g1f3 b8b7 d1d4 f6d4 f3d4 b7b2 e1g1 f8c5 d4b3 c5b6 e2e3 e8e7 f1b1 b2c2 b1c1 c2b2');
INSERT INTO evaluations VALUES(16266,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16266,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(16266,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16266,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(16266,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(16266,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(16266,36,'cp 8','d2d4 d7d5 c4d5 e6d5 c1g5 f8e7 e2e3 h7h6 g5f6 e7f6 g1e2 c6e7 e2f4 c7c6 f4h5 d8d6 a1c1 g7g6 h5f6 d6f6 f1g2 e8g8 e1g1 f8d8 d1f3 f6f3 g2f3 e7f5 f3g2 f5d6 b2b3 g6g5 f2f3 g5g4 f3g4 c8g4 h2h3 g4f5 g1f2 f5g6 g2f3 g8g7 c3e2 d8e8 e2f4');
INSERT INTO evaluations VALUES(19785,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(19785,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19785,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19785,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19785,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(19785,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(19785,36,'cp 13','d4c5 a7a5 c3c4 e7e6 b1c3 g8f6 c3a4 f6e4 c4d5 e6d5 g1f3 f8c5 a4c5 e4c5 c1d2 c5e4 d2c3 e8g8 f1e2 a5a4 e1g1 e4c3 b2c3 c8e6 a1b1 d8e7 b1b5 f8c8 h2h3 c8c7 d1b1 g7g6 f1c1 a4a3 e2d3 a8c8 c1d1 g8g7 g1h1 e7f6 b5b7 c7b7 b1b7 f6c3');
INSERT INTO evaluations VALUES(22614,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22614,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(22614,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(22614,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(22614,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(22614,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(22614,36,'cp 95','d7d5 e4e5 f7f6 g1f3 f6e5 f3e5 c6e5 d4e5 d8h4 f1e2 g8h6 e1g1 h6f7 g2g3 h4d8 f2f4 f8c5 g1g2 c8d7 b1d2 c5b6 h2h3 c7c5 g2h2 f7h6 g3g4 d7c6 d2f3 d8d7 a2a4 a7a5 d1b3 b6a7 f1d1 c5c4 b3c2 e8g8 f3d4 a7d4 c3d4');
INSERT INTO evaluations VALUES(21439,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(21439,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(21439,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(21439,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(21439,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(21439,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(21439,36,'cp 114','e4e5 f7f6 g1f3 f6e5 f3e5 c6e5 d4e5 c8d7 f1d3 g7g6 e1g1 f8g7 f2f4 g8e7 b1d2 c7c5 d2f3 d8b6 f1f2 g7f8 d1e1 a7a5 b2b3 h7h6 c1e3 h6h5 a1b1 e7f5 b3b4 a5b4 c3b4 f5e3 d3g6 e8d8 e1e3 f8e7 b4c5');
INSERT INTO evaluations VALUES(40758,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(40758,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(40758,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(40758,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(40758,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(40758,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(40758,36,'cp -37','e8g8 e1g1 b8c6 d2d3 c6a5 c4b3 h7h6 g5f3 c7c6 h2h3 f8e8 c1d2 a5b3 a2b3 e7f8 c3e2 d6d5 e4d5 c6d5 d3d4 e5e4 f3e5 c8e6 e2f4 a7a5 d2e3 f8d6 c2c4 d8e7 d1c2 d6c7 f4e6 e7e6 c4d5 f6d5 c2e4 c7e5 d4e5');
INSERT INTO evaluations VALUES(36213,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(36213,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(36213,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(36213,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(36213,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(36213,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(36213,36,'cp -40','e1g1 b8c6 d2d3 c6a5 c4b3 h7h6 g5f3 f8e8 c3d5 a5b3 d5e7 d8e7 a2b3 d6d5 f1e1 d5e4 d3e4 b7b6 f3d2 c8b7 d1f3 a7a5 f3c3 a8d8 f2f3 b7c8 d2f1 f6h5 c1d2 h5f4 a1d1 d8d4 d2f4 d4d1 e1d1 e5f4 d1d4 c7c5 d4d5 c8e6 d5d1 e7g5 d1d2 f7f6');
INSERT INTO evaluations VALUES(41775,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(41775,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(41775,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(41775,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(41775,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(41775,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(41775,36,'cp 53','c1g5 d6e7 e2e3 e8g8 d1c2 b8d7 c4d5 e6d5 f1d3 h7h6 g5h4 f8e8 e1g1 c7c6 h4g3 a7a6 f1d1 d7f8 h2h3 e7d6 g3d6 d8d6 a1c1 c8e6 d1e1 f8d7 b2b3 a8c8 c2b2 d6e7 d3b1 c8c7 c3e2 f6e4 e2f4 a6a5 f3d2 d7f6 d2e4 f6e4');
INSERT INTO evaluations VALUES(41768,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(41768,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(41768,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(41768,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(41768,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(41768,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(41768,36,'cp 49','d6e7 e2e3 e8g8 d1c2 b8d7 c4d5 e6d5 f1d3 h7h6 g5h4 c7c6 e1g1 f8e8 h4g3 a7a6 a1c1 e7b4 g3h4 b4e7 c1e1 f6h5 h4e7 e8e7 h2h3 h5f6 e1e2 f6e8 f1e1 d7f6 e3e4 f6e4 c3e4 d5e4 d3e4 c8e6 e4h7 g8h8 h7d3');
INSERT INTO evaluations VALUES(45601,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(45601,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(45601,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(45601,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(45601,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(45601,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(45601,36,'cp -29','a2a3 a7a5 g1f3 b7b6 c5b6 f8d6 b1c3 d8b6 c1d2 e8g8 a1c1 b6a7 f1d3 b8d7 e3e4 e6e5 d2e3 f6g4 e1g1 g4e3 f2e3 e5d4 f3d4 d7e5 d3e2 d5e4 c3e4 d6c7 d1c2 h7h6 c2c3 f8e8 b2b4 c8g4 d4c6 e5c6 e2g4 e8e4 c3c6 c7h2 g1h2 e4g4 b4b5 g4g5 h2h1');
INSERT INTO evaluations VALUES(45992,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(45992,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(45992,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45992,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(45992,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(45992,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(45992,36,'cp -9','f8e7 e1g1 e8g8 b1d2 a7a6 a2a4 b7b6 f1e1 c8b7 c4a2 d8c7 a2b1 b6b5 d2f1 b8d7 f1g3 f8c8 d3d4 b5a4 a1a4 b7c6 a4a2 c5d4 c3d4 a6a5 c1d2 a5a4 d4d5 e6d5 e4d5 f6d5 f3d4 g7g6 d4c6 c7c6 b1e4 d7f6');
INSERT INTO evaluations VALUES(40179,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(40179,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(40179,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(40179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40179,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40179,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40179,36,'cp 0','e1g1 e8g8 f1e1 d6d5 c4b3 d5e4 d3e4 d8d1 b3d1 f8d8 d1c2 b8c6 a2a4 b7b6 b1a3 h7h6 h2h3 c8a6 c1f4 d8d7 a3b5 a8d8 f4c7 d8c8 c7h2 c8d8');
INSERT INTO evaluations VALUES(40178,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40178,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40178,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40178,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40178,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40178,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40178,36,'cp 0','e8g8 f1e1 d6d5 c4b3 d5e4 d3e4 d8d1 b3d1 f8d8 d1c2 b8c6 a2a4 b7b6 b1a3 h7h6 h2h3 c8a6 c1f4 d8d7 a3b5 a8d8 f4c7 d8c8 c7g3 g7g5 g3h2 c8d8 h2c7 d8c8');
INSERT INTO evaluations VALUES(35813,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35813,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35813,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35813,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35813,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35813,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35813,36,'cp 0','f1e1 d6d5 c4b3 d5e4 d3e4 d8d1 b3d1 f8d8 d1c2 b8c6 a2a4 b7b6 b1a3 h7h6 h2h3 c8a6 c1f4 d8d7 a3b5 a8d8 f4c7 d8c8 c7g3 g7g5 g3h2 c8d8 h2c7 d8c8');
INSERT INTO evaluations VALUES(49442,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(49442,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(49442,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(49442,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(49442,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(49442,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(49442,36,'cp -87','f6e4 f1d3 e4f6 c2c4 c7c5 d4c5 f8c5 b1c3 e8g8 e1g1 b7b6 c1g5 c5e7 d1e2 c8b7 a1d1 d7d6 d3b1 b8d7 g5f4 d8b8 f1e1 f8e8 f3g5 b8c7 g5e4 b7e4 c3e4 f6e4 b1e4 a8d8 b2b3 d7e5 d1d2 f7f5 e4f3 e5f3 e2f3 e7f6 e1d1');
INSERT INTO evaluations VALUES(49341,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(49341,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(49341,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(49341,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(49341,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(49341,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(49341,36,'cp -76','f1d3 e4f6 c2c4 c7c5 d4c5 f8c5 b1c3 d7d5 c4d5 e6d5 e1g1 e8g8 a2a3 h7h6 b2b4 c5e7 f1e1 a7a5 b4a5 b8c6 h2h3 d8a5 c1d2 c8e6 c3b5 a5d8 d2c3 f6e4 d3e4 d5e4 e1e4 d8d1 a1d1 e7a3 b5c7 a8d8 d1d8 c6d8 c7e6 d8e6 f3d4 e6c5');
INSERT INTO evaluations VALUES(49466,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49466,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49466,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(49466,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49466,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(49466,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49466,36,'cp 0','e2e4 d7d5 e4e5 f6d7 c3e2 c7c5 d2d4 c5d4 e2d4 b8c6 c1f4 f8e7 f1b5 d8b6 e1g1 a7a6 d4c6 b7c6 b5d3 b6b2 c2c4 a6a5 c4d5 c6d5 a1b1 b2a2 f3d4 c8a6 d4b5 e8g8 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(50378,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(50378,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(50378,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(50378,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(50378,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(50378,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(50378,36,'cp -25','d7d5 d3d4 c7c5 g1f3 b8c6 c2c3 e7e6 b1d2 f8d6 d4c5 d6c5 b2b4 c5d6 c1b2 e8g8 a2a3 e6e5 c3c4 d5d4 c4c5 d6c7 d1c2 d4e3 f2e3 e5e4 d2e4 f6e4 c2e4 f8e8 e4d3 d8e7 e1c1 a7a5 b4b5 e7c5 d3c2 c5c2 c1c2 c8f5 f1d3 f5d3 d1d3 c6e5');
INSERT INTO evaluations VALUES(47274,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(47274,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(47274,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(47274,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(47274,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(47274,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(47274,36,'cp -30','c2c4 d5c4 d3d4 c8e6 b1d2 c7c5 f1c4 e6c4 d2c4 c5d4 e3d4 e7e6 g1f3 f8b4 c1d2 b4d2 d1d2 e8g8 a1c1 b8c6 e1g1 a8c8 c4e5 c6e7 c1c8 d8c8 f1c1 c8a8 g2g3 f6d5 d2e2 b7b6 a2a3 h7h6 h2h3 f8c8 c1c8 a8c8 e2c4 c8c4 e5c4');
INSERT INTO evaluations VALUES(54340,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(54340,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(54340,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(54340,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(54340,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(54340,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54340,36,'cp 27','e3d4 g8f6 g1f3 f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 d8d6 b5c6 d6c6 f3e5 c6d6 g5f6 d6f6 f1e1 c8e6 d1b3 b7b6 a2a4 f6d8 a4a5 f8e8 h2h3 b6a5 b3b5 d8c7 b5a5 a8c8 a5a7 c7a7 a1a7 c8c3 a7b7 e8c8');
INSERT INTO evaluations VALUES(54994,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(54994,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(54994,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(54994,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(54994,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(54994,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(54994,36,'cp 65','d2d4 g8f6 e2e4 e7e5 g1f3 b8d7 a2a4 f8e7 f1c4 e8g8 e1g1 h7h6 h2h3 f8e8 a4a5 e5d4 f3d4 d7e5 c4b3 c7c5 d4e2 e5c6 b3a4 a7a6 e2g3 c8e6 c1f4 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 c2c3');
INSERT INTO evaluations VALUES(55323,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(55323,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(55323,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(55323,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(55323,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(55323,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(55323,36,'cp 61','c7c5 b2b4 c5b4 c3d5 f8g7 d2d4 b8c6 d4e5 d6e5 f3g5 f7f6 g5f7 e8f7 d5c7 f7e7 c7a8 d8d1 e1d1 f6f5 a8c7 g8f6 c7d5 e7d6 d5f6 g7f6 a2a3 c8e6 c4e6 d6e6 c1b2 a7a5 e4f5 g6f5 a3b4 a5b4 d1e2 h8c8 h1d1 c6e7 a1a5 c8c2 d1d2 c2c4 g2g3 e7c6 a5b5 f6g5 e2d3 c4e4 d2e2');
INSERT INTO evaluations VALUES(6977,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(6977,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(6977,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(6977,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(6977,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(6977,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(6977,36,'cp 80','d2d4 e5d4 f3d4 d7d6 f1b5 c8d7 d4c6 b7c6 b5d3 e7f6 e1g1 g8e7 d1e1 e8g8 f2f4 f6d4 c1e3 d4b6 a2a4 f7f5 e4e5 e7g6 a1d1 b6e3 e1e3 d8e7 d1e1 d6e5 f4e5 d7e6 c3e2 c6c5 b2b3 a8d8 h2h3 d8d5 e2f4 g6f4 f1f4 d5d8 g1h1 g7g5 f4f2 f8f7');
INSERT INTO evaluations VALUES(7343,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(7343,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(7343,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(7343,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(7343,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(7343,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(7343,36,'cp 4','d2d3 d7d6 e1g1 e8g8 c1g5 b4c3 b2c3 h7h6 g5h4 d8e7 f1e1 g7g5 h4g3 c8g4 a4c6 b7c6 d3d4 f6d7 h2h3 g4f3 d1f3 a6a5 f3f5 f8e8 a1b1 e8b8 a2a4 c6c5 d4e5 d7e5 b1b8 a8b8 f2f4 e5g6 f4g5 e7g5 f5g5 h6g5');
INSERT INTO evaluations VALUES(7344,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7344,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7344,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7344,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(7344,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7344,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7344,36,'cp 0','d7d6 e1g1 e8g8 c1g5 b4c3 b2c3 h7h6 g5h4 d8e7 f1e1 g7g5 h4g3 c8g4 a4c6 b7c6 d3d4 f6d7 h2h3 g4f3 d1f3 a6a5 f3e2 e7e6 d4e5 d7e5 g3e5 e6e5 e2c4 e5b5 c4d4 b5e5');
INSERT INTO evaluations VALUES(13881,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(13881,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(13881,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(13881,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(13881,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(13881,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(13881,36,'cp 25','c8e6 d1e2 f8e7 c1g5 e8g8 g5f6 e7f6 c3d5 c6a5 d5f6 d8f6 c4e6 f6e6 e1g1 e6c4 e2d2 b7b6 f1e1 a8c8 c2c3 a5b7 e1e3 f8d8 b2b3 c4e6 a1d1 b7c5 c3c4 h7h6 f3h4 e6g4 h4f5 c5e4 f5e7 g8h7 e3e4 g4e4 e7c8 d8c8');
INSERT INTO evaluations VALUES(19349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19349,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19349,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(19349,36,'cp 4','a7a6 b5c6 d7c6 g1f3 e7e5 e1g1 d8c7 f1e1 g8f6 d2d3 f8e7 a2a4 h7h6 b1a3 c8e6 a3c4 e6c4 d3c4 c7d6 d1c2 d6e6 c1e3 e8c8 a1b1 f6g4 h2h3 g4e3 e1e3 g7g5 b2b4 h6h5 b4b5 g5g4 b5c6 g4f3 b1b7 f3g2 e3e1 e6h3');
INSERT INTO evaluations VALUES(29416,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(29416,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(29416,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(29416,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(29416,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(29416,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(29416,36,'cp 54','c2c4 c7c6 d1b3 d8b6 c4c5 b6b3 a2b3 g7g5 f4c7 b8d7 b3b4 a7a6 h2h3 f6e4 b1c3 f8g7 f1d3 f5g6 d3e4 d5e4 f3d2 f7f5 g2g4 f5f4 d2e4 a8c8 c7a5 d7f6 e4f6 g7f6 e1e2 h6h5 f2f3 f4e3 e2e3 e8d7 h1d1 h5g4 h3g4');
INSERT INTO evaluations VALUES(31127,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31127,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31127,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31127,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31127,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31127,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31127,36,'cp 0','g8f6 h2h3 g4f3 e2f3 d5d4 e4e5 d4c3 e5f6 c3b2 c1b2 e7f6 d2d4 f8e7 e1g1 e8g8 a1b1 b8d7 c2c4 f8e8 g2g3 d8c7 g1g2 b7b6 d1a4 a8c8 f1e1 c6c5 b1d1 c5d4 b2d4 h7h6 c4c5 d7c5 d4c5 c7c5 d1d7 c5b4 a4d1 b4c5 d1a4');
INSERT INTO evaluations VALUES(35128,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(35128,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(35128,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(35128,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(35128,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(35128,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(35128,36,'cp 67','f8c5 b1c3 g8e7 c3b5 c5b6 d2d4 e5d4 e4e5 f6e6 b5d4 b6d4 d1d4 b8c6 d4c3 e7d5 c3b3 d5b4 e1g1 e6b3 c2b3 b7b6 a2a3 c8a6 e2a6 b4a6 b3b4 c6e7 c1g5 c7c6 f3d4 a6c7 g5e7 e8e7 d4f5 e7e6 f5g7 e6e5 a1d1 c7d5 g7h5 e5f5 h5g3 f5g6 f2f3 f7f5 g1f2 a8e8 d1d4 a7a5 h2h4 f5f4 g3e4 a5b4 a3b4');
INSERT INTO evaluations VALUES(33239,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(33239,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(33239,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33239,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(33239,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(33239,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(33239,36,'cp 55','c2c3 b8c6 b2b4 c5b6 b4b5 c6a5 d2d4 d7d6 d1c2 g8e7 d4e5 d6e5 c3c4 c7c6 e1g1 h7h6 c4c5 b6c7 b5c6 b7c6 f1d1 e7g6 g2g3 e8g8 c1e3 c8g4 b1d2 f8d8 g1g2 g6f8 h2h3 g4h5 d2f1 f8e6 d1d8 a8d8 a1d1 h5g6 h3h4 e6d4 e3d4 e5d4');
INSERT INTO evaluations VALUES(37771,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(37771,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(37771,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(37771,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(37771,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(37771,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(37771,36,'cp -17','e7e6 c4e2 g8e7 c2c3 b7b6 e1g1 e8g8 d2d4 c8b7 e4e5 c5d4 f3d4 a7a6 e2f3 b7d5 c3c4 d5f3 d4f3 d7d5 c4d5 e7d5 b1c3 d5c3 b2c3 d8c8 c1a3 f8d8 d1e1 b8d7 a1d1 d7c5 d1d8 c8d8 a3c5 b6c5 f3d2 g7f8 d2e4');
INSERT INTO evaluations VALUES(37626,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(37626,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(37626,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(37626,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(37626,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(37626,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(37626,36,'cp 0','c4e2 g8e7 c2c3 b7b6 b1a3 e8g8 e1g1 d7d6 d2d3 c8b7 d1e1 a7a6 e1h4 b6b5 a3c2 d8d7 f4f5 e6f5 f3g5 h7h6 g5h3 g8h7 h3g5');
INSERT INTO evaluations VALUES(39532,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(39532,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(39532,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(39532,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(39532,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(39532,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(39532,36,'cp 133','b1c3 d7d6 c1e3 a7a6 h2h3 b7b6 a2a4 e7e6 d1d2 c8b7 f1c4 b8d7 e1g1 g8f6 d4d5 e6e5 c4d3 h6h5 f3g5 e8g8 f2f3 g7h6 h3h4 d7c5 b2b4 c5d3 c2d3 c7c6 d5c6 b7c6 b4b5 c6d7 d3d4 a6b5 a4b5 d8e7 d4e5 d6e5 e3b6');
INSERT INTO evaluations VALUES(39531,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(39531,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(39531,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(39531,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(39531,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(39531,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(39531,36,'cp 104','e7e6 c1e3 d7d5 d1d2 d5e4 c3e4 b8d7 e1c1 a7a6 h2h4 b7b6 e4c3 g8e7 f1c4 d7f6 h4h5 g6g5 f3e5 b6b5 c4b3 c8b7 f2f4 g5f4 e3f4 c7c5 d4c5 d8d2 d1d2 e7c6 h1e1 e8g8 e5f7 f8f7 b3e6 a8e8 e6f7 g8f7 e1e8 f7e8 d2e2 e8f7 c3e4 f6e4 e2e4 c6d4');
INSERT INTO evaluations VALUES(38982,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(38982,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(38982,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(38982,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(38982,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(38982,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(38982,36,'cp 111','c1e3 d7d5 d1d2 d5e4 c3e4 b8d7 e1c1 a7a6 g2g3 b7b6 d4d5 d7f6 d5e6 d8d2 e4d2 c8e6 f1c4 e6c4 d2c4 f6d7 e3d4 g7d4 h1e1 e8f8 f3d4 d7f6 a2a4 g8e7 a4a5 e7d5 e1e5 b6a5 d4c6 d5b6 c6a5 b6d7 e5e1 d7b6 f2f3 f8g7 e1e7 f6d5 c4b6 d5e7 b6a8 h8a8 a5c4');
INSERT INTO evaluations VALUES(42473,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42473,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42473,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(42473,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(42473,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(42473,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42473,36,'cp -6','c7c5 e2e3 c5d4 e3d4 d7d5 g1f3 e8g8 f1d3 d5c4 d3c4 a7a6 a2a3 b4d6 f4g5 b7b5 c4a2 d6e7 e1g1 c8b7 d4d5 f6d5 a2d5 e6d5 g5e7 d8e7 c3d5 e7d8 d5e3 b8c6 a1c1 c6e7 d1d8 f8d8 c1c7 b7f3 c7e7 f3h5 f2f3 h5g6 f1d1');
INSERT INTO evaluations VALUES(47194,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(47194,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47194,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47194,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47194,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47194,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47194,36,'cp 0','e7e6 c2c4 b7b6 g1f3 f8d6 e1g1 c8b7 b2b3 e8g8 b1c3 b8d7 a2a4 f6e4 c1b2 e4c3 b2c3 a7a5 c3b2 d8e7 a1c1 a8d8 c4d5 e6d5 e2b5 d8c8 b5d3 g7g6 c1c2 c7c6 d1e2 f8e8 h2h3 c8d8 c2c1 d8c8');
INSERT INTO evaluations VALUES(48348,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48348,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48348,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48348,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48348,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48348,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48348,36,'cp 0','c2c4 b7b6 g1f3 f8d6 e1g1 c8b7 b2b3 e8g8 c4d5 e6d5 c1a3 c7c5 b1c3 b8d7 a1c1 a8c8 e2d3 f6e4 d4c5 e4c3 c1c3 b6c5 c3c2 f8e8 a3b2 h7h6 d3b5 a7a6 b5d3 d7f6 d3f5 c8a8 f5d3');
INSERT INTO evaluations VALUES(48218,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(48218,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(48218,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(48218,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(48218,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(48218,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48218,36,'cp -39','f8e7 e2e3 a7a6 f1d3 b8d7 g1f3 c7c5 e1g1 b7b5 d4c5 d7c5 b2b4 c5d7 e3e4 d8c7 d1d2 c8b7 e4d5 a8c8 d5e6 f7e6 c3e2 b7f3 g2f3 d7e5 e2d4 c8d8 d3e2 e5c4 d4e6 c7e5 e6g7 e8f8 f3f4 e5a1 g7e6 f8f7 e6d8 h8d8 e2c4 b5c4 d2e2 a1a3 e2c4 f7g7');
INSERT INTO evaluations VALUES(40937,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(40937,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(40937,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(40937,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(40937,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(40937,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(40937,36,'cp -17','e2e3 a7a6 f1d3 b8d7 g1f3 c7c5 d4c5 d7c5 e1g1 c8d7 b2b4 c5a4 c3a4 d7a4 d1e2 e8g8 c2c4 d5c4 d3c4 h7h6 g5h4 f6e4 h4e7 d8e7 e2b2 e4d6 c4e2 f8d8 a1c1 a4b5 f1d1 b5e2 b2e2 a8c8 h2h3 d6e4 d1d8 e7d8 e2d1 d8d1 c1d1 g7g5 d1d7 c8c1 g1h2');
INSERT INTO evaluations VALUES(48490,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(48490,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(48490,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(48490,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48490,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(48490,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(48490,36,'cp 30','f1g2 f8b4 c1d2 a7a5 d1c2 c7c5 a2a3 b4d2 b1d2 c5d4 d2c4 e8g8 e1g1 d4d3 c2d3 d8d3 e2d3 b8c6 a1c1 f6d5 f1e1 a8a6 f3e5 c6e5 c4e5 f7f6 g2d5 e6d5 e5f3 a6e6 f3d4 e6e1 c1e1 c8d7 e1c1 f8c8 c1c8 d7c8 d4b5 c8d7 b5d6 b7b5 f2f4 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(57554,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(57554,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(57554,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(57554,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(57554,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(57554,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(57554,36,'cp 90','b8c6 g1f3 d7d5 f4g3 c8g4 b1d2 e7e6 e2e3 f8d6 c2c4 g8e7 a2a3 e8g8 f1d3 d8d7 b2b4 g4f5 d3e2 c6d8 f3h4 f5e4 f2f3 e4g6 h4g6 d6g3 h2g3 h7g6 e1f2 c7c6 c4d5 e6d5 d2b3 d8e6 a1c1 e7f5');
INSERT INTO evaluations VALUES(23381,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(23381,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(23381,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(23381,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(23381,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(23381,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(23381,36,'cp 101','g1f3 d7d5 c2c4 e7e5 c4d5 e5f4 d5c6 b7c6 d1c2 f8d6 c2c6 c8d7 c6c1 g8e7 g2g3 e8g8 f1g2 a8b8 b1c3 d7f5 c1d2 g7g5 e1g1 d8d7 a1c1 g8h8 d4d5 h7h5 f3d4 f5h3 f1d1 h5h4 c3e4 f4g3 h2g3 h4g3 e4d6 h3g2 g1g2 c7d6');
INSERT INTO evaluations VALUES(23380,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(23380,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(23380,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(23380,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(23380,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(23380,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(23380,36,'cp 68','d7d5 f4g3 c8g4 b1d2 h7h5 h2h3 g4f3 d2f3 e7e6 e2e3 d8d7 a2a3 g8e7 c2c4 e7f5 g3h2 f8d6 h2d6 f5d6 c4c5 d6e4 d1c2 g7g5 f1d3 e8c8 b2b4 g5g4 f3d2 f6f5 b4b5 c6a5 c2a4 b7b6 d2b3 g4h3');
INSERT INTO evaluations VALUES(10028,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(10028,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(10028,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(10028,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10028,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10028,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10028,36,'cp 0','c6d4 f3d4 c5d4 c2c3 d4b6 c1g5 c7c6 b5a4 h7h6 g5h4 d7d5 b1d2 c8e6 d1e2 b6c7 a4c2 d8d7 h4g3 d5e4 d3e4 d7e7 d2f3 e7c5 c2b3 c5e7 b3c2');
INSERT INTO evaluations VALUES(16929,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(16929,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(16929,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(16929,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(16929,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(16929,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(16929,36,'cp 213','e5d4 d1d4 f8d6 c1e3 e8g8 f2f3 f8e8 b1c3 d6e5 d4d2 c7c6 e1c1 c6d5 e4d5 a7a6 e3d4 b7b5 d4e5 e8e5 d2d4 d8c7 f1d3 c8b7 f3f4 e5e8 h1e1 b5b4 d4b4 e8e1 d1e1 a6a5 b4e7 c7f4 c1b1 f4h2 c3e4 b7d5 e4f6 g7f6 e7d7 h2g2');
INSERT INTO evaluations VALUES(16922,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(16922,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(16922,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(16922,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(16922,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(16922,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(16922,36,'cp 241','d1d4 f8d6 b1c3 e8g8 c1g5 f8e8 g5f6 d8f6 d4f6 g7f6 e1c1 a7a6 f2f3 d6e5 c3e2 g8f8 g2g4 b7b5 e2g3 c7c6 g3h5 c6d5 e4d5 d7d6 f1d3 b5b4 d3h7 a6a5 h7d3 a5a4 d3b5 e8d8 b5c6 a8a5 f3f4 c8g4 f4e5 g4h5');
INSERT INTO evaluations VALUES(16921,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(16921,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(16921,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(16921,33,'cp 204',NULL);
INSERT INTO evaluations VALUES(16921,34,'cp 232',NULL);
INSERT INTO evaluations VALUES(16921,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(16921,36,'cp 249','f8d6 f2f3 d8e7 b1c3 d6e5 d4d3 c7c6 c1g5 d7d6 e1c1 e8g8 g5h4 e7c7 h4g3 c6d5 g3e5 d6e5 e4d5 c8d7 d3d2 c7c5 f1d3 f8e8 h1e1 b7b5 d5d6 a8c8 c3e4 f6e4 d3e4 b5b4 g2g4 c5b6 c1b1 c8d8 h2h4 d7e6 d2d3 g7g6');
INSERT INTO evaluations VALUES(10740,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(10740,31,'cp 245',NULL);
INSERT INTO evaluations VALUES(10740,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(10740,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(10740,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(10740,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(10740,36,'cp 252','b1c3 d8e7 f2f3 d6e5 d4d3 c7c6 c1g5 h7h6 g5h4 e7d6 d5c6 d7c6 d3d6 e5d6 h4f2 h6h5 c3e2 h5h4 e1c1 d6c7 e2d4 c8e6 f1d3 a7a6 d4e6 f7e6 d3c4 e6e5 f2c5 f6h5 d1d2 b7b5 c4b3 a8d8 d2d8 e8d8 h1d1 d8c8 b3e6 c8b7');
INSERT INTO evaluations VALUES(10739,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(10739,31,'cp 258',NULL);
INSERT INTO evaluations VALUES(10739,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(10739,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(10739,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(10739,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(10739,36,'cp 247','d8e7 f2f3 d6e5 d4d3 e8g8 c1e3 f8e8 e1c1 d7d6 e3d4 a7a6 g2g4 b7b5 c3e2 b5b4 d3d2 a6a5 e2g3 e5d4 d2d4 a8b8 h2h4 a5a4 c1b1 f6d7 g3f5 e7e5 f1c4 e5d4 f5d4 d7e5 c4e2 g8f8 b1c1 c8d7 d1e1 e5g6 h4h5 g6f4 h5h6');
INSERT INTO evaluations VALUES(17703,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(17703,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17703,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17703,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(17703,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(17703,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(17703,36,'cp 33','b5c6 b7c6 g1f3 e7e6 c2c4 c6c5 c4d5 e6d5 b1c3 c5d4 e3d4 f8d6 c3d5 g8e7 d5c3 e8g8 e1g1 a8b8 f1e1 c8b7 h2h3 h7h6 f3e5 e7g6 a2a3 d6e5 d4e5 f8e8 e5e6 e8e6 d1d8 b8d8 e1e6 f7e6 c1e3 g6h4 f2f3 h4f5 a1d1 d8d1 c3d1 b7d5 e3c5 f5d6 c5d6');
INSERT INTO evaluations VALUES(17702,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(17702,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(17702,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(17702,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(17702,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(17702,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(17702,36,'cp 36','b7c6 c2c4 e7e6 g1f3 c6c5 c4d5 e6d5 b1c3 c5d4 e3d4 f8d6 c3d5 g8e7 d5c3 c8b7 e1g1 e8g8 c1g5 a8b8 f1e1 b7f3 d1f3 b8b2 c3d5 f7f6 g5c1 b2b8 d5e7 d6e7 f3d3 d8d7 c1f4 e7d6 f4d6 d7d6 h2h3 b8b4 a1d1 f8d8 d3c3 b4d4 d1d4 d6d4');
INSERT INTO evaluations VALUES(17968,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(17968,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(17968,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17968,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(17968,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(17968,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(17968,36,'cp 39','g1f3 e7e6 c2c4 c6c5 c4d5 e6d5 b1c3 c5d4 e3d4 f8d6 c3d5 g8e7 d5c3 e8g8 e1g1 a8b8 f1e1 c8b7 c1g5 b7f3 d1f3 b8b2 c3d5 f7f6 g5c1 b2b8 d5e7 d6e7 f3d3 b8b6 d3c4 f8f7 c1d2 d8d7 g2g3 b6c6 c4b3 c6b6 b3f3 b6b8 d2e3 b8e8');
INSERT INTO evaluations VALUES(17969,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(17969,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(17969,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(17969,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(17969,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17969,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(17969,36,'cp 40','e7e6 c2c4 c6c5 c4d5 e6d5 b1c3 c5d4 e3d4 f8d6 c3d5 g8e7 d5c3 e8g8 e1g1 e7g6 f1e1 c8b7 d1d3 b7f3 d3f3 d8h4 g2g3 h4d4 c1e3 d4b4 e1e2 g6e5 f3e4 a8d8 a2a3 b4e4 c3e4 f8e8 g1g2 f7f6 e4d6 c7d6 a1e1 e5c4 e3d4 e8e2 e1e2');
INSERT INTO evaluations VALUES(29202,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29202,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(29202,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(29202,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(29202,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(29202,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29202,36,'cp 26','g3e4 b8d7 c1d2 f8e7 e1c1 e8g8 e4f6 d7f6 c1b1 c6c5 h1g1 d8d5 g2g4 f6e4 c2c4 e4f2 c4d5 f2d3 g4g5 g8h7 g5h6 g7h6 d2h6 h7h6 d1d3 e6d5 f3e5 c5d4 d3g3 e7d6 e5f7 f8f7 g3g6 h6h5 g6d6 f7f5 d6d7 f5g5 g1d1 a8g8 d7h7 h5g4 h7b7 g4f3 d1d4 f3e2 b7a7');
INSERT INTO evaluations VALUES(34482,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(34482,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(34482,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(34482,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(34482,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(34482,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(34482,36,'cp 185','b1c3 d5d8 c1e3 c5d4 f3d4 a7a6 d1f3 f8b4 e1c1 b8d7 d4f5 e6f5 c3d5 g8f6 d5b4 d8e7 b4d5 f6d5 f3d5 d7f6 d5d6 e7d6 d1d6 c8e6 f1e2 f6d5 e2f3 d5e3 f2e3 e8e7 d6b6 a8b8 h1d1 a6a5 b2b3 g7g5 d1d4 h7h5 b6b7 b8b7 f3b7 h8b8');
INSERT INTO evaluations VALUES(34477,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(34477,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(34477,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(34477,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(34477,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(34477,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(34477,36,'cp 175','d5d8 c1e3 c5d4 f3d4 a7a6 d1f3 f8b4 e1c1 b8d7 d4f5 e6f5 c3d5 g8f6 d5b4 d8e7 b4d5 f6d5 f3d5 d7f6 d5d6 e7d6 d1d6 c8e6 f1e2 f6e4 d6d4 e8e7 f2f3 e4f6 b2b3 a8d8 c2c4 d8d4 e3d4 h8d8 h1d1 f6d7 c1b2 f5f4 d1e1 f7f6 b2c3 e7f7 e2d3 d7e5 d4e5 f6e5 e1e5 g7g6');
INSERT INTO evaluations VALUES(37648,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(37648,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(37648,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(37648,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(37648,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(37648,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(37648,36,'cp 55','g1f3 g8f6 c1g5 d7d5 e2e3 e8g8 a1c1 h7h6 g5h4 b4e7 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 e1d1 e7c5 d1e2 b8c6 h1d1 c5e7 c4b5 f8d8 h4f6 e7f6 d1d8 c6d8 c3e4 f6b2 c1c7 b2a3 f3e5 a7a6 b5e8 b7b5 e5f7 d8f7 e8f7 g8f8 f7h5 a3e7');
INSERT INTO evaluations VALUES(37646,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(37646,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(37646,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(37646,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(37646,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(37646,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(37646,36,'cp 47','g8f6 c1g5 d7d5 e2e3 h7h6 g5h4 e8g8 d1c2 b4e7 f1e2 b8d7 e1g1 b7b6 c4d5 c6d5 f1c1 c8b7 e2d3 a7a6 c3e2 e7d6 h4g3 d6g3 h2g3 a8c8 c2b3 d8e7 b3a4 a6a5 d3b5 f8d8 a4b3 c8c1 a1c1 d8c8 c1c8 b7c8 b3a4 e7b4 a4b4');
INSERT INTO evaluations VALUES(39862,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(39862,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(39862,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(39862,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(39862,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(39862,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(39862,36,'cp 58','d1b3 c6c5 a2a3 b4a5 d4c5 b8a6 b3c2 a5c3 c2c3 a6c5 g2g4 e8g8 b2b4 c5a4 c3c2 b7b5 g4g5 f6h5 c4b5 a4b6 c2e4 d8c7 c1d2 c8b7 e4g4 g7g6 f1g2 e6e5 e1g1 d7d5 f1c1 b6c4 e2e3 a8c8 a3a4 c7b6 g4d7 c8c7 a4a5 c7d7 a5b6 a7b6 d2e1 f8c8 c1c4 d5c4');
INSERT INTO evaluations VALUES(40372,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(40372,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(40372,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(40372,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(40372,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(40372,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(40372,36,'cp 0','h7h6 g5h4 e8g8 d1d2 f6e4 c3e4 d6d5 d4b5 a7a6 b5c3 d5e4 d2d8 f8d8 h4e7 d8d7 a1d1 d7d1 e1d1 b7b5 c4b3 b8c6 b3d5 g7c3 d5c6 a8a7 e7a3 c3d4 c6e4 d4f2 d1e2 f2h4 h1d1 c8g4 e4f3 g4e6 f3d5 e6g4');
INSERT INTO evaluations VALUES(41845,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(41845,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(41845,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(41845,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(41845,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(41845,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(41845,36,'cp 198','e5f3 e8g8 f1e2 b8c6 e1g1 h7h6 c2c3 f8e8 c1f4 f6h5 f4g3 a7a6 b1d2 c5a7 d3d4 c6e7 b2b4 e7g6 a2a4 g6f4 e2c4 c7c6 f1e1 d6d5 e4d5 c6d5 e1e8 d8e8 c4f1 c8g4 d1b3 a8c8 f3e5 g4e6 b4b5 a6b5 g3f4 h5f4 b3b5 e8b5');
INSERT INTO evaluations VALUES(41908,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(41908,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(41908,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(41908,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(41908,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(41908,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(41908,36,'cp 192','e8g8 f1e2 h7h6 e1g1 f8e8 f1e1 b8c6 c1f4 f6h5 f4d2 a7a6 c2c3 c5a7 e2f1 f7f5 e4f5 c8f5 d2e3 a7e3 e1e3 e8e3 f2e3 g8h8 b1d2 h5f6 d1b3 c6a5 b3a4 c7c5 a4f4 f5h7 a1e1 d8b6 e3e4 a8e8 e4e5 d6e5 f3e5');
INSERT INTO evaluations VALUES(43000,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43000,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(43000,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43000,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(43000,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(43000,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43000,36,'cp 24','c2c4 d7d5 e2e3 e8g8 b1c3 b8d7 c4c5 c7c6 f1d3 b7b6 b2b4 a7a5 a2a3 c8a6 e1g1 d8c8 h2h3 c8b7 c3a4 a6d3 d1d3 a8a7 c5b6 d7b6 a4c5 b7a8 d3d1 b6c4 f3e5 c4e5 f4e5 f8c8 a1b1 f6d7 c5d7 a7d7 d1b3 a5b4 a3b4');
INSERT INTO evaluations VALUES(43387,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(43387,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(43387,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43387,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43387,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(43387,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(43387,36,'cp 18','d2d4 c5d4 f3d4 e7e5 d4b3 c8e6 d1f3 b8d7 f1d3 a8c8 c1d2 d7b6 e1g1 f8e7 a2a4 e8g8 f1d1 d8d7 a4a5 b6c4 d2c1 d7d8 a1a4 f8e8 a4b4 d8c7 c3d5 e6d5 e4d5 e5e4 d3e4 f6e4 f3e4 e7f6 e4d3 c4b2 c1b2 f6b2 c2c4 b2e5 d3f3');
INSERT INTO evaluations VALUES(43386,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43386,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(43386,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43386,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43386,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(43386,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(43386,36,'cp 35','c5d4 f3d4 e7e5 d4b3 c8e6 c1e3 b8d7 g2g4 a8c8 d1d2 h7h5 g4g5 c8c3 d2c3 f6e4 c3b4 d6d5 b4b7 f8e7 f1g2 e4d6 b7c6 d5d4 e3d2 e8g8 d2a5 d8e8 e1g1 d6f5 c2c3 d4c3 a5c3 e7g5 a1e1 e8e7 c3e5 d7e5 e1e5 g5f6 e5e4 f6b2');
INSERT INTO evaluations VALUES(43422,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(43422,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(43422,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(43422,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(43422,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(43422,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(43422,36,'cp 37','f3d4 e7e5 d4b3 c8e6 c1e3 b8c6 d1f3 a8c8 e1c1 c6a5 b3c5 a5c4 f1c4 e6c4 c5d3 d8a5 a2a3 c4e6 e3d2 f8e7 f3e2 a5c7 f2f4 e8g8 c1b1 e5f4 d3f4 c7c4 e2e3 b7b5 g2g4 f6d7 h3h4 a6a5 g4g5 b5b4 a3b4 a5b4 b2b3 c4c7');
INSERT INTO evaluations VALUES(54325,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(54325,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54325,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54325,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54325,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54325,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(54325,36,'cp 5','e6d5 b1c3 b8c6 f1b5 f8b4 g1e2 g8e7 e1g1 e8g8 a2a3 b4c3 e2c3 c8e6 f1e1 a7a6 b5c6 e7c6 c3a4 f8e8 c1e3 c6a5 e3d2 a5c4 b2b3 c4d2 d1d2 d8d6 a4c5 b7b6 c5e6 e8e6 e1e6 d6e6 h2h3 e6d7 a1b1 a8e8 b1e1 e8e1 d2e1 h7h6');
INSERT INTO evaluations VALUES(54585,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(54585,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(54585,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(54585,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54585,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54585,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(54585,36,'cp 2','b1c3 b8c6 f1b5 f8b4 g1e2 g8e7 h2h3 c8f5 e1g1 e8g8 f1e1 h7h6 c1e3 a8c8 a2a3 a7a6 b5d3 b4c3 e2c3 f5d3 d1d3 f8e8 c3a4 e7g6 a4c5 c8c7 e3d2 c7e7 e1e7 e8e7 d3c3 g6f8 a1e1 e7e1 d2e1 d8f6 c3b3');
INSERT INTO evaluations VALUES(54583,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54583,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54583,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54583,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54583,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(54583,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54583,36,'cp 9','b8c6 f1b5 f8b4 g1e2 g8e7 e1g1 e8g8 h2h3 c8e6 a2a3 b4d6 c1f4 a8c8 f4d6 d8d6 a1c1 c6a5 d1a4 a5c6 b5c6 e7c6 a4b5 b7b6 b5d3 c6e7 c3b5 d6d7 c1c8 f8c8 b5c3 e7g6 f1c1 h7h6 d3d2 c8c6 c1e1 c6c4 e2f4');
INSERT INTO evaluations VALUES(55412,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(55412,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(55412,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(55412,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(55412,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(55412,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(55412,36,'cp 192','d4e5 b8d7 e5d6 f8d6 d1e2 g8e7 b1c3 e8g8 c1e3 e7g6 e1c1 d8e7 g2g3 d7b6 f3d2 d6b4 c3b1 g6e5 a2a3 b4d2 b1d2 b6a4 e2b5 c8d7 b5b3 e5g4 f2f4 g4e3 b3e3 f8d8 f1e2 b7b5 h1e1 b5b4');
INSERT INTO evaluations VALUES(55398,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(55398,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(55398,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(55398,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(55398,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(55398,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(55398,36,'cp 167','b8d7 c1f4 d6e5 f3e5 f8b4 b1d2 d7e5 f4e5 g8f6 c2c3 b4d6 d2c4 d6e5 d1d8 e8d8 c4e5 d8e7 f2f3 c8e6 e1f2 h8d8 f2e3 a7a5 f1e2 f6d7 e5d7 d8d7 f3f4 f7f5 e4f5 e6f5 h1d1 a8e8 e3f2 e8d8 d1d7 d8d7 a1d1 d7d1 e2d1');
INSERT INTO evaluations VALUES(55729,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(55729,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(55729,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(55729,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(55729,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(55729,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(55729,36,'cp 56','c2c4 g8f6 g1f3 g7g6 h2h3 c7c5 f1d3 f8g7 b1c3 e8g8 e1g1 d6d5 c4d5 e6d5 d4c5 b8d7 c3a4 f6e4 d1c2 d8e7 c5c6 b7c6 c2c6 a8b8 c6c7 h7h5 d3e4 d5e4 f3d4 c8b7 b2b3 f8c8 c1a3 c8c7 a3e7 b8c8 e7g5 d7e5 a1d1 e5d3');
INSERT INTO evaluations VALUES(55720,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(55720,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(55720,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(55720,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(55720,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(55720,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(55720,36,'cp 35','g8f6 b1c3 d6d5 g1f3 f8e7 b2b3 e8g8 c1b2 b8d7 f1d3 b7b6 e1g1 c8b7 d1e2 c7c5 c4d5 e6d5 a1c1 f8e8 f1d1 a7a6 d3f5 a8c8 d4c5 b6c5 c3a4 c8c7 b2c3 c7c6 f5h3 e7f8 e2d2 d8e7 h3d7 f6d7 d2d5');
INSERT INTO evaluations VALUES(48044,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(48044,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(48044,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(48044,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(48044,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(48044,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(48044,36,'cp 40','b1c3 d6d5 g1f3 f8e7 b2b3 e8g8 c1b2 b8d7 f1d3 b7b6 e1g1 c8b7 c4d5 f6d5 d1e2 e7d6 c3b5 d5b4 d3c4 d6e7 b5c3 a7a6 a2a3 b4d5 c3d5 e6d5 c4d3 e7d6 e2c2 g7g6 f3e5 d8e7 f2f4 a8c8 f4f5 d7e5 d4e5 d6e5 f5f6');
INSERT INTO evaluations VALUES(56284,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(56284,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56284,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(56284,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(56284,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(56284,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56284,36,'cp 0','g8f6 d2d4 f8d6 f1d3 e8g8 e1g1 h7h6 b1c3 b8c6 c3b5 d6e7 c1f4 a7a6 b5c3 c6b4 f1e1 b4d3 d1d3 f8e8 c3e2 e7f8 e2g3 e8e1 a1e1 a6a5 f3e5 c7c6 c2c3 a5a4 a2a3 f6d7 g3h5 d7f6');
INSERT INTO evaluations VALUES(48743,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48743,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48743,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(48743,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(48743,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(48743,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(48743,36,'cp 15','d2d4 h7h6 b1c3 f8d6 f1d3 b8c6 e1g1 e8g8 c3b5 d6e7 c1f4 a7a6 b5c3 c6b4 f1e1 b4d3 d1d3 f8e8 c3e2 e7f8 e2g3 e8e1 a1e1 c7c6 d3b3 a6a5 a2a4 f6e8 f3e5 e8d6 e5d3 d6c4 b3c3 c8d7 b2b3 c4d6 f4e5 b7b6 e5f4 b6b5 d3c5');
INSERT INTO evaluations VALUES(1600,30,'cp -115',NULL);
INSERT INTO evaluations VALUES(1600,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(1600,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(1600,33,'cp -119',NULL);
INSERT INTO evaluations VALUES(1600,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(1600,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(1600,36,'cp -109','d5d4 c1d2 e7e5 g1f3 d4d8 d1c2 f8d6 a1d1 h7h6 e2e4 g8f6 f1e2 e8g8 d2e3 d8e7 a2a3 f8d8 e1g1 e7e8 h2h3 c8e6 b2b4 a7a6 f1e1 b7b5 e2d3 a8c8 c3e2 c6e7 e2g3 e7g6 g3f5 d6f8 c2b2 c7c6 d3b1 d8d1 e1d1');
INSERT INTO evaluations VALUES(1632,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(1632,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(1632,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(1632,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(1632,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(1632,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(1632,36,'cp -100','c1d2 e7e5 g1f3 d4g4 e2e4 f8c5 h2h3 g4g6 d1a4 c5d6 c3d5 g8e7 f3h4 g6e6 f1e2 e6d7 e2g4 d7d8 g4c8 a8c8 e1g1 d8d7 d5e3 c6d4 a4a7 d7c6 h4f5 d6c5 a7a5 b7b6 a5a6 c6e4 f5g7 e8f8 g7h5 e4g6 h5g3 h7h5 e3c4');
INSERT INTO evaluations VALUES(13293,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(13293,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(13293,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(13293,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(13293,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(13293,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(13293,36,'cp 161','d4d5 c6b8 f1e2 g7g6 a2a4 f8g7 h2h3 e8g8 c1f4 b7b6 d1d2 c8a6 e2b5 a6b5 a4b5 d8c7 e1g1 a7a5 f4h2 b8d7 d2e2 f6h5 f1e1 a8e8 g2g4 g7c3 b2c3 h5f6 e4e5 f6d5 e5e6 d7e5 f3e5 d6e5 h2e5 c7b7 c3c4 d5b4');
INSERT INTO evaluations VALUES(13345,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(13345,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(13345,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(13345,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(13345,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(13345,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(13345,36,'cp 178','c6b8 a2a4 g7g6 h2h3 f8g7 c1f4 e8g8 f1e2 b7b6 e1g1 a7a6 d1d2 f6e8 a1d1 b8d7 f4h6 d7e5 f3h2 g7h6 d2h6 f7f6 f2f4 e5f7 h6h4 c8d7 d1e1 e8g7 h4g3 g8h8 g3e3 g6g5 f4g5 f7g5 h2f3 g5f3 f1f3 f8g8 b2b3 b6b5 a4b5 d7b5 c3b5');
INSERT INTO evaluations VALUES(16488,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(16488,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(16488,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(16488,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(16488,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(16488,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(16488,36,'cp -103','e5d4 f1d3 f8b4 c2c3 d4c3 b1c3 d7d5 e4e5 f6e4 e1g1 b4c3 b2c3 c8g4 f1e1 d8d7 d3e4 d5e4 d1d7 g4d7 e1e4 d7f5 e4h4 e8c8 f3d4 d8d5 c1e3 c6d4 c3d4 b7b6 a2a4 c8b7 f2f3 f5d7 h4f4 d7e6 h2h4 h7h5 f4e4 h8a8 e4f4 a7a5 g2g4 b6b5 a4b5');
INSERT INTO evaluations VALUES(16695,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(16695,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(16695,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(16695,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(16695,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(16695,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(16695,36,'cp -108','f1d3 f8b4 c2c3 d4c3 b1c3 d7d5 e4e5 f6e4 e1g1 b4c3 b2c3 c8g4 f1e1 d8d7 d3e4 d5e4 d1d7 g4d7 e1e4 d7f5 e4h4 e8c8 f3d4 d8d5 f2f3 c6d4 c3d4 b7b6 a2a4 c8b7 h4f4 f5e6 h2h4 h7h5 g1f2 a7a5 f2e3 h8d8 a1b1 c7c6 b1a1 g7g6 c1b2 b6b5 a4b5 d5b5');
INSERT INTO evaluations VALUES(16604,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(16604,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(16604,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(16604,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(16604,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(16604,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(16604,36,'cp -66','f1g2 d7d5 d2d3 f8c5 h2h3 c8e6 b1d2 d8e7 e2e3 e8c8 a2a3 c5b6 g1e2 d5d4 e3d4 e5d4 d2f3 h8e8 e1g1 e6b3 e2d4 c6d4 c2b3 e7e2 b2d4 b6d4 f3d4 e2d1 f1d1 d8d4 g1f1 e8d8 f1e2 c7c6 a1c1 f6d5 g2d5 d4d5 e2e3 d8d7 d3d4 a7a6 b3b4');
INSERT INTO evaluations VALUES(17109,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(17109,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17109,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(17109,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17109,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(17109,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(17109,36,'cp 42','e2e3 e5e4 d2d3 f8b4 g1e2 d7d5 e1g1 d5c4 d3e4 f5e4 d1d8 c6d8 c1d2 a7a5 a2a3 b4c3 e2c3 c8f5 c3b5 d8e6 f1c1 e8c8 d2c3 c7c6 b5d4 e6d4 c3d4 f5e6 g2f1 h8e8 f1c4 e6c4 c1c4 d8d5 a1c1 c8c7 c4c5 e8d8 h2h3 d5c5 c1c5 d8d5 c5d5 c6d5 g3g4 c7d6 g1g2');
INSERT INTO evaluations VALUES(17108,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(17108,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(17108,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(17108,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(17108,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(17108,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(17108,36,'cp 36','e5e4 d2d3 f8b4 g1e2 d7d5 e1g1 d5c4 d3e4 f5e4 d1d8 c6d8 c3e4 f6e4 g2e4 c7c6 f1d1 a7a5 c1d2 d8f7 a1c1 f7e5 a2a3 c8g4 a3b4 g4e2 d1e1 e2f3 d2c3 f3e4 c3e5 e8g8 c1c4 e4f3 h2h3 f8f5 e5c3 a5b4 c4b4 c6c5 b4f4 f5f4');
INSERT INTO evaluations VALUES(21046,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(21046,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(21046,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(21046,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(21046,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(21046,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(21046,36,'cp -28','d5d4 c3b1 e7e5 f1c4 f8d6 c2c3 h7h6 c3d4 e5d4 c4b5 c8d7 b5c6 d7c6 d2d3 g8e7 e1g1 e8g8 h2h3 d6c5 f3e5 c6e8 b1d2 f7f6 e5f3 e7c6 a2a3 e8f7 b2b4 c5e7 d2b3 a7a5 b4a5 d8d7 a3a4 g8h8 a1b1');
INSERT INTO evaluations VALUES(21142,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(21142,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(21142,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(21142,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(21142,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(21142,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(21142,36,'cp -16','c3b1 e7e5 f1c4 f8d6 c2c3 d8f6 d2d3 h7h6 b1d2 g8e7 e1g1 e8g8 a2a3 a7a6 c4a2 e7g6 c3d4 e5d4 f3e1 d6f4 b2b4 g8h8 d2f3 f4d6 c1d2 c8e6 d1c2 e6d7 a1d1 f6e7 a2d5 f7f5 e4f5 d7f5 d5e4');
INSERT INTO evaluations VALUES(25304,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(25304,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(25304,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(25304,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(25304,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(25304,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(25304,36,'cp -55','c6d4 c1e3 g4f3 g2f3 f6h5 g1h1 g7g6 c3e2 d4e2 d1e2 d8f6 f1g1 h5f4 e2e1 f8h6 e1b4 b7b6 b4b5 e8f8 b5c6 f8g7 c6c7 b6b5 e3f4 h6f4 c4b3 a7a5 a2a4 h8c8 c7b7 a8b8 b7d7 b5a4 a1a4 c8c5 d7h3 d6d5 e4d5 b8b3 c2b3 c5d5 b3b4 a5b4 a4b4 d5d3 h1g2');
INSERT INTO evaluations VALUES(25400,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(25400,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(25400,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(25400,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(25400,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(25400,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(25400,36,'cp -57','c1e3 g4f3 g2f3 f6h5 g1h1 g7g6 c3e2 d4e2 d1e2 d8f6 f1g1 h5f4 e2e1 f8h6 e1b4 b7b6 b4b5 e8f8 b5d7 f8g7 d7c7 b6b5 c4b3 a7a5 e3f4 h6f4 a2a4 h8c8 c7d7 b5b4 d7b7 f4g5 g1g3 g5h4 g3g2 f6f3 a1g1 h4g5 b7d5 a8a6 d3d4 g5h4 d4e5 d6e5 d5e5 h4f6 e5b5 a6c6 e4e5 f6e7 b5a5 e7c5');
INSERT INTO evaluations VALUES(26605,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(26605,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(26605,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(26605,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(26605,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26605,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26605,36,'cp 25','e7e6 e2e3 f8d6 f4g3 e6e5 d4e5 d6e5 f1e2 e5g3 h2g3 g8f6 c2c4 d5c4 d2c4 d8e7 a2a3 e8g8 d1c2 f8d8 b2b4 d8d5 a1d1 g4f5 c2b2 d5d1 e2d1 a7a6 e1g1 f5d3 d1e2 d3e2 b2e2 a8d8 c4d2 d8d5 f1e1 h7h6 e3e4 d5d8 e4e5 f6d7');
INSERT INTO evaluations VALUES(26839,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(26839,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26839,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(26839,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(26839,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(26839,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(26839,36,'cp 25','e2e3 f8d6 f4g3 e6e5 d4e5 d6e5 f1e2 e5g3 h2g3 g8f6 c2c4 d5c4 d2c4 e8g8 a1c1 d8d1 e2d1 a8d8 a2a3 a7a6 c4d2 f8e8 e1g1 d8d6 d1a4 g4d7 f1d1 g8f8 b2b4 c6e5 a4d7 d6d7 f3e5 e8e5 d2b3 d7d1 c1d1');
INSERT INTO evaluations VALUES(26838,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26838,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(26838,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26838,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(26838,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26838,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26838,36,'cp 14','f8d6 f4g5 d8g5 f3g5 g4d1 a1d1 c6e7 g5f3 g8f6 c2c4 c7c6 f1d3 a7a5 d1b1 f6d7 g2g3 h7h6 h2h4 g7g6 e1g1 e8g8 a2a3 g8g7 f1c1 b7b6 e3e4 d5e4 d2e4 d6c7 c1c2 f8b8 e4c3 a5a4');
INSERT INTO evaluations VALUES(38780,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38780,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(38780,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38780,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(38780,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38780,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(38780,36,'cp 18','a2a3 b4c3 b2c3 g8f6 c3c4 e8g8 g1f3 c7c5 c4d5 e6d5 d4c5 f6e4 f1d3 e4c5 e1g1 b8c6 a1b1 d8e7 h2h3 c8e6 c1b2 a8c8 d3e2 c5e4 d1e1 f8d8 e2d3 b7b6 f3d4 e4c5 f2f4 f7f6 a3a4 c5d3 c2d3 c6d4 b2d4 e6f5 b1b3 e7d7');
INSERT INTO evaluations VALUES(38788,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(38788,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(38788,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(38788,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(38788,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(38788,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38788,36,'cp 12','b4c3 b2c3 g8f6 c3c4 e8g8 g1f3 c7c5 d4c5 f6e4 c4d5 e6d5 f1d3 e4c5 e1g1 b8c6 a1b1 d8e7 c1b2 c8g4 d3e2 f8d8 f3d4 g4e2 d4e2 e7e4 d1d2 e4c4 f1d1 c5e6 d2d3 c4d3 c2d3 d5d4 h2h3 a8c8 b2d4 c6d4 e2d4 e6d4 e3d4 b7b6 b1c1 c8c1');
INSERT INTO evaluations VALUES(43365,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43365,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43365,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43365,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(43365,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43365,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(43365,36,'cp 25','c8f5 f1d3 f5d3 d1d3 e7e6 e3e4 d5e4 c3e4 f6e4 d3e4 d8a5 e1f1 b8d7 g2g3 f8e7 c1d2 a5c7 d2f4 e7d6 f4d6 c7d6 f1g2 h7h6 h1d1 e8g8 d1d2 a6a5 a2a4 d6b4 e4c2 f8c8 a1c1 a8b8 h2h4 b7b5 c4b5 c6b5 c2c8 b8c8 c1c8 g8h7 a4b5');
INSERT INTO evaluations VALUES(28528,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(28528,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(28528,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(28528,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(28528,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(28528,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(28528,36,'cp 36','f1d3 f5d3 d1d3 e7e6 e3e4 f6e4 c3e4 d5e4 d3e4 f8b4 e1f1 b8d7 g2g3 h7h6 c1f4 e8g8 f1g2 d7f6 e4e2 b4d6 f4e5 d8e7 a1d1 a8d8 b2b3 f6d7 e5d6 e7d6 h2h4 d6c7 d1d2 b7b6 h1c1 a6a5 c1d1 d7f6 f3e5 f6d7 e2e4 d7e5');
INSERT INTO evaluations VALUES(48322,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(48322,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(48322,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(48322,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(48322,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(48322,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(48322,36,'cp -49','c7c5 e2e3 b8c6 f1g2 f8e7 b2b3 e8g8 e1g1 b7b6 d1e2 c8b7 c1b2 a8c8 a1d1 c5d4 f3d4 c6d4 d1d4 e7c5 d4d1 d8e7 c3b1 f8d8 d1d2 c5b4 c2c3 b4c5 f1d1 h7h6 c3c4 b7a6 b1c3 c5b4 d2d4 b4d6 c3b5 a6b5 c4b5 d6e5');
INSERT INTO evaluations VALUES(46228,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(46228,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(46228,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(46228,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(46228,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(46228,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(46228,36,'cp -45','f1g2 c5d4 f3d4 f8b4 e1g1 e8g8 e2e4 b4c3 b2c3 d5e4 c3c4 a7a6 d4b3 d8c7 b3d2 b8c6 c1b2 e6e5 d2e4 f6e4 g2e4 c8e6 d1f3 a8d8 f1e1 c6d4 b2d4 d8d4 e4b7 e6c4 f3c6 c7c6 b7c6 f7f6 e1e4 f8d8 e4d4 d8d4 a1b1 c4a2 b1b7 d4d2');
INSERT INTO evaluations VALUES(4598,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4598,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4598,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4598,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(4598,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(4598,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(4598,36,'cp 0','f6e4 c3e4 d7d5 e4c3 d5c4 d4d5 e5e4 f3d2 c6e5 d2e4 f7f5 c1f4 e5g6 f4c7 d8c7 d5d6 c7d7 d6e7 d7e7 e4d2 b7b5 f1e1 e7c5 d2f3 b5b4 d1d5 c5d5 c3d5 c8b7 d5e7 g6e7 e1e7 b7f3 g2f3 f8d8 a1e1 d8d2 e7b7 a7a5 e1e7 g8h8 g1f1 a8g8 e7e2 d2d1 e2e1 d1d2');
INSERT INTO evaluations VALUES(5015,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(5015,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(5015,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(5015,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(5015,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(5015,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(5015,36,'cp 12','d7d5 f1d3 g8f6 e1g1 a7a5 e4e5 f6d7 d1g4 b4f8 g4g3 g7g6 c3a4 f8g7 f2f4 c8a6 d3a6 a8a6 a1e1 e8g8 b2b3 d8c7 g3e3 f8b8 e1c1 c6c5 c2c4 d5d4 e3e4 c7b7 e4e2 a6a8 a4b2 a5a4 b2a4 h7h5 e2f3 g7f8');
INSERT INTO evaluations VALUES(5696,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(5696,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(5696,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(5696,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(5696,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(5696,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(5696,36,'cp 106','d4d5 c6b8 f1d3 g8f6 h2h3 b8d7 c1g5 e8g8 g1f3 e7e5 d3c2 a7a6 f3d2 c7c6 e1g1 c6d5 c4d5 b7b5 a1c1 h7h6 g5e3 d7b6 c3e2 c8d7 b2b3 b5b4 c2d3 a6a5 e2g3 a5a4 f1e1 h6h5 d3f1 a4b3 a2b3');
INSERT INTO evaluations VALUES(14471,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(14471,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(14471,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(14471,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(14471,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(14471,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(14471,36,'cp 97','b1c3 g7g6 d4d5 c6b8 e2e4 f8g7 f1e2 e8g8 e1g1 a7a5 c1e3 b8a6 d1c2 e7e5 d5e6 c8e6 f3d4 d8e7 f2f3 c7c6 a1d1 a6c5 c2d2 a8d8 e3g5 a5a4 f1e1 e6c8 h2h3 f8e8 e2f1 e7f8 d2f2 h7h6 g5e3 f6h5 f2d2 h5g3');
INSERT INTO evaluations VALUES(14468,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(14468,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(14468,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(14468,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(14468,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(14468,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(14468,36,'cp 85','g7g6 d4d5 c6b8 e2e4 f8g7 f1e2 e8g8 e1g1 e7e6 c1g5 h7h6 g5e3 e6d5 c4d5 f8e8 e2d3 c8g4 h2h3 g4f3 d1f3 b8d7 f3e2 a7a6 a1e1 d7c5 e2c2 c5d3 c2d3 c7c5 d5c6 b7c6 e1d1 d8e7 d3d6 e7d6 d1d6');
INSERT INTO evaluations VALUES(21026,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(21026,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21026,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(21026,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(21026,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21026,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21026,36,'cp 41','c8f5 e2e3 c6b4 f1b5 c7c6 b5a4 d8a5 a4b3 g8f6 a2a3 e7e6 f3e5 h7h6 e1g1 b4a6 c2c4 f8d6 f2f4 a5c7 g2g4 f5h7 c4c5 d6e7 b3c2 a6b8 c2h7 f6h7 b2b4 a7a5 c1b2 a5b4 a3b4 a8a1 d1a1 e8g8 b2c3 h7f6 a1a4 b8d7 h2h3 d7e5 d4e5 f6e4 d2e4 d5e4');
INSERT INTO evaluations VALUES(26569,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(26569,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(26569,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(26569,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(26569,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(26569,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(26569,36,'cp 52','e2e3 e7e6 f1b5 a7a6 b5c6 b7c6 e1g1 a6a5 c2c4 a5a4 f1e1 g8f6 f3e5 f8d6 e5c6 d8c8 b2b4 c8b7 b4b5 d5c4 d2c4 b7b5 c6e5 e8g8 c1a3 f8d8 f2f3 f5g6 d1e2 f6d5 a1c1 d6e7 a3e7 d5e7 c4d2 b5e2 e1e2 a8b8 e5g6 h7g6 c1c5 e7d5 d2e4 g8h8 c5c1');
INSERT INTO evaluations VALUES(21397,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21397,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21397,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(21397,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(21397,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(21397,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(21397,36,'cp 11','f8d6 g1f3 c6d4 f3d4 e6e5 d3b5 e8f8 f4g3 e5d4 d1d4 c7c6 b5e2 d6g3 h2g3 d8b6 b1c3 b6d4 e3d4 g7g6 e1d2 f8g7 e2d3 h7h5 a1e1 g8f6 f2f3 c8d7 c3e2 a7a5 e2c3 b7b5 e1e5 a8e8 a2a3 e8e5 d4e5 f6h7 b2b4 a5b4 a3b4 h7g5');
INSERT INTO evaluations VALUES(5920,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(5920,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(5920,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(5920,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(5920,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(5920,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(5920,36,'cp 11','g1f3 c6d4 f3d4 e6e5 d3b5 e8f8 f4g3 e5d4 d1d4 c7c6 b5e2 d6g3 h2g3 d8b6 b1c3 b6d4 e3d4 g7g6 e1d2 f8g7 a1e1 h7h5 e2d3 g8f6 f2f3 c8d7 b2b3 b7b6 a2a3 a7a5 c3a4 a8b8 b3b4 a5b4 a3b4 c6c5 a4c3 c5b4 c3a2 b4b3 c2b3 b8a8 a2b4');
INSERT INTO evaluations VALUES(22798,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(22798,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(22798,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(22798,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(22798,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(22798,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(22798,36,'cp -78','d7d5 b1c3 d5d4 c3d5 g8e7 f1g2 c8e6 d5e7 f8e7 c1d2 d8d7 f2f4 e5f4 g3f4 f7f5 e4f5 e6f5 d1e2 f5g4 g1f3 e8c8 e1c1 d7d5 c1b1 c6b4 d2b4 e7b4 e2e4 h8e8 e4d5 d8d5 f3e5 d5e5 f4e5 g4d1 h1d1 e8e5 g2h3 c8b8 d1g1 g7g6 a2a3');
INSERT INTO evaluations VALUES(25313,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(25313,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(25313,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(25313,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(25313,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(25313,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(25313,36,'cp 58','c2c3 d8d7 a2a4 f8e7 h2h3 g4e6 c4e6 f7e6 b1d2 e8g8 d2c4 a7a6 c1d2 d7e8 b2b4 f6d7 d1e2 e7f6 c4e3 c6e7 e3g4 e8f7 g2g3 f7g6 b4b5 h7h5 g4h2 a6b5 a4b5 d7c5 d2e3');
INSERT INTO evaluations VALUES(25308,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(25308,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(25308,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(25308,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(25308,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(25308,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(25308,36,'cp 50','f8e7 h2h3 g4e6 c4e6 f7e6 b2b4 a7a6 b1d2 e8g8 a2a4 d8e8 d1b3 c6d8 f3h2 f6d7 d2f3 h7h6 a4a5 e8g6 c1e3 g8h8 g2g3 g6f7 b3a4 d7f6 g1g2 d8c6 b4b5 c6b8 a4c4 c7c6 b5c6 b8c6');
INSERT INTO evaluations VALUES(24056,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24056,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(24056,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(24056,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(24056,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(24056,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(24056,36,'cp 49','h2h3 g4e6 c4b3 e8g8 f1e1 e6b3 d1b3 d8c8 c1g5 f8e8 b1d2 h7h6 g5f6 e7f6 d2f1 c6d8 a1d1 d8e6 d3d4 e6f4 b3c2 c8e6 d4d5 e6d7 f1e3 g7g6 a2a4 a7a5 h3h4 f4h5 g2g3 d7h3 f3h2 f6g7 h2g4 h5f6 g4f6 g7f6 d1b1 h6h5');
INSERT INTO evaluations VALUES(36453,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36453,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36453,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36453,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(36453,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(36453,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36453,36,'cp 11','g1f3 e7e6 f1d3 e6d5 c4d5 c7c6 e1g1 c6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 c1f4 a8c8 a1d1 d8c7 d1c1 c7b6 b2b3 f8e8 c3e4 c8c1 e4f6 g7f6 e1c1 h7h5 d3d2 f6d4 f4g5 f7f6 g5f4 d4c5 f4g3 b6b5 c1d1 b5d7 g1h1 c5b6');
INSERT INTO evaluations VALUES(39549,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(39549,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(39549,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(39549,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39549,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39549,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39549,36,'cp 0','e7e6 b1c3 g8e7 c4b3 d7d5 c1f4 c7c5 h2h4 h7h5 d1f3 b8c6 f3g3 e8g8 g1f3 c6d4 f3d4 c5d4 f4c7 d8d7 c3e2 a7a5 a2a4 d5e4 d3e4 b7b6 c7e5 g7e5 g3e5 d4d3 e1c1 c8a6 e2f4 e7c6 e5g5 c6d4 f4h5 d7c7 h5f6 g8g7 f6h5');
INSERT INTO evaluations VALUES(39000,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(39000,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(39000,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(39000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39000,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(39000,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(39000,36,'cp -12','c4b3 d7d5 e4d5 e6d5 g1e2 g8f6 c2c3 e8g8 e1g1 c7c5 c1e3 f6g4 e3f4 b8c6 d3d4 c5c4 b3c2 b7b5 h2h3 g4f6 a2a3 c8d7 b1d2 a7a5 d2f3 f8e8 f1e1 b5b4 a3b4 a5b4 c2a4 d8b6 a4c6 a8a1 d1a1');
INSERT INTO evaluations VALUES(40108,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40108,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(40108,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(40108,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(40108,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40108,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(40108,36,'cp 22','f1d3 e8g8 h2h3 c8e6 e1g1 b8d7 f1e1 h7h6 g5h4 f8e8 a2a3 a7a5 e1e3 a5a4 a1b1 d8b6 h4g3 e7f8 g3h2 b6d8 c3e2 f6e4 c2c4 f8d6 b1c1 d6h2 f3h2 f7f5 c4d5 e6d5 h2f3 d7f6 f3e5 d8b6 c1b1 e4d6 d3c2 d6c4');
INSERT INTO evaluations VALUES(40112,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(40112,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40112,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40112,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40112,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(40112,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(40112,36,'cp 15','e8g8 h2h3 c8e6 e1g1 b8d7 f1e1 h7h6 g5f4 f8e8 a2a3 a7a6 c3e2 f6e4 c2c3 e6f5 e2g3 e4g3 f4g3 f5d3 d1d3 d7f6 e1e2 e7f8 g3e5 f6e4 a1e1 e8e6 c3c4 d8d7 e5f4 f8d6 f4d6 d7d6 e2e3 d5c4 d3c4 d6d5 c4c2 a8e8 f3e5');
INSERT INTO evaluations VALUES(35771,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(35771,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(35771,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(35771,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(35771,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(35771,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35771,36,'cp 12','h2h3 c8e6 e1g1 b8d7 f1e1 f8e8 c3e2 h7h6 g5h4 f6e4 h4e7 e8e7 e2g3 e4g3 f2g3 d8c7 g1h2 a8e8 d1d2 d7f6 e1e5 e6c8 d2f4 e7e5 f3e5 c6c5 c2c3 c7b6 f4d2 c5d4 c3d4 f6e4 d3e4 d5e4 a1e1 f7f6 e5c4 b6c7 c4e3');
INSERT INTO evaluations VALUES(51347,30,'cp 249',NULL);
INSERT INTO evaluations VALUES(51347,31,'cp 242',NULL);
INSERT INTO evaluations VALUES(51347,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(51347,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(51347,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(51347,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(51347,36,'cp 249','d2d4 b8c6 d4e5 d6e5 d1d8 e8d8 f1b5 c8b7 b5c6 b7c6 f3e5 c6e8 c1f4 g8e7 a2a4 d8c8 a4a5 b6a5 a1a5 f7f6 e5d3 e7c6 a5a4 c6e5 a4a1 e5d3 c2d3 f8b4 f2f3 b4c3 b2c3 a7a5 e1f2 a5a4 h1b1 e8f7 c3c4 g7g5 f4d2 h8d8 f2e2 a4a3 d2e3 a8a4 b1b3 a3a2 b3b2 a4a3 b2a2 a3d3');
INSERT INTO evaluations VALUES(51345,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(51345,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(51345,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(51345,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(51345,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(51345,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(51345,36,'cp 227','b8c6 d4e5 d6e5 d1d8 e8d8 f1b5 c8d7 b5c6 d7c6 f3e5 c6e8 c1f4 g8e7 a2a4 d8c8 a4a5 b6a5 a1a5 f7f6 e5d3 e7c6 a5a4 c8b7 c3d5 c6e5 a4a1 c7c6 d5c3 e5d3 c2d3 f8b4 e1e2 c6c5 h1c1 e8d7 f4e3 a7a5 f2f3 h8d8 b2b3 f6f5 e4e5 d7e6 c3a4 e6b3 a4c5 b4c5 c1c5');
INSERT INTO evaluations VALUES(6134,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(6134,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(6134,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(6134,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(6134,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(6134,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(6134,36,'cp 87','f1e2 g8f6 g1f3 e7e5 e1g1 e8g8 c1e3 f8e8 d4d5 a7a5 f3d2 g8h8 e3g5 h7h6 g5e3 e8f8 a2a3 b7b6 d1c2 f6g8 b2b4 f7f5 f2f3 h6h5 e3f2 g7h6 f1d1 f5e4 c3e4 d7f6 e2d3 f6e4 d2e4 c8f5 c4c5 g8f6 f2h4 b6c5 b4c5 f5e4 d3e4');
INSERT INTO evaluations VALUES(15299,30,'cp 232',NULL);
INSERT INTO evaluations VALUES(15299,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(15299,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(15299,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(15299,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(15299,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(15299,36,'cp 227','b1c3 c8e6 g5e6 f7e6 d1g4 d8d6 c1d2 f8e7 e1g1 a7a6 f1e1 b7b5 c4b3 e8g8 g4e6 d6e6 b3d5 e6d5 c3d5 e7d6 d5c3 b5b4 c3e4 a6a5 g1f1 h7h6 a1c1 c6e7 g2g3 e7f5 c2c3 f8b8 c1b1 f5e7 e1d1 a5a4 d1c1 b4c3 c1c3 a4a3 b2b3 e7d5 c3c4 g8h8 e4c3 d5b4 f1e2 b4c2 b1c1');
INSERT INTO evaluations VALUES(20669,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20669,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20669,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20669,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20669,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20669,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20669,36,'cp 0','g8f6 d2d3 f8e7 f1e1 e8g8 a2a4 d6d5 c4b3 d5e4 d3e4 d8d1 b3d1 f8d8 d1c2 b7b6 b1a3 c8b7 c1f4 c6a5 f4c7 d8d7 c7f4');
INSERT INTO evaluations VALUES(13932,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13932,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13932,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13932,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13932,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13932,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13932,36,'cp 0','d2d3 f8e7 b1d2 e8g8 f1e1 c6a5 c4b5 c8d7 b5d7 d8d7 a2a4 a5c6 d2c4 d7c7 c1f4 f6h5 f4c1');
INSERT INTO evaluations VALUES(28812,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(28812,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(28812,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(28812,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(28812,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(28812,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(28812,36,'cp 3','c2c4 e7e6 f1d3 d7d5 e1g1 f8d6 b1c3 e8g8 b2b3 b8d7 c1b2 a7a6 d1c2 c7c5 c4d5 e6d5 a1c1 a8c8 c2e2 b6b5 d4c5 d7c5 d3f5 c8c7 e2d1 c5e4 f3d4 g7g6 c3e4 d5e4 f5h3 d6e5 c1c7 d8c7 g2g3 f6d5 d1d2 f8d8 f1d1');
INSERT INTO evaluations VALUES(28801,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(28801,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(28801,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(28801,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(28801,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28801,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28801,36,'cp 7','e7e6 f1d3 d7d5 e1g1 f8d6 b2b3 e8g8 c4d5 f6d5 c1b2 b8d7 b1d2 c7c5 a1c1 a8c8 h2h3 d6e7 d1e2 c5d4 f3d4 c8c1 f1c1 d7c5 d3c2 d8d7 d2e4 f8d8 e4c5 e7c5 c2e4 d7e7 c1d1 h7h6 e4f3 e7g5');
INSERT INTO evaluations VALUES(42777,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42777,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42777,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42777,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42777,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42777,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42777,36,'cp 0','e8g8 a2a3 b4c3 d2c3 f8e8 g1f3 e5e4 d3e4 f6e4 d1d4 e4c3 d4c3 d7d6 e2e3 b8d7 f1e2 a7a5 b2b3 b7b6 e1g1 c8b7 f3d4 d7e5 f2f3 d8g5 e3e4 e5g6 f1e1 g6f4 e2f1 f4h3 g1h1 h3f2 h1g1');
INSERT INTO evaluations VALUES(37274,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37274,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37274,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37274,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37274,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37274,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37274,36,'cp 0','a2a3 b4c3 d2c3 f8e8 g1f3 e5e4 d3e4 f6e4 d1d4 e4c3 d4c3 d7d6 e2e3 b8d7 f1e2 a7a5 b2b3 b7b6 e1g1 c8b7 f3d4 d8f6 a1d1 d7c5 c3c2 b7e4 c2d2 e4b7');
INSERT INTO evaluations VALUES(49234,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(49234,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(49234,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(49234,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(49234,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(49234,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(49234,36,'cp -30','c7c5 d4d5 e6d5 c4d5 d7d6 b1c3 f8e7 h4g3 f6h5 e2e3 h5g3 h2g3 e8g8 g1f3 b8d7 f3d2 f8e8 a2a4 e7f8 a4a5 g7g6 f1b5 d8e7 e1g1 h6h5 d2c4 e8d8 b5d7 c8d7 d1f3 a7a6 c4b6 a8b8 a1c1 d7b5 f1e1 d8e8 e3e4 f8g7 c3b5 a6b5');
INSERT INTO evaluations VALUES(51833,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(51833,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(51833,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(51833,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(51833,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(51833,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(51833,36,'cp -31','d7d5 c4b3 a7a5 a2a4 d5e4 d3e4 d8d1 e1d1 g8f6 b1d2 b8d7 g1f3 f8d6 h1e1 e8g8 b3c4 d7c5 b2b3 c8d7 c1a3 b7b5 c4d3 h7h6 d1e2 f8e8 e1d1 d6f8 f3e1 g7g5 g2g3 h6h5 e1g2 c5e6 f2f3 e6d4 e2f2 g5g4 a3f8 g4f3 d2f3');
INSERT INTO evaluations VALUES(53746,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(53746,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(53746,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(53746,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(53746,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(53746,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(53746,36,'cp -38','c4b3 a7a5 a2a4 d5e4 d3e4 d8d1 e1d1 b8d7 g1f3 g8f6 h1e1 f8d6 f3d2 d7c5 b3c4 h7h5 f2f3 h5h4 c4f1 c8e6 d2c4 e6c4 f1c4 f6h5 d1e2 h5f4 e2f1 c5e6 c4e6 f4e6 c1e3 f7f6 b1a3 d6c5 e3c5 e6c5 a3c4');
INSERT INTO evaluations VALUES(56596,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(56596,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(56596,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(56596,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(56596,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(56596,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(56596,36,'cp -27','c8e6 c2c3 f7f6 d1a4 b8c6 f1a6 b7a6 a4c6 e8f7 c3d4 g8e7 c6a6 d8d4 g1e2 d4b6 a6b6 a7b6 d2d4 e5d4 e2d4 a8a2 a1a2 e6a2 e1g1 c7c5 d4b5 e7c8 b5c3 a2c4 f1d1 f8e7 g3f5 h8d8 f5e7 d8d1 c3d1 f7e7 f2f3 e7d7 g1f2 c8d6 d1e3 c4b3 f2e2 d7c6 c1d2');
INSERT INTO evaluations VALUES(598,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(598,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(598,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(598,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(598,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(598,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(598,36,'cp 62','g4e5 f3e5 c6e5 e2e3 b4d2 d1d2 d7d6 f1e2 e8g8 e1g1 f8e8 a1c1 a7a5 b2b3 e5d7 e2d3 d7c5 d3c2 f7f5 b3b4 a5b4 a3b4 c5e4 c2e4 e7e4 c4c5 d6c5 c1c5 c7c6 d2c3 e4e7 b4b5 c8d7 f1d1 c6b5 c5c7 b5b4 c3c2 e7f6 c7b7');
INSERT INTO evaluations VALUES(606,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(606,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(606,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(606,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(606,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(606,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(606,36,'cp 60','f3e5 c6e5 e2e3 b4d2 d1d2 d7d6 f1e2 e8g8 e1g1 a7a5 b2b3 f8e8 a1d1 c8f5 f1e1 h7h6 d1c1 f5g6 e2f1 e5d7 f2f3 d7c5 d2c3 f7f5 c1d1 g6f7 h2h3 b7b6 g1h2 e7h4 c3b2 h4e7 d1d2 e7h4 f4g3 h4g5 b3b4 a5b4 a3b4 c5d7 e3e4 f5f4 g3f2');
INSERT INTO evaluations VALUES(592,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(592,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(592,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(592,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(592,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(592,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(592,36,'cp 39','c6e5 e2e3 b4d2 d1d2 d7d6 f1e2 b7b6 e1g1 a7a5 b2b4 e5d7 e3e4 e8g8 f4g5 e7e6 f2f3 c8b7 f1e1 a5b4 a3b4 h7h6 g5h4 f8e8 e2f1 e6e5 a1d1 b7c6 h4f2 e5g5 f2e3 g5g6 d2b2 d7e5 g1h1 c6d7 c4c5 d6c5 b4c5 b6c5 e3c5 d7c6 b2c3');
INSERT INTO evaluations VALUES(614,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(614,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(614,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(614,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(614,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(614,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(614,36,'cp 67','e2e3 b4d2 d1d2 d7d6 f1e2 a7a5 e1g1 e8g8 b2b3 f8e8 a1d1 h7h6 f1e1 c8f5 d1c1 f5g6 e2f1 e5d7 f2f3 f7f5 d2d5 g8h8 d5d4 h8g8 f1d3 d7e5 d3c2 b7b6 h2h3 e7f7 c2b1 f7f6 f4g3 g6f7 f3f4 e5d7 d4f6 d7f6 b1f5');
INSERT INTO evaluations VALUES(7290,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(7290,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(7290,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(7290,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(7290,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(7290,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(7290,36,'cp 162','b5c6 d7c6 f3e5 c8e6 d2d3 d8e7 g1h1 f6d7 e5d7 e7d7 f2f4 f7f5 e4f5 e6f5 f1e1 c5e7 d1e2 e8d8 c1d2 b7b6 c3e4 d8c8 d2c3 e7h4 e1f1 h8f8 h2h3 c8b7 e2h5 h4d8 c3e5 a6a5 h1h2 g7g6 h5e2 h7h5 e4g3 d8e7 g3f5 f8f5');
INSERT INTO evaluations VALUES(7284,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(7284,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(7284,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(7284,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(7284,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(7284,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(7284,36,'cp 176','d7c6 f3e5 h7h6 a2a3 c5a7 d2d3 d8e7 e5c4 c8e6 c1e3 e6c4 e3a7 c4d3 e4e5 f6d5 c3d5 c6d5 d1d3 a8a7 d3d5 e8g8 a1e1 f8d8 d5e4 a6a5 h2h3 a7a6 f2f4 d8d2 f4f5 a6b6 f5f6 e7c5 g1h1 b6e6 e4b7 c5d5 b7d5 d2d5 f6g7 e6e5 e1e5 d5e5 f1f2 c7c5 h1h2 g8g7 b2b3 e5e1');
INSERT INTO evaluations VALUES(7200,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(7200,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(7200,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(7200,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(7200,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(7200,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(7200,36,'cp 175','f3e5 h7h6 a2a3 c5a7 d2d3 d8e7 e5c4 c8e6 c1e3 e6c4 e3a7 c4d3 e4e5 f6d5 d1d3 a8a7 c3d5 c6d5 d3d5 e8g8 a1e1 f8d8 d5e4 a6a5 h2h3 a7a6 c2c3 a6c6 e1e2 e7c5 f1e1 c5c4 e4c4 c6c4 e5e6 f7e6 e2e6 b7b6 e1e2 g8f8 f2f3 c4c5 e6e7 c5c4 g1h2 a5a4');
INSERT INTO evaluations VALUES(7193,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(7193,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(7193,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(7193,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(7193,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(7193,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(7193,36,'cp 162','h7h6 a2a3 c5a7 d2d3 d8e7 e5c4 c8e6 c1e3 e6c4 e3a7 c4d3 e4e5 f6d5 c3d5 c6d5 d1d3 a8a7 d3d5 e8g8 a1e1 f8d8 d5e4 a6a5 f2f4 a7a6 f4f5 d8d2 e4b7 e7c5 g1h1 c5c6 b7b8 g8h7 f1g1 c6c2 e5e6 f7e6 f5e6 d2e2 e1e2 c2e2');
INSERT INTO evaluations VALUES(8966,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(8966,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(8966,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(8966,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(8966,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(8966,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(8966,36,'cp 55','e2e3 d5c4 f1c4 f6d5 e1g1 d5f4 e3f4 e8g8 f1e1 d8d6 g2g3 b4c3 b2c3 b7b6 d1c2 c8b7 f3g5 g7g6 h2h4 c6e7 c4d3 c7c5 d3e4 b7c6 e4c6 d6c6 d4c5 c6c5 h4h5 a8d8 g5e4 c5c6 a1d1 e7f5 d1d3 c6c7 c3c4 d8d3 c2d3 f8d8 e4f6 g8h8 d3c3 f5d4');
INSERT INTO evaluations VALUES(18366,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(18366,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(18366,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(18366,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(18366,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(18366,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(18366,36,'cp 194','c1f4 g8e7 e1c1 a7a6 c3a4 b6b5 a4b6 a8a7 b6c8 e7c8 h2h4 h7h5 h1h3 a7b7 f1e2 c8b6 c1b1 d8e7 h3c3 e7b4 d4d3 d7d5 d3g3 b4e4 e2d3 e4a4 f4e5 b6c4 c3b3 c4e5 g3e5 e8d8 e5g5 f7f6 g5g3');
INSERT INTO evaluations VALUES(18367,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(18367,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(18367,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(18367,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(18367,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(18367,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(18367,36,'cp 185','g8e7 e1c1 e7c6 d4d3 f8c5 d3g3 e8g8 c1b1 f7f6 f1c4 d8e8 a2a3 e8g6 c3b5 g6g3 f4g3 c6e5 c4e2 a7a6 b5d6 c5d6 d1d6 b6b5 f2f3 e5c4 d6d4 d7d5 e4d5 e6d5 d4d5 f8e8 h1e1 c8e6 e2c4 b5c4 d5a5 e6f7 e1e8 f7e8 b1c1 e8b5 h2h4 a8d8 h4h5 d8d7');
INSERT INTO evaluations VALUES(19351,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(19351,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(19351,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(19351,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(19351,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(19351,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(19351,36,'cp 54','g1f3 e7e6 e1g1 g8e7 f1e1 a7a6 b5f1 d7d5 e4d5 e7d5 d2d4 c5d4 f3d4 d5b4 d4c6 d8d1 e1d1 b7c6 b1a3 a6a5 c1d2 f8e7 a3c4 c8a6 c4a5 a6f1 g1f1 a8a5 a2a3 b4c2 d2a5 c2a1 d1a1 e7d6 a3a4 e8d7 f1e2 f7f6 a5c3 h8a8 a4a5');
INSERT INTO evaluations VALUES(21105,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(21105,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(21105,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(21105,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(21105,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(21105,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(21105,36,'cp -100','d8d4 b1d2 d4f6 g1f3 e7e5 c4d5 f8d6 d5c6 b7c6 d2c4 h7h6 e1g1 g8e7 c1e3 e8g8 d1c2 c6c5 c4d2 f8e8 e3c5 d6c5 c2c5 e7g6 d2c4 c8g4 c5e3 g4f3 e3f3 f6f3 g2f3 g6h4 a1d1 h4f3 g1g2 f3g5 d1d7 g5e4 d7c7 a8c8 c7c8 e8c8 c4e5 c8c2 e5d3 c2d2 f1e1 e4f6 e1e3 f6d5 e3e8 g8h7 d3e5 d2b2 e5f7 b2a2');
INSERT INTO evaluations VALUES(23003,30,'cp -355',NULL);
INSERT INTO evaluations VALUES(23003,31,'cp -340',NULL);
INSERT INTO evaluations VALUES(23003,32,'cp -343',NULL);
INSERT INTO evaluations VALUES(23003,33,'cp -340',NULL);
INSERT INTO evaluations VALUES(23003,34,'cp -363',NULL);
INSERT INTO evaluations VALUES(23003,35,'cp -378',NULL);
INSERT INTO evaluations VALUES(23003,36,'cp -397','d8g5 e5g4 d4b5 d2d4 g5g6 e1g1 b5d4 b1c3 d4c6 g4e3 g8f6 f1e1 e8d8 e3d5 f8c5 c1e3 c5e3 e1e3 d7d6 d5f6 g6f6 c3d5 f6h6 e4e5 c8f5 e5d6 h6d6 d1d2 d8c8 d5e7 c6e7 d2d6 c7d6 e3e7 f5e6 a1d1 h8d8 a2a3 a7a5 b2b3 e6g4');
INSERT INTO evaluations VALUES(28843,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(28843,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28843,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28843,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28843,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(28843,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28843,36,'cp 0','d8d7 d2d4 c5d4 d1d4 g7g6 c1g5 f8g7 d4d2 h7h6 g5f4 b8c6 e1c1 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(32561,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32561,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32561,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32561,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32561,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32561,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32561,36,'cp 0','d2d4 c5d4 d1d4 g7g6 c1g5 f8g7 d4d2 h7h6 g5f6 g7f6 a1d1 b8c6 e1g1 f6g7 c3d5 e7e6 d5e3 a8d8 c2c3 e8g8 d2d3 d7c7 h2h4 c6e7 d3e2 c7c6 e3c2 c6b6 c2e3');
INSERT INTO evaluations VALUES(31435,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(31435,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31435,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(31435,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31435,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(31435,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31435,36,'cp 33','g8f6 d1b3 b7b6 c4d5 e6d5 g1f3 f8d6 f1e2 b6b5 e1g1 e8g8 f1c1 a7a5 b3d1 f8e8 f4d6 d8d6 f3h4 f5g6 b2b3 b8d7 h4g6 h7g6 e2d3 a5a4 b3a4 b5b4 c3e2 d7b6 d1c2 e8c8 a2a3 b4a3 c2b3 b6a4 a1a3 c8b8 b3a2 a4b6 a3a7');
INSERT INTO evaluations VALUES(34740,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(34740,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(34740,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(34740,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(34740,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(34740,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(34740,36,'cp 165','d2d4 b8d7 b1c3 d5e4 c1f4 d7e5 f4e5 g8f6 f1e2 c8f5 e1g1 c7c6 f2f3 e4f3 f1f3 f5g4 f3e3 g4e6 d1e1 e7d7 a1d1 e8c8 c3b5 b7b6 b5c7 f8d6 e5d6 d7d6 c7e6 f7e6 e2c4 h8e8 c4e6 c8c7 c2c3 a7a5 h2h3 h7h6 e6c4 e8e3');
INSERT INTO evaluations VALUES(34739,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(34739,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(34739,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(34739,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(34739,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(34739,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(34739,36,'cp 143','f7f6 e5d3 d5e4 d3f4 b8c6 b1c3 c8f5 g2g4 f5g6 c1e3 e8c8 d4d5 e7b4 d1d2 c6e5 e1c1 g6f7 a2a3 b4a5 c3e4 a5d2 e3d2 g8e7 c2c4 e7g6 h2h4 e5f3 d2e3 d8e8 e4d2 f3e5 f4d3 e5g4 e3a7 f8d6 a7d4 g6f4 c4c5 d6e5 d4e5 f6e5 d3f4 e5f4');
INSERT INTO evaluations VALUES(34698,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(34698,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(34698,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(34698,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(34698,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(34698,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(34698,36,'cp 155','e5d3 d5e4 d3f4 c8f5 b1c3 b8c6 g2g4 f5g6 c1e3 e8c8 d4d5 e7b4 d1d2 c6e5 e1c1 g6f7 a2a3 b4a5 c3e4 a5d2 e3d2 g8e7 c2c4 b7b5 f4e6 f7e6 d5e6 e5c4 f1c4 b5c4 e4g3 d8d4 g4g5 e7d5 g5f6 d5f6 d2e3 d4d1 h1d1 f8e7 e3a7');
INSERT INTO evaluations VALUES(38527,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(38527,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(38527,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(38527,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(38527,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(38527,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(38527,36,'cp 169','c1g5 f7f6 g5e3 g8h6 h2h3 h6f7 f1c4 c7c6 d1d3 d8e7 e1c1 e8g8 d4e5 d6e5 d3d6 e7d6 d1d6 f8e8 h1d1 g8f8 d6d2 c8e6 c4e6 e8e6 h3h4 g7h6 e3h6 f7h6 h4h5 h6f7 f3h4 f8g7 f2f3 g6g5 h5h6 f7h6');
INSERT INTO evaluations VALUES(42501,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(42501,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(42501,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42501,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(42501,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(42501,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(42501,36,'cp 27','d2c3 f6e4 a1c1 b7b6 g2g3 c8b7 d4d5 e4c3 c1c3 d7d6 f1g2 e6e5 e2e4 a7a5 g1e2 b8a6 b2b3 a6c5 d1d2 b7c8 e1g1 c8g4 e2c1 g4d7 f2f4 e8g8 c1d3 d8e7 f4f5 f7f6 h2h3 g8h8 g1h2 c5d3 c3d3 b6b5 f1c1 f8b8');
INSERT INTO evaluations VALUES(42493,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(42493,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(42493,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(42493,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42493,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(42493,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(42493,36,'cp 3','f6e4 d1c2 e4c3 c2c3 b7b6 g1f3 c8b7 g2g3 d7d6 f1g2 e8g8 e1g1 b8d7 f1e1 d7f6 a1d1 c7c5 f3h4 b7g2 h4g2 d8c7 e2e4 c5d4 c3d4 a8d8 b2b3 a7a6 g2e3 b6b5 d1d2 b5c4 e3c4 d6d5 e4d5 d8d5 d4a1 d5d2 c4d2');
INSERT INTO evaluations VALUES(43923,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43923,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(43923,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43923,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43923,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43923,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43923,36,'cp 0','b3c4 g7g6 b1c3 f8g7 a1b1 e8g8 e2e4 e7e6 g1f3 d7d6 f1e2 e6d5 c4d5 f8e8 f3d2 d8a5 b1b3 f6e4 c3e4 e8e4 b3a3 a5b4 a3b3');
INSERT INTO evaluations VALUES(43921,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(43921,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43921,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43921,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43921,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43921,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43921,36,'cp 0','g7g6 b1c3 f8g7 a1b1 e8g8 e2e4 e7e6 g1f3 d7d6 f1e2 e6d5 c4d5 f8e8 f3d2 d8a5 b1b3 f6e4 c3e4 e8e4 b3a3 a5b4 a3b3');
INSERT INTO evaluations VALUES(43905,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(43905,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(43905,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43905,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43905,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43905,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43905,36,'cp 0','b1c3 f8g7 a1b1 e8g8 g1f3 e7e6 e2e4 c8a6 c1g5 e6d5 e4e5 d8a5 g5d2 f6g4 c3d5 a5a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(45861,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45861,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45861,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45861,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45861,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45861,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45861,36,'cp 5','b7b5 c4b3 f8e7 e1g1 e8g8 c3e2 a7a5 a2a4 b5b4 e2g3 c6c5 b3c4 b8c6 c2c3 d8c7 f1e1 c8e6 h2h3 a8b8 c4b5 c6a7 b5a6 a7c6 f3g5 e6c8 a6c8 f8c8 g3f5 c7d7 f2f4 e7f8 g5f3 d7e6 c3c4 f6d7 g1h1 h7h6');
INSERT INTO evaluations VALUES(52356,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(52356,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(52356,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(52356,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(52356,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52356,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(52356,36,'cp -7','e2e3 g8f6 g1f3 d7d5 c1b2 c8g4 c2c4 b8d7 c4d5 c6d5 f1e2 e7e6 a2a3 h7h6 b1c3 a7a6 h2h3 g4h5 e1g1 f8d6 d1b3 e8g8 a1c1 d8e7 c1c2 a8c8 d2d3 c8c7 f1c1 h5g6 c3b1 c7c2 c1c2 e6e5 b1d2 d7b6');
INSERT INTO evaluations VALUES(54378,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(54378,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(54378,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(54378,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(54378,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(54378,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(54378,36,'cp 23','c1f4 f8d6 f4g3 d8b6 d1c2 c8d7 e2e3 g8f6 b1c3 e8g8 f1d3 d6g3 h2g3 h7h6 c2e2 b8c6 e1g1 f8c8 a2a3 d7e8 a1c1 b6d8 d3b1 a7a6 f3e5 c8c7 g3g4 a8c8 e5d3 c6a5 d3c5 d8e7 f2f4 a5c4 e3e4 d5e4 c3e4');
INSERT INTO evaluations VALUES(54363,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54363,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(54363,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54363,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(54363,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(54363,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54363,36,'cp 12','f8d6 f4d6 d8d6 e2e3 g8f6 f1d3 e8g8 e1g1 c8d7 b1c3 b8c6 a1c1 a7a6 d1e2 f8c8 a2a3 c8c7 c1c2 a8c8 f1c1 c6e7 f3e5 d7e8 f2f4 b7b5 e2e1 f6d7 e5d7 e8d7 c3e2 g7g6 c2c7 c8c7 c1c7 d6c7 e1d2 c7b6 g2g4');
INSERT INTO evaluations VALUES(55041,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(55041,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(55041,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(55041,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(55041,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(55041,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(55041,36,'cp -26','d5d4 g1f3 c7c5 f1g2 b8c6 d2d3 g8f6 e2e3 e7e5 e1g1 f8e7 e3d4 c5d4 c1g5 e8g8 a2a3 a7a5 b1d2 c8e6 g5f6 e7f6 d1e2 d8c7 f1e1 f8e8 a1b1 a5a4 b2b4 a4b3 b1b3 e8b8 e1a1 h7h6 b3b5 a8a5 b5a5 c7a5 h2h3');
INSERT INTO evaluations VALUES(2210,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(2210,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(2210,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(2210,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(2210,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(2210,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2210,36,'cp 0','c3d5 f6d8 c2c3 c6e7 d5e3 c5e3 f2e3 c7c6 d3d4 d6d5 c4d3 d5e4 d3e4 e5d4 e3d4 c8f5 e4f5 e7f5 d1d3 f5d6 f3e5 d8c7 d3g3 a8e8 a1e1 e8e6 e5c4 e6g6 g3f4 g6f6 f4g3 f6g6');
INSERT INTO evaluations VALUES(2212,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2212,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2212,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2212,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2212,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2212,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2212,36,'cp 0','f6d8 c2c3 c6e7 d5e3 c5e3 f2e3 c7c6 d3d4 d6d5 c4d3 d5e4 d3e4 e5d4 e3d4 c8f5 e4f5 e7f5 d1d3 f5d6 a1e1 d8c7 c3c4 a8e8 f3e5 f7f6 e5g6 e8e1 f1e1 f8e8 e1e8 d6e8 g1f2 c7h2 d3e4 g8f7 g6h8 f7f8 h8g6');
INSERT INTO evaluations VALUES(3655,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3655,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3655,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3655,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3655,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3655,36,'cp 7','c2c3 c6e7 d5e3 c5e3 f2e3 c7c6 d3d4 d6d5 c4d3 d5e4 d3e4 e5d4 e3d4 c8f5 e4f5 e7f5 d1d3 f5d6 f3e5 f7f6 e5c4 d6c4 d3c4 d8d5 c4d5 c6d5 a1e1 f8e8 e1e8 a8e8 f1f5 e8e2 b2b3 e2e3 f5d5 e3c3 d5d8 g8h7 d8d7 c3c2 d7b7 c2a2');
INSERT INTO evaluations VALUES(5642,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(5642,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(5642,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(5642,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(5642,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(5642,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(5642,36,'cp 23','f4e5 d6e5 e1g1 g8f6 d2d3 c6a5 c4b3 a5b3 a2b3 c8e6 f3g5 d8d6 b1c3 e8g8 d1e1 d6c6 h2h3 f6d7 g5e6 f7e6 f1f8 d7f8 c1e3 a7a5 e1e2 b7b6 a1f1 f8g6 g2g3 a8f8 g1g2 e7c5 c3d1 c5d4 h3h4 f8f1 g2f1 d4e3 d1e3 c6c5');
INSERT INTO evaluations VALUES(10326,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10326,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10326,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10326,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10326,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10326,36,'cp 0','c6e7 f1d3 d7d6 e1g1 c7c6 d5c6 b4c3 c6b7 c8b7 b2c3 f6e4 d3e4 b7e4 f1e1 e4d5 c1a3 a8c8 d1d3 c8c6 f3e5 d6e5 a1d1 f7f6 a3e7 e8e7 d3d5 d8d5 d1d5 h8c8 d5a5 c8c7 f2f4 c6c3 f4e5 c3c2 e5f6 e7f6 e1d1 c2e2 a5a6 f6e5 a6a5 e5e4 a5a4 e4e3 a4a3 e3f4 d1f1 f4g5 a3g3 g5h6 g3h3 h6g6 h3g3');
INSERT INTO evaluations VALUES(13159,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13159,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(13159,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(13159,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(13159,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(13159,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(13159,36,'cp 20','c8g4 c1f4 e7e6 h2h3 g4f3 d1f3 f8d6 b1d2 e8g8 e1g1 f8e8 f4g5 h7h6 g5h4 d6c7 a1d1 d8d6 h4g3 d6d8 f1e1 c7g3 f3g3 f6h5 g3f3 h5f6 d2b3 d8c7 e1e3 e8e7 f3e2 e6e5 d4e5 e7e5 e3e5 c6e5 b3d4 a8e8 d4f5 e5d3 e2d3 e8e6 d3d4 a7a6');
INSERT INTO evaluations VALUES(18725,30,'cp -185',NULL);
INSERT INTO evaluations VALUES(18725,31,'cp -189',NULL);
INSERT INTO evaluations VALUES(18725,32,'cp -176',NULL);
INSERT INTO evaluations VALUES(18725,33,'cp -181',NULL);
INSERT INTO evaluations VALUES(18725,34,'cp -176',NULL);
INSERT INTO evaluations VALUES(18725,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(18725,36,'cp -189','g8f6 h5d1 d7d5 e4d5 e6d5 c4b3 b7b5 c2c3 f8d6 g1f3 e8g8 e1g1 h7h6 h2h3 c8f5 d3d4 c5c4 b3c2 f5c2 d1c2 b5b4 f1e1 a7a5 b1d2 a5a4 d2f1 f8e8 c1d2 f6e4 f1e3 b4b3 c2b1 a8a5 a2b3 c4b3 b1d1 d8c7 e3f5 d6f8 f5e3');
INSERT INTO evaluations VALUES(12347,30,'cp -187',NULL);
INSERT INTO evaluations VALUES(12347,31,'cp -191',NULL);
INSERT INTO evaluations VALUES(12347,32,'cp -161',NULL);
INSERT INTO evaluations VALUES(12347,33,'cp -184',NULL);
INSERT INTO evaluations VALUES(12347,34,'cp -181',NULL);
INSERT INTO evaluations VALUES(12347,35,'cp -169',NULL);
INSERT INTO evaluations VALUES(12347,36,'cp -176','h5d1 d7d5 e4d5 e6d5 c4b3 b7b5 c2c3 f8d6 g1f3 h7h6 h2h3 e8g8 e1g1 c8f5 f1e1 d8d7 d3d4 c5c4 b3c2 f5c2 d1c2 a8e8 e1e8 f8e8 b1d2 e8e2 c2d1 e2e6 a2a3 d6c7 f3h4 d7e8 h4f3 c6e7 d2f1 a7a5 c1d2 e7f5 a3a4 f6e4 f3e5 c7e5 d4e5 e6e5 g2g4 e4d2 d1d2');
INSERT INTO evaluations VALUES(12371,30,'cp -182',NULL);
INSERT INTO evaluations VALUES(12371,31,'cp -179',NULL);
INSERT INTO evaluations VALUES(12371,32,'cp -188',NULL);
INSERT INTO evaluations VALUES(12371,33,'cp -182',NULL);
INSERT INTO evaluations VALUES(12371,34,'cp -187',NULL);
INSERT INTO evaluations VALUES(12371,35,'cp -192',NULL);
INSERT INTO evaluations VALUES(12371,36,'cp -177','d7d5 e4d5 e6d5 c4b3 b7b5 c2c3 f8d6 g1f3 e8g8 e1g1 h7h6 h2h3 c8f5 d3d4 c5c4 b3c2 f5c2 d1c2 f8e8 f1e1 a7a5 e1e8 d8e8 c1e3 a5a4 c2c1 b5b4 e3d2 e8e2 c3b4 f6e4 c1f1 e2f1 g1f1 e4d2 f3d2 d6b4 d2f3 a8b8 a2a3 b4a5 a1a2 f7f6 b1c3 a5c3 b2c3');
INSERT INTO evaluations VALUES(19960,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19960,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19960,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19960,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19960,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(19960,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(19960,36,'cp -2','g1f3 e7e6 b1c3 a7a6 a2a3 f8e7 c4a2 g8f6 d2d3 e8g8 c1d2 b7b5 e1g1 d6d5 e4e5 f6d7 c3e2 a6a5 e2g3 g8h8 c2c3 b5b4 a2b1 c8a6 f1e1 g7g6 a3b4 a5b4 d2e3 b4c3 b2c3 d5d4 c3d4 c6d4 f3d4 c5d4');
INSERT INTO evaluations VALUES(26539,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26539,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(26539,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(26539,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(26539,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(26539,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26539,36,'cp 24','g8f6 c2c4 e7e6 g1f3 f8d6 f4g3 e8g8 f1e2 d5c4 e2c4 d8e7 c4b5 d6g3 h2g3 c6b4 e1g1 a7a6 b5e2 c7c5 d4c5 e7c5 f3d4 f8d8 a2a3 b4c2 a1c1 a8c8 e2f3 d8d7 g3g4 f5d3 g4g5 c5g5 c1c2 c8c2 d4c2 d3f1 g1f1 g5b5 f3e2');
INSERT INTO evaluations VALUES(28298,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(28298,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28298,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(28298,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28298,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(28298,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(28298,36,'cp 5','e2e3 e8g8 f1e2 h7h6 g5f4 f6h5 e1g1 e7b4 a1c1 d7d6 f4g3 b4c3 c1c3 h5g3 h2g3 b8d7 e2d3 c7c5 d1e2 a8c8 e3e4 c5d4 f3d4 a7a6 f1d1 c8c7 d4f3 d8e7 e2e3 f8d8 d3e2 d7c5 e4e5 f7f5 d1d6 d8d6 e5d6 e7d6');
INSERT INTO evaluations VALUES(28297,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(28297,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(28297,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(28297,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28297,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(28297,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(28297,36,'cp 0','e8g8 f1e2 h7h6 g5f4 f6h5 e1g1 e7b4 d1b3 h5f4 e3f4 b4c3 b3c3 d8f6 g2g3 d7d6 a1d1 b8c6 d1d2 c6e7 f3h4 a7a5 f1d1 f8e8 e2f3 b7f3 h4f3 a5a4 d1e1 d6d5 c4c5 b6c5 c3c5 e8c8 g1g2 e7f5 e1e2 f6e7');
INSERT INTO evaluations VALUES(27783,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(27783,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(27783,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(27783,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(27783,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(27783,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(27783,36,'cp 5','a1c1 h7h6 g5f4 f6h5 f4g3 d7d6 f1e2 e7f6 e1g1 c7c5 g3h4 g7g5 h4g3 d8e7 e2d3 h5g3 h2g3 f6g7 d3e4 b7e4 c3e4 f7f5 e4c3 b8d7 d1e2 a8d8 f1e1 a7a6 e1d1 e7f7 a2a3 f7g6 b2b4 d7f6 c1b1 f6d7 f3d2');
INSERT INTO evaluations VALUES(34299,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(34299,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(34299,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34299,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(34299,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(34299,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(34299,36,'cp 76','d5a5 c1d2 g8f6 f1c4 c8f5 g1f3 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 d7b6 c4b3 a7a5 a2a3 f8e7 c3e4 f6e4 e2e4 e7f6 c2c3 a8d8 e4g4 b6d5 h4h5 h8h5 h1h5 g6h5 g4h5 g7g6 h5e2 b7b5 g2g4 a5a4 b3c2 e8f8 c1b1');
INSERT INTO evaluations VALUES(34525,30,'cp 762',NULL);
INSERT INTO evaluations VALUES(34525,31,'cp 771',NULL);
INSERT INTO evaluations VALUES(34525,32,'cp 762',NULL);
INSERT INTO evaluations VALUES(34525,33,'cp 767',NULL);
INSERT INTO evaluations VALUES(34525,34,'cp 771',NULL);
INSERT INTO evaluations VALUES(34525,35,'cp 763',NULL);
INSERT INTO evaluations VALUES(34525,36,'cp 763','e2b5 c8d7 b5c6 b8c6 g1e2 e7e5 e1g1 e8c8 a2a3 g7g6 b2b4 f8g7 d2d3 c6d4 c1d2 d4e2 d1e2 g8e7 a3a4 e7f5 b4b5 h7h5 e2d1 c8b8 a4a5 h5h4 h2h3 g7h6 f1e1 f7f6 c3e4 h6d2 d1d2 d7b5 e4f6');
INSERT INTO evaluations VALUES(39149,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(39149,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(39149,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(39149,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(39149,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(39149,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(39149,36,'cp 42','e2e4 d7d5 e4e5 g8e7 a2a3 b4c3 b2c3 c7c5 h2h4 c8d7 h4h5 b8c6 h5h6 g7h6 a1b1 d8a5 c1d2 e8c8 g1f3 d8g8 g2g3 a5a3 b1a1 a3b2 d4c5 e7f5 a1a4 c6e5 a4a7 e5f3 d1f3 b2b1 f3d1 b1d1 e1d1 c8c7 d2f4 c7c6 c3c4 c6c5 a7b7');
INSERT INTO evaluations VALUES(39156,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(39156,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(39156,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(39156,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(39156,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(39156,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(39156,36,'cp 36','d7d5 e4e5 g8e7 a2a3 b4c3 b2c3 c7c5 h2h4 b8c6 h4h5 d8a5 c1d2 a5c7 g1f3 h7h6 f1d3 c8d7 a3a4 c6a5 e1g1 a5c4 d2c1 c4b6 a4a5 b6a4 d4c5 a4c3 d1d2 c3e4 d2f4 e8g8 c1a3 e7c6 a5a6 b7a6 d3e4 d5e4 f4e4 c6e7 e4g4 d7c6 a3b4');
INSERT INTO evaluations VALUES(39762,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(39762,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(39762,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(39762,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(39762,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(39762,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(39762,36,'cp 86','d1d2 h7h5 h2h3 b8c6 e1c1 c8d7 c3d5 a8c8 d4c6 c8c6 c1b1 f6d5 e4d5 c6c8 f1d3 h5h4 d2f2 d8a5 e3d4 e8f8 d4g7 f8g7 h1e1 a5d5 d3g6 d5g5 g6e4 c8c7 d1d5 f7f5 f3f4 g5g3 f2d2 d7e6 e4f3 e6d5 f3d5');
INSERT INTO evaluations VALUES(41994,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(41994,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41994,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41994,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41994,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(41994,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41994,36,'cp 12','b1c3 h7h6 g5f6 d8f6 d1a4 b8c6 e2e3 e8g8 a1c1 f8d8 c4d5 e6d5 f1d3 c8g4 d3e2 a7a6 e1g1 b4f8 a4d1 c6a7 e2d3 c7c6 h2h3 g4e6 c3a4 f8d6 a4c5 d6c5 c1c5 a7b5 a2a4 b5d6 b2b4 d8e8 f3e5 g7g6 b4b5 c6b5 a4b5 a6b5 d3b5');
INSERT INTO evaluations VALUES(48804,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(48804,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48804,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(48804,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(48804,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48804,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48804,36,'cp 0','f6e4 c3e4 d5e4 f3e5 f8d6 e5c4 e8g8 f1e2 b8c6 c2c3 c6e7 c4d6 d8d6 e1g1 c8f5 f1e1 c7c6 e2f1 f8e8 a2a4 a7a5 d1b3 e7d5 c1d2 h7h6 c3c4 d5b4 d4d5 f5d7 d2c3 c6d5 a1d1 b4d3 f1d3 e4d3 e1e8 a8e8 d1d3');
INSERT INTO evaluations VALUES(49896,30,'cp -176',NULL);
INSERT INTO evaluations VALUES(49896,31,'cp -163',NULL);
INSERT INTO evaluations VALUES(49896,32,'cp -161',NULL);
INSERT INTO evaluations VALUES(49896,33,'cp -189',NULL);
INSERT INTO evaluations VALUES(49896,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(49896,35,'cp -169',NULL);
INSERT INTO evaluations VALUES(49896,36,'cp -159','d7d5 c4d3 e5d4 f3d4 e4c5 e1g1 f8e7 f1e1 e8g8 c1f4 f8e8 c2c3 b8d7 d3c2 d7f6 h2h3 c5e6 d4e6 c8e6 b1d2 e7d6 d1f3 h7h6 e1e2 d6f4 f3f4 f6h5 f4e5 g7g6 e5d4 d8f6 d2f3 f6d4 f3d4 e6d7 a1e1 e8e2');
INSERT INTO evaluations VALUES(49073,30,'cp -151',NULL);
INSERT INTO evaluations VALUES(49073,31,'cp -151',NULL);
INSERT INTO evaluations VALUES(49073,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(49073,33,'cp -147',NULL);
INSERT INTO evaluations VALUES(49073,34,'cp -140',NULL);
INSERT INTO evaluations VALUES(49073,35,'cp -160',NULL);
INSERT INTO evaluations VALUES(49073,36,'cp -165','c4d3 e5d4 f3d4 e4c5 e1g1 f8e7 c1f4 e8g8 f1e1 c5d3 d1d3 b8a6 b1d2 c7c5 d4b5 c8e6 d2f3 d5d4 c2c3 d4c3 d3c3 d8b6 a2a4 a6b4 a1d1 e7f6 c3c1 a8c8 f4g5 f6g5 f3g5 e6d5 c1f4 c8d8 f4f5 b6g6 f5g6 h7g6 b5a7 b4c2 e1e7 f7f6');
INSERT INTO evaluations VALUES(53966,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(53966,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(53966,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(53966,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(53966,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(53966,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(53966,36,'cp -101','g8f6 h5h4 f8e7 c4b5 c8d7 b5d7 d8d7 d2d3 b8c6 h4g3 e8c8 b1c3 h8g8 c1d2 d6d5 e4e5 f6e8 h2h4 e8c7 c3e2 d5d4 a2a3 c8b8 h4h5 a7a6 e1f1 f7f6 h1h3 d8f8 a1e1 c7d5 e2f4 d5f4 d2f4 f6f5 f3g5 e7g5 f4g5');
INSERT INTO evaluations VALUES(45981,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(45981,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(45981,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(45981,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(45981,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(45981,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(45981,36,'cp -101','h5h4 f8e7 c4b5 c8d7 b5d7 d8d7 d2d3 b8c6 h4g3 e8c8 b1a3 h7h6 c1e3 h8g8 e1c1 g7g5 h2h3 f6h5 g3h2 f7f5 e4f5 e6f5 e3d2 e7f6 c1b1 f5f4 g2g3 g8f8 h2g2 f6g7 g3g4 h5f6 d2c3 f6d5 c3g7 d7g7');
INSERT INTO evaluations VALUES(45982,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(45982,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(45982,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(45982,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(45982,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(45982,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(45982,36,'cp -98','f8e7 c4b5 c8d7 b5d7 d8d7 d2d3 b8c6 h4g3 e8c8 c1d2 d6d5 e4e5 f6e8 b1c3 h8g8 h2h4 e8c7 c3e2 d5d4 a2a3 c8b8 h4h5 c7d5 e2f4 f7f5 e1f1 a7a6 h1h3 d8c8 f3g1 b8a7 g1f3 d5f4 g3f4 g7g5 h5g6 h7g6');
INSERT INTO evaluations VALUES(40173,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(40173,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(40173,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(40173,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(40173,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(40173,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(40173,36,'cp -94','c4b5 c8d7 b5d7 d8d7 d2d3 b8c6 h4g3 e8c8 b1a3 h7h6 c1e3 g7g5 e1c1 h8g8 h2h3 f6h5 g3h2 f7f5 c1b1 f5f4 e3c1 d6d5 h1e1 a7a6 c2c3 h5f6 e4d5 e6d5 a3c2 e7d6 h2g1 c8b8 g1f1 d7f7 a2a3 g8f8 b1a1 b8a8 f1e2');
INSERT INTO evaluations VALUES(54062,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(54062,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(54062,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54062,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(54062,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(54062,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54062,36,'cp 12','f8c5 b1c3 g8f6 f1b5 b8c6 c1g5 a7a6 b5a4 c5e7 e4d5 f6d5 c3d5 e6d5 g5e7 d8e7 d1e2 c8g4 a4c6 b7c6 e2e7 e8e7 f3e5 g4d7 e1c1 f7f6 e5d7 e7d7 c2c4 d7d6 d1d4 d6c5 h1d1 d5c4 c1c2 h8e8 c2c3 e8e2 d4c4 c5b6 d1d2 a8e8 g2g3 e2d2 c3d2 e8e5 c4f4 b6b5 d2d3 e5d5 d3c3');
INSERT INTO evaluations VALUES(38122,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(38122,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38122,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(38122,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(38122,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38122,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(38122,36,'cp 21','b1c3 g8f6 e4d5 e6d5 c1g5 e8g8 g5f6 d8f6 d1d5 c5b4 d5d4 f6e6 d4e3 e6e3 f2e3 b4c3 b2c3 b8d7 f1c4 d7b6 c4d3 c8d7 e1d2 g7g6 h1b1 d7c6 b1b4 b6d5 b4d4 a8e8 a1e1 g8g7 a2a4 d5f6 a4a5 f6e4 d3e4 e8e4 c3c4 f8c8 d4e4 c6e4');
INSERT INTO evaluations VALUES(5431,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(5431,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(5431,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(5431,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(5431,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(5431,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(5431,36,'cp -21','g8f6 d2d4 e8g8 c1e3 c5c4 b5c6 b7c6 f3e5 d8c7 b2b3 e7d6 b1c3 f8e8 e3f4 c8f5 b3c4 d5c4 f4g3 f5e6 h2h3 a8b8 d1f3 e6d5 c3d5 c6d5 e5g4 f6g4 g3d6 c7d6 h3g4 h7h6 a2a4 e8e6 g2g3 b8e8 e1e6 e8e6 c2c3 d6c6 a4a5');
INSERT INTO evaluations VALUES(8113,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(8113,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(8113,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(8113,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(8113,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(8113,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(8113,36,'cp -9','d2d4 e8g8 c1e3 c5c4 b5c6 b7c6 f3e5 d8c7 b2b3 f8e8 h2h3 c8e6 e3f4 e7d6 d1f3 c4b3 a2b3 c7b6 f3d3 f6e4 b1d2 f7f6 e5c4 d5c4 d3e4 e6f7 d2c4 f7c4 e4e8 a8e8 e1e8 g8f7 f4d6 f7e8 d6c5 b6b5 b3c4 b5c4 a1a7 c4c2 a7g7 c2c1 g1h2 c1d2 h2g3 h7h5 g7a7 d2c3 g3h2');
INSERT INTO evaluations VALUES(10059,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(10059,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(10059,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(10059,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(10059,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(10059,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(10059,36,'cp -30','h7h6 g5h4 a7a6 f3e5 c6e5 d3d4 c5e7 d4e5 f6e4 f1d3 e7h4 d3e4 e8g8 e1g1 f8e8 f1e1 d7d6 e5d6 e8e6 g2g3 e6d6 d1e2 h4f6 b1a3 a8b8 a3c4 d6e6 e2f1 g7g6 e4f3 e6e1 a1e1 c8e6 f1e2 h6h5 h2h4 f6g7 a2a4 c7c6 c4d2 a6a5 d2e4');
INSERT INTO evaluations VALUES(9737,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(9737,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(9737,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(9737,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(9737,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(9737,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9737,36,'cp -37','g5h4 a7a6 f3e5 c6e5 d3d4 c5e7 h4f6 e7f6 d4e5 f6e5 g2g3 d7d6 b1a3 d8f6 a3c4 c8e6 d1d3 g7g5 c4e5 f6e5 d3d4 f7f6 h2h4 e8c8 d4e5 f6e5 a2a4 g5g4 f1e2 h8f8 a4a5 h6h5 a1d1 e6d7 e2d3 d7c6 e1g1 d6d5 d1e1 d5d4 c3d4 d8d4');
INSERT INTO evaluations VALUES(12791,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(12791,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(12791,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(12791,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(12791,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(12791,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(12791,36,'cp -42','e7e6 b1c3 d7d5 e4d5 e6d5 c4a2 f8e7 g1e2 e8g8 e1g1 c8e6 f1e1 e7d6 e2f4 d8d7 f4d5 f6d5 a2d5 e6d5 c3d5 d6h2 g1h2 d7d5 h2g1 a8d8 c1f4 f7f6 f4d2 g8f7 d2c3 d5f5 d1e2 f8e8 e2d2 a7a6 e1e8 d8e8 a1e1 e8e1 d2e1 h7h5 e1d2');
INSERT INTO evaluations VALUES(15972,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(15972,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(15972,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(15972,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(15972,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(15972,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(15972,36,'cp 41','f6e4 d3e4 c7c5 c1e3 d8c7 c2c3 f8d6 d4c5 d6c5 d1d2 e8g8 e1c1 c5e3 d2e3 d7f6 e4c2 b7b6 h1e1 c8b7 f3e5 a8d8 g2g3 b6b5 h2h4 a7a6 a2a3 f6d5 e3d3 d5f6 d3e2 h7h5 f2f3 d8d5 d1d5 b7d5 g3g4 h5g4 f3g4 f8d8');
INSERT INTO evaluations VALUES(15931,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(15931,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(15931,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(15931,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(15931,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(15931,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(15931,36,'cp 30','d3e4 c7c5 c1e3 d8c7 c2c3 d7f6 e4c2 f8e7 d1e2 b7b6 d4c5 e7c5 e3c5 c7c5 e1c1 e8g8 h2h4 c8b7 f3e5 b7a6 e2a6 c5e5 c1b1 g7g6 d1d4 f8d8 d4d8 a8d8 a6a7 e5c5 f2f3 c5f2 a7e7 d8d2 e7f6 d2c2 f6d8 g8g7');
INSERT INTO evaluations VALUES(15930,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(15930,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(15930,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(15930,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(15930,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(15930,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(15930,36,'cp 46','c7c5 c1e3 f8d6 c2c3 d8c7 d4c5 d6c5 d1d2 d7f6 e4c2 c5e3 d2e3 e8g8 e1c1 b7b5 d1d4 c8b7 h1d1 a7a6 f3e5 f8d8 g2g4 f6d5 e3e4 g7g6 e4e2 d5b6 h2h4 d8d5 f2f4 a8d8 d4d5 d8d5 g4g5 d5d1 e2d1 b7d5 d1d4 d5a2');
INSERT INTO evaluations VALUES(32760,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(32760,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(32760,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(32760,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(32760,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(32760,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(32760,36,'cp 5','c5d4 d1d4 g8f6 d4d3 g7g6 e1g1 b8a6 b1c3 f8g7 d3e2 a8c8 e4e5 d6e5 f3e5 d7f5 e2b5 f6d7 b5b7 d7e5 b7a6 e8g8 a6a7 f5c2 a7e7 f8e8 e7b4 e5d3 b4b3 c2b3 a2b3 d3b2 c1b2 g7c3 b2c3 c8c3 a1e1 e8e1 f1e1 c3b3 h2h4 b3b5');
INSERT INTO evaluations VALUES(38052,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(38052,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(38052,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(38052,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(38052,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(38052,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(38052,36,'cp 63','c4d5 e6d5 e4e5 c6c5 a2a3 b4a5 b2b4 c5d4 c3b5 a5c7 f2f4 g8e7 g1f3 b8c6 f1d3 a7a6 b5c7 d8c7 c1b2 c8f5 f3d4 c6d4 b2d4 e8g8 d4c5 f8e8 e1g1 c7d7 c5e7 f5d3 d1d3 e8e7 a1d1 a8d8 d3d4 d7e6 f1e1 e6g4 h2h3 g4d7 g1h2 h7h6 d1d2 d7e6 e1d1');
INSERT INTO evaluations VALUES(38036,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(38036,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(38036,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38036,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(38036,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(38036,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(38036,36,'cp 55','e6d5 e4e5 g8e7 g1f3 c8g4 a2a3 b4a5 f1e2 e8g8 e1g1 c6c5 d4c5 a5c3 b2c3 b8d7 d1d4 g4e6 a3a4 d8c7 c1a3 f8c8 f3g5 d7c5 a3c5 c7c5 g5e6 f7e6 e2g4 c5c6 h2h4 c8c7 a4a5 h7h6 f1b1 e7f5 g4f5 e6f5 h4h5');
INSERT INTO evaluations VALUES(40980,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(40980,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(40980,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40980,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(40980,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(40980,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(40980,36,'cp 12','c7c5 b1d2 e8g8 f3e5 f6d7 f1d3 d7e5 f4e5 b8d7 e5f4 d8b6 d1c2 d7f6 d4c5 e7c5 e1g1 h7h6 c3c4 c8d7 a2a3 a7a5 f4e5 f8c8 h2h3 c5f8 f1c1 d5c4 d2c4 b6d8 e5c3 b7b5 c4e5 d7e8');
INSERT INTO evaluations VALUES(40221,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40221,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40221,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40221,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(40221,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(40221,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40221,36,'cp 25','b1d2 e8g8 f3e5 f6d7 f1d3 d7e5 f4e5 b8d7 e5f4 d8b6 d1c2 d7f6 f4e5 h7h6 d4c5 b6c5 e5d4 c5c7 f2f4 e7c5 e1g1 c8d7 d2b3 c5d4 e3d4 a7a5 a2a4 f8c8 b3d2 f6h5 g2g3 h5f6 d2f3 d7c6 f3e5 c6e8 c2e2');
INSERT INTO evaluations VALUES(40220,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40220,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(40220,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40220,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(40220,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(40220,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(40220,36,'cp 10','e8g8 f3e5 c5c4 f1e2 b7b5 a2a4 b5b4 e2f3 c8a6 e1g1 d8c8 a1c1 b8d7 e3e4 b4c3 b2c3 a6b7 e5d7 f6d7 e4d5 b7d5 f3d5 e6d5 f1e1 f8e8 c1b1 c8c6 d2f3 d7f6 f3e5 c6c7 d1f3 e7d6 g2g3 f6e4 b1b5 f7f6 e5g4 d6f4 b5d5 e4d2 e1e8 a8e8 f3f4 c7f4 g3f4');
INSERT INTO evaluations VALUES(35834,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(35834,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(35834,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35834,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(35834,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(35834,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(35834,36,'cp 28','f3e5 c5c4 f1e2 b7b5 a2a4 b5b4 e2f3 c8a6 e1g1 d8c8 a1c1 b8d7 e3e4 b4c3 b2c3 a6b7 e5d7 f6d7 a4a5 d5e4 f3e4 b7e4 d2e4 c8a6 c1a1 d7f6 e4f6 e7f6 f1e1 a8d8 f4e5 d8d5 e5f6 g7f6 d1g4 d5g5 g4f3 g5a5 f3g4 g8h8 a1b1 e6e5 d4e5 a5e5 e1e5 f6e5');
INSERT INTO evaluations VALUES(44753,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(44753,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(44753,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(44753,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(44753,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(44753,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(44753,36,'cp 45','c5d4 f3d4 f8g7 f2f3 d8b6 c4b5 c8d7 c1e3 b6a5 b5e2 e8g8 d1d2 f8c8 d4b3 a5c7 g2g4 b7b5 c3b5 c7c2 d2c2 c8c2 b5d4 c2c8 e1f2 a7a5 h1c1 b8c6 d4c6 d7c6 c1c2 a5a4 b3d4 c6e8 a1c1 c8c2 c1c2 a8b8 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(44785,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(44785,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(44785,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(44785,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(44785,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(44785,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(44785,36,'cp 28','f3d4 f8g7 f2f3 d8b6 c4b5 c8d7 c1e3 b6c7 d1d2 a7a6 b5d3 b8c6 e1c1 h7h5 h2h3 b7b5 d4c6 d7c6 c1b1 a8b8 c3e2 a6a5 e2d4 b5b4 h1e1 a5a4 c2c3 b4b3 a2a3 f6d7 d3e2 d7c5 d4c6 c7c6 d2d5 c6c7 f3f4 e7e6 d5d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(53452,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(53452,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53452,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(53452,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(53452,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53452,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(53452,36,'cp 35','e3d4 e7e6 g1f3 f8d6 e1g1 g8e7 f1e1 e8g8 b1c3 h7h6 d1d3 b8a6 a2a3 a6c7 c4b3 b7b6 d3e4 c8a6 e4g4 f7f5 g4h5 d8e8 h5e8 f8e8 b3e6 c7e6 e1e6 a8d8 c1d2 a6c4 e6e1 g8f7 h2h4 e7c6 e1e8 d8e8');
INSERT INTO evaluations VALUES(56921,30,'cp -236',NULL);
INSERT INTO evaluations VALUES(56921,31,'cp -237',NULL);
INSERT INTO evaluations VALUES(56921,32,'cp -232',NULL);
INSERT INTO evaluations VALUES(56921,33,'cp -215',NULL);
INSERT INTO evaluations VALUES(56921,34,'cp -216',NULL);
INSERT INTO evaluations VALUES(56921,35,'cp -185',NULL);
INSERT INTO evaluations VALUES(56921,36,'cp -177','d8h4 e1f1 b8c6 d2d4 c6e5 d4e5 f4f3 c1f4 b7b5 c4b5 f8c5 d1d2 g8e7 b1c3 c7c6 c3a4 f3g2 d2g2 c5d4 b5e2 e7g6 f4g3 h4g5 h2h4 g5e3 g3f2 e3f4 h4h5 g6e5 a1d1 d4e3 g2g3 f4e4 h1g1 e3f2 g3f2 d7d5 d1e1 e4c2 e2g4 c2d3 e1e2 c8g4 g1g4 e8c8 g4g3 d3d1 f1g2 d1a4 e2e5 h8e8 f2f5 c8b8 e5e8 d8e8');
INSERT INTO evaluations VALUES(35014,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(35014,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(35014,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(35014,33,'cp -189',NULL);
INSERT INTO evaluations VALUES(35014,34,'cp -178',NULL);
INSERT INTO evaluations VALUES(35014,35,'cp -191',NULL);
INSERT INTO evaluations VALUES(35014,36,'cp -185','e1f1 b8c6 d2d4 c6e5 d4e5 f4f3 c1f4 b7b5 c4b5 f8c5 d1d2 g8e7 b1c3 c7c6 c3a4 f3g2 d2g2 c5d4 b5e2 e7g6 f4g3 h4g5 h2h4 g5e3 g3f2 e3f4 h4h5 g6e5 a1d1 d4f2 g2f2 f4f2 f1f2 e8e7 f2g3 h8g8 c2c4 a8b8 b2b3 g8g5 c4c5 b8b4 a4c3 c8b7 d1d2 e5f3 e2f3 g4f3 g3f3');
INSERT INTO evaluations VALUES(35013,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(35013,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(35013,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(35013,33,'cp -184',NULL);
INSERT INTO evaluations VALUES(35013,34,'cp -222',NULL);
INSERT INTO evaluations VALUES(35013,35,'cp -229',NULL);
INSERT INTO evaluations VALUES(35013,36,'cp -200','b8c6 d2d4 c6e5 d4e5 f4f3 c1f4 b7b5 c4b5 f8c5 d1d2 g8e7 b1c3 c7c6 c3a4 f3g2 d2g2 c5d4 b5e2 e7g6 f4g3 h4e7 h2h3 h7h5 h3g4 h5h4 a1d1 d4e5 g3e5 e7e5 b2b3 d7d6 g2h2 e8e7 h2e5 g6e5 g4g5 h4h3 f1f2 c8d7 a4c3 a8g8 d1g1 d7e6 g1g3 h8h4 a2a3');
INSERT INTO evaluations VALUES(2295,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(2295,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(2295,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(2295,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(2295,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(2295,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(2295,36,'cp 4','c1g5 c7c6 f1c4 d8e8 a2a4 a7a5 h2h4 f8c5 f2f3 h7h5 c3d1 f6d7 e1e2 f7f6 g5c1 d7b6 c4b3 b6d7 d1f2 c5d4 c2c3 d4a7 h1d1 e8e7 g2g3 g7g6 f2d3 d7c5 d3c5 a7c5 c1e3 c5e3 e2e3 c8e6 b3e6 e7e6 a1a3 b7b6 a3b3 h8b8 d1d3 g6g5 h4g5');
INSERT INTO evaluations VALUES(2290,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(2290,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(2290,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(2290,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(2290,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(2290,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(2290,36,'cp 2','c7c6 f1c4 d8e8 a2a4 a7a5 h2h4 f8c5 f2f3 h7h5 e1e2 f6d7 c3d1 f7f6 g5d2 d7b6 c4b3 e8e7 d1f2 b6d7 f2d3 c5d4 b3a2 h8d8 a2c4 d7c5 d3c5 d4c5 h1d1 c5d4 c2c3 d4a7 b2b4 a5b4 c3b4 d8d4 c4b3 c8e6 b3e6');
INSERT INTO evaluations VALUES(4934,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(4934,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(4934,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(4934,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(4934,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(4934,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(4934,36,'cp 90','d2d4 e5d4 c2c3 c6a5 d1d4 a5b3 a2b3 f7f5 e4f5 g8f6 e1g1 c8b7 f1e1 e8g8 b1d2 f6d5 d4g4 e7d6 b3b4 d5f6 g4h3 f8e8 d2f1 e8e1 f3e1 f6d5 c1e3 d8e7 h3g4 d5f6 g4g5 e7f7 f2f3 f6d5 e1d3 d5e7 f5f6 f7f6 g5f6 g7f6 e3c5 e7f5 g1f2');
INSERT INTO evaluations VALUES(20888,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(20888,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20888,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20888,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20888,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20888,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(20888,36,'cp 17','c8g4 f1b5 e7e6 b1d2 g8h6 h2h3 g4h5 d2b3 a7a6 b5c6 b7c6 c1e3 h6f5 d1e2 h5f3 e2f3 f8e7 f3h5 g7g6 h5g4 a8b8 e1g1 e8g8 c2c3 d8c8 g4e2 b8b5 b3d4 f5d4 c3d4 c8b7 b2b3 c6c5 d4c5 e7c5 e3c5 b5c5 e2d2');
INSERT INTO evaluations VALUES(26480,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(26480,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(26480,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26480,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(26480,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(26480,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26480,36,'cp 11','f1b5 e7e6 b1d2 g8h6 h2h3 g4h5 d2b3 a7a6 b5c6 b7c6 c1e3 h6f5 d1e2 d8d7 e1g1 f5e3 e2e3 f8e7 f3d2 h5g6 c2c3 e8g8 b3c5 e7c5 e3c5 g6d3 f1c1 d7f7 c5e3 f7g6 c3c4 d5c4 d2c4 d3e4 e3g3 g6g3 f2g3 h7h5 b2b3 f8f5 c1e1 e4d5 c4e3 f5e5');
INSERT INTO evaluations VALUES(21709,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(21709,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(21709,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(21709,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(21709,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(21709,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(21709,36,'cp -43','g8f6 e4d5 f6d5 f1e2 f8b4 c1d2 d5c3 d2c3 b4c3 b2c3 f7f5 e1g1 e8g8 f1e1 g8h8 f3d2 d8f6 e2f3 c8d7 d2c4 a8e8 a1b1 b7b6 d1e2 e5e4 d3e4 f6c3 b1b3 c3f6 e4e5 f6e7 e2e3 c6d8 e3c3 d8e6 c4e3 e7g5 e3d5 c7c6 d5b4 f5f4 b4d3 c6c5');
INSERT INTO evaluations VALUES(29803,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(29803,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(29803,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(29803,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(29803,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(29803,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(29803,36,'cp 85','f1d3 f5d3 d1d3 e7e6 c2c4 d5f6 b1c3 c7c6 e1g1 f8e7 d3e2 e8g8 c1f4 b8d7 f1d1 f6h5 f4e5 f8e8 h2h4 h5f6 e5g3 d8a5 f3e5 a8d8 a2a3 d7e5 d4e5 f6d7 b2b4 a5c7 c4c5 b7b6 c5b6 a7b6 a1c1 h7h6 h4h5 c7a7 e2b2 f7f5 g3f4 d7f8');
INSERT INTO evaluations VALUES(29799,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(29799,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(29799,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(29799,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(29799,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(29799,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(29799,36,'cp 80','f5d3 d1d3 e7e6 c2c4 d5f6 b1c3 b8d7 e1g1 f8e7 c1f4 e8g8 f1e1 c7c6 h2h3 f8e8 a2a3 d7f8 f4e5 f8g6 a1d1 a7a6 e5g3 b7b5 c4c5 d8d7 f3e5 g6e5 g3e5 f6d5 c3e4 f7f6 e5h2 a6a5 d3e2 a5a4 h2d6 e7d6 e4d6 e8e7 e2g4 g8h8 g4f3 d7c7');
INSERT INTO evaluations VALUES(29817,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(29817,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(29817,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(29817,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(29817,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(29817,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(29817,36,'cp 80','d1d3 e7e6 e1g1 c7c6 c2c4 d5f6 b1c3 f8e7 c1f4 b8d7 f1d1 e8g8 d3e2 f8e8 h2h3 d8a5 a2a3 a5h5 f4h2 h7h6 b2b4 d7f8 e2f1 a8d8 f3e5 h5f5 d1d3 f5h7 f1e2 f8g6 a1e1 a7a6 f2f4 e7d6 g1h1 d6c7 g2g4 c7e5 f4e5');
INSERT INTO evaluations VALUES(29816,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(29816,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(29816,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(29816,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(29816,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(29816,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(29816,36,'cp 78','e7e6 e1g1 c7c6 c2c4 d5f6 b1c3 f8e7 d3e2 e8g8 c1f4 b8d7 f1d1 f8e8 h2h4 d8a5 f3e5 a8d8 a2a3 h7h6 e2f3 d7e5 d4e5 f6h7 c3e4 d8d1 a1d1 e8d8 e4d6 e7d6 e5d6 a5f5 f3g3 f5h5 d1d3 h7f8 f4e3 a7a5 e3d4 f8g6 d4c3 h5f5 d6d7 e6e5');
INSERT INTO evaluations VALUES(31487,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(31487,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(31487,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(31487,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(31487,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(31487,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(31487,36,'cp -14','c6c5 c3e2 c5d4 e2d4 f8c5 g1f3 b8c6 c2c3 g8e7 e1g1 c5d4 c3d4 a7a5 c1d2 h7h6 f1c1 d8b6 a2a4 e7f5 d2c3 e8g8 h2h3 f5e7 d3b5 b6a7 b5d3 f8c8 c3d2 a7a6 d3b5 e7f5 d2c3 a6a7 b5d3 c6b4 d3d2 a7b6 c3b4 b6b4 d2b4 a5b4 c1c8 a8c8');
INSERT INTO evaluations VALUES(31646,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(31646,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(31646,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(31646,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(31646,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(31646,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(31646,36,'cp -12','g1f3 c5d4 c3b5 f8c5 b5d4 g8e7 c2c3 h7h6 c1e3 b8d7 e1g1 e8g8 a1d1 c5d4 e3d4 a8c8 d3e2 d8c7 f1e1 c7c4 e2d2 d7c5 d4c5 c4c5 a2a3 c5a5 f3d4 e7c6 d2e3 c6d4 d1d4 b7b5 e1d1 c8c4 g2g3 f8c8');
INSERT INTO evaluations VALUES(31645,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(31645,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(31645,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(31645,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(31645,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(31645,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(31645,36,'cp -5','c5d4 c3e2 f8c5 e2d4 g8e7 c2c3 h7h6 c1e3 b8d7 h2h4 a8c8 e1c1 e8g8 c1b1 d8b6 d3e2 f7f6 e5f6 d7f6 f3e5 f6e4 b1a1 b6d6 e5d3 c5d4 e3d4 e7c6 d4e3 d6d7 d3f4 d7f7 h1g1 c6e5 f2f3 e4g3 e2e1 g3f5 e3c1');
INSERT INTO evaluations VALUES(40176,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(40176,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(40176,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(40176,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(40176,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(40176,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(40176,36,'cp -24','a7a6 a2a4 b8c6 f1e1 e8g8 h2h3 b7b6 c4b3 c8d7 d3d4 c5d4 f3d4 d8c7 d4f3 c6a5 b3a2 b6b5 a4b5 a6b5 e4e5 d6e5 f3e5 b5b4 e5d7 c7d7 d1d7 f6d7 c3d5 e7d6 d5e3 d7b6 a2b1 g7g6 e1d1 d6e7 c2c3 a5b3 a1a8 f8a8 c3b4');
INSERT INTO evaluations VALUES(39689,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(39689,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(39689,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(39689,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(39689,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(39689,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(39689,36,'cp -18','a2a4 b8c6 f1e1 e8g8 h2h3 d8c7 c1f4 c8d7 d1e2 f8d8 c3d1 a8b8 c4b3 b7b5 a4b5 a6b5 d1e3 b8a8 c2c3 a8a1 e1a1 c7b7 b3c2 h7h6 c2d1 e7f8 f4g3 d7e8 e2d2 d8a8 a1a8 b7a8 d1e2 a8b7 e4e5 d6e5 f3e5 c6e5');
INSERT INTO evaluations VALUES(39693,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(39693,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(39693,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(39693,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(39693,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(39693,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(39693,36,'cp -25','b8c6 c1f4 e8g8 f1e1 f6g4 c3b1 d8b6 d1c1 g4e5 f3e5 d6e5 f4g3 b6c7 c2c3 g8h8 b1d2 f8d8 c1d1 a8b8 d1f3 e7g5 f3e2 b7b5 a4b5 a6b5 c4b3 g5e7 h2h3 f7f6 b3c2 e7d6 d2f3 c7e7 a1a2 e7d7 c2b3 c8b7 e2e3 d7e7');
INSERT INTO evaluations VALUES(41209,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(41209,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41209,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(41209,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(41209,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41209,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41209,36,'cp 20','b1c3 c8f5 e2e3 f6e4 c3e4 f5e4 f1d3 e4d3 d1d3 h7h6 g5f4 b8d7 f4g3 e7e5 a1d1 e8g8 d4e5 d7e5 f3e5 d6e5 e1g1 d8d3 d1d3 f8d8 f1d1 d8d3 d1d3 g8f8 d3d7 a8c8 f2f3 f8e8 d7d1 a7a6 e3e4 c8d8 d1d8 e8d8 g1f2 f7f6 f2e2 d8c8 g3f2 g7f8');
INSERT INTO evaluations VALUES(45288,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(45288,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(45288,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(45288,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(45288,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(45288,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(45288,36,'cp -21','c8d7 b5e2 b8c6 g1f3 d7g4 f3e5 g4e2 c3e2 e7e6 e5d3 b7b5 c2c3 h7h6 e1g1 a7a5 a2a3 f8d6 f1e1 e8g8 c1f4 d6f4 e2f4 d8d6 f4h5 f6h5 d1h5 b5b4 a3b4 a5b4 g2g3 f8c8 e1b1 b4c3 b2c3 a8a3 b1c1 c8a8 a1a3 a8a3 d3c5 e6e5 h5e2');
INSERT INTO evaluations VALUES(28876,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(28876,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(28876,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(28876,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(28876,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(28876,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(28876,36,'cp -19','b5e2 b8c6 g1f3 d7g4 f3e5 g4e2 c3e2 e7e6 e5d3 b7b5 c2c3 h7h6 c1f4 f8d6 d3c5 d6f4 e2f4 b5b4 d1a4 a8c8 c3b4 e8g8 f4e2 d8b6 a1c1 c6b4 e1g1 b4c6 c1c2 c6e7 f1c1 c8b8 c5a6 b8a8 b2b4 f8c8 g2g3 c8c2 c1c2 a8c8');
INSERT INTO evaluations VALUES(49251,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(49251,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(49251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49251,36,'cp 18','f8c5 f2f4 d7d6 g1f3 a7a6 d1e2 b8c6 f4e5 d6e5 c1e3 c5e3 e2e3 e8g8 e3c5 d8d6 c5d6 c7d6 a2a4 c8e6 e1c1 c6d4 f3d4 e5d4 c3e2 e6c4 d3c4 a8e8 e2d4 f6e4 h1f1 g7g6 h2h3 f7f5 f1e1 g8g7 b2b4 e4c3 d4e6 e8e6 e1e6 c3d1 c1d1');
INSERT INTO evaluations VALUES(50946,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(50946,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(50946,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(50946,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(50946,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(50946,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(50946,36,'cp -10','d5c4 e2e3 b7b5 a2a4 b5b4 c3b1 c8b7 f1c4 c7c5 e1g1 g8f6 b1d2 c5d4 f3d4 b8d7 b2b3 d7c5 c1b2 f8d6 c4e2 d8c7 h2h3 e8g8 a1c1 c7e7 c1c4 f6d5 d1b1 f8d8 e2f3 a8c8 f1c1 g7g6 d2e4 e6e5 e4d6 e7d6 b1c2');
INSERT INTO evaluations VALUES(54249,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54249,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54249,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54249,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54249,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(54249,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(54249,36,'cp 0','e4d5 e6d5 d3d4 b8c6 f1b5 c5d4 d1d4 g8f6 c1g5 f8e7 g5f6 e7f6 d4c5 f6c3 b2c3 d8e7 c5e7 e8e7 e1c1 c8e6 g1f3 a8c8 h1e1 h8d8 f3d4 c6d4 c3d4 e7d6 b5d3 c8c3 c1d2 c3a3 d1a1 d8c8 e1b1 b7b6 b1b3 a3a4 c2c3 e6d7 f2f3 g7g6 b3b4 a4a3 b4b3');
INSERT INTO evaluations VALUES(54143,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54143,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54143,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54143,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54143,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54143,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54143,36,'cp 0','e6d5 d3d4 b8c6 f1b5 c5d4 d1d4 g8f6 c1g5 f8e7 g5f6 e7f6 d4c5 f6c3 b2c3 d8e7 c5e7 e8e7 e1c1 c8e6 g1f3 a8c8 h1e1 e7d6 e1e3 h8d8 d1e1 h7h6 f3d4 e6d7 b5d3 g7g6 e3f3 c6e5 f3f6 d6e7 f6f3');
INSERT INTO evaluations VALUES(54550,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(54550,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54550,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54550,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54550,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54550,36,'cp -9','d3d4 b8c6 f1b5 c5d4 d1d4 g8f6 c1g5 f8e7 g5f6 e7f6 d4c5 f6c3 b2c3 d8e7 c5e7 e8e7 e1c1 c8e6 g1f3 e7d6 h1e1 d6c5 b5e2 b7b5 f3d4 h8b8 d4e6 f7e6 e2g4 e6e5 g4d7 e5e4 f2f3 e4f3 g2f3 b8b7 d7c6 c5c6 e1e5 b7d7 f3f4 a8d8 f4f5 d7f7 h2h4 g7g6 f5g6 h7g6 h4h5 g6h5 e5h5');
INSERT INTO evaluations VALUES(6530,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(6530,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6530,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6530,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(6530,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6530,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6530,36,'cp 11','c5d6 c2c4 g8f6 d2d3 a7a6 a2a3 e8g8 f1e2 b7b5 a4c3 c6d4 c4b5 d4b5 e1g1 f8e8 d1c2 c8b7 c1e3 c7c5 c3b5 a6b5 e3c5 a8c8 b2b4 d6c5 b4c5 d7d6 c2b2 c8c5 f1c1 d8b6 c1c5 b6c5 f3h4 d6d5 h4f3');
INSERT INTO evaluations VALUES(6665,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(6665,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(6665,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(6665,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(6665,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(6665,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(6665,36,'cp 158','c2c3 b4e7 c1f4 g8f6 f1d3 e8g8 b1d2 d7d5 e1g1 d5e4 d2e4 f6d5 f4d2 c8f5 f3h4 f5e4 d3e4 e7h4 d1h5 g7g6 h5d5 d8d5 e4d5 a8d8 d5f3 g8g7 g2g3 h4f6 f1f2 d8d7 a1f1 f6d8 f1e1 h7h6 h2h4 c6e7 f3b7 c7c5 b7a6 c5d4 a6b5 d7d6 g1g2 e7f5 d2f4 d6e6 e1e6');
INSERT INTO evaluations VALUES(6664,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(6664,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(6664,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(6664,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(6664,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(6664,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(6664,36,'cp 156','b4e7 c1f4 g8f6 f1d3 d7d5 e4e5 f6e4 b1d2 c8f5 d1e2 g7g5 d2e4 d5e4 d3e4 f5e4 e2e4 g5f4 d4d5 d8d7 d5c6 d7c6 e4f4 e8c8 e1g1 c6d5 g1h1 c8b8 h2h3 h7h5 f1f2 a7a6 f3d4 h8e8 a1d1 d5a2 f4f7 a2f7 f2f7 h5h4');
INSERT INTO evaluations VALUES(7017,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(7017,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(7017,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(7017,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(7017,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(7017,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(7017,36,'cp 31','g1f3 d7d6 f1b5 d6e5 f3e5 g7e5 b5c6 b7c6 d4e5 d8d1 e1d1 h7h5 f2f3 g8h6 d1e2 h6f5 h1d1 h5h4 e2f2 h8h5 b2b3 f5g7 c1b2 c6c5 b1c3 g7e6 c3d5 c8b7 d5e3 a8d8 d1d8 e8d8 a1f1 b7c6 f1d1 d8c8 b2c3 c6b7 d1d2 b7c6');
INSERT INTO evaluations VALUES(13094,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(13094,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(13094,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(13094,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(13094,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(13094,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(13094,36,'cp -31','d8b6 b2b3 g7g6 h2h3 f8g7 g1e2 e8g8 b1d2 f6h5 f4h2 e7e5 d4e5 c6e5 d3c2 f8e8 e1g1 c8d7 a1b1 b6a6 c3c4 d5c4 h2e5 g7e5 d2c4 d7c6 c4e5 e8e5 b3b4 a8e8 b4c5 e5g5 g2g4 h5f6 e2d4 g5c5 d4c6 a6c6 c2b3 c6e4 d1d4 e4d4');
INSERT INTO evaluations VALUES(801,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(801,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(801,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(801,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(801,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(801,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(801,36,'cp -32','b2b3 g7g6 h2h3 f8g7 g1f3 f6e4 a2a4 c8f5 e1g1 e8g8 a1a2 a8c8 a4a5 b6d8 a5a6 b7b6 f1e1 f7f6 f4h2 e7e5 d3b5 g6g5 a2e2 h7h5 d4e5 f6e5 c3c4 g5g4 h3g4 h5g4 f3d2 g4g3 f2g3 e4d2 b1d2 d5d4 e3d4 d8d4 g1h1 c6b4 d2f3 d4d1 e1d1 c8d8 d1d8 f8d8');
INSERT INTO evaluations VALUES(15939,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(15939,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(15939,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(15939,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(15939,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(15939,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(15939,36,'cp 34','c4d5 e6d5 d1c2 c7c6 b1c3 f8e7 e2e3 e8g8 f1d3 h7h6 g5h4 f8e8 e1g1 d7f8 f3e5 f6e4 d3e4 d5e4 h4e7 d8e7 a1c1 f7f5 a2a4 c8e6 c3e2 f8d7 e5d7 e7d7 a4a5 a7a6 b2b4 e6d5 c2b2 d5f7 c1c5 f7d5');
INSERT INTO evaluations VALUES(25706,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(25706,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(25706,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(25706,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(25706,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(25706,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(25706,36,'cp 232','b5c6 d7c6 d4e5 d6e5 d1d8 a8d8 f3e5 c6d7 e5d7 d8d7 e1e2 g7g6 c1f4 f8g7 c3d5 g8f6 d5f6 g7f6 c2c3 f6g7 h1d1 d7e7 e2f1 e8g8 d1e1 b7b5 f2f3 f8b8 a2a3 a6a5 f4e3 h7h5 a1d1 b8c8 d1d2 g7e5 g2g3 a5a4 e1d1 e5g7 d2d3 c8e8 d3d5');
INSERT INTO evaluations VALUES(25699,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(25699,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(25699,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(25699,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(25699,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(25699,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(25699,36,'cp 216','d7c6 d4e5 d6e5 d1d8 a8d8 f3e5 c6d7 f2f3 d7e6 e5d3 f7f6 e1f2 h7h5 h2h4 e8d7 b2b3 d7c8 c3e2 e6f7 h1d1 g8e7 c1a3 e7c6 a3f8 h8f8 d1d2 a6a5 a1h1 f8e8 h1d1 g7g5 e2g3 b7b6 g3f5 f7g6 f5e3 g6f7 h4g5 f6g5 c2c3 f7e6 f2e2 g5g4 d3f4 g4f3 g2f3');
INSERT INTO evaluations VALUES(25661,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(25661,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(25661,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(25661,33,'cp 229',NULL);
INSERT INTO evaluations VALUES(25661,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(25661,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(25661,36,'cp 227','d4e5 d6e5 d1d8 a8d8 f3e5 c6d7 f2f3 d7e6 e5d3 f7f6 e1f2 g8e7 c1f4 d8c8 f4e3 c8d8 h2h4 e7c6 h1d1 e8f7 d3f4 f8d6 f4e6 f7e6 c3e2 e6f7 c2c3 h7h5 e2d4 c6e7 a2a4 h8e8 b2b4 e7g6 g2g3 g6e5 d4f5 e5c4');
INSERT INTO evaluations VALUES(25658,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(25658,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(25658,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(25658,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(25658,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(25658,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(25658,36,'cp 213','d6e5 d1d8 a8d8 f3e5 c6d7 f2f3 d7e6 e5d3 f7f6 e1f2 e8d7 h2h4 d7c8 b2b3 g8e7 c1a3 e7c6 a3f8 h8f8 d3f4 e6g8 h1d1 g7g5 h4g5 f6g5 d1d8 f8d8 f4d5 h7h5 a1d1 h5h4 f2e3 g8f7 a2a3 f7e6 d5f6 e6f7 d1d8 c8d8 f6d5');
INSERT INTO evaluations VALUES(25657,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(25657,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(25657,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(25657,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(25657,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(25657,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(25657,36,'cp 228','d1d8 a8d8 f3e5 c6d7 f2f3 d7e6 e5d3 f7f6 c1e3 e8d7 e3c5 g8e7 d3f4 e6f7 a1d1 d7c8 d1d8 c8d8 e1f2 h7h5 h2h4 d8c8 c5a3 b7b6 b2b3 c8b7 c3e2 e7c6 a3f8 h8f8 h1c1 g7g5 h4g5 f6g5 f4d3 a6a5 c2c4 h5h4 f2e3 f7e6 e2c3 c6b4 d3b4 a5b4 c3d5 h4h3 g2h3');
INSERT INTO evaluations VALUES(23552,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(23552,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(23552,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(23552,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(23552,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(23552,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(23552,36,'cp 208','a8d8 f3e5 c6d7 f2f3 d7e6 e5d3 f7f6 h2h4 e6c4 c1f4 e8d7 h4h5 d7c8 e1c1 g8e7 d3f2 d8d1 f2d1 c4e6 d1e3 h8g8 g2g4 e7c6 e3f5 g7g6 h5g6 h7g6 f5e3 c6e7 h1h7 g8g7 h7h8 g7g8 h8h1 f6f5 f4g5 f5e4 c3e4');
INSERT INTO evaluations VALUES(61,30,'cp 210',NULL);
INSERT INTO evaluations VALUES(61,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(61,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(61,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(61,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(61,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(61,36,'cp 234','f3e5 c6d7 f2f3 d7e6 e5d3 e8d7 h2h4 d7c8 d3f4 e6d7 g2g4 g8e7 e1f2 e7g6 h4h5 g6f4 c1f4 d7e6 f4e3 g7g6 c3e2 f8d6 e2d4 e6d7 a1d1 f7f6 d4e2 h8e8 b2b3 a6a5 c2c4 d7e6 c4c5 d6e5 e3d4 e8f8 d4e5 f6e5 d1d8 f8d8 f2e3 g6g5');
INSERT INTO evaluations VALUES(60,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(60,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(60,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(60,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(60,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(60,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(60,36,'cp 238','c6d7 f2f3 d7e6 e5d3 f7f6 h2h4 e6c4 e1e2 f8d6 e2f2 g8e7 c1e3 h7h5 a1d1 e7g6 g2g3 b7b6 b2b3 c4f7 c3e2 e8g8 d1d2 g6e5 h1d1 d6a3 e2d4 e5d3 d2d3 b6b5 d4c6 d8d6 c6e7 g8h7 e7f5 d6d3 d1d3 f7e6 f5d4');
INSERT INTO evaluations VALUES(27127,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(27127,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(27127,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(27127,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(27127,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(27127,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(27127,36,'cp 15','g8f6 e1g1 f8e7 c3e2 d6d5 e2g3 e8g8 c2c3 a7a5 d1e2 e7d6 d3d4 d5e4 f3e5 c6e5 d4e5 d6e5 g3e4 f6e4 e2e4 d8d5 e4e2 e5d6 c1e3 a5a4 a2a3 a8a5 e3d4 h7h6 a1d1 e6e5 d4e3 d5c6 f1e1 a5d5 d1d5 c6d5');
INSERT INTO evaluations VALUES(27261,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(27261,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(27261,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(27261,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(27261,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(27261,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(27261,36,'cp 57','d7b5 c3b5 c7c6 b5c3 d8b6 c1e3 b6d4 e3d4 h7h5 a1d1 g8e7 f2f3 e7g6 e1g1 f7f6 g2g3 f8e7 g1g2 a7a5 a2a4 g6f8 c3e2 e8f7 d4b6 f8e6 f3f4 g7g6 f4f5 g6f5 e4f5 e6g7 f1f3 h8e8 e2f4 e7d8 b6f2 d8c7 d1d3 d6d5');
INSERT INTO evaluations VALUES(27078,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(27078,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(27078,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(27078,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(27078,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(27078,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(27078,36,'cp 68','c3b5 c7c6 b5c3 d8b6 d4d3 g8f6 c1e3 b6c7 e1c1 f8e7 f2f3 e8g8 g2g4 b7b5 g4g5 f6d7 c1b1 b5b4 c3a4 c7a5 b2b3 d7b6 a4b2 a8d8 h2h4 d6d5 d3e2 a5b5 e2b5 c6b5 e3b6 a7b6 d1d5 d8d5 e4d5 f7f6 h1e1 g8f7 b2d3 f6g5 h4g5 e7g5 d3e5 f7g8 d5d6');
INSERT INTO evaluations VALUES(27075,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(27075,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(27075,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(27075,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(27075,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(27075,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(27075,36,'cp 74','c7c6 b5c3 d8b6 d4c4 f8e7 c1e3 b6a5 e1c1 g8f6 h1g1 e8g8 g2g4 f8d8 c1b1 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 c6d5 c4b3 b7b6 g1d1 a8d8 a2a3 h7h6 b1a2 e7f6 c2c3 g7g5 e3d4 f6d4 d1d4 a5c5 b3a4 b6b5 a4d1 c5c7 h2h3 a7a6 d1d2 c7c6 a2b1 c6e6 f2f4 g5f4 d4f4 d8e8 f4d4 e6e3 d2g2 e8d8 b1c2 d8e8 g2d5 e3h3');
INSERT INTO evaluations VALUES(30538,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(30538,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(30538,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(30538,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(30538,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(30538,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(30538,36,'cp 125','d2d4 a7a6 f1c4 e7e6 d4d5 e6e5 a2a4 g8f6 f3d2 b8d7 d1e2 g7g6 g2g4 f6g8 d2b3 f8e7 h2h3 h7h5 h1g1 e7g5 c1d2 g5d2 b3d2 d8g5 e1c1 g8f6 b2b3 e8e7 c1b2 h5g4 h3g4 h8h3 f2f3 g5f4 g1h1');
INSERT INTO evaluations VALUES(31426,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31426,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(31426,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(31426,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(31426,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31426,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(31426,36,'cp 6','g8f6 d3g6 h7g6 d1e2 d8b6 c2c3 c6c5 d4c5 f8c5 c1g5 b6a6 h2h4 a6e2 e1e2 b8c6 a1d1 a8c8 f3d2 c5e7 d2c4 c8d8 d1d8 c6d8 g5f6 g7f6 h4h5 e8f8 f2f4 g6h5 g3h5 b7b5 c4e3 f6f5 e2f3 b5b4 c3b4 e7b4 e3c4');
INSERT INTO evaluations VALUES(29326,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29326,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29326,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29326,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(29326,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(29326,36,'cp 9','c1g5 c6c5 d1e2 b8c6 d3g6 h7g6 d4c5 f8c5 h2h4 c5e7 e1g1 e8g8 c2c3 d8c7 a1d1 a8d8 g3e4 c7a5 a2a3 a5f5 e4g3 f5a5 d1c1 a5c7 b2b4 d8d5 f1e1 a7a6 c3c4 d5d7 g3e4 f8d8 g2g3 f6e4 e2e4 e7f6 g1g2 d7d3 b4b5');
INSERT INTO evaluations VALUES(33224,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(33224,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(33224,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(33224,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(33224,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(33224,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(33224,36,'cp 80','e2e3 d7d6 f1g2 g8e7 g1e2 a7a5 e1g1 e8g8 h2h3 c5a7 d2d4 c7c6 b2b3 b8a6 a2a3 f6e6 c1d2 c8d7 a1c1 a8d8 g1h2 a7b6 f2f4 e6g6 c4c5 e5d4 c5b6 d4c3 d2c3 d6d5 g3g4 f7f5 g4g5 a6c5 c3a5 c5e4 e2d4 f8e8 g2e4 f5e4');
INSERT INTO evaluations VALUES(33221,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(33221,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(33221,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(33221,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(33221,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(33221,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(33221,36,'cp 75','g8e7 f1g2 d7d6 g1e2 a7a5 e1g1 e8g8 d2d4 c5b6 c3a4 b6a7 e2c3 b8d7 c3b5 a7b8 b2b3 f6g6 c1a3 e5e4 f2f3 e4f3 d1f3 e7c6 c4c5 a8a6 c5d6 c7d6 f3f5 g6f5 f1f5 c6b4 a1c1 b4a2 c1c8 f8c8 g2b7 c8c6 g1f1 d7f6 f1e2 a2b4 a4c3 c6b6 b7a6 b6a6 a3b4 a5b4');
INSERT INTO evaluations VALUES(35423,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(35423,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(35423,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(35423,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35423,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(35423,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(35423,36,'cp 18','b8d7 e1c1 f8b4 c3d5 b4d6 d5f6 g7f6 g5e3 d8e8 g1e2 d6c5 d1d3 a7a5 e2g3 c7c6 f1e2 c5e3 d3e3 d7c5 e3a3 b7b6 b2b4 c5b7 h1d1 e8e7 a3c3 c6c5 b4a5 b7a5 e2b5 c8e6 c3d3 a8d8 d3d8 h8d8 d1d8 e7d8 a2a4 a5b7 c1d2 b7d6');
INSERT INTO evaluations VALUES(38166,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(38166,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(38166,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(38166,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38166,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(38166,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(38166,36,'cp 33','b4d2 d1d2 f7f6 e5f6 g8f6 f1d3 b8c6 g1f3 e8g8 b1c3 d8d6 h2h3 e6e5 d4e5 c6e5 d3e2 e5f3 e2f3 f8e8 c3e2 c8f5 e1g1 a8d8 e2d4 f5g6 d2a5 a7a6 a1c1 d8c8 f1d1 h7h6 a2a3 g6e4 c1c8 e8c8 f3e2 e4g6 e2g4 f6g4 h3g4 c8c4 a5d2');
INSERT INTO evaluations VALUES(38176,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(38176,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38176,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(38176,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(38176,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(38176,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(38176,36,'cp 37','d1d2 f7f6 g1f3 f6e5 f3e5 g8e7 f2f4 e8g8 b1c3 c8d7 f1d3 b8c6 e5f3 d8b6 e1g1 c6d4 d3h7 g8h7 d2d4 b6d4 f3d4 a8c8 c3e2 h7g8 f1f3 e7g6 h2h3 a7a5 a1f1 e6e5 f4e5 g6e5 f3f8 c8f8 f1d1 d7c6 d4e6 f8c8 e2d4 g8f7 d1f1 f7g8 e6f4 c8e8 f1d1 a5a4 g1h2 e5c4 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(38175,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38175,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38175,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(38175,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(38175,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(38175,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(38175,36,'cp 28','f7f6 g1f3 f6e5 f3e5 g8e7 d2f4 h8f8 f4d2 a7a6 f1d3 b8c6 b1c3 c6e5 d4e5 d8c7 d2g5 e7c6 f2f4 g7g6 e1c1 c7e7 g5g3 e7b4 h1f1 c8d7 a2a3 b4e7 c3a4 e8c8 c1b1 c8b8 d1c1');
INSERT INTO evaluations VALUES(38503,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(38503,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(38503,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(38503,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(38503,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(38503,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(38503,36,'cp 82','c1f4 g8f6 d1d2 e8g8 e1c1 c8g4 f1e2 b8d7 h2h3 g4f3 e2f3 c7c6 g2g4 d8a5 h1e1 d7b6 e1e5 f6d5 c3d5 a5d2 f4d2 b6d5 e5e1 e7e6 c2c3 a7a5 c1c2 a5a4 a2a3 b7b5 h3h4 g7f6 h4h5 f6e7 d1c1 e7d6 c2b1 a8b8 e1h1 b8c8 h5g6 h7g6');
INSERT INTO evaluations VALUES(39472,30,'cp 235',NULL);
INSERT INTO evaluations VALUES(39472,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(39472,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(39472,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(39472,34,'cp 223',NULL);
INSERT INTO evaluations VALUES(39472,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(39472,36,'cp 243','c3d5 b4e7 d2d4 c7c6 d5e7 d8e7 c1e3 g8h6 e3h6 g7h6 d1d2 d7d5 e4d5 e5e4 f3h4 c6d5 d2h6 c8e6 c2c3 h8f8 f1e2 b8c6 e2h5 e8d7 g2g3 a8d8 h4g2 d7c8 h5e2 e7d7 e1c1 c8b8 c1b1 a7a6 b1a1 b8a7 d1c1 d8c8 h6d2 c6e7 g2f4 e7g6 f4e6');
INSERT INTO evaluations VALUES(44875,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(44875,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(44875,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(44875,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(44875,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(44875,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(44875,36,'cp 136','d5c6 b8c6 f1b5 d8b6 b5a4 c8g4 e1g1 e7e6 d2d4 b6a6 b1c3 f8b4 d1e2 a6a5 a4c6 b7c6 c3e4 a5d5 a2a3 b4a5 e4g3 g4f3 g2f3 e8g8 f1d1 a5c7 c2c4 d5d6 f3f4 a8d8 c1e3 a7a5 b2b3 g7g6 d1d3 h7h5 e2f3 f6g4 g3e4 d6e7 a1d1 e7a3 e4c5 c7b6');
INSERT INTO evaluations VALUES(44844,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(44844,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(44844,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(44844,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(44844,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(44844,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(44844,36,'cp 134','b8c6 f1b5 d8b6 b5a4 c8g4 e1g1 e7e6 d2d4 b6a6 b1c3 f8b4 d1e2 a6a5 a4c6 b7c6 c3e4 a5d5 e4g3 b4e7 f1d1 a7a5 c1f4 f6h5 g3h5 g4h5 c2c4 h5f3 g2f3 d5d7 d1d3 e8g8 a1d1 h7h6 g1g2 a5a4 h2h3 a8e8 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(47210,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(47210,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(47210,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(47210,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(47210,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(47210,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(47210,36,'cp -39','c7c5 e2e3 e7e6 d4c5 f8c5 g1f3 e8g8 f1d3 c5e7 e1g1 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d7c5 d1e2 d8c7 f1e1 f8d8 c1d2 c8d7 d2c3 a8c8 a1d1 c5e4 e2e4 d7c6 d1d8 c8d8 e4g4 e7f8 c3f6 d8d7 f6c3 a7a6 f3e5 d7d5 e5c6 c7c6');
INSERT INTO evaluations VALUES(48781,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(48781,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(48781,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(48781,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(48781,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(48781,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(48781,36,'cp 137','e4d5 f8d6 d2d3 e8g8 g1e2 c7c6 e1g1 f6d5 c3d5 c6d5 c4d5 b8c6 a2a3 c8g4 f2f3 d8b6 g1h1 g4d7 c2c4 c6e7 e2c3 d6c5 b2b4 c5d4 c1g5 f8e8 a1c1 d4c3 c4c5 b6c7 g5e7 e8e7 c1c3 a8d8 d5b3');
INSERT INTO evaluations VALUES(49482,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(49482,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(49482,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(49482,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(49482,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(49482,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(49482,36,'cp -62','f8b4 e2e3 b4c3 b2c3 d7d6 c3c4 c7c6 g1f3 e8g8 f1e2 d8a5 d1d2 a5a4 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3h4 c6d5 c4d5 f6e4 d2e1 a4c2 f2f3 e4c3 f1f2 c3d5 e3e4 d5f6 c1g5 b8d7 a1d1 c2c7 e1d2 a8d8 d2d6 c7d6 d1d6 d7c5 d6d8 f8d8 e2b5 g8h8 g5e3 c5e6 h4f5 g6f5 e4f5 e6d4');
INSERT INTO evaluations VALUES(42696,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(42696,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(42696,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(42696,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(42696,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(42696,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(42696,36,'cp -74','e2e3 b4c3 b2c3 d7d6 c3c4 c7c6 g1f3 e8g8 f1e2 d8a5 d1d2 a5a4 e1g1 f8d8 c1b2 c8g4 h2h3 g4f3 g2f3 b8d7 d5c6 a4c6 a1d1 d7b6 f3f4 b6c4 e2c4 c6c4 f4e5 f6e8 b2a3 c4a2 d2d3 d6d5 a3c5 a8c8 c5d4 c8c2 d1a1 a2c4 d3c4 d5c4');
INSERT INTO evaluations VALUES(53485,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(53485,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(53485,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(53485,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(53485,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(53485,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(53485,36,'cp 51','g8f6 d2d4 b8d7 a2a4 f8e7 f1e1 e8g8 h2h3 f8e8 a4a5 e5d4 f3d4 e7f8 c1f4 a7a6 f4g3 d7c5 e4e5 d6e5 g3e5 c5d7 e5g3 e8e1 d1e1 d7c5 e1e3 d8e7 e3e7 f8e7 a1e1 e7f8 f2f3 c8d7 g1f1 a8e8 e1d1 f8e7 g3h2 g8f8 g2g4 e8c8 h2g3 c5e6 d4e6 d7e6 c4e6 f7e6 f1e2');
INSERT INTO evaluations VALUES(45556,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(45556,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(45556,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(45556,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(45556,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(45556,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(45556,36,'cp 65','d2d4 b8d7 h2h3 f8e7 a2a4 e8g8 a4a5 d8c7 f1e1 f8e8 c4f1 d7f8 d4d5 f8g6 f3d2 c8d7 a5a6 c6d5 e4d5 b7b6 c3b5 c7d8 d2f3 e7f8 c1e3 f8e7 c2c4 d7b5 c4b5 e7f8 d1b3 d8d7 a1c1 g6e7 f3d2 f6d5 d2e4');
INSERT INTO evaluations VALUES(45563,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(45563,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(45563,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(45563,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(45563,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(45563,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(45563,36,'cp 51','b8d7 a2a4 f8e7 h2h3 e8g8 a4a5 f8e8 f1e1 d8c7 c4f1 d7f8 d4d5 c8d7 f3d2 e8b8 d2c4 c6d5 e4d5 b7b5 a5b6 a7b6 a1a8 b8a8 c1e3 a8b8 c4a3 c7d8 d1d3 f8g6 d3c4 f6e8 c4b4 g6h4 a3c4 b6b5 c4a5 h4f5 c3b5 f5e3 e1e3');
INSERT INTO evaluations VALUES(54818,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(54818,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(54818,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(54818,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(54818,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(54818,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(54818,36,'cp 30','g8f6 b1c3 f8g7 d1d2 e8g8 e2e4 c8g4 e1c1 b8c6 f4e3 e7e5 d4e5 d6e5 d2d8 a8d8 d1d8 f8d8 f3d2 g4e6 f2f3 a7a6 f1c4 e6c4 d2c4 b7b5 c4d2 c6d4 d2b3 d4b3 a2b3 g7f8 e3g5 f8e7 c3d1 g8f8 c2c3 f6h5 g5h6 f8e8 c1c2 e7c5 g2g3');
INSERT INTO evaluations VALUES(57373,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(57373,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(57373,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(57373,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(57373,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(57373,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(57373,36,'cp 61','d2d4 f6e5 d4e5 b8c6 g1f3 d7d5 e5d6 f8d6 f1d3 g8f6 e1g1 e8g8 b1d2 h7h6 h2h3 e6e5 a2a3 a7a5 f1e1 c8e6 d1e2 d8e8 d2e4 f6e4 d3e4 e8f7 c1e3 e6f5 e4f5 f7f5 e2c4 f5f7 c4e4 f7f5 e4f5 f8f5 a1d1 a5a4');
INSERT INTO evaluations VALUES(57372,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(57372,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(57372,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(57372,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(57372,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(57372,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(57372,36,'cp 58','f6e5 d4e5 b8c6 g1f3 d7d5 e5d6 f8d6 f1d3 g8f6 e1g1 e8g8 h2h3 c6b4 d3c4 g8h8 b1c3 e6e5 a2a3 b4c6 c4d3 d8e8 f1e1 c8e6 b2b4 e6f7 c3e4 a8d8 b4b5 c6e7 f3g5 f7g8 e4f6 g7f6 g5e4 e8f7 c1b2 e7g6 e4d6');
INSERT INTO evaluations VALUES(57337,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(57337,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(57337,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(57337,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(57337,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(57337,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(57337,36,'cp 52','d4e5 b8c6 g1f3 d7d5 e5d6 f8d6 f1d3 g8f6 b1d2 e8g8 c2c3 h7h6 d2e4 f6e4 d3e4 d8f6 c1e3 e6e5 e1g1 c6e7 d1e2 g8h8 a1d1 c8f5 f3d2 f6f7 a2a4 a7a6 e4b7 a8b8 b7a6 b8b2 a6c4 e7d5 c4d3 d5f4 e3f4 e5f4');
INSERT INTO evaluations VALUES(57334,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(57334,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(57334,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(57334,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(57334,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(57334,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(57334,36,'cp 52','b8c6 g1f3 d7d5 e5d6 f8d6 f1d3 g8f6 e1g1 e8g8 b1d2 h7h6 f1e1 e6e5 c2c3 c8e6 d2e4 f6e4 d3e4 d8f6 d1c2 f6f7 e4g6 f7f6 c1e3 c6e7 g6e4 e6f5 c2b3 f6f7 b3f7 g8f7 e4b7 a8b8 b7a6 b8b2 a6c4 f7f6 c4b3 f5e6 b3e6 f6e6');
INSERT INTO evaluations VALUES(2571,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(2571,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(2571,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(2571,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(2571,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(2571,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(2571,36,'cp 39','b7b5 a4c2 d6d5 d2d4 e5d4 e4e5 f6e4 c3d4 c6b4 c1e3 b4c2 d1c2 c8f5 e1c1 f7f6 c2c6 e7b4 a2a3 b4a5 b1d2 e4d2 e3d2 a5b6 d2b4 f8e8 e5f6 d8f6 c6d5 f5e6 d5c6 e6c4 c6f6 g7f6 c1e1 e8e2 e1e2 c4e2 b4c5 e2f3 g2f3 b6c5 d4c5 g8f7 a1c1 a6a5');
INSERT INTO evaluations VALUES(2733,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(2733,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(2733,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(2733,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(2733,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(2733,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(2733,36,'cp 55','a4c2 d6d5 d2d4 e5d4 e4e5 f6e4 c3d4 c6b4 c1e3 b4c2 d1c2 e7b4 b1d2 c8f5 e1c1 e4d6 e3g5 f5c2 g5d8 f8d8 c1c2 d6c4 d2f1 c7c5 a1d1 c5d4 d1d4 b4c5 d4d1 a8c8 b2b3 c4b6 g2g4 c5b4 c2c8 b6c8 f1e3 c8e7 f3d4 b4c5 g1g2');
INSERT INTO evaluations VALUES(2730,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(2730,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(2730,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(2730,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(2730,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(2730,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(2730,36,'cp 47','d6d5 d2d4 f6e4 d4e5 c8f5 c1e3 d8d7 b1d2 a8d8 d1b1 e4d2 e3d2 f5c2 b1c2 d8e8 a2a4 c6d8 d2g5 d8e6 g5e7 d7e7 c2d2 e7d7 a4b5 a6b5 f3d4 c7c6 b2b4 f7f6 d4f3 d7d8 a1a7 h7h6 f3d4 e6d4 c3d4 f6e5 e1e5 d8f6 d2e3 e8e5 d4e5');
INSERT INTO evaluations VALUES(12910,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12910,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12910,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12910,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12910,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12910,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12910,36,'cp 6','d1e2 e7e6 h2h3 g4h6 c4b5 h6f5 b5c6 b7c6 d2d3 d7d5 e5d6 f8d6 e1g1 e8g8 b1d2 f5d4 e2d1 e6e5 d2c4 f7f6 c4d6 d8d6 f3d2 d6d5 b2b3 c8f5 f1e1 a8d8 d2e4 f5g6 c1b2 d4e6 f2f3 h7h6 d1d2 d5d7 d2f2 g6e4');
INSERT INTO evaluations VALUES(15407,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(15407,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(15407,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(15407,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(15407,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(15407,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(15407,36,'cp 25','g1f3 f8c5 c1b2 c5d6 e1g1 e8g8 d2d4 f8e8 e2e4 c8g4 h2h3 g4h5 d1d3 d8d7 f3d2 f7f6 d2c4 d6c5 d4d5 b7b5 c4d2 c6a5 d2b3 a5b3 a2b3 a7a5 b2c1 h5f7 c1e3 c5e3 d3e3 a5a4 c3c4 b5c4 b3c4 e8b8 c4c5 b8b3 c5c6 b3e3 c6d7');
INSERT INTO evaluations VALUES(15406,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(15406,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15406,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(15406,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(15406,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(15406,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(15406,36,'cp 13','f8c5 c1b2 e8g8 d2d4 c5d6 e1g1 f8e8 e2e4 f7f6 d1c2 b7b6 f3h4 c6a5 h4f5 a5c4 a2a4 c8e6 f1d1 d8d7 c2e2 d7f7 f5d6 c4d6 d4d5 e6d7 c3c4 d6b7 b2c1 a7a5 c1e3 g8h8 d1b1 f7e7 g2f3 e7f7 c4c5 b7c5 e3c5 b6c5');
INSERT INTO evaluations VALUES(9141,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(9141,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(9141,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(9141,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(9141,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(9141,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(9141,36,'cp 33','c1b2 e8g8 d2d4 c5d6 e1g1 f8e8 a2a4 c8g4 a4a5 a8b8 h2h3 g4h5 d1c2 e5d4 c3d4 h5g6 c2c4 d8e7 e2e3 c6b4 c4c3 g6e4 f3d2 e4g2 g1g2 b4d5 c3b3 d5f6 b2a3 b7b5 a5b6 a7b6 a3d6 c7d6 b3b5 e7b7 g2g1 e8c8 f1c1 d6d5 d2f3 c8c1 a1c1 b8c8 c1c8 b7c8 b5b6');
INSERT INTO evaluations VALUES(15597,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(15597,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(15597,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(15597,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(15597,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(15597,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(15597,36,'cp 8','g7g6 e1g1 f8g7 c2c4 c8g4 h2h3 g4f3 g2f3 f6d7 b1c3 a7a5 f3g2 e8g8 c1e3 d7c5 f2f4 e5f4 g3f4 c5e6 g1h1 g7d4 d1d2 e6g7 a1d1 f7f5 c3d5 g8h8 d2f2 d4e3 f2e3 g7h5 e4f5 f8f5 h1h2 f5f7 d1e1 d8h4 e3f2 h4f2 f1f2 c6d4');
INSERT INTO evaluations VALUES(26526,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(26526,31,'cp 269',NULL);
INSERT INTO evaluations VALUES(26526,32,'cp 343',NULL);
INSERT INTO evaluations VALUES(26526,33,'cp 329',NULL);
INSERT INTO evaluations VALUES(26526,34,'cp 333',NULL);
INSERT INTO evaluations VALUES(26526,35,'cp 342',NULL);
INSERT INTO evaluations VALUES(26526,36,'cp 344','c4d5 d8d5 b1c3 d5d6 e3e4 f5c8 d4d5 c6b8 g1f3 e7e5 c3b5 d6d8 f3e5 g8f6 d1b3 f8d6 b5d6 c7d6 e5d3 e8g8 f2f3 f6h5 g2g4 h5f6 f1e2 b7b6 c1f4 b8a6 d3f2 f8e8 b3c3 a6c5 c3d2 a7a5 e1g1 h7h5 g4g5 f6h7 f4e3 c8a6 e2a6');
INSERT INTO evaluations VALUES(29166,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(29166,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(29166,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(29166,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(29166,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(29166,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(29166,36,'cp 105','f3e5 g4d7 e2e4 e7e6 e4d5 e6d5 d1b3 d7c6 c1f4 d8b6 b3b6 a7b6 a2a3 b8d7 f1b5 d7e5 d4e5 d5d4 e1c1 d4c3 e5f6 c3b2 c1b1 g7f6 h1e1 f8e7 f4d6 e8c8 e1e7 c6b5 e7c7 c8b8 c7f7 b8a7 f7f6 h8e8 d1d2 e8e5 g2g3 b5a4 f6f4 a4c6 b1b2 e5b5 b2c3 b5b1 f4f8 d8f8 d6f8 c6a4 f8b4 b1a1 c3d4');
INSERT INTO evaluations VALUES(29189,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(29189,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(29189,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(29189,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(29189,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(29189,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(29189,36,'cp 99','g4c8 c1f4 e7e6 e2e3 b8c6 a1c1 c8d7 f1d3 f8e7 e1g1 a8c8 d1f3 h7h6 h2h3 e8g8 a2a3 f6h7 f3g4 c6e5 d4e5 g8h8 b2b4 a7a5 c3e2 c8c1 f1c1 a5b4 a3b4 d8b6 b4b5 h7g5 f4g5 e7g5 g4d4 b6a5 c1a1 a5d2 b5b6 g5e7');
INSERT INTO evaluations VALUES(35841,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(35841,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(35841,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(35841,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(35841,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(35841,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(35841,36,'cp 43','f1d3 c8d7 h2h3 d7c6 e1g1 b8d7 a1c1 e7d6 g5f4 d8b8 c3e2 f8c8 c1c2 d7b6 d1d2 d6f4 e2f4 c6e8 f1c1 c8c2 c1c2 a7a5 f3e5 b6d7 e5g4 f6g4 h3g4 h7h6 a2a3 d7f6 f2f3 e8c6 f4e2 b8d8 d2c3 a8c8');
INSERT INTO evaluations VALUES(35842,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(35842,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(35842,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(35842,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(35842,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(35842,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(35842,36,'cp 34','b8c6 e1g1 c8d7 h2h3 f6e8 g5f4 a8c8 a1c1 e8d6 f3e5 d7e8 d1c2 f7f5 c2b3 d8b6 b3b6 a7b6 e5f3 h7h6 c3b5 d6b5 d3b5 g7g5 f4h2 f5f4 e3f4 g5f4 f1e1 e8d7 c1c2 e7d6 a2a3 c8c7 e1c1 f8c8');
INSERT INTO evaluations VALUES(3103,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(3103,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(3103,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(3103,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(3103,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(3103,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(3103,36,'cp 36','e1g1 c8d7 h2h3 f6e8 g5f4 a8c8 a1c1 e8d6 d1c2 f7f5 c2b3 c6a5 b3d1 d7e8 d3e2 h7h6 f4h2 a5c6 f3e1 c6b4 a2a3 b4c6 e1d3 d6c4 c3b1 e7d6 h2d6 c4d6 b1d2 d8e7 d2f3 a7a5 d1b3 d6c4 d3f4 g7g5');
INSERT INTO evaluations VALUES(45226,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(45226,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(45226,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(45226,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(45226,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(45226,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(45226,36,'cp 67','e7e6 b1c3 e6d5 c4d5 g7g6 e2e4 f8g7 f3d2 b8d7 f1e2 e8g8 e1g1 f8e8 a2a4 a7a6 h2h3 g6g5 d2c4 f6e4 c3e4 e8e4 c4d6 e4d4 d1b3 d7e5 d6c8 a8c8 b3g3 d4d5 c1g5 d8g5 g3g5 e5f3 e2f3 d5g5 f3b7 c8b8 b7a6 b8b2 a1b1 b2a2 b1b8 g7f8 a6b5 g8g7 h3h4 g5f5 b8b7 f5f6 h4h5 c5c4 h5h6 f6h6');
INSERT INTO evaluations VALUES(46031,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(46031,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(46031,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(46031,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(46031,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(46031,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(46031,36,'cp 90','b1c3 e6d5 c4d5 g7g6 c1f4 a7a6 a2a4 f8g7 h2h3 e8g8 e2e3 b8d7 f1e2 f6e8 e1g1 d7e5 f3d2 b7b6 f4g3 f7f5 d1c2 c8d7 a1b1 d8e7 f1e1 g6g5 f2f4 e5g6 e2d3 g8h8 g1h1 e8c7 d2c4 b6b5 a4b5 c7b5 g3f2 b5c3');
INSERT INTO evaluations VALUES(46029,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(46029,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(46029,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(46029,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(46029,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(46029,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(46029,36,'cp 82','g7g6 e2e4 f8g7 f1d3 e6d5 e4d5 e8g8 h2h3 f8e8 c1e3 b8d7 e1g1 a7a6 d1d2 b7b5 c4b5 d7b6 b5a6 c8a6 d3a6 a8a6 b2b3 f6e4 c3e4 g7a1 f1a1 e8e4 d2d3 e4e3 d3a6 e3f3 g2f3 b6d5 g1h2 d8f6 a1e1 f6f4 h2g1 g8g7 a6c4 f4f3 c4e4 f3e4 e1e4 d5c3 e4e1 c3a2 e1d1 d6d5 d1d5 a2c1 d5c5 c1b3 c5d5 g7f6 d5d6 f6e7');
INSERT INTO evaluations VALUES(46906,30,'cp 386',NULL);
INSERT INTO evaluations VALUES(46906,31,'cp 379',NULL);
INSERT INTO evaluations VALUES(46906,32,'cp 386',NULL);
INSERT INTO evaluations VALUES(46906,33,'cp 390',NULL);
INSERT INTO evaluations VALUES(46906,34,'cp 390',NULL);
INSERT INTO evaluations VALUES(46906,35,'cp 378',NULL);
INSERT INTO evaluations VALUES(46906,36,'cp 390','g5f7 e8f7 d3g6 f7g6 d1d8 b8c6 d8d1 e7e5 c2c3 g6h7 e1g1 c8f5 b1d2 f8e7 d1b3 a8b8 d2f3 h8f8 c1e3 f5e4 f3d2 e4g6 a1d1 e7d6 d2c4 d6e7 a2a4 a7a6 a4a5 h7h8 b3a4 g6h7 h2h3 h7g8 e3c1 b8d8 c4e3 e7c5 b2b4 c5e7 d1d8 f8d8');
INSERT INTO evaluations VALUES(46610,30,'cp 359',NULL);
INSERT INTO evaluations VALUES(46610,31,'cp 363',NULL);
INSERT INTO evaluations VALUES(46610,32,'cp 374',NULL);
INSERT INTO evaluations VALUES(46610,33,'cp 371',NULL);
INSERT INTO evaluations VALUES(46610,34,'cp 378',NULL);
INSERT INTO evaluations VALUES(46610,35,'cp 363',NULL);
INSERT INTO evaluations VALUES(46610,36,'cp 375','e8f7 d3g6 f7g6 d1d8 b8c6 d8d1 e7e5 c2c3 g6h7 c1e3 c8f5 b1d2 f8d6 e1g1 a7a6 d2c4 d6e7 a2a4 h8f8 b2b4 f6e4 d1b3 f5e6 b3b1 e6f5 b1b2 e4d6 c4d6 e7d6 a1d1 f5g6 f2f3 c6e7 c3c4 e7f5 e3c5 b7b6 c5d6 c7d6');
INSERT INTO evaluations VALUES(35482,30,'cp 382',NULL);
INSERT INTO evaluations VALUES(35482,31,'cp 381',NULL);
INSERT INTO evaluations VALUES(35482,32,'cp 382',NULL);
INSERT INTO evaluations VALUES(35482,33,'cp 385',NULL);
INSERT INTO evaluations VALUES(35482,34,'cp 378',NULL);
INSERT INTO evaluations VALUES(35482,35,'cp 386',NULL);
INSERT INTO evaluations VALUES(35482,36,'cp 375','d3g6 f7g6 d1d8 b8c6 d8d1 e7e5 c2c3 g6h7 c1e3 c8f5 b1d2 f8d6 e1g1 a7a6 d2c4 d6e7 a2a4 f5g6 a4a5 g6f7 d1b3 a8b8 b3a4 h8f8 a1d1 h7h8 f1e1 f7g8 e3c1 e7c5 h2h3 b8e8 c1e3 c5e7 f2f3 e5e4 c4d2 g8h7 d2e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(35483,30,'cp 378',NULL);
INSERT INTO evaluations VALUES(35483,31,'cp 382',NULL);
INSERT INTO evaluations VALUES(35483,32,'cp 378',NULL);
INSERT INTO evaluations VALUES(35483,33,'cp 382',NULL);
INSERT INTO evaluations VALUES(35483,34,'cp 375',NULL);
INSERT INTO evaluations VALUES(35483,35,'cp 382',NULL);
INSERT INTO evaluations VALUES(35483,36,'cp 358','f7g6 d1d8 b8c6 d8d1 c8g4 f2f3 a8d8 b1d2 g4c8 c2c3 c8f5 d1a4 g6h7 d2e4 f6e4 f3e4 f5g6 a4b5 g6e4 e1g1 e7e6 c1e3 f8d6 a1e1 c6e5 e3f4 e4d3 b5b7 d3f1 f4e5 f1d3 e5d6 d8d6 b7c7 h8d8 c7a7 d3g6 h2h3 d6d2 a7c7 d8f8 c7b6 d2c2 b6b5 f8f2');
INSERT INTO evaluations VALUES(35481,30,'cp 378',NULL);
INSERT INTO evaluations VALUES(35481,31,'cp 383',NULL);
INSERT INTO evaluations VALUES(35481,32,'cp 386',NULL);
INSERT INTO evaluations VALUES(35481,33,'cp 390',NULL);
INSERT INTO evaluations VALUES(35481,34,'cp 386',NULL);
INSERT INTO evaluations VALUES(35481,35,'cp 393',NULL);
INSERT INTO evaluations VALUES(35481,36,'cp 382','d1d8 b8c6 d8d1 e7e5 c1e3 g6h7 c2c3 c8f5 b1d2 f8e7 d1b3 a8b8 e1g1 h8e8 f2f3 f5d3 e3f2 b7b5 f1e1 b8d8 h2h3 a7a6 a2a4 b5b4 c3b4 d3g6 b4b5 a6b5 a4b5 c6d4 b3c3 d4b5');
INSERT INTO evaluations VALUES(35389,30,'cp 374',NULL);
INSERT INTO evaluations VALUES(35389,31,'cp 389',NULL);
INSERT INTO evaluations VALUES(35389,32,'cp 370',NULL);
INSERT INTO evaluations VALUES(35389,33,'cp 371',NULL);
INSERT INTO evaluations VALUES(35389,34,'cp 361',NULL);
INSERT INTO evaluations VALUES(35389,35,'cp 371',NULL);
INSERT INTO evaluations VALUES(35389,36,'cp 378','b8c6 d8d1 e7e5 c2c3 g6h7 c1e3 c8f5 b1d2 f8d6 d2c4 d6e7 e1g1 a7a6 a2a4 h8d8 d1b3 b7b6 a1d1 d8f8 h2h3 h7h8 c4d2 a8d8 d2f3 e7d6 b3c4 c6e7 c4a6 f5c2 a6b5 c2d1 f1d1 e7f5 e3d2 d6c5 f3e5 h8h7 b5b3 f6e4 d2e1 d8d1 b3d1');
INSERT INTO evaluations VALUES(51444,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(51444,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(51444,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(51444,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(51444,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(51444,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(51444,36,'cp 112','g1f3 b8c6 g2g3 f8b4 c3d5 b4d6 f1g2 g8e7 d5c3 e8g8 e1g1 a8b8 d2d3 d6b4 c1d2 d7d6 a2a3 b4c3 d2c3 d8e8 e2e3 e7g6 b2b4 f7f5 h2h4 c6d8 a3a4 d8e6 b4b5 c8b7 a4a5 e8f7 a5a6 b7a8 d3d4 e5e4 f3d2 b8e8 d4d5');
INSERT INTO evaluations VALUES(51443,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(51443,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(51443,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(51443,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(51443,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(51443,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(51443,36,'cp 91','b8c6 g2g3 f8b4 c3d5 b4d6 d2d4 g8e7 d4e5 c6e5 f3e5 d6e5 f1g2 a8b8 c1f4 e5f4 d5f4 e8g8 e1g1 c8b7 e2e4 c7c5 d1d3 e7g6 f4e2 b7a6 f2f4 b6b5 e2c3 b5c4 d3d6 d8c8 a1d1 f8d8 f4f5 b8b6 f5g6 b6d6 g6f7 g8f8 d1d6');
INSERT INTO evaluations VALUES(52420,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52420,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52420,36,'cp 8','b8c6 c1b2 g8f6 b1c3 d7d5 c4d5 f6d5 g1f3 e7e5 e2e3 d5c3 d2c3 c8e6 d1d8 a8d8 f1b5 f7f6 f3d2 e8d7 e1c1 f8d6 b5c4 d7e7 c4e2 h7h5 d2e4 h5h4 f2f3 e6f5 e4d6 d8d6 d1d6 e7d6 h1d1 d6e7 e3e4 f5g6');
INSERT INTO evaluations VALUES(19396,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19396,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19396,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19396,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19396,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19396,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19396,36,'cp 0','c1b2 g8f6 b1c3 d7d5 c4d5 f6d5 g1f3 d5c3 b2c3 f7f6 d1c2 e7e5 e2e3 c8e6 f1d3 d8d7 e1g1 e8c8 d3e4 e6d5 e4f5 d5e6');
INSERT INTO evaluations VALUES(19397,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19397,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19397,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19397,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19397,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19397,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19397,36,'cp 0','g8f6 g1f3 d7d5 c4d5 f6d5 b1c3 d5c3 b2c3 f7f6 e2e3 e7e5 d1c2 c8e6 f1d3 d8d7 a1d1 e8c8 d3e4 e6d5 e4f5 d5e6');
INSERT INTO evaluations VALUES(56369,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(56369,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(56369,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(56369,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(56369,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(56369,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(56369,36,'cp 167','d4e5 g8e7 c1f4 h7h6 h2h4 c7c5 e2e4 d5e4 d1d8 e8d8 f3d2 e7d5 g2g3 b8c6 b1c3 d5c3 b2c3 e4e3 f4e3 d8c7 f2f4 c8e6 e1f2 f7f6 e5f6 g7f6 a2a4 a8d8 f1d3 f6f5 a1e1 b7b6 d2f3 d8e8 c3c4 f8g7 h4h5 c6b4 e3d2 g7d4 f3d4 b4d3 c2d3 c5d4');
INSERT INTO evaluations VALUES(56315,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(56315,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(56315,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(56315,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(56315,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(56315,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(56315,36,'cp 141','b8c6 c1f4 f7f6 b1c3 f8b4 e2e3 g7g5 f4g3 g5g4 f3d4 c6e5 f2f4 e5f7 f4f5 h7h5 f1d3 d8e7 e1g1 c7c6 g3f2 b4c3 b2c3 f7d6 e3e4 d6e4 d3e4 d5e4 d1e2 g8h6 f2g3 h6f7 a1e1 e8g8 e2e4 e7e4 e1e4');
INSERT INTO evaluations VALUES(21653,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(21653,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(21653,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(21653,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(21653,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(21653,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(21653,36,'cp 150','c1f4 f7f6 e2e3 g7g5 f4g3 g5g4 f3d4 c6e5 b1c3 f8b4 f2f4 e5f7 f4f5 h7h5 d4b5 b4c3 b2c3 f7e5 d1d5 d8d5 b5c7 e8f7 c7d5 c8f5 g3e5 f6e5 a1b1 b7b6 f1d3 g8f6 d5f6 f5d3 c2d3 f7f6 d3d4 f6e6 e1e2 h5h4 b1b5 e5d4 c3d4 h4h3 g2g3 h8f8 b5e5 e6d7 e5g5 a8c8 e2d3 f8f2 g5g7 d7d6 g7a7');
INSERT INTO evaluations VALUES(3051,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3051,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3051,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3051,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3051,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3051,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3051,36,'cp 0','d7d5 c3d5 f6d5 c4d5 d8d5 c1e3 g7b2 a1b1 b2f6 d1a4 d5d7 e3c5 b7b6 c5b6 c6d4 a4d1 a7b6 f3d4 a8a2 d1b3 a2d2 b3b4 d2a2');
INSERT INTO evaluations VALUES(8001,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(8001,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8001,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8001,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(8001,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(8001,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(8001,36,'cp 21','f1e1 d7d6 d2d4 c5d4 f3d4 c6d4 d1d4 e8g8 d4d3 c8d7 c4b3 a8c8 c1g5 b7b5 a2a3 c8b8 a1d1 d7c6 c3d5 c6d5 e4d5 f8e8 c2c3 f6d7 d3d2 g7f6 g5e3 d8c7 e1e2 a7a5 d1e1 c7b7 e3d4 b8c8 h2h3 f6d4 c3d4 d7f6 d2a5 f6d5');
INSERT INTO evaluations VALUES(14426,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14426,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14426,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14426,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14426,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14426,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14426,36,'cp -3','d5c3 b2c3 e7e5 d4d5 c6a5 f1d3 f8d6 c1e3 e8g8 g1f3 b7b6 e1g1 a5b7 f3d2 c8d7 a2a4 d6c5 d3a6 c5e3 a6b7 a8b8 f2e3 b8b7 c3c4 d8e7 d1b3 a7a5 h2h3 f7f6 g1h1 b7b8 f1f3 g8h8 b3d3 e7d6');
INSERT INTO evaluations VALUES(14455,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14455,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14455,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(14455,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14455,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14455,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14455,36,'cp 0','b2c3 e7e5 d4d5 c6a5 c1e3 f8d6 g1f3 e8g8 f1d3 d8e7 f3d2 c8d7 a2a4 a8c8 e1g1 b7b6 d3a6 c8d8 a6d3');
INSERT INTO evaluations VALUES(14454,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14454,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14454,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14454,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14454,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14454,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14454,36,'cp 0','e7e5 d4d5 c6a5 c1e3 f8d6 g1f3 e8g8 f1d3 d8e7 f3d2 c8d7 e1g1 b7b6 d1e2 a5b7 a2a4 a7a5 d3b5 a8e8 f1d1 b7c5 e3c5 d6c5 b5d7 e7d7 d2b3 c5d6 b3d2');
INSERT INTO evaluations VALUES(15395,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15395,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15395,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15395,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15395,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15395,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15395,36,'cp 0','d4d5 c6a5 c1e3 f8d6 g1f3 b7b6 f1d3 d8e7 e1g1 e8g8 f3d2 c8d7 d1e2 c7c6 c3c4 a5b7 a2a4 a8c8 h2h3 c8c7 f1d1 f8c8 d1b1 d6b4 d2f1 b4d6');
INSERT INTO evaluations VALUES(14751,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14751,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(14751,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(14751,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(14751,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(14751,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(14751,36,'cp 60','e6d5 c1g5 f8e7 e2e3 f6h5 g5e7 c6e7 f1e2 c7c6 b2b4 e8g8 e1g1 h5f6 a2a4 c8g4 h2h3 g4f3 e2f3 a7a6 b4b5 a6b5 a4b5 d8d6 b5c6 a8a1 d1a1 b7c6 c3e2 g7g6 e2f4 f6e4 a1a5 e4g5 f3e2 f8e8 f4d3 g5e4');
INSERT INTO evaluations VALUES(20628,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20628,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20628,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20628,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(20628,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(20628,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(20628,36,'cp 14','d2d4 d6d5 f1b5 g8f6 e1g1 f8d6 b1c3 e8g8 c1g5 c8e6 d4c5 d6c5 g5h4 a7a6 b5c6 b7c6 d1d3 a6a5 h2h3 f8e8 a1e1 a5a4 f3d4 e6d7 e1e8 d7e8 f1d1 e8d7 d4f3 a4a3 b2b3 d7e6 f3d4 e6d7 d1e1 c5f8 d4f3 h7h6 f3e5');
INSERT INTO evaluations VALUES(21672,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(21672,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(21672,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(21672,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(21672,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(21672,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(21672,36,'cp 87','f3e5 c6e5 d4e5 d5d4 g2g3 g8e7 f1g2 e7c6 e1g1 c6e5 d1a4 c7c6 f1d1 c8e6 c1e3 f8c5 b2b4 d8e7 e3f4 c5d6 b1d2 e8g8 c4c5 d6c7 d2b3 g7g5 f4e5 c7e5 b3d4 e6g4 d1d2 e5d4 d2d4 e7e2 d4e4 e2b2 a1e1 g4e6 b4b5 c6b5 a4b4 b2b4 e4b4 a8c8 e1c1 b7b6 c5c6 h7h6 c6c7 e6a2');
INSERT INTO evaluations VALUES(21867,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21867,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21867,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21867,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21867,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21867,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21867,36,'cp 0','g8f6 b1c3 f8e7 c1d2 e8g8 e1c1 f8e8 g1f3 a7a6 b5c6 b7c6 h1e1 c6c5 e3f4 f6h5 f4e3');
INSERT INTO evaluations VALUES(22629,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(22629,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(22629,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(22629,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(22629,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(22629,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(22629,36,'cp -20','c2c3 d7d5 e4d5 e6d5 d2d4 c8e6 g1h3 f8d6 h3f4 g8f6 h2h3 e8g8 f1d3 d8d7 e1g1 a8e8 b1d2 c6e7 f4e6 d7e6 f3d1 e7g6 d2f3 f6e4 f3g5 e6d7 f1e1 e4f6 e1e8 f8e8 g5f3 c7c6 d1c2 g6f4 d3f1 f6e4 c1f4 d6f4 g2g3 f4d6 g1g2 g7g6 f1d3 h7h5 a1e1');
INSERT INTO evaluations VALUES(22624,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(22624,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(22624,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(22624,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(22624,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(22624,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(22624,36,'cp -25','d7d5 e4d5 e6d5 d2d4 c8e6 g1h3 g8f6 h3f4 f8d6 h2h3 e8g8 f1d3 d8d7 e1g1 a8e8 b1a3 a7a6 a3c2 c6e7 f4e6 f7e6 c2e3 c7c5 e3g4 c5d4 c3d4 f6e4 f3d1 d6b8 g4e5 d7d6 c1e3 e7g6 f2f4 g6h4 d1e1 d6e7 g2g3 h4f5 e3f2 g7g5 d3e4 d5e4 e1e4');
INSERT INTO evaluations VALUES(42753,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(42753,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(42753,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(42753,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(42753,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42753,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42753,36,'cp 1','f6e4 c3e4 d7d5 c2c3 b4a5 d1a4 c7c6 c4b3 d5e4 a4e4 e8g8 e1g1 b8d7 d2d4 d7f6 e4h4 f8e8 d4e5 e8e5 e2g3 f6d5 h4d8 a5d8 c1d2 h7h6 c3c4 d5f6 d2c3 e5e8 b3c2 c8e6 b2b3 b7b5 g3f5 d8b6 c3f6 e6f5 c2f5 g7f6 c4b5 c6b5 f5d7 e8e5 g1h1 a8d8');
INSERT INTO evaluations VALUES(47455,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(47455,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(47455,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(47455,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(47455,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(47455,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(47455,36,'cp 5','e7e5 f1g2 d5b6 d2d3 b8c6 g1f3 f8e7 c1e3 e8g8 e1g1 c8e6 d3d4 e5d4 f3d4 c6d4 e3d4 c7c6 c3e4 e6c4 g2f3 f8e8 b2b3 c4d5 e2e3 d8d7 e4c3 d5f3 d1f3 d7e6 c3e2 b6d5 a1d1 e7f8 d4b2 a8d8 e2d4 e6g6 d4f5');
INSERT INTO evaluations VALUES(48200,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48200,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48200,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(48200,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(48200,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(48200,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48200,36,'cp 0','c7c5 g5f6 d8f6 c4d5 c5d4 f1b5 c8d7 b5d7 b8d7 d1d4 f6d4 e3d4 e6d5 b1c3 d7f6 e1e2 f8b4 a1c1 e8g8 e2d3 a8e8 g1h3 b4d6 g2g3 g7g6 h3g5 g8g7 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(49437,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49437,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49437,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49437,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49437,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(49437,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(49437,36,'cp 0','d7d5 e2e3 f8e7 b2b3 e8g8 c1b2 b7b6 c2c4 c7c5 c4d5 f6d5 d4c5 e7c5 b1d2 b8d7 a1c1 c8b7 f1d3 a7a5 e1g1 d8e7 c1a1 e7d8 d3b5 d5c7 b5d3');
INSERT INTO evaluations VALUES(52282,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(52282,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52282,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(52282,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(52282,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52282,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52282,36,'cp 10','b8c6 f1g2 f8g7 e2e3 e7e6 g1e2 g8e7 d2d4 c5d4 e3d4 d7d5 c4d5 e7d5 e1g1 e8g8 f1e1 d8b6 h2h4 c6d4 c3d5 d4e2 e1e2 e6d5 d1d5 c8e6 d5b7 b6b7 g2b7 a8b8 b7f3 h7h6 a2a4 a7a5 g1g2 f8c8 c1f4 b8b2 e2b2 g7b2 a1b1 b2g7 b1b8 c8b8');
INSERT INTO evaluations VALUES(18392,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(18392,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(18392,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(18392,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(18392,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(18392,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(18392,36,'cp 144','f8b4 c1d2 b4d2 b1d2 d7d6 d1h5 d8e7 a2a4 g8f6 h5h4 e8g8 f2f4 a7a5 e1g1 b5a4 b3a4 c6c5 f4f5 h7h6 g2g4 d6d5 f1e1 d5e4 d2e4 c8b7 e4f6 e7f6 h4f6 g7f6 e1e7 b7d5 g1f2 f8d8 b2b3 h6h5 h2h3 g8g7 a1e1 h5g4 h3g4 d8h8 f2g3 a8b8 e7d7 b8d8 d7a7');
INSERT INTO evaluations VALUES(24902,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(24902,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(24902,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24902,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24902,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(24902,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(24902,36,'cp 35','e7e6 e1g1 f8d6 f4g3 e8g8 c2c4 c6e7 f1c1 c7c6 g3d6 d8d6 c4c5 d6c7 b2b4 e7g6 b1d2 e6e5 d4e5 g6e5 f3e5 c7e5 b4b5 a8c8 d2f3 e5e7 f3d4 c6b5 d3b5 c8c7 c1c2 f8c8 a1c1 a7a6 b5b3 g7g6 c5c6 b7c6 c2c6 c7c6 d4c6 e7d6');
INSERT INTO evaluations VALUES(34592,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(34592,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(34592,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(34592,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(34592,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(34592,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(34592,36,'cp 66','e4c3 d5a5 g1f3 g8f6 d2d4 c8f5 c1d2 e7e6 f1c4 f8b4 a2a3 b4c3 d2c3 a5b6 d1e2 e8g8 e1c1 f6d5 c4d5 e6d5 f3e5 b8d7 g2g4 d7e5 d4e5 f5e6 f2f4 c7c5 f4f5 d5d4 f5e6 d4c3 e6f7 f8f7 b2c3 a8e8 e2c4 e8e5 d1d7 e5e6 h1d1 f7d7 d1d7');
INSERT INTO evaluations VALUES(38718,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38718,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38718,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(38718,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38718,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(38718,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(38718,36,'cp 32','e2e3 c7c6 d2d4 c5b6 b1c3 c8e6 b2b3 e5d4 e3d4 g8f6 g1e2 e8g8 e2f4 b6a5 d1d2 e6f5 e1g1 f8e8 c1a3 b8a6 d4d5 f5e4 d2b2 a5c3 b2c3 c6d5 f4d5 f6d5 c4d5 a8c8 c3d4 e4g2 g1g2 a6c7 a1e1 e8e1 f1e1 c7b5 d4b4 b5a3 b4a3 a7a5 a3b2 d8d7');
INSERT INTO evaluations VALUES(38717,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38717,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(38717,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(38717,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38717,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(38717,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38717,36,'cp 22','c7c6 d2d4 c5b6 b1c3 c8e6 b2b3 e5d4 e3d4 g8f6 g1e2 e8g8 e1g1 d6d5 e2f4 e6g4 d1c2 d5c4 h2h3 c4b3 a2b3 g4d7 d4d5 c6d5 f4d5 f6d5 c3d5 b6d4 c1a3 d4a1 a3f8 b8c6 f8c5 a1d4 f1d1 d4c5 c2c5 d7e6 b3b4 d8f8 c5b5 e6d5 g2d5 f8e7 d5c6 b7c6 b5c6');
INSERT INTO evaluations VALUES(39141,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(39141,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(39141,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(39141,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(39141,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(39141,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(39141,36,'cp 119','d2d4 c5b6 c1e3 b8c6 d1d2 d7d5 g1f3 g8e7 a2a3 e6e5 e4d5 e5d4 d5c6 d4e3 d2d8 e8d8 e1c1 d8e8 c6b7 c8b7 f2e3 b7f3 g2f3 b6e3 c1b1 c7c6 c3e4 e3f4 f1c4 a8d8 d1d8 e8d8 c4f7 h8f8 f7c4 d8c7 e4c5 c7d6 c5e6 f8f7 e6g5 f7f5 g5h7 f4e3 h2h4 e7d5 c4d3 f5f3');
INSERT INTO evaluations VALUES(39138,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(39138,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(39138,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(39138,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(39138,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(39138,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(39138,36,'cp 85','c5b6 c1e3 b8c6 a2a3 d7d5 g1f3 g8e7 d1d3 e6e5 e4d5 e5d4 d5c6 d4e3 d3d8 e8d8 e1c1 d8e8 c6b7 c8b7 f2e3 b7f3 g2f3 b6e3 c1b1 c7c6 c3e4 e3f4 f1c4 a8d8 d1d8 e8d8 c4f7 e7f5 f7e6 f5e3 h1g1 g7g6 h2h4 h8e8 e6b3 e8e7 c2c3 e3f5 e4g5 f4e3 g1d1 d8c7 g5e6 c7b6');
INSERT INTO evaluations VALUES(45161,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(45161,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(45161,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(45161,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(45161,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(45161,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(45161,36,'cp -29','c8g4 f1e2 g7g6 e1g1 b8c6 b1c3 f8g7 c1e3 e8g8 d1d2 a7a6 h2h3 g4f3 e2f3 e7e6 a2a4 a8b8 g1h1 g8h8 a1e1 f6d7 f3d1 d8e7 b2b3 b7b6 c3e2 b6b5 c4b5 a6b5 a4b5 b8b5 d3d4 c5d4 e2d4 c6d4 e3d4 g7d4 d2d4 h8g8 b3b4');
INSERT INTO evaluations VALUES(50440,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50440,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(50440,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(50440,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(50440,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(50440,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(50440,36,'cp -5','g7g5 h4g3 h7h5 e2e3 d7d5 f1d3 h5h4 g3e5 f7f6 f2f3 e4d6 e5d6 d8d6 b1c3 b8d7 a2a4 a7a5 g1e2 d7b6 e1g1 e7e5 d3g6 e8d8 e3e4 c8d7 d1d3 c7c6 g1h1 h4h3 g2g4 d8c7 e4d5 b6d5 c3e4 d6e7 a1b1 c7b8 d4e5 e7e5 f1d1');
INSERT INTO evaluations VALUES(50280,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(50280,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(50280,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(50280,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(50280,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(50280,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(50280,36,'cp -10','h4g3 h7h5 e2e3 d7d5 f1d3 h5h4 g3e5 f7f6 f2f3 e4d6 e5d6 d8d6 b1c3 c7c6 d1d2 b8d7 g1e2 d7b6 e3e4 c8d7 e4d5 c6d5 a2a4 e8c8 c3b5 d7b5 a4b5 c8b8 d2a5 b6c8 c2c3 e7e6 e1d1 d6d7 h2h3 f8d6 d1c2 d8e8 a5a4 d7c7 c2b1 f6f5 h1e1 e6e5 d3f5');
INSERT INTO evaluations VALUES(50281,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(50281,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(50281,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(50281,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(50281,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(50281,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(50281,36,'cp -9','h7h5 e2e3 d7d5 f1d3 h5h4 g3e5 f7f6 f2f3 e4d6 e5d6 d8d6 b1c3 b8c6 d1d2 e7e6 e1c1 c6e7 f3f4 g5f4 e3f4 c8d7 g1f3 e8c8 g2g4 h4g3 h2g3 h8g8 h1g1 c7c5 d4c5 d6c5 g3g4 c8b8 c1b1 f8g7 g4g5 f6g5 g1g5 g7f6 g5g8 d8g8 f3e5 f6e5 f4e5 e7c6 d2f4');
INSERT INTO evaluations VALUES(54881,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(54881,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(54881,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(54881,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(54881,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(54881,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(54881,36,'cp 76','c7c5 g1f3 c5d4 d1b3 e7e6 e1g1 g8f6 f3d4 a7a6 f1d1 d8c7 b1c3 f8g7 e3e4 e8g8 c4e2 e6e5 d4f3 b8c6 c1e3 c8e6 c3d5 a8c8 b3a4 c7b8 d5b6 c8d8 h2h3 d8d1 a1d1 f8d8 d1d8 b8d8 e2c4 e6c4 a4c4 d8e7 b6d5 e7d8 e3g5 h7h6');
INSERT INTO evaluations VALUES(56126,30,'cp -455',NULL);
INSERT INTO evaluations VALUES(56126,31,'cp -452',NULL);
INSERT INTO evaluations VALUES(56126,32,'cp -444',NULL);
INSERT INTO evaluations VALUES(56126,33,'cp -451',NULL);
INSERT INTO evaluations VALUES(56126,34,'cp -445',NULL);
INSERT INTO evaluations VALUES(56126,35,'cp -439',NULL);
INSERT INTO evaluations VALUES(56126,36,'cp -455','c7c6 d5e3 d4e3 c1e3 g8f6 f1e2 c6c5 g1f3 b8c6 e1g1 f8e7 c2c3 e8g8 a2a3 a7a5 b2b4 a5b4 a3b4 a8a1 d1a1 c5b4 c3b4 c6b4 f1c1 f6g4 e3a7 b4c6 h2h3 c6a7 a1a7 g4f6 f3e5 e7d6 e5f3 d6c7 c1b1 g8h8 b1c1 h7h6 a7c5 c7b6 c5c8');
INSERT INTO evaluations VALUES(53683,30,'cp -443',NULL);
INSERT INTO evaluations VALUES(53683,31,'cp -453',NULL);
INSERT INTO evaluations VALUES(53683,32,'cp -448',NULL);
INSERT INTO evaluations VALUES(53683,33,'cp -448',NULL);
INSERT INTO evaluations VALUES(53683,34,'cp -455',NULL);
INSERT INTO evaluations VALUES(53683,35,'cp -450',NULL);
INSERT INTO evaluations VALUES(53683,36,'cp -452','d5e3 d4e3 c1e3 g8f6 f1e2 f8d6 g1f3 e8g8 c2c3 c6c5 a2a3 b8c6 b2b4 b7b6 h2h3 c6e7 b4c5 b6c5 e1g1 e7g6 d1c2 g8h8 f1b1 d8c7 b1b2 a8b8 b2b8 c7b8 a1b1 b8c7 f3d2 c8e6 c2a4 h7h6 a4a6 e6c8 a6a4 d6e7');
INSERT INTO evaluations VALUES(57610,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(57610,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(57610,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(57610,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(57610,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(57610,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(57610,36,'cp 72','g1f3 c7c6 d2d4 g7g6 c2c4 f8g7 b1c3 g8f6 f1g2 d7d6 e1g1 e8g8 d1c2 b8a6 a1b1 a6b4 c2d1 a7a5 a2a3 b4a6 b2b4 a5b4 a3b4 a6c7 b4b5 h7h6 b5c6 b7c6 e2e3 c8d7 c1b2 d8e8 c3e2 e8f7 f3d2 f8c8 d1b3 c7e6');
INSERT INTO evaluations VALUES(1345,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(1345,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(1345,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(1345,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(1345,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(1345,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(1345,36,'cp 25','d7c6 b1d2 g8f6 f1d3 e7e5 e1g1 b6c7 a2a4 f8c5 d2b3 c5d6 a4a5 c8e6 d1e2 e8g8 c1g5 d6e7 g5f6 e7f6 b3c5 a8e8 c5e6 e8e6 a5a6 b7b6 f1d1 e6d6 g2g3 g7g6 c2c3 h7h5 g1g2 f8d8 d3c4 d6d1 a1d1 d8d1 e2d1');
INSERT INTO evaluations VALUES(1387,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(1387,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(1387,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(1387,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(1387,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(1387,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(1387,36,'cp 26','f1d3 e7e5 e1g1 g8f6 b1d2 b6c7 a2a4 f8c5 d2b3 c5e7 c1e3 e8g8 d1e2 f6g4 e3d2 g4f6 h2h3 c8e6 d2e3 b7b6 b3d2 f6d7 d3c4 e6c4 d2c4 f8d8 c2c3 b6b5 a4b5 c6b5 c4a5 a7a6 c3c4 b5b4 f1d1 h7h6 d1d3 d7c5');
INSERT INTO evaluations VALUES(1463,30,'cp -166',NULL);
INSERT INTO evaluations VALUES(1463,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(1463,32,'cp -175',NULL);
INSERT INTO evaluations VALUES(1463,33,'cp -166',NULL);
INSERT INTO evaluations VALUES(1463,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(1463,35,'cp -180',NULL);
INSERT INTO evaluations VALUES(1463,36,'cp -176','c5d4 e3d4 c6d4 d1d4 f8c5 d4h4 c8d7 f1d3 g7g5 h4g3 h7h5 h2h4 g5h4 g3h4 g8h6 b3b4 c5e7 h4d4 h6g4 b1d2 a8c8 g1f3 b6d4 c3d4 c8c3 d3e2 e7b4 e1g1 b4e7 a2a4 e8d8 a4a5 d8c7 a5a6 b7b6 d2b1 c3b3 b1d2 b3a3 a1a3 e7a3 f3g5 g4h6 d2f3 h8c8 g5h7 h6f5 h7f6');
INSERT INTO evaluations VALUES(1517,30,'cp -174',NULL);
INSERT INTO evaluations VALUES(1517,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(1517,32,'cp -181',NULL);
INSERT INTO evaluations VALUES(1517,33,'cp -166',NULL);
INSERT INTO evaluations VALUES(1517,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(1517,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(1517,36,'cp -157','c3d4 f8b4 b1d2 b4c3 d1g4 g8e7 a1c1 c3b2 c1b1 b2d4 e3d4 b6d4 g4d4 c6d4 f2f4 c8d7 g1f3 e7c6 f3d4 c6d4 b1c1 e8e7 e1f2 d4c6 d2f3 h8c8 f1d3 c6b4 d3b1 h7h6 a2a3 c8c1 h1c1 b4c6 b3b4 a7a6 b1d3 f7f6 g2g3 a8c8 b4b5 a6b5');
INSERT INTO evaluations VALUES(1503,30,'cp -185',NULL);
INSERT INTO evaluations VALUES(1503,31,'cp -160',NULL);
INSERT INTO evaluations VALUES(1503,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(1503,33,'cp -165',NULL);
INSERT INTO evaluations VALUES(1503,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(1503,35,'cp -163',NULL);
INSERT INTO evaluations VALUES(1503,36,'cp -175','f8b4 b1d2 b4c3 d1g4 g8e7 a1c1 c3b2 c1b1 b2d4 e3d4 b6d4 g4d4 c6d4 f2f4 c8d7 g1f3 e7c6 f3d4 c6d4 d2f3 d4f3 g2f3 e8g8 b1c1 d7c6 e1f2 f7f6 f1h3 g8f7 h1e1 g7g6 e5f6 f8e8 b3b4 f7f6 b4b5 c6b5 c1c7 b5c6 c7h7');
INSERT INTO evaluations VALUES(135,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(135,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(135,32,'cp -158',NULL);
INSERT INTO evaluations VALUES(135,33,'cp -185',NULL);
INSERT INTO evaluations VALUES(135,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(135,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(135,36,'cp -165','b1d2 b4c3 d1g4 g8e7 a1c1 c3b2 c1b1 b2d4 e3d4 b6d4 g4d4 c6d4 f2f4 c8d7 g1f3 e7c6 b1c1 f7f6 f1d3 f6e5 f4e5 d4f3 d2f3 e8e7 h2h4 a8f8 h1h3 f8f4 h3g3 g7g6 g3g5 e7d8 e1e2 d7e8 e2e3 h8f8 g5g3 e8d7 a2a3 c6e7 b3b4 a7a5 b4a5 f4a4 c1b1 d8c8 a5a6');
INSERT INTO evaluations VALUES(134,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(134,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(134,32,'cp -174',NULL);
INSERT INTO evaluations VALUES(134,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(134,34,'cp -158',NULL);
INSERT INTO evaluations VALUES(134,35,'cp -172',NULL);
INSERT INTO evaluations VALUES(134,36,'cp -175','b4c3 d1g4 g8e7 a1c1 c3b2 c1b1 b2d4 e3d4 b6d4 g4d4 c6d4 f2f4 f7f6 g1f3 f6e5 f4e5 e7c6 b1c1 d4f3 d2f3 c8d7 f1d3 e8e7 h2h4 a8f8 h1h3 f8f4 h3g3 g7g6 g3g5 e7d8 e1e2 h8f8 h4h5 g6h5 g5h5 f4g4 e2f2 d7e8 h5h7 c6e5 d3c2 e8d7 c1h1 g4f4 h7h8 d8e7 h1h7 e7d6 h8f8 f4f8');
INSERT INTO evaluations VALUES(3499,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(3499,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(3499,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(3499,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(3499,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(3499,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(3499,36,'cp 46','d4d5 c6b4 d3e2 e7e6 c1e3 d8e7 e1g1 e6d5 e4d5 f6e4 c3e4 e7e4 d1d2 e4c2 d2b4 c2e2 e3d4 g7d4 f3d4 e2e5 a1d1 a7a5 b4c3 e5f6 d4b5 f6c3 b5c3 f8e8 f2f3 c8d7 g1f2 f7f5 g2g3 e8e5 f1e1 a8e8 e1e5 e8e5 b2b3 g8f7 d1b1 e5e7 b1c1 h7h6 c1d1');
INSERT INTO evaluations VALUES(5229,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(5229,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(5229,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(5229,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(5229,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(5229,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(5229,36,'cp 2','d2d4 c5d4 f3d4 d8c7 e1g1 g8f6 f1e1 d7d6 d4c6 b7c6 c1f4 e6e5 f4d2 e8g8 d1e2 c8e6 a1d1 h7h6 c4b3 e6g4 f2f3 g4d7 e2d3 a8b8 g1h1 d7e6 d3e2 e6d7 d2c1 f6h5 c1e3 b8b7 e2d2 d7e6 b3e6 f7e6 b2b3 d6d5 c3a4 h5f4');
INSERT INTO evaluations VALUES(5236,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5236,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(5236,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(5236,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5236,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(5236,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5236,36,'cp 0','c5d4 f3d4 d8c7 e1g1 g8f6 f1e1 e8g8 d4c6 b7c6 e4e5 f6d5 d1h5 a8b8 c1d2 b8b2 c4b3 f7f6 c3a4 b2b3 a2b3 f6e5 c2c4 f8f5 h5e8 f5f8');
INSERT INTO evaluations VALUES(5241,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5241,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(5241,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(5241,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5241,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(5241,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(5241,36,'cp 7','f3d4 d8c7 e1g1 g8f6 c4b3 a7a6 f1e1 d7d6 d4c6 c7c6 a2a4 e8g8 c1f4 b7b6 g1h1 c8b7 f2f3 f8d8 d1e2 f6d7 a1d1 d7c5 d1d2 a8c8 e1d1 h7h6 h2h3 b6b5 a4b5 a6b5 e2b5 c6b5 c3b5 d6d5 b3a2 d5e4 d2d8 c8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(5240,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(5240,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(5240,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(5240,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(5240,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5240,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5240,36,'cp 13','d8c7 e1g1 g8f6 c4b3 a7a6 f1e1 d7d6 d4c6 c7c6 d1d3 e8g8 d3g3 f6h5 g3e3 c8d7 c1d2 c6c7 c3e2 h5f6 e2d4 d7c6 d4c6 b7c6 e3g3 a6a5 d2c3 a5a4 b3c4 f6h5 g3f3 h5f6 e4e5 f6d7 e5d6 e7d6');
INSERT INTO evaluations VALUES(10487,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(10487,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(10487,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(10487,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(10487,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(10487,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(10487,36,'cp -45','e8g8 e1c1 f8e8 e3f4 b4c3 d2c3 e8e4 f4d2 e4e6 f1d3 d7d5 g1f3 d8d6 d3f5 e6e7 c3f6 d6f6 f5c8 a8c8 d2d5 c8e8 d1d2 b7b6 h1d1 h7h6 a2a3 g7g6 c1b1 g8g7 d5b5 a7a5 h2h3 h6h5 b5b3 e7e2 b3b5 e2e6 b5g5 f6g5 f3g5 e6e1 g5f3 e1d1 d2d1 e8e2');
INSERT INTO evaluations VALUES(4470,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(4470,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(4470,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(4470,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(4470,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(4470,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(4470,36,'cp -45','e1c1 f8e8 e3f4 c6e5 c3d5 b4d2 d1d2 f6d5 e4d5 e5g6 f4d4 e8e1 d2d1 d8g5 d4d2 e1d1 c1d1 g5f6 d1c1 d7d6 g2g3 c8f5 f2f4 f5e4 f1g2 e4g2 d2g2 f6d4 g1e2 d4e3 c1b1 a8e8 e2c3 g6e7 c3d1 e3d4 d1c3 g8f8 h1d1 d4c5 g2e4 e7d5 e4d5 c5d5 c3d5');
INSERT INTO evaluations VALUES(4469,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(4469,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(4469,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(4469,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(4469,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(4469,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(4469,36,'cp -45','f8e8 e3f4 b4c3 d2c3 e8e4 f4d2 e4e8 d2f4 d7d6 c3f6 d8f6 f4f6 g7f6 g1e2 e8e5 e2c3 c8g4 f2f4 e5c5 d1d5 g8g7 d5c5 d6c5 f1d3 c6d4 c3e4 a8e8 e4c5 d4e2 d3e2 e8e2 c5b7 e2g2 b7c5 g4f5 c5d3 g2e2 a2a3');
INSERT INTO evaluations VALUES(23549,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(23549,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(23549,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(23549,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(23549,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(23549,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(23549,36,'cp -48','e3f4 b4c3 d2c3 e8e4 f4d2 d7d5 f2f3 e4e8 g1e2 c6e5 e2f4 c7c6 f1e2 e5g6 f4g6 f7g6 h1e1 d8d6 e2d3 c8f5 d3f5 g6f5 e1e8 a8e8 d2g5 d6e6 d1e1 e6d7 e1e8 f6e8 g5f4 b7b6 h2h4 c6c5 c3e5 d7e6 h4h5 d5d4 b2b3 h7h6 c1b2 e8f6 e5f6');
INSERT INTO evaluations VALUES(15284,30,'cp -598',NULL);
INSERT INTO evaluations VALUES(15284,31,'cp -597',NULL);
INSERT INTO evaluations VALUES(15284,32,'cp -605',NULL);
INSERT INTO evaluations VALUES(15284,33,'cp -582',NULL);
INSERT INTO evaluations VALUES(15284,34,'cp -587',NULL);
INSERT INTO evaluations VALUES(15284,35,'cp -600',NULL);
INSERT INTO evaluations VALUES(15284,36,'cp -592','d8g5 d5c6 g5g2 h1f1 c8h3 d1e2 g2c6 f1g1 c6c2 b1a3 c2f5 b2b3 e8c8 a3c4 f7f6 g1g3 c8b8 c1b2 f5e6 c4e3 h7h5 e1c1 h5h4 g3g1 b7b6 e3c2 e6c6 c1b1 h3f5 d2d3 c6e8 e2f3 f5c8 d3d4 c8b7 f3h3 g7g6 d4e5 d8d1 g1d1 f6e5 h3c3 f8d6');
INSERT INTO evaluations VALUES(968,30,'cp -594',NULL);
INSERT INTO evaluations VALUES(968,31,'cp -569',NULL);
INSERT INTO evaluations VALUES(968,32,'cp -569',NULL);
INSERT INTO evaluations VALUES(968,33,'cp -582',NULL);
INSERT INTO evaluations VALUES(968,34,'cp -578',NULL);
INSERT INTO evaluations VALUES(968,35,'cp -595',NULL);
INSERT INTO evaluations VALUES(968,36,'cp -594','d5c6 g5g2 h1f1 c8h3 d1e2 g2c6 f1g1 c6c2 b1a3 c2f5 g1g3 e8c8 a3c4 e5e4 b2b3 h7h5 c1b2 f7f6 b2c3 a7a6 e1c1 h3g4 c4e3 f8a3 c1c2 g4e2 e3f5 h5h4 g3e3 e2d1 c2d1 g7g6 f5h4 h8h4 c3f6 h4f4 f6d8 c8d8 h2h3 a3b4 e3e2 e4e3 e2e3 f4f2 d1e1 f2d2');
INSERT INTO evaluations VALUES(15508,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15508,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(15508,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(15508,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(15508,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(15508,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(15508,36,'cp 1','f8e7 d2d3 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 e1g1 d4e6 c3e2 e8g8 c4a2 d8c7 f2f4 e5f4 e2f4 d6d5 e4d5 c7b6 g1h1 e6f4 c1f4 f6d5 a2d5 c6d5 f4e5 f7f6 e5h2 a8c8 a1b1 e7c5 d1f3 b6c6 c2c3 f8e8 h2g1 c5g1 h1g1 e8e5 b1e1 c8e8');
INSERT INTO evaluations VALUES(8604,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(8604,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(8604,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(8604,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8604,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(8604,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(8604,36,'cp 19','d2d3 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 a3a4 a7a5 e1g1 e8g8 c1e3 d4e6 c4b3 g8h8 g2g3 d8d7 d1f3 a8e8 g1g2 f6g8 b3e6 f7e6 f3d1 d6d5 c3b1 f8f7 e3b6 e7d8 b6d8 e8d8 b1d2');
INSERT INTO evaluations VALUES(8605,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(8605,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(8605,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(8605,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(8605,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(8605,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(8605,36,'cp 20','c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 c1e3 e8g8 e1g1 d4e6 d1f3 d8c7 c4b3 b7b5 c3e2 a7a5 a3a4 b5b4 e2g3 g7g6 b3e6 f7e6 f3e2 d6d5 c2c4 a8e8 e2c2 d5d4 e3h6 f8f7 f2f4 e5f4 h6f4 e6e5 f4g5');
INSERT INTO evaluations VALUES(18416,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18416,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(18416,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18416,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(18416,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18416,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18416,36,'cp 0','g8f6 e4e5 f6d5 b1c3 f8g7 c3d5 c6d5 d4d5 a8b8 f1c4 e8g8 e1g1 c8b7 d5d4 d8c8 c4b3 d7d6 d4a7 b7g2 g1g2 c8g4 g2h1 g4f3 h1g1 f3g4');
INSERT INTO evaluations VALUES(12005,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12005,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12005,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12005,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12005,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12005,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12005,36,'cp 0','e4e5 f6d5 b1c3 f8g7 c3d5 c6d5 d4d5 a8b8 f1c4 e7e6 d5c5 c8b7 e1g1 b8c8 c5b4 d8c7 c1g5 g7f8 b4b3 b7g2 g5f6 g2f1 c4f1 h8g8 a1d1 f8e7 d1d7 c7d7 f1b5 c8c6 b5c6 d7c6 b3b8 e8d7 b8g8 c6c2 g8f7 c2d1 g1g2 d1g4 g2h1 g4d1');
INSERT INTO evaluations VALUES(12002,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12002,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12002,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12002,36,'cp 0','f6d5 b1c3 f8g7 c3d5 c6d5 d4d5 a8b8 f1c4 e7e6 d5c5 c8b7 e1g1 b8c8 c5b4 d8c7 c1g5 g7f8 b4b3 b7g2 g5f6 g2f1 c4f1 h8g8 a1d1 f8e7 d1d7 c7d7 f1b5 c8c6 b5c6 d7c6 b3b8 e8d7 b8g8 c6c2 g8f7 c2d1 g1g2 d1g4 g2h1 g4d1');
INSERT INTO evaluations VALUES(12007,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12007,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12007,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12007,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12007,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12007,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12007,36,'cp 0','b1c3 f8g7 c3d5 c6d5 d4d5 a8b8 f1c4 e7e6 d5c5 c8b7 e1g1 b8c8 c5b4 d8c7 c1g5 g7f8 b4b3 b7g2 g5f6 g2f1 c4f1 h8g8 a1d1 f8e7 d1d7 c7d7 f1b5 c8c6 b5c6 d7c6 b3b8 e8d7 b8g8 c6c2 g8f7 c2d1 g1g2 d1g4 g2h1 g4d1');
INSERT INTO evaluations VALUES(21979,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(21979,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(21979,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(21979,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(21979,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(21979,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(21979,36,'cp 60','d2d4 g8f6 d4d5 c6b8 f3d2 f8e7 a2a4 e8g8 a4a5 b8a6 e1g1 h7h6 b1c3 c7c6 d2c4 c6d5 e4d5 c8f5 f2f4 e5f4 f1f4 f5g6 c1e3 f8e8 d1d2 a6b4 a1c1 f6d7 f4f1 e7f8 h2h3 d7f6 g2g4 g8h8 e3d4 b4d5 c3d5 f6d5');
INSERT INTO evaluations VALUES(22364,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(22364,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(22364,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(22364,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(22364,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(22364,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(22364,36,'cp 84','d2d4 e5d4 f3d4 g8f6 b1c3 d7b6 c4d3 f8e7 a2a4 a7a5 e1g1 e8g8 h2h3 f6d7 f1e1 d7c5 b2b3 e7f6 d3f1 f8e8 c1f4 c8d7 d1d2 d7c6 f2f3 c5e6 d4e6 f7e6 f4e3 g8h8 g2g3 f6c3 d2c3 e6e5 e3b6 c7b6 a1d1 d8f6 c3e3 e8f8');
INSERT INTO evaluations VALUES(23266,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(23266,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(23266,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(23266,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(23266,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(23266,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23266,36,'cp 47','f5e4 c3e4 g8f6 e4f6 d8f6 d1e2 f8e7 b5c6 b7c6 f3e5 c8b7 d2d4 e8c8 c2c3 c6c5 e1g1 e7d6 f2f4 c5d4 c3d4 f6e6 c1e3 e6d5 f1f2 h8e8 e2c2 g7g6 a1c1 c8b8 c2c4 d5c4 e5c4 b7d5 c4d6 c7d6 e3d2 d8c8 c1c8 e8c8 f4f5 g6f5 f2f5 d5a2');
INSERT INTO evaluations VALUES(23262,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(23262,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(23262,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(23262,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(23262,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(23262,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(23262,36,'cp 36','c3e4 g8f6 e4f6 d8f6 d1e2 f8e7 b5c6 b7c6 f3e5 c8b7 d2d4 e8c8 c2c3 c6c5 e1g1 e7d6 c1d2 d8e8 a1e1 f6e6 e2c4 c5d4 c4e6 e8e6 e5f7 d4c3 d2c3 h8g8 f7d6 e6d6 f2f4 d6g6 f1f2 b7d5 b2b3 g6g4 h2h3 g4g6 g1h2 g6e6 e1e6 d5e6');
INSERT INTO evaluations VALUES(23260,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(23260,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(23260,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(23260,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(23260,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(23260,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23260,36,'cp 35','g8f6 e4f6 d8f6 d1e2 f8e7 b5c6 b7c6 f3e5 c8b7 e1g1 e8c8 d2d4 c6c5 c1e3 h8e8 c2c3 e7d6 f2f4 c5d4 c3d4 f6e6 a1c1 e6d5 f1f2 c8b8 e2c2 e8e7 c2c4 d8e8 e3d2 b8a8 c4d5 b7d5 e5c4 e7e4 c4d6 c7d6 c1c7 d5c6 b2b3 e4d4 d2c3 d4d1 f2f1 d1f1 g1f1 e8e4');
INSERT INTO evaluations VALUES(17095,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(17095,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(17095,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(17095,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(17095,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(17095,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(17095,36,'cp 49','e4f6 d8f6 d1e2 f8e7 b5c6 b7c6 f3e5 c8b7 e1g1 e8c8 d2d4 c6c5 c1e3 h8e8 c2c3 e7d6 f2f4 c5d4 c3d4 f6e6 e2c2 e6d5 f1f2 e8e7 a1c1 c8b8 c2c4 e7e8 c4c3 b8a8 b2b4 d5b5 a2a3 g7g6 f2b2 e8e7 c1a1 b5d5 a1e1 d5b5 c3c2 b5d5 b4b5 d8e8 a3a4');
INSERT INTO evaluations VALUES(24369,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(24369,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(24369,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(24369,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(24369,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(24369,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(24369,36,'cp 93','d1f3 d8d7 b1c3 d7g4 f3d3 f8e7 f2f4 e8g8 h2h3 g4e6 c1e3 a7a5 b2b3 f8e8 a1e1 a5a4 f4f5 e6e5 e3d4 e5a5 c3a4 f6d7 e1d1 e7f6 c2c4 f6d4 d3d4 h7h6 a4c3 a5b4 d4d3 d7c5 d3e3 c5d7 e3g3 g8h8 g1h2 a8a3 f1e1 b4a5 e1e2');
INSERT INTO evaluations VALUES(25512,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(25512,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(25512,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(25512,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(25512,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(25512,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(25512,36,'cp 42','h7h6 b1d2 a7a6 b5a4 b7b5 a4c2 c6e7 d3d4 e7g6 f1e1 f8e7 d2f1 e8g8 h2h3 c7c5 f1g3 d8c7 g3f5 f8e8 f5e7 e8e7 d4d5 e7e8 f3h2 g8h8 c1e3 e8g8 d1d2 a6a5 f2f3 c5c4 b2b3 g6f4 b3c4 c7c4 h2g4 f6g4 h3g4');
INSERT INTO evaluations VALUES(30872,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(30872,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(30872,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(30872,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(30872,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(30872,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(30872,36,'cp 91','b5d7 d8d7 d4e2 b8c6 b1c3 f7f5 e4f5 g8f6 e1g1 d7f5 c1e3 a8d8 c3d5 f6d5 d1d5 f5c2 e2c3 c2g6 f2f4 f8e7 a1d1 h8f8 b2b4 a7a6 d5b3 d8d7 a2a4 e7d8 h2h3 c6e7 f1e1 g6f7 c3d5 e7d5 d1d5 f7f5 f4e5 d6e5');
INSERT INTO evaluations VALUES(35953,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(35953,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(35953,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(35953,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(35953,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(35953,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(35953,36,'cp 52','d4d5 c8g4 e1g1 e7e6 d3e2 e6d5 e4d5 b8d7 a2a4 g4f3 e2f3 f6e8 c1d2 f7f5 g2g3 d7e5 f3e2 e8f6 f2f4 e5f7 g1g2 f8e8 e2f3 d8d7 h2h3 e8e7 d1c2 a8e8 a1e1 e7e1 f1e1 e8e1 d2e1 a7a6 f3d1');
INSERT INTO evaluations VALUES(44055,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(44055,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(44055,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(44055,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(44055,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(44055,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(44055,36,'cp 97','d7d6 f1d3 d8c7 e4e5 d6e5 d4e5 f6d5 c3e4 b8d7 c2c4 d5b4 c1f4 b4d3 d1d3 c7a5 e4d2 f7f5 e5f6 g7f6 d3e2 e6e5 e1g1 f8g7 d2e4 e8g8 f4d2 a5c7 d2c3 d7b6 a1d1 c8e6 b2b3 e6g4 h2h3 g4f3 g2f3 c7e7 e4g3 f6f5 f3f4 e5e4 c3g7 e7g7 g1h2 g8h8 e2h5 g7f7');
INSERT INTO evaluations VALUES(45890,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(45890,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(45890,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(45890,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(45890,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(45890,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(45890,36,'cp 123','f1d3 d8a5 e4e5 d6e5 d4e5 f6d5 c1d2 d5c3 d2c3 f8b4 c3b4 a5b4 d1d2 b4d2 f3d2 b8d7 f2f4 d7c5 d3e2 e8e7 e1c1 h8d8 d2c4 d8d1 h1d1 f7f6 a2a3 f6e5 f4e5 b7b6 b2b4 c5b7 d1d4 c8d7 c1b2 g7g6 d4h4 h7h5 e2d3 a8g8 g2g3');
INSERT INTO evaluations VALUES(44569,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(44569,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(44569,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(44569,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(44569,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(44569,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(44569,36,'cp -12','g7g6 e4e5 f6d5 d3d4 c5d4 d1d4 e7e6 f1c4 b8c6 d4e4 f8g7 g1f3 d7d6 e5d6 d8d6 e1g1 e8g8 f1d1 d5f6 d1d6 f6e4 d6d1 b7b6 c1f4 c8b7 b1d2 e4d2 d1d2 f8d8 c4e2 c6a5 a1d1 d8d2 f3d2 a8d8 b2b4 a5c6 d2e4 d8d1 e2d1 c6e7 e4d6');
INSERT INTO evaluations VALUES(45975,30,'cp -193',NULL);
INSERT INTO evaluations VALUES(45975,31,'cp -187',NULL);
INSERT INTO evaluations VALUES(45975,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(45975,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(45975,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(45975,35,'cp -184',NULL);
INSERT INTO evaluations VALUES(45975,36,'cp -171','h7h6 g5f3 f6e4 d2d4 d6d5 c4d3 b8c6 c1e3 f8e7 c2c3 c5d4 c3d4 e7d6 d3e4 d5e4 f3d2 e8g8 d2e4 d6h2 g1h2 d8h4 h2g1 h4e4 b1c3 e4g6 d1f3 g8h7 a1d1 c8d7 f3g3 g6g3 f2g3 a8c8 d4d5 e6d5 d1d5');
INSERT INTO evaluations VALUES(45574,30,'cp -188',NULL);
INSERT INTO evaluations VALUES(45574,31,'cp -184',NULL);
INSERT INTO evaluations VALUES(45574,32,'cp -186',NULL);
INSERT INTO evaluations VALUES(45574,33,'cp -175',NULL);
INSERT INTO evaluations VALUES(45574,34,'cp -180',NULL);
INSERT INTO evaluations VALUES(45574,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(45574,36,'cp -176','g5f3 f6e4 f1e1 d6d5 c4b5 c8d7 b5d7 d8d7 c2c4 b8c6 c4d5 d7d5 b1c3 e4c3 d2c3 d5d1 e1d1 f8e7 c1e3 b7b6 g1f1 a8d8 a2a3 e7f6 d1d8 e8d8 a3a4 d8c7 f1e2 f6e7 g2g3 f7f6 h2h4 h6h5 a1g1 e7d6 g3g4 c6e7 g4g5 e7f5 g5f6 g7f6 f3d2 c7c6');
INSERT INTO evaluations VALUES(45576,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(45576,31,'cp -172',NULL);
INSERT INTO evaluations VALUES(45576,32,'cp -171',NULL);
INSERT INTO evaluations VALUES(45576,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(45576,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(45576,35,'cp -190',NULL);
INSERT INTO evaluations VALUES(45576,36,'cp -170','f6e4 f1e1 d6d5 c4b5 c8d7 b5d7 d8d7 c2c4 b8c6 c4d5 d7d5 b1c3 e4c3 d2c3 d5d1 e1d1 f8e7 c1e3 g7g5 a2a3 g5g4 f3d2 e8c8 d2c4 h6h5 d1d8 h8d8 a3a4 d8d5 g2g3 c8d7 g1g2 c6e5 c4e5 d5e5 h2h3 g4h3 g2h3 d7c6 a1h1 e5e4 b2b3 f7f5 c3c4 h5h4 h3g2 h4g3');
INSERT INTO evaluations VALUES(51578,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(51578,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(51578,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(51578,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(51578,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(51578,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(51578,36,'cp 98','b6c5 g1f3 c8b7 b1c3 a7a6 f1c4 e7e6 c1f4 b8c6 d1e2 c6d4 e2d3 d8f6 f4g3 g8e7 e1c1 e7c6 c3d5 e6d5 e4d5 d4f3 d5c6 b7c6 g2f3 f8e7 h1e1 e8g8 g3e5 f6h6 f3f4 e7f6 c1b1 f6e5 f4e5 h6f4 b2b3 f4f2 d3d6 c6b5 e1f1 f2h4 c4b5 a6b5 d6c5 a8e8 d1d2 b5b4 d2d7 h4h2 c5b4');
INSERT INTO evaluations VALUES(51621,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(51621,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(51621,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(51621,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(51621,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(51621,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(51621,36,'cp 104','g1f3 c8b7 b1c3 a7a6 c1f4 e7e6 f1c4 g8f6 d1e2 b8c6 e1c1 c6a5 f3e5 f8e7 h1e1 d8b6 c4b3 d7d6 b3a4 e8f8 e5f3 a8b8 b2b3 f6e8 c3d5 b7d5 e4d5 b6b4 f4g3 c5c4 d5e6 b4a3 c1d2 a3b4 c2c3 b4a3 d2e3 c4b3 a2b3 a5b3 e2c4 b3a5 c4d4 f7e6 e3e2 e7f6 g3d6 a3d6 d4d6 e8d6 d1d6 f6c3 e1d1 f8e7 d6a6');
INSERT INTO evaluations VALUES(53256,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(53256,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(53256,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(53256,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(53256,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(53256,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(53256,36,'cp -47','d4c5 g8f6 g1f3 e7e6 c1e3 d8c7 b1c3 c8d7 a2a3 f8c5 e3c5 c7c5 d1d4 c5c7 e2e3 e8g8 d4d2 f8c8 f1d3 b8a6 e1g1 a6c5 h2h3 a8b8 f3e5 b7b5 e5d7 f6d7 c3b5 c7b6 a3a4 a7a6 b5d4 b6b2 f1b1 b2b1 a1b1 b8b1 g1h2 b1a1 d4b3 a1a2 b3c5');
INSERT INTO evaluations VALUES(54137,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(54137,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(54137,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(54137,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(54137,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(54137,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(54137,36,'cp -33','e6d5 f1b5 b8c6 b1c3 g8f6 b5c6 b7c6 e1g1 f8e7 d2d3 e8g8 f3e5 d8c7 d1f3 c8a6 c1d2 f8e8 f1e1 f6d7 g1h1 d7e5 f4e5 e7f8 d2f4 e8e6 h2h3 c7d7 f3f2 f7f6 c3a4 d5d4 b2b3 c5c4 e5f6 e6f6 d3c4 c6c5 f2g3');
INSERT INTO evaluations VALUES(54737,30,'cp 261',NULL);
INSERT INTO evaluations VALUES(54737,31,'cp 242',NULL);
INSERT INTO evaluations VALUES(54737,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(54737,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(54737,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(54737,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(54737,36,'cp 274','e2e4 b8c6 f1c4 e7e5 e1g1 c8g4 d1b3 c6a5 b3a4 c7c6 c4e2 b7b5 a4c2 e5d4 b1d2 d8d7 h2h3 a8d8 e2d3 g4e6 e4e5 c6c5 e5f6 g8f6 f3g5 e6d5 d2f3 d5f3 g5f3 c5c4 f1e1 f8e7 d3g6 e8f8 f3e5 d7c8 g6h7 h8h7 e5g6 f8f7 e1e7 f7g8 c1f4 d4d3 c2c3');
INSERT INTO evaluations VALUES(54739,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(54739,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(54739,32,'cp 239',NULL);
INSERT INTO evaluations VALUES(54739,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(54739,34,'cp 245',NULL);
INSERT INTO evaluations VALUES(54739,35,'cp 257',NULL);
INSERT INTO evaluations VALUES(54739,36,'cp 258','c7c5 d4d5 g7g6 f1c4 g8h6 b1c3 a7a6 d1a4 c8d7 a4a3 d7g4 a3c5 h6f7 c4e2 b8d7 c5d4 f8g7 c1e3 a8c8 d4b4 g4f3 g2f3 f6f5 h2h4 b7b5 h4h5 g6g5 e4f5 e8g8 b4a3 g8h8 a1d1 f7d6 h5h6 g7f6 a3a6');
INSERT INTO evaluations VALUES(55388,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(55388,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(55388,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(55388,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(55388,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(55388,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(55388,36,'cp 113','f4g3 g8f6 b1d2 c8d7 c2c4 h7h5 h2h4 g5g4 c4d5 f6d5 a1c1 d7c6 a2a3 f8d6 g3d6 c7d6 e3e4 d5e7 f1d3 d6d5 e4e5 d8a5 c1c3 c6a4 d1c1 b8c6 c3c5 a5b6 g1e2 a7a6 b2b4 a4b5 d3b5 a6b5 e1g1 e7f5 d2b3 b6d8 e2c3 c6d4 b3d4 f5d4');
INSERT INTO evaluations VALUES(55389,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(55389,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(55389,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(55389,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(55389,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(55389,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(55389,36,'cp 104','g8f6 b1d2 b8d7 f1d3 c7c5 c2c3 f8e7 h2h3 d8b6 d1c2 e7d6 g3d6 b6d6 g1f3 h7h6 e1g1 e8g8 e3e4 c5d4 c3d4 d5e4 d3e4 d7b6 f1e1 b6d5 h3h4 g5h4 f3e5 c8d7 d2c4 d6e7 e4d5 f6d5 e5d7 e7d7 c4e5 a8c8 e5d7 c8c2 d7f8');
INSERT INTO evaluations VALUES(57321,30,'cp -189',NULL);
INSERT INTO evaluations VALUES(57321,31,'cp -196',NULL);
INSERT INTO evaluations VALUES(57321,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(57321,33,'cp -193',NULL);
INSERT INTO evaluations VALUES(57321,34,'cp -195',NULL);
INSERT INTO evaluations VALUES(57321,35,'cp -198',NULL);
INSERT INTO evaluations VALUES(57321,36,'cp -177','d8h4 e1d2 f8c5 c2c4 d7d6 b1c3 g8e7 g1f3 h4h6 d2c2 c8g4 d3d4 c5b6 f1e2 b8c6 d4d5 g4f3 e2f3 c6d4 c2b1 e8g8 g2g3 d4f3 d1f3 h6h3 f3f4 f7f5 f4f1 h3g4 e4f5 e7f5 f1e2 g4e2 c3e2 a8e8 e2f4 f5e3 f4e6 f8f3 c1e3 f3e3 b1c2 e3e2 c2c3');
INSERT INTO evaluations VALUES(4921,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(4921,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(4921,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(4921,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(4921,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(4921,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(4921,36,'cp 44','a4b3 g8f6 a2a4 c8b7 d2d3 e8g8 b1c3 c6a5 f3e5 a5b3 c2b3 d7d5 c1g5 c5e7 e5g4 f6g4 g5e7 d8e7 d1g4 e7b4 a4b5 a6b5 a1a8 b7a8 g4d7 d5e4 d3e4 a8e4 c3e4 b4e4 d7c7 e4e6 c7c3 f8c8 c3d3 g7g6 g2g3 e6f6 f1d1 f6b2 d3b5');
INSERT INTO evaluations VALUES(14817,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(14817,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14817,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14817,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14817,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(14817,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14817,36,'cp 46','e2e3 f8d6 f4g3 e8g8 c2c4 b7b6 f1e2 c6e7 e1g1 a7a5 a1c1 c8a6 f1e1 a8c8 g3h4 e7f5 h4f6 d8f6 d1b3 d5c4 d2c4 a6c4 c1c4 c8b8 c4c6 f5e7 c6c2 c7c5 g2g3 c5d4 f3d4 g7g6 e1c1 d6c5 d4f3 f8d8 a2a3 e7d5');
INSERT INTO evaluations VALUES(22631,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(22631,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22631,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(22631,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(22631,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(22631,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(22631,36,'cp -14','d7d5 b1c3 d5d4 c3e2 f7f5 d2d3 f8c5 a2a3 a7a5 g2g3 g8e7 f1g2 a5a4 c1d2 e8g8 e2c1 e6e5 e1g1 e5f4 g3f4 e7g6 g1h1 c5e7 c1e2 g8h8 a1c1 c8e6 e4e5 d8d7 d1e1 a8e8 e1f2 e6d5 c1e1 e8a8 f2g3 a8e8');
INSERT INTO evaluations VALUES(25480,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(25480,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(25480,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(25480,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(25480,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(25480,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(25480,36,'cp 56','c1g5 d8e7 b5e2 d7e6 f3e5 h7h6 g5f6 e7f6 e2b5 a7a6 b5c6 b7c6 e5c6 f8d6 d1h5 e8g8 e1g1 f6f4 g2g3 f4g5 h5g5 h6g5 f1e1 g5g4 c6e5 g7g5 e1e2 g8g7 c3a4 f8e8 a1e1 e6f5 a4c5 d6c5 d4c5 a8b8 e5d3 e8e4 b2b4 e4e2');
INSERT INTO evaluations VALUES(25478,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(25478,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(25478,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(25478,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(25478,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(25478,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(25478,36,'cp 36','d8e7 b5e2 d7e6 f3e5 h7h6 g5f6 e7f6 e2b5 a7a6 b5c6 b7c6 e1g1 f8d6 e5c6 e8g8 d1h5 f6g5 h5g5 h6g5 a2a3 f7f6 c6a5 g8f7 a5b7 a8b8 b7d6 c7d6 c3d1 a6a5 f2f3 f6f5 f1e1 f5f4 e1e2 a5a4 d1c3 f8c8 a1e1 b8b2 c3a4');
INSERT INTO evaluations VALUES(25553,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(25553,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(25553,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(25553,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(25553,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(25553,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(25553,36,'cp 94','b5d7 d8d7 d1d4 f8e7 f1d1 e7d8 a2a4 d7e6 d4b4 a8b8 c1e3 b7b6 b4b5 f6d7 d1e1 e8g8 a1d1 f7f5 e3d4 c7c5 e4f5 e6f5 d4e3 d7e5 e1e2 g8h8 c3d5 d8h4 e3f4 b8e8 f2f3 e8e6 f4e5 e6e5 b5c4 h4g5 e2e5 f5e5');
INSERT INTO evaluations VALUES(25445,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(25445,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(25445,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(25445,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(25445,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(25445,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(25445,36,'cp 72','d8d7 d1d4 f8e7 f1d1 e7d8 a2a4 d7e6 d4b4 a8b8 c1e3 b7b6 b4b5 f6d7 e3d4 e8g8 f2f3 a7a5 b2b3 d8f6 d4f6 d7f6 d1d4 f6d7 b5c4 d7e5 c4e6 f7e6 c3b5 f8f7 a1e1 g8f8 e1d1 e5g6 d4c4');
INSERT INTO evaluations VALUES(27465,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(27465,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(27465,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(27465,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(27465,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(27465,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(27465,36,'cp 90','d1d4 f8e7 f1d1 e7d8 a2a4 d7e6 a4a5 e8g8 f2f3 f8e8 c1e3 e6c8 d4d2 h7h6 b2b3 d8e7 d2f2 a7a6 c3e2 f6d7 a1a4 e7f8 h2h3 b7b5 a5b6 d7b6 a4a5 b6d7 e2g3 d7e5 f3f4 e5c6 a5a4 a6a5 a4c4');
INSERT INTO evaluations VALUES(26936,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(26936,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(26936,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(26936,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(26936,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(26936,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(26936,36,'cp 39','d6e5 b1d2 f8d6 c2c3 d8e7 h2h3 g4d7 e1g1 h7h6 c4d5 g8f6 d2c4 e8g8 f1e1 f8e8 a2a4 c6d8 c4d6 e7d6 d5b3 d6d1 b3d1 b7b6 d1c2 c7c5 a4a5 d8c6 a5b6 a7b6 c1e3 e8d8 e1d1 a8a1 d1a1 d7e6 f3d2 c6a5 b2b3 a5c6 f2f3 g8h8 d2c4');
INSERT INTO evaluations VALUES(28965,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(28965,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(28965,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28965,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28965,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(28965,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(28965,36,'cp 12','f5d3 d1d3 e7e6 g1f3 f8e7 e1g1 e8g8 b2b3 b8d7 f1e1 d8c7 e3e4 d5e4 c3e4 a8d8 e4f6 d7f6 d3e2 h7h6 c1b2 f8e8 g2g3 c6c5 a1d1 c5d4 b2d4 b7b6 d1d3 f6d7 e1d1 d7c5 d4e5 c7c6 d3d8 e8d8 d1d8 e7d8 e5c3 c5e4 c3b2 d8e7');
INSERT INTO evaluations VALUES(29030,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(29030,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(29030,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(29030,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(29030,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(29030,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29030,36,'cp 13','d1d3 e7e6 g1f3 f8b4 e1g1 e8g8 c1d2 b4a5 f1d1 f8e8 h2h3 b8d7 a2a4 a5c7 e3e4 d5e4 c3e4 e6e5 d2g5 e5d4 e4f6 d7f6 d3f5 d8d7 f3d4 d7f5 d4f5 f6e4 g5e3 a8d8 e3a7 g7g6 f5e3 c7e5 d1d8 e8d8 a1d1 d8a8 a7d4 e5d4 d1d4 e4c5 b2b4');
INSERT INTO evaluations VALUES(29029,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29029,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29029,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(29029,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(29029,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(29029,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29029,36,'cp 15','e7e6 g1f3 f8b4 e1g1 e8g8 c1d2 b4a5 a1b1 b8d7 f1d1 f8e8 h2h3 a7a6 d2e1 a8c8 a2a3 h7h6 b2b3 a5c7 e3e4 d5e4 c3e4 f6e4 d3e4 d8f6 e1c3 f6f4 e4d3 d7f6 g2g3 f4e4 b1c1 c8d8 d3e4 f6e4 c3e1 a6a5');
INSERT INTO evaluations VALUES(29285,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29285,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29285,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29285,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(29285,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(29285,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(29285,36,'cp 17','g1f3 f8b4 c1d2 e8g8 e1g1 b4a5 f1d1 b8d7 h2h3 f8e8 a1b1 h7h6 a2a4 a5c7 e3e4 d5e4 c3e4 d8e7 e4f6 e7f6 d2c3 e6e5 d1e1 a8d8 b1d1 f6e7 d4e5 d7e5 d3f5 d8d1 e1d1 e5f3 f5f3 e7e6 f3g4 e6g4 h3g4 e8d8 d1d8 c7d8 g1f1 f7f6 f1e2 g8f7 f2f3 d8e7 e2d3 a7a6 d3e4');
INSERT INTO evaluations VALUES(38189,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(38189,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(38189,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(38189,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(38189,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(38189,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(38189,36,'cp 71','d1d5 b8d7 d5e4 g8e7 g1f3 e8g8 e2e3 d7f6 e4c4 c5d6 c3b5 c8g4 c1d2 a8c8 c4d3 d6c5 d3d8 f8d8 h2h3 g4e6 b5c3 e7d5 a2a3 d5c3 d2c3 f6e4 c3d4 c5e7 f1e2 e4c5 d4c5 e7c5 e1g1 g7g6 f1c1 g8g7 c1d1 d8d1 a1d1');
INSERT INTO evaluations VALUES(38183,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(38183,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(38183,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(38183,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(38183,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(38183,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(38183,36,'cp 65','b8d7 d5e4 g8e7 e2e3 e8g8 g1f3 d7f6 e4c4 c5d6 f1e2 a7a6 c3e4 f6e4 c4e4 f8e8 c1d2 e7c6 e4c2 c6b4 c2d1 c8f5 e1g1 f5c2 d1c1 c2e4 d2b4 d6b4 c1c4 b4f8 f1d1 d8f6 c4d4 f6e7 d4b6 g7g6 a1c1 e8c8 f3d4 f8g7 h2h3 c8c1 d1c1 a8d8 c1c7 g7d4 e3d4 e7g5');
INSERT INTO evaluations VALUES(44147,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(44147,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(44147,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44147,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(44147,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(44147,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(44147,36,'cp 36','f1c4 d7d6 g1f3 f8e7 d1e2 e8g8 e1g1 b8d7 a2a3 a7a6 c1d2 d6e5 d4e5 d7b6 c4d3 c8d7 b1c3 d5c3 d2c3 b6d5 e2e4 g7g6 c3d2 d7c6 f3d4 e7g5 d4c6 b7c6 d2g5 d8g5 g2g3 h7h5 e4e2 d5f4 e2e3 f4h3 g1g2 g5e3');
INSERT INTO evaluations VALUES(48898,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(48898,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(48898,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(48898,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(48898,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(48898,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(48898,36,'cp -51','d8d1 e1d1 f6e4 c1e3 f7f6 f2f3 e4d6 e5c4 c8f5 b1a3 b8c6 c2c3 d6f7 a3b5 e8c8 d1e1 h7h5 b5d4 c6d4 e3d4 f5e6 h2h4 f7h6 a1d1 f8e7 d4f2 b7b6 f1e2 a7a5 a2a3 c8b7 c4e3 e7c5 e3c2 h6f5 f2c5 d8d1 e1d1 b6c5 d1c1 h8d8 f3f4 e6d5 h1d1 f5h4');
INSERT INTO evaluations VALUES(33913,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(33913,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(33913,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(33913,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(33913,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(33913,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(33913,36,'cp -43','e1d1 f6e4 c1e3 f7f6 e5f3 b8c6 f1d3 e4c5 d3e2 c8e6 d1c1 e8c8 b1c3 c5d7 a2a4 f8c5 a4a5 a7a6 h1d1 c5e3 f2e3 d7c5 f3d4 e6d7 d4c6 d7c6 d1d8 h8d8 e2f3 c6f3 g2f3 g7g5 b2b4 c5e6 a1a4 h7h5 b4b5 a6b5 c3b5 c8d7 b5c3 h5h4 c3d5 f6f5 a5a6 b7a6');
INSERT INTO evaluations VALUES(33912,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(33912,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(33912,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(33912,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(33912,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(33912,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(33912,36,'cp -49','f6e4 c1e3 f7f6 e5f3 b8c6 f1d3 e4c5 d3e2 c8f5 d1c1 e8c8 b1c3 a7a5 a2a3 g7g5 h1d1 d8d1 c1d1 f8e7 c3b5 c5e6 a1c1 h8d8 d1e1 e7c5 e2c4 c5e3 f2e3 h7h6 e1f2 e6c5 b5d4 c6d4 f3d4 f5g6 c1d1');
INSERT INTO evaluations VALUES(33950,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(33950,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(33950,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(33950,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(33950,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(33950,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(33950,36,'cp -46','c1e3 f7f6 e5f3 b8c6 f1d3 e4c5 d3e2 c8e6 d1c1 e8c8 b1c3 c5d7 e2b5 e6f7 a2a3 f8d6 b5c6 b7c6 b2b3 h8e8 c1b2 c8b7 h1e1 d7b6 e3d2 c6c5 h2h3 b6d5 c3d5 f7d5 a1d1 a7a6 e1e8 d8e8 d1e1 e8e1 f3e1 d6e5 d2c3');
INSERT INTO evaluations VALUES(51039,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(51039,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(51039,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(51039,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(51039,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(51039,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(51039,36,'cp 21','c7c5 d2d4 c5d4 f3d4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 f8e7 a2a3 d8c7 f1e1 c7e5 c1d2 g8f6 f2f4 e5c7 e4e5 f6d5 f4f5 e6f5 d3f5 g7g6 c3d5 c6d5 f5e4 d5e4 e2e4 e8g8 a1d1 f8e8 d2h6 a8d8 g1h1 c7c6 d1d5 e7f8 h6f8 g8f8 h2h3 f8g8');
INSERT INTO evaluations VALUES(54060,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(54060,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(54060,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(54060,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(54060,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(54060,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(54060,36,'cp 28','c4d5 f8c5 d1c2 d8b6 e2e3 e6d5 b1c3 g8f6 c1d2 c8d7 g1f3 e8g8 f1e2 c5d6 e1g1 b8c6 f1d1 a8c8 d2e1 d7e6 a1c1 a7a6 c2b1 h7h6 c3a4 b6a7 b2b4 f6d7 b4b5 a6b5 e2b5 d6a3 c1c3 a3b4 b5c6 b4c3 c6d7 c3e1 d7c8 f8c8');
INSERT INTO evaluations VALUES(54054,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(54054,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(54054,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(54054,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(54054,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(54054,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(54054,36,'cp 25','f8c5 d1c2 d8b6 e2e3 e6d5 b1c3 g8f6 c1d2 c8d7 g1f3 e8g8 f1e2 c5d6 e1g1 b8c6 a1c1 a8c8 f1d1 h7h6 c2b1 f8e8 d2e1 d7e6 a2a3 d6b8 b2b4 b6d8 h2h3 d8d6 c3b5 d6d7 e2f1 e6f5 b1a2 f5e4 b5d4 a7a6 d4c6 c8c6 c1c6 d7c6');
INSERT INTO evaluations VALUES(55688,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(55688,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(55688,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55688,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55688,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(55688,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(55688,36,'cp -14','b1c3 g8f6 h2h3 f8e7 c1e3 e8g8 d1c2 c7c5 a1d1 c8d7 e4e5 d6e5 f3e5 d8b6 e5d7 b8d7 f1e2 d7e5 e1g1 a8c8 f1e1 e5c6 e2f1 c6d4 e3d4 c5d4 e1e7 d4c3 b2c3 f8e8 e7e8 c8e8 c3c4 g7g6 c2d2 f6e4 d2f4 e8e7 g2g3 g8g7 d1d5 b6f6 f4f6 g7f6 f1g2 e4c3 d5d6 f6e5');
INSERT INTO evaluations VALUES(55686,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55686,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(55686,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(55686,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(55686,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(55686,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(55686,36,'cp -39','g8f6 c1e3 f8e7 d1c2 e8g8 e1c1 c7c6 c1b1 d8a5 h2h3 b8a6 f3d4 c6c5 d4b3 a5b6 a2a3 c8e6 b3d2 a8d8 f2f4 d6d5 f4f5 e6f5 e4f5 d5d4 c3a4 b6c6 e3g1 b7b5 a4c5 a6c5 d2b3 c6b6 d1d4 c5b3 d4d8 b6d8 c2b3 a7a6 f1d3 f8e8 g1e3 e7f8');
INSERT INTO evaluations VALUES(10927,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(10927,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(10927,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(10927,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(10927,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(10927,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(10927,36,'cp 36','d2d4 e5d4 f3d4 e8g8 d4c6 b7c6 e4e5 f6e8 f1d3 f7f6 c1f4 f6e5 f4e5 e7f6 d1e2 d7d5 e1g1 e8d6 a1e1 f6e5 e2e5 c8f5 d3f5 f8f5 e5d4 d8f6 d4c5 d6b7 c5e7 a8f8 f2f3 b7d6 e7f6 f8f6 e1e7 f6f7 e7e6 f5f6 e6f6 f7f6 f1e1');
INSERT INTO evaluations VALUES(17983,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(17983,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(17983,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(17983,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(17983,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(17983,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(17983,36,'cp 24','a4b3 f8e7 a2a4 b5b4 c3d5 g8f6 d2d4 c8g4 d4e5 c6e5 d5b4 e8g8 b4d3 f6d7 d3e5 d7e5 c1f4 e7f6 f4e5 f6e5 h2h3 g4f3 d1f3 a8b8 g2g3 g7g6 a1b1 b8b4 b3d5 b4b2 b1b2 e5b2 e1g1 d8e8 a4a5 c7c6 d5a2 e8e5 f1b1 e5c3 g1g2 c3c2');
INSERT INTO evaluations VALUES(21892,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21892,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(21892,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21892,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(21892,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(21892,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(21892,36,'cp 8','f8e7 g5e7 g8e7 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 h2h4 c8e6 b1a3 e7c6 f1c4 d8e7 f2f3 a8d8 c2c3 g7g6 h4h5 f7f5 c4e6 e7e6 a1d1 d8d1 e1d1 g6g5 d1e2 c6e7 h1d1 f5e4 f3e4 e7c8 a3c4 b7b5 c4e3 c8d6 d1d5 c7c6');
INSERT INTO evaluations VALUES(5625,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5625,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(5625,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(5625,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5625,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(5625,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5625,36,'cp 12','g5e7 g8e7 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 a2a4 f7f6 h2h4 c8e6 b1d2 e7c6 f1c4 d8e7 c2c3 h7h5 c4e6 e7e6 d2c4 a8d8 e1e2 c6e7 a4a5 e7c8 a1d1 g7g6 f2f3 f6f5 g2g3 h8f8 b2b3 b7b6 d1d8 f8d8 a5b6 a7b6 h1a1 f5e4 f3e4');
INSERT INTO evaluations VALUES(24994,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(24994,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(24994,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(24994,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(24994,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(24994,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(24994,36,'cp 50','g4f3 g2f3 e7e5 d4d5 c6d4 h1g1 f6d7 c3b5 d4b5 c4b5 a7a6 b5a4 g7g6 e1e2 f8d6 c1g5 f7f6 g5e3 e8g8 d1c2 g8h8 h2h4 f8f7 a1c1 d8e7 a4d7 e7d7 a2a4 a8f8 h4h5 g6g5 h5h6 f8g8 c2b3 b7b6 a4a5 b6a5 c1c6 d7h3 c6a6 h3h6 a6a5 h6g6 a5a7');
INSERT INTO evaluations VALUES(24996,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(24996,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(24996,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(24996,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(24996,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(24996,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(24996,36,'cp 60','g2f3 e7e5 d4d5 c6d4 h1g1 f6d7 c3b5 d4b5 c4b5 a7a6 b5a4 g7g6 e1e2 f8d6 c1g5 f7f6 g5e3 e8g8 h2h4 g8h8 a1c1 d8e7 d1c2 f8g8 a4d7 e7d7 a2a4 a8f8 c2b3 b7b6 b3c2 f6f5 g1g5 h7h6 g5g1 h6h5 g1g5 f8f7 e3d2 a6a5 d2e3 f5e4');
INSERT INTO evaluations VALUES(25765,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(25765,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(25765,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(25765,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(25765,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(25765,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(25765,36,'cp 69','a4c2 g8f6 d2d4 f8e7 b1d2 e8g8 f1e1 f8e8 d2f1 h7h6 a2a3 e7f8 b2b3 d7g4 d4d5 c6e7 h2h3 g4d7 c1e3 e7g6 c3c4 f6h5 g1h2 b5c4 b3c4 a8b8 f1g3 h5g3 f2g3 c7c6 d1d2 c6c5 a1b1 b8b1 e1b1 d8c7 c2d3 e8b8');
INSERT INTO evaluations VALUES(25760,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(25760,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(25760,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(25760,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(25760,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(25760,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(25760,36,'cp 65','f8e7 d2d4 g8f6 f1e1 e8g8 a2a3 h7h6 h2h3 f8e8 b2b3 e7f8 b1d2 a6a5 d2f1 b5b4 d4d5 c6e7 c3c4 c7c5 g1h2 e7g6 g2g3 g6e7 c1d2 g7g6 a1a2 d8c8 f3g1 f8g7 d1f3 f6h7 e1a1 c8b7 g3g4 e8b8 a3b4 a5b4 h3h4 a8a2 a1a2 b8a8');
INSERT INTO evaluations VALUES(29132,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29132,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(29132,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29132,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29132,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29132,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(29132,36,'cp 0','e7e6 d1b3 d8c7 f3h4 f5g6 h4g6 h7g6 g2g3 b8c6 f1g2 c7d7 e1g1 c6a5 b3a4 d7a4 c3a4 f8d6 h2h4 e8e7 b2b3 f6e4 c1b2 a5c6 g2e4 d5e4 a1c1 a8c8 a4c3 f7f5 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(32053,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(32053,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(32053,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(32053,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(32053,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(32053,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(32053,36,'cp 164','h2h3 g4h5 b1c3 e5d4 d1d4 b8c6 d4e3 h5f3 g2f3 g7g6 f3f4 d8d7 c1d2 f8g7 c3d5 e8c8 d2c3 g7c3 e3c3 h8h7 e1c1 c6e7 d5b4 c7c5 b4d3 d7c7 c4b3 e7c6 e4e5 c6d4 h1e1 g8e7 e5d6 c7d6 c3c4 e7d5 d3c5 d4b3 c5b3 d6c7 b3c5 c7f4 c4f4 d5f4 d1d8 c8d8 c5b7 d8c7 e1e7 c7b6 b7d6 f4h3');
INSERT INTO evaluations VALUES(32052,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(32052,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(32052,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(32052,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(32052,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(32052,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(32052,36,'cp 175','g4h5 b1c3 e5d4 d1d4 b8c6 d4e3 h5f3 g2f3 g7g6 f3f4 d8d7 c1d2 f8g7 c3d5 e8c8 d2c3 g7c3 e3c3 h8h7 e1c1 c6e7 d5b4 c7c5 b4d5 e7d5 c4d5 d7c7 h3h4 g8e7 h4h5 c8b8 a2a3 a7a6 c1b1 f7f5 h5g6 e7g6 c3g3 g6e7 d5e6 f5e4 f4f5 d6d5 f5f6 c7g3 f2g3 d5d4 f6e7 h7e7 h1h6 e4e3 b1c1 e3e2');
INSERT INTO evaluations VALUES(35003,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(35003,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(35003,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(35003,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35003,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(35003,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(35003,36,'cp 17','h5d1 d7d6 b1c3 c8e6 c3d5 f6d8 d2d4 c7c6 d5e3 e6c4 e3c4 b7b5 c4e3 b8d7 a2a4 g8f6 d4e5 d6e5 a4b5 c6b5 e3d5 f8g7 d5f6 d7f6 d1d8 e8d8 c1g5 a7a5 e1c1 d8c7 f2f3 h8d8 g1e2 b5b4 c1b1 d8d1 h1d1 c7c6 e2c1 f6d7 c1d3 g7f8 b2b3 a5a4 g5e3 f7f5 e4f5 a4b3 c2b3 g6f5');
INSERT INTO evaluations VALUES(35594,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(35594,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(35594,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(35594,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(35594,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(35594,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(35594,36,'cp 56','c1f4 f8e8 a2a4 b8c6 f1e1 c6e5 c4f1 c8g4 d1d2 e5g6 f4e3 g4d7 a4a5 f6g4 c3d5 g4e3 d5e7 g6e7 e1e3 e7c6 f1c4 d8g5 c4d5 c6d4 d2d4 d7c6 a1e1 c6d5 e4d5 e8e3 e1e3 g5f5 d4c4 f5d7 h2h3 h7h6 e3b3 a8b8 c4c3 b7b5');
INSERT INTO evaluations VALUES(35804,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(35804,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(35804,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(35804,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(35804,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(35804,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(35804,36,'cp 81','d4f5 c8f5 e4f5 b8d7 f1e1 h7h6 c4f1 d7e5 c3d5 f6d5 d1d5 e5c6 c2c3 d8b6 a1b1 e7f6 c1f4 a8d8 g2g4 c6e7 d5d3 e7c6 f1g2 f8e8 e1e8 d8e8 a2a4 c6e5 d3b5 b6b5 a4b5 e5g4 g2b7 f6e5 b7c6 e8c8 f4g3 e5g3 h2g3 g4e5 c6g2 c5c4 f2f4');
INSERT INTO evaluations VALUES(35801,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(35801,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(35801,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(35801,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(35801,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(35801,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(35801,36,'cp 65','c8f5 e4f5 b8d7 f1e1 h7h6 c4f1 d7e5 c1f4 e5c6 c3d5 f6d5 d1d5 d8b6 a2a4 a8c8 a1b1 c8d8 c2c3 d8d7 g2g3 e7f6 h2h4 a7a6 f1d3 f8d8 d3c4 c6e5 f4e5 d6e5 d5e4 b6a5 b1a1 a5b6 a1a2 d7d2 a4a5 b6c6 e4c6 b7c6 c4f1');
INSERT INTO evaluations VALUES(27794,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(27794,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(27794,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(27794,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(27794,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(27794,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(27794,36,'cp 66','e4f5 b8c6 g2g4 h7h6 f2f4 d6d5 c3d5 f6d5 c4d5 c6b4 d5b7 d8d1 f1d1 a8d8 c1d2 b4c2 a1b1 e7f6 b7e4 c2b4 g4g5 h6g5 f4g5 f8e8 e4f3 f6d4 g1g2 b4d5 g2g3 e8e5 g3g4 g7g6 f5g6 f7g6 h2h4 d5e3 d2e3 e5e3 b1c1 d8d6 c1c2 g8g7');
INSERT INTO evaluations VALUES(27793,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(27793,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(27793,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(27793,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(27793,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(27793,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(27793,36,'cp 90','b8d7 f1e1 h7h6 c4f1 d7b8 c3d5 b8c6 c1f4 f6d5 d1d5 d8b6 a2a4 e7g5 f4g5 h6g5 d5d6 b6b2 d6c5 b2f6 a1d1 a8d8 d1d8 f8d8 c5e3 a7a5 e3e4 f6d4 g2g3 g8f8 e4d4 c6d4 e1b1 d4f5 b1b7 f5d6 b7b3 g5g4 f1b5 d6b5 a4b5');
INSERT INTO evaluations VALUES(40485,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(40485,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(40485,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(40485,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40485,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40485,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(40485,36,'cp 24','b4c3 b2c3 e8g8 a3a4 f8e8 c1a3 d8c7 a1c1 b8d7 f1e2 a7a6 c4c5 e6e5 e1g1 e5e4 f3d2 d7f8 c1b1 f8g6 c3c4 g6h4 g1h1 h7h6 a3b4 c8e6 b4c3 c7c8 b1b2 e6f5 c4d5 c6d5 d2b1 c8d7 c3a5 g8h8 d1b3 a8b8 f1c1 f5g4 e2f1 d7f5');
INSERT INTO evaluations VALUES(40490,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40490,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(40490,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40490,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40490,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(40490,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(40490,36,'cp 14','b2c3 e8g8 a3a4 d8c7 c1a3 f8e8 a1c1 b8d7 f1e2 a7a6 c4c5 e6e5 e1g1 e5e4 f3d2 d7f8 c3c4 f8g6 c1b1 g6h4 g1h1 h7h6 b1b6 c8e6 c4d5 c6d5 f1g1 e6d7 d2f1 e8e6 b6e6 d7e6 f1g3 h4f5 g3f5 e6f5 a3b4 f5d7 b4e1 d7c6 d1b3 c7d7 a4a5 c6b5 e2b5 a6b5');
INSERT INTO evaluations VALUES(40488,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40488,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(40488,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40488,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(40488,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40488,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40488,36,'cp 14','e8g8 a3a4 f8e8 c1a3 h7h6 a1c1 a7a6 c4c5 b8d7 f1e2 e6e5 e1g1 e5e4 f3d2 d7f8 c3c4 f8g6 c1b1 g6h4 g1h1 a8b8 f1g1 c8e6 c4d5 d8d5 a3b4 d5g5 b4c3 h4f5 d1c2 e6d5 h2h3 e8e7 d2c4 g5h4 c3e1 b8e8 c4d6 f5d6');
INSERT INTO evaluations VALUES(36108,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36108,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36108,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(36108,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36108,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36108,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36108,36,'cp 25','a3a4 h7h6 c1a3 f8e8 a1c1 a7a6 c4c5 b8d7 f1e2 e6e5 e1g1 e5e4 f3d2 d7f8 c3c4 f8g6 c1b1 g6h4 g1h1 h6h5 b1b6 f6d7 b6b2 d7f6 c4d5 c6d5 b2b6 e8e6 d1b3 e6b6 b3b6 d8b6 c5b6 c8g4 f2f3 e4f3 g2f3 g4h3');
INSERT INTO evaluations VALUES(42871,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(42871,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(42871,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(42871,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(42871,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(42871,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(42871,36,'cp 170','c4f7 e8f7 f3g5 f7g8 d1g4 h7h6 g5f3 d7d6 g4g6 c8d7 e1g1 d7e8 g6f5 d8d7 f5d7 e8d7 c1e3 c5e3 f2e3 g8h7 b1c3 d7e6 f1f2 b8d7 f3h4 a7a5 a1f1 h8c8 h2h3 a5a4 g2g4 d7c5 g1g2 a4a3 b2b3 c8e8 c3d5 e6d5 e4d5 e5e4 d3e4 e8e4');
INSERT INTO evaluations VALUES(43881,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(43881,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(43881,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(43881,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(43881,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(43881,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(43881,36,'cp 66','c8b7 e4e5 f6d5 c3d5 b7d5 c1g5 d8c8 g1h3 d5b7 g5h4 h7h6 h4g3 e7e6 h3f4 b8c6 f1c4 b7a6 c4a6 c8a6 c2c3 a6c4 d1c2 c6b4 c2e4 b4d5 f4d5 c4d5 e4d5 e6d5 e1g1 f8e7 h2h4 a7a6 a2a4 e8g8 a4a5');
INSERT INTO evaluations VALUES(45347,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(45347,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(45347,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(45347,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(45347,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(45347,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(45347,36,'cp -53','c5d4 c3d5 e7e5 c1g5 f8e7 g5f6 e7f6 e2e3 d4e3 d5e3 e8g8 f1d3 b8a6 e1g1 a6c5 d3c2 e5e4 f3d2 f8e8 d1e2 f6b2 a1d1 d8h4 g2g3 h4e7 d2e4 c5e4 c2e4 e7e4 e2b2 e4f3 d1d4 b7b6 d4f4 f3c6 b2b5 c6b5 c4b5 c8h3 f1c1 a8d8 f4h4 h3d7 c1c7 d7b5 c7a7');
INSERT INTO evaluations VALUES(47086,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(47086,31,'cp -130',NULL);
INSERT INTO evaluations VALUES(47086,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(47086,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(47086,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(47086,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(47086,36,'cp -139','c7c6 b5d3 e7e5 g1f3 f8d6 e1g1 e8g8 f1e1 b8d7 h2h3 f8e8 e4d5 c6d5 d3f1 h7h6 c3e2 a7a6 e2g3 d7f8 d2d4 e5e4 f3e5 f8d7 c2c4 d7e5 d4e5 d6e5 c4d5 e5g3 f2g3 d8d5 d1d5 f6d5 a2a3 c8d7 f1c4 d5b6 c4b3 d7e6 c1e3 e6b3 e3b6 a6a5 a3a4 f7f6 b6e3 e8e5 e3d4');
INSERT INTO evaluations VALUES(44852,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(44852,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(44852,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(44852,33,'cp -119',NULL);
INSERT INTO evaluations VALUES(44852,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(44852,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(44852,36,'cp -127','b5d3 e7e5 g1f3 f8d6 e1g1 e8g8 f1e1 b8d7 h2h3 f8e8 e4d5 c6d5 d3f1 a7a6 c3e2 d7f8 d2d3 h7h6 a2a3 f8g6 c2c4 d5d4 f3d2 f6h7 b2b4 d6c7 a3a4 h7g5 e2g3 f7f5 d2f3 g5h7 d1c2 h7f6 a4a5 c8e6 c1d2 a8c8 c2b2 d8d7 b4b5');
INSERT INTO evaluations VALUES(44971,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(44971,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(44971,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(44971,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(44971,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(44971,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(44971,36,'cp -128','e7e5 g1f3 f8d6 e1g1 e8g8 f1e1 f8e8 h2h3 b8d7 e4d5 c6d5 d3f1 a7a6 c3e2 d7f8 d2d3 h7h6 c1d2 f8g6 c2c4 d5d4 b2b4 b7b6 e2g3 c8b7 a2a3 d6c7 g3e4 b7c6 f3h2 a6a5 e4g3 d8d7 g3e4 f6e4 d3e4 a5b4 a3b4');
INSERT INTO evaluations VALUES(48789,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(48789,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(48789,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(48789,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(48789,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(48789,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(48789,36,'cp 116','d4e5 d5c4 d1d8 e8d8 e5f6 g7f6 b1c3 h8g8 c1f4 c7c6 g2g3 c8e6 g1e2 b8d7 a1d1 a7a5 e2d4 d8c8 f2f3 a8a6 e1f2 b7b5 g3g4 a5a4 a2a3 b5b4 a3b4 f8b4 d4e2 b4c5 f4e3 c5e3 f2e3 c8c7 d1a1 g8a8 a1a3 d7c5 h2h4 a6b6 h1b1 b6b4 e2d4 c7d6 h4h5 c5d7 c3e2');
INSERT INTO evaluations VALUES(48763,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(48763,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(48763,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(48763,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(48763,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(48763,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(48763,36,'cp 123','d5c4 d1d8 e8d8 e5f6 g7f6 b1c3 c7c6 c1f4 c8e6 g1e2 h8g8 g2g3 e6h3 e2d4 b8d7 a1d1 f8c5 e1e2 d8c8 f4e3 d7e5 f2f3 h3g2 h1g1 c5d4 g1g2 d4c3 b2c3 g8d8 e3d4 a7a5 f3f4 e5d7 g3g4 b7b5 g4g5 d8g8 h2h4 c6c5 d4f2 c8c7');
INSERT INTO evaluations VALUES(51340,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(51340,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(51340,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(51340,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(51340,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(51340,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(51340,36,'cp 134','d1a4 d8d7 a4c2 e7e6 g1f3 f8d6 b1d2 g8f6 d2c4 e8g8 f3e5 d7e7 f1d3 c8b7 e1g1 f8c8 e3e4 b8a6 c1d2 a6b4 d2b4 d6b4 c4a5 b4a5 a1a5 f6d7 c2c3 d7e5 a5e5 a7a5 f1a1 a5a4 e5c5 e7d6 a1d1 c8d8 d3c2 h7h6 c5c7 a8c8 c7c5 c8c5 c3c5 d6c5 d4c5 d8c8 f2f3 b7a6 c2a4 c8c5 d1d8 g8h7 b2b4');
INSERT INTO evaluations VALUES(51338,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(51338,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(51338,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(51338,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(51338,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(51338,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(51338,36,'cp 120','d8d7 a4c2 e7e6 g1f3 f8d6 b1d2 g8f6 d2c4 e8g8 f3e5 d7e7 f1d3 c8b7 e1g1 f8c8 e3e4 b8a6 e5c6 b7c6 a1a6 c6b5 a6a5 b5c4 d3c4 d6b4 a5g5 c7c5 d4d5 e6d5 e4d5 f6e4 g5g4 e4d6 c4d3 g7g6 b2b3 a7a5 h2h4 a5a4 h4h5 a4b3 c2b3 c5c4 b3b4 c4d3 c1b2 a8a2 b4d4 a2b2 d4b2 e7e2 b2e2 d3e2 f1a1');
INSERT INTO evaluations VALUES(51356,30,'cp -181',NULL);
INSERT INTO evaluations VALUES(51356,31,'cp -197',NULL);
INSERT INTO evaluations VALUES(51356,32,'cp -189',NULL);
INSERT INTO evaluations VALUES(51356,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(51356,34,'cp -185',NULL);
INSERT INTO evaluations VALUES(51356,35,'cp -181',NULL);
INSERT INTO evaluations VALUES(51356,36,'cp -187','a7a5 c5b6 f8b4 c1d2 c7b6 g1f3 g8f6 d2b4 a5b4 e2e3 e8g8 b1d2 d8d6 f1e2 c8a6 e2a6 a8a6 e1g1 b8d7 d1b1 f8a8 f1c1 h7h6 c1c2 a6a3 b1b2 b6b5 f3e5 d7b6 c2c6 d6b8 b2b4 b6c4 d2c4 b5c4 b4b8 a8b8');
INSERT INTO evaluations VALUES(51129,30,'cp -170',NULL);
INSERT INTO evaluations VALUES(51129,31,'cp -179',NULL);
INSERT INTO evaluations VALUES(51129,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(51129,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(51129,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(51129,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(51129,36,'cp -188','c5b6 f8b4 c1d2 c7b6 e2e3 g8f6 d2b4 a5b4 g1f3 e8g8 b1d2 d8d6 f1e2 c8d7 d1b1 f8c8 e1g1 a8a3 b1b2 d7a4 f1c1 a3c3 e2f1 d6f8 f3e1 h7h6 d2f3 f6e4 f3e5 b6b5 e5d3 b8a6 a2a3 c3c1 d3c1 e4d6 a3b4 d6c4 b2e2 f8b4 e1d3 b4c3 e2a2 c4d6 a2a3 c3a3 a1a3');
INSERT INTO evaluations VALUES(1562,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(1562,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(1562,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(1562,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(1562,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(1562,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(1562,36,'cp -23','b1c3 c5d4 d1d4 a7a6 b5c6 b7c6 e1g1 c6c5 d4a4 a5a4 c3a4 c8d7 b2b3 d7a4 b3a4 g8e7 a1b1 e7c6 b1b7 a8c8 c1d2 f8e7 f1b1 e7d8 d2e3 h7h6 h2h3 d5d4 e3c1 d8a5 c1a3 a5b4 a3b4 c5b4 f3e1 e8g8 e1d3 c8b8 b7b8 f8b8 b1b3 g8f8 g1f1 f8e8 f1e2');
INSERT INTO evaluations VALUES(1560,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(1560,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(1560,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(1560,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(1560,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(1560,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(1560,36,'cp -22','c5d4 f3d4 a5c7 d1e2 g8e7 e1g1 a7a6 d4b3 e7g6 b5c6 b7c6 f2f4 a6a5 e2f2 a5a4 b3c5 f8c5 f2c5 a8a5 c5f2 e8g8 c1d2 f7f6 e5f6 f8f6 c3e2 a5b5 b2b3 e6e5 e2c3 b5b7 c3a4 g6f4 d2f4 f6f4 f2c5 b7b5 c5c3 f4f1 a1f1 c6c5 a2a3');
INSERT INTO evaluations VALUES(1559,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(1559,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(1559,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(1559,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(1559,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(1559,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(1559,36,'cp -17','d1d4 a7a6 b5c6 b7c6 e1g1 c6c5 d4f4 g8e7 f3g5 e7f5 g5f7 e8f7 g2g4 a5b4 f4f3 h7h5 g4f5 b4g4 f3g4 h5g4 f2f3 g4f3 f5e6 f7e6 f1f3 d5d4 c3e2 e6e5 c1f4 e5d5 a1e1 a8a7 f4g3 c8g4 e2f4 d5c6 f3f2 f8d6 f4d3 h8h5 e1e8 g4h3 e8d8 d6g3 h2g3 h5g5 f2e2 a7d7 d8d7 h3d7 d3e5 c6c7 e5d7');
INSERT INTO evaluations VALUES(3371,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(3371,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(3371,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(3371,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(3371,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(3371,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(3371,36,'cp 35','d2d4 f6e4 c3e4 d6d5 f3e5 d5c4 e5c6 b7c6 f1e1 f8e8 c1f4 c8f5 e4g3 f5e6 c2c3 d8d7 d1a4 a7a5 e1e2 e7d6 f4d6 c7d6 a1e1 h7h6 g3f1 e8b8 a4a3 b8b5 f1d2 a8b8 d2e4 b5d5 e2e3 g8h7 b2b3 c4b3 a2b3 d7c7 e4d2 c6c5 d2f3 e6f5 a3a2');
INSERT INTO evaluations VALUES(3402,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(3402,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(3402,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(3402,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(3402,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(3402,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(3402,36,'cp 21','f6e4 c3e4 d6d5 f3e5 d5c4 e5c6 b7c6 c1f4 c8f5 f1e1 f8e8 d1d2 a8b8 e4g3 f5e6 b2b3 d8d7 g3e4 f7f6 e1e3 e7f8 a1d1 d7f7 d1e1 e6d5 e4c5 f7g6 f2f3 e8e3 e1e3 f8d6 f4d6 c7d6 c5a6 b8e8 e3e8 g6e8');
INSERT INTO evaluations VALUES(3794,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(3794,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(3794,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(3794,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(3794,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(3794,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(3794,36,'cp -16','h7h6 g5d2 c6e7 d2e3 c5e3 f2e3 e7g6 b1c3 c7c6 c4b3 a7a5 d1e1 g8h7 a2a4 c8e6 f3d2 e6b3 d2b3 d6d5 f1f2 d5e4 c3e4 f6e4 d3e4 d8c7 e1e2 f7f6 b3d2 f8d8 b2b3 g6f8 d2f1 d8d7 f1g3 a8d8 a1f1 c7b6 g1h2 d7d2 e2e1 d2f2 e1f2');
INSERT INTO evaluations VALUES(3585,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(3585,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(3585,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(3585,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(3585,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(3585,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(3585,36,'cp -10','g5d2 f8e8 a2a4 d6d5 e4d5 f6d5 b1c3 d5c3 d2c3 c8e6 f3e5 c6e5 c4e6 e8e6 d3d4 e6d6 d1e2 c5d4 c3d4 d6d4 e2e5 d8d6 e5d6 d4d6 f1d1 a8d8 d1d6 d8d6 a1e1 g8f8 e1e4 d6d2 e4c4 c7c6 g1f1 f8e7 f1e1 d2d5 c4b4 b7b6 b2b3 b6b5 a4b5 c6b5 c2c4 b5c4 b4c4 e7d6 c4f4');
INSERT INTO evaluations VALUES(12333,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(12333,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(12333,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(12333,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(12333,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(12333,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(12333,36,'cp -60','h7h6 g5f4 a7a6 a2a4 d7d5 c4b3 d5e4 c3e4 f8e7 e4f6 e7f6 c2c3 c6e5 f4e5 f6e5 g1f3 e5f6 e1g1 b7b6 d1e2 e8g8 f3e5 d8c7 a1e1 a8b8 e5g4 f6g5 e2e5 g5f4 e5c7 f4c7 g4e5 g7g6 e5c6 b8a8 d3d4 c5d4 c3d4 g8g7 e1c1 c7d6 c6e5 c8b7 e5d7 f8d8 d7b6');
INSERT INTO evaluations VALUES(13121,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(13121,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(13121,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(13121,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(13121,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(13121,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(13121,36,'cp -18','f6e4 g5f4 d8a5 d1b3 e7e6 e2e3 f7f6 f4g3 c8d7 f3d2 e4c3 b2c3 a8c8 f1d3 c6d8 a1c1 f8a3 c1c2 a3e7 c2c1 d7a4 b3b2 a5b6 d2b3 a4b5 b3c5 b5d3 b2b6 a7b6 c5d3 d8c6 e1e2 e8d7 h2h4 c8a8 c1b1 a8a3 b1b6 a3a2 b6b2');
INSERT INTO evaluations VALUES(21629,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(21629,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(21629,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21629,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21629,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21629,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(21629,36,'cp 19','d2d4 g8f6 f1b5 f8b4 b5c6 b7c6 e1g1 e8g8 f3e5 c8a6 f1e1 f8e8 c1g5 d8d6 g5f6 d6f6 a2a3 b4c3 b2c3 f6d6 d1h5 e8e7 e1e3 f7f6 e5g4 a6c8 e3e7 d6e7 g4e3 c8e6 h5f3 e7d6 h2h4 a7a5 a1e1 a8e8 e3f5 d6d7 h4h5 h7h6');
INSERT INTO evaluations VALUES(50334,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(50334,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(50334,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(50334,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(50334,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(50334,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(50334,36,'cp -43','d7d5 e2e3 c7c5 g1f3 e7e6 f1e2 c5c4 d3d1 b7b5 b2b3 c8b7 c2c3 b8d7 e1g1 a7a6 a2a4 d8c7 d1c2 d7b6 a4b5 a6b5 a1a8 b7a8 c2a2 f8d6 b1d2 e8g8 c1a3 a8c6 a3d6 c7d6 f1a1 f8b8 f3e5 c6e8 a2a3 d6a3 a1a3 f6e4 d2e4');
INSERT INTO evaluations VALUES(778,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(778,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(778,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(778,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(778,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(778,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(778,36,'cp 0','e7f5 c1b2 c8d7 g2g4 f5e7 b1c3 h7h5 c3a4 b6c7 g4h5 e7f5 f1d3 h8h5 h2h4 c7d8 a4c5 f8c5 d4c5 f5h4 b4b5 c6a5 b2c3 a5c4 d3c4 d5c4 f3h4 h5h4 d1f3 h4h1 f3h1 d7b5 h1h8 e8e7 h8h4 g7g5 h4g5 e7e8 g5g8 e8e7');
INSERT INTO evaluations VALUES(14656,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(14656,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(14656,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(14656,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14656,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(14656,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(14656,36,'cp 61','f8e7 h2h3 e8g8 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 d7c6 f4e3 f6d7 d4d2 e7f6 e3d4 e6c4 e2c4 c6c4 b2b3 c4d4 d2d4 f6d4 d1d4 a8e8 f1d1 d7c5 a4a5 c5e6 d4b4 e6c5 f2f3 f7f5 b4c4 e8d8 c3d5');
INSERT INTO evaluations VALUES(8455,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(8455,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(8455,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(8455,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(8455,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(8455,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(8455,36,'cp 66','d4c6 b7c6 c4d3 a7a5 f2f4 a5a4 d1f3 e8g8 c1e3 f8e8 h2h3 c8a6 a1d1 d8c8 e4e5 f6d5 d3a6 c8a6 e3d4 d5c3 d4c3 c6c5 a2a3 e7f8 d1e1 a8d8 e5e6 e8e6 e1e6 f7e6 f4f5 e6f5 f3f5 a6c4 f5g6 c4d5 f1f8 g8f8 c3g7 f8e7 g7f6 e7d7 f6d8 d7d8');
INSERT INTO evaluations VALUES(14955,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(14955,31,'cp 387',NULL);
INSERT INTO evaluations VALUES(14955,32,'cp 368',NULL);
INSERT INTO evaluations VALUES(14955,33,'cp 375',NULL);
INSERT INTO evaluations VALUES(14955,34,'cp 378',NULL);
INSERT INTO evaluations VALUES(14955,35,'cp 375',NULL);
INSERT INTO evaluations VALUES(14955,36,'cp 378','f3e5 c8d7 e5d7 d8d7 e4e5 f6e4 d2d4 f7f6 f2f3 e4g5 f3f4 g5e4 c2c3 a7a6 b5a4 f6f5 b1d2 e4d2 c1d2 g7g6 d2e3 f8e7 c3c4 e8c8 c4c5 d7e6 d1d2 h7h5 b2b4 c6a7 a4d1 h5h4 a2a4 c8d7 d2c2 c7c6 d1e2 d8b8 c2d2 d7e8 h2h3 a7c8');
INSERT INTO evaluations VALUES(14945,30,'cp 367',NULL);
INSERT INTO evaluations VALUES(14945,31,'cp 378',NULL);
INSERT INTO evaluations VALUES(14945,32,'cp 362',NULL);
INSERT INTO evaluations VALUES(14945,33,'cp 364',NULL);
INSERT INTO evaluations VALUES(14945,34,'cp 394',NULL);
INSERT INTO evaluations VALUES(14945,35,'cp 374',NULL);
INSERT INTO evaluations VALUES(14945,36,'cp 378','c8d7 e5d7 d8d7 e4e5 f6g8 d2d4 a7a6 b5e2 g8e7 e2g4 e7f5 c2c3 g7g6 d1d3 c6d8 c1e3 f8e7 b1d2 h7h5 g4h3 d8e6 h3f5 g6f5 d3f5 e6g7 f5d7 e8d7 d2f3 a8g8 h2h3 g7f5 e3f4 h5h4 g1h2 f5g7 f4e3 g7f5 a1e1 g8g6 e3f4 h8g8 f1g1 f5g7 f4d2 g7h5');
INSERT INTO evaluations VALUES(15132,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15132,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15132,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15132,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(15132,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15132,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15132,36,'cp 0','e4d5 f6d5 f1g2 d5c3 b2c3 f8c5 e1g1 e8g8 f1e1 d8f6 d2d3 c5b6 c1e3 c8g4 h2h3 g4h5 g3g4 h5g6 f3d2 a8d8 a1b1 f6e7 d2c4 f7f6 d1e2 f8e8 b1b5 g6f7 g2e4 f7g6');
INSERT INTO evaluations VALUES(15002,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15002,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15002,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15002,36,'cp 0','f6d5 f1g2 d5c3 b2c3 f8c5 e1g1 e8g8 f1e1 c5b6 d2d4 e5d4 c1a3 d4c3 a3f8 d8d1 a1d1 g8f8 f3g5 c8f5 g2e4 f5g4 e4f3 g4f5');
INSERT INTO evaluations VALUES(15346,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15346,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15346,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15346,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15346,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15346,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15346,36,'cp 0','f1g2 d5c3 b2c3 f8c5 e1g1 e8g8 f1e1 c5b6 d2d4 e5d4 c1a3 d4c3 a3f8 d8d1 a1d1 g8f8 f3g5 c8f5 g2e4 f5g4 e4f3 g4f5');
INSERT INTO evaluations VALUES(15866,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(15866,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(15866,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(15866,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15866,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15866,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15866,36,'cp 0','c6d4 f3h5 g7g6 d5f7 e8e7 f7g6 h7g6 h5h8 d4c2 e1e2 d8d5 h8h7 e7d8 h7h8');
INSERT INTO evaluations VALUES(15881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15881,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15881,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15881,36,'cp 0','f3h5 g7g6 d5f7 e8e7 f7g6 h7g6 h5h8 d4c2 e1e2 d8d5 h8h7 e7d8 h7h8');
INSERT INTO evaluations VALUES(17473,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(17473,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(17473,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(17473,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(17473,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(17473,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(17473,36,'cp 49','b7c6 c1f4 g8f6 b1d2 d7d6 e4e5 f6d5 f4g3 d6e5 g3e5 c8f5 g2g4 f5g6 f1g2 f7f6 e5g3 h7h5 g4h5 g6h5 g2f3 g7g6 d1e2 e8f7 f3h5 h8h5 e1c1 d8a5 c1b1 a5b5 e2f3 h5f5 c2c4 f5f3 c4b5 f3f5 b5c6 a8c8 d2c4 c8c6 c4e5 f6e5 d1d5 f8h6 d5e5');
INSERT INTO evaluations VALUES(18192,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(18192,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(18192,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(18192,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(18192,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(18192,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(18192,36,'cp 48','b1c3 d7d6 f1c4 e7e6 d1e2 f8e7 e4e5 d6d5 c4d3 c6c5 b2b3 d8c7 e1g1 c8b7 f1e1 h7h5 c1f4 h5h4 h2h3 g8h6 e2d2 c5c4 b3c4 d5c4 d3e4 a8d8 d2e2 b7e4 c3e4 h6f5 f4g5 c7e5 g5e7 f5e7 e2c4 e8g8');
INSERT INTO evaluations VALUES(18191,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(18191,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(18191,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(18191,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(18191,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(18191,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(18191,36,'cp 38','e7e6 f1d3 d7d5 e1g1 g8f6 d1f3 f8e7 e4e5 f6d7 f3g3 d8c7 f1e1 g7g6 c1g5 f7f6 e5f6 c7g3 h2g3 d7f6 c3a4 e8f7 c2c4 a6a5 c4d5 c6d5 a1c1 h8d8 g5f4 c8a6 d3a6 a8a6 a4c5 e7c5 c1c5 a5a4 f4e5 d8e8 e5f6 f7f6 c5d5 e6d5 e1e8');
INSERT INTO evaluations VALUES(18844,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(18844,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(18844,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(18844,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(18844,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(18844,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(18844,36,'cp 37','g1e2 d7d5 e4d5 e6d5 e1g1 g8f6 d2d4 c5d4 e2d4 c8g4 d1d3 f8e7 b1c3 e8g8 h2h3 g4d7 c1f4 d8c8 f1e1 c6d4 d3d4 e7c5 d4d3 d5d4 c3e4 f6e4 g2e4 h7h6 d3f3 d7c6 e4c6 b7c6 g1g2 c8d7 e1e4 a7a5 a1d1 f8e8 c2c3 e8e4 f3e4 a8d8 c3d4 c5d4');
INSERT INTO evaluations VALUES(18845,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(18845,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(18845,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(18845,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(18845,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(18845,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(18845,36,'cp 31','d7d5 e4d5 e6d5 e1g1 g8f6 d2d4 c5d4 e2d4 f8e7 c2c4 e8g8 c4d5 f6d5 b1c3 d5c3 b2c3 c6d4 c3d4 e7d6 f1e1 a8b8 d1h5 d8f6 c1g5 f6g6 h5g6 h7g6 g5e7 d6e7 e1e7 f8d8 d4d5 g8f8 e7c7 c8f5 a1c1 a7a5 h2h3 f5d7 g3g4 b7b5 c7a7 a5a4 g4g5 b8a8 a7a8 d8a8');
INSERT INTO evaluations VALUES(31540,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(31540,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31540,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(31540,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(31540,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31540,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31540,36,'cp 0','f5g6 h2h4 h7h5 f1d3 g6d3 c2d3 d8b6 c1e3 b6b2 e1g1 g8e7 b1d2 b2a3 d2f3 e7g6 e3c1 a3a5 c1g5 b8d7 a1b1 b7b6 f3h2 f8e7 g5e7 g6e7 h2f3 g7g6 f3g5 a5a2 d1f3 h8f8 f3f4 a2a4 b1a1 a4c2 f4e3 a7a5 f1c1 c2b2 c1b1 b2c2');
INSERT INTO evaluations VALUES(31417,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31417,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31417,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31417,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31417,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31417,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31417,36,'cp 0','h2h4 h7h5 f1d3 g6d3 c2d3 d8b6 c1e3 b6b2 e1g1 g8e7 b1d2 b2a3 d1c2 b8a6 f1b1 b7b6 d2f3 a6b4 c2d1 e7g6 e3d2 c6c5 d4c5 b6c5 d3d4 b4c6 d4c5 a8c8 b1c1 f8e7 a1b1 e7h4 f3h4 g6h4 g3h5 h4f5 g2g4');
INSERT INTO evaluations VALUES(31414,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31414,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31414,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31414,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31414,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31414,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31414,36,'cp 0','h7h5 f1d3 g6d3 c2d3 d8b6 c1e3 b6b2 e1g1 g8e7 b1d2 b2a3 d2f3 e7g6 e3c1 a3a5 c1g5 b8d7 a2a4 a8c8 a1c1 d7b6 c1a1');
INSERT INTO evaluations VALUES(31367,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31367,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31367,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31367,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31367,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31367,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31367,36,'cp 0','f1d3 g6d3 c2d3 d8b6 c1e3 b6b2 e1g1 g8e7 b1d2 b2a3 d2f3 e7g6 e3c1 a3a5 c1g5 b8d7 a2a4 a8c8 a1c1 d7b6 c1a1');
INSERT INTO evaluations VALUES(32246,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32246,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32246,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32246,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32246,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(32246,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32246,36,'cp 0','g1h3 b8c6 c2c3 f8e7 h3f4 e6c8 h2h3 g8f6 c1e3 e7d6 e2f3 c6e7 b1d2 c7c6 f1d3 e7g6 g2g3 e8g8 f4h5 f6h5 f3h5 f8e8 e1c1 b7b5 h1e1 d8d7 h3h4 d7h3 d3f1 h3g4 f1e2 g4d7');
INSERT INTO evaluations VALUES(32739,30,'cp 310',NULL);
INSERT INTO evaluations VALUES(32739,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(32739,32,'cp 326',NULL);
INSERT INTO evaluations VALUES(32739,33,'cp 333',NULL);
INSERT INTO evaluations VALUES(32739,34,'cp 355',NULL);
INSERT INTO evaluations VALUES(32739,35,'cp 366',NULL);
INSERT INTO evaluations VALUES(32739,36,'cp 335','e1g1 g7g6 f1e1 e6d6 e2b5 b8d7 d2d4 f8g7 d1e2 e8f8 b5c4 h7h6 c1d2 c7c5 a1d1 c5d4 c3b5 d6c6 b5d4 g7d4 c4b5 c6d5 d2c3 d7e5 f3h4 e5d3 e2d3 d4c3 d3c3 d5b5 c3h8 a8e8 d1d4 b5g5 c2c3 b7c6 h8e5 g5e5 e1e5 g8f6 f2f3 e7e6 e5e1 f6d5');
INSERT INTO evaluations VALUES(32738,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(32738,31,'cp 317',NULL);
INSERT INTO evaluations VALUES(32738,32,'cp 328',NULL);
INSERT INTO evaluations VALUES(32738,33,'cp 358',NULL);
INSERT INTO evaluations VALUES(32738,34,'cp 351',NULL);
INSERT INTO evaluations VALUES(32738,35,'cp 359',NULL);
INSERT INTO evaluations VALUES(32738,36,'cp 343','g8f6 d2d4 e6d6 f3e5 e7e6 f1e1 f8e7 e2f3 c7c6 c3e4 f6e4 f3e4 b8d7 c1f4 d7e5 f4e5 d6d7 e5g7 h8g8 g7e5 e8c8 e4h7 c6c5 h7g8 d8g8 f2f3 f7f6 e5g3 c5d4 d1d3 e7c5 g1h1 b7d5 a1d1 c8b7 e1e4 e6e5 c2c3 g8g3 c3d4 d7h7 d4c5 d5e4 d3e4 h7e4 f3e4 g3g4 c5b6');
INSERT INTO evaluations VALUES(38315,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38315,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38315,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38315,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38315,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(38315,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(38315,36,'cp 21','f1d3 g8f6 h2h3 c7c5 d4c5 e7c5 b1c3 e8g8 e1g1 a7a6 a2a3 f8e8 b2b4 c5a7 f1e1 c8e6 c1e3 a7e3 e1e3 b8c6 c3e2 d8c7 d1d2 e6d7 e3e8 a8e8 d2f4 c7f4 e2f4 e8c8 g1f1 g7g6 f4e2 c6d8 f3d4 d8e6 f2f3 f6e8 h3h4 e8d6 f1f2 e6d4');
INSERT INTO evaluations VALUES(38313,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(38313,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38313,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(38313,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38313,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38313,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38313,36,'cp 15','g8f6 h2h3 e8g8 e1g1 c7c5 d4c5 e7c5 b1c3 a7a6 a2a3 f8e8 c1f4 b8c6 f1e1 c8e6 b2b4 c5f8 c3e2 f6e4 c2c4 e6f5 f4e3 d5c4 d3c4 d8d1 e1d1 e4d6 c4b3 a6a5 e2d4 a5b4 d4c6 b7c6 a3b4 a8a1 d1a1 e8b8 e3c5 d6e4 c5f8 g8f8');
INSERT INTO evaluations VALUES(44186,30,'cp -195',NULL);
INSERT INTO evaluations VALUES(44186,31,'cp -188',NULL);
INSERT INTO evaluations VALUES(44186,32,'cp -190',NULL);
INSERT INTO evaluations VALUES(44186,33,'cp -196',NULL);
INSERT INTO evaluations VALUES(44186,34,'cp -194',NULL);
INSERT INTO evaluations VALUES(44186,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(44186,36,'cp -198','b8c6 b1c3 c6d4 f3d1 d7d5 e4d5 e6d5 c4e2 d8e7 d2d3 e7e6 e1f1 f8e7 g1f3 d4f5 h2h3 e8g8 a2a4 d5d4 f3g5 e6d7 c3e4 f6d5 e2g4 h7h6 g5f3 d7c7 c1d2 b7b6 f1g1 c8e6 d1e2 a8e8 c2c4 d4c3 b2c3 a7a6 a1b1 e6d7 b1e1 b6b5');
INSERT INTO evaluations VALUES(12376,30,'cp -174',NULL);
INSERT INTO evaluations VALUES(12376,31,'cp -189',NULL);
INSERT INTO evaluations VALUES(12376,32,'cp -199',NULL);
INSERT INTO evaluations VALUES(12376,33,'cp -191',NULL);
INSERT INTO evaluations VALUES(12376,34,'cp -188',NULL);
INSERT INTO evaluations VALUES(12376,35,'cp -179',NULL);
INSERT INTO evaluations VALUES(12376,36,'cp -177','b1c3 c6d4 f3d1 d7d5 e4d5 e6d5 c4e2 d8e7 d2d3 c8d7 a2a4 g7g6 c3b5 d4b5 a4b5 f8g7 g1f3 e8g8 e1g1 d7b5 f1e1 f8e8 c2c3 e7d7 d3d4 e8e2 e1e2 b5e2 d1e2 c5c4 c1f4 a7a5 f4e5 d7e6 e2c2 b7b5 h2h4 f6e4 f3g5 e4g5 h4g5 a5a4 e5g7 g8g7 c2d1 e6c6 d1f3');
INSERT INTO evaluations VALUES(12353,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(12353,31,'cp -173',NULL);
INSERT INTO evaluations VALUES(12353,32,'cp -175',NULL);
INSERT INTO evaluations VALUES(12353,33,'cp -207',NULL);
INSERT INTO evaluations VALUES(12353,34,'cp -200',NULL);
INSERT INTO evaluations VALUES(12353,35,'cp -200',NULL);
INSERT INTO evaluations VALUES(12353,36,'cp -186','c6d4 f3d1 d7d5 e4d5 e6d5 c4e2 d8e7 d2d3 c8d7 a2a4 e8c8 c1g5 d8e8 e1f1 e7e6 g5f6 g7f6 e2h5 d4c6 g1f3 d5d4 c3e4 h7h6 e4d2 f8d6 d2c4 d6c7 f3d2 f6f5 d2b3 h8g8 h5f3 e6e7 d1e2 e7e2 f3e2 b7b6 a4a5 c6b4 a5b6 a7b6 a1a8 c8b7 a8e8 g8e8');
INSERT INTO evaluations VALUES(47973,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(47973,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(47973,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(47973,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(47973,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(47973,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(47973,36,'cp 93','c8e6 c4b3 b8d7 h2h3 e6b3 a2b3 h6g8 d3d4 c7c6 b1c3 g8f6 g2g4 h7h6 h1g1 f8e7 h3h4 h6h5 g4g5 f6g4 d4d5 d7f8 f3d2 f8g6 g1h1 f7f6 d2c4 f6g5 h4g5 e8g8 h1h5 g4f2 d1e2 g6f4 c1f4 f8f4 d5c6 b7c6 c4e5 e7g5');
INSERT INTO evaluations VALUES(17290,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(17290,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(17290,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(17290,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(17290,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(17290,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(17290,36,'cp 32','g1f3 d7d6 d2d4 c8f5 e2e3 c6b4 b1a3 e7e6 c2c3 b4d5 f1d3 c7c6 e1g1 f8e7 c3c4 d5b6 a3c2 f5d3 d1d3 d6d5 c4c5 b6d7 c2e1 e8g8 d3c2 d8c7 e1d3 a7a5 f1c1 f8e8 b2c3 f6e4 a2a4 b7b6 c5b6 c7b6 c3b2 d7f6 h2h3 e8c8');
INSERT INTO evaluations VALUES(17289,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(17289,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(17289,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(17289,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(17289,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(17289,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(17289,36,'cp 22','d7d6 d2d4 c8f5 e2e3 c6b4 b1a3 e7e6 c2c3 b4c6 f1d3 d6d5 e1g1 f8e7 c3c4 c6b4 d3f5 e6f5 a3c2 e8g8 c2e1 f8e8 f3e5 c7c6 a2a3 b4a6 d1c2 g7g6 e1d3 f6d7 g2g3 a6c7 f1d1 d7e5 d4e5 c7e6 a3a4 d5c4 b3c4');
INSERT INTO evaluations VALUES(20662,30,'cp -595',NULL);
INSERT INTO evaluations VALUES(20662,31,'cp -582',NULL);
INSERT INTO evaluations VALUES(20662,32,'cp -601',NULL);
INSERT INTO evaluations VALUES(20662,33,'cp -598',NULL);
INSERT INTO evaluations VALUES(20662,34,'cp -598',NULL);
INSERT INTO evaluations VALUES(20662,35,'cp -623',NULL);
INSERT INTO evaluations VALUES(20662,36,'cp -609','d8g5 e1g1 g5d8 c2c3 g8f6 d2d3 f8e7 c4b3 e8g8 b1d2 g8h8 f3e2 e6e5 h2h3 c8e6 b3e6 f7e6 d2f3 f6g8 b2b4 c5b4 c3b4 c6b4 a1b1 b4c6 c1e3 d8c8 g1h1 g8f6 e2d1 d6d5 e4d5 e6d5 e3g5 e5e4 g5f6');
INSERT INTO evaluations VALUES(1570,30,'cp -616',NULL);
INSERT INTO evaluations VALUES(1570,31,'cp -598',NULL);
INSERT INTO evaluations VALUES(1570,32,'cp -605',NULL);
INSERT INTO evaluations VALUES(1570,33,'cp -595',NULL);
INSERT INTO evaluations VALUES(1570,34,'cp -588',NULL);
INSERT INTO evaluations VALUES(1570,35,'cp -599',NULL);
INSERT INTO evaluations VALUES(1570,36,'cp -605','e1g1 g5d8 c2c3 g8f6 d2d3 f8e7 f3e2 e8g8 c1g5 g8h8 b1d2 f6g8 f2f4 e7g5 f4g5 d8g5 d2f3 g5d8 a1d1 c6e5 f3e5 d6e5 h2h3 c8d7 g1h1 d7c6 c4b3 a8c8 b3c2 d8g5 a2a3 g8f6 c3c4 f6h5 e2g4 g5g4');
INSERT INTO evaluations VALUES(1571,30,'cp -619',NULL);
INSERT INTO evaluations VALUES(1571,31,'cp -612',NULL);
INSERT INTO evaluations VALUES(1571,32,'cp -614',NULL);
INSERT INTO evaluations VALUES(1571,33,'cp -622',NULL);
INSERT INTO evaluations VALUES(1571,34,'cp -615',NULL);
INSERT INTO evaluations VALUES(1571,35,'cp -625',NULL);
INSERT INTO evaluations VALUES(1571,36,'cp -609','g5d8 d2d3 f8e7 f3e2 g8f6 c2c3 e8g8 b1d2 g8h8 d2f3 c8d7 a2a3 f6g8 b2b4 c6e5 f3e5 d6e5 f2f4 b7b5 c4a2 c5b4 a3b4 a7a5 b4a5 d8a5 g1h1 a5c3 c1b2 c3c8 f4f5 f7f6 a2b3 e6f5 a1a8 c8a8');
INSERT INTO evaluations VALUES(22141,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(22141,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(22141,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(22141,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(22141,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(22141,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(22141,36,'cp 33','g8f6 b1c3 c7c6 a2a4 f8e7 h2h3 e8g8 e1g1 c8e6 f1d1 d6d5 e4d5 f6d5 c3d5 d8d5 d4f4 d5h5 c1e3 e7f6 f4b4 b7b6 a4a5 b6a5 b4d2 f6d8 d1e1 d8b6 b2b4 b6e3 e1e3 a5b4 d2b4 e6d5 b4a3 h5g5 e3g3');
INSERT INTO evaluations VALUES(15770,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(15770,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(15770,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(15770,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(15770,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(15770,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(15770,36,'cp 30','e1g1 c7c6 a2a4 f8e7 b1c3 e8g8 f1d1 c8e6 a4a5 b7b5 a5b6 d8b6 d3f1 d6d5 e4d5 f6d5 c3d5 c6d5 a1a6 b6d4 d1d4 a8b8 d4a4 e7f6 a4a2 e6f5 c1f4 b8c8 b2b4 f5c2 a6a7 f6d4 a7c7 c8c7 f4c7 c2e4 c7f4 f8c8 b4b5 d4b6');
INSERT INTO evaluations VALUES(15771,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15771,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(15771,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(15771,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(15771,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(15771,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(15771,36,'cp 43','c7c6 b1c3 f8e7 a2a4 e8g8 a4a5 c8e6 f1d1 b7b5 a5b6 d8b6 d3f1 d6d5 e4d5 f6d5 c3d5 c6d5 c1e3 b6d4 e3d4 a7a5 a1a2 f8c8 f1d3 a5a4 d1a1 a4a3 b2a3 a8a4 c2c3 c8a8 a2b2 e7f6 a1b1 a8c8 b2b8 f6d4 b8c8 e6c8 b1b8 d4c5 b8c8 c5f8 d3h7 g8h7 c8f8');
INSERT INTO evaluations VALUES(26346,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(26346,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(26346,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(26346,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(26346,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(26346,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(26346,36,'cp 69','f1b5 g7g6 e1g1 f8g7 b5c6 b7c6 d2d3 g8f6 a1b1 e8g8 h2h3 g4f3 d1f3 f6d7 c1d2 e7e6 g1h1 d8c7 f4f5 a8e8 d2f4 d7e5 f3f2 e6f5 e4f5 c7d7 c3e4 g6f5 e4g3 e5g6 f4d2 f5f4 g3h5 g7e5 h5f4 f7f5 b2b3 f8f6 b1d1 f6f8 f2f3 f8f7 d1e1 e8f8');
INSERT INTO evaluations VALUES(26319,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(26319,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(26319,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(26319,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(26319,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(26319,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(26319,36,'cp 71','g7g6 d2d3 g8f6 h2h3 g4f3 d1f3 f6d7 b5c6 b7c6 c1d2 f8g7 b2b3 d8a5 c3a4 a5a6 a1c1 d7b6 a4c3 a6a3 e4e5 e8g8 e1g1 d6e5 f4e5 g7e5 c1e1 e5f6 e1e2 a3b2 c3e4 f6g7 e4c5 g7d4 d2e3 d4e3 f3e3 b6d5 e3e5 b2e5 e2e5 e7e6');
INSERT INTO evaluations VALUES(27433,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(27433,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(27433,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(27433,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(27433,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(27433,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(27433,36,'cp 175','c3b5 a8c8 e6f8 e7f8 e1g1 f8f7 d2d3 e8g8 f2f3 h7h6 c1e3 d6d5 e4d5 f6d5 d1d2 d5e3 d2e3 c8d8 f1f2 c6d4 b5c3 f7g6 g1h1 f8e8 c3e4 a7a5 a2a3 b7b5 e3e1 g6b6 h2h3 b5b4 a3b4 a5b4 h1h2 g8h8 f2d2');
INSERT INTO evaluations VALUES(27512,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27512,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27512,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27512,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27512,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27512,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27512,36,'cp 0','b5c6 b7c6 d4c5 b6c5 e1g1 g8e7 c3e2 e7f5 e2g3 f5g3 h2g3 a7a5 b2b3 a5a4 c1e3 c5a5 c2c4 f8e7 a2a3 a4b3 d1b3 e8g8 b3b7 a5d8 a3a4 a8b8 b7a7 b8a8');
INSERT INTO evaluations VALUES(27509,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27509,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27509,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27509,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27509,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27509,36,'cp 0','b7c6 d4c5 b6c5 e1g1 g8e7 c3e2 e7f5 b2b3 a7a5 c2c4 d5c4 g2g4 f5e7 f3d2 e7g6 d2c4 g6e5 c1e3 e5f3 g1g2 f3h4 g2g1');
INSERT INTO evaluations VALUES(32366,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(32366,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(32366,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(32366,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(32366,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(32366,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(32366,36,'cp 58','g8f6 b1c3 c7c6 e1g1 b8d7 d1e2 f6e4 f3d2 e4d2 c1d2 f5g6 e3e4 d7b6 c4d3 f8e7 d2e3 e8g8 d3c2 f7f5 f2f3 a7a5 c2b3 g6f7 e2c2 g8h8 a1e1 e7d6 a2a3 d8c7 h2h3 f7g8 e3f2 a8d8 g1h1 d6e7 c2e2 e7f6 e4e5 f6e7');
INSERT INTO evaluations VALUES(33887,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(33887,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(33887,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(33887,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(33887,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(33887,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(33887,36,'cp 72','c2c4 d5d8 b1c3 c7c5 c1e3 c5d4 f3d4 a7a6 g2g3 e6e5 d4b3 d8d1 a1d1 b8c6 f1g2 f8b4 e1g1 b4c3 b2c3 c8g4 d1b1 g4f5 b1b2 f5e4 b3c5 e4g2 g1g2 c6a5 b2b4 b7b5 c4b5 a6b5 b4b5 e8g8 e3g5 a5c4 g5f6 g7f6 c5d7 a8a2 d7f6 g8g7');
INSERT INTO evaluations VALUES(33883,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(33883,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(33883,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(33883,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(33883,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(33883,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(33883,36,'cp 74','d5d8 b1c3 c7c5 c1e3 c5d4 f3d4 a7a6 g2g3 e6e5 d4e2 d8c7 a1c1 f8c5 c3d5 f6d5 c4d5 c7a5 d1d2 a5d2 e1d2 c5e3 d2e3 b8d7 f1g2 e8g8 h1d1 f8d8 h2h3 d7f6 g3g4 h7h6 e2g3 a8b8 b2b3 g8f8 d5d6 f6e8 g3e4 c8e6 c1c7 d8d6 c7f7 f8f7 e4d6 e8d6 d1d6');
INSERT INTO evaluations VALUES(36320,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(36320,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(36320,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(36320,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(36320,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(36320,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(36320,36,'cp -11','d5c4 e1g1 a7a6 a2a4 c7c5 d4c5 b8c6 e2c4 d8d1 f1d1 e7c5 c1d2 b7b6 d2e1 c8b7 c4e2 f8d8 a1c1 d8d1 c3d1 f6d7 f3d2 a8c8 d2b3 c5e7 a4a5 b6b5 f2f3 c6b4 c1c8 b7c8 g2g3 c8b7 e3e4 f7f5 d1f2 f5e4 f2e4');
INSERT INTO evaluations VALUES(36376,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(36376,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(36376,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(36376,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(36376,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(36376,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(36376,36,'cp -12','e1g1 a7a6 e2c4 b7b5 c4d3 c8b7 e3e4 c7c5 d4c5 b8d7 b2b4 a6a5 c1a3 a5b4 a3b4 d7c5 d3b5 d8d1 f1d1 c5e4 b4e7 e4c3 e7f8 g8f8 a2a4 b7f3 g2f3 c3d1 a1d1 f6d5 d1d4 a8a5 d4c4 g7g5 h2h4 g5h4 c4h4 d5c3 h4h7 c3b5 a4b5 a5b5 h7h8 f8g7 h8a8 b5b1 g1h2 b1d1 a8a7');
INSERT INTO evaluations VALUES(38793,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(38793,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(38793,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(38793,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(38793,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(38793,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(38793,36,'cp 11','b8c6 a2a3 b4c3 e2c3 d5e4 d4d5 e6d5 d1d5 c8e6 d5e4 g8f6 e4a4 d8d4 f1b5 d4a4 b5a4 e8c8 a4c6 b7c6 e1g1 e6f5 c1e3 a7a5 a1c1 f6d5 c3d5 d8d5 f1e1 h8d8 f2f3 c6c5 g1f2 a5a4 h2h4 h7h5 f2g3 d8d7 c2c4 d5e5 e3f4 e5e1 c1e1');
INSERT INTO evaluations VALUES(39034,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(39034,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(39034,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(39034,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(39034,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(39034,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(39034,36,'cp 78','b8c6 c1e3 d7d6 h2h3 a7a6 b1c3 g8f6 d1d2 e8g8 e1c1 e6e5 d4e5 d6e5 d3c4 d8d2 f3d2 c6d4 f2f4 d4c6 c4d5 c8e6 d2f3 e5f4 e3f4 a8c8 h1e1 f8d8 d5c6 d8d1 e1d1 b7c6 f3g5 e6c4 d1d4 f6h5 d4c4 h5f4 g2g3');
INSERT INTO evaluations VALUES(42825,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(42825,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(42825,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(42825,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(42825,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(42825,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(42825,36,'cp 112','e4e5 d7d5 e5f6 d5c4 d1d8 e8d8 f6g7 h8g8 c1g5 d8e8 e1c1 f7f6 g5f6 e8f7 c3d5 c7c6 d5b4 f7f6 g1f3 f6g7 f3e5 a7a5 b4c2 b8a6 h1e1 c8e6 e5c6 a6c5 c6e5 g7h8 g2g3 g8d8 c2d4 e6g8 c1b1 a8a6 d4f5 c5d3 e1e3 d8d5 e5d3 c4d3 g3g4 a6g6 f2f3 g8e6 e3d3 e6f5 g4f5 d5f5');
INSERT INTO evaluations VALUES(43341,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(43341,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(43341,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(43341,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(43341,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(43341,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(43341,36,'cp 69','c1e3 f8g7 f2f3 b7b5 c4b3 b8d7 d1d2 h7h5 e1g1 d8c7 f1c1 c8b7 a2a4 b5a4 c3a4 e8g8 c2c4 h5h4 h2h3 f6h5 b3d1 h5g3 b2b4 e7e6 d1b3 f8d8 a1b1 d7c5 b3c2 c5d7 a4c3 d7f6 d4e2 f6h5 e2g3 h5g3 c2b3 a8c8 c3e2 d8d7 e2g3');
INSERT INTO evaluations VALUES(43339,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(43339,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(43339,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(43339,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(43339,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(43339,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(43339,36,'cp 51','f8g7 f2f3 b8d7 d1d2 b7b5 c4b3 h7h5 e1g1 c8b7 a2a4 b5a4 c3a4 a8b8 g1h1 e8g8 a4c3 d7c5 b3a2 b7a8 a1b1 d8c7 h2h3 a6a5 d4b5 c7d7 a2c4 a8c6 b5d4 e7e6 f1d1 f8d8 e3h6 g7h8 b2b3 a5a4');
INSERT INTO evaluations VALUES(39755,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(39755,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(39755,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(39755,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(39755,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(39755,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(39755,36,'cp 43','f2f3 b8d7 d1d2 b7b5 c4b3 h7h5 e1g1 c8b7 a2a4 b5a4 c3a4 a8b8 f1d1 d8c7 d4e2 e8g8 e2f4 d7e5 e3b6 c7c8 b6a7 b8a8 a7e3 a8b8 g1h1 f6d7 b3d5 b7c6 d5c6 c8c6 f4d5 f8e8 e3g1 e5c4 d2g5 c4b2 a4b2 g7b2 d5e7 e8e7 g5e7 b2a1 d1a1 d7e5 f3f4');
INSERT INTO evaluations VALUES(39756,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(39756,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(39756,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(39756,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(39756,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(39756,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(39756,36,'cp 51','b7b5 c4b3 b8d7 d1d2 h7h5 e1g1 c8b7 a2a4 b5a4 c3a4 a8b8 f1d1 d8c7 g1h1 e8g8 a4c3 f8e8 e3h6 e7e6 h6g7 g8g7 b3a4 e8c8 d4e2 b7c6 a4c6 c7c6 b2b3 d6d5 e4d5 f6d5 c3d5 e6d5 e2d4 c6f6 a1a5 c8c5 a5c5 d7c5 d4e2');
INSERT INTO evaluations VALUES(49342,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(49342,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(49342,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(49342,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(49342,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(49342,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(49342,36,'cp -27','d8f6 g1f3 c7c5 e2e3 c5d4 e3d4 b8c6 f1b5 a7a6 b5c6 b7c6 e1g1 c6c5 d4d5 f6b2 b1d2 b2b7 d5e6 d7e6 d2c4 a8a7 f3e5 f8e7 c4d6 e7d6 d1d6 b7c7 f1d1 c7d6 d1d6 a7c7 a2a4 e8e7 a1d1 g7g5 e5c6 e7f6 c6b8 h7h5 a4a5 c8b7 b8a6');
INSERT INTO evaluations VALUES(33963,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(33963,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(33963,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(33963,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(33963,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(33963,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(33963,36,'cp -30','g1f3 c7c5 e2e3 b8c6 f1b5 c5d4 e3d4 a7a6 b5c6 b7c6 e1g1 c6c5 d4d5 f6b2 b1d2 b2b7 d5e6 d7e6 d2c4 a8a7 f3e5 f8e7 c4d6 e7d6 d1d6 b7c7 a1d1 c7d6 d1d6 a7c7 a2a4 c5c4 f1b1 e8e7 d6b6 f7f6 e5c6 e7f7 c6b8 h8d8 c2c3 e6e5 b8a6');
INSERT INTO evaluations VALUES(55889,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(55889,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(55889,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55889,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(55889,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(55889,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(55889,36,'cp -60','d5d4 c3e2 c7c5 d2d3 b8c6 f3f4 f7f5 e2g3 f8d6 g1h3 d6c7 f1e2 g8e7 e2f3 e8g8 e1g1 h7h6 c1d2 a7a5 d1e2 a5a4 c2c4 g8h8 b2b3 a8a6 e4e5 c7a5 d2a5 d8a5 b3a4 a5a4 f1b1 c6b4 a2a3 b4c2 a1a2 c2e3 f3b7 c8b7 b1b7');
INSERT INTO evaluations VALUES(56114,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(56114,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(56114,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(56114,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(56114,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(56114,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(56114,36,'cp -87','b8c6 f1e2 a7a6 b5a3 f8a3 b2a3 g8e7 e1g1 h7h6 a3a4 d8d6 d2d3 c8e6 a1b1 e8c8 c1d2 g7g5 c2c3 d4c3 d2c3 e7g6 g2g3 g5g4 f3d2 h6h5 a4a5 h5h4 d2c4 d6c5 d1b3 h4g3 h2g3 h8h3 b3b7 c8d7 g1g2 d8h8 f1h1 h3h1 b1h1 h8h1 g2h1 c5f2 c4b6 d7e7 b7c6 f2e2 b6d5 e7f8');
INSERT INTO evaluations VALUES(21550,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(21550,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(21550,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(21550,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(21550,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(21550,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(21550,36,'cp -68','f1e2 a7a6 b5a3 f8a3 b2a3 g8e7 e1g1 e8g8 a3a4 d8d6 d2d3 a6a5 c1d2 h7h6 f3h4 g7g5 h4f3 e7g6 h2h3 c8e6 f3h2 d6d7 e2h5 g8g7 f1e1 b7b6 h5g6 g7g6 h2f1 f7f6 f1g3 g6h7 a2a3 f8f7 c2c3 a8g8 a1b1 g5g4');
INSERT INTO evaluations VALUES(548,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(548,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(548,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(548,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(548,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(548,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(548,36,'cp -40','b1c3 c6e7 e1g1 e8g8 c3a4 c5d6 a4c3 c7c6 d3d4 e7g6 c4b3 e5d4 d1d4 f6e7 a1d1 d6c7 f1e1 d7d6 h2h3 g6e5 f3e5 d6e5 d4d3 f8d8 d3f3 c8d7 d1d3 d7e8 e1d1 d8d3 d1d3 a8d8 c3d1 d8d4 d1e3 d4d3 c2d3 c7b6 e3f5 e7b4');
INSERT INTO evaluations VALUES(546,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(546,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(546,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(546,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(546,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(546,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(546,36,'cp -46','c6e7 e1g1 e8g8 c3a4 c5d6 d3d4 e7g6 a4c3 c7c6 c4b3 e5d4 d1d4 f6d4 f3d4 d6c5 a1d1 d7d6 f1e1 f8d8 h2h3 g8f8 a2a4 c8d7 c3e2 d8e8 e2g3 a8d8 c2c3 g6f4 e1e3 g7g5 e3f3 a7a5 d4e2 f4e2 g3e2 d6d5 e2g3 g5g4 h3g4');
INSERT INTO evaluations VALUES(3902,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(3902,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(3902,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(3902,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(3902,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(3902,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(3902,36,'cp 188','a7a5 e5f3 a5a4 h2h3 b7b5 f1e1 h7h6 a2a3 f8e8 c1f4 c5f8 b1c3 c8d7 f3e5 c6c5 d1f3 b5b4 c3d5 f6d5 e4d5 f7f6 e5g6 f8d6 f3g3 g8h7 f4d6 c7d6 g6f4 e8e5 f4e6 d7e6 e1e5 d6e5 d5e6 b4a3 a1a3 d8d4 g3f3 a8f8 c2c3 d4d6 a3a4 d6e6 a4a7');
INSERT INTO evaluations VALUES(9364,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9364,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9364,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9364,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9364,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9364,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9364,36,'cp 0','e8g8 e1g1 a7a5 c1e3 c5e3 f2e3 c6e7 b1c3 e7g6 a3a4 c8e6 b2b3 c7c6 c4e6 f7e6 c3e2 h7h6 d1d2 f6d7 g1h2 d8e7 f1f2 f8f7 a1f1 b7b5 e2g3 b5a4 b3a4 d7c5 f1a1 c5d7 a1f1');
INSERT INTO evaluations VALUES(3845,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3845,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3845,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(3845,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(3845,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3845,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(3845,36,'cp 5','e1g1 h7h6 b1c3 a7a6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 f1f2 c6e7 d1e2 f8f7 a1d1 d8d7 d3d4 e5d4 e3d4 f6h5 g1h2 a8f8 g2g3 h5f6 d1f1 f6h7 e4e5 d6e5 d4e5 e7f5 c3e4 f5d4 e4c5 d7c6 f3d4 f7f2 f1f2 c6c5');
INSERT INTO evaluations VALUES(19231,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(19231,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(19231,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(19231,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(19231,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(19231,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(19231,36,'cp 69','d2d4 e7e6 d4c5 f8c5 d1e2 c6e5 b1d2 e5c4 d2c4 b7b5 c4e5 a7a5 g1f3 c8a6 b2b3 d7d6 e5d3 c5b6 e4e5 f7f6 e5f6 g8f6 e2e6 d8e7 e6e7 e8e7 c1g5 h7h6 g5f6 e7f6 e1c1 f6f7 h1e1 h8c8 c1b2 c8f8 f3d4 b6d4 c3d4 b5b4 d3f4');
INSERT INTO evaluations VALUES(21602,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21602,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21602,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21602,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21602,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(21602,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21602,36,'cp 6','a2a3 c8f5 b1c3 g8f6 g1f3 f8d6 f3h4 f5e6 d1c2 e8g8 h4f5 e6f5 c2f5 f8e8 f1e2 a7a6 e1g1 c6a7 e2d3 c7c6 a3a4 a7c8 g1h1 c8e7 f5h3 c6c5 d4c5 d6c5 d3e2 e7g6 f1d1 d5d4 e3d4');
INSERT INTO evaluations VALUES(22787,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(22787,31,'cp -171',NULL);
INSERT INTO evaluations VALUES(22787,32,'cp -177',NULL);
INSERT INTO evaluations VALUES(22787,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(22787,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(22787,35,'cp -194',NULL);
INSERT INTO evaluations VALUES(22787,36,'cp -166','d7d5 d2d4 e5d4 c3d4 d5e4 d4d5 f8b4 b1c3 d8e7 f1e2 c6e5 f3e4 b4c3 b2c3 e7h4 e1f1 h4e4 d1d4 g8f6 d4e4 f6e4 g1f3 e5g4 c3c4 e8g8 c1f4 c7c6 d5c6 b7c6 f1g1 g4f6 f3g5 e4g5 f4g5 c8g4 e2f3 g4f3 g2f3 f6d7 g5f4 f8d8 g1f2 f7f6 h1d1 d7f8 d1d8');
INSERT INTO evaluations VALUES(28824,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28824,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28824,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28824,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28824,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28824,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28824,36,'cp 0','e7e6 e4e5 f6d5 c3e4 f7f5 e5f6 d5f6 c2c3 f6e4 d3e4 b7e4 f3g5 e4b7 d1c2 g7g6 g5h7 h8h7 c2g6 h7f7 c1g5 f8e7 h2h4 b8c6 g6g8 f7f8 g8g6');
INSERT INTO evaluations VALUES(32832,30,'cp 282',NULL);
INSERT INTO evaluations VALUES(32832,31,'cp 270',NULL);
INSERT INTO evaluations VALUES(32832,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(32832,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(32832,34,'cp 301',NULL);
INSERT INTO evaluations VALUES(32832,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(32832,36,'cp 274','f2f3 g4h5 g2g4 h5g6 f3f4 f7f5 g1f3 e6b6 e1g1 g8f6 g4g5 f6e4 c3a4 b6d6 c2c4 g6h5 c1e3 e7e6 a4c3 e4c3 b2c3 b8c6 d4d5 c6d8 d1e1 f8e7 a1d1 d6a3 e3d4 e8g8 f3e5 h5e2 e1e2 e6d5 e2h5 d8e6 g5g6 h7g6 e5g6 e6d4 c3d4 d5c4');
INSERT INTO evaluations VALUES(46883,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(46883,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(46883,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(46883,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(46883,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(46883,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(46883,36,'cp 100','c4d5 f6d5 g1f3 d5c3 b2c3 c7c5 a1b1 c5d4 c3d4 e7e6 e2e4 f8e7 f1d3 e8g8 e1g1 b7b6 c1f4 c8a6 d3a6 b8a6 f3e5 e7d6 b1c1 a8c8 d1e2 c8c1 f1c1 a6b8 e2g4 d6e5 f4e5 d8g5 g4g5 h6g5 f2f3 f8d8 e5b8 d8b8 g1f2 b8d8 f2e3 d8d7 a2a4 g8h7');
INSERT INTO evaluations VALUES(51087,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(51087,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51087,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(51087,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(51087,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(51087,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(51087,36,'cp 48','e7e6 b1c3 d7d5 g1f3 g8f6 c1f4 f8d6 f4d6 d8d6 e2e3 e8g8 a1c1 d6e7 f1e2 c7c6 e1g1 b8d7 f1e1 h7h6 h2h3 f8e8 d1c2 b7b6 c4d5 e6d5 c3a4 c8b7 e2d3 a8c8 b2b3 c6c5 c2b2 e8d8 d3f5 g7g6 f5d7 f6d7');
INSERT INTO evaluations VALUES(53705,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(53705,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(53705,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(53705,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(53705,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(53705,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(53705,36,'cp 225','f3e5 d8d5 d2d4 g8f6 c2c4 f8b4 b1c3 d5a5 c1d2 a5d8 d2e3 e8g8 f1d3 b8d7 e5f3 c6c5 e1g1 c5d4 e3d4 b7b6 a2a3 b4c3 d4c3 d7c5 d3c2 c8b7 f1e1 d8d1 a1d1 b7f3 g2f3 c5e6 b2b3 f8d8 g1f1 e6g5 c3f6 g7f6 c2e4 g5e4 f3e4 g8g7 f1g2 a7a5 d1d8 a8d8');
INSERT INTO evaluations VALUES(53687,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(53687,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(53687,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(53687,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(53687,34,'cp 230',NULL);
INSERT INTO evaluations VALUES(53687,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(53687,36,'cp 210','d8d5 d2d4 f8b4 c2c3 b4d6 e5f3 g8f6 f1e2 e8g8 e1g1 f8e8 f1e1 c8g4 h2h3 g4h5 c3c4 d5a5 c1d2 a5c7 c4c5 d6f8 b1c3 b8d7 a2a3 a7a5 d2e3 b7b6 b2b4 c7b7 a1c1 a5b4 a3b4 h7h6 e2d3 b6c5 b4c5 a8a3 d1c2');
INSERT INTO evaluations VALUES(259,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(259,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(259,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(259,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(259,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(259,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(259,36,'cp 200','d2d4 e5d4 e4e5 f6d8 c3d4 c5b4 b1c3 a7a6 b5a4 d7d5 e5d6 d8d6 e1g1 g8e7 d4d5 b7b5 a4b5 a6b5 c3b5 d6d7 d5c6 d7c6 a2a4 e8g8 c1f4 c8b7 f4c7 c6e6 h2h3 e7f5 b5d4 f5d4 d1d4 b7f3 g2f3 b4a5 c7g3 e6b3 a1a3 b3b7 d4e5 f8c8 e5b5 b7b5 a4b5');
INSERT INTO evaluations VALUES(2053,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(2053,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(2053,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(2053,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(2053,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(2053,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(2053,36,'cp 110','g8f6 b1c3 d7d6 c3d5 e7d8 f1e1 f8e7 h2h3 e8g8 d5e7 d8e7 d2d4 e5d4 f3d4 c8d7 c1f4 f8e8 d4b5 d7e6 c4f1 a7a6 b5c3 a8d8 f4g5 h7h6 g5h4 g7g5 h4g3 c6e5 f2f4 g5f4 g3f4 g8h8 f4h6 e8g8 h6f4 e6h3 c3d5 f6d5 e4d5');
INSERT INTO evaluations VALUES(1108,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(1108,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(1108,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(1108,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(1108,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(1108,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(1108,36,'cp 115','b1c3 d7d6 c3d5 e7d8 f1e1 f8e7 h2h3 e8g8 d5e7 d8e7 d2d4 e5d4 f3d4 c8d7 c1f4 f8e8 d4b5 c6e5 c4f1 e5g6 f4h2 d7b5 f1b5 c7c6 b5f1 d6d5 e4d5 e7e1 d1e1 e8e1 a1e1 f6d5 c2c4 d5f6 e1d1 g6f8 f2f3 f8e6 g1f2 a7a5 h2e5 a5a4 d1d6 h7h6 a2a3 e6c5');
INSERT INTO evaluations VALUES(2665,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(2665,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(2665,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(2665,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(2665,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(2665,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(2665,36,'cp 17','d3d4 c5a7 f3e5 c6e5 d4e5 f6e4 c4d5 e4c5 c1f4 c7c6 d5b3 c5b3 d1b3 a7c5 c3c4 f7f6 b1d2 f6e5 f4e5 d7d6 e5g3 d8b6 b3b6 c5b6 g3d6 f8d8 c4c5 b6c5 d6c5 d8d2 f1d1 d2d5 d1d5 c6d5 f2f3 c8e6 a1e1 a8c8 c5b6 c8c6 b6a5 g8f7 a5c3 d5d4 c3d4 e6a2');
INSERT INTO evaluations VALUES(6640,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(6640,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(6640,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(6640,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(6640,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(6640,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(6640,36,'cp 75','b2c3 d7d6 f1b5 c8d7 d4e5 d6e5 d1d3 f7f6 c1a3 a7a6 b5c4 g8e7 e1c1 d8c8 c4d5 e7d5 e4d5 c6e7 h1e1 e7g6 f3d2 e8f7 g2g3 h8d8 c3c4 d7f5 d2e4 b7b6 h2h4 h7h5 d3e2 c8d7 e4c3 f5g4 f2f3 g4f5 c4c5 b6c5 a3c5 a8b8 e2c4');
INSERT INTO evaluations VALUES(11422,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11422,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11422,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11422,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11422,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11422,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(11422,36,'cp 0','b7b5 a4b3 c6a5 d2d4 e5d4 c3d5 a5b3 a2b3 c7c5 c1g5 f8e7 d5e7 d8e7 b3b4 h7h6 g5h4 e7e6 e4e5 d6e5 f1e1 f6d7 f3e5 d7e5 d1h5 e8g8 e1e5 e6c4 h4e7 f8e8 e7c5 c8b7 f2f3 d4d3 c2d3 c4d3 a1e1 e8e5 h5e5 a8d8 e5e7 b7c6 e7c7 d3d2 g1f1 d2d3');
INSERT INTO evaluations VALUES(11647,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11647,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11647,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11647,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11647,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11647,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11647,36,'cp 0','a4b3 c6a5 d2d4 e5d4 c3d5 a5b3 a2b3 c7c5 c1g5 f8e7 d5e7 d8e7 b3b4 h7h6 g5h4 e7e6 f1e1 f6d7 e4e5 d6e5 f3e5 d7e5 d1h5 e8g8 e1e5 e6c4 h4e7 f8e8 e7c5 c8b7 f2f3 d4d3 c2d3 c4d3 a1e1 e8e5 h5e5 a8d8 e5e7 d3d2 e1e2 d2d1 e2e1');
INSERT INTO evaluations VALUES(11638,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11638,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11638,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11638,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11638,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11638,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11638,36,'cp 0','c6a5 d2d4 e5d4 c3d5 a5b3 a2b3 c7c5 c1g5 f8e7 d5e7 d8e7 b3b4 h7h6 g5h4 e7e6 f1e1 f6d7 e4e5 d6e5 f3e5 d7e5 d1h5 e8g8 e1e5 e6c4 h4e7 f8e8 e7c5 c8b7 f2f3 d4d3 c2d3 c4d3 a1e1 e8e5 h5e5 a8d8 e5e7 d3d2 e1e2 d2d1 e2e1');
INSERT INTO evaluations VALUES(12921,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(12921,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(12921,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12921,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(12921,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(12921,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(12921,36,'cp 28','f1c4 e7e6 g1f3 d7d6 e1g1 f8e7 d1e2 e8g8 f1d1 c6a5 c4d3 d5b4 b1c3 b4d3 d1d3 b7b6 b2b4 a5c4 f3d2 c4d2 c1d2 f7f5 e5f6 e7f6 d2f4 f6e7 f4g3 c8d7 c3e4 d7c6 a1e1 c6d5 d3c3 f8f7 a2a4 d8f8');
INSERT INTO evaluations VALUES(12917,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(12917,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(12917,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(12917,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12917,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12917,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(12917,36,'cp 16','e7e6 g1f3 d7d6 d1e2 f8e7 e1g1 e8g8 e2e4 d5b4 e5d6 e7d6 b1c3 h7h6 c1d2 b4d5 c4d5 e6d5 e4d5 d6b4 d5c4 b4c3 d2c3 c8e6 c4b5 e6d5 b5d3 f8e8 f1e1 e8e1 a1e1 c6e7 a2a3 a8c8 c3d2 d5e6 d3b5 e6d5 f3e5 f7f6');
INSERT INTO evaluations VALUES(14082,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(14082,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(14082,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(14082,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(14082,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14082,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14082,36,'cp 45','c7c6 d5c6 b7c6 b5a4 d8c7 c1e3 a7a5 d1e2 c8a6 e2d2 e7c8 c3d5 f6d5 e4d5 a6b5 a4b5 c6b5 a2a4 b5b4 d2d3 c8b6 e1g1 f8e7 f1c1 b6c4 c2c3 b4b3 f3d2 c4e3 d3b5 e8f8 f2e3 g7g6 b5d3 f8g7 c3c4 a8b8 c4c5 d6c5 d2c4 e5e4 d3e4 b8b4 g1h1 e7f6 d5d6 c7d7 e4d3 b4c4 c1c4 f6b2');
INSERT INTO evaluations VALUES(12984,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(12984,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(12984,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(12984,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(12984,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12984,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(12984,36,'cp 30','d5c6 b7c6 b5a4 d8c7 d1d3 a7a5 c1e3 c8a6 d3d2 e7c8 c3d5 f6d5 e4d5 a6b5 a4b5 c6b5 e1g1 f8e7 a2a4 b5b4 c2c3 b4c3 d2c3 c7b7 a1c1 e8g8 c3c6 a8b8 c6b7 b8b7 c1c6 f7f5 g2g3 b7b2 f1c1 b2b8 g1g2 e7f6 f3g5 f6g5 e3g5 f5f4 g3f4 e5f4 f2f3 c8b6');
INSERT INTO evaluations VALUES(12938,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(12938,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(12938,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(12938,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(12938,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(12938,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(12938,36,'cp 14','b7c6 b5a4 d8c7 c1e3 a7a5 d1d3 c8a6 d3d2 e8c8 e1c1 d6d5 e4d5 e7d5 c3d5 f6d5 a2a3 f7f6 d2e1 a6b5 a4b3 a5a4 b3a2 b5a6 c1b1 f8e7 e3d2 e7c5 e1e4 d5b6 d2e3 c5e3 e4e3 c8b8 d1d2 c6c5 h1d1 c5c4 e3e2 h8e8 b1a1 d8d2 e2d2 a6b7');
INSERT INTO evaluations VALUES(12022,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(12022,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(12022,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(12022,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(12022,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(12022,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(12022,36,'cp 51','b5a4 d8c7 c1g5 c8e6 a4b3 h7h6 g5f6 g7f6 e1g1 e7g6 b3e6 f7e6 f3h4 g6h4 d1h5 c7f7 h5h4 h8g8 h4h3 f8e7 a1d1 a8b8 h3h6 f6f5 b2b3 f5f4 f1e1 e8d8 h6h3 d8c7 c3a4 f4f3 g2g3 b8f8 a4b2 d6d5 e4d5');
INSERT INTO evaluations VALUES(15530,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(15530,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(15530,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(15530,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(15530,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(15530,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(15530,36,'cp 18','b1c3 g7g6 g2g3 f8g7 f1g2 c8g4 h2h3 g4f3 g2f3 h7h5 h3h4 c6d4 f3g2 c7c6 e1g1 a7a5 g2h3 d8e7 c1e3 d4e6 g1g2 e8g8 f2f4 e5f4 g3f4 f6h7 d1e1 g7f6 g2h1 f6h4 e1d2 h4f6 c3e2 f6g7 a1b1 e7h4 h1g2 e6c5 f4f5');
INSERT INTO evaluations VALUES(16553,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(16553,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(16553,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(16553,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(16553,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(16553,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(16553,36,'cp -128','b1c3 d7d5 c1g5 d5d4 c3e2 c8e6 a2a3 h7h6 g5d2 g7g5 g2g4 h6h5 d2g5 h5g4 e2g3 f8e7 h2h3 d8d7 d1d2 g4f3 g1f3 e8c8 e1c1 b7b5 h3h4 f6g4 g5e7 d7e7 f1h3 g4e3 h3e6 e7e6 c1b1 e3d1 d2d1 c8b7 f3g5');
INSERT INTO evaluations VALUES(16523,30,'cp -115',NULL);
INSERT INTO evaluations VALUES(16523,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(16523,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(16523,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(16523,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(16523,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(16523,36,'cp -144','d7d5 c1g5 d5d4 c3e2 c8e6 g2g3 h7h6 g5d2 d8d7 f3f4 e5f4 e2f4 e6g4 f1e2 e8c8 e2g4 f6g4 h2h3 g4f6 f4h5 f8d6 h5f6 d6g3 e1e2 g7f6 d1f1 h8g8 e2d1 f6f5 f1f5 d7f5 e4f5 c6e7 d2h6 e7f5 h6d2 g3h4 g1f3 f5g3 h1g1');
INSERT INTO evaluations VALUES(15139,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(15139,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(15139,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(15139,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(15139,34,'cp -120',NULL);
INSERT INTO evaluations VALUES(15139,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(15139,36,'cp -117','c1g5 d5d4 c3e2 c8e6 g2g3 h7h6 g5d2 f8e7 f1h3 d8c8 h3e6 c8e6 a2a3 f6d7 d1b1 f7f5 e4f5 e6f5 g3g4 e7h4 e2g3 f5g6 b1a2 h6h5 g4h5 g6h5 a2e6 c6e7 g1e2 h5f3 e1c1 f3f7 e6f7 e8f7 g3e4 a8f8 c2c3 e7f5 c3d4 e5d4 e2g3 h4g3 h2g3 h8h1');
INSERT INTO evaluations VALUES(20648,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(20648,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(20648,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(20648,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(20648,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(20648,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(20648,36,'cp 83','a7a6 b5d3 f8e7 b1c3 g8f6 e1g1 e8g8 h2h3 b7b5 c3e2 c8b7 e2g3 a8c8 d1e2 c6b4 d3b1 b4c6 f1d1 f8e8 c1e3 d8c7 e3d2 c7b6 d2g5 c6a5 g5e3 a5c4 e3c1 c4a5 b1d3 e7f8 c1g5 f6d7 b2b3 a5c6 g5e3');
INSERT INTO evaluations VALUES(17657,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(17657,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(17657,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(17657,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(17657,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(17657,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(17657,36,'cp 81','b5d3 f8e7 e1g1 g8f6 b1c3 e8g8 h2h3 b7b5 c3e2 c8b7 e2g3 a8c8 d1e2 c6b4 d3b1 f8e8 c1e3 b4c6 f1e1 h7h6 e3d2 e7f8 b1d3 d8b6 d2e3 c6b4 d3b1 b6d8 a2a3 b4c6 b1d3 c6a5 a1d1 a5c4 e3c1 d8b6 d3b1 a6a5');
INSERT INTO evaluations VALUES(28748,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(28748,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(28748,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(28748,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28748,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(28748,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(28748,36,'cp 0','e1g1 d7d5 c2c4 f8d6 b2b3 e8g8 c4d5 f6d5 c1b2 b8d7 b1d2 c7c5 a1c1 a8c8 d2c4 d6c7 d3b1 f7f5 d1e2 d8e7 a2a4 a7a6 b2a3 f8d8 c4d2 c7d6 d2c4');
INSERT INTO evaluations VALUES(28747,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(28747,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(28747,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(28747,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28747,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28747,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28747,36,'cp 0','d7d5 b2b3 f8d6 c2c4 e8g8 c4d5 f6d5 b1d2 b8d7 c1b2 c7c5 a1c1 a8c8 d2c4 d6b8 d3b1 f7f5 d1e2 d8e7 f1d1 f8d8 b2a3 e7f6 a3b2');
INSERT INTO evaluations VALUES(31123,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(31123,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(31123,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(31123,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(31123,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(31123,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(31123,36,'cp 26','g8f6 f1d3 e7e6 e1g1 f6e4 c3e4 d5e4 d3e4 g4h5 h2h3 b8d7 g2g4 h5g6 e4g6 h7g6 g1g2 d8c7 c2c4 f8d6 d1d3 e8g8 c1g5 d7f6 b2b3 c6c5 d4d5 e6d5 c4d5 a8d8 a1d1 d6f4 g5f6 g7f6 f3d2 f4d2 d1d2 d8d7 d3g3 c7g3 g2g3 f8d8 f1d1');
INSERT INTO evaluations VALUES(31858,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(31858,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31858,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(31858,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(31858,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(31858,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31858,36,'cp 0','g8f6 c2c4 e7e6 b1c3 f8b4 c4d5 e6d5 d1a4 b8c6 c1g5 h7h6 g5f6 d8f6 e2e3 e8g8 a1c1 a7a5 a2a3 b4c3 c1c3 c6e7 f1e2 c7c6 e1g1 e7c8 c3c5 c8b6 a4b3 b6d7 c5c3 f8b8 a3a4 f5e4 f3d2 e4f5');
INSERT INTO evaluations VALUES(36359,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(36359,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36359,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36359,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(36359,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(36359,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(36359,36,'cp 34','c2c4 d5c4 d1c2 a7a6 a2a4 c8d7 c2c4 d7c6 c1g5 a6a5 b1c3 b8d7 f1d1 h7h6 g5f6 d7f6 c4b3 a8b8 b3c2 c6f3 g2f3 c7c6 e2e3 f6d5 c2e2 d5b4 h2h4 d8c7 c3e4 b8d8 e4d2 c6c5 a1c1 d8d7 h4h5 d7d8 g1g2 b7b6');
INSERT INTO evaluations VALUES(37066,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(37066,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(37066,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(37066,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(37066,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(37066,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(37066,36,'cp 37','c4d5 e6d5 f1d3 c7c5 g1e2 b7b6 e1g1 c8a6 d3a6 b8a6 d1d3 a6c7 d4c5 b6c5 c3c4 a8b8 f1d1 d5c4 d3c2 d8e7 c2c4 c7e6 c1d2 f8d8 d2e1 d8d1 a1d1 e7b7 e2c3 h7h6 d1b1 b7c7 b1b8 c7b8 h2h3 b8c7 c4a4 c7b7 f2f3 f6d7 g1f2 d7e5 a4b5 b7b6 b5b6');
INSERT INTO evaluations VALUES(37018,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(37018,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(37018,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(37018,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(37018,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37018,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(37018,36,'cp 26','e6d5 f1d3 c7c5 g1e2 b7b6 e1g1 c8a6 d3a6 b8a6 d1d3 a6c7 d4c5 b6c5 c3c4 a8b8 f1d1 d5c4 d3c2 d8e7 c2c4 f8d8 c1d2 c7e6 e2c3 h7h6 d2e1 d8d1 a1d1 b8b6 c3a4 b6b8 e1c3 b8d8 d1c1 f6d7 h2h4 d8b8 c4g4 d7b6 h4h5 e7g5 g4g5 h6g5');
INSERT INTO evaluations VALUES(37119,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(37119,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(37119,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(37119,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(37119,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(37119,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(37119,36,'cp 27','f1d3 b7b6 g1e2 c7c5 e1g1 c8a6 d3a6 b8a6 d1d3 a6c7 d4c5 b6c5 c3c4 a8b8 f1d1 d5c4 d3c2 d8e7 c2c4 f8d8 c1d2 c7e6 d2e1 h7h6 d1d8 e7d8 e2c3 b8b2 a1c1 d8b8 c3a4 b2b5 h2h3 b8b7 g1h1 b7c6 h1h2 c6d6 h2g1 d6c6 e1c3 f6e4 f2f3 e4c3 a4c3');
INSERT INTO evaluations VALUES(46259,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46259,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(46259,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(46259,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46259,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46259,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46259,36,'cp 0','f1e2 b8d7 e1g1 b7b6 c2c4 d5c4 b1d2 f8e7 d2c4 e8g8 d4c5 d7c5 f3d4 f6d5 d1c2 c8d7 c4e5 e7f6 e5d7 c5d7 f1d1 a8c8 c2e4 d7c5 e4c2');
INSERT INTO evaluations VALUES(47734,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47734,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(47734,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(47734,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(47734,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(47734,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(47734,36,'cp 52','f8e7 h2h3 c8e6 d4d5 e6f5 b1c3 e7f6 f3d4 f5g6 d1e2 d8e7 e2e7 e8e7 c1e3 h8e8 e1c1 b8d7 g2g4 e7f8 b2b3 e8e7 f1d3 g6d3 d1d3 d7c5 d3d1 a8e8 c1c2 f6d4 d1d4 b6d7 g4g5 e7e3 f2e3 e8e3 d4h4 e3g3 h4h7 g3g5 h7h8 f8e7 h1e1 d7e5');
INSERT INTO evaluations VALUES(40731,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(40731,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(40731,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(40731,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(40731,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(40731,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(40731,36,'cp 86','f1d3 c8g4 c1e3 g4h5 b1c3 c7c6 d1b3 e8g8 e1c1 b8d7 h2h3 h5g6 d3g6 h7g6 g2g4 a7a5 f3d2 a5a4 b3c2 a4a3 b2b3 f8e8 h3h4 d7f8 d4d5 c6d5 e3b6 d8b6 c3d5 b6f2 d2e4 f2f3 d5c7 f3g4 c7a8 e8a8');
INSERT INTO evaluations VALUES(50631,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(50631,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(50631,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(50631,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(50631,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(50631,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(50631,36,'cp 78','c1e3 g8f6 f1e2 f8g7 d1d2 b8c6 e1c1 c8d7 f2f3 h7h5 c1b1 a8c8 h2h3 h5h4 f3f4 b7b5 e2f3 c6a5 b2b3 d8c7 c3d5 f6d5 e4d5 c7c3 h1e1 e8f8 d2c3 c8c3 e3c1 g7d4 d1d4 d7f5 d4d2 c3c7 e1e2 f7f6 c1b2 c7c8 f3g4 f5g4');
INSERT INTO evaluations VALUES(50630,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(50630,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(50630,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(50630,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(50630,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(50630,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(50630,36,'cp 57','g8f6 f1e2 f8g7 d1d2 c8d7 f2f3 h7h5 e1c1 b8c6 c1b1 a8c8 h2h3 h5h4 f3f4 b7b5 e2f3 d8c7 d4b3 e8g8 h1e1 f8e8 d2f2 c7b8 f4f5 c6e5 f5g6 f7g6 f2h4 b5b4 c3d5 f6d5 e4d5 e5f3 g2f3 d7f5 d1d2 b8b5 e3d4 b5c4 h4f2 g7d4 b3d4 c4d5');
INSERT INTO evaluations VALUES(51401,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(51401,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(51401,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(51401,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(51401,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(51401,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(51401,36,'cp 49','d2d4 d7d5 c4d5 e6d5 g1f3 g8f6 f3e5 f8e7 c1f4 e8g8 e2e3 c8b7 f1d3 c7c5 e1g1 b8c6 a1c1 d8c8 b2b3 c8e6 e5c6 e6c6 d1f3 a8d8 f1d1 f6e4 f4e5 c6g6 c3e2 f8e8 d3b5 b7c6 e2f4 g6h6 b5d3 e7f8 f3e2 f8d6 e5d6');
INSERT INTO evaluations VALUES(55368,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(55368,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(55368,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(55368,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(55368,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(55368,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(55368,36,'cp 84','f8g7 h2h4 h7h6 c1f4 b8d7 e4e5 d6e5 d4e5 a7a6 d1d2 g8e7 a1d1 b7b5 f1d3 c8b7 d3e4 b7c6 d2e3 c6e4 e3e4 d8c8 e1g1 e8g8 e4e2 d7b6 c3e4 b6d5 f4h2 c8b7 c2c3 b7c6 f1e1 c6c4 e2c4 b5c4 f3d4 a8b8');
INSERT INTO evaluations VALUES(1820,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(1820,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(1820,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(1820,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(1820,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(1820,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(1820,36,'cp -71','h4e4 f1e2 d7c6 e1g1 c8e6 b1c3 e4h4 g2g3 h4d8 c3e4 f8e7 c1e3 d8d5 e4c5 e7c5 d1d5 c6d5 e3c5 g8f6 a2a4 a7a5 f1e1 f6d7 c5d4 e8g8 f2f3 c7c5 d4f2 b7b6 e2b5 d7f6 c2c3 h7h6 g3g4 f8d8 b5d3 e6d7 h2h4 d7c6 f2g3 d5d4 g3c7 d8d7');
INSERT INTO evaluations VALUES(4800,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(4800,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(4800,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(4800,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(4800,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(4800,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4800,36,'cp 0','g8e7 d2d4 f7f6 d1b3 a6a5 a2a4 e7g6 b1d2 d8e7 d2c4 e7f7 c4d6 c7d6 b3f7 e8f7 c1e3 c8e6 f1b1 d6d5 f3d2 h8b8 b2b4 a5b4 b1b4 d5e4 d2e4 b7b6 e4d6 f7e7 d6e4');
INSERT INTO evaluations VALUES(10618,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(10618,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(10618,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(10618,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(10618,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(10618,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(10618,36,'cp 140','c4f7 e8f7 f3g5 f7g8 d1g4 d7d6 g4h5 g7g6 h5d1 h7h6 g5f3 g8h7 c2c3 h8f8 b2b4 c5b6 c1e3 g6g5 a2a4 c8e6 b4b5 g5g4 f3d2 c6e7 a4a5 b6e3 f2e3 f8f1 d2f1 e7g6 c3c4 d8g5 b1c3 a8f8 a5a6 b7b6 d1e2 h6h5 f1d2 h7h6 a1f1 f8f1 d2f1 h5h4 c3d5 e6d5 e4d5 h4h3 g2g3 g5f5 e3e4');
INSERT INTO evaluations VALUES(23670,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(23670,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(23670,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(23670,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(23670,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(23670,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(23670,36,'cp 124','g2g4 c6d4 e3d4 d8a5 h2h4 d7e6 a2a3 h7h5 g4h5 f6h5 d4g7 g8g7 f3f4 f8c8 f4f5 c8c3 d2d4 h5f6 d4c3 a5c3 b2c3 g6f5 e4f5 e6f5 f1d3 f5e6 d1g1 g7h6 g1g5 a8c8 c1b2 b7b6 a3a4 a7a5 d3f5 e6f5 g5f5 h6g7 f5g5 g7h6 h1g1 e7e6 g5g7 c8c7 c3c4 d6d5 c4d5 f6d5');
INSERT INTO evaluations VALUES(23666,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(23666,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(23666,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(23666,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(23666,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(23666,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(23666,36,'cp 142','c6d4 e3d4 d8a5 h2h4 d7e6 a2a3 h7h5 g4h5 f6h5 d4g7 g8g7 f3f4 f8c8 f4f5 c8c3 d2d4 h5f6 d4c3 a5c3 b2c3 g6f5 h1g1 g7h6 e4f5 e6f5 f1d3 f5e6 g1g5 a8c8 c1b2 b7b6 d3f5 e6f5 g5f5 h6g6 d1f1 c8c7 f5g5 g6h6 f1f4 c7c8 a3a4 c8f8 g5f5 h6g6 a4a5 b6a5 f5g5 g6h6 g5a5 f8b8 b2c1 b8b7 f4a4 b7c7 a5a7');
INSERT INTO evaluations VALUES(23952,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(23952,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(23952,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23952,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(23952,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(23952,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(23952,36,'cp 77','d2d4 e8g8 h2h3 b7b5 a4c2 f8e8 a2a3 e7f8 b1d2 h7h6 b2b3 c6a5 c1b2 d7e6 d4d5 e6d7 b3b4 a5b7 c3c4 c7c5 d5c6 d7c6 c4b5 c6b5 c2b3 a6a5 a3a4 b5d7 b4b5 b7c5 b2a3 c5d3 e1e3 d3c5 a1c1 c5b3 d1b3');
INSERT INTO evaluations VALUES(23950,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(23950,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(23950,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(23950,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(23950,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(23950,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(23950,36,'cp 56','e8g8 h2h3 e5d4 c3d4 d6d5 e4e5 f6e4 b1c3 d7f5 c3e2 c6a5 e2f4 a8c8 g2g4 f5e6 a4c2 c7c5 f4d5 e6d5 c2e4 d5e4 e1e4 d8d5 d1e2 a5c6 d4c5 e7c5 c1g5 c8e8 g5h4 h7h6 h4g3 f7f5 e2c4 e8d8 g4f5 f8f5 g1g2 d5c4 e4c4 c5b6 c4f4 f5f4 g3f4 g7g5 f4g3 g8f7 a1c1 d8d3');
INSERT INTO evaluations VALUES(23608,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(23608,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(23608,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23608,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(23608,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(23608,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(23608,36,'cp 82','h2h3 b7b5 a4c2 h7h6 a2a3 f8e8 b2b3 e7f8 b1d2 c6a5 b3b4 a5c6 d2b3 a6a5 d4d5 a5a4 b3d2 c6a7 c3c4 c7c6 c1b2 a7c8 d5c6 d7c6 c2d3 b5c4 d2c4 c8b6 f3d2 c6b5 a1c1 b6c4 d2c4 a8c8 d3f1 c8c6 d1f3 b5a6');
INSERT INTO evaluations VALUES(23607,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(23607,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(23607,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(23607,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(23607,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(23607,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(23607,36,'cp 62','f8e8 a4c2 b7b5 a2a3 h7h6 b2b3 e7f8 b1d2 c6a5 a1b1 c7c5 d4d5 g7g6 c3c4 f6h5 c2d3 h5f4 d3f1 b5b4 g1h2 a5b7 g2g3 f4h5 b1a1 f8g7 f1d3 h5f6 d2f1 f6h7 g3g4 a6a5 a3a4 h7g5 h2g2 g7f6 d3c2 g5f3 d1f3 f6g7 f1e3');
INSERT INTO evaluations VALUES(25695,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(25695,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25695,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(25695,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(25695,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(25695,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(25695,36,'cp 34','d8d7 f3c3 f8d6 d2d3 g8e7 c1e3 b7b6 b2b4 a6a5 a2a3 e7g6 b1d2 e8g8 d2c4 a5b4 a3b4 h7h6 f1b1 f8d8 g1h2 g6f4 e3f4 e5f4 c4d6 c7d6 c3d4 d7c7 c2c3 g8h7 a1a8 d8a8 b4b5 c6c5 d4d5 a8a3');
INSERT INTO evaluations VALUES(39062,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(39062,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(39062,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(39062,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(39062,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(39062,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(39062,36,'cp 86','d2d4 e5d4 f3d4 g8e7 c1g5 h7h6 g5f4 d7d6 d1d2 b8c6 f4e3 c8e6 e1c1 d8d7 d4e6 f7e6 f1c4 e8c8 a2a3 c8b8 f2f4 d6d5 e4d5 g7c3 d2c3 e6d5 g2g3 d5d4 c3d3 e7f5 h1e1 h8e8 e3d2 h6h5 c1b1 a7a6 d3b3 b8a7 e1e6 e8e6 c4e6 d7d6 d2c1 d6e7 d1e1');
INSERT INTO evaluations VALUES(39058,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(39058,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(39058,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(39058,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(39058,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(39058,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(39058,36,'cp 91','e5d4 f3d4 g8e7 c1g5 h7h6 g5f4 b8c6 f4e3 d7d6 d1d2 c8e6 d4e6 f7e6 e1c1 d8d7 a2a3 e8c8 c1b1 d6d5 e3c5 d7e8 f1d3 c8b8 e4d5 e7d5 c3d5 e6d5 h1e1 e8f7 f2f4 f7f6 d2c1 h8e8 g2g4 a7a6 e1e8 d8e8 h2h4 f6f7 c1d2 b7b6 c5f2 f7f6');
INSERT INTO evaluations VALUES(39066,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(39066,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(39066,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(39066,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(39066,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(39066,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(39066,36,'cp 104','f3d4 g8e7 d1d3 e8g8 c1f4 b8c6 d4c6 e7c6 d3d2 d7d5 c3d5 g7b2 a1b1 b2e5 f1c4 e5f4 d2f4 c8e6 e1g1 d8d6 f4g5 d6d8 h2h4 d8g5 h4g5 b7b6 d5f6 g8g7 c4d5 e6d5 e4d5 c6d4 b1e1 h7h6 f2f4 a8d8 c2c4 d4f5 g2g4 f5d6 e1e7 h6g5 f4g5 d6c4');
INSERT INTO evaluations VALUES(39065,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(39065,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(39065,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(39065,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(39065,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(39065,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(39065,36,'cp 82','g8e7 d1d3 e8g8 c1f4 b8c6 d4c6 e7c6 d3e3 d7d6 e1c1 c8e6 h2h4 h7h5 f4h6 d8f6 h6g7 f6g7 f1e2 a8e8 e3d2 c6e5 f2f4 e5g4 h1f1 e6d7 d1e1 g8h7 g2g3 a7a6 e2g4 d7g4 c3d5 c7c6 d5e3 e8e4 e3g4 e4e1 f1e1 h5g4');
INSERT INTO evaluations VALUES(46321,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(46321,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(46321,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46321,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(46321,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(46321,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(46321,36,'cp 18','b7b5 f3e5 b8d7 h2h4 d7e5 d4e5 f6d7 d1f3 g7g6 f1e2 d8c7 f3g3 h7h6 a2a4 b5b4 e3e4 c8b7 c3b4 f8b4 e1g1 b4d2 f4d2 d7c5 e4d5 c5e4 g3f4 e4d2 f4d2 b7d5 a1c1 c7e5 e2c4 d5c4 c1c4 e8g8 d2c2 a8d8 f1c1 e5f5 c2f5 e6f5');
INSERT INTO evaluations VALUES(50851,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(50851,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(50851,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50851,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(50851,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50851,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(50851,36,'cp 24','c7c5 g1f3 g8f6 e1g1 e7e6 d4c5 f8c5 d1d8 e8d8 c4e2 d8e7 b2b3 b8d7 b1d2 b7b6 c1b2 c8b7 a2a3 b6b5 a3a4 h8b8 a4b5 a6b5 a1a8 b7a8 f1c1 a8d5 g1f1 c5b4 c1c7 d5a8 c7a7 e7e8 f3e5 d7e5 b2e5');
INSERT INTO evaluations VALUES(54106,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(54106,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(54106,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(54106,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(54106,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(54106,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(54106,36,'cp 96','c3e4 c5d4 d1d4 d8d4 f3d4 a7a6 c1f4 g8f6 e4d6 f8d6 f4d6 c8d7 f2f3 h7h5 e1f2 b8c6 a1d1 e8c8 f1c4 d7e8 d6c5 c6d4 d1d4 d8d4 c5d4 e8a4 c2c3 h5h4 h1e1 h8h5 e1e5 h5e5 d4e5 c8d7 c4e2 a4c6 e5d4 f6e8 a2a3 f7f6 f3f4 d7e7');
INSERT INTO evaluations VALUES(54928,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(54928,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(54928,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(54928,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(54928,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54928,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(54928,36,'cp 23','c8f5 f1d3 f5d3 d1d3 c7c6 e3e4 e7e6 e1g1 d5e4 d3e4 f8d6 e4e2 g8f6 b1d2 e8g8 d2c4 b7b5 c4d6 d8d6 b2b3 b8d7 c2c4 b5c4 b3c4 c6c5 d4c5 d6c5 c1e3 c5h5 f1e1 f8d8 a1c1 d7e5 f3e5 h5e5 c4c5 f6d5 c5c6 d5e3 e2e3');
INSERT INTO evaluations VALUES(31755,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(31755,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(31755,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(31755,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(31755,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(31755,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(31755,36,'cp 35','c2c4 e7e6 d1b3 b7b6 f1d3 f5d3 b3d3 g8f6 e1g1 f8d6 b2b3 e8g8 b1c3 b8d7 c1b2 f8e8 c4d5 e6d5 c3b5 a7a6 b5d6 c7d6 f1c1 b6b5 c1c6 f6e4 f3e1 b5b4 f2f3 e4f6 c6a6 a8a6 d3a6 e8e3 e1c2 e3e6 c2b4 d8e8 a6d3 e6e2 b4c2');
INSERT INTO evaluations VALUES(54951,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(54951,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(54951,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(54951,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(54951,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(54951,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(54951,36,'cp -66','g8f6 h5e2 c7c5 g2g3 g7g6 f1g2 f8g7 g1f3 e8g8 c2c3 b8c6 e1g1 e7e5 d2d3 d6d5 c1g5 d5d4 c3d4 c5d4 f1c1 h7h6 g5d2 a7a5 b1a3 f8e8 c1c5 d8b6 c5c2 g7f8 a3c4 b6d8 a1c1 a5a4 f3e1 c8d7 f2f4 f6g4 g2f3 b7b5 f3g4 d7g4 e2g4 b5c4');
INSERT INTO evaluations VALUES(46974,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(46974,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(46974,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(46974,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(46974,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(46974,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(46974,36,'cp -56','h5e2 c7c5 g2g3 g7g6 g1f3 f8g7 f1g2 b8c6 c2c3 e8g8 e1g1 e7e5 d2d3 d6d5 c1g5 d5d4 c3d4 c5d4 f1c1 c8g4 b1a3 h7h6 g5d2 f6d7 e2f1 a8c8 f3e1 d8e7 a3c4 g4e6 g2h3 f7f5 e4f5 g6f5 f1e2 c8e8 h3g2 f5f4 c4a5 c6a5 d2a5');
INSERT INTO evaluations VALUES(10348,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(10348,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(10348,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(10348,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(10348,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(10348,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(10348,36,'cp -80','e5d4 c3d5 b4d2 d1d2 e8g8 f1d3 d7d6 e1g1 h7h6 d5f6 d8f6 d3b5 c6e5 f3d4 c7c5 d4e2 a7a6 b5d3 c5c4 f2f4 f6d8 f4e5 d6e5 d2e3 c4d3 c2d3 c8g4 h2h3 g4e2 e3e2 d8a5 a2a3 a5b6 g1h1 a8c8 a1c1 c8c7 c1c7 b6c7 e2d2 f8c8 d2e3 c8d8');
INSERT INTO evaluations VALUES(10485,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(10485,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(10485,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(10485,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(10485,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(10485,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(10485,36,'cp -73','c3d5 b4d2 d1d2 e8g8 f1d3 d7d6 e1g1 h7h6 h2h3 f8e8 d2f4 f6d7 f4h4 d7e5 h4d8 e5f3 g2f3 e8d8 g1h2 d8d7 f3f4 c6e7 f4f5 e7d5 e4d5 d7e7 f5f6 g7f6 a1e1 e7e5 f2f4 e5e1 f1e1 c8d7 e1e7 a8d8 e7e4 b7b5 e4d4 f6f5 h2g2 a7a5 g2f2');
INSERT INTO evaluations VALUES(14753,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(14753,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(14753,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(14753,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(14753,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(14753,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(14753,36,'cp -3','e6d5 b1c3 f8d6 c3b5 e8g8 b5d6 d8d6 f1e2 f8e8 e1g1 c8g4 c1d2 f6e4 d2e1 c6e7 f3e5 g4e2 d1e2 c7c5 f2f3 e4f6 e1c3 c5d4 e3d4 d6b6 a1e1 e7f5 e2f2 a7a5 a2a3 f5d6 g1h1 e8e6 e5d3 e6e1 f1e1 h7h6 h2h3 d6c4 e1e7 b6b5');
INSERT INTO evaluations VALUES(16419,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(16419,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(16419,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(16419,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(16419,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(16419,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(16419,36,'cp -40','g1f3 f6e4 c4d5 e4f6 d5c6 d7c6 f4e5 f6d5 e1g1 f8e7 d2d4 c8f5 a2a3 e8g8 c2c4 d5b6 d1e2 f7f6 c4c5 b6d5 b1c3 f5g4 e5f6 e7f6 e2d3 d8e8 c3d5 c6d5 c1f4 e8e4 d3e4 d5e4 f3e5 g4e2 f1f2 e2d3 f4e3 a8d8 c5c6 f6e5 f2f8 g8f8 d4e5 b7b6 g1f2 d8e8 h2h3 e8e5');
INSERT INTO evaluations VALUES(16418,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(16418,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(16418,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(16418,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(16418,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(16418,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(16418,36,'cp -29','f6e4 c4d5 e4f6 d5c6 d7c6 f4e5 f6d5 d2d4 c8f5 a2a3 b7b5 e1g1 f8e7 f3h4 f5c8 h4f5 c8f5 f1f5 d8d7 d1d3 a8d8 b1c3 e8g8 f5f3 c6c5 c3b5 a7a6 b5c3 d5c3 b2c3 c5d4 f3h3 g7g6 c1f4 f8e8 h3f3 d4c3 d3d7 d8d7 f3c3 g6g5 f4g3 e8d8 a1e1 h7h5 e5e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(31865,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(31865,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(31865,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31865,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(31865,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31865,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31865,36,'cp 12','c2c4 e7e6 b1c3 c7c6 c1f4 g8f6 e2e3 f8d6 f4d6 d8d6 g1f3 b8d7 f1e2 e8g8 e1g1 h7h6 a1c1 f8d8 e2d3 d5c4 d3c4 d6e7 c4d3 f5d3 d1d3 e6e5 d3c2 e5d4 f3d4 d7c5 f1d1 c5e4 c3e4 f6e4 a2a3');
INSERT INTO evaluations VALUES(38883,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(38883,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(38883,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(38883,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(38883,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(38883,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(38883,36,'cp -24','b7b5 a2a4 c7c6 f1e2 c8b7 e1g1 a7a6 c3a2 b4e7 e4e5 c6c5 d4c5 d8d1 f1d1 e7c5 a4b5 a6b5 c1e3 c5e3 f2e3 b8d7 a2c3 g8h6 c3b5 e8e7 b5d6 b7f3 g2f3 c4c3 b2c3 d7e5 d6e4 h8b8 g1f2 h6f5 e4c5 g7g5 f3f4 g5f4 e3f4 e5g6 f2f3 a8a1 d1a1 b8b2 a1a7 e7f8 e2d3 b2h2 d3f5');
INSERT INTO evaluations VALUES(37552,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(37552,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(37552,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(37552,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(37552,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(37552,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(37552,36,'cp -28','a2a4 c7c6 c1g5 g8e7 f1e2 h7h6 g5h4 e8g8 e1g1 c8b7 d1c2 a7a6 b2b3 c4b3 c2b3 a6a5 f1d1 b8d7 a1c1 a8c8 c3a2 b4d6 e4e5 d6c7 a4b5 c6b5 b3b5 c8b8 b5c4 c7b6 a2c3 d8e8 h4e7 e8e7 d4d5 f8c8 d5d6 e7f8 c4f4 g7g5 f4g3 b7f3 g2f3 f8g7 f3f4 g5f4 g3g7 g8g7');
INSERT INTO evaluations VALUES(41299,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(41299,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(41299,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(41299,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(41299,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41299,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(41299,36,'cp 40','c1e3 c7c5 d4c5 f6d7 c5d6 g7b2 b1d2 e7d6 a1b1 b2g7 g1h3 e8g8 h3f4 b8c6 f1e2 d7c5 e1g1 b7b6 d2b3 c5a4 f4d5 a4c3 d5c3 g7c3 d1c1 c3e5 f1d1 d8c7 c1d2 f8e8 e2f1 c8b7 h2h3 h7h5 d2f2 a8d8 b3d2 e5c3 b1c1 c3g7 d2b3 c6e5 g1h1');
INSERT INTO evaluations VALUES(45127,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(45127,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(45127,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(45127,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(45127,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(45127,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(45127,36,'cp -54','b8c6 g1e2 c6e5 c4b5 c8d7 b5d7 d8d7 f3e3 e5c6 d2d3 e7e6 e1g1 f8e7 d3d4 d6d5 e4d5 f6d5 c3d5 d7d5 d4c5 e8c8 e2c3 d5c5 e3c5 e7c5 c3e4 c5e7 c1e3 d8d5 e4c3 d5d7 a1d1 h8d8 d1d7 d8d7 c3e4 b7b6 f2f4 f7f5 e4g5 e7g5 f4g5 c6b4 c2c3 b4d3 b2b3');
INSERT INTO evaluations VALUES(13236,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(13236,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(13236,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(13236,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(13236,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(13236,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(13236,36,'cp -41','d2d3 e7e6 c4b3 c6d4 f3g3 a7a5 g1f3 c8d7 a2a4 d4b3 c2b3 f8e7 e1g1 e8g8 c1g5 f6h5 g5e7 d8e7 g3g4 g7g6 g4h4 e7h4 f3h4 f7f5 h4f3 d7c6 a1d1 f8e8 f1e1 a8d8 d3d4 c6e4 c3e4 f5e4 e1e4 h5f6');
INSERT INTO evaluations VALUES(51164,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(51164,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(51164,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(51164,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(51164,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(51164,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(51164,36,'cp 54','c8b7 d1c2 e7e6 b1c3 g8f6 e4e5 f6g8 g1f3 b7f3 g2f3 b8c6 c1e3 c6b4 c2d2 b5c4 f1c4 c7c6 a2a3 b4d5 c3d5 c6d5 c4d3 g7g6 a1c1 f8g7 e1f1 g8e7 f1g2 e8g8 h2h4 h7h5 b2b4 d7d6 f3f4 d8d7 d2e2 g7h6 h1g1 a8c8 d3a6 c8c1 g1c1 d6e5 f4e5 h6e3 e2e3 e7f5');
INSERT INTO evaluations VALUES(30356,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(30356,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(30356,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(30356,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(30356,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(30356,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(30356,36,'cp 73','d1c2 e7e6 b1c3 g8f6 e4e5 f6g8 g1f3 b5c4 f1c4 b7f3 g2f3 b8c6 c1e3 c6b4 c2e2 c7c6 e1g1 g8e7 a2a3 b4d5 c3d5 e7d5 c4a6 d8h4 a6c4 d5f4 e3f4 h4f4 e2e3 f4h4 f3f4 g7g6 e3g3 h4h5 b2b4 f8e7 g1h1 e8g8 f1g1 f8b8 c4d3 e7f8 d3e4 h5e2 g3e3 e2c4 g1b1 f8h6');
INSERT INTO evaluations VALUES(53261,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(53261,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(53261,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(53261,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(53261,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(53261,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(53261,36,'cp -23','e7e6 g1f3 g8f6 d4c5 b8c6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 h2h3 h7h6 a2a3 a7a6 c1d2 c8e6 b2b4 c5a7 e2g3 f8e8 d1e2 f6d7 a1c1 d7f8 f1d1 f8g6 c2c4 d4c3 c1c3 c6d4 f3d4 a7d4');
INSERT INTO evaluations VALUES(55022,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(55022,31,'cp -179',NULL);
INSERT INTO evaluations VALUES(55022,32,'cp -160',NULL);
INSERT INTO evaluations VALUES(55022,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(55022,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(55022,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(55022,36,'cp -164','e7e5 d2d4 e5d4 g1f3 c7c5 f1g2 g8e7 c2c3 d4c3 b1c3 b8c6 e1g1 c8e6 d1a4 d5d4 c3e4 e7f5 c1f4 f8e7 f1c1 d8b6 f3e5 e8g8 e5d3 f5h4 g2h1 c5c4 c1c4 e6c4 a4c4 f8e8 a1c1 h4g6 f4g3 e7f8 h2h4 a8d8 h1g2 g6e5 c4c2 c6b4 c2d1 e5d3 e2d3 b4d5 h4h5 b6b2');
INSERT INTO evaluations VALUES(56321,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(56321,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(56321,32,'cp -170',NULL);
INSERT INTO evaluations VALUES(56321,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(56321,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(56321,35,'cp -163',NULL);
INSERT INTO evaluations VALUES(56321,36,'cp -159','d2d4 e5d4 d1d4 b8c6 d4d1 c8f5 g1f3 d8d7 f1g2 h7h6 f3d4 f5h3 g2h3 d7h3 d4c6 b7c6 c1f4 h6g5 f4c7 a8c8 c7g3 f7f5 e2e3 e8f7 b1c3 f5f4 d1f3 f8d6 c3e2 c8e8 e3f4 d6f4 e1d1 e8e4 f3b3 g8f6 g3f4 g5f4 b3h3 h8h3');
INSERT INTO evaluations VALUES(17513,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(17513,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(17513,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(17513,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(17513,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(17513,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(17513,36,'cp -29','b7b5 c4d3 e7e6 e1g1 c8b7 c1f4 a8c8 d1e2 f8c5 a1c1 c6e7 f4g3 g8f6 e4e5 f6h5 d3e4 h5g3 h2g3 b7c6 e4c6 e7c6 f1d1 e8g8 a2a4 b5a4 c3a4 c5e7 e2a6 c8a8 a6b6 d8c8 b6b3 a8b8 b3e3 c8b7 a4c5 e7c5 c1c5 h7h6 e3e2 c6b4 c5c4 f8c8 d1c1 c8c4 c1c4');
INSERT INTO evaluations VALUES(24124,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(24124,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(24124,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(24124,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(24124,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(24124,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(24124,36,'cp 78','c3d5 c6d4 c2c3 d4e6 g5f6 g7f6 d5e3 c5e3 f2e3 c7c6 d1d2 g4f3 g2f3 d6d5 e4d5 c6d5 c4b3 d5d4 c3d4 e5d4 e1c1 d4e3 d2e3 d8d4 h1e1 d4e3 e1e3 h8g8 b3e6 f7e6 e3e6 e8f7 d1e1 a8c8 c1d2 c8c7 e6e4 g8g5 e1e2 g5b5 b2b4 b5h5 a2a4 h5h3 d2e3 c7d7 a4a5 h3h5');
INSERT INTO evaluations VALUES(25696,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(25696,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(25696,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(25696,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(25696,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(25696,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25696,36,'cp 51','d1f3 g8e7 d2d3 c6c5 f3g3 e7g6 b1a3 f7f6 a3c4 d8d7 c1e3 f8d6 a2a3 b7b5 c4d6 c7d6 h3h4 f6f5 f2f4 e8g8 f4e5 g6e5 e4f5 f8f5 f1f5 d7f5 a1f1 f5g6 g3g6 e5g6 e3f4 g6f4 f1f4 a8d8 g1f2 d6d5 a3a4');
INSERT INTO evaluations VALUES(27206,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27206,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27206,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(27206,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(27206,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(27206,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(27206,36,'cp 31','g8f6 d2d3 h7h6 c1e3 a7a6 b5a4 b7b5 a4c2 d6d5 e1g1 f8d6 e4d5 f6d5 d3d4 d5e3 f2e3 d8e7 b1d2 e8g8 d2e4 f7f5 d4e5 c6e5 e4d6 c7d6 f3e5 e7e5 d1d4 d7e6 a1d1 a8d8 a2a4 g7g6 d4f4 g8g7 f1e1 e6c4 a4b5 c4b5 f4f2 b5c6 d1d4 e5e7 b2b4');
INSERT INTO evaluations VALUES(25508,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(25508,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(25508,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(25508,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(25508,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(25508,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(25508,36,'cp 33','d2d3 h7h6 c1e3 a7a6 b5a4 b7b5 a4c2 d6d5 e1g1 f8d6 e4d5 f6d5 d3d4 d5e3 f2e3 d8e7 b1d2 e8g8 a2a4 b5b4 g1h1 b4c3 b2c3 a8b8 a4a5 f7f5 e3e4 e5d4 c3d4 f5e4 d2e4 c6b4 c2b3 g8h8 f1e1 e7d8 d4d5 d7f5 e4d6 d8d6 f3e5 d6f6 a1c1');
INSERT INTO evaluations VALUES(25511,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(25511,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(25511,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(25511,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(25511,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(25511,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(25511,36,'cp 18','h7h6 c1e3 f8e7 b1d2 e8g8 b5a4 a7a5 e1g1 f8e8 a4b3 e7f8 f1e1 d7e6 d3d4 e6b3 d1b3 e5d4 c3d4 f6e4 d2e4 e8e4 b3b7 d8e8 b7c7 e4e7 c7b6 a8b8 b6a6 c6b4 a6e2 a5a4 b2b3 b8c8 a1c1 c8c1 e1c1 a4b3 a2b3 b4d5');
INSERT INTO evaluations VALUES(28497,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(28497,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(28497,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(28497,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(28497,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(28497,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(28497,36,'cp 159','d1a4 b8c6 f3e5 a8b8 e5c6 b7c6 a4c6 f6d7 e1g1 e8g8 h2h3 d7f6 c1d2 b4c3 d2c3 f6d5 c6c5 d5c3 b2c3 d8d6 c5a7 f5e4 a7c5 d6c5 d4c5 b8a8 a2a4 a8a5 c4b5 c7c6 b5e2 e4c2 e2f3 a5a4 a1a4 c2a4 f1a1 f8a8 a1b1 h7h6 b1b6 a8a5 f3c6 a4c6 b6c6 a5a1 g1h2');
INSERT INTO evaluations VALUES(31079,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(31079,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(31079,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31079,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(31079,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31079,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(31079,36,'cp 16','b8d7 b1d2 g8f6 d1b3 d8c8 g1f3 h7h6 f3e5 d7e5 f4e5 f6d7 e5g3 e7e6 f1e2 a7a5 a2a4 f8e7 e1g1 e8g8 f1d1 b7b6 e2f3 c8a6 e3e4 f5h7 d1e1 d7f6 e4d5 c6d5 f3e2 h7d3 e2d3 a6d3 d2f3 d3a6 f3e5 a6b7 b3b5 f8c8');
INSERT INTO evaluations VALUES(32785,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(32785,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(32785,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(32785,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(32785,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(32785,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(32785,36,'cp 56','c1d2 g8f6 f1c4 b8d7 d1e2 a5c7 c4b3 e7e6 f3h4 f5g6 e1c1 a7a5 h4g6 h7g6 g2g3 d7b6 a2a3 f8e7 h2h4 b6d5 c3d5 c6d5 b3a4 e8f8 c2c3 a8c8 c1b1 f6e4 a4b5 c7b6 h4h5 e4d2 d1d2 e7f6 d2c2 h8h5 h1h5 g6h5 e2h5 f8e7');
INSERT INTO evaluations VALUES(32786,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(32786,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(32786,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(32786,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(32786,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(32786,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(32786,36,'cp 50','g8f6 f1c4 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 b7b5 c4b3 d7b6 g2g4 e8c8 d2g5 f8e7 c1b1 b5b4 c3a4 b6a4 b3a4 c8b7 e2c4 c7b6 b1a1 f6d5 g5e7 d5e7 c2c3 b6a5 c4b4 a5b4 c3b4 b7c7 a2a3 c7b7');
INSERT INTO evaluations VALUES(34719,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(34719,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(34719,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(34719,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(34719,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(34719,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(34719,36,'cp 103','g8f6 d2d4 b8c6 b1c3 c8g4 c4b5 a7a6 b5c6 b7c6 d1d3 g4f3 d3f3 e7e6 d4e5 d6e5 c1g5 f6d7 f1d1 f8d6 g5e3 a6a5 h2h3 e8g8 c3a4 e6g6 f3g4 d7f6 g4g6 h7g6 f2f3 f6d7 d1d3 f7f5 d3c3 d7f6 c3c4 f5e4 f3e4 f8b8 e3c5 d6c5 c4c5');
INSERT INTO evaluations VALUES(41469,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(41469,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(41469,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(41469,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(41469,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(41469,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(41469,36,'cp 0','g1f3 c7c5 e1g1 e8g8 h2h3 d8b6 b1c3 c5c4 d3e2 b6b2 d1d2 b8c6 a1b1 b2a3 f3e5 e7e6 c3b5 a3a5 d2a5 c6a5 b5c7 a8b8 c7a6 b8a8');
INSERT INTO evaluations VALUES(50612,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(50612,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(50612,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50612,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50612,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50612,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50612,36,'cp 0','e7e6 g1f3 d7d5 d2d4 c5d4 f3d4 d5e4 c3e4 b8d7 d4b3 d8c7 c1d2 g8f6 f1d3 f8e7 d1f3 e8g8 e1c1 f6d5 e4c3 d5b4 d2e3 d7f6 c1b1 c8d7 e3d4 d7c6 f3g3 b4d3 d1d3 f6h5 g3g4 h5f6 g4g5 h7h6 g5h4 f6e4 h4g4 e4f6');
INSERT INTO evaluations VALUES(51434,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(51434,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(51434,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51434,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(51434,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(51434,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(51434,36,'cp 23','d2d4 c8b7 f1d3 c7c5 g1f3 c5d4 f3d4 b8a6 b1c3 a6c5 e1g1 c5d3 c2d3 f8c5 c1e3 g8f6 d1e2 e8g8 a2a3 f6e8 f4f5 e6e5 d4f3 d8e7 a1e1 e8f6 e3c5 b6c5 f3d2 g8h8 g1h1 d7d6 d2c4 f8d8 f1f3 b7a6 e2d2');
INSERT INTO evaluations VALUES(51426,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51426,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(51426,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(51426,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(51426,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(51426,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(51426,36,'cp 20','c8b7 f1d3 c7c5 g1f3 c5d4 f3d4 b8a6 b1c3 a6c5 e1g1 c5d3 c2d3 f8c5 c1e3 g8f6 d1e2 e8g8 a1d1 a7a6 h2h3 a8c8 d4b3 c5e3 e2e3 h7h6 a2a3 d7d6 c3e2 c8c2 d1d2 c2d2 b3d2 f8e8 d2f3 f6d7 e2g3 d8e7 f4f5 e6e5 g3h5 a6a5 b2b4');
INSERT INTO evaluations VALUES(30688,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(30688,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(30688,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30688,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(30688,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(30688,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30688,36,'cp 28','f1d3 c7c5 g1f3 c5d4 f3d4 b8a6 b1c3 a6c5 e1g1 c5d3 d1d3 d8c7 f4f5 a7a6 c1g5 f8c5 g1h1 g8f6 g5h4 c5e7 a1e1 e6e5 h4f6 e7f6 d4f3 e8g8 c3d5 b7d5 e4d5 f8e8 f3d2 e5e4 d2e4 f6b2 d5d6 c7c6');
INSERT INTO evaluations VALUES(30687,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(30687,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(30687,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(30687,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(30687,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(30687,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30687,36,'cp 7','c7c5 g1f3 c5d4 f3d4 b8a6 b1c3 a6c5 e1g1 c5d3 d1d3 d8c7 d4b5 c7c6 g1h1 a7a6 b5d4 c6c7 f4f5 g8f6 c1g5 f6g4 g5f4 e6e5 f4g3 d7d6 d4f3 b6b5 g3h4 f8e7 h2h3 e7h4 h3g4 h4e7 c3d5 b7d5 d3d5 a8c8 c2c3 h7h6 a2a4');
INSERT INTO evaluations VALUES(55013,30,'cp -234',NULL);
INSERT INTO evaluations VALUES(55013,31,'cp -235',NULL);
INSERT INTO evaluations VALUES(55013,32,'cp -223',NULL);
INSERT INTO evaluations VALUES(55013,33,'cp -216',NULL);
INSERT INTO evaluations VALUES(55013,34,'cp -225',NULL);
INSERT INTO evaluations VALUES(55013,35,'cp -228',NULL);
INSERT INTO evaluations VALUES(55013,36,'cp -210','e7e5 d1c2 b8a6 e2e3 a6c5 d2d4 c5d7 d4e5 d7e5 b1d2 f8d6 g1f3 e5f3 d2f3 g8f6 f1d3 e8g8 b2b3 f8e8 c1b2 c7c6 e1g1 f6e4 f3d2 e4d2 c2d2 c8g4 a1c1 g4h5 d2c2 h5g6 d3g6 h7g6 c2c3 d6e5 c3c2 d8e7 b2e5');
INSERT INTO evaluations VALUES(56121,30,'cp -207',NULL);
INSERT INTO evaluations VALUES(56121,31,'cp -221',NULL);
INSERT INTO evaluations VALUES(56121,32,'cp -223',NULL);
INSERT INTO evaluations VALUES(56121,33,'cp -232',NULL);
INSERT INTO evaluations VALUES(56121,34,'cp -215',NULL);
INSERT INTO evaluations VALUES(56121,35,'cp -229',NULL);
INSERT INTO evaluations VALUES(56121,36,'cp -208','d1c2 b8a6 e2e3 a6c5 d2d4 e5d4 e3d4 c5e6 g1f3 f8d6 f1d3 g8f6 e1g1 e8g8 b1c3 c7c6 f1e1 e6f4 c1f4 d6f4 f3e5 g7g6 c3a4 f6g4 g2g3 g4e5 d4e5 f4h6 f2f4 f8e8 g1g2 h6f8 h2h3 d8e7 a1c1 c8d7 a4c5 e7c5 c2c5 f8c5 c1c5 b7b6 c5c1');
INSERT INTO evaluations VALUES(7636,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(7636,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(7636,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(7636,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(7636,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(7636,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(7636,36,'cp -42','g5f6 e7f6 c2c3 e8g8 b1d2 d7d5 f1d3 c6c5 e1g1 c7c6 e4e5 f6e5 d3h7 g8h7 d1h5 h7g8 h5e5 f8e8 e5f4 c8e6 b2b3 a7a5 f1e1 a5a4 a1c1 d8b8 f4h4 b8d6 d2f3 a4b3 a2b3 f7f6 e1e3 e6f5 e3e8 a8e8 b3b4 f5e4 b4c5 d6c5 f3d4 e4g6');
INSERT INTO evaluations VALUES(15036,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15036,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15036,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15036,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15036,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(15036,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(15036,36,'cp 4','f1d3 f8d6 a2a3 e8g8 g1e2 c6e7 d1c2 a7a6 e1g1 f8e8 c1d2 e7g6 e2g3 f6g4 c3e2 c7c6 h2h3 g4f6 a1c1 g6h4 f2f3 d8c7 g1f2 c7d7 e3e4 d6g3 f2g3 h4f3 e4e5 f3d2 c2d2 f6e4 d3e4 d5e4 d2g5');
INSERT INTO evaluations VALUES(24980,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(24980,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(24980,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(24980,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(24980,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(24980,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(24980,36,'cp 199','d4d5 c6b8 d1b3 b8d7 b3b7 f5e4 b1d2 e7e5 d2e4 f6e4 b7b3 f8d6 b3c2 f7f5 f3d2 e4d2 c1d2 e8g8 e1c1 a8b8 c1b1 g8h8 f2f3 d8g5 h1g1 d6c5 c4d3 c5e3 d2e3 g5e3 d3f5 d7f6 g1e1 f6d5 e1e3');
INSERT INTO evaluations VALUES(33896,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(33896,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(33896,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(33896,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(33896,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(33896,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(33896,36,'cp -23','g1f3 f8b4 d1d2 e8g8 a2a3 b4a5 b2b4 a5b6 e2e3 c7c6 f1e2 b8d7 e3e4 e6e5 e4d5 e5d4 c3e4 f6e7 d5c6 b7c6 e4g3 d7f6 e1g1 e7c7 e2d3 c6c5 b4c5 b6c5 f1e1 c8b7 d2g5 h7h6 g5e5 a8c8 g3f5 c7e5 e1e5 f8e8 e5e8 c8e8 f3d4 f6e4');
INSERT INTO evaluations VALUES(33895,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(33895,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(33895,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(33895,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(33895,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(33895,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(33895,36,'cp -26','f8b4 d1d2 b4a5 a2a3 e8g8 b2b4 a5b6 e2e3 c7c6 f1e2 b8d7 e3e4 e6e5 e4d5 e5d4 c3e4 f6e7 d5c6 b7c6 e4g3 c6c5 e1g1 d7f6 e2d3 e7c7 g3e4 f6e4 d3e4 a8b8 d2g5 c8e6 g5e5 c7e5 f3e5 b8e8 f1e1 g7g6 g2g3 g8g7 e5c6 e6d7 b4b5 f7f5 e4d5 e8e1 a1e1 g7f6');
INSERT INTO evaluations VALUES(44056,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(44056,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(44056,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(44056,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(44056,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(44056,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(44056,36,'cp 146','e4e5 f6g8 f1d3 d7d5 c3e2 c6c5 c2c3 c5d4 c3d4 c8d7 a2a3 d8b6 g1f3 d7b5 d3c2 b8c6 a1b1 g8e7 e1g1 e7f5 c2f5 e6f5 c1e3 f8e7 f1e1 b5e2 e1e2 b6d8 b1c1 a7a5 g2g3 e8g8 h2h4 d8d7 e2c2 a5a4 c2c3 f8b8 e3d2 c6a5 c3c7');
INSERT INTO evaluations VALUES(44033,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(44033,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(44033,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(44033,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(44033,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(44033,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(44033,36,'cp 152','f6g8 f1d3 d7d5 c3e2 c6c5 c2c3 c5d4 c3d4 c8d7 g1f3 d8b6 e2c3 b8c6 e1g1 g8e7 f1e1 h7h6 h2h4 a8c8 h4h5 c6a5 a1b1 a7a6 c1e3 a5c4 e3f4 b6d8 a2a3 b7b5 a3a4 b5b4 c3e2 d8b6 b2b3 c4a3 b1c1');
INSERT INTO evaluations VALUES(49938,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(49938,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(49938,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(49938,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(49938,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(49938,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(49938,36,'cp 46','b7b6 f1e2 f8b4 b1d2 b4e7 e1g1 e8g8 b2b3 c8b7 c1b2 c7c5 d1c2 c5d4 f3d4 b8c6 d4c6 b7c6 e2f3 d8c7 h2h3 a8c8 a1c1 c8d8 c1d1 f8e8 f3c6 d7c6 b2c3 c6c5 e3e4 c7f4 f1e1 f5e4 d2e4 d8d1 e4f6 e7f6 e1d1 f6c3 c2c3');
INSERT INTO evaluations VALUES(50247,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(50247,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(50247,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(50247,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(50247,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(50247,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(50247,36,'cp -12','d2d4 d7d5 c2c4 e7e6 e2e3 c7c5 c4d5 e6d5 f1b5 b8c6 e1g1 f8g7 b1c3 e8g8 d4c5 d8e7 c3d5 f6d5 d1d5 f8d8 d5b3 c8e6 b3a3 e6d5 f1f2 d5f3 f2f3 d8d1 f3f1 d1d5 a3a4 d5c5 b5e2 a8d8 e2f3 c6a5 a1b1 a5c4');
INSERT INTO evaluations VALUES(50201,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(50201,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(50201,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(50201,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(50201,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(50201,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(50201,36,'cp -19','d7d5 c2c4 e7e6 e2e3 c7c5 c4d5 e6d5 f1b5 b8d7 e1g1 f8g7 b5e2 e8g8 b1c3 c5d4 e3d4 d7b6 f3e5 f6e4 a2a4 f7f6 e5d3 f8e8 a4a5 e4c3 b2c3 b6c4 a5a6 b7b6 g2g4 g8h8 d3b4 g7f8 e2d3 f6f5 g4g5 f8b4 c3b4');
INSERT INTO evaluations VALUES(46781,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(46781,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(46781,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(46781,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(46781,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(46781,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(46781,36,'cp -7','c2c4 d5c4 e2e3 c7c5 b1c3 c5d4 d1d4 d8d4 f3d4 f8g7 d4b5 b8a6 f1c4 e8g8 e3e4 a6b4 e1g1 b4c2 a1b1 a7a6 b5c7 a8a7 c4b3 b7b5 c7b5 a6b5 b3c2 b5b4 c3a4 a7c7 c2b3 f6e4 c1e3 c8f5 f1c1 c7d7 a4b6 g7d4 e3d4 d7d4 b6d5 f8e8 c1c7 e4d2 c7e7 e8e7 d5e7 g8h8');
INSERT INTO evaluations VALUES(55936,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55936,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(55936,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55936,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(55936,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(55936,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(55936,36,'cp 0','e6e5 c1b2 b8c6 g1f3 e5e4 f3d4 c6d4 b2d4 h7h5 d2d3 f7f5 d3e4 d5e4 f2f3 h5h4 f3e4 h4g3 h2g3 h8h1 g2h1 d8g5 d1d3 g5c1 d3d1');
INSERT INTO evaluations VALUES(4564,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(4564,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(4564,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(4564,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(4564,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(4564,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(4564,36,'cp 24','e1g1 d7d6 a2a4 h7h6 g5f6 e7f6 c3d5 g7g6 h2h3 f6g7 c2c3 a8b8 f1e1 c6e7 d5e7 d8e7 b2b4 c8d7 a4a5 c7c6 d1d2 b8c8 a1d1 b7b5 a5b6 a7b6 d3d4 e5d4 c3d4 d7e6 c4e6 e7e6 d4d5 c6d5 e4d5 e6f6 d2e3 f8d8 e3b6 f6b2');
INSERT INTO evaluations VALUES(4562,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(4562,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(4562,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(4562,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(4562,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(4562,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(4562,36,'cp 25','d7d6 a2a4 h7h6 g5f6 e7f6 c3d5 g7g6 c2c3 f6g7 h2h3 a8b8 b2b4 c6e7 d5e7 d8e7 f1e1 c8d7 a4a5 c7c6 d1d2 b8c8 d2e3 a7a6 a1b1 c8c7 d3d4 e5d4 c3d4 d7e6 b1c1 e6c4 c1c4 e7d7 c4c1 f8e8 e3d3 d7e6 e1e3 c7e7 c1d1 h6h5 d1e1 e6d7 d3c2 e8c8');
INSERT INTO evaluations VALUES(3365,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3365,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3365,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3365,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(3365,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(3365,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(3365,36,'cp 17','a2a4 h7h6 g5d2 a7a5 h2h3 c8e6 d1e2 d8d7 a1d1 c6b4 d3d4 e6c4 e2c4 e5d4 f3d4 f8e8 c4e2 d6d5 e4d5 f6d5 e2f3 d5c3 b2c3 b4a6 d4f5 d7c6 f3c6 b7c6 f5e7 e8e7 d2e3 e7e4 d1d4 e4d4 c3d4 f7f6 c2c3 g8f7 e3d2 a8d8');
INSERT INTO evaluations VALUES(5989,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5989,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(5989,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5989,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5989,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5989,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5989,36,'cp 0','d2d3 g8f6 e1g1 e8g8 h2h3 a7a6 b1d2 c8e6 c2c3 c5a7 f1e1 f6d7 c4b3 d8f6 d2f1 f6g6 f3h4 g6f6');
INSERT INTO evaluations VALUES(5986,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5986,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(5986,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5986,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5986,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(5986,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5986,36,'cp 0','g8f6 b1c3 c6d4 f3d4 c5d4 d1f3 c7c6 c3e2 d4b6 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(11678,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(11678,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(11678,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11678,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11678,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(11678,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(11678,36,'cp 39','e1g1 d7d6 a2a4 c8d7 c1d2 f8e7 a4b5 a6b5 b1c3 a8a1 d1a1 d8b8 a1a2 e8g8 f1a1 c6b4 a2a7 b4c6 a7b8 f8b8 h2h3 g8f8 g2g4 c6d8 g1g2 f8e8 g4g5 h6g5 f3g5 c7c5 h3h4 d8e6 c3d5 e6d4 d5f6 e7f6 b3f7 e8e7 a1a7 d4c2 g2g3 b8f8 h4h5');
INSERT INTO evaluations VALUES(11677,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11677,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(11677,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11677,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(11677,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(11677,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(11677,36,'cp 33','d7d6 a2a4 c8d7 c1d2 b5b4 f1e1 f8e7 c2c3 e8g8 b3c4 b4c3 d2c3 d7e6 b1d2 d8d7 h2h3 e6c4 d2c4 d7e6 a1c1 f8b8 d1c2 f6h5 c4e3 h5f4 d3d4 e5d4 f3d4 c6d4 c3d4 e7f6 d4f6 e6f6 b2b3 b8e8 c1d1 f4e6 e3f5 a8b8 c2c4');
INSERT INTO evaluations VALUES(18040,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(18040,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(18040,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(18040,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(18040,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(18040,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(18040,36,'cp 132','a2a4 b5a4 a1a4 c8a6 d2d3 h7h6 c1e3 g8f6 e1g1 f8e7 b1c3 e8g8 h2h3 a6c8 c3d5 d7d6 d1d2 c8e6 c2c3 f6d7 a4a1 e7f6 a1a3 d7b6 f1a1 a8b8 b3a2 f8e8 d5f6 d8f6 a2e6 f6e6 b2b4 a5b4 c3b4 c6e7 b4b5');
INSERT INTO evaluations VALUES(20317,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(20317,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(20317,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(20317,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(20317,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(20317,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(20317,36,'cp -57','g1f3 c5d4 c1f4 h7h6 h2h4 f8b4 b1d2 g8e7 f1d3 d5c4 d3c4 e7d5 f4g3 d5e3 f2e3 d4e3 e1g1 b4d2 f3d2 e3d2 f1f2 d8d4 c4b3 e8g8 d1d2 d4d2 f2d2 b7b6 a1c1 c8b7 d2d7 c6a5 b3d1 b7d5 c1c7 a8d8 d1a4 d8d7 c7d7 d5a2 b2b4 a5c4 d7a7');
INSERT INTO evaluations VALUES(22136,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(22136,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(22136,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(22136,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(22136,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(22136,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(22136,36,'cp 196','c2c3 g8f6 c3d4 c7c6 d2d3 d6d5 c4b3 f8b4 a2a3 b4a5 b1c3 e8g8 f2f3 c8e6 c1e3 g7g6 a1c1 f6h5 f3f4 h5g7 b3c2 f7f5 e4e5 b7b5 c3e2 d8d7 b2b4 a5d8 d1d2 a7a5 h2h3 a5b4 a3b4 d8b6 g1h2');
INSERT INTO evaluations VALUES(22132,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(22132,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(22132,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(22132,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(22132,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(22132,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(22132,36,'cp 194','g8f6 c3d4 c7c6 c4b3 d6d5 d2d3 f8b4 a2a3 b4a5 b1c3 e8g8 f2f3 c8e6 c1e3 g7g6 d1d2 f6h5 f3f4 h5g7 b3c2 d8d7 h2h3 a8d8 b2b4 a5b6 a1d1 b6c7 c3a4 d7e7 d1e1 g8h8 f1f3 f7f6 e3f2 d8e8 e4e5 g7f5 g2g4 f5g7 b4b5 f6e5 d4e5');
INSERT INTO evaluations VALUES(29545,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(29545,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(29545,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29545,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29545,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29545,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(29545,36,'cp 3','e2e3 e7e6 f1d3 f5d3 c2d3 c7c5 d1b3 d8d7 d4c5 f8c5 g1f3 b8c6 e1g1 e8g8 f3e5 c6e5 f4e5 f8c8 b1d2 c5e7 a2a4 f6e8 d2f3 e8d6 f1c1 h7h6 e5c3 c8c7 c3a5 c7c1 a1c1 a8c8 c1c3 e7f6 c3c8 d6c8 a5c3 f6c3 b3c3');
INSERT INTO evaluations VALUES(35182,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(35182,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(35182,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(35182,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(35182,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(35182,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(35182,36,'cp 202','g1f3 h4c4 e2e4 c4e6 c3d5 f8d6 f1c4 e6g6 d2d3 h7h5 e1g1 h5h4 d3d4 h4h3 g2g3 b8c6 d4e5 c6e5 f3e5 d6e5 c1f4 e5f4 d5f4 g6g5 f1e1 g8e7 e4e5 g5h6 a1c1 e8g8 c4f1 c7c6 f1h3 e7g6 f4g6 h6g6 d1d6 a7a5 c1d1 g6d6 e5d6 a8a6 e1e7 f8d8 d1e1 g8f8 e1e4 c6c5 e4f4 f7f6');
INSERT INTO evaluations VALUES(35181,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(35181,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(35181,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(35181,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(35181,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(35181,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(35181,36,'cp 216','h4c4 e2e4 c4e6 c3d5 f8d6 f1c4 e6g6 d2d3 h7h5 e1g1 c7c6 f3h4 g6h7 h4f5 h7g6 f5d6 g6d6 d5e3 g8e7 c4b3 d6c7 f2f4 d7d5 f4e5 c8e6 d3d4 c7d7 e4d5 c6d5 c1d2 b8c6 d2c3 e7f5 e3f5 e6f5 b3d5 d7d5 f1f5 e8c8 d1f1 c6d4 f5f4 d4e6 f4f7 d5c5 g1h1 c5c6 a1d1 d8d1 f1d1');
INSERT INTO evaluations VALUES(37394,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(37394,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37394,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37394,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(37394,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37394,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(37394,36,'cp 12','b1c3 d7d5 a1c1 d5c4 e2e3 c8e6 f3g5 e6d5 e3e4 h7h6 e4d5 h6g5 f4g5 f6d5 f1c4 d5b6 c4b3 d8d4 d1d4 g7d4 c3b5 d4b2 c1c2 b2e5 b5c7 e5c7 c2c7 b8c6 c7b7 e7e6 b7c7 c6d4 b3d1 e6e5 h2h4 d4e6 c7c6 f8c8 c6c8 a8c8 h1h3 e6g5 h4g5 b6d7 h3d3 d7c5 d3d6 c5e4 d6a6 e4g5 a6a7 c8c1');
INSERT INTO evaluations VALUES(38800,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(38800,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(38800,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(38800,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(38800,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(38800,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(38800,36,'cp 74','a2a3 b4c3 d2c3 d5e4 d1d8 e8d8 f3g5 d8e7 g5e4 g8f6 e4f6 g7f6 c1f4 e6e5 f4g3 c8f5 f2f4 b8d7 e1c1 a8g8 d1e1 e7d6 f1c4 f5e6 f4e5 f6e5 e1d1 d6e7 g3h4 f7f6 c4e6 e7e6 d1d7 e6d7 h4f6 h7h5 f6h8 g8h8 c1d2 h5h4 d2e3 b7b5 h1a1 h8g8 e3f2 g8f8 f2g1 f8f4 g2g3 h4g3 h2g3 f4f3 g1g2 e5e4 a1f1 f3e3');
INSERT INTO evaluations VALUES(39220,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(39220,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(39220,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(39220,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(39220,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(39220,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(39220,36,'cp 19','g1f3 d7d6 a2a3 a7a5 b1c3 b8c6 e2e3 g8f6 f1e2 e8g8 e1g1 h7h6 b2b3 c8f5 c1b2 f8e8 h2h3 c5b6 f3d2 c6b8 b3b4 c7c6 d2e4 f6e4 c3e4 d8e7 e4g3 f5g6 e2h5 b8d7 h5g6 f7g6 d1b3 a5a4 b3c2 d6d5 a1d1 d7f6 c4c5');
INSERT INTO evaluations VALUES(39490,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(39490,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(39490,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(39490,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(39490,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(39490,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(39490,36,'cp 31','e2e4 d7d5 c4d5 e6d5 e4e5 c7c5 g1f3 b8c6 f1b5 c8d7 e1g1 c6e5 f3e5 d7b5 f1e1 g8f6 b1c3 b5c6 d4c5 e8g8 c1e3 f6d7 e5d3 e7f6 d3f4 f6c3 b2c3 f8e8 f4d5 d7c5 d5b4 d8d1 e1d1 c5a4 a1c1 h7h6 d1d4 a4c5 c1d1 c5e6 b4c6');
INSERT INTO evaluations VALUES(39491,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(39491,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39491,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(39491,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(39491,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(39491,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(39491,36,'cp 28','d7d5 c4d5 e6d5 e4e5 c7c5 g1f3 b8c6 f1b5 c8d7 b5c6 b7c6 e1g1 c5d4 f3d4 e7c5 b1c3 g8e7 c3a4 c5b6 b2b4 e8g8 c1e3 d8c7 f1e1 e7f5 d4f5 d7f5 a1c1 f8e8 e3d4 e8e6 a4c3 c7e7 b4b5 e7g5 d4b6 a7b6 b5c6 e6c6 c3d5 c6c1 d1c1 g5c1 e1c1 f5e6');
INSERT INTO evaluations VALUES(43205,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(43205,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43205,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43205,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43205,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(43205,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(43205,36,'cp 18','b1d2 d7d5 e2e3 e8g8 f1e2 b8d7 h2h3 c7c5 g1f3 b7b6 e1g1 c8b7 a2a4 a7a6 d1b1 f8e8 b2b3 e7e6 f1d1 d8e7 f3e5 d7e5 f4e5 g7f8 b1b2 f6d7 e5g3 a8c8 d2f3 f8g7 g3h2 a6a5 b2c2 e6e5 e2b5 e8d8 f3e5');
INSERT INTO evaluations VALUES(48738,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(48738,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(48738,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(48738,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(48738,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48738,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48738,36,'cp -5','e1g1 f8e7 c2c3 e8g8 h2h3 c7c5 d3c2 d5d4 c3d4 c5d4 d2d3 b8c6 b1d2 c8e6 c2b3 e6d5 f1e1 a7a5 d2e4 f6e4 d3e4 d5b3 d1b3 d8d7 c1d2 a5a4 b3d3 f8d8 a2a3 c6a5 d2a5 a8a5 e1e2 d7b5 a1d1 b5d3 d1d3');
INSERT INTO evaluations VALUES(48737,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(48737,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(48737,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(48737,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(48737,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48737,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(48737,36,'cp -19','f8e7 h2h3 e8g8 c2c3 e7d6 d3c2 b8c6 f1e1 c6e7 d2d4 e7g6 c1g5 h7h6 g5e3 f8e8 b1d2 c8d7 c2g6 f7g6 c3c4 c7c6 d1b3 d8b6 f3e5 b6b3 a2b3 g6g5 e5d7 f6d7 d2f3 d7f6 e3d2 e8e1 d2e1 g8f7 e1c3 a7a6 f3e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(41110,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(41110,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(41110,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(41110,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(41110,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(41110,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(41110,36,'cp -7','c2c3 e8g8 h2h3 c7c5 d3c2 d5d4 c3d4 c5d4 d2d3 b8c6 f1e1 c8e6 b1d2 h7h6 c2b3 e6d5 d2c4 e7d6 c4d2 d6c7 d2e4 f6e4 d3e4 d5b3 d1b3 d8d7 c1d2 f8e8 a1c1 c7b6 b3d3 c6e5 f3e5 e8e5 b2b4 d7b5 d3b5 e5b5');
INSERT INTO evaluations VALUES(49457,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(49457,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(49457,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(49457,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(49457,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(49457,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(49457,36,'cp 52','e4e5 f6d5 d2d4 b7b5 f1b5 c7c5 c2c4 c5d4 d1d4 d8c7 g1e2 f8c5 d4d3 d5b6 a2a4 a7a6 a4a5 c7b7 a5b6 b7g2 h1f1 a6b5 a1a8 g2a8 c4b5 c5b6 b1a3 d7d5 e5d6 e8g8 a3c4 b8d7 c4b6 d7b6 f1g1 b6d7 b2b4 c8b7 c1b2 f7f6 d3c4 b7d5 c4c7 f8f7 b5b6');
INSERT INTO evaluations VALUES(49350,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(49350,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(49350,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(49350,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(49350,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(49350,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(49350,36,'cp 25','f6d5 d2d4 b7b5 f1b5 c7c5 c2c4 c5d4 d1d4 d8a5 c1d2 d5b4 d4e4 a7a6 e4a8 a5c7 a8e4 a6b5 d2b4 c8b7 e4e2 f8b4 b1c3 b5c4 g1f3 c7b6 f3d2 b8c6 d2c4 b6d4 a1d1 d4f4 g2g3 f4g5 a2a3 b4c3 b2c3 e8g8 e1g1 b7a6 e2e4 a6c4 e4c4 c6e5 c4e2 f8c8 g1h1 c8c3');
INSERT INTO evaluations VALUES(51416,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51416,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(51416,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(51416,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(51416,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(51416,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(51416,36,'cp 37','e3e4 c8b7 f1d3 c7c5 c2c3 b8c6 g1f3 c5d4 c3d4 c6b4 e1g1 g8e7 d4d5 e7g6 b1c3 b4d3 d1d3 f8c5 c1e3 e8g8 f1e1 c5e3 d3e3 h7h6 a1d1 d8f6 f3d4 e6d5 e4e5 f6g5 e3g5 h6g5 c3b5 f7f6 e5f6 f8f6 g2g3 a7a6 b5c3 b6b5 a2a3 a8c8 h2h4 g5g4');
INSERT INTO evaluations VALUES(54908,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(54908,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(54908,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(54908,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(54908,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(54908,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(54908,36,'cp 93','g1f3 g8f6 b1c3 a7a6 f1d3 e7e6 h2h3 c7c5 d4d5 f8e7 a2a4 e6d5 e4d5 b8d7 c1f4 d7b6 d3e2 c8f5 a4a5 b6d7 e1g1 e8g8 f1e1 f8e8 f3d2 b7b5 a5b6 d7b6 d2f1 f6e4 a1a6 e7f6 c3e4 f5e4 a6a8 d8a8 c2c4 f6b2 f4d6 b2d4 f1g3 e4g6');
INSERT INTO evaluations VALUES(54907,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(54907,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(54907,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(54907,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(54907,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(54907,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(54907,36,'cp 96','g8f6 b1c3 e7e6 f1d3 a7a6 d1e2 b8c6 e4e5 f6d7 a2a3 d6e5 d4e5 d7c5 e1g1 c5d3 f1d1 f7f5 e5f6 d8f6 d1d3 f6f7 c1d2 f8e7 c3e4 c8d7 d2c3 e8c8 a1e1 c8b8 b2b4 d7c8 d3e3 h8f8 f3e5 c6e5 c3e5 e7f6 e4c5 f6e5 e3e5');
INSERT INTO evaluations VALUES(11335,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(11335,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(11335,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(11335,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(11335,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(11335,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(11335,36,'cp 71','f8d6 e1g1 f6g4 d3c2 e6e5 d4e5 g4e5 f3e5 d6e5 f1e1 e8g8 d2f3 e5f4 e3f4 c8g4 d1d3 g7g6 c2b3 d8f6 b3d5 a8d8 a1d1 g4e6 c3c4 f6f4 d3c3 f8e8 g2g3 f4g4 a2a3 g4f5 f3h4 f5g5 g1g2 d8d5 c4d5 e6d5 h4f3');
INSERT INTO evaluations VALUES(7139,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(7139,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(7139,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(7139,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(7139,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(7139,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(7139,36,'cp 88','e1g1 e8g8 d3c2 c6e7 f1e1 d6f4 e3f4 b7b6 f3e5 c7c5 g2g4 c5d4 g4g5 d4c3 b2c3 h6g5 f4g5 f6e4 d2e4 d5e4 d1h5 g7g6 h5g4 d8c7 c2e4 a8b8 g4f4 c8b7 a1d1 b7e4 d1d7 e7d5 f4e4 c7c3 e5f7 f8f7 e4e6 b8f8 d7d5 g8g7 d5d6');
INSERT INTO evaluations VALUES(14725,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(14725,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(14725,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(14725,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(14725,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(14725,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(14725,36,'cp 140','g1f3 f8d6 f1c4 c6a5 c4d3 c7c5 e1g1 c8d7 a1b1 a8c8 d1e2 a7a6 e4e5 d6e7 d4c5 d8c7 h2h4 h7h6 e2e4 d7c6 e4g4 c6f3 g2f3 c7e5 g4a4 a5c6 b1b7 e8g8 c1f4 e5c5 a4e4 g7g6 f4h6 f8d8 g1g2 c5c3');
INSERT INTO evaluations VALUES(14724,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(14724,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(14724,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(14724,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(14724,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(14724,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(14724,36,'cp 140','f8d6 f1c4 c6a5 c4d3 c7c5 e1g1 c8d7 d1e2 c5d4 c3d4 e8g8 e4e5 d6e7 e2e4 g7g6 c1h6 d7c6 e4e3 c6f3 h6f8 e7f8 g2f3 d8h4 e3e4 h4g5 g1h1 a8d8 f1d1 d8d5 a1b1 g5d8 d3e2 a7a6 a2a4 d8d7 f3f4 b7b5 a4b5 a6b5 e4e3 a5c4 e2c4 b5c4 b1b8');
INSERT INTO evaluations VALUES(14851,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14851,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(14851,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14851,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(14851,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(14851,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(14851,36,'cp 18','f8d6 b1d2 e6e5 d4e5 c6e5 f3e5 d6e5 d2f3 c8g4 c2c4 c7c6 h2h3 g4f3 d1f3 d8e7 a1b1 a8d8 f1d1 e8g8 c1d2 g7g6 a2a3 d8d7 d2b4 e5d6 b4e1 d6e5 c4d5 f6d5 d1d2 d5f6 d3f1 d7d2 e1d2 f8d8 d2b4 e7e6 b4e1');
INSERT INTO evaluations VALUES(9257,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9257,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9257,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(9257,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9257,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(9257,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9257,36,'cp 15','b1d2 e6e5 f3e5 c6e5 d4e5 d6e5 d2f3 c8g4 c2c4 c7c6 h2h3 g4f3 d1f3 d8e7 c4d5 f6d5 f1d1 e8g8 a1b1 g7g6 c1d2 a8d8 d3c4 d8d7 d2e1 f8d8 g2g3 h7h5 h3h4 g8g7 c4b3 a7a6 e1a5 e5c7 a5c7 d5c7 d1d7 d8d7 b1d1 d7d1 f3d1 c7e6 g1g2 e6c5');
INSERT INTO evaluations VALUES(19178,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(19178,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(19178,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(19178,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(19178,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(19178,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(19178,36,'cp 33','g8f6 d5d3 d7d5 e4d5 f6d5 b1c3 d8e7 f1e2 d5b4 d3d1 c8g4 c1e3 a8d8 e2b5 g4d7 d1e2 d7b5 e2b5 e7d7 b5e2 d7e6 e1g1 f8e7 a2a3 b4c6 e2b5 b7b6 f1e1 e6c8 e3g5 f7f6 g5f4 e8g8 b5c4 g8h8 a1d1 d8d1 c3d1 f8d8 d1e3 h7h6 c4f7 c8d7 f7h5 e7f8 h2h4 d7e8 h5e8 d8e8 c2c3 h8g8 h4h5');
INSERT INTO evaluations VALUES(12660,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(12660,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12660,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(12660,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(12660,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(12660,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(12660,36,'cp 52','d5e5 f8e7 b1c3 e8g8 f1e2 d7d6 e5g3 d6d5 e4d5 f8e8 g3d3 a7a6 e1g1 b7b5 e2f3 e7d6 c1g5 c5c4 d3d2 h7h6 g5f4 b5b4 c3d1 d6f4 d2f4 f6d5 f4c4 c8e6 c4d4 a6a5 f3d5 e6d5 d1e3 d5e4 d4d8 a8d8 f1e1 f7f5 f2f3 f5f4 e3c4 e4c2 e1e8 d8e8 c4a5 e8e2');
INSERT INTO evaluations VALUES(12645,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(12645,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(12645,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(12645,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(12645,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(12645,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(12645,36,'cp 25','f8e7 b1c3 e8g8 f1e2 d7d6 e5g3 d6d5 e4d5 f8e8 g3d3 a7a6 e1g1 b7b5 e2f3 e7d6 c1g5 c5c4 d3d2 h7h6 g5f4 b5b4 c3e2 f6e4 d2d4 e4g5 f4g5 d8g5 f1e1 c8f5 e2g3 e8e1 a1e1 f5c2 d4c4 g5d2 e1f1 a8b8 g3e4 c2e4 f3e4 d2b2 c4a6 b2f6 g2g3 g7g6 g1g2 f6e5 a6d3');
INSERT INTO evaluations VALUES(23065,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(23065,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(23065,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(23065,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(23065,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(23065,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(23065,36,'cp 28','d4e6 g2g3 d7d6 e5c4 g7g6 f1g2 f8g7 b1c3 g8e7 e1g1 e8g8 f2f4 f7f5 e4e5 d6e5 c4e5 c7c6 c1e3 g6g5 d1h5 g7e5 f4e5 e7g6 h5e2 d8c7 a1e1 c7g7 e3f2 f8f7 c3a4 f5f4 a4c5 e6c5 f2c5 c8e6 c2c3 a8e8');
INSERT INTO evaluations VALUES(23886,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(23886,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(23886,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(23886,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(23886,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(23886,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(23886,36,'cp 109','c4e6 f7e6 d1g4 d8f6 g4e6 g8e7 e6f6 g7f6 g5f3 c6b4 b1a3 b4a6 b2b3 d6d5 e1e2 e8c8 c2c3 c5b6 b3b4 c7c5 a1b1 c5b4 c3b4 h8g8 g2g3 c8b8 c1h6 a6c7 h1c1 g8g6 h6e3 b6e3 e2e3 h7h5 f3h4 g6g8 e3e2 c7e6 b4b5 d5e4 d3e4 g8g4 f2f3 g4g8 a3c2');
INSERT INTO evaluations VALUES(23884,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(23884,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(23884,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(23884,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(23884,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(23884,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(23884,36,'cp 109','f7e6 d1g4 d8f6 g4e6 g8e7 e6f6 g7f6 g5f3 c6b4 b1a3 b4a6 b2b3 d6d5 e1e2 e8c8 c2c3 c5b6 b3b4 c7c5 a1b1 h8g8 g2g3 c5b4 c3b4 c8b8 a3b5 d8c8 a2a3 d5e4 d3e4 g8d8 c1d2 c8c4 b5c3 a6c7 h1c1 h7h5 c1d1 e7c8 d2e1 c8d6 e2f1 c7e6');
INSERT INTO evaluations VALUES(23898,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(23898,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(23898,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(23898,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(23898,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(23898,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(23898,36,'cp 115','d1g4 d8f6 g4e6 g8e7 e6f6 g7f6 g5f3 c6b4 b1a3 b4a6 b2b3 e8c8 c2c3 c5b6 b3b4 d6d5 e1e2 c7c5 a1b1 c5b4 c3b4 c8b8 g2g3 d8d7 c1h6 a6c7 h1c1 h8d8 h6e3 h7h5 e3b6 a7b6 b1b2 e7c6 a3c2 d5e4 d3e4 c7b5 c2e3 b5d6 f3d2 c6d4 e2f1 b6b5 c1c3 h5h4 g3h4 d7g7');
INSERT INTO evaluations VALUES(34258,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(34258,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(34258,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(34258,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(34258,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(34258,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(34258,36,'cp 41','a5c5 e2e4 g8f6 f1d3 b8c6 g1f3 g7g6 b2b4 c5d6 e1g1 f8g7 d1e2 d6c7 b4b5 c6e5 f3e5 c7e5 f2f4 e5h5 e4e5 h5e2 d3e2 f6e4 e2f3 f7f5 e5f6 e4f6 c1e3 d7d5 e3d4 c8f5 f1e1 e8f7 b1d2 h8d8 g2g4 f5d3 a2a4 a8c8 a1c1');
INSERT INTO evaluations VALUES(34252,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(34252,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(34252,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(34252,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(34252,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(34252,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(34252,36,'cp 43','e2e4 g8f6 f1d3 b8c6 g1f3 g7g6 d1e2 f8g7 e1g1 c5h5 c1f4 d7d5 b1d2 d5e4 d2e4 e8g8 e4f6 g7f6 e2e3 f8d8 f1e1 h5d5 d3f1 c8g4 f3d2 g6g5 e3g3 d5f5 f4e3 h7h6 h2h3 g4h5 f2f4 a8c8 d2f3 b7b5 g3f2 h5g6');
INSERT INTO evaluations VALUES(37379,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37379,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37379,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(37379,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37379,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37379,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37379,36,'cp 0','c7c6 e3e4 c6c5 g1e2 d7d6 a1b1 b8c6 e1g1 a8b8 d2d3 a7a6 a2a3 b7b5 c4b5 a6b5 b2b4 c8g4 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 e4e5 f6e8 b1a1 a3b2 a1b1');
INSERT INTO evaluations VALUES(35684,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35684,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35684,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35684,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35684,36,'cp 0','e3e4 d7d6 g1e2 c6c5 e1g1 b8c6 d2d3 a7a6 a1b1 a8b8 a2a3 b7b5 c4b5 a6b5 b2b4 c8g4 c1e3 f6d7 f2f3 g4e6 f3f4 c6d4 d1d2 d4b3 d2b2 b3d4');
INSERT INTO evaluations VALUES(37419,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(37419,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37419,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37419,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(37419,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37419,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37419,36,'cp 5','g1f3 d7d6 a2a4 b8d7 h2h3 b7b6 f1c4 a7a6 e1g1 c8b7 f1e1 d6d5 c4f1 c7c5 c3b1 c5c4 c2c3 f6e4 b1d2 b6b5 d1c2 e4d2 c2d2 b7c6 a4b5 a6b5 f3e5 d7e5 f4e5 g7e5 d4e5 d8c7 f2f4');
INSERT INTO evaluations VALUES(37420,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37420,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37420,36,'cp 0','d7d6 a2a4 b8d7 h2h3 b7b6 f1c4 a7a6 e1g1 c8b7 f1e1 d6d5 c4f1 c7c5 c3b1 f6e4 b1d2 e4d2 d1d2 e7e6 c2c3 f8e8 f4g5 d8c7 g5f4');
INSERT INTO evaluations VALUES(37729,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(37729,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(37729,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(37729,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(37729,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(37729,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(37729,36,'cp 71','d7d5 c4d5 c6d5 c1f4 b8c6 e2e3 g8f6 h2h3 e8g8 f1e2 f6e4 a1c1 c8e6 e1g1 a8c8 d1b3 d8d7 b3a3 h7h6 f4h2 g6g5 f1d1 f7f5 e2d3 d7d8 f3e5 c6e5 h2e5 a7a6 a3b4 g7e5 d4e5 e4c3 c1c3 c8c3 b4c3 d8b6 c3a3 f8c8');
INSERT INTO evaluations VALUES(47117,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(47117,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(47117,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(47117,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(47117,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(47117,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47117,36,'cp 0','h7h6 g5f6 e7f6 e2e3 b8d7 d1c2 f8d6 c3c4 d5c4 f1c4 g7g6 b1c3 f6f5 c3b5 d6b4 b5c3');
INSERT INTO evaluations VALUES(48963,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(48963,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(48963,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(48963,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(48963,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(48963,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(48963,36,'cp -91','d8d5 g1f3 b8c6 b1c3 f8b4 c1d2 b4c3 d2c3 f7f6 a2a3 c8f5 f3h4 f5d7 d1h5 g7g6 h5f3 d5e6 e1c1 e8c8 f3e3 e6a2 h4f3 d7e6 h1e1 e6d5 g2g3 h8e8 f3d2 f6f5 f2f3 b7b6 e3f2 g6g5');
INSERT INTO evaluations VALUES(49186,30,'cp -315',NULL);
INSERT INTO evaluations VALUES(49186,31,'cp -349',NULL);
INSERT INTO evaluations VALUES(49186,32,'cp -347',NULL);
INSERT INTO evaluations VALUES(49186,33,'cp -417',NULL);
INSERT INTO evaluations VALUES(49186,34,'cp -417',NULL);
INSERT INTO evaluations VALUES(49186,35,'cp -382',NULL);
INSERT INTO evaluations VALUES(49186,36,'cp -401','d8e7 f7e5 e7e5 d2d4 e5f5 f2f3 e4f6 f1d3 f5e6 e1f2 e6f7 c1f4 b8c6 h1e1 f8e7 f4c7 c6d4 b1c3 d4e6 c7g3 e8g8 d3c4 g8h8 c3e4 d7d5 c4d5 f8d8 e4f6 e7f6 f2g1 f6b2 a1b1 b2c3 d1d3 c3e1 b1e1 f7f8 c2c4 e6f4 g3f4 f8f4');
INSERT INTO evaluations VALUES(50052,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(50052,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(50052,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(50052,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(50052,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(50052,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(50052,36,'cp 101','g1f3 d7d5 b1c3 a7a6 d1c2 d5c4 e2e3 b8d7 f1c4 b7b5 c4e2 c8b7 e1g1 e7d5 e3e4 d5f6 e4e5 f6d5 c1g5 f8e7 c3d5 b7d5 g5e7 d8e7 c2c7 e8g8 f1c1 h7h6 c7a5 d7b8 c1c5 b8c6 a5c3 e7d7 a1d1 a8c8 a2a3 c6e7 h2h3 d5c4 e2d3 c4d5 f3h4 c8c5 d4c5');
INSERT INTO evaluations VALUES(53478,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(53478,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(53478,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(53478,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(53478,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(53478,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(53478,36,'cp 113','e2e4 f8b4 f1d3 d5e4 d3e4 g8f6 e4c2 c6c5 a2a3 b4a5 e1g1 e8g8 d4c5 a5c3 b2c3 b7b6 f3d4 c8b7 c5b6 a7b6 c1f4 b8c6 f1e1 d8d7 d1d3 f8d8 a1d1 a8a3 d4b5 d7d3 c2d3 a3a4 f4c7 d8d7 c7b6 c6a5 c4c5 b7c6 b5d6 f6d5 d1b1 d5c3');
INSERT INTO evaluations VALUES(2034,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(2034,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(2034,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(2034,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(2034,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(2034,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(2034,36,'cp 92','d7d6 f1b5 c8d7 b1c3 c6e5 b5d7 e7d7 c1f4 f7f6 f4e3 g8e7 e1g1 a7a6 f3e5 d6e5 d1e2 e7g6 c3d5 d7f7 f1d1 f8d6 c2c4 c7c6 d5b6 a8d8 c4c5 d6c7 d1d8 c7d8 b6c4 d8c7 a1d1 e8g8 h2h3 f8e8 b2b4 g6f8 a2a4 f8e6 c4d6 c7d6 c5d6 e8d8 e3b6 d8d7 b4b5 a6b5 a4b5 h7h6 e2d3 c6b5');
INSERT INTO evaluations VALUES(5922,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(5922,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(5922,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(5922,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(5922,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(5922,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(5922,36,'cp 0','h7h5 b1d2 h5h4 g3d6 c7d6 c3c4 d5c4 f1c4 d6d5 c4b5 c8d7 b5c6 d7c6 a2a4 g8f6 h2h3 d8b6 b2b3 f6e4 g1f3 b6a6 f3e5 a6a5 e5f3');
INSERT INTO evaluations VALUES(10906,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(10906,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(10906,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(10906,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(10906,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(10906,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(10906,36,'cp -13','f6e4 c3e4 d7d5 c4d3 d5e4 d3e4 c8g4 h2h3 g4h5 c2c3 e8g8 d1c2 g8h8 e1g1 e7d6 d2d4 e5d4 f3d4 c6d4 c3d4 h5g6 c1d2 d8d7 e4g6 h7g6 c2b3 b7b6 a1c1 f8e8 b3f3 h8g8 f1e1 a7a5 h3h4 e8e1 c1e1 a8e8 e1e8 d7e8 g2g3 e8b5 d2c3 b5c4 f3a8 g8h7 a8f3');
INSERT INTO evaluations VALUES(14298,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(14298,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(14298,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(14298,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(14298,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(14298,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(14298,36,'cp -43','f1b5 c8d7 g1e2 e7e6 e1g1 f8d6 c2c4 d5c4 e3e4 c6a5 b5a4 b7b5 a4c2 d6e7 b1c3 b5b4 e4e5 b4c3 e5f6 e7f6 b2c3 c7c5 d4c5 a8c8 a1b1 e8g8 c1e3 d8c7 g1h1 f8d8 d1e1 d7c6 e3f4 e6e5 f4g3');
INSERT INTO evaluations VALUES(14168,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(14168,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(14168,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(14168,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(14168,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(14168,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(14168,36,'cp -50','e7e6 g1e2 c8d7 c2c4 d5c4 b5c4 c6a5 c4d3 c7c5 b1c3 f8e7 e1g1 e8g8 g1h1 c5d4 e2d4 d8b6 d4c2 f8d8 e3e4 a5c6 f3f4 c6b4 c1e3 e7c5 e3c5 b6c5 c2b4 c5b4 d1e2 d7c6 f4f5 e6e5 d3c4 h7h6 a2a3 b4b6 a1d1 d8d1');
INSERT INTO evaluations VALUES(20647,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(20647,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(20647,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(20647,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(20647,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(20647,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(20647,36,'cp 30','d8c7 d1f3 c8d7 d4c6 b7c6 b5e2 g8f6 f3g3 f8e7 e1g1 e8g8 c1h6 f6e8 a1d1 g8h8 h6g5 e7g5 g3g5 e6e5 g5e7 h8g8 d1d3 a8c8 b2b3 a7a5 f2f4 e5f4 e7h4 d7e6 g1h1 h7h6 h4f4 f7f5 f4d2 c8d8');
INSERT INTO evaluations VALUES(22021,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(22021,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(22021,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(22021,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(22021,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(22021,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(22021,36,'cp 16','c6e5 f3e5 d6e5 f2f4 d8e7 f4e5 c8e6 c4d5 c7c6 d5e6 e7e6 d1d4 g8e7 b1d2 e7g6 d2c4 a8d8 c4d6 f8d6 e5d6 e8g8 d4a7 d8d6 a7b7 e6e4 b7b3 g6h4 b3g3 d6g6 c1g5 e4c2 g3h4 c2c5 h4f2 c5g5 a2a4 g5a5 b2b3 g6f6 f2c2 f6f1 a1f1');
INSERT INTO evaluations VALUES(22739,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(22739,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(22739,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(22739,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(22739,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22739,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22739,36,'cp -4','e5e4 f3d4 c6d4 e3d4 d8f6 d4d5 f8c5 d1e2 f6e7 d2d3 e4d3 e2e7 g8e7 f1d3 d7d6 c1e3 c8f5 d3f5 e7f5 e3c5 d6c5 b1c3 e8d7 e1d2 h8e8 h1e1 b7b6 e1e8 a8e8 a1e1 e8e1 d2e1 c7c6 e1e2 c6d5 c3d5 d7d6 g2g3 f5d4 e2e3 d6e5 d5c3 g7g5 f2f3 f7f6');
INSERT INTO evaluations VALUES(32655,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(32655,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(32655,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(32655,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(32655,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(32655,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(32655,36,'cp 21','b8c6 h2h3 g4h5 g2g4 h5g6 e2b5 e7e6 d2d3 f8b4 c1d2 e8g8 b5c6 b7c6 a2a3 b4d6 c3e4 a5d5 e4f6 g7f6 f1e1 g8g7 d2c3 f8g8 f3d2 h7h5 d1f3 d5f3 d2f3 h5g4 h3g4 g7f8 c3f6 g6h5 f3e5 h5g4 g1f1 d6e5 f6e5 f8e7 f2f4 f7f6 e5d4 g4f5 f1f2 g8g4 f2e3 a8g8 d4c5 e7f7 e1h1 g4g2');
INSERT INTO evaluations VALUES(27414,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(27414,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(27414,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(27414,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(27414,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(27414,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(27414,36,'cp 39','h2h3 g4h5 g2g4 h5g6 e2b5 e7e6 f3e5 e8c8 e5c6 b7c6 b5c6 h7h5 g4g5 a5g5 c6g2 f8d6 d2d3 g5e5 f2f4 e5c5 g1h1 h5h4 d1f3 g6h5 f3b7 c8d7 c1d2 h5g6 b2b4 c5b6 b7b6 a7b6 c3b5 f6h5 h1g1 h5g3 f1e1 d6e7 a2a4 e7f6 a1b1 h8h5 b1c1 d8h8 d3d4 g3f5 c2c3 f5d6 b5d6 c7d6 b4b5');
INSERT INTO evaluations VALUES(35702,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(35702,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(35702,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(35702,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(35702,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(35702,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(35702,36,'cp 9','d7d5 c4d5 f6d5 d2d4 b8c6 c3d5 d8d5 c1e3 c5d4 f3d4 d5c4 g2c6 b7c6 d1c1 c8a6 c1c4 a6c4 a1c1 g7d4 e3d4 c4e2 f1e1 e2f3 e1e7 f8e8 e7e3 f3d5 a2a4 a7a5 e3e8 a8e8 d4c3 e8a8 c3e1 a8b8 e1a5 b8b2 a5c3 b2e2 a4a5');
INSERT INTO evaluations VALUES(38802,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(38802,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38802,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(38802,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(38802,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(38802,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(38802,36,'cp 0','d5c4 e2e3 b7b5 a2a4 g8f6 a4b5 f6d5 a1c1 d5f4 e3f4 c8b7 g1f3 d8d6 f1c4 d6f4 c4e2 e8g8 e1g1 b8d7 g2g3 f4f6 f3h4 f8d8 e2f3 a8b8 f3b7 b8b7 c1a1 h7h6 h4f3 f6f5 f3h4');
INSERT INTO evaluations VALUES(38874,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38874,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38874,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38874,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38874,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38874,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38874,36,'cp 0','e2e3 b7b5 a2a4 g8e7 a4b5 e7d5 a1c1 d5f4 e3f4 c8b7 g1f3 d8d6 f1c4 d6f4 c4e2 e8g8 e1g1 b8d7 g2g3 f4f6 f3h4 f8d8 e2f3 a8b8 f3b7 b8b7 c1a1 h7h6 h4f3 f6f5 f3h4');
INSERT INTO evaluations VALUES(38873,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38873,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38873,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38873,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38873,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(38873,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38873,36,'cp 0','b7b5 a2a4 g8e7 a4b5 e7d5 a1c1 d5f4 e3f4 c8b7 g1f3 d8d6 f1c4 d6f4 c4e2 e8g8 e1g1 b8d7 g2g3 f4d6 d1c2 h7h6 c3e4 d6d5 e4c3');
INSERT INTO evaluations VALUES(46986,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(46986,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46986,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(46986,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(46986,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(46986,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46986,36,'cp 35','c7c6 e2e4 e7e5 f1g2 c8e6 d2d3 f8e7 g1f3 a7a6 e1g1 e8g8 h2h3 c6c5 f3h4 b7b5 b2b3 b5b4 c3e2 a6a5 a2a4 f6e8 h4f5 e7f6 a1b1 b8c6 f5e3 e6d7 e3d5 e8c7 f2f4 c7d5 e4d5 c6d4 e2d4 e5d4 c1d2 g7g6 b1a1 g8h8 g3g4 f6g7 g1h2');
INSERT INTO evaluations VALUES(47593,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(47593,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(47593,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(47593,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(47593,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(47593,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(47593,36,'cp 92','c1f4 a7a6 e2e3 f8g7 f1d3 d5c4 d3c4 b7b5 c4e2 c8b7 e1g1 b8d7 a1c1 a8c8 b2b4 e8g8 a2a3 d7b6 f4g3 d8e7 g3h4 h7h6 e2d3 g6g5 h4g3 f6d7 d3e4 b7e4 c3e4 b6c4 e4d2 c4d2 f3d2 e6e5 d2b3 e5d4 e3d4 d7b6');
INSERT INTO evaluations VALUES(48198,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(48198,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(48198,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(48198,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(48198,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48198,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48198,36,'cp -29','e6d5 g1f3 f8e7 e1g1 e8g8 b2b3 a7a5 c1b2 b8a6 a2a3 c8f5 f3h4 f5g4 d2d3 f8e8 h2h3 g4h5 b1d2 f6d7 g2f3 h5f3 d2f3 c7c6 d1c2 e7f8 a1e1 d8b6 h4f5 a6c5 f5d4 c5e6 d4e6 e8e6 e2e4 d7c5 e4d5 c6d5');
INSERT INTO evaluations VALUES(50600,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(50600,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(50600,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(50600,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(50600,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(50600,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50600,36,'cp 49','c5d4 c3d4 d7d5 e4e5 c8f5 b1c3 b8c6 f1d3 f5d3 d1d3 e7e6 g1f3 d8b6 h2h4 g8e7 e1g1 a8c8 a1b1 h7h6 h4h5 e7f5 c3d5 b6d8 d5c3 c6d4 f1d1 d4f3 d3f3 d8b6 c1f4 f8e7 g2g3 c8d8 d1d8 e7d8 c3e4 e8g8 e4d6 f5d6 e5d6 d8g5 b2b3 g5f4 f3f4 f8d8 b1d1');
INSERT INTO evaluations VALUES(50625,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(50625,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(50625,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(50625,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(50625,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(50625,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(50625,36,'cp 42','c3d4 d7d5 e4e5 c8f5 f1d3 f5d3 d1d3 e7e6 b1c3 b8c6 g1f3 g8e7 h2h4 d8b6 e1g1 a8c8 a1b1 h7h6 h4h5 e7f5 c3d5 b6d8 d5c3 c6d4 f1d1 f8c5 g2g4 f5h4 f3d4 c5d4 c1f4 d4c3 b2c3 d8d3 d1d3 c8c4 b1b4 c4b4 c3b4 g7g5 f4g3 e8g8 g3h4 g5h4 d3d7 b7b5');
INSERT INTO evaluations VALUES(50621,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(50621,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(50621,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(50621,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(50621,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(50621,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50621,36,'cp 30','d7d5 e4e5 c8f5 b1c3 b8c6 f1d3 f5d3 d1d3 e7e6 g1f3 a8c8 h2h4 f8b4 e1g1 b4c3 b2c3 c6a5 h4h5 h7h6 a1b1 c8c6 f3d2 g8e7 c1a3 e8g8 g2g4 f7f5 e5f6 f8f6 d3g3 e7c8 f2f4 c8d6 a3d6 d8d6 f1f3 b7b5 g4g5 h6g5 f4g5 d6g3 f3g3 f6f4 b1f1 f4h4 h5h6 a5c4 d2f3 h4e4');
INSERT INTO evaluations VALUES(52071,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(52071,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(52071,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(52071,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(52071,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(52071,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(52071,36,'cp 24','e3d4 d8b6 b2b3 f8d6 f4e3 b6a5 e3d2 a5d8 f1d3 g8f6 g1f3 d6e7 e1g1 b7b6 d2f4 e8g8 f1e1 f6h5 f4e3 c8b7 a2a3 h5f6 e3f4 b8c6 b3b4 d7d5 b1d2 a8c8 c2c3 e7d6 f3e5 c6e7 a1c1 e7g6 e5g6 h7g6 d1f3 d8c7 f4e5 a7a5 e5f6');
INSERT INTO evaluations VALUES(52061,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(52061,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52061,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52061,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(52061,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52061,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(52061,36,'cp 8','d8b6 b2b3 f8d6 f4d6 b6d6 g1f3 b8c6 f1e2 g8f6 e1g1 e8g8 c2c4 b7b6 b1c3 c8b7 c3b5 d6b8 d4d5 e6d5 c4d5 c6e7 d5d6 e7d5 a1c1 f8c8 d1d4 a7a6 b5c7 d5c7 d6c7 c8c7 d4b6 f6d5 b6d4 c7c1 f1c1 d5f4 c1e1 f4e2 e1e2 b7f3');
INSERT INTO evaluations VALUES(57185,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(57185,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(57185,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(57185,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(57185,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(57185,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(57185,36,'cp -104','d7d5 c2c3 b8d7 d2d4 c7c6 d4e5 d7e5 c1f4 e5g6 f4g5 d8b6 d1b3 f8c5 b3b6 a7b6 h4h5 g6f8 b1d2 f8e6 g1f3 e6g5 f3g5 g8f6 f1h3 f6g4 e2e3 b6b5 g5f3 e8g8 a2a3 f8e8 f3d4 h7h6 e1f1 a8a6 d2f3 c5b6 f1g2 c8d7 h1d1 e8a8 d1d2');
INSERT INTO evaluations VALUES(56420,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(56420,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(56420,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(56420,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(56420,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(56420,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(56420,36,'cp -78','c2c3 b8d7 d2d4 c7c6 d4e5 d7e5 c1f4 e5g6 f4g5 d8b6 d1b3 f8c5 b3b6 a7b6 h4h5 g6f8 e2e3 f8e6 g5f4 g8f6 f1e2 e6f4 g3f4 e8e7 g1f3 c8f5 b1d2 g7g6 f3e5 c5d6 a2a4 f5c2 d2f3 c2e4 h5g6 h7g6 h1h8 a8h8 e1d2 d6e5 f3e5 f6d7 e5d7 e7d7');
INSERT INTO evaluations VALUES(1998,30,'cp -500',NULL);
INSERT INTO evaluations VALUES(1998,31,'cp -519',NULL);
INSERT INTO evaluations VALUES(1998,32,'cp -518',NULL);
INSERT INTO evaluations VALUES(1998,33,'cp -510',NULL);
INSERT INTO evaluations VALUES(1998,34,'cp -515',NULL);
INSERT INTO evaluations VALUES(1998,35,'cp -517',NULL);
INSERT INTO evaluations VALUES(1998,36,'cp -543','g5g2 f3d4 g2h1 c4f1 h1e4 d1e2 e4e2 f1e2 d7d5 d4b5 e8d8 b5d4 c7c6 d2d3 f8d6 h2h4 g8f6 b1c3 h8e8 c1e3 d8e7 d4b3 f6g4 e2g4 c8g4 b3a5 g4c8 a5b3 f7f6 e1d2 e7f8 h4h5 f8f7 a1g1 b7b6 b3d4 c8d7 h5h6 g7g5 d4f3 g5g4');
INSERT INTO evaluations VALUES(2006,30,'cp -520',NULL);
INSERT INTO evaluations VALUES(2006,31,'cp -517',NULL);
INSERT INTO evaluations VALUES(2006,32,'cp -506',NULL);
INSERT INTO evaluations VALUES(2006,33,'cp -517',NULL);
INSERT INTO evaluations VALUES(2006,34,'cp -525',NULL);
INSERT INTO evaluations VALUES(2006,35,'cp -512',NULL);
INSERT INTO evaluations VALUES(2006,36,'cp -524','f3d4 g2h1 c4f1 h1e4 d1e2 e4e2 f1e2 d7d5 d4b5 e8d8 b5d4 c7c6 d2d3 f8d6 c1e3 g8f6 h2h4 h8e8 e1d2 d8e7 b1a3 e7f8 c2c3 c8d7 h4h5 h7h6 a1g1 b7b6 e2f3 c6c5 d4e2 e8e5 a3c4 d5c4 f3a8 c4d3 e2g3 f6h5 g3h5 e5h5 d2d3 d6e5 d3c4 h5h4 c4b3');
INSERT INTO evaluations VALUES(2004,30,'cp -512',NULL);
INSERT INTO evaluations VALUES(2004,31,'cp -530',NULL);
INSERT INTO evaluations VALUES(2004,32,'cp -521',NULL);
INSERT INTO evaluations VALUES(2004,33,'cp -503',NULL);
INSERT INTO evaluations VALUES(2004,34,'cp -521',NULL);
INSERT INTO evaluations VALUES(2004,35,'cp -531',NULL);
INSERT INTO evaluations VALUES(2004,36,'cp -522','g2h1 c4f1 h1e4 d1e2 e4e2 f1e2 d7d5 d4b5 e8d8 d2d3 c7c6 b5d4 f8d6 h2h4 g8f6 c1e3 h8e8 e1d2 d8e7 b1a3 e7f8 c2c3 c8d7 e2f3 d7g4 a3c2 g4f3 d4f3 c6c5 a2a4 a7a6 a4a5 a8c8 a1g1 h7h6 d3d4 f6e4 d2e1 c8c6 d4c5 d6c5 f3d4 c6g6 g1g6');
INSERT INTO evaluations VALUES(2003,30,'cp -521',NULL);
INSERT INTO evaluations VALUES(2003,31,'cp -530',NULL);
INSERT INTO evaluations VALUES(2003,32,'cp -535',NULL);
INSERT INTO evaluations VALUES(2003,33,'cp -536',NULL);
INSERT INTO evaluations VALUES(2003,34,'cp -521',NULL);
INSERT INTO evaluations VALUES(2003,35,'cp -523',NULL);
INSERT INTO evaluations VALUES(2003,36,'cp -528','c4f1 h1e4 d1e2 e4e2 f1e2 d7d5 d4b5 e8d8 d2d3 c7c6 b5d4 f8d6 h2h4 g8f6 a2a4 h8e8 b1c3 d8e7 a4a5 c8d7 d4b3 f6g4 a1a4 e7f8 c1f4 d6e7 h4h5 f8g8 e1f1 g4f6 h5h6 c6c5 a5a6 b7b6 a4a1 d5d4 e2f3 d4c3 f3a8 e8a8 b2c3 g7h6');
INSERT INTO evaluations VALUES(8428,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(8428,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(8428,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(8428,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(8428,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(8428,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(8428,36,'cp 62','g5f4 e8g8 a1c1 b7b6 a2a3 d5c4 f1c4 c6a5 c4b5 c7c6 b5d3 c6c5 e1g1 c8b7 d4c5 b6c5 d1e2 a8c8 f1d1 d8b6 c3a4 b6b3 d3c2 b7f3 e2f3 b3b5 e3e4 b5b7 f4d2 a5c4 d2c3 c4b6 a4b6 a7b6 a3a4 b6b5 e4e5 f6d5 f3d3 g7g6 d3b5 b7c7 b5a5 c5c4');
INSERT INTO evaluations VALUES(10874,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(10874,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(10874,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10874,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10874,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(10874,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10874,36,'cp -12','h7h6 g5f6 e7f6 a2a4 d7d6 b1d2 e8g8 c2c3 c6e7 e1g1 c7c6 f1e1 d6d5 c4b3 e7g6 d3d4 e5d4 f3d4 f6d4 c3d4 c8e6 d2f3 d5e4 e1e4 e6d5 e4e3 a7a5 b3d5 d8d5 d1b3 f8d8 g2g3 d5b3 e3b3 a8b8 h2h4 d8d5');
INSERT INTO evaluations VALUES(10824,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(10824,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(10824,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(10824,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10824,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(10824,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(10824,36,'cp 0','g5f6 e7f6 a2a4 d7d6 b1d2 e8g8 e1g1 c6e7 f1e1 e7g6 d2f1 c7c6 f1e3 g6f4 c2c3 d6d5 c4a2 g7g6 h2h4 f6h4 e4d5 h4f6 d5c6 b7c6 d3d4 e5e4 f3d2 f4d3 e1e2 d3f4');
INSERT INTO evaluations VALUES(11944,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(11944,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(11944,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(11944,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(11944,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(11944,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11944,36,'cp 48','d7d6 c4d5 c6d5 d1d5 c8e6 d5c6 e6d7 c6e4 d6e5 e4e5 d8e7 e5e3 e8c8 e1g1 d7c6 e3h3 e7d7 h3d7 d8d7 c1e3 a7a5 b1d2 a5a4 d2f3 c6f3 g2f3 f8e7 a1d1 h8d8 d1d7 d8d7 b2b3 d7d6 f1e1 d6a6 e1e2 g7g6 f3f4 e7d6 e2d2 d6b4 d2d4 a4b3 a2b3 b4c5 d4c4 c5e3');
INSERT INTO evaluations VALUES(12915,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(12915,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(12915,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12915,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(12915,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12915,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12915,36,'cp 8','c5d4 c3d4 d7d6 f1c4 d5b6 c4b5 d6e5 f3e5 c8d7 e5d7 d8d7 b1c3 e7e6 e1g1 a7a6 b5c6 d7c6 f1e1 f8e7 d1g4 e7f6 c3e4 b6d5 c1f4 e8g8 f4e5 f6e5 d4e5 f7f5 e5f6 d5f6 g4e2 c6e4 e2e4 f6e4 e1e4 a8d8 a1e1 d8d2');
INSERT INTO evaluations VALUES(16352,30,'cp -305',NULL);
INSERT INTO evaluations VALUES(16352,31,'cp -301',NULL);
INSERT INTO evaluations VALUES(16352,32,'cp -310',NULL);
INSERT INTO evaluations VALUES(16352,33,'cp -317',NULL);
INSERT INTO evaluations VALUES(16352,34,'cp -323',NULL);
INSERT INTO evaluations VALUES(16352,35,'cp -321',NULL);
INSERT INTO evaluations VALUES(16352,36,'cp -318','f6g8 f1b5 d7d5 e1g1 g8e7 b5a4 c8d7 f1e1 d7e6 c3e2 e7g6 e1f1 d8d7 c2c3 f8e7 f2f4 e8g8 a4c2 f7f5 h2h3 c6a5 g1h2 b7b6 c2a4 c7c6 d1d3 d7c7 a4c2 c6c5 a1b1 a5c6 d3g3 c7d7 b2b3 b6b5 c1a3 b5b4 c3b4 c6b4 a3b4');
INSERT INTO evaluations VALUES(22672,30,'cp -322',NULL);
INSERT INTO evaluations VALUES(22672,31,'cp -305',NULL);
INSERT INTO evaluations VALUES(22672,32,'cp -313',NULL);
INSERT INTO evaluations VALUES(22672,33,'cp -318',NULL);
INSERT INTO evaluations VALUES(22672,34,'cp -323',NULL);
INSERT INTO evaluations VALUES(22672,35,'cp -310',NULL);
INSERT INTO evaluations VALUES(22672,36,'cp -313','d4d5 c6e5 d1e2 d8e7 c1e3 a7a6 e1c1 d7d6 d1e1 f7f5 e3d4 e8d8 f2f4 e5d7 e2d1 e7f7 f1e2 g8f6 e2f3 f8e7 e1e3 h8e8 h1e1 d7c5 c1b1 c8d7 d1e2 f6e4 f3e4 c5e4 c3e4 f5e4 e3e4 e7h4 e4e8 f7e8 e2e8 d7e8 g2g3 h4e7 d4g7 e8g6 f4f5');
INSERT INTO evaluations VALUES(19560,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(19560,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(19560,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(19560,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(19560,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(19560,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(19560,36,'cp 53','g7g6 h2h3 f8g7 e1g1 d7d6 c1e3 g8f6 d1d2 d8a5 b2b4 a5d8 a2a4 e8g8 f1d1 c8e6 d3c2 d6d5 f3g5 d8c7 g5e6 f7e6 b1a3 f6e4 c2e4 d5e4 d2c2 c7e5 a3c4 e5c3 c2e4 c6b4 e4e6 g8h8 a1c1 c3f6 e6f6 g7f6 d1d7 b4a2 c1c2 a8d8 d7d8 f8d8 c2a2');
INSERT INTO evaluations VALUES(19328,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(19328,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(19328,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(19328,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(19328,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(19328,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(19328,36,'cp 62','e4e5 f8g7 c1f4 f7f6 e5f6 g8f6 e1g1 e8g8 b1d2 d7d5 f1e1 f6e4 f4e3 e4d2 d1d2 e7e5 d3c2 e5e4 f3d4 c6e5 c2b3 g8h8 d4e2 e5d3 e2f4 d3e1 a1e1 d8d6 b3d5 c8f5 h2h3 a8d8 e1d1 g7e5 c3c4 f5d7 f4e2 d7c6 e3a7 c6d5');
INSERT INTO evaluations VALUES(23294,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(23294,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(23294,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(23294,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(23294,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(23294,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(23294,36,'cp 19','g8f6 d2d3 f8b4 c1d2 d7d6 a2a3 b4c3 d2c3 e8g8 b2b4 d8e8 d1d2 h7h6 f2f4 c6d4 a1b1 c8d7 c3d4 e5d4 g1f3 d7c6 b4b5 c6f3 g2f3 a8b8 e1g1 f6g4 f3g4 f5g4 f1e1 a7a6 a3a4 b8a8 a4a5 e8g6 d2b2 g6f6 e2e3 d4e3 b2f6 g7f6');
INSERT INTO evaluations VALUES(24494,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(24494,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(24494,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(24494,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(24494,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(24494,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(24494,36,'cp 67','d1b3 d8c8 a1c1 f8e7 c4d5 f6d5 f4g3 d5c3 b2c3 c6c5 f1e2 c5d4 e3d4 e8g8 e1g1 f8d8 a2a4 b7b6 f1d1 h7h6 a4a5 b6a5 c1a1 d7f6 a1a5 f6e4 c3c4 e4g3 h2g3 c8c7 b3a2 f5e4 f3e5 e7f6 e2f3 e4f3 e5f3 a8c8 c4c5');
INSERT INTO evaluations VALUES(27009,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(27009,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(27009,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(27009,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(27009,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(27009,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(27009,36,'cp 46','e2b5 g8f6 d2d4 f6d7 d4d5 c6d4 b5e2 d4e2 d1e2 f8e7 h2h3 g4f3 e2f3 e7g5 c1g5 d8g5 h3h4 g5f6 f3e2 f6h6 f2f3 h6f4 e1f2 h7h5 g2g3 f4f6 a2a4 f6g6 e2e3 e8g8 e3g5 g6h7 g5f5 h7f5 e4f5 g7g6 f5g6 f7g6 a4a5 d7f6 a1a4 g8g7 h1d1');
INSERT INTO evaluations VALUES(32730,30,'cp 382',NULL);
INSERT INTO evaluations VALUES(32730,31,'cp 393',NULL);
INSERT INTO evaluations VALUES(32730,32,'cp 405',NULL);
INSERT INTO evaluations VALUES(32730,33,'cp 408',NULL);
INSERT INTO evaluations VALUES(32730,34,'cp 401',NULL);
INSERT INTO evaluations VALUES(32730,35,'cp 405',NULL);
INSERT INTO evaluations VALUES(32730,36,'cp 395','b1c3 c7c6 c3d5 c6d5 c4b5 b8c6 c1g5 f6g6 b3d5 f8b4 c2c3 g8e7 g5e7 e8e7 b5c6 a8d8 d5e5 e7f8 a1d1 b4d6 e5b5 g6e6 e1g1 g7g6 c6d5 e6e7 f2f4 h7h5 g1h1 f8g7 b5e2 h8e8 a2a3 a7a5 b2b4 d6c7 e2f2 b6b5 f2e2 e7d7 e4e5 a5b4 a3b4 c7b8 h2h3 d8c8 e2d2 c8d8 d2f2 e8e7 f2e2 d7c7');
INSERT INTO evaluations VALUES(32732,30,'cp 380',NULL);
INSERT INTO evaluations VALUES(32732,31,'cp 408',NULL);
INSERT INTO evaluations VALUES(32732,32,'cp 409',NULL);
INSERT INTO evaluations VALUES(32732,33,'cp 400',NULL);
INSERT INTO evaluations VALUES(32732,34,'cp 396',NULL);
INSERT INTO evaluations VALUES(32732,35,'cp 379',NULL);
INSERT INTO evaluations VALUES(32732,36,'cp 423','c7c6 c3d5 c6d5 c4b5 b8c6 c1g5 f6g6 b3d5 f8b4 c2c3 g8e7 g5e7 e8e7 b5c6 a8d8 d5e5 e7f8 a1d1 d8d1 e1d1 g6d6 e5d5 d6d5 e4d5 b4d6 h2h4 h7h5 h1e1 g7g6 d1c2 h8g8 g2g3 g6g5 h4g5 g8g5 e1e8 f8g7 e8d8 d6e7 d8d7 e7c5 d7a7 g5g6 c6e8 g6f6 f2f4 c5f2 d5d6 f2g3 d6d7 f6d6 a7a8 h5h4 d7d8q d6d8 a8d8 g3f4 d8d7 g7f6');
INSERT INTO evaluations VALUES(32734,30,'cp 422',NULL);
INSERT INTO evaluations VALUES(32734,31,'cp 400',NULL);
INSERT INTO evaluations VALUES(32734,32,'cp 415',NULL);
INSERT INTO evaluations VALUES(32734,33,'cp 427',NULL);
INSERT INTO evaluations VALUES(32734,34,'cp 420',NULL);
INSERT INTO evaluations VALUES(32734,35,'cp 428',NULL);
INSERT INTO evaluations VALUES(32734,36,'cp 421','c3d5 c6d5 c4b5 b8c6 c1g5 f6g6 b3d5 f8b4 c2c3 g8e7 g5e7 e8e7 b5c6 a8d8 d5e5 e7f8 a1d1 d8d1 e1d1 g6d6 e5d5 d6d5 e4d5 b4d6 h2h4 h7h5 h1e1 g7g6 d1c2 h8g8 b2b4 g6g5 h4g5 g8g5 e1e8 f8g7 e8d8 g5g6 g2g3 g6f6 c2d3 a7a5 a2a3 d6e5 f2f4 e5c7 d8d7 h5h4 g3h4 c7f4 b4a5 b6a5 c3c4 f6h6 c4c5 h6h4 d7a7 g7f6 d5d6 h4h3 d3c4 h3a3 d6d7');
INSERT INTO evaluations VALUES(40375,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40375,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40375,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40375,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40375,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40375,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40375,36,'cp 0','h7h6 g5h4 e8g8 e1c1 b8c6 d4b3 c8e6 c1b1 a8c8 f2f3 a7a6 c3d5 e6d5 e4d5 c6e5 h4f6 g7f6 d2h6 b7b5 h2h4 e5c4 b3d4 f6g7 h6g5 g7f6');
INSERT INTO evaluations VALUES(40979,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40979,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(40979,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40979,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40979,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40979,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(40979,36,'cp 25','b1d2 e8g8 g1f3 c7c5 f1d3 f6h5 d4c5 h5f4 e3f4 b8d7 d1c2 d8c7 g2g3 h7h6 b2b4 e6e5 f4e5 d7e5 e1g1 e5d3 c2d3 b7b6 d3d5 c8b7 d5e5 c7d7 f3d4 b6c5 b4c5 a8c8 d2e4 b7e4 e5e4 c8c5 f1d1 f8d8 d4b3 d7d1 a1d1 d8d1 g1g2 c5c7 b3d4 c7c3');
INSERT INTO evaluations VALUES(40978,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40978,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(40978,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40978,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40978,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40978,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40978,36,'cp 0','e8g8 g1f3 c7c5 f1d3 f6h5 d4c5 h5f4 e3f4 e7c5 e1g1 b8c6 d2b3 c5b6 f1e1 d8d6 d1d2 c8d7 b3d4 f7f6 a2a4 a7a5 d4b5 d6c5 d2c2 h7h6 f3h4 c6e7 h4f3');
INSERT INTO evaluations VALUES(36341,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(36341,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36341,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(36341,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36341,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36341,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36341,36,'cp 19','f1d3 c7c5 h2h3 b8c6 g1f3 e7d6 d4c5 d6c5 e1g1 c5d6 f4d6 d8d6 c3c4 f8d8 a1c1 d5c4 d3c4 h7h6 a2a3 c8d7 b2b4 a8c8 d1b3 c6e5 f3e5 d6e5 d2f3 e5c7 b3b2 c7b6 c4e2 a7a6 f1d1 d7b5 c1c8 d8c8');
INSERT INTO evaluations VALUES(36344,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36344,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36344,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36344,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36344,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36344,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36344,36,'cp 16','c7c5 g1f3 f6h5 d4c5 h5f4 e3f4 e7c5 e1g1 b8c6 g2g3 g7g6 a2a4 f8e8 d2b3 c5d6 f1e1 a7a6 h2h4 e6e5 f3e5 d6e5 f4e5 c6e5 b3d4 c8g4 d3e2 d8d7 e2g4 e5g4 e1e8 a8e8 g1g2 e8e4 d1d2 h7h5 d4f3 g8g7 b2b3 g4f6 a4a5 d7b5');
INSERT INTO evaluations VALUES(41018,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(41018,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(41018,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(41018,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(41018,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41018,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41018,36,'cp 17','c7c5 f1c4 c5d4 f3d4 e8g8 e1g1 c8d7 d4f3 h7h6 g5f4 d8b6 d1e2 d7c6 f3e5 f6d7 e5c6 b8c6 f1d1 f8d8 a1c1 d7e5 c4b3 d8d1 c1d1 a8d8 h2h3 e5g6 f4g3 d8d1 b3d1 e7f6 e2c2 c6e7 c2b3 b6b3 d1b3 f6c3 b2c3');
INSERT INTO evaluations VALUES(47765,30,'cp -189',NULL);
INSERT INTO evaluations VALUES(47765,31,'cp -196',NULL);
INSERT INTO evaluations VALUES(47765,32,'cp -189',NULL);
INSERT INTO evaluations VALUES(47765,33,'cp -180',NULL);
INSERT INTO evaluations VALUES(47765,34,'cp -177',NULL);
INSERT INTO evaluations VALUES(47765,35,'cp -199',NULL);
INSERT INTO evaluations VALUES(47765,36,'cp -193','f6e4 c2c4 a7a5 c1e3 b8d7 f1d3 e4c5 d3c2 f8e7 e1g1 e8g8 b1c3 f7f5 a2a3 e5e4 f3d2 d7e5 d1e2 e7g5 f2f4 e4f3 d2f3 g5e3 e2e3 e5c4 e3f4 c4b2 a1b1 d8f6 c3b5 c8d7 b5c7 b2d3 c2d3 c5d3 f4e3 a8c8 c7e6 d7e6 e3e6 f6e6 d5e6 d3c5 f3d4 g7g6');
INSERT INTO evaluations VALUES(47983,30,'cp -204',NULL);
INSERT INTO evaluations VALUES(47983,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(47983,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(47983,33,'cp -194',NULL);
INSERT INTO evaluations VALUES(47983,34,'cp -187',NULL);
INSERT INTO evaluations VALUES(47983,35,'cp -182',NULL);
INSERT INTO evaluations VALUES(47983,36,'cp -200','c2c4 a7a5 c1e3 b8d7 f1d3 e4c5 d3c2 g7g6 b1c3 f8g7 h2h4 f7f5 d1e2 e5e4 f3d2 e8g8 e1c1 a5a4 a2a3 d7f6 f2f3 f6h5 e2f2 b7b6 e3d4 d8e7 d4g7 e7g7 f3e4 f5e4 f2e3 c8f5 d2e4 a8e8 h1f1 c5e4 c3e4 f5e4 c2e4 h5f6');
INSERT INTO evaluations VALUES(49553,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49553,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(49553,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(49553,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(49553,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(49553,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(49553,36,'cp -26','c1b2 b8c6 a2a3 d7d5 c4d5 f6d5 e2e3 f8d6 d2d3 d8e7 b1d2 g7g5 b3b4 a7a6 a1c1 h7h5 d2e4 c8d7 d3d4 e8c8 e4d6 e7d6 g1f3 e5d4 b2d4 c6d4 d1d4 d7a4 d4c5 g5g4 c5d6 d8d6 f3d4 h8e8 f1e2 c8b8 h2h4 g4g3 e2h5 g3f2 e1f2 d5e3');
INSERT INTO evaluations VALUES(49554,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(49554,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(49554,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(49554,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(49554,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(49554,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49554,36,'cp -9','b8c6 g1f3 e5e4 f3d4 f8c5 d4f5 e8g8 e2e3 d7d5 c4d5 c6b4 f5g3 f6d5 d1c1 b7b6 g3e4 f8e8 b1c3 d5c3 c1c3 c5d4 c3b4 d4b2 a1b1 a7a5 b4b5 e8e4 b1b2 c8b7 b5d3 d8e7 f1e2 e4e5 f2f3 e5c5 e1f2 a8d8 d3b1 e7h4 g2g3 h4f6 b2c2 b7e4');
INSERT INTO evaluations VALUES(50242,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(50242,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(50242,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50242,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50242,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(50242,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(50242,36,'cp 6','b1c3 d7d6 d3d4 f8g7 c1e3 e8g8 d1d2 c7c6 g1f3 b8d7 a2a4 f6g4 e3g5 h7h6 g5h4 g4f6 a4a5 e7e5 d4e5 d6e5 f1c4 g6g5 h4g3 d8e7 e1g1 f6h5 f1d1 f8e8 d2e3 h5f4 h2h3 d7f8 c3a4 c8e6 c4f1 f8d7 e3c3 g7f8 d1e1 f8g7 f1c4 g8h8 c4e6');
INSERT INTO evaluations VALUES(50240,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(50240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50240,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50240,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50240,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(50240,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50240,36,'cp 2','d7d6 d3d4 f8g7 g1f3 e8g8 c1e3 b8d7 a2a4 f6g4 e3c1 e7e5 c1g5 g4f6 d4e5 d6e5 f1c4 h7h6 g5c1 f8e8 e1g1 c7c6 h2h3 d8c7 c1e3 d7f8 d1e2 c8e6 f1d1 b7b6 b2b3 a8d8 c4e6 f8e6 d1d8 e8d8 a1d1 d8d1 e2d1 f6h5 f3d2 h5f4');
INSERT INTO evaluations VALUES(46724,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46724,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46724,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46724,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(46724,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(46724,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(46724,36,'cp 4','d3d4 f8g7 g1f3 e8g8 c1e3 c7c5 d4c5 d8a5 f1d3 d6c5 e1g1 c8g4 h2h3 g4f3 d1f3 f6d7 f3g3 b8c6 d3b5 g7c3 b5c6 b7c6 b2c3 a5c3 g3c7 d7f6 e3h6 f8e8 f2f3 c3c2 c7a5 c2b2 a5c5 b2b6 c5e3 f6d7 a1b1 b6e3 h6e3 e8c8 f1c1 c8c7 b1b3');
INSERT INTO evaluations VALUES(50735,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(50735,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(50735,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(50735,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50735,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50735,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50735,36,'cp 2','a2a4 g8f6 e1g1 f8e7 c2c3 b7b6 c1f4 e8g8 d1e2 c8b7 b1d2 b8d7 h2h3 f8e8 f1d1 h7h6 f4h2 e7f8 d1e1 b7c6 c4b3 b6b5 d3d4 b5a4 b3a4 c6a4 a1a4 c5d4 c3d4 d8b6 d2c4 b6c6 e1a1 c6e4 e2e4 f6e4 a4a6 a8a6');
INSERT INTO evaluations VALUES(57262,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(57262,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57262,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(57262,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57262,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57262,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57262,36,'cp 0','e2e3 g8f6 a2a3 d7d5 c2c4 c7c6 c4d5 c6d5 f1b5 b8c6 g1e2 f8d6 d2d3 e8g8 d3e4 d5e4 b5c6 b7c6 b1d2 a7a5 d2c4 d6c7 d1d8 f8d8 c4a5 c7a5 b4a5 c8a6 b2d4 f6d5 e1d2 c6c5 d4c5 d5b6 e2d4 b6c4 d2e1 a8c8 c5b4 c4b2 e1d2 b2c4');
INSERT INTO evaluations VALUES(57261,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57261,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57261,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57261,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(57261,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(57261,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57261,36,'cp 0','g8f6 a2a3 d7d5 c2c4 c7c6 c4d5 c6d5 g1e2 f8d6 d2d3 a7a5 d3e4 d5e4 b4b5 e8g8 d1d4 d8e7 b1c3 b8d7 c3d5 e7d8 d5f6 d8f6 d4f6 d7f6 e2d4 f6d5 f1c4 d5b6 c4b3 a5a4 b3a2 c8d7 e1g1 d6e5 b2c3 e5d6');
INSERT INTO evaluations VALUES(1552,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(1552,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(1552,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(1552,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(1552,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(1552,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(1552,36,'cp 101','d4c5 a5c5 f1d3 c5b6 e1g1 f7f6 c3c4 f6e5 b1c3 g8f6 c1g5 d5d4 c3e4 f8e7 a2a3 a7a5 g5h4 e8g8 h4g3 g7g6 e4f6 e7f6 f1e1 c8d7 h2h4 b6c5 d1c2 c5e7 c4c5 a5a4 h4h5 g6h5 f3e5 c6e5 g3e5 d7c6 e5d6 e7g7 d3e4 f6e7 e4c6 e7d6 c5d6 b7c6 c2c6');
INSERT INTO evaluations VALUES(2669,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(2669,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2669,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2669,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(2669,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2669,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2669,36,'cp 0','d2d3 d7d6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1e1 d8e8 c3b1 e8g6 b1d2 f8f7 e1h4 a8f8 f3g5 f7e7 g5f3 e7f7');
INSERT INTO evaluations VALUES(2671,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2671,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2671,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2671,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2671,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2671,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2671,36,'cp 0','d7d6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1e1 d8e8 e1g3 e8h5 h2h3 a8e8 f1f2 e8e7 a1f1 e7f7 f3g5 f7e7');
INSERT INTO evaluations VALUES(2616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2616,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2616,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2616,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2616,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2616,36,'cp 0','c1e3 b7b5 c4a2 c5e3 f2e3 c6e7 d3d4 c8b7 d4e5 f6e4 c3e4 b7e4 c2c4 e7f5 d1e2 d6e5 a1d1 d8f6 c4b5 a6b5 f3d2 e4b7 g2g4 g7g6 g4f5 f6g5 g1f2 g5h4 f2g1');
INSERT INTO evaluations VALUES(7491,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(7491,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(7491,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(7491,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(7491,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(7491,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(7491,36,'cp 27','a2a4 b5b4 c3d5 e8g8 d5f6 e7f6 b3d5 c8b7 d2d3 d8c8 h2h4 c6a5 d5b7 c8b7 c1g5 b7c6 d1d2 f6g5 h4g5 c6c5 e1g1 f7f6 g5f6 f8f6 d3d4 e5d4 d2d4 c5d4 f3d4 a8e8 f2f3 a5c4 b2b3 c4e3 f1e1 f6d6 e1e3 d6d4 e3d3 d4d3 c2d3 c7c5 a1c1 d7d6 g2g3');
INSERT INTO evaluations VALUES(13578,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(13578,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(13578,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(13578,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(13578,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13578,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13578,36,'cp 0','f1d3 f8d6 f3e5 d8c7 a2a4 c5c4 d3c2 b7b6 e1g1 c8b7 d1e2 c6a5 b1d2 e8g8 e2f3 a8e8 a1b1 e8c8 f1e1 c8a8 b1a1 a8e8 a1b1 e8a8');
INSERT INTO evaluations VALUES(32437,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(32437,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(32437,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(32437,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(32437,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(32437,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(32437,36,'cp 195','d2d4 d8d7 f1c4 g8f6 d4e5 d6e5 f3e5 d7d1 e1d1 b8c6 e5c6 b7c6 h1e1 e8d7 e1e5 f5g6 e5e2 f8d6 c4e6 d7d8 d1e1 h8e8 e6b3 d8d7 h2h3 e8e2 e1e2 f6e4 c3e4 g6e4 f2f3 e4g6 c1d2 a7a5 a2a4 d6e5 a1d1 d7c8 d2c1 c8b7 e2f2 c6c5 g2g4 e5f6 c2c3');
INSERT INTO evaluations VALUES(32435,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(32435,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(32435,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(32435,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(32435,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(32435,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(32435,36,'cp 187','d8d7 f1c4 b8c6 e1g1 g8f6 c4b5 a7a6 b5c6 b7c6 d4e5 d6e5 f3e5 d7d1 f1d1 f8d6 e5c6 f5c2 d1e1 e8f7 c6e5 d6e5 e1e5 h8e8 e5e8 a8e8 c1e3 e8e7 a1c1 c2d3 b2b3 c7c6 c1d1 d3c2 d1d6 e7e6 d6d2 c2f5 d2d4 f6e4 c3a4 e6d6 d4d6 e4d6 a4c5 f5c8 g1f1 d6f5 b3b4 f7g8 a2a4');
INSERT INTO evaluations VALUES(39402,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(39402,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(39402,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(39402,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(39402,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(39402,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(39402,36,'cp 173','c2c3 b4e7 c1f4 g8f6 f1d3 e8g8 d1c2 d7d5 e4e5 f6h5 f4e3 f7f5 b1d2 g8h8 e1c1 b8c6 d1f1 c8e6 g2g3 c6a5 f3e1 e7g5 e3g5 d8g5 c1b1 a5c4 d2c4 d5c4 d3e2 c7c6 h1g1 g7g6 b1a1 g5h6 e1g2 g6g5 e2h5 h6h5 g1h1 h5g4 h2h4');
INSERT INTO evaluations VALUES(39401,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(39401,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(39401,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(39401,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(39401,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(39401,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(39401,36,'cp 168','b4e7 c1f4 g8f6 f1d3 e8g8 b1d2 d7d6 d1e2 f6h5 g2g3 b8d7 e1c1 c7c5 d3c2 f8e8 h1e1 c5d4 f3d4 d7e5 d4f5 g7g6 d2f3 d8c7 f3d4 e7f8 c2b3 c7d8 f4e5 d6e5 d4b5 d8g5 f5e3 c8h3 e2f3');
INSERT INTO evaluations VALUES(43873,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(43873,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(43873,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(43873,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(43873,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43873,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43873,36,'cp 0','c8b7 c2c4 e7e6 a2a3 c7c5 b1c3 c5d4 d1d4 b8c6 d4d3 f6h5 f4g5 f8e7 g5e7 d8e7 c3b5 e8g8 b5d6 c6a5 d6b7 a5b7 g2g3 b7c5 d3c2 a8c8 f1g2 f7f5 e1g1 h5f6 f3d4 c5e4 c2d3 e4c5');
INSERT INTO evaluations VALUES(48946,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(48946,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48946,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48946,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48946,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48946,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48946,36,'cp 0','f8d6 g1f3 e8g8 e1g1 b8c6 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 c4d5 g4f3 d5f3 c6d4 f3b7 a8b8 c3d5 b8b7 h4f6 g7f6 d1d4 c7c6 d4f6 d8f6 d5f6 b7b2 e1c1 c6c5 f6e4 d6e7 e4c5 f8c8 c5e4 c8c1 a1c1 b2a2 g2g4 h8g7 e4g3 e7d8 g3f5 g7g6 c1c6 d8f6 f5h4 g6g7 h4f5');
INSERT INTO evaluations VALUES(49964,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(49964,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(49964,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(49964,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(49964,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(49964,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(49964,36,'cp 36','d7d5 c2c4 c7c6 e1g1 f8d6 f3e5 b8d7 c1f4 d6e7 b1d2 d7e5 f4e5 e8g8 d2f3 a7a5 d1c2 f6d7 e5f4 g7g5 f4d2 f8f7 a1c1 h7h6 f3e1 h6h5 e1d3 f7g7 a2a3 a5a4 f2f3 d7b6 d3e5 c8d7 d2e1 b6c4 e5c4 d5c4');
INSERT INTO evaluations VALUES(49127,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(49127,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(49127,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(49127,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(49127,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(49127,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(49127,36,'cp 53','c2c4 c7c6 c1f4 f8e7 d1c2 b8d7 b1d2 f6e4 f3e5 d7e5 f4e5 e8g8 d2e4 f5e4 e1g1 e7f6 e5f6 g7f6 f2f3 f6f5 c2d2 a7a5 f1f2 e4f3 g2f3 d5c4 d2c3 d8b6 a1c1 b6b4 c3b4 a5b4 c1c4 a8a2 c4b4 a2a5 e2e3 e6e5 d4e5 a5e5');
INSERT INTO evaluations VALUES(49115,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(49115,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(49115,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(49115,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(49115,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(49115,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(49115,36,'cp 45','c7c6 c1f4 b8d7 b1d2 f8e7 d1c2 f6e4 f3e5 e8g8 e5d3 d7f6 h2h4 e4d2 f4d2 f6e4 a1c1 e7f6 d2e3 d8a5 b2b4 a5d8 e1g1 e4d6 c4d5 e6d5 a2a4 d6c4 b4b5 d8e7 e3g5 f6g5 h4g5 e7g5 b5c6 b7c6 d3e5 c4e5 d4e5');
INSERT INTO evaluations VALUES(46519,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(46519,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(46519,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(46519,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(46519,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(46519,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(46519,36,'cp 46','b1c3 f8d6 d1b3 a7a5 c4c5 d6c7 c1f4 e8g8 e1g1 b8d7 c3a4 a8b8 a1c1 f6e4 f1d1 c7f4 g3f4 f8f6 b3e3 f6h6 b2b3 d7f8 a4b2 c8d7 g1h1 d7e8 d1g1 e8h5 f3g5 d8e7 g2e4 f5e4 f2f3 e4f3 e2f3 e6e5 d4e5');
INSERT INTO evaluations VALUES(46515,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(46515,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(46515,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(46515,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(46515,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(46515,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(46515,36,'cp 50','f8d6 d1b3 e8g8 c1f4 a7a5 c4c5 d6c7 c3a4 b8d7 e1g1 f8e8 e2e3 e6e5 f3e5 d7e5 f4e5 c7e5 d4e5 e8e5 a4b6 a8a6 a2a3 e5e8 b3d3 f6e4 b2b4 c8e6 d3d4 e6f7 f1c1 a5b4 a3b4 a6a1 c1a1 d8c7 f2f3 e4g5 a1a7 g5e6 d4c3 g7g5 f3f4 g5f4 e3f4');
INSERT INTO evaluations VALUES(40663,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(40663,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(40663,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(40663,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(40663,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(40663,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(40663,36,'cp 61','c1f4 d6f4 g3f4 e8g8 e2e3 d5c4 f3d2 f6d5 d2c4 b8d7 c4d2 d7b6 d2f3 c8d7 a1c1 d7e8 f3e5 b6d7 e5d3 d8e7 d1b3 d5b6 e1g1 g8h8 c3a4 e8f7 b3c2 d7f6 f1e1 f8d8 c2c5 e7d7 a4b6 a7b6 c5b6 a8a2');
INSERT INTO evaluations VALUES(53412,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(53412,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(53412,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(53412,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(53412,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(53412,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(53412,36,'cp -82','g8f6 f1b5 b8c6 d2d4 c8g4 f3g3 e7e6 g1f3 f8d6 c1f4 d6f4 g3f4 g4f3 b5c6 b7c6 g2f3 d8b8 f4b8 a8b8 b2b3 e8e7 c3a4 a7a5 e1c1 g7g6 h2h4 f6d7 d1d3 h8c8 d3c3 b8b4 h1d1 e6e5 d4e5 b4h4 d1d5 c6d5 c3c8 d7e5 c8c5 h7h5 a4c3 h4h1 c1b2 e5f3 c5d5 h5h4');
INSERT INTO evaluations VALUES(45475,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(45475,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(45475,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(45475,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(45475,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(45475,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(45475,36,'cp -72','f1b5 c8d7 b5d7 d8d7 d2d4 b8c6 c3e2 e7e6 c2c3 f8d6 c1f4 e8g8 f3e3 a8e8 g1f3 f6g4 e3d2 d6f4 d2f4 e6e5 d4e5 c6e5 e1g1 e5f3 f4f3 g4e5 f3f4 e5c4 a1d1 e8e2 f4c4 d5c4 d1d7 e2b2 a2a4 a7a6 a4a5 b2b5 f1e1 b5a5 d7b7');
INSERT INTO evaluations VALUES(55238,30,'cp -200',NULL);
INSERT INTO evaluations VALUES(55238,31,'cp -200',NULL);
INSERT INTO evaluations VALUES(55238,32,'cp -201',NULL);
INSERT INTO evaluations VALUES(55238,33,'cp -226',NULL);
INSERT INTO evaluations VALUES(55238,34,'cp -225',NULL);
INSERT INTO evaluations VALUES(55238,35,'cp -206',NULL);
INSERT INTO evaluations VALUES(55238,36,'cp -207','d8d4 b1c3 c7c6 h2h3 g8f6 c1e3 d4d8 g2g4 c8e6 g4g5 f6d5 e1c1 d8a5 g1e2 b8d7 e3d2 g7g6 e2d4 f8g7 d4e6 f7e6 h3h4 h8f8 c3d5 a5d5 f3d5 e6d5 h4h5 a8d8 c1b1 b7b6 c2c3 d7c5 d2e3 g7e5 h5g6 h7g6 e3c5 b6c5 f1d3 d8d6');
INSERT INTO evaluations VALUES(34650,30,'cp -216',NULL);
INSERT INTO evaluations VALUES(34650,31,'cp -211',NULL);
INSERT INTO evaluations VALUES(34650,32,'cp -192',NULL);
INSERT INTO evaluations VALUES(34650,33,'cp -203',NULL);
INSERT INTO evaluations VALUES(34650,34,'cp -209',NULL);
INSERT INTO evaluations VALUES(34650,35,'cp -174',NULL);
INSERT INTO evaluations VALUES(34650,36,'cp -238','b1c3 e7e5 a2a3 d4h4 g2g3 h4f6 f3e2 f8d6 c1e3 g8e7 g1f3 b8c6 e1c1 e8g8 e2g2 c8g4 f1e2 e7f5 e3f2 f6h6 c1b1 f5e3 f2e3 h6e3 c3d5 e3h6 h2h3 g4e6 h3h4 e6g4 h1f1 f7f5 f3g5 g4e2 g2e2 c6d4 e2d3 c7c6');
INSERT INTO evaluations VALUES(56142,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(56142,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56142,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(56142,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(56142,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56142,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(56142,36,'cp 0','d8d5 d2d4 e4d3 d1d3 d5e6 f1e2 c8d7 g1f3 b8c6 e1g1 g8f6 b1c3 e8c8 c1e3 a7a6 a1d1 f8d6 f3g5 e6e8 d3c4 c6e5 c4d4 e5c6 d4a4 c6e5');
INSERT INTO evaluations VALUES(34913,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34913,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(34913,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(34913,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34913,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(34913,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34913,36,'cp -4','d2d4 e4d3 d1d3 d5e6 c1e3 f8c5 b1c3 c5e3 d3e3 e6e3 f2e3 c8e6 e1c1 a7a6 g1f3 b8d7 c3e2 g8f6 e2f4 e8c8 f4e6 f7e6 f3g5 d7c5 d1d8 h8d8 f1c4 d8e8 b2b4 h7h6 b4c5 h6g5 c5c6 b7b5 c4e2 f6d5 h1e1 d5e3 e2b5 e3g2 e1e2 a6b5 e2g2 e6e5 g2g5');
INSERT INTO evaluations VALUES(5639,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(5639,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(5639,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(5639,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(5639,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5639,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(5639,36,'cp 51','d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 e1g1 c6d4 d1d4 c7c6 a2a4 c8e6 c4e6 f7e6 e4e5 f6d5 e5d6 e7d6 c3e4 d6c7 d4c4 f8f5 a1a3 d8e7 a3b3 a8f8 c4e2 c7f4 c2c4 f4c1 f1c1 d5f4 e2e3 a7a6 g1h1 b7b5 a4b5 c6b5 c4b5 a6b5');
INSERT INTO evaluations VALUES(6702,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(6702,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(6702,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(6702,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(6702,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(6702,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(6702,36,'cp -33','g8e7 e1g1 e8g8 d4d5 c6a5 f1e1 d7d6 a1c1 a5c4 d2c4 c8g4 h2h3 g4f3 d1f3 h7h6 e1d1 f8e8 c4e3 a8c8 c1c3 d8d7 f3g4 e8d8 c3a3 a7a6 a3b3 c7c6 g4d7 d8d7 d5c6 b7c6 b3b6 c8a8 e3f5 e7f5 e4f5 a8c8 b6a6 d6d5 a2a4');
INSERT INTO evaluations VALUES(10263,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(10263,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(10263,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(10263,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(10263,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(10263,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(10263,36,'cp -34','d7d5 e4d5 f6d5 f1e2 e8g8 e1g1 f8e8 c2c4 d5f4 c1f4 e5f4 b2b4 c6d4 b4c5 d4e2 g1h1 e2d4 b1c3 c8f5 c3d5 d4e6 d1d2 c7c6 d5f4 e6c5 a1d1 d8c7 f3d4 f5d7 d4e2 a8d8 f2f3 c5e6 d3d4 b7b6 f1e1 e6f4 d2f4 c7f4 e2f4 e8e1 d1e1 g8f8 e1b1 h7h6 g2g4 g7g5 f4h5');
INSERT INTO evaluations VALUES(17645,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(17645,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(17645,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(17645,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(17645,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(17645,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(17645,36,'cp 57','b5c3 g8f6 f1e2 c8e6 c1e3 g7g6 e1g1 f8g7 h2h3 e8g8 b1d2 c6d4 f1e1 d8b6 a1b1 f8c8 e2d3 h7h6 d3f1 f6d7 d2f3 e6c4 f1c4 c8c4 c3d5 d4f3 d1f3 b6d8 a2a4 c4a4 b1c1 b7b5 c1c7 a4c4 e1c1 c4c7 c1c7 d7f6 e3b6');
INSERT INTO evaluations VALUES(17649,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(17649,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(17649,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(17649,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17649,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(17649,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(17649,36,'cp 41','g8f6 f1d3 f8e7 e1g1 e8g8 h2h3 c8e6 c3d5 f6d7 d5e7 d8e7 b1c3 d7c5 b2b3 b7b5 c1e3 c5d3 d1d3 b5c4 b3c4 c6b4 d3e2 a8c8 c3d5 e6d5 c4d5 c8c2 e2d1 c2c7 e3d2 b4c2 a1b1 e7h4 d1g4 h4g4 h3g4 c2a3 b1b3 a3b5 a2a4 b5d4 b3b6 d4e2 g1h1 c7c2 d2e3');
INSERT INTO evaluations VALUES(11224,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(11224,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(11224,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(11224,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(11224,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(11224,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(11224,36,'cp 56','c1e3 f6g4 e3c1 g4f6 f1d3 g7g6 e1g1 f8g7 c1g5 h7h6 g5e3 e8g8 f2f3 c8e6 c3d5 c6d4 b1c3 e6d5 c3d5 f6d5 c4d5 h6h5 d1d2 a8c8 e3d4 e5d4 f3f4 c8c5 g2g3 d8c7 a1e1 f8c8 h2h4 c7d7 g1g2 b7b5 f1f2 g8h7 g2h2');
INSERT INTO evaluations VALUES(18605,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(18605,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(18605,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(18605,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(18605,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(18605,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(18605,36,'cp 126','d4d5 g7g6 f1e2 f8g7 a2a4 a7a6 b1c3 g8f6 h2h3 e8g8 e1g1 b7b6 c1e3 a8b8 e2d3 f6e8 d1d2 e8c7 e3h6 b6b5 a4b5 a6b5 h6g7 g8g7 b2b4 e7e5 d5e6 c7e6 b4c5 d7e5 f3e5 d6e5 c5c6 d8d6 f1b1 d6c6 b1b5 b8b5');
INSERT INTO evaluations VALUES(18606,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(18606,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(18606,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(18606,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(18606,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(18606,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(18606,36,'cp 126','g8f6 b1c3 g7g6 f1e2 a7a6 a2a4 f8g7 e1g1 e8g8 c1g5 h7h6 g5e3 b7b6 h2h3 g8h7 e3f4 f6e8 f4g3 a8b8 e2c4 d7e5 f3e5 d6e5 d1e2 e8d6 c4d3 b8a8 f1e1 c8b7 b2b3 d8d7 a1d1 e7e6 d5e6 d7e6 c3d5 b7d5 e4d5');
INSERT INTO evaluations VALUES(24831,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24831,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(24831,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24831,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24831,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(24831,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24831,36,'cp 44','e2e3 e7e6 c2c4 f8d6 f4g5 e8g8 f1e2 h7h6 g5h4 d6e7 e1g1 a7a5 a2a3 f6e4 h4e7 c6e7 d2e4 f5e4 f3d2 e4g6 c4d5 e6d5 b2b4 a5b4 a3b4 a8a1 d1a1 c7c6 a1a7 d8d6 b4b5 c6b5 e2b5 g6f5 d2f3 f5e4 f1c1 f8b8 f3e5');
INSERT INTO evaluations VALUES(34118,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(34118,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(34118,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(34118,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(34118,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(34118,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(34118,36,'cp 131','d2d4 g8f6 g1f3 e6d6 e1g1 e7e6 d1d3 c7c5 f1d1 c5d4 f3d4 b8d7 a2a4 d6c7 d3g3 c7g3 h2g3 f8b4 c1d2 b4c5 d2e1 d7e5 a4a5 c8d7 d4f3 e5f3 e2f3 d7c6 c3a4 c5e7 a4b6 a8d8 d1d8 e8d8 a1d1 d8c7 e1c3 f6e4 c3g7');
INSERT INTO evaluations VALUES(51237,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(51237,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(51237,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(51237,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(51237,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(51237,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51237,36,'cp 31','c4d5 a8a7 a2a4 g8f6 d2d3 b5b4 d5b3 d7d6 c2c3 b4c3 b2c3 e7e6 e1g1 b8c6 c1f4 f8e7 b1d2 d6d5 f1e1 e8g8 a1b1 h7h6 h2h3 a6a5 d1c2 c8b7 f4g3 b7a8 c2d1 a7b7 e4d5 f6d5 d2e4 d5f6 b3c4 b7b1 d1b1 f6d5 c4b5');
INSERT INTO evaluations VALUES(51249,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(51249,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(51249,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(51249,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(51249,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(51249,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51249,36,'cp 31','a8a7 a2a4 g8f6 d2d3 b5b4 d5b3 d7d6 c2c3 b4c3 b2c3 e7e6 e1g1 f8e7 c1f4 b8c6 f1e1 e8g8 h2h3 d6d5 b1d2 a6a5 d1c2 h7h6 a1b1 c8d7 f4g3 a7b7 b1c1 d8b6 e4d5 e6d5 c2a2 d7e6 e1e6 f7e6 c1e1 f6h5 b3d5');
INSERT INTO evaluations VALUES(2,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(2,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(2,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(2,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(2,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(2,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(2,36,'cp 29','a2a4 g8f6 d2d3 b5b4 d5b3 d7d6 c2c3 b4c3 b2c3 e7e6 e1g1 f8e7 f1e1 e8g8 h2h3 b8c6 c1f4 d6d5 b1d2 a6a5 f4g3 h7h6 a1b1 f8e8 f3e5 c6e5 g3e5 d5e4 d2e4 f6e4 d3e4 d8d1 e1d1 a7d7 b3c4 d7d1 b1d1 e8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(52815,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(52815,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(52815,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(52815,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(52815,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(52815,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(52815,36,'cp -30','d5e4 b1c3 g8f6 c3e4 f6e4 g2e4 e7e5 d2d3 f8c5 e4g2 e8g8 d1e2 f8e8 g1f3 b8a6 e1g1 f7f6 c1e3 c8g4 c2c3 c5e3 e2e3 d8b6 f1e1 b6e3 e1e3 a6c5 h2h3 g4f3 g2f3 c5a4 a1b1 g8f7 d3d4 e5d4 c3d4 a8d8 e3e8 f7e8 b1e1 e8f8 b2b3 a4c3');
INSERT INTO evaluations VALUES(52929,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(52929,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(52929,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(52929,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(52929,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(52929,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(52929,36,'cp -32','b1c3 g8f6 c3e4 f6e4 g2e4 e7e5 d2d3 f8c5 d1e2 e8g8 e4g2 f8e8 g1f3 c8g4 h2h3 g4e6 e1g1 f7f6 c1e3 c5d6 a1e1 b8d7 b2b3 d7f8 f3d2 d6b4 a2a3 b4a3 e1a1 a3b4 a1a7 a8a7 e3a7 b7b6 d2c4 b4c5');
INSERT INTO evaluations VALUES(55499,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(55499,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(55499,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(55499,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(55499,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(55499,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(55499,36,'cp 76','g8f6 c1e3 c7c6 h2h3 c6d5 e4d5 h7h6 f3d2 f8e7 d1f3 b8a6 e1c1 c8d7 g2g4 a8c8 d2b3 a6b4 a2a3 c8c3 b2c3 e5e4 f3g3 b4d5 e3d4 d8c7 c3c4 d5b6 d4c5 e4e3 c5d6 e7d6 g3d6 c7d6 d1d6 d7c6 d6c6 b7c6 f2e3 b6d7 f1g2 d7e5 h1d1 e5c4 g2c6 e8e7 b3d4 h8c8 d1d3 c4e5 d3c3 e7d8');
INSERT INTO evaluations VALUES(47762,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(47762,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(47762,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(47762,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(47762,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(47762,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(47762,36,'cp 87','h2h3 b8d7 c1e3 f8e7 g2g4 h7h6 d1d2 c7c6 e1c1 d8a5 a2a3 a8b8 c1b1 c6d5 c3d5 a5d2 f3d2 f6d5 e4d5 a7a6 f2f4 b7b5 d2b3 e7d8 f1g2 e8g8 e3d2 f8e8 d2b4 d8c7 f4f5 d7f6 b3a5 c7a5 b4a5');
INSERT INTO evaluations VALUES(10729,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(10729,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(10729,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(10729,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(10729,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(10729,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(10729,36,'cp 37','e1g1 d6c5 b1c3 e8g8 c1g5 c5e7 h2h3 h7h6 g5h4 d7d6 a2a4 a7a5 d3d4 e5d4 f3d4 f6e4 h4e7 c6e7 c3e4 d6d5 d1f3 d5c4 a1d1 e7d5 f1e1 c7c6 f3g3 f8e8 e4d6 e8e1 d1e1 c8d7 d6c4 b7b5 c4e5 b5a4 c2c4 d5f6 d4c6 d7c6 e5c6');
INSERT INTO evaluations VALUES(14841,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14841,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14841,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14841,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14841,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14841,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14841,36,'cp 0','f8e7 e1g1 e8g8 a2a4 h7h6 f3e5 c6e5 d4e5 f6d7 e2e4 c7c6 f2f4 e7c5 g1h1 a7a5 b2b3 b7b6 e4d5 e6d5 c3e2 c8a6 f1e1 c5f2 e1f1');
INSERT INTO evaluations VALUES(34707,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(34707,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(34707,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(34707,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(34707,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(34707,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(34707,36,'cp 67','g7g6 b1c3 g8f6 a2a4 f8g7 c1g5 h7h6 g5h4 g6g5 h4g3 c7c6 f1c4 c6d5 c3d5 f6d5 c4d5 c8e6 e1g1 b8d7 d5e6 e7e6 f1e1 a8c8 h2h4 g5g4 f3h2 h6h5 h2f1 c8c4 f1d2 c4c6 b2b3 e8g8 d2f1 d7c5 b3b4 c5d7 f1e3 f8c8 d1d3');
INSERT INTO evaluations VALUES(34775,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(34775,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(34775,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(34775,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(34775,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(34775,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(34775,36,'cp 85','c8g4 c4e2 b8c6 c2c3 g8e7 d2d3 h7h5 b1a3 g4f3 e2f3 g7g5 f3h5 f6h6 g2g4 e7g6 d1b3 e8c8 b3f7 g6h4 f2f3 f8e7 f7c4 d8f8 c1e3 f8f3 f1f3 h4f3 g1h1 h6f6 a1f1 f3d2 f1f6 d2c4 e3g5 c4b2 f6h6 h8f8 g5e7 c6e7 d3d4 b2d3 g4g5 d3f2 h1g2 f2e4 g5g6 e4c3 g6g7 f8g8 h5g4 c8b8');
INSERT INTO evaluations VALUES(32921,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(32921,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(32921,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(32921,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(32921,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(32921,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(32921,36,'cp 63','c4e2 b8c6 c2c3 g8e7 d2d3 h7h6 b1d2 g4d7 b2b4 g7g5 d2c4 a7a6 a2a4 g5g4 f3e1 h6h5 e1c2 d6d5 c4e3 d5d4 c3d4 c6d4 c1b2 f8g7 c2d4 e5d4 f2f4 g4f3 e2f3 f6h6 e3c4 d7e6 b2c1 h6g6 e4e5 e6c4 d3c4 g7e5 f3b7');
INSERT INTO evaluations VALUES(32928,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(32928,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(32928,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(32928,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(32928,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(32928,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(32928,36,'cp 86','b8c6 c2c3 h7h5 b1a3 a7a6 a3c2 f6g6 d2d3 f8e7 c2e3 g8f6 f1e1 e8c8 c3c4 g4f3 e2f3 c6d4 a1b1 f6d7 b2b4 e7g5 e3d5 c7c6 d5e3 g6f6 b4b5 d4f3 g2f3 a6b5 c4b5 g5e3 c1e3 c6c5 a2a4 d7f8 a4a5 f8e6 a5a6');
INSERT INTO evaluations VALUES(36351,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36351,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36351,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(36351,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36351,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36351,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(36351,36,'cp 0','c7c5 d4c5 f6d7 f4b8 d7b8 a2a3 e7c5 e1g1 a7a5 c2c4 c5e7 c4d5 e6d5 d2b1 b8c6 b1c3 c8e6 d3b1 g7g6 b1a2 e7f6 d1c2 a8c8 a1d1 c6e5 f3d4 e5c4 h2h3 d8c7 a2c4 c7c4 c2b1 c4a6 c3e2 a6b6 e2c3 b6a6');
INSERT INTO evaluations VALUES(35837,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35837,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(35837,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(35837,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35837,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(35837,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(35837,36,'cp 9','d4c5 f6d7 f4b8 d7b8 e1g1 a7a5 a2a3 e7c5 c2c4 c5e7 c4d5 e6d5 d2b1 b8c6 b1c3 c8e6 d3b1 g7g6 d1d3 d8b6 c3d5 e6d5 d3d5 b6b2 a1a2 b2c3 d5d3 c3d3 b1d3 f8d8 d3e2 a5a4 f1b1 c6a5 b1b5 g8g7 g2g3 d8c8 g1g2 c8c5 b5c5');
INSERT INTO evaluations VALUES(38831,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(38831,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38831,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(38831,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(38831,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(38831,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(38831,36,'cp 13','e3f4 b7b6 c2c3 c7c5 b1d2 d8c7 g2g3 g8f6 f1b5 b8c6 d1e2 e8g8 e1g1 c6e7 f1e1 e7f5 b5d3 a7a5 a2a4 f5d6 f3e5 c8b7 d2b1 f6e4 f2f3 e4f6 b1a3 b7c6 d3c2 a8c8 g3g4 f6d7 e2f2 c5d4 c3d4 d7e5 f4e5');
INSERT INTO evaluations VALUES(38830,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(38830,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38830,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(38830,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(38830,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38830,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(38830,36,'cp 13','b7b6 b1c3 g8e7 h2h4 h7h5 g2g3 c7c5 f1h3 b8d7 e1g1 e8g8 f1e1 d7f6 c3b1 e7c6 a2a3 a7a5 b1d2 d8d6 a1c1 c6d4 f3d4 c5d4 d2f3 c8d7 d1d4 a8c8 c2c3 d6c5 h3f1 c8c7 f1d3 c5d4 c3d4 f8c8 c1c7 c8c7 f3e5 f6e8 e5d7');
INSERT INTO evaluations VALUES(48256,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(48256,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(48256,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(48256,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(48256,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(48256,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(48256,36,'cp -102','c7c5 e2e3 b8c6 c4d5 e6d5 f1b5 f8d6 g1e2 e8g8 d4c5 d6c5 e1g1 f8e8 c3a4 c5d6 c1d2 c8d7 a1c1 d8e7 b5c6 b7c6 b2b4 e7e5 g2g3 e5e7 a4c5 a7a5 c5d7 e7d7 b4a5 c6c5 d1c2 c5c4 f1d1 d6c5 g1g2 c5e3');
INSERT INTO evaluations VALUES(50640,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(50640,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(50640,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(50640,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(50640,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(50640,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(50640,36,'cp 75','g1f3 d5e4 c3e4 g8f6 f1d3 f6e4 d3e4 b8d7 e4d3 d7f6 h2h3 c8f5 d3f5 d8a5 c2c3 a5f5 e1g1 e7e6 f3e5 f8d6 g2g4 f5e4 f1e1 e4d5 c3c4 d5a5 c1f4 a5c7 c4c5 d6e5 f4e5 c7d7 d1f3 e8c8 a2a4 h7h5 g4g5 d7d5 f3f4 f6d7 e5d6 h8e8');
INSERT INTO evaluations VALUES(50639,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(50639,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(50639,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(50639,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(50639,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(50639,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(50639,36,'cp 70','d5e4 c3e4 g8f6 f1d3 f6e4 d3e4 b8d7 e4d3 d7f6 f3e5 c8e6 c1e3 g7g6 e1g1 f6d5 e3d2 f8g7 d1e2 d5f6 c2c3 e8g8 f2f4 e6f5 d3f5 g6f5 d2e3 d8d5 a2a4 a6a5 c3c4 d5e6 f1e1 f6e4 a1d1 g7e5 f4e5 e6g6 e3f4 g8h8');
INSERT INTO evaluations VALUES(51703,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(51703,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(51703,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(51703,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(51703,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(51703,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(51703,36,'cp 174','c3b5 d6d5 c4b3 b8d7 b5c3 d5d4 c3e2 f8d6 e2g3 d7c5 d2d3 c5b3 a2b3 g8e7 f3d2 e8g8 d2c4 d6c7 e1g1 e7g6 c1d2 c6c5 d1e1 h7h6 g3e2 c8e6 f2f4 e6c4 b3c4 e5f4 d2f4 a7a5 b2b3 a8a6 f4d2 g8h7 e2f4 g6f4 d2f4 a6f6 e1g3 c7f4 f1f4 f6f4 g3f4');
INSERT INTO evaluations VALUES(51973,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(51973,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(51973,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(51973,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(51973,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(51973,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(51973,36,'cp -12','g8f6 b1c3 d8a5 f1d3 d7d5 e3d2 a5b6 d4c5 f8c5 d1e2 d5e4 d3e4 f6e4 c3e4 e8g8 d2c3 c5e7 e1c1 b8c6 g1f3 c6b4 e2c4 a7a6 a2a3 b4d5 c3d4 b6d8 e4c3 b7b5 c4e2 c8b7 e2e5 e7f6 e5g3 f6d4 d1d4 d5c3 d4d8 c3e2 c1d2 e2g3 d8f8 a8f8 h2g3 b7d5 h1e1');
INSERT INTO evaluations VALUES(57434,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(57434,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(57434,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(57434,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(57434,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(57434,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(57434,36,'cp 64','d3d4 e5d4 d1d4 b8c6 d4e3 f6f5 e4f5 f8e7 f1d3 d7d5 g1e2 g8f6 e1g1 e8g8 e3h3 c6b4 f1e1 b4d3 c2d3 g7g6 e2d4 f6h5 c1h6 c8f5 d4f5 f8f5 g2g4 h5f4 h3e3 e7g5 g4f5 g5h6 f5g6 d8g5 e3g3 h7g6 g3g5 h6g5 b1c3 c7c6 d3d4 g5f6');
INSERT INTO evaluations VALUES(6405,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(6405,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(6405,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(6405,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(6405,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(6405,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(6405,36,'cp 15','c6e5 b1c3 e5f3 d1f3 g8e7 b2b3 e8g8 c1b2 c5d4 e1c1 d7d6 h2h4 e7c6 c3e2 d4b2 c1b2 f7f5 e4f5 c8f5 f3c3 d6d5 e2g3 f5g4 f1e2 g4e2 g3e2 f8f2 e2d4 c6d4 c3d4 d8f6 d4f6 g7f6 h1f1 f2f1 d1f1 g8g7 g2g4 a8e8 d2d3 a7a5 f1f5 c7c6 a2a3 e8e2 b2c3');
INSERT INTO evaluations VALUES(6845,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(6845,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(6845,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6845,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(6845,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(6845,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(6845,36,'cp 12','b1c3 e5f3 d1f3 d7d6 d2d3 g8e7 c1e3 c5e3 f3e3 e8g8 d3d4 d6d5 e1c1 c7c6 d1e1 c8e6 f1d3 c6c5 e4d5 e7d5 c3d5 d8d5 e3e4 d5g5 c1b1 g7g6 e4b7 c5d4 b7b5 g5b5 d3b5 a8b8 b5d3 e6d5 d3e4 f8d8 e4d5 d8d5 e1e7 d5g5 g2g3 a7a5 h1e1');
INSERT INTO evaluations VALUES(10226,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(10226,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(10226,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(10226,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10226,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(10226,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(10226,36,'cp 27','d2d4 e5d4 b2b4 c5b6 b4b5 c6e7 e4e5 f6d5 c3c4 b6a5 c1d2 a5d2 c2d2 d5b6 d2d4 d7d6 b1c3 c8e6 a1d1 e8g8 f1d3 e7g6 e5d6 c7d6 c3e4 d6d5 e4g5 e6d7 c4c5 d8e7 e1f1 h7h6 g5h3 b6c4 d3c4 d7h3 c4d3 h3e6 h2h4 f8c8 d3g6 f7g6 d1c1 a7a6 b5b6 c8e8 f1g1 a8c8 h4h5 g6g5');
INSERT INTO evaluations VALUES(10716,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(10716,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(10716,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(10716,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(10716,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(10716,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(10716,36,'cp 40','e1g1 a7a6 b5a4 e8g8 h2h3 b7b5 a4b3 c6a5 d3d4 f8e8 d4e5 d6e5 f3e5 e8e5 b1c3 c8b7 c1f4 a5b3 a2b3 e5e8 e4e5 b5b4 c3a4 f6d5 f4g3 d7d6 e5d6 c7d6 d1d2 d8f6 c2c3 b4c3 b2c3 e8e4 f1e1 a8e8 f2f3 e4e1 a1e1 e8e6 c3c4 e6e1 d2e1 d5f4');
INSERT INTO evaluations VALUES(10715,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10715,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(10715,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10715,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(10715,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(10715,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(10715,36,'cp 28','a7a6 b5a4 e8g8 c2c3 f8e8 c1g5 d6e7 c3c4 d7d6 b1c3 c8g4 g5e3 f6d7 h2h3 g4f3 d1f3 d7f8 g2g3 f8e6 f3g4 a8b8 b2b4 e8f8 a1b1 c6d4 c3e2 d4e2 g4e2 c7c6 f2f4 e5f4 g3f4 e7f6 e2f2 b7b5 a4b3 b5c4 b3c4');
INSERT INTO evaluations VALUES(11922,30,'cp 248',NULL);
INSERT INTO evaluations VALUES(11922,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(11922,32,'cp 262',NULL);
INSERT INTO evaluations VALUES(11922,33,'cp 264',NULL);
INSERT INTO evaluations VALUES(11922,34,'cp 263',NULL);
INSERT INTO evaluations VALUES(11922,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(11922,36,'cp 259','f3g5 d7d5 c3d5 h7h6 d5f6 d8f6 g5f7 f8c5 d1f3 f6f3 g2f3 h8f8 c4b5 e8f7 b5c6 a8b8 c6b5 f7e7 b5e2 c8h3 d2d3 f8f6 h1g1 g7g5 c2c3 b8f8 b2b4 c5d6 c1e3 a7a5 b4b5 h3e6 e1d2 e6f7 g1h1 f7e8 d3d4 e8h5 h2h4 h5f3 h4g5 h6g5 e2f3 f6f3');
INSERT INTO evaluations VALUES(11921,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(11921,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(11921,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(11921,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(11921,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(11921,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(11921,36,'cp 249','d7d5 c3d5 h7h6 d5f6 d8f6 g5f7 f8c5 d1f3 f6f3 g2f3 h8f8 c4b5 e8f7 b5c6 a8b8 h1g1 g7g5 a2a3 f7e7 b2b4 c5d6 e1e2 c8e6 d2d3 f8f6 c6d5 e6d5 e4d5 b8f8 c1e3 f6f3 g1g3 f3f5 c2c4 e7d7 g3h3 f5f6 h3h5 a7a5 c4c5 d6e7 h2h4 g5h4 h5e5 b6c5 b4c5 f8b8 f2f4 h4h3 a1h1');
INSERT INTO evaluations VALUES(19786,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19786,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19786,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(19786,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19786,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(19786,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(19786,36,'cp 10','e7e6 f1d3 g8f6 e1g1 f8e7 b1d2 e8g8 d4c5 e7c5 b2b4 c5d6 b4b5 c6a5 c3c4 d5c4 d3c4 a5c4 d2c4 d6c5 d1b3 d8e7 c1b2 b7b6 c4e5 c8b7 e5c6 b7c6 b5c6 e7c7 b3a4 a8c8 b2f6 g7f6 a4g4 g8h8 f3d4 f6f5 g4h4 c7e7 h4e7 c5e7 a2a4 f8d8');
INSERT INTO evaluations VALUES(20415,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(20415,31,'cp -178',NULL);
INSERT INTO evaluations VALUES(20415,32,'cp -160',NULL);
INSERT INTO evaluations VALUES(20415,33,'cp -175',NULL);
INSERT INTO evaluations VALUES(20415,34,'cp -159',NULL);
INSERT INTO evaluations VALUES(20415,35,'cp -155',NULL);
INSERT INTO evaluations VALUES(20415,36,'cp -163','c6e5 c1f4 f8d6 f1b5 e8f8 d1e2 e5g6 f4d6 d8d6 e1c1 g8f6 g2g3 a7a6 b5d3 d6e5 f2f4 e5e2 c3e2 g6e7 c2c3 g7g6 d4f3 c8d7 e2d4 e7c6 d3c2 a8c8 h1e1 h7h5 d4c6 d7c6 a2a4 a6a5 d1d2 f8e7 d2e2 f6e4 f3e5 e4c5 b2b3 h5h4 c1b2');
INSERT INTO evaluations VALUES(20778,30,'cp -156',NULL);
INSERT INTO evaluations VALUES(20778,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(20778,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(20778,33,'cp -148',NULL);
INSERT INTO evaluations VALUES(20778,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(20778,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(20778,36,'cp -143','c1f4 e5g6 f4g3 a7a6 h2h4 f8d6 h4h5 g6e7 d1g4 e8f8 g3d6 d8d6 g4h4 e6e5 d4b3 h7h6 e1c1 g8f6 f1e2 c8e6 f2f4 e5e4 c1b1 a8c8 h4f2 f8g8 h1g1 e6g4 e2g4 f6g4 f2g3 g4f6 g3h4 c8c3 b2c3 g8h7 g2g4 d6f4 b3d4 h8c8 d1f1 f4g5 h4g5 h6g5 d4f5 e7f5 f1f5');
INSERT INTO evaluations VALUES(20777,30,'cp -158',NULL);
INSERT INTO evaluations VALUES(20777,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(20777,32,'cp -156',NULL);
INSERT INTO evaluations VALUES(20777,33,'cp -157',NULL);
INSERT INTO evaluations VALUES(20777,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(20777,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(20777,36,'cp -155','f8d6 f1b5 e8f8 d1e2 e5g6 f4d6 d8d6 e1c1 g8f6 g2g3 a7a6 b5d3 d6e5 f2f4 e5e2 c3e2 g6e7 h1e1 c8d7 d4f3 g7g6 c2c3 a8c8 d3c2 h7h5 e2d4 e7c6 a2a4 f8e7 f3e5 c6d4 d1d4 a6a5 c1b1 h8d8 h2h3 d7c6 g3g4 f6d7 e5f3 h5g4 h3g4');
INSERT INTO evaluations VALUES(22516,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(22516,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(22516,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(22516,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(22516,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(22516,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(22516,36,'cp 30','b1c3 c6a5 c4b3 d7d6 f2f4 e5f4 c1f4 a5b3 a2b3 g8e7 g1e2 e7g6 e1g1 f8e7 f4e3 e8g8 e2g3 e7f6 d1h5 c7c6 d3d4 f8e8 b3b4 a7a6 b4b5 c8e6 b5c6 b7c6 g3f5 e6f5 h5f5 d8e7 f5f2 e7e6 g1h1 d6d5 e4e5 f6e7');
INSERT INTO evaluations VALUES(22513,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(22513,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(22513,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22513,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(22513,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(22513,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(22513,36,'cp 32','c6a5 c4b3 g8e7 f2f4 a5b3 a2b3 e5f4 c1f4 d7d6 g1e2 e7g6 f4e3 f8e7 e1g1 e8g8 c3d5 e7h4 e2f4 g6f4 d5f4 h4g5 d1e1 c8d7 e3d4 g5f4 f1f4 a7a6 e1f2 f7f6 h2h3 d8e7 a1f1 a8e8 f4f3 e7e6 f3g3 g8h7 g1h2');
INSERT INTO evaluations VALUES(25103,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(25103,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(25103,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(25103,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(25103,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(25103,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(25103,36,'cp 151','g5e6 f7e6 e1g1 d8f6 c4b5 f6g6 b1d2 f8d6 b5c6 b7c6 d2e4 e8g8 g1h1 d5f4 g2g3 f4d5 d1e2 a7a5 a2a4 h7h6 c1d2 d6b4 f2f3 f8f5 f1f2 a8b8 d2e3 b4d6 a1b1 g8h8 c2c3 d5f6 f2g2 f6e4 d3e4 f5f8 b1f1');
INSERT INTO evaluations VALUES(25097,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(25097,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(25097,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(25097,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(25097,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(25097,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(25097,36,'cp 120','f7e6 b1d2 d8d7 e1g1 e8c8 a2a4 f8e7 a4a5 a7a6 c2c3 h7h6 d2e4 c6a7 c1e3 c8b8 d1h5 g7g5 f1e1 a7b5 h5g6 b5d6 e4c5 d7c6 c4d5 e6d5 g6e6 h8e8 e6e5 d6c4 c5a6 c6a6 d3c4 e7c5 e5g3 a6c4 h2h3 c5e3 e1e3 d5d4 g3g4 e8e3 f2e3');
INSERT INTO evaluations VALUES(25418,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(25418,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(25418,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(25418,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(25418,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(25418,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(25418,36,'cp 150','e1g1 d8d7 b1d2 e8c8 a2a4 h7h6 d2e4 f8e7 a4a5 a7a6 c2c3 c6a7 c4b3 c8b8 b3a4 a7c6 b2b4 d5f6 b4b5 c6a5 b5a6 d7d3 d1d3 d8d3 a4c2 d3d5 a6b7 b8b7 c1e3 f6e4 c2e4 a5c4 f1b1 b7c6 a1a4 c4d6 e4f3 c6d7 c3c4 d5d3 c4c5 d6f5 c5c6 d7d6');
INSERT INTO evaluations VALUES(26546,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26546,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(26546,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(26546,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(26546,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(26546,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(26546,36,'cp 49','g8f6 g1f3 e7e6 c2c4 f8d6 f4g5 d6e7 b1c3 e8g8 c4d5 f6d5 g5e7 d8e7 d1b3 d5b6 f1e2 a7a5 c3a4 b6a4 b3a4 c6b8 a1c1 c7c6 f3e5 f8d8 e1g1 b8d7 e5d7 d8d7 c1c5 d7d5 f1c1 f5e4 f2f3 e7g5 a4b3 d5c5 c1c5 e4d5 b3b7 a8f8');
INSERT INTO evaluations VALUES(24836,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(24836,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(24836,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(24836,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(24836,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(24836,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(24836,36,'cp 54','g1f3 e7e6 c2c4 f8d6 f4g5 d6e7 b1c3 e8g8 a1c1 h7h6 g5f4 e7d6 f4g3 c6e7 g3h4 c7c6 h4f6 g7f6 f1d3 f5g4 d3e2 g4f5 e1g1 g8h8 f3h4 f5h7 g2g3 d5c4 e2c4 f6f5 d1h5 e7g8 c4d3 h8g7 h5e2 g8f6 f1d1 d6c7 d3c2 d8d7');
INSERT INTO evaluations VALUES(24837,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(24837,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(24837,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(24837,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(24837,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(24837,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(24837,36,'cp 30','e7e6 c2c4 f8d6 f4g5 d6e7 f1d3 f6e4 g5e7 c6e7 d1c2 e8g8 e1g1 c7c6 f3h4 g7g6 c4c5 h7h5 b2b4 a7a6 a3a4 g8g7 b1c3 e4c3 c2c3 f5d3 c3d3 e7f5 h4f3 g6g5 b4b5 g5g4 b5c6 g4f3 c6b7 a8b8 d3a6');
INSERT INTO evaluations VALUES(25169,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(25169,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(25169,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(25169,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25169,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(25169,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(25169,36,'cp 49','c2c4 f8d6 f4g5 d6e7 b1c3 e8g8 c4d5 e6d5 f1d3 f6e4 g5e7 c6e7 d3e4 f5e4 c3e4 d5e4 f3d2 f8e8 d2e4 e7f5 e4c3 f5d4 e1g1 d4e6 d1b3 a8b8 b3a4 a7a6 a1d1 d8f6 d1d2 h7h6 f1d1 f6e5 h2h3 e8d8 d2d8 e6d8 a4c4 c7c6');
INSERT INTO evaluations VALUES(26680,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(26680,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(26680,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(26680,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(26680,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(26680,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(26680,36,'cp 84','b1d2 e7e6 d2c4 g8f6 a2a3 f8e7 b2b3 a7a5 c1b2 e8g8 a1c1 h7h6 f1d3 g4f5 e1g1 f5d3 d1d3 d8d5 d3c2 f8d8 f1e1 f6e4 e1d1 d5f5 f3e1 e4f6 c2f5 e6f5 c1c2 e7d6 g1f1 c6e7 d1c1 e7d5 g2g3 c7c6 e1d3 f6e4 f1e2 e4f6 c4d6 d8d6');
INSERT INTO evaluations VALUES(27738,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(27738,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(27738,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(27738,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(27738,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(27738,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(27738,36,'cp -87','c4e6 f7e6 d2d4 e6f7 e1g1 f8e7 d1f3 b8c6 f3b3 f7e8 d4d5 c6e5 b1c3 e5g6 b3b7 h8f8 b7b3 e8f7 f2f4 d8b8 g2g3 b8b6 b3b6 c7b6 c1e3 f7g8 e3d4 a8c8 a1d1 f6d7 h2h4 e7f6 c3b5 c8c2 b5d6 f6d4 d1d4 c2b2 a2a4 b2a2 e4e5 g6e7 h4h5 h7h6');
INSERT INTO evaluations VALUES(27737,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(27737,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(27737,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(27737,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(27737,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(27737,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(27737,36,'cp -76','f7e6 d2d4 e6f7 e1g1 f8e7 d1f3 b8c6 f3b3 f7e8 c2c3 d8d7 b1d2 d6d5 e4e5 f6h5 f2f4 g7g6 b3b7 e8f7 b7a6 h8f8 b2b4 f7g8 d2b3 c6d8 c1d2 c7c6 b3a5 a8c8 a2a4 f8f7 a6d3 h5g7 f1f2 d8b7 a5b3 c8b8 d2e3 b7d8 d3e2 d8e6 g2g3');
INSERT INTO evaluations VALUES(27739,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(27739,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(27739,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(27739,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(27739,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(27739,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(27739,36,'cp -67','d2d4 e6f7 b1c3 f8e7 f2f4 h8f8 d1f3 f7g8 c1e3 d8d7 e1c1 b8a6 c1b1 d7g4 f3f1 c7c6 h2h3 g4d7 g2g4 a6c7 f1g2 b7b5 g4g5 f6h5 f4f5 b5b4 c3e2 d6d5 g2g4 d5e4 g4h5 d7f5 h1f1 f5f1 d1f1 f8f1 e3c1 e4e3 g5g6 h7g6 h5g6 c7d5 g6c6 a8d8 c6e6 g8h8 e6g4 d5f6 g4h4 h8g8 e2g3 f1g1 g3e2 g1d1 h4g3 f6d5 g3e5 e7f6');
INSERT INTO evaluations VALUES(30445,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(30445,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(30445,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(30445,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(30445,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(30445,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(30445,36,'cp 140','c4f7 e8f7 f3g5 f7e8 g5e4 g8f6 b1c3 b8c6 e1g1 d7d5 e4f6 e7f6 d1h5 g7g6 h5d5 d8d5 c3d5 e8f7 c2c3 f8d6 d2d4 a8d8 h2h4 h8e8 g2g3 c6e7 d5e3 h7h5 g1g2 c7c5 d4d5 d6e5 f1d1 g6g5 h4g5 f6g5 a2a4 h5h4 a1a3 f7g6 a4a5 h4g3 f2g3 e7f5 e3f5 g6f5 d1f1 f5g6 a5b6');
INSERT INTO evaluations VALUES(32394,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(32394,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(32394,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(32394,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(32394,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(32394,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(32394,36,'cp 20','f1d3 f5d3 d1d3 c7c6 c2c4 d5c4 d3c4 g8f6 e1g1 f8e7 f1e1 e8g8 c4b3 d8c7 c1f4 c7f4 e1e7 b8d7 b1c3 b7b6 a1e1 a8e8 e7e3 c6c5 b3a4 e8e3 e1e3 c5d4 g2g3 d7c5 a4d4 f4f5 g1g2 f8c8 h2h3 c5e6 d4e5 f5e5 e3e5');
INSERT INTO evaluations VALUES(32392,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(32392,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32392,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(32392,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32392,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(32392,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(32392,36,'cp 13','f5d3 d1d3 c7c6 e1g1 f8d6 c1g5 g8e7 b1d2 e8g8 f1e1 b8a6 g5e7 d6e7 g2g3 f8e8 d2f1 e7f6 e1e8 d8e8 d3b3 a8b8 a1e1 e8d7 b3a3 b8a8 f3e5 d7c7 c2c3 a8e8 e5d3 e8e1 d3e1 f6e7 a3a4 c7b6 e1d3 g7g6 a4d1 a6c7 f1e3 h7h5 d1f3 c7e6');
INSERT INTO evaluations VALUES(32403,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(32403,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(32403,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(32403,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32403,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(32403,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(32403,36,'cp 18','d1d3 c7c6 c2c4 d5c4 d3c4 f8e7 e1g1 g8f6 f1e1 e8g8 c4b3 d8c7 c1f4 c7f4 b3b7 b8d7 e1e7 f8b8 b7a6 f4c1 e7e1 c1b2 b1d2 b2b5 a6a3 b5b4 a3c1 c6c5 a1b1 b4a5 d2c4 a5a6 b1b3 c5d4 b3a3 a6b5 f3d4 b5d5 e1d1 b8c8 a3c3 f6e4 c4e3 c8c3 e3d5 c3c1 d1c1');
INSERT INTO evaluations VALUES(32402,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(32402,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(32402,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(32402,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(32402,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(32402,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(32402,36,'cp 15','c7c6 e1g1 f8d6 c1g5 g8e7 f1e1 e8g8 b1d2 f7f6 g5h4 d8d7 h4g3 b8a6 g3d6 d7d6 h2h4 a6c7 h4h5 a8e8 d3b3 b7b6 d2e4 d6d7 e4g3 e7c8 h5h6 g7g6 c2c4 d5c4 b3c4 c7d5 g3e4 c8d6 e4d6 d7d6 e1e8 f8e8 a2a3');
INSERT INTO evaluations VALUES(42638,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(42638,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(42638,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(42638,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(42638,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(42638,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(42638,36,'cp -12','b8c6 c2c4 d7d6 b1c3 a7a5 f1e2 h7h6 e1g1 f6h7 c1e3 h7g5 c3b5 g5e6 d3d4 e5d4 f3d4 c6d4 b5d4 d8f6 d4e6 c8e6 e3c5 d6c5 d1a4 e6d7 a4a3 e8g8 a3c5 f6b2 e2f3 d7c6 f1b1 b2d2 c5e5 f8e8 e5c7 d2d4 c7g3 d4c4 e4e5 c6f3 g2f3 b7b6 b1b6');
INSERT INTO evaluations VALUES(46047,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(46047,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(46047,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(46047,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(46047,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(46047,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(46047,36,'cp 42','a7a6 f1e2 f8e7 f2f4 e6e5 f4e5 d6e5 d4f5 d8d1 a1d1 c8f5 e4f5 e7b4 e1g1 b4c3 b2c3 b8c6 g2g4 h7h6 e3b6 a8c8 h2h4 c6b8 d1d3 b8d7 b6a5 b7b6 a5b4 d7c5 d3e3 e8g8 e3e5 f8e8 e5e8 c8e8 e2f3 a6a5 b4c5 b6c5 f1d1 f6e4 c3c4 e4c3 d1d6 c3a2');
INSERT INTO evaluations VALUES(6046,30,'cp 292',NULL);
INSERT INTO evaluations VALUES(6046,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(6046,32,'cp 271',NULL);
INSERT INTO evaluations VALUES(6046,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(6046,34,'cp 263',NULL);
INSERT INTO evaluations VALUES(6046,35,'cp 304',NULL);
INSERT INTO evaluations VALUES(6046,36,'cp 309','d4c6 b7c6 e3c5 d6c5 d1d8 e8d8 b1d2 g8f6 e1c1 a7a5 d2b3 d8e7 b3c5 h7h5 h2h4 a5a4 f2f3 a8a5 c5d3 f6e8 d3b4 c8b7 a2a3 e8d6 b4a2 h8h6 a2c3 c6c5 f1e2 h6g6 h1g1 g6g3 c1d2 g7g6 d2e3 b7a6 e2a6 a5a6');
INSERT INTO evaluations VALUES(5790,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(5790,31,'cp 282',NULL);
INSERT INTO evaluations VALUES(5790,32,'cp 285',NULL);
INSERT INTO evaluations VALUES(5790,33,'cp 285',NULL);
INSERT INTO evaluations VALUES(5790,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(5790,35,'cp 288',NULL);
INSERT INTO evaluations VALUES(5790,36,'cp 281','b7c6 e3c5 d6c5 d1d8 e8d8 b1d2 a7a5 d2b3 a5a4 b3c5 a8a5 e1c1 d8e7 c5d3 g8f6 f2f3 g7g5 f1e2 h8g8 b2b4 a4b3 a2b3 c8a6 d3b4 a6b7 c1d2 g8a8 d2e3 c6c5 b4d3 f6d7 h2h4 g5g4 d3f4 d7e5 d1d2 g4f3 g2f3 b7c6 f4d3 e5d3 e2d3 a5a2 h1b1 c6d7');
INSERT INTO evaluations VALUES(4966,30,'cp 285',NULL);
INSERT INTO evaluations VALUES(4966,31,'cp 306',NULL);
INSERT INTO evaluations VALUES(4966,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(4966,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(4966,34,'cp 276',NULL);
INSERT INTO evaluations VALUES(4966,35,'cp 276',NULL);
INSERT INTO evaluations VALUES(4966,36,'cp 273','e3c5 d6c5 d1d8 e8d8 b1d2 a7a5 e1c1 d8e7 d2b3 a5a4 b3c5 a8a5 c5d3 g8f6 a2a3 h8d8 f2f3 c8a6 d3b4 d8d1 c1d1 a6b7 b4a2 h7h5 h2h4 f6e8 f1d3 e8d6 d1d2 c6c5 a2c3 b7c6 h1h3 e7f8 e4e5 d6c8 h3g3 c5c4 d3c4 a5e5 g3g5 e5g5 h4g5 f8e7 d2e3 c8d6');
INSERT INTO evaluations VALUES(4949,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(4949,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(4949,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(4949,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(4949,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(4949,35,'cp 275',NULL);
INSERT INTO evaluations VALUES(4949,36,'cp 276','d6c5 d1d8 e8d8 b1d2 a7a5 d2b3 a5a4 b3c5 d8e7 f2f3 g8f6 e1c1 a8a5 c5d3 g7g5 f1e2 h8g8 b2b4 a4b3 a2b3 c8a6 d3b4 a6b7 c1d2 g8a8 h2h4 g5g4 b4d3 c6c5 d3e5 e7e6 e5g4 f6g4 f3g4 a5a2 e2c4 e6e5 h1e1 a8g8 g4g5 b7e4 c4d3 f7f5 d2c3 e5f4 d3e4 f5e4');
INSERT INTO evaluations VALUES(4946,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(4946,31,'cp 277',NULL);
INSERT INTO evaluations VALUES(4946,32,'cp 272',NULL);
INSERT INTO evaluations VALUES(4946,33,'cp 277',NULL);
INSERT INTO evaluations VALUES(4946,34,'cp 280',NULL);
INSERT INTO evaluations VALUES(4946,35,'cp 282',NULL);
INSERT INTO evaluations VALUES(4946,36,'cp 275','d1d8 e8d8 b1d2 a7a5 d2b3 d8e7 e1c1 a5a4 b3c5 a8a5 c5d3 g8f6 f2f3 g7g5 d1e1 f6d7 h2h4 h7h5 e1e3 g5g4 d3b4 c8b7 e3c3 g4f3 g2f3 d7e5 b4d3 e5d3 f1d3 h8g8 c1d2 c6c5 d2e3 g8g2 b2b4 a4b3 a2b3 c7c6 d3f1 g2g8 e3f4 b7c8 c3c4 c8e6');
INSERT INTO evaluations VALUES(2240,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(2240,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(2240,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(2240,33,'cp 291',NULL);
INSERT INTO evaluations VALUES(2240,34,'cp 286',NULL);
INSERT INTO evaluations VALUES(2240,35,'cp 278',NULL);
INSERT INTO evaluations VALUES(2240,36,'cp 279','e8d8 b1d2 a7a5 e1c1 d8e7 d2b3 a5a4 b3c5 a8a5 c5d3 g8f6 a2a3 h8d8 f2f3 c8a6 d3b4 d8d1 c1d1 a6b7 d1d2 f6e8 b4a2 e8d6 f1d3 c6c5 a2c3 h7h5 h1e1 g7g6 g2g3 a5a8 f3f4 b7c6 d2c1 h5h4 g3g4 c5c4 d3f1 g6g5 f4g5 a8g8');
INSERT INTO evaluations VALUES(2340,30,'cp 290',NULL);
INSERT INTO evaluations VALUES(2340,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(2340,32,'cp 273',NULL);
INSERT INTO evaluations VALUES(2340,33,'cp 283',NULL);
INSERT INTO evaluations VALUES(2340,34,'cp 277',NULL);
INSERT INTO evaluations VALUES(2340,35,'cp 285',NULL);
INSERT INTO evaluations VALUES(2340,36,'cp 295','b1d2 d8e7 d2b3 a7a5 b3c5 a5a4 a2a3 a8a5 c5d3 g8f6 e1c1 c8a6 d3b4 a6b7 f2f3 h7h5 h2h4 f6d7 d1d3 h8h6 f1e2 h6g6 h1g1 g6d6 d3e3 c6c5 b4d3 g7g6 g1e1 b7a6 g2g3 d6d4 c2c3 d4d6 e4e5 d6b6 d3f4 a6e2');
INSERT INTO evaluations VALUES(6628,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(6628,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(6628,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(6628,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(6628,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(6628,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(6628,36,'cp 37','c3d5 d7d6 f1g2 a7a5 a2a3 b4c5 e2e3 c6e7 d2d4 e5d4 e3d4 c5a7 d5e7 g8e7 g1f3 e8g8 e1g1 c7c6 f1e1 h7h6 b2b3 f8e8 a1a2 d6d5 a2e2 a7b8 f3h4 c8g4 g2f3 g4f3 h4f3 d8d7 h2h4 b8c7 d1d3 e7g6 a3a4 e8e2 e1e2 d7g4 c1d2');
INSERT INTO evaluations VALUES(8979,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(8979,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(8979,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(8979,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(8979,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(8979,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(8979,36,'cp 52','h7h6 h2h3 e8g8 e1g1 d5c4 d3c4 a7a6 a2a3 b4d6 c1d2 d8e7 c4d3 e6e5 d4d5 c6a7 e3e4 c7c6 f1e1 c8d7 d5c6 d7c6 f3h4 c6d7 d3c4 f8c8 c4a2 d7e6 a2e6 e7e6 h4f5 d6f8 d1f3 c8d8 e1d1 a7b5 c3b5 a6b5 d2c3 e6c6');
INSERT INTO evaluations VALUES(10181,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(10181,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(10181,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(10181,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(10181,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(10181,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(10181,36,'cp 32','a2a3 a7a5 c1g5 h7h6 g5h4 c5e7 e2e3 d7d6 f1e2 e8g8 h2h3 c8f5 e1g1 f6d7 h4g3 d7c5 d3d4 e5d4 e3d4 c5e4 c3d5 e7f6 g3f4 f8e8 f4e3 c6e7 d5f6 e4f6 d4d5 f5d7 e2d3 e7f5 e3d2 f6e4 d3e4 e8e4 d1c2 e4e8 a1e1 e8e1 f1e1');
INSERT INTO evaluations VALUES(10407,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10407,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(10407,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10407,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(10407,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(10407,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(10407,36,'cp 13','b1c3 b4c5 f3e5 c6e5 d3d4 c5b4 d4e5 f6e4 d1d4 e4c3 b2c3 b4e7 f1e2 e8g8 e1g1 d7d6 c1f4 c8e6 f1d1 d8c8 e5d6 c7d6 d4e4 c8c7 c4c5 c7c5 e4b7 f8e8 c3c4 a8c8 f4e3 c5a5 b7a7 a5a7 e3a7 e6c4 e2f3 e7f6 a7d4 f6d4 d1d4 c4e6 a2a3');
INSERT INTO evaluations VALUES(11211,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11211,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11211,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11211,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11211,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11211,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11211,36,'cp 0','g7f6 b5a3 f6f5 f1d3 a8b8 d1h5 f5f4 g2g3 h8g8 c3d5 g8g6 g3f4 c8g4 h5h7 e5f4 a3c4 b7b5 c4d2 g6h6 h7g8 c6e5 h2h3 d8h4 h1g1 h6g6 g1g4 e5g4 d5c7 e8e7 c7d5');
INSERT INTO evaluations VALUES(11190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11190,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11190,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11190,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11190,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11190,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(11190,36,'cp 0','b5a3 f6f5 f1d3 h8g8 g2g3 f5e4 d3e4 c8e6 e1g1 g8g4 f2f3 g4g7 g1h1 f7f5 e4d5 d8e7 a3c4 e8c8 d5e6 e7e6 c3d5 c6e7 d1d3 e7d5 d3d5 e6d5 c4b6 c8c7 b6d5 c7d7 a1d1 d7e6 f1e1 d8c8 e1e2 g7g8 d5f4 e6f7 f4g2 f7e6');
INSERT INTO evaluations VALUES(11191,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11191,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11191,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11191,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11191,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11191,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11191,36,'cp 0','f6f5 f1d3 h8g8 g2g3 f5e4 d3e4 c8e6 e1g1 g8g4 f2f3 g4g7 g1h1 f7f5 e4d5 d8f6 a3c4 e8c8 d5e6 f6e6 c3d5 c6e7 c4e3 e7d5 d1d5 e6d5 e3d5 c8d7 a1d1 d7e6 f1e1 d8c8 c2c3 g7g8 d5f4 e6f7 f4d5');
INSERT INTO evaluations VALUES(11192,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11192,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11192,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11192,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11192,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11192,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11192,36,'cp 0','f1c4 d8g5 c4d5 g5g2 e4f5 g2h3 a3c4 c8f5 c4d6 f8d6 d5c6 e8e7 c3d5 e7f8 c6b7 a8b8 d5e3 d6c5 d1d5 c5e3 d5e5 e3f2 e1f2 h3h4 f2f1 h4c4 f1f2 c4c2 f2f3 c2d3 f3f4 d3d2 f4f5 d2f2 e5f4 f2c5 f4e5');
INSERT INTO evaluations VALUES(11366,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(11366,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(11366,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(11366,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(11366,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11366,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(11366,36,'cp 25','a2a3 f8d6 f1d3 d6f4 e3f4 b7b6 c3e2 c6e7 c2c3 a6a5 e1g1 c7c5 d3c2 d8c7 f1e1 e8g8 d1c1 e7f5 e2g3 f5d6 f3e5 h7h5 h2h4 c5d4 c3d4 g7g6 c2d3 c7c1 e1c1 f8d8 c1c6 c8a6 c6b6 a6d3 e5d3 d6c4 b6b7');
INSERT INTO evaluations VALUES(11379,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(11379,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(11379,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(11379,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(11379,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11379,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(11379,36,'cp 12','f8d6 f1d3 d6f4 e3f4 b7b6 c3e2 e8g8 h2h4 c6e7 h4h5 c7c5 c2c3 d8c7 f3e5 f6e4 d1c2 f7f6 e5f3 e7f5 d3e4 d5e4 c2e4 c8d7 e1g1 a8e8 f1e1 d7c6 e4d3 c7f7 d4c5 b6c5 e2g3 f5g3 f2g3 f7h5 e1e3 a6a5 a1e1 h5g4 f3d2 c6d5 c3c4 e8b8');
INSERT INTO evaluations VALUES(18079,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(18079,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(18079,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(18079,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(18079,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(18079,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(18079,36,'cp 145','c8b7 a2a4 f8d6 d2d3 g8e7 a4b5 a6b5 d1h5 e8g8 a1a8 b7a8 h5b5 c7c5 b1a3 a8c6 b5a6 h7h6 g2g3 d6c7 c1d2 d8a8 a6a8 f8a8 b3c4 a8b8 f1b1 d7d5 e4d5 e7d5 c4b3 g8f8 a3c4 b8a8 b1e1 d5f6 f2f4 c6d5 h2h3 d5e6 g3g4 e6d5');
INSERT INTO evaluations VALUES(25606,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(25606,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(25606,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(25606,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(25606,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(25606,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(25606,36,'cp 156','a2a4 f8d6 d2d3 g8e7 a4b5 a6b5 a1a8 b7a8 d1h5 e8g8 h5b5 c7c5 g2g3 a8c6 b5a6 d6c7 b1a3 d8a8 a6a8 f8a8 c1g5 e7c8 g5d2 g8f8 f1b1 h7h6 g1f1 c8e7 b3c4 d7d5 e4d5 e7d5 c4b5 c6b5 a3b5 a8a2 b5a3 g7g5 a3c4 f7f5 b1e1 f8f7');
INSERT INTO evaluations VALUES(18828,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(18828,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(18828,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(18828,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(18828,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(18828,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(18828,36,'cp -18','d7d5 e1g1 g8f6 e4d5 e6d5 d3d4 c5d4 f3d4 f8d6 c1e3 e8g8 b1d2 a7a6 c2c3 h7h6 f1e1 f8e8 d2f1 c6a5 h2h3 c8d7 d4f3 d7f5 e2d3 f6e4 f3d4 f5d7 d1c2 a8c8 f1d2 e4d2 c2d2 a5c4 d3c4 d5c4 a1d1');
INSERT INTO evaluations VALUES(20296,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(20296,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20296,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20296,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20296,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(20296,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(20296,36,'cp -12','e1g1 g8f6 e4d5 e6d5 d3d4 c5d4 f3d4 f8d6 c2c4 e8g8 b1c3 d5c4 d4c6 b7c6 e2c4 c8g4 d1c2 g4h5 h2h3 h5g6 c2a4 d8c7 c4b3 a7a5 b3c2 a8b8 c2g6 h7g6 a4c2 f8d8 c1e3 d6h2 g1h1 h2f4 a1e1 f4e3 e1e3 c7b6 c3a4 b6b4');
INSERT INTO evaluations VALUES(32496,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32496,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(32496,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(32496,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(32496,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32496,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32496,36,'cp 0','b5d7 d8d7 e4d5 e6d5 d2d4 f8d6 e1g1 g8e7 b1c3 e8g8 d1d3 c7c6 c3e2 b8a6 c1f4 e7f5 f4d2 a6b4 d3c3 b4a6');
INSERT INTO evaluations VALUES(32463,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32463,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32463,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32463,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32463,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(32463,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(32463,36,'cp -12','d8d7 e4d5 e6d5 e1g1 b8c6 d2d4 f7f6 c2c3 f8d6 b2b3 g8e7 c3c4 b7b6 b1c3 e8g8 c3b5 a8e8 h2h3 e7g6 a1b1 d6f4 f1e1 e8e4 b5c3 e4e1 f3e1 d5c4 e1f3 f8e8 b3c4 c6a5 c4c5 b6c5');
INSERT INTO evaluations VALUES(32952,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32952,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32952,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32952,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(32952,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(32952,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(32952,36,'cp 0','e4d5 e6d5 e1g1 b8c6 d2d4 f7f6 c2c3 f8d6 b2b3 g8e7 c3c4 b7b6 b1c3 e8g8 c1b2 f8e8 f1e1 a8d8 h2h3 d5c4 b3c4 e7g6 e1e8 d7e8 c3e2 e8e4 e2c3');
INSERT INTO evaluations VALUES(35144,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(35144,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(35144,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(35144,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(35144,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(35144,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(35144,36,'cp 65','b1c3 f8b4 g1e2 g8e7 e1g1 e8g8 a2a3 b4c3 e2c3 d7d6 f2f4 b8c6 f4f5 c8d7 c1e3 c6e5 c4a2 g8h8 a1c1 f6h4 d1d2 f7f6 b2b4 a7a6 h2h3 f8d8 e3f4 d6d5 f4e5 f6e5 c3d5 e7d5 e4d5 a8c8 d2e3 d7b5 f1e1 d8e8 c1c5 h4d4 e3d4 e5d4');
INSERT INTO evaluations VALUES(35143,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(35143,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(35143,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(35143,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(35143,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(35143,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(35143,36,'cp 63','f8b4 g1e2 g8e7 e1g1 b8c6 a2a3 b4c3 e2c3 e8g8 c3b5 d7d5 c4d5 c8e6 d5e6 f7e6 c1e3 a8c8 b5c3 b7b6 d1e2 c8d8 a1d1 c6d4 e3d4 d8d4 d1d4 f6d4 c3b5 d4e5 b5a7 e7g6 g2g3 e5g5 a7b5 g6f4 e2c4 c7c5 b5d6 g5g4');
INSERT INTO evaluations VALUES(33273,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(33273,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(33273,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(33273,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(33273,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(33273,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(33273,36,'cp 61','g1e2 g8e7 e1g1 e8g8 a2a3 b4c3 e2c3 b8c6 c3b5 d7d5 c4d5 c8e6 d5e6 f7e6 c1e3 a8c8 b5c3 h7h6 d1b3 c6a5 b3a2 c7c5 f2f4 c5c4 e3a7 a5b3 a1d1 c8d8 a7e3 e7c6 a2b1 c6d4 e4e5 f6e7 c3e4 b7b5 e4d6 d4f5 d6f5 e6f5 d1d8 e7d8');
INSERT INTO evaluations VALUES(33272,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(33272,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(33272,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33272,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(33272,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(33272,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(33272,36,'cp 50','g8e7 e1g1 e8g8 a2a3 b4c3 e2c3 b8c6 f2f4 d7d5 c4d5 c8e6 d5e6 f7e6 c1e3 a8d8 d1a4 e7c8 a1d1 c8b6 a4b3 d8d1 f1d1 h7h6 c3e2 g8h7 h2h3 e6e5 f4f5 f6h4 e3c5 f8d8 d1d8 c6d8 e2g3 h4f6 g1h2 f6f7 b3d3 d8c6 c5b6 a7b6 d3b5 f7d7');
INSERT INTO evaluations VALUES(36747,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36747,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36747,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36747,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(36747,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36747,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(36747,36,'cp 23','f1e2 c7c5 b1d2 b8d7 h2h3 b7b6 a2a4 c8b7 e1g1 a7a6 d1b3 a8c8 f1d1 e7e6 c3c4 a6a5 c4d5 b7d5 b3a3 c5d4 e3d4 f8e8 e2b5 g7f8 a3d3 d5c6 d2c4 c6b5 a4b5 f6d5 f4g3 f8b4 f3e5 d7e5');
INSERT INTO evaluations VALUES(39433,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(39433,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(39433,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(39433,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(39433,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(39433,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(39433,36,'cp -89','f4g3 e1g1 d7d5 e4d5 g3h2 g1h1 c8h3 f3h4 d8h4 d1e1 h4e7 e1e7 g8e7 f1e1 e8d8 d2d3 b8d7 c1f4 e7g6 f4h2 f7f6 d3d4 g6h4 b1d2 d7b6 c4d3 b6d5 d3e4 d5e7 d2b3 a8b8 b3c5 b7b6 c5a6 b8c8 e4b7 e7f5 c2c4 f5d4 b7c8 h3c8 a6c7 h4f3 a1d1 f3e1 d1d4 c8d7');
INSERT INTO evaluations VALUES(39423,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(39423,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(39423,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(39423,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(39423,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(39423,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(39423,36,'cp -68','e1g1 d7d5 e4d5 g3h2 g1h1 c8h3 f3h4 d8h4 d1e1 h4e7 e1e7 g8e7 f1e1 b8d7 d2d3 e8d8 c1f4 e7g6 f4h2 d7f6 d5d6 c7d6 h2d6 h3d7 c4f7 a8c8 b1c3 c8c6 d6g3 g6e7 g3f4 h8f8 f7b3 c6c5 e1e5 c5e5 f4e5 f6g4 e5g3 e7f5 g3f4 h7h6 a1f1 f5d4');
INSERT INTO evaluations VALUES(39422,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(39422,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(39422,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(39422,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(39422,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(39422,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(39422,36,'cp -71','g3h2 g1h1 d7d5 e4d5 c8h3 f3h4 d8h4 d1e1 h4e7 e1e7 g8e7 f1e1 b8d7 b2b3 e8d8 c1a3 e7g6 h1h2 h3g4 d2d4 f7f5 b1c3 a8c8 c4d3 d7f6 h2g1 a7a6 c3a4 f6d5 a4c5 c8b8 c5e6 d8c8 d3c4 d5f6 e6g7 f5f4 e1e6 b7b5 e6f6 b5c4 a1e1 h8g8 g7e6 b8b6 b3c4 g4e6 e1e6 b6e6');
INSERT INTO evaluations VALUES(39421,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(39421,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(39421,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(39421,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(39421,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(39421,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(39421,36,'cp -71','g1h1 d7d5 e4d5 g8h6 f3h4 d8h4 d1e1 h4e1 f1e1 e8d8 d2d4 h6g4 b1c3 c8f5 e1f1 f5g6 c4e2 g4f6 c1g5 b8d7 a1c1 h8e8 h1h2 a7a6 e2f3 h7h6 g5h4 a8b8 a2a4 b7b5 a4b5 a6b5 d5d6 c7d6 c3d5 g6e4 h4f6 d7f6 d5f6 g7f6 h2g3 b8c8 f3e4 e8e4 f1f6 d8e7 f6h6 e4d4 h6h7 d4d3 c2d3 c8c1');
INSERT INTO evaluations VALUES(39420,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(39420,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(39420,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(39420,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(39420,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(39420,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(39420,36,'cp -52','d7d5 e4d5 g8h6 f3h4 d8h4 d1e1 h4e1 f1e1 e8d8 b1c3 h6g4 d2d4 c8f5 c4e2 h8e8 c1f4 h7h5 e2g4 e8e1 a1e1 f5g4 h1h2 d8d7 c3b5 b8a6 d5d6 c7d6 f4d6 a8e8 e1e8 d7e8 c2c4 e8d7 d6f8 g7g5 b5d6 f7f5 d6b7 g4e2 b7d6 f5f4 d6e4 g5g4 f8h6 e2c4 h6f4 d7e6 e4c5 a6c5 d4c5 c4a2 h2g3 e6f5 f4d6 a2d5');
INSERT INTO evaluations VALUES(51535,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51535,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(51535,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(51535,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(51535,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(51535,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(51535,36,'cp 76','d7d5 e2e3 a7a6 b2b3 e7e6 c2c4 b5c4 b3c4 g8f6 c4c5 f8e7 b1c3 a6a5 a1b1 e8g8 d1c2 c7c6 c1d2 b8d7 c3a4 f6e4 f1d3 c8a6 d3a6 a8a6 e1g1 d8c7 b1b2 e6e5 f1b1 e4d2 f3d2 a6a7 a4b6 e5d4 e3d4 e7f6 d2f3 f8d8 b6d7');
INSERT INTO evaluations VALUES(51315,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(51315,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(51315,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(51315,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(51315,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(51315,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(51315,36,'cp 82','e2e3 a7a6 b2b3 e7e6 c2c4 b5c4 b3c4 g8f6 c4c5 a6a5 f1d3 f8e7 b1c3 e8g8 e1g1 c8a6 d3a6 a8a6 f3e5 c7c6 c1d2 b8d7 f2f4 d8c7 h2h3 a6a8 d2e1 h7h6 d1c2 d7e5 f4e5 f6h7 e1g3 f7f5 c2d3 f8b8 g3f4 b8b4 f1c1');
INSERT INTO evaluations VALUES(51595,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(51595,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(51595,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(51595,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(51595,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(51595,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(51595,36,'cp 74','e7e6 d2d4 c5d4 f3d4 a7a6 c1e3 d8c7 d1d2 g8f6 f2f3 h7h5 e1c1 c8b7 d4b3 b8c6 c1b1 f8e7 d2f2 b6b5 f2g3 c7g3 h2g3 e8c8 e3b6 d8e8 b6f2 e8d8 a2a4 b5b4 c3e2 d7d5 e4d5 d8d5 d1d5 f6d5 e2c1 a6a5 f1b5 d5c7 b5c4 c7d5 c1d3 c8c7');
INSERT INTO evaluations VALUES(51551,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(51551,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(51551,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(51551,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(51551,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(51551,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(51551,36,'cp 83','d2d4 c5d4 f3d4 a7a6 c1e3 c8b7 f2f3 h7h5 d1d2 d8c7 e1c1 g8f6 c1b1 f8e7 f1e2 b8c6 d4c6 d7c6 e3g5 b6b5 e4e5 a8d8 d2e3 f6d5 c3d5 d8d5 g5e7 c7e7 d1d5 e6d5 f3f4 c6c5 c2c3 g7g6 h1e1 e8g8 f4f5 d5d4 c3d4 c5d4 e3d4 b7g2');
INSERT INTO evaluations VALUES(51547,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(51547,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(51547,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(51547,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(51547,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(51547,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(51547,36,'cp 87','c5d4 f3d4 a7a6 c1e3 c8b7 f2f3 h7h5 d1d2 d8c7 e1c1 g8f6 d4b3 b8c6 c1b1 f8e7 d2f2 b6b5 f2g3 c7g3 h2g3 e8c8 e3b6 d8g8 a2a4 b5b4 c3e2 d7d5 e4d5 f6d5 b6f2 g7g5 e2d4 c6d4 f2d4 e7f6 b3a5 f6d4 d1d4 g5g4 f1e2 c8c7 a5b7 c7b7');
INSERT INTO evaluations VALUES(9766,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(9766,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9766,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(9766,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9766,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9766,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(9766,36,'cp 13','a7a6 e1g1 d7d6 c4b3 e8g8 f1e1 c8e6 d2f1 e6b3 d1b3 f8e8 c1e3 c5e3 f1e3 d8d7 e3f5 b7b6 a1d1 b6b5 f5g3 d7e6 b3e6 f7e6 d3d4 a6a5 d4e5 c6e5 f3e5 d6e5 f2f3 b5b4 g3f1 b4c3 b2c3 e8d8 f1e3 d8d1');
INSERT INTO evaluations VALUES(15531,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(15531,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(15531,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(15531,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(15531,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(15531,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(15531,36,'cp 27','c8e6 c3d5 c6e7 d5f6 g7f6 d2d4 e6c4 g2b7 a8b8 b7g2 c7c5 d4c5 d6c5 d1d8 b8d8 c1d2 d8b8 b2b3 c4e6 g1e2 f8g7 e1g1 e8g8 f1d1 f6f5 e2c3 c5c4 b3c4 e6c4 e3e4 f5e4 g2e4 c4e6 a1c1 e7f5 e4d5 f8e8 d5e6 e8e6 c3e4 f5d4 c1c7');
INSERT INTO evaluations VALUES(16324,30,'cp -373',NULL);
INSERT INTO evaluations VALUES(16324,31,'cp -362',NULL);
INSERT INTO evaluations VALUES(16324,32,'cp -359',NULL);
INSERT INTO evaluations VALUES(16324,33,'cp -383',NULL);
INSERT INTO evaluations VALUES(16324,34,'cp -376',NULL);
INSERT INTO evaluations VALUES(16324,35,'cp -393',NULL);
INSERT INTO evaluations VALUES(16324,36,'cp -407','g7g6 f5g5 h7h6 g5g3 f6e4 g3d3 e4d6 b5c6 d7c6 d3e2 f8g7 b1c3 e8g8 f2f3 f8e8 d2d3 b7b5 e2f2 b5b4 c3e4 f7f5 e4d6 c7d6 g1e2 a7a5 e1g1 c8a6 f1d1 c6c5 a2a3 d8d7 c1e3 f5f4 e3c1 g6g5 a3b4 a5b4 g1h1 d6d5 f2c5 e5e4 d3e4 a6e2 a1a8 e8a8');
INSERT INTO evaluations VALUES(20685,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(20685,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(20685,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(20685,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(20685,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(20685,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(20685,36,'cp -44','d6d5 e4d5 e6d5 d1e2 f8e7 c4b5 c8g4 d2d3 a8c8 c1e3 g8f6 h2h3 g4f3 e2f3 d8b6 b5c6 b6c6 e3d2 c6b6 c2c4 b6b2 d2c3 b2b6 c4d5 b6d6 b1d2 d6d5 d2e4 c8d8 c3f6 e7f6 e4f6 g7f6 f3g3 e8f8 g3h4 h7h6 h4f6 d5d4 f6d4 d8d4 f1c1');
INSERT INTO evaluations VALUES(21583,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(21583,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(21583,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(21583,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(21583,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(21583,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(21583,36,'cp -8','f8b4 b1c3 g8f6 c1g5 c8g4 f1e2 d5c4 e1g1 e8g8 h2h3 g4f3 e2f3 c6d4 f3b7 a8b8 b7a6 h7h6 g5f6 d8f6 a6c4 b4a5 d1h5 f6c6 h5a5 c6c4 a1b1 b8b6 a5d5 d4e2 g1h1 c4d5 c3d5 b6d6 d5e3 e2f4 f1d1 f8d8 b2b3 d6d1 b1d1 d8d1 e3d1 g8f8 h1h2 f4d3 h2g3 f8e7 d1e3 e7e6');
INSERT INTO evaluations VALUES(21699,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(21699,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(21699,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(21699,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(21699,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(21699,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(21699,36,'cp 75','d4e5 d5d4 c3d5 g8e7 c1g5 c8e6 d5f4 d8d7 f1b5 a7a6 g5e7 f8e7 f4e6 f7e6 b5c6 d7c6 d1d3 e8c8 g1f3 g7g5 e1c1 g5g4 f3e1 h8f8 d3e2 h7h5 h2h3 c6b5 e2b5 a6b5 h3g4 h5g4 e1d3 c7c5 h1h7 e7g5 c1b1 c5c4 d3c5 f8f2 c5e6');
INSERT INTO evaluations VALUES(22794,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(22794,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(22794,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(22794,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(22794,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(22794,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(22794,36,'cp -6','g1f3 g8f6 e1g1 f8c5 b1c3 e8g8 h2h3 a7a5 d3c4 d7d6 d2d3 c8e6 c4b3 c6d4 f3d4 c5d4 c3e2 d4b6 e2g3 f6d7 b3e6 f7e6 d1g4 d8f6 c1e3 b6e3 f2e3 f6h6 g1h2 b7b5 f1f8 a8f8 a1f1 f8f1 g3f1 g8f7 h2g1 a5a4 a2a3 d7c5 g4d1 f7e8');
INSERT INTO evaluations VALUES(23910,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(23910,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(23910,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(23910,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(23910,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(23910,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(23910,36,'cp 108','d2d4 c5b6 a2a4 a7a6 b5c6 d7c6 d4e5 d6e5 f3e5 d8d1 f1d1 c6e4 e5c4 g8f6 c4b6 c7b6 c1e3 e8g8 f2f3 e4b1 a1b1 b6b5 a4b5 a6b5 g1f2 f8c8 g2g4 b5b4 c3b4 c8c2 d1d2 c2d2 e3d2 h7h5 g4h5 f6h5 b1c1 a8a2 d2c3 f7f6 f2e3 g8f7');
INSERT INTO evaluations VALUES(23907,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(23907,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(23907,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(23907,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(23907,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(23907,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(23907,36,'cp 100','c5b6 a2a4 a7a6 b5c6 d7c6 d4e5 d6e5 f3e5 d8d1 f1d1 c6e4 e5c4 g8f6 c4b6 c7b6 c1e3 e8g8 f2f3 e4b1 d1b1 b6b5 a4b5 a6b5 e3c5 f8c8 a1a8 c8a8 b1d1 h7h5 c5e7 a8a2 d1d2 g8h7 e7f6 g7f6 g1f2 b5b4 c3c4 h7g6 f2e3 g6f5 g2g3 f5e5 e3d3 a2a1 d3c2 a1f1 d2d7 f1f3 d7b7 f3f2 c2b3 f2h2 b7b4 e5d4');
INSERT INTO evaluations VALUES(24213,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(24213,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(24213,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24213,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(24213,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24213,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24213,36,'cp 28','c6d4 b5d7 f6d7 d1d4 e8g8 c1f4 e7f6 d4e3 d7e5 c3d5 e5g6 e3b3 g6f4 d5f4 a8b8 a1d1 f8e8 g2g3 b7b5 g1g2 c7c6 c2c3 a7a5 b3c2 d8d7 b2b3 f6e5 f4d3 d7e6 h2h4 h7h5 c3c4 b5c4 b3c4 g7g6 c4c5 d6c5 d3c5 b8b2');
INSERT INTO evaluations VALUES(32100,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(32100,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(32100,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(32100,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(32100,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(32100,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(32100,36,'cp 113','d1f3 g8f6 a2a4 a7a6 c1e3 b8c6 b1c3 f8e7 e1g1 e8g8 a4a5 d8d7 c4b3 f6g4 c3d5 e7d8 g2g3 c6d4 e3d4 e5d4 d5f4 g4e5 f3e2 g8h8 g1g2 e5g6 f4g6 f7g6 b3a4 c7c6 f2f4 d8f6 h2h3');
INSERT INTO evaluations VALUES(32099,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(32099,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(32099,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(32099,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(32099,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(32099,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(32099,36,'cp 92','g8f6 e1g1 f8e7 a2a4 e8g8 b1c3 b8c6 c1e3 c6b4 c4b3 f6d7 a4a5 a7a6 f3g3 g8h8 c3e2 e7h4 g3h3 d7c5 e3c5 d6c5 c2c3 b4c6 f2f4 d8e7 b3d5 e5f4 e2f4 h4g5 a1e1 c6e5 d3d4 c5d4 c3d4 e5c6 d5c6 b7c6 e4e5 a8d8');
INSERT INTO evaluations VALUES(29920,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(29920,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(29920,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(29920,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(29920,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(29920,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(29920,36,'cp 122','c1e3 b8c6 a2a4 f8e7 b1c3 e8g8 e1g1 c6b4 c4b3 f6d7 a4a5 d7c5 e3c5 d6c5 f3g3 e7f6 c3d1 a8b8 d1e3 b4c6 c2c3 g8h8 e3g4 d8d6 h2h4 g7g6 h4h5 f6g5 g4e3 g5e3 f2e3 f7f5 g3h4 c6e7 b3c4 b7b5 a5b6 a7b6 a1a6 f5e4');
INSERT INTO evaluations VALUES(36893,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36893,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36893,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(36893,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(36893,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(36893,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36893,36,'cp 25','e1g1 b8c6 c1g5 d8d1 f1d1 c8g4 h2h3 g4f3 e2f3 c6d4 c3d5 f6d5 c4d5 c7c6 d5c6 b7c6 d1c1 a7a5 a1b1 f8b8 f3d1 b8b6 g5d2 g7f8 g1f1 h7h5 c1c4 f8a3 c4a4 a3b2 d2a5 b6b7 a5d2 a8d8 d2e3 d8b8 a4c4 g8g7 d1a4 b7b4 c4b4 b8b4 a4d1 b2a3');
INSERT INTO evaluations VALUES(47558,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(47558,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(47558,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(47558,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(47558,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47558,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(47558,36,'cp 52','e8f7 d1f3 f7e6 e1g1 c7c6 b1c3 e6d7 d2d4 d5c3 f3c3 d7c7 d4e5 d8e7 e5e6 e7f6 f1d1 f8d6 c3h3 h8f8 c1e3 f6f5 g2g4 f5f3 h3h5 b8d7 e6d7 c8d7 d1d6 c7d6 e3c5 d6c7 h5e5 c7c8 c5f8 f3g4 e5g3 g4c4 a1e1 c4c2 f8d6 b7b6 g3g7 c2g6 g7g6 h7g6 e1e7 a7a5 e7g7 d7f5 g7c7 c8d8 c7c6 b6b5 c6b6 b5b4 g1g2 f5b1 d6e5 b1a2 b6g6 a2f7 g6d6 d8e8 g2g3 a5a4 d6b6 a8a5');
INSERT INTO evaluations VALUES(35497,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(35497,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(35497,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(35497,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(35497,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(35497,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(35497,36,'cp 59','d1f3 f7e6 b1c3 c7c6 e1g1 e6d7 d2d4 d5c3 f3c3 d7c7 d4e5 d8e7 e5e6 e7f6 f1d1 f8d6 c3h3 d6e5 f2f4 h8d8 c1e3 e5b2 d1d8 f6d8 f4f5 b7b6 a1e1 c7b7 e3f2 d8d2 h3h5 b8a6 e1d1 d2b4 c4a6 b7a6 h5e2 a6b7 e6e7 c8f5 d1d8 b2f6 d8a8 f6e7 a8g8 e7f6 e2e8 b4b1 f2e1 b1c2');
INSERT INTO evaluations VALUES(35496,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(35496,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(35496,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(35496,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(35496,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(35496,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(35496,36,'cp 71','f7e6 e1g1 c7c6 b1c3 e6d7 d2d4 d5c3 f3c3 d7c7 d4e5 d8e7 e5e6 e7f6 f1d1 f8d6 c3h3 h8f8 c1e3 d6e5 e3c5 f8h8 d1d3 b8a6 c4a6 c8e6 h3h5 f6f5 h5f5 e6f5 d3f3 f5d7 f3e3 e5f6 a6c4 f6b2 a1d1 a8e8 e3b3 b2e5 b3b7 c7b7 d1d7 b7b8 c5a7 b8a8 a7e3 e8d8 d7a7 a8b8 c4d3 e5d4 e3d4 d8d4 a7g7 d4a4 g1f1 a4a2 g7h7 h8h7 d3h7 a2a1 f1e2');
INSERT INTO evaluations VALUES(35503,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(35503,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(35503,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(35503,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(35503,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(35503,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(35503,36,'cp 82','e1g1 c7c6 b1c3 e6d7 d2d4 d5c3 f3c3 d7c7 d4e5 d8e7 e5e6 e7f6 f1d1 f8d6 c3h3 h8f8 c1e3 d6e5 e3c5 f8h8 d1d3 b8a6 c4a6 c8e6 h3h5 a8e8 a1e1 e5f4 g2g3 f6f5 h5f5 e6f5 e1e8 h8e8 g3f4 b7a6 c5d6 c7d7 d3d2 e8e1 g1g2 d7e8 d6c5 e8f7 c5a7 e1a1 a7d4 a1a2');
INSERT INTO evaluations VALUES(52031,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(52031,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(52031,32,'cp -159',NULL);
INSERT INTO evaluations VALUES(52031,33,'cp -168',NULL);
INSERT INTO evaluations VALUES(52031,34,'cp -159',NULL);
INSERT INTO evaluations VALUES(52031,35,'cp -148',NULL);
INSERT INTO evaluations VALUES(52031,36,'cp -141','b8c6 d4e3 d8a5 b1c3 a5e5 e3e5 c6e5 c3b5 e8d8 c1f4 f7f6 f4e5 f6e5 g1f3 a7a6 f3e5 a6b5 e5f7 d8e8 f7h8 g7g6 h8g6 h7g6 f1b5 g8f6 a2a3 e8e7 e1c1 e6e5 b5d3 g6g5 c1b1 d7d6 h1e1 c8e6 f2f3 e7f7 h2h3 f8e7 c2c3 f6d5 d3e4 d5f4 e4b7 a8b8');
INSERT INTO evaluations VALUES(18870,30,'cp -147',NULL);
INSERT INTO evaluations VALUES(18870,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(18870,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(18870,33,'cp -148',NULL);
INSERT INTO evaluations VALUES(18870,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(18870,35,'cp -143',NULL);
INSERT INTO evaluations VALUES(18870,36,'cp -152','d4e3 d8a5 b1c3 a5e5 e3e5 c6e5 c3b5 e8d8 c1f4 f7f6 f4e5 f6e5 g1f3 a7a6 f3e5 a6b5 e5f7 d8e8 f7h8 g7g6 h8g6 h7g6 f1b5 g8f6 a2a4 e8f7 f2f3 d7d5 e1d1 f8d6 c2c3 e6e5 a4a5 c8e6 h1e1 e5e4 h2h3 d6g3 e1f1 f7e7 d1e2 g6g5 f3e4 f6e4 b5d3 a8h8');
INSERT INTO evaluations VALUES(18874,30,'cp -158',NULL);
INSERT INTO evaluations VALUES(18874,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(18874,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(18874,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(18874,34,'cp -156',NULL);
INSERT INTO evaluations VALUES(18874,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(18874,36,'cp -148','d8a5 b1c3 a5e5 e3e5 c6e5 c3b5 e8d8 c1f4 f7f6 f4e5 f6e5 g1f3 a7a6 f3e5 a6b5 e5f7 d8e8 f7h8 g7g6 h8g6 h7g6 f1b5 g8f6 a2a3 e8e7 e1c1 e6e5 b5d3 g6g5 c1b1 d7d6 h2h3 c8e6 h1e1 a8b8 f2f3 f6d5 d1d2 d5f4 d3e4 d6d5 e4d5 f4d5 e1e5 d5c7 e5c5 b8c8 c5g5 f8h6');
INSERT INTO evaluations VALUES(52516,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(52516,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(52516,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(52516,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(52516,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(52516,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(52516,36,'cp -87','b8c6 d4a4 e7e5 f4g3 d7d6 e2e3 f8e7 g1f3 g8f6 b1d2 f6d7 c2c3 e8g8 h2h3 f7f5 a4c2 g8h8 f1e2 d7f6 a1d1 c8d7 e1g1 d8c7 c2b3 c6d8 f1e1 d8e6 b3c2 e5e4 f3d4 e6d4 e3d4 a8c8');
INSERT INTO evaluations VALUES(19548,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(19548,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(19548,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(19548,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(19548,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(19548,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(19548,36,'cp -81','d4a4 e7e5 f4g3 b7b5 a4b3 a7a6 a2a4 b5b4 e2e4 g8f6 b1d2 f8c5 b3d3 d7d6 d2b3 e8g8 b3c5 d6c5 d3d8 f8d8 f2f3 c8e6 g1h3 h7h6 h3f2 f6h5 f2d3 c5c4 d3c5 h5g3 h2g3 c6d4 a1c1 d8c8 c5e6 d4e6 f1e2 c4c3 b2b3 c8d8 c1d1 e6d4 e2d3 a6a5 e1f2');
INSERT INTO evaluations VALUES(52824,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52824,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52824,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52824,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52824,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52824,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52824,36,'cp 0','e2e3 g8f6 c2c4 e7e6 b2b3 f8e7 c1b2 e8g8 b1c3 c6c5 f3e5 f6d7 e5f3');
INSERT INTO evaluations VALUES(52823,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52823,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(52823,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(52823,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52823,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(52823,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(52823,36,'cp -11','g8f6 c2c4 c8g4 b2b3 e7e6 f1e2 g4f5 e1g1 h7h6 b1c3 f8e7 c1b2 b8d7 d2d4 e8g8 e2d3 e7b4 a1c1 b4c3 b2c3 f5d3 d1d3 f6e4 h2h3 f8e8 c3e1 a7a5 a2a4 c6c5 d3c2 c5d4 e3d4 d7f6 c4c5 b7b6 c5c6');
INSERT INTO evaluations VALUES(52982,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(52982,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(52982,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(52982,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(52982,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(52982,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(52982,36,'cp -39','d8d6 b1c3 b8c6 g1f3 g8f6 d2d4 c5d4 c3b5 d6b8 b5d4 e7e5 d4c6 b7c6 f1c4 e5e4 d1e2 f8c5 f3g5 e8g8 g5e4 f6e4 e2e4 c8f5 e4f5 b8b4 c1d2 b4c4 f5d3 a8e8 e1f1 c4e6 d2c3 e8d8 d3f3 f8e8 h2h4 e6c4 f1g1 e8e2 a1f1 e2c2 h1h3 c5d4 f3f4 h7h6 c3d4 d8d4 f4b8 g8h7');
INSERT INTO evaluations VALUES(34347,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(34347,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(34347,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(34347,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(34347,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(34347,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(34347,36,'cp -37','g1f3 b8c6 f1e2 g8f6 e1g1 e7e5 d2d3 f8e7 b1c3 f6d5 c3d5 d6d5 f1e1 e8g8 f3d2 c6d4 d2c4 b7b5 c4e3 d5b7 a2a4 b5b4 e3c4 c8e6 c1e3 d4e2 d1e2 b7c7 e3d2 f7f6 f2f4 e6c4 d3c4 e7d6 f4f5 e5e4 e2e4 d6h2 g1h1 h2g3');
INSERT INTO evaluations VALUES(54849,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(54849,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(54849,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(54849,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(54849,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(54849,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(54849,36,'cp 0','f8g7 e2e3 g8f6 f1e2 e8g8 g1f3 c7c5 e1g1 b7b6 c3b5 b8c6 b5c7 a8b8 c7b5');
INSERT INTO evaluations VALUES(57721,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(57721,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(57721,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(57721,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(57721,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(57721,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(57721,36,'cp 39','g8f6 g2g3 e7e6 f1g2 d7d5 e1g1 f8e7 c2c4 e8g8 d1c2 a7a5 b2b3 a5a4 c1a3 b8c6 b3a4 c8d7 f1c1 c6b4 c2b3 d7a4 b3b2 c7c5 d4c5 b4c6 b1c3 d5d4 c3a4 a8a4 b2b7 d8a8 b7a8 f8a8 f3d2 a8c8 a3b2 e7c5 c1b1');
INSERT INTO evaluations VALUES(50296,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(50296,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(50296,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(50296,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50296,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(50296,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(50296,36,'cp 46','g2g3 d7d5 c2c4 e7e6 f1g2 f8e7 e1g1 e8g8 d1c2 a7a5 f1d1 c7c6 b1d2 c6c5 d4c5 b8a6 a2a3 a6c5 b2b3 b7b6 c4d5 e6d5 c1b2 c8a6 f3d4 a8c8 c2f5 d8d7 f5d7 c5d7 d4f5 f8e8 f5e7 e8e7 e2e3 d7c5');
INSERT INTO evaluations VALUES(3645,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3645,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3645,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3645,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3645,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3645,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(3645,36,'cp -9','a7a5 c2c3 f6d5 c4d5 a5a4 a2a3 c6e7 d5a2 c8e6 a2e6 f7e6 c1e3 c5e3 f2e3 e7g6 d1e2 d8e7 f3d2 g8h7 e2h5 g6h4 h5g4 b7b5 f1f8 a8f8 a1f1 f8f1 g1f1 h4g6 f1e2 g6f8 e2d1 f8d7 g4f3 d7c5 f3e2 e7h4 c3c4 c7c6 c4b5 c6b5');
INSERT INTO evaluations VALUES(15433,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(15433,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(15433,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(15433,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(15433,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(15433,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(15433,36,'cp 46','f8e7 f1e1 e8g8 c2c3 a7a6 b5a4 b7b5 a4c2 d6d5 e4d5 d8d5 d2d4 e5d4 c3d4 c8f5 b1c3 f5c2 d1c2 d5d7 a2a4 b5b4 c3e2 f8e8 c1e3 e7d6 a1c1 c6e7 f3e5 d7f5 c2b3 d6e5 d4e5 f5e5 e3f4 e5e4 f2f3 e4g6 b3b4 c7c6 b4c4 f6d5 f4g3');
INSERT INTO evaluations VALUES(8561,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8561,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(8561,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(8561,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(8561,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(8561,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(8561,36,'cp 51','f1e1 e8g8 c2c3 a7a6 b5a4 b7b5 a4c2 d6d5 d2d4 e5d4 e4e5 f6e4 c3d4 c6b4 c1e3 b4c2 d1c2 c8f5 e1c1 d8d7 c2c7 f8c8 c7d7 c8c1 e3c1 f5d7 b1d2 e4d2 c1d2 b5b4 a2a3 b4a3 b2a3 a6a5 a3a4 h7h5 d2g5 f7f6 e5f6 g7f6');
INSERT INTO evaluations VALUES(8562,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(8562,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(8562,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(8562,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(8562,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(8562,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(8562,36,'cp 40','c8d7 a2a4 e8g8 c2c3 f8e8 d2d4 e5d4 c3d4 c6b4 b5d7 d8d7 c1g5 a7a5 b1c3 h7h6 g5h4 c7c6 d1d2 d6d5 e4e5 f6h7 h4e7 d7e7 a1d1 b7b6 f3h2 h7f8 h2g4 f8e6 c3e2 c6c5 f2f4 e6d4 e2d4 c5d4 d2d4');
INSERT INTO evaluations VALUES(24207,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(24207,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(24207,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(24207,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(24207,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(24207,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(24207,36,'cp 39','a2a4 e8g8 c2c3 h7h6 d2d4 e5d4 c3d4 c6b4 b5d7 d8d7 b1c3 f8e8 c1f4 a7a5 e1e3 c7c6 f3d2 e7d8 d2c4 d6d5 c4d6 e8e6 e4e5 f6e8 d6e8 d7e8 c3e2 d8b6 e3g3 e6g6 g3f3 c6c5 d4c5 b6c5 f4e3 c5b6 e3b6 g6b6');
INSERT INTO evaluations VALUES(16160,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(16160,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(16160,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(16160,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(16160,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(16160,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(16160,36,'cp -105','d7d5 h4f6 g7f6 b1c3 d5d4 c3b1 f6f5 e4f5 c8f5 b1d2 d8d7 g2g3 e8c8 f1g2 d8e8 a2a3 f5g4 h2h3 g4e6 d1e2 h8g8 e1c1 f7f6 h1e1 c6e7 d2b3 e7d5 f3d2 c8b8 g2f3 e6h3 f3d5 d7d5 e2h5 h3d7 d2e4 f8e7 h5h6 d7g4 e4f6 e7f6 h6f6 g4d1 e1d1 b8a8 f6f5');
INSERT INTO evaluations VALUES(16166,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(16166,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16166,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(16166,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(16166,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(16166,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(16166,36,'cp -15','d1c2 d7d5 b2b4 f8d6 e4d5 f6d5 b4b5 c6a5 c3c4 d5f6 f1e2 e8g8 e1g1 f8e8 b1d2 c8f5 f1d1 b7b6 d2e4 d6f8 c1b2 f6d7 h2h3 c7c6 a2a4 a8c8 e2f1 f5h7 b2c3 a5b7 a4a5 b7a5 c2a2 a5b7 a2a7 b7c5 e4c5');
INSERT INTO evaluations VALUES(16417,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(16417,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(16417,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(16417,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(16417,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(16417,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(16417,36,'cp -34','e5f4 b1c3 f8b4 c1f4 e8g8 g1e2 d7d5 e4d5 f6d5 c4d5 d8d5 e1g1 d5d7 a2a3 b4c5 g1h1 b7b6 b2b4 c5d6 f4d6 d7d6 c3e4 d6h6 e4g3 c8d7 d1c1 h6c1 a1c1 f8e8 h1g1 f7f6 e2c3 a8c8 c3d5 d7e6 d5c3 e6f7 f1e1 e8e1 c1e1 c6d4');
INSERT INTO evaluations VALUES(23293,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(23293,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(23293,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(23293,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(23293,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(23293,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(23293,36,'cp 40','g1f3 f8b4 c3d5 b4e7 f1g2 a7a5 a2a3 d7d6 e1g1 c8d7 d2d3 g8f6 b2b3 f6d5 c4d5 c6b8 c1d2 e8g8 b3b4 e7f6 d2c3 a5b4 a3b4 a8a1 d1a1 b8a6 a1b2 d8e8 f3d2 e8h5 f1e1 d7b5 d2c4 h5g6 c4a5 h7h5 a5b7');
INSERT INTO evaluations VALUES(29692,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(29692,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29692,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(29692,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29692,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(29692,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(29692,36,'cp 2','h2h3 g4f3 d1f3 b8c6 f1b5 d8d6 b1c3 e7e6 b2b3 h7h5 c1b2 h5h4 e1c1 e8c8 b5c6 d6c6 c3e2 c8b8 c1b1 f8b4 d2d3 c6d6 g2g3 h4g3 f3g3 h8h7 e2d4 b4c5 d1e1 a7a6 d4f3 d6e7 h1g1 c5b4 e1e2 d8h8 g3g5 h7h3 g5g7 h3h1 g7f6 h1g1 f3g1');
INSERT INTO evaluations VALUES(34494,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(34494,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(34494,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(34494,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(34494,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(34494,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(34494,36,'cp 60','b1d2 c8d7 g1f3 d7a4 d2b3 a5b6 d1e2 b8d7 b3c5 d7c5 d4c5 f8c5 b2b3 a4d7 b3b4 c5f8 c1e3 b6c7 e2d2 a8c8 a1c1 g8e7 f1e2 e7g6 e1g1 g6e5 f3e5 c7e5 f2f4 e5f5 e3a7 f8e7 e2d3 f5g4 a7d4 e8g8 a2a3 g4h4 g2g3 h4h6 d2e3 e7f6');
INSERT INTO evaluations VALUES(35137,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(35137,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(35137,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(35137,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(35137,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(35137,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(35137,36,'cp 114','g1f3 d7d6 b1c3 e5f4 c3d5 f6d8 d5f4 g7g5 f4h5 g5g4 f3d4 g8f6 h5f6 d8f6 d4b5 f6e7 f1e2 f8g7 e1g1 e8g8 b5c3 g7d4 g1h1 g4g3 h2g3 c8e6 c3d5 e6d5 e4d5 b8d7 c2c3 d4e5 f1f3 a8e8 d2d4 e5g7 e2b5 a7a6 b5d7 e7d7 c1f4 d7d8 d1d2 d8f6 a1f1 f6g6');
INSERT INTO evaluations VALUES(35134,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(35134,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(35134,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(35134,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(35134,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(35134,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(35134,36,'cp 103','d7d6 b1c3 e5f4 c3d5 f6d8 d5f4 g7g5 f4h5 g5g4 f3d4 g8f6 h5f6 d8f6 d4b5 f6e7 f1e2 c7c6 b5c3 f8g7 d2d4 f7f5 e1g1 e8g8 c1f4 f5e4 d4d5 e4e3 e2g4 b8d7 d5c6 b7c6 d1d6 e7d6 f4d6 f8f1 a1f1 d7b6 g4f3 c8d7 d6f4 b6c4 f3e2 c4b2 c3e4 d7f5 e4d6 f5c2 f4e3 b2d3');
INSERT INTO evaluations VALUES(39080,30,'cp 477',NULL);
INSERT INTO evaluations VALUES(39080,31,'cp 482',NULL);
INSERT INTO evaluations VALUES(39080,32,'cp 488',NULL);
INSERT INTO evaluations VALUES(39080,33,'cp 482',NULL);
INSERT INTO evaluations VALUES(39080,34,'cp 501',NULL);
INSERT INTO evaluations VALUES(39080,35,'cp 505',NULL);
INSERT INTO evaluations VALUES(39080,36,'cp 515','f3e5 d8g5 c4f7 e8d8 d2d4 g5g2 h1f1 g2e4 d1e2 c5b4 c2c3 e4e2 e1e2 g8e7 f7h5 h8f8 e5f7 d8e8 f7d6 e8d8 d6c8 d8c8 c3b4 b8c6 f1d1 c6b4 b1c3 a7a5 c1e3 a5a4 d4d5 a8a5 h5f3 a4a3 b2b3 f8f3 e2f3 b4d5 c3d5 e7d5 d1g1 g7g5 h2h4 d5e3 f3e3 a5e5 e3d3 c7c5 d3d2 e5f5 a1f1 c5c4 h4g5 h6g5 b3c4 c8c7');
INSERT INTO evaluations VALUES(39077,30,'cp 505',NULL);
INSERT INTO evaluations VALUES(39077,31,'cp 493',NULL);
INSERT INTO evaluations VALUES(39077,32,'cp 490',NULL);
INSERT INTO evaluations VALUES(39077,33,'cp 482',NULL);
INSERT INTO evaluations VALUES(39077,34,'cp 477',NULL);
INSERT INTO evaluations VALUES(39077,35,'cp 468',NULL);
INSERT INTO evaluations VALUES(39077,36,'cp 471','d8g5 d2d4 g5g2 c4f7 e8d8 h1f1 g2e4 d1e2 c5b4 c2c3 e4e2 e1e2 g8e7 f7h5 h8f8 e5f7 d8e8 f7d6 e8d8 d6c8 e7f5 a2a3 b4c3 b2c3 b8c6 b1d2 a8c8 f1e1 c6e7 e2f1 e7d5 d2e4 f5d6 a3a4 d6e4 e1e4 c7c5 d4c5 d5f6 e4f4 d8c7 h5f3 g7g5 f4f5 a7a6 c1e3 c8e8 a1e1');
INSERT INTO evaluations VALUES(41284,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(41284,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41284,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(41284,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41284,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41284,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(41284,36,'cp 45','g1f3 c7c5 e1g1 b8c6 d4c5 d6c5 d1d8 c6d8 b1c3 c8e6 c1e3 a8c8 b2b3 f6h5 e3d2 h7h6 a1c1 d8c6 f1d1 e8g8 c3b5 h5f6 d2e3 b7b6 h2h3 f6e4 f3d2 e4d2 d1d2 a7a6 b5c3 c6d4 g3g4 f8d8 f2f4 b6b5 c3d5 e6d5 c4d5 e7e6 d5e6 d4e6 d2d8 c8d8 g1f2 d8e8');
INSERT INTO evaluations VALUES(42554,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(42554,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(42554,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(42554,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(42554,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(42554,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(42554,36,'cp -11','c7c6 g1f3 e8g8 e1g1 d7d5 e4d5 c6d5 c4b3 b8c6 d3d4 e5d4 c3d4 c5b4 h2h3 c8e6 b1c3 b4c3 b2c3 f6e4 d1d3 c6a5 f3d2 f8e8 f1e1 e6f5 d2e4 f5e4 d3d2 a5c4 b3c4 d5c4 c1a3 d8d5 f2f3 e4d3 e1e7 e8e7 a3e7 a8e8');
INSERT INTO evaluations VALUES(44833,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(44833,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(44833,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(44833,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44833,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(44833,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(44833,36,'cp 62','g7g6 g1f3 f8g7 c1e3 e8g8 a1c1 c8g4 f1e2 e7e6 h2h3 g4f3 e2f3 b8c6 b2b3 d6d5 c4d5 b6d5 c3d5 e6d5 e1g1 f8e8 d1d3 c6e7 g2g4 d8d7 c1c5 a8c8 f1e1 c8c6 c5c6 e7c6 e1c1 c6b4 d3d2 g7f8 g1g2 e8c8 c1c8 d7c8 a2a3 b4c2');
INSERT INTO evaluations VALUES(44700,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(44700,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(44700,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(44700,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(44700,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(44700,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(44700,36,'cp 56','g1f3 f8g7 c1e3 e8g8 a1c1 c8g4 f1e2 e7e6 b2b3 d6d5 h2h3 g4f3 e2f3 b8c6 c4d5 b6d5 c3d5 e6d5 e1g1 c6e7 d1d3 h7h5 f1e1 e7f5 c1c5 d8h4 g1f1 f5e3 f2e3 a8d8 c5d5 b7b6 d3e4 h4f6 d5d8 f6d8 e1d1 g7h6 e4d3 d8g5 e3e4 f8c8 e4e5');
INSERT INTO evaluations VALUES(44699,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(44699,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(44699,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(44699,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(44699,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(44699,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(44699,36,'cp 40','f8g7 h2h3 e8g8 f1e2 c8f5 e1g1 b8c6 c1f4 e7e5 f4g5 d8d7 d4e5 d6e5 c4c5 d7d1 f1d1 b6d7 g5e3 c6d4 f3d4 e5d4 e3d4 g7d4 d1d4 d7c5 a1d1 a8c8 g2g4 f5e6 g1g2 g8g7 f2f3 a7a6 e2f1 f8e8 g2g3 e8e7 d4d6 e7d7 d6d7 e6d7 f1c4 d7c6');
INSERT INTO evaluations VALUES(40263,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(40263,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(40263,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(40263,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(40263,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(40263,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(40263,36,'cp 46','h2h3 e8g8 c1g5 h7h6 g5e3 e7e5 d4e5 c8e6 d1d6 b6c4 f1c4 e6c4 f3d2 c4d3 d6d8 f8d8 f2f4 f7f6 e5f6 g7f6 e1f2 b8c6 d2e4 f6g7 h1d1 b7b6 a1c1 d8d7 d1d2 a8d8 g2g4 c6b4 a2a3 d3e4 d2d7 d8d7 a3b4 e4c6 c1c2 d7d3 f4f5 c6d7 f5g6');
INSERT INTO evaluations VALUES(45739,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(45739,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(45739,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(45739,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(45739,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(45739,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(45739,36,'cp -116','c6d5 c4b3 b8c6 c1g5 c8e6 b1c3 d8a5 g5d2 f8b4 f3d1 a5d8 g1e2 e8g8 e1g1 h7h6 e2g3 b4c5 c3a4 c5d6 c2c4 d5c4 b3c4 d8e7 d2e3 a8d8 a1c1 f8e8 a4c3 d6b4 d1e2 e6c4 d3c4 c6d4 e3d4 e5d4 e2e7 e8e7');
INSERT INTO evaluations VALUES(46488,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(46488,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(46488,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(46488,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(46488,34,'cp -110',NULL);
INSERT INTO evaluations VALUES(46488,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(46488,36,'cp -105','c4b3 b8c6 c1g5 d8a5 g5d2 f8b4 f3d1 e8g8 g1e2 f8e8 e1g1 c8g4 b1c3 d5d4 c3e4 b4d2 e4d2 a8d8 h2h3 g4e6 d2e4 f6e4 d3e4 h7h6 b3e6 e8e6 a2a3 a5b6 d1d3 a7a5 f2f4 e6e8 g1h1 b6c5 a1b1 b7b5 e2g3 e5f4 f1f4');
INSERT INTO evaluations VALUES(46495,30,'cp -115',NULL);
INSERT INTO evaluations VALUES(46495,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(46495,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(46495,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(46495,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(46495,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(46495,36,'cp -93','b8c6 c1g5 d8a5 g5d2 f8b4 f3d1 e8g8 g1e2 f8e8 e1g1 c8g4 h2h3 b4d2 b1d2 g4h5 f1e1 a8d8 d2f1 a5c7 d1d2 d5d4 f1g3 h5g6 f2f4 e5e4 d3e4 g6e4 e2c1 c6a5 d2f2 h7h5 b3a4 e8e7 c1b3 h5h4 g3e4 e7e4');
INSERT INTO evaluations VALUES(46663,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46663,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(46663,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(46663,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(46663,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(46663,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(46663,36,'cp 27','e7e5 g1f3 d8e7 f1e2 f8g7 e1g1 e8g8 b2b4 c8f5 b4b5 e5e4 f3d2 h7h5 a2a4 c7c5 b5c6 b7c6 c1a3 f8c8 d2b3 c6c5 a1b1 b8d7 h2h3 f6h7 c3d5 e7d8 d4c5 d6c5 a3b2 d8g5 g1h1 c8e8');
INSERT INTO evaluations VALUES(52298,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52298,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52298,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(52298,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(52298,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52298,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52298,36,'cp 0','b8c6 f1g2 f8g7 g1e2 d7d6 d2d3 a8b8 a2a4 a7a6 e1g1 b7b5 a4b5 a6b5 f2f4 c8d7 h2h3 g8f6 d1e1 e8g8 g3g4 b5b4 c3d1 h7h6 d1e3 d8c7 e3c4 d7e6 c4e3');
INSERT INTO evaluations VALUES(56118,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(56118,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(56118,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(56118,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(56118,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(56118,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(56118,36,'cp -63','c1b2 b8d7 e2e3 c7c5 d2d3 f8d6 g1f3 g8f6 a2a4 e8g8 f1e2 a7a6 e1g1 f8e8 c2c4 d5d4 b1d2 h7h6 f1e1 d6c7 e2f1 c7a5 g2g3 b7b6 f3h4 c8b7 e3e4 a5b4 f1h3 a6b5 a4b5 a8a1 b2a1 g7g6 a1b2 d8c7');
INSERT INTO evaluations VALUES(56119,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(56119,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(56119,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(56119,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(56119,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(56119,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(56119,36,'cp -59','b8d7 e2e3 c7c5 d2d3 f8d6 a2a4 g8f6 g1f3 e8g8 f1e2 d8c7 a4a5 f8e8 f3d2 b7b6 a5b6 c7b6 c2c4 d5d4 e2f3 c8b7 e1g1 a7a6 b1a3 d6c7 f3b7 b6b7 b5a6 a8a6 a3b5 a6a1 d1a1 h7h6 a1a7 b7a7 b5a7 e8a8');
INSERT INTO evaluations VALUES(3848,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(3848,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(3848,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(3848,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(3848,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3848,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(3848,36,'cp 5','c5b6 h2h3 f6e4 e1e4 d6d5 c1g5 f7f6 d4e5 d5c4 d1d8 f8d8 e5f6 d8d1 e4e1 d1e1 f3e1 h7h6 g5e3 b6e3 f2e3 g7f6 b1d2 c8e6 e1c2 g8f7 a1f1 a8g8 g1h2 g8g5 f1f4 g5b5 d2c4 e6c4 f4c4 b5b2 c2d4 b2a2 d4c6 b7c6 c4c6 a7a5 c6c7 f7g6 e3e4 a2c2 c7c6 g6g7 c6a6 c2c3 e4e5 f6e5 a6a5 c3e3 a5b5 h6h5 b5a5 g7g6 a5c5 g6f6 h3h4');
INSERT INTO evaluations VALUES(4581,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(4581,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(4581,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(4581,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(4581,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(4581,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(4581,36,'cp -76','f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 f8e8 d2d4 c6c5 h2h3 c5d4 c3d4 c8e6 b1c3 f6d5 e5d3 e7f8 c1d2 b7b6 d1f3 c7c6 f3g3 f7f6 c3d5 e6d5 e1e8 d8e8 d3b4 d5e6 g3f3 a8c8 a1c1 c6c5 d4c5 c8c5 c1c5 f8c5 b4d3 c5f8 d3f4 e6a2');
INSERT INTO evaluations VALUES(10388,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(10388,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(10388,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(10388,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(10388,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10388,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10388,36,'cp 0','c6a5 g1e2 a5c4 d3c4 e8g8 e2g3 d7d6 e1g1 c7c6 a2a3 b4a5 h2h3 a7a6 a3a4 c8e6 b2b3 a8c8 d1e2 h7h6 a1d1 f8e8 e2f3 d8c7 g3e2 c7d8');
INSERT INTO evaluations VALUES(11483,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(11483,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(11483,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(11483,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(11483,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(11483,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(11483,36,'cp 12','f8c5 c1e3 d7d6 c3d5 c5e3 f2e3 c8e6 d1e2 b7b5 c4a2 e8g8 h1g1 b5b4 a3b4 e6d5 a2d5 f6d5 e4d5 c6e7 e3e4 c7c6 d5c6 d8b6 c2c3 e7c6 e2f2 b6b5 f2c2 f8b8 g2g3 a6a5 b4a5 b5b2 c2b2 b8b2 a5a6 b2b6 g1f1 b6a6 a1a6 a8a6');
INSERT INTO evaluations VALUES(36746,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36746,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(36746,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(36746,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36746,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(36746,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36746,36,'cp 2','f6h5 f4g5 b7b6 g5h4 c7c5 h2h3 c8b7 f1e2 b8d7 e1g1 g7f6 h4f6 h5f6 a2a4 d8c7 d1b3 e7e5 d4e5 d7e5 f3e5 c7e5 f1d1 e5e6 a4a5 f8d8 e2f3 h7h6 b3c2 a8b8 b2b4 e6e7 a5b6 a7b6 b4c5 b6c5 d1b1 g8g7 c2b2 b7a6 b2b8 d8b8 b1b8');
INSERT INTO evaluations VALUES(41310,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41310,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41310,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(41310,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41310,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41310,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41310,36,'cp 0','e8g8 g1f3 b8c6 f1e2 f6d7 d4d5 c6e5 f3e5 d7e5 e1g1 e7e6 d5e6 c8e6 e3e4 d6d5 f1e1 f8e8 d1c2 d5e4 d2e4 d8d5 e4g3 a7a5 a2a4 e6d7 e2f1 d7c6 h2h3 h7h6 a1d1 d5a2 d1a1');
INSERT INTO evaluations VALUES(43282,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(43282,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(43282,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(43282,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(43282,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(43282,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(43282,36,'cp -13','d7d5 g1f3 c7c5 d2d3 e8g8 e1g1 f8e8 b1d2 b7b6 c2c4 b8c6 f3e5 c8b7 e5c6 b7c6 c4d5 f6d5 b2g7 g8g7 d1c1 d8d6 a2a3 a8c8 f1e1 d6f6 a1a2 d5c3 a2c2 c3b5 g2c6 c8c6 c2c4 c6e6 b3b4 c5b4 a3b4 b5d4 e2e3');
INSERT INTO evaluations VALUES(41513,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(41513,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(41513,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(41513,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(41513,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41513,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(41513,36,'cp -13','g1f3 c7c5 e1g1 e8g8 d2d3 d5d4 b3b4 d8b6 a2a3 b8c6 b1d2 f8d8 d2c4 b6b5 b4c5 c8e6 f3d2 a8c8 c2c3 d4c3 b2c3 b5c5 a1c1 f6g4 c3g7 g8g7 a3a4 b7b6 d2e4 c5d4 e4g5 e6c4 c1c4 d4c4 d3c4 d8d1 f1d1 g4f6 f2f4 f6g4 h2h3');
INSERT INTO evaluations VALUES(41512,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(41512,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(41512,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(41512,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(41512,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(41512,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(41512,36,'cp -13','c7c5 e1g1 e8g8 d2d3 f8e8 b1d2 b7b6 e2e3 b8c6 f3e5 c6e5 b2e5 c8g4 f2f3 g4e6 d1e2 d8d7 f3f4 a8d8 d2f3 e6g4 a1d1 f6h5 e5g7 h5g7 e3e4 d7c7 h2h3 d5e4 d3e4 d8d1 f1d1 g4f3 g2f3 e8d8 d1d8 c7d8 e2d3 d8d3');
INSERT INTO evaluations VALUES(44930,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(44930,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44930,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44930,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(44930,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(44930,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44930,36,'cp 8','d8b6 b2b3 e7e6 c2c3 c6c5 g1f3 f8d6 f3e5 b6c7 f1d3 e8g8 e1g1 b8c6 d2f3 b7b6 e5c6 d6f4 e3f4 c7c6 f3e5 c6c7 f1e1 c8b7 a1c1 a8c8 d1d2 f6e4 d2e3 c7e7 c1c2 f7f6 e5f3 g8h8 a2a3');
INSERT INTO evaluations VALUES(48358,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48358,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(48358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48358,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48358,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(48358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48358,36,'cp 0','b2b3 c7c5 e2e3 c5d4 e3d4 f8d6 f1d3 e8g8 e1g1 a7a5 c1b2 b7b6 d3b5 a8a7 d1e2 a7c7 c2c3 b8c6 f3e5 c6e7 a2a4 c8b7 a1c1 d8a8 c1e1 a8d8');
INSERT INTO evaluations VALUES(51027,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(51027,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51027,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(51027,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(51027,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(51027,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(51027,36,'cp 32','d7d5 c4d5 e6d5 g1f3 c7c6 c1f4 g8f6 e2e3 f6h5 f4g5 f8e7 g5e7 d8e7 d1c2 b8d7 a2a4 a6a5 h2h3 h5f6 f1d3 d7b8 e1g1 b8a6 c3a2 e8g8 f3e5 g7g6 f1e1 f6d7 e5f3 d7f6 c2d2 a6b4 a2b4 e7b4 d2b4 a5b4 a4a5 c8e6 f3g5 e6c8 a1a4 f6d7 f2f4');
INSERT INTO evaluations VALUES(57178,30,'cp -158',NULL);
INSERT INTO evaluations VALUES(57178,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(57178,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(57178,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(57178,34,'cp -140',NULL);
INSERT INTO evaluations VALUES(57178,35,'cp -148',NULL);
INSERT INTO evaluations VALUES(57178,36,'cp -143','b8c6 c2c3 h7h5 g4h5 h8h5 d2d4 g8f6 d4e5 c6e5 c1f4 e5g6 f4h2 d7d5 b1d2 f8d6 e2e3 d6h2 h1h2 c7c5 h2g2 d8e7 f1d3 g6e5 d3e2 h5h7 d1a4 c8d7 a4f4 e5g6 f4c7 d7h3 c7e7 e8e7 g1h3 h7h3 g2g1 a8h8 e1c1 h3h2 d1f1 f6e4 d2e4 d5e4 g1g5 e7e6 g5c5 g6e5 c5c7');
INSERT INTO evaluations VALUES(130,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(130,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(130,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(130,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(130,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(130,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(130,36,'cp -102','d1c2 b6c7 f1b5 g8e7 b2b4 c5b6 e1g1 e8g8 b5c6 b7c6 a2a4 a7a5 c1a3 c6c5 f1c1 e7g6 b4c5 b6c5 c3c4 c5a3 b1a3 g6e5 f3e5 c7e5 c4c5 c8d7 a1b1 d7c6 b1b6 e5c7 a3b5 c7d7 c1e1 f8c8');
INSERT INTO evaluations VALUES(131,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(131,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(131,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(131,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(131,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(131,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(131,36,'cp -82','b6c7 f1b5 g8e7 b2b4 c5b6 e1g1 e8g8 b5c6 b7c6 a2a4 a7a5 b1d2 e7g6 f1e1 f7f6 c1a3 c6c5 b4c5 b6c5 e5f6 g7f6 a3c5 c7c5 c2b3 e6e5 b3b5 c5b5 a4b5 c8e6 d2b3 f8b8 b3c5 g8f7 c5e6 f7e6 a1a4 h7h5 g2g3 g6e7 e1b1 b8c8 b1c1 c8c5 c3c4 d5c4 c1c4');
INSERT INTO evaluations VALUES(1330,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(1330,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(1330,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(1330,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(1330,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(1330,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(1330,36,'cp 177','b1c3 a7a6 b5d6 f8d6 d1d6 a5b4 d6c7 g8f6 a2a3 b4e7 c1g5 b7b5 e1c1 e8g8 f2f3 e7e6 c7d6 c6d4 d6e6 f7e6 g5e3 d4c6 c1b1 f8f7 f1e2 c8b7 h1e1 a8c8 d1d6 f6e8 d6d2 e8f6 e3b6 c6e7 a3a4 b5b4 c3a2 e7c6 d2d6 f6e8 d6d1 g8f8 a2c1 d7d6 c1b3');
INSERT INTO evaluations VALUES(1329,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(1329,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(1329,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(1329,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(1329,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(1329,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(1329,36,'cp 173','a7a6 b5d6 f8d6 d1d6 a5b4 d6c7 g8f6 a2a3 b4e7 c1g5 e8g8 e1c1 b7b5 g5e3 f6e8 c7b6 e7e6 f2f4 e5f4 e3f4 c6e7 b6d4 e7g6 f4g3 d7d6 f1e2 g6e5 d1d2 f7f6 h1d1 e6f7 c3d5 c8e6 d4b6 f7a7 b6a7 a8a7 d5e3 a7c7 e3f5 e5c4 d2d4 e6f5 e2c4 b5c4 e4f5 c4c3 b2b3 c7c5 g3d6 e8d6 d4d6 c5f5 d6a6 f5f4');
INSERT INTO evaluations VALUES(1150,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(1150,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(1150,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(1150,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(1150,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(1150,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(1150,36,'cp 177','b5d6 f8d6 d1d6 a5b4 d6c7 g8f6 a2a3 b4e7 c1g5 b7b5 e1c1 e8g8 f2f3 e7e6 c7d6 e6d6 d1d6 f6e8 d6d2 d7d6 b2b3 c8e6 g5e3 f7f5 e4f5 e6f5 a3a4 c6b4 c1b2 a8c8 a4b5 a6b5 c3b5 b4c2 e3b6 f8f7 f1c4 c8c4 b3c4 f7b7 g2g4 b7b6 g4f5');
INSERT INTO evaluations VALUES(1151,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(1151,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(1151,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(1151,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(1151,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(1151,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(1151,36,'cp 184','f8d6 d1d6 a5b4 d6c7 g8f6 a2a3 b4e7 c1g5 e8g8 e1c1 b7b5 f2f3 e7e6 c7d6 c6d4 d6e6 f7e6 g5e3 d4c6 e3c5 f8d8 a3a4 b5b4 c3a2 a6a5 c5b6 d8f8 f1b5 c6a7 b5c4 c8b7 c4b3 a7c6 c1b1 g8f7 a2c1 f8c8 g2g4 f7e7 d1d2 d7d6 h1d1 a8a6 b6e3 c6d4 e3d4 e5d4 d2d4 a6c6 g4g5 f6d7');
INSERT INTO evaluations VALUES(79,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(79,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(79,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(79,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(79,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(79,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(79,36,'cp 193','d1d6 a5b4 d6c7 g8f6 a2a3 b4e7 c1g5 b7b5 f2f3 e7e6 e1c1 e8g8 c7d6 e6d6 d1d6 f6e8 d6d2 d7d6 b2b3 f7f5 c1b2 c8e6 e4f5 f8f5 g5e3 f5f7 c3e4 f7d7 f1e2 e6f7 h1a1 d6d5 a3a4 b5b4 e4c5 d7a7 e3f2 e8f6 a1e1 f6d7 e2d3 g7g6 c5d7 a7d7 f2g3 f7e6 g3e5');
INSERT INTO evaluations VALUES(78,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(78,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(78,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(78,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(78,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(78,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(78,36,'cp 168','a5b4 d6c7 g8f6 a2a3 b4e7 c1g5 e8g8 e1c1 b7b5 g5e3 f6e8 c7b6 e7e6 f2f4 e5f4 e3f4 c6e7 b6d4 e7g6 f4g3 d7d6 f1e2 g6e5 c1b1 c8b7 d4b6 b7c6 c3d5 a8c8 b6a6 e8f6 g3e5 e6e5 d5f6 e5f6 e2g4 c8c7 a6b6 f6e7 g4f3 f8c8 b6d4 c7b7 h2h4 h7h6 d1d2 b7c7');
INSERT INTO evaluations VALUES(5295,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(5295,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(5295,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(5295,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(5295,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(5295,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(5295,36,'cp -12','g8f6 c1e3 b7b6 e1g1 c8b7 f1e1 e8g8 e4e5 f6g4 e3f4 d7d6 e5d6 e7d6 b1d2 d8d7 c4b5 a7a6 b5c6 d7c6 d2c4 a8d8 d1b3 g7e5 f4g3 e5g3 h2g3 b6b5 c4a5 c6c7 a5b7 c7b7 a2a4 f8e8 a4b5 a6b5 b3d1 b5b4 d1a4 e8e1 a1e1 h7h6 c3b4 b7b4 a4b4');
INSERT INTO evaluations VALUES(8007,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(8007,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(8007,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(8007,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(8007,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(8007,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(8007,36,'cp 0','c1e3 b7b6 b1d2 e8g8 h2h3 c8b7 e1g1 d7d6 f1e1 a8c8 a2a4 h7h6 e3f4 g8h7 f4h2 e7e5 a1b1 f6d7 c4d5 b7a6 d5c4');
INSERT INTO evaluations VALUES(8976,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(8976,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(8976,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(8976,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(8976,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(8976,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(8976,36,'cp 66','e2e3 e8g8 a2a3 b4c3 d2c3 b7b6 b2b3 f6e4 c3b2 c6a5 f1d3 c8a6 d1c2 c7c5 d4c5 f7f5 c4d5 a6d3 c2d3 d8d5 d3d5 e6d5 c5b6 a5b3 a1d1 a7b6 d1d5 b3c5 e1g1 c5a4 d5b5 a4b2 b5b2 a8a3 b2b6 a3a2 g2g3 f8a8 h2h4 g7g6 h4h5 g6h5 b6b5 a2a5 b5a5 a8a5 f3d4 e4d2 f1d1 d2e4 d1c1 a5a2 c1c8 g8f7');
INSERT INTO evaluations VALUES(17672,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17672,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(17672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17672,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(17672,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(17672,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(17672,36,'cp 20','d2d4 c5d4 f3d4 f8e7 c1e3 g8f6 e1g1 e8g8 h2h3 c8d7 f2f4 d8c7 d1e2 f6e4 c3e4 d6d5 f4f5 e6f5 c4d5 f5e4 e2h5 c7g3 d5f7 g8h8 e3f4 g3h4 h5h4 e7h4 d4c6 d7c6 f7c4 h4f6 c2c3 f6e7 f4e3 e7f6 c4b3 a8e8 a1e1 a6a5');
INSERT INTO evaluations VALUES(17676,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(17676,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(17676,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17676,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17676,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17676,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17676,36,'cp 0','c5d4 f3d4 f8e7 e1g1 g8f6 c1e3 e8g8 h2h3 c8d7 f2f4 d8c7 d1e2 a8c8 c4d3 c6b4 e2f2 e6e5 d4f3 f6h5 e3b6 c7b8 f4f5 d7c6 b6a7 b8c7 a7b6');
INSERT INTO evaluations VALUES(17692,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17692,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17692,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17692,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(17692,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(17692,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17692,36,'cp 4','f3d4 g8f6 e1g1 f8e7 c1e3 e8g8 d1e2 c8d7 d4c6 d7c6 a4a5 f6d7 f1d1 a8c8 e3d4 d8c7 e2e3 e7f6 h2h3 f8e8 c4f1 e8d8 d4f6 d7f6 d1d4 h7h6 f1e2 d6d5 e4e5 f6e4 f2f4 e4c3 e3c3 c7d7 c3e3 c6b5 c2c3 b5e2 e3e2 d7c7');
INSERT INTO evaluations VALUES(17691,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17691,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17691,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17691,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(17691,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17691,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17691,36,'cp 0','f8e7 e1g1 g8f6 c1e3 e8g8 d1e2 c8d7 g1h1 c6d4 e3d4 d7c6 a4a5 a8c8 f1d1 f6d7 b2b4 e7f6 c4b3 d7e5 h1g1 d8e7 e2e3 f6g5 e3e2');
INSERT INTO evaluations VALUES(21893,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(21893,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(21893,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(21893,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(21893,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(21893,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(21893,36,'cp -46','f8e7 g5d2 f7f5 b1c3 g8f6 g2g3 e8g8 f1g2 d8e8 e1g1 g8h8 h2h3 c8d7 f1e1 a7a5 e4f5 d7f5 d3d4 c6d4 f3d4 e5d4 c3e2 c7c5 c2c3 d4d3 e2f4 g7g5 f4d5 f6d5 g2d5 e8d7 d1h5 e7f6 d2g5 f5h3 h5h6 f6g5 h6g5 d7f5');
INSERT INTO evaluations VALUES(5626,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(5626,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(5626,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(5626,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(5626,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(5626,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(5626,36,'cp -25','g5d2 f7f5 b1c3 g8f6 g2g3 e8g8 f1g2 d8e8 e1g1 e7d8 e4f5 c8f5 f3g5 e8d7 g5e4 f5g4 f2f3 g4e6 d2e3 b7b6 e4g5 e6f5 g3g4 f5g6 f3f4 e5f4 e3f4 d6d5 f4g3 h7h6 g5h3 c6d4 h3f4 g6f7 h2h3 f8e8 d1d2 c7c5 c3e2 d4c6 c2c3');
INSERT INTO evaluations VALUES(27653,30,'cp 354',NULL);
INSERT INTO evaluations VALUES(27653,31,'cp 371',NULL);
INSERT INTO evaluations VALUES(27653,32,'cp 359',NULL);
INSERT INTO evaluations VALUES(27653,33,'cp 356',NULL);
INSERT INTO evaluations VALUES(27653,34,'cp 381',NULL);
INSERT INTO evaluations VALUES(27653,35,'cp 366',NULL);
INSERT INTO evaluations VALUES(27653,36,'cp 377','d2d4 e6f5 g2g4 f5a5 d4d5 e8c8 d5c6 d7c6 e2d3 g8h6 d1e2 h6g4 c1d2 c6f3 e2f3 g4e5 f3h3 e7e6 d3e2 e5c6 e1c1 g7g6 c1b1 f8g7 h3e3 a5b6 e3b6 a7b6 c3e4 c6d4 e2d3 h7h6 c2c3 d4f3 d2c1 g6g5 d3e2 f3h4 f2f4 f7f5 e4g3 g7f6 b1c2 d8d1 h1d1 h4g6');
INSERT INTO evaluations VALUES(27651,30,'cp 378',NULL);
INSERT INTO evaluations VALUES(27651,31,'cp 390',NULL);
INSERT INTO evaluations VALUES(27651,32,'cp 389',NULL);
INSERT INTO evaluations VALUES(27651,33,'cp 380',NULL);
INSERT INTO evaluations VALUES(27651,34,'cp 379',NULL);
INSERT INTO evaluations VALUES(27651,35,'cp 354',NULL);
INSERT INTO evaluations VALUES(27651,36,'cp 386','e6f5 g2g4 f5f6 g4g5 f6f5 d4d5 c6b4 f3d4 f5e5 e1g1 h7h6 a2a3 b4d5 d4f3 d5c3 d1d7 e8d7 f3e5 d7e6 b2c3 e6e5 a1b1 h6g5 f2f4 g5f4 c1f4 e5e6 e2c4 e6d7 b1b7 e7e5 f4e5 f8d6 e5d6 d7d6 f1d1 d6e5 b7b5 e5f6 d1f1 f6e7 f1f7 e7d6 b5d5 d6c6 d5e5 a8d8 e5e6 c6b7 f7g7 h8h6 c4a6 b7b8 e6h6 g8h6 g7g5 c7c6 a6d3 h6f7 g5h5 d8e8 h2h4 f7e5 g1f2 e5g4 f2f3');
INSERT INTO evaluations VALUES(30680,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(30680,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(30680,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(30680,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(30680,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(30680,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(30680,36,'cp 62','c2c4 f8b4 c1d2 b4d2 d1d2 d8h4 g2g3 h4h6 b1c3 h6d2 e1d2 b8c6 d4d5 c6d4 f1d3 d7d6 d2e3 c7c5 a1b1 e6e5 b2b4 b7a6 g1e2 g8f6 g3g4 d4e2 d3e2 c5b4 b1b4 f6d7 h1a1 e8e7 a2a4 h8b8 h2h4 h7h6 e2d3 g7g6 a4a5 b6a5 b4b8 a8b8 a1a5');
INSERT INTO evaluations VALUES(30652,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(30652,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(30652,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(30652,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(30652,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(30652,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(30652,36,'cp 29','f8b4 b1c3 d8h4 g2g3 b4c3 b2c3 h4e7 f1d3 d7d6 g1h3 b8c6 e1g1 e8c8 f3f4 c6a5 h3g5 h7h6 g5f3 e7d7 d1e2 c7c5 f1e1 g8f6 a1b1 f6e8 h2h3 e8c7 f3h4 b7a6 f4f5 h8e8 g1h2 d7a4 f5e6 f7e6 h4g6 a6c4 d4c5');
INSERT INTO evaluations VALUES(34669,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(34669,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(34669,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34669,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34669,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(34669,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(34669,36,'cp -18','a5b4 a1b1 b4d6 g1f3 g8f6 d2d4 g7g6 g2g3 f8g7 f1g2 d6d8 e1g1 e8g8 f1e1 b8c6 c1f4 c6a5 d1e2 c8f5 e2e7 d8e7 e1e7 c7c6 e7e5 b7b5 e5f5 g6f5 f3h4 f6e4 c3e4 f5e4 h4f5 f8e8 b1e1 a7a6 c2c3 a5c4 g2e4 c4b6');
INSERT INTO evaluations VALUES(34635,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34635,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34635,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34635,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(34635,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34635,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34635,36,'cp 0','a1b1 b4d6 d2d4 g8f6 g1f3 a7a6 g2g3 d6c6 d1d3 c8e6 f1e2 b7b5 e1g1 e6c4 d3d2 b8d7 d4d5 c6c5 c1b2 c4d5 c3d5 c5d5 d2c1 e7e6 f1d1 d5c5 b2d4 c5a3 d4b2');
INSERT INTO evaluations VALUES(34634,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34634,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34634,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34634,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34634,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34634,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(34634,36,'cp 0','b4d6 d2d4 g8f6 g1f3 g7g6 g2g3 f8g7 f1g2 d6d8 e1g1 e8g8 f1e1 b8c6 c1g5 f8e8 h2h3 c6a5 g3g4 c7c6 d1d3 f6d5 g5d2 d5f6');
INSERT INTO evaluations VALUES(38777,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(38777,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(38777,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38777,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(38777,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(38777,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(38777,36,'cp 38','b1d2 g8f6 g1f3 b4d6 f4g3 c7c5 d4c5 d6c5 c2c4 c5e7 a1c1 e8g8 a2a3 d8b6 d1c2 c8d7 f1d3 f8c8 e1g1 a7a5 c2b1 b8c6 f1d1 h7h6 f3e5 d7e8 c4d5 e6d5 d2f3 a5a4 h2h3 c8d8 g3h4 a8c8 e5g4 f6g4 h4e7 c6e7 h3g4');
INSERT INTO evaluations VALUES(44574,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(44574,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(44574,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(44574,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(44574,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(44574,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(44574,36,'cp -21','e4e5 f6g8 c2c3 b8c6 f3e3 e7e6 g1f3 d7d6 d2d4 g8e7 f1d3 e7d5 e3e4 c5d4 c3d4 d5b4 d3b5 d6e5 e1g1 e5d4 f3d4 f8e7 b1c3 e8g8 f1d1 c6d4 d1d4 b4d5 c3d5 e7c5 d4d1 e6d5 d1d5 d8b6 c1e3 c5e3 e4e3 b6e3 f2e3 c8e6 d5d4 a8c8 b5d3 g7g6 a1d1 g8g7 a2a3 c8c5 g1f2 f8c8 h2h4 c5c1 d1c1 c8c1 f2e2 c1a1 e2f3');
INSERT INTO evaluations VALUES(45652,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(45652,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(45652,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(45652,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(45652,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45652,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(45652,36,'cp 31','g1f3 f8e7 e1g1 b8d7 b2b3 b7b6 f3d2 c8b7 e2e4 f6e4 d2e4 d5e4 g2e4 d7f6 e4g2 d8d7 c1b2 c6c5 d4c5 d7d1 f1d1 b7g2 g1g2 e7c5 b1c3 e8e7 a2a3 h8d8 b3b4 c5d4 c3d5 d8d5 b2d4 d5d7 d4e3 d7d1 a1d1 a8c8 c4c5 f6d7 c5b6 a7b6 d1c1 c8c1');
INSERT INTO evaluations VALUES(49436,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(49436,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(49436,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49436,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49436,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(49436,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49436,36,'cp 0','d7d5 g1f3 c7c5 b2b3 b7b6 c2c4 c8b7 c4d5 e6d5 c1b2 b8d7 f1d3 f8d6 b1c3 e8g8 e1g1 d8e7 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(235,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(235,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(235,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(235,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(235,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(235,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(235,36,'cp -18','d7c6 d1e2 c8e6 b1c3 e8c8 c1d2 f6e7 e1c1 g8f6 c1b1 f6d7 f3f4 f7f6 d2c1 d7b6 d1d8 h8d8 e2f3 c5d4 f1e2 e6d7 c1d2 c8b8 a2a3 g7g6 h1d1 a7a6 e2d3 d7e6 h2h3 b6c4 d3c4 e6c4 d2c1 a6a5 c3e2');
INSERT INTO evaluations VALUES(201,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(201,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(201,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(201,36,'cp 0','d1e2 c8e6 b1c3 e8c8 c1d2 f6e7 c3a4 c5d6 g2g3 c8b8 a4c3 g8h6 f1g2 f7f6 e1c1 h6f7 f3f4 d6b4 c1b1 h8e8 c3a4 b4d2 d1d2 e7b4 d2d8 e8d8 a4c3 f7d6 h1d1 d8e8 e2f3 d6c4 b2b3 e6g4 f3g4 c4a3 b1b2 a3c4');
INSERT INTO evaluations VALUES(4046,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(4046,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(4046,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(4046,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(4046,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(4046,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(4046,36,'cp 86','d1d2 c7c5 g1e2 a7a6 a1d1 b7b6 d4c5 b6c5 b2b3 c8b7 g2g4 b7c6 f1g2 a8b8 d1b1 f6e8 e1g1 d8c7 e3g5 c7d8 f3f4 h7h6 g5h4 a6a5 f4f5 a5a4 g4g5 h6g5 d2g5 a4b3 a2b3 d7e5 g5e7 d8e7 h4e7');
INSERT INTO evaluations VALUES(4044,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(4044,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(4044,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(4044,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(4044,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(4044,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(4044,36,'cp 67','c7c5 g1e2 a7a6 d4c5 d7c5 a1d1 c5e6 e2c1 c8d7 f1e2 a8c8 e1g1 e6c5 f1e1 f8e8 c1b3 c5a4 d1c1 a4c3 c1c3 d7a4 b3d4 d8a5 e1c1 h7h5 h2h3 f6d7 g1h1 e7e6 a2a3 a5d8 b2b3 a4c6 b3b4 c6a4 e2d1 a4d1 c1d1 d8c7 c3c1');
INSERT INTO evaluations VALUES(6218,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(6218,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6218,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(6218,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(6218,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6218,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6218,36,'cp 0','d7d6 c2c3 g8f6 b2b4 c5b6 b1d2 c6e7 a2a4 a7a5 e1g1 e8g8 f1e1 e7g6 b4b5 f8e8 d3d4 e5d4 c3d4 d6d5 e4d5 e8e1 d1e1 f6d5 d2b3 c8e6 c1d2 c7c6 b5c6 b7c6 a1c1 d8f6 b3c5 e6h3 c5e4 f6f5 e4g3 f5f6');
INSERT INTO evaluations VALUES(5739,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(5739,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5739,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5739,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5739,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5739,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5739,36,'cp 9','e1g1 g8f6 b1c3 c6a5 c3a4 c5b6 c4b3 e8g8 c1d2 c8d7 c2c4 a5b3 a2b3 b6d4 a4c3 d4b6 f1e1 a7a5 d2c1 c7c6 d3d4 e5d4 f3d4 f8e8 c1f4 b6c5 d4f3 c5b4 f4d6 b4c3 b2c3 f6e4 d6e5 e4g5 f3g5 h6g5 d1f3 d8e7 e5d4');
INSERT INTO evaluations VALUES(5738,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5738,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5738,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5738,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5738,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5738,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5738,36,'cp 3','g8f6 c2c3 a7a6 a2a4 e8g8 c4a2 c5a7 c1e3 a7e3 f2e3 c8d7 b1d2 b7b5 a2b3 c6a5 b3c2 c7c5 d1e1 a5c6 f3h4 a8b8 a4b5 a6b5 h4f5 d7f5 f1f5 b8a8 a1a8 d8a8 c2b3 c6e7 f5f2 g8h8 g1h2 e7g8 f2f1 a8a7 e1e2 f8a8 e2f3 a8f8');
INSERT INTO evaluations VALUES(16988,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(16988,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16988,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16988,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16988,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(16988,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16988,36,'cp 0','f8c5 f3e5 d8e7 e5f3 d7d5 c3d5 e7e4 d5e3 c8g4 c4e2 d4e2 d1e2 e8c8 d2d3 e4e6 e3g4 e6e2 e1e2 f6g4 c1g5 d8e8 e2d2 c5f2 h2h3 g4e5 f3e5 e8e5 g5f4 e5f5 g2g3 f2d4 a1e1 g7g5 f4e3 d4e3 e1e3 f5f2 e3e2 f2f3 e2e3');
INSERT INTO evaluations VALUES(19720,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19720,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19720,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19720,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19720,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19720,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19720,36,'cp 0','d2d4 g8f6 g1f3 c8g4 e1g1 e7e6 h2h3 g4f3 d1f3 f8d6 c1g5 h7h6 b5c6 b7c6 g5h4 e8g8 c3a4 a8b8 b2b3 e6e5 d4e5 d6e5 a1e1 d8d6 e1e5 d6e5 h4g3 e5g5 g3b8 f8b8 f3c3 f6h5 c3f3');
INSERT INTO evaluations VALUES(25225,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(25225,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(25225,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(25225,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(25225,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(25225,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(25225,36,'cp 63','c2c4 f8d6 b1c3 e8g8 c4d5 f6d5 h2h3 g4h5 e2e4 d5b6 g3g4 h5g6 a2a3 d8d7 f1e1 h7h6 c1e3 c6a5 f3d2 a5c4 d2c4 b6c4 e3c1 c7c6 f2f4 d6c7 f4f5 e6f5 g4f5 g6h7 d1d3 c4d6 c1f4 g8h8 a1d1 a8e8 b2b4 c7b6');
INSERT INTO evaluations VALUES(25196,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(25196,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(25196,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(25196,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(25196,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(25196,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(25196,36,'cp 61','f8d6 c4d5 f6d5 h2h3 g4h5 b1c3 e8g8 e2e4 d5b6 g3g4 h5g6 a2a3 d8d7 c1e3 a8d8 f1e1 c6a5 e3g5 f7f6 e4e5 f6g5 e5d6 d7d6 f3g5 d8e8 d4d5 e6d5 g2d5 g8h8 e1e8 g6e8 d1c2 d6g6 c2g6 e8g6 g5e6 f8e8 d5a2 c7c6 f2f4 g6f7 e6c7 e8e7 a2f7 e7f7 c7e6 g7g6 a1d1');
INSERT INTO evaluations VALUES(30685,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(30685,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(30685,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(30685,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30685,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(30685,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(30685,36,'cp 23','f8b4 d1f3 g8f6 f1d3 c7c5 d4d5 e6d5 e4e5 b4c3 b2c3 f6e4 c3c4 e4g5 f3g3 g5e6 c4d5 e6f4 g3f4 b7d5 g1e2 c5c4 d3c4 d5c4 f4c4 e8g8 f2f4 b8c6 e1g1 a8c8 c4b3 d8e7 a1d1 c6a5 b3f3 f7f6 e5f6 f8f6 e2g3 a5c4 f3d5 g8h8 g3e4 c4e3 d5d7 e7d7 d1d7 e3f1');
INSERT INTO evaluations VALUES(27956,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(27956,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(27956,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(27956,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(27956,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(27956,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(27956,36,'cp 19','g1e2 g8f6 a2a3 b4c3 e2c3 b7e4 c3e4 f6e4 d1f3 d7d5 f1b5 c7c6 b5d3 e8g8 e1c1 b8d7 d3e4 d5e4 f3e4 d7f6 e4c6 a8c8 c6b5 f6d5 f4e3 d8e7 b5d3 e7c7 c1b1 b6b5 c2c3 a7a5 e3d2 c7b6 d1c1 b6b7 f2f3 f8d8 h1e1 h7h6 d3e2');
INSERT INTO evaluations VALUES(30834,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(30834,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(30834,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(30834,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(30834,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(30834,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(30834,36,'cp -21','e2e3 c7c5 b2b3 g8f6 g2g3 g7g6 c1b2 f8g7 f1g2 d7d5 d1e2 e8g8 e1g1 b8c6 c2c4 d5c4 b3c4 c6b4 d2d4 c5d4 e3d4 a8c8 b1a3 e7e6 f3e5 b7g2 e2g2 a7a6 a1b1 d8c7 f1d1 f8d8 d4d5 e6d5 b2d4 c7e7 g2b2 b4c6 e5c6 c8c6 c4d5');
INSERT INTO evaluations VALUES(30833,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(30833,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(30833,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(30833,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(30833,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(30833,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(30833,36,'cp -28','g8f6 b2b3 g7g6 g2g3 f8g7 c1b2 c7c5 f1g2 e8g8 e1g1 d7d5 d1e2 b8c6 a2a4 c6b4 b1a3 e7e6 f3e5 d8e7 c2c3 b4c6 e5c6 b7c6 d2d3 f8d8 c3c4 d8d7 b2e5 f6e8 a1d1 a8d8 a3b5 a7a6 b5c3 g7e5 f4e5');
INSERT INTO evaluations VALUES(32947,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(32947,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(32947,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(32947,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(32947,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(32947,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(32947,36,'cp 199','g1f3 g7g6 d2d4 f8g7 c1f4 g8f6 f4c7 e6c6 c7f4 e8g8 d1d2 c6b6 e1g1 b8c6 d4d5 c6a5 b2b3 f8d8 f4e5 d7e8 e5d4 b6c7 a1d1 a8c8 d4e5 c7c5 f1e1 a5c6 e5f6 e7f6 c3e4 c5b4 d2f4 f6f5 d5c6 f5e4 d1d8 c8d8');
INSERT INTO evaluations VALUES(32946,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(32946,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(32946,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(32946,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(32946,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(32946,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(32946,36,'cp 196','g7g6 d2d4 f8g7 c1f4 e6b6 d1d2 e7e6 e1c1 g8e7 f3e5 b8c6 e5d7 e8d7 c3a4 b6a5 d2a5 c6a5 a4c5 d7c8 c1b1 a5c6 c2c3 a7a5 h2h4 b7b6 c5d3 e7d5 e2f3 a5a4 a2a3 h7h6 h1e1 c6a5 f4g3 h8d8 g3e5 d8g8 h4h5 g6g5');
INSERT INTO evaluations VALUES(35368,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(35368,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(35368,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35368,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(35368,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35368,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(35368,36,'cp -12','e1g1 c7c5 d3d4 b8c6 c2c3 h7h6 d4c5 d6c5 c1f4 c8e6 h2h3 f6d5 f4g3 d5b6 b1d2 a8c8 f1e1 f8d8 e2d1 d8d7 d1c2 c8d8 e1e2 g8f8 a2a3 b6c4 d2c4 e6c4 e2e3 e7d6 f3d2 c4a6 g3d6 d7d6 d2e4 d6d5 c2b3 c5c4 b3c2 g7g6 a1e1');
INSERT INTO evaluations VALUES(35369,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(35369,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(35369,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(35369,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(35369,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(35369,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(35369,36,'cp -16','c7c5 d3d4 b8c6 c2c3 h7h6 h2h3 a7a6 f1d1 c8e6 d4c5 d6c5 c1f4 b7b5 f4d6 e7d6 d1d6 f8c8 d6d1 a8b8 b1d2 c8e8 e2f1 c5c4 b2b3 f6d5 d2e4 c4b3 a2b3 f7f5 e4d6 e8d8 c3c4 d8d6 c4d5 d6d5 d1d5 e6d5 a1a6');
INSERT INTO evaluations VALUES(41594,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(41594,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(41594,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(41594,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(41594,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(41594,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(41594,36,'cp -52','a2a4 e8g8 b1c3 c7c6 h2h3 a7a5 g1e2 b8a6 c4a6 a8a6 e1g1 f8e8 f3g3 f6h5 g3g5 d8g5 c1g5 f7f6 g5d2 f6f5 f1b1 g7f6 f2f3 f6d8 d2e1 h5f6 e1f2 d6d5 e4d5 c6d5 d3d4 e5e4 e2f4 a6c6 f4d5 f6d5 c3d5 e4f3');
INSERT INTO evaluations VALUES(43982,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(43982,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(43982,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(43982,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43982,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(43982,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(43982,36,'cp -21','b8c6 c1f4 d7c5 d1d2 f8e7 f1e2 e8g8 e1g1 c5e4 c3e4 d5e4 d2d8 f8d8 f3d2 c6e5 d2e4 e5g6 f4e3 b7b6 f1d1 c8b7 f2f3 b7e4 f3e4 e7c5 e3c5 b6c5 d1d8 a8d8 a1d1 d8d1 e2d1 g8f8 g1f2 f8e7 a2a3 g6e5 b2b3 e7d6 d1e2 f7f6 g2g3 e5f7 e2c4 g7g5 c4e2 d6e5');
INSERT INTO evaluations VALUES(48065,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(48065,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(48065,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(48065,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(48065,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(48065,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(48065,36,'cp 145','c8e6 g1f3 c7c5 f1b5 b8c6 e1g1 d8b6 b5c6 b7c6 b1c3 e8c8 c3a4 b6b5 c2c4 e6c4 f1e1 c5d4 f3d4 b5a6 c1e3 f8b4 d1c2 c4d3 c2b3 b4e1 a4c5 a6b6 b3d3 e1b4 c5a4 b6b5 d3c2 d8d4 e3d4 h8d8 a4c3 b5c4 d4f6 g7f6 c2h7 b4c3 b2c3 c4c3');
INSERT INTO evaluations VALUES(49373,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(49373,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(49373,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(49373,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(49373,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(49373,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(49373,36,'cp 35','e4e5 f6g8 d2d4 d7d6 e5d6 f8d6 b1c3 g8f6 g1f3 e8g8 f1d3 c7c5 d4c5 d6c5 e1g1 c8d7 d1e2 d7c6 f1d1 b8d7 f3e5 d7e5 d3h7 g8h7 d1d8 a8d8 e2e5 c5f2 g1f1 f2d4 e5e2 d4c3 b2c3 f6e4 c1a3 f8e8 a1d1 e4c3 e2c2 c3e4 f1g1 f7f5 h2h3');
INSERT INTO evaluations VALUES(50833,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(50833,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(50833,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(50833,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(50833,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(50833,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(50833,36,'cp 115','c4d5 g8f6 g1f3 g7g6 b1c3 f6d5 e2e4 d5c3 b2c3 c7c5 f1e2 f8g7 e1g1 b8d7 f1e1 e8g8 a2a4 c5d4 c3d4 d7f6 e2d3 c8g4 c1e3 b7b5 a4b5 a6b5 a1b1 a8b8 h2h3 g4f3 d1f3 d8d7 e1c1 f8d8 f3e2 b5b4 c1c4 b4b3 e4e5');
INSERT INTO evaluations VALUES(50827,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(50827,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(50827,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(50827,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(50827,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(50827,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(50827,36,'cp 119','g8f6 g1f3 g7g6 c1g5 f6d5 e2e4 h7h6 g5d2 d5b6 a2a4 a6a5 f1b5 c7c6 b5e2 c8g4 e1g1 f8g7 d2c3 g4f3 e2f3 e8g8 e4e5 b6d5 d1b3 b7b6 b1d2 b8a6 g2g3 a6b4 f1d1 d8c7 d2e4 a8d8 a1c1 e7e6 h2h4 c6c5 d4c5 b6c5 e4d6 d5c3');
INSERT INTO evaluations VALUES(54633,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(54633,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(54633,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(54633,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(54633,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(54633,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(54633,36,'cp 127','f1d3 c6c5 e1g1 b8c6 c2c3 f8e7 d1e2 c8d7 f1e1 c5d4 c3d4 f6e5 f3e5 g8f6 e5c6 b7c6 c1e3 d8c7 b1d2 e8g8 h2h3 e6e5 e3f4 e5f4 e2e7 c7b6 d2f3 a8e8 e7a3 f6e4 a1c1 d7f5 b2b3 h7h6 a3b2 e8e6 a2a3 a7a5 f3e5 f8e8 b3b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(54890,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(54890,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(54890,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(54890,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(54890,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54890,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(54890,36,'cp 83','g8f6 c1g5 f8g7 d1d2 e8g8 g5h6 c8g4 h6g7 g8g7 f3e5 g4f5 e1c1 b8d7 f1d3 f5d3 e5d3 f8e8 c1b1 c7c6 h2h4 h7h5 f2f3 d8a5 g2g4 a8d8 g4h5 a5h5 d2f2 e7e5 c3e4 e5d4 f2d4 h5f3 e4d6 e8e7 d6b7 d8h8 b7d6 f3e3 d4c4 h8h5 c4c6 d7b6');
INSERT INTO evaluations VALUES(46828,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(46828,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(46828,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(46828,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(46828,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(46828,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(46828,36,'cp 103','c1f4 f8g7 d1d2 e8g8 e1c1 c8g4 f1e2 b8d7 h2h3 g4f3 e2f3 c7c6 c1b1 d7b6 g2g4 f6d5 c3d5 c6d5 h3h4 b6c4 d2c1 a8c8 h4h5 e7e6 f3e2 a7a5 e2c4 c8c4 c1e3 d8d7 h5g6 f7g6 f4h6 b7b5 h6g7 d7g7 e3e6 g7f7');
INSERT INTO evaluations VALUES(2959,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(2959,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(2959,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2959,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(2959,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(2959,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2959,36,'cp 16','d5c4 d3c4 h7h6 g5h4 b7b5 c4e2 a7a6 d1c2 c6c5 d4c5 d7c5 f1d1 d8b6 b2b4 c5d7 a2a4 e7b4 a4b5 b4c3 c2c3 c8b7 c3d4 b6d4 f3d4 a6b5 d4b5 a8a1 d1a1 g7g5 h4g3 f8a8 a1a8 b7a8 f2f3 f6d5 g1f2 f7f5 g3d6 g8f7 d6a3 f7f6 e2d3 d5e7 b5d6');
INSERT INTO evaluations VALUES(2980,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2980,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2980,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2980,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(2980,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(2980,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(2980,36,'cp 11','d3c4 h7h6 g5h4 b7b5 c4e2 a7a6 f3e5 c8b7 h4f6 d7f6 e5d3 d8c7 d1c2 e7d6 h2h3 e6e5 d4e5 d6e5 d3c5 a8d8 c5b7 c7b7 f1d1 b7c7 a2a4 d8d1 a1d1 f8b8 b2b3 g7g6 e2f3 h6h5 c3e4 f6e4 f3e4 c6c5 d1c1 b8c8');
INSERT INTO evaluations VALUES(2977,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(2977,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(2977,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(2977,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(2977,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(2977,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(2977,36,'cp 26','h7h6 g5h4 b7b5 c4d3 c8b7 a1c1 a8c8 a2a3 a7a5 f3e5 d7e5 d4e5 f6d7 h4e7 d8e7 f2f4 d7c5 d3b1 b5b4 c3e2 c5d7 d1d3 g7g6 f1d1 f8d8 d3d6 e7d6 d1d6 c6c5 e2g3 c5c4 b1e4 b7e4 g3e4 g8f8 c1d1 b4a3 b2a3 c8c7 e4c3 f8e7 h2h3 c7b7 d6a6 d7c5');
INSERT INTO evaluations VALUES(6980,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(6980,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(6980,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(6980,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(6980,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(6980,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(6980,36,'cp 85','d1b3 e8f8 c4f7 g8f6 b1c3 c6a5 b3a4 f8f7 a4a5 c7c6 a5a4 h8f8 e4e5 f6d5 c3d5 c6d5 c1f4 f7g8 e1g1 d7d6 f1e1 d6e5 f4e5 c8d7 a4d4 f8f7 a1c1 d7c6 d4c3 d8b6 a2a3 e7f6 e5f6 f7f6 e1e7 a8d8 f3d4 d8f8 c1c2 f8f7 e7f7 f6f7 d4c6 b7c6 c3c6');
INSERT INTO evaluations VALUES(8572,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(8572,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(8572,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(8572,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(8572,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(8572,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(8572,36,'cp -25','c6a5 a2a3 a5c4 d3c4 h7h6 g5f6 e7f6 c3d5 c7c6 d5f6 d8f6 d1d3 e8g8 e1c1 f8d8 d1d2 c8g4 h1d1 d8d7 d3e3 a8d8 a3a4 g4f3 e3f3 f6f3 g2f3 g8f8 b2b4 f8e7 c1b2 e7e6 b2c3 a7a6 a4a5 d6d5 e4d5 c6d5 d2d5 d7d5 c4d5 d8d5 d1d5 e6d5 c3d3 f7f5 c2c4 d5d6 h2h4');
INSERT INTO evaluations VALUES(10472,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(10472,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(10472,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(10472,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(10472,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(10472,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(10472,36,'cp -27','e8g8 d4c6 b7c6 f1d3 f8e8 e1g1 d7d5 c1f4 h7h6 g1h1 a8b8 a2a3 b4c3 b2c3 c6c5 e4d5 f6d5 d1d2 d5f4 d2f4 d8d6 f4a4 c8d7 a4a7 d6e5 a1d1 d7c6 a7a6 b8b6 a6c4 c6b5 c4e4 e5e4 f3e4 c5c4 d3e2 e8e4 d1d8 b5e8 e2f3');
INSERT INTO evaluations VALUES(4467,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(4467,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(4467,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(4467,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(4467,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(4467,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(4467,36,'cp -18','d4c6 b7c6 f1d3 f8e8 e1g1 d7d5 c1g5 h7h6 g5h4 a8b8 g1h1 d8d6 h4g3 d6d7 e4d5 c6d5 f1e1 e8e1 d1e1 b4a5 e1e5 c8b7 e5f5 d7f5 d3f5 d5d4 c3e2 a5b6 a2a4 a7a5 f5d3 f6d5 g3e5 d5b4 e5d4 b4d3 d4b6 c7b6 c2d3 b7a6 e2f4 b8d8');
INSERT INTO evaluations VALUES(11102,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(11102,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(11102,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(11102,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(11102,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(11102,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(11102,36,'cp 130','b5d6 f8d6 d1d6 d8e7 d6e7 c6e7 c1g5 b7b5 f2f3 c8b7 e1c1 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e7f5 d1e1 e8d7 f3f4 a8e8 g2g3 f5d6 f1h3 d7c7 e1e3 e5f4 e3c3 d6c4 h1f1 f4g3 c3g3 f6f5 b2b3 c4e3 f1f3 e3d5 f3f5');
INSERT INTO evaluations VALUES(11110,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(11110,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(11110,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(11110,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(11110,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(11110,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(11110,36,'cp 99','f8d6 d1d6 d8e7 d6e7 c6e7 c1g5 b7b5 e1c1 c8b7 f2f3 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e7f5 f1d3 f5e3 d1d2 f6f5 d5d6 f5f4 h1e1 h8g8 d3h7 g8g7 h7e4 a8d8 g2g4 e8d7 d2d3 h6h5 e1e3 f4e3 e4f5 d7c6 d6d7 b5b4 d3e3 d8d7 f5d7 c6d7 e3e4 h5g4 f3g4 g7h7 e4e5 h7h2 e5f5 h2h1 c1d2 h1h2 d2d3 d7e7');
INSERT INTO evaluations VALUES(7054,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(7054,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(7054,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(7054,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(7054,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(7054,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(7054,36,'cp 105','d1d6 d8e7 d6e7 c6e7 c1g5 b7b5 e1c1 b5b4 g5f6 g7f6 c3a4 a8b8 d1d6 e7c6 d6f6 e8e7 f6h6 d7d6 b2b3 c6d4 a4b2 c8e6 b2c4 a6a5 c4e3 a5a4 f1c4 e7f8 g2g4 f8g7 g4g5 b8a8 c1b2 a8a5 h2h4 h8a8 h1a1 a5a7 e3d5 e6d5 c4d5 a4a3 b2c1');
INSERT INTO evaluations VALUES(11341,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11341,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(11341,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(11341,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(11341,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(11341,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(11341,36,'cp 37','a2a4 f8e7 h2h3 e8g8 c1d2 c6a5 c4a2 a5c6 f1e1 c6b4 a2c4 c8e6 a4a5 d8d7 c3e2 e6c4 d3c4 b4c6 e2g3 d7e6 b2b3 g8h7 g3f5 a8e8 g2g4 e7d8 d1e2 c6e7 g1h2 e7f5 e4f5 e6c8 g4g5 f6h5 e2e4 d8g5');
INSERT INTO evaluations VALUES(13681,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(13681,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13681,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(13681,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(13681,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13681,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(13681,36,'cp 25','e3d4 f6h5 f4e3 f8d6 f1d3 h5f4 e3f4 d6f4 e1g1 c8d7 g2g3 f4d6 a2a3 e8g8 f1e1 g8h8 d1e2 a8c8 a1d1 c6e7 d1c1 d7e8 f3g5 h7h6 d2f1 e7c6 h2h4 c8c7 f2f4 c7e7 f1h2 d8c7 g1g2 g7g6 g5f3 f7f6 h4h5 g6h5');
INSERT INTO evaluations VALUES(17617,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(17617,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(17617,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(17617,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(17617,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(17617,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(17617,36,'cp 28','b5a4 c8d7 g1f3 c5d4 c3d4 f8b4 b1c3 c6a5 e1g1 b4c3 b2c3 d7a4 d1a4 b7b5 a4c2 g8e7 a2a4 h7h6 c1a3 e8g8 f3d2 f8e8 f1b1 a5c4 d2c4 d5c4 a3e7 d8e7 b1b4 e8d8 c2b1 e7e8 h2h4 a8b8 a4b5 a6b5 a1a7 e8c6 h4h5 c6b6 b1a1 b6c6 b4b1 b8b7 a7a8');
INSERT INTO evaluations VALUES(20876,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(20876,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(20876,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(20876,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(20876,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(20876,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(20876,36,'cp 122','c8f5 f1b5 e7e6 f3e5 g8e7 b1c3 a7a6 b5a4 a8b8 c1g5 d8d6 f2f4 b7b5 a4b3 c6e5 f4e5 d6b6 e1g1 f5g6 c3e2 e7f5 g2g4 f5d4 e2d4 c7c5 d4e6 c5c4 e6d4 f8c5 g5e3 h8f8 d1d2 c4b3 c2b3 c5d4 e3d4 b6e6 h2h3 f8f1 a1f1 b8b7 f1c1 g6e4 d2c3 b7e7 c3c6 e8f7 c6e6 e7e6 g1f2');
INSERT INTO evaluations VALUES(26479,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(26479,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(26479,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(26479,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(26479,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(26479,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(26479,36,'cp 122','f1b5 e7e6 f3e5 g8e7 b1c3 a7a6 b5a4 a8b8 c1g5 d8d6 f2f4 b7b5 a4b3 c6e5 f4e5 d6b6 e1g1 f5g6 c3e2 e7f5 c2c3 f8e7 g5e7 f5e7 d1d2 c7c5 f1f2 e7f5 d4c5 b6c5 e2d4 e8d7 a2a4 h8f8 a4b5 a6b5 a1a6 f5d4 c3d4 f8f2 g1f2 c5c7 b3d1 b8b6 a6a5 d7e7 h2h4 b6b8 h4h5 g6e4 d1e2 e7f8 a5b5 b8b5');
INSERT INTO evaluations VALUES(26475,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(26475,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(26475,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(26475,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(26475,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(26475,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(26475,36,'cp 120','e7e6 f3e5 g8e7 b1c3 a7a6 b5a4 a8b8 c1g5 d8d6 f2f4 b7b5 a4b3 c6e5 f4e5 d6b6 e1g1 f5g6 c3e2 e7f5 c2c3 f8e7 g5e7 f5e7 a2a4 g6f5 a4a5 b6c6 e2g3 e8g8 g3f5 f8f5 b3c2 f5f1 d1f1 e7g6 c2g6 h7g6 f1d3 b5b4 d3g6 b4c3 b2c3 c6c3 g6e6 g8h7 e6f5 h7g8 a1d1 b8b2 f5d3 c3d3 d1d3 b2b5 d3c3 b5a5 c3c7 a5a4 c7c8 g8h7');
INSERT INTO evaluations VALUES(21281,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(21281,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21281,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(21281,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(21281,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21281,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21281,36,'cp 0','c6a5 c4b3 a5b3 a2b3 g8e7 d3d4 e7g6 e1g1 f8e7 d4e5 d6e5 d1d8 e7d8 c1e3 b7b6 f3d2 c8b7 b1c3 a7a5 d2c4 d8e7 f1d1 a8d8 c3b5 d8d7 d1d7 e8d7 a1d1 d7c8 e3b6 c7b6 c4b6 c8b8 b6d7 b8a8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(27461,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(27461,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(27461,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(27461,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(27461,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(27461,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(27461,36,'cp 159','c1f4 d7g4 d4d2 f8e7 f2f3 g4h4 g2g3 h4h3 e1c1 e8g8 c1b1 f8d8 h1e1 c7c6 d2f2 f6e8 d1d3 h3d7 a2a4 d7e6 h2h4 e8c7 f4g5 f7f6 g5f4 b7b5 a4b5 c7b5 c3b5 c6b5 e1d1 a7a5 f2d4 b5b4 d4d5 e6f7 d5f7 g8f7 f4d6 e7d6 d3d6 f7e7 d6d8');
INSERT INTO evaluations VALUES(33211,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(33211,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(33211,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(33211,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(33211,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(33211,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(33211,36,'cp 49','g1f3 b8c6 c2c3 c5b6 a2a4 a7a6 b2b4 d7d6 e1g1 g8e7 c1e3 b6e3 f2e3 e8g8 b1d2 f6h6 d1e2 c6b8 a4a5 e7g6 f1f2 b8d7 g2g4 d7b8 g4g5 h6h3 c4d5 c7c6 d5a2 c8g4 f2g2 b8d7 g2g3 h3h5 e2g2 g4e6 d2c4');
INSERT INTO evaluations VALUES(37083,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(37083,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(37083,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(37083,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(37083,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(37083,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(37083,36,'cp 50','e1g1 c7c5 a2a3 b4c3 b2c3 b7b6 c4d3 c8b7 f1e1 b7e4 d3e4 f6e4 d1c2 e4f6 d4c5 b6c5 c3c4 h7h6 h2h3 b8c6 a1b1 f6d7 c1b2 a8b8 c2e4 d8c7 e4g4 e6e5 e1d1 d7f6 g4f5 f8e8 b2c3 b8b1 f5b1 e8b8');
INSERT INTO evaluations VALUES(37081,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(37081,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(37081,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(37081,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(37081,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(37081,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(37081,36,'cp 54','c7c5 a2a3 b4c3 b2c3 b7b6 f1e1 c8b7 c4f1 d8c7 f3e5 f8d8 c1b2 b8c6 e5c6 b7c6 c3c4 d8d7 a3a4 a8d8 d1e2 h7h5 h2h3 h5h4 a4a5 b6a5 d4d5 e6d5 b2f6 g7f6 e2g4 g8f8 g4h4 f8g7 h4g4 g7f8 g4f5 f8g7 f1d3 d8h8 e3e4 d5e4 d3e4 d7d4 e4c6');
INSERT INTO evaluations VALUES(37921,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(37921,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(37921,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(37921,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(37921,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(37921,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(37921,36,'cp 32','g8f6 f1e2 e8g8 e1g1 f8e8 a2a4 a7a5 c4d5 c6d5 d1b3 e7e6 f1d1 b8c6 c1d2 d8e7 c3b5 e8d8 d1c1 f6e4 d2e1 c8d7 c1c2 d8c8 a1c1 g7f8 e2d3 e4d6 b5d6 e7d6 d3b5 c6b4 c2c8 a8c8 c1c8 d7c8 f3e5 d6c7 b3c3 c7c3 b2c3');
INSERT INTO evaluations VALUES(38969,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(38969,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(38969,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38969,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(38969,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38969,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(38969,36,'cp 58','g8f6 g1f3 e8g8 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 c7c5 e5g6 h7g6 g4g5 d8d4 d1d4 c5d4 c3b5 f6h5 b5d4 b8d7 f1d1 d7e5 c4e2 f8e8 c1e3 b4c5 a1c1 c5b6 a2a3 a8d8 e2b5 e8f8 g1g2 d8d5 b5e2 f8d8 d4f3 e5d3 e2d3 b6e3 f2e3 d5d3 d1d3 d8d3 c1c8 g8h7');
INSERT INTO evaluations VALUES(42266,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(42266,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(42266,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(42266,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(42266,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(42266,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(42266,36,'cp 58','g1f3 e8g8 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 c7c5 e5g6 h7g6 g4g5 d8d4 d1d4 c5d4 c3b5 f6h5 b5d4 b8d7 f1d1 d7e5 c4e2 b4c5 c1e3 f8d8 a2a4 a8c8 g1f1 d8d5 b2b4 c5b4 d4f3 d5a5 a1b1 e5f3 b1b4 f3g5 b4b7 g5h3 b7a7 a5a7 e3a7 h3f4 a7b6 f4e2');
INSERT INTO evaluations VALUES(42265,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(42265,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(42265,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(42265,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(42265,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(42265,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(42265,36,'cp 47','e8g8 e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 c7c5 e5g6 h7g6 g4g5 d8d4 d1d4 c5d4 c3b5 f6h5 a2a3 b4a5 b5d4 b8d7 d4f3 a8d8 b2b4 a5c7 f1e1 d7b6 c4f1 b6d5 h3h4 h5f4 f1c4 d5b6 c4a2 f4e6 g1f1 b6d5 c1d2 f8e8 e1e4 c7f4 a1d1 f4d2 d1d2 d5f4');
INSERT INTO evaluations VALUES(37123,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(37123,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(37123,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(37123,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(37123,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(37123,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(37123,36,'cp 51','e1g1 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 c7c5 e5g6 h7g6 d4c5 b8d7 d1b3 d8a5 c3d5 f6d5 c4d5 a8d8 a2a3 b4c5 c1f4 c5d4 a1d1 d7c5 b3a2 c5e6 d5e6 f7e6 f4e3 a5e5 e3d4 d8d4 d1d4 e5d4 a2e6 g8h7 e6e2 a7a5 g1g2 d4d5 g2g3 d5b3 f2f3 b7b5 f1e1 f8d8 e2f2 d8d3');
INSERT INTO evaluations VALUES(37122,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(37122,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(37122,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(37122,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(37122,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(37122,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(37122,36,'cp 37','c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 c7c5 d4c5 b8d7 e5g6 h7g6 d1b3 d8a5 c3d5 f6d5 c4d5 a8d8 a2a3 b4c5 c1f4 c5d4 a1d1 a5b6 g1g2 b6b3 d5b3 d7c5 b3a2 c5e6 f4c1 d4f6 b2b4 g6g5 c1e3 b7b6 d1d8 f8d8 f1c1 f6b2 c1c6 b2a3 a2e6 f7e6 e3g5 d8d5 c6c8 g8h7');
INSERT INTO evaluations VALUES(47109,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(47109,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(47109,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(47109,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(47109,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(47109,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(47109,36,'cp -27','d8d5 f3d5 f6d5 b1c3 d5c3 d2c3 e7e5 c1e3 c8e6 a2a4 f7f6 f2f4 e5e4 g1e2 b8c6 a4a5 e8f7 h2h4 h7h5 e1f2 a7a6 g2g3 f8e7 e2d4 c6d4 c3d4 c7c6 f1e2 a8d8 a1a4 d8d7 h1e1 g7g6 e1f1 h8d8 f1d1 d8b8 b2b3 b8e8 c2c4');
INSERT INTO evaluations VALUES(50537,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(50537,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(50537,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50537,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(50537,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50537,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50537,36,'cp 13','d2d3 d7d5 g2g3 b7b5 f1g2 d5e4 d3e4 d8d1 e1d1 g8f6 d1e2 c8b7 e4e5 f6d5 a2a4 f8e7 c1e3 e8g8 a4b5 a6b5 a1a8 b7a8 b1d2 b8d7 h1a1 a8c6 f3e1 f8a8 a1a8 c6a8 b2b3 d5c3 e2d3 c3d5 g2f3 g7g5 c2c4 b5c4 b3c4 d5b4 d3e2 g5f4 g3f4 a8f3 e1f3 b4c6 d2e4 d7b6');
INSERT INTO evaluations VALUES(4032,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(4032,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(4032,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(4032,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(4032,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(4032,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(4032,36,'cp 31','e1g1 e7e5 e3e4 e5d4 f3d4 f8e8 f2f3 c7c6 g1h1 a7a6 c1g5 h7h6 g5h4 g6g5 h4f2 d7e5 a2a4 f6h5 g2g3 h5f6 d4c2 c8h3 f1e1 f6d7 d1d6 h3e6 d6d1 e5c4 c2d4 c4b2 d1c2 g7d4 f2d4 b2c4 a1d1 d8e7 f3f4 c6c5 d4f2 a8d8 f4f5 c4a3');
INSERT INTO evaluations VALUES(4031,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(4031,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(4031,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(4031,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(4031,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(4031,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(4031,36,'cp 31','e7e5 b2b3 f8e8 d4e5 d7e5 c1a3 d8e7 d1d2 e5d7 a1d1 d7c5 d2c1 c5e4 c3e4 f6e4 a3b2 g7b2 c1b2 b7b6 f3d4 c8b7 f2f3 e4c5 e3e4 e7e5 b2d2 a7a6 a2a4 e5e7 f1e1 e7e5 a4a5');
INSERT INTO evaluations VALUES(3947,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3947,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(3947,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(3947,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(3947,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3947,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(3947,36,'cp 23','b2b3 f8e8 d1c2 e5d4 f3d4 d7c5 c1b2 a7a5 a1d1 c7c6 f1e1 d8b6 a2a3 c8g4 e2g4 f6g4 h2h3 g4f6 d1b1 b6c7 d4f3 c7e7 b1d1 a8d8 b3b4 a5b4 a3b4 c5d7 c3a4 d7e5 f3e5 d6e5 b2c3 h7h5 d1d8 e8d8 e3e4 g7f8 a4b6 e7c7');
INSERT INTO evaluations VALUES(6830,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(6830,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(6830,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(6830,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(6830,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(6830,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(6830,36,'cp 62','c7c6 b5a4 a7a5 b1d2 d7d6 c2c4 g8f6 h2h3 c8d7 f2f4 e8g8 a4b3 d8e7 a2a3 b7b5 c4b5 c6b5 d2f3 a5a4 b3c2 c5b6 f1e1 f8c8 f4f5 d7e8 c2b1 c8c7 b1a2 a8c8 c1d2 f6d7 d2g5 e7f8 a2d5 h7h6 g5h4 c7c5 e1e2 d7e5 f3e5 d6e5 f5f6');
INSERT INTO evaluations VALUES(7025,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(7025,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(7025,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(7025,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(7025,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(7025,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(7025,36,'cp 67','d7d6 d4d5 c6e5 f3e5 g7e5 e2e4 e7e6 f1d3 g8f6 e1g1 f6d7 d3c2 e8g8 f2f4 e5g7 c1e3 b7b6 d1d2 c8a6 d5e6 f7e6 b2b3 d8f6 c3b5 a6b5 c4b5 f6c3 d2c3 g7c3 a1c1 c3b4 c2b1 d7c5 g2g3 a7a6 b5a6 c5a6 c1c6 b4c5 e3c5 b6c5 f4f5 a6b4 c6c7 b4a2 f5e6 f8f1 g1f1');
INSERT INTO evaluations VALUES(5693,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(5693,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(5693,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(5693,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(5693,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(5693,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(5693,36,'cp 55','d4d5 c6e5 f3e5 g7e5 e2e4 e7e6 f1e2 g8f6 e1g1 e6d5 c4d5 e8g8 c1h6 f6g4 e2g4 d8h4 f2f4 c8g4 d1d3 h4h6 f4e5 d6e5 d3g3 g4d7 f1f6 b7b5 h2h3 b5b4 c3d1 h6d2 d1e3 d2b2 a1f1 b2d4 g3f3 g8h8 g1h2 a7a5 f6f7 f8f7 f3f7 d4e3 f7d7 e3e4 d7f7');
INSERT INTO evaluations VALUES(7456,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(7456,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7456,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7456,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(7456,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(7456,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(7456,36,'cp 0','d7d6 c3d5 h7h6 c2c3 c8e6 d5f6 d8f6 e1g1 c6e7 c1e3 c5e3 f2e3 e8g8 a2a4 a8b8 h2h3 f6g6 b3e6 g6e6 a4b5 a6b5 a1a7 f8c8 f3h4 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(8319,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(8319,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(8319,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(8319,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(8319,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(8319,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(8319,36,'cp 15','f1e1 e8g8 c4b3 c6d4 d1d4 c8d7 d4d3 a8c8 h2h3 d7c6 c1e3 a7a6 e3d4 f6d7 d4g7 g8g7 a1d1 b7b5 c3e2 d7c5 d3e3 c6e4 e2c3 e4c6 e3e7 c5b3 a2b3 f8e8 e7d8 e8e1 d1e1 c8d8 c3e2 b5b4 e2d4 c6d7 e1d1 g7f6 c2c3 b4c3 b2c3');
INSERT INTO evaluations VALUES(13884,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(13884,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13884,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13884,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(13884,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(13884,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(13884,36,'cp 37','d4b3 f6g4 e3c1 g4f6 f2f3 c8e6 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5d3 c8c7 c2c3 d5e6 c1c2 b7b5 d3e2 f8e7 b3c5 e7c5 e3c5 d8d1 h1d1 h8d8 d1d8 c6d8 c5e3 f7f5 e2d3 g7g6 b2b3 f5f4 e3c5 d8f7');
INSERT INTO evaluations VALUES(19076,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19076,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19076,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19076,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19076,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19076,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19076,36,'cp 0','g8f6 f1c4 f8b4 c2c3 b4a5 e1g1 d7d6 f1d1 h7h6 h2h3 e8g8 c4b3 c8e6 b1a3 a5b6 e3e2 d8e7 b3d5 e6d7 d5b3');
INSERT INTO evaluations VALUES(12602,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(12602,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12602,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(12602,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(12602,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12602,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(12602,36,'cp 6','f1c4 f8b4 c2c3 b4a5 h2h3 e8g8 e3e2 h7h6 e1g1 d7d6 f1d1 d8c7 c1e3 c6e7 b1a3 c8d7 a3b5 d7b5 c4b5 d6d5 f3d2 f8d8 e4d5 e7d5 d2f1 a7a6 b5a4 d5f4 e3f4 e5f4 f1d2 b7b5 a4c2 g7g6 d2e4 f6e4 d1d8 a8d8 e2e4 g8g7 a1d1 d8d1 c2d1 a5b6 d1b3 b6c5 b3d5');
INSERT INTO evaluations VALUES(24702,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(24702,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24702,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24702,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24702,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(24702,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(24702,36,'cp 0','f1b5 d8b6 d1a4 a8c8 f3e5 a7a6 b5c6 b7c6 e1g1 h7h6 a4b3 b6a7 b3a4');
INSERT INTO evaluations VALUES(24699,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24699,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24699,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24699,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24699,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24699,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24699,36,'cp 0','d8b6 d1a4 a8c8 f3e5 a7a6 b5c6 b7c6 e1g1 h7h6 a4b3 b6a7 b3a4');
INSERT INTO evaluations VALUES(26573,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26573,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26573,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26573,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26573,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26573,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26573,36,'cp 0','e7e6 c1f4 f8d6 f4g3 g8f6 e2e3 c6e7 f1d3 c7c6 d1e2 f5g4 e2d2 g4f5');
INSERT INTO evaluations VALUES(26827,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(26827,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26827,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26827,36,'cp 0','c1f4 f8d6 f4g3 c6e7 e2e3 g8f6 f1d3 c7c6 d1e2 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 f5e4 e2e4 e7f5 g3d6 f5d6 e4e2 d8a5 c2c3 f8d8 e1g1 a5b5 e2c2 b5f5 c2e2 f5b5');
INSERT INTO evaluations VALUES(26818,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(26818,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26818,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26818,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(26818,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(26818,36,'cp -5','f8d6 f4g3 g8f6 e2e3 c6e7 f1d3 e8g8 g3h4 c7c5 h4f6 g7f6 d3f5 e7f5 d4c5 d6c5 e3e4 d5e4 c3e4 c5e7 d1e2 a8c8 c2c3 f5g7 e4g3 f6f5 e1g1 d8c7 f1d1 f8d8 d1d8 c8d8 a1d1 d8d1 e2d1 c7d6 d1c2 b7b5 g3f1 g7e8 c2e2 a7a6 g2g3 d6d5 f1e3');
INSERT INTO evaluations VALUES(26767,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(26767,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(26767,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(26767,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(26767,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(26767,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(26767,36,'cp -15','d6e5 d1e2 f8d6 c2c4 g8f6 b1c3 d6b4 c1d2 b4c3 d2c3 f6e4 f3e5 e4c3 b2c3 e8g8 e5c6 b7c6 e1g1 d8f6 f1e1 f6c3 a1c1 c3b4 g2g3 h7h6 c1d1 a8d8 d1d8 f8d8 e2e7 b4d6 e7d6 c7d6 e1e7 a7a5 e7a7 d8b8 a7a6 c6c5 a6d6 b8b4 d6a6 a5a4 g1g2 b4c4');
INSERT INTO evaluations VALUES(26764,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(26764,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(26764,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(26764,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(26764,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(26764,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(26764,36,'cp -13','d1e2 f8d6 c2c4 g8f6 b1c3 d6b4 c1d2 b4c3 d2c3 f6e4 f3e5 e4c3 b2c3 e8g8 e5c6 b7c6 e1g1 d8f6 f1e1 f6c3 e2e4 g7g6 e4c6 a8e8 e1d1 e8e7 a2a4 f8e8 h2h3 g8g7 a1c1 c3e5 c6a6 e5c5 a6b5 e7e5 g2g3 e8e6 b5b2 a7a5 d1d8 c5b6 c1b1 b6b2');
INSERT INTO evaluations VALUES(36973,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36973,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36973,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36973,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36973,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36973,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36973,36,'cp 0','b7b6 f1d3 c7c5 c2c3 c8b7 d1e2 f6e4 f3e5 e4d2 e2d2 b8c6 d2e2 d8c7 e5c6 b7c6 e2h5 f7f5 e1g1 g7g6 g3d6 c7d6 h5h6 e6e5 d4e5 d6e5 h2h4 a8e8 f1e1 e5f6 a1d1 f6g7 h6f4 g7e5 f4h6');
INSERT INTO evaluations VALUES(35651,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(35651,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(35651,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(35651,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(35651,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(35651,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35651,36,'cp 0','c2c3 c8b7 f1d3 b8d7 d1c2 d8c8 e3e4 d5e4 d2e4 f6e4 d3e4 b7e4 c2e4 d7f6 e4c6 f8d8 a2a4 d6g3 h2g3 d8d6 c6c4 d6d5 c4c6');
INSERT INTO evaluations VALUES(38181,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38181,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38181,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38181,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38181,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38181,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(38181,36,'cp 12','b8c6 e1g1 g8e7 b1d2 c8g4 d2b3 c5d6 h2h3 g4h5 f1e1 e8g8 c2c3 h7h6 c1e3 f8e8 b5e2 d6c7 b3d4 a7a6 d1c2 d8d7 f3h4 h5e2 e1e2 c6d4 e3d4 e7c6 h4f3 e8e2 c2e2 c6d4 f3d4 c7b6 e2d3 b6d4 d3d4 d7b5 b2b3 a8e8 a1d1 e8e2 d4d5 b5d5 d1d5');
INSERT INTO evaluations VALUES(5588,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5588,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5588,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5588,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5588,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5588,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5588,36,'cp 0','e1g1 g8e7 b1d2 c8g4 d2b3 c5d6 h2h3 g4h5 f1e1 e8g8 c2c3 f8e8 b5e2 d6c7 b3d4 a7a6 c1e3 d8d6 g2g3 d6d7 e2f1 h5g6 d1d2 g6e4 f3g5 e4g6');
INSERT INTO evaluations VALUES(5587,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5587,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5587,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5587,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5587,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5587,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5587,36,'cp 0','g8e7 b1d2 e8g8 d2b3 c5b6 f1e1 c8g4 h2h3 g4h5 b5f1 c6b4 d1d2 b4c6');
INSERT INTO evaluations VALUES(8229,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8229,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8229,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8229,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8229,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8229,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8229,36,'cp 0','b1d2 c5b6 d2b3 e8g8 f1e1 c8g4 b5f1 g4h5 a2a4 b6c7 c2c3 f8e8 g2g3 h7h6 d1d3 h5g6 d3d1');
INSERT INTO evaluations VALUES(38556,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(38556,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(38556,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(38556,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(38556,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(38556,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(38556,36,'cp 104','b1c3 g8f6 h2h3 e8g8 c1e3 b7b6 f1d3 e6e5 d1d2 b8c6 d4d5 c6b4 d3e2 a7a5 g2g4 c8d7 f3h2 c7c6 h3h4 c6d5 c4d5 b4a6 h4h5 b6b5 h5h6 g7h8 f2f3 b5b4 c3b5 f6e8 e1g1 h8f6 g1g2 d8b8 a2a4 a6c5 e3c5 d6c5 g4g5 f6e7 d5d6 e7d6 h2g4');
INSERT INTO evaluations VALUES(38868,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(38868,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38868,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(38868,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(38868,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(38868,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(38868,36,'cp 51','b4d2 b1d2 g8f6 a1c1 e8g8 g1f3 b7b6 e1g1 c8b7 f1e1 b8d7 e3e4 h7h6 e4e5 f6d5 d2e4 a8b8 c4b5 a7a6 b5c4 d5e7 d1e2 c7c5 e4d6 b7a8 c4d3 a8d5 h2h4 e7c8 d3a6 c8d6 e5d6 c5d4 f3d4 d8h4 d4c6');
INSERT INTO evaluations VALUES(38891,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(38891,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(38891,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(38891,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(38891,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(38891,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(38891,36,'cp 62','b1d2 g8f6 g1e2 e8g8 e1g1 b7b6 a1c1 c8b7 b2b4 b8d7 d1b3 e6e5 f1d1 a7a5 a2a3 d8e7 h2h3 a5b4 b3b4 e7b4 a3b4 e5d4 e2d4 g7g6 c4b5 c7c5 b4c5 b6c5 d4c6 b7c6 b5c6 a8a2 c6d7 f6d7 d2c4 f8d8');
INSERT INTO evaluations VALUES(38889,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(38889,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(38889,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(38889,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(38889,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(38889,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(38889,36,'cp 56','g8f6 g1e2 e8g8 e1g1 b8c6 c4b5 c8d7 e2c3 d8e7 b5e2 a8d8 a2a3 a7a6 d2e4 f6e4 c3e4 e6e5 d4d5 c6a7 d1b3 d7b5 f1e1 b5e2 e1e2 a7b5 e2c2 b5d6 e4d6 c7d6 a1c1 d8d7 c2c3 h7h6 a3a4 g8h7 h2h3 f7f5 a4a5 e7f7');
INSERT INTO evaluations VALUES(43468,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(43468,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(43468,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(43468,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(43468,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(43468,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43468,36,'cp 16','c1g5 b8d7 c3d5 h7h6 g5f6 d7f6 d5f6 d8f6 f1c4 f8e7 e1g1 e8g8 c4d5 a8b8 d1e2 c8g4 a2a4 g4f3 e2f3 f6f3 g2f3 f8c8 c2c3 e7g5 f1d1 b7b5 a4b5 b8b5 b2b3 c8c3 a1a6 c3f3 d5c4 b5b8 d1d6 f3f6 g1g2 f6d6 a6d6 g8f8 d6d7');
INSERT INTO evaluations VALUES(56526,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(56526,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(56526,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(56526,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(56526,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(56526,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(56526,36,'cp 77','c7c5 e2e3 b8c6 g1f3 g8e7 f1d3 e7g6 e3d4 c6d4 d3g6 h7g6 c1e3 c8e6 e3d4 c5d4 b1d2 d4d3 d1a4 d8d7 a4d7 e6d7 e1c1 d7c6 d1e1 f8e7 d2e4 e8c8 c1d2 f7f5 e4g5 e7g5 f3g5 c6g2 g5f7 g2h1 f7h8 d8h8 e1h1 h8h3 e5e6 g6g5 d2c3 h3f3 e6e7 d3d2 c3d2 f3f2 d2e3');
INSERT INTO evaluations VALUES(10866,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10866,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(10866,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(10866,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(10866,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(10866,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(10866,36,'cp -18','f6e4 f1e1 d7d5 e5d6 e4d6 c4d5 d6f5 d5c6 b7c6 g2g4 f5d6 f3d4 d8d7 c1g5 f7f6 g5f4 e8f7 d4e6 d6b5 b1d2 h8e8 g4g5 f7g8 d1h5 f6g5 e6g5 d7g4 h5g4 c8g4 c2c3 e7f8 d2c4 e8e1 a1e1 h7h6 g5e4 g4e6 b2b3 e6c4 b3c4 b5a3 f4c7 a3c4 c7f4 a8e8 f4g3');
INSERT INTO evaluations VALUES(10962,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(10962,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(10962,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(10962,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(10962,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(10962,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(10962,36,'cp -9','c4d5 e4c5 c2c3 d4c3 b1c3 e8g8 d1e2 d7d6 f1d1 d8e8 e5d6 e7d6 e2e8 f8e8 c3b5 c6e7 b5d6 c7d6 c1f4 e8d8 f4d6 e7d5 d6c5 b7b6 c5e7 d8d7 e7a3 c8b7 d1d4 f7f6 a1d1 g8f7 h2h3 a8c8 d4d2 c8d8 g1h2 b7a6 b2b3 a6b7 a3b2 h7h6 f3d4 d5b4');
INSERT INTO evaluations VALUES(12039,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12039,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12039,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12039,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12039,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12039,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12039,36,'cp 0','e7e6 e1g1 f8e7 d1e2 e8g8 f1e1 d6d5 c4d3 d8c7 c1d2 f8e8 a1d1 a7a5 e4e5 f6d7 c3a4 d7b6 b2b3 b6a4 b3a4 c7a7 d2e3 a7b8 e3d2');
INSERT INTO evaluations VALUES(12097,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12097,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12097,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12097,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12097,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12097,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12097,36,'cp 0','e1g1 f8e7 c1f4 d6d5 c4d3 e8g8 f1e1 a7a5 e4e5 f6d7 c3a4 c8a6 d3h7 g8h7 d1h5 h7g8 e1e3 f7f6 e3h3 f6e5 h5h8 g8f7 h8h5');
INSERT INTO evaluations VALUES(12096,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12096,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12096,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12096,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12096,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(12096,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12096,36,'cp 0','f8e7 d1e2 d6d5 c4d3 e8g8 f1e1 d8c7 e4d5 c6d5 c3b5 c7b6 c2c3 c8d7 a2a4 a7a6 b5d4 e7d6 d4f3 a6a5 f3e5 h7h6 e5d7 f6d7 c1h6 g7h6 e2g4 g8h8 g4h4 h8g7 h4g4');
INSERT INTO evaluations VALUES(7710,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7710,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7710,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7710,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7710,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7710,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7710,36,'cp 0','d1e2 e8g8 f1e1 d6d5 c4d3 d8c7 c1d2 c8b7 e4e5 f6d7 d2f4 a7a5 a1d1 g7g6 c3a4 d7b6 a4c3');
INSERT INTO evaluations VALUES(18573,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(18573,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(18573,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(18573,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(18573,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(18573,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(18573,36,'cp 175','f1d3 e7e6 c2c3 c6c5 g1e2 g8e7 e1g1 c5d4 c3d4 e7c6 b1d2 f7f6 f2f4 f6f5 d2f3 d7b6 a2a3 f8e7 g1h1 c8d7 h2h3 e8g8 f1g1 d7e8 g2g4 e8g6 e2g3 a8c8 g1g2 c6a5 g4f5 e6f5 h1h2 a5c6 b2b3 b6a8');
INSERT INTO evaluations VALUES(22536,30,'cp -158',NULL);
INSERT INTO evaluations VALUES(22536,31,'cp -157',NULL);
INSERT INTO evaluations VALUES(22536,32,'cp -176',NULL);
INSERT INTO evaluations VALUES(22536,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(22536,34,'cp -174',NULL);
INSERT INTO evaluations VALUES(22536,35,'cp -183',NULL);
INSERT INTO evaluations VALUES(22536,36,'cp -184','g7g5 d2d4 d7d6 a2a4 f8g7 a4a5 g8e7 c2c3 e8g8 b1d2 e7g6 a5a6 c6e7 a6b7 c8b7 b2b4 d6d5 c4d3 d5e4 d2e4 e7d5 c1d2 d5b4 c3b4 g5g4 g1h1 g4f3 g2f3 g7d4 d2c3 d4c3 e4c3 d8d4 d3e4 d4c3 e4b7 a8d8 d1e2');
INSERT INTO evaluations VALUES(22841,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(22841,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(22841,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(22841,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(22841,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(22841,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22841,36,'cp -37','d7d5 g1f3 f8d6 b1d2 g8f6 e2e4 e8g8 a2a3 f8e8 f1e2 a7a5 e1g1 h7h6 e4d5 f6d5 f1e1 c8g4 e2f1 d8d7 h2h3 g4h5 g2g3 f7f5 d2c4 d5b6 c4d6 c7d6 f1g2 a5a4 d1d2 a4b3 c2b3 h5f7 b3b4 c6e7 f3h4 b6a4 a1c1 a4b2 d2b2 a8d8');
INSERT INTO evaluations VALUES(21725,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(21725,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(21725,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(21725,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(21725,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(21725,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(21725,36,'cp -25','g1f3 f8d6 b1d2 a7a5 g2g3 g8f6 f1g2 e8g8 e1g1 d8e7 a2a3 f8d8 f3h4 h7h6 e2e4 c8g4 d1e1 d5e4 d2e4 f6e4 e1e4 e7d7 f1e1 d8e8 e4c4 g4e6 c4c3 a8d8 g2c6 b7c6 c3a5 f7f6');
INSERT INTO evaluations VALUES(21724,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(21724,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(21724,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(21724,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(21724,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(21724,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(21724,36,'cp -32','f8d6 b1d2 g8f6 e2e4 e8g8 f1e2 a7a5 a2a3 h7h6 e1g1 f8e8 e4d5 f6d5 f1e1 d5f4 e2f1 c8f5 d2e4 f5g4 h2h3 g4h5 e4g3 h5g6 g3e4 f7f6 c2c3 d6f8 g2g3 f4e6 b3b4 a5b4 a3b4 a8a1 b2a1 g6f7 f3h4 e6g5 e4g5 h6g5 h4f3 c6e7 d3d4 e7g6 d4e5 d8d1 e1d1 g6e5 f3e5 e8e5');
INSERT INTO evaluations VALUES(25721,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(25721,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(25721,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(25721,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(25721,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(25721,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25721,36,'cp 49','g7g6 e1g1 f8g7 f1e1 g8f6 h2h3 e8g8 a4c2 f8e8 d4d5 c6e7 a2a4 b7b5 c3c4 a8b8 a4b5 a6b5 c4b5 d7b5 b1c3 b5d7 f3d2 g7h6 d2c4 h6c1 d1c1 f6h5 c2d3 f7f5 d3f1 g8g7 a1a7 h5f4 b2b4');
INSERT INTO evaluations VALUES(29775,30,'cp 340',NULL);
INSERT INTO evaluations VALUES(29775,31,'cp 341',NULL);
INSERT INTO evaluations VALUES(29775,32,'cp 345',NULL);
INSERT INTO evaluations VALUES(29775,33,'cp 325',NULL);
INSERT INTO evaluations VALUES(29775,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(29775,35,'cp 345',NULL);
INSERT INTO evaluations VALUES(29775,36,'cp 341','d1b3 g4c8 e4e5 f6d7 e5e6 f7e6 g1f3 d7f6 f1c4 b8c6 e1g1 c8d7 f1e1 c6a5 b3b4 a5c4 b4c4 c7c6 f3g5 d8b6 c1f4 b6a6 c4b3 a6b6 b3d1 h7h6 g5e6 d7e6 e1e6 e8f7 e6e3 e7e6 e3g3 h8h7 f4e5 f8e7 d1c2 f7g8 c3a4 b6d8 a1e1 d8a5 a4c3 a5b6 a2a3 a8f8 c3a4 b6a5 b2b4 a5d5 a4c5');
INSERT INTO evaluations VALUES(46839,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(46839,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(46839,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(46839,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(46839,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(46839,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(46839,36,'cp 126','e2e4 d5b6 h2h3 f8g7 b1c3 e8g8 c1e3 f7f5 d1b3 g8h8 a1d1 f5e4 f3e5 d8e8 f1e2 c7c6 a2a4 b8d7 a4a5 d7e5 d4e5 b6d5 c3d5 c8e6 e2c4 b7b5 a5b6 c6d5 c4d5 e6d5 b3d5 a7b6 e1g1 a8a5 d5e4 a5e5 e4d4 e5c5 d4d2 c5c8 d2e2 h7h5 e3d4 e7e5 d4e3');
INSERT INTO evaluations VALUES(46843,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(46843,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(46843,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(46843,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(46843,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(46843,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(46843,36,'cp 118','d5b6 h2h3 f8g7 b1c3 b8c6 f1b5 e8g8 c1e3 c6a5 d1c1 c7c6 b5e2 a5c4 e3h6 e7e5 h6g7 g8g7 e1g1 e5d4 f1d1 d8f6 d1d4 c8e6 e2c4 b6c4 b2b3 c4e5 f3e5 f6e5 c1d2 f7f6 a1d1 a8e8 b3b4 e5c7 b4b5 e8e7 b5c6 c7c6 d1c1 f8c8 c3e2 c6b5 c1c8 e6c8');
INSERT INTO evaluations VALUES(46621,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(46621,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(46621,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(46621,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(46621,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(46621,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(46621,36,'cp 129','h2h3 f8g7 b1c3 e8g8 c1e3 f7f5 d1b3 g8h8 e1c1 d8e8 e4f5 c8f5 g2g4 f5d7 f3e5 a7a5 e5d7 b8d7 a2a4 e7e6 h3h4 d7f6 h4h5 g6h5 g4h5 f6d5 h5h6 g7f6 c3d5 e6d5 f1b5 e8f7 b5d3 b6c4 c1b1 a8e8 h1g1 c4e3 f2e3 c7c6 d1e1 f7e7 e3e4 f6d4 e4d5');
INSERT INTO evaluations VALUES(47140,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47140,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(47140,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(47140,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47140,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(47140,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47140,36,'cp 0','d5d4 g1f3 b8c6 b2b4 c6b4 d1a4 b4c6 f3e5 c8d7 e5c6 b7c6 a4a5 e7e6 e1g1 f8d6 d2d3 e8g8 c1g5 a8b8 b1d2 d6b4 a5a7 h7h6 a2a3 b4c3 g5f6 g7f6 a1a2 b8a8 a7b7 f6f5 d2b1 a8b8 b7a7 b8a8');
INSERT INTO evaluations VALUES(57672,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57672,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(57672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57672,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(57672,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(57672,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(57672,36,'cp 4','c7c5 f1g2 f8g7 g1e2 b8c6 e1g1 d7d6 d2d3 e7e6 c2c3 g8e7 c1g5 e8g8 d3d4 c5d4 e2d4 c6d4 c3d4 h7h6 g5e3 d6d5 b1c3 d5e4 g2e4 e7d5 c3d5 e6d5 e4g2 c8e6 f1e1 a8c8 d1d2 h6h5 h2h4 f8e8 a2a4 d8d7 a4a5 g8h7 a1c1 b7b6 a5b6');
INSERT INTO evaluations VALUES(52310,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52310,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(52310,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(52310,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52310,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(52310,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(52310,36,'cp 1','f1g2 f8g7 g1e2 b8c6 e1g1 d7d6 d2d3 e7e6 c2c3 g8e7 d3d4 e8g8 c1g5 c5d4 e2d4 c6d4 c3d4 h7h6 g5e3 d6d5 b1c3 d5e4 g2e4 e7d5 c3d5 e6d5 e4g2 c8e6 d1d2 h6h5 h2h4 a8c8 a1c1 d8d7 b2b3 b7b6 e3g5 c8c1 f1c1 f8c8 c1c8 d7c8 g5e3 c8c6 e3g5');
INSERT INTO evaluations VALUES(52311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52311,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52311,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52311,36,'cp 0','b8c6 g1e2 f8g7 e1g1 d7d6 d2d3 e7e6 c2c3 g8e7 c1g5 e8g8 d3d4 c5d4 e2d4 c6d4 c3d4 h7h6 g5e3 d6d5 b1c3 d5e4 g2e4 e7d5 c3d5 e6d5 e4g2 c8e6 d1d2 h6h5 h2h4 d8d7 b2b3 a8c8 a1c1 b7b6 f1e1 f8e8 e3f4 c8c1 e1c1 e8c8 c1c8 d7c8 f4e3 c8c6 e3g5 c6d7 a2a4');
INSERT INTO evaluations VALUES(19288,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19288,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19288,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19288,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19288,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19288,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19288,36,'cp 0','g1e2 f8g7 e1g1 d7d6 d2d3 e7e6 c2c3 g8e7 d3d4 e8g8 c1g5 c5d4 e2d4 c6d4 c3d4 h7h6 g5e3 d6d5 b1c3 d5e4 g2e4 e7d5 c3d5 e6d5 e4g2 c8e6 a1c1 h6h5 h2h4 d8d7 b2b3 f8e8 d1d2 a8c8 f1e1 b7b6 e3g5 c8c1 e1c1 e8c8 c1c8 d7c8 g1h2 c8f8 g5f4');
INSERT INTO evaluations VALUES(19289,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(19289,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(19289,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(19289,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19289,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19289,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19289,36,'cp 4','f8g7 e1g1 d7d6 d2d3 a8b8 c2c3 e7e5 b1d2 g8e7 d2f3 e8g8 d3d4 c5d4 c3d4 e5d4 f3d4 c6d4 e2d4 e7c6 c1e3 d8f6 d4c6 b7c6 a1c1 b8b2 c1c6 b2a2 c6d6 f6e7 e3c5 a2a5 d1c1 c8a6 d6c6 a5c5 c1c5 e7c5 c6c5 a6f1 g2f1 g7d4 c5d5 d4b6 e4e5 f8d8 d5d8 b6d8 f1c4 g8g7 c4b3');
INSERT INTO evaluations VALUES(5364,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5364,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5364,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5364,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5364,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5364,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5364,36,'cp 0','e1g1 d7d6 d2d3 a8b8 c2c3 e7e5 b1d2 g8e7 d2f3 e8g8 d3d4 c5d4 c3d4 c6d4 e2d4 e5d4 f3d4 e7c6 c1e3 d8f6 d4c6 b7c6 a1c1 b8b2 c1c6 b2a2 c6d6 f6e7 d6d5 f8e8 e3c5 e7c7 c5d6 c7b6 d6c5');
INSERT INTO evaluations VALUES(7665,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(7665,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(7665,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(7665,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(7665,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(7665,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(7665,36,'cp -59','d7d5 c4d3 e8g8 e1g1 b4c3 d2c3 d5e4 d3c4 d8e7 d1d2 f8e8 a1d1 c8g4 c4e2 e7e6 h2h3 g4f5 d2d6 f6d5 d6e6 e8e6 c3d2 e6g6 e2h5 g6d6 c2c4 d5f4 d2f4 d6d1 f1d1 e5f4 h5g4 f5g4 h3g4 g7g6 d1d6 a8e8 g1f1 e8e6 d6d8 g8g7 d8a8 g7f6 a8a7 e6e8 a7c7 e8d8 c7c6 f6e5');
INSERT INTO evaluations VALUES(10017,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(10017,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(10017,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(10017,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(10017,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(10017,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10017,36,'cp 0','h7h6 g5h4 c6e7 d3d4 e5d4 e4e5 f6d5 e1g1 a7a6 b5c4 d5f4 c2c3 d7d5 e5d6 c5d6 c3d4 e8g8 b1c3 b7b5 c4b3 c8e6 f1e1 e6b3 d1b3 d8d7 h4e7 d7g4 g2g3 f4h3 g1g2 h3f4');
INSERT INTO evaluations VALUES(9726,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9726,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9726,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9726,36,'cp 0','g5h4 c6e7 d3d4 e5d4 e4e5 f6d5 e1g1 a7a6 b5c4 d5f4 c2c3 d7d5 e5d6 c5d6 c3d4 e8g8 f1e1 b7b5 c4b3 c8e6 b1c3 e6b3 d1b3 d8d7 h4e7 d7g4 g2g3 g4f3 g3f4 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(9727,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9727,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9727,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9727,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9727,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9727,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9727,36,'cp 0','c6e7 d3d4 e5d4 e4e5 f6d5 e1g1 a7a6 b5c4 d5f4 c2c3 d7d5 e5d6 c5d6 c3d4 e8g8 f1e1 b7b5 c4b3 c8e6 b1c3 e6b3 d1b3 d8d7 h4e7 d7g4 g2g3 g4f3 g3f4 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(10104,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(10104,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(10104,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(10104,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(10104,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(10104,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(10104,36,'cp -13','d7d6 d2d3 a7a5 e1g1 e8g8 a2a4 c8e6 b1a3 h7h6 c4e6 f7e6 c1e3 c5e3 f2e3 d8e8 a3b5 f8f7 h2h3 e8d7 f3h2 a8f8 h2g4 f6g4 e2g4 c6d8 c3c4 b7b6 g4h5 d8b7 d3d4 f7f1 a1f1 f8f1 g1f1 e5d4 e3d4 c7c6 b5c3 g8h7');
INSERT INTO evaluations VALUES(9341,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9341,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9341,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9341,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9341,36,'cp 0','a2a4 a7a5 d2d3 h7h6 e1g1 e8g8 h2h3 c8d7 c1e3 c5e3 f2e3 c6e7 f3h4 c7c6 b1d2 b7b5 c4b3 b5a4 a1a4 d6d5 e2f2 d8c7 a4a1 d7e6 f2g3 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(10353,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(10353,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(10353,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(10353,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(10353,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(10353,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(10353,36,'cp -35','c3d5 b4a5 d5f6 d8f6 c2c3 d7d6 a2a4 a5b6 d2d3 h7h6 f1e2 c8e6 e1g1 e8g8 b3b4 a7a5 b4b5 c6e7 d3d4 e7g6 h2h3 a8d8 c1e3 g6f4 f1e1 f4e2 d1e2 f6g6 d4e5 g6e4 e3b6 e4e2 e1e2 c7b6 a1e1 e6c4 e2d2 d6e5 e1e5 d8d2 f3d2 c4e6 f2f4 f8d8 f4f5 d8d2 f5e6 f7e6 e5e6');
INSERT INTO evaluations VALUES(13311,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(13311,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(13311,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13311,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(13311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13311,36,'cp 0','g7g6 g1e2 f8g7 e1g1 e8g8 h2h3 a8b8 a2a4 a7a6 f2f4 c8d7 g3g4 b7b5 a4b5 a6b5 d1e1 c6b4 e1d2 b4c6');
INSERT INTO evaluations VALUES(13018,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(13018,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(13018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13018,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13018,36,'cp 0','h2h3 a8b8 a2a4 a7a6 c3d5 f8g7 c2c3 e8g8 g1f3 f6d5 e4d5 c6e5 f3e5 d6e5 a4a5 b7b5 a5b6 d8b6 e1g1 e7e6 d5e6 c8e6 d1a4 f8d8 a4a6 b6a6 a1a6 d8d3 g2e4 d3d7 e4c6 d7d3');
INSERT INTO evaluations VALUES(14697,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(14697,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(14697,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(14697,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(14697,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(14697,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(14697,36,'cp -29','b7c6 e1g1 f8e7 h2h3 e8g8 b1c3 c8b7 f3h2 c6c5 d1e2 d6d5 c1g5 d5d4 g5f6 e7f6 c3b1 d8e7 b1d2 g7g6 d2c4 f6g7 h2f3 f7f6 e2e1 b7c8 b2b3 g7h6 e1a5 c8e6 a5a4 g8h8 a4a5 e6c4 b3c4 h6f4');
INSERT INTO evaluations VALUES(24664,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24664,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(24664,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24664,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24664,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(24664,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24664,36,'cp 6','h2h3 g4d7 c3d5 e7e6 d5f6 d8f6 a2a4 f8e7 e1g1 h7h6 c2c3 e8g8 d1e2 a8d8 g1h1 f6g6 f1e1 e6e5 a1b1 a7a5 c1d2 d7e6 b2b4 c5b4 c3b4 e6c4 b4b5 c4b5 a4b5 c6b4 d2b4 a5b4 b1b4 d6d5 b4b1 g6e6 e4d5 d8d5 f3e5 f8d8');
INSERT INTO evaluations VALUES(24662,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24662,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24662,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(24662,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(24662,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(24662,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(24662,36,'cp 0','g4d7 c3d5 e7e6 d5f6 d8f6 a2a4 f8e7 c2c3 h7h6 e1g1 e8g8 f3h2 f6g6 f2f4 a8e8 h2f3 f7f5 e4f5 g6f5 d3d4 c5d4 f3d4 f5c5 c4b3 c6d4 c3d4 c5f5 b3c2 f5d5 c2b3');
INSERT INTO evaluations VALUES(30708,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30708,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(30708,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(30708,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(30708,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(30708,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(30708,36,'cp 25','c2c4 c7c5 b1c3 f8e7 d2d4 c5d4 d1d4 g8f6 e1g1 d7d6 e2e4 e8g8 f1d1 b8d7 f3e1 a7a6 b2b3 a8c8 c1a3 d7c5 d4e3 d8c7 a1c1 f8d8 e3e2 c7b8 a3b2 d8e8 e1d3 c5d3 e2d3 e8d8 d3e2 f6d7 b2a3 b7c6 f2f4 d7c5');
INSERT INTO evaluations VALUES(30640,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30640,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(30640,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(30640,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(30640,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(30640,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(30640,36,'cp 38','g8f6 d2d4 f8b4 c1d2 b4d2 d1d2 e8g8 b1c3 d7d5 c4d5 e6d5 e1g1 b8d7 b2b3 f8e8 a1c1 c7c6 d2b2 a7a5 e2e3 h7h6 c1c2 d8e7 f1c1 b7a6 c3a4 a6b5 f3e5 d7e5 d4e5 e7e5 b2e5 e8e5 a4b6 a8b8 b6a4 b5a4 b3a4');
INSERT INTO evaluations VALUES(28706,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(28706,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(28706,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(28706,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(28706,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(28706,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(28706,36,'cp 36','d2d4 f8b4 c1d2 b4d2 d1d2 e8g8 b1c3 d7d5 c4d5 e6d5 e1g1 b8d7 a1c1 c7c6 c1c2 f8e8 f1c1 d8e7 e2e3 f6e4 d2d1 d7f6 f3e5 c6c5 b2b3 e4c3 c2c3 a8c8 g2h3 c8d8 c3c2 g7g6 h3g2 d8c8 e5f3 a7a5 d4c5 b6c5');
INSERT INTO evaluations VALUES(34139,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(34139,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(34139,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(34139,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(34139,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(34139,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(34139,36,'cp 70','b1c3 g8f6 a2a3 b7b6 c1e3 c8b7 f2f3 f8e7 f1e2 e8g8 e1g1 b8c6 b2b4 c6e5 a1c1 e5g6 g1h1 d7d6 d1d2 h7h6 f1d1 f8c8 d2e1 a8b8 e1f2 b7a8 e2f1 f6d7 f2d2 d7f6 b4b5 a6a5 c3d5 e6d5 c4d5 c7d8 d4c6');
INSERT INTO evaluations VALUES(34138,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(34138,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(34138,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(34138,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(34138,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(34138,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(34138,36,'cp 60','g8f6 a2a3 b7b6 c1e3 c8b7 f2f3 f8e7 f1e2 e8g8 e1g1 d7d6 d1e1 b8d7 a1c1 a8c8 e1g3 d7e5 f1d1 c7b8 b2b3 f8e8 c3a4 b7a8 b3b4 e5g6 b4b5 a6b5 d4b5 c8c6 g1h1 f6d7 a4c3 d7f6 g3f2 g6e5 f3f4 e5c4');
INSERT INTO evaluations VALUES(33523,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(33523,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(33523,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(33523,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(33523,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(33523,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(33523,36,'cp 65','a2a3 b7b6 c1e3 c8b7 f2f3 f8e7 f1e2 b8c6 a1c1 e8g8 b2b4 a6a5 d4b5 c7b8 b4a5 c6a5 c3a4 b7c6 c1b1 a8a6 c4c5 b6c5 b5c3 b8a7 e2a6 a7a6 a4c5 a6c4 d1d3 d7d5 d3c4 a5c4 e3d4 d5e4 c3e4 f6e4 c5e4 e7a3');
INSERT INTO evaluations VALUES(40917,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(40917,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40917,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40917,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(40917,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40917,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(40917,36,'cp 0','e6d5 c1f4 c7c6 d1c2 g7g6 e2e3 c8f5 f1d3 f5d3 c2d3 e8g8 h2h3 f8e8 e1g1 e7f8 d3c2 b8d7 f1d1 a7a5 a1c1 d8b6 a2a3 f6e4 c3e4 e8e4 a3a4 e4e8 f3e1 d7f6 e1d3 f6h5 f4h2 h5f6');
INSERT INTO evaluations VALUES(42395,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(42395,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(42395,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42395,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(42395,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(42395,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(42395,36,'cp 56','d3d4 f6e4 f1d3 d6e7 f3e5 e4f6 e1g1 e8g8 h2h3 d7d6 e5f3 c7c5 f1e1 c5d4 b1d2 b8c6 d2b3 d6d5 f3d4 c8d7 a2a4 f8e8 a4a5 e7d6 e1e8 d8e8 a5a6 b7a6 a1a6 h7h6 c1e3 f6e4 d3e4 d5e4 c2c3');
INSERT INTO evaluations VALUES(45107,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(45107,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(45107,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(45107,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(45107,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(45107,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(45107,36,'cp 60','d2d4 g7g6 c1e3 f8g7 h2h3 e8g8 a2a4 b8d7 f1e2 e7e5 d4e5 d7e5 e1g1 e5f3 e2f3 c8e6 d1d2 d8c7 f1d1 d6d5 e4d5 f6d5 c3d5 e6d5 f3d5 c6d5 c2c3 f8d8 d2e2 c7c6 d1d2 d8d7 a1d1 a8d8 e2b5 a7a6 b5b4 g7f8 b4b3 f8c5 e3g5 d8e8 d2d5 d7d5 b3d5 c6a4');
INSERT INTO evaluations VALUES(48088,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(48088,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(48088,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(48088,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(48088,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(48088,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(48088,36,'cp 25','c3d5 e6d5 c2c3 f8e7 f1d3 e8g8 g1e2 b8d7 e1g1 f8e8 f1e1 e7d6 c1f4 d8f6 f4d6 f6d6 d1d2 d7f6 e2f4 c8d7 f2f3 h7h6 h2h4 e8e1 a1e1 a8e8 e1e8 f6e8 g1f2 c7c6 g2g4 g8f8 f4g2 a7a5 d3c2 b7b6 a2a3 d6h2 d2f4 h2f4');
INSERT INTO evaluations VALUES(48074,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(48074,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48074,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(48074,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(48074,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(48074,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(48074,36,'cp 26','e6d5 c2c3 f8e7 f1d3 e8g8 g1e2 f8e8 e1g1 b8d7 f1e1 e7d6 c1f4 d8f6 g2g3 d7b6 f4d6 f6d6 d1d2 c8g4 e2f4 h7h6 h2h3 g4d7 h3h4 e8e1 a1e1 a8e8 e1e8 d7e8 h4h5 c7c6 g1g2 e8d7 b2b3 d6e7 a2a4 a7a5 d3c2 e7g5 d2d3 b6c8 d3h7 g8f8');
INSERT INTO evaluations VALUES(50635,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(50635,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(50635,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(50635,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50635,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(50635,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(50635,36,'cp 52','e2e4 d5e4 c3e4 g8f6 e4f6 e7f6 a2a3 c6c5 d4d5 f8d6 d1e2 e8f8 c1e3 c8f5 g1f3 b8d7 e2d2 h7h5 f1d3 f5d3 d2d3 g7g6 e1g1 f8g7 b2b3 d8c7 h2h3 h8e8 f1e1 d7e5 f3e5 e8e5 e3d2 a8e8 g1f1 h5h4 e1e5 d6e5 a1d1');
INSERT INTO evaluations VALUES(51360,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(51360,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(51360,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(51360,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(51360,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(51360,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(51360,36,'cp 62','c4d5 e6d5 g1f3 g8f6 f3e5 c7c6 g2g3 f8b4 f1g2 e8g8 e1g1 h7h6 d1b3 b4c3 b3c3 f6e4 c3c2 c8f5 e5c6 d8d6 c6b8 e4g3 e2e4 g3f1 e4f5 f1h2 b8d7 f8c8 c2d3 h2g4 d3e2 d6h2 g1f1 h2h5 d7e5 g4h2 f1e1 h5f5 c1e3 c8c2 e2d3 f5d3 e5d3 a8d8 e3f4 h2g4 a1d1 f7f6');
INSERT INTO evaluations VALUES(55943,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(55943,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55943,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55943,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55943,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55943,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55943,36,'cp 0','g2g3 g8f6 f1g2 f8e7 e1g1 e8g8 c2c4 c7c5 d3d4 d5c4 d4c5 e7c5 b1d2 c4c3 d2b3 c5e7 b2c3 b8c6 f3d4 c6e5 d1c2 e5c4 c2d3 c4e5');
INSERT INTO evaluations VALUES(55944,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55944,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55944,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55944,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55944,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55944,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55944,36,'cp 0','f8e7 f1g2 g8f6 c2c4 e8g8 e1g1 c7c5 d3d4 d5c4 d4c5 e7c5 b1d2 c4c3 d2b3 c5e7 b2c3 b8c6 f3d4 c6e5 d1c2 e5c4 c2d3 c4e5');
INSERT INTO evaluations VALUES(3061,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(3061,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(3061,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(3061,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(3061,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(3061,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(3061,36,'cp -10','d8b6 c4b3 f6e4 c3d5 b6a5 c2c3 e4c5 d4c6 d7c6 d5e7 g8h8 e7c8 a8c8 e1g1 f8e8 e3f2 c8d8 d1c2 a5b5 f1e1 h8g8 f2c5 b5c5 c2f2 c5f2 g1f2 d8d2 f2f1 e8e1 a1e1 d2b2 e1e8 g7f8 e8a8 a7a6 a8a7 b2f2 f1e1 f2g2 a7b7 g8h8 b7f7 f8g7 h2h4 g7c3 e1f1 g2h2 f7c7 a6a5 c7c6 c3b4 c6c8 h8g7');
INSERT INTO evaluations VALUES(2195,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(2195,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(2195,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(2195,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(2195,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(2195,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(2195,36,'cp -9','c4b3 f6e4 c3d5 b6a5 c2c3 e4c5 d4c6 d7c6 d5e7 g8h8 e7c8 a8c8 e1g1 f8e8 e3f2 c8d8 d1c2 a5b5 f1d1 h8g8 d1d8 e8d8 a1d1 d8e8 f2c5 b5c5 c2f2 c5e7 g2g3 g7h6 f3f4 a7a5 g1f1 e7e4 f2d4 e4e2 f1g1 h6f8 d4c4 e2c4 b3c4 b7b5 c4f1 f8c5 g1h1 a5a4 h1g2 g8g7 d1d7 e8e3 d7c7');
INSERT INTO evaluations VALUES(2196,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(2196,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(2196,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(2196,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(2196,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(2196,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(2196,36,'cp -20','f6e4 c3d5 b6a5 c2c3 e4f6 d4c6 d7c6 d5e7 g8h8 e7c8 a8c8 e1g1 c6c5 e3f2 c5c4 b3c2 f8d8 d1c1 h8g8 a2a4 a7a6 f1e1 a5c7 e1d1 d8d1 c1d1 c8d8 d1e2 g7f8 g1h1 d8e8 e2f1 b7b5 f2d4 f8g7 a1e1 e8d8 d4g1 h7h6 e1d1 d8d7 g2g3 d7d1 f1d1 c7c6 h1g2');
INSERT INTO evaluations VALUES(2197,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(2197,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(2197,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(2197,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(2197,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(2197,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(2197,36,'cp -17','c3d5 b6a5 c2c3 e4f6 d4c6 d7c6 d5e7 g8h8 e7c8 a8c8 e1g1 c6c5 a2a4 c5c4 b3c2 f8d8 d1e2 h8g8 a1d1 f6d5 e3c1 b7b5 a4b5 a5b5 e2f2 a7a5 f3f4 d5f6 d1d8 c8d8 f2f3 f6d7 f1d1 g7f6 d1d5 d7c5 g1f1 d8d7 d5d7 b5d7 f4f5 g8g7 c1e3 c5d3 f5g6 h7g6');
INSERT INTO evaluations VALUES(5848,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(5848,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(5848,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(5848,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(5848,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(5848,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(5848,36,'cp 127','g1e2 g8f6 a2a3 b4a5 b2b4 a5b6 d1b3 e8g8 e1g1 e6e5 d4d5 c6e7 a3a4 a7a6 e2g3 e7f5 g3f5 c8f5 c1a3 d8d6 a4a5 b6a7 c4e2 f6e4 b4b5 a7c5 a3c5 e4c5 b3c4 h7h6 f1c1 f8b8 h2h3 e5e4 c3a4 c5a4 a1a4 a6b5 c4b5 d6e7 e2f1');
INSERT INTO evaluations VALUES(7061,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(7061,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(7061,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(7061,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(7061,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(7061,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(7061,36,'cp 69','e7f6 a3c4 f6e7 c3d5 c8e6 c4b6 a8b8 c2c3 e8g8 f1c4 g8h8 e1g1 d8e8 c4b3 c6a5 b3a4 a5c6 d1e2 e7d8 a1d1 d8b6 d5b6 e8d8 b6d5 b7b5 a4b3 c6a5 d1d2 a5b3 a2b3 a6a5 f1d1 h7h6 f2f3 a5a4 b3b4 d8d7 h2h3 b8d8 d5e3');
INSERT INTO evaluations VALUES(7105,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(7105,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(7105,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(7105,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(7105,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(7105,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(7105,36,'cp 56','a3c4 f6e7 c3d5 c8e6 c2c3 e8g8 c4b6 a8b8 f1c4 g8h8 e1g1 d8e8 d1e2 e7d8 c4b3 c6a5 b3a4 a5c6 a1d1 d8b6 d5b6 e8d8 b6d5 b7b5 a4b3 c6a5 d1d2 h7h6 f1d1 a5b3 a2b3 a6a5 b3b4 a5a4 f2f3 d8d7 h2h3 d7b7 d5e3 e6b3 d1e1 b8d8 g1h2 d8d7 h2h1 d6d5 e3d5 b3d5 d2d5');
INSERT INTO evaluations VALUES(8149,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(8149,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(8149,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(8149,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(8149,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(8149,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(8149,36,'cp -58','h7h6 g5f4 a7a6 e4e5 f6h5 e5d6 h5f4 d6e7 d8c7 c3e2 b7b5 e2f4 c7f4 c4b3 c8b7 d1e1 a8c8 e1e3 f4d6 a1e1 c6e7 f3e5 c8d8 f2f4 e8g8 c2c3 b7d5 b3d5 e7d5 e3e4 f7f6 e5g6 f8e8 f1f3 b5b4 c3c4 d5e7 g6e7 e8e7 f3e3 a6a5 a2a4 e6e5 f4f5 g8h8 g2g4');
INSERT INTO evaluations VALUES(8389,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(8389,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8389,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8389,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8389,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8389,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(8389,36,'cp 26','f4g3 d6g3 h2g3 d8b6 d1c2 h7h6 e1g1 e8g8 b1d2 c8d7 d4c5 b6c5 e3e4 a8c8 a2a3 c5b6 a1e1 b6c7 e4d5 e6d5 e1e2 c8e8 f1e1 e8e2 e1e2 f8d8 b2b4 d8e8 e2e8 d7e8 d2f1 c6e5 f3e5 c7e5 c2d2 e8d7 d3e2 e5c7 f1e3');
INSERT INTO evaluations VALUES(8398,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(8398,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(8398,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(8398,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8398,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(8398,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(8398,36,'cp 15','d6g3 h2g3 d8b6 d1c2 h7h6 e1g1 e8g8 b1d2 a7a5 a2a4 f8d8 d4c5 b6c5 e3e4 c5b6 a1e1 b6c7 e4d5 f6d5 e1d1 b7b6 d2e4 c8b7 c2e2 c6e5 f3e5 c7e5 f1e1 d5f6 e4f6 e5f6 d3e4 b7e4 e2e4 d8d1 e1d1 a8d8');
INSERT INTO evaluations VALUES(14009,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(14009,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(14009,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(14009,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(14009,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(14009,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(14009,36,'cp 169','d4d3 a7a6 a2a4 f8e7 c1f4 e8g8 e1g1 f6h5 f4e3 h5f6 a1d1 c8e6 b2b3 d8d7 d3e2 d7c6 e3f4 f8e8 d1d2 f6d7 h2h3 a8d8 f1d1 e6c4 b3c4 h7h6 d2d3 d7f8 a4a5 f8e6 f4e3 e7g5 c3d5 g5e3 d3e3 e6d4');
INSERT INTO evaluations VALUES(14008,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(14008,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(14008,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(14008,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(14008,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(14008,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(14008,36,'cp 148','f8e7 c1f4 a7a6 a2a4 e8g8 e1g1 c8e6 b2b3 f8e8 a1d1 d8b6 d1d2 a8d8 f1d1 e6c4 b3c4 f6h5 f4e3 h5f6 d1b1 b6c6 a4a5 c6c7 b1b6 d8d7 h2h3 c7c8 b6b1 h7h6 b1a1 c8c6 f2f3 e7d8 e3f2 f6h5 c3d5 d8g5');
INSERT INTO evaluations VALUES(15582,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15582,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15582,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15582,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15582,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15582,36,'cp -7','f8e7 d3d4 e5d4 f3d4 e8g8 f1e2 d6d5 c3d5 f6d5 e4d5 c6d4 e3d4 d8d5 e1g1 f8d8 c2c3 d5g5 d1c1 g5c1 f1c1 c8e6 a2a4 a7a5 c1e1 d8d7 e2b5 c7c6 b5e2 e7d6 g2g3 g7g6 d4e3 d6c7 e1d1 d7d1 a1d1 g8f8 g1g2');
INSERT INTO evaluations VALUES(19020,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(19020,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(19020,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(19020,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19020,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(19020,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(19020,36,'cp 32','d2d3 d7d6 b1c3 g8e7 f2f4 e5f4 c1f4 e7g6 f4e3 f8e7 g1e2 e8g8 e1g1 e7f6 d1d2 a8b8 a1b1 c8e6 a2a4 a7a6 c3d5 f6d4 e3d4 c6d4 e2f4 b7b5 a4b5 a6b5 c2c3 g6f4 d5f4 d4c6 d3d4 c5d4 f4e6 f7e6 f1f8 d8f8');
INSERT INTO evaluations VALUES(19014,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(19014,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(19014,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(19014,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19014,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(19014,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19014,36,'cp 19','g8e7 b1c3 d7d6 f2f4 e5f4 c1f4 e7g6 f4e3 f8e7 g1e2 e7f6 e1g1 e8g8 d1d2 c8e6 a2a4 f6e5 e2f4 d8d7 a1d1 g6f4 g3f4 e5c3 b2c3 f7f5 e4e5 b7b6 a4a5 d6e5 a5b6 a7b6 f4e5 a8a2 d3d4 c5d4 c3d4 e6d5');
INSERT INTO evaluations VALUES(25977,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(25977,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(25977,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25977,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(25977,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(25977,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(25977,36,'cp 25','c1e3 g8f6 d1e2 f8e7 e1c1 e8g8 g2g4 c6d4 e3d4 a7a6 b5d7 f6d7 c1b1 b7b5 f2f4 b5b4 c3a4 d8a5 b2b3 a5b5 e2e3 a8c8 h2h4 b5c6 e3e2 e6e5 f4e5 d6e5 d4b2 d7f6 h1e1 f8d8 d1d8 c8d8 g4g5 f6d7 c2c3 a6a5 c3b4 a5b4 e1c1 d7c5 a4c5 e7c5 e2c4');
INSERT INTO evaluations VALUES(25975,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(25975,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(25975,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(25975,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(25975,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(25975,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(25975,36,'cp 20','g8f6 d1e2 f8e7 e1c1 e8g8 g2g4 c6d4 e3d4 a7a6 b5d7 f6d7 f2f4 b7b5 c1b1 b5b4 c3a4 d8b8 b2b3 b8b5 e2e3 a8c8 h2h4 b5c6 e3e2 e6e5 f4e5 d6e5 d4b2 d7f6 h1e1 f8d8 d1d8 c8d8 g4g5 f6d7 e2g4 d7c5 a4c5 c6c5 g4e2 a6a5 e1h1 c5c6 b2e5 a5a4');
INSERT INTO evaluations VALUES(24440,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(24440,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(24440,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24440,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24440,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24440,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24440,36,'cp 27','d1e2 f8e7 e1c1 e8g8 g2g4 c6d4 e3d4 a7a6 b5d7 f6d7 c1b1 b7b5 f2f4 b5b4 c3a4 d8a5 b2b3 a5b5 e2g2 d7c5 d4c5 d6c5 a4b2 b5c6 b2c4 a6a5 h2h4 a5a4 g4g5 f8d8 g2e2 c6c7 e4e5 c7c6 h4h5 d8d5 g5g6 d5d4 d1d4 c5d4 h1d1 h7g6 h5g6 f7g6');
INSERT INTO evaluations VALUES(28011,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(28011,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(28011,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(28011,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(28011,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(28011,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(28011,36,'cp 102','h7h6 h2h4 e7e6 f1d3 b8c6 c1f4 g8e7 d1d2 c6b4 d3e4 b4d5 h4h5 g6g5 f3g5 h6g5 f4g5 d7d6 e4d5 b7d5 c3d5 e6d5 e1c1 d6e5 h5h6 g7f8 d4e5 d8d7 h6h7 f8g7 g2g4 e8c8 f2f4 c8b8 d2c3 d7g4 d1g1 g4f5 g5e7 f5f4 c1b1 g7e5 c3d3 d8d7 g1f1 f4e4 d3e4 d5e4 f1f7');
INSERT INTO evaluations VALUES(30984,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(30984,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(30984,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(30984,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(30984,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(30984,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(30984,36,'cp 61','f1d3 f5d3 d1d3 e7e6 h4h5 c6c5 c2c3 b8c6 g1e2 a7a6 b1d2 c5d4 c3d4 d8a5 h1h3 a5b5 e1f1 g8e7 d2b3 b5d3 h3d3 e7c8 d3g3 c8b6 b3d2 a6a5 a2a4 c6b4 b2b3 b6c8 d2f3 b4c2 a1b1 c8e7 c1d2 e7c6 d2c3 a8c8');
INSERT INTO evaluations VALUES(32369,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(32369,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(32369,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(32369,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(32369,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(32369,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(32369,36,'cp 143','e2e4 f5g6 f1c4 g8f6 d1e2 f8b4 c1g5 h7h6 g5h4 g6e4 h4f6 e4f3 e2f3 g7f6 f3b7 b4c3 b2c3 b8d7 e1g1 d7b6 c4b3 a7a5 a2a4 e8g8 b7f3 f6f5 f1e1 a8b8 e1e5 b6d7 e5b5 g8g7 b3c2 b8b5 a4b5 d8g5 a1a5 f8d8 g2g3 d7f6 f3e3 g5e3 f2e3');
INSERT INTO evaluations VALUES(32371,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(32371,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(32371,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(32371,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(32371,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(32371,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(32371,36,'cp 119','f5g6 f1c4 g8f6 d1e2 f8b4 c1g5 h7h6 g5h4 g6e4 h4f6 e4f3 e2f3 g7f6 f3b7 b4c3 b2c3 b8d7 e1g1 d7b6 c4b3 a7a5 a1d1 a5a4 b3c2 e8e7 b7f3 a8a5 c2e4 d8d6 f1e1 h8b8 h2h4 b6c4 f3e2 c4b2 d1d2 d6a3 d4d5 e6e5 e1b1 a5b5 c3c4 b5b4 c4c5 b2c4 b1b4 a3b4 d2c2 c4a3');
INSERT INTO evaluations VALUES(32265,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(32265,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(32265,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(32265,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(32265,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(32265,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(32265,36,'cp 138','f1c4 g8f6 d1e2 f8b4 c1g5 h7h6 g5h4 g6e4 h4f6 e4f3 e2f3 g7f6 f3b7 b4c3 b2c3 b8d7 e1g1 d7b6 c4b3 a7a5 a2a4 e8g8 b7f3 f6f5 a1d1 d8g5 f1e1 a8b8 c3c4 b6d7 b3c2 b8b2 f3c3 f8b8 d4d5 d7c5 c3a5 b2c2 a5c5 e6e5 e1e5 g5f4 e5e1 f4c4 c5e7 c4a4');
INSERT INTO evaluations VALUES(32390,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(32390,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(32390,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(32390,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32390,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(32390,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(32390,36,'cp 24','d1b3 b8c6 g1f3 g8f6 e2e4 d5e4 f3e5 f5e6 b3b7 c6d4 f1b5 d4b5 b7b5 e6d7 b5c4 d8e7 e5d7 e7d7 c1g5 d7e6 c4a4 e6d7 c3b5 f8d6 g5f6 g7f6 a4e4 d7e6 b5d6 c7d6 e4e3 e8e7 e1c1 e6e3 f2e3 h8c8 c1b1 c8c5 g2g3 a7a5 d1d4 a8c8 a2a3 f6f5 h1d1 c5c2 d1d2 c2d2 d4d2');
INSERT INTO evaluations VALUES(32386,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(32386,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(32386,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(32386,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(32386,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(32386,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(32386,36,'cp 16','b8c6 g1f3 g8f6 e2e4 f6e4 b3b7 f5d7 b7b3 f8b4 f1d3 a8b8 b3c2 d7g4 d3e4 d5e4 c2e4 d8e7 f3e5 c6e5 d4e5 e7e6 e1g1 e8g8 h2h3 g4f5 e4f3 f8d8 c1e3 b4c3 b2c3 d8d3 a1b1 b8e8 b1b7 e6e5 b7a7 d3c3 e3f4 e5c5 a7a8 e8a8 f3a8 c5f8 a8b7 h7h6 f1d1 f8e7 f4e3');
INSERT INTO evaluations VALUES(32493,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(32493,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(32493,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(32493,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(32493,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(32493,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(32493,36,'cp 123','c2c4 c7c5 d4d5 f8e7 b1c3 e7f6 d1b3 d8b6 b3c2 g8e7 f1d3 e7g6 e1g1 e8g8 h2h3 h7h6 a2a3 b6d8 c1d2 b8a6 a1e1 a8b8 d2e3 a6c7 b2b4 b7b6 c3e2 c7e8 e1b1 b8c8 f1e1 e6d5 e4d5 g6e5 f3e5 f6e5 e2g3 d8f6 c2d2 e5g3 f2g3 d7f5 g3g4 f5d3 d2d3');
INSERT INTO evaluations VALUES(32489,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(32489,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(32489,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(32489,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(32489,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(32489,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(32489,36,'cp 105','c7c5 d4d5 f8e7 b1c3 e7f6 d1b3 d8b6 b3c2 g8e7 f1d3 e7g6 a2a3 e8g8 e1g1 b6d8 c1d2 b8a6 h2h3 h7h6 a1d1 b7b6 c3e2 a6c7 b2b4 a8c8 d2e3 e6e5 d1b1 c7e8 c2d2 f6e7 e2c3 e8f6 b4c5 b6c5 d2c2 f6h5 b1b7 c8c7 f1b1 g6f4 b7c7 d8c7');
INSERT INTO evaluations VALUES(39271,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(39271,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39271,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39271,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(39271,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(39271,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(39271,36,'cp -11','d7d6 g1f3 g8f6 f1c4 e8g8 d2d3 e5f4 c1f4 b8c6 d1d2 c8e6 c4e6 f7e6 e1c1 c5b4 f4g5 h7h6 g5h4 d8d7 h4f6 f8f6 a2a3 b4a5 d3d4 d7f7 h1e1 e6e5 d4e5 c6e5 f3e5 d6e5 b2b4 a5b6 c3d5 f6d6 e1f1 f7e6 d2e2 c7c6 d5b6 d6d1 f1d1 a7b6');
INSERT INTO evaluations VALUES(44548,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(44548,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(44548,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44548,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(44548,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(44548,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(44548,36,'cp 0','e7e6 b1d2 d7d5 g1f3 b8d7 f1d3 b7b6 e1g1 f8e7 d1e2 e8g8 e3e4 d5e4 d2e4 c8b7 e4f6 e7f6 f1d1 c5d4 f3d4 d7c5 d3c2 d8d5 f2f3 d5h5 c1e3 f8d8 d4b5 a7a6 b5d6 f6e5 d6b7 h5h2 g1f1 h2h1 f1f2 h1h4 f2f1');
INSERT INTO evaluations VALUES(45029,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(45029,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(45029,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(45029,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(45029,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(45029,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(45029,36,'cp 100','f1c4 b7b5 c4d3 c8g4 h2h3 g4h5 g2g4 h5g6 g4g5 b5b4 g5f6 b4c3 b2c3 g7f6 e1f1 b8d7 d1e2 e7e6 c3c4 d8a5 c1f4 a5a3 a1b1 h8g8 b1b3 a3a2 f4g3 a7a5 f1g2 a5a4 b3b7 g6d3 e2d3 a4a3 g2h2 g8g6 h1b1 f8e7 b1b3');
INSERT INTO evaluations VALUES(47450,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(47450,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(47450,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(47450,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(47450,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47450,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(47450,36,'cp 64','d5b6 f1e2 c8g4 d2d4 g4f3 g2f3 e7e6 c1e3 b8c6 b1c3 f8e7 d1d3 d8d7 e1c1 e8c8 a2a3 c8b8 c1b1 f7f5 d3c2 e7f6 e4e5 f6h4 f3f4 c6e7 a3a4 e7d5 a4a5 d5b4 c2b3 b6d5 e2c4 h4e7 c3d5 b4d5 a5a6 c7c6 h2h4 b8a8 a6b7 d7b7 b3b7 a8b7 h4h5 d8d7 d1c1');
INSERT INTO evaluations VALUES(46948,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(46948,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(46948,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(46948,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(46948,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(46948,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(46948,36,'cp 71','d2d4 c8g4 f1e2 g4f3 e2f3 e7e5 d4d5 f8c5 e1g1 e8g8 d1e2 b8d7 g2g3 b6c8 f3g2 c8d6 c1d2 c5d4 d2e3 f8e8 f1c1 f7f5 e4f5 d4e3 e2e3 d6f5 e3c3 f5d4 c3d3 g8h8 b1d2 d7f6 d2b3 d4b3 d3b3 e5e4 b3b7');
INSERT INTO evaluations VALUES(53900,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53900,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(53900,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(53900,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(53900,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(53900,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(53900,36,'cp 34','f1b5 c8d7 b5d7 d8d7 e1g1 b8c6 c2c3 g8f6 f1e1 f8e7 d2d4 e5d4 c3d4 d6d5 e4d5 f6d5 d4c5 e8c8 c1d2 d5b4 b1c3 e7c5 c3e4 c5e7 e1e2 h8e8 a2a3 f7f5 e4g3 b4d3 d1c2 e7f6 e2e8 d8e8 d2c3 f6c3 c2c3 d3f4 a1e1 e8e1 f3e1 f4e6');
INSERT INTO evaluations VALUES(54345,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(54345,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(54345,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(54345,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(54345,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(54345,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54345,36,'cp 27','f8d6 f4d6 d8d6 b1d2 b8c6 f1e2 a8b8 e1g1 g8e7 a2a4 e8g8 f1e1 h7h6 g2g3 d6c7 e3e4 d5e4 d2e4 c6a5 e2f1 a7a6 d1e2 a5b3 a1d1 c8d7 f3e5 d7a4 e2c4 c7c4 f1c4 b3a5 d1c1 a5c4 e5c4 f8d8 e4c5 a4c6 c4e5 c6e8 b2b4 e7d5 h2h4');
INSERT INTO evaluations VALUES(55552,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(55552,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55552,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(55552,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55552,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(55552,36,'cp 3','c7c5 b1c3 b8c6 a2a3 f8e7 f1e2 g8f6 a1b1 e8g8 g1f3 b7b6 b2b4 a7a6 e1g1 h7h6 c3d5 a8b8 d1c2 c8d7 c1d2 f6d5 e4d5 c6d4 f3d4 e5d4 b4c5 b6c5 b1b8 d8b8 f1b1 b8c7 g2g3 d7f5 g1g2 e7g5 h2h3 g5d2 c2d2');
INSERT INTO evaluations VALUES(1362,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(1362,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(1362,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(1362,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(1362,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(1362,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(1362,36,'cp 124','b1c3 g8f6 d4b3 a5c7 f2f4 e7e5 c3d5 c7d8 f1b5 d7d6 f4e5 f6d5 d1d5 c8e6 b5c6 b7c6 d5c6 e6d7 c6d5 d7e6 d5d4 d6e5 d4e5 f8d6 e5b5 e6d7 b5g5 e8g8 g5d8 f8d8 c1e3 d7c6 b3d2 a7a5 e1c1 a5a4 g2g3 f7f6 e3f4 d6c5 h1e1 c5f2 e1e2 f2d4 c2c3');
INSERT INTO evaluations VALUES(1360,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(1360,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(1360,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(1360,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(1360,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(1360,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(1360,36,'cp 121','g8f6 d4b3 a5c7 f2f4 e7e5 c3d5 c7d8 f1d3 f8e7 e1g1 e8g8 d5e7 d8e7 f4e5 f6e8 d1f3 c6e5 f3g3 d7d6 b3d4 c8d7 h2h3 a8c8 c1f4 e5g6 f4d2 g6e5 a2a4 e5d3 c2d3 e7e5 g3f2 f7f6 d2c3 d6d5 d4f5 e5e6 f2a7 d5e4 a7b7');
INSERT INTO evaluations VALUES(761,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(761,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(761,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(761,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(761,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(761,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(761,36,'cp 98','d4b3 a5c7 f2f4 e7e5 c3d5 c7d8 f1d3 d7d6 e1g1 f8e7 g1h1 e8g8 f4f5 a7a5 a2a4 c6b4 d5f6 e7f6 d3b5 c8d7 c2c3 d7b5 a4b5 a5a4 c3b4 a4b3 a1a8 d8a8 d1b3 a8a1 f1d1 f8a8 b3d5 a1a7 h2h3 a8c8 d5d6 h7h6 d6d3 a7a4 c1e3 a4b4');
INSERT INTO evaluations VALUES(1598,30,'cp 462',NULL);
INSERT INTO evaluations VALUES(1598,31,'cp 457',NULL);
INSERT INTO evaluations VALUES(1598,32,'cp 459',NULL);
INSERT INTO evaluations VALUES(1598,33,'cp 482',NULL);
INSERT INTO evaluations VALUES(1598,34,'cp 500',NULL);
INSERT INTO evaluations VALUES(1598,35,'cp 478',NULL);
INSERT INTO evaluations VALUES(1598,36,'cp 494','c3b5 e8d8 d2d4 a7a6 d4d5 a6b5 d5c6 d8e8 f3e5 g6e6 c1f4 b7c6 e1g1 g8f6 c2c4 f6d7 e5d7 c8d7 c4b5 a8c8 a1c1 c6b5 f4c7 g7g6 f1e1 f8h6 c1c2 h6g7 c2c5 e6a6 d1d2 g7f6 e1d1 e8f8 d2d7 f8g7 d7b5 a6a2 b2b3 a2b2 b5c4 c8a8 c5a5 a8a5 c7a5');
INSERT INTO evaluations VALUES(4267,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(4267,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(4267,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(4267,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4267,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4267,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(4267,36,'cp 16','b5c6 d7c6 h2h3 f8e8 e1g1 f6d7 c1e3 c5d6 f3d2 d7f8 f2f4 e5f4 e3f4 f8g6 f4d6 c7d6 d1f3 e8f8 f3e3 c8e6 d2f3 d8d7 a2a4 f7f5 a4a5 a8e8 a1e1 a7a6 e4e5 d6e5 f3e5 g6e5 e3e5 e6d5 e5f4 e8e1 f1e1');
INSERT INTO evaluations VALUES(6003,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(6003,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(6003,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(6003,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(6003,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(6003,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(6003,36,'cp 29','e2e3 c5b6 a2a3 g8f6 d2d3 c8e6 b2b4 e8g8 c1b2 d8d7 h2h3 a8d8 c3d5 e6d5 c4d5 c6e7 d1b3 h7h5 g1e2 c7c6 d5c6 b7c6 b3c2 d8c8 d3d4 e5d4 e2d4 h5h4 g3h4 c6c5 d4e2 c5b4 c2d1 d7b5 b2f6 g7f6 e1g1 b4b3 a3a4 b5b4');
INSERT INTO evaluations VALUES(5999,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(5999,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(5999,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(5999,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(5999,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(5999,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(5999,36,'cp 24','c5b6 d2d3 g8f6 g1e2 e8g8 e1g1 c6e7 b2b3 c7c6 e3e4 f8e8 h2h3 h7h6 d3d4 e5d4 e2d4 d6d5 c4d5 c6d5 e4d5 e7d5 c3d5 f6d5 c1b2 d5b4 a2a3 b6d4 b2d4 b4c6 d4e3 d8f6 a1c1 c8f5 b3b4 a8d8 d1a4 f5e4 f1d1 e4g2 d1d8');
INSERT INTO evaluations VALUES(11181,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11181,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(11181,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(11181,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(11181,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(11181,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(11181,36,'cp 55','d4c6 b7c6 b2b3 d8c7 e1g1 e7e5 d1e2 f8e7 c3a4 c8e6 c2c4 a6a5 c1e3 e8g8 a1c1 a8b8 a4c3 c7b7 f1d1 h7h6 c3a4 b7c7 f2f3 f6d7 d3c2 b8b7 e2f2 f8b8 d1d2 e7g5 e3g5 h6g5 f2g3 f7f6 c1d1 d7c5 d2d6 c5a4 d6e6 a4c5');
INSERT INTO evaluations VALUES(11137,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(11137,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11137,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(11137,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(11137,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11137,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(11137,36,'cp 25','b7c6 b2b3 g7g6 c1b2 f8g7 d1d2 f6d7 c3a4 g7b2 a4b2 d7c5 h2h4 h7h5 e1g1 e7e5 a1e1 e8g8 f2f4 d8h4 f4f5 g6f5 e4f5 f7f6 e1e3 f8f7 e3h3 h4d4 g1h2 f7g7 b2c4 d6d5 d2h6 d4g4 c4e3 g4g5 h3h5 g5h6 h5h6 c5d3 c2d3 g8f7 h6h8 c8b7 h8a8 b7a8 b3b4');
INSERT INTO evaluations VALUES(11789,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(11789,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(11789,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(11789,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(11789,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(11789,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(11789,36,'cp 41','b2b3 g7g6 c1b2 f8g7 d1d2 f6d7 c3a4 g7b2 a4b2 d7c5 e1g1 e8g8 f1e1 a6a5 d3f1 c8e6 h2h4 f7f6 e1e3 d8b6 a1e1 c5d7 f1c4 e6c4 b2c4 b6b4 e3c3 d7e5 c4e5 f6e5 h4h5 f8f6 d2e3 a8f8 f2f3 g6h5 c3c6 b4a3 c6c7');
INSERT INTO evaluations VALUES(12890,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12890,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12890,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(12890,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(12890,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(12890,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(12890,36,'cp -4','e4e5 f6g4 c1f4 d7d6 e5d6 e7e5 f4g3 f8d6 f1b5 e8g8 e1g1 c6d4 h2h3 g4h6 f1e1 h6f5 g3e5 d6e5 e1e5 c8e6 f3d4 f5d4 d1d3 d4b5 d3b5 d8d7 b5d7 e6d7 a1d1 d7c6 c3d5 f7f6 d5e7 g8f7 e5e1 f8e8 e7c6 e8e1 d1e1 b7c6 g2g4 f7g6 e1c1 a8c8 g1g2');
INSERT INTO evaluations VALUES(14270,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(14270,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(14270,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(14270,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(14270,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(14270,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(14270,36,'cp -59','e7e5 d4e5 c6e5 g1e2 f8c5 e1g1 e8g8 b2b3 f8e8 c1b2 c7c6 h2h3 d8c7 e2d4 c5d6 c3e2 a7a6 d4f5 c8f5 d3f5 c7e7 d1d2 a8d8 a1d1 d6c7 a2a4 e5g6 b2f6 e7f6 f5g6 h7g6 g2g3 f6f5 g1g2 f5e4 g2g1 c7d6 d2d3');
INSERT INTO evaluations VALUES(15111,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(15111,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(15111,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(15111,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(15111,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(15111,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(15111,36,'cp -67','d4e5 c6e5 g1e2 f8c5 e1g1 e8g8 b2b3 f8e8 c1b2 c7c6 h2h3 d8c7 e2d4 e5d3 c2d3 c5d4 e3d4 c8f5 d1f3 f5g6 c3e2 h7h6 b2c3 c7d6 e2f4 g6h7 b3b4 g7g5 f4h5 f6h5 f3h5 h7d3 f1e1 d3h7 a2a4 d6f6 e1e5 e8e5 d4e5 f6g6');
INSERT INTO evaluations VALUES(15074,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(15074,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(15074,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(15074,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(15074,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(15074,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(15074,36,'cp -78','c6e5 g1e2 f8c5 e1g1 e8g8 h2h3 f8e8 e2g3 c5f8 d3e2 c7c6 a2a4 e5g6 e2d3 f8d6 a1b1 d8e7 b2b3 g6e5 d3f5 g7g6 f5c8 a8c8 e3e4 e5d7 c1b2 d6g3 f2g3 f6e4 c3e4 e7e4 g1h2 c8d8 c2c4 d7c5 c4d5 e4d5');
INSERT INTO evaluations VALUES(18062,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(18062,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18062,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(18062,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(18062,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(18062,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(18062,36,'cp 14','d7d6 d3d4 e5d4 d1d4 g8f6 b1c3 f8e7 e4e5 c6c5 d4f4 f6g4 e1g1 e8g8 f4g3 f7f5 e5d6 c7d6 f1e1 c8b7 c1f4 g7g5 e1e7 d8e7 f4d6 f5f4 d6e7 f4g3 h2g3 f8f5 e7g5 b7f3 g2f3 f5g5 f3g4 g5g4 a1e1 g8h8 e1e5 g4c4 e5e6 c4d4 c3e4');
INSERT INTO evaluations VALUES(19144,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(19144,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(19144,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(19144,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(19144,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(19144,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(19144,36,'cp 70','d1d4 a7a6 b1c3 d8c7 c1e3 b7b5 c4b3 c7d6 a2a4 d6d4 e3d4 b5b4 c3b1 c8b7 b1d2 g8f6 f2f3 d7d5 d4f6 g7f6 e4d5 e6d5 e1c1 e8c8 h1e1 f8d6 d2f1 d6f4 c1b1 h8e8 g2g3 f4c7 f1e3 d5d4 e3f5 e8e1 d1e1 b7f3 b3f7');
INSERT INTO evaluations VALUES(21934,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(21934,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(21934,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(21934,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(21934,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(21934,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(21934,36,'cp 85','d4d5 c6b8 b1c3 b7b6 e2e4 g7g6 f1d3 f8h6 e1g1 h6c1 d1c1 g8f6 c1h6 f6g4 h6d2 g4f6 h2h3 f6h5 b2b4 e8g8 a2a3 a7a5 f1e1 b8d7 d3f1 g8h8 c3b5 c8a6 a1c1 a6b5 c4b5 a5b4 a3b4 f7f5');
INSERT INTO evaluations VALUES(21873,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(21873,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(21873,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(21873,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(21873,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(21873,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(21873,36,'cp 91','c6b8 b1c3 b7b6 e2e4 g7g6 h2h3 f8h6 c1h6 g8h6 d1c1 h6g8 f1e2 b8d7 b2b4 g8e7 e1g1 e8g8 c1h6 f7f6 a2a3 g8h8 h6d2 e7g8 a1e1 a7a6 f3h2 d8e7 f2f4 e5f4 d2f4 g6g5 f4e3 d7e5 h2f3 c8d7 f3d4 g8h6 f1f2 f6f5 e4f5 h6f5 d4f5');
INSERT INTO evaluations VALUES(23085,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(23085,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23085,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23085,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(23085,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23085,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(23085,36,'cp 20','e5g6 f4g3 g8f6 d1d4 d7d5 b1c3 f6e4 c3e4 d5e4 d4e4 f8e7 a1d1 c8d7 g1f3 e8g8 f1d3 d8c8 e1g1 e7f6 h2h4 f8e8 e4c4 d7g4 f1e1 g4f3 e1e8 c8e8 g2f3 c7c6 c4b4 a8d8 b4b7 f6h4 b7c7 h7h6 g1f1 h4g3 f2g3 g6e5 d3e2 d8d1 e2d1 e8e6');
INSERT INTO evaluations VALUES(23827,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(23827,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(23827,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(23827,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(23827,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(23827,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(23827,36,'cp 34','h2h3 g4h5 c3d5 g8e7 a2a4 e7d5 c4d5 c6b4 d5b3 a7a5 c2c3 b4c6 b3c4 d8f6 g2g4 h5g6 d1e2 f6e7 c1e3 c5e3 f2e3 c6d8 g1h2 d8e6 e2f2 e8g8 f3h4 c7c6 f2g3 g8h8 f1f2 e6c5 a1g1 c5a4 h4f5 g6f5 g4f5 f8g8 c4b3 a4c5 f5f6 g7f6');
INSERT INTO evaluations VALUES(23826,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(23826,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(23826,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(23826,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(23826,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(23826,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(23826,36,'cp 31','g4e6 c3d5 g8f6 c2c3 e6d5 e4d5 c6e7 c4b5 c7c6 d5c6 b7c6 b5a4 c5b6 f1e1 e8g8 a4c2 c6c5 d3d4 e5d4 c3d4 e7c6 c1e3 c6d4 f3d4 c5d4 e3d4 a8c8 c2b3 d6d5 d1d3 b6d4 d3d4 a7a5 e1e5 c8b8 b3d5 b8b4 d4c5 f6d5 c5d5 d8d5 e5d5 b4b2');
INSERT INTO evaluations VALUES(26116,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(26116,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(26116,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26116,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(26116,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(26116,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26116,36,'cp 24','c2c3 f8e7 f1d3 f5d3 d1d3 g7g6 e1g1 h7h5 a2a4 g8h6 c1e3 h6f5 b1d2 d8b6 e3f2 a7a5 h2h3 h5h4 b2b3 b6a6 c3c4 e8g8 d3e2 c6c5 g1h2 a8c8 a1d1 a6b6 d4c5 d7c5 f3d4 f5d4 f2d4 b6a6 e2f3 f8d8');
INSERT INTO evaluations VALUES(32890,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(32890,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(32890,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32890,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(32890,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(32890,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(32890,36,'cp 62','g8f6 f3e5 c7c6 f1c4 e7e6 g2g4 f5g6 h2h4 b8d7 e5d7 f6d7 h4h5 g6e4 e1g1 e4d5 c3d5 c6d5 c4d3 f8d6 d1f3 g7g6 h5h6 e8g8 c2c3 a8e8 c1g5 a5b6 a1e1 f7f6 g5c1 b6c7 f3g2 d6f4 c3c4 d5c4 c1f4 c7f4 d3c4 d7e5 d4e5 f4c4 e1e3 c4c7 f1e1 f6e5 e3e5 c7b6 e5e2 b6d4 b2b3');
INSERT INTO evaluations VALUES(42610,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(42610,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(42610,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(42610,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(42610,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(42610,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(42610,36,'cp 45','e5d4 e4e5 f6d5 c3d4 c5b4 b1d2 d7d6 a2a3 b4d2 c1d2 b8c6 f1c4 d5b6 d2g5 d8d7 c4a2 d6e5 e1g1 h7h6 g5e3 e5d4 f3d4 c6d4 e3d4 e8g8 f1e1 d7d6 a2b1 g7g6 b1e4 c7c6 d1d3 f8e8 a1d1 c6c5 d4e3 d6d3 e4d3 c8f5 d3f5 g6f5 e3c5 e8e1 d1e1');
INSERT INTO evaluations VALUES(42684,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(42684,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(42684,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(42684,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(42684,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(42684,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(42684,36,'cp 55','e4e5 f6d5 c3d4 c5b4 b1d2 d7d6 a2a3 b4d2 c1d2 b8c6 f1c4 d5b6 d2g5 d8d7 c4a2 h7h6 g5e3 d6e5 e1g1 e5d4 f3d4 c6d4 e3d4 e8g8 f1e1 d7d6 a2b1 g7g6 b1e4 c7c6 d1d2 f8d8 a1d1 c6c5 d4e3 d6d2 d1d2 d8d2 e3d2 c8f5 e4b7 a8b8');
INSERT INTO evaluations VALUES(46136,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(46136,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(46136,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(46136,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46136,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46136,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(46136,36,'cp 45','e6d5 f1g2 b8c6 e1g1 c5c4 f3e5 f8d6 e5c6 b7c6 b2b3 c4b3 a2b3 c8f5 b1d2 e8g8 c1a3 d6a3 a1a3 d8e7 a3a5 f8b8 f1e1 b8b6 d1c1 g7g6 e2e3 a8b8 c1a3 e7a3 a5a3 a7a6 e1a1 f5d3 h2h4 g8g7 g2f1 f6e4 d2e4 d3e4 a3a6 b6b3 a6c6');
INSERT INTO evaluations VALUES(46440,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(46440,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(46440,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(46440,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(46440,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(46440,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(46440,36,'cp 32','f1g2 b8c6 e1g1 c5c4 f3e5 f8d6 e5c6 b7c6 b2b3 c4b3 a2b3 e8g8 c1a3 d6a3 a1a3 c8f5 b1d2 d8e7 a3a2 f8b8 e2e3 h7h5 h2h4 b8b6 d1c1 g7g6 f1e1 f6e4 g2e4 f5e4 c1a3 e7a3 a2a3 a7a5 e1a1 b6b5 b3b4 b5b4 a3a5 a8a5 a1a5 b4b2');
INSERT INTO evaluations VALUES(46441,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46441,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46441,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(46441,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(46441,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(46441,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(46441,36,'cp 26','b8c6 e1g1 c5c4 f3e5 f8d6 e5c6 b7c6 b2b3 c4b3 a2b3 e8g8 c1a3 f8e8 a3d6 d8d6 e2e3 c8f5 b1d2 a7a5 a1a4 c6c5 d4c5 d6c5 d1a1 f5d7 b3b4 c5d6 a4a5 a8b8 f1d1 b8b4 d2f3 h7h6 f3d4 e8b8 a5a2 d6e5 h2h3 b4b6 a1c1 b8c8 c1a3 c8b8 d1a1 f6e4 g2e4');
INSERT INTO evaluations VALUES(13758,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(13758,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13758,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13758,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(13758,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(13758,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(13758,36,'cp 60','e1g1 f8e7 d4c5 e7c5 a2a3 f6e4 f3d2 e4d2 b1d2 e8g8 b2b4 c5b6 c1b2 c8g4 h2h3 g4e6 d2b3 d8d7 g1h2 a8c8 e2e3 f7f6 b2d4 c6d4 b3d4 b6d4 d1d4 b7b6 f1d1 f8d8 d4d2 d5d4 d2d4 d7d4 e3d4 e6b3 d1d2 b3d5');
INSERT INTO evaluations VALUES(48725,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48725,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48725,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48725,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48725,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48725,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48725,36,'cp 0','d2d4 c7c6 d1c2 g7g6 c1g5 f8e7 e2e3 c8f5 f1d3 f5d3 c2d3 e8g8 g5f6 e7f6 b2b4 a7a6 e1g1 b8d7 a2a4 b7b5 c3e2 d7b6 a4b5 a6b5 a1a8 d8a8 d3c3 a8b7 e2f4 f8a8 f4d3 b7c8 f1a1 b6c4 d3c5 f6e7 h2h3 a8a1 c3a1 h7h5 c5d3 e7d6 d3c5');
INSERT INTO evaluations VALUES(50874,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(50874,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(50874,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(50874,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(50874,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(50874,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(50874,36,'cp 37','b1c3 g8f6 e1g1 b8c6 a2a4 f8e7 c1d2 e8g8 f1e1 f8e8 a4a5 e7f8 c3d5 f6d5 c4d5 d8f6 h2h3 c6e7 d5b3 c8e6 b3e6 f6e6 c2c4 e7g6 d1c2 a8d8 d3d4 e5d4 f3d4 e6f6 d4f5 g6e7 f5e7 f8e7 a1d1 f6g6');
INSERT INTO evaluations VALUES(50871,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50871,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50871,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(50871,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(50871,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(50871,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(50871,36,'cp 33','g8f6 c1e3 f8e7 h2h3 c7c5 a2a4 b8c6 c3d5 f6d5 c4d5 e8g8 c2c3 c8e6 e1g1 e6d5 e4d5 c6b8 a1b1 d8c8 d1e2 b7b5 b2b4 c5b4 b1b4 c8c3 b4g4 f7f5 g4g6 b8d7 f1c1 c3a5 e3h6 f8f7 h6g5 e7f6 g5f6 f7f6 g6f6 d7f6 c1c6 a5a4');
INSERT INTO evaluations VALUES(51263,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(51263,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(51263,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(51263,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(51263,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(51263,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(51263,36,'cp 88','c3b5 a6b5 a1a8 c8b7 a8a1 e7e6 g1e2 b7e4 e2c3 f8b4 f1e2 g8f6 e1g1 b4c3 b2c3 e8g8 c1g5 b8d7 e2f3 d8b6 g5d2 h7h6 f1e1 e4f3 d1f3 f6d5 a1a3 d7f6 e1a1 f8c8 d2e1 b5b4 c3b4 b6d4 f3d1 d4e5 a3a8 c8a8 a1a8 g8h7 d1b1 g7g6 a8a7 h7g7 b4b5');
INSERT INTO evaluations VALUES(51261,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(51261,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(51261,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(51261,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(51261,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(51261,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(51261,36,'cp 79','a6b5 a1a8 c8b7 a8a1 e7e6 g1e2 b7e4 e2c3 f8b4 f1e2 g8f6 e1g1 b4c3 b2c3 e8g8 c1g5 b8d7 f1e1 d8b6 g5d2 f8c8 e2f3 e4d5 d1e2 d5f3 e2f3 f6d5 a1a3 d7f6 e1d1 h7h6 d2e1 d5e7 d1a1 f6d5 h2h3 e7c6 a3a8 b6b7 a8c8 b7c8 e1d2 c8b7 a1a3 c6e7 f3g4 g8h7 g4e4 e7g6 e4b1 b7b6 b1a1');
INSERT INTO evaluations VALUES(52454,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52454,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(52454,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(52454,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(52454,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52454,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52454,36,'cp -14','b8c6 f1b5 d7d5 e4d5 d8d5 d1f3 d5f3 g1f3 c8d7 e1g1 e7e6 d2d3 a8d8 c1e3 g8f6 a3c4 c6b4 b5d7 f6d7 a1c1 f8e7 e3d2 d7b8 a2a3 b4d5 f1e1 b8c6 g1f1 e7f6 f3e5 c6e5 c4e5 e8g8 b2b4 c5b4 d2b4 d5b4 a3b4 a7a6 e5c4');
INSERT INTO evaluations VALUES(19473,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(19473,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(19473,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(19473,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(19473,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(19473,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(19473,36,'cp -12','f1b5 d7d5 e4d5 d8d5 d1f3 d5f3 g1f3 c8d7 d2d3 a7a6 b5c6 d7c6 a3c4 e7e6 c1e3 c6f3 g2f3 g8f6 a2a4 f6d5 a4a5 a8d8 a1a3 f7f6 a3b3 d8d7 e1e2 e8f7 f3f4 f8e7 h1a1 h8b8 c2c3 h7h6 a1a4 g7g5 f4g5 h6g5 c4b6 d7c7 a4c4 b8g8 e3d2');
INSERT INTO evaluations VALUES(53201,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(53201,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(53201,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53201,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53201,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(53201,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(53201,36,'cp 22','a7a6 f1c4 e7e6 e1g1 g8f6 b2b3 b8d7 c1b2 f8e7 d4c5 e7c5 b1d2 e8g8 c4d3 b7b5 a2a4 b5a4 a1a4 c8b7 d1b1 c5e7 f1d1 d8b8 b1a1 b7d5 a4a2 a6a5 a2a5 b8b7 b2c3 f8c8 c3d4 d7c5 d4c5 a8a5');
INSERT INTO evaluations VALUES(54759,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(54759,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(54759,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54759,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(54759,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(54759,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(54759,36,'cp 25','g1f3 e7e6 e2e3 g8f6 f1e2 c7c5 e1g1 b8c6 c2c4 a7a6 a2a3 f8e7 b1c3 e8g8 d4c5 d5c4 c3a4 c6a5 f3d4 f6e4 a4b6 a8b8 b6c4 a5c4 e2c4 e7c5 c4e6 g8h8 e6a2 a6a5 h2h3 c8d7 b2b4 a5b4 a3b4 c5b4 c1b2 b4c5 g1h2');
INSERT INTO evaluations VALUES(54758,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(54758,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(54758,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(54758,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54758,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(54758,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(54758,36,'cp 13','e7e6 e2e3 g8f6 f1e2 f8e7 c2c4 c7c5 e1g1 b8c6 a2a3 a7a6 b1c3 e8g8 d4c5 d5c4 c3a4 c6a5 f3d4 f6e4 c1d2 e4c5 a4c5 e7c5 d2a5 d8a5 e2c4 c5d4 e3d4 f8d8 d1b3 d8d4 c4e6 c8e6 b3e6 g8h8 a1e1 d4d8 h2h3 a5b5 b2b4 d8e8 e6b3 e8e1 f1e1 a8e8 e1e8 b5e8');
INSERT INTO evaluations VALUES(57112,30,'cp -187',NULL);
INSERT INTO evaluations VALUES(57112,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(57112,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(57112,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(57112,34,'cp -187',NULL);
INSERT INTO evaluations VALUES(57112,35,'cp -184',NULL);
INSERT INTO evaluations VALUES(57112,36,'cp -189','f8b4 c2c3 b4a5 g1f3 d7d6 d2d4 e5d4 c1g5 g8e7 d1a4 c7c6 f3d4 f7f6 g5e3 e8g8 f1e2 a5b6 b1d2 f6f5 a4b3 g8h8 e4f5 e7f5 d4f5 f8f5 d2c4 b6e3 c4e3 f5f6 a1d1 b8d7 b3a3 d8f8 c3c4 d7c5 e1g1 c8e6 h2h3 f6f4 d1d2 f8e7 e2f3 e6c4 e3c4 f4c4');
INSERT INTO evaluations VALUES(39344,30,'cp -185',NULL);
INSERT INTO evaluations VALUES(39344,31,'cp -173',NULL);
INSERT INTO evaluations VALUES(39344,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(39344,33,'cp -165',NULL);
INSERT INTO evaluations VALUES(39344,34,'cp -177',NULL);
INSERT INTO evaluations VALUES(39344,35,'cp -166',NULL);
INSERT INTO evaluations VALUES(39344,36,'cp -190','g1f3 b8c6 c2c3 b4a5 f1b5 g8e7 e1g1 a7a6 b5a4 e8g8 d2d4 e5d4 c3d4 d7d5 e4d5 e7d5 a4c6 b7c6 h2h3 h7h6 c1d2 a5b6 b1a3 c6c5 d4c5 b6c5 a3c4 c8e6 a1c1 f8e8 f1e1 d8f6 d1a4 c5f8 c4e5 f8d6 e5c4 d5b6 c4b6 c7b6');
INSERT INTO evaluations VALUES(8950,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(8950,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(8950,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(8950,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(8950,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(8950,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(8950,36,'cp 69','b4d6 f1d3 f6d7 f4d6 c7d6 e3e4 d5e4 d2e4 d7f6 e1g1 f6e4 d3e4 c6e7 f1e1 h7h6 a1c1 e8g8 c3c4 d6d5 c4d5 e7d5 f3e5 c8d7 e4d5 e6d5 d1b3 d7c6 h2h3 a8c8 e5c6 b7c6 b3c3 a7a6 b2b3 f8e8 e1e8 d8e8 a2a3 c8b8 c1e1 e8d7 b3b4 b8c8');
INSERT INTO evaluations VALUES(9235,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(9235,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(9235,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(9235,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(9235,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(9235,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(9235,36,'cp 68','f1d3 f6d7 f4d6 c7d6 e3e4 d5e4 d2e4 d7f6 d1d2 h7h6 e1c1 f6e4 d3e4 d6d5 e4c2 b7b5 h1e1 a7a5 g2g4 c8a6 f3e5 c6e5 e1e5 d8d6 f2f4 b5b4 c3b4 a5b4 c2a4 e8e7 f4f5 a6c4 b2b3 a8a4 b3a4 b4b3 c1b2 b3a2 b2a1 h8b8 f5e6 d6b4 d2b4 b8b4 e6f7 e7f7 g4g5 b4a4 g5h6 g7h6 d1g1');
INSERT INTO evaluations VALUES(31117,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(31117,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(31117,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(31117,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(31117,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(31117,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(31117,36,'cp 84','f3e5 g4e6 e2e4 g8f6 e4d5 c6d5 f1d3 g7g6 e1g1 b8d7 e5f3 f8g7 f1e1 e8g8 h2h3 a8c8 a2a4 e6f5 d3f5 g6f5 a4a5 f8e8 c3e2 d7f8 c2c3 e7e6 c1e3 f8g6 f3e5 f6e4 e5d3 b7b6 a5b6 a7b6 e2f4 d8c7');
INSERT INTO evaluations VALUES(33279,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(33279,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(33279,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(33279,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(33279,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(33279,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(33279,36,'cp 187','h4e7 e5d3 c5b6 b1c3 g8f6 d3f4 c7c6 f1e2 d7d5 e4d5 b8a6 d5d6 e7d6 c3e4 f6e4 f3e4 e8d8 e1g1 b6c7 e2a6 d6f4 e4f4 c7f4 a6c4 c8f5 d2d3 f4c1 a1c1 f7f6 a2a4 h8e8 f1e1 f5d7 f2f3 e8e1 c1e1 a7a5 g1f2 b7b5 a4b5 c6b5 c4d5 a8c8 c2c3 a5a4');
INSERT INTO evaluations VALUES(34316,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(34316,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(34316,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(34316,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(34316,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(34316,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(34316,36,'cp 71','c8f5 d2d4 b8d7 c1d2 g8f6 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 b7b5 c4b3 d7b6 g2g4 f8b4 e2f3 a7a5 a2a3 b4e7 a3a4 f6d5 b3d5 b6d5 h4h5 d5c3 f3c3 b5b4 c3d3 g6h5 g4h5 a8d8 h5h6 g7g6 h6h7 d8d5 d3e2 d5d4 f2f4 e7f6');
INSERT INTO evaluations VALUES(40899,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(40899,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(40899,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(40899,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(40899,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(40899,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(40899,36,'cp 55','e8g8 f1d3 c7c5 e1g1 c5d4 c3d4 b7b6 c1f4 b8c6 d3b5 c6a5 d4d5 e6d5 e4d5 e7d6 f4g5 f7f6 g5d2 c8d7 b5d7 d8d7 d2a5 b6a5 d1d3 f8d8 a1c1 d6f8 f1d1 a8b8 g2g3 b8b5 d5d6 b5b6 d3d5 g8h8 d5a5 b6d6 d1d6 d7d6 a5a7 d6d5 a7b6 h7h5 g1g2 h5h4');
INSERT INTO evaluations VALUES(36289,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(36289,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(36289,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(36289,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(36289,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(36289,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(36289,36,'cp 51','f1d3 c7c5 e1g1 c5d4 c3d4 b7b6 h2h3 c8b7 d1e2 b8d7 c1d2 a8c8 a2a4 a7a5 a1b1 b7c6 d3a6 c8a8 a6b5 c6b5 e2b5 h7h6 f1e1 a8c8 h3h4 d8c7 e1c1 c7a7 c1c6 a7b7 b1c1 c8c6 b5c6 b7c6 c1c6 e7b4 d2b4 a5b4');
INSERT INTO evaluations VALUES(36290,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(36290,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36290,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(36290,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36290,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(36290,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(36290,36,'cp 51','c7c5 e1g1 c5d4 c3d4 b7b6 c1f4 b8c6 d3b5 c6a5 d4d5 e6d5 e4d5 e7d6 f4g5 f7f6 g5d2 c8d7 b5d7 d8d7 d2a5 b6a5 d1d3 f8d8 a1c1 d6f8 f1d1 a8b8 g2g3 b8b5 d5d6 b5b6 d3d5 g8h8 d5a5 b6d6 d1d6 f8d6 h2h4 h7h6 h4h5 h8g8 f3h4 d6e5 h4g6 d7d2');
INSERT INTO evaluations VALUES(45971,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(45971,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(45971,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(45971,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(45971,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(45971,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(45971,36,'cp -71','a7a6 c4b3 b7b5 d2d3 h7h6 g5f3 b8c6 e1g1 d6d5 e4d5 e6d5 f1e1 f8e7 a2a4 b5b4 c3e2 e8g8 c1f4 e7d6 f4d6 d8d6 h2h3 a6a5 e2g3 c8e6 c2c3 f8e8 d3d4 c5c4 b3c2 e6d7 d1d2 a8c8 g3f5 e8e1 a1e1 d7f5 c2f5 c8e8 f5c2 e8e1 d2e1 b4b3 c2d1 c6e7 e1d2');
INSERT INTO evaluations VALUES(43486,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(43486,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(43486,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(43486,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(43486,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(43486,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(43486,36,'cp -79','c4b3 d6d5 d2d3 h7h6 g5f3 b7b5 e4d5 e6d5 a2a3 f8e7 c3e2 e8g8 e1g1 b8c6 h2h3 f8e8 c2c3 e7d6 c1f4 d6f4 e2f4 d5d4 f1e1 d4c3 e1e8 d8e8 b2c3 c8b7 f4d5 f6d5 b3d5 a8d8 c3c4 c6a5 d5b7 a5b7 a1c1 e8d7 f3e5 d7f5 e5c6 d8d7 c4b5 a6b5 d3d4 c5c4 c1c3');
INSERT INTO evaluations VALUES(54344,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(54344,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(54344,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(54344,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(54344,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(54344,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(54344,36,'cp 59','b2b3 c4b3 a2b3 b8c6 b3b4 g8f6 b1d2 f8d6 f4d6 d8d6 b4b5 c6e7 g1f3 e8g8 f1d3 a7a5 b5a6 b7a6 e1g1 a6a5 c3c4 e7c6 c4c5 d6c7 d3b5 c8a6 b5c6 c7c6 f3e5 c6c7 f1e1 f8b8 d1c2 a6b5 a1a3 f6d7 e5d7 b5d7 e1a1 a5a4 d2b1 b8b4 b1c3 c7c6 h2h3 d7e8');
INSERT INTO evaluations VALUES(1577,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(1577,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(1577,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(1577,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(1577,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(1577,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(1577,36,'cp -9','c8e6 g1f3 a7a6 a2a3 e6g4 f1e2 e7e5 h2h3 g4d7 b2b4 f7f6 c1b2 g8h6 c3e4 e8c8 e1g1 f8e7 f1e1 h8e8 e2f1 d7e6 d2d3 h6f5 c2c3 f5d6 e4d6 e7d6 f3d2 d6f8 a1d1 e6f7 b2c1 h7h6 g2g3 f7h5 g3g4 h5g6 d2b3');
INSERT INTO evaluations VALUES(27608,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(27608,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(27608,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(27608,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(27608,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(27608,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(27608,36,'cp -12','g1f3 a7a6 a2a3 e6g4 f1e2 e7e5 h2h3 g4d7 b2b4 e8c8 c1b2 g8h6 c3e4 f7f6 e1g1 f8e7 f1e1 h8e8 e2c4 h6f5 a1d1 g7g5 g2g4 f5d6 e4d6 e7d6 d2d4 e5d4 e1e8 d8e8 f3d4 d6e5 c4f7 e8e7 f7d5 c6d4 b2d4 e5d4 d1d4');
INSERT INTO evaluations VALUES(27607,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(27607,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(27607,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(27607,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(27607,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(27607,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27607,36,'cp -5','a7a6 d2d3 c6b4 e1d2 b4c6 g2g3 f7f6 a2a3 e8c8 b2b4 c8b8 c1b2 g8h6 a1e1 e6d7 c3e4 e7e5 d2c1 a6a5 c2c3 a5a4 f1e2 f8e7 e2d1 d7e6 d1c2 h6f5 f3d2 h7h5 h2h4 f5h6 c2a4 d8d3 a4c2 d3d8');
INSERT INTO evaluations VALUES(6827,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(6827,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(6827,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(6827,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(6827,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(6827,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(6827,36,'cp 40','f3d4 c5d4 c2c3 d4b6 d2d4 c7c6 b5c4 d7d6 d4e5 d6e5 d1d8 b6d8 a2a4 g8f6 a4a5 b7b5 a5b6 d8b6 b1d2 e8g8 b2b3 a7a5 h2h3 f8d8 f1e1 b6c7 d2f3 h7h6 c1a3 c8a6 a3e7 d8e8 c4a6 e8e7 a6f1 f6d7 f3d2 d7b6 g2g3 a5a4 b3a4 b6a4');
INSERT INTO evaluations VALUES(6825,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(6825,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(6825,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(6825,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(6825,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(6825,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(6825,36,'cp 28','c5d4 c2c3 d4b6 d2d4 c7c6 b5c4 d7d6 d4e5 d6e5 d1d8 b6d8 a2a4 g8f6 a4a5 b7b5 a5b6 d8b6 b1d2 a7a5 b2b4 e8g8 b4a5 a8a5 c1a3 f8d8 d2b3 a5a4 a3e7 a4a1 f1a1 d8e8 e7c5 f6d7 c5b4 d7f6 c4d3 h7h6 a1a8 c8d7 a8e8 d7e8 g1f1 f6d7 f1e2 g7g6 b3a5 d7c5');
INSERT INTO evaluations VALUES(6818,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(6818,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(6818,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(6818,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(6818,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(6818,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(6818,36,'cp 46','c2c3 d4b6 d2d4 c7c6 b5c4 d7d6 d4e5 d6e5 d1d8 b6d8 a2a4 g8f6 a4a5 b7b5 a5b6 d8b6 b1d2 e8g8 h2h3 a7a5 f1e1 f8d8 a1a2 h7h6 b2b3 f6d7 c1a3 b6c7 e1a1 d7b6 c4e2 c8e6 a3e7 d8d7 e7c5 a5a4 c5b6 c7b6 d2c4 e6c4 e2c4 a8d8 b3a4 b6a5 a2a3 d7d1 a1d1 d8d1 g1h2');
INSERT INTO evaluations VALUES(6817,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(6817,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(6817,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(6817,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(6817,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(6817,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(6817,36,'cp 33','d4b6 d2d4 c7c6 b5c4 d7d6 d4e5 d6e5 d1d8 b6d8 a2a4 g8f6 a4a5 b7b5 a5b6 d8b6 b1d2 e8g8 h2h3 a7a5 f1e1 f8d8 a1a2 f6d7 c4f1 b6c7 d2c4 d7f6 c1g5 c8e6 e1a1 a8b8 f2f3 h7h6 g5h4 d8e8 h4f2 e8d8 a2a4 e6c4 a4c4 d8d2 c4c6 b8b2 c6c7');
INSERT INTO evaluations VALUES(6304,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(6304,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(6304,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(6304,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(6304,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(6304,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(6304,36,'cp 37','d2d4 c7c6 b5c4 d7d6 d4e5 d6e5 d1d8 b6d8 a2a4 g8f6 a4a5 b7b5 a5b6 d8b6 b1d2 e8g8 f1e1 a7a5 d2f3 b6c7 b2b3 h7h6 g1f1 f8d8 c1a3 f6d7 e1d1 d7b6 d1d8 c7d8 f3e5 b6c4 e5c4 a8b8 c4d2 d8f6 a1c1 f6g5 f1e2 g5d2 e2d2 b8b3');
INSERT INTO evaluations VALUES(6303,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6303,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(6303,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(6303,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(6303,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(6303,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(6303,36,'cp 23','c7c6 b5c4 d7d6 d4e5 d6e5 d1d8 b6d8 a2a4 g8f6 a4a5 b7b5 a5b6 d8b6 b1d2 a7a5 f1e1 e8g8 h2h3 h7h6 b2b3 f8d8 a1a2 b6c7 c1a3 f6d7 c4e2 a5a4 e1a1 a4b3 d2b3 d7f6 a3e7 a8a2 a1a2 d8e8 e7f6 g7f6 e2c4 c7b6 b3a5 b6a5 a2a5 c8e6 c4e6 e8e6 g1f1 g8g7 a5a8');
INSERT INTO evaluations VALUES(7901,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7901,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(7901,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(7901,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(7901,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7901,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7901,36,'cp 7','e4e5 f6g4 f1e1 d7d6 e5d6 d8d6 c3e4 d6c7 d2d4 c5d4 h2h3 g4f6 f3d4 c6d4 d1d4 e8g8 e4f6 e7f6 d4e4 c8d7 c4d3 g7g6 e4f3 f6g7 c2c3 a8c8 d3e4 b7b6 c1f4 e6e5 f4e3 d7e6 e4d5 e6d5 f3d5 f8d8 d5b3 g7f6 e1d1 c7c6 a2a4 d8d1 a1d1 c8d8 d1d8 f6d8 b3d1 d8e7 d1g4 g8f8');
INSERT INTO evaluations VALUES(19006,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19006,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(19006,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19006,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19006,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19006,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19006,36,'cp 18','g2g3 g8e7 f1g2 d7d6 f2f4 e5f4 c1f4 e7g6 f4e3 c8e6 d1d2 f8e7 g1e2 e7f6 e1g1 e8g8 a2a4 a7a6 f1f2 f6e5 a1f1 d8a5 e2f4 e5d4 e3d4 c5d4 c3d5 a5d2 f2d2 g6f4 d5f4 g7g6 h2h4 a8c8 d2f2 c6e5 a4a5 g8g7');
INSERT INTO evaluations VALUES(20108,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(20108,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(20108,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20108,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(20108,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20108,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(20108,36,'cp 5','d2d3 e5c4 d3c4 g7g6 d1d3 f8g7 b1c3 g8f6 e1g1 h7h6 g5f3 b7b6 h2h3 f6d7 f1e1 c8b7 c1d2 e8g8 a1d1 d7b8 d2f4 b8c6 f4g3 g8h7 c3d5 c6e5 f3e5 d6e5 f2f3 e7e6 d5e3 d8d3 d1d3 a8d8 e1d1 d8d4 c2c3 d4d3 d1d3');
INSERT INTO evaluations VALUES(22134,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(22134,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(22134,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(22134,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(22134,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(22134,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(22134,36,'cp 159','c2c3 g8f6 c3d4 d6d5 e4d5 f6d5 e1g1 c7c6 b1c3 f8d6 d1b3 e8g8 c4d5 c6d5 b3d5 d8b6 a1b1 c8e6 d5e4 a8e8 d4d5 e6c8 c1e3 b6c7 e4h4 b7b6 f1e1 c8b7 h2h3 c7d7 b1c1 h7h6 a2a3 d6b8 h4g4 f7f5 g4g6 f8f6 g6h5 g8h7 e3d4 e8e1 c1e1 f6g6 d4e5');
INSERT INTO evaluations VALUES(23005,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(23005,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(23005,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(23005,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(23005,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(23005,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(23005,36,'cp 188','e1g1 a7a6 b5a4 b4c6 a4c6 d7c6 f3e5 c8e6 d2d4 f7f6 e5f3 d8d7 c1e3 g8h6 b1d2 h6g4 d1e2 e8c8 c2c4 g4e3 e2e3 c8b8 d4d5 e6f7 a1c1 f8b4 a2a3 b4d2 e3d2 h8e8 d2f4 f7g6 f1e1 d7e7 d5c6 e7d6 f4d6 d8d6 c6b7 g6e4 c4c5');
INSERT INTO evaluations VALUES(23158,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(23158,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(23158,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(23158,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(23158,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(23158,35,'cp 236',NULL);
INSERT INTO evaluations VALUES(23158,36,'cp 213','f2f4 c7c6 a2a3 e5f4 c1f4 d8f6 d1c1 f8d6 f4d6 f6d6 b1c3 b7b5 g1f3 d6h6 c1h6 g8h6 b2b4 a5b7 a3a4 b5a4 f1d3 a7a5 a1a4 a8b8 b4a5 b7c5 a4a1 c5d3 c2d3 c8a6 e1g1 e8g8 f1b1 h6g4 d5c6 d7c6 d3d4 f8e8 h2h3 g4f6 b1b8');
INSERT INTO evaluations VALUES(25566,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(25566,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(25566,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(25566,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(25566,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(25566,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(25566,36,'cp 76','d1d3 h7h6 b5d7 d8d7 c1e3 f6g4 e3d2 c7c6 d5c6 b7c6 h2h3 g4f6 e1c1 e7g6 h3h4 f6g4 d3e2 h6h5 g2g3 a8b8 c1b1 f8e7 d2c1 e8g8 f3h2 g4f6 c1g5 d6d5 g5f6 e7f6 e2h5 d5d4 b2b3 d7c7 c3a4 c7a5 a4b2 g6e7');
INSERT INTO evaluations VALUES(26576,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(26576,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(26576,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(26576,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(26576,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(26576,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(26576,36,'cp 43','e7e6 c2c4 g8f6 b2b4 f8e7 c1b2 f6e4 b1d2 e8g8 c4c5 e4d2 d1d2 c6b8 f1e2 c7c6 e1g1 b8d7 a3a4 e7f6 b2c3 d8e7 d2b2 a7a6 f3e5 h7h6 f2f4 f6h4 a1a3 g8h8 e5f3 f5e4 f1c1 e4f3 e2f3 f7f5 g2g3 h4f6 c3e1');
INSERT INTO evaluations VALUES(26828,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(26828,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(26828,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(26828,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(26828,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(26828,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(26828,36,'cp 50','c2c4 g8f6 b2b4 f8e7 c1b2 f6e4 b1d2 e4d2 d1d2 c6b8 c4c5 e8g8 f1e2 b8d7 a3a4 c7c6 b2c3 a7a6 e1g1 e7f6 d2b2 d8e7 f3e5 h7h6 f2f4 f6h4 e5f3 g8h8 a1a3 f5g4 g2g3 h4f6 e2d3 g4f5 b2e2 f5d3 e2d3 f8e8 f1b1');
INSERT INTO evaluations VALUES(35019,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(35019,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(35019,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(35019,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(35019,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(35019,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(35019,36,'cp 165','b8c6 d2d4 d7d6 e1g1 c8g4 c3d5 f6d8 c2c3 g8f6 d1b3 a8b8 c1e3 a7a6 f3d2 b7b5 d5f6 d8f6 c4d5 c6e7 f2f4 e7d5 f4e5 f6g6 b3d5 g4d7 e3f4 f8e7 f1f3 g6e6 d5b3 e8g8 f3g3 g7g5 h2h4 g5g4 e5d6 c7d6 b3e6 d7e6 h4h5 e7g5 f4g5 h6g5 g1f2 b5b4 c3c4 f7f5 e4f5 e6f5');
INSERT INTO evaluations VALUES(38555,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(38555,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(38555,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(38555,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(38555,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(38555,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(38555,36,'cp 92','g1f3 g8f6 f1e2 e8g8 c1e3 e6e5 e1g1 e5d4 f3d4 b8c6 d1d2 c6d4 e3d4 c8d7 e2f3 a7a6 a1d1 d7e6 b2b3 c7c6 d4e3 b7b5 e4e5 d6e5 f3c6 d8d2 d1d2 a8c8 c3d5 e6d5 c4d5 f6e4 d2c2 e4d6 e3b6 e5e4 f1d1 f7f5 g1h1 g7f6 f2f3 f6g7 f3e4');
INSERT INTO evaluations VALUES(39008,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(39008,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(39008,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(39008,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(39008,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(39008,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(39008,36,'cp 76','b1c3 g8e7 g1f3 d7d5 c4b3 a7a5 a2a4 d5e4 c3e4 h7h6 c1f4 b8c6 e1g1 b7b6 f1e1 c8b7 d4d5 e6d5 b3d5 e8g8 d5c4 c6b4 c2c3 d8d1 a1d1 b4d5 f4g3 a8d8 d1d2 d8d7 g3e5 d7d8 h2h4 g7e5 f3e5 f7f6 e5f3 g8g7 h4h5 d5f4 d2d8 f8d8');
INSERT INTO evaluations VALUES(39255,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(39255,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(39255,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(39255,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(39255,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(39255,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(39255,36,'cp -25','c5e3 f2e3 g8f6 g1f3 e8g8 c2c4 c7c6 b1c3 d7d5 e4d5 c6d5 f3e5 b8c6 d3d4 d5c4 e5c6 b7c6 d1d2 c8e6 f1e2 a8b8 e2f3 e6d5 e1g1 d5f3 f1f3 c6c5 d4d5 f6d7 b2b3 c4b3 a2b3 d8b6 d5d6 b6b3 a1a7 b3e6 h2h3 d7e5');
INSERT INTO evaluations VALUES(39393,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(39393,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(39393,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(39393,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(39393,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(39393,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(39393,36,'cp -14','f2e3 g8f6 g1f3 e8g8 c2c4 f8e8 b1c3 c7c6 f1e2 d7d5 c4d5 c6d5 e4d5 f6d5 c3d5 d8d5 e1g1 h7h6 d3d4 b8c6 b2b3 d5e4 d1d2 c8e6 e2d3 e4g4 d4e5 a8d8 a1d1 g4h5 d2c3 e6g4 d3e2 d8c8 h2h3 g4f3 e2f3 c6e5 c3b4 e5f3 f1f3 e8e3 b4b7 c8e8 b7d5 e3e5 d5d7');
INSERT INTO evaluations VALUES(39392,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(39392,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(39392,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(39392,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(39392,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(39392,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(39392,36,'cp -11','g8f6 g1f3 e8g8 c2c4 c7c6 b1c3 f8e8 f1e2 d7d5 c4d5 c6d5 d1b3 d5d4 e3d4 e5d4 c3d5 b8c6 e1g1 f6d5 b3d5 c8e6 d5d8 a8d8 g1f2 f7f5 e4f5 e6f5 f1e1 e8f8 a2a3 d8e8 b2b4 f5g6 f2g1 h7h6 b4b5 c6e5 f3d4 e5d3');
INSERT INTO evaluations VALUES(44453,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(44453,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(44453,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(44453,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(44453,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(44453,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(44453,36,'cp -110','b8c6 e2e3 e7e5 c2c4 d7d5 c4d5 f6d5 a2a3 c8e6 g1f3 f8e7 e3e4 d5f6 d1d8 a8d8 f1b5 f6e4 f3e5 e6d5 e1g1 e8g8 b5c6 b7c6 c1e3 f8e8 b1c3 e4c3 b2c3 e7d6 e5f3 c6c5 f1d1 d5c6 a3a4 f7f6 a4a5 g8f7 a5a6 d6e7 d1d8 e8d8 f3d2');
INSERT INTO evaluations VALUES(12904,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(12904,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(12904,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(12904,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(12904,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(12904,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(12904,36,'cp -95','c2c4 e7e5 a2a3 d7d5 c4d5 d8d5 d1d5 f6d5 e2e3 d5b6 c1d2 c8e6 d2c3 f8e7 f1d3 a8c8 g1f3 f7f6 e1e2 b6a4 d3b5 a4c3 b1c3 e8f7 h1d1 a7a6 b5d3 c6a5 f3d2 f6f5 a1c1 e5e4 d3b1 e7f6 f2f3 e4f3 d2f3 h8d8 d1d8 c8d8');
INSERT INTO evaluations VALUES(44506,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(44506,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(44506,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(44506,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44506,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44506,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44506,36,'cp 0','e7e6 c1e3 b8a6 e3d4 f8c5 b1c3 d8b6 e2e4 e6e5 d4c5 a6c5 f3d2 b6b2 c3b5 e8g8 d2b3 c5e4 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(52975,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52975,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52975,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(52975,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(52975,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(52975,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(52975,36,'cp 41','d6d5 e4e5 c8f5 g1f3 e7e6 f1e2 h7h5 e1g1 g8h6 h2h3 c6c5 c1e3 b8c6 b1d2 c5c4 c2c3 f5g6 e3f2 f8e7 b2b4 a7a6 a2a4 c6a7 f3e1 h6f5 e1c2 e7h4 d1e1 h4e7 g2g4 h5g4 h3g4 f5h6 c2e3 f7f5 g4g5 h6f7');
INSERT INTO evaluations VALUES(53651,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(53651,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(53651,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(53651,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(53651,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(53651,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(53651,36,'cp 182','g1f3 g8f6 e4e5 f6d5 e1g1 f8e7 b1c3 d5c3 b2c3 c6c5 d1e2 c8d7 d4d5 e6d5 c4d5 b8c6 f1d1 d8c7 c1f4 d7e6 d5e6 f7e6 e2c4 c6d8 f3d4 c7c8 d4b5 e8g8 f4g3 a7a6 b5d6 e7d6 e5d6 b7b5 c4e4 c8c6 e4c6 d8c6 d6d7 a8d8 g3c7');
INSERT INTO evaluations VALUES(53650,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(53650,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(53650,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(53650,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(53650,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(53650,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(53650,36,'cp 177','g8f6 e4e5 f6d5 e1g1 c6c5 d1c2 d5b4 c2e4 c8d7 a2a3 d7c6 e4e2 b4d5 f1d1 a7a6 d4c5 f8c5 b1c3 e8g8 e2c2 c5e7 c3d5 c6d5 c4d5 e6d5 c2b3 d8c8 c1e3 b8c6 d1d5 c8e6 a1d1 h7h6 h2h3 b7b5 e3b6 a8c8 b3e3 f8e8 b2b4 e6f5 b6c5 c8d8 c5e7 d8d5 d1d5 c6e7 d5c5');
INSERT INTO evaluations VALUES(55188,30,'cp -127',NULL);
INSERT INTO evaluations VALUES(55188,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(55188,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(55188,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(55188,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(55188,35,'cp -134',NULL);
INSERT INTO evaluations VALUES(55188,36,'cp -138','b8c6 g1e2 c6a5 c4b3 e7e6 b1c3 a5b3 a2b3 f7f5 g2g4 f5g4 h1g1 g8f6 c1g5 f8e7 d1d2 h7h6 g5f4 e8g8 f4e5 c7c5 d4c5 e7c5 d2d8 f8d8 e5f6 g7f6 g1g4 g8f7 g4e4 e6e5 e4c4 b7b6 b3b4 c8e6 c4h4 c5f8 f2f4 f8b4 f4e5');
INSERT INTO evaluations VALUES(3824,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3824,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(3824,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3824,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3824,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3824,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3824,36,'cp 0','f6e4 d2d4 e5d4 c3d4 e4f2 g1f2 c6d4 f2g3 d6d5 f3d4 c5d6 g3f2 d8h4 f2f1 d5c4 d4f3 h4h5 f3d4');
INSERT INTO evaluations VALUES(9781,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9781,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9781,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(9781,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(9781,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9781,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9781,36,'cp 0','d7d6 d2d3 a7a6 c1e3 c5e3 f2e3 c8e6 f3d2 b7b5 c4b3 e8g8 d1f3 a8b8 e1g1 c6e7 f1f2 e7g6 a1f1 c7c6 b3e6 f7e6 c3e2 d6d5 f3g3 g8h7 d2f3 d8d6 e4d5 e6d5 b2b3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(8853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8853,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(8853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8853,36,'cp 0','d2d3 a7a6 c1e3 e8g8 e3c5 d6c5 e1g1 d8d6 a3a4 c8e6 c4e6 d6e6 f3d2 b7b6 d2c4 a8d8 c4e3 e6d6 e3c4');
INSERT INTO evaluations VALUES(15950,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(15950,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(15950,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(15950,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(15950,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(15950,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(15950,36,'cp 5','d1c2 c7c5 b1c3 d5c4 f1c4 a7a6 a2a4 d8c7 c2e2 f8d6 e1g1 e8g8 a4a5 c5d4 e3d4 h7h6 h2h3 b7b6 a5b6 d7b6 c4d3 c8b7 c1d2 b6d5 d3a6 c7b6 a6b7 a8a1 f1a1 b6b2 a1b1 d5c3 b1b2 c3e2 g1f1 e2f4 d2f4 d6f4 f1e2 f8d8 e2e1');
INSERT INTO evaluations VALUES(16882,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(16882,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(16882,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16882,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(16882,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(16882,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(16882,36,'cp -7','f6d5 e4d5 d8e7 d1e2 c7c6 b5d3 b7b6 b2b3 c8b7 e1g1 e7e2 d3e2 f8c5 d5c6 d7c6 e2g4 e8d8 g4f3 d8c7 d2d3 c5d6 c1d2 c6c5 f3b7 c7b7 f1e1 a8e8 a2a3 b7c6 h2h3 h7h5 g1f1 e8e7 a3a4 e7e1 a1e1 c6d5 h3h4');
INSERT INTO evaluations VALUES(17072,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17072,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17072,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(17072,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(17072,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(17072,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17072,36,'cp 0','e4d5 d8e7 d1e2 c7c6 b5c4 b7b5 c4d3 c8b7 e1g1 e7e2 d3e2 e8c8 b2b3 f8c5 c1b2 h8e8 e2d3 e8e5 a2a4 a7a6 d5c6 d7c6 b2a3 c5a3 a1a3 c8c7 a3a1 d8e8 b3b4 h7h5 a4b5 c6b5 f2f3 h5h4 g1f2 e8e6');
INSERT INTO evaluations VALUES(17071,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17071,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(17071,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(17071,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17071,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17071,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17071,36,'cp -1','d8e7 d1e2 c7c6 b5d3 e7e2 d3e2 b7b6 b2b3 c8b7 c1b2 f8c5 a2a3 a7a5 d5c6 b7c6 e1g1 e8g8 e2d3 f8e8 a1e1 h7h5 e1e8 a8e8 h2h4 g8f8 b3b4 a5b4 a3b4 c5b4 b2d4 b4d2 f1b1 e8e1 b1e1 d2e1 d4b6 e1c3 b6c5 f8e8 g1f1 c3f6 g2g3');
INSERT INTO evaluations VALUES(17752,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(17752,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(17752,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(17752,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(17752,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(17752,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(17752,36,'cp 41','b7c6 d2d4 f7f6 b1c3 d8d7 c1e3 g8e7 b2b4 e7g6 d1d3 c8b7 d4d5 f6f5 e4f5 g6e7 d5c6 d7c6 f1b1 e5e4 b4b5 e4d3 b5c6 b7c6 f3d4 d3c2 b1b2 c6d7 b2c2 d7f5 d4f5 e7f5 c3d5 a8c8 d5c7 e8f7 a1c1 f8e7 c2c6 f5e3 f2e3 e7g5 c1c3 g5f6 c3c4 f6e5 c6a6 h8f8 e3e4 f7e7');
INSERT INTO evaluations VALUES(31477,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31477,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31477,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(31477,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31477,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31477,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(31477,36,'cp 14','c1e3 g8e7 b1d2 e7f5 f1d3 f5e3 f2e3 c6c5 g2g4 h5g6 d3g6 h7g6 d1e2 d8a5 c2c3 f8e7 d4c5 a5c5 e3e4 d5e4 e2e4 b8d7 e1c1 c5c7 d2c4 d7c5 e4e3 e8g8 h3h4 a8d8 h4h5 b7b5 h5g6 b5c4 d1d8 f8d8');
INSERT INTO evaluations VALUES(31944,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(31944,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(31944,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(31944,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(31944,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(31944,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(31944,36,'cp 88','c2c4 f7f6 e5f3 e7e6 b1c3 c7c6 e2e3 b8d7 f1e2 f5g6 h2h4 f8d6 h4h5 g6f7 e3e4 g8e7 h5h6 g7g5 e1g1 e8g8 f1e1 g5g4 f3h4 g4g3 e4d5 e6d5 c4d5 g8h8 c3e4 g3f2 g1f2 d6b4 c1d2 c6d5 d2b4 d5e4 b4e7 d8e7 h4f5 e7b4 d1c2 f7g6 c2e4 g6f5 e4f5');
INSERT INTO evaluations VALUES(31943,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(31943,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(31943,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(31943,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(31943,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(31943,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(31943,36,'cp 94','f7f6 e5f3 c7c6 b1c3 e7e6 e2e3 b8d7 f1e2 f5g6 h2h4 f8d6 h4h5 g6f7 e3e4 d5e4 c3e4 d6c7 d1b3 g8e7 h5h6 g7g5 c1d2 e8g8 e1c1 b7b5 c1b1 b5c4 b3c4 g8h8 b1a1 a8b8 c4c2 e7d5 c2c6 d7b6 c6c2 d5f4 d2f4 c7f4 g2g3 b8c8');
INSERT INTO evaluations VALUES(34803,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(34803,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(34803,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(34803,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(34803,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(34803,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(34803,36,'cp 143','e5d4 c3b5 d6e7 d1e2 b8a6 c1f4 e7e2 f1e2 f8b4 e1f1 g8f6 f4c7 e8g8 c7e5 d4d3 e2d3 b4e7 a1e1 c8e6 b5d4 e6d7 h2h4 f8e8 d3f5 g7g6 f5d7 f6d7 e5f4 e7f6 e1e8 a8e8 g2g3 d7e5 f4e5 f6e5 f3e5 e8e5 c2c3 a6c5 f1g2 c5a4 h1b1 e5d5 a2a3');
INSERT INTO evaluations VALUES(34809,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(34809,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(34809,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(34809,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(34809,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(34809,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(34809,36,'cp 169','c3b5 d6e7 f1e2 a7a6 b5d4 c7c5 d4b3 g8f6 c1f4 e7e4 f4d6 e4b4 c2c3 b4b6 d6f8 h8f8 e1g1 c8e6 f1e1 b8d7 e2c4 e8c8 c4e6 f7e6 b3d2 b6c6 d1e2 e6e5 a1d1 c8b8 f3e5 d7e5 e2e5 b8a7 d2c4 d8d1 e1d1 f8e8 e5g3 e8e7 b2b3 c6c7 g3c7 e7c7 c4d6 a7b8');
INSERT INTO evaluations VALUES(37894,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(37894,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(37894,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(37894,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(37894,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(37894,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(37894,36,'cp 96','g2g4 d6d5 e4d5 g8f6 d5c6 b8c6 d4c6 d8d1 c3d1 b7c6 f1g2 c8d7 d1e3 e8c8 g4g5 f6g4 e3g4 d7g4 c1e3 c8c7 e1g1 a7a5 h2h3 g4e6 a1d1 h7h6 g5h6 g7h6 d1d8 e7d8 b2b3 h8g8 g1h1 h6h5 f1d1 d8g5 d1d3 g5e3 d3e3 g8d8 e3c3 d8d2 c3c6 c7d7 c6c5 d2f2 c5h5 f2c2 h5a5');
INSERT INTO evaluations VALUES(41272,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(41272,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(41272,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41272,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(41272,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41272,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(41272,36,'cp 51','e8g8 e2e4 e7e5 d4d5 f6h5 g2g3 f7f5 e4f5 g6f5 f3g5 h5f6 h1g1 b8a6 c1e3 c7c6 d1d2 d8a5 a2a3 c6d5 c4d5 c8d7 g5e6 d7e6 d5e6 d6d5 e3h6 d5d4 d2g5 a5c7 g3g4 f5f4 c3b5 c7e7 f1c4 g8h8 g5g7 e7g7 h6g7 h8g7');
INSERT INTO evaluations VALUES(36544,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(36544,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(36544,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(36544,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(36544,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(36544,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(36544,36,'cp 49','e2e4 e7e5 d4d5 f6h5 g2g3 f7f5 e4f5 g6f5 c1g5 h5f6 f3h4 b8a6 f1d3 e5e4 d3e2 d8e8 d1d2 f6d7 e1g1 d7e5 f2f3 e4f3 h4f3 a6c5 a1e1 c5e4 c3e4 f5e4 f3e5 f8f1 e1f1 e8e5 g1g2 c8d7 b2b3 a8e8 d2f4 e8f8 f4e5 g7e5 f1f8 g8f8 g5h6 f8g8 g3g4 c7c6 a2a4 c6c5 h6g5 g8g7 h3h4 h7h6 g5e7');
INSERT INTO evaluations VALUES(42176,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(42176,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(42176,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(42176,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(42176,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(42176,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(42176,36,'cp 135','e8g8 c4d3 h7h6 e1g1 c7c5 e3e4 b8c6 e4e5 f6d5 d4c5 d5c3 d1c2 c3d5 c2c4 c8d7 a1b1 c6e7 b1b7 a8b8 b7b8 d8b8 c1d2 f8d8 f1b1 b8c7 d3c2 d7c6 h2h3 e7g6 c2g6 f7g6 f3d4 c7d7 d4c6 d7c6 c4d4 g8h7 b1c1 d8c8');
INSERT INTO evaluations VALUES(37087,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(37087,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(37087,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(37087,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(37087,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(37087,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(37087,36,'cp 145','c4d3 h7h6 e1g1 c7c5 e3e4 b8c6 d3c2 c5d4 c3d4 c6e7 c1a3 f8e8 a1b1 b7b6 f1e1 c8a6 c2a4 b6b5 a4b5 a6b5 b1b5 e7g6 g2g3 a8b8 b5c5 d8d7 d1d3 b8c8 c5c8 d7c8 g1g2 c8a8 h2h3 e8d8 a3c5 d8c8 h3h4 a7a5 f3d2 e6e5 d2f3 a8c6 d3b1 a5a4 a2a3 f6g4 h4h5 e5d4 h5g6 c6c5');
INSERT INTO evaluations VALUES(46538,30,'cp 414',NULL);
INSERT INTO evaluations VALUES(46538,31,'cp 397',NULL);
INSERT INTO evaluations VALUES(46538,32,'cp 425',NULL);
INSERT INTO evaluations VALUES(46538,33,'cp 412',NULL);
INSERT INTO evaluations VALUES(46538,34,'cp 390',NULL);
INSERT INTO evaluations VALUES(46538,35,'cp 436',NULL);
INSERT INTO evaluations VALUES(46538,36,'cp 442','d4e5 b8c6 e5f6 g7f6 b1c3 c6a5 c4b5 a5c6 f3d4 f8g7 d4c6 b7c6 d1h5 e8f8 b5a4 c8e6 c1d2 d8e8 h5a5 e8c8 e1g1 a8b8 a4b3 c6c5 b3e6 c8e6 a5c7 b8b2 d2f4 f8f7 a1d1 h8c8 c7a7 c8d8 h2h3 b2c2 c3b5 c2a2 a7b6 d8d7 d1d6 d7d6 b5d6 f7g8 b6c5');
INSERT INTO evaluations VALUES(46534,30,'cp 445',NULL);
INSERT INTO evaluations VALUES(46534,31,'cp 443',NULL);
INSERT INTO evaluations VALUES(46534,32,'cp 412',NULL);
INSERT INTO evaluations VALUES(46534,33,'cp 431',NULL);
INSERT INTO evaluations VALUES(46534,34,'cp 402',NULL);
INSERT INTO evaluations VALUES(46534,35,'cp 415',NULL);
INSERT INTO evaluations VALUES(46534,36,'cp 424','b8d7 e5f6 d7f6 b1c3 e7c6 c1e3 c8g4 h2h3 g4f3 d1f3 d8d7 f3e2 e8c8 e1c1 d8e8 f2f3 h7h6 c4b5 a7a6 b5a4 g7g5 c3d5 f6d5 e4d5 d7e7 d1e1 e7e3 e2e3 e8e3 e1e3 c6e5 c2c3 f8g7 a4c2 h8f8 a2a4 e5c4 e3e2 b7b5 a4b5 a6b5 c2d3 c4e5 e2e3 b5b4 c1c2 b4c3 d3a6 c8d7 a6b5 d7e7 e3c3 f8a8');
INSERT INTO evaluations VALUES(50886,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(50886,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(50886,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(50886,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(50886,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(50886,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(50886,36,'cp 58','g8f6 e1g1 f8e7 c4d3 c7c5 d4c5 d6c5 e4e5 f6d5 d1e2 d8c7 c3e4 b7b5 b2b3 b5b4 c1b2 b8c6 a1d1 c8b7 f1e1 a8d8 e2d2 d5c3 b2c3 b4c3 d2c3 e8g8 a2a3 c6d4 f3d4 d8d4 e4d2 a6a5 d2c4 a5a4 b3a4 c7c6 f2f3 c6a4 c4d6 e7d6 e5d6 a4c6 e1e5 f8d8 c3c5');
INSERT INTO evaluations VALUES(51198,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(51198,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(51198,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(51198,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(51198,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(51198,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(51198,36,'cp -3','c4b3 c8b7 e1g1 d7d6 c2c3 g8f6 f1e1 b8d7 h2h3 d8c7 c1f4 f8e7 b1d2 e8g8 f4g3 f8c8 d1e2 e7f8 a2a4 c8e8 b3c2 f6h5 g3h2 g7g6 d2f1 h5f6 f3d2 a8c8 f1e3 b7c6 f2f4 f8g7 e2f2 b5b4 d2f3 b4c3 b2c3');
INSERT INTO evaluations VALUES(51210,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(51210,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(51210,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(51210,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(51210,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(51210,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(51210,36,'cp -13','c8b7 e1g1 d7d6 f1e1 b8d7 h2h3 g8f6 c2c3 d8c7 b1d2 f8e7 b3c2 e8g8 d2f1 f8c8 a2a4 e7f8 c1f4 a8b8 d1e2 e6e5 f4d2 c8e8 a4b5 a6b5 c3c4 b5c4 d3c4 g7g6 d2a5 c7c8 a5c3 f8h6 a1a7 d7f8 c2a4 f6e4');
INSERT INTO evaluations VALUES(30331,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(30331,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(30331,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(30331,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(30331,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(30331,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(30331,36,'cp -9','e1g1 d7d6 h2h3 g8f6 c2c3 b8d7 b3c2 g7g6 c1e3 f8g7 b1d2 e8g8 d3d4 d8c7 a2a4 a8c8 a4b5 a6b5 d1e2 b7c6 e3f4 c5c4 e2e3 c8a8 f1e1 e6e5 f4h6 g7h6 e3h6 a8a1 e1a1 f8a8 a1a8 c6a8 c2b1 a8c6 d4d5 c6b7 f3g5 d7f8 d2f1');
INSERT INTO evaluations VALUES(52646,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(52646,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(52646,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(52646,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(52646,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(52646,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(52646,36,'cp 114','c3d4 f8g7 b1c3 c8g4 f1b5 b8d7 c1e3 a7a6 b5e2 e7e6 e1g1 g8e7 d4d5 e6d5 c3d5 e7d5 d1d5 e8g8 d5d6 d7e5 a1d1 e5f3 g2f3 g4e6 d6d8 f8d8 b2b3 b7b5 f3f4 d8c8 d1c1 g7b2 c1c8 a8c8 f1d1 b2f6 g1g2 c8c2 d1d2 c2d2 e3d2 e6d7 e2d3 d7c6 f2f3 f6e7');
INSERT INTO evaluations VALUES(5849,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(5849,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(5849,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(5849,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(5849,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(5849,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(5849,36,'cp 81','g8f6 a2a3 b4c3 b2c3 c6a5 d1a4 c7c6 c4d3 c8d7 a4d1 c6c5 e1g1 d7c6 f1e1 c6f3 d1f3 e8g8 a1b1 a8c8 e3e4 c5d4 e4e5 f6d5 c3d4 b7b6 c1d2 a5c6 f3h3 g7g6 e1e4 d8d7 b1e1 f7f5 e5f6 d5f6 h3e6 d7e6 e4e6 c6d4 e6e7 f8f7 d3a6 c8c7 e7f7 g8f7 d2f4 c7d7 f2f3');
INSERT INTO evaluations VALUES(11525,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(11525,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(11525,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(11525,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(11525,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(11525,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(11525,36,'cp -42','f6e4 c3e4 d7d5 c4d3 f7f5 e4c3 e5e4 d3e2 e4f3 e2f3 d5d4 d1e2 e8f7 e2c4 c8e6 f3d5 d8d6 d5e6 d6e6 c4e6 f7e6 c3e2 a8e8 d2d3 e6d7 c1f4 c6e7 e1f1 e7d5 f4d2 c7c5 h2h4 f8d6 h4h5 e8e7 e2g1 h8e8 g1f3 b7b6 h1h4 d5f6 h5h6 g7g6 d2g5 e7e6 g5f6 e6f6 a1e1 e8e1 f1e1 d7e7 e1d1');
INSERT INTO evaluations VALUES(14870,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(14870,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(14870,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(14870,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(14870,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(14870,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(14870,36,'cp 55','e4e5 f6d7 f3f4 c6e7 g1f3 c7c5 a2a3 e7c6 c1e3 c5d4 f3d4 f8c5 d1d2 c6d4 e3d4 c5d4 d2d4 a7a6 h2h4 d8b6 d4b6 d7b6 a3a4 b6d7 f1d3 e8g8 e1c1 d7c5 a4a5 c5d3 d1d3 c8d7 h4h5 h7h6 g2g4 a8c8 b2b3 f7f6 e5f6 f8f6 c3e2');
INSERT INTO evaluations VALUES(14887,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(14887,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(14887,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14887,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(14887,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(14887,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(14887,36,'cp 50','f6d7 f3f4 c6e7 a2a3 c7c5 g1f3 e7c6 c1e3 d8b6 a1b1 f8e7 d1d2 e8g8 f1e2 c5d4 f3d4 b6c7 c3b5 c7d8 b2b4 a7a6 b5c3 d8c7 e1g1 b7b5 e2d3 d7b6 d2f2 c6d4 e3d4 b6c4 b1a1 c8b7 f1e1 f7f6 e5f6 e7f6 e1e6 f6d4 f2d4 c7f4');
INSERT INTO evaluations VALUES(15893,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(15893,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(15893,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(15893,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(15893,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(15893,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(15893,36,'cp 51','f3f4 c6e7 g1f3 c7c5 a2a3 e7c6 c1e3 c5d4 f3d4 f8c5 d1d2 c5d4 e3d4 c6d4 d2d4 d8b6 c3b5 b6d4 b5d4 e8e7 f1d3 d7b8 b2b4 c8d7 c2c3 f7f6 e1d2 b8c6 a3a4 f6e5 d4c6 d7c6 f4e5 c6e8 h1f1 e8g6 a4a5 g6d3 d2d3 a8f8 f1f8 h8f8 b4b5 f8c8');
INSERT INTO evaluations VALUES(15759,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(15759,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(15759,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(15759,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(15759,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(15759,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(15759,36,'cp 78','d1d4 f8e7 c1f4 e8g8 e1c1 c8e6 c1b1 f6d7 c4e2 e7f6 d4e3 d8e7 h2h4 f6e5 g2g3 d7b6 b2b3 a8d8 f4e5 d6e5 f2f4 d8d1 c3d1 e5f4 g3f4 f7f6 a2a4 g8h8 d1c3 e7b4 b1b2 f8d8 e4e5 f6e5 e3e5 e6g8 a4a5 d8d2 e5b5 b4b5 e2b5 b6d7');
INSERT INTO evaluations VALUES(34536,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(34536,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(34536,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(34536,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(34536,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(34536,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(34536,36,'cp 30','b8c6 g1f3 g8f6 d2d4 c8g4 d4d5 c6e5 c4b5 f6d7 b5e2 g4f3 e2f3 e5f3 d1f3 e7e5 d5e6 d6e6 c1e3 f8c5 e1g1 c5e3 a1e1 e8g8 e1e3 e6b6 b2b3 d7f6 h2h3 a8d8 c3a4 b6b5 c2c4 b5b4 a2a3 b4a3 f3b7 d8b8 b7f3 f8e8 e3d3 a3e7 f3c6 b8d8 a4c5 g7g6 f1d1 d8d3');
INSERT INTO evaluations VALUES(40218,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(40218,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(40218,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40218,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40218,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40218,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40218,36,'cp 23','d4c5 e8g8 c4d5 e6d5 b2b4 a7a5 b4b5 e7c5 f1d3 c8e6 e1g1 b8d7 c1b2 d8e7 h2h3 f8c8 c3e2 a5a4 a1c1 a4a3 b2d4 c5d4 f3d4 c8c1 d1c1 d7e5 e2f4 a8c8 c1a1 e6d7 f1c1 c8c5 c1c5 e7c5 d3f5 c5c4 f5d7 f6d7 g1h1');
INSERT INTO evaluations VALUES(43351,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43351,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43351,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(43351,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(43351,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(43351,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(43351,36,'cp 35','f2f3 b8c6 c1e3 d8b6 a2a3 e7e5 d4f5 b6d8 f5g3 c8e6 d1d2 f8g7 e1g1 e8g8 a1d1 a8c8 g1h1 c6d4 d2f2 f6d7 g3e2 d4c6 h2h3 b7b5 f3f4 e5f4 e2f4 g7c3 b2c3 c6e5 a3a4 b5a4 d3a6 c8b8 d1a1 e6c4 a6c4 e5c4 a1a4 c4e3 f2e3');
INSERT INTO evaluations VALUES(44681,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(44681,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(44681,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(44681,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(44681,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(44681,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(44681,36,'cp 102','a2a3 c6c5 d4c5 e7g6 h2h4 b8c6 b2b4 a7a5 h4h5 g6e5 f3e5 c6e5 c1b2 f7f6 h5h6 g7g6 b1c3 a5b4 a3b4 a8a1 d1a1 e5c6 f1b5 e8f7 a1a4 d8c7 c3e2 e6e5 f2f4 e5f4 a4b3 c8e6 b5c6 b7c6 b3c3 f8e7 e1g1 h8e8 e2f4 e6f5 f4e2 c7d7 e2g3 d5d4 c3d2 f5e6');
INSERT INTO evaluations VALUES(46144,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(46144,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(46144,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(46144,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(46144,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(46144,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(46144,36,'cp 46','c4d5 e6d5 b1c3 c8e6 g5f6 d8f6 e2e4 d5e4 c3e4 f6f4 f1b5 b8c6 d1e2 e8c8 b5c6 b7c6 e1g1 e6d5 a1e1 c8b8 e4c5 f8c5 d4c5 d5f3 e2f3 f4f3 g2f3 h8e8 f3f4 b8b7 b2b3 e8e1 f1e1 d8d7 e1e8 b7a6 e8h8 g7g6 h8h7 a6b5 h2h4 b5c5');
INSERT INTO evaluations VALUES(46126,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(46126,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(46126,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46126,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(46126,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(46126,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(46126,36,'cp 42','e6d5 b1c3 c8e6 g5f6 d8f6 e2e4 b8d7 f1b5 e8c8 e1g1 c5d4 b5d7 d8d7 e4e5 f6d8 c3b5 c8b8 d1d4 d8a5 a2a4 a5b6 d4b6 a7b6 f3d4 f8c5 a1d1 h8e8 h2h3 d7e7 d4e6 e7e6 d1d5 e6e5 d5d7 e5e7 f1d1 e7d7 d1d7 e8e1 g1h2 f7f6 d7g7 c5f2');
INSERT INTO evaluations VALUES(46409,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(46409,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(46409,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(46409,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(46409,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46409,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(46409,36,'cp 44','b1c3 c5d4 f3d4 f8e7 e2e3 e8g8 f1e2 c8e6 e1g1 b8c6 g5f4 a8c8 d4c6 c8c6 c3b5 d8b6 a2a4 a7a5 f4e5 f6e4 b5d4 c6c8 e2b5 h7h6 h2h3 f8d8 d1h5 e7f6 e5f6 e4f6 h5e2 c8c5 a1c1 d8c8 c1c5 c8c5 e2d3 e6d7 b5d7 f6d7 b2b3');
INSERT INTO evaluations VALUES(51836,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(51836,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(51836,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(51836,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(51836,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(51836,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(51836,36,'cp -77','g8f6 d2d4 e5d4 d1d4 d7d5 e4d5 c6d5 c4b5 b8c6 g1f3 f8e7 e1g1 e8g8 d4d1 c8g4 c1e3 d5d4 e3f4 d8d5 b5e2 f6h5 f4g3 a8d8 b1d2 h5g3 h2g3 e7f6 e2d3 c6e5 d3e4 d5a5 e4d3 e5d3 c2d3 f8e8 f1e1 e8e1 d1e1 a5b5 d2e4');
INSERT INTO evaluations VALUES(53925,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(53925,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(53925,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(53925,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(53925,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(53925,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(53925,36,'cp 127','b1c3 g7g6 c1e3 h7h5 d1d2 b8d7 e1c1 f8h6 e3h6 g8h6 g2g3 d7b6 f2f4 f7f6 f1e2 c8g4 h2h3 g4e2 d2e2 d8e7 g1f3 e8c8 f3h4 e7h7 f4e5 f6e5 h1f1 h7g7 a2a4 a7a6 h4f3 g6g5 a4a5 b6d7 h3h4 g5h4 g3h4 g7g4 f1g1 g4f4 c1b1');
INSERT INTO evaluations VALUES(53922,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(53922,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(53922,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(53922,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(53922,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(53922,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(53922,36,'cp 134','g7g6 f1e2 h7h5 h2h3 b8a6 g1f3 f8h6 c1h6 g8h6 d1d2 d8f6 a2a3 c8d7 e1g1 e8f8 b2b4 f8g7 f1b1 a8f8 d2g5 f6g5 f3g5 f7f6 g5e6 d7e6 d5e6 f8e8 c3b5 e8e6 b5a7 f6f5 e2f3 f5f4 b4b5 a6c7 b5b6 c7a6 a7b5 h6f7 b5c7 e6e7 c7a6 b7a6 g1h1 e7b7 a3a4');
INSERT INTO evaluations VALUES(56424,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(56424,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(56424,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(56424,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(56424,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(56424,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(56424,36,'cp -112','b8c6 c1b2 h7h5 f1b5 g8e7 b2e5 a7a6 b5c6 e7c6 e5b2 h5h4 d1e2 h4g3 f2g3 d8g5 g1f3 g5h5 d2d3 c6b4 e1d1 c8g4 b1d2 e8c8 a2a3 b4c6 d1c1 d8e8 e2f2 h5h6 h1e1 f8c5 d3d4 c5d6 b2c3 e8e6 c1b2 h8e8 f3h4 c8b8 h4f5 g4f5 f2f5 h6h2 f5d5');
INSERT INTO evaluations VALUES(21728,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(21728,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(21728,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(21728,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(21728,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(21728,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(21728,36,'cp -112','c1b2 h7h5 f1b5 f8d6 d1e2 g8e7 b1c3 c8e6 f2f4 f7f6 e2g2 d8d7 e1c1 e8c8 g1f3 e6h3 g2f2 a7a6 b5f1 e5e4 f1h3 d7h3 f3g1 h3e6 c1b1 d6c5 g1e2 c8b8 c3a4 c5a7 h2h3 b7b5 a4c3 f6f5 g3g4 f5g4 h3g4 e6g4 d1g1 g4f3 f2h2 g7g6 g1g5 d5d4 c3d1');
INSERT INTO evaluations VALUES(21729,30,'cp -102',NULL);
INSERT INTO evaluations VALUES(21729,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(21729,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(21729,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(21729,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(21729,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(21729,36,'cp -108','h7h5 f1b5 f8d6 d1e2 g8e7 b1c3 c8e6 f2f4 f7f6 e2g2 d8d7 e1c1 e8c8 g1f3 e6h3 g2f2 h8f8 c1b1 c8b8 b5f1 h3e6 a2a3 e5e4 f3h4 g7g5 h4g2 a7a6 d2d3 e4d3 f1d3 d6c5 d3e2 d7e8 h1f1 c5b6 f1e1 e6g4 e2g4 h5g4 c3a4 b6a7');
INSERT INTO evaluations VALUES(2125,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(2125,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(2125,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2125,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(2125,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(2125,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(2125,36,'cp 36','e7e6 e1g1 g8e7 d2d4 c6d4 f3d4 c5d4 d1d4 e7c6 d4e3 a7a6 b5c6 b7c6 b2b3 c8b7 c3a4 c6c5 c1a3 a8c8 a1d1 f8e7 c2c4 e8g8 d1d3 f7f6 f1d1 d7d6 a4c3 c8e8 e3e2 f8f7 a3c1 c7c8 c1f4 e6e5 f4g3 f6f5 e4f5 c8f5');
INSERT INTO evaluations VALUES(2102,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(2102,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2102,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(2102,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(2102,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(2102,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(2102,36,'cp 27','e1g1 g8e7 d2d4 c6d4 f3d4 c5d4 d1d4 e7c6 d4e3 f8e7 b5e2 a7a6 c3a4 b7b5 a4b6 a8b8 b6c8 b8c8 c2c3 e8g8 e3g3 e7d6 f2f4 c6e7 g1h1 c7c6 e2d3 f7f5 g3f3 d6c5 c1d2 f5e4 f3e4 g7g6 a1e1 e7d5 g2g3 d5f6 e4c6 c8c6');
INSERT INTO evaluations VALUES(2101,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2101,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(2101,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(2101,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(2101,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(2101,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(2101,36,'cp 20','g8e7 d2d4 c5d4 f3d4 c6d4 d1d4 e7c6 d4e3 a7a6 b5c6 b7c6 b2b3 c6c5 c1b2 c8b7 a1d1 f8e7 c3e2 e8g8 c2c4 a8d8 e2g3 a6a5 a2a4 d8e8 f2f4 f7f6 f4f5 e7d6 b2c3 e6f5 c3a5 c7b8 f1f5 d6g3 e3g3 b8g3 h2g3 b7e4');
INSERT INTO evaluations VALUES(2782,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(2782,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2782,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2782,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(2782,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(2782,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2782,36,'cp 10','a2a4 d7d5 e4d5 c6d4 f3d4 c5d4 h2h3 f6d5 d2d3 c8e6 b1d2 d5f4 b3e6 d8h4 d2e4 f4e6 d1g4 h4g4 h3g4 b5b4 c2c3 d4c5 e4c5 e6c5 e1e5 c5d3 e5d5 f8d8 d5d8 a8d8 c3b4 d3b4 c1e3 b4c2 a1c1 c2e3 f2e3 d8b8 c1c7 g7g6 c7a7 b8b2 a4a5 b2e2 a7a6 g8g7 g4g5 e2e3 g1f2 e3c3');
INSERT INTO evaluations VALUES(6425,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(6425,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(6425,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(6425,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(6425,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(6425,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(6425,36,'cp -37','c5b6 d2d4 d7d6 d4e5 d6e5 d1d8 c6d8 f3e5 g8f6 b1d2 e8g8 f1e1 f8e8 d2f3 c8e6 c4f1 f6g4 e5g4 e6g4 f3d4 g4d7 a2a4 d8c6 d4c6 d7c6 a4a5 b6c5 e4e5 h7h6 h2h4 a8d8 c1f4 b7b6 a5a6 g8f8 h4h5 c6d5 e1d1 c7c6 d1d2 d5e6 a1d1 d8d2');
INSERT INTO evaluations VALUES(13721,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(13721,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(13721,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(13721,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13721,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13721,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(13721,36,'cp 12','b1c3 f8e7 d2d4 e8g8 d4c5 e7c5 c1g5 c8e6 d1d3 c5e7 a1d1 h7h6 g5f6 e7f6 c3d5 e6d5 d3d5 d8d5 d1d5 f6b2 d5d7 c6a5 f1d1 b2f6 f3d2 f8d8 d2b3 a7a6 b5e2 d8d7 d1d7 a8d8 d7c7 a5b3 a2b3 b7b5 c7a7 d8d2 a7a8 g8h7 e2d3 g7g6 a8a6 d2d1 d3f1');
INSERT INTO evaluations VALUES(16414,30,'cp -131',NULL);
INSERT INTO evaluations VALUES(16414,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(16414,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(16414,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(16414,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(16414,35,'cp -111',NULL);
INSERT INTO evaluations VALUES(16414,36,'cp -111','f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 f8d6 e5f3 e8g8 e1g1 f8e8 d2d3 c8g4 b1d2 g4h5 d2c4 f6d5 c4d6 d8d6 h2h3 d6f6 g2g4 h5g6 f3h2 a8d8 f2f4 f6d4 g1h1 f7f5 h2f3 d4c5 f3e5 f5g4 h3g4 g6f7 c1d2 d5e3 d2e3 c5e3 e5f7');
INSERT INTO evaluations VALUES(16837,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(16837,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(16837,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(16837,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(16837,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(16837,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(16837,36,'cp -120','c4d5 e4g5 f3g5 d8g5 e1g1 g5g6 d2d4 c6d4 f1e1 d4e6 e1e5 c7c6 d5e6 f7e6 e5e3 d7d6 e3g3 g6f7 g3f3 f7c7 c1g5 e6e5 b1d2 c8e6 c2c4 f8e7 g5e7 c7e7 d2e4 e8c8 d1c2 c8b8 a1e1 e6g4 f3g3 g4f5 c2b3 e7f7');
INSERT INTO evaluations VALUES(23781,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(23781,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(23781,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(23781,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(23781,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(23781,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(23781,36,'cp 95','d4e5 g4f3 d1f3 d6e5 f3b3 b7b5 b1d2 g8f6 d2f3 d8e7 f1e1 a8d8 f3e5 e7e5 b3g3 e5g3 h2g3 e8d7 a2a4 b5b4 c1g5 h7h6 g5f6 g7f6 a1d1 d7e6 g1f1 f6f5 e4f5 e6f5 d1d8 h8d8 g3g4 f5g6 e1e4 b4b3 c2b3 d8d2 e4c4 d2b2 c4c6 f7f6 c6c3 g6g5 f2f3 g5h4 c3c6 b2b1 f1e2 b1b2 e2e3 b2b3 e3e4 b3b4 e4f5 h4g3');
INSERT INTO evaluations VALUES(23777,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(23777,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(23777,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(23777,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(23777,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(23777,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(23777,36,'cp 80','g4f3 d1f3 d6e5 f3b3 b7b5 b1d2 g8f6 d2f3 d8e7 f1e1 a8d8 f3e5 e7e5 b3g3 e5g3 h2g3 e8d7 f2f3 d7c8 c1g5 h7h6 g5f6 g7f6 g1f2 h6h5 e1e2 h5h4 g3h4 h8h4 a1e1 c8b7 e4e5 f6e5 e2e5 h4c4 e1e2 a6a5 g2g4');
INSERT INTO evaluations VALUES(23779,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(23779,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(23779,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23779,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(23779,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(23779,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(23779,36,'cp 94','d1f3 d6e5 f3b3 b7b5 b1d2 g8f6 d2f3 d8e7 f1e1 e8g8 f3e5 e7e5 b3g3 e5g3 h2g3 f6d7 f2f3 d7c5 c1e3 c5e6 g1f2 f8d8 f3f4 f7f6 a1d1 h7h6 d1d3 c6c5 d3c3 c5c4 b2b3 c4b3 c2b3 d8d6 f4f5 e6g5 e3g5 h6g5 c3c7 a8d8 e1e2');
INSERT INTO evaluations VALUES(23778,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(23778,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(23778,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(23778,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(23778,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(23778,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(23778,36,'cp 69','d6e5 f3b3 b7b5 b1d2 g8f6 d2f3 d8e7 f3e5 e7e5 f1e1 a8d8 f2f3 e8g8 c1e3 f6d7 b3a3 a6a5 b2b3 e5c3 a1d1 a5a4 e3d4 c3c2 d1c1 c2d2 e1d1 d2e2 b3a4 d7e5 a4b5 c6b5 a3b3 e5g6 d1e1 g6f4 e1e2 f4e2 g1f1 e2c1 b3c3 c1a2 c3b2');
INSERT INTO evaluations VALUES(23776,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(23776,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(23776,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(23776,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(23776,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(23776,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(23776,36,'cp 77','f3b3 b7b5 b1d2 g8f6 d2f3 d8e7 f3e5 e7e5 f1e1 a8d8 b3g3 e5g3 h2g3 e8d7 a2a4 b5b4 f2f3 c6c5 c1g5 d7c6 g1f2 d8d6 b2b3 h8d8 e4e5 d6d5 e1e2 h7h6 g5h4 g7g5 e5f6 d8d6 a1e1 g5h4 g3h4 d6f6 g2g4 c5c4 f2g3 c4b3 c2b3 d5d3');
INSERT INTO evaluations VALUES(26613,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(26613,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(26613,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(26613,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(26613,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(26613,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(26613,36,'cp 50','b1d2 a7a6 d1b3 e7e6 f3e5 c6e5 d4e5 b7b5 a2a4 c7c6 e2e4 g8e7 b3c2 a8b8 h2h3 g4h5 a4b5 a6b5 f1d3 c6c5 e4d5 e7d5 d3b5 b8b5 c2a4 d8d7 a4h4 f8e7 h4h5 d5f4 a1a8 e7d8 h5f3 f4d3 e1e2 d3c1 h1c1 b5b2 f3d3 d7d3 e2d3 e8g8 c1a1 c5c4 d2c4 b2f2 a1a2 f2a2 a8a2 f7f6 a2a7 f6e5 c4e5 d8f6 d3e4 f6e5 e4e5 f8f2 e5e6 f2e2 e6d6 e2g2 a7a8 g8f7');
INSERT INTO evaluations VALUES(26612,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(26612,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(26612,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(26612,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(26612,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(26612,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(26612,36,'cp 35','a7a6 h2h3 g4f5 d1b3 e7e6 g2g4 f5g6 f3h4 g8e7 f1g2 h7h6 e2e4 c6a5 b3a4 b7b5 a4d1 g6h7 d1e2 c7c6 f2f4 d5e4 d2e4 e7d5 h4f3 f8d6 f3e5 h7e4 g2e4 d6e5 d4e5 d8h4 e2f2 h4f2 e1f2 e8c8 h1d1 h6h5 g4g5 a5c4 a2a4');
INSERT INTO evaluations VALUES(28881,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(28881,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(28881,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(28881,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(28881,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(28881,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(28881,36,'cp -11','h7h6 g5e3 c6c5 e1g1 c5d4 f3d4 b8c6 f2f4 c6d4 e3d4 e7f5 c2c3 a7a6 b1d2 h6h5 d2b3 f8e7 d4f2 g7g6 a2a4 a8c8 f1d1 e8f8 h2h3 f8g7 a4a5 d8c7 d1d2 c7c4 b3d4 h5h4 a1d1 c4d3 d2d3 c8c4 b2b3 c4c8 d4f5 g6f5');
INSERT INTO evaluations VALUES(28879,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(28879,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(28879,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(28879,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(28879,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(28879,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(28879,36,'cp -12','g5e3 c6c5 c2c3 e7f5 b1d2 b8d7 e1g1 f8e7 g2g4 f5e3 f2e3 e8g8 f1f2 f7f6 e5f6 e7f6 a1f1 d8e7 h2h3 a8d8 e3e4 c5d4 c3d4 e6e5 e4d5 e5d4 d2e4 d7c5 e4f6 e7f6 d3d4 f6d4 f3d4 f8f2 g1f2 d8d5 f2e3 c5a4 f1c1 a4b2 c1c8 g8f7 c8c7 f7f6 c7b7 b2c4 e3e4');
INSERT INTO evaluations VALUES(41995,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(41995,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41995,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41995,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41995,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41995,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41995,36,'cp 0','b4c3 b2c3 d5c4 e2e4 c7c5 g5f6 d8f6 f1c4 e8g8 g1f3 b8c6 d1d3 f8d8 a1d1 c5d4 c3d4 c6d4 f3d4 e6e5 c4d5 e5d4 d3d4 f6d4 d1d4 c8e6 e1e2 e6d5 e4d5 a8c8 a3a4 d8d6 h1b1 b7b6 h2h4 h7h5 g2g3 g8f8 a4a5 b6a5 b1b5 c8c2 e2e3 d6b6 b5a5 b6b3 d4d3 b3d3 e3d3 c2f2');
INSERT INTO evaluations VALUES(42001,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42001,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(42001,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(42001,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42001,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42001,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42001,36,'cp -1','b2c3 d5c4 e2e4 c7c5 g5f6 d8f6 f1c4 e8g8 g1f3 f8d8 d1d3 b8c6 a1d1 c5d4 c3d4 c6d4 f3d4 e6e5 c4d5 e5d4 d3d4 f6d4 d1d4 c8e6 e1e2 e6d5 e4d5 a8c8 d5d6 f7f6 a3a4 g8f7 h1d1 d8d7 f2f4 c8c5 e2f3 h7h5 d1e1 a7a6 e1b1 c5c6 b1d1 b7b5 a4b5');
INSERT INTO evaluations VALUES(44265,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(44265,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(44265,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(44265,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(44265,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(44265,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(44265,36,'cp 78','e2e4 d7d6 f1c4 e6d5 c3d5 f8e7 g1e2 e8g8 e1g1 f6d5 c4d5 b8a6 e2f4 a6c7 a2a4 c7d5 f4d5 c8d7 d1d3 d7c6 f1d1 f8e8 c1e3 e7f6 c2c3 f6e5 f2f4 e5f6 e3f2 c6d5 d3d5 d8b6 d1d2 b6a6 f2g3 h7h6 a1d1 a6a4 d5b7');
INSERT INTO evaluations VALUES(44800,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(44800,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(44800,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(44800,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(44800,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(44800,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(44800,36,'cp 101','g1f3 a7a6 f1e2 f8g7 e4e5 f6d7 e5d6 e7d6 c1f4 e8g8 f4d6 f8e8 e1g1 d7f6 d6g3 b8c6 d4d5 c6e7 d1b3 e7f5 g3e5 g6g5 h2h3 h7h5 b3b4 c8d7 e2d3 a6a5 b4b7 g5g4 e5c7 d8c8 b7c8 a8c8 f3e5 c8c7 e5d7 f6d7');
INSERT INTO evaluations VALUES(44799,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(44799,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(44799,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(44799,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(44799,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(44799,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(44799,36,'cp 96','a7a6 h2h3 f8g7 f1d3 e8g8 e1g1 b7b5 e4e5 f6d7 d3e4 a8a7 c1e3 d7b6 b2b3 b5b4 c3e2 d6e5 f3e5 c8b7 e4b7 a7b7 e2f4 g7e5 d4e5 d8d1 f1d1 b8d7 e5e6 f7e6 f4e6 f8f6 e6g5 b7c7 a2a3 b4a3 a1a3 b6c8 d1a1 d7b8 g5e4 f6c6 e3f4 c8d6');
INSERT INTO evaluations VALUES(52184,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(52184,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(52184,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(52184,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(52184,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(52184,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(52184,36,'cp 66','g8f6 f1c4 f8c5 f3e5 d8e7 e5d3 e7e4 d1e2 e4e2 e1e2 c5d4 c2c3 d4b6 d3f4 e8g8 h1d1 b8c6 b1a3 c6e7 c4b3 d7d5 a3c2 f8e8 e2f1 c8f5 c2e3 f5e6 a2a4 h7h6 a4a5 b6c5 f2f3 b7b6 a5b6 a7b6 a1a8 e8a8 f4d5 f6d5 e3d5 e7d5');
INSERT INTO evaluations VALUES(52660,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(52660,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(52660,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52660,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(52660,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(52660,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(52660,36,'cp 74','b1c3 g8f6 c4d5 c6d5 c1f4 b8c6 e2e3 e7e6 f1d3 f8d6 f3e5 e8g8 e1g1 d6e5 d4e5 f6d7 d1h5 f7f5 e5f6 d7f6 h5e2 e6e5 f4g3 c8e6 h2h3 d8e7 g3h4 a8d8 f2f4 e5e4 d3c2 g8h8 f1d1 a7a6 a1c1 e6g8 a2a3 c6a5 e2f2 a5c4 c2b3');
INSERT INTO evaluations VALUES(52658,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(52658,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(52658,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(52658,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(52658,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(52658,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(52658,36,'cp 55','g8f6 c4d5 c6d5 c1f4 b8c6 e2e3 e7e6 f1d3 f8d6 f3e5 d6e5 d4e5 f6d7 d1h5 e8g8 e1g1 f7f5 e5f6 d7f6 h5e2 e6e5 f4g3 c8e6 a1d1 d8e7 h2h3 e6f7 f2f3 f6h5 g3h2 a8d8 d3b1 h5f6 h2g3 f7h5 e2b5 a7a6 b5b3 c6a5 b3a4 a5c6 d1c1 h5f7 a4c2 g8h8 f1d1 f7g8');
INSERT INTO evaluations VALUES(53935,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(53935,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(53935,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(53935,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(53935,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(53935,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(53935,36,'cp 156','b1d2 g8f6 a2a4 f8e7 h2h3 e8g8 g1e2 b8a6 g2g4 a6c7 e2g3 f6e8 f1d3 e7g5 d2c4 g5c1 d1c1 c8d7 c1d2 d7a4 a1a4 b7b5 a4a2 b5c4 d3c4 g7g6 h3h4 d8f6 g4g5 f6f3 d2e3 f3e3 f2e3 f7f6 e1d2 a7a5 d2d3 f6g5 h4g5');
INSERT INTO evaluations VALUES(1876,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(1876,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(1876,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(1876,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(1876,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(1876,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(1876,36,'cp 132','e1g1 d7d6 c2c3 c8g4 d1b3 g8e7 f3d4 g4d7 d4c6 e7c6 c4d5 c6d8 b3d1 f8e7 f2f4 c7c6 d5b3 d6d5 e4d5 c6d5 b3d5 e8g8 c1e3 d7c6 d1h5 c6d5 h5d5 d8c6 b1d2 a8d8 d5b5 d8d7 d2f3 f6e6 f1e1 e6d5 b5d5 d7d5 f3e5 c6e5 f4e5 d5e5 e3a7');
INSERT INTO evaluations VALUES(1875,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(1875,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(1875,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(1875,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(1875,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(1875,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(1875,36,'cp 112','d7d6 c2c3 d4d3 c4d3 c8g4 b1d2 g8e7 h2h3 g4d7 d2c4 h7h6 b2b4 g7g5 d3e2 e7g6 a2a4 f8g7 b4b5 c6e7 c1e3 g6f4 e3d4 f6g6 f1e1 e8g8 d4g7 g6g7 c4e3 f7f5 e4f5 e7f5 e3f5 d7f5 f3d4 g8h8 d4f5');
INSERT INTO evaluations VALUES(5937,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(5937,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(5937,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(5937,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(5937,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(5937,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(5937,36,'cp 17','e5d4 f3d4 g8e7 c1g5 e8g8 b5c6 c5d4 g5e7 d8e7 d1d4 b7c6 b1c3 a7a5 b2b3 f8e8 f2f3 c8d7 d4c4 e7e6 c4e6 e8e6 a1e1 e6e8 e1e2 f7f5 f1e1 f5e4 e2e4 e8e4 e1e4 a8e8 g1f2 e8e4 c3e4 c6c5 f2e3 g8f7 c2c4 d7c6 e3d3 f7e7 e4c3 c6b7 c3d5 e7d7');
INSERT INTO evaluations VALUES(5958,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(5958,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(5958,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(5958,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(5958,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5958,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(5958,36,'cp 22','f3d4 g8e7 c1g5 e8g8 b5c6 c5d4 g5e7 d8e7 d1d4 b7c6 b1c3 a7a5 b2b3 f8e8 f1e1 c8d7 h2h3 a5a4 a1d1 c6c5 d4d5 e7e5 d5e5 e8e5 f2f4 e5e8 e4e5 a4b3 a2b3 d7c6 c3d5 c6d5 d1d5 a8a2 e1e2 g8f8 e2d2 d6e5 f4e5 c5c4 b3c4 a2b2 c4c5 b2b1 g1h2');
INSERT INTO evaluations VALUES(5954,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(5954,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(5954,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(5954,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(5954,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(5954,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(5954,36,'cp 24','g8e7 c1g5 e8g8 b5c6 c5d4 d1d4 e7c6 d4d2 d8d7 f2f3 f7f5 b1c3 d7f7 b2b3 c8e6 e4f5 e6f5 a1c1 f5e6 g5h4 a8e8 f1f2 c6e7 h4e7 e8e7 c1e1 f8e8 a2a4 h7h6 h2h3 g8h8 f2e2 f7f6 a4a5 a7a6 c3e4 f6g6 g1h2');
INSERT INTO evaluations VALUES(16551,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(16551,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16551,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(16551,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(16551,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(16551,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(16551,36,'cp -25','d7d5 f1e2 f8d6 e1g1 e8g8 e4d5 f6d5 f1e1 h7h6 h2h3 c8f5 e2f1 f8e8 c2c3 a7a5 d2c4 d8f6 d1a4 e8d8 c4d6 d8d6 a4h4 f6h4 f3h4 f5h7 h4f3 f7f6 c1d2 g8f8 d3d4 e5d4 f3d4 c6d4 c3d4 h7g6 a1d1');
INSERT INTO evaluations VALUES(15157,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(15157,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(15157,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(15157,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(15157,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(15157,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(15157,36,'cp -20','f1e2 f8d6 e1g1 e8g8 h2h3 h7h6 f1e1 f8e8 e4d5 f6d5 d2c4 a7a5 e2f1 c8f5 a2a3 a5a4 c4d6 d8d6 c2c3 d6f6 c1d2 e8d8 d1e2 d5b6 d3d4 e5d4 c3d4 f5e6 d2c3 e6c4 e2c2 c4b3 c2c1 b6d5 c1d2 d5c3 b2c3 b3d5 f3e5');
INSERT INTO evaluations VALUES(21426,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(21426,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(21426,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(21426,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(21426,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(21426,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(21426,36,'cp 47','g1f3 g8f6 e1g1 f8e7 b1d2 a7a5 c2c3 e8g8 d1c2 c6b8 f1d1 c7c5 e2e4 b8c6 e4d5 e6d5 d4c5 e7c5 d2f1 f8e8 c1g5 c8e6 g5e3 c5b6 e3b6 d8b6 f1e3 a5a4 d1d2 a4a3 b2b4 a8d8 a1d1 h7h6 f3d4 c6d4 d2d4');
INSERT INTO evaluations VALUES(26202,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26202,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(26202,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(26202,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26202,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26202,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(26202,36,'cp 0','e7e6 e1g1 a7a6 d2d3 d8c7 h2h3 g8e7 c3e2 d7e5 f3e5 d6e5 f2f4 e5f4 e2f4 e8c8 d1e2 e7c6 c1e3 c7e5 f4h5 f8d6 e3f4 e5d4 f4e3');
INSERT INTO evaluations VALUES(30036,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(30036,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(30036,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(30036,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(30036,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(30036,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(30036,36,'cp 78','f8b4 f1d3 c7c5 d4c5 e8g8 g1f3 b8d7 e1g1 d7c5 f3g5 c5d3 g5e6 f7e6 c2d3 a8c8 c3e2 b4a5 c1f4 d8b6 a1b1 b6b5 d3d4 a5b6 f4e5 f6d7 e5d6 f8e8 d6g3 c8d8 h2h3 e6e5 e2c3 b5c4 b1c1 c4d4');
INSERT INTO evaluations VALUES(28476,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(28476,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(28476,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(28476,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(28476,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(28476,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(28476,36,'cp 100','g1f3 e8g8 f1d3 c7c5 d4c5 b8d7 c5c6 b7c6 e1g1 d7c5 f3g5 e6d7 f1e1 h7h6 g5f3 f6d5 c1d2 c5d3 c2d3 d5c3 d2c3 c6c5 f3e5 d7e6 e1e3 d8b6 h2h3 a8d8 d1d2 b6d6 a2a3 b4c3 d2c3 f8e8 e5f3 e6d5 e3e8 d8e8 a1e1 e8e1 f3e1');
INSERT INTO evaluations VALUES(32143,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(32143,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(32143,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(32143,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(32143,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(32143,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(32143,36,'cp 45','h2h3 g4h5 b1c3 g8f6 g3g4 h5g6 d2d3 h7h6 f1g2 f8e7 f3h4 g6h7 h4f5 h7f5 g4f5 c7c5 f2f4 b8c6 a2a4 d8d7 a4a5 e8g8 e1g1 c6d4 c3e2 g8h8 c2c3 d4e2 d1e2 b7b5 g1h1 f8g8 c1e3 e5f4 e3f4');
INSERT INTO evaluations VALUES(33375,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(33375,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(33375,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(33375,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(33375,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(33375,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(33375,36,'cp -24','h7h6 g5e3 f6d5 e3d2 e8g8 b1c3 d5c3 d2c3 f8e8 g2g3 e7f8 e1d1 c7c5 f1g2 b8c6 a2a4 c8d7 h1e1 d7e6 d1e2 a7a5 e2f1 a8d8 e1e2 b7b6 a1e1 c6b4 f1g1 e6d7 b2b3 e8e2 e1e2 d7e6 g1f1 d6d5 f3e5 g7g5 c3b4 c5b4 e5c6');
INSERT INTO evaluations VALUES(33371,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(33371,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(33371,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(33371,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(33371,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(33371,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(33371,36,'cp -9','g5e3 f6d5 e3d2 e8g8 g2g3 f8e8 b1c3 d5c3 d2c3 b8c6 f1g2 d6d5 e1g1 a7a5 f1e1 c8e6 a2a3 b7b5 f3e5 c6e5 c3e5 c7c6 h2h4 a5a4 c2c4 e8c8 c4d5 c6d5 e5c3 e7d6 g2f3 a8b8 f3d1 d6c5 c3e5 b8b6 d3d4 c5d6 e5d6 b6d6');
INSERT INTO evaluations VALUES(41774,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41774,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(41774,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(41774,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41774,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41774,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41774,36,'cp 0','b7b6 b2b3 c8b7 f1d3 e8g8 e1g1 d5c4 b3c4 c7c5 f1e1 b8d7 c1b2 c5d4 e3d4 a8c8 h2h3 d6b8 a2a4 d8c7 b2a3 b7f3 d1f3 c7h2 g1f1 f8e8 a1d1 e6e5 d4d5 e5e4 c3e4 h2h1 f1e2 h1h2');
INSERT INTO evaluations VALUES(44087,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(44087,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(44087,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(44087,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(44087,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(44087,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(44087,36,'cp -80','d7d5 e4d5 c6d5 c4a2 h7h6 g1e2 f8c5 e1g1 e8g8 b2b4 c5b6 h2h3 a7a5 c1b2 f8e8 b1d2 b6c7 c2c4 d5d4 c4c5 c8e6 a2e6 e8e6 f2f4 f6d5 d2c4 d5f4 e2f4 e5f4 b4b5 b8d7 b5b6 c7b8 d1h5 b8e5 a1c1 d8f6 c4e5 e6e5');
INSERT INTO evaluations VALUES(45811,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(45811,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(45811,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(45811,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(45811,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(45811,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(45811,36,'cp -67','e4d5 c6d5 c4a2 h7h6 g1e2 f8c5 e1g1 e8g8 b2b4 c5b6 h2h3 b6c7 b1d2 a7a5 c1b2 a5b4 a3b4 b8c6 b4b5 c6e7 c2c4 c8f5 d2f3 e7g6 d3d4 e5e4 f3e5 f6h5 e2g3 h5g3 f2g3 d5c4 a2c4 a8a1 d1a1 c7e5 d4e5 f5e6 c4e6 f7e6 f1f8 g6f8 b2a3 d8b6 g1h2 b6b5');
INSERT INTO evaluations VALUES(45744,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(45744,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(45744,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(45744,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(45744,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(45744,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(45744,36,'cp -47','c6d5 c4a2 h7h6 h2h3 b8c6 g1e2 f8c5 b2b4 c5d6 c2c4 d6c7 e1g1 e8g8 b1d2 c8e6 c1b2 d8d7 f2f4 d5c4 d2c4 e5f4 e2f4 c6d4 f4e6 f7e6 a1c1 b7b5 c4d2 c7b6 g1h1 f6d5 a2d5 e6d5 d1h5 f8f5 h5g6 a8e8 b2d4 f5f1 c1f1');
INSERT INTO evaluations VALUES(46492,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(46492,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(46492,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(46492,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(46492,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(46492,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(46492,36,'cp -49','c4a2 b8c6 g1e2 f8c5 e1g1 a7a6 c2c4 d5c4 d3c4 c8g4 d1d8 a8d8 b1c3 g4e2 c3e2 f6e4 b2b4 c6d4 e2d4 c5d4 a1b1 e4c3 b1b2 c3a2 b2a2 e8e7 g2g3 d8c8 c4c5 c8c6 c1b2 h8d8 b2d4 d8d4');
INSERT INTO evaluations VALUES(46507,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(46507,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(46507,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(46507,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(46507,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(46507,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(46507,36,'cp -70','b8c6 g1e2 f8c5 e1g1 a7a6 c1g5 c8e6 b1c3 c5a7 g1h1 d8d7 f2f4 f6g4 f4e5 g4f2 f1f2 a7f2 d3d4 h7h6 g5f4 g7g5 f4g3 f2g3 e2g3 e8c8 c3a4 c8b8 c2c3 h6h5 a4c5 d7e7 d1c2 h5h4 g3f5 e7c7 f5d6 d8d6 e5d6 c7d6 c5e6');
INSERT INTO evaluations VALUES(13806,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(13806,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(13806,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(13806,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(13806,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(13806,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(13806,36,'cp -44','g1e2 f8c5 e1g1 a7a6 c2c4 c8g4 c4d5 f6d5 h2h3 g4e6 b1c3 e8g8 a2d5 e6d5 c1e3 c5e3 f2e3 d5e6 a1c1 d8g5 f1f3 f7f5 d1e1 a8d8 c1d1 e6f7 d3d4 h7h6 g1h1 g5f6 f3f2 e5d4 e2d4 c6d4 d1d4 d8d4 e3d4 f6d4 f2f5 f8e8 e1f2 d4f2 f5f2');
INSERT INTO evaluations VALUES(44094,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44094,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44094,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44094,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44094,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(44094,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(44094,36,'cp 17','d2d4 e5d4 e3d4 d7d5 f1d3 d5c4 d3c4 f8b4 a2a3 b4c3 b2c3 e8g8 g1e2 b8d7 c4d3 c6c5 e1g1 c5d4 e2d4 d7e5 c1g5 e5d3 d1d3 d8d6 f1e1 c8d7 d3f3 d6d5 f3f4 f6h5 f4e5 d5e5 e1e5 f8e8 e5d5 d7e6 d4e6 e8e6 d5d7');
INSERT INTO evaluations VALUES(44104,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44104,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44104,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(44104,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(44104,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(44104,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44104,36,'cp 6','e5d4 e3d4 d7d5 f1d3 d5c4 d3c4 f8b4 a2a3 b4c3 b2c3 e8g8 g1e2 b8d7 c4d3 c6c5 e1g1 c5d4 c3d4 f8e8 f1e1 d7f8 d1b3 b7b6 c1d2 c8b7 d3b5 e8e6 f2f3 f8g6 a1c1 a7a6 b5d3 b6b5 a3a4 b7d5 b3b2 h7h6 a4b5 a6b5');
INSERT INTO evaluations VALUES(44573,30,'cp -145',NULL);
INSERT INTO evaluations VALUES(44573,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(44573,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(44573,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(44573,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(44573,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(44573,36,'cp -128','b8c6 b1c3 e7e6 d2d3 d7d5 f3d1 d5d4 c3b1 e6e5 a2a4 f8d6 b2b3 c8g4 g1f3 h7h6 b1d2 d6c7 g2g3 d8e7 d2c4 e8c8 f1g2 g7g5 h2h3 g4e6 h3h4 g5g4 f3g1 f6d7 g1e2 f7f6 e1g1 c8b8 c1d2 h6h5 c2c3 b8a8 f1e1 e7g7 c3d4');
INSERT INTO evaluations VALUES(12843,30,'cp -127',NULL);
INSERT INTO evaluations VALUES(12843,31,'cp -138',NULL);
INSERT INTO evaluations VALUES(12843,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(12843,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(12843,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(12843,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(12843,36,'cp -122','b1c3 e7e6 d2d3 d7d5 f3d1 d5d4 c3b1 e6e5 b2b3 f8d6 a2a4 c8g4 g1f3 a7a6 b1d2 d6c7 g2g3 h7h6 f1g2 f6d7 h2h3 g4e6 f3h4 g7g6 e1g1 d8e7 d2c4 e8c8 c1d2 c8b8 d1e2 b8a7 a1b1 d8g8 c2c3 g8b8 f1d1 a7a8 b3b4');
INSERT INTO evaluations VALUES(57223,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(57223,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(57223,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(57223,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(57223,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(57223,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(57223,36,'cp -78','d7d5 d2d4 b8c6 d4e5 c6e5 g1f3 e5f3 g2f3 g8f6 e3e4 c8e6 c1f4 c7c6 b1d2 d8a5 c2c3 e8c8 d2b3 a5a4 b3d4 d5e4 d4e6 f7e6 f3e4 a4e4 e2e4 f6e4 f1g2 e4c5 e1g1 f8d6 f4g5 d8d7 a1d1 c8c7 g2h3 h8f8 g5e3 a7a6 d1d2 a6a5 f1d1 f8d8 h3g4 b7b5 e3c5 d6c5 d2d7 d8d7');
INSERT INTO evaluations VALUES(56478,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(56478,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(56478,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(56478,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(56478,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(56478,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(56478,36,'cp -74','d2d4 b8c6 d4e5 c6e5 g1f3 e5f3 g2f3 g8f6 b1c3 c7c6 b2b3 g7g6 e2d2 f8g7 c1a3 f6d7 f3f4 d8f6 a3b2 e8g8 e1c1 f6h4 f2f3 d7c5 c3e2 g7b2 c1b2 f8e8 e2d4 c5e6 d4e6 c8e6 f1d3 c6c5 e3e4 c5c4 e4d5');
INSERT INTO evaluations VALUES(4331,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(4331,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4331,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4331,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4331,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(4331,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4331,36,'cp 0','d7d5 e4d5 f6d5 e1g1 d5b6 c4b5 f7f6 b1d2 d8d5 d2e4 c5e7 c3c4 d5f7 c4c5 b6d5 f1e1 c8d7 d3d4 e5d4 f3d4 a8d8 b5c4 c6a5 c4e2 a5c6');
INSERT INTO evaluations VALUES(3721,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(3721,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(3721,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3721,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3721,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(3721,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3721,36,'cp 0','e4d5 f6d5 e1g1 d5b6 c4b5 c5d6 f1e1 f8e8 c1e3 a7a6 b5c6 b7c6 b1d2 b6d5 d2e4 c8f5 d1a4 d8b8 a1b1 b8b5 a4d1 c6c5 f3h4 f5e6 c3c4 d5e3 e1e3 b5d7 e3g3 d6e7 d1h5 g7g6 h4f3 f7f6 f3e5 f6e5 g3g6 h7g6 h5g6 g8f8 g6h6 f8g8');
INSERT INTO evaluations VALUES(3713,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3713,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(3713,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(3713,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(3713,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3713,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3713,36,'cp 0','f6d5 e1g1 d5b6 c4b5 f7f6 b1d2 a7a6 b5c6 b7c6 d2e4 c5e7 c1e3 c8f5 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(3742,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3742,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3742,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3742,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(3742,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3742,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3742,36,'cp 0','e1g1 d5b6 c4b5 f7f6 b1d2 a7a6 b5c6 b7c6 d2e4 c5e7 f1e1 c6c5 b2b3 f8e8 c1a3 c8b7 f3h4 b7e4 d3e4 d8d1 e1d1 g7g6 g2g4 a6a5 h4g2 a5a4 g2e3 e8d8 d1d8 a8d8 g1f1 a4b3 a2b3 d8d3 a3b2 d3d2 b2c1 d2d3');
INSERT INTO evaluations VALUES(3741,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3741,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(3741,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3741,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3741,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3741,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3741,36,'cp -9','d5b6 c4b5 f7f6 b1d2 a7a6 b5c6 b7c6 d2e4 c5e7 d3d4 e5d4 f3d4 d8d5 e4g3 f8f7 d4f5 c8f5 d1d5 c6d5 g3f5 e7f8 b2b3 a6a5 f5d4 f7d7 c1e3 g8f7 a1b1 c7c5 d4b5 a8c8 a2a4 g7g6 f1d1 g6g5 f2f3 h7h5 e3f2 h5h4');
INSERT INTO evaluations VALUES(5035,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5035,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5035,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5035,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5035,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5035,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5035,36,'cp 0','d7d6 b1c3 g8f6 e4e5 d6e5 d1d8 e8d8 f1e1 c8f5 f2f3 h7h5 c1g5 e7e6 e1e5 d8c7 e5e2 f6d5 a1d1 g7c3 b2c3 d5c3 g5f4 c7b6 f4e3 b6c7');
INSERT INTO evaluations VALUES(5064,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5064,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5064,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5064,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5064,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5064,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(5064,36,'cp 4','b1c3 g8f6 c1f4 e8g8 d1d2 f6e4 c3e4 d6d5 a1d1 d8b6 f4g5 b6b2 f1e1 d5e4 e1e4 g7f6 g5f6 b2f6 d1e1 f8e8 e4e7 c8e6 e7e8 a8e8 c4b3 f6d8 d2c3 e6b3 a2b3 e8e1 c3e1 h7h6 e1e2 a7a5 h2h3 d8d5 g2g3 c6c5 h3h4 g8g7 e2e1');
INSERT INTO evaluations VALUES(10558,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(10558,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(10558,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(10558,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(10558,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(10558,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(10558,36,'cp 121','b1c3 h6g4 c4f7 e8f7 f3g5 f7g8 d1g4 c6d4 g4d1 d7d6 h2h3 h7h6 g5f3 g8h7 f3d4 c5d4 c3e2 d4b6 e1g1 c8d7 e2g3 h8f8 d1e2 d8h4 c1e3 f8f7 g1h2 a8f8 e3b6 a7b6 a1d1 g7g5 f2f3 h6h5 d3d4 g5g4 f3g4 h5g4 f1f7 f8f7 d1f1 f7f1 e2f1 g4h3');
INSERT INTO evaluations VALUES(10555,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(10555,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(10555,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(10555,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(10555,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(10555,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(10555,36,'cp 120','h6g4 c4f7 e8f7 f3g5 f7g8 d1g4 c6d4 g4d1 d7d6 h2h3 h7h6 g5f3 d4f3 d1f3 g8h7 e1g1 h8f8 f3h5 c7c6 c1e3 c5e3 f2e3 f8f1 a1f1 c8e6 g1h2 d8b6 c3d1 b6d8 c2c4 a8b8 d1c3 a7a6 h5e2 b7b5 b2b3 b5c4 d3c4 b8b7 e2d3 b7d7 f1d1 a6a5 d1d2 d8e7 c3e2 d7d8 e2g3 e7f7 g3f5');
INSERT INTO evaluations VALUES(14447,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(14447,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(14447,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(14447,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(14447,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(14447,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(14447,36,'cp 189','c3e4 d5e4 c2c3 c8f5 g1e2 e7e6 e2g3 f5g6 h2h4 h7h5 f1e2 a7a6 c1e3 d8d5 g3h5 g6h5 e2h5 c6e7 h5e2 e7f5 h4h5 f8e7 d1b3 f5e3 f2e3 d5b3 a2b3 e7g5 b3b4 f7f6 e5f6 g7f6 e1d2 f6f5 a1g1 e8e7 g2g4 a8f8 g4f5 f8f5 h1h3 h8f8 d2c2 e7d6 h3g3');
INSERT INTO evaluations VALUES(14446,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(14446,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(14446,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(14446,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(14446,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(14446,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(14446,36,'cp 172','d5e4 c2c3 g7g6 g1e2 f7f6 d1b3 f6e5 d4d5 c6a5 b3a4 c7c6 b2b4 e7e6 a4a5 d8a5 b4a5 e6d5 c1e3 b7b5 a5b6 a7b6 e3b6 f8a3 b6c7 a3b2 c7e5 e8g8 a1d1 a8a2 e5d4 f8e8 e2c1 a2a1 e1d2 c8f5 c1e2 a1a2');
INSERT INTO evaluations VALUES(14449,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(14449,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(14449,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(14449,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(14449,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(14449,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(14449,36,'cp 194','c2c3 g7g6 g1e2 f7f6 e2g3 d8d5 b2b3 f6e5 f1c4 d5d7 d4d5 c6d8 d1e2 d7g4 e2c2 g4h4 g3e4 c8f5 c4d3 f8h6 c1h6 h4h6 e1g1 e8g8 a1d1 h6f4 c2e2 f5e4 d3e4 d8f7 g2g3 f4f6 e4g2 a8d8 f1e1 b7b6 g2h3 c7c5 h3e6 g8g7 e6f7 f8f7 e2e5 f6e5');
INSERT INTO evaluations VALUES(14810,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(14810,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(14810,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(14810,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(14810,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(14810,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(14810,36,'cp 18','c2c4 f8d6 f4d6 c7d6 b1c3 c6e7 f1d3 d5c4 d3c4 c8d7 g1e2 e8g8 e1g1 d6d5 c4d3 d8b6 a1c1 b6b2 c1b1 b2a3 b1b7 e7c8 d1b3 a3b3 b7b3 c8d6 f1b1 f8c8 a2a4 a7a5 g2g4 h7h6 f2f3 f6e8 b3b6 d6c4');
INSERT INTO evaluations VALUES(19402,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(19402,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(19402,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(19402,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(19402,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(19402,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(19402,36,'cp -27','g8f6 e2e4 e7e6 c1f4 a7a6 f1e2 d7d6 g1f3 f8e7 e1g1 e8g8 a2a3 f6g4 a1b1 e7f6 d1d2 g4e5 f4e3 e5f3 e2f3 c6d4 f3d1 c8d7 b2b4 b7b5 c3e2 d7c6 e2g3 a8c8 b1c1 g7g6 e3h6 f8e8 c4b5 c6b5 a3a4 b5d7 h6e3 d8b6 b4c5 c8c5 c1c5 d6c5 g3e2 f6e7 e2d4');
INSERT INTO evaluations VALUES(12809,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(12809,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(12809,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(12809,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(12809,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(12809,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(12809,36,'cp -13','g2g3 e7e6 f1g2 d7d5 c4d5 e6d5 c1g5 f8e7 g5f6 e7f6 d1c1 c6e7 e2e3 c8f5 e3e4 d5e4 d3e4 f5e6 g1f3 e7c6 e1g1 e8g8 f1d1 d8b6 e4e5 c6e5 f3e5 f6e5 c3a4 b6a5 c1c5 a5c5 a4c5 e6g4 g2b7 a8d8 d1d5 d8d5 b7d5 e5b2');
INSERT INTO evaluations VALUES(21040,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(21040,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21040,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(21040,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(21040,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(21040,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(21040,36,'cp 51','g1f3 g8f6 e2e3 c8f5 c2c4 e7e6 b2b4 f8e7 b1d2 f6e4 d2e4 f5e4 f3d2 e4g6 f1e2 c6b8 c4d5 e6d5 d1b3 e8g8 e1g1 a7a6 a3a4 c7c6 c1a3 b8d7 b4b5 a6b5 a3e7 d8e7 a4b5 d7f6 b5c6 b7c6 f1c1 f6e4 d2e4 g6e4 b3b2 a8a1 c1a1');
INSERT INTO evaluations VALUES(24575,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24575,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(24575,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24575,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24575,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24575,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(24575,36,'cp 26','e7e6 d2d4 c5d4 f3d4 g7g6 c2c4 a7a6 b2b3 f8g7 c1b2 a8c8 d1d2 e8g8 b1c3 d8a5 a1d1 d7e5 e1e2 f8d8 c3d5 a5d2 d5f6 g7f6 e2d2 f6g5 d2e2 e5d7 g2g3 d7c5 b2a3 g5e7 f2f3 c8c7 f3f4 d8d7 d4c2 c7c8 g1g2 b7b5 c4b5 a6b5 g2f3 d7d8 c2d4');
INSERT INTO evaluations VALUES(24518,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(24518,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(24518,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24518,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(24518,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(24518,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(24518,36,'cp 25','d2d4 c5d4 f3d4 g7g6 c2c4 a7a6 b2b3 f8g7 b1c3 e8g8 c1b2 d8a5 h2h3 a8c8 a1c1 f8d8 d1f3 a5h5 f3h5 f6h5 c1d1 d7c5 g2g3 h5f6 g1g2 c8b8 e1e2 d8c8 f2f3 c5d7 e2c2 c8c7 c2d2 g7h6 f3f4 b8c8');
INSERT INTO evaluations VALUES(36610,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36610,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(36610,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(36610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36610,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(36610,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(36610,36,'cp -8','g1f3 f6h5 f4g5 d8e8 e3e4 e7e5 e1g1 b8c6 d4d5 c6d8 c3c4 f7f5 e4f5 g6f5 f3h4 e5e4 d3e2 h5f6 b1c3 d8f7 g5e3 f7e5 c3b5 f6g4 e3f4 e8e7 g2g3 e5g6 h4g6 h7g6 e2g4 f5g4 d1d2 a7a6 b5d4 c8d7 a1c1 a8e8 c1e1 e7f7 b2b3 b7b5 d4c2 f7f5 c2e3 f5h5');
INSERT INTO evaluations VALUES(38375,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(38375,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(38375,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(38375,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(38375,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(38375,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(38375,36,'cp 37','g1f3 g8f6 f1e2 e8g8 b1c3 b8a6 e1g1 c8f5 f3g5 e7e5 d4e5 d6e5 e3e4 f5c8 c1e3 d8d1 f1d1 f6g4 e3d2 g4f6 c3d5 h7h6 d5f6 g7f6 g5f3 a6c5 d2h6 f8e8 e2d3 c8g4 h6e3 g4f3 g2f3 c5e6 c4c5 a8d8 d3c4 e6d4 e3d4 e5d4 a2a4');
INSERT INTO evaluations VALUES(38826,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38826,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38826,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38826,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38826,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(38826,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(38826,36,'cp 62','b1d2 g8f6 e2e3 e8g8 d1c2 d8e7 a1c1 c8d7 c4d5 e6d5 f1d3 c7c6 e1g1 g7g6 f1e1 d7e6 f3g5 a7a5 f2f4 b8d7 d2f3 e6g4 f3e5 h7h6 g5f3 g8h8 h2h3 g4f3 e5f3 f6e4 d3e4 d5e4');
INSERT INTO evaluations VALUES(38825,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(38825,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(38825,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(38825,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(38825,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(38825,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(38825,36,'cp 47','g8f6 d1c2 e8g8 e2e3 d5c4 d2c4 b8d7 f1e2 c7c5 d4c5 d7c5 b2b4 c5d7 e1g1 d8e7 a2a3 b7b6 f1d1 c8b7 c4d6 b7d5 c2c7 d5b3 d1d3 f6d5 c7c6 a8d8 d3b3 d7b8 c6c4 d8d6 f3d4 d6d7 d4b5 a7a6 b5d4 f8d8 a1c1');
INSERT INTO evaluations VALUES(42082,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(42082,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(42082,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(42082,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(42082,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(42082,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(42082,36,'cp 47','e2e3 e8g8 a1c1 b7b6 c4d5 e6d5 d1c2 c7c6 f1d3 c8b7 e1g1 f8e8 b2b4 b8d7 a2a3 a7a5 f1e1 g7g6 c2b2 d8c7 f3g5 h7h6 g5f3 c7d6 h2h3 b6b5 c1a1 a5b4 a3b4 d6e7 b2b3 d7b6 a1a8 e8a8 e3e4 d5e4 d2e4 f6e4');
INSERT INTO evaluations VALUES(39316,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(39316,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(39316,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(39316,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(39316,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(39316,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(39316,36,'cp 137','d8e7 a2a3 e7e5 a3b4 e5e4 d1e2 e4e6 b4b5 g8f6 b1c3 e8g8 d2d4 f8e8 e2e6 d7e6 c1g5 c7c6 f2f3 a7a5 b5a6 b8a6 g5f6 g7f6 f1a6 a8a6 a1a6 b7a6 e1f2 e6e5 c3e4 g8g7 e4d6 e8d8 d6c8 d8c8 d4e5 f6e5 h1a1 c8a8 a1a5');
INSERT INTO evaluations VALUES(44471,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44471,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44471,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44471,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(44471,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(44471,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44471,36,'cp 13','d7d5 d2d4 g7g6 c4d5 c6d5 g1f3 f8g7 b1c3 f6e4 e1g1 b8c6 f3e5 e4c3 b2c3 e8g8 e5c6 b7c6 d1a4 d8b6 e2e4 b6a6 a4a3 e7e6 f1e1 a6a3 c1a3 f8d8 g2f1 d5e4 e1e4 c8d7 h2h4 g7f8 a3f8 g8f8 e4e5 d8b8 e5c5 f8e7 f1g2 b8b2 g2c6 a8c8 c6d7 c8c5 d4c5');
INSERT INTO evaluations VALUES(44658,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44658,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(44658,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44658,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44658,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44658,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44658,36,'cp 23','d5c7 d2d4 c5d4 d1d4 b8c6 d4d2 g7g6 b2b4 a7a6 c1b2 f8g7 f1e2 e8g8 b4b5 a6b5 c4b5 c6a5 e1g1 d7d5 e5d6 d8d6 d2d6 e7d6 b2g7 g8g7 b1c3 c8e6 f3d4 e6c4 e2c4 a5c4 a2a4 d6d5 f1c1 f8d8 h2h4 h7h5 c1e1 d8d7 g2g3 b7b6 g1g2 c4b2 d4b3 c7e6 a4a5 b2d3 e1d1');
INSERT INTO evaluations VALUES(46530,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(46530,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(46530,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46530,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46530,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46530,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46530,36,'cp 13','d2d4 c5d4 d1d4 b8c6 d4d2 g7g6 b2b4 d7d6 c1b2 c6e5 f3e5 d6e5 d2d8 e8d8 b2e5 f7f6 e5c3 f8h6 b1d2 d8e8 g2g3 e8f7 f2f4 h8d8 f1e2 b7b5 e1f2 b5c4 d2c4 c7b5 c3b2 c8e6 h1d1 a8c8 d1d8 c8d8 a1c1 g6g5 f4g5 h6g5 c1e1 b5d4 e2h5 f7f8 b2d4 d8d4 e1e6');
INSERT INTO evaluations VALUES(44754,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(44754,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(44754,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(44754,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44754,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(44754,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(44754,36,'cp 55','d2d4 c5d4 f3d4 b8c6 f1e2 c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 a7a6 c1e3 c8d7 a2a4 a6a5 b2b3 d7c6 a1d1 f6d7 d3c2 b7b6 h2h4 d7c5 h4h5 g7c3 c2c3 c5e4 c3e1 d8d7 h5g6 h7g6 f2f3 e4c5 e1h4 f7f6 g1f2 d7f5 f1h1 g8f7 g2g4 f5e5 h4h7 f7e8 h7g6 e8d7');
INSERT INTO evaluations VALUES(44759,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(44759,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(44759,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(44759,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(44759,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(44759,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(44759,36,'cp 50','c5d4 f3d4 b8c6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 a7a6 e1g1 c8d7 a2a4 a6a5 b2b3 d7c6 a1d1 f6d7 d3c2 d8c7 f1e1 d7c5 e2f3 c7b6 c3b5 a8d8 c2e2 d8d7 f3g4 c6b5 c4b5 e7e6 g2g3 b6c7 e2c2 f8c8 g4h3 c7b8 h3f1 b7b6');
INSERT INTO evaluations VALUES(44788,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(44788,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(44788,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(44788,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(44788,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(44788,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(44788,36,'cp 48','f3d4 b8c6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d2 a7a5 e3d4 c8d7 d2e3 a5a4 e1g1 d8a5 f1d1 f8e8 a1b1 d7c6 b2b4 a4b3 a2b3 a5b4 e4e5 f6d7 e5e6 g7d4 e6f7 g8f7 d1d4 d7f6 e2f3 b4c5 f3d5 c6d5 c3d5 e7e5 d4d1 c5e3 f2e3 f6e4');
INSERT INTO evaluations VALUES(44786,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(44786,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(44786,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(44786,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(44786,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(44786,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44786,36,'cp 61','b8c6 f1e2 c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 a7a6 c1e3 c8d7 a2a4 a6a5 b2b3 d7c6 d3c2 f6d7 a1d1 d8c7 f1e1 d7c5 e2f3 c7b6 c3b5 a8d8 c2e2 c6b5 c4b5 b6c7 e2c4 d8c8 d1c1 c7d7 f3e2 e7e6 e1d1 f8d8 g2g3');
INSERT INTO evaluations VALUES(45876,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(45876,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(45876,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(45876,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(45876,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(45876,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(45876,36,'cp 48','d2d4 b8d7 a2a4 e5d4 d1d4 f8e7 f1c4 d7c5 e1g1 e8g8 f1d1 c8e6 c1f4 a7a5 f4g3 f8e8 f3g5 d6d5 e4d5 e6d5 c3d5 c6d5 c4b5 e8f8 d4e3 h7h6 g5f3 c5e4 g3e5 e7c5 e5d4 a8c8 c2c3 c5d4 f3d4 e4d6 d1e1 d6b5 d4b5 f8e8 e3a7 e8e1 a1e1');
INSERT INTO evaluations VALUES(45874,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(45874,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(45874,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(45874,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45874,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(45874,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(45874,36,'cp 33','b8d7 a2a4 f8e7 f1c4 e5d4 d1d4 e8g8 e1g1 d7c5 f1d1 c8e6 a4a5 e6c4 d4c4 b7b5 a5b6 a7b6 a1a8 d8a8 e4e5 d6e5 f3e5 b6b5 c4f4 c5e6 f4f3 f8c8 c1e3 b5b4 c3e2 c6c5 f3a8 c8a8 e2f4 a8d8 d1d8 e6d8 e5c4 d8c6 g1f1 g8f8 b2b3 g7g6 f2f3 f8e8 f1f2');
INSERT INTO evaluations VALUES(48890,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(48890,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(48890,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(48890,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(48890,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(48890,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(48890,36,'cp -58','g1f3 f8d6 c2c4 c7c6 b2b4 e8g8 b1d2 f8e8 c1b2 b8d7 f1e2 d7f8 e1g1 f8g6 f1e1 h7h6 d1c2 c8g4 e2f1 b7b6 b2c3 a8c8 h2h3 g4d7 a3a4 e5e4 d3e4 f6e4 c4c5 b6c5 d2e4 d5e4 b4c5 e4f3 c5d6 f3g2 f1g2');
INSERT INTO evaluations VALUES(50720,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(50720,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(50720,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(50720,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(50720,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(50720,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(50720,36,'cp 86','d4f5 c8e6 f1e2 g7g6 f5e3 g8f6 e3d5 b8d7 e2c4 a8c8 d5f6 d7f6 c4b3 b7b5 c1g5 f8e7 g5f6 e7f6 a2a4 b5b4 c3d5 f6g5 a4a5 c8c5 d5b4 e6b3 c2b3 c5a5 e1g1 e8g8 d1d3 a5a1 f1a1 d8b6 b4d5 b6b5 d3c2 g8g7 g2g3 g5d8 a1a4 a6a5 g1g2');
INSERT INTO evaluations VALUES(50890,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(50890,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(50890,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(50890,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(50890,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(50890,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(50890,36,'cp 138','f8e7 f1d3 b8d7 e1g1 e6e5 c1e3 g8f6 d3c2 c7c6 d4d5 f6g4 e3c1 e8g8 h2h3 g4f6 c1e3 f8e8 a2a4 a6a5 d1e2 d7f8 g1h2 c8d7 a1c1 f6h7 g2g3 f8g6 c1d1 a8c8 b2b3 e8f8 f3e1 h7f6 e1g2 c6d5 e4d5');
INSERT INTO evaluations VALUES(53698,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53698,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(53698,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53698,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(53698,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(53698,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53698,36,'cp 0','f1c4 g8f6 b1c3 f8b4 g1f3 c6d5 e4d5 e8g8 e1g1 e5e4 f3d4 b8d7 d4e6 f7e6 d5e6 g8h8 e6d7 c8d7 c1g5 b4c3 b2c3 d8c7 g5f6 f8f6 d1d4 f6h6 g2g3 d7c6 a1d1 c7a5 c4d5 a8d8 c3c4 h6d6 d4e4 c6d5 c4d5 d6d5 d1e1 h7h6 e1e2 a5c7 e4a4 d5d4 a4b5 c7d7 b5d7 d8d7 e2e8 h8h7');
INSERT INTO evaluations VALUES(53689,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53689,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(53689,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53689,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53689,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53689,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(53689,36,'cp 0','g8f6 b1c3 f8b4 g1f3 c6d5 e4d5 e8g8 e1g1 e5e4 f3d4 b8d7 d4e6 f7e6 d5e6 g8h8 e6d7 c8d7 c1g5 b4c3 b2c3 d8c7 g5f6 f8f6 d1d4 f6d6 c4d5 d7c6 c3c4 c7a5 f1d1 a8d8 d4e4 c6d5 c4d5 d6d5 d1e1 h7h6 g2g3 b7b6 e4g6 d5e5 e1d1 e5d5');
INSERT INTO evaluations VALUES(54202,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(54202,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(54202,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54202,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54202,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54202,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(54202,36,'cp 0','g8f6 f1d3 b7b6 e1g1 b8c6 b1d2 c8b7 c1b2 a8c8 f1e1 c5d4 e3d4 f8d6 c2c3 a7a5 f3e5 e8g8 a2a4 c6e7 a1c1 c8c7 d1e2 e7f5 e1f1 d8c8 f1e1');
INSERT INTO evaluations VALUES(46223,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46223,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(46223,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(46223,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46223,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46223,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46223,36,'cp 0','f1d3 b7b6 e1g1 b8c6 c1b2 c8b7 b1d2 f8e7 a2a3 e8g8 f1e1 a7a5 c2c4 c5d4 e3d4 a8c8 d1c2 c6b8 f3e5 b8c6');
INSERT INTO evaluations VALUES(56270,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(56270,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(56270,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(56270,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(56270,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(56270,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(56270,36,'cp -37','g8f6 d2d4 b8c6 f1b5 c8g4 b5c6 b7c6 f3e3 f8e7 g1f3 e8g8 f3e5 e7d6 e1g1 f8e8 e3g3 g4f5 e5c6 d8c8 c6e5 f5c2 c1g5 c2g6 g3f3 d6e5 d4e5 f6e4 g5f4 e4c3 b2c3 c7c6 h2h4 h7h6 a1d1 g6f5 d1d2 c8e6 f3e3 e6g6 f1d1 e8b8 e3g3 g6g3 f4g3 f5e6');
INSERT INTO evaluations VALUES(48730,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(48730,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(48730,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(48730,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(48730,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(48730,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48730,36,'cp -44','d2d4 b8c6 f1b5 c8g4 f3e3 f8e7 b5c6 b7c6 g1f3 e8g8 e1g1 f8e8 f3e5 e7d6 e3g3 g4f5 c1g5 e8e6 f1e1 f5c2 e5c6 d8e8 e1e6 e8e6 c6e5 c7c6 g5f6 e6f6 f2f4 a8b8 b2b3 d6a3 a1e1 c2f5 g3f2 f6d6 g2g4 f5d7 f4f5 f7f6 e5d7');
INSERT INTO evaluations VALUES(48659,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(48659,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(48659,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(48659,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(48659,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(48659,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(48659,36,'cp -33','c8g4 f3d3 b8c6 a2a3 f8e7 f1e2 d8d7 e2g4 d7g4 g1e2 g4g6 d3g6 h7g6 c1g5 e8c8 h2h4 f6g4 g5e7 c6e7 f2f3 g4f6 e1f2 e7f5 g2g3 c7c6 a1e1 d8e8 a3a4 b7b6 c3d1 c8b7 c2c3 c6c5 b2b3 c5d4 e2d4 f5d4 c3d4');
INSERT INTO evaluations VALUES(9355,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9355,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(9355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9355,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9355,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9355,36,'cp 1','e8g8 c2c3 f6e4 d2d4 e5d4 c3d4 c5b4 e1e4 d6d5 c1g5 f7f6 g5f6 f8f6 c4b3 c6e7 b1d2 c7c6 f3e5 e7g6 d2f3 c8e6 e4e2 g6e5 f3e5 d8d6 b3c2 b4a5 d1c1 a8f8 c1e3 a5b6 a1e1 d6b4 e1d1 b4e7 e3d3 e6f5 d3c3 f5c2 c3c2');
INSERT INTO evaluations VALUES(3836,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3836,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3836,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3836,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3836,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3836,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3836,36,'cp 0','c2c3 f6e4 d2d4 e5d4 c3d4 c5b4 e1e4 d6d5 c1g5 f7f6 g5f6 f8f6 c4b3 c6e7 b1d2 c7c6 f3e5 c8e6 e4e2 e7g6 d2f3 g6e5 f3e5 d8d6 b3c2 b4a5 a1c1 a8f8 d1d3 g7g6 d3e3 a5b6 c2b1 f6f4 c1d1 f4f6');
INSERT INTO evaluations VALUES(10793,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(10793,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(10793,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(10793,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(10793,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(10793,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(10793,36,'cp 200','d7d6 e1c1 c8e6 f1b5 c7c6 b5e2 e8g8 d4d6 d8a5 a2a3 h7h6 f2f4 f8d8 d6e7 a5b6 f4f5 d8d1 h1d1 a8e8 e7d6 b6e3 d6d2 e3d2 c1d2 e6c8 e2f3 g8f8 h2h3 g7g6 g2g4 b7b6 d1h1 f6d7 d2e3 c8a6 b2b3 d7e5 f3e2 a6e2 c3e2 g6f5 g4f5 f8e7 e2f4 e7f6 h1d1');
INSERT INTO evaluations VALUES(12523,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(12523,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12523,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12523,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(12523,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(12523,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(12523,36,'cp 23','f8e7 d2d3 e8g8 c1g5 d7d6 g5f6 e7f6 c3d5 g7g6 c2c3 f6g7 a2a4 c8e6 d1d2 g8h8 b2b4 e6g4 f3e1 c5b4 c3b4 c6d4 e1c2 d4e2 g1h1 d8h4 d2e1 e2f4 f2f3 h4e1 f1e1 g4e6 c2e3 h7h5 b4b5 g7h6 g2g3 f4d5 e3d5 a8c8 a4a5 h8g7 h2h4 c8c5 e1b1 f8c8 b5b6 e6d5 c4d5 c5c1 b1c1 c8c1 a1c1 h6c1');
INSERT INTO evaluations VALUES(7937,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(7937,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(7937,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(7937,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(7937,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(7937,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(7937,36,'cp 35','d2d3 d7d6 c1g5 a8b8 g5f6 e7f6 a2a4 c6e7 c3d5 e7d5 c4d5 c8e6 d5e6 f7e6 d1e2 e8g8 f1b1 d8d7 b2b4 c5b4 b1b4 f6d8 a4a5 b7b6 a5a6 h7h6 h2h3 b8c8 b4b3 f8f7 c2c4 d8e7 e2a2 f7f8 a2a4 d7a4 a1a4');
INSERT INTO evaluations VALUES(7938,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(7938,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(7938,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(7938,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(7938,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(7938,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(7938,36,'cp 26','d7d6 c1g5 e8g8 g5f6 e7f6 c3d5 g7g6 a2a4 f6g7 c2c3 c8e6 d1d2 c6a5 c4a2 a5c6 b2b4 c5b4 c3b4 c6e7 d5e3 d6d5 f3g5 g7h6 g5e6 f7e6 d2b2 d8d6 b4b5 a8c8 a1e1 h6e3 e1e3 d5d4 e3e2 g8g7 h2h3');
INSERT INTO evaluations VALUES(8132,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8132,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(8132,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(8132,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(8132,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(8132,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(8132,36,'cp 40','c1g5 e8g8 g5f6 e7f6 c3d5 c8e6 c2c3 g7g6 a2a4 f6g7 d1d2 c6e7 d5e7 d8e7 a4a5 f8b8 f1b1 b7b5 a5b6 a7b6 c4e6 e7e6 h2h4 h7h5 b2b4 e6d7 a1a8 b8a8 b4c5 b6c5 g2g3 d6d5 g1g2 d5e4 d3e4 d7d2 f3d2 g7h6 d2c4');
INSERT INTO evaluations VALUES(12884,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(12884,31,'cp 310',NULL);
INSERT INTO evaluations VALUES(12884,32,'cp 301',NULL);
INSERT INTO evaluations VALUES(12884,33,'cp 297',NULL);
INSERT INTO evaluations VALUES(12884,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(12884,35,'cp 310',NULL);
INSERT INTO evaluations VALUES(12884,36,'cp 295','d4d5 f6e4 d5c6 b7c6 b1d2 e4f6 g1f3 g7g6 b2b3 f8g7 c1b2 e8g8 f1e2 c6c5 e1g1 c8b7 a1c1 a8c8 f1e1 e7e6 e2f1 d8e7 d2c4 d7d5 c4e5 f6e4 f3d2 e4f6 g2g3 f8d8 d1e2 f6d7 c1c2 d7f8 e5d3 g7b2 c2b2');
INSERT INTO evaluations VALUES(20835,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(20835,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(20835,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(20835,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(20835,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(20835,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(20835,36,'cp 45','g1f3 c8g4 e2e4 e7e5 e4d5 c6d4 f4e3 d8e7 f1e2 d4e2 d1e2 e7d7 a2a3 f8d6 h2h3 g4f3 e2f3 g8e7 f3h5 e7g6 e1c1 f6f5 h5e2 a7a6 c1b1 e8g8 h3h4 e5e4 h4h5 g6e5 f2f3 e4f3 g2f3 a8e8 e3c1 e5g4 e2d2 g4e3 d1e1 e3c4 d2d4 c4e5 d4f2 b7b5 c3e2 e5c4 h5h6 g7g6 e2d4 d6c5');
INSERT INTO evaluations VALUES(26903,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(26903,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(26903,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(26903,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(26903,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(26903,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(26903,36,'cp 172','d2d4 e5d4 g2g4 h5g6 f3d4 d8d7 b1c3 g8e7 f2f4 f7f6 f4f5 g6f7 c3d5 a7a6 d4e6 a8c8 b5c4 c6e5 d1e2 e7d5 c4d5 g7g5 h3h4 h7h6 d5b7 c8b8 b7d5 a6a5 g1g2 c7c6 d5b3 a5a4 b3a4 f7e6 f5e6 d7e6 c1e3 h6h5 e3d4 g5h4 d4e5 d6e5 g4h5 f8d6 a4b3');
INSERT INTO evaluations VALUES(31102,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(31102,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(31102,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(31102,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(31102,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(31102,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(31102,36,'cp 66','e4d5 c6d5 g2g4 h5g6 d2d4 e7e6 h3h4 h7h6 f3e5 g8e7 e5g6 e7g6 h4h5 g6e7 d1d3 e7c6 c1e3 d8a5 a2a3 b8d7 e1c1 d7f6 f1e2 a7a6 f2f4 b7b5 g4g5 b5b4 g5f6 b4c3 d3c3 a5c3 b2c3 g7f6 c1d2 f8a3 d1a1 a3d6 h1b1 c6e7 c3c4 d5c4 e2c4 e8g8 c4a6 g8g7 c2c4 e7f5 c4c5 d6c7');
INSERT INTO evaluations VALUES(31056,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(31056,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(31056,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(31056,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(31056,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(31056,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(31056,36,'cp 65','c6d5 g2g4 h5g6 f3e5 b8c6 d2d4 a7a6 h3h4 c6e5 d4e5 h7h5 d1d5 d8d5 c3d5 a8c8 f1g2 h5g4 h4h5 g6c2 d5b6 c8b8 c1e3 e7e6 a1c1 c2d3 c1c7 f8b4 e1d1 g8e7 a2a3 b4a5 b2b4 a5b6 e3b6 e7d5 g2d5 e6d5 b6c5 d3e4');
INSERT INTO evaluations VALUES(31337,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(31337,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(31337,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(31337,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(31337,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(31337,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(31337,36,'cp 85','g2g4 h5g6 f3e5 b8c6 d2d4 f7f6 e5g6 h7g6 c1e3 e7e6 d1d3 f6f5 e1c1 e8f7 c1b1 a8c8 f1e2 f8e7 d1g1 g8f6 a2a3 f6e4 c3e4 d5e4 d3d1 e7f6 c2c3 c6a5 f2f3 d8d5 e3c1 h8h4 g4f5 e6f5 f3e4 h4e4 e2f3 a5b3');
INSERT INTO evaluations VALUES(33583,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(33583,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(33583,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(33583,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(33583,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(33583,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(33583,36,'cp 58','c8f5 c1d2 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 b7b5 c4b3 d7b6 g2g4 f8b4 c1b1 a7a5 h4h5 a5a4 g4g5 f6h5 b3e6 f7e6 e2e6 e8d8 d4d5 c7d6 d2e3 d6e6 e3b6 d8e7 d5e6 b4c3 b2c3 e7e6 h1e1 e6f5 d1d6 a8e8 e1e8 h8e8 d6c6 h5f4 b1b2 e8e4 c6c5 e4e5 c3c4 e5c5 b6c5');
INSERT INTO evaluations VALUES(32571,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(32571,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(32571,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32571,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(32571,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(32571,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(32571,36,'cp 67','f3e5 e7e6 g2g4 f5g6 h2h4 b8d7 e5d7 f6d7 h4h5 g6e4 e1g1 e4d5 c3d5 c6d5 c4d3 f8d6 c1d2 a5b6 c2c3 e8g8 d1f3 a8e8 a1e1 b6c6 d3c2 e6e5 h5h6 g7g6 d4e5 d6b8 d2f4 d7e5 f4e5 b8e5 c2b3 c6c7 g4g5 e8d8 e1d1 d5d4 c3d4 d8d4 d1d4 e5d4 f1d1 c7e5');
INSERT INTO evaluations VALUES(34668,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34668,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(34668,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(34668,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(34668,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(34668,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(34668,36,'cp 0','g1f3 b8c6 f1b5 c8d7 d2d4 e7e5 d4d5 c6b4 b5d7 e8d7 f3e5 d7e8 e1d2 b4d5 d2e2 g8e7 h1d1 f7f6 e5f3 a8d8 c2c4 d5b4 d1d8 e8d8 a2a3 b4c6 b2b4 d8c8 b1c3 a7a5 b4b5 c6d8 e2d3 e7g6 c1e3 d8e6 g2g3 b7b6 d3c2 c8b7 c3e4 f8e7 e4c3');
INSERT INTO evaluations VALUES(34666,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34666,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34666,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34666,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34666,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34666,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34666,36,'cp 0','b8c6 f1b5 c8d7 d2d4 e7e5 d4e5 c6e5 b5d7 e5d7 b1c3 e8c8 e1g1 g8f6 f3g5 d7e5 f2f4 f8c5 g1h1 e5g4 g5f7 h8e8 f7d8 g4f2 h1g1 f2g4');
INSERT INTO evaluations VALUES(34738,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(34738,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(34738,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34738,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(34738,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(34738,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(34738,36,'cp 6','d2d4 c8f5 c1e3 b8c6 b1c3 g8f6 e1c1 e8c8 h2h3 f6e4 c3e4 f5e4 f3d2 e4g6 a2a3 h7h5 d2b3 e7d7 c1b1 f8d6 e2d2 h5h4 f1e2 c6e7 h1e1 d7c6 b3c1 a7a6 e2d3 g6d3 d2d3 c6b5 d3b5 a6b5 c1d3 e7c6 e3f4 d6f4 d3f4');
INSERT INTO evaluations VALUES(34735,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34735,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(34735,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(34735,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34735,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(34735,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34735,36,'cp -6','c8f5 c1e3 g8f6 b1c3 b8c6 e1c1 e8c8 h2h3 f6e4 c3e4 f5e4 f3d2 e4g6 a2a3 e7d7 d2b3 f8d6 c1b1 h7h5 e2d2 b7b6 f1a6 c8b8 h1e1 h8e8 d2e2 h5h4 e2d2 f7f6 e3f4 e8e1 d1e1 d8e8 e1e8 g6e8 f4d6 d7d6 d2e3 e8d7 b3c1 g7g5 c1d3');
INSERT INTO evaluations VALUES(39599,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(39599,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(39599,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(39599,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(39599,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(39599,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(39599,36,'cp 9','c2c4 c7c6 g1f3 d7d5 c4d5 c6d5 b1c3 g8f6 f1g2 e8g8 f3e5 c8f5 e1g1 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 h2h4 a7a5 c3c4 a8c8 a1c1 d8b6 e2e3 f5d3 c4c5 b6b4 a4b4 a5b4 f1d1 d3f5');
INSERT INTO evaluations VALUES(43057,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(43057,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(43057,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43057,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(43057,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(43057,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43057,36,'cp 62','b8c6 c3e4 d7d6 d1e2 h7h6 c1d2 e7d5 e1c1 d8e7 c2c4 d5b6 e2e3 d6e5 f3e5 c6e5 d4e5 g7e5 d2c3 e5c3 e3c3 e6e5 c4c5 c8f5 f1b5 c7c6 e4d6 e8f8 c5b6 c6b5 d6f5 g6f5 h1e1 b5b4 c3e5 e7e5 e1e5 a7b6 d1d7 h8g8 e5f5 g8g2 d7f7 f8g8 f7b7 g2h2 b7b6');
INSERT INTO evaluations VALUES(43412,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(43412,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(43412,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(43412,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(43412,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43412,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(43412,36,'cp 14','g7g6 c2c4 f8g7 b1c3 b8c6 d4c6 b7c6 d1a4 c8d7 h2h3 e8g8 a4c2 d7e6 c1e3 d6d5 c4d5 c6d5 f1d1 d8b8 e4d5 f6d5 c3d5 e6d5 d3g6 h7g6 d1d5 b8b2 c2b2 g7b2 a1b1 b2a3 d5a5 a3d6 g2g3 f8c8 g1g2 g8g7 h3h4 d6c7 a5c5 c7d6 c5c8 a8c8 b1b3');
INSERT INTO evaluations VALUES(51190,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(51190,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(51190,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51190,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(51190,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51190,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51190,36,'cp 0','c8b7 f1d3 d7d6 d1e2 b8d7 c1d2 g8f6 h2h4 h7h5 e1c1 a8c8 c1b1 f8e7 f2f3 d7e5 d4b3 d8c7 d2g5 c7b6 g5e3 b6c7');
INSERT INTO evaluations VALUES(30326,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30326,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30326,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(30326,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(30326,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30326,36,'cp 0','f1d3 d7d6 h2h4 g8f6 h4h5 b8d7 d1e2 a8c8 c1d2 d8b6 d2e3 b6d8');
INSERT INTO evaluations VALUES(30327,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30327,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30327,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30327,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30327,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30327,36,'cp 0','d7d6 d1e2 b8d7 h2h4 h7h5 c1d2 f8e7 e1c1 a8c8 f2f3 d7e5 d4b3 g8f6 c1b1 d8c7 d2g5 c7b8 c3a2 b8c7');
INSERT INTO evaluations VALUES(52430,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(52430,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(52430,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(52430,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(52430,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(52430,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52430,36,'cp 0','c5d4 d1d4 b8c6 d4d3 a7a6 e2e4 b7b5 d3g3 d7d6 f1d3 b5b4 c3d5 e7e6 d5e3 g8f6 g1f3 f6h5 g3g4 g7g6 e1g1 f8g7 a2a3 e8g8 f1d1 e6e5 g4g5 g7f6 g5h6 f6g7');
INSERT INTO evaluations VALUES(52522,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52522,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52522,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52522,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52522,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52522,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52522,36,'cp 0','d1d4 b8c6 d4d3 a7a6 e2e4 b7b5 d3g3 d7d6 f1d3 b5b4 c3d5 e7e6 d5e3 g8f6 g1f3 f6h5 g3h3 h5f4 h3g3');
INSERT INTO evaluations VALUES(52517,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52517,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52517,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52517,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52517,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52517,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52517,36,'cp 0','b8c6 d4d3 a7a6 e2e4 b7b5 d3g3 d7d6 f1d3 b5b4 c3d5 e7e6 d5e3 g8f6 g1f3 f6h5 g3h3 h5f4 h3g3');
INSERT INTO evaluations VALUES(19549,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19549,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19549,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19549,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19549,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19549,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19549,36,'cp 0','d4d3 a7a6 e2e4 b7b5 d3g3 d7d6 f1d3 b5b4 c3d5 e7e6 d5e3 g8f6 h2h4 f8e7 h4h5 e8g8 g1f3 d6d5 e4d5 e6d5 b2b3 h7h6 c1b2 e7d6 g3h4 d6e7');
INSERT INTO evaluations VALUES(53449,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(53449,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(53449,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(53449,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(53449,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(53449,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(53449,36,'cp -12','d8a5 g1f3 b8c6 b1c3 c8g4 d2d4 e7e6 f1e2 f8b4 d1d3 g4f5 d3e3 g8e7 e1g1 f5g4 e3d3 e7f5 c1e3 f5e3 f2e3 g4f5 d3d2 e8g8 a1c1 a5b6 a2a3 b4c3 c1c3 h7h6 f1c1 f8c8 b2b4 b6d8 c3c5 c6e7 e2d3 c8c5 b4c5 b7b6');
INSERT INTO evaluations VALUES(54764,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(54764,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(54764,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(54764,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(54764,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(54764,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(54764,36,'cp 59','e7e6 f1e2 a7a5 b2b3 f8d6 c2c4 g8f6 e1g1 e8g8 c1a3 c8d7 f3e5 d6a3 b1a3 d8e7 a3b1 b8c6 e5d7 e7d7 b1c3 d7e7 a1c1 g8h8 c4d5 e6d5 d1d3 e7e6 a2a3 a8e8 c1c2 e6d7 c3a4 f6e4 d3b5 f5f4 b5b7 f4e3');
INSERT INTO evaluations VALUES(56771,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(56771,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(56771,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(56771,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(56771,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(56771,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(56771,36,'cp 52','f1e2 a7a5 e1g1 f8d6 c2c4 g8f6 b2b3 e8g8 c1a3 c8d7 f3e5 d6a3 b1a3 d8e7 d1c1 d7e8 a3c2 b8d7 e5d7 f6d7 c1b2 e8g6 g2g3 c7c6 f2f4 h7h6 c2e1 c6c5 e1d3 b7b6 a1c1 a8d8 f1d1 g6f7');
INSERT INTO evaluations VALUES(54965,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(54965,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(54965,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54965,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54965,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(54965,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(54965,36,'cp 22','g8f6 g1f3 c8f5 c3c4 b8d7 g2g3 h7h6 f1g2 c7c6 b1c3 e7e5 d4d5 c6d5 f3h4 f5h7 c3d5 f6d5 g2d5 d8a5 d1d2 a5d2 c1d2 a8b8 h4g2 d7b6 g2e3 b6d5 e3d5 h7e4 e1g1 f8e7 a1c1 e8d7 d5e7 d7e7 f1d1 h8d8 f2f3 e4c6 b2b3 b8c8 d2e1 a7a6 e2e4 e7e6');
INSERT INTO evaluations VALUES(47012,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47012,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(47012,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(47012,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47012,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(47012,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47012,36,'cp 42','g1f3 g7g6 b1d2 f8g7 e2e4 e8g8 f1d3 c7c5 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 a2a4 b8a6 f1e1 f6d7 d3a6 a8a6 d2c4 d7c5 b2b3 b7b6 c1a3 a6a8 a3c5 b6c5 d1d2 a8a7 a1c1 f7f5 d2g5 f5e4 g5d8 f8d8 f3g5 g7h6 h2h4 h6g5 h4g5 c8f5');
INSERT INTO evaluations VALUES(47011,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(47011,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(47011,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(47011,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(47011,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(47011,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(47011,36,'cp 25','g7g6 b1d2 f8g7 e2e4 e8g8 f1d3 c7c5 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 a2a4 b8a6 d2c4 a6c5 f1e1 b7b6 c1g5 c8a6 a1c1 h7h6 g5e3 f6e4 d3e4 c5e4 e3b6 d8b8 e1e4 f7f5 e4e2 a6c4 c1c4 b8b6 c4c6 b6d8 d1d2 e5e4 f3d4 g7d4 d2d4');
INSERT INTO evaluations VALUES(56195,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(56195,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(56195,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(56195,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(56195,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(56195,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(56195,36,'cp 151','e4d5 d8d5 d1e2 c8e6 d4b5 d5d7 c1f4 b8a6 b1c3 g8f6 a1d1 d7c8 c3e4 f6e4 e2e4 f8b4 c2c3 b4e7 f1d3 a6c5 b5c7 c8c7 f4c7 c5e4 d3e4 a8c8 c7f4 c8c4 d1d4 c4d4 c3d4 g7g5 f4d2 e8g8 d2c3 f8b8 d4d5 e6d7 e1d2 f7f5 e4c2 b8e8 h1e1 g8f7 d2d3 g5g4 c2b3 e7d6 e1e8 f7e8 c3d4 d6h2 d4a7');
INSERT INTO evaluations VALUES(56133,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(56133,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(56133,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(56133,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(56133,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(56133,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(56133,36,'cp 146','d8d5 d1e2 c8e6 d4b5 d5d7 c1f4 b8a6 b1c3 g8f6 a1d1 d7c8 c3e4 f6e4 e2e4 f8b4 c2c3 b4e7 f1d3 a6c5 b5c7 c8c7 f4c7 c5e4 d3e4 a8c8 c7e5 e8g8 b2b3 b7b6 e1g1 f8d8 f1e1 d8d1 e1d1 f7f6 e5d4 e7c5 d4c5 c8c5 c3c4 g8f7 e4d3 h7h6 f2f4 g7g5 g2g3 h6h5 g1f2 h5h4 d1d2 h4g3 h2g3 f7g7 f4g5 c5g5');
INSERT INTO evaluations VALUES(34894,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(34894,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(34894,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(34894,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(34894,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(34894,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(34894,36,'cp 195','d1e2 f8e7 c1e3 d5a5 b1c3 g8f6 e1c1 e8g8 d4b3 a5e5 f2f4 e5e6 b3d4 e6b6 h2h3 f8e8 d4b3 b6e6 b3c5 e7c5 e3c5 e6e2 f1e2 c8d7 c5d4 d7c6 d4f6 g7f6 e2f3 c6f3 g2f3 b8c6 c3d5 e8e2 h1g1 g8f8 d5f6 f8e7 f6d5 e7f8 d5c7 a8c8 c7d5 c8d8 d5f6 f8e7 f6h7 d8d1 c1d1 c6d4 h7g5 f7f6 g5e4 e2c2 g1g7 e7e6 g7b7 c2h2 b7a7 h2b2 a7a6 e6e7');
INSERT INTO evaluations VALUES(56668,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(56668,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(56668,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(56668,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(56668,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(56668,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(56668,36,'cp 48','d2d4 f8d6 f1c4 h7h6 e1g1 b8c6 d4d5 c6e7 c3e4 e7f5 f3d4 g8e7 d4f5 c8f5 e4d6 c7d6 c1f4 e8g8 f1f2 f5e4 d1d4 f7f5 a1e1 e7g6 f4g3 b7b5 c4f1 a7a6 c2c4 b5c4 f1c4 f8e8 c4d3 e4d3 e1e8 d8e8 d4d3 f5f4 g3f4 g6f4 f2f4 e8e5 f4e4');
INSERT INTO evaluations VALUES(56659,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(56659,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(56659,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(56659,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(56659,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(56659,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(56659,36,'cp 41','f8d6 f1c4 b8c6 e1g1 h7h6 d4d5 c6e7 c3e4 e7f5 f3d4 g8e7 d4f5 c8f5 e4d6 c7d6 f1f4 e8g8 a2a4 a7a6 c4f1 b7b5 c1e3 a8c8 c2c3 f5d7 a4b5 a6b5 d1d2 f8e8 f4e4 c8a8 a1e1 e7f5 e4e8 d7e8 e3f4 g7g5 f1d3 e8d7 d3f5 d7f5 f4e3 d8f6 e3d4 f6g6 d2e3 f5d7');
INSERT INTO evaluations VALUES(8618,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(8618,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(8618,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(8618,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(8618,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(8618,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(8618,36,'cp 62','d4d5 c6b8 e1g1 e8g8 d1e2 a7a5 c1e3 f6g4 e3d2 b8d7 a2a4 g4f6 a1d1 c7c6 d2g5 d7b6 c4b3 h7h6 g5e3 c6d5 e3b6 d8b6 c3d5 f6d5 d1d5 b6c6 d5b5 c6c7 e2c4 c7c6 c4d3 c6c7 b3d5 a8b8 b5b3 c8d7 b3c3 d7c6');
INSERT INTO evaluations VALUES(8569,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(8569,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(8569,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(8569,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(8569,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(8569,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8569,36,'cp 51','c6b8 a2a4 e8g8 e1g1 b8d7 d1e2 a7a5 c1e3 f6g4 e3d2 g4f6 a1d1 c7c6 d2g5 d7b6 c4b3 h7h6 g5e3 c6d5 e3b6 d8b6 c3d5 f6d5 d1d5 b6c7 d5d3 c8d7 f1d1 a8d8 d3c3 c7b6 f3d2 d7c6 d2f1 e7g5 e2d3 g8h8 f1e3 g5e3 d3e3 b6e3 c3e3 g7g6 b3d5 c6a4');
INSERT INTO evaluations VALUES(11264,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(11264,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11264,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11264,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(11264,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11264,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(11264,36,'cp 37','d4c6 b7c6 f1e1 f8e7 e4e5 d6e5 d1d8 e8d8 e1e5 f6d5 c1d2 e7f6 e5h5 d8e7 c3a4 e6e5 a1e1 g7g6 h5e5 f6e5 e1e5 c8e6 c4b3 h8b8 e5e1 e7f8 f2f3 b8e8 g1f2 d5f6 c2c4 f6d7 d2b4 f8g8 b4c3 e8c8 e1d1 a6a5 d1d4 h7h5 d4d6');
INSERT INTO evaluations VALUES(12913,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(12913,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(12913,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(12913,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(12913,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(12913,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(12913,36,'cp 144','d2d4 e7e6 f1d3 d7d5 e1g1 a7a6 f1e1 g4h6 d4c5 f8c5 c1h6 g7h6 d1d2 c5f8 b1a3 f8g7 a3c2 d8c7 c2b4 c6b4 c3b4 c8d7 d3c2 a8c8 a1c1 c7b6 e1e3 e8e7 h2h3 c8c4 a2a3 h8g8 c2h7 c4c1 d2c1 g8c8 h7c2 d7b5');
INSERT INTO evaluations VALUES(12911,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(12911,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(12911,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(12911,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(12911,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(12911,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(12911,36,'cp 145','e7e6 f1d3 d7d5 e1g1 a7a6 d1e2 f7f5 h2h3 g4h6 c1h6 g7h6 b1d2 f8e7 e2e3 c5d4 f3d4 e8g8 d2f3 f5f4 e3d2 g8h8 g1h1 f8f7 f1g1 d8f8 c3c4 c6d4 f3d4 f8g7 c4d5 g7e5 d4e6 c8e6 d5e6 e5e6');
INSERT INTO evaluations VALUES(16363,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(16363,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(16363,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(16363,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(16363,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(16363,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(16363,36,'cp -40','e5d4 e4e5 d8e7 d1e2 h7h6 g5h4 g7g5 e5f6 e7e2 f1e2 g5h4 f3h4 h6h5 b1d2 d7d5 e1c1 h8h6 h1e1 c8e6 f2f4 e8c8 g2g3 f8d6 d2b3 d8h8 f4f5 e6d7 e2f3 h6f6 f3d5 c6e7 d5e4 c7c5 b3d2 b7b5 e4d3 e7d5 d2e4 f6h6 e4d6 h6d6 d3e4');
INSERT INTO evaluations VALUES(19167,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(19167,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(19167,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(19167,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(19167,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(19167,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(19167,36,'cp 191','d7d5 d3d5 d8d5 e4d5 g8f6 c2c4 f8d6 b1c3 c8f5 f2f3 f6d7 c1e3 h7h5 b2b3 e8c8 h2h4 d6e5 a1c1 h8e8 e1d2 b7b6 f1d3 f5d3 d2d3 d7f8 c3e4 f7f5 e4c3 f8g6 e3g5 g6f4 g5f4 e5f4 c1e1 c8d7 e1e2 f4g3 d3c2 e8e2 c3e2 g3d6 h1e1');
INSERT INTO evaluations VALUES(22885,30,'cp -896',NULL);
INSERT INTO evaluations VALUES(22885,31,'cp -898',NULL);
INSERT INTO evaluations VALUES(22885,32,'cp -905',NULL);
INSERT INTO evaluations VALUES(22885,33,'cp -917',NULL);
INSERT INTO evaluations VALUES(22885,34,'cp -921',NULL);
INSERT INTO evaluations VALUES(22885,35,'cp -925',NULL);
INSERT INTO evaluations VALUES(22885,36,'cp -934','d8h4 g2g3 h4e4 d1e2 e4h1 g1f3 d7d5 d2d3 c8f5 f3e5 e8c8 e5c6 b7c6 c1f4 g8f6 e2g2 f8b4 c2c3 h1g2 f1g2 h8e8 e1d2 b4c5 d3d4 c5d6 b1a3 d6a3 b2a3 c8d7 g2f3 d8b8 f4e5 b8b2 d2e3 b2c2 e3f4 f5e4 a1f1 e4f3 f1f3 c2h2 f3f1 h2a2 e5f6 g7f6');
INSERT INTO evaluations VALUES(29681,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29681,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29681,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(29681,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29681,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(29681,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(29681,36,'cp 23','d1d3 b8d7 b2b3 e7e6 e1g1 f8e7 c2c4 e8g8 b1c3 f8e8 f1d1 c7c6 e3e4 d5e4 c3e4 d8c7 e4f6 e7f6 c1e3 e6e5 f3g5 f6g5 e3g5 h7h6 g5e3 e5e4 d3c3 f7f5 d4d5 d7f6 d5d6 c7f7 a2a4 e8d8 b3b4 f7e6 c4c5 f6d5 c3b3');
INSERT INTO evaluations VALUES(29680,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(29680,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(29680,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29680,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29680,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(29680,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(29680,36,'cp 9','b8d7 b2b3 e7e6 e1g1 f8e7 c2c4 e8g8 b1c3 a8c8 a1b1 f8e8 c1d2 c7c6 f1d1 d8c7 e3e4 d5e4 c3e4 c8d8 e4f6 e7f6 d2e3 d7e5 f3e5 f6e5 g2g3 e5f6 d3e4 d8d7 d1d2 e8d8 b1d1 c7a5 e4f3 f6e7 f3g4 e7f6 h2h3 g8h8 g4e2 a5f5');
INSERT INTO evaluations VALUES(30635,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(30635,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(30635,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(30635,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(30635,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(30635,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(30635,36,'cp 67','d2d4 f8b4 f2f3 d8h4 g2g3 b4c3 b2c3 h4e7 f1d3 d7d6 g1h3 b8c6 e1g1 g8f6 f3f4 e6e5 d4d5 c6b8 c4c5 b6c5 c3c4 b7c8 h3f2 b8d7 a1b1 a7a5 d1e2 h7h5 f4f5 h5h4 c1g5 h4g3 h2g3 d7b6 g1g2 h8h5 e2e3 c8a6 f1h1 h5h1');
INSERT INTO evaluations VALUES(35296,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(35296,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(35296,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(35296,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(35296,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(35296,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(35296,36,'cp 86','f1e2 e7d8 g1f3 d7d5 c2c4 f8e7 b1c3 g8f6 e1g1 e8g8 f1e1 c8e6 f3g5 b8c6 g5e6 f7e6 e2f3 d5c4 f3c6 b7c6 d1e2 d8d7 e2c4 f6d5 c3e4 d5b6 c4c3 d7d5 b2b3 a8e8 h2h3 a7a5 c1d2 e7d6 c3c2 d5d4');
INSERT INTO evaluations VALUES(35295,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(35295,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(35295,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(35295,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(35295,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(35295,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(35295,36,'cp 67','e7d8 g1f3 d7d5 e1g1 f8e7 c2c4 g8f6 b1c3 e8g8 f1e1 b8c6 h2h3 c8e6 c4c5 f6e4 e2b5 e7f6 d1a4 d8d7 c1f4 a7a6 b5c6 b7c6 c3e4 d5e4 e1e4 e6d5 e4e3 f8e8 a1e1 e8e3 f2e3 a8e8 f3e5 f6e5 f4e5 f7f6 e5h2');
INSERT INTO evaluations VALUES(36371,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(36371,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(36371,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(36371,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(36371,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(36371,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(36371,36,'cp 54','d1e2 c7c5 d4c5 b8c6 e1g1 e7c5 e3e4 e6e5 f1d1 c6d4 f3d4 e5d4 h2h3 c8e6 e4e5 f6d7 c3e4 c5b6 e4d6 e6c4 d6c4 d7c5 c4d6 c5e6 e2g4 b6c7 c1f4 c7d6 e5d6 d8f6 f4g3 f8d8 g3h4 g7g5 h4g3 d8d7 g1h1 a8d8 f2f4 e6g7 f4f5');
INSERT INTO evaluations VALUES(38998,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(38998,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(38998,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(38998,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(38998,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(38998,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(38998,36,'cp -32','d7d5 e4d5 e6d5 c4b3 d8e7 d1e2 a7a5 d3d4 c7c6 a2a4 c8g4 e2e7 g8e7 b1d2 c6c5 d4c5 e8g8 h2h3 g4f5 e1g1 b8a6 c2c3 a6c5 b3d1 e7c6 d2b3 c5d3 b3d4 g7d4 f3d4 c6d4 c3d4 f8c8 d1b3 f5e6 f1d1 d3c1 a1c1 g8f8 g2g4');
INSERT INTO evaluations VALUES(41222,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(41222,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(41222,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(41222,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(41222,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(41222,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(41222,36,'cp -44','e8g8 d2d3 h7h6 g5f3 e7e6 d3d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 c7c5 c2c3 c5d4 c3d4 b8d7 e1g1 d7f6 d1d3 f6e4 d3e4 d8f6 c1f4 f6f5 e4f5 e6f5 d4d5 g7b2 f4h6 b2a1 h6f8 a1f6 f8d6 c8d7 f1c1 a8d8 d6f4 a7a6 c1c7 d7b5 d5d6 b5c6 g1f1 d8d7 a2a3 g8g7 f3e5 c6b5 f1g1 d7d6');
INSERT INTO evaluations VALUES(36489,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(36489,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(36489,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(36489,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(36489,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(36489,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(36489,36,'cp -34','d2d3 h7h6 g5f3 e7e6 e1g1 c7c5 d3d4 c5d4 d1d4 d6d5 e4d5 f6d5 d4e4 d5c3 b2c3 c8d7 c1f4 d7c6 e4e3 c6f3 e3f3 b8c6 a1b1 c6e5 f4e5 g7e5 f3e4 e5c3 b1b7 d8f6 f1d1 c3a5 e4d4 f6f5 b7b5 a5b6 b5f5 b6d4 f5f3 d4b6 d1d7 a8c8 c4d3 f8d8 d7d8 b6d8 g2g3 d8b6');
INSERT INTO evaluations VALUES(36490,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(36490,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(36490,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(36490,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(36490,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(36490,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(36490,36,'cp -51','a7a6 a2a4 h7h6 g5f3 c7c5 e1g1 e7e6 c4b3 b8c6 f1e1 d6d5 c1f4 b7b6 e4d5 e6d5 d1d2 g8h7 h2h3 d5d4 c3e4 f6e4 d3e4 c8e6 b3e6 f7e6 b2b3 e6e5 f4g3 b6b5 a4b5 a6b5 a1a8 d8a8 e1b1 a8a7 d2d1 f8e8');
INSERT INTO evaluations VALUES(49057,30,'cp -317',NULL);
INSERT INTO evaluations VALUES(49057,31,'cp -343',NULL);
INSERT INTO evaluations VALUES(49057,32,'cp -298',NULL);
INSERT INTO evaluations VALUES(49057,33,'cp -302',NULL);
INSERT INTO evaluations VALUES(49057,34,'cp -329',NULL);
INSERT INTO evaluations VALUES(49057,35,'cp -301',NULL);
INSERT INTO evaluations VALUES(49057,36,'cp -339','d8h4 g2g3 e4g3 h2g3 h4h1 b1c3 c8e6 d2d4 b8c6 c1e3 f8d6 e1c1 d6e5 d4e5 d5d4 c3b5 e8c8 b5d4 c6d4 e3d4 h7h5 e2e3 h1h2 e3a3 d8d4 d1d4 h2f2 d4d1 c8b8 f1b5 a7a6 b5d7 e6d7 d1d7 f2g3 a3c5 h8c8 b2b3 g3f3 c5e7 f3e3 c1b2 h5h4');
INSERT INTO evaluations VALUES(52322,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(52322,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(52322,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(52322,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(52322,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(52322,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(52322,36,'cp 93','e2e4 d7d6 b1c3 g8f6 g1f3 f8g7 a2a4 e8g8 f1c4 c8g4 h2h3 g4f3 d1f3 b8d7 c4e2 a7a6 e1g1 d8c7 c1e3 c7b6 e3f4 c5c4 f1b1 b6b4 a4a5 b7b5 a5b6 d7b6 e2f1 f6d7 f4g5 d7f6 f3e2 a6a5 g5d2 f6d7 f2f4');
INSERT INTO evaluations VALUES(52321,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(52321,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(52321,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(52321,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(52321,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(52321,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(52321,36,'cp 93','d7d6 b1c3 f8g7 f1c4 g8f6 g1f3 e8g8 a2a4 e7e6 e1g1 h7h6 f1e1 e6d5 c3d5 f8e8 c1f4 b8c6 h2h3 c8e6 f4d6 e6d5 c4d5 f6d5 d1d5 c6b4 d5c5 b4c2 c5c2 d8d6 e1e3 d6b4 c2e2 a7a6 e4e5 a8c8 g2g3 b4c4 e2c4 c8c4 b2b3 c4c2');
INSERT INTO evaluations VALUES(52632,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(52632,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(52632,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(52632,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(52632,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(52632,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(52632,36,'cp 100','b1c3 g8f6 f1c4 f8g7 g1f3 e8g8 e1g1 b8a6 c1f4 f6h5 f4e3 a6c7 a2a4 b7b6 d1d3 a7a6 f1e1 a8b8 h2h3 d8d7 d3d2 b6b5 a4b5 a6b5 c4f1 b5b4 c3a4 c8a6 f1a6 c7a6 e4e5 d6e5 a4c5 a6c5 e3c5 f8d8 a1d1 d7c7 c5b4');
INSERT INTO evaluations VALUES(54387,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(54387,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(54387,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(54387,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(54387,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(54387,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(54387,36,'cp 42','e3d4 g8e7 f1d3 e7g6 f4g3 f8d6 g1e2 b8c6 b1d2 e6e5 d4e5 c6e5 d3c2 e8g8 h2h3 e5c6 g3d6 d8d6 d2f3 f8e8 e1g1 c8d7 f1e1 c6e5 f3e5 e8e5 d1d2 a8e8 e2d4 g6f4 h3h4 d6f6 g2g3 e5e1 a1e1 e8e1 d2e1 f4e6 d4b3 d7b5 e1e3 a7a6 b3c5 e6c5 e3c5');
INSERT INTO evaluations VALUES(54361,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(54361,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(54361,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(54361,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(54361,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(54361,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(54361,36,'cp 35','g8e7 f1d3 e7g6 f4g3 f8d6 g1e2 b8c6 b1d2 e6e5 d4e5 c6e5 d3c2 e8g8 h2h3 e5c6 g3d6 d8d6 e1g1 d6f6 f1e1 f8e8 c2g6 h7g6 e2d4 c8d7 e1e8 a8e8 d2f3 c6d4 d1d4 f6d4 f3d4 g8f8 f2f3 f8e7 a1e1 e7d6 e1e8 d7e8 g1f2 d6e5 h3h4 f7f6 g2g3 e8d7 d4b3');
INSERT INTO evaluations VALUES(5811,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(5811,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(5811,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(5811,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(5811,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(5811,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(5811,36,'cp 62','c1d2 g8f6 g1f3 e8g8 a2a3 b4d6 c4c5 d6e7 f1b5 f6e4 e1g1 c8d7 b5d3 f7f5 b2b4 e7f6 c3e2 a7a6 d2e1 b7b6 h2h3 d8e8 d3c2 g7g5 f3h2 f6g7 a1b1 e6e5 c2b3 d7e6 d4e5 c6e5 c5b6 c7b6 b3d5 a8d8 d5e6 e8e6 d1a4 g5g4');
INSERT INTO evaluations VALUES(5810,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(5810,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(5810,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(5810,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(5810,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(5810,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(5810,36,'cp 51','g8f6 a2a3 b4c3 d2c3 e8g8 d1c2 b7b6 c4d5 f6d5 c3d2 c8b7 g1f3 a8c8 f1e2 c6b8 e1g1 c7c5 d4c5 b8d7 f1c1 c8c5 c2d1 c5c1 a1c1 h7h6 b2b4 a7a6 h2h3 d5f6 d2c3 d8a8 c3b2 f8c8 c1c8 a8c8 b4b5 a6b5 f3d4 b7d5 d4b5 c8d8 b5c3 d7c5 c3d5 f6d5');
INSERT INTO evaluations VALUES(7439,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(7439,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(7439,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(7439,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(7439,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(7439,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(7439,36,'cp 71','c3d5 e8g8 c2c3 b4e7 e1g1 f6e4 d2d4 e5d4 f1e1 e4f6 f3d4 f6d5 b3d5 e7g5 d4f5 g5c1 d1c1 d8f6 c1f4 g7g6 f5h6 g8g7 f4d2 a8b8 h6g4 f6h4 e1e4 f7f5 d5c6 f5e4 d2d4 f8f6 c6d7 c8d7 g4f6 h4f6 d4d7 g7h6 a1d1 b8f8 d7d4 f6d4 d1d4 f8e8 d4d7 c7c6 d7d6 e4e3 f2e3 e8e3 d6c6 e3e2');
INSERT INTO evaluations VALUES(19958,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(19958,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(19958,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(19958,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(19958,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19958,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(19958,36,'cp -16','g7g6 g1f3 f8g7 e1g1 e7e6 d2d4 c5d4 c3b5 a7a6 b5d4 c6d4 f3d4 g8f6 e4e5 d6e5 f4e5 f6d5 c2c3 e8g8 d4f3 d8b6 d1d4 b6d4 c3d4 c8d7 c4d5 e6d5 c1g5 f8c8 a1c1 g7f8 g1f2 h7h6 g5d2 a6a5 h2h4 a5a4 c1c8 a8c8 f1c1 c8c1 d2c1 g8g7 f3e1 d7f5');
INSERT INTO evaluations VALUES(25136,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(25136,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(25136,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(25136,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(25136,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25136,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(25136,36,'cp 43','f8b4 a1c1 f6e4 c4d5 e6d5 f1d3 e8g8 e1g1 b4c3 b2c3 c6a5 f3g5 a5c4 g5e4 c4b2 d1c2 b2d3 c2d3 f5e4 d3b5 b7b6 f2f3 c7c6 b5c6 e4d3 f1d1 d3c4 c6a4 b6b5 a4c2 d8d7 h2h3 h7h6 g1h2 a8c8 e3e4 d5e4 f3e4 f7f5 f4e5 f5e4 c2e4 f8e8');
INSERT INTO evaluations VALUES(24088,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(24088,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(24088,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(24088,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(24088,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(24088,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(24088,36,'cp 44','a1c1 f6e4 c4d5 e6d5 f1d3 e8g8 e1g1 b4c3 b2c3 c6a5 f3g5 f8e8 f2f3 e4g5 d3f5 a5c4 f1e1 b7b5 c1b1 a7a6 f5d3 g5e6 f4g3 c7c5 d3c4 b5c4 h2h4 d8a5 d1d2 a8a7 d2b2 a7a8 g1h2 h7h6 b2c2 e8e7 c2d2 a8d8 b1b2 e6f8 e1b1 d8e8 d4c5 a5c5');
INSERT INTO evaluations VALUES(29755,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(29755,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(29755,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(29755,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(29755,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(29755,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(29755,36,'cp 56','b8c6 d4d5 c6e5 c1e3 a7a6 h2h3 e5f3 g2f3 g4h5 f1c4 e7e5 d5e6 d8d1 a1d1 f7e6 e1e2 f8b4 e3d4 e8e7 h1g1 g7g6 a2a3 b4c3 d4c3 f6e4 c3h8 a8h8 d1d3 h8f8 c4a2 f8f6 b2b4 e4d6 g1e1 g6g5 e2f1 h5g6 d3e3 g6f5 f1g1 c7c6 a3a4 f6h6 e3e5 h6f6 a2b1 f5b1 e1b1 d6f5');
INSERT INTO evaluations VALUES(33977,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(33977,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(33977,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(33977,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(33977,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(33977,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(33977,36,'cp -2','d7d5 b1c3 e4e2 c3e2 h7h6 d2d4 c8f5 c1f4 b8d7 e1c1 f5h7 e2c3 e8c8 f1d3 h7d3 d1d3 c7c6 d3e3 d8e8 e3e8 f6e8 h1e1 e8d6 h2h4 d7f6 c3d1 f6h5 f4h2 g7g6 d1e3 h5g7 c1d1 f8e7 c2c3');
INSERT INTO evaluations VALUES(37452,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37452,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37452,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37452,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37452,36,'cp 0','d7d5 c2c4 c7c5 d4c5 d5c4 d3c4 d8c7 d1e2 f6e4 f3d4 b8c6 e2c2 c7e5 b1d2 e4d2 c1d2 c6d4 e3d4 e5d4 a1e1 d4b2 c2b2 g7b2 d2h6 f8e8 c4b5 e8d8 e1e7 b2f6 e7c7 f6e5 c7e7');
INSERT INTO evaluations VALUES(36776,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(36776,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36776,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36776,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36776,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36776,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36776,36,'cp 0','c2c4 d5c4 d3c4 c7c5 d4c5 b8d7 c5c6 d7b6 c6b7 c8b7 c4e2 f6e4 b1d2 e4d2 d1d2 d8d2 f3d2 b6a4 d2f3 b7f3 e2f3 a8b8 a1b1 a4c3 b1a1');
INSERT INTO evaluations VALUES(37606,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(37606,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(37606,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(37606,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(37606,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(37606,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(37606,36,'cp 164','d7d5 c1f4 g8e7 h2h4 b8d7 f1d3 e6e5 d4e5 d5d4 f3d4 d7e5 d3e2 f7f5 d1d2 f5e4 c3e4 e8g8 e1c1 e7f5 d4f5 d8d2 d1d2 c8f5 e4d6 f5e6 f4e3 b7b5 c4b5 c6b5 h4h5 e6a2 h5h6 g7f6 e2b5 a2b3 f2f4 e5f7 d6e4 f6e7 d2d7 a8c8 c1b1 b3c2 b1a1');
INSERT INTO evaluations VALUES(38021,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(38021,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(38021,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(38021,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(38021,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(38021,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(38021,36,'cp 194','c1f4 g8e7 h2h4 e8g8 h4h5 c6c5 d4c5 d5e4 c3e4 b8c6 h5g6 f7g6 d1d8 f8d8 f4c7 d8d7 c7d6 h7h6 g2g4 b7b6 g4g5 h6h5 e1c1 e7f5 f1d3 c6d4 f3d4 g7d4 d3c2 b6c5 d6c5 c8b7 f2f3 a7a5 h1e1 b7e4 f3e4 a8d8 e4f5 d4c5 f5e6');
INSERT INTO evaluations VALUES(42812,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(42812,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(42812,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(42812,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(42812,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(42812,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(42812,36,'cp -22','e8g8 d2d4 f6e4 d4e5 d7d6 f1d3 e4c5 e5d6 d8d6 e1g1 c5d3 d1d3 d6d3 c2d3 b8c6 c1f4 f8d8 d3d4 d8d7 f3e5 c6e5 d4e5 d7d8 f1d1 c8e6 f2f3 g8f8 h2h4 b7b6 g1f2 f8e8 h4h5 a7a5 h5h6 g7h6 f4h6 d8d1 a1d1');
INSERT INTO evaluations VALUES(43728,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(43728,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(43728,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(43728,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(43728,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(43728,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(43728,36,'cp 128','c3d5 e7e6 d5f6 d8f6 a2a4 b5a4 a1a4 f6d8 b2b4 f8e7 b4b5 e8g8 f1d3 c8b7 d1a1 b8d7 b5a6 d7c5 e1g1 e7f6 a4a5 b7e4 d3e4 c5e4 f2f3 e4c5 a6a7 c5b7 a5a6 b7c5 a6a2 d6d5 g1h1 e6e5 d4b5 d5d4 e3g1 d8d7 f1b1 c5e6 a2a5 f6e7 a1a4 e7c5 b1e1 c5b6 a5a6');
INSERT INTO evaluations VALUES(48048,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(48048,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(48048,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(48048,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(48048,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(48048,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(48048,36,'cp 115','b1c3 f8e7 g1f3 e8g8 f1d3 e6e5 h2h3 e5d4 f3d4 b8c6 c1e3 c6d4 e3d4 f6d7 e1g1 d7e5 d3e2 e5c6 d4e3 e7f6 a1c1 f8e8 f2f3 b7b6 d1d2 c8b7 c3d5 f6h4 c1d1 c6e5 e3f4 b7d5 c4d5 d8f6 f4e3 e5g6 e2b5 e8e7');
INSERT INTO evaluations VALUES(49884,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(49884,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(49884,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(49884,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(49884,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(49884,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(49884,36,'cp -115','e4c3 d2c3 f7f6 e1g1 b8c6 b2b4 a7a6 a2a4 d7d6 a4a5 d8e7 f3d4 e5d4 f1e1 c6e5 c3d4 c8e6 d4e5 d6e5 d1e2 e7d7 c1e3 f8d6 h2h3 d7f7 c4e6 f7e6 b4b5 e8g8 a1b1 a6b5 e2b5 f8b8 b5a4 e6f7 e1d1');
INSERT INTO evaluations VALUES(49895,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(49895,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(49895,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(49895,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(49895,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(49895,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(49895,36,'cp -75','d2c3 f7f6 e1g1 b8c6 b2b4 d7d6 a2a4 d8e7 f3d4 e5d4 f1e1 c6e5 c3d4 c8e6 c4f1 e7f7 d4e5 f6e5 c1g5 h7h6 g5h4 g7g5 h4g3 h6h5 h2h3 g5g4 h3h4 f8g7 f1b5 c7c6 b5f1 a8d8 b4b5 e8g8 d1d2 e6c4 f1c4 f7c4 b5c6 b7c6 d2e3 c4f7 a1b1');
INSERT INTO evaluations VALUES(49894,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(49894,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(49894,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(49894,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(49894,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(49894,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(49894,36,'cp -74','f7f6 e1g1 b8c6 b2b4 d7d6 a2a4 a7a6 a1b1 c6e7 f1e1 c7c6 b4b5 a6b5 a4b5 d6d5 c4b3 e7f5 b5c6 b7c6 c1b2 f8c5 c3c4 d5c4 b3c4 d8d1 b1d1 c5b4 e1e2 f5e7 h2h3 e7g6 e2e3 b4e7 e3c3 a8a7 c4b3 c6c5 d1e1 g6f4');
INSERT INTO evaluations VALUES(52954,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(52954,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(52954,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(52954,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(52954,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(52954,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(52954,36,'cp 85','d2d4 d6d5 e2e3 g8f6 g1f3 g7g6 h2h3 f8g7 f1d3 e8g8 e1g1 f8e8 d1b3 d8b6 c1d2 c8f5 d3f5 b6b3 a2b3 g6f5 f1c1 e7e6 c4d5 c6d5 c3b5 b8a6 d2e1 f6e4 b5c7 a6c7 c1c7 g7f8 f3e5 e8e7 c7c2 f7f6 e5d3 e4d6 e1b4 e7d7 a1c1 a7a5 b4e1 b7b6 c1a1 f8h6');
INSERT INTO evaluations VALUES(53535,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(53535,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(53535,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(53535,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(53535,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(53535,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(53535,36,'cp -32','b7b6 c5b6 a7b6 g1f3 g8f6 f1d3 f8b4 c1d2 b4d6 b1c3 c8b7 e1g1 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d7f6 e4d3 e8g8 d1e2 c6c5 d4c5 d6c5 f1d1 d8e7 d2e3 c5e3 e2e3 f8d8 d3c4 b7d5 c4d5 f6d5 e3e5 h7h6 a2a3 a8c8');
INSERT INTO evaluations VALUES(51689,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(51689,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(51689,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(51689,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(51689,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(51689,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(51689,36,'cp -36','c5b6 a7b6 g1f3 b8d7 f1d3 f8d6 d1c2 c8b7 b2b3 d8e7 e1g1 g8f6 c1b2 e8g8 a2a4 c6c5 f1e1 f8e8 h2h3 h7h6 b1d2 e8d8 a1c1 a8c8 c2d1 e6e5 d4e5 d7e5 f3e5 d6e5 b2e5 e7e5 d1c2 e5e7 e3e4 d5e4 d2e4');
INSERT INTO evaluations VALUES(4449,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(4449,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(4449,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(4449,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(4449,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(4449,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(4449,36,'cp 35','c1g5 h7h6 g5h4 d7d6 c4b3 d8e7 e1g1 g8h7 f3d2 g7g5 h4g3 c6a5 d3d4 f8g8 f1e1 c8g4 f2f3 g4e6 d4e5 d6e5 g3e5 e6b3 e5f6 e7f6 c2b3 f6c3 d2f1 a8d8 d1e2 a5c6 a1c1 c3b4 e2f2 b4b6 e1d1 d8d1 c1d1 g8d8 d1d8 b6f2 g1f2 c6d8 f1e3 h7g6 e3f5 d8c6 g2g3 a7a5 f3f4');
INSERT INTO evaluations VALUES(11735,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(11735,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(11735,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(11735,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(11735,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(11735,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(11735,36,'cp 203','f3e5 f8d6 d2d4 e8g8 f2f3 f8e8 e5d3 d6f8 c1e3 d7d5 e4e5 c8f5 b1c3 f6d7 f3f4 c6c5 d3c5 d7c5 d4c5 a8b8 b2b3 d5d4 d1d4 d8d4 e3d4 b8d8 d4f2 f5c2 a1c1 c2d3 f1d1 c7c6 g2g3 f7f6 e5f6 g7f6 d1d2 d3f5 d2d8 e8d8');
INSERT INTO evaluations VALUES(12409,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(12409,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(12409,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(12409,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(12409,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(12409,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(12409,36,'cp 118','d4d5 c6e7 b1c3 d7d6 d5e6 c8e6 f1e2 g7g6 e1g1 f8g7 e3e4 e8g8 c1f4 d8b6 d1c2 f6g4 a1d1 g4e5 b2b3 e5f3 e2f3 g7e5 f4h6 f8d8 f3e2 b6a5 h6d2 e7c6 f2f4 c6d4 c2d3 e5g7 f4f5 g6f5 e2h5 g8h8 c3d5 f5e4 d2a5');
INSERT INTO evaluations VALUES(12512,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(12512,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(12512,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(12512,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(12512,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(12512,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(12512,36,'cp 76','b5c6 d7c6 f3e5 f6e4 c3e4 d8d5 d1e2 d5e5 f2f4 e5e7 f4f5 g7g6 e1g1 c8f5 e4d6 e8d7 e2e7 f8e7 d6f5 g6f5 f1f5 d7e6 f5f3 a8e8 d2d3 e7f6 c2c3 f6e7 c1e3 f7f6 a1f1 h8f8 e3h6 f8g8 h2h3 e8d8 f1f2 d8d5 g1f1 b7b5 g2g4 d5e5 b2b3 a7a5 h6d2');
INSERT INTO evaluations VALUES(12291,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(12291,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(12291,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(12291,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(12291,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(12291,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(12291,36,'cp 62','d7c6 f3e5 f6e4 c3e4 d8d5 e1g1 d5e5 d2d4 c5d4 f1e1 f8e7 c1g5 e8g8 e4f6 e7f6 e1e5 f6e5 d1e2 e5c7 a2a4 h7h6 g5h4 c8f5 e2e7 c7d8 e7c5 d8h4 c5f5 h4f6 f5d7 b7b6 g2g3 c6c5 d7b5 f8d8 a4a5 a8c8 a5b6 a7b6 a1a4 d8d6 g1g2 g8f8 a4a6 c8b8 a6a7 f8g8');
INSERT INTO evaluations VALUES(13309,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(13309,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(13309,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(13309,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(13309,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(13309,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(13309,36,'cp -24','e7e5 f3d2 c8e6 d2c4 f8e7 c4e3 c6d4 a2a4 e8g8 e2f3 d8d7 e1g1 b7b6 c1d2 a7a6 c3d5 e6d5 e4d5 b6b5 g2g3 d4f3 d1f3 g7g6 b2b3 f6h5 a4b5 a6b5 b3b4 c5b4 d2b4 h5f6 g1g2 h7h5 h2h3 f6h7 g3g4 h7g5 f3e2 e7f6 a1a8 f8a8 f1b1');
INSERT INTO evaluations VALUES(13867,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(13867,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(13867,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(13867,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(13867,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(13867,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(13867,36,'cp -22','f3d2 f8e7 d2c4 e8g8 a2a4 c8e6 c4e3 c6d4 e2f3 d8d7 e1g1 b7b6 c3d5 e6d5 e4d5 d4f3 d1f3 g7g6 e3c4 h7h5 c1d2 h5h4 b2b3 f6h5 g2g4 h4g3 f2g3 f7f5 d2h6 h5g7 a1e1 a8e8 e1e2 e7f6 f1e1 g8h7');
INSERT INTO evaluations VALUES(14475,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(14475,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(14475,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(14475,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(14475,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14475,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(14475,36,'cp -9','e7e5 f4g3 g7g6 g1f3 f6e4 d4e5 e4g3 h2g3 c6e5 f3e5 d6e5 b1d2 f8g7 d1a4 c7c6 e1c1 d8e7 f1e2 a7a5 a4h4 h7h5 d2e4 e7h4 e4d6 e8e7 h1h4 f7f5 e3e4 f5f4 g3g4 g7f6 h4h1 h5h4 a2a4 h8d8 d6c8 d8c8 f2f3 c8d8 e2d3 e7f7 d3c4 f7g7 b2b3 f6e7 d1d8 a8d8');
INSERT INTO evaluations VALUES(15543,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15543,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15543,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(15543,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(15543,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15543,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(15543,36,'cp 0','f4g3 g7g6 g1f3 f6e4 d4e5 e4g3 h2g3 c6e5 f3e5 d6e5 b1d2 f8g7 d1a4 c7c6 e1c1 d8e7 f1e2 a7a5 a4h4 e7h4 h1h4 c8e6 e2c4 e6f5 c4e2');
INSERT INTO evaluations VALUES(15547,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15547,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15547,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(15547,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15547,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15547,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(15547,36,'cp -12','f6e4 g1f3 g7g6 d4e5 e4g3 h2g3 c6e5 f3e5 d6e5 b1d2 f8g7 d1a4 d8d7 f1b5 c7c6 b5e2 d7e7 e1c1 a7a5 a4h4 h7h5 d2e4 e7h4 e4d6 e8e7 g3h4 a5a4 a2a3 h8d8 d6c8 a8c8 e3e4 d8d1 c1d1 g7f8 g2g4 h5g4 e2g4 c8d8 d1c2 e7f6 h4h5 f6g5 g4e2 g6h5 h1h5 g5f4');
INSERT INTO evaluations VALUES(14692,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(14692,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(14692,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14692,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(14692,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(14692,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(14692,36,'cp 72','b7c6 f3e5 d5e4 d2d4 f8b4 e1g1 b4c3 b2c3 e8g8 f1e1 c6c5 c1a3 f8e8 a3c5 f6d7 e5d7 d8d7 d4d5 c8b7 c3c4 b7a6 d1d4 d7a4 e1e3 a4c4 e3c3 c4d4 c5d4 e8d8 c3c7 d8d5 d4a7 d5a5 a7e3');
INSERT INTO evaluations VALUES(21332,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(21332,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(21332,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(21332,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(21332,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(21332,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(21332,36,'cp 172','c4e2 b7b6 b2b4 a5b7 e2b5 c8d7 b5a6 a8b8 e1g1 g7g6 b1c3 f8g7 d1e2 g8e7 c3b5 e8g8 b5a7 g8h7 a2a4 f7f5 a1a3 d8e8 a6b5 b8a8 b5d7 e8d7 a7b5 c7c6 d5c6 d7c6 e4f5 e7f5 c1b2 b7d8 a3c3 c6d7 c3c7 d7e6 b5c3 a8c8 c7c8 e6c8 c3d5 d8e6 d5b6 e6f4');
INSERT INTO evaluations VALUES(22415,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(22415,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(22415,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(22415,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(22415,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(22415,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(22415,36,'cp 72','e7e6 g1f3 g8f6 a2a3 a7a6 c1f4 c7c6 e4e5 d6e5 f3e5 f6d5 c3d5 e6d5 f1d3 d7e5 f4e5 f8d6 e5g7 h8g8 d1e2 c8e6 g7e5 d6e5 e2e5 g8g2 e1c1 g2f2 d1f1 f2f1 h1f1 e8d7 e5f4 d8e7 d3h7 a8h8 h7f5 e6f5 f4f5 d7c7 f5f7 e7f7');
INSERT INTO evaluations VALUES(25895,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(25895,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(25895,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(25895,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(25895,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(25895,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(25895,36,'cp 75','e1g1 g8f6 d2d4 e7e6 c2c4 c6a5 b2b3 f8e7 b1c3 d7d5 c4d5 e6d5 c1f4 e8g8 a1c1 h7h6 f3e5 c7c5 h2h4 f8e8 e2e3 e7d6 f1e1 d8e7 e5d3 c5c4 b3c4 d5c4 f4d6 e7d6 g2b7 a5b7 d3f4 a8d8 d1f3 b7a5 g3g4 d6c6 d4d5 c6d7 g4g5 f6g4 c3e4 h6g5');
INSERT INTO evaluations VALUES(25894,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(25894,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(25894,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(25894,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(25894,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(25894,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(25894,36,'cp 94','g8f6 d2d4 e7e6 c2c4 c6a5 b2b3 f8e7 b1c3 d7d5 c4d5 e6d5 c1f4 e8g8 a1c1 h7h6 f3e5 c7c5 h2h4 a8c8 g2h3 c8a8 e2e3 c5c4 h3g2 a8c8 b3c4 a5c4 c3b5 c4e5 f4e5 c8c1 d1c1 a7a6 b5c3 b6b5 e5f6 e7f6 c1b1 d8d6 b1b3 f8d8 a2a3 b7c6 c3e2');
INSERT INTO evaluations VALUES(27838,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(27838,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(27838,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(27838,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(27838,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(27838,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(27838,36,'cp 36','h2h3 g4f3 d1f3 b8c6 f3d1 e7e6 c4b3 e6e5 d4e5 d6e5 c1g5 c6d4 c3d5 c7c6 d5f6 g7f6 g5h6 f8e8 c2c3 d4b3 d1b3 d8e7 f1d1 e8d8 h6e3 b7b6 b3c4 d8d1 a1d1 e7c7 a2a4 g8g7 b2b4 a8d8 d1d8 f6d8 g2g3 h7h6 g1g2 h6h5 c4a6 h5h4 g3g4 c7d7 g4g5 d8e7');
INSERT INTO evaluations VALUES(27837,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(27837,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27837,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(27837,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(27837,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(27837,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(27837,36,'cp 33','g4f3 d1f3 b8c6 c3e2 e7e6 c4b5 a7a6 b5c6 b7c6 c1g5 h7h6 g5f4 f6d7 a1d1 d8b8 b2b3 c6c5 d4c5 d7c5 f3e3 g6g5 f4h2 c5d7 g1h1 b8b6 e3d2 a6a5 f2f3 b6b4 c2c3 b4b7 d2c2 b7b5 e2d4 b5b7 a2a4');
INSERT INTO evaluations VALUES(27840,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(27840,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(27840,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(27840,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(27840,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(27840,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(27840,36,'cp 37','d1f3 b8c6 c1e3 e7e5 d4d5 c6d4 f3d1 f6d7 d1d2 d7b6 c4d3 f7f5 f2f3 c7c6 a2a4 a7a5 e3f2 g8h8 d5c6 b7c6 c3e2 c6c5 e2d4 e5d4 e4f5 g6f5 d3b5 b6d5 c2c3 f5f4 b5c6 d5e3 c6a8 d8a8 c3d4 g7d4 f1e1 d6d5 f2e3 f4e3');
INSERT INTO evaluations VALUES(27839,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(27839,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(27839,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(27839,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(27839,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(27839,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(27839,36,'cp 23','b8c6 c1e3 e7e5 d4d5 c6d4 f3d1 c7c5 f1e1 f6d7 c3b1 f7f5 c2c3 f5e4 b1d2 d4f5 d2e4 f5e3 e1e3 g7h6 e3f3 d7f6 c4d3 f6e4 f3f8 d8f8 d3e4 f8f4 d1e2 f4d2 e2d2 h6d2 a2a4 c5c4 g1f1 a8b8 a1a2 b7b6 g2g3 a7a5 h3h4 d2h6 a2a3 b8c8 f1e2 g8g7');
INSERT INTO evaluations VALUES(35988,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35988,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35988,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35988,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35988,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35988,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35988,36,'cp 0','f6g4 e3g5 h7h6 g5c1 b8c6 d4e2 g4f6 a2a4 g8h7 c1e3 f6g4 e3c1');
INSERT INTO evaluations VALUES(37984,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(37984,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(37984,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(37984,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(37984,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(37984,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(37984,36,'cp 79','e2e4 g7c3 b2c3 e7e5 a2a4 g8f6 f1d3 a7a5 g1e2 b8d7 h2h4 b7b6 e2g3 h7h5 g3f1 f6g8 f2f3 d8e7 d1b3 c8a6 c1e3 e8c8 f1d2 g8h6 d2b1 f7f6 b1a3 h6f7 a3b5 c8b8 e3f2 g6g5 e1c1 a6b5 b3b5 g5g4');
INSERT INTO evaluations VALUES(41021,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(41021,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(41021,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(41021,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(41021,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41021,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(41021,36,'cp 30','g1f3 b7b5 g5f6 e7f6 c3b5 c8b7 e4e5 f6e7 f1c4 e8g8 e1g1 b8d7 b2b3 d7b6 d1d3 a7a6 b5c3 c7c5 d4c5 e7c5 c3e4 b6c4 d3c4 c5e7 f1d1 d8a5 d1d7 a8c8 c4e2 c8c7 d7d4 b7d5 a1d1 a5b5 e2e3 c7c2 d4d2 c2d2 f3d2 f8d8 d1c1 d5b7');
INSERT INTO evaluations VALUES(46348,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46348,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46348,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(46348,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46348,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(46348,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(46348,36,'cp 0','f1d3 a7a6 g1f3 b7b5 d1e2 f8e7 e1g1 c8b7 c1d2 e8g8 h2h3 b8c6 f1c1 d8b6 d2e1 e7d6 c3b1 f8c8 a2a3 c6a5 c1c8 a8c8 b1d2 h7h6 b2b4 a5c4 d2b3 b7c6 b3c5 c6e8 e1d2 c4b2 d3b1 b2c4');
INSERT INTO evaluations VALUES(48156,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(48156,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(48156,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48156,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48156,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(48156,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(48156,36,'cp -11','b8d7 b2b3 f8b4 c2c3 b4d6 b1d2 e8g8 e1g1 c7c6 b5d3 e6e5 f3e5 d7e5 d4e5 d6e5 c1b2 c8g4 d3e2 g4f5 d2f3 e5d6 e2d3 f5g4 c3c4 d5c4 d3c4 d8e7 h2h3 g4h5 d1c2 h5f3 g2f3 d6e5 b2e5 e7e5 f1d1 h7h6 g1g2 a8d8 a1c1 f8e8 d1d8 e8d8 c1d1 d8d1 c2d1');
INSERT INTO evaluations VALUES(54399,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(54399,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(54399,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(54399,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(54399,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(54399,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(54399,36,'cp -12','d2d4 f8d6 g1f3 g8f6 e1g1 e8g8 b1c3 a7a6 f1e1 d8c7 c1g5 b8d7 a2a4 b7b6 g5h4 c8b7 h4g3 d6g3 h2g3 b7c6 c3e2 a6a5 e2f4 b6b5 c2c3 f8b8 f4d3 b8b7 d1e2 h7h6 a1a2 f6e4 e1a1 a8c8 b3c2');
INSERT INTO evaluations VALUES(54367,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(54367,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54367,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(54367,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54367,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(54367,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(54367,36,'cp 0','f8d6 g1f3 g8f6 e1g1 e8g8 b1c3 a7a6 f1e1 h7h6 c1d2 b7b5 a2a3 b8c6 d1c1 c6a5 d2f4 c8b7 c3e2 d6e7 b3a2 f6e4 e2g3 a5c4 a2c4 b5c4 f3e5 g8h7 f2f3 e4f6 c2c3 e7d6 c1b1 h7g8 b1c1 g8h7');
INSERT INTO evaluations VALUES(2645,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2645,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2645,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2645,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2645,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2645,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2645,36,'cp 0','d7d6 c3e2 c8e6 c4e6 f7e6 e2g3 a6a5 c2c3 d8e8 d3d4 c5b6 c1e3 e8g6 f3h4 g6f7 h4f3');
INSERT INTO evaluations VALUES(2683,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(2683,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(2683,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(2683,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(2683,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(2683,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(2683,36,'cp 118','a4c6 b7c6 f3e5 f8e8 e5f3 h7h6 b1c3 a6a5 c1f4 d7d5 e4e5 f6d7 d3d4 c5e7 f4e3 e7f8 f1e1 d7b6 c3e2 b6c4 e3c1 c8f5 e2f4 a5a4 f4d3 f7f6 b2b3 f6e5 d4e5 c4b6 f3d4 f5d3 d1d3 c6c5 d4f3 a4b3 c2b3 d8d7 c1d2 c5c4 b3c4 b6c4 d2c3');
INSERT INTO evaluations VALUES(3738,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3738,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(3738,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(3738,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(3738,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3738,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(3738,36,'cp 15','b1c3 d5f6 g5e4 f6e4 c3e4 c5e7 d1h5 c6d4 h5e5 d4c2 a1b1 g8h8 e5h5 c8e6 c4e6 f7e6 h5b5 b7b6 b5c4 c2d4 f1e1 d8d5 c1e3 d5c4 d3c4 d4c2 e1e2 c2e3 e2e3 a8d8 g2g3 h8g8 b1c1 f8f5 g1g2 g8f7 h2h4 h7h6 f2f3 d8d4 b2b3');
INSERT INTO evaluations VALUES(4108,30,'cp -298',NULL);
INSERT INTO evaluations VALUES(4108,31,'cp -283',NULL);
INSERT INTO evaluations VALUES(4108,32,'cp -278',NULL);
INSERT INTO evaluations VALUES(4108,33,'cp -272',NULL);
INSERT INTO evaluations VALUES(4108,34,'cp -273',NULL);
INSERT INTO evaluations VALUES(4108,35,'cp -273',NULL);
INSERT INTO evaluations VALUES(4108,36,'cp -275','f8f7 c4f7 g8f7 d2d3 f7g8 e1g1 h7h6 c3a4 c5e7 c2c4 d7d6 h2h3 d8e8 a4c3 f6h5 c3d5 e7d8 c1e3 e8g6 g1h1 c6e7 f1g1 e7d5 c4d5 a7a5 d3d4 e5d4 e3d4 c8d7 a1c1 a5a4 e4e5 d6e5 d4e5 h5f6 e5c7');
INSERT INTO evaluations VALUES(4646,30,'cp -254',NULL);
INSERT INTO evaluations VALUES(4646,31,'cp -246',NULL);
INSERT INTO evaluations VALUES(4646,32,'cp -273',NULL);
INSERT INTO evaluations VALUES(4646,33,'cp -250',NULL);
INSERT INTO evaluations VALUES(4646,34,'cp -282',NULL);
INSERT INTO evaluations VALUES(4646,35,'cp -262',NULL);
INSERT INTO evaluations VALUES(4646,36,'cp -272','c4f7 g8f7 d2d3 d7d5 c1g5 d5e4 g5f6 g7f6 c3e4 c5e7 e4g3 f7g8 c2c3 f6f5 e1g1 f5f4 g3e4 d8d5 b2b4 c8f5 d1f3 g8h8 f3e2 a8d8 a1d1 a7a5 b4a5 d5a5 d1a1 b7b5 a2a3 h8g7 g1h1 f5g6 f2f3 a5a6 f1d1 a6b6 a3a4 b5b4');
INSERT INTO evaluations VALUES(4637,30,'cp -271',NULL);
INSERT INTO evaluations VALUES(4637,31,'cp -282',NULL);
INSERT INTO evaluations VALUES(4637,32,'cp -272',NULL);
INSERT INTO evaluations VALUES(4637,33,'cp -277',NULL);
INSERT INTO evaluations VALUES(4637,34,'cp -268',NULL);
INSERT INTO evaluations VALUES(4637,35,'cp -273',NULL);
INSERT INTO evaluations VALUES(4637,36,'cp -271','g8f7 d2d3 d7d5 c1g5 d5e4 g5f6 g7f6 c3e4 c5e7 e4g3 f7g8 e1g1 f6f5 c2c3 f5f4 g3e4 d8d5 d1f3 g8h8 b2b4 c8f5 g1h1 a7a5 b4a5 c6a5 f1e1 f5g6 f3e2 b7b5 f2f3 a5c6 e1c1 h7h6 c1d1 a8a3 e2b2 h8g7 h2h3 a3a5 c3c4 d5d4 b2d4 c6d4');
INSERT INTO evaluations VALUES(11071,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(11071,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(11071,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11071,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(11071,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(11071,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(11071,36,'cp -24','e1g1 b7b5 c4b3 d7d5 c1g5 d5e4 c3e4 f8e7 d1e2 f6e4 g5e7 d8e7 e2e4 c8b7 e4e3 e8g8 c2c3 e7d6 b3c2 c6e7 d3d4 c5d4 e3d4 d6c7 d4h4 h7h6 h4g3 f8c8 a1d1 c7g3 h2g3 b7d5 c2b3 c8c7 f1e1 a8d8 b3d5 e7d5');
INSERT INTO evaluations VALUES(11070,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(11070,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(11070,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(11070,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(11070,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(11070,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(11070,36,'cp -27','d7d5 c4b3 b7b5 c1g5 d5e4 c3e4 f8e7 d1e1 f6e4 e1e4 c8b7 g5e7 d8e7 e4e3 c6a5 f3e5 e7d6 a1e1 a5b3 a2b3 e8g8 f2f4 a6a5 h2h3 f7f6 e5f3 f8e8 f1f2 a8d8 f2e2 h7h6 f3d2 d6b6 e3f2 d8d4 d2e4 b7e4 e2e4 d4e4');
INSERT INTO evaluations VALUES(12627,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(12627,31,'cp 246',NULL);
INSERT INTO evaluations VALUES(12627,32,'cp 265',NULL);
INSERT INTO evaluations VALUES(12627,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(12627,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(12627,35,'cp 245',NULL);
INSERT INTO evaluations VALUES(12627,36,'cp 258','e4e5 f6g8 c3e4 d8b6 e4d6 f8d6 d4d6 c6c5 f1d3 b6c6 c1f4 b7b6 e1c1 c8b7 h1e1 e8c8 f2f3 c6d6 e5d6 g8f6 d3c4 d8e8 f4e5 b7c6 c4f7 e8f8 f7c4 c8b7 c4d3 b6b5 b2b3 f8e8 e5g3 g7g6 c1d2 f6h5 g3f2 h5f4 f2c5 f4g2 e1e7 g2h4 d3e4 h4f3 e4f3 e8e7 d6e7');
INSERT INTO evaluations VALUES(13220,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(13220,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(13220,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(13220,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(13220,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(13220,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(13220,36,'cp 95','e4e5 d6e5 f3e5 c8d7 b5c6 d7c6 e5c6 b7c6 f1e1 e7e6 b2b3 f8e7 c1b2 e8g8 b1c3 f6d5 c3a4 d8c7 b2e5 c7c8 d1e2 f8d8 a1d1 e7f8 e2g4 d5b4 g4c4 c8a6 c4a6 b4a6 a2a3 c5c4 b3b4 f7f6 e5c3 g8f7 a4b2 c6c5 b4b5');
INSERT INTO evaluations VALUES(13213,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(13213,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(13213,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(13213,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(13213,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(13213,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(13213,36,'cp 97','d6e5 b5c6 b7c6 f3e5 f6d7 e5d7 d8d7 b2b3 e7e5 c1a3 f8d6 d1e2 e8g8 b1c3 f8e8 c3e4 d6c7 a1e1 f7f6 d2d3 d7d5 a3c5 f6f5 c2c4 d5d7 e4c3 c8a6 f2f3 a8d8 e1d1 d7f7 e2f2 a6c8 d3d4 f7g6 d4e5 c7e5 d1d8 e8d8');
INSERT INTO evaluations VALUES(13656,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(13656,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13656,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(13656,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(13656,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13656,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(13656,36,'cp 20','f8d6 f4g3 h7h6 f1d3 d6g3 h2g3 d8d6 a1c1 e8g8 d1e2 f8d8 e1g1 c8d7 f1d1 a7a6 a2a3 a8c8 c1c2 e6e5 d4e5 c6e5 e2d2 d7g4 d3e2 e5c6 d2e1 g4e6 f3d4 c6e5 d1c1 b7b5 e1d1 d6b6 d4e6 b6e6 e2f3 c8c4 c3e2 e5f3 g2f3');
INSERT INTO evaluations VALUES(8405,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(8405,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(8405,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(8405,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(8405,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8405,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(8405,36,'cp 32','f4g3 h7h6 f1d3 d6g3 h2g3 d8d6 e1g1 e8g8 a1c1 f8d8 d1e2 c8d7 a2a3 a7a6 c1c2 e6e5 d4e5 c6e5 e2d1 a8c8 f3e5 d6e5 d3e2 h6h5 c2d2 d7g4 e2f3 b7b5 f1e1 g7g6 d2d4 c8c4 a3a4 g8g7 a4b5 a6b5 c3b5 d8b8');
INSERT INTO evaluations VALUES(27940,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27940,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27940,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27940,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27940,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27940,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27940,36,'cp 0','d7d6 f1d3 b8d7 c1d2 d6e5 f3e5 d7e5 d4e5 d8h4 d3b5 e8f8 e1g1 g8h6 b5d3 h6f5 d1e2 f5d4 e2e3 a8d8 a1d1 h7h5 c3e4 b4e7 f1e1 d4f5 e3e2 f5d4');
INSERT INTO evaluations VALUES(34484,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34484,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(34484,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(34484,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(34484,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(34484,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(34484,36,'cp 23','g1f3 c5d4 f1c4 d5d7 e1g1 b8c6 d2b3 g8f6 b3d4 c6d4 f3d4 d7c7 d1e2 a7a6 c1g5 f8e7 c4b3 e8g8 f1e1 b7b5 c2c3 c7b6 a1d1 f8e8 d4f3 h7h6 g5f6 e7f6 f3e5 a8b8 e2h5 b6c7 b3c2 f6e5 h5e5 c7e5');
INSERT INTO evaluations VALUES(34483,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(34483,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(34483,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(34483,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(34483,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(34483,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(34483,36,'cp 20','c5d4 f1c4 d5d7 e1g1 b8c6 d2b3 a7a6 b3d4 c6d4 d1d4 d7d4 f3d4 c8d7 f1d1 a8c8 c4f1 g8f6 g2g3 f8c5 c2c3 f6g4 h2h3 c5d4 d1d4 g4e5 f1g2 d7c6 g2c6 e5c6 d4g4 g7g6 c1g5 c6e5 g5f6 e5g4 f6h8 g4h6 h8f6');
INSERT INTO evaluations VALUES(36549,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36549,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36549,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(36549,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36549,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(36549,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(36549,36,'cp 1','g1f3 f8e8 d3e2 e7e6 e1g1 c7c5 a2a3 b7b6 b2b4 c8b7 d1c2 b8c6 d4c5 d6c5 b4c5 b6c5 c1b2 d8e7 a1b1 a8b8 c3a4 b7a8 f3d2 f6d7 b2g7 b8b1 f1b1 g8g7 a4c3 e8d8 h2h3 c6a5 b1d1 d7b6 d2b3 d8d1 c3d1 a5b3 c2b3 a8c6 b3c2');
INSERT INTO evaluations VALUES(39915,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(39915,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39915,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(39915,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(39915,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39915,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(39915,36,'cp 18','g1e2 d7d5 a2a3 b4a5 d4c5 d5c4 c1d2 b7b6 e2g3 b6c5 f1c4 c8b7 e1g1 e8g8 d1c2 b8d7 a1d1 d8b6 c3e4 a8b8 h2h3 f8d8 d2a5 b6a5 e4f6 d7f6 d1d8 b8d8 f1d1 d8d1 c2d1 a5b6 g3h5 f6h5 d1h5 b6c6 c4f1 c6c7 h5d1 h7h6 d1c2 b7d5');
INSERT INTO evaluations VALUES(39914,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(39914,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(39914,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(39914,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(39914,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(39914,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(39914,36,'cp 16','d7d5 a2a3 b4a5 c1d2 c5d4 e2d4 e8g8 c4d5 f6d5 c3d5 a5d2 d1d2 d8d5 a1d1 c8d7 f1e2 b8c6 e1g1 c6d4 d2d4 d5d4 d1d4 d7c6 f1c1 f8d8 d4d8 a8d8 e2f3 c6f3 g2f3 a7a5 f3f4 d8b8 g1f1 b7b5 b2b4 a5b4 a3b4 g7g6 c1c7 g8g7 f1g2 h7h5');
INSERT INTO evaluations VALUES(43901,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(43901,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43901,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43901,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43901,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43901,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43901,36,'cp 0','c1b2 b5c4 b3c4 f8g7 g2g3 d7d6 g1f3 e8g8 f1g2 b8d7 e1g1 a8b8 b2c3 d7b6 b1d2 d8e8 a1b1 c8a6 d1c2 e8a4 c2b3 a4d7 b3c2');
INSERT INTO evaluations VALUES(43902,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43902,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43902,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43902,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43902,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43902,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43902,36,'cp -11','b5c4 e2e4 c4b3 e4e5 d8a5 b1c3 f6e4 g1e2 f8g7 a2b3 e4c3 e2c3 a5d8 c3e4 e8g8 d5d6 b8c6 f1c4 c6e5 d1d5 e5c6 b2g7 g8g7 d6e7 d8e7 e1g1 e7e5 e4c5 a7a5 f1d1 e5d5 c4d5 a8b8 d5c6 d7c6 a1a5 f8e8 f2f3 c8f5 g1f1 b8b5 a5b5 c6b5 g2g3 g7f6 f1f2');
INSERT INTO evaluations VALUES(48732,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(48732,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(48732,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(48732,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(48732,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(48732,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(48732,36,'cp -66','d5d4 c3e4 f6e4 g2e4 b8a6 e4g2 f8e7 g1f3 e8g8 e1g1 c7c6 e2e3 e7f6 f3d4 f6d4 e3d4 d8d4 d2d3 c8e6 f1e1 d4d7 b2b3 f8d8 c1b2 h7h6 d1d2 a6c7 d2b4 c7b5 e1e3 a7a5 b4c5 b5d4 a1e1 a5a4 c5c3 a4b3 a2b3 a8a2 c3d4 d7d4 b2d4');
INSERT INTO evaluations VALUES(50007,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(50007,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(50007,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(50007,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(50007,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(50007,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(50007,36,'cp 125','d7d5 g2g3 f8e7 f1g2 c7c5 c1f4 e8g8 e1g1 b8c6 e2e3 h7h6 f3e5 d8e8 c2c4 c6e5 f4e5 d5c4 d1e2 c8e6 g2b7 a8d8 b7g2 d8d3 e3e4 f6g4 e5c3 e8g6 b1a3 h6h5 h2h3 g4f2 f1f2 f8f2 e2f2 d3g3 c3e5 g3h3 a1f1 e7h4');
INSERT INTO evaluations VALUES(51391,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(51391,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(51391,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(51391,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(51391,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(51391,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(51391,36,'cp 31','d2d4 c8b7 b1c3 d7d6 g1f3 a7a6 f1d3 d6e5 f3e5 g8f6 d3e2 f8d6 e2f3 d8c8 e1g1 e8g8 c1f4 b6b5 f4g3 b8d7 a2a4 b5b4 e5d7 f6d7 c3e4 d6g3 h2g3 b7d5 a4a5 c7c5 d4c5 d7c5 e4d6 c8c6 f3d5 c6d5 d1d5 e6d5');
INSERT INTO evaluations VALUES(53031,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(53031,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(53031,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(53031,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(53031,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(53031,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(53031,36,'cp 25','f1g2 g7g6 e2e3 f8g7 b1c3 b8c6 g1e2 e7e5 a2a3 g8e7 e1g1 a7a5 a1b1 e8g8 d2d3 a8b8 c3d5 h7h5 d5e7 d8e7 e2c3 c8e6 h2h3 e7d7 g1h2 f7f5 f2f4 c6e7 c1d2 b7b6 b2b4 a5b4 a3b4 e5e4 d1c2 c5b4 b1b4');
INSERT INTO evaluations VALUES(53032,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(53032,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(53032,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(53032,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(53032,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(53032,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53032,36,'cp 21','b8c6 g1f3 e7e5 e1g1 g7g6 a2a3 f8g7 b1c3 a7a5 d2d3 g8e7 c1g5 f7f6 g5d2 e8g8 f3e1 c8e6 c3d5 a5a4 e2e3 b7b5 d5e7 c6e7 g2a8 d8a8 c4b5 f8b8 e3e4 b8b5 d2c3 d6d5 a1c1 d5d4 c3d2 e6h3 c1c4 h3f1 g1f1 a8c8 f1g2 b5b2 c4a4 g7f8 e1f3 h7h6');
INSERT INTO evaluations VALUES(19984,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(19984,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19984,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(19984,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(19984,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(19984,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(19984,36,'cp 32','g1f3 e7e5 e1g1 g7g6 a2a3 f8g7 b1c3 a7a5 d2d3 g8e7 f3e1 c8e6 c1g5 e8g8 g5e7 d8e7 e1c2 f7f5 a1b1 g8h8 c2e3 g6g5 b2b4 c5b4 e3d5 e7f7 a3b4 a5b4 d5b4 e5e4 b4d5 a8d8 d1d2 h7h6 c3b5 e4d3 f2f4 c6e7 d5e7 f7e7 e2d3');
INSERT INTO evaluations VALUES(19983,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(19983,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(19983,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(19983,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19983,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(19983,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19983,36,'cp 25','e7e5 b1c3 g7g6 a2a3 f8g7 e1g1 g8e7 a1b1 a7a5 d2d3 e8g8 c1g5 c8e6 g5e7 d8e7 f3e1 g8h8 e1c2 f7f5 c2e3 g6g5 e3d5 e7f7 b2b4 a5b4 a3b4 c5b4 d5b4 e5e4 c3b5 a8d8 d3e4 e6c4 b4c6 b7c6 b5d4 f5e4 d4c6 d8e8 c6a5 c4a6');
INSERT INTO evaluations VALUES(53976,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(53976,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53976,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53976,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53976,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(53976,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(53976,36,'cp 0','g1f3 g8f6 e1g1 f8e7 a2a4 b8c6 c4b3 c6a5 b3a2 a7a6 c1f4 d8b6 b1d2 b6b2 d2b3 a5c6 f4c1 b2c3 c1d2 c3b2');
INSERT INTO evaluations VALUES(56911,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(56911,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(56911,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(56911,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(56911,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(56911,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(56911,36,'cp 94','f8g7 d2d4 e5d4 f3d4 g8e7 d1d3 e8g8 c1f4 b8c6 d4c6 e7c6 e1c1 d7d6 d3e3 c8e6 f2f3 d8e7 h2h4 h7h5 f4h6 g7h6 e3h6 e7f6 h6e3 a7a6 f1e2 a8e8 f3f4 b7b5 g2g4 h5g4 h4h5 g6g5 f4g5 f6e5 g5g6 f7g6 h5g6');
INSERT INTO evaluations VALUES(57443,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57443,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(57443,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(57443,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(57443,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57443,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57443,36,'cp 0','b8c6 g1f3 d7d5 f1e2 c8e6 e1g1 d8d7 d2d4 e5e4 f3d2 f6f5 c2c4 g8f6 a2a3 f8e7 f2f3 c6a5 b2c3 a5c6');
INSERT INTO evaluations VALUES(23245,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23245,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23245,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23245,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23245,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23245,36,'cp 12','g1f3 d7d5 f1e2 c8e6 e1g1 d8d7 c2c4 d5d4 e3d4 e5d4 d2d3 g8h6 b1a3 h6f5 a3c2 h8g8 f1e1 e8c8 e2f1 e6f7 e1e4 f5d6 e4h4 d6f5 h4h7 f7g6 h7h3 g6e8 b3b4 g7g5 b4b5 c6a5 f3d4 g5g4 d4f5 g4h3 d1f3 h3g2 f1g2 d7d3');
INSERT INTO evaluations VALUES(4301,30,'cp -355',NULL);
INSERT INTO evaluations VALUES(4301,31,'cp -350',NULL);
INSERT INTO evaluations VALUES(4301,32,'cp -334',NULL);
INSERT INTO evaluations VALUES(4301,33,'cp -325',NULL);
INSERT INTO evaluations VALUES(4301,34,'cp -329',NULL);
INSERT INTO evaluations VALUES(4301,35,'cp -324',NULL);
INSERT INTO evaluations VALUES(4301,36,'cp -334','h7h6 g5f7 f8f7 c4f7 g8f7 c1e3 d7d6 f3d1 f7g8 b1c3 c8d7 e1g1 c6d4 f2f3 c5b6 g1h1 d4e6 d1d2 g8h7 c3e2 f6h5 c2c3 b6e3 d2e3 a7a5 a1d1 d8e7 d3d4 h5f4 e2f4 e6f4 d1d2 f4g6 f1d1 a5a4 d4e5 g6e5 b2b3 h7g8 c3c4 a4b3 a2b3 e5g6 e3d4 d7c6');
INSERT INTO evaluations VALUES(4150,30,'cp -323',NULL);
INSERT INTO evaluations VALUES(4150,31,'cp -318',NULL);
INSERT INTO evaluations VALUES(4150,32,'cp -319',NULL);
INSERT INTO evaluations VALUES(4150,33,'cp -325',NULL);
INSERT INTO evaluations VALUES(4150,34,'cp -310',NULL);
INSERT INTO evaluations VALUES(4150,35,'cp -331',NULL);
INSERT INTO evaluations VALUES(4150,36,'cp -316','g5f7 f8f7 c4f7 g8f7 c1e3 d7d6 f3d1 f7g8 b1c3 c8d7 d1d2 c6d4 e1g1 d8e7 f2f3 a8f8 g1h1 g8h7 a1e1 e7f7 e3g1 d7c6 c3e2 d4e2 g1c5 d6c5 e1e2 f6h5 g2g3 b7b6 b2b3 h7h8 a2a4 a7a5 d2c3 f7e7 c3d2 c6d7');
INSERT INTO evaluations VALUES(5410,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(5410,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(5410,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(5410,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(5410,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(5410,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(5410,36,'cp 62','g8f6 d4d5 f6e4 d5c6 g7b2 b1d2 e4c3 d1c2 b2a1 c2c1 c3d5 c1a1 f7f6 e3h6 d7c6 f1c4 b7b5 c4b3 c6c5 e1g1 c5c4 b3c2 c8b7 f1e1 a7a6 d2e4 d8b6 a2a4 e8c8 a4b5 a6b5 e1b1 g6g5 h2h4 g5h4 e4c3 b5b4 c3d5 d8d5 h6e3 b6a5 a1a5 d5a5 b1b4 h4h3 b4c4 c8b8 c2e4 h3g2 e4b7 b8b7 g1g2');
INSERT INTO evaluations VALUES(11320,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(11320,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(11320,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(11320,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(11320,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(11320,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(11320,36,'cp 25','c6b4 c2c4 b4d3 d1d3 e7e6 e1g1 d5c4 d3c4 f8d6 f4g5 e8g8 b1d2 c8d7 f1e1 h7h6 g5h4 d6e7 a2a3 d7b5 c4c2 a8c8 b2b4 b7b6 h4f6 e7f6 d2e4 f6e7 h2h3 a6a5 b4a5 b5a6 f3e5 c7c5 a5b6 c5d4 c2b3 d8d5 b3d5 e6d5');
INSERT INTO evaluations VALUES(11482,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(11482,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(11482,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(11482,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11482,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(11482,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(11482,36,'cp 56','d2d4 f6e4 f3e5 c6e5 d4e5 e4c3 b2c3 d7d6 e1g1 f8e7 d1h5 d6d5 c4d3 d8d7 f2f4 g7g6 h5f3 d7c6 c1e3 c6c3 f3d5 c7c6 d5d4 c3d4 e3d4 c6c5 d4e3 b7b5 c2c4 b5c4 d3c4 c8f5 c4e2 e8g8 g2g4 f5e4 e2c4 f8d8 a1d1 a8b8 d1d8 b8d8 g1f2 h6h5 g4h5 g6h5 f2e2 d8b8 f4f5 b8b2 e3d2 e4c2 c4a6');
INSERT INTO evaluations VALUES(17038,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(17038,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(17038,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(17038,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(17038,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(17038,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(17038,36,'cp 25','f8d6 b1c3 d8e7 d4e3 e8g8 c1d2 d6c5 e3g3 d7d5 d2g5 e7e6 g5f6 e6f6 c3d5 f6b2 e1g1 c5d6 a1b1 b2d4 c2c3 d4d5 e4d5 d6g3 h2g3 b7b6 c3c4 f8e8 a2a4 c8d7 a4a5 a8b8 b1a1 c7c6 a5b6 a7b6 a1a7 e8d8 f1e1 c6d5 c4d5 g8f8 d3e4 b6b5 f2f3');
INSERT INTO evaluations VALUES(19604,30,'cp 338',NULL);
INSERT INTO evaluations VALUES(19604,31,'cp 327',NULL);
INSERT INTO evaluations VALUES(19604,32,'cp 333',NULL);
INSERT INTO evaluations VALUES(19604,33,'cp 354',NULL);
INSERT INTO evaluations VALUES(19604,34,'cp 335',NULL);
INSERT INTO evaluations VALUES(19604,35,'cp 348',NULL);
INSERT INTO evaluations VALUES(19604,36,'cp 341','f2f4 e5g6 h2h4 h7h5 g1f3 d7d6 f1d3 d8c7 e4e5 h8h6 d3b5 c8d7 d1e2 d7b5 e2b5 c7d7 b1c3 e8c8 b5a5 b7b6 a5a6 d7b7 a6b7 c8b7 f3g5 g6h8 e5e6 f7e6 g5e6 d8e8 f4f5 h6f6 e1g1 g7g6 c1g5 f6f5 f1f5 g6f5 a1f1 f8h6 f1f5 b7c8 c3b5 c8b8 g2g3 a7a6 b5c3 h8g6 g5h6 g8h6 f5h5 h6g4 h5h7');
INSERT INTO evaluations VALUES(19606,30,'cp 332',NULL);
INSERT INTO evaluations VALUES(19606,31,'cp 337',NULL);
INSERT INTO evaluations VALUES(19606,32,'cp 350',NULL);
INSERT INTO evaluations VALUES(19606,33,'cp 352',NULL);
INSERT INTO evaluations VALUES(19606,34,'cp 332',NULL);
INSERT INTO evaluations VALUES(19606,35,'cp 356',NULL);
INSERT INTO evaluations VALUES(19606,36,'cp 330','e5g6 h2h4 h7h5 g1f3 d7d6 f1d3 d8c7 c2c4 g8h6 f4f5 g6e5 f3e5 d6e5 d1h5 g7g6 f5g6 c7d6 b1c3 c8d7 d3e2 d6g6 h5g6 f7g6 a2a4 h6f7 c1e3 a8c8 e1d2 f8h6 e3h6 h8h6 a4a5 e8f8 d2c2 h6h8 a1e1 c8e8 g2g4 f8g7 g4g5');
INSERT INTO evaluations VALUES(19576,30,'cp 360',NULL);
INSERT INTO evaluations VALUES(19576,31,'cp 325',NULL);
INSERT INTO evaluations VALUES(19576,32,'cp 366',NULL);
INSERT INTO evaluations VALUES(19576,33,'cp 357',NULL);
INSERT INTO evaluations VALUES(19576,34,'cp 338',NULL);
INSERT INTO evaluations VALUES(19576,35,'cp 336',NULL);
INSERT INTO evaluations VALUES(19576,36,'cp 344','h2h4 h7h5 g1f3 d7d6 f1d3 d8c7 c2c4 g8h6 f4f5 g6e5 f3e5 d6e5 d1h5 g7g6 f5g6 c7d6 b1c3 c8d7 d3e2 d6g6 h5g6 f7g6 c1e3 b7b6 a2a4 a7a5 e1f2 h6f7 g2g4 f8g7 a1f1 f7d6 f2g3 h8h7 g4g5 e8c8 b2b3 d8f8 f1f8 g7f8 h4h5 g6h5 h1h5 h7h5 e2h5 f8g7 g3g2');
INSERT INTO evaluations VALUES(26734,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(26734,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(26734,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(26734,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(26734,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(26734,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(26734,36,'cp 71','h2h3 g4d7 a2a3 f8e7 c1e3 g8f6 c4a2 a7a5 d1d2 d7e6 c3d5 b7b5 a1c1 e8g8 c2c4 b5c4 a2c4 e6d5 e4d5 c6b8 f3h4 f6h7 h4f5 e7g5 e1g1 g5e3 f2e3 h6h5 c4b5 g7g6 f5g3 b8d7 b5c6 a8a7 b2b4 d7f6 b4b5 g8g7 d3d4');
INSERT INTO evaluations VALUES(26732,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(26732,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(26732,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(26732,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(26732,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(26732,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(26732,36,'cp 64','g4d7 a2a3 f8e7 c1e3 g8f6 g2g4 f6h7 d1e2 e7g5 f3g5 h7g5 c3d5 g5e6 c2c3 c6e7 h3h4 c7c6 d5e7 d8e7 e1c1 c6c5 c1b1 d7c6 f2f3 e8c8 d3d4 c5d4 c3d4 e6d4 e2f2 d6d5 e4d5 c6d5 c4d5 d8d5 e3d4 e5d4');
INSERT INTO evaluations VALUES(33025,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(33025,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(33025,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(33025,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(33025,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(33025,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(33025,36,'cp 95','d3d4 c5b6 a2a4 g8e7 c1e3 c8g4 d4e5 d6e5 a4a5 b6c7 h2h3 g4e6 f3d2 b8d7 d1e2 e7c8 c4e6 f6e6 d2c4 c8d6 c4d6 c7d6 f1d1 d7f6 f2f3 e8g8 e2f2 a7a6 c3a4 a8d8 d1d3 d8d7 e3c5 d6e7 c5e7 e6e7 f2c5 e7c5 a4c5 d7d3 c5d3 f6d7 a1d1 f8d8');
INSERT INTO evaluations VALUES(34035,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34035,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34035,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34035,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34035,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34035,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34035,36,'cp 0','h5e2 d7d6 d2d3 g7g6 g1f3 f8g7 a2a4 e8g8 e1g1 f8e8 f3g5 e8f8');
INSERT INTO evaluations VALUES(34086,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(34086,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(34086,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(34086,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(34086,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(34086,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(34086,36,'cp 119','f1d3 c8g4 h2h3 g4h5 g2g4 h5g6 d3g6 h7g6 f3g5 b8d7 d1f3 e7e6 e1c1 d7b6 c1b1 e8c8 d2c1 d8d7 h3h4 b6c4 f3e2 b7b5 c3e4 f6e4 g5e4 c4b6 e2f3 b6d5 c2c3 a7a5 h4h5 g6h5 g4h5 c8b7 d1g1 d7d8 a2a3 d8c8');
INSERT INTO evaluations VALUES(34087,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(34087,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(34087,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(34087,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(34087,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(34087,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(34087,36,'cp 116','c8g4 h2h3 g4h5 g2g4 h5g6 d3g6 h7g6 f3g5 e7e6 d1f3 b8d7 e1c1 d7b6 c1b1 e8c8 d2c1 d8d7 c3e4 f6e4 g5e4 b6c4 f3e2 b7b5 h3h4 c4b6 e2f3 c8b7 c1f4 c7d8 f4g3 a7a5 h1e1 b6d5 a2a3 d8b6 b1a1 b5b4 c2c4 b4c3 e4c3 b6b3 d1d3 d5c3 d3c3');
INSERT INTO evaluations VALUES(32718,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(32718,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(32718,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(32718,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(32718,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(32718,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(32718,36,'cp 117','h2h3 g4h5 g2g4 h5g6 d3g6 h7g6 f3g5 b8d7 d1f3 e7e6 e1c1 d7b6 d2f4 c7d7 f4e5 f6d5 h1e1 d5c3 f3c3 f8e7 g5e4 f7f6 e5g3 d7d5 b2b3 h8h3 e4c5 d5g5 c1b2 g5g4 c5b7 e8f7 c3c6 a8c8 b7d6 e7d6 c6d6 h3h5 c2c4 c8c4 d4d5 c4e4 e1e4 g4e4 d6c7 f7g8 d5e6 e4e6 c7a7 e6c6');
INSERT INTO evaluations VALUES(34584,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34584,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(34584,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(34584,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(34584,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(34584,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(34584,36,'cp 8','b1c3 d5e5 c1e3 g8f6 d1d3 c8d7 e1c1 e7e6 f2f4 e5a5 a2a3 b8c6 c3e4 a8d8 d3b3 f8e7 e4f6 e7f6 c1b1 c6e7 b3b7 e7d5 b7a7 a5a7 e3a7 d7c6 g5e4 f6e7 h1g1 e8g8 g2g3 d5f4 f1c4 c6e4 g3f4 g7g6 g1e1 e4c6 a7e3 d8b8 e3d4 f8d8 a3a4 d8d4 d1d4 c6a4');
INSERT INTO evaluations VALUES(34943,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34943,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34943,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34943,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34943,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34943,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34943,36,'cp 0','b8c6 b1c3 d5d6 f1e2 c8e6 e1g1 e8c8 f1e1 g8f6 a1b1 h7h6 e2f1 g7g5 b2b4 e6g4 b4b5 c6d4 f1e2 g4f3 e2f3 g5g4 f3g4 f6g4 d1g4 f7f5 g4h3 d6e6 c3e2 d4c2 e2f4 e6f6 f4h5 f6e6');
INSERT INTO evaluations VALUES(39174,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(39174,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(39174,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(39174,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(39174,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(39174,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(39174,36,'cp 69','b2c3 g8f6 e2e3 e8g8 f1d3 c7c5 g1e2 d7d6 e3e4 b8c6 e1g1 e6e5 h2h3 c5d4 c3d4 e5d4 c1b2 f8e8 f1e1 c8e6 e2d4 c6d4 b2d4 a8c8 a1c1 f6d7 d3f1 f7f6 d1b3 b7b6 g1h1 e6f7 b3a3 d8c7 d4a1 d7c5 f2f3 h7h6 a1d4 f7e6 a3e3 c7d7');
INSERT INTO evaluations VALUES(39173,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(39173,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(39173,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(39173,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(39173,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(39173,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(39173,36,'cp 62','g8f6 d1c2 c7c5 e2e4 d7d5 e4e5 f6e4 g1e2 c5d4 e2d4 b8c6 c4d5 e6d5 f1d3 e4c5 d4c6 b7c6 c1a3 d8a5 a3c5 a5c5 e1g1 g7g6 h2h4 e8g8 h4h5 c5e7 f2f4 c8g4 h5g6 f7g6 c2f2 a8b8 a1b1 b8b1 f1b1 g4e6 a2a4 f8f7 a4a5 g8g7 b1e1 g7h8 a5a6 h8g7');
INSERT INTO evaluations VALUES(42497,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(42497,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(42497,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(42497,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(42497,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(42497,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(42497,36,'cp 86','d1c2 d7d5 c4d5 d8d5 f2f3 d5a5 e1f2 c7c5 e2e4 b8c6 c1e3 b7b6 a2a4 c8a6 f1a6 a5a6 g1e2 e8g8 c2b2 f8c8 b2b5 a6a5 h1b1 f6d7 h2h3 a5b5 a4b5 c6a5 e3f4 f7f6 h3h4 h7h6 h4h5 a5c4 f4e3 e6e5 d4d5 c4d6 g2g4');
INSERT INTO evaluations VALUES(43116,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(43116,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(43116,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(43116,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(43116,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(43116,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(43116,36,'cp -67','d7d6 b5c3 e8g8 g1f3 c7c5 d4c5 d6d5 e2e3 b8d7 f1e2 d7c5 f4e5 b7b6 a2a4 c8b7 a4a5 b6a5 e1g1 c5d7 e5g3 a8c8 f3d4 d7c5 g3e5 f6e4 e5g7 e4c3 b2c3 g8g7 d4b3 c5b3 c2b3 e7e5 c3c4 c8c5 e2f3 d8a8 f3d5 b7d5 c4d5 a8d5 d1d5 c5d5');
INSERT INTO evaluations VALUES(41184,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(41184,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(41184,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(41184,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(41184,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(41184,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(41184,36,'cp -69','b5c3 e8g8 g1f3 c7c5 d4c5 d6d5 e2e3 b8d7 f1e2 d7c5 f4e5 b7b6 a2a4 c5d7 e5g3 c8b7 a4a5 b6a5 e1g1 f6e4 c3e4 d5e4 f3d4 d7c5 g3h4 a5a4 e2b5 d8c7 b5a4 f8d8 d1d2 a8c8 h2h3 d8d6 d2c3 a7a6 h4g3 e7e5 b2b4 c5a4 c3c7 c8c7 a1a4 e5d4 g3d6 c7c2 e3d4 g7d4 d6f4 e4e3 f4e3 d4e3 f2e3 c2g2 g1h1');
INSERT INTO evaluations VALUES(50178,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(50178,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(50178,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(50178,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50178,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(50178,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(50178,36,'cp 25','d2d4 f8g7 g1f3 e8g8 f1e2 c7c5 d4d5 e7e6 b1c3 d7d6 f3d2 f8e8 e3e4 b8a6 e1g1 b7b6 d5e6 e8e6 f2f3 c8b7 d2b3 a6c7 c1g5 h7h6 g5h4 g6g5 h4f2 f6h5 f1e1 g7c3 b2c3 h5f4 e2f1 d8f6 d1d2 a8d8 a1d1');
INSERT INTO evaluations VALUES(53091,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53091,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53091,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53091,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53091,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53091,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53091,36,'cp 0','g8f6 f1e2 b8c6 g1f3 e7e6 a2a4 f8e7 c2c3 a7a6 e1g1 e8g8 c1d2 d8b6 b1a3 d6d5 e4e5 f6d7 g1h1 f7f6 d3d4 b6b2 a1b1 b2a3 b1a1 a3b2 a1b1 b2a2 b1a1');
INSERT INTO evaluations VALUES(55955,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(55955,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(55955,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(55955,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(55955,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55955,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(55955,36,'cp -33','b8d7 c1f4 c7c5 c2c3 d8b6 d1b3 g8f6 b1a3 a7a6 b3b6 d7b6 a3b1 c5d4 c3d4 f6e4 f2f3 e4d6 h2h4 f7f6 e5d3 c8d7 b1c3 a8c8 a1c1 d6f5 e2e3 b6c4 e1f2 c4b2 d3b2 f8a3 c1c2 a3b4 f1d3 c8c3 d3f5 e6f5');
INSERT INTO evaluations VALUES(56981,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(56981,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(56981,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(56981,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(56981,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(56981,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(56981,36,'cp 12','b8c6 d3d4 e5d4 f3d4 f8c5 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 c6d5 c1f4 c7c6 b1d2 g8f6 e2e7 e8e7 d2f3 h8e8 e1g1 e7f8 f1e1 e8e1 a1e1 c8e6 c2c3 g7g5 f4g3 f6e4 d3e4 d5e4 e1e4 e6a2 e4a4 a2e6');
INSERT INTO evaluations VALUES(57491,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(57491,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(57491,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(57491,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(57491,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(57491,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(57491,36,'cp 95','f5e4 f3e5 d8g5 d2d4 g5g2 d1h5 g7g6 c4f7 e8d8 f7g6 g2h1 e1e2 h1c1 e5f7 d8e8 f7h8 h7g6 h5g6 e8d8 h8f7 d8e7 b1c3 c1c2 e2e1 d7d6 c3d5 e7d7 g6g8 e4e3 f2e3 f8e7 f7g5 b8a6 g8e6 d7c6 d5e7 c6b6 e6b3 c2b3 a2b3 a6b4 e7c8 a8c8 e1f2 c8f8 f2e2 f8e8 a1c1 b4d5 e3e4 d5e7 c1h1 d6d5 e4e5 e7f5 e2d3 e8g8 h1g1 b6b5 d3c3 b7b6 b3b4 f5e7 h2h4 g8h8 g1h1 e7f5 h1f1 f5h4 f1f7 c7c6 e5e6 h4g6 f7a7');
INSERT INTO evaluations VALUES(57470,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(57470,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(57470,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(57470,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(57470,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(57470,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(57470,36,'cp 102','f3e5 d8g5 d2d4 g5g2 d1h5 g7g6 c4f7 e8d8 f7g6 g2h1 e1e2 h1c1 e5f7 d8e8 f7h8 h7g6 h5g6 e8d8 h8f7 d8e7 b1c3 c1c2 e2e1 d7d6 c3d5 e7d7 g6g8 e4e3 f2e3 f8e7 f7g5 b8a6 g8e6 d7c6 d5e7 c6b6 e6b3 c2b3 a2b3 a6b4 e7c8 a8c8 e1e2 c8e8 a1c1 b4d5 e3e4 d5e7 e2e3 d6d5 c1h1 d5e4 e3e4 e7c6 e4d3 c6b4 d3d2 a7a5 h2h4 e8g8 g5e6 g8g2 d2e3 g2b2 e3e4 b2e2 e4f5 b4c6');
INSERT INTO evaluations VALUES(57460,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(57460,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(57460,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(57460,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(57460,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(57460,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(57460,36,'cp 72','d7d5 d1h5 g7g6 e5g6 h7g6 h5h8 e8f7 h8d4 g8f6 c4e2 b8c6 d4e3 c8f5 b1a3 d8d7 e1g1 a8e8 b2b4 f8b4 c2c3 b4e7 a3b5 a7a6 b5d4 c6e5 e3h6 e7d6 h6h4 c7c5 d4f5 g6f5 d2d4 e4d3 e2h5 f6h5 h4h5 f7f6 c1g5 f6e6');
INSERT INTO evaluations VALUES(12040,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12040,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12040,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12040,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12040,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12040,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12040,36,'cp 0','b1c3 e7e6 c1f4 d6d5 c4d3 f8e7 e1g1 e8g8 f1e1 a7a5 e4e5 f6d7 c3a4 c8a6 c2c4 d5c4 d3h7 g8h7 d1h5 h7g8 e1e3 f7f6 e3h3 f6e5 h5h8 g8f7 h8h5');
INSERT INTO evaluations VALUES(12711,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(12711,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(12711,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(12711,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12711,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(12711,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(12711,36,'cp 46','d2d4 c5d4 f3d4 d8a5 e1g1 f8g7 d4b3 a5c7 c1g5 e8g8 d1d2 d7d6 g5f6 e7f6 c3b5 c7e7 f1e1 f8d8 b5c3 c8e6 c3d5 e7d7 c2c3 f6f5 e4f5 e6f5 d5e3 f5e6 c4d5 d8e8 a1d1 e6d5 d2d5 a8d8 h2h3 c6e7 d5d2 d6d5');
INSERT INTO evaluations VALUES(16205,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(16205,31,'cp 265',NULL);
INSERT INTO evaluations VALUES(16205,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(16205,33,'cp 273',NULL);
INSERT INTO evaluations VALUES(16205,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(16205,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(16205,36,'cp 280','c2c3 a7a6 b5a4 d7d6 d2d4 c8d7 d4d5 c6e7 a4d7 d8d7 h3g4 h5g4 f3h2 g7g6 h2g4 e7g8 f2f3 f7f5 g4f2 g8f6 c1g5 f6h5 b1d2 f8e7 f3f4 e8c8 g5e7 d7e7 f4e5 d6e5 f1e1 h5g3 d2f1 g3e4 e1e4 f5e4 d1d2 h8h5 a1e1 e7c5 b2b4 c5d5 d2d5 d8d5 e1e4');
INSERT INTO evaluations VALUES(20688,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20688,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(20688,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(20688,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20688,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20688,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20688,36,'cp 10','c5d4 c3d4 d6d5 e4d5 e6d5 c4d3 g8f6 b1c3 c8g4 h2h3 g4f3 d1f3 c6d4 f3e3 d4e6 d3b5 f6d7 c3d5 a7a6 b5c4 f8c5 e3e2 e8g8 e1g1 b7b5 c4b3 e6d4 e2h5 f8e8 c1e3 e8e5 h5d1 d4b3 d1b3 d7f6 e3c5 d8d5 b3d5 e5d5 c5e3 d5d8 f1d1 h7h6 a1c1 f6d5 e3c5 d5f4');
INSERT INTO evaluations VALUES(20717,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(20717,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(20717,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(20717,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(20717,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20717,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(20717,36,'cp 9','c3d4 d6d5 e4d5 e6d5 c4b3 f8d6 b1c3 g8e7 c1g5 f7f6 g5d2 c8g4 b3d5 e7d5 c3d5 d6e5 d5e3 g4f3 d1f3 d8d4 d2c3 d4f4 f3d5 e5c3 b2c3 f4e5 d5e5 c6e5 e1c1 a8d8 c1c2 e8f7 f2f4 e5g6 g2g3 g6e7 h1e1 h8e8 a2a4 b7b6 c3c4 g7g6 d1d8 e8d8 c2c3');
INSERT INTO evaluations VALUES(20714,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20714,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20714,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20714,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20714,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(20714,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20714,36,'cp 0','d6d5 e4d5 e6d5 c4b3 f8d6 c1g5 f7f6 g5d2 d8e7 d2e3 e7f7 e1g1 g8e7 b1c3 e8g8 f1e1 c6a5 h2h3 a5b3 d1b3 a7a6 c3a4 a8b8 e3d2 e7c6 a4b6 c8e6 e1e2 b8e8 a1e1 h7h6 b6a4 f7d7 a4b6 d7f7');
INSERT INTO evaluations VALUES(21420,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21420,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21420,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(21420,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21420,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(21420,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(21420,36,'cp 16','g8f6 b1d2 c6b4 d3e2 c7c5 e1g1 b4c6 c2c3 a7a5 a2a3 f8e7 d4c5 e7c5 a1b1 c5d6 b2b4 a5b4 a3b4 e8g8 b4b5 c6e5 c3c4 e5c4 d2c4 d5c4 c1b2 b7b6 d1d4 c8b7 f1d1 d6c5 d4c4 d8e7 f3e5 f8c8 e2f3 b7f3 g2f3');
INSERT INTO evaluations VALUES(34895,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(34895,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(34895,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(34895,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(34895,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(34895,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(34895,36,'cp -75','b8c6 d4b5 d5e5 d1e2 a7a6 d2d4 e5e7 b5a3 f7f5 g2g3 c8e6 f1g2 e8c8 e1g1 g8f6 f2f3 e4f3 e2f3 g7g6 f3e2 e6d5 e2e7 f8e7 g2d5 f6d5 a3c4 e7f6 a2a4 h7h5 b1d2 b7b5 a4b5 a6b5 c4a5 c6a5 a1a5 b5b4 d2f3 b4c3 b2c3 d5c3 c1g5 f6g5 f3g5 h8e8 a5a8 c8b7 a8d8 e8d8');
INSERT INTO evaluations VALUES(42785,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(42785,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42785,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42785,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(42785,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(42785,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42785,36,'cp -13','f6e4 f3e5 d7d5 c4b3 e8g8 e1g1 f8e8 f2f3 e4f6 d2d4 c7c5 c3c4 b8c6 c1b2 c6d4 b2d4 c5d4 d1d4 c8e6 c4d5 e6d5 a1d1 d8c7 e5g4 f6g4 f3g4 d5e6 d4f4 a8c8 d1d4 c7f4 f1f4 b7b6 b3e6 e8e6 d4d7 e6e2 d7a7 e2c2');
INSERT INTO evaluations VALUES(42903,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42903,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(42903,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(42903,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(42903,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(42903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42903,36,'cp -11','f3e5 d7d5 c4b3 e8g8 e1g1 f8e8 f2f3 e4f6 d2d4 c7c5 c3c4 b8c6 c1b2 c6d4 b2d4 c5d4 d1d4 c8e6 a1d1 d8c7 c4c5 e8d8 f1e1 a8c8 e5d3 e6f5 d3f4 c7c5 d4c5 c8c5 f4d5 f6d5 d1d5 c5d5 b3d5 f5c2 d5b7 g8f8 g2g3 c2g6 b7e4 g6e4 e1e4');
INSERT INTO evaluations VALUES(42891,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42891,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42891,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42891,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42891,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42891,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42891,36,'cp 0','d7d5 c4b3 e8g8 e1g1 a7a5 a2a4 f8e8 f2f3 e8e5 f3e4 e5e4 d1f3 c8e6 d2d3 e4e5 c1f4 e5f5 f3g3 c7c6 f4c7 f5f1 a1f1 d8f8 c7d6 f8c8 d6e5 c8f8');
INSERT INTO evaluations VALUES(44192,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44192,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44192,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44192,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(44192,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(44192,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(44192,36,'cp 24','e4e5 f6g8 g1f3 b8c6 d2d4 c5d4 f3d4 c6e5 d4b5 a7a6 b5d6 f8d6 d1d6 f7f6 c1e3 g8e7 e3b6 e7f5 d6b4 e5c6 b4c5 d7d6 b6d8 d6c5 d8b6 c6d4 e1c1 c8d7 b6c5 e6e5 f1d3 d7c6 f2f4 f5e3 d1d2 e3g2 f4e5 f6e5 h1g1 g2f4 g1g7 e8c8 c5b6 h8g8 g7g8 f4d3 d2d3 d8g8 b6d4 e5d4 d3d4');
INSERT INTO evaluations VALUES(50115,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(50115,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50115,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(50115,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(50115,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50115,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50115,36,'cp 0','f6e4 g5d2 e7e5 d4e5 d7d6 e5d6 d8d6 b1c3 d6d2 d1d2 e4d2 e1d2 c8e6 e2e3 b8c6 d2c2 e8c8 g1h3 c6b4 c2b3 b4c6 b3c2');
INSERT INTO evaluations VALUES(51067,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(51067,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(51067,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(51067,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(51067,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(51067,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(51067,36,'cp 198','g1f3 g7g6 h2h3 f8g7 b1c3 c7c6 f1d3 d7d5 e4d5 c6d5 c1f4 g8f6 c3b5 e8g8 e1g1 c8d7 b5c7 a8a7 a2a4 b8c6 c2c3 d8c8 f1e1 d7f5 d3f1 f5e4 c7b5 a7a6 f3e5 e4f5 f2f3 f8d8 d1e2 c6e5 f4e5 f5d7 e2d2 d8f8 e5g3 d7b5 f1b5');
INSERT INTO evaluations VALUES(51066,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(51066,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(51066,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(51066,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(51066,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(51066,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(51066,36,'cp 191','g7g6 b1c3 c7c6 f1d3 d7d6 c1e3 c8g4 h2h3 g4f3 d1f3 b8d7 g2g4 e7e6 f3g2 f8h6 f2f4 d8b6 e1c1 e6e5 c3e2 e5f4 e2f4 h5g4 h3g4 h6f4 e3f4 h8h1 g2h1 b6d4 d3f1 d4f6 f4d6 d7e5 h1h2 e5g4 h2h7 e8c8 f1h3 d8d6 h3g4 c8c7 h7g8');
INSERT INTO evaluations VALUES(52868,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(52868,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(52868,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(52868,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(52868,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(52868,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(52868,36,'cp 82','b7b5 c4b3 g8f6 e1g1 a7a5 a2a3 e7e6 d1e2 f8d6 d2d4 e8g8 c3e4 f6e4 e2e4 d8f6 c2c3 f6g6 e4e2 g6h5 f1e1 b8d7 b3e6 d7b6 e6a2 c8g4 h2h3 g4f3 e2f3 h5f3 g2f3 f8e8 c1e3 b6d5 h3h4 b5b4 a2d5 c6d5 a3b4 a5b4 g1g2 b4c3 b2c3 d6e7 h4h5');
INSERT INTO evaluations VALUES(54186,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54186,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54186,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54186,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54186,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(54186,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(54186,36,'cp -25','a7a6 g1f3 g8f6 d4c5 f8c5 f1d3 b8d7 e1g1 b7b5 e3e4 b5b4 e4e5 b4c3 e5f6 d7f6 b2c3 e8g8 d1e2 c5d6 f4d6 d8d6 f1b1 d6c7 e2e5 a8a7 e5d4 a6a5 f3e5 f6d7 e5d7 c8d7 c3c4 d5c4 d3c4 a7a8 c4e2 a8c8 c2c4 f8d8 d4e3 h7h6 b1c1 c7c5 e3c5 c8c5 a1b1 d7e8 f2f4');
INSERT INTO evaluations VALUES(4182,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(4182,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4182,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(4182,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(4182,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(4182,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(4182,36,'cp 15','a7a6 f1e1 d7d6 d2f1 c8e6 c4e6 f7e6 f1g3 d8e8 h2h3 f6d7 c1e3 c5e3 e1e3 c6e7 d1b3 d7c5 b3c2 e7g6 d3d4 c5d7 d4e5 d7e5 f3e5 g6e5 a1f1 e5g6 g3e2 e6e5 e3g3 g6f4 e2f4 e5f4 g3f3 e8e5 f1d1 a8e8 c2b3 f8f7');
INSERT INTO evaluations VALUES(10114,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(10114,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10114,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(10114,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(10114,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(10114,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10114,36,'cp 0','d7d5 e4d5 f6d5 g1f3 d5b6 c4b5 e8g8 e1g1 c5d6 f1e1 f8e8 c1e3 a7a6 b5c6 b7c6 b1d2 b6d5 d2e4 c8f5 d1a4 d8d7 f3h4 f5e6 d3d4 d5e3 e1e3 e5d4 c3d4 d6f4 e3f3 f4d6');
INSERT INTO evaluations VALUES(26197,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(26197,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(26197,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(26197,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(26197,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(26197,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(26197,36,'cp 140','f3h4 f5e4 e5e6 f7e6 b1d2 g8f6 f2f3 e4f5 g2g4 f5g6 h4g6 h7g6 f3f4 d8b6 c2c3 c6c5 f1d3 e8c8 d1e2 c5d4 c3d4 c8b8 d2f3 f6g4 a2a4 a7a6 a4a5 b6c6 c1d2 d7f6 a1c1 c6d7 f3g5 d8c8 c1c8 b8c8 g5e6 c8b8 f4f5 b8a7 d2f4 g6f5 d3f5');
INSERT INTO evaluations VALUES(28034,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(28034,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(28034,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28034,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(28034,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(28034,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(28034,36,'cp 28','c2c4 d7d6 b1c3 g8f6 d1c2 b8d7 f1e2 e7e5 f4g5 e5d4 f3d4 e8g8 e1g1 f8e8 a1d1 a7a6 e2f3 b7f3 d4f3 d8c8 h2h3 c8b7 a2a4 f6e4 c3d5 e4g5 f3g5 d7f6 b2b4 b6b5 a4a5 b5c4 c2c4 f6d5 d1d5');
INSERT INTO evaluations VALUES(34658,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(34658,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(34658,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(34658,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(34658,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(34658,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(34658,36,'cp 200','f1e2 c8g4 e1g1 b8c6 d2d4 e8c8 b1c3 e4f5 d4d5 g4f3 e2f3 e7e6 f3e4 f5e5 f1e1 e6d5 e4d3 e5d6 c3b5 d6d7 d1f3 g7g6 c1f4 a7a6 b5c7 g8f6 a1d1 f8d6 c7a8 d6c5 f4c7 f6g4 a8b6 c5b6 c7b6 g4e5 f3f6 d8e8 d3f1 d7e6 f6f4 e6g4 b6e3 g4f4');
INSERT INTO evaluations VALUES(34659,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(34659,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(34659,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(34659,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(34659,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(34659,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(34659,36,'cp 186','c7c6 e1g1 e4f4 d2d4 f4c7 f3e5 b8d7 c1f4 d7e5 f4e5 c7d8 c2c4 f7f6 e5f4 e7e6 b1c3 g8e7 f1e1 e7g6 f4g3 f6f5 e2h5 f8d6 g3d6 d8d6 d1d2 e8g8 h5g6 h7g6 a1d1 c8d7 d2e3 a8e8 a2a3 d6e7 c3a4 g6g5 a4c5 b7b6');
INSERT INTO evaluations VALUES(35756,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(35756,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(35756,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(35756,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(35756,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(35756,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(35756,36,'cp 34','e1g1 b7b6 b2b3 b8d7 c1b2 c8b7 d1e2 d8c7 f1d1 a8d8 c4d5 e6d5 d1c1 a7a5 c1c2 c7b8 a2a4 f8e8 e2d1 d8c8 h2h3 e7d6 b2a3 d6c7 f3d2 c8d8 a1c1 d7f8 c3e2 f8e6 c2c6 b7c6 c1c6 d8c8');
INSERT INTO evaluations VALUES(35755,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(35755,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(35755,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(35755,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(35755,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35755,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(35755,36,'cp 41','b7b6 b2b3 b8d7 c1b2 c8b7 d1c2 c6c5 c4d5 e6d5 f1d1 f8e8 a1c1 a8c8 c2e2 e7d6 a2a4 f6e4 d3b5 e4c3 b2c3 d8e7 a4a5 e8d8 b5a6 b7a6 e2a6 e7e6 d4c5 d7c5 a6b5 b6a5 c3a5 a7a6 b5f1 c5b3 a5d8 b3c1 d1c1 c8d8');
INSERT INTO evaluations VALUES(37671,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(37671,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(37671,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(37671,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(37671,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(37671,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(37671,36,'cp 48','d8b6 c3a4 b6a5 c2c3 c5d4 d1d4 g8f6 a4c5 e8g8 c5b3 a5h5 f1e2 h5g6 f2f3 b8c6 d4f2 d7d5 e4d5 f6d5 e1g1 e6e5 f1e1 c8f5 e2b5 a8e8 c1d2 a7a6 b5f1 e8e7 f2g3 g6d6 a1d1 f5c2 d1c1 c2b3 a2b3 c6a5 b3b4 a5b3 c1d1 d6b6 g1h1 b3d2');
INSERT INTO evaluations VALUES(32979,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(32979,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(32979,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(32979,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(32979,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(32979,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(32979,36,'cp 51','c3a4 b6a5 c2c3 c5d4 d1d4 g8f6 a4c5 e8g8 f2f3 b8c6 d4f2 d7d5 c1g5 f6d7 c5b3 a5c7 e4d5 e6d5 f1e2 f8e8 g5h4 c7e5 h4g3 e5e7 e1c1 e7g5 c1b1 d7b6 b1a1 c8f5 e2b5 b6c4 h2h4 g5e7 b3d4 c6d4 d1d4 c4e3 d4d2 d5d4 h1c1 e8d8 d2d4 d8d4 c3d4');
INSERT INTO evaluations VALUES(32982,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(32982,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(32982,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(32982,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(32982,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(32982,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(32982,36,'cp 37','b6a5 c2c3 c5d4 d1d4 g8f6 a4c5 e8g8 f2f3 b8c6 d4f2 a5c7 f1b5 e6e5 c5b3 d7d6 c1e3 c8e6 e1g1 h7h6 f1d1 d6d5 e4d5 e6d5 b3c5 c6e7 b5d3 d5c6 c3c4 a8d8 d3c2 d8d1 a1d1 f8d8 d1d8 c7d8 b2b3 b7b6 c5a6 e5e4 a6b4 e4f3 b4c6');
INSERT INTO evaluations VALUES(38095,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(38095,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38095,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(38095,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(38095,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(38095,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(38095,36,'cp 28','g8f6 f1d3 d5c4 d3c4 b7b5 c4d3 b8d7 e1g1 e8g8 d1c2 c8b7 e3e4 e6e5 f3e5 d7e5 d4e5 d6e5 h2h3 b5b4 c3a4 e5d4 f1e1 c6c5 c1e3 a8c8 e3d4 c5d4 c2d2 f8e8 e4e5 d8d5 f2f4 f6h5 d3e4 d5b5 e4b7 b5b7 d2d4 e8d8 d4e4 b7b5 e4f3 d8d2 f3h5 b5a4');
INSERT INTO evaluations VALUES(38372,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(38372,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(38372,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38372,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(38372,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38372,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(38372,36,'cp 29','g8f6 f1e2 e8g8 e3e4 c7c5 g1f3 e7e6 e1g1 b8c6 c1e3 f6g4 e3g5 f7f6 g5c1 f6f5 d4d5 e6d5 c1g5 d8b6 d1d5 g8h8 d5d2 f5e4 c3e4 b6b2 d2b2 g7b2 a1e1 b7b6 e4d6 c8d7 e2d3 g4e5 d3e4 e5f3 e4f3 f8f3 g2f3 d7h3 e1b1 b2g7');
INSERT INTO evaluations VALUES(38967,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(38967,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(38967,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(38967,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(38967,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(38967,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(38967,36,'cp 173','b1d2 c7c5 a2a3 b4d2 d1d2 b8c6 e2e3 c8g4 f1e2 d8d7 e1g1 a8d8 e3d4 g4f3 e2f3 c6d4 d2d3 g8e7 f3e4 d7g4 h2h3 g4h5 f1e1 h5e5 c1d2 e5c7 b2b4 h7h6 a1b1 e8g8 b4c5 d4e6 d3e3 e6c5 d2b4 b7b6 b4c5 c7c5 e3c5 b6c5 b1b7 e7c8 b7c7 c8d6 e4d5 d6f5 c7c5 f5d4');
INSERT INTO evaluations VALUES(45893,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(45893,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(45893,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45893,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(45893,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45893,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(45893,36,'cp 0','a7a6 b5a3 b7b5 c3d5 b8d7 c2c4 f6d5 d1d5 d7b6 d5d1 b5c4 a3c4 c8e6 c4b6 d8b6 d1a4 e6d7 a4b3 b6c6 f1c4 c6e4 c1e3 e4f5 c4d3 f5e6 d3c4 e6f5');
INSERT INTO evaluations VALUES(43439,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43439,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(43439,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43439,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(43439,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(43439,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43439,36,'cp 11','b5a3 b7b5 c3d5 f6d5 d1d5 a8a7 c2c4 c8e6 d5d1 b5b4 a3c2 a6a5 c1e3 a7a8 g2g3 b8d7 b2b3 f8e7 f1g2 d8c7 e1g1 e8g8 f2f4 d7c5 f4f5 e6c8 g3g4 c7d8 d1e2 c8b7 e3c5 d6c5 a1d1 d8b6 c2e3 e7g5 e3d5 b6h6 f5f6 g5f6 d5f6 g7f6 d1d6');
INSERT INTO evaluations VALUES(43472,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(43472,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(43472,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(43472,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(43472,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(43472,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(43472,36,'cp 2','b7b5 c3d5 b8d7 c2c4 f6d5 d1d5 d7b6 d5d1 b5c4 a3c4 c8e6 c4b6 d8b6 d1a4 e6d7 a4b3 b6c6 f1c4 c6e4 c1e3 e4f5 b3b7 a8d8 e1g1 f8e7 e3b6 d8c8 c4a6 e8g8 b6c7 f5e6 b7b6 c8a8 a2a4 d6d5 b6e6 f7e6 a6b5 d7b5 a4b5 e5e4 a1a8 f8a8 f1c1 g8f7 b2b3 f7e8');
INSERT INTO evaluations VALUES(43745,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43745,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43745,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43745,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(43745,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43745,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(43745,36,'cp 12','c3d5 b8d7 c2c4 f6d5 d1d5 d7b6 d5d1 b5c4 a3c4 c8e6 c4b6 d8b6 d1a4 e6d7 a4b3 b6c6 f1c4 c6e4 c1e3 e4f5 b3b7 a8d8 e3b6 d8c8 c4a6 f8e7 e1g1 e8g8 b6c7 f5g6 a2a4 e5e4 f1e1 d6d5 b7b6 c8a8 a6b5 d7b5 a4b5 g6b6 c7b6 f8c8 a1a8 c8a8');
INSERT INTO evaluations VALUES(48509,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(48509,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(48509,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(48509,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(48509,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(48509,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(48509,36,'cp -29','f1d3 f8e7 a2a3 b8d7 d1e2 c7c5 c1e3 e8g8 e1c1 f6d5 c3d5 e6d5 d4c5 d7c5 d3c4 c8e6 e3c5 e7c5 c4d5 e6d5 e2e5 d8b6 e5d5 c5e7 d5d4 b6g6 d4f2 e7f6 c2c3 h7h6 f2c2 g6g4 h2h3 g4c4 h1e1 a8d8 d1d8 f8d8 c2e2 c4c5 g2g4 g7g6 c1c2 f6g7');
INSERT INTO evaluations VALUES(48513,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(48513,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(48513,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(48513,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(48513,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(48513,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(48513,36,'cp -63','c7c5 c1e3 c5d4 f3d4 f8b4 e1g1 e8g8 g1h1 b8d7 d1e2 b4c3 b2c3 f6d5 e3c1 d7c5 c1a3 b7b6 c3c4 d5c3 e2e3 c8b7 a3c5 b6c5 d4b3 c3a4 b3c5 a4c5 e3c5 d8c8 c5d6 c8c6 d6c6 b7c6 a1e1 g7g6 e1e5 f8d8 h1g1 a7a6 f1e1 a8c8 h2h3 c6b7 e5a5');
INSERT INTO evaluations VALUES(49368,30,'cp -155',NULL);
INSERT INTO evaluations VALUES(49368,31,'cp -156',NULL);
INSERT INTO evaluations VALUES(49368,32,'cp -183',NULL);
INSERT INTO evaluations VALUES(49368,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(49368,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(49368,35,'cp -166',NULL);
INSERT INTO evaluations VALUES(49368,36,'cp -151','f6e4 e1g1 d7d5 c4e2 c7c5 d2d4 c5d4 f3d4 f8c5 d4b3 c5b6 c2c4 b8c6 c4d5 e6d5 b1c3 e4c3 b2c3 e8g8 e2f3 c8e6 c1a3 f8e8 a3c5 c6e5 c5b6 d8b6 b3d4 a8c8 a1b1 b6c7 f1e1 b7b6 e1e3 e5c4 e3e2 h7h6 d1e1 c7d6 d4b5 d6d7 b5d4 c4d6 d4e6 f7e6 f3g4 d6e4');
INSERT INTO evaluations VALUES(49340,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(49340,31,'cp -159',NULL);
INSERT INTO evaluations VALUES(49340,32,'cp -155',NULL);
INSERT INTO evaluations VALUES(49340,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(49340,34,'cp -153',NULL);
INSERT INTO evaluations VALUES(49340,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(49340,36,'cp -157','d2d4 d7d5 c4e2 c7c5 e1g1 c5d4 f3d4 f8c5 d4b3 c5b6 c2c4 b8c6 c4d5 e6d5 b1c3 e4c3 b2c3 e8g8 b3d4 c6d4 c3d4 c8f5 c1e3 a8c8 d1b3 f5c2 b3b5 c2e4 a1c1 d8f6 b5d7 c8b8 f1d1 h7h6 e2d3 f8d8 d7h3 e4d3 d1d3 d8e8');
INSERT INTO evaluations VALUES(50748,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(50748,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(50748,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(50748,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50748,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50748,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50748,36,'cp 0','d8c7 e1g1 g8f6 c4e2 f8e7 d2d4 c5d4 d1d4 c8d7 c1e3 b8c6 d4b6 c7b6 e3b6 c6b4 b6a5 b4c6');
INSERT INTO evaluations VALUES(50877,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(50877,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(50877,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(50877,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(50877,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(50877,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(50877,36,'cp 118','d2d4 e5d4 c2c3 d4d3 f1e1 b8c6 e4e5 c8e6 c4e6 f7e6 d1d3 d8d7 e5d6 e8c8 d3c4 f8d6 c4e6 d7e6 e1e6 d6c5 b1d2 g8f6 d2b3 d8d1 f3e1 c5b6 c1f4 d1a1 b3a1 f6g4 f4g3 h8d8 a1b3 d8d1 e6e8 c6d8 e8e2 a6a5 b3d2 a5a4 d2e4 d8f7 h2h3 g4e5 g1h2 e5c4 b2b3 c4d6 e4d6 f7d6 e1f3 a4b3 a2b3 d6f5');
INSERT INTO evaluations VALUES(50880,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(50880,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(50880,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(50880,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(50880,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(50880,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(50880,36,'cp 129','e5d4 c2c3 d4d3 f1e1 b8c6 e4e5 c8e6 c4e6 f7e6 d1d3 d8d7 e5d6 e8c8 d3g6 f8d6 g6e6 d7e6 e1e6 g8f6 b1d2 h8e8 e6e8 d8e8 d2c4 d6f8 c1e3 b7b5 c4d2 g7g5 h2h3 f6d5 a1e1 f8g7 d2b3 d5e3 e1e3 e8e3 f2e3 b5b4 b3d4 b4c3 b2c3 c6e7 g1f2');
INSERT INTO evaluations VALUES(52120,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52120,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52120,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52120,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52120,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(52120,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52120,36,'cp 19','f1c4 b8c6 b1c3 g7g6 g1e2 f8g7 d2d3 d7d6 c3d5 g8e7 c1g5 h7h6 g5e7 c6e7 d5e7 d8e7 e2c3 e8g8 c3d5 e7d8 e1g1 h6h5 a2a4 g7h6 d5e3 g8g7 a4a5 a8b8 d1e2 d8g5 f1b1 c8g4 e2e1 h5h4 e3g4 g5g4 h2h3 g4d7');
INSERT INTO evaluations VALUES(52538,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(52538,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(52538,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(52538,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(52538,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(52538,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(52538,36,'cp -65','g7g6 d2d4 f8g7 g1f3 d7d6 f1d3 c8g4 c2c3 g8f6 e1g1 b8c6 h2h3 g4f3 d1f3 e8g8 b1d2 a8b8 f3e2 a7a6 f1b1 d8c7 d2f3 b8a8 a3c1 e7e6 a1a3 b7b5 b1a1 c7b7 c1e3 f8b8 e2a2 a6a5 d4d5 e6d5 e4d5 c6e7 a3a5 a8a5 a2a5');
INSERT INTO evaluations VALUES(52769,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(52769,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(52769,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(52769,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52769,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(52769,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(52769,36,'cp -6','g8f6 g1f3 c8f5 f1d3 e7e6 e1g1 b8d7 c3c4 f5g6 d1c2 d5c4 c2c4 f8d6 b1d2 e8g8 d3g6 h7g6 b2b3 e6e5 c4c2 d8e7 d4e5 d7e5 c1b2 a8d8 a1d1 f8e8 h2h3 d6c7 b2e5 c7e5 f3e5 e7e5 d2f3 e5f5 c2e2 f6e4 e2c2 e4c5 d1d8 e8d8 f1c1 f5c2 c1c2 d8d1 g1h2');
INSERT INTO evaluations VALUES(44947,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(44947,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(44947,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(44947,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44947,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44947,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44947,36,'cp 0','f1d3 b8d7 f2f4 f6e4 g1f3 d7f6 c3c4 d5c4 d3c4 c8f5 e1g1 e7e6 c4d3 f8e7 f3e5 h7h6 d1f3 d8d5 d3c4 d5d8 c4d3');
INSERT INTO evaluations VALUES(10695,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(10695,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(10695,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(10695,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(10695,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(10695,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(10695,36,'cp 58','d7d5 b3c3 d5c4 d4d5 c6e7 f1e1 e8g8 c1a3 e7d5 c3c4 c7c6 a3f8 d8f8 a1b1 a7a5 f3e5 a5a4 h2h3 h7h6 e5d3 d5c7 d3f4 c7e6 f4h5 a8a5 h5g3 a4a3 c4c3 a5d5 e1e3 e6d4 c3a3 b7b5 a3f8 g8f8 g3e4 c8e6 b1c1');
INSERT INTO evaluations VALUES(12814,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12814,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12814,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12814,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(12814,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12814,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(12814,36,'cp 10','c5d4 f3d4 e7e6 a2a3 f8e7 e2e4 e8g8 d4b3 d7d6 c1e3 a7a5 f1e2 a5a4 b3d4 d6d5 e4d5 e6d5 d4c6 b7c6 e1g1 d5c4 e2c4 c8g4 d1c2 d8d7 a1c1 g4f5 c2e2 f5g4 f2f3 g4e6 f1d1 f6d5 g1h1 f8d8');
INSERT INTO evaluations VALUES(12874,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(12874,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(12874,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(12874,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(12874,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12874,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(12874,36,'cp 12','f3d4 e7e6 a2a3 f8e7 e2e4 e8g8 d4f3 d8a5 f1d3 d7d5 e1g1 d5c4 d3c4 f8d8 d1e2 c6e5 f3e5 a5e5 g2g3 c8d7 c1f4 e5h5 e2h5 f6h5 f4e3 h5f6 c4e2 d7c6 f2f3 a7a6 f1d1 f6d7 d1d2 g7g5 a1d1 h7h5 h2h3 g8g7 f3f4 g5f4 g3f4 d7f6 d2d8 a8d8 e4e5 d8d1 e2d1 f6d5 c3d5 c6d5 d1h5');
INSERT INTO evaluations VALUES(22604,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(22604,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(22604,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(22604,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(22604,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(22604,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(22604,36,'cp 53','g8f6 e2e4 f8b4 f1d3 d7d5 e1g1 b4c3 e4d5 c6d4 b2c3 d4f3 d1f3 d8d5 f3g3 b7b6 a2a4 c8b7 d3b5 c7c6 b5e2 f6e4 g3g7 e8c8 c3c4 d5f5 e2g4 f5g6 g7g6 h7g6 g4e2 c6c5 f1e1 e6e5 c1b2 f7f6 f2f3 e4d2 f3f4 h8h4 g2g3 h4h5 a1d1 d2f3 e2f3 b7f3 d1d8 c8d8 f4e5 f6e5 b2e5 d8c8 e5f4 c8b7');
INSERT INTO evaluations VALUES(27438,30,'cp 691',NULL);
INSERT INTO evaluations VALUES(27438,31,'cp 687',NULL);
INSERT INTO evaluations VALUES(27438,32,'cp 694',NULL);
INSERT INTO evaluations VALUES(27438,33,'cp 736',NULL);
INSERT INTO evaluations VALUES(27438,34,'cp 732',NULL);
INSERT INTO evaluations VALUES(27438,35,'cp 747',NULL);
INSERT INTO evaluations VALUES(27438,36,'cp 748','d5c6 e8d8 c6b7 a8b8 b7d5 g5g6 d5e6 g6e6 e1g1 f8d6 f3e3 a7a6 b2b3 h8e8 c2c4 e5e4 c1b2 e6h6 e3h6 g7h6 b1c3 d6e5 f1e1 e5d4 c3a4 d4b2 a4b2 b8b6 a1d1 h6h5 b2a4 b6c6 d2d3 e4e3 f2e3 c6g6 g1f2 h5h4 a4c3 d8e7 d3d4');
INSERT INTO evaluations VALUES(27437,30,'cp 698',NULL);
INSERT INTO evaluations VALUES(27437,31,'cp 675',NULL);
INSERT INTO evaluations VALUES(27437,32,'cp 678',NULL);
INSERT INTO evaluations VALUES(27437,33,'cp 671',NULL);
INSERT INTO evaluations VALUES(27437,34,'cp 675',NULL);
INSERT INTO evaluations VALUES(27437,35,'cp 676',NULL);
INSERT INTO evaluations VALUES(27437,36,'cp 694','e8d8 c6b7 a8b8 b7d5 g5f6 d5e6 f6e6 e1g1 f8d6 f3e3 a7a6 b2b3 h8e8 c2c4 c7c6 c1b2 d8e7 f1e1 e7f8 e3e4 f8g8 e4c6 b8d8 g2g3 d6b4 c6e6 e8e6 e1e5 e6e5 b2e5 f7f6 e5c3 b4c3 b1c3 d8d2 b3b4');
INSERT INTO evaluations VALUES(31886,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(31886,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(31886,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(31886,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(31886,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(31886,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31886,36,'cp 22','c2c4 g8f6 g5f6 e7f6 e2e3 f8b4 b1c3 c7c5 c4d5 d8d5 d1a4 b8c6 a4b4 c6b4 c3d5 b4d5 g1f3 a8c8 a1c1 c5d4 c1c8 f5c8 f3d4 e8e7 e1d2 h8d8 f1d3 g7g6 h1c1 f6f5 d3e2 c8e6 e2f3 f5f4 d2e2 f4e3 f2e3 f7f5 c1c5 b7b6 c5c6 d5f4 e3f4 d8d4 e2e3 d4b4 b2b3');
INSERT INTO evaluations VALUES(42802,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(42802,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42802,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(42802,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42802,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(42802,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(42802,36,'cp 15','e8g8 d2d3 e5e4 c1g5 e4d3 e2d3 d7d6 g1e2 b8d7 e1g1 h7h6 g5f6 d7f6 a2a4 f8e8 a4a5 a8b8 d1d2 c8d7 a5a6 d7c6 a6b7 c6g2 g1g2 b8b7 e2f4 d8d7 f4d5 f6d5 c4d5 c7c6 d5c6 d7c6 g2g1 c6c5 c3c4');
INSERT INTO evaluations VALUES(37277,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(37277,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(37277,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(37277,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(37277,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(37277,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(37277,36,'cp 26','d2d3 c7c6 g1f3 f8e8 e1g1 h7h6 e2e4 d7d5 e4d5 c6d5 c4d5 d8d5 c3c4 d5a5 c1d2 a5c7 d2c3 b8c6 f1e1 c8f5 d1d2 e8e7 d2b2 a8e8 a1d1 b7b6 h2h3 c7d6 b2b5 f5d7 b5b3 d7f5 e1e3 d6d7 d3d4 c6d4 f3d4 e5d4 d1d4 d7c7 e3e7 c7e7');
INSERT INTO evaluations VALUES(43033,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(43033,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(43033,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(43033,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(43033,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(43033,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(43033,36,'cp 125','f3e5 e8g8 d2d3 d7d5 e4d5 b7b5 c4b3 e7d6 f2f4 c7c6 d1f3 c6d5 b1c3 b8d7 d3d4 d6b4 e1g1 b4c3 f3c3 a7a5 a2a4 b5b4 c3f3 c8b7 c2c3 b4c3 b2c3 a8c8 c1a3 f8e8 f1e1 d8c7 a1c1 d7b6 f3f2 b7c6 b3c2 b6a4 c3c4 d5c4 c2a4 c6a4 c1c4 c7c4 e5c4 c8c4 e1e8 a4e8 f2d2 e8d7 a3e7 a5a4 e7f6 g7f6');
INSERT INTO evaluations VALUES(43009,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(43009,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(43009,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(43009,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(43009,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(43009,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(43009,36,'cp 140','e8g8 d2d3 d7d5 e4d5 b7b5 c4b3 e7d6 f2f4 c7c6 a2a4 b5b4 b1d2 c6d5 d2f3 f8e8 e1g1 b8d7 c1e3 d7c5 f3g5 e8f8 e3d4 c5b3 c2b3 h7h6 g5f3 f6h5 f3e1 h5f4 f1f4 d8e8 f4f2 d6e5 d4c5 e5f6 c5f8 e8f8 d1d2 c8g4 d3d4 a8e8 e1c2');
INSERT INTO evaluations VALUES(37311,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(37311,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(37311,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(37311,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(37311,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(37311,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(37311,36,'cp 140','d2d3 d7d5 e4d5 b7b5 c4b3 e7d6 f2f4 c7c6 a2a4 b5b4 b1d2 c6d5 d2f3 f8e8 e1g1 b8a6 c1e3 a6c5 f3g5 e8f8 e3d4 c8b7 d1e2 a8c8 a1e1 h7h6 g5f3 c8c7 e2f2 d8c8 f2h4 c5b3 c2b3 b7a6 h2h3 f6d7 e5d7 c8d7 f3e5 d7d8 h4g3 f7f6 e5g6 f8e8 g6h4');
INSERT INTO evaluations VALUES(37310,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(37310,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(37310,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(37310,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(37310,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(37310,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(37310,36,'cp 125','d7d5 e4d5 b7b5 c4b3 e7d6 f2f4 c7c6 a2a4 b5b4 b1d2 f8e8 e1g1 c6d5 d2f3 d8b6 g1h1 b8c6 f1e1 c6a5 c1e3 d6c5 e3c5 b6c5 b3a2 c8g4 h2h3 g4f3 d1f3 a8d8 a1c1 c5d4 c2c3 b4c3 b2c3 d4a4 f3f2 f6h5 d3d4 f7f6 e5d3 a5c4 d3c5 e8e1 c1e1');
INSERT INTO evaluations VALUES(46680,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(46680,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46680,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46680,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46680,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(46680,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(46680,36,'cp -1','f8g7 h2h3 f6d7 f1e2 b8c6 g1f3 e8g8 a2a4 e7e5 f4h2 g8h8 e1g1 f7f5 d4e5 d7e5 b1d2 d8e7 b2b4 g6g5 b4b5 c6d8 f3e5 d6e5 e2c4 c8e6 g1h1 e5e4 f2f3 e4f3 d1f3 e7c5 f3e2 e6c4 d2c4 d8f7 a1d1 a8e8');
INSERT INTO evaluations VALUES(1877,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(1877,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(1877,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(1877,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(1877,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(1877,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(1877,36,'cp 7','f8c5 c1e3 d7d6 f1e2 f6g6 d4b5 c5e3 b5c7 e8d7 c7a8 g6g2 h1f1 e3f4 b1d2 g2h2 d1a4 f4d2 e1d2 h2f4 d2c2 d7d8 a1d1 c8d7 f2f3 d8c8 a4a3 c8b8 d1d6 g8f6 f1d1 b8a8 d6f6 f4f6 d1d7 c6b8 d7d5 h7h5 a3c5 h5h4 d5f5 f6g6 e2c4 h4h3 f5f7');
INSERT INTO evaluations VALUES(9410,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9410,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9410,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9410,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9410,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9410,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(9410,36,'cp 0','c2c3 a7a6 b2b4 c5a7 a2a4 h7h6 h2h3 e8g8 f1e1 f6h7 b1d2 d8f6 d2f1 f6g6 f3h4 g6f6');
INSERT INTO evaluations VALUES(9404,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9404,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9404,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9404,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9404,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9404,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9404,36,'cp 0','a7a6 h2h3 e8g8 f1e1 h7h6 b2b4 c5a7 b1d2 f6h7 a2a4 d8f6 d2f1 f6g6 f3h4 g6f6');
INSERT INTO evaluations VALUES(14190,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14190,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14190,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14190,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14190,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(14190,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14190,36,'cp 10','f6e4 c4d5 e4g5 d5c6 g5f3 g2f3 a8b8 e2e3 b7c6 d1c2 e7e5 d4e5 d8d5 f1g2 f8b4 b1c3 d5e5 e1g1 b4c3 f3f4 b8b2 c2b2 c3b2 f4e5 b2a1 f1a1 e8e7 a1c1 c8e6 c1c6 h8b8 g2e4 b8b4 e4d3 e7d7 c6a6 g7g6 a2a4 e6d5 a6a7 d5c6 f2f4 b4a4');
INSERT INTO evaluations VALUES(26529,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(26529,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(26529,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(26529,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(26529,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26529,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(26529,36,'cp 11','e7e6 f1d3 f5d3 d1d3 f8d6 g1f3 d6f4 e3f4 d8d6 g2g3 g8h6 e1g1 e8g8 c3d1 h6f5 d1e3 c6e7 f1e1 b7b6 c2c3 c7c5 a2a3 a8c8 a1d1 c8c7 e3f5 e7f5 e1e2 c5d4 f3d4 f5d4 d3d4 g7g6 d4d2 g8g7 d1e1 d6c6 h2h4 h7h5 d2d4 g7g8');
INSERT INTO evaluations VALUES(26803,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(26803,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26803,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(26803,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(26803,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26803,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(26803,36,'cp 21','f1d3 f5d3 d1d3 f8d6 g1f3 d6f4 e3f4 d8d6 g2g3 g8h6 e1g1 e8g8 c3d1 h6g4 f1e1 g4f6 d1e3 c6e7 a2a3 f8e8 a1c1 d6b6 c2c4 d5c4 e3c4 b6a6 d3d2 a8c8 c4e5 h7h6 g3g4 f6d5 g4g5 h6h5 h2h4');
INSERT INTO evaluations VALUES(26964,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(26964,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(26964,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26964,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(26964,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26964,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26964,36,'cp 10','d8f6 c2c3 g4f3 d1f3 f6f3 g2f3 g7g6 a2a4 f8h6 c1e3 f7f5 a4a5 a7a6 b2b4 e8e7 g1g2 a8f8 h2h4 c6d8 f1g1 d8e6 c4d5 f8b8 g2f1 e6f4 d5b3 f5e4 d3e4 b8f8 b1d2 g8f6 d2c4 f8c8 c4a3 f6h5 g1h1 h5g7 b4b5 h8f8 b5a6 b7a6 a3c4');
INSERT INTO evaluations VALUES(29366,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(29366,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(29366,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(29366,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29366,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(29366,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(29366,36,'cp -3','d8b6 f3h4 f5e4 f2f3 e4g6 h4g6 h7g6 d1b3 b8c6 g2g4 f8e7 a2a4 b6b3 d2b3 f6d7 e1f2 e7h4 f2g2 c5c4 b3d2 e6e5 f4g3 h4g3 g2g3 e5d4 e3d4 e8c8 b2b3 c4b3 d2b3 d8e8 a1c1 e8e3 g3f2 h8e8 h2h4 c8b8 h4h5 g6h5 g4h5 d7f6 b3c5 e8e7 h5h6 g7h6');
INSERT INTO evaluations VALUES(34963,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(34963,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(34963,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(34963,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(34963,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(34963,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(34963,36,'cp 61','f2f4 e5f4 d2d3 c7c6 c1f4 g8f6 e1c1 f8e7 g1f3 e8g8 d1e1 f8e8 e2d1 h7h6 a2a3 b8d7 f3d2 d7f8 d2c4 c8e6 c4d6 e7d6 f4d6 e6g4 d1g4 f6g4 c1d2 g4f6 h1f1 f8g6 h2h3 a8d8 e1e8 d8e8 f1e1 e8e1 d2e1 h6h5 e1d2 f6d7');
INSERT INTO evaluations VALUES(35135,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(35135,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(35135,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(35135,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(35135,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(35135,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(35135,36,'cp 73','e5f4 b1c3 f4g3 h2g3 d7d5 c3d5 f6g6 e1f2 g8f6 d5f6 g6f6 d2d4 c8g4 c1g5 f6b6 d1d3 b8c6 c2c3 f7f6 g5e3 e8c8 d3b5 b6b5 f1b5 h7h5 a1f1 f8d6 f3d2 c6e7 a2a4 c7c6 b5d3 d6c7 b2b4 e7g6 d2c4 g6e5 c4e5 c7e5');
INSERT INTO evaluations VALUES(35153,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(35153,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(35153,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(35153,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(35153,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(35153,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(35153,36,'cp 82','d2d4 f4g3 h2g3 d7d5 b1c3 f8b4 e4d5 c8g4 f1e2 b8d7 e1g1 b4c3 b2c3 f6d6 d1e1 e8c8 c3c4 g8f6 c1f4 d6a3 e2d3 d8e8 e1c1 a3c1 a1c1 g4f3 f1f3 h7h5 c1f1 h5h4 a2a4 h4g3 f4g3 h8h5 f3f5 h5f5 d3f5 c8d8 a4a5 e8e3 g3h4 e3a3 c4c5 a3a5 f1b1 a5a4 c5c6 b7c6 d5c6');
INSERT INTO evaluations VALUES(44870,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(44870,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(44870,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(44870,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(44870,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(44870,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(44870,36,'cp 131','d5c6 b8c6 g1f3 c8g4 f1e2 g4f3 e2f3 c6d4 f3b7 a8b8 b7f3 e7e6 b1d2 d4f3 d2f3 d8d1 e1d1 f6e4 d1e2 f8c5 h1f1 e8e7 c2c3 a7a5 f3d2 e4d2 e2d2 g7g5 d2c2 f7f6 c1d2 h7h5 a1e1 c5d6 h2h3 d6f4 e1e2 e6e5 c3c4 f4d2 e2d2 h5h4 f1d1');
INSERT INTO evaluations VALUES(44842,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(44842,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(44842,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(44842,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(44842,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(44842,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(44842,36,'cp 125','b8c6 g1f3 c8g4 f1e2 g4f3 e2f3 c6d4 f3b7 a8b8 b7f3 g7g6 c2c3 d4f3 d1f3 d8b6 b1d2 f8g7 e1g1 e8g8 f3e2 b6c7 a2a4 f6d7 e2e7 f8e8 e7h4 d7c5 h4g3 c7g3 f2g3 c5d3 d2c4 e8e2 f1d1 d3f2 d1d7 f2e4');
INSERT INTO evaluations VALUES(13192,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(13192,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(13192,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(13192,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(13192,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(13192,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(13192,36,'cp 133','g1f3 c8g4 f1e2 g4f3 e2f3 c6d4 f3b7 a8b8 b7f3 e7e6 e1g1 f8c5 b1d2 d4f3 d1f3 e8g8 c2c4 d8c7 b2b3 b8d8 d2e4 f6e4 f3e4 c5d4 a1b1 e6e5 c1e3 f7f5 e4c2 f5f4 e3d4 e5d4 c4c5 f4f3 c2c4 g8h8 f1d1 c7f4 g2g3 f4f5 c5c6 f5h3 c4f1 h3e6 c6c7');
INSERT INTO evaluations VALUES(13190,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(13190,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(13190,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(13190,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(13190,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(13190,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(13190,36,'cp 113','c8g4 f1e2 g4f3 e2f3 c6d4 f3b7 a8b8 b7f3 g7g6 e1g1 f8g7 b1a3 e8g8 c2c3 d4f3 d1f3 d8d7 f1d1 d7e6 a3c2 e6e4 f3e4 f6e4 a2a4 a7a5 g1f1 e4c5 a1a2 c5b3 c2a3 e7e5 c1e3 f7f5 a3c4 f5f4 e3a7 b8d8 c4d6 f8f6 a2a3 f6d6 d1d6 b3d2 d6d2 d8d2');
INSERT INTO evaluations VALUES(24653,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(24653,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(24653,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(24653,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(24653,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(24653,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(24653,36,'cp 125','f1e2 g4f3 e2f3 c6d4 f3b7 a8b8 b7f3 g7g6 e1g1 f8g7 c2c3 d4f3 d1f3 e8g8 b1a3 d8c7 h2h3 f8d8 f1e1 c7d7 a3c2 f6d5 e1d1 d7a4 c2e3 d5f6 d1d8 b8d8 f3b7 f6e4 b7e7 g7f6 e7b7 d8d1 e3d1 a4d1 g1h2 e4f2 b7a7');
INSERT INTO evaluations VALUES(53633,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(53633,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(53633,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(53633,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(53633,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(53633,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(53633,36,'cp 66','c6c5 f1d3 b8c6 g1f3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f8e7 b1c3 e8g8 a1c1 f5e3 f2e3 c8d7 d1c2 h7h6 c2d2 f7f5 e5f6 f8f6 d3c2 d7e8 e3e4 d5e4 c3e4 f6f3 f1f3 c6d4 f3f2 d8b6 g1h1 a8d8');
INSERT INTO evaluations VALUES(54292,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(54292,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(54292,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(54292,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(54292,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(54292,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(54292,36,'cp 66','g1f3 b8c6 f1d3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f5e3 f2e3 f8e7 b1c3 c8d7 e3e4 d5e4 d3e4 d8b6 a2a3 a8c8 c3a4 b6d8 b2b4 b7b6 a1c1 e8g8 a4c3 f7f5 e5f6 e7f6 d1d3 h7h6 c3b5 c6e7 b5a7 c8c7 a7b5 d7b5 d3b5 c7c1');
INSERT INTO evaluations VALUES(3842,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3842,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(3842,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(3842,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3842,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(3842,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(3842,36,'cp 17','c3a4 c5b6 a4b6 a7b6 e1g1 h7h6 h2h3 c8e6 c4e6 f7e6 f1e1 d8e8 c2c3 e8g6 b2b4 f8f7 a3a4 f6d7 e1f1 g6h5 c1e3 d6d5 d1e2 d7f6 f1e1 a8f8 e3d2 f6d7 f3h2 h5h4 e1f1 c6e7 e2g4 h4g4 h2g4 h6h5 g4e3 d7f6');
INSERT INTO evaluations VALUES(4480,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4480,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(4480,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(4480,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4480,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4480,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4480,36,'cp 0','d7d5 e4d5 f6d5 a1c1 d5f6 d2b3 d8d6 c4b5 c6e7 b3c5 c7c6 b5c4 b7b6 c5d3 c8g4 d3e5 g4f3 e5f3 g7g6 f1e1 a8c8 c4d3 g8g7 g2g3 c8c7 d1b3 d6d5 d3c4 d5d6');
INSERT INTO evaluations VALUES(3729,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3729,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3729,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3729,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3729,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3729,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3729,36,'cp 0','e4d5 f6d5 a1c1 d5f6 d2b3 d8d6 f1e1 b7b6 f3e5 c8b7 d1e2 a8d8 c4b5 c6d4 b3d4 d6d4 b5c6 b7c8 c6b5');
INSERT INTO evaluations VALUES(3726,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3726,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3726,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3726,36,'cp 0','f6d5 a1c1 d5f6 d2b3 d8d6 f1e1 b7b6 f3e5 c8b7 d1e2 a8d8 c4b5 c6d4 b3d4 d6d4 b5c6 b7c8 c6b5');
INSERT INTO evaluations VALUES(12549,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(12549,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(12549,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(12549,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12549,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(12549,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(12549,36,'cp 41','d7d6 e1g1 f8e7 a2a3 e8g8 a1b1 h7h6 d2d3 a8b8 b2b4 a7a6 b4c5 d6c5 c1b2 c8e6 b2a1 d8c7 f3d2 f8d8 b1b2 c7a5 f1e1 d8c8 b2b1 b7b5 c3d5 e7f8 f2f4 f6g4 e2e3 a5d8 h2h3 g4f6 d5f6 g7f6 g3g4 b5c4 b1b8 c8b8 d3c4 c6a5 f4e5');
INSERT INTO evaluations VALUES(13862,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(13862,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(13862,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(13862,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(13862,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(13862,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(13862,36,'cp 45','e1g1 f8e7 d2d3 e8g8 a2a3 h7h6 a1b1 a7a6 f3d2 a8b8 b2b3 c8f5 b3b4 f5e6 b4c5 d6c5 g2c6 b7c6 b1b8 d8b8 d1a4 e6h3 f1d1 b8c8 d2b3 c8e6 f2f3 f6d7 a4a6 e5e4 d3e4 e7f6 e4e5 f6e5 c3e4 h3f5 b3c5 d7c5 e4c5');
INSERT INTO evaluations VALUES(13861,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(13861,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(13861,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(13861,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(13861,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(13861,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(13861,36,'cp 41','f8e7 d2d3 e8g8 a1b1 h7h6 a2a3 a7a6 f3d2 c8e6 c3d5 a8b8 b2b3 d8d7 e2e3 e7d8 c1b2 e6g4 f2f3 g4h5 d1e1 h5g6 e1e2 f8e8 d5f6 d8f6 d2e4 f6d8 f3f4 f7f5 e4c3 b7b5 h2h3 c6e7 g1h2 g6f7 g3g4 e7g6 g2d5');
INSERT INTO evaluations VALUES(8136,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(8136,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(8136,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(8136,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(8136,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(8136,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(8136,36,'cp 57','d2d3 h7h6 a2a3 e8g8 a1b1 a7a6 f3d2 a8b8 b2b4 c8e6 b4c5 d6c5 g2c6 b7c6 b1b8 d8b8 d1a4 b8b6 c3d1 f8a8 c1b2 f6d7 b2c3 b6c7 d1e3 e7d8 f1b1 a6a5 e3g2 e6h3 f2f3 f7f5 g2e3 d7b6 a4c2 d8g5 e3d1 b6d7 a3a4');
INSERT INTO evaluations VALUES(8137,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(8137,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(8137,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(8137,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(8137,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(8137,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(8137,36,'cp 40','h7h6 a2a3 e8g8 a1b1 a8b8 b2b4 a7a6 f3d2 c8e6 b4c5 d6c5 g2c6 b7c6 b1b8 d8b8 d1a4 b8b6 c3d1 a6a5 c1b2 f6d7 b2c3 b6c7 d1e3 e7g5 e3g2 f8b8 d2e4 g5e7 f1a1 e7d8 a4c2 e6h3 a1b1 b8b1 c2b1 f7f5 e4d2 h3g4 g2e3');
INSERT INTO evaluations VALUES(13334,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13334,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13334,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13334,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13334,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13334,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13334,36,'cp 7','b1c3 g7g6 d2d4 c6d4 f3d4 c5d4 d1d4 f8g7 e1g1 c8d7 a2a4 e8g8 d4b4 a7a5 b4b3 d7c6 c1e3 f6d7 c3d5 a8b8 d5b6 d7c5 e3c5 d6c5 a1d1 d8c7 b6d5 c7d6 f1e1 c5c4 b3c4 g7b2 e4e5 b2e5 d5b6 d6c7 g2c6 c7b6 e1e5 b6c6 c4c6 b7c6 e5e7');
INSERT INTO evaluations VALUES(13305,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(13305,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13305,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(13305,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(13305,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(13305,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(13305,36,'cp 0','g7g6 d2d4 c6d4 f3d4 c5d4 d1d4 f8g7 e1g1 e8g8 a2a4 d8c7 d4d3 c8e6 c3d5 f6d5 e4d5 e6f5 g2e4 f5e4 d3e4 a8b8 e4e2 a7a6 a4a5 b7b5 a5b6 b8b6 a1a6 b6a6 e2a6 c7c2 a6b7 c2e4 c1g5 g7f6 g5h6 f6g7');
INSERT INTO evaluations VALUES(13614,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13614,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13614,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13614,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13614,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(13614,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13614,36,'cp 0','e2e4 d5e4 d3e4 f6e4 b1a3 f8e7 d1e2 e4f6 a3c4 e8g8 c2c3 f6d5 f3e5 d8c7 f1d1 c6e5 c4e5 e7d6 e5c4 d6e7');
INSERT INTO evaluations VALUES(19074,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(19074,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(19074,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(19074,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(19074,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(19074,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(19074,36,'cp 47','b1c3 f8e7 c3d5 g8f6 d5f6 e7f6 g1e2 d7d6 e2c3 e8g8 c3d5 f6g5 c1g5 d8g5 c2c3 c8e6 g2g3 a8d8 h2h4 g5h5 d5e3 c6e7 f1g2 f7f5 e4f5 e7f5 e3f5 e6f5 d3d5 g8h8 e1g1 b7b6 a1d1 h7h6 c3c4 a7a5 d5d2 f5e6');
INSERT INTO evaluations VALUES(21694,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(21694,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(21694,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(21694,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(21694,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(21694,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(21694,36,'cp -99','e5e4 f3e5 c6e7 f2f3 f7f6 e5g4 c8g4 f3g4 h7h5 g4g5 h5h4 c2c4 h4g3 h2g3 h8h1 g2h1 e7f5 g3g4 f5g3 d1a4 c7c6 c4d5 b7b5 a4a6 d8d5 b1c3 f8b4 a6b7 a8d8 b7g7 d5d4 c1e3 d4e3 g7g8 b4f8 g8g6 e8d7 a1d1 f8d6 h1e4 d7c7 d1d3 e3c1 d3d1 c1g5 g6g5 f6g5 e4g2 b5b4 c3a4 d8h8 g2f3');
INSERT INTO evaluations VALUES(21615,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(21615,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(21615,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(21615,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(21615,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(21615,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(21615,36,'cp -75','f3e5 c6e7 f2f3 f7f6 e5g4 c8g4 f3g4 h7h5 g4g5 d8d7 c2c4 h5h4 c4d5 f6f5 b1c3 e7d5 d1b3 e8c8 b3d5 d7d5 c3d5 d8d5 g2h3 f8b4 e1f2 b4d6 c1f4 g8e7 e2e3 c7c6 a1d1 c8d7 h3f1 b7b5 f1e2 a7a5 f4d6 d5d6 g3h4 h8h4 f2g3 h4h8 h2h4 e7d5 g3f2');
INSERT INTO evaluations VALUES(21634,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(21634,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(21634,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(21634,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(21634,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(21634,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(21634,36,'cp -93','c6e7 f2f3 f7f6 e5g4 c8g4 f3g4 h7h5 g4g5 d8d7 c2c4 h5h4 c4d5 f6f5 d1b3 e7d5 b1c3 e8c8 c3d5 d7d5 g2h3 d5d7 c1f4 f8d6 f4d6 c7d6 e1c1 g8e7 c1b1 c8b8 d1g1 a7a6 e2e3 h4g3 h2g3 d6d5 h3f1 b8a7 f1e2 d8c8 b3a3 d7c7 a3c5 c7c5 d4c5 h8h1 g1h1 c8c5 h1h7');
INSERT INTO evaluations VALUES(21987,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(21987,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(21987,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(21987,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21987,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21987,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21987,36,'cp 13','g1f3 e5f4 d2d4 g7g5 d4d5 c6e5 f1b5 c8d7 b5d7 e5d7 h2h4 f8g7 h4g5 d8e7 d1e2 e8c8 c1f4 g7c3 b2c3 d8e8 e1c1 e7e4 e2e4 e8e4 d1e1 g8f6 g5f6 e4f4 h1h7 h8f8 h7f7 f4f6 f7g7 d7b6 e1d1 a7a6 d1d4 f6f4 d4f4 f8f4 g7g8 c8d7 g8g5 b6c4 c1d1');
INSERT INTO evaluations VALUES(22024,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(22024,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22024,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22024,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22024,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22024,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22024,36,'cp 0','b1c3 g7g6 h2h4 g8f6 c3d5 f8g7 c1g5 c8e6 h4h5 h7h6 g5f6 e6d5 e4d5 d8f6 d5c6 f6b2 e1d2 b2a1 c6b7 a8b8 c4c6 e8d8 g1f3 b8b7 c6b7 a1c3 d2c1 c3a1');
INSERT INTO evaluations VALUES(22679,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(22679,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(22679,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(22679,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(22679,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(22679,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(22679,36,'cp 80','c6e5 e2e4 f8c5 c1f4 d7d6 d1d2 g8f6 e1c1 c8d7 g1f3 d8e7 f3d4 h7h6 f2f3 a7a6 d4b3 c5b6 h2h4 e8c8 c1b1 d7e6 h4h5 c8b8 f4g3 f6d7 f1e2 d8e8 f3f4 e5c4 e2c4 e6c4 e4e5 c4e6 e5d6 e7d6 d2d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(23087,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(23087,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(23087,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(23087,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(23087,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(23087,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(23087,36,'cp 58','f2f4 e5c6 e2e4 f8c5 g1f3 g8f6 e4e5 f6g4 f1c4 e8g8 d1e2 d7d6 h2h3 g4h6 c1e3 c5e3 e2e3 h6f5 e3f2 d6e5 f4e5 c8e6 c4d3 c6d4 e1c1 d4f3 f2f3 d8g5 c1b1 f5g3 c3e4 g3e4 f3e4 g7g6 h3h4 g5h5 d3e2 e6f5 e4e3 f5g4 e2g4 h5g4 h4h5 a8d8');
INSERT INTO evaluations VALUES(24903,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(24903,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(24903,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(24903,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(24903,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(24903,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24903,36,'cp 43','d1d3 e7e6 e1g1 f8d6 f4g5 h7h6 g5f6 d8f6 c2c4 d5c4 d3c4 e8g8 b1c3 e6e5 d4d5 c6e7 e3e4 a8c8 c4b3 c8b8 a1d1 b7b5 d1d2 a7a6 h2h3 f6g6 f1e1 f7f6 b3c2 f6f5 e4f5 g6f5 c3e4 e7g6 d2d1 f5f7');
INSERT INTO evaluations VALUES(26889,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(26889,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(26889,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(26889,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(26889,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(26889,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(26889,36,'cp 126','b1d2 d8e7 g2g3 c6e5 f3e5 e7e5 h2h3 g4e6 f1g2 e8c8 d2f3 e5c5 e1g1 g8f6 c1f4 f8d6 f4d6 c5d6 a1c1 c7c5 b2b4 e6f5 d1a4 c8b8 b4c5 d6c5 f3d2 f5d7 a4d1 d7f5 d1b3 d8d7 g3g4 f5g6 c1b1 c5e7 b3f3 h8e8 c4c5 g6b1 f1b1 f6d5 c5c6');
INSERT INTO evaluations VALUES(26888,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(26888,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(26888,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(26888,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(26888,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(26888,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(26888,36,'cp 129','d8e7 g2g3 c6e5 f3e5 e7e5 h2h3 g4e6 f1g2 e8c8 e1g1 g8e7 d2f3 e5h5 g3g4 h5a5 b2b4 a5a6 f3d4 e6c4 c1b2 c4d5 d1c2 d5g2 a1c1 e7d5 g1g2 d8d7 e2e3 a6g6 d4f5 h7h5 c2e4 h5g4 h3g4 g6e6 e4e6 f7e6 f5d4 f8d6 f1h1 h8e8 d4b5 c8b8');
INSERT INTO evaluations VALUES(29544,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29544,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29544,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(29544,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29544,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29544,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(29544,36,'cp 8','e7e6 c2c4 c7c5 e2e3 b8c6 c4d5 e6d5 d4c5 f8c5 a1c1 c5b6 f1b5 e8g8 e1g1 a8c8 a2a3 a7a6 b5a4 d8e7 b2b4 a6a5 d1b3 a5b4 a4c6 b7c6 a3b4 h7h6 f3d4 b6d4 e3d4 f8e8 f4e5 f6d7 e5g3 e7e2 d2f3 e2b5 b3c3 b5d3 c3d3 f5d3 f1e1');
INSERT INTO evaluations VALUES(30100,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(30100,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(30100,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(30100,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(30100,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(30100,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30100,36,'cp 11','c2c4 c7c5 e2e3 b8c6 c4d5 e6d5 d4c5 f8c5 a1c1 c5b6 f1b5 e8g8 e1g1 a8c8 a2a3 h7h6 h2h3 f6e4 d2e4 f5e4 b5d3 d8e7 d3e4 e7e4 f4d6 f8d8 d6c5 c6e5 f3e5 c8c5 c1c5 b6c5 e5f3 c5b6 d1d2 d8c8 f1c1 c8c4 b2b3 c4c5 c1c5 b6c5 d2c3');
INSERT INTO evaluations VALUES(30144,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(30144,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(30144,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(30144,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(30144,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(30144,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(30144,36,'cp 131','c4d5 e6d5 g5f6 g7f6 d1b3 c7c6 b3b7 d8b6 b7b6 a7b6 e2e3 b8d7 c3e2 f8d6 e2c1 e8e7 f1d3 f5e6 g1e2 f6f5 e1g1 a8a5 e2g3 e7f6 c1e2 h8a8 a2a3 c6c5 a1b1 c5c4 d3c2 b6b5 f2f3 d7b6 e2c3 b5b4 a3b4 d6b4 e3e4 f5e4 f3e4 f6g7 e4e5 b6d7 g1h1');
INSERT INTO evaluations VALUES(30277,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(30277,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(30277,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(30277,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(30277,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(30277,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(30277,36,'cp 0','c7c5 f3g5 d8d7 f2f4 b8c6 f4e5 d6e5 e1g1 f8e7 g1h1 h7h6 g5f3 e8g8 c1e3 a8d8 d1e1 c6d4 e1d1');
INSERT INTO evaluations VALUES(33103,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(33103,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(33103,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(33103,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(33103,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(33103,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(33103,36,'cp -32','d6b4 b1d2 b4b2 c2c4 g8f6 f1e2 e8g8 e1g1 b2a3 d1c2 b7b6 c4d5 f6d5 f1c1 c8b7 f3e5 f7f6 e5f3 b8c6 e2d3 f6f5 d3b5 c6a5 d2c4 a5c4 c2c4 a3e7 b5c6 b7c6 c4c6 a8c8 h2h3 e7d6 a2a3 h7h6 a3a4 g8h7 f3e5 d6c6 e5c6');
INSERT INTO evaluations VALUES(39140,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(39140,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(39140,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(39140,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(39140,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(39140,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(39140,36,'cp 78','d7d5 d2d4 c5b4 e4d5 e6d5 f1d3 g8e7 e1g1 e8g8 c3e2 b4d6 e2g3 e7g6 c2c3 b8d7 c1g5 d7f6 d1c2 c7c5 d4c5 d6c5 a1e1 d8d6 g3f5 c8f5 d3f5 a8e8 g2g3 h7h6 g5f6 d6f6 g1g2 g6e5 f5h7 g8h8 f3e5 e8e5 e1e5 f6e5 h7d3 c5d6 h2h4 f8e8');
INSERT INTO evaluations VALUES(41335,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41335,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41335,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(41335,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(41335,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(41335,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(41335,36,'cp 9','e3e4 e8g8 h2h3 c7c5 d4c5 d6c5 f1d3 b8c6 e1g1 b7b6 e4e5 f6d5 c3d5 d8d5 f1e1 c8b7 d3e4 d5d1 e1d1 f8d8 c1g5 g8f8 g5f4 f8e8 c2c3 d8d1 a1d1 a8d8 d1d8 c6d8 e4b7 d8b7 g1f1 e8d7 f1e2 b7a5 e2d3 d7e6 d3e4 f7f6 e5f6 e7f6 g2g4 a5c6');
INSERT INTO evaluations VALUES(45375,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45375,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(45375,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(45375,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(45375,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45375,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45375,36,'cp 11','d4c5 a7a5 b2b4 e7e6 b1d2 a5b4 c3b4 b7b6 c5b6 f8b4 a1b1 d8b6 a2a3 b4d2 c1d2 b6c7 d2b4 c8d7 f1d3 b8c6 d1c1 a8c8 c1d2 c6b4 a3b4 e8g8 g1e2 e6e5 e1g1 c7d6 h2h3 c8b8 f1d1 f8c8 d1c1 c8c1 e2c1 d5d4 e3d4 d6d4 c1e2 f6e4 d2e3 d4e3 f2e3 e4d6 e2c3');
INSERT INTO evaluations VALUES(46091,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(46091,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46091,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(46091,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46091,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(46091,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46091,36,'cp 0','f8c5 c4d5 e6d5 e2e3 b8c6 a2a3 e8g8 b2b4 c5d6 c3b5 d6e7 f1d3 f6e4 c1b2 a7a5 e1g1 a5b4 a3b4 a8a1 d1a1 c6b4 d3e2 e7f6 b2f6 e4f6 b5d4 b4c6 h2h3 h7h6 f1b1 f6e4 e2d3 g8h8 d4e2 d8d6 b1b6 d6c7 b6b1');
INSERT INTO evaluations VALUES(48005,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(48005,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(48005,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(48005,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(48005,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(48005,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(48005,36,'cp 43','d2d4 f8e7 d4e5 d6e5 d1d8 e7d8 f1e1 e4d6 e1e5 d8e7 c4d5 c7c6 d5b3 c8g4 b1d2 b8d7 e5e2 d7c5 h2h3 c5b3 a2b3 g4e6 d2f1 d6f5 g2g4 f5h4 f3d4 e8g8 d4e6 f7e6 c1e3 a7a6 f1g3 a8d8 e3b6 d8d5 e2e6 e7f6 a1b1 h4f3 g1g2 f3h4 g2f1 h4f3 b6e3 f6g5 b1a1 g5e3 e6e3 f3d4 c2c4 d4c2 c4d5 c2e3 f1g1 e3c2');
INSERT INTO evaluations VALUES(54706,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(54706,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(54706,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(54706,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(54706,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(54706,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(54706,36,'cp -11','b8c6 f4g3 e7e5 d4e5 f6e5 e2e4 g8f6 e4d5 d8d5 d1d5 f6d5 f1c4 d5f4 g3f4 e5f4 b1c3 c8d7 e1c1 e8c8 g1f3 f8c5 h1f1 h7h6 h2h3 a7a6 c4f7 c5a7 f7h5 h8f8 a2a3 d7e6 f1e1 e6f5 e1e2 c6d4 f3d4 a7d4 c3e4 d4b6 d1d8 f8d8');
INSERT INTO evaluations VALUES(20847,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(20847,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(20847,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(20847,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(20847,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(20847,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(20847,36,'cp -12','f4g3 e7e5 d4e5 f6e5 e2e4 g8f6 e4d5 d8d5 d1d5 f6d5 f1c4 d5f4 g3f4 e5f4 b1c3 c8d7 e1c1 e8c8 g1f3 f8b4 c3d5 b4d6 f3g5 c6e5 c4b3 f4f3 g2f3 d8f8 g5e4 f8f3 e4d6 c7d6 d5c3 c8c7 h1g1 g7g6 c3e4 d7c6 e4d6 h8d8 g1e1 d8d6 e1e5 d6d1 c1d1 f3h3 d1e2 h3h2 a2a3 h2h1 b3d5');
INSERT INTO evaluations VALUES(55203,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(55203,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(55203,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(55203,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(55203,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(55203,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(55203,36,'cp -39','b8c6 e2e3 e7e5 e3d4 e5d4 f1e2 g8e7 h2h4 h7h5 b1d2 e7g6 e1g1 f8e7 a2a3 a7a5 f1e1 c8g4 g2g3 f7f5 f3g5 d8d7 f2f4 e8g8 d2f3 a8e8 c1d2 a5a4 g1f2 c6d8 b2b4 e7g5 h4g5 b7b6 b4b5 d8b7 d2b4 b7c5 b4c5 b6c5 d1a4');
INSERT INTO evaluations VALUES(21106,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(21106,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(21106,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(21106,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(21106,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(21106,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(21106,36,'cp -39','g2g3 e7e5 f1g2 f8c5 e1g1 g8e7 e2e3 c8g4 h2h3 g4h5 e3d4 h5f3 g2f3 c5d4 b1a3 e8g8 a3c2 a7a5 c2d4 d8d4 c1g5 a8b8 g5e7 c6e7 f1e1 b8d8 d1d2 b7b6 b2b4 a5b4 d2b4 e7f5 g1g2 f8e8 a1d1 d4d6 b4d6 d8d6');
INSERT INTO evaluations VALUES(1972,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(1972,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(1972,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(1972,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(1972,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(1972,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(1972,36,'cp 187','b5c4 g8e7 d2d3 g5f6 a2a4 d7d6 c2c3 e7c6 b1a3 f8e7 c1d2 e8g8 c3d4 c6d4 d2c3 d6d5 c4d5 c7c6 d5b3 e7a3 a1a3 f6d6 b3c4 c8e6 c3d4 d6d4 c4e6 f7e6 d1b3 d4d7 a4a5 a8d8 b3b4 e6e5 a3c3 f8f7 h2h3 d7e7 b4e7 f7e7');
INSERT INTO evaluations VALUES(1999,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(1999,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(1999,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(1999,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(1999,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(1999,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(1999,36,'cp 163','g5f6 d2d3 g8e7 a2a4 d7d6 b1a3 a7a6 c2c3 c8e6 c4e6 f7e6 d1b3 e8c8 a3c2 d6d5 c2d4 e7g6 d4f3 g6e5 f3e5 f6e5 f2f4 e5h5 d3d4 h5e2 e4e5 f8e7 b3d1 e2d1 f1d1 c7c5 c1e3 c5d4 e3d4 d8f8 g2g3 g7g5 f4g5 e7g5 g1g2 h7h5 d4c5 f8g8 d1f1 h8h7 f1f3 h5h4 a1f1 h4g3 f3f8 g8f8 f1f8 c8c7 g2g3');
INSERT INTO evaluations VALUES(1966,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(1966,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(1966,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(1966,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(1966,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(1966,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(1966,36,'cp 192','d2d3 g8e7 a2a4 d7d6 b1d2 h7h5 c2c3 c8g4 d1b3 e8c8 a4a5 h5h4 h2h3 g4e6 a5a6 b7b6 f1e1 d4c3 b2c3 d6d5 e4d5 e6d5 c4d5 e7d5 d3d4 d5f4 d2e4 f6f5 c1f4 f5f4 e1e3 f4f5 a1e1 f5d5 c3c4 d5d4');
INSERT INTO evaluations VALUES(1963,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(1963,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(1963,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(1963,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(1963,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(1963,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(1963,36,'cp 188','g8e7 a2a4 d7d6 c2c3 e7c6 b1a3 f8e7 c1d2 e8g8 a3b5 e7d8 c3d4 a7a6 b5c3 c6d4 c3d5 f6h4 d2c3 d4c6 f2f4 c6e7 d5e3 c7c6 g2g3 h4h6 e3g4 c8g4 d1g4 d6d5 c4b3 f7f5 g4e2 g8h8 a1e1 h6d6 e4e5 d6e6 a4a5 c6c5 d3d4 c5c4 b3a4 a8c8');
INSERT INTO evaluations VALUES(3696,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(3696,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(3696,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(3696,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(3696,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(3696,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(3696,36,'cp 71','d1c2 b4d6 e1g1 c6b4 c2b1 b4c6 h2h3 e6e5 f3e5 c6e5 d4e5 d6e5 b1c2 c7c6 f2f4 e5c7 c4d3 c8e6 f1f3 d8e7 e3e4 g8h8 a1e1 a8d8 e4e5 f6d5 a2a3 f7f5 c3d5 e6d5 f3g3 c7b6 g1h2 b6f2 d2b4 f2g3 h2g3 e7e6 b4f8 d8f8 g3h2 a7a6 c2c5 f8e8 c5e3');
INSERT INTO evaluations VALUES(7409,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(7409,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(7409,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(7409,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(7409,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(7409,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(7409,36,'cp 18','c5b6 c1e3 e8g8 b1d2 h7h6 h2h3 e5d4 c3d4 c6a5 b3c2 c8b7 f1e1 f8e8 b2b4 a5c4 d2c4 b5c4 d4d5 c7c6 c2a4 a8c8 e3b6 d8b6 f3d4 c6d5 a4e8 c8e8 a1c1 e8e4 d4f3 b7c6 a2a3 c6b5 e1e4 d5e4 d1d4 b6c7 f3e1 d6d5 e1c2 g8h7 c2e3 c7d7 g1f1 d7e6 f1e1 b5a4 e1e2 a4b5 e2d2 e6d7 c1a1 b5a4 d2c3');
INSERT INTO evaluations VALUES(7425,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(7425,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(7425,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(7425,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(7425,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(7425,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(7425,36,'cp 27','a2a4 c8b7 c1g5 h7h6 g5f6 d8f6 b3d5 e8g8 b1a3 e5d4 c3d4 a8a7 a3c2 b5a4 a1a4 a6a5 f1e1 c6e7 d5b7 a7b7 c2e3 f8b8 g2g3 c7c6 e3c4 f6e6 b2b3 b6d8 a4a1 d6d5 e4d5 e6d5 e1e5 d5d7 c4a5 d8a5 e5a5 b7b3 a5a8 g8h7 d1c2 e7g6 a8a7');
INSERT INTO evaluations VALUES(8490,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(8490,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(8490,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(8490,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(8490,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(8490,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(8490,36,'cp 62','f6h5 b1d2 h5f4 e3f4 c6b8 h2h4 b8d7 g2g3 c7c5 d4c5 d7c5 d2b3 e8g8 b3c5 e7c5 e1g1 d8b6 d1e2 c8d7 f1e1 f8d8 a1d1 a8c8 h4h5 h7h6 f3e5 d7a4 d1d2 d5d4 e2e4 f7f5 e4e2 a4e8 e5g6 d4c3 b2c3 e8f7 g6e5 c8c7');
INSERT INTO evaluations VALUES(12043,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12043,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(12043,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12043,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12043,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12043,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(12043,36,'cp 0','g7g6 b2b3 f8g7 c1b2 e8g8 e1g1 e7e5 f1e1 d8c7 d1d2 f8d8 c3a4 f6h5 b2c3 a7a5 d2e3 c8e6 d3f1 h5f4 e3b6 c7e7 c3a5 e7g5 g1h1 d8b8 b6c7 b8c8 c7b6');
INSERT INTO evaluations VALUES(12026,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12026,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(12026,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12026,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(12026,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(12026,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(12026,36,'cp 13','b2b3 f8g7 c1b2 e8g8 e1g1 f6d7 c3a4 d8a5 b2g7 g8g7 h2h3 f7f6 f2f4 d7b6 a4b6 a5c5 g1h1 a7b6 a2a4 e7e5 f4e5 c5e5 d1d2 c8e6 a1d1 f8e8 d2f2 e8b8 d3e2 b6b5 a4b5 c6b5 f2f4 e5f4 f1f4 b8c8 d1d6 c8c2 d6e6 c2e2 e6e7 g7f8 e7h7');
INSERT INTO evaluations VALUES(12398,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12398,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12398,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(12398,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12398,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(12398,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(12398,36,'cp 1','d7d5 c4d5 e6d5 d2d4 c5d4 e3d4 h7h6 a2a3 f8d6 g1e2 c8g4 e1g1 e8g8 c3d5 f6d5 g2d5 d6c7 d5g2 c6d4 d1d4 g4e2 d4d8 a8d8 f1e1 e2a6 g2f3 c7b6 c1e3 b6e3 e1e3 d8d7 a1e1 g7g6 h2h4 g8g7 e1c1 f8d8 e3c3 b7b6 c3c7 a6b5 c7d7');
INSERT INTO evaluations VALUES(13540,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13540,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13540,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(13540,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(13540,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(13540,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(13540,36,'cp 0','c4d5 e6d5 d2d4 c5d4 e3d4 h7h6 a2a3 c8g4 g1e2 f8d6 e1g1 e8g8 c3d5 f6d5 g2d5 f8e8 c1e3 d6g3 d5f7 g8f7 h2g3 d8d7 d1b3 g4e6 b3a4 a8d8 a1c1 e6h3 a4b3 h3e6');
INSERT INTO evaluations VALUES(13518,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13518,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13518,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13518,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(13518,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13518,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13518,36,'cp -3','e6d5 d2d4 c5d4 e3d4 h7h6 a2a3 c8g4 g1e2 f8d6 e1g1 e8g8 c3d5 f6d5 g2d5 d6c7 d5g2 c7b6 h2h3 g4e2 d1e2 c6d4 e2g4 d8c8 g4c8 d4e2 g1h2 a8c8 g2b7 c8c1 a1c1 e2c1 f1c1 b6f2 c1c2 f2e3 b2b4 f8d8 c2c8 e3b6 c8d8 b6d8 b7d5 g7g6 h2g2');
INSERT INTO evaluations VALUES(13766,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13766,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13766,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13766,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13766,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13766,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(13766,36,'cp 0','d2d4 c5d4 e3d4 h7h6 a2a3 f8d6 g1e2 e8g8 e1g1 d6c7 c1f4 c7b6 f1e1 f8e8 f4e3 c8g4 h2h3 g4f5 e2f4 c6e7 h3h4 f5e4 g2h3 d8d6 c3b5 d6d8');
INSERT INTO evaluations VALUES(13745,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13745,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13745,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13745,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13745,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13745,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13745,36,'cp 0','c5d4 e3d4 h7h6 a2a3 f8d6 g1e2 e8g8 e1g1 d6c7 c1f4 c7b6 f4e3 c8g4 h2h3 g4f5 e2f4 f5e4 f4h5 f6h5 d1h5 f7f5 e3h6 g7h6 c3e4 f5e4 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(18959,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(18959,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(18959,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(18959,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(18959,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(18959,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(18959,36,'cp 76','c1e3 g8f6 b1c3 h7h6 f3d2 d7d6 c5d6 d8d6 c3b5 d6e7 c2c4 c8e6 g2g3 e6c4 d2c4 e7b4 b5c3 b4c4 f1g2 f8c5 e3c5 c4c5 e1g1 e8g8 e2e3 f8d8 d1c2 a8c8 a2a3 f6d5 a1c1 d5c3 c2c3 c5c3 c1c3 a7a5 c3c5 g7g6 f1c1 d8d2 c5b5 c8c7 g2d5 g8f8 c1c5 f8e7 d5c6 b7c6');
INSERT INTO evaluations VALUES(22057,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(22057,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(22057,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(22057,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(22057,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(22057,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(22057,36,'cp 79','a7a6 b5d3 g8f6 c2c3 g7g6 a2a4 f8g7 e1g1 a6a5 f1e1 e6e5 b2b4 a5b4 d4d5 c6e7 c3b4 e8g8 d3f1 c7c6 d5c6 e7c6 b4b5 c6a5 c1a3 f8e8 b1c3 g7f8 a3b4 b7b6 c3d5 f6d5 e4d5 c8b7 h2h3');
INSERT INTO evaluations VALUES(24661,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24661,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(24661,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(24661,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(24661,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(24661,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24661,36,'cp 12','h2h3 g4f3 d1f3 e7e6 c4b5 d8c7 f3g3 f6h5 b5c6 c7c6 g3h4 h5f6 d2d3 f8e7 a2a4 e8g8 c1d2 h7h6 h4g3 f6h5 g3f3 h5f6 f1e1 a8d8 c3e2 c6c7 e2g3 d6d5 f3e2 d5e4 d3e4 d8d4 d2c3 d4d7 a4a5 f8d8');
INSERT INTO evaluations VALUES(24660,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(24660,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24660,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24660,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24660,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24660,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24660,36,'cp 13','g4f3 d1f3 e7e6 c4b5 d8c7 d2d3 f8e7 f3g3 e8g8 c1h6 f6h5 g3g4 c6d4 g4h5 g7h6 b5a4 e7f6 h5h6 c7e7 h6h5 g8h8 f2f4 f6g7 a1e1 a8b8 a4b3 b7b5 e4e5 d6e5 c3e4 c5c4 d3c4 b5c4 b3c4 d4c2 e4g5 h7h6 e1e5 b8b2');
INSERT INTO evaluations VALUES(24941,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(24941,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(24941,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24941,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(24941,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(24941,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(24941,36,'cp 50','b1d2 a7a6 d1a4 b7b5 a4c2 e7e6 a2a4 b5b4 f1d3 f8d6 h2h3 g4h5 e1g1 b4c3 b2c3 c6a5 c3c4 d5c4 d3c4 a5c4 c2c4 e8g8 c1a3 d8e7 c4a2 h5f3 d2f3 c7c5 f1c1 c5d4 f3d4 f8c8 d4c6 e7d7 a3d6 d7d6 a2c4 g8h8 a4a5 h7h6 a1b1 f6d5 c4c5 d6c5 c1c5');
INSERT INTO evaluations VALUES(24939,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24939,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(24939,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(24939,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24939,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(24939,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(24939,36,'cp 48','a7a6 d1a4 b7b5 a4c2 e7e6 a2a4 b5b4 f1d3 f8d6 e1g1 b4c3 b2c3 e8g8 h2h3 g4h5 a4a5 f8e8 c2a4 d8d7 c3c4 h5g6 d3g6 h7g6 c1b2 a8b8 f1b1 b8b4 a4a2 e8b8 b2c3 b4b1 a1b1 b8b1 a2b1 d7c8 b1b3 d5c4 d2c4 c8b8 b3b8 c6b8 f3e5 f6d7 e5d3 d7f6 f2f3 b8c6 e3e4 f6d7 d4d5 e6d5');
INSERT INTO evaluations VALUES(30925,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30925,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30925,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(30925,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(30925,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30925,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(30925,36,'cp 0','g8f6 b5d7 d8d7 d2d3 b8c6 e1g1 e7e6 h2h3 f8e7 c1f4 b7b6 b1d2 e8g8 f1e1 d7b7 d1e2 d6d5 f4g3 a8d8 c2c3 h7h6 f3e5 c6e5 g3e5 d8c8 a1d1 c8d8');
INSERT INTO evaluations VALUES(36846,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(36846,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(36846,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36846,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(36846,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36846,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36846,36,'cp 0','d6e5 c1g5 d8d1 a1d1 b8a6 g5e3 b7b6 e1g1 c8b7 e2d3 f6g4 e3c1 a6c5 d3c2 g4f6 c3d5 c5e4 f3e5 b7d5 c4d5 f8e8 e5c6 e4d6 g2g3 a7a5 a2a4 e8e2 c2d3 e2e8');
INSERT INTO evaluations VALUES(42130,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42130,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(42130,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42130,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(42130,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(42130,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(42130,36,'cp 20','h7h6 g5f6 d8f6 e4e5 f6e7 a2a3 b4c3 b2c3 c7c5 d1g4 e8g8 c4d3 b8d7 g1e2 f8d8 g4f4 d7f8 h2h4 f8g6 f4g3 c8d7 h4h5 g6f8 e1g1 a7a6 a1b1 b7b5 f2f4 c5d4 c3d4 g8h8 f4f5 e6f5 d3f5 d7f5 f1f5 f8e6 b1f1 e6g5 g1h1 e7e6 g3g4 b5b4');
INSERT INTO evaluations VALUES(41661,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(41661,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(41661,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(41661,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(41661,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(41661,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(41661,36,'cp 37','g5f6 d8f6 e4e5 f6e7 a2a3 b4c3 b2c3 b8c6 g1e2 b7b6 e2g3 c8b7 e1g1 e8c8 d1e2 c8b8 c4a6 b7a8 a3a4 c6a5 f1b1 a8d5 g3f1 f7f6 e5f6 g7f6 c3c4 d5a8 d4d5 h8e8 b1e1 a5b3 a1b1 b3d4 e2e4 e7c5 e1d1 e6e5 f1g3 e8g8 g1h1 a8b7 a6b7 b8b7 e4d3');
INSERT INTO evaluations VALUES(41635,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(41635,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(41635,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(41635,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(41635,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(41635,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41635,36,'cp 42','d8f6 e4e5 f6e7 a2a3 b4c3 b2c3 c8d7 g1e2 c7c5 e2g3 e8g8 e1g1 c5d4 c3d4 f8d8 d1g4 d7a4 g3e4 b8c6 e4d6 d8d6 e5d6 e7d6 f1b1 c6d4 b1b7 d6c6 b7f7 g8f7 g4d4 a8c8 d4f4 f7g8 c4e6 c6e6 f4a4 e6c4 a4d1 c4c2 d1c2 c8c2 g2g3 a7a5 a1e1 c2a2 e1e3');
INSERT INTO evaluations VALUES(33410,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(33410,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(33410,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(33410,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(33410,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(33410,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(33410,36,'cp 32','e4e5 f6e7 a2a3 b4c3 b2c3 b8c6 g1e2 b7b6 e2g3 c8b7 e1g1 e8c8 d1e2 c8b8 c4a6 b7a8 a3a4 h6h5 a4a5 f7f5 e5f6 g7f6 h2h4 e6e5 f1e1 e7d7 d4e5 c6e5 e2a2 d7d5 a2d5 d8d5 a6f1 d5c5 g3e4 a8e4 e1e4 c5c3 a5a6');
INSERT INTO evaluations VALUES(43149,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(43149,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(43149,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(43149,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(43149,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(43149,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(43149,36,'cp -38','d7d5 c4b5 c7c6 b5e2 c6c5 e1g1 e8g8 d4c5 d8c7 a2a3 a7a5 c1d2 c7c5 d2c3 c8g4 c3d4 c5c7 d4e5 c7c8 h2h3 g4f3 e2f3 b8d7 e5h2 e7e6 b1c3 d7b6 c3b5 c8c6 b5d4 c6c5 a3a4 f8d8 d1e2 a8c8 c2c3 f6e4');
INSERT INTO evaluations VALUES(48061,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(48061,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(48061,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(48061,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(48061,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(48061,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(48061,36,'cp 75','d2d4 b8d7 f1d3 e6e5 e1g1 f8e7 h2h3 e8g8 a2a4 f8e8 f1e1 e5d4 f3d4 h7h6 d3c4 e7f8 c1f4 c7c6 f4g3 a7a6 c4a2 d7c5 d1f3 c5e6 a2e6 c8e6 g3h4 g7g5 h4g3 f6d7 a1d1 d8f6 f3f6 d7f6 d4e6 e8e6 f2f3 a8d8 a4a5 f8g7');
INSERT INTO evaluations VALUES(53194,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53194,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(53194,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53194,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(53194,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(53194,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53194,36,'cp 0','b2b4 d4e3 f2e3 c5b4 d2d4 g8f6 f1d3 c8g4 b1d2 e7e6 d1a4 b8c6 f3e5 a8c8 e1g1 f8d6 e5c6 c8c6 c4c5 d6b8 d2e4 f6d5 d3b5 e8g8 b5c6 b7c6 a4c6 d8h4 e4d6 d5c3 c1d2 c3e2 g1h1 e2g3 h1g1');
INSERT INTO evaluations VALUES(53189,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53189,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53189,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(53189,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(53189,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(53189,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53189,36,'cp -9','g8f6 b4b5 b8d7 e3d4 c5d4 f3d4 d7c5 d4b3 c5d3 f1d3 d8d3 d1e2 c8f5 c1b2 a8c8 b1a3 f6d7 a1c1 e7e5 a3c2 c8c4 c2e3 c4c1 b3c1 d3b1 e2c4 b1b2 e3f5 d7b6 c4c7 b2b5 c1b3 b5d7 c7b8 d7c8 b8e5 c8e6 e5e2 e6e2 e1e2 e8d7 f5e3 d7c7 d2d4 g7g6 h1c1 c7b8 b3c5 f8g7 e2d3 h8d8');
INSERT INTO evaluations VALUES(55971,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(55971,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(55971,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(55971,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(55971,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(55971,35,'cp -128',NULL);
INSERT INTO evaluations VALUES(55971,36,'cp -129','c7c5 c2c3 g8e7 f3e2 b8c6 g1f3 d8c7 b1a3 a7a6 a3c2 c5c4 c2d4 c6d4 c3d4 b7b5 a2a3 c7b6 d2d3 e7f5 d3c4 d5c4 g2g4 f5e7 c1e3 e7d5 f1g2 f8e7 e1g1 c8b7 f3d2 e8g8 d2e4 a8d8 a3a4 d8c8 g4g5 d5e3 f2e3 b7e4 g2e4 e7g5 g1h1');
INSERT INTO evaluations VALUES(54320,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(54320,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(54320,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(54320,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(54320,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(54320,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(54320,36,'cp -127','c2c3 g8e7 f3e2 b8c6 g1f3 d8c7 b1a3 a7a6 a3c2 c5c4 c2d4 c6d4 c3d4 b7b5 g2g3 c7b6 f1g2 e7c6 e1g1 c6d4 f3d4 b6d4 f1d1 f8e7 d2d3 c8b7 d3c4 d4c4 e2e3 e8g8 e3b6 c4c6 c1e3 f8c8 a1c1 c6b6 e3b6 g7g6 h2h4 b5b4 c1c7 c8c7 b6c7 a8c8 d1c1 e7f8');
INSERT INTO evaluations VALUES(56787,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(56787,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(56787,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(56787,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(56787,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(56787,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(56787,36,'cp -25','g8f6 g1f3 c7c5 f1b5 c8d7 d1e2 d7b5 e2b5 d8d7 b5d7 b8d7 e1g1 h7h6 c1e3 a8c8 b1d2 e8f7 c2c4 b7b6 d4c5 f8c5 a1c1 c5e3 f2e3 f7e7 b2b3 h8d8 h2h3 c8c5 b3b4 c5c7 c4d5 c7c1 d5d6 e7d6 f1c1 d7e5');
INSERT INTO evaluations VALUES(49131,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(49131,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(49131,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(49131,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(49131,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49131,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(49131,36,'cp -13','g1f3 c7c5 f1b5 c8d7 d1e2 d7b5 e2b5 d8d7 b5d7 b8d7 e1g1 h7h6 c1e3 c5d4 e3d4 f8d6 f1e1 e8f7 b1d2 b7b5 a2a3 a7a5 e1e2 h8c8 c2c3 b5b4 a3b4 a5b4 a1e1 a8a6 h2h4 b4c3 b2c3 d6c5 f3e5 d7e5 d4e5 f6d7 d2f3 c5e7 e5d4 e7f6 h4h5');
INSERT INTO evaluations VALUES(49130,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49130,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(49130,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(49130,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(49130,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(49130,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(49130,36,'cp -18','c7c5 f1b5 c8d7 d1e2 d7b5 e2b5 d8d7 b5d7 b8d7 e1g1 h7h6 c1e3 c5d4 e3d4 f8d6 c2c4 e6e5 f1e1 e8g8 c4c5 d7c5 d4e5 d6e5 f3e5 f6e4 b1c3 e4c3 b2c3 c5e4 c3c4 a8e8 c4d5 e8e5 f2f3 e5d5 e1e4 f8c8 a2a4 c8c2 h2h4 b7b6 e4g4 h6h5 g4g5 d5g5 h4g5 g8f7 a4a5 f7g6 a5b6');
INSERT INTO evaluations VALUES(46528,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(46528,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(46528,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(46528,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(46528,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(46528,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(46528,36,'cp -16','f1b5 c8d7 d1e2 d7b5 e2b5 d8d7 b5d7 b8d7 e1g1 h7h6 c1e3 c5d4 e3d4 f8d6 b1d2 e8f7 f1e1 b7b5 c2c3 h8e8 d4e5 d6e5 f3e5 d7e5 e1e5 a7a5 a1e1 b5b4 c3c4 e8c8 b2b3 a5a4 e5e3 a4b3 a2b3 a8a3 d2f3 f6g4 e3e6 d5c4 b3c4');
INSERT INTO evaluations VALUES(3380,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(3380,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(3380,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(3380,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(3380,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(3380,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(3380,36,'cp 29','c4b3 c8e6 b3a4 f6d7 d2d4 e5d4 c3d4 d6d5 e4e5 d7b6 a4b3 d8d7 b1c3 a7a5 c1e3 c6b4 a2a3 a5a4 a3b4 a4b3 d1b3 h7h6 h2h3 b6c4 b4b5 f8b8 e1c1 c7c6 b5c6 b7c6 a1a8 b8a8 c3e2 e7f8 f3d2');
INSERT INTO evaluations VALUES(16005,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(16005,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(16005,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(16005,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(16005,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(16005,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(16005,36,'cp 85','b1c3 g7g6 d1c2 f8g7 e2e4 e7e5 f1e2 e5d4 f3d4 e8g8 e1g1 f8e8 c1e3 c7c6 a1d1 d8e7 e3g5 d7c5 f2f3 a7a5 b2b3 h7h6 g5h4 e7f8 h4f2 f6h5 c2d2 g7e5 g2g3 c8h3 f1e1 h5f6 g1h1 h6h5 f2e3 a5a4 b3b4 c5e6 d4e6 h3e6 e2f1 a4a3 f3f4');
INSERT INTO evaluations VALUES(17956,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(17956,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(17956,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(17956,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(17956,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(17956,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(17956,36,'cp 90','f8c5 c2c3 c5a7 d2d4 d7d6 b2b4 c8g4 c1e3 e5d4 c3d4 g8e7 d4d5 b7b5 a4b5 a6b5 d5c6 a7e3 f2e3 e7c6 d1d5 d8d7 f3d4 a8a6 d5b5 a6b6 b5c4 c6b4 h2h3 g4h5 d4e6 b4c6 c4d5 h5f7 e6g7 e8d8 d5f5 d7f5 g7f5 b6b2 a2a4 h8g8 f1f2 b2f2');
INSERT INTO evaluations VALUES(20240,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(20240,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(20240,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(20240,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(20240,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(20240,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(20240,36,'cp 169','e4d5 e6d5 c3d5 g8e7 d1e2 c8e6 d5e3 d8d7 c4e6 d7e6 d2d3 e6d7 b2b3 e8c8 e1g1 e7g6 e3c4 f8d6 c4d6 d7d6 c1b2 f7f6 f1e1 h8e8 e2d2 g6f4 g1f1 f4d5 a2a3 h7h5 e1e8 d8e8 a1e1 e8e1 d2e1 d6d7');
INSERT INTO evaluations VALUES(20218,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(20218,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(20218,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(20218,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(20218,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(20218,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(20218,36,'cp 167','e6d5 d1e2 g8e7 c3d5 c8e6 d5e3 d8d7 d2d3 e6c4 e3c4 e8c8 c1e3 e7d5 e1c1 f8e7 h1e1 h8e8 c1b1 f7f6 f3d2 f6f5 d2b3 b7b6 e3d2 e7f6 e2f3 g7g5 e1e8 d8e8 f3h5 e8g8 c4e3 d5c7 d1e1 g8e8 g2g3 c7d5 e3d5 e8e1 d2e1 d7d5 e1c3');
INSERT INTO evaluations VALUES(20498,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(20498,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(20498,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(20498,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(20498,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(20498,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(20498,36,'cp 161','d1e2 g8e7 c3d5 c8e6 d5e3 d8d7 c4e6 d7e6 e2c4 e6c4 e3c4 a8d8 d2d3 e7d5 c1d2 f7f6 a2a4 h7h5 e1c1 e8f7 g2g3 a7a6 a4a5 f8d6 c4d6 d8d6 h1e1 h8d8 e1e4 d5b4 d2b4 c5b4 f3d2 d6d5 d2b3 d5f5 f2f4 c6a5 e4b4 a5b3 b4b3 b7b5 c1d2 d8c8');
INSERT INTO evaluations VALUES(20911,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(20911,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(20911,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(20911,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(20911,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(20911,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(20911,36,'cp 85','c2c4 e7e6 a2a3 g8f6 b2b4 f8g7 c1b2 e8g8 b1d2 c8d7 f1e2 c6e7 f3e5 a7a5 b4b5 a5a4 e1g1 d5c4 e2c4 d7e8 a1b1 c7c6 b2a1 f6d5 b5c6 e7c6 c4d5 e6d5 b1b7 c6e5 d4e5 e8c6 b7b4 f8e8 e5e6 e8e6 a1g7 g8g7');
INSERT INTO evaluations VALUES(22870,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(22870,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(22870,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(22870,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(22870,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(22870,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(22870,36,'cp -43','g8f6 d2d4 d7d5 d4e5 c6e5 b1d2 f8d6 g1f3 c8g4 e1g1 e8g8 h2h3 g4h5 b2b3 d8e7 c1b2 a8d8 d1c1 e5f3 g2f3 h5f3 d2f3 f6e4 c2c4 e7e6 g1g2 d5c4 c1c4 e6c4 b3c4 f8e8 a1d1 h7h6 f3d2 e4c5 d2f3 c5a4 b2a1 b7b6 d1d5 f7f6 f1d1');
INSERT INTO evaluations VALUES(25346,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(25346,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(25346,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(25346,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(25346,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(25346,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(25346,36,'cp 42','h2h3 g4f3 e2f3 c6d4 c1e3 d4e6 e1g1 c7c6 c3e2 d6d5 e4d5 f6d5 f3d5 d8d5 c2c4 d5a5 a2a3 f8d6 b2b4 a5c7 d3d4 e5d4 e2d4 e6d4 e3d4 e8g8 d1h5 f8d8 a1d1 d6f8 f1e1 a7a5 d4e5 d8d1 e1d1 c7e7 e5d6 e7e6 d6f8 a8f8 h5a5 e6c4 a5c7');
INSERT INTO evaluations VALUES(25724,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(25724,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(25724,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(25724,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25724,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25724,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(25724,36,'cp 18','f8e7 a4c6 d7c6 d2d4 e7f6 c1e3 g8e7 d1e2 d8c8 a1d1 e8g8 d4e5 d6e5 e3c5 f8e8 c5e7 e8e7 c3d5 c6d5 d1d5 c8e6 b2b3 c7c6 d5d1 e7d7 d1d7 e6d7 f1d1 d7c7 g2g3 a8d8 g1g2 d8d1 e2d1 f6e7 a2a4 g7g6 d1d3 f7f6 f3d2 c7d6 d3e2 g8g7 d2c4 d6d4 e2f3');
INSERT INTO evaluations VALUES(26618,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26618,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(26618,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26618,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(26618,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(26618,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(26618,36,'cp 37','e7e5 h2h3 g4f3 d2f3 f8d6 c2c4 e5d4 f3d4 g8f6 f1e2 d6b4 c1d2 b4d2 d1d2 c6d4 d2d4 d5c4 d4c4 e8g8 e1g1 d8e7 f1d1 a8d8 a1c1 c7c6 d1d4 g7g6 c1d1 f8e8 b2b3 g8g7 d4d8 e8d8 d1d8 e7d8 c4b4 d8b6 b4c3 h7h5 a2a3 b6c7 e2f3 a7a6');
INSERT INTO evaluations VALUES(26878,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(26878,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(26878,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26878,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(26878,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(26878,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(26878,36,'cp 35','h2h3 g4f3 d2f3 f8d6 c2c4 e5d4 f3d4 g8f6 f1e2 d6b4 c1d2 b4d2 d1d2 c6d4 d2d4 d5c4 d4c4 e8g8 e1g1 d8e7 f1d1 f8d8 e2f3 c7c6 a2a4 a7a6 a4a5 g7g6 g2g3 d8d7 d1d7 f6d7 c4c3 d7e5 f3e2 a8d8 b2b4 e7f6 g1g2 f6e7 a1d1 d8d1 e2d1');
INSERT INTO evaluations VALUES(29536,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29536,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29536,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(29536,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(29536,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(29536,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(29536,36,'cp 11','d1b3 d8c8 g1f3 e7e6 b1d2 h7h6 e2e3 a7a5 a2a4 b8d7 h2h3 f8e7 f1e2 e8g8 e1g1 b7b6 a1c1 c7c6 c3c4 c8b7 f1d1 f8c8 f3e5 d7e5 f4e5 f6d7 e5f4 e7b4 e2d3 f5d3 b3d3 d7f6 f4e5 b7e7 b2b3 f6d7 e5f4 g8h8 c4d5');
INSERT INTO evaluations VALUES(35213,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(35213,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(35213,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(35213,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(35213,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(35213,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(35213,36,'cp 158','f1d3 c7c6 d1e2 d7d6 b1c3 e5e7 d3c4 b8d7 c1f4 d7e5 e1c1 g7g6 c4b3 g8f6 f4g5 h7h6 g5e3 c8g4 f2f3 g4e6 f3f4 e5d7 e2d3 d7b6 g1f3 e8c8 h1e1 f8g7 e3b6 a7b6 c3d5 c6d5 e4d5 f6d5 b3d5 e7f6 d3b5 f6f4 c1b1 h8e8 a2a4 g6g5 a4a5 g5g4 a5b6 g4f3 d5f3 g7e5 b5a5');
INSERT INTO evaluations VALUES(44782,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(44782,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(44782,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(44782,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(44782,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(44782,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(44782,36,'cp 98','g1f3 f8g7 f1c4 c8g4 h2h3 g4f3 d1f3 b8d7 c4e2 a7a6 a2a4 e8g8 e1g1 c5c4 a4a5 b7b5 a5b6 d8b6 a1a4 a6a5 g1h1 d7c5 a4c4 f6d7 f3e3 d7e5 c4c5 b6c5 e3c5 d6c5 c3a4 e5d7 e2b5 d7b6 a4b6 a8b8 b6d7 b8b5 d7f8 g8f8 f2f4 c5c4 f1d1 a5a4');
INSERT INTO evaluations VALUES(44781,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(44781,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(44781,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(44781,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(44781,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(44781,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(44781,36,'cp 88','f8g7 f1c4 e8g8 e1g1 b8a6 f1e1 a6c7 h2h3 e7e6 d5e6 c8e6 c4f1 d6d5 e4e5 f6e4 c3e4 d5e4 d1d8 a8d8 f3g5 g7e5 g5e4 c5c4 e4g5 e5d4 c2c3 d4c5 c1f4 d8d7 g5e6 c7e6 f4h6 f8c8 f1c4 c5f2 g1f2 c8c4 a1d1 d7d1 e1d1 a7a6 d1d6 c4c8 d6d7');
INSERT INTO evaluations VALUES(40370,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(40370,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(40370,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(40370,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(40370,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(40370,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(40370,36,'cp 112','f1c4 b8a6 e1g1 e8g8 a2a4 a6b4 f1e1 e7e6 d5e6 c8e6 c4e6 f7e6 c1g5 d8d7 c3b5 f6e8 c2c3 b4c6 d1d3 h7h6 g5h4 g6g5 h4g3 a7a6 b5a3 a8d8 a1d1 d7f7 a3c4 g5g4 f3h4 d6d5 e4d5 e6d5 c4e3 e8f6 e3f5 f6e4 d3c2 f8e8 f2f3 g4f3 g2f3 e4g3 h2g3 d5d4 e1e8 d8e8 f5d6');
INSERT INTO evaluations VALUES(45128,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(45128,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(45128,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(45128,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(45128,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(45128,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(45128,36,'cp -13','a7a6 a2a4 e7e6 d2d3 f8e7 g1f3 b8c6 e1g1 e8g8 f1e1 b7b6 c1f4 h7h6 c4b3 c8d7 c3b1 c6a5 b3a2 b6b5 b1c3 d7c6 a4b5 a6b5 c3e2 c6b7 a2b1 b5b4 c2c3 d8b6 f3d2 f8d8 d3d4 b4c3 b2c3 c5d4 c3d4');
INSERT INTO evaluations VALUES(43374,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(43374,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(43374,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(43374,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43374,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(43374,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(43374,36,'cp 0','a2a4 e7e6 g1f3 d6d5 e4d5 e6d5 c4a2 c8e6 e1g1 f8d6 d2d4 c5c4 f3g5 e8g8 b2b3 d8a5 c3e2 b7b5 b3c4 b5c4 c1f4 a5c7 g5e6 f7e6 f4d6 c7d6 c2c3 b8d7 e2g3 g7g6 d1d2 e6e5 d4e5 d7e5 a2b1 e5d7 b1c2 d6c7 d2d4 d7c5 a1e1 a8d8');
INSERT INTO evaluations VALUES(45346,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45346,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45346,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(45346,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(45346,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(45346,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(45346,36,'cp 7','c5d4 e3d4 b8c6 g1f3 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 e1g1 e7e6 a1c1 f8e7 f1d1 d8a5 b2b3 e8g8 c3a4 f8d8 c1c4 h7h6 g2g3 e7f6 g1g2 c8c7 h2h4 c7d7 a4c5 d7d4 c5b7 a5d5 f3d5 d8d5 d1d4 f6d4 c4c6 d4b6 b7d6');
INSERT INTO evaluations VALUES(49280,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(49280,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(49280,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(49280,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(49280,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(49280,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(49280,36,'cp 135','f8e7 f1d3 d7d5 e5d6 c7d6 e1g1 b8d7 f1e1 d7f6 c1e3 e8g8 b1c3 f8e8 a2a4 a7a5 f3d2 a8b8 h2h3 h7h6 a1c1 e7f8 d3b1 c8d7 d1b3 b6c8 d4d5 e6e5 g1h2 b7b6 b3c2 c8e7 g2g4 e7g6 d2e4 f6e4');
INSERT INTO evaluations VALUES(50002,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(50002,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(50002,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(50002,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(50002,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(50002,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(50002,36,'cp 236','d2d3 e4d3 f1d3 f8e7 c1f4 e8g8 b1c3 d7d5 g2g4 e7c5 d4b3 f8e8 e1f1 c5b4 d1f3 b8c6 a1d1 c6e5 f4e5 e8e5 g4g5 b4c3 b2c3 f6e8 h2h4 d8e7 b3d2 g8h8 f3f4 c8f5 d3f5 e8d6 f5d3 a8f8 f4g3 d6f5 d3f5 e5f5 d1e1 f5f2 g3f2 f8f2 f1f2 e7c5 e1e3');
INSERT INTO evaluations VALUES(50001,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(50001,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(50001,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(50001,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(50001,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(50001,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(50001,36,'cp 215','e4d3 f1d3 f8e7 c1f4 e8g8 e1g1 b8c6 d4c6 b7c6 c2c4 d7d5 b1d2 g8h8 f1e1 e7b4 a2a3 b4d6 f4e3 c6c5 e3g5 c7c6 g5h4 d8c7 h4g3 a8b8 b2b3 c8d7 d2f3 b8e8 d1c2 e8e1 a1e1 f8e8 e1e2 d6g3 h2g3 h8g8 e2e8 d7e8 c2e2 c7b6 e2b2');
INSERT INTO evaluations VALUES(50066,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(50066,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(50066,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(50066,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50066,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(50066,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50066,36,'cp 16','b1c3 d7d5 d2d4 d5e4 c3e4 e7d5 f1c4 c8e6 e1g1 f8e7 c4b3 e8g8 d1e2 a7a5 c2c4 d5e3 c1e3 f4e3 e4c3 e7f6 e2e3 a5a4 c3a4 e6g4 d4d5 b8d7 a1e1 f8e8 e3e8 d8e8 e1e8 a8e8 f1e1 g8f8 e1e8 f8e8 g1f1 g4f3 g2f3 f6d4 f1g2');
INSERT INTO evaluations VALUES(50406,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(50406,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(50406,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(50406,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(50406,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(50406,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(50406,36,'cp 22','d5c3 d2c3 d7d6 g1f3 b8c6 e5d6 d8d6 f1d3 c8g4 h2h3 g4h5 d1e2 e8c8 e1g1 h5f3 e2f3 c6e5 f3e3 e5d3 c2d3 d6d3 e3a7 d3a6 a7e3 e7e6 e3f3 f7f5 c1g5 d8d3 f3e2 h7h6 g5f4 d3d5 e2a6 b7a6 f1e1 g7g5 f4e5 h8g8 c3c4 d5d3 e5c3 g8g6 a1d1 d3d1 e1d1 f8d6 g1f1 g5g4 h3g4');
INSERT INTO evaluations VALUES(50427,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(50427,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(50427,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50427,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(50427,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(50427,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(50427,36,'cp 39','d2c3 d7d6 g1f3 b8c6 e5d6 d8d6 f1d3 c8g4 h2h3 g4h5 d1e2 e8c8 e1g1 h5f3 e2f3 c6e5 f3e3 e5d3 c2d3 d6d3 e3a7 d3a6 a7e3 d8d7 a2a4 e7e6 e3g3 f7f6 c1e3 f8d6 g3h4 d6e5 a4a5 h7h6 f1e1 h8d8 h4g4 f6f5 g4g6 a6c4 e3d4 e5d4 c3d4 c4d4 g6e6');
INSERT INTO evaluations VALUES(50426,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(50426,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(50426,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(50426,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(50426,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(50426,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(50426,36,'cp 29','d7d6 g1f3 b8c6 e5d6 d8d6 f1d3 c8g4 h2h3 g4h5 d1e2 e8c8 e1g1 h5f3 e2f3 c6e5 f3e3 e5d3 c2d3 d6d3 e3a7 d3a6 a7e3 e7e6 e3f3 f7f5 c1g5 d8d5 f1e1 f8d6 a1c1 h7h6 c3c4 d5e5 g5f4 e5e1 c1e1 h8e8 f4d6 a6d6 f3c3 e6e5 c4c5 d6e6');
INSERT INTO evaluations VALUES(51448,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(51448,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(51448,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(51448,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(51448,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(51448,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(51448,36,'cp 119','b8c6 f1c4 f8c5 e1g1 g8f6 c3d5 e8g8 c2c3 c6a5 c4d3 c5d6 f1e1 c7c6 d5f6 d8f6 b2b4 a5b7 d3f1 d6c7 d2d4 d7d6 a2a4 h7h6 c1e3 a7a5 b4b5 c6c5 g2g3 c8e6 f3h4 g7g5 h4f5 f8d8 d1h5 e6f5 e4f5 d6d5 d4c5 b7c5 e3c5 b6c5');
INSERT INTO evaluations VALUES(52174,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(52174,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(52174,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(52174,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(52174,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(52174,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(52174,36,'cp 42','g8f6 g1f3 b8c6 f3g5 d7d5 e4d5 c6d4 e1g1 f8d6 g5e4 f6e4 c3e4 e8g8 c1e3 d4f5 a1c1 f5e3 f2e3 d8e7 d1b3 a8b8 e4d6 e7d6 c4e2 c8d7 e3e4 b7b6 h2h3 b8c8 e2b5 d7b5 b3b5 g7g6 b2b4 f8d8 a2a3 c8c7 c1c6 c7c6 b5c6 d6c6 d5c6');
INSERT INTO evaluations VALUES(44357,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(44357,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(44357,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44357,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(44357,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(44357,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(44357,36,'cp 48','g1f3 b8c6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8c5 d1h5 g7g6 h5d1 c5b6 g5e4 f7f6 a2a4 e8g8 c1e3 c8f5 d5d6 g8g7 f1e1 f8f7 a1c1 a8c8 h2h4 d8d7 h4h5 g6h5 a4a5 b6a5 e3d4 f5g4 f2f3 e5d4 e4f6 d7f5 f6g4 d4c3 g4e3 f5e5 b2c3 c8c3 c1c3 a5c3');
INSERT INTO evaluations VALUES(44356,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(44356,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(44356,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(44356,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(44356,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(44356,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(44356,36,'cp 48','b8c6 f3g5 d7d5 c4d5 f6d5 e4d5 c6d4 e1g1 f8c5 d1h5 g7g6 h5d1 f7f6 g5e4 c5b6 c1e3 e8g8 a2a4 c8f5 d5d6 g8g7 a4a5 b6a5 e3d4 e5d4 d1d4 d8b6 d4d5 a5c3 e4c3 a8d8 a1d1 f8e8 c3a4 b6c6 d5c6 b7c6 d1d2 e8e5 a4c3 f5e6 f1d1 d8d7 b2b4 g7f7 f2f4 e5f5 g2g3 f7g8');
INSERT INTO evaluations VALUES(52479,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(52479,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(52479,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(52479,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(52479,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(52479,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(52479,36,'cp -29','d7d5 g1f3 b8c6 d2d4 c5d4 e3d4 c8g4 f1g2 e7e6 h2h3 g4f5 e1g1 a8c8 c2c3 h7h6 f1e1 g8f6 c1f4 f8d6 f4d6 d8d6 f3e5 e8g8 b1d2 c8c7 d1e2 f5h7 h3h4 f8c8 e5g4 f6g4 e2g4 h7f5 g4e2 b7b6 d2f3 c6e7 f3e5');
INSERT INTO evaluations VALUES(53319,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(53319,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(53319,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(53319,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(53319,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(53319,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(53319,36,'cp -18','g1f3 g8f6 d2d4 b8c6 f1g2 c8f5 e1g1 a8c8 c2c4 d5c4 d1a4 c5d4 f3d4 f5d7 a4c4 e7e5 d4c6 d7c6 c4e2 c6g2 g1g2 d8d5 g2g1 f8c5 b1c3 d5e6 e3e4 c5d4 c1g5 d4c3 b2c3 e8g8 a1c1 c8c6 g5f6 e6f6 f1d1 f8c8');
INSERT INTO evaluations VALUES(56381,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(56381,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(56381,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(56381,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(56381,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(56381,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(56381,36,'cp 3','e4d5 g8f6 f1b5 c8d7 b5d7 d8d7 c2c4 f8d6 g1e2 c7c6 b1c3 e8g8 e1g1 c6d5 c4d5 b8a6 d2d4 a6b4 d4e5 d6e5 c1f4 d7f5 f4e5 f5e5 f1e1 b4d5 c3d5 e5d5 d1d5 f6d5 a1d1 f8d8 d1d2 d5b4 e1d1 d8d2 d1d2 g8f8 f2f3 a8c8 d2d7 b4a2 d7b7 a7a5 b7b5 a2c3 e2c3 c8c3 b5a5 c3b3 a5a8 f8e7 a8a7 e7e6');
INSERT INTO evaluations VALUES(56160,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56160,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(56160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56160,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(56160,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56160,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56160,36,'cp -5','g8f6 f1b5 c8d7 b5d7 d8d7 c2c4 f8d6 b1c3 c7c6 d2d4 c6d5 g1f3 e5d4 c3d5 d7e6 d1e2 b8c6 c1b2 e8c8 f3d4 e6e2 d4e2 f6d5 c4d5 d6b4 e1f1 d8d5 b2g7 h8d8 g2g3 d5d2 g7h6 d2c2 h6g5 d8d6 h2h4 d6d7 g5f4 b4d6 f4e3 d6e5 a1c1 c2a2 e2c3 a2b2 c3a4 b2b3 e3a7 d7d2 a7b6');
INSERT INTO evaluations VALUES(57569,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(57569,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57569,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(57569,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(57569,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(57569,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57569,36,'cp 0','g8f6 g1f3 e7e6 d5e6 b8c6 e6d7 c8d7 c1f4 f6e4 b1d2 g7g5 d2e4 g5f4 e4c3 d8f6 c3d5 f6b2 a1b1 b2a2 d5c7 e8d8 c7d5 c6d4 d5c3 d4c2 e1d2 a2a5 d1c2 a8c8 b1b7 f8g7 b7d7 d8d7 c2a4 a5a4 c3a4 h8d8 e2e3 f4e3 f2e3 d7e6 f1d3 c8c4 f3d4 d8d4 e3d4 c4a4 h1e1 e6d5 d3f5 h7h6 e1e7 g7d4 f5g4 d4c3 d2c3 a4g4 e7a7 g4g2 a7h7 g2g4 c3d3 d5e5 h7e7 e5d5');
INSERT INTO evaluations VALUES(50069,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(50069,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(50069,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(50069,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(50069,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50069,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50069,36,'cp 0','g1f3 c7c6 c2c4 e7e6 d5e6 f8b4 c1d2 b4d2 b1d2 d7e6 c4c5 d8e7 a1c1 b8d7 d1c2 e6e5 e2e4 f6e4 d2e4 f5e4 c2e4 d7c5 e4e5 e7e5 f3e5 c5d7 e5c4 d7b6 f2f3 e8e7 e1f2 c8e6 c4a5 a8b8 f1d3 h8d8 d3h7 d8d2 f2g3 d2b2 h1e1 e7f6');
INSERT INTO evaluations VALUES(4035,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(4035,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(4035,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4035,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4035,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4035,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4035,36,'cp 23','e7e5 e3e4 e5d4 f3d4 d7c5 d3c2 c7c6 f2f3 f8e8 g1h1 d6d5 c4d5 c6d5 d4b3 d5d4 c3e2 c5e6 a2a4 a7a6 a4a5 a8b8 a1a3 f6h5 f3f4 b7b5 f4f5 e6g5 b3c5 g7f8 c1g5 d8g5 b2b4 c8b7 c5b7 b8b7 f5g6 h7g6 a3f3 f8g7 c2b3 h5f6 d1d4 f6e4');
INSERT INTO evaluations VALUES(3949,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(3949,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(3949,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3949,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(3949,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(3949,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3949,36,'cp 22','e3e4 e5d4 f3d4 d7c5 d3c2 c7c6 f2f3 f8e8 g1h1 d6d5 c4d5 c6d5 d4b3 d5d4 c3e2 c5e6 a2a4 a7a6 a4a5 a8b8 e2f4 d8d6 f4d3 d6d8 b3c5 e6c5 d3c5 f6d7 c5d3 d7e5 c1f4 b7b5 b2b4 d8e7 d3e5 g7e5 f4e5 e7e5 d1d2 c8b7 f1d1 b8c8 a1c1 e8d8 c2d3 c8c1 d2c1 d8c8 c1d2 e5e7 h2h3');
INSERT INTO evaluations VALUES(4974,30,'cp 356',NULL);
INSERT INTO evaluations VALUES(4974,31,'cp 344',NULL);
INSERT INTO evaluations VALUES(4974,32,'cp 360',NULL);
INSERT INTO evaluations VALUES(4974,33,'cp 355',NULL);
INSERT INTO evaluations VALUES(4974,34,'cp 364',NULL);
INSERT INTO evaluations VALUES(4974,35,'cp 389',NULL);
INSERT INTO evaluations VALUES(4974,36,'cp 370','d4g7 d8f6 g7f6 g8f6 c1g5 c5e7 c3b5 e8d8 e1c1 h8g8 g5f4 d7d6 e4e5 f6e4 e5d6 c7d6 f4e3 a7a6 b5d4 e7g5 f2f4 g5f6 f1c4 d8c7 g2g3 g8f8 c4d5 c8g4 d5a8 f8a8 d1e1 h7h5 d4b3 e4c5 h2h3 c5b3 a2b3 g4f3 h1g1 d6d5 c2c3 a6a5 c1d2 f3e4 g3g4 h5g4 h3g4 a8h8 g4g5 f6g7 e3d4 g7d4');
INSERT INTO evaluations VALUES(4967,30,'cp 375',NULL);
INSERT INTO evaluations VALUES(4967,31,'cp 383',NULL);
INSERT INTO evaluations VALUES(4967,32,'cp 349',NULL);
INSERT INTO evaluations VALUES(4967,33,'cp 361',NULL);
INSERT INTO evaluations VALUES(4967,34,'cp 371',NULL);
INSERT INTO evaluations VALUES(4967,35,'cp 350',NULL);
INSERT INTO evaluations VALUES(4967,36,'cp 358','d8f6 g7f6 g8f6 c1g5 c5e7 c3b5 e8d8 f2f3 d7d6 e1c1 c8d7 b5d4 h8g8 g5h6 a7a6 g2g4 b6b5 h6e3 c7c6 g4g5 f6h5 f3f4 d8c7 f1e2 h5g7 f4f5 f7f6 h2h4 c6c5 d4b3 d7c6 g5f6 e7f6 e3f4 a8e8 f4d6 c7b6 d6c5 b6c7 c5d6 c7b6 e2d3 c6e4 h1f1 e4d3');
INSERT INTO evaluations VALUES(98,30,'cp 365',NULL);
INSERT INTO evaluations VALUES(98,31,'cp 359',NULL);
INSERT INTO evaluations VALUES(98,32,'cp 375',NULL);
INSERT INTO evaluations VALUES(98,33,'cp 371',NULL);
INSERT INTO evaluations VALUES(98,34,'cp 384',NULL);
INSERT INTO evaluations VALUES(98,35,'cp 392',NULL);
INSERT INTO evaluations VALUES(98,36,'cp 371','g7f6 g8f6 c1g5 c5e7 c3b5 e8d8 f2f3 d7d6 e1c1 h7h6 g5d2 a7a6 b5d4 c8d7 g2g4 c7c5 d4e2 d8c7 e2f4 d7c6 f1c4 h8f8 c4e2 f8h8 h2h4 b6b5 c1b1 h8g8 d2c1 a8e8 b2b3 g8h8 c1b2 h8g8 a2a4 b5a4 e2c4 a4b3 c4b3 g8g7 f4d5 f6d5 b2g7 d5e3 b3f7');
INSERT INTO evaluations VALUES(97,30,'cp 365',NULL);
INSERT INTO evaluations VALUES(97,31,'cp 348',NULL);
INSERT INTO evaluations VALUES(97,32,'cp 350',NULL);
INSERT INTO evaluations VALUES(97,33,'cp 354',NULL);
INSERT INTO evaluations VALUES(97,34,'cp 371',NULL);
INSERT INTO evaluations VALUES(97,35,'cp 355',NULL);
INSERT INTO evaluations VALUES(97,36,'cp 384','g8f6 c1g5 c5e7 c3b5 e8d8 e1c1 h8g8 g5f4 d7d6 e4e5 f6e4 e5d6 c7d6 f4e3 a7a6 b5d4 e7g5 f2f4 g5f6 f1c4 c8b7 g2g3 g8f8 d4f5 d8c7 c4d5 b7d5 d1d5 h7h5 h1d1 a8d8 e3d4 h5h4 g3h4 f8g8 f5h6 g8g2 h6f7 d8d7 f7d6 f6d4 d6e8 c7c6 d5d4 d7d4 d1d4 g2g1 d4d1 g1g2 d1e1 e4d2 e8f6 d2f3 e1e6 c6c7 c2c3');
INSERT INTO evaluations VALUES(443,30,'cp 382',NULL);
INSERT INTO evaluations VALUES(443,31,'cp 381',NULL);
INSERT INTO evaluations VALUES(443,32,'cp 381',NULL);
INSERT INTO evaluations VALUES(443,33,'cp 375',NULL);
INSERT INTO evaluations VALUES(443,34,'cp 366',NULL);
INSERT INTO evaluations VALUES(443,35,'cp 362',NULL);
INSERT INTO evaluations VALUES(443,36,'cp 380','c1g5 c5e7 c3b5 e8d8 e1c1 h8g8 g5f4 d7d6 e4e5 f6e4 e5d6 c7d6 f4e3 a7a6 b5d4 e7g5 f2f4 g5f6 f1c4 d8c7 g2g3 g8f8 c4d5 c8b7 d4f5 a8d8 g3g4 f6h8 d5b7 c7b7 d1d3 d8d7 h1e1 f8e8 e3d4 h8d4 d3d4 d6d5 f5g3 e4d6 e1e8 d6e8 h2h4 e8g7');
INSERT INTO evaluations VALUES(5248,30,'cp 358',NULL);
INSERT INTO evaluations VALUES(5248,31,'cp 406',NULL);
INSERT INTO evaluations VALUES(5248,32,'cp 383',NULL);
INSERT INTO evaluations VALUES(5248,33,'cp 403',NULL);
INSERT INTO evaluations VALUES(5248,34,'cp 389',NULL);
INSERT INTO evaluations VALUES(5248,35,'cp 422',NULL);
INSERT INTO evaluations VALUES(5248,36,'cp 435','e5g7 e7f6 g7g3 g8e7 b1c3 d7d5 f1e2 h7h5 e4d5 e7f5 g3f4 e8g8 e1g1 f8e8 e2h5 f6e5 f4f3 d8f6 c3e4 e5h2 g1h1 e8e4 f3e4 h2d6 e4f3 f6h4 f3h3 h4f6 c1d2 f5d4 h5g4 c8g4 h3g4 g8f8 a1e1 f6h8 g4h3 h8h3 g2h3 d4f3 e1e2 b7b5 d2e3 b5b4 e3h6 f8g8');
INSERT INTO evaluations VALUES(5227,30,'cp 449',NULL);
INSERT INTO evaluations VALUES(5227,31,'cp 443',NULL);
INSERT INTO evaluations VALUES(5227,32,'cp 448',NULL);
INSERT INTO evaluations VALUES(5227,33,'cp 425',NULL);
INSERT INTO evaluations VALUES(5227,34,'cp 390',NULL);
INSERT INTO evaluations VALUES(5227,35,'cp 428',NULL);
INSERT INTO evaluations VALUES(5227,36,'cp 439','e7f6 g7g3 g8e7 b1c3 d7d5 f1e2 h7h5 e4d5 e7f5 g3f4 d8e7 d5d6 f6c3 b2c3 e7d6 f4d6 f5d6 c1f4 e8e7 e1c1 d6e4 h1e1 c8e6 e2f3 f7f5 f3e4 f5e4 e1e4 e7f7 d1d6 h8e8 f4e3 a8c8 c3c4 c8c6 d6c6 b7c6 e4e5 e8g8 e5h5 g8g2 h5c5 g2g4 c5c6 e6c4 c6c7 f7f6');
INSERT INTO evaluations VALUES(5082,30,'cp 440',NULL);
INSERT INTO evaluations VALUES(5082,31,'cp 441',NULL);
INSERT INTO evaluations VALUES(5082,32,'cp 442',NULL);
INSERT INTO evaluations VALUES(5082,33,'cp 442',NULL);
INSERT INTO evaluations VALUES(5082,34,'cp 449',NULL);
INSERT INTO evaluations VALUES(5082,35,'cp 454',NULL);
INSERT INTO evaluations VALUES(5082,36,'cp 439','g7g3 g8e7 b1c3 d7d5 f1e2 c8e6 c3b5 h8g8 b5c7 e8d7 c7e6 f7e6 g3h3 d8c7 e4d5 e7d5 e1g1 c7e5 e2f3 a8d8 h3h7 g8g7 h7e4 e5e4 f3e4 b7b5 g2g3 c5c4 g1g2 g7g8 a2a3 a7a5 f1e1 d8f8 e1d1 d7c7 c2c3 f6e7 f2f3 e7c5 e4d5 e6d5 d1d5 c7c6 d5e5 a5a4');
INSERT INTO evaluations VALUES(5081,30,'cp 462',NULL);
INSERT INTO evaluations VALUES(5081,31,'cp 471',NULL);
INSERT INTO evaluations VALUES(5081,32,'cp 472',NULL);
INSERT INTO evaluations VALUES(5081,33,'cp 467',NULL);
INSERT INTO evaluations VALUES(5081,34,'cp 464',NULL);
INSERT INTO evaluations VALUES(5081,35,'cp 463',NULL);
INSERT INTO evaluations VALUES(5081,36,'cp 443','g8e7 b1c3 d7d5 f1e2 h7h5 e4d5 e7f5 g3f4 d8e7 d5d6 f6c3 b2c3 f5d6 e1g1 c8d7 f1e1 e8c8 c1e3 d6f5 e2f3 h8h6 a1b1 h6b6 b1b6 a7b6 e3d2 e7d6 f3h5 d6f4 d2f4 f5g7 h5e2 d8h8 e2f3 g7e6 f4e3 d7c6 f3c6 b7c6');
INSERT INTO evaluations VALUES(7607,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(7607,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(7607,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(7607,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(7607,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(7607,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(7607,36,'cp -67','d8e7 f2f4 d7d6 d1e2 d6e5 e2e5 c8f5 a2a3 b4d6 e5e7 d6e7 c1e3 e8c8 f1e2 a7a5 a1c1 f6g4 e2g4 f5g4 h2h3 g4f5 e1f2 h8e8 f2f3 c6c5 h1e1 g7g6 e1e2 f5d7 c1e1 d7c6 f3f2 h7h5 g2g3 a5a4 e3c1 e7f6 e2e8 c6e8 c1e3 e8d7 e1d1 f6c3 b2c3');
INSERT INTO evaluations VALUES(447,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(447,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(447,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(447,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(447,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(447,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(447,36,'cp -65','f2f4 d7d6 d1e2 d6e5 e2e5 c8f5 a2a3 b4d6 e5e7 d6e7 c1e3 f5c2 f1e2 e8c8 e3a7 h8e8 e1g1 c8b7 a1c1 c2d3 a7f2 d3e2 c3e2 f6e4 f1d1 g7g6 f2e1 e7c5 g1f1 c5e3 d1d8 e8d8 c1c2 d8d1 e2g3 e3d2 g3e4 d1e1 f1f2 e1e4 c2d2 e4f4 f2e2 c6c5 e2e3 f4a4');
INSERT INTO evaluations VALUES(14166,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(14166,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(14166,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(14166,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14166,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(14166,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(14166,36,'cp 42','g1f3 e7e6 e1g1 a7a6 b5c6 b7c6 c2c4 c6c5 c4d5 d8d5 b1c3 d5b7 f3e5 f8d6 d1a4 f6d7 e5c4 e8g8 e3e4 h7h6 c4d6 c7d6 f1d1 d7b6 a4c2 c5d4 d1d4 d6d5 e4d5 b6d5 c3d5 e6d5 c1e3 c8e6 c2d2 f8c8 h2h3 b7d7 d4b4 a6a5 b4b6 a5a4 a1d1');
INSERT INTO evaluations VALUES(16465,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(16465,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(16465,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(16465,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(16465,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(16465,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(16465,36,'cp -4','d7d5 c4d5 f6d5 d1c2 d5c3 b2c3 d8g5 g1f3 g5g6 c2g6 h7g6 d2d4 f8d6 f1d3 c8d7 c1b2 d7e6 a1c1 e6f5 d3b5 f5d7 b5d3 e8c8 h2h4 d8e8 f3g5 c6d8 g5e4 d6e7 d4e5 e7h4 c3c4 h4e7 e1e2 a7a6 h1h8 e8h8 f2f4');
INSERT INTO evaluations VALUES(15100,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(15100,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(15100,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(15100,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15100,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(15100,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(15100,36,'cp 0','c4d5 f6d5 d1c2 d5c3 b2c3 f8d6 d2d4 d8g5 g1f3 g5g6 c2g6 h7g6 f1d3 c8d7 c1b2 d7f5 d3b5 f5d7');
INSERT INTO evaluations VALUES(19420,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(19420,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(19420,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(19420,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(19420,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(19420,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(19420,36,'cp 161','d4d5 c6b8 e2e4 d7d6 g1f3 g7g6 a2a4 g8f6 f1b5 c8d7 b5d3 f8g7 e1g1 e8g8 b1a3 a7a6 f1e1 d7g4 h2h3 g4f3 d1f3 b8d7 f3d1 d7e5 d3f1 d8c7 c1f4 f6d7 f4e3 d7f6 e3c1 f6h5 f1e2 h5f6 d1c2 a8e8 c1f4 e7e6 d5e6 e8e6 a1d1');
INSERT INTO evaluations VALUES(21985,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(21985,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21985,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(21985,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(21985,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(21985,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(21985,36,'cp 47','d2d4 e5d4 f3d4 g8f6 b1c3 f8e7 d4c6 b7c6 f1d3 a8b8 b2b3 f6g4 c1b2 e8g8 e1g1 g4e5 c3e2 e7f6 a1b1 e5d3 b2f6 d8f6 d1d3 f8e8 e2g3 a7a5 b1d1 f6g5 d3c3 g5e5 c3e5 e8e5 f2f4 e5c5 d1d2');
INSERT INTO evaluations VALUES(22583,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(22583,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(22583,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(22583,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(22583,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(22583,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(22583,36,'cp -17','d7d5 e4d5 e6d5 c4b3 g8f6 g1e2 c6a5 e1g1 f8d6 c1f4 e8g8 f4d6 d8d6 f1e1 c8g4 d1d2 g4e2 e1e2 f8e8 b1c3 e8e2 c3e2 b7b6 a1e1 a8e8 e2c3 c7c6 g2g3 a5b3 e1e8 f6e8 a2b3 h7h6 c3e2 g7g5 c2c3 e8g7 d2e3 a7a5 e3e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(21373,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(21373,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(21373,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(21373,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(21373,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(21373,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(21373,36,'cp -17','e4d5 e6d5 c4b3 c6a5 g1e2 a5b3 a2b3 g8f6 e1g1 f8e7 f1e1 e8g8 e2g3 f8e8 h2h3 e7d6 e1e8 f6e8 b1d2 c8e6 d2f3 c7c6 f3g5 d8d7 c1d2 d6g3 f2g3 f7f6 g5f3 e8d6 d2f4 d6e4 g3g4 e6f7 f3d2 a7a5 d1e2 e4d6 e2f2 f7g6');
INSERT INTO evaluations VALUES(21361,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(21361,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(21361,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(21361,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(21361,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21361,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(21361,36,'cp -22','e6d5 c4b3 g8f6 g1e2 c6a5 e1g1 f8d6 c1f4 e8g8 f4d6 d8d6 h2h3 c8f5 b1d2 a8e8 f1e1 a7a6 d2f3 f6e4 c2c3 a5b3 a2b3 f7f6 b3b4 e8e7 e2c1 f8e8 c1d3 g7g5 d3c5 c7c6 c5e4 f5e4 f3d2 e4g6 e1e7 d6e7 d2b3 e7e2 d1d2 e2d2');
INSERT INTO evaluations VALUES(21582,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(21582,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(21582,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(21582,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21582,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(21582,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(21582,36,'cp -14','c4b3 g8f6 g1e2 c6a5 e1g1 f8d6 c1f4 e8g8 f1e1 f8e8 f4d6 d8d6 h2h3 c8f5 b1d2 e8e7 e2g3 e7e1 d1e1 a8e8 e1c1 f5g6 c2c3 c7c6 d2f3 f6h5 g3h5 g6h5 g2g4 h5g6 f3e5 e8e7 c1f4 g6e4 g4g5 f7f6 g5f6 d6f6 f4f6 g7f6 e5g4 a5b3');
INSERT INTO evaluations VALUES(23468,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(23468,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23468,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23468,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23468,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(23468,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23468,36,'cp 0','d7d5 d3d4 c8f5 e2e3 c6b4 f1b5 c7c6 b5a4 e7e6 e1g1 a7a5 a2a3 b7b5 a3b4 a5b4 a4b5 c6b5 d1e2 a8a1 e2b5 d8d7 b5b8 d7d8');
INSERT INTO evaluations VALUES(26052,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(26052,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(26052,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(26052,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(26052,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(26052,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(26052,36,'cp 24','a7a6 b5a4 g8f6 f1e1 b7b5 a4c2 d7g4 d2d4 c5d4 c3d4 e7e5 c1e3 e5d4 e3d4 f8e7 d4e3 g4h5 h2h3 e8g8 g2g4 h5g6 b1c3 a8c8 a1c1 c6a5 c2d3 f8e8 f3h4 a5c4 h4g6 h7g6 d3c4 c8c4 b2b3 c4c8 c3d5 c8c1 e3c1 f6d5 d1d5 e7f6');
INSERT INTO evaluations VALUES(25621,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(25621,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(25621,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25621,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(25621,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25621,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(25621,36,'cp 24','b5a4 g8f6 f1e1 b7b5 a4c2 d7g4 a2a4 g7g6 d2d4 c5d4 a4b5 a6b5 a1a8 d8a8 b1a3 a8b7 c3d4 f8g7 a3b5 b7b5 c2a4 b5a6 b2b4 g4d7 b4b5 a6a8 b5c6 d7c6 a4c6 a8c6 d4d5 c6a6 e4e5 d6e5 f3e5 e8g8 e5c6 e7e6 c6e7 g8h8 d5e6 f7e6 d1b3 f6d7 c1g5');
INSERT INTO evaluations VALUES(30249,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(30249,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(30249,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(30249,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(30249,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(30249,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(30249,36,'cp 119','b8c6 c4b5 d8d6 g5f6 g7f6 b1c3 e8c8 b5c6 d6c6 e1g1 f8c5 c3d5 f6f5 f3f5 c6e6 f5f3 c7c6 d5e3 d8d2 a1d1 h8d8 d1d2 d8d2 b2b3 a7a6 g2g3 c5e3 f2e3 d2c2 f3f7 e6d6 f7f8 d6f8 f1f8 c8c7 f8f7 c7b6 g3g4 c2a2 f7h7 a2b2 g4g5 b2b3 g5g6 b3e3');
INSERT INTO evaluations VALUES(35031,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(35031,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(35031,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(35031,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(35031,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(35031,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(35031,36,'cp 118','h4d8 g1f3 d7d5 c2c3 c7c5 f1d3 a7a6 e1g1 b8c6 f1e1 c8d7 a2a3 h7h6 d3c2 d8c7 c1e3 c5c4 b1d2 c6a5 h2h4 e8c8 h4h5 g8e7 f3h4 h8g8 d1g4 c8b8 g4g3 g7g5 h5g6 f7g6 h4g6 e7g6 c2g6 a5b3');
INSERT INTO evaluations VALUES(56995,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(56995,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(56995,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(56995,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(56995,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(56995,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(56995,36,'cp 124','g1f3 d7d5 f1d3 c7c5 c2c3 b8c6 e1g1 c8d7 f1e1 a7a6 a2a3 c5c4 d3c2 h7h6 b1d2 d8c7 a1b1 c6a5 h2h4 g8e7 d2f1 e8c8 c1f4 a5c6 g2g3 c6a5 f1h2 c8b8 h2g4 e7g6 f4c1 g6e7 h4h5 e7f5 c1f4');
INSERT INTO evaluations VALUES(56994,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(56994,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(56994,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(56994,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(56994,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(56994,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(56994,36,'cp 123','d7d5 f1d3 c7c5 c2c3 b8c6 e1g1 c8d7 d4c5 f8c5 b2b4 c5b6 b1d2 g8e7 a2a4 e7g6 f1e1 e8g8 c1a3 f8e8 b4b5 c6a5 g2g3 a8c8 a3b4 b6c5 h2h4 g6f8 f3d4 b7b6 a1b1 f7f5 d1e2 d8e7 f2f4 e8d8 d2f3 d7e8 g1g2 a5b7 e1h1 h7h6 h4h5 e7f7 b1e1 g8h8');
INSERT INTO evaluations VALUES(35486,30,'cp -160',NULL);
INSERT INTO evaluations VALUES(35486,31,'cp -186',NULL);
INSERT INTO evaluations VALUES(35486,32,'cp -180',NULL);
INSERT INTO evaluations VALUES(35486,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(35486,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(35486,35,'cp -156',NULL);
INSERT INTO evaluations VALUES(35486,36,'cp -192','c7c5 f1c4 d6d5 e4d5 f8d6 e1g1 b8d7 b1c3 a7a6 c1g5 d8c7 f1e1 h8f8 g5f4 f7g8 d1d2 c5d4 d2d4 b7b5 f4d6 c7d6 c4f1 d7c5 d4e5 d6e5 e1e5 c8d7 d5d6 a8c8 a1d1 f8e8 e5e8 c8e8 h2h4 g8f8 f2f3 c5b7 a2a4 b5a4 g2g4 d7c6 g1f2 e8d8 f1a6 b7d6 a6f1 f8e7 d1e1 e7f7 g4g5');
INSERT INTO evaluations VALUES(43224,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43224,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43224,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(43224,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43224,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43224,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43224,36,'cp 13','c1b2 e8g8 e2e3 b7b6 f1e2 c7c5 e1g1 c5d4 f3d4 c8b7 c2c4 d7d5 b1d2 b8a6 a1c1 a8c8 d4f3 a6c5 b3b4 c5e4 d1a4 d5c4 d2c4 f6d5 b2g7 g8g7 a4b3 b6b5 c4a5 d8d7 b3b2 e4c3 h2h3 g7g8 a5b7 d7b7 a2a3');
INSERT INTO evaluations VALUES(43230,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(43230,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43230,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43230,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(43230,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(43230,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43230,36,'cp 0','g1f3 e8g8 f1e2 b7b6 b2b4 d7d5 b1d2 b8d7 e1g1 c7c5 b4c5 b6c5 c1a3 c5c4 a1b1 f8e8 f3e5 d7e5 d4e5 f6d7 f2f4 d8a5 a3b4 a5b6 e2f3 b6e3 g1h1 d7b6 f3e2 b6d7');
INSERT INTO evaluations VALUES(43231,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43231,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(43231,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43231,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(43231,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43231,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(43231,36,'cp 0','b7b6 b1d2 d7d5 b2b4 c7c5 f1b5 c8d7 b5e2 c5c4 a2a4 d7f5 f3e5 b8d7 e5c6 d8c8 b4b5 d7b8 c6e5 b8d7');
INSERT INTO evaluations VALUES(47197,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(47197,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(47197,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(47197,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(47197,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(47197,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(47197,36,'cp -17','c8f5 g1f3 e7e6 c2c4 f8e7 d1b3 b7b6 f1d3 f5g6 e1g1 e8g8 f3e5 c7c5 d4c5 b6c5 d3g6 h7g6 b3b7 b8d7 e5c6 d8e8 c4d5 e6d5 e3e4 e7d6 e4d5 d7b6 d2f3 e8c8 b7c8 f8c8 c1d2 f6d5 c6e5 a7a5 a1c1 c5c4 h2h3 d6e7 f1e1 a5a4');
INSERT INTO evaluations VALUES(29581,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(29581,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(29581,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(29581,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(29581,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(29581,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(29581,36,'cp -16','c2c4 e7e6 g1f3 f8e7 f3h4 f5e4 h4f3 e4g6 f1e2 e8g8 e1g1 b8d7 f3h4 c7c5 h4g6 h7g6 c4d5 f6d5 d2f3 c5d4 d1d4 a8c8 f1d1 d8c7 e3e4 e7f6 d4d3 d5b6 c1e3 f6b2 a1b1 d7e5 d3b5 e5f3 g2f3 b2f6 d1c1 c7e5 b5e5 c8c1 e3c1');
INSERT INTO evaluations VALUES(50758,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(50758,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50758,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(50758,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(50758,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(50758,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50758,36,'cp 37','c4d3 f8e7 c2c4 c8d7 b1c3 b8c6 d4c6 d7c6 d1c2 e7f6 e1g1 f6d4 f1e1 g8f6 c1e3 d4e3 e1e3 e8g8 c2d2 d8a5 d3f1 f8d8 a1d1 d8d7 a2a3 d6d5 e4d5 e6d5 b2b4 a5b6 c4c5 b6c7 d2d4 a8d8 e3g3 f6h5');
INSERT INTO evaluations VALUES(51366,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(51366,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(51366,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(51366,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(51366,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(51366,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(51366,36,'cp 84','c7c5 c2c3 d8d7 h2h4 c8a6 f1a6 b8a6 h4h5 a6b8 h1h3 b8c6 e1f1 f7f6 c1f4 g8e7 b1a3 g7g5 e5f6 g5f4 f6e7 f8e7 d1e2 e8g8 a1e1 c5d4 c3d4 f8f6 e1c1 f6f8 h5h6 e7f6 a3b5 a8c8 c1e1 a7a6 b5c3 c6d4 f3d4 f6d4 e2e6 d7e6 e1e6 f8f6 e6e7 f6f7 c3d5 c8c1 f1e2 c1c2 e2d1 c2b2');
INSERT INTO evaluations VALUES(51687,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(51687,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(51687,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(51687,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(51687,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(51687,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(51687,36,'cp 77','c2c3 d8d7 h2h4 c8a6 f1a6 b8a6 d1e2 a6c7 a2a4 h7h6 h4h5 c5d4 f3d4 f8c5 d4c2 f7f6 b1d2 f6e5 d2f3 c5d6 f3e5 d6e5 e2e5 g8f6 c1f4 a8c8 c2d4 e8g8 f2f3 d7e8 e1f1 e8f7 h1h4 f6d7 e5d6 d7c5');
INSERT INTO evaluations VALUES(51685,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(51685,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(51685,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(51685,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(51685,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(51685,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(51685,36,'cp 86','d8d7 h2h4 c8a6 f1a6 b8a6 h4h5 a6b8 e1f1 b8c6 h1h3 f7f6 c1f4 g8e7 b1a3 g7g5 e5f6 g5f4 f6e7 f8e7 d1e2 c5d4 c3d4 e7a3 b2a3 e8c8 a1e1 d8e8 e1c1 c8b7 f3e5 c6e5 e2e5 e8c8 c1c8 h8c8 f1g1 d7a4 g1h2 c8c1 e5g7 b7a6 g7h7 a4d4 h7d3 d4d3 h3d3');
INSERT INTO evaluations VALUES(54786,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(54786,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(54786,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(54786,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(54786,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(54786,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(54786,36,'cp 74','e4g5 e7e5 g1h3 b8c6 f1c4 g8h6 c2c3 d8f6 d2d4 f5f4 d4e5 c6e5 c4b3 c8f5 e1g1 f8d6 c1f4 e8c8 b3c2 e5d3 f4d6 d8d6 c2d3 d6d3 d1a4 c8b8 a1d1 d3d1 f1d1 h8d8 d1d8 f6d8 a4d4 d8e8 d4d2 h6f7 b2b3 f7e5 h3f4 h7h6 g5f3 e5f3 g2f3 e8d7 d2d7');
INSERT INTO evaluations VALUES(56415,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(56415,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(56415,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(56415,33,'cp -143',NULL);
INSERT INTO evaluations VALUES(56415,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(56415,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(56415,36,'cp -167','b8c6 b1c3 c8e6 d2d4 e5e4 c1f4 h7h5 g4h5 g8f6 f4g5 d8d6 e2e3 f6h5 g1e2 f8e7 h3h4 e6g4 d1d2 f7f6 g5f4 d6d7 f4h2 e8c8 e1c1 g7g5 h4g5 f6g5 d1g1 d8f8 d2e1 c8b8 c1b1 a7a6 e2g3 h5f6 b1a1 g4e6');
INSERT INTO evaluations VALUES(57499,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(57499,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(57499,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(57499,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(57499,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(57499,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(57499,36,'cp 13','f8b4 c1d2 g8f6 g1f3 d7d6 g2g3 e8g8 f1g2 a7a5 e1g1 d8e8 a2a3 b4c3 d2c3 a5a4 e2e3 b8d7 d1c2 e5e4 d3e4 f6e4 a1c1 e8f7 c3b4 f8e8 f3d2 d7c5 b4c5 e4c5 d2f3 c5e4 f3d4 h7h6 d4e2 e4c5');
INSERT INTO evaluations VALUES(3434,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(3434,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(3434,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(3434,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(3434,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(3434,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(3434,36,'cp 84','e1c1 c8e6 f2f4 h7h6 g5h4 c7c6 c4e6 f7e6 e4e5 f6d5 h4e7 d8e7 e5d6 e7d6 g2g3 d6c7 a2a3 a7a5 d4c5 b7b6 c5d4 a8d8 h1e1 b6b5 d4e4 f8e8 e4g6 d5f6 d1d8 c7d8 c3e4 f6e4 e1e4 d8d7 e4e5 a5a4 f4f5 d7f7 e5e6 f7g6 e6g6 e8e2 g6c6 e2h2');
INSERT INTO evaluations VALUES(3433,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(3433,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(3433,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(3433,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(3433,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(3433,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(3433,36,'cp 70','c8e6 f2f4 h7h6 g5h4 c7c6 c4e6 f7e6 e4e5 f6d5 h4e7 d8e7 e5d6 e7d6 g2g3 d6c7 a2a3 a7a5 h1e1 a8d8 c3d5 e6d5 d4c3 a5a4 c3b4 b7b5 e1e7 c7c8 d1e1 d8e8 c1d2 e8e7 b4e7 c8h3 e7e6 h3e6 e1e6 f8f6 e6e5 f6d6 e5e7 g8f8 e7c7 g7g5 f4f5 g5g4 d2e3 d6f6 e3d4 f6f5 c7c6 f5f2');
INSERT INTO evaluations VALUES(4022,30,'cp 221',NULL);
INSERT INTO evaluations VALUES(4022,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(4022,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(4022,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(4022,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(4022,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(4022,36,'cp 208','e4e5 f6g4 h2h3 d6e5 h3g4 e5d4 c3e4 d7b6 c4b3 c8g4 f1e1 c7c6 d1d3 d8c7 a2a4 a8d8 c1d2 b6d5 a4a5 d5f4 d2f4 c7f4 e4d2 g4f5 d3c4 f4d6 d2e4 d6c7 e4g3 f5g4 f3h2 g4c8 c4c5 e7e6 c5a7 d4d3 c2d3 d8d3 b3c4 d3g3 f2g3');
INSERT INTO evaluations VALUES(13695,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(13695,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(13695,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(13695,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(13695,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(13695,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(13695,36,'cp 62','d2d3 f8c5 b5c6 b7c6 e1g1 e8g8 b1d2 h7h6 g5f3 f8e8 b2b3 f6d5 c1b2 f7f6 d2e4 c5f8 f1e1 c8e6 a2a3 c6c5 f3d2 d8d7 d2c4 a8d8 b2c3 f6f5 c3a5 f5e4 c4e5 d7d6 e5c4 d6d7 a5d8 d7d8 d3e4 d5b6 d1d8 e8d8 a1d1 d8d1 e1d1');
INSERT INTO evaluations VALUES(18479,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(18479,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(18479,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(18479,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(18479,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(18479,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(18479,36,'cp -80','c2c3 c8a6 g1e2 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 d8b6 d1c2 a8c8 f1d1 f8e7 c3a4 b6a5 c1d2 a5b5 d1e1 b5c4 c2c4 a6c4 d2c3 c6c5 a4c5 e7c5 d4c5 c4e2 e1e2 c8c5 f2f3 h7h5 a1c1 e8g8 e2e1 f8b8 b2b4 c5c6 c3d2 f5d4');
INSERT INTO evaluations VALUES(18477,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(18477,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(18477,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(18477,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(18477,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(18477,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(18477,36,'cp -87','c5d4 c3d4 g8e7 g1e2 e7f5 b1c3 c8a6 e1g1 f8e7 f1e1 a6e2 e1e2 a8c8 c3a4 e8g8 e2c2 d8a5 c2c3 a5b4 c1e3 c6c5 a4c5 e7c5 a1c1 b4b2 d1c2 b2b5 c3c5 c8c5 d4c5 b5c6 h2h3 d5d4 e3d2 f8c8 c2d3 f5h4 d3g3 h4g6 h3h4 c6d5');
INSERT INTO evaluations VALUES(18489,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(18489,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(18489,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(18489,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(18489,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(18489,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(18489,36,'cp -89','c3d4 g8e7 g1e2 e7f5 b1c3 d8b6 e1g1 c8a6 d1c2 a8c8 f1d1 f8e7 c3a4 b6a5 c1d2 a5b5 d1e1 b5d3 c2c3 d3c3 d2c3 c6c5 a4c5 e7c5 d4c5 a6e2 e1e2 c8c5 f2f3 h7h5 a1c1 e8g8 e2e1 f8c8 c3d2 d5d4 g1f2 c5c2 f2e2 f5e7 b2b3 c2a2 c1c8 e7c8 e2d3 a2b2 e1c1 c8b6 c1c7 b2b3 d3d4');
INSERT INTO evaluations VALUES(18488,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(18488,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(18488,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(18488,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(18488,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(18488,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(18488,36,'cp -73','g8e7 g1e2 e7f5 b1c3 c8a6 e1g1 d8b6 d1c2 f8e7 c3a4 b6b5 c1e3 b5c4 f1c1 c6c5 c2c4 a6c4 e2g3 f5e3 f2e3 c5d4 e3d4 e8g8 a4c5 e7g5 c1c3 g5d2 c3f3 f8c8 g1f2 d2b4 b2b3 c4b5 a1c1 b4a3 c1c2 a7a5 g3h5 a5a4 f3g3 g7g6 h5f6 g8g7');
INSERT INTO evaluations VALUES(33165,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(33165,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(33165,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(33165,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(33165,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33165,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33165,36,'cp 0','h4e4 f1e2 b7b6 e1g1 e4b7 f1e1 d7d5 e2d3 g8f6 f3e5 c7c5 c1g5 c5d4 d3b5 e8f8 d1f3 d4c3 g5f6 g7f6 f3f6 h8g8 f6d8 f8g7 d8g5 g7f8');
INSERT INTO evaluations VALUES(33174,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(33174,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(33174,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(33174,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(33174,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(33174,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(33174,36,'cp 28','f1e2 d7d6 e1g1 b4c3 b2c3 g8f6 a2a4 e8g8 f1e1 b8d7 a4a5 e4c6 c3c4 a7a6 h2h3 f8e8 e2d3 b7b6 d4d5 c6b7 d5e6 e8e6 c1e3 d7c5 f3d4 c5d3 d1d3 e6e5 e1b1 e5e8 d4e2 f6d7 a5b6 c7b6 d3d6 d7e5 b1b6 e5c4 b6b7 c4d6 b7b1 c8e6 e2d4 e6d5');
INSERT INTO evaluations VALUES(33173,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(33173,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(33173,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33173,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(33173,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(33173,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(33173,36,'cp 11','b4c3 b2c3 d7d6 e1g1 g8f6 f1e1 e8g8 a2a4 e4f5 h2h3 b8c6 g2g4 f5a5 e2d3 e6e5 c1d2 e5e4 c3c4 a5b6 c4c5 b6b2 d3e4 f6e4 e1e4 c6b4 d2b4 b2b4 e4e7 d6c5 e7c7 c5d4 d1d4 b4d4 f3d4 f8d8 d4f5 c8f5 g4f5 d8c8 c7b7 c8c2 a4a5 c2c5 f5f6 g7f6 a5a6 h7h6');
INSERT INTO evaluations VALUES(35413,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(35413,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35413,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(35413,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(35413,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(35413,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(35413,36,'cp 2','e8d8 f3d2 a7a5 a2a4 h7h5 f1c4 g8e7 d2b3 c8d7 b3c5 e7c6 c5d7 b8d7 c2c3 c6a7 e1e2 a7c8 h1d1 c8d6 c4d3 c7c6 c1e3 d6c8 d3c4 d8c7 h2h4 c8b6 e3b6 d7b6 b1d2 a8e8 c4f7 e8e7 f7b3 g7g5 g2g3');
INSERT INTO evaluations VALUES(37381,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(37381,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37381,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(37381,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(37381,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(37381,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(37381,36,'cp 22','f1g2 d7d5 c4d5 f6d5 e1g1 c7c5 d2d4 c5d4 f3d4 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1e3 c8h3 e3a7 h3f1 a7b8 f1e2 b8f4 g7c3 a1b1 e7e5 f4h6 c3d2 h6d2 d8d2 b1a1 g8g7 a2a4 g6g5 a4a5 e2a6 c6g2 d2c2');
INSERT INTO evaluations VALUES(45040,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(45040,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(45040,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(45040,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(45040,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(45040,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(45040,36,'cp -17','e7e5 d4e5 d8d1 c3d1 f6e4 g1f3 c8e6 d1e3 b8d7 f1d3 d7c5 d3c4 e8c8 a2a4 f8e7 e1g1 h8e8 f1e1 e4g5 c4e6 c5e6 f3g5 e7g5 e3c4 g5c1 e1c1 e6d4 c1d1 c8c7 c2c3 d4e2 g1f1 d8d1 a1d1 e2f4 g2g3 f4g6 d1e1 e8e6 b2b4 b7b5 a4b5 c6b5 c4d6 e6e5 e1e5 g6e5 d6b5 c7b6 b5d6');
INSERT INTO evaluations VALUES(46414,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46414,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46414,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(46414,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(46414,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(46414,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(46414,36,'cp -8','c4b3 f8e7 g1e2 e8g8 e1g1 c8e6 e2f4 b8c6 f4e6 f7e6 c3e2 e7d6 c2c3 f6g4 g2g3 d8f6 f2f3 g4e5 f3f4 e5d3 d1d3 c5c4 b3c4 d6c5 e2d4 c6b4 d3e2 c5d4 c3d4 f6d4 c1e3 d4c4 e2c4 d5c4 e3c5 b4c2 c5f8 c2a1 f8g7 g8g7 f1a1 a8d8 a1c1 b7b5');
INSERT INTO evaluations VALUES(46955,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(46955,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(46955,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(46955,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(46955,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(46955,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(46955,36,'cp -92','d8d6 b1c3 b8c6 g1f3 c8g4 f1e2 e7e5 d2d3 e8c8 a2a3 h7h6 c1e3 c8b8 f3d2 g4e6 e1g1 d6d7 d2c4 d7e8 d1e1 c6d4 a1c1 e6c8 e1d1 h6h5 f1e1 d4e2 d1e2 f6g4 d3d4 e5e4 c1d1 f7f5 d4d5 g4e3 e2e3');
INSERT INTO evaluations VALUES(53534,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(53534,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(53534,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(53534,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(53534,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(53534,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(53534,36,'cp -36','b7b6 c5b6 a7b6 e2e4 d5e4 c3e4 f8b4 e4c3 g8f6 g1f3 e8g8 f1d3 c8a6 e1g1 b4c3 b2c3 a6d3 d1d3 b8d7 c1f4 a8a4 f4d6 f8e8 f3e5 b6b5 c3c4 b5c4 e5c4 d7b6 c4b6 d8b6 d6c5 b6a5 a2a3 f6d5 a1b1 a5c3 d3c3 d5c3 b1b2 h7h6 f1a1 e8a8 a1c1');
INSERT INTO evaluations VALUES(53579,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(53579,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53579,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(53579,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(53579,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(53579,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(53579,36,'cp 7','d5c4 f1g2 b7b5 f3e5 c8b7 a2a4 a7a6 b2b3 c4b3 c1b2 b8d7 d1b3 d7e5 d4e5 g8e7 b1c3 e7d5 a4b5 a6b5 a1a8 d8a8 c3d5 c6d5 b3b5 b7c6 b5b6 f8e7 e1g1 a8b7 b6b7 c6b7 b2d4 e8d7 f1b1 b7c6 e2e3 h8a8 g2f1 c6a4 f1e2 d7e8 b1b7');
INSERT INTO evaluations VALUES(53659,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(53659,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(53659,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(53659,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(53659,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53659,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(53659,36,'cp 12','f1g2 b7b5 a2a4 c8b7 f3e5 a7a6 b2b3 c4b3 c1b2 b8d7 d1b3 d7e5 d4e5 g8e7 a4b5 a6b5 a1a8 d8a8 b1c3 e7d5 c3d5 c6d5 e1g1 a8a6 f1a1 a6b6 e2e3 f8c5 g2f1 e8g8 b3b5 b6b5 f1b5 f8c8 a1c1 c5f8 c1c8 b7c8 b2d4 f8e7 f2f4 h7h6 g1f2 g8h8 b5f1');
INSERT INTO evaluations VALUES(5819,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(5819,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(5819,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(5819,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(5819,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(5819,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(5819,36,'cp 106','b1d2 b4e7 a2a3 a7a5 f1d3 f7f5 c4d5 e6d5 d1c2 g8h6 d2b1 c6b8 b1c3 c7c6 e1g1 e8g8 c1d2 b8d7 c3e2 d7f6 b2b4 f6e4 f1b1 e4d2 c2d2 e7d6 b4b5 c8d7 h2h3 h6f7 h3h4 f7h6 e2f4 h6g4 g2g3 d8e7 b5c6 b7c6 a3a4 g4f6 b1b6 f6e4');
INSERT INTO evaluations VALUES(9298,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(9298,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(9298,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(9298,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(9298,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(9298,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(9298,36,'cp 17','c2c3 e8g8 h2h3 h7h6 d2d4 c5b6 b5a4 c8d7 a4c2 f8e8 a2a4 a7a5 c1e3 e5d4 e3d4 c6d4 c3d4 f6h5 b1c3 c7c5 c3d5 c5d4 e4e5 d7e6 d5b6 d8b6 d1d4 b6d4 f3d4 d6e5 e1e5 h5f4 d4e6 e8e6 e5e6 f7e6 a1d1 a8a6 d1d7 a6b6 b2b3 b6b4 g2g3 f4h3 g1g2 h3g5');
INSERT INTO evaluations VALUES(9290,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(9290,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(9290,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(9290,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(9290,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9290,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(9290,36,'cp 20','e8g8 h2h3 h7h6 a2a4 a7a5 d2d3 c8d7 b1d2 c6e7 b5d7 f6d7 d1b3 f8e8 d2c4 e7g6 c1d2 c5a7 b3c2 d8f6 c4e3 g6h4 f3h4 f6h4 b2b4 a7e3 e1e3 c7c6 b4b5 e8e6 d3d4 d6d5 e4d5 c6d5 d4e5 d7e5 a1e1 h4f6');
INSERT INTO evaluations VALUES(11791,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11791,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11791,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11791,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11791,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11791,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(11791,36,'cp 35','c3d5 f6f5 f1d3 c8e6 e1g1 f8g7 d1h5 f5f4 g1h1 e8g8 a3b1 h7h6 b1d2 f7f5 f2f3 g8h8 a2a4 e6f7 h5h3 f7d5 e4d5 c6e7 f1g1 d8a5 d2b3 a5b4 a4b5 e5e4 d3e2 a6b5 a1a8 f8a8 f3e4 b4e4 e2d3 e4d5 b3d2 d5c5 c2c3 c5e5');
INSERT INTO evaluations VALUES(11790,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11790,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11790,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11790,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11790,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(11790,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(11790,36,'cp 0','f6f5 f1d3 c8e6 e1g1 e6d5 e4d5 c6e7 c2c3 f8g7 d1h5 e5e4 d3c2 e8g8 a1e1 a8c8 g1h1 c8c5 c2b3 e7d5 b3d5 c5d5 e1e4 f8e8 e4e2 e8e2 h5e2 d8d7 f2f4 d7e6 e2f2 d5d3 a3c2 h7h6 a2a3 a6a5 h1g1 e6d5 c2e3 d5e4 f1e1 d6d5 e3c2 e4c4 c2e3 c4e4');
INSERT INTO evaluations VALUES(11793,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11793,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(11793,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(11793,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11793,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11793,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(11793,36,'cp 6','c2c3 f8g7 e4f5 c8f5 a3c2 e8g8 c2e3 f5e6 f1d3 f7f5 e1g1 g8h8 d1h5 e5e4 d3c2 c6e7 a1d1 e6f7 h5h3 e7d5 e3d5 d8g5 h3e3 g5h4 g2g3 h4g4 c2b3 a8e8 d5f4 f7b3 a2b3 g7e5 d1d5 e5f4 e3f4 g4f4 g3f4 f8g8 g1h1 e4e3 f2e3 e8e3 d5d6 e3e2 d6a6 e2b2 f1d1');
INSERT INTO evaluations VALUES(12770,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(12770,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(12770,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(12770,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(12770,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(12770,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(12770,36,'cp -78','f6g4 d1e2 f7f6 e5f6 g4f6 c3e4 d7d5 e4f6 g7f6 c2c3 e7e5 d2d4 c5d4 f3d4 d8c7 d4c6 b7c6 e2h5 c7f7 h5h4 h8g8 h4a4 f7b7 g2g3 e8f7 a4c2 f7g7 f1g2 b7a6 b2b3 c8e6 c1d2 e6f7 c3c4 f7g6 c2d1 a8d8 e1g1 f8c5 c4d5 c6d5 a1c1 c5b6 a2a4 d5d4 f1e1 g8f8 d2b4 f8e8');
INSERT INTO evaluations VALUES(12912,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(12912,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(12912,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(12912,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(12912,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(12912,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(12912,36,'cp -55','d1e2 f7f6 e5f6 g4f6 c3e4 d7d5 e4f6 g7f6 c2c3 e7e5 d2d4 c5d4 f3d4 h7h5 d4c6 b7c6 e2c2 e8f7 f1e2 f8c5 c1e3 c5b6 c3c4 h5h4 e1g1 h4h3 g2g3 c8e6 c4c5 b6c7 f2f4 d5d4 f4e5 c7e5 e3f4 d8d5 e2f3 d4d3 f4e5 d5e5');
INSERT INTO evaluations VALUES(15739,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15739,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(15739,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15739,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(15739,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(15739,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(15739,36,'cp 7','d6e5 d1d8 e8d8 c1g5 c7c6 f1c4 d8e8 a2a4 a7a5 f2f3 f6d7 c3d1 h7h5 h2h4 f8c5 g5d2 e8e7 e1e2 d7b6 c4b3 g7g6 d1f2 b6d7 f2d3 c5d4 c2c3 d4a7 h1d1 f7f6 b3c4 d7c5 d3c5 a7c5 d2e3 c5e3 e2e3 h8d8');
INSERT INTO evaluations VALUES(15831,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(15831,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(15831,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15831,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(15831,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(15831,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(15831,36,'cp 13','d1d8 e8d8 c1g5 c7c6 f1c4 d8e8 a2a4 a7a5 f2f3 f6d7 c3d1 h7h5 h2h4 f8c5 g5d2 f7f6 e1e2 e8e7 d1f2 d7b6 c4b3 b6d7 f2d3 c5d4 c2c3 d4a7 h1d1 g7g6 g2g3 d7c5 d3c5 a7c5 d2e3 c5e3 e2e3 c8h3 b3c4 h3e6 c4e6 e7e6 f3f4 f6f5');
INSERT INTO evaluations VALUES(2255,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(2255,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(2255,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(2255,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(2255,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(2255,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(2255,36,'cp 6','e8d8 c1g5 c7c6 f1c4 d8e8 a2a4 a7a5 f2f3 f6d7 c3d1 h7h5 h2h4 f8c5 e1e2 f7f6 g5c1 d7b6 c4b3 b6d7 d1f2 c5d4 c2c3 d4a7 h1d1 e8e7 g2g3 d7c5 b3c4 c8e6 c4e6 e7e6 c1e3 c5b3 a1a3 b3c5 f2d3 c5d3 d1d3 a7e3 e2e3 a8d8 a3b3');
INSERT INTO evaluations VALUES(16031,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(16031,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(16031,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(16031,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(16031,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(16031,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(16031,36,'cp 32','c2c4 e7e6 g1f3 f8b4 b1d2 c7c5 d4c5 e8g8 a2a3 b4c5 c4d5 f6d5 f4g3 b7b6 f1b5 c8b7 e3e4 d5f6 e4e5 f6d5 d2e4 c5e7 e4d6 e7d6 e5d6 a7a6 b5a4 d5f6 e1g1 b6b5 a4c2 b7d5 a1c1 a8c8 f3e5 d7e5 g3e5');
INSERT INTO evaluations VALUES(16533,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(16533,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(16533,32,'cp -143',NULL);
INSERT INTO evaluations VALUES(16533,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(16533,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(16533,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(16533,36,'cp -114','d7d5 e4d5 d8d5 f3d5 f6d5 g1f3 c8f5 d2d3 e8c8 g2g4 f5g6 f3h4 f8e7 h4g6 h7g6 a2a4 c6a5 b1d2 d5f4 d2c4 a5c4 d3c4 e7g5 c1e3 f7f5 g4f5 g6f5 a4a5 a7a6 a1d1 d8d1 e1d1 h8h4 c4c5 c8d7 b2b4 g5f6 d1c2 e5e4 b4b5 a6b5 f1b5 c7c6 b5c4 d7c7 a5a6 b7a6 c4a6 f4g2');
INSERT INTO evaluations VALUES(15143,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(15143,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(15143,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(15143,33,'cp -134',NULL);
INSERT INTO evaluations VALUES(15143,34,'cp -108',NULL);
INSERT INTO evaluations VALUES(15143,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(15143,36,'cp -106','e4d5 d8d5 f3d5 f6d5 g1f3 c8f5 d2d3 a8d8 g2g4 f5e6 c1g5 f8e7 g5e3 e8g8 b1d2 f7f5 g4f5 f8f5 e1c1 d8f8 f1g2 d5f4 g2f1 e6d5 h1g1 f5h5 e3f4 f8f4 f1g2 g7g6 c1c2 g8f8 d2e4 d5e4 d3e4 f4e4');
INSERT INTO evaluations VALUES(19589,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(19589,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(19589,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(19589,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(19589,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(19589,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(19589,36,'cp -61','g8f6 g1f3 e5c4 b3c4 d7d5 e4d5 d8d5 c4e2 c8g4 b1c3 d5c6 a2a4 e7e6 e2b5 g4f3 g2f3 a8c8 d2d3 f6d5 b5c6 c8c6 c1d2 f7f6 h2h4 h7h5 e1e2 e8f7 c3d5 e6d5 f3f4 f8e7 f4f5 e7d6 e2f3 c5c4 d2f4 d6f4 f3f4 c4d3 c2d3');
INSERT INTO evaluations VALUES(22862,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(22862,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(22862,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(22862,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(22862,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(22862,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(22862,36,'cp -8','e2e4 d7d5 d1a4 f7f6 d2d3 g8e7 b1d2 a7a6 d3d4 e5d4 f3d4 c8d7 f1e2 d5e4 d2e4 c6d4 a4d4 d7c6 d4e3 d8d5 f2f3 d5e5 e4g3 e5e3 c1e3 h7h5 h2h4 e7d5 e3d2 e8c8 e1c1 f8c5 g3f5 d8d7 b2b4 c5f2 h1f1 f2a7 c3c4 d5b6');
INSERT INTO evaluations VALUES(26016,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(26016,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(26016,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(26016,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(26016,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(26016,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(26016,36,'cp 53','c2c4 e7e6 c4c5 g7g6 e1g1 f8g7 b1c3 g8e7 c1f4 e8g8 f1e1 e7f5 c3e2 f8e8 d1d2 a7a6 b5c6 d7c6 a1d1 f7f6 g2g4 f5e7 g4g5 e7f5 e2g3 f6g5 f4g5 d8c7 g3f5 e6f5 f3e5 e8e6 e5c6 c7c6 e1e6 c6e6 d1e1 e6c6 d2f4 a8e8 e1e8 c6e8 h2h3 h7h5 g1h2 g8h7 b2b3');
INSERT INTO evaluations VALUES(31815,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(31815,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(31815,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(31815,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(31815,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(31815,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(31815,36,'cp 128','b1c3 c7c6 d5c6 b8c6 f1b5 a8c8 g1f3 a7a6 b5c6 c8c6 e1g1 e7e6 f1e1 c6c8 c3e4 g8f6 e4f6 g7f6 d2d3 h8g8 c1f4 d8b6 c2c3 f5g4 f4g3 b6b2 c3c4 b2b6 a1b1 b6c6 d3d4 g4f3 g2f3 c6d7 d4d5 c8c4 d5e6 d7d1 e6f7 e8f7 b1b7 f7g6 e1d1');
INSERT INTO evaluations VALUES(31814,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(31814,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(31814,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(31814,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(31814,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(31814,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(31814,36,'cp 123','c7c6 d5c6 b8c6 f1b5 a8c8 g1f3 a7a6 b5c6 c8c6 e1g1 e7e6 f1e1 c6c8 h2h3 h7h6 c3e4 f8e7 c2c3 g8f6 e4f6 e7f6 d2d4 d8c7 f3h2 e8g8 h2g4 f6e7 d1f3 f5g6 g4e5 g6h7 c1f4 f7f6 e5g4 c7d7 f3e2 f8d8 g4e3 h7g6 a1d1');
INSERT INTO evaluations VALUES(34478,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(34478,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(34478,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(34478,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34478,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34478,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34478,36,'cp 9','g8f6 f1e2 c5d4 d1d4 b8c6 d4d5 f6d5 e1g1 g7g6 e2c4 f8g7 f1d1 e8g8 c4d5 e6d5 c1e3 f8d8 h2h3 c8f5 f3d4 c6d4 e3d4 d8d6 d4g7 g8g7 b1d2 f5c2 d1c1 c2a4 d2b3 a4b3 a2b3 g7f8 c1d1 a7a6 b3b4 f8e7 f2f3');
INSERT INTO evaluations VALUES(46385,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(46385,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(46385,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(46385,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(46385,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(46385,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(46385,36,'cp -49','c8d7 b5d7 b8d7 e1g1 f8d6 c2c4 d5d4 f1e1 e8g8 c1g5 a7a5 b1d2 h7h6 g5f6 d8f6 d2e4 f6e7 d1a4 d6c7 a4b5 b7b6 e4g3 e7d8 b5d5 f8e8 g3e4 d8e7 f3h4 a8d8 h4f5 e7e6 d5e6 e8e6 f2f3 g7g6 f5h6 g8g7 h6g4 f7f5 e4g5 e6e7 g4f2 d7c5 g2g4');
INSERT INTO evaluations VALUES(28857,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(28857,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(28857,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(28857,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(28857,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(28857,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(28857,36,'cp -36','b5d7 b8d7 e1g1 f8d6 c2c4 d5d4 c1g5 e8g8 b1d2 d8c7 f1e1 a7a5 g5f6 d7f6 d2e4 f6d7 c4c5 d6c5 a1c1 f8e8 a2a3 a5a4 c1c4 b7b5 c4c1 a8c8 b2b4 a4b3 d1b3 c7b7 c1b1 c5f8 b3b5 b7b5 b1b5 f8a3 b5b7 d7f6 e4f6 g7f6 f3d2');
INSERT INTO evaluations VALUES(28841,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(28841,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(28841,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(28841,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(28841,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(28841,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(28841,36,'cp -39','b8d7 e1g1 f8b4 c2c3 b4d6 c3c4 d5d4 f1e1 e8g8 c1g5 d8c7 b1d2 a7a5 g5f6 d7f6 d2e4 f6e4 e1e4 f8e8 e4e2 f7f6 a2a3 g8h8 g2g3 c7c6 f3d2 a5a4 e2e1 d6c7 b2b4 a4b3 d1b3 a8a5 b3d1 c7d6 d2e4 f6f5 e4d6 c6d6 d1f3');
INSERT INTO evaluations VALUES(24532,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(24532,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(24532,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(24532,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(24532,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(24532,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(24532,36,'cp -38','e1g1 f8d6 c2c4 d5d4 c1g5 e8g8 b1d2 d8c7 f1e1 a7a5 g5f6 d7f6 d2e4 f6e4 e1e4 f8e8 e4e2 f7f6 d1c2 g8h8 f3d2 a5a4 d2e4 d6e7 e4d2 b7b5 a2a3 e7d6 g2g3 c7c6 c2d1 d6f8 c4b5 c6b5 d2f3 e8b8 e2c2 b5d5 a1c1');
INSERT INTO evaluations VALUES(24531,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(24531,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(24531,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(24531,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(24531,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(24531,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(24531,36,'cp -37','f8d6 c2c4 d5d4 c1g5 e8g8 b1d2 d8c7 f1e1 a7a5 g5f6 d7f6 d2e4 f6e4 e1e4 f8e8 a2a3 f7f6 g2g3 a5a4 e4e1 f6f5 f3d2 g7g6 b2b4 a4b3 d1b3 d6c5 a3a4 b7b6 b3b5 e5e4 d2b3 e8e5 b3c5 c7c5 g1g2 g8f7 a1d1 c5b5 a4b5');
INSERT INTO evaluations VALUES(47057,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(47057,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(47057,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47057,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(47057,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(47057,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(47057,36,'cp 22','c2c4 c7c6 g2g3 d6d5 g1f3 c8g4 f1g2 e7e6 e1g1 f8e7 d2d4 b8d7 b1c3 e8g8 h2h3 g4f5 f3d2 f8e8 e2e4 d5e4 c3e4 d8c7 d1e2 f6e4 d2e4 d7f6 e4d2 h7h6 f1e1 a8d8 d2f3 c6c5 a1d1 f5e4 d4c5 e7c5 f3e5 e4g2 g1g2 c5b4 d1d8 e8d8 e1d1 d8d1');
INSERT INTO evaluations VALUES(48512,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(48512,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(48512,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(48512,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(48512,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(48512,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(48512,36,'cp 80','f8e7 c1g5 e8g8 d1e2 b8c6 e1c1 c6b4 h1e1 h7h6 g5h4 b4d3 e2d3 f8e8 d3c4 f6d5 h4e7 e8e7 f3e5 d5c3 c4c3 c8d7 c3f3 b7b6 h2h4 d7e8 h4h5 a8c8 d4d5 d8d6 e5c6 e8c6 d5c6 d6c5 e1e3 c8f8 c1b1 a7a5 a2a3 c5g5 e3c3 g5b5 c3d3 b5g5');
INSERT INTO evaluations VALUES(41046,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(41046,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(41046,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(41046,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(41046,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(41046,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(41046,36,'cp 66','c1e3 b8d7 d1e2 a7a6 f3e5 c7c5 e1c1 d8c7 f2f4 e8g8 c1b1 b7b5 d4c5 e7c5 e5d7 f6d7 e3c1 c8b7 h1f1 d7f6 g2g4 c5b4 c3e4 f6e4 d3e4 b7e4 e2e4 b4e7 a2a3 f8c8 h2h4 b5b4');
INSERT INTO evaluations VALUES(57071,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(57071,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(57071,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(57071,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(57071,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(57071,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(57071,36,'cp -19','d7d5 g1f3 g8f6 d3d4 b7b6 f1b5 c7c6 b5d3 b8d7 e1g1 f8d6 b1d2 e8g8 b2b3 e6e5 d4e5 d7e5 d3e2 d8e7 c2c4 e5f3 d2f3 d5c4 e2c4 c8g4 c1b2 a8d8 d1c2 g4f3 g2f3 d6e5 f1d1 e5b2 c2b2 f8e8 a1c1 f6d7 f3f4 d7c5');
INSERT INTO evaluations VALUES(55946,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(55946,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(55946,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(55946,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(55946,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(55946,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(55946,36,'cp -13','g1f3 f8d6 d3d4 g8f6 b2b3 b7b6 c1b2 e8g8 f1d3 c7c5 e1g1 b8d7 b1d2 c8b7 c2c4 c5d4 f3d4 d7e5 d3e2 d5c4 d2c4 e5c4 e2c4 d8e7 a2a4 f8d8 d1e2 h7h6 a1c1 d6a3 b2a3 e7a3 c4a6 b7d5 e2c2 a3e7 d4c6 d5c6 c2c6');
INSERT INTO evaluations VALUES(9855,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(9855,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9855,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(9855,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(9855,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(9855,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(9855,36,'cp 29','d2c3 d7d6 c4b5 e8g8 b5c6 b7c6 f3d2 f6h7 e1g1 d8g5 f2f4 e5f4 d1f3 f7f6 f3f4 g5f4 f1f4 c6c5 d2c4 c8e6 c4a5 f8e8 b2b3 h7f8 a1e1 f8g6 f4f2 a7a6 h2h3 g8f7 e1f1 d6d5 e4d5 e6d5 f1e1 a8d8');
INSERT INTO evaluations VALUES(9848,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(9848,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(9848,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(9848,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(9848,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9848,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(9848,36,'cp 21','d7d6 c4b5 c8d7 d1e2 d8e7 e1c1 a7a6 b5c6 d7c6 f3d2 e8c8 b2b3 h8e8 d2c4 b7b6 a2a4 f6d7 c4e3 d7c5 f2f3 a6a5 h1e1 f7f6 d3d4 e5d4 c3d4 c8b7 e2c4 c5e6 d4c3 e6c5 h2h4 c5e6 c3b2 e7d7 c1b1 e6c5');
INSERT INTO evaluations VALUES(8884,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(8884,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(8884,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(8884,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(8884,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(8884,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(8884,36,'cp 20','c4b5 c8d7 d1e2 d8e7 e1c1 a7a6 b5c6 d7c6 h1e1 e8c8 f3d2 h8e8 f2f3 b7b6 d2c4 f6d7 b2b3 d7c5 e2f2 f7f6 f3f4 a6a5 c4e3 c6b7 a2a4 c8b8 c1b1 e7d7 c3b2 b6b5 a4b5 d7b5 b2a3 c5e6 f4f5 e6d4 b1a2 a5a4 b3b4');
INSERT INTO evaluations VALUES(13231,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(13231,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(13231,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(13231,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(13231,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(13231,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(13231,36,'cp -48','e7e6 e1g1 f8e7 c2c3 e8g8 f1e1 d6d5 c4b3 d5e4 g5e4 f6d5 b1a3 a8b8 d3d4 c5d4 c3d4 b7b5 a3c2 c8b7 e4c3 d5c3 b2c3 c6a5 c2e3 a5b3 a2b3 a7a5 c3c4 e7f6 c4b5 f6d4 c1a3 d4a1 a3f8 d8d1 e1d1 g8f8');
INSERT INTO evaluations VALUES(22013,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22013,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22013,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22013,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22013,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22013,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22013,36,'cp 0','b1a3 c8e6 e1g1 g8f6 c4e6 f7e6 f3g5 e6e5 d1b3 d8d7 a1c1 c6a5 b3c3 a5c6');
INSERT INTO evaluations VALUES(26766,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(26766,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(26766,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(26766,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(26766,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(26766,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(26766,36,'cp -18','c6d7 e1g1 f8e7 b1c3 g8f6 a2a4 c7c6 c1g5 f6d5 g5e7 d5c3 e7d8 c3d1 f1d1 a8d8 d1d6 f7f6 d6d1 e8e7 f3e1 d7e6 f2f3 c6c5 e1d3 c5c4 d3f2 d8d4 g1f1 h8d8 f1e1 a7a5 c2c3 d4d6 d1d6 d8d6 f2d1 d6b6 a1b1 f6f5 e4f5 e6f5 d1e3 f5b1 e3d5 e7d6 d5b6');
INSERT INTO evaluations VALUES(27265,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(27265,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(27265,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(27265,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(27265,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(27265,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(27265,36,'cp -17','e1g1 f8e7 b1c3 g8f6 a2a4 c7c6 c1g5 f6d5 g5e7 d5c3 e7d8 c3d1 f1d1 a8d8 d1d6 f7f6 a1d1 e8e7 a4a5 d7e6 d6d2 d8d2 f3d2 h8d8 g1f1 b7b5 a5b6 a7b6 f1e2 d8a8 h2h4 g7g6 e2d3 b6b5 f2f3 c6c5 d1h1 c5c4 d3e3 b5b4 h1b1 f6f5');
INSERT INTO evaluations VALUES(29626,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(29626,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(29626,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(29626,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(29626,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(29626,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(29626,36,'cp 83','g4f3 g2f3 c7c6 h2h4 a7a6 e2e4 d5e4 d1b3 b8d7 c1f4 e7e5 f4e5 d7e5 d4e5 f6d7 e1c1 d8c7 c3e4 d7c5 b3c3 c5e4 f3e4 f8c5 f1h3 a8d8 f2f4 e8g8 h4h5 c7e7 a2a3 d8d1 h1d1 f8d8 b2b4 c5f2 c3f3 d8d1 c1d1 f2d4 c4c5 g7g6 d1c2 e7h4 e5e6 f7e6');
INSERT INTO evaluations VALUES(29636,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(29636,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(29636,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(29636,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(29636,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(29636,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(29636,36,'cp 88','g2f3 c7c6 e2e4 d5e4 f3e4 e7e5 d4d5 f8b4 c1d2 b4c3 d2c3 f6e4 d1g4 e4c3 g4g7 h8f8 g7e5 d8e7 e5e7 e8e7 b2c3 c6d5 f1g2 f8c8 g2d5 c8c7 e1c1 e7f8 c1c2 b8d7 h1e1 a8e8 e1e8 f8e8 d1d3 e8f8 d3h3 d7f6 c2d3 b7b6 h3f3 f8e7 d3d4 c7c5 f3e3 e7f8 e3e2 f6d7 f2f4 c5a5 d5c6 d7f6 c6b5 a5a3 d4e5 f6h5 f4f5 a3c3');
INSERT INTO evaluations VALUES(32972,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(32972,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(32972,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(32972,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(32972,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(32972,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(32972,36,'cp 76','d7d6 d3d4 c5b6 a2a4 g8e7 a4a5 b6d4 f3d4 e5d4 c3e2 d6d5 c4b3 e8g8 e4d5 f8d8 d1d4 f6d4 e2d4 e7d5 b3c4 c6c5 d4b3 b8a6 c4a6 b7a6 b3c5 c8f5 c2c3 a8c8 c5b3 f5d3 f1e1 d8e8 c1d2 e8e1 d2e1 d5f4 b3d4 c8e8 f2f3 f4e2 d4e2 e8e2 a1d1 d3c2 d1d8 g8h7 e1f2 c2b3 f2a7 e2b2 a7e3 b2b1 g1f2');
INSERT INTO evaluations VALUES(36677,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(36677,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(36677,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(36677,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(36677,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(36677,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(36677,36,'cp 72','f1e1 b8c6 c2c3 e7e5 h2h3 a7a5 d3c2 h7h6 c1e3 e5d4 c3d4 c6b4 b1c3 b4c2 d1c2 f8e8 c2d2 g8h7 e3f4 c7c6 a1c1 a5a4 a2a3 c8e6 d4d5 c6d5 e4d5 e6f5 f3d4 f5d7 f4d6 d7h3 d6e5 h3d7 d4f3 a8c8 d5d6 c8c6 c1d1');
INSERT INTO evaluations VALUES(37354,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37354,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37354,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37354,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37354,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37354,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37354,36,'cp 0','h7h6 g5f6 g7f6 d1c2 b8c6 f2f4 d7d6 a1d1 f6g7 g1f3 d8e8 c3d5 e8d8');
INSERT INTO evaluations VALUES(40008,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40008,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(40008,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40008,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40008,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(40008,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40008,36,'cp 0','c5d4 d1d4 b8c6 d4d1 d7d6 f1e2 c8f5 e1g1 f6e4 c3e4 f5e4 f3g5 e4f5 e3e4 f5d7 a1b1 c6d4 b2b3 d8a5 a2a4 e8g8 c1e3 d4e2 d1e2 a7a6 e3d2 a5d8 f1c1 b7b5 c4b5 a6b5 a4a5 d8b8 b3b4 f8c8 h2h3 h7h6 g5f3 b8b7');
INSERT INTO evaluations VALUES(40026,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40026,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40026,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40026,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40026,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40026,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40026,36,'cp 0','f3d4 b8c6 d1b3 b7b6 f1e2 e8g8 e1g1 c8b7 f1d1 e7e6 d4b5 f6e8 a1b1 a8c8 c1d2 a7a6 b5a3 c6a5 b3b4 a5c6');
INSERT INTO evaluations VALUES(44243,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(44243,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(44243,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(44243,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(44243,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(44243,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(44243,36,'cp 83','f1d3 c5d4 c3d4 b8c6 g1e2 c6b4 b1c3 d7d5 e4e5 b4d3 d1d3 f6g8 h2h4 c8d7 h4h5 h7h6 c1d2 g8e7 a1c1 e7f5 c3b5 d7b5 d3b5 d8d7 b5d3 a8c8 c1c8 d7c8 e1g1 c8d7 g2g4 f5h4 f2f4 f8e7 f4f5 e6f5 g4f5 e7g5');
INSERT INTO evaluations VALUES(44279,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(44279,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44279,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44279,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44279,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(44279,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44279,36,'cp 12','f8b4 c1d2 b4d2 b1d2 e8g8 f1d3 b7b6 b2b4 c8a6 d3e2 a6b7 e1g1 d7d6 a2a3 b8c6 d1b3 c6e7 h2h3 h7h6 f1d1 a7a5 b3b2 a5b4 a3b4 d8c7 d1e1 e7g6 e2d3 g6f4 d3f1 f4g6 a1c1 g6e7 b4b5 e7g6 e1d1 a8a7');
INSERT INTO evaluations VALUES(48223,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(48223,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(48223,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(48223,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(48223,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(48223,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48223,36,'cp -33','c7c5 b4c5 f8c5 d2d4 d8b6 d1c1 c5e7 a3a4 e8g8 b1d2 b6c7 g1f3 c8d7 f1d3 b8c6 c2c3 f8c8 e1g1 c6a5 c1b1 h7h6 f3e5 d7e8 b2a3 e7a3 a1a3 f6d7 e5d7 e8d7 f1c1 e6e5 d4e5 c7e5 d2f3 e5d6 a3a1 c8c5 b1b4');
INSERT INTO evaluations VALUES(46152,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(46152,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(46152,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(46152,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(46152,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(46152,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(46152,36,'cp -18','b4c5 f8c5 g1f3 e8g8 c2c4 b8c6 d2d4 c5e7 b1d2 c6a5 b2c3 a5c4 d2c4 d5c4 f1c4 b7b6 e1g1 c8b7 d1b3 a8c8 f3e5 c8c7 f1c1 f6e4 c3b4 e7b4 a3b4 e4d2 b3d3 d8g5 g2g3 d2c4 c1c4 c7c4 e5c4 g5d5 e3e4 d5b5 a1a7 b7e4 d3e4 b5c4 a7b7 c4b4 g1g2 b4a5 e4e5');
INSERT INTO evaluations VALUES(48390,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48390,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48390,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48390,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(48390,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(48390,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(48390,36,'cp -11','e4e5 f6d7 c2c3 g7g5 h2h3 f8g7 d3c2 d7e5 f3e5 g7e5 d2d4 e5f4 h3h4 h7h5 h4g5 d8g5 c1f4 g5f4 b1d2 f4h6 d1f3 b8c6 e1c1 e6e5 d4e5 c8e6 c2a4 h5h4 h1h2 e8c8 a4c6 b7c6 d1h1 h6g5 f3f6 g5f6 e5f6 h4h3 g2h3 d8g8');
INSERT INTO evaluations VALUES(48433,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(48433,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48433,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48433,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48433,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48433,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(48433,36,'cp -23','f6d7 c2c3 g7g5 d3c2 g5g4 f3g1 d7e5 d2d4 e5g6 d1g4 c7c5 g1f3 b8c6 g4h5 c5d4 e1g1 f8g7 f1e1 d8d6 c1h6 g6f4 h6f4 d6f4 h5d5 e8g8 d5h5 f4h6 h5h6 g7h6 f3d4 c6d4 c3d4 h6g7 b1c3 g7d4 a1d1 d4b6 g1f1 a8b8 c3e4 f8d8 d1d8 b6d8 e4d6 d8e7 d6c8 b8c8');
INSERT INTO evaluations VALUES(56375,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(56375,31,'cp 266',NULL);
INSERT INTO evaluations VALUES(56375,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(56375,33,'cp 305',NULL);
INSERT INTO evaluations VALUES(56375,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(56375,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(56375,36,'cp 302','d4e5 c8e6 b1c3 f8b4 c1g5 b4c3 b2c3 d8d7 e4d5 e6d5 d1d3 h7h6 e1c1 h6g5 d3d5 d7d5 d1d5 g8e7 d5d1 g5g4 f3g5 h8h5 h2h4 b8c6 f1c4 e8f8 g5f7 e7g6 d1d7 g6e5 f7e5 c6e5 d7c7 e5c4 c7c4 h5f5 h1e1 f5f2 c4g4 a8c8 e1e3 c8d8 e3e4');
INSERT INTO evaluations VALUES(56308,30,'cp 289',NULL);
INSERT INTO evaluations VALUES(56308,31,'cp 289',NULL);
INSERT INTO evaluations VALUES(56308,32,'cp 306',NULL);
INSERT INTO evaluations VALUES(56308,33,'cp 302',NULL);
INSERT INTO evaluations VALUES(56308,34,'cp 303',NULL);
INSERT INTO evaluations VALUES(56308,35,'cp 308',NULL);
INSERT INTO evaluations VALUES(56308,36,'cp 296','c8e6 b1c3 f8b4 c1g5 b4c3 b2c3 d8d7 e4d5 e6d5 d1d3 h7h6 e1c1 h6g5 d3d5 d7d5 d1d5 g5g4 f3g5 h8h5 g5f7 b8d7 d5d4 d7e5 f7e5 h5e5 d4g4 e5e1 c1b2 e8f8 g2g3 g8f6 g4c4 c7c6 f1g2 e1e2 h1f1 f6d7 g2f3 e2e5 b2c1 f8g8 c4f4 d7c5');
INSERT INTO evaluations VALUES(56957,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(56957,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(56957,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(56957,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(56957,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(56957,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(56957,36,'cp 52','e2e4 d7d5 b1c3 f8b4 g1e2 b8c6 e4e5 g8e7 d1d2 e7g6 f4e3 c6a5 e2g3 c7c5 a2a3 c5d4 e3d4 b4c3 d4c3 a5c6 e1c1 d8g5 f1e2 g5d2 d1d2 g6f4 e2f1 c8d7 g3e2 f4g6 f2f4 h6h5 g2g3 h5h4 f1h3 h4g3 h2g3 e8c8 b2b3 g6e7 c1b2 c8b8 e2d4 c6d4 c3d4 e7c6');
INSERT INTO evaluations VALUES(5695,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(5695,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(5695,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(5695,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(5695,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(5695,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(5695,36,'cp 39','c8g4 c1e3 e7e5 d4d5 c6d4 f1e2 d4e2 d1e2 g7h6 a2a4 h6e3 e2e3 g4f3 e3f3 g8f6 a4a5 e8g8 e1g1 a7a6 f3e3 f6g4 e3g3 g4f6 b2b4 f6d7 g3e3 d8e7 f1c1 f7f5 c4c5 d6c5 c3a4 f5e4 a4c5 d7c5 b4c5 a8d8 c1d1 d8d7 a1c1 e7f7 d5d6');
INSERT INTO evaluations VALUES(6532,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(6532,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(6532,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(6532,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(6532,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(6532,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(6532,36,'cp -31','d7d5 f5g7 e8f8 g7h5 d8h4 h5g3 g8f6 f1e2 h8g8 b1c3 d5e4 e1g1 c8e6 c1e3 c5d6 c3b5 d6e5 b5d4 c6d4 e3d4 a8d8 d4e5 d8d1 a1d1 h4g5 e5c7 h7h5 c7d6 f8e8 h2h4 g5h4 e2b5 f6d7 d1d4 h4g5 b5d7 e6d7 d4e4 d7e6 d6f4 g5g6 f4e3 e8d8 f1d1 d8c8 d1d3 b7b6 e4d4 g8h8 d3c3 c8b7 d4h4 h8d8 h4h5');
INSERT INTO evaluations VALUES(13288,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(13288,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(13288,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(13288,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(13288,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(13288,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(13288,36,'cp 42','g1f3 g7g6 d2d4 c5d4 f3d4 c8d7 e1g1 f8g7 b2b3 e8g8 d4c2 a7a6 c1b2 d8a5 a1c1 a8b8 b2a1 f8c8 c2e3 c6b4 d1d2 b4c6 h2h3 b7b5 f2f4 b5c4 e3c4 a5c5 g1h1 c5d4 d2e1 d7e6 g3g4 d4a7 f4f5 e6c4 b3c4 c6e5 g4g5');
INSERT INTO evaluations VALUES(13287,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(13287,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(13287,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(13287,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(13287,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(13287,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(13287,36,'cp 30','c8d7 d2d4 c5d4 f3d4 g7g6 d4c2 f8g7 a1b1 a8c8 e1g1 h7h5 h2h4 e8g8 b2b3 d7f5 e2e4 f5g4 f2f3 g4d7 a2a4 a7a6 c1e3 d8a5 d1d2 d7e6 c2d4 c6d4 e3d4 f6d7 d4g7 g8g7 g1h2 d7e5 h2h1 g7g8 b1c1 a5c5 f1d1 e5c6');
INSERT INTO evaluations VALUES(15295,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(15295,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(15295,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(15295,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(15295,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(15295,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(15295,36,'cp -59','d8g5 c4d5 g5g6 d2d3 c8d7 h2h4 f7f5 h4h5 g6f6 c3b5 e8c8 d5c6 b7c6 b5c3 g7g5 c1e3 h8g8 f2f3 c6c5 d1d2 d7c6 e1c1 g5g4 f3f4 e5f4 e3f4 g4g3 h1f1 g8g4 a2a3 f8d6 f4d6 d8d6 c1b1 c5c4 d2e2 d6e6 e2d2 g4d4 c3e2');
INSERT INTO evaluations VALUES(970,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(970,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(970,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(970,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(970,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(970,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(970,36,'cp -50','c4d5 c6d4 d2d3 g5g6 c1e3 c8g4 d1d2 e8c8 f2f3 g4h3 e1g1 f8b4 d5c4 h8e8 a2a3 b4e7 g1h1 h3e6 c4e6 g6e6 c3e2 d4f5 e3f2 h7h5 d2a5 a7a6 a1e1 g7g5 b2b4 h5h4 e2c3 f7f6 b4b5 f5g3 h2g3 h4g3 b5a6 g3f2 a6b7 c8b7 a5b5 b7c8');
INSERT INTO evaluations VALUES(15522,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(15522,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(15522,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(15522,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(15522,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(15522,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(15522,36,'cp -12','e5d4 f3g5 d6d5 e4d5 d8e7 c4e2 f6d5 e1g1 h7h6 g5f3 e7f6 f1e1 c8e6 e2b5 f8d6 d1d4 f6f5 b1d2 e8g8 d4d3 f5d3 b5d3 f8e8 h2h3 a8d8 d3f1 d6f8 d2b3 a7a6 c1d2 g7g5 h3h4 f7f6 f3d4 c6d4 b3d4 e6f7 e1e8 d8e8 g2g3 a6a5 a1d1 a5a4 c3c4 d5b4 h4g5');
INSERT INTO evaluations VALUES(16131,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(16131,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(16131,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16131,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16131,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16131,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16131,36,'cp 8','f8c5 b2b4 c5b6 a2a4 a7a6 e1g1 d7d6 a4a5 b6a7 d2d3 e8g8 c1e3 a7e3 f2e3 b7b5 a5b6 c7b6 b1d2 c6e7 h2h3 d8c7 c4b3 c8d7 c3c4 e7c6 f3h4 c7d8 b4b5 a6b5 a1a8 d8a8 c4b5 c6a5 b3c4 a8d8 h4f5 d7e6 c4e6');
INSERT INTO evaluations VALUES(9763,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(9763,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(9763,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(9763,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(9763,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9763,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9763,36,'cp 14','b2b4 c5b6 a2a4 a7a6 e1g1 d7d6 a4a5 b6a7 d2d3 c8e6 c4e6 f7e6 c1e3 e8g8 e3a7 c6a7 d3d4 e5d4 f3d4 d8d7 c3c4 a7c6 d4c6 d7c6 b1d2 e6e5 f1d1 f8f7 f2f3 f6h5 g2g3 a8f8 f3f4 h5f6 f4f5 f8e8 g1g2 c6d7 a1a3');
INSERT INTO evaluations VALUES(20146,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(20146,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(20146,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(20146,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(20146,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(20146,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(20146,36,'cp -95','b7c6 c2c3 c8a6 g1e2 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 d8b6 d1c2 a8c8 f1d1 f8e7 c3a4 b6a5 e2c3 e8g8 b2b3 f8d8 c2d2 h7h6 c1b2 a5b4 a1c1 c6c5 a4c5 e7c5 d4c5 c8c5 c3e2 b4d2 d1d2 c5c1 b2c1 d5d4 e2g3 f5g3 h2g3 d8d5');
INSERT INTO evaluations VALUES(22901,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(22901,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22901,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(22901,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(22901,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(22901,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(22901,36,'cp -6','f8e7 c1g5 d7d6 g5e7 d8e7 e4e7 g8e7 e2e3 c6b4 e1d2 c8e6 b1c3 e8c8 g1f3 b4c6 c3b5 h7h6 f1e2 h8e8 b2b3 a7a6 b5d4 e6d7 d4c6 d7c6 e2d3 e7g6 a1g1 g6e5 d3f5 c6d7 f5d7 e5d7 d2e2 d6d5 c4d5 d7b6 g1c1 b6d5');
INSERT INTO evaluations VALUES(6991,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(6991,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(6991,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(6991,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(6991,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(6991,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(6991,36,'cp -9','c1g5 d7d6 g5e7 d8e7 e4e7 g8e7 e2e3 c6b4 e1d2 c8e6 g1f3 e8c8 b1c3 b4c6 c3b5 c6e5 b5a7 c8b8 a7b5 e5c4 f1c4 e6c4 b5c7 b8c7 h1c1 d6d5 b2b3 b7b5 b3c4 b5c4 f3d4 e7c6 d4b5 c7b6 a2a4 b6c5 c1b1 d8b8 d2c3 g7g5');
INSERT INTO evaluations VALUES(23368,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(23368,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(23368,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(23368,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(23368,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(23368,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(23368,36,'cp 37','h7h6 g5e3 d7d6 c2c4 f7f5 e4f5 g8f6 f1d3 e7f5 g1e2 f6g4 e3d2 d8h4 g2g3 h4h3 e2g1 h3g2 d1f3 g2f3 g1f3 c8d7 b1c3 e8c8 e1g1 f5d4 f3e1 g7g5 g1g2 f8g7 f2f3 g4f6 h2h3 d7f5 d3f5 d4f5 e1c2 d8f8 a1e1 f8f7 a2a4 c8b8 a4a5 h8f8');
INSERT INTO evaluations VALUES(26505,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(26505,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(26505,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(26505,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26505,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(26505,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(26505,36,'cp 26','e7e6 c2c4 d5c4 b1c3 f8d6 f4g3 g8f6 b5c4 d6g3 h2g3 e8g8 g1f3 d8e7 a2a3 f8d8 c4b5 c6b8 d1b3 c7c5 e1g1 c5d4 f3d4 f5g6 b5e2 e6e5 d4f3 b8c6 f1d1 e5e4 f3h4 a7a6 b3a4 d8b8 g3g4 b7b5 h4g6 h7g6 a4c2 c6e5');
INSERT INTO evaluations VALUES(28905,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(28905,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(28905,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(28905,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(28905,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(28905,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(28905,36,'cp 73','d1b3 d8b6 c4d5 b6b3 a2b3 f6d5 e3e4 d5c3 e4f5 c3d5 c1d2 e7e6 f5e6 f7e6 f1c4 b8d7 e1g1 e8f7 f1e1 f8d6 g2g3 h8e8 e1e2 a7a5 a1a5 a8a5 d2a5 e8a8 a5c3 b7b5 c4d3 d6b4 c3b4 d5b4 d3e4 a8a1 g1g2 g7g5 h2h4 g5g4 f3e5 d7e5 d4e5 a1c1 f2f3 g4f3 g2f3 c1c5 f3e3 c5c1 g3g4');
INSERT INTO evaluations VALUES(29809,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(29809,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29809,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29809,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(29809,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(29809,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(29809,36,'cp 25','e1g1 e7e6 d2d4 f8d6 c4b3 e8g8 h2h3 g4h5 c2c4 d5f6 c4c5 d6e7 g2g4 h5g6 b1c3 c7c6 c1f4 b7b6 c5b6 a7b6 f3e5 f6d7 e5g6 h7g6 d1f3 d7f6 f1d1 e7d6 f4d2 d6e7 a1c1 d8d4 d2e3 d4e5 e3b6 e7d6');
INSERT INTO evaluations VALUES(29808,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(29808,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(29808,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(29808,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(29808,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(29808,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(29808,36,'cp 24','e7e6 d2d4 d5f6 c2c3 f8e7 h2h3 g4h5 c4e2 e8g8 f3e5 h5e2 d1e2 c7c5 d4c5 e7c5 f1d1 d8c7 c1f4 b8c6 b1d2 f6d5 f4g3 c6e5 g3e5 c5d6 e5d6 c7d6 d2f3 d6c7 d1d4 d5f6 a1d1 a8d8 e2e5 d8d4 d1d4 c7b6 b2b3 f8d8 g2g3 h7h6 d4d8 b6d8 f3d4 b7b6 a2a4 f6d5');
INSERT INTO evaluations VALUES(30046,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30046,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(30046,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(30046,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(30046,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(30046,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(30046,36,'cp 30','d2d4 d5f6 c2c3 f8e7 h2h3 g4h5 c4e2 e8g8 f3e5 h5e2 d1e2 c7c5 d4c5 e7c5 f1d1 d8c7 c1f4 b8c6 b1d2 f6d5 f4h2 c6e5 h2e5 c5d6 e5d6 c7d6 d2f3 d6c7 d1d4 d5f6 a1d1 f8d8 e2e3 d8d4 e3d4 h7h6 c3c4 c7b6 d4b6 a7b6 a2a3 f6e4 d1d4 e4c5 d4d6 c5a4');
INSERT INTO evaluations VALUES(33160,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(33160,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(33160,32,'cp 262',NULL);
INSERT INTO evaluations VALUES(33160,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(33160,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(33160,35,'cp 245',NULL);
INSERT INTO evaluations VALUES(33160,36,'cp 282','b1c3 b8c6 c3d5 f6g6 f1d3 g8e7 e1g1 g6h5 d5e7 d6e7 d3e2 h5g6 d4d5 c6d4 f3d4 e5d4 d1d4 e8g8 c1f4 d7d6 a1e1 e7f6 d4c4 f6b2 c4c7 b2a3 e2d3 b7b6 c7c6 a3c5 f4d6 c5d6 e4e5 c8f5 d3f5 g6f5 e5d6 a8c8 c6a4 f8d8 e1e7 f5f6 e7a7 d8d6 c2c4 f6e5 a4b4');
INSERT INTO evaluations VALUES(38575,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(38575,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(38575,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(38575,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(38575,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(38575,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(38575,36,'cp 83','g8e7 h2h4 c7c5 d4c5 b8d7 c1f4 h7h6 c3b5 e8g8 d1d2 d7c5 f1d3 c8d7 b5d4 c5d3 d2d3 e7f5 d4f5 e6f5 e1g1 a8c8 f1d1 d8b6 b2b3 d7e6 c2c3 b6a5 d1c1 f8e8 f3d4 a7a6 a2a4 a5d8 d3g3 g7f8 c1d1 g8h7 a4a5 e6d7 f4d2 d8e7 d1e1 f8g7 g3f4');
INSERT INTO evaluations VALUES(41136,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(41136,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(41136,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(41136,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(41136,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(41136,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(41136,36,'cp 80','h2h4 c7c5 d4c5 b8d7 c1f4 h7h6 c3b5 e8g8 d1d2 d7c5 f1d3 c8d7 b5d4 c5d3 d2d3 e7f5 e1c1 d8c7 c1b1 f8c8 d4f5 e6f5 f3d4 c7b6 d3f3 c8e8 h1e1 a8c8 c2c3 d7e6 f3e3 e6d7 g2g3 g8h7 e3f3 d7e6 d4e6 f7e6 f3e3 b6e3');
INSERT INTO evaluations VALUES(40987,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(40987,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40987,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(40987,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40987,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40987,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40987,36,'cp 5','c7c5 f1e2 b8d7 e1g1 e8g8 b1d2 b7b6 c2c3 c8b7 a2a4 d8c8 f3e5 d7e5 f4e5 a7a6 e2d3 c5c4 d3c2 b6b5 d1e2 b7c6 e3e4 b5b4 e5f4 d5e4 d2c4 f6d5 f4d2 f7f5 c4a5 c8c7 a5c6 c7c6 c3c4 e7d6');
INSERT INTO evaluations VALUES(42972,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42972,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42972,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(42972,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42972,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(42972,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(42972,36,'cp 3','e8g8 f1d3 c7c5 c2c3 b7b6 b1d2 c8a6 d3a6 b8a6 e1g1 d7d5 d1e2 a6b8 a1d1 b8c6 f3e5 a8c8 h2h3 h7h6 g5h4 c5d4 e3d4 f6d7 e5c6 c8c6 h4g3 d8a8 d2f3 f8c8 f1e1 a8b7 h3h4 b6b5 a2a3 d7f6');
INSERT INTO evaluations VALUES(37300,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(37300,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(37300,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(37300,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(37300,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(37300,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37300,36,'cp 12','f1d3 c7c5 c2c3 b7b6 b1d2 c8b7 e1g1 d7d5 g5f4 b8d7 f3e5 f6e4 a2a4 d7e5 f4e5 e7d6 d2e4 d5e4 e5d6 d8d6 d3e2 a7a5 d4c5 d6c5 d1c2 g7g6 f1d1 f8d8 h2h3 b7c6 d1d2 g8g7 a1d1 d8d2 d1d2');
INSERT INTO evaluations VALUES(50412,30,'cp 241',NULL);
INSERT INTO evaluations VALUES(50412,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(50412,32,'cp 230',NULL);
INSERT INTO evaluations VALUES(50412,33,'cp 242',NULL);
INSERT INTO evaluations VALUES(50412,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(50412,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(50412,36,'cp 239','d2d4 f4g6 h2h4 e7e6 g1f3 f8b4 b1d2 g6e7 a2a3 b4d2 c1d2 d7d5 c4c5 b7b6 h4h5 h7h6 b2b4 c7c6 a3a4 b8d7 a1c1 f7f6 e5f6 d7f6 f1d3 b6c5 d4c5 e8g8 d1e2 d8c7 b4b5 f6g4 d2c3 e6e5 c3e5 g4e5 f3e5 c7a5 e2d2 a5a4 b5c6 c8f5 e1g1 a4d4');
INSERT INTO evaluations VALUES(50413,30,'cp 238',NULL);
INSERT INTO evaluations VALUES(50413,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(50413,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(50413,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(50413,34,'cp 258',NULL);
INSERT INTO evaluations VALUES(50413,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(50413,36,'cp 234','f4g6 h2h4 e7e6 g1f3 f8b4 b1d2 h7h6 f1d3 d7d5 c4c5 g6f4 d3f1 f4g6 a2a3 b4d2 c1d2 b7b6 b2b4 b6c5 h4h5 g6e7 d4c5 c8b7 f1b5 c7c6 b5a4 b7a6 f3d4 e8g8 d1g4 g8h8 a4c2 a6c4 f2f4 a7a5 e1f2 a5b4 a3b4 b8d7 g4f3 d8c7 d2c3 a8a2 a1a2 c4a2');
INSERT INTO evaluations VALUES(50763,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50763,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(50763,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(50763,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(50763,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(50763,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50763,36,'cp 41','g8f6 f1e2 f8e7 g2g4 d6d5 e4d5 f6d5 c3d5 d8d5 e2f3 d5c4 d1e2 c4e2 f3e2 c8d7 e1c1 b8c6 h1e1 a8c8 e2f3 b7b5 h2h3 h7h6 a2a3 c6a5 c1b1 a5c4 e3f4 e7g5 f4g5 h6g5 f3b7 c8c5 b7g2 e8e7 d4f5 e7f6 d1d7 e6f5 g4f5 c5f5');
INSERT INTO evaluations VALUES(52711,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(52711,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52711,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52711,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(52711,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(52711,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52711,36,'cp 6','c6d5 d2d4 b8c6 g1f3 c8g4 h2h3 g4h5 c1f4 e7e6 f1d3 f8d6 f4d6 d8d6 e1g1 g8e7 f1e1 e8g8 b1d2 h7h6 a2a4 d6c7 d1e2 h5g6 d3g6 e7g6 g2g3 g6e7 d2f1 e7f5 g1g2 f5d6 h3h4 a8c8 a1c1 c8d8 h4h5 d6e4');
INSERT INTO evaluations VALUES(53418,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53418,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53418,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(53418,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(53418,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53418,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(53418,36,'cp 12','d2d4 b8c6 f1d3 g8f6 c1f4 c8g4 d1b3 d8d7 b1d2 e7e6 d2f3 a7a6 f3e5 c6e5 f4e5 g4h5 g1e2 f8d6 e2f4 h5g6 b3b6 d6e5 d4e5 f6g8 h2h4 g6d3 f4d3 a8c8 h4h5 g8e7 h5h6 g7g6 e1c1 c8c6 b6e3 e7f5 e3f4');
INSERT INTO evaluations VALUES(52826,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(52826,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(52826,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(52826,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(52826,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(52826,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(52826,36,'cp -90','g4g5 h7h6 d2d4 h6g5 c1g5 d8b6 g1f3 c8f5 d1c1 g8f6 c2c4 b8d7 a2a4 f6e4 a4a5 b6b4 g5d2 e4d2 b1d2 e7e6 h2h4 a7a6 c1c3 d5c4 c3b4 f8b4 a1a4 b4e7 a4c4 f7f6 c4c3 e8c8 c3b3 f5g6 b3e3 d7f8 d2c4 g6f7 c4b6 c8b8 e3b3 c6c5');
INSERT INTO evaluations VALUES(3823,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(3823,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(3823,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(3823,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(3823,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(3823,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(3823,36,'cp -20','c5e3 f2e3 c6a5 c4b3 a5b3 a2b3 c8d7 c3e2 a7a5 d1e1 a5a4 b3a4 a8a4 a1a4 d7a4 b2b3 a4d7 h2h3 h7h6 e1a5 d8b8 e2g3 f8e8 f3h4 b8d8 h4f3 b7b6 a5d2 b6b5 d2a5 d8c8 g1h2 c8b8 f1a1 g7g6 a5a6 b5b4 h2g1 g8g7 a1a5 h6h5 a6a7 b8c8');
INSERT INTO evaluations VALUES(3884,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(3884,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(3884,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(3884,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(3884,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(3884,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(3884,36,'cp -12','f2e3 c6a5 c4b3 a5b3 a2b3 a7a5 d1e1 c8d7 c3e2 a5a4 b3a4 a8a4 a1a4 d7a4 b2b3 a4d7 h2h3 h7h6 e1a5 b7b6 a5d2 d8a8 e2g3 b6b5 d2c3 a8b8 c3d2 f8e8 d2a5 b8c8 f3h4 c8d8 h4f3 d8b8 f1a1 b8c8 g1h2 g7g6 c2c4 b5c4');
INSERT INTO evaluations VALUES(3882,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(3882,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(3882,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(3882,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(3882,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(3882,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(3882,36,'cp -19','c6a5 c4b3 a5b3 a2b3 a7a5 d1e1 c8d7 c3e2 a5a4 b3a4 a8a4 a1a4 d7a4 b2b3 a4d7 h2h3 f8e8 e2g3 h7h6 e1c3 d8b8 c3d2 b7b5 g1h2 g8h7 f3h4 b8d8 h4f3 g7g6 b3b4 h7g7 f1a1 d8e7 a1a5 d7c6 a5a6 c6b7 a6a7');
INSERT INTO evaluations VALUES(5462,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5462,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5462,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5462,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5462,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5462,36,'cp 0','e7e6 c4b3 g8e7 c3e2 e8g8 c2c3 d6d5 e4d5 e7d5 c1g5 d5f6 d1c1 b7b6 e2g3 d8c7 f1e1 c8a6 g5f4 c7e7 f4g5');
INSERT INTO evaluations VALUES(7601,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(7601,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(7601,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(7601,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(7601,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(7601,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(7601,36,'cp 38','d7d6 b1c3 f6d7 c1f4 e8g8 d1d2 d7e5 d3e2 a7a5 c3a4 c5a7 a1e1 f7f5 f4e5 d6e5 d2d8 f8d8 e2c4 g8h8 e4f5 c8f5 e1e5 f5c2 a4c5 d8d2 f1e1 a8f8 e5e2 d2e2 e1e2 a7c5 e2c2 f8d8 g2g3 g7g6 g1f1 c5b4 f2f4 a5a4 f1g2 d8d4 b2b3');
INSERT INTO evaluations VALUES(7562,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(7562,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(7562,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(7562,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(7562,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(7562,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(7562,36,'cp 49','b1c3 f6d7 c3a4 c5d4 c2c3 d4f6 f2f4 e8g8 d1c2 g7g6 c1e3 d7b6 a1e1 f8e8 e3f2 c6c5 e4e5 f6g7 d3e4 a8b8 a4b6 a7b6 f2g3 d6e5 f4e5 c8a6 f1f3 e8e6 e1d1 d8e8 e4d5 a6b7 d5e6 b7f3 e6f7 e8f7 g2f3 f7f3 d1e1 f3c6 e5e6 b8e8');
INSERT INTO evaluations VALUES(8596,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(8596,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(8596,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8596,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8596,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8596,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8596,36,'cp 7','c6a5 c4b3 b7b5 d3d4 c8b7 d4e5 f6e4 b3d5 d6e5 d5b7 d8d1 f1d1 a5b7 d1e1 b7d6 f3e5 e7f6 e5d3 e8c8 a2a4 h8e8 a4b5 e4c5 e1e8 d8e8 g1f1 c5d3 c2d3 d6b5 a1a5 e8e5 c1e3 e5d5 b1a3 a7a6 a5a6 f6b2 a3b5 d5b5 f1e2 c8d7 g2g3 b2f6 a6a4 f6e7 e2f3 h7h5 a4a7 b5b3 d3d4 e7d6 a7a8 g7g6 f3e4');
INSERT INTO evaluations VALUES(10943,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(10943,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(10943,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(10943,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(10943,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(10943,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(10943,36,'cp 109','e4e5 f6e4 c4d5 f7f5 e1g1 d7d6 c3d4 d6e5 d5c6 b7c6 f3e5 e7f6 c1f4 e8g8 f2f3 g7g5 f3e4 g5f4 e5c6 d8d6 e4e5 d6c6 e5f6 f8f6 b1c3 c8b7 d1d2 f6g6 d4d5 c6b6 g1h1 b6e3 a1d1 a8e8 h2h3 c7c6 d5d6 c6c5 c3d5 g6d6 d5e3 d6d2 d1d2 f4e3 d2d7 b7e4 h1g1 c5c4 d7a7 e8b8 f1d1 e3e2 d1e1 b8b2 g1f2');
INSERT INTO evaluations VALUES(10865,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(10865,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(10865,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(10865,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(10865,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(10865,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(10865,36,'cp 98','f6e4 c4d5 f7f5 c3d4 d7d6 e1g1 d6e5 d5c6 b7c6 f3e5 e7f6 c1f4 e8g8 f2f3 g7g5 f3e4 g5f4 e5c6 d8e8 e4e5 e8c6 e5f6 f8f6 b1c3 c8b7 d1d2 f6g6 d4d5 c6b6 g1h1 b6e3 a1d1 c7c6 d5d6 e3d2 d1d2 a8d8 d6d7 c6c5 c3d5 f4f3 g2f3 g8h8');
INSERT INTO evaluations VALUES(10961,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(10961,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(10961,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(10961,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(10961,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(10961,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(10961,36,'cp 103','c4d5 f7f5 e1g1 d7d6 c3d4 d6e5 d5c6 b7c6 f3e5 e7f6 c1f4 e8g8 f2f3 e4g5 f4g5 f6g5 d1b3 d8d5 b3d5 c6d5 f3f4 g5e7 b1c3 e7d6 c3d5 c8b7 d5c3 f8d8 a1d1 h7h6 f1f2 a8b8 g2g3 b7a8 b2b3 a7a5 g1f1 a8b7 f2d2 g7g5 c3a4 b7d5 a4c5 b8b5 f4g5 h6g5 d2f2 d6c5 d4c5');
INSERT INTO evaluations VALUES(13382,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(13382,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(13382,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(13382,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(13382,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(13382,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(13382,36,'cp 155','d4d5 c6b8 g1f3 g7g6 f3d4 f8g7 f1e2 e8g8 e1g1 e7e6 d5e6 c8e6 d4e6 f7e6 d1b3 d8e7 c1e3 b8c6 a1d1 a8b8 h2h3 a7a6 a2a4 f8e8 b3a3 b8d8 e2c4 f6h5 c3e2 g8h8 a4a5 d6d5 a3e7 c6e7 e4d5 e6d5 e3b6');
INSERT INTO evaluations VALUES(14178,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(14178,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(14178,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(14178,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(14178,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(14178,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(14178,36,'cp -44','e7e5 d4e5 f6e4 c1e3 f8c5 e3c5 e4c5 e2e3 c8g4 d1c2 c5d7 b1c3 d7e5 f3e5 c6e5 h2h3 g4c8 e1c1 c7c6 e3e4 d8g5 c1b1 e8g8 g2g3 d5e4 c3e4 g5e7 f2f4 e5g6 e4f2 c8d7 f4f5 g6e5 f2g4 a8d8 d1e1 f7f6 g4e5 f6e5 g3g4 d7e8 f1c4 e8f7 c4f7 f8f7 e1d1 h7h6 a2a3 g8h8 d1d8 e7d8 h1d1 d8e7 d1e1');
INSERT INTO evaluations VALUES(15108,30,'cp 295',NULL);
INSERT INTO evaluations VALUES(15108,31,'cp 274',NULL);
INSERT INTO evaluations VALUES(15108,32,'cp 280',NULL);
INSERT INTO evaluations VALUES(15108,33,'cp 280',NULL);
INSERT INTO evaluations VALUES(15108,34,'cp 295',NULL);
INSERT INTO evaluations VALUES(15108,35,'cp 280',NULL);
INSERT INTO evaluations VALUES(15108,36,'cp 287','d4e5 f6e4 d1d5 c8f5 f3d4 c6d4 e3d4 e4c3 d5d8 a8d8 b2c3 f5c2 f1c4 f8e7 e1e2 f7f6 f2f4 c2e4 c1e3 c7c5 h1g1 c5d4 c3d4 e7a3 a1b1 e4b1 g1b1 b7b6 b1b3 a3e7 d4d5 f6e5 f4e5 h8f8 c4b5 e8f7 d5d6 e7d6 b5c4 f7e8 e5d6 d8d6 g2g4 e8e7');
INSERT INTO evaluations VALUES(15072,30,'cp 299',NULL);
INSERT INTO evaluations VALUES(15072,31,'cp 274',NULL);
INSERT INTO evaluations VALUES(15072,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(15072,33,'cp 278',NULL);
INSERT INTO evaluations VALUES(15072,34,'cp 258',NULL);
INSERT INTO evaluations VALUES(15072,35,'cp 278',NULL);
INSERT INTO evaluations VALUES(15072,36,'cp 301','f6e4 d1d5 e4c3 d5d8 c6d8 b2c3 d8c6 c1b2 c8g4 f1d3 g4f3 g2f3 c6e5 f3f4 e5d3 c2d3 e8c8 e1e2 d8d5 c3c4 d5a5 a2a4 a5h5 a1g1 h8g8 g1g5 h5h3 g5g3 h3h5 b2d4 a7a6 g3g5 h5h3 c4c5 f8e7 g5g3 h3h5 c5c6 g7g6 c6b7 c8b7 d4e5 g8e8 h1b1 b7c6 h2h3 h5f5 e5c3');
INSERT INTO evaluations VALUES(15372,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(15372,31,'cp 288',NULL);
INSERT INTO evaluations VALUES(15372,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(15372,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(15372,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(15372,35,'cp 304',NULL);
INSERT INTO evaluations VALUES(15372,36,'cp 310','d1d5 e4c3 d5d8 c6d8 b2c3 d8c6 c1b2 f8c5 f1d3 e8g8 e1e2 f8e8 c3c4 a8b8 h2h3 g7g6 d3e4 c5f8 a1b1 f8g7 h1d1 c6a5 b2d4 b7b6 c4c5 a5c4 c5b6 a7b6 e4c6 e8d8 a2a4 c8e6 b1a1 c4a5 c6b5 c7c6 b5d3 a5c4 d1b1 c6c5 d4c3 e6d5 e3e4 d5e6 d3c4 e6c4 e2e1 g7f8');
INSERT INTO evaluations VALUES(17701,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(17701,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(17701,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(17701,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(17701,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(17701,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(17701,36,'cp 100','c2c4 e7e6 b1d2 f8d6 f4g3 g8f6 a1c1 e8g8 f1d3 c6a5 c4d5 e6d5 g3h4 f8e8 e1g1 c8g4 f1e1 g7g5 h4g3 f6e4 d2f1 a8c8 d1a4 a5c6 g3d6 c7d6 a4b3 d8d7 h2h3 g4e6 f3h2 f7f5 f2f3 e4f6 f1g3 c6e7 c1c8 e8c8 b3b4');
INSERT INTO evaluations VALUES(21377,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(21377,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(21377,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(21377,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(21377,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(21377,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(21377,36,'cp 48','e2e3 f8d6 f4g3 g8f6 b1c3 e8g8 g1f3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 e1g1 c8f5 a1c1 f5d3 d1d3 a7a5 c1c2 c7c6 f1c1 d6g3 h2g3 d8d6 f3e1 f8c8 d3d1 h7h6 e1d3 c8c7 a2a3 a8e8 d1f3 e7g6 b2b4 a5b4 d3b4 g6e7 f3f4 d6d7 b4d3 f7f6');
INSERT INTO evaluations VALUES(24218,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(24218,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(24218,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(24218,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(24218,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(24218,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(24218,36,'cp 131','a7a6 b5a4 e8g8 b1c3 d7g4 h2h3 g4h5 a2a3 b7b5 a4c2 d8d7 c1e3 f8e8 a1c1 a8c8 b2b4 h7h6 d1e2 e7d8 e2f1 h5f3 g2f3 c6e7 g1h1 e7g6 f1g2 g6h4 g2g3 f6h5 g3g4 d7g4 h3g4 h5f6 c2d1 c8a8 d1e2 c7c6 h1g1 f6d7 c3b1 d6d5 c1c6 d5e4');
INSERT INTO evaluations VALUES(29957,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(29957,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(29957,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(29957,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(29957,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(29957,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(29957,36,'cp 71','h2h3 g4f3 d1f3 f8e7 f1c4 b8c6 c4b5 f6d7 b5c6 b7c6 f4f5 h7h5 b2b3 d7f6 c1b2 d6d5 d2d3 h5h4 c3a4 e7b4 c2c3 b4a5 e1g1 a5b6 a4b6 a7b6 a2a4 d5e4 d3e4 d8d2 f1f2 d2g5 b3b4 g5g3 c3c4 g3f3 f2f3 c6c5 b2e5');
INSERT INTO evaluations VALUES(32338,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(32338,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32338,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(32338,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(32338,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(32338,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32338,36,'cp 0','c7c5 g1f3 g8f6 e1g1 b8c6 d4c5 f8c5 c2c4 e8g8 b1c3 d5c4 d3c4 d8a5 f1d1 a8c8 a2a3 c5e7 c4b3 a5a6 f4d6 c6a5 b3b4 e7d6 d1d6 a5c6 b4b3 c6a5');
INSERT INTO evaluations VALUES(37227,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37227,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(37227,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37227,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37227,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(37227,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37227,36,'cp 0','g1f3 d7d6 e1g1 b8d7 d2d3 c7c6 a2a4 h7h6 c1e3 a7a5 c4a2 f8e8 f3d2 b7b6 d1f3 d7f8 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(37223,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37223,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37223,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37223,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37223,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37223,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37223,36,'cp -14','d7d6 e1g1 f8e8 d2d3 c7c6 f3g5 e8e7 g5f3 h7h6 a2a3 d6d5 c4b3 c5b6 f3h4 e7e8 d1f3 c8e6 h4f5 g8h8 f1e1 b6c7 b3a2 d5d4 c3e2 b8d7 a2e6 f7e6 f5g3 h8g8 g3h5 g8h7 h5f6 d7f6 g1h1 a7a5 c1d2 d8e7 e1g1');
INSERT INTO evaluations VALUES(37794,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(37794,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(37794,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(37794,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(37794,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(37794,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(37794,36,'cp 93','d4d5 d7d6 e2e4 e7e6 b1c3 e6d5 c4d5 g8f6 f3d2 e8g8 f1e2 f8e8 a2a4 b8d7 e1g1 g6g5 f1e1 h7h6 d1c2 d7e5 d2f1 a7a6 f1g3 a8b8 c1d2 e5g6 c3d1 b7b5 d1e3 b5b4 e2d3 f6d7 e3f5 d7e5 d3e2 d8f6 f5g7 g8g7');
INSERT INTO evaluations VALUES(37839,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(37839,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(37839,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(37839,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(37839,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37839,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(37839,36,'cp 56','b7b5 c4b5 a7a6 b5a6 c8a6 g2g3 d7d6 f1g2 g8f6 b1c3 b8d7 e1g1 d7b6 f3d2 e8g8 a2a4 b6d7 a1b1 a8b8 d2f3 d8b6 b2b3 b6b4 f1e1 c5c4 c3a2 b4b7 b3b4 b7d5 a2c3 d5a8 e2e4 a6b7 d1e2 d6d5 e4e5 f6e4 c3e4 d5e4');
INSERT INTO evaluations VALUES(41182,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(41182,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(41182,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(41182,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(41182,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(41182,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(41182,36,'cp 17','c7c6 b5e2 e8g8 a2a4 b8d7 h2h3 e7e5 d4e5 d6e5 e2c4 d8a5 c1d2 a5b4 b2b3 b4e7 e1g1 d7c5 d1e2 c8e6 b3b4 e6c4 e2c4 c5a6 a1b1 f8d8 d2g5 a6c7 g5e3 f6e8 b4b5 e8d6 c4b4 a7a5 b4a3 e7e6 b5c6 b7c6 f1d1 d6c4');
INSERT INTO evaluations VALUES(40285,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40285,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(40285,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(40285,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40285,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40285,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(40285,36,'cp 24','b5e2 e8g8 a2a4 b8d7 a4a5 d8c7 c1e3 e7e5 e1g1 e5d4 e3d4 f8e8 e2c4 g7h6 h2h3 d7e5 f3e5 d6e5 d4c5 c8e6 c4e6 e8e6 d1d3 a8d8 d3c4 b7b6 a5b6 a7b6 c5b4 h6f8 b4f8 g8f8 f1d1 d8d1 a1d1 b6b5 c4c5 f8g7');
INSERT INTO evaluations VALUES(42803,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(42803,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(42803,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(42803,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42803,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42803,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42803,36,'cp 25','b2c3 e8g8 d2d3 c7c6 g1f3 f8e8 e1g1 d7d5 d1b3 h7h6 c4d5 c6d5 c3c4 d5c4 b3c4 b8c6 c4b5 c8d7 b5b2 a8c8 c1e3 b7b6 a1c1 d8e7 f1d1 e7b4 b2a1 b4a3 h2h3 f6d5 e3d2 e5e4 f3e1 e4d3 e1d3 e8e2');
INSERT INTO evaluations VALUES(44720,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(44720,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(44720,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(44720,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(44720,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(44720,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(44720,36,'cp 110','g1f3 c8f5 f1e2 d8c7 f3h4 f5g6 c2c3 b8d7 e1g1 f8g7 f1e1 e7e6 g2g3 e8g8 a2a4 a8d8 a4a5 f8e8 d1b3 b7b6 a5b6 a7b6 c1f4 e6e5 f4e3 f6f5 d4e5 g7e5 h4g2 e5d6 b3a2 d7e5 e3g5 d8b8 b2b4 e5g4 e2f1 b6b5');
INSERT INTO evaluations VALUES(44719,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(44719,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(44719,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(44719,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(44719,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(44719,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(44719,36,'cp 101','c8g4 f1e2 b8d7 e1g1 e7e6 c2c4 f8g7 d1b3 d8b6 b3a3 f6f5 c1g5 c6c5 a1d1 e8g8 d4c5 d7c5 g5e3 f8c8 b2b4 b6a6 a3a6 c5a6 a2a3 e6e5 h2h3 g4h5 g2g4 h5g6 f3h4 f5f4 e3c1 c8d8 h4g6 h7g6 d1d8 a8d8 c4c5 a6c7 e2f3 c7e6 f3b7');
INSERT INTO evaluations VALUES(28899,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(28899,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(28899,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(28899,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(28899,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(28899,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(28899,36,'cp 96','f1e2 b8d7 e1g1 e7e6 c2c4 f8g7 a2a4 e8g8 a1a3 f6f5 h2h3 g4h5 f1e1 d8c7 a3d3 f5f4 d3d2 a8d8 b2b4 f8e8 a4a5 h5g6 c1b2 c6c5 d1b3 c5d4 f3d4 a7a6 d4f3 e6e5 f3h4 e5e4 h4g6 h7g6 c4c5 g7b2 b3b2');
INSERT INTO evaluations VALUES(45395,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(45395,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45395,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(45395,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(45395,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(45395,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45395,36,'cp 17','g1f3 e7e6 c2c4 b8c6 c4d5 e6d5 f1b5 f8d6 d2d4 c5d4 f3d4 e8g8 e1g1 f8e8 b1d2 c8d7 h2h3 a8c8 d4c6 b7c6 b5d3 h7h6 d1f3 d8e7 f1d1 d7e6 a1c1 d6a3 b2f6 e7f6 f3f6 g7f6 c1c2 a7a5 d2f3 c6c5 e3e4');
INSERT INTO evaluations VALUES(47779,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(47779,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(47779,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(47779,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(47779,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(47779,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(47779,36,'cp -10','f8e7 a2a4 h7h6 g5h4 b8c6 h4g3 c8g4 g1f3 c6d4 b1d2 f6h7 c2c3 d4f3 d2f3 h7g5 h2h4 g5f3 g2f3 g4h5 d3d4 e7f6 e1f1 c7c6 d1d3 e8g8 d4e5 d6e5 d3d8 f8d8 f1e2 g8f8 a1d1 f8e7 b2b4 d8d1 h1d1 a8d8 d1d8 e7d8 c4b3 d8c7 a4a5 b7b5 a5b6 c7b6');
INSERT INTO evaluations VALUES(40754,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(40754,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(40754,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40754,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40754,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40754,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40754,36,'cp 0','a2a4 h7h6 g5h4 b8d7 b1c3 c7c6 g1f3 d7c5 h2h3 c8e6 a4a5 e6c4 h4f6 e7f6 d3c4 e8g8 e1g1 d8d7 d1d2 c5e6 f1d1 f8d8 d2e3 e6d4 e3d3 d4e6');
INSERT INTO evaluations VALUES(49791,30,'cp 315',NULL);
INSERT INTO evaluations VALUES(49791,31,'cp 312',NULL);
INSERT INTO evaluations VALUES(49791,32,'cp 325',NULL);
INSERT INTO evaluations VALUES(49791,33,'cp 326',NULL);
INSERT INTO evaluations VALUES(49791,34,'cp 319',NULL);
INSERT INTO evaluations VALUES(49791,35,'cp 312',NULL);
INSERT INTO evaluations VALUES(49791,36,'cp 301','h5e5 d8e7 e5e7 f8e7 d2d4 h6g8 b1c3 d7d6 c1e3 c7c6 f2f3 g8f6 g2g4 e8g8 g1e2 b7b5 a2a3 b8a6 e2f4 f6d7 h2h4 a6c7 e1f2 a7a6 f1d3 d7b6 a3a4 b6a4 c3a4 b5a4 a1a4 d6d5 h4h5 c8d7 e4d5 c6d5 a4a5 e7b4 a5a1 a6a5 c2c3');
INSERT INTO evaluations VALUES(49789,30,'cp 315',NULL);
INSERT INTO evaluations VALUES(49789,31,'cp 303',NULL);
INSERT INTO evaluations VALUES(49789,32,'cp 301',NULL);
INSERT INTO evaluations VALUES(49789,33,'cp 311',NULL);
INSERT INTO evaluations VALUES(49789,34,'cp 311',NULL);
INSERT INTO evaluations VALUES(49789,35,'cp 310',NULL);
INSERT INTO evaluations VALUES(49789,36,'cp 320','d8e7 e5e7 f8e7 d2d4 c7c6 b1c3 h6g8 c1e3 d7d6 f2f3 g8f6 g2g4 e8g8 g1e2 b7b5 a2a3 b8a6 e2f4 a6c7 g4g5 f6d7 h2h4 c8b7 e1c1 a7a5 a3a4 b5a4 c3a4 f8b8 c2c3 b7a6 f1a6 a8a6 f4d3 c6c5 d4c5 d7c5 e3c5 d6c5 a4c5 a6c6 c5d7 b8b5 d3e5');
INSERT INTO evaluations VALUES(34060,30,'cp 304',NULL);
INSERT INTO evaluations VALUES(34060,31,'cp 298',NULL);
INSERT INTO evaluations VALUES(34060,32,'cp 299',NULL);
INSERT INTO evaluations VALUES(34060,33,'cp 314',NULL);
INSERT INTO evaluations VALUES(34060,34,'cp 305',NULL);
INSERT INTO evaluations VALUES(34060,35,'cp 313',NULL);
INSERT INTO evaluations VALUES(34060,36,'cp 317','e5e7 f8e7 d2d4 h6g8 b1c3 d7d6 c1e3 g8f6 f2f3 e8g8 g1e2 c7c6 g2g4 b7b5 a2a3 b8a6 e2f4 a6c7 h2h4 f6d7 f1d3 d7b6 d4d5 c8b7 e1c1 c6d5 e4d5 b6c4 d3c4 b5c4 g4g5 f8e8 e3d4 e8b8 d1e1 b8e8 h1g1 a7a6');
INSERT INTO evaluations VALUES(50517,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(50517,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(50517,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(50517,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(50517,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(50517,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50517,36,'cp 36','f1d3 d7d5 e4e5 c8d7 g1f3 b8c6 a2a3 d8b6 d3c2 a8c8 e1g1 c5d4 c3d4 g7g5 h2h3 h7h5 c1g5 b6b2 b1d2 c6d4 c2d3 d7b5 a1b1 b2a3 f3d4 a3d3 d4b5 a6b5 b1b3 d3c2 d1c2 c8c2 f1a1 f8h6 g5h6 h8h6 d2f3 e8f8 a1a8 f8g7 a8b8 g8e7 b8b7 e7c6 b3b5');
INSERT INTO evaluations VALUES(50826,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(50826,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(50826,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(50826,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(50826,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(50826,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(50826,36,'cp -85','b5a3 e7e5 a3c4 b8c6 g1f3 f7f6 d2d3 c8e6 c2c3 d8d7 f1e2 e8c8 c3d4 e5d4 d1a4 g7g5 e1g1 g8e7 b2b4 e6c4 d3c4 g5g4 f3e1 c6e5 a4d7 d8d7 e1d3 e5d3 e2d3 e7c6 a2a3 c6e5 f1d1 c7c5 a1b1 e5d3 d1d3 f8d6 f2f3 g4f3');
INSERT INTO evaluations VALUES(50864,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(50864,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(50864,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(50864,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(50864,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(50864,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(50864,36,'cp -94','e7e5 d2d3 c8e6 g1f3 f8b4 c1d2 b4d2 d1d2 f7f6 f1e2 g8e7 c2c3 c7c5 e2d1 e8g8 d1b3 e6b3 a2b3 b8c6 e1g1 d8d7 c3d4 c6d4 f3d4 d7d4 a3c2 d4d6 b3b4 c5b4 d2b4 d6b4 c2b4 a6a5 b4c2 e7c6 f1d1 f8d8 a1c1 d8d7 c2e3 a5a4 g2g3');
INSERT INTO evaluations VALUES(50995,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(50995,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(50995,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(50995,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(50995,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(50995,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(50995,36,'cp -81','d2d3 c8e6 g1f3 f8b4 c1d2 b4d2 d1d2 f7f6 c2c3 c7c5 f1e2 b8c6 e1g1 g8e7 e2d1 b7b5 d1b3 e6b3 a2b3 e8g8 c3d4 c5d4 b3b4 f6f5 a3c2 e7g6 e4f5 f8f5 d2e2 d8d5 f1d1 g6f4 e2e4 f5f6 f3g5 f6g6 e4d5 f4d5 g5e4 d5b4 c2b4');
INSERT INTO evaluations VALUES(50993,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(50993,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(50993,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(50993,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(50993,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(50993,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(50993,36,'cp -76','b8c6 f1e2 h7h6 a3c4 c8e6 g1f3 e6c4 d3c4 g8f6 e2d3 a6a5 e1g1 f8c5 a2a3 f6d7 f3e1 d8e7 d3e2 c5d6 e1d3 d7c5 d3c5 d6c5 e2d3 a5a4 f2f4 h6h5 f4f5 f7f6 h2h3 e8c8 g1h2 c8b8 c1d2 c5d6 b2b4 a4b3');
INSERT INTO evaluations VALUES(57648,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57648,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(57648,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(57648,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(57648,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(57648,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57648,36,'cp 6','e7e5 b1c3 f8g7 g2g3 c7c6 d2d4 e5d4 e3d4 d7d5 c4d5 g8e7 c1g5 h7h6 g5e7 d8e7 g1e2 e8g8 f1g2 c8g4 e1g1 c6d5 h2h3 g4e2 d1e2 b8c6 e2e7 c6e7 f1e1 e7c6 a1d1 a8d8 g2d5 g7d4 d5c6 b7c6 e1e7 d4c3 d1d8 f8d8 b2c3 a7a5 e7c7 g8g7 c7c6');
INSERT INTO evaluations VALUES(1190,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(1190,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(1190,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(1190,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(1190,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(1190,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(1190,36,'cp -121','e1d1 c8e6 h1e1 e4g2 e5f7 g8e7 e1e6 g7g6 h5h4 g2f3 e6e2 f3f7 h4h3 f8g7 b1c3 e8g8 c3e4 h7h6 h3d7 e7d5 d7f7 f8f7 d2d3 f7f3 c1e3 a8f8 d1d2 f3h3 a1h1 g6g5 e4c5 g7b2 c2c4 d5f4 e3f4 f8f4 h1b1 b2a3 c5e6 f4f3 b1b3 a3f8 e6f8 g8f8 b3b7 f3d3 d2c2 d3d7 b7a7 h3h2 a7a6 g5g4 a6c6');
INSERT INTO evaluations VALUES(1192,30,'cp -139',NULL);
INSERT INTO evaluations VALUES(1192,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(1192,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(1192,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(1192,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(1192,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(1192,36,'cp -122','c8e6 h1e1 e4g2 e5f7 g8e7 e1e6 g7g6 h5h4 g2f3 e6e2 f3f7 h4h3 f8g7 b1c3 e8g8 c3e4 h7h6 h3d7 e7d5 d7f7 f8f7 a2a4 f7f4 c2c4 d5b4 a1a3 a8d8 a3g3 g8h7 d2d3 f4h4 e4g5 h6g5 c1g5 h4d4 g5d8 d4d8 e2e7 b4d3 e7c7 d3b2 d1e2 b2a4 c7b7 a4c5 b7e7 d8d7');
INSERT INTO evaluations VALUES(4105,30,'cp -286',NULL);
INSERT INTO evaluations VALUES(4105,31,'cp -283',NULL);
INSERT INTO evaluations VALUES(4105,32,'cp -283',NULL);
INSERT INTO evaluations VALUES(4105,33,'cp -274',NULL);
INSERT INTO evaluations VALUES(4105,34,'cp -280',NULL);
INSERT INTO evaluations VALUES(4105,35,'cp -294',NULL);
INSERT INTO evaluations VALUES(4105,36,'cp -293','f8f7 g5f7 g8f7 d2d3 h7h6 c3a4 c5b4 c2c3 b4e7 c3c4 e7b4 a4c3 d7d6 e1g1 c8g4 f2f3 g4h5 c3d5 b4c5 g1h1 f7g8 c1d2 a7a5 a2a3 c5d4 d1e1 h5f7 f3f4 e5f4 f1f4 d4b2 a1b1 b2e5 f4f1 b7b6 d2c3 e5c3 e1c3 f6d5 c4d5 c6e5 d3d4');
INSERT INTO evaluations VALUES(4643,30,'cp -284',NULL);
INSERT INTO evaluations VALUES(4643,31,'cp -287',NULL);
INSERT INTO evaluations VALUES(4643,32,'cp -288',NULL);
INSERT INTO evaluations VALUES(4643,33,'cp -267',NULL);
INSERT INTO evaluations VALUES(4643,34,'cp -265',NULL);
INSERT INTO evaluations VALUES(4643,35,'cp -271',NULL);
INSERT INTO evaluations VALUES(4643,36,'cp -270','g5f7 g8f7 d2d3 d7d5 c1g5 d5e4 c3e4 c5e7 e1g1 f6e4 g5e7 f7e7 d3e4 d8d1 f1d1 c8e6 f2f3 c6d4 g1f2 a7a5 d1d3 b7b6 a2a4 a8f8 h2h4 e6d7 d3d2 c7c5 f2e3 f8d8 e3f2 d7e6 f2e3 h7h5 a1d1');
INSERT INTO evaluations VALUES(4640,30,'cp -286',NULL);
INSERT INTO evaluations VALUES(4640,31,'cp -284',NULL);
INSERT INTO evaluations VALUES(4640,32,'cp -274',NULL);
INSERT INTO evaluations VALUES(4640,33,'cp -291',NULL);
INSERT INTO evaluations VALUES(4640,34,'cp -284',NULL);
INSERT INTO evaluations VALUES(4640,35,'cp -280',NULL);
INSERT INTO evaluations VALUES(4640,36,'cp -293','g8f7 d2d3 d7d5 c1g5 d5e4 g5f6 g7f6 c3e4 c5e7 e1g1 f6f5 d1h5 f7g8 e4g3 f5f4 g3f5 e7g5 f5h6 g5h6 h5h6 d8f8 h6h4 c8f5 a1d1 f8d6 c2c3 f5d3 d1d2 a8d8 f1d1 d6g6 f2f3 g8f7 h4g4 g6g4 f3g4 e5e4 d2f2 c6e5');
INSERT INTO evaluations VALUES(5992,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(5992,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(5992,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(5992,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(5992,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5992,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5992,36,'cp 7','f4e5 d6e5 d1e2 c6d4 f3d4 c5d4 c2c3 d4b6 d2d3 d8h4 e1d2 c8g4 e2f1 e8c8 d2c2 a7a5 b1d2 f7f6 c4f7 d8d7 f7e8 d7d8 g2g3 h4h3 f1h3 g4h3 e8h5 b6c5 d2c4 b7b6 c1e3 c5e3 c4e3 g8e7 a2a4 c8b7 h5g4 h3g4 e3g4 h7h5 g4e3 h5h4 a1f1');
INSERT INTO evaluations VALUES(6677,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(6677,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(6677,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(6677,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(6677,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(6677,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(6677,36,'cp -34','b1d2 g8f6 c2c3 b4c5 f1e2 e8g8 g1f3 f8e8 e1g1 d7d5 e4d5 d8d5 d2b3 d5d1 f1d1 c5b6 d1e1 a7a5 c1f4 a5a4 b3d4 c6d4 f3d4 b6d4 c3d4 c7c6 a2a3 c8e6 a1c1 a8d8 f4e5 e6b3 e2f3 f6d5 g1f1 f7f6 e5g3 e8e1 f1e1 g8f7');
INSERT INTO evaluations VALUES(27301,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(27301,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(27301,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(27301,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(27301,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(27301,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(27301,36,'cp 75','c1e3 g8f6 b1c3 g7g6 d1d2 d7b6 b2b3 f8g7 e1c1 d8e7 h1e1 e8c8 c3d5 e7d7 d5f6 g7f6 a2a4 d7c6 e3d4 f6d4 d2d4 c6c5 d4c5 d6c5 d1d8 c8d8 f3e5 d8c8 f2f3 h8f8 c1d2 f7f6 e5g4 a7a5 g4e3 f8d8 d2c3 c7c6 h2h4 h7h5 e1g1 b6d7 e3c4 b7b6 g1e1');
INSERT INTO evaluations VALUES(28012,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(28012,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(28012,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(28012,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(28012,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(28012,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(28012,36,'cp 98','b1c3 e7e6 c1f4 g8e7 h2h4 e8g8 h4h5 c7c5 d4c5 b6c5 f1d3 e7d5 c3d5 b7d5 c2c4 d5f3 d1f3 b8c6 d3e4 c6e5 f4e5 g7e5 e4a8 d8a8 f3a8 f8a8 e1c1 d7d6 h1h3 a7a5 h3b3 g8g7 d1h1 g6g5 b3b7 h7h6 c1c2 g5g4 f2f3 g4g3 b7b5 a8f8 b5a5');
INSERT INTO evaluations VALUES(31390,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(31390,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(31390,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31390,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(31390,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(31390,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31390,36,'cp 41','f1d3 f5d3 d1d3 g8e7 e1g1 h6h5 f3h4 g7g6 c1e3 b8a6 h2h3 e7f5 h4f5 e6f5 b2b3 a6c7 c2c4 c7e6 b1d2 f8e7 d2f3 a7a5 a2a4 d8d7 g1h1 e8g8 f1d1 f8c8 d3d2 b7b6 a1c1 c8b8 c1c2 b6b5 a4b5 c6b5 c4c5 d7c6 c2a2 a5a4 b3b4');
INSERT INTO evaluations VALUES(31389,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(31389,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(31389,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(31389,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(31389,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(31389,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31389,36,'cp 29','g8e7 e1g1 f5d3 d1d3 h6h5 c1e3 e7f5 e3f2 f8e7 b1d2 a7a5 c2c3 d8b6 h2h3 b8d7 f1e1 a5a4 d2f1 c6c5 d3c2 g7g6 e1c1 h5h4 f1e3 f5e3 f2e3 c5c4 g1h2 b6a6 b2b4 b7b5 a2a3 d7b8 c1g1 e8d7 g2g4 h4g3 h2g2');
INSERT INTO evaluations VALUES(31408,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31408,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(31408,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31408,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(31408,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(31408,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(31408,36,'cp 32','f1d3 h7d3 d1d3 g8f6 c1d2 f8e7 e1c1 b8d7 c1b1 e8g8 g3e4 c6c5 g2g4 f6g4 d3e2 f7f5 e4c5 d7c5 d4c5 d8e8 f3e1 e7c5 e1d3 c5d6 h1g1 g8h7 d2c3 f8f7 d1e1 a8d8 f2f3 g4f6 e2g2 f7d7 g2g6 e8g6 h5g6 h7g8 f3f4');
INSERT INTO evaluations VALUES(31407,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(31407,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(31407,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31407,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(31407,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(31407,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(31407,36,'cp 27','h7d3 d1d3 g8f6 c1d2 b8d7 e1c1 f8e7 g3e4 e8g8 c1b1 c6c5 g2g4 f6g4 d3e2 d8b6 f3e5 d7e5 d4e5 f7f5 e5f6 g4f6 d1g1 f6e4 e2e4 e7f6 c2c3 g8h8 g1g6 b6d6 d2h6 d6d5 e4d5 e6d5 h1g1 f8f7 h6d2 a8e8 h5h6 d5d4 h6g7 f6g7 b1c1 b7b5 c1d1 d4c3 b2c3');
INSERT INTO evaluations VALUES(31399,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(31399,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(31399,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31399,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(31399,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(31399,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(31399,36,'cp 29','d1d3 b8d7 c1d2 f8e7 e1c1 g8f6 c1b1 e8g8 g3e4 c6c5 g2g4 f6g4 d3e2 d8b6 f3e5 d7e5 d4e5 f7f5 e5f6 g4f6 d1g1 f6e4 e2e4 e7f6 c2c3 g8h8 g1g6 b6d6 d2h6 d6d5 e4d5 e6d5 h6d2 h8h7 h1g1 f8f7 h5h6 d5d4 f2f3 a8e8 b1c2 b7b5 c3d4 f6d4');
INSERT INTO evaluations VALUES(31398,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(31398,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(31398,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31398,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(31398,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31398,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(31398,36,'cp 9','g8f6 c1d2 f8e7 e1c1 e8g8 g3e4 b8d7 c1b1 c6c5 g2g4 f6g4 d3e2 d8b6 f3e5 d7e5 d4e5 f7f5 e5f6 g4f6 h1g1 f6e4 e2e4 e7f6 b2b3 a8d8 d2h6 d8d1 g1d1 b6b4 e4b4 c5b4 h6e3 b7b6 d1d6 f6h4 d6e6 h4f2 e3f2 f8f2 h5h6 g7h6 e6h6 g8f7 h6h7 f7e6 h7a7 e6d6 a7g7 d6c6 g7g4 c6c5 g4g5 c5c6 g5g6 c6c5');
INSERT INTO evaluations VALUES(36294,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(36294,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36294,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36294,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36294,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36294,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36294,36,'cp -8','e6d5 f1d3 f8e8 e1g1 e7d6 f1e1 c8g4 e3e4 d5e4 c3e4 f6e4 e1e4 g4h5 c1g5 d8c8 h2h3 b8c6 e4h4 h5g6 d1b3 c8d7 d3g6 h7g6 g5f4 d6e7 h4g4 c6a5 b3c2 a5c6');
INSERT INTO evaluations VALUES(37939,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(37939,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(37939,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(37939,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(37939,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(37939,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(37939,36,'cp 76','f1e2 g8f6 b1c3 e8g8 h2h3 e7e5 d4d5 b8a6 c1e3 c6d5 c4d5 f6e8 a2a3 f7f5 b2b4 f5f4 e3d2 c8d7 c3b5 h7h6 a3a4 a6c7 b5c7 d8c7 a1c1 c7d8 e2d3 g6g5 a4a5 a8c8 c1c8 d8c8 b4b5 g7f6 d1b3');
INSERT INTO evaluations VALUES(38888,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(38888,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(38888,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(38888,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(38888,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(38888,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(38888,36,'cp 154','b2c3 g8f6 c1a3 b8d7 g1f3 b7b6 d1a4 c8b7 f3e5 a7a6 e5c6 d8c8 e1g1 f6e4 c4b3 e4d6 d4d5 e6d5 b3d5 d7f6 d5f3 c8d7 f1d1 e8g8 a3d6 c7d6 a4b4 a6a5 b4d6 b7c6 f3c6 d7d6 d1d6 a8c8 c6f3 c8c3 d6b6 f6d7 b6a6 f8c8 g2g3 c3c5 a1b1 d7e5 f3b7 c8b8 a6a7 b8b7 b1b7');
INSERT INTO evaluations VALUES(40369,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(40369,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(40369,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(40369,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(40369,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(40369,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(40369,36,'cp 101','e2e4 e7e6 f1d3 e8g8 h2h3 e6d5 e4d5 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1c2 e4f6 e3d2 b8a6 a2a3 a6c7 a1e1 e8e1 f1e1 d8d7 c2b3 b7b6 g2g3 a8e8 g1g2 e8e1 f3e1 d7e8 b3c2 f6d7 f2f4 g8f8 b2b3 h7h6 e1f3 b6b5 f3h2 d7b6 f4f5 b5c4 b3c4');
INSERT INTO evaluations VALUES(40779,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(40779,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40779,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40779,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(40779,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40779,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(40779,36,'cp 20','c7c5 b1c3 e8g8 d2d3 b8c6 a2a3 h7h6 h2h3 c8e6 c3d5 e6d5 e4d5 c6d4 f3d4 c5d4 c2c3 d4c3 b2c3 f6d7 a1b1 d7b6 c4b3 e7g5 c1b2 b6d7 d3d4 a8c8 a3a4 f8e8 g2g3 g5f6 h3h4 b7b6 d1h5 d8c7 b1a1 c8b8 h5e2 e5d4');
INSERT INTO evaluations VALUES(43309,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(43309,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(43309,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(43309,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(43309,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(43309,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(43309,36,'cp -12','b5d7 c8d7 c3e2 d7b5 c2c3 b8c6 h2h4 b5e2 d1e2 c5d4 f3d4 a8c8 d4f3 h7h6 c1e3 d8c7 e1f1 c6e7 g2g3 e7f5 h4h5 f5e3 e2e3 f8c5 e3e2 c5e7 f3d4 c7c4 e2c4 c8c4 f2f4 e8g8');
INSERT INTO evaluations VALUES(43308,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(43308,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(43308,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(43308,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43308,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43308,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(43308,36,'cp 0','c8d7 c3e2 d7b5 c2c3 b8c6 h2h4 b5e2 d1e2 c5d4 f3d4 a8c8 d4f3 h7h6 c1e3 d8c7 e1f1 c6a5 f1g1 a5c6 g1f1');
INSERT INTO evaluations VALUES(28527,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(28527,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(28527,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(28527,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28527,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28527,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28527,36,'cp 0','c3e2 d7b5 c2c3 b8c6 h2h4 b5e2 e1e2 c5d4 c3d4 f7f6 c1e3 f8e7 h4h5 f6e5 f3e5 c6e5 d4e5 e8g8 h5h6 d8d7 h6g7 d7b5 d1d3 b5b2 d3d2 b2b5 d2d3');
INSERT INTO evaluations VALUES(44888,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(44888,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44888,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(44888,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(44888,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44888,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(44888,36,'cp 15','e7e6 d2d4 b8d7 d1c2 f8d6 f1d3 d5c4 d3c4 e8g8 e1g1 e6e5 b2b3 d8e7 a2a4 e5e4 f3g5 d6h2 g1h2 f6g4 h2g1 e7g5 c2e4 d7b6 c4e2 b6d5 c3d5 c6d5 e4f3 g4f6 e3e4 g5g6 e4e5 c8g4 e2d3 g4f3 d3g6 f7g6 e5f6 f3e4 f6g7 g8g7 c1a3 f8e8 a1c1 a8c8 f1e1 c8c1 e1c1');
INSERT INTO evaluations VALUES(45620,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(45620,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(45620,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(45620,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(45620,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45620,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(45620,36,'cp 20','d2d4 b8d7 d1c2 f8d6 f1d3 d5c4 d3c4 e8g8 e1g1 a7a6 f1d1 b7b5 c4d3 d8c7 c3e4 f6e4 d3e4 d7f6 e4d3 c6c5 d4c5 c7c5 a2a4 c8b7 c2c5 d6c5 a4b5 a6b5 a1a8 f8a8 d3b5 b7d5 f3e5 d5b3 d1f1 a8b8 b5c4 b3c4 e5c4 b8b4 c4e5 f6d5');
INSERT INTO evaluations VALUES(46085,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(46085,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(46085,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(46085,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(46085,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(46085,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(46085,36,'cp -74','b8d7 b5e2 a7a6 c3b1 d8c7 b1d2 b7b5 a2a4 b5b4 c2c4 b4c3 b2c3 c5c4 d1c2 f8d6 c2a2 e6e5 d4e5 d7e5 c1a3 c8f5 a3d6 c7d6 e1g1 e8g8 f1d1 e5f3 d2f3 d6c7 a1c1 a8b8 h2h3 f8d8 d1d4 f6e4 f3d2 e4c5 e2f3 f5e6 a2a3 c5b3');
INSERT INTO evaluations VALUES(48204,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(48204,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(48204,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(48204,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(48204,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(48204,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(48204,36,'cp -33','f8e7 g1f3 b8d7 f1d3 c7c5 a2a4 e8g8 e1g1 b7b6 f3e5 d7e5 d4e5 f6d7 g5e7 d8e7 f2f4 a7a6 b2b3 f7f6 d1h5 g7g6 h5h4 c8b7 e5f6 d7f6 h4g5 a8c8 h2h3 c5c4 b3c4 e7c5 a1e1 d5c4 g5c5 c8c5 d3e2 f6e4 e2f3 e4c3 f3b7 c3a4 b7a6 a4c3');
INSERT INTO evaluations VALUES(40927,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(40927,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(40927,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(40927,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(40927,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(40927,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(40927,36,'cp -29','g1f3 b8d7 f1d3 c7c5 a2a4 e8g8 e1g1 a7a6 b2b3 h7h6 g5h4 b7b6 d1e1 c8b7 f3e5 d7e5 d4e5 f6d7 h4e7 d8e7 f2f4 f8d8 a1b1 a8c8 h2h3 c5c4 b3c4 d5c4 d3e4 d7c5 e4b7 e7b7 e1e2 c5e4 c3e4 b7e4 b1b6 c4c3 b6d6 e4a4');
INSERT INTO evaluations VALUES(48471,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(48471,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48471,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48471,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48471,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48471,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48471,36,'cp 0','c7c5 g1f3 a7a6 e1g1 b7b5 c4e2 b8d7 d4c5 f8c5 b2b4 c5e7 b1d2 c8b7 c1b2 e8g8 a1c1 a8c8 c1c8 d8c8 d1a1 c8a8 b2d4 f8c8 f1c1 c8c1 a1c1 h7h6 h2h3 b7d5 c1a1 e7d6 d4b2 a8b7 e2d1 b7a8');
INSERT INTO evaluations VALUES(55571,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(55571,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(55571,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(55571,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(55571,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(55571,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(55571,36,'cp 122','d4e5 b8c6 g1f3 c8g4 b1d2 d6e5 f4e5 g8f6 e5g3 f8d6 e2e3 d6g3 h2g3 d8e7 c2c3 e8c8 d1a4 c8b8 f1b5 g4d7 e1g1 a7a6 b5e2 h7h5 f1d1 h8e8 a4h4 d7g4 d2b3 d8d1 e2d1 c6e5 d1e2 f6e4 h4e7 e8e7 a1d1 b8c8 c3c4 f7f5 d1c1 g7g5 c4c5 e5c6 a2a3 c6e5 c1d1');
INSERT INTO evaluations VALUES(55515,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(55515,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(55515,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(55515,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(55515,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(55515,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(55515,36,'cp 106','b8c6 g1f3 c8g4 e5d6 d8f6 d6d7 g4d7 d1c1 f6g6 g2g3 e8c8 f1g2 h7h6 e1g1 g6e6 c1e3 g8e7 e3b3 e7d5 f4d2 d5b6 b3e6 d7e6 b2b3 g7g5 e2e3 b6d5 c2c4 d5b4 d2c3 h8g8 f3e1 f8g7 a2a3 b4d3 c3g7 g8g7');
INSERT INTO evaluations VALUES(56128,30,'cp -501',NULL);
INSERT INTO evaluations VALUES(56128,31,'cp -505',NULL);
INSERT INTO evaluations VALUES(56128,32,'cp -503',NULL);
INSERT INTO evaluations VALUES(56128,33,'cp -509',NULL);
INSERT INTO evaluations VALUES(56128,34,'cp -504',NULL);
INSERT INTO evaluations VALUES(56128,35,'cp -499',NULL);
INSERT INTO evaluations VALUES(56128,36,'cp -519','c7c6 d5c3 d4c3 b2c3 g8f6 d2d3 f8c5 g1f3 e8g8 f1e2 f8e8 e1g1 c5b6 a1b1 b8d7 h2h3 d7f8 f1e1 h7h6 e2f1 b6c7 a2a4 g8h8 g2g3 c6c5 f1g2 c7a5 c1d2 f8g6 d3d4 e5d4 e4e5 f6d5 c3d4 d5c3 d2c3 a5c3 d4c5 c3e1 d1d8');
INSERT INTO evaluations VALUES(53685,30,'cp -499',NULL);
INSERT INTO evaluations VALUES(53685,31,'cp -499',NULL);
INSERT INTO evaluations VALUES(53685,32,'cp -512',NULL);
INSERT INTO evaluations VALUES(53685,33,'cp -513',NULL);
INSERT INTO evaluations VALUES(53685,34,'cp -495',NULL);
INSERT INTO evaluations VALUES(53685,35,'cp -495',NULL);
INSERT INTO evaluations VALUES(53685,36,'cp -494','d5c3 d4c3 b2c3 g8f6 d2d3 f8c5 g1f3 e8g8 f1e2 f8e8 e1g1 c5b6 h2h3 b8d7 f1e1 h7h6 a2a4 d7f8 e2f1 f8g6 d3d4 b6a5 a1a3 e5d4 e4e5 f6h7 c3d4 a5e1 d1e1 c6c5 d4c5 d8c7 e1b4 g6e5 f3e5 e8e5 a3e3 e5e3 c1e3 c8e6 e3d2 h7f6');
INSERT INTO evaluations VALUES(3468,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3468,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(3468,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(3468,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(3468,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3468,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(3468,36,'cp 32','f6h5 h2h4 c8g4 d3e2 h5f4 e3f4 f7f6 d2f1 e7e5 f1e3 g4f3 e2f3 e5e4 f3e2 c6e7 h4h5 f6f5 d1d2 d8d6 g2g3 c7c6 e1c1 b7b5 f2f3 a7a5 f3e4 d5e4 c1b1 a8b8 g3g4 d6f4 e3f5 f4d2 f5e7 g8f7 d1d2 f7e7 h5g6 h7g6 d4d5 c6d5 h1h7 f8f7 d2d5 b5b4 c3c4 b8d8 d5a5');
INSERT INTO evaluations VALUES(4355,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4355,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4355,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4355,36,'cp 5','d7d6 e1g1 c6d4 c1e3 d4f3 d1f3 c5e3 f3e3 c7c6 c4b3 d8b6 e3b6 a7b6 b1c3 c8e6 a1d1 e6b3 c2b3 f8d8 d3d4 f6d7 d4d5 d7c5 b3b4 c5b3 f2f4 b3d4 g1f2 c6c5 c3e2 d4e2 f2e2 c5b4 a3b4 a8a2 d1d2 e5f4 f1f4 h7h6');
INSERT INTO evaluations VALUES(5018,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(5018,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(5018,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(5018,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(5018,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(5018,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(5018,36,'cp -23','g8f6 f1d3 e8g8 e1g1 f8e8 a2a3 b4f8 d1f3 a7a5 f3g3 g8h8 a1d1 h7h6 f1e1 a5a4 d3c4 d8e7 c3b1 d7d6 d2b4 e8d8 b1d2 c8g4 d1b1 d8e8 c4f1 h8g8 d2c4 e7c7 g3d3 e8e6 b4c3');
INSERT INTO evaluations VALUES(7658,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(7658,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(7658,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(7658,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(7658,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(7658,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7658,36,'cp -6','f1d3 e8g8 e1g1 f8e8 a2a3 b4f8 d1f3 d7d5 f1e1 e8e6 d2g5 c8b7 e4d5 c6d5 d3f5 e6e8 a1d1 e5e4 f3h3 h7h6 g5f6 d8f6 c3d5 b7d5 d1d5 f6b2 h3e3 a8d8 d5d8 e8d8 f5e4 b2a3 e3a3 f8a3 e1a1 a3b2 a1b1 b2f6 g1f1 a7a5 b1b5 d8d1 f1e2');
INSERT INTO evaluations VALUES(7660,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(7660,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(7660,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(7660,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(7660,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7660,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(7660,36,'cp -6','e8g8 e1g1 f8e8 a2a3 b4f8 d1f3 c8b7 d2g5 d7d5 f1d1 h7h6 g5f6 d8f6 f3f6 g7f6 e4d5 c6d5 d3f5 e8d8 d1d3 f8e7 a1d1 d5d4 c3e4 b7c8 g2g4 a8b8 b2b4 h6h5 h2h3 c8a6 d3d2 h5g4 h3g4 a6c4 e4g3 a7a5 b4a5 b8a8 f5d3 c4e6 a5a6 e6g4 d1e1 g4e6 f2f4');
INSERT INTO evaluations VALUES(2889,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(2889,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(2889,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(2889,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(2889,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(2889,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(2889,36,'cp -20','e1g1 f8e8 a2a3 b4f8 d1f3 c8b7 f3g3 d7d5 f1e1 d8c7 b2b4 e8e6 a1d1 a8e8 g3f3 h7h6 f3e2 d5d4 c3a4 c6c5 a4c5 f8c5 b4c5 f6d7 c2c3 d7c5 c3d4 e5d4 f2f3 e6c6 d1c1 e8c8 d2b4 c5d3 c1c6 c7c6 e2d3');
INSERT INTO evaluations VALUES(7212,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(7212,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(7212,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(7212,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(7212,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(7212,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(7212,36,'cp 213','e5c4 e8g8 c4d6 c7d6 b1c3 d6d5 e1g1 f6e8 c1f4 d5e4 c3e4 c8f5 f1e1 f5g6 a2a4 e8c7 a4a5 c7e6 f4g3 d8d7 e4d6 e6d4 h2h4 d4f5 d6e4 h7h6 h4h5 f5g3 e4g3 g6f5 g3f5 d7f5 e1e7 f8d8 e7b7 c6c5');
INSERT INTO evaluations VALUES(7235,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(7235,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(7235,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(7235,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(7235,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(7235,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(7235,36,'cp 218','e8g8 c4d6 c7d6 e1g1 a6a5 b1c3 b7b5 a2a3 d6d5 c1f4 h7h6 h2h3 f8e8 f1e1 d5d4 c3e2 c6c5 e2g3 d8b6 f4d2 c8b7 d1f3 a5a4 f3f4 a8a6 f4h4 b6d8 f2f4 b7c8 g3h5 a6c6 h5f6 c6f6 h4f2 f6g6 f4f5');
INSERT INTO evaluations VALUES(15916,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(15916,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(15916,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(15916,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(15916,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(15916,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(15916,36,'cp 86','g1f3 g7g6 f1e2 f8g7 d1c2 e8g8 e1g1 e5d4 f3d4 f8e8 c1e3 a7a6 a1d1 d8e7 e3g5 h7h6 g5h4 c7c5 d4b3 g6g5 h4g3 f6e4 c3e4 e7e4 e2d3 e4e6 d3h7 g8h8 h7f5 e6f6 d1d6 f6e7 h2h3 d7e5 d6d5 b7b6 f5c8 a8c8 f1d1 c8c6 d1e1 e7c7 a2a4 c6g6');
INSERT INTO evaluations VALUES(15915,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(15915,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(15915,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(15915,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(15915,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(15915,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(15915,36,'cp 82','g7g6 f1e2 f8g7 e1g1 e8g8 c1e3 e5d4 f3d4 c7c6 d1c2 f8e8 a1d1 a7a6 h2h3 d8e7 e3g5 h7h6 g5h4 d7c5 e2f3 a6a5 f1e1 a5a4 e1e3 h6h5 h4g3 e8d8 e3e1 f6h7 f3e2 h5h4 g3h2 g7e5 f2f4 e5g7 f4f5 h7g5 e2f1');
INSERT INTO evaluations VALUES(17196,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(17196,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(17196,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(17196,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(17196,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(17196,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(17196,36,'cp 55','d7d6 a2a4 g7g6 f1b5 c8d7 d1d3 a7a6 c1g5 f8g7 b5d7 d8d7 f3d2 h7h6 g5f6 g7f6 a4a5 h6h5 e1g1 f6g5 b2b4 e8g8 d2f3 g5h6 h2h4 c7c6 d5c6 d7c6 f1d1 a8c8 a1a3 f8d8 a3b3 f7f5 f3g5 h6g5 h4g5 f5e4 c3e4 c6c2 d3c2 c8c2 d1d6 d8d6');
INSERT INTO evaluations VALUES(20056,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(20056,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(20056,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(20056,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(20056,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(20056,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(20056,36,'cp -13','g1f3 a7a6 e1g1 g8f6 c1f4 c8g4 h2h3 g4f3 d1f3 e7e6 f1d1 d8a5 c4e2 f8e7 f4d6 e7d6 d1d6 e8g8 a1d1 a8d8 f3e3 d8d6 d1d6 a5c7 d6d2 h7h6 a2a3 f8e8 d2d1 e8c8 f2f4 c6e7 e2f3 e6e5 f4e5 c7e5 e3d4 e7g6 d4e5 g6e5 c3d5 c8c2 d5f6 g7f6 d1d6');
INSERT INTO evaluations VALUES(27596,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(27596,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(27596,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(27596,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(27596,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(27596,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(27596,36,'cp 37','e1g1 c5d4 f3d4 f8c5 d4b3 g8e7 a2a4 e8g8 b1d2 e7f5 b3c5 b6c5 d2b3 c5c4 b3d4 b7b5 a4a5 b5b4 d4f5 e6f5 c1e3 b4c3 b2c3 f8c8 d1d2 c6d7 f1b1 c4c3 d2c3 c8c3 b1b7 d7c8 b7a7 a8a7 e3a7 c8a6 a7b6 h7h6 h2h3 c3c2 f2f3 c2d2 a1a4 d2c2 h3h4');
INSERT INTO evaluations VALUES(27595,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(27595,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27595,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(27595,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27595,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(27595,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(27595,36,'cp 35','c5d4 f3d4 f8c5 d4b3 g8e7 b3c5 b6c5 c1e3 c5a5 f1e1 c6a4 d1d3 a4b5 d3d2 b5a4 f2f4 e7f5 d2f2 f5e3 f2e3 a4c2 b1d2 e8g8 e1c1 c2e4 a2a4 a5a6 c1e1 a8c8 e3f2 h7h6 d2e4 d5e4 e1e4 c8c4 e4c4');
INSERT INTO evaluations VALUES(33560,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33560,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(33560,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33560,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(33560,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(33560,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33560,36,'cp 0','b1c3 e7e6 g1f3 a7a6 d4c5 f8c5 d1d2 c5b4 f1d3 d7d6 e1g1 b6a5 e3e4 b8d7 a2a3 b4c3 d2c3 a5c3 b2c3 d7c5 f3d2 e6e5 f4g5 c8e6 g5f6 g7f6 c3c4 e8e7 f1b1 h8b8 b1b6 c5d7 b6b4 d7c5');
INSERT INTO evaluations VALUES(33558,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33558,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33558,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33558,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33558,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33558,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33558,36,'cp 0','e7e6 g1f3 a7a6 d4c5 f8c5 d1d2 c5b4 f1d3 d7d6 e3e4 b8d7 a2a3 b6a5 e1g1 b4c3 d2c3 a5c3 b2c3 d7c5 f3d2 e6e5 f4g5 c8e6 g5f6 g7f6 c3c4 e8e7 f1b1 h8b8 b1b6 c5d7 b6b1');
INSERT INTO evaluations VALUES(40571,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40571,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40571,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(40571,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40571,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40571,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40571,36,'cp 12','f1b5 b8c6 b5c6 b7c6 f4d6 d8d6 d1a4 e8g8 e1g1 c5d4 c3d4 c6c5 f1c1 c5d4 a4d4 c8d7 b1d2 f8b8 h2h3 b8b4 d4e5 d6e5 f3e5 b4b2 d2b3 d7b5 c1c7 f6e4 c7f7 a8f8 f7f8 g8f8 a1c1 f8e7 f2f4 e7d6 b3d4 a7a6 d4b5 a6b5 c1c6 d6e7 c6c7 e7f8 e5c6 g7g6 c6d4 b2a2 d4e6 f8e8 c7h7');
INSERT INTO evaluations VALUES(42028,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(42028,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(42028,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(42028,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(42028,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(42028,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(42028,36,'cp 41','e4e5 f6d7 f3f4 b4e7 g1f3 c7c5 c1e3 b8c6 d1d2 a7a6 f1d3 c5d4 f3d4 d7c5 e1c1 c6d4 e3d4 b7b5 c1b1 b5b4 c3e2 c5d3 c2d3 a6a5 g2g4 c8d7 h2h4 e8g8 h4h5 a5a4 d1c1 d8a5 h5h6 g7g6 d4e3 b4b3 d2a5 a8a5 e3b6 a5a8 b6c5 e7c5');
INSERT INTO evaluations VALUES(42122,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(42122,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(42122,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(42122,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(42122,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42122,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(42122,36,'cp 17','f6d7 f3f4 b4e7 c1e3 c7c5 g1f3 b8c6 d1d2 a7a6 f1d3 e8g8 d4c5 d7c5 e3c5 e7c5 e1c1 h7h6 c1b1 c5b4 a2a3 b4a5 d2e3 a5c3 b2c3 d8e7 b1b2 c8d7 g2g4 f7f6 h2h4 f6e5 f3e5 c6e5 e3e5 e7f6 e5f6 f8f6 h1f1 a8e8');
INSERT INTO evaluations VALUES(42514,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(42514,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(42514,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(42514,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(42514,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(42514,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(42514,36,'cp 82','b2c3 f6e4 d1g4 f7f5 g4g7 d8f6 g7f6 e4f6 f2f3 b7b6 c3c4 c8a6 e1c1 b8c6 g1h3 c6e7 d2c3 e8f7 h3g5 f7g6 h2h4 h7h6 g5h3 g6f7 d4d5 e6d5 g2g4 d5c4 h3f4 a8d8 f1h3 a6b7 g4g5 h6g5 h4g5 f6h5 g5g6 e7g6 f4g6 f7g6 h1g1 g6f7 c3h8 d8h8 d1d7 f7e6');
INSERT INTO evaluations VALUES(43690,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43690,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43690,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43690,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(43690,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43690,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(43690,36,'cp 26','g1f3 b8c6 a2a4 f8c5 e1g1 d7d6 a4a5 e8g8 c2c3 h7h6 c1e3 c5e3 f2e3 c6e7 b1d2 e7g6 d3d4 f8e8 h2h3 c8d7 d1b3 d8e7 c4d3 e5d4 e3d4 g6f4 d3c2 b7b5 a5b6 c7b6 e4e5 d6e5 f3e5 d7e6 d2c4 b6b5 f1f4 e6c4 e5c4 b5c4');
INSERT INTO evaluations VALUES(43689,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43689,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(43689,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43689,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(43689,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43689,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(43689,36,'cp 15','b8c6 a2a4 f8c5 e1g1 e8g8 c1e3 c5e3 f2e3 d7d6 a4a5 c6e7 d1e1 e7g6 h2h3 c7c6 c4b3 d6d5 e1b4 c8e6 f3g5 f6d7 g5e6 f7e6 f1f8 d8f8 b4f8 g8f8 b1d2 f8e7 e4d5 e6d5 e3e4 g6f4 e4d5 c6d5 a1e1 a8d8 d3d4');
INSERT INTO evaluations VALUES(44524,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(44524,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(44524,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(44524,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(44524,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(44524,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(44524,36,'cp -31','e7e6 b1c3 b8c6 c1f4 a7a6 g2g3 d7d5 c4d5 e6d5 f1g2 c8e6 g1f3 d5d4 c3e2 h7h6 e1g1 g7g5 f4d2 f8e7 a2a3 f6d7 b2b4 a8c8 b4c5 d7c5 d2b4 e8g8 h2h4 b7b6 h4g5 h6g5 f3h2 c6e5 b4c5 c8c5 f2f4 e5g4 h2g4 e6g4');
INSERT INTO evaluations VALUES(45600,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45600,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45600,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45600,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(45600,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(45600,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45600,36,'cp 0','c1f4 f6e4 d1c2 e4c3 c2c3 b8d7 g1f3 a7a5 a2a3 a5a4 a1c1 f8e7 e2e3 f7f6 f1e2 e8g8 e1g1 e6e5 f4g3 e5e4 f3d2 f6f5 c1b1 e7h4 g3d6 h4e7');
INSERT INTO evaluations VALUES(45597,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(45597,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(45597,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45597,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45597,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45597,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45597,36,'cp 0','f6e4 c3e4 d5e4 e2e3 b8d7 g1e2 e6e5 d4e5 d7c5 d1d8 e8d8 e2d4 d8e8 h2h4 c5e6 f1c4 e6f4 e3f4 f8c5 e1c1 e8e7 h1e1 c8g4 c4e2 g4c8 e2f1 c8g4 f1e2');
INSERT INTO evaluations VALUES(46458,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(46458,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(46458,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(46458,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(46458,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(46458,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(46458,36,'cp -12','d2d4 b8c6 b1c3 h7h6 e1g1 f8e7 c1e3 c5c4 f3e5 e7b4 c3a4 e8g8 e5c6 b7c6 b2b3 c4b3 a2b3 a7a5 e2d3 f6e4 d1f3 c8a6 e3f4 f8e8 d3a6 a8a6 f3d3 a6a7 c2c3 b4d6 f4d6 e4d6 a1e1 a7e7 e1e7 d8e7 a4c5 e7e2 d3e2 e8e2');
INSERT INTO evaluations VALUES(46437,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(46437,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(46437,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(46437,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(46437,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(46437,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(46437,36,'cp -12','b8c6 e1g1 c5d4 f3d4 f8d6 b1c3 e8g8 e2f3 d6e5 c1e3 h7h6 a1c1 a7a6 h2h3 e5c7 d4c6 b7c6 c3e2 f6d7 e2f4 d7e5 f3e2 c8f5 c2c4 e5c4 e2c4 c7f4 e3f4 d5c4 f1e1 d8d1 e1d1 f5e6 d1d6 f8b8 d6e6 f7e6 f4b8 a8b8 c1c4 b8b2 c4c6 b2a2');
INSERT INTO evaluations VALUES(13756,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(13756,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(13756,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(13756,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(13756,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(13756,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(13756,36,'cp -13','b1c3 c5d4 f3d4 f8d6 e1g1 e8g8 e2f3 a7a6 f1e1 h7h6 h2h3 d6b8 d4c6 b7c6 c3e2 b8a7 e2g3 d8b6 d1d2 c8e6 b2b3 a8d8 c1b2 d5d4 a1d1 c6c5 c2c3 c5c4 c3d4 c4b3 a2b3 e6b3 d1b1 b3a2 b1a1 a2d5 b2a3 b6d4 d2d4 a7d4 a3f8 d4a1 f3d5 d8d5 e1a1');
INSERT INTO evaluations VALUES(48862,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48862,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(48862,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(48862,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(48862,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(48862,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(48862,36,'cp -40','d2d3 d5d4 c3e4 f6e4 g2e4 f8d6 e4g2 c7c5 e2e3 b8c6 g1f3 c8g4 h2h3 g4f5 e1g1 e8g8 a2a3 f7f6 e3d4 c5d4 f3d2 d8d7 d2e4 a8e8 e4d6 d7d6 f2f4 d6d7 g1h2 f5e6 d1f3 a7a5 c1d2 e5f4 f3f4 g7g5');
INSERT INTO evaluations VALUES(48863,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(48863,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(48863,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(48863,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(48863,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(48863,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(48863,36,'cp -54','d5d4 c3e4 f6e4 g2e4 f8d6 e4g2 c7c5 g1f3 b8c6 e2e3 c8g4 h2h3 g4f5 e1g1 e8g8 f1e1 d8c8 g3g4 f5d7 f3g5 f7f5 g4f5 d7f5 e3d4 c5d4 d1f3 c8c7 c1d2 a8e8 a2a3 h7h6 g5e4 f5e4 f3e4 c7f7 e4d5 e8e6');
INSERT INTO evaluations VALUES(906,30,'cp 878',NULL);
INSERT INTO evaluations VALUES(906,31,'cp 878',NULL);
INSERT INTO evaluations VALUES(906,32,'cp 886',NULL);
INSERT INTO evaluations VALUES(906,33,'cp 898',NULL);
INSERT INTO evaluations VALUES(906,34,'cp 890',NULL);
INSERT INTO evaluations VALUES(906,35,'cp 894',NULL);
INSERT INTO evaluations VALUES(906,36,'cp 898','d5c6 c8d7 c6b7 a8b8 b7d5 e7f6 f3e3 f6h6 e3e2 d7c6 e2e5 e8d7 e1g1 f8d6 d5c6 d7c6 e5e4 c6d7 e4a4 d7c8 g2g3 h6e6 b1c3 h7h5 c3b5 b8b6 b5a7 c8b7 a7b5 h8a8 b5d6 b6d6 a4b3 e6b3 c2b3 g7g5 g1g2 d6b6 b3b4 h5h4 g3g4 h4h3 g2h3 b6b4');
INSERT INTO evaluations VALUES(905,30,'cp 888',NULL);
INSERT INTO evaluations VALUES(905,31,'cp 894',NULL);
INSERT INTO evaluations VALUES(905,32,'cp 880',NULL);
INSERT INTO evaluations VALUES(905,33,'cp 873',NULL);
INSERT INTO evaluations VALUES(905,34,'cp 888',NULL);
INSERT INTO evaluations VALUES(905,35,'cp 871',NULL);
INSERT INTO evaluations VALUES(905,36,'cp 882','c8d7 c6b7 a8b8 b7d5 e7f6 f3e3 b8d8 b1c3 d7c6 c3e4 f6g6 d5c6 g6c6 e1g1 f8e7 d2d3 h8f8 e3h3 e8f7 h3h7 c6g6 h7h3 f8h8 h3f3 f7g8 f3g3 g6g3 f2g3 e7d6 c1e3 d8b8 b2b3 a7a6 h2h3 b8f8 g3g4 f8f7 f1f7 g8f7 a2a4 h8f8 g1f2 a6a5 f2e2 f8h8 e3d2 d6b4 g4g5 f7g6 d2b4');
INSERT INTO evaluations VALUES(7529,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(7529,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(7529,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(7529,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(7529,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(7529,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(7529,36,'cp 43','d2d4 f6d7 d4e5 d6e5 d1d3 e7d6 f1d1 e8g8 c1g5 d8e8 f3h4 d7c5 d3g3 g8h8 g5e3 g7g6 h4f3 f7f6 e3c5 d6c5 c3a4 c5b6 f3d2 e8e7 a4b6 a7b6 g3c3 c6c5 a2a4 f8d8 b2b3 h8g7 h2h3 c8b7 f2f3 b7a6 d2f1 a6f1 d1d8 e7d8 g1f1 g7h8');
INSERT INTO evaluations VALUES(7527,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(7527,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(7527,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(7527,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(7527,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7527,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(7527,36,'cp 25','f6d7 c1e3 a7a5 b2b3 e8g8 h2h3 f8e8 f1e1 h7h6 d4e5 d6e5 c3a4 c8a6 f3d2 a6b5 a4b2 e7b4 c2c4 b5a6 d1c2 d8e7 b2d3 d7f8 e1d1 f7f6 d2f1 f8e6 f1g3 e7f7 d3b4 a5b4 a2a3 c6c5 a3b4 c5b4 g3f5 a6b7 a1a8');
INSERT INTO evaluations VALUES(8299,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(8299,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(8299,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(8299,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(8299,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(8299,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(8299,36,'cp -10','f1e1 e8g8 a2a3 b7b6 h2h3 f6d7 b1c3 c8b7 a1b1 a7a6 c3e2 a6a5 a3a4 b7a6 c1e3 b6b5 a4b5 a6b5 e2c3 b5a6 d3d4 c5d4 f3d4 g7d4 e3d4 c6d4 d1d4 d8b6 d4b6 d7b6 b1a1 b6c4 e1b1 a6b7');
INSERT INTO evaluations VALUES(8300,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(8300,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(8300,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8300,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(8300,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(8300,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(8300,36,'cp -5','e8g8 e4e5 f6e8 e5d6 e8d6 c2c3 e7e5 b1a3 d8d7 a3c4 d6c4 d3c4 d7d1 e1d1 e5e4 f3d2 f7f5 f2f3 g6g5 f3e4 f5f4 g3f4 g5f4 d2f3 c8e6 g1f2 c6e5 b2b3 e5g4 f2e2 g7c3 a1b1 a8e8 h2h3 g4e5 c1b2 c3b2 b1b2 e5c6 h3h4 c6b4 d1d6 g8h8 d6d1');
INSERT INTO evaluations VALUES(3210,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3210,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(3210,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(3210,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(3210,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3210,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3210,36,'cp 0','e4e5 f6e8 e5d6 e8d6 c2c3 a8b8 c1f4 f8e8 f3d2 c8d7 b1a3 e7e5 f4e3 b7b6 d2e4 d8c7 a1c1 d7e6 e4g5 e6d7');
INSERT INTO evaluations VALUES(9571,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(9571,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(9571,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(9571,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(9571,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(9571,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(9571,36,'cp 27','f3d2 c7c6 a2a4 e8g8 a4a5 c6d5 e4d5 a7a6 f1e2 b7b6 a5b6 d7b6 d2b3 c8d7 e1g1 d7b5 e2d3 b6c4 d1e2 d8c7 h2h3 f8b8 f1d1 h7h6 c3b5 a6b5 a1a8 b8a8 b3d2 c4b6 d2e4 b6d5 d3b5 a8a1 e4g3 e7f8 c1d2 d5f4 d2f4 a1d1 e2d1 e5f4');
INSERT INTO evaluations VALUES(11751,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(11751,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(11751,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(11751,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(11751,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(11751,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(11751,36,'cp 119','d2d4 e5d4 f3d4 c6a5 b1c3 c8b7 d1h5 e7g6 d4f5 a5b3 a2b3 g6e7 f5g3 d7d6 c3d5 c7c6 d5f4 g7g6 h5f3 c6c5 f3c3 h8h7 f1d1 f8g7 c3e1 e7c6 c2c3 c6a5 a1a3 c5c4 b3c4 a5c4 a3a1 e8f8 f4d5 b7d5');
INSERT INTO evaluations VALUES(11750,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(11750,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(11750,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(11750,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(11750,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(11750,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(11750,36,'cp 112','e5d4 f3d4 c6a5 b1c3 c8b7 d1h5 a5b3 a2b3 e7g6 d4f5 g6e7 f5g3 d7d6 c3d5 c7c6 d5f4 g7g6 h5f3 c6c5 f3c3 h8h7 f1e1 f8g7 c3d3 g7e5 h2h4 d8d7 f4d5 e7d5 e4d5 e8c8 c2c4 f7f5 b3b4 b5c4 d3c4 d7b5 c4b5 a6b5 b4c5 b7d5 h4h5 g6h5 g3f5 h7f7');
INSERT INTO evaluations VALUES(12308,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(12308,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(12308,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(12308,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(12308,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(12308,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(12308,36,'cp 149','c3c4 d5e7 d4d5 c6d4 f3d4 c5d4 f1e2 e6d5 e1g1 d7d6 e5d6 d8d6 b1a3 a7a6 f1e1 c8e6 c4d5 d6d5 e2c4 d5c6 d1b3 a8d8 c4e6 f7e6 e1e6 c6d5 b3d5 d8d5 a3c4 e8d7 e6b6 d7c8 c1f4 e7c6 b6a6 b7a6 c4b6 c8b7 b6d5 c6b4 d5b4 f8b4 g1f1 h8d8 a2a3 b4a5 a1d1 d4d3 f4e5 d8d7 b2b4 a5d8 f1e1 d8b6');
INSERT INTO evaluations VALUES(12307,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(12307,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(12307,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(12307,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(12307,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(12307,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(12307,36,'cp 152','d5b4 d4c5 a7a5 b1c3 f8c5 a2a3 b4a6 c1g5 c5e7 h2h4 a6c5 c3b5 e8g8 f1d3 b7b6 d3c2 f7f6 e5f6 e7f6 d1e2 c8b7 b5d6 d8b8 e1c1 c6d4 d1d4 f6d4 d6b7 f8f3 e2f3 b8b7 f3b7 c5b7 h1d1 d4f2 d1d7 b7c5 d7d6 h7h6 g5f4 a5a4 f4e5 c5b3 c1b1 a8f8 h4h5');
INSERT INTO evaluations VALUES(23978,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(23978,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(23978,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(23978,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(23978,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(23978,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(23978,36,'cp 77','f2f4 e8g8 d1f3 d8c8 h2h3 a7a5 c1e3 c8b7 b2b3 d6d5 e4e5 e7b4 c3e2 f6e4 e2g3 f7f5 g3e4 f5e4 f3g4 b4c3 a1d1 b7c8 g4g3 f8f5 e3c5 c3b4 g3e3 c8e6 c2c4 a8d8 c5d4 g7g5 f4g5 f5f1 d1f1 d5c4 g5g6 e6d5 d4a7 d8f8 g6h7 g8h7 f1f4 f8f4 e3f4 c4b3 a2b3 h7g8 a7f2 d5b3 f4g4 g8f8');
INSERT INTO evaluations VALUES(26139,30,'cp -206',NULL);
INSERT INTO evaluations VALUES(26139,31,'cp -222',NULL);
INSERT INTO evaluations VALUES(26139,32,'cp -206',NULL);
INSERT INTO evaluations VALUES(26139,33,'cp -221',NULL);
INSERT INTO evaluations VALUES(26139,34,'cp -234',NULL);
INSERT INTO evaluations VALUES(26139,35,'cp -221',NULL);
INSERT INTO evaluations VALUES(26139,36,'cp -221','c3c4 c5d4 d1d4 g8f6 c4d5 f8c5 d4d3 f6d5 b1c3 d5c3 d3c3 e8g8 e1g1 d8b6 a2a3 d7f6 c3e5 f8d8 c1g5 d8d5 e5e2 b6b3 a1c1 a8d8 c1c3 b3a4 g5f6 g7f6 g2g3 a4g4 c3c4 g4g6 f1c1 c5b6 c4c8 d8c8 c1c8 g8g7 g1g2 h7h6');
INSERT INTO evaluations VALUES(26761,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(26761,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(26761,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(26761,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26761,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(26761,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(26761,36,'cp 22','h2h4 d8d7 b1c3 h7h5 e1f1 g8e7 h1h3 e7f5 g1f3 c6b5 c3b5 d7b5 f1g1 c7c5 c2c3 a8c8 c1g5 a7a5 a1c1 f8e7 c1b1 b7b6 d1d2 b5a4 b2b3 a4c6 g5e7 f5e7 d4c5 b6c5 c3c4 a5a4 b1d1 c8d8 d2g5 e7f5 c4d5 d8d5 d1d5 c6d5');
INSERT INTO evaluations VALUES(27062,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(27062,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(27062,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(27062,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(27062,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(27062,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(27062,36,'cp 92','c4e6 f7e6 d1b3 d8c8 f3g5 c6d8 d2d4 g8f6 e1g1 f8e7 d4e5 d6e5 g5f3 e7d6 c1e3 a7a5 b1d2 e8g8 a2a4 h7h6 h2h3 b7b6 f1e1 d8f7 a1d1 f8e8 d2c4 c8b7 b3b5 b7e4 e3d4 e4d5 f3e5 d5b5 a4b5 a8d8 e5c6 d8d7 d4f6 g7f6');
INSERT INTO evaluations VALUES(27046,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(27046,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(27046,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(27046,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(27046,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(27046,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(27046,36,'cp 72','f7e6 d1b3 d8c8 f3g5 c6d8 d2d4 g8f6 e1g1 f8e7 d4e5 d6e5 g5f3 e7d6 f1e1 h7h6 c1e3 e8g8 b1d2 a7a5 h2h3 b7b6 a2a4 d8f7 d2c4 c8e8 f3d2 d6e7 e1d1 f6d7 b3b5 e7c5 b5c6 c5e3 c4e3 d7c5 c6c7 c5a4 d2c4 a8b8 c7d7 b6b5 d7e8');
INSERT INTO evaluations VALUES(27128,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(27128,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(27128,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(27128,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(27128,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(27128,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(27128,36,'cp 80','d1b3 d8c8 f3g5 c6d8 d2d4 g8f6 d4e5 d6e5 c1e3 f8d6 f2f3 e8g8 b1d2 h7h6 g5h3 f6d7 h3f2 a7a5 e1g1 d6c5 e3c5 d7c5 b3b5 c5d7 a2a4 d8f7 f2d3 f8d8 f1d1 c8b8 d2c4 f7d6 b5b3 b8a7 g1f1 d7c5 d3c5 a7c5 c4d6 d8d6');
INSERT INTO evaluations VALUES(31106,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(31106,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(31106,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(31106,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(31106,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(31106,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(31106,36,'cp 90','g4f3 g2f3 g8f6 e2e4 d5e4 f3e4 e7e5 d4d5 b8d7 d1c2 f8c5 h1g1 e8g8 c1g5 c6d5 e4d5 c5d4 f1h3 d8a5 e1c1 d4c3 c2c3 a5a2 g5h6 g7g6 h3d7 f6d7 h6f8 a8f8 c1d2 d7b6 d1a1 a2c4 c3c4 b6c4 d2d3 c4d6 g1c1 f8a8 c1c7 a7a5 c7d7 a8a6 d3e3 a5a4');
INSERT INTO evaluations VALUES(31111,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(31111,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(31111,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(31111,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(31111,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(31111,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(31111,36,'cp 98','g2f3 g8f6 e2e4 d5e4 f3e4 e7e5 d4d5 f8c5 h1g1 e8g8 c1h6 f6e8 d1f3 c6d5 c3d5 g8h8 h6d2 b8c6 e1c1 c6d4 f3g4 d4e6 g4g3 e8d6 d2e3 d6e4 g3e5 d8h4 d5c7 c5e3 f2e3 h4h6 e5e4 e6c7 e4f4 h6b6 f4e5 c7e8 e5g3 b6f6 c1b1 a8d8 d1d8 f6d8');
INSERT INTO evaluations VALUES(33788,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(33788,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(33788,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(33788,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(33788,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(33788,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(33788,36,'cp 162','f1e2 b8c6 c1g5 c6d4 d1d4 e7e5 d4e3 c7c6 e1c1 f8e7 f2f4 e5c5 e3c5 d6c5 f4f5 f6d7 g5f4 e7f6 d1d2 e8g8 h1d1 f8e8 d2d7 c8d7 e4e5 f6e5 f4e5 d7f5 e5c7 a8c8 c7d6 b7b6 d1f1 c8d8 d6c5 f5c2 c5b6 a7b6 c1c2 e8e6 a2a3');
INSERT INTO evaluations VALUES(34925,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(34925,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(34925,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(34925,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(34925,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(34925,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(34925,36,'cp -47','d5d4 d1d4 e5d4 c3b5 f8b4 c1d2 b4d2 e1d2 b8a6 b5d4 c7c5 d4b5 c8e6 a2a3 g8f6 g1f3 e8e7 e2e3 c5c4 d2e1 a6c5 f3d4 h8c8 d4e6 c5e6 f1e2 a7a6 b5c3 b7b5 f2f3 e6c5 e1f2 a8b8 e3e4 f6d7 h1d1 c5b3 a1b1 g7g6 f3f4 d7c5 f2e3 c8d8 d1d8 b8d8');
INSERT INTO evaluations VALUES(34957,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(34957,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(34957,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(34957,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(34957,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(34957,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(34957,36,'cp -39','d1d4 e5d4 c3b5 b8a6 b5d4 f8b4 c1d2 b4d2 e1d2 c7c5 d4b5 g8f6 e2e3 c8e6 a2a3 e8e7 g1f3 c5c4 f3d4 a6c5 d2e1 h8c8 d4e6 c5e6 f1e2 a7a6 b5c3 b7b5 e1d2 e6c5 d2c2 c8d8 a1d1 h7h5 h2h4 a8b8 d1d8 b8d8 h1d1 d8d1 c2d1 e7d6 d1d2 c5e4 c3e4 f6e4 d2e1');
INSERT INTO evaluations VALUES(34956,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(34956,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(34956,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(34956,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(34956,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(34956,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(34956,36,'cp -45','e5d4 c3b5 b8a6 b5d4 f8b4 c1d2 b4d2 e1d2 c7c5 d4b5 c8e6 a2a3 g8f6 e2e3 e8e7 g1f3 c5c4 f3d4 a6c5 d2e1 h8c8 d4e6 c5e6 f1e2 a7a6 b5c3 b7b5 e1d2 a8b8 h1d1 e6c5 d2e1 h7h6 a1b1 c8d8 f2f3 d8d7 d1d7 f6d7 b1d1 d7b6 h2h4 b6a4 c3a4 c5a4 d1d2');
INSERT INTO evaluations VALUES(34975,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(34975,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(34975,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(34975,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(34975,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(34975,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(34975,36,'cp -55','c3b5 f8b4 c1d2 b4d2 e1d2 b8a6 b5d4 c7c5 d4b5 c8e6 a2a3 c5c4 e2e3 a6c5 d2e1 e8e7 g1f3 g8f6 f3d4 h8c8 d4e6 c5e6 g2g3 a8b8 f1h3 c8c6 a1b1 a7a6 b5c3 b7b5 e1e2 e6c5 h1d1 a6a5 h3g2 c6c8 g3g4 b5b4 g4g5 b4c3 g5f6 g7f6 b2c3 b8b3');
INSERT INTO evaluations VALUES(34885,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(34885,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(34885,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(34885,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(34885,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(34885,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(34885,36,'cp -44','b8a6 b5d4 f8b4 c1d2 b4d2 e1d2 c7c5 d4b5 c8e6 a2a3 g8f6 e2e3 e8e7 g1f3 c5c4 d2e1 a6c5 f3d4 h8c8 d4e6 c5e6 f1e2 a7a6 b5c3 b7b5 e1d2 a8b8 h1d1 e6c5 d2e1 h7h6 a1b1 c8d8 f2f3 d8d7 d1d7 f6d7 b1d1 d7b6 e1d2 b8d8 d2c2 d8d1 e2d1 e7d6 c2d2');
INSERT INTO evaluations VALUES(37564,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(37564,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(37564,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(37564,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(37564,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(37564,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(37564,36,'cp 108','g1f3 c8b7 b1c3 e7e6 f1d3 h7h6 e1g1 g8e7 c3e4 d7d6 e5d6 c7d6 c1f4 e7f5 g2g4 d6d5 e4c3 f5d6 c3b5 d6b5 d3b5 b8c6 c2c3 e8g8 f1e1 c6a5 b5d3 a8c8 d1e2 a5c4 h2h4 f8e8 f4g3 c4d6 a1d1 d6c4 a2a4 d8d7 f3e5 g7e5 g3e5 c4e5 e2e5');
INSERT INTO evaluations VALUES(37563,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(37563,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(37563,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(37563,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(37563,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(37563,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(37563,36,'cp 90','e7e6 h2h4 h7h6 b1c3 c8b7 c1f4 g8e7 d1d2 d7d6 e1c1 a7a6 f1d3 b7f3 e5d6 c7d6 g2f3 b8c6 d4d5 e6d5 h1e1 e8f8 h4h5 g6g5 f4h2 c6d4 e1e7 d8e7 c3d5 e7e6 d3e4 a8c8 d2b4 d4e2 c1d2 e2f4 h2f4 g5f4 d2c1 g7e5 d5b6 c8d8 b6d5 d8c8 b4b6');
INSERT INTO evaluations VALUES(37553,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(37553,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(37553,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(37553,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(37553,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(37553,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(37553,36,'cp 99','h2h4 h7h6 b1c3 c8b7 c1f4 g8e7 d1d2 d7d6 e1c1 a7a6 f1d3 b7f3 e5d6 c7d6 g2f3 b8c6 d4d5 e6d5 h1e1 e8f8 h4h5 g6g5 f4h2 d5d4 c3e2 b6b5 c1b1 g7f6 f3f4 d8d7 f4g5 h6g5 f2f4 g5g4 e2g3 d6d5 f4f5 c6e5 h5h6 e5f3 d2f4 e7g8 g3h5 h8h6 h5f6 h6f6 f4g4');
INSERT INTO evaluations VALUES(39470,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(39470,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(39470,32,'cp 285',NULL);
INSERT INTO evaluations VALUES(39470,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(39470,34,'cp 294',NULL);
INSERT INTO evaluations VALUES(39470,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(39470,36,'cp 261','c2c3 b4e7 f1c4 d7d6 c1e3 e8f8 b1d2 b8c6 d1e2 d8e8 h2h3 h7h5 e1c1 g7g5 d2f1 a7a5 f1g3 a5a4 c1b1 g8h6 h3h4 g5g4 f3e1 a4a3 b2b3 e8g6 c4d3 c8d7 e1c2 f6f5 e3h6 g6h6 d4d5 f5e4 g3e4 c6a5 f2f3 g4f3 e2f3 f8e8 c2a3');
INSERT INTO evaluations VALUES(39469,30,'cp 269',NULL);
INSERT INTO evaluations VALUES(39469,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(39469,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(39469,33,'cp 253',NULL);
INSERT INTO evaluations VALUES(39469,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(39469,35,'cp 282',NULL);
INSERT INTO evaluations VALUES(39469,36,'cp 275','b4e7 f1c4 d7d6 c1e3 e8f8 d1e2 d8e8 b1d2 h7h5 e1c1 a7a5 h2h3 g7g5 h3h4 g5g4 f3e1 b8c6 e1c2 a5a4 a2a3 f8g7 c1b1 h8h7 d2f1 g7h8 f1g3 c6a5 c4a2 a5b3 f2f4 c7c6 d4e5 f6e5 f4e5 d6e5 c2b4 c8e6 b4d3 e7f6 d3c5 b3c5 e3c5 b7b5 a2e6 e8e6 g3f5');
INSERT INTO evaluations VALUES(43414,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(43414,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(43414,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(43414,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(43414,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(43414,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(43414,36,'cp 55','c2c4 e7e6 f1e2 b8c6 b1c3 f8e7 e1g1 e8g8 g1h1 d6d5 c4d5 e6d5 d4c6 b7c6 e4e5 f6d7 f3f4 d7c5 c1e3 a8b8 a1c1 c8f5 b2b3 c5e6 e2g4 f5g4 d1g4 d5d4 f1d1 c6c5 f4f5 e7g5 e3g5 d8g5 g4g5 e6g5 h2h4 d4c3');
INSERT INTO evaluations VALUES(43403,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(43403,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(43403,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(43403,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(43403,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(43403,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(43403,36,'cp 35','b8c6 b1c3 e7e6 f1e2 f8e7 e1g1 d6d5 c4d5 e6d5 d4c6 b7c6 e4e5 f6d7 f3f4 e8g8 c1e3 e7c5 e3f2 c5f2 f1f2 d7c5 d1d4 d8e7 a1c1 f8d8 c3a4 c5a4 d4a4 c8f5 a4d4 d8c8 c1c5 e7a7 h2h3 a6a5 e2g4 f5g4 h3g4 h7h6 f2c2 c8e8 d4d2 a7d7 c5c6 d7g4 c2c3 a8d8');
INSERT INTO evaluations VALUES(44466,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44466,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44466,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(44466,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(44466,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44466,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44466,36,'cp 0','d7d5 d2d4 e7e6 g1f3 d5c4 f3e5 b8d7 e5c4 b7b5 c4e5 c8b7 f1g2 b5b4 e5d7 d8d7 c3e4 f6e4 g2e4 c6c5 e4b7 d7b7 e1g1 c5d4 d1d4 a8d8 d4e5 d8d5 e5e4 f8e7 c1e3 e8g8 a1d1 e7f6 b2b3 f8d8 d1d5 d8d5 e4c2 h7h6 f1c1 f6c3 c1d1 a7a5 d1d5 b7d5 f2f3 c3d4 e3d4 d5d4 g1f1');
INSERT INTO evaluations VALUES(44890,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44890,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44890,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44890,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44890,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44890,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44890,36,'cp 8','d2d4 d5c4 g1f3 e7e6 f3e5 b8d7 e5c4 b7b5 c4e5 d7e5 d4e5 d8d1 e1d1 f6d5 f1g2 f8c5 f2f4 e8g8 c1d2 a7a5 c3d5 c6d5 a1c1 c5d4 b2b4 a5b4 d2b4 f8d8 a2a3 c8d7 c1c7 a8c8 c7c8 d8c8 d1d2 f7f6 e5f6 d4f6 g2f3 g8f7 h1c1 c8c1 d2c1');
INSERT INTO evaluations VALUES(44903,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44903,36,'cp 0','d5c4 g1f3 e7e6 f3e5 b8d7 e5c4 b7b5 c4e5 c8b7 f1g2 b5b4 e5d7 d8d7 c3e4 f6e4 g2e4 c6c5 e4b7 d7b7 e1g1 c5c4 d1a4 b7d7 a4a6 a8c8 c1f4 f8e7 a1c1 c8c6 a6a4 c6c8');
INSERT INTO evaluations VALUES(45485,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(45485,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(45485,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(45485,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(45485,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(45485,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(45485,36,'cp -8','g1f3 b8c6 d2d4 c8f5 b1c3 e7e6 f3h4 f5e4 f2f3 e4g6 e1g1 d8b6 e2e3 f8d6 c3a4 b6d8 h4g6 h7g6 f1f2 b7b5 a4c5 d6c5 d4c5 e8g8 a2a3 a7a5 g2f1 d8b8 d1e2 b5b4 e2b5 f8c8 b5b8 c8b8 f2c2 b4a3 b2a3');
INSERT INTO evaluations VALUES(45484,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(45484,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(45484,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(45484,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(45484,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(45484,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(45484,36,'cp -26','b8c6 d2d4 c8f5 b1c3 e7e6 e1g1 h7h6 c1f4 f8d6 f3e5 a8c8 a1c1 e8g8 d1d2 d6e7 h2h3 d8b6 c3a4 b6b5 e5c6 c8c6 c1c6 b7c6 a4c3 b5a6 f1c1 f8c8 g2f1 a6b7 c3a4 f6d7 d2d1 f5h7 d1b3 b7b3 a2b3 c6c5 d4c5 d7c5');
INSERT INTO evaluations VALUES(13179,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(13179,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(13179,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(13179,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(13179,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(13179,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(13179,36,'cp -11','d2d4 c8f5 d1b3 d8b6 b3b6 a7b6 e1g1 e7e6 c1e3 f5b1 f1b1 f8d6 e3d2 f6e4 e2e3 g7g5 d2c3 g5g4 f3d2 e4d2 c3d2 f7f5 g2f1 h7h5 b1c1 h5h4 f1b5 a8c8 a2a3 e8f7 c1c3 c6e7 c3c8 e7c8 a1c1 h4g3');
INSERT INTO evaluations VALUES(13172,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(13172,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(13172,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(13172,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(13172,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(13172,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(13172,36,'cp -14','c8f5 b1c3 e7e6 f3h4 f5e4 f2f3 e4g6 e1g1 d8b6 h4g6 h7g6 e2e3 f8d6 c3a4 b6d8 c1d2 e8g8 d2e1 e6e5 e1f2 a8c8 a1c1 f8e8 d1b3 c6a5 b3b5 e5d4 e3d4 c8c6 b5d3 d8c7 c1c6 c7c6 a4c3 d6b4 d3b5 b4c3 b5c6 a5c6 b2c3');
INSERT INTO evaluations VALUES(54974,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54974,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54974,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(54974,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(54974,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(54974,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(54974,36,'cp 2','g8f6 d1c2 e7e5 g1f3 d6d5 d2d4 f6e4 f3e5 f8d6 f1d3 e8g8 b1d2 e4f6 e1g1 c7c5 f1e1 d8c7 d2f3 c5c4 d3f1 b7b5 b2b3 b8c6 e5c6 c7c6 a2a4 c4b3 a4b5 c6b5 f1b5 b3c2 b5d3 c8e6 d3c2 f8c8 c1a3 d6f8 a3f8 g8f8 f3g5 e6g4 c2h7');
INSERT INTO evaluations VALUES(47037,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(47037,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(47037,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(47037,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47037,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(47037,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(47037,36,'cp 18','d1c2 e7e5 d2d4 d6d5 g1f3 f6e4 f3e5 f8d6 f1d3 e8g8 b1d2 e4f6 e1g1 c7c5 f1e1 c5c4 d3f1 d8c7 d2f3 b7b5 b2b3 b8c6 e5c6 c7c6 f3e5 c6c7 a2a4 c4b3 c2b3 b5a4 a1a4 a8b8 b3c2 f6e4 f1d3 c7c3 c2c3 e4c3 a4a7 c3b5 d3b5 b8b5 c1a3 d6a3 a7a3');
INSERT INTO evaluations VALUES(56819,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(56819,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(56819,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(56819,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(56819,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(56819,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(56819,36,'cp 217','d4e5 c8e6 b1d2 b8d7 e5f6 g8f6 f1e2 d7c5 g1f3 f8d6 f4d6 d8d6 f3d4 e8g8 e1g1 g8h8 c2c4 e6g8 a1c1 a8d8 b2b4 c5e6 c4c5 d6e7 d2f3 c7c6 e2d3 f6e4 d4e6 e7e6 f3d4 e6h6 b4b5 f8f2 f1f2 e4f2 g1f2');
INSERT INTO evaluations VALUES(56813,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(56813,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(56813,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(56813,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(56813,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(56813,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(56813,36,'cp 200','c8e6 b1d2 b8d7 e5f6 g8f6 f1e2 d7c5 g1f3 f8d6 e1g1 d6f4 e3f4 d8d6 b2b4 c5e4 d2e4 f6e4 g2g3 e4c3 d1d3 c3e2 d3e2 e8g8 f1e1 e6g4 e2d3 g4f3 d3f3 c7c6 a2a3 a7a5 c2c3 a5b4 a3b4 a8a1 e1a1 d6e7 h2h4 f8e8 h4h5 d5d4 a1d1 d4c3 f3c3 e8d8 c3b3 g8f8');
INSERT INTO evaluations VALUES(1825,30,'cp -166',NULL);
INSERT INTO evaluations VALUES(1825,31,'cp -166',NULL);
INSERT INTO evaluations VALUES(1825,32,'cp -158',NULL);
INSERT INTO evaluations VALUES(1825,33,'cp -177',NULL);
INSERT INTO evaluations VALUES(1825,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(1825,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(1825,36,'cp -175','f1e2 d7c6 e2d3 c8e6 b1c3 g8f6 c1f4 e8c8 e1c1 h7h6 a2a3 f8d6 f4d2 h8e8 h2h3 c6c5 h1e1 c5c4 d3f1 a7a6 g2g4 c7c6 d2e3 b7b5 f1g2 c8c7 f2f4 a6a5 f4f5 e6c8 e3f2 c8b7 g2f3 d6f4 c1b1 d8d1 c3d1 e8e5 e1e5 f4e5');
INSERT INTO evaluations VALUES(1824,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(1824,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(1824,32,'cp -173',NULL);
INSERT INTO evaluations VALUES(1824,33,'cp -179',NULL);
INSERT INTO evaluations VALUES(1824,34,'cp -168',NULL);
INSERT INTO evaluations VALUES(1824,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(1824,36,'cp -164','d7c6 e2d3 c8e6 c1f4 e8c8 b1c3 g8f6 e1c1 h7h6 f2f3 f8e7 h1e1 h8e8 f4e5 g7g6 e5f4 g6g5 f4e5 f6d7 e5d4 d7c5 d3e2 a7a6 a2a3 h6h5 e2f1 h5h4 g2g3 a6a5 c1b1 c5d7 c3e4 g5g4 f3g4 e6g4');
INSERT INTO evaluations VALUES(1754,30,'cp -156',NULL);
INSERT INTO evaluations VALUES(1754,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(1754,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(1754,33,'cp -170',NULL);
INSERT INTO evaluations VALUES(1754,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(1754,35,'cp -166',NULL);
INSERT INTO evaluations VALUES(1754,36,'cp -163','e2d3 c8e6 c1f4 e8c8 b1c3 g8f6 e1c1 h7h6 h1e1 f8d6 f4d2 a7a6 h2h3 b7b5 a2a3 c8b7 c3e4 f6e4 d3e4 h8e8 d2c3 d6f4 c1b1 d8d1 e1d1 g7g6 e4f3 f4d6 d1e1 e8d8 c3g7 h6h5 g7f6 d8a8 g2g4 e6d5 f3d5 c6d5 g4h5 g6h5 f6g7 d6f4');
INSERT INTO evaluations VALUES(7564,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(7564,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(7564,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(7564,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(7564,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(7564,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(7564,36,'cp 13','e1g1 e8g8 h2h3 c7c6 d1f3 f8e8 c1f4 b4d6 f1e1 c8e6 c3e2 d8b8 b2b3 a7a5 a1d1 a5a4 c2c4 d6f4 e2f4 a4b3 a2b3 d5c4 b3c4 e6d7 e1e8 d7e8 f3e3 b8b2 e3e7 h7h6 f4e2 b2a3 e7a3 a8a3 c4c5');
INSERT INTO evaluations VALUES(7563,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(7563,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(7563,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(7563,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(7563,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(7563,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(7563,36,'cp 5','e8g8 h2h3 c7c6 d1f3 f8e8 c1f4 a7a5 c3a4 c8e6 a1d1 b4d6 f1e1 d8b8 f4d6 b8d6 b2b3 a8b8 d3f1 d6b4 e1e2 h7h6 c2c3 b4a3 f3e3 e6f5 e3c5 a3c5 e2e8 f6e8 a4c5 g8f8 b3b4 a5b4 c5a6 b8a8 c3b4 f8e7 b4b5 c6b5 f1b5 f5e6 a2a4 e7d6 a4a5');
INSERT INTO evaluations VALUES(2859,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(2859,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(2859,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2859,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(2859,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(2859,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(2859,36,'cp 0','h2h3 c7c6 d1f3 f8e8 c1g5 h7h6 g5h4 a7a5 f1e1 c8e6 e1e2 a5a4 a2a3 b4d6 a1e1 g7g5 h4g3 g8g7 e2e5 d8b8 f3e3 g7f8 e3f3');
INSERT INTO evaluations VALUES(9955,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(9955,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(9955,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(9955,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(9955,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(9955,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(9955,36,'cp 46','c6d7 c8d7 d2c3 d7c6 d1e2 c6e4 c1g5 e8g8 g5f6 e4f3 f6d8 f3e2 d8c7 e2h5 f2f3 f7f6 c7d6 f8c8 e1c1 a7a5 h1e1 c8c4 e1e3 a8c8 d1d5 b7b6 g2g3 c8d8 d5d2 c4c6 d6a3 c6c8 d2d8 c8d8 b2b3 b6b5 a3c5 g8f7 c3c4 b5c4');
INSERT INTO evaluations VALUES(12865,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12865,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12865,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12865,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(12865,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12865,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(12865,36,'cp 0','g1f3 e7e6 e1g1 d7d6 d1e2 a7a6 f1d1 d8c7 c1f4 f8e7 a1c1 e8g8 h2h3 f6d7 c4b3 c7b8 f4g3 h7h6 g1h2 a8a7 h2g1 a7a8');
INSERT INTO evaluations VALUES(13065,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(13065,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(13065,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(13065,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(13065,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(13065,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(13065,36,'cp 89','d8c7 b5c6 c7c6 d1f3 c8d7 f3c6 d7c6 f2f3 e7e5 g1e2 f8c5 d2d3 h7h5 a2a3 h5h4 h2h3 a7a5 a3a4 f6d5 c3d5 c6d5 e2c3 d5c6 c1d2 f7f6 c3e4 c5e7 d2c3 e8f7 e1d1 h8d8 h1e1 e7b4 e1e2 f7g6 e4d2 c6d5 d2f1 d5c6 f1e3 d8c8 e3c4');
INSERT INTO evaluations VALUES(13580,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(13580,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(13580,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(13580,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(13580,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(13580,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(13580,36,'cp -10','g1f3 f6e4 e1g1 f8e7 b1d2 f7f5 d3b5 e8g8 b5c6 b7c6 f3e5 e7d6 d1a4 c8d7 d2e4 f5e4 a4a6 d6e5 f4e5 f8f1 a6f1 c5c4 f1f4 h7h6 b2b4 c4b3 a2b3 d8b6 c1a3 b6b3 a3d6 g8h7 h2h3 d7e8 f4g4 b3c3 a1a3 c3e1 g1h2 e8g6 g4e6 a8e8');
INSERT INTO evaluations VALUES(17936,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(17936,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(17936,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(17936,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(17936,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(17936,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(17936,36,'cp 167','d7d6 c1f4 g8e7 e1c1 e7c6 d4e3 c8e6 f1e2 f8e7 e4e5 e8g8 h2h4 d8b8 e5d6 e7d6 h4h5 b8a7 f4d6 a7e3 f2e3 c7d6 d1d6 a8c8 e2f3 f8e8 h1h4 c6a5 d6b6 e6f5 h4f4 e8e5 a2a3 f5e6 f4e4 e5c5 e4d4 g8f8 f3b7 a5b7 b6b7 c5h5');
INSERT INTO evaluations VALUES(17824,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(17824,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(17824,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(17824,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(17824,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(17824,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(17824,36,'cp 180','c1f4 g8e7 e1c1 e7c6 d4e3 c8e6 h2h4 d8d7 c3d5 f8e7 h4h5 h7h6 e3g3 e8c8 g3g7 e6d5 e4d5 c6e5 g7g3 d8e8 g3e3 e7f6 g2g3 d7a4 e3b3 a4b3 a2b3 c8d8 f1e2 e5d7 e2g4 d7e5 g4f5 e5f3 f4e3 f6g5 e3g5 f3g5');
INSERT INTO evaluations VALUES(23036,30,'cp 457',NULL);
INSERT INTO evaluations VALUES(23036,31,'cp 460',NULL);
INSERT INTO evaluations VALUES(23036,32,'cp 455',NULL);
INSERT INTO evaluations VALUES(23036,33,'cp 462',NULL);
INSERT INTO evaluations VALUES(23036,34,'cp 459',NULL);
INSERT INTO evaluations VALUES(23036,35,'cp 471',NULL);
INSERT INTO evaluations VALUES(23036,36,'cp 465','f1b5 c7c6 b5e2 d7d6 c2c4 f8e7 g1f3 g8f6 e1g1 e8g8 a2a4 f8e8 d2d3 c8f5 g1h1 h7h6 b1d2 f5h7 c4c5 f6d7 c5d6 e7d6 d2c4 d6c7 a4a5 h7f5 e3e4 f5h7 d1b3 b7b5 a5b6 a7b6 a1a8 d8a8 f1a1 a8b7 a1c1 d7c5 b3c2 c5d7');
INSERT INTO evaluations VALUES(23014,30,'cp 465',NULL);
INSERT INTO evaluations VALUES(23014,31,'cp 450',NULL);
INSERT INTO evaluations VALUES(23014,32,'cp 444',NULL);
INSERT INTO evaluations VALUES(23014,33,'cp 455',NULL);
INSERT INTO evaluations VALUES(23014,34,'cp 455',NULL);
INSERT INTO evaluations VALUES(23014,35,'cp 462',NULL);
INSERT INTO evaluations VALUES(23014,36,'cp 465','c7c6 b5e2 d7d6 c2c4 g8f6 g1f3 f8e7 a2a4 e8g8 b1c3 c8f5 e1g1 f8e8 g1h1 a7a6 a4a5 a8c8 d2d4 h7h6 a1a2 e7f8 f3g1 d8c7 e2d3 e5e4 d3e2 d6d5 c4c5 g8h7 a2a1 g7g6 g2g4 f5d7 f2f3 e4f3 e2f3');
INSERT INTO evaluations VALUES(19096,30,'cp 459',NULL);
INSERT INTO evaluations VALUES(19096,31,'cp 457',NULL);
INSERT INTO evaluations VALUES(19096,32,'cp 465',NULL);
INSERT INTO evaluations VALUES(19096,33,'cp 465',NULL);
INSERT INTO evaluations VALUES(19096,34,'cp 464',NULL);
INSERT INTO evaluations VALUES(19096,35,'cp 475',NULL);
INSERT INTO evaluations VALUES(19096,36,'cp 469','b5e2 d7d6 c2c4 f8e7 g1f3 g8f6 e1g1 e8g8 b1c3 c8f5 a2a4 f8e8 g1h1 e7f8 d2d4 a8c8 a4a5 a7a6 a1a2 h7h6 f3g1 c8b8 b2c1 b7b5 a5b6 d8b6 g1f3 b6b3 c1d2 b3d1 c3d1 a6a5 d4e5 d6e5 a2a5 b8b1 f3e5');
INSERT INTO evaluations VALUES(19102,30,'cp 455',NULL);
INSERT INTO evaluations VALUES(19102,31,'cp 475',NULL);
INSERT INTO evaluations VALUES(19102,32,'cp 460',NULL);
INSERT INTO evaluations VALUES(19102,33,'cp 467',NULL);
INSERT INTO evaluations VALUES(19102,34,'cp 458',NULL);
INSERT INTO evaluations VALUES(19102,35,'cp 462',NULL);
INSERT INTO evaluations VALUES(19102,36,'cp 461','d7d6 c2c4 f8e7 g1f3 g8f6 b1c3 c8f5 e1g1 e8g8 g1h1 f8e8 a2a4 a8c8 a4a5 h7h6 d2d4 d8c7 f3g1 e5d4 d1d4 e7f8 h2h3 a7a6 d4d2 d6d5 c4d5 e8d8 e2f3 c6d5 c3e2 f6e4 f3e4 d5e4 b2d4 d8d4 e2d4 f5h7 g1e2');
INSERT INTO evaluations VALUES(23340,30,'cp 855',NULL);
INSERT INTO evaluations VALUES(23340,31,'cp 858',NULL);
INSERT INTO evaluations VALUES(23340,32,'cp 858',NULL);
INSERT INTO evaluations VALUES(23340,33,'cp 865',NULL);
INSERT INTO evaluations VALUES(23340,34,'cp 855',NULL);
INSERT INTO evaluations VALUES(23340,35,'cp 854',NULL);
INSERT INTO evaluations VALUES(23340,36,'cp 861','d1h5 g7g6 h5e5 d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 g7c3 b2c3 d7d6 c1g5 h7h6 g5e3 c8e6 f1d3 a7a6 e1d2 g8f6 f2f3 h6h5 h1b1 b7b6 c3c4 f6d7 f3f4 g6g5 f4f5 e6f7 d2c3 g5g4 a2a4 d7e5 d3e2 a6a5 h2h4 e8d7 e3f4 a8g8 f4e5 d6e5 g2g3 d7d6 e2d3');
INSERT INTO evaluations VALUES(23339,30,'cp 851',NULL);
INSERT INTO evaluations VALUES(23339,31,'cp 855',NULL);
INSERT INTO evaluations VALUES(23339,32,'cp 863',NULL);
INSERT INTO evaluations VALUES(23339,33,'cp 853',NULL);
INSERT INTO evaluations VALUES(23339,34,'cp 859',NULL);
INSERT INTO evaluations VALUES(23339,35,'cp 859',NULL);
INSERT INTO evaluations VALUES(23339,36,'cp 848','g7g6 h5e5 d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 g8e7 h2h4 h7h6 c1d2 d7d6 c3b5 e8d7 d2c3 g7c3 b5c3 c7c6 e1d2 d7c7 f2f3 b7b5 f1d3 c8d7 a2a4 b5b4 c3e2 d6d5 e4d5 e7d5 d3e4 d5b6 e4g6 a8d8 e2d4 b6c4 d2c1 d8f8 h4h5 c4e5 a4a5');
INSERT INTO evaluations VALUES(23356,30,'cp 863',NULL);
INSERT INTO evaluations VALUES(23356,31,'cp 865',NULL);
INSERT INTO evaluations VALUES(23356,32,'cp 857',NULL);
INSERT INTO evaluations VALUES(23356,33,'cp 867',NULL);
INSERT INTO evaluations VALUES(23356,34,'cp 875',NULL);
INSERT INTO evaluations VALUES(23356,35,'cp 871',NULL);
INSERT INTO evaluations VALUES(23356,36,'cp 875','h5e5 d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 g8e7 h2h4 h7h6 c1d2 d7d6 c3b5 e8d7 d2c3 g7c3 b5c3 b7b6 f1b5 c7c6 b5d3 d7c7 e1d2 c8d7 f2f3 a7a5 a1f1 h6h5 f1e1 a8f8 e4e5 d7e6 e5d6 c7d6 a2a4 e7f5 b2b3 d6d7 d3f5 g6f5');
INSERT INTO evaluations VALUES(23355,30,'cp 863',NULL);
INSERT INTO evaluations VALUES(23355,31,'cp 865',NULL);
INSERT INTO evaluations VALUES(23355,32,'cp 863',NULL);
INSERT INTO evaluations VALUES(23355,33,'cp 871',NULL);
INSERT INTO evaluations VALUES(23355,34,'cp 859',NULL);
INSERT INTO evaluations VALUES(23355,35,'cp 872',NULL);
INSERT INTO evaluations VALUES(23355,36,'cp 863','d8e7 e5h8 e7g7 h8g7 f8g7 b1c3 d7d6 c3d5 e8d7 f1e2 g8f6 d5f6 g7f6 h2h4 h7h5 c2c3 d7e7 c1g5 c8e6 e1d2 a7a5 d2e3 e7f7 f2f4 a8f8 a2a4 f7g7 a1d1 g7f7 d1f1 d6d5 e4e5 f6e7 g5e7 f7e7 e2d3');
INSERT INTO evaluations VALUES(29518,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(29518,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(29518,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(29518,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(29518,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(29518,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(29518,36,'cp 19','g1f3 e7e6 e2e3 f8d6 f4d6 c7d6 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 b8c6 e1g1 e8g8 b5c6 b7c6 d1e2 d8e7 f1c1 f8c8 h2h3 h7h6 e3e4 a7a5 c1c3 c6c5 a1c1 c5d4 f3d4 c8c3 c1c3 b6a4 c3c1');
INSERT INTO evaluations VALUES(32579,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(32579,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(32579,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(32579,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(32579,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(32579,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(32579,36,'cp 106','h2h3 g4f3 d1f3 b8d7 e1c1 e7e6 f1c4 a5b6 c3e2 f8e7 g2g4 e8g8 h1e1 f6e8 c4b3 e8d6 e2g3 a7a5 a2a4 f8e8 c1b1 e7f8 h3h4 b6a6 h4h5 b7b5 h5h6 g7g6 g3e4 d6e4 f3e4 d7b6 e4c6 b5a4 b3a4 b6a4 c6a4');
INSERT INTO evaluations VALUES(36522,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36522,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36522,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(36522,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36522,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(36522,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(36522,36,'cp 28','d2d4 c8f5 d1b3 b8c6 b3b7 a7a5 b7b3 c6b4 f3h4 b4c2 h4f5 c2a1 b3d3 g6f5 g2a8 d8a8 b1c3 f8b8 d3f5 a8a6 c1e3 e7e6 f5b1 a6c4 b1a1 a5a4 b2b3 a4b3 f1b1 f6d7 a2b3 b8b3 b1b3 c4b3 a1a8 d7f8 a8c6 b3b6 c6b5 g7d4 b5g5 f8g6 e3d4 b6d4');
INSERT INTO evaluations VALUES(44392,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44392,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44392,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(44392,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(44392,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(44392,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44392,36,'cp 23','g1f3 a7a6 a2a4 b8c6 d2d4 c6d4 f3d4 c5d4 d1d4 e7c6 d4e3 d8c7 e3g3 c7g3 h2g3 f8b4 c1d2 d7d6 c3d1 b4d2 e1d2 b7b6 d1e3 c8b7 f2f3 c6e5 c4e2 e8e7 h1h4 h7h6 b2b3 a8c8 h4h5 c8c5 g3g4 e5d7 h5c5');
INSERT INTO evaluations VALUES(44391,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44391,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44391,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44391,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(44391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44391,36,'cp 2','a7a6 d2d4 b7b5 c4d3 c5d4 f3d4 e7c6 c3e2 c8b7 e1g1 f8e7 c1e3 e8g8 f2f4 c6d4 e2d4 b8c6 c2c3 e7c5 d4c2 c5e3 c2e3 b5b4 d1d2 b4c3 b2c3 d8c7 a1c1 a8c8 c3c4 a6a5 c4c5 c6b4 f4f5 b4d3 d2d3 a5a4 f5f6');
INSERT INTO evaluations VALUES(44544,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(44544,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(44544,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(44544,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(44544,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(44544,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(44544,36,'cp -39','c5d4 d1d4 d7d6 d4d2 h7h6 b1c3 e7e5 f4g3 c8e6 e2e4 a7a6 f1d3 f8e7 e1g1 b8d7 a2a4 e8g8 f1d1 a8c8 a4a5 d8c7 g3h4 f8e8 h2h3 c7c6 d1e1 e7d8 a1a3 d7c5 a3a1 e8f8 h4f6 d8f6 d3f1 f6e7 d2e2 c6c7 f3d2');
INSERT INTO evaluations VALUES(44629,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(44629,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(44629,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(44629,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(44629,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(44629,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(44629,36,'cp -27','d1d4 d7d6 d4d2 h7h6 b1c3 e7e5 f4g3 f8e7 e2e4 c8e6 f1d3 a7a6 e1g1 e8g8 a2a4 d8c7 f1e1 e7d8 g3h4 b8d7 h2h3 a8c8 a4a5 c7c6 e1d1 c6c5 c3e2 g7g5 h4g3 c5c6 e2c3 f8e8 a1a3 d7c5 d2e2 c5d3 c2d3 f6h5 g3h2');
INSERT INTO evaluations VALUES(44798,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44798,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44798,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44798,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44798,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44798,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44798,36,'cp 0','b8c6 c1e3 f6g4 e3g1 e7e5 f4e5 g4e5 d1d2 f8g7 f1e2 e8g8 e1c1 c6d4 g1d4 c8e6 c1b1 a8c8 h2h3 a7a6 g2g4 d8h4 d2e1 h4e1 h1e1 b7b5 a2a3 h7h6 a3a4 b5a4 c3a4 c8c7 a4c3 f8b8 b1c1 b8c8 c1b1');
INSERT INTO evaluations VALUES(13042,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13042,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13042,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13042,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13042,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13042,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13042,36,'cp 0','h2h3 c6d4 d1d4 f8g7 d4f2 e8g8 f1d3 c8d7 e1g1 d7c6 a2a3 b7b5 c1d2 f6d7 f4f5 a7a5 c3b5 c6b5 d3b5 g7b2 a1b1 b2f6 d2h6 f6g7 h6g5 g7f6');
INSERT INTO evaluations VALUES(45164,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(45164,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(45164,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(45164,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(45164,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(45164,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(45164,36,'cp 58','d2d4 c5d4 f3d4 g7g6 e2e4 b8c6 f1e2 c6d4 d1d4 f8g7 e1g1 c8e6 c1e3 d8a5 a1b1 e8g8 d4d2 f8c8 b2b3 a7a6 a2a4 f6d7 e3d4 g7d4 d2d4 a5g5 g1h1 g5f4 g2g3 f4f6 d4d2 c8c5 b1c1 a8c8 f2f4 d7b6 c1d1 g8h8 d2e3 e6d7 b3b4');
INSERT INTO evaluations VALUES(45170,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(45170,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(45170,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(45170,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(45170,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(45170,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(45170,36,'cp 48','c5d4 f3d4 b8c6 e2e4 c6d4 d1d4 g7g6 f1e2 f8g7 c1e3 e8g8 d4d3 a7a5 e1g1 a5a4 e3d4 d8a5 f2f4 c8g4 e4e5 g4e2 d3e2 f6d7 c3d5 e7e6 b2b4 a4b3 a2b3 a5d8 a1a8 d8a8 d5f6 g7f6 e5f6 a8a3 e2e3 e6e5 f4e5 d6e5 d4a1 a3c5 e3c5 d7c5');
INSERT INTO evaluations VALUES(45968,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(45968,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(45968,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(45968,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45968,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(45968,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(45968,36,'cp -28','h7h6 g5h4 g7g5 h4g3 f6h5 d3d4 h5g3 h2g3 f8g7 c2c3 c5d4 f3d4 b8c6 d4c2 a7a6 b1d2 b7b5 c4e2 a8b8 c2e3 c8b7 d1c2 d8b6 a2a4 c6e5 a4b5 a6b5 d2b3 e8g8 e1g1 b5b4 b3d4 b8c8 f1d1 b4c3 b2c3 b6c5');
INSERT INTO evaluations VALUES(48314,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(48314,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48314,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(48314,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48314,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(48314,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48314,36,'cp 6','g1f3 b7b6 e2e3 c8b7 f1d3 f8d6 b1d2 e8g8 c2c4 b8d7 e1g1 c7c5 c4d5 f6d5 d2e4 d6e7 e4g3 a7a6 d1e2 c5d4 f3d4 b6b5 f1d1 d5f6 a2a4 b5a4 b3a4 d8a5 d4b3 a5c7 g3h5 f6h5 e2h5 d7f6 h5a5 f8c8 a5c7 c8c7');
INSERT INTO evaluations VALUES(48313,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(48313,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(48313,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48313,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48313,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(48313,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48313,36,'cp 0','b7b6 e2e3 c8b7 b1d2 f8e7 f1e2 b8d7 e1g1 e8g8 a2a4 a7a5 c2c4 a8c8 a1c1 c7c5 c4d5 f6d5 d2c4 d5b4 f3e5 d7e5 c4e5 b4a2 c1a1 a2b4');
INSERT INTO evaluations VALUES(54201,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(54201,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(54201,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(54201,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(54201,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(54201,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(54201,36,'cp -35','c2c3 g8f6 e2e3 f8e7 f1d3 b7b6 d1e2 e8g8 e1g1 f6e4 b2b3 c8b7 c1b2 b8d7 b1d2 a8c8 f3e5 e4d2 e2d2 d7f6 d2e2 c5d4 c3d4 e7d6 a1c1 d8e7 e5g4 f6e4 g4f2 f8d8 f2e4 d5e4 d3c4 g7g6 c4a6 b7a6 e2a6 e7d7 a2a4 f7f6 c1c8');
INSERT INTO evaluations VALUES(54198,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(54198,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(54198,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(54198,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(54198,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(54198,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(54198,36,'cp -29','g8f6 e2e3 f8e7 f1d3 e8g8 e1g1 b7b6 d1e2 f6e4 b2b3 c8b7 c1b2 b8d7 b1d2 a8c8 f3e5 e4d2 e2d2 d7f6 d2e2 e7d6 d4c5 d6c5 a2a4 a7a5 b3b4 c5d6 b4b5 d6c5 f1d1 d8e7 a1c1 f8d8 c3c4 d5c4 c1c4 f6d5 b2d4');
INSERT INTO evaluations VALUES(588,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(588,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(588,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(588,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(588,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(588,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(588,36,'cp 73','b4c3 b2c3 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 h7h6 h2h4 c8e6 g2b7 a8e8 b7c6 e8b8 c6f3 e2c2 c4c5 d6c5 f4c7 b8c8 c7e5 c5d4 e5d4 c8d8 d4a7 c2a2 a1a2 e6a2 f3g2 d8d2 a7e3 d2c2 f1e1 a2d5 g2d5 f6d5');
INSERT INTO evaluations VALUES(590,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(590,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(590,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(590,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(590,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(590,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(590,36,'cp 59','b2c3 f7f6 e5f6 g4f6 d5d3 d7d6 g2g3 e8g8 f1g2 f8e8 e1g1 e7e2 d3e2 e8e2 f3d4 c6d4 c3d4 c8e6 g2b7 a8b8 b7f3 e2c2 c4c5 f6d5 f1c1 b8b2 c1c2 b2c2 f3d1 c2c3 f4c1 d6c5 c1b2 c3d3 d4c5 d5c3 d1b3 g8f7 g1f1 g7g5 b3e6 f7e6 a1e1 e6d7 c5c6 d7d6 e1e8 d6c6 e8e7 c3a2 e7h7');
INSERT INTO evaluations VALUES(3876,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3876,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3876,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3876,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3876,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3876,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3876,36,'cp 0','f6d5 c4d5 a7a5 a2a4 c6d4 f3d4 c5d4 c2c3 d4a7 d1f3 d8h4 h2h3 h7h6 c1e3 c8e6 d5e6 f7e6 f3e2 a7e3 f2e3 c7c5 f1f8 a8f8 a1f1 h4g5 f1f8 g8f8 e2f2 f8e7 c3c4 g5f6 f2f6 e7f6 g1h2 f6g6 g2g4 h6h5 h2g3 h5g4 g3g4 g6h6 h3h4');
INSERT INTO evaluations VALUES(3897,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3897,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3897,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3897,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3897,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3897,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3897,36,'cp 0','c4d5 a7a5 a2a4 c6d4 f3d4 c5d4 c2c3 d4a7 d1f3 d8h4 h2h3 h7h6 c1e3 c8e6 d5e6 f7e6 f3e2 a7e3 f2e3 c7c5 f1f8 a8f8 a1f1 f8f7 f1f7 g8f7 e2f2 h4f6 f2e1 f6g5 g1h2 f7e7 c3c4 b7b6 e1f2 g5f6 f2f6 e7f6 g2g3 f6f7 h2g2 g7g6 g2f3 f7g8');
INSERT INTO evaluations VALUES(3895,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3895,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3895,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3895,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3895,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3895,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3895,36,'cp 0','a7a5 a2a4 c6d4 f3d4 c5d4 c2c3 d4a7 d1f3 d8h4 h2h3 h7h6 c1e3 c8e6 d5e6 f7e6 f3e2 a7e3 f2e3 c7c5 f1f8 a8f8 a1f1 f8f7 f1f7 g8f7 e2f2 h4f2 g1f2 f7g6 f2f3 g6g5 g2g3 g5h5 f3g2 h5g6 c3c4 g6f7 g3g4 f7e7 g2g3 g7g6 g3h2 e7f8 h2g1');
INSERT INTO evaluations VALUES(6386,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6386,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(6386,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6386,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6386,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(6386,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6386,36,'cp 28','d8e7 e5f3 e7e4 b5e2 c6d4 f3d4 e4d4 e1g1 d7d5 c2c3 d4f6 d2d4 c5d6 c3c4 d5c4 b1c3 g8e7 c3e4 d6h2 g1h2 f6h4 h2g1 h4e4 f1e1 e4h4 e2c4 e8g8 b2b3 c8g4 d1d2 f8e8 d4d5 g4d7 c1b2 e7f5 a1c1 c7c6 c4d3 h4g4 f2f3 g4g6 d5c6 d7c6 d2f4 e8e1 c1e1 g6g3 f4g3');
INSERT INTO evaluations VALUES(8968,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(8968,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(8968,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(8968,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(8968,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(8968,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(8968,36,'cp 96','b1d2 f6e4 a2a3 b4d6 f4g3 e8g8 d2e4 d5e4 f3d2 f7f5 c4c5 d6e7 b2b4 e7f6 f1c4 g8h8 b4b5 c6d4 a1c1 e6e5 e3d4 d8d4 d1e2 f5f4 d2b3 d4d7 e2e4 d7f5 c4d3 f5h5 e1g1 f4g3 h2g3 f6g5 e4e5 f8e8 e5c7 g5c1 f1c1 c8e6');
INSERT INTO evaluations VALUES(9124,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(9124,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(9124,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(9124,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(9124,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(9124,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(9124,36,'cp 40','e4d5 f6d5 f3g5 e8g8 d1f3 c6e7 c1a3 c7c6 e1g1 h7h6 g5e4 b7b5 c4b3 a7a5 f1e1 d5f4 e4d2 a5a4 a3e7 d8e7 f3f4 a4b3 d2b3 e5f4 e1e7 c8e6 e7c7 e6d5 b3d4 a8a6 g1f1 f8a8 a2a3 g7g6 f1e2 d5g2 f2f3 a6a3 a1a3 a8a3 d4c6 g2h3 c6d8 h3e6 d8e6 f7e6 c7c5 a3a1 c5b5 a1h1');
INSERT INTO evaluations VALUES(9105,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(9105,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(9105,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(9105,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(9105,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(9105,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(9105,36,'cp 38','f6d5 f3g5 e8g8 d1f3 c6e7 c1a3 c7c6 e1g1 h7h6 g5e4 b7b5 c4b3 a7a5 f1e1 d5f4 e4d2 a5a4 a3e7 d8e7 f3f4 a4b3 d2b3 e5f4 e1e7 c8e6 e7c7 e6d5 b3d4 a8a6 g1f1 f8e8 f2f3 e8a8 f1e2 a6a2 a1a2 a8a2 e2d2 a2a1 d4c6 d5c6 c7c6 a1g1 d3d4 g1g2 d2d3 g2h2 d4d5 g7g5 c3c4 b5c4 c6c4');
INSERT INTO evaluations VALUES(9185,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(9185,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(9185,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(9185,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(9185,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(9185,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(9185,36,'cp 47','f3g5 e8g8 d1f3 c6e7 c1a3 c7c6 e1g1 h7h6 g5e4 b7b5 c4b3 a7a5 f1e1 d5f4 e4d2 a5a4 a3e7 d8e7 f3f4 a4b3 d2b3 e5f4 e1e7 c8e6 e7c7 e6d5 b3d4 f8e8 h2h4 a8a6 d4f5 g8h7 c3c4 b5c4 d3c4 d5e6 f5d6 e8a8 d6f7 a6a2 a1a2 a8a2 f7d6 h7g6 c7c6 a2c2 f2f3 g6h5');
INSERT INTO evaluations VALUES(19055,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19055,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19055,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(19055,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(19055,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19055,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19055,36,'cp 0','c2c3 g8f6 f3g5 d7d5 e4d5 c6a5 c4b5 c8d7 d1e2 f8d6 b2b4 h7h6 g5f3 e8g8 b4a5 d4d3 e2b2 a8c8 e1g1 e5e4 f3d4 f8e8 c1d2 a7a6 b5d7 d8d7 b2b3 f6d5 b1a3 c8c5 a3c4 d6h2 g1h2 d7c7 h2g1 c5c4 a1b1 d5c3 d2c3 c4c3 b3b7 c7e5 b7b6 e4e3 f2e3 e5e3 g1h1');
INSERT INTO evaluations VALUES(19054,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19054,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(19054,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19054,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19054,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19054,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19054,36,'cp 0','g8f6 f3g5 d7d5 e4d5 c6a5 c4b5 c8d7 d1a4 f8e7 c3d4 e8g8 b5d7 f6d7 g5f3 e5d4 e1g1 a8c8 b1a3 d7f6 f1d1 d8d5 d1d4 d5f5 d4f4 f5h5 f4h4 h5f5');
INSERT INTO evaluations VALUES(12583,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12583,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12583,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12583,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12583,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12583,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12583,36,'cp 0','f3g5 d7d5 e4d5 c6a5 c4b5 c8d7 d1a4 f8e7 g5f3 e8g8 b5d7 f6d7 c3d4 e5d4 e1g1 a8c8 b1a3 d7f6 f1d1 d8d5 d1d4 d5f5 d4f4 f5d5');
INSERT INTO evaluations VALUES(19210,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19210,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19210,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19210,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19210,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19210,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19210,36,'cp 0','c5d4 c2c3 d4c3 b1c3 f8g7 e1g1 d7d6 e4e5 c8g4 e5d6 d8d6 d1a4 g4f3 g2f3 g8f6 c1f4 d6c5 f4e3 c5d6');
INSERT INTO evaluations VALUES(19207,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19207,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19207,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19207,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19207,36,'cp 0','c2c3 d4c3 b1c3 f8g7 e1g1 d7d6 e4e5 c8g4 e5d6 d8d6 d1a4 g4f3 g2f3 g8f6 c1f4 d6c5 f4e3 c5d6');
INSERT INTO evaluations VALUES(24833,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24833,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24833,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24833,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24833,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24833,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(24833,36,'cp 0','f6h5 f4d2 h5f6 d2f4');
INSERT INTO evaluations VALUES(24911,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24911,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24911,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24911,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24911,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24911,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24911,36,'cp 0','e7e6 h2h3 g4h5 f1b5 a7a6 b5c6 b7c6 g2g4 h5g6 f3e5 f6d7 e5c6 d8c8 e3e4 c8b7 e4d5 b7b2 c3a4 b2b5 a4c3');
INSERT INTO evaluations VALUES(25205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25205,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(25205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25205,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25205,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25205,36,'cp 0','h2h3 g4h5 f1b5 f8b4 e1g1 b4c3 b2c3 e8g8 b5c6 b7c6 g2g4 h5g6 f3e5 f6e4 e5c6 d8e8 c6e5 e4c3 d1e1 c3a4 e1e2 a4c3');
INSERT INTO evaluations VALUES(25896,30,'cp -211',NULL);
INSERT INTO evaluations VALUES(25896,31,'cp -217',NULL);
INSERT INTO evaluations VALUES(25896,32,'cp -212',NULL);
INSERT INTO evaluations VALUES(25896,33,'cp -230',NULL);
INSERT INTO evaluations VALUES(25896,34,'cp -217',NULL);
INSERT INTO evaluations VALUES(25896,35,'cp -232',NULL);
INSERT INTO evaluations VALUES(25896,36,'cp -230','e5d7 c3c4 c5d4 d1d4 d5c4 c1e3 g8f6 d4c4 a8c8 c4e2 f6d5 e1g1 d5e3 e2e3 d8a5 f1d1 f8e7 b1c3 d7f6 e3e5 a5b6 e5e2 e8g8 g2g3 a7a6 a1c1 f8d8 d1d8 c8d8 c1c2 h7h6 f3e5 b6d6 h2h4 f6d7 e5d7');
INSERT INTO evaluations VALUES(26010,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(26010,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(26010,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(26010,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(26010,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(26010,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26010,36,'cp -9','b7c6 c2c4 e7e5 d4f3 d7e6 e1g1 g8f6 d1e2 d8c7 b1c3 f8e7 f1d1 h7h6 h2h3 e8g8 c1e3 c7b8 b2b3 f6d7 f3d2 b8b4 a1c1 f7f5 e4f5 f8f5 d2f1 f5f7 f1g3 b4b7 c3e4 d6d5 c4d5 c6d5 e4c5 d7c5 e3c5 e7c5 c1c5 b7b6');
INSERT INTO evaluations VALUES(25836,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25836,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(25836,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(25836,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25836,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25836,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25836,36,'cp 0','c2c4 e7e5 d4f3 d8a5 b1c3 d7e6 d1d3 g8f6 f3g5 e6c8 g5f3');
INSERT INTO evaluations VALUES(32384,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(32384,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(32384,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(32384,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(32384,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(32384,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(32384,36,'cp 206','b1c3 b8d7 h2h3 g4f3 d1f3 f8d6 b2b3 g8e7 c1b2 f7f5 g2g4 e8g8 e1c1 e7g6 g4f5 g6h4 f3g4 f8f5 c3e4 h7h5 g4e2 d8f8 e2e3 h4f3 f1e2 f5f4 d1g1 f3g1 h1g1 a8d8 e2c4 d7b6 e4g5 b6c4 b3c4 d8e8 g5e6 f4f3 e3h6');
INSERT INTO evaluations VALUES(37819,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(37819,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(37819,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(37819,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(37819,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(37819,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(37819,36,'cp -12','g8f6 g2g3 e8g8 f1g2 b8c6 c3d5 a7a5 d5f6 g7f6 a2a4 d7d5 e1g1 c8g4 h2h3 g4e6 f3g5 f6g5 c1g5 d5e4 g5h6 f8e8 d3e4 d8b6 d1c1 a8d8 f2f4 d8d7 g1h2 e8d8 b2b3 f7f6 a1b1 b6c7 f1f2 e6f7 c1e3 b7b6 e4e5 c6d4 b1b2');
INSERT INTO evaluations VALUES(38712,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(38712,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(38712,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(38712,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(38712,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(38712,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(38712,36,'cp 60','b8d7 d2d4 c5b4 d1b3 b4a5 b3a3 c7c5 f1e2 g8f6 e1g1 e8g8 f1d1 f8e8 c3b5 a5b4 a3b3 f6e4 d4c5 b4c5 f3d2 e4d2 c1d2 a7a6 b5c3 d7f6 c3a4 d8c7 a4c5 d6c5 d2e1 c8d7 f2f3 a6a5 e3e4 f6h5 b3d3 d7a4 d3d6 c7d6 d1d6 h5f4');
INSERT INTO evaluations VALUES(39350,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(39350,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(39350,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(39350,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(39350,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(39350,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(39350,36,'cp 7','g8f6 g1f3 b4c3 d2c3 d7d6 c1g5 h7h6 g5h4 a7a6 f3d2 b8d7 f2f3 d7f8 a2a4 c8e6 h4f2 f8g6 c4e6 f7e6 d1e2 e8g8 a4a5 d8d7 f2e3 g6f4 e2f1 f4g6 g2g3 g6e7 f1c4 e7c6 c4e2 d7f7 g3g4 d6d5 g4g5');
INSERT INTO evaluations VALUES(48036,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(48036,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(48036,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(48036,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(48036,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(48036,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(48036,36,'cp 44','f8e7 f2f4 h7h6 g5f6 e7f6 g1f3 c7c5 f1b5 c8d7 e4e5 d6e5 d4e5 f6e7 d1d2 d8a5 b5d7 b8d7 e1c1 d7b6 d2e2 e8g8 f3d2 a8d8 d2b3 a5b4 g2g3 b6c4 d1d8 f8d8 h1d1 d8d1 c3d1 c4a5 b3d2 a5c6 a2a3 c6d4 e2d3 b4b6 d1e3 b6c6 c2c3 c6h1 e3f1 d4f5 d3e2');
INSERT INTO evaluations VALUES(40880,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(40880,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(40880,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(40880,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(40880,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40880,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(40880,36,'cp 61','g1f3 h7h6 g5f4 a7a6 f1d3 c7c5 d4c5 d6c5 h2h3 b8c6 e1g1 b7b5 e4e5 f6d5 c3d5 e6d5 c2c3 c8e6 d1e2 d8d7 f1d1 e8g8 d3b1 e6f5 d1d5 d7d5 b1f5 a8d8 a2a4 c5c4 g2g4 b5b4 e5e6 e7f6 f4c7 b4c3 c7d8 f8d8 b2c3 f6c3 a1c1');
INSERT INTO evaluations VALUES(40879,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40879,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(40879,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(40879,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(40879,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(40879,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(40879,36,'cp 31','b8d7 d1e2 e8g8 e1c1 f6d5 e4d5 e7g5 f3g5 d8g5 e2d2 g5d2 c1d2 d7f6 d5e6 c8e6 f2f3 g7g6 f1d3 a7a5 a2a4 c7c6 d1b1 h7h5 h1e1 g8g7 h2h4 f8d8 b2b4 a5b4 b1b4 a8a7 e1a1 d8a8 a1b1 f6d5 c3d5 e6d5 c2c4 d5e6 d3c2');
INSERT INTO evaluations VALUES(52198,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(52198,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(52198,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(52198,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(52198,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(52198,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(52198,36,'cp 170','b1c3 d8c7 g1e2 h7h5 e2g3 g7g6 f1e2 d7d6 g3f1 b8d7 f1e3 d7b6 a2a4 a7a5 e3c4 f6f5 c1g5 b6c4 e2c4 f5e4 c3e4 c8f5 f2f3 f8e7 h2h4 f5e4 f3e4 e7g5 h4g5 e8c8 d5c6 c7c6 c4d5');
INSERT INTO evaluations VALUES(52197,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(52197,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(52197,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(52197,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(52197,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(52197,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(52197,36,'cp 160','d8c7 g1e2 h7h5 e2g3 g7g6 f1e2 d7d6 g3f1 b8d7 a2a4 a7a5 c1e3 f6f5 f2f3 f8e7 f1d2 f5f4 e3f2 g8h6 d2c4 h6f7 d1d2 a8b8 g2g3 e8g8 g3f4 e5f4 d2f4 f7e5 f4g3 g8h7 e1c1 e5c4');
INSERT INTO evaluations VALUES(52950,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(52950,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52950,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(52950,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(52950,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(52950,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(52950,36,'cp 15','c4b3 g8f6 d2d3 e7e6 g1f3 b7b5 e1g1 f8e7 f1e1 c8b7 d1e2 e8g8 d3d4 d8c7 e4e5 d6e5 d4e5 f6d5 c2c3 b8d7 b3c2 f8d8 e2e4 d7f8 h2h4 c6c5 e4g4 g8h8 a2a4 a7a6 h4h5 c5c4 a4b5 a6b5');
INSERT INTO evaluations VALUES(55921,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(55921,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(55921,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(55921,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(55921,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(55921,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(55921,36,'cp -13','d5e4 d3e4 d8d1 e1d1 b8d7 f1d3 b7b6 g1f3 c8b7 d1e2 e8c8 c1e3 h7h6 b1c3 f8b4 c3b5 g8f6 a2a3 b4e7 b5a7 c8b8 a7b5 f6e4 a3a4 e4f6 a4a5 b6a5 a1a5 f6d5 e3a7 b8c8 g2g3 g7g5 a7d4 d5f4 g3f4 g5g4 f3d2 b7h1 a5a7 h1b7 d3e4 c7c6 d4h8 d8h8');
INSERT INTO evaluations VALUES(56088,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56088,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(56088,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(56088,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56088,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(56088,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(56088,36,'cp -19','d3e4 d8d1 e1d1 b8d7 f1d3 b7b6 d1e2 c8b7 g1f3 g7g6 h1d1 e8c8 a2a4 a7a5 b1c3 f8g7 e4e5 f7f6 d3c4 g8h6 c3b5 c8b8 b5d4 d7c5 c4b5 f6e5 f4e5 h6f7 c1f4 d8f8 c2c3 g6g5 f3g5 f7g5 f4g5 b7g2 g5e7 f8f7 e7c5 b6c5 d4c6 g2c6');
INSERT INTO evaluations VALUES(4406,30,'cp 617',NULL);
INSERT INTO evaluations VALUES(4406,31,'cp 618',NULL);
INSERT INTO evaluations VALUES(4406,32,'cp 621',NULL);
INSERT INTO evaluations VALUES(4406,33,'cp 632',NULL);
INSERT INTO evaluations VALUES(4406,34,'cp 624',NULL);
INSERT INTO evaluations VALUES(4406,35,'cp 640',NULL);
INSERT INTO evaluations VALUES(4406,36,'cp 636','d4c6 b7c6 e3c5 d7d6 c5d4 c6c5 d4f6 d8f6 d1d2 a8b8 b2b3 c8b7 f1d3 d6d5 e1g1 c5c4 b3c4 d5e4 c3e4 b7e4 d3e4 f8e8 a1e1 a7a6 e1e3 e8e7 e4f3 e7e3 f2e3 f6e7 f3d5 b8f8 g1h1 h7h6 d2f2 c7c5 f2f3 a6a5 f3g3 a5a4');
INSERT INTO evaluations VALUES(4251,30,'cp 613',NULL);
INSERT INTO evaluations VALUES(4251,31,'cp 633',NULL);
INSERT INTO evaluations VALUES(4251,32,'cp 605',NULL);
INSERT INTO evaluations VALUES(4251,33,'cp 625',NULL);
INSERT INTO evaluations VALUES(4251,34,'cp 613',NULL);
INSERT INTO evaluations VALUES(4251,35,'cp 636',NULL);
INSERT INTO evaluations VALUES(4251,36,'cp 636','b7c6 e3c5 d7d6 c5d4 c6c5 d4f6 d8f6 d1d2 a7a5 f1e2 a8b8 b2b3 c8b7 e1g1 b8e8 e2d3 f6h4 f2f3 g8h8 c3e2 f7f5 e4f5 a5a4 e2f4 b7c8 a1e1 c8d7 e1e8 f8e8 b3a4 h4f6 a4a5 d7f5 d3f5 f6f5 a5a6');
INSERT INTO evaluations VALUES(9755,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9755,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9755,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9755,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9755,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9755,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9755,36,'cp 0','e8g8 e1g1 d7d6 c3a4 c5b6 a4b6 a7b6 a2a3 d8e7 c1e3 c8e6 c4e6 e7e6 c2c4 c6e7 f1e1 f6d7 d3d4 e7g6 e3d2 e6f6 d2e3');
INSERT INTO evaluations VALUES(4169,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4169,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4169,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4169,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4169,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4169,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4169,36,'cp 0','e1g1 d7d6 c3a4 c5b6 a4b6 a7b6 a2a3 d8e8 c1e3 c8e6 c4b5 e8e7 h2h3 c6a7 b5a4 a7c6');
INSERT INTO evaluations VALUES(4170,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4170,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4170,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4170,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4170,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4170,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4170,36,'cp 0','d7d6 c3a4 c5b6 a4b6 a7b6 a2a3 d8e8 c1e3 c8e6 c4b5 e8e7 h2h3 c6a7 b5a4 a7c6');
INSERT INTO evaluations VALUES(3608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3608,36,'cp 0','c3a4 c5b6 a4b6 a7b6 a2a3 d8e8 c1e3 c8e6 c4b5 e8e7 h2h3 c6a7 b5a4 a7c6');
INSERT INTO evaluations VALUES(10409,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(10409,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(10409,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10409,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10409,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10409,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10409,36,'cp 0','b4c3 b2c3 e5e4 f2f3 c6e5 f3e4 e5d3 e1e2 f6g4 d1a4 g4e5 g1f3 e8g8 f3e5 d3e5 d2d4 e5g4 c4c5 d7d6 h2h3 c8d7 a4c2 g4f6 e2f2 f6h5 c5d6 d8g5 g3g4 g5h4 f2g1 h4e1 g1h2 e1g3 h2g1');
INSERT INTO evaluations VALUES(10450,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10450,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10450,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10450,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10450,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(10450,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(10450,36,'cp -11','b2c3 e5e4 f2f3 c6e5 f3e4 e5d3 e1e2 f6g4 d1a4 g4e5 g1f3 e8g8 f3e5 d3e5 d2d4 e5g4 e4e5 d7d6 h2h3 g4h6 e3e4 d6e5 c4c5 c8d7 a4c4 d8f6 h1f1 f6g6 e2f2 a7a6 c4d5 d7b5 f1e1 a8e8 d4e5 b5c6 d5d4 e8d8 d4c4 c6b5 c4b3 d8d3 c1f4 g6c6 a2a4 c6c5 f4e3 c5c4 b3c4 b5c4 a1c1 c4e6');
INSERT INTO evaluations VALUES(10448,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(10448,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10448,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10448,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10448,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10448,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10448,36,'cp -9','e5e4 f2f3 c6e5 f3e4 e5d3 e1e2 f6g4 d1a4 g4e5 g1f3 e8g8 f3e5 d3e5 d2d4 e5g4 e4e5 d7d6 h2h3 g4h6 e3e4 d6e5 c4c5 g8h8 d4e5 d8e7 c5c6 b7c6 a4a3 h6g8 a3e7 g8e7 c1f4 e7g6 h1d1 c8e6 a1b1 e6a2 b1b7 f8e8 b7c7');
INSERT INTO evaluations VALUES(14066,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(14066,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(14066,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(14066,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(14066,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(14066,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(14066,36,'cp 71','a7a6 b5e2 c6d4 f3d4 e5d4 d1d4 e7c6 d4d3 f8e7 b1c3 c6e5 d3g3 e8g8 f2f4 e5c6 c3d5 e7h4 g3d3 c6e7 d5e3 f7f5 e2f3 g8h8 e4f5 e7f5 f3e4 g7g6 c1d2 f5e3 d2e3 h4f6 c2c3 c7c6 f4f5 c8f5 e4f5 g6f5 f1f5 f6e5 f5f8 d8f8 a1f1');
INSERT INTO evaluations VALUES(18787,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(18787,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(18787,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(18787,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(18787,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(18787,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18787,36,'cp 13','d7d5 c4d5 e6d5 f1e2 g8f6 e1g1 c5d4 f3d4 f8d6 b2b3 d8e7 c1b2 e7e5 g2g3 e8g8 b1d2 c6d4 b2d4 e5e7 d4f6 e7f6 e2g4 g7g6 g4c8 a8c8 d2f3 f8d8 d1e2 d6c5 a1c1 d5d4 f3d4 c5d4 c1c8 d8c8 e3d4 f6d4 f1d1 d4c3 e2e4 c8c7 g1g2 h7h6 a2a3 g8g7');
INSERT INTO evaluations VALUES(29839,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(29839,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(29839,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(29839,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(29839,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(29839,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(29839,36,'cp 78','c1e3 c7c5 a2a3 f8e7 h2h3 g4e6 b2b4 e6c4 d3c4 b8c6 e1g1 c5b4 a3b4 c6b4 e3a7 e8g8 d1e2 d8d7 f1b1 b4c6 a7e3 a8a1 b1a1 f8c8 a1b1 c8a8 c3d5 a8a3 f3d2 c6d4 e3d4 f6d5 d4e5 d5c3 e5c3 a3c3 d2f1 e7f6 f1e3 d7c6 b1d1 c6e4');
INSERT INTO evaluations VALUES(31757,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(31757,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(31757,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(31757,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(31757,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(31757,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(31757,36,'cp 77','b1c3 g8f6 g1f3 b8a6 h2h3 h7h6 e2e3 c7c6 f1d3 f5d3 d1d3 e7e6 e3e4 d6d5 e1g1 f8e7 c1f4 e8g8 f1d1 d5e4 c3e4 f6e4 d3e4 e7d6 f4d2 d8f6 g2g3 f8e8 d2c3 a8d8 g1g2 f6f5 e4f5 e6f5 d4d5 d6c5 a2a3 c6d5 d1d5 d8d5 c4d5');
INSERT INTO evaluations VALUES(31756,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(31756,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(31756,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(31756,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(31756,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(31756,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(31756,36,'cp 66','g8f6 g1f3 b8a6 h2h3 h7h6 e2e3 e7e6 f1d3 f5d3 d1d3 c7c6 e3e4 d6d5 e1g1 f8e7 c1f4 e8g8 a1d1 d5e4 c3e4 f6e4 d3e4 e7d6 f3e5 d8f6 a2a3 a8d8 f1e1 f6f5 e4e3 a6c7 b2b4 f8e8 f4g3 f7f6 e5d3 b7b6 a3a4 b6b5 a4b5 c6b5 g3d6');
INSERT INTO evaluations VALUES(34063,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(34063,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(34063,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(34063,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(34063,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(34063,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(34063,36,'cp 24','d7d6 c4f7 e8d8 e1g1 d6e5 d4e5 c8d7 f7d5 e7e5 f1e1 e4f2 d1d2 e5f6 d2f2 f6f2 g1f2 d8c8 b1c3 a7a5 a2a3 a8a6 f2g1 a6b6 b2b3 g7g6 c1e3 f8g7 e3b6 c7b6 e1e3 g7d4 a1e1 h8e8 d5e4 d7c6 g1f1 e8f8 f1e2 d4e3 e2e3 b8d7 e3d4 c8c7 e4d5 d7f6 d5f3 f6d7 e1e7 c6f3 g2f3 f8f3 e7h7');
INSERT INTO evaluations VALUES(33803,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(33803,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(33803,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(33803,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(33803,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(33803,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33803,36,'cp 29','c4f7 e8d8 e1g1 d6e5 d4e5 c8d7 f7d5 e7e5 f1e1 e4f2 d1d2 e5f6 d2f2 f6f2 g1f2 d8c8 a2a3 a7a5 b1c3 a8a6 f2g1 a6g6 c3e4 f8e7 d5f7 g6c6 c1d2 h8f8 f7b3 b7b6 d2c3 c6g6 b3c4 d7c6 e1e2 c6b7 e4d2 e7f6 c3f6 g6f6 a1e1 f8d8 c4e6 f6e6 e2e6 d8d2 e6e8 d2d8');
INSERT INTO evaluations VALUES(40006,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40006,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(40006,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40006,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40006,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(40006,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40006,36,'cp 18','b1c3 d7d5 d2d4 e8g8 d4c5 d8a5 e1g1 d5c4 d1d4 b8c6 d4c4 c8e6 c4h4 a5c5 c1e3 c5b4 h4b4 c6b4 e3c5 b4c6 f3d4 c6d4 c5d4 f6e8 d4c5 e8d6 f1d1 g7c3 b2c3 a8c8 c5a7 d6b5 a7d4 b5c3 d4c3 c8c3 g2b7 c3c2 a2a4 c2e2 b7d5 f8b8 d5e6 e2e6');
INSERT INTO evaluations VALUES(40403,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40403,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(40403,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(40403,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40403,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(40403,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40403,36,'cp 0','e8g8 d3e2 b8c6 e1g1 a7a6 a2a4 c8d7 h2h3 a8c8 f1e1 d8c7 d4f3 c6a5 d1d2 f8d8 a1d1 d7c6 e2d3 d6d5 c3d5 c6d5 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(36019,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36019,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36019,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36019,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36019,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36019,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36019,36,'cp 0','d3e2 b8c6 e1g1 c8d7 f1e1 a8c8 h2h3 a7a6 a2a4 d8c7 d4f3 c6a5 d1d2 f8d8 a1d1 d7c6 e2d3 d6d5 c3d5 c6d5 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(44359,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(44359,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44359,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(44359,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(44359,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44359,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44359,36,'cp 40','d4b5 d7d5 c4d5 f8c5 b5c3 e8g8 g2g3 d8b6 e2e3 c8f5 c3a4 b6c7 a4c5 c7c5 b1c3 e5e4 f1g2 b8a6 e1g1 f8e8 d1d4 c5d4 e3d4 a6b4 c1g5 b4d5 f1e1 h7h6 g5f6 d5f6 a1d1 a8d8 d4d5 d8d7 d1d4 e8e5 f2f3 e4e3 f3f4 e5e8 g2f3 h6h5');
INSERT INTO evaluations VALUES(44305,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(44305,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(44305,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(44305,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(44305,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(44305,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(44305,36,'cp 37','d7d5 c4d5 f8c5 b5c3 e8g8 g2g3 d8b6 e2e3 c8f5 c3a4 b6d6 a4c5 d6c5 b1c3 e5e4 f1g2 b8a6 e1g1 a6c7 c1d2 c7d5 d1b3 f8d8 c3d5 f6d5 b3b7 a8b8 b7a6 b8b2 f1c1 c5e7 a6a5 f5g4 d2e1 h7h6 h2h3 g4f3 g2f3 e4f3 c1d1 d5c7 e1c3 b2c2');
INSERT INTO evaluations VALUES(47862,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47862,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(47862,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(47862,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(47862,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(47862,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(47862,36,'cp 19','b1c3 c7c5 f2f4 a7a6 f1c4 b8c6 g1f3 e5f4 c3d5 f8e7 d5f4 e8g8 d2d3 b7b5 c4d5 f6d5 f4d5 e7h4 f3h4 d8h4 g2g3 h4h6 d1f3 c8e6 f3e3 h6e3 d5e3 f7f5 h1f1 f5e4 f1f8 a8f8 d3e4 f8d8 a2a4 c6b4 a4b5 a6b5 e1c1 g8f7 c1b1 b4c6 e3d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(47864,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(47864,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(47864,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(47864,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(47864,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(47864,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47864,36,'cp 0','c7c5 f2f4 a7a6 f1c4 b7b5 c4d5 a8a7 g1f3 b8d7 e1g1 e5f4 c3e2 g7g6 e2f4 f8h6 f3e1 e8g8 e1d3 h6g7 d1f3 a6a5 a2a4 b5a4 a1a4 d7b6 a4a3 c8a6 h2h3 f6d5 b2g7 d5f4 g7f8 f4d3 f8h6 d8h4 f3f6 h4h6 f6d8 g8g7 d8f6');
INSERT INTO evaluations VALUES(56176,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(56176,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(56176,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56176,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56176,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56176,36,'cp 0','c4d5 g8f6 d5f7 e8f7 d1d8 f8b4 d8d2 b4d2 b1d2 h8d8 g1f3 f7g8 f3e5 f6d7 e1c1 d7e5 b2e5 b8a6 f2f3 b7b6 d2b3 c8d7 d1d2 d7a4 d2d8 a8d8 h1d1 d8d1 c1d1 c7c5 d1d2 a6b4 b3c1 b4c6 e5g3 c6b4');
INSERT INTO evaluations VALUES(56122,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56122,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56122,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56122,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56122,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(56122,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56122,36,'cp 0','g8f6 d5f7 e8f7 d1d8 f8b4 d8d2 b4d2 b1d2 h8d8 g1f3 f7g8 f3e5 f6d7 a1c1 d7e5 b2e5 b8a6 e1e2 b7b6 e2e3 a6c5 e5d4 c5e6 d2f3 c8b7 h1d1 d8d6 d4c3 d6d1 c1d1 e6c5 d1d4 c5e6');
INSERT INTO evaluations VALUES(48553,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(48553,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48553,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(48553,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48553,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(48553,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(48553,36,'cp 2','d5f7 e8f7 d1d8 f8b4 d8d2 b4d2 b1d2 h8d8 g1f3 f7g8 f3e5 f6d7 a1c1 d7e5 b2e5 b8a6 e1e2 b7b6 e2e3 a6c5 e5d4 c5e6 d4a1 c8a6 d2c4 a6c4 c1c4 c7c5 c4c1 b6b5 h1d1 c5c4 d1d5 a7a6 a1c3 d8d5 e4d5 e6c5 e3d4 c5d3 c1c2 g8f7 c3a5 f7e7 a2a3 a8c8 f2f3 d3c5 c2c3');
INSERT INTO evaluations VALUES(47553,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(47553,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(47553,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(47553,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(47553,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47553,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47553,36,'cp 12','e8f7 d1d8 f8b4 d8d2 b4d2 b1d2 b8a6 g1f3 h8e8 f3e5 f7g8 e1e2 a6c5 a1c1 c5e4 d2e4 f6e4 c1c7 e4f6 h1d1 c8e6 a2a3 a8b8 f2f3 f6d5 c7c1 e8c8 g2g4 c8c1 d1c1 b8c8 c1c8 e6c8 e2d3 d5e7 d3d4 e7g6 e5g6 h7g6 d4e5 c8d7 b2c1 d7c6 f3f4 g8f7 e5d6 c6b5 d6c7 b5c6');
INSERT INTO evaluations VALUES(35494,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(35494,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(35494,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35494,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(35494,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(35494,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(35494,36,'cp 8','d1d8 f8b4 d8d2 b4d2 b1d2 b8a6 g1f3 h8e8 f3e5 f7g8 e1e2 a6c5 a1c1 c5e4 d2e4 f6e4 c1c7 e4f6 h1d1 c8e6 a2a3 a8b8 f2f3 f6d5 c7c1 b8c8 e2f2 c8c1 d1c1 e8c8 c1c8 e6c8 b2d4 b7b6 e5c6 a7a5 h2h4 h7h6 d4b2 c8a6 g2g4 d5f4 f2g3 f4d3 b2c3');
INSERT INTO evaluations VALUES(35390,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(35390,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(35390,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(35390,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(35390,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(35390,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(35390,36,'cp 5','f8b4 d8d2 b4d2 b1d2 b8a6 g1f3 h8e8 f3e5 f7g8 e1e2 a6c5 a1c1 c5e4 d2e4 f6e4 c1c7 c8e6 h1d1 e4f6 a2a3 a8b8 f2f3 f6d5 c7c1 b8c8 c1c8 e8c8 d1c1 d5f4 e2f2 f4d5 c1c8 e6c8 b2d4 b7b6 h2h4 d5e7 g2g4 e7g6 e5g6 h7g6 f2g3 c8b7 g3f4 b7c6 d4a1 c6b7 f4e3 a7a6');
INSERT INTO evaluations VALUES(35392,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(35392,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(35392,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(35392,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(35392,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(35392,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(35392,36,'cp 11','d8d2 b4d2 b1d2 b8a6 g1f3 h8e8 f3e5 f7g8 e1e2 a6c5 a1c1 c5e4 d2e4 f6e4 c1c7 c8e6 h1d1 a8b8 a2a3 e4f6 d1d6 f6e4 d6d4 e4f6 f2f3 f6d5 c7c1 b8c8 d4c4 b7b6 c4c8 e8c8 e2f2 a7a6 c1c8 e6c8 b2d4 c8f5 g2g3 b6b5 e5c6 f5d7 c6b4 d5b4 a3b4 g8f7 d4a1');
INSERT INTO evaluations VALUES(35385,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(35385,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(35385,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(35385,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35385,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(35385,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(35385,36,'cp 7','b4d2 b1d2 b8a6 g1f3 h8e8 f3e5 f7g8 e1e2 a6c5 a1c1 c5e4 d2e4 f6e4 c1c7 e4f6 h1d1 c8e6 a2a3 a8b8 f2f3 f6d5 c7c1 b8c8 h2h4 c8c1 d1c1 e8c8 c1c8 e6c8 g2g4 d5f4 e2f2 f4g6 e5g6 h7g6 b2d4 b7b6 d4a1 c8a6 f2e3 g8f7 e3e4 a6b7 e4f4 b7c6 h4h5');
INSERT INTO evaluations VALUES(35387,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(35387,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(35387,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35387,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(35387,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35387,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(35387,36,'cp 12','b1d2 b8a6 g1f3 h8e8 f3e5 f7g8 e1e2 a6c5 a1c1 c5e4 d2e4 f6e4 c1c7 e4f6 h1d1 c8e6 a2a3 a8b8 f2f3 f6d5 c7c5 b8c8 d1c1 d5f4 e2f2 c8d8 b2a1 e8e7 c5c7 e7c7 c1c7 d8d2 f2e3 f4d5 e3d2 d5c7 a1b2 c7d5 g2g3 d5e7 d2d3 e7c6 e5c6 b7c6 d3e3 a7a5 a3a4');
INSERT INTO evaluations VALUES(35386,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(35386,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(35386,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(35386,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(35386,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(35386,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(35386,36,'cp 9','b8a6 g1f3 h8e8 f3e5 f7g8 e1e2 a6c5 a1c1 c5e4 d2e4 f6e4 c1c7 e4f6 h1d1 c8e6 a2a3 a8b8 f2f3 f6d5 c7c1 b8c8 e2f2 c8c1 d1c1 e8c8 g2g4 c8c1 b2c1 a7a6 f2e2 d5b6 e2d3 b6d7 d3d4 d7e5 d4e5 e6d7 c1d2 g8f7 e5d6 d7b5 f3f4 h7h6 f4f5 b5e2');
INSERT INTO evaluations VALUES(2481,30,'mate 2',NULL);
INSERT INTO evaluations VALUES(2481,31,'mate 2',NULL);
INSERT INTO evaluations VALUES(2481,32,'mate 2',NULL);
INSERT INTO evaluations VALUES(2481,33,'mate 2',NULL);
INSERT INTO evaluations VALUES(2481,34,'mate 2',NULL);
INSERT INTO evaluations VALUES(2481,35,'mate 2',NULL);
INSERT INTO evaluations VALUES(2481,36,'mate 2','d8d5 f3d5 c8e6 d5e6');
INSERT INTO evaluations VALUES(2161,30,'mate 2',NULL);
INSERT INTO evaluations VALUES(2161,31,'mate 2',NULL);
INSERT INTO evaluations VALUES(2161,32,'mate 2',NULL);
INSERT INTO evaluations VALUES(2161,33,'mate 2',NULL);
INSERT INTO evaluations VALUES(2161,34,'mate 2',NULL);
INSERT INTO evaluations VALUES(2161,35,'mate 2',NULL);
INSERT INTO evaluations VALUES(2161,36,'mate 2','f3d5 c8e6 d5e6');
INSERT INTO evaluations VALUES(2160,30,'mate 1',NULL);
INSERT INTO evaluations VALUES(2160,31,'mate 1',NULL);
INSERT INTO evaluations VALUES(2160,32,'mate 1',NULL);
INSERT INTO evaluations VALUES(2160,33,'mate 1',NULL);
INSERT INTO evaluations VALUES(2160,34,'mate 1',NULL);
INSERT INTO evaluations VALUES(2160,35,'mate 1',NULL);
INSERT INTO evaluations VALUES(2160,36,'mate 1','c8e6 d5e6');
INSERT INTO evaluations VALUES(27701,30,'mate 1',NULL);
INSERT INTO evaluations VALUES(27701,31,'mate 1',NULL);
INSERT INTO evaluations VALUES(27701,32,'mate 1',NULL);
INSERT INTO evaluations VALUES(27701,33,'mate 1',NULL);
INSERT INTO evaluations VALUES(27701,34,'mate 1',NULL);
INSERT INTO evaluations VALUES(27701,35,'mate 1',NULL);
INSERT INTO evaluations VALUES(27701,36,'mate 1','d5e6');
INSERT INTO evaluations VALUES(12394,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12394,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(12394,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(12394,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12394,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12394,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12394,36,'cp 0','d7d5 c4d5 e6d5 e4d5 c6b4 d2d3 f8e7 g1e2 e8g8 d3d4 b4d5 c3d5 f6d5 e1g1 c8e6 d4c5 e7c5 e2d4 d8b6 d4e6 f7e6 d1e2 f8f2 f1f2 a8f8 c1f4 g7g5 g2d5 e6d5 g1g2 c5f2 f4g5 b6g6 a1f1 g6g5 f1f2 f8f2 g2f2 g8g7 e2e6 g5d2 f2f3 d2h2 e6e5 g7g6 e5e6');
INSERT INTO evaluations VALUES(24637,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(24637,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(24637,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(24637,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(24637,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(24637,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(24637,36,'cp -5','h2h3 g4f3 e2f3 e7e6 c3e2 f8d6 e1g1 e8g8 c2c3 d8c7 d1d3 a8b8 a2a4 h7h6 g2g3 c6a5 c1f4 a5c4 b2b3 d6f4 e2f4 c4a5 a1b1 b8c8 f1c1 a5c6 h3h4 f8d8 f4h5 f6h5 f3h5 g7g6 h5e2 h6h5 d3e3 c6e7 e2d3 e7f5');
INSERT INTO evaluations VALUES(25525,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(25525,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(25525,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(25525,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(25525,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(25525,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25525,36,'cp 25','h7h6 g5f4 f8e7 d4c6 b7c6 b5d3 e8g8 e1g1 d7e6 c3e2 f8e8 f1e1 a7a5 f4d2 d6d5 e4e5 f6e4 e2d4 e7c5 d2e3 c5d4 e3d4 c6c5 d4e3 c5c4 d3f1 d8e7 f2f3 e4c5 b2b3 c4b3 a2b3 a5a4 d1d2 a4b3 a1a8 e8a8 c2b3 c5d7');
INSERT INTO evaluations VALUES(26608,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(26608,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(26608,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(26608,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(26608,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(26608,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(26608,36,'cp -5','e7e5 d4e5 c6e5 f1e2 g4f3 g2f3 c7c6 f3f4 e5d7 e3e4 d5e4 c3e4 d7c5 d1d8 a8d8 e4c5 f8c5 c1e3 c5e3 f2e3 f7f5 a1d1 d8d1 e1d1 g8f6 e2f3 a7a5 d1e2 e8e7 h1g1 g7g6 h2h4 h8a8 g1d1 a8g8 a2a3 f6d7 d1d4 d7c5 b2b4 c5e6 d4c4');
INSERT INTO evaluations VALUES(26877,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26877,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26877,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26877,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26877,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26877,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(26877,36,'cp 0','h2h3 g4f3 d1f3 g8f6 f1b5 e5e4 f3e2 f8e7 b5c6 b7c6 e2a6 d8d6 b2b3 e8g8 c1d2 d6e6 c3e2 f8b8 d2a5 e6d6 e2c3 h7h5 c3a4 d6d7 e1c1 h5h4 c1b1 b8b5 a4c3 b5b8');
INSERT INTO evaluations VALUES(28894,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(28894,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(28894,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(28894,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28894,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(28894,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(28894,36,'cp 29','e1g1 b8d7 d3e2 h7h6 c2c3 f5h7 b1d2 a7a5 f3e1 c6c5 d2f3 h7e4 e2d3 c5d4 c3d4 d8b6 d1e2 e4d3 e1d3 e7f5 c1e3 f8e7 g2g4 f5e3 f2e3 b6d8 a1c1 e8g8 d3f4 d7b6 b2b3 a8c8 g1g2 c8c1 f1c1 d8d7 h2h4 f8c8');
INSERT INTO evaluations VALUES(28893,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(28893,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28893,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(28893,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28893,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(28893,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28893,36,'cp 0','b8d7 d3e2 f5g6 f3h4 c6c5 c2c3 d8b6 b1a3 a8c8 d4c5 d7c5 a3b5 e7c6 c1e3 b6d8 g2g3 a7a6 b5d4 c6e5 f2f4 e5c4 e2c4 d5c4 f4f5 e6f5 d1e2 f8e7 e3g5 e8f8 g5e7 d8e7 e2c4 e7e3 g1g2 e3e4 g2g1');
INSERT INTO evaluations VALUES(35152,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(35152,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(35152,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(35152,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(35152,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(35152,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(35152,36,'cp 109','b1c3 d7d6 c3d5 f6d8 d5f4 g7g5 f4d5 g5g4 f3d4 f8g7 c2c3 g8e7 f1c4 e7d5 c4d5 d8h4 g2g3 h4h3 d1e2 e8g8 d2d3 a7a5 c1f4 g7d4 c3d4 b8c6 d5c6 b7c6 e2d2 f7f6 e1c1 c8d7 c1b1 h3h5 d1c1 a5a4 d4d5 c6c5 d3d4 h5g6 d4c5 d6c5 b1a1');
INSERT INTO evaluations VALUES(41701,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41701,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41701,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(41701,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41701,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(41701,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(41701,36,'cp 19','c1g5 c8e6 e2e3 e8g8 f1d3 h7h6 g5h4 b8d7 e1g1 c7c6 a2a3 a7a5 d1c2 f8e8 a1c1 e6g4 f1e1 d8b8 h4g3 d6g3 h2g3 b8d6 f3h2 g4h5 d3f5 d7f8 c3a4 h5g6 f5g6 f8g6 a4c5 d6c7 h2f3 g6f8 f3e5 f8d7 c5d3 e8e7 f2f3 a8e8');
INSERT INTO evaluations VALUES(41680,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41680,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41680,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(41680,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(41680,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(41680,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(41680,36,'cp 14','c8e6 e2e3 h7h6 g5h4 b8d7 f1d3 c7c6 e1g1 e8g8 f3d2 f8e8 a1c1 e6g4 d1b3 d8b8 h4g3 d6g3 h2g3 a7a5 f1e1 g4h5 c3a4 b8c7 a2a3 e8e7 b3c2 c7d6 a4c5 d7c5 d4c5 d6e6 f2f3 h5g6 d3g6 f7g6 e3e4 e6e5 e4d5 e5d4 g1h1 e7e1 c1e1 d4d5');
INSERT INTO evaluations VALUES(42077,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(42077,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(42077,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(42077,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(42077,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(42077,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(42077,36,'cp 58','b2c3 e8g8 f1d3 d5c4 d3c4 b7b6 c1a3 f8e8 e1g1 c8b7 d1e2 b8d7 f1d1 a8c8 c4d3 f6e4 a3b2 c7c5 c3c4 c5d4 e3d4 e4f6 f3e5 d7e5 e2e5 d8c7 e5e3 e8d8 a1c1 d8d7 d3f1 d7d8 b2a3 d8e8');
INSERT INTO evaluations VALUES(43884,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(43884,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(43884,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(43884,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(43884,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(43884,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(43884,36,'cp 166','e4e5 f6g8 d2d4 e7e6 f1d3 c8a6 d3a6 b8a6 d1e2 d8c8 a2a3 d7d5 e5d6 c7d6 d4d5 a6c7 d5e6 f7e6 c1f4 c8a6 e2f3 e8c8 e1c1 g8f6 g1e2 a6b7 f3g3 d8d7 c1b1 f8e7 f4d6 e7d6 d1d6 d7d6 g3d6 b7g2 h1g1 h8d8 g1g2 d8d6 g2g7');
INSERT INTO evaluations VALUES(44315,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(44315,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(44315,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(44315,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(44315,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(44315,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(44315,36,'cp 124','g1f3 f8d6 b1c3 b8c6 f3h4 c6d4 c1g5 d6e7 h4f5 d4f5 e4f5 a8b8 g5f6 e7f6 d1f3 d7d6 c4d5 e8g8 h2h4 h7h6 f3h3 b7b5 g2g4 e5e4 d3e4 b5b4 c3d1 c8b7 d5b7 b8b7 g4g5 f6d4 d1e3 f8e8 e1c1 e8e4 g5h6 d8f6 h6g7 d4b2 c1b1');
INSERT INTO evaluations VALUES(44313,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(44313,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(44313,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(44313,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(44313,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(44313,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(44313,36,'cp 114','f8d6 b1c3 b8c6 f3h4 d6f8 c1g5 d7d6 h4f5 h7h6 g5f6 d8f6 c3d5 f6d8 f5e3 a8b8 e1g1 f8e7 g2g3 b7b5 c4b3 c8h3 f1e1 h6h5 c2c3 g7g6 a2a4 b5b4 b3c4 b4c3 b2c3 e8f8 c4b5 c6a5 f2f4 e5f4 d5f4 h3d7 b5d7 d8d7 f4d5 f8g7 e1f1 a5c6');
INSERT INTO evaluations VALUES(52626,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(52626,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(52626,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52626,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52626,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52626,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52626,36,'cp 0','f8g7 f1b5 c8d7 e1g1 d7b5 c3b5 b8c6 d2d3 d8d7 c2c4 g8h6 h2h3 f7f5 a1b1 e8g8 c1e3 e7e6 g1h1 a7a6 b5c3 c6d4 a2a4 a8e8 d1d2 d4b3 d2f2 b7b6 f2c2 b3d4 c2d2 d4b3');
INSERT INTO evaluations VALUES(56704,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(56704,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(56704,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(56704,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(56704,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(56704,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(56704,36,'cp 195','d2d4 d6d5 d1e2 f8b4 c2c3 b4e7 d4e5 b8c6 b1d2 f6e5 f3e5 d8d6 e5c6 d6c6 e2h5 g7g6 e4d5 c6d6 h5e5 g8f6 e5d6 e7d6 d2c4 e8d8 c4d6 c7d6 f1b5 c8d7 b5e2 h8e8 e1f2 f6g4 e2g4 d7g4 f2g3 g4f5');
INSERT INTO evaluations VALUES(3058,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(3058,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(3058,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(3058,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3058,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3058,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3058,36,'cp 0','f6e4 c3e4 d7d5 d4c6 b7c6 c4d3 d5e4 d3e4 d8d1 f1d1 c8e6 c2c3 f8c8 a2a4 a7a5 d1d2 g7f6 g2g3 a8b8 e3f4 b8b7 f4h6 f6g7 h6f4');
INSERT INTO evaluations VALUES(3085,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3085,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3085,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3085,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3085,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3085,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3085,36,'cp 5','c4f7 f8f7 c3e4 g7d4 e3d4 d7d5 e4g3 h7h5 h2h4 e7e5 d4e3 d8h4 d1d5 h4g4 d5d1 h5h4 d1g4 c8g4 g3e4 c6b4 e3g5 b4c2 a1c1 c2d4 e4f6 f7f6 g5f6 d4e2 g1h2 e2c1 f1c1 a8c8 c1c8 g4c8 f6e5 g8f7 f2f3 c8e6 b2b3 e6d7 e5d4 b7b6 d4f2 g6g5 f2e3 f7f6 e3d4 f6f5 d4c3 d7e6');
INSERT INTO evaluations VALUES(4805,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(4805,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4805,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4805,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(4805,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(4805,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(4805,36,'cp -12','g8e7 d2d4 f7f6 c1e3 e8g8 b1d2 e5d4 f3d4 c6c5 d4e2 c8e6 f2f4 d8e8 e2g3 a8d8 d1e1 b7b6 b2b3 e7c6 c2c3 e8d7 e1e2 d6e7 a1c1 f8e8 f1d1 d7c8 d2c4 d8d1 c1d1 e7f8 e3f2 e8d8 d1d8 c6d8 c4e3 b6b5 a2a4 c8d7 a4b5 a6b5');
INSERT INTO evaluations VALUES(5797,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(5797,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(5797,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(5797,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(5797,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(5797,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(5797,36,'cp 67','b4d6 b1d2 b7b6 f1d3 g8f6 f3e5 d6e5 d4e5 f6d7 d3b5 c6e7 d1g4 e8g8 h2h4 e7g6 f4g5 g6e5 g5d8 e5g4 d8c7 a7a6 b5e2 h7h5 c3c4 g4f6 c7g3 a6a5 c4d5 e6d5 a1c1 c8a6 e2a6 a8a6 e1e2 f8e8 h1d1 d7c5 d2f3');
INSERT INTO evaluations VALUES(5918,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(5918,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(5918,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(5918,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(5918,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(5918,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(5918,36,'cp 66','f1d3 g8f6 b1d2 b7b6 f3e5 d6e5 d4e5 f6d7 d3b5 c6e7 d1g4 e8g8 h2h4 e7g6 f4g5 g6e5 g5d8 e5g4 d8c7 a7a6 b5e2 f7f5 c3c4 g4f6 c7d6 f8d8 a1c1 c8b7 f2f3 a6a5 e1f2 b7a6 d6e7 d8e8 e7f6 d7f6 c4d5 a6e2 f2e2 e6d5 e2f2 a8c8 d2b3 f6d7 c1c8 e8c8 b3d4 d7f6');
INSERT INTO evaluations VALUES(6924,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(6924,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(6924,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(6924,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(6924,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(6924,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(6924,36,'cp 208','d2d4 c5b6 e1g1 d7d6 a2a4 a7a6 b2b4 c8g4 h2h3 g4f3 d1f3 e5d4 f3h5 g7g6 h5d1 g8e7 c3d4 f6f5 b1c3 b6d4 c1h6 f5e4 d1b3 d6d5 c3d5 d8d6 a1e1 e8c8 h6f4 d4e5 d5e7 c6e7 f4g5 e5f4 e1d1 f4g5 d1d6 d8d6 b4b5 a6a5 f1e1 e7f5 e1e4 h8d8 g2g3 f5d4');
INSERT INTO evaluations VALUES(6929,30,'cp 233',NULL);
INSERT INTO evaluations VALUES(6929,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(6929,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(6929,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(6929,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(6929,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(6929,36,'cp 206','c5b6 a2a4 a7a5 e1g1 d7d6 d1b3 e8f8 c1e3 g7g6 b1a3 f8g7 a3b5 g8h6 d4e5 b6e3 e5f6 g7f6 f2e3 f6g7 h2h3 h8f8 f1f2 c6e5 a1f1 d8e7 b5d4 e5f3 d4f3 e7e4 c4d5 e4d3 f1d1 d3f5 e3e4 f5f4 d1d3 g7h8 d5b7 c8b7 b3b7 f4c1 g1h2');
INSERT INTO evaluations VALUES(7021,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(7021,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(7021,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(7021,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(7021,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(7021,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(7021,36,'cp 51','d2d4 d7d6 e1g1 g8f6 d4d5 c6b8 b1c3 e8g8 a2a4 e7e6 f1e1 e6d5 c3d5 c8g4 h2h3 g4f3 d1f3 b8d7 c1g5 c7c6 d5f6 g7f6 g5f6 d8f6 f3f6 d7f6 a1d1 f8d8 f2f3 d8d7 g1f2 a8d8 h3h4 h7h5 g2g3 d7e7 c2c3 a7a5 b2b4');
INSERT INTO evaluations VALUES(7022,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(7022,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(7022,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(7022,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(7022,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(7022,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(7022,36,'cp 43','d7d6 b1c3 g8f6 d4d5 c6b8 a2a4 e8g8 e1g1 e7e6 f1e1 e6d5 c3d5 c8g4 h2h3 g4f3 d1f3 b8d7 c2c3 d7e5 f3e2 e5c4 e2c4 f6d5 c4d5 c7c6 d5b3 d8c7 b3c2 a7a5 c1f4 g7e5 f4e3 f8e8 a1d1 a8d8');
INSERT INTO evaluations VALUES(5691,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(5691,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(5691,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(5691,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(5691,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(5691,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(5691,36,'cp 59','b1c3 g8f6 d4d5 c6b8 h2h3 e8g8 d1e2 e7e6 d5e6 c8e6 c4e6 f7e6 e4e5 d6e5 e2e5 d8d6 e5e2 f6d5 c3e4 d6a6 e2a6 b8a6 c2c3 b7b5 a2a3 c7c5 c1e3 c5c4 e1c1 a6c7 e3c5 f8d8 c5d6 a7a5 d6e5 g7e5 f3e5 d5f4 d1d8 a8d8 e5c6 d8d7 c1c2 c7d5 c6a5');
INSERT INTO evaluations VALUES(5690,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(5690,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(5690,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(5690,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(5690,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(5690,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(5690,36,'cp 44','g8f6 d4d5 c6b8 h2h3 e8g8 a2a4 e7e6 e1g1 a7a5 c1g5 h7h6 g5e3 e6e5 d1d2 g8h7 a1e1 b8a6 f3h2 f6d7 c4a6 a8a6 f2f4 e5f4 e3f4 d7e5 b2b3 a6a8 h2f3 c8d7 f3e5 d6e5 f4e3 b7b6 f1f3 d8e7 c3b5 d7b5 a4b5');
INSERT INTO evaluations VALUES(8738,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(8738,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(8738,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(8738,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(8738,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(8738,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(8738,36,'cp 58','d4d5 c6b8 h2h3 e8g8 d1e2 f8e8 e1g1 e7e6 f3g5 e6e5 g5f3 e8f8 c1g5 h7h6 g5e3 f6h5 e2d2 g8h7 c4e2 b8d7 f3h2 h5f4 e3f4 e5f4 d2f4 d7c5 a1e1 d8g5 f4g5 h6g5 c3b5 c8d7 b5c7 a8c8 c7b5 d7b5 e2b5 g7b2 h2f3 a7a6 f3g5 h7g8');
INSERT INTO evaluations VALUES(8820,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(8820,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8820,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8820,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(8820,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(8820,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(8820,36,'cp 25','c2c3 a7a5 e1g1 g7g5 h4g3 e8g8 b1d2 c5a7 f1e1 g5g4 f3h4 f6h5 d2f1 d8g5 f1e3 a7e3 e1e3 h5f4 d3d4 c6e7 d1e1 c8e6 c4e6 f7e6 e1d2 h6h5 d4e5 d6e5 e3d3 e7c6 a1d1 f8f7 d3e3 b7b5 e3e1 a5a4 f2f3 a4a3 f3g4 h5g4');
INSERT INTO evaluations VALUES(8817,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(8817,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(8817,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(8817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8817,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8817,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(8817,36,'cp 24','a7a5 b1d2 c5a7 e1g1 e8g8 f1e1 c8e6 a2a3 a8b8 b2b4 g7g5 h4g3 b7b5 c4e6 f7e6 d3d4 g5g4 f3h4 e5d4 c3d4 a7d4 a1c1 c6e5 d2f3 d4b2 f3e5 d6e5 h4g6 d8d1 c1d1 f8d8 g3e5 d8d1 e1d1 b2e5 g6e5 a5b4 a3b4 f6e4 e5g4 h6h5 g4e5 b8f8 d1f1 e4d6 f1e1 d6c4');
INSERT INTO evaluations VALUES(9845,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(9845,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(9845,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(9845,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(9845,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(9845,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(9845,36,'cp -14','b4c3 b2c3 d7d5 e4d5 f6d5 e1g1 c8g4 c4b5 d8f6 b5c6 b7c6 f1e1 e8c8 c1d2 h8e8 h2h3 g4f3 d1f3 f6f3 g2f3 g7g5 c3c4 d5f4 d2f4 e5f4 e1e4 f7f5 e4e1 c8d7 a1b1 d8b8 b1b8 e8b8 e1e5 b8b2 e5f5 b2c2 h3h4 d7e6 f5a5 c2d2 a5a7 d2d3 a7c7');
INSERT INTO evaluations VALUES(9864,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(9864,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(9864,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(9864,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(9864,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9864,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9864,36,'cp -20','b2c3 d7d5 e4d5 f6d5 e1g1 e8g8 c1d2 d8d6 f1e1 d5f6 d1c1 c8g4 f3h4 g4e6 c4b5 e6d7 b5c6 b7c6 f2f4 a8e8 f4e5 e8e5 e1e5 d6e5 h4f3 e5d6 d2f4 d6c5 f4e3 c5f5 e3a7 c6c5 c3c4 d7c6 c1f1 f8e8 a1e1');
INSERT INTO evaluations VALUES(9862,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9862,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(9862,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(9862,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(9862,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(9862,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(9862,36,'cp -17','d7d5 e4d5 f6d5 e1g1 e8g8 c1d2 d5f6 f1e1 f8e8 a1b1 a7a6 h2h3 d8d6 d1c1 b7b5 c4b3 f6d7 c3c4 c8b7 c1d1 d7f6 c4b5 a6b5 c2c4 b5c4 b3c4 b7a6 d2b4 d6d7 b1c1 a6c4 c1c4 f6d5 c4c5 d7d6 d1b3 c6b4 a3b4 c7c6 c5a5 d6b4 b3b4 d5b4 a5a8 e8a8 f3e5');
INSERT INTO evaluations VALUES(10594,30,'cp 436',NULL);
INSERT INTO evaluations VALUES(10594,31,'cp 505',NULL);
INSERT INTO evaluations VALUES(10594,32,'cp 466',NULL);
INSERT INTO evaluations VALUES(10594,33,'cp 473',NULL);
INSERT INTO evaluations VALUES(10594,34,'cp 523',NULL);
INSERT INTO evaluations VALUES(10594,35,'cp 501',NULL);
INSERT INTO evaluations VALUES(10594,36,'cp 510','d1e2 d7d5 e5d6 c8f5 b1d2 e8g8 d2e4 c7d6 e4g3 f5g6 e2d1 c6e5 c4d3 d8d7 a2a4 a7a6 f3h4 a8e8 g1h1 c5b6 h4g6 h7g6 c1f4 b6c7 b2b4 e5d3 c2d3 c7b8 g3e2 d7g4 f2f3 g4h4 d1d2 b8c7 f4g3 h4f6 e2g1');
INSERT INTO evaluations VALUES(21170,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(21170,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(21170,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(21170,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(21170,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(21170,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21170,36,'cp 52','d2d4 e7e6 b1c3 g8f6 g1f3 d6d5 c4d3 c6b4 e4e5 b4d3 d1d3 f6g8 e1g1 a7a5 c3e2 b7b6 d3d1 g8e7 e2g3 h7h6 f1e1 a5a4 b2b3 e7c6 c1d2 c8a6 a1c1 f8a3 c1b1 a3f8 g3h5 d8d7 b1c1 e8d8 c2c4 f8a3 c4d5 d7d5');
INSERT INTO evaluations VALUES(21497,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(21497,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(21497,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(21497,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(21497,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(21497,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(21497,36,'cp 96','a7a6 f1c4 b7b5 c4d3 c6b4 d3e2 g8f6 e3e4 c7c5 e1g1 c5d4 f3d4 c8b7 a2a4 f8d6 a4b5 e8g8 b5a6 b4a6 d4b5 d6e7 d1d8 f8d8 f2f3 e7c5 g1h1 a6b4 a1a8 d8a8 f1d1 a8a1 c1e3 a1d1 c3d1 f6d7 h1g1 c5e3 d1e3 b7a6 g1f1 d7c5');
INSERT INTO evaluations VALUES(23145,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(23145,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(23145,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(23145,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(23145,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(23145,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(23145,36,'cp 95','d8f6 d4d3 f8b4 c1d2 g8e7 e2e4 e8g8 f1e2 c7c6 a2a3 b4a5 e1c1 a5c7 d2e3 d7d5 e4d5 c8f5 d3d2 e7d5 e3d4 c7e5 c3d5 c6d5 g2g4 f5d7 g4g5 f6f4 d4e5 f4e5 f2f4 e5c7 e2d3 d7g4 d1f1 g7g6 f4f5 a8e8 f5g6 f7g6');
INSERT INTO evaluations VALUES(29944,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(29944,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29944,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(29944,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29944,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(29944,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29944,36,'cp 0','c7c5 e1g1 f8e7 f3d2 g4e6 d2c4 e8g8 f2f4 e5f4 c1f4 d6d5 e4d5 f6d5 c3d5 e6d5 c2c3 b8c6 e2h5 g7g6 h5g4 e7g5 f4g3 g5h4 g3d6 h4e7');
INSERT INTO evaluations VALUES(33296,30,'cp 562',NULL);
INSERT INTO evaluations VALUES(33296,31,'cp 557',NULL);
INSERT INTO evaluations VALUES(33296,32,'cp 557',NULL);
INSERT INTO evaluations VALUES(33296,33,'cp 555',NULL);
INSERT INTO evaluations VALUES(33296,34,'cp 559',NULL);
INSERT INTO evaluations VALUES(33296,35,'cp 557',NULL);
INSERT INTO evaluations VALUES(33296,36,'cp 551','h4e4 e5f3 e4e7 b1c3 g8f6 d1e2 e7e2 c3e2 d7d6 h2h3 e8g8 d2d3 c7c5 f1g2 b8c6 h1f1 h7h6 c1d2 c8d7 a1e1 c6b4 d2b4 c5b4 f3d2 f8e8 e2d4 h6h5 f2g1 e8e1 f1e1 b4b3 a2b3 d6d5 d2f3 a8e8 e1e8 d7e8 g1f2 g8f8 b3b4 f6d7 c2c3 d7b6 f3d2');
INSERT INTO evaluations VALUES(33299,30,'cp 560',NULL);
INSERT INTO evaluations VALUES(33299,31,'cp 571',NULL);
INSERT INTO evaluations VALUES(33299,32,'cp 566',NULL);
INSERT INTO evaluations VALUES(33299,33,'cp 575',NULL);
INSERT INTO evaluations VALUES(33299,34,'cp 578',NULL);
INSERT INTO evaluations VALUES(33299,35,'cp 559',NULL);
INSERT INTO evaluations VALUES(33299,36,'cp 553','e5f3 e4e7 b1c3 g8f6 d1e2 e7e2 c3e2 d7d6 h2h3 b7b6 d2d3 e8g8 f1g2 c8b7 g3g4 f6d7 c1d2 f7f5 e2f4 d7e5 f3e5 b7g2 f2g2 d6e5 f4e6 f8f7 h1f1 b8c6 f1f5 f7f5 g4f5 c6d4 a1f1 d4c2 f1c1 c2d4 e6d4 e5d4 c1c7 a8f8');
INSERT INTO evaluations VALUES(33306,30,'cp 542',NULL);
INSERT INTO evaluations VALUES(33306,31,'cp 532',NULL);
INSERT INTO evaluations VALUES(33306,32,'cp 521',NULL);
INSERT INTO evaluations VALUES(33306,33,'cp 557',NULL);
INSERT INTO evaluations VALUES(33306,34,'cp 559',NULL);
INSERT INTO evaluations VALUES(33306,35,'cp 561',NULL);
INSERT INTO evaluations VALUES(33306,36,'cp 540','e4e7 b1c3 g8f6 d1e2 e7e2 f1e2 e8g8 d2d3 d7d6 c1g5 b8d7 f3d2 h7h6 g5f6 d7f6 e2f3 c7c6 h1e1 d6d5 c3d1 h6h5 f2g1 c8g4 d1f2 f8e8 g1f1 e8e1 a1e1 g4f3 d2f3 h5h4 g3g4 f6h7 c2c3 f7f5 g4f5');
INSERT INTO evaluations VALUES(34338,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(34338,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(34338,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(34338,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(34338,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(34338,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(34338,36,'cp 90','c1e3 g8f6 f1d3 c8g4 f2f3 g4h5 d1d2 d6c7 g1e2 h5g6 e1c1 g6d3 d2d3 e7e6 c1b1 b8d7 e3c1 f8e7 h2h4 b7b5 g2g4 d7b6 c1f4 e7d6 f4d6 c7d6 c3e4 d6e7 g4g5 f6d5 f3f4 e8g8 h4h5 a7a5 h5h6 g7g6');
INSERT INTO evaluations VALUES(34386,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(34386,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(34386,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(34386,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(34386,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(34386,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(34386,36,'cp 86','d5c4 e2e4 c8e6 g1e2 b8c6 e2f4 a8d8 c1d2 a5c5 f4e6 f7e6 d1a4 g8f6 d2e3 c5b4 a4b4 c6b4 f1c4 b4c2 e1e2 c2a1 h1a1 e8f7 f2f4 g7g6 f4f5 g6f5 e4f5 f8g7 c4e6 f7f8 e3a7 g7h6 a1d1 d8d1 e2d1 h6f4 h2h3 f8g7 a7b6 f4e5 d1c2 f6h5 c3d5 h8a8 d5c7 a8b8 b2b4 h5f4 a2a4 f4e6');
INSERT INTO evaluations VALUES(34385,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(34385,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(34385,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(34385,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(34385,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(34385,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(34385,36,'cp 124','e2e4 c8e6 g1f3 b8c6 f3g5 a8d8 c1d2 a5c5 g5e6 f7e6 d1a4 g8f6 d2e3 c5b4 a4b4 c6b4 f1c4 b4c2 e1e2 c2a1 h1a1 e8f7 f2f4 g7g6 f4f5 g6f5 e4f5 f8g7 c4e6 f7f8 e3a7 g7h6 a1d1 d8d1 e2d1 h6f4 h2h3 f4e5 d1c2 f8g7 a7b6 h8a8 g2g4 g7f8 b6e3 e5c3 c2c3 a8d8 c3c2 f6d5 e3h6 f8e8');
INSERT INTO evaluations VALUES(36855,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(36855,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(36855,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(36855,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(36855,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(36855,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(36855,36,'cp 39','d2d4 b8c6 d4d5 c6b8 e2e4 a7a5 a2a3 b8d7 c1g5 h7h6 g5e3 f6g4 e3d2 d7c5 b2b4 a5b4 a3b4 a8a1 d1a1 c5b3 a1a2 b3d2 a2d2 d8e7 h2h3 g4f6 f1a1 c7c6 d2e3 c6d5 c3d5 f6d5 c4d5 h6h5 a1c1 f7f5 e4f5 c8f5 f3d2 e7d7 d2e4 f5h3 g2h3 d7h3 c1c7');
INSERT INTO evaluations VALUES(36858,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(36858,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(36858,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(36858,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(36858,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(36858,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(36858,36,'cp 28','b8c6 d4d5 c6b8 e2e4 a7a5 a2a3 b8d7 c1g5 h7h6 g5e3 f6g4 e3d2 d7c5 b2b4 a5b4 a3b4 a8a1 d1a1 c5b3 a1b2 b3d2 b2d2 d8e7 h2h3 g4f6 f1a1 c7c6 d5c6 b7c6 b4b5 c6b5 c4b5 c8e6 d2e3 f8b8 a1a6 e7d8 f3d2 d6d5 e4d5 f6d5 g2d5 e6d5 d2e4 d5e4 e3e4 d8d4 a6a8 b8a8 e4a8 g8h7 a8c6 f7f5');
INSERT INTO evaluations VALUES(42070,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(42070,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(42070,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(42070,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(42070,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(42070,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(42070,36,'cp 46','b7b6 g1f3 e8g8 f1e2 d5c4 e2c4 d8e7 a1c1 b8d7 b2b3 c8b7 e1g1 c7c5 d1e2 f8d8 f1d1 h7h6 d4c5 d7c5 c3d4 a8c8 f3e5 c5d7 h2h3 f6e8 e2b2 e8d6 c4f1 d6f5 d4c3 d7e5 c3e5 f7f6 e5c3 d8d7 a2a4 d7c7 c3e1 c7c1');
INSERT INTO evaluations VALUES(42452,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(42452,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(42452,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(42452,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(42452,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42452,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(42452,36,'cp 32','c2c3 b4e7 b1d2 c7c5 g1f3 f6h5 f4b8 a8b8 d4c5 e7c5 d2e4 d8c7 f3d4 h5f6 d4b5 c7e5 e4c5 e5c5 d1d4 c5d4 c3d4 a7a6 b5c3 d7d5 f2f3 e8e7 h2h4 c8d7 e1d2 h7h5 f1d3 f6e8 h1c1 e8d6 c3e2 h8c8 c1c5 d7b5 a1c1 b5d3 d2d3 c8c5 c1c5');
INSERT INTO evaluations VALUES(42450,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(42450,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(42450,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(42450,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(42450,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42450,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(42450,36,'cp 31','b4e7 f1d3 c7c5 b1d2 d7d5 g1f3 f6h5 d4c5 h5f4 e3f4 e7c5 e1g1 e8g8 g2g3 a7a5 d2b3 c5e7 f1e1 b8c6 a2a4 c8d7 h2h4 h7h5 d1e2 g7g6 d3b5 a8c8 e2e3 e7f6 b5c6 c8c6 b3d4 c6c4 b2b3 c4c8 g1g2 d8c7');
INSERT INTO evaluations VALUES(42995,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(42995,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(42995,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(42995,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(42995,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(42995,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(42995,36,'cp 35','b1d2 c7c5 f1d3 d7d5 g1f3 f6h5 f4e5 b8d7 h2h4 d8b6 a1b1 d7e5 d4e5 g7g6 g2g4 h5g7 d1a4 c8d7 a4f4 h7h6 e1e2 b6c7 b2b3 d7c6 b3b4 h6h5 b4b5 c6d7 f3g5 e7g5 h4g5 e8c8 f4f6 c5c4 f6g7 c4d3 e2d3');
INSERT INTO evaluations VALUES(44872,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(44872,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44872,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44872,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44872,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(44872,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(44872,36,'cp -12','c6d5 c3e2 c8f5 c2c3 d8c7 d2d4 e7e6 c1f4 f8d6 f4d6 c7d6 e2g3 e8g8 g3f5 e6f5 f1d3 f8e8 e1f1 f6e4 g2g3 b8c6 f1g2 h7h6 h1e1 g7g5 f3d2 e8e7 a2a4 a8e8 f2f3 e4f6 e1e7 c6e7 d2f1 f5f4 d1c2 g5g4');
INSERT INTO evaluations VALUES(45921,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(45921,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(45921,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(45921,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(45921,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(45921,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(45921,36,'cp 103','e4e5 d8e7 f1b5 b8c6 e1g1 d6e5 f1e1 c8g4 h2h3 g4f3 d1f3 e7e6 f3b3 e6b3 b5c6 b7c6 a2b3 f6d7 a1a6 f7f6 c1e3 d7b8 a6a4 f8e7 b1d2 a7a6 e1a1 a8a7 d2e4 f6f5 e3c5 a7d7 a4b4 e8f7 c5e7 f7e7 e4c5 d7d2 b4b7 e7f6 a1a4 h8d8 a4c4 h7h5 f2f3');
INSERT INTO evaluations VALUES(48800,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(48800,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(48800,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(48800,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48800,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(48800,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(48800,36,'cp -20','e5d4 e3d4 b8c6 g1f3 f8b4 f1d3 c8g4 e1g1 e8g8 c1g5 d5c4 d3c4 h7h6 g5h4 b4a5 a1c1 f8e8 h2h3 g4f3 d1f3 d8d4 h4f6 d4f6 f3f6 g7f6 c4b5 a5c3 c1c3 c6d4 b5e8 d4e2 g1h2 e2c3 e8c6 c3a2 c6b7 a8b8 f1a1 b8b7 a1a2 c7c5 b2b3 b7b3 a2a7 g8f8 a7c7');
INSERT INTO evaluations VALUES(48640,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(48640,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(48640,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(48640,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48640,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48640,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(48640,36,'cp -12','e3d4 f8b4 g1f3 e8g8 f1e2 d5c4 e2c4 b8c6 e1g1 c8g4 c1g5 h7h6 g5h4 b4a5 a1c1 f8e8 h2h3 g4f3 d1f3 d8d4 h4f6 d4f6 f3f6 g7f6 c4b5 a8d8 b5c6 b7c6 f1d1 d8d1 c1d1 g8g7 g1f1 f6f5 c3a4 e8e6 a4c5 e6d6 d1c1 a5b6 f1e2 g7f6 b2b4');
INSERT INTO evaluations VALUES(49212,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(49212,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(49212,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(49212,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(49212,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(49212,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(49212,36,'cp 193','f3e5 f8g7 d2d4 e8g8 f1d3 d7d6 e5f3 c7c5 d4c5 d6c5 e1g1 b8c6 h2h3 c8e6 c1e3 d8e7 d1c1 f6d7 e3g5 f7f6 g5e3 d7e5 f3e5 c6e5 d3e2 f6f5 f1e1 g8h8 e4f5 g6f5 e2h5 e6f7 h5f7 f8f7 g1h1 e5g6 e3g5 e7d7');
INSERT INTO evaluations VALUES(49200,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(49200,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(49200,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(49200,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(49200,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(49200,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(49200,36,'cp 183','f8g7 d2d4 e8g8 f1d3 d7d5 c1g5 d8d6 g5f6 d6f6 c3d5 f6d8 e5f3 c7c6 d5e3 g7d4 f3d4 d8d4 c2c3 d4c5 e1g1 f8e8 d3c2 c8e6 d1d4 b8d7 d4c5 d7c5 f1d1 a7a5 f2f4 f7f6 g1f2 e6f7 e3g4 g8g7 e4e5 f6e5 g4e5 f7e6 d1d4 a8d8 a1d1 d8d4 c3d4 c5d7 d4d5 c6d5 e5d7 e6d7 d1d5 d7g4');
INSERT INTO evaluations VALUES(52069,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(52069,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52069,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52069,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(52069,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(52069,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(52069,36,'cp 2','g1f3 b8c6 d4d1 g8f6 f1d3 g7g6 b1c3 f8g7 e1g1 e8g8 h2h3 d7d5 e4d5 f6d5 c3d5 e6d5 f1e1 c8e6 a2a3 f8e8 c2c3 d5d4 c3c4 d8d7 c1f4 e6f5 d1d2 a8c8 a1d1 f5d3 d2d3 h7h6 f4d2 b7b6 b2b3 c8d8 b3b4 g6g5 b4b5');
INSERT INTO evaluations VALUES(53732,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(53732,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(53732,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53732,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(53732,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(53732,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53732,36,'cp 28','d2d4 f8d6 f1d3 g8e7 c3e2 c8f5 e1g1 e8g8 f1e1 f5d3 d1d3 b8d7 c1f4 d6c7 h2h4 d7f6 c2c3 f6e4 g2g3 f8e8 f4c7 d8c7 e2f4 e7g6 f4g6 h7g6 f3e5 e8e7 e1e2 a8e8 a1e1 e4f6 g1g2 f6d7 e5g4 e7e2 e1e2 e8e2 d3e2');
INSERT INTO evaluations VALUES(54028,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(54028,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(54028,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(54028,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(54028,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(54028,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(54028,36,'cp 134','c3d4 f8e7 f1d3 a7a6 b1c3 g8f6 e1g1 b8c6 h2h3 e8g8 d1e2 c8d7 c1f4 f8e8 f1e1 c6d4 f3d4 e6e5 f4e3 e5d4 e3d4 d7c6 a1d1 f6d7 d3b1 e7f6 e2e3 f6h4 a2a3 d8g5 e3g5 h4g5 d4e3 g5e7 b1a2 d7c5 c3d5 c6d5 a2d5 e7f6 e3c5 d6c5 d5b7');
INSERT INTO evaluations VALUES(20367,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(20367,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(20367,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20367,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20367,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20367,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20367,36,'cp 0','e6d5 b1c3 f8b4 h2h3 g8e7 f1e2 e8g8 e1g1 c8f5 f1e1 a8c8 c1f4 e7g6 f4g5 d8d7 a1c1 h7h6 g5d2 b4e7 e2f1 f8d8 d2e3 c6a5 f1b5 a5c6');
INSERT INTO evaluations VALUES(20552,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20552,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20552,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(20552,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20552,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20552,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20552,36,'cp 0','b1c3 f8b4 f1d3 g8e7 e1g1 e8g8 f1e1 c8f5 d3f1 f5g4 f1e2 a8c8 h2h3 g4f5 e2f1 f5e4 f3g5 e4f5');
INSERT INTO evaluations VALUES(20551,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20551,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20551,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(20551,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(20551,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(20551,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(20551,36,'cp 5','f8b4 f1d3 g8e7 e1g1 e8g8 c1g5 h7h6 g5h4 d8a5 a1c1 c8g4 h2h3 g4f3 d1f3 c6d4 f3h5 a8e8 h4e7 b4e7 h5d5 a5d5 c3d5 e7g5 c1d1 d4e2 g1h1 e2f4 d5f4 g5f4 b2b3 f4e5 d3c4 e8d8 c4d5 d8d7 f1e1 e5f6 d5e4 d7d1 e1d1 f8d8');
INSERT INTO evaluations VALUES(21606,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21606,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(21606,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(21606,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21606,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(21606,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(21606,36,'cp 17','f8d6 b1c3 g8f6 c1g5 d6e7 f3e5 e8g8 e1g1 c8e6 e5c6 b7c6 c3a4 f8e8 f1e1 e6f5 e2d3 f5d3 d1d3 d8d6 g5d2 f6e4 e1e2 e4d2 d3d2 d6b4 d2b4 e7b4 g1f1 e8e2 f1e2 g7g6 e2f1 b4d6 g2g3 h7h5 a1e1');
INSERT INTO evaluations VALUES(5905,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5905,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(5905,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(5905,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(5905,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(5905,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(5905,36,'cp 2','b1c3 c6e7 c3b5 g8f6 e1g1 e8g8 a2a4 f8e8 b5d6 d8d6 a1a3 e7g6 e2d3 c7c6 a4a5 f6e4 f1e1 c8f5 f3e5 d6f6 d3e4 d5e4 e5g6 f5g6 c1e3 a8d8 a3b3 d8d7 d1d2 f6d6 d2b4 d6b4 b3b4 f7f6 e1a1 g6f7 c2c4 h7h6 b2b3 g7g5 a5a6');
INSERT INTO evaluations VALUES(24882,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(24882,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(24882,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(24882,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(24882,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(24882,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(24882,36,'cp 41','c2c4 e7e6 c4c5 e6e5 d4e5 f6e4 b1c3 e4c3 b2c3 f8c5 f1d3 c5b6 e1g1 d8d7 a2a4 c6a5 f4g3 e8c8 d3e2 d7e6 g1h2 h5f3 g2f3 c8b8 f3f4 f7f6 d1c2 g7g6 a1d1 f6e5 f4e5 h7h5 h3h4 a5c6 c3c4 c6e5 c4d5 d8d5');
INSERT INTO evaluations VALUES(26051,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(26051,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(26051,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(26051,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(26051,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(26051,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(26051,36,'cp -13','g8f6 e1g1 e7e6 d3d4 c6d4 b5d7 d8d7 f3d4 c5d4 d1d4 a7a6 c1g5 f8e7 f1d1 a8c8 g5f6 e7f6 d4d6 d7d6 d1d6 f6c3 b2c3 c8c3 d6b6 c3c2 b6b7 e8g8 b7a7 f8d8 a7a6 g7g5 g2g3 d8d2 a1f1 g8g7 a6a5 h7h6 a2a4 c2a2 a5a8 g5g4 a4a5 d2c2 a5a6');
INSERT INTO evaluations VALUES(24478,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(24478,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(24478,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(24478,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24478,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(24478,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(24478,36,'cp -10','e1g1 e7e6 d3d4 c5d4 f3d4 c6d4 b5d7 d8d7 d1d4 a7a6 f1d1 a8c8 d4d3 f8e7 c1f4 c8d8 d3e3 d7c7 e4e5 d6e5 f4e5 c7c6 d1d8 e7d8 a1d1 d8e7 e3g3 e8g8 e5d6 e7d6 g3d6 c6d6 d1d6 f8c8 g1f1 g8f8 f2f3 c8c7 d6d8 f8e7 d8a8 g7g5 h2h3 h7h5 a2a4 g5g4 h3g4');
INSERT INTO evaluations VALUES(24477,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(24477,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(24477,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(24477,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(24477,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(24477,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(24477,36,'cp -8','e7e6 d3d4 c5d4 f3d4 f8e7 c1e3 a7a6 b5d3 e8g8 h2h3 c6d4 e3d4 d7c6 a2a3 d8c7 d1e2 b7b5 g1h1 g7g6 a1d1 f6h5 e2d2 c7b7 h1h2 a8d8 f2f3 b7c7 d2f2 d6d5 e4e5 f7f6 f3f4 f6e5 d4e5 e7d6');
INSERT INTO evaluations VALUES(26678,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(26678,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(26678,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(26678,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(26678,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(26678,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(26678,36,'cp 60','g8f6 f1c4 e7e6 a2a3 a7a6 h2h3 g4h5 c4e2 h7h6 b2b4 f8d6 e1g1 e8g8 c1b2 d8e8 a1c1 e6e5 d4d5 h5f3 e2f3 c6b8 d1c2 a6a5 b4b5 b8d7 a3a4 d7b6 c3e4 f6e4 f3e4 e8d7 e4h7 g8h8 h7f5 d7e7 g2g3 g7g6 f5e4 f7f5 e4f3 h8g7 f3g2');
INSERT INTO evaluations VALUES(27157,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(27157,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(27157,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(27157,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(27157,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(27157,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(27157,36,'cp -49','b7c6 c2c3 c6c5 g1e2 g8e7 e1g1 e7g6 c1e3 f8e7 b1d2 d8b8 d1c2 c5d4 c3d4 e8g8 d2f3 a7a5 f1c1 f8c8 c2d2 a5a4 h2h3 c7c5 d4c5 g6e5 f3e5 b8e5 c1c3 e5c7 a1c1 d7c6 f2f4 f7f6 e2d4 e6e5 f4e5 f6e5');
INSERT INTO evaluations VALUES(30622,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(30622,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(30622,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(30622,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(30622,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30622,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30622,36,'cp 0','c7c5 g1f3 a7a6 e1g1 b8c6 d3d4 c5d4 f3d4 c6d4 d1d4 g8f6 e4e5 d7d5 e5f6 d5c4 d4c4 g7f6 f1d1 d8c8 c4e2 c8c6 f2f3 h8g8 g1h1 f8d6 c3e4 d6e5 e2d3 c6d5 d3e2');
INSERT INTO evaluations VALUES(33484,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(33484,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(33484,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(33484,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(33484,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(33484,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(33484,36,'cp 53','g2g3 b4d2 d1d2 b8c6 a2a3 d7d5 f1g2 e8g8 e1g1 d5c4 d2c3 f8d8 b1d2 e6e5 d4e5 f6d5 c3c4 c6e5 f3e5 e7e5 f1d1 c7c6 d2f3 e5e7 f3d4 h7h6 d1d3 d5b6 c4c3 b6d5 c3c2 c8g4 a1d1 g4h5 d3d2 h5g6 d4f5 g6f5 c2f5 d5f6 f5c2 d8d2 c2d2 g7g6 b2b4');
INSERT INTO evaluations VALUES(33485,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(33485,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(33485,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(33485,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(33485,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(33485,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(33485,36,'cp 37','b4d2 d1d2 b8c6 a2a3 d7d5 f1g2 e8g8 e1g1 d5c4 d2c3 e6e5 d4e5 f6g4 c3c4 c6e5 f3e5 e7e5 b1c3 c7c6 f1d1 c8e6 c4b4 e5c7 d1d6 g4f6 a1d1 h7h6 b4c5 f8e8 b2b4 c7b6 c5d4 e6b3 d1d2 a8c8 e2e4 b6d4 d6d4');
INSERT INTO evaluations VALUES(36750,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(36750,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(36750,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(36750,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36750,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36750,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36750,36,'cp 0','c7c5 b1d2 b8c6 h2h3 d8b6 a1b1 f6d7 e1g1 e7e5 d4e5 d7e5 d3c2 f8e8 e3e4 d5e4 c2e4 c8e6 d1a4 h7h6 f1d1 e6d7 f3e5 c6e5 a4b3 d7e6 b3a4');
INSERT INTO evaluations VALUES(45167,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(45167,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(45167,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(45167,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(45167,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(45167,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(45167,36,'cp 28','b8c6 e2e3 c8e6 b2b3 d6d5 d2d4 e6g4 g1e2 c5d4 e3d4 e7e6 c4d5 f6d5 c3d5 e6d5 e1g1 f8d6 h2h3 g4e6 e2f4 d6f4 c1f4 e8g8 d1d2 a8c8 a1c1 d8f6 f1d1 h7h6 g3g4 f6g6 c1c3 f8e8 g1h2 h6h5 g4h5 g6h5 c3g3 c6e7 b3b4');
INSERT INTO evaluations VALUES(49455,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49455,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(49455,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49455,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49455,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49455,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49455,36,'cp 0','d2d4 d7d5 e4e5 f6d7 c2c3 c7c5 f1d3 b8c6 g1e2 d8b6 f3e3 f7f6 f2f4 f8e7 b1a3 e8g8 e3h3 g7g6 h3g3 c5d4 d3g6 d4c3 g6h7 g8h7 g3h3 h7g7 h3g4 g7f7 g4h5 f7g7');
INSERT INTO evaluations VALUES(52899,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(52899,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(52899,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(52899,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(52899,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(52899,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(52899,36,'cp -75','e4f3 g1f3 g8f6 f1d3 c8g4 c1e3 e7e6 e1g1 f8e7 d1e1 e8g8 g1h1 b8d7 e1g3 g4h5 g3h4 g7g6 h4g3 h5f3 g3f3 f6h5 a1e1 e7g5 e3g1 h5g7 c3e4 g5e7 f3e2 d7f6 g2g4 f6e4 d3e4 f7f5 e4f3 e7g5 g1f2 f8f7');
INSERT INTO evaluations VALUES(52892,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(52892,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(52892,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(52892,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(52892,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(52892,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(52892,36,'cp -71','g1f3 g8f6 f1d3 c8g4 c1e3 e7e6 e1g1 b8d7 d1e1 g4f3 f1f3 f8d6 e1h4 h7h6 g1h1 d8e7 a1e1 d6c7 a2a3 e8c8 f3f1 h8e8 h4h3 c8b8 e3g1 e7d6 b2b4 a7a6 c3e2 e6e5 e1d1 b8a8 h3g3 g7g6 g3h4 c7b8 d4e5 d6e5 h4c4 e5e6 c4e6');
INSERT INTO evaluations VALUES(52891,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(52891,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(52891,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(52891,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(52891,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(52891,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(52891,36,'cp -80','g8f6 f1d3 c8g4 c1e3 e7e6 e1g1 f8d6 d1e1 g4f3 f1f3 b8d7 g1h1 d8e7 e1h4 h7h6 a1f1 e8c8 a2a3 h8e8 e3g1 c8b8 b2b4 b8a8 h4h3 d6b8 b4b5 c6c5 d4c5 d7c5 b5b6 a7a6 h3h4 c5d3 c2d3 e7d7 h4a4 d7a4 c3a4');
INSERT INTO evaluations VALUES(45030,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(45030,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(45030,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(45030,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(45030,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(45030,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(45030,36,'cp -80','f1d3 c8g4 c1e3 e7e6 e1g1 b8d7 d1e1 g4f3 f1f3 f8d6 g1h1 d8e7 e1h4 h7h6 a1f1 e8c8 e3g1 h8e8 h2h3 d6c7 a2a3 c8b8 h4f2 b8a8 b2b4 c7b8 b4b5 c6b5 c3b5 a7a6 a3a4 e6e5 b5c7 b8c7 d4e5 c7b6 e5f6 e7e5 f2g3 b6g1 h1g1 e5d4 g1h1 g7g6 g3c7 d4e5 c7e5 d7e5 f3g3 a8b8 g3e3 e5d3 e3d3 d8d3 c2d3');
INSERT INTO evaluations VALUES(55529,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(55529,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(55529,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(55529,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(55529,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(55529,35,'cp -154',NULL);
INSERT INTO evaluations VALUES(55529,36,'cp -155','h7h6 f3g1 g8f6 h5f3 d6d5 e4d5 c7c6 f3g3 c6d5 g3e5 f8e7 e5e2 c8g4 g1f3 e8g8 d2d4 f8e8 c1e3 b8c6 c2c3 e7d6 e2d3 g4h5 f1e2 d8c7 h2h3 e8e3 d3e3 a8e8 e3d2 h5f3 g2f3 f6h5 e1d1 d6f4 d2c2 c6d4 c3d4 c7c2 d1c2 e8e2 c2b3 e2f2 b1c3 f2f3 h1d1 f3h3');
INSERT INTO evaluations VALUES(55566,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(55566,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(55566,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(55566,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55566,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(55566,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(55566,36,'cp 17','e5d4 d1d4 b8c6 d4d2 a7a5 g2g3 a5a4 f1g2 c8e6 d2c2 e6c4 g1f3 f8e7 c3a4 c4e6 e1g1 g8f6 a4c3 e8g8 f1d1 f8e8 f3g5 e6c8 c2b3 e8f8 g5f3 f8e8 a2a3 h7h6 b3c2 e7f8 e2e4 c8g4 h2h3 g4h5 c1f4 h5g6 f3d4 c6e5 b2b3');
INSERT INTO evaluations VALUES(55669,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(55669,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(55669,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(55669,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(55669,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55669,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(55669,36,'cp 25','d1d4 b8c6 d4d2 g8f6 e2e4 g7g6 f1d3 f8g7 b2b3 e8g8 g1e2 a7a6 e1g1 b7b5 c4b5 c6b4 b5b6 c7b6 c1a3 b4d3 d2d3 f8e8 a3d6 b6b5 a1d1 b5b4 d6b4 d8d3 d1d3 a6a5 b4c5 c8a6 d3d2 g7h6 d2d4 a8c8 c5b6 a6e2 c3e2 c8c2 f1e1 f6e4 b6a5 c2a2 e2c3 a2a5 d4e4 e8e4 c3e4 h6g7 f2f3');
INSERT INTO evaluations VALUES(55668,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55668,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55668,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55668,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(55668,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(55668,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(55668,36,'cp 9','b8c6 d4d2 g8f6 e2e4 g7g6 b2b3 f8g7 f1d3 a7a6 g1e2 e8g8 e1g1 b7b5 c4b5 c6b4 b5b6 c7b6 c1a3 b4d3 d2d3 f8e8 a3d6 b6b5 a1d1 b5b4 d6b4 d8d3 d1d3 a6a5 b4c5 c8a6 d3d2 a8c8 c5d4 f6e4 c3e4 e8e4 d4e3 a6e2 d2e2 g7h6 f1d1 h6e3 f2e3 c8c3 g1f2 a5a4');
INSERT INTO evaluations VALUES(880,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(880,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(880,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(880,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(880,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(880,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(880,36,'cp 201','b1c3 d5d6 e2e4 e7e5 d4d5 c6b8 f1e2 f8e7 c3b5 d6d8 f3e5 e8g8 b5c3 f8e8 e5f7 g8f7 e4e5 f6d7 e5e6 f7g8 e6d7 b8d7 e1g1 d7c5 b2b4 c5d7 c1f4 a7a6 a2a3 d7f6 d1d2 c8f5 f1e1 e7d6 e2f3 d8d7 a1d1 f5g4 f3g4 d7g4 f4d6 c7d6 h2h3');
INSERT INTO evaluations VALUES(877,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(877,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(877,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(877,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(877,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(877,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(877,36,'cp 200','d5d6 e2e4 e7e5 d4d5 c6b8 f1e2 f8e7 c3b5 d6d8 f3e5 e8g8 b5c3 f8e8 e5f7 g8f7 e4e5 f6d7 e5e6 f7g8 e6d7 b8d7 e1g1 d7f6 c1f4 e7d6 d1d4 c8f5 f1e1 a7a6 h2h3 d8d7 e2f3 f5g6 a2a3 d6f4 d4f4 d7d6 f4d6 c7d6 a1d1 e8e1 d1e1 g8f8 g2g4 a8d8 g1g2 f6d7 h3h4 d7c5 h4h5 g6f7');
INSERT INTO evaluations VALUES(3381,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(3381,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(3381,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(3381,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3381,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(3381,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3381,36,'cp 0','c6a5 c4b5 a7a6 b5a4 c7c5 a4c2 b7b5 d3d4 e5d4 c3d4 c5d4 f3d4 f8e8 b2b4 a5c4 a2a4 c8d7 b1c3 d8b6 a4b5 a6b5 a1a8 e8a8 f1e1 e7f8 c1g5 c4b2 d1d2 b2c4');
INSERT INTO evaluations VALUES(3449,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(3449,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3449,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3449,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3449,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3449,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3449,36,'cp 17','c4b5 c7c5 b5a4 a5c6 c1e3 a8b8 b1d2 b7b5 a4c2 d8b6 f1e1 c8e6 f3g5 e6c8 f2f4 h7h6 g5f3 e5f4 e3f4 c8e6 f4e3 b5b4 d3d4 d6d5 e4e5 f6d7 d2f1 b4c3 b2c3 c5d4 e3f2 d7c5 f3d4 c6d4 f2d4');
INSERT INTO evaluations VALUES(3898,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(3898,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(3898,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(3898,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(3898,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(3898,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(3898,36,'cp -45','f8e8 e5f3 f6e4 c3e4 e8e4 d2d3 e4e8 f1e1 c8g4 e1e8 d8e8 c1f4 e8e7 h2h3 g4f3 d1f3 e7f6 c2c3 c5d6 f4e3 f6f3 g2f3 a8e8 d3d4 e8e6 a1d1 g8f8 d4d5 c6d5 d1d5 f8e7 g1f1 g7g6 f1e2 e7d7 a2a4 b7b6 d5d4 f7f5 e2d3 c7c5 d4h4 h7h5 b2b4');
INSERT INTO evaluations VALUES(23534,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(23534,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(23534,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(23534,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(23534,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(23534,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(23534,36,'cp -40','e5f3 f6e4 c3e4 e8e4 d2d3 e4e8 f1e1 c8g4 e1e8 d8e8 c1f4 e8e7 h2h3 g4f3 d1f3 e7f6 c2c3 c5d6 f4e3 f6f3 g2f3 a8e8 d3d4 g8f8 a1d1 e8e6 d4d5 c6d5 d1d5 a7a6 g1f1 f8e7 d5d4 e7d7 a2a4 b7b6 d4h4 h7h6 b2b4 b6b5 a4b5 a6b5 f1e2 e6e5 h4g4 g7g6 f3f4 e5e4');
INSERT INTO evaluations VALUES(4806,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(4806,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(4806,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4806,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(4806,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4806,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4806,36,'cp 12','d2d4 e5d4 d1d4 f7f6 c1e3 g8e7 b1d2 e7g6 e1c1 c8e6 d2c4 e6c4 d4c4 d8e7 f3d2 e8c8 h1e1 g6f4 c4f1 g7g5 a2a3 f4g6 g2g3 h8e8 f2f4 g5f4 g3f4 d6c5 e3c5 e7c5 f1f3 c8b8 c1b1 c5e7 f3g3 e8g8 h3h4 h7h5 g3f3 g6h4 f3h5 h4g6 h5f5');
INSERT INTO evaluations VALUES(13583,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(13583,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(13583,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(13583,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(13583,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(13583,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(13583,36,'cp 11','c1b2 b7b6 b1d2 f8e7 f3e5 c8b7 e1g1 e8g8 e5c6 b7c6 d1e2 a7a5 a2a4 a8c8 d4c5 e7c5 d3a6 c8c7 f1d1 d8e7 c2c4 f8d8 c4d5 d8d5 a1c1 h7h6 c1c2 d5d8 d1c1 c6d5 h2h3 f6e4 e2g4 e7g5 g4g5 e4g5 a6e2 g5e4 d2e4');
INSERT INTO evaluations VALUES(13584,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(13584,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(13584,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13584,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(13584,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13584,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(13584,36,'cp 0','b7b6 b1d2 c8b7 e1g1 f8d6 c2c4 c5d4 c4d5 f6d5 f3d4 c6d4 b2d4 e8g8 d2c4 d6c7 d1g4 d5f6 g4h3 h7h6 a1c1 a8c8 f1d1 d8e7 d4b2 f8d8 b2a3 e7e8 a3b2');
INSERT INTO evaluations VALUES(14309,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14309,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(14309,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(14309,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(14309,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(14309,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(14309,36,'cp 39','g1f3 c8f5 b2b4 c6b8 c2c4 e7e6 c4c5 c7c6 b1d2 b8d7 f1e2 f6e4 d2e4 f5e4 e1g1 f8e7 f3d2 e4g6 d2b3 e8g8 c1d2 a7a6 f2f3 d8c8 a1c1 e6e5 d4e5 d7e5 b3a5 e5d7 d2e1 h7h6 e1g3 f8e8 d1d2 e7f6 f1e1 g6h7');
INSERT INTO evaluations VALUES(14586,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(14586,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(14586,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14586,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(14586,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14586,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(14586,36,'cp 48','d5b6 c4b5 f8d6 d3d4 e5d4 f1e1 c8e6 b5c6 b7c6 f3d4 e8g8 d4e6 f7e6 d1e2 b6d5 g2g3 d6c5 e1f1 e6e5 b1d2 c5b6 d2e4 d5f6 e4f6 d8f6 a2a4 a7a5 c1e3 f6e6 a1e1 e6b3 e2g4 g8h8 e3b6 c7b6 e1e2 b6b5 g1g2 b5b4 f1c1 b3a4 e2e5 a4b3 c3b4');
INSERT INTO evaluations VALUES(17959,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(17959,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(17959,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(17959,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(17959,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(17959,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(17959,36,'cp 2','b5c6 d7c6 g1f3 e7e5 e1g1 c8g4 h2h3 h7h5 d2d3 d8f6 b1d2 g8e7 f1e1 e7g6 d3d4 g6f4 d4e5 f6g6 f3h4 g4d1 h4g6 f4g6 e1d1 e8c8 e5e6 f7f6 d1e1 d8e8 d2f3 f8d6 f3d4 g6f8 e6e7 e8e7 d4f5 e7d7 f5d6 c7d6 c1e3 g7g5 e3d4 d7f7 d4c3 f8g6 a1d1 c8c7 f2f3 h8e8 d1d2 b7b5');
INSERT INTO evaluations VALUES(17958,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(17958,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(17958,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17958,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17958,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(17958,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(17958,36,'cp 1','d7c6 d2d3 e7e5 c1e3 g8e7 b1c3 e7g6 d1d2 a6a5 e1c1 f8d6 d3d4 e8g8 g1e2 a5a4 h2h4 h7h5 d4e5 g6e5 e3c5 e5c4 d2d4 c8e6 f2f4 f7f6 c5d6 d8d6 a2a3 d6d4 e2d4 e6d7 d4f3 d7g4 d1e1 b7b5 f3d4');
INSERT INTO evaluations VALUES(17731,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17731,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17731,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17731,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17731,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17731,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17731,36,'cp 0','d1f3 g8f6 d2d3 e7e5 b1c3 c8e6 g1e2 d8d6 c1e3 e8c8 e1c1 c6c5 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(26148,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(26148,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(26148,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(26148,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(26148,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(26148,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(26148,36,'cp -132','c5d4 e1g1 g8e7 d1d4 e7g6 c1e3 d8c7 b1c3 a7a6 a1e1 a8c8 d4d2 h7h6 e1e2 f8e7 e3d4 e7c5 d4c5 c7c5 f1e1 e8g8 g2g3 c5b6 c3a4 b6c7 a4c3 g6e7 c3b1 f8d8 c2c3 b7b5 a2a3 d7b6 g1g2 b6c4 d2c1 e7c6');
INSERT INTO evaluations VALUES(26163,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(26163,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(26163,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(26163,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(26163,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(26163,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(26163,36,'cp -117','d1d4 g8e7 e1g1 e7g6 c1e3 d8c7 b1c3 f8c5 d4d3 e8g8 e3c5 c7c5 f1e1 a8c8 a1c1 f8d8 d3b5 c5c7 b5e2 a7a6 c3b1 c7b6 b2b3 c8c7 c2c3 h7h6 g2g3 b6c5 h2h4 d8c8 c1d1 b7b5 a2a3 c5a7 g1g2 a7b7 g2g1 g6e7 f3d4 e7c6 d4f3');
INSERT INTO evaluations VALUES(26162,30,'cp -126',NULL);
INSERT INTO evaluations VALUES(26162,31,'cp -123',NULL);
INSERT INTO evaluations VALUES(26162,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(26162,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(26162,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(26162,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(26162,36,'cp -134','g8e7 e1g1 e7g6 c1e3 d8c7 b1c3 f8c5 d4d3 e8g8 e3c5 c7c5 f1e1 a8c8 a1c1 f8d8 c3b1 d5d4 d3e4 c5b6 e4d4 g6e5 d4b6 e5f3 g2f3 d7b6 e1d1 b6d5 d1d2 g8f8 c1d1 f8e7 c2c3 d5f4 b1a3 d8d2 d1d2 g7g5 g1f1 b7b6 a3b5 a7a6 b5d6');
INSERT INTO evaluations VALUES(29478,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(29478,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(29478,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(29478,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(29478,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(29478,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(29478,36,'cp 212','g1f3 c7c6 f3g5 e6g4 f2f3 g4h5 f1c4 h7h6 g5h3 g7g5 h3f2 b8d7 g2g4 h5g6 h2h4 g5h4 f3f4 e7e6 f4f5 e6f5 e4f5 d8e7 e1f1 d7b6 d1b3 e8c8 f5g6 f7g6 c4e6 c8b8 c1f4 b8a8 a1d1 g6g5 f4e5 f8g7 f1g2 h6h5 h1e1 h5g4 e6f5 b6d5 c3d5 f6d5 f5g4 g7e5 d4e5 d5f4 g2h1 d8d1 f2d1');
INSERT INTO evaluations VALUES(29476,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(29476,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(29476,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(29476,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(29476,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(29476,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(29476,36,'cp 184','c7c6 f3g5 e6g4 f2f3 g4h5 f1c4 h7h6 g5h3 g7g5 e1g1 e7e5 d4e5 d8d1 c3d1 f6d7 e5e6 d7e5 e6f7 h5f7 c4f7 e5f7 f3f4 b8d7 f4g5 f8c5 d1f2 h6g5 h3g5 f7e5 g5e6 c5b6 g1h1 e8f7 e6g5 f7e7 f2h3 a8f8 h3f4 f8g8 g5f3 e5f3 g2f3 e7e8 c1d2 b6d4 a1d1 d7e5 d2c3 d4c3');
INSERT INTO evaluations VALUES(28929,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(28929,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(28929,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(28929,33,'cp 216',NULL);
INSERT INTO evaluations VALUES(28929,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(28929,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(28929,36,'cp 187','f3g5 e6g4 f2f3 g4h5 f1c4 h7h6 g5h3 g7g5 e1g1 e7e5 d4e5 d8d1 c3d1 f6d7 e5e6 d7e5 e6f7 h5f7 c4f7 e5f7 d1e3 b8d7 h3f2 e8c8 f1d1 f8c5 g1f1 d7e5 d1d8 h8d8 f1e2 f7d6 a2a4 c5d4 a1b1 c8c7 b2b3 b7b5 c1b2 d4b2 b1b2 b5a4 b3a4 e5c4 e3c4');
INSERT INTO evaluations VALUES(34855,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(34855,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(34855,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(34855,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(34855,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(34855,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(34855,36,'cp 144','e2e4 c7c5 d4d5 g8f6 c1d2 a5d8 g1f3 a7a6 d2f4 e6d5 e4d5 f8d6 d1e2 e8f8 e2e3 b7b5 f1d3 d6f4 e3f4 g7g6 e1c1 f8g7 h1e1 b8d7 d5d6 c8b7 g2g4 c5c4 d3e4 f6e4 c3e4 h8e8 e4g5 d8f6 f4f6 d7f6 e1e8 a8e8 d6d7 e8d8 f3e5');
INSERT INTO evaluations VALUES(35253,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(35253,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(35253,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(35253,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(35253,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(35253,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(35253,36,'cp 121','e1g1 d7d6 d2d4 g8f6 f1e1 b8c6 c4b5 a7a6 b5c6 b7c6 b1a3 f6h7 d4e5 d6e5 d1d3 h7g5 c1g5 h6g5 d3c3 e7c5 c3e5 c5e5 f3e5 f8a3 b2a3 g5g4 e5c6 h8h5 e4e5 h5h6 c6d4 h6b6 a1d1 c7c5 d4b3 c5c4 b3c5 b6b2 e5e6 c8e6 c5e6');
INSERT INTO evaluations VALUES(35252,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(35252,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(35252,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(35252,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(35252,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(35252,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(35252,36,'cp 133','g8f6 d2d4 d7d6 b1c3 c7c6 a2a4 c8g4 d4d5 c6d5 c3d5 f6d5 c4d5 e7d7 c1e3 f8e7 b2b4 g4e6 c2c4 e8g8 b4b5 e6d5 c4d5 f7f5 d1d3 f5f4 e3d2 a7a6 f1b1 a6b5 d3b5 a8a7 h2h3 d7b5 a4b5 a7a1 b1a1 b8d7 a1a7 f8b8 d2b4 g8f7 f3d2 d7c5 b4c5 d6c5');
INSERT INTO evaluations VALUES(37181,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(37181,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(37181,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(37181,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(37181,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37181,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(37181,36,'cp 0','d7d5 g1f3 d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 d1e2 c8b7 c1g5 b8d7 f3e5 b4c3 b2c3 a8c8 a1c1 d7e5 d4e5 c8c5 c4d3 d8c7 g5f6 g7f6 e2g4 g8h8 g4h4 f6f5 h4f6 h8g8 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(37048,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(37048,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(37048,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(37048,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(37048,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37048,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37048,36,'cp 0','g1f3 d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 b7b6 d1e2 c8b7 f1d1 b4c3 b2c3 d8c7 c1b2 b7f3 e2f3 c7c4 f3a8 b8c6 a8b7 f6d5 h2h3 f8b8 b7d7 b8d8 d7b7');
INSERT INTO evaluations VALUES(39826,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(39826,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(39826,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(39826,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(39826,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(39826,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(39826,36,'cp 17','g2g3 c8b7 g1f3 b4c3 d2c3 f6e4 c3d2 e8g8 f1g2 e4d2 d1d2 d7d6 e1g1 b8d7 f3h4 b7g2 h4g2 d7f6 f2f3 c7c5 e2e4 c5d4 d2d4 d8b8 f1d1 f8d8 a1c1 f6d7 b2b3 h7h6 g2e1 d7c5 d4e3 b8c7 e1d3');
INSERT INTO evaluations VALUES(43562,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43562,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43562,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43562,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(43562,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(43562,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43562,36,'cp 24','c4d5 e6d5 f1e2 c5d4 e3d4 h7h6 f3e5 b8c6 e1g1 f8d6 f1e1 e8g8 c1f4 c8e6 e5c6 b7c6 f4d6 d8d6 c3a4 e6f5 a4c5 f6e4 e2d3 e4c5 d3f5 c5d7 f5d3 a6a5 a1c1 a8b8 b2b3 f8e8 e1e8 b8e8 g2g3 d7f6 d1d2');
INSERT INTO evaluations VALUES(43560,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43560,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(43560,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(43560,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(43560,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(43560,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(43560,36,'cp 14','e6d5 b2b3 b8c6 c1b2 c8g4 f1e2 g4f3 e2f3 c5d4 e3d4 f8b4 e1g1 b4c3 b2c3 e8g8 f1e1 f8e8 d1d3 d8b6 a1c1 h7h6 g2g3 b6b5 d3f5 e8e7 e1e7 c6e7 f5b1 g7g6 g1g2 e7f5 c3b2 g8g7 c1c5 b5d7 b1d3 a8d8');
INSERT INTO evaluations VALUES(46129,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(46129,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46129,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(46129,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(46129,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(46129,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(46129,36,'cp -21','e6d5 c4b3 f8e7 g1e2 e8g8 e1g1 c8e6 e2f4 b8c6 f4e6 f7e6 c3e2 e7d6 e2f4 d8d7 c2c3 d7f7 d3d4 c5c4 b3c2 e6e5 f4e2 a8e8 d4e5 c6e5 e2d4 d6c5 c1e3 e5g4 e3f4 c5d4 c3d4 f6e4 d1g4 f7f4 g4f4 f8f4 g2g3 f4f3 a1e1 g8h8 h2h3');
INSERT INTO evaluations VALUES(50210,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(50210,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(50210,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50210,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(50210,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(50210,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(50210,36,'cp 0','f8g7 g1f3 e8g8 f1e2 c7c5 d4c5 a7a5 a2a3 a5a4 b1d2 b7b6 e1g1 b6c5 e3e4 d8c7 e2b5 c7a5 b5e2');
INSERT INTO evaluations VALUES(53971,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(53971,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(53971,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(53971,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(53971,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(53971,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(53971,36,'cp -16','b8c6 g1f3 g8f6 d3d4 c5d4 f3d4 f8e7 e1g1 e8g8 c1e3 d8c7 c3b5 c7b8 b5c3 a7a6 a2a4 c8d7 c4d3 c6e5 h2h3 e5d3 c2d3 b7b5 a4b5 a6b5 a1a8 b8a8 c3b5 a8b8 b5a3 d6d5 e4d5 f6d5 a3c4 e7f6 d1d2 f8d8 d4f3 d5e3');
INSERT INTO evaluations VALUES(54432,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(54432,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(54432,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(54432,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(54432,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(54432,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(54432,36,'cp 134','c3d4 b8c6 b1c3 g8e7 h2h4 e7f5 g2g4 f5e7 f1d3 h7h6 c3b5 e7g8 a2a3 c8d7 b5c3 c6a5 h4h5 g8e7 a1b1 a5c4 e1f1 e7c6 h1h3 d8a5 h3g3 e8c8 b2b4 a5c7 d1b3 c8b8 d3c4 d5c4 b3c4 d8c8 c1d2 c7d8 c4d3 c6e7 b4b5 g7g6 d4d5');
INSERT INTO evaluations VALUES(54420,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(54420,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(54420,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(54420,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(54420,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(54420,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(54420,36,'cp 140','b8c6 b1c3 g8e7 h2h4 e7f5 g2g4 f5e7 f1d3 c6a5 h4h5 c8d7 e1f1 h7h6 a1b1 e7c6 h1h3 d8c7 a2a3 e8c8 c3b5 c7b6 b2b3 c8b8 c1e3 d7e8 b5d6 b6c7 b3b4 a5c4 d3c4 d5c4 d6e8 d8e8 b1c1 c6e7 d1a4 e8d8 f3d2 e7d5 c1c4');
INSERT INTO evaluations VALUES(20423,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(20423,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(20423,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(20423,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(20423,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(20423,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(20423,36,'cp 137','b1c3 g8e7 h2h4 h7h6 h4h5 c8d7 f1e2 e7f5 g2g4 f5e7 e1f1 c6a5 h1h3 d8b6 a1b1 a8c8 e2d3 a5c4 c3e2 e7c6 b2b3 c4a3 b1a1 f8b4 g4g5 c6e7 g5h6 g7h6 c1d2 h8g8 d2h6 d7b5 h3g3 g8g3 e2g3 a3c2 a1c1');
INSERT INTO evaluations VALUES(57252,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(57252,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(57252,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(57252,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(57252,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(57252,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(57252,36,'cp -82','d7d5 e2e3 b8c6 d2d4 g8f6 f1b5 e5d4 e3d4 f8d6 g1e2 e8g8 e1g1 f6g4 g2g3 f8e8 e2f4 a7a6 b5e2 g4f6 f1e1 c8f5 c2c3 h7h6 c1e3 c6e7 b1d2 e7g6 f4h5 f6h5 e2h5 g6h4 g3h4 d8h4');
INSERT INTO evaluations VALUES(56504,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(56504,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(56504,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(56504,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(56504,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(56504,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(56504,36,'cp -86','e2e3 g8f6 d2d4 e5d4 e3d4 f6e4 c2c3 f8d6 f1d3 e8g8 g1e2 c8g4 c1a3 d8h4 d3e4 d5e4 a3d6 c7d6 e1g1 b8c6 h2h3 g4h5 d1e1 f7f5 f2f3 h4e1 f1e1 e4f3 g2f3 h5f3 e2f4 f3e4 b1d2 d6d5 h3h4 a8c8 g1f2 f8e8 a4a5 g8f7 e1e3 c6d8 a1g1 a7a6');
INSERT INTO evaluations VALUES(4239,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(4239,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4239,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(4239,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(4239,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4239,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4239,36,'cp 5','b1d2 d7d6 a2a4 a7a5 b4b5 c6e7 e1g1 e7g6 c4b3 d6d5 e4d5 c8f5 d2c4 f6d5 d1d2 d5f4 c4e5 g6e5 d2f4 e5f3 f4f3 d8d3 b3d5 c7c6 f1d1 d3f3 d5f3 f8d8 c1f4 d8d1 a1d1 a8c8 b5c6 b7c6 d1d2 h7h6 h2h4 f5e6 g2g3 b6c5 g1g2');
INSERT INTO evaluations VALUES(6503,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(6503,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6503,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(6503,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(6503,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(6503,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(6503,36,'cp 20','f3e5 d8e7 e5c6 e7e4 d1e2 e4e2 f1e2 d7c6 b1c3 c8e6 d2d3 f7f5 c1e3 c5d6 e3d4 g8f6 d4f6 g7f6 e2h5 e8f8 h3h4 e6f7 h5f3 a8e8 e1f1 h7h5 a1b1 f7e6 a2a4 a7a5 c3e2 f5f4 e2d4');
INSERT INTO evaluations VALUES(6792,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(6792,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(6792,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(6792,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(6792,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(6792,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(6792,36,'cp 220','d1d4 d8e7 b1c3 g8f6 f2f3 d6e5 d4d3 e8g8 c1e3 d7d6 e1c1 c7c5 d5c6 b7c6 e3d4 f8d8 d4e5 e7e5 d3d4 e5g5 d4d2 g5a5 c1b1 a5c7 f1c4 c8e6 c4e6 f7e6 h1e1 a8b8 d2e3 a7a5 b2b3 e6e5 d1d3 f6h5 g2g3 h5f6 b1a1 d6d5 e4d5 f6d5 e3c5 d5b4 d3d8 b8d8 e1e5 b4c2 a1b2');
INSERT INTO evaluations VALUES(6791,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(6791,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(6791,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(6791,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(6791,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(6791,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(6791,36,'cp 206','g8f6 b1c3 e8g8 f2f3 c7c6 c1g5 f8e8 g5f6 d8f6 d4f6 g7f6 e1c1 d6f4 c1b1 f4e5 c3e2 c6d5 e4d5 f6f5 e2g3 d7d6 f1d3 f5f4 g3h5 c8d7 g2g3 f4g3 h2g3 a8c8 g3g4 h7h6 a2a3 g8f8 c2c3 a7a5 d1g1 c8c5 f3f4 c5d5 f4e5 d5d3');
INSERT INTO evaluations VALUES(14300,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14300,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(14300,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14300,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(14300,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14300,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(14300,36,'cp 55','e2e3 g7g6 b2b3 f8g7 c1b2 e8g8 f1b5 c6b8 b5e2 c8f5 c2c4 c7c6 e1g1 b8d7 a2a4 h7h5 f3h4 e7e6 h4f5 e6f5 b3b4 f8e8 b4b5 a8c8 b5c6 b7c6 a1c1 g7f8 c4d5 c6d5 e2b5 c8c1 b2c1 d8c7 d1b3 e8b8 c1a3');
INSERT INTO evaluations VALUES(14770,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14770,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14770,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14770,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14770,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14770,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(14770,36,'cp 55','e2e3 f8d6 f4g3 e8g8 g1f3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c6e7 d1c2 h7h6 e1g1 c7c6 f1c1 c8e6 a2a3 a7a5 b2b3 f8e8 c2c3 d8d7 g3d6 d7d6 h2h3 e8c8 b3b4 e6f5 f3e5 f5d3 e5d3 d6d8 c3b2 d8c7 d3e5 a5b4 a3b4 a8a1 b2a1');
INSERT INTO evaluations VALUES(22201,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(22201,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(22201,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(22201,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(22201,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(22201,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(22201,36,'cp 88','d1c2 f8e7 h2h3 c7c6 f1a6 b7a6 d5c6 d8b6 c2a4 c8e6 e1g1 a8c8 f1e1 b6c6 a4c6 c8c6 b2b3 e7d8 c3c4 g8f6 e1e2 d8a5 c1d2 a5b6 d2b4 f6d7 b1c3 a6a5 b4a3 b6c5 a3c5 d7c5 c3b5 e8e7 b5a7 c6a6 a7b5 h8c8 f3d2');
INSERT INTO evaluations VALUES(22429,30,'cp -486',NULL);
INSERT INTO evaluations VALUES(22429,31,'cp -498',NULL);
INSERT INTO evaluations VALUES(22429,32,'cp -502',NULL);
INSERT INTO evaluations VALUES(22429,33,'cp -494',NULL);
INSERT INTO evaluations VALUES(22429,34,'cp -499',NULL);
INSERT INTO evaluations VALUES(22429,35,'cp -509',NULL);
INSERT INTO evaluations VALUES(22429,36,'cp -513','e8f7 f2f4 e5f4 g1f3 d7d5 e4d5 d8d5 b1c3 d5f5 d2d4 g8f6 e1g1 f7g8 d1e2 f8d6 c3b5 c8d7 f3h4 f5h5 e2c4 h5d5 c4d5 f6d5 c1d2 g7g5 h4f3 h7h6 b5d6 c7d6 a1e1 d5b4 c2c4 b4d3 d4d5 c6e5 f3e5 d6e5 d2c3 a8e8 d5d6 d3e1 f1e1');
INSERT INTO evaluations VALUES(2502,30,'cp -494',NULL);
INSERT INTO evaluations VALUES(2502,31,'cp -513',NULL);
INSERT INTO evaluations VALUES(2502,32,'cp -510',NULL);
INSERT INTO evaluations VALUES(2502,33,'cp -488',NULL);
INSERT INTO evaluations VALUES(2502,34,'cp -483',NULL);
INSERT INTO evaluations VALUES(2502,35,'cp -488',NULL);
INSERT INTO evaluations VALUES(2502,36,'cp -494','b1c3 g8f6 g1e2 d7d5 e4d5 f6d5 e1g1 f7g8 d2d4 d5c3 b2c3 c8e6 a1b1 b7b6 f2f4 e5d4 c3d4 e6a2 b1b2 a2c4 c2c3 d8d7 f4f5 a8e8 f1f3 h7h5 e2f4 d7f5 f4d3 f5d5 d3f4 d5d6 b2f2 h8h6 f4d3 d6g6 f3f8 e8f8 c1h6 g6h6');
INSERT INTO evaluations VALUES(22437,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(22437,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(22437,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(22437,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(22437,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(22437,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(22437,36,'cp 183','d4d5 c6b8 b1c3 d7d6 f1b5 c8d7 d5e6 f7e6 d1d4 e6e5 d4d3 g8f6 b5c4 f8g7 a2a3 d8e7 e1g1 b8c6 c1g5 d7e6 c3d5 e6d5 e4d5 c6d8 a1e1 d8f7 c4b5 e8f8 b5a4 e5e4 d3b3 a8b8 g5f4 g7h6 f4h6 f7h6 f3d4');
INSERT INTO evaluations VALUES(25327,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(25327,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(25327,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(25327,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(25327,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(25327,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(25327,36,'cp 12','c4b5 e5f4 d2d4 d6d5 e4d5 f6d5 e1g1 f8e7 h2h3 g4h5 d1e2 d5c3 b2c3 e8g8 c1f4 e7d6 e2d2 h5g6 b5c4 d6f4 d2f4 d8d6 f4d2 c6a5 c4d3 a8e8 a1e1 a5c6 d3c4 b7b6 d2f2 e8e1 f1e1 c6a5 c4d3 f7f6 f3h4 g6d3 c2d3 g7g6 h4f3 a5c6 c3c4 c6b4');
INSERT INTO evaluations VALUES(26147,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(26147,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(26147,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(26147,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(26147,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(26147,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(26147,36,'cp 70','c2c4 d5c4 d3c4 c5d4 c4d4 d8b6 d4e4 b6c6 e4e2 c6a6 e2a6 b7a6 c1e3 f8c5 e3c5 d7c5 b1d2 g8e7 d2c4 e8g8 f1d1 a8b8 b2b3 e7d5 c4d6 c5d7 a1c1 f8d8 d6e4 b8c8 g2g3 h7h6 g1f1 c8c1 d1c1 d5b4 c1c4 b4a2 c4c7');
INSERT INTO evaluations VALUES(27618,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(27618,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(27618,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(27618,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(27618,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(27618,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(27618,36,'cp 159','b1c3 d5a5 d4d5 e8c8 c1d2 d8d5 c3d5 a5d5 h2h3 g4f5 e2e3 e7e6 a1c1 f5e4 f1e2 f8d6 e1g1 g8f6 d2c3 d5a2 f3d2 a2d5 d2e4 f6e4 c3g7 h8g8 d1d5 e6d5 g7h6 g8g6 h6f4 d6f4 e3f4 f7f5 f1d1 c6e7 g2g4 c7c6 g4g5 a7a6 f2f3 e4g3 g1f2 g3h5');
INSERT INTO evaluations VALUES(27614,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(27614,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(27614,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(27614,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(27614,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(27614,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(27614,36,'cp 153','d5a5 d4d5 e8c8 c1d2 d8d5 c3d5 a5d5 h2h3 g4f3 g2f3 g8f6 e2e3 e7e6 f1e2 f8d6 d1a4 f6d7 a4c4 g7g6 h3h4 h7h5 e1c1 d7b6 d2c3 d5c4 e2c4 d6e5 c3e5 c6e5 c4e2 c7c6 f3f4 e5d7 c1c2 d7f6 h1g1 c8c7 e2f3 f6d5 f3d5');
INSERT INTO evaluations VALUES(27633,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(27633,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(27633,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(27633,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(27633,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(27633,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(27633,36,'cp 149','d4d5 e8c8 c1d2 d8d5 c3d5 a5d5 h2h3 g4f5 e2e3 e7e6 a1c1 f5e4 f1c4 d5h5 c4e2 g8f6 d1a4 a7a6 d2c3 e4d5 f3d2 h5g5 e2f3 c6e5 c3e5 g5e5 d2c4 d5c4 c1c4 f8d6 a4c2 c7c6 e1e2 f6d5 c4e4 e5f6 h1d1 h8d8 e4c4 d6c7 b2b4 f6e5 a2a4 c8b8');
INSERT INTO evaluations VALUES(29719,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(29719,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(29719,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(29719,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(29719,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(29719,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(29719,36,'cp -71','f6e4 d1f3 e4c3 b2c3 e7e6 f3b7 b8d7 f1d3 a8b8 b7f3 f5d3 c2d3 f8d6 g1e2 b8b5 e2g3 d8h4 f3f2 f7f5 g3e2 h4f2 e1f2 e8f7 c3c4 b5b8 h2h3 c7c5 c1a3 b8b6 h1b1 h8d8 b1b6 a7b6 d4c5 d7c5 f2e3 f5f4 e2f4 d6f4 e3f4');
INSERT INTO evaluations VALUES(29825,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(29825,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(29825,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(29825,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(29825,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(29825,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(29825,36,'cp -69','d1f3 e4c3 b2c3 e7e6 f3b7 b8d7 f1d3 f5d3 c2d3 a8b8 b7f3 f8d6 g1e2 b8b5 e2g3 e8g8 e1g1 d8h4 c1f4 d6f4 f3f4 h4f4 f1f4 f8b8 g3e4 h7h6 a1e1 b5a5 f4f2 a5a3 f2c2 b8b5 c3c4 b5b7 e4c3 c7c6 d4d5 c6d5 c4d5 e6d5 c3d5 a3d3 d5e7 g8h7');
INSERT INTO evaluations VALUES(35453,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(35453,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(35453,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35453,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35453,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35453,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35453,36,'cp 0','f7f6 g5d2 b8d7 b1c3 c7c6 a2a4 a7a5 f2f4 g8h6 g1f3 e5f4 d2f4 f8c5 e1c1 h6g4 e4e5 g4f2 e5e6 f2d1 e6d7 d1f2 d7c8r a8c8 h1g1 f2g4 g1h1');
INSERT INTO evaluations VALUES(35467,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35467,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35467,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35467,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35467,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35467,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35467,36,'cp 0','g5d2 b8d7 b1c3 c7c6 a2a4 a7a5 f2f4 e5f4 d2f4 f8c5 g1f3 g8h6 e1c1 h6g4 e4e5 g4f2 e5e6 f2d1 e6d7 d1f2 d7c8q a8c8 h1g1 f2g4 g1h1');
INSERT INTO evaluations VALUES(39168,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(39168,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(39168,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(39168,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(39168,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(39168,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(39168,36,'cp 47','e2e4 b8c6 d2d4 d7d5 c4d5 e6d5 e4e5 g8e7 f1e2 b4a5 g1f3 e8g8 e1g1 h7h6 c3a4 a5b6 a4b6 a7b6 a2a3 c8g4 h2h3 g4f3 e2f3 f7f6 e5f6 f8f6 f3g4 e7f5 c1d2 c6d4 d2c3 c7c5 g4f5 f6f5 c3d4 c5d4 d1d4 d8f6 d4d2 d5d4 a1d1 f5f4 f1e1');
INSERT INTO evaluations VALUES(41916,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(41916,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41916,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(41916,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(41916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41916,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41916,36,'cp 8','a2a3 b4c3 b2c3 e8g8 c3c4 b8c6 f1e2 f6d7 c1g5 c6e7 f3h4 d7c5 e1g1 f7f6 g5e3 a7a6 a1b1 b7b6 h2h3 g7g5 h4f3 c5e6 f3h2 e6f4 e2g4 c8g4 h2g4 g8h8 e3d2 f6f5 g4e3 a6a5 a3a4 f4e6 e4f5 e7f5 e3f5 f8f5');
INSERT INTO evaluations VALUES(44761,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44761,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44761,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44761,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(44761,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(44761,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(44761,36,'cp 1','d2d4 c5d4 f3d4 f8g7 c1e3 e8g8 d1d2 d6d5 e4d5 f6d5 e1c1 d5e3 d2e3 d8a5 d4b3 a5e5 e3e5 g7e5 c3d5 e7e6 d5e7 g8g7 e7c8 f8c8 c2c3 e5f6 f2f4 c8c7 g2g4 b8d7 h3h4 h7h5 g4h5 g6h5 f1e2 d7b6 e2h5 b6d5 d1g1 g7h8 c1b1');
INSERT INTO evaluations VALUES(46039,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(46039,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46039,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46039,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46039,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46039,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46039,36,'cp 0','e4e5 d6e5 d1d8 e8d8 f3e5 d8e7 f2f4 b8c6 e1g1 f6d7 e5d7 c8d7 c1e3 a8c8 f1d1 e7e8 c4e2 a7a6 e2f3 f8e7 c3a4 c6b4 a4b6 c8d8 f3b7 b4c2 e3f2 c2a1 b6d7 e7d6 d1d6 e8e7 f2c5 d8d7 d6d1 e7d8 c5b6 d8e7 b6c5');
INSERT INTO evaluations VALUES(46885,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(46885,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(46885,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(46885,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(46885,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(46885,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(46885,36,'cp 155','c4d5 c8f5 d1a4 c7c6 d5c6 b8c6 e2e3 a7a6 f3e5 b7b5 e5c6 d8c8 a4d1 c8c6 a2a4 a8b8 a4b5 a6b5 b1c3 b5b4 c3b5 b8b5 d1a4 b5b6 a4c6 b6c6 f1b5 e8d7 f2f3 d7c7 b5c6 c7c6 e3e4 f5d7 d4d5 c6b5 a1a8 e7e6 d5d6 b5c6 c1e3 c6d6 e1f2 e6e5 h1d1 d6c6 a8a6 c6b5 a6f6 g7f6 d1d7');
INSERT INTO evaluations VALUES(47030,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(47030,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(47030,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(47030,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(47030,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(47030,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(47030,36,'cp 10','g7g6 g1f3 f8g7 f1d3 c7c5 c2c3 e8g8 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 a2a4 b7b6 d2c4 b8a6 f1e1 a6c5 c1e3 a8b8 d1c2 f6h5 f3d2 f7f5 f2f3 c8a6 d3f1 b8c8 a1d1 f5e4 d2e4 c5e4 c2e4 h5f6 e4h4 a6c4 f1c4 e5e4 c4e2 e4f3 e2f3');
INSERT INTO evaluations VALUES(46716,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46716,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(46716,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46716,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46716,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(46716,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(46716,36,'cp 19','g1f3 f8g7 f1b5 c7c6 b5c4 d6d5 e4d5 c6d5 c4b3 b8c6 e1g1 d8c7 f1e1 e8g8 c2c3 b7b5 f3e5 b5b4 e5c6 c7c6 c3b4 c6b6 d2f3 b6b4 f3e5 c8g4 f2f3 g4e6 c1e3 b4b7 e5d3 f6d7 a1c1 f8c8 e3f2 c8c1 d1c1 b7b6 d3c5 d7c5 d4c5');
INSERT INTO evaluations VALUES(46714,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46714,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46714,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46714,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(46714,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(46714,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46714,36,'cp 13','f8g7 f1d3 c7c5 c2c3 e8g8 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 a2a4 b8a6 d2c4 b7b6 f1e1 a6c5 c1e3 c8a6 a1c1 f6e4 d3e4 c5e4 e3b6 d8b8 e1e4 a6c4 e4c4 b8b6 c1c2 e5e4 c4e4 a8c8 e4c4 c8c4 c2c4 f8b8 g2g3 b6b2 g1g2 b2b3 d1b3 b8b3 c4c7 b3b4 c7c8 g7f8');
INSERT INTO evaluations VALUES(41390,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(41390,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(41390,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(41390,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(41390,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41390,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(41390,36,'cp 41','f1d3 c7c5 c2c3 e8g8 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 d2b1 b8a6 b1c3 c8d7 a2a4 a8c8 h2h3 a6c5 c1e3 f6h5 f1e1 c5d3 d1d3 f7f5 d3d1 c8c4 a1c1 f5f4 e3d2 h5f6 b2b3 c4c8 c3b5 d8b6 b3b4 a5b4 d2b4 c8c1 d1c1 f8c8 c1a3');
INSERT INTO evaluations VALUES(41381,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(41381,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41381,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(41381,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(41381,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(41381,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(41381,36,'cp 31','c7c5 c2c3 c5d4 c3d4 e7e5 d4d5 e8g8 e1g1 a7a5 d2b1 b8d7 b1c3 d7c5 a2a4 f6h5 c1g5 f7f6 g5e3 c5d3 d1d3 f6f5 a1c1 c8d7 f1e1 a8c8 e4f5 g6f5 f3g5 d8e7 c3b5 e5e4 d3d1 c8c1 e3c1 h5f6 h2h3 h7h6 g5e6 d7e6 d5e6');
INSERT INTO evaluations VALUES(50204,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(50204,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(50204,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50204,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50204,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50204,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50204,36,'cp 1','f8g7 c1b2 e8g8 e2e3 b7b6 c2c4 c7c5 f1e2 c5d4 f3d4 c8b7 e1g1 d7d5 b1d2 b8a6 a1c1 a8c8 c4d5 f6d5 c1c8 d8c8 d1a1 a6c5 f1c1 c8a8 d4f3 g7b2 a1b2 f8c8 f3d4 d5b4 b2a3 a7a5 d2c4 c5d7 e2g4 e7e6 c4d6 c8c1 a3c1 b7g2');
INSERT INTO evaluations VALUES(55926,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(55926,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(55926,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(55926,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(55926,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(55926,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(55926,36,'cp -27','b2b3 f8e7 c1b2 g8f6 g1f3 c7c5 f1b5 b8d7 b5d3 b7b6 c2c4 c8b7 e1g1 e8g8 b1c3 d5d4 c3e2 d8c7 e3d4 c5d4 b2d4 b7f3 g2f3 a8d8 d4e3 f6h5 d3c2 e7d6 d1e1 b6b5 e1h4 g7g6 f4f5 e6f5 c2f5 b5c4');
INSERT INTO evaluations VALUES(56359,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(56359,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(56359,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(56359,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(56359,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(56359,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(56359,36,'cp 175','f4e5 b8c6 e5g3 h7h5 h2h3 g8f6 g3h4 f8e7 e2e3 f6e4 h4e7 d8e7 b1d2 e4d6 g1e2 c8e6 e2f4 g7g5 f4e6 e7e6 c2c3 f7f5 h3h4 f5f4 d1e2 e8c8 h4g5 f4e3 e2e3 e6e3 f2e3 d8g8 d2f3 d6e4 f1e2 e4g5 f3g5 g8g5 e1f2 c6e7 e2f3 h5h4 a1e1 c8b8 e3e4');
INSERT INTO evaluations VALUES(56299,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(56299,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(56299,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(56299,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(56299,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(56299,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(56299,36,'cp 154','b8c6 e5g3 h7h5 h2h3 g8f6 g3h4 f8e7 e2e3 f6e4 h4e7 d8e7 b1d2 e4d6 g1e2 c8e6 e2f4 g7g5 f4e6 e7e6 c2c3 e8c8 d2b3 f7f5 d1c2 f5f4 e1c1 f4e3 d1e1 e6f5 f2e3 f5c2 c1c2 c6e7 f1d3 e7f5 h1f1 d8f8 b3d2 g5g4 h3g4 h5g4 c2b1');
INSERT INTO evaluations VALUES(7231,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(7231,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(7231,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7231,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(7231,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7231,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(7231,36,'cp 17','b1d2 c8g4 d2c4 f6d7 h2h3 g4h5 c4a5 a8b8 c1e3 c6c5 g2g4 h5g6 a5c4 b7b5 c4d6 c7d6 a2a4 b5b4 f3d2 f7f6 b2b3 e8g8 f2f4 e5f4 e3f4 d7e5 f4e5 f6e5 f1f8 d8f8 d1e2 f8e7 d2c4 b8f8 c4e3 h7h5 a1f1 e7g5 e3c4 f8f1 g1f1');
INSERT INTO evaluations VALUES(12731,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(12731,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(12731,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(12731,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(12731,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(12731,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(12731,36,'cp 45','d8a5 d4b3 a5c7 e1g1 f8g7 f1e1 e8g8 c1g5 a7a6 c3d5 f6d5 e4d5 c6e5 c4f1 d7d6 c2c4 h7h6 g5h4 e7e6 a1c1 e6d5 d1d5 c8e6 d5d2 e5c6 h2h3 f8e8 h4g3 c7b6 g3d6 c6d4 d6e5 d4b3 a2b3 g7e5 e1e5 a8d8 d2h6 b6b3 e5e1 b3b2 c1b1 b2g7 h6g7 g8g7 b1b7');
INSERT INTO evaluations VALUES(17221,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(17221,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(17221,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(17221,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(17221,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17221,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(17221,36,'cp 20','c6e5 e2e4 e7e6 f2f4 e5g6 b1c3 f8b4 f1d3 e6d5 c4d5 e8g8 g1f3 f8e8 f4f5 g6e5 f3e5 e8e5 e1g1 d7d6 d1f3 c7c6 d5c6 b7c6 c1f4 e5e8 g1h1 c8b7 a1d1 d8e7 h2h3 b4c3 b2c3 c6c5 f1e1 b7e4 d3e4 f6e4 f5f6 e7e6 f6g7 a8d8');
INSERT INTO evaluations VALUES(17312,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(17312,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(17312,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(17312,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(17312,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(17312,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(17312,36,'cp 32','e2e4 e7e6 f2f4 e5g6 b1c3 e6d5 c4d5 f8b4 f1d3 d8e7 g1e2 b4c5 e2g3 d7d6 b2b4 c5b6 c3a4 f6e4 d3e4 f7f5 a4b6 a7b6 e1g1 f5e4 f4f5 g6e5 g3e4 e8g8 c1g5 e7f7 f5f6 f7g6 f6g7 f8f1 d1f1 c8f5 e4g3 f5d3 f1f4 g6g7 a1c1 a8f8 f4d2 f8f7 g5f4 d3g6 c1f1');
INSERT INTO evaluations VALUES(18349,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(18349,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(18349,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(18349,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(18349,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(18349,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(18349,36,'cp 68','c6a5 c4e2 e7e6 d2d4 c8b7 e2d3 c7c5 c2c3 g8f6 b1d2 f8e7 e1g1 e8g8 f1e1 a8c8 d2f1 c5d4 c3d4 e7b4 f1d2 d7d5 e4e5 f6e4 e1e2 d8d7 d2e4 d5e4 d3e4 h7h6 h2h4 b4e7 d1d3 f8d8 c1e3 a5c4 a2a3 c4e3 e4b7 d7b7 d3e3');
INSERT INTO evaluations VALUES(19019,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(19019,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(19019,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(19019,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(19019,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(19019,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(19019,36,'cp 0','g7g6 e1g1 f8g7 c2c4 d7d6 b1c3 g8e7 a1b1 c8e6 d2d3 e8g8 f3g5 e6c8 g5f3 c8e6');
INSERT INTO evaluations VALUES(22181,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(22181,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(22181,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(22181,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(22181,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(22181,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(22181,36,'cp 154','d4e5 g8e7 a2a3 c8e6 b1c3 d5e4 d1d8 a8d8 c3e4 e7c6 c1f4 d8d5 f1c4 d5d4 c4d3 e6c4 e1c1 c4d3 c2d3 d4d5 h1e1 f8c5 f4d2 c5b6 d2c3 e8e7 f2f4 h8d8 c1c2 a7a5 h2h3 h7h5 g2g3 c6d4 c3d4 b6d4 e4c3 d5c5 e1e4 e7f8 c2b1 b7b5 g3g4 g7g6 g4h5');
INSERT INTO evaluations VALUES(22178,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(22178,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(22178,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(22178,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(22178,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(22178,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(22178,36,'cp 156','g8e7 b1c3 c8e6 d1d4 d5e4 d4e4 e7d5 c1d2 d5c3 d2c3 d8d5 f1d3 f8c5 e4d5 e6d5 a2a4 a7a5 f2f3 e8e7 e1c1 b7b6 h2h4 a8d8 h4h5 d5e6 b2b3 c5e3 c3d2 e3d2 d1d2 g7g5 d3e4 d8d2 c1d2 h8d8 d2e2 h7h6 e2e3 d8b8 e4d3');
INSERT INTO evaluations VALUES(24880,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24880,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(24880,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(24880,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(24880,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24880,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(24880,36,'cp 12','a2a3 d8d7 d1d2 e8c8 e1c1 h7h6 g5e3 e7e6 f3e5 c6e5 d4e5 f6e4 c3e4 d5e4 d2d7 d8d7 d1d7 c8d7 f2f3 g4f5 h2h4 h6h5 g2g3 f8e7 f1g2 e4f3 e2f3 d7c8 g2f1 a7a6 f1e2 h8d8 h1d1 d8d5 d1d5 e6d5 e3g5 e7g5 h4g5 g7g6 b2b4 c8d7 c1d2 c7c5 b4c5 d7c6');
INSERT INTO evaluations VALUES(35468,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(35468,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35468,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35468,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35468,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35468,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(35468,36,'cp 0','b8d7 e1e2 c7c6 f3e1 a7a5 a2a4 d8c7 e1d3 d7c5 f2f3 c5d3 c2d3 f8c5 c1e3 c5e3 e2e3 c8d7 g2g3 g8e7 b1d2 e7c8 f3f4 c8d6 f4e5 f6e5 h1f1 d7g4 f1f2 a8d8 a1f1 h7h5 h2h4 g4h3 f1a1 c7b6 c4a2 h3g4 a2c4');
INSERT INTO evaluations VALUES(43554,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43554,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43554,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43554,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43554,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(43554,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43554,36,'cp 0','f8e7 c1e3 e8g8 e1g1 b8c6 h2h3 c8d7 c4d3 f8e8 f2f4 c6d4 e3d4 e6e5 d4e3 e5f4 e3f4 d8b6 g1h2 b6b2 c3d5 f6d5 e4d5 b7b5 a1b1 b2c3 f4d2 c3e5 d2f4');
INSERT INTO evaluations VALUES(39751,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39751,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39751,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39751,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39751,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39751,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39751,36,'cp 0','e1g1 b8c6 c1e3 e8g8 d1e2 c8d7 d4c6 d7c6 a4a5 f6d7 f1d1 a8c8 e2d2 d8c7 c4f1 c7b8 f2f3 f8d8 g1h1 h7h6 e3f2 d7f6 f2d4 f6d7');
INSERT INTO evaluations VALUES(47167,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(47167,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(47167,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(47167,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(47167,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(47167,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(47167,36,'cp -19','c7c5 d4c5 e7e6 a2a3 f8c5 c1b2 b8c6 b3b4 c5b6 b1d2 e6e5 b4b5 c6e7 g1f3 e5e4 f3d4 c8g4 f1e2 g4e2 d4e2 e7f5 e2g3 f5g3 h2g3 a8c8 d2b3 b6a5 b3a5 d8a5 d1d2 a5b5 b2f6 g7f6 a3a4 b5c4 h1h5 c4c2 d2c2 c8c2 h5d5 e8e7 a1d1 h8c8 d5d7 e7e6');
INSERT INTO evaluations VALUES(49151,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(49151,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(49151,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(49151,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(49151,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(49151,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(49151,36,'cp -23','c7c5 d4c5 b8c6 g1f3 f8c5 e1g1 a7a6 b5c6 b7c6 c3a4 c5e7 c2c4 e8g8 c1f4 c8b7 a1c1 c6c5 c4d5 b7d5 b2b3 d8c7 f1e1 d5c6 a4c3 f8d8 d1e2 d7f8 c3e4 f8g6 f4g3 h7h5 h2h4 c6e4 e2e4 d8d5 e4e2 c7b7 c1c4 b7b5 e2c2 a8d8 e1e3 a6a5 e3c3 b5b7 f3g5');
INSERT INTO evaluations VALUES(43981,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(43981,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(43981,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(43981,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43981,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(43981,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(43981,36,'cp -28','g1f3 b8c6 d4c5 f8c5 e1g1 e8g8 c1f4 a7a6 b5c6 b7c6 c3a4 c5e7 c2c4 c8b7 a1c1 c6c5 c4d5 b7d5 b2b3 d8c7 f1e1 d5c6 a4c3 c7b7 d1e2 a6a5 f3d2 f8d8 c3e4 a5a4 h2h4 a4b3 a2b3 c6e4 d2e4 b7b3 f4g5 e7g5 h4g5 b3d5 f2f4 d5d4 e2f2 a8b8 f2d4 c5d4');
INSERT INTO evaluations VALUES(50076,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50076,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(50076,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(50076,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(50076,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(50076,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(50076,36,'cp 32','e7e5 d2d4 e5d4 d1d4 b8c6 d4e3 d8e7 f1g2 e7e3 c1e3 f8b4 e3d2 c6e5 a2a3 b4d6 b2b3 c7c6 g1f3 e5f3 e2f3 e8f7 e1g1 d6c7 f1e1 h8e8 e1e8 f6e8 g3g4 g7g6 g4g5 d7d6 f3f4 c8e6 h2h4 c7a5 b3b4 a5b6 g2f1 f7e7 a1d1 e8c7 a3a4 a8b8 a4a5');
INSERT INTO evaluations VALUES(50842,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(50842,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(50842,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(50842,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(50842,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(50842,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50842,36,'cp 50','g2g3 c8f5 c2c4 e7e6 f1g2 g8f6 c4d5 e6d5 d1b3 b8c6 c1g5 h7h6 g5f6 d8f6 e1g1 e8c8 e2e3 c6a5 b3c3 f6b6 b2b3 f7f6 a2a3 c8b8 b1d2 a5c6 b3b4 f8d6 f1c1 g7g5 d2b3 h6h5 h2h4 f5g4 b3c5 g5h4');
INSERT INTO evaluations VALUES(51441,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(51441,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(51441,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(51441,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(51441,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(51441,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(51441,36,'cp 180','g1f3 d7d6 d2d4 e5d4 d1d4 d8d7 c1f4 c8b7 b1c3 b8c6 d4d2 e8c8 f3g5 g8h6 e1c1 c8b8 h1e1 f7f6 g5e6 d8e8 f4h6 g7h6 e6f8 h8f8 f2f4 d7g7 g2g3 h6h5 c4e2 g7h6 c1b1 h5h4 g3g4 c6e7 a2a3 h4h3 e2b5 e8d8 d2f2 f8g8 e1g1');
INSERT INTO evaluations VALUES(54187,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(54187,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(54187,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(54187,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(54187,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(54187,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(54187,36,'cp -9','g8f6 e2e3 a7a6 d4c5 f8c5 f1d3 b8d7 e3e4 e6e5 f4g5 d8b6 e1g1 d5e4 c3e4 f6e4 d3e4 h7h6 g5h4 e8g8 e4f5 b6c7 c2c3 f8e8 f1e1 d7f8 f5e4 f7f6 h4g3 e8d8 d1e2 c7f7 b2b4 c5a7 h2h3 c8e6 a1d1 f8g6 d1d8 a8d8 e2c2');
INSERT INTO evaluations VALUES(56719,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(56719,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(56719,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56719,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(56719,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(56719,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(56719,36,'cp 17','f8e7 g5c1 c8e6 f3g5 e7g5 d1h5 e8d7 c1g5 g8f6 h5e2 h7h6 g5d2 b8c6 b1c3 d8e7 e1c1 a8d8 e2e1 c6d4 c1b1 d7c8 d2e3 c7c5 b2b3 c8b8 f2f3 h8f8 c3a4 a7a6 f1d3 f6h5 c2c3 d4c6 e1d2 h5f4 d3c2 b8c7 a4b2 b7b5 h2h4 e7f7 b1a1 c7b7 b2d3');
INSERT INTO evaluations VALUES(38206,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38206,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38206,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(38206,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38206,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38206,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(38206,36,'cp 33','g5e3 c8e6 d1e2 b8d7 f3g5 e7g5 e2h5 g7g6 h5g5 d8g5 e3g5 h7h6 g5e3 g6g5 b1c3 g8f6 f2f3 a7a5 e1c1 e8c8 f1e2 d8g8 h1f1 f6h5 g2g3 b7b6 b2b3 h5f6 c3d5 e6d5 e4d5 c8b7 a2a3');
INSERT INTO evaluations VALUES(786,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(786,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(786,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(786,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(786,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(786,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(786,36,'cp -102','e7f5 b1c3 f5d4 f3d4 b6d4 c3b5 d4d1 f1d1 a8b8 c1e3 a7a6 b5d6 f8d6 e5d6 e8d7 a1c1 d7d6 e3c5 d6c7 c5e3 h8d8 e3f4 d8d6 b2b4 c7d7 b4b5 a6b5 e2b5 b8a8 a2a4 d7e7 f4d6 e7d6 c1b1 c8d7 f2f4 a8a5 b5d3 g7g6 d3b5');
INSERT INTO evaluations VALUES(10135,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(10135,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10135,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(10135,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(10135,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(10135,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10135,36,'cp 12','d2d3 d7d6 b1c3 c6d4 e2d4 c5d4 c3e2 d4c5 h2h3 c7c6 c2c3 e8g8 c1g5 h7h6 g5f6 d8f6 f3f6 g7f6 g2g4 b7b5 c4b3 a7a5 d3d4 e5d4 e2d4 c5d4 c3d4 f8e8 f2f3 d6d5 e1g1 d5e4 f3e4 e8e4 f1f6 c8e6 b3e6 f7e6 f6h6 e4e2 g4g5 a8d8 h6g6 g8f7');
INSERT INTO evaluations VALUES(15102,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15102,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(15102,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(15102,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(15102,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15102,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(15102,36,'cp -6','e5d4 f3d4 c6d4 d1d4 c8e6 c4d5 f6d5 c1d2 c7c6 f1c4 d5c7 c4e6 c7e6 d4d8 a8d8 e1e2 f7f5 a1c1 e8f7 h1d1 g7g5 c3b1 f8e7 d2c3 d8d1 c1d1 h8d8 d1c1 c6c5 a2a4 h7h5 g2g3 b7b6 b1d2 e7f6 c3f6 f7f6 h2h4 g5g4 f2f4');
INSERT INTO evaluations VALUES(16797,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(16797,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(16797,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(16797,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(16797,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(16797,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(16797,36,'cp 59','f8c5 c3e4 d8e7 e4c5 e7c5 e2e3 g4e5 c1d2 e5f3 d1f3 d7d6 f1e2 c8e6 a1c1 e8g8 e1g1 a7a5 d2c3 f7f6 f3e4 f8e8 f1d1 a5a4 h2h4 c5f5 e4f5 e6f5 d1d5 c6e5 c4c5 f5e6 d5d1 d6c5 c3e5 f6e5 c1c5 c7c6 c5e5');
INSERT INTO evaluations VALUES(19957,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(19957,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19957,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19957,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19957,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(19957,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(19957,36,'cp -5','b1c3 g7g6 g1f3 f8g7 e1g1 g8f6 d2d3 c8g4 h2h3 g4f3 d1f3 e8g8 c1g5 c6d4 f3d1 e7e6 c4a2 a8b8 c3e2 d4e2 d1e2 b7b5 c2c3 d8c7 a1d1 b5b4 c3b4 c5b4 g5f4 f8d8 d3d4 c7b7 a2b1 b4a3 b2a3 b7b2 b1c2 b2a2 e4e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(19953,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(19953,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(19953,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(19953,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(19953,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(19953,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(19953,36,'cp -9','g8f6 c4a2 g7g6 d2d3 f8g7 f2f4 c8g4 g1e2 f6h5 e1g1 c6d4 c1e3 e8g8 f1f2 g7h6 e3c1 e7e5 f4e5 h6c1 d1c1 d6e5 c1d2 g4e2 c3e2 d4e2 d2e2 h5f4 e2e3 d8b6 a2b3 a8d8 f2f1 f4d3 f1f7 f8f7 b3f7 g8f7 c2d3');
INSERT INTO evaluations VALUES(13259,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13259,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13259,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13259,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13259,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13259,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13259,36,'cp -3','c4a2 e7e6 g1e2 f8e7 d2d4 c5d4 e2d4 c6d4 d1d4 e8g8 d4d3 c8d7 c1f4 d8b6 e1g1 d7c6 f4e3 b6c7 e3d4 f8d8 a1e1 a7a6 f2f4 b7b5 g1h1 a8c8 d3e3 c7b7 b2b4 h7h6 a2b3 f6h7 e3f2 h7f6');
INSERT INTO evaluations VALUES(20793,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(20793,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(20793,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(20793,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(20793,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(20793,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(20793,36,'cp 67','f1b5 c8d7 e1g1 c6e5 f3e5 d7b5 d1h5 g7g6 e5g6 h7g6 h5h8 b5f1 h8g8 f1d3 g8e6 d8e7 e6e7 f8e7 b1c3 a8d8 h2h4 e7b4 c1d2 e8d7 c3a4 b4d6 a1e1 d8c8 a4c3 d6b4 f2f3 b4c3 d2c3 c8h8 g2g3 d3c4 a2a3 h8e8 e1e8 d7e8 g1f2 e8f7 a3a4 b7b6 f2g2 c4d3 c3d2');
INSERT INTO evaluations VALUES(20788,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(20788,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(20788,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(20788,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(20788,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(20788,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(20788,36,'cp 73','c8d7 e1g1 c6e5 f3e5 d7b5 d1h5 g7g6 e5g6 h7g6 h5h8 b5f1 h8g8 f1d3 g8e6 d8e7 e6e7 f8e7 b1c3 a8d8 c1d2 g6g5 h2h3 e7b4 a1c1 b4a5 d2e3 e8f7 c3a4 d3b5 c1c5 a7a6 a4c3 a5c3 c5c3 d8e8 c3c7 e8e7 c7e7 f7e7 f2f4 g5f4 e3f4 a6a5 b2b3 e7e6 f4c7 a5a4 b3a4 b5a4 g1f2 b7b5 c7d8 b5b4 f2g3 e6f7 h3h4 a4c2');
INSERT INTO evaluations VALUES(21457,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(21457,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21457,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(21457,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(21457,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(21457,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(21457,36,'cp 44','d2d4 f8e7 b2b3 g8f6 c1b2 e8g8 e1g1 b7b6 c2c4 c8b7 b1d2 c6a5 c4d5 e6d5 a1c1 e7d6 e2e3 f8e8 f3e5 d8e7 b2c3 c7c5 d2f3 c5d4 c3d4 a8c8 c1c8 e8c8 d1a1 f6e4 e5d3 f7f6 f1c1 c8c1 a1c1 a5c6 d4b2 c6b4 d3f4 d6f4 e3f4 b4d3');
INSERT INTO evaluations VALUES(21940,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(21940,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21940,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(21940,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(21940,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(21940,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(21940,36,'cp 93','e2e4 e5d4 f3d4 f8e7 f1b5 c8d7 d4c6 b7c6 b5d3 e7f6 e1g1 g8e7 c1e3 f6c3 b2c3 e8g8 d1d2 e7g6 f2f4 f8e8 e3f2 c6c5 f2g3 d7c6 a1e1 g6f8 c3c4 a8b8 e4e5 d6e5 f4e5 f8e6 d2f2 e8f8 e1e3 b8b2 h2h3 d8d4 g1h2 b2a2 g3h4 g8h8 h4e7');
INSERT INTO evaluations VALUES(25152,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(25152,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(25152,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(25152,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(25152,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(25152,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(25152,36,'cp -9','f8d6 b1d2 h7h6 f1b5 e8g8 e1g1 d6f4 e3f4 c6e7 f3e5 b7b6 f1e1 c7c5 d2f1 a7a6 b5d3 f5d3 d1d3 f6d7 e5d7 d8d7 d3d2 d7c7 g2g3 a6a5 f1e3 b6b5 a2a3 c5c4 e3g4 b5b4 a3b4 a5b4 c3b4 c7b6 d2c3 a8c8');
INSERT INTO evaluations VALUES(24104,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(24104,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(24104,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(24104,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(24104,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(24104,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(24104,36,'cp -17','b1d2 h7h6 f4h2 e8g8 f1e2 d8e7 e1g1 e6e5 h2e5 c6e5 d4e5 d6e5 f3e5 e7e5 d1b3 c7c5 f1d1 b7b6 d2f3 e5e7 c3c4 d5c4 b3c4 f8d8 c4c3 d8d1 a1d1 a8e8 a2a3 f6e4 c3c4 e8d8 c4c1 d8d1 c1d1 e7f6 d1c1 e4d6 f3d2 f6g6 g1h2 f5d3');
INSERT INTO evaluations VALUES(29160,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(29160,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(29160,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(29160,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(29160,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(29160,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29160,36,'cp 8','c4d5 f6d5 h2h3 g4f3 f1b5 b8c6 d1f3 e7e6 e1g1 a7a6 b5c6 b7c6 c1e3 f8e7 f1c1 a8b8 c1c2 e8g8 b1d2 a6a5 d2c4 e7f6 c4e5 f6e5 d4e5 d5e3 f3e3 b8b4 b2b3 a5a4 b3a4 b4a4 c2c6 d8a8 c6c2 h7h6 g1h2 f8d8 a1b1 a4a2 c2a2 a8a2 b1b7 a2a4');
INSERT INTO evaluations VALUES(29275,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29275,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(29275,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29275,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29275,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(29275,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29275,36,'cp 2','b8d7 d1b3 a8b8 c4d5 c6d5 e1g1 f8e7 h2h3 g4h5 c1d2 f6e4 f1d1 e4d2 f3d2 h5e2 c3e2 d8b6 e2f4 e8g8 d1c1 f8c8 f4d3 b6a6 c1c2 h7h6 a1c1 c8c2 c1c2 d7b6 c2c1 b8a8 d2f3 a8c8 c1c8 b6c8 f3e5 e7d6 b3c2 c8e7');
INSERT INTO evaluations VALUES(24501,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24501,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24501,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24501,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24501,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24501,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24501,36,'cp 12','d1b3 a8b8 c4d5 c6d5 e1g1 f8e7 h2h3 g4h5 c1d2 f6e4 f1d1 e4d2 f3d2 h5e2 c3e2 d8b6 e2c1 e8g8 c1d3 f8c8 a2a4 c8c7 b3b6 d7b6 d1c1 b8c8 c1c7 c8c7 a4a5 b6c4 d2c4 c7c4 a5a6 b7b6 g1f1 f7f6 f1e2 g8f7');
INSERT INTO evaluations VALUES(31082,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31082,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31082,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31082,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(31082,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(31082,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31082,36,'cp 12','f5d3 d1d3 g8f6 g1f3 e7e6 e1g1 f8d6 f4d6 d8d6 d3b3 d6c7 c2c4 e8g8 h2h3 b8d7 f1c1 h7h6 b1c3 d5c4 b3c4 e6e5 c1c2 f8e8 a1c1 a7a6 c1d1 e5d4 f3d4 a8d8 b2b4 d7e5 c4c5 d8d7 a2a3 g7g6 c3a4 e8d8 c2c1');
INSERT INTO evaluations VALUES(31113,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31113,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(31113,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31113,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(31113,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(31113,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31113,36,'cp 19','c2d3 e7e6 g1f3 d8b6 d1c2 g8f6 b1c3 f6h5 f4e5 b8d7 h2h4 h5f6 e1g1 f8e7 e5f4 e8g8 e3e4 a8c8 c2e2 f8e8 a1d1 f6g4 g2g3 h7h5 g1g2 b6a6 a2a3 f7f6 e4e5 c6c5 d4c5 d7c5 e5f6 e7f6 d3d4 a6e2');
INSERT INTO evaluations VALUES(31176,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(31176,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(31176,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31176,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(31176,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31176,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(31176,36,'cp 11','d2d4 e7e6 f1d3 f5d3 c2d3 g8e7 g1f3 e7f5 e1g1 b8d7 a1b1 h7h5 g2g3 f8e7 h2h4 a7a5 c3e2 d7f8 c1g5 f8g6 g5e7 d8e7 a2a3 e8g8 d1d2 b7b6 b2b4 a5b4 a3b4 f8c8 f1c1 a8a3 b1b2 g8f8 b4b5 c6c5 d4c5 c8c5');
INSERT INTO evaluations VALUES(34246,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(34246,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(34246,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(34246,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(34246,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(34246,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(34246,36,'cp 43','c5c7 c3d5 c7d6 e3f4 e7e5 f4g3 g8f6 d5f6 d6f6 e2e3 b8c6 a2a3 d7d6 g1e2 f6g6 e2c3 f8e7 f1d3 f7f5 d3c4 c8e6 c4e6 g6e6 f2f3 h7h5 d1d3 h5h4 g3f2 h4h3 g2g3 e8g8 e1c1 a8c8 h1g1 e5e4 f3e4 c6e5 d3d5 e6d5 c3d5');
INSERT INTO evaluations VALUES(35352,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(35352,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(35352,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(35352,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(35352,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(35352,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(35352,36,'cp 46','c3d5 c7d6 c2c4 e7e6 d5c3 b8c6 g1f3 g8f6 a2a3 b7b6 c3b5 d6b8 g2g3 f8c5 e3f4 e6e5 f4d2 c8b7 b2b4 c5e7 f1g2 e8g8 e1g1 f8d8 b5c3 d7d5 c4d5 f6d5 c3d5 d8d5 d1b3 c6d4 f3d4 d5d4 g2b7 b8b7 d2e3 d4d7 f1d1 h7h6 d1d7 b7d7 a1d1 d7g4 b4b5 g4e2');
INSERT INTO evaluations VALUES(35044,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(35044,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(35044,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(35044,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(35044,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(35044,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(35044,36,'cp 179','f4c7 h4e7 e2e3 d7d5 c7b8 a8b8 b1d2 g8f6 f1d3 e7b4 d1b1 b7b5 c2c3 b4a5 b1d1 f8d6 a2a4 b5a4 a1a4 a5c7 d1a1 a7a5 a4a5 b8b2 g1f3 e8g8 a1b2 c7a5 e1g1 c8d7 c3c4 f8b8 b2a1 a5a1 f1a1 d5c4 d2c4 d6f8 a1a7 g8h8');
INSERT INTO evaluations VALUES(42635,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(42635,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42635,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42635,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(42635,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(42635,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42635,36,'cp 0','g1f3 d7d6 g2g3 h7h6 f1g2 e8g8 e1g1 f8e8 f1e1 a7a5 d3d4 c5a7 h2h3 c8d7 b1a3 e5d4 f3d4 b8c6 a3c2 a5a4 g1h2 c6a5 c2a3 a7c5 a3c2');
INSERT INTO evaluations VALUES(42632,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42632,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(42632,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42632,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42632,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42632,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42632,36,'cp 0','d7d6 g2g3 a7a5 f1g2 b8c6 e1g1 e8g8 a2a4 h7h6 b1a3 f8e8 f1e1 f6g4 e1f1');
INSERT INTO evaluations VALUES(44140,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(44140,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(44140,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(44140,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(44140,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(44140,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(44140,36,'cp 39','d4c5 d8c7 g1f3 f8c5 f1d3 b8c6 d1e2 f7f6 c3c4 d5b4 e5f6 b4d3 e2d3 c6e5 f3e5 c7e5 c1e3 c5e3 f2e3 e5f6 b1c3 e8g8 e1c1 f6e5 d3d4 e5d4 e3d4 b7b6 d1d2 c8b7 h1e1 f8f6 b2b3 a8f8 c3b5 a7a6 b5d6 b7c6 a2a4 h7h5 g2g3 f6f2 e1e2 f2f1 c1c2 c6f3 e2e5 f3g4 c2c3 a6a5 d6e4 f8f3 c3b2');
INSERT INTO evaluations VALUES(47423,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(47423,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(47423,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(47423,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(47423,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(47423,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(47423,36,'cp -127','a7a6 f1c4 b8c6 b1c3 c6d4 f3d1 b7b5 c4e2 c8b7 g1f3 c7c5 e1g1 e7e6 d2d3 f8e7 c1f4 e8g8 f3d4 c5d4 c3b1 a8c8 b1d2 f6d5 f4e5 e7d6 e5d6 d8d6 e2f3 d6c7 a1c1 f8d8 f1e1 h7h6 d1e2 d5f4 e2e5 c7e5 e1e5 b7f3 d2f3');
INSERT INTO evaluations VALUES(49851,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49851,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(49851,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(49851,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(49851,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49851,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(49851,36,'cp -8','d7d6 b1c3 b8c6 d4d2 a7a5 e2e3 d6e5 d2d8 e8d8 h2h3 g4f6 g1f3 f8d6 f1e2 a5a4 e1g1 d8e7 e3e4 h8d8 c1g5 e7f8 c3d5 a4a3 b2b3 f6d5 c4d5 c6e7 f3d2 f7f5 d2c4 f5e4 f1c1 h7h6 c4d6 c7d6 g5e7 f8e7 c1c7 d8d7');
INSERT INTO evaluations VALUES(47997,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(47997,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47997,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(47997,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(47997,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(47997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47997,36,'cp -12','b1c3 b8c6 d4d2 a7a5 e2e3 d6e5 d2d8 e8d8 h2h3 g4f6 g1f3 c8f5 c1d2 d8c8 e1c1 h7h6 g2g4 f5e6 d2e1 a5a4 c3d5 f8d6 d5f6 g7f6 a2a3 f6f5 g4f5 e6f5 f3d2 c6a5 h3h4 f5e6 f1h3 e6h3 h1h3 d6e7 h3h1');
INSERT INTO evaluations VALUES(51016,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(51016,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(51016,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(51016,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(51016,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(51016,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(51016,36,'cp 104','f1d3 d7d5 b1c3 d5e4 c3e4 g8f6 e4f6 d8f6 g1f3 h7h6 e1g1 f8d6 f1e1 b8d7 d1e2 f6e7 f3e5 e8g8 e5c4 e7h4 c4d6 c7d6 c1e3 d7f6 c2c4 e6e5 f2f3 e5d4 e3f2 h4g5 a1d1 c8f5 d3f5 g5f5 e2d3 f5d3 d1d3 d6d5 c4c5 f8e8 e1e8 f6e8 d3b3 e8c7 b3b7 c7e6 c5c6 d4d3 c6c7');
INSERT INTO evaluations VALUES(51532,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(51532,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(51532,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(51532,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(51532,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(51532,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(51532,36,'cp 38','c8b7 g1f3 e7e6 e2e3 b5b4 a2a3 a7a5 a3b4 a5b4 a1a8 b7a8 c2c4 b4c3 b1c3 g8f6 f1e2 f8e7 d1a4 a8b7 e1g1 e8g8 b2b4 f6h5 f4g3 d7d6 a4a7 d8c8 g3h4 e7h4 f3h4 h5f6 f1a1 b8d7 a7a3 d7b6 a3b2 c8d7 h4f3 f8a8 a1a8 b7a8 f3d2 f6d5 c3d5 a8d5 b4b5');
INSERT INTO evaluations VALUES(30845,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(30845,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(30845,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(30845,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(30845,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(30845,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(30845,36,'cp 56','g1f3 g8f6 e2e3 b5b4 a2a3 a7a5 a3b4 a5b4 a1a8 b7a8 f1e2 e7e6 c2c4 b4c3 b1c3 f8e7 d1a4 a8b7 e1g1 e8g8 b2b4 d7d6 a4a7 d8c8 f1a1 h7h6 h2h3 b8d7 f3d2 c8a8 e3e4 f8c8 b4b5 c8b8 a7a5 a8a5 a1a5');
INSERT INTO evaluations VALUES(52643,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(52643,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(52643,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(52643,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(52643,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(52643,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(52643,36,'cp 104','g1f3 f8g7 b1c3 b8c6 h2h3 e7e6 f1b5 g8e7 e1g1 e8g8 f1e1 c8d7 c1e3 c6b4 b5f1 d6d5 e4e5 e7f5 e3g5 f7f6 e5f6 g7f6 g5f6 d8f6 d1d2 a8d8 a1d1 f5h4 f3e5 b4c6 e5g4 f6f4 d2f4 f8f4 f1b5 g8g7 b5c6 d7c6 e1e6 h4f5 f2f3 g7h8 g1h1 f4d4 d1d4 f5d4');
INSERT INTO evaluations VALUES(2355,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(2355,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(2355,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(2355,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(2355,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(2355,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(2355,36,'cp 108','c1d2 b4d6 f2f3 f6d7 f3f4 d7c5 d2e3 f7f6 e1c1 c8d7 e4e5 f6e5 f4e5 h8e8 e3d4 d6e7 f1c4 d8c8 c4f7 e8f8 h1f1 d7g4 d1d2 c5e6 c3e4 e6d4 d2d4 f8d8 d4d8 c8d8 h2h3 g4d7 e5e6 d7e8 f7e8 d8e8 f1f7');
INSERT INTO evaluations VALUES(2356,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(2356,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(2356,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(2356,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(2356,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(2356,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(2356,36,'cp 78','c8e6 a2a3 b4c5 f1e2 f6d7 f2f4 f7f6 e1c1 d8e8 g2g4 a8d8 a3a4 h7h5 f4f5 e6f7 g4g5 d7e5 g5g6 f7c4 d2f4 c4e2 c3e2 e5g4 f4c7 d8d7 d1d7 e8d7 c7g3 h5h4 g3e1 c5e3 c1b1 h4h3 e1g3 h8e8 h1d1 d7c8 e2c3 e8d8 d1d8 c8d8 c3e2 d8e7 a4a5 e3f2 g3f4 g4e3 c2c3');
INSERT INTO evaluations VALUES(4222,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(4222,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(4222,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(4222,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(4222,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(4222,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(4222,36,'cp 34','h2h3 d7d6 g2g4 f6h7 a2a3 a7a6 h1g1 a8b8 h3h4 b7b5 c4b3 g7g6 g4g5 h6h5 d1d2 a6a5 c3d5 a5a4 b3a2 c8e6 g1f1 d8d7 f3h2 e6d5 a2d5 g8g7 d3d4 f7f6 d2c3 c6e7 d4e5 e7d5 e4d5 f6e5 e1c1 f8f1 d1f1 b8f8');
INSERT INTO evaluations VALUES(6501,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(6501,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(6501,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(6501,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(6501,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(6501,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(6501,36,'cp -6','g8f6 f1g2 d7d5 e4d5 f6d5 e1g1 e8g8 f1e1 c8g4 h2h3 g4h5 c2c3 a7a5 b1d2 f7f6 d1b3 c5b6 f3e5 f6e5 b3d5 g8h8 d3d4 d8d5 g2d5 e5d4 d2c4 d4c3 c4b6 c3b2 c1b2 c7b6 a1c1 h5f7 d5c6 b7c6 c1c6 f7a2 c6b6 h8g8 b6b7 f8f7 b7e7 a5a4 e7e8 f7f8 e8f8 g8f8 b2a3 f8g8 f2f3 a2d5 g1f2 a8c8 g3g4 c8c2 e1e2 c2c3 e2e3 c3e3 f2e3 d5b3 e3e2');
INSERT INTO evaluations VALUES(7495,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7495,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7495,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7495,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(7495,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(7495,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(7495,36,'cp 0','d7d5 b1d2 d5e4 d2e4 c6a5 f3e5 a5b3 d1b3 c8e6 b3c2 d8d5 e5f3 a8d8 f3d4 c7c5 d4e6 f7e6 e1g1 d5d3 c2d3 d8d3 e4g5 e6e5 f1e1 e8g8 g1f1 e7d6 g5e4 d6e7');
INSERT INTO evaluations VALUES(7840,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7840,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(7840,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7840,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(7840,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(7840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7840,36,'cp -9','f1d3 b4c3 b2c3 e6e5 f3d2 d7d6 e1g1 e8g8 d2c4 c8e6 c1g5 e6c4 d3c4 h7h6 g5f6 d8f6 f1e1 c6d8 e1e3 a8c8 c4b3 d8e6 g2g3 c8c7 h2h4 e6c5 d1h5 f6e7 h5e2 g7g6 h4h5 g8g7 a1d1 f8c8 h5g6 f7g6 g1g2 c5b3 a2b3 c7c3 e3c3 c8c3');
INSERT INTO evaluations VALUES(7842,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7842,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7842,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7842,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7842,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7842,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7842,36,'cp -5','e6e5 f3d2 b4c3 b2c3 d7d6 e1g1 c8e6 d2c4 e8g8 c1g5 h7h6 g5f6 d8f6 c4d6 b7b6 d1e2 c6a5 c3c4 f8d8 d6f5 e6f5 e4f5 a8c8 a1e1 c8c5 e1b1 a5b7 b1b5 c5b5 c4b5 b7c5 f1d1 c5d3 d1d3 d8d3 c2d3 f6g5 e2e3 g5f5 e3e4 f5e4 d3e4 g8f8 g1f1 f8e7');
INSERT INTO evaluations VALUES(15572,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(15572,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(15572,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(15572,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(15572,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(15572,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(15572,36,'cp 50','d2d4 f8e7 d4d5 c6b8 a2a4 e8g8 f3d2 c7c6 d2c4 b8a6 c4e3 a6c5 f2f3 c6d5 c3d5 f6d5 e3d5 c8e6 e1g1 e7g5 c1g5 d8g5 e2c4 a8c8 d1e2 e6d5 c4d5 c5e6 d5e6 f7e6 f1d1 g5d8 c2c3 d8b6 g1h1 a7a5 d1d3 c8c6 g2g3 b6c5 h1g2 b7b5 a4b5 c5b5');
INSERT INTO evaluations VALUES(18909,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(18909,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(18909,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18909,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18909,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18909,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(18909,36,'cp 27','g8f6 b1c3 f8b4 e1g1 b4c3 b2c3 f6e4 e2d3 d7d5 d4c6 b7c6 c1a3 c6c5 f1e1 e8g8 d3e4 d5e4 a3c5 f8e8 e1e4 c8b7 d1d8 e8d8 e4g4 d8d2 a1c1 e6e5 c5e3 d2d7 g4a4 f7f6 f2f3 a7a5 e3b6 g8f7 a4a5 a8a5 b6a5');
INSERT INTO evaluations VALUES(12486,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(12486,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(12486,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12486,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(12486,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12486,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(12486,36,'cp 22','b1c3 f8b4 e1g1 b4c3 b2c3 f6e4 e2d3 d7d5 d4c6 b7c6 c1a3 c6c5 f1e1 e8g8 d3e4 d5e4 a3c5 f8e8 e1e4 c8b7 d1d8 e8d8 e4a4 f7f6 h2h4 e6e5 a1b1 b7c6 a4a7 a8a7 c5a7 d8d2 b1c1 c6a4 c3c4 g8f7 f2f3 a4c2 c4c5 c2f5 a2a3 d2a2');
INSERT INTO evaluations VALUES(29438,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(29438,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(29438,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(29438,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(29438,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(29438,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(29438,36,'cp 32','g1f3 b8d7 f1d3 c7c6 e1g1 f5d3 d1d3 f6h5 c2c4 g7g6 b1c3 h5f4 e3f4 f8g7 a1d1 e8g8 f1e1 e7e6 g2g3 d7f6 g1g2 a7a6 b2b3 a8c8 h2h4 d8a5 b3b4 a5c7 c4c5 c8d8 c3e4 d6c5 e4f6 g7f6 b4c5 d8d5 d3b3 f8b8 b3c2');
INSERT INTO evaluations VALUES(29439,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29439,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(29439,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(29439,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(29439,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(29439,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(29439,36,'cp 15','b8d7 f1d3 f5d3 d1d3 g7g6 e3e4 f8g7 e1g1 e8g8 b1c3 c7c6 a2a4 d8c7 f1d1 f8e8 h2h3 e7e5 d4e5 d6e5 f4e3 g7f8 d3c4 d7b6 c4e2 b6d7 f3d2 d7c5 d2c4 c5e6 e2f3 f6h5 d1d2 b7b6 g2g3 a8d8 d2d8');
INSERT INTO evaluations VALUES(35150,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(35150,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(35150,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(35150,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(35150,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(35150,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(35150,36,'cp 80','d4c4 b8c6 a2a3 f6e6 b1c3 g8f6 g1f3 f8e7 c1f4 e6c4 f1c4 d7d6 e1c1 c8e6 c4e6 f7e6 e4e5 d6e5 f4e5 c6e5 f3e5 a8d8 d1d8 e7d8 c3b5 a7a6 b5d4 f6d7 e5d3 e8f7 f2f4 h8e8 d4f3 d7b6 g2g3 h7h6 h1e1 d8f6');
INSERT INTO evaluations VALUES(35292,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(35292,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(35292,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(35292,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(35292,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(35292,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(35292,36,'cp 90','b1c3 g8f6 c1e3 d7d5 c4d5 e7d7 g1e2 f6d5 e4d5 f8d6 c3e4 e8g8 e1g1 f8e8 e4d6 d7d6 e3f4 d6b4 f4c7 b4b2 e2f4 c8f5 a1c1 b8d7 d5d6 d7f6 f4d5 f6d5 d1d5 f5e6 d5a5 e6d7 c1b1 b7b6 b1b2 b6a5 f1b1 h7h6 b1d1 a5a4 h2h4 f7f6 h4h5 a7a6 b2b7 d7c6');
INSERT INTO evaluations VALUES(35291,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(35291,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(35291,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(35291,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(35291,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(35291,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(35291,36,'cp 100','g8f6 c1e3 d7d5 c4d5 e7d7 g1e2 f6d5 e4d5 f8d6 c3e4 e8g8 e1g1 b8a6 e4d6 d7d6 e3f4 d6d7 e2c3 b7b6 d1f3 c8b7 a1d1 f7f6 a2a3 c7c6 d5c6 d7c6 f3g3 a6c5 d1d6 c5e4 d6c6 e4g3 c6c7 g3h5 f4d6 f8d8');
INSERT INTO evaluations VALUES(34051,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(34051,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(34051,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(34051,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(34051,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(34051,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(34051,36,'cp 88','c1e3 d7d5 c4d5 e7d7 g1e2 f6d5 e4d5 f8d6 c3e4 e8g8 e1g1 f8e8 e4d6 d7d6 e3f4 d6b4 f4c7 b4b2 e2g3 b8a6 c7d6 c8d7 d1f3 b2b6 d6f4 b6g6 h2h3 b7b6 a1c1 a8c8 c1c8 d7c8 f3a3 a6c5 a3a7 h7h5 d5d6 h5h4 a7b6 h4g3 b6c5 g3f2 c5f2');
INSERT INTO evaluations VALUES(40139,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40139,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40139,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(40139,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40139,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(40139,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(40139,36,'cp 26','e8g8 a2a4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 e7f6 c2c3 f6d4 c3d4 b8d7 e4c2 f8e8 d4d5 c6d5 d1d5 d8c7 d5d3 d7f6 c1g5 f6e4 g5e3 c8f5 f1c1 e4d6 d3b3 c7e7 a4a5 f5c2 b3c2 h7h6 c2d3 e7e5 a5a6 b7b6 h2h3 d6f5');
INSERT INTO evaluations VALUES(42760,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42760,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42760,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42760,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(42760,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(42760,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(42760,36,'cp 0','b4a5 g1f3 d7d6 e1g1 e8g8 f1e1 c7c5 c3d5 c8e6 d5f6 d8f6 d1e2 b8c6 c2c3 e6c4 e2c4 a7a6 d2d3 b7b5 c4a2 a8b8 c1e3 h7h6 e1d1 f8c8 f3d2 c6e7 g1h1 e7g6 a2d5 g6e7');
INSERT INTO evaluations VALUES(43464,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43464,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43464,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43464,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43464,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43464,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43464,36,'cp 0','f8e7 e1g1 e8g8 h2h3 c8e6 c1e3 b8c6 a2a3 h7h6 d1d2 d8c7 d3e2 f8c8 f1d1 e7f8 a1c1 c7d8 a3a4 c6b4 e2f3 b7b5 a4b5 a6b5 c3b5 b4c2 c1c2 e6b3 c2c8 a8c8 d1c1 b3e6 c1c8 d8c8');
INSERT INTO evaluations VALUES(39678,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(39678,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39678,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(39678,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(39678,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(39678,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(39678,36,'cp 13','e1g1 e8g8 c1e3 c8e6 f1e1 h7h6 h2h3 b7b5 a2a4 b5b4 c3d5 f6d5 e4d5 e6d5 d1g4 e7g5 d3e4 d5e4 e3g5 d8g5 g4e4 b8d7 e4b4 g5g6 a1d1 g6c2 d1d2 c2c7 b4d6 c7d6 d2d6 f8b8 b3a5 d7c5 b2b3 e5e4 d6d4 c5b3 a5b3 b8b3 d4e4 a6a5 g1h2 b3d3');
INSERT INTO evaluations VALUES(44515,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(44515,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(44515,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44515,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44515,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(44515,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(44515,36,'cp 35','d4d5 f6e4 g5f4 d8b6 f4c1 e7e6 f2f3 b6a5 c2c3 e4f6 e2e4 d7d6 a2a4 e6d5 e4d5 f8e7 b1d2 a5d8 d2c4 b8d7 g1e2 d7e5 e2g3 e5c4 f1c4 e8g8 c1f4 c8d7 e1g1 a7a6 a4a5 f8e8 f1e1 e7f8 e1e8 d7e8 d1d2 e8d7 h2h3 h7h6 d2d3 b7b5 a5b6 d8b6 b2b3');
INSERT INTO evaluations VALUES(44589,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44589,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(44589,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(44589,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(44589,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(44589,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(44589,36,'cp 15','f6e4 g5f4 d8b6 f4c1 e7e6 f2f3 b6a5 c2c3 e4f6 e2e4 d7d6 a2a4 e6d5 e4d5 b8d7 b1d2 f8e7 d2c4 a5c7 g1e2 d7e5 e2g3 c8d7 c4e3 h7h5 g3f5 e7f8 f3f4 e5g4 h2h3 g4e3 f5e3 e8c8 f1e2 f6e4 e1g1 e4g3 f1e1 f8e7 e3c4 e7h4 e2d3 g3f5 d3f5 d7f5 e1e2 c7d7');
INSERT INTO evaluations VALUES(48203,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(48203,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(48203,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(48203,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(48203,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(48203,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(48203,36,'cp -81','d5c4 d1a4 b8d7 b1c3 f8e7 g1f3 a7a6 e2e3 e8g8 g5f6 g7f6 a4c4 b7b5 c4a2 c7c5 d4c5 e7c5 f1e2 c8b7 a1d1 d8c7 e1g1 d7e5 a2b3 a8c8 f3e5 c7e5 d1d3 f6f5 f1d1 e5g7 g2g3 c5b6 d3d7 b7c6 d7d6 b6a5');
INSERT INTO evaluations VALUES(54896,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(54896,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(54896,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(54896,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(54896,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(54896,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(54896,36,'cp 93','d2d4 f8g7 c1f4 g8f6 d1d2 e8g8 f4h6 g7h6 d2h6 c8g4 e1c1 g4f3 g2f3 b8d7 f3f4 g8h8 f1e2 c7c6 h2h4 f8g8 h4h5 d8f8 h6g5 g6h5 g5h4 a8d8 e2h5 f8h6 h5f7 h6h4 h1h4 g8g4 h4h3 g4f4 d1e1 f6g8 f7e6 f4d4 e1h1 d7f8');
INSERT INTO evaluations VALUES(1822,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(1822,31,'cp -172',NULL);
INSERT INTO evaluations VALUES(1822,32,'cp -169',NULL);
INSERT INTO evaluations VALUES(1822,33,'cp -168',NULL);
INSERT INTO evaluations VALUES(1822,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(1822,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(1822,36,'cp -164','e4e2 f1e2 d7c6 e2d3 c8e6 b1c3 g8f6 c1d2 f6g4 d3e2 e8c8 h2h3 g4e5 e1c1 e5c4 d2f4 f8d6 d1d4 c4e5 h1e1 h8e8 e2h5 e6f5 f4e3 f5g6 h5g6 e5g6 c3e4 c6c5 d4a4 a7a6 e4g5 b7b5 a4a6 c8b7 a6a5 b7b6 g5f7 b6a5');
INSERT INTO evaluations VALUES(1858,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(1858,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(1858,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(1858,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(1858,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(1858,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(1858,36,'cp 52','f6g6 d2d4 g6e4 c1e3 d7d5 b1d2 e4g6 c4d5 e5d4 f3d4 g8f6 d5c6 b7c6 e1g1 f8d6 d2c4 f6g4 c4d6 g6d6 d4f3 g4e3 f1e1 e8g8 e1e3 c6c5 h2h3 c8b7 d1e2 a8d8 a1e1 h7h6 b2b3 b7c6 f3e5 f8e8 e2h5 d6d5 h5g4 d5d6 c3c4 d6f6 g4g3 f6d6 e5c6 e8e3 e1e3 d6c6 e3e7 d8d7 e7e8 g8h7');
INSERT INTO evaluations VALUES(7685,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(7685,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(7685,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(7685,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(7685,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(7685,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(7685,36,'cp 63','e4e5 f6g4 d1g4 d7d5 g4e2 d5c4 f2f4 d8b6 a1b1 c6c5 c1e3 b6c6 e1g1 c8e6 b1e1 e8g8 e3d2 a8d8 c3e4 f7f5 e4c3 a7a5 a2a4 d8d7 c3b5 e6d5 d2c3 e7e6 b5d6 c6a4 e1a1 a4c6 a1a5 f8a8 a5a8 c6a8 d6c4 a8c6 h2h3 g7f8 g1h2 d5c4 e2c4');
INSERT INTO evaluations VALUES(9314,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(9314,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(9314,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(9314,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(9314,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(9314,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(9314,36,'cp 58','c3d5 c8e6 c2c3 e6d5 e4d5 c6b8 d1c2 h7h6 g5h4 c7c6 e1c1 e8g8 d3d4 e5d4 c3d4 c5b6 g2g4 c6d5 c1b1 b8c6 g4g5 h6g5 h4g5 f8e8 a2a3 a8c8 f1h3 c8c7 c2f5 g8f8 h1g1 e8e4 g5e3 d8d7 f5f6 g7f6 h3d7 c7d7 h2h4 f6f5 h4h5 f5f4 e3c1 f7f6 g1g6 b6d4 f3d4 e4d4 g6f6 f8g7 d1d4 g7f6 d4f4 f6g5');
INSERT INTO evaluations VALUES(9426,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(9426,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(9426,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(9426,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(9426,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(9426,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(9426,36,'cp 66','c8e6 c2c3 e6d5 e4d5 c6b8 d1c2 e8g8 e1c1 c7c6 d3d4 e5d4 c3d4 c5b6 f1d3 h7h6 d3h7 g8h8 h2h4 c6d5 c1b1 d8c8 g5f6 c8c2 h7c2 g7f6 c2b3 b8c6 b3d5 f6f5 d1d2 b6a5 d2e2 a5b6 h1c1 c6d4 f3d4 b6d4 d5b7 a8e8 e2d2 d4c5 b7d5 e8b8 c1c4 h8g7 c4f4 g7g6 g2g4 f5g4');
INSERT INTO evaluations VALUES(10271,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(10271,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(10271,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(10271,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10271,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(10271,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(10271,36,'cp 40','f3e5 c5f2 e1f2 c6e5 d3d4 e5g6 c2c4 e8g8 b1c3 c7c6 h1f1 d7d5 f2g1 h7h6 c4d5 c6d5 d1b3 g8h8 c1d2 a7a5 a1c1 c8e6 c3b5 f6e4 b3e3 g6e7 a2a4 e4d2 e3d2 a8c8 c1c8 d8c8 f1c1 e7c6 d2f4 h8g8 g1f2 c8d7 f4c7 d7c7 b5c7 c6d4 c7d5');
INSERT INTO evaluations VALUES(10946,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(10946,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(10946,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(10946,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(10946,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(10946,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(10946,36,'cp 86','d4c6 d7c6 c1f4 a7a5 d1d8 e7d8 e1c1 h7h5 h2h4 f6d7 a2a4 d7c5 f2f3 g7g6 f1e2 c8d7 f4e5 f7f6 e5f4 d7e6 b2b3 c5d7 f4g3 e8f7 e2c4 d7e5 c4e6 f7e6 c3e2 h8e8 f3f4 e5g4 f4f5 g6f5 e2f4 e6f7 e4f5');
INSERT INTO evaluations VALUES(10852,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(10852,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(10852,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(10852,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(10852,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(10852,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(10852,36,'cp 67','d7c6 c1f4 a7a5 d1d8 e7d8 h2h4 h7h5 e1c1 f6d7 f1e2 d7c5 f4e5 f7f6 e5g3 c8e6 h1e1 e8f7 f2f4 e6g4 f4f5 g4e2 c3e2 h8e8 e2f4 e8e4 e1e4 c5e4 f4h5 d8e7 g3c7 a8h8 h5g3 h8h4 g3e4 h4e4 c7a5 e4f4 d1d7 f4f5');
INSERT INTO evaluations VALUES(8666,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(8666,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(8666,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(8666,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(8666,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(8666,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(8666,36,'cp 63','c1f4 a7a5 d1d8 e7d8 e1c1 h7h5 h2h4 f6d7 f1e2 d7c5 f4e5 f7f6 e5g3 g7g6 h1e1 c8e6 f2f4 e8f7 e2f3 f6f5 e4f5 e6f5 g3f2 c5e6 d1d7 f7e8 d7d2 e8f7 c3e4 h8e8 e4g3 e6g7 e1e8 f7e8 g3f5 g7f5 g2g3 d8e7');
INSERT INTO evaluations VALUES(12399,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(12399,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12399,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(12399,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(12399,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(12399,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(12399,36,'cp 9','a2a3 d7d5 d2d4 d5c4 f1c4 a7a6 c4a2 f8e7 e1g1 b7b5 d1c2 c5d4 e3d4 c8b7 d4d5 f6d5 c3d5 e6d5 a2b1 d5d4 c1d2 d4d3 c2d3 d8d3 b1d3 e8g8 d2c3 b5b4 a3b4 c6b4 d3c4 b7f3 g2f3 f8c8 c3b4 e7b4 a1a6 a8a6 c4a6 c8d8 a6e2 d8d2');
INSERT INTO evaluations VALUES(16051,30,'cp -516',NULL);
INSERT INTO evaluations VALUES(16051,31,'cp -501',NULL);
INSERT INTO evaluations VALUES(16051,32,'cp -532',NULL);
INSERT INTO evaluations VALUES(16051,33,'cp -505',NULL);
INSERT INTO evaluations VALUES(16051,34,'cp -528',NULL);
INSERT INTO evaluations VALUES(16051,35,'cp -521',NULL);
INSERT INTO evaluations VALUES(16051,36,'cp -551','c6d4 c4f7 e8e7 b3c4 b7b5 c4d3 e7f7 g1f3 c7c5 f3e5 f7g7 c2c3 d4e6 d3b5 d8c7 e5f3 a8b8 b5a4 e6f4 e1g1 c8b7 d2d3 f4d3 b1d2 b8e8 a4b5 d3c1 f1c1 f6e4 c1e1 f8d6 d2e4 b7e4 f3g5 e4c6 b5c4 e8e1 a1e1 h8f8 g2g3 c7b7 g5e4 c6e4 c4e4 b7b2 e1e2 b2c3 g1g2 c3f6');
INSERT INTO evaluations VALUES(16643,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(16643,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(16643,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(16643,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16643,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16643,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16643,36,'cp 0','e1g1 f8e7 e4e5 f6h5 b1c3 h5f4 c3d5 f4e6 a1c1 b7b5 c4b5 a8b8 d1e2 c6b4 d5e7 d8e7 b5c4 c8a6 f3d4 e8g8 c4a6 b4a6 d4f5 e7g5 b2a3 g5f5 a3f8 e6f4 e2a6 f5g4 g2g3 f4e2 g1g2 e2f4');
INSERT INTO evaluations VALUES(16642,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16642,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16642,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16642,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16642,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16642,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16642,36,'cp 0','f8e7 e4e5 f6h5 b1c3 h5f4 c3d5 f4e6 a1c1 b7b5 c4b5 a8b8 d1e2 c6b4 d5e7 d8e7 b5c4 c8a6 f3d4 e8g8 c4a6 b4a6 d4f5 e7g5 b2a3 g5f5 a3f8 e6f4 e2a6 f5g4 g2g3 f4e2 g1g2 e2f4');
INSERT INTO evaluations VALUES(16843,30,'cp -210',NULL);
INSERT INTO evaluations VALUES(16843,31,'cp -206',NULL);
INSERT INTO evaluations VALUES(16843,32,'cp -205',NULL);
INSERT INTO evaluations VALUES(16843,33,'cp -179',NULL);
INSERT INTO evaluations VALUES(16843,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(16843,35,'cp -211',NULL);
INSERT INTO evaluations VALUES(16843,36,'cp -194','d7d5 c4b5 e5d4 e1g1 f8d6 f3d4 c8d7 b1d2 e8g8 d2e4 d5e4 f1e1 f8e8 d4c6 b7c6 b5f1 h7h6 c1e3 f7f5 g2g3 d8f6 d1h5 d7e6 e3d4 f6f7 h5f7 e6f7 c3c4 a7a5 f1e2 d6b4 e1d1 a5a4 h2h4 b4d6 c4c5');
INSERT INTO evaluations VALUES(16933,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(16933,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(16933,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(16933,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(16933,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(16933,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(16933,36,'cp 117','c1e3 c7c5 f2f3 d7d6 f1d3 d8b6 e3c1 b6b4 d1d2 f6h5 c3b1 b4d2 e1d2 c8d7 b1a3 h5f4 d3f1 b7b5 c2c3 b5b4 a3c4 d4b5 d2c2 f7f5 g1e2 f4h5 e2g3 h5g3 h2g3 f5e4 f3e4 f8e7 c1e3 b4c3 b2c3 e8g8 c4a5 b5c7 a5c6 e7d8 a2a4 f8f6 f1e2 f6g6 c6d8 a8d8');
INSERT INTO evaluations VALUES(16931,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(16931,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(16931,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(16931,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(16931,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(16931,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(16931,36,'cp 117','c7c5 f2f3 f8e7 c3b1 d8b6 b2b3 c5c4 e3d4 e5d4 f1c4 e7b4 e1f1 e8g8 g2g3 b4c5 f1g2 d7d6 b1d2 f6d7 a2a3 a7a5 c4d3 b6c7 g1h3 h7h6 h3f4 d7e5 a3a4 c8d7 d2c4 c7d8 h1f1 f8e8 g2h1 c5b4 f4g2 b7b5 a4b5');
INSERT INTO evaluations VALUES(18314,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(18314,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(18314,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(18314,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(18314,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(18314,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(18314,36,'cp 133','c2c3 a7a6 b5c4 b6b5 c4b3 g8f6 d2d4 f6e4 e1g1 d7d5 d4e5 c8e6 b3c2 f8e7 b1d2 e4d2 d1d2 e8g8 f1e1 d8d7 h2h3 f8e8 d2f4 f7f6 e5f6 e7f6 c1e3 e6f7 f3g5 f7g6 c2g6 h7g6 f4h4 f6g5 h4g5 d7f7 e3c5 e8e4 e1e4 d5e4 a1d1 a8e8 g5g3 e8e6');
INSERT INTO evaluations VALUES(25375,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(25375,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(25375,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(25375,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(25375,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(25375,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(25375,36,'cp 34','c4b5 f6d7 c3e2 e6g4 d2d4 g4f3 g2f3 c6d4 e2d4 e5d4 d1d4 a7a6 b5e2 d8f6 d4f6 d7f6 a2a4 g7g5 f1d1 h7h6 a1a3 f8e7 b2b4 f6h5 a3c3 c7c6 c3b3 h5f4 e2f1 b7b5 a4a5 a8b8 c1e3 e8g8 c2c4 b5c4 f1c4 d6d5 e3f4 d5c4');
INSERT INTO evaluations VALUES(26700,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(26700,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(26700,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(26700,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26700,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(26700,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(26700,36,'cp 0','b4d3 c2d3 c7c6 b1c3 e7e6 e3e4 f5g6 h2h4 d5e4 d3e4 g8f6 d1b3 f6e4 b3b7 d8c8 b7b3 c8a6 b3a4 a6a4 c3a4 e4f6 f3e5 f6d5 f4d2 f7f6 e5g6 h7g6 e1e2 e8d7 a1c1 f8d6 g2g3 a8b8 h1d1 h8h5 b2b3 d5e7 c1c4 h5d5 d1c1 e7f5 d2c3 d6a3');
INSERT INTO evaluations VALUES(26711,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26711,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(26711,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26711,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(26711,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(26711,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26711,36,'cp 25','c2d3 c7c6 b1c3 e7e6 h2h4 g8f6 f3e5 h7h5 e1g1 f8d6 f1e1 e8g8 f4g5 c6c5 d1b3 c5d4 e3d4 d8e7 a1c1 a8c8 a2a3 g8h7 g2g3 f5g6 c3e2 g6f5 c1c3 c8c3 b3c3 e7c7 g5f6 g7f6 c3c7 d6c7 e5d7 f8d8 d7f6 h7g6');
INSERT INTO evaluations VALUES(26710,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(26710,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(26710,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(26710,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(26710,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26710,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(26710,36,'cp 10','c7c6 b1c3 e7e6 h2h4 g8f6 f3e5 h7h5 e1g1 f6g4 e3e4 g4e5 f4e5 f5g6 f1e1 f8d6 e4d5 d6e5 d4e5 c6d5 d1a4 d8d7 d3d4 d7a4 c3a4 e8e7 f2f3 h8c8 a1c1 c8c4 c1c4 d5c4 g1f2 a8d8 f2e3 f7f6 a4c3 a7a6 a2a3 b7b5 e5f6 g7f6 g2g4 d8h8 e1h1 g6f7 g4g5 f6g5 h4g5 f7g6 h1h4 h8g8 d4d5 e6d5 c3d5 e7f7 d5f4 g8e8 e3f2');
INSERT INTO evaluations VALUES(28019,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(28019,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(28019,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(28019,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(28019,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(28019,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(28019,36,'cp 121','b1c3 d7d6 d1e2 a7a6 c4b3 e7e6 h2h4 h7h6 c1f4 g8e7 e1c1 b8d7 c1b1 b6b5 a2a3 d7b6 h4h5 g6g5 f3g5 h6g5 f4g5 g7h6 f2f4 b5b4 a3b4 a6a5 g5f6 h8h7 d4d5 a5b4 c3b5 b6d7 e4e5 d7f6 e5f6 e7d5 b3d5 b7d5 d1d5 h6f4 e2e4 h7h8 e4f4 e6d5');
INSERT INTO evaluations VALUES(28265,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(28265,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28265,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28265,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28265,36,'cp 15','e8g8 f1d3 d7d5 c4d5 e6d5 a1c1 b4e7 e1g1 c7c5 d4c5 b6c5 f1e1 f8e8 e3e4 d5e4 d3e4 f6e4 c3e4 b8a6 d1a4 d8d5 c1d1 d5c6 a4c6 b7c6 f3e5 c6b7 e5d7 b7d5 d2c3 d5a2 d1a1 a2c4 e1e3 f7f6 c3f6 g7f6 e4f6 g8f7 f6e8 f7e8 a1e1 e8d7 e3e7 d7c6 e7h7 a6b4');
INSERT INTO evaluations VALUES(27775,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(27775,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(27775,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(27775,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(27775,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27775,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(27775,36,'cp 24','f1d3 d7d5 c4d5 e6d5 a1c1 b4e7 e1g1 c7c5 d4c5 b6c5 f1e1 f8e8 d1b3 d8b6 e3e4 c5c4 b3b6 a7b6 d3b1 f6e4 c3e4 d5e4 c1c4 e4f3 c4c7 b7d5 e1e7 e8d8 d2c3 f3g2 b2b4 b8a6 c7d7 d8d7 e7d7 d5e6 d7d6 a6c7 b1e4');
INSERT INTO evaluations VALUES(27777,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(27777,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(27777,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(27777,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(27777,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(27777,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(27777,36,'cp 13','d7d5 c4d5 e6d5 a1c1 b4e7 e1g1 c7c5 d4c5 b6c5 f1e1 f8e8 d1b3 d8b6 e3e4 b6b3 a2b3 f6e4 c3d5 b7d5 d3e4 d5e4 e1e4 b8c6 d2e3 f7f5 e4e6 c6b4 g1f1 f5f4 e3d2 b4d3 c1a1 e7f6 e6e8 a8e8 d2c3 d3b2 c3f6 g7f6 a1a7 c5c4 b3c4 b2c4 a7d7 c4e5');
INSERT INTO evaluations VALUES(32354,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(32354,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(32354,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(32354,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(32354,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(32354,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(32354,36,'cp -14','g8e7 c2c3 c7c5 g1e2 b8c6 e1g1 a7a6 a2a3 f5d3 d1d3 e7f5 b1d2 d8b6 d2f3 f8e7 c1e3 c5d4 c3d4 f5e3 f2e3 e8g8 e2c3 c6a5 b2b4 a5c4 e3e4 b6c6 a1c1 a8c8 f3d2 c4d2 e4d5 e6d5 d3d2 c6d7 c3e2 c8c1 e2c1 f7f6 c1d3 f6e5 f1f8 e7f8 d4e5 d7e6');
INSERT INTO evaluations VALUES(38851,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(38851,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(38851,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(38851,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(38851,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(38851,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(38851,36,'cp 174','d8d7 d1g4 f7f6 g1f3 b7b6 a2a4 b8c6 f1d3 c8b7 e5f6 g7f6 e1g1 e8c8 f1e1 d8e8 g4h3 c8b8 f3h4 d7f7 h4g6 h7g6 h3h8 g8e7 h8h3 e7f5 f2f3 e6e5 d4e5 f6e5 c1g5 c6a5 e1e2 f5d6 h3g3 e5e4 d3b5 d6b5 a4b5 f7d7 a1b1 a5c4 g5f4 a7a5 g3g6');
INSERT INTO evaluations VALUES(40789,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(40789,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(40789,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(40789,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(40789,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(40789,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(40789,36,'cp 32','d2d4 e5d4 d1d4 e8g8 b2b3 f8e8 g1f3 b8c6 d4d2 a7a5 c1b2 a5a4 a1c1 h7h6 e1g1 a4b3 a2b3 c8f5 f1d1 d8d7 c1a1 a8a1 d1a1 f6e4 c3e4 f5e4 b3b4 d7e6 b4b5 c6e5 f3e5 e4g2 g1g2 d6e5 d2d5 e6d5 c4d5 f7f6 a1a7 e8b8 b2a3 g8f7 e2e4 e7a3 a7a3');
INSERT INTO evaluations VALUES(46768,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(46768,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(46768,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46768,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46768,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46768,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(46768,36,'cp 16','f8g7 b1d2 e8g8 f1e2 b8d7 h2h3 c7c5 g1f3 b7b6 e1g1 a7a6 e2d3 c8b7 d3c2 b6b5 a2a4 c5c4 d1e2 f8e8 f3e5 d7e5 f4e5 e7e6 a4b5 a6b5 a1a8 b7a8 e2f3 f6d7 e5g7 g8g7 f1a1 a8c6 a1a6 c6b7 a6d6 d8e7');
INSERT INTO evaluations VALUES(50645,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(50645,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(50645,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(50645,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(50645,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(50645,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(50645,36,'cp -60','e7e6 d2d3 d6d5 e4d5 e6d5 c4b3 f8e7 g5f3 g8f6 e1g1 e8g8 c1g5 h7h6 g5h4 b8c6 f1e1 c8f5 a2a4 f5e6 b1c3 a8c8 h2h3 c6b4 d1e2 c5c4 d3c4 d5c4 b3a2 f8e8 e1d1 d8c7 h4g3 c7b6 f3d4 b4a2 a1a2 e7b4 e2f3 b4c3 b2c3 e6d5 f3f5 g8h8 a4a5 b6c5 a2a1');
INSERT INTO evaluations VALUES(50728,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(50728,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(50728,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(50728,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(50728,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(50728,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(50728,36,'cp -62','d2d3 d6d5 e4d5 e6d5 c4b3 h7h6 g5f3 d8e7 d1e2 e7e2 e1e2 g8f6 c2c4 d5d4 h1e1 f8d6 b3a4 e8d8 b2b4 b7b6 e2d1 d8c7 b1d2 b8c6 b4c5 b6c5 a4c6 c7c6 f3e5 d6e5 e1e5 f6g4 e5e2 c8f5 h2h3 g4f6 d2f3 h8e8 e2e8 a8e8 d1d2 g7g5 c1a3 f6d7 a1e1 e8e1 f3e1 f5g6 e1c2');
INSERT INTO evaluations VALUES(50727,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(50727,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(50727,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(50727,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(50727,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(50727,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(50727,36,'cp -60','d6d5 e4d5 e6d5 c4b3 g8f6 e1g1 f8e7 c2c4 d5d4 b3a4 b8d7 f1e1 e8g8 g5f3 h7h6 b1d2 b7b6 b2b4 c8b7 a1b1 e7d6 b4c5 d7c5 a4c2 d8c7 d2f1 f6g4 f1g3 a8d8 f3h4 f8e8 e1e8 d8e8 h4f5 h6h5 h2h3 d6g3 f5g3 c7c6 f2f3 g4e3 c1e3 e8e3');
INSERT INTO evaluations VALUES(50860,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(50860,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(50860,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(50860,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(50860,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(50860,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(50860,36,'cp 73','g8f6 f3e5 c8e6 f1d3 g7g6 c3e2 f6d7 e5f3 e6g4 f3g5 e7e6 h2h3 g4f5 e1g1 f5d3 d1d3 b8c6 f1e1 f8g7 g5f3 e8g8 c1g5 d7f6 a1d1 f8e8 g5f4 b7b5 e2c1 c6a5 d3e2 f6d5 f4g3 a5c4 c1d3 a8c8 c2c3 d8d7 f3e5 c4e5 d3e5 g7e5');
INSERT INTO evaluations VALUES(53929,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(53929,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(53929,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(53929,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(53929,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(53929,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(53929,36,'cp 115','b1c3 g7g6 e2e4 b8d7 h2h4 h7h5 g2g3 f8h6 f2f3 h6c1 d1c1 g8h6 g1h3 d8e7 f1d3 f7f6 c1d2 h6f7 h3f2 d7f8 f3f4 c8d7 e1c1 e8c8 c1b1 c8b8 d3c2 g6g5 f4f5 d7e8 a2a4 f8d7 b1a1 b8a8 a4a5 g5g4');
INSERT INTO evaluations VALUES(53926,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(53926,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(53926,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(53926,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(53926,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(53926,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(53926,36,'cp 109','g7g6 e2e4 h7h5 h2h3 b8a6 g1f3 c8d7 g2g3 f8h6 a2a3 h6c1 d1c1 g8h6 c1d2 d8e7 f1g2 f7f6 e1g1 e7g7 c3b5 d7b5 c4b5 a6b8 a3a4 b8d7 f1c1 e8g8 a4a5 g6g5 b2b4 h6f7 d2b2 a8c8 f3d2 c5b4 b2b4');
INSERT INTO evaluations VALUES(55642,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(55642,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(55642,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(55642,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(55642,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(55642,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55642,36,'cp 18','g7g6 f1g2 f8g7 c2c4 g8e7 b1c3 e8g8 e1g1 b8c6 e2e3 c8e6 f3g5 e6c8 a1b1 a7a5 a2a3 h7h6 g5f3 c8e6 b2b4 a5b4 a3b4 d8d7 b4b5 c6d8 c1b2 e6h3 g2h3 d7h3 d3d4 e5d4 f3d4 d8e6 d1f3 e6d4 e3d4 e7f5 f3b7 f5d4 b7g2 h3f5 g2d5 f8e8 d5f5');
INSERT INTO evaluations VALUES(12880,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(12880,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(12880,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(12880,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(12880,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(12880,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(12880,36,'cp -14','d7d5 e4d5 d8d5 d4c6 d5c6 b1c3 a7a6 c1e3 e7e5 f1d3 f8a3 d1c1 e8g8 e1g1 a3b4 c1e1 c8e6 e1h4 b4c5 e3c5 c6c5 h4f2 c5c7 f1e1 h7h6 a1d1 a8d8 f2e3 f8e8 a2a3 e6d5 d3f1 d5c6 d1d8 c7d8 e3c5 e5e4');
INSERT INTO evaluations VALUES(17526,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17526,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17526,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(17526,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(17526,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(17526,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17526,36,'cp -5','f1e2 d7d6 d1b3 e7e6 e1g1 d8c7 f1d1 g8f6 c1f4 f8e7 b3a3 e6e5 f4g5 d6d5 a3a4 d5e4 g5f6 e7f6 c3d5 c7d8 d5e3 d8a5 a4e4 c8e6 e2c4 e8g8 c4e6 f7e6 d1d7 a5b4 e4b4 c6b4 d7b7 b4d3 f3d2 f6h4 g2g3 f8f2 d2c4 h4g5 a1d1 g5e3 d1d3 e3c5 b2b4 f2c2 b4c5 c2c4');
INSERT INTO evaluations VALUES(22586,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(22586,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(22586,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(22586,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(22586,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(22586,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(22586,36,'cp 7','b1c3 c6d4 c3e2 d4c6 g1f3 d7d5 e4d5 e6d5 c4d5 d8d5 e2c3 d5h5 f1e2 g8f6 d2d4 f8d6 d4d5 c6e5 c1f4 e5f3 e2f3 c8g4 f4d6 c7d6 d1a4 e8f8 f3g4 h5g4 a4g4 f6g4 h2h3 g4f6 e1c1 h7h5 g2g3 h8h6 d1d4 h6g6 d4c4 g6g5 h1d1 g5e5 d1d4 a8e8 c1d2');
INSERT INTO evaluations VALUES(22584,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(22584,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(22584,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(22584,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(22584,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(22584,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(22584,36,'cp 13','c6d4 c3e2 d4c6 g1f3 d7d5 c4d5 e6d5 e4d5 d8d5 e2c3 d5h5 f1e2 f8d6 d2d4 g8f6 d4d5 c6e5 c1f4 e5f3 e2f3 c8g4 d1a4 e8f8 f3g4 h5g4 f4d6 c7d6 a4g4 f6g4 h2h3 g4f6 e1c1 h7h5 g2g3 a8c8 d1d4 a7a6 h1e1 h8h6 c1b1 h6g6 d4b4 c8c7 a2a4 g6g5 b4b6 c7d7 e1d1');
INSERT INTO evaluations VALUES(24368,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(24368,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(24368,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(24368,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(24368,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(24368,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(24368,36,'cp 90','e1g1 f8e7 f2f4 e8g8 d1f3 a7a5 c1e3 f8e8 a1d1 d8c8 h2h3 c8e6 e3f2 a5a4 f1e1 a8a5 b2b4 a4b3 c2b3 e7f8 g1h1 h7h6 f2h4 f8e7 h4g3 e7f8 d1d4 c6c5 d4c4 e6c8 e4e5 f6d7 c4e4 a5a3 g3h4 d6e5 f4e5 a3a6 h4g3 a6e6 a2a4 c8b7 h1h2');
INSERT INTO evaluations VALUES(24367,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(24367,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(24367,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(24367,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(24367,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(24367,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(24367,36,'cp 70','f8e7 f2f4 a7a5 c1e3 e8g8 h2h3 a5a4 d1f3 d8b8 e3d4 b8b7 a1d1 c6c5 d4f2 f8e8 b2b3 a4b3 a2b3 f6d7 f1e1 d7b6 f3d3 e7f8 g1h2 c7c6 c3e2 d6d5 e2g3 c5c4 d3f3 d5e4 g3e4 b7c7 f2b6 c7b6 b3c4 b6a6 e4d2 e8e1 d1e1');
INSERT INTO evaluations VALUES(26272,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(26272,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(26272,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(26272,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(26272,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(26272,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(26272,36,'cp -44','f1b5 d8b6 b5c6 b6c6 d2d3 e7e6 h2h3 g4f3 d1f3 g8e7 e1g1 e8c8 c1e3 h7h5 e3f2 g7g6 b1c3 e7f5 c3e2 f8e7 c2c3 c8b8 a2a4 d8c8 e2g3 h8g8 g3f5 g6f5 g1h2 h5h4 b2b3 g8g7 f1c1 c8g8 c1g1 c6c8 d3d4 c5d4 f2d4 g7g3 f3f2 g3c3 d4c3 c8c3 a1c1 c3b3');
INSERT INTO evaluations VALUES(26253,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(26253,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(26253,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(26253,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(26253,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(26253,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(26253,36,'cp -43','d8b6 b5c6 b6c6 h2h3 g4f3 d1f3 e7e6 d2d3 g8e7 e1g1 e8c8 c1e3 h7h5 e3f2 g7g6 f2h4 d8d7 b1d2 c8b8 h4f6 h8g8 c2c3 c6b5 f1b1 b5a4 b2b3 a4a5 b1c1 e7f5 f6g5 f8e7 g5e7 f5e7 f3f2 a5a6 d3d4 c5d4 c3d4 g8c8 c1c8 e7c8');
INSERT INTO evaluations VALUES(28146,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(28146,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(28146,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28146,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28146,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28146,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(28146,36,'cp 5','g8f6 c2c4 e8g8 b1c3 h7h6 c4d5 f6d5 f1d3 d5c3 b2c3 f5d3 d1d3 b8d7 a2a4 d6g3 h2g3 c7c5 e1g1 d8c7 f1b1 a8b8 g3g4 d7f6 g4g5 h6g5 f3g5 f8d8 g2g3 d8d5 g5e4 f6e4 d3e4 b8d8 a4a5 c5d4 c3d4 d8d7 g1g2 c7c6 e4g4 b7b6 a5b6 d5f5 g2g1 a7b6');
INSERT INTO evaluations VALUES(28457,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(28457,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(28457,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28457,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(28457,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28457,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28457,36,'cp 5','c2c4 e8g8 b1c3 h7h6 f1e2 d5c4 e1g1 c7c5 d4c5 d6c5 f3e5 d8e7 e5c4 f8d8 d1b3 b8c6 e2f3 c6b4 a1c1 b4d3 c1d1 c5b4 c4e5 b4c3 b3c3 a8c8 c3b3 d3c5 b3a3 f6e4 d1d8 e7d8 f1c1 a7a5 c1d1 d8f6 e5c4 e4g3 h2g3');
INSERT INTO evaluations VALUES(35950,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(35950,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(35950,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(35950,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(35950,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(35950,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(35950,36,'cp 106','d4d5 e7e6 f1d3 e6d5 e4d5 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 e3f4 e4f6 d1b3 b8a6 f1e1 d8d7 d3f1 e8e1 a1e1 a8b8 f4d2 b8e8 e1e8 f6e8 g2g3 b7b6 g1g2 a6b8 d2c1 d7e7 b3a4 e8f6 f1d3 a7a6 c1g5 e7d7 a4c2');
INSERT INTO evaluations VALUES(37032,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(37032,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(37032,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(37032,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(37032,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(37032,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(37032,36,'cp -14','f1d3 b7b6 a2a3 b4c3 d2c3 f6e4 c3d2 b8d7 e1g1 c7c5 a3a4 c8b7 d2e1 a7a5 c2c3 e4d6 f3e5 b7a6 d3a6 a8a6 f2f3 d8c7 e1g3 d7e5 g3e5 c7c6 d1e2 a6a8 f1e1 c5d4 c3d4 f8c8 e3e4 d5e4 f3e4 c8e8 a1d1 a8c8 e5g3 h7h6');
INSERT INTO evaluations VALUES(37033,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(37033,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(37033,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(37033,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(37033,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(37033,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(37033,36,'cp 0','b7b6 a2a3 b4c3 d2c3 f6e4 c3d2 c7c5 d4c5 e4c5 e1g1 c8b7 f3d4 b8c6 d2c3 c5d3 c2d3 c6d4 c3d4 f8e8 f2f4 f7f6 d1g4 a8c8 f1f3 d8e7 f3g3 b7a6 a1d1 c8c2 h2h3 a6b5 g4h4 e6e5 d4c3 e7f7 f4e5');
INSERT INTO evaluations VALUES(42561,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(42561,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(42561,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(42561,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(42561,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(42561,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(42561,36,'cp -33','c5e3 f2e3 c7c6 c4b3 d7d6 b1c3 b7b5 g1f3 a7a5 a2a3 e8g8 e1g1 d8b6 d1d2 c8e6 b3e6 f7e6 a3a4 b8d7 c3d1 b5b4 b2b3 d6d5 h2h3 c6c5 c2c4 b4c3 d1c3 a8d8 c3b5 d5e4 f3g5 h7h6 g5e4 f6e4 d3e4 d7f6');
INSERT INTO evaluations VALUES(42792,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(42792,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(42792,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(42792,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(42792,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(42792,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(42792,36,'cp -36','f2e3 c7c6 c4b3 d7d6 g1f3 b8d7 b1c3 d7c5 d3d4 c5b3 a2b3 d8e7 d1d3 e8g8 e1c1 f8e8 d4e5 d6e5 d3d6 e7d6 d1d6 c8g4 h1f1 g8f8 h2h3 g4f3 g2f3 f8e7 d6d1 e8d8 b3b4 g7g6 b4b5 c6b5 c3b5 a7a6 b5c3 a8c8 c3d5 f6d5 e4d5');
INSERT INTO evaluations VALUES(42791,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(42791,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(42791,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(42791,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(42791,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(42791,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(42791,36,'cp -35','c7c6 c4b3 d7d6 g1f3 b8d7 b1c3 d7c5 d3d4 c5e4 d4e5 e4c3 b2c3 d6e5 f3e5 c8e6 b3e6 f7e6 e1g1 e8g8 d1d4 d8e7 a1d1 f8d8 d4h4 e7c7 e5g4 f6g4 h4g4 c7e5 g4f4 e5f4 e3f4 g8f7 g2g4 b7b5 g1f2 d8d5 f2e3 f7f6 d1d4 a8d8 e3e4 a7a5 h2h4 h7h6 d4d5 d8d5');
INSERT INTO evaluations VALUES(46057,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(46057,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(46057,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(46057,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(46057,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(46057,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(46057,36,'cp 124','g1f3 f8e7 f1d3 e8g8 e1g1 b8c6 h2h3 d8b6 c3a4 b6d8 c1e3 b7b6 a4c3 c8b7 a1c1 a8c8 f1e1 c6a5 f3d2 a5c6 d3b1 a7a6 a2a3 b6b5 b2b4 f8e8 d2f3 f6d7 d1d3 d7b6 e4e5 g7g6 e5d6 e7d6 c3e4 b6d5 e4c5 b7a8 e3h6 d5f4');
INSERT INTO evaluations VALUES(46056,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(46056,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(46056,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(46056,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(46056,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(46056,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(46056,36,'cp 118','f8e7 f1d3 e8g8 e1g1 b8c6 a2a3 c8d7 h2h3 a8c8 f1e1 c6a5 c1f4 d6d5 e4e5 f6e4 d3e4 d5e4 c3e4 d7c6 d1d3 c6e4 d3e4 c8c6 a1c1 d8d7 e1d1 c6c1 f4c1 d7a4 e4d3 f8d8 c1g5 a5c6 g5e3 h7h6 d1c1 d8d7 f3d2 a7a6 b2b3 a4b5 d3b5 a6b5');
INSERT INTO evaluations VALUES(40208,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(40208,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(40208,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(40208,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(40208,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(40208,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(40208,36,'cp 105','f1d3 b8c6 e1g1 e8g8 h2h3 d8b6 c3a4 b6c7 c1e3 c8d7 a4c3 a8c8 f1e1 f8d8 a1c1 c7b8 e3d2 h7h6 d1e2 f6h7 e1d1 h7g5 d2e3 g5f3 e2f3 e7g5 e3g5 h6g5 c3e2 c6b4 d3b1 e6e5 d4d5 c8c1 d1c1 d8c8 e2c3 a7a6 a2a3');
INSERT INTO evaluations VALUES(40209,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(40209,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(40209,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(40209,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(40209,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(40209,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(40209,36,'cp 107','b8c6 e1g1 e8g8 a2a3 c8d7 c1f4 a8c8 a1c1 c6d4 f3d4 e6e5 f4e3 e5d4 e3d4 a7a6 h2h3 f8e8 f2f4 d7c6 d3b1 f6d7 d1f3 e7h4 g1h2 f7f5 f3h5 h4f6 b1a2 g8h8 d4f6 d7f6 h5f5 f6e4 c3e4 c6e4 c1c8 e4f5 c8d8 e8d8');
INSERT INTO evaluations VALUES(46447,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46447,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46447,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46447,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46447,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46447,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46447,36,'cp 0','d2d4 c5c4 b3a4 c8d7 a4d7 b8d7 e1g1 f8e7 f3h4 e8g8 h4f5 f8e8 d1f3 f6e4 f3g4 g7g6 f2f3 e4f6 f5h6 g8f8 g4f4 d7b8 c1d2 b8c6 b1c3 d8b8 f4h4 f6h5 h4g4 e7f6 g4d7 c6e7 d7g4');
INSERT INTO evaluations VALUES(48793,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(48793,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(48793,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(48793,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(48793,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(48793,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(48793,36,'cp -116','d5d4 c3e4 f6e4 g2e4 f8d6 d2d3 e8g8 e4g2 c7c5 g1h3 b8c6 e1g1 c8g4 f2f4 e5f4 c1f4 f8e8 f1f2 g4h3 g2h3 d6f4 f2f4 e8e5 d1f1 d8e7 h3g2 g7g6 f4e4 a8d8 e4e5 c6e5 a1e1 b7b6 f1f4 g8g7 g2d5 d8d7');
INSERT INTO evaluations VALUES(48913,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(48913,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(48913,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(48913,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(48913,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(48913,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(48913,36,'cp -93','c3e4 f6e4 g2e4 f8d6 d2d3 e8g8 e4g2 h7h6 g1f3 c7c5 c1d2 b8c6 e1g1 c8f5 d1b3 d8d7 e2e3 d4e3 f2e3 a7a5 f3h4 f5e6 h4f3 a5a4 b3c2 a4a3 b2b3 f7f5 d2c3 f5f4 f3h4 f4e3 c2e2 a8d8 e2e3 d6c7 f1f8 d8f8');
INSERT INTO evaluations VALUES(48820,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(48820,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(48820,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(48820,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(48820,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(48820,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48820,36,'cp -9','d2d4 e5d4 d1d4 b8c6 f1b5 c8d7 b5c6 d7c6 e4e5 f6d7 g1e2 f8c5 d4g4 d8e7 g4g7 e7e5 g7e5 d7e5 c1f4 f7f6 e1c1 e8c8 e2d4 d8g8 d4b3 c5b4 f4e5 f6e5 c3d5 c6d5 d1d5 g8g2 d5e5 h8g8 c2c3 b4d6 e5f5 g2g1 h1g1 g8g1 c1c2 g1g2 b3d2 g2h2 c2d3 c8d7 f3f4 h7h5 d2c4 d7e6 c4d6 c7d6 f5g5 h5h4 g5h5 h4h3 b2b4 h2h1 d3e4 h1e1 e4f3');
INSERT INTO evaluations VALUES(48806,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(48806,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(48806,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48806,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(48806,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48806,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(48806,36,'cp -20','e5d4 d1d4 b8c6 f1b5 c8d7 b5c6 d7c6 e4e5 f6d7 g1e2 f8c5 d4g4 d7e5 g4g7 e5g6 h2h4 f7f6 c1f4 d8e7 g7e7 g6e7 f4c7 h8g8 g2g3 e8f7 e1c1 a8e8 e2d4 c5d4 d1d4 e7f5 d4b4 f5g3 c7g3 g8g3 c3b5 a7a5 b4b3 a5a4 b3b4 e8e2 b5d4 e2f2 d4c6 b7c6 b4b6 h7h5 b6c6 g3f3 c6d6 f2f1 h1f1 f3f1 c1d2 f1f4 d2e2 f4h4 d6d5 h4e4 e2f3');
INSERT INTO evaluations VALUES(49249,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(49249,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(49249,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(49249,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(49249,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(49249,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(49249,36,'cp 67','d7d6 d2d4 b8d7 a2a4 c7c6 h2h3 f8e7 e1g1 e8g8 f1e1 f8e8 a4a5 e7f8 c1e3 e5d4 e3d4 d8c7 c4f1 c6c5 d4f6 d7f6 c3b5 c7b8 d1d3 c8d7 f3d2 d7b5 d3b5 b8c7 b5d3 a8d8 d2c4');
INSERT INTO evaluations VALUES(49727,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(49727,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(49727,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(49727,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(49727,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(49727,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(49727,36,'cp -112','f8b4 b1d2 e8g8 e1g1 d7d6 d1c2 b8d7 d2b3 d7b6 c4d3 b6a4 b2d4 h7h6 a1d1 c7c5 d4a1 c8e6 a2a3 c5c4 b3d4 c4d3 c2a4 b4c5 d4e6 f7e6 a4b3 d8e8 e4e5 d6e5 f3e5 a8d8 d1d3 d8d3 e5d3 c5b6 a1f6 f8f6 b3c4 e8d7 g2g3 d7d5 c4d5 e6d5 g1g2 f6f7 f1e1');
INSERT INTO evaluations VALUES(42829,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(42829,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(42829,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(42829,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(42829,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(42829,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(42829,36,'cp -110','b1c3 e8g8 d1b3 d7d5 e1c1 d5c4 b3b4 b8a6 b4c4 c8e6 d1d8 e6c4 d8d4 c4e6 f3g5 a8d8 g5e6 f7e6 h1d1 d8d4 d1d4 a6b8 f2f3 b8c6 d4c4 f8d8 c3a4 f6d7 a4c5 d7c5 c4c5 d8d6 f3f4 g8f7 c1c2 g7g6 g2g4 f7e8 g4g5 e8d7 c5c4 b7b6 a2a4 d7c8');
INSERT INTO evaluations VALUES(42823,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(42823,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(42823,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(42823,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(42823,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(42823,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(42823,36,'cp -126','e8g8 d1b3 d7d5 e1c1 d5c4 b3b4 a7a5 b4c4 c8e6 d1d8 e6c4 d8f8 g8f8 f3e5 c4e6 h1d1 b8d7 e5d7 f6d7 c3d5 a8c8 d5f4 f8e7 f4d5 e6d5 d1d5 c7c5 b2g7 c8g8 g7e5 g8g2 e5d6 e7e6 d6c5 d7c5 d5c5 g2f2 c5a5 f2h2 a5b5 h2a2 b5b6 e6e5 b6b7 a2f2 b7e7 e5f6 e7e8 f6g5 c1d1 h7h5 d1e1 f2f6 e4e5 f6f5 e5e6 f7e6 e8e6');
INSERT INTO evaluations VALUES(51741,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(51741,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(51741,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(51741,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(51741,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(51741,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(51741,36,'cp 206','b1c3 c5d4 d1d4 f7f6 c1f4 b8c6 d4d2 f8b4 a2a3 d8a5 f4d6 b4d6 d2d6 a7a6 f1e2 b7b5 b2b4 a5d8 a3a4 b5a4 a1a4 c8b7 b4b5 a6b5 a4a8 b7a8 c3b5 g8e7 b5c7 e8f7 e1g1 e7c8 d6g3 c8b6 c7b5 d8e7 f1d1 e7c5 g3c7 h8b8 b5d6 f7g8 f3d4 c5c3');
INSERT INTO evaluations VALUES(51739,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(51739,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(51739,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(51739,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(51739,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(51739,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(51739,36,'cp 192','c5d4 d1d4 f7f6 c1f4 b8c6 d4d2 f8b4 a2a3 d8a5 f4d6 b4d6 d2d6 a7a6 f1e2 b7b5 e1g1 a5d8 b2b4 g8h6 d6d2 h6f7 a3a4 b5a4 c3a4 e8g8 f1b1 d7d6 c2c4 c8b7 c4c5 d8e7 c5d6 f7d6 a4c5 f8d8 d2a2 a6a5 b4b5 c6b4 a2e6 e7e6 c5e6 d8c8 e4e5 d6e4 e5f6 e4c3 f6f7 g8f7');
INSERT INTO evaluations VALUES(53910,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(53910,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(53910,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(53910,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(53910,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(53910,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(53910,36,'cp 105','g8f6 b1c3 b8c6 c1f4 c8e6 d1d2 d8b6 e1c1 f8e7 f2f3 a7a5 c3a4 b6a7 d2f2 a8d8 f1b5 e8g8 c1b1 f6e8 b5c6 b7c6 e4e5 e6b3 a2b3 d6d5 h1e1 d5d4 f4d2 a7a6 f3f4 a6b5 f2f1 e8c7 f4f5 f8e8 f5f6 e7f8 f6g7 f8g7 f1b5 c6b5 a4c5');
INSERT INTO evaluations VALUES(6496,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6496,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(6496,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(6496,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(6496,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6496,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6496,36,'cp -6','d3c4 g8f6 d2d3 d7d5 e4d5 f6d5 e1g1 e8g8 b1d2 d5b6 c4b3 a7a5 a2a4 c8g4 h2h3 g4h5 g2g4 h5g6 d2e4 c5e7 e4g3 b6d7 f1e1 d7c5 b3c4 e5e4 d3e4 d8d1 e1d1 c5e4 g3e4 g6e4 f3e1 e7d6 c1g5 c6e5 c4e2 a8e8 g5f4');
INSERT INTO evaluations VALUES(7225,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(7225,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(7225,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(7225,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(7225,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(7225,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(7225,36,'cp -28','c8g4 c3b1 g4h5 c1e3 d8e7 b1d2 f6g4 d2c4 g4e3 f2e3 f7f6 c4d6 c7d6 d1e1 e8g8 f3h4 h5g4 h2h3 g4e6 e1g3 e7c7 f1f2 a6a5 a2a4 c7b6 b2b3 d6d5 a1f1 g8h8 e4d5 c6d5 h4f5 e6f5 f2f5 a8c8 f5f2');
INSERT INTO evaluations VALUES(23947,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(23947,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(23947,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(23947,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(23947,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(23947,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(23947,36,'cp -27','c3b1 g4h5 c1e3 d8e7 b1d2 f6g4 d2c4 g4e3 f2e3 f7f6 c4d6 c7d6 d1e1 h5f7 a2a4 e8g8 a4a5 e7c7 f3h4 f7e6 h4f5 d6d5 f5g3 a8d8 f1f2 e6f7 e1e2 f7g6 e2g4 c7d7 g3f5 d5e4 d3e4 g8h8 h2h3 d7c7');
INSERT INTO evaluations VALUES(18904,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(18904,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(18904,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(18904,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(18904,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(18904,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(18904,36,'cp -18','g1f3 a7a6 c4e2 g8f6 g2g3 f8c5 f1g2 d7d6 e1g1 e8g8 e4e5 c6e5 f3e5 d6e5 e2e5 d8b6 e5e2 f8e8 a2a4 c8d7 b1d2 e6e5 d2e4 f6e4 g2e4 d7c6 a4a5 b6c7 e4c6 c7c6 c2c3 a8d8 c1e3 c5e3 e2e3 c6b5 f1d1 b5b2');
INSERT INTO evaluations VALUES(18903,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(18903,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(18903,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18903,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(18903,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(18903,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(18903,36,'cp -27','a7a6 c4e2 g8f6 g2g3 f8c5 f1g2 d7d6 e1g1 e8g8 e4e5 d6e5 f3e5 c6e5 e2e5 d8b6 e5e2 f8e8 a2a4 h7h6 b1d2 e6e5 d2e4 f6e4 e2e4 a8b8 c2c3 c8e6 a4a5 b6c7 f1e1 e8d8 c1e3 f7f5 e4c2 c5e3 e1e3 e5e4');
INSERT INTO evaluations VALUES(19047,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(19047,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(19047,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(19047,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(19047,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(19047,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(19047,36,'cp 75','c6d4 g1f3 d4f3 d1f3 g8f6 c1e3 b7b6 f1b5 f8c5 e1g1 e8g8 h2h3 d7d6 b1c3 a8b8 a1c1 f6e8 b5d3 c8d7 a2a3 d8f6 f3e2 f6g6 g1h2 e8f6 b2b4 c5e3 e2e3 f6h5 c3e2 b8c8 c1c8 f8c8 f1c1 c8c1 e3c1');
INSERT INTO evaluations VALUES(19182,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(19182,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(19182,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(19182,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(19182,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(19182,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(19182,36,'cp 54','g1f3 d4f3 d1f3 g8f6 c1e3 b7b6 b1c3 f8c5 a2a3 c5e3 f3e3 e8g8 f1e2 d7d6 h2h3 c8d7 e1g1 a8c8 f1c1 c8c7 b2b3 d8b8 e2a6 b6b5 c3a2 c7c5 c1c5 d6c5 e3c5 f6e4 c5e3 b8d6 e3e4 d6a6 a2c3 f8e8 b3b4 a6b6 e4e3 b6e3');
INSERT INTO evaluations VALUES(19181,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(19181,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(19181,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(19181,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(19181,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(19181,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(19181,36,'cp 60','d4f3 d1f3 g8f6 c1e3 b7b6 a2a3 f8c5 b1c3 c5e3 f3e3 e8g8 f1e2 d7d6 e1g1 c8d7 a1c1 a8c8 b2b4 c8c7 a3a4 h7h6 h2h3 a7a5 b4a5 b6a5 c3b5 c7c1 f1c1 d8b8 e2f1 f8c8 c1c8 d7c8 e3c3 f6e4 c3a5 c8d7 b5a7 b8b7 a7c6');
INSERT INTO evaluations VALUES(19914,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(19914,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(19914,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(19914,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(19914,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(19914,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(19914,36,'cp -37','g8f6 b1c3 a7a6 b5d3 e7e6 a2a4 f8e7 g1f3 e8g8 e1g1 d8c7 h2h3 b7b6 d1e2 c8b7 f1d1 c6b4 c1f4 a8c8 e2d2 f8d8 a1c1 h7h6 d3f1 f6h5 f4h2 c7c5 d2e2 h5f6 h2f4 e7f8 d1d2 d6d5 e4d5 f6d5 c3d5 b4d5 f4e5 d5b4');
INSERT INTO evaluations VALUES(30179,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(30179,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(30179,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(30179,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(30179,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(30179,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(30179,36,'cp 24','h2h3 g4f3 d2f3 b8d7 c1d2 f8e7 c2c4 e8g8 f1e2 d5c4 e2c4 f6e4 e1g1 e4d2 d1d2 c7c5 a1c1 c5d4 f3d4 d7f6 d2c2 d8b6 c4b3 a8b8 g2g3 g7g6 c2c7 b6c7 c1c7 e7d8 c7c4 d8b6 g1g2 b6d4 c4d4');
INSERT INTO evaluations VALUES(31997,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(31997,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(31997,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(31997,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(31997,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(31997,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(31997,36,'cp 37','g8f6 b1c3 e7e6 f1c4 c7c5 d1a4 b8d7 f3e5 c5d4 e3d4 a7a6 e5g4 f6g4 c4e2 b7b5 a4b3 g4f6 e2f3 a8c8 a2a4 b5b4 c3e4 f6e4 f3e4 a6a5 c1f4 d8b6 e1g1 f8e7 f1c1 e8g8 b3d3 d7f6 e4f3 f8d8 f4e3 f6d5 f3d5 e6d5 d3b5 c8b8 b5b6 b8b6 c1c7 e7d6');
INSERT INTO evaluations VALUES(33085,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(33085,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(33085,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(33085,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(33085,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(33085,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(33085,36,'cp 65','b1c3 c7c6 e1g1 c8g4 c1g5 f6g5 c4f7 e8e7 f3g5 g4d1 c3d1 h7h6 f7g8 h6g5 g8b3 b8d7 d1e3 h8h7 c2c3 c5e3 f2e3 a8h8 f1f7 e7e8 a1f1 h7h2 g1f2 h2h1 f1h1 h8h1 f7g7 h1b1 f2f3 b1b2 d3d4 d7f6 g7g5 b2f2 f3f2 f6e4 f2f3 e4g5 f3g4 g5e4 d4e5 d6e5 c3c4 e4d6 g4g5 b7b5 c4b5 c6b5 g5f6 d6c4 e3e4 a7a5 g2g4 a5a4 b3d1 c4b2 d1e2');
INSERT INTO evaluations VALUES(33081,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(33081,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(33081,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(33081,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(33081,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(33081,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(33081,36,'cp 81','c7c6 e1g1 c8g4 c1g5 f6g5 f3g5 g4d1 c4f7 e8e7 c3d1 h7h6 f7g8 h6g5 g8b3 b8d7 d1e3 h8h7 c2c3 c5e3 f2e3 a8h8 f1f7 e7e8 a1f1 h7h2 g1f2 d7c5 f2g3 h2h1 f1f3 g7g6 b3c2 h1c1 d3d4 g5g4 f3f1 c5d7 c2b3 c1f1 f7f1 h8f8 f1f8 e8f8 g3g4 f8g7 g4g5 d7f6 b3c2 f6h7 g5g4 b7b5 c2d3 a7a5 b2b4');
INSERT INTO evaluations VALUES(35101,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(35101,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(35101,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(35101,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(35101,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(35101,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(35101,36,'cp 113','d7d6 g1f3 c8e6 c4b3 c7c5 d2d3 e6b3 a2b3 b8c6 e1g1 f6d8 f3d2 f8e7 d2c4 g8f6 c4e3 d8d7 f2f4 e5f4 f1f4 e8g8 f4f2 e7d8 h2h3 f8e8 f2f1 a7a6 c1d2 a6a5 c3d5 f6d5 e3d5 c6d4 d2c3 b7b5 c3d4 c5d4 d1h5 e8e5');
INSERT INTO evaluations VALUES(42423,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(42423,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(42423,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(42423,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(42423,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42423,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(42423,36,'cp 28','d1c2 f7f5 c1d2 e4d2 c2d2 d7d6 g2g3 d8f6 f1g2 b8c6 e1g1 e6e5 a1d1 a7a5 d2e3 b4c3 e3c3 e8g8 d4e5 d6e5 f3e1 e5e4 c3f6 g7f6 f2f3 c8e6 d1c1 a8d8 f3e4 f5e4 g2e4 d8d2 e4c6 b7c6 f1f2 d2b2 e2e4 b2f2 g1f2 f6f5');
INSERT INTO evaluations VALUES(42424,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(42424,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(42424,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(42424,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(42424,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42424,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(42424,36,'cp 24','f7f5 f3d2 d7d5 d2e4 f5e4 c1f4 e8g8 e2e3 c7c5 a2a3 b4c3 b2c3 c5d4 c3d4 d8a5 c2d2 a5d2 e1d2 b8c6 d2c3 e6e5 c4d5 e5d4 e3d4 c6e7 f4g3 e7d5 c3d2 c8e6 f1c4 a8d8 c4d5 e6d5 d2e3 d5c6 h1c1 f8f5 a3a4 a7a6 c1c2 h7h6 h2h4 d8d7 a4a5 f5b5 g3e5');
INSERT INTO evaluations VALUES(52023,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(52023,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(52023,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(52023,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(52023,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(52023,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52023,36,'cp 36','d7d5 c2c4 d5d4 e2e3 b8c6 e1g1 f8d6 d2d3 g8f6 e3d4 c5d4 c1g5 h7h6 g5f6 d8f6 b1d2 e8g8 d2e4 f6d8 e4d6 d8d6 a1c1 c8d7 c4c5 d6c7 a2a3 a7a5 b2b3 f8d8 f1e1 d7e8 d1e2 a8b8 h2h4 b7b6 f3d2 c6e7 b3b4 a5b4 a3b4 b6c5 c1c5');
INSERT INTO evaluations VALUES(52124,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52124,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52124,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52124,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(52124,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(52124,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(52124,36,'cp 8','g8e7 f2f4 d7d6 c1e3 e5f4 e3f4 e7g6 f4e3 f8e7 d1d2 b8c6 g1f3 e8g8 e1c1 g6e5 c3d5 e5g4 c1b1 g4e3 d2e3 f8e8 d3d4 c5d4 f3d4 e7g5 e3e1 c8e6 d4e6 f7e6 d5b4 d8a5 b4d3 a5e1 d3e1');
INSERT INTO evaluations VALUES(167,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(167,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(167,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(167,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(167,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(167,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(167,36,'cp 6','c2c4 d5c4 b3c4 g8f6 c4c5 d6c5 b2e5 e8g8 b5c6 b7c6 e1g1 c5d6 e5g3 c6c5 f1e1 c8a6 b1c3 a6d3 d1a4 c5c4 g3h4 e7e6 f3d4 e6e5 f2f4 e5h5 h4f6 g7f6 a4c6 h5g6 d4b5 g8h8 c6f3 a8b8 b5d6 c7d6 a1d1 f6f5 f3g3');
INSERT INTO evaluations VALUES(6758,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(6758,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(6758,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(6758,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(6758,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(6758,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(6758,36,'cp 137','c6e7 b1d2 g8f6 f1d3 c7c6 d2c4 d8c7 d5c6 e7c6 c1g5 d6e7 d3c2 e8g8 e1g1 b7b6 f1e1 h7h6 g5h4 d7d6 h4f6 e7f6 d1d6 c7d6 c4d6 c8g4 e1d1 a8d8 d6b5 a7a6 b5c7 d8c8 c7d5 c6d4 d5f6 g7f6 d1d4 e5d4 f3d4 g4e6 f2f4 f8d8 d4e6 f7e6 c2b3 c8c6 f4f5 a6a5 b3e6 g8h8');
INSERT INTO evaluations VALUES(7011,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(7011,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(7011,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(7011,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(7011,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(7011,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(7011,36,'cp 166','b1d2 g8f6 f1d3 c7c6 d2c4 d8c7 d5c6 e7c6 c1g5 d6e7 d3c2 e8g8 g5f6 e7f6 d1d6 c7d8 f3e5 b7b5 e5c6 d7c6 d6d8 f6d8 c4e5 c6c5 e1c1 b5b4 c3b4 c5b4 e5c6 d8f6 e4e5 c8b7 c6a5 b7g2 h1g1 f6g5 c1b1 g2f3 g1g5 f3d1 c2d1 a8d8 d1b3 d8d2');
INSERT INTO evaluations VALUES(18046,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(18046,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(18046,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(18046,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(18046,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(18046,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(18046,36,'cp 94','c4d5 g8e7 e1g1 b5b4 c3a4 d7d6 d5c4 e7g6 d2d4 c6d4 f3d4 e5d4 d1d4 c8d7 b2b3 c7c5 d4e3 f8e7 a2a3 d7a4 b3a4 e8g8 c1b2 e7f6 b2f6 d8f6 f2f4 a6a5 a1d1 a8b8 a3b4 b8b4 c4b5 f6e7 f1e1 f8b8 g2g3 c5c4 d1d5 h7h6 e1d1');
INSERT INTO evaluations VALUES(18042,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(18042,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(18042,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(18042,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(18042,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(18042,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(18042,36,'cp 110','b5b4 c3e2 g8f6 e1g1 f6d5 e4d5 e5e4 f3e1 c6e5 e2g3 f8e7 d1e2 d7d6 e2e4 a6a5 a2a3 b4a3 e1f3 e5f3 e4f3 e8g8 a1a3 c8d7 b2b3 e7f6 d2d3 a5a4 b3a4 a8a4 a3a4 d7a4 c2c4 d8b8 g3e4 f6e5 f3e3 b8b6 e3b6 c7b6 c1e3 f7f5 d3d4 e5d4 e4d6 d4e5 d6b7 f8c8 f2f4 e5c3 b7d6');
INSERT INTO evaluations VALUES(18856,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(18856,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(18856,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18856,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(18856,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(18856,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(18856,36,'cp -20','d7d5 d2d3 f8d6 e1g1 g8e7 g2g3 f7f6 c2c3 e8g8 c1e3 b7b6 e3f2 c8b7 b1d2 d8d7 a2a3 e6e5 f4e5 d6e5 d1a4 a8d8 f3e5 f6e5 f2e3 c6d4 f1f8 g8f8 a1f1 f8g8 a4d7 d4e2 g1f2 d8d7 f2e2 d5e4 d3e4 b7a6 c3c4 e7c6 f1b1 c6d4 e3d4 d7d4 b2b3 b6b5 c4b5 a6b5 e2e3 b5a6 d2f3 d4d3 e3f2');
INSERT INTO evaluations VALUES(20307,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(20307,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(20307,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(20307,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(20307,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(20307,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(20307,36,'cp -21','d2d3 f8d6 e1g1 g8e7 g2g3 f7f6 c2c3 e8g8 c1e3 b7b6 a2a3 c8b7 e3f2 d8c7 b1d2 g8h8 f1e1 a8d8 e2f1 e6e5 f4f5 d5d4 d2c4 g7g6 f3h4 b7c8 f1h3 f8g8 c4d6 c7d6 a1c1 d6d7 b2b4 g6f5 b4c5 b6c5');
INSERT INTO evaluations VALUES(18986,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(18986,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(18986,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(18986,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18986,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(18986,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18986,36,'cp 13','g8f6 f1e2 d7d5 c4d5 f6d5 e1g1 d5c3 d2c3 d8d1 f1d1 f7f6 e3e4 c8e6 c1e3 f8e7 f3d2 e8f7 f2f3 a8d8 d2c4 b7b6 d1d8 h8d8 g1f2 g7g5 a2a4 h7h5 g2g3 h5h4 f2g2 d8h8 a1d1 h8b8 g3h4 g5h4 g2f2 c6a5 c4a5 b6a5');
INSERT INTO evaluations VALUES(24843,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(24843,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24843,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(24843,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24843,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24843,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(24843,36,'cp 16','f1b5 e7e6 f3e5 a7a6 b5c6 b7c6 e1g1 d8c8 c2c4 c8b7 c4c5 h7h5 d1b3 b7b3 a2b3 h5h4 b1c3 h4h3 g2g3 f6g4 e5g4 f5g4 c3d1 f7f6 d1f2 g4e2 f1e1 e2f3 c1d2 g7g5 a1a3 f8e7 d2a5 e8d7 a5c3 d7c8 a3a5 c8b7 a5a1 g5g4 a1a4 a8b8 e1c1 f3e2 c3e1 e2b5');
INSERT INTO evaluations VALUES(25926,30,'cp 294',NULL);
INSERT INTO evaluations VALUES(25926,31,'cp 298',NULL);
INSERT INTO evaluations VALUES(25926,32,'cp 314',NULL);
INSERT INTO evaluations VALUES(25926,33,'cp 303',NULL);
INSERT INTO evaluations VALUES(25926,34,'cp 288',NULL);
INSERT INTO evaluations VALUES(25926,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(25926,36,'cp 301','e4d5 e6d5 c3d5 c6d4 f3d4 c5d4 d1e2 f8e7 e1g1 e8f8 f1d1 d7g4 e2g4 d8d5 c2c4 h7h5 g4d4 d5d4 d1d4 g8f6 c1d2 e7c5 d4f4 h5h4 a1d1 h4h3 g2g3 c5d6 f4f5 d6e7 d2e1 a7a6 b5a4 a8c8 f5f4 g7g5 f4d4 e7c5 d4d8 c8d8 d1d8 f8g7 d8h8 g7h8');
INSERT INTO evaluations VALUES(25923,30,'cp 284',NULL);
INSERT INTO evaluations VALUES(25923,31,'cp 302',NULL);
INSERT INTO evaluations VALUES(25923,32,'cp 278',NULL);
INSERT INTO evaluations VALUES(25923,33,'cp 282',NULL);
INSERT INTO evaluations VALUES(25923,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(25923,35,'cp 296',NULL);
INSERT INTO evaluations VALUES(25923,36,'cp 273','e6d5 c3d5 c6d4 f3d4 c5d4 d1e2 f8e7 e1g1 e8f8 f1d1 d7g4 e2g4 d8d5 c2c4 h7h5 g4d4 d5d4 d1d4 g8f6 c1d2 e7c5 d4f4 h5h4 h2h3 h8h5 a1e1 c5d6 f4f3 a7a6 b5a4 h5c5 b2b3 a8d8 f3d3 d6h2 g1h2 d8d3 d2b4 b7b6 e1e5 f8g8 b4c5 b6c5 e5c5 d3d2 c5c8 g8h7');
INSERT INTO evaluations VALUES(25962,30,'cp 289',NULL);
INSERT INTO evaluations VALUES(25962,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(25962,32,'cp 310',NULL);
INSERT INTO evaluations VALUES(25962,33,'cp 288',NULL);
INSERT INTO evaluations VALUES(25962,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(25962,35,'cp 297',NULL);
INSERT INTO evaluations VALUES(25962,36,'cp 307','c3d5 c6d4 f3d4 c5d4 d1e2 f8e7 e1g1 e8f8 f1d1 d7g4 e2g4 d8d5 c2c4 h7h5 g4d4 d5d4 d1d4 g8f6 c1d2 h5h4 b2b4 h4h3 c4c5 h3g2 a1d1 a8c8 b5e2 b7b6 d2e3 b6c5 b4c5 g7g5 g1g2 h8h4 d4h4 g5h4 e2f3 e7c5 d1c1 f6d7 f3g4 f8e7 e3g5 f7f6 c1e1 e7d8 g5h4 c5d6 h4f6 d7f6 g4c8');
INSERT INTO evaluations VALUES(26899,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(26899,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(26899,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(26899,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(26899,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(26899,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(26899,36,'cp 73','d6e5 d1d8 a8d8 b5c6 b7c6 c1e3 a7a5 h2h3 g4h5 g2g4 h5g6 b1d2 f7f6 e1e2 h7h5 a1g1 g6f7 g4g5 a5a4 f3e1 g8e7 e1d3 e7g6 b2b3 d8a8 b3b4 f8d6 a2a3 h8g8 h3h4 e8e7 g5f6 g7f6 d3c5 g6f4 e3f4 e5f4 g1g8 f7g8 h1g1 g8f7 c5d3 f7e6 g1g7 e7f8');
INSERT INTO evaluations VALUES(26859,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(26859,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(26859,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(26859,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(26859,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(26859,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(26859,36,'cp 97','b5c6 b7c6 d1d8 a8d8 c1e3 a7a5 h2h3 g4h5 a2a4 f7f6 b1d2 h5f7 e1c1 f8d6 d2b3 f7b3 c2b3 g8e7 c1b1 e8f7 b1a2 d8d7 d1c1 h8d8 c1c4 d6b4 h1c1 f7e8 e3c5 b4c5 c4c5 d8a8 c1c2 d7d1 f3d2 d1e1 d2c4 e1e4');
INSERT INTO evaluations VALUES(33030,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33030,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(33030,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(33030,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(33030,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(33030,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33030,36,'cp 0','d1d2 c8d7 b1c3 g8e7 f1d3 e7g6 h2h4 b6c7 d2e2 b8c6 c3b5 c7b8 d3g6 f7g6 c1e3 c5e7 e3g5 e7c5');
INSERT INTO evaluations VALUES(35110,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(35110,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(35110,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(35110,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(35110,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(35110,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(35110,36,'cp 141','g1f3 b8c6 e2e3 f6d8 d2d4 d7d6 f1e2 g8f6 d4d5 c6b8 e3e4 a7a5 c1e3 f6g4 e3d2 b8a6 a2a3 c8d7 e1g1 g4f6 b2b4 f8e7 d1b3 e8g8 h2h3 a5b4 a3b4 c7c5 b4b5 a6b4 a1a8 d8a8 b3b2 h7h6 f1a1 a8e8 b2b3 e8b8 e2f1 f6h7 g2g3 f7f5 f1g2');
INSERT INTO evaluations VALUES(35108,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(35108,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(35108,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(35108,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(35108,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(35108,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(35108,36,'cp 129','b8c6 e2e3 f6d8 d2d4 d7d6 f1e2 g8f6 d4d5 c6b8 e3e4 a7a5 h2h3 b7b6 c1e3 b8a6 g2g4 g7g6 a2a3 f8g7 f3d2 c8d7 d1c2 h7h5 h1g1 h5g4 h3g4 a6c5 e1c1 h8h3 c1b1 e8f8 e3g5 f8e7 c3b5 d7b5 c4b5');
INSERT INTO evaluations VALUES(35566,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(35566,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35566,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35566,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(35566,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35566,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35566,36,'cp 0','a2a4 c8d7 g5e3 b8c6 h2h3 a8c8 c4d3 f8e8 d4c6 d7c6 f1e1 f6d7 d3f1 h7h6 a4a5 d8c7 d1d2 d7c5 e3d4 e7g5 d4e3 g5h4 e3d4 h4g5');
INSERT INTO evaluations VALUES(41230,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(41230,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(41230,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(41230,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41230,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(41230,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41230,36,'cp 14','f2f4 b8c6 a2a3 e7e5 g1f3 c8g4 h2h3 g4f3 d1f3 f6h5 f4f5 c6d4 f3f2 g6f5 c1e3 f5f4 e3d4 e5d4 c3e2 d8g5 g2g4 f4g3 f2f7 e8d8 c4e6 h5f6 h1f1 h8f8 f1f5 g5f5 f7f8 g7f8 e6f5 c7c5 e2g3 d8c7 e1f2 f8h6 f5g4 a8e8 f2e2 h6g5 a1f1 b7b5 g3h5 f6h5 g4h5 e8e7');
INSERT INTO evaluations VALUES(45369,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(45369,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(45369,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(45369,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(45369,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(45369,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(45369,36,'cp -13','g1f3 e7e6 d4c5 f8c5 f1d3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a6 h2h3 h7h6 b2b4 c5a7 e2g3 c8e6 c1d2 f6d7 d1e2 a8c8 a1d1 c6e7 f3h4 e7g6 h4f5 d7f6 e2f3 a7b8 a3a4 c8c6 b4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(45657,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(45657,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(45657,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45657,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45657,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(45657,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(45657,36,'cp 14','f8d6 b1d2 e8g8 b2b4 d8e7 f1e2 b8d7 c1b2 e6e5 d4e5 d7e5 e1g1 e5c4 e2c4 d5c4 b2f6 g7f6 d2c4 d6c7 f3d4 e7e4 d1e2 f8d8 f1d1 e4g4 e2g4 c8g4 f2f3 g4e6 a1c1 d8d5 d4e6 f7e6 e3e4 d5d8 h2h3 b7b5 c4e3 d8d1 c1d1');
INSERT INTO evaluations VALUES(40553,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(40553,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40553,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(40553,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40553,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(40553,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(40553,36,'cp 23','b1d2 b8d7 b2b4 d8e7 c1b2 e8g8 c4c5 d6c7 f1d3 e6e5 d4e5 d7e5 b2e5 c7e5 f3e5 e7e5 e1g1 c8g4 d1c2 g4h5 a1e1 h5g6 f2f4 g6d3 c2d3 e5h5 e3e4 d5e4 d2e4 h5f5 d3f3 f6e4 f3e4 f5f6 e4e5 f6e5 e1e5 f7f6 e5e6 a8d8 f1e1 a7a5 e6e7 f8f7 e7e8 f7f8 e8d8 f8d8 e1e7 d8d3 e7b7 d3a3 b4b5');
INSERT INTO evaluations VALUES(48295,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(48295,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(48295,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(48295,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48295,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48295,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(48295,36,'cp 7','g1f3 c7c5 e2e3 d8b6 a1b1 f8d6 f4d6 b6d6 d4c5 d6c5 c2c4 e8g8 b1c1 b8c6 c4d5 c5d5 d1b3 d5b3 d2b3 c8d7 b3c5 f8c8 c5d7 f6d7 e1d2 c6e5 f1b5 e5f3 g2f3 d7b6 h2h4 c8c1 h1c1 a8c8 c1c8 b6c8 d2c3 c8d6 b5d3 b7b6 c3d4 g8f8 d3h7 g7g6 e3e4');
INSERT INTO evaluations VALUES(53159,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(53159,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53159,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(53159,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53159,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(53159,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53159,36,'cp 11','b8c6 e2e3 e7e5 e3d4 e5d4 f1d3 f8c5 e1g1 g8e7 b1d2 e8g8 f1e1 c8f5 d2e4 c5b4 c1d2 e7g6 h2h3 f8e8 e4c5 e8e1 f3e1 b4d2 d1d2 f5d3 c5d3 d8f6 b2b3 a8e8 e1f3 f6f5 a1e1 e8e1 f3e1 f7f6 f2f4 g8f7 g2g4 f5e4 g1f2 a7a5 e1f3 h7h5');
INSERT INTO evaluations VALUES(19747,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(19747,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(19747,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(19747,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19747,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19747,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(19747,36,'cp 28','e2e3 e7e5 f3e5 c6e5 e3d4 e5c6 c1e3 g8h6 d4d5 h6f5 d5c6 d8d1 e1d1 f5e3 f2e3 b7c6 f1d3 f8c5 d1c2 e8g8 b1c3 f8e8 c3e4 c5e7 h2h3 a7a5 e4c3 c8e6 g2g4 g7g6 b2b3 e7c5 a1d1 a8d8 h1e1 h7h6 d3e4 d8c8 c2b2 c5b4 d1d4 c8c7');
INSERT INTO evaluations VALUES(19746,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(19746,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19746,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(19746,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(19746,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(19746,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(19746,36,'cp 20','e7e5 e3d4 e5d4 f1d3 f8c5 e1g1 g8e7 b1d2 e8g8 d2e4 c5b6 h2h3 e7g6 f1e1 c8f5 c4c5 f5e4 e1e4 b6c5 c1d2 f8e8 d1e2 c5b4 d2b4 c6b4 a1e1 e8e4 d3e4 d4d3 e2d2 b4c6 d2d3 d8d3 e4d3 g6f4 d3e4 f4e6 e4d5 a8e8 d5e6');
INSERT INTO evaluations VALUES(20487,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(20487,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(20487,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20487,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(20487,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(20487,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(20487,36,'cp 39','e3d4 e5d4 f1d3 f8c5 e1g1 g8e7 b1d2 e8g8 d2e4 b7b6 f1e1 c5b4 c1d2 b4d2 d1d2 c8g4 f3g5 g4f5 a1d1 c6e5 f2f4 e5d3 d2d3 h7h6 g5f3 d8c7 f3d4 c7f4 d3d2 f4d2 d1d2 f8d8 e4g3 f5e6 e1e2 a8c8 d4e6 d8d2 e2d2 f7e6 d2d7');
INSERT INTO evaluations VALUES(20486,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(20486,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(20486,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(20486,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(20486,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20486,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(20486,36,'cp 26','e5d4 f1d3 f8c5 e1g1 g8e7 b1d2 e8g8 d2e4 b7b6 f1e1 c5b4 c1d2 b4d2 d1d2 c8g4 f3g5 g4f5 a1d1 d8d7 e4g3 f5d3 d2d3 e7g6 h2h4 h7h6 h4h5 h6g5 h5g6 a8e8 e1e8 f8e8 a2a3 g5g4 g6f7 d7f7 b2b4 f7e6 d3c2 e8d8 g3f5 c6e5 f5d4 e6c4');
INSERT INTO evaluations VALUES(20488,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(20488,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(20488,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(20488,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(20488,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(20488,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(20488,36,'cp 25','f1d3 f8c5 e1g1 g8e7 b1d2 b7b6 d2e4 e8g8 f1e1 c5b4 c1d2 b4d2 e4d2 c8f5 d3f5 e7f5 a2a3 d8f6 d2e4 f6g6 c4c5 f8e8 c5b6 a7b6 d1b3 c6e5 e4d2 e5f3 d2f3 g6f6 e1e8 a8e8 a1d1 g7g6 b3b5 e8c8 h2h3 c8c5 b5d3 g8g7');
INSERT INTO evaluations VALUES(53190,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53190,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53190,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(53190,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53190,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53190,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53190,36,'cp 13','b8c6 d2d3 a7a6 c1e3 e7e6 d3d4 c6d4 f3d4 c5d4 d1d4 d8d4 e3d4 c8d7 c3e4 g8h6 e4g3 d7c6 f2f3 h8g8 c2c3 e8c8 e1f2 f8d6 a1d1 g7g6 d4f6 d8d7 c4b3 d6c5 f2e2 c6b5 c3c4 d7d1 h1d1 b5c6 d1d2 h6f5 g3e4 c6e4 f3e4 c5e7 e4f5 e7f6 f5e6');
INSERT INTO evaluations VALUES(56316,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(56316,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(56316,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(56316,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(56316,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(56316,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(56316,36,'cp -35','c7c5 f1b5 c8d7 b5d7 d8d7 g1f3 b8c6 e1g1 g8e7 d2d4 c5d4 b1d2 e7g6 d2b3 f8e7 f1e1 e8g8 b3d4 f8e8 c1d2 c6d4 f3d4 e7c5 d4f3 h7h6 a1c1 a8c8 d1b3 g6f8 h2h3 e8d8 d2a5 b7b6 a5d2 f8e6 b3d3 d5d4 b2b4 c5f8 c1c8 d8c8');
INSERT INTO evaluations VALUES(54564,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(54564,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(54564,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(54564,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(54564,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(54564,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(54564,36,'cp -25','f1b5 c8d7 b5d7 d8d7 g1f3 b8c6 e1g1 g8e7 d2d4 c6d4 f3d4 c5d4 b1d2 e7c6 d2b3 f8e7 f1e1 e8g8 b3d4 e7c5 d4c6 b7c6 c1e3 c5e3 e1e3 d5d4 a1c1 d7d5 e3e2 a8d8 d1b3 d5a5 c1d1 c6c5 f2f4 a5b6 b3b6 a7b6 a2a4 f8e8');
INSERT INTO evaluations VALUES(56654,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(56654,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(56654,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56654,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(56654,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(56654,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(56654,36,'cp 9','e3d4 g8f6 f1c4 f8d6 g1f3 e8g8 e1g1 b8c6 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 a1c1 a8b8 c3e4 d6e7 e4f6 e7f6 h4f6 d8f6 c4d5 f8e8 e1e8 b8e8 h2h3 g4f3 d1f3 f6f3 g2f3 e8e2 d5c6 b7c6 c1c6 e2b2 d4d5 b2a2 c6c7 a2d2 c7a7 d2d5 a7f7 d5d1 g1g2 d1d2 f7c7 h8h7');
INSERT INTO evaluations VALUES(56651,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(56651,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(56651,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(56651,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(56651,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56651,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56651,36,'cp 7','g8f6 f1c4 f8d6 g1f3 b8c6 e1g1 e8g8 c1g5 h7h6 g5h4 c8g4 f1e1 g8h8 h2h3 g4f3 d1f3 g7g5 h4g3 d6g3 f3g3 d8d4 c4b3 d4d6 g3d6 c7d6 f2f4 h8g7 f4g5 h6g5 e1f1 c6d4 a1d1 d4b3 d1d6 a8d8 d6f6 b3d2 f1f2 d8d7 g2g4 d7d3 g1g2 b7b6 f6c6 d3d7 f2f5 f7f6 c6e6 f8d8');
INSERT INTO evaluations VALUES(48949,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(48949,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(48949,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(48949,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(48949,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(48949,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(48949,36,'cp 13','f1c4 f8d6 g1f3 b8c6 e1g1 e8g8 h2h3 h7h6 c1e3 c8f5 f3h4 f5h7 d1f3 f6e8 h4f5 d8f6 c4d3 h7f5 f3f5 f6f5 d3f5 e8f6 c3e4 d6e7 a1c1 a8d8 e4f6 e7f6 f5e4 f6d4 e4c6 b7c6 e3d4 d8d4 c1c6 d4d2 c6c7 f8b8 c7a7 b8b2 a2a4 g7g6 a7a8 g8g7 a4a5 b2a2 a5a6 a2a4');
INSERT INTO evaluations VALUES(57493,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(57493,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(57493,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(57493,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(57493,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(57493,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(57493,36,'cp 26','d7d6 e4f5 c8f5 g1e2 f5d3 e2g3 d3f1 g3f1 b8c6 f1g3 d8d7 e1g1 g8f6 d2d3 f8e7 c3d5 e8g8 c1e3 e7d8 d1b3 b7b6 a1d1 f8f7 d5c3 d8e7 d3d4 c6a5 b3b5 d7b5 c4b5 a5c4 e3c1 e5d4 d1d4 c4e5 b2b3 a8e8 d4d1 e7f8 h2h3 e5g6 c1e3 f6d7 c3d5 d7f6');
INSERT INTO evaluations VALUES(56739,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(56739,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(56739,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(56739,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(56739,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(56739,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(56739,36,'cp 28','f1d3 g8e7 d3c2 c7c5 e4f5 c8f5 c2f5 e7f5 d2d3 d8d7 g1e2 f8e7 e2g3 b8c6 a2a3 e8g8 d1g4 c6d4 e1g1 a8d8 g3f5 d4f5 g4e4 e7f6 c1e3 b7b6 a1b1 f8f7 b2b4 d8f8 c3d5 f5e7 b4b5 e7d5 e4d5 f6d8 h2h3 g8h8');
INSERT INTO evaluations VALUES(3387,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3387,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(3387,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(3387,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(3387,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(3387,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3387,36,'cp 21','a2a4 c8e6 b1d2 e6c4 d2c4 d6d5 c4e5 c6e5 f3e5 d5e4 d3e4 f8e8 d1d3 e7c5 d3d8 a8d8 e5d3 c5d4 c1g5 e8e4 f1e1 e4e1 a1e1 h7h6 g5f6 d4f6 e1e4 g8f8 e4c4 d8d7 g1f1 f8e8 f1e2 d7e7 e2f3 e8d7 g2g3 b7b6 b2b3 a7a5 h3h4 d7d6 c4f4 e7e6 h4h5');
INSERT INTO evaluations VALUES(7428,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(7428,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(7428,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(7428,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(7428,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(7428,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(7428,36,'cp 46','a2a4 b5b4 f3e5 c6e5 d3d4 c5d4 d1d4 d8e7 c1f4 e5c6 d4d3 d7d6 b1d2 e8g8 f4g3 c8b7 f1e1 f8e8 a1d1 f6d7 b3d5 d7c5 d3c4 a6a5 f2f4 b7a6 c4a2 b4b3 d2b3 c6b4 a2b1 c7c6 b3c5 d6c5 d5a2 b4a2 b1a2 a8d8 a2a1 d8d1 a1d1 e8d8 d1h5 d8d2');
INSERT INTO evaluations VALUES(9307,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(9307,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(9307,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(9307,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9307,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(9307,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(9307,36,'cp 12','h7h6 g5h4 g7g5 h4g3 a7a6 b1d2 c5a7 a2a4 e8g8 e1g1 f6h7 d3d4 e5d4 c3d4 c6d4 f3d4 a7d4 d1c2 d4g7 e4e5 g7e5 g3e5 d6e5 a1e1 d8f6 c4d3 f6g7 c2c7 h7f6 d3c4 f8e8 f2f3 c8e6 c4e6 e8e6 c7b7 a8d8 d2c4 f6d5');
INSERT INTO evaluations VALUES(8810,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(8810,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(8810,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(8810,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(8810,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8810,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(8810,36,'cp 23','g5h4 g7g5 h4g3 a7a6 b1d2 c5a7 a2a4 e8g8 e1g1 g5g4 f3h4 f6h5 g1h1 d8g5 h4f5 c8f5 e4f5 g8h8 f5f6 h5f6 f2f3 d6d5 c4b3 a8d8 d1e1 g5g6 g3h4 d8d6 e1e2 g6h5 h4f6 d6f6 f3g4');
INSERT INTO evaluations VALUES(12530,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(12530,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(12530,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(12530,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12530,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(12530,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12530,36,'cp 14','f8e7 e1g1 e8g8 f1e1 d7d6 a2a4 c8e6 b1a3 d8c7 c1g5 a8d8 a3c2 d6d5 e4d5 f6d5 g5e7 c7e7 f3e5 c6e5 e1e5 e7f6 e5e4 e6f5 c4d5 d8d5 e4e3 h7h6 d1e2 f6g6 c2e1 f8d8 a1d1 g6c6 b2b3 c6b6 e2c2 b6g6 h2h3 b7b6 c2e2 d5d6 e3e8 g8h7 e8e7 g6f6 e7a7 f6c3 a7f7');
INSERT INTO evaluations VALUES(7940,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(7940,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7940,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(7940,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(7940,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7940,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7940,36,'cp 14','e1g1 e8g8 f1e1 d7d6 a2a4 h7h6 c1d2 c8e6 c4e6 f7e6 b1a3 a7a6 h2h3 d8d7 a3c4 d7c7 d1e2 b7b5 a4b5 a6b5 c4e3 c7b7 e3g4 f6d7 e1b1 g8h7 b2b4 a8a1 b1a1 f8a8 e2d1 a8a1 d1a1 c5b4 c3b4 d7f8 a1a2 b7d7');
INSERT INTO evaluations VALUES(13575,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(13575,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13575,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13575,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(13575,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13575,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(13575,36,'cp 20','c2c3 a7a6 f3e5 f8d6 e5c6 b7c6 f4d6 d8d6 d1a4 c5d4 c3d4 c8d7 f1d3 e6e5 d4e5 d6e5 a4c2 e8g8 e1g1 a6a5 h2h3 e5g5 d2f3 g5h5 d3e2 f8c8 a1c1 c6c5 c2d1 d7e6 b2b3 h5h6 e2d3 h6h5 f1e1 g8h8 g1h1 h7h6 c1c2');
INSERT INTO evaluations VALUES(14543,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(14543,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(14543,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(14543,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(14543,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(14543,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(14543,36,'cp 131','g5f6 g7f6 c4d5 e6d5 e2e3 a7a6 g2g3 c6e7 d1c2 c7c6 f3h4 h6h5 a2a3 f8h6 h2h3 h6g5 h4g2 c8f5 f1d3 d8c8 h3h4 g5h6 c3a4 f5d3 c2d3 c8g4 d3b3 a8b8 e1g1 e8g8 b3d1 f8d8 a4c5 e7f5 c5d3 g4d1 f1d1 f5d6 g2f4 h6f4 d3f4');
INSERT INTO evaluations VALUES(14542,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(14542,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(14542,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(14542,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(14542,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(14542,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(14542,36,'cp 143','g7f6 c4d5 e6d5 e2e3 a7a6 g2g3 h6h5 f1g2 c6e7 e1g1 c7c6 d1b3 f6f5 f1e1 e7g6 h2h4 f8d6 f3g5 d6e7 g5h3 e7d6 c3a4 a8b8 a4c5 e8f8 a1c1 f8g7 a2a3 a6a5 b3d1 d6c7 c5d3 b8a8 c1c3 c8d7 c3b3 b7b6 d3f4 g6f4 h3f4');
INSERT INTO evaluations VALUES(16011,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(16011,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(16011,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(16011,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(16011,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(16011,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(16011,36,'cp 38','b1c3 c7c5 d4d5 d8a5 a2a4 f6e4 d1d3 d7f6 f4d2 a5b4 e2e3 e7e5 d5e6 c8e6 c3e4 b4e4 f3g5 e4d3 f1d3 e6d7 e1g1 d6d5 c2c4 h7h6 g5f3 e8c8 f3e5 d7e6 d2a5 b7b6 a5c3 f6e4 e5c6 d8d7 c4d5 e6d5 d3b5 e4c3 b2c3');
INSERT INTO evaluations VALUES(16899,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(16899,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(16899,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(16899,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(16899,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(16899,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(16899,36,'cp 121','f3d4 e5d4 e4e5 f6d5 d2d3 d5b6 f1e1 a7a6 b5c4 b6c4 d3c4 d7d5 e5d6 c8e6 d6c7 d8c7 d1d4 f8d6 b1c3 e8g8 c3d5 e6d5 c4d5 c7c2 c1e3 a8d8 g2g3 h7h5 g1g2 c2f5 a1d1 f8e8 d4d3 f5d3 d1d3 d8c8 a2a3 b7b5 g2f1 f7f6 h2h3 g8f7 e1e2');
INSERT INTO evaluations VALUES(16893,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(16893,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(16893,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(16893,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(16893,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(16893,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(16893,36,'cp 130','e5d4 d2d3 c7c6 e4e5 c6b5 c1g5 f8e7 e5f6 e7f6 f1e1 e8f8 g5f6 d8f6 d1h5 b7b6 h5b5 c8b7 b5d7 b7c6 d7c7 g7g6 f2f3 f8g7 c7e7 h7h5 b1d2 h8c8 a1c1 c6d5 e7f6 g7f6 a2a3 c8c6 c2c4 d4c3 c1c3 a8c8 e1c1 h5h4 g1f2 g6g5 f2e3 f6f5 c3c6 d5c6 b2b4 f7f6');
INSERT INTO evaluations VALUES(17220,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(17220,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(17220,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(17220,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(17220,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(17220,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(17220,36,'cp 13','d2d3 e7e5 g1f3 f8c5 c2c3 e8g8 e1g1 d7d6 f1e1 c5b6 h2h3 c6e7 d3d4 d6d5 f3e5 f6e4 b1d2 e4d6 b5d3 f7f6 e5f3 c8f5 d2b3 f5d3 d1d3 f8e8 a2a4 a7a5 c1f4 d6e4 h3h4 d8d7 f3d2 e4d2 b3d2 d7g4 f4g3 g4f5 d3f3 f5f3 d2f3 e7f5');
INSERT INTO evaluations VALUES(18677,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18677,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18677,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18677,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(18677,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18677,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18677,36,'cp 1','d7c6 d2d3 f8d6 c1e3 g8f6 a2a4 c8e6 e1g1 h7h6 a4a5 f6d7 d1e1 b7b5 a5b6 a7b6 a1a8 d8a8 e1c3 e8g8 b1d2 b6b5 b2b3 a8b7 f1a1 f8a8 a1a8 b7a8 c3b2 d6f8 d2f1 f7f6 h2h3 f8d6 f3h4 a8a6 f1g3 c5c4 b3c4 b5c4 h4f5 d6f8');
INSERT INTO evaluations VALUES(20735,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(20735,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20735,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20735,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20735,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20735,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20735,36,'cp 0','d2d3 f8d6 c1e3 g8f6 a2a4 c8e6 e1g1 a7a5 b1a3 d8e7 f3d2 f6g4 d2c4 g4e3 c4d6 e7d6 f2e3 f7f6 d1e1 b7b6 e1c3 e8c8 a3c4 e6c4 c3c4 c8b7 f1f2 d8d7 a1f1 h8d8 f2f5 d6e7 h2h3 d7d6 f5f2 d6d7');
INSERT INTO evaluations VALUES(18825,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(18825,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(18825,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(18825,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(18825,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(18825,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(18825,36,'cp -12','d7d5 b1d2 g8f6 g2g3 f8e7 f1g2 b7b5 e4d5 e6d5 e1g1 e8g8 f1e1 a7a5 d2f1 c8e6 d3d4 c5d4 f3d4 c6d4 d1d4 f6e4 g2e4 d5e4 d4e4 f8e8 c1e3 e7f6 a1d1 d8c8 e4f4 e6a2 e3d4 f6d4 e1e8 c8e8 d1d4 a2e6 f1e3');
INSERT INTO evaluations VALUES(20295,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(20295,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(20295,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(20295,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(20295,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(20295,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(20295,36,'cp 0','b1d2 f8e7 g2g3 g8f6 f1g2 e8g8 e1g1 b7b5 f1e1 c8b7 e4d5 f6d5 a2a4 a7a6 d2b3 h7h6 h2h4 d8c7 d1e2 f8d8 f3d2 a8c8 d2e4 c7b6 a4a5 b6a7 c1h6 g7h6 e2g4 g8h7 g4h5 h7g7 h5g4');
INSERT INTO evaluations VALUES(21695,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(21695,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(21695,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(21695,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(21695,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(21695,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(21695,36,'cp 118','d4e5 c6e5 d1d5 d8d5 g2d5 g8e7 d5g2 e7c6 g1f3 e5f3 g2f3 c8d7 c2c3 e8c8 c1f4 f8c5 b1d2 h8e8 e2e3 h7h6 h2h4 c6e5 f4e5 e8e5 e1c1 a7a5 a2a3 a5a4 f3e2 e5e7 g3g4 f7f6 g4g5 b7b5 g5h6 g7h6 d2f3 e7e8');
INSERT INTO evaluations VALUES(21655,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(21655,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(21655,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(21655,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(21655,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(21655,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(21655,36,'cp 134','c6e5 d1d5 d8d5 g2d5 g8e7 d5e4 e7c6 g1f3 e5f3 e4f3 c8d7 c2c3 f8c5 c1f4 e8c8 b1d2 h8e8 e2e3 c6e5 f3e2 d7c6 f4e5 e8e5 h1f1 c6a4 h2h4 g7g6 d2f3 e5e7 h4h5 d8e8 f3d4 a7a5 f1h1 c5b6 h1h4 a4d7 h4f4 f7f5 e1c1 c7c6 d4f3 b6c5 h5g6 h7g6 f4h4');
INSERT INTO evaluations VALUES(22186,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(22186,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(22186,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(22186,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(22186,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(22186,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(22186,36,'cp 131','d1d5 d8d5 g2d5 g8e7 d5e4 e7c6 g1f3 e5f3 e4f3 c8d7 c2c3 f8c5 c1f4 e8c8 b1d2 h8e8 e2e3 c6e5 f3e2 d7c6 f4e5 e8e5 h1g1 h7h6 e1c1 c6a4 b2b3 a4d7 c1b2 c5f8 d2f3 e5e8 f3d4 c7c5 d4b5 a7a6 b5d6 f8d6 d1d6 c8c7 g1d1');
INSERT INTO evaluations VALUES(21794,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(21794,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(21794,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(21794,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(21794,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(21794,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(21794,36,'cp 179','d4d5 c6b8 f3e5 b7b5 b1c3 f8b4 a2a4 d8e7 e5g4 b8a6 a4b5 c8g4 d1g4 g8f6 g4h4 a6c5 f1c4 f6e4 h4e7 e8e7 e1g1 e4d6 c4e2 h8e8 c1e3 e7f8 f1d1 c5b3 a1a4 b4c5 e3c5 b3c5 a4a3 e8e5 f2f3 c5d7 d1a1 d7b6 f3f4 e5e7 a3a7 a8a7 a1a7 b6d5 c3d5 e7e2 a7a8 d6e8');
INSERT INTO evaluations VALUES(21568,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(21568,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(21568,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(21568,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(21568,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(21568,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(21568,36,'cp 170','c6b8 f3e5 b7b5 b1c3 f8b4 a2a4 d8e7 e5g4 b8a6 a4b5 c8g4 d1g4 g8f6 g4h4 a6c5 f1c4 e7e4 h4e4 f6e4 e1g1 e4d6 c4e2 c5b3 a1a6 e8g8 c1e3 f8e8 e2d1 b3a5 c3a2 e8e3 f2e3 b4d2 e3e4 a5c4 d1f3 d2e3 g1h1 a8e8 a2b4 c4d2 f1d1 d2e4 f3e4 e8e4 b4c6 d6b5 d1e1 b5d6 g2g3 g7g6 b2b4');
INSERT INTO evaluations VALUES(25459,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(25459,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(25459,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(25459,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(25459,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(25459,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25459,36,'cp 52','b7c6 b5a4 f8e7 e1g1 e8g8 d1d3 f6g4 c1d2 g4e5 d3e2 c6c5 f2f4 d7a4 c3a4 e5c6 a1d1 f8e8 d2e3 d8c8 a4c3 a8b8 b2b3 e7f6 e4e5 c6d4 e2f2 d6e5 c3e4 c8f5 f4e5 f5e4 e3d4 c5d4 e5f6 c7c5 f1e1 e4e3 f6g7 e3f2 g1f2 e8e1');
INSERT INTO evaluations VALUES(24380,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(24380,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(24380,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(24380,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(24380,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(24380,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(24380,36,'cp 54','b5a4 f8e7 d1d3 e8g8 e1g1 f6g4 b2b3 e7f6 c1d2 g4e5 d3e2 c6c5 a4d7 e5d7 a1e1 f8e8 g1h1 a7a5 e2b5 f6d4 b5d3 d7e5 d3g3 e5c6 c3d5 c6b4 c2c3 b4d5 c3d4 d5f6 e4e5 d6e5 d4e5 d8d2 e5f6 g7g6');
INSERT INTO evaluations VALUES(25710,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25710,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(25710,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(25710,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(25710,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(25710,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(25710,36,'cp 60','b5c6 d7c6 d3d4 f7f5 d4e5 d6e5 f3e5 d8d1 e1d1 c6e4 c3e4 f5e4 d1e2 f8d6 e5c4 e8c8 h1d1 g8f6 c4d6 d8d6 c1f4 d6c6 a1c1 h8e8 d1d4 a6a5 f4e3 b7b6 c2c3 h7h6 c1h1 e8g8 h2h3 g8e8 h1d1 g7g5 d1h1 f6h5 g2g3 c6c5 h3h4 h5f6 h4g5 h6g5');
INSERT INTO evaluations VALUES(25702,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(25702,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(25702,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(25702,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(25702,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(25702,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(25702,36,'cp 49','d7c6 d3d4 f7f5 d4e5 d6e5 f3e5 d8d1 e1d1 c6e4 c3e4 f5e4 d1e2 g8f6 c1e3 f8d6 e5c4 e8c8 a1d1 d6e7 d1d8 h8d8 c4d2 d8d6 h1d1 d6e6 d2b3 e6e5 c2c3 g7g6 c3c4 f6g4 e3d4 e5e6 h2h3 g4e5 d4e5 e6e5 d1d4 c7c5 d4d1 c8c7 b3d2 e7g5');
INSERT INTO evaluations VALUES(31956,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(31956,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(31956,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(31956,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(31956,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(31956,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31956,36,'cp 19','e7e6 d2d4 c7c6 c4d3 f5d3 d1d3 g8f6 c1g5 b8d7 e1c1 f8b4 c1b1 b4c3 d3c3 f6e4 c3e3 e4g5 f3g5 d7f6 f2f3 f6d5 e3e5 e8g8 g5e4 d8e7 h2h4 a8d8 h4h5 h7h6 e4c5 b7b6 c5d3 e7c7 e5c7 d5c7 d3b4');
INSERT INTO evaluations VALUES(32362,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(32362,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(32362,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(32362,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(32362,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(32362,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(32362,36,'cp 33','d2d4 c7c6 c4d3 f5d3 d1d3 g8f6 c1g5 b8d7 e1c1 f8b4 c3e4 b4e7 e4f6 d7f6 c1b1 d8b6 h2h4 a8d8 c2c3 c6c5 g5e3 b6a5 d3c2 f6d5 c2e2 e8g8 g2g4 a5c7 h4h5 b7b5 h5h6 g7g6 d4c5 d5e3 e2e3 d8d1 h1d1 c7c5 d1d7 c5e3');
INSERT INTO evaluations VALUES(32364,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(32364,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(32364,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(32364,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(32364,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(32364,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(32364,36,'cp 26','c7c6 e1g1 f8e7 c4b3 b8d7 h2h3 h7h6 c3e2 g8f6 c1f4 f5e4 e2g3 e4h7 f1e1 e8g8 c2c4 f8e8 a2a3 d7f8 b3c2 f8g6 f4d2 d8c7 b2b4 a7a5 d1e2 a5b4 a3b4 b7b6 c2d3 a8a1 e1a1 e8d8 d2c3');
INSERT INTO evaluations VALUES(31563,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(31563,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(31563,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(31563,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(31563,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(31563,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(31563,36,'cp 43','c4d3 f5d3 d1d3 g8f6 c1g5 f8b4 g5f6 d8f6 e1c1 b4c3 d3c3 b8d7 c1b1 e8g8 g2g4 d7b6 h2h4 c6c5 f3g5 c5d4 c3d3 f6g6 g5e4 f7f5 g4f5 e6f5 e4g5 g6c6 h1e1 a8e8 e1e8 f8e8 d3f5 c6g6 f5b5 g6c6 b5d3 h7h6 g5f3');
INSERT INTO evaluations VALUES(36801,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(36801,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(36801,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(36801,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(36801,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(36801,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(36801,36,'cp 99','e1g1 a7a6 e3e4 b7b5 c4d3 c7c5 d4c5 c8b7 h2h3 f6d7 c1e3 g7c3 b2c3 d8c7 c3c4 b5b4 a2a3 b8c6 d1b1 a6a5 f1d1 b4a3 a1a3 c6b4 c5c6 b7c6 d3e2 f8d8 f3d4 d7f6 f2f3 f6h5 c4c5 h5f4 e2f1 e7e5 d4c6 d8d1 c6e7 c7e7 b1d1 e7c7 d1d6 f4e6 d6c7 e6c7');
INSERT INTO evaluations VALUES(36799,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(36799,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(36799,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(36799,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(36799,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(36799,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(36799,36,'cp 78','a7a6 e3e4 b7b5 c4d3 c7c5 d4c5 c8b7 h2h3 f6d7 c1e3 g7c3 b2c3 d8c7 c3c4 b5b4 a2a3 b8c6 a3b4 c6b4 c5c6 b4c6 c4c5 a6a5 d3b5 d7e5 f3e5 c6e5 d1c2 f8b8 f2f4 b7c6 b5e2 e5d7 e2c4 c6b5 f1c1 b5c4 c2c4 b8c8 c1c3 c7c6 c4d5 a5a4 f4f5 c6d5 e4d5');
INSERT INTO evaluations VALUES(37667,30,'cp 307',NULL);
INSERT INTO evaluations VALUES(37667,31,'cp 306',NULL);
INSERT INTO evaluations VALUES(37667,32,'cp 304',NULL);
INSERT INTO evaluations VALUES(37667,33,'cp 319',NULL);
INSERT INTO evaluations VALUES(37667,34,'cp 300',NULL);
INSERT INTO evaluations VALUES(37667,35,'cp 302',NULL);
INSERT INTO evaluations VALUES(37667,36,'cp 308','d2d4 g8e7 d4e5 d6c5 c3a4 d7d5 e4d5 c6d5 a4c5 d8a5 c2c3 a5c5 f1d3 c8f5 d3f5 e7f5 e1g1 b8c6 d1d3 f5e7 f1e1 c5c4 d3d2 h7h6 b2b3 c4g4 h2h3 g4g6 c1a3 e8g8 a1d1 f8d8 a3e7 c6e7 f3d4');
INSERT INTO evaluations VALUES(37666,30,'cp 315',NULL);
INSERT INTO evaluations VALUES(37666,31,'cp 300',NULL);
INSERT INTO evaluations VALUES(37666,32,'cp 297',NULL);
INSERT INTO evaluations VALUES(37666,33,'cp 311',NULL);
INSERT INTO evaluations VALUES(37666,34,'cp 309',NULL);
INSERT INTO evaluations VALUES(37666,35,'cp 306',NULL);
INSERT INTO evaluations VALUES(37666,36,'cp 296','d6b4 d4e5 d7d5 e5d6 d8d6 d1d6 b4d6 c1e3 g8e7 e1c1 d6c7 a2a4 b8d7 f3d2 e8g8 g2g3 f8e8 f2f4 d7f6 e3f2 h7h5 d1e1 c7a5 h2h3 c8d7 f1d3 a8d8 f2a7 a5c3 b2c3 d8a8 a7c5 a8a4 d2b3 b7b5 f4f5');
INSERT INTO evaluations VALUES(44284,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(44284,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(44284,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44284,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44284,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(44284,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44284,36,'cp 11','e3d4 f8b4 c1d2 b4d2 b1d2 e8g8 f1d3 b7b6 b2b4 d7d6 e1g1 c8b7 f1e1 a7a5 a2a3 b8d7 d3f1 d8c7 d1b3 f8c8 e1c1 c7d8 f3e1 f6e4 d2e4 b7e4 b3e3 d7f6 h2h3 c8c7 e1d3 h7h6 c4c5 d6c5 d4c5 e4d3 f1d3 b6c5 c1c5 c7c5');
INSERT INTO evaluations VALUES(45143,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(45143,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(45143,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(45143,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(45143,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(45143,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(45143,36,'cp -51','b8c6 g1e2 c6e5 c4b5 c8d7 b5d7 d8d7 f3e3 e7e6 d2d3 e5c6 e1g1 d6d5 a2a3 d7c7 e3f3 a8d8 b1c3 c6d4 e2d4 c5d4 c3e2 d5e4 d3e4 c7c2 e4e5 c2e4 e5f6 e4f3 g2f3 d4d3 e2g3 d3d2 c1d2 d8d2 g3e4 d2d4 f1c1 g7f6 c1c7 f8g7 c7b7 f6f5 e4c3 e8g8 b7a7');
INSERT INTO evaluations VALUES(13257,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(13257,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(13257,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(13257,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(13257,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(13257,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(13257,36,'cp -40','g1e2 c6e5 c4b5 c8d7 b5d7 d8d7 f3e3 e7e6 e1g1 e5c6 d2d3 d6d5 a2a3 f8e7 e4e5 d5d4 e3g3 f6d5 c2c4 d4c3 b1c3 e8c8 f1d1 d5c3 b2c3 d7d5 f2f4 f7f6 c3c4 d5d7 c1e3 f6e5 f4e5 d7c7 d3d4 c5d4 e2d4 c6d4 d1d4 d8d4 e3d4 h8d8 d4f2 e7c5 a1c1');
INSERT INTO evaluations VALUES(45680,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(45680,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(45680,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45680,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(45680,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(45680,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45680,36,'cp 23','c2c4 f8e7 b2b3 e8g8 d1c2 c6c5 d4c5 d8a5 c2d2 a5d2 b1d2 e7c5 e1g1 b8c6 c1b2 f8d8 f1d1 c5e7 a1c1 c8d7 a2a3 a7a5 f3e1 d7e8 c4d5 f6d5 d2e4 d8c8 e1d3 h7h6');
INSERT INTO evaluations VALUES(48191,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(48191,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(48191,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(48191,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(48191,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(48191,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(48191,36,'cp 0','e6d5 d2d4 c7c6 d1c2 g7g6 c1g5 f8e7 e2e3 c8f5 f1d3 f5d3 c2d3 e8g8 e1g1 b8d7 h2h3 a7a5 d3c2 f8e8 g5f4 e7b4 a2a3 b4f8 a3a4 f8b4 c3a2 b4f8');
INSERT INTO evaluations VALUES(50367,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(50367,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(50367,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(50367,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(50367,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(50367,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(50367,36,'cp -33','d7d5 d2d4 c7c5 d4c5 e7e6 f1d3 f8c5 g1f3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 d1e2 a8c8 f1d1 c6d4 f3d4 c5d4 c1e3 e6b3 d3c2 b3c2 e2c2 d8b6 c2e2 d4e3 e2e3 b6e3 f2e3');
INSERT INTO evaluations VALUES(51295,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(51295,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(51295,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(51295,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(51295,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51295,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(51295,36,'cp 8','f1b5 c7c6 b5d3 g8f6 g1f3 e7e6 e1g1 c8b7 c2c4 f8e7 b2b3 e8g8 c1b2 c6c5 d4c5 d5c4 d3c4 e7c5 d1e2 d8e7 a2a3 a7a5 f1d1 b8d7 b1d2 f8d8 c4b5 f6e4 d2e4 b7e4 a3a4 h7h6 f3d2 e4b7 d2c4 d7f6 c4e5 d8d1 a1d1 a8d8 h2h3 d8d1 e2d1 f6d5 e5d7');
INSERT INTO evaluations VALUES(52492,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52492,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52492,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52492,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(52492,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52492,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52492,36,'cp 0','b8c6 f1b5 g8f6 b1c3 c6d4 g1f3 a7a6 b5e2 d7d6 d2d3 e7e6 f3d4 c5d4 c3b1 d8c7 f2f4 c8d7 c2c4 d4c3 b2c3 c7c5 c1a3 c5e3 a3c1');
INSERT INTO evaluations VALUES(19509,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19509,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19509,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19509,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19509,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19509,36,'cp 0','f1b5 e7e6 g1f3 g8e7 d2d4 c5d4 f3d4 a7a6 b5e2 c6d4 d1d4 e7c6 d4d3 d7d5 c2c3 c6e5 d3d4 e5c6');
INSERT INTO evaluations VALUES(53276,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53276,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53276,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53276,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53276,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53276,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(53276,36,'cp 0','e2e3 e7e6 g1f3 g8f6 a2a3 b7b6 c2c4 c5d4 f3d4 c8b7 c4d5 f6d5 f1b5 b8d7 e1g1 a7a6 b5c6 b7c6 d4c6 d8c7 c6d4 f8e7 d4f5 e7f8');
INSERT INTO evaluations VALUES(5965,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(5965,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(5965,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(5965,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(5965,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(5965,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(5965,36,'cp 23','c5b6 e1g1 g8f6 a2a4 a7a6 d2d3 e8g8 b1d2 h7h6 f1e1 c8e6 h2h3 d8d7 c4b3 e6b3 d2b3 f8e8 a4a5 b6a7 c1e3 a7e3 e1e3 a8d8 d1c2 c6e7 a1e1 e7g6 d3d4 d7a4 e1a1 a4d7 d4e5 d6e5 b3c5 d7c8 a1e1 d8d6 g1h2');
INSERT INTO evaluations VALUES(9288,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9288,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(9288,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(9288,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(9288,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9288,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(9288,36,'cp 28','c3a4 c5b6 e1g1 e8g8 a4b6 a7b6 f1e1 h7h6 h2h3 c8d7 a2a4 f8e8 b2b3 c6e7 b5d7 f6d7 d3d4 b6b5 d4e5 d6e5 a4a5 c7c5 d1d3 e7c6 d3b5 a8a5 a1a5 d8a5 b5a5 c6a5 c1e3 a5c6 c2c3 e8d8 e1d1 d7f6 d1d8 c6d8 f3d2 f6d7 d2c4 d8e6 f2f3 b7b5');
INSERT INTO evaluations VALUES(10014,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(10014,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10014,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10014,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(10014,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(10014,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(10014,36,'cp 8','c6e5 d2d4 a7a6 b5e2 c5d6 d4e5 d6e5 c1g5 h7h6 g5h4 c7c6 e1g1 d8c7 g1h1 e8g8 e2d3 d7d5 e4d5 f6d5 c3d5 c6d5 h4g3 c8d7 d1d2 e5g3 h2g3 c7b6 b2b3 d7b5 d3b5 a6b5 d2d5 f8c8 d5d3 c8c5 h1g1 b6c6 f1c1 c5c3 d3d2 b5b4 c1f1 c3c2 d2b4 c2a2 a1a2 a8a2 b4d4 g8h7');
INSERT INTO evaluations VALUES(10753,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10753,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(10753,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10753,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10753,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(10753,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10753,36,'cp 7','d2d4 a7a6 b5e2 c5d6 d4e5 d6e5 c1g5 h7h6 g5h4 c7c6 e1g1 d8c7 g1h1 e8g8 e2d3 d7d5 e4d5 f6d5 c3d5 c6d5 h4g3 e5g3 h2g3 c7e5 b2b3 e5g5 d1c1 g5f6 h1g1 c8d7 c1d2 a8c8 d2b4 d7c6 a1e1 f8e8 b4d2 f6g5 e1e8 c8e8 d2c3 g5d8 f1d1 d8b6');
INSERT INTO evaluations VALUES(10751,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10751,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10751,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10751,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(10751,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(10751,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10751,36,'cp 7','a7a6 b5e2 c5d6 d4e5 d6e5 c1g5 h7h6 g5h4 c7c6 e1g1 d8c7 g1h1 e8g8 e2d3 d7d5 e4d5 f6d5 c3d5 c6d5 h4g3 e5g3 h2g3 c7e5 a1b1 c8d7 d1d2 e5f6 h1g1 f8e8 c2c3 d7c6 d2f4 f6f4 g3f4 g7g6 b1d1 d5d4 c3d4 a8d8 d3c4 c6d5 c4d5 d8d5');
INSERT INTO evaluations VALUES(10037,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(10037,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(10037,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(10037,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(10037,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(10037,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(10037,36,'cp -31','c6d4 e1g1 e8g8 f3d4 c5d4 d2d3 c7c6 b5a4 d7d5 e4d5 f6d5 c3d5 d8d5 a4b3 d5d6 a1b1 c8e6 b3e6 d6e6 c1d2 a8d8 f1e1 f7f6 h2h3 d8d6 c2c3 d4b6 d2e3 f8d8 e3b6 a7b6 e1e3 h7h6 d1e2 e6d7 b1d1 c6c5 e2f3 b6b5 f3g4 d7g4');
INSERT INTO evaluations VALUES(10199,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(10199,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(10199,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(10199,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(10199,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(10199,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(10199,36,'cp -38','f1e2 d7d5 b1d2 e8g8 e1g1 d8e7 c1b2 d5d4 a2a3 a7a5 f3h4 g7g6 g1h1 b7b5 g2g3 c8h3 h4g2 c5d6 d2f3 c6d8 f3g5 h3c8 f2f4 c7c5 a3a4 b5a4 a1a4 d8c6 a4a1 c6b4 b2c1 c8d7 c1d2 h7h6 g5f3 a5a4');
INSERT INTO evaluations VALUES(14252,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(14252,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(14252,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(14252,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(14252,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14252,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(14252,36,'cp 26','c8f5 a2a3 e7e6 c2c4 f6e4 b1d2 e4d2 d1d2 h7h5 e2e3 d8f6 b3b4 a7a6 f1d3 h5h4 h2h3 f6g6 d3f5 g6f5 e1g1 d5c4 d2e2 f5d3 e2d3 c4d3 f3e1 d3d2 e1d3 e8c8 a1d1 f8d6 d1d2 f7f6 d2c2 c6e7 a3a4 e7d5 b4b5 a6b5');
INSERT INTO evaluations VALUES(24845,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24845,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(24845,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24845,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(24845,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(24845,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24845,36,'cp 24','a2a3 e7e6 c2c4 c6b8 b1c3 b8d7 f3h4 f5e4 c4d5 e4d5 h4f3 d5f3 g2f3 g7g6 e2e3 f8g7 f1g2 f6d5 d1c2 d5c3 b2c3 e8g8 f3f4 c7c6 e1g1 d7f6 c3d2 a7a5 a1d1 d8c7 g2f3 f6d5 g1h1 f8d8 c2c5 b7b6 c5c2 c7e7');
INSERT INTO evaluations VALUES(19890,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(19890,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19890,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(19890,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(19890,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(19890,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(19890,36,'cp 13','c3d4 c8f5 f1d3 f5d3 d1d3 e7e6 g1f3 g8e7 b1c3 e7f5 a2a3 f8e7 e1g1 f7f6 e5f6 e7f6 c3e2 d8d7 c1g5 d7f7 g5f6 f7f6 a1d1 e8g8 d1d2 a8c8 h2h3 f5d6 e2g3 f6g6 f1e1 d6c4 d3g6 h7g6 d2e2 f8f3 g2f3 c6d4');
INSERT INTO evaluations VALUES(24243,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(24243,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(24243,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(24243,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(24243,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(24243,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(24243,36,'cp 63','d4c6 b7c6 b5c4 g7g6 d1e2 a6a5 e4e5 d6e5 e2e5 f8g7 f1e1 a5a4 h2h3 e8g8 e5e7 d7f5 g2g4 f5c2 e7e2 f8e8 c1e3 c2e4 c3e4 e8e4 e2c2 d8e8 e3d2 e8e5 c4f1 a8b8 d2c3 e5f4 f1g2 e4e1 a1e1 h7h5 g2c6 h5g4 c2a4 f4a4 c6a4 g4h3');
INSERT INTO evaluations VALUES(24241,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24241,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(24241,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(24241,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(24241,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(24241,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24241,36,'cp 57','b7c6 b5c4 g7g6 d1e2 f8g7 c4a6 d7g4 e2d3 f6d7 a6b7 d7e5 d3g3 a8a7 b7c6 e5c6 g3g4 g7c3 b2c3 d8d7 g4e2 e8g8 c1h6 f8a8 h2h3 d7e6 f1b1 a7a2 a1a2 e6a2 b1b3 a2a5 e2c4 a5h5 h6f4 c6e5 c4c7 a8a1 g1h2 e5f3 h2g3 g6g5 c7c8 g8g7 c8g4 h5g4 g3g4 g5f4 g4f3 a1a2 b3b7 g7f8 f3f4');
INSERT INTO evaluations VALUES(29509,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(29509,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(29509,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(29509,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(29509,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(29509,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(29509,36,'cp 122','f1b5 c7c6 d5c6 b8c6 b5c6 b7c6 g1f3 g4h5 h2h3 d8c7 b2b3 f6d5 c1b2 e7e6 c3d5 c6d5 d1e2 h5f3 g2f3 a8c8 e2b5 c7d7 b5d7 e8d7 h1g1 f7f6 a1c1 e6e5 e1e2 g7g5 g1g4 f8d6 g4a4 c8c7 d2d4 h8c8 c2c4 e5d4 b2d4 d5c4 a4c4 c7c4 c1c4');
INSERT INTO evaluations VALUES(30479,30,'cp 671',NULL);
INSERT INTO evaluations VALUES(30479,31,'cp 666',NULL);
INSERT INTO evaluations VALUES(30479,32,'cp 664',NULL);
INSERT INTO evaluations VALUES(30479,33,'cp 686',NULL);
INSERT INTO evaluations VALUES(30479,34,'cp 687',NULL);
INSERT INTO evaluations VALUES(30479,35,'cp 683',NULL);
INSERT INTO evaluations VALUES(30479,36,'cp 683','b8d7 b7b5 e7e6 f1c4 g8f6 g1f3 a8b8 b5a6 f8b4 c1d2 b4e7 c4b5 e8g8 b1c3 f6e8 a6a2 e8f6 b5e2 e7d6 e1g1 b8b7 e2a6 b7b6 a6d3 d8b8 d2c1 a7a5 a2a5 f6g4 h2h3 g4h2 f3h2 d6h2 g1h1 d7f6 e3e4 h2d6');
INSERT INTO evaluations VALUES(25849,30,'cp 717',NULL);
INSERT INTO evaluations VALUES(25849,31,'cp 698',NULL);
INSERT INTO evaluations VALUES(25849,32,'cp 701',NULL);
INSERT INTO evaluations VALUES(25849,33,'cp 699',NULL);
INSERT INTO evaluations VALUES(25849,34,'cp 701',NULL);
INSERT INTO evaluations VALUES(25849,35,'cp 695',NULL);
INSERT INTO evaluations VALUES(25849,36,'cp 717','b7b5 g8f6 f1c4 d8c7 g1f3 a8b8 b5a6 b8b6 a6a2 e7e6 e1g1 f8b4 c1d2 e8g8 b1c3 a7a5 e3e4 e6e5 c4f7 g8h8 d2e3 e5d4 e3d4 b4c5 d4c5 d7c5 f7c4 c5e4 c3e4 f6e4 a2a5 c7c5 b2b3 b6c6 a5c5 c6c5 a1a7 e4c3 f1a1');
INSERT INTO evaluations VALUES(25843,30,'cp 706',NULL);
INSERT INTO evaluations VALUES(25843,31,'cp 714',NULL);
INSERT INTO evaluations VALUES(25843,32,'cp 709',NULL);
INSERT INTO evaluations VALUES(25843,33,'cp 710',NULL);
INSERT INTO evaluations VALUES(25843,34,'cp 721',NULL);
INSERT INTO evaluations VALUES(25843,35,'cp 709',NULL);
INSERT INTO evaluations VALUES(25843,36,'cp 691','g8f6 g1f3 a8b8 b5a6 e7e6 f1c4 f8b4 c1d2 b4d2 b1d2 e8g8 e1g1 d8b6 a6b6 a7b6 c4b5 b8b7 f1c1 h7h6 g2g3 f8d8 g1g2 f6d5 e3e4 d5e7 b2b4 d7f6 c1c3 e7g6 h2h3 b7b8 a1a7 b8a8 a7b7 a8a1 b7b6 g6e7 c3c7');
INSERT INTO evaluations VALUES(39036,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(39036,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(39036,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(39036,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(39036,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(39036,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(39036,36,'cp 75','d7d5 e4e5 c7c5 c2c3 b8c6 f1b5 g8e7 b5c6 e7c6 e1g1 c8d7 f1e1 c5d4 c3d4 h7h6 b1c3 c6e7 c3e2 e7f5 e3d2 e8g8 a1c1 d7b5 e2g3 d8d7 g3f5 e6f5 h2h4 f8c8 d2f4 c8c1 d1c1 a8c8 c1d2 g8h7 b2b3 b5a6 g1h1 b7b6');
INSERT INTO evaluations VALUES(41322,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41322,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41322,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41322,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41322,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41322,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41322,36,'cp 6','g1f3 e8g8 c2c3 b8c6 f1e2 d8e8 h2h3 e7e5 d4e5 c6e5 e1g1 e5f3 e2f3 a7a5 a2a4 f6d7 e3e4 a8b8 d2b3 b7b6 b3d4 c8a6 f1e1 d7c5 e1e3 a6b7 d1c2 e8e7 b2b3 g7d4 c3d4 c5e6 f4h6 e6d4');
INSERT INTO evaluations VALUES(41754,30,'cp 310',NULL);
INSERT INTO evaluations VALUES(41754,31,'cp 316',NULL);
INSERT INTO evaluations VALUES(41754,32,'cp 324',NULL);
INSERT INTO evaluations VALUES(41754,33,'cp 350',NULL);
INSERT INTO evaluations VALUES(41754,34,'cp 344',NULL);
INSERT INTO evaluations VALUES(41754,35,'cp 341',NULL);
INSERT INTO evaluations VALUES(41754,36,'cp 355','e5f3 e8g8 d3e4 d5e4 f3d4 c7c5 d4b3 a7a5 c1e3 a5a4 b3d2 b8c6 b1a3 c6d4 e2h5 g7g6 h5d1 d6c7 c2c3 d4f5 e3c5 f8e8 d2c4 d8f6 c5d4 f6e7 a3b5 f5d4 b5d4 a4a3 b2a3 e4e3 c4e3 e7a3 d1d2 c7a5');
INSERT INTO evaluations VALUES(41750,30,'cp 356',NULL);
INSERT INTO evaluations VALUES(41750,31,'cp 322',NULL);
INSERT INTO evaluations VALUES(41750,32,'cp 329',NULL);
INSERT INTO evaluations VALUES(41750,33,'cp 329',NULL);
INSERT INTO evaluations VALUES(41750,34,'cp 338',NULL);
INSERT INTO evaluations VALUES(41750,35,'cp 340',NULL);
INSERT INTO evaluations VALUES(41750,36,'cp 343','e8g8 d3e4 d5e4 f3d4 d8f6 c1e3 c7c5 d4b3 f6b2 c2c3 b2e2 f1e2 b7b6 b1d2 c8b7 d2c4 d6e7 e1c1 b7a6 h1e1 f7f5 e3f4 b8c6 f2f3 e7h4 g2g3 h4f6 c1c2 g7g5 f4c1 f5f4 g3f4 g5f4 d1d6 a8c8 f3e4 f4f3 e2d3 f3f2 e1f1 f6g7 c1e3 c6d4 c3d4 c5d4 b3d4');
INSERT INTO evaluations VALUES(46674,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(46674,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(46674,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(46674,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(46674,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(46674,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(46674,36,'cp 27','f8g7 b1c3 e8g8 d3e2 b8a6 c1e3 c7c6 d1d2 e7e5 g1f3 f8e8 d4d5 f6g4 e3g5 d8b6 e1g1 h7h6 g5h4 a6c5 a1d1 a7a5 f1e1 g4f6 h4f6 g7f6 d2h6 b6b2 h6e3 b2b6 d5c6 b7c6 d1d6 f6g7 e3d2');
INSERT INTO evaluations VALUES(41295,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(41295,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(41295,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41295,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(41295,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(41295,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(41295,36,'cp 38','b1c3 e8g8 g1f3 c8g4 d3e2 g4f3 e2f3 f6d7 c1e3 b8c6 d4d5 c6a5 f3e2 d7e5 d1a4 c7c6 e1g1 c6d5 c3d5 e7e6 d5f4 b7b6 a1d1 a8c8 b2b3 a5b7 b3b4 e5c4 e2c4 c8c4 a4a6 c4c7 d1c1 c7d7 f1d1 d8e7 h2h3 e6e5 f4e2');
INSERT INTO evaluations VALUES(51767,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(51767,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(51767,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(51767,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(51767,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(51767,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(51767,36,'cp -6','b5e2 d7d5 d2d3 c6c5 g1f3 b8c6 e1g1 g8f6 f1e1 f8d6 e2f1 e6e5 e4d5 f6d5 d3d4 c5d4 f3d4 e8g8 d4c6 b7c6 b1d2 d6c7 c2c4 d5f4 d1a4 f4e6 a4c6 a8b8 c6a4 c8d7 a4d1 f7f5 d2b1 e6d4 b1c3 d7c6 c3b5 d4b5 c4b5 c6b5 f1b5 b8b5');
INSERT INTO evaluations VALUES(51816,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(51816,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51816,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(51816,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(51816,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51816,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(51816,36,'cp 25','b1c3 d7d5 g1f3 d5d4 c3b1 f8e7 d2d3 g8f6 a2a4 e8g8 g2g3 c6c5 b1a3 b7b6 f1g2 a7a6 c1d2 b8c6 e1g1 a8b8 a3c4 d8c7 d1e2 c8b7 f4f5 e6e5 g1h1 b6b5 c4a5 b7a8 a4b5 a6b5 a5c6 a8c6 a1a6 c7b7');
INSERT INTO evaluations VALUES(52736,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52736,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52736,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52736,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52736,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(52736,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(52736,36,'cp 10','e7e5 d2d4 e5d4 d1d4 g8f6 g1f3 f8e7 c4d5 c6d5 f1e2 b8c6 d4d3 c8e6 e1g1 e8g8 b2b3 f6e4 c3e4 e6f5 e4f6 e7f6 d3f5 f6a1 c1a3 d8f6 f5f6 a1f6 a3f8 g8f8 f1c1 d5d4 e3d4 a8d8 g1f1 f6d4 e2b5 d4b2 c1c2 c6d4 f3d4 b2d4');
INSERT INTO evaluations VALUES(53753,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(53753,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53753,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(53753,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(53753,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(53753,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53753,36,'cp 28','d2d4 e5d4 e3d4 g8f6 f1d3 f8d6 g1f3 e8g8 e1g1 d5c4 d3c4 c8f5 h2h3 h7h6 f3e5 b8d7 f1e1 d7b6 c4b3 b6d5 d1f3 f5e6 c1d2 d5c7 e5c4 f8e8 c4d6 d8d6 d2f4 d6d7 f4c7 d7c7 b3e6 e8e6 e1e6 f7e6 f3e3 c7f7 a1e1 a8e8 e3e5 a7a6 e1e2 f7d7 c3e4 f6e4 e5e4');
INSERT INTO evaluations VALUES(53756,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53756,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(53756,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(53756,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(53756,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(53756,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(53756,36,'cp 19','e5d4 e3d4 g8f6 f1d3 d5c4 d3c4 f8b4 a2a3 b4c3 b2c3 e8g8 g1e2 b8d7 c4d3 c6c5 e1g1 c5d4 c3d4 f8e8 a3a4 d7f8 d1b3 f8g6 a4a5 b7b6 c1d2 c8g4 f2f3 g4e6 b3b2 a8c8 a5b6 a7b6 a1c1 c8c1 f1c1 f6d5 c1b1');
INSERT INTO evaluations VALUES(52796,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(52796,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(52796,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(52796,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(52796,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(52796,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(52796,36,'cp -64','d5d4 c3e2 e7e5 d2d3 f8d6 f3f4 e5f4 e2f4 c6c5 g1f3 g8e7 g2g3 b8c6 f1g2 e8g8 e1g1 c8g4 d1e2 f7f6 g2h3 g4h3 f4h3 d8d7 c1f4 e7g6 e2g2 a8e8 b2b3 e8e7 f4d6 d7d6 f1f2 f8e8 a2a4 g6e5 f3e5 c6e5 h3f4 e5g4');
INSERT INTO evaluations VALUES(6476,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(6476,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(6476,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(6476,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(6476,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(6476,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(6476,36,'cp -51','g1f3 g8f6 b1c3 e8g8 f1d3 a7a6 c3e2 d7d5 e2g3 c5a7 e1g1 d5e4 g3e4 c8f5 e4f6 d8f6 d3f5 f6f5 d2d3 f7f6 f1e1 a8d8 h2h3 d8d5 e1e4 f5d7 d1e2 a7c5 g1f1 f8d8 a1e1 a6a5 a2a3 h7h6 e4c4 d7d6 a3a4 d6d7 e2e4 c5b4 c2c3 b4f8 d3d4 e5d4 c3d4');
INSERT INTO evaluations VALUES(8038,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(8038,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(8038,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(8038,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(8038,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(8038,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(8038,36,'cp 24','b1c3 e8g8 c4b3 d7d5 e4d5 c6a5 d1d2 a5b3 d4b3 b7b5 c3b5 d8d5 d2d5 f6d5 e3d4 g7h6 e1g1 c8f5 c2c4 f5d3 c4d5 d3b5 f1d1 b5c4 d4e3 h6e3 f2e3 c4b3 a2b3 f8b8 a1a3 b8b5 g1f2 a7a5 e3e4 f7f6 f2e3 g8f7 d1c1 a8b8 c1c3 h7h5 h2h3 h5h4');
INSERT INTO evaluations VALUES(15266,30,'cp 349',NULL);
INSERT INTO evaluations VALUES(15266,31,'cp 361',NULL);
INSERT INTO evaluations VALUES(15266,32,'cp 358',NULL);
INSERT INTO evaluations VALUES(15266,33,'cp 373',NULL);
INSERT INTO evaluations VALUES(15266,34,'cp 365',NULL);
INSERT INTO evaluations VALUES(15266,35,'cp 378',NULL);
INSERT INTO evaluations VALUES(15266,36,'cp 372','d1d5 a7a6 d5d8 e8d8 a1d1 c8d7 b5d4 c6d4 d1d4 e4c5 g2g3 a6a5 f1g2 f8e7 g1f3 c5e6 d4d3 c7c6 f4e3 d8c7 e3b6 c7b6 d3d7 h8e8 e2e3 e6c5 d7d4 b6c7 d4g4 g7g6 g4f4 e8f8 e1e2 c5e6 f4e4 a5a4 f3d2 f8d8 f2f4 b7b5 h1d1 e6c5');
INSERT INTO evaluations VALUES(18589,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(18589,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(18589,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(18589,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(18589,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(18589,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(18589,36,'cp 96','f1e2 g8f6 b1c3 g7g6 g1f3 f8g7 c1e3 e8g8 e1g1 a7a6 d1c2 b7b5 f1d1 c8b7 h2h3 b5c4 e2c4 d6d5 c4b3 d5e4 c3e4 f6e4 c2e4 d7f6 e4c2 e7e6 e3g5 d8d6 d1e1 h7h6 g5d2 f6d7 f3e5 d7e5 d4e5 d6e7 a1c1 f8d8');
INSERT INTO evaluations VALUES(19287,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(19287,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(19287,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19287,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(19287,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(19287,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(19287,36,'cp -11','f8g7 e1g1 d7d6 c2c3 e7e5 d2d3 g8e7 b1a3 e8g8 a3c2 d6d5 a2a4 h7h6 b2b4 c8e6 c1a3 d5e4 d3e4 d8d1 f1d1 c5b4 c3b4 e6b3 b4b5 b3c2 b5c6 c2d1 c6b7 a8b8 a3e7 d1f3 e7f8 g7f8 g2f3 b8b7 a4a5 b7b5 a5a6 f8c5 a1c1 c5b6 f3e2 b5b2 c1c8 g8g7');
INSERT INTO evaluations VALUES(21409,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(21409,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(21409,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(21409,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(21409,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(21409,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(21409,36,'cp 15','g8f6 c2c4 f8e7 a2a3 e8g8 b1c3 b7b6 b2b4 d5c4 f1c4 c6b8 e1g1 c7c5 f3e5 c5d4 c3b5 d4e3 c1e3 c8b7 d1d8 e7d8 e5f7 b7d5 f7e5 d5c4 e5c4 f6d5 a1d1 b8c6 g2g3 a7a6 b5d4 d5e3 c4e3 c6d4 d1d4 a6a5 d4e4 a5b4 a3b4 d8e7 e3c2 a8a4 f1e1 b6b5 e4e6 e7b4 c2b4 a4b4');
INSERT INTO evaluations VALUES(23124,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(23124,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(23124,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(23124,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(23124,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(23124,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23124,36,'cp 17','d4f3 d1f3 c7c6 e1g1 f8c5 f3g3 d7d6 g3g7 d8f6 g7f6 g8f6 b1c3 h8g8 g1h1 a7a5 a2a3 c8e6 d2d3 e8e7 c1d2 b7b5 a4b3 e6b3 c2b3 b5b4 a3b4 a5b4 a1a8 g8a8 c3a4 c5d4 g2g3 c6c5 h1g2 h7h5 h2h3 e7e6 f1a1 a8g8 a4b6 d4b2 a1a7 b2d4 g2f3 g8b8 b6c4 f6d7 d2e3');
INSERT INTO evaluations VALUES(23126,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(23126,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(23126,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(23126,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(23126,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(23126,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(23126,36,'cp 21','d1f3 c7c6 e1g1 f8c5 f3g3 d7d6 g3g7 d8f6 g7f6 g8f6 b1c3 a7a5 a2a3 c8e6 d2d3 h8g8 g1h1 e8e7 c1d2 b7b5 a4b3 b5b4 a3b4 a5b4 a1a8 g8a8 c3a4 e6b3 c2b3 c5d4 g2g3 c6c5 h1g2 e7e6 g2f3 h7h5 h2h3 a8g8 f1g1 f6d7 g3g4');
INSERT INTO evaluations VALUES(23125,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(23125,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(23125,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(23125,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(23125,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(23125,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(23125,36,'cp 13','c7c6 e1g1 f8c5 f3g3 d8f6 b1c3 g8e7 d2d3 d7d6 c1e3 c5e3 f2e3 f6h6 d3d4 e8g8 d4e5 d6e5 a1e1 c8e6 g3e5 e7g6 e5g3 h6h5 a4b3 h5c5 b3e6 f7e6 f1f8 a8f8 h2h4 c5h5 g3h3 f8f6 e1d1 e6e5 h3g3 h5h4 g3h4 g6h4');
INSERT INTO evaluations VALUES(27099,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(27099,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(27099,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(27099,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(27099,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(27099,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(27099,36,'cp 46','f7e6 b1c3 d8d7 e1g1 g8e7 c1e3 e7c6 d4c4 g7g6 c3e2 f8g7 e2f4 c6d8 c4b3 e8g8 a2a4 b7b6 c2c3 d7f7 a4a5 g7h6 f4d3 h6e3 f2e3 f7d7 f1f8 g8f8 a5b6 c7b6 e4e5 d8f7 a1d1 a8e8 b3a3 d7e7 e5d6 f7d6 d3e5 d6f5 a3e7 e8e7 g1f2 f5h6 d1d8 f8g7 d8c8 g7f6');
INSERT INTO evaluations VALUES(27136,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(27136,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(27136,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(27136,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(27136,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(27136,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(27136,36,'cp 62','e1g1 d8d7 b1c3 g8f6 c1g5 f8e7 e4e5 d6e5 d4e5 h7h6 g5h4 e7d6 e5e2 e8g8 h4f6 f8f6 a1d1 f6f5 g2g3 a8e8 c3e4 d7c6 d1d4 a7a5 d4c4 c6d7 f1d1 f5d5 e2g4 d7e7 e4c3 d5f5 g1g2 e8d8 f2f3 d6c5 d1d8 e7d8');
INSERT INTO evaluations VALUES(27137,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(27137,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(27137,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(27137,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(27137,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(27137,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(27137,36,'cp 52','d8d7 b1c3 g8e7 d4c4 e8c8 f1d1 e7c6 b2b4 f8e7 a2a4 e7f6 a1b1 h8f8 c1e3 g7g5 a4a5 c8b8 a5a6 b7b6 b4b5 c6e5 c4b3 e5g4 c3e2 g4e3 b3e3 d7f7 c2c4 f7g7 b1c1 h7h5 e2d4 f6d4 d1d4 g7e5 e3d3');
INSERT INTO evaluations VALUES(28355,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(28355,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(28355,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(28355,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(28355,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(28355,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(28355,36,'cp 80','d1b3 d8a5 a1c1 e8g8 f3e5 b8d7 e5d7 f6d7 a2a3 b4c3 b3c3 a5c3 c1c3 c6c5 c4d5 c5d4 e3d4 e6d5 f1e2 f8c8 g2g4 c8c3 b2c3 f5e6 e1d2 d7b6 h1b1 e6d7 h2h4 d7c6 b1b4 f7f6 f4c7 g8f7 c7g3 b6c4 e2c4 d5c4 b4c4');
INSERT INTO evaluations VALUES(29736,30,'cp 265',NULL);
INSERT INTO evaluations VALUES(29736,31,'cp 283',NULL);
INSERT INTO evaluations VALUES(29736,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(29736,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(29736,34,'cp 277',NULL);
INSERT INTO evaluations VALUES(29736,35,'cp 272',NULL);
INSERT INTO evaluations VALUES(29736,36,'cp 294','h2h4 h7h6 f1c4 c7c5 d4d5 e7e5 d1b3 d8b6 g2g4 h5g6 b3b6 a7b6 g1e2 e8d8 h4h5 g6h7 c1e3 f6e8 e2g3 f8e7 c3b5 d8d7 e1c1 a8a4 c4b3 a4a8 g3e2 e8d6 e2c3 b8a6 b3a4 a6c7 b5d6 d7d6 b2b3 e7f6 h1g1 f6g5 e3g5 h6g5');
INSERT INTO evaluations VALUES(33290,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(33290,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(33290,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(33290,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(33290,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(33290,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(33290,36,'cp 51','c5b6 g1f3 h4h5 b1c3 g8e7 c3e2 e7g6 c1f4 g6f4 e2f4 h5h6 f4d5 c7c6 d5b6 a7b6 a2a4 d7d6 f1f2 c8e6 c4d3 e8g8 d1c1 h6c1 h1c1 b8a6 c2c3 a6c7 a1a3 g7g6 c1e1 c7e8 h2h3 e8f6 g2g4 f6d7 f2g3 f7f6 b2b4 f8e8 c3c4 e6f7 g4g5');
INSERT INTO evaluations VALUES(33154,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(33154,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(33154,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(33154,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(33154,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(33154,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(33154,36,'cp 54','g1f3 h4h5 b1c3 g8e7 c3e2 e7g6 c1f4 g6f4 e2f4 h5h6 d1d2 d7d6 c2c3 e8g8 f4d5 h6d2 f3d2 c8e6 a2a4 c7c6 d5b6 a7b6 f1e2 b8d7 e2d3 f8e8 c4e6 e8e6 h1e1 a8e8 e1e2 h7h5 a1e1 e8a8 b2b3 f7f6 h2h3 g8f7 d2c4 h5h4 e1f1');
INSERT INTO evaluations VALUES(33153,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(33153,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(33153,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(33153,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(33153,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(33153,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(33153,36,'cp 45','h4h5 b1c3 g8e7 c3e2 e7g6 c1f4 g6f4 e2f4 h5h6 d1d2 d7d6 f1e2 c7c5 f4d5 h6d2 e2d2 b6d8 e4e5 c5d4 e5d6 b8c6 d5c7 d8c7 d6c7 c8e6 h1e1 e8d7 c4e6 f7e6 e1e4 a8f8 a1d1 d7c7 e4e6 f8f4 d1e1 f4g4 e1e2 h8f8 f3e5 c6e5 e6e5 c7d7 d2d3 f8f4 e5e7 d7c6');
INSERT INTO evaluations VALUES(39809,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(39809,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39809,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(39809,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(39809,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(39809,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(39809,36,'cp 21','e8g8 c4d5 f6d5 g5e7 d8e7 c3d5 e6d5 a1c1 c8b7 f1e2 c7c5 d4c5 b6c5 e1g1 f8c8 d1a4 a7a5 f1d1 b8a6 e2b5 e7f6 b5a6 f6a6 a4d7 a8b8 d7f5 a6d6 b2b3 h7h6 h2h3 d6e6 f5e6 f7e6 f3e5 c8c7 h3h4 b8c8 h4h5 g8f8 e5g6 f8f7');
INSERT INTO evaluations VALUES(41928,30,'cp 374',NULL);
INSERT INTO evaluations VALUES(41928,31,'cp 402',NULL);
INSERT INTO evaluations VALUES(41928,32,'cp 388',NULL);
INSERT INTO evaluations VALUES(41928,33,'cp 384',NULL);
INSERT INTO evaluations VALUES(41928,34,'cp 386',NULL);
INSERT INTO evaluations VALUES(41928,35,'cp 412',NULL);
INSERT INTO evaluations VALUES(41928,36,'cp 403','d2d4 c5b4 c1g5 g4f6 f1d3 h7h6 g5h4 c8g4 e1g1 g7g5 h4g3 b8d7 c3a4 d6d5 e4d5 f6d5 a2a3 b4e7 f1e1 e8g8 c2c4 d5f4 a4c3 c7c6 g3f4 g5f4 d3c2 e7g5 d1d3 f7f5 h2h3 g4f3 d3f3 d8f6 a1d1 a8e8 e1e8 f8e8');
INSERT INTO evaluations VALUES(41927,30,'cp 384',NULL);
INSERT INTO evaluations VALUES(41927,31,'cp 410',NULL);
INSERT INTO evaluations VALUES(41927,32,'cp 407',NULL);
INSERT INTO evaluations VALUES(41927,33,'cp 393',NULL);
INSERT INTO evaluations VALUES(41927,34,'cp 398',NULL);
INSERT INTO evaluations VALUES(41927,35,'cp 416',NULL);
INSERT INTO evaluations VALUES(41927,36,'cp 400','c5b4 c1g5 g4f6 f1d3 h7h6 g5h4 g7g5 h4g3 c8g4 e1g1 b8d7 c3a4 d6d5 e4d5 f6d5 c2c4 d5f4 a2a3 b4e7 f1e1 e8g8 a4c3 c7c6 d3c2 e7d6 c4c5 d6c7 g3f4 c7f4 h2h3 g4h5 g2g4 h5g6 c2g6 f7g6 d1b3 f8f7');
INSERT INTO evaluations VALUES(41931,30,'cp 421',NULL);
INSERT INTO evaluations VALUES(41931,31,'cp 414',NULL);
INSERT INTO evaluations VALUES(41931,32,'cp 408',NULL);
INSERT INTO evaluations VALUES(41931,33,'cp 408',NULL);
INSERT INTO evaluations VALUES(41931,34,'cp 429',NULL);
INSERT INTO evaluations VALUES(41931,35,'cp 419',NULL);
INSERT INTO evaluations VALUES(41931,36,'cp 405','c1g5 f7f6 g5f4 d8e7 f1d3 b8c6 f4d2 g4h6 e1g1 b4c3 d2c3 c8g4 h2h3 g4h5 g2g4 h5f7 f1e1 e8g8 d3f1 e7d7 d4d5 c6e5 f1g2 g8h8 f3d4 e5g6 d1d2 c7c5 d4f5 h6f5 e4f5 g6h4 g2h1 b7b6 e1e6 d7c7 a1e1 a8e8 e1e3 f7e6 d5e6');
INSERT INTO evaluations VALUES(46897,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(46897,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46897,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(46897,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46897,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46897,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(46897,36,'cp 4','e2e3 c7c5 g1f3 c5d4 e3d4 b8c6 c2c3 d8b6 d1b3 f6h5 b1a3 c8f5 b3b6 a7b6 a3b5 e8d7 h4g3 h5g3 h2g3 e7e6 e1d2 f8e7 f1d3 f5d3 d2d3 h6h5 b2b4 g7g5 d3c2 h5h4 g3g4 f7f6 c2b3 e7d6 a2a4 c6e7 h1e1 e7g6 f3g1 g6f4');
INSERT INTO evaluations VALUES(46895,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46895,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46895,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46895,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46895,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46895,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46895,36,'cp 0','c7c5 g1f3 c5d4 e3d4 b8c6 c2c3 d8b6 d1b3 f6h5 b1a3 c8f5 b3b6 a7b6 a3b5 e8d7 h4g3 h5g3 h2g3 e7e6 f3e5 c6e5 d4e5 f8c5 b2b4 c5e7 b5d4 a8a3 e1c1 a3a2 f1b5 d7c7 d4f5 e6f5 h1f1 h8d8 c1b1 d8a8 d1d5 a2a1 b1b2 a8a2 b2b3 a2a3 b3b2');
INSERT INTO evaluations VALUES(49652,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49652,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(49652,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(49652,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(49652,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(49652,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(49652,36,'cp -29','f6e4 d2d3 e4f6 f3e5 f8d6 e5c4 d6e7 f1g2 d7d5 c4e5 e8g8 e1g1 f8e8 f1e1 h7h6 b1d2 e7d6 d2f3 a7a5 a2a3 a5a4 b2b4 a4b3 c2b3 c7c6 c1b2 f6h7 d1c2 b8d7 e5d7 e8e1 a1e1 c8d7 b3b4 d6e7 h2h4 e7f6 c2d2 h7f8 b2f6 d8f6');
INSERT INTO evaluations VALUES(49907,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49907,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(49907,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(49907,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(49907,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(49907,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(49907,36,'cp -19','d2d3 e4f6 f3e5 f8d6 e5c4 d6e7 f1g2 e8g8 e1g1 d7d5 c4e5 f8e8 b1c3 c7c6 f1e1 a7a5 e5f3 b8d7 c3e2 d7f8 b2b3 c8g4 h2h3 g4f3 g2f3 d8d7 f3g2 f8e6 a2a3 e7c5 c1d2 h7h6 a3a4 c5b4 d2b4 a5b4 d1d2 c6c5');
INSERT INTO evaluations VALUES(56454,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(56454,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(56454,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(56454,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(56454,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(56454,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(56454,36,'cp -90','g8f6 g1f3 f8d6 e3e4 e8g8 e4d5 f6d5 g2g3 f8e8 b1d2 d6f8 f1g2 c8f5 d2e4 h7h6 e1g1 a7a5 f1e1 b8c6 c1e3 b7b6 a2a3 f5e6 e4d2 d8d7 d2c4 f7f6 d1a4 e6f5 a1d1 f5g6 f3h4 g6f7 d1d2 a8d8 h4f5');
INSERT INTO evaluations VALUES(48870,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(48870,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(48870,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(48870,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(48870,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(48870,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(48870,36,'cp -85','g1f3 f8d6 e3e4 e8g8 e4d5 f6d5 g2g3 f8e8 b1d2 d6f8 f1g2 c8f5 d2e4 h7h6 e1g1 a7a5 f1e1 b8d7 c3c4 d5f6 d1c2 f5g6 c1e3 f6g4 a1d1 c7c6 g2h3 g6h5 c4c5 d8c7 d3d4 g4e3 e1e3 a8d8 d4e5 d7e5 d1d8 e8d8 f3e5');
INSERT INTO evaluations VALUES(57359,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(57359,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(57359,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(57359,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(57359,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(57359,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(57359,36,'cp 42','g8f6 c2c4 f8e7 f1d3 e8g8 b1c3 b7b6 e1g1 c8b7 a1b1 a7a5 a2a3 d8e8 b2b4 a5b4 a3b4 b8c6 c4c5 c6d8 c3b5 a8c8 d1e2 d7d6 c1a3 f6d5 f1c1 d8f7 b5c3 c8a8 c3d5 b7d5 a3b2 g7g5 c5b6 g5g4 f3d2 c7b6');
INSERT INTO evaluations VALUES(5827,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(5827,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(5827,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(5827,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(5827,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(5827,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(5827,36,'cp 96','h2g3 d8e7 c2c4 g8f6 b1d2 e8g8 a1c1 f8d8 f1d3 e6e5 c4d5 d8d5 c1c6 b7c6 f3e5 h7h6 e1g1 c6c5 d3c4 c8e6 d1c1 c5d4 e5c6 e7d6 c4d5 e6d5 c6d4 c7c5 d4b3 a8b8 f1d1 d6b6 c1c3 d5e6 d1a1 f6d5 c3c2 b8c8 a2a3 c8b8 b3c5 b6b2 a1a2 b2c2');
INSERT INTO evaluations VALUES(10930,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(10930,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(10930,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(10930,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(10930,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(10930,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(10930,36,'cp -37','d4f5 h7h6 f5e7 h6g5 e7c6 d7c6 d1d8 e8d8 f1d3 g5g4 b1d2 d8e7 e1g1 c8e6 f2f4 g4f3 d2f3 a8d8 h2h3 h8h5 g1f2 c6c5 b2b3 f6d7 d3b5 d7e5 a1d1 d8d1 f1d1 e5f3 f2f3 g7g5 b5e2 h5h4 f3e3 b7b6 a2a4 f7f6');
INSERT INTO evaluations VALUES(12277,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(12277,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(12277,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(12277,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(12277,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(12277,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(12277,36,'cp 54','f3g1 d7d6 g1e2 g7g6 e1g1 f8g7 f2f4 e5f4 e2f4 c6e5 c4b3 e8g8 a2a4 g8h7 c1d2 e5c6 h2h3 c6d4 b3c4 c8e6 c3d5 f6d5 f4d5 f7f5 a4a5 a8b8 c2c3 d4c6 e4f5 f8f5 f1f5 e6f5 d5e3 f5d7 c4d5 c6e5 d5e4 d8g5');
INSERT INTO evaluations VALUES(13418,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(13418,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(13418,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(13418,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(13418,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(13418,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(13418,36,'cp 54','c2c3 c8g4 h2h3 g4f3 d1f3 d8d5 f1e2 e7e6 a2a4 a7a5 f3g3 e8c8 e2f3 d5f5 g3g5 d8e8 g5f5 e6f5 e1d1 f6e4 d1c2 f8d6 h1f1 h7h5 h3h4 g7g6 c1e3 f5f4 e3d2 e4d2 c2d2 c8c7 d2d3 d6e7 f1h1 e7f6 a1d1 e8d8 d1e1 h8e8');
INSERT INTO evaluations VALUES(17242,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(17242,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(17242,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(17242,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(17242,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(17242,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(17242,36,'cp 19','f1g2 e7e5 b1c3 f8c5 g1f3 d7d6 e1g1 e8g8 e2e3 c8f5 d2d3 d8d7 d3d4 c5b6 c3d5 f6d5 c4d5 c6e7 a2a4 a7a5 d4e5 f5e4 b2b3 e7d5 e5d6 d7d6 c1a3 b6c5 a3c5 d6c5 a1c1 c5e7 f3d2 e4g2 g1g2 c7c6 d1f3 d5f6 f1d1');
INSERT INTO evaluations VALUES(17243,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(17243,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(17243,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(17243,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(17243,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(17243,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(17243,36,'cp 23','e7e5 b1c3 f8c5 g1f3 d7d6 e1g1 e8g8 d2d3 c6d4 c3a4 d4f3 g2f3 c5d4 c1g5 c8h3 f3g2 h3g2 g1g2 h7h6 g5f6 d8f6 e2e3 d4b6 d1f3 f6f3 g2f3 f7f5 a4b6 a7b6 a2a4 g8f7 b2b4 g7g5 a4a5 b6a5 b4a5 f7e6 f1b1 e5e4 d3e4 f5e4 f3e4 f8f2 b1b7');
INSERT INTO evaluations VALUES(20756,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(20756,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(20756,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(20756,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(20756,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(20756,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(20756,36,'cp 141','c1f4 f8e7 f1b5 c8d7 b5d7 d8d7 d1d3 g8f6 b1c3 e8g8 e1c1 d7e6 h1g1 a8d8 f2f3 b7b5 d3b5 d6d5 c3d5 f6d5 e4d5 d8d5 a2a3 a7a6 b5a4 d5d4 d1d4 c5d4 g1d1 h7h5 c1b1 e6e2 d1d4 e2g2 d4d7 e7f6 a4b3 g2h3 d7a7 h3g2 b1a2 h5h4 a7c7 g2e2 c7d7 e2f1');
INSERT INTO evaluations VALUES(22091,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(22091,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(22091,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(22091,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(22091,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(22091,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(22091,36,'cp 73','a7a5 d2e3 g8f6 a2a3 b4a6 f3d2 c8d7 h2h3 f8e7 g2g4 e8g8 g4g5 f6e8 h3h4 f7f6 h1g1 f6g5 h4g5 g7g6 b1c3 e8g7 f1b5 g7h5 b5d7 d8d7 d1g4 d7g4 g1g4 f8f7 d2c4 b7b6 c3b5 f7f3 b2b3 f3h3 e1c1 a8f8 c4d2 h5f4 f2f3 f4h5');
INSERT INTO evaluations VALUES(17743,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(17743,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(17743,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(17743,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(17743,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(17743,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17743,36,'cp 82','a2a3 b4a6 d2e3 g8f6 f3d2 c7c6 b1c3 f6g4 d2c4 g4e3 c4e3 f8e7 g2g3 d8b6 e3c4 b6c7 h2h4 a6c5 c4e3 a8b8 f1h3 c8h3 h1h3 h7h5 d1e2 g7g6 h3h1 e7f6 b2b4 c5d7 e1f1 f6g7 f1g2 e8g8 b4b5');
INSERT INTO evaluations VALUES(17744,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(17744,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(17744,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(17744,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(17744,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(17744,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(17744,36,'cp 89','b4a6 d2e3 g8f6 f3d2 c7c6 b1c3 f6g4 d2c4 g4e3 c4e3 f8e7 g2g3 e7g5 e3f5 e8g8 h2h4 c8f5 f1a6 a8a6 e4f5 g5f6 e1g1 d8d7 d1f3 f6d8 a1d1 g7g6 f5g6 h7g6 d5c6 a6c6 d1d5 f7f5 f1d1 f5f4 d5e5 d7h3 e5e2 f4g3 f3g3 h3h4 g3g6 g8h8');
INSERT INTO evaluations VALUES(23081,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23081,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(23081,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23081,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23081,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23081,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23081,36,'cp 0','d7d5 e4d5 e5f3 d1f3 f8c5 f1b5 c8d7 b5d7 d8d7 d2d3 g8f6 c1e3 d7e7 e1d2 e7e3 f3e3 c5e3 d2e3 f6d5 e3d2 f7f6 a2a4 e8d7 h1e1 h7h5 b1c3 d5c3 d2c3 a8e8 b2b4 h5h4 c3d4 h8h5 e1e8 d7e8 a1e1 e8f7 c2c4 h5g5 e1e2 g5g4 d4d5 g4g5');
INSERT INTO evaluations VALUES(23867,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(23867,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(23867,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(23867,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(23867,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(23867,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(23867,36,'cp 55','g4h5 b2b4 c5b6 a2a4 a7a6 d2d3 g8e7 g2g4 h5g6 f3h4 f7f6 d1f3 g6f7 c4f7 e8f7 g4g5 e7g6 f3h5 f7g8 h4f5 g6e7 g5f6 e7f5 h5f5 d8f6 f5f6 g7f6 g1h2 b6a7 b1d2 g8f7 d2c4 c6e7 f2f4 h8g8 c1e3 a7e3 c4e3 a8f8 f4f5');
INSERT INTO evaluations VALUES(23859,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(23859,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(23859,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(23859,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(23859,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(23859,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(23859,36,'cp 47','b2b4 c5b6 a2a4 a7a6 d2d3 g8e7 b1d2 e8g8 g2g4 h5g6 c4b3 b6a7 f3h4 g8h8 d2f3 d8d7 c1e3 f7f6 d1d2 a7e3 f2e3 g6f7 b3f7 f8f7 f1f2 d6d5 e4d5 e7d5 a1e1 a8f8 d2c2 c6d8 d3d4 e5d4 e3d4 d8e6');
INSERT INTO evaluations VALUES(26994,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(26994,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(26994,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(26994,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(26994,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(26994,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(26994,36,'cp -81','g4f3 e2f3 c6d4 c1e3 d4e6 d1e2 c7c6 f3g4 e6d4 e2d3 g8f6 g4h3 d4e6 e1c1 d8c7 h3e6 f7e6 f2f4 e5f4 e3f4 e6e5 f4g5 f8e7 g5f6 e7f6 d3d6 c7d6 d1d6 h8f8 c3a4 f6e7 d6e6 f8f2 a4c5 e8f7 e6e5 e7c5 e5c5 f2g2 c5f5 f7e6 f5f3 a8d8 f3d3 d8d3 c2d3 b7b6 d3d4');
INSERT INTO evaluations VALUES(27003,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(27003,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(27003,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(27003,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(27003,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(27003,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(27003,36,'cp -75','e2f3 c6d4 c1e3 d4e6 d1d2 g8f6 e1c1 h7h5 h2h3 c7c6 c1b1 f8e7 a2a3 g7g6 g2g3 d8c7 f3e2 e6c5 e3c5 d6c5 d2e3 b7b5 f2f4 b5b4 a3b4 c5b4 c3a4 h5h4 e2c4 h4g3 f4f5 g6f5 e4f5 a8d8 e3g3 d8d1 h1d1 e8f8 d1g1');
INSERT INTO evaluations VALUES(27001,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(27001,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(27001,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(27001,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(27001,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(27001,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(27001,36,'cp -68','c6d4 c1e3 d4e6 d1d2 h7h5 e1c1 c7c6 g2g4 h5g4 f3g4 d8e7 h2h4 g8f6 g4e6 e7e6 c1b1 f6h5 f2f3 f8e7 a2a4 b7b6 d2d3 a8c8 d3a6 c8c7 a6d3 g7g6 e3g5 e8f8 c3e2 c7d7 f3f4 h5f4 e2f4 e5f4 g5f4');
INSERT INTO evaluations VALUES(47430,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(47430,31,'cp -104',NULL);
INSERT INTO evaluations VALUES(47430,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(47430,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(47430,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(47430,35,'cp -98',NULL);
INSERT INTO evaluations VALUES(47430,36,'cp -92','d8d5 d1f3 d5d7 d2d3 b8c6 c1e3 b7b6 b1c3 c8b7 h2h4 e7e6 e1c1 e8c8 h4h5 c6d4 f3g4 d4f5 e3g5 f7f6 g5d2 h8g8 g4h3 g7g5 h5g6 h7g6 g1e2 f8g7 c1b1 g6g5 h3g4 c8b8 f2f3 g8h8 h1h8 d8h8 e2g3 f5d4');
INSERT INTO evaluations VALUES(33762,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(33762,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(33762,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(33762,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(33762,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(33762,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(33762,36,'cp -87','d1f3 d5d6 d2d3 b8c6 g1e2 e7e5 c1d2 c8e6 b1c3 e8c8 e1c1 c6d4 f3g3 d4e2 c3e2 e6a2 d2c3 f7f6 b2b3 d6a6 g3g4 c8b8 d3d4 e5e4 h1e1 f8a3 c3b2 a3b4 e2c3 a6a5 e1e3 f6f5 g4f4 h8e8 f2f3 e4f3 e3f3 b4e7 h2h3 g7g6 d1e1 b8a8');
INSERT INTO evaluations VALUES(53176,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53176,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(53176,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53176,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53176,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53176,36,'cp 0','e7e6 d2d3 g8e7 c1f4 e7d5 f4g3 c8d7 e1g1 f8e7 a2a4 d7c6 b1d2 e8g8 c4b3 a7a6 a4a5 b7b5 a5b6 d5b6 d2c4 b6c4 b3c4 b8d7 f3d2 d7b6 b2b3 h7h6 d1g4 b6d5 c4d5 d8d5 d2e4 g8h7 f2f4 a6a5 h2h3');
INSERT INTO evaluations VALUES(54080,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(54080,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(54080,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(54080,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54080,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54080,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54080,36,'cp 0','d2d3 g8e7 c1f4 e7d5 f4g3 c8d7 e1g1 f8e7 a2a4 e8g8 b1d2 a7a6 d2e4 d7c6 f3d2 b8d7 d1g4 h7h6 h2h4 b7b5 c4d5 c6d5 b2b3 g8h7 g4h3 h7h8 h3g4');
INSERT INTO evaluations VALUES(54079,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54079,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54079,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54079,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54079,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54079,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54079,36,'cp 0','g8e7 c1f4 e7d5 f4g3 c8d7 e1g1 f8e7 a2a4 d7c6 b1d2 e8g8 d2e4 a7a6 f3d2 b8d7 d1g4 h7h6 h2h4 b7b5 c4d5 c6d5 b2b3 g8h7 g4h3 h7h8 h3g4');
INSERT INTO evaluations VALUES(54782,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(54782,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(54782,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(54782,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(54782,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(54782,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(54782,36,'cp 92','g1f3 e7e6 f1c4 g8f6 d1e2 f8b4 c1d2 b8c6 e1c1 e8g8 c1b1 g8h8 a2a3 b4c3 d2c3 f6e4 c3d2 e6e5 d4e5 d8e8 c4d5 c8e6 d5e4 f5e4 e2e4 e6f5 e4e3 a8d8 h1e1 f5e6 b2b3 e8g6 e3g5 g6g5 f3g5 e6g4 g5f3 d8e8 d2c3 h8g8');
INSERT INTO evaluations VALUES(54780,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(54780,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(54780,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(54780,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(54780,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(54780,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(54780,36,'cp 74','e7e6 f1c4 g8f6 d1e2 f8b4 c1d2 e8g8 e1c1 b8c6 a2a3 b4c3 d2c3 d8d6 h1g1 g8h8 g2g4 f6e4 g4f5 e4c3 e2g2 d6f4 d1d2 f4h6 b2c3 e6e5 c4e6 e5d4 c3d4 h6f6 f3g5 c8e6 g5e6 f8f7 g2h3 f6f5 h3f5 f7f5 g1g7 a8e8 d2g2 c6e7 g2g5 h7h6 g5f5 e7f5');
INSERT INTO evaluations VALUES(55593,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(55593,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(55593,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55593,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55593,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55593,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55593,36,'cp 8','b1c3 g8f6 c1b2 c7c5 f2f4 a7a6 f1c4 e5f4 d1f3 f8e7 g1e2 b8c6 c3d5 c6e5 f3f4 b7b5 d5f6 e7f6 c4d5 a8a7 e1g1 e8g8 f4e3 e5g4 b2f6 g4e3 f6d8 e3d5 d8a5 d5e7 d2d4 c8g4 e2f4 c5d4 a5b6 a7d7 b6d4 e7c6 d4b6 g8h8 a2a4 f8e8 a4b5 a6b5');
INSERT INTO evaluations VALUES(55715,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(55715,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(55715,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(55715,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(55715,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(55715,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(55715,36,'cp 51','d2d4 d6d5 g2g3 g8f6 f1g2 f8e7 g1f3 b8d7 e1g1 e8g8 d1d3 b7b6 c4d5 e6d5 c1f4 c8b7 f1d1 f6h5 f4d2 h5f6 a1c1 f8e8 d2f4 c7c5 f3e5 c5d4 c3b5 d7c5 d3d4 c5e6 d4d2 e7f8 e5c6 b7c6 c1c6 d8d7 c6e6 d7e6 b5c7');
INSERT INTO evaluations VALUES(55956,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(55956,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(55956,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(55956,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(55956,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(55956,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(55956,36,'cp -105','d5e4 b1c3 g8f6 d2d3 e4d3 f1d3 b8d7 f2f4 f8d6 e5c4 e8g8 c4d6 c7d6 c1e3 d7c5 d1e2 c5d3 e2d3 b7b6 e1c1 c8b7 e3d4 f6d5 c3d5 b7d5 f4f5 d8g5 d1d2 e6f5 d4e3 g5e7 e3f4 d5e4 d3d6 e7d6 f4d6 f8e8 h1d1 a8d8');
INSERT INTO evaluations VALUES(55978,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(55978,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(55978,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(55978,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(55978,34,'cp -108',NULL);
INSERT INTO evaluations VALUES(55978,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(55978,36,'cp -119','b1c3 g8f6 d2d3 e4d3 f1d3 b8d7 f2f4 f8d6 e5c4 d6b4 c1d2 d7c5 d3e2 b7b6 c4e5 b4c3 d2c3 e8g8 d1d4 c8b7 e1c1 c5e4 d4e3 d8e7 e2f3 e4c3 e3c3 c7c5 h1e1 e7c7 g2g3 a8d8 f3b7 c7b7 a2a3 d8d1 e1d1 b7c7 c3f3 f6d5 e5c4 d5e7');
INSERT INTO evaluations VALUES(56163,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(56163,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(56163,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(56163,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(56163,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(56163,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(56163,36,'cp -36','d8d5 d2d4 b8c6 g1f3 g8f6 c1e3 e5d4 c3d4 f8b4 b1c3 d5a5 d1b3 e8g8 f1c4 a5h5 e1g1 b4d6 b3d1 c8g4 h2h3 g4f3 d1f3 h5f3 g2f3 f8e8 c4b5 a7a6 b5c6 b7c6 f1c1 f6d5 c3d5 c6d5 c1c2 e8e6 a1c1 g8f8 g1f1 f8e8 f3f4 e6e4 f1g2 g7g6 b2b3');
INSERT INTO evaluations VALUES(34942,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(34942,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(34942,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(34942,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(34942,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(34942,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(34942,36,'cp -36','d2d4 g8f6 g1f3 b8c6 c1e3 e5d4 c3d4 f8b4 b1c3 d5a5 d1b3 e8g8 f1c4 a5h5 e1g1 b4d6 b3d1 c8g4 h2h3 g4f3 d1f3 h5f3 g2f3 c6e7 c3e4 f6d5 e4d6 c7d6 f1e1 f8c8 a1c1 c8c6 e3d2 g8f8 g1g2 h7h6 a2a4 d5c7 b2b3 a8c8 d2b4 c7d5 b4d2 a7a6 g2h2 d5c7 d2b4 b7b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(34927,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(34927,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(34927,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(34927,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(34927,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(34927,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(34927,36,'cp -28','g8f6 g1f3 b8c6 c1e3 e5d4 c3d4 f8b4 b1c3 d5a5 d1b3 e8g8 f1c4 a5h5 e1g1 b4d6 b3d1 c8g4 h2h3 g4f3 d1f3 h5f3 g2f3 c6e7 c3e4 f6d5 e4d6 c7d6 a1c1 f8c8 e3d2 g8f8 a2a4 h7h6 g1g2 c8c6 f1e1 a8c8 b2b3 a7a6 g2h2 d5c7');
INSERT INTO evaluations VALUES(2298,30,'cp 306',NULL);
INSERT INTO evaluations VALUES(2298,31,'cp 283',NULL);
INSERT INTO evaluations VALUES(2298,32,'cp 291',NULL);
INSERT INTO evaluations VALUES(2298,33,'cp 296',NULL);
INSERT INTO evaluations VALUES(2298,34,'cp 283',NULL);
INSERT INTO evaluations VALUES(2298,35,'cp 298',NULL);
INSERT INTO evaluations VALUES(2298,36,'cp 331','c6a5 g5f7 d8e8 f7h8 a5c4 d2d3 e7b4 c2c3 c4b2 c3b4 b2d1 e1d1 f8b4 b1a3 a7a6 a3c2 b4c5 f2f4 d7d6 f4e5 d6e5 c1b2 c8g4 d1d2 a8d8 a1e1 f6d7 c2e3 g4e6 h1f1 e8e7 e3f5 e6f5 f1f5 c5b4 b2c3 b4c3 d2c3 d8h8 e1f1 h8f8 f5f8 d7f8 f1f5 f8d7 f5h5 h7h6 g2g4 c7c6 g4g5 h6g5 h5g5 e7f6 g5f5 f6e6 d3d4 e5d4 c3d4');
INSERT INTO evaluations VALUES(5645,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(5645,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(5645,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(5645,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(5645,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(5645,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(5645,36,'cp 33','g8f6 d4d5 c6b8 c4d3 b8a6 b2b4 c7c6 d5c6 b7c6 e1g1 e8g8 f1e1 f8e8 h2h3 h7h6 a2a3 a6c7 b1d2 e7f8 d1c2 a7a5 d2c4 a5b4 c3b4 c8a6 d3f1 a6b5 c1e3 b5a4 c2b2 a4b5 f3d2 c7e6 a1b1 a8b8 e1d1 d8c7 c4a5 b5f1 d1f1');
INSERT INTO evaluations VALUES(8619,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(8619,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(8619,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(8619,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(8619,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(8619,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(8619,36,'cp 31','d4d5 c6b8 c4d3 b8a6 b2b4 c7c6 d5c6 b7c6 h2h3 a6c7 e1g1 e8g8 b1d2 a7a5 f1e1 a5b4 c3b4 c8a6 d3a6 a8a6 c1b2 d8a8 a2a3 f8b8 d2c4 f6e8 b2c3 c7b5 d1c2 h7h6 c2b3 b5c3 b3c3 e8c7 c3c2 c7e6 e1d1');
INSERT INTO evaluations VALUES(9723,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(9723,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(9723,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(9723,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(9723,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(9723,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(9723,36,'cp -31','d8f6 b1c3 c6e7 e1g1 e8g8 c3a4 c5b4 a2a3 b4d6 d3d4 e5d4 d1d4 f6d4 f3d4 c7c6 a4c3 f8d8 a3a4 a7a5 a1d1 g8f8 h2h3 d6c5 e4e5 d7d5 e5d6 c5d6 c3e4 d6f4 e4c5 f4c7 d4f3 e7g6 d1d8 c7d8 f1d1 d8c7 c5d7 f8e8');
INSERT INTO evaluations VALUES(15969,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(15969,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(15969,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(15969,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(15969,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(15969,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(15969,36,'cp 17','h7h6 g5h4 f8e7 e4f6 e7f6 h4f6 d8f6 d1d2 e8g8 d2e3 f8d8 f1d3 c7c5 e1c1 c5d4 e3d4 f6d4 f3d4 d7c5 d3e2 c8d7 e2f3 a8b8 h1e1 g8f8 b2b3 a7a5 g2g3 b8c8 f3g2 b7b5 f2f4 b5b4 c1b2 d7e8 a2a4 g7g5 f4f5');
INSERT INTO evaluations VALUES(16273,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(16273,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(16273,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(16273,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(16273,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(16273,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(16273,36,'cp 45','d7d5 b1c3 f8d6 c4c5 d6e7 a1b1 a7a5 a2a3 e8g8 b2b4 f6e4 d1c2 a5b4 a3b4 e4c3 c2c3 e6e5 d4e5 f7f6 e5e6 c8e6 f1d3 d8d7 e1g1 e6f5 f1d1 g8h8 b4b5 c6d8 d3f5 d7f5 c1b2 d8e6 c5c6 b7b6 c3b3 f8d8 f3d4 e6d4 b2d4');
INSERT INTO evaluations VALUES(19286,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19286,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19286,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(19286,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(19286,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(19286,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(19286,36,'cp -5','f1g2 f8g7 e1g1 d7d6 c2c3 e7e5 d2d3 g8e7 a2a3 a7a6 f1e1 e8g8 b2b4 c8g4 b1d2 a8c8 d2b3 b7b6 b4c5 b6c5 a1b1 c8b8 c1e3 f7f5 h2h3 g4f3 g2f3 f5f4 e3c1 g8h8 f3g2 g7h6 g3g4 b8b6 b3d2 b6b1 d2b1 d8a5 d1c2 f8b8');
INSERT INTO evaluations VALUES(20969,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(20969,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(20969,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(20969,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(20969,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(20969,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(20969,36,'cp 28','a2a3 a7a5 b4b5 c6a7 e2e3 c8d7 c2c4 d5c4 a3a4 c7c6 b5c6 a7c6 f1c4 g8f6 g1f3 e7e6 e1g1 f8b4 d1b3 e8g8 d2d4 a8c8 b1c3 h7h6 c3b5 c6e7 f1c1 e7d5 f3e5 d7e8 h2h3 f6e4 c4d3 e4d2 c1c8 d8c8 a1c1 d2b3 c1c8 e8b5 c8f8 g8f8 d3b5');
INSERT INTO evaluations VALUES(21677,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(21677,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(21677,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(21677,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(21677,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(21677,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(21677,36,'cp -56','e5d4 e3d4 c8f5 g1f3 d8f6 f1d3 g8e7 e1g1 e8c8 f3e5 h7h5 f1e1 g7g5 e5c6 f6c6 b1c3 c6g6 e1e5 f5d3 d1d3 g6d3 c2d3 g5g4 a1e1 e7c6 e5f5 c6b4 e1d1 c7c6 a2a3 b4c2 c3d5 d8d5 f5d5 c6d5 d1c1 c8d7 c1c2 f7f5 b2c1 f8d6 g2g3 h5h4 g1f1 h4h3 b3b4 h8f8 f1e2 f5f4 c1f4 d6f4');
INSERT INTO evaluations VALUES(22089,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(22089,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(22089,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(22089,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(22089,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(22089,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(22089,36,'cp 94','b4a6 d1c2 f8e7 c3c4 g8f6 b1c3 c7c6 h2h3 e8g8 a2a3 c6d5 c4d5 c8d7 c1e3 a8c8 e3a7 f6h5 c2d1 a6c5 f3d2 g7g6 d2c4 e7h4 c4b6 d8g5 b6d7 c5d7 d1d2 g5f6 a7e3 d7c5 f1c4 b7b5 c4a2 h5f4 e1g1 h4g5 a1d1');
INSERT INTO evaluations VALUES(24951,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(24951,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(24951,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(24951,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(24951,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(24951,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(24951,36,'cp 69','f1b5 e7e6 h2h3 g4h5 c2c4 f8e7 b5c6 b7c6 d1a4 e8g8 a4c6 d5c4 d2c4 h5g6 f3d2 f6d5 a2a3 e7d6 e1g1 a7a5 f1e1 f7f6 c6a4 d5b6 c4b6 c7b6 e3e4 d8e8 a4e8 f8e8 e4e5 d6e7 e5f6 e7f6 d2f3 a8c8 c1d2 g6c2 f3e5 c2a4 a1c1 a4b3 f2f3 a5a4 d2b4 c8c1 e1c1 e8d8');
INSERT INTO evaluations VALUES(34337,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(34337,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(34337,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(34337,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(34337,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(34337,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(34337,36,'cp 78','g8f6 g2g3 c8f5 c1f4 d6b4 f1d3 f5d3 d1d3 e7e6 e1c1 b8d7 c1b1 f6d5 c3e4 d5f4 g3f4 f8e7 f3e5 d7e5 f4e5 e8c8 c2c4 c8b8 a2a3 b4a5 h2h4 d8d7 c4c5 d7d5 h4h5 h8d8 e4d6 e7d6 c5d6 g7g6 b1a2 a5b5 d3b5 c6b5 h5g6 h7g6');
INSERT INTO evaluations VALUES(39033,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(39033,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(39033,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(39033,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(39033,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(39033,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(39033,36,'cp 85','g1f3 d7d6 f1d3 e6e5 d4e5 d6e5 e1g1 d8e7 a2a4 g8f6 d1c2 e8g8 a4a5 a7a6 f1d1 b7b6 a5b6 c7b6 h2h3 b8d7 c3c4 c8b7 b1c3 d7c5 b2b3 a8c8 c1a3 f8d8 c3d5 f6d5 c4d5 g7f8 d3f1 a6a5 f3d2');
INSERT INTO evaluations VALUES(39099,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(39099,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(39099,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(39099,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(39099,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(39099,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(39099,36,'cp 20','f1c4 d8e7 d2d3 g8f6 h5h4 d6c5 b1c3 c7c6 c4b3 d7d6 f2f4 a7a5 g1f3 b8d7 e1d1 e5f4 c1f4 d7e5 f3e5 d6e5 f4g5 h7h6 h1f1 h8g8 g5d2 c8g4 d1e1 e8c8 h2h3 g4h5 f1f6 g7f6 h4h5 g8g2 e1d1 b7b5 a2a4 b5b4 c3e2 g2f2');
INSERT INTO evaluations VALUES(39098,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(39098,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(39098,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39098,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(39098,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(39098,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(39098,36,'cp 22','d8e7 b1c3 g8f6 h5h4 d6b4 g1e2 d7d6 d2d3 c8e6 a2a3 b4c3 e2c3 b8c6 c1g5 h7h6 c4b5 h8g8 b5c6 b7c6 g5d2 g7g5 h4g3 e8c8 e1c1 f6d7 d2e3 c8b7 h2h4 g5h4 g3h4 e7h4 h1h4 g8g2 h4h6 d8g8 d1h1 d7f8 b2b4 a7a5 b4a5 c6c5 h6h8 b7c6 h8g8');
INSERT INTO evaluations VALUES(39512,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(39512,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(39512,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(39512,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(39512,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(39512,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(39512,36,'cp 141','f3e5 g8f6 b1c3 e8g8 d2d4 e7b4 c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g2f3 b8c6 e5c6 b7c6 e1c1 f8d8 h2h4 b4e7 f1e2 a7a5 c3a4 d7d6 f3f4 c8d7 b2b3 g8f8 e4e5 g7g6 h4h5 g6g5 e5d6 c7d6 a4b6 a8a7 b6d7 d8d7 c2c3 g5f4 d4d5 c6d5 d1d5');
INSERT INTO evaluations VALUES(39507,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(39507,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(39507,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(39507,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(39507,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(39507,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(39507,36,'cp 148','g8f6 b1c3 e8g8 d2d4 e7b4 c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g2f3 b8c6 e5c6 b7c6 e1c1 a8b8 h2h4 b4e7 b2b3 d7d6 f3f4 d6d5 f2f3 f8d8 c3a4 d5e4 f3e4 c8g4 d1d3 e7d6 f4f5 d8e8 f1g2 g4e2 d3d2 d6f4 a4c5 b8d8 c2c3 e2g4 h1g1');
INSERT INTO evaluations VALUES(42107,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(42107,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(42107,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(42107,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(42107,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(42107,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(42107,36,'cp 28','b8c6 g1f3 d8e7 c2c4 d5c4 d3c4 e8g8 f3e5 f8d8 e5c6 b7c6 d1b3 c6c5 b3a3 a8b8 d4c5 f6e4 b1c3 e7c5 a3c5 e4c5 b2b3 c8b7 e1g1 g8f8 f1c1 f8e7 f2f3 h7h6 c4f1 d8d2 c3b1 d2d5 f1e2 b7a6 e2a6 c5a6');
INSERT INTO evaluations VALUES(9011,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(9011,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(9011,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(9011,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(9011,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(9011,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(9011,36,'cp 31','f2f4 d8e7 a2a3 c8d7 b1d2 e8c8 d1e2 e7e8 e1c1 h7h5 g1f3 c6e7 e3e4 f6e4 d2e4 d5e4 d3e4 c8b8 h1e1 f7f6 c1b1 d7c8 e2f2 e8f7 d1d2 h8e8 c2c4 g7g6 f3h4 f7g7 e4c2 a7a6 d2e2 g6g5 f4g5 g7g5 h4f3 g5h6');
INSERT INTO evaluations VALUES(49971,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(49971,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(49971,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(49971,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(49971,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(49971,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(49971,36,'cp 117','e2e4 f5e4 c3e4 f8e7 g5f6 e7f6 c2c3 e8g8 g1f3 b8c6 h2h4 d7d5 e4f6 d8f6 d1d2 a7a6 e1c1 c8d7 d1e1 a8e8 e1e3 g8h8 d2e1 f6h6 f3g5 e6e5 d4e5 d5d4 c3d4 c6d4 f1c4 d7f5 e5e6 d4c2 g5f7 f8f7 e6f7 e8f8 e1c3 c2e3 f2e3 f5e6 h1d1 e6c4 c3c4 h6e3 c1b1 h7h6 c4c7');
INSERT INTO evaluations VALUES(49974,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(49974,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(49974,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(49974,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(49974,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(49974,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(49974,36,'cp 108','f5e4 c3e4 f8e7 g5f6 e7f6 c2c3 e8g8 g1f3 b8c6 h2h4 d7d5 e4f6 d8f6 d1d2 a7a6 e1c1 f6f4 h4h5 h7h6 f1e2 a6a5 g2g3 f4d2 d1d2 f8f6 e2d1 c8d7 f3h4 c6e7 h4g2 e7f5 g2f4 f5d6 d2e2 d6e4 d1c2 d7b5 e2e3 a5a4 h1e1 b5c4 b2b3 a4b3 a2b3 a8a1 c1b2 a1e1');
INSERT INTO evaluations VALUES(49976,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(49976,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(49976,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(49976,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(49976,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(49976,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(49976,36,'cp 114','c3e4 f8e7 g5f6 e7f6 c2c3 e8g8 g1f3 b8c6 h2h4 d7d5 e4f6 d8f6 d1d2 a7a6 e1c1 c8d7 d1e1 a8e8 h4h5 h7h6 d2e3 f6f4 f1d3 e8e7 d3c2 e7f7 h1h4 f4e3 f2e3 b7b6 e3e4 a6a5 c1d2 g8h8 d2e3 d7e8 h4h1 c6e7 e4e5 f7f4');
INSERT INTO evaluations VALUES(49975,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(49975,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(49975,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(49975,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(49975,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(49975,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(49975,36,'cp 97','f8e7 g5f6 e7f6 c2c3 e8g8 g1f3 b8c6 h2h4 d7d5 e4f6 d8f6 d1d2 a7a6 e1c1 a8b8 d1e1 b7b5 f1d3 b5b4 c3c4 a6a5 h4h5 h7h6 c4c5 a5a4 d3b1 a4a3 b2b3 c8a6 b1g6 a6b5 d2e3 e6e5 d4e5 f6e7 c1b1 f8d8 f3d4 e7c5 d4b5 c5e3 e1e3 b8b5 f2f4 b5c5');
INSERT INTO evaluations VALUES(42952,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(42952,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(42952,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(42952,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(42952,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(42952,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(42952,36,'cp 112','g5f6 e7f6 c2c3 e8g8 g1f3 b8c6 h2h4 d7d5 e4f6 g7f6 d1d2 e6e5 d4e5 c6e5 e1c1 c7c6 f1e2 d8d6 h4h5 c8g4 d1e1 e5f3 g2f3 g4f5 e2d3 f5d3 d2d3 g8h8 h5h6 f8g8 d3f5 g8g5 f5e6 d6e6 e1e6 g5g6 e6e7 b7b5 e7c7 f6f5 c1d2 a7a5 f3f4 a5a4 d2d3');
INSERT INTO evaluations VALUES(42928,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(42928,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(42928,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(42928,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(42928,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(42928,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(42928,36,'cp 92','e7f6 c2c3 e8g8 g1f3 b8c6 h2h4 d7d5 e4f6 g7f6 d1d2 e6e5 d4e5 c6e5 e1c1 c7c6 h4h5 c8g4 f1e2 d8d6 d1e1 g4d7 f3h4 a8e8 g2g4 e5f7 a2a3 f7g5 e2d3 e8e1 h1e1 f8e8 e1e8 d7e8 h4f5 d6h2 c1b1 e8d7 c3c4 d7f5 d3f5 d5c4 d2d8 g8g7 h5h6 h2h6 d8e7 g5f7');
INSERT INTO evaluations VALUES(42909,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(42909,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(42909,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(42909,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(42909,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(42909,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(42909,36,'cp 100','c2c3 e8g8 g1f3 b8c6 f1d3 d7d5 e4f6 g7f6 e1g1 e6e5 f3h4 d8e7 h4f5 c8f5 d3f5 g8h8 d1h5 e5d4 a1d1 e7f7 h5h3 a8d8 c3d4 c6e7 d1e1 e7f5 h3f5 d8e8 g2g3 c7c6 h2h4 f8g8 h4h5 e8e1 f1e1 g8g5 f5h3 f7h5 h3c8 g5g8 c8b7');
INSERT INTO evaluations VALUES(56624,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(56624,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(56624,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(56624,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(56624,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(56624,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(56624,36,'cp -11','f3e5 b8d7 d1e2 d7e5 f4e5 d8d4 b1c3 d4e5 c3e4 c8f5 d2d3 e8c8 e4g3 e5e2 f1e2 f5e6 e2f3 g8e7 c1d2 e7c6 f3c6 b7c6 e1g1 f8c5 g1h1 f7f6 a1e1 h8e8 b2b3 e6f7 g3f5 c5f8 h2h3 h7h5 e1e8 d8e8 f1e1 e8e1 d2e1 c6c5 g2g4 f7d5 h1g1 h5g4 h3g4');
INSERT INTO evaluations VALUES(56519,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(56519,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(56519,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(56519,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(56519,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(56519,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(56519,36,'cp -23','b8d7 d1e2 d7e5 f4e5 d8d5 b1c3 d5e5 c3e4 c8e6 e2b5 e5b5 f1b5 c7c6 b5d3 f8e7 b2b3 g8f6 c1b2 f6e4 d3e4 f7f6 e1c1 a8d8 h2h3 e6d5 d1e1 e8f7 e1e2 d5e4 e2e4 d8d5 h1e1 e7d6 b2d4 h7h5 c1d1 d5g5 e1e2 b7b6 d4e3 g5f5 e2f2 f5f2 e3f2 g7g5 d1e2');
INSERT INTO evaluations VALUES(5624,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(5624,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(5624,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(5624,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(5624,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(5624,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(5624,36,'cp 43','d6e5 d1d8 e7d8 b1c3 a7a6 c1e3 g8f6 a2a3 d8e7 c3d5 e7d6 d5f6 g7f6 e1c1 c6a5 c4e2 c8e6 f3h4 a5c4 e3h6 h8g8 g2g4 d6f8 h6f8 e8f8 e2c4 e6c4 h2h3 c4e6 d1d3 h7h5 h4f5 e6f5 g4f5 f8g7 h1d1');
INSERT INTO evaluations VALUES(5618,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(5618,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5618,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(5618,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(5618,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(5618,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(5618,36,'cp 54','d1d8 e7d8 b1c3 a7a6 c1e3 g8f6 e1c1 d8e7 c3d5 e7d6 d5f6 g7f6 f3h4 c8g4 f2f3 g4e6 c4e2 h7h5 g2g3 e8c8 c1b1 c6e7 c2c3 a6a5 b1c2 h8g8 e2d3 c7c6 a2a4 c8c7 h4f5 e6f5 e4f5 e7d5 e3c1 d5b6 b2b3 b6d5 d3e4 d6c5 d1d3 b7b6 c2b2 g8h8 h1d1');
INSERT INTO evaluations VALUES(2241,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(2241,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(2241,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(2241,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(2241,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(2241,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(2241,36,'cp 45','e7d8 c1e3 g8f6 b1c3 a7a6 e1c1 d8e7 c3d5 e7d6 d5f6 g7f6 f3h4 c8g4 f2f3 g4e6 c4e2 h7h5 g2g3 e8c8 c1b1 c6e7 h1f1 e7g6 h4f5 d6f8 d1d8 c8d8 e2d3 h5h4 g3g4 g6f4 f1d1 d8c8 e3f4 e5f4 h2h3 c7c6 c2c3 f8c5 d3c2 c5e3 c2b3 e6f5 e4f5 h8d8 d1d8 c8d8');
INSERT INTO evaluations VALUES(2267,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(2267,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(2267,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(2267,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(2267,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(2267,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(2267,36,'cp 47','c1e3 d8e7 b1c3 g8f6 c3d5 e7d6 d5f6 g7f6 f3h4 c8e6 c4b5 e8c8 e1c1 c6e7 b5d3 h7h5 a2a3 a7a5 d1d2 c7c6 g2g3 c8c7 c2c4 b7b6 h4f5 e7f5 e4f5 e6c8 d3c2 c8a6 h1d1 h5h4 e3b6 c7b6 d2d6 d8d6');
INSERT INTO evaluations VALUES(13252,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(13252,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(13252,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(13252,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(13252,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(13252,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(13252,36,'cp -32','g1e2 e7e6 c4b3 f8e7 e1g1 e8g8 a2a4 c6a5 b3a2 a7a6 c2c3 c8d7 b1d2 b7b5 a2b1 d8c7 h2h3 b5a4 b1c2 f8b8 c2a4 d7a4 a1a4 f6d7 a4a2 a5b3 d2b3 b8b3 e4e5 b3b7 e5d6 e7d6 c1f4 d6f4 e2f4 a6a5 f4h5 a5a4 h5g3 f7f5 f1a1');
INSERT INTO evaluations VALUES(14909,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(14909,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(14909,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(14909,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(14909,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(14909,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(14909,36,'cp 208','g1f3 f8e7 e1g1 a7a6 a2a3 b7b5 c4d3 c8b7 f1e1 a8c8 b2b4 e8g8 c1b2 f6d7 a1c1 d7b6 d1e2 c6b8 c1d1 b8d7 b2c1 f8e8 e4e5 d7f8 c3e4 b6c4 e4c5 b7d5 c5a6 c8a8 a6c5');
INSERT INTO evaluations VALUES(23138,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(23138,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(23138,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(23138,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(23138,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(23138,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(23138,36,'cp 55','d7d5 e4d5 g8f6 e1g1 f6d5 d1h5 c7c6 d2d3 f8e7 b1d2 e8g8 c4d5 d8d5 h5d5 c6d5 d2f3 e7c5 h2h3 f8e8 b2b3 e8e2 f1e1 e2e1 f3e1 c8d7 c1b2 a8c8 e1f3 f7f6 a1c1 h7h5 f3d4 g8f7 d4e2 g7g5 g1f1 h5h4 c2c4 d5c4');
INSERT INTO evaluations VALUES(47733,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47733,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(47733,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(47733,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(47733,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(47733,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(47733,36,'cp 48','f8e7 f1d3 b8c6 g1e2 e8g8 h2h3 c6b4 d3b1 e7g5 b2b3 g5c1 d1c1 c7c6 e1g1 d6d5 c3d1 b4a6 d1e3 a6c7 a2a4 b6d7 c1d2 d7f6 b1c2 a7a5 a1e1 b7b5 a4b5 c6b5 c4c5 b5b4 c2d3 c8a6 d3a6 a8a6 e3f5');
INSERT INTO evaluations VALUES(40730,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(40730,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(40730,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(40730,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(40730,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(40730,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(40730,36,'cp 62','f1d3 b8c6 g1e2 e8g8 h2h3 c6b4 d3b1 e7g5 b2b3 g5c1 d1c1 c7c6 c3d1 b6d7 c1d2 b4a6 d1e3 d7f6 e1g1 d6d5 b1d3 a6c7 a1e1 a7a5 a2a4 f8e8 f2f4 b7b5 a4b5 d5c4 b3c4 c6b5 c4b5 c7d5 e3c4 d5b4 c4e5 b4d3 d2d3 c8e6 f4f5');
INSERT INTO evaluations VALUES(52239,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(52239,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(52239,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(52239,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(52239,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(52239,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(52239,36,'cp 189','c5d4 f3d4 b8c6 b1c3 a7a6 c1e3 e7e6 d4c6 b7c6 f1d3 a8b8 c3a4 d8a5 c2c3 g8h6 e1g1 h6f7 b2b4 a5c7 a2a3 f8e7 c3c4 d7d6 c4c5 e8g8 d3e2 g8h8 d1c2 e6e5 a1d1 a6a5 a4b6 a5b4 a3b4 d6c5 b4c5 c8e6 e2c4 e6c4 c2c4 b8d8 h2h3 d8d1 f1d1');
INSERT INTO evaluations VALUES(52243,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(52243,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(52243,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(52243,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(52243,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(52243,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(52243,36,'cp 206','f3d4 b8c6 b1c3 a7a6 d4c6 d7c6 d1f3 e7e5 f1c4 g8e7 e1g1 e7g6 f1d1 d8e7 c4b3 c8e6 c1e3 a8b8 c3a4 e7f7 b3e6 f7e6 h2h3 e8f7 d1d3 b7b5 a4b6 f8e7 b6d7 b8d8 a1d1 h7h5 e3c5 e7c5 d7c5 d8d3');
INSERT INTO evaluations VALUES(52242,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(52242,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(52242,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(52242,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(52242,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(52242,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(52242,36,'cp 184','b8c6 b1c3 e7e6 d4b5 a7a6 b5d6 f8d6 d1d6 d8e7 d6g3 h7h5 h2h4 d7d5 c1f4 e7f7 e4d5 e6d5 e1c1 g8e7 c3d5 e7d5 f1c4 c8e6 c4d5 e6d5 d1d5 e8g8 g3b3 a8e8 h1d1 e8e2 f4g3 f8e8 d5h5 f7b3 a2b3 g8f7 d1d7 f7g8 c2c3 a6a5 d7b7');
INSERT INTO evaluations VALUES(55087,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(55087,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(55087,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(55087,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(55087,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(55087,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(55087,36,'cp -50','g8f6 e2e3 a7a6 g2g3 b7b5 f1g2 c8b7 d1e2 e7e6 g1f3 c7c5 e3e4 d5e4 f3e5 b8c6 e5c6 b7c6 c3e4 f6e4 g2e4 a8c8 d2d3 f8e7 c1e3 e8g8 e4c6 c8c6 e1c1 d8c7 c1b1 c5c4 d3c4 c6c4 a2a3 f8c8');
INSERT INTO evaluations VALUES(47240,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(47240,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(47240,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(47240,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(47240,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(47240,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(47240,36,'cp -56','e2e3 c7c5 g1f3 a7a6 d1e2 e7e6 a2a4 c5c4 d2d4 c4d3 c2d3 b8c6 d3d4 f6e4 e2d1 f8b4 d1c2 c8d7 f1d3 f7f5 e1g1 e8g8 c1d2 e4d2 c2d2 d7e8 f3g5 d8d6 g5f3 a8c8 d3e2 g8h8 a1c1 h7h6 f3e5');
INSERT INTO evaluations VALUES(55790,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(55790,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(55790,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(55790,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(55790,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(55790,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(55790,36,'cp -18','g8f6 b4b5 f8d6 c2c4 e8g8 g1f3 b8d7 c4d5 e6d5 d1c2 a7a6 a2a4 f8e8 f1e2 d7e5 f3e5 d6e5 b2e5 e8e5 b5a6 d5d4 b1a3 b7a6 e1g1 c8f5 e2d3 f5g6 d3g6 h7g6 a1c1 c7c5 a3c4 e5e8 d2d3 d4e3 c4e3 f6d7 f1d1 a8b8 d3d4 c5d4 d1d4');
INSERT INTO evaluations VALUES(48221,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(48221,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(48221,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(48221,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(48221,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48221,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(48221,36,'cp -12','b4b5 f8d6 g1f3 e8g8 a2a4 c7c6 c2c4 b8d7 d2d4 d6b4 b2c3 a7a5 b5a6 b4c3 b1c3 d8a5 d1d2 b7b6 c4d5 f6d5 c3d5 a5d5 f1e2 h7h6 e1g1 c8a6 f1c1 a6e2 d2e2 c6c5 c1b1 a8a5 d4c5 d7c5 b1b6 c5a4 b6b1 a4c3 e2e1 a5a1 b1a1 f8c8 f3d4');
INSERT INTO evaluations VALUES(48161,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(48161,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(48161,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(48161,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48161,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48161,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(48161,36,'cp -10','f8d6 g1f3 e8g8 f1e2 a7a6 a2a4 b8d7 d2d4 c7c5 b5c6 b7c6 b1d2 c6c5 c2c4 c5d4 f3d4 e6e5 d4f3 a8b8 b2a3 d6a3 a1a3 d5d4 e3d4 e5e4 f3e5 d7e5 d4e5 f6d7 e1g1 d7c5 d2b3 d8d1 e2d1 c5d3 b3a5 c8e6 d1c2 d3e5 c2e4');
INSERT INTO evaluations VALUES(57335,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(57335,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(57335,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(57335,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(57335,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(57335,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(57335,36,'cp 109','d7d6 g1f3 a7a6 f1c4 b8c6 e5d6 f8d6 e1g1 g8f6 f1e1 e8g8 a2a3 h7h6 c4a2 g8h7 a2e6 c8e6 e1e6 d8d7 e6e3 a8d8 b1c3 g7g5 e3e1 d8e8 e1e8 d7e8 h2h3 e8h5 d1d3 g5g4 h3g4 f6g4 c3e2 f8e8 e2g3 d6g3');
INSERT INTO evaluations VALUES(1445,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(1445,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(1445,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(1445,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(1445,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(1445,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(1445,36,'cp -89','f8c5 d1c2 b6c7 f1b5 g8e7 b2b4 c5b6 e1g1 e8g8 b5c6 e7c6 c2e2 f7f6 a2a4 a7a5 e5f6 g7f6 c1h6 f8f7 b4a5 c6a5 b1d2 e6e5 f3h4 g8h8 h6e3 b6e3 e2e3 c8e6 a1b1 f7g7 b1b5 g7g8 f1b1 c7f7 b5b6 g8g4 g2g3');
INSERT INTO evaluations VALUES(3353,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(3353,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(3353,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(3353,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(3353,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(3353,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(3353,36,'cp 21','a7a6 b5a4 b7b5 a4c2 d6d5 e4d5 d8d5 d2d4 e5d4 c3d4 c8f5 b1c3 f5c2 d1c2 d5d7 a2a4 a8b8 a4b5 a6b5 c1f4 b5b4 c3e2 b4b3 c2c4 b8b4 c4d3 f8e8 f4d2 b4b8 e2f4 e7d6 e1e8 d7e8 d3c4 e8d7 f4d3 d7f5 a1e1 f5d5 e1c1 c6e7 d2g5 d5c4 c1c4 b8a8 f3d2 a8a1 c4c1 a1a2');
INSERT INTO evaluations VALUES(2565,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(2565,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(2565,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(2565,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(2565,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(2565,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(2565,36,'cp 48','b5a4 b7b5 a4c2 c8b7 d2d4 e5d4 c3d4 c6b4 c2b3 b7e4 e1e4 f6e4 a2a3 a6a5 a3b4 a5b4 a1a8 d8a8 b3c2 e4f6 b1d2 a8b7 d1e2 b7c6 d2b3 c6c4 e2d1 c4d5 g2g4 h7h6 b3d2 f8e8 c2d3 d5b7 g4g5 h6g5 f3g5 c7c5 d2e4 c5c4');
INSERT INTO evaluations VALUES(4902,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4902,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4902,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(4902,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(4902,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(4902,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4902,36,'cp 17','g8f6 a2a4 c8b7 f1e1 c5b6 d2d4 e8g8 c1g5 h7h6 g5h4 e5d4 c3d4 g7g5 h4g3 f8e8 d4d5 c6e7 b3c2 e7g6 a4b5 a6b5 a1a8 d8a8 b1c3 b5b4 c3a4 b4b3 c2b3 e8e4 e1e4 f6e4 a4b6 c7b6 h2h4 a8e8 g3h2 g5h4 f3d4 h4h3 d4f5 g6e5 h2e5 e8e5');
INSERT INTO evaluations VALUES(12721,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(12721,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(12721,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(12721,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12721,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(12721,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(12721,36,'cp 36','g1f3 f8g7 d2d4 c5d4 f3d4 e8g8 e1g1 d8a5 e2e3 d7d6 c1d2 a5b6 f1e1 c8g4 d1b3 b6b3 a2b3 f8c8 h2h3 g4d7 a1a3 a8b8 d4c6 b7c6 c3e2 a7a6 e1a1 h7h5 a1a2 d6d5 d2c3 f6e4 b3b4 c6c5 c3g7 g8g7 b4c5 c8c5 c4d5 d7b5');
INSERT INTO evaluations VALUES(18572,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(18572,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(18572,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(18572,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(18572,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(18572,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(18572,36,'cp 122','e7e6 c2c3 c6c5 f1d3 g8e7 e1g1 e7c6 c1e3 d8b6 b2b3 f8e7 b1d2 e8g8 a2a3 a7a5 f1e1 c5d4 c3d4 b6d8 h2h3 f7f6 e5f6 d7f6 f3e5 c8d7 e5c6 d7c6 e3g5 d8d7 d2f3 e7d6 f3e5 d6e5 e1e5 c6b5 d3b5 d7b5 e5e6 f6e4 g5e3 a5a4 b3a4 b5a4');
INSERT INTO evaluations VALUES(20282,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(20282,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(20282,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(20282,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(20282,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(20282,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(20282,36,'cp 71','c5d4 f3d4 g8f6 e4d5 e6d5 d1d2 f8b4 f2f3 e8g8 e1c1 c8e6 a2a3 b4a5 d4b3 a5b6 f1b5 b6e3 d2e3 d8d6 g2g3 f6d7 h1e1 f8e8 b5c6 b7c6 c3e4 d6e5 e4c5 e5f6 c5d3 h7h6 e3f4 e6f5 e1e8 a8e8 b3d4 f5d3 f4f6');
INSERT INTO evaluations VALUES(21680,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(21680,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(21680,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(21680,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(21680,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(21680,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(21680,36,'cp 148','f3e5 c6e5 d4e5 d5d4 c3b5 c7c5 e2e3 a7a6 b5a3 g8e7 a3c4 e7c6 e3d4 c5d4 f1d3 c8e6 e1g1 b7b5 c4d6 f8d6 e5d6 d8d6 a2a4 e8g8 d1h5 g7g6 h5h4 e6c4 c1f4 c6e5 f1e1 f7f6 d3e4 a8c8 h2h3 b5a4 a1a4 f8e8 f4g3 g8h8 g1h1 f6f5');
INSERT INTO evaluations VALUES(22393,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(22393,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(22393,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(22393,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(22393,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(22393,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(22393,36,'cp 96','f1d3 c7c5 c2c3 g8e7 g1e2 d8b6 a2a3 f7f6 f2f4 e7f5 b2b4 c5c4 d3f5 e6f5 a3a4 f8e7 b1a3 d7f8 a3c2 f8e6 c2e3 e6c7 e1g1 c8d7 g2g4 f6e5 d4e5 h7h5 a4a5 b6g6 g4g5 e8c8 e2d4 h5h4 g1h1 c7b5 d4b5 d7b5');
INSERT INTO evaluations VALUES(27953,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(27953,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(27953,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(27953,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(27953,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(27953,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(27953,36,'cp 67','f7f5 f2f3 d8h4 g2g3 h4h5 e4f5 b8c6 a2a3 b4e7 d2e3 e8c8 d1d2 e7f6 d2f2 e6f5 g1e2 g8e7 e1c1 h5f7 h2h4 c6a5 c1b1 c8b8 e3c1 e7d5 h1f1 d5c3 e2c3 g7g6 g3g4 a5c6 c3e2 h8e8 d1e1 a7a6 g4f5 g6f5 c1g5 c6e7');
INSERT INTO evaluations VALUES(28509,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(28509,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(28509,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(28509,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(28509,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(28509,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(28509,36,'cp 30','e1g1 e7e6 f1e1 d8d7 a2a4 b5b4 c3a2 f8e7 a2b4 b7e4 f2f3 d7b7 b4a6 b8a6 f3e4 e8g8 d1f3 a6c5 e4e5 f6e4 e5d6 e7d6 d4b5 d6e5 b3c4 b7c6 c2c3 f7f5 b5d4 e5d4 c3d4 c5a4 c4d3 a4b6 a1b1 a8a4 d3e4 f5e4 f3e4 c6e4 e1e4 b6d5 c1d2 f8b8 e4e6 a4d4');
INSERT INTO evaluations VALUES(31642,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31642,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31642,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31642,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31642,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31642,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31642,36,'cp 0','g4f3 e2f3 b8c6 e1g1 g8e7 b1a3 c5d4 c3d4 e7f5 a3c2 h7h5 f3e2 a8c8 a1b1 g7g6 e2d3 d8h4 c1e3 f5e3 f2e3 f8h6 f1f3 c6e7 f3h3 h4g4 d3e2 g4g5 h3g3 g5h4 g3h3');
INSERT INTO evaluations VALUES(38170,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(38170,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(38170,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(38170,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(38170,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(38170,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(38170,36,'cp 101','a3b4 d4c3 d1g4 c3b2 c1b2 d8e7 b2a3 f7f6 b4b5 e7c7 g4b4 g8e7 e5f6 g7f6 f1e2 e8g8 g1f3 f8f7 e1g1 a7a5 b4b2 e6e5 c2c4 c8e6 c4d5 e7d5 a1c1 c7d8 f3d2 b8d7 g2g3 g8h8 d2e4 d8g8 e2f3 e6g4 f3g2 f7g7 e4d6 g4e6 f1d1 d7b6 d6e4 a8d8 e4c5 e6f5 g2f3');
INSERT INTO evaluations VALUES(38160,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(38160,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(38160,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(38160,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(38160,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(38160,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(38160,36,'cp 111','d4c3 d1g4 c3b2 c1b2 d8e7 b2a3 f7f6 b4b5 e7c7 g4b4 g8e7 e5f6 g7f6 f1e2 e8g8 g1f3 f8f7 e1g1 a7a5 b4b2 e6e5 c2c4 c8e6 c4d5 e7d5 a1c1 c7d8 f3d2 b8d7 g2g3 f6f5 d2c4 d8f6 c4d6 f7g7 d6b7 d5f4 b7d6 f4h3 g1h1 e6d5 f2f3 f5f4 d6e4 f6e6 g3g4 d5e4 f3e4');
INSERT INTO evaluations VALUES(38163,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(38163,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(38163,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(38163,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(38163,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(38163,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(38163,36,'cp 134','d1g4 c3b2 c1b2 d8e7 b2a3 f7f6 b4b5 e7c7 g4b4 g8e7 e5f6 g7f6 g1f3 a7a5 b4b2 b8d7 c2c4 b7b6 f1e2 d5c4 a1c1 c8b7 c1c4 d7c5 a3c5 b6c5 b5b6 c7d6 b2f6 h8f8 f6h6 e8c8 e1g1 c8b8 f1c1 e7f5 h6h7 d8d7 h7h5 b7a6 c4e4 a6b7 c1d1 f5d4 e4d4 c5d4');
INSERT INTO evaluations VALUES(38161,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(38161,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(38161,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(38161,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(38161,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(38161,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(38161,36,'cp 144','c3b2 c1b2 d8e7 b2a3 f7f6 b4b5 e7c7 g4b4 g8e7 e5f6 g7f6 g1f3 a7a5 b4b2 b8d7 c2c4 d7e5 f3d2 e8g8 a1c1 c7d8 f1e2 f8f7 e1g1 c8d7 c4d5 e6d5 f1d1 a8c8 d2f1 c8c1 b2c1 d8c8 c1c8 d7c8 f2f4 e5g6 g2g3 c8e6 f1e3 g6f8 e2h5 f7g7 g1f2 f8d7 h5f3 g7f7 h2h4');
INSERT INTO evaluations VALUES(38319,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(38319,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(38319,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(38319,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(38319,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(38319,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(38319,36,'cp -72','g8f6 d2d4 e8g8 e2d1 e7b4 f1e2 b4c3 b2c3 f6e4 c1b2 b8d7 c3c4 d7b6 c4d5 b6a4 c2c4 a4b2 d1b3 b2c4 e2c4 c8g4 f3e5 g4h5 b3h3 h5g6 e1g1 e4d2 f1c1 d2c4 e5g6 h7g6 c1c4 d8d5 c4c7 d5d4 a1c1 b7b5 h3c3 d4c3 c1c3 a7a5 g2g3 b5b4 c3c4 f8d8 g1g2 d8d2');
INSERT INTO evaluations VALUES(40602,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(40602,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(40602,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40602,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40602,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40602,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40602,36,'cp 0','c4d5 e6d5 f1e2 b8c6 g1f3 e8g8 e1g1 a7a6 b2b3 c5a7 c1b2 c8f5 a1c1 f8e8 e2d3 f5g4 d3e2');
INSERT INTO evaluations VALUES(40595,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40595,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40595,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40595,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40595,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40595,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40595,36,'cp 0','e6d5 f1e2 e8g8 g1f3 b8c6 e1g1 a7a6 b2b3 c5a7 c1b2 c8f5 a1c1 d8d6 c3b1 f8e8 f3d4 f5g6 d4c6 b7c6 b1d2 a6a5 d2c4 d6c7 c4d2');
INSERT INTO evaluations VALUES(40647,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40647,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40647,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40647,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40647,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40647,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40647,36,'cp 0','f1e2 b8c6 g1f3 e8g8 e1g1 a7a6 b2b3 c5a7 c1b2 c8f5 a1c1 f8e8 e2d3 f5g4 d3e2');
INSERT INTO evaluations VALUES(43911,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(43911,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(43911,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(43911,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(43911,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(43911,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(43911,36,'cp -39','b5b4 c3a4 d7d6 g2g3 g7g6 f1g2 b8d7 a2a3 a7a5 b2b3 f8g7 c1b2 e8g8 g1f3 f8e8 e1g1 a8a6 d1c2 e7e6 d5e6 e8e6 a1d1 e6e8 f1e1 c8b7 f3d2 b7g2 g1g2 d8a8 e2e4 d7e5 f2f3 f6d7 f3f4 e5c6 b2g7 g8g7 d2f3 b4a3 c2c3 c6d4 f3d4 c5d4 c3d4 g7g8');
INSERT INTO evaluations VALUES(43918,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(43918,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(43918,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(43918,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(43918,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(43918,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(43918,36,'cp -25','c3a4 d7d6 g2g3 g7g6 f1g2 d8c7 g1f3 f8g7 e1g1 e8g8 f1e1 f6e4 d1c2 f7f5 f3g5 e4g5 c1g5 f8f7 a1d1 b8d7 b2b3 d7f6 f2f3 c8d7 a4b2 a7a5 g5c1 a5a4 b2d3 f7f8 c1b2 a4b3 a2b3 a8a2 d1a1 f8a8 a1a2 a8a2 d3c1 a2a8 c1d3');
INSERT INTO evaluations VALUES(43926,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(43926,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(43926,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(43926,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(43926,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(43926,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(43926,36,'cp -28','d7d6 g2g3 g7g6 f1g2 d8c7 g1f3 f8g7 e1g1 e8g8 f1e1 f6e4 d1c2 f7f5 f3g5 e4g5 c1g5 f8f7 a1d1 b8d7 b2b3 d7f6 g5c1 c8d7 c1b2 a8f8 f2f4 c7a5 e2e3 d7a4 b3a4 f6g4 e3e4 g7b2 c2b2 f5e4 g2e4 a5a4 e4c2 a4e8 c2d3 g4f6 e1e2 g8h8 d1e1 a7a5');
INSERT INTO evaluations VALUES(47833,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47833,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(47833,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(47833,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(47833,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(47833,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(47833,36,'cp 31','c7c6 f1g2 f8e7 h2h3 a7a6 e1g1 e8g8 a2a4 c6c5 d2d3 b8c6 g1h2 a8b8 c1d2 c6b4 a4a5 h7h6 f3e1 b7b5 a5b6 d8b6 e1c2 c8b7 d2c1 a6a5 c2e3 b7g2 h2g2 f8e8 g2h2 b4c6 c3d5 f6d5 c4d5');
INSERT INTO evaluations VALUES(45869,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(45869,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(45869,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(45869,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45869,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(45869,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(45869,36,'cp 40','f1g2 f8e7 d1c2 c8g4 e1g1 e8g8 f1d1 c6c5 h2h3 g4h5 d2d3 f6e8 a2a3 a7a5 b2b3 b8c6 e2e3 e8c7 c1b2 d8d7 g1h2 f7f5 d1g1 e7f6 f3d2 a8e8 a1e1 g7g5 f2f4 g5g4 c3d5 f6g7 d5c7 d7c7 h3g4 h5g4 g2d5 g8h8 d5f3 g4f3 d2f3 g7f6 c2g2 c6e7 f3h4 c7d7 g1h1');
INSERT INTO evaluations VALUES(45870,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(45870,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45870,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(45870,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(45870,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(45870,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(45870,36,'cp 38','f8e7 d1c2 c8g4 e1g1 e8g8 f1d1 c6c5 h2h3 g4h5 d2d3 f6e8 a2a3 a7a5 e2e3 f7f5 b2b3 e8c7 g1h2 e7f6 d1g1 b8c6 c1b2 d8d7 f3d2 g8h8 a1e1 a8e8 c3d5 c7d5 c4d5 c6e7 a3a4 f5f4 g2e4 h5f7 c2c4 f7g6 e3f4 e5f4 b2f6 f4g3 f2g3 f8f6 g1f1 f6f8 f1f8 e8f8');
INSERT INTO evaluations VALUES(40134,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40134,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(40134,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(40134,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(40134,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40134,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(40134,36,'cp 39','d1c2 c8g4 d2d3 e8g8 e1g1 d8c8 e2e4 f8e8 c1e3 c6c5 f3e1 b8c6 f2f4 g4d7 e1f3 a8b8 c2d1 a7a6 c3d5 d7g4 d1d2 e5f4 g3f4 b7b5 b2b3 b5c4 b3c4 g4f3 g2f3 f6d5 c4d5 c6d4 e3d4 c5d4 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(50951,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50951,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(50951,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(50951,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(50951,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(50951,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(50951,36,'cp -12','c7c5 g1f3 g8f6 d4c5 f8c5 f1d3 b8d7 e3e4 e6e5 f4g5 d8b6 e1g1 d5e4 c3e4 f6e4 d3e4 e8g8 g5h4 f8e8 d1d3 h7h6 a1d1 c5f8 b2b3 d7c5 d3e3 b6c7 e4d5 c8e6 e3e5 c7e5 f3e5 e6d5 d1d5 c5e4 f1e1 e4c3 d5d7 c3a2 g1f1 a2b4 e5f7 b4c2 e1c1 c2b4 d7b7');
INSERT INTO evaluations VALUES(51286,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(51286,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(51286,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(51286,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(51286,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(51286,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(51286,36,'cp 67','e7e6 g1f3 g8f6 c4d5 e6d5 f3e5 f8e7 c1f4 e8g8 e2e3 c7c6 f1d3 f6d7 e5f3 c8a6 d3a6 b8a6 a1c1 b6b5 e1g1 f8e8 c3e2 a8c8 e2g3 g7g6 f4h6 d7b6 b2b3 c6c5 d4c5 a6c5 f3e5 a7a6 e5g4 c5e6 c1c8 d8c8');
INSERT INTO evaluations VALUES(55525,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(55525,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(55525,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(55525,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(55525,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(55525,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(55525,36,'cp -61','g7g6 h5d1 c7c6 d2d4 e5e4 g1e2 d6d5 c4b3 b7b5 a2a4 b5b4 e1g1 g8f6 b1d2 f8g7 c2c3 a7a5 f2f3 e4f3 g2f3 e8g8 e3e4 f6h5 f1f2 g8h8 d2f1 f7f5 e4f5 c8f5 f1g3 d8h4 g3h5 h4h5 e2g3 h5h4 g3f5 f8f5 b3c2 f5f7 f2g2 b8d7');
INSERT INTO evaluations VALUES(55321,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(55321,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(55321,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(55321,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(55321,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(55321,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(55321,36,'cp -50','h5d1 f8g7 d2d4 e5e4 g1e2 g8f6 e1g1 e8g8 a2a4 a7a5 c2c3 d6d5 c4b3 c7c6 c3c4 b8a6 c4d5 c6d5 c1d2 h7h5 b1a3 a6c7 h2h3 f8e8 a1c1 c8d7 e2c3 a8c8 a3b5 c7b5 a4b5 d7e6 c3a4 c8c1 d1c1 g7h6 a4c5 b7b6');
INSERT INTO evaluations VALUES(56310,30,'cp -727',NULL);
INSERT INTO evaluations VALUES(56310,31,'cp -725',NULL);
INSERT INTO evaluations VALUES(56310,32,'cp -717',NULL);
INSERT INTO evaluations VALUES(56310,33,'cp -732',NULL);
INSERT INTO evaluations VALUES(56310,34,'cp -732',NULL);
INSERT INTO evaluations VALUES(56310,35,'cp -741',NULL);
INSERT INTO evaluations VALUES(56310,36,'cp -736','d8h4 g2g3 h4e4 d1e2 e4h1 e2g2 h1g2 f1g2 c8e6 b1c3 g8e7 g1f3 e7f5 a2a3 f8e7 d2d4 e8g8 c1d2 b8d7 c3e2 d7b6 e2f4 b6c4 g2h3 c7c6 e1c1 a7a6 d1f1 h7h6 f1f2 g8h7 d2b4 a8e8 b4e7 e8e7 f4e6 f7e6 f3e1 g7g6 e1d3 c4e3 h3f5 e3f5 a3a4');
INSERT INTO evaluations VALUES(34892,30,'cp -721',NULL);
INSERT INTO evaluations VALUES(34892,31,'cp -713',NULL);
INSERT INTO evaluations VALUES(34892,32,'cp -721',NULL);
INSERT INTO evaluations VALUES(34892,33,'cp -701',NULL);
INSERT INTO evaluations VALUES(34892,34,'cp -718',NULL);
INSERT INTO evaluations VALUES(34892,35,'cp -725',NULL);
INSERT INTO evaluations VALUES(34892,36,'cp -732','g2g3 h4e4 d1e2 e4h1 e2g2 h1g2 f1g2 c8e6 b1c3 g8e7 g1f3 e7f5 g3g4 f5e7 g4g5 e7f5 c3e2 b8c6 h2h3 c6b4 e1d1 c7c5 e2f4 e8c8 c2c3 b4c6 d2d3 f8e7 c1d2 c8c7 f4e6 f7e6 d1c2 d5d4 a1f1 h8f8 h3h4 g7g6 a2a4 f5e3 d2e3 d4e3');
INSERT INTO evaluations VALUES(34891,30,'cp -732',NULL);
INSERT INTO evaluations VALUES(34891,31,'cp -721',NULL);
INSERT INTO evaluations VALUES(34891,32,'cp -728',NULL);
INSERT INTO evaluations VALUES(34891,33,'cp -744',NULL);
INSERT INTO evaluations VALUES(34891,34,'cp -732',NULL);
INSERT INTO evaluations VALUES(34891,35,'cp -736',NULL);
INSERT INTO evaluations VALUES(34891,36,'cp -732','h4e4 d1e2 e4h1 e2g2 h1g2 f1g2 c8e6 b1c3 g8e7 g1f3 e7f5 a2a3 f8e7 d2d4 e8g8 c1d2 b8d7 e1c1 d7b6 g3g4 f5h4 f3h4 e7h4 g4g5 b6c4 d2f4 c7c6 g2f1 a7a5 f1c4 d5c4 c3e4 f8d8 f4g3 h4g3 h2g3 g7g6 c2c3 e6d5 d1e1 g8g7 e4c5 b7b6');
INSERT INTO evaluations VALUES(34893,30,'cp -717',NULL);
INSERT INTO evaluations VALUES(34893,31,'cp -725',NULL);
INSERT INTO evaluations VALUES(34893,32,'cp -725',NULL);
INSERT INTO evaluations VALUES(34893,33,'cp -706',NULL);
INSERT INTO evaluations VALUES(34893,34,'cp -705',NULL);
INSERT INTO evaluations VALUES(34893,35,'cp -709',NULL);
INSERT INTO evaluations VALUES(34893,36,'cp -710','d1e2 e4h1 e2g2 h1g2 f1g2 c8e6 b1c3 g8e7 g1f3 e7f5 a2a3 f8e7 d2d4 e8g8 c1d2 b8d7 c3e2 a7a5 e2f4 d7b6 g2h3 b6c4 e1c1 c7c6 a3a4 b7b5 b2b3 e7a3 c1b1 c4d2 d1d2 b5a4 f4e6 f7e6 f3h4 a3b4 d2d3 a4b3 c2b3 g7g6 h4f5 g6f5 d3f3 b4e7 g3g4 f5f4 b1c2 g8g7');
INSERT INTO evaluations VALUES(57052,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(57052,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(57052,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(57052,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(57052,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(57052,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(57052,36,'cp -29','d7d5 b2b3 c7c5 c1b2 g8f6 f1e2 f8e7 g1f3 e8g8 e1g1 b7b6 f3e5 c8b7 e2f3 b6b5 d2d3 f6d7 c2c4 d7e5 f4e5 e7g5 f1e1 b8c6 c4d5 e6d5 d3d4 c5d4 e3d4 d8b6 b1a3 b5b4 a3c2 c6e7 b2c1 g5c1 a1c1 a8c8 g1h1 e7f5 d1e2');
INSERT INTO evaluations VALUES(57208,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(57208,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(57208,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(57208,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(57208,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(57208,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(57208,36,'cp -72','d7d5 d3e2 c7c6 d2d4 e5e4 c2c4 f8d6 b1c3 g8f6 c1d2 e8g8 c4d5 c6d5 d1b3 d6e7 g1h3 c8h3 g2h3 d8d7 f2f3 e4f3 e2f3 f8d8 e1g1 b8c6 c3e2 d7h3 e2f4 h3d7 g1h1 f6e4 f3e4 d5e4 b3c2 e7d6 c2e4 c6e7 f4h5 d7h3 h5f4 d6f4 f1f4 h3d7');
INSERT INTO evaluations VALUES(56455,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(56455,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(56455,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(56455,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(56455,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(56455,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(56455,36,'cp -88','c2c4 b8c6 c4d5 d8d5 d3f1 c8f5 d1a4 d5e6 b1c3 e8c8 g1f3 g8f6 f1c4 e6e8 e1g1 a7a6 d2d4 e5d4 f3d4 c6d4 a4e8 f6e8 e3d4 f5g6 c1e3 e8d6 c4b3 d6f5 a1d1 f8d6 c3e4 c8b8 e4d6 d8d6 f1e1 h8d8 d4d5 f5e3 f2e3 d6f6 a2a3 d8e8');
INSERT INTO evaluations VALUES(1854,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(1854,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(1854,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(1854,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(1854,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(1854,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(1854,36,'cp -22','g7g6 h5g5 f8h6 g5f6 g8f6 b1c3 d7d6 d2d3 h6c1 a1c1 h8f8 e1g1 c8g4 f3d2 g6g5 f2f3 g4d7 c3e2 e8c8 c2c3 h7h5 b2b4 h5h4 a2a4 f6h5 d3d4 c6e7 b4b5 e7g6 a4a5 c8b8 c1b1 h5f4 b5b6 a7b6 d4e5 d6e5 e2f4');
INSERT INTO evaluations VALUES(1788,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(1788,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(1788,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(1788,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(1788,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(1788,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(1788,36,'cp -15','h5g5 f8h6 g5f6 g8f6 b1c3 d7d6 d2d3 h6c1 a1c1 h8f8 f3g1 c6d4 g1e2 d4e2 c3e2 a7a5 a2a4 e8e7 f2f3 f6d7 c1a1 d7b6 c4b3 c7c6 d3d4 e5d4 e2d4 f7f5 e1g1 f5e4 f3e4 b6d7 f1f8 e7f8 a1f1 f8e7 d4e6');
INSERT INTO evaluations VALUES(1787,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(1787,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(1787,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(1787,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(1787,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(1787,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(1787,36,'cp -24','f8h6 g5f6 g8f6 b1c3 d7d6 d2d3 h6c1 a1c1 h8f8 f3g1 c6d4 g1e2 d4e2 c3e2 e8e7 a2a4 a7a5 f2f4 c8g4 h1f1 h7h5 e2g1 h5h4 e1d2 c7c6 f1f2 f6d7 c1f1 f7f6 b2b3 a8d8 d2e3 d7c5 g1f3 e5f4 e3f4 g4f3 f4f3 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(2796,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(2796,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(2796,32,'cp 215',NULL);
INSERT INTO evaluations VALUES(2796,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(2796,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(2796,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(2796,36,'cp 197','e4e5 f6e8 c3d4 c5e7 b1c3 c6a5 b3c2 d7d5 h2h3 c7c6 f1e1 e8c7 c3e2 a5c4 a2a4 b5a4 a1a4 a6a5 b2b3 c4b6 a4a1 c7e6 c1d2 a5a4 b3a4 b6c4 d2c3 g7g6 f3h2 c8a6 e2g3 e7g5 a4a5 c6c5 h2g4 c5d4 c3d4 a6b7 g4f6 g5f6 e5f6');
INSERT INTO evaluations VALUES(3745,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(3745,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(3745,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3745,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3745,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(3745,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3745,36,'cp 13','a2a4 h7h6 f1e1 c8e6 b1d2 f8e8 c2c3 d5f4 d2e4 e6c4 c1f4 c5f8 f3e5 c6e5 d3c4 d8d1 a1d1 e5c4 b2b3 c4a5 d1b1 e8e7 g2g4 a8d8 g1g2 b7b6 f2f3 e7d7 h3h4 f7f6 g2f2 g8f7 e1e2 a5b7 b3b4 g7g5 h4g5 h6g5 f4e3 c7c5 b4c5 b7c5 e4c5 f8c5 e3c5');
INSERT INTO evaluations VALUES(9157,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(9157,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9157,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(9157,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9157,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(9157,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(9157,36,'cp 17','e1g1 e8g8 a2a4 h7h6 f1e1 c8e6 b1d2 f8e8 c2c3 d5f4 d2e4 e6c4 c1f4 c5f8 f3e5 c6e5 d3c4 e5c4 b2b3 d8d1 a1d1 c4a5 d1b1 e8e7 g2g4 a8d8 f2f3 d8d5 g1g2 f7f5 c3c4 d5d3 g4f5 d3b3 b1b3 a5b3 e1b1 b3a5 c4c5 e7f7 f4e5 a5c4 e5c3 b7b6 f5f6');
INSERT INTO evaluations VALUES(9156,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(9156,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9156,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(9156,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(9156,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9156,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(9156,36,'cp 23','e8g8 a2a4 h7h6 b1d2 c8e6 f1e1 f8e8 c2c3 d5f4 d2e4 e6c4 c1f4 c5f8 f3e5 c6e5 d3c4 d8d1 a1d1 e5c4 b2b3 c4a5 d1b1 e8e7 g2g4 a8d8 g1g2 a5c6 g2f3 g7g5 f4g3 f8g7 b1d1 d8e8 e1e3 b7b6 b3b4 a7a5 b4b5 c6e5 g3e5 g7e5 f3g2 e5g7');
INSERT INTO evaluations VALUES(12168,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12168,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12168,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(12168,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12168,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12168,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12168,36,'cp 0','f8d6 f1d3 e8g8 e1g1 e6e5 c4d5 c6d5 d4e5 d7e5 d3e2 e5f3 e2f3 c8e6 c3b5 d6e5 b5d4 d8b6 d4e6 b6e6 d1b3 g7g5 c1d2 g5g4 f3e2 f6e4 a1d1 f8d8 d2c3 e5h2 g1h2 e6h6 h2g1 g4g3 f2g3 h6e3 g1h2 e3h6 h2g1');
INSERT INTO evaluations VALUES(7755,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7755,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7755,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7755,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7755,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7755,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7755,36,'cp 0','f1d3 e8g8 e1g1 e6e5 c4d5 c6d5 d4e5 d7e5 d3e2 e5f3 e2f3 c8e6 c3b5 d6e5 b5d4 d8b6 d4e6 b6e6 d1b3 g7g5 c1d2 g5g4 f3e2 f6e4 a1d1 f8d8 d2c3 e5h2 g1h2 e6h6 h2g1 g4g3 f2g3 h6e3 g1h2 e3h6 h2g1');
INSERT INTO evaluations VALUES(15084,30,'cp -239',NULL);
INSERT INTO evaluations VALUES(15084,31,'cp -244',NULL);
INSERT INTO evaluations VALUES(15084,32,'cp -232',NULL);
INSERT INTO evaluations VALUES(15084,33,'cp -251',NULL);
INSERT INTO evaluations VALUES(15084,34,'cp -250',NULL);
INSERT INTO evaluations VALUES(15084,35,'cp -259',NULL);
INSERT INTO evaluations VALUES(15084,36,'cp -240','f6e4 g5e4 d5e4 e1g1 d8d6 f2f3 e4f3 d1f3 f7f5 b1a3 c8d7 a3c4 d6g6 d2d3 e8c8 b3a4 c6d4 a4d7 d8d7 f3h3 g6e6 c1e3 g7g6 e3d4 d7d4 a1e1 f8g7 f1f2 e6d5 f2e2 d5c5 g1h1 e5e4 b2b3 e4d3 c2d3 d4d8 h3g3 d8d7 c4e5 g7e5 e2e5');
INSERT INTO evaluations VALUES(15375,30,'cp -251',NULL);
INSERT INTO evaluations VALUES(15375,31,'cp -263',NULL);
INSERT INTO evaluations VALUES(15375,32,'cp -265',NULL);
INSERT INTO evaluations VALUES(15375,33,'cp -262',NULL);
INSERT INTO evaluations VALUES(15375,34,'cp -240',NULL);
INSERT INTO evaluations VALUES(15375,35,'cp -262',NULL);
INSERT INTO evaluations VALUES(15375,36,'cp -252','g5e4 d5e4 e1g1 d8d6 f2f3 e4f3 d1f3 f7f5 c2c3 c8d7 d2d4 e8c8 c1g5 f8e7 g5e7 d6e7 d4d5 e7c5 f3f2 c5f2 g1f2 c6e7 c3c4 e5e4 b1c3 h8e8 a1e1 e7g6 b3c2 g6e5 c4c5 a7a5 f2e3 c7c6 d5d6 b7b6 c5b6 e5c4 e3d4 c4d6 c2a4 g7g6 h2h3 c8b7 d4c5');
INSERT INTO evaluations VALUES(15384,30,'cp -251',NULL);
INSERT INTO evaluations VALUES(15384,31,'cp -254',NULL);
INSERT INTO evaluations VALUES(15384,32,'cp -274',NULL);
INSERT INTO evaluations VALUES(15384,33,'cp -265',NULL);
INSERT INTO evaluations VALUES(15384,34,'cp -258',NULL);
INSERT INTO evaluations VALUES(15384,35,'cp -255',NULL);
INSERT INTO evaluations VALUES(15384,36,'cp -262','d5e4 e1g1 d8d6 f2f3 e4f3 d1f3 f7f5 c2c3 c8d7 d2d4 e8c8 c1g5 f8e7 g5e7 d6e7 d4d5 e7c5 f3f2 c5f2 g1f2 c6e7 c3c4 e5e4 b1c3 h8e8 b3c2 g7g5 a1e1 e7g6 c4c5 h7h5 b2b4 g6e5 h2h3 h5h4 f2e3 e5g6 b4b5 g6f4 f1f2 c7c6 b5c6 b7c6 d5d6 e8e5 e3d4 d8e8');
INSERT INTO evaluations VALUES(15399,30,'cp -247',NULL);
INSERT INTO evaluations VALUES(15399,31,'cp -263',NULL);
INSERT INTO evaluations VALUES(15399,32,'cp -267',NULL);
INSERT INTO evaluations VALUES(15399,33,'cp -266',NULL);
INSERT INTO evaluations VALUES(15399,34,'cp -246',NULL);
INSERT INTO evaluations VALUES(15399,35,'cp -240',NULL);
INSERT INTO evaluations VALUES(15399,36,'cp -255','e1g1 d8d6 f2f3 e4f3 d1f3 f7f5 c2c3 c8d7 d2d4 e8c8 c1g5 f8e7 g5e7 d6e7 d4d5 e7c5 f3f2 c5f2 g1f2 c6e7 c3c4 e5e4 b1c3 h8e8 b3c2 g7g5 a1e1 h7h5 c2d1 e8h8 f2g1 h5h4 b2b4 e7g6 d1e2 g6f4 c4c5 d7e8 e2c4 e8h5 c3b5 h4h3 g2g3 f4g2 d5d6 c7d6 b5d6 c8b8 e1c1 g2e3 f1e1 e3c4 c1c4');
INSERT INTO evaluations VALUES(18142,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(18142,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(18142,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(18142,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(18142,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(18142,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18142,36,'cp 40','g8e7 c3a4 e7g6 d2d3 f7f6 c1e3 d7d6 c2c3 f8e7 h2h4 e8g8 h4h5 g6e5 f3e5 f6e5 d1g4 g8h8 b2b3 c8d7 a4b2 h7h6 a1e1 f8f7 b2c4 d8c7 f2f3 f7f8 g4g3 e7f6 g3h3 f6e7 e1e2 d6d5 c4b2 d5d4 e3d2 d4c3 d2c3');
INSERT INTO evaluations VALUES(24572,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24572,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(24572,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24572,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24572,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24572,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(24572,36,'cp 3','e7e6 c2c4 a7a6 b1c3 f8e7 f3e1 d7b8 a2a4 b8c6 f2f4 e8g8 g2g4 f6e8 f4f5 a8b8 e1f3 e8c7 g1h1 d8d7 a4a5 b7b5 a5b6 b8b6 c3a4 b6b7 c1d2 e7f6 d2c3 f6c3 b2c3 d7e7 d1d2 f8b8 d2g2 f7f6 h2h4 g8h8 g4g5 e6f5 e4f5 e7d7 a1e1');
INSERT INTO evaluations VALUES(24517,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24517,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(24517,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24517,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(24517,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(24517,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(24517,36,'cp 18','c2c4 a7a6 b2b3 f8e7 c1b2 e8g8 d1e2 b7b5 b1c3 d8a5 a1d1 a8c8 e2d2 f8d8 h2h3 d7b8 e4e5 d6e5 f3e5 b8c6 e5c6 c8c6 c3d5 a5d2 d5e7 g8f8 d1d2 f8e7 f1d1 c6d6 d3d4 b5c4 b3c4 d8b8 d4d5 h7h6 f2f4 e6d5');
INSERT INTO evaluations VALUES(32073,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(32073,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(32073,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(32073,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(32073,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(32073,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(32073,36,'cp 99','d1f3 b8d7 h2h4 h7h5 f3b3 d7c5 f1b5 c7c6 d5c6 c5b3 c6c7 d8d7 b5d7 e8d7 a2b3 g8f6 b1c3 d7c7 h1h3 f8e7 c1g5 c7d7 h3d3 h8c8 f2f3 c8c6 b3b4 d7e6 b4b5 c6c5 e1c1 a7a6 g5f6 g7f6 b5a6 a8a6 g2g4 h5g4 f3g4 a6a8 d3g3 a8a1 c1d2 a1d1 d2d1 c5c8');
INSERT INTO evaluations VALUES(32072,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(32072,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(32072,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(32072,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(32072,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(32072,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(32072,36,'cp 96','b8d7 f3b3 g8f6 b1c3 d7c5 f1b5 f6d7 b5d7 d8d7 b3c4 f8e7 c1e3 e8g8 a2a4 h7h6 e3c5 d6c5 c4b5 d7c8 b5e2 c8d7 a4a5 e7d6 e2d3 f8b8 e1g1 g7g6 c3d1 b7b5 a5b6 a7b6 d1e3 b6b5 c2c4 b5c4 e3c4');
INSERT INTO evaluations VALUES(35898,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(35898,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(35898,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(35898,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(35898,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(35898,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(35898,36,'cp 60','f1e1 d6d5 c4b3 c8g4 h2h3 g4f3 d1f3 f6e4 c3e4 d5e4 e1e4 b8d7 c2c3 e7e6 e4e2 a7a5 a2a4 c6c5 d4c5 d7c5 b3c2 d8b6 a1b1 c5d7 c1e3 b6a6 e3f4 a8d8 b1d1 d7f6 c2d3 a6c6 f3c6 b7c6 e2d2 f6d5 f4g3 h7h6 c3c4 d5b4 g3c7 d8d3 d2d3 b4d3 d1d3 f8a8');
INSERT INTO evaluations VALUES(37389,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(37389,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(37389,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(37389,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(37389,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(37389,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(37389,36,'cp -20','c7c5 g1f3 c5d4 f3d4 f8e8 f4g3 b8c6 f1e2 c6d4 e3d4 d7d5 e1g1 d5c4 e2c4 d8b6 c3b5 c8e6 c4e6 b6b5 e6b3 b5b6 d1f3 a7a5 g3e5 e8f8 a1e1 a5a4 b3c4 b6b4 f3e2 e7e6 a2a3 b4b6 e1b1 f6d7 e5g7 g8g7 b2b3 d7f6');
INSERT INTO evaluations VALUES(38966,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(38966,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(38966,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(38966,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(38966,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(38966,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(38966,36,'cp 129','a7a5 d2b4 a5b4 d1d4 b8c6 d4d8 e8d8 e2e3 g8e7 a2a4 e7g6 b1d2 g6e5 f1e2 e5f3 e2f3 c6e5 f3e4 d8e7 b2b3 h8d8 e1e2 e5d7 e4c2 b7b6 d2f3 c8b7 h1d1 b7f3 e2f3 d7c5 d1d4 d8d6 h2h4 g7g6 d4d6 e7d6 f3e2 f7f5 a1d1 d6e7 e2f3 a8d8 d1d8 e7d8 g2g4 f5g4 f3g4 d8e7 f2f4');
INSERT INTO evaluations VALUES(41032,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(41032,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41032,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(41032,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41032,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(41032,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(41032,36,'cp 28','e3e4 c7c5 g1f3 c5d4 f3d4 e8g8 e1g1 d8c7 d1e2 b8c6 d4c6 b7c6 f1d1 c8b7 c1e3 f8d8 a1c1 h7h6 c4d3 f6d7 c3a4 d7e5 h2h3 d8d3 d1d3 e5d3 e2d3 a8d8 d3c4 c7a5 a4c5 e7c5 e3c5 b7a6 c4b4 a5b4 c5b4 d8d4 a2a3 d4e4 c1c6');
INSERT INTO evaluations VALUES(41492,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(41492,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(41492,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(41492,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(41492,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(41492,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(41492,36,'cp -35','c7c5 a2a4 a7a6 e1g1 e8g8 f1e1 d8c7 h2h3 f8d8 c1d2 c5c4 d3f1 b8c6 b2b3 e7e5 b3c4 d5c4 d1e2 e5d4 f3d4 c6d4 e3d4 c8e6 d2e3 a8c8 e2d2 f6d5 c3d5 e6d5 c2c3 b7b5 a4b5 a6b5 a1b1 c7d7 e3h6 g7h6 d2h6 d8e8 h6f4 h7h5 e1e5');
INSERT INTO evaluations VALUES(43054,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(43054,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(43054,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(43054,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(43054,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(43054,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(43054,36,'cp -24','b8c6 h2h4 h7h6 g5f4 d7d6 b1c3 d6e5 d4e5 d8d1 a1d1 c8d7 f1e2 a7a6 c3e4 e8c8 e4c5 e7d5 f4g3 d5b6 c2c3 c6e7 f3d4 b6a4 c5a4 d7a4 d1d2 e7d5 h4h5 g6g5 g3h2 d5f4 h2f4 g5f4 d4f3 d8d2 e1d2 a4c6 d2c2');
INSERT INTO evaluations VALUES(46970,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(46970,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(46970,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(46970,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(46970,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46970,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(46970,36,'cp 7','e7e6 e2e3 h7h6 g5h4 g7g5 h4g3 f6h5 c2c4 f8g7 b1c3 c7c5 d1d2 h5g3 h2g3 b8c6 a1d1 g5g4 d4d5 e6d5 c4d5 c6e5 f3e5 g7e5 h1h5 d8f6 f1b5 e8f8 b5d3 c8d7 d2c2 a7a6 a2a4 f8g7 c3e4 f6g6 h5e5 d6e5 c2c5 d7a4');
INSERT INTO evaluations VALUES(47059,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(47059,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(47059,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(47059,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(47059,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(47059,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(47059,36,'cp 44','d2d4 g7g6 f1g2 f8g7 e1g1 e8g8 c2c4 c8f5 b1c3 f6e4 c3a4 c7c5 d4d5 d8e8 d1b3 f5d7 b3b7 b8c6 d5c6 d7c6 b7a6 c6a4 f3g5 e4g5 c1g5 a8c8 a6a7 a4c6 g5e7 c6g2 e7f8 g2f1 f8g7 f1e2 g7c3 e2c4 a1e1 c4e6 b2b3 e8d7 a7d7 e6d7 a2a4 d7e6');
INSERT INTO evaluations VALUES(48106,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(48106,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(48106,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(48106,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(48106,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(48106,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(48106,36,'cp 80','g1f3 c7c5 f1c4 b8c6 e1g1 c5d4 f3d4 d5c3 b2c3 c8d7 f1e1 a8c8 c4b3 c6d4 c3d4 d7c6 e1e3 h7h5 e3d3 f8e7 d4d5 c6d5 b3a4 e8f8 a4b3 d8d7 b3d5 e6d5 d3d5 d7e6 d5d7 b7b6 c1b2 f8g8 d7a7 h8h6 d1f3 e7c5 a1d1 h6g6 a7a8 c8a8 f3a8 g8h7 a8e4 e6g4');
INSERT INTO evaluations VALUES(48105,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(48105,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(48105,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(48105,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(48105,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(48105,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(48105,36,'cp 76','c7c5 f1c4 b8c6 e1g1 c5d4 f3d4 d5c3 b2c3 c8d7 d1e2 f8e7 a1b1 c6d4 c3d4 d7c6 c4b5 e8g8 b5c6 b7c6 b1b7 f8e8 c1e3 a7a5 e2f3 d8d5 f1c1 a8b8 b7a7 b8a8 f3d5 c6d5 a7d7 a8d8 d7c7 e7a3 c1c2 h7h6 h2h4 d8b8');
INSERT INTO evaluations VALUES(49600,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(49600,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(49600,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(49600,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(49600,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(49600,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(49600,36,'cp -74','d2d3 h7h6 h2h3 d7d5 e4d5 f6d5 c2c4 d5e7 f1g2 b8c6 c1e3 h6h5 b1c3 c6d4 g4h5 e7f5 e3d4 e5d4 d1e2 c8e6 c3e4 f5h4 e1f1 h4g2 f1g2 f8e7 g1f3 h8h5 f3d4 e6h3 h1h3 h5h3 g2h3 d8d4 h3g2 e8c8 e2g4 c8b8 a1e1 e7f6 e4f6 d4f6 g4f3 f6f3 g2f3 d8d3 f3e4');
INSERT INTO evaluations VALUES(49597,30,'cp -110',NULL);
INSERT INTO evaluations VALUES(49597,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(49597,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(49597,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(49597,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(49597,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(49597,36,'cp -102','h7h6 h2h3 d7d5 e4d5 f6d5 c2c4 f8b4 c1d2 b4d2 d1d2 d5f4 b1c3 b7b6 e1c1 c8b7 h1h2 d8g5 d2e3 e8g8 d3d4 e5d4 d1d4 f4e6 d4d1 b8a6 e3g5 h6g5 c3d5 f8d8 f2f3 a6c5 h2d2 a7a5 g1e2 d8d7 e2c3 a8d8 c3e4 g8f8 e4c5 b6c5 f1d3 b7d5');
INSERT INTO evaluations VALUES(51895,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51895,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51895,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51895,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51895,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51895,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51895,36,'cp 0','f8c5 c3e4 g8f6 e4c5 d8a5 c1d2 a5c5 e2e3 b8c6 a2a3 a7a5 c2c4 e8g8 g1f3 e6e5 b2b4 c5e7 b4b5 e5e4 f3h4 c6e5 h4f5 e7e6 f5d4 e6e7');
INSERT INTO evaluations VALUES(37684,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(37684,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37684,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37684,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37684,36,'cp 16','e2e4 b8c6 f1d3 a7a6 g1f3 d7d6 e1g1 g8f6 c3a4 c5a7 c2c4 e8g8 a4c3 f6g4 c1f4 g4e5 d3e2 e5f3 e2f3 e6e5 f4e3 c6d4 f3e2 d8h4 g1h1 c8e6 e2d3 b7b5 b2b3 b5b4 c3e2 d4c6 e3a7 a8a7 a2a3 b4a3 a1a3 a6a5 d1d2 a7d7 d3b1 h4e7 e2c3');
INSERT INTO evaluations VALUES(54608,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(54608,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(54608,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(54608,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(54608,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(54608,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(54608,36,'cp -32','b8c6 g1e2 f8c5 e1g1 g8e7 b1c3 e8g8 c3a4 c5b4 d3d4 e5e4 c2c4 d5c4 b3c4 c8g4 a2a3 b4a5 a4c3 e7f5 c3e4 f5d4 g1h1 d4e2 c4e2 d8d1 e2d1 g4d1 f1d1 a5b6 b2b4 f7f6 c1e3 b6e3 f2e3 f8d8 e4c5 b7b6');
INSERT INTO evaluations VALUES(20555,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(20555,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(20555,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(20555,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(20555,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(20555,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(20555,36,'cp -31','g1e2 f8c5 b1c3 g8e7 e1g1 e8g8 c3a4 c5d6 a4c3 c8e6 d3d4 e5e4 f2f3 e4f3 f1f3 c6a5 h2h3 a8c8 c1g5 a5c4 b3c4 c8c4 g5f4 d6f4 f3f4 e6f5 d1d2 h7h6 f4f2 f8e8 a1e1 a7a6 b2b3 c4c8 e2f4 d8d7 e1e3 f5g6 f4g6 e7g6');
INSERT INTO evaluations VALUES(54892,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(54892,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(54892,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(54892,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(54892,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(54892,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(54892,36,'cp 100','c1g5 f8g7 g1f3 g8f6 d1d2 e8g8 g5h6 c7c6 h2h4 c8g4 h6g7 g8g7 f3e5 h7h5 f2f3 g4e6 e1c1 b8d7 d2g5 d7e5 d4e5 f6d5 f1c4 d8c7 c4d5 e6d5 c3d5 c6d5 g2g4 f8h8 h1g1 h5g4 f3g4 a8c8 c2c3 f7f6 e5f6 e7f6 g5d5 c8d8 d5h1 d8d1 g1d1 c7f4 d1d2 h8d8 h1b7 g7h8 b7g2 d8d2 g2d2 f4g4 d2d8 h8g7');
INSERT INTO evaluations VALUES(5660,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(5660,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(5660,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(5660,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(5660,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(5660,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(5660,36,'cp 65','d1b3 c6a5 c4f7 e8f8 b3a4 f8f7 a4a5 g8f6 e4e5 d6e5 f3e5 f7g8 e1g1 c7c5 a5a4 c8f5 f1e1 h7h6 h2h3 d8d5 c3d4 c5d4 e5f3 a8c8 a4d4 e7c5 d4d2 g8h7 b1c3 d5d2 c1d2 h8e8 d2e3 c5a3 e1e2 a7a6 e3d4 e8e2 c3e2 a3d6 a2a3 f5e4 e2c3 e4f3 g2f3 f6h5 d4e3 h5f4 a1d1 f4h3 g1f1');
INSERT INTO evaluations VALUES(5948,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(5948,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(5948,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(5948,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(5948,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(5948,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(5948,36,'cp 15','g8e7 c2c3 a7a6 b5a4 c5a7 d2d4 e8g8 a4c2 e5d4 c3d4 d6d5 e4d5 c6b4 c2b3 b4d5 b1c3 c8e6 c1g5 f7f6 g5d2 e6f7 c3e4 b7b6 f1e1 d8d7 a1c1 f8e8 a2a3 g8h8 e4c3 a8d8 c3d5 e7d5 e1e8 f7e8 b3d5 d7d5 c1c7 a7b8 c7c1 h7h6 d1c2 b8d6 d2b4 d6b4 a3b4 h8g8');
INSERT INTO evaluations VALUES(7677,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(7677,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(7677,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(7677,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(7677,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7677,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(7677,36,'cp -3','d8b6 g5f6 e7f6 a1b1 f6e7 e1g1 b6c7 b2b4 a8b8 d1f3 e8g8 b4b5 c8b7 f3d3 g7g6 f1d1 b8c8 b5c6 b7c6 c4d5 e7c5 d5c6 d7c6 c3a4 f8d8 d3a6 d8d1 b1d1 c5f8 g2g3 c8b8 g1g2 h7h5 h2h4 g8g7 a6d3 f8e7');
INSERT INTO evaluations VALUES(7775,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(7775,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(7775,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(7775,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(7775,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(7775,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(7775,36,'cp 49','f1d3 d5c4 d3c4 e8g8 e1g1 b7b5 c4d3 b5b4 c3e4 c6c5 a1c1 c8b7 f3d2 f6d5 e4d6 b7c8 d6c8 a8c8 d3a6 c8c6 a6b7 c6b6 b7d5 e6d5 b2b3 b6c6 h2h3 f8e8 d1f3 d7f8 f1d1 f8e6 d4c5 c6c5 c1c5 e6c5 f4e5 d8d7 f3e2 d7e6 d2f3 e8c8');
INSERT INTO evaluations VALUES(7776,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(7776,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(7776,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(7776,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(7776,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(7776,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(7776,36,'cp 35','f6h5 f4e5 h5f6 e5g3 e8g8 c4d5 e6d5 h2h3 g7g6 a2a3 d7b6 e1g1 c8e6 f1e1 f6h5 g3h2 f8e8 d1c2 h5g7 e1e2 g7f5 a1e1 f5d6 h2g3 e7f6 b2b3 e6f5 a3a4 f5d3 c2d3 d6e4 g3h2 e4c3 d3c3 a8c8 h2g3 b6d7');
INSERT INTO evaluations VALUES(8333,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(8333,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8333,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(8333,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(8333,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(8333,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(8333,36,'cp 14','e8g8 d1d2 c6d4 e3d4 c8e6 f2f4 a8c8 a1d1 a7a6 h2h3 b7b5 a2a3 e6c4 e2c4 c8c4 d2e3 d8a8 b2b3 c4c6 f4f5 a8b7 g1h1 c6c8 d1d3 a6a5 d4f6 g7f6 c3d5 c8c2 d5f6 e7f6 f5g6 h7g6 f1f6 f8e8 d3d4 b7c6');
INSERT INTO evaluations VALUES(11641,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11641,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(11641,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11641,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(11641,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(11641,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(11641,36,'cp 29','f8e7 d2d4 e8g8 h2h3 c8b7 f1e1 f8e8 b1d2 e7f8 a2a4 e5d4 c3d4 h7h6 b3c2 c6b4 c2b1 c7c5 d4d5 f6d7 d2f1 f7f5 e4f5 e8e1 d1e1 d8e7 f1e3 b4d5 a4b5 a6b5 a1a8 b7a8 e1a5 d5b6 a5b5 a8f3 g2f3 d6d5 e3g2 c5c4 b1c2');
INSERT INTO evaluations VALUES(7393,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(7393,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(7393,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(7393,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(7393,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(7393,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(7393,36,'cp 38','d2d4 e8g8 h2h3 f8e8 b1d2 c8b7 b3c2 e7f8 f1e1 h7h6 d4d5 c6b8 b2b3 a6a5 c3c4 c7c6 d2f1 b8d7 c1e3 a8c8 f1g3 b7a6 f3d2 c6d5 c4d5 b5b4 a2a3 d8c7 a1a2 c8b8 a3b4 b8b4 d2f3 d7c5 e3d2 b4b5 g3f5');
INSERT INTO evaluations VALUES(7385,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(7385,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(7385,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(7385,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7385,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(7385,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(7385,36,'cp 35','e8g8 h2h3 f8e8 b1d2 c8b7 f1e1 e7f8 b3c2 h7h6 d4d5 c6b8 b2b3 c7c6 c3c4 b8d7 d2f1 a6a5 c1e3 b7a6 a1c1 a8c8 f3d2 g7g6 a2a3 b5c4 b3c4 c6d5 c4d5 d7b6 d1f3 f6d7 c2b3 a5a4 b3a2 d8e7 e1d1 d7c5 e3c5 c8c5 c1c5');
INSERT INTO evaluations VALUES(15202,30,'cp 241',NULL);
INSERT INTO evaluations VALUES(15202,31,'cp 280',NULL);
INSERT INTO evaluations VALUES(15202,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(15202,33,'cp 270',NULL);
INSERT INTO evaluations VALUES(15202,34,'cp 274',NULL);
INSERT INTO evaluations VALUES(15202,35,'cp 269',NULL);
INSERT INTO evaluations VALUES(15202,36,'cp 282','d1d8 c6d8 f3g5 h7h6 e5f6 h6g5 c1g5 c8f5 f1c4 d8e6 c4e6 f7e6 f6g7 f8g7 e1c1 g7c3 b2c3 e8f7 d1d7 f7g6 h2h4 h8h7 d7h7 g6h7 h1d1 e6e5 d1d5 a8e8 c1d2 h7g7 c3c4 g7f7 d2e3 b7b6 d5d1 f7e6 d1g1 e8f8 g2g4 f5g6 h4h5 f8f3 e3d2 g6e8 g5e3 f3h3 c4c5 e6f7 h5h6 e8d7 g4g5 f7g6 c2c4 d7e6 c5b6');
INSERT INTO evaluations VALUES(2249,30,'cp 261',NULL);
INSERT INTO evaluations VALUES(2249,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(2249,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(2249,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(2249,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(2249,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(2249,36,'cp 276','c6d8 f3g5 h7h6 e5f6 h6g5 c1g5 d8e6 f6g7 f8g7 g5e3 f7f5 e1c1 c8d7 f1c4 g7c3 b2c3 b7b5 c4b3 f5f4 e3d2 e8c8 d1e1 d8g8 e1e4 e6c5 e4f4 c5b3 a2b3 g8g2 h2h4 d7c6 h4h5 g2f2 f4f2 c6h1 h5h6 h8h7 d2g5 a7a6 f2f4 h1d5 c3c4 b5c4 b3c4');
INSERT INTO evaluations VALUES(2378,30,'cp 257',NULL);
INSERT INTO evaluations VALUES(2378,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(2378,32,'cp 256',NULL);
INSERT INTO evaluations VALUES(2378,33,'cp 291',NULL);
INSERT INTO evaluations VALUES(2378,34,'cp 279',NULL);
INSERT INTO evaluations VALUES(2378,35,'cp 259',NULL);
INSERT INTO evaluations VALUES(2378,36,'cp 253','f3g5 f6g4 c3d5 e8d7 f1e2 d7c6 d5c3 g4e5 g5e4 c8e6 e2b5 c6b6 c1e3 c7c5 b2b4 a7a6 b4c5 b6a7 b5e2 d8c6 e4d6 f8d6 c5d6 b7b6 e1c1 h8d8 h1e1 a8b8 e3f4 d8d7 f4g3 b8d8 f2f4 e5g6 e2f3 c6a5 c3d5 a5c4 d5c7 e6f5 e1e8 d8e8 c7e8');
INSERT INTO evaluations VALUES(16556,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(16556,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(16556,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(16556,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(16556,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(16556,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16556,36,'cp 0','c2c4 a7a6 f1g2 f8c5 g1e2 f6g4 e1g1 d8f6 d3d4 e5d4 e2f4 g4e5 f4h5 f6g6 h5f4');
INSERT INTO evaluations VALUES(20637,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(20637,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(20637,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(20637,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(20637,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(20637,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(20637,36,'cp 176','d4d5 c6e7 b1c3 e7g6 h2h4 f8e7 f1b5 c8d7 b5d7 d8d7 h4h5 g6f8 a2a4 d7d8 h5h6 g7g6 a4a5 g8f6 a5a6 d8c8 a6b7 c8b7 f3d2 f8d7 d2c4 d7c5 c4a5 b7b4 e1g1 c5e4 a1a4 e4c3 b2c3 b4b5 c3c4 b5d7 f1e1 e8g8 a5c6 f8c8 c6e7 d7e7 c1g5 a7a6');
INSERT INTO evaluations VALUES(20630,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(20630,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(20630,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(20630,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(20630,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(20630,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(20630,36,'cp 157','c6b8 b1c3 g8f6 h2h3 f8e7 c1e3 e8g8 f1d3 a7a5 e1g1 b8a6 f3d2 f6d7 d2c4 b7b6 c3b5 d7c5 a2a4 a6c7 c4b6 c7b5 a4b5 d8b6 b2b4 c8d7 b4c5 d6c5 d1b3 f8b8 f1b1 h7h6 d3f1 b8b7 g1h1 a8b8 b3a3 f7f5 a3a5 f5f4 e3d2 b6a5');
INSERT INTO evaluations VALUES(23376,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(23376,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(23376,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(23376,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(23376,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(23376,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(23376,36,'cp -36','e7g6 h2h4 g6f4 g2g3 f4g6 h4h5 g6e7 c1g5 g8f6 g5f6 g7f6 d1f3 d7d6 f3f6 h8g8 b1c3 f8g7 f6g5 c7c6 d5c6 b7c6 e1c1 d6d5 e4d5 c8g4 f1e2 e7d5 g5d2 g4e6 h5h6 g7h8 e2b5 c6b5 c3d5 g8g3 g1e2 g3g2 d2e3 e6d5 e2c3 d8g5 e3g5 g2g5 d1d5 b5b4 c3b5 a8c8 h1d1');
INSERT INTO evaluations VALUES(38653,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(38653,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(38653,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(38653,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38653,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38653,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(38653,36,'cp 16','f4d6 d8d6 g1f3 g8f6 e2e3 e8g8 b1c3 d5c4 f1c4 c7c5 e1g1 b8c6 c3b5 d6e7 d4c5 e7c5 a1c1 f8d8 c4e2 c5e7 d1c2 a7a6 b5c3 c8d7 h2h3 a8c8 f1d1 h7h6 c2b1 d7e8 d1d8 e7d8 b2b4 c6e7 a2a3 d8b6 b1b2 c8c7');
INSERT INTO evaluations VALUES(38601,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(38601,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38601,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(38601,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(38601,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38601,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(38601,36,'cp 15','d8d6 g1f3 g8f6 e2e3 e8g8 f1e2 d6e7 b1d2 b7b6 a1c1 b8d7 c4d5 f6d5 e1g1 c7c5 d4c5 d7c5 a2a3 f8d8 b2b4 c5d7 f3d4 c8b7 d2c4 a8c8 d1d2 d7f6 e2f3 h7h6 d2b2 f6d7 f1d1 b7a6 b4b5 a6b7 a3a4 a7a5 b5a6 b7a6');
INSERT INTO evaluations VALUES(33100,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(33100,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(33100,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(33100,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(33100,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33100,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(33100,36,'cp 16','c4d5 e6d5 e2e3 g8f6 d1c2 e8g8 b1c3 d6e7 g1f3 f8d8 a1c1 c8g4 f3e5 g4h5 f1e2 h5e2 c2e2 c7c6 e1g1 a7a5 f2f3 b8d7 e5d3 d7f8 f1e1 f8e6 h2h3 f6e8 e2f2 g7g6 c3a4 e8g7 c1c3 a8c8 a2a3 c8a8 a4b6');
INSERT INTO evaluations VALUES(42053,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42053,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(42053,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42053,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42053,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42053,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(42053,36,'cp 0','d5c4 d3c4 c7c5 g1f3 c5d4 e3d4 a7a6 c4d3 e8g8 e1g1 b7b5 a2a4 b5a4 a1a4 a6a5 f3e5 c8b7 c1g5 b8d7 e5d7 d8d7 g5f6 g7f6 d1h5 f6f5 h5g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(43878,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43878,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43878,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43878,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43878,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(43878,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(43878,36,'cp 0','c8b7 f1d3 e7e6 e1g1 d7d5 c2c4 f8d6 b2b3 e8g8 c4d5 f6d5 b1d2 b8d7 d2c4 d5c3 d1c2 b7f3 g2f3 d8g5 g1h1 g5h5 f3f4 h5f3 h1g1 f7f5 c4e5 d6e5 d4e5 d7c5 c2c3 c5d3 c3d3 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(44962,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(44962,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(44962,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(44962,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(44962,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(44962,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(44962,36,'cp -94','d5e4 f3e5 b8d7 f1e2 g7g6 e1g1 f8g7 e2c4 d7e5 d4e5 d8d1 f1d1 f6g4 e5e6 f7f5 a2a4 g4e5 c4b3 b7b6 h2h3 g7f6 c1f4 e8g8 a4a5 b6b5 f4h6 f8e8 d1d7 e5c4 d7c7 c8e6 c7c6 f6c3 b2c3 g8f7 h6e3 e8d8 b3c4 e6c4 c6c7 f7e6 e3a7 d8d2 a5a6 d2c2');
INSERT INTO evaluations VALUES(45042,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(45042,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(45042,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(45042,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(45042,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(45042,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(45042,36,'cp -112','f3e5 b8d7 c1e3 d7b6 g2g3 g7g6 f1g2 c8f5 e1g1 f8g7 d1e2 h7h5 a1d1 b6d5 c3d5 d8d5 c2c4 d5e6 e3f4 a8d8 f1e1 h5h4 e2e3 h4h3 g2h1 e8g8 b2b3 f6g4 e5g4 f5g4 d1d2 e6d7 e3e4 g4f5 e4f3 g7d4');
INSERT INTO evaluations VALUES(46232,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(46232,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(46232,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(46232,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(46232,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(46232,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(46232,36,'cp -8','b8d7 f1e2 b7b6 b2b3 c8b7 e1g1 f8e7 c3c4 e8g8 c1b2 c5d4 f3d4 a8c8 c4d5 f6d5 a1c1 c8c1 d1c1 d8a8 c1a1 f8c8 f1c1 d5f6 d4f3 b7d5 c1c8 a8c8 b2d4 h7h6 h2h3 c8a8 a1b2 a7a6 a2a3 b6b5 b3b4 a8b7 e2d1 d7b6 b2a1 b6c4');
INSERT INTO evaluations VALUES(47225,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(47225,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(47225,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(47225,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(47225,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(47225,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(47225,36,'cp -43','d5e4 c3e4 e7e5 e4f6 d8f6 d3e4 b8c6 d2d3 c8f5 g1f3 f5e4 d3e4 f6g6 e1g1 g6e4 f3e5 c6e5 f1e1 e4f5 d1d3 f5d3 e1e5 e8d7 c2d3 f8d6 e5f5 f7f6 c1e3 a8e8 a1c1 h7h5 h2h3 e8e6 f5d5 a7a6 g1f1 c7c6 d5d4 h8e8 b2b4 e6e5 a2a4 g7g5 h3h4 g5g4 g2g3 e5d5 d4d5');
INSERT INTO evaluations VALUES(47801,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(47801,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(47801,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(47801,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(47801,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(47801,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(47801,36,'cp -50','f6e4 d2d4 d6d5 d1e2 c8g4 e1g1 f8e7 d4e5 e4f2 c4b3 f2e4 e2b5 c7c6 b5b7 d8b6 b7b6 a7b6 f3d4 e4c5 b3c2 c5e6 h2h3 g4h5 c1e3 e6d4 c3d4 b8a6 b1c3 h5g6 c2g6 h7g6 a2a3 a6c7 a1c1 f7f6 c3e2 e8d7 f1f3 f6f5 b2b4 c7e6 e2f4');
INSERT INTO evaluations VALUES(48001,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(48001,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(48001,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(48001,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(48001,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(48001,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(48001,36,'cp -37','d2d4 d6d5 d1e2 c8g4 e1g1 e5d4 c4b3 f8e7 e2b5 b8d7 b5d5 e4d6 c1f4 e8g8 f1e1 a7a5 b3c2 f8e8 d5d4 g4f3 g2f3 e7f6 e1e8 d8e8 d4d3 e8e1 d3f1 e1e6 b1d2 d7e5 f1e2 a8e8 a1d1 e6a2 d2e4 a2c4 e2c4 d6c4 b2b3 e5f3 g1g2 c4e5 e4f6 g7f6 c2e4 f3h4 g2g3');
INSERT INTO evaluations VALUES(51754,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(51754,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(51754,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(51754,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(51754,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(51754,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(51754,36,'cp 104','f8b4 f1d3 d7d5 e1g1 g8e7 e4e5 c6c5 d4c5 b8d7 c3b5 b4c5 c2c3 a7a6 b5d4 c5b6 d3c2 d7c5 f1e1 c8d7 a2a4 e8g8 a4a5 b6a7 c1f4 c5e4 c2e4 d5e4 e1e4 d8c7 d1e2 a7d4 f3d4 a8c8 f4e3 f8d8 h2h4 d7c6 d4c6 c7c6 f2f3 d8d5 h4h5 e7f5 e3b6');
INSERT INTO evaluations VALUES(37639,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(37639,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(37639,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(37639,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(37639,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(37639,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(37639,36,'cp 141','f1d3 d7d5 e1g1 g8e7 f1e1 e8g8 c1d2 b8d7 d1c1 c6c5 e4d5 c5d4 d5e6 d4c3 b2c3 b4a5 e6d7 c8d7 c1b2 d7c6 f3e5 c6d5 a1d1 d8c7 c3c4 a5d2 d1d2 d5e6 b2d4 f8d8 d4e4 e6f5 e4f3 f7f6 e5g4 f5d3 c2d3 e7c6 g4e3 c6d4 f3h5 c7d6 h2h4');
INSERT INTO evaluations VALUES(56520,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(56520,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(56520,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(56520,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(56520,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(56520,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(56520,36,'cp 162','f8d6 d2d4 d6e5 d4e5 d8d1 c3d1 b8c6 f2f4 h7h5 f1d3 g8h6 c2c3 c8e6 d1f2 e8c8 e1g1 h6g4 d3c2 g4f2 g1f2 c6e7 c1e3 g7g6 f1e1 b7b6 a2a3 a7a5 a1d1 e6g4 d1d8 h8d8 c2a4 g4d7 a4d7 d8d7 f2e2 a5a4 g2g3 d7d8 e3f2 e7f5');
INSERT INTO evaluations VALUES(5787,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(5787,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(5787,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(5787,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(5787,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(5787,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(5787,36,'cp -83','b7c6 c2c4 g8e7 c4c5 d6c5 b2e5 e8g8 g1f3 e7g6 e1g1 c8g4 e5g3 f7f5 d1c2 c5d6 c2c6 g6e7 c6c2 g4f3 g2f3 f5f4 e3f4 e7g6 b1c3 g6f4 a1e1 d8g5 c3b5 f4h3 g1g2 g5h5 b5d4 a8e8 e1e8 f8e8 c2c6 h3f4 g3f4 d6f4');
INSERT INTO evaluations VALUES(4947,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(4947,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(4947,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(4947,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(4947,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(4947,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(4947,36,'cp -62','g1f3 f7f6 c2c4 g8h6 d2d3 e8g8 b1c3 c8e6 d1c2 d8d7 e1c1 h6f7 h2h3 f7d8 c1b1 e6f7 c3a4 a7a5 b1a1 f8e8 c4c5 d6f8 d3d4 f7g6 c2d2 g6e4 d4e5 e4f3 g2f3 f6e5 d1g1 d7f7 h3h4 f7f3 h4h5 d8f7 d2c2');
INSERT INTO evaluations VALUES(4948,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(4948,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(4948,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(4948,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(4948,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(4948,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(4948,36,'cp -78','f7f6 c2c4 g8h6 b1c3 c8e6 d1c2 e8g8 d2d3 d8d7 c3a4 f8e8 e1c1 h6f7 h2h3 f7d8 c1b1 e6f7 b1a1 a7a5 c4c5 d6e7 d1c1 h7h6 d3d4 e5e4 f3g1 f6f5 g1e2 d8e6 g2g3 f7h5 e2f4 h5f3 h1g1 g7g5');
INSERT INTO evaluations VALUES(8036,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(8036,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(8036,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(8036,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(8036,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(8036,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(8036,36,'cp 50','d4c6 b7c6 e4e5 f6g4 d1g4 d7d5 g4e2 d5c4 f2f4 d8b6 a1b1 c6c5 e1g1 b6c6 c1d2 e8g8 b1e1 c8a6 e2f3 c6f3 f1f3 a6b7 f3f2 a8b8 c3d1 f7f6 d2c3 b8d8 e5f6 e7f6 f2e2 g7h6 f4f5 g6f5 d1e3 h6e3 e2e3 d8d7 g1f2 b7d5 e3e7 d7e7');
INSERT INTO evaluations VALUES(10721,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(10721,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(10721,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(10721,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(10721,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(10721,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(10721,36,'cp 102','c6e7 a2a3 e8g8 c1g5 f6e8 d1d2 a7a6 h2h4 b7b5 h4h5 c7c6 h5h6 g7g6 d5c6 d7c6 a3a4 c8e6 g5e7 d8e7 a4b5 c6b5 f1b5 e8c7 b5c6 a8b8 e1g1 b8b2 c3d5 e6d5 c6d5 b2b8 d5b3 d6b4 d2e3 g8h8 a1d1 a6a5 g2g3 f7f6');
INSERT INTO evaluations VALUES(11044,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(11044,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(11044,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(11044,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(11044,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(11044,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(11044,36,'cp 104','a2a3 h7h6 c1e3 b7b6 d1d3 f6g4 c3b5 e8g8 b5d6 c7d6 f1e2 a7a5 e1g1 g4e3 f2e3 a5a4 c2c4 d8c7 a1c1 c7c5 c1c3 c8a6 h2h3 a8b8 d3d2 c5a5 f1c1 g7g6 b2b4 a4b3 c3b3 a5c5 d2b4 c5b4 a3b4 b8a8 c1f1 g8g7 b3a3 a6b7 f1a1 e7g8 a3a8 f8a8 a1a8');
INSERT INTO evaluations VALUES(14433,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(14433,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(14433,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14433,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14433,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(14433,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(14433,36,'cp 52','f4g3 h5g3 h2g3 g7g6 c2c4 d5c4 f1c4 f8g7 d1b3 e7e6 c4b5 c8d7 e1g1 e8g8 b1c3 d8e7 f1d1 f8d8 a1c1 c6a5 b3c2 c7c6 b5e2 c6c5 d4d5 a8c8 c2d2 e7f8 b2b3 e6d5 c3d5 a5c6 f3g5 h7h5 g5e4 d7f5 e2f3 f5e4 f3e4 c8b8');
INSERT INTO evaluations VALUES(14434,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(14434,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14434,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(14434,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(14434,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(14434,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(14434,36,'cp 46','h5g3 h2g3 g7g6 c2c4 d5c4 f1c4 f8g7 d1b3 e7e6 c4b5 c8d7 e1g1 e8g8 b1c3 d8e7 f1d1 f8d8 a1c1 d7e8 a2a3 a7a6 b5d3 e6e5 d4e5 c6e5 f3e5 e7e5 b3b7 c7c6 b7b4 a8b8 b4a4 b8b2 c3e4 e5e7 a4a6 b2a2 a6c4');
INSERT INTO evaluations VALUES(14436,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(14436,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(14436,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(14436,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(14436,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14436,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(14436,36,'cp 58','h2g3 g7g6 c2c4 d5c4 f1c4 f8g7 d1b3 e7e6 c4b5 c8d7 e1g1 e8g8 b1c3 c6a5 b3c2 d7b5 c3b5 c7c6 b5c3 a5c4 c3e4 c4d6 e4c5 h7h5 a1c1 d8c7 c2b3 c7e7 a2a4 b7b6 c5d3 f8c8 b3a3 c8c7 c1c2 d6c8');
INSERT INTO evaluations VALUES(14435,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(14435,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(14435,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(14435,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(14435,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14435,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(14435,36,'cp 38','g7g6 c2c4 d5c4 f1c4 f8g7 d1b3 e7e6 c4b5 c8d7 b1c3 e8g8 e1g1 c6a5 b3a3 d7b5 c3b5 a5c4 a3c5 c4d6 a1c1 h7h5 f1d1 a7a6 b5c3 f8e8 c3a4 c7c6 c5c2 d8c7 a4c5 a6a5 f3e5 e8d8 c2b3 a5a4 c5a4 c7a5 a4b6 a5a2 b3c2 a8a5');
INSERT INTO evaluations VALUES(16350,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(16350,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(16350,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(16350,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(16350,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(16350,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(16350,36,'cp -102','c6e5 d1e2 d8e7 c1f4 d7d6 b1c3 e5c6 d4c6 b7c6 e1c1 e7e2 f1e2 c8e6 e2f3 e8d7 h2h3 a7a5 g2g4 h7h6 b2b3 a8e8 f4e3 d6d5 c1b2 d7c8 c3a4 e6d7 c2c4 f8d6 e3d4 h6h5 c4d5 c6d5 d4f6 g7f6 d1d5 h5g4 h3g4 h8h1 f3h1');
INSERT INTO evaluations VALUES(16953,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(16953,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(16953,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(16953,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(16953,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(16953,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(16953,36,'cp -88','d1e2 d8e7 c1f4 d7d6 b1c3 e5c6 d4c6 b7c6 e1c1 e7e2 f1e2 c8e6 e2f3 e8d7 h1e1 a7a5 h2h3 f8e7 a2a3 h7h6 f4e5 h8e8 c3a4 e6d5 f3d5 c6d5 e5f6 e7f6 e1e8 a8e8 d1d5 e8e1 c1d2 e1g1 g2g3 c7c5 d2e2 d7c6 d5d1 g1d1 e2d1 d6d5 f2f3 c5c4 b2b3');
INSERT INTO evaluations VALUES(21450,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21450,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21450,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21450,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(21450,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(21450,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21450,36,'cp 0','b1d2 g8f6 a2a3 a7a5 e4e5 f6d7 d3d4 f7f6 e5f6 d8f6 d2b1 f8d6 b1c3 h7h6 f1e2 e8g8 e1g1 a5a4 c3b5 d7b6 c1d2 c8d7 d2c3 d7e8 f3e5 c6e5 b5d6 c7d6 d4e5 d6e5 f2f4 b6d7 f4e5 f6g5 d1d4 f8f1 a1f1 e8g6 e2b5 g5e7 b5a4 d7c5');
INSERT INTO evaluations VALUES(21575,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(21575,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(21575,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(21575,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(21575,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(21575,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(21575,36,'cp 112','c6d4 h2h3 f8d6 f3d4 e5d4 c3e4 g8f6 f1b5 c8d7 e4d6 c7d6 b5d7 d8d7 e1g1 f6d5 d1f3 d5e7 b2b3 e8g8 c1b2 e7c6 f3f4 d7c7 f1c1 c7b6 f4d6 a8d8 d6f4 h7h6 f4f5 c6e7 f5f3 e7c6 c1e1 b6a5 c2c3 c6e5 f3g3 e5g6 c3d4 a5d2 b2c3 d2f4 g3f4 g6f4 c3b4 f8e8 e1e8 d8e8 b4d2 f4d3 g1f1');
INSERT INTO evaluations VALUES(22171,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(22171,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(22171,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(22171,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(22171,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(22171,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(22171,36,'cp 117','h2h3 f8d6 c3b5 d4f3 d1f3 g8e7 b2b3 e8g8 c1b2 d6c5 f3c3 c5d6 b5d6 c7d6 f1c4 c8f5 e1g1 b7b5 c4b5 a8c8 c3f3 f5c2 b5c4 c2g6 f1e1 e7f5 a1c1 d8b6 d2d3 f8e8 f3g4 f5d4');
INSERT INTO evaluations VALUES(22290,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(22290,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(22290,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(22290,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(22290,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(22290,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(22290,36,'cp 178','e1g1 a7a6 b5c4 h7h5 c1e3 g8e7 d4e5 d6e5 d1e2 e7g6 b1c3 f8d6 h2h3 c6a5 c4d3 c8e6 c3d5 e8g8 f1d1 b7b5 a2a4 a5c4 d3c4 b5c4 e2c4 g6e7 c4e2 e6d5 e4d5 e7f5 e3d2 d8e8 e2e4 f5e7 b2b4 a8b8 e4e2 d6b4 d2b4 b8b4 e2a6 e5e4');
INSERT INTO evaluations VALUES(30937,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(30937,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(30937,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(30937,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(30937,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(30937,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(30937,36,'cp -60','e7e6 g1f3 d7b5 e2b5 d8d7 b5e2 b8c6 d2d4 f8d6 e1g1 g8e7 f1e1 e8g8 b1d2 f8e8 d2f1 e7g6 c1e3 h7h6 e3d2 a8c8 a1d1 d6b8 c2c3 f7f6 h2h4 e6e5 d4e5 f6e5 c3c4 e5e4 c4d5 e4f3 d5c6 d7c6 e2f3 c6f3 e1e8 c8e8 g2f3 g6h4');
INSERT INTO evaluations VALUES(32242,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(32242,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(32242,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(32242,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(32242,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(32242,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(32242,36,'cp 25','g1e2 g8f6 e2f4 d8d7 f4e6 d7e6 f1e2 f8b4 e1g1 e8g8 c3b5 b4a5 c1f4 f8e8 e2d3 e6c8 d1f3 a7a6 b5c3 a5c3 b2c3 c7c5 f4e5 b8d7 a1e1 c8c6 c3c4 b7b5 e5f6 e8e1 f1e1 b5c4 d3c4 d7f6 d4c5 c6c5 c4d3 c5d6 c2c4');
INSERT INTO evaluations VALUES(34870,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(34870,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(34870,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(34870,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(34870,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(34870,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(34870,36,'cp 50','b8c6 f1c4 e6g6 c3b5 f8d6 d2d4 c8g4 e1g1 e8c8 h2h3 h7h5 d4d5 d6c5 c1e3 a7a6 b5d6 c7d6 d5c6 g4h3 f3h4 g6f6 c6b7 c8b8 d1f3 f6h4 f3h3 h4h3 g2h3 f7f5 e3c5 d6c5 a1e1 e5e4 f2f3 g8f6 f3e4 f5e4 c4a6 d8d5 a6c4 d5d2 e1e2 h8d8 a2a4 d2e2');
INSERT INTO evaluations VALUES(36122,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(36122,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(36122,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(36122,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36122,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36122,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36122,36,'cp 13','h2h3 f8e8 f1e1 e8e1 d1e1 h7h6 c1e3 d8c7 e1c1 c8e6 c3e2 b8d7 b2b3 c6c5 d4c5 d7c5 e3f4 a7a6 f4d6 c7d6 c1f4 d6f4 e2f4 c5d3 f4d3 e6f5 d3e1 g7g5 f3d4 f5h7 a2a4 g8g7 f2f3 f6h5 d4e2 a8e8 g1f2');
INSERT INTO evaluations VALUES(36121,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(36121,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36121,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36121,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(36121,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36121,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(36121,36,'cp 19','f8e8 f1e1 e8e1 d1e1 b8d7 c3e2 d8c7 e2g3 d7f8 c1g5 d6g3 g5f6 g3d6 f6e5 d6e5 f3e5 c7b6 e1c3 f7f6 e5f3 c8d7 a1b1 d7e8 b1e1 e8f7 a2a4 a7a5 f3d2 b6c7 d2b3 b7b6 b3d2 a8e8 e1e8 f7e8 d2f1 g7g6 f1e3 f8e6 b2b3 g8g7 g2g3 e8f7 h3h4 c7d6');
INSERT INTO evaluations VALUES(43415,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43415,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43415,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43415,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43415,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43415,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43415,36,'cp 0','e7e5 d4b3 c8e6 f1d3 f8e7 e1g1 e8g8 d1f3 e6g4 f3g3 g4e6');
INSERT INTO evaluations VALUES(43456,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43456,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43456,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43456,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43456,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43456,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43456,36,'cp 0','d4b3 c8e6 f1d3 f8e7 d1f3 e8g8 e1g1 e6g4 f3g3 g4e6');
INSERT INTO evaluations VALUES(43609,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(43609,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(43609,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(43609,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(43609,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(43609,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(43609,36,'cp 71','d2d4 b8d7 a2a4 f8e7 c1e3 e8g8 h2h3 e5d4 e3d4 d7b8 e4e5 d6e5 f3e5 c8f5 f1e1 b8c6 e5c6 b7c6 e1e2 e7b4 d4e5 d8d1 c3d1 f8e8 e5f6 g7f6 e2e8 a8e8 d1e3 f5c8 g2g4 a6a5 a1d1 h6h5 g1g2 g8g7 c4d3 h5h4 e3c4 c8e6 b2b3 e6c4 d3c4');
INSERT INTO evaluations VALUES(43617,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43617,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(43617,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(43617,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(43617,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(43617,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(43617,36,'cp 65','b8d7 a2a4 f8e7 h2h3 e8g8 c1e3 e5d4 e3d4 f8e8 a4a5 d7f8 d1d3 c8e6 c3d5 f6d7 f1d1 f8g6 d3b3 d8c8 c4f1 e6d5 e4d5 e7f6 d4f6 d7f6 c2c4 a8b8 f3d4 c8d7 b3c2 c7c5 d4f5 b8d8 d1d2 f6e4');
INSERT INTO evaluations VALUES(46291,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(46291,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(46291,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(46291,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(46291,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(46291,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(46291,36,'cp -2','e6e5 g5f6 g7f6 d4c2 d5d4 c3d5 b8c6 e2e3 c8e6 e3d4 d8a5 d1d2 a5d2 e1d2 e8c8 a1d1 b7b5 b2b3 e6d5 c4d5 d8d5 f1d3 e5d4 d3e4 f8h6 d2d3 d5c5 e4c6 c5c3 d3d4 c3c2 c6b5 c2f2 d4c3 c8b8 d1d7 h6f4 h1d1 f2a2 d7f7 f4e5 c3b4 a7a5 b4c5 h8c8 c5b6 e5c7 b6c6 a2g2 d1d7 c7e5 c6b6');
INSERT INTO evaluations VALUES(49089,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(49089,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(49089,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(49089,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(49089,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(49089,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(49089,36,'cp 135','e4e5 c8e6 b1c3 b8d7 e5g3 d7f6 d2d4 c7c6 f1e2 f8e7 e1g1 e8g8 f1d1 e6f5 e2d3 e7d6 g3h4 d8d7 c1g5 d6e7 h4f4 f5g6 d3g6 f7g6 f4d2 a8e8 d1e1 h7h6 g5f4 e7d6 f4g3 g8h7 e1e8 f8e8 d2d3 d6b4 h2h3 d7f5 d3f5 g6f5 g3e5 b4c3 b2c3 f6e4 c3c4');
INSERT INTO evaluations VALUES(49085,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(49085,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(49085,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(49085,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(49085,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(49085,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(49085,36,'cp 112','c8e6 b1c3 b8d7 e5g3 d7f6 d2d4 c7c6 f1e2 f8e7 e1g1 e8g8 f1d1 e6f5 e2d3 f5g6 h2h3 d8d7 g3f3 g6h5 g2g4 h5g6 c1f4 a8e8 f4e5 h7h6 g1g2 f6d5 c3d5 c6d5 d1e1 e7h4 e1e2 g6d3 f3d3 e8e6 a1e1 f8e8 c2c3 d7c6 e5g3 e6e2 e1e2 h4g3 e2e8 c6e8 g2g3');
INSERT INTO evaluations VALUES(50073,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(50073,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(50073,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(50073,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(50073,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(50073,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(50073,36,'cp 144','d7d5 f1b5 c7c6 b5d3 b8a6 d3a6 b7a6 e1g1 d8c7 h2h3 c8f5 f1e1 e7e6 f3h4 f8e7 h4f5 e6f5 d1e2 e8f8 d2d4 e7d6 e2f3 g7g6 c1h6 f8f7 c2c4 a8e8 b1d2 c7b8 f3d3 e8e1 a1e1 h8e8 e1e8 f6e8 b2b3 d6h2 g1h1 h2f4 h6f4 b8f4 d2f3 e8c7 d3c3');
INSERT INTO evaluations VALUES(46587,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(46587,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(46587,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(46587,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(46587,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(46587,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(46587,36,'cp 144','f1b5 c7c6 b5d3 b8a6 d3a6 b7a6 e1g1 c8f5 f1e1 d8c7 h2h3 e7e6 f3h4 f8e7 h4f5 e6f5 d1e2 e8f8 d2d4 e7d6 e2a6 f8f7 b1d2 f6e4 d2f3 h8b8 b2b3 f7g8 c1a3 d6f4 a6d3 a7a5 a3c1 f4d6 a2a4 b8e8 c2c4 h7h6 c1a3 d6f4 b3b4 c7f7 c4d5 f7d5');
INSERT INTO evaluations VALUES(50692,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(50692,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(50692,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(50692,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(50692,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(50692,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(50692,36,'cp 118','c3d4 g8f6 b1c3 e7e6 f1d3 f8e7 e1g1 b8c6 a2a3 b7b5 c1f4 c6a5 d1c2 c8b7 b2b4 a5c4 a3a4 e8g8 a4b5 a6b5 c3b5 d6d5 a1a8 b7a8 e4e5 f6e4 f1a1 d8b6 b5d6 c4d6 e5d6 e7d6 f4d6 b6d6 b4b5 a8b7 f3e5 f8c8');
INSERT INTO evaluations VALUES(52390,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(52390,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(52390,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(52390,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(52390,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(52390,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52390,36,'cp 0','g8f6 e2e3 d7d5 c2c4 e7e6 b2b3 f8e7 c1b2 e8g8 b1c3 c6c5 f3e5 f6d7 e5f3');
INSERT INTO evaluations VALUES(54695,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(54695,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(54695,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(54695,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(54695,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(54695,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(54695,36,'cp 114','e4f5 c8f5 d2d4 g8f6 f1b5 c7c6 b5d3 d8d7 e1g1 g7g6 f1e1 b8a6 d3f5 g6f5 d1e2 e8c8 c2c4 d8e8 b1c3 f8g7 c1f4 h8g8 a1d1 e7e6 a2a3 f6h5 f4c1 g7f6 b2b4 a6c7 g2g3 d7f7 b4b5 c6c5 d4c5 f6c3 c5d6 c3e1 d1e1');
INSERT INTO evaluations VALUES(56325,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(56325,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(56325,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(56325,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(56325,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(56325,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(56325,36,'cp 147','h5e5 g8e7 e4d5 b8d7 e5e3 d7f6 f1b5 c8d7 b5d7 d8d7 c2c4 d7g4 d2d3 g4g2 e3f3 g2g6 b1c3 e8c8 c1f4 e7f5 e1c1 f8c5 f4e3 f5e3 f2e3 g6h6 d3d4 c5b4 g1e2 h8e8 d1d3 h6g6 h1d1 c8b8 e2g3 h7h5 h2h3 b4c3 d3c3 f6e4 g3e4');
INSERT INTO evaluations VALUES(56318,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(56318,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(56318,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(56318,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(56318,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(56318,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(56318,36,'cp 133','g8e7 e4d5 b8d7 e5e3 d7f6 f1b5 c8d7 b5d7 d8d7 b1c3 g7g6 g1f3 f8g7 d2d4 e8g8 e3d3 f6d5 c3d5 d7d5 e1g1 e7f5 f1d1 f8d8 c2c4 d5d6 d4d5 c7c6 c1g5 d8d7 d5c6 b7c6 d3e2 f5d4 f3d4 g7d4 a1c1 c6c5 e2e4 a8c8 b2b3 d6c6 e4c6 c8c6 d1d3 f7f6 g5d2');
INSERT INTO evaluations VALUES(46572,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(46572,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(46572,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(46572,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(46572,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(46572,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(46572,36,'cp 154','e4d5 b8d7 e5e3 d7f6 b1c3 f6d5 c3d5 d8d5 g1f3 c8e6 f1d3 e8c8 e1g1 c8b8 f1e1 d5d7 f3e5 d7c8 d3c4 e6c4 e5c4 e7f5 e3f3 f7f6 d2d3 h7h5 c1d2 f5d4 f3d1 g7g5 d2c3 c8d7 c3d4 d7d4 d1f3 f8c5 e1e4 d4d5 c4e3 d5c6 a1e1 c5e3 e4e3 c6f3');
INSERT INTO evaluations VALUES(14648,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(14648,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(14648,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(14648,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(14648,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(14648,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(14648,36,'cp 169','d4e5 d6e5 d1d8 e8d8 c4f7 f8d6 b1c3 d8e7 f7b3 c6a5 h2h3 g7g5 c1e3 b7b6 f3d2 h8d8 f1e1 d6b4 d2f3 a5b3 a2b3 b4c3 b2c3 e7e6 f3d2 a7a5 f2f3 c8d7 g1f2 a5a4 c3c4 d7c6 g2g3 f6d7 f2e2 a4a3 d2b1 a3a2 b1c3');
INSERT INTO evaluations VALUES(14606,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(14606,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(14606,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(14606,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(14606,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(14606,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(14606,36,'cp 176','d6e5 d1d8 e8d8 c4f7 c8g4 b1d2 f8d6 b2b3 d8d7 c1b2 g7g5 f7c4 a8e8 h2h3 g4h5 f3e1 a7a6 a2a4 d7c8 f2f3 d6c5 g1h1 h5g6 e1d3 c5a7 f1d1 c6d4 d2f1 f6h5 d1d2 h5f4');
INSERT INTO evaluations VALUES(14573,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(14573,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(14573,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(14573,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(14573,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(14573,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(14573,36,'cp 178','d1d8 e8d8 c4f7 f8d6 b1c3 d8e7 f7b3 h8d8 f3h4 c8e6 c1e3 e6b3 a2b3 e7f7 h4f5 d6f8 f2f3 c6d4 f1f2 d4f5 e4f5 f8b4 f2e2 a7a5 c3e4 b7b5 g1f2 a5a4 b3a4 a8a4 a1a4 f6e4 f3e4 b5a4 c2c3 b4d6 f2f3 d8b8 e2c2 b8b3 g2g4 a4a3 b2a3 b3a3 f3e2');
INSERT INTO evaluations VALUES(14760,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(14760,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(14760,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(14760,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(14760,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(14760,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14760,36,'cp 26','h7h6 g5h4 f8e7 a2a3 a7a6 f1d3 b7b6 e1g1 c8b7 a1c1 e8g8 c3e2 f6d7 h4e7 c6e7 c2c3 c7c5 d3b1 d8c7 d1c2 g7g6 h2h4 d7f6 b2b4 c5b4 c3b4 c7c2 c1c2 f8c8 f3e5 c8c2 b1c2 a8c8 f1c1 e7f5 c2f5 c8c1 e2c1 g6f5');
INSERT INTO evaluations VALUES(15574,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(15574,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(15574,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(15574,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(15574,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(15574,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(15574,36,'cp 37','g7g6 d2d4 f8g7 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 a2a4 c7c6 a4a5 c8e6 c3a2 g7f8 c1d2 f6d7 a2b4 d8c7 b4d3 f8d6 b2b4 a8c8 e1g1 c6c5 b4b5 c7b8 f1d1 h8e8 h2h4 c5c4 d3b4 d6c5 b4d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(17027,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(17027,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(17027,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(17027,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(17027,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(17027,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(17027,36,'cp 112','e1g1 d7d6 c2c3 c8g4 f3f4 c7c6 c3d4 d6d5 e4d5 c6d5 c4b5 g4d7 b5d7 d8d7 d2d3 f8e7 f4f3 e8g8 b1c3 h7h6 h2h3 f8e8 c1f4 a7a5 f1e1 a8a6 a2a3 b7b5 e1e5 e7d6 e5e8 d7e8 f4d6 a6d6 g2g3 e8d7 g1g2 d6e6 g3g4 b5b4 a3b4 a5b4 c3e2 e6c6 f3g3 c6c2 a1a8 g8h7');
INSERT INTO evaluations VALUES(20606,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20606,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20606,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20606,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20606,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20606,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(20606,36,'cp 0','g2g3 g7g6 b1c3 g8e7 a1b1 f8g7 f1g2 e8g8 e1g1 c8e6 c1d2 a8b8 a2a3 a7a5 h2h3 h7h6 g1h2 b7b6 f3h4 c6d4 h4f3');
INSERT INTO evaluations VALUES(21303,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(21303,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(21303,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(21303,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(21303,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(21303,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(21303,36,'cp 103','g8f6 d4c6 b7c6 d1e2 f8e7 c1e3 f6g4 e3d4 e8g8 e1c1 d8e8 f2f3 g4e5 d4e5 d6e5 c3a4 e7d6 c1b1 a7a5 d1d3 c8e6 d3b3 e8e7 h1d1 f8d8 a2a3 e7e8 c4e6 e8e6 b3d3 d8b8 b1a1 d6e7 e2d2 b8d8 d2c3 d8d4 a4c5 e7c5 c3c5');
INSERT INTO evaluations VALUES(14657,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(14657,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(14657,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(14657,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(14657,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(14657,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(14657,36,'cp 72','d4c6 b7c6 d1e2 f8e7 c1e3 e8g8 e1c1 f8e8 c4b3 a7a5 e2c4 e8f8 f2f3 d8e8 b3a4 c8d7 h2h4 c6c5 g2g4 d7a4 c3a4 a8b8 a4c3 f6d7 g4g5 d7b6 c4d3 b6a4 c3a4 e8a4 c1b1 a4b4 e3c1 c5c4 d3d4 b8b6 c2c3 b4b5 b1a1 f8b8 g5h6 e7f6');
INSERT INTO evaluations VALUES(14541,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(14541,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(14541,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(14541,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(14541,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(14541,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(14541,36,'cp 74','b7c6 d1e2 f8e7 c1e3 e8g8 e1c1 f8e8 c4b3 a7a5 e2c4 e8f8 f2f3 d8e8 b3a4 c8d7 h1e1 c6c5 e3f4 a8b8 e4e5 b8b4 c4d3 f6h5 a4d7 h5f4 d7e8 f4d3 d1d3 f8e8 e5d6 c7d6 c1b1 g8f8 d3e3 b4b8 c3d5 e7g5 f3f4 e8e3 e1e3 g5f6 e3b3 b8b3 a2b3 f6d4 c2c3 d4f2 b1c2 h6h5 f4f5');
INSERT INTO evaluations VALUES(11814,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(11814,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(11814,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(11814,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(11814,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(11814,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(11814,36,'cp 70','d1e2 f8e7 c1e3 e8g8 e1c1 f8e8 h2h3 e7f8 h1e1 d8e7 f2f3 c8e6 c4b3 e6b3 c2b3 e7e6 e2d3 a7a5 g2g4 d6d5 e3d4 d5e4 c3e4 f6e4 e1e4 e6g6 e4e8 a8e8 d3g6 f7g6 d4c3 f8b4 d1d7 b4c3 b2c3 e8e3 c3c4 e3f3 d7c7 f3h3 c7c6 h3h2 c6c8 g8f7');
INSERT INTO evaluations VALUES(29344,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(29344,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(29344,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(29344,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(29344,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(29344,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(29344,36,'cp 132','g2g4 h7h6 a2a3 b8a6 d2d3 a6c7 c1d2 g7g6 e1c1 f8g7 c1b1 d8e7 f3g3 c7e6 c4e6 f7e6 f2f4 e5f4 d2f4 e6e5 f4c1 e8c8 d1g1 f6d7 h3h4 d8g8 h4h5 g7f6 g3f2 c8b8 c1e3 b7b6 c3d1 f6g5 e3g5 e7g5 d1e3 d7c5 f2f3');
INSERT INTO evaluations VALUES(29567,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(29567,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(29567,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(29567,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(29567,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(29567,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(29567,36,'cp -46','g1f3 h7h6 f1g2 e7e6 e1g1 f8e7 c3b1 b8d7 c2c4 e8g8 d1b3 b7b6 c4d5 e6d5 b3a4 a7a5 a4c6 a8c8 c1f4 d7b8 c6b7 f8e8 f3e5 c7c6 f1c1 g7g5 f4e3 f6g4 e5c6 b8c6 c1c6 e7b4 c6c8 f5c8 b7d5 d8d5 g2d5 g4e3 f2e3 e8e3');
INSERT INTO evaluations VALUES(29565,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(29565,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(29565,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(29565,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(29565,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(29565,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(29565,36,'cp -61','h7h6 f1g2 e7e6 c3b1 b8d7 e1g1 f8e7 b2b3 c7c6 c2c4 a7a5 b1c3 e8g8 c1b2 d8b8 a2a4 f8e8 e2e3 b8d8 f3d2 e7b4 c3a2 b4f8 a1c1 f5d3 f1e1 d3h7 g2f1 d8b6 d2f3 a8d8 a2c3 f8b4 f1d3 h7d3 d1d3');
INSERT INTO evaluations VALUES(38852,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(38852,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(38852,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(38852,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(38852,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(38852,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(38852,36,'cp 158','b2c3 d8d7 d1g4 f7f6 g1f3 b7b6 a2a4 b8c6 f1d3 c8b7 e5f6 g8f6 g4h3 e8c8 e1g1 c8b8 c1f4 f6e4 f3e5 c6e5 f4e5 e4d2 f1e1 d2c4 h3g3 c4e5 g3e5 a7a5 e5e6 d7c6 e6c6 b7c6 g2g3 d8e8 e1e8 c6e8 f2f3 e8g6 a1e1 g6d3');
INSERT INTO evaluations VALUES(48096,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(48096,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(48096,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(48096,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(48096,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(48096,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(48096,36,'cp 28','d2d4 d5c3 b2c3 b8d7 f1d3 b7b6 d3b5 f8d6 f3e5 e8g8 d1f3 a8b8 e5c6 c8b7 c6d8 b7f3 d8e6 f7e6 g2f3 d7f6 c1d2 f6d5 c3c4 d5e7 e1e2 c7c6 b5a4 f8f5 d2e3 b8f8 h2h4 f5f3 h4h5 f3f6 a1g1 d6f4 e2d3 f4e3 f2e3 c6c5 c2c3 g8f7 e3e4 f6f3 d3d2 f3f4 d2e3');
INSERT INTO evaluations VALUES(48086,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(48086,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(48086,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(48086,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(48086,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(48086,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(48086,36,'cp 40','f8b4 c1d2 e8g8 f1d3 c7c5 e1g1 c5d4 c3d5 b4d2 d1d2 e6d5 d2f4 b8c6 f1e1 c8d7 f3d4 d8b6 d4c6 b7c6 b2b4 a7a5 f4d6 f8d8 b4a5 b6a5 a2a4 d7e8 d6e5 e8d7 e5d4 h7h6 e1e7 g8f8 e7e2 d8e8 e2e8 d7e8 d4f4 a5c5 a1e1');
INSERT INTO evaluations VALUES(50486,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(50486,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50486,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(50486,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50486,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50486,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(50486,36,'cp 7','d2d4 d7d5 e4d5 e6d5 c4b3 g8f6 g1f3 c5c4 b3c2 f8d6 d1e2 d8e7 e2e7 e8e7 e1g1 h8e8 f1e1 e7f8 e1e8 f8e8 c1g5 b8d7 b1d2 h7h6 g5h4 f6h5 d2f1 d7f6 f1e3 c8e6 h4f6 h5f6 f3e5 h6h5 f2f4 g7g6 f4f5 g6f5 e3f5');
INSERT INTO evaluations VALUES(50503,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50503,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50503,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(50503,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50503,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(50503,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50503,36,'cp 0','d7d5 e4d5 e6d5 c4b3 c5c4 b3c2 f8d6 g1f3 g8e7 e1g1 e8g8 c1g5 d8c7 f1e1 e7g6 b1d2 h7h6 c2g6 f7g6 g5h4 b8c6 h4g3 c8f5 g3d6 c7d6 f3e5 c6e5 e1e5 f5d3 d2f3 d3e4 f3d2');
INSERT INTO evaluations VALUES(52210,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(52210,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(52210,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(52210,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(52210,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(52210,35,'cp 235',NULL);
INSERT INTO evaluations VALUES(52210,36,'cp 219','a2a3 d7d6 f1c4 g7g6 c1e3 f8h6 d1d2 h6e3 d2e3 d8e7 b1d2 b8d7 h2h3 d7b6 c4b3 c8e6 a3a4 b6d7 a4a5 e6b3 e3b3 a8b8 b3e3 d6d5 d4e5 f6e5 e4d5 c6d5 e1g1 g8f6 e3a7 e8g8 f1e1 e5e4 a7d4 b8e8 a5a6 b7a6 a1a6 d7e5 f3e5 e7e5');
INSERT INTO evaluations VALUES(52684,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(52684,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(52684,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(52684,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(52684,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(52684,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(52684,36,'cp 45','e7e6 f1e1 f8e7 c2c3 g8f6 c4f1 b7b6 d2d4 c8b7 b1d2 c5d4 c3d4 b8d7 f1d3 a7a5 h2h3 e8g8 a2a3 e6e5 d3b5 e5d4 f3d4 f8e8 d2c4 e7f8 d4f5 f6e4 f2f3 e4f6 c1f4 b7a6 b5a6 a8a6 c4d6 e8e6 a1c1 d7c5 e1e6 c5e6 f4e5');
INSERT INTO evaluations VALUES(53842,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(53842,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(53842,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(53842,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(53842,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(53842,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(53842,36,'cp 99','c2c4 g8f6 b1c3 f8e7 f1e2 e8g8 c1f4 b8a6 d1c2 f8e8 a1d1 e7f8 e1g1 a6c5 f2f3 a7a5 f4e3 d8e7 e3g5 a5a4 f1e1 g7g6 c2d2 f8g7 d4c2 e7d8 g5e3 d8b6 d1b1 e8d8 e1d1 c8e6 b2b3 a4b3 a2b3 f6d7 g1h1');
INSERT INTO evaluations VALUES(55960,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(55960,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(55960,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(55960,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(55960,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(55960,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(55960,36,'cp 66','c7c5 c2c3 b8c6 f1d3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f8e7 b1c3 f5e3 f2e3 c8d7 g1h1 e8g8 e3e4 c6b4 e4d5 b4d5 c3d5 e6d5 d1b3 a8b8 b3d5 d7c6 d5b3 c6f3 f1f3 d8d4 e5e6 f7f5 f3f5 f8f5 d3f5 g7g6 f5h3 d4e5 g2g3 g8g7 h3g2 b8f8');
INSERT INTO evaluations VALUES(56016,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(56016,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(56016,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56016,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56016,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(56016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56016,36,'cp 0','e2e3 c7c5 g1f3 b8c6 d4c5 d8d1 e1d1 g8f6 f1c4 f8c5 b1c3 e8e7 d1e2 c5d6 f4d6 e7d6 a1c1 c8d7 c3b5 d6e7 h1d1 h8c8 b5d6 c8c7 d6b5');
INSERT INTO evaluations VALUES(56015,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(56015,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56015,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(56015,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56015,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(56015,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(56015,36,'cp -10','c7c5 g1f3 b8c6 d4c5 d8d1 e1d1 f8c5 f1c4 g8f6 b1d2 f6h5 f4c7 e8d7 c7e5 c6e5 f3e5 d7e7 d1e2 c5d6 d2f3 h5f6 h1d1 d6b8 a2a4 f6d7 e5d7 c8d7 c4b5 d7b5 a4b5 h8c8 d1c1 c8c1 a1c1 b8d6 f3d2 a7a5 b5a6 a8a6 h2h3 h7h6 g2g3 a6b6 b2b3 b6a6 d2e4 a6a3 e4d6');
INSERT INTO evaluations VALUES(56900,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56900,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56900,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56900,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(56900,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(56900,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56900,36,'cp 0','g8f6 d2d4 b8c6 c1g5 f8e7 g5f6 e7f6 d4d5 c6a5 c4d3 b7b6 b1c3 a5b7 d3a6 b7d6 a6c8 a8c8 f3e2 e8g8 g1f3 d8e7 h2h4 b6b5 h4h5 c7c6 e1c1 b5b4 c3a4 c6c5 c1b1 c5c4 c2c3 b4c3 a4c3 c8b8 h5g6 f7g6 b1a1 f6g7 f3d2 f8c8 a2a3');
INSERT INTO evaluations VALUES(49207,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(49207,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(49207,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(49207,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49207,36,'cp 0','d2d4 b8c6 c1g5 f8e7 g5f6 e7f6 d4d5 c6a5 c4e2 b7b6 b1c3 a7a6 h2h4 h7h5 d5d6 c7d6 c3d5 f6g7 e1c1 c8b7 g2g4 a5c6 g4h5 c6d4 f3g2 b7d5 e4d5 d4e2 g1e2 h8h5 e2g3 h5h4 g3f5 h4h1 f5g7 e8f8 g7e6 d7e6 d1h1 f8e7 g2g5 e7d7 d5e6 d7e6 g5g4 e6e7 g4g5');
INSERT INTO evaluations VALUES(5906,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5906,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(5906,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5906,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5906,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5906,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5906,36,'cp 0','f1b5 f7f6 e1g1 g8h6 c2c4 a7a6 c4d5 a6b5 d5c6 e8g8 d2d3 b7c6 d1c2 c6c5 a2a4 c8b7 a4b5 b7f3 g2f3 a8a1 b2a1 d8e8 c2c4 g8h8 b1d2 e8h5 g1g2 e5e4 f3f4 h5g4 g2h1 g4e2 a1c3 h6g4 h1g1 g4h2 g1h2 e2h5 h2g1 h5g4 g1h1 g4h3 h1g1');
INSERT INTO evaluations VALUES(7753,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(7753,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(7753,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(7753,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(7753,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(7753,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(7753,36,'cp 51','e3e4 d5e4 c3e4 f6e4 d3e4 e6e5 e1g1 e5d4 e4c2 e8g8 d1d4 d6c5 d4e4 g7g6 c1h6 f8e8 e4d3 c5f8 h6f8 d7f8 d3c3 c8g4 f3d2 f8e6 f1e1 e6f4 d2e4 e8e4 e1e4 f4e2 e4e2 g4e2 c4c5 d8d7 a1e1 e2g4 c2b3 a8e8 e1e8 d7e8 h2h3 g4e6 c3e5 h7h6 b3e6 e8e6 e5e6 f7e6 g1f1 g8f7 f1e2 f7f6 e2e3 f6e5 h3h4');
INSERT INTO evaluations VALUES(10056,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10056,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10056,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(10056,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(10056,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(10056,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(10056,36,'cp 24','g1f3 d7d6 e1g1 c6d4 a2a4 d4f3 d1f3 h7h6 g5e3 c5e3 f3e3 e8g8 b1d2 c8e6 c4e6 f7e6 c2c3 d8e7 d3d4 e5d4 c3d4 b7b6 a1e1 a8c8 e3b3 f8d8 h2h3 c7c5 d2f3 f6h5 g2g3 h5f6 e4e5 f6d5 e5d6 e7d6');
INSERT INTO evaluations VALUES(11950,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11950,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(11950,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(11950,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(11950,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(11950,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(11950,36,'cp 47','d8e7 e2e4 f7f6 e1g1 e7e5 e4e5 f6e5 f1e1 d7d6 d2d4 c8f5 d4e5 e8c8 e5d6 f8d6 c1g5 d8e8 g5e3 f5c2 e1c1 c2d3 b1d2 d6e5 c1c6 e5b2 a1d1 d3b5 c6c5 a7a6 d1b1 b2e5 a2a4 b5e2 b1e1 e2d3 d2c4 d3c4 c5c4 e5d6 g2g3 e8e6 c4c6 c8b7 c6c2 h8e8 e1d1 h7h5 h2h4 g7g6 d1b1 b7c8');
INSERT INTO evaluations VALUES(12850,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12850,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12850,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12850,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(12850,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(12850,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(12850,36,'cp 1','b1c3 e7e6 f1b5 c6d4 e4e5 d4b5 c3b5 a7a6 b5c3 f6d5 c3d5 e6d5 d2d4 d7d6 e1g1 f8e7 c1e3 b7b6 d1d2 e8g8 a1d1 f8e8 d4c5 d6c5 d2d5 d8d5 d1d5 c8e6 d5d2 e6a2 f4f5 a2c4 f1d1 a8d8 d1a1 c4b5 d2d8 e8d8 b2b3 b5d7 g2g4 d7c6 g1f2 a6a5');
INSERT INTO evaluations VALUES(12848,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12848,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(12848,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12848,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12848,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12848,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12848,36,'cp 3','e7e6 f1b5 c6d4 e4e5 d4b5 c3b5 a7a6 b5c3 f6d5 c3d5 e6d5 d2d4 d7d6 c1e3 b7b6 e1g1 f8e7 d1d3 d6e5 f4e5 c5c4 d3d2 e8g8 f1f2 b6b5 e3g5 e7g5 d2g5 d8g5 f3g5 h7h6 g5f3 a6a5 f3h4 b5b4 a2a3 c8g4 h4f5 g4f5 f2f5 c4c3 b2c3');
INSERT INTO evaluations VALUES(12954,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(12954,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(12954,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(12954,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(12954,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(12954,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(12954,36,'cp -47','h7h6 g5c1 e7f5 d4c5 f8c5 f1d3 d8c7 e1g1 c6e5 d3f5 e5f3 d1f3 e6f5 c1f4 c7c6 f1e1 c8e6 b1d2 e8g8 f3g3 g8h7 d2b3 b7b6 f4e5 f7f6 b3d4 c6d7 d4e6 d7e6 e5f4 e6d7 b2b4 c5e7 g3d3 g7g5 f4e3 e7d6 a2a4 h7g6 a4a5 d6c7 a5b6 a7b6 a1d1 a8e8');
INSERT INTO evaluations VALUES(12937,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(12937,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(12937,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(12937,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(12937,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(12937,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(12937,36,'cp -44','g5c1 e7f5 d4c5 f8c5 f1d3 d8c7 e1g1 c6e5 d3f5 e5f3 d1f3 e6f5 c1f4 c7c6 f1e1 c8e6 b1d2 e8g8 d2b3 b7b6 f3g3 g8h7 f4e5 f7f6 b3d4 c6d7 d4e6 d7e6 e5f4 e6d7 b2b4 c5e7 g3d3 a8c8 g2g3 f8d8 f4e3 e7d6 e3d4 d8e8 e1e8 c8e8 a2a4 d6e5 a4a5 e5d4 d3d4 e8e4 d4d3 d7e6 g1g2 e4c4 a5b6 a7b6');
INSERT INTO evaluations VALUES(14122,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(14122,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(14122,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(14122,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(14122,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(14122,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(14122,36,'cp 117','c1e3 a7a6 d1d2 f8g7 d4d5 c6b8 e3h6 g7h6 d2h6 e7e5 h2h3 b8d7 f1e2 d8e7 e1c1 f6g8 h6e3 f7f5 e4f5 g6f5 g2g4 d7c5 b2b4 c5d7 g4f5 d7b6 f3g5 c8f5 f2f4 e8c8 f4e5 d6e5 d5d6 d8d6 d1d6 e7d6 g5f7 d6d4 e3d4 e5d4 f7h8 d4c3');
INSERT INTO evaluations VALUES(14850,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(14850,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(14850,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(14850,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(14850,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(14850,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14850,36,'cp -26','f8d6 b1d2 e6e5 d4e5 c6e5 f3e5 d6e5 d1c2 e8g8 b2b3 f8e8 e1g1 c7c6 h2h3 e5c7 f1d1 d8e7 d2f1 f6e4 c3c4 d5c4 b3c4 h7h6 c1b2 b7b6 a1b1 e4c5 d3h7 g8h8 b1c1 c8e6 h7f5 e6f5 c2f5 a8d8 b2d4 c5e4 f1d2 e4d2');
INSERT INTO evaluations VALUES(9256,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9256,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9256,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9256,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9256,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9256,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9256,36,'cp 3','b1d2 e6e5 d4e5 c6e5 f3e5 d6e5 d1c2 e8g8 b2b3 a7a5 c1a3 f8e8 d2f3 e5d6 a3d6 d8d6 h2h3 c8d7 e1g1 a5a4 c3c4 d5c4 d3c4 d6e7 f1d1 a4b3 a2b3 a8a1 d1a1 d7c6 f3d4 c6e4 c2c3 e4d5 d4f5 e7e4 c4d5 e4d5 f5h4 d5d6 h4f3 f6e4');
INSERT INTO evaluations VALUES(9255,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9255,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9255,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9255,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9255,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9255,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9255,36,'cp 0','e6e5 d4e5 c6e5 f3e5 d6e5 d1c2 e8g8 b2b3 d8e7 e1g1 b7b6 d2f3 e5d6 c1b2 c8b7 a1d1 a8d8 c3c4 d5c4 d3c4 b7f3 g2f3 d6e5 b2c1 e5h2 g1h2 e7e5 h2g1 e5g5 g1h1 g5h4 h1g2 h4g5 g2h1');
INSERT INTO evaluations VALUES(18050,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(18050,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(18050,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(18050,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(18050,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(18050,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(18050,36,'cp 35','c4b3 g8f6 e1g1 c8b7 d2d3 f8e7 c2c4 b5c4 b3c4 e8g8 b1c3 d7d6 a2a3 c6d4 f3d4 e5d4 c3d5 f6d5 e4d5 d8d7 d1f3 a6a5 a3a4 c7c5 c1f4 d7d8 b2b3 e7g5 f1e1 b7c8 e1e2 g5f4 f3f4 c8a6 a1e1 a6c4 b3c4 a8b8 h2h3 b8b3 f4f3 d8f6 f3f6 g7f6');
INSERT INTO evaluations VALUES(18661,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(18661,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(18661,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(18661,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(18661,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(18661,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(18661,36,'cp 40','a2a4 d7d6 h2h3 a7a6 f3h2 g8f6 h2g4 f6g4 h3g4 f8e7 b1c3 c8e6 c1e3 c6b4 f2f3 e7g5 e3g5 d8g5 d1d2 e8c8 e1c1 c8b8 b2b3 g5e7 c1b2 d6d5 c4d5 e6d5 c3d5 b4d5 e4d5 d8d5 g4g5 h6h5 d1e1 e7c7 e1e4 f7f5 g5f6 g7f6 g2g3');
INSERT INTO evaluations VALUES(18812,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18812,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18812,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(18812,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18812,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(18812,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18812,36,'cp 3','c6d4 e1g1 a7a6 d2d3 g8e7 c1g5 d8c7 f3d4 c5d4 c3b1 e7g6 c2c3 f8c5 d1c1 h7h6 g5d2 c7d8 c3d4 c5d4 d2c3 d4c3 b1c3 b7b5 d3d4 c8b7 c1e3 a8c8 a2a3 e8g8 f1d1 g6e7 d4d5 e6d5 e4d5 f8e8 e3f3 e7g6 f3g4 d8g5 g4g5 h6g5');
INSERT INTO evaluations VALUES(19009,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(19009,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(19009,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(19009,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(19009,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19009,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(19009,36,'cp -10','g8f6 c3c4 a8b8 b1c3 d7d6 c1d2 f8e7 f1e2 a7a6 a2a4 e8g8 e1g1 c8d7 c3d5 b7b6 h2h3 g8h8 b2b3 a6a5 d5e7 d8e7 f3e1 f6g8 f2f4 f7f5 e4f5 d7f5 e1c2 f5d7 e2f3 e5f4 f3c6 d7c6 f1f4 f8f4');
INSERT INTO evaluations VALUES(12565,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12565,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(12565,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12565,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(12565,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12565,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12565,36,'cp -1','c3c4 d7d6 f1e2 a8b8 b1c3 f8e7 c1d2 e8g8 e1g1 c8d7 h2h3 a7a6 c3d5 b7b5 b2b3 h7h6 a2a3 f8e8 a1b1 a6a5 b1a1 e7f8 a1c1 b5c4 b3c4 f6d5 c4d5 c6d4 f3d4 e5d4 c1b1 b8b1 d1b1 d8c7 b1c2 e8b8 f2f4');
INSERT INTO evaluations VALUES(22233,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(22233,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(22233,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(22233,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(22233,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(22233,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(22233,36,'cp 53','f6e5 f3e5 c6e5 d4e5 a7a6 h2h4 c7c5 d1h5 g7g6 h5e2 d8c7 h4h5 f8g7 c1f4 g8e7 g2g3 b7b5 f1g2 b5b4 c3d1 c8b7 h5h6 g7f8 f4g5 d5d4 g2b7 c7b7 f2f3 e7f5 g5f6 h8g8 h1h3 b7c6 g3g4 f5e7 d1f2 e7d5 f2e4 d5f4 e2h2 f4h3 h2h3 c6d5 f6h4');
INSERT INTO evaluations VALUES(22219,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(22219,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(22219,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(22219,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(22219,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(22219,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(22219,36,'cp 57','d4e5 g8h6 c1f4 h6f7 h2h4 f8b4 a2a3 b4c3 b2c3 d8e7 d1d2 c8d7 c3c4 e8c8 d2e3 c8b8 f1d3 d5c4 d3c4 b7b6 e1g1 d7c8 a3a4 c6a5 c4e2 c8b7 a1d1 h7h6 h4h5 d8d1 f1d1 a5c6 f3d4 c6d4 d1d4 h8d8');
INSERT INTO evaluations VALUES(22215,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(22215,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(22215,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(22215,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(22215,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(22215,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(22215,36,'cp 50','g8h6 c1f4 f8b4 a2a3 b4c3 b2c3 d8e7 f1d3 c8d7 d1d2 h6f7 f4g3 e8c8 a3a4 c8b8 d2e3 h7h5 h2h4 c6a5 e1g1 d8g8 g3f4 c7c5 f3g5 b8a8 c3c4 a5c4 d3c4 d5c4 f1d1 d7c6 d1d6 c6d5 a1d1 f7d6 e5d6 e7d7 e3c5 b7b6');
INSERT INTO evaluations VALUES(24089,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(24089,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(24089,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(24089,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(24089,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24089,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24089,36,'cp 41','a7a6 f1e2 e8g8 c3c4 f8e8 d1b3 d8e7 a1c1 c6a5 b3c3 a5c4 d2c4 d5c4 c3c4 f5e4 e1g1 c7c6 a2a3 a6a5 c4c3 a5a4 f3e5 e4d5 e2d3 d5b3 f1e1 f6d7 e5d7 e7d7 d3c2 d7d5 e3e4 d5b5 c2b3 b5b3');
INSERT INTO evaluations VALUES(26514,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(26514,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(26514,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(26514,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(26514,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(26514,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(26514,36,'cp -26','a8c8 e2e3 e7e6 f1d3 f8b4 b5c3 b4c3 b2c3 g8e7 f2f3 f5d3 c2d3 c6a5 d1a4 c7c6 g1e2 b7b6 e1g1 e8g8 f4g3 e7g6 a1c1 c6c5 d4c5 c8c5 f1d1 d8e7 h2h4 h7h5 g3f2 c5c6 a4b4 f8c8 b4e7 g6e7 e3e4 a5b7 e4d5 e7d5 c3c4');
INSERT INTO evaluations VALUES(59,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(59,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(59,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(59,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(59,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(59,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(59,36,'cp -25','e2e3 e7e6 f1d3 f8b4 b5c3 b4c3 b2c3 g8e7 f2f3 f5d3 c2d3 e8g8 e3e4 d8d7 f4e3 c6a5 g1e2 d5e4 f3e4 f7f5 e4f5 f8f5 e2g3 f5f7 e3f2 c8f8 e1g1 h7h6 d1d2 b7b6 a1e1 e7g6 d2e2 g6f4 e2e4 d7d5 f2e3 d5e4 d3e4 f4d3 e1d1 d3b2 f1f7');
INSERT INTO evaluations VALUES(30835,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30835,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(30835,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(30835,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(30835,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30835,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(30835,36,'cp 0','g8f6 g1f3 d7d5 g2g3 e7e6 f1g2 g7g6 c2c4 d5c4 b3c4 f8g7 e1g1 e8g8 d1e2 c7c5 d2d4 b8d7 f1d1 c5d4 e3d4 a8c8 b1d2 d8c7 a1c1 f8e8 d2f1 b7c6 f1e3 c7b7 h2h3 c6e4 b2a3 e4c6 a3b2');
INSERT INTO evaluations VALUES(34874,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(34874,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(34874,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(34874,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(34874,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(34874,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(34874,36,'cp 13','b8c6 g2g3 b7b6 c3b5 e6e7 c2c4 g8f6 f1g2 c8b7 e1g1 e8c8 f1e1 a7a6 b5c3 e7d7 f3e5 c6e5 e1e5 d7d3 c1g5 f8c5 g5f6 g7f6 g2b7 c8b7 e5d5 d8d5 c3d5 d3c4 d5f6 c4d4 d1f3 b7a7 a1d1 d4b2 d1d7 h8c8 f6d5 c5d6 g1g2 f7f5 f3f5 a7b8 f5h7 b2a2');
INSERT INTO evaluations VALUES(1690,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(1690,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(1690,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(1690,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(1690,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(1690,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(1690,36,'cp 43','g2g3 b7b6 c3b5 e6e7 c2c4 c8b7 f1h3 a7a6 b5c3 c6d4 c3d5 b7d5 c4d5 d4f3 d1f3 e7f6 f3e4 f8d6 c1d2 g8e7 e1g1 e8g8 a1e1 a8e8 h3g2 f6f5 d3d4 f5e4 e1e4 e7f5 d4e5 e8e5 e4c4 e5e7 f1d1 a6a5 d2c3 f8e8 g1f1');
INSERT INTO evaluations VALUES(38304,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(38304,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(38304,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(38304,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(38304,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(38304,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(38304,36,'cp 67','g1f3 g8f6 f1d3 f6e4 d3e4 b8d7 c2c3 d7f6 e4c2 c7c5 e1g1 c5d4 f3d4 c8d7 f1e1 e8g8 c1g5 d7c6 d4c6 b7c6 d1e2 d8b6 a1d1 f8d8 g5c1 a7a5 g2g3 e7c5 c2b3 h7h6 d1d8 a8d8 e1d1 d8d1 b3d1 b6b7 h2h4 c5d6 d1c2 d6c7 e2d3');
INSERT INTO evaluations VALUES(38303,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(38303,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(38303,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(38303,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(38303,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(38303,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(38303,36,'cp 72','g8f6 f1d3 f6e4 d3e4 c7c5 e1g1 b8d7 c2c3 d7f6 e4c2 e8g8 d1d3 c5d4 f3d4 g7g6 f1e1 d8b6 d4f3 c8d7 f3e5 d7c6 d3e2 a7a5 c1h6 b6b2 h6f8 a8f8 c3c4 e7b4 e1b1 b2d4 e5c6 b7c6 b1d1 d4b2 g2g3 g8g7 g1g2');
INSERT INTO evaluations VALUES(41049,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(41049,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(41049,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(41049,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(41049,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(41049,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(41049,36,'cp 64','f1d3 f6e4 d3e4 c7c5 e1g1 c5d4 f3d4 b8d7 c2c3 e8g8 e4c2 d7f6 d1d3 c8d7 f1e1 d8b6 c1g5 g7g6 h2h4 f8e8 d3f3 d7c6 d4c6 b6c6 f3c6 b7c6 a1d1 a8b8 g5c1 h7h5 g2g3 a7a5 g1g2 f6d5 c2e4 e8c8 c3c4 d5f6 e4c2 c8d8 d1d8 b8d8');
INSERT INTO evaluations VALUES(41047,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(41047,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(41047,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(41047,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(41047,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(41047,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(41047,36,'cp 41','b8d7 e4f6 d7f6 c1f4 e8g8 a2a3 c7c5 d4c5 e7c5 d1e2 d8e7 f3e5 f8d8 f4g5 h7h6 g5h4 d8d4 h4g3 c8d7 c2c3 d4d6 b2b4 c5b6 e1g1 d7c6 c3c4 d6d4 e5c6 b7c6 f1d1 c6c5 g3e5 d4d7 e2f3 a8d8 b4b5 b6c7 e5c7 d7c7');
INSERT INTO evaluations VALUES(9595,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(9595,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(9595,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(9595,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(9595,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(9595,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(9595,36,'cp 49','d1e2 f6e4 e2e4 d7f6 e4e2 e8g8 c1g5 c7c5 d4c5 d8a5 c2c3 a5c5 e1c1 h7h6 g5e3 c5c7 e3d4 f8d8 d3c2 c8d7 h1g1 a7a6 f3e5 e7c5 e2c4 b7b6 d4c5 b6c5 e5d7 f6d7 h2h4 d7f6 g2g4 d8d1 g1d1 a8b8 a2a4 c7b6 b2b3 g7g6 g4g5');
INSERT INTO evaluations VALUES(40539,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40539,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40539,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40539,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(40539,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(40539,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(40539,36,'cp 29','f4d6 d8d6 d1c2 d6e7 e2e3 e8g8 a1c1 b8d7 h2h3 f6e4 f1d3 e4c3 c2c3 d5c4 c3c4 e6e5 c4c2 g7g6 e1g1 e5d4 f3d4 d7f6 c2c5 e7c5 c1c5 f8d8 b2b4 g8f8 f1d1 f6d7 c5c2 a7a5 b4b5 c6c5 d4b3 b7b6');
INSERT INTO evaluations VALUES(40503,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40503,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40503,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40503,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40503,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40503,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(40503,36,'cp 23','d8d6 e2e3 e8g8 a1c1 d6e7 d1c2 b8d7 h2h3 f6e4 a2a4 e4c3 c2c3 a7a5 f1e2 d5c4 c3c4 e6e5 e1g1 f8e8 c4b3 e5d4 f3d4 d7f6 e2f3 h7h6 f1d1 f6e4 b3c2 e4g5 f3e2 g5e4 e2d3 e4f6 c2c5 e7c5 c1c5 c8d7 b2b3 g8f8 g2g4');
INSERT INTO evaluations VALUES(33366,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(33366,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(33366,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(33366,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(33366,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(33366,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(33366,36,'cp 30','e2e3 e8g8 a1c1 d6e7 d1c2 b8d7 h2h3 f6e4 c3e4 d5e4 c2e4 e7b4 f3d2 b4b2 c1b1 b2a3 f1e2 e6e5 e4c2 a3e7 d4d5 d7c5 e1g1 c6d5 c4d5 c8d7 d2b3 a8c8 b3c5 c8c5 c2b3 b7b6 b3a3 e7g5 e2f3 d7h3 b1c1 f8c8 c1c5 c8c5 a3a7');
INSERT INTO evaluations VALUES(33365,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(33365,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(33365,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(33365,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(33365,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(33365,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(33365,36,'cp 22','e8g8 a2a3 d5c4 f1c4 c6c5 c4e2 c8d7 d4c5 d6c5 a1c1 c5e7 e1g1 f8d8 d1c2 h7h6 f1d1 d7e8 d1d8 e7d8 c3e4 f6e4 c2e4 e8c6 e4b1 c6d5 c1c3 b8d7 b1c1 d8b6 b2b4 d7f6 h2h3 a8d8 f3d4 a7a6');
INSERT INTO evaluations VALUES(35361,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(35361,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(35361,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(35361,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(35361,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(35361,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(35361,36,'cp 24','d1c2 d6e7 h2h3 b7b6 c4d5 c6d5 f3e5 c8b7 f1d3 b8c6 e5c6 b7c6 e1g1 f8c8 f1c1 h7h6 c2b3 f6e8 c1c2 e8d6 a1c1 d6c4 a2a3 c6b7 c3e2 c4d6 c2c3 c8c3 c1c3 e7d8 b3a4 a7a5 a4c2 a8c8 f2f3 d8g5 e2f4');
INSERT INTO evaluations VALUES(42207,30,'cp 266',NULL);
INSERT INTO evaluations VALUES(42207,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(42207,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(42207,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(42207,34,'cp 282',NULL);
INSERT INTO evaluations VALUES(42207,35,'cp 249',NULL);
INSERT INTO evaluations VALUES(42207,36,'cp 264','e4d5 b8d7 e5d7 c8d7 d2d4 c5b4 c2c3 b4e7 b1d2 e8g8 e1g1 d7g4 f2f3 g4f5 f1e1 f8e8 d2e4 f6d5 d1b3 f5e6 b3b7 a8b8 b7a7 b8a8 a7b7 a8b8 b7c6 d5b4 c3b4 d8d4 c1e3 d4c4 c6c4 e6c4 a2a3 f7f5 e4d2 c4d5 e3f4 b8b7 h2h3');
INSERT INTO evaluations VALUES(42201,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(42201,31,'cp 266',NULL);
INSERT INTO evaluations VALUES(42201,32,'cp 260',NULL);
INSERT INTO evaluations VALUES(42201,33,'cp 266',NULL);
INSERT INTO evaluations VALUES(42201,34,'cp 279',NULL);
INSERT INTO evaluations VALUES(42201,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(42201,36,'cp 253','b8d7 e5d7 c8d7 d2d4 c5b4 c2c3 b4e7 e1g1 e8g8 b1d2 d7g4 d1b3 a8b8 h2h3 g4h5 f1e1 f8e8 c4f1 e7d6 e1e8 d8e8 c3c4 a7a5 d2f3 b7b5 c4c5 a5a4 b3d3 d6f8 d5d6 c7d6 c1f4 e8d7 a1c1 h5g6 d3e2 g6e4 b2b4 a4b3 a2b3 e4c6 c5d6 f6d5 f4g3 f8d6');
INSERT INTO evaluations VALUES(45455,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(45455,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(45455,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(45455,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(45455,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45455,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(45455,36,'cp 11','c8f5 c2c4 d5c4 f1c4 e7e6 d1b3 f8e7 b3b7 f5e4 b7b3 b8c6 b1d2 a8b8 b3d1 e4f3 d2f3 d8a5 c1d2 e7b4 e1g1 b4d2 f3d2 c6d4 d2f3 d4f3 d1f3 e8g8 b2b3 f8d8 a2a4 h7h6 f1d1 a5b6 d1d8 b8d8 a1d1 d8d1 f3d1 f6e4 d1c2 b6d4 c4d3 e4d6 d3e2 d4a1 g1h2 a1e5 h2h1');
INSERT INTO evaluations VALUES(29142,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(29142,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29142,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(29142,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29142,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(29142,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(29142,36,'cp 23','g2g4 f5d7 c1f4 d8b6 f1g2 e7e6 e1g1 f8d6 f4d6 b6d6 c2c3 d6b6 d1b3 e8g8 b3b6 a7b6 f3e5 d7c6 b1d2 f6e8 f2f4 f7f6 e5c6 b7c6 a2a4 e8d6 f1e1 g8f7 h3h4 b8d7 g2f3 g7g6 g1f2 c6c5 c3c4');
INSERT INTO evaluations VALUES(48774,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(48774,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(48774,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(48774,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(48774,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(48774,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(48774,36,'cp -62','f8e7 g1f3 h7h6 g5f6 e7f6 b1d2 a7a5 e1g1 b8c6 c2c3 e8g8 d1a4 c8e6 e2e4 d5e4 d3e4 c6b8 f1d1 c7c6 a4c2 d8c7 a2a4 f6e7 g2f1 b8d7 f1c4 e6g4 g1g2 g4h5 c4e2 h5g6 d2c4 d7c5 f3d2 a8d8 e2g4 b7b5 a4b5 c6b5 c4a5');
INSERT INTO evaluations VALUES(41113,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(41113,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(41113,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(41113,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(41113,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(41113,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(41113,36,'cp -56','b1d2 a7a5 g1f3 b8c6 c2c3 e8g8 d1a4 h7h6 g5f6 e7f6 e1g1 g7g5 e2e4 d5d4 c3d4 g5g4 f3e5 c6d4 f2f4 f6e5 f4e5 c8e6 f1f2 c7c5 a4d1 d8g5 d2f1 a8d8 d1d2 d4c6 f1e3 c6e5 e3d5 g5d2 f2d2 g8g7 a1c1 b7b6 c1c3 d8d6 d3d4 c5d4 d2d4');
INSERT INTO evaluations VALUES(53040,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(53040,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(53040,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(53040,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(53040,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(53040,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53040,36,'cp 31','d4c5 e7e5 f4g3 d8a5 d1d2 a5c5 e2e4 g8f6 b1c3 b8c6 f2f3 c8e6 a2a3 d6d5 c3d5 f6d5 g3f2 c6d4 e4d5 c5d5 g1e2 f8c5 e2d4 c5d4 f2d4 e5d4 f1d3 e8g8 d3e4 d5c5 d2b4 c5g5 e1g1 f8d8 b4b7 a7a5 b7c7 a8c8 c7g3 g5b5 b2b3 g7g6 a1d1');
INSERT INTO evaluations VALUES(56877,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(56877,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(56877,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(56877,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(56877,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(56877,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(56877,36,'cp 136','h2h4 d7d5 d2d4 d5e4 c3e4 c8d7 c1g5 f8e7 g1f3 d7c6 d1e2 d8d5 e4c3 d5d6 e1c1 b8d7 g2g3 e7g5 h4g5 g8e7 f1g2 e8c8 c3b5 c6b5 e2b5 d6c6 b5c6 e7c6 d1e1 c6e7 a2a4 a7a5 c2c3 c7c6 c3c4 c8c7 c1c2 c6c5 c2c3');
INSERT INTO evaluations VALUES(57281,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(57281,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(57281,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(57281,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(57281,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(57281,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(57281,36,'cp -41','f8b4 b1d2 c7c5 a2a3 b4d2 c1d2 d7d5 c4d5 g8f6 a1c1 d8d5 e2e3 d4e3 d2e3 b7b6 d1d5 f6d5 b2b4 d5e3 f2e3 b8d7 f1b5 e8g8 e1g1 c5b4 a3b4 d7f6 b5c6 a8b8 c1a1 c8b7 f3d4 b7c6 d4c6 b8b7 f1c1 f8e8 c6a7 e8e3 c1c8 e3e8 c8e8 f6e8 a7c6 g8f8 a1a8');
INSERT INTO evaluations VALUES(57382,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(57382,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(57382,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(57382,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(57382,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(57382,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(57382,36,'cp 116','c7c6 b5a4 b8a6 c2c4 f8c5 d2d4 c5b6 c4c5 b6c7 b1c3 d7d6 c5d6 d8d6 c1e3 c8g4 d4d5 g8e7 h2h3 g4d7 e1g1 c6d5 e4d5 d7a4 d1a4 d6d7 a4b3 c7d6 f3d2 e8g8 d2c4 e7f5 c3e4 f5e3 b3e3 d6b4 d5d6');
INSERT INTO evaluations VALUES(52195,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(52195,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(52195,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(52195,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(52195,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(52195,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(52195,36,'cp 125','b5a4 b8a6 c2c4 f8d6 d2d4 d6c7 e1g1 g8e7 a2a3 e8g8 b2b4 g8h8 b1c3 d7d6 h2h3 c8e6 d1e2 a6b8 d4d5 e6g8 f1d1 b8d7 c1e3 d7b6 a4b3 f6f5 d5c6 b7c6 f3e5 d8e8 e5f3 f5e4 c3e4 d6d5 c4d5 b6d5 d1e1 e7g6 e3d4 g6f4');
INSERT INTO evaluations VALUES(52214,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(52214,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(52214,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(52214,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(52214,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(52214,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(52214,36,'cp 109','b8a6 c2c4 f8d6 d2d4 d6c7 a2a3 g8e7 e1g1 d7d6 h2h3 e8g8 b2b4 g8h8 b1c3 c8e6 d1e2 a6b8 f1d1 b8d7 d4d5 e6g8 c4c5 c6d5 e4d5 a7a6 a1b1 a8c8 c3e4 e7d5 a4b3 f6f5 b3d5 g8d5 d1d5 f5e4 f3g5 d7f6 g5e6 d8e7 e6c7 e7c7 c5d6');
INSERT INTO evaluations VALUES(2813,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(2813,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2813,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(2813,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(2813,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(2813,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(2813,36,'cp 10','d7d5 b1d2 c8e6 f1e1 d5e4 d2e4 e6b3 a2b3 f6e4 d3e4 d8d1 e1d1 f7f6 g1f1 f8d8 c1e3 d8d1 a1d1 c6d8 g2g3 g8f7 f1e2 d8e6 d1d5 f7e8 h2h4 c7c6 d5d1 a6a5 d1a1 e8f7 f3e1 a5a4 b3a4 a8a4 a1a4 b5a4 e1d3 h7h5 f2f3');
INSERT INTO evaluations VALUES(2708,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(2708,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2708,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(2708,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(2708,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(2708,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2708,36,'cp 12','b1d2 c8e6 f1e1 d5e4 d2e4 f6e4 d3e4 e6b3 a2b3 d8d1 e1d1 f7f6 c1e3 f8d8 d1d5 g8f7 f3d2 f7e6 d5d8 e7d8 b3b4 a6a5 g1f1 a5a4 f1e2 d8e7 a1a2 c6d8 b2b3 d8b7 b3a4 a8a4 a2a4 b5a4 e2d3 e6d7 d3c2 h7h5 c2b2 g7g6 d2c4 d7e6');
INSERT INTO evaluations VALUES(10826,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(10826,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(10826,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(10826,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(10826,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(10826,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(10826,36,'cp 74','d2d4 e5d4 f3d4 e8g8 c1f4 d7d6 h2h3 f6d7 f4e3 c6d4 e3d4 d7b6 c4b3 c8e6 b3e6 f7e6 d1g4 e6e5 d4b6 a7b6 a1d1 c7c6 a2a4 d8c8 g4e2 c8d7 e2e3 d7c7 b2b3 a8d8 e3d3 g8h8 d3e2 c7d7 e2e3 d7c7 d1d3 h8h7 f1d1 h7h8 e3e2 f8f7');
INSERT INTO evaluations VALUES(10841,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(10841,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(10841,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(10841,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(10841,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(10841,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(10841,36,'cp 64','e5d4 f3d4 e8g8 h2h3 d7d6 c1f4 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 f6d7 c4e6 f7e6 f4e3 e7f6 d4c4 f8e8 f2f4 f6c3 c4c3 d7f6 c3d3 d8e7 e3d4 a8d8 d1e1 e6e5 d4c3 e7e6 g1h2 e5f4 f1f4 f6d7 d3g3 d7e5 e1f1 d8d7');
INSERT INTO evaluations VALUES(10844,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(10844,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(10844,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(10844,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(10844,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(10844,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(10844,36,'cp 67','f3d4 e8g8 c1f4 d7d6 h2h3 f6d7 f4e3 c6d4 e3d4 d7e5 c4e2 f7f5 e4f5 c8f5 f2f4 e5d7 e2c4 g8h8 c3d5 c7c5 d5e7 d8e7 d4c3 d7b6 c4b5 d6d5 c3e5 e7f7 d1d2 d5d4 a1e1 f7g6 g1h2 a7a6 b5d3 f5d3 d2d3 g6d3 c2d3 h8g8 e5d6');
INSERT INTO evaluations VALUES(10843,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(10843,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(10843,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(10843,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(10843,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(10843,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(10843,36,'cp 56','e8g8 c1f4 d7d6 h2h3 c6d4 d1d4 a7a6 a2a4 c8e6 a1d1 f6d7 c4e6 f7e6 f4e3 e7f6 d4c4 f8e8 f2f4 f6c3 c4c3 d7f6 c3d3 d8e7 e3d4 a8d8 d3b3 f6e4 b3b7 e4g3 f1e1 g3h5 b7a6 h5f4 a6c6 e7g5 c6f3 e6e5 d4e3 g5g6 c2c3 e8f8 f3g4 g6g4 h3g4');
INSERT INTO evaluations VALUES(11443,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(11443,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(11443,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(11443,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(11443,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(11443,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(11443,36,'cp -14','f8b4 e1g1 e8g8 d4e5 b4c3 e5f6 c3f6 c2c3 f8e8 d1d8 f6d8 f1d1 c6c5 c1f4 c8e6 a2a4 e6b3 d1e1 a8c8 f3d2 b3e6 e1d1 g7g5 f4g3 h7h5 h2h3 b7b6 f2f3 d8e7 f3f4 g5f4 g3f4 e8d8 d2f3 d8d1 a1d1 e6b3 d1a1 f7f6 f3h4 g8f7 h4f5 e7f8');
INSERT INTO evaluations VALUES(14236,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(14236,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(14236,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(14236,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14236,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(14236,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(14236,36,'cp 34','e2e3 c8f5 c2c4 e7e6 g1f3 f8d6 f4g5 e8g8 f1e2 h7h6 g5h4 d6e7 e1g1 f6e4 h4e7 d8e7 a2a3 f8d8 a1c1 c6b8 b2b4 e4d2 d1d2 c7c6 c4d5 e6d5 a3a4 b8d7 b4b5 c6b5 a4b5 d8c8 d2b2 f5e4 f3d2 d7f6');
INSERT INTO evaluations VALUES(14744,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14744,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(14744,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(14744,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(14744,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(14744,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14744,36,'cp 0','e2e3 g7g6 b1c3 a7a6 b2b3 f8g7 f1e2 e8g8 c1b2 f8e8 a1c1 e6e5 f3e5 c6e5 d4e5 e8e5 c5c6 b7c6 e2f3 d8d6 c3a4 e5e7 d1c2 d6b4 b2c3 b4d6 c3d4 d6b4 d4c3');
INSERT INTO evaluations VALUES(18024,30,'cp -353',NULL);
INSERT INTO evaluations VALUES(18024,31,'cp -362',NULL);
INSERT INTO evaluations VALUES(18024,32,'cp -361',NULL);
INSERT INTO evaluations VALUES(18024,33,'cp -356',NULL);
INSERT INTO evaluations VALUES(18024,34,'cp -379',NULL);
INSERT INTO evaluations VALUES(18024,35,'cp -390',NULL);
INSERT INTO evaluations VALUES(18024,36,'cp -363','c7c5 d4c3 c5c4 b3c4 b5c4 c3c4 f8e7 c4d3 c8e6 e1g1 d6d5 e4d5 d8d5 d3d5 e6d5 b1c3 d5e6 c3e2 a8d8 b2b3 e7f6 a1b1 g8e7 c1a3 d8d2 f1e1 d2c2 e2f4 a6a5 h2h3 e7g6 f4h5 f6e5 e1e5 g6e5 h5g7 e8d7 b1d1 d7c8 d1c1 e6b3 a2b3');
INSERT INTO evaluations VALUES(18244,30,'cp -359',NULL);
INSERT INTO evaluations VALUES(18244,31,'cp -358',NULL);
INSERT INTO evaluations VALUES(18244,32,'cp -371',NULL);
INSERT INTO evaluations VALUES(18244,33,'cp -376',NULL);
INSERT INTO evaluations VALUES(18244,34,'cp -370',NULL);
INSERT INTO evaluations VALUES(18244,35,'cp -373',NULL);
INSERT INTO evaluations VALUES(18244,36,'cp -344','d4c3 c5c4 a2a4 c4b3 a4b5 g8f6 e1g1 f8e7 b5a6 e8g8 a6a7 c8e6 c2b3 f6e4 c3c2 e4c5 c1e3 d8b6 a1a3 f8c8 b1c3 b6b7 f1d1 e6b3 a3b3 c5b3 c2f5 b3c5 c3d5 e7d8 b2b4 c5e6 d1a1 e6c7 d5c7 c8c7 b4b5 c7e7 f5f3 b7f3 g2f3 e7b7');
INSERT INTO evaluations VALUES(23238,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(23238,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(23238,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(23238,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(23238,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(23238,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(23238,36,'cp 14','f8b4 d3d4 d7d5 e4d5 d8d5 c1d2 b4c3 d2c3 e5e4 f3d2 g8e7 f1c4 d5g5 d2e4 g5g2 e4g3 h7h5 f2f3 c8f5 d1e2 g2e2 g3e2 f5c2 h1g1 g7g6 e1f2 e8c8 e2f4 g6g5 d4d5 g5f4 d5c6 e7d5 c4d5 d8d5 c6b7 c8b7 a1e1 h8d8 g1g7 d8d6 g7f7 c2f5 f7f6 d6f6 c3f6 d5d2 e1e2 d2e2 f2e2 b7a8 a2a3');
INSERT INTO evaluations VALUES(6937,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(6937,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(6937,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(6937,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(6937,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(6937,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6937,36,'cp 12','d3d4 d7d5 e4d5 d8d5 c1d2 b4c3 d2c3 e5e4 f3d2 g8e7 f1c4 d5g5 d1e2 g5g2 e2e4 g2e4 d2e4 c8f5 f2f3 e8c8 e1c1 e7d5 d1g1 d5b6 c4b5 c6d4 c3d4 d8d4 g1g7 f5e4 f3e4 d4e4 h1g1 c7c6 b5d3 e4f4 g7g8 h8g8 g1g8 c8c7 g8g7 b6d7 g7h7 c7d6 h2h4 f6f5 c1d2 d7c5 d2e3 c5d3 c2d3 f4b4 b2b3 a7a5 d3d4 d6e6 h7h6 e6f7 h4h5 a5a4 e3f4 a4b3 a2b3 b4b3');
INSERT INTO evaluations VALUES(24150,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24150,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(24150,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24150,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24150,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24150,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(24150,36,'cp 25','c4e6 f7e6 c3a4 c5b6 e1g1 e8g8 c1d2 h7h6 b2b4 d8e8 a4b6 a7b6 a2a4 e8h5 c2c3 h5g6 d2e3 f8f7 g1h2 g6h5 f3g1 h5g6 g1e2 f6h5 b4b5 c6e7 d1d2 g6f6 e2g1 h5f4 f2f3 e7g6 g2g3');
INSERT INTO evaluations VALUES(24146,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(24146,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24146,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24146,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(24146,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24146,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(24146,36,'cp 27','f7e6 c3a4 c5b6 e1g1 e8g8 c1d2 d8e8 a4b6 a7b6 a2a4 h7h6 b2b4 e8h5 c2c3 h5g6 d2e3 f8f7 d1e2 a8f8 f3h4 g6h5 e2h5 f6h5 h4f3 h5f4 e3f4 f7f4 f3d2 g7g5 a4a5 b6a5 b4b5 c6b8 a1a5 b8d7 d2b3 g8f7 a5a7 f8b8 f1a1 f4f6');
INSERT INTO evaluations VALUES(24170,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24170,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24170,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24170,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24170,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(24170,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24170,36,'cp 15','c3a4 c5b6 e1g1 e8g8 c1d2 d8e8 a4b6 a7b6 a2a4 h7h6 b2b4 f6h5 g1h2 e8f7 c2c3 h5f4 d2e3 f7h5 f3g1 h5f7 g2g3 f4h5 h2g2 f7g6 d1g4 g6f7 g4e2 f7g6 e2d1 c6b8 d1g4 g6f7 c3c4 h5f6 g4d1 b8d7 d1c2 f7g6 g2h2 c7c5');
INSERT INTO evaluations VALUES(26455,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(26455,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(26455,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(26455,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(26455,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(26455,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(26455,36,'cp 115','a2a4 g4d7 a4a5 a7a6 h2h3 g8f6 f3d2 h7h5 f1d3 h5h4 d2f3 e7g6 e1g1 f8e7 c1e3 f6h5 c3e2 e8g8 a1a3 c7c6 c2c4 d8e8 d1d2 f7f5 f1e1 f5f4 e3b6 e7d8 b2b4 d8b6 a5b6 c6c5 b4c5 d6c5 g1h2 e8e7');
INSERT INTO evaluations VALUES(32136,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(32136,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(32136,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(32136,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(32136,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(32136,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(32136,36,'cp 25','b8d7 d2d4 g8f6 e1g1 c7c6 d1c2 g4h5 c1e3 h5g6 e2d3 f8e7 h2h3 d6d5 f3e5 g6e4 b1d2 e4d3 e5d3 e8g8 e3f4 a7a5 a2a4 f8e8 a1e1 h7h6 e1e2 e7f8 e2e8 f6e8 f1e1 d8b6 g2g3 e8c7 g1g2 a8e8 e1e8 c7e8 f4e3 b6c7');
INSERT INTO evaluations VALUES(32515,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32515,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32515,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32515,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(32515,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(32515,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(32515,36,'cp 0','d1e2 d8e7 e1g1 e7e2 b5e2 b8c6 b1c3 g8f6 d2d4 e8c8 c1g5 d7f5 a1c1 f8e7 e2b5 h7h6 g5e3 c6a5 f3e5 f5e6 c3e2 f6g4 e2f4 g4e5 d4e5 g7g5 f4e6 f7e6 e3a7 b7b6 b5a6 c8d7 a6b5 d7c8');
INSERT INTO evaluations VALUES(32516,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32516,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32516,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32516,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32516,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32516,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(32516,36,'cp 0','d8e7 e1g1 c7c6 b5d3 e7e2 d3e2 f8d6 f1e1 g8e7 b2b3 e8g8 c1a3 e7c8 d2d4 f8d8 c2c4 d6a3 b1a3 d5c4 b3c4 d7f5 f3h4 f5e6 h4f3');
INSERT INTO evaluations VALUES(34850,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(34850,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(34850,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(34850,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(34850,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(34850,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(34850,36,'cp 44','d2d4 f8b4 g1f3 g8e7 c1d2 b4c3 d2c3 b8c6 e1c1 f7f6 g2g4 a7a5 g4g5 c8d7 h1g1 c6b4 c1b1 e7f5 d1e1 h7h6 g5f6 g7f6 b2b3 e8f7 a2a4 b7b6 c3d2 d7c6 b1b2 c6e4 c2c3 b4d5 c3c4 d5b4 d2f4 h8c8');
INSERT INTO evaluations VALUES(37607,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(37607,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(37607,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(37607,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(37607,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(37607,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(37607,36,'cp 163','e4e5 d7d5 e5d6 b7b6 g1f3 g8f6 g2g3 e8g8 f1g2 c8a6 d1a4 c6c5 e1g1 a6b7 d4c5 b6c5 f1d1 d8b6 a4c2 b7c6 f3e5 c6g2 g1g2 f6d7 e5d7 b8d7 a1b1 b6a6 b2b3 a8d8 c1g5 f7f6 g5f4 f6f5 c3b5 e6e5 f4g5 a6c6 f2f3 e5e4 f3e4');
INSERT INTO evaluations VALUES(38778,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(38778,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(38778,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(38778,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(38778,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(38778,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(38778,36,'cp 36','b1d2 b4d6 f4g3 g8f6 e2e3 e8g8 f1d3 c7c5 c2c3 b8c6 d1e2 g7g6 e1g1 d6g3 h2g3 b7b6 f1b1 c8b7 b2b4 c5b4 c3b4 d8e7 a2a3 a8c8 b1c1 f8d8 a1b1 e6e5 d4e5 c6e5 d3a6 e5c6 a6b7 e7b7 b4b5 c6a5 f3d4 b7e7 c1c8');
INSERT INTO evaluations VALUES(41131,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(41131,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(41131,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(41131,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(41131,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(41131,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(41131,36,'cp 130','h2h4 h7h6 c1e3 b8d7 f1e2 d7f6 f3h2 c8d7 e1g1 e6e5 d4e5 d6e5 b2b4 e8g8 b4b5 c7c6 a2a4 d7e6 d1c1 d8a5 e3c5 a5c7 c1a3 e7c8 f1d1 f8e8 a4a5 f6d7 h4h5 d7c5 a3c5 g6h5 a5a6 c8b6 a6b7 c7b7 c5c6 b7c6 b5c6');
INSERT INTO evaluations VALUES(41275,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(41275,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(41275,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(41275,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(41275,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41275,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(41275,36,'cp 49','f1g2 e8g8 g1f3 c7c5 d4d5 e7e6 c1f4 e6d5 c4d5 f8e8 e1g1 c8f5 f3h4 f5g4 d1d2 b7b5 c3b5 e8e2 d2c1 f6e4 g2e4 e2e4 b5d6 e4e7 c1c5 b8d7 c5a3 d7f6 h4g2 e7d7 g2e3 g7f8 e3g4 f6g4 a3a4 f8d6 f4d6 d7d6 a4g4 d6d5 a1d1 d8a5 d1d5');
INSERT INTO evaluations VALUES(42104,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(42104,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42104,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42104,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42104,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(42104,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(42104,36,'cp 23','d8e7 b1d2 e6e5 d4e5 e7e5 g1f3 e5e7 c3c4 c8e6 d1b3 c7c6 c4d5 e6d5 f1c4 d5c4 d2c4 e8g8 b3a3 e7a3 c4a3 b8d7 e1e2 a8d8 h1d1 d7b6 d1d4 f8e8 a1c1 b6d5 a3c4 c6c5 d4d1 h7h6 c4a5 b7b6 a5c6 d8d7 f3e5 d7c7 a2a3 f6e4');
INSERT INTO evaluations VALUES(43419,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(43419,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(43419,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(43419,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(43419,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(43419,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(43419,36,'cp 129','g1f3 e7e6 f1d3 f8e7 e1g1 b8c6 a2a3 c8d7 h2h3 e8g8 c1d2 a8c8 f1e1 b7b5 b2b4 d8b6 e4e5 d6e5 d4e5 f6d5 c3d5 e6d5 d2e3 d5d4 e3d2 d7e6 a1c1 b6d8 d1e2 d8d7 e2e4 g7g6 e4f4 f8e8 h3h4 e6f5 d3e4 f5e4 f4e4 d7e6');
INSERT INTO evaluations VALUES(44623,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(44623,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(44623,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(44623,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44623,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(44623,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44623,36,'cp 15','f1d3 d7d5 h2h3 b8c6 g1f3 c6b4 e1g1 b4d3 d1d3 e7e6 f1e1 f8e7 c1g5 e8g8 a2a4 b7b6 f3e5 c8b7 b1d2 a7a5 d3g3 f6e8 g5e7 d8e7 c2c3 e8d6 e5d3 a8c8 d3f4 e7c7 h3h4 g8h8 h4h5 d6e4 d2e4 d5e4 g3h4');
INSERT INTO evaluations VALUES(51596,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(51596,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(51596,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(51596,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(51596,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(51596,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(51596,36,'cp 70','g1f3 a7a6 d2d4 c5d4 f3d4 e7e6 c1e3 d8c7 d1d2 g8f6 f2f3 f8b4 g2g4 h7h6 e1c1 b4c3 d2c3 c7c3 b2c3 d7d5 e4e5 f6d7 f3f4 d7c5 f1g2 c8b7 f4f5 b8d7 h1e1 e8g8 g4g5 h6g5 e3g5 f8e8 f5f6 c5e4 g2e4 d5e4 g5f4 g7f6 e5f6 d7f6 e1g1 g8h7 f4e5 f6d7 g1g7 h7h8 g7f7 d7e5 f7b7');
INSERT INTO evaluations VALUES(52854,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(52854,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(52854,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(52854,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(52854,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52854,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(52854,36,'cp -11','c8f5 d3d4 e7e6 c2c3 c6c5 f1b5 b8d7 g1e2 a8c8 e1g1 a7a6 b5d3 g8e7 b1a3 d8c7 c1g5 f5d3 d1d3 e7c6 g5d2 c5d4 c3d4 f8b4 a3b1 d7b6 b2b3 e8g8 d2b4 c6b4 d3d2 b4c6 b1c3 f7f6 e5f6 f8f6');
INSERT INTO evaluations VALUES(31179,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(31179,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(31179,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(31179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31179,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(31179,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(31179,36,'cp 0','c2c3 e7e6 d3d4 c6c5 f1b5 b8c6 g1e2 d8b6 b5a4 f7f6 e1g1 f6e5 d4e5 e8c8 a4c6 b6c6 b2b4 d5d4 c3d4 c5b4 d1b3 c8b8 c1g5 f8e7 g5e7 g8e7 b3b4 f5d3 f1c1 c6d7 b1c3 e7c6 b4b2 c6d4 e2f4 d3f5 f4e2');
INSERT INTO evaluations VALUES(52857,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(52857,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(52857,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(52857,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(52857,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(52857,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52857,36,'cp 51','d2d4 c6c5 g1f3 c5d4 d1d4 b8c6 d4f4 g8h6 b1c3 a7a6 f1d3 g7g6 c3e2 c8g4 e2d4 g4f3 d4f3 e7e6 c2c3 d8c7 e1g1 h6f5 f4a4 h7h6 h2h4 e8c8 d3f5 g6f5 c1f4 f8g7 f1e1 c8b8');
INSERT INTO evaluations VALUES(53916,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(53916,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(53916,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(53916,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(53916,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(53916,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(53916,36,'cp 65','g7g6 b1c3 f8g7 f1d3 b8d7 g1f3 d8e7 d1e2 d7e5 f3e5 g7e5 a1b1 c8d7 e1g1 f7f5 f1e1 e8f7 c1e3 b7b6 e2d2 g8f6 h2h3 h8e8 e3g5 e7f8 g5h6 f8h8 c3e2 f6h5 b2b4 c5b4 b1b4 h8f6 g2g3 f7g8 h6g5 f6f7 b4b1 h5f6 f2f4');
INSERT INTO evaluations VALUES(57069,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(57069,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(57069,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(57069,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(57069,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(57069,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57069,36,'cp 0','d7d5 g2g3 g8f6 f1g2 f8e7 c2c4 c7c5 e1g1 e8g8 d3d4 c5d4 c4d5 d8d5 f3d4 d5c5 d4b3 c5f5 b3d4');
INSERT INTO evaluations VALUES(3582,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3582,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(3582,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(3582,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3582,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3582,36,'cp 0','g5f6 d8f6 c3d5 f6d8 c2c3 c6a5 b2b4 a5c4 b4c5 c4a5 e1g1 f7f5 d5e3 f5e4 d3e4 c8e6 d1e2 b7b6 a1d1 d8e7 e3d5 e6d5 d1d5 b6c5 e2a6 f8f3 g2f3 e7g5 g1h2 g5f4 h2g1');
INSERT INTO evaluations VALUES(4621,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(4621,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(4621,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(4621,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(4621,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(4621,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(4621,36,'cp 50','d7d6 c1f4 a7a6 a2a4 c8e6 a1d1 d8d7 c4e2 a8d8 d4e3 d7c6 h2h3 f6d7 c3d5 f8e8 d5e7 e8e7 d1d4 f7f6 f2f3 e6f7 a4a5 d7f8 d4b4 d8e8 e3c3 c6c3 b2c3 f7d5 e2a6 b7a6 e4d5 e7e2 b4b7 e8c8 b7a7 f8g6 f4g3 e2c2 a7a6 g6e7 f1d1 c2c3 a6a7 c3c2');
INSERT INTO evaluations VALUES(5591,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(5591,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(5591,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(5591,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(5591,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(5591,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(5591,36,'cp 57','d1d5 d8e7 d5e4 c8f5 e4e7 g8e7 c1d2 c5b6 e1c1 b6f2 f1c4 e8g8 h1f1 f2b6 d1e1 h7h6 b2b4 a8e8 a2a3 a7a6 c1b2 f5g4 h2h3 g4f3 f1f3 e8d8 b2c1 e7g6 c3d5 b6d4 f3f1 g6e5 c4b3 b7b5 d2e3 f8e8 e3f2 d4f2 f1f2');
INSERT INTO evaluations VALUES(5590,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(5590,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(5590,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(5590,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(5590,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(5590,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(5590,36,'cp 55','d8e7 d5e4 c8f5 e4e7 g8e7 c1d2 c5b6 e1c1 b6f2 f1c4 e8g8 h1f1 f2b6 d1e1 h7h6 c3a4 b6c7 a4c5 b7b6 c5d3 a8c8 b2b3 c7d6 c1b1 c8c7 a2a4 e7g6 d2c3 c7e7 e1e7 c6e7 f1f2 f5e4 g2g3 e7d5 c3b2 d5e3');
INSERT INTO evaluations VALUES(11580,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(11580,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(11580,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(11580,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(11580,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11580,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(11580,36,'cp 13','e7c6 e1g1 f8e7 d2d4 d7d6 c3d5 e8g8 c2c3 f8e8 d4e5 d6e5 d1d3 c8e6 f1d1 e7f8 c1e3 a6a5 h2h3 f7f6 f3h4 a5a4 d3e2 d8c8 a1c1 e8d8 d1d2 c6e7 c3c4 d8d7 c1d1 c8e8 e2f3 e8f7 h4f5 g8h8 f5e7 f8e7 d5e7 d7d2 d1d2');
INSERT INTO evaluations VALUES(11558,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(11558,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(11558,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(11558,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(11558,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(11558,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(11558,36,'cp 12','d2d4 e5d4 f3d4 f8b4 d4e2 d7d6 e1g1 b4c3 e2c3 e8g8 c1f4 f7f5 d1d5 g8h8 f2f3 d8f6 d5d2 c8e6 e4f5 e6f5 f4g3 h7h6 c3d5 f6b2 a1b1 b2d4 d2d4 c6d4 b1b7 f5e4 c2c4 e4d3 f1e1 d3c4 d5f4 f8f4 g3f4 d4e2 e1e2 c4e2 b7c7');
INSERT INTO evaluations VALUES(11554,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(11554,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(11554,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(11554,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(11554,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(11554,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(11554,36,'cp 12','e5d4 f3d4 f8b4 d4e2 e8g8 e1g1 d7d6 c1e3 b4c3 e2c3 f7f5 f2f3 c8e6 d1d2 d8f6 e4f5 f6f5 b2b3 a8e8 e3f2 f5g6 f2g3 g6f7 c3e2 e6d7 c2c4 c6e7 a1e1 d7c6 e2f4 e7g6 e1e8 f8e8 f4e2 b7b6 e2c3 h7h6 f1e1 e8e1 d2e1 g6f4 e1d2');
INSERT INTO evaluations VALUES(11568,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(11568,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(11568,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(11568,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(11568,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(11568,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(11568,36,'cp 14','f3d4 f8b4 d4e2 d7d6 e1g1 b4c3 e2c3 e8g8 f2f3 f7f5 c1e3 d8f6 d1d2 c8e6 e4f5 f6f5 b2b3 c6e7 e3f2 e6d7 a1d1 a8e8 f1e1 f5f7 c3e2 e7f5 c2c4 d7c6 h2h3 e8e7 e2f4 f8e8 e1e7 f5e7 d1e1 h7h6 f4e2');
INSERT INTO evaluations VALUES(13258,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(13258,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(13258,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(13258,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(13258,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(13258,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(13258,36,'cp -16','c8g4 c4b5 e7e6 h2h3 g4h5 g2g4 h5g6 d2d3 f6d7 c1g5 d8c7 d1e2 h7h6 g5h4 f7f6 b5c6 b7c6 f3d2 c7b7 b2b3 g6f7 h4g3 d6d5 h3h4 f8e7 f2f4 e6e5 e1g1 e7d6 c3a4 e5f4 e4d5 d6e5 d5c6 b7c6 g3f4 e8g8 f4e5 a8e8 d2e4 e8e5 e2g2 f8e8 a4c3 c6c7');
INSERT INTO evaluations VALUES(14708,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(14708,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(14708,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(14708,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(14708,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(14708,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(14708,36,'cp 115','c1h6 g7h6 b1d2 f8g7 f1b5 c8d7 a2a4 c6e7 e1g1 e7g6 f1e1 e8g8 c2c3 b7b6 d2f1 d7b5 a4b5 d8e8 d1b3 g8h8 f1g3 f7f6 e5f6 g7f6 b3c2 g6f4 c2d2 f6g5 f3g5 h6g5 g3e2 e8b5 e2f4 f8f4 e1e6 a8f8 e6e2 a7a5 g2g3 f4e4');
INSERT INTO evaluations VALUES(15315,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(15315,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(15315,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15315,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15315,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15315,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15315,36,'cp 0','c8f5 e1g1 d8d7 d1e2 f7f6 d3d4 e5e4 c4d5 d7d5 b1c3 d5e6 f3g5 f6g5 d4d5 e6e5 d5c6 b7c6 c1d2 f8d6 f2f4 g5f4 d2f4 e5c5 f4e3 c5e5');
INSERT INTO evaluations VALUES(19758,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(19758,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(19758,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(19758,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(19758,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(19758,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(19758,36,'cp 41','f2f4 e7e6 c1e3 d5d4 e3f2 f8c5 b1d2 g8e7 a2a3 e7d5 d1f3 e8g8 g1h3 f7f6 e5f6 d5f6 f2h4 c5e7 f1d3 f6d5 h4e7 d8e7 e1g1 g7g6 f1e1 e7c7 f3g3 d5f4 h3f4 c7f4 g3f4 f8f4 d3c4 f4f5 e1e6 c8e6 c4e6 f5f7 d2f3 g8g7 e6f7 g7f7 g1f2 f7e6 a1e1 e6d5');
INSERT INTO evaluations VALUES(21916,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(21916,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(21916,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(21916,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(21916,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(21916,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(21916,36,'cp 31','g1e2 c6d4 e2d4 e5d4 c3e2 c7c5 d2d3 g8e7 b2b4 a7a6 e2f4 c5b4 f1e2 g7g6 d1a4 e7c6 a2a3 b4a3 a1b1 f8g7 f4d5 a8a7 a4a3 e8g8 e1g1 c8e6 d5b6 a6a5 b6d5 f7f5 f2f4 g8h8 c1d2 e6d5 e4d5 c6e7 e2f3 b7b6');
INSERT INTO evaluations VALUES(24461,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24461,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(24461,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(24461,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(24461,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(24461,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(24461,36,'cp 10','b7c6 b5c4 g7g6 e1g1 f8g7 h2h3 e8g8 c1e3 d8c8 d1d3 d7e6 c4e6 c8e6 a1d1 a7a5 f2f4 a5a4 e3d4 f8d8 d3f3 e6c8 e4e5 f6d5 e5d6 d5c3 d4c3 d8d6 d1d6 e7d6 c3g7 g8g7 c2c4 c8d7 f3d3 d6d5 f1d1 a8e8 c4d5 c6d5 d3d4 f7f6');
INSERT INTO evaluations VALUES(24392,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(24392,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24392,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(24392,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(24392,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24392,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24392,36,'cp 13','b5c4 g7g6 h2h3 f8g7 c1e3 e8g8 e1g1 d8c8 d1d3 d7e6 c4e6 c8e6 b2b3 d6d5 e4d5 f6d5 c3d5 c6d5 a1c1 e6c6 c2c4 d5d4 e3d2 e7e5 b3b4 e5e4 d3b3 a7a5 b4b5 c6c5 a2a4 f7f5 d2f4 a8e8 b5b6 g7e5 f4e5 e8e5 b6b7 f8b8 f1d1 f5f4 c1b1 d4d3 b3c3 f4f3');
INSERT INTO evaluations VALUES(30806,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(30806,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(30806,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30806,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(30806,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(30806,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30806,36,'cp 1','d7d5 f1g2 e7e6 e1g1 b8d7 c2c4 d5c4 c1g5 d8c8 d1a4 h7h6 g5f4 g8f6 a4c4 c7c5 b1d2 f8e7 d4c5 b6c5 c4c2 e8g8 f1d1 f8d8 a1c1 f6d5 d2c4 d5f4 g3f4 c8c7 f3e5 d7e5 f4e5 b7g2 g1g2 d8d1 c1d1 a8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(31386,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(31386,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(31386,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31386,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(31386,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(31386,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(31386,36,'cp 25','e1g1 c6c5 c1e3 c5d4 e3d4 g8e7 c2c4 d5c4 d3c4 b8c6 b1c3 e7f5 a1d1 d8c8 d4c5 f8c5 c4c5 f5e7 f3d2 e8g8 d2c4 b7b6 c5b5 c8c7 c4d6 a7a6 b5e2 a8d8 f2f4 b6b5 c3e4 e7d5 e2f2 c7a7 d1d2 a7f2 f1f2 d8d7 e4c5 d7c7 d6e4 f8c8');
INSERT INTO evaluations VALUES(31387,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(31387,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(31387,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31387,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(31387,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(31387,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31387,36,'cp 27','c6c5 c1e3 c5d4 e3d4 b8c6 c2c4 d5c4 d3c4 g8e7 b1c3 e7f5 f1d1 d8c8 d4c5 f8c5 c4c5 f5e7 c3b5 e8g8 b5d6 c8c7 a1c1 c7b6 c5b6 a7b6 a2a3 f8d8 d6b7 d8d1 c1d1 e7g6 d1e1 a8b8 b7d6 b8d8 g1f1 g6e5 f3e5 c6e5 e1e5 d8d6 f1e2 f7f5 a3a4 g8f8 e5b5 f8e7 b2b4 d6c6 e2d2 g7g5 g2g3 e7d6');
INSERT INTO evaluations VALUES(31392,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31392,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(31392,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(31392,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(31392,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(31392,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(31392,36,'cp 36','f1d3 g8e7 c3e2 f5d3 d1d3 c6c5 c2c3 c5d4 c3d4 d8a5 e1f1 a5a6 d3d1 b8c6 h2h4 h6h5 g2g3 e7f5 f1g2 a8c8 c1d2 f8e7 d2c3 e7b4 a1c1 b4c3 c1c3 a6a2 d1d2 a2a6 e2f4 a6b6 h1c1 a7a5 f4e2 e8d7 f3g5 f5d4');
INSERT INTO evaluations VALUES(31393,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31393,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(31393,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31393,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(31393,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31393,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31393,36,'cp 34','g8e7 c3e2 f5d3 d1d3 c6c5 e1g1 e7f5 c2c3 b8c6 c1d2 a7a5 h2h3 c5d4 f3d4 f5d4 c3d4 f8b4 d2e3 e8g8 e2f4 a5a4 a2a3 b4e7 f4h5 g8h8 d3d1 d8d7 d1d2 h8h7 d2d3 h7h8 e3d2 c6a5 d2a5 a8a5 a1c1 d7b5');
INSERT INTO evaluations VALUES(34782,30,'cp 223',NULL);
INSERT INTO evaluations VALUES(34782,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(34782,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(34782,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(34782,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(34782,35,'cp 230',NULL);
INSERT INTO evaluations VALUES(34782,36,'cp 230','d2d4 b8c6 c1g5 f6g6 h2h4 f7f6 h4h5 g6f7 g5e3 a7a6 d1d2 g8h6 b2b3 e5d4 f3d4 c6d4 d2d4 c7c6 e1c1 f8e7 c3a4 c8e6 a4b6 a8d8 e3h6 g7h6 f2f4 f7g7 f1c4 e6c4 b6c4 e8g8 c4e3 d6d5 e4d5 c6d5 e3d5 e7a3 c1b1 g8h8 d4c3');
INSERT INTO evaluations VALUES(38270,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(38270,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(38270,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(38270,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(38270,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(38270,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(38270,36,'cp 84','b1c3 c7c5 d4d5 e7f6 f1d3 f6c3 b2c3 g8f6 g1e2 b8d7 d3b1 e6e5 e2g3 d7f8 g3f1 f8g6 h2h4 h7h6 b1c2 c8g4 f2f3 g4d7 a2a4 e8g8 e1f2 f6h7 h4h5 g6e7 g2g4 g8h8 f2g2 b7b6 f1g3 e7g8 d1e2 f8e8 e2d2 d8e7 a1b1 a8c8');
INSERT INTO evaluations VALUES(38268,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(38268,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(38268,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(38268,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(38268,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(38268,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(38268,36,'cp 87','c7c5 d4d5 e7f6 g1e2 g8e7 e2g3 e8g8 f1d3 b8d7 e1g1 f6c3 b2c3 e6e5 a2a4 b7b6 f2f3 h7h6 f1f2 a7a5 c1e3 d7f6 d1d2 g8h8 a1b1 g7g5 g3f5 e7f5 e4f5 f6h5 h2h4 h5f4 g2g4 h8g7 f2h2 f8h8 h4g5 h6g5 h2h8 d8h8 b1b6 h8h3 f5f6 g7g8');
INSERT INTO evaluations VALUES(40606,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(40606,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(40606,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40606,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(40606,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(40606,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40606,36,'cp 0','a2a3 e8g8 b2b4 c5d6 c4d5 e6d5 c1b2 a7a5 b4a5 d8a5 b2c3 a5d8 f1d3 b8c6 e1g1 f6e4 c3b2 c8f5 b1c3 c6e5 d3b1 f5g6 c3e4 e5f3 g2f3 d5e4 d1d4 d6h2 g1h2 d8h4 h2g1 h4g5 g1h1 g5h5 h1g1');
INSERT INTO evaluations VALUES(44973,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(44973,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(44973,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44973,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(44973,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(44973,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(44973,36,'cp -13','e4e5 d5d4 e5f6 d4c3 b2c3 e7f6 d3d4 f8d6 f1d3 e8g8 d1f3 c6c5 g1e2 b8c6 e1g1 c5d4 c3d4 c6b4 c1f4 d6f4 e2f4 b4d3 f3d3 d8d6 f4e2 b7b5 f1e1 c8d7 e2c3 a8c8 c3b5 d7b5 d3b5 d6d4 a1d1 d4c3 b5a6 g7g6 a6a7 f8e8 e1e8 c8e8');
INSERT INTO evaluations VALUES(45707,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(45707,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(45707,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(45707,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(45707,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(45707,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(45707,36,'cp 192','g1f3 b7b5 c4d3 b5b4 c3a4 c8a6 d3a6 b8a6 e1g1 f8e7 c1f4 d8a5 f3e5 a5b5 a1c1 a8c8 d4d5 e6d5 e4d5 f6d5 d1g4 e8g8 f4h6 e7f6 e5d7 g8h8 d7f6 g7h6 f6h5 f8g8 g4d4 f7f6 h5f6 d5f6 d4f6 g8g7 b2b3 b5g5 f6f3 g5g6 c1c4 g7f7 f3e2 a6c7 c4g4 g6f6 g4b4 c7e6 e2b2 c8d8 b2f6 f7f6 b4e4 d8d2');
INSERT INTO evaluations VALUES(45706,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(45706,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(45706,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(45706,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(45706,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(45706,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(45706,36,'cp 141','b7b5 c4d3 b5b4 c3a4 c8a6 d3a6 b8a6 e1g1 a8c8 f3e5 a6b8 c1g5 f8e7 a1c1 e8g8 f1e1 h7h6 g5f4 e7d6 d1b3 f6h5 f4e3 h5f6 a4c5 d6c5 d4c5 f6e4 b3b4 e4f6 b4a4 f6d5 b2b4 d5e3 e1e3 d8c7 e5c4 b8d7 a2a3 d7f6 h2h3 f6d5 e3f3 c8d8 c4d6 d8d7 c1e1');
INSERT INTO evaluations VALUES(48482,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48482,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(48482,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(48482,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(48482,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(48482,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(48482,36,'cp 0','e2e3 c7c5 g1f3 c5d4 d1d4 d8d4 f3d4 a7a6 f1c4 b8d7 f4g3 h7h5 h2h4 f8b4 f2f3 b7b5 c4b3 c8b7 e1g1 e8g8 f1d1 f8d8 e3e4 d7c5 g3f2 b4a5 b3c2 a8c8 a2a3 c5d7 c2b3 d7c5');
INSERT INTO evaluations VALUES(48480,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48480,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(48480,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48480,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(48480,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48480,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48480,36,'cp 0','c7c5 g1f3 c5d4 d1d4 d8d4 f3d4 a7a6 f1c4 b8d7 f4g3 h7h5 h2h4 f8b4 f2f3 b7b5 c4b3 c8b7 e1g1 d7c5 f1d1 e8g8 e3e4 a8c8 b3c2 c5d7 c2b3');
INSERT INTO evaluations VALUES(48564,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(48564,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(48564,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(48564,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(48564,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(48564,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(48564,36,'cp -48','f6d5 d1e2 f8e7 e1g1 b8c6 b1c3 d5c3 d2c3 d8d6 b2b4 c8g4 b4b5 g4f3 f1f3 c6e5 f3f4 e5c4 f4c4 e8g8 g1h1 d6d7 c4d4 d7f5 d4e4 e7d6 a2a4 a8d8 h2h3 h7h6 c1d2 f5d7 d2f4 d6f4 e4f4 f8e8 e2c4 d7e6 a1f1 e6c4 f4c4 e8e7 c4d4 d8e8');
INSERT INTO evaluations VALUES(48978,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(48978,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(48978,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(48978,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(48978,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(48978,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(48978,36,'cp -36','d1e2 f8e7 e1g1 b8c6 b1c3 d5c3 d2c3 d8d6 b2b4 c8g4 b4b5 g4f3 f1f3 c6e5 f3f4 e5c4 f4c4 e8g8 a2a4 f8e8 c4e4 d6d7 c1e3 e7f6 e3d4 e8e4 e2e4 a8e8 e4d3 f6d4 c3d4 e8e6 a4a5 d7e7 h2h3 g7g6 a5a6 b7a6 b5a6 g8g7');
INSERT INTO evaluations VALUES(53637,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(53637,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(53637,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(53637,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(53637,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(53637,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(53637,36,'cp 74','c2c3 c6c5 f1d3 b8c6 g1f3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f5e3 f2e3 f8e7 b1c3 c8d7 e3e4 d8b6 c3a4 b6d8 a2a3 a8c8 a4c3 d5e4 d3e4 c6a5 d1e2 e8g8 g1h1 a5c4 f1d1 b7b5 d4d5 d8b6 d5d6 e7d8 d1e1 g7g6 e4d3 d7c6 a3a4 c6f3 e2f3 b5a4 c3a4 b6b4 d3c4 c8c4');
INSERT INTO evaluations VALUES(57478,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(57478,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(57478,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(57478,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(57478,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(57478,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(57478,36,'cp 147','f3e5 g8f6 e5g4 c7c6 g4f6 d8f6 c3e4 f6e6 d1e2 d7d5 e4g5 e6e2 f1e2 f8d6 e2h5 g7g6 h5d1 e8g8 e1g1 c6c5 c2c3 g8g7 g5f3 b8c6 d2d4 c8f5 d4c5 d6c5 d1b3 a8d8 c1g5 d8d7 g5e3 d5d4 e3d2 h7h6 c3d4 c6d4 d2c3 g7h7 f3d4 c5d4 c3d4 d7d4 a1d1 d4d3 d1d3');
INSERT INTO evaluations VALUES(2720,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(2720,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(2720,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2720,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(2720,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(2720,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(2720,36,'cp 23','c2c3 f8e8 f3g5 e8f8 d2d4 c8b7 g5f3 f8e8 b1d2 e7f8 a2a3 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 c2d3 f6h5 d3f1 h5f4 d1c2 a6a5 a1c1 f4e6 a3a4 b5a4 b3a4 b7a6 f1a6 a8a6 b2a3 e5d4 c3d4 d6d5 a3f8 d7f8 d2b3 e6f4 c2d2 f8e6 e4d5');
INSERT INTO evaluations VALUES(12763,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(12763,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(12763,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(12763,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12763,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(12763,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(12763,36,'cp 0','d8a5 b1c3 c6d4 b5c4 b7b5 c4b3 d4b3 c2b3 b5b4 c3a4 e7e6 c1g5 c8a6 e1g1 a5b5 f3e5 d7d6 g5f6 g7f6 e5c4 h8g8 f1e1 a8d8 d1h5 a6b7 h5h7 g8g6 e1e3 g6h6 h7g8 h6g6');
INSERT INTO evaluations VALUES(18659,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(18659,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(18659,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(18659,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(18659,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(18659,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(18659,36,'cp 107','e1g1 g8f6 f1e1 d8c7 b1a3 a7a6 c4f1 b7b5 a3c2 f8e7 d2d4 d7d6 d4d5 c6a5 b2b3 c5c4 c2a3 e8g8 b3c4 b5c4 f3d2 c8d7 a1b1 f8b8 b1b4 a5b7 a3c4 a6a5 b4b6 b7c5 b6b8 a8b8 c1a3 d7b5 d1e2 e7f8 h2h3 c5a4 e1c1 a4c5 a3c5 c7c5');
INSERT INTO evaluations VALUES(18658,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(18658,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(18658,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(18658,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(18658,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(18658,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(18658,36,'cp 104','g8f6 f1e1 d8c7 b1a3 a7a6 c4f1 d7d6 d2d4 b7b5 b2b3 e5d4 c3d4 f8e7 c1b2 e8g8 a3c2 c8g4 a1c1 g4h5 d1d3 d6d5 e4d5 f6d5 d4c5 a8d8 c2d4 c6b4 d3e4 b4a2 c1a1 h5g6 e4e5 c7e5 f3e5 a2b4 e5g6 f7g6');
INSERT INTO evaluations VALUES(12279,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(12279,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(12279,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(12279,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(12279,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(12279,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(12279,36,'cp 88','f1e1 d8c7 b1a3 a7a6 c4f1 b7b5 a3c2 f8e7 a2a4 b5a4 d2d4 c5d4 c3d4 e8g8 a1a4 d7d5 e4d5 f6d5 f3e5 c6e5 d4e5 c8e6 c2d4 e7c5 d1d3 d5b6 a4a1 e6c4 d3e4 c5d4 e4d4 c4f1 e1f1 f8d8 d4e3 d8e8 f2f4 b6d5 e3f2 c7d7');
INSERT INTO evaluations VALUES(25533,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(25533,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(25533,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(25533,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(25533,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(25533,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(25533,36,'cp 174','b1c3 h7h6 h2h3 f8e7 a2a3 e8g8 e1g1 f8e8 b5c4 e7f8 f1e1 c6a5 c4a2 c7c5 d4c5 d6c5 c1f4 a5c6 e4e5 f6h7 d1c2 c6d4 f3d4 c5d4 c3e2 h7g5 e2d4 g5e6 a2e6 d7e6 d4e6 e8e6 a1d1 d8b6 d1d7 a8d8 f4e3 e6e5 e1d1 d8d7 e3b6 d7d1 c2d1 a7b6 d1d4 e5e6 b2b4');
INSERT INTO evaluations VALUES(25531,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(25531,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(25531,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(25531,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(25531,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(25531,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(25531,36,'cp 188','h7h6 h2h3 f8e7 e1g1 e8g8 b5d3 f8e8 f1e1 c6b4 d3c4 d6d5 e4d5 b7b5 c3b5 b4d5 d1d3 c7c6 b5c3 e7d6 e1e8 d8e8 c1d2 d5b4 d3f1 d7f5 a2a3 b4d5 a1e1 e8b8 b2b4 d5c3 d2c3 f5e4 f3e5 e4d5 f1d3 b8c7 e5g4 f6g4 h3g4 c7d7 c4d5');
INSERT INTO evaluations VALUES(28082,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(28082,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(28082,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(28082,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(28082,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(28082,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(28082,36,'cp 62','d1b3 d8e7 a2a3 b4a5 c1d2 g8f6 f3h4 a5c3 d2c3 f5e4 h4f3 a7a5 f3e5 e8g8 f2f3 d5c4 f1c4 e4d5 e3e4 b7b5 c4d5 e6d5 e1g1 f8e8 a1e1 e7e6 e5d3 d5e4 b3e6 e8e6 d3c5 e6e8 c5e4 f6e4 f3e4 b8d7 d4d5 c6d5 e4d5 a5a4 d5d6 e8e1 f1e1 f7f6');
INSERT INTO evaluations VALUES(37803,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(37803,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(37803,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(37803,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(37803,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(37803,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(37803,36,'cp 100','d4d5 d7d6 g1f3 g8f6 f1c4 c8g4 h2h3 g4f3 d1f3 b8d7 c4e2 a7a6 a2a4 d8c7 e1g1 e8g8 f1d1 c5c4 f3e3 c7c5 a4a5 c5c7 c1d2 b7b5 a5b6 d7b6 c3a4 f6d7 d2a5 f8b8 f2f4 c7a7 a4b6 d7b6 e4e5 d6e5 f4e5 b6d7 e3a7');
INSERT INTO evaluations VALUES(37841,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(37841,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(37841,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(37841,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(37841,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(37841,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(37841,36,'cp 87','d7d6 g1f3 g8f6 f1c4 e8g8 e1g1 b8a6 a2a4 a6c7 d1d3 a7a6 c1f4 c8d7 a4a5 c7b5 c3a4 d8a5 a4c5 a5c7 c5d7 f6d7 c4b5 a6b5 d3b5 d7f6 b5d3 f6h5 a1a8 f8a8 f4c1 h5f6 c1e3 a8a4 b2b3 a4a2 c2c4 f6d7 e3d4 g7d4 f3d4 c7a5 f1b1');
INSERT INTO evaluations VALUES(37987,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(37987,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(37987,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(37987,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(37987,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(37987,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(37987,36,'cp 101','f1c4 b8a6 g1f3 g8f6 a2a4 e8g8 e1g1 a6c7 c1f4 f6h5 f4e3 b7b6 f1e1 e7e5 f3d2 h5f6 f2f3 c8d7 c4f1 f6e8 d1b1 f7f5 b2b4 g7f6 b4c5 b6c5 e3f2 a8b8 b1d1 f6h4 f2h4 d8h4 f1d3 h4h5 d2c4 h5g5 d3f1 g5e7 d1d2');
INSERT INTO evaluations VALUES(38587,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(38587,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(38587,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(38587,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(38587,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(38587,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(38587,36,'cp 7','e1g1 g8f6 f1e1 e8g8 c2c3 c7c6 h2h3 c5b6 b1d2 b8d7 c4b3 f8e8 d3d4 d7f8 b3c2 b6c7 d2f1 d6d5 d4e5 f6e4 f1g3 f8g6 g3e4 d5e4 d1d8 e8d8 c2e4 g6e5 f3e5 c7e5 e4c6 e5c3 b2c3 b7c6 c1e3 c8e6 e3d4');
INSERT INTO evaluations VALUES(39775,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(39775,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(39775,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(39775,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(39775,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(39775,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(39775,36,'cp 32','h7h6 g5f4 d5c4 f1c4 b7b5 c4e2 b8d7 f3e5 d7e5 f4e5 e8g8 e2f3 a8b8 e5g3 e7d6 a1c1 d8e7 g3h4 f8d8 c3e4 g7g5 e4d6 c7d6 h4g3 c8b7 e1g1 b8c8 c1c8 b7c8 h2h3 c8b7 f3e2 f6e4 g3h2 e4f6 f2f3 e6e5 d1d2 e5d4 e3d4');
INSERT INTO evaluations VALUES(42410,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(42410,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(42410,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(42410,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(42410,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(42410,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(42410,36,'cp 88','e8g8 c3d5 b4d2 d1d2 e6d5 f2f4 d7d6 e1c1 f7f6 g1f3 b8c6 f1b5 c8f5 h1e1 f5e4 d2e3 d8e7 f3d2 e4f5 b5c6 b7c6 h2h3 f6e5 d4e5 d6e5 g2g4 f5d7 f4e5 c6c5 e5e6 d7a4 e3e5 a8d8 d2b3 a4b3 a2b3 d5d4 g4g5 d8e8 h3h4 e7d6 e6e7 d6e5');
INSERT INTO evaluations VALUES(43830,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43830,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43830,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43830,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43830,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43830,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(43830,36,'cp 14','c8b7 b1c3 d7d5 c4d5 f6d5 e2e3 f8e7 c3d5 d8d5 f1d3 c7c5 e3e4 d5d7 d4c5 e8g8 c5b6 a7b6 d1e2 f8d8 d3b5 e7b4 c1d2 b4d2 e2d2 d7d2 f3d2 b8c6 a1d1 c6d4 b5d3 b7c6 d1c1 b6b5 e1g1 b5b4 a3b4 a8b8 d2f3 b8b4 f3d4 b4d4 c1c6 d4d3 f2f3 h7h6 b2b4 d3d2');
INSERT INTO evaluations VALUES(28726,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(28726,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(28726,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(28726,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(28726,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28726,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(28726,36,'cp 15','b1c3 d7d5 c4d5 f6d5 e2e3 f8e7 f1b5 c7c6 b5d3 c6c5 c3d5 d8d5 e3e4 d5d6 d4c5 b6c5 c1e3 e8g8 a1c1 b8d7 e1g1 f8d8 d1e2 a8c8 f1d1 h7h6 e3d2 d6c7 c1c2 d7f8 h2h4 f8d7 d1c1 e7d6 d3a6 b7a6 e2a6 c7b8 d2a5');
INSERT INTO evaluations VALUES(43876,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(43876,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(43876,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(43876,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(43876,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(43876,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(43876,36,'cp 97','e2e4 c8b7 e4e5 f6d5 c3d5 b7d5 c1g5 h7h6 g5h4 d7d6 g1f3 d8d7 f1d3 e7e6 e1g1 f8e7 h4g3 e8g8 c2c3 b8c6 d3c2 d6e5 d4e5 a8d8 d1e2 c6a5 f1d1 d7c8 f3d4 c8b7 e2g4 a5c6 d4b5 a7a6 c3c4 a6b5');
INSERT INTO evaluations VALUES(44838,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(44838,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(44838,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(44838,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(44838,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(44838,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(44838,36,'cp 136','b7c6 b5e2 e7e5 d2d3 d8c7 b1c3 f8e7 g1f3 c6c5 f3d2 b8c6 e2f3 e8g8 b2b3 a8b8 e1g1 c6d4 f1e1 c8e6 d2c4 h7h6 g2g3 f8e8 c1d2 e7d6 f3g2 e6g4 f2f3 g4e6 c3e4 f6e4 f3e4 d4c6 c4e3 d6e7 c2c3 e8d8');
INSERT INTO evaluations VALUES(43930,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(43930,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(43930,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(43930,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(43930,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(43930,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(43930,36,'cp 139','b5e2 e7e5 d2d3 d8c7 b1c3 f8e7 g1f3 c6c5 f3d2 b8c6 e2f3 e8g8 b2b3 c8e6 f3c6 c7c6 d1f3 c6a6 c1b2 a8d8 e1g1 f6d5 c3d5 e6d5 f3e3 f7f6 f2f3 d8d7 e3f2 d5f7 f3f4 e5f4 f2f4 d7d6 f4g3 d6e6 g3f2 f8d8 d2c4 h7h6');
INSERT INTO evaluations VALUES(45766,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(45766,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(45766,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(45766,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(45766,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(45766,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(45766,36,'cp -26','f8d6 d1c2 e8g8 f1d3 f8e8 e1g1 d8e7 f1e1 f6e4 f3e5 d6e5 d4e5 e7e5 f2f3 e4c3 d3h7 g8h8 b2c3 b8d7 h7d3 d7c5 c1a3 c5d3 c2d3 h8g8 d3d4 b7b6 d4e5 e8e5 a3d6 e5e6 d6f4 c8a6 a2a4 a6c4 g2g4 b6b5 h2h4 a7a5 e1b1 e6e8');
INSERT INTO evaluations VALUES(40517,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(40517,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(40517,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(40517,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(40517,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(40517,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(40517,36,'cp -20','d1c2 e8g8 f1d3 f8e8 e1g1 d8e7 f1e1 f6e4 f3e5 d6e5 d4e5 e7e5 f2f3 e4c3 d3h7 g8h8 b2c3 b8d7 h7d3 d7c5 c1a3 c5d3 c2d3 e5g5 a3d6 b7b6 d6f4 c8f5 f4g5 f5d3 g1f2 a7a5 g5f4 e8e6 h2h4 b6b5 a2a3 d3c4 e3e4 d5e4 e1e4 e6e4 f3e4 a5a4 f4e3');
INSERT INTO evaluations VALUES(48478,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(48478,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(48478,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(48478,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(48478,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(48478,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(48478,36,'cp -60','g1e2 f8e7 g2g3 e6e5 f1g2 e5d4 e2d4 e8g8 e1g1 c8g4 d1d2 e4e3 f2e3 c7c6 h2h3 g4h5 d4f5 b8d7 b2b3 f8e8 c1b2 h5g6 a1d1 d8c7 f5e7 e8e7 g3g4 h7h6 d2d4 c7a5 e3e4 a8e8 d4d2 a5b6 d2f2 b6f2 f1f2 d7c5 b2a3 e7e5 a3b2 g6e4 c3e4 f6e4');
INSERT INTO evaluations VALUES(49840,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(49840,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(49840,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(49840,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(49840,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(49840,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(49840,36,'cp 72','g4e5 f2f4 e5c6 b1c3 f8c5 d1g4 e8g8 f4f5 d7d6 g1f3 c5b4 c1d2 b8d7 e1c1 f7f6 d2h6 f8f7 c3d5 c6e5 g4g3 b4c5 f1e2 c7c6 d5c3 g8h8 h6f4 c5b4 f4d2 f7e7 a2a3 b4c5 c1b1 d8b6 f3e5 d6e5 c3a4 b6c7 a4c5 d7c5');
INSERT INTO evaluations VALUES(49871,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(49871,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(49871,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(49871,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(49871,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(49871,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(49871,36,'cp 66','f2f4 e5c6 a2a3 d7d6 g1f3 c8g4 f1e2 g7g6 c1d2 f8h6 b1c3 e8g8 e1g1 b8d7 a1c1 h6g7 d2e1 g4f3 e2f3 c6d4 e1f2 d4f3 d1f3 c7c6 c1d1 d8e7 f2d4 a8d8 d4g7 g8g7 d1d4 a7a5 f3f2 g7g8 f1d1 d7f6 h2h3 f8e8');
INSERT INTO evaluations VALUES(49872,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(49872,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(49872,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(49872,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(49872,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(49872,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(49872,36,'cp 64','e5c6 a2a3 d7d6 g1f3 g7g6 c1d2 f8h6 f1e2 c8g4 b1c3 b8d7 e1g1 e8g8 a1c1 f8e8 h2h3 g4f3 e2f3 c6d4 d2e3 d4f3 d1f3 c7c6 c1d1 d8e7 f1e1 h6g7 e3f2 a7a5 c3a4 g7f8 d1d2 a8d8 b2b3 d8b8 d2d1 b8d8 d1d3 f8g7 e1d1');
INSERT INTO evaluations VALUES(52964,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(52964,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52964,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(52964,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(52964,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(52964,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(52964,36,'cp 34','g8f6 b1c3 d6d5 d1d3 d8a5 f3d2 b8d7 e2e4 d5e4 c3e4 f6e4 d3e4 d7f6 e4d3 g7g6 c2c3 c8f5 d2c4 f5d3 c4a5 d3f1 h1f1 e8c8 e1c1 f6d5 f4g3 f8g7 f1e1 h7h5 h2h4 g7f6 c1c2 d8d7 a5c4 e7e6 a2a4');
INSERT INTO evaluations VALUES(45099,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(45099,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(45099,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(45099,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(45099,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(45099,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(45099,36,'cp 43','b1c3 d8a5 f4d2 b8d7 e2e4 e7e5 f1d3 a5c7 a2a4 f8e7 h2h3 e8g8 e1g1 f8e8 f1e1 h7h6 a4a5 e5d4 f3d4 d7e5 d3f1 c8d7 f2f4 e5g6 a5a6 a8b8 a6b7 c7b7 d2e3 e7d8 d4f3 d8b6 d1d6 b6e3 e1e3 b7b6 a1e1 f6e4 d6d7 e4c3 d7d4 b6d4 f3d4 e8e3 e1e3 c3d5');
INSERT INTO evaluations VALUES(53161,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(53161,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(53161,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(53161,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(53161,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(53161,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(53161,36,'cp 95','d1d8 e8d8 g1f3 g8f6 b1c3 b8c6 e2e4 e7e6 f1c4 f8c5 e1g1 h7h6 b2b3 d8e7 c1b2 h8d8 c3a4 c5b6 a4b6 a7b6 f1e1 c8d7 b2c3 e7f8 c4f1 d8c8 f3e5 c6e5 c3e5 b6b5 f2f3 f8g8 e5f4 c8c2 e1e2 a8c8 a1d1 c2e2 f1e2');
INSERT INTO evaluations VALUES(35399,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(35399,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(35399,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(35399,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(35399,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(35399,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(35399,36,'cp 89','e8d8 g1f3 g8f6 b1c3 b8c6 e2e4 e7e6 f1c4 f8c5 e1g1 h7h6 c3a4 c5e7 f1d1 d8e8 a4c3 a7a6 c1e3 f6g4 e3b6 c8d7 c4b3 e7d8 c3a4 g4f6 e4e5 f6d5 b6d8 a8d8 a4c5 d7c8 c5e4 e8e7 e4d6 g7g5 b3d5 e6d5 d1d5');
INSERT INTO evaluations VALUES(35435,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(35435,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(35435,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(35435,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(35435,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(35435,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(35435,36,'cp 91','g1f3 g8f6 b1c3 b8c6 e2e4 e7e6 f1c4 f8c5 e1g1 h7h6 c3a4 c5e7 f1d1 d8e8 a4c3 g7g5 c4b5 g5g4 f3e5 c8d7 c1f4 a8c8 a1c1 a7a6 b5c6 d7c6 e5c6 c8c6 e4e5 f6h5 f4d2 f7f6 c3e4 f6e5 c1c6 b7c6 d2c3 e8f7 d1d7 h8d8 e4d6 f7f8 d7d8 e7d8 c3e5');
INSERT INTO evaluations VALUES(925,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(925,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(925,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(925,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(925,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(925,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(925,36,'cp 0','d2d4 f6d7 d1e2 d7e5 d4e5 e7e5 b1c3 c8d7 f2f4 e5a5 c1d2 e8c8 e1c1 d7e6 c1b1 f8b4 a2a3 b4a3 b2a3 a5a3 e2e3 d8d2 e3d2 a3b4 b1c1 b4a3');
INSERT INTO evaluations VALUES(924,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(924,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(924,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(924,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(924,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(924,36,'cp 0','f6d7 f2f4 d7e5 d4e5 e7h4 g2g3 h4h3 e1f2 g7g5 d1f1 g5f4 f1h3 c8h3 c1f4 e8c8 b1d2 f8c5 f2e2 h3g4 e2e1 g4h3');
INSERT INTO evaluations VALUES(1620,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(1620,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(1620,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(1620,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(1620,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(1620,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(1620,36,'cp 34','e5e2 f1e2 c8f5 e2b5 e7e6 f3d4 g8e7 d4f5 e7f5 b5c6 b7c6 d2d3 h7h5 c3e4 f7f6 c1d2 f8e7 h2h4 e6e5 e1c1 a8d8 g2g3 e8f7 d1e1 h8e8 h1f1 e7d6 d2a5 d8d7 a5c3 d7d8 e4d2 d6c5 e1e4 f5d6 e4a4 c5b6 a4a6 d8a8 d2e4');
INSERT INTO evaluations VALUES(1639,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(1639,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(1639,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(1639,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(1639,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(1639,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(1639,36,'cp 27','f1e2 c8f5 e2b5 e7e6 f3d4 g8e7 d4f5 e7f5 b5c6 b7c6 d2d3 h7h5 c1d2 f7f6 e1c1 e8f7 h2h4 f8e7 g2g3 h8d8 c3e4 a7a5 a2a4 d8d4 b2b3 e7a3 c1b1 d4d5 h1e1 a8d8 b1a2 a3e7 a2b2 f5d4 d2c3 e6e5 e4d2 e7c5 d2c4 d4f3 e1e2 c5b6 c3d2');
INSERT INTO evaluations VALUES(15610,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15610,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(15610,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(15610,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(15610,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(15610,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(15610,36,'cp 44','f1e2 g7g6 a2a4 f8g7 e1g1 e8g8 f4e5 c6e5 f3e5 d6e5 a4a5 c8e6 c1e3 d8d6 b1c3 c7c6 d1e1 f6d7 e1f2 a7a6 c3a4 f7f5 e4f5 g6f5 g1h1 g8h8 e2f3 e6d5 a4b6 e5e4 f3e2 d7b6 e3b6 d6e5 c2c4 d5e6 d3e4 f5f4');
INSERT INTO evaluations VALUES(16559,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(16559,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(16559,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(16559,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(16559,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(16559,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(16559,36,'cp -16','c2c4 f8c5 g2g3 d7d6 b1c3 a7a6 g1e2 b7b5 f1g2 a8b8 e1g1 e8g8 b2b3 c8d7 g1h2 c6d4 a1b1 f8e8 c1b2 b5c4 d3c4 a6a5 d1d3 h7h5 e2d4 c5d4 c3b5 d4b2 b1b2 h5h4 b2c2 f6h7 b5c3 h4g3 f2g3 h7g5 d3e2');
INSERT INTO evaluations VALUES(23739,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(23739,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23739,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(23739,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(23739,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(23739,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(23739,36,'cp 25','c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 f1e1 h7h6 c2c3 d8d7 c1d2 d7e8 a2a4 e8g6 d2e3 f8f7 b2b4 a8f8 f3h4 g6h5 d1h5 f6h5 h4f3 h5f4 e3f4 f7f4 e1d1 g7g5 d1d2 g8g7 f3h2 h6h5 h2f3 g5g4 h3g4 h5g4 f3h2 f8a8 g2g3');
INSERT INTO evaluations VALUES(23735,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(23735,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(23735,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(23735,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(23735,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(23735,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(23735,36,'cp 24','f7e6 c3a4 c5b6 c1d2 d8e8 a4b6 a7b6 a2a4 e8h5 b2b4 h7h6 c2c3 f8f7 d1b3 f6e4 d3e4 f7f3 b3e6 h5f7 e6d5 f3f6 f2f3 f7d5 e4d5 c6e7 c3c4 f6f8 a4a5 e7f5 f1b1 b6a5 b4a5 b7b6 a5b6 c7b6 g1f2 a8a1 b1a1 f8c8 a1a4 f5d4 a4b4 d4f5');
INSERT INTO evaluations VALUES(23746,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(23746,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(23746,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(23746,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(23746,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(23746,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(23746,36,'cp 27','c3a4 c5b6 c2c3 c6e7 a4b6 a7b6 a2a4 e7g6 f1e1 h7h6 d3d4 f6d7 d4e5 g6e5 f3e5 d7e5 f2f4 e5g6 g2g3 g6e7 c1e3 d8e8 e3f2 e7c6 d1e2 e8g6 g1h2 g6f7 f2e3 c6b8 e2c4 b8c6 b2b4 f7d7 e3f2 d7f7 c4e2 c6e5 h2g2 e5d7 c3c4');
INSERT INTO evaluations VALUES(24732,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24732,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24732,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(24732,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24732,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(24732,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(24732,36,'cp 26','h2h3 g4h5 c1f4 f8d6 f4d6 d8d6 f1e1 e8g8 b1d2 c6e7 a2a4 f8c8 a4a5 d6c7 d1c1 e7c6 a5a6 b7b6 d3b5 h5f3 d2f3 c6a5 c1c2 a5c4 f3e5 c4d6 c2d3 d6b5 d3b5 f6e4 b5e2 h7h6 e2e3 e4d6 e5d3 a8b8 h3h4 d6c4 e3e2 b6b5 g2g4 c4d6');
INSERT INTO evaluations VALUES(24729,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24729,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(24729,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24729,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24729,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24729,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24729,36,'cp 30','g4h5 c1f4 f8d6 f4d6 d8d6 f1e1 f6d7 b1d2 e8g8 e1e3 c6e7 a2a4 h7h6 a4a5 a8c8 d2b3 c8c7 d1d2 h5g6 b3c1 g6d3 c1d3 d7f6 e3e1 f6e4 d2e2 e7f5 f3d2 e4d2 e2d2 d6d7 g2g3 d7b5 d3c5 f8c8 d2e2');
INSERT INTO evaluations VALUES(24717,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24717,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24717,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(24717,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(24717,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(24717,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(24717,36,'cp 25','c1f4 f8d6 f4d6 d8d6 b1d2 e8g8 f1e1 f6d7 d1c2 h5g6 d3g6 h7g6 c2d3 f8e8 a1d1 e8e7 d2f1 d7f6 f1h2 e7c7 d3e2 a8c8 h3h4 a7a5 g2g3 a5a4 a2a3 c6a5 f3e5 d6a6 e5d3 a5c4 h2f3 f6e4 g1g2 c4d6');
INSERT INTO evaluations VALUES(34117,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(34117,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(34117,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(34117,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(34117,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(34117,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(34117,36,'cp 126','e6d6 c3e4 d6d8 d2d4 g8f6 d1d3 f6e4 d3e4 b8d7 c2c4 d7f6 e4c2 g7g6 e1g1 f8g7 c1f4 c8f5 c2d2 c7c5 a1d1 e8g8 h2h3 d8b6 d4c5 b6c5 b2b4 c5c8 f3d4 f6e4 d2e3 e7e5 d4f5 g6f5 f4h2 f8d8 d1d8 c8d8 g2g4 d8d4 g4f5 d4e3 f2e3');
INSERT INTO evaluations VALUES(34143,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(34143,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(34143,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(34143,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(34143,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(34143,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(34143,36,'cp 12','f8b4 c1d2 g8f6 f1g2 b8c6 d4c6 d7c6 f2f4 b7b5 a2a3 b4e7 e4e5 f6d5 c3d5 c6d5 d2b4 c8d7 b4e7 e8e7 e1g1 h7h5 d1d2 a6a5 c2c3 c7c5 f1f2 g7g6 a1d1 a8c8 d2d4 c5d4 d1d4 f7f6 e5f6 e7f6 f2d2 c8a8 d2d1 h8c8 g1f2');
INSERT INTO evaluations VALUES(43159,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(43159,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43159,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43159,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(43159,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(43159,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(43159,36,'cp 16','d2d4 d7d5 g1f3 e8g8 c4d5 f6d5 f1c4 d5b6 c4b3 c7c5 e1g1 c8g4 d4c5 b6d7 h2h3 g4f3 d1f3 d7c5 f1d1 b8d7 f3e2 g7c3 b2c3 d8c7 c1b2 f8d8 c3c4 a8c8 d1d4 c5e6 d4d2 d7f6 a1d1 b7b6 d2d8 c8d8 d1d8 c7d8 b3d1 e6c5 e2c2 d8c7 d1e2 f6d7 c2d1 e7e6 e2f3 e6e5 b2a3');
INSERT INTO evaluations VALUES(47060,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(47060,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(47060,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(47060,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(47060,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(47060,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(47060,36,'cp 31','e7e5 c2c4 c7c5 b1c3 f8e7 e1g1 e8g8 a1b1 b8c6 d2d3 h7h6 a2a3 a7a6 f3d2 a8b8 b2b4 c8e6 b4c5 d6c5 c1b2 d8d7 b2a1 e6h3 d2e4 f6e4 d3e4 b7b5 c4b5 a6b5 b1b5 b8b5 c3b5 d7d1 f1d1 h3g4 g2f3 g4f3 e2f3 f8b8 b5c7 e7g5 f3f4');
INSERT INTO evaluations VALUES(47915,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(47915,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(47915,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(47915,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(47915,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(47915,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(47915,36,'cp 42','c2c4 c7c5 e1g1 f8e7 d2d3 e8g8 b1c3 b8c6 f3e1 a7a6 e1c2 c8e6 c3d5 a8b8 b2b3 d8d7 e2e4 b7b5 c1g5 f6d5 e4d5 e7g5 d5c6 d7e7 c2e3 b5c4 d3c4 g5e3 f2e3 e5e4 g2e4 e6c4 e4h7 g8h7 b3c4 e7e3 g1h1 e3e4 d1f3 e4c4');
INSERT INTO evaluations VALUES(50865,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(50865,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(50865,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(50865,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(50865,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(50865,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(50865,36,'cp 86','d2d4 g8f6 f3e5 c8e6 c1e3 b8d7 f2f4 g7g6 d1f3 c7c6 f1d3 f8g7 e1g1 e8g8 a1e1 d7e5 d4e5 f6d5 c3d5 d8d5 f3f2 d5a2 e3c5 a2b2 e1b1 b2a2 c5e7 f8e8 b1b7 a6a5 f4f5 g6f5 d3f5 e6f5 f2f5 a2e6 e7d6 e6f5');
INSERT INTO evaluations VALUES(51634,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(51634,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(51634,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(51634,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(51634,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(51634,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(51634,36,'cp -34','c5b6 a7b6 g1f3 c8a6 a2a3 g8f6 b1c3 e7e6 c1g5 b8d7 e2e3 d8c8 f1d3 f8e7 g5f6 d7f6 e1g1 c8b7 a3a4 e8g8 d1e2 a6d3 e2d3 a8a6 b2b3 f8a8 a1c1 a6a5 c1c2 h7h6 f1a1 e7d6 a1b1 f6d7 h2h3 b7a6 d3a6');
INSERT INTO evaluations VALUES(6230,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(6230,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(6230,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6230,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(6230,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(6230,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(6230,36,'cp 36','c2c3 c5b6 d2d4 d7d6 d4e5 c8g4 c4b5 d6e5 d1d8 a8d8 b1d2 f7f6 a3a4 a7a6 b5e2 g8e7 d2c4 c6a5 c4a5 b6a5 f3d2 g4e2 e1e2 a5b6 b2b3 e8f7 d2c4 d8d7 c4b6 c7b6 c1e3 h8d8 h1d1 d7d1 a1d1 d8d1 e2d1 b6b5 a4b5 a6b5 f2f3 f7e6 d1c2 e7c8 e3f2');
INSERT INTO evaluations VALUES(6324,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(6324,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(6324,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(6324,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(6324,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6324,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(6324,36,'cp 50','d2d4 e5d4 c3d4 d7d6 h2h3 g8f6 e1g1 f6e4 f1e1 e8g8 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 c8f5 e4e1 d8d6 a4a5 b6c5 d1a4 c5b4 c1g5 f7f6 g5d2 b4c3 b2c3 e7d5 a4b3 c7c5 a1d1 f5e6 c3c4 d5b4');
INSERT INTO evaluations VALUES(6326,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(6326,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(6326,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(6326,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(6326,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6326,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(6326,36,'cp 53','e5d4 c3d4 d7d6 d4d5 c6e5 f3e5 d6e5 d1e2 g8f6 c1e3 c8d7 e1g1 c7c6 b1c3 b6e3 e2e3 d8b6 e3b6 a7b6 a2a4 c6d5 e4d5 e8e7 d5d6 e7d6 c4f7 e5e4 f2f3 h8f8 f7c4 d7c6 f1d1 d6c5 c4e2 a8e8 a1c1 e4f3 e2f3 c6f3');
INSERT INTO evaluations VALUES(6340,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(6340,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(6340,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(6340,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(6340,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(6340,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(6340,36,'cp 41','c3d4 d7d6 c4b5 c8d7 b1c3 g8e7 h2h3 e8g8 e1g1 h7h6 c1e3 f7f5 d4d5 c6e5 f3e5 d6e5 d5d6 c7d6 d1d6 d7b5 d6e6 g8h7 c3b5 f8f6 e6e5 e7g6 e5d5 b6e3 f2e3 f5f4 d5d8 a8d8 e3f4 g6f4 f1f2 f6b6 f2f4 b6b5 b2b3 b5c5 f4f2 d8d7 a1e1 c5e5 a2a3');
INSERT INTO evaluations VALUES(6336,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(6336,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(6336,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(6336,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(6336,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(6336,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(6336,36,'cp 40','d7d6 e1g1 g8f6 b1c3 e8g8 c4b3 c8g4 c1e3 f8e8 d1d3 g4f3 g2f3 f6h5 g1h1 d8f6 b3d1 c6e7 d3d2 d6d5 f1g1 g7g6 g1g5 a8d8 f3f4 h5g7 d1b3 c7c6 e4e5 f6e6 c3e4 d5e4 b3e6 g7e6 a1d1 d8d5 b2b3 e6g5 f4g5');
INSERT INTO evaluations VALUES(9465,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(9465,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(9465,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(9465,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(9465,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(9465,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(9465,36,'cp 159','c1g5 f7f6 g5d2 c6a5 c4b3 a5b3 a2b3 h7h5 h2h3 g4h6 d3d4 e5d4 c3a4 a7a6 a4c5 d6c5 f3h4 h6f7 f2f4 c8d7 b3b4 d7c6 h4g6 h8h7 f1e1 f7d6 f4f5 e8f7 g6f4 d8e8 b4c5 d6e4 c2c3 f7g8 c3d4 e8f7 a1c1 a8e8 f4g6 h5h4 d1g4');
INSERT INTO evaluations VALUES(14362,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(14362,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(14362,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(14362,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(14362,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(14362,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(14362,36,'cp -49','e7e5 d4e5 d8d1 e1d1 f6e4 d1e2 g7g5 c4d3 e4c5 f4g5 h8g8 g5f6 c5d3 e2d3 c6b4 d3d2 g8g2 b1c3 c8g4 g1e2 f8e7 a1g1 a8d8 d2c1 g2g1 h1g1 b4d3 c1c2 d3f2 e2f4 g4f5 c2b3 e7f6 e5f6 f2g4 e3e4 d8d4 f4d5 f5e6 c3b5 e6d5 e4d5 d4f4 g1e1 e8d8 e1e7 f4f6 d5d6 f6d6 e7f7 d6d3 b3a4 d3d7 f7f8 d8e7');
INSERT INTO evaluations VALUES(15368,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(15368,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(15368,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(15368,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(15368,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(15368,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(15368,36,'cp 44','e1g1 f8e7 b1d2 e4d2 c1d2 c8g4 h2h3 g4f3 d1f3 c6d4 f3h5 d4e6 a1e1 g7g6 d3g6 f7g6 h5e5 e8g8 e5e6 f8f7 e1e2 c7c6 f1e1 e7f6 d2h6 f6g7 h6g7 g8g7 a2a4 d8f6 e6f6 f7f6 g1f1 a8f8 f2f3 h7h5 a4a5 h5h4 f1f2 b7b5 e2e6 f6e6');
INSERT INTO evaluations VALUES(15367,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(15367,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(15367,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15367,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(15367,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(15367,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(15367,36,'cp 50','f8e7 b1d2 e4d2 c1d2 c8g4 c2c3 e8g8 h2h3 g4h5 g2g3 d8d7 g1g2 h5g6 d1c2 e7d6 a1e1 a7a6 d3g6 h7g6 h3h4 b7b5 f1h1 f8e8 c2d3 e8e1 f3e1 d7g4 h4h5 g6h5 h1h4 g4g6 d3f3 c6e7 e1d3 c7c6 h4h5 g6e4 h5h1 f7f6 d2f4 e7f5 h1e1 e4f3 g2f3');
INSERT INTO evaluations VALUES(8537,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(8537,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(8537,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(8537,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(8537,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(8537,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(8537,36,'cp 44','b1d2 e4d6 c2c3 c8f5 d3c2 f5c2 d1c2 e7f6 f1e1 c6e7 d2f1 e8g8 c1f4 c7c6 f4d6 d8d6 e1e2 d6d7 a1e1 a8e8 f1e3 e7c8 e3g4 c8d6 g4f6 g7f6 f3d2 e8e2 e1e2 f8e8 d2f1 e8e2 c2e2 d7e6 e2f3 g8f8 g2g3 f8e7 f1e3 e6e4 f3e4 d5e4 b2b3 e7d7 h2h3 f6f5 g1g2');
INSERT INTO evaluations VALUES(22988,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(22988,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(22988,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(22988,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(22988,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(22988,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(22988,36,'cp 33','f8b4 d1d3 h7h6 h2h4 b4c3 d3c3 d7d6 c1e3 g8f6 f3d2 f6g4 h4h5 g6e7 e1c1 e8g8 f2f3 g4e3 c3e3 f7f5 h1g1 f5f4 e3f2 g8h8 g2g4 e7g8 g4g5 h6g5 f2g2 h8h7 g2g5 d8g5 g1g5 g8f6 f1e2 c8d7 d1g1 f8f7 b2b3 h7h6 g5g2 a8f8 c1b2');
INSERT INTO evaluations VALUES(6810,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(6810,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(6810,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(6810,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(6810,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(6810,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(6810,36,'cp 66','d1d3 a7a6 h2h4 d7d6 h4h5 g6e7 h5h6 g8h6 c1h6 g7h6 a2a3 b4c5 g2g3 h6h5 h1h5 c8g4 h5h2 e7g6 e1c1 d8f6 f1e2 e8c8 d1h1 c8b8 c3d1 h7h5 c1b1 d8g8 f3h4 g6h4 h2h4 c5b6 d1e3 g4f3 h1h3 f3e2 d3e2 b6e3 e2e3 g8g5 e3d2 f6g6');
INSERT INTO evaluations VALUES(26091,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(26091,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26091,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26091,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26091,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26091,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26091,36,'cp 0','d1e2 g8f6 h2h4 h7h6 e1g1 f8e7 f3e5 d7e5 d4e5 f6d5 c4d5 d8d5 h4h5 g6h7 f1d1 d5a5 c2c4 a5c7 c1f4 e8g8 g3e4 h7e4 e2e4 a7a5 f4e3 a5a4 g2g3 f8d8 g1g2 c7a5 b2b3 e7f8 e4f4 a4b3 a2b3 a5a1 d1a1 a8a1 f4g4 g8h8 g4f4');
INSERT INTO evaluations VALUES(30218,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(30218,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(30218,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(30218,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(30218,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(30218,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(30218,36,'cp 13','c2c4 d5c4 d3c4 f8d6 f4g3 e8g8 e1g1 c7c5 b1c3 b8c6 d4c5 d6c5 c4e2 a8c8 a1c1 h7h6 d1b3 d8b6 b3b6 c5b6 f1d1 f6e4 c3e4 g6e4 f3d2 e4d5 d2c4 b6c7 g3c7 c8c7 a2a3 f8d8 f2f3 b7b6 g1f2 d5c4 d1d8 c6d8 c1c4 c7c4');
INSERT INTO evaluations VALUES(33230,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33230,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33230,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33230,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(33230,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(33230,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(33230,36,'cp 1','b8c6 b1c3 g8e7 c3d5 e7d5 e4d5 c6d4 f3d4 c5d4 b2d4 e5d4 f1d3 e8g8 e1g1 d7d6 f2f4 f8e8 d1f3 c8d7 h2h3 e8e7 a1e1 e7e1 f1e1 g7g6 a2a4 g8f8 e1f1 a8e8 g2g4 f6h4 g1h2 a7a5 f3g3 h4h6 f1f2 h6g7 g3h4 h7h6 d3c4 f7f5 h2g3 f5g4 h3g4');
INSERT INTO evaluations VALUES(35904,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(35904,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(35904,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(35904,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(35904,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(35904,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(35904,36,'cp 45','e2e4 c6c5 e1g1 c5d4 f3d4 c8g4 f2f3 g4d7 c1e3 b8c6 d4c6 d7c6 d1e2 a7a6 f1d1 d8a5 f3f4 a8c8 c4c5 c8d8 c5d6 e7d6 e3d4 f8e8 e2e3 a5b4 b2b3 f6g4 e3d2 g7d4 d2d4 b4d4 d1d4 g4e3 d4d3 e3c2');
INSERT INTO evaluations VALUES(35912,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(35912,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(35912,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(35912,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(35912,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(35912,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(35912,36,'cp 64','c6c5 e1g1 c5d4 f3d4 b8c6 d4e2 a7a6 b2b3 c8d7 a2a4 d8c8 c1e3 d7h3 d1d2 h3g2 g1g2 c8c7 a1d1 c7a5 c3d5 a5d2 d5f6 g7f6 d1d2 f8e8 f1d1 a8c8 f2f3 h7h5 g2f2 c6e5 e2f4 f6g5 h2h4 g5h6 f2e2 e5c6 f4d5 h6e3 e2e3 g8f8 f3f4');
INSERT INTO evaluations VALUES(39132,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(39132,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39132,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(39132,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(39132,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(39132,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(39132,36,'cp -16','d7d5 e4d5 e6d5 c4e2 g8f6 d2d4 c5d6 e1g1 h7h6 b1c3 a7a6 e2d3 e8g8 c3e2 f8e8 h2h3 c8d7 c1f4 d7b5 a2a4 b5d3 d1d3 b8c6 c2c3 c6e7 f1e1 e7g6 f4d6 d8d6 e2g3 d6f4 a4a5 f6e4 g3h5 f4d6 d3c2 c7c6');
INSERT INTO evaluations VALUES(43052,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(43052,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(43052,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(43052,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(43052,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(43052,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(43052,36,'cp 52','g5g7 e7f6 g7h6 d7d5 g1f3 d8e7 f1b5 c7c6 b5e2 c8g4 e1g1 h8g8 d2d3 e4c5 c2c3 b8d7 h6e3 e7e3 f2e3 e8c8 b1d2 d8e8 d2b3 f6g5 b3c5 d7c5 g1f2 g4f3 f2f3 g8g6 g2g3 g6f6 f3g2 f6f1 g2f1 g5e3 c1e3 e8e3 f1f2 e3e8 a1f1');
INSERT INTO evaluations VALUES(46647,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46647,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(46647,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(46647,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(46647,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(46647,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46647,36,'cp -14','d2d4 f8g7 f3e2 e8g8 g1f3 c7c5 d4d5 c8g4 e1g1 b8d7 b1d2 d7e5 h2h3 g4f3 d2f3 e5c4 e2c4 f8e8 c1g5 b7b5 c4d3 c5c4 d3e2 h7h6 g5e3 f6e4 e3h6 g7h6 e2e4 e7e5 d5e6 e8e6 e4d5 e6e2 d5d1 d8e7 d1d5 e7e4 d5e4 e2e4 f1e1 e4e1 a1e1');
INSERT INTO evaluations VALUES(47991,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(47991,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(47991,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(47991,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(47991,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(47991,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(47991,36,'cp 58','d8e7 d2d4 d6e5 d4e5 b8c6 b1c3 e7e5 e4e5 c6e5 c1f4 f8d6 f4g3 f7f6 e1c1 h7h5 h2h4 c8d7 c3b5 d7b5 f1b5 e8f7 h1e1 h8d8 e1e4 g7g6 e4d4 d6e7 c2c3 d8d4 c3d4 e5g4 b5c4 f7f8 d1d3 e7d6 g3d6 c7d6 f2f3 a8c8 d3c3 g4f2 a2a3 f8e7');
INSERT INTO evaluations VALUES(50101,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50101,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(50101,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(50101,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(50101,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(50101,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50101,36,'cp 15','d7d6 g1h3 e7e5 e1g1 e5e4 d3e2 d6d5 c2c4 c7c6 c1d2 b8a6 c4d5 c6d5 b1c3 a6c7 f2f3 f8d6 f3e4 f5e4 h3f4 e8g8 a1c1 c8d7 d1b3 d7c6 d2e1 d8d7 e1h4 g8h8 a2a4 a7a6 c1a1 a8c8 h4g3 a6a5 c3b5 c7b5');
INSERT INTO evaluations VALUES(51292,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(51292,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(51292,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(51292,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(51292,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51292,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(51292,36,'cp 56','b1c3 e7e6 c3b5 b8a6 g1f3 c7c6 b5c3 g8f6 e2e3 c8b7 h2h4 a6b8 f1d3 b8d7 h4h5 h7h6 f3e5 d7e5 f4e5 f6d7 e5f4 a7a6 d1f3 c6c5 c3e2 d7f6 f4e5 f8e7 c2c3 e8g8 f3g3 b6b5 e5f6 e7f6 d4c5 d8a5 a2a3 f6e7 e2d4 e7c5');
INSERT INTO evaluations VALUES(51916,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(51916,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(51916,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(51916,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(51916,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(51916,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(51916,36,'cp 42','b4c5 e6e5 d1h5 b8c6 f1c4 g7g6 h5d1 f8c5 g1f3 g8f6 b1c3 d7d6 e1g1 c6d4 h2h3 e8g8 d2d3 c8e6 c4e6 d4e6 a1b1 b7b6 c3d5 f6d5 e4d5 e6d4 f3d2 d4f5 d2e4 f7f6 a2a4 d8d7 a4a5 b6a5 g2g4 f5h4 e4c5 d6c5 c2c4 g6g5');
INSERT INTO evaluations VALUES(51892,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(51892,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(51892,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(51892,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(51892,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(51892,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51892,36,'cp 43','e6e5 d1h5 b8c6 f1c4 g7g6 h5d1 g8f6 b1c3 f8c5 g1f3 e8g8 e1g1 d7d6 h2h3 c6d4 d2d3 c8e6 c4e6 d4e6 a1b1 b7b6 c3a4 f6h5 c1h6 f8e8 a4c5 e6c5 c2c4 c5e6 g1h2 a8c8 f3g1 h5g7 g1e2 d8h4 h6e3 f7f5 f2f4 h4h5 b1b5 f5e4 d3e4 c8c4');
INSERT INTO evaluations VALUES(2986,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(2986,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(2986,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(2986,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(2986,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(2986,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(2986,36,'cp 38','h2h3 f6e4 g5f4 d7f6 f3e5 c8f5 d1c2 f5g6 e1g1 e4c3 b2c3 g6d3 c2d3 f6e4 a2a4 f7f6 e5f3 b7b6 f1c1 d8d7 f3d2 e4d2 d3d2 a7a6 d2a2 b6b5 a4b5 c6b5 e3e4 f8d8 e4d5 d7d5 c3c4 b5c4 c1c4 a6a5 f4g3 a5a4 c4a4 d5a2 a1a2 a8a4 a2a4');
INSERT INTO evaluations VALUES(22005,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(22005,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(22005,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(22005,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22005,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(22005,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(22005,36,'cp -20','g8h6 g5f3 d8f6 c4b5 f6g6 e1g1 c8g4 d1d3 g4f3 d3f3 f8e7 c2c3 e8g8 b5c6 b7c6 c3d4 d6d5 b1c3 d5e4 f3e4 h6f5 c1f4 a8b8 a1b1 f8e8 f1e1 h7h5 f4c7 b8c8 c7e5 f7f6 e5f4 e7d6 e4f3 f5d4 f3g3 e8e1 b1e1 g6g3 f2g3 d6b4 g1f1 d4b5 e1c1 b5c3 b2c3 b4a5 c3c4');
INSERT INTO evaluations VALUES(15674,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(15674,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(15674,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15674,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(15674,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(15674,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(15674,36,'cp 0','g5f3 h6g4 c4b5 c8d7 e1g1 f8e7 b5c6 b7c6 f3d4 e8g8 d4c6 d7c6 d1g4 e7f6 b1c3 a8b8 g4f3 f8e8 c3d5 c6d5 e4d5 f6b2 c1b2 b8b2 f3c3 b2b5 c3d3 b5b2 d3c3');
INSERT INTO evaluations VALUES(27022,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(27022,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(27022,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(27022,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(27022,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(27022,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(27022,36,'cp 55','f1b5 g8e7 h2h3 g4f3 d1f3 a7a6 b5a4 e5f4 f3f4 e7g6 f4f2 f8e7 e1g1 e8g8 d2d3 e7h4 g2g3 h4f6 c3d5 f6d4 c1e3 d4e3 d5e3 g6e5 g1g2 b7b5 a4b3 c6d4 a1d1 d4b3 a2b3 d8d7 e3f5 f7f6 h3h4 a6a5 d3d4 e5g4 f2f4 h7h5 d4d5');
INSERT INTO evaluations VALUES(26923,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(26923,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(26923,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(26923,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(26923,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(26923,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(26923,36,'cp 52','g8e7 h2h3 g4f3 d1f3 a7a6 b5a4 b7b5 a4b3 c6d4 f3f2 d4b3 a2b3 e5f4 f2f4 e7g6 f4f5 f8e7 a1a6 c7c6 a6a8 e7h4 e1e2 d8a8 d2d3 b5b4 c3d1 e8g8 c1d2 a8b7 h1f1 h4f6 f5g4 d6d5 e4d5 c6d5 d1e3 f8e8 f1f6 g7f6 e2f2 b7c6 g4b4 h7h5');
INSERT INTO evaluations VALUES(33289,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(33289,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(33289,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(33289,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(33289,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(33289,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(33289,36,'cp 40','d2d4 c5b6 g1f3 h4h5 b1c3 g8e7 c3e2 e7g6 c1f4 g6f4 e2f4 h5h6 f4d5 c7c6 d5b6 a7b6 a2a4 d7d6 f1f2 c8e6 c4d3 e8g8 d1d2 h6d2 f3d2 f7f5 f2g3 f5e4 d2e4 e6f5 h1e1 b8d7 e4d6 f5d3 c2d3 d7f6 e1e7 f6h5 g3g4 h5f6 g4h3 f6d5 e7b7');
INSERT INTO evaluations VALUES(38104,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(38104,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(38104,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(38104,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(38104,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(38104,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(38104,36,'cp -30','g1f3 g8f6 d2d4 e8g8 b1c3 a7a6 d4c5 d6c5 d1e2 b7b5 c4d3 c8b7 e1g1 b8d7 c3e4 c5e7 e4f6 e7f6 c2c3 d7c5 d3c2 c5a4 f3e5 d8d5 f2f4 a8d8 h2h3 d5c5 c1e3 c5c7 c2a4 b5a4 e3d4 d8d6 a1e1 f8d8 e2f2 f6e7 f2g3 a4a3');
INSERT INTO evaluations VALUES(38103,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(38103,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(38103,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(38103,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(38103,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(38103,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(38103,36,'cp -39','b8d7 d2d4 g8f6 e1g1 a7a6 d4c5 d6c5 c4d3 e8g8 d1e2 b7b5 a2a4 b5b4 c1f4 c8b7 b1d2 f6h5 f4e5 c5e7 d3e4 b7e4 d2e4 d7e5 f3e5 h5f4 e2f3 f4g6 e5c6 d8d5 c6e7 g6e7 c2c3 b4c3 b2c3 d5c6 f3e2 e7f5 f1e1 h7h6');
INSERT INTO evaluations VALUES(40904,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(40904,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(40904,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(40904,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(40904,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(40904,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(40904,36,'cp -31','e7f6 g1f3 e8g8 d1d2 c7c5 d4c5 d5e4 c3e4 f6b2 a1b1 b2f6 e4f6 d8f6 f1d3 b8d7 c5c6 d7c5 e1g1 b7c6 d3c4 c5e4 d2d3 e4g5 f3g5 f6g5 d3e4 f8d8 c4d3 g7g6 h2h4 g5f6 h4h5 f6g5 e4c6 c8d7 c6f3 g5h5 f3h5 g6h5 b1b7 d7a4 f1b1 a4c6 b7c7');
INSERT INTO evaluations VALUES(41970,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(41970,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(41970,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(41970,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(41970,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(41970,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(41970,36,'cp -16','g1f3 e8g8 d1d2 c7c5 d4c5 d5e4 c3e4 f6b2 a1b1 b2f6 e4f6 d8f6 f1d3 b8d7 c5c6 b7c6 e1g1 d7c5 d3c4 e6e5 d2e3 e5e4 f3d2 f6e7 b1e1 f8e8 f2f3 c8e6 c4e6 c5e6 d2e4 e7d8 g1h1 d8d5 e1d1 d5a2 e3c3 a8b8 c3c6 e8c8 c6d7 a2c2 d7a7');
INSERT INTO evaluations VALUES(41300,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41300,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41300,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(41300,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41300,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41300,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(41300,36,'cp 14','e8g8 f1e2 b8c6 g1f3 e7e5 d4e5 d6e5 f3e5 d8d1 a1d1 f6e4 e5c6 g7c3 b2c3 b7c6 d1d3 a8b8 e1g1 c8e6 f1e1 e6c4 d3e3 c4e2 e3e4 f8e8 e4e8 b8e8 f2f3 e8e6 g1f2 e2c4 e1e6 c4e6 a2a3 g8f8 f2e3 f8e8 e3d4 e8d7 f4e5 a7a5 e5f6 a5a4 d4c5 e6d5');
INSERT INTO evaluations VALUES(41878,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(41878,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(41878,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(41878,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(41878,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(41878,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(41878,36,'cp -2','h2h3 b8c6 d2d3 e8g8 f1e2 c6d4 f3d4 c5d4 c1e3 d4e3 f2e3 c7c6 d1d2 c8e6 e1g1 d6d5 e4d5 c6d5 c4c5 d5d4 c3b5 f6e8 e2f3 a8c8 b2b4 a7a6 b5a3 b7b6 a1c1 e8f6 b4b5 c8c5 c1c5 b6c5 b5a6 d4e3 d2e3 d8a5 a3c4 e6c4 d3c4 a5a6 e3e5 a6c4');
INSERT INTO evaluations VALUES(41877,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(41877,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(41877,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(41877,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(41877,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(41877,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(41877,36,'cp -21','e8g8 d2d3 a7a5 f1e2 b8c6 e1g1 c8d7 c1e3 c6d4 d1d2 c7c6 e2d1 f6e8 c3a4 d4f3 d1f3 c5e3 d2e3 a8b8 b2b3 c6c5 a4c3 e8c7 f3d1 c7e6 d1g4 b7b5 c4b5 d7b5 g4e6 f7e6 c3b5 b8b5 a1c1 h7h6 c1c4 b5b4 f1c1 d8b6 e3d2');
INSERT INTO evaluations VALUES(42695,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(42695,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42695,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42695,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42695,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42695,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(42695,36,'cp 2','f6d5 c4d5 c7c6 d5b3 d7d5 c2c3 b4c5 g1f3 d5e4 f3e5 e8g8 d2d4 e4d3 e1g1 c5b6 d1d3 d8d3 e5d3 c8f5 d3e5 b8d7 e5d7 f5d7 c1f4 f8e8 f1e1 d7f5 h2h3 h7h6 a1d1 g7g5 g2g4 e8e1 d1e1 f5g6 f4g3 a8d8 e1e7 d8d2 e7b7 d2b2 b7e7');
INSERT INTO evaluations VALUES(42890,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42890,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42890,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(42890,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42890,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(42890,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42890,36,'cp 13','c4d5 e8g8 c2c3 c7c6 d5b3 b4c5 g1f3 d7d5 f3e5 d5e4 d2d4 e4d3 e1g1 c5b6 d1f3 c8e6 b3e6 f7e6 f3d3 d8d3 e5d3 b8d7 f1e1 a8d8 c1e3 f8f5 h2h3 b6e3 e1e3 d7b6 a1d1 b6c4 e3e4 f5d5 e4c4 d5d3 d1d3 d8d3 g1f1 d3d2 c4b4 b7b5 a2a4 b5a4');
INSERT INTO evaluations VALUES(42887,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(42887,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42887,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(42887,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(42887,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(42887,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(42887,36,'cp 8','e8g8 c2c3 b4c5 g1f3 c7c6 d5b3 d7d5 f3e5 d8g5 d2d4 g5g2 d1f3 g2f3 e5f3 d5e4 f3g5 c5e7 g5e4 a7a5 a2a4 c8e6 b3e6 f7e6 h1g1 b8d7 c1h6 f8f7 e1c1 d7b6 b2b3 b6d5 c1c2 a8d8 g1g4 g7g6 h6g5 e7g5 g4g5 f7f5 g5f5 e6f5');
INSERT INTO evaluations VALUES(51317,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(51317,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(51317,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(51317,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(51317,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(51317,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(51317,36,'cp 85','a2a4 b5b4 f1c4 e7e6 d2d3 g8f6 e3e4 c7c5 e1g1 f8e7 b1d2 e8g8 d2b3 c8b7 d1e2 d8c7 c1g5 b8d7 a4a5 a8c8 f1e1 c7b8 a5a6 b7a8 b3a5 h7h6 g5h4 e7d6 a5b7 d7e5 b7d6 b8d6 f3e5 d6e5 h4g3 e5d4 g3f4 f6d7 f4e3 d4d6 a1a5 f8d8 f2f3 a8c6 e1a1 d6e5 e2f2');
INSERT INTO evaluations VALUES(51332,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(51332,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(51332,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(51332,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(51332,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(51332,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(51332,36,'cp 67','b5b4 f1c4 e7e6 e1g1 g8f6 d2d3 f8e7 b1d2 e8g8 d2b3 c8a6 c4a6 b8a6 d1c2 c7c5 e3e4 a8c8 c1e3 d8b6 f1c1 f6g4 e3f4 g4f6 c2e2 a6b8 c1c4 b8d7 h2h3 f8d8 a4a5 b6a6 a1c1 h7h6 f4e3 c8c7 e2c2 d8c8 d3d4 c5d4 c4c7 c8c7 c2c7');
INSERT INTO evaluations VALUES(53182,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53182,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(53182,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(53182,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(53182,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53182,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(53182,36,'cp 0','f1b5 c8d7 e5e6 f7e6 a2a4 g8h6 g1f3 h6f5 c3d4 d7b5 a4b5 f5d4 f3d4 c5d4 d1h5 g7g6 h5e5 d8d5 e5h8 d5g2 h1f1 g2e4 e1d1 e4g2 d1e1 g2e4');
INSERT INTO evaluations VALUES(54963,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(54963,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(54963,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(54963,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(54963,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(54963,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54963,36,'cp 0','d6d5 e2e3 c7c5 f1d3 g8f6 g1f3 c5d4 e3d4 g7g6 c2c3 c8f5 d3f5 g6f5 d1e2 e7e6 c1e3 b8c6 b1d2 h7h5 f3e5 f8d6 e1g1 a8c8 a1d1 h5h4 d2f3 f6e4 f3d2');
INSERT INTO evaluations VALUES(55930,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(55930,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(55930,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(55930,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(55930,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(55930,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(55930,36,'cp -51','g1f3 c7c5 d2d4 b8c6 c2c4 d5c4 d1a4 c5d4 f3d4 d8d4 g2c6 c8d7 c1e3 d7c6 a4c6 d4d7 c6d7 e8d7 b1d2 e6e5 d2c4 d7e6 e1g1 g8f6 g4g5 f6d5 e3d2 f8e7 a1c1 h8d8 g1g2 b7b6 g2f3 f7f5 e2e4 d5f4 c4e5 d8d2 f3f4 f5e4 f4e4');
INSERT INTO evaluations VALUES(56077,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(56077,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(56077,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(56077,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56077,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56077,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56077,36,'cp 0','b1c3 g8f6 d1e2 f8e7 c3e4 e8g8 g1f3 b7b6 e1c1 c8b7 e4f6 e7f6 d2d4 a7a5 a2a4 b8c6 f3e5 f6e5 d4e5 d8g5 e2d2 g5e7 d2d7 e7g5');
INSERT INTO evaluations VALUES(56078,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56078,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56078,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56078,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56078,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56078,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56078,36,'cp 0','g8f6 d1e2 f8e7 c3e4 e8g8 g1f3 b7b6 d2d4 c8b7 e4f6 e7f6 e1c1 a7a5 a2a4 b8c6 f3e5 f6e5 d4e5 d8g5 e2d2 g5e7 d2d7 e7g5');
INSERT INTO evaluations VALUES(48538,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48538,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48538,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48538,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48538,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48538,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48538,36,'cp 0','d1e2 f8e7 c3e4 e8g8 g1f3 b7b6 e4f6 e7f6 d2d4 c8b7 e1c1 a7a5 a2a4 b8c6 f3e5 f6e5 d4e5 d8g5 e2d2 g5e7 d2d7 e7g5');
INSERT INTO evaluations VALUES(48539,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48539,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48539,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48539,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48539,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48539,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48539,36,'cp 0','f8e7 c3e4 e8g8 g1f3 b7b6 e4f6 e7f6 d2d4 c8b7 e1c1 a7a5 a2a4 b8c6 f3e5 f6e5 d4e5 d8g5 e2d2 g5e7 d2d7 e7g5');
INSERT INTO evaluations VALUES(41052,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41052,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41052,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41052,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41052,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41052,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41052,36,'cp 0','c3e4 e8g8 g1f3 b7b6 e1c1 c8b7 e4f6 e7f6 d2d4 a7a5 a2a4 b8c6 f3e5 f6e5 d4e5 d8g5 e2d2 g5e7 d2d7 e7g5');
INSERT INTO evaluations VALUES(56954,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(56954,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(56954,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(56954,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(56954,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(56954,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(56954,36,'cp 6','d7d5 e4d5 e6d5 b1c3 g8f6 c4d5 f6d5 f1c4 d5c3 b2c3 f8d6 g1e2 b8d7 e1g1 d7f6 e2g3 e8g8 d1f3 a8b8 a2a4 c8e6 c4e6 f7e6 f1e1 d8d7 g3e4 f6e4 f3e4 f8f6 g2g3 b7b6 f2f4 h6h5 a4a5 b6b5 e1e2 h5h4 c1a3 d6a3 a1a3 b5b4 c3b4');
INSERT INTO evaluations VALUES(2995,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(2995,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(2995,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(2995,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(2995,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(2995,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(2995,36,'cp 66','e2e4 e7e5 h2h3 e5d4 f3d4 f8e8 c1e3 a7a5 d1c2 a5a4 a1d1 d8a5 f1e1 d7c5 e3f4 g7f8 d4f3 e8d8 f4d2 d8e8 c3d5 a5d8 d5f6 d8f6 d2c3 f6e7 f3d4 f8g7 f2f4 c8d7 g1h2 h7h5 c2d2 e7f8 a2a3 d7e6 d4e6 c5e6 f4f5 g7c3 d2c3');
INSERT INTO evaluations VALUES(2997,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(2997,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(2997,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(2997,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(2997,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(2997,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(2997,36,'cp 64','e7e5 h2h3 d8a5 c1e3 e5d4 f3d4 d7b6 d1d3 a5a6 b2b3 d6d5 f1d1 d5c4 d3c2 f8d8 f2f4 c8d7 e3f2 c6c5 d4e2 a8c8 e4e5 d7f5 c2b2 f6e8 c3e4 f5e4 g2e4 e8c7 b3c4 c7e6 e4d5 b6c4 b2b7 a6b7 d5b7 c8b8 d1d8 b8d8 e2c3 g6g5 b7d5 c4b6');
INSERT INTO evaluations VALUES(2922,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(2922,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(2922,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(2922,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(2922,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(2922,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(2922,36,'cp 83','h2h3 e5d4 f3d4 f8e8 c1e3 a7a5 d1c2 a5a4 a1d1 d8a5 b2b3 d7c5 e3f4 a4b3 a2b3 a5b6 f1e1 g7f8 f4e3 f8g7 d1b1 b6b4 e1d1 f6d7 e3d2 b4b6 d2f4 d7e5 f4e3 b6b4 c3e2 c5e6 d1d2 e6d4 e2d4 a8a3 c2d1 b4a5 d4c2 a3a2');
INSERT INTO evaluations VALUES(2923,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(2923,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(2923,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(2923,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(2923,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(2923,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(2923,36,'cp 74','a7a5 d1c2 f8e8 c1e3 e5d4 f3d4 a5a4 f1d1 d7b6 c4c5 b6c4 e3f4 d8a5 c5d6 f6h5 d4e2 c8e6 f4c1 e8d8 e2d4 a4a3 b2b4 a5b4 e4e5 g7e5 a1b1 b4d6 d4c6 b7c6 d1d6 c4d6 g3g4 h5f4 c1f4 e5f4');
INSERT INTO evaluations VALUES(4903,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(4903,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(4903,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(4903,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(4903,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(4903,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(4903,36,'cp 79','a2a4 b5b4 d2d4 e5d4 c3d4 c5a7 c1e3 c6a5 e1g1 a5b3 d1b3 g8f6 b1d2 e8g8 b3b4 f8e8 f1c1 h7h6 b4b3 a8b8 b3c2 c8b7 c2c7 d8c7 c1c7 a7b6 c7c1 f6e4 b2b4 b7d5 d2e4 d5e4 f3d2 b6d8 e3f4 b8b4 f4d6 b4b2');
INSERT INTO evaluations VALUES(9270,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(9270,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(9270,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(9270,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(9270,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(9270,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(9270,36,'cp 137','e3e4 e6e5 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 c4f7 e5c3 b2c3 f6e4 e1g1 c8d7 f1e1 e4d6 f7d5 a7a5 c3c4 h8e8 c1g5 d8c8 c4c5 a5a4 g5d2 e8e1 a1e1 d6e8 e1e7 a8a6 d5b7 c8b7 e7d7 a6e6 g2g3 b7c6 d7d4 a4a3 d4d3 e8f6 d2e3 f6d5 e3d4 g7g6 d3a3 c6b5 a3a7 e6e2 g1f1 e2d2');
INSERT INTO evaluations VALUES(9844,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(9844,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(9844,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(9844,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(9844,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(9844,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9844,36,'cp -12','b4c3 b2c3 f6e4 d1e1 e4f6 f3e5 e8g8 d2d3 c6a5 c4a2 d7d6 e5f3 c7c5 e1d2 f8e8 d2f4 c8e6 a2e6 e8e6 f4g3 g8h8 c1d2 d8d7 c3c4 a5c6 f1e1 a8e8 d2c3 b7b6 e1e6 e8e6 a1e1 e6e1 f3e1 f6e8 e1f3 h8h7 g1f1 d6d5 c4d5 c6e7 g3b8 e7d5 c3b2 e8f6 f3e5 d7b5');
INSERT INTO evaluations VALUES(9863,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(9863,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(9863,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(9863,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(9863,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(9863,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9863,36,'cp -7','d2c3 d7d6 d1e2 c8g4 a3a4 a7a5 h2h3 g4e6 c4b5 f6d7 f3h2 e8g8 f2f4 e5f4 c1f4 c6e5 h2f3 e5f3 f1f3 f7f6 b2b3 g8h8 e2d2 d8e7 g1h1 e6g8 a1e1 g8h7 f3e3 d7c5 d2e2 a8d8 f4g3 c5d7 b3b4 c7c6 b5d3 f8e8 b4a5');
INSERT INTO evaluations VALUES(9859,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(9859,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(9859,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(9859,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(9859,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(9859,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(9859,36,'cp -9','d7d6 d1e2 c8g4 a3a4 a7a5 h2h3 g4e6 c4b5 f6d7 f3h2 e8g8 f2f4 e5f4 c1f4 c6e5 h2f3 e5f3 f1f3 f7f6 b2b3 g8h8 a1e1 d8e7 e2d2 d7e5 f3f1 b7b6 g1h1 e6d7 b5d3 d7c6 c3c4 e5d7 f4g3 a8e8 f1f4 f8f7 f4g4 d7e5 g4h4');
INSERT INTO evaluations VALUES(10456,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(10456,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(10456,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(10456,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(10456,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(10456,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10456,36,'cp -7','d8e7 d1e2 e8g8 g2g3 d7d5 f1g2 c8g4 e1g1 a8d8 e4d5 f6d5 f1e1 f7f6 h2h3 g4h5 d3d4 e7f7 d4e5 f6e5 g3g4 d5f4 e2e3 h5g6 a1d1 d8d1 e1d1 f4h3 g2h3 f7f3 e3f3 f8f3 h3g2 f3f4 d1d7 f4g4 g1h2 g4f4 g2c6 b7c6 d7g7 g8g7 c3e5 g7f8 e5f4 g6c2 f4c7 f8f7 h2h1 c2f5 b2b3');
INSERT INTO evaluations VALUES(10554,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(10554,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(10554,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(10554,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(10554,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(10554,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(10554,36,'cp 145','c3d5 h6g4 e1g1 c5e7 d2d4 d7d6 h2h3 g4f6 d5e7 d8e7 f1e1 e8g8 c2c3 a7a5 c1g5 h7h6 g5h4 a5a4 d1d2 g8h7 g2g4 g7g6 c4d5 h7g7 h4g3 f6d7 f3h4 f7f6 f2f4 d7b6 d4e5 f6e5 f4f5 g6f5 e4f5 b6d5 d2d5');
INSERT INTO evaluations VALUES(15538,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(15538,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(15538,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(15538,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(15538,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(15538,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(15538,36,'cp 51','g7g6 e2e4 f8g7 g1e2 c8e6 a1b1 d8d7 e1g1 a7a5 f2f3 f6h5 b2b3 f7f5 e4f5 g6f5 d3d4 e5d4 c3b5 e8g8 e2d4 e6f7 f3f4 h5f6 c1b2 c6d4 b2d4 f7h5 d1d2 c7c6 d4f6 f8f6 b5d4 d6d5 b1e1 a8e8 e1e8 h5e8');
INSERT INTO evaluations VALUES(19336,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(19336,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(19336,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(19336,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(19336,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(19336,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(19336,36,'cp 44','e7e6 d2d4 c5d4 f3d4 g8f6 c1f4 d7d6 d4c6 b7c6 d1e2 f8e7 e1c1 e8g8 g2g4 d6d5 h1g1 e7b4 g4g5 b4c3 b2c3 f6e4 c4d3 d8e7 d3e4 e7a3 c1d2 d5e4 g5h6 g7g6 e2e4 f8d8 d2e1 a3c3 f4d2 c3f6 g1g3 c8a6 d2c3 f6f5 e4f5 e6f5 c3f6 d8d1 e1d1 a6b7');
INSERT INTO evaluations VALUES(18650,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(18650,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(18650,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(18650,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(18650,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(18650,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(18650,36,'cp 35','d2d4 c5d4 f3d4 d8c7 d4c6 d7c6 e1g1 g8f6 a2a4 f8c5 g1h1 e6e5 d1e1 c5d6 f2f3 e8g8 c3d1 f8d8 d1e3 f6h5 g2g3 h5f6 g3g4 d6c5 c1d2 c7e7 a1d1 c5e3 e1e3 c8e6 c4d3 f6d7 d2c3 d7c5 c3e5 c5d3 d1d3 d8d3 e3d3');
INSERT INTO evaluations VALUES(19419,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19419,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19419,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19419,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(19419,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(19419,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19419,36,'cp 14','c5d4 c3d4 d7d5 c1f4 g8f6 b1c3 c8f5 d1b3 c6a5 b3a4 f5d7 a4c2 a8c8 e2e3 b7b5 a2a3 e7e6 f1d3 f8e7 e1g1 a7a6 c2e2 f6h5 f4e5 e8g8 h2h3 a5c4 f1c1 h5f6 a3a4 b5b4 e5f6 e7f6 c3d1 a6a5 d3c4 d5c4 c1c4 c8c4');
INSERT INTO evaluations VALUES(37431,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(37431,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(37431,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(37431,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(37431,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37431,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(37431,36,'cp 0','g1f3 c7c5 f1e2 d8b6 d1c1 d7d5 c2c3 b8c6 a2a4 c8f5 b1d2 c5d4 e3d4 c6b4 c3b4 f8c8 c1d1 f5c2 a4a5 b6e6 d1c1 c2d3 c1d1');
INSERT INTO evaluations VALUES(44331,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(44331,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(44331,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(44331,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(44331,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(44331,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(44331,36,'cp 44','e1g1 d7d5 e4d5 f8e7 c2c3 f6d5 c3d4 d5b6 c4b5 b8c6 f3e5 c8d7 e5d7 b6d7 b1c3 e8g8 d4c5 e7c5 d1d5 d8c7 f1d1 f8d8 c1g5 c5e7 g5e3 d7f6 d5c4 a7a6 b5c6 b7c6 b2b3 c7b7 e3d4 h7h6 a1c1 d8d7 d4e3 d7d8 d1d8 a8d8 h2h3 a6a5 c3e4');
INSERT INTO evaluations VALUES(44740,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(44740,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(44740,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44740,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(44740,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(44740,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44740,36,'cp 0','c1g5 f8g7 f2f4 b8c6 g1f3 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 d8b6 a1b1 e7e6 f4f5 g6f5 c3e2 d6d5 g5f6 g7f6 e4d5 a8e8 d5e6 f7e6 c2c3 d4e2 d1e2 g8h8 g1h1 f6d8 e2e5 d8f6');
INSERT INTO evaluations VALUES(51843,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(51843,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(51843,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51843,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(51843,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51843,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(51843,36,'cp 7','g8f6 d2d4 e5d4 d1d4 f8e7 f1e2 e8g8 g1f3 d7d5 e1g1 d5c4 d4c4 d8c7 f1d1 b8a6 h2h3 a6c5 c4f4 c7f4 e3f4 a7a5 g2g4 a5a4 g1g2 a4a3 b2a3 c8e6 f3d4 f6d5 d4e6 f7e6 c3d5 e6d5 e2f3 e7d6 f4f5');
INSERT INTO evaluations VALUES(53546,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53546,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53546,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53546,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53546,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(53546,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(53546,36,'cp 0','e6d5 c1g5 f8e7 g5e7 d8e7 e2e3 c8f5 b1c3 g8f6 f1d3 f5d3 d1d3 e8g8 e1g1 b8d7 d3c2 f8e8 c2b3 e8b8 b3c2 b8e8');
INSERT INTO evaluations VALUES(55842,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(55842,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(55842,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(55842,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(55842,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(55842,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(55842,36,'cp -51','c7c5 g1f3 g8f6 f1g2 c5d4 f3d4 f8b4 e1g1 e8g8 e2e4 b4c3 b2c3 e6e5 d4b3 c8g4 d1d2 d5e4 c1a3 f8e8 d2d8 e8d8 a3e7 d8c8 e7f6 g7f6 g2e4 b8c6 f2f4 c8c7 f4e5 f6e5 a1e1 a8e8 e4g2 g8g7 b3d4 c6d4 c3d4 c7c2 e1e5 e8e5');
INSERT INTO evaluations VALUES(2795,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(2795,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(2795,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(2795,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(2795,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(2795,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(2795,36,'cp 199','c5e7 e4e5 f6e8 b1c3 c6a5 b3c2 d7d5 c3e2 c7c6 h2h3 f7f5 f1e1 e8c7 c2d3 c8d7 d1c2 a8c8 b2b3 a5b7 a2a4 a6a5 c1e3 b5b4 a1c1 g7g6 e2f4 c7e6 f4e6 d7e6 c2d2 d8b6 e3g5 e7g5 f3g5 c6c5 f2f4');
INSERT INTO evaluations VALUES(4142,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4142,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(4142,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(4142,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4142,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(4142,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(4142,36,'cp 14','g5h4 d7d6 b1d2 a7a5 h2h3 c8e6 a2a3 d8e7 e1g1 e6c4 d2c4 e7e6 b2b4 a5b4 a3b4 c5a7 h4f6 e6f6 d1b3 f6e6 c4d2 e6b3 d2b3 c6e7 g2g3 f7f5 e4f5 g7g6 g1g2 g6f5 d3d4 a7b6 d4e5 d6e5 f1e1 a8a1');
INSERT INTO evaluations VALUES(4158,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(4158,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(4158,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(4158,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(4158,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(4158,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4158,36,'cp 18','c5e7 b1d2 d7d6 a2a4 g8h8 h4g3 f6g8 e1g1 f7f5 e4f5 c8f5 a4a5 a7a6 f1e1 e7f6 h2h3 f5g6 d2f1 g8e7 f1e3 d8d7 e3g4 e7f5 d1b3 a8b8 d3d4 e5d4 c4e6 d7d8 e6f5 g6f5 g4f6 d8f6 f3d4 f5g6 e1e3 c6d4 c3d4 f8e8 e3e8 g6e8 a1e1 f6f7 b3f7 e8f7 e1e7');
INSERT INTO evaluations VALUES(4547,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(4547,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(4547,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(4547,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(4547,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(4547,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(4547,36,'cp 18','b1d2 d7d6 a2a4 g8h8 h4g3 f6g8 e1g1 f7f5 e4f5 c8f5 a4a5 a7a6 f1e1 e7f6 h2h3 f5g6 d2f1 g8e7 f1e3 d8d7 e3g4 e7f5 d1b3 a8b8 c4e6 d7d8 b3d1 f6g5 e6f5 g6f5 d3d4 e5d4 f3d4 c6d4 d1d4 d8d7 h3h4 g5e7 g4e3 e7f6 d4d2 b8e8 a1d1 f5e4 e3d5 e4d5');
INSERT INTO evaluations VALUES(10909,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10909,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(10909,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(10909,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10909,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(10909,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(10909,36,'cp -5','d7d5 e4d5 f6d5 e1g1 c8g4 b1d2 f7f6 f1e1 d8d7 c4b5 a7a6 b5c6 d7c6 f3e5 g4d1 e5c6 d1c2 c6e7 d5e7 d2e4 e8c8 e4f6 e7c6 f6h5 c6d4 c1f4 c2d3 e1e7 d4e2 g1h1 e2f4 h5f4 g7g5 f4e6 d8e8 e7c7 c8b8 c7c3 e8e6 c3d3 e6e2 d3f3 g5g4 f3f5 e2b2 g2g3 h8c8 h1g2 c8c2 a1f1 b2a2 f5f7 a2a3');
INSERT INTO evaluations VALUES(12432,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(12432,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12432,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(12432,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(12432,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12432,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(12432,36,'cp -12','f8e7 g1e2 e8g8 e1g1 d7d6 d3d4 c5d4 e2d4 h7h6 f1e1 c6d4 d1d4 c8d7 a2a4 a7a6 h2h3 d7c6 c1e3 d8c7 a4a5 f6d7 d4d2 a8b8 e3d4 e7g5 d2d3 g5f6 c3e2 b8c8 d4f6 d7f6 e2d4 f8d8 d4c6 c7c6');
INSERT INTO evaluations VALUES(7920,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(7920,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(7920,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(7920,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7920,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7920,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7920,36,'cp 0','g1e2 e8g8 e1g1 d7d6 d3d4 c5d4 e2d4 h7h6 f1e1 c6d4 d1d4 c8d7 a2a4 a7a6 h2h3 d7c6 a4a5 d8c7 c1e3 a8c8 d4c4 f6d7 e1d1 d7c5 e3d4 c5d7');
INSERT INTO evaluations VALUES(14842,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14842,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14842,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14842,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14842,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14842,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14842,36,'cp 0','g1f3 f8e7 c1f4 f6h5 f4c1');
INSERT INTO evaluations VALUES(16541,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(16541,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(16541,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(16541,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(16541,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(16541,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(16541,36,'cp -28','d7d5 b1d2 f8c5 g1f3 a7a5 f1e2 e8g8 e1g1 c8e6 f1e1 h7h6 e4d5 e6d5 d2f1 f8e8 c1e3 c5d6 f1g3 d5e6 d3d4 e5e4 f3d2 d6g3 f2g3 c6e7 d2f1 e7d5 e3f2 c7c6 g3g4 b7b5 d1c1 a5a4 f1g3 a4a3 b2a3 e4e3 f2e3 d5e3 c1e3 a8a3');
INSERT INTO evaluations VALUES(15152,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(15152,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(15152,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(15152,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(15152,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(15152,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(15152,36,'cp -20','b1d2 f8c5 g1f3 a7a5 f1e2 e8g8 e1g1 c8e6 f1e1 h7h6 e4d5 e6d5 d2f1 c5d6 f1g3 d5e6 e2f1 f8e8 c1d2 f6d7 d2e3 a5a4 g3e4 d6f8 e4g3 d7f6 d1c2 f6d5 d3d4 d5e3 f2e3 a4a3 b2b4 e5d4 e3d4 f8d6');
INSERT INTO evaluations VALUES(17940,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(17940,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(17940,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(17940,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(17940,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(17940,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(17940,36,'cp 102','f8b4 c1d2 a6a5 e1g1 b4d2 b1d2 c7c6 a2a3 d7d6 d1e1 b7b5 a4b3 a5a4 b3a2 c8e6 a2e6 f7e6 e4e5 d6d5 b2b4 g8h6 d2f3 h6f5 g2g4 f5h4 f3d4 d8g5 e1e2 g5f4 c2c3 e8g8 f2f3 f8f7 e2f2 g7g5 a1e1 a8f8 f2e3 h7h6 e1e2 f4e3 e2e3 f7f4 d4e6 f4f3 e3f3 f8f3 d3d4 f3c3');
INSERT INTO evaluations VALUES(28830,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(28830,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(28830,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(28830,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(28830,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(28830,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(28830,36,'cp 45','c2c4 c7c5 b1c3 g7g6 e1g1 f8g7 d2d4 c5d4 d1d4 d7d6 d4h4 h7h6 f1d1 b8d7 b2b3 a8c8 c1b2 e8g8 a1c1 c8c5 f3e1 d8a8 g2b7 a8b7 e1d3 c5c8 h4f4 d7c5 d3b4 a7a6 f4d2 f6e4 d2d5 c8c7 d5b7 c7b7 f2f3 e4c3 b2c3 g7c3 c1c3');
INSERT INTO evaluations VALUES(28794,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(28794,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(28794,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(28794,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(28794,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(28794,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(28794,36,'cp 34','c7c5 b1c3 g7g6 d2d4 c5d4 d1d4 f8g7 e1g1 d7d6 d4h4 h7h6 b2b3 b8d7 f1d1 d7c5 c1b2 f6e4 c3e4 g7b2 a1b1 b2g7 e4c5 b6c5 b3b4 c5b4 b1b4 b7c6 c4c5 e8g8 c5d6 e7d6 h4d8 f8d8 f3d4 c6g2 g1g2 d8e8 e2e3 a8c8 b4b7 g7d4 d1d4 c8c2 b7a7 e8e3 d4d6');
INSERT INTO evaluations VALUES(28624,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(28624,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(28624,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(28624,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(28624,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(28624,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(28624,36,'cp 38','b1c3 g7g6 e1g1 f8g7 d2d4 c5d4 d1d4 d7d6 d4h4 h7h6 b2b3 b8d7 c1e3 d7c5 a1c1 d8d7 f1d1 c5e4 c3e4 f6e4 e3d4 e8g8 h4f4 a8c8 f4e3 f8e8 d4g7 g8g7 f3e5 d7f5 e5d3 e4c5 g2b7 c5b7 f2f3 b7c5 d3f4 f5e5 e3e5 d6e5 f4d3 c5d3 d1d3');
INSERT INTO evaluations VALUES(34025,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(34025,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(34025,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(34025,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(34025,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(34025,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(34025,36,'cp 106','f1e2 f6g4 e1g1 g4e5 f3d4 e7d8 b1c3 f8c5 d4f5 e8g8 c3d5 a7a5 c1h6 g7h6 d1c1 f7f6 c1h6 f8f7 e2h5 d8f8 h6f4 e5g6 f5h6 g8g7 h5g6 h7g6 h6f7 f8f7 d5c7 a8a6 c7a6 b8a6 a1d1 d7d5 f4d2 c8e6 d2a5 c5b4 a5b5 b4d6 f1e1 a6c5 b5b4 c5e4 b4d4 d6c5');
INSERT INTO evaluations VALUES(34024,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(34024,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(34024,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(34024,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(34024,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(34024,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(34024,36,'cp 125','f6g4 c1g5 e7c5 e1g1 b8c6 c2c3 g4e5 f1e1 e5f3 e2f3 f8e7 g5e7 c6e7 c3d4 c5d6 d1e2 d6e6 e2d2 e6f6 d2e3 d7d5 f3d5 c8e6 d5e6 f6e6 e3e6 f7e6 e1e6 a8d8 b1c3 d8d7 e6e4 e8g8 a1e1 g8f7 g1f1 f8d8 g2g3 e7c6 d4d5 c6e7 h2h4 h7h6 e4b4 c7c6 d5c6 e7c6 b4f4 f7g8 e1e2 d7d4 f4d4 c6d4 e2e4 d4c6 h4h5 d8d2 e4e2');
INSERT INTO evaluations VALUES(36701,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(36701,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(36701,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(36701,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(36701,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(36701,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(36701,36,'cp 78','e7e5 h6g7 g8g7 f2f3 d8e7 e1c1 h7h5 h2h3 b8c6 g1e2 h5h4 d2e3 e5d4 e2d4 c6d4 d1d4 c8d7 f1e2 d7c6 e3d2 a7a6 e2c4 a8e8 h1e1 b7b5 c4b3 f8h8 c3e2 e7e5 a2a3 f6d7 e1d1 d7c5 b3a2 c6d7 d2c3 e5f6');
INSERT INTO evaluations VALUES(37302,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(37302,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(37302,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(37302,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(37302,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(37302,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(37302,36,'cp 78','e2e4 d7d5 e4e5 f6e4 c4d5 e6d5 f1d3 c8f5 e1g1 b8c6 f1e1 e4c3 b2c3 d8d7 h2h3 f5d3 d1d3 d7e6 a2a4 b7b6 c1a3 e7a3 a1a3 c6a5 a3a1 e6h6 g1h2 a8e8 g2g3 h6g6 d3b5 c7c6 b5e2 a5c4');
INSERT INTO evaluations VALUES(37306,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(37306,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(37306,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(37306,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(37306,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(37306,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(37306,36,'cp 59','d7d5 e4e5 f6e4 c4d5 e6d5 f1d3 c8f5 c1e3 b8c6 e1g1 e4c3 b2c3 d8d7 h2h3 b7b6 d1e2 f5d3 e2d3 d7e6 a1e1 c6a5 f3d2 c7c6 f2f4 f7f5 e5f6 e6f6 d2f3 f6g6 d3g6 h7g6 f3e5 f8f6 g2g4 e7d6 f4f5 g6f5 e3g5 d6e5 e1e5');
INSERT INTO evaluations VALUES(39817,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39817,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39817,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39817,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39817,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39817,36,'cp 0','h7h6 g5h4 g7g5 h4g3 f6e4 d1c2 b4c3 b2c3 c8b7 e2e3 e4g3 h2g3 h6h5 f1d3 d8e7 a1b1 g5g4 f3h4 d7d6 e1g1 b8c6 e3e4 e8c8 b1b5 a7a6 b5b2 d8d7 f1b1 h8d8 b1e1 e7f6 e1b1');
INSERT INTO evaluations VALUES(39815,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39815,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39815,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39815,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(39815,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39815,36,'cp 0','g5h4 g7g5 h4g3 f6e4 d1c2 f7f5 e2e3 c8b7 f1d3 b4c3 b2c3 d7d6 c4c5 b6c5 c2b3 b7d5 c3c4 e4g3 c4d5 g3h1 d5e6 g5g4 b3d5 g4f3 d5a8 c5d4 e3d4 e8g8 a8f3 d8f6 a1c1 f6d4 c1c7 d4a1 e1d2 a1a2 c7c2 a2a5 d2e2 a5e5 e2f1 e5a1 f1e2');
INSERT INTO evaluations VALUES(39816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39816,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(39816,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39816,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(39816,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(39816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39816,36,'cp 9','c8b7 d1c2 b4c3 c2c3 g7g5 h4g3 f6e4 c3c2 e4g3 h2g3 d8f6 e2e3 h6h5 e1c1 g5g4 f3e1 b8c6 e1d3 e8c8 h1h4 d7d6 d3f4 e6e5 d4e5 d6e5 d1d8 c6d8 f4d5 b7d5 c4d5 d8b7 c1b1 c8b8 f1b5 b7d6 b5c6 f6f5 a2a3 f5c2 b1c2 d6e4 f2f3 g4f3 h4e4 f3g2');
INSERT INTO evaluations VALUES(28222,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(28222,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(28222,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(28222,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28222,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28222,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(28222,36,'cp 0','d1c2 b4c3 c2c3 g7g5 h4g3 f6e4 c3c2 e4g3 h2g3 d8f6 e2e3 b7f3 g2f3 f6f3 h1h2 b8c6 c2a4 f3f5 f1g2 e8c8 c4c5 c8b8 a1c1 h6h5 e1e2 h5h4 g2c6 h4g3 c5b6 c7b6 h2h8 d8h8 c6d7 f5f2 e2d3 g3g2 a4c6 f2f1 d3d2 f1f2');
INSERT INTO evaluations VALUES(42478,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(42478,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42478,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(42478,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42478,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42478,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42478,36,'cp 3','c7c5 d4d5 d7d6 e2e4 e8g8 g1e2 b7b5 e2f4 a7a6 a2a3 b4a5 d5e6 f7e6 f1e2 b8c6 e1g1 c6d4 c1e3 e6e5 f4d5 c8e6 c4b5 a6b5 e2b5 f6d5 e4d5 e6f5 b5d3 a8b8 a1b1 f5d3 d1d3 b8b3 g1h1 d8h4 e3f2 h4h6 f2e3');
INSERT INTO evaluations VALUES(39921,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39921,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39921,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(39921,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(39921,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(39921,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(39921,36,'cp 3','d4d5 e8g8 e2e4 d7d6 g1e2 b7b5 e2f4 e6d5 f4d5 f6d5 d1d5 d8b6 c4b5 c8b7 d5d2 b8d7 a2a3 b4a5 f1e2 f7f5 b2b4 c5b4 c3a4 b4b3 e2c4 g8h8 a4b6 a5d2 e1d2 d7b6 c4b3 f5e4 f3e4 a7a6 b5a6 b7e4 h1e1 e4g2 c1b2 g2d5 b3d5 b6d5 e1g1 d5f6 a1e1 a8a6 e1e7 f8g8 b2f6 g7f6 g1g8 h8g8');
INSERT INTO evaluations VALUES(39937,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39937,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39937,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39937,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39937,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39937,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39937,36,'cp 0','d7d6 e2e4 b7b5 g1e2 b5c4 e2f4 e6e5 f4e2 b8d7 e2g3 d7b6 f1e2 c8d7 d1c2 d7a4 c2d2 a4d7');
INSERT INTO evaluations VALUES(42589,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(42589,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(42589,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42589,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42589,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(42589,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(42589,36,'cp -19','d7d5 c4d5 f6d5 g1f3 b8c6 f1g2 d5b4 e1g1 b4d3 f3e1 c8f5 a2a3 c5b6 b2b4 d8d7 c3e4 e8c8 d1c2 f5g6 c2d3 d7d3 e1d3 d8d3 c1b2 f7f5 e4c3 d3d2 c3a4 h8e8 b2c3 d2d6 g2h3 c8b8 a1c1 a7a6 a4b6 c7b6 f1d1 e8d8 d1d6 d8d6 h3g2 b8a7 c1b1 g6f7 g2c6 b7c6');
INSERT INTO evaluations VALUES(42944,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(42944,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(42944,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(42944,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(42944,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(42944,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(42944,36,'cp 52','c2c4 e8g8 f1d3 f6h5 b1c3 d7d6 e1g1 c7c5 d3e2 c5d4 f3d4 h5f4 e3f4 b8c6 d4f3 g8h8 d1d2 e6e5 a1d1 e7f6 g2g3 e5e4 f3e1 d8b6 b2b3 c8e6 e1c2 a8d8 c3d5 e6d5 c4d5 c6e7 c2e3 d8c8 e3c4 b6c7 d1c1 a7a6 a2a4 c7c5 a4a5 c5d5 d2d5 e7d5 c4d6');
INSERT INTO evaluations VALUES(49631,30,'cp -228',NULL);
INSERT INTO evaluations VALUES(49631,31,'cp -200',NULL);
INSERT INTO evaluations VALUES(49631,32,'cp -191',NULL);
INSERT INTO evaluations VALUES(49631,33,'cp -216',NULL);
INSERT INTO evaluations VALUES(49631,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(49631,35,'cp -197',NULL);
INSERT INTO evaluations VALUES(49631,36,'cp -222','d7d5 d1e2 d5e4 d3e4 f6e4 e2e4 f8d6 g1f3 f7f5 e4e2 e8g8 d2d3 b7b6 e1g1 d8f6 a2a4 c8a6 b2b4 b8d7 b4b5 a6b7 b1d2 a7a6 c1b2 a8e8 d2c4 d6c5 b2a3 c5a3 a1a3 a6a5 f1e1 e5e4 d3e4 e8e4 e2a2 e4e1 f3e1 b7d5 a2d2 d5c4 d2d7');
INSERT INTO evaluations VALUES(48829,30,'cp -228',NULL);
INSERT INTO evaluations VALUES(48829,31,'cp -208',NULL);
INSERT INTO evaluations VALUES(48829,32,'cp -209',NULL);
INSERT INTO evaluations VALUES(48829,33,'cp -225',NULL);
INSERT INTO evaluations VALUES(48829,34,'cp -223',NULL);
INSERT INTO evaluations VALUES(48829,35,'cp -195',NULL);
INSERT INTO evaluations VALUES(48829,36,'cp -208','d1e2 d5e4 d3e4 f6e4 e2e4 f8d6 b1a3 e8g8 a3c4 b8d7 d2d4 e5d4 c4d6 d4c3 g1e2 d7c5 e4c2 d8d6 e1g1 f8e8 e2c3 b7b6 c2d2 d6g6 d2g5 g6g5 c1g5 f7f6 g5e3 c8a6 f1d1 c5d3 c3d5 a8c8 b2b3 d3e5 d5c3 c7c5 d1d6 c5c4 d6d1 h7h6 b3c4 a6c4');
INSERT INTO evaluations VALUES(50357,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(50357,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(50357,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(50357,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(50357,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(50357,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(50357,36,'cp -37','c7c5 g1f3 g7g6 b1c3 f8g7 e3e4 e8g8 f1e2 b8c6 e1g1 d7d6 d2d3 a7a6 a2a4 a8b8 d1e1 b7b5 a4b5 a6b5 c3d1 c6b4 d1e3 f6g4 e2d1 g4e3 e1e3 b4c6 e3e1 c8g4 h2h3 g4f3 d1f3 b5b4 c2c3 d8b6 g1h1 b4c3 b2c3 b6b5 f3e2 b5b3 c1d2 b3b2 a1b1 b2a3 e1c1 a3c1');
INSERT INTO evaluations VALUES(54173,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(54173,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(54173,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54173,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54173,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54173,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(54173,36,'cp 10','c4d5 g8f6 d5e6 c8e6 e2e3 d8b6 b1c3 b6b2 f1b5 b8c6 g1e2 c5d4 a1b1 b2a3 d1d4 f8e7 d4a4 e8g8 a4a3 e7a3 e1g1 a7a5 f1d1 a3b4 b5c6 b7c6 f4d6 f8d8 d6b4 d8d1 c3d1 a5b4 b1b4 e6a2 d1c3 c6c5 b4b6 a2c4 f2f3 c4e2 c3e2 f6d5 b6b5 d5e3 b5c5 a8d8 e2c3 e3d1 c3d5 g8f8');
INSERT INTO evaluations VALUES(54796,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(54796,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(54796,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(54796,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(54796,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(54796,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(54796,36,'cp 130','c8f5 e2e3 b8c6 h2h4 e7e6 f1e2 h7h6 c2c4 f8d6 c4d5 e6d5 e2h5 e8f8 g3d6 d8d6 b1c3 c6b4 e1f1 g8e7 g1e2 f5h7 f1g1 f8g7 e2g3 e7f5 h1h3 a8d8 a2a3 b4c6 a1c1 a7a6 g3f5 h7f5 h3g3 c6e7 c3e2 c7c6 d1b3 d6d7 b3c3 h8f8 e2f4 e7g6');
INSERT INTO evaluations VALUES(9505,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(9505,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(9505,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(9505,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(9505,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(9505,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(9505,36,'cp 80','d1d8 e8d8 c1d2 c8e6 a2a3 b4c5 f1e2 b7b5 f2f4 d8c8 f4f5 e6d7 e1c1 h8e8 e2f3 a7a5 d1e1 b5b4 c3b1 c5d6 g2g4 a8b8 g4g5 f6g8 f3e2 c8b7 h1f1 f7f6 e2h5 e8e5 d2f4 e5e7 a3b4 d6f4 f1f4 f6g5 f4g4 a5b4 b1d2 g8f6 g4g5 f6h5 g5h5 b8a8 h5h7 a8a1 d2b1');
INSERT INTO evaluations VALUES(2245,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(2245,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(2245,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(2245,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(2245,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(2245,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(2245,36,'cp 81','e8d8 c1d2 c8e6 f1e2 a7a5 a2a3 b4e7 f2f4 d8c8 e1c1 e6g4 e4e5 g4e2 c3e2 f6d5 g2g4 b7b5 h2h4 h8d8 g4g5 c8b7 e2g3 b5b4 a3a4 c6c5 f4f5 d5b6 b2b3 c5c4 f5f6 g7f6 g3f5 e7f8 e5f6 d8d5 f5e3 d5d4 e3c4 b6c4 b3c4');
INSERT INTO evaluations VALUES(10754,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(10754,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(10754,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(10754,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(10754,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(10754,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(10754,36,'cp 75','d2d4 a7a6 b5e2 d7d6 d4c5 f6e4 c5d6 d8d6 b1d2 c8f5 d2e4 f5e4 d1d6 c7d6 f1d1 e4c2 d1d6 e8g8 c1f4 f8e8 a1c1 c2e4 f2f3 e4c6 g1f2 e5g6 f4e3 h7h6 g2g3 g6e7 d6d2 a8c8 a2a3 e8d8 d2d1 d8d1 c1d1 e7d5 e3c5 d5f6 c5d4');
INSERT INTO evaluations VALUES(10752,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(10752,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(10752,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(10752,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(10752,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(10752,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(10752,36,'cp 72','a7a6 b5e2 f6e4 d4c5 d7d6 c5d6 d8d6 b1d2 c8f5 d2e4 f5e4 d1d6 c7d6 f1d1 e4c2 d1d6 e8g8 c1f4 f8e8 a1c1 c2e4 f2f3 e5g6 f4g3 e4c6 e2c4 g6e7 d6d2 e7f5 g3f4 a8d8 c4d3 f5e7 c1d1 e7g6 d3c2 d8d2 f4d2 e8d8 g1f2 h7h6 g2g3 g6e7 d2a5 d8d1');
INSERT INTO evaluations VALUES(14494,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(14494,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(14494,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(14494,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(14494,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(14494,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(14494,36,'cp 152','f2f4 e5d7 g1f3 c7c6 d5c6 b7c6 e4e5 f6g4 d1e2 e7e6 h2h3 g4h6 g2g4 d6e5 f4e5 a8b8 c3e4 d7c5 c1g5 d8a5 g5d2 a5b5 e4c5 f8c5 b2b3 b5e2 f1e2 h6g8 e1c1 g8e7 h3h4 h7h5 g4h5 e7f5 d2g5 c8d7 h1h3 a7a5 h5h6 f5h6 f3d2 a5a4 d2c4 a4b3 h3b3');
INSERT INTO evaluations VALUES(16377,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(16377,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(16377,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(16377,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(16377,34,'cp -133',NULL);
INSERT INTO evaluations VALUES(16377,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(16377,36,'cp -138','h5e2 d7d5 g1f3 f8c5 c1e3 d5e4 d3e4 f6e4 e3c5 e4c5 e2e3 d8d6 b1d2 f7f6 e1c1 d6e7 d2b3 c5b3 e3b3 e7b4 f1b5 b4b3 b5c6 b7c6 a2b3 e8f7 h1e1 c8e6 e1e4 e6d5 e4a4 a7a5 f3e1 h7h5 f2f3 h5h4 h2h3 f6f5 e1d3 f7f6 c1d2 h8d8');
INSERT INTO evaluations VALUES(16674,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(16674,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(16674,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(16674,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(16674,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(16674,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(16674,36,'cp -125','f6e4 f1d3 e4f6 e1g1 c6d4 c3d4 f8e7 d1c2 e8g8 d4d5 h7h6 b1c3 c7c5 f1e1 f8e8 c1f4 d7d6 h2h3 c8d7 a2a4 e7f8 e1e8 d8e8 f4g3 e8e7 a4a5 a8b8 c2b3 e7d8 g3h4 g7g6 g2g4 f8g7 h4g3 d8e7 c3b5 d7b5 d3b5 a7a6');
INSERT INTO evaluations VALUES(16864,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(16864,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(16864,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(16864,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(16864,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(16864,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(16864,36,'cp -94','f1d3 e4f6 e1g1 c6d4 c3d4 d7d5 c1g5 f8e7 d1c2 e8g8 b1d2 f8e8 a1e1 h7h6 g5h4 f6h5 h4e7 e8e7 d2f3 h5f4 e1e7 d8e7 f1e1 c8e6 d3h7 g8h8 h7f5 e7d8 f5e6 f4e6 f3e5 h8g8 c2b3 b7b6 b3f3 d8f6 f3f6 g7f6');
INSERT INTO evaluations VALUES(19117,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(19117,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(19117,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(19117,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(19117,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(19117,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19117,36,'cp 23','g8f6 c2c4 f8e7 d2d3 e8g8 b1d2 a7a6 a4b3 d7d6 a2a4 a8b8 h2h3 b7b5 a4b5 a6b5 c4b5 b8b5 b3c4 b5b8 d2f3 c6c5 b2b3 c8e6 c1f4 d8d7 f3g5 e6c4 b3c4 b8b6 g5f3 f8b8 d1e2 f6e8 f3d2 e8c7 a1a7 b8b7 a7b7 b6b7 f4g3 c7e6 f2f4');
INSERT INTO evaluations VALUES(23046,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(23046,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(23046,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(23046,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(23046,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(23046,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(23046,36,'cp 141','b4a6 b1c3 g8f6 f2f4 d7d6 f1a6 b7a6 g1f3 f6d7 e1g1 f8e7 f4e5 d6e5 d1d3 e7d6 c1e3 e8g8 f3d2 d8h4 h2h3 d7c5 d3c4 c5b7 d2f3 h4e7 a1e1 b7c5 g1h1 c5d7 a3a4 a8b8 b2b3 b8b4 c4e2 d7c5 f3d2 f7f5 e4f5 c8f5 d2c4 e7h4');
INSERT INTO evaluations VALUES(23173,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(23173,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(23173,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(23173,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(23173,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(23173,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(23173,36,'cp 149','b1c3 d7d6 f1a6 b7a6 d1d3 f8e7 g1e2 g8f6 f2f3 e8g8 c1e3 f6d7 e2g3 e7g5 e3f2 d7c5 d3c4 c5d7 b2b3 d7b6 c4d3 a6a5 a3a4 g5h4 e1g1 g7g6 f2e3 b6d7 d3d2 c8a6 c3b5 f7f5 e4f5 g6f5 d2a5 a6b5 a5b5 f5f4');
INSERT INTO evaluations VALUES(25708,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(25708,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(25708,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(25708,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(25708,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(25708,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(25708,36,'cp 180','b5c6 d7c6 d4e5 c6b5 f1e1 d6e5 d1d8 a8d8 a2a4 b5c6 f3e5 g8e7 f2f3 f7f6 e5d3 e8d7 b1c3 d7c8 b2b4 c6e8 b4b5 e8f7 c1e3 f7c4 e1d1 e7g6 g1f2 f8d6 h2h3 g6e5 a1b1 h8e8 d3b2 a6b5 c3b5 c4a2 b1a1 a2f7 b5d6 d8d6 d1d6 c7d6 a1d1 c8c7 a4a5 e5c6 e3b6 c7d7 d1b1');
INSERT INTO evaluations VALUES(25700,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(25700,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(25700,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(25700,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(25700,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(25700,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(25700,36,'cp 198','d7c6 d4e5 c6b5 f1e1 d6e5 d1d8 a8d8 a2a4 b5d7 f3e5 g8e7 e5d7 d8d7 c1e3 e7c6 b1d2 f8e7 d2b3 f7f6 g1f1 c6e5 a1d1 d7d1 e1d1 e5c4 e3c1 e8f7 b3d4 h8d8 f1e2 g7g6 d4f3 d8d1 e2d1 c4d6 f3d2 b7b5 a4b5 a6b5 b2b3 f6f5 e4f5 d6f5');
INSERT INTO evaluations VALUES(25662,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(25662,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(25662,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(25662,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(25662,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(25662,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(25662,36,'cp 165','d4e5 c6b5 f1e1 d6e5 d1d8 a8d8 a2a4 b5c6 f3e5 g8e7 f2f3 f7f6 e5d3 g7g5 b1c3 e8f7 c1e3 f8g7 a1d1 h7h6 e3c5 a6a5 c3b5 c6b5 a4b5 b7b6 c5e7 f7e7 d1d2 h6h5 g1f2 h5h4 c2c3 h8e8 e1d1 f6f5 e4f5 e7f6 g2g4 h4g3 h2g3 f6f5 g3g4 f5g6 d2e2 e8e2 f2e2');
INSERT INTO evaluations VALUES(25659,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(25659,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(25659,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(25659,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(25659,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(25659,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(25659,36,'cp 184','c6b5 f1e1 d6e5 d1d8 a8d8 a2a4 b5c4 f3e5 c4e6 b1c3 g8e7 c1f4 e7g6 e5g6 h7g6 f4c7 d8c8 c7b6 f8d6 h2h3 e8g8 e1d1 d6b4 c3d5 e6d5 e4d5 c8c2 a1c1 c2c1 d1c1 b4d6 g1f1 f8e8 b6c5 e8c8 c5e3 c8e8 c1d1 f7f6 f1e2 g8f7 e2d3 g6g5 e3d4 e8c8 a4a5 f7e7 d1e1 e7d7 d4b6 d6b4 e1e3 b4d6 e3e6');
INSERT INTO evaluations VALUES(27428,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(27428,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(27428,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(27428,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(27428,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(27428,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(27428,36,'cp -19','e8c8 b2b4 e7e5 b4b5 g4f3 e2f3 d5b5 a2a4 b5a6 c1b2 a6c4 d2d3 c4e6 f1e1 f8c5 a1b1 e6f6 b2c3 f6d6 f3c6 d6c6 d1g4 c8b8 c3e5 f7f6 e5g3 h7h5 g4e4 c6e4 e1e4 b7b6 g1f1 b8b7 f2f3 h8e8 b1e1 e8e4 e1e4 g7g5 g3f2 c5d6 h2h4 f6f5');
INSERT INTO evaluations VALUES(21,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(21,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(21,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(21,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(21,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(21,36,'cp -21','b2b4 e7e5 b4b5 g4f3 e2f3 d5b5 a2a4 b5a6 c1b2 f7f5 a4a5 a6b5 f3c6 b5c6 b2e5 f8d6 f1e1 d6e5 e1e5 g7g6 d2d3 h8e8 e5e8 d8e8 d1d2 a7a6 h2h3 c8b8 c2c4 e8d8 a1d1 c6c5 g2g3 c5d4 d2b4 d4f6 d1b1');
INSERT INTO evaluations VALUES(30513,30,'cp 698',NULL);
INSERT INTO evaluations VALUES(30513,31,'cp 690',NULL);
INSERT INTO evaluations VALUES(30513,32,'cp 701',NULL);
INSERT INTO evaluations VALUES(30513,33,'cp 690',NULL);
INSERT INTO evaluations VALUES(30513,34,'cp 694',NULL);
INSERT INTO evaluations VALUES(30513,35,'cp 690',NULL);
INSERT INTO evaluations VALUES(30513,36,'cp 705','f3b7 b8d7 b7b5 a8b8 b5a6 e7e6 f1c4 f8b4 c1d2 g8f6 g1f3 d8e7 e1g1 e8g8 d2b4 e7b4 b2b3 f8c8 a6a5 b4a5 a1a5 b8a8 b1d2 d7b8 f1c1 b8c6 a5c5 c6e7 f3e5 f6d5 c1a1 c8c5 d4c5 d5b4 f2f3 e7c6 e5c6 b4c6');
INSERT INTO evaluations VALUES(32811,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(32811,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(32811,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(32811,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32811,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(32811,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(32811,36,'cp 24','c2c3 a5b5 d3c2 c6c5 b1d2 c5d4 f3d4 b5d7 d2f3 f8c5 c1e3 b8c6 c2e2 g8e7 h2h4 c5d4 e3d4 h7h6 h4h5 c6d4 f3d4 d7a4 g2g3 e7c6 h1h4 a4a5 a2a3 e8g8 e1c1 a8c8 d4c2 a5c7 f2f4 c7b6 c1b1 g8h8 b1a1');
INSERT INTO evaluations VALUES(32810,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32810,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(32810,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(32810,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(32810,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(32810,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(32810,36,'cp 8','a5b5 d3c2 c6c5 c1e3 c5d4 f3d4 b5d7 b1d2 b8c6 d4c6 d7c6 d2b3 g8e7 b3d4 c6c4 c2b3 c4b3 a2b3 e7g6 d4f3 a7a6 b3b4 f8e7 e1e2 e8d7 h1d1 a8c8 g2g3 h7h5 b4b5 a6b5 a1a7 c8b8 f3e1 h8c8 e1d3 c8c4');
INSERT INTO evaluations VALUES(34031,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(34031,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(34031,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(34031,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(34031,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(34031,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34031,36,'cp 12','e7e5 f3e5 f6e4 d2d3 e4f6 b1c3 d7d6 e5f3 c8g4 c1g5 g4f3 g5f6 f3g4 c3d5 c7c6 d5e3 g4e6 f6c3 b8a6 h2h4 h7h5 f2f3 d6d5 d3d4 a6c7 a2a4 a7a5 e1f2 e8d7 f1d3 c7e8 c3d2 e8d6 b2b3 g7g6 g2g3 f8h6 a1e1 h8e8 c2c3 h6e3 d2e3');
INSERT INTO evaluations VALUES(33970,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(33970,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(33970,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33970,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(33970,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(33970,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(33970,36,'cp 11','f3e5 f6e4 d2d3 e4f6 b1c3 d7d6 e5f3 c8g4 f1e2 b8c6 h2h3 g4e6 c1e3 a7a6 e1c1 d6d5 c3a4 e8c8 c2c3 d8e8 d3d4 h7h6 a4c5 e6f5 f3h4 f5h7 e2d3 g7g5 d3h7 f6h7 h4f5 c6e7 f5e7 f8e7 a2a4 h7f8 b2b4 f8g6 c5d3 g6h4');
INSERT INTO evaluations VALUES(33965,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(33965,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(33965,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(33965,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(33965,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(33965,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(33965,36,'cp 10','f6e4 d2d3 e4f6 b1c3 d7d6 e5f3 c8g4 f1e2 b8c6 h2h3 g4e6 c1e3 a7a6 e1c1 d6d5 c3a4 e8c8 d3d4 h7h6 a4c5 e6f5 e2d3 f5d3 c5d3 f8d6 c2c3 f6e4 d1e1 d8e8 g2g3 f7f6 c1c2 c8d7 f3d2 f6f5 e3f4 d6f4 g3f4');
INSERT INTO evaluations VALUES(33990,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(33990,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33990,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33990,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33990,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33990,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33990,36,'cp 13','d2d3 e4f6 b1c3 d7d6 e5f3 c8g4 c1g5 g4f3 g5f6 f3g4 f6g5 b8c6 e1d2 g4e6 c3e2 d6d5 g5e3 c6e5 e3f4 f8d6 g2g3 e8c8 f1g2 h7h6 e2d4 g7g5 f4e3 e5g4 d4b5 g4e3 f2e3 c7c6');
INSERT INTO evaluations VALUES(34310,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(34310,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(34310,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(34310,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(34310,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(34310,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(34310,36,'cp 39','d2d3 g8f6 g1f3 b8d7 c1f4 g7g6 e1c1 f6d5 c3d5 c6d5 d3d4 h7h5 e2d3 d7b6 h1e1 c8g4 f3e5 f8h6 f4h6 h8h6 f2f3 g4f5 d3f5 g6f5 b2b3 e7e6 c1d2 e8e7 a2a4 a8g8 e1e2 a7a5 h2h4 b6d7 e5d3 h6h8 d1b1 g8c8 c2c3 h8g8');
INSERT INTO evaluations VALUES(37817,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37817,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(37817,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37817,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37817,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37817,36,'cp 2','d2d3 e7e6 g1e2 b8c6 f1g2 d7d6 e1g1 g8e7 c1g5 a8b8 a2a4 a7a6 d1d2 h7h6 g5e3 b7b5 a4b5 a6b5 d3d4 b5b4 c3a4 c5d4 e2d4 c6e5 d2e2 e8g8 b2b3 c8b7 c2c4 b4c3 a4c3 d6d5 f1d1 d8d7 d4b5 b7c6 b5d4');
INSERT INTO evaluations VALUES(42489,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(42489,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(42489,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(42489,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(42489,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(42489,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(42489,36,'cp 12','c1d2 b4e7 g1f3 e8g8 f1g2 d7d5 e1g1 c7c6 d1c2 b7b6 d2f4 c8b7 f1d1 b8d7 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 d8c7 e2e3 a8d8 a1c1 d7f6 c4e5 c6c5 d4c5 d8d1 c1d1 b7g2 g1g2 e7c5 c2e2 f8d8 d1d8 c7d8 e2f3 c5d6 e5c6 d8d7 a2a4 a7a6 b2b4 d6c7');
INSERT INTO evaluations VALUES(42488,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(42488,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(42488,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(42488,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(42488,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(42488,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(42488,36,'cp 12','b4e7 g1f3 e8g8 f1g2 d7d5 d1c2 c7c6 e1g1 b7b6 f1d1 b8d7 d2f4 c8b7 b1c3 d5c4 f3d2 f6d5 d2c4 d5f4 g3f4 d8c7 e2e3 a7a6 a1c1 a8d8 c3e4 h7h6 a2a3 c6c5 d4c5 b6c5 e4c3 b7g2 g1g2 d7f6 d1d8 f8d8 c2a4 c7b7 f2f3');
INSERT INTO evaluations VALUES(53889,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(53889,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53889,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(53889,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(53889,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(53889,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(53889,36,'cp 38','g7g6 a2a3 a7a5 d2d3 b8c6 c3d5 h7h6 c1d2 g8f6 b2b4 a5b4 a3b4 a8a1 d5f6 d8f6 d1a1 c6b4 a1a8 f6d8 d2b4 c5b4 g1f3 f8e7 a8a4 c8d7 a4b4 d6d5 b4b7 d8a5 e1f1 d5c4 b7b8 e7d8 b8e5 a5e5 f3e5 c4d3 e2d3 d7f5 g2c6 e8f8 d3d4 d8c7 f1e2 c7e5 d4e5 f8g7');
INSERT INTO evaluations VALUES(55155,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(55155,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(55155,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(55155,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(55155,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(55155,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(55155,36,'cp -51','e7e5 e2e3 f8d6 d2d4 e5e4 c2c4 c7c6 b1c3 g8f6 g1e2 h7h5 c4d5 c6d5 e2f4 d6f4 e3f4 b8c6 f1e2 c6e7 d1b3 e8g8 c1e3 b7b6 e1g1 e7f5 f1c1 f8e8 b3d1 h5h4 e2g4 d8d7 g4f5 d7f5 c1c2 c8a6 a1c1 a8c8 c3e2');
INSERT INTO evaluations VALUES(57011,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57011,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57011,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57011,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57011,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(57011,36,'cp 0','g8f6 e2e3 d7d5 d2d4 c7c5 g1f3 d5c4 f1c4 a7a6 d4c5 d8d1 e1d1 f8c5 d1e2 c5e7 b2b4 b8d7 c1b2 b7b5 c4d3 c8b7 h1c1 e7d6 b1d2 e8g8 g2g3 a8d8 e3e4 e6e5 b2c3 b7c6 c3b2 c6b7');
INSERT INTO evaluations VALUES(49389,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(49389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49389,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49389,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(49389,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49389,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49389,36,'cp 0','d2d4 d7d5 e2e3 c7c5 g1f3 d5c4 f1c4 a7a6 e1g1 b7b5 c4d3 c5d4 e3d4 c8b7 a3a4 b5a4 d1a4 b7c6 a4c4 f8d6 b1c3 e8g8 c1e3 c6f3 g2f3 b8d7 a1a6 d7b6 a6a8 d8a8 c4a6 a8f3 a6b6 f3g4 g1h1 g4f3');
INSERT INTO evaluations VALUES(5855,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(5855,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(5855,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(5855,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(5855,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(5855,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(5855,36,'cp 204','b1c3 g8e7 a2a3 b4a5 b2b4 a5b6 c1e3 e8g8 d1d3 h7h6 a1d1 e7g6 h2h4 e6e5 h4h5 g6f4 e3f4 e5f4 d4d5 c6e7 e4e5 a7a5 c3a4 b6a7 b4a5 c8g4 a4c3 e7f5 c3e4 c7c6 e1g1 c6d5 d3d5 d8d5 d1d5 a8d8 e4d6 f5d6');
INSERT INTO evaluations VALUES(18008,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(18008,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18008,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18008,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18008,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(18008,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18008,36,'cp 0','c6c5 c3e2 g8f6 d2d3 f8e7 f3d2 e8g8 d2c4 c8e6 e2g3 f6d7 c1d2 e7g5 b2b3 f8e8 d2g5 d8g5 c4e3 a6a5 a2a4 d7f8 d1f3 g5f4 f3e2 f4g5');
INSERT INTO evaluations VALUES(24002,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24002,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(24002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24002,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(24002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24002,36,'cp 12','h2h3 e8g8 a2a4 a7a6 d4c6 d7c6 b5c6 b7c6 a4a5 f6d7 d1d3 d8c8 c3a4 c6c5 a4c3 c8b7 c3d5 d7e5 d3b3 a8b8 b3a4 f8e8 f1d1 h7h6 d5b6 e8d8 b6d5');
INSERT INTO evaluations VALUES(27637,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(27637,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(27637,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(27637,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(27637,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(27637,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(27637,36,'cp 52','h2h3 g4h5 a2a3 e7e6 a1b1 a5b6 b2b4 h5f3 e2f3 c6d4 e1g1 g8f6 c1b2 d4f3 d1f3 b6c6 f3c6 b7c6 f1e1 e8d7 c3e4 f6e4 e1e4 f7f6 b2c3 f8d6 b1e1 e6e5 g1f1 a7a6 e4c4 h8e8 f2f3 h7h5 a3a4 a8b8 d2d4 e5d4 e1e8 d7e8 c4d4 e8d7 d4c4 b8b6');
INSERT INTO evaluations VALUES(33162,30,'cp 619',NULL);
INSERT INTO evaluations VALUES(33162,31,'cp 648',NULL);
INSERT INTO evaluations VALUES(33162,32,'cp 658',NULL);
INSERT INTO evaluations VALUES(33162,33,'cp 662',NULL);
INSERT INTO evaluations VALUES(33162,34,'cp 651',NULL);
INSERT INTO evaluations VALUES(33162,35,'cp 671',NULL);
INSERT INTO evaluations VALUES(33162,36,'cp 640','d2d4 c5b6 f3e5 g8e7 e5c4 d6c6 c4b6 c7b6 f1b5 c6g6 d5c7 e8d8 c7a8 g6g2 h1f1 d7d5 f2f3 g2h2 d1e2 h2e2 b5e2 f7f6 c1f4 b8a6 f1g1 g7g5 f4h2 h8e8 e2a6 b7a6 a8c7 d5e4 c7e8 d8e8 f3e4 e8d7 d4d5 e7g6 e1f2 c8b7 a2a4');
INSERT INTO evaluations VALUES(43037,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(43037,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43037,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(43037,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(43037,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(43037,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43037,36,'cp 47','d7d6 g1f3 c7c5 b2b3 b8c6 e1g1 c8f5 d2d3 d8d7 f3g5 e8g8 a1b1 f6e8 c3d5 e7g5 c1g5 e8c7 e2e4 f5g4 f2f3 g4h5 g5e3 a8c8 d1d2 c7d5 c4d5 c6d4 g3g4 h5g6 h2h3 b7b5 f3f4 e5f4 e3f4 b5b4 g1h2 f7f6 b1c1');
INSERT INTO evaluations VALUES(49219,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(49219,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(49219,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(49219,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(49219,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(49219,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(49219,36,'cp -30','e2e3 d7d5 d2d4 b8d7 g2g3 c7c6 f1g2 f6f5 g1e2 d7f6 c2c4 h7h5 h2h4 f8b4 b1c3 d8a5 d1d3 f6e4 a1c1 e8g8 e1g1 c8e6 a2a4 a8c8 f1d1 e4c3 e2c3 f5f4 e3f4 e6f5 d3e3 f8e8 e3d2 g8g7 c4d5');
INSERT INTO evaluations VALUES(56376,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(56376,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(56376,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(56376,33,'cp -119',NULL);
INSERT INTO evaluations VALUES(56376,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(56376,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(56376,36,'cp -109','d5e4 d4e5 d8d1 e1d1 e4f3 g1f3 b8c6 f1d3 g8e7 b1d2 e7g6 c2c3 c8d7 h1e1 e8c8 d1c2 d8e8 d2c4 b7b5 e5e6 e8e6 e1e6 f7e6 c4d2 c6e5 d3e2 f8e7 f3e5 g6e5 d2f3 e5f3 e2f3 h8f8 c1e3 a7a5 a1d1 e7d6 h2h3 h7h6 b2b3 d7e8 c2b2 h6h5 e3d4 e6e5 d4e3');
INSERT INTO evaluations VALUES(1802,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(1802,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(1802,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(1802,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(1802,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(1802,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(1802,36,'cp 97','e1g1 f8c5 b1c3 g8e7 c1e3 c5b6 e3b6 a7b6 c3b5 e8d8 d3d4 d7d6 d4d5 c6a7 b5c3 e7g6 a2a4 g6f4 f3e1 c8d7 a4a5 a7c8 a5a6 b7a6 c4a6 f6g6 f2f3 g6g5 g1h1 h6h5 g2g3 f4h3 e1g2 g5h6 g2h4 g7g5 h4f5 h6f6 f5e3 h5h4 g3g4 h3f4 d1d2');
INSERT INTO evaluations VALUES(1801,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(1801,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(1801,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(1801,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(1801,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(1801,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(1801,36,'cp 48','g8e7 c2c3 d7d6 c1e3 g7g5 d3d4 a7a6 b1d2 f8g7 a2a4 e5d4 f3d4 e7g6 d4c6 b7c6 e3d4 g6e5 c4d3 f6e6 d4e5 e6e5 d2c4 e5f4 d1c2 e8g8 a1e1 a8b8 g2g3 f4f6 e4e5 d6e5 c4e5 c8h3 d3a6 g8h8 a6d3 h3f1 e1f1');
INSERT INTO evaluations VALUES(6918,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(6918,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(6918,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(6918,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(6918,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(6918,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(6918,36,'cp 114','c2c3 a7a6 d2d4 c5a7 b5c4 d7d6 c1e3 g8e7 f1e1 c6a5 d1a4 a5c6 a4c2 c6a5 c4f1 a5c6 a2a4 g7g5 d4d5 a7e3 e1e3 c6b8 c3c4 g5g4 f3h4 e8g8 b1c3 f6f5 e4f5 e7f5 h4f5 c8f5 f1d3 d8g5 c4c5 b8d7 d3f5 g5f5 c2f5 f8f5 c5d6 c7d6');
INSERT INTO evaluations VALUES(6916,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(6916,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(6916,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(6916,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(6916,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(6916,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(6916,36,'cp 112','a7a6 b5c4 c5a7 a2a4 d7d6 d2d4 g8e7 h2h3 e7g6 c1e3 d8e7 b1d2 c6d8 a4a5 c8e6 d1a4 e7d7 a4d7 e6d7 b2b4 d8f7 f1c1 g6e7 d2f1 h8f8 c4d3 f6f5 d4e5 f7e5 f3e5 a7e3 f1e3 d6e5 e4f5 d7f5 e3f5');
INSERT INTO evaluations VALUES(13896,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(13896,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(13896,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(13896,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(13896,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(13896,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(13896,36,'cp 93','c1g5 c8f5 e4f5 f8e7 g2g3 c6d4 f1g2 d4f5 d1d3 g7g6 e1c1 f6g4 g5e7 d8e7 c3d5 e7d8 d1d2 e8g8 g2h3 g4h6 h1d1 g8g7 d5e3 d8a5 e3f5 h6f5 h3f5 a5a2 c2c3 g6f5 d3f5 a2e6 f5e6 f7e6 d2d6 f8f2 d6e6 a8f8 e6e5 f2h2 e5e7 g7g6 e7b7 f8f2');
INSERT INTO evaluations VALUES(13897,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(13897,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(13897,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(13897,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(13897,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(13897,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(13897,36,'cp 93','c8f5 e4f5 f8e7 g2g3 c6d4 f1g2 f6g8 g5c1 g8f6 g2b7 a8b8 b7g2 e8g8 e1g1 d4f5 b2b3 d8c8 c1b2 e7d8 a1c1 f8e8 c3d5 f6d5 g2d5 g7g6 c2c4 h7h5 d1d3 d8b6 h2h4 g8g7 b3b4 b6d4 b2d4 e5d4 d3d2 e8e5 b4b5');
INSERT INTO evaluations VALUES(24740,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(24740,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(24740,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(24740,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(24740,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(24740,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(24740,36,'cp 100','e4f5 f8e7 g2g3 d8d7 f1g2 c6d4 e1g1 d4f5 g5f6 e7f6 g2h3 g7g6 c3d5 f6g7 d5e3 e8g8 h3f5 g6f5 d1h5 f5f4 e3f5 d7e6 a1d1 a8d8 h5g4 e6f6 d1d5 g8h8 f1d1 f6g6 g4g6 h7g6 f5d6 d8d7 g1g2 f4g3 h2g3 h8g8 c2c4 b7b6 g3g4 f8d8 g4g5 g8f8 b2b4 f7f5 g5f6 g7f6 c4c5 b6c5 b4c5');
INSERT INTO evaluations VALUES(24739,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(24739,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(24739,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(24739,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(24739,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(24739,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(24739,36,'cp 85','f8e7 g2g3 c6d4 f1g2 d4f5 d1d3 g7g6 h2h4 e8g8 g5f6 e7f6 e1c1 h7h5 g2e4 f5d4 c1b1 d8d7 f2f3 a8c8 g3g4 c8c3 b2c3 d6d5 c3d4 d5e4 d3b3 e5d4 g4h5 d7f5 f3e4 f5h5 d1f1 h5e5 h4h5 d4d3 c2d3 f6g7 h5g6 e5a1 b1c2');
INSERT INTO evaluations VALUES(24031,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(24031,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(24031,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(24031,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(24031,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(24031,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(24031,36,'cp 80','g2g3 c6d4 f1g2 f6g8 g5c1 d8c8 d1d3 c8f5 g2e4 f5g4 h2h3 g4e6 c3b5 e6d7 b5d4 e5d4 d3d4 g8f6 e4g2 d6d5 e1g1 e8g8 c1e3 b7b6 d4d3 a8d8 a1d1 f8e8 e3d4 d7c6 c2c3 f6e4 f1e1 e7f6 d4f6 e4f6 e1e8 c6e8 g2f3 e8e6 g1g2 d8d7 d3d2 d5d4 c3d4 e6a2');
INSERT INTO evaluations VALUES(21948,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(21948,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(21948,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(21948,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(21948,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(21948,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(21948,36,'cp 66','e5d4 f3d4 g8f6 b1c3 f8e7 d4c6 b7c6 f1e2 e8g8 e1g1 d6d5 e4e5 f6d7 f2f4 d7b6 f4f5 b6d7 f5f6 g7f6 e5f6 e7f6 e2d3 d7e5 d1h5 e5d3 c2d3 d8d6 f1f3 d6e5 h5h6 f6g7 h6c6 a8b8 c6d5 c7c6 d5e5 g7e5 d3d4 e5g7 b2b3 f8e8 e3f2 c8g4 f3g3 g4h5 a1f1');
INSERT INTO evaluations VALUES(23042,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(23042,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(23042,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(23042,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(23042,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(23042,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(23042,36,'cp 116','b4a6 f1a6 b7a6 g1e2 g8f6 d1d3 a6a5 e1g1 f8c5 c3c4 e8g8 h2h3 f6h5 b1c3 d8h4 c1e3 c5e3 d3e3 d7d6 a1e1 c8d7 f2f4 h5f4 e2f4 e5f4 f1f4 h4g5 e1f1 f7f6 b2b3 f8b8 c3e2 a5a4 e2d4 a4b3 a2b3 g5e5');
INSERT INTO evaluations VALUES(23171,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(23171,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(23171,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(23171,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(23171,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(23171,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(23171,36,'cp 111','f1a6 b7a6 g1e2 g8f6 d1d3 a6a5 e1g1 f8c5 c3c4 d7d6 b1c3 e8g8 h2h3 c5b6 b2b3 f6d7 c3a4 d8h4 e2g3 d7c5 a4c5 b6c5 c1d2 f7f5 e4f5 e5e4 d3c3 c8f5 d2e3 c5b6 c4c5 d6c5 e3c5 h4f6 c3c4 f8f7 c5b6 c7b6 g3f5 f6f5 d5d6 a8c8 c4d4 c8e8 a1e1');
INSERT INTO evaluations VALUES(23169,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(23169,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(23169,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(23169,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(23169,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(23169,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(23169,36,'cp 111','b7a6 g1e2 g8f6 d1d3 a6a5 c3c4 f8c5 e1g1 d7d6 b1c3 e8g8 h2h3 f6d7 b2b3 d8h4 c3a4 c5b6 c1a3 c8a6 e2g3 g7g6 g1h2 b6d4 a1e1 a6c8 f2f3 d7b6 a3b2 d4b2 a4b2 f7f5 e4f5 c8f5 d3e3 b6d7 b2a4 d7f6 a4c3 a8e8 c4c5 d6c5');
INSERT INTO evaluations VALUES(18338,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(18338,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(18338,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(18338,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(18338,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(18338,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(18338,36,'cp 100','g1e2 g8f6 d1d3 a6a5 c3c4 f8c5 e1g1 d7d6 b1c3 e8g8 h2h3 c5b6 c1d2 f6d7 c3a4 d7c5 a4c5 b6c5 e2g3 c5d4 d2c3 d4b6 g1h2 d8h4 c3d2 a5a4 f2f4 b6d4 a1b1 a8b8 b2b3 a4b3 a2b3 a7a5 f4f5 a5a4 b3b4 c8d7 d2e3 h4f6 c4c5');
INSERT INTO evaluations VALUES(29785,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(29785,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29785,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(29785,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29785,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(29785,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29785,36,'cp 14','g1f3 e7e6 c2c4 d5c4 d3c4 f8d6 f4g3 d6g3 h2g3 c7c5 b1c3 e8g8 d4c5 d8a5 e1g1 a5c5 c4e2 b8d7 a1c1 c5b6 c3a4 b6a5 d1b3 a7a6 a2a3 f8d8 f1d1 h7h6 f3d4 e6e5 d4f3 b7b5 a4c5 d7c5 c1c5 d8d1 b3d1');
INSERT INTO evaluations VALUES(29786,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29786,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29786,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(29786,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(29786,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(29786,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29786,36,'cp 15','e7e6 c2c4 f8d6 d3g6 h7g6 d1b3 d6f4 b3b7 b8d7 e3f4 c7c5 b1c3 d5c4 d4c5 e8g8 b7b5 a8b8 b5c4 d8a5 e1g1 d7c5 c4e2 a5b4 f1c1 b4f4 g2g3 f4f5 c1c2 f8d8 a1d1 d8d1 c3d1 b8d8 d1e3 f5h5 b2b3 c5d3 c2d2 d8d7 d2d3 d7d3 e2d3 h5f3');
INSERT INTO evaluations VALUES(34610,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(34610,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(34610,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(34610,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(34610,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(34610,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(34610,36,'cp 89','f1e2 e5g5 e2f3 g5c5 b1a3 e7e5 g1e2 g8f6 d2d4 c5a5 c1d2 f8b4 d4e5 b4d2 d1d2 a5e5 e1g1 e8g8 f1e1 c7c6 e2g3 e5c7 a1d1 c8e6 a3c2 b8d7 c2d4 d7e5 d4e6 f7e6 f3e2 e5g6 d2e3 c7b6 e3c3 a8d8 h2h3 d8d1 e1d1 f8d8 d1d8 b6d8 e2f3 d8d7 b2b3');
INSERT INTO evaluations VALUES(34611,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(34611,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(34611,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(34611,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(34611,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(34611,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(34611,36,'cp 81','e5g5 e2f3 g5c5 b1a3 e7e5 g1e2 g8f6 d2d4 c5a5 c1d2 f8b4 d4e5 b4d2 d1d2 a5e5 e1g1 e8g8 f1e1 c7c6 e2g3 e5c7 a3c2 f8d8 c2d4 b8a6 d2c3 c8d7 a2a3 d8e8 b2b4 e8e1 a1e1 a8e8 c4c5 e8e1 c3e1 c7d8 d4b3 a6c7 b3a5 d8b8 e1e5 d7e6 g3f5 e6f5 e5f5 h7h6 g2g3 c7e6 a5c4');
INSERT INTO evaluations VALUES(37373,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(37373,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(37373,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(37373,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(37373,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(37373,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(37373,36,'cp -13','f6e4 c3e4 d7d5 c4d3 d5e4 d3e4 c7c5 c2c3 b8d7 e1g1 c5d4 f3d4 e7e5 d4b5 d7c5 d1d8 f8d8 e4f3 c8f5 c1e3 g7f8 f1d1 f5d3 b5c7 a8c8 e3c5 f8c5 c7d5 d3f5 d1e1 c5d6 a1d1 g8g7 g2g3 b7b5 d5e3 f5e6 f3d5 g7f6 d5b3 d6c5 e3d5 e6d5 d1d5 d8d5 b3d5 b5b4 c3c4 g6g5 e1d1 g5g4');
INSERT INTO evaluations VALUES(38761,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(38761,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(38761,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(38761,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(38761,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(38761,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(38761,36,'cp 128','d2d4 c5b6 b2b4 a7a5 b4b5 c7c5 c2c3 b8d7 f1d3 f7f6 e1g1 f6e5 f3e5 d7e5 d4e5 g8e7 a2a4 e8g8 b1d2 e7f5 d1g4 c8d7 c3c4 d7e8 c4d5 d8d5 a1a3 e8g6 d3e4 d5d4 h2h4 a8e8 h4h5 f5h6 g4e2 g6e4 e2e4 b6c7 e4d4 c5d4 d2c4 h6f5');
INSERT INTO evaluations VALUES(38759,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(38759,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(38759,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(38759,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(38759,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(38759,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(38759,36,'cp 115','c5e7 c2c3 a7a6 b2b4 c8d7 f1e2 d7b5 e1g1 b5e2 d1e2 b8d7 b1d2 d7b6 d2b3 d8d7 e2c2 e7d8 b3a5 d7c8 a2a4 g8e7 c2d3 h7h6 h2h4 g7g6 h4h5 g6h5 b4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(41316,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(41316,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(41316,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(41316,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(41316,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(41316,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(41316,36,'cp -22','b8c6 g1f3 e8g8 e1g1 f6h5 f4g5 f7f6 g5h4 e7e5 f1e1 g8h8 b1d2 d8e8 e3e4 a7a5 a2a4 c8d7 d3b5 c6d8 b5d7 e8d7 d4e5 d6e5 h4g3 d8e6 d2c4 d7e7 d1d5 c7c6 d5d6 e7d6 c4d6 e6c5 e1d1 h5f4');
INSERT INTO evaluations VALUES(41918,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(41918,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(41918,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(41918,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(41918,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(41918,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(41918,36,'cp 14','e1g1 e8g8 c3d5 f6d5 c4d5 b4c5 c2c3 b8c6 a2a4 a7a5 d2d4 e5d4 f3d4 c6d4 c3d4 c5b4 c1d2 c7c6 d5b3 b4d2 d1d2 d8b6 a1a3 c8e6 b3e6 f7e6 h2h3 f8f7 d2c3 a8f8 f2f3 b6d8 f1e1 d8b6 a3b3 b6d8 c3e3 d6d5 e1e2 d8h4 e3d2 h4g3 e2e3');
INSERT INTO evaluations VALUES(45517,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(45517,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(45517,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(45517,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(45517,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(45517,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(45517,36,'cp -21','b1c3 e4c3 d2c3 b8c6 f2f4 g7g6 f1d3 f8g7 g1f3 e8g8 e1g1 d5d4 c3d4 c6d4 c2c3 d4f5 d3f5 c8f5 f1e1 f8e8 h2h3 d8c7 a2a4 b7b6 f3d4 f5d7 e3f2 a8d8 c1e3 e7e5 f4e5 g7e5 e1f1 d7e6 d4e6 e8e6 a4a5 e5g3 f2f3 b6a5 a1a5 e6e3 f3e3 c7a5 e3g3');
INSERT INTO evaluations VALUES(48225,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(48225,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(48225,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(48225,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(48225,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(48225,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(48225,36,'cp -18','e2e3 c7c5 b4c5 f8c5 g1f3 e8g8 c2c4 b8c6 d2d4 c5e7 b1d2 c6a5 b2c3 a5c4 d2c4 d5c4 f1c4 b7b6 e1g1 c8b7 d1b3 f6e4 c3e1 a8c8 c4d3 e4f6 a3a4 b7f3 g2f3 f6d5 f3f4 a7a5 f4f5 e7b4 e1b4 a5b4 f5e6 f7e6');
INSERT INTO evaluations VALUES(49210,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(49210,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(49210,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(49210,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(49210,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(49210,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(49210,36,'cp 25','f8g7 g1f3 b8c6 e1g1 e8g8 d2d3 d7d6 a1b1 a7a5 b2b3 h7h6 e2e4 f6h7 d1d2 g6g5 h2h4 c6d4 h4g5 d4f3 g2f3 h7g5 f3g2 f7f5 e4f5 c8f5 d2d1 c7c6 c3e4 g5e6 d1h5 d8d7 b3b4 d7f7 h5f7 f8f7 b4b5 d6d5 b5c6 d5e4 c6b7 a8b8 d3e4');
INSERT INTO evaluations VALUES(51081,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(51081,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(51081,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(51081,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(51081,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(51081,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(51081,36,'cp 141','a2a3 d7d6 b1c3 c7c6 h2h3 e7e5 g2g4 g8f6 g4g5 f6d7 h3h4 f8e7 g1f3 e8g8 c1e3 b8a6 f1e2 b7b5 h4h5 b5b4 a3b4 a6b4 h5h6 e5d4 f3d4 g7g6 a1a4 a8a4 c3a4 e7g5 c2c3 g5e3 f2e3 b4a6 d4c6 d8f6 d1d4 f8e8 d4f6 d7f6 a4b6 c8b7 c6a5 a6c5 a5b7 c5b7 b2b4 e8e4');
INSERT INTO evaluations VALUES(51082,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(51082,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(51082,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(51082,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(51082,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(51082,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(51082,36,'cp 135','d7d6 b1c3 c7c6 h2h3 g8f6 c1e3 e7e5 g2g4 f8e7 g4g5 f6d7 g1f3 e8g8 h3h4 b7b5 f1e2 b8a6 h4h5 b5b4 a3b4 a6b4 h5h6 g7g6 e1g1 d6d5 d4e5 d7c5 e3c5 e7c5 d1d2 d8d7 d2f4 d7g4 f4g4 c8g4 e4d5 b4d5 c3e4 g4f3 e2f3 c5d4 c2c4 d5e7 a1d1 d4e5 e4f6 e5f6 g5f6 e7f5 f3c6');
INSERT INTO evaluations VALUES(55560,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(55560,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55560,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(55560,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55560,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(55560,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(55560,36,'cp 9','g7g6 g2g3 f8g7 f1g2 g8e7 g1f3 b8c6 a1b1 a7a5 e1g1 e8g8 e2e3 d6d5 c4d5 e7d5 a2a3 h7h6 d1c2 d5c3 b2c3 a5a4 f3d2 c8f5 e3e4 f5e6 b1b7 c6a5 b7b4 d8d7 b4a4 f8d8 d2c4 a5c4 a4a8 d8a8 d3c4 d7c6');
INSERT INTO evaluations VALUES(56515,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(56515,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(56515,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(56515,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(56515,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(56515,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(56515,36,'cp -43','d8e7 e5c4 b8c6 c3d5 e7d8 d5f4 g8f6 d2d3 f8e7 c4e3 e8g8 d3e4 d8d1 e3d1 f6e4 f4d5 e7d8 c1e3 c8f5 f1e2 e4f6 d5f6 d8f6 c2c3 a7a5 a2a4 f5e6 e3f4 c6e5 d1e3 a8d8 f4e5 f6e5 e3c4 d8d5 e2f3 d5c5 c4e5 c5e5 e1d2 e6b3 a1a3 b3d5 f3d5 e5d5 d2c2');
INSERT INTO evaluations VALUES(6482,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(6482,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(6482,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(6482,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(6482,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(6482,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(6482,36,'cp -20','g8f6 f1g2 d7d6 e1g1 a7a5 d2d3 e8g8 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 d1d2 c8e6 c2c3 f8e8 f1d1 e6d5 e4d5 c6e7 d1e1 d8d7 d3d4 e5d4 f3d4 c5d4 d2d4 e7g6 b2b3 e8e1 a1e1 a8e8 e1e4 b7b6 f2f4 e8e4 d4e4 g6e7 c3c4 e7f5 e4d3 g7g6');
INSERT INTO evaluations VALUES(16002,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(16002,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(16002,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(16002,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(16002,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(16002,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(16002,36,'cp 47','e7e5 g1f3 f8e7 e1g1 e8g8 a2a4 c7c6 h2h3 h7h6 f1e1 f8e8 c4a2 a7a5 c1d2 e5d4 f3d4 e7f8 d1f3 d7e5 f3g3 f6h5 g3e3 d8b6 a1d1 b6b2 f2f4 e5g6 e3f3 h5f4 d2f4 g6f4 f3f4');
INSERT INTO evaluations VALUES(15912,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(15912,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(15912,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(15912,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(15912,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(15912,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(15912,36,'cp 63','g1f3 f8e7 a2a4 e8g8 e1g1 c7c6 f1e1 h7h6 h2h3 f8e8 a4a5 e5d4 f3d4 e7f8 c1f4 d7e5 c4b3 e5d7 b3a2 d7c5 d1f3 c5e6 a2e6 c8e6 e4e5 d6e5 f4e5 f6d7 e5g3 d7c5 e1d1 d8c8 b2b4 c5d7 a5a6 f8b4 a6b7 c8b7 d4c6 a8c8');
INSERT INTO evaluations VALUES(20826,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(20826,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(20826,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(20826,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(20826,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(20826,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(20826,36,'cp 43','c8f5 b5c6 b7c6 g1f3 e7e6 e1g1 g8e7 f3h4 d8d7 b1d2 f6e5 d2b3 e7g6 h4f5 e6f5 d4e5 f8e7 f2f4 g6f8 c1e3 f8e6 d1d3 e8g8 a1d1 a7a5 a2a4 g7g6 b3d4 e6d4 d3d4 f8d8 d4c3 e7b4 c3d3 b4e7 b2b3 d7e6 h2h3 c6c5 d3e2 h7h5 f1e1 d5d4 e3d2');
INSERT INTO evaluations VALUES(25409,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(25409,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(25409,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(25409,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(25409,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(25409,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(25409,36,'cp 94','e1g1 f8e7 c4e6 f7e6 e4e5 d6e5 d4e5 d8d6 e5b5 d6c6 f1e1 e8f7 b5c6 b7c6 e1e2 h8b8 b2b3 f6d5 c3a4 b8d8 c1b2 d5b6 a4c3 e7f6 a1e1 d8d6 b2a1 a8d8 g1f1 a7a5 g2g3 c6c5 c3b5 d6c6 a1e5 f6e5 e2e5 b6d7 e5e3 d7f6 e3c3 d8d5 e1e2 c5c4 c3c4');
INSERT INTO evaluations VALUES(26703,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26703,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(26703,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(26703,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(26703,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(26703,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26703,36,'cp 30','a7a6 c2c3 b4c6 f1d3 f5d3 d1d3 e7e6 e1g1 f8d6 f4g3 g8f6 c3c4 e8g8 a1c1 c6e7 d3b3 a8b8 a3b1 c7c6 b3c2 b8a8 c2e2 f8e8 b1c3 e7g6 f1d1 d6g3 h2g3 d8e7 b2b3 a8d8 c3a4 e6e5 d4e5 g6e5 c4d5 d8d5 d1d5 f6d5');
INSERT INTO evaluations VALUES(29651,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(29651,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(29651,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29651,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29651,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29651,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29651,36,'cp 12','e7e6 e2e3 f8d6 f4d6 d8d6 c2c4 e8g8 f1e2 b8d7 e1g1 a7a5 h2h3 g4h5 d1c2 c7c6 a2a3 d6e7 a1c1 h5g6 e2d3 g6d3 c2d3 h7h6 f1e1 e7d6 c1c2 a5a4 e3e4 d5e4 d2e4 f6e4 d3e4 d7f6 e4e3 f8d8 c4c5 d6c7 e3c1 f6d5 c2c4 b7b5 c5b6 c7b6');
INSERT INTO evaluations VALUES(30169,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(30169,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(30169,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(30169,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(30169,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(30169,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(30169,36,'cp 22','e2e3 f8d6 h2h3 g4h5 f4d6 d8d6 c2c4 e8g8 f1e2 c7c5 d4c5 d6c5 a1c1 c5e7 c4d5 f6d5 e1g1 f8d8 d1b3 b8c6 d2e4 h7h6 f1d1 a8c8 e4c3 d5f6 d1d8 c8d8 c1d1 d8d1 e2d1 a7a6 b3b6 e7b4 b6b4 c6b4 a2a3 b4d3 b2b4 h5f3 d1f3');
INSERT INTO evaluations VALUES(29741,30,'cp 337',NULL);
INSERT INTO evaluations VALUES(29741,31,'cp 328',NULL);
INSERT INTO evaluations VALUES(29741,32,'cp 339',NULL);
INSERT INTO evaluations VALUES(29741,33,'cp 331',NULL);
INSERT INTO evaluations VALUES(29741,34,'cp 335',NULL);
INSERT INTO evaluations VALUES(29741,35,'cp 324',NULL);
INSERT INTO evaluations VALUES(29741,36,'cp 332','f3e5 g4e6 c4e6 f7e6 e1g1 d8c8 d2d4 g7g6 d1f3 f8g7 c3e4 e8g8 e4g5 f6h5 f3h3 c7c5 g5e6 c5d4 e5g4 f8f5 h3b3 g8h8 g4h6 f5f6 e6d4 b8a6 f1e1 e7e5 d4f3 c8c7 f3e5 a8f8 h6g4 f6f5 c1d2 h5f6 b3h3 f5h5 h3g3 f6g4 e5g4 c7g3 h2g3 h5b5 e1e7');
INSERT INTO evaluations VALUES(30501,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(30501,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(30501,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(30501,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(30501,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(30501,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(30501,36,'cp 81','d2d4 c5d4 f3d4 a7a6 c1e3 e7e6 f2f3 d8c7 d1d2 h7h5 e1c1 g8f6 c1b1 f8e7 d4b3 b8c6 d2f2 b6b5 f2g3 c7g3 h2g3 d7d6 g3g4 e8c8 g4g5 f6d7 a2a4 b5b4 c3e2 d7c5 e2d4 c6d4 d1d4 c5b3 c2b3 d6d5 e4d5 e6d5 d4d1 h5h4 f3f4 e7d6 e3b6 d8d7');
INSERT INTO evaluations VALUES(30498,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(30498,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(30498,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(30498,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(30498,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(30498,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(30498,36,'cp 90','c5d4 f3d4 a7a6 c1g5 d8c7 c3d5 c7e5 g5e3 b7d5 e4d5 e5d5 f1e2 e7e5 d4f5 d5d1 a1d1 b8c6 e2f3 g7g6 f5d6 f8d6 d1d6 g8e7 e3g5 a8a7 g5f6 e8g8 e1g1 e7f5 d6d2 f8e8 f3c6 d7c6 g2g4 f5h6 f2f3 g8f8 f1e1 h6g8 f6e5 f7f6 e5g3 e8e1 g3e1 g8e7 g1f2 f8f7 d2d6 a7c7 h2h3 e7d5');
INSERT INTO evaluations VALUES(30924,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(30924,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(30924,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30924,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30924,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30924,36,'cp 0','d7b5 e2b5 d8d7 b5d7 e8d7 c2c3 e7e6 e1e2 g8f6 d2d3 d6d5 h1d1 b8c6 g2g3 f8d6 b1a3 h8e8 e4d5 e6d5 e2f1 c6e5 f3e1 e5c6');
INSERT INTO evaluations VALUES(35323,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(35323,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(35323,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(35323,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(35323,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(35323,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(35323,36,'cp 113','g8f6 f1d3 g7g6 h2h3 f8g7 e1g1 e8g8 f1e1 b7b6 d1e2 c8b7 e2e7 b8d7 c3e4 c6c5 e7d6 c7d6 e4d6 b7f3 g2f3 f6e8 e1e7 e8d6 e7d7 a8d8 d7d8 f8d8 d4c5 b6c5 a1b1 c5c4 d3e2 c4c3 d2c3 g7c3 b2c3 d8c8 b1b3 c8c7 g1g2 g8g7 e2d3 d6c4 b3b4 c4d6 c3c4 h7h6');
INSERT INTO evaluations VALUES(36226,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36226,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36226,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36226,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36226,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36226,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(36226,36,'cp 0','e1g1 c7c6 c4b3 c8g4 f1e1 b8d7 b1d2 f8e8 h2h3 g4h5 d2f1 e7f8 f1g3 h5g6 f3h4 d7c5 c1g5 c5e6 g5c1');
INSERT INTO evaluations VALUES(36225,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36225,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36225,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36225,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36225,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36225,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36225,36,'cp 0','c7c6 f1e1 b8d7 c4b3 h7h6 a2a4 f8e8 d3d4 e7f8 b3c2 d8c7 h2h3 b7b6 b1d2 a7a5 d2f1 c8a6 f1g3 a8d8 c1e3 d6d5 d4e5 f6e4 e3f4 g7g5 f3g5 h6g5 g3e4 g5f4 d1g4 f8g7 e4f6 g8f8 f6h7 f8g8');
INSERT INTO evaluations VALUES(37402,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(37402,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(37402,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(37402,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(37402,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(37402,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(37402,36,'cp 37','f1g2 b8c6 g1f3 d7d5 c4d5 f6d5 e1g1 e7e5 c1g5 d5c3 b2c3 d8d7 a1c1 e5d4 c3d4 c6d4 f3d4 g7d4 e2e3 d4e5 d1d7 c8d7 g2b7 a8b8 b7d5 f8c8 d5b3 e5d6 f1d1 a7a5 e3e4 a5a4 b3d5 a4a3 g5f6 d7e6 c1c3 e6d5 d1d5 b8b2');
INSERT INTO evaluations VALUES(38366,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(38366,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38366,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(38366,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(38366,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38366,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38366,36,'cp 14','e7e5 a1b1 a7a5 d2d3 g8e7 e2e4 b8c6 g1e2 c6d4 e2d4 e5d4 c3e2 a5a4 e1g1 e8g8 c1f4 c8g4 h2h3 g4e6 f4d2 e6d7 b2b4 a4b3 a2b3 b7b5 d1c2 c7c5 d2g5 h7h6 g5e7 d8e7 e2f4 f8b8 c2d2 e7d8 f4d5 d7c6');
INSERT INTO evaluations VALUES(38524,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(38524,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(38524,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(38524,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(38524,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38524,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(38524,36,'cp 20','g1f3 b8c6 e1g1 g8e7 d2d3 e8g8 e2e3 h7h6 a1b1 a7a5 a2a3 c8g4 h2h3 g4e6 b2b4 a5b4 a3b4 d8d7 b4b5 c6d8 g1h2 f7f5 f3d2 c7c6 c1b2 e6f7 f2f4 d8e6 c3a4 d7c7 b5b6 c7d7 d2f3 e5f4 e3f4 g7b2 a4b2 g6g5');
INSERT INTO evaluations VALUES(39836,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(39836,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(39836,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(39836,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(39836,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(39836,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(39836,36,'cp -18','c8a6 f3d2 d7d5 d1a4 c7c6 e2e4 d5c4 a4c2 e8g8 f1c4 a6c4 d2c4 c6c5 c1e3 f6g4 a1d1 c5d4 c3d4 b6b5 h2h3 g4e3 c4e3 d8a5 c2d2 a5a3 d4d5 b8a6 e1g1 a3b4 d2e2 b4e4 e2b5 a6c7 b5d7 c7d5 e3d5 e6d5');
INSERT INTO evaluations VALUES(41478,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41478,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(41478,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(41478,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41478,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41478,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41478,36,'cp 0','e2e3 c7c5 c2c3 b8c6 f1e2 d8b6 d1c1 e8g8 e1g1 c8f5 b1d2 a8c8 a2a4 f6d7 g2g4 f5e6 c1b1 c5d4 f3d4 a7a5 a1a3 g7e5 a3b3 b6c7 f4h6 e5g7 h6f4');
INSERT INTO evaluations VALUES(48101,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(48101,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(48101,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(48101,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(48101,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(48101,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(48101,36,'cp 62','c7c5 g1f3 b8c6 e1g1 c5d4 f3d4 d5c3 b2c3 c8d7 f1e1 c6d4 c3d4 a8c8 c4b3 d7c6 e1e3 h7h5 e3d3 f8e7 d4d5 c6d5 c1e3 d8c7 b3d5 e6d5 a1c1 c7d7 c1c8 d7c8 d3d5 e8g8 e3a7 c8c4 d1b3 c4c1 d5d1 c1c6 a7e3 f8d8 h2h3 d8d1 b3d1 c6e6 d1a4');
INSERT INTO evaluations VALUES(3075,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(3075,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(3075,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(3075,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3075,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(3075,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(3075,36,'cp 16','d4b3 d7d6 f2f3 d8c7 g2g4 c8e6 e1c1 f8d8 b3d4 c6d4 e3d4 c7a5 a2a3 a8b8 g4g5 f6h5 c3d5 a5d2 d1d2 e6d5 e4d5 b7b5 d4g7 g8g7 d2d4 f7f6 f3f4 f6g5 f4g5 d8f8 f1e2 h5f4 e2g4 f8f7 h1e1 a7a6 g4d7 b8f8 h2h4 h7h6 e1e4 f4h5 c1b1 h5g3');
INSERT INTO evaluations VALUES(12021,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(12021,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12021,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(12021,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(12021,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12021,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12021,36,'cp 17','c2c4 d5c7 c1d2 g7g6 f1e2 f8g7 d2c3 e8g8 e1g1 d7d5 e5d6 d8d6 d1d6 e7d6 c3g7 g8g7 f1d1 d6d5 c4d5 c7d5 b1d2 c8e6 d2e4 f8d8 e4c5 d5f4 e2f1 e6d5 b2b3 a8b8 c5a6 b8b6 a6c7 b6b7 c7d5 c6d5 d1d2');
INSERT INTO evaluations VALUES(12020,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12020,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(12020,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(12020,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(12020,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(12020,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(12020,36,'cp 11','d5c7 f1e2 g7g6 c1d2 f8g7 d2c3 e8g8 e1g1 d7d5 e5d6 d8d6 d1d6 e7d6 c3g7 g8g7 f1d1 d6d5 b1c3 f8d8 c4d5 c7d5 c3a4 c8f5 a1c1 h7h5 f2f3 d5f4 e2f1 f5d3 g1f2 a8b8 b2b3 h5h4 g2g3 d3f1 f2f1 f4e6 g3h4 d8d1 c1d1 b8h8');
INSERT INTO evaluations VALUES(14905,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(14905,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(14905,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(14905,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(14905,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(14905,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(14905,36,'cp 69','f8d6 f2f4 f6d5 g1e2 d5c3 e2c3 e8g8 e1g1 b7b6 c1d2 g8h8 d1c2 c8b7 a1d1 d8d7 c3e4 f7f5 e4d6 d7d6 c2b3 a7a5 c4e6 c6e7 e6c4 b7e4 c4d3 e4d3 b3d3 e7d5 e3e4 f5e4 d3e4 a8e8 e4f3 h8g8 f4f5 c7c6 f1f2 d6f6 h2h3 f6d4 d2a5 d4c5');
INSERT INTO evaluations VALUES(9272,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(9272,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(9272,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(9272,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(9272,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(9272,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(9272,36,'cp 75','f2f4 f6d5 g1e2 d5c3 e2c3 e8g8 e1g1 b7b6 c1d2 a7a6 d4d5 b6b5 c4b5 a6b5 d5c6 b5b4 c3e4 b4a3 b2a3 c8a6 f1e1 f8e8 d2c3 d6f8 d1f3 f7f5 e4g3 d8e7 c3b4 e7f6 g3h5 f6f7 b4f8 e8f8 e1d1 f8d8 a3a4 d8d5 d1d5 e6d5 a4a5 a6c4');
INSERT INTO evaluations VALUES(18968,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(18968,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(18968,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(18968,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18968,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(18968,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(18968,36,'cp 37','g1f3 g8f6 f3g5 d7d5 e4d5 f6d5 d2d4 c5d4 g5f7 e8f7 d1f3 d8f6 c4d5 c8e6 d5e6 f7e6 f3e2 e6d7 e1g1 d7c7 b1d2 a8d8 d2e4 f6e6 c1g5 f8e7 g5e7 e6e7 a1d1 c7b8 f1e1 h8e8 h2h3 a7a6 d1d3 d4c3 e4c3 h7h6 a2a3 d8d4 b2b4 e8d8 d3g3 e7d7 b4b5 a6b5');
INSERT INTO evaluations VALUES(22554,30,'cp 667',NULL);
INSERT INTO evaluations VALUES(22554,31,'cp 667',NULL);
INSERT INTO evaluations VALUES(22554,32,'cp 657',NULL);
INSERT INTO evaluations VALUES(22554,33,'cp 660',NULL);
INSERT INTO evaluations VALUES(22554,34,'cp 656',NULL);
INSERT INTO evaluations VALUES(22554,35,'cp 658',NULL);
INSERT INTO evaluations VALUES(22554,36,'cp 671','f3e5 d7d5 d1d4 g8f6 b1c3 f8d6 c4d5 f6d5 c3d5 e8g8 e5c4 c7c6 c1h6 f7f6 e1c1 g7h6 d4e3 c6d5 d1d5 c8e6 d5d6 e6c4 d6d8 a8d8 e3h3 f8f7 b2b3 c4b5 h1e1 d8e8 h3h6 f7g7 e4e5 b5d7 e1d1 d7e6 h6f6 g7g2 d1d6 e6f7 e5e6 g2g7 e6e7');
INSERT INTO evaluations VALUES(29782,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(29782,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(29782,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(29782,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(29782,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(29782,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(29782,36,'cp 68','f3e5 g4e2 d1e2 b8d7 d2d4 e7e6 c1g5 f8e7 h2h4 h7h6 g5d2 d7e5 d4e5 f6d5 e1c1 d8d7 c3d5 e6d5 h4h5 e8c8 f2f4 d5d4 h1h3 d7e6 c1b1 c7c5 h3g3 h8g8 e2e4 e6d5 d1e1 d5e4 e1e4 f7f6 g3g6 c5c4 f4f5 f6e5 e4e5');
INSERT INTO evaluations VALUES(29903,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(29903,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29903,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(29903,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(29903,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(29903,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(29903,36,'cp 27','f8e7 b1d2 e8g8 h2h3 g4h5 a2a3 c7c6 f1e1 b8d7 d2f1 d7c5 f1g3 h5g6 c1e3 c5e6 g3f5 d8d7 f5e7 d7e7 d1d2 h7h6 c4a2 a7a5 b2b4 a5b4 a3b4 f6h5 g2g3 g8h7 g1h2 f7f5 e4f5 g6f5 f3h4 e6g5 e3g5 e7g5 d2g5 h6g5 h4f5 f8f5 h2g2 f5f8');
INSERT INTO evaluations VALUES(28382,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(28382,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(28382,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(28382,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(28382,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(28382,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(28382,36,'cp 31','b1d2 b8d7 a2a3 c7c6 f1e1 e8g8 h2h3 g4h5 d2f1 d7c5 c4a2 d8d7 c2c3 f8e8 f1g3 h5g6 c1e3 e7f8 b2b4 c5e6 c3c4 f6h5 a3a4 a7a5 b4b5 e6c5 a1b1 h5g3 f2g3 g8h8 g3g4 f8e7 g2g3 h7h5 g1g2 h5g4 h3g4 d7g4 e1h1 g6h5');
INSERT INTO evaluations VALUES(31244,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(31244,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31244,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31244,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(31244,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31244,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(31244,36,'cp 30','g4f3 d1f3 e7e6 c4b5 b8d7 b5e2 g7g6 f3g3 d8c7 e1g1 f8g7 b1a3 a7a6 a3c4 d7e5 d2d3 d6d5 c4d2 g8e7 c2c3 a8c8 f2f4 e5d7 e4e5 e8g8 h3h4 f7f6 g3h3 c7b6 e5f6 d7f6 d2f3 f6h5 f3g5 e7f5 g2g4 c5c4 g1g2 h5g3 g4f5 g3e2 f5g6 h7g6 g5e6');
INSERT INTO evaluations VALUES(31248,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(31248,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(31248,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(31248,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(31248,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(31248,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31248,36,'cp 34','d1f3 e7e6 c4b5 b8d7 b5e2 g7g6 f3g3 d8c7 d2d3 f8g7 c2c3 g8e7 b1a3 a7a6 a3c4 d7e5 c1e3 f7f5 c4e5 g7e5 f2f4 e5g7 e1g1 e8g8 a1d1 g8h8 g3f2 b7b5 e2f3 a8c8 g1h1 c5c4 d3c4 f5e4 f3e4 c7c4 e4d3 c4a2 e3b6');
INSERT INTO evaluations VALUES(31247,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31247,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31247,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(31247,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(31247,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(31247,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(31247,36,'cp 34','e7e6 c4b5 b8d7 b5e2 d7e5 f3g3 g8f6 d2d3 f8e7 e1g1 e8g8 f2f4 e5c6 g1h1 b7b5 c1e3 b5b4 b1d2 a8b8 c2c3 b4c3 b2c3 d6d5 e4e5 f6d7 c3c4 d5c4 d2c4 c6d4 e3d4 c5d4 f4f5 e6f5 f1f5 g7g6 a1f1 d8c7');
INSERT INTO evaluations VALUES(32793,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(32793,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(32793,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(32793,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(32793,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(32793,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(32793,36,'cp 51','e7e6 h2h3 g4f3 e2f3 f8e7 e1g1 g8f6 a2a3 e8g8 c3e2 a5a6 c2c3 b8d7 f1e1 f8e8 e2f4 a8d8 d1e2 a6e2 e1e2 a7a5 a3a4 e7d6 f4d3 h7h6 g2g3 e6e5 d4e5 d7e5 d3e5 d6e5 c1e3');
INSERT INTO evaluations VALUES(32816,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(32816,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(32816,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(32816,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(32816,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(32816,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(32816,36,'cp 65','h2h3 g4f3 e2f3 b8d7 e1g1 g8f6 c1f4 f8e7 a2a3 e8g8 d1d3 a5b6 b2b4 a7a5 b4b5 f6d5 f4g3 d7f6 b5c6 b6c6 c3d5 e6d5 f1e1 f8e8 c2c4 c6c4 d3c4 d5c4 e1c1 a8a7 c1c4 b7b5 c4c1 b5b4 a3b4 e7b4 c1c7 a7c7 g3c7 e8c8 c7a5 b4a5 a1a5 c8c1 g1h2');
INSERT INTO evaluations VALUES(35167,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(35167,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(35167,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(35167,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(35167,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(35167,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(35167,36,'cp 178','g1f3 g5e7 d2d4 d7d6 e1g1 c8e6 b1a3 g8f6 d4e5 d6e5 f3e5 b8d7 e5d7 e7d7 d1d7 e8d7 c4e6 f7e6 f1d1 d7c8 f2f3 f8a3 b2a3 b7b6 c1b2 h8e8 d1d3 e6e5 a1d1 c8b7 a3a4 a7a5 g2g4 g7g5 h2h3 h7h6 g1g2 e8e7 d3d8 a8d8 d1d8 f6d7');
INSERT INTO evaluations VALUES(37623,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(37623,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(37623,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(37623,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(37623,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(37623,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(37623,36,'cp 75','d2d4 c5d4 f3d4 a7a6 c4b3 b7b5 a2a4 b5b4 c2c3 c8b7 f1e1 g8f6 e4e5 f6d5 b1d2 e8g8 d2e4 b4c3 b2c3 g7e5 e4c5 d8b6 e1e5 b6c5 c1h6 f8e8 c3c4 c5c7 e5d5 e6d5 c4d5 c7e5 d4f3 e5h5 h6f4 d7d6 a1c1 e8c8 c1c8 b7c8 f4d6 b8d7');
INSERT INTO evaluations VALUES(39024,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(39024,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(39024,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(39024,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(39024,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(39024,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(39024,36,'cp 44','d7d6 c2c4 b8d7 b1c3 e6e5 f4g3 g8e7 f1d3 e5d4 e3d4 d7f6 e1g1 e8g8 h2h3 c8f5 f1e1 f5d3 d1d3 d8d7 d4d5 e7f5 g3f4 a8e8 b2b4 h7h6 a2a4 e8e1 a1e1 f8e8 e1e8 d7e8 g2g3 h6h5 g1g2 a7a6 f4d2 f6d7 g3g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(38557,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(38557,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(38557,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(38557,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(38557,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(38557,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(38557,36,'cp 53','f1e2 b8d7 h2h3 g8e7 c2c4 e6e5 f4h2 e5d4 e3d4 e8g8 b1c3 b7b6 d1d2 c8b7 a1d1 e7f5 e1g1 d8f6 c3b5 f8c8 h2f4 h7h6 g2g4 a7a6 b5c3 f5h4 f3h4 f6h4 g1h2 c8e8 f4g3 h4d8 d4d5 d7c5 f1e1 a6a5 d2c2 a5a4 h2g2 g7c3');
INSERT INTO evaluations VALUES(39821,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(39821,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(39821,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(39821,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(39821,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(39821,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(39821,36,'cp 18','g1e2 c8a6 a2a3 b4c3 e2c3 d7d5 b2b3 e8g8 f1e2 c7c5 d4c5 d5c4 d1d8 f8d8 b3c4 b8d7 c5c6 d7c5 c3b5 a8c8 c1b2 c8c6 b5a7 c6c7 a7b5 c7c8 b2e5 c5d3 e2d3 d8d3 e1e2 d3b3 h1d1 a6b5 c4b5 b3b5 e5d4 f6d5 a1c1');
INSERT INTO evaluations VALUES(43996,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(43996,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(43996,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(43996,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43996,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(43996,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(43996,36,'cp -26','f2f4 b8c6 g1f3 c6d4 f3d4 c5d4 c3e2 d8a5 c1d2 a5b6 e1g1 d7c5 d2e1 c5d3 c2d3 f8e7 a1b1 e8g8 e1f2 c8d7 f2d4 b6a5 b1a1 f8c8 f1f3 e7c5 d4c5 c8c5 d3d4 c5c4 f3c3 c4c3 e2c3 a8c8 d1d2 h7h6 a2a3 d7b5 h2h3 b5a6 a1d1 a5b6 g1h1 c8c4');
INSERT INTO evaluations VALUES(49182,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(49182,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(49182,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(49182,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(49182,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(49182,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(49182,36,'cp 60','f2f4 c7c5 c2c3 b8c6 g1f3 e7e6 d3d4 f8e7 f1e2 c5d4 c3d4 d7b6 b1c3 e8g8 e1g1 c8d7 c1e3 d7e8 a1c1 f7f6 e5f6 e7f6 f3e5 b6c4 e5c4 d5c4 e2c4 e8f7 c3e2 d8d7 e3f2 a8c8 g1h1 h7h6 c4b3 f8d8 c1c3');
INSERT INTO evaluations VALUES(49318,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(49318,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(49318,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(49318,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(49318,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(49318,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(49318,36,'cp 37','c7c5 d2d4 c5d4 f3d4 d8a5 d1d2 a5c5 c4e2 b8c6 d4c6 b7c6 a2a3 d5e7 f2f4 e7f5 b1c3 a7a5 e2d3 d7d5 e5d6 c5d6 d3f5 e6f5 d2d6 f8d6 c1e3 c8e6 e1c1 e8c8 g2g3 d6c7 d1d8 h8d8 h1e1 g7g6 c3a4 h7h5 e3b6 c7b6 a4b6 c8c7 b6a4 d8d4 b2b3 c7d6 e1e3 e6d5');
INSERT INTO evaluations VALUES(44135,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44135,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44135,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(44135,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(44135,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44135,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(44135,36,'cp 39','d2d4 c5d4 f3d4 b8c6 e1g1 d7d6 e5d6 f8d6 c4d5 e6d5 f1e1 d6e7 c1g5 f7f6 g5e3 e8g8 a2a3 f6f5 d1d2 e7f6 d4c6 b7c6 e3d4 c8d7 b1c3 f6d4 d2d4 f5f4 f2f3 d8b6 c3e2 a8e8 d4b6 a7b6 b2b4 f8f6 e2d4 g8f7 e1e8 f7e8 g1f2 c6c5 b4c5 b6c5');
INSERT INTO evaluations VALUES(44136,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44136,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(44136,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44136,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(44136,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(44136,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(44136,36,'cp 33','c5d4 f3d4 d8a5 d1d2 a5c5 c4e2 b8c6 d4c6 b7c6 a2a3 d5e7 f2f4 a7a5 b1c3 e7f5 e2d3 d7d5 e5d6 c5d6 d3f5 e6f5 d2d6 f8d6 c1e3 c8a6 e1c1 e8c8 g2g3 d6c7 e3c5 h8e8 d1d8 c7d8 h1d1 g7g5 d1d6 g5f4 g3f4 e8e6 c3a4 e6d6 c5d6 d8f6');
INSERT INTO evaluations VALUES(50550,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(50550,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(50550,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(50550,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(50550,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(50550,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(50550,36,'cp -44','g1f3 b7b5 c4d3 g8e7 h2h4 d7d6 e1g1 b8d7 b2b4 e7c6 a2a4 c6b4 a4b5 f8e7 d3e2 a6a5 c1f4 d7f6 e4e5 f6d5 c3d5 b4d5 f4g3 e8g8 d1d4 c8b7 e2d3 d6e5 f3e5 f7f5 f1d1 e7f6 d3e2 d8e7 e2f3 f8c8 d4a4');
INSERT INTO evaluations VALUES(52461,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(52461,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(52461,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(52461,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(52461,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(52461,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(52461,36,'cp -97','d7d5 b1c3 e7e6 d2d3 b8c6 e2e4 d5d4 c3e2 e6e5 e2g3 h7h6 h2h4 g7g6 a2a3 f8e7 h4h5 c8e6 c1d2 e7h4 h5g6 f7g6 b2b4 d8e7 b4c5 e8c8 g1f3 e6g4 a1b1 h4g3 f2g3 g8f6 h1h6 h8h6 d2h6 e7c5');
INSERT INTO evaluations VALUES(53299,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(53299,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(53299,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(53299,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(53299,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(53299,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(53299,36,'cp -93','b1c3 d5d4 c3e4 b8c6 h2h3 h7h5 g4g5 e7e5 g1f3 g8e7 f3h4 e7g6 h4g6 f7g6 c2c3 c8e6 d1c2 d8d7 h3h4 a8c8 e4g3 c6e7 e1g1 e7f5 c2e4 c8c7 d2d3 f8d6 g3f5 g6f5 e4f3 c7c6 c1d2 g7g6 c3d4 c5d4 a1c1 c6b6 b2b3 b6b5 e2e3 d4e3 d2e3 e8g8 e3a7 e6d5');
INSERT INTO evaluations VALUES(54058,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(54058,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54058,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54058,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(54058,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54058,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54058,36,'cp 0','f8c5 c4d5 e6d5 g1f3 g8f6 b1c3 b8c6 f1e2 a7a6 e1g1 e8g8 b2b3 c5a7 c1b2 f8e8 a1c1 c8f5 h2h3 d8d7 c3a4 d7d6 b2f6 g7f6 a4c3 a8d8 e2d3 f5g6 d3g6 h7g6 c3e2 e8e4 e2c3 e4e6 c3e2 e6e4');
INSERT INTO evaluations VALUES(38117,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38117,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38117,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38117,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38117,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38117,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38117,36,'cp 0','c4d5 e6d5 g1f3 g8f6 a2a3 e8g8 b1c3 b8c6 b2b4 d5d4 c3b5 c5b6 b5d4 c6d4 f3d4 c8g4 f1e2 g4e2 d1e2 b6d4 e3d4 a7a5 b4a5 f8e8 c1e3 d8a5 e2d2 a5b5 d2e2');
INSERT INTO evaluations VALUES(38111,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38111,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38111,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38111,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38111,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38111,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38111,36,'cp 0','e6d5 a2a3 g8f6 g1f3 b8c6 b1c3 e8g8 b2b4 d5d4 c3b5 c5b6 b5d4 c6d4 f3d4 c8g4 f1e2 g4e2 d1e2 b6d4 e3d4 a7a5 b4a5 f8e8 c1e3 d8a5 e2d2 a5a6 d2e2');
INSERT INTO evaluations VALUES(38195,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38195,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38195,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38195,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38195,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38195,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38195,36,'cp 0','a2a3 g8f6 g1f3 b8c6 b1c3 e8g8 b2b4 d5d4 c3b5 c5b6 b5d4 c6d4 f3d4 c8g4 f1e2 g4e2 d1e2 b6d4 e3d4 a7a5 b4a5 f8e8 c1e3 d8a5 e2d2 a5a6 d2e2');
INSERT INTO evaluations VALUES(6478,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6478,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6478,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6478,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6478,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6478,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6478,36,'cp 0','d7d6 c3a4 g8e7 f1c4 e8g8 c2c3 a7a5 a4c5 d6c5 d2d3 d8d6 c1e3 b7b6 e1g1 h7h6 a2a3 c8e6 d1a4 e6d7 a4c2 a8d8 f1d1 d6g6 g1h1 d7e6 c2a4 e6d7');
INSERT INTO evaluations VALUES(6010,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6010,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6010,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6010,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6010,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6010,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6010,36,'cp 0','c3a4 g8e7 f1c4 e8g8 c2c3 a7a5 a4c5 d6c5 d2d3 d8d6 e1g1 h7h6 c1e3 b7b6 a2a3 c8e6 d1a4 e6d7 a4c2 d7e6');
INSERT INTO evaluations VALUES(7366,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(7366,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7366,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(7366,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(7366,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(7366,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(7366,36,'cp 2','d2d3 d7d6 a4c6 b7c6 d3d4 e5d4 f3d4 c8d7 e1g1 e8g8 f1e1 f8e8 d1f3 h7h6 h2h3 e7f8 c1d2 a6a5 b2b3 a5a4 a1d1 a4b3 a2b3 f6h7 f3g3 d8f6 d4e2 h7g5 g3d3 f6g6 e2f4 g6f6 d2e3 f6e5 f2f3 g5e6 f4e6 e5e6 f3f4 e6g6 g1h2');
INSERT INTO evaluations VALUES(7367,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7367,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7367,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7367,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7367,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7367,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7367,36,'cp 0','d7d6 a4c6 b7c6 d3d4 e5d4 f3d4 c8d7 e1g1 e8g8 f1e1 f8e8 d1f3 h7h6 c1d2 a6a5 b2b3 e7f8 a2a4 d8b8 d4f5 d7f5 f3f5 b8b6 a1d1 e8e6 d2e3 b6b4 e3d2');
INSERT INTO evaluations VALUES(8552,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(8552,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(8552,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(8552,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(8552,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(8552,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(8552,36,'cp 21','d3d4 e5d4 f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 e1g1 e8g8 f2f4 f8e8 f1e1 a7a5 c1e3 a5a4 a1b1 f6d7 d1f3 e7h4 e1e2 h4f6 e3f2 f6c3 f3c3 c6c5 e4e5 d8e7 b1d1 d6e5 f4e5 e7e6 g1h1 f7f6 f2c5 d7e5 c5d4 e6a2 d4e5 e8e5 e2e5 f6e5 c3e5');
INSERT INTO evaluations VALUES(10042,30,'cp -302',NULL);
INSERT INTO evaluations VALUES(10042,31,'cp -301',NULL);
INSERT INTO evaluations VALUES(10042,32,'cp -336',NULL);
INSERT INTO evaluations VALUES(10042,33,'cp -322',NULL);
INSERT INTO evaluations VALUES(10042,34,'cp -325',NULL);
INSERT INTO evaluations VALUES(10042,35,'cp -309',NULL);
INSERT INTO evaluations VALUES(10042,36,'cp -327','c6e5 d2d4 e5c4 d4c5 d7d5 b2b3 d5d4 c3e2 c4e5 d1d4 d8d4 e2d4 c8d7 f2f3 e8c8 c1b2 h7h5 d4e2 d8e8 h2h4 f6h7 e1f2 h7f8 e2f4 f7f6 g2g3 f8g6 a1d1 d7c6 h1e1 g6f4 g3f4 e5d7 b3b4 b7b6 c5b6 a7b6 f4f5 d7e5 b2e5 e8e5');
INSERT INTO evaluations VALUES(10762,30,'cp -310',NULL);
INSERT INTO evaluations VALUES(10762,31,'cp -313',NULL);
INSERT INTO evaluations VALUES(10762,32,'cp -320',NULL);
INSERT INTO evaluations VALUES(10762,33,'cp -300',NULL);
INSERT INTO evaluations VALUES(10762,34,'cp -320',NULL);
INSERT INTO evaluations VALUES(10762,35,'cp -310',NULL);
INSERT INTO evaluations VALUES(10762,36,'cp -310','d2d4 e5c4 d4c5 d7d5 b2b3 d5d4 c3e2 c4e5 d1d4 d8d4 e2d4 c8d7 f2f3 e8c8 c1b2 h8e8 a1d1 f6h5 e1f2 h5f4 h2h4 h7h5 c5c6 b7c6 g2g3 f4e6 d4e6 e8e6 d1d4 a7a6 h1d1 f7f6 d4a4 c6c5 a4a5 c5c4 d1d5 d7b5 d5d8 c8d8 b2c3 b5d7 a5c5 g7g6');
INSERT INTO evaluations VALUES(10764,30,'cp -319',NULL);
INSERT INTO evaluations VALUES(10764,31,'cp -309',NULL);
INSERT INTO evaluations VALUES(10764,32,'cp -315',NULL);
INSERT INTO evaluations VALUES(10764,33,'cp -304',NULL);
INSERT INTO evaluations VALUES(10764,34,'cp -301',NULL);
INSERT INTO evaluations VALUES(10764,35,'cp -306',NULL);
INSERT INTO evaluations VALUES(10764,36,'cp -312','e5c4 d4c5 d7d5 b2b3 d5d4 c3e2 c4e5 d1d4 d8d4 e2d4 c8d7 f2f3 e8c8 c1b2 h7h5 h2h4 f6h7 e1f2 f7f6 a2a4 e5c6 h1d1 h7f8 d4c6 d7c6 d1d8 c8d8 b3b4 d8c8 b4b5 c6d7 b2d4 f8g6 c2c4 h8d8 c5c6 b7c6 d4a7 g6e5 b5b6 e5c4 b6c7 c8c7 a4a5 c7b7');
INSERT INTO evaluations VALUES(10777,30,'cp -287',NULL);
INSERT INTO evaluations VALUES(10777,31,'cp -316',NULL);
INSERT INTO evaluations VALUES(10777,32,'cp -305',NULL);
INSERT INTO evaluations VALUES(10777,33,'cp -297',NULL);
INSERT INTO evaluations VALUES(10777,34,'cp -293',NULL);
INSERT INTO evaluations VALUES(10777,35,'cp -289',NULL);
INSERT INTO evaluations VALUES(10777,36,'cp -291','d4c5 d7d5 b2b3 d5d4 c3e2 c4e5 d1d4 d8d4 e2d4 c8d7 f2f3 e8c8 c1b2 e5c6 e1f2 c6d4 b2d4 f6e8 a1d1 f7f5 e4f5 d7f5 c2c3 e8f6 h1e1 f5g6 e1e7 h8e8 e7e8 g6e8 d1e1 e8g6 c5c6 b7c6 d4a7 d8d5 a7d4 c6c5 d4e3 d5d3 c3c4');
INSERT INTO evaluations VALUES(10760,30,'cp -297',NULL);
INSERT INTO evaluations VALUES(10760,31,'cp -323',NULL);
INSERT INTO evaluations VALUES(10760,32,'cp -312',NULL);
INSERT INTO evaluations VALUES(10760,33,'cp -310',NULL);
INSERT INTO evaluations VALUES(10760,34,'cp -321',NULL);
INSERT INTO evaluations VALUES(10760,35,'cp -306',NULL);
INSERT INTO evaluations VALUES(10760,36,'cp -326','d7d5 b2b3 d5d4 c3e2 c4e5 d1d4 d8d4 e2d4 c8d7 f2f3 e8c8 c1b2 h8e8 e1f2 h7h5 a2a4 h5h4 b3b4 h4h3 g2h3 e8h8 a1g1 h8h3 g1g3 h3g3 h2g3 d7a4 b2c3 e5c6 d4c6 a4c6 c3f6 g7f6 h1h7 c6e8 f2e3 d8d1 h7h6 c8d7 h6f6 d7e7 f6h6');
INSERT INTO evaluations VALUES(14921,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(14921,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(14921,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(14921,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(14921,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(14921,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(14921,36,'cp 71','c6a5 f1c4 a5c4 d1a4 f6d7 a4c4 h7h6 c4e2 f8d6 c1e3 a7a6 a1d1 b7b5 e1g1 c8b7 e4e5 b5b4 e5d6 b4c3 d6c7 d8c7 d1c1 e8g8 c1c3 c7a5 f1c1 f8c8 f3d2 c8c3 c1c3 d7f6 c3c5 a5b6 b2b4 f6d7 d2c4 b6d8 c5a5 d8c7 e3d2 c7c6 f2f3');
INSERT INTO evaluations VALUES(17797,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(17797,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(17797,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(17797,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(17797,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(17797,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(17797,36,'cp 92','c1e3 g8f6 h2h3 e5d4 f3d4 f8e7 d4c6 b7c6 d1e2 e8g8 e1c1 a8b8 g2g4 f6d7 f2f4 e7f6 e3a7 b8b7 a7d4 f6d4 d1d4 d7c5 e2e3 a6a5 g4g5 b7b4 h3h4 b4d4 e3d4 f7f6 f1c4 g8h8 d4d2 d8e7 h1e1 c8e6 c4b3 e6g8 b3g8 h8g8 c1b1');
INSERT INTO evaluations VALUES(17815,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(17815,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(17815,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(17815,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(17815,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(17815,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(17815,36,'cp -25','d8d3 d1e2 d3e2 e1e2 f7f6 d2d3 h7h5 c1e3 a6a5 a2a4 h5h4 h2h3 b7b6 b1d2 g7g5 b2b3 c6c5 h1c1 h8h7 d2c4 c8a6 c1h1 g8e7 g2g4 e7g6 f3d2 e8c8 c4b2 c8b8 d2c4 f8e7 h1d1 a6b7 a1b1 d8h8 d1d2 g6f4 e3f4 e5f4');
INSERT INTO evaluations VALUES(19454,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19454,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(19454,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(19454,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(19454,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(19454,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(19454,36,'cp -22','g8f6 f1g2 d7d5 e4d5 f6d5 g1f3 g7g6 e1g1 f8g7 f1e1 e8g8 b1d2 b7b6 d2c4 c8b7 a2a4 e7e6 c2c3 d8c7 d1e2 h7h6 a4a5 a8d8 f3e5 b6b5 e5c6 b7c6 c4e5 c6a8 c1d2 b5b4 a5a6 f8e8 c3b4 c5b4 d3d4 c7b6');
INSERT INTO evaluations VALUES(21847,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(21847,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(21847,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(21847,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(21847,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(21847,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(21847,36,'cp -44','d2d3 g7g6 h5g5 f7f6 g5h4 c8d7 b5c6 d7c6 b1c3 f8g7 g1e2 d8e7 a2a4 a7a6 e1g1 f6f5 h4e7 g8e7 c1g5 h7h6 g5e7 e8e7 b2b4 e7d8 f2f3 h6h5 b4b5 c6e8 b5b6 c7c6 e4f5 g6f5 f3f4 e8g6 a4a5 e5f4 e2f4 g7c3');
INSERT INTO evaluations VALUES(22848,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(22848,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(22848,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(22848,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(22848,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(22848,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(22848,36,'cp -61','d7d5 d2d3 c8e6 a2a3 d8d7 c2c4 g8f6 c1g5 h7h5 b1d2 f6h7 g1f3 f7f6 c4d5 e6d5 g5e3 g7g5 b3b4 g5g4 f3h4 d5g2 h4g2 e8c8 d1a4 h7g5 b4b5 c6d4 e3g5 f6g5 a4a7 d7b5 a1b1 b5c6 a7a8 c8d7 a8b7 c6b7 b1b7 f8a3 g2e3 d8b8 b7b8 h8b8 e1g1 d4e2 g1g2 b8b2');
INSERT INTO evaluations VALUES(27892,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(27892,31,'cp -152',NULL);
INSERT INTO evaluations VALUES(27892,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(27892,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(27892,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(27892,35,'cp -163',NULL);
INSERT INTO evaluations VALUES(27892,36,'cp -156','g8f6 d3e2 e8g8 e1g1 h7h6 f1e1 b8c6 c3d2 d7d6 h2h3 a8c8 c2c3 c6e7 f3h2 e4g6 e2a6 c8b8 a6f1 a7a5 a2a3 f6e4 d2f4 e7d5 f4c1 b6b5 h2f3 a5a4 c3c4 b5c4 f1c4 d5b6 c4b5 d8f6 b5c6 d6d5 c6a4 b6a4');
INSERT INTO evaluations VALUES(28077,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(28077,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(28077,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(28077,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(28077,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(28077,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(28077,36,'cp 87','f1b5 b8c6 b5c6 b7c6 b1d2 f7f6 e1g1 f6e5 d4e5 g8h6 h2h3 g4f3 d2f3 e8g8 c1e3 d8d7 f1e1 h6f7 e3c5 f7d8 f3d4 d8e6 d4e6 d7e6 d1d2 a7a5 a1c1 a5a4 c5d4 f8b8 c1c5 b8b5 c5b5 c6b5 d2b4 a8b8 a2a3 g7f8 e1c1 b8c8 c1c8 e6c8 b4b5 c8c1 b5f1');
INSERT INTO evaluations VALUES(28479,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(28479,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(28479,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(28479,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(28479,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(28479,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(28479,36,'cp -29','f6e4 g1e2 e6c4 d1c2 e4d6 a2a3 b4a5 c1f4 e8g8 e1c1 b8d7 e2g3 a5c3 f4d6 c7d6 f1c4 a8c8 c2d3 d6d5 c4a2 d8a5 c1c2 c3d4 c2b1 d4f6 d3d5 a5b6 d5d2 f8d8 g3e4 d7c5 e4f6 b6f6 d2c2 d8e8 d1e1 b7b6 e1e8 c8e8 f2f3 g7g6 h1d1 g8g7 a2c4 h7h5');
INSERT INTO evaluations VALUES(28467,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(28467,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(28467,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(28467,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(28467,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(28467,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(28467,36,'cp -13','g1e2 e6c4 d1c2 e4d6 a2a3 b4a5 c1f4 e8g8 e1c1 b8d7 e2g3 c4f1 h1f1 a5c3 c2c3 d7b6 c3f3 c7c6 f1e1 b6d5 f4d6 d8d6 e1e5 a8d8 g3f5 d6c7 h2h4 g7g6 f5g3 f7f6 e5e4 f8e8 d1e1 e8e4 e1e4 g8f7 c1b1 c7d7 h4h5 f6f5 e4e5');
INSERT INTO evaluations VALUES(28842,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(28842,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(28842,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(28842,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(28842,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(28842,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(28842,36,'cp -30','b8d7 e1g1 f8g7 d1e2 e8g8 f1d1 d8c7 a2a4 a7a6 d4f3 f8c8 h2h3 d7b6 c1g5 e7e6 f3d2 f6e8 a1a3 b6c4 d2c4 c7c4 e2d3 g7c3 b2c3 c4c5 g5c1 a6a5 d3e3 c5c6 e3d3 c8c7 c1f4 a8b8 f4d6 c6d6 d3d6 e8d6 d1d6 b8c8');
INSERT INTO evaluations VALUES(34425,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(34425,31,'cp -140',NULL);
INSERT INTO evaluations VALUES(34425,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(34425,33,'cp -147',NULL);
INSERT INTO evaluations VALUES(34425,34,'cp -138',NULL);
INSERT INTO evaluations VALUES(34425,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(34425,36,'cp -127','b8c6 g1f3 c5d4 e3d4 c6d4 d1d4 b6d4 f1b5 e8e7 f3d4 f7f6 f2f4 a7a6 b5e2 f6e5 f4e5 g7g6 c2c4 f8g7 d4f3 g8h6 e1g1 h8d8 c4d5 e6d5 b1c3 h6g4 f1e1 c8e6 f3d4 g7h6 e2g4 e6g4 c3e2 h6e3 g1h1 g4e2 d4e2 e7e6 e2c3 e3f4 a1d1');
INSERT INTO evaluations VALUES(1460,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(1460,31,'cp -159',NULL);
INSERT INTO evaluations VALUES(1460,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(1460,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(1460,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(1460,35,'cp -141',NULL);
INSERT INTO evaluations VALUES(1460,36,'cp -157','g1f3 c5d4 e3d4 c6d4 d1d4 b6d4 f1b5 c8d7 b5d7 e8d7 f3d4 g7g5 e1g1 g8e7 b1d2 e7g6 c2c4 f8c5 d4c2 d5c4 d2c4 b7b5 c4d6 c5d6 e5d6 h8c8 c2d4 a7a6 a1d1 c8c5 g2g3 a8c8 d4f3 g5g4 b3b4 c5d5 d1d5 e6d5 f3g5 g6e5 f2f4 g4f3 g5f3 f7f6 f3e5 f6e5');
INSERT INTO evaluations VALUES(35026,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(35026,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(35026,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(35026,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(35026,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(35026,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(35026,36,'cp 62','f8b4 c1d2 c7c6 a2a3 b4a5 d3d4 d7d6 c3e2 a5b6 d4e5 d6e5 e2g3 c8g4 h2h3 g4f3 d1f3 g7g6 h3h4 h6h5 f3e2 b8d7 f2f3 f6e6 e2c4 e6e7 e1c1 g8f6 g3e2 e8g8 g2g4 h5g4 h4h5 f6h5 f3g4 h5f6 e2c3 b6d4 d2g5 d7c5 c4e2 c5e6');
INSERT INTO evaluations VALUES(35247,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35247,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35247,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35247,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(35247,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35247,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35247,36,'cp 0','h5e2 f8g7 a2a4 g8f6 d2d3 e8g8 g1f3 d7d6 e1g1 f8e8 f3g5 e8f8');
INSERT INTO evaluations VALUES(37438,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37438,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37438,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37438,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37438,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37438,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37438,36,'cp 17','e2e3 c7c5 f1e2 d8b6 d1c1 d7d5 c2c3 b8c6 e1g1 c8f5 b1d2 a8c8 d2b3 f6e4 f1d1 f8d8 f3g5 c5d4 e3d4 e7e5 g5e4 d5e4 d4e5 d8d1 c1d1 c8d8 d1e1 c6e5 f4e3 b6c7 b3d4 f5d7 a1d1 h7h6 d4b5 d7b5 d1d8 c7d8 e2b5 e5d3 e1d2');
INSERT INTO evaluations VALUES(39219,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(39219,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(39219,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(39219,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39219,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(39219,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(39219,36,'cp 10','g8f6 g2g3 e8g8 f1g2 d7d6 g1f3 c7c6 e1g1 a7a5 d3d4 e5d4 f3d4 a5a4 e2e3 a4a3 b2b3 f8e8 c3a4 b8a6 a4c5 d6c5 d4c2 c8f5 d1d8 e8d8 c2a3 a6b4 f2f3 d8d3 e3e4 f5e6 f1e1 f6h5 f3f4 f7f6 e4e5 f6e5 f4e5 a8d8 c1g5');
INSERT INTO evaluations VALUES(39522,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(39522,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(39522,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(39522,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(39522,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(39522,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(39522,36,'cp 62','g1f3 b8c6 f1c4 f6e5 d4e5 c6e5 f3e5 g7e5 d1d5 d7d6 d5f7 e8d7 c4e6 d7c6 e6b3 c6d7 b1d2 d8e8 d2c4 e5c3 b2c3 e8f7 c4e5 d6e5 b3f7 d7c6 c1a3 e7e6 f7e8 c8d7 e8d7 c6d7 e1c1 d7e8 d1e1 g8f6 e1e5 e8f7 h1e1 h8e8 c3c4 f6d7 e5e3 a8d8 e3f3 f7g8 f3d3 d7b6 e1e4 d8d3 c2d3 b6d7 c1c2 g8f7 a3b2 b7b6 e4e1 e6e5 b2c3 a7a6 f2f3 e8e6 a2a4');
INSERT INTO evaluations VALUES(39521,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(39521,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(39521,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(39521,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(39521,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(39521,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(39521,36,'cp 41','b8c6 f1c4 f6e5 d4e5 c6e5 f3e5 g7e5 d1d5 d7d6 d5f7 e8d7 c4e6 d7c6 e6b3 c6d7 b1a3 d8e8 a3c4 e5c3 b2c3 e8f7 c4e5 d6e5 b3f7 g8f6 c1a3 f6e4 e1c1 e4d6 f7d5 c7c6 d5b3 a7a5 h1e1 a5a4 b3c4 d7c7 c4d3 d6f7 a3e7 h8e8 e7b4 c8f5 c3c4 e5e4 d3e2 a8d8 c4c5 c7c8 b4c3 d8d1 e2d1 c8d7 c3f6');
INSERT INTO evaluations VALUES(40291,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(40291,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(40291,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(40291,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(40291,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(40291,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(40291,36,'cp 67','c4b3 b7b5 e4e5 d6e5 f3e5 e8g8 e1g1 c8b7 f1e1 b8d7 d1f3 a7a5 a2a3 d7e5 d4e5 f6d5 b3d5 c6d5 c3b5 d8b6 f3d3 b7a6 a3a4 a8b8 d3d5 f8d8 d5f3 a6b5 a4b5 b6b5 e5e6 f7e6 c2c3 e6e5 h2h4 d8d3 a1a5 d3f3 a5b5 b8b5 g2f3 g8f7 g1f1');
INSERT INTO evaluations VALUES(44020,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(44020,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(44020,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(44020,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(44020,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(44020,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(44020,36,'cp 121','e7e6 f3g5 f8e7 h2h4 h7h6 g5e6 f7e6 d1h5 e8f8 h1h3 d8e8 h3f3 d7f6 h5g4 b8d7 c1d2 f8g8 g4e6 e8f7 e6f7 g8f7 e1c1 d7f8 e5f6 e7f6 c3e2 c8g4 f3b3 b7b6 b3c3 a8c8 c3a3 c8a8 f2f3 g4d7 g2g3 f8e6 f1h3 h8e8 h3f5 g7g6 f5e6 e8e6 e2f4 e6e8 h4h5 g6g5 f4d3 g5g4');
INSERT INTO evaluations VALUES(44156,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(44156,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(44156,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(44156,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(44156,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(44156,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(44156,36,'cp -58','d7d5 e5d6 d8d6 g1f3 a7a6 a2a4 b8c6 e1g1 b7b6 d2d3 d6c7 c1g5 c8b7 f1e1 f8e7 d1c1 c6d4 f3d4 c5d4 c3e2 a8d8 c1f4 c7c6 f4g3 f6h5 g3g4 e7g5 g4g5 h7h6 g5g4 e8g8 f2f3 h5f6 g4h4 c6d6 c4b3 g8h8 e2f4');
INSERT INTO evaluations VALUES(46283,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(46283,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(46283,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(46283,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(46283,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(46283,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(46283,36,'cp -54','e5d6 d8d6 g1f3 a7a6 d2d4 b7b5 c4d3 c8b7 d4c5 d6c5 d1e2 b8d7 e1g1 f8e7 c3e4 c5d5 f1e1 f6e4 d3e4 d5e4 e2e4 b7e4 e1e4 d7c5 e4e2 f7f6 c1e3 e8f7 e3c5 e7c5 a1d1 a8d8 e2d2 d8d2 f3d2 h8d8 g1f1 f6f5 g2g3 c5e7 f1e2 e6e5 d2f3 d8d1 e2d1');
INSERT INTO evaluations VALUES(48549,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(48549,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(48549,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(48549,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(48549,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(48549,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(48549,36,'cp -61','c7c6 b5e2 d5d4 c3b1 f8e7 e1g1 e8g8 e2c4 c6c5 d2d3 b8c6 a2a4 h7h6 c1f4 e7d6 f4g3 d8c7 b1d2 c8f5 f3h4 f5d7 f1e1 d6g3 h2g3 f8e8 h4f3 e8e1 d1e1 a8e8 e1f1 c6b4 c4b3 b7b6 c2c3 d4c3 b2c3 b4d5 a1c1 b6b5 b3d5 f6d5 a4b5 d7b5');
INSERT INTO evaluations VALUES(45722,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(45722,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(45722,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(45722,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(45722,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(45722,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(45722,36,'cp -65','b5e2 d5d4 c3b1 f8e7 e1g1 e8g8 e2c4 c6c5 d2d3 b8c6 a2a4 h7h6 c1f4 e7d6 f4g3 d8c7 f1e1 c8g4 b1d2 f8e8 g3d6 c7d6 h2h3 g4d7 e1e8 a8e8 c2c3 d4c3 b2c3 c6a5 c4b5 e8d8 d3d4 c5d4 f3d4 a5c6 d2c4 d6c7 c4e3 c6d4 c3d4 d7e6 b5f1');
INSERT INTO evaluations VALUES(48899,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(48899,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(48899,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(48899,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(48899,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(48899,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(48899,36,'cp 78','f1c4 c8e6 c4e6 f7e6 d1e2 d8d4 e5g4 b8c6 g4f6 d4f6 b2b3 f8d6 c1b2 e8g8 e1c1 f6f2 e2c4 f2f5 d1e1 a8e8 c1b1 f5c5 c4c5 d6c5 c3e4 c5e7 e1e2 f8f7 d2d4 e8f8 h1e1 f7f1 c2c3 f1e1 e2e1 h7h6 b1c2 e7d6 g2g3 g8f7 e4d6 c7d6 b2a3 f8d8 e1e2 d6d5');
INSERT INTO evaluations VALUES(48893,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(48893,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(48893,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(48893,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(48893,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(48893,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(48893,36,'cp 56','c8e6 c4e6 f7e6 d1e2 d8d4 e5g4 b8c6 g4f6 d4f6 b2b3 f8d6 c1b2 e8g8 e1c1 f6f2 e2c4 f2f5 h1f1 f5e5 f1e1 a8e8 c3e4 e5d5 c4e2 e6e5 c1b1 h7h6 e2h5 d5f7 h5h3 f7e6 h3g3 a7a6 e1f1 f8f1 d1f1 e8f8 f1f8 d6f8 h2h3 c6d4 d2d3 e6f7 e4d2 f8d6 a2a3');
INSERT INTO evaluations VALUES(30028,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(30028,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(30028,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(30028,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(30028,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(30028,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(30028,36,'cp 55','c4e6 f7e6 d1e2 d8d4 f2f4 b8d7 e5d7 e8d7 b2b3 f8d6 g2g3 e6e5 c1b2 e5f4 e2b5 d7c8 e1c1 d4c5 g3f4 c5b5 c3b5 d6f4 d1f1 f6h5 b2g7 h8d8 g7c3 c8d7 c3d4 d7c6 d4e3 f4e3 d2e3 c6b5 f1f5 b5b6 f5h5 d8d7 h1f1 a8g8 f1f4 g8g2 f4e4 d7d2 e4b4 b6a6 b4c4 c7c6');
INSERT INTO evaluations VALUES(56445,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(56445,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(56445,32,'cp -141',NULL);
INSERT INTO evaluations VALUES(56445,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(56445,34,'cp -140',NULL);
INSERT INTO evaluations VALUES(56445,35,'cp -146',NULL);
INSERT INTO evaluations VALUES(56445,36,'cp -144','e5e4 f3g1 g8f6 d2d4 c7c5 g1e2 b8c6 b2b3 h7h5 c1b2 f8d6 d4c5 d6c5 b1a3 c6e5 e2d4 e8g8 a3c2 f8e8 d1d2 c5d6 h2h3 c8d7 b2a3 d6a3 c2a3 d8e7 a3c2 e8c8 f1e2 a7a6 e1g1 e7c5 c2b4 c5c3 d2c3 c8c3 a2a3');
INSERT INTO evaluations VALUES(56256,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(56256,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(56256,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(56256,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(56256,34,'cp -139',NULL);
INSERT INTO evaluations VALUES(56256,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(56256,36,'cp -133','f3g1 g8f6 d2d3 b8d7 b1d2 d7e5 d1c2 e5d3 f1d3 e4d3 c2d3 g7g6 g1f3 f8g7 b2b4 e8g8 c1b2 a7a5 e1g1 c8f5 d3e2 f8e8 a2a3 f6e4 f3d4 f5g4 f2f3 e4d2 e2d2 g4d7 f1e1 h7h5 a1c1 h5h4 h2h3 d8g5 g1h1 b7b5 b4a5 c7c5 d4b3 a8c8 e3e4 g5d2 b3d2');
INSERT INTO evaluations VALUES(56789,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(56789,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(56789,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(56789,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(56789,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(56789,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(56789,36,'cp 50','g5d2 c7c5 e2e3 c5d4 e3d4 b8c6 f1d3 f8d6 c2c4 g8e7 e1g1 e8g8 f1e1 c8d7 b1c3 a8c8 b2b3 d7e8 e1e6 e8h5 c3e2 h5f3 g2f3 d8d7 e6e3 e7f5 d3f5 d7f5 c4c5 d6f4 e2f4 f5f4 d2c3 c8e8 d1d3 e8e3 d3e3 f4g5 g1h1 g5h5 a1e1');
INSERT INTO evaluations VALUES(2461,30,'cp -184',NULL);
INSERT INTO evaluations VALUES(2461,31,'cp -199',NULL);
INSERT INTO evaluations VALUES(2461,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(2461,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(2461,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(2461,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(2461,36,'cp -82','e1d1 c2a1 c3d5 d8h4 d5b6 e6e7 b6c8 a8c8 d2d3 e7d8 g2g3 h4f6 f3b7 f8d6 c1e3 c7c6 h1e1 h8e8 e3a7 c8c7 a7b6 d8d7 b6c7 d6c7 b7a7 g7g5 a7e3 e8b8 b2b4 c7b6 e3e5 f6e5 e1e5 b8f8 e5e2 b6f2 d3d4 f8f3 d4d5 f2g1 d5c6 d7c6 a3a4 h7h5 e2e6 c6d7 e6h6 f3c3 h6h7 d7d6 h7h6 d6e5');
INSERT INTO evaluations VALUES(2462,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(2462,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(2462,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(2462,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(2462,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(2462,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(2462,36,'cp -85','c2a1 c3d5 d8h4 d5b6 e6e7 b6c8 a8c8 d2d3 e7d8 g2g3 h4f6 f3b7 f8d6 c1e3 c7c6 h1e1 h8e8 e3a7 c8c7 a7b6 d8d7 b6c7 d6c7 b7a7 g7g5 a7e3 e8b8 b2b4 c7b6 e3e5 f6e5 e1e5 b8f8 e5e2 f8f2 e2f2 b6f2 d1d2 c6c5 b4c5 d7c6 d2e2 f2g1 h2h4 g5h4 g3h4 a1c2 a3a4 g1c5 e2f3 c6d6');
INSERT INTO evaluations VALUES(2460,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(2460,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(2460,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(2460,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(2460,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(2460,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(2460,36,'cp -62','c3d5 d8h4 d5b6 e6e7 b6c8 a8c8 d2d3 e7d8 g2g3 h4f6 f3b7 f8d6 c1e3 d8d7 e3a7 c8b8 a7b8 h8b8 c4b5 d7e7 b7b8 f6f3 d1e1 f3h1 e1e2 h1h2 b8e8 e7f6 e8d8 d6e7 d8c7 h2h5 e2f1 h5h1 f1e2 h1d5 b5c4 d5d6 c7c8 a1c2 c8h3 c2d4 e2f1 d4f3 h3h7 d6d4 h7g8 f3d2 f1g2 d2c4 g8c4 d4b2 c4c6 f6g5 c6e4 b2d4 f2f4 g5f6 e4d4 e5d4 a3a4 f6e6 a4a5 e7c5 a5a6 c5a7 g2f3');
INSERT INTO evaluations VALUES(2452,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(2452,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(2452,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(2452,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(2452,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(2452,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(2452,36,'cp -73','d8h4 d5b6 e6e7 b6c8 a8c8 d2d3 e7d8 g2g3 h4f6 f3b7 f8d6 c1e3 c7c6 h1e1 h8e8 e3a7 c8c7 a7b6 d8d7 b6c7 d6c7 b7a7 g7g5 a7e3 e8b8 b2b4 c7b6 e3e5 f6e5 e1e5 b8f8 e5e2 b6f2 d3d4 d7d6 e2e6 d6c7 e6e7 c7d8 e7e2 f8f6 d4d5 f2g1 e2a2 c6d5 c4d5 d8c7 d5c4 g1d4 a2d2 d4c3 d2d5 f6f2 d5c5 c7b6 c5b5 b6a7 b5a5 a7b8');
INSERT INTO evaluations VALUES(11122,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(11122,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11122,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(11122,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(11122,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(11122,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(11122,36,'cp 30','d8c7 d4c6 e7e6 d1e2 d7c6 a2a4 f8d6 h2h3 b7b6 e1g1 c8b7 c4d3 f6d7 c1e3 d6h2 g1h1 h2f4 e3f4 c7f4 a1d1 e8g8 e2d2 f4d2 d1d2 d7c5 f1d1 f8d8 h1g1 g8f8 e4e5 c5d3 d2d3 d8d3 d1d3 c6c5');
INSERT INTO evaluations VALUES(15593,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15593,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(15593,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(15593,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(15593,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(15593,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(15593,36,'cp 11','c2c4 g7g6 c1e3 f8g7 b1c3 f6d7 d3d4 e5d4 f3d4 e8g8 d1d2 d7c5 f1e2 g7d4 e3d4 c6d4 d2d4 f7f5 e4f5 c8f5 a1d1 a7a5 e1g1 d8f6 f2f3 a8e8 d4f6 f8f6 g1f2 f5e4 d1d4 e4c6 b2b3 h7h6 h3h4 g8f7 f1e1 f6f5 d4d2 f5e5 e1d1 e5e3 c3d5');
INSERT INTO evaluations VALUES(16235,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(16235,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(16235,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(16235,33,'cp -120',NULL);
INSERT INTO evaluations VALUES(16235,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(16235,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(16235,36,'cp -111','b7c6 d1e2 d8e7 e1g1 e4d6 f1e1 e7e2 c4e2 f8e7 b1c3 e8d8 c1d2 a7a5 c3a4 e7f6 a1d1 h8e8 e2d3 e8e1 d1e1 g7g6 b2b3 d6b5 d3e4 d7d5 e4d3 b5d6 g1f1 c8f5 d3f5 d6f5 c2c4 d5c4 b3c4 f5d6 e1d1 d8c8 d1e1 f6d4 c4c5 d6c4');
INSERT INTO evaluations VALUES(11954,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(11954,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(11954,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(11954,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(11954,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(11954,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(11954,36,'cp -113','d1e2 d8e7 e1g1 e4d6 b1c3 e7e2 c4e2 f8e7 f1e1 e8d8 a1b1 h8e8 b2b3 a7a5 e2d3 h7h6 c3a4 e7f6 c1d2 d6b7 d3e4 d7d5 e4f3 e8e1 b1e1 c8d7 e1d1 d8c8 f3e2 b7d6 d2f4 d6e4 f4e3 c8b7 c2c4 d7e6 e2f3');
INSERT INTO evaluations VALUES(11953,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(11953,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(11953,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(11953,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(11953,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(11953,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(11953,36,'cp -95','d8e7 e1g1 e4d6 b1c3 e7e2 c4e2 f8e7 c1d2 a7a5 f1e1 e8d8 b2b3 d6f5 a1d1 d7d6 e2d3 c8e6 e1e4 d8d7 e4a4 d6d5 c3e2 e7b4 c2c3 b4d6 c3c4 g7g6 d2c3 h8e8 c3a5 h7h5 g1f1 d5c4 b3c4 c6c5 d3e4 a8a6 e2f4');
INSERT INTO evaluations VALUES(17636,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(17636,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(17636,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(17636,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(17636,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(17636,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(17636,36,'cp 110','b1c3 g8e7 h2h4 c8d7 a2a3 d8b6 c3a4 b6c7 f1d3 e7f5 d3f5 e6f5 c1f4 c6d8 a1c1 c7a5 a4c3 d8e6 f4d2 f8e7 d1c2 f5f4 c2b3 d7c6 c3e2 a5b5 c1c3 e8g8 e2f4 e6f4 d2f4 a8c8 e1d2 b5a5 h1c1 h7h6 g2g3 c6b5 b3d5 c8c3 c1c3');
INSERT INTO evaluations VALUES(17633,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(17633,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(17633,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(17633,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(17633,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(17633,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(17633,36,'cp 117','g8h6 c1h6 g7h6 f1d3 f8g7 e1g1 f7f6 e5f6 d8f6 f3e5 e8g8 f2f4 f6e7 c3e2 c8d7 d1d2 g8h8 a1e1 h6h5 d2e3 a8c8 a2a3 c6e5 d4e5 g7h6 g2g3 e7g7 e1c1 c8c1 e3c1 f8c8 c1e3 g7g4 g1g2 h6f8 h2h3 f8c5 h3g4');
INSERT INTO evaluations VALUES(27615,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(27615,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(27615,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(27615,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(27615,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(27615,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(27615,36,'cp 11','d5h5 d4d5 e8c8 f1e2 g4f3 e2f3 h5e5 c1e3 e7e6 d1d3 g8f6 e1c1 f8b4 d5c6 d8d3 c6b7 c8b8 d1d3 b4c3 d3c3 f6d5 c3c5 e5f6 f3d5 e6d5 c5d5 h8e8 h1d1 b8b7 g2g3 e8e5 d5d4 e5e6 a2a4 e6d6 d4d6 c7d6 d1d3 f6e6 b2b3 f7f6 h2h4 e6e5 c1b1 a7a5 e3d4 e5f5 b1b2 h7h5');
INSERT INTO evaluations VALUES(28107,30,'cp 312',NULL);
INSERT INTO evaluations VALUES(28107,31,'cp 306',NULL);
INSERT INTO evaluations VALUES(28107,32,'cp 299',NULL);
INSERT INTO evaluations VALUES(28107,33,'cp 307',NULL);
INSERT INTO evaluations VALUES(28107,34,'cp 302',NULL);
INSERT INTO evaluations VALUES(28107,35,'cp 297',NULL);
INSERT INTO evaluations VALUES(28107,36,'cp 301','c4f7 e8f7 f3g5 f7e8 d1g4 g7d4 g4e6 b8c6 c2c3 d4f6 g5f7 d8c8 e6c8 a8c8 f7h8 f6h8 a2a4 c6a5 c1g5 h7h6 g5e3 a5c4 e1e2 g8f6 b1d2 c4e3 e2e3 e8f7 g2g3 a7a5 h2h4 h8g7 f2f4 b7b6 a1d1 f6d7 d2f3 d7c5 h4h5 c5a4 h5g6 f7g8');
INSERT INTO evaluations VALUES(29988,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(29988,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(29988,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(29988,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(29988,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(29988,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(29988,36,'cp 68','h2h3 g4f5 f1d3 d6d5 e1g1 f8e7 c2c4 d5c4 d3c4 e8g8 b1c3 e4c3 b2c3 b8c6 f1e1 h7h6 c1f4 e7d6 f3e5 c6e5 f4e5 d6e5 e1e5 d8f6 d1e2 f5d7 c4d3 g7g6 e2e4 d7c6 e4e3 a8e8 f2f4 e8e5 f4e5 f6g5 e3g5 h6g5 a1e1 g8g7 d3c4');
INSERT INTO evaluations VALUES(31028,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(31028,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(31028,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(31028,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(31028,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(31028,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(31028,36,'cp -46','e7e6 h2h4 h7h5 f4f5 e6f5 g4g5 c6c5 c2c3 g8e7 g1e2 f5f4 f1h3 g6f5 g5g6 f5g6 e2f4 b8c6 e5e6 c5d4 e1g1 g6f5 d1f3 d8d6 e6f7 e8f7 f4d5 g7g6 d5e7 f8e7 h3f5 g6f5 f3f5 f7e8 c1f4 h8g8 g1h1 d6g6 f5g6 g8g6 b1d2 a8d8 a1e1 d4c3 b2c3 e8d7 d2e4 e7h4 e1d1 d7e7 d1d8 c6d8');
INSERT INTO evaluations VALUES(32828,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(32828,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(32828,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(32828,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(32828,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(32828,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(32828,36,'cp 24','b8c6 h2h3 g4f3 e2f3 e8c8 c3e4 d6g6 d2d3 e7e5 e1g1 h7h5 e4g5 f8e7 f3c6 e7g5 c6f3 g8e7 d1e1 g5c1 a1c1 g6d6 e1e3 c8b8 f1e1 f7f6 c1d1 e7c6 c2c3 g7g5 f3c6 d6c6 d3d4 e5d4 d1d4 d8e8 d4e4 e8e4 e3e4 c6e4 e1e4 h8d8 h3h4 g5h4 e4h4');
INSERT INTO evaluations VALUES(27639,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(27639,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(27639,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(27639,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27639,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(27639,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(27639,36,'cp 33','h2h3 g4f3 e2f3 e8c8 c3b5 d6e5 d1e2 e5e2 e1e2 g8f6 f3c6 b7c6 b5c3 f6d5 h1e1 e7e5 d2d3 d5b4 e2d1 f8d6 c3e4 f7f5 e4d2 c6c5 a2a3 b4d5 d1e2 d5f4 e2f1 f4e6 a1b1 h8e8 d2c4 e6d4 c1g5 d6e7 g5e7 e8e7 e1e5 e7e5 c4e5 d4c2 b1c1 d8e8 c1c2 e8e5');
INSERT INTO evaluations VALUES(36551,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36551,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36551,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(36551,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36551,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36551,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36551,36,'cp 1','e7e5 b1c3 f8e8 d3c2 b8d7 a2a4 a7a5 b2b3 e5d4 e3d4 d7b8 h2h3 b8a6 c1g5 a6b4 a1c1 h7h6 g5e3 d6d5 f3e5 c7c5 f2f4 d5c4 b3c4 c5d4 e3d4 b4c6 d4e3 d8d1 c1d1 c6e5 f4e5 f6d7 c3d5 d7e5 d5b6');
INSERT INTO evaluations VALUES(39667,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(39667,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(39667,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(39667,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(39667,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(39667,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(39667,36,'cp 48','g5f6 e7f6 f1c4 c8e6 d1e2 b8d7 e1g1 a8c8 c3d5 d7b6 d5f6 d8f6 c4e6 f6e6 b2b3 b6d7 c2c4 e8g8 h2h3 f8d8 f1d1 d7c5 f3h4 g7g6 h4f3 h7h6 d1d2 g8g7 f3e1 b7b5 c4b5 a6b5 e1c2 d6d5 d2d5 d8d5');
INSERT INTO evaluations VALUES(40653,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40653,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40653,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40653,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40653,36,'cp 13','e1g1 e8g8 b1c3 b8c6 h2h3 c8e6 f1e1 h7h6 a2a3 a7a6 c1f4 f8e8 d1d2 a8c8 a1d1 b7b5 f3e5 c6e5 f4e5 f6d7 e5f4 d8f6 f4e3 c5e3 e1e3 d7c5 e3g3 d5d4 c3e4 c5e4 d3e4 e6h3 d2d4 f6d4 d1d4 h3e6 c2c3 g7g5 e4d3 e6b3 d4d6 e8e1 g1h2');
INSERT INTO evaluations VALUES(40652,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40652,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40652,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40652,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40652,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40652,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40652,36,'cp 0','e8g8 b1c3 b8c6 h2h3 c8e6 f1e1 h7h6 c1f4 f8e8 c3b5 d8b6 f4c7 c5f2 g1h2 b6c5 c7d6 c5b6');
INSERT INTO evaluations VALUES(36165,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36165,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36165,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36165,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36165,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36165,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36165,36,'cp 0','b1c3 b8c6 a2a3 c8g4 h2h3 g4h5 c1g5 f8e8 a1b1 c6e5 d3e2 h5f3 e2f3 d5d4 c3e4 e5f3 d1f3 e8e4 g5f6 d8d5 f3g3 g7g6 b1e1 a8e8 g3f3 d5c6 e1e4 c6e4 f3e4 e8e4 f1d1 e4e2 b2b4 c5b6 f6d4 b6d4 d1d4 e2c2 d4d7 c2a2 d7b7 a2a1 g1h2 a1a3 b4b5 a3b3 b7b8 g8g7 b8b7 g7g8');
INSERT INTO evaluations VALUES(42223,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(42223,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(42223,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(42223,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(42223,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(42223,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(42223,36,'cp 56','d1f3 c8e6 g5f6 d8f6 f3f6 g7f6 a2a3 b4a5 g1f3 b8c6 f1b5 h8g8 e1f1 a5c3 b2c3 e6f5 f3e1 h7h5 f2f3 a7a6 b5e2 e8d7 f1f2 c6a5 e1d3 f5d3 e2d3 b7b5 a3a4 a5c4 h1b1 d7c6 b1e1 g8e8 e1e8 a8e8 a4b5 a6b5 a1a6 c4b6');
INSERT INTO evaluations VALUES(50172,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(50172,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(50172,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(50172,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(50172,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(50172,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(50172,36,'cp 32','c7c5 d2d4 c5d4 e3d4 f8g7 d4d5 e8g8 f1e2 e7e6 g1f3 e6d5 c4d5 d7d6 c1f4 f8e8 e1g1 a7a6 d1b3 f6h5 f4g5 h5f6 b3b4 b8d7 b4f4 d7c5 a1d1 b7b5 b2b4 c5a4 c3a4 b5a4 f1e1 c8b7 e2c4 e8e1 d1e1 a8c8 h2h3 f6d5');
INSERT INTO evaluations VALUES(51062,30,'cp 239',NULL);
INSERT INTO evaluations VALUES(51062,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(51062,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(51062,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(51062,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(51062,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(51062,36,'cp 248','e1g1 b8c6 d2d4 d7d6 a2a4 c8g4 c2c3 g7g6 h2h3 g4f3 d1f3 f8g7 f3d1 g8h6 a4a5 d8e7 d1a4 e7d7 a4b3 c6d8 f2f4 h6f7 c1e3 d7e7 b3a4 e7d7 a4d7 e8d7 b1d2 g7h6 c4d5 c7c6 d2c4 a8b8 c4b6 d7c7 d5b3 h8e8 a1e1 e5f4 e3f4 h6f4');
INSERT INTO evaluations VALUES(53776,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(53776,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53776,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(53776,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(53776,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(53776,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(53776,36,'cp -3','g1f3 f8d6 e1g1 g8f6 b1c3 a7a5 e2e4 d5e4 c3e4 f6e4 d3e4 e8g8 d1d3 d8e7 c1e3 b8a6 a1d1 d6c5 d1d2 f7f6 f1d1 c8g4 e3c5 a6c5 d3e3 b7b6 g2f1 f8d8 a2a3 d8d2 d1d2 g8f7 f3e1 a8d8 d2d8 e7d8 b2b4 a5b4');
INSERT INTO evaluations VALUES(53775,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(53775,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(53775,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(53775,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(53775,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(53775,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(53775,36,'cp -6','f8d6 e1g1 g8f6 b1c3 e8g8 e2e4 d5d4 c3e2 c6c5 f3d2 b8c6 a2a4 d6c7 f2f4 e5f4 e2f4 f6g4 d2c4 g4e5 c4e5 c6e5 c2c4 c8g4 d1c2 f7f6 f4d5 a8b8 b2b4 c5b4 d5b4 b7b6 c1f4 d8d7 b4d5 g4e6 c2b3 c7d6 b3b5 h7h6 f1f2 d6c5 f4e5 f6e5 b5d7 e6d7');
INSERT INTO evaluations VALUES(53807,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(53807,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(53807,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(53807,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(53807,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(53807,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(53807,36,'cp 73','d2d4 b8d7 a2a4 g8f6 g1e2 f8e7 e1g1 e8g8 h2h3 f8e8 c1e3 e5d4 e3d4 b7b6 f2f4 d7c5 e2g3 d6d5 e4d5 c6d5 c4b5 c8d7 d4f6 e7f6 c3d5 d7b5 a4b5 c5e4 d5f6 d8f6 g3e4 e8e4 c2c3 e4f4 d1d7 f6e6 d7e6 f4f1 g1f1 f7e6 f1e2 h7h6');
INSERT INTO evaluations VALUES(55690,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(55690,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(55690,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(55690,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(55690,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(55690,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(55690,36,'cp -86','g7g5 d2d4 g5g4 c1f4 g4f3 d1f3 b8c6 f1b5 d8h4 f4g3 h4g4 f3f2 f8g7 e1g1 c8d7 a1d1 e8c8 b5e2 g4g6 d4d5 c6b8 f2a7 d8e8 e2b5 g7c3 b5d7 c8d7 b2c3 g8f6 e4e5 f6e4 a7b7 e4g3 b7b5 d7c8 h2g3 b8d7 b5a6 c8d8 e5d6 g6d6 a6d6 c7d6 f1f7 e8e7 f7e7 d8e7');
INSERT INTO evaluations VALUES(57315,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(57315,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(57315,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(57315,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(57315,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(57315,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(57315,36,'cp 70','f1d3 f8c5 b1d2 a7a5 e4e5 d7d5 e5d6 c5d6 d2e4 g8f6 d3b5 c8d7 e4d6 c7d6 d1e2 d8e7 b5d3 b8c6 e2e7 e8e7 g1e2 c6e5 d3c2 h8e8 c1g5 d7c6 f2f3 h7h6 g5f6 e7f6 e1c1 f6e7 e2f4 e5c4 f4d5 c6d5 d1d5 a5a4 h1e1 e7f8 e1e8 a8e8 c2a4 e8e2 a4b3 c4b2 d5d6 e2g2');
INSERT INTO evaluations VALUES(57314,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(57314,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(57314,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(57314,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(57314,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(57314,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(57314,36,'cp 87','f8c5 b1d2 b8c6 b2b4 c5b6 g1f3 d7d6 a2a4 a7a6 e1g1 g8f6 h2h3 c8e6 d2c4 e6c4 d3c4 d8e7 a4a5 b6a7 c4d5 f6d5 e4d5 c6e5 f3e5 d6e5 d1a4 e7d7 a4b3 e8g8 c3c4 a7d4 c1b2 d4b2 b3b2 f7f5 a1e1 e5e4 f2f3 e4f3 f1f3 a8e8 e1d1');
INSERT INTO evaluations VALUES(1071,30,'cp 817',NULL);
INSERT INTO evaluations VALUES(1071,31,'cp 820',NULL);
INSERT INTO evaluations VALUES(1071,32,'cp 825',NULL);
INSERT INTO evaluations VALUES(1071,33,'cp 811',NULL);
INSERT INTO evaluations VALUES(1071,34,'cp 809',NULL);
INSERT INTO evaluations VALUES(1071,35,'cp 834',NULL);
INSERT INTO evaluations VALUES(1071,36,'cp 825','f7h8 g7g5 c4f7 e8d8 f7b3 e7h7 h8f7 d8e8 b1c3 d7d6 c3b5 h7e4 e1f1 e8d7 h2h4 g5g4 d2d3 e4h7 d3d4 e5e4 f1g1 a7a6 f7g5 h7e7 b5c3 d7e8 b3f7 e8d8 c3d5 f6d5 f7d5 c8d7 c2c3 d8c8 g5e4 c8b8 c1f4 b8a7 g2g3 a8e8 g1h2 d7e6 d5c6 b7c6');
INSERT INTO evaluations VALUES(627,30,'cp 832',NULL);
INSERT INTO evaluations VALUES(627,31,'cp 825',NULL);
INSERT INTO evaluations VALUES(627,32,'cp 823',NULL);
INSERT INTO evaluations VALUES(627,33,'cp 818',NULL);
INSERT INTO evaluations VALUES(627,34,'cp 821',NULL);
INSERT INTO evaluations VALUES(627,35,'cp 821',NULL);
INSERT INTO evaluations VALUES(627,36,'cp 825','g7g5 c4f7 e8d8 f7b3 d7d6 h8g6 e7g7 g6f8 g7f8 d2d3 f8g7 c1e3 c8d7 e1g1 a7a5 a2a4 d8c8 b1c3 b7b6 c3d5 c6e7 d5f6 g7f6 d1h5 c8b7 h5f7 f6f7 b3f7 a8f8 f7c4 e7g6 f2f3 b7b8 c2c3 b8b7 g1f2 g6f4 b2b4 a5b4 c3b4');
INSERT INTO evaluations VALUES(2681,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(2681,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2681,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2681,36,'cp 0','b7b5 a4b3 d7d6 d2d4 c5b6 c1g5 a8b8 h2h3 h7h6 g5h4 e5d4 c3d4 g7g5 h4g3 f6h5 f3e5 c6e5 d1h5 b6d4 h5h6 e5d3 e4e5 d6e5 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(2793,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2793,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2793,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2793,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2793,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2793,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2793,36,'cp 0','a4b3 d7d6 d2d4 c5b6 c1g5 h7h6 g5h4 e5d4 b3d5 d4c3 b1c3 c8b7 e4e5 d6e5 f3e5 c6e5 d5b7 d8d1 a1d1 a8a7 e1e5 a7b7 h4f6 g7f6 e5e7 b7b8 c3d5 g8g7 e7d7 f8e8 d5e7 c7c5 e7f5 g7h7 g2g4 h7g6 d1d6 e8e4 g1f1 e4g4 f5e7 g6g7 e7f5');
INSERT INTO evaluations VALUES(3588,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(3588,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(3588,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(3588,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(3588,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(3588,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(3588,36,'cp -28','g5f3 c6a5 c4b3 a5b3 a2b3 a7a5 h2h3 f8e8 c1e3 c5e3 f2e3 c8e6 d1e1 c7c6 e1g3 b7b5 f3h4 f6h5 g3f2 e8f8 h4f3 h5f6 f2e1 g8h7 e1g3 e6d7 g3f2 h7g8 c3e2 d7e6 f3h4 d6d5 e4d5 c6d5 h4f3 f6d7 d3d4 f7f6');
INSERT INTO evaluations VALUES(6874,30,'cp 570',NULL);
INSERT INTO evaluations VALUES(6874,31,'cp 557',NULL);
INSERT INTO evaluations VALUES(6874,32,'cp 564',NULL);
INSERT INTO evaluations VALUES(6874,33,'cp 567',NULL);
INSERT INTO evaluations VALUES(6874,34,'cp 573',NULL);
INSERT INTO evaluations VALUES(6874,35,'cp 586',NULL);
INSERT INTO evaluations VALUES(6874,36,'cp 581','c3d4 e5d4 g1f3 g8f6 f1d3 e8g8 e1g1 h7h6 c1f4 d7d6 b1d2 f6h5 f4g3 c7c6 g1h1 a7a5 a2a4 c8d7 g3h4 d8e8 f1e1 e8c8 d3f1 f8e8 a1c1 h5f4 h4g3 f4h5 f1d3 h5g3 h2g3 c8d8 d5c6 b7c6');
INSERT INTO evaluations VALUES(6872,30,'cp 568',NULL);
INSERT INTO evaluations VALUES(6872,31,'cp 571',NULL);
INSERT INTO evaluations VALUES(6872,32,'cp 559',NULL);
INSERT INTO evaluations VALUES(6872,33,'cp 576',NULL);
INSERT INTO evaluations VALUES(6872,34,'cp 575',NULL);
INSERT INTO evaluations VALUES(6872,35,'cp 548',NULL);
INSERT INTO evaluations VALUES(6872,36,'cp 572','e5d4 g1f3 g8f6 f1d3 e8g8 e1g1 h7h6 c1f4 d7d6 f4g3 c7c6 d5c6 b7c6 b1d2 f6h5 e4e5 h5g3 h2g3 d6e5 f3e5 c5d6 f1e1 c8e6 a1c1 e6d5 e5c6 d5c6 c1c6 d6b4 a2a3 b4d2 d1d2 f8e8 e1e8 d8e8 c6c7');
INSERT INTO evaluations VALUES(10771,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(10771,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(10771,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(10771,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(10771,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(10771,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(10771,36,'cp 154','f3e5 e8g8 e5f3 c7c6 c1e3 d4f3 d1f3 c5e3 f2e3 d7d6 e1g1 c8e6 c4e6 f7e6 f3h3 e6e5 h3e6 g8h8 h2h3 h7h6 f1f3 d8b6 e6b3 b6b3 a2b3 a7a5 c3e2 h8g8 e2g3 g7g6 g3f1 g8f7 f1d2 b7b5 g1f2 f7e7 g2g4 f8f7 f2e2 f6d7 f3f7 e7f7 d2f3');
INSERT INTO evaluations VALUES(11408,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11408,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(11408,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11408,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(11408,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11408,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(11408,36,'cp 40','b5c6 b7c6 d3d4 e5d4 f3d4 c8d7 d1f3 f8e7 e1g1 e8g8 f1e1 f8e8 h2h3 h7h6 c1f4 c6c5 d4e2 d7c6 e2g3 a6a5 a1d1 e7f8 a2a4 f6d7 b2b3 g7g6 f3d3 f8g7 g3e2 d7f8 c3d5 f8e6 f4g3 a8b8 f2f3 c6d5 d3d5 e6d4');
INSERT INTO evaluations VALUES(11359,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(11359,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(11359,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11359,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(11359,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(11359,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(11359,36,'cp 33','b7c6 d3d4 e5d4 f3d4 c8d7 d1f3 f8e7 e1g1 e8g8 f1e1 f8e8 h2h3 h7h6 c1d2 e7f8 a1d1 d6d5 e4d5 c6d5 c3d5 f6d5 f3d5 d7h3 d5d8 e8d8 d2h6 h3g4 f2f3 g4f3 d4f3 g7h6 g1f2 g8g7 d1d8 a8d8 e1e4 d8d5 f2e2 d5a5 a2a3 f8d6 e2d3 a5h5 d3c4 h5h1 f3d4 g7g6 c2c3 h1a1');
INSERT INTO evaluations VALUES(11668,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(11668,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(11668,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(11668,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(11668,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11668,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(11668,36,'cp 28','d3d4 e5d4 f3d4 c8d7 e1g1 f8e7 d1f3 e8g8 f1e1 f8e8 h2h3 h7h6 c1d2 e7f8 a1d1 d6d5 e4e5 f6h7 f3g3 c6c5 d4f3 d5d4 c3e2 d7c6 c2c3 c6f3 g3f3 e8e5 c3d4 e5d5 e2f4 d5d4 d2c3 d4d1 e1d1 d8e8 f4d5 h7g5 f3f5 e8e4 f5e4 g5e4 d5c7');
INSERT INTO evaluations VALUES(17391,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(17391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17391,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(17391,36,'cp 18','d3d4 c5d4 f3d4 d8c7 d4c6 b7c6 d1g4 d7d5 f1d3 g8f6 g4g3 c7g3 h2g3 e6e5 f2f3 c8e6 c1d2 f8d6 g3g4 f6d7 e4d5 c6d5 e1c1 h7h6 d1e1 d7c5 d3f5 a8b8 f3f4 d5d4 c3e4 c5e4 e1e4 e6f5 g4f5 f7f6 g2g4');
INSERT INTO evaluations VALUES(17509,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(17509,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(17509,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(17509,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(17509,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(17509,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(17509,36,'cp -23','g8f6 d1e2 e7e6 c4b3 d7d5 d2d3 c5c4 e4d5 c4b3 d5c6 b3c2 e2c2 f6d5 c1d2 f8e7 e1g1 e8g8 f1d1 d8c7 b1c3 c7c6 c2b3 f8d8 f3e5 c6c5 e5c4 c5d4 d2e3 d5e3 f2e3 d4h4 d3d4 a8b8 d4d5 e6d5 c3d5 h4g5 d5e7 g5e7 d1d8 e7d8');
INSERT INTO evaluations VALUES(17910,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(17910,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(17910,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(17910,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(17910,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(17910,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(17910,36,'cp 43','d7d6 d2d4 c8g4 c3e2 g8f6 d4d5 c6e7 e2g3 c7c6 d5c6 e7c6 h2h3 g4e6 f1d3 h7h6 e1g1 f8e7 b2b3 e8g8 c2c4 d8c7 c1e3 f8e8 g3e2 b7b5 e2c3 b5c4 d3c4 e6c4 b3c4 a8b8 d1d3 c7b7 f1d1 e7f8 c3d5 f6d5');
INSERT INTO evaluations VALUES(26657,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(26657,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(26657,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26657,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26657,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(26657,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(26657,36,'cp 24','c2c3 e7e6 d2d4 f7f6 b1d2 f6e5 f3e5 c6e5 d4e5 c7c6 d2b3 d8c7 c1f4 c7f7 b3d4 g8e7 f4e3 f5g6 f1e2 e7f5 d4f5 f7f5 d1b3 b7b5 a2a4 b5b4 c3b4 f5e5 b3c3 e5c3 b2c3 e8d7 h2h4 h7h5 e1g1 f8d6 g2g3 h8b8 b4b5 c6c5 e2f3 g6d3 f1e1 a7a6 b5a6 a8a6 f3h5 d3f5 a4a5');
INSERT INTO evaluations VALUES(26928,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(26928,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(26928,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(26928,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(26928,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(26928,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(26928,36,'cp 82','c6d4 d1d3 c7c5 d5c6 d4c6 h2h3 g4e6 f3g5 g8f6 c1e3 d8a5 e3d2 c6b4 d3b5 a5b5 f1b5 b4c6 g5e6 f7e6 b5c4 c6d4 e1c1 a8c8 c4d3 f8e7 c1b1 e8g8 d2e3 d4c6 g2g3 a7a6 d3e2 b7b5 h3h4 h7h5 a2a4 b5a4 c3a4 f6e4');
INSERT INTO evaluations VALUES(27081,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(27081,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(27081,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(27081,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(27081,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(27081,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(27081,36,'cp 92','d1d3 c7c5 d5c6 d4c6 c3d5 f8e7 h2h3 g4h5 c2c3 g8f6 f1e2 e8g8 c1e3 f6d5 d3d5 h5g6 a1d1 h7h6 f3d2 g8h8 d2f1 f7f5 e4f5 g6f5 e2d3 f5d3 d1d3 d8a5 d5a5 c6a5 b2b3 b7b6 f1g3 h8h7 e1g1 a8d8');
INSERT INTO evaluations VALUES(44420,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(44420,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(44420,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44420,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(44420,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(44420,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(44420,36,'cp 40','d4d5 e7e6 b1c3 f8g7 g1f3 d7d6 h2h3 e6d5 c4d5 e8g8 f1d3 b8a6 a2a4 a6b4 e3e4 f6d7 d3b1 d8e7 e1g1 d7e5 c1f4 f7f6 f3e5 f6e5 f4e3 c8d7 d1d2 f8f7 a4a5 a8f8 f2f3 e7h4 d2e1 g7f6 g1h1 h4e1 f1e1');
INSERT INTO evaluations VALUES(46279,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46279,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46279,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46279,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46279,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46279,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46279,36,'cp 0','b8c6 e2e4 d5e4 d3e4 d8d1 f1d1 f6e4 b1a3 f8e7 f3e5 c6e5 g2e4 e8g8 c1f4 e7f6 f4e3 f6e7');
INSERT INTO evaluations VALUES(50117,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(50117,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50117,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(50117,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(50117,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(50117,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(50117,36,'cp 21','e7e6 b1d2 f8e7 g5f6 e7f6 e2e4 e8g8 f1d3 d7d5 e4d5 e6d5 e1g1 b8c6 c2c3 d8d6 d3c2 b7b6 f1e1 c8d7 d2f1 a8e8 f1e3 c6e7 f3e5 c7c6 e5d7 d6d7 g2g3 g7g6 e3g2 e7c8 e1e8 f8e8 g2f4 d7f7 d1d2 c8d6 g1g2 g6g5');
INSERT INTO evaluations VALUES(49973,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(49973,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(49973,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(49973,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(49973,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(49973,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(49973,36,'cp 22','b1d2 f8e7 g5f6 e7f6 e2e4 e8g8 f1d3 d7d5 e4d5 e6d5 e1g1 c7c5 c2c3 d8b6 d4c5 b6b2 d2b3 b8c6 d3f5 c8f5 d1d5 g8h8 d5f5 a8d8 a1d1 b2a2 d1d8 f8d8 b3d4 a2d5 f5d5 d8d5 f1b1 h8g8 b1b7 d5c5 d4c6 c5c6 b7b8 g8f7 b8a8 f7e6 a8a7 c6c3');
INSERT INTO evaluations VALUES(51807,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(51807,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(51807,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51807,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(51807,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51807,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(51807,36,'cp 4','d7d5 b1c3 d5e4 c3e4 b8d7 f1d3 g8f6 e4f6 d7f6 g1f3 c6c5 c1d2 c5d4 d1e2 d8d5 d3c4 d5e4 e1c1 e4e2 c4e2 f8c5 d2e1 f6d5 g2g3 b7b6 f3d4 c8b7 e2f3 e8c8 d4b3 c5e7 e1f2 d5b4 f3b7 c8b7 a2a3 b4c6 f2d4 h8g8 d4e3 h7h5 b3d2 h5h4');
INSERT INTO evaluations VALUES(53618,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53618,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53618,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53618,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(53618,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(53618,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(53618,36,'cp 13','b1c3 d5e4 c3e4 b8d7 g1f3 g8f6 f1d3 f6e4 d3e4 d7f6 e4d3 c6c5 d1e2 c5d4 c1d2 c8d7 f4f5 f8c5 f5e6 d7e6 f3g5 d8e7 e1g1 e8c8 g5e6 f7e6 a1e1 h8e8 d2g5 e6e5 g1h1 c5d6 g5f6 g7f6 e2f3 c8b8 f3f6 d6b4 e1e4 b4d6');
INSERT INTO evaluations VALUES(56326,30,'cp -147',NULL);
INSERT INTO evaluations VALUES(56326,31,'cp -159',NULL);
INSERT INTO evaluations VALUES(56326,32,'cp -156',NULL);
INSERT INTO evaluations VALUES(56326,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(56326,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(56326,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(56326,36,'cp -149','b8d7 c1b2 g8f6 h5d1 f8d6 d2d3 e8g8 b1d2 f8e8 g1e2 d5d4 e2g3 f6d5 e3d4 e5d4 f1e2 c7c5 e1g1 d6c7 e2f3 d7e5 f1e1 c8e6 f3d5 d8d5 d1h5 e5g6 h5d5 e6d5 e1e8 a8e8 g1f1 b7b6 d2e4 c7f4 a1e1 d5c6 f2f3');
INSERT INTO evaluations VALUES(6795,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(6795,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(6795,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(6795,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(6795,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(6795,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(6795,36,'cp 191','e5d4 d1d4 g8f6 b1c3 d8e7 f2f3 d6e5 d4d3 e8g8 c1e3 c7c6 e1c1 c6d5 c3d5 f6d5 d3d5 e7f6 d5b5 a7a6 b5b6 d7d6 f1c4 c8e6 c4e6 f7e6 c1b1 f6g6 d1d2 a8c8 a2a3 h7h6 b6b4 b7b5 b4b3 g6f6 c2c3');
INSERT INTO evaluations VALUES(10097,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10097,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10097,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10097,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(10097,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10097,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10097,36,'cp 0','e8g8 a2a3 c6d4 c4a2 c7c6 g1f3 f8e8 e1g1 h7h6 c3a4 c5f8 a4c3');
INSERT INTO evaluations VALUES(4324,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4324,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4324,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4324,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4324,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4324,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4324,36,'cp 0','g1e2 a7a5 e1g1 d7d6 c1e3 c5e3 f2e3 c8e6 b2b3 c6b8 d1d2 c7c6 a2a4 d6d5 e4d5 c6d5 c4b5 b8c6 d3d4 d8c7 f1f6 g7f6 a1f1 e5e4 b5c6 b7c6 d2e1 f6f5 e1h4 c7d8 h4g3 g8h8 g3e5 h8g8');
INSERT INTO evaluations VALUES(12114,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(12114,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(12114,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(12114,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(12114,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(12114,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(12114,36,'cp 55','d2f3 d8b6 a2a3 f8e7 h2h4 a7a5 b2b3 b6c7 a1b1 d7b6 h4h5 c5d4 c3d4 c8d7 f1d3 a5a4 g1e2 c6a5 b3b4 a5b3 e1g1 g7g6 f3d2 b3d2 c1d2 b6c4 b1c1 b7b5 c1c3 c7b6 g1h2 e8c8 d1c2 h8g8 h5h6 c8b8 f1c1 d8c8 g2g4 b8a8');
INSERT INTO evaluations VALUES(12113,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(12113,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(12113,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12113,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(12113,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(12113,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(12113,36,'cp 48','d8b6 a2a3 a7a5 b2b3 b6c7 a1b1 d7b6 g1h3 c6e7 f1e2 a5a4 e1g1 a4b3 e2b5 c8d7 d1b3 b6c4 b5d7 c7d7 b3b7 d7b7 b1b7 h7h6 g1h1 c4a5 b7b6 a5c4 b6b5 e7c6 f4f5 c5d4 f3d4 c6d4 c3d4 f8a3 f5e6 e8g8 e6f7 f8f7 f1f7 g8f7 c1a3 c4a3');
INSERT INTO evaluations VALUES(726,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(726,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(726,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(726,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(726,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(726,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(726,36,'cp 51','a2a3 a7a5 b2b3 f8e7 h2h4 b6c7 a1b1 d7b6 h4h5 c8d7 h5h6 g7g6 d4c5 e7c5 f1d3 a5a4 b3b4 c5g1 h1g1 c6a7 c1e3 d7b5 e1f2 b5d3 d1d3 b6c4 e3a7 a8a7 b1a1 a7a6 g1c1 f7f6 c1f1 e8g8 f2g1 f6e5 f3e5 c7b6 g1h2 b6e3 d3e3 c4e3 f1f3 e3f5');
INSERT INTO evaluations VALUES(21774,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(21774,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(21774,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21774,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21774,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21774,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21774,36,'cp 0','f7f6 e5f6 g8f6 g2g3 h7h6 g5d2 c8f5 f1g2 d8d7 e1g1 e8c8 b2b4 f8b4 d2b4 c6b4 d1b3 c7c5 f3e5 d7c7 a2a3 b4a6 b1d2 h8e8 f2f4 f6g4 e5g4 f5g4 f1b1 b7b6 d2f3 a6b8 f3e5 g4e2 g2h3 c8b7 h3g2');
INSERT INTO evaluations VALUES(22271,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22271,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22271,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22271,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22271,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22271,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22271,36,'cp 0','e5f6 g8f6 g2g3 h7h6 g5f6 d8f6 f1g2 c8g4 b1d2 e8c8 e1g1 d4d3 e2e3 f8b4 d1a4 c8b8 h2h3 g4h5 d2e4 f6f8 e4d2');
INSERT INTO evaluations VALUES(22239,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22239,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22239,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22239,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22239,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22239,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22239,36,'cp 0','g8f6 g2g3 h7h6 g5f6 d8f6 f1g2 c8g4 e1g1 e8c8 b1d2 d4d3 e2e3 f8b4 d1a4 h8f8 a2a3 b4d2 f3d2 g4e2 g2c6 e2f1 c6b7 c8b7 a1f1 b7a8 c4c5 f6b2 d2b3 d3d2 a4e4 c7c6 e4c6 a8b8 c6b5 b8a8');
INSERT INTO evaluations VALUES(15865,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15865,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15865,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15865,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15865,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15865,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(15865,36,'cp 0','g2g3 h7h6 g5f6 d8f6 f1g2 c8f5 b1d2 g7g5 d1a4 e8c8 e1g1 d4d3 e2e3 c8b8 a1d1 f8e7 a4b3 f5c8 f3e1 c8f5 e1f3');
INSERT INTO evaluations VALUES(25518,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25518,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(25518,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(25518,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(25518,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(25518,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(25518,36,'cp 50','d2d4 c6d4 b5d7 d8d7 f3d4 e5d4 d1d4 f8e7 c2c4 e8g8 b1c3 d7e6 c1e3 c7c6 b2b3 f6d7 a1d1 d7c5 f2f3 a8d8 f1e1 e7h4 e3f2 h4f2 d4f2 e6e5 c3e2 a7a5 e2d4 g7g6 f2d2 a5a4 b3b4 c5e6 d4e6 f7e6 a2a3 d8d7 d2d4 e5d4 d1d4 e6e5 d4d2 g8h8');
INSERT INTO evaluations VALUES(35169,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(35169,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(35169,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(35169,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(35169,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(35169,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(35169,36,'cp 42','f8c5 d1e2 b8c6 g1f3 h4e7 b1c3 g8f6 a2a4 e8g8 c1g5 c5b4 e1g1 b4c3 b2c3 h7h6 g5h4 c6a5 c4a2 d7d6 f3d2 g7g5 h4g3 c8g4 f2f3 g4e6 a2e6 e7e6 f1e1 f6h5 d2f1 f7f5 e4f5 f8f5 f1e3 f5f7 d3d4 a8e8 d4e5 h5g3 h2g3 e6e5 e2b5 b7b6');
INSERT INTO evaluations VALUES(33288,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(33288,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(33288,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(33288,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(33288,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(33288,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(33288,36,'cp 50','d1e2 b8c6 g1f3 h4e7 e1g1 g8f6 c1g5 c6d4 f3d4 c5d4 c2c3 d4b6 g1h1 d7d6 a2a4 c7c6 a4a5 b6c7 f2f3 h7h6 g5h4 g7g5 h4f2 d6d5 c4b3 e8g8 b1d2 c8e6 e4d5 f6d5 f2g3 a8e8 e2f2 a7a6 f1e1 f7f6 d2e4 f6f5 e4c5 f5f4 c5e6 f4g3');
INSERT INTO evaluations VALUES(38696,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(38696,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(38696,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(38696,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(38696,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(38696,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(38696,36,'cp 19','c3a4 c5b6 a4b6 a7b6 g1e2 g8f6 f2f3 c8e6 c4e6 f7e6 d2d4 b8c6 d4e5 c6e5 e1g1 b6b5 e2f4 d8e7 a2a3 e8g8 d1e2 c7c6 b2b3 f6d7 f4h5 e7h4 h5g3 e5g6 c1e3 h4e7 f3f4 d6d5 e4e5 a8a3 a1a3 e7a3');
INSERT INTO evaluations VALUES(39122,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(39122,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(39122,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(39122,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(39122,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(39122,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(39122,36,'cp 105','f4d6 c7d6 e2e4 g8f6 b1c3 e8g8 d1d2 a7a6 d4d5 b7b5 a2a3 d8a5 f2f3 e6d5 c3d5 a5d2 e1d2 f6d5 e4d5 c8b7 c2c4 f8c8 a1c1 b5c4 f1c4 a6a5 g1e2 b8a6 b2b3 a6c5 b3b4 a5b4 a3b4 c5a4 c4b3 c8c1 h1c1 a4b6 e2c3 a8a3 c1b1 g8f8 h2h4 g7g6 c3b5 a3a8 g2g3 f8e7 b1e1 e7f8');
INSERT INTO evaluations VALUES(40791,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(40791,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(40791,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(40791,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(40791,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(40791,35,'cp 256',NULL);
INSERT INTO evaluations VALUES(40791,36,'cp 251','d4e5 d6e5 d1d8 e7d8 f3e5 c8e6 e5d3 b8d7 c1d2 c7c6 f2f3 d8c7 e1c1 e8c8 b2b3 d7e5 d3c5 h7h5 h2h4 e5g6 c3e2 c8b8 c5d3 a7a5 e2f4 g6f4 d2f4 c7f4 d3f4 b8c7 f1e2 b7b5 h1e1 b5b4 d1d8 h8d8 f4e6 f7e6');
INSERT INTO evaluations VALUES(40746,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(40746,31,'cp 247',NULL);
INSERT INTO evaluations VALUES(40746,32,'cp 239',NULL);
INSERT INTO evaluations VALUES(40746,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(40746,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(40746,35,'cp 260',NULL);
INSERT INTO evaluations VALUES(40746,36,'cp 248','d6e5 d1d8 e7d8 f3e5 c8e6 e5d3 b8d7 f2f3 c7c6 b2b3 d8c7 c3e2 e8c8 c1e3 d7e5 e2f4 h7h6 f4e6 f7e6 e1c1 e5d3 f1d3 c8b8 d3c4 e6e5 a2a4 c7b6 e3d2 a7a6 c2c3 a6a5 c1c2 b8c7 b3b4 h8f8 h1f1 a5b4 c3b4 b6d4 a4a5 b7b5 a5b6 c7b6');
INSERT INTO evaluations VALUES(41114,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(41114,31,'cp 245',NULL);
INSERT INTO evaluations VALUES(41114,32,'cp 248',NULL);
INSERT INTO evaluations VALUES(41114,33,'cp 271',NULL);
INSERT INTO evaluations VALUES(41114,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(41114,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(41114,36,'cp 252','d1d8 e7d8 f3e5 c8e6 e5d3 b8d7 a2a4 a7a5 f2f3 c7c6 c1e3 d8b6 e3b6 d7b6 b2b3 e8c8 e1c1 f6d7 f1e2 c8c7 h2h4 f7f6 f3f4 g7g6 e2f3 h8e8 h1e1 e6f7 c3e2 b6c8 e2d4 c8d6 g2g3 h7h5 e4e5 f6e5 d3e5 d7e5 f4e5');
INSERT INTO evaluations VALUES(42576,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42576,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(42576,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(42576,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42576,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42576,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42576,36,'cp 8','e5d4 e4e5 d7d5 e5f6 d5c4 d1e2 c8e6 f6g7 h8g8 c1g5 c5e7 g5e7 e8e7 e2e4 b8c6 e4h7 e7d7 b1c3 d8f6 c3e4 f6g7 h7g7 g8g7 a1d1 e6g4 e4f6 d7e7 f6g4 g7g4 e1g1 g4f4 f1e1 e7f8 b2b3 c4b3 a2b3 a8d8 g1f1 d8d5 d1d2 a7a5 e1d1 d5c5 h2h4 c6b4 d2d4 f4d4 f3d4 b4c2 d4c2 c5c2 h4h5 f8g7');
INSERT INTO evaluations VALUES(42637,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(42637,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(42637,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(42637,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(42637,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42637,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(42637,36,'cp 29','f3e5 d8e7 e5g4 f6e4 e1g1 d7d5 b1c3 e4c3 d2c3 c8e6 g4e3 e8g8 d1f3 c7c6 c3c4 d5c4 e3c4 b8d7 f3g3 b7b5 c4a5 e6d5 c2c4 b5c4 a5c4 d7f6 c1e3 d5c4 e3c5 e7c5 a1c1 c4d3 c1c5 d3f1 g1f1 f6e4 g3e5 e4c5 e5c5 f8c8 b2b4 a7a5 b4a5 g7g6');
INSERT INTO evaluations VALUES(42531,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42531,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(42531,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(42531,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(42531,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(42531,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42531,36,'cp 27','d8e7 e5g4 f6e4 e1g1 d7d5 b1c3 e4c3 d2c3 c8e6 f1e1 e8g8 h2h3 b8d7 c1f4 c5d6 d1f3 d7c5 f3g3 d6f4 g3f4 c5d3 c2d3 a8e8 a2a4 e7d7 g4e5 d7d6 d3d4 f7f6 e5d3 d6f4 d3f4 e6f7 a4a5 e8e4 f4d3 f8e8 g1f1 c7c6 f2f3 e4e1 d3e1 f7g6 f1f2 h7h5');
INSERT INTO evaluations VALUES(50300,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(50300,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(50300,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(50300,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50300,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(50300,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(50300,36,'cp 62','d7d6 e5d6 e7d6 d2d4 f8e7 f1e2 e8g8 e1g1 b8c6 h2h3 e7f6 b1c3 c8f5 c1f4 f8e8 b2b3 h7h6 d1d2 a7a5 e2d3 f5d3 d2d3 a5a4 a1e1 e8e1 f1e1 a4b3 a2b3 c6b4 d3d2 b4c6 c3e4 b6d7 f4g3 a8a3 d4d5 c6e5 f3d4 e5g6');
INSERT INTO evaluations VALUES(50586,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(50586,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(50586,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(50586,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(50586,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(50586,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(50586,36,'cp -40','e7e6 d2d3 b7b5 c4b3 d7d5 c2c3 d5e4 d3e4 g8f6 h5e2 b8d7 b3c2 c5c4 g1f3 e6e5 e1g1 c8b7 f1d1 d8c7 a2a4 f8c5 a4b5 a6b5 a1a8 b7a8 f3d2 c7a7 b2b4 c5b6 b1a3 a8c6 d2f1 e8g8 f1e3 g7g6 e3d5 c6d5 e4d5 a7a6 g1h1 b6c7 d5d6 c7d6');
INSERT INTO evaluations VALUES(50480,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(50480,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(50480,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(50480,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(50480,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(50480,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(50480,36,'cp -40','c4b3 g8f6 h5e2 b7b5 d2d3 c8b7 g1f3 d7d5 e4d5 f6d5 c1g5 d5f6 e1g1 f8e7 b1d2 h7h6 g5h4 b8d7 f1e1 e8g8 a2a4 d8b6 c2c3 f8e8 b3c2 e7f8 d2e4 f6e4 d3e4 c5c4 a1d1 b6c7 e4e5 b5b4 d1d4 a8c8 h4g3 b4c3');
INSERT INTO evaluations VALUES(54223,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54223,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54223,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54223,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54223,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(54223,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54223,36,'cp 0','g1f3 g8f6 c2c4 d5c4 f1c4 a7a6 c4e2 b8d7 e1g1 b7b5 b2b3 c8b7 c1b2 c5d4 d1d4 f8c5 d4h4 e8g8 b1d2 h7h6 f1c1 a8c8 a3a4 b5b4 h4g3 f6h5 g3h3 h5f6');
INSERT INTO evaluations VALUES(54224,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54224,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(54224,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(54224,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54224,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54224,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54224,36,'cp 0','g8f6 c2c4 d5c4 f1c4 a7a6 e1g1 b7b5 c4d3 c8b7 a3a4 b5b4 b1d2 f8e7 d2b3 b8d7 d1e2 e8g8 e3e4 c5d4 c1f4 f6h5 f4d2 h5f6 d2f4');
INSERT INTO evaluations VALUES(54555,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54555,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(54555,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(54555,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(54555,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(54555,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(54555,36,'cp 1','d3d4 b8c6 f1b5 f8d6 d4c5 d6c5 b1c3 g8e7 c1e3 c5e3 f2e3 e8g8 d1d2 a7a6 b5d3 d8c7 e1c1 c8e6 c1b1 a8d8 c3e2 c6e5 e2d4 e5d3 d4e6 c7b6 d2d3 b6e6 h1e1 h7h6 a2a3 f8e8 f3d4 e6g6 d1d2 g6d3 c2d3 e7c6 d4f5 c6e5');
INSERT INTO evaluations VALUES(5147,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(5147,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5147,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(5147,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(5147,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(5147,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(5147,36,'cp 7','d7d5 c4d3 d5e4 d3e4 g8e7 b1c3 e8g8 c1e3 e7d5 e3g5 d8d6 e4d5 e6d5 c3b5 d6b8 d1d2 c8g4 g5f4 b8d8 f3e5 d8b6 b5c3 g4e6 e5c6 b7c6 a1d1 a7a5 c3a4 b6b5 a4c5 f8e8 a2a3 e6g4 a3a4 b5c4 b2b3 c4b4 f2f3 g4f5 d2b4 a5b4');
INSERT INTO evaluations VALUES(5553,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(5553,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(5553,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5553,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(5553,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(5553,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(5553,36,'cp 4','c4d3 d5e4 d3e4 g8e7 b1c3 e8g8 c1e3 e7d5 e4d5 e6d5 h2h3 c8f5 d1d2 f8e8 e3h6 g7h8 f1e1 d8d7 a1d1 a8d8 a2a3 f7f6 b2b4 d8c8 f3h2 e8e1 d1e1 f5e6 c3e2 c6e7 e2g3 e7f5 g3f5 e6f5 h2g4 g6g5 f2f4 f5g4');
INSERT INTO evaluations VALUES(9060,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(9060,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9060,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9060,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(9060,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9060,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9060,36,'cp 0','c8e6 f6g7 h8g8 c1g5 c5e7 g5e7 d8e7 f3d4 a8d8 c2c3 g8g7 d1f3 c6d4 c3d4 c7c6 b1c3 d8d4 a1d1 d4d1 f3d1 g7g6 d1d4 e8f8 d4a7 b7b5 a7a8 e7e8 a8a5 e8e7');
INSERT INTO evaluations VALUES(24086,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(24086,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24086,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24086,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(24086,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(24086,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(24086,36,'cp 20','f6g7 h8g8 c1g5 c5e7 g5e7 d8e7 f3d4 a8d8 c2c3 g8g7 d1f3 c6d4 c3d4 c7c6 b1c3 d8d4 a1d1 d4d1 f3d1 e8f8 d1d4 g7g6 g2g3 e7f6 d4a7 b7b5 a2a4 e6h3 a7e3 f6g5 e3e8 f8g7 a4b5 c6b5 c3b5 g6b6 e8e5 g5e5 e1e5 b6a6 f2f4 a6a1 g1f2 a1f1 f2e3 f1b1 e5g5 g7f6 b5c3 b1b2 c3e4 f6e7 g5e5 e7f8 e5h5 b2h2 h5h7');
INSERT INTO evaluations VALUES(9797,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(9797,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(9797,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(9797,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(9797,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(9797,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(9797,36,'cp -13','d7d6 c3a4 a7a6 a4c5 d6c5 f1e2 d8d6 f3d2 c8e6 h4f6 g7f6 c2c3 e8c8 d1a4 f6f5 e4f5 e6f5 d2e4 f5e4 a4e4 h8f8 e2g4 c8b8 e1c1 f7f5 g4f5 d6f6 g2g4 c6e7 d1d2 e7f5 g4f5 f6f5 e4f5 f8f5 h1e1 d8f8 c1d1 b7b6 e1f1 h6h5 d2e2 h5h4 h2h3');
INSERT INTO evaluations VALUES(11477,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(11477,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11477,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(11477,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11477,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(11477,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(11477,36,'cp 39','c2c3 d7d6 a2a4 f8e7 f1e1 e8g8 b1d2 c6a5 c4a2 c7c5 d2c4 a5c4 a2c4 c8d7 a4a5 d7c6 c4b3 d8c7 f3h4 f6e4 d3e4 e7h4 d1g4 h4f6 c1h6 b7b5 a5b6 c7b6 a1a3 f8b8 h6c1 b6c7 h2h3 c6d7 g4d1 c5c4 b3a4 d7b5 a4c2');
INSERT INTO evaluations VALUES(12707,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(12707,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(12707,32,'cp 268',NULL);
INSERT INTO evaluations VALUES(12707,33,'cp 288',NULL);
INSERT INTO evaluations VALUES(12707,34,'cp 282',NULL);
INSERT INTO evaluations VALUES(12707,35,'cp 307',NULL);
INSERT INTO evaluations VALUES(12707,36,'cp 310','b1c3 d7d6 e5g3 c8e6 c1g5 d8a5 f1b5 e7c6 e1c1 h7h6 g5f4 e8c8 b5c6 b7c6 f4d6 e6a2 d6f8 h8f8 g3g7 a2e6 c1b1 f8g8 g7f6 a5b6 d1d8 b6d8 f6d8 c8d8 g2g3 d8c7 b1c1 a7a5 h1e1 c5c4 c3a4 h6h5 c1d2 h5h4 d2e3 f7f6 g3h4 g8g2 a4c5 e6c8 e1a1 g2h2 a1a5 h2h4');
INSERT INTO evaluations VALUES(14910,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(14910,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(14910,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(14910,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(14910,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(14910,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(14910,36,'cp 186','b1c3 f8b4 c4d3 f6d7 a2a3 b4c3 b2c3 b7b6 e1g1 c8b7 f1e1 e8g8 d1c2 c6e7 c3c4 e6e5 d4d5 a7a5 a3a4 e7g6 c2c3 d7c5 d3c2 d8d6 c1a3 b7c8 f3d2 c8d7 d2b3 f8b8 a3c5 b6c5');
INSERT INTO evaluations VALUES(19609,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(19609,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(19609,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(19609,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(19609,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(19609,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(19609,36,'cp 151','c1e3 d7d6 b1c3 g8f6 d1d2 e5g4 e3g5 h7h6 g5h4 g7g5 h4g3 c8d7 e1c1 a8c8 f1e2 f8g7 f2f3 g4e5 h2h4 f6h5 g3f2 h5f4 e2f1 d8a5 c1b1 d7e6 f2e3 a7a6 a2a3 e5c4 f1c4 c8c4 d4b3 a5c7 e3d4 e8g8 d4g7 g8g7 g2g3');
INSERT INTO evaluations VALUES(20371,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(20371,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(20371,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(20371,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(20371,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(20371,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(20371,36,'cp 34','f8b4 f1d3 d5c4 d3c4 g8f6 e1g1 h7h6 a2a3 b4a5 b2b4 a5b6 d4d5 e6d5 c3d5 f6d5 c4d5 e8g8 d1b3 c6d4 f3d4 b6d4 a1a2 a7a5 c1e3 d4e3 f2e3 c8e6 d5e6 f7e6 b3e6 g8h8 f1f8 d8f8 a2f2 f8d8 h2h3 d8d3 e6e5 a5b4 a3b4 d3b1 g1h2 b1b4 f2f7 a8g8 f7d7 b4e1 d7b7 e1d2 e5g3 d2d1 b7b8 g8b8 g3b8 h8h7');
INSERT INTO evaluations VALUES(5576,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(5576,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(5576,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(5576,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(5576,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5576,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(5576,36,'cp 33','f1d3 d5c4 d3c4 g8f6 e1g1 h7h6 a2a3 b4a5 b2b4 a5b6 d4d5 e6d5 c3d5 f6d5 c4d5 e8g8 d1b3 c6d4 f3d4 b6d4 a1a2 d8f6 c1e3 a7a5 a2e2 a5b4 e3d4 f6d4 e2e4 d4f6 b3b4 c8d7 b4b7 f6d6 b7a8 f8a8 d5a8 d6a3 e4d4 a3a8 d4d7 a8a5 d7d1 a5a3 h2h3 a3a5 f1e1 g8h8 d1b1 a5a2');
INSERT INTO evaluations VALUES(21370,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(21370,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(21370,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(21370,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21370,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(21370,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(21370,36,'cp 17','f8e7 g5e7 d8e7 e2e4 g8f6 e4e5 f6e4 d1d3 e7b4 a1b1 e4c3 d3c3 b4c3 b2c3 b7b6 h2h4 h7h5 e1d2 c8d7 f1e2 c6e7 g2g3 a7a6 h1e1 d7b5 a2a4 b5e2 e1e2 a8c8 f3g5 e8g8 e2e1 e7f5 g5h3 f8d8 h3f4');
INSERT INTO evaluations VALUES(5609,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(5609,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(5609,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(5609,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5609,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(5609,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(5609,36,'cp 28','g5e7 d8e7 e2e4 g8f6 e4e5 f6e4 d1d3 e7b4 a1b1 b4a5 f1e2 b7b6 e1g1 e4c3 b2c3 c8d7 c3c4 d5c4 d3c4 a8c8 c4d3 c6e7 f1d1 d7a4 d1e1 c7c5 f3d2 e8g8 d2c4 a5a6 d3f3 a4c6 c4d6 a6a2 d6c8 f8c8 f3h3 e7g6 d4c5 a2c2 c5b6 a7b6');
INSERT INTO evaluations VALUES(26185,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26185,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26185,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(26185,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(26185,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26185,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26185,36,'cp 0','g8f6 d1e2 e7e6 e1g1 f8e7 f3e5 d7b6 c2c3 b6c4 e2c4 f6d7 e5g6 h7g6 f1e1 d7f6 c4e2 d8c7 c1g5 a8d8 g3f1 e8g8 a1d1 d8d7 g2g3 f8d8 g5f4 e7d6 f4g5');
INSERT INTO evaluations VALUES(27747,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(27747,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(27747,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(27747,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(27747,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(27747,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(27747,36,'cp 30','e1c1 b8d7 g2g4 b7b5 g4g5 f6h5 c1b1 d7b6 b3a5 a8c8 a2a3 g7g6 h2h4 d8c7 c3d5 e6d5 e4d5 b6d5 d2d5 c7a5 f1d3 a5c7 c2c4 h5g3 h1g1 g3f5 d3f5 g6f5 c4c5 f5f4 c5d6 c7c2 b1a2 e7d6 e3f4 e5f4 d5d6 c8d8 d6d8 f8d8 d1d8 g8g7 h4h5 c2c6 h5h6 g7g6 d8g8 g6f5');
INSERT INTO evaluations VALUES(27746,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(27746,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(27746,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(27746,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(27746,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(27746,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(27746,36,'cp 28','b8d7 g2g4 b7b5 g4g5 f6h5 c1b1 d7b6 b3a5 a8c8 a2a3 g7g6 h2h4 d8c7 c3d5 e6d5 e4d5 b6d5 d2d5 c7a5 f1d3 a5c7 c2c4 b5c4 d1c1 c7c6 d5c6 c8c6 d3c4 h5g3 c4d3 c6c1 h1c1 g3f5 e3f2 h7h6 d3a6 h6g5 h4g5 e7g5 c1c6 g5e7 a3a4 f5d4 f2d4 e5d4 a6c4');
INSERT INTO evaluations VALUES(23601,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(23601,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(23601,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(23601,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(23601,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(23601,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(23601,36,'cp 22','g2g4 b7b5 g4g5 f6h5 c3d5 e6d5 e4d5 f7f6 g5f6 e7f6 b3a5 d8e8 c1b1 f6d8 a5b7 d8c7 d2c3 e8b8 c3c6 f8f3 e3c5 f3f6 c5d6 c7d6 c6d7 h5f4 b7d6 b8d6 d7d6 f6d6 d1e1 f4d5 c2c4 b5c4 f1c4 a8e8 h1g1 g8h8 g1g5 d5b4 a2a3 b4d3 e1e4 e8f8 c4d3 d6d3 g5e5');
INSERT INTO evaluations VALUES(23600,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(23600,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(23600,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(23600,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(23600,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23600,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(23600,36,'cp 25','b7b5 g4g5 f6h5 c1b1 d7b6 b3a5 d8c7 a2a3 a8b8 c3d5 e6d5 e4d5 b6d5 d2d5 c7a5 f1c4 a5d8 c4a2 e7g5 e3g5 d8g5 d5d6 g5f6 d6f6 h5f6 d1d6 b8a8 h1g1 g8h8 d6b6 a8a7 c2c4 b5c4 a2c4 a6a5 g1e1 a7e7 b6b5 f8c8 c4b3 e5e4 f3e4');
INSERT INTO evaluations VALUES(23635,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(23635,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(23635,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(23635,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(23635,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(23635,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(23635,36,'cp 37','a2a3 d8c7 g4g5 f6h5 c3d5 e6d5 e4d5 d7b6 c1b1 a8b8 b3a5 b6d5 d2d5 c7a5 f1c4 a5d8 c4a2 e7g5 e3g5 d8g5 d5d6 g5f6 d6f6 h5f6 d1d6 b8a8 h1g1 g8h8 d6b6 a8a7 g1e1 e5e4 f3e4 f6d7 b6d6 d7e5 e1f1 g7g6 f1f6 h7h5 d6a6 a7a6 f6a6 e5f3 e4e5 f3h2 a6b6 h5h4 b6b5 h2f1 e5e6 f7e6 a2e6');
INSERT INTO evaluations VALUES(30753,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(30753,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(30753,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(30753,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(30753,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(30753,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(30753,36,'cp 141','b1c3 f8g7 h2h4 h7h6 d4d5 a7a6 h4h5 g6g5 f3d4 g7e5 g2g3 c7c5 d4e2 d7d6 f2f4 e5g7 f4g5 h6g5 c1g5 b6b5 d1d2 g7e5 h5h6 g8f6 f1g2 b8d7 c3d1 b5b4 g5f4 d8b6 f4e5 d7e5 d1e3 a6a5 e3f5 a5a4 e1c1 e5c4');
INSERT INTO evaluations VALUES(30770,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(30770,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(30770,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30770,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(30770,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(30770,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(30770,36,'cp 9','c7c5 c2c3 e7e6 d2d4 b8c6 f1d3 c5d4 c3d4 c6b4 d3e2 a8c8 b1c3 b4d5 c1d2 g8e7 e1g1 d5c3 b2c3 e7f5 e2d3 f8e7 c3c4 d7d5 e5d6 f5d6 a1c1 e8g8 f1e1 d6e8 d2c3 e8f6 f3e5 f6d7 e5d7 d8d7 d4d5 e6d5 d1e2 c8e8 e2h5');
INSERT INTO evaluations VALUES(31494,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(31494,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(31494,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(31494,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(31494,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(31494,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31494,36,'cp 0','c6c5 c2c3 b8c6 e1g1 d8b6 b1d2 c5d4 c3d4 c6b4 d3b3 b4c6');
INSERT INTO evaluations VALUES(31655,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(31655,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(31655,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31655,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(31655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31655,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31655,36,'cp 17','c2c3 b8c6 b1d2 c5d4 c3d4 c6b4 d3b3 d8b6 e1g1 b4c6 b3d3 c6b4 d3c3 g8e7 a2a3 b4c6 d2f3 b6b5 c3c2 a8c8 c2d1 h7h5 b2b4 e7f5 a1b1 f8e7 e2g3 f5g3 f2g3 a7a5 f3g5 e7g5 c1g5 a5b4');
INSERT INTO evaluations VALUES(36512,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36512,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36512,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36512,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36512,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36512,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36512,36,'cp 0','d2d3 c7c5 a1b1 b8c6 g1e2 a7a6 e1g1 a8b8 a2a3 c8g4 b2b4 b7b5 c4b5 a6b5 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 e4e5 f6e8 b1a1 a3b2 a1b1 b2a3');
INSERT INTO evaluations VALUES(36514,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36514,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36514,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36514,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36514,36,'cp 0','c7c5 a1b1 a7a6 g1e2 b8c6 e1g1 a8b8 a2a3 c8g4 b2b4 b7b5 c4b5 a6b5 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 e4e5 f6e8 b1a1 a3b2 a1b1');
INSERT INTO evaluations VALUES(44822,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(44822,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(44822,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(44822,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(44822,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(44822,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44822,36,'cp 71','e7e6 f1d3 f8b4 e1g1 e8g8 c1d2 c6c5 a2a3 b4c3 d2c3 b8d7 b2b4 d5c4 d3c4 d7b6 c4e2 b6d5 c3e1 c5d4 d1d4 b7b6 e1d2 c8b7 f1c1 d5e7 d4b2 a8c8 c1c8 d8c8 f3d4 e6e5 d4b5 f8d8 f2f3 c8a8');
INSERT INTO evaluations VALUES(45546,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(45546,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(45546,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(45546,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(45546,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(45546,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(45546,36,'cp 75','f1d3 f8b4 e1g1 e8g8 c1d2 d5c4 d3c4 b8d7 d1c2 b7b5 c4e2 b4d6 f1d1 d8e7 a1c1 a7a6 d2e1 e6e5 c3e4 f6e4 c2e4 d7f6 e4c6 c8b7 c6c2 a8c8 c2b1 c8c1 d1c1 b7e4 b1a1 e5d4 f3d4 e7e5 g2g3 f8e8 b2b3 e4b7 e2f3 b7f3 d4f3');
INSERT INTO evaluations VALUES(45547,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(45547,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(45547,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(45547,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(45547,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(45547,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(45547,36,'cp 69','d5c4 d3c4 b7b5 c4b3 b5b4 c3e2 b8d7 e1g1 f8d6 f1e1 c6c5 e3e4 f6e4 e2f4 d7f6 b3a4 c8d7 f3e5 d7a4 d1a4 e8f8 f2f3 e4g5 h2h4 g5h7 a4b3 d8c7 e5g6 f8g8 f4e6 f7e6 b3e6 c7f7 g6h8 f7e6 e1e6 a8d8 h8g6 g8f7 e6d6 d8d6 c1f4 d6d4 g6e5 f7g8');
INSERT INTO evaluations VALUES(50410,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(50410,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(50410,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(50410,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(50410,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(50410,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(50410,36,'cp 203','d2d4 b4a6 f2f4 c7c6 g1f3 d7d5 b1c3 a6c7 f1e2 g7g6 e1g1 b8a6 d1b3 f8g7 c4d5 c7d5 c3d5 c6d5 e2b5 c8d7 b3d5 a6c7 b5d7 d8d7 d5b7 e8g8 f1f2 f8b8 b7e4 a7a5 f2c2 a5a4 c1d2 c7d5 e4e2');
INSERT INTO evaluations VALUES(53954,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(53954,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(53954,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(53954,33,'cp 246',NULL);
INSERT INTO evaluations VALUES(53954,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(53954,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(53954,36,'cp 250','e2e4 g8f6 b1c3 f8e7 g1e2 b8c6 e2f4 e8g8 f1e2 c8e6 c1e3 a8c8 e1g1 c6e5 d1d2 a7a6 a2a4 d8c7 f1d1 e5c4 e2c4 c7c4 f2f3 b7b5 a4b5 a6b5 e3d4 b5b4 c3e2 c4c2 d2b4 c2b3 b4b3 e6b3 d1d3 b3c4 d3c3 c4b5 c3c8');
INSERT INTO evaluations VALUES(54997,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(54997,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(54997,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(54997,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(54997,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(54997,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(54997,36,'cp 8','g2g3 c7c5 c2c4 g7g6 b1c3 f8g7 e2e3 e7e6 g1e2 b8c6 d3d4 c5d4 e3d4 g8e7 f1g2 e8g8 e1g1 c8d7 b2b3 d6d5 c4d5 e7d5 c3d5 e6d5 c1e3 d7g4 d1d2 g4e2 d2e2 c6d4 e3d4 g7d4 a1d1 d8b6 g2d5 d4c5 e2c4 a8c8 c4e4 c8e8 e4f3 e8e7');
INSERT INTO evaluations VALUES(3174,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3174,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3174,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3174,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(3174,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3174,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3174,36,'cp 3','c1e3 b7b6 d1d2 c8b7 e3g5 g8h8 g5h6 e7e6 c4b3 e6e5 a1e1 g7h6 d2h6 f6g8 h6d2 d8d7 f3h2 f7f5 f2f4 a8e8 b3a4 a7a6 a4c6 b7c6 c3d5 c6d5 e4d5 g8f6 h2f3 e5f4 c2c4 e8e3 e1e3 f4e3 d2e3 f8e8 e3f2 h7h6 f1e1 e8e1 f2e1 h8g7');
INSERT INTO evaluations VALUES(3175,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3175,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3175,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3175,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3175,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3175,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3175,36,'cp 0','b7b6 d1d2 c8b7 f1e1 e7e6 e3g5 a8c8 a2a4 d8d7 g5h6 a7a6 c4b3 d7c7 a1d1 c8d8 d2g5 c6e5 f3h2 e5c6');
INSERT INTO evaluations VALUES(7183,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(7183,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(7183,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(7183,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(7183,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(7183,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(7183,36,'cp 96','d2d4 c5a7 d4e5 f6e4 d1e2 d6d5 c1e3 c8g4 e3a7 a8a7 h2h3 g4f3 e2f3 e8g8 f3e3 a7a8 c3c4 c6e5 c4d5 f7f5 b1d2 e4d2 e3d2 e5c4 d2d4 b7b5 a4b3 d8f6 d4d3 f6b2 a1b1 b2f6 b3c4 b5c4 d3c4 f6d6 f1d1 a8e8 b1c1 e8e7 c4d3 h7h6 c1c6 d6b4 c6e6 e7e6 d5e6');
INSERT INTO evaluations VALUES(7180,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(7180,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(7180,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(7180,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(7180,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(7180,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(7180,36,'cp 90','c5a7 d4e5 f6e4 d1e2 d6d5 c3c4 e4c5 a4c6 b7c6 c1e3 c5e6 e3a7 a8a7 b1c3 e8g8 f1d1 d5d4 c3e4 d8e7 f3d4 e6d4 d1d4 e7e5 a1d1 a7a8 e4g3 e5e2 g3e2 a6a5 e2f4 f8e8 f2f3 a5a4 g1f2 h7h6 h2h3 g8f8 d1d2 c8e6 h3h4 a8b8 f2g3 e6f5 a2a3 f5e6');
INSERT INTO evaluations VALUES(15707,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(15707,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(15707,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(15707,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(15707,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(15707,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(15707,36,'cp -53','d8f6 b1c3 e4c3 d2c3 f8d6 f1e1 c8e6 d1d4 e8g8 c1f4 f8e8 e5d3 f6d4 c3d4 e6f5 e1e8 a8e8 f4d6 c7d6 a1e1 g8f8 e1e8 f8e8 d3e1 c6c5 d4c5 d6c5 f2f3 e8d7 g1f2 d7d6 f2e3 b7b5 e1d3 f5e6 a2a3 e6f5 d3e1 a7a5 e1d3 d6d5 g2g4');
INSERT INTO evaluations VALUES(20348,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(20348,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(20348,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(20348,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(20348,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(20348,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(20348,36,'cp -28','g1f3 g8e7 f1d3 f7f5 e1g1 a7a6 d3e2 d5d4 d2d3 a8b8 c3c4 b7b5 b2b3 b5c4 d3c4 c8b7 c1a3 d8a5 d1e1 a5b6 e1c1 e7g6 f3e1 f8e7 e1d3 e8g8 b1d2 b6a5 e2h5 f8c8 h5g6 h7g6 d2f3');
INSERT INTO evaluations VALUES(20421,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(20421,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(20421,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(20421,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(20421,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20421,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(20421,36,'cp 9','d8b6 d1d2 c8d7 b1c3 a8c8 a1c1 c6a5 f1d3 a5c4 d3c4 c8c4 g1e2 g8e7 e1g1 e7f5 e2g3 f8b4 g3f5 e6f5 a2a3 b4c3 c1c3 c4c3 d2c3 e8g8 f1c1 d7e6 e3d2 h7h6 c3b4 b6b4 d2b4 f8c8 c1c8 e6c8 f2f4 g8h7 g1f2 h7g6 b4f8 c8d7 f2e3 a7a6 h2h3 g6h7 e3d2');
INSERT INTO evaluations VALUES(20712,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(20712,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(20712,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(20712,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(20712,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20712,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(20712,36,'cp 0','g8f6 f1e1 f8e7 c4f1 e8g8 c2c4 d6d5 e4d5 e6d5 b1c3 c8g4 d1d2 c6d4 d2d4 d5c4 d4d8 e7d8 f1c4 a8c8 c4b3 c8c6 c1e3 a7a6 c3a4 d8e7 a4b6 e7d8');
INSERT INTO evaluations VALUES(13986,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(13986,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(13986,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(13986,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(13986,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13986,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13986,36,'cp 8','f1e1 f8e7 c4f1 e8g8 c2c4 d6d5 e4d5 e6d5 d4c6 b7c6 c4d5 f6d5 b1d2 c8e6 d2e4 h7h6 h2h3 a7a5 c1d2 d5f6 e4c3 f6d5 c3d5 d8d5 d1c2 e7f6 d2c3 d5c5 e1c1 f6c3 c2c3 c5c3 c1c3 f8b8 b2b3 a5a4 b3a4 a8a4 c3c6 a4a2');
INSERT INTO evaluations VALUES(21637,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(21637,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(21637,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(21637,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(21637,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(21637,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(21637,36,'cp 165','c4d5 c6e5 g1f3 e5f3 e2f3 f8d6 b1c3 g8e7 f1d3 e8g8 e1g1 h7h6 f1e1 a7a6 d3e4 b7b5 a2a3 f8e8 d1c2 e7g6 c3e2 g6h4 c1d2 f7f5 e4d3 c8d7 a1c1 d8f6 e2g3 e8e1 d2e1 a8f8 e1b4 d6b4 a3b4 f5f4 g3e4 f6g6');
INSERT INTO evaluations VALUES(21563,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(21563,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(21563,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(21563,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(21563,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(21563,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(21563,36,'cp 154','c6e5 b1c3 g8f6 c1f4 e5g6 f4g3 h7h5 h2h3 f8d6 g3d6 d8d6 g1f3 e8g8 e2e3 f8d8 f1c4 g6e5 f3e5 d6e5 d1e2 a7a6 a2a4 e5g5 e2f3 c8d7 e1g1 b7b5 a4b5 a6b5 c4d3 a8a1 f1a1 g5e5 a1c1 b5b4 c3e4 e5b2 e4f6 b2f6 f3f6 g7f6 c1c7');
INSERT INTO evaluations VALUES(22165,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(22165,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(22165,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(22165,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(22165,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(22165,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(22165,36,'cp 163','b1c3 g8f6 c1f4 e5g6 f4g3 h7h5 h2h3 f8d6 g3d6 d8d6 g1f3 e8g8 e2e3 f8d8 f1c4 g6e5 f3e5 d6e5 d1e2 b7b5 c4b5 f6d5 c3d5 e5d5 e1g1 d5g5 g1h1 a8b8 b5c4 c8b7 f2f3 b7d5 c4d3 d8d6 e3e4 d5b7 d3c4 b8d8 a1d1 d6d2 d1d2 d8d2 f3f4 d2e2 f4g5');
INSERT INTO evaluations VALUES(28845,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(28845,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(28845,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(28845,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28845,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28845,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28845,36,'cp 0','d8d7 b1c3 g7g6 b2b3 f8g7 c1b2 e8g8 d1e2 b8c6 d4c6 d7c6 e1c1 f8e8 f2f3 a8c8 c3b5 c6c5 b2d4 c5g5 d4e3 g5e5 e3d4');
INSERT INTO evaluations VALUES(32565,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32565,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32565,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32565,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32565,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32565,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32565,36,'cp 0','d1e2 g7g6 b2b3 f8g7 c1b2 e8g8 c2c4 d7g4 f2f3 g4g5 d4c2 g5a5 b2c3 a5c5 c3d4 c5a5');
INSERT INTO evaluations VALUES(31628,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31628,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31628,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31628,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(31628,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31628,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31628,36,'cp 0','b8d7 c1e3 c5c4 e1g1 a7a6 b5d7 d8d7 f3h4 f5g6 c3e2 g8e7 e2f4 e7f5 h4f5 g6f5 f4h5 d7a4 b2b3 a4c6 h2h3 f5g6 a2a4 c4b3 c2b3 c6c2 d1f3 c2e4 f3d1');
INSERT INTO evaluations VALUES(32423,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(32423,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(32423,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(32423,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(32423,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(32423,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(32423,36,'cp 154','h2h3 g4f3 d1f3 b8d7 f3b3 d7c5 b3b5 d8d7 b5c4 g7g6 c1e3 f6f5 b1c3 f5f4 e3c5 d6c5 a2a4 e8c8 c4b3 g8f6 e1c1 a7a6 f2f3 d7d6 c3b1 d6b6 b3b6 c7b6 b1d2 f6d7 h3h4 f8e7 c2c3 h7h6 c1c2 g6g5 h4h5 h8g8 f1e2 g8g7 d2c4 g7h7 d5d6 e7f6 d1d5 b6b5 a4b5');
INSERT INTO evaluations VALUES(34802,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(34802,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(34802,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(34802,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(34802,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(34802,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(34802,36,'cp 142','c8e6 c3b5 d6c5 c4e6 f7e6 a2a4 a7a6 d2d4 c5b6 f3e5 g8f6 b5c3 b8c6 e5c6 b6c6 e1g1 e8c8 d1d3 c8b8 c1e3 h7h6 f1b1 f8b4 c3e2 f6g4 c2c3 b4d6 h2h3 g4e3 d3e3 h8f8 b2b4 c6c4 b4b5 a6a5 b5b6 d8c8 g2g3 f8f5 b6c7 c8c7 e2f4');
INSERT INTO evaluations VALUES(35087,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(35087,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(35087,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(35087,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(35087,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(35087,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(35087,36,'cp 26','f8c5 g1f3 d7d6 f1d3 f6g6 e1g1 g8f6 f3h4 g6h5 d1h5 f6h5 b1c3 a7a5 f1e1 c8d7 a2a3 b8a6 h2h3 h7h6 c1d2 h5f6 g2g4 g7g5 h4f5 d7f5 e4f5 e8e7 g1g2 c5d4 c3b5 d4c5 d3c4 c7c6 b5c3 a6c7 d5c6 b7c6');
INSERT INTO evaluations VALUES(33259,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(33259,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(33259,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(33259,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(33259,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(33259,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(33259,36,'cp 25','g1f3 d7d6 b1c3 a7a5 f1d3 f6e7 e1g1 g8f6 a2a3 e8g8 c3a4 c5a7 c2c4 c8g4 h2h3 g4h5 a4c3 b8d7 d3c2 h7h6 c1d2 c7c6 d1e2 c6d5 c4d5 a8c8 a1e1 h5f3 e2f3 a7d4 f3e2 d7c5');
INSERT INTO evaluations VALUES(33258,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33258,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(33258,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(33258,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(33258,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(33258,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(33258,36,'cp 20','d7d6 f1d3 f6g6 e1g1 g8f6 f3h4 g6h5 d1h5 f6h5 b1c3 a7a5 a2a3 h5f4 d3b5 c7c6 b5c4 f4g6 h4f5 c8f5 e4f5 g6e7 d5c6 b7c6 c3e4 b8d7 c1e3 e7f5 e3c5 d6c5 a1e1 e8e7 e4c5 f7f6 c5d7 e7d7 c4d3 f5d6 f2f4 a8e8 f4e5 e8e5 e1e5');
INSERT INTO evaluations VALUES(39810,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(39810,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(39810,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39810,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(39810,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39810,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(39810,36,'cp 25','g5f6 e7f6 c4d5 e6d5 g2g3 e8g8 f1g2 c8b7 e1g1 b8a6 e2e3 c7c5 a1c1 a6c7 d4c5 b6c5 c3a4 c5c4 b2b3 c7e6 f3d2 c4b3 d2b3 a8c8 a4c5 e6c5 b3c5 b7c6 d1g4 d8e7 f1d1 f6b2 c1b1 b2f6 c5b3 c6d7 g4e2 d7f5 b1c1 f8d8 c1c8 d8c8 g2d5');
INSERT INTO evaluations VALUES(42543,30,'cp -314',NULL);
INSERT INTO evaluations VALUES(42543,31,'cp -303',NULL);
INSERT INTO evaluations VALUES(42543,32,'cp -317',NULL);
INSERT INTO evaluations VALUES(42543,33,'cp -327',NULL);
INSERT INTO evaluations VALUES(42543,34,'cp -313',NULL);
INSERT INTO evaluations VALUES(42543,35,'cp -291',NULL);
INSERT INTO evaluations VALUES(42543,36,'cp -291','b8c6 g1e2 d7d5 c4d5 f6d5 e4d5 c6b4 b1a3 b4d5 d2d3 c8e6 h1g1 d8d7 c1d2 c5a3 b2a3 e8c8 f3e4 h8e8 e1c1 d5f6 e4e3 d7a4 e3c5 d8d5 c5b4 a4c6 g4g5 a7a5 b4c3 d5c5 c3b2 c5b5 b2a1 c6c5 d2e3 c5a3 c1d2 f6g4 e2c3 b5b4 h2h3 g4e3 f2e3 e8d8 c3e2 b4b5 e2c3 b5b6 d1b1');
INSERT INTO evaluations VALUES(44587,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44587,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(44587,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44587,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(44587,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44587,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(44587,36,'cp 5','b8c6 c2c4 d7d5 c4d5 f6d5 d2d4 c5d4 f3d4 d5b4 d4c6 d8d1 e1d1 b4c6 c1e3 e7e5 g2c6 b7c6 f2f3 h7h5 b1d2 c8e6 d1e1 f7f5 h2h4 a8b8 d2b3 a7a5 b3a5 f8b4 e3d2 b4d2 e1d2 b8b2 d2d3 e8d7 h1c1 b2a2 a1a2 e6a2 a5c6 d7d6 e2e4 f5e4 f3e4 a2e6 d3e3 h8a8');
INSERT INTO evaluations VALUES(12855,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12855,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(12855,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(12855,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(12855,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(12855,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(12855,36,'cp 33','e1g1 e7e5 e2e4 f8e7 b1c3 d7d6 d2d3 e8g8 a2a4 c8e6 f3g5 e6g4 f2f3 g4c8 f3f4 e5f4 g3f4 f8e8 d1d2 e7f8 b2b3 h7h6 g5f3 d6d5 e4d5 f6d5 c3d5 d8d5 c1b2 d5d6 f3e5 c8f5 a1e1 a8d8 d2f2 f7f6 e5c6 b7c6');
INSERT INTO evaluations VALUES(12854,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12854,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(12854,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(12854,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(12854,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(12854,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(12854,36,'cp 10','e7e5 e2e4 f8e7 b1c3 e8g8 d2d3 d7d6 a2a4 b7b6 h2h3 c8e6 f3g5 e6c8 f2f4 c8b7 g1h2 e5f4 c1f4 c6d4 f1f2 a8b8 f4d2 a7a6 g5f3 d4f3 g2f3 f6d7 d2f4 d7e5 f4e5 d6e5 d1e2 d8c7 a1f1 e7g5 f3g2 b8d8');
INSERT INTO evaluations VALUES(4153,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(4153,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(4153,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(4153,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(4153,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(4153,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4153,36,'cp 21','c5b6 a2a4 a7a6 f1e1 d7d6 b1d2 c8e6 c4b3 e6b3 d1b3 f8e8 c1b2 d6d5 a1d1 d5e4 d2e4 f6e4 d3e4 d8f6 a4a5 b6a7 d1d5 a8d8 b2c1 f6e6 c1e3 a7e3 e1e3 d8d6 c3c4 c6d4 b3d1 d4f3 e3f3 e8d8 f3d3 e6e7 d5d6 d8d6 d3d6 c7d6');
INSERT INTO evaluations VALUES(4128,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(4128,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(4128,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(4128,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(4128,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(4128,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(4128,36,'cp 34','a2a4 a7a6 b1d2 d7d6 h2h3 c8e6 f1e1 b6a7 c4b3 e6b3 d1b3 f8e8 d2f1 d6d5 e1d1 d8d6 c1e3 a7e3 f1e3 a8d8 b4b5 c6e7 b3c2 c7c6 c3c4 d6e6 b5a6 b7a6 c4c5 f6d7 a4a5 d8a8 a1c1 a8b8 d3d4 e5d4 e4d5 e7d5 e3d5');
INSERT INTO evaluations VALUES(4132,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(4132,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(4132,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(4132,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(4132,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(4132,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(4132,36,'cp 31','a7a6 f1e1 d7d6 b1d2 f6g4 e1f1 g4f6 h2h3 c8e6 f1e1 b6a7 c4b3 f8e8 b3e6 f7e6 a1b1 e8f8 b4b5 a6b5 a4b5 c6e7 d2c4 e7g6 d3d4 f6h5 c1e3 h5f4 b5b6 c7b6 d4e5 d6d5 c4b6 a7b6 b1b6 d5e4 d1d8 f8d8 f3d4 f4d3');
INSERT INTO evaluations VALUES(2649,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(2649,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(2649,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(2649,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(2649,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(2649,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(2649,36,'cp 20','b1d2 d7d6 h2h3 c8e6 c4e6 f7e6 d2c4 b6a7 c1e3 a7e3 f2e3 d8e7 d1c2 a6a5 b4b5 c6d8 d3d4 f6d7 d4e5 d8f7 e5d6 f7d6 c4d6 c7d6 a1d1 a8c8 e4e5 d6d5 d1d4 e7f7 d4f4 f7h5 c2f2 f8f4 e3f4 d7c5 f3d4 c5e4 f2f3 h5f3 f1f3 e4c3 g1h2 c3a4 d4e6 a4c5 e6c5 c8c5');
INSERT INTO evaluations VALUES(4173,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4173,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(4173,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(4173,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(4173,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(4173,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(4173,36,'cp 0','a7a5 h2h3 d7d6 c3e2 c8e6 c4e6 f7e6 e2g3 d6d5 c2c3 d8d7 c1d2 d7f7 e4d5 e6d5 d3d4 e5d4 c3d4 c5b6 d2c3 f7g6 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(8574,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(8574,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8574,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(8574,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8574,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(8574,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(8574,36,'cp 0','a2a4 c8e6 b1d2 d6d5 c4b5 f6d7 g5e7 d8e7 e1g1 e8g8 f1e1 a8d8 c2c3 d5e4 e1e4 f7f6 e4e1 e6d5 b2b4 e7f7 d1c2 f8e8 a4a5 a7a6 b5a4 f7g6 a1b1 g8h8 f3h4 g6g4 h4f3');
INSERT INTO evaluations VALUES(9497,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(9497,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(9497,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(9497,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(9497,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(9497,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9497,36,'cp -10','d8e7 e5d3 b4c3 d2c3 e7e4 d1e2 e4e2 e1e2 c8f5 h1e1 e8c8 c1g5 d8d6 a1d1 h8e8 e2f1 e8e1 d1e1 f5d3 c2d3 d6d3 e1e7 d3d7 g5f6 g7f6 e7e8 d7d8 e8e7 d8d1 f1e2 d1b1 b2b4 b1b2 e2e3 b2a2 e7f7 a2c2 f7f6 c2c3 e3d4 c3b3 d4c5 a7a5 b4a5 b3b5 c5c4 b5a5 f6f8 c8d7 f8f7 d7d6 f7f6 d6e7 f6h6 a5g5 h6h7 e7d6 h7h6 d6e5 h6h7 g5g4 c4c3 g4g2 h7c7 e5d5 c7b7');
INSERT INTO evaluations VALUES(10460,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(10460,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(10460,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(10460,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(10460,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(10460,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(10460,36,'cp 23','c3c4 e8g8 g2g3 a7a6 a2a4 d7d6 f1g2 a8b8 e1g1 b7b5 a4b5 a6b5 c4b5 b8b5 h2h3 h7h6 c1e3 c8e6 f3d2 d8d7 g1h2 b5b2 d1c1 b2b5 c2c4 b5b7 f2f4 d7e7 c1c3 f8b8 f1b1 b7b1 a1b1 b8b1 d2b1 e7d8 f4f5 e6d7 c3d2');
INSERT INTO evaluations VALUES(10482,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(10482,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(10482,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(10482,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(10482,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(10482,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(10482,36,'cp -86','b1d2 f6e4 d4d5 d8e7 f1e2 e7c5 e1g1 e4d2 f3d2 c5d5 e2c4 d5f5 f1e1 b4e7 d2f1 d7d6 f1e3 f5d7 b2b4 e8g8 c1b2 c6e5 c4b3 c7c6 b4b5 d7d8 b5c6 b7c6 a1c1 a8b8 b2e5 d6e5 d1f3 e7b4 e1d1 d8g5 f3c6 b4a3 c1c2 g8h8 e3d5');
INSERT INTO evaluations VALUES(16485,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(16485,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(16485,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(16485,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(16485,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(16485,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(16485,36,'cp -136','d7d5 f1b5 e5d4 g1e2 d5e4 e2d4 a7a6 b5a4 c8d7 a4c6 d7c6 d4c6 d8d1 e1d1 b7c6 b1d2 e8c8 d1c2 d8e8 h1e1 f8d6 d2e4 d6h2 e1h1 h2e5 h1h4 f6e4 h4e4 e5d6 c1e3 e8e4 f3e4 f7f6 a1f1 h7h5 f1f5 h5h4 f5a5 c8b7 a5f5 h8e8 c2d3 d6e5 e3d4 g7g6');
INSERT INTO evaluations VALUES(15123,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(15123,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(15123,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(15123,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(15123,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(15123,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(15123,36,'cp -131','f1b5 e5d4 g1e2 d5e4 e2d4 c8d7 b5c6 d7c6 d4c6 d8d1 e1d1 b7c6 b1d2 e8c8 d1c2 e4f3 d2f3 d8e8 c1d2 f8d6 a1e1 f6e4 h1f1 c6c5 g2g3 h8f8 d2c1 c8d7 f3d2 e4d2 c1d2 f7f6 c2d3 h7h5 e1e8 f8e8 d2f4 c5c4 d3c4 e8e2 b2b4 d6f4 f1f4 e2a2 h2h4 a2g2 f4f5 g2g3 f5h5 g3g2');
INSERT INTO evaluations VALUES(24549,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(24549,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(24549,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(24549,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(24549,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(24549,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(24549,36,'cp 41','c1f4 e7e6 e1c1 f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e2 d8d5 f3e5 d5e4 e2e4 f6e4 f4e3 e7d6 c2c3 e4f6 b1c2 a7a5 a2a4 f6d5 g2g4 d6e5 d4e5 d5e3 f2e3 f7f5 e5f6 f8f6 h1g1 f6f2 c2b3 a8f8 d1d7 f8f7 g1d1 f2g2 d7f7 g8f7 d1d7 f7f6');
INSERT INTO evaluations VALUES(24548,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(24548,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24548,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(24548,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(24548,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24548,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24548,36,'cp 31','e7e6 e1c1 f8e7 g3e4 f6e4 d3e4 d7f6 e4d3 d8d5 c2c4 d5a5 c1b1 f6g4 f4g3 e8g8 f3e5 g4e5 g3e5 a8d8 g2g4 e7d6 f2f4 f7f6 e5d6 d8d6 d3e4 a5d8 h1g1 c6c5 d4d5 e6d5 d1d5 d6d5 e4d5 d8d5 c4d5 g8f7 g1c1 f8c8 a2a3 b7b5 f4f5');
INSERT INTO evaluations VALUES(24490,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(24490,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(24490,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(24490,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(24490,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24490,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24490,36,'cp 51','e1c1 f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e1 d8d5 f3e5 d5e4 e1e4 f6e4 e5d3 e4f6 g2g4 f8d8 c2c3 a8c8 h1g1 b7b6 g4g5 f6h5 g5h6 g7g6 f4e3 c8c7 d3e5 e7f6 a2a4 g8h7 b2b4 f6e5 d4e5 d8d5 d1d5 e6d5 b1c2 c7d7');
INSERT INTO evaluations VALUES(24489,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(24489,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(24489,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(24489,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(24489,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(24489,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24489,36,'cp 36','f8e7 c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e2 d8d5 f3e5 d5e4 e2e4 f6e4 h1e1 e4f6 g2g4 f8c8 e1g1 e7d6 f4e3 f6d5 e3c1 b7b5 f2f4 c6c5 g1e1 c5c4 g4g5 c4c3 g5h6 c3b2 c1b2 g7h6 f4f5 e6f5 e1g1 g8f8 d1f1');
INSERT INTO evaluations VALUES(24015,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(24015,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24015,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(24015,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(24015,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(24015,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(24015,36,'cp 40','c1b1 e8g8 g3e4 f6e4 d3e4 d7f6 e4e1 d8d5 f3e5 d5e4 f4e3 f6g4 e3c1 e4e1 h1e1 g4e5 d4e5 f8d8 c2c3 b7b5 d1d8 a8d8 b1c2 d8d5 b2b3 g8h7 f2f4 e7c5 g2g4 g7g6 e1e2 h7g7 e2d2 d5d2 c2d2 c5b6 d2d3 c6c5 d3e4 g6h5 g4h5 c5c4 c1a3');
INSERT INTO evaluations VALUES(31948,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31948,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31948,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(31948,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(31948,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(31948,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(31948,36,'cp -24','e7e6 d2d4 c7c5 c2c3 b8d7 f1d3 f5d3 d1d3 g8e7 e1g1 e7c6 a2a3 d8b6 b2b4 c5c4 d3c2 a7a5 c2b2 a5b4 a3b4 a8a1 b2a1 f8e7 a1a8 e7d8 b1a3 e8g8 a8c8 b6c7 c8c7 d8c7 f1e1 f8a8 a3b5 c7b8 g2g3 a8a2 b5d6');
INSERT INTO evaluations VALUES(32360,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(32360,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(32360,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(32360,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(32360,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(32360,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(32360,36,'cp -15','d2d4 c7c5 c2c3 b8c6 f1b5 a8c8 e1g1 f5g6 b5c6 c8c6 f1e1 c5c4 g2g3 f8e7 a2a4 h7h5 h2h4 d8b6 b1a3 a7a6 a3c2 b6b3 c2e3 b3d1 e1d1 c6c8 e3g2 e7d8 c1e3 g8e7 g2f4 g6f5');
INSERT INTO evaluations VALUES(32063,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(32063,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(32063,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32063,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(32063,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(32063,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(32063,36,'cp 49','b8c6 b1c3 c6a5 d1e2 g8f6 c1d2 c7c6 a2a3 a5c4 d3c4 f8e7 e1c1 f6d7 f3e1 f7f6 e1g2 g6f7 g2e3 e7f8 f2f3 h7h5 h3h4 d8b6 e3f5 b6a6 d2e3 a6c4 f5d6 f8d6 e2c4 f7c4 d1d6 h5g4 b2b3 c4f7 f3g4 e8c8 g4g5 d7f8 d6d8 c8d8 g5f6 g7f6 e3a7');
INSERT INTO evaluations VALUES(26897,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(26897,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(26897,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(26897,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(26897,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(26897,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(26897,36,'cp 41','b1c3 c6a5 d1e2 g8f6 c1d2 c7c6 a2a3 a5c4 d3c4 f6d7 e1c1 f8e7 f3e1 f7f6 e1g2 g6f7 g2e3 e7f8 e3f5 g7g6 f5g3 d8a5 c3d5 a5d8 d5e3 d8c7 h3h4 f8e7 h4h5 e8c8 h1h2 d7c5 d1h1 h8g8 h5g6 h7g6 h2h7 g8f8 h1h6 c5e6 c1b1 c8b8 h7f7 f8f7 h6g6');
INSERT INTO evaluations VALUES(26896,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(26896,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(26896,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(26896,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(26896,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(26896,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(26896,36,'cp 41','c6a5 d1e2 g8f6 c1d2 f6d7 e1c1 c7c6 a2a3 a5c4 d3c4 f8e7 d2e3 d8c7 c1b1 h7h5 f3d2 e8c8 f2f3 h5g4 h3g4 f7f6 d2b3 g6f7 a3a4 d7b6 b3d2 b6d7 b2b3 c7a5 b1b2 d7c5 d1a1 c5e6 h1h2 e6d4 e2f2 h8h2 f2h2 f7e6 a1h1 a5b4');
INSERT INTO evaluations VALUES(32269,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(32269,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(32269,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(32269,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(32269,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(32269,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(32269,36,'cp 28','e6d5 d1b3 b8c6 g1f3 g8f6 e2e4 d5e4 f3e5 f5e6 b3b7 c6d4 f1b5 d4b5 b7b5 e6d7 b5c4 d8e7 e5d7 e7d7 c1g5 d7e6 c4a4 e6d7 a4d7 f6d7 c3e4 f8b4 e1e2 f7f5 a2a3 b4e7 g5e7 e8e7 e4c3 c7c6 c3a4 h8e8 h1c1 d7e5 a4c5 a8d8 b2b4 g7g5 c1d1 e7f6 d1d8 e8d8 a1c1');
INSERT INTO evaluations VALUES(37374,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37374,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37374,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37374,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37374,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37374,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(37374,36,'cp 0','c7c5 g1e2 b8c6 e1g1 d7d6 d2d3 a7a6 a2a3 c8g4 a1b1 g4e2 c3e2 b7b5 b2b4 b5c4 d3c4 f6d7 f2f4 c5b4 a3b4 a6a5 b4b5 c6b4 c1e3 a8c8 b1c1 g7b2 c1b1');
INSERT INTO evaluations VALUES(37517,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(37517,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(37517,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(37517,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(37517,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(37517,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(37517,36,'cp 83','d1d2 g8f6 f1e2 b8d7 f2f3 h7h5 h2h3 b7b6 e1c1 c8b7 g2g4 b6b5 a2a3 d7c5 c1b1 a8c8 h1e1 d8c7 d4b3 e8g8 e3h6 h5h4 h6g7 g8g7 b3c5 d6c5 e4e5 c7e5 e2b5 e5c7 b5e2 e7e6 d2g5 c5c4 d1d4 c7g3 e1d1 g3h3 f3f4 h3g3 f4f5');
INSERT INTO evaluations VALUES(37518,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(37518,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(37518,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(37518,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(37518,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(37518,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(37518,36,'cp 57','g8f6 f2f3 h7h5 h2h3 b8d7 f1e2 b7b6 e1c1 c8b7 g2g4 b6b5 a2a3 d7b6 c1b1 d8c7 d4b3 f6d7 e3d4 g7d4 d2d4 e8c8 f3f4 c8b8 g4h5 g6h5 h1g1 d7c5 f4f5 d8c8 b3c5 d6c5 d4e3 c7e5 g1g5 b5b4 a3b4 c5b4 e3b6 b4c3 e2h5 c3b2 h5f7 h8h3');
INSERT INTO evaluations VALUES(37611,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(37611,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(37611,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(37611,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(37611,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(37611,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(37611,36,'cp -85','g8e7 e1g1 b8c6 c4b3 h7h6 g5f3 e8g8 b1c3 b7b6 f1e1 d7d6 c3e2 e6e5 e2c3 c6a5 b3d5 e7d5 c3d5 a5c6 c1d2 f7f5 h2h3 g6g5 f3h2 c8e6 c2c4 d8d7 d2c3 d7f7 f2f3 a8e8 d1d2 f5f4 b2b4 h6h5 b4c5 d6c5 a2a4 g8h7 a4a5');
INSERT INTO evaluations VALUES(38765,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(38765,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(38765,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(38765,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(38765,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(38765,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(38765,36,'cp 86','e6d5 f1d3 g8e7 c1a3 e8g8 d1h5 g7g6 h5f3 b8d7 g1e2 c7c5 d4c5 d7e5 f3f6 e7c6 f6d8 f8d8 a1d1 b7b6 e1g1 e5d3 c2d3 d5d4 c3c4 c8a6 e2g3 f7f5 d1b1 a8b8 c5b6 a7b6 a3c1 b6b5 c1f4 b8b7 c4b5 b7b5 b1c1 c6e5 a2a4');
INSERT INTO evaluations VALUES(38945,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(38945,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(38945,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(38945,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(38945,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(38945,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(38945,36,'cp 98','f1d3 g8e7 c1a3 e8g8 d1h5 g7g6 h5f3 b8d7 g1e2 c7c5 d4c5 d7e5 f3f4 e5d3 c2d3 c8f5 a1d1 d8a5 a3c1 a5c5 e1g1 f7f6 f1e1 f8f7 c1e3 c5c6 h2h4 f5d7 e2g3 a8c8 e3d4 b7b6 c3c4 e7f5 g3f5 d7f5 h4h5 d5c4 d3c4 c6c4 d1c1');
INSERT INTO evaluations VALUES(38775,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38775,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38775,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(38775,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(38775,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38775,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(38775,36,'cp 36','b4d6 g1f3 g8f6 b1d2 d6f4 e3f4 c7c5 d4c5 d8c7 g2g3 c7c5 f1d3 c5b6 d2b3 e8g8 e1g1 c8d7 d1e2 a7a5 b3d4 b8a6 f1e1 a6c5 d3c2 d7a4 c2a4 c5a4 a1b1 f6e4 g1g2 a4c5 h2h4 f8e8 b1d1 a8d8 f3e5 f7f6');
INSERT INTO evaluations VALUES(38661,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(38661,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(38661,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(38661,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(38661,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(38661,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(38661,36,'cp 41','b1d2 g8f6 g1f3 e8g8 f3e5 c7c5 f1d3 b8c6 d1f3 d8c7 e5c6 c7c6 e1g1 d6f4 f3f4 c6b6 d4c5 b6c5 e3e4 d5e4 d2e4 f6e4 d3e4 c5b6 b2b4 c8d7 f1d1 d7c6 d1d6 a8d8 e4c6 d8d6 f4d6 b6c6 d6c6 b7c6 a2a4 f8d8 b4b5 c6b5 a4b5');
INSERT INTO evaluations VALUES(38970,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(38970,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(38970,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(38970,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(38970,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(38970,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(38970,36,'cp 44','b1c3 g8f6 g1f3 e8g8 e1g1 c8g4 h2h3 g4h5 f1e1 b8d7 c1g5 c7c6 g2g4 h5g6 f3e5 a7a5 d1f3 d7b6 c4f1 a5a4 a2a3 b4d6 a1c1 d6c7 g5h4 f8e8 h4g3 c7d6 f1g2 h7h6 e5g6 e8e1 c1e1 f7g6 g3e5 b6c4 f3g3 d8d7 e5d6 c4d6');
INSERT INTO evaluations VALUES(39437,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(39437,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(39437,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(39437,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(39437,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(39437,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(39437,36,'cp 145','d1d4 g8f6 d4e5 d8e7 e5e7 e8e7 b1c3 d7d6 c1g5 b8d7 e1c1 h7h6 g5e3 a7a6 f2f3 b7b5 a2a3 c8b7 h2h4 h6h5 e3f2 d7e5 h1h3 h8e8 h3g3 g7g6 f2e3 e7f8 g3g5 e5d7 b2b3 a8d8 f1e2 d7e5');
INSERT INTO evaluations VALUES(39665,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(39665,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(39665,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(39665,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(39665,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(39665,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(39665,36,'cp 34','d1d3 b8d7 a2a4 h7h6 g5e3 b7b6 f1e2 c8b7 e1g1 e8g8 e2f3 a8c8 f1d1 c8c7 b3d2 d8c8 h2h3 f8d8 d3e2 e7f8 e2f1 a6a5 a1c1 d8e8 f1d3 d7c5 d3b5 b7c6 b5b6 c5a4 c3a4 c6a4 b2b3 a4d7 b6a5 c7c2 d2c4 c2c1 d1c1 c8b8 a5b6');
INSERT INTO evaluations VALUES(43751,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43751,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43751,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43751,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43751,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43751,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43751,36,'cp 0','f8e7 d1f3 d8b6 c1e3 b6b7 f3g3 b5b4 c3a4 b8d7 f1e1 e8g8 e3h6 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(39802,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39802,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39802,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39802,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39802,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39802,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39802,36,'cp 0','d1f3 d8b6 c1e3 b6b7 f3g3 b5b4 c3a4 b8d7 f2f3 e8g8 d4e2 d7e5 f1d1 c8d7 a4b6 a8b8 c2c3 b4c3 e2c3 d7c6 a1c1 a6a5 c3a4 h7h6 g1h1 b7a6 b3c2 c6a4 b6a4 f8c8 g3f2 e5g6 f2d2 d6d5 e4d5 f6d5 c2g6 f7g6 e3d4 c8c1 d1c1');
INSERT INTO evaluations VALUES(46012,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(46012,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(46012,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(46012,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(46012,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(46012,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(46012,36,'cp 69','d2d4 c5d4 f3d4 f8e7 f1e2 e8g8 e1g1 a7a6 d1d3 b7b6 d3g3 c8b7 c1h6 f6e8 h6e3 b8d7 f2f3 e8f6 a1c1 f8e8 f1d1 a8c8 b2b4 d8c7 g3f2 c7b8 a2a3 b7a8 c3a4 c8c7 d4b3 c7c6 b4b5 a6b5 c4b5 c6c1 d1c1 d7c5 a4c5 d6c5 a3a4 e7d6');
INSERT INTO evaluations VALUES(46014,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(46014,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(46014,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(46014,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(46014,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(46014,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(46014,36,'cp 57','c5d4 f3d4 f8e7 f1e2 a7a6 d1c2 b8c6 c1e3 e8g8 e1g1 c8d7 f1d1 c6e5 b2b3 a8c8 c2d2 d8c7 a1c1 e5g4 e3g5 h7h6 g5h4 g4e5 h4g3 f8d8 h2h3 e5c6 g1h1 c6d4 d2d4 c7a5 d4d2 a5b6 f2f4 d7c6 e2f3 f6d7 c1b1 b6c5 g3f2');
INSERT INTO evaluations VALUES(46045,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(46045,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(46045,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(46045,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(46045,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(46045,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(46045,36,'cp 66','f3d4 f8e7 f1e2 e8g8 e1g1 a7a6 d1d3 b7b6 d3g3 c8b7 c1h6 f6e8 h6e3 e8f6 f2f3 b8d7 a1c1 f8e8 b2b4 a8c8 f1d1 d8c7 a2a3 c7b8 d4b3 b7c6 b3d2 c6a8 g3f2 h7h6 e3d4 e7f8 g1h1 a8c6 d4e3 c8c7');
INSERT INTO evaluations VALUES(46044,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(46044,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(46044,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(46044,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(46044,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(46044,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(46044,36,'cp 80','f8e7 f1e2 e8g8 e1g1 b8c6 c1e3 c8d7 h2h3 a7a6 a2a3 d8b8 d4b3 f8c8 b3d2 e7d8 a1b1 c6e7 a3a4 d6d5 c4d5 e6d5 e3g5 d5d4 g5f6 g7f6 d2c4 b8f4 g2g3 f4g5 d1d4 d7h3 f1d1 e7c6 d4e3 h3e6 b2b3');
INSERT INTO evaluations VALUES(50869,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(50869,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(50869,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(50869,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(50869,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(50869,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(50869,36,'cp 47','c7c5 c3d5 b7b5 c4e2 g8f6 d5f6 d8f6 a2a4 b5b4 c2c3 f8e7 c3b4 e8g8 b4c5 d6c5 e2c4 b8c6 d2d3 c8e6 c4e6 f6e6 c1d2 f8d8 d1e2 a8b8 d2c3 c6d4 c3d4 e5d4 a1c1 b8b4 g2g3 a6a5 c1c2 b4a4 f3d2 d8a8 b2b3 a4b4 f1a1 a5a4 b3a4');
INSERT INTO evaluations VALUES(52924,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(52924,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(52924,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(52924,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(52924,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(52924,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(52924,36,'cp 24','d3e4 d8d1 c3d1 g8f6 f2f3 a7a5 a2a4 f6d7 g1h3 e7e5 d1e3 f7f6 h3f2 e8d8 c1d2 d8c7 b2b3 h7h5 h2h4 d7c5 e1c1 b8d7 f1c4 d7b6 c4f7 c8e6 f7e6 c5e6 f2d3 b6d7 e3c4 b7b6 f3f4 e6c5 d2e3 c5d3 d1d3 f8c5 e3c5 d7c5 d3g3');
INSERT INTO evaluations VALUES(52918,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52918,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52918,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(52918,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(52918,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52918,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52918,36,'cp 19','d8d1 c3d1 g8f6 f2f3 f6d7 h2h4 h7h5 g1h3 a7a5 a2a4 e7e5 d1e3 d7b6 c1d2 b8d7 h3f2 d7c5 e3c4 b6c4 f1c4 f7f6 d2e3 c8e6 c4e6 c5e6 f2d3 f8e7 b2b3 e8f7 e1e2 e7b4 a1d1 b4e7 e3b6 e7d8 b6f2 d8e7 g2g3');
INSERT INTO evaluations VALUES(53219,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53219,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53219,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53219,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(53219,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53219,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(53219,36,'cp -20','e2g3 a7a6 f1c4 g8f6 g1f3 b7b5 c4e2 b8c6 a2a4 b5b4 e1g1 h7h5 e4e5 h5h4 e5f6 h4g3 f2g3 e7f6 f1e1 f8e7 e2d3 a8a7 d1e2 g7g6 e2f2 e8g8 b2b3 e7d6 d3e4 c6e5 f3e5 f6e5 d2d3 f7f5 e4d5 g8g7 c1d2 f8e8');
INSERT INTO evaluations VALUES(53218,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(53218,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(53218,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53218,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(53218,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(53218,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53218,36,'cp -12','a7a6 f1c4 g8f6 d2d3 b7b5 c4b3 e7e6 d1e2 c8b7 g1h3 b8d7 e1g1 d8c7 c1f4 c7c6 a2a4 f8e7 h3g5 e8g8 g5f3 e7d8 f3d2 d7b6 a4a5 b6d7 g3h5 h7h6 h5f6 d7f6 f4g3 d8c7 g3c7 c6c7 f2f4 a8e8 h2h3 g8h8 f4f5 e6f5 f1f5');
INSERT INTO evaluations VALUES(57066,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(57066,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(57066,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(57066,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(57066,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(57066,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(57066,36,'cp -20','d7d5 d2d4 g8f6 g1f3 f8d6 b2b3 e8g8 f1e2 c7c5 e1g1 b7b6 c3c4 d5c4 b3c4 c5d4 e3d4 c8b7 f1e1 b8d7 c1a3 f6e4 d1b3 d7f6 b1c3 d6a3 b3a3 e4c3 a3c3 a8c8 a1c1 c8c7 c3a1 d8d6 h2h3 f8c8 e2f1');
INSERT INTO evaluations VALUES(55942,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(55942,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(55942,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(55942,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(55942,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(55942,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55942,36,'cp -10','d2d4 g8f6 g1f3 f8d6 f1e2 e8g8 e1g1 b7b6 b2b3 c8b7 c3c4 d5c4 e2c4 b8d7 c1b2 d8e7 b1d2 c7c5 d4c5 d7c5 d1e2 f8d8 a1c1 a7a6 a2a4 d8d7 f1d1 a8d8 d2f1 d6b8 h2h3 d7d1 c1d1 d8d1 e2d1');
INSERT INTO evaluations VALUES(13365,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(13365,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(13365,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(13365,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(13365,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(13365,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(13365,36,'cp 59','f3d4 g7g6 f1e2 c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 a7a6 c1e3 c8d7 a2a4 a6a5 d3c2 d7c6 b2b3 f6d7 a1d1 b7b6 e3g5 d7c5 c3d5 f7f6 g5f4 g8h8 f2f3 f6f5 f4g5 c6d5 e4d5 g7f6 g5e3 d8c7 g2g3 a8e8 g1h1');
INSERT INTO evaluations VALUES(17470,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(17470,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(17470,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(17470,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(17470,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(17470,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(17470,36,'cp 159','c1e3 e7e5 d4f3 f8b4 f1c4 g8f6 e1g1 b4c3 b2c3 d7d6 d1d3 d8e7 e3h6 c6a5 c4b3 c8d7 f1e1 a5b3 a2b3 f6h5 a1d1 f7f6 d3d6 e7d6 d1d6 g6g5 h2h3 e8e7 d6b6 d7c6 g2g4 h5f4 h3h4 g5h4 f3h4 h8g8 f2f3 f4e6 h6e3 e7f7 e1d1 a8c8');
INSERT INTO evaluations VALUES(17469,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(17469,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(17469,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(17469,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(17469,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(17469,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(17469,36,'cp 142','g8f6 d4c6 d7c6 d1d8 e8d8 e1c1 f6d7 e4e5 b7b5 c3e4 d8c7 e3f4 c7b6 a2a4 b5a4 d1d4 f8g7 d4b4 b6c7 e5e6 g7e5 e4g5 e5f4 b4f4 f7e6 g5e6 c7d6 e6g5 d6c7 f4a4 e7e5 a4a5 c7b6 a5a3 a6a5 f1c4 a5a4 h1e1 h8e8 a3e3 a8a5 f2f4 d7f6 e3e5 a5e5 f4e5 f6g4');
INSERT INTO evaluations VALUES(18128,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(18128,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(18128,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(18128,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(18128,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(18128,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(18128,36,'cp -37','c8b7 d2d4 b5b4 c3b1 c5d4 f3d4 g8f6 d4c6 b7c6 e4e5 f6d5 e2f3 d8c7 c2c4 b4c3 b1c3 d5c3 b2c3 c7e5 f3c6 d7c6 d1a4 e5c7 f1d1 f8d6 c1a3 d6a3 a4a3 c7e7 d1d6 e8g8 a1d1 a8d8 a3c5 d8d6 d1d6 f8d8 d6d8 e7d8 h2h3 d8c7 h3h4 g7g6 a2a4 h7h6 h4h5 g6h5 c5h5');
INSERT INTO evaluations VALUES(25594,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(25594,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(25594,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(25594,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(25594,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(25594,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(25594,36,'cp -35','d2d4 b5b4 c3b1 c5d4 f3d4 g8f6 d4c6 b7c6 e4e5 f6d5 e2f3 d8c7 c2c4 b4c3 b1c3 d5c3 b2c3 c7e5 f3c6 d7c6 d1a4 e5c7 f1d1 f8d6 c1a3 d6a3 a4a3 c7e7 d1d6 e8g8 a1d1 a8d8 g2g3 d8d6 d1d6 h7h5 h2h4 f8b8 a3c5 b8b2 a2a4 b2a2 d6c6 e7c5 c6c5 a2a4 c5h5 a4c4 g1g2');
INSERT INTO evaluations VALUES(19331,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19331,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19331,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19331,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19331,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19331,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(19331,36,'cp -8','f8g7 b1c3 g7c3 b2c3 g8f6 f1d3 d7d5 e4d5 d8d5 h2h3 c6e5 d3e2 d5d1 e2d1 e5c4 e1g1 e8g8 f1e1 f8e8 f3e5 c8e6 e5c4 e6c4 d1f3 c4d5 f3e2 a8c8 e2b5 d5c6 a2a4 a7a6 b5d3 f6d5 c1d2 e7e5 c3c4 d5f4 d3f1 f7f6 a1b1 g8f7 a4a5 f4e6');
INSERT INTO evaluations VALUES(5419,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(5419,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(5419,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(5419,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(5419,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5419,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(5419,36,'cp -4','b1c3 g7c3 b2c3 g8f6 f1d3 d7d5 e4d5 d8d5 h2h3 c6e5 d3e2 d5d1 e2d1 e5c4 e1g1 e8g8 f1e1 f8e8 f3e5 c4e5 e1e5 b7b6 a2a4 c8e6 a4a5 a8c8 a5b6 a7b6 a1a3 e8d8 d1f3 e6d5 f3e2 d5c4 e2f3 d8d7 c1g5 f6d5 f3d5 d7d5 e5d5 c4d5');
INSERT INTO evaluations VALUES(20943,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(20943,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(20943,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20943,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(20943,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(20943,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(20943,36,'cp 0','c8d7 b5c6 b7c6 g1f3 e7e6 b2b3 c6c5 e1g1 g8e7 c1a3 e7g6 d2d4 f7f6 c2c4 c5d4 a3f8 h8f8 d1d4 f6e5 f3e5 c7c5 d4b2 g6e5 b2e5 d8f6 e5f6 f8f6 f1e1 d5d4 b1d2 e8e7 d2e4 f6f5 e4g3 f5f7 g3e4');
INSERT INTO evaluations VALUES(27284,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27284,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27284,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27284,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27284,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27284,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27284,36,'cp 0','b5c6 b7c6 b2b3 c6c5 g1f3 e7e6 e1g1 g8e7 c1a3 e7g6 d2d4 f7f6 c2c4 c5d4 a3f8 h8f8 d1d4 f6e5 f3e5 c7c5 d4b2 g6e5 b2e5 d8f6 e5f6 f8f6 b1d2 e8e7 f1e1 d5d4 d2e4 f6f5 e4g3 f5f4 g3e4');
INSERT INTO evaluations VALUES(22605,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(22605,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22605,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(22605,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(22605,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(22605,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(22605,36,'cp 100','e2e4 d7d5 e4e5 f7f6 f1b5 c8d7 g1f3 d8e7 b5c6 d7c6 a2a4 e7f7 c3e2 g8e7 c2c3 c6d7 e2f4 e7c6 f4d3 f6e5 d3e5 c6e5 f3e5 f7f5 e1g1 f8d6 f1e1 h8f8 d1d2 a7a6 b2b3 e8c8 b3b4 d7e8 c1a3 d6e5 e1e5');
INSERT INTO evaluations VALUES(30260,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(30260,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(30260,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(30260,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(30260,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(30260,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30260,36,'cp 28','e1g1 f8e7 f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 d5d7 c1f4 e7d6 f4d6 c7d6 d4d5 c6a5 f3d4 f5d3 d1d3 f8e8 d4f5 e8e1 a1e1 a8e8 e1e8 d7e8 g2g3 e8e1 g1g2 e1e5 d3b5 a5c4 f5h6 g7h6 b5c4 g8g7 c4d3 h6h5 c3c4 b7b6 g2h3 e5f6 d3e3 h7h6 h3g2 h5h4 g3g4');
INSERT INTO evaluations VALUES(30259,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(30259,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(30259,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30259,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(30259,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(30259,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(30259,36,'cp 25','f8e7 f1e1 b8c6 b1d2 e4d2 d1d2 d8d7 d3b5 e8g8 b5c6 b7c6 f3e5 d7d6 d2c3 c6c5 d4c5 d6c5 e5c6 e7d6 c3c5 d6c5 c1e3 c5d6 e3f4 d6f4 c6e7 g8h8 e7f5 f8e8 f5d4 f4d6 e1e8 a8e8 g1f1 e8b8 b2b3 a7a6 a1e1 g7g6 g2g3 h8g7 a2a4 d6b4 d4c6 b8b6 c6b4');
INSERT INTO evaluations VALUES(39109,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(39109,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(39109,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(39109,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(39109,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(39109,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(39109,36,'cp 60','g8f6 d2d4 d6e7 d4e5 f6e4 f1c4 e8g8 e1g1 d7d6 f1e1 c8f5 b1d2 d6d5 f3d4 f5g6 d2e4 d5c4 d1g4 d8d7 h2h3 c7c5 d4f3 d7g4 h3g4 b8d7 c1f4 d7b6 e4d6 e7d6 e5d6 f7f6 f3d2 g6d3 g4g5 b6d5 f4g3 f6g5 d6d7 a8d8');
INSERT INTO evaluations VALUES(42393,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(42393,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(42393,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(42393,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(42393,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(42393,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(42393,36,'cp 88','d2d4 d6e7 f1d3 d7d5 f3e5 d5e4 d3c2 b8d7 d1e2 e8g8 e1g1 e7d6 e5c4 d7b6 c4d6 d8d6 c2e4 f8e8 f2f3 c8f5 b1d2 f6e4 f3e4 f5e4 d2e4 d6g6 e4f6 g6f6 e2e8 a8e8 f1f6 g7f6 g1f2 b6a4 c3c4 c7c6 b2b3 a4c3 c1e3 f6f5 a2a4 f7f6 a1c1 c3e4 f2e2 g8f7 e2d3 e8g8 g2g3 b7b6 d3c2 g8g4 e3f4 g4g8 b3b4 g8g4 a4a5 b6a5');
INSERT INTO evaluations VALUES(40010,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(40010,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(40010,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(40010,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(40010,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(40010,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(40010,36,'cp 37','d4d5 e8g8 b1c3 e7e6 f1d3 d7d6 h2h3 e6d5 c4d5 b8a6 e1g1 a6c7 e3e4 b7b5 f1e1 f8e8 a2a3 a7a5 c1f4 c5c4 d3c2 c8a6 d1c1 b5b4 a3b4 a5b4 c3d1 d8e7 a1a4 b4b3 c2b1 f6d7 f4e3 d7e5 f3e5 g7e5 f2f4 e5g7 e3f2 c7b5 a4c4');
INSERT INTO evaluations VALUES(40129,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(40129,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40129,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40129,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40129,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(40129,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40129,36,'cp 13','a2a4 d6d5 c4a2 c8e6 e1g1 d5e4 f3e5 e6a2 a1a2 e4d3 e5d3 a7a5 a2a1 e8g8 d1f3 f8e8 h2h3 h7h6 c1f4 d8d4 f4e5 d4b6 c3e2 b8d7 e5d4 c6c5 d4c3 c5c4 d3e5 d7e5 c3e5 a8d8 f1d1 e7b4 e5d4 b6c6 f3c6');
INSERT INTO evaluations VALUES(43364,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43364,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(43364,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(43364,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(43364,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(43364,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(43364,36,'cp 36','g1f3 e7e6 b2b3 f8b4 c1d2 b8d7 f1d3 b4d6 e1g1 e8g8 f1e1 f8e8 d1c2 h7h6 h2h3 e6e5 c4d5 f6d5 c3d5 c6d5 d4e5 d7e5 f3e5 e8e5 a1c1 c8d7 d2c3 a8c8 c2d1 c8c3 c1c3 d6b4 d1a1 e5e8 g1h1 b4c3 a1c3 d7c6 e1d1');
INSERT INTO evaluations VALUES(44001,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(44001,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44001,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(44001,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(44001,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(44001,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(44001,36,'cp 51','g1f3 b8c6 f1d3 c5d4 c3d4 d8b6 e1g1 a7a5 f1e1 f8e7 h2h3 a5a4 a2a3 c6d4 f3d4 b6d4 d2f3 d4b6 c1e3 d7c5 a1c1 c8d7 d3b1 b6a5 f3d4 e8g8 d1g4 f8c8 h3h4 c5b3 c1d1 b3d4 e3d4 e7c5 h4h5 c5d4 g4d4 c8c4 d4d3 g7g6 d3g3 a5c5');
INSERT INTO evaluations VALUES(51589,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(51589,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(51589,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(51589,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(51589,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(51589,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(51589,36,'cp 132','d4d5 d7d6 e2e4 g7g6 f1e2 f8g7 b1c3 g7c3 b2c3 g8f6 d1c2 e7e6 d5e6 c8e6 c1f4 e8g8 a1d1 f6e8 g1f3 d8e7 f4h6 e8g7 c2d3 b8c6 d3d6 e7d6 d1d6 c6a5 f3d2 a8d8 d6d8 f8d8 h6g5 d8d7 f2f3 g7e8 e4e5 a5c6 g5f4 g6g5 f4g3 h7h5 h2h4 g5g4 f3g4 e6g4 e2g4 h5g4 h4h5 c6a5 h5h6');
INSERT INTO evaluations VALUES(51600,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(51600,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(51600,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(51600,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(51600,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(51600,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(51600,36,'cp 126','g7g6 e2e4 d7d6 f1e2 e7e6 b1c3 f8g7 c3b5 e6e5 h2h4 h7h5 d1a4 e8f8 g1f3 c8g4 f3g5 g4e2 e1e2 g8f6 g2g3 a7a6 b5c3 g7h6 e2f1 d8e8 a4c2 b6b5 f1g2 b5c4 c2e2 b8d7 e2c4 f8g7 c4e2 a6a5 b2b3 e8e7 f2f3');
INSERT INTO evaluations VALUES(57198,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(57198,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(57198,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(57198,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(57198,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(57198,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(57198,36,'cp -70','d7d5 d2d3 d5d4 c3e4 c7c5 f1g2 b8c6 c2c3 f8e7 g1f3 g8f6 e4f6 e7f6 c3d4 c5d4 f3d2 e8g8 e1g1 c8e6 a2a3 a8c8 b2b4 b7b6 c1b2 e6d5 g2e4 d8d7 e4d5 d7d5 d1b3 c6e7 a1c1 c8c1 f1c1 f6g5 f2f4 g5h6 b3d5 e7d5');
INSERT INTO evaluations VALUES(56438,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(56438,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(56438,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(56438,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(56438,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(56438,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(56438,36,'cp -55','d2d3 d5d4 c3b1 f8d6 g1f3 c7c5 f1g2 b8c6 e1g1 g8f6 c2c3 e8g8 c3d4 e5d4 e2e3 d4e3 c1e3 h7h6 b1c3 f6g4 e3d2 c8f5 c3e4 d6e7 d1e2 g4f6 d2c3 f6d5 f1d1 f8e8 f3e5 d5c3 e4c3 e7f6 e5c6 e8e2 c6d8 f6c3 b2c3 a8d8');
INSERT INTO evaluations VALUES(57677,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(57677,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(57677,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(57677,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(57677,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(57677,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(57677,36,'cp -45','d2d4 d7d5 g4g5 h7h6 h2h4 h6g5 c1g5 f7f6 g5e3 c8f5 c2c4 e7e6 b1c3 c7c6 e3f4 b8d7 f1g2 f8b4 d1b3 d8a5 g1f3 d5c4 b3c4 d7b6 c4b3 b6d5 f4d2 g8e7 e1g1 a5b6 c3d5 e7d5 d2b4 b6b4 b3b4 d5b4');
INSERT INTO evaluations VALUES(1585,30,'cp 498',NULL);
INSERT INTO evaluations VALUES(1585,31,'cp 512',NULL);
INSERT INTO evaluations VALUES(1585,32,'cp 521',NULL);
INSERT INTO evaluations VALUES(1585,33,'cp 510',NULL);
INSERT INTO evaluations VALUES(1585,34,'cp 501',NULL);
INSERT INTO evaluations VALUES(1585,35,'cp 507',NULL);
INSERT INTO evaluations VALUES(1585,36,'cp 518','d2d4 c6b4 d4d5 e6f5 e1g1 g8f6 c3b5 b4d5 c2c4 a7a6 b5d4 f5h5 c4d5 h5d5 c1f4 c7c5 d4c2 d5d1 f1d1 g7g6 c2e3 f8g7 e3c4 e8g8 c4b6 c8f5 b6a8 f8a8 f4e5 a8c8 a2a4 f5c2 d1e1 e7e6 a4a5 c2e4 f3d2 e4c6 d2c4 c8d8 e1d1 f6d5 e5g7 g8g7');
INSERT INTO evaluations VALUES(1578,30,'cp 484',NULL);
INSERT INTO evaluations VALUES(1578,31,'cp 489',NULL);
INSERT INTO evaluations VALUES(1578,32,'cp 483',NULL);
INSERT INTO evaluations VALUES(1578,33,'cp 503',NULL);
INSERT INTO evaluations VALUES(1578,34,'cp 486',NULL);
INSERT INTO evaluations VALUES(1578,35,'cp 505',NULL);
INSERT INTO evaluations VALUES(1578,36,'cp 499','c6b4 d4d5 e6f5 e1g1 g8f6 c3b5 b4d5 c2c4 a7a6 b5d4 f5d7 c4d5 d7d5 c1f4 c7c5 d4b3 d5d1 f1d1 e7e6 f3e5 f6d5 f4g3 b7b5 e2f3 f7f6 e5c6 c8d7 b3a5 d7c6 a5c6 a8c8 c6b8 a6a5 f3g4 f6f5 g4f5 e6f5 d1d5 f8e7 a1e1 e8f7 d5f5 e7f6 b8d7 h8e8 e1e8 c8e8 g1f1 e8c8 d7c5 f7g6');
INSERT INTO evaluations VALUES(5021,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(5021,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(5021,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(5021,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(5021,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(5021,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(5021,36,'cp 0','e1g1 e7e5 a1b1 g8e7 a2a3 a7a5 d2d3 e8g8 c1e3 f7f5 e3c5 d7d6 c5e3 f5f4 e3c1 h7h6 h2h3 g6g5 f3h2 d8e8 b2b4 a5b4 a3b4 h6h5 f2f3 c8d7 c1d2 g8h7 g1h1 e7g6 d2e1 g7f6 e1f2 f6d8 b1a1 a8b8 a1b1');
INSERT INTO evaluations VALUES(5020,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(5020,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(5020,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(5020,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(5020,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(5020,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(5020,36,'cp -1','e7e5 a1b1 a7a5 a2a3 g8e7 d2d3 a5a4 b2b4 a4b3 c2b3 e8g8 h2h3 h7h6 b3b4 c5b4 a3b4 d7d5 f1e1 f8e8 d1c2 c8a6 e1d1 g6g5 f3h2 d5d4 c3a4 a6b5 a4c5 g7f8 c1d2 e7g6 b1a1 f8c5 a1a8 d8a8 b4c5 a8d8 d1a1 g8h7');
INSERT INTO evaluations VALUES(6913,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(6913,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(6913,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(6913,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(6913,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(6913,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(6913,36,'cp 199','d5c5 d7d6 c5c4 c3c2 b1c3 c8g4 f3e1 d8d7 e1c2 g8f6 c4b5 b7b6 b5d3 a7a5 f2f3 g4e6 b2b3 e8d8 c1e3 d8c8 f1d1 d6d5 e4d5 h8d8 d5d6 d7d6 d3d6 d8d6 d1d6 c7d6 a1d1 c6b4 d1d6 b4c2 d6e6 c2e3 e6e3');
INSERT INTO evaluations VALUES(6910,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(6910,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(6910,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(6910,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(6910,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(6910,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(6910,36,'cp 197','d7d6 c5c4 c3c2 b1c3 g8f6 f1e1 e8f8 b2b3 d8e7 c1b2 c8e6 c4d3 f8g8 d3c2 e7f7 c3e2 a8f8 f3d4 c6b4 c2b1 e6d7 f2f3 c7c5 a2a3 b4d5 d4c2 d5e7 b3b4 g8h7 b4c5 d6c5 b1c1 d7c6 e2f4 f6h5 f4h5 f7h5');
INSERT INTO evaluations VALUES(7010,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(7010,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(7010,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(7010,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(7010,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(7010,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(7010,36,'cp 90','a7a6 g2g4 h7h6 g4g5 b7b5 g5h6 g8h6 h1g1 e7g6 c1e3 d6b4 h2h4 h6g8 a2a4 b5a4 d5d6 c8b7 a1a4 b4c3 b2c3 b7c6 g1g6 f7g6 a4c4 c7d6 d1d6 d8f6 d6e5 f6e5 f3e5 h8h4 e5c6 d7c6 c4c6 h4h1 e1e2 a8d8 c6g6 g8f6 f1g2');
INSERT INTO evaluations VALUES(7675,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(7675,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(7675,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7675,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7675,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7675,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7675,36,'cp 0','e4e5 f6d5 g5e7 d8e7 d1d4 f7f6 e1c1 f6e5 d4e5 e7f6 e5f6 d5f6 f2f4 e8e7 g2g3 d7d5 f1h3 h7h6 h1e1 e7d6 c3e2 g7g5 e2d4 h8e8 f4g5 h6g5 e1f1 e8f8 f1f2 c6c5 d4b5 d6e7 b5c7 a8b8 d1e1 g5g4 h3g4 f6g4 c7d5 e7e8 d5c7 e8e7');
INSERT INTO evaluations VALUES(7674,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(7674,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7674,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7674,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7674,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7674,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7674,36,'cp 0','f6d5 g5e7 d8e7 d1d4 f7f6 e1c1 f6e5 d4e5 e7f6 e5f6 d5f6 f2f4 e8e7 g2g3 d7d5 f1h3 e7d6 h1e1 h7h6 c3e2 h8e8 e2d4 g7g5 f4g5 h6g5 e1f1 e8f8 f1f2 c6c5 d4b5 d6e7 b5c7 a8b8 d1e1 g5g4 h3g4 f6g4 c7d5 e7e8 d5c7 e8e7');
INSERT INTO evaluations VALUES(7673,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7673,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7673,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7673,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7673,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7673,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7673,36,'cp 0','g5e7 d8e7 d1d4 f7f6 e1c1 f6e5 d4e5 e7f6 e5f6 d5f6 f2f4 d7d5 g2g3 e8e7 f1h3 e7d6 h1e1 h7h6 c3e2 h8e8 e2d4 g7g5 f4g5 h6g5 e1f1 e8f8 f1f2 c6c5 d4b5 d6e7 b5c7 a8b8 d1e1 g5g4 h3g4 f6g4 c7d5 e7e8 d5c7 e8e7');
INSERT INTO evaluations VALUES(7672,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7672,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7672,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7672,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7672,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7672,36,'cp 0','d8e7 d1d4 f7f6 e1c1 f6e5 d4e5 e7f6 e5f6 d5f6 f2f4 d7d5 g2g3 e8e7 f1h3 e7d6 h1e1 h7h6 c3e2 h8e8 e2d4 g7g5 f4g5 h6g5 e1f1 e8f8 f1f2 c6c5 d4b5 d6e7 b5c7 a8b8 d1e1 g5g4 h3g4 f6g4 c7d5 e7e8 d5c7 e8e7');
INSERT INTO evaluations VALUES(9112,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(9112,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9112,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(9112,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9112,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9112,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9112,36,'cp 5','f1d3 e8g8 c1d2 f8e8 a2a3 b4d6 h2h3 c6e7 d1c2 c7c6 e1g1 g7g6 f3h4 c8d7 f2f4 d8c7 a1c1 c6c5 d4c5 c7c5 c2b3 c5b6 b3b6 a7b6 h4f3 e7c6 c3b5 d6c5 f3d4 c6d4 b5d4 a8a4 f1d1 c5d4 e3d4 a4d4');
INSERT INTO evaluations VALUES(18456,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(18456,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(18456,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(18456,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(18456,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(18456,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(18456,36,'cp -46','e7e5 c3e2 g7g6 d2d3 f8g7 a2a3 a7a5 e1g1 g8f6 h2h3 e8g8 a1b1 h7h6 c2c4 f8e8 c1d2 d6d5 e2g3 f6d7 b2b4 a5b4 a3b4 d5d4 d1c2 c5b4 d2b4 c6c5 b4d2 d7f8 b1a1 c8d7 f1b1 d8c7 a1a2 a8a2 c2a2 f8e6 b1a1 e8b8');
INSERT INTO evaluations VALUES(18499,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(18499,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(18499,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(18499,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(18499,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(18499,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(18499,36,'cp -34','d2d3 g8e7 f3h4 g7g6 e1g1 e7g8 h4f3 g8f6 a2a3 a7a5 a1b1 f8g7 b2b4 a5b4 a3b4 c5b4 b1b4 e8g8 b4b1 c8e6 h2h3 h7h6 c1e3 g8h7 d1d2 d8a5 d3d4 e5d4 e3d4 c6c5 d4f6 g7f6 c3d5 f6g7 c2c3 e6d5 d2d5 a5c3 d5d6 f8d8 d6g3 c3d3 b1e1');
INSERT INTO evaluations VALUES(18500,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(18500,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(18500,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(18500,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(18500,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(18500,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(18500,36,'cp -49','g8e7 f3h4 g7g6 e1g1 h7h6 c3e2 f8g7 f2f4 e5f4 e2f4 a8b8 h4f3 e8g8 d1e2 d8e8 h2h3 f7f5 e4f5 e7f5 e2e8 f8e8 f4g6 c5c4 g6f4 c4d3 c2d3 g8h7 a1b1 c6c5 g1h2 c8b7 b2b3 e8e7 f4h5 g7h8 f1f2 a7a5');
INSERT INTO evaluations VALUES(19461,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19461,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19461,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(19461,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19461,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19461,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19461,36,'cp -1','g8f6 e4e5 f6d5 f1b5 d8c7 b5c6 d7c6 d2d3 c8f5 a2a3 h7h6 c1d2 e7e6 b1c3 d5c3 d2c3 b7b5 b2b3 a7a5 a3a4 c5c4 d3c4 b5c4 b3c4 a8d8 d1e2 f8c5 e1g1 e8g8 f1d1 d8d1 a1d1 f8d8 d1d8 c7d8 e2d2 d8d2 f3d2 f5c2 c3a5 c2a4 a5c3 c5e7 d2e4 c6c5');
INSERT INTO evaluations VALUES(21364,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21364,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21364,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21364,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21364,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21364,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21364,36,'cp 0','e6d5 a2a3 c8f5 g1f3 g8f6 d1b3 a8b8 f3h4 f5e6 f1d3 f6e4 h4f3 f8d6 b3c2 e8g8 b1c3 f7f5 c3b5 d6e7 b5c3');
INSERT INTO evaluations VALUES(23300,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(23300,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23300,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(23300,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(23300,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23300,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23300,36,'cp 1','g8f6 e4f5 d7d5 f3h4 f8e7 g2g3 e8g8 f1g2 c6d4 e1g1 c7c6 h4f3 c8f5 f3d4 e5d4 c3e2 e7c5 c1f4 a7a5 a2a3 a5a4 b2b4 a4b3 c2b3 a8a3 a1a3 c5a3 e2d4 f5g4 d1a1 a3b4 h2h3 g4d7 d4c2 b4c5 b3b4 c5b6 f4e3 h7h6 e3b6 d8b6 f1b1 d7f5');
INSERT INTO evaluations VALUES(17112,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17112,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17112,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17112,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17112,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17112,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17112,36,'cp 4','e4f5 d7d5 f3h4 f8e7 g2g3 e8g8 f1g2 c6d4 e1g1 c7c6 h4f3 c8f5 f3d4 e5d4 c3e2 e7c5 h2h3 d8e7 b2b4 c5b4 e2d4 f5d7 c1b2 a8e8 a2a3 b4d6 d4f3 e7f7 f1e1 e8e1 d1e1 f7h5 b2e5 f8e8 g3g4 f6g4 h3g4 d7g4 e5d6 e8e1 f3e1 h5g5 d6b8 a7a5 a1b1 g5d2 b8e5 h7h5 e1f3 g4f3 g2f3 h5h4');
INSERT INTO evaluations VALUES(23799,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(23799,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23799,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23799,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23799,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23799,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(23799,36,'cp 0','g8f6 h2h3 e8g8 a2a4 a7a6 b5d3 a8c8 d1d2 c6d4 e3d4 d8a5 d2e3 d7e6 c3e2 e6c4 f1d1 e7e5 d4c3 a5c7 c3b4 f8d8 d3c4 c7c4 b4a5 d8d7 e2c3 d6d5 e4d5 f6d5 e3d3 d5b6 d3e3 b6d5');
INSERT INTO evaluations VALUES(25000,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(25000,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(25000,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(25000,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(25000,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(25000,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(25000,36,'cp 46','e7e6 f1c4 f8b4 d1c2 e8g8 a1d1 c6e7 c4e2 b4c3 b2c3 c7c5 h2h3 g4h5 d4c5 d8c7 e1g1 h5g6 d1d4 c7c5 e2d3 a8c8 c2b1 c5c7 d4c4 c7b8 c4c8 f8c8 f3d4 h7h6 f2f3 b7b6 f1c1 e6e5 d4b5 f6h5 d3f1 e7d5');
INSERT INTO evaluations VALUES(25240,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(25240,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(25240,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(25240,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(25240,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(25240,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(25240,36,'cp 50','f1c4 f8b4 d1d3 g4h5 e1g1 e8g8 a2a3 b4c3 b2c3 h5f3 g2f3 c6e7 a1b1 b7b6 e3g5 g8h8 f3f4 f6g8 f1e1 d8d7 c4b3 a8e8 d3h3 e7g6 f4f5 e6f5 h3f5 d7f5 e4f5 f7f6 e1e8 f8e8 f5g6 f6g5 b3f7 e8f8 b1e1 h7g6 f7g6 f8f3 e1e3 f3e3 f2e3 g8e7 g6e4 c7c6 c3c4');
INSERT INTO evaluations VALUES(25239,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(25239,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(25239,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(25239,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(25239,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25239,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(25239,36,'cp 41','f8b4 d1c2 e8g8 a1d1 c6e7 c4e2 b4c3 b2c3 c7c5 h2h3 g4h5 d4c5 d8c7 e1g1 h5g6 e2d3 f6d7 d1b1 f8d8 f1d1 g6h5 d3e2 h5f3 e2f3 d7c5 e3c5 c7c5 b1b7 e7g6 d1d8 a8d8 g2g3 h7h6 h3h4 g6e5 f3e2 d8c8 g1g2 g7g6 h4h5 c5c3 c2c3 c8c3 h5g6 e5g6 b7a7 c3c2');
INSERT INTO evaluations VALUES(24096,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(24096,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(24096,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(24096,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(24096,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(24096,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(24096,36,'cp 36','d1c2 e8g8 a1d1 c6e7 c4e2 b4c3 b2c3 c7c5 h2h3 g4h5 d4c5 d8c7 e1g1 h5g6 d1d4 c7c5 e2d3 a8c8 c2b1 c5c7 d4c4 c7b8 c4c8 f8c8 f3d4 h7h6 f2f3 e6e5 d4e2 f6d7 c3c4 b7b6 e2c3 d7c5 d3c2 e7c6 f1d1 f7f6 e3c5 b6c5');
INSERT INTO evaluations VALUES(29616,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(29616,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(29616,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(29616,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(29616,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(29616,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29616,36,'cp 28','e2e3 b8d7 f1d3 f5d3 d1d3 c7c6 e1g1 e7e6 c2c4 f8e7 b1c3 e8g8 a1c1 h7h6 g5f6 d7f6 f1d1 a7a5 g2g3 e7d6 a2a3 f8e8 e3e4 d5e4 c3e4 f6e4 d3e4 d6e7 f3e5 d8c7 c4c5 a8d8 e5c4 d8d5 b2b4');
INSERT INTO evaluations VALUES(29615,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(29615,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(29615,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(29615,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(29615,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(29615,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(29615,36,'cp 14','b8d7 f1d3 f5d3 d1d3 e7e6 c2c4 c7c6 e1g1 f8e7 b1c3 e8g8 f1d1 h7h6 g5f6 d7f6 a1c1 d8c7 c1c2 e7b4 d1c1 c7e7 a2a3 b4c3 c2c3 f6e4 c3c2 a7a5 a3a4 f7f6 f3d2 e4d6 b2b3 a8c8 c4c5 d6f5 d3e2 f8d8');
INSERT INTO evaluations VALUES(33445,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(33445,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(33445,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(33445,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(33445,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(33445,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(33445,36,'cp -7','a1c1 c7c5 g1f3 c5d4 f3d4 b8c6 e2e3 c6d4 e3d4 b7b5 a2a3 b4c3 c1c3 b5c4 f1c4 a7a5 e1g1 e8g8 d4d5 c8a6 d5e6 d7e6 c4a6 a8a6 h2h3 a6b6 d1c1 f8d8 b2b3 b6b8 c1c2 d8d5 f1c1 b8d8 c3c8 g7g6 b3b4 a5b4');
INSERT INTO evaluations VALUES(34546,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(34546,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34546,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(34546,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34546,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(34546,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34546,36,'cp 0','b8c6 g1f3 e7e5 f1e2 c8d7 e1g1 e8c8 f1e1 d6g6 f3h4 g6e6 h4f3');
INSERT INTO evaluations VALUES(36627,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36627,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(36627,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(36627,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36627,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36627,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(36627,36,'cp -4','e3e4 c7c5 d4c5 d8a5 b1c3 a5c5 d1e2 b8c6 a2a3 e7e5 f4e5 d6e5 h2h3 c6d4 e2f2 f6h5 c1e3 c8e6 e1g1 f7f5 e4f5 e6f5 c3e4 c5c6 e3d4 f5e4 d3e4 c6e4 f2e3 e4e3 d4e3 e5e4 f3d4 f8e8 c2c3 h5g3 f1e1 a8d8 g1f2 g7e5 a3a4 e8f8 f2g1 f8e8');
INSERT INTO evaluations VALUES(39867,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(39867,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(39867,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(39867,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(39867,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(39867,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(39867,36,'cp 239','f3e5 d7d5 e4d5 c6d5 c4b5 b8d7 d2d4 c5b4 e1g1 e8g8 d1f3 b4c3 b2c3 d8b6 a2a4 d7e5 d4e5 f6e4 f1e1 c8e6 c1e3 b6c7 e3d4 f8c8 e1e3 c7e7 f3f4 h7h6 b5d3 e7g5 e3e4 d5e4 f4e4 g5g6 e4g6 f7g6 d3g6 c8c7 a4a5 a8f8 g6e4 f8c8 f2f4 a7a6');
INSERT INTO evaluations VALUES(39865,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(39865,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(39865,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(39865,33,'cp 243',NULL);
INSERT INTO evaluations VALUES(39865,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(39865,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(39865,36,'cp 244','d7d5 d2d4 c5b4 e4d5 c6d5 c4b5 b8d7 e1g1 e8g8 d1f3 b4c3 b2c3 d8b6 a2a4 d7e5 d4e5 f6e4 f1e1 c8e6 c1e3 b6c7 e3d4 f8c8 e1e3 c7e7 b5d3 e7h4 f3e2 c8c6 e2e1 e4c5 h2h3 h7h6 a4a5 a8c8 a1b1 b7b6 d3b5 c6c7 a5b6 a7b6 b1b4 h4d8 b5d3 c5d3 c2d3 c7c6 f2f4 d8c7 e1f2 e6f5 g2g4 f5d7 f4f5');
INSERT INTO evaluations VALUES(41702,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(41702,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(41702,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(41702,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(41702,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(41702,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(41702,36,'cp -35','c7c6 e1g1 e8g8 c1g5 f8e8 f1e1 c8e6 e2d3 b8d7 h2h3 d8c7 a2a4 a7a6 c3e2 h7h6 g5e3 e8e7 b2b3 c6c5 c2c4 d5c4 b3c4 c5d4 e2d4 a8e8 d4e6 e7e6 f3d4 e6e5 a1b1 d6c5 d3c2 b7b6 d4f5 c5e3 e1e3 e5e3 f5e3');
INSERT INTO evaluations VALUES(40519,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(40519,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(40519,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(40519,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(40519,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(40519,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(40519,36,'cp -25','e1g1 e8g8 h2h3 c8f5 e2d3 f5d3 d1d3 d8c7 c1d2 b8d7 f1e1 a7a6 a2a4 f8e8 e1e8 a8e8 a1e1 e8e1 d2e1 d7f8 g2g3 c7d7 g1g2 d7e6 c3d1 e6e4 d3e4 f6e4 d1e3 f7f6 f3d2 e4d2 e1d2 g8f7 b2b4 g7g6 b4b5 f7e6 c2c4 c6b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(40518,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(40518,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(40518,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(40518,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(40518,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(40518,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(40518,36,'cp -35','e8g8 h2h3 c8f5 e2d3 f5d3 d1d3 f8e8 c1d2 a7a5 a1e1 a5a4 e1e8 f6e8 c3e2 b8d7 b2b3 d7f6 c2c4 a4b3 a2b3 h7h6 f1c1 d8e7 e2c3 d6b4 c1e1 e7d8 c3d1 b4d2 d3d2 e8d6 d2c2 f6e4 d1e3 d8f6 e3g4 f6g6');
INSERT INTO evaluations VALUES(36118,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(36118,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(36118,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(36118,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(36118,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(36118,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(36118,36,'cp -19','h2h3 c8f5 e2d3 f5d3 d1d3 b8d7 f1e1 d8c7 c1d2 f8e8 e1e8 a8e8 a1e1 e8e1 d2e1 d7f8 g2g3 c7d7 g1g2 d7e6 e1d2 f8g6 c3e2 h7h6 d2e3 f6e4 d3b3 e6c8 e2c3 e4f6 c3e2 a7a6 b3d3 c8e6 a2a4 f6e4 d3b3 e6d7 e2c1 g6e7');
INSERT INTO evaluations VALUES(47510,30,'cp 432',NULL);
INSERT INTO evaluations VALUES(47510,31,'cp 425',NULL);
INSERT INTO evaluations VALUES(47510,32,'cp 439',NULL);
INSERT INTO evaluations VALUES(47510,33,'cp 431',NULL);
INSERT INTO evaluations VALUES(47510,34,'cp 432',NULL);
INSERT INTO evaluations VALUES(47510,35,'cp 440',NULL);
INSERT INTO evaluations VALUES(47510,36,'cp 443','f2f3 e7e6 f3e4 d8h4 e1d2 h4f4 d2c3 f4e4 a2a3 c7c5 g1f3 b8c6 d4c5 d5d4 c3b3 e4d5 c2c4 d4c3 d1d5 e6d5 b1c3 f8c5 b3a2 e8g8 f1d3 c8e6 c3e2 e6g4 e2f4 f8e8 d3b5 a7a6 b5c6 b7c6 c1d2 c5e7 a1c1 c6c5 b2b3 a8d8 h2h3 g4f3 g2f3');
INSERT INTO evaluations VALUES(47509,30,'cp 444',NULL);
INSERT INTO evaluations VALUES(47509,31,'cp 421',NULL);
INSERT INTO evaluations VALUES(47509,32,'cp 430',NULL);
INSERT INTO evaluations VALUES(47509,33,'cp 427',NULL);
INSERT INTO evaluations VALUES(47509,34,'cp 440',NULL);
INSERT INTO evaluations VALUES(47509,35,'cp 433',NULL);
INSERT INTO evaluations VALUES(47509,36,'cp 421','e7e6 f3e4 d8h4 e1d2 h4f4 d2c3 f4e4 g1f3 c7c5 a2a3 b8c6 d4c5 d5d4 c3b3 e4d5 c2c4 d4c3 d1d5 e6d5 b1c3 f8c5 b3a2 e8g8 f1d3 h7h6 c1f4 a7a6 a1c1 c8e6 c3e2 c5b6 f4d2 e6g4 h1f1 g4f3 f1f3 c6e5 f3h3 f8e8 e2f4 a8d8 c1e1 e5d3 e1e8 d8e8 h3d3');
INSERT INTO evaluations VALUES(49968,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(49968,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(49968,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(49968,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(49968,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(49968,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(49968,36,'cp 17','f8b4 g5d2 b4d2 d1d2 f6e4 d2c2 b7b6 g2g3 c8b7 f1g2 e8g8 e1g1 c7c5 b1c3 d8e7 a2a3 d7d6 a1d1 b8d7 d4c5 d6c5 c3e4 b7e4 c2a4 d7f6 f3e5 e4g2 g1g2 e7c7 e5d7 f8d8 d7f6 g7f6 f1e1 a7a5 d1d8 a8d8 e1d1 d8d1 a4d1 g8g7 b2b3 c7b7 f2f3 b7c7 h2h4 h7h5 d1d2 g7g6 d2d3 c7e5');
INSERT INTO evaluations VALUES(52022,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(52022,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(52022,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(52022,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52022,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(52022,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(52022,36,'cp 49','c2c4 b8c6 b1c3 d7d5 c4d5 e6d5 d2d4 g8f6 f1g2 c5d4 f3d4 f8c5 d4b3 c5b6 c3a4 b6c7 c1g5 e8g8 a4c3 d5d4 c3d5 c7e5 e1g1 c8e6 d5f6 e5f6 g5f6 d8f6 b3c5 f6e7 c5e6 e7e6 g2c6 b7c6 d1d4 e6e2 f1e1 e2a6 e1e7 a6b6 d4c3 a8d8 b2b3 h7h6');
INSERT INTO evaluations VALUES(54676,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(54676,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(54676,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(54676,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(54676,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(54676,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(54676,36,'cp 209','e2e4 g7g6 b1c3 c7c6 f1c4 f8g7 h2h4 g8h6 d1e2 c8g4 c1e3 b8a6 e2d2 h6f7 h4h5 g4h5 h1h5 g6h5 f3h4 a6c7 h4f5 e8g8 e3h6 d6d5 h6g7 d5c4 g7f8 d8f8 e1c1 b7b5 d1h1 f7d6 f5h6 g8h8 h1h5 a8e8 h5h1 b5b4 c3d1 d6e4');
INSERT INTO evaluations VALUES(54690,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(54690,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(54690,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(54690,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(54690,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(54690,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(54690,36,'cp 85','b1c3 g8f6 c1g5 d6d5 e2e3 e7e6 g2g4 f8b4 g4f5 e6f5 h1g1 e8g8 d1d3 c7c6 a2a3 b4d6 f3e5 g8h8 f2f4 c8e6 e1c1 d8c7 g5f6 f8f6 d3d2 b8d7 f1d3 b7b5 d2g2 a8g8 c3e2 d6e7 h2h4 b5b4 a3b4 e7b4 c2c3 b4d6');
INSERT INTO evaluations VALUES(56869,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(56869,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(56869,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(56869,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(56869,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(56869,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(56869,36,'cp 101','e2e4 f8g7 b1c3 d7d6 g1f3 b8d7 h2h4 h7h6 e4e5 d6e5 d4e5 a7a6 g2g3 b7b5 f1g2 c8b7 d1e2 g8e7 a1d1 e7d5 c3d5 b7d5 b2b3 d8e7 e1g1 d7b6 f1e1 a8d8 a2a4 b5a4 c2c4 d5f3 g2f3 e8g8 b3a4 d8d1 e1d1');
INSERT INTO evaluations VALUES(56914,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(56914,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(56914,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(56914,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(56914,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(56914,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(56914,36,'cp 56','d3d4 e5d4 d1d4 g8f6 c1g5 f8e7 b1c3 b8c6 d4a4 e8g8 h2h4 d7d6 e1c1 a7a6 f1b5 c6a7 b5e2 b7b5 a4d4 c7c5 d4d2 c8e6 g1f3 b5b4 c3d5 e6d5 g5f6 e7f6 d2d5 a7b5 e2c4 f6g7 e4e5 d6e5 f3g5 d8f6 g5e4 f6e7 d5d2 h7h6 h4h5 g6g5 d2e3 f8d8 e3c5');
INSERT INTO evaluations VALUES(3321,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(3321,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3321,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(3321,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(3321,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(3321,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(3321,36,'cp 18','c6b4 a1c1 h7h6 g5h4 b7b6 d3e2 d5c4 e2c4 b4d5 c4d5 e6d5 e1g1 c8b7 d1b3 c7c6 f3e5 f8e8 f1e1 a8c8 h2h3 f6d7 h4e7 e8e7 e5d7 e7d7 b3a3 c6c5 d4c5 c8c5 c1d1 d5d4 e3d4 c5g5 c3e4 g5g6 e4g3 d7d4 d1d4 d8d4');
INSERT INTO evaluations VALUES(4821,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4821,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(4821,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4821,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(4821,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4821,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4821,36,'cp 13','d7d6 c1e3 g8f6 c3d5 c5e3 f2e3 e8g8 d1e1 f6d5 c4d5 c8e6 d5e6 f7e6 f1f2 f8f7 e1c3 d8d7 h2h3 c6e7 a1f1 e7g6 c3d2 a8f8 b2b4 d7e7 f3h2 f7f2 f1f2 f8f2 d2f2 g6f8 h2f3 f8d7 f2d2 d7f6 a2a4 g8h8 d2c3');
INSERT INTO evaluations VALUES(4753,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(4753,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(4753,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4753,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(4753,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(4753,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4753,36,'cp 15','c1e3 g8e7 d3d4 e5d4 f3d4 e8g8 a2a3 c6d4 e3d4 c5d4 d1d4 b7b5 c4d3 c7c5 d4e3 c8b7 a1d1 d8c7 f1e1 a8d8 e3g3 f7f5 e4f5 e7f5 g3g6 c7f7 g6f7 f8f7 f2f3 g8f8 d3e4 b7e4 c3e4 d8e8 c2c3 e8e5 g1f2');
INSERT INTO evaluations VALUES(16776,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(16776,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(16776,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(16776,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(16776,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(16776,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(16776,36,'cp 50','e3e4 d5b6 g1f3 f8b4 c1d2 d8e7 b1c3 d7d6 e5d6 b4d6 e4e7 d6e7 c3b5 e7d8 e1c1 e8g8 d2f4 c8f5 f1e2 a7a6 b5d4 c6d4 f3d4 f5e4 e2f3 e4f3 d4f3 c7c6 h1e1 b6d5 f4d6 d8a5 d6f8 a5e1 f8c5 e1a5 c2c4 d5f4 b2b4 f4e6 c5d6 a5d8 a2a4 f7f6 a4a5 g8f7 c1c2 d8e7');
INSERT INTO evaluations VALUES(16775,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(16775,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(16775,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(16775,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(16775,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(16775,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(16775,36,'cp 34','d5b6 g1f3 f8b4 c1d2 d8e7 b1c3 d7d6 e5d6 b4d6 e4e7 d6e7 c3b5 e7d8 e1c1 e8g8 d2f4 c8f5 a2a3 a8c8 f1d3 f5d3 d1d3 a7a6 b5c3 d8f6 h1d1 f8e8 c3d5 b6d5 d3d5 g8f8 c2c3 e8e2 d1d2 e2d2 d5d2 c6d8 c1c2 d8e6 f4e3 f8e8 h2h4 c8d8');
INSERT INTO evaluations VALUES(17989,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(17989,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(17989,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(17989,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(17989,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(17989,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(17989,36,'cp 57','e1g1 a8b8 c2c3 d4d3 d1d3 g8f6 c1f4 f8e7 f1e1 e8g8 b1d2 f6h5 f4e3 c6e5 d3e2 e5g6 f3d4 h5f4 e2f1 d8d7 d4f5 f8e8 g2g3 f4h5 f5e7 d7e7 f2f3 c8b7 f1f2 h5f6 b3c2 c7c5');
INSERT INTO evaluations VALUES(19268,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(19268,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(19268,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(19268,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(19268,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(19268,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(19268,36,'cp 204','d4d5 c6a5 h2h4 h7h6 d5d6 a7a6 b1c3 b7b5 c1f4 c8b7 d6e7 g8e7 f4e5 h8h7 d1d2 d8b6 e1c1 e8c8 e5d6 f7f5 f3e5 f5e4 h4h5 g6g5 e5g6 e7g6 h5g6 f8d6 d2d6 b6d6 d1d6 h7h8 d6f6 h8f8 f6f8 d8f8 h1h6');
INSERT INTO evaluations VALUES(19300,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(19300,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(19300,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(19300,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(19300,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(19300,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(19300,36,'cp 235','c6a5 h2h4 h7h6 d5d6 a7a6 c1f4 b7b5 b1c3 c8b7 d6e7 f8e7 e4e5 e7f8 d1d3 g8e7 e1c1 f8g7 c3e4 b7e4 d3e4 d8c8 e4e1 a5c4 f1c4 b5c4 c2c3 c8c6 e1e2 e8g8 e2c4 f8b8 d1d6 c6b5 c4b5 a6b5 h1d1 a8a2 d6d7 a2a4 d7d8 b8d8 d1d8 g8h7');
INSERT INTO evaluations VALUES(20715,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(20715,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(20715,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20715,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(20715,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(20715,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20715,36,'cp 0','d4d3 f8e7 c4b3 g8f6 c2c4 e8g8 e1g1 b7b6 b1c3 c8b7 c1f4 d8c7 c3b5 c7b8 b5c3');
INSERT INTO evaluations VALUES(26215,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(26215,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(26215,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(26215,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(26215,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(26215,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(26215,36,'cp -8','a8c8 d2d4 e7e6 c1f4 d8b6 b1d2 g8f6 a1b1 f8d6 f4d6 b6d6 c2c3 e8g8 f1e1 d6a6 a2a3 f6e8 f3e5 d7f6 d1f3 e8d6 e5g4 f6g4 f3g4 b7b6 g4f4 f8d8 d2f3 a6d3 b1d1 d3g6 a3a4 g6f5');
INSERT INTO evaluations VALUES(26913,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(26913,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(26913,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(26913,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(26913,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(26913,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(26913,36,'cp 50','a7a6 b5e2 f8e7 d2d4 g8f6 d4d5 c6b8 f3d2 g4e2 d1e2 e8g8 a2a4 b8d7 a4a5 c7c6 c3c4 b7b6 b1c3 c6d5 e4d5 d8b8 a1a3 f8c8 a5b6 b8b6 c3a4 b6b7 b2b3 d7b6 a4c3 f6d7 c1b2 f7f5 f1a1 e7g5');
INSERT INTO evaluations VALUES(27037,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(27037,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(27037,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(27037,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(27037,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(27037,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(27037,36,'cp 176','d1a4 d6d5 f1a6 f8b4 c1d2 g4d7 a6b7 c6d4 a4d1 d4f3 d1f3 b4d2 b1d2 a8b8 b7d5 g8f6 e1g1 e8g8 d5b3 d7e6 f3c3 e6b3 a2b3 f8e8 f1e1 h7h6 a1a7 f6d5 c3g3 d8f6 e4e5 f6f4 g3f4 d5f4 a7c7 f4d3 e1d1 d3e5 c7c3 e8d8 h2h3 e5d3 d2c4 d3b4');
INSERT INTO evaluations VALUES(35758,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(35758,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(35758,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(35758,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(35758,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35758,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(35758,36,'cp 24','d1c2 b7b6 f3e5 c8b7 b1c3 f6d7 f1d1 d7e5 d4e5 d8c7 c4d5 c6d5 e2e4 d5d4 d1d4 b8c6 c3d5 e6d5 e4d5 c7e5 d4e4 e5d6 c1f4 d6f6 d5c6 b7c6 e4c4 c6g2 g1g2 f6e6 a1c1 a8d8 c2e4 e6e4 c4e4 e7c5 b2b4 f7f5 e4e6 c5b4 c1c7');
INSERT INTO evaluations VALUES(35759,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(35759,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(35759,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(35759,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(35759,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(35759,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(35759,36,'cp 33','b7b6 f1d1 b8d7 b1d2 c8a6 e2e4 f6e4 d2e4 d5e4 f3e5 d7e5 d4e5 d8c7 g2e4 h7h6 c1e3 c7e5 d1d7 e7c5 a1e1 c5e3 e1e3 e5c5 b2b3 b6b5 e4h7 g8h8 c4b5 c5b5 h7d3 b5a5 d3a6 a5a6 e3c3 a6a5 c3c6 a8d8 c6c7 d8d7 c7d7 h8g8 h2h4 a5a3 g1g2 a7a5 d7b7 a3d6 c2e4 d6d1 a2a4 f8c8 e4f3 d1f3 g2f3 c8c3 f3e4 c3c2 e4e3 c2b2 e3f3');
INSERT INTO evaluations VALUES(37818,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(37818,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(37818,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37818,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(37818,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(37818,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37818,36,'cp 0','d7d6 d2d3 b8c6 g1e2 a8b8 a2a4 a7a6 e1g1 b7b5 a4b5 a6b5 f2f4 c8d7 h2h3 g8f6 g3g4 e8g8 e2g3 c6d4 c3e2 f6e8 c2c3 d4e2 g3e2 b5b4 c1e3 b4c3 b2c3 b8b2 e3c1 b2b8');
INSERT INTO evaluations VALUES(37975,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37975,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37975,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37975,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37975,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37975,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37975,36,'cp 0','d2d3 b8c6 g1e2 g8f6 e1g1 e8g8 c3d5 f6d5 e4d5 c6d4 e2d4 c5d4 f1e1 f8e8 c1f4 c8d7 d1d2 a8c8 e1e2 d7g4 e2e4 g4f5 e4e2');
INSERT INTO evaluations VALUES(37977,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37977,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37977,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37977,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37977,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(37977,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37977,36,'cp 0','b8c6 g1e2 g8f6 e1g1 e8g8 a2a4 a7a6 c3d5 f6d5 e4d5 c6d4 e2d4 c5d4 f1e1 f8e8 h2h4 h7h5 c1g5 c8g4 d1d2 a8c8 a1c1 b7b6 g5h6 g7f6 h6g5 f6g7');
INSERT INTO evaluations VALUES(42766,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(42766,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42766,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42766,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42766,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42766,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42766,36,'cp 0','b1c3 c7c6 f2f4 d7d6 f1e2 b8d7 g1f3 b4c5 c3a4 c5b4');
INSERT INTO evaluations VALUES(44092,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(44092,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(44092,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(44092,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(44092,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(44092,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(44092,36,'cp -62','d7d5 c4d5 c6d5 e4d5 f6d5 d3d4 f8b4 d4e5 b8c6 f1b5 e8g8 g1e2 d5c3 b2c3 d8d1 e1d1 f8d8 d1c2 b4a5 e2g3 c6e5 h1e1 a7a6 b5f1 e5g6 c1g5 f7f6 g5e3 c8d7 e3d4 a8c8 c2b3 g6e5 a1d1 d7e6 b3b2 g8f7 d4e5 f6e5 g3e4 c8c6 d1d8 a5d8 f1e2 d8e7');
INSERT INTO evaluations VALUES(45813,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(45813,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(45813,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(45813,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(45813,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(45813,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(45813,36,'cp -62','c4d5 c6d5 e4d5 f6d5 d3d4 d5c3 b2c3 e5d4 f1b5 b8c6 d1d4 d8d4 c3d4 f8e7 g1f3 e8g8 e1g1 c8e6 a2a4 e6d5 c1e3 f8d8 f1c1 a8c8 f3e5 c6b4 b5c4 f7f6 c4d5 b4d5 e5d3 g8f7 g1f1 b7b6 e3d2 e7d6 g2g3 h7h5 a1b1 d6a3 c1c8 d8c8 d3f4');
INSERT INTO evaluations VALUES(46224,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46224,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46224,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(46224,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46224,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46224,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46224,36,'cp 1','c5d4 e3d4 f8d6 f1d3 b8c6 b1d2 e8g8 a2a3 b7b6 d1e2 c8b7 f3e5 a8c8 e1g1 c8c7 c2c3 a7a5 a1e1 g7g6 b3b4 c6e7 d2b3 f6e4 d3e4 d5e4 b4a5 d6e5 d4e5 b6a5 b3a5 d8d3 a5b7 c7b7 b2c1 f8c8 e2e4 d3e4 e1e4 c8c3 e4a4 e7d5 c1d2 c3c2 f1d1');
INSERT INTO evaluations VALUES(46342,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46342,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46342,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(46342,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(46342,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46342,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46342,36,'cp 21','f8d6 f4d6 d8d6 f1d3 b8c6 b1c3 c8d7 e1g1 e8g8 a1c1 f8c8 d1e2 a7a6 h2h3 d7e8 f1d1 c8c7 f3d2 a8c8 d2b3 h7h6 e2d2 c6e7 a2a3 f6d7 c3a4 d7f6 c1c7 c8c7 a4c3 d6d8 d1c1 e7c8 c3e2 c8d6 c1c7 d8c7 b3c5');
INSERT INTO evaluations VALUES(46971,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(46971,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(46971,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(46971,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(46971,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(46971,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46971,36,'cp 22','g5f6 g7f6 e2e3 f6f5 g1e2 c7c5 c2c3 e7e6 e2f4 b8d7 b1d2 d6d5 a2a4 a7a6 g2g3 f8d6 f1g2 c5d4 c3d4 d7f6 e1g1 e8g8 d2f3 c8d7 f3e5 a8c8 f4d3 d7e8 d1b3 b7b5 a4a5 f6e4 f1c1 c8c1 d3c1 d6e5 d4e5 d5d4');
INSERT INTO evaluations VALUES(46958,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(46958,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46958,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(46958,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(46958,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46958,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(46958,36,'cp 14','g7f6 e2e3 f6f5 g1e2 c7c5 c2c3 b8d7 e2f4 e7e6 b1d2 d6d5 f4h5 d8b6 b2b3 f8e7 f1d3 d7f6 e1g1 c8d7 h5f6 e7f6 d1h5 e8e7 d2f3 h8g8 a1c1 b6d6 g2g3 b7b6 f1d1 d7c6 d3f1 a8d8 f1g2 d6c7 f3h4 g8g4 g2h3');
INSERT INTO evaluations VALUES(48115,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(48115,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(48115,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(48115,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(48115,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(48115,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(48115,36,'cp 76','b1d2 f8d6 f1d3 d6f4 e3f4 e4d6 c2c3 b7b6 h2h4 c7c5 h4h5 e8g8 d1c2 h7h6 g2g4 g8h8 g4g5 c8a6 d3a6 b8a6 c2a4 d8c8 f3e5 b6b5 a4d1 b5b4 d2f3 b4c3 b2c3 c5d4 c3d4 a6b4 e1g1 c8c2 a1c1 c2d1 f1d1 f8c8 g5g6 f7f6 e5f7 d6f7 c1c8 a8c8');
INSERT INTO evaluations VALUES(48116,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(48116,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(48116,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(48116,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(48116,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(48116,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(48116,36,'cp 62','f8d6 f1d3 e4d2 d1d2 b7b6 e1c1 c8a6 d3a6 b8a6 h2h4 a6b4 c1b1 b4c6 d2d3 d8d7 h4h5 h7h6 h1e1 a7a5 g2g4 d6f4 e3f4 e8c8 a2a3 c8b7 e1e3 h8e8 d3e2 e8g8 f3e5 c6e5 f4e5 d7a4 f2f4 a4c4 e2f3 b7b8');
INSERT INTO evaluations VALUES(54792,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(54792,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(54792,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(54792,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(54792,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(54792,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(54792,36,'cp 114','f4g3 h7h5 h2h3 c8f5 c2c4 e7e6 c4d5 e6d5 b1c3 b8c6 d1b3 d8d7 e2e3 h5h4 g3h2 f8b4 a2a3 b4a5 f3d2 g8e7 a1c1 c6d8 b3d1 a5b6 f1e2 c7c6 b2b4 a8c8 c3a4 e8g8 a4c5 b6c5 b4c5 g8g7 d1b3 d8f7 e1g1 f7h6 b3a4 e7g6 a4a7 c8a8');
INSERT INTO evaluations VALUES(54795,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(54795,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(54795,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(54795,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(54795,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(54795,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(54795,36,'cp 79','h7h5 h2h3 c8f5 c2c4 e7e6 c4d5 e6d5 b1c3 b8c6 d1b3 h5h4 g3h2 f8b4 e2e3 g8e7 f3d2 b4a5 a2a3 d8d7 a1c1 c6d8 f1e2 c7c6 e1g1 e8g8 b3d1 g8g7 b2b4 a5c7 h2c7 d7c7 a3a4 d8f7 b4b5 a8c8 d2b3 b7b6 a4a5 c7d6 b5c6');
INSERT INTO evaluations VALUES(54800,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(54800,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(54800,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(54800,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(54800,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(54800,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(54800,36,'cp 103','h2h3 c8f5 c2c4 e7e6 c4d5 e6d5 d1b3 b8c6 b1c3 h5h4 g3h2 f8b4 e2e3 d8d7 a2a3 b4a5 f3d2 g8e7 a1c1 c6d8 f1e2 c7c6 e1g1 d8f7 b3d1 f7d6 d2b3 a5c3 b2c3 b7b6 h2d6 d7d6 a3a4 e8f8 c3c4 f8g7 f1e1 h8e8 c4c5 d6e6');
INSERT INTO evaluations VALUES(54979,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54979,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(54979,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(54979,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(54979,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54979,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54979,36,'cp 13','d6d5 e4d5 d8d5 g1f3 c8g4 f1e2 b8c6 h2h3 g4h5 b1c3 d5d7 d2d4 e7e6 d4d5 e6d5 d1d5 g8f6 d5d7 f6d7 c1e3 e8c8 e1c1 f7f6 a2a3 f8c5 e3c5 d7c5 b2b4 c5d7 g3g4 h5f7 g4g5 h8e8 h1g1 d7b6 g5f6 g7f6 d1d8 c8d8 f3d2 f7g6 e2f3 c6d4 f3b7 g6c2');
INSERT INTO evaluations VALUES(57204,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(57204,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(57204,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(57204,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(57204,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(57204,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(57204,36,'cp -56','g8f6 d2d4 e5e4 f2f3 e4f3 g1f3 d7d5 f1g2 f8e7 e1g1 c7c5 c3c4 d5c4 b1c3 c5d4 f3d4 e8g8 g1h1 h7h6 a1b1 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b4 f4e3 a7a5 f1d1 c8f5 d1d8 e7d8 b1d1 d8e7 d1d2 f6g4');
INSERT INTO evaluations VALUES(2080,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(2080,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(2080,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(2080,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(2080,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(2080,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(2080,36,'cp 147','c7c6 c1f4 e5g6 f4g3 g8f6 d1e2 f6h5 f3d4 h5g3 h2g3 d7d6 f2f4 c8d7 e1c1 e8c8 e2c4 d7e8 c4a4 c8b8 a4b3 h7h5 f1c4 e7c7 c3e2 h5h4 g3h4 h8h4 h1h4 g6h4 g2g3 h4g6 a2a3 b8a8 e2c3 d8d7 c1b1 c7b6 d4f5 b6b3 c4b3 d6d5 e4d5 g6e7 f5e3 a8b8 d1h1 e7d5 e3d5 c6d5 b3d5 f8c5');
INSERT INTO evaluations VALUES(2136,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(2136,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(2136,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(2136,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(2136,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(2136,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(2136,36,'cp 40','e7e6 c1e3 a7a6 d1f3 f8d6 e1c1 d6e5 c1b1 g8e7 d4c6 b7c6 e3d4 e5d4 d1d4 e8g8 f3g3 c7b6 d4d6 f7f5 f2f3 b6a5 a2a3 a8b8 b1a1 a5c5 d6d1 f5f4 g3e1 d7d5 f1d3 c5a7 e1d2 c6c5 e4d5 e6d5 b2b3 c8f5 d3f5 f8f5 c3a4 c5c4 b3c4 h7h5 c4c5');
INSERT INTO evaluations VALUES(7490,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(7490,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(7490,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7490,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(7490,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7490,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(7490,36,'cp 0','d7d6 a2a4 b5b4 c3d5 c6a5 d5e7 d8e7 d2d4 e8g8 c1g5 h7h6 g5f6 e7f6 d4e5 d6e5 d1d5 a5b3 c2b3 c8e6 d5e5 a8d8 e5f6 g7f6 a1c1 e6b3 a4a5 f8e8 f1e1 f6f5 e1e3 f5e4 e3b3 e4f3 b3b4 e8e2 g2f3 d8d2 c1c7 e2f2 b4g4 g8f8 c7a7 d2b2 a7a8 f8e7 a8a7');
INSERT INTO evaluations VALUES(7391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7391,36,'cp 3','a2a3 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 d2d3 d4b3 c2b3 e8g8 d3d4 e5d4 d1d4 d6d5 e4d5 f6d5 c3d5 d8d5 d4d5 c6d5 f1d1 f8d8 c1e3 e7f6 e3b6 d8c8 a1b1 c8c6 b6d4 f6d4 d1d4 a8d8 b1d1 c6c2 a3a4 g7g6 a4b5 a6b5 d4d5 d8d5 d1d5 c2b2 d5b5 g8g7 g1h2');
INSERT INTO evaluations VALUES(16821,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(16821,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(16821,32,'cp -164',NULL);
INSERT INTO evaluations VALUES(16821,33,'cp -162',NULL);
INSERT INTO evaluations VALUES(16821,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(16821,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(16821,36,'cp -160','d7d5 c4d3 d5e4 d3e4 c8d7 b1c3 f7f5 e4c6 d7c6 d1e2 d8d4 d2d3 e8c8 c1e3 d4b4 e1c1 c6g2 h1g1 g2c6 e2h5 c6d7 h5f7 b4d6 d3d4 e5e4 f7c4 g7g6 d4d5 d6b4 c4e2 f8g7 a2a3 b4a5 e2c4 g7c3 c4c3 a5c3 b2c3 h8f8 e3f4 d7b5 d1d4 c7c5 d5c6 b5c6 f4e3 f5f4 d4d8 c8d8');
INSERT INTO evaluations VALUES(15378,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(15378,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(15378,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(15378,33,'cp -152',NULL);
INSERT INTO evaluations VALUES(15378,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(15378,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(15378,36,'cp -152','c4d3 d5e4 d3e4 c8d7 b1c3 f7f5 e4c6 d7c6 d1e2 d8d4 d2d3 e8c8 c1e3 d4b4 e1c1 c6g2 h1g1 g2c6 e2h5 f5f4 e3d2 b4d4 d1e1 f8d6 g1g7 d4f2 e1e5 f4f3 c3d1 f2d4 e5g5 h8e8 g5g4 d4f6 h5f7 f6f7 g7f7 e8e2 g4g7 h7h5 g7h7 e2h2 d2f4 d6f4 f7f4 b7b6 f4f7 c8b8 f7c7');
INSERT INTO evaluations VALUES(17384,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(17384,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(17384,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(17384,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(17384,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(17384,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(17384,36,'cp 212','d4d5 c6a5 d5e6 f7e6 e4e5 d7d5 e5d6 d8d6 d1e2 a5c6 g2g3 e6e5 c1f4 c8g4 f4e5 c6e5 e2e5 f8e7 e5e3 g8f6 f3e5 e8g8 f1c4 g8h8 e1g1 g4h5 a1d1 d6c7 c3d5 f6d5 d1d5 a8d8 e5d3 d8d5 c4d5 f8d8 d3f4 e7g5 f4e6 g5e3 e6c7 e3d2 d5b7 h5e2 c7e6 e2f1');
INSERT INTO evaluations VALUES(17410,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(17410,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(17410,32,'cp 251',NULL);
INSERT INTO evaluations VALUES(17410,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(17410,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(17410,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(17410,36,'cp 213','c6a5 d5e6 f7e6 e4e5 d7d5 e5d6 d8d6 d1d6 f8d6 c3e4 d6e7 c1f4 b7b6 e1c1 g8f6 e4d6 e7d6 f4d6 c8b7 f3g5 b7d5 d6c7 h7h6 g5h3 a5b7 c7b6 d5a2 h3f4 g7g5 f4d3 f6d7 b6c7 a2d5 d3e5 d7e5 c7e5 h8f8 f2f3 f8f5 e5c3 c5c4 h2h4 b7d6 h4g5 h6g5 f1e2 e8c8 h1e1');
INSERT INTO evaluations VALUES(20942,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(20942,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(20942,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(20942,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(20942,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(20942,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20942,36,'cp 36','c8d7 d2d4 e7e6 e1g1 c6e5 b5e2 e5f3 e2f3 f7f5 f3h5 g7g6 h5e2 f8d6 b1d2 c7c6 a2a4 a7a5 d2f3 g8h6 c2c4 h6f7 f3e1 g6g5 f2f4 h7h5 e1d3 h5h4 c4c5 d6c7 c1d2 d8f6 d1b3 a8b8 b3c3 g5g4 b2b4 b8a8 b4a5 c7a5');
INSERT INTO evaluations VALUES(27283,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(27283,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(27283,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(27283,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(27283,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(27283,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(27283,36,'cp 40','e1g1 e7e6 d2d4 c6e5 b5e2 e5f3 e2f3 f8d6 c2c4 c7c6 b1d2 f7f5 a2a4 d8c7 h2h3 a7a5 f3e2 g8f6 d2f3 f6e4 b2b3 g7g5 c1a3 d6a3 a1a3 h7h5 h3h4 g5h4 f3e5 h4h3 e2h5 e8d8 g2g3 h3h2 g1h1');
INSERT INTO evaluations VALUES(32295,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(32295,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(32295,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(32295,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32295,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32295,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32295,36,'cp 4','g8f6 c2c4 c7c5 c4d5 e6d5 f4b8 a8b8 f1b5 f5d7 b5d7 d8d7 d4c5 f8c5 a1c1 b7b6 g1e2 e8g8 e1g1 d5d4 e2d4 c5d4 e3d4 d7d4 d1c2 b8d8 d2f3 d4f4 a2a3 d8d7 f1d1 g7g6 d1d7 f6d7 c2c7 f4a4 h2h3 d7c5 c7e7 c5e6');
INSERT INTO evaluations VALUES(36578,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(36578,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36578,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36578,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36578,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36578,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(36578,36,'cp 0','c7c5 d4d5 e7e6 g1e2 e6d5 c4d5 b8d7 a2a4 d7e5 e1g1 a7a6 e2g3 e5d3 d1d3 a8b8 c1g5 h7h6 g5e3 b7b5 a4b5 a6b5 c3b5 f6d7 f3f4 d8b6 b5a7 b6b3 d3b3 b8b3 e3c1 b3b6 a7c6 c8b7 f1d1 b7c6 d5c6 b6c6 e4e5 d7b6 d1d6 c6d6 e5d6 f8d8 c1e3 d8d6 e3c5 d6c6 c5b6 c6b6 g3e4 g7d4 g1f1');
INSERT INTO evaluations VALUES(41204,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(41204,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41204,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(41204,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(41204,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41204,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(41204,36,'cp 10','c7c5 g1f3 h7h6 g5h4 g6g5 h4g3 f6h5 f1e2 h5g3 h2g3 g5g4 f3h4 h6h5 a1c1 b8c6 d4d5 c6e5 d1c2 c8d7 e1g1 a7a6 a2a3 a8c8 a3a4 d8a5 b2b3 a5b4 c3e4 g7h6 e2d3 c8b8 e4c3 e5d3 c2d3 h6g7 c3e2 b7b5 a4b5 a6b5 h4f5 d7f5 d3f5 b4b3 c4b5 b3b5 e2f4');
INSERT INTO evaluations VALUES(51125,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(51125,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(51125,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(51125,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(51125,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(51125,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(51125,36,'cp -41','b5b4 c3d5 e7e6 d5e3 c8b7 d4d5 g8f6 d5e6 d8d1 e3d1 f7e6 f2f3 f6d7 f1c4 b8c6 c1e3 c6e5 c4e2 d7c5 g1h3 e5d3 e2d3 c5d3 e1e2 e8c8 h3g5 d8d7 d1f2 d3b2 h1b1 b2c4 g5e6 a6a5 f2d3 f8e7 b1c1 e7f6 a1b1 c4e3 e2e3 h8e8 e6c5 d7d8 c5b7 c8b7 e3e2 f6d4 b1b3');
INSERT INTO evaluations VALUES(51127,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(51127,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(51127,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(51127,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(51127,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(51127,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(51127,36,'cp -20','c3e2 b8c6 g1f3 e7e6 c1g5 f8e7 g5e7 g8e7 e2g3 c6a5 a1c1 c7c5 f1c4 c5d4 d1d4 d8d4 f3d4 c8d7 c4d3 d7a4 c1a1 a5b3 a1a4 b3d4 g3e2 d4c6 a4a6 a8a6 d3a6 e8d7 e1d2 d7c7 f2f4 c7b6 a6c4 c6a5 h1c1 h8c8 c4a2 e7c6 d2e3 b4b3 a2b1');
INSERT INTO evaluations VALUES(51209,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(51209,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(51209,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(51209,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51209,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51209,36,'cp -11','c5d4 a2a4 b5b4 f3d4 c8b7 e2d3 d8c7 d1e2 g8f6 e1g1 f8e7 c2c3 b8c6 d4c6 b7c6 c3b4 e7b4 b1c3 c7e5 f2f4 b4c5 g1h1 e5h5 e2h5 f6h5 a1b1 c5d4 c3e2 d4b6 e2c3 e8g8 b2b4 h5f6 a4a5 b6d4 b4b5 d4c3 b5c6');
INSERT INTO evaluations VALUES(51192,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(51192,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(51192,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51192,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(51192,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(51192,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(51192,36,'cp -7','a2a4 b5b4 f3d4 c8b7 e2d3 g8f6 d1e2 d8c7 e1g1 f8e7 c2c3 b8c6 d4c6 c7c6 c3b4 f6e4 b4b5 a6b5 a4b5 c6d5 d3c4 d5e5 a1a8 b7a8 b1d2 e8g8 d2e4 a8e4 c4d3 e7d6 g2g3 e4d5 e2e5 d6e5 c1e3 f8a8 f1c1 g8f8 b2b4 f8e7 b5b6 a8a3 e3c5 d7d6');
INSERT INTO evaluations VALUES(57045,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(57045,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(57045,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57045,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(57045,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(57045,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(57045,36,'cp -9','d7d5 e4d5 e6d5 d2d4 f8d6 g1f3 g8f6 e1g1 e8g8 c1g5 h7h6 g5h4 c8f5 e2d3 f5d3 d1d3 b8d7 f1e1 c7c6 b1d2 d8a5 c2c4 f8e8 c4c5 d6f4 h4f6 d7f6 a2a3 a7a6 d2b3 a5a4 g2g3 f4c7 e1e8 a8e8 a1e1 e8e1 f3e1 f6d7 b3d2 d7f8 d2f3 f8e6 b2b4 g7g6 g1g2 a6a5 e1c2');
INSERT INTO evaluations VALUES(55915,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55915,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(55915,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(55915,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(55915,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(55915,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55915,36,'cp -4','e4d5 e6d5 g1f3 g8f6 d2d4 f8d6 c1g5 e8g8 e1g1 c8f5 e2d3 f5d3 d1d3 b8d7 f1e1 h7h6 g5f6 d7f6 b1d2 c7c6 g2g3 d8d7 c2c3 a8e8 g1g2 a7a6 a2a4 d7c7 b2b4 a6a5 b4b5 e8e1 a1e1 c6b5 a4b5 c7d7 f3e5');
INSERT INTO evaluations VALUES(3477,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(3477,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(3477,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(3477,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(3477,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(3477,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(3477,36,'cp 50','d4d5 c6b8 a2a4 c8g4 h2h3 g4f3 d1f3 c7c6 a4a5 d8c7 c4e2 b8d7 f1d1 a7a6 c1e3 f8b8 e2f1 b8e8 f3e2 e7e6 d5c6 c7c6 a1a4 a8c8 a4c4 c6c4 e2c4 c8c4 f1c4 e8c8 c4b3 g7f8 d1d4 d7c5 b3c4 f6d7 c4f1 b7b5 a5b6 d7b6 d4b4');
INSERT INTO evaluations VALUES(3910,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(3910,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(3910,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(3910,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(3910,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(3910,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(3910,36,'cp 201','f3e5 f6d7 e5f3 c5d6 b1c3 h7h6 f1e1 c6c5 c3d5 b7b5 e4e5 c8b7 c2c4 d6e7 d5f4 d7b6 e5e6 e7d6 f3e5 d6e5 e1e5 f7e6 d1e2 d8d6 e5e6 d6d7 c1e3 f8e8 e6e8 a8e8 b2b3 b5c4 d3c4 d7f5 f2f3 b6d7 a1e1 b7c6 e2f2');
INSERT INTO evaluations VALUES(6197,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(6197,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(6197,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(6197,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(6197,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(6197,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(6197,36,'cp 93','d7d6 h2h3 g8e7 f1e1 e8g8 b1c3 g8h8 c4b3 c8d7 c1e3 f7f5 d4d5 c6e5 f3e5 d6e5 d5d6 c7d6 d1d6 b6e3 f2e3 d7c6 d6e5 e7g6 e5g3 d8h4 g3h4 g6h4 b3d5 f5e4 d5c6 b7c6 c3e4 a8d8 e1e2 f8e8 e4c5 h4g6 a1c1 g6e5 e3e4 h8g8 a2a3 a7a5 c5b7');
INSERT INTO evaluations VALUES(8952,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(8952,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(8952,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(8952,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(8952,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(8952,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(8952,36,'cp 89','b1d2 e8g8 f4g3 b7b6 a2a3 b4d2 f3d2 c6e7 e1g1 e7f5 g3e5 f6d7 e5f4 a7a5 c2c4 c8a6 d1c2 d8f6 c2a4 f6d8 f1c1 d5c4 d3c4 a6c4 d2c4 a8c8 c1d1 d7f6 c4e5 f6d5 f4g3 f7f6 e5d3 f5d6 a4c2 d8e7');
INSERT INTO evaluations VALUES(9291,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(9291,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9291,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(9291,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(9291,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9291,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(9291,36,'cp 15','e8g8 f1e1 c5b6 h2h3 c6e7 d3d4 d6d5 f3e5 f6e4 b1d2 e4d6 b5f1 c7c6 a2a4 a7a5 d2f3 b6c7 c1f4 f8e8 d1b3 f7f6 e5g4 h7h5 g4e3 e7f5 f1d3 f5e3 e1e3 e8e3 f4e3 c8e6 b3c2 d8d7 a1e1 e6f7 e3f4 g7g5');
INSERT INTO evaluations VALUES(10794,30,'cp 281',NULL);
INSERT INTO evaluations VALUES(10794,31,'cp 253',NULL);
INSERT INTO evaluations VALUES(10794,32,'cp 282',NULL);
INSERT INTO evaluations VALUES(10794,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(10794,34,'cp 269',NULL);
INSERT INTO evaluations VALUES(10794,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(10794,36,'cp 268','e4e5 c7c5 d4c5 d7d6 c5d6 d8d6 e5d6 f6e4 b1a3 e4d6 a3c4 d6c4 f1c4 c8e6 c4e6 f7e6 e1c1 a8d8 d1d8 e8d8 h1e1 d8e7 e1e5 b7b6 e5b5 e7d6 a2a4 h8f8 f2f3 h7h6 b2b3 g7g5 h2h3 f8d8 c1d2 d8f8 c2c3 f8c8 d2d3 c8d8 d3e3 d8c8 c3c4 c8c5');
INSERT INTO evaluations VALUES(10782,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(10782,31,'cp 253',NULL);
INSERT INTO evaluations VALUES(10782,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(10782,33,'cp 243',NULL);
INSERT INTO evaluations VALUES(10782,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(10782,35,'cp 249',NULL);
INSERT INTO evaluations VALUES(10782,36,'cp 256','c7c5 d4c5 d7d6 c5d6 d8d6 e5d6 f6e4 f1d3 e4d6 b1c3 c8d7 e1c1 e8c8 h1e1 h7h6 e1e5 h8e8 d1e1 c8c7 b2b3 d7c6 f2f3 e8e5 e1e5 d6c8 c1d2 a7a6 c3e2 b7b5 c2c3 c7d6 e2d4 c6d7 e5h5 c8e7 c3c4 b5c4 b3c4 e7c6 d2c3 c6d4 h5d5 d6c7 c3d4');
INSERT INTO evaluations VALUES(14780,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(14780,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(14780,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(14780,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(14780,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(14780,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(14780,36,'cp 12','d5c4 d3c4 f8d6 b1c3 e8g8 e3e4 e6e5 d4d5 c6e7 e1g1 a7a6 a2a4 c7c6 d5c6 e7c6 h2h3 h7h6 c1e3 d8e7 a1c1 f8e8 d1e2 c8e6 f1d1 a8c8 c4e6 e7e6 e2d3 d6e7 c3d5 c6b4 d5e7 e6e7 d3b3 c8c1 d1c1 b4c6 f3d2 c6d4 b3c4');
INSERT INTO evaluations VALUES(23846,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23846,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23846,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23846,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(23846,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(23846,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23846,36,'cp 8','d8e7 c1g5 e7e4 b1d2 e4g6 c2c3 h7h6 d1b3 c6b4 c3b4 h6g5 d2c4 g4f3 e2f3 h8h4 e1c1 h4c4 b3c4 d6f4 d1d2 a8d8 h1d1 d8d2 d1d2 g6d6 c4e2 g8e7 g2g3 f4d2 e2d2 d6d2 c1d2 b7b6 d2d3 e8d7 d3c4 c7c6 h2h3 d7d6 c4d4 e7f5 d4e4 f5h6 f3e2 g7g6 e2d3 h6f5 d3c4 d6e7 c4d3');
INSERT INTO evaluations VALUES(26663,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(26663,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(26663,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(26663,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(26663,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(26663,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(26663,36,'cp 78','e7e6 b1c3 g8f6 e1g1 f8e7 d1e2 e8g8 f1d1 f6e4 c1d2 e4d2 d1d2 e7b4 a1c1 b4c3 c1c3 c6e7 e3e4 f5g4 h2h3 g4f3 c3f3 d8d6 h3h4 a8d8 f3d3 e7c6 g2g3 d6b4 b2b3 b4b6 d4d5 c6e5 d3e3 e6d5 c4d5 g8h8');
INSERT INTO evaluations VALUES(26857,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(26857,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(26857,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(26857,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(26857,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(26857,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(26857,36,'cp 82','e1g1 g8f6 b1c3 f8e7 d1e2 e8g8 f1d1 f6e4 c1d2 e7d6 a1c1 e4d2 f3d2 c6e7 a2a3 f5g6 c4d3 g6d3 e2d3 c7c6 g2g3 e7d5 c3e4 d6e7 d2c4 a7a5 d3e2 d8c7 c1c2 h7h6 e4c5 a8b8 c5d3 d5f6 d1c1 f8d8');
INSERT INTO evaluations VALUES(27066,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(27066,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27066,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(27066,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(27066,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(27066,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(27066,36,'cp 47','c4e6 f7e6 d2d3 e5f4 c1f4 d8f6 f4c1 f6g6 e1g1 f8e7 b1c3 g8f6 c3e2 f6h5 f3d4 c6e5 c2c3 e8c8 e2f4 h5f4 c1f4 h8f8 d1e2 e5d7 b2b4 a7a6 a2a4 d8e8 f4g3 f8f1 a1f1 e7f6 g3f2 f6e5 h2h3 e8f8 f1b1 h7h5 a4a5 g6h6 b4b5 a6b5 d4b5 d7c5');
INSERT INTO evaluations VALUES(27050,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(27050,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(27050,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(27050,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(27050,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(27050,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(27050,36,'cp 44','f7e6 d2d3 e5f4 c1f4 g8f6 b1c3 d8d7 d3d4 e8c8 e1g1 e6e5 f4g5 e5d4 g5f6 g7f6 f3d4 c6d4 d1d4 c8b8 f1f5 c7c6 a1f1 f8e7 d4d3 h8g8 g2g3 g8g5 g1g2 a7a6 a2a4 g5g4 b2b3 g4g5 c3e2 d8e8 h2h4 g5f5 f1f5 e7d8 e2f4 d7e7 f4h5 e7e4 d3e4 e8e4 h5f6 e4e2 g2f3 e2c2 f6h7');
INSERT INTO evaluations VALUES(27134,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(27134,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(27134,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(27134,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(27134,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(27134,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(27134,36,'cp 36','d2d3 e5f4 c1f4 d8f6 f4d2 f8e7 b1c3 f6g6 e1g1 g8f6 c3e2 f6h5 c2c3 e8c8 e2d4 c6e5 d1e2 h8f8 b2b4 e5f3 f1f3 f8f3 e2f3 e6e5 d4f5 e7f6 g2g4 h5f4 d2f4 e5f4 h2h4 g6e8 d3d4 g7g5 a2a4 e8e6 h4g5 f6g5 d4d5 e6e5');
INSERT INTO evaluations VALUES(29622,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(29622,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(29622,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(29622,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(29622,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(29622,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(29622,36,'cp 53','g2g4 h5g6 f3e5 f6d7 e5g6 h7g6 f1g2 c7c6 d1e2 e7e6 c2c4 f8e7 e1g1 e8g8 b1c3 a7a5 f1d1 d7b6 c4d5 e6d5 e3e4 d5e4 g2e4 b8d7 d4d5 d8c7 e4g2 e7d6 c1e3 f8e8 a1c1 d7f6 d5c6 b7c6 e2f3 b6c4 f3c6 d6h2 g1h1 c7c6 g2c6');
INSERT INTO evaluations VALUES(30508,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(30508,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(30508,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(30508,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(30508,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(30508,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(30508,36,'cp 63','a7a6 c1e3 e7e6 f2f3 d8c7 d1d2 h7h5 e1c1 g8f6 c1b1 f8e7 f1d3 b8c6 d4c6 d7c6 e4e5 f6d5 c3d5 c6d5 e3d4 e7c5 c2c3 e8c8 d2e2 a6a5 d1c1 h5h4 f3f4 c8b8 a2a3 c7d7 g2g4 c5d4 c3d4 d8c8 f4f5 c8c1 h1c1');
INSERT INTO evaluations VALUES(32144,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(32144,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(32144,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(32144,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(32144,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(32144,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(32144,36,'cp 51','g8f6 h2h3 g4h5 d2d3 f8e7 g3g4 h5g6 h3h4 h7h6 h4h5 g6h7 g4g5 h6g5 c1g5 d6d5 f3e5 f6e4 g5f4 e7g5 d1g4 g5f4 g4f4 d8f6 f4f6 e4f6 b1c3 c7c6 h5h6 e8f8 e1d2 b8d7 e5d7 f6d7 h6g7 f8g7 g2h3 d7e5 c3e2 h7g6 f2f4 e5f3 d2e3 f3h4 a1g1');
INSERT INTO evaluations VALUES(29951,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(29951,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(29951,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(29951,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(29951,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(29951,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(29951,36,'cp 53','h2h3 g4h5 d2d3 f8e7 b1c3 h5g6 f3h4 c7c6 e1g1 e8g8 a2a4 f8e8 a4a5 b8d7 f2f4 f6h5 h4f5 g6f5 d1h5 f5g6 h5f3 f7f6 c1e3 g6f7 f3f2 e5f4 f2f4 a7a6 f4f2 e7f8 g3g4 f7g6 b2b3 g6f7 f2d2 a8c8');
INSERT INTO evaluations VALUES(33084,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(33084,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(33084,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(33084,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(33084,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(33084,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(33084,36,'cp 23','b8c6 c2c3 a7a6 b2b4 c5a7 a2a4 f6g6 b1d2 g8f6 f1e1 e8g8 d2f1 c8e6 c4a2 f6h5 g2g3 g8h8 f3h4 g6f6 d1f3 f6f3 h4f3 e6a2 a1a2 f7f5 e4f5 f8f5 f1d2 h5f6 g1g2 h8g8 b4b5 c6e7 d2e4 e7d5 c1d2 a8f8 b5a6 b7a6');
INSERT INTO evaluations VALUES(191,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(191,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(191,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(191,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(191,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(191,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(191,36,'cp 25','c2c3 a7a6 a2a4 c5a7 b2b4 f6g6 b1d2 g8f6 f1e1 e8g8 d2f1 c8e6 c4a2 g8h8 f1g3 e6g4 c1e3 a7e3 e1e3 f6h5 g3h5 g4h5 d1e1 c6d8 a4a5 d8e6 a2e6 g6e6 h2h3 f7f5 e4f5 f8f5 f3d2 e6d5 f2f3 f5f4 c3c4 d5d4 d2e4');
INSERT INTO evaluations VALUES(33145,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(33145,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(33145,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(33145,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(33145,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(33145,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(33145,36,'cp 72','g8e7 e1g1 d7d6 d3d4 c5b6 a2a4 a7a5 c1e3 e8g8 d4e5 d6e5 e3b6 c7b6 h2h3 c8e6 d1e2 e7g6 c4e6 f7e6 e2e3 f8f7 a1d1 g6f4 g1h2 b8c6 d1d6 a8f8 c3d5 f6g6 d5f4 f7f4 f3d2 g6g5 f2f3 b6b5 c2c3 b5a4 f1a1 f8d8 d6d8 g5d8 a1a4 d8c7 d2c4');
INSERT INTO evaluations VALUES(33471,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(33471,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(33471,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(33471,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(33471,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(33471,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(33471,36,'cp 172','b8c6 f1c4 c6d4 c4b3 a7a5 g1f3 d4b3 a2b3 c7c6 d2d3 d7d5 e4d5 f6d5 c3d5 c6d5 c1d2 c8g4 e1e2 e8d7 h2h3 g4f3 e2f3 h8c8 c2c3 b7b5 h1e1 a5a4 b3a4 a8a4 a1a4 b5a4 e1a1 c8a8 f3e2 h7h5 d2e3 h5h4 e3d4 f7f6 d4e3 a8a5 e2d2 e7d6');
INSERT INTO evaluations VALUES(559,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(559,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(559,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(559,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(559,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(559,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(559,36,'cp 185','f1c4 c6d4 c4b3 a7a5 g1f3 d4b3 a2b3 c7c6 d2d3 d7d5 e4d5 f6d5 c3d5 c6d5 c1d2 c8g4 f3e5 g4h5 a1a5 e8g8 e1g1 a8a5 d2a5 f7f6 e5f3 f8a8 a5b6 e7d6 f3h4 h5g4 h2h3 g4d7 h4f3 g7g5 f1b1 h7h5 b6e3 g8f7 f3d4 a8a2 g1f1 d6e5 c2c3 e5d4 e3d4');
INSERT INTO evaluations VALUES(39116,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(39116,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(39116,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(39116,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(39116,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(39116,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(39116,36,'cp 51','c7c6 e4e5 d6c7 e1g1 d7d5 e5d6 c7d6 d2d4 c8f5 d1e2 g8e7 f3h4 e8g8 c1f4 d6f4 f1f4 e7g6 f4f5 g6h4 f5f2 b8d7 c2c3 b7b5 c4b3 d7f6 b1d2 f8e8 e2d3 h4g6 f2f5 d8e7 a2a4 b5b4 a1f1 b4c3 b2c3 a8d8 d2f3 h7h6 g2g3 g8h8');
INSERT INTO evaluations VALUES(39863,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(39863,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(39863,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(39863,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(39863,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(39863,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(39863,36,'cp 188','e4e5 f6e4 a2a3 e4c3 d1g4 b4f8 b2c3 d7d6 e5d6 e6e5 g4g3 d8d6 f1d3 c8e6 g1f3 b8d7 e1g1 e8c8 f1e1 f7f6 a3a4 g7g6 a4a5 a7a6 c1e3 f8e7 d3f1 d6c7 d4e5 c6c5 e5f6 c7g3 h2g3 e7f6 f3g5 e6f5 e3f4 f6c3 g5f7 c3a1 e1a1 d7b8 f7d6 d8d6 f4d6 b8c6 d6c5 h8d8 c5b6 d8d2 f2f3 h7h5 a1e1 f5d7 e1e2 d2e2 f1e2 c6b4');
INSERT INTO evaluations VALUES(42859,30,'cp 299',NULL);
INSERT INTO evaluations VALUES(42859,31,'cp 286',NULL);
INSERT INTO evaluations VALUES(42859,32,'cp 280',NULL);
INSERT INTO evaluations VALUES(42859,33,'cp 314',NULL);
INSERT INTO evaluations VALUES(42859,34,'cp 283',NULL);
INSERT INTO evaluations VALUES(42859,35,'cp 313',NULL);
INSERT INTO evaluations VALUES(42859,36,'cp 310','d1d5 d8h4 g2g3 c5f2 e1f1 h4g4 b1d2 e4d2 c1d2 f2g1 d5f7 e8d8 h2h3 g4g6 f7g6 h7g6 f1g1 b8c6 a1e1 c6d4 e1f1 d7d6 e5d6 c7d6 d2a5 b7b6 a5c3 c8e6 c4d3 d4c6 c3g7 h8g8 g7c3 d8c7 f1f6 e6a2 b2b3 a7a5 f6f7 c7b8 g1g2 a5a4 d3e4 a8a7 f7f6 b8c7 h1d1 g8d8 f6g6 a4b3 c3f6 d8d7 c2b3 a2b3');
INSERT INTO evaluations VALUES(42886,30,'cp 305',NULL);
INSERT INTO evaluations VALUES(42886,31,'cp 307',NULL);
INSERT INTO evaluations VALUES(42886,32,'cp 298',NULL);
INSERT INTO evaluations VALUES(42886,33,'cp 297',NULL);
INSERT INTO evaluations VALUES(42886,34,'cp 309',NULL);
INSERT INTO evaluations VALUES(42886,35,'cp 297',NULL);
INSERT INTO evaluations VALUES(42886,36,'cp 320','d8h4 g2g3 c5f2 e1f1 h4g4 b1d2 e4d2 c1d2 f2g1 d5f7 e8d8 h2h3 g4g6 f7g6 h7g6 f1g1 b8c6 a1e1 c6d4 e1f1 d7d5 e5d6 c7d6 d2a5 b7b6 a5c3 c8e6 c4d3 d4c6 c3g7 h8g8 g7c3 d8c7 f1f6 e6d5 h1h2 d5a2 b2b3 a8e8 f6f7 c7b8 h2f2 c6e5 c3e5 e8e5 f2f6 e5d5 f6g6 g8g6 f7f8 b8c7 d3g6 d5d2');
INSERT INTO evaluations VALUES(43589,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43589,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43589,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43589,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43589,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43589,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43589,36,'cp 0','c8f5 f1e2 e7e6 b1d2 f8d6 f4d6 d8d6 g1f3 c7c5 c3c4 b8c6 d4c5 d6c5 c4d5 c5d5 e1g1 e8g8 d1b3 d5d6 d2c4 d6e7 f1d1 a8d8 d1d8 f8d8 a1d1 d8d5 c4b6 d5c5 b3a3 e7f8 b6c4 f5e4 c4d6 c5c2 d1d2 c2c1 d2d1');
INSERT INTO evaluations VALUES(28562,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28562,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28562,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28562,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(28562,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28562,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(28562,36,'cp 12','b1d2 e7e6 f1e2 f8d6 f4d6 d8d6 g1f3 c7c5 c3c4 b8c6 d4c5 d6c5 c4d5 c5d5 e1g1 e8g8 a2a3 f8d8 d2b3 f5e4 d1d5 e4d5 b3c5 f6d7 f1c1 d7c5 c1c5 d8d7 a1c1 a8d8 b2b4 g8f8 b4b5 a6b5 e2b5 f8e7 g1f1 d5f3 g2f3');
INSERT INTO evaluations VALUES(50370,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(50370,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(50370,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(50370,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(50370,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(50370,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(50370,36,'cp -10','d7d5 g1f3 e7e6 d2d4 f8d6 f1d3 b8d7 b1d2 e8g8 e1g1 f8e8 c3c4 e6e5 c4c5 d6f8 f3e5 d7e5 d4e5 f6g4 d1c2 g4e5 d3h7 g8h8 h7f5 c8f5 c2f5 f8c5 d2b3 c5b6 c1d2 h8g8 d2c3 d8e7 a1d1 c7c6 b3d4 a8d8 f5h3 b6d4 d1d4 e5g6 h3g3 c6c5');
INSERT INTO evaluations VALUES(47264,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(47264,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(47264,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(47264,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(47264,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(47264,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(47264,36,'cp -10','g1f3 c8f5 f1e2 h7h6 e1g1 b8d7 d2d4 e7e6 c3c4 c7c6 e2d3 d5c4 d3f5 e6f5 d1c2 g7g6 a2a4 a7a5 b1d2 d7b6 f3e5 f8d6 d2c4 b6c4 e5c4 e8g8 c2b3 d6b4 f1d1 d8e7 c1d2 f6d5 d2b4 d5b4 a1c1 g8g7 c4e5 h6h5 e5c4 f8e8 c4e5 b4d5');
INSERT INTO evaluations VALUES(52483,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(52483,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(52483,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(52483,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(52483,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(52483,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(52483,36,'cp -68','g8f6 d2d4 d7d5 f3d1 b8c6 g1f3 c8f5 f1d3 f5d3 d1d3 c5c4 d3e2 e7e6 e1g1 b7b5 a2a3 f8e7 b1c3 d8b6 a1b1 e8g8 b2b3 b6a6 b3c4 b5c4 e3e4 f8b8 b1b8 a8b8 e4d5 e6d5 c1g5 a6a5 e2d2 h7h6 g5f4 b8b7 f3e5 a5b6 e5c6 b6c6');
INSERT INTO evaluations VALUES(56734,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(56734,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(56734,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(56734,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(56734,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(56734,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(56734,36,'cp 217','d2d4 c8f5 b1c3 d8d7 f1c4 g8f6 d4e5 d6e5 f3e5 d7d1 e1d1 b8c6 e5c6 b7c6 h1e1 e8d7 h2h3 a8d8 e1e2 f8b4 c1d2 h8e8 c4d3 f5d3 c2d3 e8e2 d1e2 d8b8 d2e3 b4d6 b2b3 b8e8 g2g3 d6a3 a1f1 d7c8 c3e4 f6e4 d3e4 e8e4 e2f3 e4e8 e3a7 a3d6');
INSERT INTO evaluations VALUES(56730,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(56730,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(56730,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(56730,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(56730,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(56730,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(56730,36,'cp 191','c8f5 b1c3 d8d7 f1c4 b8c6 e1g1 g8f6 c4b5 a7a6 b5c6 b7c6 d4e5 d6e5 f3e5 d7d1 f1d1 f8d6 e5c6 f5c2 d1e1 e8f7 c6e5 d6e5 e1e5 h8e8 e5e8 a8e8 c1e3 c7c6 a1c1 c2d3 c1d1 d3c2 d1d2 c2f5 d2d4 f6e4 c3a4 e4f6 a4c5 f6d5 e3d2 d5c7 h2h3 c7b5 d4a4 e8e2 d2e3 b5c7 g2g4');
INSERT INTO evaluations VALUES(32436,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(32436,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(32436,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(32436,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(32436,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(32436,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(32436,36,'cp 206','b1c3 d8d7 f1c4 b8c6 e1g1 g8f6 c4b5 a7a6 b5c6 b7c6 d4e5 d6e5 f3e5 d7d1 f1d1 f8d6 e5c6 f5c2 d1e1 e8f7 c6e5 d6e5 e1e5 h8e8 e5e8 a8e8 c1e3 c7c6 a1c1 c2f5 c1d1 e8b8 b2b3 b8e8 h2h3 f5g6 g1h2 e8e6 d1d4 g6c2 d4d2 c2g6 c3a4 f6d5 e3d4 e6e8 a4c5');
INSERT INTO evaluations VALUES(3606,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(3606,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3606,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3606,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(3606,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3606,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(3606,36,'cp 0','a7a5 c3e2 c8e6 c4e6 f7e6 e2g3 d6d5 c2c3 d8d7 c1d2 d7f7 e4d5 e6d5 d3d4 e5d4 c3d4 c5b6 d2c3 f6h7 a1c1 f7g6 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(4383,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4383,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(4383,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(4383,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(4383,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4383,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(4383,36,'cp 15','c3a4 c5b6 a4b6 a7b6 c2c4 d7d6 e1g1 h7h6 a2a4 f6d7 c1d2 d7c5 b2b4 c5e6 f1e1 f7f5 e4f5 f8f5 b4b5 c6d4 f3d4 e6d4 e2g4 f5f8 g4c8 a8c8 d2e3 d4f5 d1g4 d8h4 g4h4 f5h4 a4a5 b6a5 b5b6 c7b6 e3b6 c8a8 b6a5 h4g6');
INSERT INTO evaluations VALUES(4451,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4451,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4451,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4451,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4451,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4451,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4451,36,'cp 0','d7d6 a2a4 c6e7 h2h3 e7g6 d1e1 c7c6 c4b3 f8e8 f3g5 e8f8');
INSERT INTO evaluations VALUES(3885,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3885,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3885,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(3885,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3885,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(3885,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(3885,36,'cp 2','a2a4 c6e7 b1c3 e7g6 d3d4 c7c6 d1d3 d8a5 h2h3 e5d4 e3d4 c8e6 c4e6 f7e6 a1e1 a5b6 c3e2 a8e8 g1h2 b6c7 e2g3 e6e5 d3c4 g8h8 c4c3 a7a6 a4a5 h8g8 c3c4 g8h8 d4d5 g6f4 d5c6 c7c6');
INSERT INTO evaluations VALUES(5318,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(5318,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5318,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5318,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(5318,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(5318,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(5318,36,'cp 0','a7a6 a2a4 d7d6 d2d3 a8b8 c1e3 g8f6 g1e2 f6g4 e3c1 c6b4 c3d5 b4c6');
INSERT INTO evaluations VALUES(21747,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(21747,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(21747,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(21747,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(21747,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(21747,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(21747,36,'cp -39','d2d4 e5e4 f3d2 f7f5 c3c4 g8f6 b1c3 c8e6 a2a3 d8d7 b2b4 h7h5 d1a4 h5h4 h2h3 g7g5 f1e2 f8g7 d2b3 c6e7 b3c5 d7a4 c3a4 e6c8 b4b5 b7b6 c5a6 c8a6 b5a6 d5c4 a4c3 c7c6 e2c4 e8d7 c1d2 a8c8 c4b3 g5g4 c3e2 f6d5 h3g4 f5g4');
INSERT INTO evaluations VALUES(21896,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(21896,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(21896,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(21896,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(21896,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(21896,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(21896,36,'cp -62','f1c4 g7g6 h5g5 g8f6 d2d3 f8g7 b1c3 h7h6 g5d2 c8g4 f3g1 d8d7 f2f3 g4e6 c4e6 d7e6 g1e2 d6d5 e1g1 e8g8 g1h1 a8d8 b2b3 a7a6 a2a4 f8e8 c1b2 e6d7 a1e1 d5e4 f3e4 b7b5 e2g1 c6d4 b2a3 f6h5 g1f3 d4e6 c3d5 c7c6 d5e3');
INSERT INTO evaluations VALUES(26048,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(26048,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(26048,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(26048,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(26048,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(26048,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(26048,36,'cp 147','d4d5 c6e5 f3e5 d7b5 c2c4 b5d7 e5d7 d8d7 e1g1 g7g6 b1c3 f8g7 f1e1 g8f6 f2f4 d7g4 d1g4 f6g4 e4e5 g4h6 e5d6 h6f5 c3e4 e7e6 d5e6 f7e6 g1f1 b7b6 a1b1 e8c8 b2b4 c5b4 e1d1 g7d4 b1b4 e6e5 b4a4 f5d6 e4d6 d8d6 a4a7 h8d8 a7h7');
INSERT INTO evaluations VALUES(26062,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(26062,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(26062,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(26062,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(26062,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(26062,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(26062,36,'cp 125','c6e5 f3e5 d7b5 c2c4 b5d7 e5d7 d8d7 b1c3 g7g6 e1g1 f8g7 d1e2 g8h6 c1g5 e8g8 e2d2 h6g4 h2h3 g4e5 b2b3 f7f5 a1e1 e5f7 g5e3 f5e4 c3e4 a8e8 d2d3 f7e5 d3e2 e5f7 e3d2 e7e6 e2d3 b7b6 d5e6 e8e6');
INSERT INTO evaluations VALUES(26078,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(26078,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(26078,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(26078,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(26078,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(26078,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(26078,36,'cp 134','b5d7 e5d7 c2c4 g7g6 a2a4 g8f6 b1c3 f8g7 h2h3 e8g8 e1g1 f8e8 c1e3 f6h5 d1c2 d7e5 f3e5 g7e5 e3g5 f7f6 g5d2 g8h8 g2g3 d8d7 g1h2 g6g5 a4a5 h5g7 a1e1 b7b6 a5a6 e8g8 c2a4 d7e8 a4e8 a8e8');
INSERT INTO evaluations VALUES(28754,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(28754,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(28754,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(28754,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(28754,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(28754,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(28754,36,'cp 38','c2c4 f8b4 c1d2 b4d2 d1d2 e8g8 e1g1 d7d5 c4d5 e6d5 b2b4 c7c6 b1c3 b8d7 f1c1 f8e8 d2b2 a7a5 a2a3 d8e7 f3d2 h7h6 e2e3 e8c8 b2b3 a5b4 a3b4 b6b5 e3e4 d7b6 e4e5 f6d7 f2f4 c8e8 c1e1 b6c4 a1a8 e8a8 d2c4 b5c4');
INSERT INTO evaluations VALUES(37296,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(37296,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(37296,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(37296,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(37296,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(37296,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(37296,36,'cp 26','e2e3 b7b6 f1e2 c8b7 e1g1 c7c5 d1c2 f6e4 g5e7 d8e7 a1d1 d7d6 c3e4 b7e4 c2d2 f8d8 f3e1 e4b7 e2f3 b7f3 e1f3 b8d7 d4c5 d7c5 f3d4 a8c8 d2e2 h7h6 b2b4 c5e4 d1c1 e7f7 f2f3 e4f6 f1d1 a7a5 b4b5 g8h8 d4c6 d8d7 e2c2 d6d5 c4d5 d7d5 d1d5');
INSERT INTO evaluations VALUES(37295,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37295,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37295,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(37295,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(37295,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(37295,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(37295,36,'cp 12','b7b6 f1d3 c8b7 e1g1 c7c5 d4d5 f6d5 c4d5 e7g5 d5e6 g5f6 d1b3 g8h8 a1d1 f6c3 b2c3 d8e7 d3c4 b7f3 g2f3 d7e6 c4e6 b8c6 e6d5 c6a5 b3a4 e7g5 g1h1 a8d8 f3f4 g5h4 d5g2 d8d1 a4d1 f8d8 d1c2 g7g6 c3c4 d8d6 g2d5 a5c6 a2a3 c6e7 f1d1 h4f6');
INSERT INTO evaluations VALUES(40366,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(40366,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(40366,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(40366,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(40366,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(40366,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(40366,36,'cp 84','e7e6 h2h3 e8g8 g1f3 e6d5 e4d5 f8e8 c1e3 g7h6 e1g1 h6e3 f2e3 d8e7 e3e4 b8d7 d1d2 d7e5 f3e5 e7e5 f1f3 e8f8 d2f4 e5f4 f3f4 g8g7 a1f1 f6e8 g2g4 h7h6 g1f2 c8d7 f2e3 f7f6 e3d2 e8c7 f4f2 a8b8 c3d1 b8e8 d1e3 e8e5 h3h4 c7a6 f2f3 h6h5 g4h5');
INSERT INTO evaluations VALUES(42608,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(42608,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(42608,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(42608,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(42608,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(42608,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(42608,36,'cp 70','e5d4 e3d4 c5e7 d4d5 e8g8 f1d3 e7c5 g1e2 a7a6 e1g1 d7d6 c1g5 h7h6 g5h4 b8d7 c3e4 g7g5 e4c5 d7c5 h4g3 c5d3 d1d3 f6h5 f2f3 d8f6 g3f2 f6g6 d3c3 c8d7 a1e1 a8e8 b2b3 b7b6 c3d2 f7f5 e2d4 f5f4 e1e8 f8e8 f1e1');
INSERT INTO evaluations VALUES(42677,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(42677,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(42677,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(42677,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(42677,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(42677,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(42677,36,'cp 50','e3d4 c5e7 d4d5 a7a5 f1d3 e8g8 g1e2 e7c5 e1g1 d7d6 b2b3 b8d7 d3c2 f8e8 h2h3 d7e5 e2g3 c8d7 g3e4 f6e4 c3e4 c5b6 d1h5 d8e7 c1b2 f7f5 e4g5 h7h6 g5e6 d7e6 d5e6 e7e6 c2f5 e6f7 h5f7 g8f7 f5c2 a5a4');
INSERT INTO evaluations VALUES(42676,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(42676,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(42676,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(42676,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(42676,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(42676,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(42676,36,'cp 47','c5e7 d4d5 e8g8 f1d3 a7a5 g1e2 e7c5 e1g1 d7d6 d3c2 b8d7 h2h3 f8e8 e2g3 d7e5 b2b3 c8d7 a1b1 e5g6 a2a3 c5a7 d1d2 c7c6 c1b2 c6d5 c4d5 b7b5 b1e1 e8e1 f1e1 a5a4 b3a4 b5a4 g3e4 a8b8 c2a4 f6e4 e1e4 a7c5 e4e1 d8b6 a4d7 b6b2 d2b2 b8b2');
INSERT INTO evaluations VALUES(48155,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(48155,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(48155,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(48155,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(48155,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(48155,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(48155,36,'cp -56','b8d7 b5d3 e6e5 d4e5 d7e5 g1e2 f8c5 e1g1 e8g8 h2h3 f8e8 e2g3 c7c6 b2b3 e5d3 c2d3 d5d4 c3e4 f6e4 d3e4 d4e3 c1e3 c5e3 f2e3 d8a5 d1f3 c8e6 a1d1 h7h6 f3f2 a8d8 g3f5 d8d1 f1d1 a5b6 g1h2 e6f5 e4f5 e8e3');
INSERT INTO evaluations VALUES(48572,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(48572,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(48572,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(48572,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(48572,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(48572,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(48572,36,'cp 127','f8e7 c3e4 e8g8 e4f6 e7f6 d2d4 d8d5 c2c3 b8d7 e2b5 d7b6 f1e2 c8e6 e1g1 d5b5 e2b5 e6d5 f1e1 c7c6 b5d3 d5f3 g2f3 a7a5 a2a4 f8d8 c1e3 g7g6 f3f4 b6d5 d3f1 g8g7 e3d2 d5e7 f1g2 e7f5 g2h3 d8d7 h3f5 g6f5');
INSERT INTO evaluations VALUES(41058,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(41058,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(41058,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(41058,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(41058,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(41058,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(41058,36,'cp 120','c3e4 f6d5 e4c3 d5b4 d2d3 c7c5 a2a3 b4c6 c1g5 e8g8 g5e7 c6e7 e1c1 c8g4 e2e3 b8c6 e3c5 g4f3 g2f3 c6d4 f1h3 e7c6 h1g1 g7g6 g1e1 b7b6 c5c4 d8g5 c1b1 a8d8 f3f4 g5f4 h3g2 d8c8 g2c6 c8c6 e1e4 f4e4 c3e4 c6c4 d3c4 d4c6 c2c3');
INSERT INTO evaluations VALUES(41057,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(41057,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(41057,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(41057,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(41057,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(41057,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(41057,36,'cp 109','f6d5 e4c3 d5b4 d2d3 c7c5 a2a3 b4c6 c1g5 f7f6 g5f4 c6d4 f3d4 c5d4 c3e4 e8g8 e2d1 b8c6 f1e2 f6f5 e4g3 h7h6 e1g1 e7g5 f4d2 g5d2 d1d2 f5f4 g3e4 c6e5 f2f3 b7b6 a1e1 c8b7 d2b4 b7e4 d3e4 d8c7 e1d1 a8d8 d1d4');
INSERT INTO evaluations VALUES(48716,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(48716,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(48716,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(48716,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(48716,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(48716,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(48716,36,'cp -58','f8d6 g1f3 c7c5 d2d3 e8g8 e1g1 c8e6 c2c3 h7h6 d3d4 c5c4 b3c2 b8c6 f1e1 d8c7 b1d2 f8e8 d2f1 e6d7 f1e3 b7b5 a2a3 e8e6 e3f5 e6e1 f3e1 d7f5 c2f5 a8e8 c1e3 a7a5 d1f3 b5b4 a3b4 a5b4 e3d2 b4b3 g2g3 c6e7 e1g2 e7f5 f3f5');
INSERT INTO evaluations VALUES(41725,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(41725,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(41725,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(41725,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(41725,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(41725,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(41725,36,'cp -61','g1f3 e8g8 e1g1 c7c5 c2c3 c8f5 d2d4 c5c4 b3c2 f5c2 d1c2 b8d7 b2b3 a8c8 c1a3 d6a3 b1a3 d8a5 a3b1 c4b3 c2b2 b3a2 a1a2 a5b6 b2b6 a7b6 a2a7 c8b8 f3e5 f8c8 e5d7 f6d7 f1e1 g8f8 e1e3 d7f6 b1a3 b8a8 a7a8 c8a8 a3b5 f6e4 b5c7');
INSERT INTO evaluations VALUES(53550,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53550,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53550,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53550,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53550,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53550,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(53550,36,'cp 15','e6d5 d2d4 f8d6 f1d3 g8e7 c3e2 c8f5 d3f5 e7f5 e2g3 f5e7 e1g1 e8g8 f1e1 b8d7 c1g5 f7f6 d1e2 d7b6 g5d2 d8d7 b2b3 f8e8 c2c4 d5c4 b3c4 d6g3 h2g3 e7g6 e2d3 d7f7 c4c5 b6d5 d3b3 g6f8 a2a4 h7h6 a4a5 e8e1 a1e1');
INSERT INTO evaluations VALUES(56238,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(56238,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(56238,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(56238,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(56238,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(56238,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(56238,36,'cp -16','c8e6 e4d5 e6d5 c1f4 d8e7 d4e5 b8a6 f1b5 c7c6 b5a6 e7e5 f4e5 e8c8 a6e2 d8e8 g1f3 d5f3 e2f3 e8e5 e1f1 g8f6 b1d2 f8c5 d2c4 e5e6 b2b4 c5e7 g2g3 h8d8 f1g2 c8c7 a1d1 d8d1 h1d1 g7g6 c4d2 f6d7 a2a4 f7f5 d2c4 e7f6 d1d3 e6e1 a4a5 e1b1');
INSERT INTO evaluations VALUES(57006,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(57006,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(57006,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(57006,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(57006,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(57006,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(57006,36,'cp -24','d7d5 g1f3 d5c4 d1a4 c8d7 a4c4 c7c5 b2b3 g8f6 c1b2 b7b5 c4c2 b8c6 e2e3 f8d6 f1e2 e8g8 e1g1 a7a6 b1c3 d8c7 f1d1 f8d8 h2h3 a8c8 a1c1 d6e7 f3d2 c7b8 c3e4 f6e4 d2e4 c6b4 c2c3 f7f6');
INSERT INTO evaluations VALUES(55804,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(55804,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(55804,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55804,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(55804,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(55804,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(55804,36,'cp 0','d3d4 g8f6 g1f3 f8e7 b1c3 e8g8 c1g5 d5c4 e2e4 b7b5 g5f6 e7f6 c3b5 c8b7 e4e5 f6e7 f1c4 e7b4 b5c3 b7f3 d1f3 d8d4 e1g1 b8d7 f1d1 d4c4 d1d7 a8d8 a1d1 d8d7 d1d7 b4c3 f3c3 c4a2 h2h3 a2b1 g1h2 h7h6 d7c7 a7a5 c7a7 b1f5 c3d4 f8b8 a7a5 f5b1 a5c5 b1b2 d4b2 b8b2 c5c8 g8h7');
INSERT INTO evaluations VALUES(2572,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(2572,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(2572,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(2572,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(2572,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(2572,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(2572,36,'cp 52','h2h3 b7b5 a4c2 d6d5 d2d4 e5d4 e4e5 f6e4 c3d4 c6b4 c1e3 b4c2 d1c2 f7f5 e5f6 e7f6 b1d2 c8f5 a1c1 a8c8 c2c6 d8d6 d2b3 f5d7 c6c2 c7c6 b3c5 e4c5 d4c5 d6e6 e3d4 e6f7 f3e5 f6e5 e1e5 d7f5 c2d2 c8e8 c1e1 f7g6 e5e8 f8e8');
INSERT INTO evaluations VALUES(4195,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(4195,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(4195,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4195,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4195,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4195,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4195,36,'cp 25','d2d3 d7d6 c3a4 c5b6 f1e1 c8e6 c4e6 f7e6 c2c3 d8e8 b2b4 e8g6 a4b6 a7b6 a3a4 f6h5 h2h3 f8f6 b4b5 c6a5 c1e3 a8f8 f3h2 h5f4 e3f4 f6f4 a1a2 g6g5 c3c4 f4f7 g2g3 g5e7 e1e3 e7f6 e3f3 f6d8 h3h4 f7f3 h2f3 d8f6');
INSERT INTO evaluations VALUES(4716,30,'cp -314',NULL);
INSERT INTO evaluations VALUES(4716,31,'cp -296',NULL);
INSERT INTO evaluations VALUES(4716,32,'cp -326',NULL);
INSERT INTO evaluations VALUES(4716,33,'cp -299',NULL);
INSERT INTO evaluations VALUES(4716,34,'cp -295',NULL);
INSERT INTO evaluations VALUES(4716,35,'cp -305',NULL);
INSERT INTO evaluations VALUES(4716,36,'cp -314','d7d5 c1e3 c5e7 e4d5 f6d5 b1c3 f7g8 d1e2 c8e6 a1e1 e6f7 e2g4 d5c3 b2c3 e7f6 h2h3 h7h5 g4e4 d8d7 h3h4 f7d5 e4g6 d7g4 g6g4 h5g4 e3g5 g8f7 c3c4 d5e6 g5f6 g7f6 f2f4 g4f3 f1f3 e6g4 f3f2 f6f5 a2a3 a8d8 e1e4 f5f4 h4h5');
INSERT INTO evaluations VALUES(4700,30,'cp -310',NULL);
INSERT INTO evaluations VALUES(4700,31,'cp -306',NULL);
INSERT INTO evaluations VALUES(4700,32,'cp -296',NULL);
INSERT INTO evaluations VALUES(4700,33,'cp -294',NULL);
INSERT INTO evaluations VALUES(4700,34,'cp -311',NULL);
INSERT INTO evaluations VALUES(4700,35,'cp -299',NULL);
INSERT INTO evaluations VALUES(4700,36,'cp -309','c1e3 c5e7 e4d5 f6d5 d1e2 c8e6 b1c3 f7g8 a1e1 e6f7 h2h3 d5c3 b2c3 d8d7 e2d2 b7b5 f2f4 e5f4 f1f4 e7d6 f4f1 a7a5 e1e2 a8f8 e2f2 d6a3 f2f3 c6e5 f3f5 e5g6 d2f2 a5a4');
INSERT INTO evaluations VALUES(5437,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(5437,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(5437,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(5437,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(5437,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(5437,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(5437,36,'cp -51','g5f3 a7a6 c4b3 c6a5 e1g1 b7b5 b1d2 g8f6 f1e1 c8b7 c2c3 a5b3 a2b3 e8g8 d2f1 a6a5 f1g3 h7h6 h2h3 d8c7 c1d2 b7c6 c3c4 c7b6 d1e2 f8e8 d2c3 b5b4 c3d2 g8h7 d2e3 b6c7 e3d2 f6d7 d3d4 c5d4 f3d4');
INSERT INTO evaluations VALUES(5441,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(5441,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(5441,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(5441,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(5441,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(5441,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(5441,36,'cp -51','a7a6 c4b3 c6a5 b1d2 g8f6 e1g1 b7b5 f1e1 e8g8 d2f1 c8b7 c2c3 a5b3 a2b3 a6a5 f1g3 b7c6 d3d4 c5d4 f3d4 c6d7 g3h5 f6h5 d1h5 d8b6 c1e3 b6b7 e3g5 f7f6 g5d2 f8e8 h5f3 a5a4 b3b4 e7f8 h2h3 e6e5 d4f5 d7e6 d2e3');
INSERT INTO evaluations VALUES(7539,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(7539,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(7539,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(7539,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(7539,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(7539,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(7539,36,'cp 37','e8g8 h2h3 a7a5 b2b3 c8e6 d1e2 d6d5 c4d3 d5e4 c3e4 e6f5 c1b2 f8e8 b2f6 e7f6 a1d1 f5e4 d3e4 d8e7 d1e1 e7c5 e2f3 a8d8 e4d3 c5d5 e1e8 d8e8 f3f4 f6e5 f1e1 d5e6 f4e4 g7g6 g2g3 g8g7 e1e2 e5f6 g1g2 e6e4 e2e4 e8e4 d3e4');
INSERT INTO evaluations VALUES(8551,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8551,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(8551,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(8551,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8551,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(8551,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(8551,36,'cp 24','e8g8 h2h3 a7a6 b5c6 b7c6 d3d4 f6d7 c1e3 a6a5 d4e5 d7e5 f3e5 d6e5 c3a4 f7f5 d1d8 f8d8 f1d1 d8d1 a1d1 f5e4 e3c5 e7f6 a4c3 c8e6 b2b3 g8f7 c3e4 a5a4 f2f3 a4b3 a2b3 a8a2 d1d2 e6f5 g1f1 h7h5 f1f2 f7e6 f2e2 f5e4 f3e4');
INSERT INTO evaluations VALUES(3352,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3352,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(3352,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(3352,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(3352,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3352,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(3352,36,'cp 36','h2h3 h7h6 f1e1 c8d7 c1d2 f8e8 a2a4 e7f8 c3e2 a7a6 b5c4 c6e7 a4a5 e7g6 c4b3 c7c6 e2c3 a8b8 d3d4 e5d4 f3d4 e8e5 d1f3 e5a5 a1a5 d8a5 d2h6 g6e5 f3g3 f6h5 g3g5 h5f6 b3a2 b8e8 d4f5 d7f5 g5f5 g7h6 f5f6 f8g7');
INSERT INTO evaluations VALUES(3351,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3351,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3351,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(3351,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(3351,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(3351,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(3351,36,'cp 5','h7h6 f1e1 c8d7 c1d2 f8e8 a2a4 e7f8 a4a5 c6b4 b5c4 c7c6 c3a4 d6d5 c2c3 d5c4 c3b4 c6c5 a4c5 f8c5 b4c5 c4d3 d2c3 d7c6 f3e5 f6e4 d1d3 d8d3 e5d3 c6b5 a1d1 e4c3 b2c3 b5d3 e1e8 a8e8 d1d3 e8e1 g1h2 e1a1 d3d7 a1a5 d7b7 a5c5 b7a7 c5c3 a7a8 g8h7');
INSERT INTO evaluations VALUES(10107,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(10107,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(10107,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(10107,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(10107,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(10107,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(10107,36,'cp -101','e8g8 d2d3 d7d5 c4b3 a7a5 g1e2 a5a4 b3d5 f6d5 e4d5 c6e7 d3d4 c5b6 e1g1 e5d4 f1d1 d8d5 f3d5 e7d5 e2d4 f8e8 b1a3 b6d4 c3d4 c8f5 c1d2 f5d3 a1c1 f7f6 a3c2 d3c2 c1c2 c7c6 g1f1 g8f7 a2a3 h7h5 c2c4 e8d8 d1e1 g7g5');
INSERT INTO evaluations VALUES(4330,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(4330,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(4330,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(4330,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(4330,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(4330,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(4330,36,'cp -86','d2d3 d7d5 c4b3 a7a5 g1e2 a5a4 b3d5 f6d5 e4d5 c6e7 d3d4 c5b6 e1g1 e5d4 f1d1 f8e8 b1a3 d8d5 f3d5 e7d5 e2d4 b6d4 c3d4 c8f5 c1d2 f5d3 a1c1 f7f6 a3c4 e8e7 c4e3 d5f4 g1h2 e7d7 d2c3 a4a3 b2a3 a8a3 d1d2 c7c6 c3b4 a3a4 a2a3 d7d4 c1d1');
INSERT INTO evaluations VALUES(4336,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(4336,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(4336,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(4336,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(4336,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(4336,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(4336,36,'cp -84','d7d5 c4b3 a7a5 g1e2 a5a4 b3d5 f6d5 e4d5 c6e7 d3d4 c5b6 e1g1 e5d4 f1d1 d8d5 f3d5 e7d5 e2d4 f8d8 b1a3 c7c6 c3c4 d5b4 c1e3 b6c7 a3c2 b4c2 d4c2 c8f5 c2d4 f5g6 b2b4 a4b3 d4b3 d8d1 a1d1 g6f5 g2g4 f5c8 g1g2 a8a2');
INSERT INTO evaluations VALUES(10542,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(10542,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(10542,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(10542,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(10542,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(10542,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(10542,36,'cp 20','b2b4 a5b6 d2d3 d7d6 c3a4 e8g8 e1g1 c8e6 c4e6 f7e6 c2c3 c6e7 f1e1 e7g6 d3d4 e5d4 c3d4 d6d5 e4d5 d8d5 a4c5 a8e8 c1e3 b6c5 d4c5 c7c6 f3d2 a7a6 d1b3 d5b3 d2b3 f8f7 e3g5 f6d5');
INSERT INTO evaluations VALUES(10538,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(10538,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(10538,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(10538,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(10538,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(10538,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(10538,36,'cp 21','a5b6 d2d3 d7d6 c3a4 e8g8 e1g1 h7h6 f1e1 c8e6 c4e6 f7e6 c2c3 d8e8 h2h3 e8g6 a4b6 a7b6 a3a4 f8f7 f3h4 g6h5 d1h5 f6h5 h4f3 c6e7 f3h2 e7c6 c1e3 h5f4 e3f4 f7f4 g2g3 f4f7 e1b1 g8f8 h3h4 f8e7 h2g4 e7d7 g1g2 a8f8 b1d1 f7f3 h4h5 c6e7 d1d2 f3f7 c3c4');
INSERT INTO evaluations VALUES(9926,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(9926,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(9926,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(9926,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(9926,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(9926,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(9926,36,'cp 27','d2d3 d7d6 c3a4 e8g8 e1g1 c6e7 c1g5 e7g6 a4b6 a7b6 f3h4 g6f4 f1e1 f4e6 g5c1 f6e4 e1e4 f7f5 h4f5 f8f5 d1g4 d8f6 g4h4 f6h4 e4h4 g8h8 c4d5 f5f6 c1e3 h7h6 a3a4 g7g5 h4e4 e6f4 d5b3 c8e6 e3f4 g5f4 b3e6 f6e6 c2c4 e6e8');
INSERT INTO evaluations VALUES(20383,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(20383,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(20383,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(20383,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20383,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(20383,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(20383,36,'cp 26','g8f6 f1d3 f6h5 f4e3 f8d6 e1g1 h5f4 e3f4 d6f4 b1d2 f4d6 f1e1 e8g8 g2g3 c8d7 a2a3 g8h8 a1c1 a8c8 b2b4 d7e8 d1e2 c6e7 h2h4 e7g8 c3c4 d5c4 d2c4 d6b8 d3e4 d8c7 c4a5 c7c1 e1c1 c8c1 g1g2');
INSERT INTO evaluations VALUES(20982,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(20982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20982,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(20982,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(20982,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(20982,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(20982,36,'cp 3','c8f5 e2e3 e7e6 f1d3 f5d3 d1d3 g8f6 g1e2 d8d7 c2c3 f8d6 b1d2 d6f4 e2f4 e6e5 f4e2 e8g8 d3c2 e5e4 e1g1 c6e7 c3c4 c7c6 f1c1 f8c8 a2a4 h7h6 c2b3 e7g6 c1c2 a7a5 c4d5 c6d5 c2c5 c8c5');
INSERT INTO evaluations VALUES(26544,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26544,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(26544,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(26544,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(26544,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(26544,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(26544,36,'cp 6','e2e3 e7e6 f1d3 f5d3 d1d3 f8d6 g1e2 g8f6 e1g1 e8g8 f4g3 c6b4 d3b3 b7b6 c2c4 c7c5 b1d2 c5d4 g3d6 d8d6 e2d4 b4c6 d4f3 f8c8 c4d5 d6d5 f1d1 d5b3 d2b3 c8c7 a1c1 a8c8 a2a3 h7h6 c1c2 a7a6');
INSERT INTO evaluations VALUES(26537,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26537,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26537,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26537,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26537,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(26537,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(26537,36,'cp 12','e7e6 f1d3 f5d3 d1d3 f8d6 g1e2 g8f6 e1g1 e8g8 f1d1 d6f4 e2f4 d8d6 c2c4 e6e5 f4d5 e5d4 e3e4 f8e8 b1d2 c6b4 d3f3 b4d5 c4d5 c7c6 d5c6 b7c6 a1c1 c6c5 f3a3 f6e4 d2e4 e8e4 c1c5');
INSERT INTO evaluations VALUES(21532,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(21532,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(21532,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(21532,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(21532,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(21532,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(21532,36,'cp 53','e5d4 e1g1 d5e4 f3d4 c8d7 b1c3 f7f5 b5c6 b7c6 f2f3 c6c5 d4b3 c5c4 b3d4 f8c5 c1e3 d8f6 b2b4 c4b3 a2b3 g8e7 f3e4 e8g8 d1d3 c5b6 c3a4 d7a4 a1a4 f6e5 d4f3 e5d6 d3c4 g8h8 e3b6 d6b6 g1h1 f5e4 c4e4 e7f5 f1e1 h7h6 f3d4 b6g6 d4f5 g6f5 e4c4 f5f7');
INSERT INTO evaluations VALUES(21861,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21861,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(21861,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21861,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(21861,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(21861,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21861,36,'cp 0','c8d7 g1f3 a7a6 b5c6 b7c6 b1c3 f7f5 e3e4 g8f6 d1e2 f5e4 c3e4 f6d5 e1g1 d5f4 e2e3 f4d5');
INSERT INTO evaluations VALUES(27220,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27220,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27220,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27220,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27220,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27220,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27220,36,'cp 0','g1f3 f7f5 b5e2 g7g5 d2d4 g5g4 d4e5 g4f3 e2f3 f8g7 f3c6 d7c6 d1h5 e8f8 h5f5 f8e8 f5h5 e8f8');
INSERT INTO evaluations VALUES(22531,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(22531,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(22531,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22531,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(22531,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(22531,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22531,36,'cp 39','d7d6 b2b4 c8g4 b1c3 g8f6 d1b3 d8d7 c3d5 c6e5 d5f6 g7f6 f3d4 e5c4 b3c4 c7c5 f2f3 g4h3 g2h3 h8g8 g1h1 c5d4 c4d4 d7h3 f1f2 a7a6 c1f4 a8c8 b4b5 a6b5 d4b6 h3d7 a1b1 b5b4 b6b4 d6d5 b4b7 d7b7 b1b7');
INSERT INTO evaluations VALUES(21301,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(21301,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(21301,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(21301,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(21301,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21301,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(21301,36,'cp 39','b2b4 c8g4 b1c3 g8f6 d1b3 d8d7 c3d5 c6e5 d5f6 g7f6 f3d4 e5c4 b3c4 c7c5 d4b5 g4h3 c1f4 h8g8 f4g3 h3e6 c4e2 a7a6 b5c3 c5b4 c3d5 e6d5 e4d5 f8e7 e2d2 g8g6 d2b4 e8f8 a1b1 b7b5 b1c1 f8g8 c1c6 d7f5 b4b3 e7f8 g3d6 g6g5 d6f8 a8f8 g2g3 f5d5 c6f6 d5b3');
INSERT INTO evaluations VALUES(22982,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(22982,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(22982,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(22982,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(22982,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(22982,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(22982,36,'cp -35','d1e2 f8b4 f4f5 g6e7 c1g5 h7h6 g5h4 b4c3 b2c3 c7c6 e1c1 d8a5 e2g4 h8h7 d5d6 e7d5 d1d5 c6d5 c1d2 h6h5 g4f4 b7b6 e5e6 e8f8 f1e2 c8a6 e2f3 a8c8 e6d7 a5c3 d2d1 c3c2 d1e1 g8f6 d7c8q a6c8 f4e5 f8g8 h4f6 g7f6 e5e8 g8g7 d6d7 c8d7 e8d7 c2c1 f3d1 c1e3 e1f1 e3d3 f1f2');
INSERT INTO evaluations VALUES(25501,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(25501,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25501,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(25501,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25501,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(25501,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(25501,36,'cp 36','f8e7 b5c6 d7c6 d1e2 e8g8 d2d4 e5d4 f3d4 c6d7 c1f4 f8e8 a1d1 c7c6 d4f3 b7b5 e4e5 e7f8 e2d3 d6e5 f3e5 d7e6 d3d8 e8d8 d1d8 a8d8 e5c6 d8c8 c6d4 e6d7 c3b5 d7b5 d4b5 c8c2 f1b1 f8c5 f4e3 c5e3 f2e3 a7a6 b5d6 h7h6 g1f1 f6d5 d6f5 c2d2 a2a4');
INSERT INTO evaluations VALUES(24195,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(24195,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(24195,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(24195,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(24195,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24195,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(24195,36,'cp 44','f1e1 e8g8 b5c6 d7c6 d2d4 f8e8 d1d3 e7f8 c1g5 h7h6 g5h4 g7g5 h4g3 f6h5 a1d1 f8g7 g3h2 h5f4 d3c4 e5d4 f3d4 d6d5 e4d5 e8e1 d1e1 c6d5 c3d5 d8d5 c4d5 f4d5 d4f5 b7b5 b2b3 g7f8 c2c4 b5c4 b3c4 d5c3 a2a3 f8a3 f5h6 g8f8 h2c7 a3b4 h6f5 a7a5 c7d6 b4d6 f5d6');
INSERT INTO evaluations VALUES(26299,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(26299,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(26299,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(26299,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(26299,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26299,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(26299,36,'cp 12','h2h3 g4h5 d1b3 d8d7 f3e5 c6e5 d4e5 a7a6 c3c4 d5c4 f1c4 e7e6 e1g1 b7b5 c4d3 a8d8 d3e4 f7f5 e5f6 g8f6 b1d2 f8c5 e4f3 h5g6 f1d1 f6d5 d2e4 g6e4 f3e4 e8g8 c1e3 c5e3 f2e3 g8h8 b3a3 d7b7 a3c5 h7h6 a1c1 b7b8 e4d5 d8d5 d1d5 e6d5 c1d1 f8e8 d1d5 b8g3 d5d2 g3e1 g1h2 e1e3 c5e3');
INSERT INTO evaluations VALUES(26494,30,'cp 390',NULL);
INSERT INTO evaluations VALUES(26494,31,'cp 405',NULL);
INSERT INTO evaluations VALUES(26494,32,'cp 400',NULL);
INSERT INTO evaluations VALUES(26494,33,'cp 421',NULL);
INSERT INTO evaluations VALUES(26494,34,'cp 397',NULL);
INSERT INTO evaluations VALUES(26494,35,'cp 415',NULL);
INSERT INTO evaluations VALUES(26494,36,'cp 406','d4d5 e6d5 c3d5 e7e6 d5c3 d8d1 e1d1 c6e5 g1f3 a8d8 c1d2 e5d3 f1d3 c4d3 a1c1 c7c5 c3e4 b7b6 c1c3 f8e7 c3a3 c5c4 a3a7 g8f6 e4c3 e7d6 a7a4 d8c8 h1f1 e8g8 e3e4 f6g4 h2h3 g4e5 f3e5 d6e5 a4b4 f8d8 f2f4 e5d4 c3b5 c8a8 b5d4 d8d4 a2a3');
INSERT INTO evaluations VALUES(26487,30,'cp 403',NULL);
INSERT INTO evaluations VALUES(26487,31,'cp 409',NULL);
INSERT INTO evaluations VALUES(26487,32,'cp 405',NULL);
INSERT INTO evaluations VALUES(26487,33,'cp 394',NULL);
INSERT INTO evaluations VALUES(26487,34,'cp 392',NULL);
INSERT INTO evaluations VALUES(26487,35,'cp 410',NULL);
INSERT INTO evaluations VALUES(26487,36,'cp 413','e6d5 c3d5 e7e6 d5c3 d8d1 e1d1 e8c8 c1d2 f8b4 d1e1 b7b5 g1f3 a7a6 a2a4 b4c3 d2c3 g8f6 f1e2 f6d5 a4b5 a6b5 f3d4 c6d4 c3d4 c8b7 d4g7 d8a8 a1d1 h8g8 g7d4 c7c6 e1f1 d5b4 g2g3 b4c2 d4c5 a8a1 d1a1 c2a1');
INSERT INTO evaluations VALUES(26665,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(26665,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(26665,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(26665,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(26665,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(26665,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(26665,36,'cp 248','e3e4 f5d7 g1f3 e7e6 f1c4 f8b4 e1g1 g8e7 a2a3 b4c3 b2c3 e8g8 f1e1 h7h6 h2h4 e6e5 d1d3 d7g4 f3h2 g4h5 c4a2 g8h8 d4d5 c6b8 g2g4 h5g6 f2f4 e5f4 c1f4 g6h7 d3g3 b8a6 h2f3 a8c8 f4e3 h7e4 e3d4 e4f3 g3f3 e7g6 h4h5 g6h4 f3g3 d8g5');
INSERT INTO evaluations VALUES(28165,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(28165,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(28165,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(28165,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(28165,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(28165,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(28165,36,'cp 54','c4d5 e6d5 d1b3 b8c6 c1g5 f7f6 g5d2 d8d7 a1c1 g8e7 e2e3 e8g8 h2h3 a7a5 f1e2 c6d8 a2a3 b4c3 b2c3 d8f7 c3c4 c7c6 e1g1 d5c4 e2c4 b7b5 c4d3 f5d3 b3d3 f7e5 f3e5 f6e5 d2c3 e5d4 c3d4 e7f5 d3b3 d7f7 b3f7 f8f7 c1c6 b5b4');
INSERT INTO evaluations VALUES(29971,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(29971,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(29971,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(29971,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(29971,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(29971,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(29971,36,'cp 145','b8c6 d1d3 d6d5 e1c1 d5e4 c3e4 f6d5 f4d2 f8e7 h2h3 g4h5 c2c4 d5b4 d3e3 b4a2 c1b1 a2b4 d4d5 b4d5 c4d5 d8d5 f1d3 e8c8 d3c2 h5g6 d2c3 d5e6 h1e1 d8d1 e1d1 h8e8 d1e1 c8b8 f3d4 c6d4 e3d4 e6c6 e1e3 f7f5 e4f2 e7f6 e3e8 g6e8 d4d3 e8g6 c3f6 c6f6 d3e3');
INSERT INTO evaluations VALUES(25358,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(25358,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(25358,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(25358,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(25358,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(25358,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(25358,36,'cp 148','d1d3 d6d5 e1c1 d5e4 c3e4 f6d5 f4d2 f8e7 h2h3 g4h5 c2c4 d5b4 d3e3 b4a2 c1b1 a2b4 d4d5 b4d5 c4d5 d8d5 e3d3 e8c8 d3d5 d8d5 f1c4 d5d8 d2c3 e7b4 c3b4 c6b4 e4g5 d8d1 h1d1 h8f8 d1e1 h7h6 g5e4 f8d8 f3h4 d8d4 c4b3 b7b6 g2g4 d4d3 e4c3 d3h3 g4h5 h3h4 b3f7 b4d3 e1e8 c8d7 e8g8 d3e5 f7b3 h4h5 g8g7 d7d8');
INSERT INTO evaluations VALUES(30712,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(30712,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(30712,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(30712,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(30712,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(30712,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(30712,36,'cp 62','e4d5 g8f6 b1c3 f8b4 g1e2 d8e7 c1e3 f6g4 e3d4 e8g8 a2a3 b4c3 d4c3 b8a6 d1d4 f7f5 h2h3 g4f6 e1c1 a6c5 c1c2 c7c6 f2f3 c6d5 e2f4 e7f7 c4d5 g7g5 d4f6 f7f6 c3f6 f8f6 f4e2 a8c8 e2c3 c5a4 c2b3 a4c3 b2c3 d7d6 c3c4 f6f7 f1d3 b7a6 h1e1 g8f8 b3c3');
INSERT INTO evaluations VALUES(32847,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(32847,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(32847,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(32847,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(32847,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(32847,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(32847,36,'cp 91','b1c3 d5d8 d2d4 e7e6 f1d3 f5d3 d1d3 c7c6 c1d2 f8b4 e1c1 g8f6 h2h4 d8a5 a2a3 b4c3 d2c3 a5b5 h4h5 b5d3 d1d3 b8d7 f3e5 f6e4 c3e1 a8d8 e5c4 d7b6 c4a5 e4d6 f2f3 b6c4 a5c4 d6c4 a3a4 c4d6 b2b3 d6f5 e1f2 f5e7 c2c4 b7b5 h5h6 b5c4 h6g7 h8g8 b3c4');
INSERT INTO evaluations VALUES(32846,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(32846,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(32846,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(32846,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32846,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(32846,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(32846,36,'cp 76','d5d8 d2d4 e7e6 f1d3 f5d3 d1d3 c7c6 c1g5 d8c7 e1c1 b8d7 f3e5 f8d6 c3e4 g8f6 e4d6 c7d6 h1e1 d7e5 d4e5 d6d3 d1d3 f6d5 c2c4 h7h6 g5d2 d5b6 b2b3 b6d7 d3g3 g7g5 d2c3 e8c8 g3f3 h8h7 e1d1 c6c5 a2a3 g5g4 f3d3 h7h8 c1c2 d7b8 b3b4 d8d3 d1d3 c5b4 a3b4');
INSERT INTO evaluations VALUES(34456,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(34456,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(34456,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(34456,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(34456,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(34456,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(34456,36,'cp 88','c8e6 c4b3 b8d7 h2h3 e6b3 a2b3 f6d8 g2g4 f8e7 c1e3 h7h6 d3d4 g8f6 d1e2 b7b5 h1g1 a7a5 h3h4 h6h5 g4h5 f6h5 e1c1 d8c7 c1b1 g7g6 d4e5 d6e5 f3g5 e7b4 e3d2 d7f6 f2f4 b4c3 d2c3 h5f4 e2f2 a5a4 b3b4');
INSERT INTO evaluations VALUES(36554,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(36554,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(36554,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(36554,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36554,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(36554,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(36554,36,'cp 31','b1c3 b8d7 e1g1 e7e5 e3e4 f8e8 d4d5 a7a5 f3e1 d7c5 f2f3 f6h5 e1d3 c5d3 e2d3 e8f8 g1h1 c8d7 c1e3 b7b6 g2g3 f7f5 e4f5 g6f5 e3d2 g8h8 d1e2 d8f6 c3b5 d7b5 c4b5 f6f7 a1e1 f7d5 d3c4 d5c5');
INSERT INTO evaluations VALUES(36560,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(36560,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(36560,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(36560,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(36560,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36560,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(36560,36,'cp 62','b8a6 g1f3 e7e5 e1g1 d8e7 d1c2 e5d4 f3d4 a6b4 c2d2 c7c5 d4b5 f6e4 c3e4 e7e4 a1d1 e4c2 e2f3 c2b2 d2b2 g7b2 b5d6 b2d4 e3d4 c5d4 d1d4 b4c6 d4d2 a8b8 c4c5 f8d8 h2h4 c8e6 h4h5 g8f8 f1b1 f8e7 f3e4 f7f5 e4d5 e6d5 d2d5 e7e6');
INSERT INTO evaluations VALUES(49494,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49494,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(49494,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(49494,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(49494,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49494,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49494,36,'cp 0','d2d4 d7d6 e5d3 b8c6 c2c3 g7g6 c1g5 f8g7 b1d2 e8g8 f1e2 f8e8 e1c1 h7h6 g5f6 g7f6 d3f4 f6g5 h1e1 g8g7 h2h3 d8f6 f4d5 f6d8');
INSERT INTO evaluations VALUES(49845,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(49845,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(49845,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(49845,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(49845,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(49845,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(49845,36,'cp -92','f8c5 e2e3 d7d6 g1f3 e8g8 b1c3 d6e5 d1d8 f8d8 c3d5 b8a6 h2h3 g4f6 f3e5 f6e4 f1d3 e4g3 h1g1 f7f6 e5f3 a6b4 e1f2 b4d3 f2g3 c7c6 g1d1 c8f5 d1d3 f5d3 d5c7 d3c4 c7a8 d8a8 c1d2 c5b6 a2a4 a7a5 f3d4 a8e8 g3f3 c4d5 f3f2 h7h6 g2g4 d5e4 a1c1 e8d8 d2c3 d8d7');
INSERT INTO evaluations VALUES(42863,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(42863,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(42863,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(42863,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(42863,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(42863,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(42863,36,'cp -89','e2e3 d7d6 g1f3 e8g8 b1c3 d6e5 d1d8 f8d8 c3d5 b8a6 h2h3 g4f6 f3e5 f6e4 f1d3 e4g3 h1g1 f7f6 e5f3 a6b4 e1f2 b4d3 f2g3 c7c6 g1d1 c8f5 d1d3 f5d3 d5c7 d3c4 c7a8 d8a8 c1d2 c5b6 e3e4 a7a5 a2a3 a8d8 a1e1 d8d3 e1c1 c4a6 c1c3 a5a4 c3d3 a6d3 e4e5 g8f7 d2c3 f7e6 e5f6');
INSERT INTO evaluations VALUES(51159,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(51159,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(51159,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(51159,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(51159,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(51159,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(51159,36,'cp 172','g1f3 e7e6 f1d3 c8a6 e1g1 g8e7 f1e1 a6d3 d1d3 e7g6 g2g3 f8b4 c2c3 b4e7 h2h4 h7h5 c3c4 b8c6 b1c3 e8g8 a2a3 d7d6 c1e3 d8d7 d3e2 f7f5 e4f5 f8f5 c3e4 a8f8 f3g5 e7g5 e4g5 f5g5 h4g5 h5h4 e2g4 h4g3 f2g3');
INSERT INTO evaluations VALUES(51158,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(51158,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(51158,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(51158,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(51158,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(51158,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(51158,36,'cp 149','e7e6 f1d3 c8a6 e1g1 d7d6 c2c4 g8e7 b1c3 b8d7 c1e3 g7g6 d1d2 f8g7 a1d1 e8g8 b2b3 e7c6 d3e2 a6b7 d4d5 c6b4 a2a3 b4a6 f3d4 a6c5 d2c2 e6e5 d4b5 f7f5 e4f5 g6f5 f2f4 e5f4 e3f4 g7e5 f4e5 d7e5 b5d4 d8h4 d4f5');
INSERT INTO evaluations VALUES(52615,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(52615,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(52615,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(52615,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(52615,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(52615,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(52615,36,'cp 237','d4c5 d6c5 d1d8 e8d8 b1c3 f8g7 c1d2 c8d7 e1c1 b8c6 c3a4 b7b6 f1b5 c6d4 b5d7 d8d7 d2e3 d7c7 f3d4 c5d4 e3d4 g7d4 d1d4 a8d8 d4d8 c7d8 a4c3 g8f6 h1d1 d8c8 f2f3 h7h5 h2h4 f6d7 c3e2 d7e5 e2f4 e5c4 b2b3 c4e3 d1d2 c8c7 f4d3');
INSERT INTO evaluations VALUES(52593,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(52593,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(52593,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(52593,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(52593,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(52593,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(52593,36,'cp 234','d6c5 d1d8 e8d8 b1c3 f8g7 c1d2 b8c6 e1c1 c8d7 c3a4 b7b6 f1b5 c6d4 b5d7 d8d7 d2e3 d7c6 f3d4 c5d4 e3d4 e7e5 d4e3 f7f5 a4c3 g8f6 f2f3 h8d8 a2a4 f5f4 e3d2 d8d4 b2b3 g7f8 c3b5 d4d8 d2c3 f6d7 d1d5 a7a6 h1d1 f8c5 b5d6 d7f6 c3e5 f6d5 e4d5 c6d7 d6e4 d8f8 c1b2 f8f5');
INSERT INTO evaluations VALUES(54826,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(54826,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(54826,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(54826,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(54826,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(54826,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(54826,36,'cp 48','f8g7 f1d3 g8f6 g1f3 b8d7 e1g1 e8g8 f1e1 e7e5 c1g5 h7h6 g5h4 g6g5 h4g3 f6h5 d3c2 d7b6 d4e5 h5g3 h2g3 d6e5 d1e2 a7a5 b1d2 a5a4 d2f1 c8e6 f3h2 b6c4 b2b3 c4a3 f1e3 a3c2 e2c2 d8e7 c3c4 a4b3 a2b3 a8a1 e1a1 e7c5 h2g4 h6h5');
INSERT INTO evaluations VALUES(56938,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(56938,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(56938,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(56938,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(56938,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(56938,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(56938,36,'cp -55','f3e5 d8h4 g2g3 f4g3 d1g4 h4g4 e5g4 b8c6 c2c3 d7d5 e4d5 c6d4 c3d4 c8g4 h2g3 e8c8 b1c3 g8f6 c1g5 f8e7 f1e2 g4e2 e1e2 f6d5 g5e7 d5e7 a1f1 f7f5 f1f4 h7h5 f4h4 h8h6 h4h5 h6b6 b2b3 d8d4 h1d1 d4d1 c3d1 c8d7 e2f3 b6a6 d1c3 a6a5 h5h7');
INSERT INTO evaluations VALUES(56922,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(56922,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(56922,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(56922,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(56922,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(56922,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(56922,36,'cp -72','d8h4 g2g3 f4g3 d1g4 h4g4 e5g4 b8c6 c2c3 d7d5 e4d5 c6d4 c3d4 c8g4 h2g3 e8c8 b1c3 g8f6 c1g5 f8e7 f1e2 g4e2 e1e2 f6d5 g5e7 d5e7 a1f1 f7f5 e2d3 h7h6 c3e2 d8d6 h1h4 d6b6 d3c2 b6g6 f1h1 a7a5 c2c1 a5a4 c1b1 c8d7 a2a3');
INSERT INTO evaluations VALUES(35015,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(35015,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(35015,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(35015,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(35015,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(35015,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(35015,36,'cp -67','g2g3 f4g3 d1g4 h4g4 e5g4 b8c6 c2c3 g3h2 g4h2 d7d6 c1e3 h7h5 b1d2 c6e7 f1c4 f8h6 e3h6 g8h6 e1c1 e7g6 d1g1 h5h4 g1f1 c8h3 f1g1 h3d7 h2f3 h4h3 g1f1 g6f4 f3g1 f4h5 f1e1 h5g3 h1h2 a7a6 c4e2 g3e2 e1e2 e8c8 g1h3 h6g4 h2h1');
INSERT INTO evaluations VALUES(5426,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(5426,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(5426,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(5426,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(5426,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(5426,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(5426,36,'cp 49','f1d3 c8d7 e1g1 c5d4 c3d4 d8b6 d3e2 g8h6 c1h6 g7h6 d1d2 f7f6 e5f6 e7f6 f1d1 e8g8 b1c3 c6e7 e2d3 f6g7 d2e3 d7e8 d1e1 g8h8 e3e6 b6e6 e1e6 e7c6 e6d6 c6d4 f3d4 g7d4 a1f1 d4g7 c3d5 e8c6 d5e7 a8d8 d6d8 f8d8 d3c4 g7b2 e7c6 b7c6 f1b1');
INSERT INTO evaluations VALUES(7832,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(7832,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(7832,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(7832,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(7832,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(7832,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(7832,36,'cp -37','d7d5 a2a3 b4c3 b2c3 e8g8 d4c6 b7c6 f1d3 d8a5 e1g1 a5c3 d1e2 f8e8 e2e1 c3e1 f1e1 a7a5 e4d5 c6d5 d3b5 e8e7 a3a4 c8d7 c2c4 d5c4 b5c4 d7e8 e1c1 h7h6 e3d2 e7d7 d2e1 d7d8 c4e2 f6d5 c1c5 d5b4 g1f2 g8h8 c5c7 d8d7 c7c5 d7b7 e2c4');
INSERT INTO evaluations VALUES(8360,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(8360,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(8360,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(8360,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(8360,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(8360,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(8360,36,'cp -23','a2a3 b4c3 b2c3 e8g8 d4c6 b7c6 f1d3 d8a5 e1g1 a5c3 d1e1 c3e1 f1e1 a7a5 a1b1 c8a6 d3a6 a8a6 e4e5 f6d7 b1b7 f8d8 e3g5 d7c5 b7c7 d8b8 g5e3 c5a4 e3a7 b8f8 e1b1 a4c3 b1b7 c3b5 c7d7 a6a7 b7a7 b5a7 d7a7 f8b8 a7a5 h7h6 a5c5');
INSERT INTO evaluations VALUES(9323,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9323,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9323,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9323,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9323,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9323,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9323,36,'cp 1','c5b6 e1g1 e8g8 a2a4 a7a5 b4b5 c6e7 b1d2 e7g6 d3d4 c8g4 d1c2 d8d7 f1e1 f8e8 h2h3 g4f3 d2f3 e5d4 c1b2 d4c3 c2c3 d7e7 a1d1 b6c5 e4e5 g6e5 f3e5 d6e5 e1e5 e7e5 c3e5 e8e5 b2e5 c5d6 e5f6 g7f6 b5b6 a8c8 c4b3 c8d8 b6c7 d6c7 d1d8 c7d8 b3c2 g8g7 g1f1 d8e7 c2e4 g7g8 e4d5 e7b4 d5b7 h7h5 b7f3 h5h4 f1e2');
INSERT INTO evaluations VALUES(12545,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12545,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(12545,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12545,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12545,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12545,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12545,36,'cp 0','f1e2 f8e7 e1g1 e8g8 f3e1 d7d6 b1c3 c8e6 e1c2 f6d7 e2g4 d7f6');
INSERT INTO evaluations VALUES(12544,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12544,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12544,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12544,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12544,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12544,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12544,36,'cp 0','f8e7 e1g1 e8g8 f3e1 d7d6 b1c3 a7a6 a2a4 c8e6 e1c2 c6d4 a1b1 d4b3 c2e3 b3d4');
INSERT INTO evaluations VALUES(13427,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(13427,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(13427,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(13427,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(13427,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(13427,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(13427,36,'cp 66','e7e6 c1g5 f8e7 e1c1 e8g8 f2f4 a7a6 c4b3 b7b5 h1e1 d8c7 g5f6 e7f6 d4d6 c7d6 d1d6 g7g5 g2g3 g5f4 g3f4 f8d8 d6d8 f6d8 e4e5 c8b7 c3e4 b7e4 e1e4 d8b6 c2c3 a8d8 c1c2 g8g7 a2a4 g7g6 c2b1 d8d2 a4b5 a6b5 b3c2 g6g7 e4b4 b6g1 f4f5 e6f5 c2f5 g1h2 b4b5 d2e2');
INSERT INTO evaluations VALUES(14022,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(14022,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(14022,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(14022,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(14022,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(14022,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(14022,36,'cp 75','c1g5 f8e7 e1c1 e8g8 f2f4 a7a6 c4b3 b7b5 h1e1 d8c7 c1b1 h7h6 g5f6 e7f6 d4d6 c7d6 d1d6 a8b8 d6d3 a6a5 e4e5 f6e7 a2a4 b5a4 c3a4 g7g5 g2g3 b8b4 b1a2 c8a6 d3d7 a6b5 d7e7 b5a4 f4g5 a4b3 c2b3 h6g5 e7a7 b4d4 a7a5');
INSERT INTO evaluations VALUES(14792,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(14792,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(14792,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(14792,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(14792,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(14792,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(14792,36,'cp 71','b1c3 f8e7 c1f4 f6e4 e2e3 e4c3 b2c3 e8g8 c4d5 e6d5 f1d3 e7d6 f4d6 d8d6 e1g1 c6e7 c3c4 d5c4 d3c4 c7c5 d4c5 d6c5 d1b3 b7b6 a1c1 c5d6 f1d1 d6f6 f3d4 c8g4 d1d3 e7f5 h2h3 f5d4 d3d4 g4e6 c4e6 f7e6');
INSERT INTO evaluations VALUES(15959,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(15959,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(15959,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(15959,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(15959,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(15959,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(15959,36,'cp 16','c7c5 c2c3 d8b6 d1c2 f8d6 b1d2 d6f4 e3f4 b6c7 g2g3 e8g8 e1g1 a7a5 a2a4 c5d4 f3d4 d7c5 d2f3 c8d7 f3e5 d7e8 f2f3 h7h6 f1e1 f6d7 e1e2 c7b6 g1g2 a8d8 g3g4 b6c7 h2h4 d8c8 h4h5 d7e5 e2e5');
INSERT INTO evaluations VALUES(12211,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(12211,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(12211,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(12211,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12211,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12211,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12211,36,'cp 11','c2c3 d8b6 d1c2 f8d6 b1d2 e8g8 e1g1 a7a5 f4g3 b6c7 c3c4 d6g3 h2g3 c5d4 f3d4 b7b6 f1c1 c8a6 c4d5 c7c2 d3c2 f6d5 c2b3 f8c8 d4c6 g8f8 b3d5 e6d5 d2b3 c8c7 c1c2 a6c4 c6d4 g7g6 f2f3 a8c8 a1c1');
INSERT INTO evaluations VALUES(12210,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(12210,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(12210,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(12210,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(12210,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12210,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(12210,36,'cp 9','f6h5 f4e5 h5f6 b1d2 f8e7 e1g1 e8g8 e5g3 f6h5 d1e2 h5g3 h2g3 g7g6 e3e4 c5d4 c3d4 d5e4 d3e4 a8b8 a1c1 d7f6 f1e1 f6e4 d2e4 b7b6 e2e3 c8b7 e3f4 b7e4 f4e4 b8c8 e4b7 a7a5 c1c8 d8c8');
INSERT INTO evaluations VALUES(16036,30,'cp -482',NULL);
INSERT INTO evaluations VALUES(16036,31,'cp -491',NULL);
INSERT INTO evaluations VALUES(16036,32,'cp -478',NULL);
INSERT INTO evaluations VALUES(16036,33,'cp -491',NULL);
INSERT INTO evaluations VALUES(16036,34,'cp -497',NULL);
INSERT INTO evaluations VALUES(16036,35,'cp -512',NULL);
INSERT INTO evaluations VALUES(16036,36,'cp -522','e8f7 d2d4 e5d4 e4e5 d8e7 d1e2 f6d5 e1g1 h7h6 c2c3 d7d6 c3d4 f7g8 b1c3 d5c3 b2c3 e7f7 h2h3 c8e6 f1e1 d6e5 d4e5 a8e8 e1d1 e6f5 d1e1 h8h7 e2b5 f5e6 b5b7 e6d5 b7b5 g7g5 e1e3 f7d7 b5b1 d5f3 e3f3 c6e5');
INSERT INTO evaluations VALUES(21039,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(21039,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(21039,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(21039,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(21039,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21039,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(21039,36,'cp -41','e7e5 d4e5 c6e5 c1f4 e5g6 f4h2 g8f6 e2e3 f8d6 h2d6 d8d6 g1f3 e8g8 c2c4 g6e5 c4d5 e5f3 d1f3 d6b6 b2b4 a7a5 b4b5 b6c5 f1d3 f8d8 a1a2 d8d5 a2c2 c5d6 d3c4 d5f5 f3d1 d6d1 e1d1 f6e4 d1e1 f5c5 c4d3 c5c2 d3c2 e4d6 c2d3 c8d7 b1c3 c7c6 b5c6 d7c6');
INSERT INTO evaluations VALUES(21697,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(21697,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(21697,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(21697,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(21697,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(21697,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(21697,36,'cp -31','d4e5 c6e5 c1f4 e5g6 f4h2 f8d6 h2d6 d8d6 e2e3 g8f6 g1f3 e8g8 c2c4 g6e5 b1c3 c8e6 c4d5 f6d5 f1e2 d5c3 d1d6 c7d6 b2c3 a8c8 e1d2 e6d5 h1b1 c8c5 b1b4 f8c8 e3e4 d5c6 a3a4 b7b6 b4d4 g8f8 d4d6 c6e4 f3e5 c5e5 g2g4 g7g6 e2d3');
INSERT INTO evaluations VALUES(22042,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(22042,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(22042,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(22042,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(22042,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(22042,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(22042,36,'cp -4','f1b5 g8f6 b5c6 b7c6 e1g1 f8e7 e4e5 f6g4 c1f4 g4e5 f3e5 d6e5 f4e5 d8d1 f1d1 e8g8 e5c7 f8e8 c3e4 e7f8 f2f3 c8f5 e4d6 e8e7 d6f5 e7c7 a1c1 a7a5 f5d6 a5a4 g1f1 g7g6 d6e4 c7b7 b2b3 a4b3 a2b3 a8c8 e4d6 f8d6 d1d6');
INSERT INTO evaluations VALUES(22048,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(22048,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(22048,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(22048,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(22048,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(22048,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(22048,36,'cp 0','g7g6 b1c3 f8g7 b2b3 c8g4 c1b2 g7h6 e3d3 g8f6 c3d5 h6g7 h2h3 g4f3 d3f3 c6e5 b2e5 d6e5 f1c4 c7c6 d5f6 d8f6 f3f6 g7f6 e1e2 a8d8 h1d1 e8g8 b3b4 d8d1 a1d1 f8d8 a2a4 d8d1 e2d1 f6d8 c2c3 a7a5 f2f3 g8f8 d1e2 h7h5 e2d3 h5h4 d3e2 d8b6 e2d3 f8g7 d3e2');
INSERT INTO evaluations VALUES(37559,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(37559,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(37559,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(37559,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(37559,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(37559,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(37559,36,'cp 144','h2h4 h7h6 c1f4 c8b7 d4d5 g8f6 d1d2 b8a6 f1a6 b7a6 e1c1 a6b7 h1e1 a8c8 c1b1 f6g4 e4e5 e8g8 f3g1 d7d6 e5d6 g7c3 d2c3 e6e5 d6c7 c8c7 c3g3 e5f4 g3g4 d8d6 g1h3 d6c5 h4h5 g6g5 h3g5 c5c2 b1a1 h6g5 g4g5 g8h7 g5f4');
INSERT INTO evaluations VALUES(38701,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(38701,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(38701,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(38701,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(38701,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38701,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(38701,36,'cp 11','g1f3 g8f6 e1g1 h7h6 f1e1 e8g8 c2c3 c5b6 b1d2 c7c6 h2h3 b8d7 d3d4 f8e8 c4f1 a7a5 d1c2 a5a4 d2c4 b6c7 d4e5 d6e5 c1e3 b7b5 c4d2 c7b6 b2b4 a4b3 a2b3 a8a1 e3b6 d8b6 e1a1 c8b7 b3b4 c6c5');
INSERT INTO evaluations VALUES(39479,30,'cp 303',NULL);
INSERT INTO evaluations VALUES(39479,31,'cp 300',NULL);
INSERT INTO evaluations VALUES(39479,32,'cp 271',NULL);
INSERT INTO evaluations VALUES(39479,33,'cp 310',NULL);
INSERT INTO evaluations VALUES(39479,34,'cp 294',NULL);
INSERT INTO evaluations VALUES(39479,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(39479,36,'cp 299','f3e5 d8e7 d2d4 c5d4 d1d4 b8c6 d4e3 e7e5 e3e5 c6e5 b1c3 g8f6 c1f4 d7d6 f4e5 d6e5 f1d3 c8d7 f2f3 e8c8 e1c1 h8e8 h1e1 d7c6 d3c4 d8d1 e1d1 b7b6 c4e6 c6d7 e6d7 f6d7 c3e4 d7f6 g2g4 f6e4 f3e4 e8f8 c1d2 f8d8 d2e2 d8d1 e2d1 c8d8 d1e2 d8e7 g4g5 a7a5 e2d3');
INSERT INTO evaluations VALUES(39477,30,'cp 308',NULL);
INSERT INTO evaluations VALUES(39477,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(39477,32,'cp 266',NULL);
INSERT INTO evaluations VALUES(39477,33,'cp 302',NULL);
INSERT INTO evaluations VALUES(39477,34,'cp 300',NULL);
INSERT INTO evaluations VALUES(39477,35,'cp 282',NULL);
INSERT INTO evaluations VALUES(39477,36,'cp 263','d8e7 d2d4 c5d4 d1d4 b8c6 d4e3 c6e5 b1c3 g8f6 c1d2 d7d6 e1c1 c8f5 d1e1 e8f8 f1c4 e7d7 c4b3 h7h5 f2f3 a7a5 a2a4 a8e8 e3a7 b7b6 c3b5 e8e7 h2h4 d7d8 a7a6 f6d7 d2g5 d7c5 g5e7 f8e7 a6a7 c5b3 c2b3 d8d7 c1d2 c7c6 b5d4 d7a7 d4c6 e7f7 c6a7');
INSERT INTO evaluations VALUES(41868,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41868,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(41868,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41868,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(41868,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41868,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41868,36,'cp -7','b1c3 d8e7 g1e2 c7c6 h2h3 c8e6 a2a4 a7a5 e2g3 b8d7 g3f5 e6f5 f3f5 d7f8 c1e3 f8e6 e3c5 e6c5 f5g5 e8g8 e1g1 c5e6 c4e6 f7e6 c3e2 h7h6 g5e3 f6h5 a1d1 d6d5 d3d4 e5d4 e2d4 c6c5 d4e2 d5d4 e3d3 h5f4 e2f4 f8f4');
INSERT INTO evaluations VALUES(45765,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(45765,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(45765,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(45765,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(45765,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(45765,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(45765,36,'cp -23','f1d3 f8d6 g1e2 e8g8 e1g1 f8e8 f2f3 c6c5 g1h1 b8c6 d4c5 d6c5 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 d1c2 d8b6 e2f4 c6d4 c2c3 c8e6 c1e3 a8c8 c3d3 c8d8 e4h7 g8h8 h7e4 d4c6 e3c5 b6c5 f4e6 e8e6 d3b3 c5e7 a1c1');
INSERT INTO evaluations VALUES(45775,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(45775,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(45775,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(45775,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45775,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(45775,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(45775,36,'cp -19','f8d6 g1e2 e8g8 c1d2 f8e8 e1g1 b8d7 d1c2 a7a5 a2a3 b7b5 e2g3 d7b6 e3e4 d6g3 h2g3 b5b4 a3b4 a5b4 a1a8 b6a8 c3a2 d5e4 d3c4 d8d4 a2b4 a8b6 c4b3 d4d6 b4c6 c8e6 b3e6 d6e6 f1c1 h7h5 b2b4 f6g4 b4b5 h5h4 g3h4 g4f2');
INSERT INTO evaluations VALUES(48807,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(48807,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(48807,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(48807,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(48807,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(48807,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(48807,36,'cp 118','d4e5 f6e4 d1d5 e4c3 d5d8 e8d8 b2c3 f8c5 c1d2 b8c6 f2f4 f7f6 e1c1 c8g4 g1f3 d8e7 f1e2 f6e5 h1e1 h8e8 f4e5 e7f8 d2f4 a8d8 d1d8 e8d8 f3g5 c5f2 e2g4 f2e1 g5e6 f8f7 e6d8 c6d8 c1b2 h7h6 b2b3 c7c6 g4c8 d8e6 f4e3 b7b6');
INSERT INTO evaluations VALUES(48768,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(48768,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(48768,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(48768,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(48768,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(48768,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(48768,36,'cp 114','f6e4 d1d5 e4c3 d5d8 e8d8 b2c3 f8c5 f1d3 b8c6 f2f4 f7f6 g1f3 f6e5 f4f5 c8d7 f3g5 h8f8 g2g4 h7h6 g5e4 c5e7 c1e3 e7h4 e1d2 b7b6 h1g1 d8e8 a1f1 a8d8 d2c1 h4e7 h2h3 c6a5 f5f6 e7f6 e4f6 f8f6 f1f6 g7f6 e3h6 d7e6 g1f1 e6c4 d3c4 a5c4 f1f6 d8d5 h3h4');
INSERT INTO evaluations VALUES(49063,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(49063,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(49063,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(49063,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(49063,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(49063,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(49063,36,'cp 123','d1d5 e4c3 d5d8 e8d8 b2c3 f8c5 f1d3 b8c6 f2f4 f7f6 g1f3 f6e5 f4f5 c8d7 f3g5 h8f8 g2g4 h7h6 g5e4 c5e7 c1e3 e7h4 e1d2 c6e7 c3c4 d7c6 h1f1 c6e4 d3e4 c7c6 a1d1 d8c7 d2e2 b7b6 e2f3 h4g5 e3f2 a8d8 d1d8 f8d8 a2a4 d8d2 c4c5');
INSERT INTO evaluations VALUES(49306,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(49306,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(49306,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(49306,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(49306,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(49306,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(49306,36,'cp 199','h2h4 h7h6 a2a3 a7a6 b1c3 d7d6 h4h5 g6e7 d1e2 b8d7 c1e3 b7b5 e1c1 d7f6 f3d2 d6d5 e4e5 f6d7 g2g4 b5b4 a3b4 a8b8 f2f4 b8b4 c1b1 e7c6 e2f2 f8e7 d2b3 a6a5 f1b5 c6a7 b5a4 c7c6 f4f5 a7b5 a4b5 c6b5 f2g3');
INSERT INTO evaluations VALUES(51276,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(51276,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(51276,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(51276,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(51276,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(51276,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(51276,36,'cp 82','e2e4 g7g6 b1c3 f8g7 g1f3 g8f6 h2h3 e8g8 c1g5 b8a6 d1d2 e7e5 f1d3 e5d4 f3d4 a6c5 e1g1 h7h6 g5e3 c5d3 d2d3 c8b7 f2f3 d8d7 a1d1 a8e8 d4e2 f6h5 g2g4 h5f6 e3d4 d7d8 c3d5 c7c6 d5e3 e8e7 e2g3 f8e8 b2b3');
INSERT INTO evaluations VALUES(51277,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(51277,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(51277,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(51277,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(51277,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(51277,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(51277,36,'cp 84','g7g6 b1c3 f8g7 c1g5 g8f6 f2f4 f6h5 f4f5 c7c5 g1f3 c5d4 f3d4 c8b7 f1e2 b8c6 d4c6 g7c3 b2c3 b7c6 e1g1 d8c7 c4c5 d6c5 e4e5 h5g7 f5f6 g7e6 g5h6 e7f6 f1f6 c7e5 d1f1 e8c8 f6f7 d8d7 f7d7 c6d7 a1e1 b6b5 e2f3 e5c3 e1c1 c3d4 g1h1 e6f4');
INSERT INTO evaluations VALUES(54222,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(54222,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(54222,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(54222,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54222,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(54222,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(54222,36,'cp 0','e2e4 d5e4 d4c5 d8d1 e1d1 f8c5 c3e4 c5e7 g1f3 g8f6 e4f6 e7f6 f1d3 c8d7 a3a4 d7c6 c2c3 b8d7 c1e3 e8g8 d3b5 c6d5 d1c2 f8d8 h1d1 b7b6 b5f1 d5e4 f1d3 e4d5');
INSERT INTO evaluations VALUES(279,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(279,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(279,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(279,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(279,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(279,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(279,36,'cp 28','d7d6 a2a4 a7a6 b2b4 c5b6 d2d3 f6g6 b1d2 g8f6 f3h4 g6h5 d1h5 f6h5 f1e1 h5f6 h4f3 h7h6 g2g3 c8h3 a4a5 b6a7 d2f1 f6g4 c1e3 a7e3 f1e3 g4e3 e1e3 g7g5 c4d5 h3d7 d3d4 f7f6 d4e5 f6e5 f3d2 e8c8');
INSERT INTO evaluations VALUES(190,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(190,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(190,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(190,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(190,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(190,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(190,36,'cp 47','b2b4 c5b6 a2a4 a7a6 d2d3 f6g6 f1e1 g8f6 b1d2 e8g8 f3h4 g6h5 d1h5 f6h5 g2g3 h7h6 g1g2 c8e6 h4f3 h5f6 c4e6 f7e6 d2c4 f6d7 c4b6 d7b6 c1e3 b6d7 f3d2 f8f7 e1b1 b7b6 b1b2 g7g5 h2h3 c6e7 d2c4 a8f8 b4b5');
INSERT INTO evaluations VALUES(4201,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4201,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(4201,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(4201,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4201,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4201,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(4201,36,'cp 12','b1c3 a7a6 c1e3 d7d6 e3c5 d6c5 c3d5 d8d6 d5e3 b7b5 c4a2 f8d8 a3a4 a8b8 h2h3 c8e6 a2e6 d6e6 d1e2 f6h5 e3d5 h5f4 d5f4 e5f4 f1d1 b5b4 c2c3 b4c3 b2c3 c6e5 a1c1 e5f3 e2f3 e6e5 d3d4 c5d4 c3d4 d8d4 d1d4 e5d4 f3f4 d4a4 c1c7 b8f8 g1h2 a4d4');
INSERT INTO evaluations VALUES(4875,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4875,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4875,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4875,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4875,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4875,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4875,36,'cp 0','g8f6 d2d3 e8g8 a4c6 d7c6 a2a4 f8e8 e1g1 c5d6 h2h3 h7h6 c3e2 f6h5 e2g3 h5f4 f1e1 f4g6 f3h2 g6f4');
INSERT INTO evaluations VALUES(10137,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10137,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(10137,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(10137,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10137,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10137,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10137,36,'cp 0','e8g8 d2d3 c6d4 g1f3 d7d6 c1e3 c5b6 e1g1 d4f3 d1f3 c7c6 c3e2 c8e6 e2g3 b6e3 f3e3 d8b6 c4e6 f7e6 e3d2 f8f7 b2b4 a8f8 a3a4 c6c5 b4b5 b6c7 c2c4 c7e7 a4a5 h7h6 f1d1 f6d7 d1f1');
INSERT INTO evaluations VALUES(4351,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4351,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4351,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4351,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4351,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4351,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4351,36,'cp 0','g1f3 d7d6 d2d3 c6d4 f3d4 c5d4 d1f3 c8e6 c3e2 e6c4 d3c4 f6d7 e1g1 d4b6 f3g4 d7f6 g4h4 f6d7');
INSERT INTO evaluations VALUES(10605,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(10605,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(10605,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(10605,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(10605,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(10605,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(10605,36,'cp 93','f3e5 c6e5 f2f4 e5c6 f1d3 d8h4 g2g3 h4h3 c1d2 e8g8 c3e4 c5e7 e4f2 h3e6 e1g1 e7b4 a2a3 b4d2 d1d2 d7d6 f1e1 a7a5 f2e4 e6e7 e4c3 c8h3 d3e4 e7d7 b2b3 g8h8 e4f3 a8e8 e3e4 h3g4 f3g2 g4h3 g2h3 d7h3 d2d3 f7f6 c3d5');
INSERT INTO evaluations VALUES(10582,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(10582,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(10582,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(10582,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(10582,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(10582,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(10582,36,'cp 85','c6e5 f2f4 e5g6 f1d3 e8g8 e1g1 b7b6 c1d2 c8b7 d1c2 a7a5 h2h3 f7f5 d3f5 f8f5 c2f5 b7g2 f1f2 g2b7 a1f1 d8e7 c3d5 e7d6 f5e4 b7d5 e4d5 d6d5 c4d5 g6e7 e3e4 c5f2 f1f2 a8f8 d2c3');
INSERT INTO evaluations VALUES(10805,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(10805,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(10805,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(10805,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(10805,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(10805,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(10805,36,'cp 75','f2f4 e5g6 f1d3 e8g8 e1g1 b7b6 a2a3 a7a5 a1b1 c8b7 b2b4 c5e7 d1c2 f8e8 f1d1 a5b4 a3b4 e7f6 c3d5 b7c6 d1f1 c6a4 c2d2 a4c6 g1h1 d7d6 d2c2 c6d5 c4d5 d8d7 e3e4 d7e7 f1e1 e7d7 b4b5 d7e7 g2g3 e7d7 d3f1 h7h6 f1g2 f6d4 h2h3 d4c5');
INSERT INTO evaluations VALUES(10694,30,'cp 532',NULL);
INSERT INTO evaluations VALUES(10694,31,'cp 544',NULL);
INSERT INTO evaluations VALUES(10694,32,'cp 540',NULL);
INSERT INTO evaluations VALUES(10694,33,'cp 531',NULL);
INSERT INTO evaluations VALUES(10694,34,'cp 532',NULL);
INSERT INTO evaluations VALUES(10694,35,'cp 525',NULL);
INSERT INTO evaluations VALUES(10694,36,'cp 533','c4f7 e8f8 c1g5 c6d4 b3a3 f8f7 g5d8 h8d8 f3g5 f7f6 f1a1 h7h6 a3d3 f6g5 d3g3 g5f6 g3h4 f6f7 h4d8 d4e6 d8h4 d7d6 h4h5 f7f8 a1e1 c8d7 h5f3 f8g8 f3b7 a8e8 h2h3 e6g5 e1c1 g5e6 c1c3 e8f8 c3e3 f8f7 b7a8 f7f8 a8d5 g8h8 d5b7 f8e8 a2a3 h8g8 b7b3 g8h8 e3e1 e8f8 b3b7 h8g8 b7d5 g8h8 d5b3 e6f4 e1e7 d7e6 b3b7 e6d5');
INSERT INTO evaluations VALUES(9652,30,'cp 548',NULL);
INSERT INTO evaluations VALUES(9652,31,'cp 540',NULL);
INSERT INTO evaluations VALUES(9652,32,'cp 541',NULL);
INSERT INTO evaluations VALUES(9652,33,'cp 544',NULL);
INSERT INTO evaluations VALUES(9652,34,'cp 517',NULL);
INSERT INTO evaluations VALUES(9652,35,'cp 540',NULL);
INSERT INTO evaluations VALUES(9652,36,'cp 533','e8f8 c1g5 c6d4 b3a3 f8f7 g5d8 h8d8 f3g5 f7f6 f1a1 h7h6 a3d3 f6g5 d3g3 g5f6 g3h4 f6f7 h4d8 d4e6 d8h4 d7d6 h4h5 f7f8 a1e1 c8d7 h5f3 f8g8 f3b7 a8e8 h2h3 g8h7 b7b1 h7g8 b1b3 g8h7 e1e3 e6c5 b3b1 h7g8 e3e8 d7e8 b1b8 g8f8 b8c7 c5d7 c7a7 d7c5 a7c7 c5e6 c7d6 f8f7 f2f4 g7g6 d6d5 f7e7 g1f2 e8d7 f4f5 g6f5');
INSERT INTO evaluations VALUES(2511,30,'cp 532',NULL);
INSERT INTO evaluations VALUES(2511,31,'cp 533',NULL);
INSERT INTO evaluations VALUES(2511,32,'cp 532',NULL);
INSERT INTO evaluations VALUES(2511,33,'cp 536',NULL);
INSERT INTO evaluations VALUES(2511,34,'cp 542',NULL);
INSERT INTO evaluations VALUES(2511,35,'cp 538',NULL);
INSERT INTO evaluations VALUES(2511,36,'cp 559','c1g5 c6d4 b3a3 f8f7 g5d8 h8d8 f3g5 f7f6 f1a1 h7h6 a3d3 f6g5 d3g3 g5f6 g3h4 f6f7 h4d8 d4e6 d8h4 d7d6 h4h5 f7f8 a1e1 c8d7 h5f3 f8g8 f3b7 a8e8 h2h3 g8h7 b7a7 d6d5 a7b7 e6f4 b7b1 f4g6 e1e8 d7e8 b1b8 e8c6 b8c7 c6b5 c7d6 b5c4 a2a4 g6h8 a4a5 h8f7 d6d7 f7g5 f2f4 g5e4 d7f5 h7g8 f5e6 g8h7 a5a6 c4a6 e6a6 e4f6 a6d3 h7g8');
INSERT INTO evaluations VALUES(2510,30,'cp 561',NULL);
INSERT INTO evaluations VALUES(2510,31,'cp 559',NULL);
INSERT INTO evaluations VALUES(2510,32,'cp 561',NULL);
INSERT INTO evaluations VALUES(2510,33,'cp 566',NULL);
INSERT INTO evaluations VALUES(2510,34,'cp 555',NULL);
INSERT INTO evaluations VALUES(2510,35,'cp 550',NULL);
INSERT INTO evaluations VALUES(2510,36,'cp 553','c6d4 b3a3 f8f7 g5d8 h8d8 f3g5 f7f6 f1a1 h7h6 a3d3 f6g5 d3g3 g5f6 g3h4 f6f7 h4d8 d4e6 d8h4 d7d6 h4h5 f7f8 a1e1 c8d7 h5f3 f8g8 f3b7 a8e8 h2h3 e6f8 e1e8 d7e8 b7c7 d6d5 c7b8 e8d7 b8a7 d5d4 a2a4 d7a4 a7a4 g8f7 a4a7 f7f6 a7d4 f6f7 d4d5 f8e6 d5f5 f7e7 f2f4 g7g6 f5g6 e6f4 g6h6 f4e6 g1f2 e7f7 h3h4 f7e7 h4h5 e7d6');
INSERT INTO evaluations VALUES(13943,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(13943,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(13943,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(13943,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(13943,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(13943,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(13943,36,'cp -13','f8e7 e1g1 e8g8 f1e1 d8c7 c3b5 c7d7 b5c3 f8d8 c4a2 d7c7 h2h3 c8d7 d1e2 a7a6 c1f4 c6e5 f3e5 d6e5 f4g3 b7b5 c3b1 a6a5 b1d2 d7c6 a1d1 h7h6 d2f3 e7d6 g3h4 a5a4 c2c3 c5c4 d3c4 b5c4 e2c4');
INSERT INTO evaluations VALUES(8170,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(8170,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(8170,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8170,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(8170,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(8170,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(8170,36,'cp -9','e1g1 e8g8 c4a2 d8c7 f1e1 f8d8 c3e2 b7b5 c2c3 c6e5 e2f4 c5c4 d3d4 e5f3 d1f3 e6e5 f4h5 f6h5 f3h5 a7a5 c1e3 e7f6 d4d5 c8d7 a2b1 d8b8 h5d1 f6e7 b1c2 g7g6 h2h3 e7f8 d1e2 a5a4 f2f4 e5f4 e3f4');
INSERT INTO evaluations VALUES(8169,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(8169,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(8169,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(8169,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(8169,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(8169,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(8169,36,'cp -17','e8g8 f1e1 c8d7 h2h3 d8c7 c3b5 c7b8 c1f4 a7a6 f4d6 e7d6 b5d6 b8d6 e4e5 c6e5 f3e5 d7c6 c4b3 a8d8 a3a4 b7b6 d1e2 f6d5 b3d5 c6d5 e5c4 d6c6 e2g4 d5c4 g4c4 a6a5 e1e4 d8d5 b2b3 c6d6 a1e1 h7h6 e4g4 f8d8 g4e4 d8d7 c4c3');
INSERT INTO evaluations VALUES(3125,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(3125,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(3125,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(3125,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(3125,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(3125,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(3125,36,'cp -13','f1e1 d8c7 c3b5 c7d7 b5c3 b7b6 c3e2 c8b7 c4a2 c6e5 e2g3 a8d8 f3e5 d6e5 c1d2 b7a6 g3f1 c5c4 a2c4 a6c4 d3c4 d7d4 d2c3 d4c4 f1d2 c4c7 d1e2 d8c8 d2f3 f6d7 h2h4 f7f6 h4h5 f8f7 a3a4 a7a5 h5h6 g7h6 g2g3 c7c6 g1g2 e7b4 c3b4');
INSERT INTO evaluations VALUES(14072,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(14072,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(14072,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(14072,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(14072,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(14072,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14072,36,'cp 75','e1g1 a7a6 b5a4 c8d7 d2d4 e7g6 d4d5 c6b8 c3c4 f8e7 b1c3 h7h6 c1e3 e8g8 a2a3 d7a4 d1a4 b8d7 b2b4 b7b6 g2g3 d7f6 h2h3 d8c8 g1g2 f6h7 a4c2 f7f5 e4f5 f8f5 f3d2 c8d7 c2a4 d7a4 c3a4 e7g5 a4c3 a8f8 e3g5 h7g5');
INSERT INTO evaluations VALUES(15596,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(15596,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(15596,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(15596,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(15596,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(15596,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(15596,36,'cp 17','f1g2 g7g6 e1g1 f8g7 a2a4 a7a5 f1e1 e8g8 b1a3 f8e8 h2h3 d8d7 c2c3 b7b6 c1g5 c8b7 a3c4 g8h8 g5f6 g7f6 h3h4 a8d8 d1d2 f6g7 d3d4 b7a6 c4a3 e5d4 c3d4 c6b4 a3b5 c7c5 d4d5 h7h6');
INSERT INTO evaluations VALUES(19085,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(19085,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19085,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(19085,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(19085,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19085,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(19085,36,'cp 1','e2e4 f8c5 f1d3 d7d6 h2h3 g8f6 a2a3 e8g8 e1g1 c8e6 c2c4 c6d4 b1c3 a8c8 f3g5 f6d7 c1e3 d4c6 g5e6 f7e6 b2b4 c5e3 f2e3 d8e7 d3e2 f8f1 e2f1 d7f6 d1d3 a7a5 a1b1 h7h5 f1e2 a5b4 a3b4 g8h7 g1h2 c8a8');
INSERT INTO evaluations VALUES(20167,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20167,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(20167,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20167,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20167,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20167,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20167,36,'cp 0','c3d4 c8d7 b1c3 d8b6 b5c6 b6c6 g1e2 b7b5 e2f4 b5b4 c3e2 a8c8 h2h4 g8e7 h4h5 c6c2 d1c2 c8c2 a2a3 b4b3 h1h3 d7a4 e2c3 a4d7');
INSERT INTO evaluations VALUES(23114,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(23114,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(23114,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(23114,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(23114,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(23114,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(23114,36,'cp 25','e5d4 c3b5 d7d5 d1f3 a7a6 b5d4 d8e7 d2d3 g8f6 c2c3 d5e4 d3e4 e7e4 f3e4 f6e4 f1d3 e4c5 d3c4 b7b5 c4d5 c8b7 d5b7 c5b7 c1e3 c7c5 d4f3 f8d6 h2h4 e8d7 h4h5 d7c6 h5h6 g7g6 b2b3 d6c7 e1c1 f7f6');
INSERT INTO evaluations VALUES(23150,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(23150,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(23150,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(23150,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(23150,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(23150,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(23150,36,'cp 43','c3b5 c7c5 d1h5 d7d6 f1c4 d8e7 d2d3 g8f6 h5e2 a7a6 b5d6 e7d6 e4e5 d6e7 e5f6 g7f6 c4d5 c8e6 d5f3 e8c8 c1f4 e6d5 f3g4 d5e6 e1g1 h8g8 g4e6 f7e6 f4g3 e7d7 a1e1 d8e8 e2h5 f8d6 g3d6 d7d6 g2g3 e8e7 e1e2 g8g5 h5h6 g5g6 h6f8 c8c7 f2f4 e7d7 f8a8 d6d5 e2e4');
INSERT INTO evaluations VALUES(29005,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(29005,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(29005,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(29005,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(29005,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(29005,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(29005,36,'cp 40','d5c4 e2e3 d8b6 d1c2 g4f3 g2f3 b6a6 f1e2 e7e6 e1g1 b8d7 c3e4 f8e7 g5f4 e8g8 e2c4 a6a5 c2b3 d7b6 c4e2 a5h5 e4c3 c6c5 d4c5 h5c5 f1c1 c5a5 c3b5 a8c8 c1d1 c8d8 e3e4 d8d1 a1d1 f8d8 d1d8 e7d8 b5c3 d8e7 f4e3');
INSERT INTO evaluations VALUES(31751,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(31751,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(31751,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(31751,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(31751,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(31751,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(31751,36,'cp 59','c2c4 d5c4 b1c3 g8f6 e2e3 e7e6 f1c4 f8d6 f3e5 c7c6 d1b3 d8e7 c4e2 e8g8 e1g1 a7a5 a2a3 f8d8 f1c1 f6e4 c1d1 e4c3 b2c3 a5a4 b3b2 b7b5 e2f3 d8c8 e3e4 f5h7 f4g3 b8d7 e5d7 e7d7');
INSERT INTO evaluations VALUES(31849,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(31849,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(31849,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(31849,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(31849,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(31849,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(31849,36,'cp -16','g8f6 f1d3 e7e6 g1f3 f5d3 d1d3 c7c5 e1g1 b8c6 d4c5 f8c5 c3c4 a8c8 b1c3 d5c4 d3c4 c5e7 f1d1 d8c7 c4a4 e8g8 c1d2 a7a6 a1c1 f8d8 a2a3 c7b8 c3e4 f6d5 e4c3 d5b6 a4b3 b6d7 d2e1 c6e5 f3e5 d7e5');
INSERT INTO evaluations VALUES(32253,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(32253,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(32253,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(32253,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(32253,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(32253,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(32253,36,'cp 117','c7c5 d4d5 d8d5 d1d5 e6d5 b1c3 g8f6 c3d5 f6d5 c1d2 f8e7 g1f3 b8c6 e1c1 e8g8 f1b5 e7f6 h1e1 c6d4 f3d4 c5d4 b5e2 f8d8 e2f3 a8c8 c1b1 c8c5 h2h3 g7g6 a2a3 g8g7 e1e4 d5b6 d2b4 c5c7 f3e2 b6c4 e2d3 c4e5');
INSERT INTO evaluations VALUES(38492,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(38492,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(38492,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(38492,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(38492,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(38492,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(38492,36,'cp 86','g1f3 g8f6 e1g1 c7c5 d4c5 e8g8 b1c3 b8c6 e3e4 d8a5 e4e5 f6g4 e5e6 f7e6 c3e4 a5c7 h2h3 g4e5 f3e5 c6e5 c4b3 b7b6 d1c2 c7c6 c1e3 c8a6 e4g5 a6d3 b3e6 g8h8 c2b3 b6c5 e6d5 c5c4 d5c6 c4b3 c6a8 f8a8 e3d4 d3f1 g1f1 e5c6 d4g7 h8g7 a2b3');
INSERT INTO evaluations VALUES(38493,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(38493,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(38493,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(38493,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(38493,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(38493,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(38493,36,'cp 75','g8f6 e1g1 c7c5 d4c5 d8c7 b1d2 c7c5 a2a3 c5b6 b2b4 f6d7 a1a2 b8c6 c4e2 d7e5 f3e5 c6e5 a2c2 e8g8 c1b2 c8f5 e3e4 f5e6 d1a1 b6d6 f2f4 d6b6 f1f2 e5g4 e2g4 g7b2 a1b2 e6g4 d2c4 b6e6 c4e3 h7h5 h2h3 e6e4 h3g4 e4e3 c2e2 e3d3 g4h5 a8c8 h5h6 d3c3 b2c3 c8c3 e2e7');
INSERT INTO evaluations VALUES(41538,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(41538,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(41538,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(41538,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(41538,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(41538,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(41538,36,'cp 87','e1g1 c7c5 d4c5 d8c7 a2a3 c7c5 b1d2 e8g8 b2b4 c5d6 c1b2 c8e6 a1c1 b8d7 d1b3 f8c8 f1d1 h7h6 c4e6 d6e6 b3e6 f7e6 g1f1 c8c1 d1c1 f6e8 b2g7 g8g7 f1e2 a7a5 b4b5 e6e5 g2g4 e8d6 c1c7 a8d8 d2c4 d6b5 c7b7 b5c3 e2f1 g7f6');
INSERT INTO evaluations VALUES(41537,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(41537,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(41537,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(41537,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(41537,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(41537,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(41537,36,'cp 86','c7c5 d4c5 d8c7 b1d2 c7c5 a2a3 c5d6 b2b4 f6g4 a1b1 b8c6 d1b3 e8g8 f1d1 g4e5 c1b2 c8f5 f3e5 c6e5 e3e4 f5e6 c4e6 d6e6 b3e6 f7e6 b2e5 g7e5 d2f3 e5c7 d1d7 a8c8 d7e7 c7b6 h2h4 c8c2 b1f1 c2a2 e7b7 a2a3 f1d1 f8f7 b7f7 g8f7 d1d7 f7g8');
INSERT INTO evaluations VALUES(38768,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(38768,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(38768,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(38768,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(38768,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(38768,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(38768,36,'cp 95','b2c3 e6d5 f1d3 g8e7 c1a3 e8g8 d1h5 g7g6 h5f3 b8d7 g1e2 c7c5 d4c5 d7e5 f3f6 e7c6 f6d8 f8d8 a1d1 c6a5 e1g1 c8d7 e2d4 d8e8 f1e1 a8c8 d4b3 a5c4 a3c1 a7a5 f2f3 a5a4 d3f1 a4b3 a2b3 c8c5 b3c4 e5c4 e1e8 d7e8 c1f4');
INSERT INTO evaluations VALUES(40968,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40968,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40968,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40968,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40968,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40968,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40968,36,'cp 0','b8d7 g1f3 e8g8 c4d5 e6d5 f1d3 f8e8 e1g1 e7d6 f1e1 c7c6 d1c2 c6c5 d4c5 d7c5 h2h3 c5d3 c2d3 f6e4 c1d2 c8h3 g2h3 d8f6 d3e2 f6g6 g1f1 g6f5 f1g2 f5g6');
INSERT INTO evaluations VALUES(44093,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(44093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44093,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(44093,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(44093,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44093,36,'cp 0','g1f3 f8b4 c1d2 b4d2 d1d2 d7d6 f1e2 c8g4 b1c3 g4f3 e2f3 d8b6 c3a4 b6d4 a4c3');
INSERT INTO evaluations VALUES(44706,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(44706,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44706,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44706,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44706,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(44706,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(44706,36,'cp 10','d5c4 a2a4 f6d5 f4e5 f7f6 e5g3 d8a5 d1d2 d5c3 b2c3 b8d7 e2e3 d7b6 g3c7 c8e6 f1e2 f8h6 e1g1 e8g8 d2a2 a5c3 f1c1 c3a5 f3d2 f8c8 c7b6 a7b6 d2c4 g8h8 c4a5 e6a2 a5c6 e7e5 a1a2 e5d4 c1d1 b7c6 d1d4 f6f5 e2f3 a8a5 d4d6');
INSERT INTO evaluations VALUES(45151,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(45151,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(45151,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45151,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45151,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(45151,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45151,36,'cp 0','e7e6 g1f3 a7a6 d2d3 b7b5 c4b3 f8e7 e4e5 d6e5 f4e5 f6d7 c3e4 c8b7 d1e2 b8c6 c1f4 c6a5 e1g1 e8g8 c2c4 a5b3 a2b3 d8b6 f3d2 a6a5 e4c3 b5c4 d2c4 b6b3 a1a3 b3b4 a3a4 b4b3');
INSERT INTO evaluations VALUES(46010,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(46010,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(46010,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46010,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46010,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46010,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46010,36,'cp 0','d2d3 d6d5 c4b3 b8c6 g1f3 b7b5 e4d5 e6d5 c3b5 d8a5 b5c3 d5d4 d1e2 f8e7 f3e5 d4c3 e5c6 c3b2 e1f2 a5c7 h1e1 b2a1q e2e7 c7e7 e1e7 e8f8 e7f7 f8e8 f7e7 e8f8');
INSERT INTO evaluations VALUES(45457,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45457,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45457,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45457,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45457,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45457,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(45457,36,'cp 8','c8g4 e1g1 e7e6 c1f4 b8c6 c2c3 f8d6 f4d6 d8d6 f3e5 g4f5 e2b5 e8g8 d1e2 c6e7 b1d2 e7g6 g2g3 d6c7 d2f3 a7a6 b5d3 g6e7 a2a4 a6a5 f3h4 f5d3 e5d3 f8e8 h4g2 e7f5 g2e3 f5d6 h2h4 a8c8 f1e1 b7b6 e3g4');
INSERT INTO evaluations VALUES(49332,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(49332,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(49332,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(49332,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(49332,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(49332,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(49332,36,'cp 184','d7d5 g1h3 f8e7 d1g4 g7g6 d3e4 d5e4 c1e3 c7c5 d4c5 d8c7 g4f4 e8g8 b1d2 b8d7 e1c1 c7e5 f4e5 d7e5 d2e4 c8d7 f2f3 d7c6 e4d6 e7d6 c5d6 f7f6 b2b3 g8f7 h3f2 a7a5 a2a4 e5d7 c2c4 e6e5 h2h4 h7h5 c1c2 a8c8 c2b2 f8e8 h1e1');
INSERT INTO evaluations VALUES(50527,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50527,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50527,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(50527,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(50527,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(50527,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50527,36,'cp -6','b8c6 d3d4 c5d4 f3d4 d8c7 d4c6 b7c6 d1g4 d7d5 g4g3 c7g3 h2g3 g8f6 c1e3 c8b7 f1d3 f8b4 e3d2 h7h5 f2f3 b4d6 d2f4 d6f4 g3f4 e8e7 c3a4 f6d7 b2b3 h5h4 e1f2 a6a5 a1b1 c6c5 a4b2 b7a6 b1c1 a6d3 b2d3');
INSERT INTO evaluations VALUES(51433,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(51433,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(51433,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51433,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51433,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51433,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51433,36,'cp -17','c8b7 d1e2 c7c5 g2g3 b8c6 f1g2 f8e7 d2d3 g8h6 c2c3 d7d5 e1g1 e8g8 f3e5 f8e8 e5c6 b7c6 b1d2 a8c8 f1d1 d8c7 e4e5 h6f5 d2f3 h7h5 d3d4 g7g6 g2h3 c7d7 b2b3 g8g7 h3f5 e6f5 c1e3 d7e6 a2a3');
INSERT INTO evaluations VALUES(30703,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(30703,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30703,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30703,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(30703,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(30703,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30703,36,'cp -7','b1c3 d7d5 e4d5 e6d5 d2d4 g8f6 f1d3 c7c5 f3e5 c5d4 d3b5 b8d7 e5d7 f6d7 d1d4 a7a6 d4e5 d8e7 b5d7 e8d7 e5e7 f8e7 c1e3 e7h4 g2g3 d5d4 h1g1 d4c3 e1c1 d7c7 g3h4 c3b2 c1b2 g7g6 f4f5 a8d8 d1d8 h8d8 h4h5 b7e4 f5f6 c7c6 a2a4 d8e8 h5g6');
INSERT INTO evaluations VALUES(1233,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(1233,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(1233,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(1233,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(1233,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(1233,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(1233,36,'cp 82','b5a4 g8e7 d2d4 e5d4 e1g1 d7d6 c3d4 c8g4 d4d5 b7b5 d5c6 b5a4 b1c3 g7g6 f1e1 f8g7 c3d5 e7d5 e4d5 e8f8 e1e3 g7h6 e3d3 h6c1 a1c1 f8g7 c1c4 g4f3 d3f3 f6e5 g2g3 h8e8 c4a4 e5b2 f3b3 b2e2 d1e2 e8e2 b3b7 a8c8 a4a6 e2d2 g1g2 d2d5');
INSERT INTO evaluations VALUES(1237,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(1237,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(1237,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(1237,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(1237,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(1237,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(1237,36,'cp 128','d7d6 d2d4 c8d7 e1g1 f6d8 f1e1 g8e7 b1d2 g7g6 d4e5 c6e5 f3e5 d6e5 d2f3 f7f6 a4b3 e7c6 a2a4 d8e7 b3d5 c6d8 b2b4 d7e6 c1e3 e7f7 d1b3 f8d6 f3d2 c7c6 d5e6 f7e6 a4a5 e6b3 d2b3 d8e6 e3b6');
INSERT INTO evaluations VALUES(2018,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2018,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2018,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2018,36,'cp 0','g1e2 c6a5 d2d3 c7c6 b1c3 b7b5 c4b3 g8f6 f3g3 f8g7 e1g1 b5b4 c3a4 d7d5 c1e3 d5d4 e3d2 f6h5 g3g5 g7f6 g5h6 a5b3 a2b3 f6g7 h6g5 g7f6');
INSERT INTO evaluations VALUES(2870,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(2870,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(2870,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(2870,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(2870,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(2870,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(2870,36,'cp 0','e1g1 d7d6 d1d3 c5b6 c4b3 f6d7 c1f4 d7c5 d3g3 g8h8 a1d1 f7f5 e4e5 d6d5 f4e3 c8e6 b3a4 e6d7 a4b3');
INSERT INTO evaluations VALUES(2869,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(2869,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2869,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2869,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(2869,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2869,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2869,36,'cp 0','d7d6 c1g5 h7h6 g5h4 d8e7 d1d3 e7e5 h4g3 e5g5 c3a4 c5b6 a4b6 a7b6 f2f4 g5a5 e4e5 f6h5 f1e1 h5g3 h2g3 f8e8 g1h2 c8g4 e1e3 d6e5 a1e1 e5e4 e3e4 a5h5 h2g1 h5c5 g1h1 c5h5');
INSERT INTO evaluations VALUES(2856,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2856,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2856,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2856,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2856,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2856,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2856,36,'cp 0','h2h3 a7a5 c1f4 c8e6 c4b3 f6d7 d1d2 d8h4 c3a4 c5a7 a1e1 d7e5 f4e5 d6e5 b3e6 f7e6 e1e2 f8f3 d2d7 a8f8 d7e6 f8f7 e6c8 f7f8');
INSERT INTO evaluations VALUES(4410,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(4410,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(4410,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(4410,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(4410,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(4410,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(4410,36,'cp -50','f6e4 c2c3 e4f2 f1f2 c5f2 g1f2 e5e4 f3e1 c6e5 c4e2 c7c6 d5f4 d7d5 d2d4 e4d3 f4d3 e5c4 e1c2 f8e8 f2g1 c8f5 c2d4 f5g6 a2a4 a7a5 d3f4 d8b6 d1b3 b6a7 b3d1 e8e7 f4g6 h7g6 e2c4 d5c4 c1g5 e7e4 d1f1 a7b6 f1f2 a8e8 g5f4');
INSERT INTO evaluations VALUES(4500,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(4500,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(4500,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(4500,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(4500,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(4500,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(4500,36,'cp -47','c2c3 e4f2 f1f2 c5f2 g1f2 e5e4 f3e1 c6e5 c4e2 c7c6 d5f4 d7d5 d2d4 e4d3 f4d3 e5c4 e1c2 f8e8 f2g1 d8b6 c2d4 c6c5 d4f3 d5d4 b2b3 c4e3 c1e3 e8e3 c3d4 c5d4 f3d2 c8g4 e2g4 e3d3 d1e2 d3c3 d2c4 b6c7 a1e1 g7g6 e2f2 d4d3 g4d1 a8d8 f2f6 c7c5 g1f1 c5d4 f6d4 d8d4 e1e8 g8g7');
INSERT INTO evaluations VALUES(5148,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(5148,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(5148,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(5148,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(5148,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(5148,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(5148,36,'cp 175','d4d5 c6a5 c4e2 g8e7 b1c3 e8g8 d5d6 e7c6 h2h4 a7a6 h4h5 b7b5 c1f4 a5c4 e2c4 b5c4 d1d2 c6b4 f4h6 b4d3 e1f1 g7h8 h1h3 a6a5 a1b1 a8b8 f1g1 c8a6 b2b3 c4b3 a2b3 d8f6 h5g6 f7g6 e4e5 d3e5 h6f8 b8f8 c3e4 f6f4 d2f4 f8f4 e4c5 a6b5 f3g5 h8g7 g5h7 b5c6 h7g5');
INSERT INTO evaluations VALUES(12762,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(12762,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(12762,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(12762,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(12762,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(12762,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(12762,36,'cp 34','g1f3 c6d4 e4e5 d4b5 c3b5 f6d5 e1g1 a7a6 c2c4 d5b4 b5c3 d7d6 e5d6 d8d6 c3e4 d6c7 f1e1 c8g4 d2d4 e8c8 e4c5 e7e5 c5b3 e5d4 e1e4 g4h5 c1f4 f8d6 f4d6 c7d6 b3d4 b4c6 d1b3 c6d4 e4d4 d6f6 d4d5 h5f3 g2f3 d8d5 c4d5 h8d8 a1c1 c8b8');
INSERT INTO evaluations VALUES(14251,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14251,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(14251,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(14251,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(14251,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(14251,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14251,36,'cp 50','e2e3 a7a6 c1b2 c8f5 a2a3 e7e6 c2c4 c6e7 b1d2 c7c6 f1e2 e7g6 c4c5 f6e4 e1g1 e4d2 f3d2 h7h5 b3b4 d8f6 f1e1 f8e7 e2f1 f6g5 f2f3 h5h4 d2b3 e8g8 b3a5 a8b8 e3e4 d5e4 f3e4 g6f4 d1c2 g5g6 e4f5 f4h3 g1h1');
INSERT INTO evaluations VALUES(14250,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14250,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(14250,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(14250,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14250,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(14250,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(14250,36,'cp 34','a7a6 c2c4 c8f5 f1d3 e7e6 d3f5 e6f5 e1g1 f8e7 c1b2 e8g8 b1c3 f8e8 a1c1 g7g6 h2h3 e7f8 f3e5 d5c4 e5c6 b7c6 b3c4 a8b8 b2a1 c6c5 d4c5 d8d1 f1d1 f8c5 c3a4 c5e7 a1f6 e7f6 a4c5 b8a8 d1d7 e8e7 g2g4 f5g4 h3g4 e7d7 c5d7');
INSERT INTO evaluations VALUES(14636,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(14636,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(14636,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(14636,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(14636,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(14636,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(14636,36,'cp 30','c8e6 e1g1 e6c4 d3c4 g7g6 d1e2 f8g7 c1e3 e8g8 a1d1 b7b6 b2b4 d8c8 c3d5 g8h7 e2d3 f6g4 e3c1 g4f6 c1b2 c6d8 g2g3 a7a5 g1g2 d8e6 f3d2 c8e8 f2f4 f6d5 c4d5 e5f4 b2g7 e6g7 f1f4 a5b4 a3b4');
INSERT INTO evaluations VALUES(18300,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(18300,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(18300,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(18300,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(18300,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(18300,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(18300,36,'cp 18','c7c6 d5e5 d8e7 c1f4 e7e5 f4e5 f7f5 b1d2 d7d6 e5d4 g8f6 e4e5 d6e5 d4e5 c8e6 e1c1 f8c5 f1c4 e6c4 d2c4 e8g8 e5d6 c5d6 c4d6 f6g4 d1d2 a8d8 h1d1 g7g6 h2h3 g4f6 b2b3 d8d7 f2f3 d7e7 c2c4 f6h5 c1c2 g8g7 b3b4 g7f6 a2a3 h5f4');
INSERT INTO evaluations VALUES(18372,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(18372,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(18372,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(18372,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(18372,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(18372,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(18372,36,'cp 44','d5e5 d8e7 e5e7 f8e7 b1c3 d7d6 c1f4 c8e6 f1a6 e8d7 e1c1 d7c7 h2h3 b6b5 a2a4 b5b4 c3e2 g8f6 e4e5 f6d5 e5d6 e7d6 f4d6 c7d6 e2d4 a8e8 d4b3 d6c7 d1d4 c7b6 a6e2 a7a5 e2f3 e8e7 h1d1 h8e8 h3h4 g7g6 b3d2 d5c7 d2c4 e6c4 d4c4');
INSERT INTO evaluations VALUES(18373,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(18373,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18373,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18373,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(18373,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(18373,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(18373,36,'cp 30','d8e7 e5e7 f8e7 b1c3 g8f6 f2f3 e8g8 c1e3 f8d8 a1d1 d7d5 e4d5 f6d5 c3d5 d8d5 d1d5 c6d5 c2c3 g8f8 e1f2 c8e6 f1b5 e7c5 h1d1 a8c8 b5a6 c5e3 f2e3 c8c5 a6d3 h7h6 d1e1 f8e7 g2g3 c5a5 a2a3 g7g5');
INSERT INTO evaluations VALUES(18568,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(18568,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(18568,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(18568,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(18568,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(18568,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(18568,36,'cp 42','g8f6 b1d2 f6h5 f4e5 e7e6 f1d3 f8e7 d1e2 h5f6 e5f6 d7f6 e3e4 d5e4 d2e4 c8d7 e1c1 f6e4 d3e4 d8a5 c1b1 e8c8 f3e5 d7e8 e2g4 e7f6 f2f4 h7h5 g4e2 c8b8 c2c3 h5h4 e4f3 a5c7 b1a1');
INSERT INTO evaluations VALUES(12242,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(12242,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(12242,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(12242,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(12242,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(12242,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(12242,36,'cp 52','c2c4 e7e6 b1c3 f6h5 f4g5 f8e7 g5e7 d8e7 f1d3 h5f6 e1g1 e8g8 f1e1 d5c4 d3c4 b7b5 c4f1 c8b7 a2a3 e6e5 h2h3 a7a6 b2b4 a6a5 d1b3 a5b4 a3b4 g7g6 d4e5 d7e5 f3d4 f8c8 a1a8 c8a8 e1c1 e5d7 b3b1 a8d8 b1b2 d8e8 f1e2 d7b6 e2f3');
INSERT INTO evaluations VALUES(12234,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(12234,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(12234,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(12234,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(12234,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(12234,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12234,36,'cp 33','e7e6 b1c3 f6h5 f4g5 f8e7 g5e7 d8e7 f1d3 h5f6 e1g1 e8g8 e3e4 d5c4 d3c4 b7b5 c4d3 e6e5 a2a4 b5b4 c3e2 c8b7 a4a5 c6c5 d4d5 c5c4 d3c4 f6e4 e2g3 e4d6 a5a6 b7c8 c4e2 d7f6 f1e1 f8d8 d1b3 c8d7 b3b4 e5e4 f3d4 f6d5');
INSERT INTO evaluations VALUES(12155,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(12155,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(12155,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(12155,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(12155,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(12155,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12155,36,'cp 30','b1c3 f8e7 f1d3 d5c4 d3c4 e8g8 e1g1 b7b5 c4d3 c8b7 a1c1 a7a6 c3e4 f6e4 d3e4 a8c8 f3e5 d7e5 f4e5 d8d7 d1g4 f7f6 g4h3 h7h6 e5f4 f6f5 e4f3 c6c5 d4c5 c8c5 c1c5 e7c5 f1d1 d7e7 h3g3 b7f3 g3f3 f8d8 h2h3 d8d1 f3d1 g8f7 f4e5 c5b6');
INSERT INTO evaluations VALUES(19120,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(19120,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(19120,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(19120,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(19120,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(19120,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(19120,36,'cp 34','d8c7 c1e3 g8f6 b1c3 f8d6 b5d3 d6e5 d4b4 e5c3 b4c3 c7c3 b2c3 d7d6 a2a4 c8d7 a4a5 d7c6 f2f3 h7h6 h2h4 f6d7 g2g4 e8e7 e1e2 f7f6 a1b1 e7f7 e3d4 a8c8 h1g1 c8a8 e2f2 f7e7 c3c4 e6e5 d4c3 d7c5 c3b4 a8c8 d3e2');
INSERT INTO evaluations VALUES(35180,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(35180,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(35180,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(35180,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(35180,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(35180,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(35180,36,'cp 68','g1f3 h4e4 f1e2 e4f5 b1c3 e5e4 f3h4 f5f6 g2g3 f6d8 e1g1 d7d6 c3e4 f8e7 h4g2 g8f6 e4c3 d6d5 c4d5 f6d5 f1e1 e8g8 d2d4 d5f6 e2f3 e7d6 c1g5 h7h6 g5f6 d8f6 c3e4 f6d8 e4d6 d8d6 g2e3 b8c6 d4d5 c6e5 f3g2 c8d7 d1d4 f8e8 a1c1 a8c8');
INSERT INTO evaluations VALUES(39197,30,'cp 348',NULL);
INSERT INTO evaluations VALUES(39197,31,'cp 336',NULL);
INSERT INTO evaluations VALUES(39197,32,'cp 344',NULL);
INSERT INTO evaluations VALUES(39197,33,'cp 350',NULL);
INSERT INTO evaluations VALUES(39197,34,'cp 357',NULL);
INSERT INTO evaluations VALUES(39197,35,'cp 352',NULL);
INSERT INTO evaluations VALUES(39197,36,'cp 354','h5e5 c5e7 e5g7 e7f6 g7g3 b8c6 g1f3 g8e7 b1c3 d7d6 f1b5 c8d7 b5c6 b7c6 d2d4 h8g8 g3f4 f6h8 c1e3 g8g2 f3g5 f7f6 g5h7 e7g6 f4f3 g2g4 h2h3 g4h4 h1g1 d8e7 h7f6 e7f6 f3f6 h8f6 g1g6 e8f7 g6g3 a8h8 e1d2 h4h3 a1g1 h3g3 g1g3 h8h2 g3f3 f7e7 b2b3 f6h4 c3e2 e7e8 d2d3 d6d5 a2a4 d7h3 e4d5');
INSERT INTO evaluations VALUES(39195,30,'cp 367',NULL);
INSERT INTO evaluations VALUES(39195,31,'cp 362',NULL);
INSERT INTO evaluations VALUES(39195,32,'cp 350',NULL);
INSERT INTO evaluations VALUES(39195,33,'cp 361',NULL);
INSERT INTO evaluations VALUES(39195,34,'cp 336',NULL);
INSERT INTO evaluations VALUES(39195,35,'cp 355',NULL);
INSERT INTO evaluations VALUES(39195,36,'cp 355','c5e7 e5g7 e7f6 g7g3 b8c6 b1c3 g8e7 g1f3 d7d6 f1b5 h8g8 g3f4 a7a6 f4f6 a6b5 d2d4 b5b4 c3e2 e7g6 f6d8 e8d8 c1g5 d8e8 f3d2 h7h6 g5f6 b7b6 d4d5 c6e5 f2f3 c8a6 g2g3 e5d7 f6d4 f7f5 a2a3 a6e2 e1e2 b4a3 a1a3 a8a3 b2a3 e8f7 e4f5 g6e7 g3g4');
INSERT INTO evaluations VALUES(40654,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40654,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40654,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40654,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40654,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40654,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40654,36,'cp 0','a2a3 e8g8 b1c3 b8c6 b2b4 c5d6 c3b5 d6e7 f1d3 f6e4 c1b2 a7a5 e1g1 a5b4 a3b4 a8a1 d1a1 c6b4 d3e2 b4c2 a1a2 c2b4');
INSERT INTO evaluations VALUES(44318,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44318,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44318,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44318,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44318,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44318,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44318,36,'cp 0','b8c6 f1e2 f8e7 e1g1 e8g8 d2d3 a7a6 a2a4 a8b8 f3e1 c6d4 e1c2 d4b3 a1b1 d7d6 c2e3 b3d4 e3c2');
INSERT INTO evaluations VALUES(47101,30,'cp -185',NULL);
INSERT INTO evaluations VALUES(47101,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(47101,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(47101,33,'cp -164',NULL);
INSERT INTO evaluations VALUES(47101,34,'cp -182',NULL);
INSERT INTO evaluations VALUES(47101,35,'cp -180',NULL);
INSERT INTO evaluations VALUES(47101,36,'cp -190','e7e6 c1g5 h7h6 f1b5 c8d7 g5f6 d8f6 d5e6 f6e6 d1d4 b8c6 d4e4 e6e4 c3e4 a7a6 b5a4 f7f5 e4c3 e8c8 g1e2 f8c5 e1g1 h8e8 a4b3 g7g5 f1e1 f5f4 a1d1 d7f5 d1d8 e8d8 c3d5 f5c2 b3c2 d8d5 e2c3 d5d8 g1f1 c5b4 c2f5 c8b8 h2h4 g5h4 f5g4 d8d2 e1e2 b4c3 b2c3 d2e2');
INSERT INTO evaluations VALUES(48190,30,'cp -194',NULL);
INSERT INTO evaluations VALUES(48190,31,'cp -166',NULL);
INSERT INTO evaluations VALUES(48190,32,'cp -185',NULL);
INSERT INTO evaluations VALUES(48190,33,'cp -186',NULL);
INSERT INTO evaluations VALUES(48190,34,'cp -187',NULL);
INSERT INTO evaluations VALUES(48190,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(48190,36,'cp -175','c1g5 h7h6 f1b5 c8d7 g5f6 d8f6 d5e6 f6e6 d1d4 b8c6 d4e4 e6e4 c3e4 a7a6 b5d3 c6b4 g1f3 b4d3 c2d3 e8c8 e1c1 g7g5 c1c2 f8g7 h1e1 d7e6 a2a3 b7b6 e1e3 g5g4 f3d2 d8d5 f2f3 h6h5 e4c3 d5d8 d1e1 h5h4 c3e4 g7h6');
INSERT INTO evaluations VALUES(47643,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(47643,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(47643,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(47643,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(47643,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(47643,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(47643,36,'cp 48','f8e7 b1c3 e8g8 c1f4 f8e8 f1e1 e7f8 h2h3 b8d7 a2a4 c7c6 a4a5 a7a6 f4g3 d7c5 d1f3 c5e6 c4e6 c8e6 g3h4 g7g5 h4g3 f6d7 f3e3 d8f6 f2f3 a8d8 c3a4 f6g6 d4e6 g6e6 a1d1 d6d5 e3c3 d8c8 e4d5 c6d5 e1e6');
INSERT INTO evaluations VALUES(40700,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(40700,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(40700,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(40700,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(40700,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(40700,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(40700,36,'cp 63','f1e1 e8g8 c1f4 f8e8 b1c3 b8d7 h2h3 e7f8 a2a4 c7c6 a4a5 a7a6 f4g3 b7b5 a5b6 d8b6 d4b3 d7e5 c3a4 b6b8 c4f1 a6a5 b3d2 b8a7 g1h1 g7g5 h1g1 g5g4 a4c3 g4h3 g2h3 a7b8 b2b3 f8g7 d2f3 b8c7 f1g2 c8d7 f3e5 d6e5 d1f3');
INSERT INTO evaluations VALUES(40701,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(40701,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(40701,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(40701,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(40701,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(40701,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(40701,36,'cp 46','e8g8 b1c3 b8c6 h2h3 c6d4 d1d4 c7c6 a2a4 d8b6 c1e3 b6d4 e3d4 f8e8 c4f1 c8e6 f2f4 d6d5 f4f5 e6c8 g2g4 f6e4 c3e4 d5e4 e1e4 c8d7 e4e5 a7a5 b2b3 e7h4 f1d3 e8e5 d4e5 d7e8 g1f1 g8f8 e5c7 h4d8 c7g3 d8f6 a1d1 a8d8 f1e2');
INSERT INTO evaluations VALUES(36192,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(36192,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(36192,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(36192,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36192,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(36192,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(36192,36,'cp 60','b1c3 b8c6 h2h3 c6d4 d1d4 c7c6 a2a4 d8b6 d4d3 c8e6 c4e6 f7e6 a4a5 b6c7 d3g3 g8h8 e4e5 d6e5 g3e5 e7d6 e5d4 e6e5 d4c4 b7b5 a5b6 a7b6 c1e3 b6b5 c4d3 d6b4 a1a8 f8a8 e1d1 c7f7 c3e4 f6d5 c2c3 b4f8 d3e2 f8e7 e4g3 e7f8 e2g4');
INSERT INTO evaluations VALUES(36191,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(36191,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(36191,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(36191,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36191,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(36191,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(36191,36,'cp 64','b8c6 h2h3 c6d4 d1d4 c7c6 a2a4 d8b6 d4d3 c8e6 c4e6 f7e6 a4a5 b6c7 d3g3 g8h8 e4e5 d6e5 g3e5 e7d6 e5d4 e6e5 d4c4 b7b5 a5b6 a7b6 c1e3 b6b5 c4d3 d6b4 a1a8 f8a8 e1d1 c7f7 c3e4 f6d5 e4g3 a8e8 c2c3 b4f8 d1a1 f8d6 d3e4');
INSERT INTO evaluations VALUES(57043,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57043,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57043,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(57043,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(57043,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(57043,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57043,36,'cp 16','d7d5 e4d5 e6d5 g1f3 f8d6 d2d4 g8f6 f1d3 e8g8 e1g1 h7h6 b1c3 b8c6 c3b5 c6b4 b5d6 d8d6 d3b5 c8f5 b5a4 b4c6 f1e1 c6e7 f3e5 e7g6 d1f3 f5e4 f3c3 c7c6 f2f3 e4f5 c1d2 f8e8 e5g6 e8e1 a1e1 f5g6 c3a5 b7b5 a4b3 g6f5 a5c3 d6c7');
INSERT INTO evaluations VALUES(55914,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(55914,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55914,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55914,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(55914,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(55914,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(55914,36,'cp 13','e4d5 e6d5 d2d4 f8d6 g1f3 g8f6 f1d3 h7h6 b1c3 e8g8 e1g1 b8c6 c3b5 c6b4 b5d6 d8d6 d3b5 c8f5 b5a4 b4c6 c2c3 a7a6 a4c6 d6c6 f3e5 c6b6 d1b3 b6b3 a2b3 f8e8 f1e1 f5c2 b3b4 f6e4 c1f4 e4d6 e5f3 d6c4');
INSERT INTO evaluations VALUES(4514,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(4514,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(4514,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(4514,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(4514,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(4514,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(4514,36,'cp 131','d2d4 f8e8 b5c6 b7c6 c1g5 h7h6 g5h4 g7g5 h4g3 d6f8 d4e5 f6h5 b1d2 d7d6 e5d6 c7d6 d2f1 g5g4 f3d4 c8d7 d1d3 d8g5 f1e3 h5f6 e3f5 d6d5 e4d5 f6d5 f5d6 f8d6 g3d6 c6c5 d4b3 d5b4 d3c3 d7f5 b3c5 b4c2 e1e8 a8e8 a1d1 g5g7 c3d2 g7f6 b2b3');
INSERT INTO evaluations VALUES(14047,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(14047,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(14047,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(14047,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(14047,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(14047,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(14047,36,'cp 91','f7f6 e5f6 g7f6 f3h4 d8d7 d1h5 e8d8 c1e3 d7e8 h5f3 e8f7 e1c1 c8d7 c1b1 e7c8 g2g4 c8d6 h2h3 f8e7 f3g2 a7a5 e3c1 a5a4 a2a3 a8b8 c3a4 b7b6 f2f4 c6d4 d1d4 c7c5 d4d1 d7a4 f4f5 e6e5 d1d5 d8c7 f1e2 a4c6 h1d1 h8g8 e2f3');
INSERT INTO evaluations VALUES(14373,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(14373,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(14373,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(14373,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(14373,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(14373,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14373,36,'cp 0','a7a6 d4d5 e7e5 f4g3 c6a5 g3e5 b7b5 d1d4 f8e7 g2g3 e8g8 f1g2 b5b4 c3e4 f6d5 a1d1 c8e6 e4c5 e7c5 d4c5 a5b7 c5c6 b7a5');
INSERT INTO evaluations VALUES(17476,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(17476,31,'cp -108',NULL);
INSERT INTO evaluations VALUES(17476,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(17476,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(17476,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(17476,35,'cp -111',NULL);
INSERT INTO evaluations VALUES(17476,36,'cp -105','e7e6 d2d3 d7d5 c4b3 f8e7 g5f3 d5e4 d3e4 d8d1 e1d1 g8f6 b1d2 b7b5 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 c2c3 c5c4 b3c2 a8b7 d1e2 e8g8 h1d1 f8a8 c2b1 h7h6 d2f1 c6a5 f3d2 b7c6 f1e3 h6h5 b1c2 g7g5 e3f1 h5h4 g2g3 b5b4 c3b4 e7b4 g3h4 g5h4 f2f4 f6g4 f1e3 g4h2 d2c4');
INSERT INTO evaluations VALUES(17348,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(17348,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(17348,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(17348,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(17348,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(17348,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(17348,36,'cp -97','d2d3 d7d5 c4b3 h7h6 g5f3 d5e4 d3e4 d8d1 e1d1 g8f6 b1d2 g7g5 c2c3 f8d6 b3c2 d6c7 a2a4 g5g4 f3e1 b7b6 d2c4 c8b7 c1e3 e8c8 d1e2 a6a5 c2d3 f6h5 a1d1 c6e7 d3c2 f7f5 d1d8 c8d8 f2f3 g4g3 h2h4 f5f4 e3d2 e7g6 e1d3 b7a6 b2b3 d8e7 d3b2 g6e5 c2d3');
INSERT INTO evaluations VALUES(17347,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(17347,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(17347,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(17347,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(17347,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(17347,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(17347,36,'cp -93','f8e7 g5f3 d7d5 c4b3 g8f6 d1e2 b7b5 a2a4 c6a5 b3a2 c8b7 e1g1 d5e4 d3e4 c5c4 b1d2 e8g8 a4b5 a6b5 a2b1 d8c7 f3d4 c7b6 c2c3 a5c6 a1a8 f8a8 d4f3 f6d7 f1d1 d7c5 d2f1 b5b4 c1e3 a8a1 f1d2 b4c3 b2c3 b6b2 e3c5 e7c5 e4e5 b2c3 e2c4 c3b4 c4b4 c6b4');
INSERT INTO evaluations VALUES(17489,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(17489,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(17489,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(17489,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(17489,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(17489,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(17489,36,'cp -15','e7e6 d2d4 c5d4 f3d4 g8f6 b1c3 d8c7 c4e2 b7b5 d4c6 d7c6 f2f4 c8b7 e4e5 f6d7 e1g1 a8d8 d1e1 f8e7 e2f3 c6c5 c3e4 e8g8 c1d2 b5b4 g1h1 a6a5 h2h3 d7b6 a3b4 a5b4 e1e2 b7e4 f3e4 d8d4 e4f3 b6d5');
INSERT INTO evaluations VALUES(18998,30,'cp 301',NULL);
INSERT INTO evaluations VALUES(18998,31,'cp 301',NULL);
INSERT INTO evaluations VALUES(18998,32,'cp 303',NULL);
INSERT INTO evaluations VALUES(18998,33,'cp 312',NULL);
INSERT INTO evaluations VALUES(18998,34,'cp 321',NULL);
INSERT INTO evaluations VALUES(18998,35,'cp 317',NULL);
INSERT INTO evaluations VALUES(18998,36,'cp 318','d4d5 c6a5 f3e5 g8f6 b1c3 d7d6 f1b5 f6d7 e5d7 c8d7 b5d7 d8d7 d1d3 f8e7 c3e2 e8g8 c1d2 e7d8 e1g1 f7f5 a1e1 f5e4 d3e4 b7b5 b2b3 a5b7 c2c4 b5c4 e4c4 d8a5 d2a5 b7a5 c4d3 c5c4 d3g3 c4b3 a2b3 a5b7 e2d4 b7c5 d4e6');
INSERT INTO evaluations VALUES(19042,30,'cp 302',NULL);
INSERT INTO evaluations VALUES(19042,31,'cp 311',NULL);
INSERT INTO evaluations VALUES(19042,32,'cp 312',NULL);
INSERT INTO evaluations VALUES(19042,33,'cp 326',NULL);
INSERT INTO evaluations VALUES(19042,34,'cp 358',NULL);
INSERT INTO evaluations VALUES(19042,35,'cp 323',NULL);
INSERT INTO evaluations VALUES(19042,36,'cp 329','c6a5 f3e5 a7a6 e5f3 d7d6 a2a3 b7b5 b2b4 a5b7 c2c4 b5c4 f1c4 f8e7 e1g1 g8f6 b1d2 e8g8 h2h3 f6d7 c1b2 e7f6 b2f6 d8f6 c4e2 d7e5 f3e5 f6e5 f1e1 e5g5 d2f3 g5g6 g1h1 c8d7 e4e5 g6h6 a1c1 d7b5 e2b5 a6b5 e5e6 f7e6');
INSERT INTO evaluations VALUES(24181,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(24181,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(24181,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(24181,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(24181,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(24181,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(24181,36,'cp 37','b5d7 d8d7 e1g1 h7h6 g5f3 f6d5 f3e5 d7e6 f1e1 e8c8 e5f3 e6f5 d2d3 a5c6 b1c3 g7g5 c3d5 f5d5 e2e4 d5e4 e1e4 f7f5 e4e6 g5g4 f3e1 h6h5 e6h6 h8h6 c1h6 c6d4 h6e3 d4e6 e3d2 b7b5 c2c3 c7c5 h2h3 f5f4 g1f1 a7a5 a1d1 d8f8 h3g4 h5g4');
INSERT INTO evaluations VALUES(25253,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(25253,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(25253,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(25253,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(25253,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(25253,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(25253,36,'cp 7','d3d4 e5d4 f3d4 c6d4 d1d4 g4e6 b1c3 c7c6 e2e4 d5e4 c1e3 d8d4 e3d4 f8e7 a2a4 e8g8 c3e4 f6e4 g2e4 f8d8 d4e3 a7a5 a1d1 g7g6 e3b6 d8e8 f1e1 e7d8 b6d4 f7f6 e4d3 g8f7 d4e3 d8c7 d1d2 a8d8 b2b3 d8d7 h2h4 e8d8 e1d1');
INSERT INTO evaluations VALUES(26780,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(26780,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(26780,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(26780,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26780,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(26780,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26780,36,'cp 18','g8f6 c2c4 f8e7 b1c3 e8g8 c1e3 f8e8 d4d3 f6g4 e3d4 g4e5 d3e2 e7f8 a1e1 a7a6 h2h3 a8b8 b2b3 b7b5 f3d2 c6b7 f2f4 e5d7 e2d3 c7c6 g1h2 d8e7 a2a3 d7f6 b3b4 c6c5 d4f2 h7h6');
INSERT INTO evaluations VALUES(25096,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(25096,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(25096,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(25096,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25096,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25096,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(25096,36,'cp 25','c2c4 f8e7 b1c3 e8g8 f1e1 f6d7 c3d5 e7f6 d5f6 d8f6 d4f6 d7f6 e4e5 c6f3 e5f6 f3c6 f6g7 g8g7 f2f3 f8e8 c1d2 g7g6 g1f2 h7h5 h2h4 e8e1 a1e1 a8e8 e1d1 b7b6 d2c3 a7a5 g2g4 h5g4 f3g4 f7f5 d1d4 c6d7 h4h5 g6g5 g4f5 d7f5 d4d5');
INSERT INTO evaluations VALUES(27921,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(27921,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(27921,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(27921,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(27921,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(27921,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(27921,36,'cp 143','h2h4 h7h6 h4h5 g6g5 d1d2 d7d5 e4d5 e6d5 f3e5 b8c6 f2f4 c6e5 f4e5 f7f5 e1c1 g8e7 g2g3 d8d7 f1b5 e7c6 c3e2 e8c8 c1b1 c8b8 d2c3 h8f8 h1f1 d7e8 d1e1 a7a6 b5d3 b7c8 e3c1 c6e7 f1h1 e7c6 c3d2 f8f7 c2c3 g7f8 b1a1 c6a5 d2c2 f8e7');
INSERT INTO evaluations VALUES(29367,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(29367,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(29367,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29367,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29367,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(29367,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(29367,36,'cp 45','f4b8 d8b8 d1a4 e8d8 f3e5 f5g6 h2h4 b8c7 b1a3 c5d4 e3d4 f6d7 a3b5 c7b8 e5d7 d8d7 h4h5 g6f5 b5a3 d7d8 a3c2 f5c2 a4c2 b8f4 g2g3 f4f3 h1h2 a8c8 f1e2 f3f6 e1c1 g7g5 c1b1 d8c7 d1f1 f8d6 f2f4 g5f4 g3f4 a7a6 e2d3 c7b8 d3h7 f6h6 h7d3 d6f4 h2f2 f4d6');
INSERT INTO evaluations VALUES(32082,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(32082,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(32082,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32082,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(32082,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(32082,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(32082,36,'cp 79','c7c6 a2a3 b8d7 c4a2 f8e7 d1e2 h5g6 b1c3 g8f6 f3h4 a7a5 g2g4 d7c5 e2f3 c5e6 c1e3 f6d7 h4g6 h7g6 g4g5 e7g5 a2e6 f7e6 f3g4 g5e3 g4g6 e8f8 f2e3 d8e7 h3h4 b7b5 c3e2 f8g8 g6g2 a8f8 e1c1 b5b4 a3a4 h8h4 d1g1 h4h1 g1h1');
INSERT INTO evaluations VALUES(36738,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(36738,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(36738,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(36738,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(36738,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(36738,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(36738,36,'cp 20','c4d5 f6d5 e1g1 d5b6 b1c3 b8c6 d4d5 c6a5 e2e4 c7c6 f1e1 f8e8 c1f4 c6d5 e4d5 a5c4 d1c1 c8f5 b2b3 c4d6 f4e5 g7e5 e1e5 b6d7 e5e3 d7f6 c1e1 f6g4 e3e2 g4f6 h2h3 h7h5 f3d4 a8c8 e1d2 d8b6 a1c1 c8c7 e2e3 e8c8 c1e1 b6a5');
INSERT INTO evaluations VALUES(36711,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(36711,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36711,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36711,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(36711,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(36711,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36711,36,'cp 37','f6d5 e1g1 d5b6 e2e3 b8c6 b1c3 e7e5 d4d5 e5e4 d5c6 d8d1 f1d1 e4f3 g2f3 b7c6 f3c6 a8b8 c6b5 c8e6 a2a4 b6c4 c3d5 e6g4 d5e7 g8h8 d1d5 c4b2 c1b2 g7b2 a1f1 b2a3 e7c6 b8a8 d5d2 a3d6 g1g2 g4e6 c6d4 d6b4 d4e6 b4d2 e6f8 a8f8');
INSERT INTO evaluations VALUES(36816,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(36816,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36816,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(36816,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(36816,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(36816,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(36816,36,'cp 37','e1g1 d5b6 e2e3 b8c6 b1c3 e7e5 d4d5 e5e4 d5c6 d8d1 f1d1 e4f3 g2f3 b7c6 f3c6 a8b8 c6b5 c8g4 d1d2 g4e6 d2c2 b6d5 c3d5 e6d5 b5c6 d5c6 c2c6 g7b2 c1b2 b8b2 c6c7 f8d8 a2a4 d8d2 a1f1 b2b4 a4a5 d2a2 c7a7 b4a4 f1b1 a4a5 a7a5 a2a5');
INSERT INTO evaluations VALUES(36815,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(36815,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36815,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(36815,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36815,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(36815,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(36815,36,'cp 43','d5b6 b1c3 b8c6 e2e3 e7e5 d4d5 e5e4 d5c6 d8d1 f1d1 e4f3 g2f3 b7c6 f3c6 a8b8 c6b5 c8e6 a2a4 b6c4 c3d5 e6g4 d1f1 c4d6 e3e4 d6b5 a4b5 b8b5 a1a7 g4f3 f1e1 g7b2 c1b2 b5b2 a7a3 b2e2 e1e2 f3e2 a3c3 f8a8 g1g2 a8a1 d5f6 g8g7 f6e8 g7f8 e8c7');
INSERT INTO evaluations VALUES(44478,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44478,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(44478,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(44478,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(44478,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44478,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(44478,36,'cp 16','d8b6 b2b3 d7d6 h2h3 c8f5 a2a3 e7e6 e2e3 b8d7 f1d3 f5d3 d1d3 f8e7 b1d2 d6d5 c2c4 b6a6 d3c2 c6c5 e1g1 c5d4 e3d4 b7b6 a3a4 a6a5 f3e5 a8c8 d2f3 d7e5 f3e5 e8g8 f1d1 f8d8 f4g5 h7h6 g5h4 d8e8 c2d3 f6e4 h4e7 e8e7');
INSERT INTO evaluations VALUES(45656,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(45656,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(45656,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(45656,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(45656,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(45656,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45656,36,'cp 11','g1f3 b8d7 b1d2 f8e7 b2b4 e8g8 c4c5 b7b6 c1b2 a7a5 b2c3 f6e4 d2e4 d5e4 f3d2 d7f6 d1c1 f6d5 d2e4 d5c3 c1c3 a5b4 a3b4 a8a1 c3a1 b6c5 b4c5 e6e5 d4e5 d8d5 f2f3 c8f5 e4d6 f5e6 e1f2 d5c5 f1e2 e7d6 e5d6 c5d6 a1d4 d6c7 h1c1 c7h2 c1c6 f8b8 c6c1');
INSERT INTO evaluations VALUES(46573,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(46573,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(46573,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(46573,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(46573,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(46573,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(46573,36,'cp 70','g7g6 g2g3 f8g7 f1g2 c7c6 e1g1 e8g8 b2b4 e7e5 d4e5 d6e5 e2e4 d8d1 f1d1 c8e6 e4f5 g6f5 b4b5 f8e8 g2f1 b8d7 f3g5 d7f8 b5c6 b7c6 a1b1 a8d8 b1b7 e5e4 c4c5 f6d5 g5e6 f8e6 c3d5 d8d5 d1d5 c6d5 b7d7 e8d8 d7a7 e6c5 c1e3 d5d4 f1c4 g8h8');
INSERT INTO evaluations VALUES(48876,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(48876,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(48876,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(48876,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(48876,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(48876,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(48876,36,'cp -67','f8d6 e3e4 d5e4 d3e4 e8g8 b1c3 d8e7 g1f3 d6b4 e1g1 b4c3 b2c3 f8d8 d1e2 h7h6 f3h4 e7c5 e2f3 c8g4 f3e3 c5e3 c1e3 g4e6 f1e1 b8c6 f2f4 d8e8 h4f3 b7b6 a2a4 e6c4 a1d1 f6g4 e3c1 a8d8 h2h3 g4f6 d1d8 e8d8 f3e5 c6e5 f4e5');
INSERT INTO evaluations VALUES(49417,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(49417,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(49417,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(49417,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(49417,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(49417,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(49417,36,'cp -13','b7b6 g1f3 c8b7 b2b4 d7d5 f1e2 c7c5 b4c5 b6c5 e1g1 b8d7 b1d2 f8e7 d4c5 a8c8 c1b2 e8g8 c3c4 d5c4 a1c1 c8c5 e2c4 c5c7 c4b5 c7c1 d1c1 d8b8 b5e2 f8c8 c1a1 b8a8 f1c1 b7d5 c1c8 a8c8 h2h3 c8c2 e2d1 c2d3');
INSERT INTO evaluations VALUES(49792,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(49792,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(49792,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(49792,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(49792,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(49792,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(49792,36,'cp 83','d7d6 f2f4 b8c6 g1f3 c6a5 b1c3 c7c6 h2h3 f8e7 f4f5 b7b5 c1h6 g7h6 c4b3 e7h4 e1f1 b5b4 f3h4 d8h4 c3e2 h8g8 d1d2 a5b3 a2b3 c6c5 g2g3 h4f6 g3g4 h6h5 f1f2 f6h4 f2g2 h5g4 e2g3 g4h3 h1h3 h4f4 d2f2 f4f2 g2f2');
INSERT INTO evaluations VALUES(53763,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(53763,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(53763,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(53763,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(53763,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(53763,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(53763,36,'cp 171','d2d4 d5e4 c3e4 g8f6 d1e2 c8e6 e4g5 e5e4 g5e4 h7h6 e4f6 d8f6 g2g3 f8e7 f1g2 e8g8 e1g1 b8d7 c1d2 f8e8 c2c4 e6g4 e2d3 g4f5 d3b3 a7a5 d2f4 e7d6 f4d6 f6d6 b3c3 d6b4 f1e1 d7b6 f3e5 a8d8 c3b4 a5b4');
INSERT INTO evaluations VALUES(53759,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(53759,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(53759,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(53759,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(53759,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(53759,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(53759,36,'cp 147','d5e4 c3e4 g8f6 d1e2 c8e6 e4g5 e5e4 g5e4 h7h6 e4f6 d8f6 g2g3 f8e7 c1d2 e8g8 f1g2 b8d7 e1g1 f8e8 c2c4 e6g4 e2d3 f6f5 d3f5 g4f5 f1e1 e7f6 h2h3 e8e1 a1e1 a8d8 d2e3 a7a5 e3d2 a5a4 d2c3 c6c5 d4d5 f6c3 b2c3 f5d3 g2f1 d3f1 g1f1');
INSERT INTO evaluations VALUES(55416,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(55416,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(55416,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(55416,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(55416,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(55416,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(55416,36,'cp 70','d2d4 b8c6 c1e3 g8f6 h2h3 g7g6 b1d2 f8g7 f1d3 e8g8 e1g1 a7a5 f1e1 a5a4 d1c2 c8d7 a1d1 f8e8 d4d5 c6e7 c3c4 b7b6 d2f1 f6h5 f1g3 h5g3 f2g3 f7f5 g1h2 e8f8 b2b4 a4b3 a2b3 c7c5 c2d2');
INSERT INTO evaluations VALUES(55410,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(55410,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(55410,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(55410,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(55410,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(55410,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(55410,36,'cp 79','b8c6 c1e3 g8f6 f1d3 a7a6 d3c2 c8d7 e1g1 g7g5 a2a4 f6g4 e3c1 f8g7 a4a5 e8g8 h2h3 g4f6 b1d2 e5d4 c3d4 c6b4 c2b1 d7b5 a1a3 b5f1 d2f1 c7c5 d4d5 c5c4 f1g3 f8e8 f3d4 b4d3 b1d3 c4d3 d1d3');
INSERT INTO evaluations VALUES(5682,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(5682,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(5682,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(5682,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(5682,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(5682,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(5682,36,'cp 95','f1b5 g8h6 e1g1 e8g8 h2h3 f7f6 f1e1 c8f5 c2c3 h6f7 b5c6 b7c6 c1f4 a7a5 b1d2 a5a4 d1e2 g6g5 f4g3 d8b8 b2b4 a4b3 a2b3 b8b5 e2d1 b5d3 e1e3 d3c2 d1e1 g7h6 a1a8 f8a8 e5f6 e7f6 g3c7');
INSERT INTO evaluations VALUES(5674,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(5674,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(5674,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(5674,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(5674,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(5674,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(5674,36,'cp 125','g8h6 e1g1 e8g8 h2h3 f7f6 f1e1 h6f7 b1c3 f6e5 b5c6 e5d4 f3d4 e7e5 d4e6 c8e6 c6b7 d8h4 b7a8 f8a8 c3e2 c7c6 d1d3 a8f8 d3g3 h4e7 c1d2 g7f6 g3c3 f8c8 e2g3 e5e4 c3e3 f6b2 a1b1 b2e5 d2c3 e7a3 c3e5 a3e3 f2e3 f7e5');
INSERT INTO evaluations VALUES(6265,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6265,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6265,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6265,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6265,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(6265,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6265,36,'cp 0','b4c3 d2c3 d7d6 c1e3 d8f6 d1d3 g8e7 e1c1 f6g6 h1g1 c8g4 b2b4 e8c8 c1b2 c8b8 h2h3 g4e6 c4b3 a7a6 f3h4 g6f6 h4f3');
INSERT INTO evaluations VALUES(6275,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6275,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6275,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6275,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6275,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6275,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6275,36,'cp 12','d2c3 d7d6 c1e3 d8f6 d1d3 g8e7 e1c1 f6g6 h1g1 c8g4 h2h3 g4e6 c4e6 g6e6 f3d2 f7f5 f2f3 d6d5 e4d5 e6d5 d2c4 e8c8 d3e2 d5e6 g2g4 d8d1 g1d1 h8d8 d1d8 c8d8 b2b3 b7b6 c1b2 e7d5 a3a4 c6e7 e3g1 e5e4 c4e3 d5f4 e2d2 e6d6 d2d6 c7d6');
INSERT INTO evaluations VALUES(6273,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(6273,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(6273,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(6273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6273,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(6273,36,'cp 0','d8f6 d1d3 g8e7 c1e3 d7d6 e1c1 f6g6 h1g1 c8g4 b2b4 e8c8 c1b2 a7a6 f3h4 g6f6 h4f3');
INSERT INTO evaluations VALUES(13585,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(13585,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(13585,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(13585,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13585,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13585,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13585,36,'cp 0','f1d3 b7b6 e1g1 c8b7 b1d2 a8c8 d4c5 f8c5 a2a3 a7a5 d1e2 e8g8 f1d1 h7h6 h2h3 f8e8 d3b5 d8e7 c2c4 e8d8 c4d5 d8d5 b5c4 d5d8 c4b5');
INSERT INTO evaluations VALUES(14421,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(14421,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14421,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(14421,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(14421,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(14421,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(14421,36,'cp 36','g1f3 c8g4 f1b5 a7a6 b5e2 d5c3 b2c3 e7e6 e1g1 f8e7 c3c4 e8g8 c2c3 g4h5 f1e1 b7b6 h2h3 h5g6 c1e3 h7h6 e2f1 e7d6 c4c5 d6e7 c5b6 c7b6 c3c4 e7f6 f1d3 g6d3 d1d3 b6b5 c4b5 a6b5 d3b5');
INSERT INTO evaluations VALUES(15455,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(15455,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(15455,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(15455,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(15455,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15455,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(15455,36,'cp 51','c6e7 f1d3 e7g6 h2h3 h7h6 c3c4 g6f4 c1f4 e5f4 c4c5 f6d7 d3b5 f8e7 e1g1 e8g8 b5d7 c8d7 f1e1 f8e8 d1b3 a8b8 b3c2 e7f8 b1d2 d6c5 e4e5 b7b5 a1d1 d8c8 d2e4 d7f5 c2d2 f5g6 d2f4 c8f5 f4d2 b8d8 d5d6 c7d6 e4d6 f8d6 e5d6');
INSERT INTO evaluations VALUES(14093,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(14093,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(14093,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(14093,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(14093,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14093,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(14093,36,'cp 55','f1d3 e7g6 d3c2 f8e7 c3c4 e8g8 b1c3 f6h5 h2h3 h5f4 h1h2 f4h5 g2g3 h5f6 c1e3 c7c6 f3d2 c8d7 a2a4 c6d5 c4d5 d8c8 c2d3 d7h3 d2c4 h7h6 f2f3 h3d7 d1e2 c8c7 c3b5 d7b5 a4b5 a8c8 e1c1 c8b8 c1b1 f8c8');
INSERT INTO evaluations VALUES(17685,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(17685,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(17685,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(17685,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(17685,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(17685,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(17685,36,'cp 87','f2f4 g8f6 d1f3 c8d7 e1c1 d8c7 h2h3 b7b5 e4e5 d6e5 f4e5 f6d5 c3d5 e6d5 d4c6 c7c6 f3d5 c6d5 d1d5 a8c8 a2a3 d7e6 d5d3 f8c5 e3f4 h7h6 f1e2 e6f5 d3c3 e8g8 e2d3 f5e6 h1e1 f8d8 e1d1 c5e7 c3c8 d8c8 b2b4 c8d8 d3e4 d8d1 c1d1 g7g6 f4d2');
INSERT INTO evaluations VALUES(19521,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(19521,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(19521,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(19521,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(19521,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(19521,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(19521,36,'cp -36','b1c3 e7e6 g1f3 f8b4 e1g1 g8e7 c1e3 e8g8 d1b3 d8a5 f1d1 b4c3 b3c3 a5c3 b2c3 b7b6 c4b5 d7d5 e4d5 e6d5 b5c6 e7c6 d1d5 c8e6 d5d2 a8c8 h2h3 h7h6 f3d4 c6d4 e3d4 f8d8 a1b1 c8c4 b1e1 d8d5 f2f4 c4a4');
INSERT INTO evaluations VALUES(20312,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20312,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(20312,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(20312,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(20312,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(20312,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20312,36,'cp 28','d2d4 g8f6 c2c4 d5c4 d4c5 d8d1 f1d1 f8c5 b1d2 c4c3 b2c3 e8g8 d2b3 c5e7 c3c4 e6e5 c1b2 e5e4 f3d4 c6d4 b3d4 c8d7 a1b1 f8d8 d4b5 d7g4 b2f6 e7f6 g1f1 d8d1 b1d1 g4f5 c4c5 f6e7 b5c3 a8c8 g2e4 f5h3 e4g2 h3g2 f1g2 c8c5 d1d7 e7f8 c3e4 c5c2');
INSERT INTO evaluations VALUES(21090,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(21090,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(21090,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(21090,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21090,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(21090,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(21090,36,'cp -26','e7e6 d2d3 g8e7 h2h4 b7b6 b1d2 c8b7 d1e2 h7h6 a2a3 d8d7 d2e4 e7g6 c4b5 a7a6 b5c6 b7c6 h4h5 g6e7 h1h4 e8c8 e4g3 d8e8 h4f4 e7f5 g3e4 f8e7 c1d2 g7g5 h5g6 f7g6 e4f6 e7f6 e5f6 e6e5 f4e4 d7e6 e4e5 e6f6');
INSERT INTO evaluations VALUES(21478,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(21478,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(21478,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(21478,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(21478,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(21478,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(21478,36,'cp -34','h2h4 b7b6 d2d3 g8e7 h4h5 c8b7 b1d2 h7h6 d1e2 d8d7 a2a3 e8c8 d2e4 c8b8 c1d2 e7d5 c4b5 f7f5 e5f6 g7f6 e1c1 e6e5 f3h4 f8a3 b2a3 f6f5 h4g6 h8e8 f2f3 f5e4 d3e4 d5f6 d2h6 d7d6 c1b2 f6h5 h1h5 d6g6 d1h1');
INSERT INTO evaluations VALUES(21679,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(21679,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(21679,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(21679,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(21679,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(21679,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(21679,36,'cp -81','e5e4 f3d2 a7a6 a2a3 h7h5 c3e2 c6e7 c2c4 c7c6 f2f3 e4f3 d2f3 g7g6 e2c3 g8f6 f1d3 c8f5 d1e2 f8g7 e1g1 e8g8 c1d2 f5d3 e2d3 e7f5 c4d5 c6d5 f3e5 f8e8 f1f5 g6f5 d3f5 d8c8 f5f4 c8e6 a1e1 a8c8 h2h3');
INSERT INTO evaluations VALUES(21612,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(21612,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(21612,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(21612,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(21612,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(21612,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(21612,36,'cp -94','f3d2 a7a6 c3e2 c6e7 c2c4 c7c6 e2c3 g8f6 f1e2 e7f5 d1a4 c8e6 c3d5 f6d5 c4d5 e6d5 a4c2 f8b4 e1g1 b4d2 c1d2 e8g8 e2c4 d5c4 c2c4 f8e8 a1c1 f5e7 h2h3 e7d5 c4e2 d8d7 a2a3 h7h6 f2f3 e8e6');
INSERT INTO evaluations VALUES(22985,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(22985,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(22985,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(22985,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(22985,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(22985,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(22985,36,'cp 0','g8f6 d1e2 f8c5 b1c3 e8g8 g2g3 c7c6 c1e3 c5e3 e2e3 d7d6 f1g2 c6d5 c4d5 b7b5 a2a3 c8d7 e1g1 a7a5 f1c1 h7h5 h2h4 f6g4 e3d2 d8b6 f3g5 b6d4 g5f3');
INSERT INTO evaluations VALUES(17062,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17062,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17062,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17062,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17062,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17062,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17062,36,'cp 15','d1e2 c7c6 b1c3 f8b4 c1d2 e8g8 g2g3 c6d5 c4d5 d7d6 f1g2 a7a6 a2a3 b4c5 d2e3 c5e3 e2e3 c8d7 e1g1 f6g4 e3g5 g4f6 f1c1 b7b5 g5e3 f6g4 e3d2 g4f6 d2e2 a8c8 h2h3 h7h6 e2e3 h6h5 h3h4 f6g4 e3d2');
INSERT INTO evaluations VALUES(38711,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38711,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(38711,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(38711,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(38711,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(38711,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(38711,36,'cp 83','d2d4 e5d4 e3d4 c5b6 h2h3 b8c6 g1f3 g8e7 b2b4 e8g8 a2a3 a7a6 d4d5 f8e8 f1e2 c6b8 e1g1 h7h6 c1b2 c8f5 f1e1 b8d7 e2d3 f5d3 d1d3 e7g6 e1e8 d8e8 c3e4 a6a5 a1e1 e8f8 d3b3 a8e8 b3c3 d7e5 f3e5 e8e5 b4a5');
INSERT INTO evaluations VALUES(38719,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(38719,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(38719,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(38719,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(38719,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(38719,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(38719,36,'cp 59','e5d4 e3d4 c5b6 a2a3 g8e7 g1f3 e8g8 f1d3 e7f5 e1g1 f5d4 f3d4 b6d4 d3h7 g8h7 d1d4 h7g8 d4d3 c8e6 f2f4 f7f5 c1e3 d8d7 b2b3 b8a6 a1d1 d7c6 d3c2 a6c5 f1e1 e6f7 c3d5 a8e8 c2f5 f7d5 c4d5 f8f5 d5c6 b7c6');
INSERT INTO evaluations VALUES(38743,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(38743,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(38743,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(38743,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(38743,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(38743,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(38743,36,'cp 69','e3d4 c5b6 c1e3 g8e7 g1f3 b6a5 f1d3 c8f5 e1g1 e8g8 e3g5 f7f6 d3f5 e7f5 g5d2 b8c6 c3d5 a5d2 d1d2 d8d7 a1e1 a8e8 h2h3 f5e7 d5e7 e8e7 e1e7 c6e7 d4d5 c7c5 d5c6 d7c6 f1d1 c6c4 d2d6 e7g6 f3d4 c4a2 d4f5 a2b2');
INSERT INTO evaluations VALUES(38742,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(38742,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(38742,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(38742,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(38742,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(38742,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(38742,36,'cp 66','c5b6 b2b4 c7c6 g1f3 g8e7 h2h3 d6d5 c4c5 b6c7 f1d3 c8f5 d3f5 e7f5 c1g5 d8d7 e1g1 e8g8 f1e1 h7h6 g5d2 f5e7 a2a4 d7d8 b4b5 f8e8 d1e2 c7a5 c3d1 a5d2 e2d2 b8d7 d1c3 d7f6 f3e5 a7a6 b5c6 b7c6 f2f4 f6d7 a4a5 d7e5');
INSERT INTO evaluations VALUES(38786,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(38786,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(38786,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(38786,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(38786,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(38786,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(38786,36,'cp 56','b1d2 b4e7 f1d3 g8f6 e1g1 b7b6 e3e4 d5e4 d2e4 c8b7 e4f6 e7f6 d1e2 b8c6 c2c3 c6e7 f1d1 e8g8 f3g5 e7g6 e2g4 d8e7 d1e1 c7c5 d4c5 e7c5 g5e4 b7e4 g4e4 a8d8 c1e3 c5c7 g2g3 h7h6 a1d1 d8d5 d3c4 d5d1');
INSERT INTO evaluations VALUES(42377,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(42377,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(42377,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(42377,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(42377,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(42377,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(42377,36,'cp 44','f1d3 d6c5 c2c4 a7a5 a2a3 d7d6 h2h3 b8d7 g1f3 f6h5 d3c2 h7h6 b1c3 e8g8 a1b1 c5b6 b2b3 h5f4 e1g1 d8e7 c3e2 f4e2 d1e2 d7c5 c1e3 c5d7 e3d2 d7f6 f3h4 c8d7 h4f5 d7f5 e4f5 b6d4 g2g3');
INSERT INTO evaluations VALUES(44061,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(44061,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(44061,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(44061,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(44061,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(44061,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(44061,36,'cp -12','b5e2 f6e4 f3e5 d7d5 d2d3 e4f6 e1g1 f8d6 d3d4 e8g8 c1f4 d6c7 b1d2 f8e8 f1e1 b8d7 d2f3 d7e5 f4e5 c8f5 c2c3 a7a5 a2a4 f5g6 h2h3 c7d6 e2d3 g6h5 g2g4 h5g6 d3g6 h7g6 d1b3 e8e7 e5d6 d8d6 e1e7 d6e7 a1e1');
INSERT INTO evaluations VALUES(47994,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47994,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(47994,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(47994,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(47994,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(47994,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(47994,36,'cp 48','c7c6 b1d2 e4d2 d1d2 d6d5 c4d3 c6c5 e1g1 b8c6 c2c3 f8e7 h2h3 c8e6 f1e1 d8d7 d2e2 e8g8 c1e3 b7b6 a1d1 a8d8 f3g5 e6f5 d3c4 c6d4 c3d4 d5c4 e2c4 h7h6 g5f3 f5e6 c4c2 d7c8 d4c5 b6c5 b2b3 d8d1 e1d1');
INSERT INTO evaluations VALUES(49367,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(49367,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(49367,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(49367,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(49367,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(49367,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(49367,36,'cp -92','d7d5 e4d5 e6d5 c4b3 a7a5 a2a4 f8d6 g1e2 e8g8 h2h3 b8d7 d3d4 f8e8 e1g1 d7f8 c1f4 f8g6 f4d6 d8d6 b1d2 b7b6 f1e1 c8a6 d2f1 h7h5 f1e3 a6e2 d1e2 c7c6 a1d1 a8d8 c2c3 e8e7 b3c2 g6f4 e2f3 d8e8 d1a1');
INSERT INTO evaluations VALUES(48227,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(48227,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(48227,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(48227,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(48227,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(48227,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(48227,36,'cp -81','e4d5 e6d5 c4b3 a7a5 c2c4 f8b4 b1c3 d5d4 a2a3 b4c5 c3d5 e8g8 g1e2 c8g4 f2f3 g4e6 d5f6 d8f6 e1g1 b8d7 e2g3 f6h4 g3e4 c5e7 g2g3 h4h5 c1f4 c7c6 f1e1 f8e8 b3c2 b7b5 e4d6 e7d6 f4d6 h5g6 d1d2 h7h6 d6f4 e6f5 c4b5 c6b5 e1e8 a8e8 a1e1 e8e1 d2e1 f5d3 e1e8 d7f8');
INSERT INTO evaluations VALUES(48173,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(48173,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(48173,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(48173,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(48173,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(48173,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(48173,36,'cp -74','e6d5 c4b3 a7a5 c2c4 f8b4 b1c3 d5d4 a2a3 b4c5 c3d5 e8g8 g1e2 c8g4 f2f3 g4e6 d5f6 d8f6 e1g1 b8d7 e2g3 f6h4 g3e4 c5e7 f1e1 f8e8 c1d2 c7c6 b3a4 d7c5 e4c5 e7c5 e1e4 h4h5 d1e2 c5d6 f3f4 h5e2 e4e2 e6f5');
INSERT INTO evaluations VALUES(48632,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(48632,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(48632,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(48632,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(48632,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(48632,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(48632,36,'cp -60','c4b3 a7a5 c2c4 f8b4 b1c3 e8g8 g1e2 d5d4 a2a3 b4c5 c3d5 c8g4 h2h3 g4e2 d5f6 d8f6 d1e2 b8d7 e1g1 f8e8 e2d1 c5d6 b3c2 g7g6 a1b1 d6f4 d1f3 f4c1 f3f6 d7f6 f1c1 f6d7 c1e1 g8f8 e1e8 a8e8 g1f1 e8a8 b1e1 d7c5 f1e2 a5a4 e2d2 a8e8 h3h4 e8e1 d2e1 h7h6 g2g3 f8e8 c2b1 e8e7 b1c2');
INSERT INTO evaluations VALUES(50731,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(50731,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(50731,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(50731,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(50731,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(50731,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(50731,36,'cp -20','b7b5 c4b3 f8e7 c3e2 d8c7 c2c3 c5c4 b3c2 c4d3 d1d3 b8d7 e2g3 g8f6 e1g1 e8g8 f1e1 c8b7 d3e2 a8c8 a2a4 b5a4 a1a4 a6a5 h2h3 c8a8 c1d2 d7c5 a4a1 a5a4 f3d4 c7b6 d4b5 f6d7 c3c4 b7c6 d2c3 c6b5 c4b5');
INSERT INTO evaluations VALUES(51269,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(51269,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(51269,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(51269,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(51269,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(51269,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(51269,36,'cp -20','c4b3 f8e7 e1g1 g8f6 c3e2 d8c7 c2c3 c5c4 b3c2 c4d3 d1d3 b8d7 e2g3 e8g8 f1e1 c8b7 d3e2 f8c8 a2a3 d7b6 h2h3 b6c4 f3d4 g7g6 a3a4 e6e5 d4f3 b5b4 c3b4 c4a3 a1a3 c7c2 e2c2 c8c2 b4b5 f6e4 g3e4 b7e4 e1e4 c2c1 e4e1');
INSERT INTO evaluations VALUES(51272,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(51272,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(51272,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(51272,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(51272,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(51272,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(51272,36,'cp -29','g8f6 e1g1 f8e7 f1e1 b8c6 a2a4 b5b4 c3e2 e8g8 e2g3 a6a5 c2c3 c8a6 d3d4 b4c3 b2c3 c5d4 c3d4 d6d5 e4e5 f6d7 b3c2 c6b4 c2b1 a8c8 g3h5 g7g6 h5g3 c8c4 c1h6 f8e8 b1d3 c4d4 f3d4 a6d3 d4b5 d3b5 a4b5 d7c5 d1e2 d8b6 e1d1 e8c8 a1c1');
INSERT INTO evaluations VALUES(52097,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(52097,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(52097,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(52097,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(52097,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(52097,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(52097,36,'cp 85','d8e7 b1c3 g8f6 h5d1 d7d6 g1e2 g7g6 d2d3 c8e6 c1g5 f8g7 c3d5 e6d5 c4d5 e7d7 g5f6 g7f6 e2c3 b8a6 d5c4 a6c7 a2a4 e8g8 d1f3 f6g7 c3d5 a8c8 a4a5 g8h8 h2h4 h7h5 e1g1 c7d5 c4d5 c8c7 g2g3 d7e7 g1g2 g7h6 f1b1 f8d8');
INSERT INTO evaluations VALUES(52578,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(52578,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(52578,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(52578,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(52578,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(52578,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(52578,36,'cp 43','f8g7 f1d3 d5e4 c3e4 g8f6 e4f6 e7f6 e1g1 e8g8 f1e1 f8e8 e1e8 d8e8 h2h3 c8e6 d3f1 b8d7 c2c4 a8d8 c1d2 f6f5 d1a4 d8a8 d2f4 d7f6 a4b4 e8d7 a1d1 a7a5 b4b6 a5a4 f3e5 d7e7 d1e1 f6e4 f1d3 h7h5 a2a3 a8d8 f2f3 e4d6');
INSERT INTO evaluations VALUES(54332,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(54332,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54332,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54332,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(54332,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(54332,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(54332,36,'cp 0','c4f1 d5d4 f1b5 b8c6 c2c3 f8c5 b5c6 b7c6 d1a4 g8e7 c3d4 c5b6 d2d3 e8g8 c1e3 a7a5 e1g1 c8a6 b1c3 a6d3 f1e1 d3f5 f3e5 d8d6 e5c4 d6d8');
INSERT INTO evaluations VALUES(266,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(266,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(266,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(266,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(266,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(266,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(266,36,'cp -79','c5e3 f2e3 f6h4 e1d2 e8d8 d1f3 g8f6 b1c3 d7d6 f1d3 h4g5 f3f4 g5f4 e3f4 a7a6 b5a3 b7b5 c3d1 c6d4 c2c4 c7c6 h1f1 h8e8 d1f2 d8c7 a1e1 f6d7 d2c3 d4e6 g2g3 d7b6 b2b3 e6c5 h2h4 f7f6 e4e5 f6e5 f4e5 d6d5 d3h7 d5c4 b3c4 c8e6 h7c2 b5c4');
INSERT INTO evaluations VALUES(265,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(265,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(265,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(265,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(265,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(265,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(265,36,'cp -93','f2e3 f6h4 e1d2 e8d8 d1f3 g8f6 b1c3 a7a6 b5d4 c6d4 e3d4 d7d5 a1e1 d5e4 f3g3 h4h6 g3e3 h6e3 e1e3 h8e8 f1c4 c8e6 b2b3 b7b5 c4e6 e8e6 h1e1 e6d6 c3e4 d6d4 d2c3 d4e4 e3e4 f6e4 e1e4 d8d7 c3b4 f7f5 e4e5 a8f8 c2c4 b5c4 b4c4 f8f6 c4d4 g7g6 d4e3 f6c6 e5d5 d7c8 d5d2 c8b7 e3f4 c6d6 d2c2 b7b6 f4g5 d6d4 g2g3 d4d7 g5h6 c7c5 h2h4 b6c6');
INSERT INTO evaluations VALUES(264,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(264,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(264,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(264,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(264,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(264,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(264,36,'cp -84','f6h4 e1d2 e8d8 d1f3 a7a6 b5d4 c6d4 e3d4 d7d5 b1c3 g8f6 a1e1 d5e4 f3g3 h4h6 g3e3 h6e3 e1e3 h8e8 f1c4 c8e6 b2b3 b7b5 c4e6 e8e6 h1e1 e6d6 c3e4 d6d4 d2c3 d4e4 e3e4 f6e4 e1e4 d8d7 c3b4 f7f5 e4e5 a8f8 c2c4 b5c4 b4c4 f8f6 c4d4 g7g6 d4e3 f6c6 e5d5 d7c8 d5d2 c6d6');
INSERT INTO evaluations VALUES(307,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(307,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(307,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(307,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(307,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(307,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(307,36,'cp -63','e1d2 e8d8 d1f3 g8f6 b1c3 a7a6 b5d4 c6d4 e3d4 d7d5 a1e1 d5e4 f3g3 h4h6 g3e3 h6e3 e1e3 h8e8 f1c4 c8e6 b2b3 b7b5 c4e6 e8e6 h1e1 e6d6 c3e4 d6d4 d2c3 d4e4 e3e4 f6e4 e1e4 d8d7 c3b4 f7f5 e4e5 a8f8 c2c4 b5c4 b4c4 f8f6 e5a5 g7g6 h2h4 f6e6 h4h5 d7e7 a5c5 c7c6 c5a5 e6e2 a5a6 f5f4 c4d4 e2g2 a6c6 g6g5');
INSERT INTO evaluations VALUES(4824,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(4824,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(4824,32,'cp 220',NULL);
INSERT INTO evaluations VALUES(4824,33,'cp 247',NULL);
INSERT INTO evaluations VALUES(4824,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(4824,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(4824,36,'cp 260','d2d4 e5d4 c3d4 c5a7 d4d5 b7b5 c4d3 c6d4 e4e5 g8e7 d5d6 e7c6 b1c3 c7d6 e5d6 d4e6 c3d5 c8b7 f1e1 e8g8 d3e4 a8b8 c1e3 a7e3 e1e3 f7f5 e4c2 g8h8 h2h4 b8c8 c2b3 h8h7 a1c1 c6d4 c1c8 d4f3 g2f3 b7c8 d5e7 c8b7 b3e6 d7e6 e3e6 d8d7 e6g6 f8e8');
INSERT INTO evaluations VALUES(4831,30,'cp 244',NULL);
INSERT INTO evaluations VALUES(4831,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(4831,32,'cp 259',NULL);
INSERT INTO evaluations VALUES(4831,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(4831,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(4831,35,'cp 254',NULL);
INSERT INTO evaluations VALUES(4831,36,'cp 233','e5d4 c3d4 c5a7 d4d5 b7b5 c4d3 c6d4 e4e5 d7d6 f3d4 a7d4 d3b5 a6b5 d1d4 g8e7 e5d6 d8d6 b1c3 e8g8 c1f4 d6d7 a1d1 f8e8 h2h3 c8b7 d4c5 c7c6 d5d6 e7d5 c3d5 c6d5 f1e1 e8e1 d1e1 a8e8 e1e5 e8e5 f4e5 f7f6 e5g3 d7e8 f2f3 b7c6 b2b3 g8f7 g3f2 e8d7 a2a3 f7e6 f2g3 e6f7 g1f2 f7g6');
INSERT INTO evaluations VALUES(4833,30,'cp 244',NULL);
INSERT INTO evaluations VALUES(4833,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(4833,32,'cp 252',NULL);
INSERT INTO evaluations VALUES(4833,33,'cp 254',NULL);
INSERT INTO evaluations VALUES(4833,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(4833,35,'cp 227',NULL);
INSERT INTO evaluations VALUES(4833,36,'cp 249','c3d4 c5a7 d4d5 b7b5 c4d3 c6d4 e4e5 d7d6 f3d4 a7d4 d3b5 a6b5 d1d4 g8e7 e5d6 d8d6 b1c3 e8g8 c1f4 d6d7 a1d1 c8b7 d4c5 e7g6 f4g3 f7f5 f2f3 f5f4 g3f2 a8d8 f1e1 f8f5 c5b4 d7f7 b4b5 b7d5 b5b4 g6e5 c3d5 d8d5 d1d5 f7d5 a2a4 d5e6');
INSERT INTO evaluations VALUES(4832,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(4832,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(4832,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(4832,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(4832,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(4832,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(4832,36,'cp 233','c5a7 d4d5 b7b5 c4d3 c6d4 e4e5 g8e7 d5d6 e7c6 b1c3 d4e6 c3e4 a7b8 f1e1 c8b7 d3b1 c7d6 e5d6 b8a7 e4g3 e8g8 d1d3 g7g6 c1h6 d8f6 d3d2 c6d4 f3e5 a8d8 b1e4 f6e5 h6f8 g8f8 e4b7 e5d6 d2h6 f8g8 a1d1 d6f4 h6f4 e6f4 b7a6 d7d5 a6b7 g8g7');
INSERT INTO evaluations VALUES(4939,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(4939,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(4939,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(4939,33,'cp 253',NULL);
INSERT INTO evaluations VALUES(4939,34,'cp 258',NULL);
INSERT INTO evaluations VALUES(4939,35,'cp 254',NULL);
INSERT INTO evaluations VALUES(4939,36,'cp 262','d4d5 b7b5 c4d3 c6d4 e4e5 d7d6 f3d4 a7d4 d3b5 a6b5 d1d4 g8e7 e5d6 d8d6 b1c3 e8g8 c1f4 d6d7 a1d1 f8e8 d4c5 c7c6 d5d6 e7g6 f4g3 d7b7 a2a3 g6f8 c5b4 c8d7 h2h3 f8e6 f2f3 c6c5 b4e4 b7e4 f3e4 e6d4 e4e5 g8f8 f1e1 d4c2 e5e6 e8e6 e1e6 d7e6 c3b5');
INSERT INTO evaluations VALUES(5985,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5985,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5985,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(5985,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(5985,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5985,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5985,36,'cp 0','g8f6 d2d3 c6d4 h2h3 e8g8 c1e3 c5b6 e1g1 d4f3 d1f3 c7c6 c3e2 h7h6 a1d1 c8e6 c4e6 f7e6 f3g3 b6e3 g3e3 d8b6 e3b6 a7b6 f2f4 e5f4 e2f4 g8f7 f4h5 f7g6 h5f4');
INSERT INTO evaluations VALUES(12341,30,'cp -212',NULL);
INSERT INTO evaluations VALUES(12341,31,'cp -222',NULL);
INSERT INTO evaluations VALUES(12341,32,'cp -214',NULL);
INSERT INTO evaluations VALUES(12341,33,'cp -198',NULL);
INSERT INTO evaluations VALUES(12341,34,'cp -213',NULL);
INSERT INTO evaluations VALUES(12341,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(12341,36,'cp -199','h7h6 g5f3 d7d5 e4d5 e6d5 c4b3 f8d6 e1g1 e8g8 h2h3 c8e6 d3d4 c5c4 b3a4 c6e7 c2c3 e7g6 b1d2 f6e4 f3e5 g6e5 d2e4 d5e4 d4e5 d6e5 d1d8 a8d8 f1e1 f7f5 a4c2 e6d5 c1e3 b7b6 a1d1 d5c6 f2f4 e4f3 g2f3 d8d1 c2d1 e5g3 e3f2 g3f2 g1f2 f8d8 e1e6 d8d1 e6c6 d1d2 f2e3 d2b2 c6c4 b2a2 e3f4 a7a5 c4c6 a2b2 c6c7');
INSERT INTO evaluations VALUES(12271,30,'cp -209',NULL);
INSERT INTO evaluations VALUES(12271,31,'cp -198',NULL);
INSERT INTO evaluations VALUES(12271,32,'cp -206',NULL);
INSERT INTO evaluations VALUES(12271,33,'cp -215',NULL);
INSERT INTO evaluations VALUES(12271,34,'cp -199',NULL);
INSERT INTO evaluations VALUES(12271,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(12271,36,'cp -210','g5f3 d7d5 e4d5 e6d5 c4b3 f8d6 b1c3 d8e7 d1e2 c8e6 d3d4 c5d4 c3b5 d6b4 c2c3 d4c3 b2c3 b4c5 e1g1 e8g8 b5d4 f6e4 e2d3 a8c8 c1e3 c5d4 c3d4 c6a5 a1c1 e7f6 f3e5 e4d6 e3d2 c8c1 f1c1 a5c4 d2b4 f6f4 c1e1 f8c8 d3d1 d6e4 f2f3 e4d6 b3c4 d6c4 e5c4 c8c4 d1d2');
INSERT INTO evaluations VALUES(14253,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(14253,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14253,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(14253,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(14253,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14253,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(14253,36,'cp 27','g1f3 c8f5 a2a3 e7e6 c2c4 c6b8 b1c3 b8d7 f3h4 f5g4 g2g3 c7c6 f1g2 g4h5 g2f3 h5f3 h4f3 f8e7 d1c2 e8g8 e1g1 d8b6 a1d1 h7h6 f3d2 f8e8 e2e4 d5e4 c3e4 f6e4 c2e4 b6c7 e4c2 a8d8 d2f3 e7f6 f1e1 c6c5 g1g2 c5d4');
INSERT INTO evaluations VALUES(14839,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(14839,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(14839,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(14839,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(14839,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14839,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14839,36,'cp 11','e2e3 f8d6 c3b5 e8g8 c2c4 f8e8 b5d6 c7d6 c4d5 d8a5 c1d2 a5d5 d2c3 b7b5 d1b3 a7a6 b3d5 f6d5 c3d2 e6e5 a1c1 c8b7 d4e5 d6e5 f1e2 e5e4 f3h4 g7g6 e1g1 h7h6 a2a3 c6e5 c1c5 a8c8 f1c1 c8c5 c1c5 e5d3 e2d3 e4d3');
INSERT INTO evaluations VALUES(14834,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14834,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14834,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14834,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14834,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14834,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14834,36,'cp 0','f8d6 c3b5 e8g8 c2c4 f8e8 f1e2 d6f8 b5c3 d5c4 e2c4 a7a6 e3e4 b7b5 c4b3 c6a5 b3c2 c7c5 e4e5 c5d4 e5f6 d4c3 c2h7 g8h7 f3g5 h7g6 d1g4 d8d5 g5e4 g6h7 e4g5');
INSERT INTO evaluations VALUES(15849,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(15849,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(15849,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(15849,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(15849,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(15849,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(15849,36,'cp -71','e1g1 c7c6 g5f3 f8d6 f1e1 e8g8 f3e5 f8e8 d2d4 d8h4 b1d2 f7f6 d2f3 h4h5 e5d3 c8f5 c1d2 f5e4 d3e5 d6e5 d4e5 e4f3 g2f3 f6e5 c4g4 h5f7 c2c4 d5f6 g4f5 f7c4 d2c3 c4f4 e1e5 f4f5 e5f5 f6d5 a2a4 e8e2 f5g5 d5c3 b2c3 a8d8 g5a5 a7a6 a1b1 d8d2 b1b7 e2f2 a5a6 f2g2 g1f1 d2f2 f1e1 g2h2 a6c6');
INSERT INTO evaluations VALUES(17663,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17663,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17663,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17663,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17663,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17663,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17663,36,'cp 0','d3d4 c5d4 f3d4 c8d7 c1e3 g8f6 e1g1 a8c8 c4d3 c6e5 h2h3 e5c4 d3c4 c8c4 d1d3 c4c8 a2a4 f8e7 a4a5 d8c7 d4b3 e8g8 f1d1 d7c6 b3d4 c6d7');
INSERT INTO evaluations VALUES(18665,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(18665,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(18665,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(18665,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(18665,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(18665,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(18665,36,'cp 99','d7d6 b2b4 c8g4 d1b3 d8d7 c1e3 g8f6 a2a4 g4f3 g2f3 f6h7 g1h1 h7g5 f1g1 f8e7 g1g3 a8c8 c3d5 e8f8 b3d3 g5e6 b4b5 c6d4 e3d4 e6d4 f3f4 e5f4 g3g7 f8g7 d3d4 g7h7 e4e5 h8g8 c4d3 g8g6 d3g6 f7g6 d5e7 d7e7 e5d6 e7g7 a1d1 g7d4 d1d4 c8d8 a4a5 h7g7 d6d7 g7f7 d4f4 f7e6');
INSERT INTO evaluations VALUES(23469,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(23469,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(23469,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(23469,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(23469,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(23469,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(23469,36,'cp 15','c2c4 e7e5 b1c3 g8f6 g1f3 d7d5 c4d5 f6d5 g2g3 f8c5 f1g2 e8g8 e1g1 f8e8 c3d5 d8d5 a2a3 a7a5 c1d2 d5d8 d2c3 c6d4 e2e3 d4f3 g2f3 c5b6 f3e4 c7c6 f1e1 g7g6 h2h4 d8d6 d1b3 d6c7 b3a4 c8e6 a1d1 a8d8 a4c2 c7d6');
INSERT INTO evaluations VALUES(23426,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23426,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23426,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23426,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23426,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23426,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(23426,36,'cp 10','e7e5 b1c3 g8f6 g1f3 f8e7 e2e4 e8g8 f1e2 e7c5 e1g1 d7d6 a2a3 a7a5 h2h3 c8d7 c1e3 c6d4 e3d4 c5d4 f3d4 e5d4 c3b5 d7b5 c4b5 c7c6 b5c6 b7c6 b2b4 a5b4 a3b4 a8a1 d1a1 d8b6 f1b1 f8b8 a1a2 b6c7 a2b2 c6c5 b4b5 f6d7');
INSERT INTO evaluations VALUES(26122,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26122,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(26122,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(26122,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(26122,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(26122,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26122,36,'cp 32','f1e2 f7f6 e1g1 f6e5 d4e5 f8c5 b1d2 g8e7 f3d4 e8g8 b2b4 c5d4 c3d4 a7a5 b4b5 c6c5 c1a3 c5d4 d2f3 d4d3 e2d3 f5d3 d1d3 f8f4 g2g3 f4c4 f1c1 a8c8 f3g5 d7f8 c1c4 c8c4 a1c1 c4c1 a3c1 d8c7 c1a3 h7h6 g5f3 f8g6 a3b2 e7f5 g1g2 f5e7 h2h3 a5a4');
INSERT INTO evaluations VALUES(26121,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26121,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(26121,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(26121,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(26121,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(26121,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(26121,36,'cp 27','f7f6 e1g1 f6e5 d4e5 f8c5 b1d2 g8e7 f3d4 e8g8 b2b4 c5d4 c3d4 a7a5 b4b5 c6c5 c1a3 c5d4 d2f3 d4d3 e2d3 f5d3 d1d3 f8f4 g2g3 f4c4 f3g5 d7f8 a1c1 a8c8 c1c4 c8c4 f1c1 c4c1 a3c1 h7h6 g5f3 e7f5 c1a3 d8c7 a3f8 g8f8 h2h4');
INSERT INTO evaluations VALUES(26806,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26806,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26806,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(26806,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26806,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(26806,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26806,36,'cp 9','g8f6 g1f3 f8d6 f4g3 e8g8 f1e2 f8e8 e1g1 d6g3 h2g3 e6e5 d4e5 c6e5 d1b3 e5f3 e2f3 b7b6 c3c4 d5c4 d2c4 f6e4 a1c1 d8f6 g3g4 f5g6 f1d1 e4c5 b3a3 g6e4 c4d2 e4f3 d2f3 a7a5 d1d4 h7h6 a3c3 c5e6 d4c4 f6e7 f3d4 a8d8 d4e6');
INSERT INTO evaluations VALUES(31288,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31288,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(31288,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(31288,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31288,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31288,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(31288,36,'cp -14','e7e6 f1b5 b8d7 g1e2 a8c8 e1g1 a7a6 b5d3 g8e7 d3f5 e7f5 b1d2 f8e7 d2f3 e8g8 d1d3 c5d4 c3d4 d8c7 g2g4 f5h4 f3h4 e7h4 c1e3 h4e7 f1c1 c7b6 c1c2 c8c6 a1c1 f8c8 h2h3 d7f8 e2f4 f8g6 f4g6 h7g6 d3e2 c6c2 c1c2 c8c2 e2c2');
INSERT INTO evaluations VALUES(31636,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31636,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(31636,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(31636,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(31636,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(31636,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31636,36,'cp 0','d8b6 b1d2 g8f6 d4c5 f8c5 b2b4 c5d6 d2c4 d5c4 d1d6 b6d6 f4d6 f6e4 d6e5 f7f6 e5b8 a8b8 f3d4 e8e7 f1c4 h8c8 c4b3 c8c3 e1e2 e4d6 h1c1 b8c8 c1c3 c8c3 e2d2 d6e4 d2e2 e4d6');
INSERT INTO evaluations VALUES(37164,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(37164,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(37164,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(37164,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(37164,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(37164,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(37164,36,'cp -5','c7c5 d4c5 b8a6 a2a3 b4c3 c2c3 a6c5 g5f6 d8f6 c3f6 g7f6 a1d1 b7b6 e2e3 f8d8 g1e2 c8b7 e2c3 c5b3 c3b5 b7c6 b5d6 f6f5 f1e2 g8f8 f2f3 f8e7 d1d3 b3c5 d3d4 a7a5 e1f2 h7h5 b2b4 c5b7 h1d1 a5b4 a3b4 a8a2 d6b7 c6b7 d4d2 a2d2 d1d2 d7d5');
INSERT INTO evaluations VALUES(39632,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39632,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(39632,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39632,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(39632,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39632,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39632,36,'cp 0','e2e4 c7c5 f1e2 d7d6 e1g1 g8f6 d2d3 e8g8 a2a4 b8c6 d1e1 c8d7 c1d2 d8b6 d2c1');
INSERT INTO evaluations VALUES(42435,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(42435,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(42435,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(42435,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(42435,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(42435,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(42435,36,'cp 145','d2d4 c5e7 e4e5 f6d5 c3e4 d7d6 c2c3 d6e5 d4e5 c8d7 f1d3 c7c5 a2a3 b7b5 e1g1 d7c6 d1e2 a7a6 f1e1 b8d7 b2b3 c5c4 b3c4 b5c4 d3c4 e8g8 c4d3 d8c7 c3c4 d5b6 c1f4 a8d8 e2e3 c6e4 e3e4 g7g6 d3f1 f8e8 f4h6 e7f8 e4f4 f8h6 f4h6');
INSERT INTO evaluations VALUES(42434,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(42434,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(42434,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(42434,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(42434,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(42434,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(42434,36,'cp 122','c5b4 f1d3 d7d5 e4e5 f6e4 d3e4 d5e4 f3g5 b8c6 c1e3 b4c3 b2c3 b7b6 d1h5 d8e7 g5h7 e7f8 h5g4 f8a3 e1g1 c8b7 h7g5 a3c3 g5e4 c3c4 c2c3 e8c8 f2f3 f7f5 e5f6 g7f6 e4f6 c4c3 g4e6 c8b8 e6b3 c3b3 a2b3 c6d4 a1d1 d4b3 d1d8 h8d8 g1f2 a7a5 g2g4 b3d4 e3d4 d8d4 g4g5');
INSERT INTO evaluations VALUES(42458,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(42458,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(42458,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(42458,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42458,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42458,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42458,36,'cp -12','b7b6 e2e3 e8g8 f1d3 d7d5 a2a3 b4c3 d2c3 f6e4 c3d2 c7c5 d4c5 e4c5 e1g1 c5d3 c2d3 f8e8 a1c1 e6e5 h2h3 f7f6 b2b4 c8e6 c1c3 d8d7 d1c2 b8a6 f1c1 e8c8 c3c8 a8c8 c2c8 d7c8 c1c8 e6c8 f3e1 g8f7 f2f4 e5f4 e3f4');
INSERT INTO evaluations VALUES(47592,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(47592,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(47592,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(47592,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47592,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(47592,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(47592,36,'cp 31','f8g7 f1e2 e8g8 e1g1 d8e7 a1b1 a7a5 b2b3 f6e4 c3a4 c8d7 c1b2 f8c8 b1c1 d7a4 b3a4 b8a6 f3d2 e4d2 d1d2 b7b6 c4d5 e6d5 e2b5 c7c5 b5a6 a8a6 b2a3 e7d7 d4c5 b6c5 c1c5 c8c5 a3c5 a6c6 c5d4 g7d4 d2d4');
INSERT INTO evaluations VALUES(55181,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(55181,31,'cp -185',NULL);
INSERT INTO evaluations VALUES(55181,32,'cp -177',NULL);
INSERT INTO evaluations VALUES(55181,33,'cp -183',NULL);
INSERT INTO evaluations VALUES(55181,34,'cp -179',NULL);
INSERT INTO evaluations VALUES(55181,35,'cp -183',NULL);
INSERT INTO evaluations VALUES(55181,36,'cp -163','f1c4 g7g6 h5b5 c7c6 b5b3 e7e6 a2a4 g8f6 b3e3 b8d7 b1c3 d7b6 e3e2 f8g7 c4b3 e8g8 a4a5 b6d7 c3e4 f6e4 e2e4 d7c5 e4e3 c5b3 e3b3 d8g5 b3f3 g5c5 c2c3 e6e5 g1e2 c8e6 d2d3 c5b5 e1g1 f8d8 d3d4 e6c4 f1e1 c4e2 f3e2 b5e2 e1e2 e5d4');
INSERT INTO evaluations VALUES(55180,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(55180,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(55180,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(55180,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(55180,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(55180,35,'cp -155',NULL);
INSERT INTO evaluations VALUES(55180,36,'cp -154','g7g6 h5b5 c7c6 b5b3 e7e6 b3e3 b7b5 c4e2 f7f5 b1c3 f8g7 d2d3 e4d3 e3d3 d8d3 c2d3 a7a5 h2h4 a5a4 a2a3 b8d7 c1e3 g8f6 d3d4 e6e5 d4e5 d7e5 e3d4 e5f7 e2f3 a8a6 g1e2 c8e6 e1c1 e6b3 e2f4 e8g8 d1e1 f6d7 d4g7 g8g7');
INSERT INTO evaluations VALUES(54879,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(54879,31,'cp -172',NULL);
INSERT INTO evaluations VALUES(54879,32,'cp -159',NULL);
INSERT INTO evaluations VALUES(54879,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(54879,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(54879,35,'cp -160',NULL);
INSERT INTO evaluations VALUES(54879,36,'cp -162','h5b5 c7c6 b5b3 e7e6 b3e3 b7b5 c4e2 f7f5 b1c3 f8g7 d2d3 e4d3 e3d3 d8d3 c2d3 g8e7 f2f4 b8a6 c1d2 e8g8 g1f3 f8d8 a1c1 c8b7 e1g1 a8b8 d3d4 b5b4 c3a4 c6c5 e2c4 b7f3 c4e6 g8f8 f1f3 e7d5 g1f1 g7d4 e6d5 d8d5 d2e3 b8e8 e3d4');
INSERT INTO evaluations VALUES(5047,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(5047,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(5047,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5047,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5047,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5047,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5047,36,'cp 0','g8f6 e1g1 d7d5 c1f4 e8g8 f1e1 d5e4 d3e4 d8b6 a1b1 f6e4 e1e4 f8e8 f4e3 b6c7 d1f3 c8e6 e4a4 a7a5 e3f4 c7a7 f4e3 a7c7');
INSERT INTO evaluations VALUES(7691,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7691,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7691,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7691,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7691,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7691,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7691,36,'cp 0','e1g1 d7d5 c1f4 e8g8 f1e1 d8b6 a1b1 c8e6 h2h3 d5e4 d3e4 f6e4 e1e4 f8d8 d1e1 c6c5 f4e5 g7h6 e5f4');
INSERT INTO evaluations VALUES(7690,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7690,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7690,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7690,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7690,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7690,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7690,36,'cp 0','d7d5 c1f4 e8g8 f1e1 d5e4 d3e4 f6e4 e1e4 d8d1 a1d1 g7c3 b2c3 c8e6 f2f3 f8e8 a2a4 a7a5 f4g5 g8f8 g5h6 f8g8');
INSERT INTO evaluations VALUES(6385,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(6385,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(6385,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(6385,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(6385,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(6385,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(6385,36,'cp 53','b2c3 g8f6 e1g1 e8g8 f1e1 d7d6 d2d4 c8g4 h2h3 g4f3 d1f3 a7a6 b5c6 b7c6 d4e5 d6e5 c1g5 d8e7 e1b1 f8d8 b1b7 d8d6 a1b1 a8d8 b7a7 c6c5 b1b7 d6c6 g5f6 e7f6 b7c7 f6f3 g2f3 c6c7 a7c7 d8d1 g1g2 f7f6 c7c5 d1d2 c3c4 d2c2 c5c8 g8f7');
INSERT INTO evaluations VALUES(6383,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(6383,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(6383,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(6383,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(6383,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(6383,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(6383,36,'cp 61','g8f6 e1g1 e8g8 f1e1 d7d6 d2d4 c8g4 h2h3 g4f3 d1f3 a7a6 b5c6 b7c6 d4e5 d6e5 c1g5 d8e7 a2a4 f8b8 a1b1 b8b6 b1b3 a8b8 e1b1 b6b3 b1b3 b8b3 c2b3 a6a5 h3h4 e7d6 g5f6 g7f6 g1h2 h7h6 c3c4 g8g7 f3g4 g7f8 g4f5 c6c5 h4h5');
INSERT INTO evaluations VALUES(10569,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(10569,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(10569,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(10569,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(10569,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(10569,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(10569,36,'cp 31','c5e7 b5c6 d7c6 d1e2 c8f5 g2g4 f5g6 f3e5 e4f6 f1e1 h7h6 e5g6 f7g6 b1d2 f6d5 e2f3 d8d6 d2e4 d6e6 c3c4 d5b4 f3e2 e8c8 a2a3 b4a6 c1d2 d8d4 d2c3 d4d7 c3g7 h8h7 g7c3 a6c5 e4c5 e6e2 e1e2 e7c5 e2e8 d7d8 e8d8 c8d8 g1g2 h7f7 f2f3 c5d6');
INSERT INTO evaluations VALUES(11168,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(11168,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(11168,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(11168,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(11168,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11168,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(11168,36,'cp 6','c6d4 d1d4 e7e6 a2a4 f8e7 c1e3 e8g8 h2h3 c8d7 f2f4 d8c7 c4b3 a8d8 a1d1 d7c6 f4f5 d6d5 f5e6 f7e6 e4d5 f6d5 f1f8 g8f8 c3d5 c6d5 d1f1 f8g8 b3d5 d8d5 d4f4 c7f4 f1f4 e7g5 f4e4 g5e3 e4e3 g8f7 e3c3 b7b5 a4b5 a6b5 g1f2 d5d2 f2g3 e6e5 c3c5 f7f6 c5b5 d2c2');
INSERT INTO evaluations VALUES(13106,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(13106,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(13106,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(13106,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(13106,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(13106,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(13106,36,'cp -18','d4c5 e7e6 b2b4 a7a5 d3b5 c8d7 a2a4 a5b4 b5c6 d7c6 c3b4 b7b6 c1b2 b6c5 b4b5 c6b7 g1f3 f8d6 b1d2 e8g8 e1g1 f6d7 d1c2 d6c7 f1d1 f8e8 d2b3 c5c4 b3d2 c7a5 b2c3 d8c7 c3a5 a8a5 d2b3 a5a8 a4a5 c7d8 a5a6 c4b3 c2b3 b7c8');
INSERT INTO evaluations VALUES(22744,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(22744,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22744,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22744,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(22744,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(22744,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(22744,36,'cp 0','g8f6 g1f3 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 f8c5 f1c4 h7h6 e1g1 d8f6 d2d3 c5b6 c4d5 e8g8 a3a4 f8d8 h2h3 c6e7 d5b3 e7c6');
INSERT INTO evaluations VALUES(32059,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(32059,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(32059,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(32059,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(32059,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(32059,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(32059,36,'cp 46','c2c4 d5c4 f1c4 f8d6 f4g3 b8d7 d1b3 d7b6 b1c3 g8f6 c4b5 c7c6 b5e2 e8g8 e1g1 d6g3 h2g3 a7a5 f1d1 a5a4 b3a3 f8e8 a1c1 d8e7 a3e7 e8e7 a2a3 f6e4 c3e4 f5e4 f3d2 f7f5 d2e4 f5e4 f2f3 e4f3 e2f3 a8d8');
INSERT INTO evaluations VALUES(33148,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(33148,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(33148,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(33148,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(33148,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(33148,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(33148,36,'cp 76','b8c6 c1e3 c5b6 a2a4 g8e7 d1e1 d7d6 a4a5 b6c5 b1c3 e8g8 a5a6 b7b6 c3d5 e7d5 c4d5 c8d7 c2c3 a8e8 b2b4 c5e3 f2e3 c6e7 d5b7 f6e6 c3c4 e8b8 b4b5 c7c6 a1b1 c6b5 c4b5 d7c8 b7c6 e7c6 b5c6 g8h8 e1h4 e6g6');
INSERT INTO evaluations VALUES(224,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(224,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(224,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(224,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(224,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(224,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(224,36,'cp 89','c1e3 c5b6 a2a4 g8e7 c2c3 d7d6 b2b4 c8e6 c4e6 f7e6 b1d2 e8g8 a4a5 b6e3 f2e3 a7a6 d3d4 e5d4 e3d4 e7g6 g2g3 f6e7 d1c2 e7d7 f1f2 f8f7 a1f1 a8f8 g1g2 e6e5 d4d5 c6e7 d2c4 c7c6 d5c6 b7c6 f3d2 g6h4 g2g1 f7f2 f1f2 f8f2 g1f2 d7h3 g3h4 h3h2 f2f1');
INSERT INTO evaluations VALUES(34764,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(34764,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(34764,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(34764,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(34764,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(34764,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(34764,36,'cp 104','f6d8 c1e3 g8f6 f1d3 g7g6 c2c4 f8g7 b1c3 e8g8 e1g1 f6g4 e3d2 f7f5 b2b4 g4f6 d1c2 f5e4 c3e4 f6e4 d3e4 b8d7 a1e1 d7f6 f3g5 f6e4 g5e4 c8f5 f2f3 d8d7 c2b3 b7b6 h2h3 f5e4 e1e4 a7a5 b4a5 b6a5 f1b1 d7f5');
INSERT INTO evaluations VALUES(37044,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37044,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(37044,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(37044,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(37044,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37044,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37044,36,'cp 0','c7c5 d4c5 b4c5 a2a3 d5c4 f1c4 b8d7 e1g1 a7a6 c3e4 f6e4 c2e4 d7f6 e4c2 c5d6 e3e4 b7b5 c4d3 e6e5 h2h3 c8d7 f1d1 d8e7 c1e3 a8c8 c2e2 h7h6 a1c1 d7e6 b2b4 c8c1 d1c1 e7b7 c1d1 f8d8 e2c2 d8c8 c2b2 c8d8 b2c2');
INSERT INTO evaluations VALUES(36153,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36153,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36153,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(36153,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36153,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36153,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36153,36,'cp 0','d4c5 b4c5 a2a3 d5c4 f1c4 a7a6 c3e4 f6e4 c2e4 b8d7 e1g1 d7f6 e4c2 c5d6 e3e4 b7b5 c4d3 e6e5 a3a4 b5a4 a1a4 a6a5 c1e3 c8d7 a4c4 d8e7 f1c1 d7e6 c4c6 e6d7 c6c4');
INSERT INTO evaluations VALUES(37808,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(37808,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(37808,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(37808,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(37808,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(37808,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(37808,36,'cp 15','g8f6 b1c3 c5d4 f3d4 e8g8 e4e5 f6e8 h2h4 d7d6 e5d6 d8d6 d1d2 e8f6 e1c1 b8c6 f1b5 c6d4 d2d4 d6d4 e3d4 c8f5 b5d3 e7e6 d3f5 g6f5 a2a4 f8d8 f2f3 b7b6 c1b1 f6h5 c3e2 g7d4 e2d4 h5f4 g2g4 f5g4 f3g4 h7h5');
INSERT INTO evaluations VALUES(40715,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40715,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(40715,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40715,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(40715,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40715,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40715,36,'cp 13','g1f3 d5c4 e2e3 c7c5 f1c4 c5d4 f3d4 c8d7 e1g1 e8g8 a1c1 b8c6 d4f3 f6h5 h4e7 d8e7 c3e4 f8d8 c4e2 h5f6 e4f6 e7f6 d1b3 d7e8 f1d1 a8c8 c1c3 d8d1 b3d1 f6d8 d1c1 c6e7 f3e5 c8c3 c1c3 a7a6 a2a3');
INSERT INTO evaluations VALUES(41869,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41869,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(41869,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41869,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41869,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41869,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41869,36,'cp 12','c2c3 a7a5 g1f3 e8g8 e1g1 b8c6 f1e1 h7h6 c4b5 c5a7 c1e3 a7e3 e1e3 c6e7 d3d4 c7c6 b5d3 e7g6 a2a4 c8d7 d3f1 f8e8 b1d2 b7b5 d1c2 d8b6 b2b3 d7e6 e3e1 e6d7 a1b1 d7e6 a4b5 c6b5 c3c4 e5d4 c4b5 g6f4');
INSERT INTO evaluations VALUES(45599,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(45599,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(45599,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(45599,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(45599,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(45599,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(45599,36,'cp -39','b7b6 c5b6 a7b6 e2e3 f8b4 c1d2 e8g8 f1d3 c8b7 e1g1 b8d7 f1e1 h7h6 h2h3 b4d6 e3e4 f6e4 c3e4 d5e4 d3e4 d7f6 e4d3 c6c5 d4c5 d6c5 d1e2 a8a7 d2c3 a7a4 d3c2 a4f4 a1d1 b7f3 d1d8 f3e2 d8f8 g8f8 e1e2 c5d4 c3d4 f4d4');
INSERT INTO evaluations VALUES(43960,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(43960,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(43960,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(43960,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(43960,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(43960,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(43960,36,'cp -37','c5b6 a7b6 e2e3 c8b7 f1d3 f8b4 c1d2 e8g8 e1g1 b8d7 f1e1 b4d6 e3e4 d5e4 c3e4 f6e4 d3e4 d8c7 d1e2 d7f6 e4d3 f8d8 h2h3 c6c5 d4c5 d6c5 d2e3 f6d5 e3c5 b6c5 d3e4 h7h6 a2a4 c5c4 a4a5 a8a5 a1a5 c7a5 e2c4');
INSERT INTO evaluations VALUES(48102,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(48102,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(48102,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(48102,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(48102,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(48102,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(48102,36,'cp 39','b8d7 c4d5 e6d5 b1c3 d7b6 c1g5 f8e7 g5e7 d8e7 e1g1 c7c6 c3e2 e8g8 e2g3 a7a5 d1c2 b6c4 a1e1 a5a4 h2h3 f7f5 e5f6 e7f6 f3e5 c4e5 e1e5 c8d7 c2c3 f6h4 g3e2 d7f5 c3b4 b7b5 b4c5 h4f6 f2f3 f5d3 f1e1 f8f7');
INSERT INTO evaluations VALUES(49470,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(49470,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(49470,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(49470,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(49470,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(49470,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(49470,36,'cp 37','g1f3 d7d5 d2d4 b7b5 e1g1 f8e7 a2a4 b5b4 f3e5 e8g8 c2c4 b4c3 b1c3 a7a5 e2e4 c7c6 f1e1 f6d7 e5d3 b8a6 e4d5 e6d5 c3a2 f8e8 c1e3 d8b6 e3g5 d7f6 g5f6 g7f6 a1c1 c8f5 d1f3 f5g6 d3f4 a6b4 a2b4 e7b4');
INSERT INTO evaluations VALUES(53634,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(53634,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(53634,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(53634,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(53634,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(53634,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(53634,36,'cp 59','c6c5 c2c3 b8c6 g1f3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f8e7 a2a3 c8d7 b1c3 f5e3 f2e3 e8g8 c3e2 f7f6 e5f6 e7f6 e2f4 g7g6 a1c1 a8c8 d3b1 f6g7 d1d3 d8e7 b2b4 g7h6 h2h4 e7g7 b4b5 c6e7 c1c8 f8c8');
INSERT INTO evaluations VALUES(54294,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(54294,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(54294,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(54294,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(54294,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(54294,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(54294,36,'cp 62','c2c3 b8c6 g1f3 g8e7 h2h4 c5d4 c3d4 e7f5 d3f5 e6f5 b1c3 f8b4 d1d3 c8e6 e1g1 b4c3 d3c3 a8c8 c1g5 d8a5 c3e3 h7h6 g5f4 c6e7 h4h5 a5a4 b2b3 a4d7 f1e1 a7a5 a1c1 e8g8 f3h4 c8c1');
INSERT INTO evaluations VALUES(54293,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(54293,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(54293,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(54293,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(54293,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(54293,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(54293,36,'cp 69','b8c6 g1f3 g8e7 h2h4 c5d4 c3d4 e7f5 d3f5 e6f5 b1c3 f8b4 h4h5 h7h6 e1f1 b4c3 b2c3 c8e6 f1g1 a8c8 f3h4 c6e7 d1f3 e8g8 f3g3 g8h8 g1h2 f8g8 a2a4 b7b6 h1g1 c8c4 c1d2 h8h7 g3f4 d8e8 a4a5 b6b5 f4g3');
INSERT INTO evaluations VALUES(55578,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(55578,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(55578,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(55578,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(55578,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(55578,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(55578,36,'cp 36','e2e4 e5d4 c3d4 d6d5 e4d5 g8f6 b1c3 f6d5 f1c4 d5c3 b2c3 f8d6 g1f3 e8g8 e1g1 h7h6 f3d2 c8f5 d1f3 d8h4 g2g3 h4g4 f3g4 f5g4 f2f3 g4d7 f1e1 b8c6 c4d3 c6a5 d2b3 a5b3 a2b3 d7e6 b3b4 c7c6 c1e3 d6e7 g1f2 f8d8');
INSERT INTO evaluations VALUES(282,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(282,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(282,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(282,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(282,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(282,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(282,36,'cp 24','d7d6 c2c3 a7a6 b2b4 c5a7 a2a4 f6g6 b1d2 g8f6 c4b3 e8g8 f3h4 g6g4 d2f3 h7h6 h2h3 g4h5 c1e3 g7g5 e3a7 c6a7 h4f5 c8f5 e4f5 a7c6 f1e1 g5g4 h3g4 f6g4 d3d4 h5f5 b3c2 f5h5 b4b5 a6b5 a4b5 a8a1 d1a1 c6e7');
INSERT INTO evaluations VALUES(2081,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(2081,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(2081,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(2081,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(2081,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(2081,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(2081,36,'cp 152','b1c3 c7c6 c1f4 e5g6 f4g3 g8f6 d1e2 f6h5 f3d4 e7f6 e1c1 h5g3 h2g3 f8c5 d4f5 d7d6 c3a4 c8f5 e4f5 g6e7 a4c5 d6c5 e2e3 e8g8 e3c5 a8d8 f1d3 b7b6 c5e3 e7d5 e3d2 h7h6 h1e1 f8e8 a2a3 e8e1 d1e1 g8f8 f2f3 d5c7 d2f4 c7d5');
INSERT INTO evaluations VALUES(6494,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(6494,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6494,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6494,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6494,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(6494,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6494,36,'cp -7','g1f3 d7d5 b1d2 g8f6 b2b4 c5d6 c1b2 d5e4 d3e4 a7a5 a2a3 e8g8 d1c2 d8e7 d2c4 a5b4 c3b4 c8g4 f1e2 g4f3 e2f3 c6d4 b2d4 e5d4 e1g1 f6d7 g2g3 d6e5 c2d3 e5f6 a1e1 d7e5 c4e5 f6e5 f3g2 e7e6');
INSERT INTO evaluations VALUES(9379,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(9379,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(9379,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(9379,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(9379,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(9379,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9379,36,'cp -16','h2h3 e8g8 f1e2 c8d7 a2a3 c6d4 e1g1 d4f3 e2f3 c7c6 c1e3 c5e3 f2e3 a7a6 a3a4 b7b5 a4b5 a6b5 a1a8 d8a8 c4b5 c6b5 d1d2 a8a7 d3d4 b5b4 c3d5 f6d5 e4d5 e5d4 e3d4 b4b3 f1c1 h7h6 c1c3 a7b6 d2f4 f8e8 g1h2 e8c8 c3c8 d7c8');
INSERT INTO evaluations VALUES(15769,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(15769,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(15769,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(15769,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(15769,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(15769,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(15769,36,'cp 38','c7c6 e1g1 f8e7 a2a4 c8e6 c3e2 d6d5 e4e5 c6c5 d4c3 c5c4 e5f6 e7f6 e2d4 c4d3 c3d3 e8g8 c2c3 a7a6 c1e3 d8d7 f1d1 f8e8 a4a5 e6g4 d1d2 a8d8 h2h3 g4h5 d3f5 d7f5 d4f5 h5g6 g2g4 h7h5 a1d1 h5g4 h3g4');
INSERT INTO evaluations VALUES(17907,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(17907,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(17907,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17907,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17907,36,'cp 0','g8f6 c3d5 h7h6 c2c3 f8c5 f3e5 c5f2 e1f2 c6e5 d3d4 f6e4 f2g1 c7c6 d4e5 c6d5 c1e3 d8a5 d1b3 e4d2 b3d1 d2e4');
INSERT INTO evaluations VALUES(11533,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11533,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11533,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11533,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11533,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11533,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11533,36,'cp 0','c3d5 h7h6 c2c3 f8c5 f3e5 c5f2 e1f2 c6e5 d3d4 f6e4 f2g1 c7c6 d4e5 c6d5 c1e3 d8a5 d1b3 e4d2 b3c2 d2c4 e3d4 d7d6 e5d6 e8g8 f1c4 d5c4 c2f2 c8e6 f2g3 f7f6 h2h3 a5d5 d4e3 f8d8 e3h6 d5c5 g1f1 c5f5 f1g1');
INSERT INTO evaluations VALUES(18806,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(18806,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(18806,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(18806,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(18806,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(18806,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(18806,36,'cp -59','d7d5 e4e5 d8b6 f1d3 c8d7 e1g1 c5d4 a2a3 a7a5 g1h1 g8h6 b1d2 f8e7 d1e2 h6f5 a3a4 c6b4 d3f5 e6f5 d2b3 d4d3 c2d3 b4c6 e2d1 e8g8 d3d4 b6a6 c1e3 b7b6 e3d2 b6b5 d2c3 b5a4 b3c5 e7c5 d4c5 a6b5 d1d5 f8d8 d5d2 d7e6 d2f2 d8d3 f1d1');
INSERT INTO evaluations VALUES(25892,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(25892,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(25892,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(25892,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(25892,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(25892,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(25892,36,'cp 209','d4d5 c6b8 b1c3 e7e6 a2a3 g8f6 c1f4 a7a6 g2g4 f8c5 d1d2 b6b5 g4g5 f6h5 f4e3 c5e3 f2e3 g7g6 f1h3 h5g7 e4e5 b7d5 c3d5 e6d5 d2d5 b8c6 e1c1 g7e6 h1f1 d8e7 f3d4 e8g8 d4e6 f7e6 d5d7 e7d7 d1d7');
INSERT INTO evaluations VALUES(25885,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(25885,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(25885,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(25885,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(25885,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(25885,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(25885,36,'cp 182','c6b8 b1c3 e7e6 a2a3 g8f6 c1f4 a7a6 d1d3 b6b5 e1c1 f8c5 f4g5 c5e7 g5f6 e7f6 e4e5 f6e7 d3e3 e8g8 h2h4 b5b4 a3b4 e7b4 f1c4 d8e7 f3g5 b4c3 e3c3 e6d5 c3d3 g7g6 c4d5 b7d5 d3d5 b8c6 d5d7 e7c5 d7c7 a8c8 g5e4 c5e5 c7e5');
INSERT INTO evaluations VALUES(26825,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26825,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(26825,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(26825,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(26825,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(26825,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26825,36,'cp 25','e1g1 c6b4 f3e1 f8e7 b1d2 g8f6 c2c3 b4c6 e1d3 h7h6 f1e1 e8g8 e2e4 f6e4 d2e4 f5e4 g2e4 d5e4 e1e4 d8d5 d1e2 a8d8 h2h4 e7d6 c1f4 c6e7 f4d6 c7d6 d3f4 d5c6 a1e1 b7b5 g3g4 c6b7 e4e3 a7a5 e3f3 g8h8 a2a3 b5b4');
INSERT INTO evaluations VALUES(26824,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(26824,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(26824,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26824,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(26824,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(26824,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(26824,36,'cp 26','c6b4 f3e1 f8e7 c2c3 b4c6 e1d3 g8f6 f1e1 e8g8 b1d2 h7h6 e2e4 f6e4 d2e4 f5e4 g2e4 d5e4 e1e4 d8d5 d1e2 e7d6 d3f4 d5f5 f4g2 a8d8 c1f4 d8d7 a2a3 a7a6 f4d6 c7d6 e4f4 f5b5 e2b5 a6b5 g2e3 f8d8 h2h4 d8c8');
INSERT INTO evaluations VALUES(32415,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(32415,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(32415,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(32415,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(32415,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(32415,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(32415,36,'cp 30','f1b5 b8d7 e1g1 g8f6 b5c4 c7c6 d2d3 b7b5 c4b3 a7a5 a2a4 f8d6 b3a2 d7c5 a4b5 c6b5 b1c3 b5b4 c3d5 b4b3 d5f6 d8f6 c2b3 e8g8 h2h3 g4f3 f1f3 f6d8 c1e3 c5e6 b3b4 a5b4 a2e6 f7e6 f3f1 a8a1 d1a1 d8e7 a1a6 d6c5 f1f8 e7f8 a6e6 g8h8 e3c5 f8c5 g1h2');
INSERT INTO evaluations VALUES(33292,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(33292,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(33292,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(33292,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(33292,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(33292,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(33292,36,'cp 176','b1c3 d7d6 c1e3 g8e7 d1d2 b8c6 d4b5 c5e3 d2e3 e8d8 e1c1 h4h6 e3h6 g7h6 f2f4 c8d7 f1c4 d7g4 d1d3 a7a6 b5d4 c6d4 d3d4 g4e6 c4e2 h8g8 g2g3 f7f6 c3d1 d8d7 d1e3 g8g7 b2b3 a8g8 d4d1 d7c8 e2d3 g8f8 h1e1 f6f5 e3f5 e7f5 e4f5 e6f5');
INSERT INTO evaluations VALUES(33291,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(33291,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(33291,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(33291,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(33291,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(33291,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(33291,36,'cp 182','g8e7 c1e3 d7d6 d1d2 b8c6 d4b5 c5e3 d2e3 e8d8 e1c1 h4h6 e3h6 g7h6 f2f4 c8d7 f1e2 f7f5 g2g4 a7a6 b5d6 c7d6 g4f5 d8c7 f5f6 e7g6 c3d5 c7b8 h2h4 h8f8 h4h5 g6f4 d5f4 f8f6 h1f1 b8a7 f4d5 f6f1 d1f1 a8e8 d5f6 e8d8 f6d7 d8d7 f1f6');
INSERT INTO evaluations VALUES(34400,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(34400,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(34400,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(34400,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(34400,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(34400,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34400,36,'cp 81','f1c4 g8f6 f3g5 c8g4 f2f3 g4h5 h2h4 h5g6 c1e3 b8d7 d1e2 b7b5 c4b3 d7c5 e1c1 c5b3 a2b3 e7b7 g5h3 f8b4 h3f2 a7a5 f2d3 e8g8 g2g4 h7h5 g4h5 f6h5 d3e5 a5a4 c3a2 c6c5 a2b4 c5b4 e3c5 h5g3 e2e3 a4b3 c5f8 g3h1 f8b4 a8a1 c1d2');
INSERT INTO evaluations VALUES(34399,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(34399,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(34399,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(34399,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(34399,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(34399,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(34399,36,'cp 50','g8f6 f3g5 c8g4 f2f3 g4h5 d1e2 b7b5 c4b3 b8d7 c1e3 d7c5 e1c1 c5b3 a2b3 e7b7 h2h4 h5g6 g5h3 f8b4 h3f2 a7a5 f2d3 e8g8 g2g4 h7h5 e2f2 b7e7 c1b1 b4d6 e3g5 a5a4 b3b4 a4a3 g4h5 g6h5 b2b3 a3a2 c3a2 e7a7 f2a7 a8a7 g5f6 g7f6');
INSERT INTO evaluations VALUES(36459,30,'cp -142',NULL);
INSERT INTO evaluations VALUES(36459,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(36459,32,'cp -143',NULL);
INSERT INTO evaluations VALUES(36459,33,'cp -106',NULL);
INSERT INTO evaluations VALUES(36459,34,'cp -130',NULL);
INSERT INTO evaluations VALUES(36459,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(36459,36,'cp -148','d6e5 f3e5 c7c5 d4d5 e7e6 e5f3 b8a6 h2h3 e6d5 c4d5 a6b4 f1c4 f8e8 c1e3 c8f5 a1c1 a7a6 e1g1 e8e3 f2e3 b7b5 g2g4 f5g4 h3g4 b5c4 f3g5 b4d3 c1c2 d8e7 d1f3 d3e5 f3h3 h7h6 g5e4 f6g4 d5d6 e7e6 c2d2 e5d3 d2g2 f7f5 g2g4 f5g4 h3h4 a8e8 e4f6 g7f6 h4f6 e6f6');
INSERT INTO evaluations VALUES(36795,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(36795,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(36795,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(36795,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(36795,34,'cp -126',NULL);
INSERT INTO evaluations VALUES(36795,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(36795,36,'cp -123','f3e5 c7c5 d4d5 e7e6 e5f3 b8a6 h2h3 f8e8 d5d6 e6e5 f1e2 e8e6 c3b5 e5e4 f3g5 e6e8 h3h4 h7h6 g5h3 a6b4 h3f4 b4d3 f4d3 e4d3 d1d3 c8f5 d3d1 e8e6 d1b3 a8c8 c1e3 a7a6 b5c3 f6g4 e2g4 f5g4 e1g1 e6d6 b3b7 d8h4 e3f4');
INSERT INTO evaluations VALUES(40155,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(40155,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(40155,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(40155,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(40155,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(40155,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(40155,36,'cp 75','a7a6 a2a4 c8g4 f1e2 g4f3 g2f3 b8d7 e1g1 e8g8 g1h1 f8e8 d1c2 g7g6 f1g1 e7f8 c1g5 f8g7 f3f4 d8b6 f2f3 c5c4 f4f5 b6b4 a4a5 d7c5 g5f4 a8d8 a1c1 b4a5 e2c4 b7b5 c4f1 b5b4 c3e2 d8c8 g1g2 b4b3 c2d1 f6d7');
INSERT INTO evaluations VALUES(41103,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(41103,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(41103,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(41103,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(41103,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(41103,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41103,36,'cp 1','e1g1 e8g8 f1e1 c8e6 b1d2 f8e8 e2d3 e7d6 c2c3 b8d7 d2f1 h7h6 h2h3 c7c5 d3c2 d8b6 e1e2 b6c7 d4c5 d7c5 a2a4 e8e7 c1e3 b7b6 e3d4 c5e4 d4f6 e4f6 d1d3 a8e8 a1e1 f6e4 f3d4 e6d7');
INSERT INTO evaluations VALUES(41102,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41102,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41102,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41102,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(41102,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(41102,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41102,36,'cp 9','e8g8 f1e1 c8e6 b1d2 b8d7 e2d3 c7c5 d4c5 d7c5 d2b3 f8e8 d3b5 e6d7 b5d7 d8d7 c1f4 c5e4 c2c3 h7h6 h2h3 e7d6 f4d6 e4d6 d1c2 d7c7 f3d4 a7a6 b3d2 d6b5 d2f3 b5d6 e1e8 a8e8 a1d1 d6c4 d4f5 c7f4 f5g3 g7g6');
INSERT INTO evaluations VALUES(36416,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(36416,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(36416,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(36416,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(36416,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36416,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36416,36,'cp 8','f1e1 c8e6 b1d2 b8d7 e2d3 c7c5 d4c5 d7c5 d2b3 f8e8 d3b5 e6d7 b5d7 d8d7 c1f4 c5e4 c2c3 e7d6 f4d6 e4d6 d1c2 h7h6 b3d2 a7a6 h2h3 e8e1 a1e1 a8e8 e1d1 d7c6 f3d4 c6b6 d2f1 d6c4 b2b3 c4e5 f1e3 e5c6 c2d2');
INSERT INTO evaluations VALUES(42571,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(42571,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(42571,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42571,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(42571,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(42571,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(42571,36,'cp -14','d7d6 g1f3 a7a5 e1g1 e8g8 h2h3 b8c6 d2d3 c8e6 c4e6 f7e6 c1e3 b7b6 d3d4 e5d4 f3d4 d8d7 d4c6 d7c6 d1d3 f6d7 f2f4 c5e3 d3e3 b6b5 a1d1 b5b4 c3e2 b4a3 e2d4 c6c5 e3a3 c5a3 b2a3 a8e8 d4c6 e6e5 f4e5');
INSERT INTO evaluations VALUES(45866,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(45866,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(45866,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(45866,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(45866,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(45866,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(45866,36,'cp 97','d2d4 c8g4 c1e3 b8d7 f1e2 f8e7 d4d5 e8g8 h2h3 g4h5 e1g1 h5g6 f3d2 a8c8 a2a4 c6d5 c4d5 d7c5 e2f3 a7a5 d1e2 f6d7 f3g4 c5e4 d2e4 f7f5 g4f3 f5e4 f3g4 h7h6 a1c1 c8a8 c3b5 d7c5 e3c5 d6c5 f1d1 d8b6');
INSERT INTO evaluations VALUES(45873,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(45873,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(45873,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(45873,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(45873,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(45873,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(45873,36,'cp 75','c8g4 f1e2 b8d7 d4d5 c6d5 c4d5 f8e7 h2h3 g4f3 e2f3 e8g8 c1e3 a7a6 e1g1 f6e8 d1d2 h7h6 f3g4 d7b6 b2b3 e7g5 a1c1 g5e3 d2e3 e8f6 c3e2 h6h5 g4f5 b6d7 c1c3 a8c8 c3c8 d8c8 e2g3 g7g6 e3h6 c8d8 f5g6 f7g6');
INSERT INTO evaluations VALUES(50026,30,'cp 434',NULL);
INSERT INTO evaluations VALUES(50026,31,'cp 437',NULL);
INSERT INTO evaluations VALUES(50026,32,'cp 440',NULL);
INSERT INTO evaluations VALUES(50026,33,'cp 443',NULL);
INSERT INTO evaluations VALUES(50026,34,'cp 443',NULL);
INSERT INTO evaluations VALUES(50026,35,'cp 455',NULL);
INSERT INTO evaluations VALUES(50026,36,'cp 450','e2h5 g7g6 e5g6 h7g6 h5h8 b8c6 b1c3 c6b4 f2f3 d8f6 h8f6 e4f6 e1d1 d7d5 d2d4 e8f7 c1g5 c8e6 c3e2 f8d6 e2f4 e6d7 h2h4 b4c6 c2c3 a8h8 g5f6 f7f6 f4d5 f6g7 d1c2 d7e6 d5e3 c6e7 f1c4 e6d7 a1e1 b7b5 c4b3 a7a5 e3d5 e7d5 b3d5 c7c6 d5b3');
INSERT INTO evaluations VALUES(54562,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(54562,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(54562,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(54562,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(54562,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(54562,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(54562,36,'cp -40','b8c6 b1c3 g8f6 d2d3 c8e6 g1e2 f8e7 e1g1 e8g8 f1e1 d8d7 c1g5 a8e8 e2f4 e7d6 c3e2 d6e5 c2c3 h7h6 g5f6 e5f6 f4e6 e8e6 e2f4 e6e1 d1e1 f8e8 e1f1 e8e5 f1d1 f6g5 f4h5 d7f5 h5g3 f5f4 g3e2 f4f6');
INSERT INTO evaluations VALUES(55065,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(55065,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(55065,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(55065,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(55065,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(55065,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55065,36,'cp -30','c7c5 e2e3 e7e6 g1f3 b8c6 f1e2 a7a6 e1g1 g8f6 d4c5 f8c5 c2c4 d5c4 e2c4 e8g8 d2c3 b7b5 d1d8 f8d8 c4e2 c8b7 f1c1 c5e7 b1d2 b5b4 c3d4 f6d7 b2b3 c6d4 f3d4 a8c8 e2f3 b7f3 d4f3 d7c5 c1c4 c5d3');
INSERT INTO evaluations VALUES(55447,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(55447,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(55447,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(55447,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(55447,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(55447,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(55447,36,'cp 107','c7c5 d4d5 g7g6 d1a4 c8d7 a4b3 b8a6 b1c3 g8f6 a2a3 a6c7 f1e2 d7c8 e1g1 f8g7 f1d1 e6d5 e4e5 d6e5 c4d5 f6d7 d5d6 c7e6 e2c4 e8g8 c4e6 f7e6 b3e6 g8h7 c1e3 b7b6 b2b4 c8b7 b4c5 b7f3 g2f3');
INSERT INTO evaluations VALUES(55841,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(55841,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(55841,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(55841,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(55841,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(55841,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(55841,36,'cp -75','c7c5 e2e4 c5d4 d1d4 b8c6 f1b5 a7a6 b5c6 b7c6 c3a4 g8f6 g1e2 f8e7 d4b6 d5e4 f3e4 c8d7 c1f4 f6e4 e1c1 d8b6 a4b6 a8a7 h1e1 c6c5 e2g3 e4g3 f4g3 h7h6 g3f4 h6h5 f4d6 a7b7 d6e7 e8e7 b6d5 e7e8 d5f4 h5h4 e1e5 b7b5 d1d2 c5c4 f4d5');
INSERT INTO evaluations VALUES(1940,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(1940,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(1940,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(1940,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(1940,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(1940,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(1940,36,'cp 145','f3d4 e5d4 d2d3 c7c6 b5c4 f8c5 b1d2 g8e7 d1h5 d7d6 f2f4 h7h6 a2a3 c5b6 b2b4 g7g6 h5f3 d6d5 e4d5 c6d5 c4d5 e7d5 f3d5 f6e6 d5e4 e8g8 a3a4 e6e4 d2e4 f8d8 a4a5 b6c7 c1d2 c8f5 e4g3 f5d7 f4f5 c7g3 h2g3 g6f5 a5a6 b7b6 d2h6 a8c8');
INSERT INTO evaluations VALUES(1937,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(1937,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(1937,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(1937,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(1937,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(1937,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(1937,36,'cp 165','e5d4 d2d3 c7c6 b5c4 f8c5 b1d2 g8e7 d1h5 d7d6 f2f4 h7h6 a2a4 c5b4 a4a5 c8d7 a1a4 b4d2 c1d2 d6d5 c4a2 e8c8 a4a3 a7a6 b2b4 h8e8 a2b3 d5e4 d3e4 e7g6 e4e5 f6e7 b3c4 c8b8 c4d3 b8a8 d3g6 f7g6');
INSERT INTO evaluations VALUES(1944,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(1944,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(1944,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(1944,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(1944,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(1944,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(1944,36,'cp 144','d2d3 c7c6 b5c4 f8c5 b1d2 g8e7 d1h5 d7d6 f2f4 h7h6 a2a4 c8d7 a4a5 c5b4 c4b3 d6d5 e4d5 c6d5 b3d5 e7d5 h5d5 d7c6 f1e1 e8f8 d5c4 b4d2 c1d2 a7a6 e1e5 a8d8 c4c5 f8g8 c5e7 g8h7 e7f6 g7f6 e5c5 d8e8 g1f2 c6g2');
INSERT INTO evaluations VALUES(1942,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(1942,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(1942,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(1942,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(1942,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(1942,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(1942,36,'cp 145','c7c6 b5c4 f8c5 b1d2 g8e7 d1h5 d7d6 f2f4 h7h6 a2a4 c8d7 d2b3 c5b6 c1d2 b6c7 a1e1 e8g8 b3c1 d6d5 c4a2 a8e8 f4f5 c6c5 g2g4 g7g5 f5g6 f7g6 f1f6 g6h5 f6f8 g8f8 e4d5 f8g7 b2b4 c5b4 d2b4 e7c8 e1e8 d7e8 c1e2 h5g4 e2d4 e8a4');
INSERT INTO evaluations VALUES(5301,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(5301,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(5301,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(5301,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(5301,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(5301,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(5301,36,'cp -17','e7e6 d2d4 c6d4 f3d4 c5d4 c3b5 d7d6 b5d4 g8f6 d1e2 e8g8 e1g1 b7b6 f1d1 c8b7 e4e5 f6d5 c4d5 b7d5 c1f4 d6e5 f4e5 a8c8 c2c3 g7e5 e2e5 c8c5 e5e3 d8f6 d1d2 d5b7 a1e1 h7h5 h2h3 e6e5 d4c2 f8d8 e1d1 c5d5 d2d5 b7d5');
INSERT INTO evaluations VALUES(5111,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(5111,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(5111,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(5111,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(5111,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(5111,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(5111,36,'cp -12','d2d4 c5d4 c3b5 d7d5 e4d5 e6d5 c4e2 g8e7 b5d4 c6d4 f3d4 e8g8 e1g1 e7f5 d4f3 d5d4 e2d3 c8e6 f1e1 d8b6 a3a4 a8e8 a4a5 b6c6 c1f4 e6d5 e1e8 f8e8 d3f5 g6f5 f3d4 g7d4 d1d4 d5g2 d4d6 c6d6 f4d6 g2h3 d6c5 e8e4 f2f3 e4e2 a1d1 e2c2 d1d8 g8g7 c5a7 c2b2');
INSERT INTO evaluations VALUES(6615,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(6615,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(6615,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(6615,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(6615,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(6615,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(6615,36,'cp 84','d1h5 g7g6 h5f3 g8f6 c3d5 f6d5 c4d5 e8g8 c2c3 b4e7 d2d3 e7g5 c1g5 d8g5 h2h4 g5h6 h4h5 d7d6 f3f6 c6d8 g1f3 d8e6 d5c4 c8d7 h1h2 h6f4 f6e7 e6c5 h5g6 h7g6 e1e2 a8e8 e7g5 d7e6 g5f4 e5f4 c4e6 c5e6 a1h1 g8g7 h2h7 g7f6 g2g3 f8g8 g3f4 e6f4 e2e3 f4g2 e3d2 g8g7 f3d4 g7h7 h1h7');
INSERT INTO evaluations VALUES(6726,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(6726,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(6726,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(6726,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(6726,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(6726,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(6726,36,'cp -87','d8e7 d2d4 g8f6 c1a3 d7d6 d4d5 c6d4 b3a4 e7d7 a4a5 d4c2 e1e2 a7a6 h1c1 b7b6 c4b5 d7b5 a5b5 a6b5 c1c2 f6e4 a3b4 c8b7 b1d2 e4f6 c3c4 b5c4 d2c4 b7a6 a1d1 a6c4 c2c4 a8a2 e2f1 e8d7 c4c1 h8a8 b4c3 f6g4 c3e1 a2b2 h2h3 g4f6 f3g5 b2b5 g5f7 b5d5 f1g1 a8a2 f2f3 c7c5 e1c3 d5d1 c1d1');
INSERT INTO evaluations VALUES(397,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(397,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(397,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(397,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(397,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(397,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(397,36,'cp -81','d2d4 g8f6 c1a3 d7d6 d4d5 c6d4 b3a4 e7d7 a4a5 d4c2 e1e2 a7a6 h1c1 b7b6 c4b5 d7b5 a5b5 a6b5 c1c2 f6e4 a3b4 c8b7 c3c4 b5c4 b1d2 e4f6 d2c4 b7a6 a1d1 a6b5 a2a3 b5a4 c2d2 a4d1 d2d1 g7g6 c4e3 f6h5 f3g5 f7f5 g2g4 h5f4 e2f1 f5g4 e3g4 h7h5 g4f6 e8e7 f6e4 h8f8 h2h4 e7d7 e4c3 f8f5');
INSERT INTO evaluations VALUES(9302,30,'cp 514',NULL);
INSERT INTO evaluations VALUES(9302,31,'cp 498',NULL);
INSERT INTO evaluations VALUES(9302,32,'cp 530',NULL);
INSERT INTO evaluations VALUES(9302,33,'cp 513',NULL);
INSERT INTO evaluations VALUES(9302,34,'cp 525',NULL);
INSERT INTO evaluations VALUES(9302,35,'cp 524',NULL);
INSERT INTO evaluations VALUES(9302,36,'cp 521','c5b6 d4d5 a7a6 b5a4 f6e4 d5c6 e8g8 b1c3 e4c3 b2c3 d8f6 f3d4 b7c6 a4c6 a8b8 c1e3 h7h6 f1e1 f8d8 d1d2 b6c5 h2h3 c5a3 c6f3 c8d7 e1b1 c7c5 d4e2 d6d5 e2f4 d5d4 c3d4 a3b4 f4d5 f6d6 d5b4 c5b4');
INSERT INTO evaluations VALUES(8907,30,'cp 531',NULL);
INSERT INTO evaluations VALUES(8907,31,'cp 521',NULL);
INSERT INTO evaluations VALUES(8907,32,'cp 537',NULL);
INSERT INTO evaluations VALUES(8907,33,'cp 525',NULL);
INSERT INTO evaluations VALUES(8907,34,'cp 524',NULL);
INSERT INTO evaluations VALUES(8907,35,'cp 533',NULL);
INSERT INTO evaluations VALUES(8907,36,'cp 540','d4d5 c8d7 d5c6 b7c6 b5d3 e8g8 b1c3 d7g4 c1g5 h7h6 g5h4 g4h5 h2h3 g7g5 h4g3 f8e8 a1c1 d8d7 f1e1 a8b8 d3b1 g5g4 f3h4 g4h3 d1d2 h5g6 h4g6 d7g4 d2h6 f7g6 h6h3 g4h3 g2h3 f6h5 c3e2 h5g3 e2g3 b6f2 g1f2 b8b2 f2g1 c6c5 e4e5');
INSERT INTO evaluations VALUES(8906,30,'cp 523',NULL);
INSERT INTO evaluations VALUES(8906,31,'cp 551',NULL);
INSERT INTO evaluations VALUES(8906,32,'cp 524',NULL);
INSERT INTO evaluations VALUES(8906,33,'cp 540',NULL);
INSERT INTO evaluations VALUES(8906,34,'cp 538',NULL);
INSERT INTO evaluations VALUES(8906,35,'cp 529',NULL);
INSERT INTO evaluations VALUES(8906,36,'cp 525','c8d7 d5c6 b7c6 b5d3 e8g8 b1c3 f8e8 c1g5 a8b8 g5h4 h7h6 a1c1 d8e7 d3c2 d7g4 h2h3 g4f3 d1f3 b6d4 c1b1 e7e6 f1e1 e6e5 e1d1 b8b4 h4g3 e5c5 g1h1 e8b8 b1c1 b4b2 c3a4 b2c2 a4c5 c2c1 d1c1 d4c5 h1g1 f6d7 e4e5 d6e5 f3e4 b8b2');
INSERT INTO evaluations VALUES(11625,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(11625,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(11625,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(11625,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(11625,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(11625,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(11625,36,'cp 117','b1c3 e4c3 d2c3 c8e6 a2a4 b5b4 a4a5 d8d6 c3b4 f7f6 c2c3 c6d8 c1f4 c7c6 f3d4 e6f7 f4g3 d6c7 f2f4 e5e4 d1d2 f8e7 f4f5 e7d6 c3c4 d6g3 h2g3 e8g8 d2c3 c7d6 c4d5 c6d5 b3a4 h7h5 b4b5 a6b5 a4b5 f7e8 b5f1 e8c6');
INSERT INTO evaluations VALUES(11983,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(11983,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(11983,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(11983,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(11983,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(11983,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(11983,36,'cp -62','d8a5 g5d2 a5b6 e1g1 f8e7 c4b3 f6e4 d1h5 d7d5 h5e5 c8e6 d2e3 e7f6 e5f4 b6b8 f4b8 a8b8 c2c3 a7a5 b1d2 e4d2 e3d2 b8b7 f1d1 e8d7 d2e3 h8a8 c3c4 a5a4 c4d5 a4b3 d5e6 d7e6 a2b3 a8a1 d1a1 b7b3 a1c1 e6d7 c1d1 d7c8 d1d6 c8c7 e3f4 c7b6 f4e3 b6b5 g2g3 h7h5 d6d7 b3b2 d7f7 b5c4');
INSERT INTO evaluations VALUES(12682,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12682,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12682,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(12682,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12682,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12682,36,'cp 0','a7a6 c4b3 b7b5 c1e3 e7g6 h2h4 h7h5 a2a4 b5b4 c3e2 c8b7 e2f4 g6f4 e3f4 d8f6 d1c1 a8c8 e1g1 c6a5 b3a2 f6g6 f3e5 g6f6');
INSERT INTO evaluations VALUES(17725,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(17725,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(17725,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(17725,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(17725,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(17725,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(17725,36,'cp 85','c2c3 c8f5 b1d2 e7e6 h2h4 h7h6 g1e2 d8d7 e2g3 f5h7 f1e2 f7f6 f2f4 g8e7 e1g1 e7f5 g3f5 h7f5 e2h5 g7g6 h5e2 d7e7 d2f3 f5e4 b2b4 f6f5 a2a4 e8d7 c1d2 e4f3 e2f3 e7h4');
INSERT INTO evaluations VALUES(20008,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(20008,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20008,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20008,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(20008,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(20008,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20008,36,'cp 0','g8f6 f1d3 e7e5 d3c2 d6d5 e4d5 d8d5 d1e2 f8e7 d2d3 c8f5 b1d2 a8d8 d2c4 f5d3 e2d3 d5d3 c2d3 d8d3 f3e5 c6e5 c4e5 d3d6 e1g1 f6d5 e5c4 d6d7 a2a4 f7f6 g2g3 e8f7 g1g2 h8d8 f1e1 d5c7 c1e3 g7g6 c4a3 c7d5 a3c4');
INSERT INTO evaluations VALUES(13314,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(13314,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13314,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13314,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13314,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13314,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13314,36,'cp 0','f1d3 e7e5 d3c2 d6d5 e4d5 d8d5 d1e2 f8e7 d2d3 c8f5 b1d2 a8d8 d2c4 f5d3 c2d3 d5d3 e2d3 d8d3 f3e5 c6e5 c4e5 d3d6 e1g1 f6d5 f1d1 e8g8 e5c4 d6d7 g2g3 f8d8 c4e5 d7d6 e5c4');
INSERT INTO evaluations VALUES(13318,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13318,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13318,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13318,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13318,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13318,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13318,36,'cp 0','e7e5 d3c2 d6d5 e4d5 d8d5 d1e2 f8e7 d2d3 c8f5 b1d2 a8d8 d2c4 f5d3 e2d3 d5d3 c2d3 d8d3 f3e5 c6e5 c4e5 d3d6 e1g1 f6d5 f1d1 e8g8 e5c4 d6d7 g2g3 f7f6 a2a4 f8d8 d1e1 g8f7 g1g2 b7b6 c1e3 h7h5 g2f3 e7d6 a1d1 d6c7 e1e2 d5e7 d1d7 d8d7');
INSERT INTO evaluations VALUES(23277,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(23277,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(23277,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(23277,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(23277,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(23277,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(23277,36,'cp 29','f8b4 f1c4 d8e7 e1g1 b4c3 b2c3 d7d6 d1d3 g8f6 c1g5 c8e6 c4d5 e8g8 g5f6 e7f6 f1b1 a8b8 d5c6 b7c6 b1b8 f8b8 d3a6 c6c5 a6a7 b8f8 a7a6 f6f7 a2a4 e6c4 a6c6 g7g5 g1h1 g5g4 f3h4 c4e6 h4f5 e6f5 e4f5 f7f5');
INSERT INTO evaluations VALUES(25299,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(25299,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(25299,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(25299,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(25299,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(25299,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(25299,36,'cp 109','g2g4 h5g6 a2a3 f6d7 c1e3 d7b6 c4b3 h7h5 h1g1 h5g4 h3g4 d8d7 d3d4 f8e7 a3a4 e5d4 f3d4 c6a5 d1e2 e8c8 e1c1 e7f6 f2f4 f6d4 e3d4 f7f6 f4f5 g6f7 b3f7 d7f7 g4g5 c7c5 d4f2 f7c4 e2c4 a5c4 b2b3 c4e5 g5f6 g7f6 g1g7');
INSERT INTO evaluations VALUES(25296,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(25296,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(25296,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(25296,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(25296,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(25296,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(25296,36,'cp 119','h5g6 a2a3 f6d7 h3h4 h7h5 g4g5 f8e7 c1e3 d7c5 d1d2 e8g8 f3g1 c5e6 e1c1 c6d4 f2f4 b7b5 c4a2 b5b4 e3d4 e6d4 a3b4 e5f4 g1e2 d4e2 c3e2 f4f3 e2f4 g8h7 h1f1 c7c6 f1f3 d8b6 d1f1 a8b8 c2c3 a7a5 d2f2 b6f2 f3f2 a5b4 f4g6 f7g6 f2f8 e7f8 c3b4 d6d5 e4d5 c6d5 a2d5 f8b4 c1c2 b4d6');
INSERT INTO evaluations VALUES(25362,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(25362,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(25362,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(25362,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(25362,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(25362,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(25362,36,'cp 134','a2a3 c6d4 f3d4 e5d4 c3e2 d6d5 e4d5 f6d5 e2d4 d8e7 d1e2 e8c8 e2e7 f8e7 d4f3 h7h5 g4g5 f7f6 e1g1 g6f5 g1g2 c7c6 c1d2 e7d6 a1e1 h8e8 f1h1 f5d7 c4d5 c6d5 d2e3 b7b6 f3d4 e8f8 g5g6 c8c7 h3h4 d8e8 b2b4 e8c8 f2f4');
INSERT INTO evaluations VALUES(35835,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35835,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35835,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35835,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35835,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(35835,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(35835,36,'cp 19','f6h5 d4c5 h5f4 e3f4 e7c5 e1g1 b8c6 g2g3 g7g6 b2b4 c5e7 d1e2 d5d4 f3d4 c6b4 c3b4 d8d4 d2c4 e7b4 f1d1 b4c3 a1c1 d4f6 h2h4 c3d4 c4e5 d4e5 f4e5 f6e7 h4h5 b7b6 d3e4 c8b7 e4b7 e7b7 d1d6 a8c8');
INSERT INTO evaluations VALUES(37385,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(37385,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(37385,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(37385,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(37385,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(37385,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37385,36,'cp 17','c7c5 c1f4 d7d6 d1d2 b8c6 f4h6 g7h6 d2h6 c8d7 g1f3 c6d4 e1g1 d7c6 a1d1 a7a6 f3d4 c5d4 c3e4 c6e4 d3e4 e7e5 e2e3 d4e3 h6e3 a8c8 b2b3 b7b5 d1d3 b5c4 b3c4 c8c4 f1d1 g8g7 d3d6 d8a5');
INSERT INTO evaluations VALUES(39917,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(39917,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(39917,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(39917,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(39917,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(39917,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(39917,36,'cp 8','b7b6 f1d3 d7d5 e1g1 e8g8 d4c5 d5c4 d3c4 d8d1 f1d1 b4c5 c4e2 c8b7 a2a3 c5e7 f3e5 b8c6 e5c6 b7c6 c1d2 a7a6 a1c1 f8c8 f2f3 c6e8 d2e1 b6b5 e2d3 f6d7 c3e4 d7e5 d3f1 e5c4 f1c4 c8c4 c1c4 b5c4 d1c1');
INSERT INTO evaluations VALUES(42567,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(42567,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(42567,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(42567,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(42567,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(42567,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(42567,36,'cp -64','c7c6 c4b3 d7d5 b1c3 e8g8 d2d3 c5b4 c1d2 d5d4 c3b1 b4d6 f3g3 b8d7 d2g5 g8h8 b1d2 b7b5 g1e2 h7h6 g3h4 h8g8 g5f6 d7f6 e1g1 a7a5 a2a4 d6b4 d2f3 d8e7 g2g4 f6d7 h4h5 g8h7 g4g5 b4d6');
INSERT INTO evaluations VALUES(43375,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(43375,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(43375,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(43375,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(43375,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(43375,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(43375,36,'cp -25','b7b5 c4b3 e7e6 c3e2 d8c7 c2c3 c5c4 b3c2 c4d3 d1d3 b8d7 e2g3 f8e7 e1g1 e8g8 f1e1 d7e5 d3e2 c8d7 h2h3 f8c8 e1d1 b5b4 f3e5 d6e5 c1d2 d7b5 c2d3 b5d3 e2d3 c8d8 d3e2 a6a5 c3b4 a5b4 d2e3 d8d1 e2d1 c7c6');
INSERT INTO evaluations VALUES(43738,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(43738,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(43738,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(43738,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(43738,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(43738,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(43738,36,'cp -18','c4b3 e7e6 e1g1 f8e7 c3e2 d8c7 c2c3 c5c4 b3c2 c4d3 d1d3 c8b7 e2g3 b8d7 d3e2 e8g8 f1e1 f8c8 a2a3 e7f8 c1d2 g7g6 a1d1 c8e8 h2h3 a8c8 f3h2 f8g7 d2c1 h7h5 h2f3 d7b6 c2d3 e6e5 c1g5 f6h7 g5e3');
INSERT INTO evaluations VALUES(44465,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(44465,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44465,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44465,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(44465,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(44465,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(44465,36,'cp 30','d7d5 d2d4 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c8d7 f3e5 d7c6 c4e2 f8c8 e5c6 b8c6 h4f6 g7f6 d1d7 a8b8 c3e4 c5e7 f1d1 c8d8 g2g3 d8d7 d1d7 f6f5 e4c3 g8f8');
INSERT INTO evaluations VALUES(44889,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(44889,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(44889,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(44889,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44889,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(44889,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(44889,36,'cp 30','d2d4 e7e6 c1g5 h7h6 g5h4 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 f3e5 b4c3 b2c3 f6d7 h4g3 d7e5 g3e5 b7b6 c4b5 c8b7 d1d6 f8c8 c3c4 b8c6 e5c3 c6e7 f1d1 b7c6 c3b2 c6b5 c4b5');
INSERT INTO evaluations VALUES(45602,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(45602,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(45602,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(45602,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(45602,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(45602,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(45602,36,'cp -36','b7b6 c5b6 a7b6 f1d3 b8d7 d1c2 c8b7 e1g1 f8d6 b2b3 e8g8 c1b2 d8e7 f1c1 f8c8 b1d2 c6c5 c2d1 e6e5 f3e5 d7e5 d4e5 d6e5 b2e5 e7e5 a2a4 g7g6 d3f1 c8c6 d1c2 h7h6 c1d1 e5e6 h2h3 e6e7 c2b2 h6h5 d1c1 f6e8 c1b1');
INSERT INTO evaluations VALUES(43961,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(43961,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(43961,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(43961,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(43961,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(43961,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(43961,36,'cp -24','c5b6 a7b6 f1d3 b8d7 d1c2 c8b7 b2b3 f8d6 e1g1 d8e7 c1b2 e8g8 a2a4 c6c5 f1c1 f8c8 c2d1 e6e5 d4e5 d7e5 b2e5 d6e5 f3e5 e7e5 b1d2 e5e7 h2h3 f6d7 d3b5 h7h6 d2f3 d7f6 d1c2 f6e4 c2b2 c8c7 b2e5 e4d6 b5d3 g7g6 e5e7 c7e7');
INSERT INTO evaluations VALUES(46657,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(46657,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46657,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(46657,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(46657,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(46657,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(46657,36,'cp 31','f8g7 d2d4 e8g8 g1f3 c7c5 d4d5 e7e6 e1g1 f8e8 c1f4 e6e5 f4g5 h7h6 g5f6 d8f6 e2e4 f6d8 d1d2 a7a6 a2a4 b7b6 g1h1 a8a7 f3g1 a7e7 g2h3 b8d7 g1f3 h6h5 f1g1 g7f6 d2c2 f6g7 a1e1 e8f8');
INSERT INTO evaluations VALUES(49376,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(49376,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49376,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49376,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49376,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49376,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(49376,36,'cp 0','d7d5 g1f3 f8d6 d2d4 b7b6 b2b3 c8b7 f1d3 b8d7 e1g1 a7a6 d1c2 d5c4 b3c4 b7f3 g2f3 c7c5 c1a3 d8e7 c3e4 f6e4 d3e4 d6h2 g1h2 e7h4 h2g2 h4g5 g2h1 g5h4');
INSERT INTO evaluations VALUES(49627,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(49627,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(49627,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(49627,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(49627,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(49627,35,'cp -148',NULL);
INSERT INTO evaluations VALUES(49627,36,'cp -142','d7d5 e4d5 c8g4 f3d3 f8d6 d3c2 f6d5 f1c4 c7c6 d2d4 e8g8 c4e2 g4e2 g1e2 e5d4 c3d4 f8e8 e1g1 b8d7 b1c3 d8h4 h2h3 d7f6 c2d3 e8e6 c1d2 a8e8 c3d5 f6d5 e2g3 d6g3 f2g3 h4e4 d3e4 e6e4 a1e1 f7f6 e1e4 e8e4 d2c3 g8f7 g1f2 f7e6 f1b1 f6f5 a2a3 d5f6 b1e1 e6d5 e1d1 d5c4');
INSERT INTO evaluations VALUES(54171,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(54171,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54171,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(54171,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(54171,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(54171,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(54171,36,'cp -8','c5d4 e3d4 f8b4 b1c3 g8f6 g1f3 e8g8 a1c1 b7b6 f4g5 b4e7 g5f6 e7f6 c4d5 e6d5 f1e2 b8c6 e1g1 a7a6 f3e5 c6e5 d4e5 f6e5 d1d5 d8d5 c3d5 c8e6 d5b6 a8b8 b6c4 e5b2 c4b2 b8b2 e2a6 e6a2 h2h4 h7h5 a6c4 a2c4 c1c4 f8d8 c4c5 g7g6 g2g3 d8d7');
INSERT INTO evaluations VALUES(54339,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(54339,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(54339,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(54339,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(54339,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(54339,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(54339,36,'cp -1','e3d4 f8b4 b1c3 g8f6 g1f3 e8g8 a1c1 b7b6 f4g5 b4e7 g5f6 e7f6 c4d5 e6d5 f1e2 b8c6 e1g1 g7g6 d1d2 a7a6 f3e5 c6e5 d4e5 f6e5 d2d5 d8d5 c3d5 e5b2 c1b1 b2a3 d5b6 a8a7 b6c8 f8c8 e2d3 a6a5 g2g3 a5a4 g1g2 a3f8 f1d1 a4a3 b1c1 a7c7 c1c7 c8c7');
INSERT INTO evaluations VALUES(55987,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(55987,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(55987,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(55987,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55987,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55987,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55987,36,'cp 0','b8c6 g2g3 d8d5 c2c4 d4c3 b1c3 f8b4 d3d4 b7b6 c1d2 b4c3 b2c3 c8b7 f1g2 d5b5 g2f1');
INSERT INTO evaluations VALUES(21479,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21479,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21479,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21479,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21479,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21479,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21479,36,'cp 0','g2g3 d8d5 c2c4 d4c3 b1c3 f8b4 d3d4 b7b6 c1d2 b4c3 b2c3 c8b7 f1g2 d5c4 g2f1');
INSERT INTO evaluations VALUES(1795,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(1795,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(1795,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(1795,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(1795,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(1795,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(1795,36,'cp 84','e1g1 g8e7 c2c3 a7a6 b5a4 g7g5 d3d4 e5d4 c3d4 f8g7 d4d5 b7b5 a4b3 g5g4 f3e1 c6e5 b1c3 d7d6 c1e3 e8g8 c3e2 a6a5 a1c1 a5a4 b3c2 c7c5 d5c6 e7c6 b2b3 c8e6 e1d3 a4b3 a2b3 e5g6 d3f4 g6f4 e2f4');
INSERT INTO evaluations VALUES(5529,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(5529,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(5529,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(5529,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(5529,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(5529,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(5529,36,'cp 93','b1c3 g8f6 d4d5 c6a5 c4d3 a7a6 e1g1 c8g4 b2b3 b7b5 c1g5 e8g8 h2h3 g4f3 d1f3 f6d7 a1c1 a5b7 c3e2 b7c5 d3b1 a6a5 f1d1 b5b4 e2d4 g7d4 d1d4 d8b6 f3d1 f8e8 g5h6 a5a4 h6d2 a4b3 a2b3 c5a6 d2e3 a6c5');
INSERT INTO evaluations VALUES(5527,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(5527,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(5527,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(5527,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(5527,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(5527,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(5527,36,'cp 77','g8f6 d4d5 c6a5 c4d3 a7a6 e1g1 c8g4 b2b3 b7b5 c1e3 a8c8 c3e2 e7e6 d5e6 g4e6 e2f4 e6d7 f4d5 f6d5 e4d5 e8g8 a1c1 d7g4 c1c8 d8c8 h2h3 g4f3 d1f3 c8c3 f3e2 f8e8 f1b1 a5b7 b3b4 c3e5 a2a4 b5a4 d3a6 e5d5 b1d1');
INSERT INTO evaluations VALUES(8321,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(8321,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(8321,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(8321,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(8321,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(8321,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(8321,36,'cp 101','d4d5 c6a5 c4d3 a7a6 e1g1 c8g4 c1e3 a8c8 a1c1 b7b5 b2b3 c8c3 c1c3 f6d5 e4d5 g7c3 d1c1 c3g7 f3d4 g4d7 b3b4 a5c4 d3c4 b5c4 c1c4 g7d4 c4d4 f7f6 d4a7 e8f7 f1c1 d7b5 c1c7 g6g5 c7b7 h7h5 a2a4 b5a4 a7a6 a4b3 a6c6 d8a8 b4b5');
INSERT INTO evaluations VALUES(7104,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(7104,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(7104,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(7104,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(7104,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(7104,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(7104,36,'cp 25','a8b8 a3c4 b7b5 d5f6 d8f6 c4d6 e8e7 d6c8 b8c8 f1d3 h8d8 e1g1 c6d4 c2c3 d4e6 d1c2 f6g5 f1d1 h7h5 d3f1 e6d4 c2c1 d4f3 g1h1 g5c1 a1c1 f3g5 f2f3 h5h4 f1e2 h4h3 d1d8 h3g2 h1g2 c8d8 c1d1 d8h8 d1d5 f7f6 e2f1 g5h3 g2g3 h3f4 d5d1 g7g5 d1a1 h8b8 b2b4 b8c8 a1c1 c8d8 c3c4');
INSERT INTO evaluations VALUES(9015,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(9015,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(9015,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(9015,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(9015,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(9015,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(9015,36,'cp 94','h7h6 c2c4 e8g8 a1c1 f8e8 f3e5 c6e5 d4e5 f6e4 d2e4 d5e4 d1g4 d8e7 c4c5 f7f5 e5f6 e7f6 f1c4 c8d7 g4f4 f6f4 g3f4 a8d8 b2b4 d7c6 e1e2 c6d5 g2g4 c7c6 a2a4 g8f7 e2f1 a7a5 b4a5 e6e5 f4f5 d5c4 c1c4 d8d1 f1g2 d1h1 g2h1');
INSERT INTO evaluations VALUES(12901,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(12901,31,'cp -147',NULL);
INSERT INTO evaluations VALUES(12901,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(12901,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(12901,34,'cp -153',NULL);
INSERT INTO evaluations VALUES(12901,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(12901,36,'cp -173','d7d5 f1d3 e7e5 b1c3 f8b4 g1e2 d5e4 f3e4 b4c5 c3a4 d8a5 c2c3 f6g4 h1f1 c5e3 h2h3 e3c1 d1c1 g4h2 f1h1 a5a4 h1h2 c8e6 c1e3 a8d8 e2g3 e8g8 e1f2 c6a5 h2h1 a5c4 d3c4 a4c4 h1c1 c4a6');
INSERT INTO evaluations VALUES(14223,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(14223,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(14223,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(14223,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(14223,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(14223,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(14223,36,'cp 34','c8f5 g1f3 e7e6 f3h4 d8d7 e2e3 f8d6 f4d6 d7d6 f1e2 f5e4 h4f3 e4f3 e2f3 e8g8 e1g1 e6e5 f3e2 e5d4 c3d4 c6d8 g2g3 f8e8 e2d3 g7g6 a1c1 c7c6 d2f3 g8g7 g1g2 e8e7 d1c2 a7a6 f3e5 a6a5 h2h4');
INSERT INTO evaluations VALUES(24821,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(24821,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(24821,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(24821,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24821,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(24821,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(24821,36,'cp 36','g1f3 e7e6 f3h4 d8d7 e2e3 f8d6 f4d6 d7d6 f1e2 f5e4 h4f3 e4f3 d2f3 e8g8 e1g1 c6e7 a1c1 c7c5 c3c4 d5c4 d4c5 d6c5 c1c4 c5b6 d1b3 b6b3 a2b3 e7c6 f1c1 h7h6 g2g3 a8c8 b3b4 a7a6 g1g2 c8c7');
INSERT INTO evaluations VALUES(16267,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16267,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(16267,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(16267,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(16267,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(16267,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(16267,36,'cp -5','d7d5 c4d5 e6d5 d2d4 h7h6 g1h3 f8b4 e1g1 e8g8 h3f4 c6e7 d1b3 a7a5 c3d5 e7d5 f4d5 f6d5 a2a3 b4e7 b3d5 d8d5 g2d5 f8d8 d5a2 d8d4 c1e3 d4d7 f1d1 a5a4 a1c1 d7d1 c1d1 g8f8 f2f3 e7d6 g1f2 f8e7 e3f4 d6f4 g3f4 c8e6 a2e6 f7e6 f2g3 a8d8 d1d8 e7d8 g3f2 c7c6');
INSERT INTO evaluations VALUES(17558,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(17558,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(17558,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(17558,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(17558,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(17558,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(17558,36,'cp 86','b5c6 b7c6 e4e5 f7f6 e5d6 e7e5 d2d3 f8d6 f3d2 g8e7 d2e4 e8g8 e4d6 d8d6 b1d2 e7d5 f1e1 a6a5 b2b3 c8f5 d2c4 d6c7 a2a4 f5e6 d1e2 f8e8 f2f3 d5f4 c1f4 e5f4 e2f2 e6c4 d3c4 c7b6 e1e2 e8e2 f2e2 a8d8 a1e1 b6c7 h2h3 c7d6');
INSERT INTO evaluations VALUES(17543,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(17543,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(17543,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(17543,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(17543,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(17543,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(17543,36,'cp 78','b7c6 e4e5 f7f6 e5d6 e7e5 d2d3 f8d6 f3d2 g8e7 d2c4 e8g8 c4d6 d8d6 b1d2 e7d5 b2b3 c8f5 d2c4 d6e7 f2f3 a6a5 a2a4 f5e6 d1e1 e6f7 e1f2 d5c7 c1e3 c7e6 a1e1 e7a7 f3f4 e5f4 e3f4 e6f4 f2f4 f8e8 f4g4 e8e1 f1e1 a8d8 h2h3 a7c7 g4f5 c7g3 g1f1 f7c4 b3c4');
INSERT INTO evaluations VALUES(18393,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(18393,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(18393,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(18393,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(18393,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(18393,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(18393,36,'cp 166','g8f6 e4e5 f6d5 d2d3 d7d6 b3d5 c6d5 c1f4 f8e7 e5d6 e7d6 f4d6 d8d6 f1e1 c8e6 b1d2 e8g8 d2b3 e6d7 d1f3 b5b4 f3g3 d6c6 b3d4 c6c5 g3f4 a7a5 e1e2 a8e8 a1e1 e8e2 e1e2 a5a4 g1f1 h7h6 h2h4 a4a3 b2a3 b4a3 d4b3 c5c3 f4d4 c3d4 b3d4 f8c8 e2e7');
INSERT INTO evaluations VALUES(19486,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(19486,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(19486,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(19486,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(19486,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(19486,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(19486,36,'cp -17','g8f6 f1b5 d8c7 b2b3 g7g6 c1b2 f8g7 f3e5 c6e5 b2e5 c7d8 e1g1 e8g8 b5e2 d7d6 e5b2 d6d5 d2d3 f6e8 b2g7 e8g7 e3e4 d5e4 d3e4 d8d1 e2d1 b7b6 d1e2 c8b7 e2d3 a8d8 b1d2 g7e6 a2a4 g8g7 a1d1 e6d4 g1f2');
INSERT INTO evaluations VALUES(20932,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(20932,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(20932,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(20932,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(20932,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(20932,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(20932,36,'cp 42','g8f6 f1b5 g7g5 h4g3 f6e4 b1d2 e4g3 h2g3 f8g7 g1f3 e8g8 e1g1 c6b8 c2c3 c8f5 f3e1 e7e5 d4e5 g7e5 b5d3 f5d3 e1d3 e5g7 d1f3 c7c6 e3e4 d8d6 f1e1 b8d7 a1d1 f8e8 e4d5 d6d5 f3d5 c6d5 d2b3 b7b6 e1e8 a8e8 g1f1 a7a5 g3g4 a5a4');
INSERT INTO evaluations VALUES(14143,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(14143,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(14143,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(14143,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(14143,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(14143,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(14143,36,'cp 35','f1b5 g7g5 h4g3 f6e4 c2c4 d5c4 g1f3 f8g7 d1c2 e4d6 b5c6 b7c6 b1d2 c8f5 e3e4 f5e6 a1d1 e8g8 e1g1 d8e8 b2b3 f7f5 e4f5 c4b3 d2b3 e6d5 f3e5 d6f5 f1e1 e7e6 e5g4 a8d8 g3e5 e8g6 g4e3 g5g4 e3d5 c6d5');
INSERT INTO evaluations VALUES(25092,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(25092,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(25092,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(25092,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(25092,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(25092,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(25092,36,'cp 16','f8e7 e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 f8e8 d4h4 h7h6 g5f6 d8f6 h4f6 g7f6 f3d4 c6d7 g2g3 d7g4 d1d3 a8d8 d4b5 d6e5 d3d8 e8d8 f2f4 e5c3 b5c3 g8f8 h1e1 c7c6 e1e3 b7b6 e3d3 d8d3 c2d3 f8e7 c1d2 h6h5 b2b3 g4e6 c3d1 f6f5 d1e3 c6c5 d3d4 c5d4 e3c2 e7d6 d2d3');
INSERT INTO evaluations VALUES(24040,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(24040,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24040,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24040,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(24040,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24040,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24040,36,'cp 21','e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 f8e8 d4h4 h7h6 g5f6 d8f6 h4f6 g7f6 f3d4 c6d7 g2g3 d7g4 d1d3 a8d8 d4b5 d6e5 d3d8 e8d8 f2f4 e5c3 b5c3 g8f8 h1e1 c7c6 e1e3 b7b6 e3d3 d8d3 c2d3 h6h5 c1d2 f8e7 a2a3 e7d6 c3e4 d6e7 d2c3 g4e6 e4d2 f6f5 d2c4 e6c4 d3c4 c6c5 b2b4 e7d6');
INSERT INTO evaluations VALUES(24039,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24039,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24039,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24039,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(24039,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(24039,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24039,36,'cp 22','e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 f8e8 d4h4 h7h6 g5f6 d8f6 h4f6 g7f6 f3d4 c6d7 g2g3 d7g4 f2f3 g4h3 c3b5 d6e5 f3f4 e5d4 d1d4 e8e2 b2b3 a8e8 b5c3 e2g2 d4d2 g2d2 c1d2 e8d8 d2c1 d8e8 h1d1 h3g4 d1d4 e8e1 c1b2 g8g7 d4a4 a7a6 a4b4 b7b6 b4c4 e1h1 c4c7');
INSERT INTO evaluations VALUES(23687,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(23687,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(23687,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(23687,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(23687,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23687,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(23687,36,'cp 25','h1e1 f8e8 g5h4 d8d7 h2h3 e7d8 g2g4 h7h5 c3d5 c6d5 e4d5 c7c5 d5c6 d7c6 e1e8 f6e8 d4f4 d8h4 f3h4 a8c8 c2c3 c6c4 f4c4 c8c4 f2f3 c4f4 d1d3 g8f8 c1c2 g7g6 h4g2 f4f6 g4h5 g6h5 c2d2 e8g7 d2e3 g7f5 e3e2 f5e7 d3d1 e7g6 h3h4');
INSERT INTO evaluations VALUES(28086,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28086,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(28086,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(28086,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28086,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(28086,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28086,36,'cp 0','b8d7 f1d3 f5d3 c2d3 g8e7 a2a3 b4a5 b2b4 a5b6 e1g1 e8g8 c3e2 h7h6 a3a4 a7a5 a1b1 a5b4 d2b4 c6c5 d4c5 b6c5 b4d2 c5b6 d2b4');
INSERT INTO evaluations VALUES(31075,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31075,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31075,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31075,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(31075,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31075,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(31075,36,'cp 7','e7e6 g1f3 f8e7 f1d3 g8f6 e1g1 b8d7 d3f5 e6f5 h2h3 f6e4 c3e4 f5e4 f3d2 d7f8 c2c4 f8e6 f4h2 e8g8 d1b3 d8d7 d2b1 a7a5 b1c3 f8d8 c4d5 c6d5 f1c1 a8a6 b3b5 e7b4 b5d7 d8d7 c1c2 a6c6 a1c1 b4c3 c2c3');
INSERT INTO evaluations VALUES(31449,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(31449,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31449,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(31449,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31449,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31449,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(31449,36,'cp 0','f1d3 g8f6 d3f5 e6f5 d1d3 g7g6 c3e2 b8d7 g1f3 f8g7 e1g1 e8g8 a2a4 f6e4 h2h3 a8c8 a4a5 f8e8 f4h2 c6c5 c2c3 c5c4 d3d1 d7f6 f3e5 f6d7');
INSERT INTO evaluations VALUES(31672,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31672,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31672,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(31672,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31672,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31672,36,'cp 0','f1b5 b8d7 c2c4 g8e7 d4c5 a7a6 b5d7 d8d7 c1e3 a8d8 e1g1 d5c4 d1d7 d8d7 b1d2 e7d5 d2c4 d5e3 c4e3 f8c5 e3f5 e6f5 a1c1 c5b6 c1c8 d7d8 c8c3 d8d7');
INSERT INTO evaluations VALUES(32228,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(32228,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(32228,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(32228,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(32228,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(32228,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(32228,36,'cp 105','d2d4 e5d4 f3d4 b8c6 b1c3 g8f6 c1e3 a7a6 d4e6 f7e6 g2g3 g7g6 f1g2 f8g7 e1g1 e8g8 a2a4 d8e7 d1d2 c6e5 d2e2 a8e8 f2f4 e5f7 a4a5 e6e5 f4f5 g7h6 e3h6 f7h6 e2d2 h6f7 c3d5 f6d5 d2d5 c7c6 d5b3 e7c7 a1e1 g8h8 h3h4 c7a5');
INSERT INTO evaluations VALUES(32219,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(32219,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(32219,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(32219,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(32219,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(32219,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(32219,36,'cp 112','g8f6 b1c3 e5d4 f3d4 b8c6 c1e3 a7a6 d4e6 f7e6 g2g3 g7g6 f1g2 f8g7 e1g1 e8g8 d1e2 f6d7 f2f4 g8h8 e2d2 d8e7 a1d1 f8f7 b2b3 a8f8 a2a4 g7h6 b3b4 h6g7 b4b5 a6b5 a4b5 c6a5 d2d3 b7b6 c3e2 e6e5 f4f5 a5b7 g3g4 b7c5');
INSERT INTO evaluations VALUES(33755,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(33755,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(33755,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(33755,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(33755,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(33755,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(33755,36,'cp -69','b8c6 g1f3 e7e5 f1e2 c8f5 e1g1 e8c8 a2a3 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 c5e3 f2e3 h8e8 a1d1 f5c5 g1f2 c5e7 d1d8 c6d8 e2d3 g7g6 b2b4 a7a6 f1d1 f7f5 c2c4 d8f7 c4c5 c8b8 h2h3 f5f4 e3e4 e7e4 d3e4');
INSERT INTO evaluations VALUES(900,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(900,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(900,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(900,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(900,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(900,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(900,36,'cp -59','g1f3 c8f5 a2a3 e8c8 f1e2 e7e5 e1g1 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 c5e3 f2e3 f5c5 a1d1 c5e7 f1e1 h8e8 d1d8 e8d8 e3e4 f7f6 e2e3 d8e8 c2c3 c8b8 h2h3 b7b6 b2b4 h7h5 c3c4 c6e5 f3e5 e7e5');
INSERT INTO evaluations VALUES(899,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(899,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(899,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(899,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(899,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(899,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(899,36,'cp -69','c8f5 f1e2 e7e5 e1g1 e8c8 a2a3 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 h8e8 a1d1 c5e3 f2e3 g7g6 d1d8 c6d8 f1e1 f5e6 e2b5 d8c6 e3e4 f7f6 h2h3 e6d6 b2b4 d6e7 b5d3 a7a6 d3c3 h7h5 f3h4 c6e5 h4f3 e5f3 g2f3 e7d6 g1g2 e8d8 e4e5');
INSERT INTO evaluations VALUES(35022,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(35022,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(35022,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(35022,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(35022,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(35022,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(35022,36,'cp 149','d7d6 d2d4 g8e7 b1c3 c7c6 a2a4 e7g6 c1e3 b8d7 a4a5 f6d8 a5a6 f8e7 d4e5 d7e5 a6b7 c8b7 c4b3 e5g4 e3a7 e8g8 h2h3 g4f6 a7e3 a8a1 d1a1 c6c5 a1a4 d8c7 f1d1 b7c6 a4c4 c7b7 f3d2 g6e5 c4e2 f6e4 d2e4 c6e4 c3e4 b7e4 b3d5 e4f5 f2f4');
INSERT INTO evaluations VALUES(34755,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(34755,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(34755,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(34755,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(34755,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(34755,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(34755,36,'cp 164','d2d4 c8g4 b1d2 b8d7 c2c3 g7g5 d1b3 a8b8 h2h3 g4h5 c4b5 c7c6 b5e2 g8e7 b3d1 h5g6 d4e5 d7e5 a2a4 e5f3 e2f3 f8g7 d2c4 d6d5 c1e3 e8g8 e4d5 c6d5 e3d4 f6f4 g2g3 f4c7 c4e3 g7d4 d1d4 e7c6 d4d5 f8d8 d5b3 c6e5 f3g2');
INSERT INTO evaluations VALUES(34757,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(34757,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(34757,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(34757,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(34757,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(34757,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(34757,36,'cp 147','g8e7 b1c3 c7c6 a2a4 e7g6 a4a5 b8d7 d4d5 f6d8 b2b4 d7f6 a5a6 f8e7 b4b5 c6b5 a6b7 c8b7 c4b5 f6d7 c1e3 e8g8 e3a7 d8c7 d1d2 d7f6 a7e3 f6g4 g2g3 f8c8 a1a8 c8a8 f1b1 g4e3 d2e3 a8a3 b1b3 a3b3 c2b3 c7a5 e3d3 e7d8 g1g2 d8b6 h2h3 g6f8 c3a4 b7c8 b5c4 g8h8');
INSERT INTO evaluations VALUES(41934,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(41934,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(41934,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(41934,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(41934,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(41934,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(41934,36,'cp 54','d5c3 b2c3 b4e7 e2e4 c7c5 f1b5 c8d7 b5c4 d7c6 d1e2 c5d4 c3d4 e8g8 e1g1 b8d7 c4d3 e7a3 a1b1 a8c8 b1b3 a3d6 f1c1 f8e8 b3c3 h7h6 h2h3 d8e7 a2a4 d6b4 c3c2 a7a5 d2b4 e7b4 c2c4 b4e7 c4c3 e8d8 d3b5 d7b6 b5c6 b7c6 c3c6 c8c6 c1c6');
INSERT INTO evaluations VALUES(44972,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(44972,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(44972,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(44972,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(44972,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(44972,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(44972,36,'cp -41','e7e5 a2a3 f8e7 f1e2 d5d4 c3b1 d8c7 c2c3 c6c5 e1g1 b8c6 b2b4 d4c3 b1c3 e8g8 b4c5 e7c5 c1e3 c7e7 d1c1 c5e3 f2e3 f8d8 h2h3 h7h6 c1d2 a8b8 a3a4 b7b6 f1d1 c8d7 d2b2 b8c8 a1b1 c6a5 d1c1 c8c5 b2a3 d7c6');
INSERT INTO evaluations VALUES(45819,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(45819,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(45819,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(45819,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(45819,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(45819,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(45819,36,'cp -27','e4d5 c6d5 d3d4 e5e4 f3e5 f8b4 f1b5 c8d7 e1g1 e8g8 b5e2 b8c6 a2a3 b4c3 b2c3 c6e7 a1b1 b7b6 c3c4 d7e6 c4d5 f6d5 b1b3 f7f6 c2c4 f6e5 c4d5 e6d5 b3c3 d8d6 d4e5 d6e5 c1b2 e7f5 c3c4 e5e6 c4c7 f8f7 c7f7 e6f7 e2h5');
INSERT INTO evaluations VALUES(45204,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(45204,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(45204,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(45204,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(45204,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(45204,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(45204,36,'cp -10','g7g6 c2c3 b8c6 f1e2 f8g7 e1g1 e8g8 a2a3 b7b6 f1e1 e7e5 b2b4 c8b7 b4b5 c6e7 c3c4 f6e8 b1c3 e8c7 c1d2 a7a5 b5a6 a8a6 a1b1 c7e6 c3d5 e7c6 e2f1 b7c8 a3a4 c8d7 d2c3 h7h6 d1c2 g8h7 a4a5');
INSERT INTO evaluations VALUES(44772,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(44772,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(44772,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(44772,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(44772,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(44772,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(44772,36,'cp -9','c2c3 f8g7 f1e2 b7b6 e1g1 c8b7 f1e1 e8g8 e2f1 b8d7 a2a4 a7a6 c1f4 e7e5 f4h2 b6b5 b1d2 f8e8 d3d4 c5c4 d4e5 d6e5 a4b5 a6b5 a1a8 d8a8 b2b3 c4b3 d1b3 d7c5 b3b5 f6e4 d2c4 e8c8 c4b6 e4c3 b6a8 c3b5 a8b6 c8c6 b6d5');
INSERT INTO evaluations VALUES(46177,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46177,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46177,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(46177,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46177,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46177,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46177,36,'cp 0','d5c4 d3c4 c5d4 e3d4 h7h6 e1g1 f8d6 b1c3 b8c6 f1e1 e8g8 a2a3 c8d7 d4d5 e6d5 c3d5 f8e8 c1d2 c6e5 f3e5 d6e5 e1e5 e8e5 d2c3 e5d5 c4d5 f6d5 d1d5 d7c6 d5d8 a8d8 a1e1 f7f6 c3b4 d8d5 e1e3 d5d1 e3e1 d1d5');
INSERT INTO evaluations VALUES(50825,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(50825,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(50825,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(50825,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(50825,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(50825,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(50825,36,'cp 100','d2d4 b8c6 b1c3 e7e5 h2h3 g8f6 c1e3 c6d4 f3d4 e5d4 e3d4 c8e6 f2f4 d6d5 f4f5 e6c8 d1e2 f6e4 c3e4 d5e4 e2e4 d8e7 f1d3 e7e4 d3e4 f8b4 e1f2 e8g8 g2g4 f8e8 f2f3 a6a5 h3h4 a5a4 a1d1 b4d6 c2c4 a8a5 e4d5 g8f8 h1e1 f7f6 e1e8 f8e8 d4c3 a5c5');
INSERT INTO evaluations VALUES(53669,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(53669,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(53669,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(53669,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(53669,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(53669,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(53669,36,'cp 125','c3e4 b8d7 c1f4 d7f6 e4g5 h7h6 g5f7 e8f7 f3e5 f7e8 f1d3 g8e7 d1e2 g7g6 e1c1 f8g7 e5g6 f6d5 f4g3 e7g6 d3g6 e8d7 g6e4 b7b5 c1b1 c8b7 h1e1 h8e8 e4g6 d8f6 g6e8 a8e8 e2d2 b5b4 f2f3 b7a6 g3e5 f6g6 e5g7 g6g7 g2g4 e8g8');
INSERT INTO evaluations VALUES(54828,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(54828,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(54828,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(54828,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(54828,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(54828,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(54828,36,'cp 71','f8g7 b1c3 a7a6 f2f3 g8f6 d1d2 b8d7 h2h4 b7b5 e1c1 h7h5 g1h3 d7b6 h3g5 c7c6 f1d3 d8c7 d2f2 b6c4 d3c4 b5c4 f2e2 a6a5 e4e5 f6d5 e2c4 c7b7 e3d2 a8b8 c4b3 b7b3 a2b3 d6e5 d4e5 g7e5 h1e1 d5c3 d2c3 e5c3 b2c3');
INSERT INTO evaluations VALUES(56409,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(56409,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(56409,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(56409,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(56409,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(56409,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(56409,36,'cp -144','e5f4 c3d5 d8h4 e1e2 b8c6 d5c7 e8d8 g1f3 c8g4 c7a8 c6e5 d1e1 e5f3 e1h4 f3h4 e2e1 g7g5 d2d3 d8c8 c1d2 c8b8 d2c3 f7f6 a8b6 a7b6 a2a4 f8d6 a4a5 d6c7 e1f2 b6a5 c3a5 c7a5 a1a5 g8e7 h2h3 e7c6 a5d5 g4e6 d5d6 h8e8 d3d4 b8c7 d6e6 e8e6 d4d5 e6e4 d5c6 c7c6 f1e2 e4e5 h1d1');
INSERT INTO evaluations VALUES(4977,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(4977,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(4977,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(4977,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(4977,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(4977,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(4977,36,'cp 125','e3g3 g8f6 c1d2 e8g8 e1c1 d7d6 h2h4 c6e5 h4h5 h7h6 f2f3 g8h8 f1e2 f8e8 g3h4 f6h7 h4d8 e8d8 g1h3 c8h3 h1h3 h7g5 h3h1 g5e6 c3a4 h8g8 c2c3 a7a5 c1c2 a8b8 a4c5 b6c5 b2b3 c5c4 e2c4 e5c4 b3c4 e6c5 d2e3 a5a4');
INSERT INTO evaluations VALUES(4978,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(4978,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(4978,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(4978,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(4978,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(4978,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(4978,36,'cp 113','g8f6 c1d2 e8g8 e1c1 d7d5 e4d5 c5d6 g3h4 c6e5 d2g5 e5g6 g5f6 g6h4 f6d8 f8d8 g2g3 h4f5 g1f3 d6b4 f3d4 f5d4 d1d4 c7c5 d4d3 d8e8 a2a3 b4a5 d3e3 e8e3 f2e3 c8g4 f1e2 a5c3 e2g4 c3e5 h1f1 a8d8 e3e4 d8e8 g4d7 e8e7 d7b5 e5d6 b5d3 c5c4 d3c4 e7e4');
INSERT INTO evaluations VALUES(5249,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(5249,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(5249,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(5249,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(5249,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(5249,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(5249,36,'cp -12','b7c6 e1g1 e7e5 a2a3 a7a5 a1b1 d7d6 b2b4 a5b4 a3b4 c5b4 b1b4 a8a1 b4a4 a1a4 c3a4 g8f6 d2d3 d8a5 a4b2 a5c7 h2h3 e8g8 c2c4 c8e6 d1c2 h7h6 b2a4 f6d7 c1e3 g8h7 f1b1 f8b8 b1b8 c7b8 f3d2 b8a8 c2b3 a8a6');
INSERT INTO evaluations VALUES(8004,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(8004,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(8004,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(8004,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(8004,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(8004,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(8004,36,'cp 106','e4e5 f6h5 d2d4 c5d4 c3d4 d7d6 f3g5 e8g8 e5e6 d8b6 e6f7 g8h8 c4e2 c6d4 e2h5 g6h5 d1h5 c8f5 b1c3 f5g6 h5h4 g7f6 c3d5 d4e2 g1h1 b6d4 h4d4 e2d4 c1d2 h7h6 d5f6 e7f6 g5h3 h8h7 a1d1 d4f5 d2c3 g6f7 g2g4 f5e7 h3f4 f7a2 d1d6 a8d8 d6d8 f8d8');
INSERT INTO evaluations VALUES(10463,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(10463,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(10463,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(10463,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(10463,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(10463,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(10463,36,'cp -6','d2c3 d8e7 d1e2 e8g8 g2g3 d7d5 f1g2 c8g4 e1g1 a8d8 h2h3 g4f3 g2f3 d5e4 f3e4 f6e4 e2e4 e7c5 a1e1 f7f6 e4e3 c5e3 f2e3 f8e8 g1f2 b7b6 a2a4 c6e7 e3e4 a7a5 f2e2 d8d7 e1d1 e7c6 g3g4 e8d8 h3h4');
INSERT INTO evaluations VALUES(10615,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(10615,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(10615,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(10615,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(10615,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(10615,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(10615,36,'cp 70','d7d6 b2b4 c5b6 a2a4 a7a6 c1g5 g4f6 b1d2 e8g8 c4a2 c8e6 d2c4 h7h6 g5h4 b6a7 h2h3 g8h8 b4b5 a6b5 a4b5 c6e7 h4f6 g7f6 d1d2 h8g7 c4e3 a7e3 f2e3 d6d5 e4d5 e6d5 a2d5 d8d5 c3c4 d5d6 d3d4 a8a1 f1a1 e5d4 e3d4 e7f5');
INSERT INTO evaluations VALUES(9466,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(9466,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(9466,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(9466,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(9466,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(9466,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(9466,36,'cp 79','b2b4 c5b6 a2a4 a7a6 c1g5 g4f6 b1d2 e8g8 f1e1 h7h6 g5h4 c8e6 c4d5 g7g5 d5e6 f7e6 h4g3 f6h5 d2c4 b6a7 b4b5 g5g4 g3h4 d8e8 f3d2 g4g3 g1h1 g3f2 e1f1 e8g6 b5c6 b7b5 a4b5 a6b5 a1a7 a8a7 c4e3 h5f4');
INSERT INTO evaluations VALUES(11872,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11872,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(11872,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(11872,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(11872,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(11872,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11872,36,'cp 48','b1c3 f8e7 e1g1 e8g8 h2h3 a7a5 b2b3 a5a4 c1f4 a4b3 c2b3 c8e6 c4d3 f6d7 a1c1 e7f6 d3b1 d7e5 d1c2 g7g5 f4h2 f6g7 c3d5 c6c5 d5e3 g8h8 h2g3 a8a3 f1e1 d8f6 e3f5 e6f5 e4f5');
INSERT INTO evaluations VALUES(11871,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(11871,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(11871,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(11871,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(11871,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(11871,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11871,36,'cp 42','f8e7 h2h3 a7a5 e1g1 e8g8 b2b3 c8e6 d1e2 d6d5 c4d3 d5e4 c3e4 e6f5 c1b2 f8e8 b2f6 e7f6 a1d1 f5e4 d3e4 d8e7 d1e1 e7c5 e2f3 a8d8 e4d3 c5c3 e1e8 d8e8 h3h4 g7g6 h4h5 e8e1 g2g3 e1f1 g1f1 g8f8 f1g2 f6e5 a2a4 e5f6 f3f4 f6e5 f4g4');
INSERT INTO evaluations VALUES(7540,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(7540,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(7540,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(7540,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(7540,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(7540,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(7540,36,'cp 31','e1g1 e8g8 h2h3 a7a5 b2b3 f6d7 c1b2 d7c5 f2f4 a5a4 a1b1 a4b3 c2b3 e7f6 d1c2 c5e6 b1d1 e6d4 c2d3 c6c5 b3b4 c5b4 c3d5 c7c5 d5f6 d8f6 f4f5 c8b7 b2d4 c5d4 d3d4 f6e5 f5f6 b7e4 d4d6 e5d6');
INSERT INTO evaluations VALUES(19991,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19991,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(19991,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(19991,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19991,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19991,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19991,36,'cp 0','c5d4 f3d4 g7g6 c1e3 g8f6 b1c3 f6g4 e3g1 e7e5 f4e5 g4e5 d1d2 f8g7 e1c1 e8g8 c1b1 c6d4 g1d4 c8e6 a2a3 a7a6 d4e3 e5g4 e3d4');
INSERT INTO evaluations VALUES(20087,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20087,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20087,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20087,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20087,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20087,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20087,36,'cp 0','f3d4 g8f6 b1c3 g7g6 c1e3 f6g4 e3g1 e7e5 f4e5 g4e5 d1d2 f8g7 e1c1 e8g8 c1b1 c6d4 g1d4 c8e6 a2a3 a7a6 d4e3 e5g4 e3d4');
INSERT INTO evaluations VALUES(20079,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20079,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20079,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20079,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20079,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20079,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20079,36,'cp 0','g7g6 c1e3 g8f6 b1c3 f6g4 e3g1 e7e5 f4e5 g4e5 d1d2 f8g7 e1c1 e8g8 a2a3 c6d4 g1d4 c8e6 c1b1 a8c8 d4a7 e5c4 f1c4 c8c4 d2d6 d8a8 a7d4 f8d8 d6d8 a8d8 d4g7 d8g5 h2h4 g5g2 g7f6 c4c8 h4h5 b7b5 h5h6 g2f2 f6d4 f2g2 d4f6');
INSERT INTO evaluations VALUES(22392,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22392,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(22392,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(22392,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(22392,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(22392,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(22392,36,'cp 84','c7c5 c2c3 g8e7 f1d3 e7c6 e1g1 f8e7 c1e3 d8b6 b2b3 e8g8 b1d2 a7a5 a2a3 f7f6 e5f6 d7f6 f3e5 e7d6 d2f3 c5d4 c3d4 b6c7 f1e1 c8d7 d1b1 h7h6 e3f4 c6e5 f4e5 d7e8 b1b2 d6e5 f3e5 f6d7 a1c1 c7b6 b2e2 b6d4 e5f3 d4f6');
INSERT INTO evaluations VALUES(18548,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(18548,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(18548,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(18548,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(18548,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(18548,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(18548,36,'cp 100','c2c3 g8e7 f1d3 e7c6 e1g1 f8e7 c1e3 d8b6 b2b3 e8g8 b1d2 a7a5 a2a3 b6c7 f3e1 f7f5 f2f4 b7b5 d3b5 c7b6 b5a4 c8a6 f1f2 c5d4 c3d4 b6b7 h2h3 d7b6 a4c6 b7c6 d2f3 a5a4 f2c2 c6b5 c2c7 b5e8 b3a4 b6c4 e3c1 a8c8 c7c8');
INSERT INTO evaluations VALUES(18546,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(18546,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(18546,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(18546,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(18546,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(18546,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(18546,36,'cp 95','g8e7 f1d3 e7c6 e1g1 f8e7 c1e3 d8b6 b2b3 e8g8 b1d2 a7a5 a2a3 f7f5 e5f6 d7f6 f3e5 e7d6 d2f3 c5d4 c3d4 b6c7 f1e1 c8d7 e5d7 c7d7 e3c1 a8e8 a1a2 e6e5 d4e5 c6e5 f3e5 e8e5 e1e5 d6e5 a2e2 d7d6 g2g3 f6e4 c1b2 e5b2 d3e4 b2a3 e4d5 g8h8');
INSERT INTO evaluations VALUES(24219,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(24219,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(24219,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(24219,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(24219,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(24219,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(24219,36,'cp 140','c3d4 e8g8 b1c3 d7g4 h2h3 g4f3 g2f3 g8h8 f3f4 d6d5 e4e5 f6g8 f4f5 f7f6 e5e6 e7d6 b5d3 c6d4 c1e3 c7c5 e3d4 c5d4 c3d5 d6h2 g1h2 d8d5 d1g4 g8e7 a1c1 f8c8 e1g1 d5d6 h2h1 c8g8 g4e4 e7c6 d3b5 c6e7 g1d1 d6b6 b5d7 d4d3 h1g1 b6b2 e4d3 a8d8');
INSERT INTO evaluations VALUES(27253,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(27253,31,'cp 255',NULL);
INSERT INTO evaluations VALUES(27253,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(27253,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(27253,34,'cp 247',NULL);
INSERT INTO evaluations VALUES(27253,35,'cp 259',NULL);
INSERT INTO evaluations VALUES(27253,36,'cp 238','f3d4 e5d4 b5d7 d8d7 d1d4 g8f6 e1g1 f8e7 h2h3 e8g8 b1c3 f8e8 c1f4 f6h5 f4d2 c7c6 a1d1 d7c7 d4d3 h5f6 f1e1 a8c8 d5c6 b7c6 d2e3 c8d8 b2b3 f6d7 c3a4 d7f6 e3f4 c7a5 f4d2 a5h5 a4c3 h5a5 d3f3 f6d7 c3a4 a5c7 c2c4');
INSERT INTO evaluations VALUES(27252,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(27252,31,'cp 228',NULL);
INSERT INTO evaluations VALUES(27252,32,'cp 239',NULL);
INSERT INTO evaluations VALUES(27252,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(27252,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(27252,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(27252,36,'cp 234','e5d4 b5d7 d8d7 d1d4 g8f6 b1c3 f8e7 d4d3 e8g8 h2h3 f8e8 c1e3 d7d8 e1g1 f6d7 a1d1 d7e5 d3e2 e7f6 f2f4 e5d7 e3d4 a7a6 e2f2 f6h4 f2f3 d8e7 g1h2 b7b5 a2a3 h4f6 d4f6 e7f6 d1e1 e8e7 e1e2 a8e8');
INSERT INTO evaluations VALUES(27250,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(27250,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(27250,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(27250,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(27250,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(27250,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(27250,36,'cp 245','b5d7 d8d7 d1d4 g8f6 c2c4 f8e7 e1g1 f6g4 b1c3 e8g8 f2f3 g4e5 f3f4 c7c5 d5c6 e5c6 d4d3 f7f5 c1e3 f5e4 c3e4 c6b4 d3b3 d7e6 a1e1 e6e4 e3f2 e4c2 e1e7 c2b3 a2b3 b4d3 f2d4 f8f7 e7f7 g8f7 g2g3 a8e8 f1d1 d3b4 d4a7 e8e2 a7d4 e2c2');
INSERT INTO evaluations VALUES(27138,30,'cp 242',NULL);
INSERT INTO evaluations VALUES(27138,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(27138,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(27138,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(27138,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(27138,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(27138,36,'cp 222','d8d7 d1d4 g8f6 b1c3 f8e7 d4d3 e8g8 h2h3 f8e8 c1e3 d7d8 e1g1 f6d7 a1d1 d7e5 d3e2 e7f6 f2f4 e5d7 e3d4 a7a6 d4f6 d8f6 e2c4 a8c8 a2a4 f6h4 a4a5 e8e7 b2b3 e7e8 f1f3 d7f6 c4d3 b7b5 a5b6');
INSERT INTO evaluations VALUES(27656,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(27656,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(27656,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(27656,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(27656,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(27656,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(27656,36,'cp 237','d1d4 g8f6 f2f3 f8e7 c2c4 e8g8 b1c3 f6h5 d4d3 e7f6 c3e2 g7g6 e1g1 a8e8 a1b1 f6g7 g2g4 h5f6 g1g2 d7e7 c1g5 h7h6 g5h4 g6g5 h4f2 h6h5 h2h3 f6d7 d3c2 h5g4 h3g4 d7e5 c4c5 e5g6 c5d6 c7d6');
INSERT INTO evaluations VALUES(27655,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(27655,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(27655,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(27655,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(27655,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(27655,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(27655,36,'cp 236','g8f6 b1c3 f8e7 h2h3 e8g8 d4d3 f8e8 c1e3 d7d8 e1g1 f6d7 b2b3 e7f6 a1e1 d7e5 d3d2 e5g6 e3d4 f6d4 d2d4 d8g5 c3e2 g5e5 d4e5 e8e5 f2f3 a8e8 e2d4 g6f4 d4f5 g7g6 f5e3 c7c6');
INSERT INTO evaluations VALUES(27457,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(27457,31,'cp 228',NULL);
INSERT INTO evaluations VALUES(27457,32,'cp 251',NULL);
INSERT INTO evaluations VALUES(27457,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(27457,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(27457,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(27457,36,'cp 252','b1c3 f8e7 d4d3 e8g8 c1e3 c7c5 h2h3 a7a6 e1g1 f6h5 a1d1 a8e8 g2g4 h5f6 e3f4 b7b5 g1g2 h7h5 f2f3 f6h7 c3e2 h5h4 b2b3 e7g5 f4h2 d7e7 a2a4 g5f6 a4b5 a6b5 d3b5 e8a8 b5d3 f8e8 h2f4 a8a2 f4d2 h7g5');
INSERT INTO evaluations VALUES(35378,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(35378,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(35378,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(35378,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(35378,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35378,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(35378,36,'cp 26','d1d3 c7c5 e1c1 f8d8 d3e4 b8c6 f1d3 f6h6 e4e3 h6e3 f2e3 c5d4 d3e4 d4d3 e4c6 b7c6 d1d3 d8d3 c2d3 c8a6 d3d4 f7f6 c1d2 a6c4 e3e4 c4a2 d2c3 f6f5 e4f5 a2d5 f5e6 d5e6 h1a1 e6d5 f3e5 a7a5 g2g3 g8f8 e5c4 d5c4 c3c4 a8b8 c4c5 b8b2 c5c6 b2h2 a1a5');
INSERT INTO evaluations VALUES(36152,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(36152,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(36152,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(36152,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36152,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36152,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(36152,36,'cp 28','a2a3 b4c3 d2c3 c5d4 f3d4 f8e8 f1d3 e6e5 d4e2 d5c4 d3c4 d8c7 c4a2 c8e6 a2e6 e8e6 a1c1 b8d7 e1g1 d7c5 d1c2 c5e4 c2a4 e6a6 a4b3 e4c3 c1c3 a6c6 c3c6 c7c6 h2h3 a8d8 f1c1 c6d5 b3d5 f6d5 e3e4 d5f6');
INSERT INTO evaluations VALUES(36156,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36156,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36156,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(36156,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(36156,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(36156,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36156,36,'cp 16','b4c3 d2c3 c5d4 f3d4 f8e8 f1d3 e6e5 d4e2 b7b6 b2b4 c8a6 c4d5 d8d5 d3a6 b8a6 e1g1 a8d8 a1c1 a6c7 d1a4 d5b7 a4b3 c7b5 f1d1 d8d1 c1d1 b5c3 e2c3 e8c8 b3b2 b7e7 h2h3 g7g6 c3d5 f6d5 d1d5');
INSERT INTO evaluations VALUES(38435,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(38435,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(38435,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(38435,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(38435,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(38435,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(38435,36,'cp 54','d2d4 g8f6 f1d3 e8g8 e1g1 b8d7 c1e3 e7e5 b1d2 b7b6 f1e1 c8b7 h2h3 h7h6 d1c2 f8e8 a1d1 a8c8 a2a4 a7a5 d4e5 d7e5 f3e5 e8e5 e3f4 e5e6 d2f3 c7c5 f3d2 d6d5 e4e5 f6d7 d2f3 d8c7 c2d2 c8a8 f4h2 d7e5 f3e5 g7e5 h2e5 e6e5 d2h6');
INSERT INTO evaluations VALUES(41402,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(41402,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(41402,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(41402,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(41402,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(41402,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41402,36,'cp 35','c7c5 d4c5 d8a5 b1c3 a5c5 d1e2 b8c6 c1e3 c5a5 e1g1 c8g4 h2h3 g4f3 e2f3 e8g8 g1h1 f6d7 e3d2 a8c8 a1d1 a7a6 f4f5 c6e5 f3f2 a5d8 a2a4 e5c4 d3c4 c8c4 f2h4 g7f6 h4g4 d7e5 g4g3 f6g7 g3h4 g7f6 h4f2 f6g7 f2e2 g8h8 b2b3 c4c8');
INSERT INTO evaluations VALUES(42036,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42036,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42036,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(42036,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42036,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(42036,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(42036,36,'cp 5','b1d2 c7c5 d4c5 b4d2 d1d2 f6e4 d2b4 b8c6 b4a3 e6e5 f4g3 d5d4 e1c1 e8g8 f1d3 c8f5 f2f3 e4g3 h2g3 f5d3 a3d3 h7h6 g1e2 f8e8 e3d4 e5d4 d3d2 d8e7 e2d4 e7c5 d4c6 c5c4 d2c3 c4c3 b2c3 b7c6 h1e1 g7g6 c1b2 h6h5 b2a3 g8g7 a3b4 a8d8 e1e8 d8d1 e8e2 d1c1 b4c4 a7a5 g3g4 h5g4 f3g4');
INSERT INTO evaluations VALUES(42359,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(42359,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(42359,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(42359,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(42359,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(42359,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(42359,36,'cp 201','f3e5 e8g8 e1g1 d7d6 e5f3 b8c6 c2c3 c6e5 f3e5 d6e5 a2a4 a7a5 h2h3 f8e8 b1d2 c8e6 d1e2 b7b6 c4e6 e8e6 d2c4 d8d7 f1d1 a8d8 g1h2 d7e7 c1d2 h6h5 h2g1 e7d7 d2c1 g7g6 b2b3 d7e7 d1d2 h5h4');
INSERT INTO evaluations VALUES(42351,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(42351,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(42351,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(42351,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(42351,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(42351,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(42351,36,'cp 189','e8g8 e1g1 d7d6 e5f3 f8e8 h2h3 d6d5 e4d5 f6d5 b1d2 c5f8 f1e1 e8e1 d1e1 b8c6 d2e4 c8e6 c1d2 d8e7 e4c3 a8d8 a2a3 d5c3 c4e6 c3d5 e6d5 d8d5 e1e7 f8e7 d2c3 f7f6 g1f1 g8f7 f1e2 a7a5 g2g4 a5a4 a1d1');
INSERT INTO evaluations VALUES(37142,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(37142,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(37142,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(37142,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(37142,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(37142,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(37142,36,'cp 186','e1g1 d7d6 e5f3 f8e8 h2h3 d6d5 e4d5 f6d5 c4b3 c5f8 d3d4 b8c6 c2c4 d5b6 b1c3 c6a5 d1d3 a5b3 a2b3 c7c6 c1d2 a7a5 f1d1 b6d7 d1e1 e8e1 a1e1 d7f6 d2f4 c8e6 f3e5 f6d7 c3a4 d8c7 e1e3 a8e8 e3g3 d7e5 f4e5');
INSERT INTO evaluations VALUES(46130,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(46130,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(46130,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(46130,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(46130,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(46130,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(46130,36,'cp -17','e6d5 c4b5 b8d7 e1g1 f8e7 d3d4 e8g8 b1c3 h7h6 d4c5 d7c5 f1e1 c8f5 c1e3 c5e4 c3e2 a7a6 b5d3 f6g4 e2g3 e4g3 h2g3 g4e3 e1e3 f5g4 d3e2 e7c5 f3d4 d8b6 e2g4 c5d4 e3f3 b6b2 a1b1 b2a2 b1b7 d4f6 d1d3 a8e8');
INSERT INTO evaluations VALUES(46415,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46415,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46415,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(46415,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(46415,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(46415,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(46415,36,'cp -17','c4b5 b8d7 e1g1 f8e7 b1c3 e8g8 d3d4 h7h6 d4c5 d7c5 f1e1 c8f5 c1e3 c5e4 c3e2 f6g4 e3d4 a8c8 c2c3 e7c5 e1f1 d8b6 d1b3 c5d4 e2d4 e4c5 b3a3 f5e4 b5e2 f8e8 e2b5 e8e7 f1e1 a7a6 b5f1 c8e8 h2h3 g4f6 a1d1 c5e6 a3b3 b6c7 d4e6');
INSERT INTO evaluations VALUES(46394,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(46394,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(46394,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(46394,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(46394,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(46394,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(46394,36,'cp -14','b8c6 e1g1 f8e7 c1g5 e8g8 b5c6 b7c6 b1d2 f8e8 f1e1 c8f5 h2h3 h7h6 g5h4 a7a5 a2a4 d8b6 b2b3 b6b4 d2f1 f5h7 a1b1 h7g6 d1d2 a8c8 d2b4 c5b4 e1e5 e7c5 h4f6 g7f6 e5e8 c8e8 b1e1 e8e1 f3e1 g8h7 f1g3 c5b6 e1f3');
INSERT INTO evaluations VALUES(13720,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(13720,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(13720,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(13720,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(13720,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(13720,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(13720,36,'cp -9','b5c6 b7c6 e1g1 f8e7 f1e1 e8g8 c1g5 f8e8 b1d2 h7h6 g5h4 a7a5 a2a4 c8g4 h2h3 g4h5 g2g4 h5g6 f3e5 e7d6 e5g6 f7g6 e1e8 d8e8 h4f6 g7f6 d1f3 e8e5 a1b1 a8e8 d2b3 f6f5 g4f5 e8f8 f3g3 e5f5 g3d6 f5f2 g1h1 f2f3 h1h2 f3e2 h2g1 e2e3 g1h1 e3h3 d6h2 f8f1 b1f1 h3f1 h2g1 f1h3 g1h2');
INSERT INTO evaluations VALUES(47414,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(47414,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(47414,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(47414,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(47414,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(47414,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(47414,36,'cp 128','d2d4 e7e6 e2e4 a7a6 c1e3 b7b5 f1d3 c8b7 a2a3 f8e7 e1g1 e8g8 d1c2 h7h6 b2b4 a6a5 b4a5 a8a5 f1b1 b5b4 a3b4 a5a1 b1a1 e7b4 a1b1 b4c3 b1b7 c3a5 e3f4 d8c8 b7b1 b8d7 h2h3 a5b6 d3b5 c8a8 b5c6 a8a3 f4e3 f8d8 b1c1 a3a5 g2g3');
INSERT INTO evaluations VALUES(47797,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(47797,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(47797,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(47797,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(47797,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(47797,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(47797,36,'cp -12','c7c6 a2a4 f8e7 g1e2 e8g8 d2d4 c8g4 f3e3 b8d7 e1g1 f8e8 d4e5 d7e5 c4b3 e5d7 f2f3 g4e6 b3e6 f7e6 g1h1 d8b6 e2f4 b6e3 c1e3 e7f8 e3f2 e6e5 f4d3 a7a5 a1d1 d6d5 e4d5 c6d5 c3b5 a8c8 c2c3 e5e4 f3e4 f6e4');
INSERT INTO evaluations VALUES(49900,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(49900,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(49900,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49900,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(49900,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(49900,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(49900,36,'cp 28','d7d5 f3e5 f8d6 b1c3 e4c3 b2c3 b8d7 e1g1 e8g8 f1e1 c7c5 d1f3 c5c4 d3f5 d7e5 d4e5 d6c5 f5c8 a8c8 c1e3 f8e8 f3g3 d8d7 f2f4 b7b6 a1d1 c5e3 g3e3 c8c5 e3f2 c5a5 f4f5 a5a2 f5f6 g7g6 h2h4 d7f5 f2f5 g6f5 d1d5 a2c2');
INSERT INTO evaluations VALUES(49075,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(49075,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(49075,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(49075,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(49075,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(49075,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(49075,36,'cp 20','f3e5 b8d7 e5d7 c8d7 e1g1 f8d6 b1c3 d8h4 g2g3 e4c3 f1e1 d7e6 b2c3 h4h3 d1f3 e8g8 c1f4 a8d8 f4e5 b7b6 e1e3 h3g4 f3g2 g4h3 a1e1 h3g2 g1g2 d6e5 e3e5 e6d7 e5e7 d8e8 g3g4 g7g6 g2g3 e8e7 e1e7 f8d8 f2f4 g8f8 e7e5');
INSERT INTO evaluations VALUES(52457,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(52457,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(52457,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(52457,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(52457,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(52457,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(52457,36,'cp -43','g8f6 b2b3 g7g6 c1b2 f8g7 f1e2 e8g8 g1f3 e7e6 e1g1 b7b6 f3e5 c8b7 e2f3 d7d5 c2c4 f6e4 b1c3 b8c6 e5c6 b7c6 d1c2 e4c3 b2c3 g7c3 c2c3 d8d7 a1d1 d5c4 f3c6 d7c6 c3c4 a8d8 d2d4 c6e4 d4c5 e4e3 g1h1');
INSERT INTO evaluations VALUES(53417,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(53417,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(53417,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(53417,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53417,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53417,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53417,36,'cp 18','b8c6 d2d4 c8g4 h2h3 g4h5 c1f4 e7e6 f1d3 a8c8 e1g1 f8d6 f4d6 d8d6 b1d2 g8e7 f1e1 e8g8 a2a4 d6c7 d2b3 c7b6 d3c2 b6c7 d1e2 f8e8 e2e3 h5g6 c2g6 e7g6 b3c5 c6a5 c5d3 a5c4 e3e2 g6e7 g2g3 b7b6');
INSERT INTO evaluations VALUES(54722,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(54722,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(54722,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(54722,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(54722,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(54722,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(54722,36,'cp 55','d2d4 c8f5 g1f3 e7e6 c2c4 c7c6 g2g3 f8d6 f1g2 g8h6 e1g1 e8g8 b1d2 b8d7 f3e1 f5g4 e1d3 g4h5 d1e1 f6f5 d2f3 a7a5 f3e5 h6g4 h2h3 g4f6 f2f3 d8e7 e2e3 h5f7 a2a4 f8e8 c4c5 d6c7 b2c3 f7h5 b3b4 a5b4 c3b4');
INSERT INTO evaluations VALUES(55063,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(55063,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(55063,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(55063,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(55063,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(55063,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(55063,36,'cp -27','c7c5 e2e3 g8f6 g1f3 b8c6 f1b5 c5d4 e3d4 c8g4 h2h3 g4f3 d1f3 d8a5 b1c3 e7e6 e1g1 f8d6 f1d1 e8g8 b5f1 a8c8 c1g5 f6e4 c3e4 d5e4 f3e3 c6b4 c2c4 h7h6 g5h4 f7f5 e3b3 d6b8 a2a3 b4c6 c4c5 a5c7 b3e6 g8h8 h4g3 f5f4');
INSERT INTO evaluations VALUES(56392,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(56392,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(56392,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(56392,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(56392,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(56392,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(56392,36,'cp 15','e4d5 g8f6 d3b5 c8d7 b5d7 b8d7 c2c4 b7b5 d2d3 c7c6 g1f3 c6d5 c4d5 f8b4 b1d2 e8g8 e1g1 f6d5 d2e4 a7a5 d1b3 d5c7 c1e3 b4e7 a2a3 h7h6 d3d4 a5a4 b3a2 e5d4 e3d4 d7f6 e4f6 e7f6 a1d1 d8d5 a2d5 c7d5 d4f6 d5f6 f1e1 f8e8 e1e8 a8e8');
INSERT INTO evaluations VALUES(56518,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(56518,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56518,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56518,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56518,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56518,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56518,36,'cp 0','f8d6 d2d4 e4d3 e5d3 g8e7 b1c3 e8g8 c1g5 h7h6 g5h4 b8c6 e1c1 c6d4 e2e4 d4f5 c3b5 f5h4 e4h4 c8f5 b5d6 d8d6 d3b4 d6c5 b4d3');
INSERT INTO evaluations VALUES(5532,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(5532,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(5532,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(5532,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(5532,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(5532,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(5532,36,'cp 62','c1e3 g7d4 e3d4 c6d4 d1d4 g8f6 f1e2 e8g8 e1g1 c8d7 h2h3 d8b6 d4b6 a7b6 a2a4 f8c8 f2f3 g8f8 g1f2 d7e6 b2b3 f6d7 f2e3 d7c5 a1a3 f7f6 f1b1 c5d7 h3h4 h7h5 b1b2 f8g7 a3a1 g7f7 f3f4 e6g4 e2f3 g4f3 g2f3 f6f5 a1d1');
INSERT INTO evaluations VALUES(6442,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6442,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6442,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(6442,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6442,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6442,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6442,36,'cp 0','g1f3 g8f6 d2d3 d7d6 b1c3 c6d4 f3d4 c5d4 d1f3 c7c6 c3e2 d4b6 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(7826,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7826,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7826,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7826,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(7826,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(7826,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(7826,36,'cp 10','f1b5 e8g8 e1g1 d7d5 e4d5 e6d5 c1g5 c8e6 g5h4 h7h6 f1e1 a7a6 b5c6 b7c6 d1d3 f8e8 e1e2 c5e7 f3d4 e6g4 e2e5 d8c7 a1e1 e7d6 e5e8 f6e8 h4g3 g4h5 h2h3 h5g6 g3d6 c7d6 d3g3 d6f6 g3e5 f6e5 e1e5 c6c5');
INSERT INTO evaluations VALUES(14368,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(14368,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(14368,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(14368,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(14368,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(14368,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(14368,36,'cp -72','d8d4 d1e2 c6e5 c4b5 c7c6 f2f4 c8g4 g1f3 e5f3 g2f3 g4h5 c1e3 d4d7 e1g1 e7e6 a1d1 d7c7 g1h1 f8e7 b5d3 e8g8 e2g2 f8d8 c3e2 h5g6 e2g3 f6h5 f4f5 e6f5 e4f5 h5g3 g2g3 c7g3 h2g3 d8d3 d1d3 g6f5 d3d2 f5e6 a2a3 a7a5 h1g2 h7h5 f1c1 g7g6 e3d4 e7g5 f3f4');
INSERT INTO evaluations VALUES(14953,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(14953,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(14953,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(14953,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(14953,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(14953,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(14953,36,'cp -115','d5e4 f3e5 d8d5 b5c6 b7c6 d3d4 c6c5 c2c3 f8d6 d1a4 f6d7 e1g1 e8g8 a4c4 d5c4 e5c4 c8a6 b2b3 a6c4 b3c4 f7f5 b1d2 a7a5 a1b1 f8b8 b1b8 a8b8 g2g4 g7g6 f1e1 d6f4 g4f5 g6f5 f2f3 e4f3 d2f3 f4c1 e1c1 b8b2 f3g5 b2a2 g5e6');
INSERT INTO evaluations VALUES(14960,30,'cp -112',NULL);
INSERT INTO evaluations VALUES(14960,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(14960,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(14960,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(14960,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(14960,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(14960,36,'cp -99','f3e5 d8d5 b5c6 b7c6 d3d4 c6c5 c1e3 d5e6 c2c3 f8d6 d1b3 e8g8 b3e6 c8e6 b1d2 c5d4 c3d4 f8b8 b2b3 a7a5 e1g1 a5a4 e5c4 d6b4 e3g5 b4c3 a1c1 c3d4 g5f6 g7f6 d2e4 f6f5 e4d2 c7c5 d2f3 a4b3 a2b3 b8b3 f3d4 c5d4 f1d1 b3c3 c4e5 h7h6 c1a1 a8a1');
INSERT INTO evaluations VALUES(21324,30,'cp 310',NULL);
INSERT INTO evaluations VALUES(21324,31,'cp 304',NULL);
INSERT INTO evaluations VALUES(21324,32,'cp 291',NULL);
INSERT INTO evaluations VALUES(21324,33,'cp 302',NULL);
INSERT INTO evaluations VALUES(21324,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(21324,35,'cp 301',NULL);
INSERT INTO evaluations VALUES(21324,36,'cp 278','f3e5 d6e5 c4f7 e8e7 d1f3 g8f6 f7h5 c8e6 b1c3 e7d7 e1g1 a7a6 f1d1 f8d6 c1e3 d7c8 a2a4 d8d7 h2h3 f6h5 f3h5 d7f7 h5f7 e6f7 a4a5 c8d7 f2f3 a8d8 c3d5 f7e6 g1f2 e6d5 d1d5 d7e6 a1d1 c7c6 d5d3 d6e7 f2e2 d8d3 d1d3');
INSERT INTO evaluations VALUES(21322,30,'cp 291',NULL);
INSERT INTO evaluations VALUES(21322,31,'cp 296',NULL);
INSERT INTO evaluations VALUES(21322,32,'cp 299',NULL);
INSERT INTO evaluations VALUES(21322,33,'cp 282',NULL);
INSERT INTO evaluations VALUES(21322,34,'cp 281',NULL);
INSERT INTO evaluations VALUES(21322,35,'cp 285',NULL);
INSERT INTO evaluations VALUES(21322,36,'cp 269','d6e5 c4f7 e8e7 d1f3 g8f6 f7h5 c8e6 b1c3 e6f7 h5f7 e7f7 c1d2 d8c8 e1c1 c7c6 c3e2 b7b5 c1b1 c8e6 h2h3 a7a5 g2g4 a5a4 h1g1 a4a3 b2b3 a8d8 e2g3 c6c5 d2a5 d8d4 g3f5 d4d1 g1d1 e6c6 f3g3 f7g8 f2f3 c6e6 g4g5 h6g5 g3g5');
INSERT INTO evaluations VALUES(21334,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(21334,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(21334,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(21334,33,'cp 266',NULL);
INSERT INTO evaluations VALUES(21334,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(21334,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(21334,36,'cp 265','c4f7 e8e7 d1f3 g8f6 f7h5 c8e6 e1g1 e7d7 b1c3 a7a6 f1d1 f8d6 h5g6 d8e7 c3d5 e6d5 e4d5 d7c8 c1e3 c8b8 f3e2 h8f8 h2h3 a6a5 g6f5 a8a6 f5d3 a6a8 d3b5 d6c5 a1c1 a5a4 e3c5 e7c5 c2c4 e5e4 c1b1 a8a5 b2b4 a4b3');
INSERT INTO evaluations VALUES(21238,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(21238,31,'cp 270',NULL);
INSERT INTO evaluations VALUES(21238,32,'cp 269',NULL);
INSERT INTO evaluations VALUES(21238,33,'cp 269',NULL);
INSERT INTO evaluations VALUES(21238,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(21238,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(21238,36,'cp 276','e8e7 d1f3 g8f6 f7h5 c8e6 b1c3 e6f7 h5f7 e7f7 c1d2 d8c8 e1c1 c7c6 c3e2 b7b5 e2g3 c8e6 c1b1 g7g6 h1f1 f7g8 h2h4 h8h7 h4h5 g6g5 f3b3 e6b3 c2b3 h7d7 d2e3 d7f7 d1c1 a8c8 f2f3 c6c5 f1d1 g5g4 g3f5 f6h5 e3h6 g4f3 g2f3');
INSERT INTO evaluations VALUES(31231,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(31231,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(31231,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(31231,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(31231,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(31231,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(31231,36,'cp 55','g8f6 h2h3 g4h5 b1c3 h5g6 e2d3 e7e6 e1g1 f8e7 f1e1 d6d5 d1e2 e8g8 a2a3 d5e4 c3e4 f6e4 d3e4 g6e4 e2e4 b8d7 c2c4 d7f6 e4e5 d8d6 e5e2 d6c7 g2g3 a7a5 c1f4 e7d6 f4d2 c6c5 d4c5 d6c5 d2f4 c7e7 b2b3 h7h6');
INSERT INTO evaluations VALUES(31979,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(31979,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(31979,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(31979,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(31979,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(31979,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(31979,36,'cp -13','g8f6 e4f6 e7f6 d3d4 b8c6 c2c3 d8d7 g1f3 f8d6 f1d3 e8g8 e1g1 c6e7 c3c4 a8d8 f1e1 f8e8 d3f5 e7f5 e1e8 d8e8 c1d2 b7b5 c4c5 d6f8 d2c3 a7a6 a2a4 g7g5 a4b5 a6b5 h2h3 f5e7 h3h4 g5g4 f3e1 e7d5');
INSERT INTO evaluations VALUES(33341,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(33341,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(33341,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(33341,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(33341,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(33341,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(33341,36,'cp 25','e1c1 b8d7 g2g4 b7b5 g5f6 d7f6 g4g5 f6d7 f4f5 e8g8 h1g1 b5b4 c3e2 e6e5 f5f6 e5d4 f6e7 f8e8 e2d4 d7e5 f3f4 c7c5 g1g2 g7g6 h2h4 c8h3 g2g1 h3f1 g1f1 e8e7 h4h5 a8c8 h5h6 e5c4 f1f2 a6a5 b2b3 c4a3 f4f6 c5e5 f6e5 e7e5 d4f3 e5c5 d1d6 a3c2 c1d2 c2a3 d2d3 c5c1');
INSERT INTO evaluations VALUES(33340,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(33340,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(33340,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(33340,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(33340,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(33340,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(33340,36,'cp 35','b8d7 g2g4 b7b5 g5f6 d7f6 g4g5 f6d7 f4f5 e8g8 h1g1 b5b4 c3e2 e6e5 f5f6 e5d4 f6e7 f8e8 e4e5 a8a7 e5e6 f7e6 g5g6 h7g6 e2f4 d7e5 f3g3 c7e7 f4g6 e5g6 g3g6 e7f7 d1d4 f7g6 g1g6 a7f7 f1g2 f7f2 d4b4 e8e7 g2e4 g8f7 g6g2 f2g2 e4g2 e7c7');
INSERT INTO evaluations VALUES(436,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(436,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(436,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(436,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(436,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(436,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(436,36,'cp 32','g2g4 b7b5 g5f6 d7f6 g4g5 f6d7 f4f5 e8g8 h1g1 b5b4 c3e2 e6e5 f5f6 e5d4 f6e7 f8e8 e2d4 d7e5 f3f4 c7c5 g1g2 g7g6 h2h4 c8h3 g2f2 h3f1 f2f1 e8e7 h4h5 a8c8 h5h6 e5c4 f1f2 c5a5 f4f6 a5e5 f6e5 e7e5 d4f3 e5e4 b2b3 c4e5 d1d6 e5f3 f2f3 e4e2 d6d2 e2d2 c1d2');
INSERT INTO evaluations VALUES(437,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(437,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(437,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(437,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(437,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(437,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(437,36,'cp 30','b7b5 g5f6 d7f6 g4g5 f6d7 f4f5 e8g8 h1g1 b5b4 c3e2 e6e5 f5f6 e5d4 f6e7 f8e8 e2d4 d7e5 f3f4 g7g6 h2h4 d6d5 c1b1 d5e4 f1g2 c8b7 g2e4 b7e4 g1e1 e8e7 e1e4 a8c8 b2b3 c7c5 d1d2 c8e8 d2e2 a6a5 h4h5 e5d7 e4e7 e8e7 e2e7 c5e7 h5g6 h7g6 d4c6');
INSERT INTO evaluations VALUES(441,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(441,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(441,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(441,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(441,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(441,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(441,36,'cp 41','g5f6 d7f6 g4g5 f6d7 f4f5 e8g8 h1g1 b5b4 c3e2 e6e5 f5f6 e5d4 f6e7 f8e8 e2d4 d7e5 f3b3 a8b8 d4f5 c8e6 b3e3 a6a5 g1g2 a5a4 e3d4 b4b3 a2a3 e6f5 e4f5 c7e7 d4a4 b3c2 c1c2 e5c4 c2b1 c4e3 d1e1 e7d8 g2f2 e8e5 g5g6 h7g6 f5g6 f7g6 f1c4 e3c4 a4c4 d6d5');
INSERT INTO evaluations VALUES(440,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(440,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(440,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(440,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(440,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(440,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(440,36,'cp 20','d7f6 g4g5 f6d7 f4f5 e8g8 h1g1 b5b4 c3e2 e6e5 f5f6 e5d4 f6e7 f8e8 e2d4 d7e5 f3f4 g7g6 h2h4 d6d5 c1b1 d5e4 f1g2 c8b7 g2e4 b7e4 f4e4 c7e7 g1e1 a8d8 e4f4 d8d5 d1d2 e7d6 b2b3 e8d8 d4f5 g6f5 d2d5 d6d5 f4e5 d5e5 e1e5 f5f4 e5f5 d8d4 c2c4 b4c3 b1c2 f4f3 f5f3 d4h4');
INSERT INTO evaluations VALUES(38299,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(38299,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(38299,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(38299,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(38299,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(38299,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(38299,36,'cp 82','d1g4 g7g6 g1f3 h7h5 g4g3 h5h4 g3f4 b7b6 c2c3 c8a6 f1a6 b8a6 e1g1 g8h6 h2h3 h6f5 f1d1 a8c8 b1a3 d8d7 f4g4 c7c5 c1g5 e7g5 g4g5 c5d4 f3d4 d7e7 g5d2 f5d4 d2d4 e8g8 d4f4 a6c5 a3b5 c5e4 a1c1 c8c4');
INSERT INTO evaluations VALUES(44845,30,'cp 283',NULL);
INSERT INTO evaluations VALUES(44845,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(44845,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(44845,33,'cp 274',NULL);
INSERT INTO evaluations VALUES(44845,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(44845,35,'cp 275',NULL);
INSERT INTO evaluations VALUES(44845,36,'cp 272','c6b7 c8b7 g1f3 b8c6 f1b5 e7e6 e1g1 f8d6 d2d4 e8g8 c2c3 d8a5 b5a4 e6e5 b1d2 a5c7 d2c4 c6d4 f3e5 d4f3 e5f3 c7c4 d1d6 c4a4 f3d4 f8e8 c1e3 a8d8 d6g3 f6e4 g3f4 e4c5 f1d1 a4a6 h2h3 c5d3 f4g4 b7e4 b2b3 d8d6 g4e2 d6d8 c3c4 a6g6 f2f3');
INSERT INTO evaluations VALUES(43971,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(43971,31,'cp 282',NULL);
INSERT INTO evaluations VALUES(43971,32,'cp 268',NULL);
INSERT INTO evaluations VALUES(43971,33,'cp 285',NULL);
INSERT INTO evaluations VALUES(43971,34,'cp 270',NULL);
INSERT INTO evaluations VALUES(43971,35,'cp 273',NULL);
INSERT INTO evaluations VALUES(43971,36,'cp 258','c8b7 g1f3 b8c6 f1b5 e7e6 e1g1 f8d6 c2c3 e8g8 d2d4 d8b6 b5a4 a8d8 b1d2 d6b8 f1e1 b6c7 d1e2 f6d7 d2f1 h7h6 c1e3 c6e7 a1d1 e7g6 e3c1 f8e8 a4c2 e6e5 c2g6 f7g6 d4e5 d7e5 f3e5 d8d1 e2d1 e8e5 e1e5 c7e5 c1e3 e5c7 d1a4 c7c6');
INSERT INTO evaluations VALUES(28627,30,'cp 257',NULL);
INSERT INTO evaluations VALUES(28627,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(28627,32,'cp 259',NULL);
INSERT INTO evaluations VALUES(28627,33,'cp 261',NULL);
INSERT INTO evaluations VALUES(28627,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(28627,35,'cp 261',NULL);
INSERT INTO evaluations VALUES(28627,36,'cp 271','g1f3 b8c6 f1b5 e7e6 e1g1 f8d6 c2c3 e8g8 d2d4 d8a5 b5a4 a5c7 b1d2 e6e5 d2c4 c6d4 f3e5 d4f3 e5f3 c7c4 d1d6 c4a4 d6g3 f6e4 g3f4 a8e8 c1e3 a4a6 f1d1 e8e6 f3d4 e6f6 d4f5 a6e6 f5g7 g8g7 e3d4 g7g8 d4f6 e6f6 f4f6 e4f6 d1e1 f8b8 f2f3 f6d5 h2h3');
INSERT INTO evaluations VALUES(46503,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(46503,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(46503,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(46503,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(46503,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(46503,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(46503,36,'cp -97','d5d4 c3e4 f6e4 g2e4 d4e3 d2e3 d8d1 e1d1 b8a6 d1e1 a6c5 e4c2 c8d7 f2f3 a8d8 e1f2 d7c6 e3e4 c5d3 c2d3 d8d3 c1e3 f7f5 e4f5 f8e7 g1e2 e8g8 h1d1 d3d1 a1d1 f8f5 f3f4 e5f4 e3f4 f5f8 e2d4 e7c5 f2e1 f8e8 e1f1 c6d5 f1f2 c5b6');
INSERT INTO evaluations VALUES(47316,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(47316,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47316,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(47316,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(47316,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47316,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47316,36,'cp 38','h7h6 g5f4 b8c6 g1f3 e7e6 d1d2 a7a6 f1d3 c6b4 d3e2 f8d6 f3e5 b4d5 e2f3 e8g8 f4g3 b7b5 g3h4 f8e8 e1c1 d6b4 f3d5 e6d5 e5c6 b4c3 d2c3 d8d6 h4g3 f6e4 g3d6 e4c3 d1e1 e8e1 h1e1 c3a2 c1b1 a2c3 b2c3 c7d6 e1e8 g8h7 c6b4 a6a5 b4d5 c8b7');
INSERT INTO evaluations VALUES(53188,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53188,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(53188,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(53188,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(53188,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(53188,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(53188,36,'cp 11','f3e5 b8d7 e5c4 g7g6 b1c3 f8g7 g2g3 d7b6 c4e3 c8d7 f1g2 d8c7 d1e2 g8f6 a2a4 d7c6 g2c6 c7c6 e2b5 c6b5 a4b5 c5c4 c3a4 f6d5 a4b6 d5b6 e1e2 f7f5 h1d1 f5f4 g3f4 e8g8 c2c3 f8f4 d2d4 a8f8 d1f1 e7e5 d4e5 g7e5 a1a7');
INSERT INTO evaluations VALUES(53175,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(53175,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(53175,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(53175,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(53175,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(53175,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(53175,36,'cp -10','b8d7 f1b5 g8f6 d1f3 d8c7 e5d7 c8d7 b1c3 e7e6 f3g3 e8c8 d2d3 d7b5 g3c7 c8c7 c3b5 c7c6 a2a4 a7a6 b5a3 b7b6 a3c4 f8d6 c1d2 d6c7 h2h4 h7h5 e1g1 f6d5 f1e1 h8e8 g2g3 g7g6 b2b3 f7f6 e1e2 e6e5 a1e1 b6b5 a4b5 a6b5 c4e3 d5e3 d2e3');
INSERT INTO evaluations VALUES(1453,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(1453,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(1453,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(1453,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(1453,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(1453,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(1453,36,'cp 46','e4d5 e6d5 d1e2 f8e7 d4c5 b6c5 d2b3 c5d6 g2g3 g8f6 c1f4 d6d8 f1g2 e8g8 e1g1 c8f5 f3d4 f5g4 e2d2 d8d7 a1e1 f8e8 h2h4 h7h6 f2f3 g4f5 d4f5 d7f5 g1h2 f6h5 g2h3 f5f6 h3d7 e8d8 d7c6 b7c6 f4e5 f6g6 b3d4');
INSERT INTO evaluations VALUES(1447,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(1447,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(1447,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(1447,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(1447,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(1447,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(1447,36,'cp 41','e6d5 d1e2 f8e7 d4c5 b6c5 d2b3 c5d6 g2g3 g8f6 c1f4 d6d8 f1g2 e8g8 e1g1 f8e8 e2d3 c8g4 f1e1 d8c8 f4g5 g4f5 d3f1 h7h6 g5f6 e7f6 e1e8 c8e8 f1b5 e8d7 a1d1 c6e7 f3d4 a8c8 b5f1 f5g4 d1d2 f6g5 f2f4 g5f6 d4c2 g7g6 f1f2 g4e6 f2a7 e7f5');
INSERT INTO evaluations VALUES(1530,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(1530,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(1530,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(1530,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(1530,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(1530,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(1530,36,'cp 47','d1e2 f8e7 d4c5 b6c5 d2b3 c5d6 g2g3 g8f6 c1f4 d6d8 f1g2 e8g8 e1g1 f8e8 e2d3 a7a5 a2a4 h7h6 f1e1 b7b6 f3e5 c6e5 f4e5 c8d7 e5d4 a8c8 b3d2 d7e6 d2f1 e7c5 f1e3 f6e4 a1d1 d8d7 e3d5 e6d5 g2e4 d5e4 e1e4 d7a4');
INSERT INTO evaluations VALUES(2384,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(2384,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(2384,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(2384,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(2384,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(2384,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(2384,36,'cp 211','f3e5 g8f6 f2f3 f6d7 e5d7 c8d7 c1e3 d8c6 b1d2 f7f6 d2b3 e8c8 e1c1 c6e5 c4e2 b7b6 b3d4 f8d6 h2h4 h6h5 e2a6 c8b8 c2c3 h8e8 d1d2 e5c6 a6e2 a7a5 c1c2 c6d4 c3d4 a5a4 e2c4 e8e7 d2d1 b6b5 c4d3 a4a3');
INSERT INTO evaluations VALUES(2383,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(2383,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(2383,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(2383,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(2383,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(2383,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(2383,36,'cp 218','g8f6 f2f3 f6d7 e5d7 c8d7 c1e3 d8c6 b1d2 h6h5 e1c1 f7f6 d2b3 c6e5 c4e2 b7b6 b3d4 f8c5 c2c3 h5h4 h1e1 e8c8 b2b4 c5d6 d4f5 d7f5 e4f5 e5f7 e2c4 f7h6 c4e6 c8b7 h2h3 d6g3 d1d8 h8d8 e1e2 c7c6 a2a4 d8d3 c1c2');
INSERT INTO evaluations VALUES(4392,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(4392,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4392,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(4392,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(4392,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4392,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4392,36,'cp -7','e1g1 d7d6 h2h3 a7a5 c2c4 c6d4 b1c3 d4e6 d1e2 c7c6 c1e3 c5e3 e2e3 c6c5 e3d2 c8d7 b2b3 f6e8 a2a4 e8c7 a1e1 c7a6 c3d5 a6b4 g1h2 h7h6 d5c3 d7c6 c3d5 c6d5 e4d5');
INSERT INTO evaluations VALUES(4391,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(4391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4391,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(4391,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(4391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4391,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(4391,36,'cp 0','d7d6 h2h3 a7a5 c2c4 c6d4 f3d4 c5d4 b1c3 c8d7 a2a4 h7h6 c3e2 d4c5 g1h2 c7c6 e2c3 c5d4 c3e2');
INSERT INTO evaluations VALUES(3866,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3866,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3866,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(3866,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(3866,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(3866,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(3866,36,'cp 0','h2h3 a7a5 c2c4 c6d4 f3d4 c5d4 a2a4 c7c6 b1c3 c8d7 g1h2 f8e8 c3e2 d4c5 e2c3');
INSERT INTO evaluations VALUES(11691,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(11691,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(11691,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(11691,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(11691,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(11691,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(11691,36,'cp -73','d7d5 e4d5 c6d4 d2d3 c8g4 f2f3 g4h5 e1g1 f6d5 f1e1 f7f6 g5e4 f8e7 b1c3 d5b4 c1e3 d4b3 a2b3 e8g8 d1d2 f8e8 e4c5 h5g6 e3f2 e7f8 c3a2 b4c6 a2b4 c6b4 d2b4 d8d5 b4c3 a6a5 b3b4 a5a4 d3d4 e5d4 f2d4 e8e1 a1e1 g6f7 d4f2');
INSERT INTO evaluations VALUES(11613,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(11613,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(11613,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(11613,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(11613,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(11613,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(11613,36,'cp -65','e4d5 c6d4 d2d3 d4b3 a2b3 d8d5 e1g1 c8b7 d1f3 f8d6 f1e1 e8g8 g5e4 f6e4 f3e4 f7f5 e4d5 b7d5 c1d2 f8e8 b1c3 d5b7 f2f3 a8d8 h2h4 b5b4 c3d1 e5e4 f3e4 f5e4 d1f2 d6c5 d2e3 c5e3 e1e3 e4d3 e3e8 d8e8 f2d3 e8e2');
INSERT INTO evaluations VALUES(11609,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(11609,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(11609,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(11609,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(11609,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(11609,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(11609,36,'cp -75','c6d4 d2d3 d4b3 a2b3 d8d5 e1g1 c8b7 d1f3 f8d6 f1e1 e8g8 g5e4 f6e4 f3e4 f7f5 e4d5 b7d5 c1d2 h7h6 b3b4 f8e8 b1c3 d5b7 c3e2 f5f4 e2c1 f4f3 g2g3 e5e4 c1b3 e4d3 c2d3 b7d5 e1e8 a8e8 b3d4 d5b7 d2c3 d6f8 h2h4 c7c5 b4c5 f8c5 d4b3 c5b6');
INSERT INTO evaluations VALUES(12546,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(12546,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(12546,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12546,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(12546,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(12546,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(12546,36,'cp 44','g7g6 d2d3 f8g7 g1e2 e8g8 e1g1 d7d6 a1b1 a8b8 a2a3 a7a5 c3d5 c8d7 d5f6 g7f6 e2c3 f6g7 c1d2 d7f5 c3e4 f5e6 h2h4 b7b5 c4b5 b8b5 d1a4 b5b6 d2a5 c6a5 a4a5 b6b3 a5d2 d8d7 e4g5 f8b8');
INSERT INTO evaluations VALUES(18621,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(18621,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(18621,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(18621,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(18621,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(18621,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(18621,36,'cp 19','d3d4 c5d4 f3d4 g8f6 b1c3 f8g7 d4b5 d7d5 e4d5 e6d5 c4d5 e8g8 e1g1 a7a6 d5c6 b7c6 b5d4 f8e8 h2h3 d8c7 d4f3 c8b7 f1e1 c6c5 e1e8 a8e8 c1e3 b7c6 d1d3 c7b7 e3c5 c6f3 d3f3 b7b2 a1d1 b2c2 c5a7 f6e4 c3e4 c2e4 f3e4 e8e4 d1d8 g7f8 a7c5 e4e1 g1h2');
INSERT INTO evaluations VALUES(22322,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(22322,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(22322,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(22322,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(22322,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(22322,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(22322,36,'cp 170','d2d4 f8e7 d4e5 d6e5 d1e2 g8f6 e1g1 e7d6 f1d1 f5e4 c3e4 f6e4 e2e4 d8f6 c4d5 c8d7 f3g5 h7h6 g5f7 h8f8 f7d6 c7d6 c1e3 a8c8 d1d2 f6f5 a1d1 f8f6 h2h3 d7e6 e4f5 e6f5 g2g4 f5h7 h3h4 c6e7 g4g5 f6f8 d5b7');
INSERT INTO evaluations VALUES(23422,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(23422,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23422,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(23422,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(23422,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(23422,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23422,36,'cp 24','a2a3 g8f6 e2e3 d7d6 c1b2 g7g6 g1f3 f8g7 c2c4 e7e5 d2d3 c8f5 f1e2 e8g8 e1g1 a7a5 b4b5 c6b8 h2h3 b8d7 b1c3 f8e8 a1c1 e5e4 c3e4 f6e4 b2g7 g8g7 d3e4 f5e4 f3d2 d7c5 e2f3 e4f3 d2f3 d8f6 d1c2 e8e7 f1d1 a8e8 a3a4 e7d7 f3d4 g7g8');
INSERT INTO evaluations VALUES(26816,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(26816,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(26816,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(26816,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(26816,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(26816,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(26816,36,'cp 27','f8d6 f4g3 c6b4 f1b5 c7c6 b5a4 d6g3 h2g3 d8a5 a4b3 c6c5 a2a3 b4c6 d4c5 a5c5 c2c4 g8f6 c4d5 e6d5 b3a4 c5b6 d1b3 e8g8 b3b6 a7b6 a4b5 f8c8 e1e2 g7g6 h1c1 h7h5 d2b3 g8g7 b3d4 f5d7 e2f1 c6d4 f3d4 d7b5 d4b5 c8c6 f2f3 a8d8 f1e2 d8c8 b5c3 c6c4 c1d1');
INSERT INTO evaluations VALUES(28919,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(28919,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(28919,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(28919,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(28919,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(28919,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(28919,36,'cp 47','f3e5 g6h7 f1d3 h7d3 d1d3 b8d7 f2f4 e7e6 c1d2 f8e7 e1c1 d8c7 d3f3 e8g8 c2c4 f8e8 h1e1 e7f8 d2c3 b7b5 c4c5 f6d5 c3d2 a7a5 g3h5 g7g6 h5g3 h6h5 g3e4 f7f5 e4d6 d7e5 f4e5 e8b8 c1b1 c7d7 g2g3 a5a4 d1c1 a4a3 b2b3 f8e7 e1g1');
INSERT INTO evaluations VALUES(28916,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(28916,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(28916,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(28916,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(28916,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(28916,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(28916,36,'cp 49','g6h7 f1d3 h7d3 d1d3 e7e6 c1d2 b8d7 f2f4 d8c7 e1c1 f8e7 d3f3 e8g8 c1b1 a8d8 c2c4 f8e8 d2c3 e7f8 h1e1 c6c5 e5d7 f6d7 d4d5 e6d5 e1e8 d8e8 c4d5 d7b6 g3h5 c7d7 b1c1 f7f5 d5d6 b6c4 c3e5 c4e5 f4e5 e8e5 f3b3 d7f7 d6d7 f7b3');
INSERT INTO evaluations VALUES(28921,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(28921,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(28921,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(28921,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(28921,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(28921,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(28921,36,'cp 35','f1d3 h7d3 d1d3 e7e6 c1d2 f8e7 e1c1 b8d7 f2f4 e8g8 d3f3 d8c7 c2c4 f8e8 d2c3 b7b5 c4c5 f6d5 c3d2 f7f6 e5d7 c7d7 f4f5 e7f8 h1e1 d7f7 f5e6 e8e6 g3f5 e6e1 d1e1 g7g6 f5e3 d5e3 f3e3 a8d8 e3e4 d8d5 d2c3 f6f5 e4e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(42132,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42132,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42132,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(42132,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42132,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(42132,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(42132,36,'cp 15','e2e4 c7c5 f1c4 c5d4 f3d4 b4c3 b2c3 d8a5 d4b5 f6e4 d1d4 e8g8 d4e4 a7a6 c4d3 f7f5 e4e2 a6b5 e1g1 b8c6 g5d2 e6e5 d3b5 c8e6 f2f3 a5c7 f1e1 e5e4 a2a4 c6a5 f3e4 f5f4 e1f1 a5b3 a1d1 b3d2 d1d2 c7c3 e2d3 c3c5 d3d4 c5g5 d4b6 e6g4 a4a5');
INSERT INTO evaluations VALUES(42133,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(42133,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42133,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42133,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42133,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(42133,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(42133,36,'cp 12','c7c5 f1c4 c5d4 f3d4 b4c3 b2c3 d8a5 d4b5 f6e4 d1d4 e8g8 d4e4 a7a6 g5d2 a6b5 c4d3 f7f5 e4e2 b8c6 e1g1 e6e5 d3b5 c8e6 a2a4 a5c7 b5c6 c7c6 e2e5 f8e8 e5g3 a8a4 a1a4 c6a4 h2h3 a4d7 d2h6 d7f7 h6e3 f7g6 g3f4 e6d5 f2f3 h7h6 c3c4 d5c6 f1b1 e8e6 e3f2 g6g5 f4g5');
INSERT INTO evaluations VALUES(42639,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(42639,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(42639,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42639,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42639,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(42639,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(42639,36,'cp -6','d7d6 c2c3 e8g8 e1g1 a7a5 d3d4 c5a7 d4e5 d6e5 f3e5 d8d1 f1d1 f6e4 e5d3 f8e8 g1f1 b8a6 b1d2 e4d2 c1d2 c8f5 d2f4 c7c6 f4g3 a6c5 d3c5 a7c5 a2a4 c5b6 e2c4 h7h6 d1d2 a8d8 a1d1 d8d2 d1d2 e8d8 g3f4 g7g5 d2d8 b6d8 f4e3');
INSERT INTO evaluations VALUES(41896,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(41896,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(41896,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(41896,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(41896,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(41896,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(41896,36,'cp -8','c2c3 e8g8 e1g1 a7a5 d3d4 c5a7 d4e5 d6e5 f3e5 d8d1 f1d1 f6e4 e5d3 b8d7 b1d2 e4d2 c1d2 c7c6 e2f3 f8e8 d2f4 d7f6 f4d6 f6e4 f3e4 e8e4 d3c5 a7c5 d6c5 c8e6 d1d2 e6a2 f2f3 e4a4 c5b6 a8e8 h2h3 f7f6 d2d7 a2e6 a1a4 e6d7');
INSERT INTO evaluations VALUES(42853,30,'cp 273',NULL);
INSERT INTO evaluations VALUES(42853,31,'cp 287',NULL);
INSERT INTO evaluations VALUES(42853,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(42853,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(42853,34,'cp 276',NULL);
INSERT INTO evaluations VALUES(42853,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(42853,36,'cp 263','f3e5 d8g5 d2d4 g5g2 d1f3 g2f3 e5f3 c5f8 b1c3 d7d6 c3d5 c8g4 f3g5 e8d8 h2h3 g4d7 d5f4 d8e8 c1e3 f8e7 h1g1 g7g6 e1c1 b8c6 g5f3 h6g8 a2a3 c6d8 e4e5 d8e6 h3h4 e6f4 e3f4 h7h6 d4d5 e8f8 e5e6 d7e8 d1e1 f7e6 e1e6');
INSERT INTO evaluations VALUES(42849,30,'cp 280',NULL);
INSERT INTO evaluations VALUES(42849,31,'cp 272',NULL);
INSERT INTO evaluations VALUES(42849,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(42849,33,'cp 291',NULL);
INSERT INTO evaluations VALUES(42849,34,'cp 272',NULL);
INSERT INTO evaluations VALUES(42849,35,'cp 280',NULL);
INSERT INTO evaluations VALUES(42849,36,'cp 271','d8g5 d2d4 g5g2 d1f3 g2f3 e5f3 c5f8 b1c3 d7d6 c3d5 c8g4 f3g5 e8d8 f2f3 g4d7 d5f4 d8e8 h2h4 b8c6 c1e3 c6d8 e1c1 a8c8 h4h5 f8e7 g5h3 c7c6 d1g1 e7f6 c2c3 d8e6 c4b3 c6c5 d4c5 e6c5 f4d5 f6e5 f3f4 e5f6 d5f6 g7f6 f4f5 c5b3 a2b3 h6g8');
INSERT INTO evaluations VALUES(43214,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(43214,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43214,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43214,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43214,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43214,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43214,36,'cp 0','c7c5 e2e3 e8g8 c2c3 c5d4 e3d4 b8c6 f1e2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(44904,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(44904,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(44904,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(44904,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(44904,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(44904,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(44904,36,'cp -41','d5c4 a2a4 c8f5 e2e3 b8a6 f1c4 a6b4 c4b3 e7e6 g1f3 b4d3 e1f1 d3b4 g2g4 f5d3 f1g2 f8d6 a4a5 h7h6 a1a4 d3h7 d1e2 b4d5 b3c2 d5c3 b2c3 f6e4 e2c4 a8b8 c4b3 e8g8 c1a3 g8h8 a3d6 d8d6 h1b1 f7f5 c2e4 f5g4 f3d2 g4h3 g2g1 d6d8 f2f4 h7e4 d2e4 d8h4 b3c2');
INSERT INTO evaluations VALUES(47088,30,'cp -155',NULL);
INSERT INTO evaluations VALUES(47088,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(47088,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(47088,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(47088,34,'cp -159',NULL);
INSERT INTO evaluations VALUES(47088,35,'cp -130',NULL);
INSERT INTO evaluations VALUES(47088,36,'cp -169','c7c6 b5e2 d8a5 a2a3 c8f5 c1d2 a5c7 g2g4 f5g6 d2e3 h7h5 g4h5 f6h5 g1h3 b8d7 d1d2 e7e6 e1c1 e8c8 h3f4 h5f4 e3f4 f8d6 f4d6 c7d6 h2h4 d7f6 d2e3 g6f5 e3g5 d8g8 h4h5 f6d5 c3d5 c6d5 g5d2 c8b8 d1g1 d6c7 f2f3 e4f3 e2f3');
INSERT INTO evaluations VALUES(44853,30,'cp -166',NULL);
INSERT INTO evaluations VALUES(44853,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(44853,32,'cp -152',NULL);
INSERT INTO evaluations VALUES(44853,33,'cp -139',NULL);
INSERT INTO evaluations VALUES(44853,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(44853,35,'cp -148',NULL);
INSERT INTO evaluations VALUES(44853,36,'cp -153','b5e2 d8a5 c1e3 c8f5 g2g4 f5g6 a2a3 f6d5 e3d2 d5c3 d2c3 a5c7 h2h4 h7h6 d1d2 b8d7 g1h3 d7b6 d2f4 c7d7 f4d2 g6h7 e1c1 b6d5 h3f4 e7e6 f4d5 e6d5 c3b4 f8b4 a3b4 d7c7 d2e3 e8c8 h1f1 f7f5 g4f5 h7f5');
INSERT INTO evaluations VALUES(49189,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(49189,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(49189,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(49189,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(49189,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(49189,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(49189,36,'cp 78','d7d5 c1f4 f8g7 e2e3 e8g8 a1c1 a7a6 b2b4 d8e7 d1b3 c7c6 c4c5 f6e8 f1e2 b8d7 e3e4 d5e4 c3e4 e6e5 f4g5 e8f6 e4d6 e5e4 f3d2 h7h6 g5h4 d7b6 h4f6 g7f6 d2e4 b6d5 e1g1 f6d4 e2c4 c8e6 f1e1 d5f4 c4e6 f4e6 e4g3');
INSERT INTO evaluations VALUES(51465,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(51465,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(51465,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(51465,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(51465,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(51465,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(51465,36,'cp 101','e2e4 f8g7 h2h4 h7h5 b1c3 g8f6 c1g5 d7d6 f1e2 a7a6 g1f3 e8g8 d1d2 c7c5 g5h6 g7h6 d2h6 f6g4 h6d2 e7e5 d4d5 d8f6 a2a4 b8d7 c3d1 f6e7 f3g5 d7f6 f2f3 g4h6 a4a5 b6a5 d1f2 f6h7 g5h7 g8h7 d2a5');
INSERT INTO evaluations VALUES(51466,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(51466,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(51466,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(51466,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(51466,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(51466,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(51466,36,'cp 101','f8g7 h2h4 h7h5 b1c3 d7d6 g1f3 g8f6 f1e2 e8g8 d1c2 b8c6 d4d5 c6b4 c2d1 c8g4 c1g5 g4f3 e2f3 a7a5 f3e2 b4a6 d1c2 a6c5 f2f3 a5a4 c2d2 e7e6 e1g1 d8d7 g2g3 f6h7 g5e3 g7c3 d2c3 d7e7');
INSERT INTO evaluations VALUES(53811,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(53811,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(53811,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(53811,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(53811,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(53811,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(53811,36,'cp 36','c8g4 d2d4 b8d7 d4e5 d6e5 a2a4 g8f6 b1d2 g4h5 h2h3 f8c5 d1e1 e8g8 f3h4 a7a5 g1h1 c5b4 c2c3 b4d6 f2f3 h5g6 d2b3 f6h5 c1e3 d7b6 c4e2 d6e7 h4g6 h7g6 g2g3 f7f5 a1d1 d8c7 b3c5 f5e4 c5e6 c7c8 h1h2 c8e6 e3b6 g8h8 f3e4');
INSERT INTO evaluations VALUES(31596,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(31596,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(31596,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(31596,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(31596,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(31596,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(31596,36,'cp 69','d2d4 b8d7 d1d3 g4h5 d4e5 d6e5 a2a4 f8c5 b1d2 g8f6 h2h3 e8g8 f3h4 a7a5 g2g4 h5g6 h4g6 h7g6 d2f3 d8e7 g1g2 a8d8 c1g5 d7b6 d3e2 b6c4 e2c4 c5d4 c2c3 d4c5 a1d1 f8e8 d1d8 e8d8 c4e2 d8d7 h3h4 c5b6 f1d1 d7d1');
INSERT INTO evaluations VALUES(53877,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53877,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(53877,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(53877,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(53877,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(53877,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(53877,36,'cp 27','f8e7 d2d3 b8c6 b1c3 g8f6 c1g5 e8g8 g5f6 e7f6 a2a4 g7g6 c3d5 f6g7 c2c3 g8h8 d1d2 c8e6 f1e1 a8b8 d2g5 f7f6 g5h4 b7b6 h2h3 c6e7 d5e7 d8e7 f3d2 e7d7 c4e6 d7e6 d2c4 f6f5 e4f5 g6f5');
INSERT INTO evaluations VALUES(57056,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(57056,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(57056,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(57056,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(57056,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(57056,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(57056,36,'cp -62','h7h5 g4h5 d7d5 d2d4 c7c5 c2c4 d5c4 d4d5 e6d5 g2d5 b8c6 d5c6 b7c6 d1d8 e8d8 b1d2 c8e6 g1f3 h8h5 h2h4 d8d7 d2f1 e6d5 c1f4 h5f5 f4g5 f7f6 g5e3 g8e7 e1c1 d7c7 f1g3 d5f3 g3f5 f3e2 d1d2 e2f3 f5e7 f3h1 e3f4 c7b6 d2d7 a8e8 e7g6');
INSERT INTO evaluations VALUES(5799,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(5799,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(5799,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(5799,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(5799,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(5799,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(5799,36,'cp 34','f1d3 g8f6 c1d2 e8g8 a2a3 b4d6 e3e4 e6e5 c3d5 f6d5 e4d5 c6d4 f3d4 e5d4 e1g1 c8d7 d1h5 f7f5 d2g5 d8e8 h5h4 h7h6 g5d2 e8e5 a1e1 e5f6 h4f6 f8f6 h2h4 g8f8 h4h5 a8e8 g2g3 e8e1 f1e1 f6f7 d3c4 b7b6 b2b4');
INSERT INTO evaluations VALUES(7382,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(7382,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(7382,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(7382,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(7382,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(7382,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(7382,36,'cp 42','e7g6 d2d4 c5a7 a4c6 d7c6 f3e5 g6e5 d4e5 e8g8 d1e2 d8h4 b1d2 c8g4 d2f3 h4h5 h2h3 g4f3 e2f3 h5e5 c1f4 e5e7 f1e1 a8d8 f3g4 a6a5 g2g3 a5a4 g1g2 a7c5 h3h4 e7d7 g4d7 d8d7 a1d1 f8d8 d1d7 d8d7 h4h5 f7f6 e4e5 f6e5 e1e5 c5d6 e5e4 b7b5 f4d6');
INSERT INTO evaluations VALUES(7737,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(7737,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(7737,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(7737,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(7737,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(7737,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(7737,36,'cp 58','f1d3 e8g8 e1g1 d5c4 d3c4 c6c5 d1b3 c5d4 b3b4 a7a5 b4b3 d4c3 b3c3 b7b6 f1d1 h7h6 g5h4 d8c7 a1c1 c8a6 c4a6 c7c3 c1c3 a8a6 c3c7 g7g5 h4g3 d7c5 c7c6 f6d5 g3d6 f8a8 d6c5 b6c5 c6c5 a6b6 d1d2 a8b8 b2b3 a5a4 f3d4 a4b3 a2b3 g5g4 d2d1 d5e7');
INSERT INTO evaluations VALUES(8964,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(8964,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(8964,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(8964,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(8964,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(8964,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(8964,36,'cp 63','a1c1 f6e4 f1d3 e4c3 b2c3 d5c4 d3c4 b4a3 c1b1 a3d6 g1h3 e8g8 e1g1 e6e5 f4g3 g7g6 a2a4 g8g7 c4d5 e5d4 d5c6 b7c6 c3d4 c8h3 g2h3 d8d7 b1c1 a8b8 d1f3 b8b4 c1c6 b4a4 f1c1 a7a5 g3e5 d6e5 d4e5 d7f5 f3f5 g6f5 c1c5 f5f4 e3f4 a4f4 c6c7');
INSERT INTO evaluations VALUES(10230,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(10230,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(10230,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(10230,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(10230,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(10230,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(10230,36,'cp -88','d7d5 e4d5 d8d5 d3e2 a7a5 d1b3 e8g8 b3d5 f6d5 f3e5 d5f4 e2d1 c6e5 d2d4 f4g2 g1g2 e5d3 d1e2 d3c1 f1c1 c5d6 c1e1 c8e6 e2f3 c7c6 b1d2 f8e8 d2e4 d6e7 a2a4 g7g6 e4d2 e7d6 a1d1 a8d8 d2e4 d6e7 e4c5 e7c5');
INSERT INTO evaluations VALUES(10354,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(10354,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(10354,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(10354,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(10354,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(10354,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(10354,36,'cp -39','e8g8 f1d3 d7d6 e1g1 a7a6 h2h3 b4a5 f1e1 c6e7 c3e2 e7g6 e2g3 c7c5 b2c3 b7b5 a2a4 b5b4 c3b2 c8e6 d3c4 e6c4 b3c4 g6f4 d2d3 h7h6 g3f5 f4e6 a1b1 g8h7 f5e3 a8b8 g2g3 d8d7 e1f1 a5d8 c2c3 b4c3 b2c3');
INSERT INTO evaluations VALUES(4428,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(4428,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(4428,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(4428,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(4428,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(4428,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(4428,36,'cp -24','f1d3 d7d6 e1g1 a7a6 h2h3 b4a5 f1e1 c6e7 c3e2 e7g6 e2g3 c7c5 b2c3 c8e6 a2a4 h7h6 d3f1 f8e8 f1c4 a5c3 d2c3 d8d7 c4d5 e6d5 e4d5 g6f4 c3c4 g7g6 f3d2 g8g7 d1f3 h6h5 a4a5 a8d8 g3e4');
INSERT INTO evaluations VALUES(15170,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(15170,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(15170,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(15170,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(15170,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(15170,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(15170,36,'cp -81','f1g2 c8f5 d2d3 h7h5 h2h3 d8d7 g1f3 f8d6 b1d2 e8c8 f3g5 c8b8 a2a3 d7e7 e3e4 d5e4 g5e4 f5e4 d2e4 f6e4 g2e4 c6d4 e1g1 h5h4 g3g4 d4e6 b3b4 e6g5 e4g2 e5e4 f1e1 h8e8 d1e2 e4d3 e2e7 d6e7 c2d3 d8d3');
INSERT INTO evaluations VALUES(16981,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(16981,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(16981,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(16981,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(16981,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(16981,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(16981,36,'cp 188','d2d4 d7d6 f3b3 d8d7 b1d2 c7c6 f1e1 d7c7 a2a4 c8g4 f2f4 g4h5 h2h3 e5f4 e4e5 d6e5 e1e5 e8d8 d2f3 c7b6 b3d3 h5g6 d3e2 a7a5 c1f4 f8d6 a1e1 d6e5 f3e5 h8e8 c2c3 f6h5 e2g4 h5f6 g4h4 b6b2 c4f7 g6f7 e5f7 d8d7 f7e5 d7c8 e5c4 e8e1 h4e1');
INSERT INTO evaluations VALUES(17354,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(17354,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(17354,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(17354,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(17354,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(17354,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17354,36,'cp 12','d7d5 e4d5 e6d5 c4b3 f8d6 d2d4 c5c4 f1e1 g8e7 b3c2 d8c7 b2b3 c4b3 a2b3 c8g4 h2h3 g4f3 d1f3 c6d4 c3d4 c7c2 c1g5 f7f6 g5f4 d6b4 f4d2 b4d2 b1d2 c2d2 e1e2 d2d4 a1e1 e8g8 e2e7 d4c5 e7b7 a8b8 b7e7 f8f7 e7f7 g8f7 f3h5 f7g8 e1e8 b8e8 h5e8 c5f8 e8e6 g8h8 e6d5 f8g8 d5c6 g8b3 c6c8 b3g8');
INSERT INTO evaluations VALUES(19094,30,'cp 601',NULL);
INSERT INTO evaluations VALUES(19094,31,'cp 603',NULL);
INSERT INTO evaluations VALUES(19094,32,'cp 604',NULL);
INSERT INTO evaluations VALUES(19094,33,'cp 604',NULL);
INSERT INTO evaluations VALUES(19094,34,'cp 611',NULL);
INSERT INTO evaluations VALUES(19094,35,'cp 609',NULL);
INSERT INTO evaluations VALUES(19094,36,'cp 607','b5a4 b7b5 a4c2 d8c7 g1f3 a7a5 e1g1 f8c5 f1e1 g8f6 c1g5 f6g4 g5h4 f7f6 h2h3 g4h6 g1h1 e8g8 b1d2 c6d5 h4g3 d7d6 d2b3 c5b6 d1d5 h6f7 d5b5 a5a4 b3d2 a8a5 b5c4 c8a6 c4c7 b6c7 b2b3 a4b3 c2b3 g8h8');
INSERT INTO evaluations VALUES(19341,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(19341,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(19341,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(19341,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(19341,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(19341,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(19341,36,'cp 44','c5d4 e1g1 e7e6 f3d4 g8f6 d4c6 d7c6 d1d8 e8d8 f1d1 f6d7 a2a4 a7a5 b2b3 d8e8 b1d2 f8c5 c4e2 e6e5 d2c4 e8e7 e2g4 f7f6 g1f1 d7b6 g4c8 b6c8 f1e2 h6h5 h2h3 h8d8 d1d8 e7d8 c1d2 b7b6 g2g4 h5g4 h3g4 d8c7 g4g5 c8d6 c4d6 c7d6');
INSERT INTO evaluations VALUES(19343,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(19343,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(19343,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(19343,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(19343,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(19343,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(19343,36,'cp 54','f3d4 g8f6 b1c3 e7e6 c1f4 f8c5 d4b5 e8g8 f4c7 d8e7 c7d6 c5d6 d1d6 a7a6 d6e7 c6e7 b5d6 b7b5 c4e2 b5b4 c3a4 a6a5 f2f3 a8a7 e1c1 c8a6 e2f1 a6f1 h1f1 e7g6 b2b3 g6e5 c1b2 a7a6 a2a3 b4a3 b2a3');
INSERT INTO evaluations VALUES(19342,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(19342,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(19342,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(19342,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(19342,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(19342,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(19342,36,'cp 54','e7e6 b1c3 d8c7 c1e3 g8f6 d4b5 c7b8 f2f4 a7a6 b5d4 b7b5 d4c6 d7c6 e4e5 f6d5 c3d5 c6d5 c4d3 b8c7 d1e2 b5b4 e1g1 f8c5 e3c5 c7c5 e2f2 c5a5 g1h1 a5c7 a1e1 c7e7 e1e3 c8d7 h2h3 e8g8 h1h2 e7c5 f2e2 d7b5 e3g3 b5d3 c2d3 a8c8 e2g4');
INSERT INTO evaluations VALUES(19648,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(19648,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19648,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19648,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19648,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19648,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19648,36,'cp 0','f8g7 c1e3 g8f6 h2h3 a7a6 a2a4 b7b6 e1g1 c8b7 d1d2 e7e6 f1e1 e8g8 c4b3 d8c7 a1d1 a8d8 e3h6 d6d5 e4d5 e6d5 h6f4 c7c8 f4h6');
INSERT INTO evaluations VALUES(21700,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(21700,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(21700,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(21700,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(21700,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(21700,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(21700,36,'cp -74','d5e4 f3e5 c6e5 d4e5 d8d1 e1d1 g8e7 c1e3 c8f5 b1a3 a7a6 d1c2 e8c8 a3c4 d8e8 a2a4 e7c6 b2b4 c6e5 c4e5 e8e5 f1c4 f7f6 h1d1 h7h5 h2h4 f8d6 d1d4 e5e7 c4d5 h8e8 g2g3 f5g4 d4d2 e7d7 a4a5 d7d8 c2b3 d6e5 d5e4 d8d2 e3d2 c7c6');
INSERT INTO evaluations VALUES(21800,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(21800,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(21800,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(21800,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(21800,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(21800,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(21800,36,'cp -51','f3e5 c6e5 d4e5 d8d1 e1d1 c8f5 c1e3 e8c8 d1c1 a7a5 f1c4 f7f6 e5f6 g8f6 h2h3 f6d7 g2g4 f5g6 c4e6 d8e8 e6d7 c8d7 b1d2 a5a4 c1c2 d7c8 a1d1 e8e6 d2c4 g6e8 d1d5 b7b6 e3g5 f8e7 g5e7 e6e7 c4e3');
INSERT INTO evaluations VALUES(21771,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(21771,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(21771,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(21771,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(21771,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(21771,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(21771,36,'cp -64','c6e5 d4e5 d8d1 e1d1 g8e7 c1e3 c8f5 b1a3 e8c8 d1c2 a7a6 a3c4 d8e8 b2b4 e7g6 a2a4 g6e5 c4e5 e8e5 f1c4 f7f6 h1d1 h7h5 h2h4 f8d6 d1d4 a6a5 c4b3 g7g5 a1h1 c8b8 b4a5 e5a5 d4d5 a5d5 b3d5 c7c6 d5f7 g5g4 g2g3 d6e5 h1b1 b8c8 a4a5');
INSERT INTO evaluations VALUES(22191,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(22191,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(22191,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(22191,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(22191,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(22191,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(22191,36,'cp -67','d4e5 d8d1 e1d1 c8f5 c1e3 g8e7 b1a3 a7a6 d1c2 e8c8 a3c4 e7c6 a2a4 d8e8 a1d1 c6e5 d1d5 e5c4 d5f5 c4e3 f2e3 f7f6 g2g4 f8d6 f1g2 e8e7 h2h4 h8e8 g4g5 f6g5 f5g5 c8b8 h1d1 d6e5 g2h3 e5f6 g5g4 c7c6 g4f4 a6a5 d1d7 e7d7 h3d7');
INSERT INTO evaluations VALUES(22188,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(22188,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(22188,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(22188,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(22188,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(22188,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(22188,36,'cp -75','d8d1 e1d1 c8f5 c1e3 g8e7 b1a3 a7a6 d1c2 e8c8 a3c4 e7c6 b2b4 d8e8 e5e6 e8e6 a2a4 c6e5 c4e5 e6e5 f1c4 f7f6 h1d1 h7h5 h2h4 f8d6 d1d4 e5e7 b4b5 a6a5 b5b6 h8e8 c4e2 g7g6 a1b1 f5d7 e2b5 d7b5 b1b5 e7e5 b6c7 d6c7 g2g3 e5b5');
INSERT INTO evaluations VALUES(1770,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(1770,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(1770,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(1770,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(1770,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(1770,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(1770,36,'cp -68','e1d1 g8e7 c1e3 c8f5 b1a3 e8c8 d1c2 e7c6 a3b5 f5g6 f1e2 f8e7 a1d1 c8b8 h2h4 h7h5 e5e6 f7e6 g2g3 e7f6 d1d8 h8d8 h1d1 d8f8 a2a4 a7a6 b5d4 c6d4 c3d4 f8d8 a4a5 b8c8 b2b4 d8d6 d1d2 c8d7 d2d1 g6f7 b4b5 a6b5 e2b5 d7c8 d1d2 f7g6 c2c3 g6f5');
INSERT INTO evaluations VALUES(1769,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(1769,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(1769,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(1769,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(1769,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(1769,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(1769,36,'cp -77','c8f5 c1e3 g8e7 b1a3 e8c8 d1c2 e7c6 a3b5 f5g6 f1e2 f8e7 a1d1 c8b8 e5e6 f7e6 d1d8 h8d8 h1d1 d8d1 c2d1 e7f6 e2c4 e6e5 c4e6 b7b6 g2g3 c6d8 e6d7 c7c6 b5a3 b8c7 d7h3 d8f7 a3c4 g6h5 d1d2 h5f3 d2e1 f7d6 c4d6 c7d6 e1d2');
INSERT INTO evaluations VALUES(22120,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(22120,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(22120,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(22120,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(22120,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(22120,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(22120,36,'cp 151','f3d4 e5d4 c2c3 g8f6 e1g1 d6d5 e4d5 f6d5 c3d4 c7c6 b1c3 f8d6 d1b3 e8g8 c4d5 c6d5 b3d5 d8b6 c1e3 f8d8 d5b5 b6b5 c3b5 c8f5 f1d1 d6e7 a1c1 a7a6 b5c3 e7f6 c3e2 f5e6 b2b3 a6a5 e2f4 f6d4 f4e6 f7e6 c1c4 d4e3 f2e3 b7b5 c4c7 a5a4');
INSERT INTO evaluations VALUES(22312,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(22312,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(22312,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(22312,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(22312,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(22312,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(22312,36,'cp 166','f1c4 g7g6 b2b4 a7a6 a2a4 g8h6 e1g1 h6f7 c4a2 d8e7 b4b5 c6d8 b5b6 c8e6 b6c7 e7c7 c3c4 f8h6 c1b2 e8g8 b1c3 a8c8 a2b3 c7d7 a4a5 e6c4 b3c4 c8c4 d1b3 d7e6 f1b1 g8h8 c3d5 c4c6 b2c3 c6c8 b1b2');
INSERT INTO evaluations VALUES(22307,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(22307,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(22307,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(22307,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(22307,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(22307,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(22307,36,'cp 159','g7g6 e1g1 d8e7 b2b4 f8h6 b1d2 h6g7 b4b5 c6d8 c4d5 g8h6 c1a3 h6f7 c3c4 c8e6 c4c5 e8g8 c5d6 f7d6 d4e5 f6e5 d2c4 e6d5 c4d6 c7d6 d1d5 e7e6 d5d6 e6d6 a3d6 f8e8 f1d1 a8c8 a1c1 c8c1 d1c1 g7h6 c1c3 d8f7');
INSERT INTO evaluations VALUES(23858,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(23858,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(23858,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(23858,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(23858,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(23858,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(23858,36,'cp 49','a2a3 g8e7 g2g4 h5g6 c1e3 d8d7 d1d2 e8c8 e1c1 h7h5 h1g1 f7f6 d3d4 e5d4 e3d4 h5g4 h3g4 c5d4 f3d4 c6d4 d2d4 c8b8 c4b5 c7c6 b5c4 e7c8 c1b1 c8b6 f2f4 b6c4 d4c4 g6f7 c4d4 h8h4 d4f2 h4g4 g1g4 d7g4 d1g1 g4h5 g1g7');
INSERT INTO evaluations VALUES(28564,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(28564,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(28564,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(28564,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(28564,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(28564,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(28564,36,'cp 47','c2c4 e7e6 d1b3 b8c6 b1d2 f8d6 f4d6 d8d6 a1c1 e8g8 f1e2 c6a5 b3c3 a5c4 d2c4 d5c4 c3c4 d6b6 c4c3 f5g4 h2h3 g4f3 e2f3 c7c6 e1g1 f8d8 a2a3 g7g6 c1c2 a6a5 f1b1 d8d7 g2g3 h7h5 h3h4 g8g7 c2c1 f6d5 c3c2 b6c7 g1g2');
INSERT INTO evaluations VALUES(30275,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(30275,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(30275,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(30275,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(30275,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30275,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(30275,36,'cp -2','c7c5 a2a3 b8c6 b2b4 b7b6 d2d3 f8e7 b4b5 c6d4 c1g5 e8g8 a3a4 f6d5 f3d4 c5d4 c3d5 e6d5 g5e7 d8e7 e4d5 a8c8 a4a5 c8c5 a5b6 a7b6 d1g4 c5b5 g4e6 e7e6 d5e6 f8e8 f2f4 e5f4 e6e7 b5e5 f1f4 h7h6 h2h3 e5e7 f4d4 e7e2 d4d6 e2c2 d6b6');
INSERT INTO evaluations VALUES(32383,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(32383,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(32383,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(32383,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(32383,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(32383,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(32383,36,'cp 26','d2d4 g8f6 f1b5 b8d7 e1g1 a7a6 b5d7 d8d7 c2c4 g4f3 d1f3 g7g5 f1e1 f8e7 b1d2 e8c8 f3d3 h8e8 d2f3 e7b4 f3e5 f6g4 c1d2 e8e5 e1e5 g4e5 d4e5 b4d2 d3d2 d7f5 d2d4 g5g4 a1d1 h7h5 d5d6 f5c2 d1d2 c2f5 b2b4 f7f6 d4c5 d8d7 c5a7 c7d6 e5d6 f5b1 g1f2');
INSERT INTO evaluations VALUES(38389,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(38389,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(38389,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(38389,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(38389,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(38389,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(38389,36,'cp -12','b8d7 f4g5 g8f6 g1f3 h7h6 g5h4 e7e5 f1e2 e8g8 e1g1 d8e7 d1c2 g6g5 h4g3 f6h5 d4e5 h5g3 h2g3 d7e5 f3e5 d6e5 e3e4 g5g4 b1d2 h6h5 d2c4 c8e6 a1d1 a8d8 c4e3 g7h6 e3f5 e6f5 e4f5 h6g5 e2d3 e7c5 f5f6 g5f6');
INSERT INTO evaluations VALUES(41038,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(41038,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(41038,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(41038,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(41038,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(41038,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(41038,36,'cp 109','c7c5 e1g1 c5d4 f3d4 b8c6 d4c6 b7c6 d1d8 e7d8 f1d1 d8a5 f2f3 e8e7 c1f4 h8d8 g1f1 d8d1 a1d1 f6d7 f4d6 e7d8 g2g3 a5c7 d6a3 d8e8 c4e2 c7d8 f3f4 a7a5 c3a4 d8e7 a3d6 c8a6 e2a6 a8a6 f1e2 d7b6 a4c5 a6a8 h2h3 a5a4 d6e7');
INSERT INTO evaluations VALUES(46399,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46399,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(46399,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(46399,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46399,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(46399,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46399,36,'cp 7','f8c5 c1g5 e8g8 g5f6 d8f6 d1d5 c5b4 d5d4 f6e7 d4e3 b4c3 b2c3 e7e3 f2e3 b8d7 f1c4 d7f6 c4d3 c8e6 e3e4 e6d7 e1c1 d7c6 h1e1 a8e8 e4e5 f6d5 d3c4 d5f4 c4f1 f4g6 e5e6 e8e6 e1e6 c6f3 e6e1 f3d1 e1d1 g6e5 d1d5');
INSERT INTO evaluations VALUES(47033,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47033,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47033,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(47033,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47033,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47033,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(47033,36,'cp 0','c7c5 g1f3 b8c6 d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 f1c4 f8e7 e1g1 e8g8 c4b3 d8c7 d1e2 h7h6 c1e3 a7a5 f1e1 c8a6 e2f3 a8b8 a1d1 b8b4 e3d2 b4d4 d2c1 d4b4');
INSERT INTO evaluations VALUES(45192,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(45192,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(45192,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(45192,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(45192,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(45192,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45192,36,'cp 11','g1f3 b8c6 d2d4 c5d4 f3d4 e7e5 d4c6 b7c6 f1d3 f8e7 d1e2 e8g8 e1g1 c8e6 b2b3 d8c7 c1g5 a8d8 a1d1 d6d5 d1e1 h7h6 g5h4 d5d4 c3a4 g7g5 h4g3 e7b4 e1d1 a7a5 a4b2 b4c3 b2a4');
INSERT INTO evaluations VALUES(48647,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(48647,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(48647,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48647,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(48647,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(48647,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48647,36,'cp -10','f8d6 f4g3 e8g8 f1e2 c7c5 e1g1 b8c6 g3d6 d8d6 d4c5 d6c5 b1d2 c8f5 c2c3 f8e8 a2a4 c5b6 b2b4 a7a6 a4a5 b6c7 f1e1 h7h6 a1c1 f5g4 h2h3 g4h5 d2f1 a8d8 f3d4 h5e2 e1e2 f6e4 f1e3 c7f4 d4c6 b7c6 d1d4 f4g5 b4b5');
INSERT INTO evaluations VALUES(41692,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(41692,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(41692,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(41692,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(41692,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(41692,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(41692,36,'cp -8','f4g3 e8g8 f1e2 c7c5 g3d6 d8d6 d4c5 d6c5 e1g1 b8c6 b1d2 c8f5 c2c3 f8e8 a2a4 c5b6 b2b4 a7a6 a4a5 b6c7 f1e1 a8d8 d2b3 f5g4 a1c1 h7h6 b3d4 g4f3 d4f3 f6e4 d1c2 c7f4 e2d3 e4g5 e1e8 d8e8 f3g5 h6g5');
INSERT INTO evaluations VALUES(50249,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(50249,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(50249,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(50249,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50249,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50249,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50249,36,'cp 14','d2d4 f8g7 c1b2 b7b6 b1c3 d7d5 g1f3 e8g8 d1d2 c7c6 f3e5 d8c7 f2f4 b6b5 f1d3 a7a5 e1g1 b8d7 a1c1 c8a6 d2e2 c7b7 g2g4 d7e5 f4e5 f6d7 g4g5 a5a4 h2h4 a4a3 b2a1 h7h5 g5h6 g7h6 h4h5 b5b4 c3a4 f7f5 c2c4 b4c3 a1c3 a6d3');
INSERT INTO evaluations VALUES(50203,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(50203,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(50203,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(50203,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(50203,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(50203,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(50203,36,'cp 1','f8g7 c1b2 e8g8 f1e2 b7b6 g1f3 c7c5 e1g1 c8b7 c2c4 c5d4 f3d4 b8c6 d4c6 d7c6 d1c2 c6c5 b1c3 e7e6 a1d1 d8e7 d1d2 f8d8 f1d1 f6e8 d2d3 b7c6 e2f3 c6f3 g2f3 d8d3 c2d3 g7f6 f3f4 e7b7 d3e2 b7e7 d1d2 a8d8 d2d8 e7d8');
INSERT INTO evaluations VALUES(43221,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43221,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43221,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43221,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43221,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43221,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43221,36,'cp 1','c1b2 e8g8 f1e2 b7b6 c2c4 c8b7 e2f3 d7d5 g1e2 b8d7 e1g1 c7c6 e2f4 e7e6 d1e2 d5c4 b3c4 d8c7 b1d2 e6e5 f4d3 e5d4 b2d4 c6c5 d4b2 b7f3 e2f3 a8d8 a1d1 f6e8 b2g7 g8g7 d2e4 d7e5 d3e5 c7e5');
INSERT INTO evaluations VALUES(50700,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50700,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(50700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50700,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50700,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(50700,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(50700,36,'cp 0','c3e2 b8c6 c2c3 e7e6 g2g3 f8d6 f1g2 b7b5 g1f3 a6a5 e1g1 c8a6 f1e1 g8e7 e2f4 e8g8 f3g5 d6f4 c1f4 h7h6 g5f3 b5b4 h2h4 b4c3 b2c3 e7g6 f4e3 g6e7');
INSERT INTO evaluations VALUES(51323,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(51323,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(51323,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(51323,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(51323,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(51323,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(51323,36,'cp 50','e7e6 a2a4 c7c6 a4b5 c6b5 b2b3 f8e7 b3c4 b5c4 f3e5 g8f6 f1c4 d8c7 c1a3 e8g8 e1g1 e7a3 b1a3 b8c6 a3b5 c7b6 c4d3 c6b4 d3e2 c8a6 b5c7 b6c7 e2a6 b4a6 a1a6 f8c8 e5c6 c7d7 d1a4 h7h6 f1c1 d7b7 a4a3 g8h8');
INSERT INTO evaluations VALUES(55755,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(55755,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(55755,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(55755,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(55755,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55755,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(55755,36,'cp -10','c7c6 b5d3 d5e4 d3e4 g8f6 b1c3 f8b4 e4f3 e8g8 g1e2 b8d7 e1g1 e6e5 e2g3 d8a5 c3e4 f6e4 f3e4 e5d4 d1d4 a5b6 d4d1 d7f6 c1e3 b4c5 e3c5 b6c5 e4d3 c8e6 d1f3 a8d8 a1e1 f8e8 b2b3 h7h6 a2a4 c5a5 f3f4');
INSERT INTO evaluations VALUES(53529,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(53529,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(53529,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(53529,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(53529,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(53529,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(53529,36,'cp -12','b5d3 d5e4 d3e4 g8f6 b1c3 f8b4 e4f3 e8g8 g1e2 b8d7 e1g1 e6e5 e2g3 d8a5 c3e4 f6e4 f3e4 e5d4 d1d4 a5b6 d4d1 d7f6 e4f5 b4c5 f5c8 a8c8 c1g5 b6b2 a1b1 b2e5 g5f6 e5f6 g3e4 f6e7 e4c5 e7c5 b1b7 c8d8 d1c1 h7h6');
INSERT INTO evaluations VALUES(7733,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(7733,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(7733,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(7733,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7733,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(7733,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(7733,36,'cp 76','e1g1 h7h6 g5f3 c8a6 d2d4 a6f1 d1f1 e5e4 d4c5 e4f3 f1e1 e8f8 b1c3 d8d4 e1d1 d4d1 c3d1 a8d8 c2c3 a5c4 g2f3 g7g5 g1f1 h8g8 b2b4 g8g6 a4c2 g6g8 a2a4 h6h5 b4b5 d8d5 b5b6 a7b6 c5b6 c4b6 c1a3 c6c5 a4a5 b6c4 d1e3 c4a3 e3d5');
INSERT INTO evaluations VALUES(7732,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(7732,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(7732,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(7732,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(7732,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(7732,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(7732,36,'cp 24','h7h6 g5f3 c8a6 d2d4 a6f1 d1f1 e5e4 d4c5 e4f3 f1e1 e8f8 b1c3 d8d4 e1d1 d4d1 c3d1 a8e8 c1e3 f6d5 b2b3 e8e6 e3d2 a5b7 d1e3 d5e3 d2e3 f3g2 b3b4 f8e7 c2c3 h8e8 a4c2 g7g5 a1d1 e8d8 d1e1 e7f8 g1g2 e6e8 g2f1 d8d5 f2f4 g5g4 e3d4 h6h5 e1e8 f8e8 f1f2');
INSERT INTO evaluations VALUES(12811,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(12811,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(12811,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(12811,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(12811,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(12811,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(12811,36,'cp 34','b1c3 d7d5 d2d4 c5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 f1g2 f6d5 d1e3 c8a6 e3d5 c6d5 g2d5 a8c8 c1d2 e7e6 d5e4 f8c5 d2c3 f7f6 e2e3 c8c7 h2h4 a6b7 e4b7 c7b7 e1c1 h7h5 h1g1 e8f7 g3g4 e6e5 g4h5 f7e6 d1d2 b7c7 c1c2');
INSERT INTO evaluations VALUES(14934,30,'cp -369',NULL);
INSERT INTO evaluations VALUES(14934,31,'cp -371',NULL);
INSERT INTO evaluations VALUES(14934,32,'cp -374',NULL);
INSERT INTO evaluations VALUES(14934,33,'cp -374',NULL);
INSERT INTO evaluations VALUES(14934,34,'cp -370',NULL);
INSERT INTO evaluations VALUES(14934,35,'cp -345',NULL);
INSERT INTO evaluations VALUES(14934,36,'cp -366','c2c3 d4d3 f6g7 d8e7 c1e3 f8g7 e1g1 c8e6 f1e1 e8c8 f3g5 e6d5 d1g4 e7d7 g4d7 d8d7 b1d2 a7a5 a2a4 h8g8 g2g3 h7h5 g5f3 b7b6 e3f4 c6e7 f3e5 g7e5 f4e5 f7f5 f2f3 e7g6 g1f2 h5h4 e5f6 d7f7 f6g5 f7h7 g3h4 g6h4 e1g1 h4g6 h2h4 g6h4 g5h4 g8g1 a1g1 h7h4');
INSERT INTO evaluations VALUES(15914,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(15914,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(15914,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(15914,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(15914,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(15914,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(15914,36,'cp 92','e2e4 g7g6 f1e2 f8g7 c1e3 d8e7 d1c2 e8g8 e1g1 e5d4 f3d4 c7c6 a1d1 f8e8 f2f3 a7a6 e3g5 d7c5 g5h4 h7h6 g1h1 g6g5 h4f2 a6a5 f1e1 f6h5 g2g3 a5a4 c2d2 h5f6 e2f1 f6h5 f2e3 g7e5 f3f4 g5f4 g3f4 e5g7 d2f2 c5e4 c3e4 e7e4 f1g2');
INSERT INTO evaluations VALUES(17428,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17428,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(17428,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17428,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(17428,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(17428,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(17428,36,'cp 82','b1c3 d8c7 a2a3 f8d6 c1e3 g8f6 f1e2 e8g8 h2h3 d6f4 d1d2 f4e3 d2e3 d7d5 e4d5 e6d5 c4d5 c6e7 e2f3 c7d6 e1g1 f6d5 e3d2 d5c3 d2c3 e7d5 c3c2 d5f4 a1d1 d6g6 c2g6 h7g6 f1e1 f8d8 e1e7 g8f8 e7c7 a8b8 d1d2 f4e6 d4e6 c8e6 d2d8 b8d8 c7b7');
INSERT INTO evaluations VALUES(18569,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(18569,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(18569,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(18569,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(18569,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(18569,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(18569,36,'cp 48','b1d2 g8f6 e2e3 f6h5 f4e5 e7e6 f1d3 f8e7 h2h4 h5f6 e5h2 e8g8 c2c3 c6c5 f3e5 a7a5 h2f4 c5c4 d3c2 b7b5 e5c6 d8e8 h4h5 c8b7 c6e7 e8e7 h5h6 g7g6 f4g5 e7d6 e1g1 f6h5 g2g4 h5f6 d1e2 e6e5 f2f3 f8e8 d4e5 d7e5 e2h2 b5b4 g5f4 f6d7');
INSERT INTO evaluations VALUES(19488,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(19488,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(19488,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19488,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19488,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19488,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(19488,36,'cp 0','c1b2 e7e5 c2c4 g7g6 b1c3 g8f6 h2h4 h7h5 g1f3 f8g7 f3g5 e8g8 f1e2 d7d6 d2d3 d6d5 c3d5 f6d5 c4d5 d8d5 e2h5 g6h5 d1h5 d5d3 e3e4 f8d8 h5f7 g8h8 f7g6 h8g8 g6f7');
INSERT INTO evaluations VALUES(21149,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21149,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21149,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(21149,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(21149,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21149,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(21149,36,'cp 17','f1b5 e7e5 g1f3 f8c5 b5c6 b7c6 d2d3 g8e7 d1e2 c8g4 b1d2 e7g6 d2c4 e8g8 c1e3 c5b4 c4d2 a7a5 a2a3 b4d6 a3a4 d8e7 e1c1 g4e6 d2c4 e6c4 d3c4 f8b8 c1b1 d6c5 b2b3 c5e3 e2e3 f7f6 b1a2 g6f4 g2g3 f4e6 h1e1 c6c5 d1d5 c7c6');
INSERT INTO evaluations VALUES(21193,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(21193,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(21193,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(21193,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21193,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(21193,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21193,36,'cp 33','g1f3 e7e5 f4e5 d6e5 f1b5 f8c5 b5c6 b7c6 d2d3 g8e7 d1e2 c8g4 b1d2 e7g6 d2c4 e8g8 c1d2 d8e7 e1c1 a7a5 a2a4 f8b8 b2b3 f7f6 c1b1 g4e6 g2g3 g6f8 b1a2 c5d4 h2h3 f8d7 g3g4 d7b6 h1f1 h7h6 f3d4 e5d4 d1e1');
INSERT INTO evaluations VALUES(23333,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(23333,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(23333,32,'cp 257',NULL);
INSERT INTO evaluations VALUES(23333,33,'cp 264',NULL);
INSERT INTO evaluations VALUES(23333,34,'cp 273',NULL);
INSERT INTO evaluations VALUES(23333,35,'cp 277',NULL);
INSERT INTO evaluations VALUES(23333,36,'cp 264','f3e5 f6e5 d1h5 e8e7 c1d2 d7d6 d2a5 g8f6 h5e2 e7f7 b1a3 g7g6 f2f3 f8h6 a3c4 h8f8 c4e3 f6h5 e1c1 b7b6 a5d2 c8d7 c1b1 a7a5 e2e1 b6b5 c2c4 b5c4 d1c1 a5a4 f1c4 d8b8 c1c2 b8a7 a2a3 f7g8 d2c1 f8b8 g2g3 h6g7 h2h4 a7c5 c4e2');
INSERT INTO evaluations VALUES(27230,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(27230,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(27230,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(27230,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(27230,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(27230,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(27230,36,'cp 6','g1e2 g7g6 b2b3 f8g7 b1c3 c6b4 b5d7 d8d7 e3d2 f7f5 c1b2 e8c8 e1c1 b4c6 h1e1 g8e7 e2g3 h8f8 e4f5 e7f5 g3e4 c6d4 c3d5 d8e8 c1b1 c8b8 f2f3 d7f7 d5e3 h7h6 e3f5 d4f5 b2g7 f7g7 e4g3 f5g3 h2g3 h6h5 e1e8 f8e8 d1e1 e8e1 d2e1 g7e5 e1e5');
INSERT INTO evaluations VALUES(28952,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(28952,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(28952,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(28952,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(28952,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(28952,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(28952,36,'cp 51','d1b3 d8b6 c4c5 b6b3 a2b3 f5b1 a1b1 g7g6 e2e3 f8g7 b3b4 e8g8 b1a1 f6e4 f1d3 b8d7 h2h3 a7a6 d3e4 d5e4 f3d2 e7e5 d4e5 d7e5 f4e5 g7e5 d2e4 e5b2 a1b1 b2e5 e1e2 f7f5 e4d6 e5d6 c5d6 g8f7 g2g4 f8d8 h1d1 b7b6 g4f5 g6f5');
INSERT INTO evaluations VALUES(29018,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(29018,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(29018,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(29018,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(29018,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(29018,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(29018,36,'cp 44','g4f3 g2f3 d5c4 e2e4 b7b5 a2a4 e7e6 a4b5 c6b5 c3b5 f8b4 b5c3 e8g8 f1c4 d8c7 d1d3 b8c6 c4b5 c6e7 e1g1 a7a5 c1e3 f6h5 c3e2 e7g6 e4e5 c7b7 f1c1 a8b8 b5c6 b7e7 f3f4 e7h4 d3e4 h4h3 e4f3 h3h4 c1c4 g6e7 a1c1 e7c6 c4c6');
INSERT INTO evaluations VALUES(29028,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(29028,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(29028,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(29028,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(29028,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(29028,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(29028,36,'cp 50','g2f3 d8c7 c1d2 d5c4 e2e4 b8d7 f1c4 e7e5 d1b3 e8c8 d4d5 c8b8 a1c1 d7c5 b3c2 c6d5 c3d5 c7d7 d5f6 g7f6 d2e3 c5e6 c4e6 f7e6 e1g1 f8d6 g1h2 d8c8 c2d2 h8d8 c1c8 d7c8 f1d1 d6c7 d2c2 d8d1 c2d1 c8e8 b2b3');
INSERT INTO evaluations VALUES(31923,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31923,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31923,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31923,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(31923,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31923,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(31923,36,'cp -12','e7e6 b1d2 g8f6 f1d3 c7c5 d4c5 f8c5 f3f4 e8g8 c2c3 b8c6 e1g1 h7h6 d1e2 d8c7 g2g3 e6e5 f4e5 c6e5 d2b3 c5b6 c1f4 a8e8 f1e1 e5d3 e2d3 c7d7 g1g2 e8e1 a1e1 d7a4 e1e2 a4a2 d3b5 a2b1 b3d2 b1g6 d2f3 f8d8 f4e5 f6e4 e5d4');
INSERT INTO evaluations VALUES(33728,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(33728,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(33728,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(33728,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(33728,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(33728,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(33728,36,'cp 20','a7a6 c1d2 a5c5 d2e3 c5c6 g1f3 e7e6 e1g1 b7b5 c4b3 c8b7 d3d4 b8d7 d1e2 c6d6 d4d5 f6d5 c3d5 b7d5 a2a4 f8e7 b3d5 d6d5 a4b5 a6b5 a1a8 d5a8 e2b5 c7c6 b5e2 e8g8 f1d1 a8b7 b2b3 d7f6');
INSERT INTO evaluations VALUES(34038,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(34038,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(34038,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(34038,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(34038,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(34038,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(34038,36,'cp 21','h5e5 f6e4 e5e7 f8e7 d2d3 e4f6 g2g3 d7d5 b1c3 e8g8 f1g2 c7c6 e1g1 e7d6 f1e1 b8d7 c1d2 f8e8 e1e8 f6e8 f3d4 e8c7 a2a3 d7f8 a3a4 a7a5 d4b3 f8e6 c3e2 d6b4 d2e3 b4d6 h2h4 c8d7 e2d4 e6d4 b3d4');
INSERT INTO evaluations VALUES(39023,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(39023,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(39023,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(39023,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(39023,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(39023,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(39023,36,'cp 60','g1f3 d7d6 f1e2 b8d7 h2h3 g8e7 c2c4 e8g8 b1c3 b7b6 f4h2 e7f5 d1c2 c8b7 e1g1 f5h4 f3h4 d8h4 f1d1 h4d8 a1c1 a7a5 e2f1 f8e8 b2b3 f7f5 f2f3 d8g5 h2f4 g5e7 c3b5');
INSERT INTO evaluations VALUES(41764,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(41764,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(41764,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(41764,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(41764,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41764,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41764,36,'cp 37','b1d2 e4g3 h2g3 c7c5 c2c4 c5d4 f3d4 b8c6 c4d5 e6d5 f1b5 c8d7 d1b3 c6d4 e3d4 e8g8 b5d7 d8d7 b3d5 a8d8 e1g1 d6e7 d5d7 d8d7 d2b3 e7f6 f1d1 f8d8 a1c1 g7g6 d1d2 d7d6 c1c7 f6d4 b3d4 d6d4 d2d4 d8d4 c7b7 d4d1 g1h2 d1d2 f2f3 h7h5 a2a4 a7a5 b7b5 g8g7');
INSERT INTO evaluations VALUES(41765,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(41765,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41765,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(41765,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(41765,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(41765,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41765,36,'cp 32','e4g3 h2g3 c7c5 d4c5 d6c5 c2c4 d5c4 d1c2 h7h6 f1e2 b8d7 e1g1 e8g8 d2c4 b7b6 f1d1 d8c7 a1c1 a7a5 a2a3 f8d8 c2b1 a5a4 c4d6 d7f6 d6c8 d8d1 e2d1 c7c8 b1c2 f6d5 g1h1 c8d7 f3e5 d7e8 e5d3 c5e7 d1f3 a8d8 d3e5 d5f6');
INSERT INTO evaluations VALUES(45414,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45414,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(45414,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45414,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45414,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(45414,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45414,36,'cp 0','b8c6 c2c3 h7h6 g5h4 d8b6 d1b3 e7e6 h4g3 f8e7 b3b6 a7b6 b1d2 c8d7 g3c7 e7d8 c7d6 d8e7');
INSERT INTO evaluations VALUES(46882,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(46882,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46882,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(46882,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(46882,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46882,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(46882,36,'cp 21','e7e6 g2g3 f8e7 f1g2 e8g8 e1g1 d5c4 f3e5 b8c6 g2c6 b7c6 e5c6 d8e8 c6e7 e8e7 d1a4 a7a5 a4c4 c8a6 c4c6 f6d5 c3d5 e6d5 f1e1 e7b4 c1e3 b4b2 c6d5 a5a4 d5c5 a4a3 a1b1 b2a2 b1a1 a2c4 a1a3 c4c5 d4c5 a6c4 a3a8 f8a8 f2f3 a8a2 g1f2 a2c2 c5c6 c4d5');
INSERT INTO evaluations VALUES(47346,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(47346,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(47346,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(47346,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(47346,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(47346,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(47346,36,'cp -75','b7b5 f1e2 e7e6 e2f3 f6d5 g1e2 f8b4 b1c3 a7a6 e1g1 c7c6 f4g3 b8d7 a2a3 b4e7 a3a4 c8b7 c3e4 e8g8 e4d6 d8b6 a4a5 b6a7 e3e4 d5f6 b2b3 c4b3 d1b3 c6c5 d6b7 c5c4 b3e3 a7b7 e2f4 a8d8 e4e5 f6d5 f3d5 e6d5 e3f3 d7b8 f3g4 b5b4 e5e6 f7e6 f4e6');
INSERT INTO evaluations VALUES(48997,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(48997,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(48997,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(48997,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(48997,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(48997,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(48997,36,'cp 168','d8g5 c4d5 g5f6 d5b7 c8b7 f3b7 f6c6 b7c8 e8e7 c8h3 c6e6 h3f3 e6c6 f3a3 e7e8 a3h3 h7h5 d2d3 c6d7 h3d7 b8d7 h2h3 a7a5 c1e3 f7f5 b1c3 f8b4 e3d2 e8f7 e1c1 b4d6 h1f1 d7c5 f2f4 e5e4 d3e4 c5e4 c3e4 f5e4 d2e3 h8e8 d1d4 h5h4 c2c3 a5a4 c1c2');
INSERT INTO evaluations VALUES(33924,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(33924,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(33924,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(33924,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(33924,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(33924,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(33924,36,'cp 184','c4d5 g5f6 d5b7 c8b7 f3b7 f6c6 b7c8 e8e7 c8h3 c6e6 h3a3 e7d7 a3a4 e6c6 a4g4 c6e6 g4e6 d7e6 d2d3 b8c6 b1d2 f8c5 f2f3 f7f5 e1d1 g7g5 c2c3 h7h5 b2b4 c5d6 d2c4 g5g4 h1e1 g4f3 g2f3 e6f6 c4d6 c7d6 b4b5 c6e7 a2a4 f5f4 d1c2');
INSERT INTO evaluations VALUES(33920,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(33920,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(33920,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(33920,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(33920,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(33920,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(33920,36,'cp 177','g5g6 d5b7 c8b7 f3b7 g6c6 b7c8 e8e7 c8h3 c6e6 h3e6 e7e6 d2d3 h7h5 c1e3 b8c6 b1d2 c6b4 e1d1 f7f6 h1e1 g7g5 a2a3 b4d5 c2c3 f8d6 d1c2 a7a5 a3a4 e6f7 a1d1 d5f4 g2g3 f4e6 d2c4 g5g4 h2h3 g4h3 e1h1 h5h4 h1h3 h4g3 h3g3 h8h2 d1d2 d6c5');
INSERT INTO evaluations VALUES(49000,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(49000,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(49000,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(49000,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(49000,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(49000,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(49000,36,'cp 21','d5c3 b2c3 f8d6 e1g1 e8g8 d2d4 e5d4 c1g5 d8d7 c3d4 h7h6 g5h4 d7c6 c4d3 c8g4 c2c4 b8d7 c4c5 d6f4 d1b3 g4f3 g2f3 d7f6 f1e1 a8d8 a1d1 b7b6 d3b5 c6d5 h4g3 f6h5 b3d5 d8d5 b5c4 d5d7 c5c6 d7d8 c4b3 h5g3 h2g3 f4d6 f3f4 g8h8 e1e4');
INSERT INTO evaluations VALUES(49087,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(49087,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(49087,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49087,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(49087,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(49087,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(49087,36,'cp 32','b2c3 f8d6 e1g1 e8g8 d2d4 e5d4 c1g5 d6e7 g5e7 d8e7 c3d4 e7f6 f1e1 c8g4 a1b1 b7b6 h2h3 g4f3 d1f3 f6f3 g2f3 g7g6 h3h4 b8d7 e1e7 a8d8 b1d1 g8g7 a2a4 c7c6 c2c3 d7f6 e7a7 d8a8 a7b7 f6d5 c4d5 c6d5 g1g2 a8a4');
INSERT INTO evaluations VALUES(49086,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(49086,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(49086,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(49086,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(49086,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(49086,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(49086,36,'cp 38','f8d6 e1g1 e8g8 d2d4 e5d4 c1g5 d6e7 g5e7 d8e7 c3d4 e7f6 f1e1 c8g4 a1b1 b7b6 c4d5 c7c6 d5e4 b8d7 h2h3 g4f3 d1f3 a8d8 c2c3 f6f3 e4f3 c6c5 e1e7 c5d4 c3d4 f8e8 e7e8 d8e8 b1c1 e8d8 c1c7 d7f8 d4d5 d8d7 c7c8 g7g5 a2a4 h7h6 g2g3 g8g7 g1g2 f8g6 f3e4');
INSERT INTO evaluations VALUES(49454,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(49454,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(49454,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(49454,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(49454,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(49454,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(49454,36,'cp -97','b8c6 c2c3 d7d5 e4d5 e6d5 d2d4 c8g4 f3e3 f8e7 h2h3 g4h5 f1e2 h5g6 g1f3 e8g8 e3d2 e7d6 e1g1 c6e7 d2d1 c7c6 f1e1 f8e8 f3e5 e7f5 c1f4 f6e4 e2g4 d8b6 d1b3 h7h5 g4e2 d6e5 b3b6 a7b6 d4e5 b6b5');
INSERT INTO evaluations VALUES(50131,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(50131,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(50131,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(50131,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(50131,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(50131,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(50131,36,'cp 77','h2h4 h7h6 d1f3 d7d6 b1c3 e7e6 e1c1 d8e7 e3e4 b8c6 e4f5 g6f5 f1c4 c6a5 c4d3 c8d7 d1e1 e8c8 d3f5 a5c6 d4d5 f6d5 c3d5 e7f7 f5g4 e6d5 f4d2 f7f3 g4f3 d5d4 f3e4 c6e5 g1e2 c7c5 e2f4 h6h5 f4d5 e5g4 h1f1 d7b5 f2f3 b5f1 f3g4 f1c4 g4h5 c4d5 e4d5');
INSERT INTO evaluations VALUES(54727,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(54727,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(54727,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54727,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(54727,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(54727,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54727,36,'cp 0','c8g4 d2d4 f6e5 d4e5 b8c6 f1b5 e7e6 b1d2 g8h6 h2h3 g4h5 d2b3 d8d7 e1g1 h6f7 g2g4 h5g6 f3d4 f7e5 f1e1 g6e4 c1e3 a7a6 b5c6 e5c6 f2f3 e4g6 d4c6 d7c6 b3d4 c6d6 d4e6 d6g3 g1f1 g3h3 f1g1 h3g3 g1f1');
INSERT INTO evaluations VALUES(56413,30,'cp -139',NULL);
INSERT INTO evaluations VALUES(56413,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(56413,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(56413,33,'cp -157',NULL);
INSERT INTO evaluations VALUES(56413,34,'cp -169',NULL);
INSERT INTO evaluations VALUES(56413,35,'cp -149',NULL);
INSERT INTO evaluations VALUES(56413,36,'cp -155','d2d4 b8c6 b1c3 c8e6 d4e5 c6e5 h2h3 h7h5 g1f3 e5f3 d1f3 h5g4 h3g4 e6g4 f3g2 h8h1 g2h1 g8f6 c1d2 d8d6 h1g2 e8c8 f2f3 g4f5 e1c1 a7a6 c1b1 g7g6 c3e2 f8h6 g2g1 d6b6 e2d4 f5d7 f1e2 d8e8 d2c1 c7c5 d4b3');
INSERT INTO evaluations VALUES(1210,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(1210,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(1210,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(1210,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(1210,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(1210,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(1210,36,'cp -131','d1e2 c8f5 b1c3 e4e2 e1e2 e8c8 d3e1 g8e7 d2d3 c6c5 c1e3 e7c6 a1d1 f8d6 e2d2 h8e8 d2c1 c6d4 e1f3 f5g4 e3d4 c5d4 c3e4 d6f4 c1b1 f7f5 e4g3 g7g6 a2a3 g4f3 g2f3 e8e5 h2h4 d8e8 h4h5 c8d7 h5g6 h7g6 h1h7 d7d6 h7g7');
INSERT INTO evaluations VALUES(1211,30,'cp -142',NULL);
INSERT INTO evaluations VALUES(1211,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(1211,32,'cp -140',NULL);
INSERT INTO evaluations VALUES(1211,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(1211,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(1211,35,'cp -141',NULL);
INSERT INTO evaluations VALUES(1211,36,'cp -141','c8f5 b1c3 e4e2 e1e2 e8c8 d3e1 d8e8 e2d1 g8e7 h2h3 c6c5 d2d3 e7c6 c1e3 e8d8 b2b3 h7h6 d1d2 f5e6 c3e2 f8e7 e1f3 e7f6 e3c5 f6a1 h1a1 e6d5 a1g1 d5f3 g2f3 g7g6 c5e3 a6a5 a2a4 b7b6 e2c3 c6e5 f3f4 e5c6 g1e1');
INSERT INTO evaluations VALUES(6793,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(6793,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(6793,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(6793,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(6793,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(6793,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(6793,36,'cp 203','a2a3 b4a6 b1d2 g8f6 f1a6 b7a6 e1g1 e8g8 f1e1 f6e8 b2b4 f7f6 d2b3 f8f7 c2c4 d6e7 b3a5 d7d6 a5c6 d8d7 f3h4 g7g6 h4f3 d7g4 d1a4 e7f8 c6a7 a8a7 a4e8 a7a8 h2h3 g4d7 e8d7 c8d7 c1e3 f6f5 f3g5 f7g7 a1c1 f5f4 e3d2 h7h6 g5e6 d7e6 d5e6');
INSERT INTO evaluations VALUES(6794,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(6794,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(6794,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(6794,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(6794,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(6794,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(6794,36,'cp 210','b4a6 b1c3 g8f6 g2g4 d6c5 f1a6 f6g4 h1g1 c5f2 e1e2 f2g1 d1g1 g4f6 a6d3 d7d6 c1e3 d8e7 a1f1 c8h3 f1e1 f6h5 e2d2 e8g8 d2c1 a7a6 d3e2 h5f4 e3f4 e5f4 e2d3 a8e8 e4e5 h3d7 e5d6 e7d6 c3e4 d6b6 g1g5 f7f6 g5f4');
INSERT INTO evaluations VALUES(6909,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(6909,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(6909,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(6909,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(6909,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(6909,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(6909,36,'cp 221','b1c3 g8f6 g2g4 d6c5 h1g1 d7d6 g4g5 f6g8 h2h4 c8d7 h4h5 d8e7 f3h4 h7h6 d1f3 e8c8 c1d2 c5d4 f1a6 b7a6 e1c1 d8f8 h4f5 d7f5 f3f5 c8b8 f5f3 h6g5 d2g5 e7e8 g5e3 d4c3 b2c3 g7g6 c3c4 h8h5 c4c5 e8a4 c5d6');
INSERT INTO evaluations VALUES(8801,30,'cp 294',NULL);
INSERT INTO evaluations VALUES(8801,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(8801,32,'cp 300',NULL);
INSERT INTO evaluations VALUES(8801,33,'cp 290',NULL);
INSERT INTO evaluations VALUES(8801,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(8801,35,'cp 300',NULL);
INSERT INTO evaluations VALUES(8801,36,'cp 305','d1b3 c8e6 b3b7 d5e7 b7b5 c5d6 c4e6 f7e6 b5a4 d8d7 b1d2 e8g8 d2e4 e7g6 c1e3 a8d8 b2b3 a7a5 a1d1 d6e7 g1h1 d7e8 f1e1 g8h8 f3g1 f8f5 a4c4 e7d6 f2f3 e8d7 g1e2 d6e7 e1g1 f5f8');
INSERT INTO evaluations VALUES(9802,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(9802,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(9802,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(9802,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(9802,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(9802,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(9802,36,'cp 28','f3e5 c6e5 d3d4 c5d6 d4e5 d6e5 c3b5 a7a6 f2f4 e5b2 c1b2 a6b5 e4e5 f6e4 e1g1 d7d5 e5d6 e8g8 d1d4 e4f6 e2b5 c8d7 b5c4 d7f5 c4b3 d8d6 d4d6 c7d6 b2f6 g7f6 f1d1 a8a6 d1d2 f8c8 a1e1 g8f8 e1e2 a6c6 h2h3 h6h5 b3d5 c6c2 d5b7 c2d2 e2d2 c8c7 b7f3 f5e6 d2d6');
INSERT INTO evaluations VALUES(13942,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(13942,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(13942,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(13942,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13942,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13942,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13942,36,'cp 0','d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 e1g1 a7a6 c4d3 c6d4 e3d4 b7b5 d1f3 a8b8 g1h1 d8c7 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(19450,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(19450,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19450,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(19450,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19450,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(19450,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(19450,36,'cp -8','g7g6 e1g1 f8g7 f1e1 e7e5 d3c4 d7d6 b1c3 g8e7 c3d5 h7h6 d2d3 e8g8 a2a4 g8h7 c1d2 f7f5 a4a5 c8d7 h2h3 f5f4 b2b4 c5b4 a5a6 b7b6 d5b4 a8c8 b4c6 e7c6 c4d5 c6e7 c2c4 e7d5 c4d5');
INSERT INTO evaluations VALUES(28996,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28996,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(28996,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28996,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(28996,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(28996,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28996,36,'cp -3','h7h6 c2c4 e7e6 d1b3 d8b6 c4c5 b6b3 a2b3 b8d7 b3b4 a7a6 c1e3 e8c8 f1c1 f5b1 c1b1 f8e7 e3d2 g7g5 d2c3 g5g4 f3d2 h6h5 b1e1 h8g8 e2e4 d5e4 d2c4 h5h4 g2e4 h4h3 e4c2 f6d5 c4e5 d7e5');
INSERT INTO evaluations VALUES(28908,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28908,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28908,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28908,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28908,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28908,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28908,36,'cp 0','c2c4 e7e6 b1c3 b8d7 f3d2 f8e7 e2e4 d5e4 d2e4 f6e4 g2e4 f5e4 c3e4 e8g8 c1f4 d7f6 f1e1 d8a5 f4e5 f6d7 e5f4');
INSERT INTO evaluations VALUES(28907,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28907,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28907,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28907,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28907,36,'cp -4','e7e6 b1c3 b8d7 f3d2 f8e7 b2b3 e8g8 c1b2 f8e8 e2e4 d5e4 d2e4 d8c7 d1e2 a8d8 a1d1 f6e4 c3e4 d7f6 b2c3 c7b6 d1d2 f6e4 g2e4 f5e4 e2e4 d8d7 d2d3 e7b4 c3b4 b6b4 f1d1 b4a3 d1d2 e8d8 e4e3 a3b4 h2h4 b7b5');
INSERT INTO evaluations VALUES(29198,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29198,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29198,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29198,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29198,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29198,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29198,36,'cp 0','b1c3 b8d7 f3d2 f8e7 e2e4 f6e4 d2e4 d5e4 g2e4 f5e4 c3e4 e8g8 c1f4 d7f6 f1e1 d8b6 f4e5 f6d7 e5f4');
INSERT INTO evaluations VALUES(29197,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29197,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(29197,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29197,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29197,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29197,36,'cp 0','b8d7 f3d2 f8e7 e2e4 f6e4 c3e4 d5e4 d2e4 d7f6 f1e1 e8g8 c1f4 d8b6 c4c5 b6b2 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(30253,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(30253,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(30253,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(30253,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(30253,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(30253,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(30253,36,'cp 98','f8c5 a2a3 b8c6 c2c3 a7a5 c1g5 h7h6 g5h4 e8g8 f1d1 d8e7 b1d2 a5a4 h2h3 f8d8 d2f1 c6a5 d1d8 a8d8 c4b5 a5b3 a1d1 d8d1 f3d1 g7g5 h4g3 f6e4 d1f3 e4g3 f1g3 e5e4 g3e4 c5b6 b5a4 b3c5 a4c2 c5e4 c2e4 c7c6 f3f5 e7d6 e4c2 d6e6 g1f1 e6f5 c2f5 b6a5 f1e2 g8g7 a3a4');
INSERT INTO evaluations VALUES(36613,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36613,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(36613,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(36613,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(36613,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(36613,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(36613,36,'cp -3','c7c5 b1c3 b8c6 d4c5 d6c5 e1g1 d8b6 a1b1 c8e6 f3g5 c6b4 g5e6 b6e6 d1f3 b7b6 f1d1 b4d3 d1d3 a8d8 d3d8 f8d8 b1d1 d8d1 f3d1 h7h6 a2a4 g8h7 d1f3 e6d7 e3e4 d7e6 f4c7 e6c6 c7h2 c5c4 g2g4 c6e6');
INSERT INTO evaluations VALUES(40189,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(40189,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(40189,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(40189,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(40189,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(40189,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(40189,36,'cp 122','f1d3 e8g8 g1e2 f6e8 b2b3 a7a6 e1g1 e6e5 d1d2 b8d7 f2f3 g7g6 d2h6 d7f6 g1h1 g8h8 h6d2 e8g7 c3d1 f6d7 d1f2 a8b8 a2a4 b7b6 a1a2 f7f5 f3f4 e5f4 e4f5 g7f5 e2f4 d7e5 c1b2 e7g5');
INSERT INTO evaluations VALUES(40190,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(40190,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(40190,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(40190,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(40190,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(40190,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(40190,36,'cp 131','e8g8 g1e2 f6e8 b2b3 a7a6 e1g1 b8d7 f2f4 e6e5 f4f5 e7g5 g1h1 g5c1 d1c1 e8f6 e2g1 h7h6 c1d2 b7b5 c4b5 a6b5 d3b5 c8a6 a2a4 a6b5 a4b5 d8b8 a1a4 b8b7 f1a1 f8b8 g1f3 d7b6 a4a6 c5c4 b3b4 b6d7');
INSERT INTO evaluations VALUES(40237,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(40237,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40237,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(40237,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40237,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(40237,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40237,36,'cp 12','c4d5 f6d5 f1d3 e8g8 e1g1 b8c6 f1e1 e7f6 d3e4 c6e7 f3e5 c8d7 c3d5 e6d5 e4d3 d7f5 c1e3 a8c8 e5g4 d8b6 d3f5 e7f5 d1f3 f5e3 g4e3 b6d4 e3d5 d4b2 a1b1 c8e8 e1c1 b2a2 d5f6 g7f6 f3b7 e8b8 b7b8 f8b8 b1b8 g8g7 h2h3 f6f5 b8e8 a2d2 c1b1 d2g5');
INSERT INTO evaluations VALUES(40225,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(40225,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40225,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40225,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(40225,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(40225,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(40225,36,'cp 20','f6d5 f1d3 b8c6 e1g1 e8g8 f1e1 e7f6 d3e4 c6e7 d1d3 h7h6 f3e5 c8d7 e5g4 d7c6 d3g3 h6h5 g4f6 d5f6 e4c2 d8d4 h2h3 e7g6 c2g6 f7g6 e1e6 g8h7 c1e3 d4c4 e6e5 a8d8 a1c1 c4b4 e5g5 h5h4 g3f4 b4f4 e3f4 f8e8 g5c5');
INSERT INTO evaluations VALUES(40210,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(40210,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40210,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40210,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(40210,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(40210,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40210,36,'cp 6','f1d3 b8c6 e1g1 e8g8 f1e1 e7f6 d3e4 c6e7 f3e5 c8d7 d1h5 g7g6 h5f3 d7c6 e4d5 e7d5 c1h6 f6g7 h6g7 g8g7 e5c6 b7c6 b2b3 d8f6 f3f6 d5f6 a1c1 f8d8 c3e2 a8c8 c1c4 f6d5 g2g3 h7h5 e1c1 d5e7 g1g2 d8d7 c4c5 d7d5');
INSERT INTO evaluations VALUES(40211,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40211,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(40211,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(40211,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(40211,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(40211,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40211,36,'cp 6','b8c6 e1g1 e8g8 f1e1 e7f6 d3e4 c6e7 f3e5 c8d7 c3d5 e6d5 e4d3 d7f5 c1e3 a8c8 e5g4 d8d6 h2h3 f8e8 a1c1 c8c1 g4f6 d6f6 e3c1 f5d3 d1d3 e8d8 d3b3 b7b6 b3g3 e7f5 c1g5 f5g3 g5f6 g7f6 f2g3 d8c8 e1e7 c8c2');
INSERT INTO evaluations VALUES(45637,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45637,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45637,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45637,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45637,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45637,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45637,36,'cp 0','d5c4 f3e5 b8d7 e5c4 b7b5 c4e5 c8b7 f1g2 b5b4 e5d7 d8d7 c3e4 f6e4 g2e4 c6c5 e4b7 d7b7 e1g1 a8c8 c1f4 c5c4 d1a4 b7d7 a4a6 c8c6 a6a4');
INSERT INTO evaluations VALUES(45711,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45711,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45711,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45711,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45711,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45711,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45711,36,'cp 0','f3e5 b8d7 e5c4 b7b5 c4e5 c8b7 f1g2 b5b4 e5d7 d8d7 c3e4 f6e4 g2e4 c6c5 e4b7 d7b7 e1g1 a8c8 c1f4 c5c4 d1a4 b7d7 a4a6 c8c6 a6a4');
INSERT INTO evaluations VALUES(46170,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(46170,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(46170,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(46170,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(46170,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46170,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46170,36,'cp 0','b1c3 c5d4 f3d4 b8d7 d1a4 f8c5 d4b3 e8g8 b3c5 d7c5 a4a3 b7b6 c4d5 f6d5 c3d5 e6d5 a1d1 d5d4 e2e3 c5e6 f4d6 d4e3 f2e3 f8e8 d6g3 d8g5 f1c4 e6f8 e1g1 c8e6 g3f4 g5e7 f4d6 e7g5');
INSERT INTO evaluations VALUES(46335,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(46335,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(46335,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(46335,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(46335,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(46335,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(46335,36,'cp 36','e2e3 f8d6 f4g3 e8g8 g1f3 d8b6 a1b1 d6g3 h2g3 b8c6 f1d3 h7h6 a2a3 f8d8 f3e5 c6e5 d4e5 f6d7 c3a4 b6c7 f2f4 d7c5 a4c5 c7c5 d1e2 c8d7 g3g4 a8c8 g4g5 h6g5 f4g5 g7g6 e2f2 c5c1 b1c1 c8c1 e1d2 c1h1 f2f6 d8c8 d3g6 f7g6 f6g6 g8f8 g6f6 f8g8 g5g6');
INSERT INTO evaluations VALUES(46332,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(46332,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(46332,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(46332,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(46332,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46332,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(46332,36,'cp 27','f8d6 f4g3 h7h6 f1d3 d6g3 h2g3 b8c6 g1f3 d8d6 e1g1 e8g8 a1c1 c8d7 a2a3 f8d8 d1e2 a7a6 f1d1 d7e8 c1c2 a8c8 d1c1 d6b8 b2b4 c6e7 f3e5 e7c6 f2f4 c6e5 f4e5 f6d7 g1h2 d7b6 b4b5 a6b5 c3b5 e8b5 d3b5 c8c2');
INSERT INTO evaluations VALUES(50299,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(50299,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(50299,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(50299,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(50299,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(50299,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50299,36,'cp 51','d7d6 e5d6 e7d6 d2d4 f8e7 f1d3 b8c6 g1e2 e8g8 e1g1 c8g4 d1c2 g7g6 a2a3 e7f6 c1e3 f6g7 a1d1 c6e7 h2h3 g4f5 b2b3 f5d3 c2d3 d6d5 e3g5 f8e8 e2f4 c7c6 c4d5 d8d7 d5c6 b7c6 d1d2 e7f5 f1d1 b6d5 f4d5');
INSERT INTO evaluations VALUES(46920,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(46920,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(46920,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(46920,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(46920,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(46920,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(46920,36,'cp 60','e5d6 e7d6 d2d4 f8e7 f1d3 e8g8 g1e2 b8c6 e1g1 c8g4 d1c2 g7g6 a2a3 g4e2 c3e2 e7f6 c1e3 f8e8 b2b4 a7a6 a1d1 d6d5 c4c5 b6d7 c2d2 f6g7 d1e1 d7f6 f2f3 d8d7 h2h3 e8e7 e3f2 c6d8 e2f4 e7e1 f1e1 d8c6 d3b1 a8e8 f4e2 e8e7');
INSERT INTO evaluations VALUES(46913,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(46913,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(46913,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(46913,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(46913,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(46913,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(46913,36,'cp 51','e7d6 d2d4 f8e7 f1d3 b8c6 g1e2 e8g8 h2h3 f8e8 e1g1 g7g6 b2b3 c6b4 d3b1 c7c6 a2a3 b4a6 a3a4 b6d7 a4a5 e7f8 b1e4 a6b4 e4f3 d6d5 c4d5 d7f6 d5c6 b7c6 c1a3 c8b7 a3b4 f8b4 d1d3 f6d5 c3d5 c6d5');
INSERT INTO evaluations VALUES(51924,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(51924,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(51924,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(51924,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(51924,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(51924,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(51924,36,'cp -82','d7d5 c4b5 c8d7 b5d7 d8d7 e4d5 d7d5 e1g1 g8e7 d2d3 e7f5 b1d2 b8c6 d2c4 f8e7 a2a4 b7b6 c1d2 a8d8 h3f4 d5d7 f1e1 e8g8 f4h5 c6b4 d2c3 f5d4 c3d4 d7d4 d1e2 g7g6 h5g3 d4d7 g3e4 b4c6 e4d2 c6d4 e2d1 d7d5 c4e3');
INSERT INTO evaluations VALUES(54158,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(54158,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(54158,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(54158,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(54158,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(54158,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(54158,36,'cp -69','c4b5 c8d7 b5d7 d8d7 e4d5 d7d5 e1g1 b8c6 d2d3 g8e7 b1a3 e7f5 a3c4 f8e7 a2a4 b7b6 h3f4 d5d7 f4h5 e8g8 c1d2 a8e8 h5g3 f5h4 d2c3 h4g6 f2f4 f7f5 d1e2 e7d8 a1e1 c6d4 c3d4 d7d4 e2f2 d4d7 b2b3 d8c7 g3e2 e6e5 f4e5 g6e5');
INSERT INTO evaluations VALUES(53057,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53057,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53057,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53057,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53057,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(53057,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53057,36,'cp 12','g8f6 g1f3 b8c6 f1c4 e7e6 e1g1 f8e7 d2d4 c5d4 f3d4 a7a6 a2a4 c8d7 c1e3 d8c7 f2f4 f6e4 c3e4 d6d5 c4d5 e6d5 e4c3 c6b4 d1d2 e8g8 a4a5 f8e8 d4b3 d7f5 e3b6 c7c8 b3d4 e7f6 d4f5 c8f5 f1d1 b4c2 c3d5');
INSERT INTO evaluations VALUES(56532,30,'cp 269',NULL);
INSERT INTO evaluations VALUES(56532,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(56532,32,'cp 256',NULL);
INSERT INTO evaluations VALUES(56532,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(56532,34,'cp 276',NULL);
INSERT INTO evaluations VALUES(56532,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(56532,36,'cp 250','b8c6 f1b5 f8b4 c2c3 d4c3 b5c6 b7c6 d1d8 e8d8 b1c3 g8e7 c1e3 e7g6 e1c1 d8e8 c3e2 f7f6 e5f6 g7f6 f3d4 b4d6 d4c6 c8b7 c6d4 b7e4 f2f3 e4b7 d4f5 d6f8 e2g3 b7c8 e3h6 e8f7 h6f8 h8f8 d1d2 c8f5 g3f5 a8d8 h1d1 d8d2');
INSERT INTO evaluations VALUES(21783,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(21783,31,'cp 282',NULL);
INSERT INTO evaluations VALUES(21783,32,'cp 288',NULL);
INSERT INTO evaluations VALUES(21783,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(21783,34,'cp 273',NULL);
INSERT INTO evaluations VALUES(21783,35,'cp 259',NULL);
INSERT INTO evaluations VALUES(21783,36,'cp 270','f1b5 f8b4 c2c3 d4c3 d1d8 e8d8 b1c3 g8e7 c1e3 b4c3 b2c3 d8e8 e1c1 c8d7 b5e2 e7g6 d1d2 c6e5 f3e5 g6e5 f2f4 e5g6 f4f5 g6e5 e3f4 f7f6 f4e5 f6e5 d2d5 d7c6 d5e5 e8f8 h1d1 a8e8 e5c5 g7g6 f5f6 e8e4 e2f3 e4e6 d1d8 e6e8 d8e8 f8e8 f3c6 b7c6 c5c6');
INSERT INTO evaluations VALUES(21542,30,'cp 293',NULL);
INSERT INTO evaluations VALUES(21542,31,'cp 294',NULL);
INSERT INTO evaluations VALUES(21542,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(21542,33,'cp 294',NULL);
INSERT INTO evaluations VALUES(21542,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(21542,35,'cp 283',NULL);
INSERT INTO evaluations VALUES(21542,36,'cp 282','f8b4 c2c3 d4c3 d1d8 e8d8 b1c3 g8e7 c1e3 b4c3 b2c3 d8e8 e1c1 a7a6 b5a4 c8d7 a4b3 a8d8 d1d2 h7h6 h2h4 d7e6 d2d8 e8d8 b3e6 f7e6 h4h5 g7g5 h5g6 e7g6 h1h5 d8d7 c1d2 b7b6 d2e2 b6b5 g2g4 b5b4 c3b4 c6b4 a2a4 b4a2 e3d2 c7c5 a4a5 d7c6 h5h6 h8h6 d2h6 a2c3 e2e3');
INSERT INTO evaluations VALUES(57036,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(57036,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(57036,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(57036,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(57036,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(57036,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(57036,36,'cp -23','d7d5 e4d5 e6d5 g1f3 f8d6 e1g1 g8e7 f1e1 h7h6 d3f1 e8g8 d2d4 c8e6 b1d2 b8d7 f3e5 f8e8 d2f3 c7c6 c1f4 d7f6 e5d3 d6f4 d3f4 e6g4 h2h3 g4f3 d1f3 d8b6 c2c3 b6b2 e1b1 b2a3 b1b7 f6e4 a1e1 a3c3 f3c3 e4c3');
INSERT INTO evaluations VALUES(55899,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55899,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(55899,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(55899,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55899,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(55899,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(55899,36,'cp -18','e4d5 e6d5 g1f3 f8d6 e1g1 g8e7 f1e1 e8g8 d3f1 e7g6 d2d4 c8e6 b2b3 b8d7 c2c4 c7c6 b1c3 d7f6 h2h3 e6f5 c1g5 d6b4 g5d2 b4a3 d2c1 a3e7 c4d5 c6d5 c1g5 h7h6 f1d3 h6g5 d3f5 g6h4 f3h4 g5h4 f5d3');
INSERT INTO evaluations VALUES(57472,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(57472,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(57472,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(57472,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(57472,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(57472,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(57472,36,'cp 30','d7d5 e4g3 e5e4 f3d4 g8f6 d2d3 c7c5 d4f5 b8c6 d3e4 f6e4 c2c3 c8f5 g3f5 d8d7 f5g3 e4f6 f1d3 d7e7 d3e2 e8c8 e1g1 c8b8 c1f4 b8a8 e2b5 e7d7 d1a4 f8d6 h2h3 d5d4 f4d6 d7d6 a1e1 f6d5 g3f5 d6g6 b5c6 g6f5 c6d5 f5d5 c3d4');
INSERT INTO evaluations VALUES(56841,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(56841,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(56841,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(56841,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(56841,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(56841,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(56841,36,'cp 14','e4g3 e5e4 f3d4 f8c5 d4b3 c5b6 c2c4 c7c6 d2d3 e4d3 f1d3 d8e7 d1e2 e7e2 d3e2 g8f6 e1g1 e8g8 c1e3 b6e3 f2e3 b8d7 c4d5 f6d5 f1f8 g8f8 a1f1 f8e8 e3e4 d5f6 b3d4 d7e5 d4f3 e5f3 g2f3 c8e6 b2b3 a8d8 f1d1 a7a5 g1f2 d8d1 e2d1 e8d7 f3f4');
INSERT INTO evaluations VALUES(1991,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(1991,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(1991,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(1991,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(1991,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(1991,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(1991,36,'cp 13','f8b4 c1d2 g8f6 d1e2 e8g8 f2f3 d7d5 a2a3 b4a5 e1c1 d5e4 c3e4 a5d2 e4d2 e5e2 f1e2 f8e8 e2b5 c7c6 b5c4 g8f8 a3a4 h7h5 d2b3 c8e6 c4e6 e8e6 h1e1 e6e1 d1e1 a8c8 c2c3 b7b6 b3d4 f6e8 c1c2 e8d6 e1e2 c6c5 d4b5 d6b5 a4b5');
INSERT INTO evaluations VALUES(379,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(379,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(379,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(379,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(379,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(379,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(379,36,'cp 19','c1d2 g8f6 d1e2 e8g8 f2f3 d7d5 a2a3 b4a5 e1c1 d5e4 c3e4 a5d2 e4d2 e5e2 f1e2 f8e8 e2b5 c7c6 b5c4 g8f8 d2b3 c8e6 c4e6 e8e6 h1e1 e6e1 d1e1 a8e8 e1e8 f6e8 c1d2 f8e7 b3a5 e8d6 d2d3 e7d7 a5c4 d6c4 d3c4 b7b6 c4d4 f7f6 f3f4 g7g6 g2g4 d7e7 d4e4');
INSERT INTO evaluations VALUES(7359,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(7359,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(7359,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(7359,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(7359,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(7359,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(7359,36,'cp 70','c2c3 e8g8 a4c2 d6e7 d2d4 d7d6 f1e1 c8g4 d4d5 c6b8 c1e3 b8d7 h2h3 g4h5 b1d2 c7c6 c3c4 a8c8 a2a3 c6d5 c4d5 h5g6 f3h4 f6h5 h4g6 h7g6 d2f3 b7b5 a3a4 h5f4 d1d2 e7g5 a4b5 a6b5 e1d1');
INSERT INTO evaluations VALUES(11424,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(11424,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(11424,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(11424,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(11424,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(11424,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(11424,36,'cp -20','b7b5 a4b3 c6a5 e1g1 f8e7 c1d2 a5b3 a2b3 c8b7 f1e1 e8g8 c3e2 f8e8 e2g3 c7c5 h2h3 a6a5 c2c4 b5b4 f3h2 f6d7 h2g4 d7f8 e1f1 f8g6 g3f5 e7f8 g2g3 g6e7 f5e7 f8e7 g4e3 e7g5 e3f5 g5d2 d1d2 b7c6');
INSERT INTO evaluations VALUES(11648,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(11648,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(11648,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(11648,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(11648,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(11648,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(11648,36,'cp -4','a4b3 c6a5 e1g1 f8e7 c1d2 a5b3 a2b3 c8b7 f1e1 e8g8 c3e2 f8e8 e2g3 e7f8 c2c4 g7g6 b3b4 b5c4 d3c4 d8d7 a1a5 f8g7 d1c1 a8b8 b2b3 h7h5 d2h6 f6h7 h6g7 g8g7 c1d2 d7e7 g3f1 h5h4 h2h3 h7f8 f3g5 f7f6');
INSERT INTO evaluations VALUES(11640,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(11640,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(11640,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(11640,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(11640,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(11640,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(11640,36,'cp -13','c6a5 e1g1 f8e7 f1e1 e8g8 c1d2 c8b7 c3e2 a5b3 a2b3 f8e8 e2g3 e7f8 c2c4 g7g6 b3b4 b5c4 d3c4 b7c6 h2h3 d8b8 f3h2 h7h5 d1c2 b8b6 a1a5 a8b8 b2b3 f8g7 e1a1 h5h4 a5a6 h4g3 a6b6 g3h2 g1h2 b8b6 f2f3');
INSERT INTO evaluations VALUES(13532,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(13532,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(13532,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(13532,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(13532,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(13532,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(13532,36,'cp -75','e4d5 e6d5 c4a2 d5d4 d1e2 f8e7 c3e4 d4d3 e4f6 g7f6 e2e4 d3c2 e1g1 c8e6 a2e6 f7e6 b2b4 d8d7 b4b5 c6d8 a3a4 e6e5 a1a2 d8e6 a2c2 e8c8 f1e1 h7h5 g2g3 h8g8 e4h7 e7d6 h7d7 d8d7 c1b2 b7b6 c2c4 d6c7 b2c3 g8g4 c4g4 h5g4 f3h4 e6g5 g1g2 c8b7 h4f5');
INSERT INTO evaluations VALUES(13512,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(13512,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(13512,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(13512,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(13512,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(13512,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(13512,36,'cp -59','e6d5 c4a2 d5d4 d1e2 f8e7 c3e4 f6e4 e2e4 d4d3 e1g1 e8g8 c2c3 g7g6 b2b4 c8f5 e4f4 c5b4 a3b4 b7b5 c1b2 e7f6 b2a3 f8e8 f1e1 e8e1 a1e1 h7h5 a3b2 a7a5 b4a5 a8a5 a2b3 a5a8 h2h3 c6a5 g2g4 a5b3 g4f5 a8a4 f4g3 a4a2 f5g6 a2b2 g6f7 g8f7');
INSERT INTO evaluations VALUES(13740,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(13740,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(13740,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(13740,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(13740,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(13740,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(13740,36,'cp -33','c4a2 c8g4 e1g1 d5d4 c3e4 f6e4 f1e1 f8e7 e1e4 g4f5 e4e1 d4d3 c2d3 f5d3 b2b3 a7a6 c1b2 e8g8 e1e3 b7b5 h2h3 d8d5 b3b4 c5c4 a2b1 d3b1 a1b1 f8d8 d1e2 e7f8 b2c3 d8d7 b1e1 a8d8 f3e5 d7e7 e2g4 d5d6 e5c6 e7e3 e1e3 d6c6 e3e4 h7h6 e4e3');
INSERT INTO evaluations VALUES(13774,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(13774,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(13774,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(13774,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(13774,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(13774,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(13774,36,'cp -45','c8g4 e1g1 d5d4 c3e4 f6e4 f1e1 f8e7 e1e4 g4f5 e4e1 d4d3 c2d3 f5d3 b2b3 e8g8 c1b2 e7f6 b2c3 c6d4 a1c1 d8d5 a2b1 d3b1 c1b1 f8e8 c3d4 e8e1 d1e1 f6d4 f3d4 d5d4 e1e3 d4d5 b1e1 h7h6 h2h3 b7b6 e3e4 d5e4 e1e4 a8d8 g1f1 d8d2 e4e8 g8h7 e8e7 h7g6 e7a7 d2b2 b3b4 c5b4 a3b4 b2b4');
INSERT INTO evaluations VALUES(18399,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(18399,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(18399,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18399,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(18399,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(18399,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(18399,36,'cp -36','f8c5 f1d3 d7d6 e1g1 g8f6 b1c3 e8g8 h2h3 a7a5 a1b1 c8e6 f1e1 h7h6 c1e3 c5d4 e3d4 e5d4 c3e2 c6c5 f2f4 e6d7 b2b3 d8c7 e2g3 g7g6 d1d2 f8e8 a2a3 d7c6 b3b4 c5b4 a3b4 a5b4 b1b4 f6d7 d2f2 d7c5 f2d4 c5d3');
INSERT INTO evaluations VALUES(22984,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(22984,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(22984,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(22984,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(22984,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(22984,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(22984,36,'cp 22','f8c5 c3a4 c5b4 c1d2 b4d2 d1d2 d7d6 a4c3 f7f5 e4f5 c8f5 e1c1 d8h4 f2f3 e8c8 g2g3 h4h6 d2h6 g8h6 h2h4 h6g8 h4h5 g6e7 g3g4 f5d7 f1d3 d8f8 d1f1 c7c6 d5c6 d7c6 c3e4 c6e4 d3e4 g8f6 g1e2 h7h6 e2c3 c8b8 c3b5 d6d5 c4d5 e7d5 b5d6 f6e4');
INSERT INTO evaluations VALUES(6808,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(6808,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(6808,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(6808,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(6808,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(6808,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(6808,36,'cp 49','a2a3 a7a5 c3a4 c5a7 d5d6 c7d6 a4c3 a7c5 h2h4 h7h6 h4h5 g6f8 c3b5 a8a6 c1d2 g8f6 d1f3 f8e6 g1e2 a6c6 e1c1 e8g8 e2c3 g8h8 g2g4 b7b6 f1d3 c8a6 c1b1 c6c8 f3e2 a6b5 c3b5 f6h7 f2f3 d8f6 d2e1 e6d4 b5d4');
INSERT INTO evaluations VALUES(23501,30,'cp 728',NULL);
INSERT INTO evaluations VALUES(23501,31,'cp 725',NULL);
INSERT INTO evaluations VALUES(23501,32,'cp 736',NULL);
INSERT INTO evaluations VALUES(23501,33,'cp 740',NULL);
INSERT INTO evaluations VALUES(23501,34,'cp 740',NULL);
INSERT INTO evaluations VALUES(23501,35,'cp 740',NULL);
INSERT INTO evaluations VALUES(23501,36,'cp 740','c4f7 e8e7 b1c3 g8h8 c3d5 f6d5 e4d5 c6d4 c2c3 h7h6 d2d3 h6g5 c1g5 e7f7 g5d8 d4f5 d1f3 d7d6 g2g4 c8d7 d8g5 g7g6 e1c1 f8h6 g5e3 h6e3 f2e3 h8h4 g4f5 g6f5 f3f2 a8h8 b2b3 f7g7 c3c4 a7a5 d1d2 h8h5 c1b2 a5a4 f2f3 h4h3 f3e2 g7g8 d3d4 e5d4 e3d4 h5h4 e2e7');
INSERT INTO evaluations VALUES(23500,30,'cp 710',NULL);
INSERT INTO evaluations VALUES(23500,31,'cp 717',NULL);
INSERT INTO evaluations VALUES(23500,32,'cp 744',NULL);
INSERT INTO evaluations VALUES(23500,33,'cp 732',NULL);
INSERT INTO evaluations VALUES(23500,34,'cp 736',NULL);
INSERT INTO evaluations VALUES(23500,35,'cp 751',NULL);
INSERT INTO evaluations VALUES(23500,36,'cp 732','e8e7 b1c3 g8h8 c3d5 f6d5 e4d5 c6d4 c2c3 h7h6 d2d3 h6g5 c1g5 e7f7 g5d8 d4f5 d1f3 d7d6 g2g4 f7e8 d8c7 f5h4 f3e4 e8d7 c7a5 b7b6 a5b4 c8b7 e1c1 g7g5 d3d4 a8e8 h1e1 d7c8 d4e5 c8b8 f2f4 d6e5 b4f8 e8f8 f4f5 b8a8 c1b1 f8f6 e1f1 h8c8 h2h3 f6f7 f5f6');
INSERT INTO evaluations VALUES(2507,30,'cp 744',NULL);
INSERT INTO evaluations VALUES(2507,31,'cp 744',NULL);
INSERT INTO evaluations VALUES(2507,32,'cp 744',NULL);
INSERT INTO evaluations VALUES(2507,33,'cp 748',NULL);
INSERT INTO evaluations VALUES(2507,34,'cp 755',NULL);
INSERT INTO evaluations VALUES(2507,35,'cp 759',NULL);
INSERT INTO evaluations VALUES(2507,36,'cp 736','b1c3 g8h8 c3d5 f6d5 e4d5 c6d4 c2c3 h7h6 d2d3 h6g5 c1g5 e7f7 g5d8 d4f5 d1f3 d7d6 g2g4 f7e8 d8c7 f5h4 f3e4 e8d7 c7a5 b7b6 a5b4 a7a5 b4a3 d7c7 e4c4 c7b8 c4c6 b8a7 b2b4 c8b7 c6a4 a7b8 a4d7 a5a4 e1c1 b7d5 d3d4 h4f3 c1b1 e5d4 c3d4 h8h2 h1h2 f3h2');
INSERT INTO evaluations VALUES(25317,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(25317,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25317,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(25317,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(25317,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25317,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25317,36,'cp 47','c6d4 f3d1 g7g6 d2d3 f8g7 c3e2 d4e2 d1e2 e8g8 a2a4 c7c6 a4a5 d8e7 c4b3 f6d7 c1e3 d7c5 b3a2 c5e6 c2c3 e7d7 f1e1 d6d5 b2b4 f8d8 e2f3 h7h6 a2b3 h6h5 e1d1 d5d4 c3d4 e6d4 e3d4 e5d4 a1b1');
INSERT INTO evaluations VALUES(25401,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(25401,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(25401,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(25401,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(25401,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(25401,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(25401,36,'cp 37','f3d1 c7c6 c3e2 f6e4 e2d4 e5d4 f1e1 d6d5 d2d3 f8e7 c4d5 d8d5 e1e4 f7f5 e4e1 e8f7 c1f4 h7h6 c2c4 d4c3 b2c3 e7f6 a1b1 b7b6 c3c4 d5d7 d1h5 g7g6 h5f3 h6h5 c4c5 b6c5 f4e3 h8e8 e1c1 d7d5 f3d5 c6d5 e3c5 g6g5 a2a4 f7g6 b1b7');
INSERT INTO evaluations VALUES(25402,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(25402,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(25402,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(25402,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(25402,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(25402,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(25402,36,'cp 34','c7c6 c3e2 f6e4 e2d4 e5d4 f1e1 d6d5 d2d3 f8e7 c4d5 d8d5 e1e4 f7f5 e4e1 e8f7 c1f4 a8e8 f4g5 e7g5 d1h5 g7g6 h5g5 d5a5 g1f1 a5c7 g2g3 f7g7 a2a4 c6c5 g5d2 h8f8 a4a5 c7c6 f1g1 c6d6 e1e8 f8e8 a1e1 e8e1 d2e1 g7f7 e1e2 h7h5 h3h4 d6d5 e2d2 b7b5 d2h6 c5c4 h6h7 f7f6');
INSERT INTO evaluations VALUES(26985,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(26985,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(26985,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(26985,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(26985,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(26985,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(26985,36,'cp 80','d8d7 d4d5 c6e7 b1d2 d7c8 h2h3 g4d7 a2a4 g8f6 c4d3 e7g6 d2f1 f8e7 f1g3 e8g8 e1g1 a7a5 c3c4 h7h6 a1a3 c8d8 d3c2 f8e8 f1e1 g6f4 c1e3 b7b6 g3f5 e7f8 d1d2 f4g6 f3h2 d7f5 e4f5 g6f4 e3f4 e5f4 e1e8 d8e8 d2f4 f6d7 f4d2');
INSERT INTO evaluations VALUES(32578,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(32578,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(32578,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(32578,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(32578,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(32578,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(32578,36,'cp 69','f3e5 g4e2 d1e2 e7e6 a2a3 f8e7 c1d2 e8g8 e1c1 a5a6 e2f3 b8d7 h2h4 a8d8 d2g5 h7h6 c1b1 a6a5 e5c4 a5f5 f3f5 e6f5 g5f4 f6g4 d1d2 f8e8 d4d5 d7b6 c4b6 a7b6 f4c7 d8d7 c7b6 e7d8 b6d8 e8d8 h4h5 c6d5 h1d1 d5d4');
INSERT INTO evaluations VALUES(36477,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36477,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36477,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36477,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(36477,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36477,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36477,36,'cp 10','c7c5 g1f3 h7h6 g5c1 e7e6 e1g1 b8c6 d4d5 c6d4 d5e6 c8e6 f3d4 c5d4 c3b5 a8c8 d1d4 a7a6 b5d6 f6e4 d4e4 d8d6 e4b7 e6c4 e2c4 c8c4 a1b1 f8b8 b7f3 d6e6 c1f4 b8b2 b1b2 g7b2 f4h6 e6e4 f3e4 c4e4 f1b1 b2e5 b1d1 e4e2 h6e3 e2a2');
INSERT INTO evaluations VALUES(42305,30,'cp -164',NULL);
INSERT INTO evaluations VALUES(42305,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(42305,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(42305,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(42305,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(42305,35,'cp -121',NULL);
INSERT INTO evaluations VALUES(42305,36,'cp -119','d7d5 g4h5 d8e7 f1b5 c7c6 b5e2 f6d4 g1f3 d4f2 e1f1 h8g8 b1c3 b8d7 d2d4 d7f6 h5e5 e7e5 d4e5 e4c3 b2c3 f6e4 f3d4 f2h4 e2d3 c8g4 c1e3 e8c8 d4f3 h4e7 c3c4 c8b8 c4d5 c6d5 a2a4 d8c8 a4a5 h7h5 a5a6 b7b6 e3f4 g4e6 f3d4 h5h4 d4e6 f7e6 f1e2 g8g2 e2f3 g2g6 h1f1');
INSERT INTO evaluations VALUES(41574,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(41574,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(41574,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(41574,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(41574,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(41574,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(41574,36,'cp -106','g4h5 d8e7 f1b5 c7c6 b5e2 f6d4 g1f3 d4f2 e1f1 h8g8 d2d4 b8d7 b1c3 d7f6 h5e5 c8e6 c3e4 f6e4 e2d3 f7f6 e5h5 e8d8 d3e4 d5e4 f1f2 e4f3 g2g3 e6f7 h5a5 b7b6 a5d2 e7d7 d2d3 g8e8 h2h3 f7g6 d3f3 d7e6 h1f1 e6h3 f1h1 h3e6 c1f4 g6e4 f3e3 e4h1 e3e6 e8e6 a1h1 e6e7 h1h6 d8d7');
INSERT INTO evaluations VALUES(44191,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(44191,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44191,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44191,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(44191,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(44191,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(44191,36,'cp 7','b8c6 f1e2 d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 d5c4 d4c6 d8d1 c3d1 b7c6 e2c4 f6d5 c1d2 c8e6 e1g1 f8e7 f1e1 e8g8 d1c3 e7c5 c4b3 d5c3 d2c3 e6b3 a2b3 f8b8 e1e5 b8b5 a1e1 f7f6 e5e6 b5b3 e6c6 c5b6 h2h3 a8d8 e1e2 g8f7');
INSERT INTO evaluations VALUES(46944,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(46944,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(46944,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(46944,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(46944,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(46944,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(46944,36,'cp 113','g1f3 g7g6 f1e2 f8g7 b1c3 e8g8 h2h3 e7e5 f3e5 b8c6 e5c6 b7c6 c4c5 b6d5 c3d5 d8d5 c1e3 c8e6 e1g1 a7a5 b2b3 f8d8 e2c4 d5e4 c4e6 e4e6 f1e1 e6d5 d1c2 d8b8 a1d1 b8e8 f2f3 e8e7 g1f1 a8b8 e3f2 e7e1 d1e1 a5a4 c2c4 d5c4 b3c4 a4a3');
INSERT INTO evaluations VALUES(46943,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(46943,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(46943,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(46943,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(46943,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(46943,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(46943,36,'cp 81','g7g6 b1c3 f8g7 f1e2 e8g8 h2h3 c7c5 d4d5 e7e6 c1g5 g7c3 b2c3 f7f6 g5e3 b8a6 e1g1 e6d5 c4d5 d8d5 f3d2 d5f7 e2a6 b7a6 d2e4 c8b7 e4d6 f7d5 d6b7 d5b7 e3c5 f8e8 c5d4 b7c6 d1d2 e8e6 f1e1 a8e8 e1e6 e8e6 a1b1 e6e7 d2f4 b6d7 f4h4 a6a5 h4h6 a5a4 d4a7 c6c3');
INSERT INTO evaluations VALUES(46618,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(46618,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(46618,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(46618,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(46618,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(46618,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(46618,36,'cp 87','b1c3 f8g7 f1e2 e8g8 h2h3 c7c5 d4d5 e7e6 c1g5 g7c3 b2c3 f7f6 g5e3 b8a6 e1g1 e6d5 c4d5 d8d5 f3d2 d5f7 f1e1 c8d7 e2a6 b7a6 e3c5 f8e8 d1b3 d7a4 b3b4 a4c6 c3c4 e8e1 a1e1 a8e8 c5e3 f7e7 b4a5 e7b7 f2f3 g8f7 a5c3 b6d7 d2b3 b7c7 a2a3 c7e5 c3c1');
INSERT INTO evaluations VALUES(46617,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(46617,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(46617,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(46617,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(46617,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(46617,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(46617,36,'cp 99','f8g7 f1e2 e8g8 h2h3 c7c5 d4d5 e7e6 c1g5 g7c3 b2c3 f7f6 g5e3 b8a6 e1g1 e6d5 c4d5 d8d5 f3d2 d5f7 e2a6 b7a6 e3c5 f8e8 c5d4 c8b7 f2f3 a8d8 f1e1 b7d5 e1e8 d8e8 d2f1 h7h5 f1e3 d5e6 d1f1 b6c4 a1e1 e8c8 a2a4 e6d7 f1f2 c4e3 f2e3 c8e8');
INSERT INTO evaluations VALUES(41177,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(41177,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(41177,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(41177,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(41177,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(41177,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(41177,36,'cp 116','f1e2 e8g8 h2h3 c7c5 d4d5 e7e6 c1g5 g7c3 b2c3 f7f6 g5e3 b8a6 e1g1 e6d5 c4d5 d8d5 f3d2 d5f7 e2a6 b7a6 e3c5 f8e8 c5d4 c8b7 f2f3 b7d5 d2e4 d5e4 f3e4 b6d7 d1a4 e8e6 a1d1 e6e4 a4a6 e4e6 a6b7 d7b6 b7f3 a8f8 f3f2 f7e8 f1e1 e6e1 d1e1 e8c6 a2a4 b6c4 f2a2');
INSERT INTO evaluations VALUES(57292,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57292,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57292,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(57292,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57292,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57292,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57292,36,'cp -5','d7d5 c2c4 c7c6 b1c3 g8f6 f3e4 f6e4 c3e4 d5e4 d1c2 f8b4 c1d2 b4d2 c2d2 c8e6 b2b3 d8h4 g2g3 h4e7 g1h3 e6g4 f1g2 b8d7 h3f2 d7f6 f2g4 f6g4 e1g1 e8g8 g2h3 g4h6 d4d5 a8d8 a1d1 d8d6 d2d4 f8d8 d4a7 c6d5 d1d5 d6d5 c4d5 d8d5 a7b8 d5d8');
INSERT INTO evaluations VALUES(56231,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56231,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(56231,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(56231,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(56231,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56231,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(56231,36,'cp 0','c2c4 c7c6 b1c3 g8f6 f3e4 f6e4 c3e4 d5e4 d1c2 f8b4 c1d2 b4d2 c2d2 c8e6 b2b3 a7a5 g1e2 a5a4 a1d1 a4b3 a2b3 e6g4 d1b1 e8g8 e2g3 b8d7 f1e2 g4e2 d2e2 f7f5 e1g1 g7g6 e2d2 d8a5 d2b2 a5a3 b2d2');
INSERT INTO evaluations VALUES(11696,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(11696,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11696,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(11696,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(11696,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(11696,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11696,36,'cp 28','d7d6 c2c3 f8e7 h2h3 c8b7 b1d2 e8g8 e1g1 f8e8 f1e1 e7f8 a2a4 h7h6 d4d5 c6b8 a4b5 a6b5 a1a8 b7a8 d2f1 b8d7 f3h2 c7c6 h2g4 c6d5 e4d5 a8b7 f1e3 f6g4 d1g4 d7f6 g4f3 b7c8 e3f1 c8b7 f3d3 d8d7 f1g3 e8a8 c1e3 g7g6 g3e4 f6e4 d3e4');
INSERT INTO evaluations VALUES(11628,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11628,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(11628,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(11628,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(11628,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11628,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11628,36,'cp 37','c2c3 f8e7 h2h3 e8g8 e1g1 c8b7 f1e1 f8e8 b1d2 e7f8 b3c2 g7g6 a2a3 c6b8 b2b4 b8d7 c1b2 c7c5 b4c5 d6c5 d4e5 d7e5 f3e5 e8e5 c3c4 e5e6 e4e5 f8h6 b2c3 f6d7 c2e4 b7e4 d2e4 h6f4 d1f3 f4e5 e4c5 d7c5 e1e5 e6e5 c3e5 b5c4');
INSERT INTO evaluations VALUES(11973,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(11973,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(11973,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(11973,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(11973,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(11973,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(11973,36,'cp -28','d7d5 f1d3 h7h6 g5d2 f8d6 d1e2 e6e5 e1g1 e8g8 g1h1 f8e8 a1d1 c8e6 f2f3 d8e7 b2b3 a8d8 e2f2 d6b4 f2e1 e6c8 e1h4 c8d7 c3a4 b4d2 d1d2 d7c8 h4f2 f6h5 f1d1 h5f4 d3f1 f7f5 f2c5 f5e4 c5e7 e8e7 f3e4');
INSERT INTO evaluations VALUES(16050,30,'cp -562',NULL);
INSERT INTO evaluations VALUES(16050,31,'cp -570',NULL);
INSERT INTO evaluations VALUES(16050,32,'cp -585',NULL);
INSERT INTO evaluations VALUES(16050,33,'cp -579',NULL);
INSERT INTO evaluations VALUES(16050,34,'cp -579',NULL);
INSERT INTO evaluations VALUES(16050,35,'cp -600',NULL);
INSERT INTO evaluations VALUES(16050,36,'cp -617','c6d4 f3d1 d7d5 c4e2 h7h5 c2c3 d4e2 g1e2 c8g4 d2d4 f6e4 b1d2 f8g7 d2e4 d5e4 d1a4 c7c6 c1e3 e8g8 h1g1 e5d4 c3d4 d8d7 a4c4 c6c5 a1d1 c5d4 d1d4 g4e2 g1g6 e2c4 g6g7 g8g7 d4d7 b7b6 b2b3 c4e6 e3d4 g7h7 d7d6 h5h4 d4b2 b6b5');
INSERT INTO evaluations VALUES(16077,30,'cp -625',NULL);
INSERT INTO evaluations VALUES(16077,31,'cp -609',NULL);
INSERT INTO evaluations VALUES(16077,32,'cp -607',NULL);
INSERT INTO evaluations VALUES(16077,33,'cp -600',NULL);
INSERT INTO evaluations VALUES(16077,34,'cp -606',NULL);
INSERT INTO evaluations VALUES(16077,35,'cp -605',NULL);
INSERT INTO evaluations VALUES(16077,36,'cp -578','f3d1 d7d5 c4e2 h7h5 c2c3 d4e2 g1e2 c8g4 d2d4 f6e4 b1d2 f8g7 d2e4 d5e4 h2h3 g4f3 h1g1 d8d7 d4e5 d7h3 c1e3 e8g8 d1b3 g8h7 e5e6 f7e6 a1d1 a8d8 d1d8 f8d8 e2f4 h3h2 g1f1 h5h4 f4e6 d8d1 b3d1 f3d1 e1d1 h2d6 d1c1 d6e6 a2a3 e6c4 f1g1 b7b5');
INSERT INTO evaluations VALUES(18483,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(18483,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(18483,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(18483,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(18483,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(18483,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(18483,36,'cp -15','c6d5 c1f4 f8d6 f1b5 e8e7 d1d2 d6f4 d2f4 d8d6 f4d6 e7d6 e1c1 a7a6 b5e2 g8f6 h1e1 c8d7 f2f3 h7h5 b2b3 h8c8 c1b2 d6e7 g2g4 h5g4 f3g4 g7g5 e2f3 e7d6 h2h4 c8g8 h4g5 g8g5 c3e4 f6e4 e1e4 a8b8 c2c4 f7f5 g4f5 g5f5');
INSERT INTO evaluations VALUES(18514,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(18514,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(18514,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(18514,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(18514,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(18514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18514,36,'cp -15','c1f4 f8d6 f1b5 e8e7 d1d4 d6f4 d4f4 d8d6 f4d6 e7d6 e1c1 g8f6 h1e1 c8b7 b2b3 h7h5 h2h4 h8c8 c1b2 a8b8 b5f1 d6e7 f2f3 g7g6 d1d2 a7a5 a2a4 c8c5 c3b5 b7a6 c2c3 c5c8 b5a7 c8c7 a7b5 c7c5 f1d3 f6d7 b3b4 c5c8');
INSERT INTO evaluations VALUES(18587,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(18587,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(18587,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(18587,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(18587,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(18587,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(18587,36,'cp 96','g8f6 f1e2 g7g6 g1f3 f8g7 e1g1 e8g8 c1e3 a7a6 h2h3 b7b5 d1c2 b5c4 e2c4 d6d5 c4b3 d5e4 c3e4 f6e4 c2e4 a6a5 e4c6 d7b6 e3f4 a8a7 c6c5 a5a4 b3c4 a7d7 f3e5 g7e5 f4e5 b6c4 c5c4 d8a5 f1c1 c8b7 f2f3 a5a6 c4a6');
INSERT INTO evaluations VALUES(12258,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(12258,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(12258,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(12258,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(12258,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(12258,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(12258,36,'cp 101','g1f3 g7g6 f1e2 f8g7 e1g1 e8g8 c1e3 f6g4 e3d2 a7a5 h2h3 g4f6 d1c2 e7e5 d4e5 d7e5 f3e5 d6e5 d2e3 c8e6 c3a4 f6d7 a1d1 d8e7 c4c5 f7f5 e2c4 e6c4 c2c4 e7f7 c4d3 a8d8 f2f3 f5f4 e3f2 g6g5');
INSERT INTO evaluations VALUES(12257,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(12257,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(12257,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(12257,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(12257,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(12257,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(12257,36,'cp 91','g7g6 c1e3 f8g7 f1e2 f6g4 e3c1 g4f6 e1g1 e8g8 c1e3 f6g4 e3d2 e7e5 d4d5 a7a5 h2h3 g4f6 d2e3 d7c5 d1c2 d8c7 a1c1 f6h5 f1e1 c8d7 c1d1 h5f4 e2f1 c6d5 e3f4 d5e4 c3d5 c7d8 f4g5 f7f6 f3e5 d7f5 g5e3 f6e5 g2g4 f5d7 d5c3');
INSERT INTO evaluations VALUES(18971,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(18971,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(18971,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(18971,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(18971,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(18971,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(18971,36,'cp 107','b1c3 d7d6 g1e2 g8f6 f2f4 f8e7 e1g1 a7a6 a2a4 c8d7 e2g3 e5f4 c1f4 e8g8 d1d2 c6d4 a4a5 d4c6 c3d5 f6d5 e4d5 c6e5 f4e5 d6e5 g3e4 d8c7 d2e2 e7d6 a1e1 f7f6 g2g4 g8h8 f1f2 d7e8 h2h4 e8g6 e1a1 a8d8 c2c3 f6f5 e4g5 f5f4 g5e6 c7e7 e6f8');
INSERT INTO evaluations VALUES(21368,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(21368,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21368,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(21368,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(21368,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(21368,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(21368,36,'cp 17','e6d5 d2d4 g8f6 f1b5 f8b4 e1g1 e8g8 c3e2 b4d6 e2g3 f8e8 c2c3 a7a6 b5d3 h7h6 c1d2 c6a5 b2b3 a5c6 f1e1 e8e1 d1e1 d8e7 e1e7 c6e7 c3c4 d5c4 b3c4 c8d7 a1e1 b7b6 f3e5 d7e8 h2h3 a6a5 d2f4');
INSERT INTO evaluations VALUES(27676,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(27676,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(27676,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(27676,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(27676,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(27676,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(27676,36,'cp 58','g8f6 f1e1 a7a5 c2c4 e7g6 h2h3 f8e7 b1c3 e8g8 c1e3 b7b6 f3d2 f6e8 a1c1 g6h4 b2b3 f7f5 e4f5 d7f5 d1g4 f5f7 g4e6 e8f6 e6f7 g8f7 g2g3 h4f5 c3b5 f8c8 d2f3 h7h6 e3d2 f6d7 g1g2 d7c5 c1c3 h6h5 c3c2 e5e4 f3g5 e7g5');
INSERT INTO evaluations VALUES(27475,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(27475,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(27475,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(27475,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(27475,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(27475,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(27475,36,'cp 46','f1e1 a7a5 h2h3 h7h6 c2c4 e7g6 b1d2 f8e7 d2f1 a5a4 f1g3 e8g8 c1e3 c7c6 d1d3 f8b8 b2b4 b7b5 d5c6 d7c6 a1c1 e7f8 a2a3 b5c4 c1c4 c6b5 d3c2 b8d8 g3f5 d6d5 e4d5 b5d5 e3c5 e5e4 f3d4 g6e5 c4c3 e5d3');
INSERT INTO evaluations VALUES(31798,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(31798,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(31798,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(31798,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(31798,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(31798,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(31798,36,'cp 175','b1c3 c7c6 e2e4 e6g4 c1e3 g8f6 f1c4 e7e6 d1b3 b7b5 c4d3 b8d7 e4e5 f6d5 c3d5 c6d5 b3b5 f8e7 e1g1 e8g8 b2b3 a8b8 b5a4 a7a5 f3d2 e7b4 a2a3 b4c3 a1b1 d8e7 h2h3 g4f5 d3f5 e6f5 d2f3 d7b6 a4c6 f8c8 c6d6');
INSERT INTO evaluations VALUES(31792,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(31792,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(31792,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(31792,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(31792,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(31792,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(31792,36,'cp 169','c7c6 e2e4 e6g4 c1e3 e7e6 f1c4 g8f6 d1b3 b7b5 c4d3 b8d7 e4e5 f6d5 c3d5 c6d5 b3b5 f8e7 b2b3 e8g8 e1g1 a8b8 b5c6 g4f3 g2f3 b8c8 c6a4 c8c3 f1d1 f7f6 a4a6 d8b6 a6b6 d7b6 f3f4 e7a3 f4f5 f6e5 d4e5 e6f5 g1h1 f5f4');
INSERT INTO evaluations VALUES(34673,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(34673,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(34673,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(34673,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34673,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(34673,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(34673,36,'cp 11','b1c3 e7e5 g1f3 g8f6 d2d3 b8c6 c1d2 a5b4 f1e2 b4d6 d2e3 f8e7 e1g1 f6g4 e3d2 g4f6 f1e1 e8g8 h2h3 f8e8 d2e3 c8f5 d3d4 e5d4 f3d4 c6d4 d1d4 c7c6 g2g4 f5g6 d4d6 e7d6 a1d1 d6e5 g1g2 h7h6 e2f3 e5c3 b2c3');
INSERT INTO evaluations VALUES(34672,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(34672,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(34672,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(34672,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(34672,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(34672,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(34672,36,'cp 5','e7e5 g1f3 b8c6 d2d3 g8f6 c1d2 a5c5 f1e2 c5d6 e1g1 f8e7 h2h3 e8g8 f1e1 f8e8 e2f1 e7f8 c3b5 d6d8 d2g5 a7a6 b5c3 f8c5 c3e4 c5b4 e1e2 c8f5 d3d4 f5e4 e2e4 e5d4 e4e8 d8e8 g5f6 g7f6 f1d3 b4d6 a2a3 g8g7');
INSERT INTO evaluations VALUES(43349,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(43349,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(43349,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(43349,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(43349,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(43349,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(43349,36,'cp 65','c1e3 f8g7 d1d2 c8d7 f2f3 h7h5 e1c1 b8c6 h2h3 h5h4 c1b1 b7b5 h1e1 a8c8 d4b3 d7e6 e3g5 f6d7 c3d5 d7b6 c2c3 b6a4 g5e3 d8d7 b3d4 c6d4 e3d4 e8f8 d5b4 a6a5 b4c2 d7b7 f3f4 g7d4 c2d4 b5b4 c3b4 b7e4 b1a1');
INSERT INTO evaluations VALUES(43342,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(43342,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(43342,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(43342,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(43342,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(43342,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(43342,36,'cp 78','f8g7 d1d2 b8d7 f2f3 d8c7 g2g4 b7b5 a2a4 b5b4 c3a2 c8b7 a2b4 e8g8 a4a5 e7e6 e1g1 f8e8 c2c3 d7c5 a1a3 a8b8 f1b1 h7h5 g4g5 f6d7 e2f1 e8c8 b1d1 g7e5 b4d3 c5d3 f1d3');
INSERT INTO evaluations VALUES(39759,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(39759,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(39759,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(39759,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(39759,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(39759,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(39759,36,'cp 80','d1d2 b8c6 e1c1 c8d7 c1b1 a8c8 f2f3 e8g8 h2h4 b7b5 d4c6 d7c6 e3h6 b5b4 h6g7 g8g7 c3d5 c6d5 e4d5 c8c5 h4h5 c5d5 d2b4 a6a5 b4c3 d5d1 e2d1 d8b6 h5h6 g7g8 h1e1 b6a7 a2a4 f8b8 g2g4 e7e5 d1e2');
INSERT INTO evaluations VALUES(39761,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(39761,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(39761,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(39761,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(39761,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(39761,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(39761,36,'cp 64','b8c6 e1c1 c8d7 f2f3 h7h5 h1e1 e8g8 h2h3 b7b5 c3d5 h5h4 c1b1 a8c8 d4c6 d7c6 d5f6 g7f6 e3g5 f6g5 d2g5 e7e5 g5d2 g8g7 d2d6 d8d6 d1d6 f7f5 e2d3 f8f6 d6f6 g7f6 c2c3 c6d7 e1d1 a6a5 b1c2');
INSERT INTO evaluations VALUES(7100,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(7100,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(7100,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(7100,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7100,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(7100,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(7100,36,'cp 96','e1c1 c8d7 f2f3 h7h5 h1e1 b7b5 c3d5 a8c8 c1b1 e8g8 h2h3 f6d5 e4d5 c6d4 e3d4 d7f5 d1c1 g7d4 d2d4 d8c7 e2d3 f5d3 d4d3 c7b7 f3f4 c8c5 c1d1 f8c8 f4f5 g6g5 e1e2 f7f6 d1d2 b5b4 d3f3 h5h4 e2e6 g8f8');
INSERT INTO evaluations VALUES(43785,30,'cp -191',NULL);
INSERT INTO evaluations VALUES(43785,31,'cp -186',NULL);
INSERT INTO evaluations VALUES(43785,32,'cp -181',NULL);
INSERT INTO evaluations VALUES(43785,33,'cp -178',NULL);
INSERT INTO evaluations VALUES(43785,34,'cp -180',NULL);
INSERT INTO evaluations VALUES(43785,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(43785,36,'cp -174','a7a5 b4b5 b6c5 c1a3 b8d7 d4c5 f8c5 a3c5 d7c5 g1f3 e8g8 f1e2 c7c6 b5c6 d8d6 a2a4 d6c6 b1c3 c8a6 c3b5 f6e4 e1g1 f8c8 a1c1 c6b6 f3d4 e4d6 b5d6 b6d6 e2a6 d6a6 d4b5 h7h6 d1d4 c5b3 c1c8 a8c8 d4b2 b3c5 b2a3 c5e4 f1c1 e4f6');
INSERT INTO evaluations VALUES(43699,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(43699,31,'cp -166',NULL);
INSERT INTO evaluations VALUES(43699,32,'cp -182',NULL);
INSERT INTO evaluations VALUES(43699,33,'cp -168',NULL);
INSERT INTO evaluations VALUES(43699,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(43699,35,'cp -182',NULL);
INSERT INTO evaluations VALUES(43699,36,'cp -183','b4b5 b6c5 c1a3 b8d7 d4c5 f8c5 a3c5 d7c5 f1e2 c7c6 b5c6 d8b6 g1f3 e8g8 b1c3 b6c6 e1g1 c8a6 f3d4 c6d7 e2a6 a8a6 a1c1 a6a8 d1e2 f8c8 e2b5 a8b8 b5d7 f6d7 c3b5 c5e4 b5a7 c8c5 d4b3 c5c3 c1c3 e4c3 b3a5');
INSERT INTO evaluations VALUES(46185,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(46185,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(46185,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(46185,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(46185,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(46185,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(46185,36,'cp -64','c5d4 f3d4 e6e5 d4f3 d5d4 f3e5 f8e7 c3d5 f6d5 c4d5 d8a5 c1d2 a5d5 e5f3 b8c6 d2c3 e7c5 e2e3 c8g4 e3d4 e8c8 f1e2 g4f3 e2f3 h8e8 e1f1 d5c4 f3e2 c4e6 d1c2 c5d4 a1e1 c8b8 h2h4 e6f6 e2f3 e8e1 c3e1 c6e5 h1h3 e5f3 h3f3 f6h4 f1g1 d4e5 c2c5 h4h2 g1f1 e5b2');
INSERT INTO evaluations VALUES(46317,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(46317,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(46317,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(46317,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(46317,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(46317,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(46317,36,'cp -50','f3d4 e6e5 d4f3 d5d4 f3e5 f8e7 c3d5 f6d5 c4d5 d8a5 c1d2 a5d5 e5d3 c8f5 d3f4 d5d7 e2e3 e8g8 f1c4 b8c6 e1g1 e7g5 f4e2 g5f6 e3d4 c6d4 d2c3 d4e2 d1e2 f6c3 b2c3 d7c6 e2e5 c6c4 e5f5 b7b6 f5f3 a8c8 f1d1 h7h6 g2g3 c4a4 d1e1 a4b3 f3b7 c8a8 e1e3 b3b2 a1d1 b2a3 d1d7 a3a1 g1g2');
INSERT INTO evaluations VALUES(47138,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47138,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47138,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47138,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47138,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47138,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47138,36,'cp 0','d5c4 d1a4 b8c6 f1g2 f6d7 a4c4 e7e5 e1g1 f8e7 b2b3 e8g8 c1b2 a7a5 c4c2 c6b4 c2c4');
INSERT INTO evaluations VALUES(48688,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(48688,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(48688,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(48688,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(48688,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(48688,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(48688,36,'cp -36','f8d6 b1c3 c7c6 g1f3 e8g8 e1g1 f8e8 f3e5 b8d7 e5d7 c8d7 c1g5 d6e7 a1c1 h7h6 g5f6 e7f6 e2e3 f6e7 a2a3 g7g6 h2h4 d7f5 f1e1 d8d7 c3a4 b7b6 b2b4 a8c8 a4c3 a7a5 d1b3 a5b4 a3b4 f5e6 b4b5 c6b5');
INSERT INTO evaluations VALUES(49518,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(49518,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(49518,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(49518,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(49518,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(49518,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(49518,36,'cp -84','h7h6 g5f3 f6e4 d2d3 e4f6 f3e5 d7d5 f1e2 f8d6 c1f4 c7c5 d3d4 e8g8 d4c5 d6c5 e1g1 b8c6 b1c3 f8e8 e5c6 b7c6 h2h3 c8f5 e2d3 f6e4 d1f3 e4c3 b2c3 f5e4 d3e4 d5e4 f3h5 c5b6 f4e3 e8e6 a1d1 d8c7 h5f5 c7e7 f1e1 a8e8 e3b6 a7b6');
INSERT INTO evaluations VALUES(50675,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(50675,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(50675,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(50675,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(50675,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50675,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(50675,36,'cp 0','g8f6 b1c3 e7e6 d2d4 c5d4 f3d4 f8e7 e1g1 b8c6 c1e3 e8g8 h2h3 c8d7 f2f4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 d8c7 d4b3 c6b4 c2c3 b4d5 e4d5 e6d5 d1d5 b7b5 d5e5 c7e5 f4e5 b5a4 b3c5 e7c5 e3c5 f8c8 c5d6');
INSERT INTO evaluations VALUES(51763,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(51763,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(51763,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(51763,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(51763,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(51763,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(51763,36,'cp 134','d4d5 d7d6 f1d3 g8f6 e1g1 f8e7 c2c4 e6e5 b1c3 f6h7 f3e1 b8d7 g2g3 d7f6 a2a4 c8h3 e1g2 h6h5 a4a5 g7g6 f1e1 a7a6 d3f1 e8g8 f2f3 h3g2 f1g2 b7b6 a5b6 d8b6 c3a4 b6d8 h2h4 f8e8 g2h3 a6a5');
INSERT INTO evaluations VALUES(53528,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53528,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(53528,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53528,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(53528,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53528,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53528,36,'cp 0','b5d3 g8f6 g1f3 f8d6 b1d2 b8d7 e1g1 e8g8 f1e1 c6c5 c2c3 h7h6 e3e4 d5e4 d2e4 f6e4 e1e4 c5d4 c1h6 g7h6 e4d4 d6b8 d4g4 g8h8 d1d2 d8f6 d3g6 h8g7 g6c2 g7h8');
INSERT INTO evaluations VALUES(53599,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53599,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53599,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53599,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53599,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53599,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53599,36,'cp 0','g8f6 g1f3 c6c5 b2b3 b7b6 e1g1 c8b7 c1b2 b8c6 c2c4 c5d4 e3d4 f8e7 f1e1 a8c8 a2a3 e8g8 b1c3 d5c4 b3c4 c6a5 f3e5 a5c6');
INSERT INTO evaluations VALUES(45676,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45676,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45676,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45676,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45676,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45676,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45676,36,'cp 1','g1f3 c6c5 b2b3 b7b6 e1g1 c8b7 c1b2 b8c6 b1d2 f8e7 f3e5 e8g8 e5c6 b7c6 d1e2 c6b7 d2f3 a8c8 c2c4 c5d4 f3d4 e7b4 f1d1 d8e7 a2a3 d5c4 d3c4 b4c5 b3b4 c5d6 a1c1 a7a5 d4b5 d6b8 c4d3 a5b4');
INSERT INTO evaluations VALUES(45677,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(45677,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(45677,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(45677,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45677,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45677,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(45677,36,'cp 7','c6c5 b2b3 b7b6 e1g1 c8b7 b1d2 f8e7 c1b2 b8c6 f3e5 e8g8 e5c6 b7c6 d1e2 c6b7 d4c5 e7c5 a1d1 d8e7 e3e4 d5e4 d2e4 b7e4 d3e4 f6e4 e2e4 a8d8 e4g4 f7f6 d1d8 f8d8 f1d1 g8f7 g1f1 e6e5 d1d8 e7d8 g4e2 d8d5 a2a4');
INSERT INTO evaluations VALUES(54172,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(54172,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54172,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54172,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(54172,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54172,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(54172,36,'cp -9','c5d4 f3d4 b8d7 d1a4 d8b6 d4b3 g8f6 c4d5 f6d5 f4d2 f8e7 e2e4 d5b4 b1a3 a7a6 f1e2 e8g8 e1g1 d7e5 a4a5 b6a5 b3a5 b7b5 d2c3 b4c6 f1d1 c6a5 c3e5 c8b7 f2f3 f8c8 d1d7 e7c5 g1f1 f7f6 e5g3 b7c6 d7c7 c8c7 g3c7 a5b7 a3c2 g8f7 c2e1 a8c8 c7g3 f7e7 e1d3 c5d6 g3d6 e7d6 a1c1');
INSERT INTO evaluations VALUES(56107,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(56107,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(56107,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(56107,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(56107,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(56107,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(56107,36,'cp -30','c7c6 b5e2 d5e4 f3e5 d8g5 e5g4 f7f5 d2d4 g5g6 g4e3 g8f6 g2g3 c8e6 e3g2 g6f7 g2f4 e6c4 e2c4 f7c4 c2c3 b8d7 b1d2 c4f7 d1b3 d7b6 a2a4 f7b3 d2b3 e8f7 a4a5 b6c4 f4g2 f8e7 g2e3 c4e3 c1e3 f6d5 h1f1 h7h5 h2h4 b7b5 e1e2 a8d8 e3d2 f7e6 d2e3 g7g6 b3d2');
INSERT INTO evaluations VALUES(53680,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(53680,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(53680,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(53680,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(53680,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(53680,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53680,36,'cp -21','b5e2 d5e4 f3e5 d8g5 e5g4 f7f5 d2d4 g5g6 g4e3 h7h5 g2g3 h5h4 e3c4 b8d7 c1f4 g8f6 b1c3 h4g3 f2g3 f8b4 d1d2 f6d5 a2a3 b4c3 b2c3 e8g8 h2h4 d5f4 d2f4 b7b5 c4e5 d7e5 d4e5 c8e6 h4h5 g6h6 f4h6 g7h6 a3a4 a7a5 e1d2 b5a4 a1a4 f8d8 d2e3 d8d5');
INSERT INTO evaluations VALUES(57592,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(57592,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(57592,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(57592,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(57592,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(57592,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(57592,36,'cp 25','d7d6 c1g5 c7c5 g2g3 b8c6 f1g2 g7g6 d1d2 h7h6 g5e3 g8f6 b1c3 f5e4 c3e4 f6g4 g1e2 f8g7 e1g1 e8g8 c2c3 g4e3 f2e3 c8f5 a1e1 g6g5 d3d4 c5d4 e3d4 d6d5 e4f2 e7e6 g3g4 f5g6 e2g3 d8d6 g3h5 a8e8 h5g7 g8g7 f2d3 f8f1 e1f1 b7b6 f1e1 d6d7 d3e5 c6e5 e1e5 g6f7 d2e2 f7g6');
INSERT INTO evaluations VALUES(419,30,'cp 863',NULL);
INSERT INTO evaluations VALUES(419,31,'cp 863',NULL);
INSERT INTO evaluations VALUES(419,32,'cp 855',NULL);
INSERT INTO evaluations VALUES(419,33,'cp 867',NULL);
INSERT INTO evaluations VALUES(419,34,'cp 862',NULL);
INSERT INTO evaluations VALUES(419,35,'cp 878',NULL);
INSERT INTO evaluations VALUES(419,36,'cp 873','d1h5 g7g6 h8g6 h7g6 h5g6 e8d8 d2d3 e7f6 g6f6 e4f6 c1g5 f2c5 g5f6 c5e7 f6e7 c6e7 b1d2 d7d5 c4b3 c8e6 a1e1 d8d7 e1e5 a8f8 d2f3 d7d6 f1e2 c7c5 h1e1 e7g6 e5g5 g6f4 e2d2 f8f6 a2a4 a7a6 b3a2 d5d4 a2e6 f4e6 g5g4 b7b5 a4a5 f6f5 h2h4 e6f4');
INSERT INTO evaluations VALUES(417,30,'cp 867',NULL);
INSERT INTO evaluations VALUES(417,31,'cp 863',NULL);
INSERT INTO evaluations VALUES(417,32,'cp 867',NULL);
INSERT INTO evaluations VALUES(417,33,'cp 865',NULL);
INSERT INTO evaluations VALUES(417,34,'cp 873',NULL);
INSERT INTO evaluations VALUES(417,35,'cp 858',NULL);
INSERT INTO evaluations VALUES(417,36,'cp 871','g7g6 h8g6 h7g6 h5g6 e8d8 d2d3 e7f6 g6f6 e4f6 c1g5 f2c5 g5f6 c5e7 f6e7 c6e7 b1d2 b7b5 c4b5 a7a6 b5a4 e7f5 f1f2 d7d6 g2g4 f5d4 h2h4 c8b7 h4h5 d8e7 f2e3 d4e6 h5h6 e6f4 d2e4 f4d5 e3d2 a8h8 g4g5 d5b6 b2b3 b7e4 d3e4 b6a4');
INSERT INTO evaluations VALUES(420,30,'cp 877',NULL);
INSERT INTO evaluations VALUES(420,31,'cp 878',NULL);
INSERT INTO evaluations VALUES(420,32,'cp 881',NULL);
INSERT INTO evaluations VALUES(420,33,'cp 873',NULL);
INSERT INTO evaluations VALUES(420,34,'cp 886',NULL);
INSERT INTO evaluations VALUES(420,35,'cp 867',NULL);
INSERT INTO evaluations VALUES(420,36,'cp 882','h8g6 h7g6 h5g6 e8d8 d2d3 e7f6 g6f6 e4f6 c1g5 f2c5 g5f6 c5e7 f6e7 c6e7 b1d2 b7b5 c4b5 a8b8 b5c4 e7f5 f1f2 f5d4 a1c1 d7d6 d2f3 d4f3 f2f3 c8b7 f3g3 d6d5 c4b3 d8e7 h2h4 b8g8 g3f2 c7c5 h1e1 e7d6 h4h5 g8f8 f2g3 f8g8 g3f3 g8g5');
INSERT INTO evaluations VALUES(2586,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(2586,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2586,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(2586,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(2586,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2586,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2586,36,'cp 12','c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d3d4 e5d4 e3d4 e6e5 d4e5 c6e5 d1d4 f8e8 a1d1 d8e7 f3e5 d6e5 d4e3 a8d8 c3d5 f6d5 e4d5 e5e4 c2c4 c7c6 f1f4 c6d5 c4d5 e7g5 d5d6 g5e5 d1d4 e5c5 e3c3 c5b6 g1h2 d8d6 f4e4 e8e4 d4e4 d6d1 c3c8 g8h7 c8f5 b6g6');
INSERT INTO evaluations VALUES(2587,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(2587,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2587,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(2587,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(2587,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(2587,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(2587,36,'cp 7','c5e3 f2e3 c8e6 c4e6 f7e6 d3d4 e5d4 e3d4 e6e5 d4e5 c6e5 d1d4 e5g6 a1e1 f8f7 c3d5 f6d5 e4d5 d8d7 d4c4 a8f8 f3d4 g6e5 c4e2 c7c6 f1f7 d7f7 d4e6 f8e8 e1f1 f7e7 c2c4 e5c4 e2c4 c6d5 c4d5 e7e6 d5a5 e6c4 a5f5 c4c5 f5c5 d6c5');
INSERT INTO evaluations VALUES(3006,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(3006,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(3006,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(3006,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(3006,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(3006,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(3006,36,'cp 112','f8e8 c1f4 d7f8 h2h3 f6h5 f4d2 e7e5 f1e1 b7b6 a2a3 h5f4 e2f1 h7h5 b2b4 f8h7 a1b1 g7h6 b4c5 b6c5 g1h2 g6g5 g2g3 g5g4 f3g1 h6g7 b1b3 f4g6 f1e2 d8d7 e1f1 h7f6 f2f3 g4h3 f1f2 f6h7 e2d3 d7d8 a3a4 e8f8 g1h3');
INSERT INTO evaluations VALUES(3366,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(3366,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(3366,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(3366,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(3366,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3366,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(3366,36,'cp 8','a2a4 h7h6 g5f6 e7f6 c3d5 g7g6 c2c3 f6g7 e1g1 g8h7 a4a5 a8b8 f1e1 c6e7 d5e7 d8e7 d3d4 c7c6 d4d5 c8d7 h2h3 f7f5 a5a6 f5e4 f3d2 b7b5 c4b3 c6c5 d2e4 c5c4 b3c2 f8f7 e4g3 b8f8 d1e2 e7f6 g3e4 f6e7 b2b4 c4b3');
INSERT INTO evaluations VALUES(4680,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4680,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(4680,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(4680,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(4680,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(4680,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(4680,36,'cp 19','f3e5 f7e6 b3b4 c7c5 b4a3 c5d4 e5f3 d8f6 e1g1 e6f7 a3d3 h8d8 f3d4 e4c5 d3e3 c5e6 f1d1 d8e8 d4e2 b7b6 b1c3 c8b7 c3d5 b7d5 d1d5 f6b2 a1d1 a8d8 e2c3 d8d5 c3d5 f7g8 a2a3 e6c7 d5e7 g8f8 e3f4 b2f6 e7g6 h7g6 f4c7 f6f7 c7f7 f8f7 d1d7 f7g8 d7a7');
INSERT INTO evaluations VALUES(4681,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(4681,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(4681,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(4681,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(4681,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(4681,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(4681,36,'cp 0','f7e6 b3b4 d8f8 b4f8 h8f8 f2f3 e4g5 b1c3 c7c6 h2h4 g5f7 e5f7 f8f7 e1f2 c8d7 a1e1 e6d6 e1e3 a7a5 h1d1 b7b6 c3a4 a8b8 a2a3 h7h6 h4h5 f7f4 e3b3 d6c7 b3e3');
INSERT INTO evaluations VALUES(5359,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(5359,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(5359,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(5359,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(5359,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(5359,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(5359,36,'cp -19','d7d6 e1g1 g8f6 c2c3 e8g8 a2a4 h7h6 h2h3 e7e5 c3c4 a8b8 f3e1 f6e8 e1c2 e8c7 b1c3 a7a6 b2b3 f7f5 e4f5 g6f5 a1b1 f5f4 b3b4 c5b4 c2b4 c6d4 f1e1 c8d7 b4d5 f4f3 g2f1 d7c6 c1e3 b7b5 a4b5 a6b5 d5c7 d8c7');
INSERT INTO evaluations VALUES(5512,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(5512,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(5512,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(5512,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(5512,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(5512,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(5512,36,'cp -14','e1g1 g8f6 a2a4 e8g8 b1d2 h7h6 d2c4 c8e6 f1e1 d8d7 c1d2 e6h3 c4a5 h3g2 g1g2 c6a5 d2a5 a7a6 a5c3 b7b5 d1d2 g8h7 h2h3 d7b7 c3f6 g7f6 c2c3 a8b8 a4b5 b7b5 e1b1 a6a5 d2d1 b5a6 a1a2 f6g7');
INSERT INTO evaluations VALUES(5511,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(5511,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(5511,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(5511,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(5511,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(5511,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(5511,36,'cp -13','g8f6 a2a4 e8g8 b1a3 h7h6 a3c4 c8e6 f1e1 d8d7 c1d2 a8d8 c4a5 c6a5 d2a5 b7b6 a5c3 e6g4 d1d2 e7e5 b2b4 f6h7 b4c5 b6c5 a4a5 g4e6 e1b1 d7c7 c3b2 d8b8 b2c3 f8d8');
INSERT INTO evaluations VALUES(8016,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(8016,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(8016,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(8016,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(8016,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(8016,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(8016,36,'cp 14','g1e2 e7e6 b2b3 e8g8 d2d4 c5d4 e3d4 d7d5 e1g1 c8d7 a1b1 b7b6 c1g5 d5c4 d4d5 c6e5 e2f4 a8c8 c3e4 h7h6 g5f6 g7f6 d5e6 d7e6 f4e6 f7e6 d1e2 c4b3 b1d1 d8e7 e4f6 e7f6 a2b3 c8e8 d1d6 e8e7 f1e1 e5f3 g2f3 f6f3 e2d2');
INSERT INTO evaluations VALUES(8017,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(8017,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8017,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(8017,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8017,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(8017,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(8017,36,'cp 13','e7e6 b2b3 e8g8 d2d4 c5d4 e3d4 d7d5 e1g1 c8d7 a1b1 b7b6 c1g5 d5c4 d4d5 c6e5 e2f4 a8c8 b3c4 h7h6 g5f6 d8f6 d5e6 d7e6 c3d5 f6d8 f4e6 f7e6 d5f4 d8d1 f1d1 g8f7 g2h3 c8c6 c4c5 e5f3 g1f1 f3d4 c5b6 a7b6 h3g2 c6d6 h2h3 f8d8 g2e4 g6g5 e4g6 f7g8 f4h5 d4f5 d1d6 d8d6 h5g7 g8g7');
INSERT INTO evaluations VALUES(12030,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(12030,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(12030,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(12030,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12030,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(12030,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(12030,36,'cp 33','e4e5 f6d5 e1g1 c8g4 f1e1 e7e6 h2h3 g4f3 d1f3 d6e5 e1e5 d8f6 f3g3 f8e7 e5e2 f6g6 g3f3 e8g8 d2d3 d5c3 b2c3 e7f6 c1d2 c5c4 d3d4 g6c2 d2g5 c2d3 e2e3 d3g6 g5f6 g6f6 f3c6 f8c8 c6a4 f6f4 a1d1 f4c7 e3e5 c7c6 a4c6 c8c6');
INSERT INTO evaluations VALUES(13029,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(13029,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13029,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13029,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(13029,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13029,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13029,36,'cp 9','f8g7 f1e1 e8g8 c4b3 c6d4 d1d4 c8d7 d4d3 d7c6 c1e3 a8c8 h2h3 a7a6 e3d4 f6d7 a1d1 g7d4 d3d4 b7b5 a2a4 d8b6 d4b6 d7b6 a4b5 a6b5 e4e5 d6d5 e5e6 f7f6 c3e2 f8d8 c2c3 g8g7 e2d4 c6e8 f2f4 b6c4 b3c4 b5c4');
INSERT INTO evaluations VALUES(13562,30,'cp 304',NULL);
INSERT INTO evaluations VALUES(13562,31,'cp 297',NULL);
INSERT INTO evaluations VALUES(13562,32,'cp 315',NULL);
INSERT INTO evaluations VALUES(13562,33,'cp 312',NULL);
INSERT INTO evaluations VALUES(13562,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(13562,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(13562,36,'cp 295','c3b5 c5d4 b5c7 e8d7 f3d4 e6e5 d4c6 d8c7 c6e5 d7e6 e5f3 f8d6 f4g5 c7b6 g5f6 g7f6 f1e2 b6b4 d1d2 b4b2 e1g1 b2b4 c2c3 b4e4 f1b1 e6e7 b1b5 h8d8 e2d3 e4g4 b5d5 g4g7 g1h1 c8e6 d5h5 h7h6 f3d4 e7f8 h2h3 a8c8 d3e4 b7b6 d4c6 c8c6 e4c6 g7g6 d2e2');
INSERT INTO evaluations VALUES(13491,30,'cp 308',NULL);
INSERT INTO evaluations VALUES(13491,31,'cp 295',NULL);
INSERT INTO evaluations VALUES(13491,32,'cp 294',NULL);
INSERT INTO evaluations VALUES(13491,33,'cp 306',NULL);
INSERT INTO evaluations VALUES(13491,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(13491,35,'cp 307',NULL);
INSERT INTO evaluations VALUES(13491,36,'cp 306','c5d4 b5c7 e8d7 f3d4 e6e5 d4c6 d8c7 c6e5 d7e6 e5f3 f8d6 f4g5 e6e7 g5f6 g7f6 c2c3 c8e6 f3d4 d6e5 f1e2 h7h5 h2h4 c7b6 e1g1 e5d4 c3d4 e7f8 d1d2 f8g7 f1c1 b6d8 c1c3 a8c8 a1c1 c8c3 d2c3 d8b6 c3a3 e6g4 e2d3 h8c8 c1c8 g4c8 a3c5 b6c5');
INSERT INTO evaluations VALUES(17973,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(17973,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(17973,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(17973,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(17973,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(17973,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(17973,36,'cp 138','d8g5 d1f3 g5c5 f3c3 c8b7 c6d4 b7e4 f2f3 e4b7 d2d3 c5e5 d4e2 e5c3 b1c3 h7h5 h2h4 g8e7 e2g3 g7g6 c1g5 f8g7 e1c1 f7f6 g5f4 a8b8 d1e1 e8f7 e1e2 h8e8 h1e1 b7c6 c3e4 a6a5 c2c4 a5a4 f4d6 c7d6 e4d6 f7f8 d6e8 b8e8 c1b1');
INSERT INTO evaluations VALUES(19166,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(19166,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(19166,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(19166,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(19166,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(19166,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(19166,36,'cp 191','a7a6 c3g3 g8e7 b1c3 d7d5 e4d5 e7f5 g3e5 f8e7 f1d3 f7f6 e5e4 f5d6 e4f3 e8g8 e1g1 c5c4 d3e2 c8f5 e2c4 f5c2 c4d3 c2d3 f3d3 a8c8 c1f4 d8b6 a1b1 b6b4 f4d6 e7d6 f1e1 b4c4 d3f5 c8e8 g2g3 c4d4 e1e8 f8e8 g1g2 d4b4 f5f3');
INSERT INTO evaluations VALUES(20041,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20041,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20041,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20041,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20041,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20041,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20041,36,'cp 0','g8f6 a2a4 a7a6 f1e2 g7g6 c1e3 f8h6 b1c3 e8g8 e1g1 e7e6 d1e1 h6g7 e1h4 f6d7 h4d8 f8d8 a4a5 c6d4 e2d1 a8b8 g1f2 b7b5 a5b6 d7b6 g2g4 b6a8 c3a4 a8b6');
INSERT INTO evaluations VALUES(13343,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13343,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13343,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13343,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13343,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13343,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13343,36,'cp 0','a2a4 g7g6 f1e2 a7a6 e1g1 f8g7 d1e1 e8g8 h2h3 a8b8 c2c3 b7b5 a4b5 a6b5 c1e3 c8b7 b1d2 e7e6 e1f2 b5b4 f1c1 b4c3 b2c3 c6e7 a1a7 d8c7 c1b1 e7c6 a7a3 c6e7');
INSERT INTO evaluations VALUES(20764,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(20764,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(20764,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(20764,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(20764,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(20764,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(20764,36,'cp 178','d4d3 g8f6 d3b3 d6d5 e4d5 f8d6 c1g5 e8g8 b1d2 h7h6 g5f6 d8f6 d2e4 f6g6 e4d6 g6d6 e1g1 c8d7 a2a4 a8c8 f1e1 d7f5 c2c3 f8d8 a1d1 c8c5 b3a2 a7a6 c4b3 f5g6 a2a3 b7b6 f2f3 a6a5 c3c4 d6f6 h2h3 f6f4 e1e2 d8e8 d1e1 f4d4 e2f2 g6f5 e1e2');
INSERT INTO evaluations VALUES(23719,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(23719,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(23719,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(23719,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(23719,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(23719,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(23719,36,'cp 18','c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 c2c3 d8e8 a3a4 b6b5 c1e3 b5a4 a1a4 a8a4 d1a4 c6d4 a4e8 d4f3 g2f3 f8e8 f1a1 g8f8 a1a7 e8b8 b2b4 f8e7 h3h4 f6h5 g1g2 e7d7 b4b5 d7e8 c3c4 e8f7 g2f1 f7e7 f1e2 h5f4 e3f4 e5f4 d3d4 e6e5 d4d5 g7g6');
INSERT INTO evaluations VALUES(23716,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(23716,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(23716,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(23716,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23716,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(23716,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(23716,36,'cp 13','f7e6 c3a4 c5b6 a4b6 a7b6 a3a4 d8e8 c2c3 c6e7 b2b4 b6b5 a4a5 e7g6 c1d2 e8c6 f1e1 f8f7 a1c1 a8f8 c3c4 c6d7 c4b5 d7b5 d3d4 f6h5 f3e5 g6e5 d4e5 b5e5 d2e3 h5f4 c1c4 e5g5 e3f4 g5f4 d1c2 f4e5 e1f1 c7c6 c2c3 e5c3 c4c3 d6d5 e4d5 e6d5');
INSERT INTO evaluations VALUES(23745,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(23745,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(23745,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(23745,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23745,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(23745,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(23745,36,'cp 12','c3a4 c5b6 b2b4 d8d7 b4b5 c6d4 a4b6 a7b6 c2c4 d4f3 d1f3 d7e8 a3a4 e8g6 f3d1 f6h5 g1h2 g8h7 c1e3 f8f7 d1g4 g6g4 h3g4 h5f6 h2h3 f6d7 h3g3 h7g6 a1a3 f7f8 f1a1 a8a5 f2f3 g6f7 g3f2 d7c5 e3d2 a5a7 f2e2');
INSERT INTO evaluations VALUES(23963,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(23963,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(23963,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(23963,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(23963,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(23963,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(23963,36,'cp 43','h2h3 g4h5 d2d3 e8g8 b1d2 c6a5 b3c2 c7c5 d2f1 a8b8 f1g3 h5g6 f3h4 f6e8 h4f5 g6f5 g3f5 e7g5 f5e3 a5c6 c2b3 e8f6 d1e2 h7h6 e1d1 g5e3 c1e3 c6a5 b3c2 a5c6 f2f4 e5f4 e3f4 f8e8');
INSERT INTO evaluations VALUES(23962,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(23962,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(23962,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(23962,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(23962,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(23962,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(23962,36,'cp 55','g4h5 d2d3 c6a5 b3c2 c7c5 b1d2 e8g8 d2f1 h5g6 f1g3 a8b8 d3d4 c5d4 c3d4 b8c8 a2a4 e5d4 a4b5 a6b5 f3d4 d6d5 d4f5 d5e4 f5e7 d8e7 a1a5 e7c7 a5b5 f8d8 c1d2 c7c2 d1c2 c8c2 d2c3 e4e3 e1e3 d8d1 g1h2 h7h5 c3f6 g7f6 g3h5 c2f2 e3e8 g8h7 b5b8 g6h5 e8h8 h7g6 b8g8 g6f5 h8h5 f5e6');
INSERT INTO evaluations VALUES(23959,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(23959,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(23959,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(23959,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(23959,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(23959,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(23959,36,'cp 48','d2d3 c6a5 b3c2 c7c5 b1d2 e8g8 d2f1 h5g6 f1g3 f6d7 f3h2 a5c6 h2g4 a6a5 a2a4 b5b4 c2b3 d7b6 g3f5 g6f5 e4f5 a8c8 c1e3 g8h8 d1f3 e7g5 e1f1 c8c7 a1e1 f7f6 e3g5 f6g5 g2g3 d8d7 b3e6 d7d8 g1g2 c6e7 h3h4 g5h4');
INSERT INTO evaluations VALUES(23960,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(23960,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(23960,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(23960,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(23960,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(23960,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(23960,36,'cp 48','c6a5 b3c2 c7c5 b1d2 e8g8 d2f1 h5g6 c1g5 f6d5 g5d2 d5c7 f1e3 a5c6 a2a4 d8d7 c2b3 a8b8 a4b5 a6b5 e3d5 e7d8 d5c7 d7c7 b3d5 c6e7 d5a2 h7h6 d1e2 g8h8 b2b4 c5b4 c3b4 e7c6 a2d5 c7d7 e1c1 c6e7 d5b3 d8b6 d3d4');
INSERT INTO evaluations VALUES(32111,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(32111,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(32111,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(32111,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(32111,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(32111,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(32111,36,'cp 31','d2d4 b8c6 d4d5 g4f3 g2f3 c6d4 b1c3 g7g6 c3b5 d4b5 c4b5 d8h4 d1a4 b7b6 a4a6 f8h6 a6b7 a8d8 c1h6 g8h6 a1c1 h4f4 b7c6 e8f8 c6c3 d8d7 h2h4 f7f5 c3e3 d7e7 f1e2 h6f7 c1c6 f8g7 h4h5 f5e4 f3e4 f7g5 e3f4');
INSERT INTO evaluations VALUES(34627,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(34627,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(34627,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(34627,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(34627,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(34627,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(34627,36,'cp 100','e7e6 d2d4 g5d8 g1e2 g8f6 e1g1 b8d7 f1e1 c7c6 c1f4 f8e7 c3e4 f6e4 f3e4 e8g8 c2c3 d7f6 e4c2 d8b6 a2a4 b6b2 a1b1 b2a2 f4e5 a2d5 e2f4 d5d8 d1f3 f6d5 f3h3 f7f5 h3g3 e7f6 e5f6 d8f6 c2b3 d5f4 g3f4 g8h8 a4a5');
INSERT INTO evaluations VALUES(38288,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38288,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(38288,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38288,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(38288,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(38288,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(38288,36,'cp 30','b1c3 g8f6 c1f4 c7c5 d4c5 e7c5 e2e3 b8c6 c4d5 f6d5 c3d5 e6d5 a2a3 c5b6 a1c1 e8g8 f1d3 d8f6 b2b4 c8f5 b4b5 c6e7 d3f5 f6f5 e1g1 f8c8 f4e5 c8c1 d1c1 d5d4 e5d4 f5b5 c1a1 b6d4 f3d4 b5d7 d4f3 a8c8 f1b1 e7f5');
INSERT INTO evaluations VALUES(39870,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39870,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39870,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(39870,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(39870,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(39870,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39870,36,'cp -3','g1f3 d7d6 e1g1 e8g8 c3e2 h7h6 e2g3 b8d7 c2c3 c5b6 c4b3 f8e8 f1e1 d6d5 h2h3 b6c7 b3c2 a7a6 a2a4 a6a5 a1b1 d7f8 d3d4 f6e4 d4e5 e4g3 f2g3 f7f6 e5f6 e8e1 d1e1 d8f6 f3d4 f6f7 e1f2 f7f2 g1f2 c8d7 c1f4 c7b6 b1e1 g8f7 f2f3');
INSERT INTO evaluations VALUES(39869,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39869,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39869,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(39869,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(39869,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39869,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39869,36,'cp -16','d7d6 a2a4 e8g8 e1g1 h7h6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1d2 b8d7 h2h3 a7a5 f1f2 a8b8 c3e2 d8c7 d2c3 c7b6 b2b3 b8a8 c3d2 b6b4 e2c3 f8f7 a1f1 a8f8 c3e2 d6d5 e2g3 b4d2 f2d2 b7b5 e4d5 e6d5');
INSERT INTO evaluations VALUES(41030,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(41030,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(41030,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(41030,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(41030,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(41030,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(41030,36,'cp 76','g1f3 e8g8 d1c2 c7c5 d4c5 h7h6 a1d1 d8a5 g5f6 g7f6 c2c1 g8g7 f1e2 a5c5 e1g1 b8c6 a2a3 a7a5 c1f4 f8g8 c3a4 c5c2 e2b5 c2b3 a4b6 b3b5 b6a8 b5b2 d1d3 e6e5 f4d2 b2d2 f3d2 c6d4 d3g3 g7f8 g3g8 f8g8 d2c4 b7b5 c4a5');
INSERT INTO evaluations VALUES(41029,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(41029,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(41029,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(41029,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(41029,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(41029,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(41029,36,'cp 97','e8g8 d1c2 c7c5 d4c5 h7h6 a1d1 d8a5 g5f4 a5c5 a2a3 b8d7 f1d3 e6e5 f4g3 a7a6 e1g1 b7b5 b2b4 c5c7 d1c1 c8b7 c2b2 c7b8 f1d1 f8c8 d3b1 e7f8 b1a2 c8c6 f3d4 c6c7 c3d5 c7c1 b2c1 b7d5 e4d5 b8b7 c1c6 b7c6 d4c6 f8d6');
INSERT INTO evaluations VALUES(36368,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(36368,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(36368,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(36368,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(36368,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(36368,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(36368,36,'cp 104','d1c2 c7c5 d4c5 h7h6 a1d1 d8a5 g5f4 a5c5 a2a3 b8d7 f1d3 e6e5 f4g3 a7a6 e1g1 b7b5 b2b4 c5c7 f1e1 f8e8 c2b3 c8b7 d3b1 c7c4 b3c4 b5c4 f3e5 d7e5 g3e5 a6a5 c3d5 a5b4 d5e7 e8e7 e5d6 e7e6 d6b4 b7e4 b1e4 f6e4 d1d4 e4f6 e1e6 f7e6 d4c4 f6d5 h2h4');
INSERT INTO evaluations VALUES(43119,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43119,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43119,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(43119,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43119,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(43119,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(43119,36,'cp 10','b1c3 d7d6 e2e4 e7e6 g1f3 e8g8 f1d3 c7c6 e1g1 c6d5 e4d5 e6d5 c4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 c1g5 d8b6 g5e3 b6b4 a2a3 b4b3 d3d1 b3d1 a1d1 f8e8 e3f4 e8e1 d1e1 f6e8 e1e7 b7b5 c3e4');
INSERT INTO evaluations VALUES(43118,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(43118,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(43118,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43118,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43118,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43118,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43118,36,'cp 0','d7d6 e2e4 e8g8 g1f3 e7e6 f1d3 c7c6 e1g1 c6d5 c4d5 e6d5 e4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f1e1 d7e5 f3d1 e5d3 d1d3 a7a6 a2a4 a8c8 c1e3 d8a5 e3f4 a5b6 d3d2 b6b4 a4a5 f8e8 e1e8 f6e8 a1a4 b4b3 a4a3 b3b4');
INSERT INTO evaluations VALUES(44561,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44561,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44561,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44561,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44561,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44561,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44561,36,'cp 14','e4e5 f6d5 b1c3 d5b4 g1f3 g7g6 c3a4 f8g7 c2c3 b4c6 a4c5 c6e5 f3e5 g7e5 f1e2 e5g7 d2d4 d7d6 c5e4 d6d5 e4c5 b8d7 c5d7 c8d7 c1f4 a8c8 d1d2 d8a5 a1c1 d7b5 e1g1 b5e2 d2e2 e7e6 f4d6 a5d8 e2b5 d8d7 b5b4 c8c6 d6e5 g7e5 d4e5 d7c7');
INSERT INTO evaluations VALUES(53432,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(53432,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(53432,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(53432,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(53432,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(53432,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(53432,36,'cp -12','b8c6 d2d4 c8f5 g1f3 e7e6 e1g1 g8f6 b1c3 f8d6 f3h4 f5g4 h2h3 g4h5 c1g5 a8c8 g2f3 h5f3 h4f3 e8g8 e2e3 h7h6 g5f6 d8f6 a1c1 f6f5 g1g2 c6a5 c3b5 d6b8 c1c8 f8c8 b5c3 a7a6 d1e2 a5c4 f1c1 b7b5 g3g4 f5h7');
INSERT INTO evaluations VALUES(53640,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(53640,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(53640,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(53640,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(53640,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53640,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(53640,36,'cp 21','d2d4 c6c5 g1f3 g8h6 c2c3 b8c6 b1a3 c5d4 c3d4 d8b6 a3c2 h6f5 f1d3 c6b4 c2b4 f8b4 e1f1 h7h5 c1d2 c8d7 d3f5 e6f5 h2h3 b4e7 d2c3 a8c8 f1g1 h5h4 g1h2 e8g8 d1d2 b6g6 d2f2 g6h6 f2e3 c8c4 h1g1 d7b5 g1e1 h6g6 f3d2 c4c7 a2a4 b5a6 e3f2 f8c8');
INSERT INTO evaluations VALUES(53965,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(53965,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(53965,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(53965,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(53965,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(53965,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(53965,36,'cp -52','b8c6 c2c3 g7g6 h5e2 f8g7 a2a4 g8e7 g1f3 e8g8 e1g1 d6d5 c4a2 h7h6 b1a3 e6e5 a3c2 a7a5 c2e3 c8e6 f1e1 g8h7 g1f1 f8e8 e4d5 e7d5 e3d5 e6d5 a2d5 d8d5 f3d2 f7f5 f2f3 e8d8 d2c4 d5d3 e2d3 d8d3 c1e3');
INSERT INTO evaluations VALUES(55324,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(55324,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(55324,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(55324,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(55324,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(55324,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(55324,36,'cp 54','f8g7 f3g5 g8h6 a2a4 c7c6 b1c3 b8a6 f2f4 a6c7 f4e5 d6e5 c1e3 f7f6 d1c1 d8d6 g5f3 h6f7 e1g1 c8e6 c4e6 c7e6 a4a5 e8g8 c1d2 e6d4 d2f2 d4f3 f2f3 g7h6 c3a4 h6e3 f3e3 g8g7 a4c5 d6e7 h2h3 a8d8 c5b3 a7a6 e3b6 f7g5');
INSERT INTO evaluations VALUES(38521,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(38521,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(38521,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38521,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(38521,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(38521,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(38521,36,'cp 59','f3g5 g8h6 b1c3 c7c6 h2h3 b7b5 c4b3 b8a6 a2a3 e8g8 c1e3 d6d5 e4d5 h6f5 g5f3 f5e3 f2e3 a6c5 d5c6 c5b3 c2b3 d8d6 e1g1 d6c6 b3b4 c6b6 d1e2 c8b7 e2f2 f7f5 e3e4 b6f2 g1f2 a7a6 a1d1 a8e8 f1e1 h7h6 e4f5 g6f5 d3d4 e5e4 f3e5 g7e5');
INSERT INTO evaluations VALUES(38518,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(38518,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(38518,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(38518,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(38518,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(38518,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(38518,36,'cp 38','g8h6 b1c3 c7c6 a2a4 b8a6 f2f4 a6c7 f4e5 d6e5 c1e3 f7f6 d1c1 d8d6 g5f3 h6f7 c4f7 e8f7 e1g1 h8d8 a4a5 f7g8 c1e1 c7b5 c3a4 c8g4 c2c3 d6e7 h2h3 g4e6 e1e2 e6f7 g1h1 b5c7 b2b4 c7e6 a4c5 e6c5 e3c5 e7d7 f1d1 f7e6 c3c4 g7h6');
INSERT INTO evaluations VALUES(41604,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(41604,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(41604,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(41604,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(41604,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(41604,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(41604,36,'cp 62','b1c3 c7c6 h2h3 b7b5 c4b3 b8a6 a2a3 d8e7 c1e3 f7f6 g5f3 h6f7 d3d4 a6c7 d1d2 e8g8 e1g1 a7a5 a1d1 c8e6 d4d5 c6d5 c3d5 c7d5 b3d5 e6d5 d2d5 a8b8 c2c3 f8c8 f3d2 e7b7 d5d3 c8d8 c3c4 a5a4 d1c1 f6f5 e4f5 b5c4 d2c4 g6f5 d3f5 d6d5 c4a5');
INSERT INTO evaluations VALUES(55879,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(55879,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(55879,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(55879,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(55879,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(55879,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(55879,36,'cp 101','b1d2 g8f6 e4e5 f6d7 c2c3 c7c5 f1d3 b8c6 e1g1 a7a5 d3c2 c5d4 c3d4 b7b5 f1e1 d7b6 d2f1 f8b4 c1g5 d8c7 e1e2 h7h6 g5c1 a5a4 f1g3 b6c4 a1b1 b4f8 b2b3 a4b3 a2b3 c4a3 c1a3 a8a3 g3h5 c7b6 d1d2 c8d7 d2f4');
INSERT INTO evaluations VALUES(57687,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(57687,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(57687,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(57687,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(57687,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(57687,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(57687,36,'cp 64','e2e4 g8f6 e4e5 f6d5 c3c4 d5b6 c4c5 b6d5 f1c4 d5f4 d1f3 f4e6 d2d4 d7d6 b1c3 f8g7 c4e6 f7e6 f3h3 e8g8 c1h6 d6c5 d4c5 d8d4 g1f3 f8f3 g2f3 d4e5 h6e3 b8c6 a1d1 e5f5 h3f5 g7c3 b2c3 e6f5 h2h4 g8f7 e3g5 c8e6 h4h5 a8c8 h5g6 h7g6 e1d2 c8d8 d2c2 d8d1 h1d1 e6a2');
INSERT INTO evaluations VALUES(8818,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(8818,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(8818,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(8818,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(8818,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(8818,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(8818,36,'cp -50','c8d7 c2c3 a7a6 b1d2 c5a7 b2b4 g7g5 h4g3 f6h5 f3h2 h5f4 g3f4 g5f4 d1h5 d8e7 a2a4 a6a5 b4b5 c6d8 c4d5 e7g5 h5g5 h6g5 d2f3 f7f6 e1e2 c7c6 d5c4 a8c8 h2g4 e8e7 h1c1 a7b6 c1b1 c6b5 a4b5 c8a8 c4d5 a5a4 d3d4');
INSERT INTO evaluations VALUES(9304,30,'cp 523',NULL);
INSERT INTO evaluations VALUES(9304,31,'cp 503',NULL);
INSERT INTO evaluations VALUES(9304,32,'cp 501',NULL);
INSERT INTO evaluations VALUES(9304,33,'cp 527',NULL);
INSERT INTO evaluations VALUES(9304,34,'cp 513',NULL);
INSERT INTO evaluations VALUES(9304,35,'cp 536',NULL);
INSERT INTO evaluations VALUES(9304,36,'cp 517','c3d4 f6e4 b1c3 e4c3 b2c3 c5b6 d4d5 e8g8 d5c6 d8f6 f3d4 b7c6 b5c6 a8b8 c1e3 f8d8 d1h5 h7h6 g1h1 b6c5 a2a4 c8a6 c6b5 a6b7 a1e1 a7a6 b5c6 d8e8 c6e8 b8e8 h5g4 c5d4 e3d4 e8e1 f1e1 f6g5 g4g5 h6g5 e1e7 b7d5 e7c7 f7f6 c7e7 d5f7 f2f3');
INSERT INTO evaluations VALUES(9397,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9397,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(9397,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(9397,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(9397,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9397,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9397,36,'cp 12','c3a4 c5b6 c2c3 d6d5 a4b6 a7b6 d1c2 c8e6 b2b4 e8g8 f1e2 d5e4 d3e4 f6e8 f3g5 e6d7 e1g1 e8d6 g5f3 d8e8 a2a4 c6d8 a4a5 d7c6 e2d3 f7f5 f3d2 d8e6 e4f5 e6f4 d2e4 f4d3 e4d6 e8d7 c2d3 c7d6 c1e3 b6b5 f2f3 f8f5');
INSERT INTO evaluations VALUES(10590,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(10590,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(10590,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(10590,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(10590,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(10590,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(10590,36,'cp 129','c3d4 c6d4 f3d4 d7d5 c4e2 g4f2 e1f2 d8h4 g2g3 h4d4 d1d4 c5d4 f2g2 d4e5 h1e1 f7f6 e2h5 g7g6 h5f3 c8e6 b1c3 e8c8 h2h4 c7c6 c1d2 h7h6 a1c1 g6g5 h4h5 h8e8 b2b4 d5d4 c3e4 g5g4 f3d1 d4d3 d1b3 e6b3 a2b3 f6f5 e4f2 e5b2 c1d1 e8e2 d2h6');
INSERT INTO evaluations VALUES(10586,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(10586,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(10586,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(10586,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(10586,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(10586,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(10586,36,'cp 145','c6d4 f3d4 d7d5 c4e2 g4f2 e1f2 d8h4 g2g3 h4d4 d1d4 c5d4 f2g2 d4e5 h1e1 f7f6 e2h5 g7g6 h5f3 c8e6 b1c3 c7c6 c1f4 e5f4 g3f4 e8f7 b2b4 a8e8 c3a4 e8e7 g2f2 h7h5 a1c1 e6g4 e1e7 f7e7 a4c5 b7b6 c5b3 e7d6 h2h4 g4d7 c1g1 d7e8 b3d4 c6c5 d4b3 h8h7 b4c5 b6c5 b3d2 e8f7');
INSERT INTO evaluations VALUES(15494,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(15494,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(15494,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(15494,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(15494,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(15494,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(15494,36,'cp 33','g1f3 f8e7 e1g1 e8g8 a2a4 a7a5 f1e1 c8e6 c4e6 f7e6 b1a3 d8d7 a3c2 c6d8 c1d2 d8f7 h2h3 f7h8 c2a3 h8g6 a3c4 f6h5 f3e5 d6e5 d1h5 d7d3 h5e2 d3e2 e1e2 g6f4 d2f4 e5f4 a1e1 f4f3 e2e3 f3g2 g1g2 f8f6');
INSERT INTO evaluations VALUES(16373,30,'cp -225',NULL);
INSERT INTO evaluations VALUES(16373,31,'cp -232',NULL);
INSERT INTO evaluations VALUES(16373,32,'cp -231',NULL);
INSERT INTO evaluations VALUES(16373,33,'cp -231',NULL);
INSERT INTO evaluations VALUES(16373,34,'cp -223',NULL);
INSERT INTO evaluations VALUES(16373,35,'cp -219',NULL);
INSERT INTO evaluations VALUES(16373,36,'cp -217','f6e4 g5d8 e8d8 b1c3 e4d6 h2h4 h7h5 c3d5 f7f6 c2c3 d6f5 d2d4 e5d4 f1d3 d7d6 d5f4 d4c3 b2c3 f5e7 f3d4 c6d4 c3d4 c8f5 d3f5 e7f5 d4d5 d8d7 a1b1 b7b6 f2f3 f5e7 c1b2 g7g6 b2f6 f8h6 f6g5 h6g5 h4g5 a8f8 g2g3');
INSERT INTO evaluations VALUES(16819,30,'cp -217',NULL);
INSERT INTO evaluations VALUES(16819,31,'cp -218',NULL);
INSERT INTO evaluations VALUES(16819,32,'cp -219',NULL);
INSERT INTO evaluations VALUES(16819,33,'cp -214',NULL);
INSERT INTO evaluations VALUES(16819,34,'cp -213',NULL);
INSERT INTO evaluations VALUES(16819,35,'cp -241',NULL);
INSERT INTO evaluations VALUES(16819,36,'cp -211','g5d8 e8d8 b1c3 e4d6 h2h4 h7h5 c3d5 f7f6 d5e3 a7a5 c2c3 a5a4 d2d4 e5d4 f3d4 a4a3 d4c6 b7c6 b2b3 d6e4 c1d2 f8c5 f1d3 h8e8 d3e4 e8e4 e1c1 d7d5 d1e1 e4e5 f2f3 c8e6 e3c2 c5d6 c2d4 e6d7 b3b4 c6c5 b4c5 d6c5 d2f4 e5e1 h1e1 a8a4');
INSERT INTO evaluations VALUES(19652,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(19652,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19652,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(19652,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19652,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19652,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19652,36,'cp 0','c2c3 g8f6 e1g1 f8g7 a2a4 e8g8 f1e1 h7h6 c1e3 c8g4 b1d2 c6e5 h2h3 g4f3 d2f3 e5c4 d3c4 f6e4 e3h6 g7h6 e1e4 g8g7 d1d5 e7e6 d5b7 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(22113,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22113,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(22113,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(22113,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(22113,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(22113,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22113,36,'cp 74','d2d4 e5f3 d1f3 d8f6 b1c3 f6f3 g2f3 c8d7 f1c4 f7f5 h1g1 f5e4 f3e4 e8c8 c1g5 g8f6 e1c1 d8e8 d1e1 f6g4 c4f7 f8e7 g5d2 e8f8 f7h5 g4h2 g1g7 e7d8 e1h1 h2f3 d2h6 f3d4 g7g1 f8g8 h6g7 g8g7 g1g7 d4e6 g7g1 d8g5 c1b1 h7h6 h5g4 h8e8 g1f1 e6f4 g4d7 c8d7 a2a4 e8f8 h1g1 c7c6');
INSERT INTO evaluations VALUES(22111,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22111,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(22111,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(22111,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(22111,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(22111,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(22111,36,'cp 66','e5f3 d1f3 d8f6 b1c3 f6f3 g2f3 c8d7 f1c4 f7f5 h1g1 f5e4 f3e4 e8c8 c1g5 g8f6 e1c1 d8e8 d1e1 f6g4 c4f7 f8e7 g5d2 e8f8 f7h5 g4h2 g1g7 e7d8 e1h1 h2f3 d2h6 f3d4 g7g1 f8g8 h6g7 g8g7 g1g7 d4e6 g7g1 e6g5 h5g4 h7h6 c1b1 c7c6 h1h5 d8f6 g4f5 c8c7 g1h1 f6g7 c3e2 h8f8 f5d7');
INSERT INTO evaluations VALUES(22139,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(22139,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(22139,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(22139,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(22139,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(22139,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(22139,36,'cp 76','d1f3 d8f6 b1c3 f6f3 g2f3 c8d7 f1c4 f7f5 h1g1 f5e4 f3e4 e8c8 c1g5 g8f6 e1c1 d8e8 d1e1 f6g4 c4f7 f8e7 g5d2 e8f8 f7h5 g4h2 g1g7 e7d8 e1h1 h2f3 d2h6 f3d4 g7g1 f8g8 h6g7 g8g7 g1g7 d4e6 g7f7 e6g5 f7f4 g5h3 f4f1 h7h6 c3d5 h3g5 h5g6');
INSERT INTO evaluations VALUES(22138,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(22138,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(22138,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(22138,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(22138,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(22138,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(22138,36,'cp 69','d8f6 b1c3 f6f3 g2f3 c8d7 f1c4 f7f5 h1g1 f5e4 f3e4 e8c8 c1g5 g8f6 e1c1 d8e8 d1e1 f6g4 c4f7 f8e7 g5d2 e8f8 f7h5 g4h2 g1g7 e7d8 e1h1 h2f3 d2h6 f3d4 g7g1 f8g8 h6g7 g8g7 g1g7 d4e6 g7g1 d8g5 c1b1 h7h6 a2a4 c8d8 h5g4 h8e8 g1f1 d7c6 b2b4 g5d2 c3d5 e6g5 h1h6 g5e4');
INSERT INTO evaluations VALUES(24123,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(24123,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(24123,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(24123,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(24123,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(24123,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(24123,36,'cp -125','c6d4 g5e3 d4f3 g2f3 g4h3 f1e1 f6h5 g1h1 d8h4 e1g1 e8c8 c3e2 c5e3 f2e3 c8b8 d1e1 h4f6 f3f4 g7g6 e2g3 f6h4 g3h5 h4h5 e1d1 h5h6 d1f3 f7f5 f4e5 d6e5 g1g3 h3g4 f3g2 g4h5 a1g1 h8f8 e4f5 g6f5 c4d5 f5f4 e3f4 e5f4 g3g8 f4f3 g2g7 h6f4 g7g3 f4g3 g1g3 f8g8');
INSERT INTO evaluations VALUES(24867,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(24867,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(24867,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(24867,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(24867,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(24867,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(24867,36,'cp -2','e7e6 c1g5 h7h6 g5f6 d8f6 e2e3 a7a6 f1d3 g7g5 d3f5 f6f5 d1e2 e8c8 e1c1 f5g6 f3d2 h6h5 d2b3 c8b8 e3e4 h8g8 f2f3 f8g7 e2f2 f7f5 e4e5 g7f8 f2e2 f8e7 c1b1 g5g4 e2d3 g6f7 c3e2 g4h3 g2h3');
INSERT INTO evaluations VALUES(25182,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(25182,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(25182,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25182,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25182,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25182,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(25182,36,'cp 0','c1g5 h7h6 g5f6 d8f6 e2e3 f8e7 f1d3 h6h5 d1e2 f5d3 e2d3 e8c8 e1c1 g7g5 c1b1 g5g4 f3e1 a7a6 d3e2 f6g6 h3g4 h5g4 e1d3 c8b8 f2f4 g4f3 g2f3 d8g8 f3f4 e7d6 e2f2 c6e7 h1h8 g8h8 d1g1 g6h5 g1g7 h5h1 g7g1');
INSERT INTO evaluations VALUES(27113,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(27113,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(27113,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(27113,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(27113,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(27113,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(27113,36,'cp 47','b1c3 d8d7 h2h4 e8c8 c1f4 d7f7 d1d2 c8b8 h4h5 d8c8 a2a3 c6d7 e1c1 h7h6 c1b1 g8e7 d2d3 f7e8 d3e2 h8g8 g2g4 e7c6 f3h4 f6e5 f4e5 c6e5 e2e5 f8d6 e5e2 e8f7 h4g6 c7c5 d4c5 c8c5 e2e3 c5c8');
INSERT INTO evaluations VALUES(29645,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(29645,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(29645,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(29645,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29645,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29645,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29645,36,'cp 2','g4e2 d1e2 e7e6 g1f3 c7c5 e1g1 b8c6 b1d2 h7h6 f4e5 c5d4 e3d4 f8e7 e5f6 e7f6 c2c3 e8g8 g2g3 a7a5 a2a4 d8b6 e2b5 b6c7 f1e1 f6e7 b5e2 e7d6 d2f1 c7d7 f1e3 a8e8 e3g4 e8c8 h2h4 c6e7 g4e3 c8c7 h4h5');
INSERT INTO evaluations VALUES(29642,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(29642,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29642,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(29642,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(29642,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(29642,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(29642,36,'cp 4','d1e2 e7e6 g1f3 f8d6 f4g3 b8d7 c2c4 d5c4 e2c4 a8c8 b1c3 d6g3 h2g3 c7c5 d4d5 f6d5 c3d5 e6d5 c4d5 e8g8 a1d1 d7f6 d5d8 f8d8 e1e2 b7b5 d1d8 c8d8 f3d2 f6g4 d2b3 c5c4 b3c5 g4f6 h1c1 h7h6 f2f3 d8c8 b2b4 c4b3');
INSERT INTO evaluations VALUES(29641,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29641,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29641,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29641,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(29641,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29641,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(29641,36,'cp 0','c7c5 g1f3 e7e6 e1g1 b8c6 b1d2 a8c8 c2c3 f8e7 d4c5 e7c5 e3e4 e8g8 f4g5 c5e7 g5f6 e7f6 e4d5 d8d5 d2e4 f6e7 f1d1 d5a5 d1d7 c8c7 b2b4 a5b6 d7d3 c7c8 d3d7');
INSERT INTO evaluations VALUES(30459,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(30459,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(30459,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(30459,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(30459,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(30459,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(30459,36,'cp 120','g1f3 e7e6 d4d5 g8f6 d5c6 b8c6 a4b5 c6a5 c1g5 d8c7 d1a4 a5b3 a1a2 a8c8 a4a7 c7a7 a2a7 f8b4 b5b6 f6e4 g5f4 e4c3 b2c3 b4c3 e1d1 b3a5 d1c2 c3b4 b6b7 a5b7 a7b7 d7a4 c2b1 b4c5 f1c4 e8g8 c4b5 a4b5 b7b5 c5f2 b5b2 f2c5 h1c1 c5a3 c1c8 f8c8 b2c2 c8d8 c2c4 a3e7 b1c2 e7f6 f4e5 h7h6 e5f6 g7f6 f3d4');
INSERT INTO evaluations VALUES(31298,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(31298,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(31298,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31298,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31298,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(31298,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(31298,36,'cp 6','e7e6 b1c3 f8b4 c4d5 d8d5 f1e2 g8e7 e1g1 d5a5 c1d2 e8g8 a2a3 b4c3 b2c3 f8d8 f1e1 b8c6 h2h3 h7h6 e2f1 a5c7 f3h4 f5h7 d1g4 g8h8 h4f3 a8c8 f3e5 c6e5 e1e5 h7f5 g4f3 e7c6 e5e1 h8g8 d2f4 c7e7');
INSERT INTO evaluations VALUES(31683,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31683,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(31683,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(31683,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31683,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(31683,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31683,36,'cp 19','c4d5 d8d5 b1c3 f8b4 c1d2 b4c3 b2c3 g8f6 f1e2 e8g8 c3c4 d5d6 e1g1 b8c6 d2c3 a7a6 d1c1 b7b5 d4d5 e6d5 c4b5 a6b5 e2b5 f8c8 a2a4 f6e4 c3b2 c6a5 c1d1 a5c4 b2d4 f5g4 h2h3 g4f3 d1f3 e4d2 f3g4 d6g6 g4g6 h7g6');
INSERT INTO evaluations VALUES(31880,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(31880,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(31880,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31880,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(31880,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(31880,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(31880,36,'cp 33','c2c4 c7c6 d1b3 d8c8 b1c3 e7e6 f3h4 f5g6 d2d4 g8f6 h4g6 h7g6 g2g3 a7a5 c1d2 b8d7 a2a3 f8d6 f1d3 d6c7 e1c1 e8g8 c1b1 d5c4 b3c4 c7d6 c3e4 f6e4 d3e4 d7f6 e4g2 f8e8 h2h4 e6e5 d4e5 d6e5 c4c2 c8e6');
INSERT INTO evaluations VALUES(31883,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(31883,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(31883,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(31883,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(31883,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(31883,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(31883,36,'cp 32','c2c4 c7c6 d1b3 d8c7 g1f3 e7e6 d2d3 g8f6 e1g1 f8d6 c4d5 e6d5 e2e4 d5e4 d3e4 f5e4 f3g5 e4g2 f1e1 e8f8 g1g2 h7h6 g5f3 b8d7 c1e3 c6c5 b1a3 c7c6 a1d1 f8g8 b3b5 a8c8 e3d2 a7a6 b5c6 c8c6 a3c4 g8h7 d2c3 d6c7 e1e7 b7b5 f3e5 c7e5 c4e5');
INSERT INTO evaluations VALUES(33509,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(33509,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(33509,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(33509,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(33509,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(33509,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(33509,36,'cp 117','g2g3 c8g4 h2h3 g4h5 c1f4 d6b4 g3g4 h5g6 d1c1 e7e6 f4c7 b8c6 a2a3 b4e7 c7f4 e7d8 f1g2 f8d6 e1g1 e8g8 f3e5 d8c7 g2c6 b7c6 e5g6 h7g6 f4d6 c7d6 c1e3 a8b8 b2b3 f6d5 c3d5 e6d5 f1e1 b8e8 e3e8 f8e8 e1e8 g8h7 e8e3 d6f4 a1d1 f7f5 e3g3 g6g5 g4f5');
INSERT INTO evaluations VALUES(42716,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42716,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42716,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42716,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42716,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42716,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42716,36,'cp 3','e8g8 d2d3 f8e8 e5c4 d7d5 c4e3 d5e4 f1e2 e4d3 c2d3 c8e6 e1g1 c7c5 c1a3 d8a5 a3b2 b8c6 c3c4 c6d4 b2d4 c5d4 e3c2 a8d8 a1b1 a5c7 e2f3 b7b6 h2h3 h7h6 f1e1 c7d6 c2b4 d8c8 d1d2 e6d7 a2a3 a7a5 e1e8 f6e8 b4c2 e8f6 f3b7');
INSERT INTO evaluations VALUES(43426,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(43426,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(43426,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(43426,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(43426,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(43426,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(43426,36,'cp 35','e3c1 g4f6 c1e3 f6g4 e3g5 h7h6 g5h4 g7g5 h4g3 f8g7 h2h3 g4e5 h3h4 g5h4 h1h4 b8c6 d4f5 g7f6 h4h3 d8a5 d1d2 c8f5 e4f5 c6d4 g3h4 f6h4 d2d4 h4f6 d4d5 a5b6 d5b3 b6b3 a2b3 a8c8 e1c1 e5c6 c3d5');
INSERT INTO evaluations VALUES(44624,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(44624,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44624,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44624,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(44624,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(44624,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44624,36,'cp 0','b8c6 d4a4 e7e6 g1f3 d8c7 c3b5 c7b8 f1d3 a7a6 b5d4 c6b4 c1d2 b4d3 c2d3 f8e7 d2b4 e7b4 a4b4 d7d6 e1g1 e8g8 d4e2 h7h6 h2h3 c8d7 f1c1 b7b5 b4a5 b8d8 a5c7 d8b8');
INSERT INTO evaluations VALUES(12885,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12885,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(12885,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12885,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(12885,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12885,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(12885,36,'cp 0','d4a4 e7e6 g1f3 d8c7 f1d3 a7a6 c3b5 c7b8 b5d4 c6b4 c1d2 b4d3 c2d3 f8e7 d2b4 e7b4 a4b4 a6a5 b4d2 e8g8 h2h3 d7d6 e1g1 h7h6 f1e1 c8d7 d4c2 b7b5 c2e3 b8b6 a2a3 f8e8 e1c1 e8c8 e3g4 f6e8 g4e3');
INSERT INTO evaluations VALUES(51906,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(51906,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(51906,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(51906,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(51906,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(51906,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(51906,36,'cp -73','b8c6 g1f3 g8e7 c3e4 e7g6 e4d6 f8d6 e5d6 g6h4 c2c3 b7b6 d2d4 h4f3 d1f3 c5d4 c3d4 c8b7 c1e3 e8g8 a1d1 d8h4 a2a3 c6a5 f3g3 h4g3 h2g3 a8c8 f2f3 h7h6 g3g4 b7d5 f1b5 a5c4 b5c4 c8c4 e1d2 f7f5 g4f5 f8f5 d1c1 c4c1 h1c1 d5c6 d2d3 g7g5 e3d2 f5d5 d2c3 d5d6 d3e3');
INSERT INTO evaluations VALUES(18717,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(18717,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(18717,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(18717,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(18717,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(18717,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(18717,36,'cp -62','g1f3 g8e7 c3e4 e7g6 e4d6 f8d6 e5d6 g6h4 f3h4 d8h4 c2c3 b7b6 f1e2 h4f4 d2d4 f4d6 d4c5 d6c5 e1g1 e8g8 c1e3 c5e7 e2d3 c8b7 d1f3 g8h8 a1d1 d7d6 f3h3 f7f5 d3c2 c6e5 h3g3 e5c4 e3c1 e6e5 f1e1 e7f6 b2b3 c4a5 c1a3 a8d8 g3d3 f6g6 f2f3');
INSERT INTO evaluations VALUES(3579,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(3579,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(3579,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3579,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3579,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(3579,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3579,36,'cp 0','d8f6 c3d5 f6d8 c2c3 a7a6 h2h3 b7b5 c4b3 c6e7 d3d4 e5d4 c3d4 c5a7 a2a4 c8b7 d5c3 b5b4 c3d5 b7d5 b3d5 e7d5 e4d5 a6a5 d1d3 a7b6 b2b3 d8f6 g2g3 f6g6 d3c4 f8e8 a1e1 g6f5 g1g2 g8f8 g3g4 f5g6 f1h1 h6h5 f3h4 g6g5 h4f3');
INSERT INTO evaluations VALUES(9122,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(9122,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(9122,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(9122,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(9122,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9122,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(9122,36,'cp -25','e4d5 f6d5 g2g3 b4c3 b2c3 d8e7 d1e2 e8g8 f1g2 e7a3 e1g1 d5c3 e2e1 c3b5 f3e5 f8e8 g2c6 b7c6 e1e4 c8h3 f1e1 a3d6 a1c1 c6c5 c2c3 f7f6 e4c4 e8e6 e5g4 a8e8 c4b5 h3g4 e1e6 e8e6 c1e1 e6e1 d2e1 g4d7 b5c4 d7e6 c4e4 d6d7 a2a4 e6d5 e4f4 c5c4 d3c4 d5c6');
INSERT INTO evaluations VALUES(9103,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(9103,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(9103,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(9103,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(9103,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(9103,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(9103,36,'cp -30','f6d5 g2g3 b4c3 b2c3 d8e7 d1e2 e8g8 f1g2 e7a3 e1g1 d5c3 e2e1 c3b5 f3e5 f8e8 g2c6 b7c6 e1e4 c8h3 f1e1 a3d6 a1c1 c6c5 c2c3 f7f6 e4c4 h3e6 c4b5 f6e5 e1e3 a8b8 b5a4 h7h6 c3c4 e5e4 e3e4 d6d3 a4c2 d3c2 c1c2 e6f5 e4e8 b8e8 c2b2 e8e2 d2c3 e2e4 b2b8 g8h7 f2f3 e4c4 b8b3 f5h3 g1f2 c4a4 b3b8 a4a2 b8b2 a2b2 c3b2 h3e6 b2c3 e6d5');
INSERT INTO evaluations VALUES(9180,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(9180,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(9180,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(9180,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(9180,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(9180,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(9180,36,'cp -31','c3d5 d8d5 d2b4 c6b4 f1e2 c8f5 d1d2 b4c6 d2g5 f5e6 e1g1 e8g8 c2c3 a8d8 b2b4 h7h6 g5e3 a7a5 a2a3 f8e8 h2h3 f7f6 f1e1 e6f7 e3c1 f7g6 d3d4 e5d4 c3d4 g8h7 b4a5 c6d4 f3d4 d5d4 e2f3 c7c6 e1d1 d4e5 a5a6 d8d1 c1d1 b7a6 f3c6');
INSERT INTO evaluations VALUES(9172,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(9172,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(9172,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(9172,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(9172,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(9172,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(9172,36,'cp -25','d8d5 d2b4 c6b4 d1d2 b4c6 f1e2 c8f5 d2g5 f5e6 e1g1 e8g8 f1d1 f8e8 c2c3 a8d8 g5g3 a7a5 h2h3 d5b5 d1d2 f7f6 d3d4 b5b6 d4e5 d8d2 f3d2 c6e5 b2b3 a5a4 a1e1 g8h8 g3f4 a4b3 a2b3 e6b3 e1b1 e5g6 f4f3');
INSERT INTO evaluations VALUES(9471,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(9471,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(9471,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(9471,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(9471,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(9471,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(9471,36,'cp 142','d1e2 d8e7 e5d3 c5d4 f2f3 e8g8 c2c3 d4b6 a2a4 a7a6 b1a3 f8e8 a3c4 b6a7 c4e3 f6h5 g2g4 h5f6 e3f5 c8f5 g4f5 a8d8 e1d1 e7d7 d3f4 f6d5 f4g2 c6c5 d1c2 d7f5 e2f1 f5g6 d2d3 d5e7 g2f4 g6f6 h2h4 a7b8 f1g2 b7b5 g2g5 f6g5 h4g5 c7c6 a4a5 c5c4');
INSERT INTO evaluations VALUES(10596,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(10596,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(10596,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(10596,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(10596,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(10596,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(10596,36,'cp 109','c5b4 b1d2 d8e7 a2a3 b4d2 d1d2 g4e5 f3e5 c6e5 c4c5 f7f6 a1c1 e8g8 f4e5 f6e5 f1c4 g8h8 c4a2 d7d6 c5d6 c7d6 e1g1 c8f5 f1d1 f8d8 a2d5 a8c8 h2h3 h7h6 b2b4 b7b6 c1c8 d8c8 e3e4 f5e6 d5e6 e7e6 d2d6 e6d6 d1d6 c8c3 d6d5 c3a3 d5e5 a7a5 b4a5 b6a5 e5c5');
INSERT INTO evaluations VALUES(13827,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13827,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13827,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13827,36,'cp 0','a7a6 b5e2 h7h6 c1e3 f8e7 f3d2 e8g8 d2c4 b7b5 c4b6 a8b8 b6c8 d8c8 e1g1 b5b4 c3d5 f6e4 e3a7 b8a8 a7e3');
INSERT INTO evaluations VALUES(11212,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11212,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11212,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11212,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11212,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11212,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11212,36,'cp 0','b5e2 f8e7 c1g5 c8e6 e1g1 a8c8 d1d2 d8a5 g5f6 e7f6 a1d1 e8g8 c3d5 f6d8 d2a5 c6a5 c2c3 g7g6 h2h4 a5c4 b2b3 c4a3 d1d2 a3b5 e2b5 a6b5 f3g5 e6d7 d5b4 d8e7 b4d5');
INSERT INTO evaluations VALUES(25889,30,'cp 232',NULL);
INSERT INTO evaluations VALUES(25889,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(25889,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(25889,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(25889,34,'cp 232',NULL);
INSERT INTO evaluations VALUES(25889,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(25889,36,'cp 217','d4d5 c6b8 c1g5 c7c6 g1f3 g8f6 d1d2 d7d6 e1c1 a7a6 g2g3 b6b5 f1h3 h7h6 g5f6 e7f6 h1e1 f8e7 d5c6 b8c6 f3h4 g7g6 c1b1 c6e5 b2b3 e8g8 d2h6 f8e8 h6e3 e5d7 e3d2 d7c5 h4g6 f7g6 d2h6 e7f8 h6g6 f8g7 h3f5 g8f8 c3d5 b7d5 d1d5 d8c7 e1d1 a8d8');
INSERT INTO evaluations VALUES(31099,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31099,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(31099,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(31099,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(31099,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31099,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(31099,36,'cp 30','e1g1 e7e6 d2d3 f5g6 c2c4 d5c4 d3c4 d8d1 f1d1 b8d7 b1c3 f8e7 c1e3 g8f6 h2h3 a7a5 a1c1 a5a4 f3h4 e7c5 e3g5 c5b4 h4g6 h7g6 c1c2 d7c5 g5e3 f6d7 a2a3 b4a5 c3e4 c5e4 g2e4 e8c8 h3h4');
INSERT INTO evaluations VALUES(31098,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31098,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(31098,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31098,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(31098,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31098,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31098,36,'cp 12','e7e6 d2d3 f5g6 c2c4 d5c4 d3c4 d8d1 f1d1 g8f6 b1c3 b8d7 c1f4 f8e7 a2a3 a7a5 b2b4 e8g8 b4b5 c6b5 c4b5 a8c8 f3e5 f8d8 e5g6 h7g6 c3a4 c8c4 a4b2 c4c5 a3a4 f6d5 g2d5 e6d5 a1c1 b7b6 b2d3 c5c4 c1c4 d5c4');
INSERT INTO evaluations VALUES(31917,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(31917,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(31917,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(31917,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(31917,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(31917,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(31917,36,'cp 59','g2f3 g8f6 e2e4 e7e6 c1e3 c7c6 f1d3 f8b4 h1g1 g7g6 e1f1 d8c7 e4e5 b4c3 b2c3 f6g8 a1b1 b8d7 d1d2 g8e7 e3g5 c6c5 d4c5 e7c6 f3f4 d7c5 h2h4 c6e7 g5f6 h8f8 d3b5 e7c6 c3c4 c5e4 d2e3 e4f6 e5f6 e8c8 b5c6 c7c6 e3a7 d5c4');
INSERT INTO evaluations VALUES(32855,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(32855,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(32855,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(32855,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(32855,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(32855,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(32855,36,'cp 156','d2d4 e5a5 h2h4 h7h6 c1d2 b8d7 g1h3 e7e5 d4e5 e8c8 f3f4 h5e2 d1e2 f8b4 e1g1 g8e7 g1h2 a5b6 d2e3 b6c6 c3b5 b4c5 e3c5 c6c5 c2c3 e7f5 b2b4 c5e3 f1e1 e3e2 e1e2 f5h4 b5a7 c8b8 a7b5 h4f5 b5d4 f5d4 c3d4 d7b6 a1d1');
INSERT INTO evaluations VALUES(32854,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(32854,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(32854,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(32854,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(32854,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(32854,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(32854,36,'cp 185','e5a5 h2h4 h7h6 c1d2 b8d7 g1h3 e7e5 d4e5 e8c8 f3f4 h5g6 h4h5 g6f5 h3f2 f8c5 e2f3 a5b6 f2e4 f5e4 c3e4 g8e7 e4c5 d7c5 d1e2 c8b8 e2f2 e7c6 d2c3 b6b5 f3e2 b5b6 a1d1 c5e4 f2b6 c7b6 d1d8 h8d8 e2d3 e4c3 b2c3');
INSERT INTO evaluations VALUES(37725,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(37725,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37725,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(37725,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(37725,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(37725,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(37725,36,'cp 41','d7d5 e4d5 c6d5 c4b3 g8f6 c2c3 b8c6 e1g1 c8f5 c1f4 e7e6 h2h3 e8g8 b1d2 h7h6 f4h2 a8c8 f1e1 g6g5 f3e5 c6e5 h2e5 f6e4 e5g7 g8g7 d2e4 f5e4 e1e3 a7a6 b3c2 e4c2 d1c2 c8c4 c2e2 b7b5 g1h1 d8c7 a1e1');
INSERT INTO evaluations VALUES(37917,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(37917,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(37917,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(37917,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(37917,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(37917,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(37917,36,'cp 49','e4d5 c6d5 c4b3 b8c6 e1g1 c8g4 c2c3 g4f3 d1f3 g8f6 f3e2 e8g8 c1f4 c6a5 b1d2 a5b3 a2b3 a7a5 h2h3 d8d7 d2f3 b7b6 f3e5 d7b7 e5d3 f8c8 g2g4 e7e6 f2f3 b6b5 h3h4 f6d7 g4g5 b7c6');
INSERT INTO evaluations VALUES(37915,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(37915,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(37915,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(37915,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(37915,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(37915,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(37915,36,'cp 34','c6d5 c4b3 g8f6 e1g1 e8g8 c2c3 b8c6 f1e1 c8f5 h2h3 a8c8 d1e2 h7h6 c1f4 g6g5 f4h2 e7e6 b1d2 a7a6 d2f1 f6e4 a1d1 f5g6 f1g3 b7b5 g3e4 g6e4 a2a3 c6a5 b3a2 a5c4 a2c4 b5c4 f3e5 a6a5 f2f3 g7e5 h2e5 e4d3 d1d3 c4d3 e2d3 d8b6');
INSERT INTO evaluations VALUES(38012,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(38012,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(38012,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(38012,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38012,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(38012,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(38012,36,'cp 41','c4b3 g8f6 c2c3 e8g8 e1g1 b8c6 h2h3 b7b5 a2a4 d8a5 c1d2 a7a6 f3e5 c6e5 d4e5 f6e4 b3d5 e4d2 b1d2 a8b8 a4b5 a5d8 d2b3 b8b5 d5c4 b5e5 d1d8 f8d8 f1d1 d8d1 a1d1 c8e6 c4e6 e5e6 b3c5 e6b6 d1d8 g7f8');
INSERT INTO evaluations VALUES(40749,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(40749,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(40749,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(40749,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(40749,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(40749,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(40749,36,'cp 120','e5g3 b8c6 f1e2 e8g8 d2d3 g8h8 g1h3 c6d4 e2d1 d6d5 e1g1 c7c6 a2a3 e7c5 c1e3 c5b6 a3a4 d5e4 d3e4 f8e8 a4a5 b6c7 g3h4 h8g8 h3f4 f6g4 h4d8 e8d8 d1g4 c8g4 h2h3 g4c8 f1c1 f7f5 a1a4 f5e4 a4d4 d8d4 e3d4 c7f4');
INSERT INTO evaluations VALUES(40856,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(40856,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(40856,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(40856,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(40856,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(40856,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(40856,36,'cp 131','b8c6 f1e2 e8g8 d2d3 g8h8 g1h3 d6d5 e1g1 c6d4 e2d1 c7c6 a2a3 e7c5 h3g5 h7h6 g5f3 c8e6 f3d4 c5d4 e4d5 f6d5 c3d5 d8d5 d1f3 d5a5 c2c3 d4f6 a3a4 a8d8 d3d4 a5f5 f1e1 e6d5 f3d5 d8d5 c1e3');
INSERT INTO evaluations VALUES(43285,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(43285,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(43285,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43285,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43285,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43285,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43285,36,'cp -18','e8g8 c2c4 d7d5 c4d5 f6d5 e1g1 b8c6 a2a3 a7a5 d1c2 c8g4 b1d2 d8d7 d2b3 g4f3 g2f3 e7e6 e2e3 b7b6 b3d2 a5a4 d2c4 b6b5 c4d2 f8c8 d3d4 c6a5 b2b4 a4b3 d2b3 a5c4 b3c5 d7e7 c1b2 c7c6 f1c1');
INSERT INTO evaluations VALUES(37474,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(37474,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(37474,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(37474,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(37474,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(37474,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(37474,36,'cp -12','c2c4 d7d5 e1g1 d5c4 d3c4 d8d1 f1d1 f6e4 f3d4 f8d8 g2e4 c7c5 b1c3 c5d4 e2e3 b8c6 e3d4 c6d4 c1f4 c8g4 e4b7 g4d1 c3d1 d4e2 g1f1 e2f4 b7a8 f4d3 a8f3 d3b2 d1b2 g7b2 a1d1 d8d1 f3d1 b2d4 f2f4 f7f5 f1e2 g8f7 e2f3 h7h6 f3g2 d4c5 g2f3 f7f6 f3e2 e7e6 d1a4');
INSERT INTO evaluations VALUES(47142,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(47142,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(47142,32,'cp -141',NULL);
INSERT INTO evaluations VALUES(47142,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(47142,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(47142,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(47142,36,'cp -150','d5c4 e2e3 c4b3 d1b3 e7e6 f1e2 b8d7 g1f3 b7b6 e1g1 c8b7 f1d1 f8e7 a2a4 e8g8 c1a3 a7a6 b1c3 a8b8 a1b1 h7h6 f3d2 c7c5 d4c5 b6c5 b3c2 d8c7 a3b2 f6d5 c3d5 e6d5 e3e4 e7d6 e2f3 d6h2 g1h1 d5e4 d2e4 b7e4 f3e4 h2e5 b2e5 d7e5');
INSERT INTO evaluations VALUES(47498,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(47498,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(47498,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(47498,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(47498,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(47498,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(47498,36,'cp 88','f1d3 e4f6 c2c4 e7e6 c4c5 b7b6 b2b4 a7a5 b1c3 a5b4 c3b5 b8a6 c5b6 c7b6 a1c1 b4b3 a2a3 b3b2 c1c2 f8e7 c2b2 e8g8 g1f3 c8d7 e1g1 d7b5 b2b5 a6b8 a3a4 b8d7 b5b1 f6h5 f4e5 h5f6 d1b3 a8c8 f1c1 c8c1 b1c1 d7e5 f3e5 e7d6 c1c6 d6e5');
INSERT INTO evaluations VALUES(51541,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(51541,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(51541,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(51541,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(51541,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(51541,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(51541,36,'cp 92','e2e4 c8b7 f1b5 b7e4 e1g1 c7c6 b5a4 d8a5 b1c3 e4g6 f1e1 e7e6 a2a3 g8f6 f3e5 f8e7 b2b4 a5b6 d2d4 e8g8 c1f4 d7d5 c3e2 b6a6 c2c4 d5c4 a1c1 g6d3 f4e3 d3e2 e1e2 f6d5 e5c4 d5b6 a4b3 b6c4 c1c4');
INSERT INTO evaluations VALUES(53037,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53037,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53037,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53037,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53037,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(53037,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53037,36,'cp 0','g8f6 b1c3 d8a5 f2f3 c5d4 d1d4 b8c6 d4d2 e7e6 g1h3 f8e7 f1e2 a5c5 h3f4 a7a6 b2b3 e8g8 c1b2 c8d7 f4d3 c5a5 d3f2 a8b8 a1d1 b7b5 e1g1 b5c4 e2c4 a5b6 d2e2 c6d4 e2d3 d4c6');
INSERT INTO evaluations VALUES(55235,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(55235,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(55235,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(55235,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(55235,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(55235,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(55235,36,'cp 144','b1c3 e7e6 e3e4 c7c5 d4d5 e6d5 e4d5 f8d6 c3e4 d6e7 c1f4 g8f6 e4c3 e7d6 d1e2 e8f8 e2e3 f6g4 e3d2 d6f4 d2f4 d8f6 f4g3 c8f5 h2h3 g4e5 f2f4 e5g6 e1c1 b8d7 g3f2 h7h5 g2g3 a8d8 g1f3 g6e7 d5d6 e7c6 f1b5');
INSERT INTO evaluations VALUES(55231,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(55231,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(55231,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(55231,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(55231,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(55231,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(55231,36,'cp 113','e7e6 e3e4 c7c5 d4d5 e6d5 d1d5 d8d5 e4d5 g8f6 c1f4 f8e7 e1c1 e8g8 g1f3 b8d7 f1e2 e7d8 f3d2 a7a6 e2f3 b7b5 h1e1 c5c4 d5d6 a8b8 d2e4 f6e4 f3e4 c8b7 e4f5 d7f6 d6d7 b8a8 d1d2 b7g2 f4d6 g2f3 d6f8 g8f8');
INSERT INTO evaluations VALUES(55837,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(55837,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(55837,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55837,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55837,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(55837,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55837,36,'cp 0','b7b6 e2e3 g8f6 g1f3 c8b7 f1d3 f8d6 e1g1 e8g8 c2c4 b8d7 b1d2 c7c5 c4d5 f6d5 a1c1 a8c8 d1e2 c5d4 f3d4 c8c1 f1c1 d8b8 h2h3 d7c5 d3c2 d6h2 g1h1 h2d6');
INSERT INTO evaluations VALUES(55901,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(55901,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(55901,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(55901,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(55901,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(55901,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(55901,36,'cp -51','b8c6 f3f4 d5e4 d3e4 d8d1 e1d1 e6e5 g1f3 g8f6 f1d3 f8c5 h2h3 c8e6 f4e5 f6d7 c1f4 h7h6 b1c3 e8c8 d1e2 a7a6 a1f1 g7g5 f4e3 c5f8 c3d5 d7e5 f3e5 c6e5 e3d4 f8g7 d5e7 c8b8 e7f5 e6f5 d4e5');
INSERT INTO evaluations VALUES(56266,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(56266,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(56266,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(56266,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(56266,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(56266,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(56266,36,'cp -22','g8f6 g1f3 f8e7 d2d4 e8g8 f1d3 b8c6 c2c3 c8g4 h2h3 g4h5 b1d2 e7d6 e1g1 c6e7 c3c4 h5g6 d1c2 g6d3 c2d3 c7c6 a1e1 f6h5 d3c2 e7g6 f3e5 h5f4 d2f3 d6e5 d4e5 d5c4 c2c4 f4d3 e1d1 d3b2 d1d8 b2c4 d8a8 f8a8');
INSERT INTO evaluations VALUES(48722,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(48722,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48722,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(48722,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48722,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48722,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(48722,36,'cp -17','g1f3 f8e7 d2d4 e8g8 f1d3 b8c6 c2c3 e7d6 e1g1 c6e7 b1d2 c8f5 d3f5 e7f5 c3c4 g7g6 d1c2 c7c6 a1e1 a7a5 a2a4 f8e8 f3e5 d6b4 e5d3 e8e1 f1e1 d5c4 d3b4 c4b3 c2c3 a5b4 c3b4 d8d6 b4b7 a8b8 b7a6 d6d5 a6a7 b8d8 a7c7 f5d4');
INSERT INTO evaluations VALUES(48721,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(48721,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(48721,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(48721,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(48721,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(48721,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(48721,36,'cp -28','f8e7 d2d4 e8g8 f1d3 b8c6 c2c3 e7d6 e1g1 c6e7 c3c4 c7c6 f1e1 c8g4 b1c3 f8e8 h2h3 g4h5 d3f1 d8d7 g2g3 a8d8 f1g2 h7h6 d1d3 d6b4 g3g4 h5g6 d3f1 d7c7 f3e5 g6h7 a2a3 b4d6 c4c5');
INSERT INTO evaluations VALUES(8314,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8314,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8314,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8314,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(8314,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(8314,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8314,36,'cp 0','f6g4 d4c6 b7c6 e3f4 g4e5 c4e2 c8e6 b2b3 e5c4 e2c4 e6c4 b3c4 g7c3 a1b1 e8g8 d1d3 d8a5 b1b7 f8e8 f4e3 c3b4 f1b1 b4c5 e3c5 a5c5 g2g3 a7a5 a2a4 a8a7 b7b8 a7a8');
INSERT INTO evaluations VALUES(10219,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(10219,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(10219,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(10219,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(10219,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(10219,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(10219,36,'cp -60','e8g8 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 g2g3 c6e7 d5e7 d8e7 f1h3 b7b5 a2a3 a7a5 e1g1 c8b7 c2c3 d7d6 f1e1 b5b4 a3b4 a5b4 h3g2 b4c3 b2c3 b7c6 d3d4 a8a1 e1a1 e5d4 c3d4 c5b6 e4e5 e7e6 e2d3 f8d8 e5d6 e6d6');
INSERT INTO evaluations VALUES(12006,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(12006,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(12006,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(12006,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(12006,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(12006,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(12006,36,'cp 113','e4e5 f6h5 f1c4 d8a5 f2f4 f8h6 d1f3 a8b8 a2a3 d7d5 e5d6 a5c5 b2b3 h5f6 c1e3 c5d6 e1g1 e8g8 e3a7 b8a8 a7e3 c8f5 b3b4 h6g7 e3c5 d6c7 f1e1 e7e6 c5f8 a8f8 a1c1 e6e5 g1h1 e5e4 f3g3 f6h5 g3e3 c7f4');
INSERT INTO evaluations VALUES(12003,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(12003,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(12003,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(12003,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(12003,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(12003,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(12003,36,'cp 123','f6h5 f1c4 d8a5 f2f4 a8b8 d1f3 f8h6 a2a3 h5g7 c1d2 e8g8 e1c1 g7f5 f3d3 a5c7 g2g4 d7d5 c4b3 c7b6 c3a4 b6b5 d3b5 c6b5 a4c5 f5d4 g4g5 d4b3 c5b3 h6g7 d2b4 f8e8 d1d3 h7h6 h1g1 h6g5 f4g5 e7e6 d3c3 g7f8');
INSERT INTO evaluations VALUES(12470,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(12470,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(12470,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(12470,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(12470,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12470,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(12470,36,'cp 37','f8b4 f1d3 d7d5 c4d5 e6d5 e1g1 e8g8 c1g5 b4c3 b2c3 c8g4 h2h3 g4h5 a1b1 h7h6 g5h4 b7b6 b1b5 d8d6 a2a4 a8c8 f1e1 c6a5 e1e5 h5g6 h4f6 d6f6 b5d5 g6d3 d1d3 c8c7 e5e2 f6c6 d3b5 c6c3 d5d7 c7d7 b5d7 c3c1 g1h2 a5c4 d7a7');
INSERT INTO evaluations VALUES(7850,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(7850,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(7850,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(7850,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(7850,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(7850,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(7850,36,'cp 44','f1d3 d7d5 c4d5 e6d5 e1g1 e8g8 c1g5 b4c3 b2c3 c8g4 a1b1 h7h6 g5h4 b7b6 f1e1 d8d6 h4g3 d6d7 d3b5 f8c8 h2h3 g4h5 f3e5 h5d1 e5d7 f6d7 b1d1 a7a6 b5d3 d7f6 g3h4 f6h5 g1h2 g7g5 h4g3 c6a5 e1e5 h5f6 d1c1 a5c4 d3c4 c8c4');
INSERT INTO evaluations VALUES(12925,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(12925,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(12925,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(12925,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(12925,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(12925,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(12925,36,'cp 200','c3d4 e7e6 f1b5 d7d6 h2h3 g4h6 c1h6 g7h6 e1g1 f8g7 b1c3 e8g8 b5c6 b7c6 e5d6 d8d6 c3e4 d6b4 f1e1 b4b2 f3e5 g7e5 d4e5 b2e5 a1c1 f7f5 e4c5 e5f6 d1b3 f8e8 c5d3 c8d7 d3e5 a8d8 b3a3 d7c8 c1c3 f5f4 c3c6');
INSERT INTO evaluations VALUES(12924,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(12924,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(12924,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(12924,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(12924,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(12924,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(12924,36,'cp 157','d7d5 h2h3 g4h6 b1c3 c8f5 d1b3 d8d7 f1b5 e7e6 e1g1 f8e7 c1h6 g7h6 c3a4 e8g8 b3e3 g8h8 a4c5 e7c5 d4c5 d7e7 b5c6 b7c6 f3d4 a8c8 b2b4 f8g8 g1h2 e7g5 g2g4 f5g4 e3g5 g8g5 h3g4 g5g4 d4f3 g4b4 f1g1 b4b2 h2g3 c8g8 g3f4 b2f2 g1g8 h8g8 f4e3');
INSERT INTO evaluations VALUES(13959,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(13959,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13959,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(13959,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(13959,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13959,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(13959,36,'cp 18','d3d4 c5d4 f3d4 f8e7 f1e2 e8g8 e1g1 d6d5 e4d5 e6d5 c1e3 c6d4 e3d4 d5c4 e2c4 c8g4 d1a4 g4h5 c4b3 a7a6 d4e3 b7b5 a4h4 d8c7 f1e1 h5g6 a1c1 c7b7 h4d4 f8d8 d4b6 b7b6 e3b6 d8d7 e1d1 g6f5');
INSERT INTO evaluations VALUES(14387,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(14387,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(14387,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(14387,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(14387,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(14387,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(14387,36,'cp 210','e2e4 e7e6 a2a3 a7a6 f1d3 f6d7 c1e3 e6e5 d4d5 c6a7 e1g1 f8d6 f3d2 d6c5 d1e2 c5e3 e2e3 e8g8 a3a4 h7h6 b2b4 d8g5 e3g5 h6g5 f1d1 f8e8 a1c1 d7f6 d2b3 g5g4 d3e2 a8b8 b3c5 c7c6 a4a5 e8d8 d5d6 c8d7 c3a4 a7b5 e2b5');
INSERT INTO evaluations VALUES(18455,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(18455,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(18455,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(18455,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(18455,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(18455,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(18455,36,'cp -29','c5d4 d1d4 e7e5 d4a4 c8d7 c2c4 f7f6 e1g1 g8e7 a4d1 d7e6 d1a4 d8d7 b1c3 d6d5 c4d5 c6d5 a4d7 e8d7 e4d5 e6d5 f3d2 d5b7 d2b3 e7d5 f2f4 d5c3 b2c3 b7a6 f1d1 d7e6 f4e5 f6e5 c1e3 a8c8 e3a7 f8e7 a1b1 c8c3 b3c5 c3c5 b1b6 e6f7 b6a6');
INSERT INTO evaluations VALUES(18465,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(18465,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(18465,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(18465,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(18465,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(18465,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(18465,36,'cp -54','d1d4 e7e5 d4a4 c8d7 c2c4 f7f6 e1g1 g8e7 a4d1 d7e6 d1a4 d8d7 b1c3 d6d5 c4d5 c6d5 a4d7 e8d7 e4d5 e6d5 f3d2 d5c6 d2b3 d7e6 f2f4 e7f5 f4e5 f6e5 f1e1 a7a5 c1f4 f8d6 a1d1 h8d8 d1d6 f5d6 e1e5 e6f7 e5a5 d6e4 a5a8 c6a8 c3e4 a8e4');
INSERT INTO evaluations VALUES(20419,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(20419,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(20419,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(20419,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(20419,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(20419,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(20419,36,'cp -49','g8h6 b1c3 h6f5 g1f3 c8d7 g2g3 d8b6 c3e2 h7h5 f1h3 f8b4 e1f2 f5h6 h1f1 a8c8 f2g1 c6e7 a2a3 b6a6 c1d2 b4d2 d1d2 d7b5 f1e1 b5e2 e1e2 e8g8 e2e1 a6b6 d2b4 b6c7 b4d2 e7f5 a1c1 c7d7 g1f2 a7a6 c1c5 b7b6 c5c1 c8c4');
INSERT INTO evaluations VALUES(30240,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30240,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(30240,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(30240,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30240,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30240,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30240,36,'cp 0','f8d6 c1g5 c7c6 f3e5 g4e2 d1e2 e8g8 e1c1 f8e8 f2f4 d6b4 e2f3 b4c3 b2c3 b8d7 e5d7 d8d7 g5f6 g7f6 h1e1 e8e1 d1e1 a8e8 e1e8 d7e8 f3g3 g8h8 f4f5 e8d8 a2a3 h7h6 c1d2 h8h7 a3a4 h7h8 g3h4 h8h7');
INSERT INTO evaluations VALUES(31601,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(31601,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(31601,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(31601,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(31601,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(31601,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(31601,36,'cp 169','d4e5 g4f3 d1f3 d6e5 h2h4 b8d7 g2g4 f8c5 g4g5 g8e7 c1d2 e7c8 e1c1 c8b6 a2a4 a7a5 c1b1 d8e7 b2b3 c5b4 f1h3 e8g8 h4h5 d7c5 h3f5 c5e6 f5e6 f7e6 f3g3 b6d7 d2c1 b4c5 c1b2 c5f2 g3g4 d7c5 c3e2 f2e3 g5g6 h7h6');
INSERT INTO evaluations VALUES(31586,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(31586,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(31586,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(31586,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(31586,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(31586,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(31586,36,'cp 182','g4f3 d1f3 d6e5 h2h4 b8d7 g2g4 f8c5 g4g5 g8e7 c1d2 e7c8 e1c1 d8e7 c3a4 c5d4 c1b1 b7b5 a4c3 c8b6 c3e2 e8c8 e2d4 e5d4 d2c1 c8b7 d1d4 d7e5 f3c3 d8d4 c3d4 h8d8 d4e3 b6c4 f1c4 e5c4 e3f4 e7c7 f2f3 c7f4 c1f4 b7c8');
INSERT INTO evaluations VALUES(31588,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(31588,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(31588,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(31588,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(31588,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(31588,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(31588,36,'cp 166','d1f3 d6e5 g2g4 b8d7 h2h4 f8c5 g4g5 g8e7 c1d2 e7c8 e1c1 d8e7 c3a4 c5d4 c1b1 d7b6 a4c3 c8d6 c3e2 d4c5 e2g3 e8c8 d2a5 g7g6 b2b4 d6c4 f3g4 c8b8 f1c4 d8d1 h1d1 b6c4 d1d7 e7e6 g4e6 f7e6 b4c5 c4a5 h4h5 b7b5 d7g7 g6h5 g3h5 a5c4 b1c1 a7a5 c2c3 b8a8');
INSERT INTO evaluations VALUES(31587,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(31587,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(31587,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(31587,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(31587,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(31587,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(31587,36,'cp 179','d6e5 g2g4 f8c5 h2h4 b8d7 g4g5 g8e7 c1d2 e7c8 e1c1 c8d6 f1h3 e8g8 c1b1 d8e7 h4h5 d7b6 d2c1 a8d8 d1d3 d6c4 f3g3 d8d6 h1d1 f8e8 a2a3 d6d3 d1d3 c4d6 g5g6 f7g6 h5g6 h7h6 g3g4 e7f6 c3e2 b6c4 e2g3');
INSERT INTO evaluations VALUES(31582,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(31582,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(31582,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(31582,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(31582,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(31582,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(31582,36,'cp 182','h2h4 b8d7 g2g4 f8c5 g4g5 g8e7 c1d2 e7c8 e1c1 c8d6 f1h3 d7b6 c1b1 e8g8 d2c1 d8e7 h4h5 a8d8 d1d3 d6c4 f3g3 d8d4 h1d1 f8e8 h3g4 d4d3 c2d3 c4d6 g5g6 h7h6 f2f4 f7f6 c3e2 b6d7 g3h3 d7f8 f4e5 f6e5 d1f1 g8h8 e2g3 f8e6 g4e6 e7e6 h3e6 e8e6');
INSERT INTO evaluations VALUES(35335,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(35335,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(35335,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(35335,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(35335,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(35335,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(35335,36,'cp 121','c2c4 g8f6 b1c3 e7e5 c1e3 b8d7 h2h3 f8e7 f1d3 d7f8 e1g1 f8g6 f1e1 e8g8 a1c1 a7a5 c3a4 f8e8 d1c2 c8d7 c4c5 e5d4 e3d4 d6c5 a4c5 e7c5 c2c5 g6f4 d3b1 f4e6 c5e5 c7e5 d4e5 e6g5 f3g5 e8e5 g5f3 e5h5 e1d1 a5a4');
INSERT INTO evaluations VALUES(35331,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(35331,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(35331,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(35331,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(35331,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(35331,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(35331,36,'cp 124','b8d7 c1e3 g8f6 b1c3 f6g4 e3d2 g4f6 h2h3 e7e5 d2e3 f8e7 f1d3 d7f8 e1g1 f8g6 a1c1 a7a5 c3a4 e8g8 f1e1 h7h6 d1c2 f6d7 c1d1 b7b6 a4c3 d7f6 a2a3 f6h5 b2b4 h5f4 d3f1 a5b4 a3b4 g8h8');
INSERT INTO evaluations VALUES(37457,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(37457,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(37457,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(37457,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(37457,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(37457,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(37457,36,'cp 25','c2c4 c7c5 d4c5 b8a6 e1g1 a6c5 b2b4 c5e4 c1b2 a7a5 a2a3 a5b4 a3b4 a8a1 b2a1 b7b6 f3d2 e4d2 b1d2 c8b7 a1d4 d8c7 d1c2 f8c8 f1c1 d7d6 h2h3 c8a8 e2f1 h7h5 c2b2 b7c6 f1e2 e7e5 d4c3 f6d7 c1d1 d7f6 b4b5 c6b7 d1a1 a8a1 b2a1 f6d7 c3b4');
INSERT INTO evaluations VALUES(38355,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(38355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38355,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(38355,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38355,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38355,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(38355,36,'cp 8','g8f6 b1c3 c7c5 d2d4 c5d4 f3d4 e8g8 f1e1 b8c6 c4b3 c6d4 d1d4 a7a6 a2a4 f6d7 d4b4 d7c5 a4a5 b7b5 a5b6 a8b8 c1e3 b8b6 b4a3 b6c6 b3c4 c8b7 c4d5 c6c7 d5b7 c7b7 b2b4 c5e6 c3d5 g7a1 e1a1 e6c7 e3b6 d8d7 b6c7 b7c7 d5c7 d7c7 a3a6');
INSERT INTO evaluations VALUES(41237,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(41237,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(41237,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41237,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(41237,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41237,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41237,36,'cp 13','b1c3 c7c5 d2d4 c5d4 f3d4 e8g8 f1e1 b8c6 c4b3 c6d4 d1d4 c8d7 d4d3 a8c8 c1e3 a7a6 h2h3 d7c6 e3d4 f6d7 d4g7 g8g7 f2f3 d8b6 g1h1 b6c5 d3d2 a6a5 a2a3 f8e8 a1d1 b7b5 f3f4 b5b4 a3b4 c5b4 e4e5');
INSERT INTO evaluations VALUES(41366,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(41366,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(41366,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(41366,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(41366,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(41366,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(41366,36,'cp 52','g1f3 e8g8 e1g1 e7e5 d4e5 d6e5 f3e5 f6e4 c3e4 d8d1 f1d1 g7e5 c2c3 f8e8 c1g5 g8g7 d1e1 b8c6 e2b5 h7h6 g5e3 g6g5 f2f3 f7f6 e3f2 c8f5 e4c5 a7a6 b5c4 c6a5 c4f1 h6h5 b2b4 a5c6 c5b7 e5c3 e1e8 a8e8');
INSERT INTO evaluations VALUES(42443,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42443,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42443,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(42443,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(42443,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42443,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42443,36,'cp 0','h7h6 g5f6 d8f6 d1d3 d7d5 e1c1 e8g8 f2f3 c7c5 e4d5 e6d5 d4c5 b4c3 d3c3 f6c3 b2c3 c8e6 g1e2 b8d7 e2f4 f8c8 c3c4 c8c5 c4d5 e6f5 f1d3 f5d3 d1d3 a8c8 d3d2 d7b6 d5d6 b6c4 d2d4 c4a3 c1b2 a3c4 b2c1');
INSERT INTO evaluations VALUES(42343,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(42343,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42343,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(42343,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(42343,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42343,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(42343,36,'cp 27','g5f6 d8f6 d1d3 d7d5 e1c1 e8g8 c1b1 b8d7 e4d5 d7b6 a2a3 b4c3 d3c3 b6d5 c3g3 b7b6 g1f3 c8b7 f3e5 f6f4 f1d3 f4g3 h2g3 d5e7 f2f3 f7f6 e5c4 g8f7 c2c3 f8h8 a3a4 a7a5 b1c2 h6h5 d1e1 a8d8 b2b3 e7f5 d3f5 e6f5 c4e3 g7g6 e3c4');
INSERT INTO evaluations VALUES(42341,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(42341,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(42341,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(42341,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(42341,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(42341,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(42341,36,'cp 11','d8f6 d1d3 d7d5 e1c1 e8g8 c1b1 c7c5 e4d5 c5d4 d3d4 f6d4 d1d4 b4c3 b2c3 e6d5 d4d5 c8e6 d5d6 b8d7 f1d3 f8c8 g1e2 d7b6 d6d4 g8f8 e2f4 e6d7 c3c4 c8c5 f4d5 d7c6 d5e3 b6a4 d4f4 c5a5 h1e1 a4c3 b1b2 c3a4 b2c1 a4b6 a2a3 a8d8 d3g6 c6e8 g2g3');
INSERT INTO evaluations VALUES(33442,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(33442,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(33442,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(33442,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(33442,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33442,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(33442,36,'cp 22','d1d3 d7d5 e1c1 e8g8 c1b1 c7c5 e4d5 c5d4 d3d4 f6d4 d1d4 b4c3 b2c3 e6d5 d4d5 c8e6 d5d6 b8d7 f1d3 d7b6 d6d4 e6d5 f2f3 a8c8 g1e2 f8e8 h1d1 g8f8 b1b2 e8e5 d4b4 d5c4 e2d4 e5a5 d4b3 a5d5 b2c1 c4d3 d1d3 d5d3 c2d3 c8c3 c1d2');
INSERT INTO evaluations VALUES(43147,30,'cp 305',NULL);
INSERT INTO evaluations VALUES(43147,31,'cp 314',NULL);
INSERT INTO evaluations VALUES(43147,32,'cp 312',NULL);
INSERT INTO evaluations VALUES(43147,33,'cp 326',NULL);
INSERT INTO evaluations VALUES(43147,34,'cp 313',NULL);
INSERT INTO evaluations VALUES(43147,35,'cp 332',NULL);
INSERT INTO evaluations VALUES(43147,36,'cp 343','e4e5 c7c6 e5f6 g7f6 c4b3 d7d5 g1e2 e7e5 d2d3 a7a5 a2a4 e8g8 c1h6 f8e8 e1g1 b8a6 b1d2 a6c5 f3e3 b7b6 f2f4 e5e4 d3d4 c5e6 a1e1 e6g7 e2g3 b6b5 c2c3 b5b4 h6g5 b4c3 b2c3 a8b8 g5f6 d8f6 f4f5 c8f5 f1f2 g7e6 e1f1 b8b6');
INSERT INTO evaluations VALUES(43122,30,'cp 298',NULL);
INSERT INTO evaluations VALUES(43122,31,'cp 322',NULL);
INSERT INTO evaluations VALUES(43122,32,'cp 324',NULL);
INSERT INTO evaluations VALUES(43122,33,'cp 330',NULL);
INSERT INTO evaluations VALUES(43122,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(43122,35,'cp 318',NULL);
INSERT INTO evaluations VALUES(43122,36,'cp 331','c7c6 e5f6 g7f6 g1e2 d7d5 c4b3 e7e5 d2d3 a7a5 a2a4 e8g8 e1g1 b8d7 c1h6 f8e8 f3e3 b7b6 b1d2 d7c5 f2f4 e5e4 d3d4 c5b7 e2g3 b7d6 a1e1 c8e6 h6g5 h7h5 c2c4 d5c4 b3c2 d6f5 g3f5 f6g5 f4g5 g6f5');
INSERT INTO evaluations VALUES(44462,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(44462,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(44462,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(44462,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(44462,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(44462,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(44462,36,'cp 35','e4e5 f6d5 g1f3 d7d6 e5d6 e7e6 d2d4 f8d6 b1d2 d6e7 e1g1 e8g8 f1e1 b8d7 d2e4 b7b6 d1d3 c8b7 f3e5 f7f5 e4g3 d7e5 e1e5 d8d7 d3e2 e7f6 e5e6 f6d4 c2c3 a8e8 e6e8 f8e8 e2d1 d4c5 g3h5 c5d6 g2g3');
INSERT INTO evaluations VALUES(45963,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(45963,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45963,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(45963,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45963,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45963,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(45963,36,'cp 0','c8d7 b5e2 b8c6 c1e3 f8e7 g2g4 h7h6 h1g1 a8c8 h2h4 c6e5 d1d2 d6d5 e4d5 c8c3 d2c3 f6e4 c3b3 e4c5 b3c3');
INSERT INTO evaluations VALUES(28850,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(28850,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28850,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28850,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28850,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28850,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(28850,36,'cp 0','c1e3 f8e7 d1e2 a7a6 b5d3 b8c6 e1g1 e8g8 f2f4 b7b5 a2a3 d8c7 g1h1 c6d4 e3d4 d7c6 a1d1 a8d8 h2h3 c6b7 e2f2 d8e8 d4b6 c7c8 b6d4');
INSERT INTO evaluations VALUES(47262,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(47262,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(47262,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(47262,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(47262,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(47262,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(47262,36,'cp -50','c7c5 e2e3 b8c6 g2g3 e7e5 f1g2 f8e7 e1g1 c8e6 f3g5 e6g4 f2f3 g4d7 f3f4 e5f4 g3f4 d7g4 d1e1 d5d4 c3b5 d4e3 c1e3 e8g8 e1f2 g4f5 h2h3 d8b6 b5a3 h7h6 a3c4 b6c7 g5f3 f6d5 f3e5 a8d8 e5c6 c7c6 a1e1');
INSERT INTO evaluations VALUES(49965,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(49965,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(49965,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(49965,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(49965,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(49965,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(49965,36,'cp 43','c2c4 d7d5 g1h3 f8d6 e1g1 e8g8 b2b3 c7c6 c1f4 d6e7 b1d2 h7h6 f4b8 a8b8 h3f4 e7d6 f4g6 f8e8 d2f3 b7b6 e2e3 c6c5 f3e5 c8b7 a1c1 b8c8 d1e2 g8h7 f1d1 d8c7 c4d5 b7d5 g2d5 f6d5 h2h4 d6e5 g6e5');
INSERT INTO evaluations VALUES(51023,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(51023,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(51023,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(51023,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(51023,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(51023,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(51023,36,'cp -28','d7d5 e4d5 e6d5 c4b3 c7c5 c2c3 d8e7 d1e2 e7e2 e1e2 g8f6 d2d4 c5c4 b3c2 f8d6 h1e1 e8g8 e2f1 c8e6 f3e5 b8c6 b1d2 c6e7 d2f3 h7h6 c1f4 f8d8 h2h3 b7b5 a2a3 a6a5 f1g1 d8e8 f4g3 e6f5 c2f5 e7f5 g3h2');
INSERT INTO evaluations VALUES(52042,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(52042,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(52042,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(52042,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(52042,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(52042,35,'cp -128',NULL);
INSERT INTO evaluations VALUES(52042,36,'cp -141','b8a6 b1c3 a6c5 e1g1 b7b6 c3d5 c8b7 f1e1 a8c8 a1c1 g8f6 d5f6 g7f6 c4d5 f8e7 d5b7 c5b7 c1b1 b7a5 d1e2 a5c4 b2d4 h8g8 g2g3 e8f8 e1c1 d7d5 c1d1 g8g6 b1c1 d5e4 e2e4 f6f5 e4e2 c4d6 d4b2 c8c1');
INSERT INTO evaluations VALUES(54145,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(54145,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(54145,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54145,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54145,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(54145,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54145,36,'cp 0','e6d5 d3d4 b8c6 f1b5 f8d6 d4c5 d6c5 e1g1 g8e7 b1d2 c8g4 d2b3 c5d6 h2h3 g4h5 b3d4 e8g8 b5e2 f8e8 c1e3 h7h6 c2c3 d8d7 d1d2 a7a6 e3f4 c6d4 f3d4 d6f4 d2f4 h5e2 d4e2 e7f5 f4d2 e8e4 f2f3 e4e5 e2d4 a8e8 d4f5 d7f5');
INSERT INTO evaluations VALUES(54791,30,'cp 321',NULL);
INSERT INTO evaluations VALUES(54791,31,'cp 359',NULL);
INSERT INTO evaluations VALUES(54791,32,'cp 350',NULL);
INSERT INTO evaluations VALUES(54791,33,'cp 360',NULL);
INSERT INTO evaluations VALUES(54791,34,'cp 378',NULL);
INSERT INTO evaluations VALUES(54791,35,'cp 361',NULL);
INSERT INTO evaluations VALUES(54791,36,'cp 377','d1h5 e8d7 f4g3 e7e6 h2h4 d8e8 h4g5 e8h5 h1h5 b8c6 c2c4 d5c4 b1c3 f6g5 h5g5 h7h6 g5h5 g8f6 h5h4 f8d6 g3d6 c7d6 f1c4 a7a6 g1e2 b7b5 c4d3 c8b7 e2f4 a8g8 f2f3 h6h5 e1f2 c6b4 d3b1 a6a5');
INSERT INTO evaluations VALUES(54987,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(54987,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(54987,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(54987,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(54987,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(54987,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(54987,36,'cp -55','h2h3 h7h5 g4g5 e7e5 d2d4 e5d4 g1f3 d6d5 f1g2 f8b4 c1d2 b4c5 c2c3 g8e7 c3d4 c5b6 b1c3 b8c6 e2e3 e7g6 d1e2 e8g8 e1c1 c6e7 f3e1 h5h4 c1b1 c7c6 b1a1 f8e8 e2h5 a7a5 e1d3 b6c7 d2c1 a5a4 a2a3 c8f5 h5e2 c7d6 e3e4 d5e4');
INSERT INTO evaluations VALUES(56442,30,'cp -153',NULL);
INSERT INTO evaluations VALUES(56442,31,'cp -140',NULL);
INSERT INTO evaluations VALUES(56442,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(56442,33,'cp -152',NULL);
INSERT INTO evaluations VALUES(56442,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(56442,35,'cp -157',NULL);
INSERT INTO evaluations VALUES(56442,36,'cp -127','h7h5 h2h4 g8f6 d2d3 b8c6 f1g2 c8f5 g1f3 d8d7 b2b4 f8d6 e1g1 e8g8 f3g5 e5e4 d3e4 f6e4 g5e4 d5e4 b1d2 f5g4 d1b3 g4e6 c3c4 f7f5 a2a3 c6e5 b3c2 e5d3 c1b2 c7c5 b4c5 d6c5 d2b3 f8c8 f1d1 c5b6 c4c5 d7f7 d1d3 e4d3 c2d3 b6c5 b3c5 c8c5 d3d4 a8c8 a1d1');
INSERT INTO evaluations VALUES(57104,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(57104,31,'cp -123',NULL);
INSERT INTO evaluations VALUES(57104,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(57104,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(57104,34,'cp -113',NULL);
INSERT INTO evaluations VALUES(57104,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(57104,36,'cp -116','f8c5 c2c3 b8c6 g1f3 d4c3 b1c3 d7d6 e5d6 d8d6 d1d6 c7d6 c3b5 e8e7 c1g5 f7f6 g5f4 a7a6 b5c7 a8a7 a1d1 b7b5 c7d5 e7f7 d5c3 g8e7 f4d6 c5b6 f1e2 e7f5 d6f4 c8e6 e1g1 c6d4 f3d4 f5d4 e2d3 h8d8 f4e3 a7d7 g1h1 h7h6 h2h3 b6a7 b2b3 f6f5 c3e2 g7g5 e2d4 a7d4 e3d4 d7d4 d3e2 f7g7 d1d4');
INSERT INTO evaluations VALUES(57452,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(57452,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(57452,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(57452,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(57452,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(57452,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(57452,36,'cp 75','d7d5 e4d5 f8d6 e1g1 g8e7 d2d4 g7g5 c4b3 e8g8 c2c4 b7b6 b1c3 c8f5 c3b5 b8d7 c1d2 f8f7 a1c1 f5g6 c4c5 b6c5 d2a5 d8b8 b5d6 c7d6 d4c5 d6c5 a5c7 b8c7 d5d6 c7d8 d6e7 d8e7 f1e1 e7d8 d1d6 g8g7 b3f7 g6f7 e1e7 d7e5 d6c5 e5f3 g2f3 a8c8 c5c8');
INSERT INTO evaluations VALUES(366,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(366,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(366,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(366,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(366,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(366,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(366,36,'cp 154','b1c3 f6g6 d1c2 g8e7 f2f3 e8g8 f1d3 b4a5 e1g1 d7d5 a1c1 a5b6 g1h1 g8h8 c3d5 e7d5 e4d5 g6d6 c2b3 c8d7 e3f2 d6f6 f2g3 a8c8 f1d1 f8e8 a2a4 h8g8 g3e5 f6g5 d3b5 c7c6 f3f4 g5g6 d5c6 d7c6 b5c6 c8c6');
INSERT INTO evaluations VALUES(365,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(365,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(365,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(365,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(365,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(365,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(365,36,'cp 166','f6g6 f2f3 g8e7 h2h4 h7h5 d1b3 b4c3 b3c3 c7c6 f1d3 d7d5 e1g1 e8g8 a1e1 c8e6 e3f4 f8e8 f1f2 a8d8 d3c2 d5e4 f3e4 g6f6 f4e5 f6h6 e5g3 d8d7 a2a4 e7g6 g1h2 e8d8 e1d1 g6f8 d1d2 e6g4 g3f4 h6e6 a4a5 e6e7 f4g5');
INSERT INTO evaluations VALUES(2458,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(2458,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(2458,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(2458,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(2458,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(2458,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(2458,36,'cp -113','b4c2 e1d1 c2a1 c3d5 d8h4 d5b6 e6e7 b6c8 a8c8 d2d3 e7d8 g2g3 h4f6 f3b7 f8d6 c1e3 d8d7 b7e4 c8b8 e4g4 d7e8 b2b4 a7a5 d1d2 a5b4 a3a4 f6g6 g4g6 h7g6 h1a1 b4b3 a1b1 b3b2 d2c2 d6a3 c4b5 c7c6 b5c6 e8e7 c6b5 h8c8 c2b3 c8c1 b1b2 a3b2 b3b2 c1c8 b2b3 e7d6 b3b4 d6d5 h2h4 b8b7 b4a5 c8c2 e3b6 d5d6 a5a6 b7e7 b6d8 e7f7 d8b6 c2b2 b6e3 b2b4 a6b6');
INSERT INTO evaluations VALUES(4506,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(4506,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(4506,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(4506,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(4506,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(4506,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(4506,36,'cp 41','d4d5 c6a5 c4d3 f7f5 d5d6 e4d6 f1e1 f8e8 c1g5 e8e1 d1e1 d8e8 e1e8 d6e8 a1e1 e8d6 g5d8 b7b6 d8c7 d6e4 f3d4 g7g6 d3e4 f5e4 e1e4 g8f7 d4f3 a5c6 e4e2 d7d5 e2d2 c8f5 d2d5 a8c8 c7f4 c8d8 d5d8 c6d8 g2g4 f5b1 a2a3 d8c6 g1g2 c6a5 g2g3 a5c4');
INSERT INTO evaluations VALUES(5130,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5130,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(5130,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(5130,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5130,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(5130,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5130,36,'cp 16','h2h4 h7h6 e1g1 g8e7 e2e3 d7d5 c4d5 e6d5 d2d4 c5d4 f3d4 e8g8 f1e1 f8e8 c1d2 c6d4 e3d4 c8e6 d2f4 e7f5 d1d2 d8d7 c3e2 a8c8 a1c1 c8c1 e1c1 e8c8 c1c8 d7c8 f4e5 f7f6 e5f4 g6g5 h4g5 h6g5 f4e3 c8d7 a2a3 e6f7 d2d3 g7f8 e2c3');
INSERT INTO evaluations VALUES(6645,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(6645,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(6645,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(6645,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(6645,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6645,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6645,36,'cp 16','b2c3 d7d5 e4d5 d8d5 f1e2 g8f6 e1g1 e8g8 c3c4 d5d6 a2a4 f8e8 c1a3 d6d8 f1e1 h7h6 f3d2 c6d4 a4a5 b7b6 e2f3 a8b8 a3b2 c8f5 h2h3 d8d6 a5b6 a7b6 a1a7 d4f3 d2f3 e5e4 d3e4 e8e4 d1d6 c7d6 b2f6 g7f6 a7c7 e4e1 f3e1');
INSERT INTO evaluations VALUES(6644,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(6644,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(6644,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(6644,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(6644,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(6644,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(6644,36,'cp 13','d7d5 e4d5 d8d5 f1e2 g8f6 c3c4 d5d6 e1g1 e8g8 a2a4 f8e8 c1a3 d6d8 f1e1 c8f5 f3h4 f5d7 a3b2 c6d4 h4f3 c7c5 f3d4 e5d4 c2c3 d8b6 d1d2 a7a5 h2h3 h7h6 c3d4 c5d4 e1b1 b6d6 b2a3 d6e5 e2f3 d7c6 f3c6 b7c6 b1e1 e5f5 e1e8 a8e8');
INSERT INTO evaluations VALUES(8245,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8245,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8245,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8245,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8245,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8245,36,'cp 0','c1e3 e8g8 e1g1 b7b6 d1d2 c8b7 e3h6 e7e6 f1e1 a7a6 a2a3 d8c7 c3e2 a8d8 b2b4 c6e5 f3e5 d6e5 h6g7 g8g7 d2c3 b6b5 c4b3 d8c8 f2f3 c7b6 g1h1 b6c7');
INSERT INTO evaluations VALUES(11340,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(11340,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(11340,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(11340,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11340,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(11340,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11340,36,'cp 30','d3d4 c6d4 f3d4 e5d4 d1d4 f8e7 a2a4 e8g8 e1g1 c8e6 f1d1 e6c4 d4c4 c7c6 a4a5 f8e8 c1f4 e7f8 c4d3 d8c8 f4e3 c8c7 e3b6 c7e7 f2f3 a8c8 d1d2 f6d7 b6f2 d7c5 f2c5 d6c5 d3d7 e7e5 d7f5 e5f5 e4f5 c5c4 a1d1');
INSERT INTO evaluations VALUES(12575,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(12575,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(12575,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(12575,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(12575,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(12575,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(12575,36,'cp -17','d8a5 g5d2 a5b6 b1c3 f8c5 d1e2 d7d6 c3a4 b6c7 a4c5 d6c5 d2g5 c8e6 g5f6 g7f6 c2c3 e8c8 e2e3 c7b6 b2b3 c5c4 e3b6 a7b6 f1c4 e6c4 b3c4 h8g8 g2g3 d8d3 f3h4 g8g4 f2f3 g4g8 a1c1 g8d8 e1g1 c6a5 f3f4 d8d7 f4e5 f6e5 h4f3 a5c4 f3g5 d3d2 f1f7 d7f7 g5f7');
INSERT INTO evaluations VALUES(15388,30,'cp -397',NULL);
INSERT INTO evaluations VALUES(15388,31,'cp -420',NULL);
INSERT INTO evaluations VALUES(15388,32,'cp -404',NULL);
INSERT INTO evaluations VALUES(15388,33,'cp -407',NULL);
INSERT INTO evaluations VALUES(15388,34,'cp -402',NULL);
INSERT INTO evaluations VALUES(15388,35,'cp -411',NULL);
INSERT INTO evaluations VALUES(15388,36,'cp -431','f8c5 d2d4 c6d4 c1e3 d4b3 a2b3 f7f6 e3c5 e4c5 f3e5 f6e5 d1h5 g7g6 h5e5 e8f7 e5f4 f7g8 e1e5 c5e6 f4g3 g8g7 b1d2 d8f6 a1e1 h8f8 e5d5 g7g8 d2e4 f6g7 e4c5 c7c6 c5e6 c8e6 d5e5 e6f7 h2h4 a8e8 e5e8 f7e8 h4h5 g7b2 g3e3 e8f7 c2c3 g6h5 b3b4 f8e8 e3g3 g8f8 e1e8 f7e8 g3e3 e8f7');
INSERT INTO evaluations VALUES(9171,30,'cp -433',NULL);
INSERT INTO evaluations VALUES(9171,31,'cp -416',NULL);
INSERT INTO evaluations VALUES(9171,32,'cp -400',NULL);
INSERT INTO evaluations VALUES(9171,33,'cp -410',NULL);
INSERT INTO evaluations VALUES(9171,34,'cp -418',NULL);
INSERT INTO evaluations VALUES(9171,35,'cp -417',NULL);
INSERT INTO evaluations VALUES(9171,36,'cp -425','d2d4 c6d4 c1e3 d4b3 a2b3 f7f6 e3c5 e4c5 f3e5 f6e5 d1h5 g7g6 h5e5 e8f7 e5f4 f7g8 e1e5 c5e6 f4g3 g8g7 b1d2 d8f6 a1e1 h8f8 e5d5 g7g8 d5e5 c8d7 d2e4 f6g7 g3c3 a8e8 f2f3 e6f4 e5e8 d7e8 c3c4 g8h8 e4f2 e8c6 f2g4 f4d5 c4c5 h7h5 c2c4 h5g4 c4d5 g4f3 d5c6 f3f2 c5f2 f8f2 g1f2 g7b2 e1e2 b2f6 f2e3 f6c6 e3d3 c6d6 d3c4 d6h2');
INSERT INTO evaluations VALUES(9170,30,'cp -440',NULL);
INSERT INTO evaluations VALUES(9170,31,'cp -421',NULL);
INSERT INTO evaluations VALUES(9170,32,'cp -451',NULL);
INSERT INTO evaluations VALUES(9170,33,'cp -449',NULL);
INSERT INTO evaluations VALUES(9170,34,'cp -428',NULL);
INSERT INTO evaluations VALUES(9170,35,'cp -449',NULL);
INSERT INTO evaluations VALUES(9170,36,'cp -432','c6d4 c1e3 d4b3 a2b3 f7f6 d1d3 e8g8 c2c4 c8e6 b1d2 c5b4 c4d5 e4d2 e3d2 b4c5 d2e3 c5e3 e1e3 d8d5 d3c3 d5d7 h2h3 f8d8 b3b4 c7c6 a1e1 d7d6 e3e2 d6f8 f3h4 d8d4 c3d4 e5d4 e2e6 f8b4 h4f5 d4d3');
INSERT INTO evaluations VALUES(16169,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(16169,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(16169,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(16169,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(16169,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(16169,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(16169,36,'cp 12','e1g1 d7d5 e4d5 d8d5 a2a3 c8f5 b2b4 e8c8 c1b2 e5e4 d3e4 d5d1 f1d1 d8d1 e2d1 f6e4 c2c4 h8g8 d1c2 g7g6 f3h4 e4d6 h4f5 d6f5 g1f1 c6d4 c2e4 f8g7 b1c3 g8e8 a1d1 c7c6 c4c5 e8d8 e4d3 d4e6 d3f5 g6f5 d1d8 c8d8 c3a4 g7d4');
INSERT INTO evaluations VALUES(16168,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(16168,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(16168,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(16168,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16168,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16168,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16168,36,'cp 8','d7d5 e4d5 d8d5 b1c3 d5d8 c1e3 c8e6 d3d4 e5d4 f3d4 c6d4 e3d4 f8e7 d1d3 e8g8 a1d1 c7c5 d4e5 d8d3 e2d3 a8d8 a2a4 c5c4 d3e2 e6f5 d1d8 f8d8 f1d1 d8d1 e2d1 e7c5 g1f1 f6e4 c3e4 f5e4 e5c3 e4d5 d1g4');
INSERT INTO evaluations VALUES(18223,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(18223,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(18223,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(18223,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(18223,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(18223,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(18223,36,'cp 48','f1c4 e7e6 d1e2 g8f6 c1g5 f8e7 e1c1 h7h6 g5h4 d8c7 h1e1 e8g8 e4e5 d6e5 e2e5 c7e5 e1e5 g7g5 h4g3 f6h5 c3a4 h5g3 h2g3 a8b8 c4e2 f8d8 d1d8 e7d8 e5c5 c8d7 e2f3 b8c8 c5c3 a6a5 a4c5 d7e8 c3b3 d8e7 c5d3 e7d6');
INSERT INTO evaluations VALUES(18219,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(18219,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(18219,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(18219,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(18219,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(18219,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(18219,36,'cp 40','e7e6 d1e2 g8f6 c1g5 d8c7 e1c1 f8e7 h1e1 c8b7 e4e5 d6e5 f2f4 h7h6 g5f6 e7f6 c3e4 f6e7 e2h5 e8g8 h5e5 c7e5 f4e5 c6c5 g2g3 b7e4 e1e4 a8d8 d1f1 a6a5 c2c3 g7g6 h2h4 g8g7 c1c2 h6h5 a2a3 d8b8 c4a6 f8d8 e4a4 g6g5 a4a5 g5h4');
INSERT INTO evaluations VALUES(18243,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(18243,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(18243,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(18243,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(18243,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(18243,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(18243,36,'cp 46','d1e2 g8f6 c1g5 f8e7 e1c1 d8c7 h1e1 c8b7 g2g3 e8g8 e4e5 d6e5 e2e5 c7e5 e1e5 c6c5 c4e2 a8d8 d1d8 f8d8 c3d1 d8a8 g5f6 g7f6 e5e3 f6f5 e3b3 b7c6 d1e3 c6b5 e2c4 e7d6 c4b5 a6b5 b3b5 a8a2 e3c4 a2a1 c1d2 d6e7 h2h4 a1h1 b5b7 g8f8 c4e5 h7h5 d2e2 h1a1 b7c7 f8e8 f2f4');
INSERT INTO evaluations VALUES(19772,30,'cp 247',NULL);
INSERT INTO evaluations VALUES(19772,31,'cp 266',NULL);
INSERT INTO evaluations VALUES(19772,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(19772,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(19772,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(19772,35,'cp 259',NULL);
INSERT INTO evaluations VALUES(19772,36,'cp 260','c4d5 c6d4 f3d4 c5d4 d1d4 e7e6 e2e4 g8f6 b1c3 d8b6 d4b6 a7b6 d5e6 f7e6 f2f3 c8d7 a2a3 f8c5 c1f4 e8e7 e1c1 h7h6 f4e5 d7c6 f1b5 c6b5 c3b5 a8a5 b5d4 e7f7 d4c2 c5e7 e5g3 f6h5 c1b1 h5g3 h2g3');
INSERT INTO evaluations VALUES(20118,30,'cp 283',NULL);
INSERT INTO evaluations VALUES(20118,31,'cp 260',NULL);
INSERT INTO evaluations VALUES(20118,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(20118,33,'cp 271',NULL);
INSERT INTO evaluations VALUES(20118,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(20118,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(20118,36,'cp 281','f2f4 e5g6 g1f3 e7e5 d5e6 f7e6 b1c3 g8f6 d1b3 c8d7 c1e3 d8a5 b3b7 a8d8 e3d2 a5b6 b7b6 a7b6 f1d3 f8e7 g2g3 e8g8 e1e2 d6d5 e4e5 f6e4 c3e4 d5e4 d3e4 d7b5 e2e1 b5a6 d2e3 e7c5 e3c5 b6c5 a1d1 g6e7 d1d8');
INSERT INTO evaluations VALUES(20119,30,'cp 285',NULL);
INSERT INTO evaluations VALUES(20119,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(20119,32,'cp 291',NULL);
INSERT INTO evaluations VALUES(20119,33,'cp 273',NULL);
INSERT INTO evaluations VALUES(20119,34,'cp 261',NULL);
INSERT INTO evaluations VALUES(20119,35,'cp 276',NULL);
INSERT INTO evaluations VALUES(20119,36,'cp 277','e5d7 b1c3 a7a6 g1f3 b7b5 f3d4 g8f6 c1e3 d7c5 f1e2 d8d7 d4c6 e7e6 a1c1 e6d5 e4d5 f8e7 e2f3 b5b4 e3c5 b4c3 c1c3 d6c5 c3e3 e8g8 e3e7 d7d6 e7e5 a6a5 e1g1 a5a4 h2h3 g7g6 f1e1 c8d7 d1c2 g8g7 e1c1 a4a3 b2b3 f8e8 g1h1 g7g8 e5e8 a8e8 c2c5');
INSERT INTO evaluations VALUES(20810,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(20810,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(20810,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(20810,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(20810,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(20810,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(20810,36,'cp 19','g7g6 b1c3 f8g7 a2a4 g8f6 b2b4 a7a6 e1g1 h7h6 a4a5 e8g8 f1e1 f6h7 a1a3 g8h8 f3d2 f7f5 c4f1 h7f6 f2f3 c8d7 c3a4 f5e4 f3e4 b7b5 a5b6 c7b6 d2f3 b6b5 a4b2 e7d5 e4d5 e5e4 f3d4 f6d5 e1e4 d8b6 c2c3');
INSERT INTO evaluations VALUES(21522,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(21522,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(21522,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(21522,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(21522,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(21522,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(21522,36,'cp 21','c2c4 d5c4 g1e2 c8d7 e1g1 g8f6 b5c4 d8e7 b1c3 e8c8 d1c2 c8b8 a1c1 c6b4 c2b1 d7c8 e2g3 d8d2 f1d1 d2d1 c1d1 b4d3 c4d3 e4d3 b1d3 h7h5 g3f5 e7c5 f5d4 f8d6 d4b5 h8d8 c3a4 c5f5 d3f5 c8f5 b5d6');
INSERT INTO evaluations VALUES(36716,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(36716,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(36716,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(36716,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(36716,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(36716,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(36716,36,'cp 83','g5f6 g7f6 c4d5 e7e6 d5e6 c8e6 e2e3 c7c5 c3e4 b8d7 e4f6 d7f6 f1e2 d8a5 d1d2 a5d2 e1d2 e6d5 d4c5 f6e4 d2e1 e4c5 h2h4 c5e6 h4h5 f8c8 h1h4 c8c2 h4b4 a7a5 b4b6 g6h5 a2a3 a8c8 e2d3 c2c6 b6b5 c6c5 b5c5 e6c5 d3f5 c5e6 e1e2 c8c5 f5d3 d5c6');
INSERT INTO evaluations VALUES(42680,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(42680,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(42680,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(42680,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(42680,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(42680,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(42680,36,'cp 63','c5b4 c1d2 b4d2 b1d2 e8g8 d1c2 b8c6 a2a3 d7d5 e4e5 f6h5 g2g3 g7g6 f1g2 h5g7 e1g1 a7a5 c2c3 c6e7 h2h3 c7c6 f1e1 a5a4 b2b4 a4b3 a3a4 b3b2 c3b2 g7e6 d2b3 b7b6 h3h4 e6g7 a4a5 b6a5 b2c3 c8f5 b3a5 d8c7 g2f1');
INSERT INTO evaluations VALUES(42817,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(42817,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(42817,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(42817,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(42817,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(42817,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(42817,36,'cp 60','c1d2 b4d2 b1d2 e8g8 d1c2 b8c6 a2a3 d7d5 e4e5 f6h5 g2g3 a7a5 f1g2 g7g6 e1g1 h5g7 c2c3 c6e7 f1e1 c7c6 h2h4 c8g4 f3h2 g4f5 d2b3 b7b6 b3c1 c6c5 c1e2 h7h5 c3d2 g7e6 a1d1 a8c8 h2f1 f5g4 f2f3 g4f5 f1e3 d8d7 f3f4 c5d4 e2d4 e6d4 d2d4 f5e6 d4b6');
INSERT INTO evaluations VALUES(46249,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(46249,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46249,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46249,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(46249,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(46249,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(46249,36,'cp 0','e1g1 b8c6 c2c4 d5c4 d4c5 d8d1 f1d1 f8c5 e2c4 c8d7 b1d2 e8g8 a2a3 c5e7 c4e2 f6d5 e3e4 d5f6 b2b4 f8d8 c1b2 a8c8 a1c1 a7a6 g1f1 h7h6 g2g3 c6a7 b2d4 a7b5 d4b2');
INSERT INTO evaluations VALUES(46248,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(46248,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46248,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(46248,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46248,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46248,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(46248,36,'cp 0','b8c6 c2c4 d5c4 d4c5 d8d1 f1d1 f8c5 e2c4 c8d7 a2a3 c6a5 c4a2 d7a4 d1d3 a4b5 d3d1');
INSERT INTO evaluations VALUES(13594,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(13594,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13594,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13594,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(13594,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13594,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(13594,36,'cp 6','f1e1 b7b6 c2c4 f8d6 c4d5 e6d5 d4c5 b6c5 b1c3 e8g8 b2b3 c8b7 c1b2 f8e8 c3b5 d6e7 e2f1 h7h6 b5c3 e7f8 a1c1 a8c8 c3e2 f6e4 e2f4 d5d4 e3d4 c5d4 d1d3 c6b4 d3d4 b4a2 c1c8 d8d4 b2d4 e8c8 d4a7 f8b4 e1d1');
INSERT INTO evaluations VALUES(47481,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(47481,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(47481,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(47481,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(47481,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(47481,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(47481,36,'cp 81','c3e4 d5e4 e2e3 c7c6 d1d2 c8f5 g1e2 b8d7 e2g3 f5g6 h2h4 f7f6 g5f4 h7h6 h4h5 g6h7 g3e2 e7e5 d4e5 f6e5 f4g3 d8b6 e1c1 e8c8 a2a3 f8e7 d2c3 b6c7 c1b1 h7f5 e2c1 e7d6 c1b3 d7f6 f1e2 f5e6');
INSERT INTO evaluations VALUES(47480,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(47480,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(47480,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(47480,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(47480,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(47480,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(47480,36,'cp 64','d5e4 e2e3 c8f5 g1e2 c7c6 d1d2 b8d7 e2g3 f5g6 e1c1 h7h6 g5h4 d7f6 f1e2 d8c7 c2c4 e8c8 d2c2 e7e5 h4f6 g7f6 e2g4 c8b8 d4e5 h6h5 d1d8 c7d8 h1d1 d8a5 g4h3 h5h4 g3f5 g6h5 d1d2 f8b4 d2d7 a5a2 e5f6 h5g6 c2e4 b8a8');
INSERT INTO evaluations VALUES(47521,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(47521,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(47521,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(47521,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(47521,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(47521,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(47521,36,'cp 82','e2e3 c8f5 g1e2 c7c6 d1d2 b8d7 g5h4 d7f6 h2h3 h7h6 e1c1 f5h7 g2g4 e7e6 f1g2 f8e7 h4f6 g7f6 e2c3 f6f5 g4f5 e6f5 d4d5 d8d6 g2f1 e8c8 f1c4 e7f6 d5c6 d6c6 c4d5 c6b6 d2e2 f6c3 e2c4 b6c7 b2c3 c7c4');
INSERT INTO evaluations VALUES(47520,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(47520,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(47520,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(47520,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(47520,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(47520,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(47520,36,'cp 76','c8f5 g2g4 f5e6 g5f4 b8c6 h2h3 h7h5 g4h5 d8d7 a2a3 f7f6 g1e2 e6f7 e2g3 e8c8 h3h4 c8b8 d1e2 e7e5 d4e5 c6e5 f4e5 f6e5 f1h3 d7a4 e1g1 f8e7 f1d1 f7d5 h3f5 a4e8 f5e4 d5e4 d1d8 e7d8 g3e4 d8h4');
INSERT INTO evaluations VALUES(48021,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(48021,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48021,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48021,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48021,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48021,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48021,36,'cp 0','f8e7 e1g1 e8g8 c2c4 b8c6 d2d3 e4f6 b1c3 d6d5 d3d4 d5c4 e2c4 c8g4 c1e3 g4h5 a1c1 a7a6 h2h3 d8d7 f1e1 f8e8 c4e2 e7d6 e3g5 f6e4 g5e3');
INSERT INTO evaluations VALUES(40871,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40871,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(40871,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(40871,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(40871,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40871,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40871,36,'cp 7','e1g1 e8g8 f1e1 d6d5 d2d3 e4f6 d3d4 h7h6 b1c3 f8e8 c1f4 c8f5 e2d3 f5d3 d1d3 c7c6 c3e2 e7f8 e2g3 e8e1 a1e1 b8d7 b2b3 d8a5 a2a4 a8e8 e1e8 f6e8 f4d2 a5d8 c2c4 d5c4 b3c4 e8c7 h2h3 c7e6 d2c3 d8c7 a4a5 e6f4');
INSERT INTO evaluations VALUES(40870,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40870,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40870,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40870,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(40870,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40870,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40870,36,'cp 5','e8g8 f1e1 d6d5 d2d3 e4f6 d3d4 h7h6 b1c3 f8e8 c1f4 c8f5 e2d3 f5d3 d1d3 c7c6 c3e2 e7f8 c2c3 d8b6 d3c2 b8d7 e2g3 e8e1 a1e1 a8e8 h2h3 b6d8 g3f5 e8e1 f3e1 d8a5 a2a3 d7b6 e1d3 f6d7 f4c7 a5a4 c2a4 b6a4');
INSERT INTO evaluations VALUES(36278,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36278,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36278,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36278,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36278,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36278,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36278,36,'cp 0','f1e1 d6d5 d2d3 e4f6 d3d4 h7h6 c2c4 c7c5 b1c3 c8e6 c4d5 f6d5 c3d5 e6d5 c1e3 c5d4 f3d4 e7f6 d4f5 d5e6 f5d4');
INSERT INTO evaluations VALUES(49515,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(49515,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(49515,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(49515,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(49515,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(49515,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(49515,36,'cp -24','e5d4 d1d4 b8c6 d4e3 h7h6 g5f6 d8f6 b1c3 f8b4 g1e2 d7d6 f2f3 f6h4 g2g3 h4g5 f3f4 g5a5 e1c1 c8g4 h2h3 g4e2 f1e2 e8c8 e2c4 b4c3 e3c3 a5c3 b2c3 h8f8 c4d3 d8e8 d1e1 e8e7 c1d2 f7f6 a2a4 f8e8 h3h4 c6a5 g3g4 c7c6');
INSERT INTO evaluations VALUES(54808,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(54808,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(54808,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(54808,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(54808,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(54808,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(54808,36,'cp -18','g8f6 d2d4 f8g7 f3e2 e8g8 g1f3 c8g4 c2c3 b8d7 h2h3 g4f3 e2f3 e7e5 c1e3 e5d4 c3d4 d6d5 c4d5 c7c5 e1g1 f6d5 e4d5 c5d4 e3f4 f8e8 b1d2 d7e5 f3b3 d8d7 f1e1 a8d8 a1d1 b7b6 d2c4 d7d5 c4e5 g7e5 b3d5 d8d5 e1e5 d5e5');
INSERT INTO evaluations VALUES(57175,30,'cp -166',NULL);
INSERT INTO evaluations VALUES(57175,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(57175,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(57175,33,'cp -175',NULL);
INSERT INTO evaluations VALUES(57175,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(57175,35,'cp -159',NULL);
INSERT INTO evaluations VALUES(57175,36,'cp -167','d7d5 d2d4 b8c6 b1c3 c8e6 d4e5 c6e5 h2h3 h7h5 g1f3 e5f3 d1f3 h5g4 h3g4 e6g4 f3g2 h8h1 g2h1 g8f6 c1d2 d8d6 h1g2 e8c8 f2f3 g4e6 c3b5 d6e5 e1c1 a7a6 b5d4 e6d7 d4b3 d8e8 d1e1 e5d6 a2a3 f6h5 d2b4 d6g3 g2g3 h5g3 f1d3 f8b4 a3b4');
INSERT INTO evaluations VALUES(2971,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(2971,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(2971,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(2971,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(2971,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(2971,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(2971,36,'cp 40','b2b3 b7b6 c1b2 c8b7 d1e2 d8c7 a1c1 d5c4 d3c4 c6c5 d4c5 c7c5 c3b5 c5h5 f1d1 f8d8 h2h3 h7h6 b5d6 b7d5 c4d5 f6d5 d6e4 h5g6 e4g3 e7f6 b2f6 d7f6 f3d4 d5e7 e3e4 g6g5 d4f3 d8d1 c1d1 g5f4');
INSERT INTO evaluations VALUES(2968,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(2968,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(2968,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(2968,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(2968,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(2968,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(2968,36,'cp 24','b7b6 d1c2 h7h6 c1b2 c8b7 c2e2 e7d6 a1d1 f8e8 e3e4 d5e4 c3e4 f6e4 d3e4 d8c7 e2c2 d7f6 e4d3 c7e7 h2h3 d6a3 b2a3 e7a3 c4c5 a8d8 f1e1 b7a6 d3a6 a3a6 f3e5 b6c5 d4c5 d8d5 d1d5 e6d5 e5f3 e8e1 f3e1 a6b7 f2f3 a7a6');
INSERT INTO evaluations VALUES(3792,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(3792,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3792,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(3792,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3792,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(3792,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(3792,36,'cp 0','h7h6 g5h4 a7a5 f1e1 g7g5 h4g3 g5g4 g3h4 g4f3 d1f3 g8g7 f3g3 g7h7 g3f3');
INSERT INTO evaluations VALUES(3584,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3584,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3584,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(3584,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(3584,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(3584,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(3584,36,'cp 12','g5h4 a7a5 b1d2 c5a7 c4b3 g7g5 h4g3 f6h7 g1h1 g8h8 d2c4 f7f5 e4f5 c8f5 c4e3 f5g6 b3c2 d8d7 d3d4 g6c2 e3c2 e5d4 c2d4 a8e8 d4c6 d7c6 f3d4 c6d5 f2f3 h7f6 g3f2 d5f7 d1c2 f6d5 a1e1 a7b6 f2g1 e8e5 e1e5 d6e5 d4b5 b6g1 f1g1 f7f4 c2d3 f4f7 d3e2 f8e8 g1e1');
INSERT INTO evaluations VALUES(3596,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(3596,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(3596,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(3596,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(3596,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3596,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3596,36,'cp 6','g7g5 h4g3 a7a6 b1d2 c5a7 a2a4 g8g7 f1e1 g5g4 f3h4 f6h5 d2f1 d8g5 f1e3 a7e3 e1e3 h5f4 d1d2 c8e6 c4a2 e6a2 h4f5 g7h7 a1a2 d6d5 b2b4 d5e4 d3e4 a8d8 d2c2 f7f6 g3h4 g5g6 f2f3 f8f7 f3g4 g6g4 e3f3 c6e7 h2h3 g4h5 f5e7 f7e7 h4f6');
INSERT INTO evaluations VALUES(9120,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(9120,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(9120,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(9120,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(9120,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(9120,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(9120,36,'cp 47','e4d5 c6a5 c4b5 c8d7 b5d3 c7c6 g5e4 e8g8 e4f6 d8f6 d5c6 a5c6 a2a3 b4e7 c3d5 f6g5 d5e7 c6e7 e1g1 d7c6 f2f3 a8d8 d3e4 c6e4 d2d3 g5g6 f3e4 f7f5 d1e2 f5f4 b2b4 e7c6 c1b2 a7a6 a3a4 h7h6 g1h1 g6d6 b4b5 c6d4 e2f2 d6e6 b2d4 d8d4 f2e2 e6d6 g2g3 g8h7');
INSERT INTO evaluations VALUES(9102,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(9102,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(9102,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(9102,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(9102,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(9102,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(9102,36,'cp 40','c6a5 c4b5 c8d7 b5d3 c7c6 g5e4 e8g8 e4f6 d8f6 d5c6 a5c6 a2a3 b4d6 c3e4 f6h6 e4d6 h6d6 e1g1 a8c8 f1e1 f8d8 b2b3 d6d4 a1b1 d7g4 d3e2 g4e2 e1e2 d4d6 d2d3 c6d4 e2d2 f7f6 c1b2 d4e6 d1g4 d6c6 b1e1 b7b5 h2h4 e6f4 h4h5 h7h6 e1e4 c6d7 g4d7 d8d7');
INSERT INTO evaluations VALUES(9981,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(9981,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(9981,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(9981,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(9981,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(9981,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(9981,36,'cp 63','e2e3 b4c3 c2c3 e8g8 b2b3 f8e8 c1b2 a7a5 g1f3 d7d6 a1d1 d8e7 d4d5 c6b8 d5e6 f7e6 f1d3 b8d7 e1g1 d7c5 d3c2 e6e5 f3d2 c8g4 f2f3 g4h5 e3e4 h5g6 f1f2 c5e6 d2f1 f6h5 f1e3 h5f4 c3d2 e7g5 g1h1 h7h5 b2c3 h5h4 h2h3');
INSERT INTO evaluations VALUES(11428,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(11428,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(11428,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(11428,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(11428,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(11428,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(11428,36,'cp -16','b7b5 a4b3 c6a5 e1g1 f8e7 b1c3 e8g8 f3h2 d8d7 f2f4 c8b7 f4e5 d6e5 c1e3 a8d8 h2g4 a5b3 a2b3 f6g4 d1g4 d7g4 h3g4 h7h6 c3d5 b7d5 e4d5 d8d5 a1a6 e7c5 e3c5 d5c5 c2c4 b5c4 b3c4 f8b8 b2b4 b8b4 a6a8 g8h7 f1f7 e5e4 d3e4 c5g5 f7f4 b4c4 g1f2 g5e5 a8a7 c4c2 f2f3 c2c3 f3f2');
INSERT INTO evaluations VALUES(11651,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(11651,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(11651,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(11651,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(11651,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(11651,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(11651,36,'cp -12','a4b3 c6a5 e1g1 f8e7 b1c3 e8g8 f3h2 d8d7 f2f4 c8b7 f4e5 d6e5 c1e3 d7d6 d1f3 a8d8 c3e2 a5b3 a2b3 f6e8 f3g3 f7f6 h2g4 g8h8 e2c3 c7c5 g1h1 e8c7 e3g1 d6d7 g4e3 f8f7 g3f3 e7f8 f3g3 f8d6 g3f3 c7e6 f3g4 e6d4 g4d7 d8d7');
INSERT INTO evaluations VALUES(11630,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(11630,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(11630,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(11630,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(11630,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(11630,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(11630,36,'cp -8','c6a5 e1g1 f8e7 b1c3 e8g8 f3h2 d8d7 f2f4 a5b3 a2b3 c8b7 f4e5 d6e5 c1e3 a8d8 d1f3 d7d6 c3e2 f6e8 f3g3 f7f6 h2g4 g8h8 e2c3 c7c5 g1h1 e8c7 e3g1 d6d7 g4e3 f8f7 g3g4 d7g4 h3g4 e7d6 c3d5 c7e6 e3f5 d6f8 g1e3 h8g8 e3g1 b7d5');
INSERT INTO evaluations VALUES(11670,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(11670,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(11670,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(11670,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(11670,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(11670,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11670,36,'cp 0','e1g1 f8e7 b1c3 e8g8 f3h2 a5b3 a2b3 d8d7 f2f4 c8b7 f4e5 d6e5 c1e3 d7e6 d1f3 f6d7 g1h1 e6d6 f3g4 d7f6 g4f3');
INSERT INTO evaluations VALUES(14055,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14055,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14055,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14055,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14055,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14055,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14055,36,'cp 0','g1f3 e5e4 f3e5 a7a6 d1h5 c8e6 e5c6 e7c6 b5c6 b7c6 e1g1 f8d6 d2d3 e4d3 c2d3 e8g8 f1c1 e6d7 b2d4 f8e8 d4c5 e8e6 c5d6 c7d6 b1c3 e6h6 h5f3 h6f6 f3d1 f6h6 d1f3');
INSERT INTO evaluations VALUES(14635,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(14635,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14635,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14635,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(14635,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(14635,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(14635,36,'cp 37','d2d4 e5d4 f3d4 f8e7 e1g1 e8g8 c1f4 f8e8 f1e1 c8g4 d4c6 b7c6 d1d3 f6h5 f4e3 e7g5 h2h3 g4e6 c4a6 e6d7 a1d1 g5f6 d3d2 f6e5 a6f1 a7a5 g2g3 e5c3 d2c3 e8e4 f1g2 e4e8 g2c6 d7c6 c3c6 h5f6 e3d4 d8d7 c6d7 f6d7 e1e8 a8e8');
INSERT INTO evaluations VALUES(15564,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(15564,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15564,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(15564,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15564,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(15564,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15564,36,'cp 0','d6d5 b1d2 d5d4 c2c3 c8g4 a2a3 d4c3 b2c3 a7a5 f1e2 f8d6 b3b4 a5b4 a3b4 g4f3 d2f3 d6b4 c3b4 c6b4 a1a8 d8a8 e1g1 b4c6 d3d4 e5d4 e4e5 f6e4 e5e6 f7e6 f3d4 e8g8 d4e6 e4c3 d1d3 c3e2 d3e2 a8e8 f1e1 f8f6 e2g4 e8g6 g4e2');
INSERT INTO evaluations VALUES(16663,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16663,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16663,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(16663,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(16663,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16663,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(16663,36,'cp 0','f8b4 d1e2 d8e7 g1f3 d7d5 a2a3 b4c3 b2c3 c6a5 c4d5 f6d5 e2e7 e8e7 f1d3 a5b3 c1g5 f7f6 a1b1 b3d4 c3d4 f6g5 f3g5 h7h6 g5f3 d5f4 d3e4 f4g2 e1d2 g2f4 d2e3 f4g2');
INSERT INTO evaluations VALUES(10512,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10512,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10512,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10512,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10512,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10512,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10512,36,'cp 0','d1e2 d8e7 g1f3 d7d5 a2a3 b4c3 b2c3 c6a5 c4d5 f6d5 e2e7 e8e7 f1d3 a5b3 c1g5 f7f6 a1b1 b3d4 c3d4 f6g5 f3g5 h7h6 g5f3 d5f4 d3e4 f4g2 e1d2 g2f4 d2e3 f4g2');
INSERT INTO evaluations VALUES(21085,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(21085,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(21085,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(21085,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(21085,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(21085,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(21085,36,'cp -27','c6e5 d4e5 c8f5 g2g3 c7c6 f1g2 g7g6 c1f4 f8h6 c2c4 h6f4 g3f4 e7e6 c4d5 e6d5 b1d2 g8h6 h2h4 h6g4 d1b3 d8b6 b3b6 a7b6 e2e4 d5e4 g2e4 g4h6 h4h5 e8c8 f2f3 d8d4 h5g6 h7g6 e4f5 g6f5');
INSERT INTO evaluations VALUES(21223,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(21223,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(21223,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(21223,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(21223,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(21223,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(21223,36,'cp -18','d4e5 c8f5 g2g3 c7c6 f1g2 g7g6 c2c4 d5c4 d1a4 f5e6 b1a3 f8g7 e1g1 g8h6 a3c4 e8g8 f1d1 d8c7 c1h6 g7h6 a4a5 b7b6 a5c3 e6c4 c3c4 c7e5 g2c6 a8d8 c4b5 e5f6 c6g2 h6g7 a1b1 e7e6 e2e3 d8d1 b1d1 f6b2 b5b2');
INSERT INTO evaluations VALUES(21222,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(21222,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(21222,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(21222,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(21222,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(21222,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(21222,36,'cp -21','c8f5 g2g3 c7c6 f1g2 g7g6 c1f4 f8h6 d1c1 d8a5 c2c3 h6f4 c1f4 f7f6 b1d2 a5c7 e2e4 d5e4 g2e4 c7e5 f4e5 f6e5 e1c1 f5e6 h1e1 g8f6 e4c2 f6d7 d2e4 e8c8 e4g5 e6d5 f2f4 e5f4 g3f4 d8f8 e1e7 f8f4 g5h7 f4f2');
INSERT INTO evaluations VALUES(21554,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(21554,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(21554,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(21554,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(21554,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(21554,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(21554,36,'cp 146','g8e7 d5c4 e7g6 e1g1 f8e7 b1c3 e8g8 c3d5 g8h8 d5e7 d8e7 d2d3 c6a5 c4d5 c7c6 d5b3 a5b3 a2b3 b7b6 c1e3 c6c5 f3d2 c8e6 d2c4 a8d8 f2f3 h7h6 d1d2 d8d7 e3f2 f7f6 c4e3 g6f4 g2g3 f4h3 g1g2 h3f2 g2f2 a7a5');
INSERT INTO evaluations VALUES(23076,30,'cp -364',NULL);
INSERT INTO evaluations VALUES(23076,31,'cp -377',NULL);
INSERT INTO evaluations VALUES(23076,32,'cp -366',NULL);
INSERT INTO evaluations VALUES(23076,33,'cp -370',NULL);
INSERT INTO evaluations VALUES(23076,34,'cp -378',NULL);
INSERT INTO evaluations VALUES(23076,35,'cp -383',NULL);
INSERT INTO evaluations VALUES(23076,36,'cp -363','e5g6 b1c3 f8b4 f1c4 g8f6 e4e5 d7d5 c4b5 c8d7 b5d7 f6d7 e1g1 b4c3 b2c3 e8g8 f2f4 f7f5 c1a3 g6e7 a1b1 d7b6 d1e2 g8h8 a3c5 f8e8 e2h5 g7g6 h5g5 e7c6 g2g4 d8g5 f4g5 f5g4 f1f7 b6a4 f7c7 a4c5 d4c5 c6e5 b1b7 e5f3 g1g2 f3g5');
INSERT INTO evaluations VALUES(22998,30,'cp -350',NULL);
INSERT INTO evaluations VALUES(22998,31,'cp -367',NULL);
INSERT INTO evaluations VALUES(22998,32,'cp -377',NULL);
INSERT INTO evaluations VALUES(22998,33,'cp -373',NULL);
INSERT INTO evaluations VALUES(22998,34,'cp -371',NULL);
INSERT INTO evaluations VALUES(22998,35,'cp -390',NULL);
INSERT INTO evaluations VALUES(22998,36,'cp -366','b1c3 f8b4 h2h4 h7h5 f1c4 g8e7 e1g1 d7d5 e4d5 g6h4 c1g5 h4f5 c3e4 f7f6 g5f4 e8f8 c2c3 b4d6 f4d2 h5h4 f1e1 f5h6 e4d6 c7d6 d2h6 h8h6 d1f3 g7g5 c4b3 c8d7 c3c4 e7g6 c4c5 g6f4 c5d6');
INSERT INTO evaluations VALUES(25434,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(25434,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(25434,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(25434,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25434,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(25434,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(25434,36,'cp 27','d2d4 e5d4 f3d4 d8d7 d4c6 d7c6 d1d3 f8e7 c1g5 h7h6 d3h3 e8f7 g5f4 e6e5 f4d2 h8f8 h3d3 f7g8 b2b3 a7a5 a1d1 c6a6 c3b5 a6c6 d2e3 a5a4 c2c4 a4b3 a2b3 a8a2 b5c3 a2b2 d1b1 b2b1 f1b1 c6d7 c3d5');
INSERT INTO evaluations VALUES(25432,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(25432,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(25432,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(25432,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(25432,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(25432,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(25432,36,'cp 19','e5d4 f3d4 d8d7 d1d3 f8e7 d4c6 d7c6 c3b5 e8g8 b5d4 c6d7 f1e1 d6d5 d4f3 e7c5 e4d5 e6d5 c1e3 c5e3 e1e3 a8e8 a1e1 a7a6 h2h3 e8e3 e1e3 f8e8 e3e8 d7e8 d3f5 c7c6 a2a4 h7h6 f3e5 f6e4 f5f4 e8e7');
INSERT INTO evaluations VALUES(29735,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(29735,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(29735,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(29735,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(29735,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(29735,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(29735,36,'cp 211','e7e5 g1e2 b8d7 g2g4 f6g4 f3g4 d8h4 e1d2 h5g4 d1e1 h4e1 d2e1 e5d4 e2d4 d7e5 c4e2 f8c5 c1e3 e8c8 a1d1 g4e2 e1e2 e5c4 c3a4 c5b4 b2b3 c4e3 e2e3 h8e8 h2h4 a7a6 h4h5 b7b5 d4c6 b5a4 c6d8 e8d8');
INSERT INTO evaluations VALUES(32208,30,'cp 244',NULL);
INSERT INTO evaluations VALUES(32208,31,'cp 259',NULL);
INSERT INTO evaluations VALUES(32208,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(32208,33,'cp 245',NULL);
INSERT INTO evaluations VALUES(32208,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(32208,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(32208,36,'cp 256','c4e6 g7g6 e6f7 e8f7 h5f3 f7g7 g1e2 b8c6 f3b3 b7b6 d2d3 g8f6 e1g1 d8d7 b1c3 c6a5 b3a4 d7a4 c3a4 a5c6 a4c3 d6d5 c3d5 f6d5 e4d5 c6b4 c1d2 b4d5 a2a3 f8c5 b2b4 c5d6 a1e1 c7c5 b4b5 h8f8 f2f3 d5f6 h2h3');
INSERT INTO evaluations VALUES(32200,30,'cp 241',NULL);
INSERT INTO evaluations VALUES(32200,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(32200,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(32200,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(32200,34,'cp 233',NULL);
INSERT INTO evaluations VALUES(32200,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(32200,36,'cp 243','g7g6 e6f7 e8f7 h5f3 f7g7 d2d3 d8d7 b1c3 b8c6 c1e3 g8f6 g1e2 f8e7 h2h4 h7h5 f3g3 d7g4 f2f3 g4g3 e2g3 a7a6 g3e2 b7b5 a2a3 g7f7 e1g1 f7e6 b2b4 h8f8 g2g3 d6d5 e4d5 f6d5 c3d5 e6d5 g1g2 d5e6');
INSERT INTO evaluations VALUES(32020,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(32020,31,'cp 241',NULL);
INSERT INTO evaluations VALUES(32020,32,'cp 238',NULL);
INSERT INTO evaluations VALUES(32020,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(32020,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(32020,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(32020,36,'cp 248','e6f7 e8f7 h5f3 f7g7 g1e2 b8c6 f3b3 b7b6 d2d3 g8f6 b1c3 d8d7 e1g1 c6a5 b3b5 c7c6 b5a4 a5b7 b2b4 f8e7 a4b3 h8f8 h2h3 a7a5 b4b5 b7c5 b3b2 c6b5 b2b5 d7b5 c3b5 d6d5 b5c7 a8c8 c7d5 f6d5 e4d5 e5e4 c1b2 g7g8');
INSERT INTO evaluations VALUES(34683,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(34683,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(34683,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(34683,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(34683,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(34683,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(34683,36,'cp 47','e7e5 f1g2 g8f6 g1f3 f8e7 e1g1 e8g8 d2d3 b8c6 c1g5 c6d4 c3e4 d4f3 g2f3 c7c6 e4f6 e7f6 g5f6 g7f6 a2a3 c8e6 b2b4 a5b6 d1d2 g8g7 a1c1 b6d4 c1c5 f8d8 d2e3 a7a6 f1c1 h7h6 e3d4 d8d4 g1f1 a8d8 f3g2 d8d6 f1e1 g7f8 c1c3 d6d8 c5a5');
INSERT INTO evaluations VALUES(36408,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(36408,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36408,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36408,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(36408,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(36408,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(36408,36,'cp -5','c8g4 h2h3 g4h5 d3e2 f8e8 c1f4 b8c6 f3e5 h5e2 c3e2 e7d6 e5d3 d6f4 d3f4 c6e7 e2g3 c7c6 f4d3 e7c8 c2c3 c8d6 f1e1 e8e1 d1e1 a7a5 a2a4 f6e4 e1e3 d8g5 g3f1 g5e3 f1e3 e4d2 a1e1 g8f8 d3c5 a8e8 c5d7 f8g8 d7b6 d2c4');
INSERT INTO evaluations VALUES(27827,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(27827,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(27827,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(27827,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27827,36,'cp 8','h2h3 g4h5 g2g4 h5g6 f3e5 g6d3 d1d3 c7c6 c3e2 f6e4 c1f4 b8d7 e2g3 d7e5 d4e5 e4c5 d3d2 c5e6 g3f5 f8e8 f4g3 d8b6 a2a4 a7a5 g1g2 e7f8 b2b3 d5d4 d2d3 b6c5 f1e1 b7b5 f5d6 c5d5 d3f3 d5f3 g2f3 f8d6 e5d6');
INSERT INTO evaluations VALUES(27826,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27826,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27826,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27826,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27826,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27826,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27826,36,'cp 0','g4h5 g2g4 h5g6 f3e5 g6d3 d1d3 c7c6 c3e2 f6e4 c1f4 b8d7 e2g3 e4g3 f4g3 d7f6 d3b3 d8c8 f1e1 f8e8 c2c4 a7a5 b3f3 c8e6 e5d3 e6d7 d3e5');
INSERT INTO evaluations VALUES(27825,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(27825,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(27825,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(27825,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27825,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27825,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(27825,36,'cp 0','g2g4 h5g6 f3e5 g6d3 d1d3 c7c6 c3e2 f6e4 c1f4 f7f6 e5c4 b8a6 c4e3 e7d6 h3h4 d6f4 e2f4 d8d6 e3g2 a8e8 c2c3 a6c7 f2f3 e4g3 f1e1 g7g5 f4g6 h7g6 d3g6 g8h8 h4g5 g3e2 e1e2 e8e2 g6h6 h8g8 h6g6');
INSERT INTO evaluations VALUES(40875,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(40875,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(40875,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(40875,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(40875,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(40875,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40875,36,'cp 5','e8g8 c1e3 b8d7 e1g1 d7c5 e3c5 d6c5 f3e5 c8e6 f1e1 e7d6 g2g3 g7g6 d3c4 d6e5 c4e6 e5g3 h2g3 f7e6 e1e6 d8d1 a1d1 a8d8 d1d8 f8d8 e6e7 b7b6 e7c7 d8d2 b2b4 d2c2 b4c5 b6c5 a2a4 c2c3 c7a7 c5c4 a4a5 c3c1 g1g2 c4c3 a7c7 c1a1 c7c3');
INSERT INTO evaluations VALUES(36281,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(36281,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(36281,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(36281,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36281,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(36281,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(36281,36,'cp 12','c1e3 b8d7 e1g1 d7c5 e3c5 d6c5 a2a4 g7g6 f3e5 e7f6 f2f4 c8e6 e5g4 f6g7 f4f5 e6d7 d1f3 d8g5 a1e1 c5c4 d3c4 g5f5 f3f5 d7f5 g4e3 f5e4 e3d5 e4d5 c4d5 a8d8 d5b7 d8b8 b7d5 b8b2 e1e7 b2c2 f1f7 f8f7 e7f7 c2c3 h2h4 g8h8');
INSERT INTO evaluations VALUES(36282,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(36282,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(36282,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(36282,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(36282,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36282,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(36282,36,'cp 9','b8d7 e1g1 d7c5 e3c5 d6c5 a2a4 e7d6 f3d2 c8d7 g2g3 g7g6 d3c4 g8g7 f1e1 d7c6 c4b5 d8d7 b5c6 d7c6 d1f3 c6f3 d2f3 a8e8 f3d2 e8e1 a1e1 f8d8 d2c4 b7b6 f2f4 d6f8 g1f2 a7a6 c4e3 d8d2 e1e2 d2e2 f2e2 f7f5 e3d5 f8d6 c3c4 g7f7 e2d2 f7e6');
INSERT INTO evaluations VALUES(41076,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(41076,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(41076,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(41076,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(41076,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(41076,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(41076,36,'cp -69','e8g8 b1c3 e7b4 g5f6 d8f6 e2e3 c8e6 f1d3 b8d7 e1g1 a8e8 f3g5 g7g6 g5e6 f7e6 f2f4 c7c5 c3b5 c5d4 b5d4 b4d2 e3d2 f6d4 g1h1 d7c5 b2b4 c5e4 d2e1 e4d6 e1d2 d6c4 d2f2 d4f2 f1f2 e6e5 g2g3 c4d6 a1d1 d6e4 d3e4 d5e4 c2c4 e5f4 f2f4 f8f4 g3f4');
INSERT INTO evaluations VALUES(36397,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(36397,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(36397,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(36397,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(36397,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(36397,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(36397,36,'cp -77','b1c3 e7b4 g5f6 d8f6 e2e3 c8e6 f1d3 b8d7 e1g1 b4d6 a1e1 d6f4 e3e2 c7c6 c3d1 a8e8 d1e3 g7g6 c2c4 f6d8 e2c2 d7f6 c4c5 e6c8 e3d1 f4c7 e1e8 f8e8 f1e1 c8g4 f3e5 g4d1 c2d1 c7e5');
INSERT INTO evaluations VALUES(36387,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(36387,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(36387,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(36387,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(36387,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(36387,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(36387,36,'cp -72','e7b4 g5f6 d8f6 e2e3 c8e6 f1d3 b8d7 e1g1 b4d6 a1e1 d6f4 e3e2 c7c6 c3d1 a8e8 d1e3 f6d8 c2c3 g7g6 e2d1 d7f6 e1e2 f4d6 g2g3 g8g7 f1e1 e6d7 d1d2 a7a5 h2h4 h7h6 e3g2 e8e2 e1e2 d7g4 e2e3 c6c5');
INSERT INTO evaluations VALUES(43053,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(43053,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(43053,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(43053,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(43053,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(43053,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(43053,36,'cp 65','g3g7 e7f6 g7h6 d7d5 d2d3 e4d6 b1c3 d6f5 h6f4 b8c6 g1f3 e8g8 f1e2 f8e8 f4d2 c6d4 e1g1 c7c6 f3d4 f5d4 e2d1 d8e7 d2e3 e7d8 e3g3 g8h8 g3f4 a7a5 a1b1 h8g8 c1e3 f6e5 f4g5 d8g5 e3g5 a5a4 a2a3 e5g7 g5d2 c8f5 d2e3 f5g6 c3e2 d4f5 e3d2');
INSERT INTO evaluations VALUES(45290,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(45290,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(45290,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(45290,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(45290,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45290,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(45290,36,'cp -17','c8d7 b5e2 e7e6 d2d4 b8c6 c1f4 f8e7 e1g1 e8g8 f1e1 a8c8 e2d3 f6h5 f4d2 e7d6 c3b5 d6b8 f3e5 g7g6 d2h6 h5g7 e5d7 d8d7 c2c3 f8e8 b5a3 e6e5 d4e5 c6e5 d3b5 e5c6 e1e8 c8e8 a3c2 g7f5 h6d2 a7a6 b5f1 h7h5 c2e3 g8g7 a2a4 h5h4');
INSERT INTO evaluations VALUES(28878,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(28878,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(28878,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(28878,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(28878,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(28878,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(28878,36,'cp -10','b5e2 d7g4 h2h3 g4f3 e2f3 e7e6 e1g1 b8c6 d2d4 f8d6 c3b5 d6e7 a2a4 e8g8 c2c3 a7a6 b5a3 h7h6 c1f4 e7d6 f4d6 d8d6 f1e1 a8c8 a3c2 g7g6 c2e3 h6h5 e3f1 h5h4 f3e2 d6f4 e2d3 c8c7 d1e2 f8c8 f1e3 g8g7');
INSERT INTO evaluations VALUES(46704,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(46704,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(46704,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(46704,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(46704,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(46704,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(46704,36,'cp 41','f8g7 g1f3 e8g8 e1g1 e7e5 d4e5 d6e5 f3e5 f6e4 c3e4 d8d1 f1d1 g7e5 c2c3 f8e8 e2b5 b8c6 c1g5 g8g7 d1e1 h7h6 g5e3 g6g5 e4c5 f7f6 f2f3 e5d6 c5e4 d6e5 e3f2 a7a6 b5c6 b7c6 e4c5 a8b8 f2g3 b8b2 c5d3 b2c2 d3e5');
INSERT INTO evaluations VALUES(47156,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(47156,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(47156,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(47156,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(47156,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(47156,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(47156,36,'cp -114','c7c5 e2e3 e7e6 c2c3 b8c6 b1d2 f8e7 f4g3 h7h5 f1d3 h5h4 g3f2 e6e5 d4c5 e7c5 d1e2 d8c7 e1c1 a7a5 c1b1 a5a4 e3e4 c5f2 e2f2 a4a3 b2b3 c8e6 g1e2 c7a5 b1a1 d5d4 c3d4 e5d4 d2c4 e6c4 d3c4 e8g8 e2d4 c6d4 f2d4 b7b5 c4d5 f6d5 e4d5 f8d8 d4e3');
INSERT INTO evaluations VALUES(56402,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(56402,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(56402,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(56402,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(56402,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(56402,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56402,36,'cp 15','e4d5 g8f6 b1c3 c7c6 d5c6 b8c6 f1b5 c8d7 g1e2 f8e7 e1g1 e8g8 b5c4 e5e4 c3e4 c6e5 f3c3 e5c4 c3c4 a8c8 e4f6 e7f6 c4b3 f8e8 e2f4 f6e5 d2d4 e5d4 c1e3 d4e3 f2e3 d7c6 a1d1 d8g5 c2c4 e8e4 f4d5 h7h6 f1f3 e4g4 f3f2 c8e8 b3c3');
INSERT INTO evaluations VALUES(56169,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(56169,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(56169,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(56169,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(56169,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(56169,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(56169,36,'cp 10','g8f6 b1c3 c7c6 d5c6 b8c6 f1b5 c8d7 g1e2 f8e7 e1g1 e8g8 b5c4 e5e4 c3e4 c6e5 f3c3 e5c4 e4f6 e7f6 c3c4 a8c8 c4b3 f8e8 e2f4 f6e5 d2d4 e5d4 c1e3 d4e3 f2e3 d7c6 a1d1 d8g5 f4d5 c6d5 d1d5 g5e3 b3e3 e8e3 d5d7 f7f6 g1f2 c8e8 f1d1 e3e2 f2f1 e8e7 d7e7 e2e7');
INSERT INTO evaluations VALUES(48615,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(48615,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(48615,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(48615,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48615,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48615,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(48615,36,'cp 5','b1c3 c7c6 d5c6 b8c6 f1b5 c8d7 g1e2 f8e7 e1g1 e8g8 b5c4 e5e4 c3e4 c6e5 f3c3 e5c4 c3c4 a8c8 e4f6 e7f6 c4b3 f8e8 e2f4 f6e5 d2d4 e5d4 c1e3 d4e3 f2e3 d7c6 a1d1 d8g5 f4d5 c6d5 d1d5 g5e3 b3e3 e8e3 f1d1 g8f8 g1f2 e3e6 d5d8 e6e8 d8d2 e8e6 h2h3 a7a6 a2a4 f8e8 a4a5 h7h6 c2c3');
INSERT INTO evaluations VALUES(56632,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(56632,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(56632,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(56632,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(56632,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(56632,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(56632,36,'cp 24','f2f4 e5f4 c1f4 g8f6 g1f3 b8c6 d3d4 f8b4 f1b5 e8g8 e1g1 c8g4 b5c6 b7c6 d1d3 g4h5 f4g5 d8d6 g5f6 d6f6 c3e4 f6e6 a2a3 b4d6 a1e1 h5g6 f3h4 e6d7 h4g6 h7g6 e1e2 a8e8 f1e1 a7a5 c2c3 c6c5 d4d5 e8b8 d3c4 f8e8 e4c5 e8e2 c5d7 e2e1 g1f2');
INSERT INTO evaluations VALUES(57600,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(57600,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(57600,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(57600,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(57600,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(57600,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(57600,36,'cp 49','c1b2 g8f6 c2c4 e7e6 g1f3 b7b6 g2g3 c8b7 f1g2 g7g6 d2d4 f8g7 e1g1 e8g8 b1c3 a7a5 e2e3 d8e7 a1c1 f6e4 h2h4 d7d6 f3g5 e4c3 b2c3 b7g2 g1g2 b8a6 d4d5 g7c3 c1c3 a6c5 d5e6 h7h6 g5h3 e7e6 h3f4 e6e5 c3c1');
INSERT INTO evaluations VALUES(57601,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(57601,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(57601,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(57601,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(57601,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(57601,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(57601,36,'cp 34','g8f6 e2e3 e7e6 f1e2 d7d5 g1f3 f8d6 e1g1 e8g8 c2c4 a7a6 d1c2 b8d7 d2d4 b7b6 b2a3 d8e7 a3d6 c7d6 b1c3 c8b7 f1c1 g8h8 a2a4 a6a5 c3b5 a8c8 c2b2 d5c4 c1c4 b7d5 c4c7 c8c7 b5c7 f8c8 a1c1');
INSERT INTO evaluations VALUES(50111,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(50111,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(50111,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(50111,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50111,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(50111,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(50111,36,'cp 54','c2c4 e7e6 g2g3 c7c6 e2e3 d7d5 b1c3 f8b4 g1f3 e8g8 a1c1 b8d7 c4d5 f6d5 c3d5 e6d5 f1g2 d8e7 e1g1 a7a5 d1c2 d7f6 f1e1 c8d7 f3d4 a8e8 c1d1 f6e4 f2f3 e4g5 b2c3 b4c5 d1c1 c5d4 c3d4 g5e6 d4b2');
INSERT INTO evaluations VALUES(6135,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(6135,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(6135,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(6135,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(6135,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(6135,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(6135,36,'cp 94','b1c3 g8f6 f1e2 e8g8 e1g1 e7e5 c1e3 f8e8 d4d5 f6g4 e3d2 a7a5 f3e1 g4f6 e1d3 b7b6 g1h1 e8f8 d1e1 f6e8 f2f3 g7f6 d2h6 f6g7 h6e3 f7f5 e4f5 g6f5 g2g4 d7c5 g4f5 c8f5 d3c5 b6c5 f1g1 g8h8 c3e4 e8f6 e3g5 d8e8 e1h4 e8g6 e2d3 f6e4 f3e4');
INSERT INTO evaluations VALUES(12519,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(12519,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(12519,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(12519,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(12519,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12519,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12519,36,'cp 0','d7d6 g2g3 f8e7 f1g2 a8b8 e1g1 e8g8 a2a4 c8g4 h2h4 f6e8 g5e7 c6e7 d1d2 e8f6 f3h2 g4d7 c3b5 d8b6 b5a3 b6c7 a3b5');
INSERT INTO evaluations VALUES(13844,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13844,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13844,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13844,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13844,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13844,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13844,36,'cp -7','g2g3 f8e7 a2a4 e8g8 f1g2 a8b8 e1g1 c8g4 h2h4 f6e8 g5e7 c6e7 d1d2 e8f6 c3d1 h7h6 b2b4 e7c6 b4c5 d6c5 d1e3 g4e6 d2c3 d8d6 e3c4 e6c4 c3c4 b7b6 f1d1 b8d8 c2c3 d6d7 c4a2 g8h8 a2b2 f6g4');
INSERT INTO evaluations VALUES(13214,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(13214,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(13214,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(13214,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(13214,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13214,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(13214,36,'cp 34','c5d4 f3d4 c8d7 c1g5 e7e6 d1d3 f8e7 e1c1 e8g8 b5c6 b7c6 d4b3 d7c8 f2f4 d8c7 g5f6 g7f6 d3h3 g8h8 d1d3 f8g8 d3g3 g8g6 f4f5 e6f5 e4f5 d6d5 c1b1 e7d6 b3d4 d6f4 h1e1 f4g3 h2g3 g6g7 h3h4 c8d7 h4f6 a8g8');
INSERT INTO evaluations VALUES(13211,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(13211,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(13211,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(13211,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(13211,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(13211,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(13211,36,'cp 32','f3d4 c8d7 c1g5 e7e6 d1d3 f8e7 e1c1 c6d4 b5d7 d8d7 d3d4 a8c8 g5f6 e7f6 d4d6 d7d6 d1d6 f6c3 b2c3 c8c3 h1d1 e8g8 d6d7 b7b5 d7a7 f8c8 d1d2 g7g5 a7d7 g8g7 c1b2 h7h5 d7b7 c3c5 f2f3 g7f6 a2a3 f6g6 h2h3 c8c6 b7d7 c5c3 d7d4');
INSERT INTO evaluations VALUES(15817,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(15817,31,'cp 213',NULL);
INSERT INTO evaluations VALUES(15817,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(15817,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(15817,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(15817,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(15817,36,'cp 237','c1g5 d5e4 d4d8 e8d8 e1c1 f8d6 c3e4 d8e7 f1d3 c8e6 f2f4 h8c8 e4d6 c7d6 d3h7 e6a2 h7e4 c8c5 e4b7 a8b8 h1e1 e7d7 g5f6 g7f6 b7f3 a2b3 d1d2 b8c8 f3d1 c5a5 d2e2 b3e6 b2b3 a5c5 e2e3 a7a5 c1d2 a5a4 b3a4 c5c4 a4a5 c4f4 a5a6 f4f2 e1e2');
INSERT INTO evaluations VALUES(17507,30,'cp -325',NULL);
INSERT INTO evaluations VALUES(17507,31,'cp -335',NULL);
INSERT INTO evaluations VALUES(17507,32,'cp -320',NULL);
INSERT INTO evaluations VALUES(17507,33,'cp -349',NULL);
INSERT INTO evaluations VALUES(17507,34,'cp -355',NULL);
INSERT INTO evaluations VALUES(17507,35,'cp -355',NULL);
INSERT INTO evaluations VALUES(17507,36,'cp -344','b7b5 d2d4 b5a4 d4d5 c6d4 f3d4 c5d4 d1d4 d7d6 d4a4 c8d7 a4a3 g8f6 b1d2 e7e5 f2f3 f8e7 d2c4 e8g8 c1e3 a6a5 c4b6 a8a6 b6a4 d7a4 a3a4 d8d7 a4d7 f6d7 b2b3 f7f5 e4f5 a6a8 c2c4 f8f5 e1c1');
INSERT INTO evaluations VALUES(18172,30,'cp -362',NULL);
INSERT INTO evaluations VALUES(18172,31,'cp -348',NULL);
INSERT INTO evaluations VALUES(18172,32,'cp -340',NULL);
INSERT INTO evaluations VALUES(18172,33,'cp -347',NULL);
INSERT INTO evaluations VALUES(18172,34,'cp -344',NULL);
INSERT INTO evaluations VALUES(18172,35,'cp -359',NULL);
INSERT INTO evaluations VALUES(18172,36,'cp -351','a4b3 c5c4 d2d4 c4b3 a2b3 d7d6 d4d5 c6e5 f3e5 d6e5 c2c4 g8f6 b1c3 e7e6 c1g5 c8b7 g5f6 d8f6 d5d6 b5b4 c3a4 b7e4 e1g1 f6g5 f2f3 e4c6 c4c5 g7g6 a4b6 g5e3 g1h1 a8d8 d1c2 e5e4 f3e4 f8d6 b6c4 e3c5 c4d6 c5d6 a1d1');
INSERT INTO evaluations VALUES(18171,30,'cp -352',NULL);
INSERT INTO evaluations VALUES(18171,31,'cp -347',NULL);
INSERT INTO evaluations VALUES(18171,32,'cp -363',NULL);
INSERT INTO evaluations VALUES(18171,33,'cp -359',NULL);
INSERT INTO evaluations VALUES(18171,34,'cp -356',NULL);
INSERT INTO evaluations VALUES(18171,35,'cp -367',NULL);
INSERT INTO evaluations VALUES(18171,36,'cp -367','c5c4 e1g1 c4b3 a2b3 d7d6 d2d4 g8f6 d4d5 c6e5 f3d4 f6e4 f1e1 e4f6 c2c4 g7g6 c4b5 c8b7 d4c6 d8c8 f2f4 a6b5 f4e5 a8a1 e5f6 a1b1 d1d3 b1a1 d3b5 b7c6 d5c6 c8c7 c1d2 a1a8 f6e7 f8g7 d2c3 g7c3 b2c3 a8c8 e1a1 c7e7 c6c7 e8f8');
INSERT INTO evaluations VALUES(18167,30,'cp -345',NULL);
INSERT INTO evaluations VALUES(18167,31,'cp -355',NULL);
INSERT INTO evaluations VALUES(18167,32,'cp -346',NULL);
INSERT INTO evaluations VALUES(18167,33,'cp -344',NULL);
INSERT INTO evaluations VALUES(18167,34,'cp -339',NULL);
INSERT INTO evaluations VALUES(18167,35,'cp -363',NULL);
INSERT INTO evaluations VALUES(18167,36,'cp -363','d2d4 c4b3 a2b3 d7d6 d4d5 c6e5 f3e5 d6e5 c2c4 g8f6 e1g1 e7e6 d1f3 f8e7 h2h3 c8b7 c1e3 e8g8 f1d1 e6d5 e4d5 g8h8 b1d2 e7b4 a1c1 a8c8 d2e4 f6e4 f3e4 d8e7 c4c5 b4c5 e3c5 c8c5 e4b4 c5c7 b4e7 c7e7 d5d6 e7d7 c1c7 f8d8 c7d7 d8d7');
INSERT INTO evaluations VALUES(18521,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(18521,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(18521,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(18521,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(18521,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(18521,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(18521,36,'cp 62','e2e4 d5e4 c3e4 g8f6 f1d3 c6c5 e1g1 c5d4 f3d4 f8e7 d4b5 f6e4 d3e4 e8g8 c1f4 e6e5 f4e3 a7a6 b5c3 d7f6 d1c2 f6e4 c2e4 d8d6 f1d1 d6e6 d1d5 f7f5 e4e5 e6e5 d5e5 e7f6 e5a5 f6c3 b2c3 c8e6 c4c5 f8d8 f2f3');
INSERT INTO evaluations VALUES(19819,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(19819,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(19819,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(19819,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(19819,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(19819,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(19819,36,'cp -84','c8g4 b1d2 e7e6 h2h3 g4h5 f1d3 c6d4 g2g4 h5g6 f3d4 g6d3 c2d3 c5d4 d2f3 g8e7 c1g5 h7h6 g5e7 f8e7 d1a4 d8d7 a4d4 a8c8 e1g1 e8g8 f1c1 d7b5 a2a4 b5a5 g1g2 c8c6 c1c6 b7c6 d4c3 a5b6 d3d4 f8c8 a1c1 c6c5 d4c5 c8c5 c3d2 b6b3 c1c5 e7c5 d2c3 b3c3 b2c3 c5b6 f3d4 b6a5 f2f4 a5c3 d4c6 c3d2 c6a7 d2f4 a7c6 f4g5');
INSERT INTO evaluations VALUES(26270,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(26270,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(26270,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(26270,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(26270,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(26270,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(26270,36,'cp -76','b1d2 e7e6 h2h3 g4h5 f1d3 c6d4 g2g4 h5g6 f3d4 g6d3 c2d3 c5d4 d2f3 g8e7 c1g5 d8b6 g5e7 f8e7 d1a4 b6c6 a4d4 e8g8 e1g1 f8c8 g1g2 c6a6 f1c1 c8c6 d4e3 a8c8 c1c6 b7c6 b2b3 c6c5 e3d2 a6b5 a1c1 a7a5 d2c2 b5c6 h3h4 a5a4 h4h5 h7h6');
INSERT INTO evaluations VALUES(25166,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(25166,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(25166,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(25166,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(25166,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(25166,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(25166,36,'cp 19','e2e3 f8d6 f3e5 a7a6 f1d3 f5d3 c2d3 c6e7 a1c1 e8g8 e1g1 f6d7 c3e2 a8c8 b2b4 e7g6 f4g5 f7f6 e5g6 h7g6 g5f4 e6e5 d4e5 f6e5 f4g3 d8e7 d1b3 c7c6 e3e4 e7f7 e4d5 f7d5 b3d5 c6d5 d3d4 d6b8 d4e5 b8e5 f1d1 c8c1 d1c1 e5g3');
INSERT INTO evaluations VALUES(25164,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(25164,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(25164,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(25164,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(25164,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(25164,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25164,36,'cp 0','f8d6 f3e5 d6e5 d4e5 f6d7 f1b5 f5g6 h2h4 h7h5 d1d2 c6e7 b5e2 e7c6 e2b5');
INSERT INTO evaluations VALUES(24107,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24107,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(24107,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24107,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24107,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24107,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24107,36,'cp 0','f4g5 d8d7 f3h4 f5g6 h4g6 h7g6 d1f3 f6h7 g5f4 h7f6 f4d6 d7d6 e1c1 g6g5 g2g3 e8c8 h2h3 c6a5 f1e2 a5c6');
INSERT INTO evaluations VALUES(29973,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(29973,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(29973,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(29973,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(29973,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(29973,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(29973,36,'cp 115','c1f4 b8c6 d1d3 d6d5 e1c1 f8e7 e4e5 f6h5 f4e3 d8d7 f1e2 e8c8 a2a3 c8b8 h1e1 a7a6 c3a2 g7g6 c1b1 f7f6 d3b3 c6a5 b3c3 a5c6 a2c1 h5g7 e5f6 e7f6 h2h3 g4f5 g2g4 f5e4 f3g5 f6g5 e3g5 d8e8 e2d3 g7e6 g5f6');
INSERT INTO evaluations VALUES(30020,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(30020,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30020,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(30020,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(30020,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(30020,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(30020,36,'cp 48','e2e3 b8d7 f1d3 c7c6 h2h3 f8e7 g1e2 e8g8 e1g1 f8e8 d1b3 d7b6 b3c2 h7h6 g5f4 a7a5 f4g3 e6d7 a1c1 e7f8 g3h4 b6c8 a2a3 c8d6 c3a4 d7c8 c1a1 c8e6 a1e1 e6c8 h4f6 d8f6 a4b6 a8a7 b6c8 e8c8 e2c3');
INSERT INTO evaluations VALUES(30017,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(30017,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(30017,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(30017,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(30017,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(30017,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(30017,36,'cp 41','b8d7 f1d3 c7c6 h2h3 f8e7 g1e2 e8g8 e1g1 f8e8 d1c2 h7h6 g5f4 b7b5 f4h2 a8c8 e2f4 d7f8 f4e6 f8e6 h2e5 f6d7 e5g3 d7f6 c3e2 e7d6 g3h4 a7a6 a2a4 e6g5 g1h1 g7g6 f1c1 d8e7 a4b5 a6b5 e2g3 h6h5');
INSERT INTO evaluations VALUES(30771,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(30771,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(30771,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(30771,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(30771,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(30771,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(30771,36,'cp 17','d2d4 e7e6 b1c3 a7a6 g1f3 d7d6 f1d3 d6e5 f3e5 g8f6 d3e2 a8a7 c1e3 b8c6 e2f3 d8a8 e5c6 b7c6 d4d5 f6d5 c3d5 c6d5 f3d5 a8d5 d1d5 e6d5 e1c1 c7c6 e3b6 a7b7 b6e3 f7f6 c2c4 d5c4 d1d4 c4c3 b2b3 f8a3 c1b1 a3c5 d4d3 b7d7 d3c3 c5e3 c3e3 e8f7');
INSERT INTO evaluations VALUES(31324,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(31324,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(31324,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(31324,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(31324,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(31324,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(31324,36,'cp -51','b8c6 g1f3 e7e6 e1g1 g8f6 h2h3 h7h6 b1c3 f8c5 c3e2 e8g8 c1f4 b7b5 c2c3 a7a5 a2a3 f5h7 d1d2 d8b6 b3c2 f8c8 b2b4 c5e7 c2b3 f6d7 f4e3 b6b7 e2d4 c6d4 e3d4 e7d6 a1c1 b7c7 b3a2 c8d8 f1e1');
INSERT INTO evaluations VALUES(33829,30,'cp 411',NULL);
INSERT INTO evaluations VALUES(33829,31,'cp 400',NULL);
INSERT INTO evaluations VALUES(33829,32,'cp 403',NULL);
INSERT INTO evaluations VALUES(33829,33,'cp 400',NULL);
INSERT INTO evaluations VALUES(33829,34,'cp 405',NULL);
INSERT INTO evaluations VALUES(33829,35,'cp 423',NULL);
INSERT INTO evaluations VALUES(33829,36,'cp 449','e1g1 b8d7 e5d7 e7d7 f1e1 f8e7 d1h5 c7c6 b1c3 e8g8 c3d5 c6d5 c4d5 e7f6 c2c3 g7g6 h5f3 f8e8 e1e3 e8e3 f2e3 d7e7 d2d4 f6g7 e3e4 c8e6 d5e6 e7e6 c1e3 a8d8 e4e5 b7b5 h2h3 f7f6 e5f6 e6f6 f3b7 d8e8 a1e1 f6f7 b7f7 g8f7');
INSERT INTO evaluations VALUES(34647,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(34647,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(34647,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(34647,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(34647,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(34647,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(34647,36,'cp 170','c1e3 e7e5 b1d2 e4g4 d4e5 g8e7 g1f3 e7f5 d1e2 b8c6 e1c1 f8e7 d2b3 c8e6 a2a3 e8g8 h2h3 g4a4 e2b5 a4b5 f1b5 f5e3 f2e3 f8d8 b5c6 b7c6 b3a5 d8d1 h1d1 e6d5 c2c4 d5e4 d1d4 e4f5 f3h4 f5e6 a5c6 e7c5 d4e4 e6d7 c6b4 a8e8 c1c2 g8f8');
INSERT INTO evaluations VALUES(34646,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(34646,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(34646,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(34646,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(34646,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(34646,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(34646,36,'cp 163','e7e5 b1d2 e4g4 d4e5 g8e7 g1f3 e7f5 d1e2 b8c6 e1c1 c8e6 d2b3 a8d8 d1d8 e8d8 h2h3 g4a4 f3g5 d8c8 g5e6 f7e6 a2a3 a4e4 e3d2 e4e5 e2e5 c6e5 g2g4 f5h4 f2f4 e5d7 f1c4 e6e5 f4f5 f8d6 h1f1 e5e4 b3d4');
INSERT INTO evaluations VALUES(39566,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(39566,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(39566,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(39566,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(39566,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(39566,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(39566,36,'cp 24','d2d4 g8f6 g1f3 e8g8 f1e2 c7c5 d4d5 e7e6 b1c3 d7d6 e1g1 e6d5 c4d5 c8g4 h2h3 g4f3 e2f3 a7a6 a2a4 b8d7 c1d2 a8b8 a4a5 b7b5 a5b6 d8b6 f3e2 b6b2 a1a6 f6e8 c3a4 b2b7 d2c3 g7c3 a4c3 c5c4 e2c4');
INSERT INTO evaluations VALUES(41804,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41804,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(41804,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(41804,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41804,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41804,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41804,36,'cp 0','f1d3 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 c7c5 e1g1 d8c7 c2c3 h7h6 e4c2 e8g8 d1d3 d7f6 d4c5 d6c5 f1e1 f8d8 c1f4 d8d3 f4c7 d3d7 c7e5 d7d8 e5c7');
INSERT INTO evaluations VALUES(41805,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41805,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41805,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41805,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41805,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41805,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41805,36,'cp 0','b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 c7c5 e1g1 e8g8 c2c3 h7h6 e4c2 d8c7 d1d3 d7f6 d4c5 d6c5 f1e1 f8d8 c1f4 d8d3 f4c7 d3d7 c7e5 d7d8 e5c7');
INSERT INTO evaluations VALUES(42728,30,'cp 290',NULL);
INSERT INTO evaluations VALUES(42728,31,'cp 306',NULL);
INSERT INTO evaluations VALUES(42728,32,'cp 305',NULL);
INSERT INTO evaluations VALUES(42728,33,'cp 302',NULL);
INSERT INTO evaluations VALUES(42728,34,'cp 304',NULL);
INSERT INTO evaluations VALUES(42728,35,'cp 273',NULL);
INSERT INTO evaluations VALUES(42728,36,'cp 270','d4e5 f6e4 d1g4 e4c3 g4b4 c3d5 b4g4 e8f8 c1g5 d7d6 g4g3 d5e7 e1c1 b8c6 e5d6 c7d6 g5d2 e7f5 g3f4 d8h4 g2g4 f5e7 h2h3 h7h5 f1g2 h4f6 f4f6 g7f6 d2c3 h5g4 d1d6 c8e6 h3g4 h8h1 g2h1 e6g4 c3f6 a8d8 d6d8 c6d8 c1d2 g4d7 g1e2 d7c6 h1c6 d8c6 e2d4 c6b4 a2a3 b4d5 f6e7 f8e7 d2d3 d5f4 d3e4 f4e6 d4e6 e7e6 e4f4 e6d5');
INSERT INTO evaluations VALUES(43384,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43384,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43384,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43384,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43384,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(43384,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(43384,36,'cp 5','c5d4 f3d4 e7e6 a2a4 f8e7 c1e3 e8g8 e1g1 b8c6 h2h3 c8d7 f2f4 a8c8 c4b3 c6a5 e4e5 f6e8 d1e2 a5b3 c2b3 d6e5 f4e5 f7f5 e5f6 f8f6 d4f3 d7c6 a1d1 d8c7 e3d4 c6f3 f1f3 f6f3 e2f3 e8f6 c3e4 c8d8 d4f6 d8d1 f3d1 g7f6');
INSERT INTO evaluations VALUES(43400,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43400,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43400,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43400,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(43400,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(43400,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(43400,36,'cp 9','f3d4 e7e6 a2a4 f8e7 c1e3 e8g8 e1g1 b8c6 h2h3 c8d7 c4d3 d8c7 d1e2 d6d5 e4d5 e6d5 d4c6 d7c6 e3d4 a8e8 e2f3 f6e4 c3e4 d5e4 d3e4 c6e4 f3e4 e7c5 e4d3 c5d4 d3d4 c7c2 f1c1 c2e4 c1c4 e4d4 c4d4 e8e7 g1f1 f8e8 d4b4 g7g6 a1c1 g8g7 b4b6 e7d7 b2b4 e8d8 a4a5 d8e8');
INSERT INTO evaluations VALUES(44615,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44615,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(44615,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44615,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44615,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(44615,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44615,36,'cp 17','c2c4 e7e6 b1c3 d7d5 e2e3 d5c4 f1c4 a7a6 e3e4 b7b5 c4b3 b8d7 c1e3 c8b7 e1g1 d7c5 e4e5 c5b3 d4b3 d8d1 f1d1 f6d5 c3d5 b7d5 d1c1 f8e7 f2f3 h7h6 g1f2 e8g8 e3c5 e7c5 c1c5 a8c8 c5c8 f8c8 a1c1 d5c4 f2e3 c8c7');
INSERT INTO evaluations VALUES(45605,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(45605,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(45605,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(45605,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(45605,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(45605,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(45605,36,'cp -18','e6d5 f1d3 f8d6 g1e2 e8g8 c1d2 f8e8 e1g1 b7b6 e2g3 c6c5 d3b5 c8d7 b5d7 b8d7 a1c1 a8c8 d1f3 d6g3 h2g3 f6e4 c1c2 d8g5 f3f4 g5f4 g3f4 c5d4 e3d4 g7g6 f2f3 e4d2 c2d2 d7f6 g1f2 g8g7 f1h1 c8c4');
INSERT INTO evaluations VALUES(47852,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47852,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(47852,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(47852,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(47852,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(47852,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(47852,36,'cp 32','e5d4 f3d4 g7g6 b1c3 f8g7 c1g5 h7h6 g5h4 e8g8 e1g1 c7c6 g1h1 b8d7 f2f4 d8c7 a2a4 b7b6 d1e2 f8e8 a1e1 d7c5 f4f5 g6g5 h4g3 c8b7 h2h4 g5g4 e1d1 c7d7 b2b4 c5d3 c2d3 c6c5 d4b5 c5b4');
INSERT INTO evaluations VALUES(49191,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(49191,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(49191,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(49191,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(49191,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(49191,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(49191,36,'cp 110','b1c3 d7d5 c1f4 f8g7 e2e3 e8g8 f1e2 d5c4 e2c4 b7b6 e1g1 c8b7 f1e1 c7c5 d4c5 d8d1 a1d1 b6c5 e3e4 b8c6 e4e5 f6h5 f4c1 c6b4 c4e2 f7f6 d1d7 b7c6 d7d6 f6e5 d6e6 h5f4 c1f4 e5f4 f3e5 a8e8 e6c6 g7e5 c6c5 e5d4 e2c4 g8h8 e1e8 d4c5');
INSERT INTO evaluations VALUES(51943,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(51943,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(51943,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51943,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(51943,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51943,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(51943,36,'cp 21','g1f3 g8f6 f1e2 b8c6 e1g1 d7d5 d2d4 d5c4 d4c5 d8d1 f1d1 f8c5 e2c4 c8d7 c4e2 c5e7 b1c3 e8g8 e3e4 f8d8 c1e3 d7e8 f3d4 a7a6 d4c6 d8d1 c3d1 e8c6 f2f3 c6b5 e2b5 a6b5 a2a3 b5b4 a3a4 a8c8');
INSERT INTO evaluations VALUES(51944,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(51944,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(51944,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(51944,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(51944,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(51944,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(51944,36,'cp 13','g8f6 d2d4 c5d4 e3d4 f8b4 b1c3 e8g8 f1d3 d7d5 c4d5 e6d5 e1g1 h7h6 c1f4 b8c6 h2h3 f8e8 f1e1 e8e1 d1e1 c8e6 a2a3 b4d6 c3e2 f6e4 f3e5 c6e7 a1c1 a8c8 c1c8 e6c8 e5f3 c8f5 f4e5 d8b6 e5d6 e4d6');
INSERT INTO evaluations VALUES(44201,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(44201,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44201,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(44201,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44201,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(44201,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(44201,36,'cp 30','b1c3 d7d5 d2d4 b8c6 a2a3 a7a6 d4c5 f8c5 b2b4 c5a7 c1b2 e8g8 f1d3 d5c4 d3c4 d8e7 e1g1 a7b8 d1c2 c6e5 c4e2 c8d7 f3e5 b8e5 f2f4 e5c3 c2c3 f8c8 c3e5 d7c6 f1d1 h7h6 a1c1 c8d8 d1d8 e7d8 c1d1 d8b6');
INSERT INTO evaluations VALUES(44196,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(44196,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(44196,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(44196,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44196,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44196,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(44196,36,'cp 13','b8c6 d2d4 d7d5 a2a3 d5c4 f1c4 a7a6 c4a2 f8e7 e1g1 b7b5 d1c2 c5d4 e3d4 c8b7 d4d5 e6d5 c1g5 e8g8 a1d1 c6a5 g5f6 e7f6 a2d5 b7d5 c3d5 g7g6 b2b3 a8c8 c2a2 g8g7 h2h3 f8e8 g1h1 a5c6 b3b4');
INSERT INTO evaluations VALUES(54431,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(54431,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(54431,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(54431,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(54431,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(54431,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(54431,36,'cp -70','d1d4 b8c6 d4f4 d5d4 c3e4 d8a5 c1d2 a5e5 f4e5 c6e5 f2f4 e5c6 f1d3 g8f6 g1f3 a7a6 e1c1 f6d5 h1e1 f8e7 e4f2 e8g8 d3e4 f8d8 f2d3 f7f5 e4d5 d8d5 d3e5 a6a5 e5c6 b7c6 c2c4 d5d8 f3g5 e7g5 f4g5 c6c5 e1e5 c8b7 e5c5 a5a4 d2f4 b7g2 c5c7 g2e4');
INSERT INTO evaluations VALUES(54424,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(54424,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(54424,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(54424,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(54424,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(54424,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(54424,36,'cp -60','b8c6 d4f4 d5d4 c3e4 d8a5 c1d2 a5e5 f4e5 c6e5 f2f4 e5c6 f1d3 g8f6 g1f3 a7a6 e1c1 f6d5 h1e1 f8e7 e4f2 e8g8 d3e4 f8d8 g2g4 a8b8 g4g5 b7b5 h2h4 b5b4 h4h5 a6a5 f2d3 a5a4 d3e5 c6e5 f3e5 c8b7 g5g6 h7g6 h5g6 f7f5 e4f3 b4b3 a2b3 a4b3 c2b3 e7f6 c1b1 d8c8');
INSERT INTO evaluations VALUES(20428,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(20428,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(20428,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(20428,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(20428,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(20428,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(20428,36,'cp -55','d4f4 d5d4 c3e4 d8a5 c1d2 a5e5 f4e5 c6e5 f2f4 e5c6 f1d3 g8f6 g1f3 a7a6 e1c1 f6d5 h1e1 f8e7 e4f2 e8g8 d3e4 f8d8 g2g4 a6a5 c1b1 a8a6 g4g5 a6b6 b1a1 f7f5 e4d5 d8d5 f2d3 a5a4 h2h4 c8d7 a1b1 e7d6 h4h5 d7e8 e1h1 b6b5');
INSERT INTO evaluations VALUES(55651,30,'cp -580',NULL);
INSERT INTO evaluations VALUES(55651,31,'cp -585',NULL);
INSERT INTO evaluations VALUES(55651,32,'cp -588',NULL);
INSERT INTO evaluations VALUES(55651,33,'cp -604',NULL);
INSERT INTO evaluations VALUES(55651,34,'cp -594',NULL);
INSERT INTO evaluations VALUES(55651,35,'cp -598',NULL);
INSERT INTO evaluations VALUES(55651,36,'cp -628','d8h4 e1e2 h4e4 e2f2 e4d4 f2e2 c8f5 d2d3 d6d5 c2c3 d4b6 d1b3 g8f6 g1f3 b8d7 b3b6 a7b6 f3h4 f5e6 e2e1 f8d6 f1e2 h7h6 h4f3 e8d8 a2a3 e5e4 d3e4 f6e4 f3d4 h8e8 d4e6 e8e6 b1d2 e4c5');
INSERT INTO evaluations VALUES(56365,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(56365,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(56365,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(56365,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(56365,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(56365,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(56365,36,'cp 126','d4e5 f7f6 e5f6 g8f6 c1g5 c8e6 e2e3 f8d6 g1f3 d8e7 f1e2 b8c6 b1d2 e8c8 e2b5 c6e5 f3e5 d6e5 d2f3 e5d6 g5f6 g7f6 f3d4 c8b8 d4e6 e7e6 d1f3 c7c6 b5d3 h7h5 g2g3 e6e7 h2h4 b8a8 d3c2 a7a6 a1d1 d6c7');
INSERT INTO evaluations VALUES(7232,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(7232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7232,36,'cp 0','d2d4 f6e4 d1e1 f7f5 d4e5 d6e7 b1c3 e7b4 e1e3 b4c5 e3e1');
INSERT INTO evaluations VALUES(7219,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7219,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7219,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7219,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7219,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7219,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7219,36,'cp 0','f6e4 d1e1 f7f5 d4e5 d6e7 b1c3 e7b4 e1e3 b4c5 e3e1');
INSERT INTO evaluations VALUES(9389,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9389,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(9389,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9389,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9389,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(9389,36,'cp 8','e8g8 e1g1 a7a5 f1e1 c8e6 c1g5 e6b3 a2b3 h7h6 g5h4 g7g5 h4g3 f8e8 b1a3 d6d5 e4d5 d8d5 a3c4 e5e4 d3e4 d5d1 a1d1 f6e4 g1f1 e4g3 h2g3 e8d8 f3d2 g8g7 d2e4 c5f8 g3g4 g7g6 e4g3 b7b5 d1d8 a8d8 c4e5 c6e5 e1e5 d8d1 f1e2');
INSERT INTO evaluations VALUES(3860,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(3860,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3860,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3860,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3860,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3860,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3860,36,'cp 0','e1g1 a7a5 h2h3 c8e6 b3c2 d6d5 e4d5 d8d5 c1g5 e5e4 d3e4 f6e4 g5h4 f8e8 d1d5 e6d5 f1d1 c6e7 h4e7 c5f2 g1h2 f2g3 h2g1');
INSERT INTO evaluations VALUES(3859,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3859,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3859,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3859,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3859,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3859,36,'cp 0','a7a5 f1e1 f6g4 e1e2 d8f6 h2h3 g4h6 e2e1 g8h8 b1a3 f6g6 f3h4 g6f6');
INSERT INTO evaluations VALUES(12686,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(12686,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(12686,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(12686,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(12686,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(12686,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(12686,36,'cp 49','d2d4 c6d4 f3d4 c5d4 d1d4 e7c6 d4c3 b7b6 f1d1 d8c7 a2a4 c7e5 c4b5 e5c3 b1c3 c8b7 a4a5 b6a5 c1e3 f8e7 f2f3 a7a6 b5a4 h7h5 c3e2 a8c8 c2c3 c6e5 e3f4 e5c6 e2c1 h5h4 h2h3 g7g5 f4e3 d7d6 c1b3 e8g8 e3b6 c8b8');
INSERT INTO evaluations VALUES(13882,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(13882,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(13882,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(13882,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(13882,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(13882,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(13882,36,'cp 34','d4f3 c8e6 d1e2 f8e7 e1g1 e8g8 c1g5 a8c8 c4b3 f6d7 g5e7 c6e7 f1d1 c8c6 f3d2 a7a6 d2c4 d8c7 c4e3 d7f6 f2f3 b7b5 d1d2 e7g6 e2f2 g6f4 a1d1 f8c8 g1h1 b5b4 c3a4 c7e7 c2c4 b4c3 a4c3 h7h6 e3f5 e6f5 e4f5 e7c7 g2g3 f4h3');
INSERT INTO evaluations VALUES(17272,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(17272,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(17272,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(17272,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(17272,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(17272,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(17272,36,'cp -13','d7d5 e4d5 f6d5 g1f3 e7e5 f3e5 c6e5 d1e2 f7f6 d3d4 f8b4 c1d2 e8g8 d4e5 b4c3 b2c3 f6e5 c3c4 d5b6 d2c3 f8e8 a1d1 d8f6 c4c5 f6c6 e2b5 c6b5 f1b5 c7c6 e1g1 c6b5 c5b6 a7b6 d1d5 a8a2 f1c1 c8f5 d5b5 e8c8 c3b2 c8c6 b5b4 f5c2 b2e5 c2a4 c1c6 a4c6 b4b6');
INSERT INTO evaluations VALUES(14323,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(14323,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(14323,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(14323,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(14323,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(14323,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(14323,36,'cp -21','e4d5 f6d5 g1f3 e7e5 f3e5 c6e5 d1e2 f7f6 d3d4 f8b4 c1d2 e8g8 d4e5 b4c3 b2c3 f8e8 e1c1 e8e5 e2b5 d8e7 b5b2 c8g4 f2f3 g4h5 f1d3 h5f7 h1e1 d5f4 d3f1 f4e2 f1e2 e5e2 e1e2 e7e2 b2b7 a8f8 g2g3 e2h2 d2f4 h2e2 b7c7 f7g6 d1d2 e2f3 d2d8 f3f1 c1b2 f1b5 b2a3 f8d8 c7d8 b5e8');
INSERT INTO evaluations VALUES(17830,30,'cp 490',NULL);
INSERT INTO evaluations VALUES(17830,31,'cp 484',NULL);
INSERT INTO evaluations VALUES(17830,32,'cp 481',NULL);
INSERT INTO evaluations VALUES(17830,33,'cp 482',NULL);
INSERT INTO evaluations VALUES(17830,34,'cp 489',NULL);
INSERT INTO evaluations VALUES(17830,35,'cp 505',NULL);
INSERT INTO evaluations VALUES(17830,36,'cp 508','d1h5 g7g6 e5g6 h7g6 h5h8 d8e7 h8h4 c8e6 d2d3 e8c8 b1c3 e7f7 h4g3 g8e7 c1f4 d8d7 f4e3 f6f5 e1c1 c8b8 h2h4 e7c8 e4e5 c8e7 a2a3 e7d5 c3d5 d7d5 d3d4 d5d7 h4h5 g6h5 g3g5 b7b6 g5h5 f5f4 h5f7 d7f7 e3d2 c6c5');
INSERT INTO evaluations VALUES(18696,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(18696,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(18696,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(18696,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(18696,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(18696,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(18696,36,'cp -29','g8e7 g1f3 a7a6 b5e2 d7d5 e1g1 d5e4 c3e4 b7b6 b2b3 e7f5 c1b2 f8e7 e4g3 e8g8 f3e5 c6e5 g3f5 e6f5 f4e5 c8e6 a2a4 d8d7 e2d3 b6b5 d1f3 a8d8 d3f5 e6f5 f3f5 d7f5 f1f5 d8d2 f5f2 d2d7 g2g3 f8d8 b2c1 c5c4 b3c4 b5c4 c1e3 h7h5');
INSERT INTO evaluations VALUES(22917,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(22917,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(22917,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(22917,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(22917,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(22917,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(22917,36,'cp 16','d7d5 b1c3 g8f6 f1b5 f8b4 e1g1 e8g8 b5c6 b7c6 f3e5 b4c3 b2c3 c8f5 f1e1 d8d6 c1f4 f8e8 d1f3 f5e4 f3e2 e4f5 f2f3 f6h5 f4e3 f7f6 e5d3 f5d3 e2d3 a7a5 c3c4 h5f4 d3c3 f4e6 e3f2 d5c4 c3c4 d6d5 c4c3 e6f4');
INSERT INTO evaluations VALUES(24727,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(24727,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(24727,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(24727,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(24727,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(24727,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(24727,36,'cp -9','h2h3 g4f3 e2f3 f8d6 c3b5 d6e7 a2a4 e8g8 c2c3 h7h6 b5a3 a7a6 a3c2 d8c7 g2g3 e7d6 c2e3 a8d8 d1d3 f8e8 f1e1 d6f8 d3d1 e6e5 d4e5 c6e5 f3g2 f8c5 e3f5 c7b6 c1e3 e5c4 e3c5 b6c5 d1d4 c5d4 f5d4 e8e1 a1e1 c4b2');
INSERT INTO evaluations VALUES(29313,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(29313,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(29313,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29313,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(29313,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(29313,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(29313,36,'cp 17','e1g1 f8d6 g3e4 f6e4 d3e4 b8d7 c2c4 e8g8 c1g5 d6e7 g5e7 d8e7 f1d1 h7h6 e4e2 a7a5 d1d3 e7d6 g2g3 f8d8 h2h4 a5a4 a1c1 d6b4 d3d2 a8c8 c1d1 b4a5 d4d5 c6d5 c4d5 d8e8 e2e3 e6d5 d2d5 e8e3 d5a5 e3f3');
INSERT INTO evaluations VALUES(29314,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(29314,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(29314,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(29314,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(29314,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(29314,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(29314,36,'cp 18','f8d6 g3e4 f6e4 d3e4 b8d7 c1g5 d6e7 g5f4 e8g8 c2c4 d7f6 e4e2 c6c5 f1d1 c5d4 f4e5 d8a5 e5d4 f8d8 b2b3 e7c5 d4f6 g7f6 f3e1 f6f5 a1c1 d8d1 c1d1 a8d8 d1d3 d8d4 d3d4 c5d4 g1f1 d4f6 g2g3 h7h6 e1d3 b7b6 f1g2 g8g7');
INSERT INTO evaluations VALUES(30418,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(30418,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(30418,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(30418,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(30418,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(30418,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(30418,36,'cp 197','b1a3 b5d7 f1c4 e7e6 g1e2 f8b4 e2c3 b8c6 a3c2 a7a5 d4d5 c6e5 c4e2 b4d6 d5e6 f7e6 e1g1 g8e7 f2f4 e5g6 c1e3 e8g8 g2g3 a8b8 e2c4 e7c6 b2b3 d8e7 e4e5 g6e5 f4e5 c6e5 a1a5 d7c6 c2d4 d6c5 c4e6 g8h8 e6d5 f8f1 g1f1 c5b6 d5c6 b6a5');
INSERT INTO evaluations VALUES(32250,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(32250,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(32250,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32250,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(32250,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(32250,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(32250,36,'cp 27','f3d4 d8e7 d4e6 e7e6 g2g3 g8f6 c2c3 e6e2 e1e2 f8d6 d2d4 e8g8 e2f3 f8e8 a2a4 c7c5 c1e3 f6e4 b1d2 c5d4 c3d4 e4d2 e3d2 b8c6 d2e3 c6b4 a1c1 b4a2 c1c2 a8c8 c2c8 e8c8 f1d3 a2b4 d3f5 c8c7 h1c1 c7c1');
INSERT INTO evaluations VALUES(33943,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(33943,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(33943,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(33943,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(33943,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(33943,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(33943,36,'cp -61','g1f3 c8f5 f1e2 b8c6 e1g1 e8c8 c1e3 f8e7 a2a3 d5e6 c2c3 e5e4 f3d4 c6d4 c3d4 e4d3 e2d3 f5d3 d1d3 d8d5 b2b4 g7g5 a1c1 e6d7 c1c2 c8b8 d3b3 h7h5 a3a4 e7d6 c2c1 d7f5 b4b5 h8e8 h2h3 b7b6 c1c4 g5g4 h3h4 f5d7 f1e1');
INSERT INTO evaluations VALUES(34984,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(34984,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(34984,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(34984,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(34984,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(34984,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(34984,36,'cp 29','g1f3 g8f6 g2g3 f8d6 f1g2 e8g8 e1g1 c7c6 f1e1 f8e8 h2h3 c8f5 g3g4 f5g6 g4g5 f6d5 c3e4 a5c7 d3d4 e5d4 e4d6 e8e1 d1e1 c7d6 e1e8 d6f8 e8f8 g8f8 f3d4 b8a6 g2d5 c6d5 c1f4 a8d8 f4d2 a6b8 a1e1 b8c6 d4c6 b7c6 d2b4 f8g8 e1e7 g6c2 b4c5 c2f5 e7a7 f5h3');
INSERT INTO evaluations VALUES(34983,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(34983,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(34983,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(34983,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(34983,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34983,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(34983,36,'cp 25','g8f6 g2g3 f8d6 f1g2 e8g8 e1g1 h7h6 f1e1 c7c6 f3d2 c8e6 d2c4 a5c7 c4d6 c7d6 b2b3 b8d7 d1d2 f8e8 c1b2 a8d8 e1e2 d6c7 a1e1 b7b5 a2a4 e6g4 f2f3 g4f5 c3d1 a7a6 d1f2 c6c5 f3f4 e5f4 e2e8 d8e8 e1e8 f6e8 d2f4');
INSERT INTO evaluations VALUES(33918,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(33918,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(33918,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(33918,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(33918,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(33918,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(33918,36,'cp 34','c1d2 f8d6 f1e2 c7c5 e1g1 b8c6 d2g5 c8e6 f3d2 c6d4 d2c4 a5a6 g5f6 g7f6 c4d6 a6d6 c3e4 d6e7 c2c3 d4e2 d1e2 e8c8 e2e3 b7b6 f2f4 c5c4 f4e5 d8d3 e3f4 f6e5 f4e5 h8g8 f1f2 e7c7 e5f6 g8g6 f6h8 c7d8 h8d8 d3d8 e4f6 h7h6 a1e1 a7a5 f6h5 g6g5 h5f4 g5f5 f4e6 f7e6 f2f5');
INSERT INTO evaluations VALUES(35062,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(35062,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(35062,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(35062,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(35062,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(35062,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(35062,36,'cp 143','f6d8 g1f3 d7d5 f1d3 c7c5 c2c3 b8c6 e1g1 a7a6 f1e1 c8d7 a2a3 c5c4 d3c2 h7h6 h2h4 d8c7 b1d2 c6a5 h4h5 g8e7 f3h4 e7c6 f2f4 f8e7 d2f3 a5b3 c2b3 c4b3 d1b3 e8c8 c1e3 c6a5 b3c2 c8b8 f4f5 a5c4 f5e6 d7e6 h4f5 d8c8 f5g7 c4e3 e1e3');
INSERT INTO evaluations VALUES(35591,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(35591,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(35591,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(35591,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(35591,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(35591,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(35591,36,'cp 167','g2g4 h7h5 h2h3 b8c6 e1c1 c6d4 e3d4 h5h4 d2g5 b7b5 g5h4 c8b7 g4g5 f6h5 d4g7 g8g7 h4f2 d8a5 a2a3 g7g8 c3a2 b5b4 a2b4 a5g5 c1b1 g5g3 f2g3 h5g3 h1g1 g3f1 d1f1 g8g7 b4d3 a8e8 h3h4 a6a5 h4h5');
INSERT INTO evaluations VALUES(36553,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(36553,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(36553,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36553,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36553,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36553,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36553,36,'cp 22','c8f5 d1b3 d8c8 b1c3 a7a5 f3h4 f5d7 e3e4 a5a4 b3d1 e7e5 h4f3 a4a3 b2b3 e5d4 f3d4 b8c6 d4c2 f6h5 c1d2 h5f4 a1c1 f8e8 f1e1 f4e2 d1e2 f7f5 c3d5 d7e6 e2d1 f5e4 e1e4 c6e5 d1e2 e6f7 d2c3 c7c6');
INSERT INTO evaluations VALUES(41279,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(41279,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(41279,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41279,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41279,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41279,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41279,36,'cp 0','e7e5 b1c3 d8e7 e3e4 b8c6 d4e5 c6e5 e1g1 e8g8 f3e5 d6e5 c1g5 c7c6 d1c1 f8d8 d3e2 c8e6 h2h3 e7b4 b2b3 h7h6 g5d2 b4f8 d2e3 g8h7 c1c2 h6h5 e3g5 d8d4 g5e3');
INSERT INTO evaluations VALUES(42540,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(42540,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(42540,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(42540,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(42540,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(42540,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(42540,36,'cp -26','h7h6 g5h4 b8c6 c2c3 d7d6 b2b4 c5b6 f1e2 g7g5 h4g3 c8e6 b1d2 d8e7 e1g1 f6h5 d2c4 e6c4 d3c4 e8c8 a2a4 a7a5 a1b1 c8b8 b4a5 c6a5 f3d2 h5f4 d2b3 a5b3 b1b3 b8a8 g3f4 e5f4 b3b5 h6h5 d1d5');
INSERT INTO evaluations VALUES(43595,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(43595,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(43595,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(43595,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(43595,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43595,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43595,36,'cp -16','g1f3 e7e6 f1d3 c7c5 d4c5 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d7c5 d1d8 e8d8 f3d2 c5e4 d2e4 c8d7 c1e3 d7c6 e1c1 d8e8 e4d6 f8d6 d1d6 a8d8 d6d8 e8d8 h1d1 d8e7 g2g3 f7f6 h2h4 h8c8 b2b3 e6e5 c1b2 c6f3 d1d3 f3e4 d3c3 c8c6 c3c6 e4c6 c2c4 e7f7 a2a3');
INSERT INTO evaluations VALUES(43596,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(43596,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(43596,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(43596,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43596,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(43596,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(43596,36,'cp -20','e7e6 f1d3 c7c5 d4c5 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d7c5 d1d8 e8d8 f3d2 c5e4 d2e4 c8d7 c1e3 d7c6 e1c1 d8e8 e4d6 f8d6 d1d6 a8d8 d6d8 e8d8 h1d1 d8e7 g2g3 f7f6 h2h4 h8c8 d1d2 c8g8 b2b3 e6e5 a2a4 e7e6 f2f4 e6f5 f4e5 f6e5 d2d6 f5g4');
INSERT INTO evaluations VALUES(43664,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(43664,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(43664,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(43664,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43664,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(43664,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(43664,36,'cp -16','f1d3 c7c5 d4c5 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d7c5 d1d8 e8d8 f3d2 c5e4 d2e4 c8d7 c1e3 d8e8 e1c1 d7c6 f2f3 h7h5 e4d6 f8d6 d1d6 a8d8 d6d8 e8d8 h1d1 d8e7 e3c5 e7f6 c1d2 h8d8 d2e1 d8c8 c5d4 e6e5 d4b6 g7g5 d1d8 c8d8 b6d8 f6f5 e1f2 g5g4 f3g4 f5g4 c2c4 g4f5 g2g3');
INSERT INTO evaluations VALUES(43665,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(43665,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(43665,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(43665,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(43665,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(43665,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(43665,36,'cp -23','c7c5 d4c5 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d7c5 d1d8 e8d8 f3d2 c5e4 d2e4 c8d7 c1e3 d8e8 e1c1 d7c6 e4d6 f8d6 d1d6 a8d8 d6d8 e8d8 e3c5 h7h5 h1e1 d8d7 e1d1 d7e8 h2h4 f7f6 g2g3 e8f7 c5e3 h8g8 d1d4 f7e7 d4c4 e6e5 f2f4 e7e6 c4c5 c6f3 f4e5 f6e5 c5c7');
INSERT INTO evaluations VALUES(43639,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(43639,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(43639,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(43639,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(43639,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(43639,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(43639,36,'cp -29','d5c4 a2a4 b8c6 e2e3 c6a5 f3d2 c7c5 d4c5 f8c5 f1c4 c8d7 e1g1 c5e7 c4e2 d8b6 a1b1 b6b4 d2f3 h7h6 g5f4 a5c6 f3e5 c6e5 f4e5 d7c6 d1d4 e8g8 f1d1 f8d8 d4b4 e7b4 d1d4 d8d4 e5d4 f6d7 b1c1 a8d8 c3a2 b4f8 e2f3');
INSERT INTO evaluations VALUES(45859,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(45859,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(45859,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(45859,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45859,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(45859,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(45859,36,'cp -12','b7b5 c4b3 f8e7 d2d4 d8c7 e1g1 e8g8 f1e1 b8d7 h2h3 b5b4 c3a4 e5d4 f3d4 d7e5 c2c3 c6c5 d4f5 c8f5 e4f5 c5c4 b3c2 a7a5 f2f4 e5d3 c2d3 c4d3 d1d3 f8e8 c1e3 c7c6 b2b3 d6d5 e3d4 a8c8 a1d1 e7d6 e1e8 c8e8 c3b4');
INSERT INTO evaluations VALUES(45901,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(45901,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(45901,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(45901,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(45901,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(45901,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(45901,36,'cp 38','h7h6 h2h3 c8e6 f3h2 b8c6 h2g4 f8e7 g4e3 f6d7 a2a4 e8g8 c3d5 e7g5 e1g1 d7b6 d5b6 a7b6 e3d5 g5c1 d1c1 c6d4 c1d2 e6d5 c4d5 d8d7 f2f4 a8a4 a1a4 d7a4 f4e5 d6e5 c2c3 a4c2 d2c2 d4c2 f1f5 c2e3 f5e5 f8d8 g1f2 e3d1 f2f3 d1b2');
INSERT INTO evaluations VALUES(45570,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(45570,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(45570,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(45570,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(45570,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45570,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(45570,36,'cp 36','f3g1 a7a6 a2a4 c8e6 g1e2 d6d5 e4d5 f6d5 e1g1 f8e7 f2f4 e8g8 f4e5 b8c6 c4d5 e6d5 c1f4 d5e6 f4g3 d8d7 e2f4 c6e5 f4e6 d7e6 f1e1 e7d6 c3e4 f7f6 g3f2 a8c8 e4d6 e6d6 f2g3 d6d4 g1h1 d4d7 g3e5');
INSERT INTO evaluations VALUES(46677,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46677,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46677,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(46677,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46677,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46677,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46677,36,'cp -1','b8d7 g1f3 f8g7 h2h3 e8g8 f1c4 b7b6 a2a4 a7a6 e1g1 c8b7 f1e1 d6d5 c4f1 c7c5 c3b1 f6e4 b1d2 e4d2 d1d2 e7e6 c2c3 d8e7 b2b3 f8d8 a1d1 a8c8 d2b2 d8e8 f4h2 e8d8');
INSERT INTO evaluations VALUES(48356,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(48356,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(48356,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(48356,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(48356,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(48356,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(48356,36,'cp -67','f3f4 f8e7 g1f3 c7c5 f3e5 e8g8 b1d2 b7b6 c2c3 d8c7 d1f3 f6e8 f1d3 c8a6 d3a6 b8a6 e1g1 f7f6 e5d3 e8d6 f3e2 a8c8 d3f2 c7b7 f1e1 f8e8 b2b3 c5d4 e3d4 e7f8 c1b2 a6b8 a1c1 b8c6 b2a3 b7d7 a3d6 f8d6 g2g3 g8h8');
INSERT INTO evaluations VALUES(48381,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(48381,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(48381,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(48381,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(48381,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(48381,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(48381,36,'cp 101','e4e5 f6d7 c2c3 c7c5 f1d3 f8e7 e1g1 b8c6 c1e3 d8b6 b2b3 e8g8 b1d2 c5d4 c3d4 f7f5 e5f6 d7f6 f3e5 c8d7 e5d7 f6d7 d2f3 d7f6 a1c1 e7a3 c1c3 a7a5 d1e2 c6b4 d3b1 b6a6 f1e1 a6e2 e1e2');
INSERT INTO evaluations VALUES(48429,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(48429,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(48429,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(48429,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(48429,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(48429,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(48429,36,'cp 89','f6d7 f1d3 c7c5 c2c3 b8c6 e1g1 f8e7 c1e3 d8b6 b2b3 e8g8 b1d2 c5d4 c3d4 f7f5 e5f6 d7f6 f3e5 c8d7 d2f3 d7e8 e5c6 e8c6 a2a4 e7d6 f3e5 a7a5 a1c1 a8c8 g2g3 d6a3 c1c3 a3b4 c3c2 b4a3 e5c6 b7c6 d1e2 b6b3');
INSERT INTO evaluations VALUES(49166,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(49166,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(49166,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(49166,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(49166,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(49166,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(49166,36,'cp 99','c2c3 c7c5 f1d3 b8c6 e1g1 f8e7 c1e3 d8b6 b2b3 e8g8 b1d2 f7f6 e5f6 d7f6 f3e5 c5d4 c3d4 e7a3 d2f3 c6b4 e3c1 a3c1 d1c1 c8d7 d3b1 a8c8 c1d2 d7e8 f1e1 b4c6 h2h3 e8h5 e5c6 c8c6 f3e5 c6c7');
INSERT INTO evaluations VALUES(50560,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50560,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(50560,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(50560,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(50560,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(50560,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(50560,36,'cp 12','g8f6 b1c3 d8c7 e1g1 b8c6 c1e3 f8e7 f2f4 d7d6 g1h1 e8g8 d1e1 c6d4 e3d4 b7b5 e1g3 c8b7 a2a3 a8d8 e2f3 g8h8 a1d1 h7h6 d1d3 f8g8 e4e5 f6h7 e5d6 e7d6 f3b7 c7b7 f1d1 b5b4 a3b4 d6b4 g3h3');
INSERT INTO evaluations VALUES(43328,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(43328,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(43328,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(43328,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43328,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(43328,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(43328,36,'cp 31','d1d3 d8c7 c2c4 d7d5 e4d5 f6d5 c1d2 d5b4 d3b3 b4c6 d4c6 b8c6 d2e3 c7a5 b1d2 f8c5 e1g1 c6d4 b3d3 d4e2 d3e2 c5e7 d2e4 e8g8 a1d1 a5c7 c4c5 e6e5 a2a3 c8e6 b2b4 a8d8 e4c3 d8d1 f1d1 f8d8');
INSERT INTO evaluations VALUES(50974,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(50974,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(50974,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(50974,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(50974,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(50974,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50974,36,'cp 5','g8f6 a2a4 c7c5 g1f3 c5d4 e3d4 f8e7 e1g1 b8c6 f1e1 e8g8 c1e3 c8d7 d4d5 e6d5 c3d5 f6d5 d1d5 d7e6 d5d8 c6d8 a1c1 a8c8 c4e6 d8e6 g1f1 e7f6 b2b4 b7b5 c1c8 f8c8 a4b5 a6b5 e1c1 c8c1 e3c1');
INSERT INTO evaluations VALUES(51351,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(51351,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(51351,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(51351,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(51351,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(51351,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(51351,36,'cp 122','g1f3 c8b7 b1c3 g8f6 d1c2 g7g6 f1e2 f8g7 c1e3 e8g8 a1d1 b8d7 e1g1 d8e7 f1e1 c7c5 d4d5 a8e8 h2h3 e6e5 c2d2 f6h5 g2g3 b7c8 f3h2 d7f6 g1g2 a7a6 e3g5 e7d7 h2g4 g8h8 e1h1 f6g8 g4e3 h5f6 f2f3 f6h5 d1e1 g7h6 g5h6');
INSERT INTO evaluations VALUES(52076,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(52076,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(52076,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(52076,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(52076,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52076,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(52076,36,'cp 35','e3d4 g8f6 a2a4 b7b6 c2c3 f8e7 f1b5 e8g8 e1g1 c8b7 f1e1 a7a6 b5d3 d7d6 h2h3 f8e8 c1f4 b8d7 b1a3 e6e5 d4e5 d6e5 f4e5 e7a3 e5f6 d8f6 e1e8 a8e8 a1a3 b7f3 d1f3 e8e1 d3f1 f6f3 g2f3 e1b1 g1g2 g7g6 f1a6 b1b2 a6b5 d7c5');
INSERT INTO evaluations VALUES(52063,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52063,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(52063,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(52063,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(52063,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(52063,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(52063,36,'cp 29','g8f6 f1d3 f8e7 e1g1 b7b6 c1f4 e8g8 a2a4 c8a6 d3b5 a6b7 h2h3 a7a6 b5d3 d7d6 f1e1 b8d7 b1a3 f8e8 c2c3 e7f8 f4h2 g7g6 d3f1 f6d5 a3c2 a8c8 f3d2 d8c7 d2c4 b7c6 c4e3 c7b7 e3d5 c6d5 c2e3 d5c6 a1c1 d7f6');
INSERT INTO evaluations VALUES(44298,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(44298,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(44298,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(44298,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(44298,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(44298,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(44298,36,'cp 39','f1d3 f8e7 e1g1 b7b6 c1f4 e8g8 a2a4 f6h5 f4c1 h5f6 b1a3 c8a6 a3b5 a6b7 f1e1 a7a6 b5a3 d8c7 c2c3 d7d6 h2h3 b8d7 c1f4 b7c6 f4g3 c7b7 a3c4 c6f3 d1f3 b7f3 g2f3 d6d5 c4e3 g7g6 h3h4 f8c8 h4h5 e7f8 h5g6 h7g6');
INSERT INTO evaluations VALUES(54317,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54317,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(54317,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54317,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54317,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54317,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54317,36,'cp 0','c2c3 d5d4 g2g3 d4d3 d1b3 b8c6 b3b5 d8c7 f1g2 c7d8');
INSERT INTO evaluations VALUES(1950,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(1950,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(1950,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(1950,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(1950,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(1950,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(1950,36,'cp 94','b1c3 e5f3 g2f3 f8b4 c1d2 g8e7 h2h4 d7d5 e4d5 e8g8 d1e2 c7c6 d5c6 f6c6 e1c1 c8g4 c3e4 a8c8 c2c3 g4h5 h1g1 f7f5 e4g5 f8f6 d2e3 h7h6 d1d4 b4d6 d4c4 c6d7 c4c8 e7c8 g5h3 c8b6 e2d2 g8h7 f1d3 f6g6 g1g6 h5g6 d3c2 g6f7 b2b3 d6a3 c1b1 d7d2 e3d2');
INSERT INTO evaluations VALUES(8355,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(8355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8355,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(8355,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(8355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8355,36,'cp 0','a2a3 e8g8 b2b4 c5d6 c4d5 e6d5 c3b5 d6e7 f1d3 f6e4 c1b2 a7a5 e1g1 a5b4 a3b4 a8a1 d1a1 c6b4 d3e2 e7f6 b2f6 e4f6 f1b1 b4c6 b5d4 f6e4 h2h3 d8c7 e2d3 e4c5 d3e2');
INSERT INTO evaluations VALUES(9514,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(9514,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(9514,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(9514,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(9514,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(9514,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(9514,36,'cp -35','c8e6 d1e2 h7h6 g5e3 d8e7 b1d2 f6d7 a2a3 a7a5 d3d4 e8c8 d4e5 d7e5 f3d4 e5g4 d4e6 e7e6 e1c1 g4e3 e2e3 e6a2 e4e5 d6e7 d2b1 c8b8 g2g3 f7f6 e5e6 f6f5 d1d8 h8d8 e3b3 a2b3 c2b3 b8c8 c1c2 d8e8 f2f4 e7d6 b1d2');
INSERT INTO evaluations VALUES(10186,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(10186,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(10186,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(10186,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(10186,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(10186,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(10186,36,'cp 28','a2a3 d7d5 b2b4 c5e7 c4d5 f6d5 c1b2 a7a6 d2d3 c8e6 b1d2 d8d7 g1f3 f7f6 h2h4 e8c8 d3d4 e5d4 f3d4 c6d4 b2d4 d7b5 d1e2 h8e8 e2b5 a6b5 e1g1 c7c6 f1c1 d5c7 d2e4 e6d5 g2h3 c8b8 e4c5 e7c5 d4c5 d5f3 h3f1 d8d2 a3a4 b5a4 a1a4 e8d8');
INSERT INTO evaluations VALUES(10477,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(10477,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(10477,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(10477,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(10477,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(10477,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(10477,36,'cp -93','b4c3 b2c3 f6e4 d4c6 d7c6 d1d8 e8d8 c3c4 c6c5 c1e3 b7b6 e1c1 e4d6 e3g5 f7f6 g5f4 d8e7 f1d3 c8e6 a3a4 a8d8 h2h4 e7f7 h4h5 g7g5 f4g3 f6f5 g3e5 h8e8 h1e1 e6d7 a4a5 b6a5 g2g4 d7c6 d3f5 d6f5 d1d8 e8d8 g4f5 d8e8 e1e3 e8e7 c1b2 c6d7 f5f6 e7e6 e5c7 e6e3');
INSERT INTO evaluations VALUES(10526,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(10526,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(10526,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(10526,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(10526,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(10526,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(10526,36,'cp -97','b2c3 f6e4 d4c6 d7c6 d1d8 e8d8 c3c4 b7b6 c1e3 c6c5 e1c1 e4d6 e3g5 f7f6 g5f4 d8e7 f1d3 c8e6 d1d2 e7f7 f4d6 c7d6 d3e2 a8d8 h1d1 f7e7 c1b2 g7g5 a3a4 f6f5 a4a5 g5g4 b2c3 d8d7 d1a1 d6d5 c4d5 d7d5 d2d5 e6d5 f2f3 h7h5 a5b6 a7b6 e2d3 e7f6 a1a6 g4f3');
INSERT INTO evaluations VALUES(10525,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(10525,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(10525,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(10525,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(10525,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(10525,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(10525,36,'cp -109','f6e4 d4c6 d7c6 d1d8 e8d8 c3c4 b7b6 c1e3 c6c5 e1c1 e4d6 e3g5 d8e8 g5f4 e8e7 f1d3 c8e6 a3a4 e7d7 a4a5 a8e8 f4d6 c7d6 c1b2 d7c7 d1a1 d6d5 a5b6 a7b6 a1a7 c7c6 h1a1 d5c4 d3e4 e6d5 e4d5 c6d5 a7f7 e8e2 g2g3 d5c6 a1a7 c6b5 f7g7 e2f2 g7h7');
INSERT INTO evaluations VALUES(10703,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(10703,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(10703,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(10703,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(10703,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(10703,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(10703,36,'cp -86','d1f3 d8e7 f1e2 d7d6 d4c6 b7c6 e1g1 f7f5 f1e1 e8g8 e2d3 e7f7 a1b1 f8e8 d3e4 f5e4 e1e4 f7f3 g2f3 g8f7 c3c4 c6c5 c1g5 e8e4 f3e4 c8a6 f2f3 f7e6 g1f2 a6c4 h2h4 e6d7 h4h5 d7c6 h5h6 g7h6 g5h6 c4f7 b1c1 c6b5 c1b1 b5c4');
INSERT INTO evaluations VALUES(11348,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(11348,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(11348,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(11348,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(11348,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(11348,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(11348,36,'cp 20','c8e6 c4e6 f7e6 b2b4 f8e7 c3e2 e8g8 e1g1 f6h5 e2g3 d8e8 c2c3 e8g6 g3h5 g6h5 c1e3 d6d5 a1c1 f8f7 f3h2 h5d1 f1d1 a6a5 h2f3 e7d6 h3h4 a5b4 a3b4 a8a3 g1f1 b7b5 f1e2 f7f8 g2g4 f8a8 h4h5 g8f7 d1d2 a8a4 g4g5');
INSERT INTO evaluations VALUES(15782,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(15782,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(15782,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(15782,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(15782,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(15782,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(15782,36,'cp 34','c7c5 c1e3 d8c7 d1e2 a7a6 e1g1 b7b6 c2c4 c5d4 e3d4 c8b7 d4c3 a8d8 a1d1 f8e7 b2b4 e8g8 f1e1 f6d7 f3e5 d7e5 c3e5 c7c6 e2g4 g7g6 a2a3 a6a5 d3e4 d8d1 g4d1 c6c8 e4b7 c8b7 e5c3 f8d8 d1e2 b7c7 e2e4 a5b4 a3b4');
INSERT INTO evaluations VALUES(14023,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(14023,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(14023,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(14023,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14023,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(14023,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(14023,36,'cp 34','c1e3 d8c7 d1e2 f8e7 e1c1 c8d7 d4c5 e7c5 e3c5 c7c5 f3e5 d7c6 f2f4 e8g8 h1e1 g7g6 g2g4 f6d5 e2d2 d5b4 d3c4 f8d8 d2c3 c6d5 a2a3 b4c6 c4d5 c5c3 b2c3 e6d5 e5c6 b7c6 e1e7 d8e8 e7c7 e8e2 c7c6 e2h2 d1d5 a8e8 g4g5');
INSERT INTO evaluations VALUES(17587,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(17587,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(17587,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(17587,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17587,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(17587,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(17587,36,'cp 28','g8f6 d4c6 b7c6 e1g1 g7g6 c1f4 d8c7 d1d2 f8g7 b2b3 e8g8 f2f3 f6h5 f4e3 c8b7 a1d1 a8d8 c3a4 c6c5 e2c4 e7e6 g2g4 h5f6 e3f4 f6e8 h2h4 b7c6 g1h1 c6b5 h4h5 b5c4 b3c4 c7c6 a4c3 d8d7 f4h6 g7h6 d2h6');
INSERT INTO evaluations VALUES(20273,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(20273,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(20273,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(20273,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(20273,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(20273,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(20273,36,'cp -20','b2b3 g8f6 f1d3 f8d6 e1g1 e8g8 c3b5 d6e7 c2c4 d5c4 b3c4 b7b6 c1b2 c8b7 d4c5 c6b4 c5c6 b4c6 a1c1 a7a6 b5d4 d8c7 d4c6 c7c6 d1e2 f6d7 d3b1 a8c8 f1d1 f8d8 e2c2 f7f5 c2e2 e7f6 b2f6');
INSERT INTO evaluations VALUES(22626,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(22626,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(22626,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(22626,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(22626,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(22626,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22626,36,'cp 0','d7d5 e4e5 d5d4 d1e2 f7f5 c2c4 f8b4 c1d2 a7a5 g2g3 b7b6 f1g2 d8d7 e1g1 g8e7 d2b4 a5b4 f3h4 c8b7 b1d2 e8g8 f2f4 c6d8 h2h3 c7c5 g3g4 b7g2 e2g2 d7c6 g2g3 c6d7');
INSERT INTO evaluations VALUES(29496,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(29496,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(29496,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(29496,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(29496,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(29496,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(29496,36,'cp 141','f1b5 c7c6 d5c6 b8c6 g1f3 d8c7 e1g1 e7e6 b5c6 b7c6 d2d3 h7h6 f3d2 f6g4 g2g3 f8e7 d2c4 e8g8 c1d2 g4f6 d1e2 c7d7 f1e1 f6d5 a2a3 f8d8 g3g4 f5h7 e2f3 f7f6 c3e4 c6c5 d2a5 d8c8');
INSERT INTO evaluations VALUES(29534,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(29534,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(29534,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(29534,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(29534,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(29534,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29534,36,'cp 0','e7e6 d1b3 f8d6 f4d6 c7d6 e2e3 d8e7 c3c4 d5c4 f1c4 e8g8 e1g1 b8c6 c4e2 h7h6 f1c1 f8c8 b1c3 d6d5 b3d1 f5g4 f3e1 g4f5');
INSERT INTO evaluations VALUES(30089,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30089,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30089,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30089,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30089,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30089,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30089,36,'cp 0','e2e3 f8d6 f4d6 c7d6 d1b3 d8e7 c3c4 d5c4 f1c4 e8g8 c4e2 b8c6 b1c3 d6d5 e1g1 h7h6 f1c1 f8c8 b3d1 f5g4 f3e1 g4f5');
INSERT INTO evaluations VALUES(30790,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30790,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(30790,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(30790,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30790,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(30790,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30790,36,'cp 8','e7e6 g1f3 g8f6 f1d3 d7d5 e1g1 f8d6 b2b3 e8g8 c4d5 e6d5 c1a3 c7c5 b1c3 f8e8 a1c1 b8d7 c3e2 d8e7 a3b2 a7a6 f1e1 a8d8 d1c2 f6e4 d4c5 d7c5 e2d4 c5d3 c2d3 e7d7 c1c2 d8c8 e1c1 c8c2 c1c2 f7f6 d4c6 d7e6');
INSERT INTO evaluations VALUES(30645,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30645,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30645,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30645,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(30645,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(30645,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(30645,36,'cp 1','g1f3 d7d5 c4d5 e6d5 d1a4 c7c6 f1d3 g8f6 e1g1 b8d7 b1d2 f8d6 e3e4 d5e4 d2e4 f6e4 d3e4 d8c7 c1g5 e8g8 a1c1 c6c5 e4b7 c7b7 f1d1 f8e8 g5e3 a8d8 h2h3 d6f8 a4c2 h7h6 c2f5 c5d4 f3d4 d7c5 b2b4 c5e4 f5b5');
INSERT INTO evaluations VALUES(36497,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36497,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36497,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36497,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36497,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36497,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36497,36,'cp 0','c7c5 h2h3 a7a6 a2a4 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(36609,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(36609,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(36609,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(36609,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(36609,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(36609,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(36609,36,'cp -33','e7e5 f4g5 e5d4 c3d4 c7c5 g1e2 b8c6 e1g1 c5d4 e3d4 h7h6 g5h4 f8e8 f2f3 d8c7 h4f2 c6b4 d3b1 c8e6 a2a3 b4d5 d2e4 f6e4 b1e4 e8e7 e4d5 e6d5 e2c3 d5c4 f1e1 e7e1 d1e1 a7a6 e1d2 c4b3 a1c1 a8c8 c1e1 c7c4 c3e4 c4c2 d2c2');
INSERT INTO evaluations VALUES(37428,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(37428,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(37428,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(37428,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(37428,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(37428,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(37428,36,'cp -21','d7d6 g1f3 b8c6 e1g1 f6h5 f4g5 f7f6 g5h4 e7e5 b1d2 d8e8 e3e4 g8h8 f1e1 a7a5 a2a4 g7h6 h4g3 e8d7 d3b5 d7g7 d2c4 c8g4 h2h3 g4f3 d1f3 h5g3 b5c6 b7c6 f3g3 h6f4 g3d3 f6f5 g1h1 g7e7 c4e3');
INSERT INTO evaluations VALUES(38290,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(38290,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38290,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(38290,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(38290,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(38290,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(38290,36,'cp 30','g8f6 c2c4 e8g8 b1c3 b7b6 f1d3 c7c5 e1g1 b8c6 c4d5 f6d5 c3d5 e6d5 b2b3 c6b4 d3e2 c8a6 e2a6 b4a6 f3e5 c5d4 e5c6 d8e8 c6d4 e8d7 f4e5 a6c5 h2h3 a8c8 d4e2 c5e6 e5d4 e7c5 a1b1 c5d4 e2d4 e6d4 d1d4');
INSERT INTO evaluations VALUES(40759,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(40759,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(40759,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(40759,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(40759,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(40759,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(40759,36,'cp -33','e8g8 a2a3 h7h6 g5f3 c7c6 c4a2 a7a5 e1g1 b7b5 h2h3 f8e8 f1e1 b8d7 c1d2 c8b7 b1c3 e7f8 f3h2 b5b4 c3e2 d6d5 e2g3 b7c8 a3b4 a5b4 h2f3 d8c7 d2e3 c8b7 c2c3 c6c5 c3b4 c5b4 e4d5 b7d5 a2d5 f6d5 a1a8');
INSERT INTO evaluations VALUES(36215,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(36215,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(36215,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(36215,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(36215,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(36215,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(36215,36,'cp -26','a2a3 c7c6 c4a2 d6d5 e1g1 d8c7 b1c3 h7h6 g5f3 c8e6 f1e1 b8d7 h2h3 a7a6 d3d4 d5e4 a2e6 e4f3 e6d7 f6d7 d1f3 e7d6 d4e5 d7e5 f3f5 e5g6 c1e3 d6h2 g1h1 h2e5 a1d1 e5c3 b2c3 a8d8 f5g4 d8d1 e1d1 f8e8');
INSERT INTO evaluations VALUES(44322,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(44322,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(44322,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(44322,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(44322,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(44322,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(44322,36,'cp 24','b1c3 b8c6 f2f4 d7d6 f1e2 f8e7 g1f3 e5f4 c1f4 e8g8 e1g1 f6g4 c3d5 g4e5 f3e5 d6e5 f4e3 c8e6 e2g4 b7b6 g1h1 c6d4 e3d4 c5d4 d1f3 g7g6 g4e6 f7e6 d5e7 d8e7 f3g3 f8f1 a1f1 a8f8 f1f8 e7f8 h2h4 f8h6 b2b4 g8f7 g3g5 h6g5 h4g5 f7e7 b4b5 e7d7 h1g1 d7e7 g1f2 e7d6 f2e2 d6d7 e2d1');
INSERT INTO evaluations VALUES(44320,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(44320,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(44320,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44320,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(44320,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(44320,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(44320,36,'cp 30','b8c6 f2f4 d7d6 f1e2 g7g6 g1f3 f8g7 e1g1 e8g8 f4e5 d6e5 c3d5 d8d6 c1g5 f6h5 d1d2 c8g4 g1h1 g4f3 f1f3 h5f4 a1f1 c6d4 f3f2 f4e2 g5e7 d6d7 e7f8 a8f8 f2e2 d4e2 d2e2 f7f5 g2g3 f5f4 f1f2 b7b6 g3f4 e5f4 f2f4 f8f4 d5f4');
INSERT INTO evaluations VALUES(12543,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(12543,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(12543,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(12543,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(12543,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12543,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(12543,36,'cp 26','f2f4 d7d6 f1e2 e5f4 c1f4 f8e7 g1f3 e8g8 e1g1 f6g4 c3d5 g4e5 f3e5 d6e5 f4e3 c8e6 e2g4 b7b6 g4f5 c6d4 e3d4 c5d4 f5e6 f7e6 d5e7 d8e7 d1h5 e7d6 h2h4 h7h6 b2b3 f8f1 a1f1 a8f8 f1f8 g8f8 g1h2 d6d8 h2g3 d8f6 h5f3 f6f3 g3f3 a7a5 a2a3 f8e7 f3g4 g7g6 h4h5 g6g5 b3b4 e7d8');
INSERT INTO evaluations VALUES(48073,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(48073,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(48073,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(48073,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(48073,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48073,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48073,36,'cp -26','e6d5 g1f3 c7c6 a2a3 f8d6 c1g5 d8a5 g5d2 a5b6 d1c2 e8g8 e2e3 b8d7 f1d3 d7f6 e1g1 f8e8 f3e5 b6d8 f2f4 f6e4 d3e4 d5e4 c2b3 e8e7 d2b4 c8e6 b3c3 d6b4 c3b4 f7f6 e5c4 b7b6 a1c1 e7d7 b4c3 e6d5 c3c2 g8h8 h2h3');
INSERT INTO evaluations VALUES(49161,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(49161,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(49161,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(49161,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(49161,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(49161,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(49161,36,'cp 59','c7c5 f1d3 b8c6 g1f3 c5d4 c3d4 d8b6 e1g1 a7a5 f1e1 f8e7 h2h3 a5a4 a1b1 c6d4 f3d4 b6d4 d2f3 d4b4 a2a3 b4b6 d3c2 d7c5 d1d4 c8d7 d4g4 g7g6 c1h6 d7b5 g4d4 b5c4 e1c1 b6a6 f3g5 c5d3 c2d3 c4d3');
INSERT INTO evaluations VALUES(56151,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56151,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56151,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56151,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(56151,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56151,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56151,36,'cp 14','d8d5 d2d4 e4d3 e5d3 b8c6 b1c3 d5a5 f1e2 c8f5 e1g1 e8c8 c1d2 g8f6 f1e1 h7h5 h2h3 f5g6 a2a4 a5f5 c3b5 a7a6 e2f3 f8c5 d3c5 f5c5 f3c6 b7c6 b5c3 c5c4 a1c1 c4d4 d2e3 d4d1 c3d1 f6d5 e3d4 f7f6 d4c5 h8e8 e1e8 d8e8 g1f1 c8b7 d1e3 e8e4 b2b3 e4e8 e3d5 c6d5 f2f3 g6h7');
INSERT INTO evaluations VALUES(34920,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(34920,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(34920,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(34920,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(34920,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(34920,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(34920,36,'cp 23','d2d4 e4d3 e5d3 b8c6 b1c3 d5a5 f1e2 c8f5 e1g1 e8c8 c1d2 c6d4 f1e1 d4e2 d1e2 f5d3 c2d3 g8f6 c3e4 a5f5 e2e3 c8b8 e3g5 f5g5 d2g5 h7h6 g5f6 g7f6 e4g3 f8d6 a1d1 d6g3 f2g3 b8c8 e1e7 d8d7 d1e1 h8d8 e7d7 c8d7');
INSERT INTO evaluations VALUES(56598,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(56598,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(56598,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56598,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(56598,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(56598,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(56598,36,'cp 0','f1c4 b8c6 d2d3 c6a5 b1d2 f8d6 g1e2 g8e7 e1g1 e8g8 f2f4 e5f4 e2f4 d6f4 f1f4 a5c4 d2c4 e7g6 f4f2 c8e6 b2b3 d8d7 a2a4 b7b6 d1f3 f7f5 e4f5 g6h4 f3e4 h4f5 c4e5 d7d6 e5c4 d6d7');
INSERT INTO evaluations VALUES(1234,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(1234,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(1234,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(1234,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(1234,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(1234,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(1234,36,'cp 34','b5a4 f8c5 e1g1 d7d6 b1c3 g8e7 c1g5 f6g6 g5e7 e8e7 c3d5 e7d8 a4c6 b7c6 b2b4 c5a7 d5e3 c8h3 d1e2 h7h5 c2c4 h3g4 c4c5 g6f6 c5d6 c7d6 e2c2 g4d7 a1b1 d8e8 g1h1 g7g5 a2a4 g5g4');
INSERT INTO evaluations VALUES(1238,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(1238,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(1238,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(1238,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(1238,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(1238,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(1238,36,'cp 37','f8c5 c2c3 b7b5 a4c2 d7d6 b2b4 c5a7 a2a4 g8e7 c1g5 f6g6 c2b3 e8g8 e1g1 b5a4 a1a4 g8h8 g5e3 f7f5 f3h4 g6f6 h4f5 e7f5 e4f5 c6e7 e3a7 a8a7 b1d2 f6f5 d2e4 c8d7 a4a1 f5g6 g1h1 g6h6 d1e2');
INSERT INTO evaluations VALUES(3495,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(3495,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(3495,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(3495,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(3495,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(3495,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(3495,36,'cp 88','d4d5 c6b4 f1e2 f8e8 e1g1 e7e6 a2a3 b4a6 d5e6 e8e6 e4e5 f6d7 c1g5 d8f8 e5d6 e6d6 d1e1 d7b6 b2b4 c7c5 b4b5 a6c7 g5f4 c8f5 f4d6 f8d6 g2g4 f5e6 a1c1 b6c4 c3e4 d6f4 f3g5 c4a3 e2d3 e6g4 h3g4 f4g4 g1h1');
INSERT INTO evaluations VALUES(5377,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(5377,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(5377,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(5377,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(5377,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(5377,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(5377,36,'cp -19','d7d6 f1b5 d6e5 f3e5 g7e5 d4e5 d8d1 e1d1 c8d7 b5c6 d7c6 f2f3 h7h5 d1e1 h5h4 e1f2 h8h5 h1e1 g8h6 b1a3 h6f5 c1d2 f5d4 d2c3 d4e6 a3c2 a8c8 a1d1 c6a4 d1d2 a4c2 d2c2 b7b6 c2d2 e6c7 e1e4 c7b5 e5e6 f7f6 d2d7 b5c3 b2c3 c8c3 d7a7 e8f8 e4d4 c3c8 f2e1');
INSERT INTO evaluations VALUES(5518,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(5518,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(5518,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(5518,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(5518,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(5518,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(5518,36,'cp -29','f1b5 d6e5 f3e5 g7e5 b5c6 b7c6 d4e5 d8a5 b1c3 a5e5 c1e3 g8h6 d1f3 e8g8 e1g1 h6f5 f1e1 e5c7 e3c5 c8e6 c3e4 f8e8 e4g3 e6d5 f3a3 f5d6 e1e2 f7f6 h2h4 h7h5 a3e3 g8g7 b2b3 e7e5 a1d1 e8d8 f2f3 d6f7 e2d2 f6f5 d1e1 f5f4');
INSERT INTO evaluations VALUES(6170,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(6170,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(6170,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(6170,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(6170,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(6170,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(6170,36,'cp 150','h2h4 h7h6 d4d5 c6e5 d5e6 d7e6 c1f4 e5f3 d1f3 e6e5 a1d1 c8d7 f4g3 h6h5 f1c4 g8h6 e1g1 c7c6 f3d3 d8e7 f2f3 d7e6 c4e6 e7e6 g3f2 g7f6 a2a4 e6e7 g2g3 g6g5 h4g5 f6g5 g1g2 h5h4 g3g4 h4h3 g2h2 g5f4 h2h1 e8g8 c3e2 f4g5 d3d6 f8d8 d6e7 g5e7 a4a5');
INSERT INTO evaluations VALUES(7327,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(7327,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7327,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7327,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7327,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7327,36,'cp 0','e8g8 a4c6 d7c6 e1g1 c5d6 h2h3 h7h6 c3e2 f8e8 e2g3 c6c5 b2b3 b7b5 a2a4 c8e6 c1b2 f6d7 d1e2 d8f6 b2c3 e8d8 f1e1 d8e8');
INSERT INTO evaluations VALUES(7611,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(7611,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(7611,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(7611,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(7611,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(7611,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(7611,36,'cp -54','h7h6 g5h4 e8g8 f1d3 d7d5 e1g1 b4c3 b2c3 g7g5 h4g3 d5e4 d3c4 d8e7 f2f4 f6g4 d1e2 g8h7 h2h3 e7c5 g3f2 g4f2 f1f2 c8f5 f4g5 f5g6 h3h4 h6g5 h4h5 g5g4 e2g4 c5h5 g4h5 g6h5 a1f1 h7g7 f2f6 e4e3 f6c6 f7f5 f1e1 f5f4 c6c7 g7h6');
INSERT INTO evaluations VALUES(7571,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(7571,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(7571,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(7571,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(7571,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(7571,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(7571,36,'cp -26','g5h4 e8g8 f1d3 d7d5 e1g1 b4c3 b2c3 g7g5 h4g3 d5e4 d3c4 d8e7 d1d4 f8d8 d4e5 e7e5 g3e5 f6d5 a1d1 c8f5 c4d5 c6d5 e5c7 d8d7 c7e5 f5g4 d1b1 g4e2 f1e1 e2c4 b1b2 f7f5 e1b1 a8e8 e5d4 g8f7 a2a4 f5f4 h2h3 c4e2 b2b4 f7g6 b4b7 d7b7 b1b7');
INSERT INTO evaluations VALUES(7894,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(7894,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(7894,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(7894,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(7894,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(7894,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(7894,36,'cp -42','a7a6 a2a4 h7h6 g5f4 d7d6 h2h3 b7b6 e1g1 c8b7 f4h2 d8d7 f1e1 e8g8 c4a2 f8d8 a2b1 a8c8 c3e2 b6b5 c2c3 b5b4 f3d2 a6a5 d2c4 d7c7 c4e3 e7f8 b1c2 c7b6 h2g3 d6d5 e4e5 f6d7 d3d4 c5d4 c3d4 b7a6');
INSERT INTO evaluations VALUES(9201,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(9201,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(9201,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(9201,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(9201,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(9201,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(9201,36,'cp -40','d2d3 e5e4 f3d2 e4e3 f2e3 c6e5 b1c3 c7c6 d2f3 e5g4 d1d2 d8e7 e3e4 d5e4 d3e4 e8g8 e4e5 g4e5 e1c1 d6c7 c1b1 f8e8 f3d4 c8d7 h2h3 e5g6 d2e1 c7d6 e2e4 a8d8 h1f1 h7h6 g3g4 d6f4 d4f3 d7e6 d1d8 e8d8');
INSERT INTO evaluations VALUES(12798,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(12798,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(12798,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(12798,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(12798,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(12798,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(12798,36,'cp -35','e7e6 f1e2 d7d5 e4d5 e6d5 c4d5 f6d5 e1g1 f8e7 d3d4 e8g8 d4c5 h7h6 e2c4 c8e6 c4d5 d8d5 d1d5 e6d5 c1e3 f7f5 b1c3 d5f3 g2f3 f5f4 e3d2 c6e5 g1g2 e7c5 c3e4 c5d4 f1d1 e5g6 g2f1 b7b6 d2c3 d4c3 e4c3 a8d8 c3e4 g6h4 d1d8 f8d8');
INSERT INTO evaluations VALUES(13657,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(13657,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(13657,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(13657,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(13657,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(13657,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(13657,36,'cp 36','e2e3 f8d6 f4g3 h7h6 f1d3 d6g3 h2g3 e8g8 e1g1 c8d7 a1c1 d8e7 a2a3 f8d8 c1c2 a7a6 d1e2 e6e5 d4e5 c6e5 e2d2 d7g4 f3d4 a8c8 f1c1 b7b5 c3e2 f6e4 d3e4 d5e4 c2c8 d8c8 c1c8 g4c8 d2c2 e5c4 a3a4 c8d7 a4b5 d7b5 e2c3 c4d6');
INSERT INTO evaluations VALUES(14262,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(14262,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(14262,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(14262,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(14262,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(14262,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(14262,36,'cp -96','e7e5 f1b5 f8d6 g1e2 e8g8 e1g1 c6e7 d4e5 d6e5 f3f4 e5d6 e2g3 c8g4 b5e2 g4e2 c3e2 d6c5 e2d4 e7c8 b2b3 c8d6 c1b2 d8d7 d1d3 f8e8 a1e1 f6e4 d4f3 e4g3 h2g3 d7e6 b2d4 c5b6 f3g5 e6f5 g3g4 f5d3 c2d3 b6d4 e3d4 d6b5 g5f3');
INSERT INTO evaluations VALUES(15109,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(15109,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(15109,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(15109,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(15109,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(15109,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(15109,36,'cp -113','f1b5 f8d6 g1e2 e8g8 e1g1 c6e7 d4e5 d6e5 f3f4 e5d6 e2g3 c8g4 b5e2 g4e2 c3e2 d6c5 e2d4 e7c8 b2b3 c8d6 c1b2 d8d7 d1d3 f8e8 a1e1 f6e4 d4f3 e4g3 h2g3 d7e6 b2d4 c5b6 d4b6 a7b6 a2a4 h7h6 f3d4 e6d7 d4b5 d6e4 b5c3 e4c3 d3c3 e8e4 g1h1 a8e8 h1h2');
INSERT INTO evaluations VALUES(15898,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(15898,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(15898,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(15898,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(15898,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(15898,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(15898,36,'cp 58','g7g6 f1d3 f8g7 g1e2 c7c6 f2f3 e8g8 c1e3 d7c5 d3c2 c6d5 c4d5 f6h5 e1g1 a7a5 d1d2 c8d7 a2a4 a8c8 c3b5 d7b5 a4b5 b7b6 e2c3 g8h8 g1h1 d8d7 f1g1 h8g8 g2g3 c8c7 d2g2 g7f6 c2d1 f8c8 g2d2 d7h3 g1g2');
INSERT INTO evaluations VALUES(18275,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(18275,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(18275,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(18275,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(18275,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(18275,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(18275,36,'cp 44','g8f6 b1c3 f8e7 e1g1 e8g8 h2h3 a7a5 b2b3 f6d7 c1b2 d7c5 f2f4 a5a4 a1b1 a4b3 c2b3 c8e6 c4e6 c5e6 a2a4 e7f6 g1h2 a8b8 c3e2 f6b2 b1b2 b8b4 e2g3 e6d4 b2d2 c6c5 d2d3 f8e8 g3h5 e8e4 d3g3 d4f5');
INSERT INTO evaluations VALUES(19520,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(19520,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(19520,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(19520,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(19520,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(19520,35,'cp -111',NULL);
INSERT INTO evaluations VALUES(19520,36,'cp -131','c3b2 c1b2 a7a6 e1g1 e7e6 b1c3 b7b5 c4b3 c8b7 a2a4 b5b4 c3d5 g8f6 a1c1 f8e7 d5e7 c6e7 b2f6 g7f6 d1d4 e7c6 d4d6 a8c8 f1d1 h8g8 a4a5 g8g4 f3d2 d8e7 d2c4 c6e5 c4e5 e7d6 c1c8 b7c8 d1d6 f6e5 f2f3 g4f4 d6b6 e8d8 g1f2 d8e7 b6b4');
INSERT INTO evaluations VALUES(19523,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(19523,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(19523,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(19523,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(19523,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(19523,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(19523,36,'cp -112','c1b2 a7a6 e1g1 e7e6 b1c3 b7b5 c4d3 c8b7 a2a4 b5a4 e4e5 a4a3 b2a3 f8a3 a1a3 d8c7 f1e1 g8e7 d3b1 c6a5 f3d2 a5c4 d2c4 c7c4 d1d6 a8c8 b1d3 e7f5 d3c4 f5d6 c4a6 b7a6 e5d6 a6b7 c3a4 e8g8 a4b6 c8c6 b6d7 f8d8 d7e5 c6d6 f2f3 f7f6 e5g4 b7c6 a3a7 d6d1 e1d1');
INSERT INTO evaluations VALUES(19522,30,'cp -134',NULL);
INSERT INTO evaluations VALUES(19522,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(19522,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(19522,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(19522,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(19522,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(19522,36,'cp -136','e7e6 e1g1 g8f6 b1c3 d7d6 d1e2 a7a6 a1d1 c8d7 c3d5 e6d5 e4d5 c6e7 f1e1 d7f5 c4b3 a8c8 b3a4 f5d7 a4b3 d7g4 h2h3 g4f3 e2f3 h7h5 b2d4 f6d7 e1e2 h8h6 d1e1 b7b5 a2a4 b5a4 b3a4 c8c4 d4g7 c4a4 g7h6 f8h6 e2e7 d8e7 e1e7 e8e7 f3e2 e7d8 e2h5 a4a1 g1h2 h6f4 g2g3');
INSERT INTO evaluations VALUES(18883,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(18883,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(18883,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(18883,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(18883,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(18883,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(18883,36,'cp -119','e1g1 g8f6 b1d2 d7d6 d1e2 f8e7 e4e5 d6e5 f3e5 e8g8 d2f3 f6d5 a1d1 e7f6 e2e4 d8e7 d1d3 c6b4 d3d2 b7b5 c4b5 c8b7 a2a3 d5c3 e4e1 c3b5 d2d7 e7e8 d7b7 b4d5 a3a4 b5d6 b7d7 d6f5 b2a3 f6e7 e1a1 e7a3 a1a3 a8d8 d7d8 e8d8');
INSERT INTO evaluations VALUES(18882,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(18882,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(18882,32,'cp -119',NULL);
INSERT INTO evaluations VALUES(18882,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(18882,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(18882,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(18882,36,'cp -137','a7a6 c4b3 b7b5 b1c3 c6a5 f1e1 g8e7 a2a4 b5b4 c3e2 c8b7 e2g3 e7g6 b3a2 b4b3 a2b3 d8b6 b3d5 b7d5 b2d4 b6b3 e4d5 b3d5 d1e2 a5b3 a1d1 b3d4 d1d4 d5a5 e1d1 f8c5 d4d7 e8g8 h2h4 a5a4 h4h5 g6f4 e2e4 a4e4 g3e4 c5b6 h5h6 f4d5 h6g7 g8g7');
INSERT INTO evaluations VALUES(20442,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(20442,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(20442,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(20442,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(20442,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(20442,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(20442,36,'cp -66','g8e7 f1d3 e7g6 e1g1 c8d7 d3g6 h7g6 a4g4 d8c7 f1e1 c6b4 b1a3 b4c2 a3c2 c7c2 f3d4 c2a4 b2b3 a4a5 c1g5 a8c8 h2h4 f8c5 e1e2 c5b6 a1d1 a5a3 g2g3 a3b4 g1g2 a7a5 e2d2 b4a3 g4f4 e8g8 g3g4 a5a4');
INSERT INTO evaluations VALUES(26704,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(26704,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(26704,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(26704,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(26704,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(26704,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(26704,36,'cp 44','f1d3 b4d3 c2d3 c7c6 e3e4 f5g4 c1e3 e7e6 h2h3 g4f3 d1f3 f8d6 f3g4 d6f8 e1g1 g7g6 g4d1 g8e7 c3e2 f8g7 d1d2 e8g8 b2b4 e7c8 e4e5 f7f6 f2f4 f8f7 a2a4 c8e7 a4a5 d8d7 g2g4 f6f5 g1h1 f5g4 h3g4 h7h5 g4h5 e7f5 e3f2 g6h5 f1g1');
INSERT INTO evaluations VALUES(26705,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(26705,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(26705,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(26705,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(26705,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(26705,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(26705,36,'cp 38','b4d3 c2d3 c7c6 e3e4 f5g4 e1g1 e7e6 h2h3 g4f3 d1f3 f8d6 f3g4 d6f8 c1e3 g7g6 a2a4 a7a5 a1b1 h7h5 g4f3 f8e7 c3e2 g8f6 b1c1 f6d7 e3d2 d7b6 b2b3 b6d7 f3e3 d7f8 e4e5 d8d7 g1h1 g6g5 f2f4');
INSERT INTO evaluations VALUES(27411,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(27411,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(27411,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(27411,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(27411,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(27411,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(27411,36,'cp 50','h2h3 g4f3 e2f3 e8c8 f3c6 b7c6 d1f3 e7e5 f3f5 f6d7 e1g1 g7g6 f5e4 e5d4 e4d4 f8b4 d4c4 b4c3 b2c3 d7e5 c4b3 a5b5 f1e1 e5c4 c1g5 d8e8 a2a4 b5b3 c2b3 c4d6 g5e3 a7a5 a1d1 d6f5 e3c5 e8e1 d1e1 h8d8 g2g4 f5g7 e1e7 g7e6 c5e3 d8d7 e7d7 c8d7 g1f1 c6c5 f1e2 d7c6 e2d3 c6d5 g4g5 e6g7 c3c4 d5c6');
INSERT INTO evaluations VALUES(28656,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(28656,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(28656,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(28656,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(28656,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(28656,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(28656,36,'cp 58','c4d5 f8b4 b1c3 b7d5 e1g1 b4c3 b2c3 e8g8 f1e1 b8d7 g2f1 f6e4 f3d2 e4d2 c1d2 d7f6 a1c1 d5c4 f1g2 a8c8 d2g5 c7c5 d1a4 c4d5 g2d5 e6d5 d4c5 c8c5 a4a7 c5a5 a7b7 a5a2 g5f6 d8f6 b7d5 a2a5 d5d2 h7h6');
INSERT INTO evaluations VALUES(28649,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(28649,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(28649,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(28649,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28649,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(28649,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(28649,36,'cp 50','f8b4 c1d2 b4d2 f3d2 b7d5 d2f3 e8g8 e1g1 b8d7 b1c3 d5b7 f1e1 f6e4 a1c1 a7a5 c3b5 d7f6 d1c2 c7c6 b5c3 c6c5 e1d1 e4c3 c2c3 c5d4 c3d4 d8d4 f3d4 b7g2 g1g2 f8d8 d4c6 d8d1 c1d1 g8f8 d1d6 f6d5 e2e4 d5f6 e4e5 f6d5 d6d7');
INSERT INTO evaluations VALUES(29605,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(29605,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(29605,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29605,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29605,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(29605,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(29605,36,'cp 26','e7e6 c2c4 c7c6 d1b3 d8b6 d2d3 b8d7 c4d5 e6d5 b3c2 b6c5 c2a4 f8e7 b1c3 e8g8 c1e3 c5d6 e3f4 d6c5 a1c1 f5g4 h2h3 g4f3 g2f3 c5b6 a4c2 d5d4 c3b1 d7c5 f4e5 c5d7 e5f6 d7f6 b1d2 a7a5 f3g2 a8e8 a2a3 f6d5 g2d5 c6d5');
INSERT INTO evaluations VALUES(30141,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(30141,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30141,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(30141,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(30141,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(30141,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(30141,36,'cp 37','c2c4 c7c6 d1b3 d8b6 d2d3 b8d7 c4d5 e6d5 b3c2 b6c5 c2a4 f8e7 b1c3 e8g8 c1e3 c5d6 e3f4 d6c5 a1c1 c5b6 a4c2 h7h6 e2e4 d5e4 d3e4 f5h7 f1d1 a8d8 f3d2 d7c5 d2c4 b6b4 c2e2 d8d1 c1d1 f8d8 d1d8 e7d8 f4d6 d8b6 d6e7 f6d7 c4b6 b4b6 e2d2 a7a5');
INSERT INTO evaluations VALUES(43590,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43590,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(43590,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(43590,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(43590,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43590,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(43590,36,'cp 0','c8g4 g1f3 c7c5 b1d2 e7e6 c2c3 f8e7 h2h3 g4h5 e1g1 e8g8 d1e2 b8c6 f1e1 f8e8 a1d1 d8b6 d1b1 b6d8');
INSERT INTO evaluations VALUES(47850,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(47850,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(47850,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(47850,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47850,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(47850,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(47850,36,'cp 44','b1d2 e5d4 c3d4 d8e7 f1d3 f6e4 d2e4 d6d5 c1g5 e7e6 f2f3 f8e7 d1d2 b8c6 g1e2 d5e4 d3e4 e7g5 d2g5 e8g8 e1g1 e6e7 g5c5 e7c5 d4c5 c8e6 f1e1 a8d8 e4c6 b7c6 e2f4 f8e8 g1f2 a7a5 f4e6 f7e6 e1e2 d8d5 a1e1 g8f7 b2b3 d5c5');
INSERT INTO evaluations VALUES(48326,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(48326,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(48326,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(48326,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(48326,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(48326,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(48326,36,'cp -43','c1f4 c7c5 e2e3 a7a6 g1f3 b7b5 f1e2 c8b7 f3e5 b8d7 e1g1 a8c8 c3b1 c5c4 e5d7 d8d7 a2a3 f8e7 c2c3 e8g8 b1d2 a6a5 e2f3 b7c6 f4e5 b5b4 d1e2 d7b7 e5f6 e7f6 e3e4 b4b3 e4e5 f6g5 f1e1 g8h8 f3g4 g5e7 g1h1');
INSERT INTO evaluations VALUES(48439,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(48439,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(48439,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(48439,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(48439,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(48439,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(48439,36,'cp 88','f6d5 f1c4 c7c5 g1f3 d5c3 b2c3 c5d4 c3d4 c8d7 e1g1 d7c6 a1b1 f8e7 d4d5 e6d5 c4d3 e8g8 d1c2 h7h6 f3d4 b8d7 c2e2 d7c5 d3f5 f8e8 e2g4 e7f8 c1h6 e8e5 h6g5 d8d6 g5f4 g7g6 d4f3 f8g7 g4g3 a8e8 f3e5 g7e5 f4e5 d6e5 g3e5 e8e5 f5g4 c5d3');
INSERT INTO evaluations VALUES(50846,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50846,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50846,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50846,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50846,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50846,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50846,36,'cp 0','c7c5 c2c3 b8c6 f1e2 c5d4 c3d4 c8f5 g1f3 e7e6 e1g1 f5e4 c1e3 g8e7 f3h4 e7g6 h4f3');
INSERT INTO evaluations VALUES(55374,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(55374,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(55374,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(55374,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(55374,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(55374,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(55374,36,'cp 111','h2h4 h7h6 g1f3 f8g7 c1f4 g8f6 d1c2 e8g8 a1d1 b8c6 c2c1 h6h5 e2e3 c6e7 c3b5 f6e8 f1d3 e8d6 c4c5 d6b5 d3b5 c8d7 b5d3 e7f5 b2b3 d7c6 e1g1 d8e7 f1e1 f8e8 a2a4 f5h4 f3h4 e7h4 f4c7');
INSERT INTO evaluations VALUES(56205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56205,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(56205,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56205,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56205,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(56205,36,'cp 0','g8f6 g1f3 h7h6 f1d3 f8d6 e1g1 e8g8 f1e1 c7c6 c2c3 f8e8 b1d2 b8d7 d1c2 d7f8 h2h3 a7a5 a2a4 c8d7 a1d1 d6c7 d2f1 f6h5 f1h2 h5f6');
INSERT INTO evaluations VALUES(48675,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48675,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48675,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48675,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48675,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48675,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48675,36,'cp 0','g1f3 c7c6 f1d3 f8d6 e1g1 e8g8 h2h3 b8d7 b1d2 f8e8 f1e1 d7f8 c2c3 h7h6 d1c2 c8d7 a2a4 a7a5 a1d1 a8c8 d2f1 f6h5 f1h2 h5f6');
INSERT INTO evaluations VALUES(48676,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48676,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48676,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48676,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48676,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48676,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48676,36,'cp -13','c7c6 f1d3 f8d6 e1g1 e8g8 b1d2 f8e8 e3g5 c8e6 c2c4 h7h6 g5h4 d5c4 d3c4 b8d7 f1e1 e6d5 e1e8 d8e8 d1b3 e8c8 h4f6 d7f6 c4d5 f6d5 d2c4 d6c7 g2g3 d5b6 c4b6 a7b6 a2a3 c8d7 a1e1 a8e8 e1e8 d7e8 h2h3');
INSERT INTO evaluations VALUES(57063,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(57063,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(57063,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(57063,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(57063,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(57063,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(57063,36,'cp -25','d7d5 d2d4 g8f6 g1f3 c7c5 d4c5 f8c5 f1d3 b8c6 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 h2h3 c8e6 a2a3 a7a6 e2g3 h7h6 c1d2 c5a7 c2c4 d4c3 d2c3 f6d7 d1e2 d8f6 a1d1 b7b5 b2b4 a8c8 f1e1 f8e8 c3b2');
INSERT INTO evaluations VALUES(7851,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(7851,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(7851,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(7851,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(7851,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(7851,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(7851,36,'cp 94','a2a3 b4c3 b5c3 d7d6 d1d3 e8g8 c1g5 h7h6 g5h4 c8e6 e1c1 a8c8 d3d6 d8a5 h4f6 f8d8 f6d8 c8d8 d6d8 c6d8 f1b5 g8h7 f2f3 a7a6 b5d7 a5c5 d7e6 d8e6 d1d5 c5f2 d5d2 f2h4 h2h3 a6a5 h1d1 h4g5 c1b1 e6f4 h3h4 g5h4 d2d7 h4f2 d7f7 f2g2 c3d5 f4e6 d5e3 g2h2');
INSERT INTO evaluations VALUES(8085,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(8085,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8085,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(8085,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8085,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(8085,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(8085,36,'cp 13','h2h3 e8g8 e1g1 e7d6 d4c5 d6c5 c3c4 c6b4 b1c3 b4d3 d1d3 b7b6 c4d5 f6d5 c3d5 d8d5 d3d5 e6d5 f1d1 c8e6 a2a3 f8c8 f4e5 f7f6 e5c3 g8f7 d1d2 h7h5 f3d4 c5d4 d2d4 a7a5 f2f3 g7g5 g2g4 h5h4 a1d1 c8c4 d4d2 a8c8 c3d4 c4c6 g1f2');
INSERT INTO evaluations VALUES(11758,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(11758,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(11758,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(11758,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(11758,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(11758,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(11758,36,'cp 53','h7h6 e1g1 g7g6 d3d4 e5d4 f3d4 f8g7 d4c6 e7c6 b1c3 e8g8 c1e3 d7d6 d1d2 g8h7 a1d1 c8b7 b3d5 d8d7 h2h3 a8e8 f1e1 d7c8 f2f3 c6e5 d5b7 c8b7 b2b3 e8d8 c3d5 c7c6 d5b6 f8e8 e3f2 e5d7 b6d7 d8d7 f2d4 d6d5 e4e5 e8d8');
INSERT INTO evaluations VALUES(14695,30,'cp -563',NULL);
INSERT INTO evaluations VALUES(14695,31,'cp -551',NULL);
INSERT INTO evaluations VALUES(14695,32,'cp -563',NULL);
INSERT INTO evaluations VALUES(14695,33,'cp -578',NULL);
INSERT INTO evaluations VALUES(14695,34,'cp -578',NULL);
INSERT INTO evaluations VALUES(14695,35,'cp -563',NULL);
INSERT INTO evaluations VALUES(14695,36,'cp -587','b7c6 f3e5 d8g5 d2d4 g5g2 h1f1 f7f6 e5c4 c8g4 c4e3 g2f1 e3f1 g4d1 e1d1 f6f5 f1g3 g7g6 g3f1 e8c8 d1e2 d8d4 b2b3 f8g7 c1b2 d4d7 b2g7 d7g7 a1d1 g7d7 d1d7 c8d7 f1e3 f5f4 e3c4 h8d8 c4d2 c6c5 e2e1 d8e8 d2c4');
INSERT INTO evaluations VALUES(11848,30,'cp -575',NULL);
INSERT INTO evaluations VALUES(11848,31,'cp -584',NULL);
INSERT INTO evaluations VALUES(11848,32,'cp -578',NULL);
INSERT INTO evaluations VALUES(11848,33,'cp -574',NULL);
INSERT INTO evaluations VALUES(11848,34,'cp -594',NULL);
INSERT INTO evaluations VALUES(11848,35,'cp -588',NULL);
INSERT INTO evaluations VALUES(11848,36,'cp -575','f3e5 d8g5 d2d4 g5g2 h1f1 f7f6 e5c4 c8g4 c4e3 g2f1 e3f1 g4d1 e1d1 f6f5 d1e2 f8d6 f1e3 e8g8 e3c4 d6h2 b2b3 f8f6 c1b2 f6g6 a1h1 h2d6 c4e5 d6e5 d4e5 a8e8 e5e6 e4e3 h1d1 e3f2 e6e7 e8e7 e2f2 g6e6 f2f3 g8f7');
INSERT INTO evaluations VALUES(11847,30,'cp -599',NULL);
INSERT INTO evaluations VALUES(11847,31,'cp -580',NULL);
INSERT INTO evaluations VALUES(11847,32,'cp -590',NULL);
INSERT INTO evaluations VALUES(11847,33,'cp -590',NULL);
INSERT INTO evaluations VALUES(11847,34,'cp -575',NULL);
INSERT INTO evaluations VALUES(11847,35,'cp -572',NULL);
INSERT INTO evaluations VALUES(11847,36,'cp -594','d8g5 d2d4 g5g2 h1f1 f7f6 e5c4 c8g4 c4e3 g2f1 e3f1 g4d1 e1d1 f6f5 d1e2 f8d6 f1e3 e8g8 c1d2 d6h2 e3c4 h2d6 a2a4 f5f4 b2b4 a8e8 c4d6 c7d6 c2c4 e8c8 a1b1 h7h5 a4a5 h5h4 b4b5 e4e3 d2b4 c6b5 b4d6 f4f3 e2e3 f8d8 c4b5 d8d6 e3f3');
INSERT INTO evaluations VALUES(684,30,'cp -587',NULL);
INSERT INTO evaluations VALUES(684,31,'cp -594',NULL);
INSERT INTO evaluations VALUES(684,32,'cp -559',NULL);
INSERT INTO evaluations VALUES(684,33,'cp -591',NULL);
INSERT INTO evaluations VALUES(684,34,'cp -578',NULL);
INSERT INTO evaluations VALUES(684,35,'cp -590',NULL);
INSERT INTO evaluations VALUES(684,36,'cp -594','d2d4 g5g2 h1f1 f7f6 e5c4 c8g4 c4e3 g2f1 e3f1 g4d1 e1d1 f6f5 f1g3 g7g6 d1e2 e8f7 g3f1 h7h6 c1d2 f8d6 c2c4 a8d8 d2c3 h8e8 f1d2 g6g5 a1e1 f5f4 e2f1 e4e3 f2e3 e8e3 e1e3 f4e3 d2f3 e3e2 f1e1');
INSERT INTO evaluations VALUES(17218,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(17218,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(17218,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(17218,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(17218,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(17218,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(17218,36,'cp -10','c6d4 b5e2 d7d5 e4e5 f6d7 f2f4 d7b8 g1f3 b8c6 f3d4 c6d4 e2d3 c7c5 c3e2 c8f5 e2d4 f5d3 d4e6 d8b6 e6c5 b6c5 c2d3 c5b4 d1b3 b4b3 a2b3 a7a5 d3d4 e7e6 d2d3 f8e7 c1d2 e8d7 e1e2 h8c8 h1c1 c8c1 a1c1 h7h5 g2g3');
INSERT INTO evaluations VALUES(25044,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(25044,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(25044,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(25044,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(25044,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(25044,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(25044,36,'cp -8','e7e6 f1d3 f8b4 b5c3 e8g8 d3f5 e6f5 d1d3 g7g6 f4g5 c7c6 g1f3 d8d6 a2a3 b4c3 d3c3 f6e4 c3b3 e4g5 f3g5 a8b8 h2h4 h7h5 g5f3 g8g7 e1g1 a6c7 f1c1 f8e8 b3b4 c7b5 c2c4 d6b4 a3b4 d5c4');
INSERT INTO evaluations VALUES(25412,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(25412,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25412,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(25412,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(25412,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25412,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25412,36,'cp 0','f1d3 f8b4 b5c3 f5d3 c2d3 c7c5 g1e2 c5d4 e2d4 e8g8 e1g1 d8d7 f4e5 b4e7 a1c1 a8c8 a2a3 a6b8 b2b4 h7h6 d1b3 b8c6 d4c6 d7c6 c3e2 c6b5 e2d4 b5a6 d4f3 a6b5');
INSERT INTO evaluations VALUES(27543,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(27543,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(27543,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(27543,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(27543,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(27543,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(27543,36,'cp 66','g8e7 b1c3 e7f5 c3a4 b6a5 c1d2 a5d8 d2c3 a8c8 e2d3 f8e7 d3f5 e6f5 a4c5 b7b6 c5d7 d8d7 a1c1 e8g8 c3d2 c6a5 h2h4 c8c1 d2c1 a5c6 g2g3 f8c8 d1e2 c6d8 e2d3 c8c4 h4h5 h7h6 c1d2 c4c8 f1e1 c8c4 b2b3');
INSERT INTO evaluations VALUES(27374,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(27374,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(27374,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(27374,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(27374,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(27374,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(27374,36,'cp 55','b1c3 e7f5 c3a4 b6a5 c1d2 a5d8 d2c3 a8c8 e2d3 b7b5 d3f5 b5a4 f5d3 a4a3 b2a3 f8a3 d1d2 a7a5 f1b1 e8g8 h2h4 d8c7 b1b3 a3e7 a1b1 c6b4 a2a3 b4d3 c3a5 c7a7 b3d3 d7a4 a5b4 c8c2 d2f4 f8c8 d3d2 c2c4 b4e7');
INSERT INTO evaluations VALUES(27372,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(27372,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(27372,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(27372,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(27372,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(27372,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(27372,36,'cp 63','e7f5 c3a4 b6d8 c1g5 f8e7 g5e7 d8e7 a1c1 a8c8 d1d2 e8g8 g2g4 f5h4 f3h4 e7h4 a4c5 c8c7 f2f4 f7f6 g1g2 f6e5 f4e5 f8f1 c1f1 h4h6 c5d7 c7d7 d2h6 g7h6 e2b5 d7g7 b5c6 b7c6 g2h3 g7b7 b2b3 b7b4 f1f6 b4d4 f6e6 d4d3 h3g2 d3d2 g2g3 d2a2 e6h6 a2e2 g3f4 e2f2 f4e3 f2a2 e3d3 a2g2 h6c6 g2g4');
INSERT INTO evaluations VALUES(27365,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(27365,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(27365,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(27365,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(27365,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(27365,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(27365,36,'cp 55','c3a4 b6d8 c1g5 f8e7 g5e7 d8e7 a1c1 a8c8 d1d2 e8g8 g2g4 f5h4 f3h4 e7h4 a4c5 c8c7 f2f4 f7f6 g1g2 f6e5 f4e5 f8f1 c1f1 h4h6 c5d7 c7d7 d2h6 g7h6 e2b5 d7f7 f1f7 g8f7 b5c6 b7c6 h2h4 a7a5 g2g3 f7g6 h4h5 g6g7 g3f4 g7f7 f4f3 f7e8 b2b3 e8f8 f3f4 f8f7 f4e3 f7e7 e3d3 e7d7 d3d2 d7c7');
INSERT INTO evaluations VALUES(29751,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(29751,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(29751,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(29751,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(29751,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(29751,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(29751,36,'cp 75','g1f3 e7e6 f1c4 c7c5 d1a4 b8d7 f3e5 c5d4 e3d4 a7a6 e5g4 f6g4 c4e2 b7b5 a4b3 g4f6 e2f3 a8b8 c1f4 b8b6 e1g1 f8d6 f4e3 e8g8 g2g3 h7h6 a1c1 b6b8 c3e4 f6e4 f3e4 d7f6 e4f3 d8d7 c1c6 a6a5 f1c1 a5a4 b3d1 f6d5 f3d5 e6d5 d1f3 d7e6 b2b3 a4b3 a2b3 b5b4');
INSERT INTO evaluations VALUES(29752,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(29752,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(29752,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(29752,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(29752,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(29752,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(29752,36,'cp 48','e7e6 f1c4 a7a6 e1g1 f8e7 d1e2 c7c5 h2h3 g4h5 f1d1 d8c7 d4d5 e6d5 c4d5 b8c6 g2g4 h5g6 f3h4 a8d8 h4g6 h7g6 d5g2 d8d1 c3d1 e8g8 b2b3 f8d8 d1c3 b7b5 c1b2 c6b4 a2a4 b5a4 c3a4 f6h7 a4c3 e7h4 c3e4 b4d3 b2c3 c5c4 a1d1 h7f6 c3f6 h4f6 b3c4 c7c4');
INSERT INTO evaluations VALUES(30211,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(30211,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(30211,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(30211,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(30211,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(30211,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(30211,36,'cp 55','f1c4 a7a6 e1g1 f8e7 h2h3 g4h5 d1e2 c7c5 f1d1 d8c7 d4d5 e6d5 c4d5 b8c6 g2g4 h5g6 f3h4 a8d8 h4g6 h7g6 d5g2 d8d1 e2d1 e8g8 d1e2 c7e5 c1d2 f8d8 a1d1 e5e6 b2b3 g6g5 d2e1 b7b5 d1d8 c6d8 e2f3 e6d7 c3d5 f6d5 f3d5');
INSERT INTO evaluations VALUES(31118,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31118,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(31118,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(31118,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31118,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31118,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31118,36,'cp 10','b8d7 c2c4 e7e6 d1b3 d8b6 e1g1 g8f6 h2h3 g4h5 b3c2 f8e7 b2b3 h5g6 e2d3 c6c5 b1c3 c5d4 e3d4 d5c4 b3c4 e8g8 d3g6 h7g6 a1b1 b6c7 c2b3 f8c8 f3d2 e6e5 b3b7 c7b7 b1b7 e5d4 c3b5 e7c5 f1d1 d7e5 d2b3 e5c4 b3c5 c8c5');
INSERT INTO evaluations VALUES(31921,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(31921,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(31921,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(31921,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(31921,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(31921,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(31921,36,'cp 60','d1f3 c7c6 c2c4 e7e6 b1c3 g8f6 f1d3 g7g6 e1g1 f8g7 f1d1 e8g8 e3e4 d5e4 d3e4 b8d7 e4c2 f8e8 a1b1 d8e7 c1g5 h7h6 g5h4 e7b4 b2b3 b4a5 g1h1 e6e5 d4d5 c6d5 c3d5 e5e4 f3e2 a8c8 a2a4 a5c5 d5e3 c5c6 h4g3 f6h5 b3b4 h5g3 h2g3');
INSERT INTO evaluations VALUES(31920,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(31920,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(31920,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(31920,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(31920,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(31920,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(31920,36,'cp 58','c7c6 c2c4 g8f6 b1c3 e7e6 f1d3 b8d7 e1g1 f8e7 b2b3 e8g8 c1b2 d8a5 a2a3 f8e8 f3e2 g7g6 a1d1 a5b6 e2c2 e7f8 e3e4 d5e4 c3e4 f6e4 d3e4 f8g7 g2g3 a8d8 e4g2 h7h5 b2c3 d7f6 d1d3 d8d7 f1d1 e8d8 h2h3 b6c7 b3b4');
INSERT INTO evaluations VALUES(34736,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(34736,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(34736,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(34736,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(34736,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(34736,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34736,36,'cp 66','f1e2 e7d8 c2c4 f8e7 e1g1 g8f6 b1c3 e8g8 h2h3 b8c6 c4c5 f6e4 d1b3 e7f6 c1e3 e4c3 b2c3 h7h6 e2d3 b7b6 f1e1 b6c5 d4c5 a8b8 b3c2 f8e8 a1d1 c8e6 a2a3 c6e5 f3e5 f6e5 e3d4 e5d4 c3d4 d8f6 c2a4');
INSERT INTO evaluations VALUES(34734,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(34734,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(34734,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(34734,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(34734,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(34734,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(34734,36,'cp 66','e7d8 c2c4 g8f6 h2h3 f8e7 b1c3 e8g8 e1g1 b8c6 f1e1 c8e6 c4c5 f6e4 e2b5 f7f5 d1a4 c6b8 a4b3 c7c6 b5d3 d8c8 c1f4 b8d7 a1d1 e6f7 b3c2 f7h5 c3e2 g7g5 f4h2 e7f6 e2g3 h5g6 b2b4 a7a6 a2a4 g8h8 b4b5 a6b5 a4b5 c6b5 c2b3');
INSERT INTO evaluations VALUES(35066,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(35066,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(35066,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(35066,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(35066,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(35066,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(35066,36,'cp 126','b8c6 e2e3 f6d8 g1f3 d7d5 f1d3 f8e7 e1g1 g8f6 a2a3 e8g8 d1c2 d5c4 d3c4 e7d6 f1d1 d8e8 b2b4 e6e5 h2h3 g8h8 c1b2 e5e4 f3d2 c8f5 d1e1 a8d8 a1d1 a7a6 f2f3 d6g3 f3e4 f5g6 c4d3 g3e1 d1e1 c6e7 d2f3');
INSERT INTO evaluations VALUES(35687,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35687,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(35687,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(35687,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(35687,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35687,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35687,36,'cp 23','e2e4 d7d5 c4d5 c6d5 e4e5 f6e4 d2d4 d8a5 d1b3 f8d8 e1g1 b8c6 c1e3 a5b6 b3b6 a7b6 f1c1 c8f5 c3b5 f7f6 e5f6 g7f6 h2h4 a8c8 f3g5 h7h5 a2a3 e4g5 h4g5 f6g7 a3a4 c6b4 c1c3 c8c3 b2c3 b4c2 a1c1 c2e3 f2e3 e7e6');
INSERT INTO evaluations VALUES(41086,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41086,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41086,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41086,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41086,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41086,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41086,36,'cp -14','e8g8 f1d3 f8e8 e1g1 e7d6 f1e1 c7c6 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 f3e5 c8e6 c1f4 d8f6 f4g3 b8d7 d1c2 d7b6 e4h7 g8h8 h7e4 a8d8 a1d1 e6d5 a2a4 d5e4 e1e4 d6e5 d4e5 d8d1 c2d1 e8d8 d1b3 f6f5 f2f3 h8g8');
INSERT INTO evaluations VALUES(44967,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(44967,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(44967,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(44967,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(44967,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(44967,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(44967,36,'cp 40','e4e5 f6d7 f3f4 c6c5 g1f3 b8c6 c2c3 c5d4 c3d4 d7b6 f3h4 e7e6 h4f3 c8d7 b1c3 f8e7 f1d3 f7f5 b2b3 e8g8 a2a3 d7e8 e1g1 e8h5 c1e3 a8c8 c3e2 a7a5 h2h3 a5a4 b3b4 b6c4 d3c4 d5c4 e2c3 h5f3 d1f3 c6d4 f3f2 d4b3 a1d1');
INSERT INTO evaluations VALUES(45032,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(45032,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(45032,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(45032,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(45032,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(45032,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(45032,36,'cp 0','b8d7 b1c3 d8c7 g1e2 e7e5 c1e3 f8b4 a2a3 b4a5 e1g1 e8g8 e2g3 a5b6 d4e5 d7e5 e3f4 c8d7 c3a4 f8e8 a4b6 c7b6 f3e2 c6c5 f4e3 e5g4 e3f4 g4e5');
INSERT INTO evaluations VALUES(46990,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(46990,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(46990,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(46990,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(46990,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(46990,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(46990,36,'cp 46','g1f3 e7e5 b1c3 c7c6 f1g2 f8e7 e1g1 e8g8 d1c2 c8g4 f1d1 c6c5 h2h3 g4e6 f3g5 b8c6 g5e6 f7e6 d2d3 g8h8 d1f1 d8d7 c1e3 a8d8 e3g5 h7h6 g5f6 g7f6 c2d2 h8g7 a2a3 f6f5 b2b4 c6d4 d2b2 f5f4 e2e3 f4f3 e3d4');
INSERT INTO evaluations VALUES(46989,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(46989,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46989,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46989,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(46989,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46989,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(46989,36,'cp 36','e7e5 b1c3 c7c6 f1g2 f8e7 d2d3 e8g8 e1g1 a7a6 d1c2 c8f5 f1d1 d8c7 e2e4 f5e6 a2a4 a6a5 h2h3 b8a6 b2b3 f6d7 d3d4 a6b4 c2e2 f7f5 f3g5 e7g5 c1g5 f5e4 c3e4 d6d5 c4d5 b4d5 d4e5 d7e5 f2f4 h7h6 f4e5 h6g5 e4d6 d5c3');
INSERT INTO evaluations VALUES(47524,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(47524,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(47524,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(47524,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(47524,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(47524,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(47524,36,'cp 171','a2a3 b4c6 g1f3 c8g4 d4d5 c6e5 c1g5 g7g6 f1e2 e5f3 e2f3 g4f3 d1f3 f8g7 b1c3 e8g8 a1d1 b8d7 e1g1 h7h6 g5e3 f8e8 e3d4 d8c8 d4g7 g8g7 c3e4 d7f6 f3c3 c8f5 f2f3 e7e5 e4c5 b7b6 c5a6 f5d7 f1e1 d7e7 e1e3 e7d6 d1e1 c7c5 e3e5 e8e5 c3e5');
INSERT INTO evaluations VALUES(47525,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(47525,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(47525,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(47525,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(47525,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(47525,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(47525,36,'cp 170','b4c6 g1f3 c8g4 d4d5 c6e5 c1g5 g7g6 f1e2 g4f3 g2f3 f8g7 b1c3 c7c6 d1c2 c6d5 c3d5 f7f6 g5e3 e7e6 d5f4 d8e7 e1c1 b8c6 c4c5 g7h6 c1b1 h6f4 e3f4 e8g8 f4e3 f8d8 f3f4 e5f7 h2h4 d8d1 e2d1 e7d7 d1f3 a8d8 h4h5 g6h5 f3h5 f7h6 h1d1 d7e7 d1d8 e7d8');
INSERT INTO evaluations VALUES(46952,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(46952,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(46952,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(46952,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(46952,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(46952,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(46952,36,'cp 164','g1f3 c8g4 d4d5 c6e5 c1g5 g7g6 f1e2 e5f3 e2f3 g4f3 d1f3 f8g7 b1c3 b8d7 e1g1 h7h6 g5f4 e8g8 a1d1 g6g5 f4e3 d7e5 f3e2 e5g6 c3e4 c7c6 d5d6 e7d6 e4d6 b7b6 f1e1 d8f6 e3d4 g6f4 e2e3 f6g6 d4g7 g6g7 g2g3 g7g6 b2b4 a8d8 c4c5 f4d5 e3d4 b6c5 b4c5 d8d7 d4c4 d5f6 c4b3 f8d8 b3c3');
INSERT INTO evaluations VALUES(47860,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47860,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(47860,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(47860,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47860,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(47860,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47860,36,'cp 0','b1c3 g7g6 d2d4 e5d4 f3d4 f8g7 f1d3 b8d7 c1b2 e8g8 e1g1 a7a6 f1e1 f8e8 d3f1 c7c6 a2a4 d8c7 a4a5 d7c5 d1f3 f6d7 f3d1');
INSERT INTO evaluations VALUES(47863,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47863,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47863,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47863,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(47863,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(47863,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47863,36,'cp 0','g7g6 d2d4 e5d4 f3d4 f8g7 f1d3 b8d7 c1b2 e8g8 e1g1 a7a6 f1e1 f8e8 d3f1 c7c6 a2a4 d8c7 a4a5 d7c5 f2f3 b7b5 a5b6 c7b6 g1h1 a6a5 d1d2 c8b7 a1b1 b6b4 e1d1 a8d8 d2f2 b4b6 b2c1 b6b4');
INSERT INTO evaluations VALUES(56276,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(56276,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(56276,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(56276,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56276,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(56276,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(56276,36,'cp -13','g8f6 d2d4 f8d6 f1d3 d8e7 d3e2 e8g8 e1g1 c8g4 f1e1 f8e8 c1e3 c7c6 b1d2 e7c7 d1c2 b8d7 e2d3 g4h5 e3g5 h5g6 g2g3 c7a5 d3g6 h7g6 a2a4 a5c7 e1e8 a8e8 a1e1 e8e1 f3e1 d7f8 g5f6 g7f6 d2f1 f8e6 f1e3');
INSERT INTO evaluations VALUES(48736,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48736,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(48736,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(48736,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48736,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48736,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(48736,36,'cp 0','d2d4 f8d6 f1d3 d8e7 d3e2 c8g4 e1g1 e8g8 c1g5 b8d7 f1e1 f8e8 b1d2 h7h6 g5h4 d6f4 e2d3 e7e1 d1e1 e8e1 a1e1 a8e8 e1e8 f6e8 h4e7 d7b6 d3c2 f4d6 f3e5 g4e6 e7d6 e8d6 a2a4 a7a5 d2f1 f7f6 e5d3 g8f7 d3f4');
INSERT INTO evaluations VALUES(1805,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(1805,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(1805,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(1805,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(1805,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(1805,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(1805,36,'cp 201','b1c3 d7d6 e1g1 a7a6 f1e1 c8g4 c3d5 f6d8 c2c3 f8e7 d5e3 g4f3 d1f3 e7f6 e3c2 d8d7 f3g3 g7g5 c1d2 f6g7 b2b4 g8f6 c4b3 f6h5 g3f3 h5f4 a1d1 c6e7 h2h3 e8c8 d2f4 g5f4 d4e5 g7e5 b3f7 h8f8 f7b3 c8b8 c2d4 d7e8');
INSERT INTO evaluations VALUES(4131,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(4131,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(4131,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4131,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4131,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(4131,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(4131,36,'cp 0','c3d5 d7d6 a2a4 f6d5 c4d5 c6e7 d5b3 e7g6 a4a5 b6e3 f2e3 d8e7 f3d2 c8e6 b3e6 e7e6 f1f2 g6e7 h2h3 b7b5 d1e2 a7a6 c2c3 c7c5 d3d4 e5d4 e3d4 c5d4 c3d4 f7f5 a1e1 f5e4 f2f8 a8f8 e2e4 e6e4 d2e4 e7c6 d4d5');
INSERT INTO evaluations VALUES(5851,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(5851,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(5851,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(5851,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(5851,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(5851,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(5851,36,'cp 143','b1d2 g8f6 a2a3 b4e7 d1c2 e8g8 c4d3 h7h6 e1g1 c6b8 d2c4 b8d7 b2b4 b7b6 c1d2 c8a6 a3a4 a6b7 e3e4 a8c8 f1c1 f8e8 h2h3 a7a6 c2b1 b7a8 b4b5 a8e4 d3e4 f6e4 b1e4');
INSERT INTO evaluations VALUES(6481,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(6481,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(6481,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(6481,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(6481,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(6481,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(6481,36,'cp -63','g8f6 d2d3 e8g8 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 g2g3 d7d6 f1g2 c6e7 d5e7 d8e7 c2c3 a7a5 e1g1 c7c6 d3d4 c5b6 e2d3 c8g4 h2h3 g4e6 a2a4 a8e8 f1d1 e7c7 f3h4 e8d8 h4f5 e6f5 e4f5 e5d4 c3d4 d6d5 d3b3 b6a7');
INSERT INTO evaluations VALUES(10210,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(10210,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(10210,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(10210,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(10210,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(10210,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(10210,36,'cp -50','d2d3 d7d5 c1g5 c5b4 g5f6 g7f6 d3d4 d5e4 e2e4 f6f5 e4e3 f5f4 e3e2 c8g4 e1c1 b4c3 d4d5 c6b8 b2c3 b8d7 e2e4 d8e7 f1b5 f7f6 b5d7 g4d7 g2g3 f4g3 h2g3 e8c8 h1h6 e7a3 c1b1 h8f8 d1d3 d7b5 d3e3 a3c5 f3d2 d8d5 d2b3 c5c6 h6h7 b5c4 b1b2 f8d8 h7e7');
INSERT INTO evaluations VALUES(11201,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(11201,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(11201,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(11201,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11201,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(11201,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(11201,36,'cp 39','f8d6 f4g3 e8g8 e1g1 d6g3 h2g3 d8b6 d1b3 b6a7 d4c5 a7c5 e3e4 f6h5 e4d5 e6d5 f1e1 h5f6 f3e5 c8e6 d2f3 h7h6 e1e2 c6e5 e2e5 e6d7 a2a3 c5c7 a1e1 f8e8 e5e8 a8e8 b3b4 d7c6 e1e8 c6e8 d3c2 a6a5 b4d4 e8c6');
INSERT INTO evaluations VALUES(19327,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19327,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(19327,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19327,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19327,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(19327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19327,36,'cp 0','f8g7 f1c4 g7c3 b2c3 g8f6 g1e2 d8a5 c4d3 e8g8 e1g1 d7d5 e4d5 f6d5 c1d2 a5c5 d1e1 c6e5 e2d4 e5c4 d2h6 f8e8 d4b3 c5c6 b3d4');
INSERT INTO evaluations VALUES(5415,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5415,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5415,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5415,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5415,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5415,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5415,36,'cp 0','f1c4 g7c3 b2c3 g8f6 g1e2 d8a5 c4d3 e8g8 c1d2 a5c5 d2e3 c5a3 c3c4 b7b6 e1g1 c8a6 e3c1 a3a5 c1d2 a5a3');
INSERT INTO evaluations VALUES(20638,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(20638,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(20638,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(20638,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(20638,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(20638,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(20638,36,'cp 58','c2c4 f8e7 b1c3 f7f5 e4f5 c8f5 f1d3 f5e6 b3d2 g8f6 d2f3 d8d7 e1g1 c6b4 d3e2 e8g8 a2a3 b4c6 c1e3 g8h8 f3g5 e6g8 g5e4 f6e4 c3e4 d6d5 e2g4 d7d8 c4d5 d8d5 d1d5 g8d5 e4c5 c6d4 c5d7 f8d8 a1d1 d5b3 d7e5 b3d1 f1d1 d4c2 e5f7 h8g8 f7d8 c2e3 f2e3 a8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(23292,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(23292,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(23292,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(23292,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(23292,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(23292,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(23292,36,'cp 40','g2g3 g7g6 d2d4 e5e4 f3g1 g8f6 h2h4 f8b4 g1h3 b4c3 b2c3 d7d6 h3f4 c8d7 f4d5 h7h6 c1f4 f6h5 e2e3 c6e7 d1b3 e7d5 c4d5 h5f4 g3f4 b7b6 f1e2 a7a5 c3c4 e8g8 a1b1 d8e7 b3c2 g8h7 f2f3 a5a4 f3e4 e7e4 c2e4');
INSERT INTO evaluations VALUES(29072,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(29072,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(29072,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29072,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(29072,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(29072,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(29072,36,'cp 15','e7e6 f1c4 f8b4 e1g1 e8g8 f3h4 f5g6 h4g6 h7g6 d1c2 b8d7 f1d1 d8e7 a4a5 f8d8 c1d2 c6c5 c2b3 a8c8 d4c5 b4c5 c3a4 d7e5 c4e2 c5d6 h2h3 g6g5 d2e1 g5g4 h3g4 e5g4 a1c1 c8c1 d1c1 d6h2 g1f1 h2d6 c1d1 d8d7');
INSERT INTO evaluations VALUES(29307,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(29307,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29307,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(29307,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(29307,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(29307,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(29307,36,'cp 13','f1c4 f8b4 e1g1 e8g8 f3h4 f5g6 h4g6 h7g6 d1c2 b8d7 b2b3 d7b6 c4e2 a8c8 c3a2 b4e7 f1d1 c6c5 d4c5 c8c5 c2c5 e7c5 d1d8 f8d8 g1f1 f6e4 e2f3 e4d2 c1d2 d8d2 f3b7 d2b2 a2c1 c5b4 b7c6 g8f8 a4a5 b6d5 c6d5 e6d5 c1e2 b2b3');
INSERT INTO evaluations VALUES(29306,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(29306,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(29306,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(29306,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(29306,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29306,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29306,36,'cp 18','f8b4 e1g1 e8g8 f3h4 f5g6 h4g6 h7g6 d1c2 b8d7 c2b3 d8b6 f1d1 a8c8 h2h3 e6e5 b3c2 e5d4 e3d4 f6d5 c1d2 a7a5 c2b3 d7f6 c3d5 c6d5 c4d5 f6d5 b3d5 f8d8 d5b3 d8d4 d2e3 d4d1 a1d1 b6c7 d1c1 c7d8 c1c8 d8c8 b3d5 c8c7 b2b3');
INSERT INTO evaluations VALUES(28361,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(28361,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(28361,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(28361,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(28361,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(28361,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(28361,36,'cp 23','e1g1 e8g8 f3h4 f5g6 h4g6 h7g6 d1c2 b8d7 f1d1 d8e7 a4a5 f8d8 c4e2 a8c8 c2b3 c6c5 a1a4 c5d4 e3d4 b4c3 b2c3 f6e4 c3c4 e6e5 b3a3 e7a3 a4a3 e4d6 c4c5 d6f5 d4e5 d7c5 d1d8 c8d8 g2g3 d8c8 c1b2 c5e6 a5a6 c8c2');
INSERT INTO evaluations VALUES(28360,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(28360,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(28360,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28360,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(28360,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(28360,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(28360,36,'cp 16','e8g8 f3h4 f5g6 h4g6 h7g6 d1c2 b8d7 f1d1 d8e7 a4a5 f8d8 c4e2 a8c8 c2b3 c6c5 a1a4 c5d4 e3d4 b4c3 b2c3 f6e4 c3c4 e6e5 b3a3 e7a3 a4a3 e4d6 c1e3 d6c4 e2c4 c8c4 d4e5 d8c8 f2f4 d7c5 g1f2 c5e6 f2f3 b7b5 a5b6 a7b6 d1d7 b6b5 a3a7');
INSERT INTO evaluations VALUES(27800,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(27800,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(27800,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(27800,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(27800,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(27800,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(27800,36,'cp 21','f3h4 b8d7 h4f5 e6f5 h2h3 g7g6 d1b3 d8b6 c3a2 b4a5 b3d3 b6d8 b2b4 a5c7 b4b5 c6b5 a4b5 d7b6 c1a3 c7d6 a2c3 b6c4 d3c4 a7a6 a3c5 d6c5 d4c5 a6b5 c3b5 f6e4 c4d4 d8d4 e3d4 f8c8 a1a8 c8a8 f1b1');
INSERT INTO evaluations VALUES(30462,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(30462,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(30462,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(30462,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(30462,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(30462,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(30462,36,'cp 105','d4d5 e7e6 d5e6 d7e6 d1d8 e8d8 c3b5 g8f6 f2f3 b8c6 g1e2 f8b4 c1d2 b4d2 e1d2 d8e7 e2d4 h8c8 d4e6 f7e6 f1c4 a7a6 b5c3 c6d4 a1a6 a8b8 d2d3 d4c6 d3e3 b8b2 c3b5 f6d7 h1c1 d7c5 a6a2 b2b4 c4e2 b4b3 c1c3 b3c3 b5c3 c5b3 c3b5 b3c1 a2a1');
INSERT INTO evaluations VALUES(30735,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(30735,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(30735,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(30735,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(30735,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(30735,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(30735,36,'cp 94','g8f6 d1c2 d7d6 e2e4 e7e5 d4d5 b8d7 a2a3 a7a5 f1e2 d7c5 b2b3 b7c8 e1g1 c8d7 f3d2 f8h6 a1b1 e8g8 b3b4 c5b7 d2b3 h6c1 b3c1 f6e8 c1d3 d8e7 b1e1 f7f5 f2f4 a5b4 a3b4 e5f4 d3f4 f5e4 c3e4 e7e5');
INSERT INTO evaluations VALUES(31329,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31329,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31329,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31329,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31329,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31329,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31329,36,'cp 11','f3e5 b8c6 f1b5 f5d7 b5c6 d7c6 d2d4 e7e6 c1f4 g8f6 e1g1 f8e7 f1e1 f6d7 d1h5 d7e5 f4e5 e8g8 h5g4 g7g6 a2a4 a8c8 c3d1 c6d7 g4e2 f7f6 e5g3 f6f5 e2d2 e7g5 f2f4 g5h6 d1e3 h6g7 c2c3');
INSERT INTO evaluations VALUES(32761,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32761,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32761,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32761,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32761,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32761,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32761,36,'cp 0','g7g6 d2d4 c5d4 d1d4 g8f6 d4d3 f8g7 c1g5 b8c6 e1g1 h7h6 g5d2 f6g4 a1d1 g4e5 f3e5 c6e5 d3e2 a8c8 b2b3 e5g4 e2d3 g4e5 d3e2');
INSERT INTO evaluations VALUES(33170,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(33170,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(33170,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(33170,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(33170,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(33170,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(33170,36,'cp -72','b4c3 d2c3 h4e4 g1e2 b7b6 d1d2 g8f6 f2f3 e4c6 h2h4 c8a6 b2b3 d7d5 a2a4 c6d6 e1c1 b8c6 g2g4 d6d7 d2g5 h8g8 h4h5 h7h6 g5e3 e8c8 c3e1 c8b8 e1h4 g8h8 c1b2 d8c8 h1g1 f6e8 h4g3 e8d6 d1c1 c8e8 c1e1 d7e7 e3d2');
INSERT INTO evaluations VALUES(33182,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(33182,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(33182,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(33182,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(33182,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(33182,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(33182,36,'cp -45','d2c3 h4e4 g1e2 b7b6 d1d2 g8e7 f2f3 e4h4 g2g3 h4f6 e2f4 c8b7 e1c1 e7d5 f4d5 b7d5 f1e2 a7a5 b2b3 a5a4 h1e1 b8c6 c3b2 a4b3 a2b3 c6a5 d2e3 b6b5 e2b5 f6f3 c2c4 a8b8 d1d3 f3e3 d3e3 b8b5 c4b5 a5b3 c1c2 b3a5 e3a3 a5c4 a3a7');
INSERT INTO evaluations VALUES(33178,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(33178,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(33178,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(33178,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(33178,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(33178,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(33178,36,'cp -59','h4e4 g1e2 g8e7 d1d2 b7b6 f2f3 e4h4 g2g3 h4f6 e2f4 c8b7 e1c1 e7d5 f4d5 b7d5 f1e2 a7a5 b2b3 a5a4 c3b2 a4b3 a2b3 d5b7 g3g4 d7d5 h2h4 b8c6 e2b5 e8d8 h4h5 d8c8 g4g5 f6e7 f3f4 a8a5 b5c6 b7c6 d1e1 c6d7 c1b1 c8b7 h5h6 g7g6 f4f5');
INSERT INTO evaluations VALUES(33185,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(33185,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(33185,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(33185,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(33185,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(33185,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(33185,36,'cp -50','g1e2 g8e7 d1d2 b7b6 f2f3 e4h4 g2g3 h4f6 e2f4 c8b7 e1c1 e7d5 f4d5 b7d5 f1e2 h7h5 h1e1 d7d6 c1b1 b8c6 e2b5 f6g6 b2b3 e8c8 d2e2 c6e7 b5d3 g6f6 f3f4 d5b7 c3b2 d6d5 e2f2 g7g6 e1e5 e7c6 e5e1 c8b8 e1e2 c6e7 d1e1 e7f5 f2f3');
INSERT INTO evaluations VALUES(33184,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(33184,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(33184,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(33184,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(33184,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(33184,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(33184,36,'cp -53','g8e7 d1d2 b7b6 f2f3 e4h4 g2g3 h4f6 e2f4 c8b7 e1c1 e7d5 f4d5 b7d5 f1e2 h7h5 h1e1 d7d6 c1b1 b8c6 b2b3 f6h6 f3f4 c6e7 c3b2 e8c8 b1a1 c8b8 d1c1 d5e4 c2c4 h6g6 b3b4 f7f5 e1g1 g6f7 h2h4 g7g6 g1e1 h8e8 b4b5');
INSERT INTO evaluations VALUES(35102,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(35102,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(35102,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35102,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(35102,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(35102,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(35102,36,'cp 30','f8c5 g1f3 b8c6 c2c3 d7d6 b2b4 c5b6 c4b3 g8e7 b1d2 e7g6 e1g1 g6f4 d2c4 c8g4 g1h1 e8g8 c4e3 g4f3 d1f3 f4d3 f3f6 g7f6 e3d5 d3f2 f1f2 b6f2 c1h6 a7a5 b4b5 a5a4 d5f6 g8h8 b3d5 c6a5 h6f8 a8f8 a1f1 c7c6 b5c6 b7c6 d5c6 a5c6 f1f2 f8d8 h2h3 c6e7');
INSERT INTO evaluations VALUES(36651,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(36651,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(36651,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(36651,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(36651,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(36651,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(36651,36,'cp 125','e1c1 f6g4 e3f4 c7c5 d4c5 b8c6 f2f3 g7c3 b2c3 g4e5 c3c4 c8e6 d2c3 f7f6 c5d6 e7d6 c4c5 a8c8 d1d6 d8e7 g1h3 f8d8 d6d8 c6d8 f4e3 e6h3 f3f4 e5d7 g2h3 e7e4 f1c4 g8g7 h1d1 d7c5 e3c5 c8c5 d1d7 g7h6 c3f6 d8e6 c4b3 e4f4 f6f4 e6f4 d7b7 a7a5 b3g8 c5e5 b7h7 h6g5 c1b2 e5e2 h7a7 g5f6 a7a5');
INSERT INTO evaluations VALUES(38242,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(38242,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(38242,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(38242,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(38242,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(38242,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(38242,36,'cp -22','h7h6 b1c3 g7g5 d2d4 b8c6 g2g3 c8h3 g3f4 g5g4 f3e1 h3f1 c4f1 e7f6 c1e3 g8e7 d1d3 f6g7 a1d1 d6d5 d3b5 a7a6 b5e2 d8d7 c3d5 e7d5 e4d5 d7d5 f1g2 d5d7 e1d3 e8g8 d3e5 d7f5 e5g4 c6e7 g4e5 g8h8 c2c4 a8d8');
INSERT INTO evaluations VALUES(39281,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(39281,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(39281,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(39281,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(39281,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(39281,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39281,36,'cp -9','d7d5 d2d4 e5d4 e3d4 c5d6 g1f3 g8f6 d1e2 d6e7 e1g1 e8g8 c2c4 b8c6 b1c3 d5c4 e2c4 c8e6 c4a4 h7h6 f1e1 f8e8 c1f4 c6b4 e1e6 f7e6 a1e1 e7d6 f4d6 d8d6 c3b5 d6e7 b5c7 e7c7 a4b4 a8d8 f3e5 c7b6 b4b6');
INSERT INTO evaluations VALUES(41475,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(41475,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41475,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(41475,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41475,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(41475,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41475,36,'cp 0','c7c5 e1g1 b8c6 h2h3 c5d4 e3d4 d8b6 b1c3 b6b2 c3b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(41981,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(41981,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41981,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(41981,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(41981,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(41981,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(41981,36,'cp 31','e6d5 e2e3 h7h6 g5h4 g7g5 h4g3 f6e4 d1c2 h6h5 f2f3 e4g3 h2g3 b8c6 e1c1 d8d6 g3g4 h5h4 g1h3 b4c3 c2c3 h8g8 e3e4 d5e4 d4d5 c6e7 f1c4 c8d7 h1e1 e4f3 g2f3 e8c8 c3e3 g8e8 e3a7 c7c6 d5c6 e7c6 d1d6 e8e1 c1d2 c6a7 d2e1 f7f5 g4f5 d7f5 d6d8 c8d8 h3g5 d8e7 e1f2 e7f6');
INSERT INTO evaluations VALUES(42342,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(42342,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(42342,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42342,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(42342,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(42342,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(42342,36,'cp -22','b4c3 b2c3 d8f6 g2g3 b8c6 g1f3 e8g8 f1g2 d7d6 e1g1 e6e5 a2a4 c6a5 d1d3 c7c5 d4c5 d6c5 f3d2 c8f5 d3e3 a8d8 g2d5 b7b6 e3f3 f5e6 f3f6 g7f6 e2e4 g8g7 f2f4 f6f5 f4e5 f5e4 d2e4 a5c4 d5c4 e6c4 f1d1 d8d1 a1d1 c4b3');
INSERT INTO evaluations VALUES(44176,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(44176,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(44176,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(44176,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(44176,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(44176,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(44176,36,'cp -50','h5h4 d7d6 f1b5 c8d7 a2a4 f8e7 d2d3 b8c6 b5c6 d7c6 h4h3 e6e5 c1g5 c5c4 b1c3 c4d3 c2d3 e8g8 e1g1 h7h6 g5f6 e7f6 h3g3 d8b6 f1b1 b6b4 f3e1 b4b3 a1a3 b3e6 b2b4 f6e7 b4b5 c6d7 c3d5 e7d8 d5e3 a7a6 b5a6 b7a6 e3f5 e6f6 f5e3 d7e6 g3f3 a8c8 f3f6 d8f6');
INSERT INTO evaluations VALUES(52300,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(52300,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(52300,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52300,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(52300,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52300,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52300,36,'cp 0','f8g7 g2g3 b8c6 f1g2 d7d6 g1e2 e7e5 c3d5 g8e7 c1g5 d8a5 g5d2 a5d8');
INSERT INTO evaluations VALUES(37820,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37820,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37820,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37820,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37820,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37820,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37820,36,'cp 0','g2g3 b8c6 f1g2 a8b8 a2a4 d7d6 g1e2 a7a6 e1g1 b7b5 a4b5 a6b5 f2f4 c8d7 d1e1 g8f6 h2h3 e8g8 g3g4 c6b4 e1d2 b4c6');
INSERT INTO evaluations VALUES(52508,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(52508,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(52508,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(52508,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(52508,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(52508,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(52508,36,'cp -90','e7e5 e2e3 d4e3 c1e3 g8f6 b1c3 b8c6 d1c2 f8e7 a2a3 e8g8 e1c1 d7d6 c1b1 g7g6 h2h3 c8e6 g2g4 a8c8 f1e2 b7b5 c4b5 c6a5 f3e5 e6b3 c2c1 f6e4 e5c6 a5c6 c3e4 c6b4 c1d2 b3c2 b1a1 c2e4 d2b4 e4h1 d1h1 d6d5 b4b3 e7c5 h1d1 f8e8 d1d5 d8e7 a1a2');
INSERT INTO evaluations VALUES(56670,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(56670,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(56670,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56670,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(56670,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(56670,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(56670,36,'cp 34','g1f3 g8f6 f1c4 f8d6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 f1e1 e7d6 c3b5 b8c6 b2b3 c8g4 b5d6 c7d6 f3h4 c6b4 a1b1 b4a2 c1a3 a7a5 a3d6 f8d8 d6e7 d8d2 h2h3 g4h5 g2g4 b7b5 c4f1 h5g6 h4g6 h7g6 f1b5 a2c3 b5c6 c3b1 c6a8 b1c3');
INSERT INTO evaluations VALUES(56669,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(56669,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(56669,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(56669,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(56669,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(56669,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(56669,36,'cp 19','f8d6 f1c4 d8e7 c4e2 g8f6 e1g1 e8g8 d2d4 h7h6 f1e1 b8c6 d4d5 c6e5 c1f4 e5f3 e2f3 e7d8 d1d4 a7a6 a1c1 c8f5 h2h3 f8e8 e1e8 f6e8 g2g4 f5d7 f4d6 c7d6 d4b4 a6a5 b4d4 a8c8 g1g2 c8c5 c1e1 e8f6 e1e3 b7b5 c3e2 c5c4 d4a7');
INSERT INTO evaluations VALUES(1251,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1251,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1251,36,'cp 2','e1g1 g8f6 f1e1 f8d6 d2d3 e8g8 c3e2 a8b8 e2g3 b7b5 c4a2 b5a4 c2c3 c8b7 a2b1 c6e5 a1a4 e5f3 d1f3 a6a5 f3e2 b7c6 a4a1 d6e5 e2d2 c7b6 d2g5 e5c7 b1c2 g8h8 e4e5 f6e8 g5h4 d7d6 f2f4');
INSERT INTO evaluations VALUES(1250,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1250,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(1250,36,'cp 0','g8f6 f1e1 f8d6 d2d3 e8g8 c3e2 a8b8 c4a2 b7b5 e2g3 b5a4 c2c3 c8b7 a2b1 c6e5 f3e5 d6e5 a1a4 b7c6 a4a1 a6a5 d1d2 e5d6 d2e2 d6e5');
INSERT INTO evaluations VALUES(654,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(654,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(654,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(654,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(654,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(654,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(654,36,'cp 0','f1e1 f8d6 d2d3 e8g8 c3e2 d6e7 e2f4 e7d6');
INSERT INTO evaluations VALUES(2728,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2728,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(2728,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(2728,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(2728,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(2728,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(2728,36,'cp 16','a2a4 c8b7 f1e1 c6a5 b3a2 c7c5 a4b5 a6b5 c1g5 d8d7 h2h4 h7h6 g5f6 e7f6 b1d2 a5c6 a2d5 a8a1 d1a1 b7c8 d2f1 c6e7 f1e3 e7d5 e3d5 d7d8 g2g3 c8g4 f3h2 g4e6 a1d1 e6d5 e4d5 d8d7 d1g4 d7g4 h2g4 f6e7 e1a1');
INSERT INTO evaluations VALUES(3813,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(3813,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(3813,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(3813,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(3813,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(3813,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(3813,36,'cp -13','c5b6 d2d3 c6d4 h2h3 c7c6 a1b1 d6d5 c4a2 c8e6 f3d4 b6d4 c3e2 d4b6 e2g3 b6c7 d1e2 d5e4 a2e6 e4d3 e6f7 f8f7 c2d3 d8d7 g3e4 c7b6 c1e3 f6d5 b4b5 c6b5 e3b6 a7b6 e2b2 d5f4 b2e5 f4d3 e5b5 d7b5 b1b5');
INSERT INTO evaluations VALUES(3664,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(3664,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3664,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(3664,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3664,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(3664,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3664,36,'cp -15','d2d3 c6d4 h2h3 c7c6 a1b1 d6d5 c4a2 c8e6 f3d4 b6d4 c3e2 d4b6 e2g3 b6c7 f1e1 a7a5 b4b5 h7h6 a3a4 f8e8 b5c6 b7c6 c1b2 d8d7 a2b3 a8b8 d1f3 f6h7 b2c3 h7g5 f3h5 d5e4 g3e4 e6b3 c2b3 g5e4 d3e4 d7d3 h5f3 d3f3');
INSERT INTO evaluations VALUES(5947,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(5947,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(5947,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(5947,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5947,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(5947,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(5947,36,'cp 16','c2c3 c5b6 b1a3 a7a6 b5c6 b7c6 a3c4 b6a7 c1e3 g8e7 e3a7 a8a7 d3d4 f7f6 e1g1 e8g8 c4e3 e7g6 f3e1 a6a5 f2f3 c8e6 d1a4 d8e8 e1d3 a7a8 f1e1 f6f5 e4f5 e6f5 e3f5 f8f5 g1h1 e8d7 a4b3 g8h8 b3c2');
INSERT INTO evaluations VALUES(14469,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(14469,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(14469,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(14469,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(14469,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(14469,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(14469,36,'cp 104','d4d5 c6b8 h2h3 g7g6 e2e4 f8g7 f1d3 e8g8 g1e2 e7e6 c1g5 e6d5 c4d5 c7c6 e1g1 b8d7 d3c2 c6d5 c3d5 h7h6 g5h4 g6g5 h4g3 f6d5 d1d5 d7e5 e2d4 d8b6 a1d1 a7a5 a2a4 e5c6 d4c6 b7c6 d5d2 c8a6 f1e1 f8d8 g1h1 b6b2');
INSERT INTO evaluations VALUES(17185,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(17185,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(17185,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(17185,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(17185,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(17185,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(17185,36,'cp 123','d2d4 e5d4 f3d4 c6d4 d1d4 e7c6 d4e3 f8b4 c1d2 d7d6 a2a3 b4a5 e1c1 c8g4 f2f3 g4e6 f3f4 f7f6 c3d5 a5d2 d1d2 e8g8 f1e2 d8d7 e3b3 c6d8 b3b4 f8e8 e2f3 g8h8 h1d1 a7a5 b4c4 d8c6 h2h3 a8b8 f3g4 b7b5 g4e6 e8e6');
INSERT INTO evaluations VALUES(17404,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(17404,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(17404,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(17404,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(17404,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(17404,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(17404,36,'cp -28','g8f6 f1e1 f6e4 e1e4 d7d5 c4d5 d8d5 b1c3 d5d8 d2d4 f8e7 d4c5 d8d1 c3d1 e7c5 c1e3 c5e7 d1c3 e8g8 a4a5 c6b4 a1c1 f7f6 c3a4 e6e5 e3c5 b4c6 c5e7 c6e7 f3d4 g8f7 d4b3 c8d7 a4b6 d7c6 e4e1 a8d8 f2f3 h7h5 e1d1 g7g5 b3c5 d8d1 c1d1');
INSERT INTO evaluations VALUES(17435,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17435,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(17435,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(17435,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(17435,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(17435,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(17435,36,'cp 55','c2c4 f8b4 b1c3 g8f6 d4c6 b7c6 e4e5 f6e4 d1d4 f7f5 e5f6 d8f6 a2a3 f6d4 e3d4 b4c3 b2c3 e6e5 d4e5 e8g8 e5d4 c6c5 d4e3 d7d6 a1c1 a8b8 f2f3 e4f6 f1d3 f6d7 e1g1 d7e5 f1d1 c8e6 d3f1 b8b6 e3g5 e5c4 f1c4 e6c4 g5e7 f8c8 e7d6 b6b2 d6f4 b2a2 f4e5 a2a3 d1d7 c4f7 g1f2 a6a5 f2g3');
INSERT INTO evaluations VALUES(17427,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(17427,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(17427,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17427,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(17427,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(17427,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(17427,36,'cp 73','g8f6 b1c3 d8c7 a2a3 f8d6 f1e2 e8g8 h2h3 c6d4 e3d4 e6e5 d4e3 d6c5 e3c5 c7c5 e1g1 d7d6 d1d3 c8e6 b2b3 a8c8 f1c1 f8d8 d3d2 c5d4 d2c2 h7h6 c1d1 d4b6 a1b1 c8c7 e2f1 d8c8 c2d3 b6a5 a3a4 a5b4');
INSERT INTO evaluations VALUES(19702,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(19702,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(19702,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(19702,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(19702,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(19702,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(19702,36,'cp 101','d2d4 c5d4 f3d4 c8d7 c1e3 g8f6 f2f4 e7e5 d4c6 b7c6 h2h3 f8e7 f4e5 d6e5 d1f3 d8c8 f3g3 d7e6 c4b3 e8g8 e1g1 e6b3 a2b3 c8e6 a1a5 f6d7 g3f2 e7b4 a5a6 d7b8 a6a7 a8a7 e3a7 b8d7 c3a4 f8a8 a7e3 e6g6 f2f5 g6f5 f1f5 f7f6 f5f1 g8h7 f1d1 d7f8 c2c3 b4e7');
INSERT INTO evaluations VALUES(19820,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(19820,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(19820,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(19820,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(19820,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(19820,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(19820,36,'cp 58','d4c5 c8g4 c2c3 e7e6 c1e3 a7a6 b1d2 c6e5 f1e2 e5c6 e1g1 g4f3 d2f3 g7g6 b2b4 f8g7 d1b3 g8e7 a1d1 e8g8 e2d3 d8c7 f1e1 f8e8 g2g3 a8d8 c3c4 d8d7 a2a4 e8d8 d3e2 d5c4 d1d7 d8d7 e2c4 c6d4 f3d4 g7d4 e3d2');
INSERT INTO evaluations VALUES(19757,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(19757,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(19757,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(19757,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(19757,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(19757,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(19757,36,'cp 51','c8g4 c2c3 e7e6 c1e3 a7a6 b1d2 c6e5 f1e2 e5c6 b2b4 g7g6 f3d4 g4e2 d1e2 f8g7 d4c6 b7c6 e3d4 g8f6 e1g1 e8g8 a2a4 f6d7 f2f4 d8c7 h2h3 g7d4 c3d4 d7f6 a1b1 f6h5 g2g3 h5g7 d2f3 f7f6 e2e3 f8e8 b1b3 g8h8');
INSERT INTO evaluations VALUES(26258,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(26258,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(26258,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(26258,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(26258,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(26258,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(26258,36,'cp 43','c2c3 e7e6 c1e3 a7a6 b1d2 c6e5 f1e2 e5c6 e1g1 g4f3 d2f3 g7g6 d1b3 d8d7 c3c4 g8f6 a1d1 f8g7 e3f4 e8g8 f4d6 f8d8 h2h3 a8c8 f1e1 c6a5 b3e3 a5c4 e2c4 d5c4 f3e5 d7e8 e5c4 f6d5');
INSERT INTO evaluations VALUES(20027,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20027,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20027,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20027,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20027,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20027,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20027,36,'cp 0','g8f6 b1c3 e7e5 e2c4 f8e7 d2d3 e8g8 c1g5 f6d7 g5e7 c6e7 a2a4 d7f6 f3d2 c8e6 f1e1 a8b8 d2f1 a7a6 f1e3 e6d7 c4a2 d7e6');
INSERT INTO evaluations VALUES(13333,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(13333,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13333,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13333,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13333,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13333,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13333,36,'cp 6','b1c3 e7e5 e2c4 f8e7 d2d3 e8g8 f1e1 h7h6 a2a3 c8e6 c3d5 d8d7 c2c3 e7d8 b2b4 c5b4 d5f6 d8f6 a3b4 a7a6 c1e3 b7b5 c4e6 d7e6 f3d2 f8b8 h2h3 a6a5 b4a5 a8a5 a1a5 c6a5 d1c2 e6d7 e1b1 a5c6 d2b3 d6d5 b3c5');
INSERT INTO evaluations VALUES(13302,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(13302,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(13302,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(13302,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(13302,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(13302,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13302,36,'cp 0','e7e5 e2c4 f8e7 f1e1 h7h6 d2d3 e8g8 a2a3 c8e6 c3d5 e6d5 e4d5 c6d4 f3d4 c5d4 a3a4 a7a6 c2c3 d4c3 b2c3 b7b5 a4b5 a6b5 a1a8 d8a8 c4b5 f6d5 b5c4 d5f6 d3d4 a8a5 d1b3 d6d5 c4f1 f6e4 d4e5 e7c5 c1e3 c5e3 e1e3 e4d2');
INSERT INTO evaluations VALUES(20722,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(20722,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(20722,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(20722,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(20722,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(20722,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(20722,36,'cp 63','g8f6 d1d2 f8e7 f2f3 e8g8 g2g4 d6d5 g4g5 c6d4 d2d4 f6h5 f3f4 b7b6 e4d5 e7c5 d4d2 c5e3 d2e3 d8c7 e3e5 c7e5 f4e5 e6d5 f1g2 h5f4 g2d5 f4d5 c3d5 f8d8 e1c1 c8g4 d1d3 g4e2 d3e3 d8d5 e3e2 g8f8 h2h4 f8e7 h4h5 e7e6 h1f1 a8e8 h5h6');
INSERT INTO evaluations VALUES(24209,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(24209,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(24209,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(24209,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(24209,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(24209,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(24209,36,'cp 37','e1g1 h7h6 g5e3 c6d4 b5d7 d8d7 e3d4 e8g8 d1d3 c7c6 a1d1 d7e6 h2h3 f8e8 f1e1 b7b5 e4e5 d6e5 e1e5 e6d7 e5e3 c6c5 d4f6 d7d3 e3d3 e7f6 c3d5 f6d8 c2c3 g8f8 d3d2 d8a5 g1f1 a8d8 d5e3 c5c4 d2d7 a5b6 e3f5 d8d7');
INSERT INTO evaluations VALUES(24208,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24208,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(24208,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24208,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24208,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(24208,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24208,36,'cp 23','h7h6 g5e3 c6d4 b5d7 f6d7 e3d4 e8g8 c3d5 f8e8 d5e7 d8e7 f2f3 a7a5 a2a4 f7f6 d1d2 e7f7 d4f2 b7b6 a1d1 e8e7 d2c3 d7f8 h2h3 a8e8 b2b3 f8e6 f1e1 g8h8 f2g3 e6c5 c3d2 c5e6 d2d5 f7g6 g1h2 g6g5');
INSERT INTO evaluations VALUES(26866,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(26866,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(26866,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(26866,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(26866,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(26866,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(26866,36,'cp 18','c2c3 d8f6 b1d2 f8d6 e1g1 g8e7 f1e1 e8g8 d3e2 a8e8 d2f1 e7f5 h2h3 g4f3 e2f3 c6e7 a2a4 c7c6 g2g3 e7c8 f1e3 d6c7 e3f5 e8e1 d1e1 f6f5 g1g2 f5d7 h3h4 f8e8 e1d1 c8d6 f3g4 f7f5 g4f3 h7h6 c1d2 d6e4');
INSERT INTO evaluations VALUES(26865,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(26865,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26865,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(26865,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(26865,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(26865,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(26865,36,'cp 25','d8f6 b1d2 f8d6 e1g1 g8e7 f1e1 h7h6 h2h3 g4f5 d3f5 f6f5 d2f1 e8g8 f1e3 f5d7 c1d2 a8e8 a2a4 d6f4 e3g4 f4d2 d1d2 e7g6 g4e5 c6e5 f3e5 g6e5 d4e5 c7c6 a1d1 f7f6 c3c4 e8e5 e1e5 f6e5 c4d5 c6d5 d2d5 d7d5 d1d5');
INSERT INTO evaluations VALUES(27067,30,'cp 559',NULL);
INSERT INTO evaluations VALUES(27067,31,'cp 547',NULL);
INSERT INTO evaluations VALUES(27067,32,'cp 573',NULL);
INSERT INTO evaluations VALUES(27067,33,'cp 578',NULL);
INSERT INTO evaluations VALUES(27067,34,'cp 553',NULL);
INSERT INTO evaluations VALUES(27067,35,'cp 543',NULL);
INSERT INTO evaluations VALUES(27067,36,'cp 567','d4d5 e6c8 d5c6 b7c6 f3g5 g8h6 d1f3 d8e7 b1c3 f7f6 g5h3 g7g6 c1d2 h6f7 f3e2 a7a5 e1c1 f8g7 f2f3 a5a4 h3f2 f7d8 h2h4 d8e6 h4h5 g6h5 d1e1 a4a3 b2b3 e6d4 e2d1 f6f5 h1h5 h7h6');
INSERT INTO evaluations VALUES(27056,30,'cp 554',NULL);
INSERT INTO evaluations VALUES(27056,31,'cp 559',NULL);
INSERT INTO evaluations VALUES(27056,32,'cp 559',NULL);
INSERT INTO evaluations VALUES(27056,33,'cp 550',NULL);
INSERT INTO evaluations VALUES(27056,34,'cp 544',NULL);
INSERT INTO evaluations VALUES(27056,35,'cp 559',NULL);
INSERT INTO evaluations VALUES(27056,36,'cp 571','e6d7 d5c6 b7c6 f3g5 g8h6 d1f3 d8e7 b1c3 f7f6 h2h4 a7a5 f3d3 a8b8 c1d2 g7g6 g5h3 b8b2 c4b3 h6f7 d2c1 b2b3 a2b3 e7d8 a1a5 d6d5 e4d5 f8b4 a5a4 d7f5 d3f3 f5e4 f3e4 b4c3 e1d1 d8d5 e4d5 c6d5 a4a8 e8d7 a8h8 f7h8 f2f3 h8f7 h3f2 h7h5 c1d2 c3d4 d1e2 f7d6 f2d3');
INSERT INTO evaluations VALUES(28070,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(28070,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28070,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28070,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(28070,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(28070,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(28070,36,'cp 0','g5e3 d8b6 d1c1 g8h6 a2a4 a7a5 h2h3 f5g6 g2g4 e8g8 f1e2 c6c5 c2c4 c5d4 e3d4 b6b4 d4c3 b4b6');
INSERT INTO evaluations VALUES(32215,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(32215,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(32215,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32215,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(32215,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(32215,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(32215,36,'cp 36','c4e6 f7e6 f3b3 d8c8 g1f3 b8c6 e1g1 f8e7 d2d3 g8f6 c1e3 e8g8 b1d2 c8d7 b3b7 a8b8 b7a6 b8b2 c2c3 f6g4 d2c4 b2b8 h2h3 f8f3 h3g4 f3f8 a6a4 b8d8 f1b1 d6d5 c4a5 d5d4 c3d4 e5d4 a5c6 d4e3 f2e3 f8f7 c6e7 d7e7');
INSERT INTO evaluations VALUES(32202,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(32202,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(32202,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(32202,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(32202,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(32202,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(32202,36,'cp 13','f7e6 f3b3 d8c8 g1f3 b8c6 d2d3 g8f6 c1e3 f8e7 b1d2 e8g8 e1g1 a7a5 a2a4 c8d7 c2c3 h7h6 d3d4 e5d4 e3d4 f6h5 a1e1 f8f7 b3b7 a8b8 b7a6 b8b2 d4e3 h5f6 h2h3 d6d5 e4e5 f6h5 e1b1 b2c2 a6d3 c2a2 b1a1 a2a1');
INSERT INTO evaluations VALUES(32447,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(32447,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(32447,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(32447,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(32447,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(32447,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(32447,36,'cp 22','f3b3 d8c8 g1f3 b8c6 e1g1 f8e7 c2c3 g8f6 f1e1 a7a5 d2d4 a5a4 b3c2 e8g8 d4e5 c6e5 f3e5 d6e5 c1e3 c8e8 b1d2 e8g6 a1d1 e7d6 f2f3 f6d5 c2d3 d5e3 d3e3 b7b6 g1h1 a8d8 e3e2 d6c5 d2c4 g6g5');
INSERT INTO evaluations VALUES(32562,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(32562,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32562,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(32562,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(32562,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(32562,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(32562,36,'cp 19','d1e2 b8c6 c2c3 e7e6 d2d4 c5d4 c3d4 d6d5 e4e5 f6e4 c1e3 f8e7 f3e1 f7f6 f2f3 e4g5 e1d3 e8g8 b1d2 a7a5 a1c1 f6e5 d4e5 c6b4 d3b4 a5b4 d2b3 g5f7 e3b6 f8c8 c1c8 a8c8 b3d4 e7d8 b6d8 f7d8 f1d1 d8c6 e2b5 c6e5 b5d7 e5d7 d4e6');
INSERT INTO evaluations VALUES(36926,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36926,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36926,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36926,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36926,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36926,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(36926,36,'cp 16','e1g1 c7c6 b1d2 c8e6 f1e1 h7h6 g5h4 b8d7 c2c3 e6g4 d1c2 d8c7 h4g3 d6g3 h2g3 f8e8 d2f1 e8e1 a1e1 a8e8 e1e8 f6e8 f1e3 g4e6 g3g4 d7f6 c2e2 e8d6 g4g5 h6g5 f3g5 e6d7 e2c2 c7b6 g2g3 b6a5 a2a3 c6c5 d4c5 a5c5 g5f3');
INSERT INTO evaluations VALUES(41179,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(41179,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(41179,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(41179,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(41179,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(41179,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(41179,36,'cp 142','h2h3 e8g8 c1e3 f7f5 e4f5 c8f5 f1e2 d8d6 e1g1 b8c6 d1d2 a8d8 a1d1 f5e6 f3g5 b6c4 d2c1 c4e3 c1e3 e6f5 d4d5 c6b4 e3a7 d6f4 g5f3 f5e4 c3e4 f4e4 a7c5 b7b6 c5c4 e4c4 e2c4 g8h8');
INSERT INTO evaluations VALUES(41370,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(41370,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(41370,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41370,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41370,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(41370,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41370,36,'cp 0','e7e5 d4e5 f6g4 e3f4 g4e5 g1e2 b8c6 e1g1 c8e6 f4e3 d8d7 e2f4 e5c4 e3c1 c4e5');
INSERT INTO evaluations VALUES(44041,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44041,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(44041,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(44041,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(44041,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(44041,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44041,36,'cp -12','d7d5 d2d4 d5c4 g1f3 b8d7 a2a4 a7a5 e1g1 f8b4 d1c2 e8g8 h2h3 h7h6 f1d1 d8e7 c3a2 b4d6 e2e4 e6e5 c2c4 e5d4 f3d4 d7c5 a2c3 f8e8 c1e3 f6e4 d4c6 b7c6 c3e4 c5e4 c4e4 e7e4 g2e4 e8e4 d1d6 c8h3 d6c6 e4e8 a1c1 h3e6 c6c7');
INSERT INTO evaluations VALUES(45621,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(45621,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(45621,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(45621,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(45621,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(45621,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(45621,36,'cp 6','d2d4 d5c4 g1f3 b8d7 e1g1 b7b5 e2e4 b5b4 e4e5 f6d5 c3e4 f8e7 c1g5 e8g8 g5e7 d8e7 f1e1 a7a5 a1c1 c8a6 g2f1 d7b6 f3d2 c4c3 b2c3 a6f1 g1f1 b4c3 e4c3 f8d8 d2e4 e7b4 c3e2 b4a3 d1b3 a3b3 a2b3 d5e7 c1a1');
INSERT INTO evaluations VALUES(45640,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(45640,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(45640,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45640,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(45640,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(45640,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(45640,36,'cp 0','d5c4 g1f3 b8d7 e1g1 b7b5 e2e4 b5b4 e4e5 f6d5 c3e4 f8e7 d1c2 a7a5 f3d2 d5b6 f1d1 e8g8 d2c4 b6c4 c2c4 d7e5 c4c2 e5d7 c1f4 d7b6 c2c6 c8d7 c6b7 d7c8');
INSERT INTO evaluations VALUES(45096,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(45096,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(45096,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(45096,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(45096,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(45096,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(45096,36,'cp 75','b1c3 e7e5 g1f3 c8g4 c1e3 b8d7 f1e2 f8e7 d4d5 c6d5 c4d5 e8g8 h2h3 g4f3 e2f3 a7a6 e1g1 f6e8 d1b3 a8b8 a2a4 d7c5 b3a3 e7g5 e3c5 d6c5 a3c5 e8d6 c5b4 g7g6 f3e2 h7h5 f1d1 g5h6 b4a3 b7b5 a4b5 a6b5 a3a6 b5b4 c3a4 d6e4 d5d6');
INSERT INTO evaluations VALUES(49828,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(49828,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(49828,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(49828,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(49828,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(49828,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(49828,36,'cp -21','d7d5 e1g1 f8d6 d3e4 d6e5 e4f3 b8c6 d2d4 e5d4 c2c3 d4b6 f3d5 e8g8 b1a3 c8e6 d5e6 d8d1 f1d1 f7e6 c1e3 b6e3 f2e3 c6e5 b2b3 g8f7 d1d4 f7e7 d4e4 e5f7 a3b5 f7d6 b5d6 c7d6 e4b4 b7b6 b4g4 g7g6 a1f1 a8c8 c3c4 f8f1 g1f1 a7a6 g4h4 h7h5 f1e2 b6b5 c4b5 a6b5');
INSERT INTO evaluations VALUES(50142,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(50142,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(50142,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(50142,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(50142,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(50142,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(50142,36,'cp -31','e7f6 e2e3 d7d5 g1f3 f8g7 c2c4 d5c4 f1c4 c7c5 e1g1 b8c6 b1c3 e8g8 a1c1 f6f5 c3e2 d8e7 d2d4 f8d8 d1e1 c5d4 e2d4 c6e5 e1e2 e5c4 c1c4 h7h6 f1c1 c8d7 e2e1 a8c8 c4c8 d8c8 c1c8 d7c8 e1c1 e7d8 h2h3 a7a6 a2a4 c8d7');
INSERT INTO evaluations VALUES(52041,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(52041,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(52041,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(52041,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(52041,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(52041,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(52041,36,'cp -41','b1c3 a7a6 g1f3 b7b5 c4d3 g8e7 e1g1 e7g6 f1e1 f8e7 h2h4 e7h4 f3h4 d8h4 e4e5 c8b7 a2a4 b5b4 c3e4 b7e4 d3e4 b8c6 g2g3 h4d8 f2f4 h7h5 c1e3 h5h4 g3g4 e8g8 d1d3 c6e7 a1c1 f7f5 e5f6 f8f6 d3d6 a8c8 c1c8 e7c8 d6b8 d8f8 e1c1 g6f4 b8c8');
INSERT INTO evaluations VALUES(52556,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(52556,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(52556,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(52556,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(52556,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(52556,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(52556,36,'cp 59','g8h6 b1c3 c8f5 g1f3 e7e6 f1d3 c6c5 e5f6 g7f6 d1e2 d8d7 c1e3 c5c4 d3f5 h6f5 g2g4 f5e3 e2e3 b8a6 g4g5 e8c8 e1c1 a6c7 h1g1 f8d6 g5f6 d7f7 f3e5 f7f6 d1f1 h8f8 a2a3 c8b8 c1b1 d6e5 e3e5 a7a6 h2h4 f8f7 e5f6 f7f6 g1g7 d8f8 g7h7 f6f4 f1f4 f8f4');
INSERT INTO evaluations VALUES(52681,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(52681,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(52681,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(52681,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(52681,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(52681,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(52681,36,'cp 38','e7e5 c3d5 g8f6 c2c3 f6d5 c4d5 d8f6 d2d3 f8e7 e1g1 b8d7 a2a4 d7b6 d5b3 c8e6 a4a5 b6d7 f3d2 e8g8 d2c4 f6g6 c4e3 a8b8 e3d5 e7g5 c1g5 g6g5 d5e3 g5e7 b3c4 d7f6 d1e2 b8e8 a1e1 e7d7 h2h3 g8h8 c4e6 f7e6');
INSERT INTO evaluations VALUES(53984,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(53984,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(53984,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(53984,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(53984,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(53984,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(53984,36,'cp -8','g8f6 d2d3 a7a6 e1g1 b7b5 c4b3 b8d7 f1e1 c8b7 h2h3 f8e7 c1f4 e8g8 b1d2 f8e8 c2c3 d8c7 d1e2 h7h6 a3a4 a8c8 f4g3 e7f8 a4b5 a6b5 d3d4 c5c4 b3c2 e6e5 d4d5 d7c5 d2f1 c8a8 f3d2 g7g6 g1h1 c7d7 f1e3 a8a1 e1a1');
INSERT INTO evaluations VALUES(46008,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(46008,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(46008,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(46008,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(46008,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(46008,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(46008,36,'cp -11','d2d3 a7a6 e1g1 b7b5 c4b3 d8c7 f1e1 c8b7 c1f4 f8e7 b1d2 e8g8 c2c3 f8e8 h2h3 b8d7 d1e2 e7f8 a3a4 h7h6 f4g3 c7b6 b3c2 g7g6 e4e5 d7e5 f3e5 d6e5 g3e5 f8g7 d3d4 b6c6 d2f3 f6d7 e5g7 g8g7 a4b5 a6b5');
INSERT INTO evaluations VALUES(55813,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(55813,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(55813,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(55813,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(55813,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(55813,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(55813,36,'cp -36','c7c5 e2e3 c5d4 e3d4 g8f6 g1f3 d5c4 b3c4 e6e5 d4d5 f8b4 f3d2 d8b6 c1a3 a7a5 f1e2 c8f5 e1g1 b8d7 a3b2 e8g8 b1c3 a5a4 a2a3 b4c5 a1a2 b6a5 d2f3 h7h6 c3b5 f8e8 d1a1 e5e4 f3e5 a5b6 e5d7 f5d7 b2f6 b6f6 a1f6');
INSERT INTO evaluations VALUES(56346,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(56346,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(56346,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(56346,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(56346,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(56346,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(56346,36,'cp 32','c4d5 d8d5 b1c3 d5d8 g1f3 b8c6 a2a3 g8f6 b2b4 f8d6 e2e3 e8g8 c1b2 f8e8 f1e2 c6e7 e1g1 e7g6 d1c2 c7c6 f1d1 h7h6 c3e4 f6e4 d3e4 d8e7 d1d2 d6c7 a1d1 c8e6 b2c3 a8d8 g2g3 e6h3 d2d8 e8d8');
INSERT INTO evaluations VALUES(56164,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(56164,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(56164,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(56164,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(56164,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(56164,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(56164,36,'cp 19','d8d5 b1c3 d5e6 g2g3 f8e7 g1f3 g8f6 f1g2 e8g8 e1g1 b8c6 c3b5 e6d7 b2b3 f8d8 c1b2 a7a6 b5a3 d7e8 d1c2 c8f5 f3h4 f5g4 a3c4 f6d5 h2h3 g4e6 h4f3 f7f6 a2a3 d5b6 c4d2 e8h5 g1h2 h5f7 b3b4 e6d5');
INSERT INTO evaluations VALUES(34944,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(34944,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(34944,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(34944,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(34944,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(34944,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(34944,36,'cp 23','b1c3 d5e6 g2g3 f8e7 g1f3 g8f6 f1g2 e8g8 e1g1 b8c6 c1g5 c6d4 f3d2 c7c6 e2e3 d4f5 g5f6 e6f6 c3e4 f6e6 a2a4 f5d6 e4c5 e6g6 d1b3 c8f5 b3c3 b7b6 c5e4 g6e6 e4d6 e7d6 g2e4 a8c8 e4f5 e6f5 d2e4 d6e7 a4a5 f5e6 a5b6');
INSERT INTO evaluations VALUES(34940,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(34940,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(34940,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(34940,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(34940,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(34940,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(34940,36,'cp 28','d5d8 g1f3 b8c6 a2a3 f8d6 b2b4 g8f6 e2e3 e8g8 c1b2 f8e8 f1e2 c6e7 e1g1 e7g6 d1c2 d8e7 f1d1 c8d7 f3d2 d7e6 d2c4 a8c8 a1c1 e8d8 g2g3 c7c6 c4d6 d8d6 c3e4 f6e4 d3e4 d6d1 e2d1 h7h6');
INSERT INTO evaluations VALUES(3504,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(3504,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(3504,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(3504,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(3504,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(3504,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(3504,36,'cp -33','e7e5 f4g5 h7h6 g5h4 d8e7 d3c2 e5d4 f3d4 c6d4 c3d4 c7c5 e1g1 f8e8 f1e1 c5d4 e3d4 e7e1 d1e1 e8e1 a1e1 c8d7 c2b3 a8e8 f2f3 b7b5 e1e8 f6e8 h4f2 a7a5 a2a3 a5a4 b3a2 g8f8 d2e4 f8e7 e4c3 e8c7 f2e3 d7e6 a2e6 e7e6 g1f2 e6d7 f2e2 d7c6 e2d3 d6d5 g2g4 h6h5');
INSERT INTO evaluations VALUES(5321,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(5321,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(5321,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(5321,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(5321,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(5321,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(5321,36,'cp -23','f1e2 d7d6 b1c3 e7e6 c1e3 c6d4 d1d2 g8e7 e1g1 e8g8 e3h6 e7c6 h6g7 g8g7 f3d4 c6d4 f2f4 c8d7 a1e1 a8b8 e2d1 b7b5 b2b3 d7c6 c3e2 d8f6 d2e3 d4e2 e1e2 f6d4 f4f5 d4e3 e2e3 h7h6 f5f6 g7h7 e4e5 f8d8 h2h4 b5b4 e5d6 d8d6 d1f3 c6f3 e3f3 a7a6');
INSERT INTO evaluations VALUES(8795,30,'cp 221',NULL);
INSERT INTO evaluations VALUES(8795,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(8795,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(8795,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(8795,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(8795,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(8795,36,'cp 225','e4d5 c6e7 f3e5 e8g8 c1f4 f6d5 f4g3 c5d6 d3d4 d6e5 g3e5 c8e6 b1d2 e7g6 f1e1 g6e5 e1e5 c7c6 h2h3 f8e8 d1f3 d8f6 f3f6 g7f6 e5e1 g8f8 a2a4 d5f4 c4f1 f6f5 g1h2 f8g7 c3c4 f4g6 e1e3 b7b6');
INSERT INTO evaluations VALUES(8793,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(8793,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(8793,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(8793,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(8793,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(8793,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(8793,36,'cp 212','c6e7 f3e5 e8g8 d3d4 c5d6 f1e1 e7d5 b1d2 c8e6 c4f1 c7c6 h2h3 f8e8 c3c4 d5e7 d2f3 f6d7 c1d2 d7e5 f3e5 e7g6 e5g6 f7g6 d2c3 d8g5 d1f3 g5f6 f3f6 g7f6 g2g3 e6f7 e1d1 d6f8 f1g2 e8d8 d4d5 c6d5 c3f6');
INSERT INTO evaluations VALUES(10311,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(10311,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(10311,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(10311,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(10311,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(10311,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(10311,36,'cp -15','g1f3 d7d5 e4d5 f6d5 e1g1 e8g8 d3e4 d5f6 e4c6 b7c6 b1c3 c8a6 d1d8 f8d8 f1e1 f6d5 c3d5 c6d5 c2c3 c5f8 b2b4 c7c5 b4c5 f8c5 a2a4 c5f8 c1e3 a6d3 f3e5 d3f5 e5c6 d8d7 e1d1 d7c7 c6d4 f5g6 d4b5 c7d7 b5d4 a7a6 d1d2 a8e8 a4a5 d7b7');
INSERT INTO evaluations VALUES(11109,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(11109,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(11109,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(11109,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(11109,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(11109,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(11109,36,'cp 26','f1c4 f8c5 e1g1 d7d6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 a2a4 e8g8 c2c3 g8h8 d1e2 a6a5 f3d2 c6e7 g1h1 e7d5 c4d5 c8e6 e2d3 a8b8 d5e6 f7e6 d2c4 d8c7 f2f3 h8g8 f1d1 b7b5 c4d6 c5d6 d3d6 c7d6 d1d6 b5a4');
INSERT INTO evaluations VALUES(11105,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(11105,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(11105,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11105,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(11105,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(11105,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(11105,36,'cp 35','f8c5 e1g1 d7d6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 e8g8 a2a4 g8h8 d1d3 a6a5 a1b1 f7f5 f3d2 c6e7 g1h1 c8d7 b2b3 a8c8 f2f3 f5f4 d5e7 d8e7 c4d5 c8c7 d2c4 b7b6 b3b4 a5b4 c3b4 c5d4 b4b5 d7e6 a4a5 e6d5 e4d5 b6a5 c4a5 c7c3 d3e4 e7c7');
INSERT INTO evaluations VALUES(12408,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(12408,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(12408,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(12408,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(12408,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(12408,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(12408,36,'cp 186','d4d5 c6a5 e2e4 d7d6 c1g5 c8d7 f1d3 e6e5 e1g1 g7g6 b2b3 a7a6 a1b1 b7b6 g1h1 f8g7 d1e2 d7g4 h2h3 h7h6 g5d2 g4d7 h1h2 e8g8 g2g3 a5b7 a2a3 f8e8 b1e1 f6h5 f3h4');
INSERT INTO evaluations VALUES(20677,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(20677,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20677,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20677,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20677,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20677,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20677,36,'cp 0','g8f6 d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 e1g1 a7a6 c4d3 c6d4 e3d4 b7b5 d1f3 c8b7 a1d1 a8c8 f3h3 h7h6 f1e1 f8e8 e4e5 d6e5 d4e5 d8b6 h3g3 f6h5 g3h3');
INSERT INTO evaluations VALUES(22137,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(22137,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(22137,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(22137,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(22137,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(22137,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(22137,36,'cp 75','g8f6 f1e2 f8e7 e1g1 e8g8 b1c3 f8e8 f2f3 c8e6 c1f4 f6d7 d4d2 e7f6 a1d1 f6e5 f4g3 d7f6 g3f2 c7c6 e2d3 d8a5 c3e2 a5d2 d1d2 d6d5 e4d5 c6d5 c4c5 e6d7 e2d4 a8c8 d2c2 a7a6 f1d1 g7g6 g2g3 f6h5 d1c1 e5g7 c5c6 b7c6');
INSERT INTO evaluations VALUES(15763,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(15763,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(15763,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(15763,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(15763,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(15763,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(15763,36,'cp 83','b1c3 f8e7 f1e2 e8g8 e1g1 f8e8 f2f3 c8e6 c1f4 f6d7 d4d2 e7f6 f4e3 d7c5 a1d1 a7a5 b2b3 h7h6 c3b5 b7b6 d2c2 d8e7 e3f2 a8d8 f1e1 e6c8 e2f1 f6e5 g2g3 c8b7 f1h3 b7c6 h3g2 e5f6 b5c3 f6e5 c3d5 c6d5 c4d5 e7f6');
INSERT INTO evaluations VALUES(15762,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(15762,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(15762,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(15762,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(15762,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(15762,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(15762,36,'cp 81','f8e7 f1e2 e8g8 e1g1 f8e8 f2f3 c8e6 c1f4 a7a5 a1d1 f6d7 d4d2 e7f6 f4e3 f6c3 d2c3 d8e7 d1e1 a5a4 f3f4 f7f6 e2f3 a8d8 e3d4 g8h8 f3d1 e7f7 b2b3 a4b3 a2b3 f7e7 d1c2 e6g8 c3e3 g8f7 d4c3 d7c5 h2h4 c5e6');
INSERT INTO evaluations VALUES(25429,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(25429,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(25429,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(25429,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(25429,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(25429,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(25429,36,'cp 139','g5e6 d8d7 e6f8 h8f8 b1c3 e8c8 e1g1 b7b6 a2a3 h7h6 c3d5 f6d5 e4d5 c6d4 a3a4 a7a6 c2c4 f8f7 a4a5 b6b5 c1e3 d4f5 h2h3 d8f8 d1g4 c8b8 f1e1 f5e3 g4d7 f7d7 f2e3 e5e4 c4b5 e4d3 b5a6 b8a7 a1d1 a7a6 d1d3 a6a5');
INSERT INTO evaluations VALUES(25425,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(25425,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(25425,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(25425,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(25425,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(25425,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(25425,36,'cp 138','d8d7 e6f8 h8f8 b1c3 b7b6 e1g1 e8c8 c1e3 c8b7 f2f3 h7h6 d1d2 a7a5 a2a3 d6d5 e4d5 f6d5 a1e1 d5f4 e3f2 g7g5 e1e4 h6h5 b2b4 h5h4 h2h3 a5b4 a3b4 f4h3 g2h3 d7h3 f2e3 c6d4 e3d4 e5d4 c3e2 f8f3 f1f3 h3f3 d2g5');
INSERT INTO evaluations VALUES(26652,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(26652,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(26652,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(26652,33,'cp -168',NULL);
INSERT INTO evaluations VALUES(26652,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(26652,35,'cp -168',NULL);
INSERT INTO evaluations VALUES(26652,36,'cp -162','c6d4 c2c3 f5d3 d1d3 d4c6 g1f3 e7e6 e1g1 g8e7 d3e2 d8d7 f1e1 h7h6 b2b4 g7g5 b4b5 c6a5 b1d2 e7g6 a1b1 f8e7 f3d4 a7a6 b5a6 b7a6 d4b3 a5b3 b1b3 a8d8 g2g3 e8g8 c1b2 c7c5 c3c4 d5d4 b2c1 g8g7 b3b6 d7c7 b6a6');
INSERT INTO evaluations VALUES(26715,30,'cp -153',NULL);
INSERT INTO evaluations VALUES(26715,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(26715,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(26715,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(26715,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(26715,35,'cp -172',NULL);
INSERT INTO evaluations VALUES(26715,36,'cp -176','c2c3 f5d3 d1d3 d4c6 f2f4 e7e6 g1f3 f8c5 b2b4 c5b6 a2a4 a7a6 a4a5 b6a7 c1e3 a7e3 d3e3 d5d4 c3d4 g8e7 b1c3 c6b4 e3d2 b4d5 c3e4 e8g8 e1g1 b7b6 g2g4 d8d7 f3h4 e7g6 h4g2 f7f5 e5f6 d5f6 e4f6 f8f6 a5b6 c7b6');
INSERT INTO evaluations VALUES(31031,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(31031,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(31031,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(31031,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31031,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31031,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(31031,36,'cp -12','h7h5 g1h3 e7e6 h3f4 g8e7 c2c3 c6c5 c1e3 b8d7 b1d2 a7a6 f1e2 a8c8 a1c1 d8b6 d1b3 b6b3 d2b3 h5g4 f3g4 c5d4 c3d4 c8c1 e3c1 g6e4 h1h3 e7c6 h4h5 f8e7 c1d2 e8g8 e2d3 e4d3 f4d3 f7f6 e5f6 d7f6 d3f2 f6e4 f2e4 d5e4 g4g5');
INSERT INTO evaluations VALUES(30952,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30952,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30952,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30952,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30952,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30952,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30952,36,'cp 0','g1h3 e7e6 h3f4 g8e7 c2c3 c6c5 c1e3 b8d7 b1d2 a8c8 d4c5 h5g4 f3g4 e7c6 f4g6 f7g6 f1d3 c6e5 d1e2 f8c5 e3c5 c8c5 e1c1 e8g8 d2b3 e5d3 d1d3 c5c6 b3d4 c6a6 d4e6 f8e8 e6d8 e8e2 d3d2 e2d2 c1d2 a6a2 d2c2 a2a4 g4g5 d7c5 b2b4 c5e4 h1d1 a4a3 d1d3 a3a2 c2b3 a2d2 d3d2 e4d2 b3c2 d2f3 d8b7');
INSERT INTO evaluations VALUES(34045,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34045,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(34045,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(34045,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(34045,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(34045,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(34045,36,'cp 1','b1c3 g7g6 g1e2 f8g7 e1g1 d7d6 d2d4 b8d7 d4d5 e8g8 a2a4 f6e8 e2g3 e7h4 c4b5 a7a6 b5e2 a6a5 f3g4 h4g4 e2g4 d7c5 g4e2 c8d7 f2f3 c5a6 e2c4 a6b4 c4b3 f7f5 g3h1 f5e4 c3e4 e8f6 h1f2 b7b6 e4f6 g7f6 f2e4 a8e8 f1e1');
INSERT INTO evaluations VALUES(35081,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(35081,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(35081,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(35081,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(35081,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(35081,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(35081,36,'cp 177','d7d6 d2d4 f6d8 f1d3 a7a6 d1e2 g8e7 h2h4 h7h6 h4h5 b8d7 c1d2 e7c6 c3d1 e6e5 c2c3 d7f6 d1e3 f6g4 e3d5 g4f6 d5f6 d8f6 d4d5 c6b8 f3h4 b8d7 d3c2 d7b6 a2a4 g7g5 h5g6 f7g6 d2e3 c8d7 a4a5 d7b5 e2f3 f6f3');
INSERT INTO evaluations VALUES(39703,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(39703,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(39703,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(39703,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(39703,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(39703,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(39703,36,'cp 9','e1c1 d8c7 a2a3 b8c6 h2h4 h7h6 g5e3 c8d7 d4c6 d7c6 g2g4 d6d5 e4d5 f6d5 c3d5 c6d5 d1d5 e6d5 f1d3 c7c6 e3f4 e7d6 h1e1 e8f8 f4d2 a8e8 e1e8 f8e8 d2c3 d5d4 d3e4 c6c4 e4d3');
INSERT INTO evaluations VALUES(39702,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(39702,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39702,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39702,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39702,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39702,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39702,36,'cp 13','d8c7 a2a3 b8c6 f3g3 c8d7 f1e2 a8c8 c1b1 e8g8 d4b3 b7b5 d1d2 c6e5 f2f4 e5c4 d2d4 f8d8 h1d1 d7e8 e2c4 b5c4 b3d2 d6d5 e4d5 e6d5 g5f6 e7f6 d4d5 f6c3 d5d8 c8d8 g3c3 c7f4 g2g3 f4d4 c3d4 d8d4 d1e1 e8c6 b1c1 h7h6');
INSERT INTO evaluations VALUES(42551,30,'cp -136',NULL);
INSERT INTO evaluations VALUES(42551,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(42551,32,'cp -159',NULL);
INSERT INTO evaluations VALUES(42551,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(42551,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(42551,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(42551,36,'cp -111','f6e4 d2d4 e5d4 d1b3 e8g8 c3d4 c5b6 b1c3 e4c3 b2c3 b8c6 c4d3 f8e8 c1e3 d7d5 h2h3 c6a5 b3c2 h7h6 e1g1 a5c4 a1e1 c8e6 f3e5 c4e5 d4e5 c7c5 f2f4 c5c4 d3h7 g8h8 e3b6 d8b6 g1h1 e6d7 h7f5 a8d8 f5d7 d8d7 c2f5 b6c6');
INSERT INTO evaluations VALUES(43045,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(43045,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(43045,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(43045,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43045,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(43045,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(43045,36,'cp -41','d2d3 d7d5 e4d5 f6d5 b1c3 c7c6 e1g1 e8g8 c4d5 c6d5 c1f4 b8c6 d1e1 c8e6 e1f2 f7f6 a1e1 e6f7 f4e3 d8d7 h2h3 f8e8 a2a3 a8d8 c3b5 a7a6 b5d4 c6d4 e3d4 e7d6 e1e8 f7e8 f3h4');
INSERT INTO evaluations VALUES(43858,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(43858,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(43858,32,'cp 265',NULL);
INSERT INTO evaluations VALUES(43858,33,'cp 261',NULL);
INSERT INTO evaluations VALUES(43858,34,'cp 287',NULL);
INSERT INTO evaluations VALUES(43858,35,'cp 297',NULL);
INSERT INTO evaluations VALUES(43858,36,'cp 300','f3e5 b8c6 d2d4 f8b4 e5c6 d7c6 f2f3 e8g8 a2a3 b4e7 c1e3 b6b5 f1d3 a7a6 e1g1 f6d7 d1d2 c6c5 d4d5 d7b6 b2b3 f8e8 f3f4 f7f5 e4e5 b5b4 a3b4 c5b4 c3e2 b6d5 d3c4 c8e6 f1d1 c7c6 e2d4 d8d7 d4e6 d7e6');
INSERT INTO evaluations VALUES(43850,30,'cp 285',NULL);
INSERT INTO evaluations VALUES(43850,31,'cp 299',NULL);
INSERT INTO evaluations VALUES(43850,32,'cp 282',NULL);
INSERT INTO evaluations VALUES(43850,33,'cp 296',NULL);
INSERT INTO evaluations VALUES(43850,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(43850,35,'cp 307',NULL);
INSERT INTO evaluations VALUES(43850,36,'cp 282','f8b4 f1c4 e8g8 e5f7 f8f7 c4f7 g8f7 e4e5 d8e7 d1e2 b8c6 e5f6 g7f6 e1d1 e7e2 c3e2 c6e5 h1f1 e5g6 a2a3 b4f8 g2g3 h7h5 d2d4 c8b7 f1e1 b7f3 c1e3 g6e7 c2c4 d7d5 c4d5 f3d5 e2c3 d5f3 d1d2 e7f5 d4d5 a8d8 d2c2 h5h4');
INSERT INTO evaluations VALUES(44820,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(44820,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44820,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(44820,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44820,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(44820,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(44820,36,'cp 52','d5c4 e2e3 b7b5 a2a4 b5b4 c3b1 c4c3 b2c3 e7e6 f1d3 a7a5 e1g1 f8e7 d1e2 e8g8 e3e4 b8d7 f1c1 d8b6 c3b4 a5b4 b1d2 c8a6 d2c4 a6c4 c1c4 f8c8 e4e5 f6d5 f4d2 c6c5 e2e4 d7f8 e4g4 g8h8 h2h4 f8d7 g4e4 d7f8 h4h5 h8g8 d2e3 d5e3');
INSERT INTO evaluations VALUES(45688,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(45688,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(45688,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(45688,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(45688,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(45688,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(45688,36,'cp 112','f6e4 c4d5 c6d5 f1d3 b8c6 d3e4 d5e4 g1e2 f8e7 c1e3 b7b6 e1g1 e8g8 c3e4 c6b4 e2g3 b4d5 g3h5 f7f5 e5f6 d5f6 e4f6 e7f6 d1g4 d8d7 a1c1 f8f7 f1d1 c8b7 h2h3 a8e8 c1c3 e8c8 e3h6 c8c3 b2c3');
INSERT INTO evaluations VALUES(46676,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(46676,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(46676,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(46676,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(46676,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(46676,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(46676,36,'cp 53','b1c3 f8g7 g1f3 e8g8 f1e2 e7e5 e1g1 c8g4 d4d5 a7a5 c1e3 b8a6 a2a3 b7b6 a1b1 a6c5 f3d2 g4e2 d1e2 a5a4 e3c5 b6c5 b2b3 g7h6 b3a4 h6d2 e2d2 f6d7 b1b7 d7b6 c3b5 a8a4 b5c7 a4c4 f2f3 c4d4 d2h6 b6c4 b7a7');
INSERT INTO evaluations VALUES(47315,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(47315,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(47315,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(47315,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(47315,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(47315,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(47315,36,'cp -28','b1c3 c7c6 e2e3 b7b5 a2a4 d8b6 f1e2 c8b7 g1f3 e7e6 e1g1 b8d7 a4a5 b6c7 b2b3 f6d5 d1c1 f8b4 c3a2 b4a5 b3c4 b5c4 e2c4 c6c5 e3e4 d5f6 g5f6 d7f6 c4b5 e8f8 c1a3 f8g8 e4e5 f6d5 d4c5 a7a6 b5a4 d5f4');
INSERT INTO evaluations VALUES(56279,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(56279,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(56279,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(56279,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(56279,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(56279,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(56279,36,'cp -63','g8f6 d3e2 f8d6 g1f3 h7h6 d2d4 e8g8 e1g1 f8e8 b1d2 c8f5 f1e1 b8d7 d2f1 f5g6 f1g3 c7c5 e2d3 e8e1 d1e1 d8b6 d3g6 f7g6 c2c3 a8e8 e1d1 c5c4 d1c2 g8h7 c1e3 b6c6 a2a4 a7a6 g3f1 f6e4 f1d2 d7f6 d2e4 e8e4 g2g3 d6c7 a1e1 e4e8 e3d2 f6e4');
INSERT INTO evaluations VALUES(48739,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(48739,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(48739,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(48739,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(48739,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(48739,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(48739,36,'cp -63','d3e2 f8d6 d2d4 e8g8 g1f3 h7h6 e1g1 f8e8 b1d2 c8f5 f1e1 b8d7 d2f1 f5g6 e2d3 f6e4 c2c4 d5c4 d3c4 c7c6 f1g3 d8f6 c4d3 e4g3 e1e8 a8e8 h2g3 g6d3 d1d3 f6e6 c1e3 e6d5 a2a3 a7a6 e3d2 d7f6 a1e1 e8e1 d2e1 g7g6 e1c3 g8g7 f3d2 a6a5');
INSERT INTO evaluations VALUES(10467,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(10467,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(10467,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(10467,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(10467,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(10467,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(10467,36,'cp 8','d7d5 d3b5 e8g8 e1g1 d5e4 b5c6 e4f3 c6f3 c7c6 f1e1 d8d1 f3d1 f8d8 d1f3 d8e8 c1d2 c8f5 a1c1 a7a5 c3c4 h7h6 h2h4 a5a4 d2c3 f6d7 g2g4 f5e6 c1d1 f7f6 c3b4 d7f8 c4c5 f8g6 h4h5 g6f4 d1d6 g8f7 e1d1');
INSERT INTO evaluations VALUES(16161,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(16161,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(16161,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(16161,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(16161,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(16161,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(16161,36,'cp -19','d7d5 e4d5 f6d5 e1g1 d5c3 b2c3 f8d6 f3d2 e8g8 d2c4 c8f5 a1b1 b7b6 f2f4 e5f4 e2f3 d8d7 c4d6 c7d6 c1f4 a8c8 d1d2 f8e8 a2a4 c6e5 f3d5 f5e6 d5b3 e6b3 c2b3 d7c6 b1e1 e5d7 c3c4 d6d5 f4g3 d7f6 e1e8 c8e8 g3f2 d5c4 d3c4');
INSERT INTO evaluations VALUES(17145,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(17145,31,'cp 282',NULL);
INSERT INTO evaluations VALUES(17145,32,'cp 290',NULL);
INSERT INTO evaluations VALUES(17145,33,'cp 316',NULL);
INSERT INTO evaluations VALUES(17145,34,'cp 281',NULL);
INSERT INTO evaluations VALUES(17145,35,'cp 321',NULL);
INSERT INTO evaluations VALUES(17145,36,'cp 319','d4e5 c6e5 f3e5 f6e5 d1f3 d7d5 e4d5 e7g6 b1d2 f8d6 d2e4 a7a6 e1g1 h8f8 f3b3 e8f7 c1g5 d8d7 c4d3 b7b6 f1e1 f7g8 e4d6 d7d6 d3g6 h7g6 a1d1 b6b5 g5e3 f8f5 f2f3 g8h7 e3f2 c8d7 f2g3 a8e8 c2c4 b5c4 b3c4 d7b5 c4c3 f5h5');
INSERT INTO evaluations VALUES(17132,30,'cp 316',NULL);
INSERT INTO evaluations VALUES(17132,31,'cp 298',NULL);
INSERT INTO evaluations VALUES(17132,32,'cp 303',NULL);
INSERT INTO evaluations VALUES(17132,33,'cp 303',NULL);
INSERT INTO evaluations VALUES(17132,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(17132,35,'cp 306',NULL);
INSERT INTO evaluations VALUES(17132,36,'cp 304','c6e5 f3e5 f6e5 d1f3 d7d5 e4d5 e7g6 b1c3 f8d6 c3e4 a7a6 c1g5 d8d7 c4d3 h8f8 f3e2 e8f7 e1c1 f7g8 g5e3 g6f4 e3f4 e5f4 h1e1 f4f3 g2f3 d7e7 e4d6 e7d6 e2e5 g7g6 e5g3 c8f5 h2h4 f5d3 d1d3 a8e8 e1e8 f8e8 g3d6 c7d6 c1d2 b7b5 d3e3 e8f8 d2e2 g8g7 e3e7 f8f7');
INSERT INTO evaluations VALUES(20158,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20158,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20158,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20158,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(20158,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20158,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20158,36,'cp -2','f8d6 e1g1 g8e7 f1e1 a7a6 b5f1 e8g8 c3e2 e7g6 c2c3 c8d7 c1d2 h7h6 a1c1 a8c8 e2g3 b7b5 f1d3 d8c7 b2b3 e6e5 d4e5 g6e5 f3e5 c6e5 d3f5 f8e8 d2f4 g7g6 f5d7 e5d7 e1e8 c8e8 f4h6');
INSERT INTO evaluations VALUES(25999,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(25999,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(25999,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(25999,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(25999,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(25999,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(25999,36,'cp 23','f8g7 d4c6 d7c6 b5c6 b7c6 d1f3 d8c7 c1f4 a8b8 b2b3 g8f6 a1d1 e8g8 c3a4 f6d7 f3e3 f8e8 f4h6 g7h6 e3h6 d7f6 f1e1 b8d8 c2c4 e7e6 h6c1 e6e5 a4c3 f6h5 c1e3 h5f4 d1d2 f4e6 b3b4 c7b6 b4b5 c6b5 e3b6 a7b6');
INSERT INTO evaluations VALUES(23802,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(23802,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(23802,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(23802,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(23802,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(23802,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(23802,36,'cp 16','d4c6 d7c6 b5c6 b7c6 d1f3 d8c7 c1f4 a8b8 b2b3 g8f6 a1d1 e8g8 c3a4 c6c5 c2c4 f6d7 f3e3 g7d4 e3e1 d7e5 f4e3 e5c6 e1d2 f7f6 h2h3 c7b7 f2f4 a7a5 e3d4 c5d4 c4c5 b7b4 d2b4 b8b4 a4b2 d6c5 b2d3 b4b6 d3c5 f8d8 c5d3');
INSERT INTO evaluations VALUES(23798,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23798,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(23798,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(23798,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23798,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(23798,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(23798,36,'cp 11','d7c6 b5c6 b7c6 d1f3 d8c7 c1f4 a8b8 b2b3 g8f6 a1d1 e8g8 c3a4 f6d7 h2h3 c7b7 f1e1 c6c5 a4c3 c5c4 c3d5 f8e8 b3b4 e7e6 d5e3 d7e5 f4e5 g7e5 a2a3 b7c7 f3e2 a7a5 e2c4 c7c4 e3c4 a5b4 c4e5');
INSERT INTO evaluations VALUES(27918,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(27918,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(27918,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(27918,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(27918,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(27918,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(27918,36,'cp 132','d7d6 c1g5 g8e7 d1d2 b8d7 e1c1 h7h6 g5e3 d7f6 h2h3 d6d5 c4d5 e6d5 e4e5 f6e4 d2c2 e4c3 c2c3 d8d7 c1b1 e8c8 d1c1 c8b8 d3b5 c7c6 b5d3 b8a8 e3d2 d8c8 h1e1 h8d8 c3a3 g7f8 a3a4 c6c5 a4d7 d8d7 d4c5 b6c5 e5e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(27903,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(27903,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(27903,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(27903,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(27903,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(27903,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(27903,36,'cp 105','c1g5 g8e7 d1d2 h7h6 g5e3 f7f5 d4d5 e6e5 g2g3 b8a6 f3h4 d8d7 f2f3 e8c8 e1c1 c8b8 a2a3 f5f4 e3f2 h6h5 g3f4 e5f4 h4g2 a6c5 d3c2 g6g5 h2h4 g5g4 g2f4 d8f8 f3g4 h5g4 c1b1 g7h6 f2e3 g4g3 d1g1 d7g4 c3e2 g4f3 e2d4 c5e4 d4f3');
INSERT INTO evaluations VALUES(28022,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(28022,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(28022,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(28022,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(28022,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(28022,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(28022,36,'cp 46','g8f6 d1c2 e8g8 e3e4 b8c6 f1e2 e7e5 d4e5 f6g4 c1f4 c6e5 f3e5 g4e5 e1g1 d7d6 a1d1 e5d7 f1e1 f8e8 f2f3 d7c5 f4e3 a7a5 e2f1 g7e5 c3b5 a5a4 g2g3 d8e7 d1d2 e5g7 f1g2 b7c6 b5c3 e7d7 f3f4 f7f5 e4f5 c6g2 g1g2 d7f5 c2f5 g6f5');
INSERT INTO evaluations VALUES(28329,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(28329,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(28329,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(28329,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(28329,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(28329,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(28329,36,'cp 48','d1c2 e8g8 e3e4 b8c6 h2h3 d7d6 c1e3 c6b4 c2b1 e7e6 a2a3 b4c6 b1c2 c6e7 f1d3 d6d5 c4d5 e6d5 e4e5 f6d7 e1g1 c7c5 c3b5 b7a6 d4c5 b6c5 e3c5 a8c8 b2b4 d7c5 b4c5 a6b5 d3b5 d8c7 a3a4 c7c5 c2c5 c8c5 f1c1 c5c1 a1c1 f8c8 c1b1 e7c6 b5c6 c8c6 b1b8 g7f8');
INSERT INTO evaluations VALUES(28923,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(28923,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(28923,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(28923,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(28923,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(28923,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(28923,36,'cp 25','f1d3 h7d3 d1d3 e7e6 c1d2 f8e7 g3e4 f6e4 d3e4 d8d5 e4d5 c6d5 g2g4 b8c6 e1e2 a8c8 c2c3 a7a6 b2b3 e8g8 a2a4 c8c7 h1c1 e7d6 a1b1 c7c8 c3c4 d6e7 e2d3 f7f5 g4g5 h6g5 c4d5 e6d5 f3g5 f8f6 c1g1');
INSERT INTO evaluations VALUES(28922,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(28922,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(28922,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(28922,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(28922,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(28922,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(28922,36,'cp 11','h7d3 d1d3 e7e6 c1d2 b8d7 e1c1 f8e7 g3e4 e8g8 c1b1 c6c5 d4c5 d7c5 e4f6 e7f6 d3d8 f8d8 d2e3 d8d1 h1d1 c5a4 f3d4 a8d8 d1d2 a7a6 c2c3 a4b6 d4b3 b6d5 c3c4 d5e3 d2d8 f6d8 f2e3 d8b6 c4c5 b6c7 b3d4 g8h7 b2b4 g7g6 b4b5 a6b5 d4b5 c7b8 b5d6 b7b6 d6f7 b6c5 f7d8');
INSERT INTO evaluations VALUES(28912,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28912,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28912,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(28912,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(28912,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(28912,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(28912,36,'cp 16','d1d3 e7e6 c1d2 b8d7 e1c1 f8e7 g3e4 e8g8 e4f6 d7f6 c1b1 c6c5 g2g4 f6g4 d3e2 d8b6 h1g1 f7f5 f3e5 c5d4 e5g4 f5g4 d2h6 g7h6 e2g4 e7g5 f2f4 d4d3 c2c4 f8f5 f4g5 f5g5 c4c5 b6c5 g4e6 g8h7 g1f1 a8f8 e6d7 h7g8 f1f8 c5f8 d7d3 f8f5 d3f5 g5f5 d1d7 f5h5 d7b7 a7a5 b7a7 h5c5 a2a4 c5f5');
INSERT INTO evaluations VALUES(28911,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(28911,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(28911,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28911,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(28911,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(28911,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(28911,36,'cp 14','e7e6 c1d2 b8d7 e1c1 f8e7 g3e4 e8g8 e4f6 d7f6 c1b1 d8d5 c2c4 d5e4 d3e4 f6e4 d2e3 f8d8 f3e5 a8c8 h1e1 e7f6 e5g4 f6e7 e3h6 e4f2 g4f2 g7h6 f2g4 g8g7 g4e5 e7f6 b1c2 d8g8 e5g4 f6e7 e1e2 b7b5 g4e5 b5c4');
INSERT INTO evaluations VALUES(30868,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(30868,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(30868,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(30868,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(30868,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(30868,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(30868,36,'cp -27','b8d7 c2c3 e7e6 c1f4 g8e7 b1d2 e7g6 f4g3 h7h5 h2h4 d8b6 d1c2 f8d6 g3d6 b6d6 g2g3 a7a5 g1f3 d6a6 d2f1 e8g8 f1e3 f8b8 f3g5 d7f6 a2a3 g6e7 g5f3 e7c6 e3g2 f6g4 c2e2 a5a4 e2a6 a8a6');
INSERT INTO evaluations VALUES(26213,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(26213,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(26213,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(26213,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(26213,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(26213,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26213,36,'cp 0','g1e2 g8f6 e1g1 e7e6 c1f4 f6h5 b1d2 f8e7 f4e3 h5f6 e2c1 b7b5 c2c3 e8g8 c1d3 a7a5 a2a3 h7h6 d1e2 d7b6 d2b3 b6c4 e3f4 f6d7 e2g4 d7f6');
INSERT INTO evaluations VALUES(34016,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(34016,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(34016,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(34016,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(34016,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(34016,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(34016,36,'cp -19','f6e4 d1e2 d7d5 b1c3 e4c3 d2c3 b8d7 e5d7 c8d7 e2e7 f8e7 c1e3 d7f5 e1c1 f5e4 a2a4 e7d6 f1d3 g7g6 d3e4 d5e4 d1d5 e8c8 h1d1 f7f5 c3c4 d6e7 c2c3 h8e8 b2b4 d8d5 c4d5 e7f6 c1d2 a7a6 g2g3 e8d8');
INSERT INTO evaluations VALUES(34071,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(34071,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(34071,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(34071,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(34071,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(34071,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(34071,36,'cp -11','d1e2 d7d5 b1c3 e4c3 d2c3 b8d7 e5d7 c8d7 e2e7 f8e7 c1e3 d7f5 f1d3 f5d3 c2d3 e7f6 e1d2 h7h5 h1e1 e8d7 a2a4 a7a5 g2g3 h8b8 d2c2 b8e8 e3f2 g7g6 c2d2 f6g7 d2c2 a8d8 e1e8 d8e8 a1e1 e8e1 f2e1 b7b6 h2h3 g7f8 e1f2 d7e6 c3c4 d5c4 d3c4');
INSERT INTO evaluations VALUES(36856,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(36856,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36856,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36856,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36856,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36856,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(36856,36,'cp 12','a7a5 h2h3 c7c6 a1b1 h7h6 b2b3 d8e7 e2e4 c6c5 f3e1 c8d7 e1c2 b8c6 a2a3 c6d4 a3a4 g8h7 b1b2 d7c6 h3h4 d4c2 b2c2 e7d7 c2b2 b7b6 d1e2 a8e8 b2d2 f6g4 f2f3 g4f6 g3g4 d7d8 e2f2 f6d7 h4h5 g6g5 g2h3');
INSERT INTO evaluations VALUES(38001,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(38001,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(38001,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(38001,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(38001,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(38001,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(38001,36,'cp 63','d1d2 b8c6 b1c3 g8f6 f2f3 e8g8 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4h5 f6h5 d4g7 g8g7 e1c1 a5e5 h1g1 c8e6 c3e2 f8h8 g1g5 e5h2 f3f4 h2h4 f4f5 e6c4 g5h5 h8h5 d2c3 g7f8 c3c4 g6f5 c4b3 h4e4 e2g3 e4f4 c1b1 h5h2 b3b7 a8e8 f1b5 f4f2 d1c1 e8d8 b7c7 f2b6 c7b6 a7b6 a2a3');
INSERT INTO evaluations VALUES(39175,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(39175,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(39175,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(39175,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(39175,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(39175,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(39175,36,'cp 13','g8f6 e2e3 d7d6 g1e2 e6e5 g2g3 e8g8 f1g2 b8c6 d4d5 c6b8 e1g1 a7a5 e2c3 b8a6 d2b3 b7b6 e3e4 c8d7 d1d2 a5a4 b3c1 a6c5 c1d3 c5d3 d2d3 h7h5 h2h3 h5h4 g3g4 d8e7 d3e3 a8a5 g1h1 f8e8 a1e1 f6h7 e3d2 a5c5 b2b3');
INSERT INTO evaluations VALUES(39548,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(39548,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(39548,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(39548,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(39548,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(39548,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(39548,36,'cp 12','e7e6 d2d4 d7d5 e4d5 e6d5 c4d3 g8f6 g1f3 d8e7 d3e2 c7c6 c1f4 f6h5 f4e3 e8g8 e1g1 f8e8 f1e1 e7c7 f3e5 h5f6 b1d2 b8d7 e5d3 d7f8 e3f4 c7d8 a2a4 a7a5 h2h3 c8f5 d3e5 f8e6 f4h2 h7h5 d2f3 f6d7 e5d7 d8d7');
INSERT INTO evaluations VALUES(41611,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(41611,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(41611,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(41611,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(41611,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(41611,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(41611,36,'cp 79','e2e3 e8g8 a1c1 b8c6 f1e2 h7h6 g5h4 c6e7 h4g3 d5c4 e2c4 e7d5 c4d3 c7c5 c3d5 d8d5 c1c5 d5a2 e1g1 b7b6 c5c1 c8b7 f3e5 a2d5 f2f3 d5d8 f1e1 a7a5 d1a4 f6d5 e5d7 f8e8 d3b5 b7a6 b5a6 a8a6 c1c4 d8a8 g3e5 d5b4 e1c1 b4d3 e5g7 g8g7');
INSERT INTO evaluations VALUES(41610,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(41610,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(41610,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(41610,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(41610,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(41610,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(41610,36,'cp 84','e8g8 a1c1 h7h6 g5h4 c7c6 f1d3 d5c4 d3c4 b8d7 e1g1 c6c5 c4b3 c5d4 e3d4 b7b6 f1e1 c8b7 c3b5 a8c8 b3c4 b7a6 d1a4 d7b8 f3e5 a6b5 a4b5 g6g5 e5f7 f8f7 h4g3 d8d7 b5d7 f6d7 c4e6 c8c1 e1c1');
INSERT INTO evaluations VALUES(36882,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(36882,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(36882,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(36882,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(36882,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(36882,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(36882,36,'cp 93','a1c1 h7h6 g5h4 c7c5 c4d5 c5d4 f3d4 e6d5 f1b5 a7a6 b5a4 c8d7 e1g1 g6g5 h4g3 b8c6 a4c6 b7c6 g3e5 f8e8 e5f6 d8f6 c3a4 a8b8 a4c5 d7c8 d1a4 b8b2 c5d3 b2b7 c1c6 c8d7 c6f6 d7a4 f6a6 g7d4 a6a4');
INSERT INTO evaluations VALUES(42470,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(42470,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(42470,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42470,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(42470,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42470,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(42470,36,'cp 9','f1g2 b8c6 g1f3 e8g8 e1g1 d7d5 a2a3 b4c3 d2c3 h7h6 c1f4 f8e8 f1e1 a7a5 c4d5 e6d5 c3c4 d5d4 b2b4 a5b4 a3b4 a8a1 d1a1 c6b4 a1d4 d8d4 f3d4 b4d3 e1a1 d3f4 g3f4 e8d8 a1a8 c7c6 d4b3 d8e8 b3c5 g7g5 f4g5 h6g5 e2e3 f6d7 c5d7');
INSERT INTO evaluations VALUES(42471,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42471,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(42471,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(42471,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(42471,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(42471,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42471,36,'cp 0','e8g8 g1f3 d7d5 a2a3 b4c3 d2c3 b8d7 c4d5 e6d5 d1c2 f8e8 e1g1 d7b6 b2b3 a7a5 a3a4 h7h6 f1e1 f6e4 f3d2 e4g5 d2f3');
INSERT INTO evaluations VALUES(37170,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37170,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(37170,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(37170,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(37170,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37170,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37170,36,'cp 5','g1f3 d7d5 a2a3 b4c3 b2c3 b8c6 e1g1 c6a5 c4d5 e6d5 d2d3 c7c5 f1e1 f8e8 h2h3 b7b6 g3g4 h7h6 c1f4 c8e6 f4g3 a8c8 g3h4 d5d4 c3d4 a5b3 d4d5 e6d5 e2e4 d5e6 a1b1 b3d4 h4g3 d4f3 d1f3 d8d4 b1d1 c8d8 f3f4 d8d7');
INSERT INTO evaluations VALUES(45041,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(45041,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(45041,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(45041,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(45041,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(45041,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(45041,36,'cp 143','e4e5 f6d5 c2c4 d5c7 b1c3 g7g6 c1e3 f8g7 d1d2 b7b5 c4b5 c6b5 g1f3 e8g8 c3b5 b8a6 b5c3 c8b7 f1c4 b7f3 g2f3 a6b4 a1d1 b4d5 e3g5 d5c3 b2c3 c7e6 g5e3 d8a5 c4b3 f8d8 c3c4 a5a3 d2e2 a7a5 e1g1 a5a4 b3c2 d8c8');
INSERT INTO evaluations VALUES(44995,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(44995,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(44995,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(44995,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(44995,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(44995,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(44995,36,'cp 183','f6d5 c2c4 d5c7 b1c3 g7g6 c1e3 f8g7 d1d2 b7b5 c4b5 c6b5 g1f3 c8b7 c3b5 b8a6 h2h4 h7h5 f1e2 e8g8 e1g1 c7b5 e2b5 a8c8 b2b3 a6c7 b5c4 d8d7 a1e1 b7d5 e3g5 d5c4 b3c4 c7e6 c4c5 f8d8 g5e3 d7d5 e1c1 c8c7 a2a4 c7b7 e3f2 d5e4');
INSERT INTO evaluations VALUES(46498,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(46498,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(46498,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(46498,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(46498,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(46498,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(46498,36,'cp -21','b8c6 d2d4 e5e4 f3e5 c8e6 e1g1 a8c8 e5c6 b7c6 f2f3 e4f3 d1f3 f8e7 c1g5 e8g8 h2h3 d8d7 a1e1 c8e8 b3a4 a7a5 f3g3 f6h5 g3h4 f7f6 h4h5 f6g5 f1f8 e7f8 h5g5 f8b4 e1f1 b4d6 c3e2 e6h3 e2g3 d6e7 g5f4 h3e6 g3f5 e6f5 f4f5 d7f5');
INSERT INTO evaluations VALUES(13803,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(13803,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(13803,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(13803,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(13803,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(13803,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(13803,36,'cp -18','d2d4 e5e4 f3e5 c8e6 e1g1 a8c8 f2f3 c6d4 b3a4 d4c6 f3e4 f8c5 g1h1 e8g8 e4d5 f6d5 e5c6 b7c6 c3e4 c5e7 d1f3 f7f5 e4c3 d5c3 b2c3 e6d5 a4b3 g8h8 b3d5 d8d5 c1e3 a7a5 e3b6 a5a4 a1e1 e7g5 f3d5 c6d5 e1e5 g5f6 e5d5 c8c3 f1f5 c3c2 f5f2 c2f2 b6f2');
INSERT INTO evaluations VALUES(47708,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(47708,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(47708,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47708,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(47708,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(47708,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47708,36,'cp 11','d1e2 c8e6 g5f6 d8f6 e2b5 b8d7 b1c3 f6d8 b5b7 a8b8 b7a7 f8d6 f1e2 b8b2 c3b5 b2c2 e2d3 c2b2 b5d6 c7d6 e1g1 e8g8 a7a3 d8b8 f1b1 b2b1 a1b1 b8a8 a3d6 a8a2 h2h3 d7f6 d6f4 f8c8 f3e5 a2a7 f4e3 a7c7 e5f3 f6h5 b1e1 c7f4 e3f4 h5f4');
INSERT INTO evaluations VALUES(48099,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(48099,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(48099,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(48099,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(48099,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(48099,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(48099,36,'cp 26','f8e7 g1f3 e8g8 d2d4 b7b6 c3d5 e6d5 c1f4 e7b4 f4d2 b4d6 e1g1 c8b7 a1c1 f8e8 d2f4 d6f4 g3f4 b8d7 f3e5 c7c5 e2e3 d7e5 f4e5 a8c8 f2f4 f7f6 g1h1 d8e7 d1a4 b7a8 a4a6 g8h8 a6e2 c8d8 e2d3 a8b7 c1c2');
INSERT INTO evaluations VALUES(48506,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(48506,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(48506,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(48506,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(48506,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(48506,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(48506,36,'cp 140','e2e4 f8e7 d1c2 e8g8 c1f4 c7c5 d4c5 b8c6 a1d1 d8a5 f4d6 e7d6 c5d6 f8d8 f1b5 c8d7 c2e2 a7a6 b5c6 d7c6 e1g1 e6e5 d1d3 d8d7 f1d1 a8e8 f3h4 g7g6 e2e3 a5d8 h4f3 e8e6 h2h3 e6d6 f3e5 d6d3 d1d3 d7d3 e5d3 d8e7 e3c5 e7c5 d3c5 b7b6 c5a6 f6e4 c3e4 c6e4 a6c7 e4b7');
INSERT INTO evaluations VALUES(49700,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(49700,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(49700,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(49700,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(49700,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(49700,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(49700,36,'cp -32','d2d4 e5d4 e3d4 d7d5 g1f3 f8d6 d1e2 d6e7 f1g2 c7c6 e1g1 e8g8 f1e1 f8e8 e2d3 c8e6 c1f4 b8d7 b1d2 h7h6 a2a3 a7a5 c2c4 d5c4 d2c4 a5a4 a1c1 e6d5 h2h4 d7b6 f3e5 b6c4 e5c4 d5g2 g1g2 f6d5 d3f3 e7f8 e1e8 d8e8 c4e3');
INSERT INTO evaluations VALUES(50768,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(50768,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(50768,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(50768,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(50768,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(50768,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(50768,36,'cp 120','f1d3 f8e7 c1e3 g8f6 b1c3 b8c6 e1g1 e8g8 a1c1 f6g4 e3d2 e6e5 d4d5 c6d4 f3d4 e5d4 c3e2 d8b6 d1b3 b6b3 a2b3 g4e5 d3b1 d4d3 e2g3 g7g5 f1d1 c8d7 d2c3 b7b5 b1d3 e5d3 d1d3 a6a5 d3d1 f8c8 h2h3 b5b4 c3d4 g5g4 g3h5 g4h3 h5f6 e7f6 d4f6 c8c1 d1c1');
INSERT INTO evaluations VALUES(52185,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(52185,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(52185,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52185,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(52185,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(52185,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(52185,36,'cp -16','g8f6 c1g5 d7d5 e2e4 d5d4 c2c3 b8c6 c3d4 e5d4 f1b5 f8b4 b1d2 h7h6 g5f6 d8f6 a2a3 b4e7 d1f3 f6e5 b5c6 b7c6 f3g3 e5g3 h2g3 e8g8 b3d4 e7f6 d4c6 c8d7 c6b4 f6b2 a1a2 b2d4 e1g1 f8e8 b4d5 a8c8 d2b3 d4c5 b3c5 c8c5 f2f3');
INSERT INTO evaluations VALUES(54761,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(54761,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(54761,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(54761,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(54761,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(54761,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(54761,36,'cp 49','g8f6 c1f4 e7e6 e2e3 a7a6 f1d3 c7c5 d4c5 f8c5 e1g1 b8c6 c3e2 d8e7 e2d4 e8g8 c2c4 c8d7 a2a3 c5a7 f1e1 a8c8 a1c1 h7h6 b2b4 d5c4 d3c4 c6d4 e3d4 f6e4 d1e2 d7e8 f4e5 e8h5 e2a2 h5f7 b4b5 a6b5 c4b5 c8c1 e1c1 a7b8 a2b3 f8d8');
INSERT INTO evaluations VALUES(46600,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(46600,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(46600,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(46600,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(46600,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(46600,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(46600,36,'cp 54','c1f4 e7e6 e2e3 a7a6 f1d3 c7c5 d4c5 f8c5 e1g1 b8c6 c3e2 d8e7 e2d4 c8d7 c2c4 c5a7 a2a3 a8c8 a1c1 e8g8 d4c6 d7c6 b2b4 d5c4 d3c4 c6d5 c4d5 f6d5 d1d3 h7h6 f4e5 a7b8 e5b2 f8d8 c1c8 d8c8 f1d1 d5b6 b2d4 b6d7 b4b5 d7f8 b5a6 b7a6');
INSERT INTO evaluations VALUES(55164,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(55164,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(55164,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(55164,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(55164,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(55164,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(55164,36,'cp -53','c8d7 a4b3 g8f6 g1f3 c7c5 f3e5 d7c8 d2d4 e7e6 c1f4 f6h5 f4e3 f8d6 g2g3 b8d7 e5d7 c5c4 b3d1 c8d7 e3c1 h5f6 f1g2 d8c7 c1g5 f6e4 g2e4 d5e4 b1d2 h7h6 g5f4 d6f4 g3f4 g7g5 e2e3 g5f4 d1g4 f4e3 f2e3 e8c8 g4g3');
INSERT INTO evaluations VALUES(56246,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(56246,31,'cp -102',NULL);
INSERT INTO evaluations VALUES(56246,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(56246,33,'cp -138',NULL);
INSERT INTO evaluations VALUES(56246,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(56246,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(56246,36,'cp -127','d5d4 c3e2 g7g5 f3d4 f8g7 c2c3 c7c5 d4f3 g5g4 e2f4 g4f3 d1f3 b8c6 f1b5 g8e7 e1g1 e8g8 f4h5 g7e5 h5f6 e5f6 f3f6 e7g6 f6f2 d8d6 b5c6 b7c6 d2d3 f7f5 c1h6 f8f7 h2h4 f5f4 h4h5 f4f3 h5g6 d6g6 f2d2 c8h3 d2g5 h3g2 g1f2 g6g5 h6g5 g2f1 a1f1 c5c4');
INSERT INTO evaluations VALUES(56507,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(56507,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(56507,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(56507,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(56507,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(56507,35,'cp -101',NULL);
INSERT INTO evaluations VALUES(56507,36,'cp -119','c8d7 a4b3 g8f6 d2d4 b8c6 g1f3 f8d6 c3c4 d5c4 f1c4 e8g8 f3e5 c6e5 d4e5 d6e5 b1d2 b7b5 c4e2 a7a5 b3c2 a5a4 e1g1 e5d6 b2b3 a4b3 c2b3 d8e7 c1b2 a8a4 f1d1 c7c5 e2b5 a4b4 b2f6 e7f6 b5d7 b4b3 a2b3 d6h2 g1h2 f6d6 h2g1 d6d7 d2e4 d7e7 e4g3 g7g6 a1c1 h7h5 c1c4 g8g7');
INSERT INTO evaluations VALUES(32518,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(32518,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(32518,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(32518,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(32518,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(32518,35,'cp -102',NULL);
INSERT INTO evaluations VALUES(32518,36,'cp -83','a4b3 g8f6 g1f3 b8c6 d2d4 f8d6 c3c4 d5c4 f1c4 e8g8 d4e5 c6e5 f3e5 d6e5 b1d2 b7b5 c4e2 a7a5 b3c2 a5a4 d2f3 e5d6 e1g1 f8e8 c1d2 b5b4 f1d1 a4a3 b2b3 d8e7 a1c1 a8a5 d2e1 a5c5 e2c4 c5h5 c2e2 h7h6 h2h3 d7h3 g2h3 h5h3 d1d4 c7c5');
INSERT INTO evaluations VALUES(1762,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(1762,31,'cp 245',NULL);
INSERT INTO evaluations VALUES(1762,32,'cp 260',NULL);
INSERT INTO evaluations VALUES(1762,33,'cp 249',NULL);
INSERT INTO evaluations VALUES(1762,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(1762,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(1762,36,'cp 247','d4a4 c7c6 c1e3 f6d8 e1c1 c8d7 h2h3 b7b5 a4d4 d8b6 d4d3 b6c7 g2g4 g8e7 d3d6 c7d6 d1d6 b5b4 c3a4 e7f5 d6d7 f5e3 d7c7 e3f1 h1f1 h7h5 c7c6 h5g4 h3g4 h8h2 a4c5 a7a5 c1d2 h2g2 d2e3 a5a4 f1d1 g2g4 c5d3 g4h4');
INSERT INTO evaluations VALUES(7898,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(7898,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(7898,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(7898,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(7898,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(7898,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(7898,36,'cp -18','b1d2 e8g8 c4b5 c6a5 b5a4 h7h6 g5f4 d7d5 h2h3 c8d7 a4d7 f6d7 e1g1 a5c6 f1e1 f8e8 c2c3 d8b6 d1c2 d7f8 a2a3 a8c8 f4h2 b6a6 h3h4 f8d7 e4d5 e6d5 d2f1 b7b5 h2g3 e7f6 e1e8 c8e8 a1e1');
INSERT INTO evaluations VALUES(13158,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(13158,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(13158,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(13158,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(13158,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(13158,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(13158,36,'cp 17','f3e5 c8f5 f1b5 d8b6 c3c4 d5c4 d1a4 a8c8 e1g1 e7e6 b1a3 a7a6 b5c6 b7c6 e5c4 b6b7 c1f4 f8e7 f1c1 e8g8 c4d6 e7d6 f4d6 f8d8 a3c4 f5d3 a4a3 d3c4 c1c4 f6d5 h2h3 h7h6 g1h1 a6a5 a1c1 f7f6 d6g3 d5e7 c4c5 d8d4 c5a5 g8f7');
INSERT INTO evaluations VALUES(13319,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(13319,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13319,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(13319,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13319,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13319,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13319,36,'cp 0','h2h3 e7e5 d3c2 d6d5 e4d5 d8d5 d1e2 f8e7 d2d3 c8f5 b1d2 a8d8 d2c4 f5d3 e2d3 d5d3 c2d3 d8d3 f3e5 c6e5 c4e5 d3d6 e1g1 f6d5 e5c4 d6d7 f1d1 d5b6 d1d7 e8d7 c4a5 b6a4 c1e3 b7b6 a1d1 d7e6 d1e1 e6d6 e1d1');
INSERT INTO evaluations VALUES(15413,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(15413,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(15413,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(15413,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(15413,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(15413,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(15413,36,'cp -38','a7a6 b5a4 b7b5 a4b3 c6a5 d2d4 a5b3 a2b3 f6d5 e4d5 f7f6 c2c4 f8e7 e1g1 e8g8 f1e1 c8b7 c1d2 c7c6 d5c6 b7c6 d1e2 d8c8 d4d5 c6d7 a1c1 f8e8 d2a5 e7f8 e2d3 c8b7 f3h4 b5c4 b3c4 b7b2 h4f5 d7f5 d3f5 b2a2');
INSERT INTO evaluations VALUES(15573,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(15573,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(15573,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(15573,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(15573,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15573,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(15573,36,'cp 37','f1g2 g7g6 d2d4 f8g7 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 a2a4 c7c6 a4a5 c8e6 c3a2 f6d7 a2b4 d7c5 e1g1 d8c7 f2f4 h8d8 c1e3 g7f8 f4e5 d8d7 f1d1 d7d1 a1d1 a8d8 d1d8 c7d8 b4d3 c5d3 c2d3 f8b4 e3a7 b4a5 d3d4 d8c7');
INSERT INTO evaluations VALUES(15618,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(15618,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(15618,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(15618,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(15618,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(15618,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(15618,36,'cp 27','f8e7 f1b5 c8d7 b5c6 b7c6 d1d3 e8g8 e1c1 f8e8 h1e1 h7h6 g5h4 d8b8 f2f4 b8b6 e4e5 f6e4 h4e7 e4c3 d3c3 e8e7 h2h3 c6c5 d4f3 a8e8 a2a3 a7a5 f3d2 d6e5 d2f3 d7c6 e1e5 c6f3 e5e7 e8e7 c3f3 e7e8 g2g4 c5c4 c1b1 a5a4 g4g5');
INSERT INTO evaluations VALUES(8638,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(8638,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8638,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(8638,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(8638,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(8638,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(8638,36,'cp 38','f1b5 c8d7 e1g1 h7h6 g5e3 c6d4 b5d7 f6d7 e3d4 e8g8 d1g4 e7f6 a1d1 f8e8 g4g3 b7b5 f2f3 f6d4 d1d4 d7f6 f1d1 d8b8 g3f2 e8e5 a2a4 a7a6 a4a5 b5b4 c3d5 f6d5 d4d5 b8e8 b2b3 a8d8 d5d4 e5a5 d4b4 d6d5 f2d2');
INSERT INTO evaluations VALUES(21909,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(21909,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21909,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21909,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(21909,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(21909,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21909,36,'cp 23','f8e7 d2d3 g8f6 e1g1 e8g8 f1e1 c8e6 b1c3 e6c4 d3c4 d8c8 c1e3 c6d8 c3d5 f6d5 c4d5 f7f5 e4f5 c8f5 c2c4 d8f7 f3d2 e7g5 d1b1 f5b1 a1b1 g5e3 e1e3 f7h6 e3e1 h6f5 b2b4 b7b6 d2e4 a7a5 c4c5 a5b4 a3b4');
INSERT INTO evaluations VALUES(22430,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(22430,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(22430,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(22430,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(22430,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(22430,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(22430,36,'cp -57','e8f7 f3e5 f7e8 c2c3 d4e6 d2d4 g8f6 e1g1 d7d6 e5d3 f6e4 d4d5 e6c5 d3f4 e4f6 b2b4 c5a6 b1d2 e8f7 d2f3 f7g8 f3d4 c7c5 d5c6 b7c6 d4c6 d8d7 b4b5 a6c7 c1e3 c8b7 d1b3 d7f7 c3c4 h7h6 a2a4 d6d5 a1c1 d5c4 b3c4 f7c4 c1c4 c7d5 f4e6 d5e3 f2e3 b7c6 c4c6');
INSERT INTO evaluations VALUES(2504,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(2504,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(2504,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(2504,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(2504,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(2504,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(2504,36,'cp -95','f3e5 f7e8 c2c3 d4c6 d1h5 g7g6 e5g6 g8f6 h5h4 h8g8 g6f8 d8e7 f8h7 f6h7 h4e7 e8e7 g2g3 d7d5 e4d5 c6e5 d2d4 e5d3 e1e2 c8f5 f2f3 a8e8 g3g4 d3c1 h1c1 e7d7 e2d2 f5b1 c1b1 h7g5 b1f1 g8h8 a1e1 h8h2 d2d3 e8e1 f1e1 g5f3 e1e2 h2h1 e2e6 f3g5 e6g6 h1h3 d3c2 g5f3 g6g7 d7d6 c2d3 b7b5 d3e2 a7a5 g7f7 f3h4 e2f2 a5a4 g4g5 d6d5 f7c7 d5d6 c7g7');
INSERT INTO evaluations VALUES(2503,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(2503,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(2503,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(2503,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(2503,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(2503,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(2503,36,'cp -92','f7e8 c2c3 d4c6 d1h5 g7g6 e5g6 g8f6 h5h4 h8g8 g6f8 d8e7 f8h7 f6h7 h4e7 e8e7 d2d3 g8g2 c1f4 d7d6 b1d2 g2g4 f4g3 h7f6 e1e2 c8e6 a1g1 a8h8 f2f3 g4g8 d3d4 e6f7 e2f2 f6h5 g3h4 e7d7 g1g8 h8g8 b2b3 b7b5 h1g1 g8g1 f2g1 h5f4 g1f2 c6e7 h4g3');
INSERT INTO evaluations VALUES(23336,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(23336,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(23336,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(23336,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(23336,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(23336,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(23336,36,'cp -86','c2c3 d4c6 d1h5 g7g6 e5g6 g8f6 h5h4 h8g8 g6f8 d8e7 f8h7 f6h7 h4e7 e8e7 d2d3 g8g2 c1f4 g2g4 f4g3 d7d6 b1d2 h7f6 e1e2 c8d7 a1g1 a8g8 e2e3 g4g7 f2f3 g7h7 h2h4 e7f7 d3d4 b7b6 c3c4 c6e7 c4c5 b6c5 d4c5 d6c5 b2b3 h7h5 d2c4 e7c6 g3c7 g8g1 h1g1 h5h4');
INSERT INTO evaluations VALUES(22866,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(22866,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(22866,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(22866,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(22866,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(22866,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(22866,36,'cp -19','d7d5 g1f3 g8f6 e1g1 c8e6 b1d2 a7a5 b2b3 f8e7 c1b2 d8d6 e2e3 e8g8 a2a3 a8d8 d3d4 e5d4 e3d4 f6e4 f1e1 f7f5 c2c4 e4d2 d1d2 d5c4 d2e2 c4c3 b2c3 e6b3 f3e5 c6e5 d4e5 d6e6 c3a5 b7b6 a5b4 c7c5 b4c3 b3d5 g2d5 e6d5 e5e6 d5c6');
INSERT INTO evaluations VALUES(23833,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(23833,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(23833,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(23833,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(23833,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(23833,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(23833,36,'cp 39','b1c3 c6e5 d1a4 e5c6 f3e5 b4c3 d2c3 d8g5 e5g4 g5g4 f1e2 g4g2 e1c1 g8f6 a4c2 a8d8 h1g1 d8d1 c2d1 g2h2 e2f3 h2d6 g1g7 d6d1 c1d1 e8e7 g7g5 h7h5 g5b5 c6d8 f3b7 h5h4 b5e5 d8e6 e5f5 h8h6 d1e2 f6d7 f5a5 h4h3 b7h1 h3h2 b2b3 h6h3 a5a7 f7f5 c3a5 f5f4 e3f4 e6f4 e2f1');
INSERT INTO evaluations VALUES(27274,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(27274,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(27274,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(27274,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27274,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(27274,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(27274,36,'cp 46','g7g6 h2h3 d7b5 c3b5 a7a6 b5c3 f8g7 d1d3 g8f6 c1g5 h7h6 g5f6 g7f6 e1c1 e7c8 c1b1 h6h5 f3d2 c8b6 d2b3 h5h4 b3a5 a8b8 d3e2 f6g5 g2g3 d8e7 a2a3 b6d7 e2g4 g5h6 h1g1 d7f6 g4f3 h6g5 g1h1 e8g8 h1g1 g8g7');
INSERT INTO evaluations VALUES(28764,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(28764,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(28764,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(28764,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(28764,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(28764,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(28764,36,'cp 92','f1d3 f8e7 h2h4 d7d6 g1f3 a7a6 d1e2 b8d7 e4e5 d6e5 d4e5 f6d5 c3e4 b6b5 e1g1 c7c5 b2b3 b5b4 a3b4 c5b4 d3a6 e8g8 c1g5 a8a6 a1a6 b7a6 e2a6 e7g5 e4g5 d8e7 a6d3 g7g6 d3c4 d7c5 f1e1 h7h6 g5e4 c5e4 e1e4 e7d7 c4e2');
INSERT INTO evaluations VALUES(29924,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(29924,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(29924,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(29924,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(29924,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(29924,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(29924,36,'cp 79','b8d7 d2d3 f8e7 c1e3 e8g8 b1d2 d7c5 a1d1 c5e6 d2b3 c7c5 c2c3 d8c7 a2a4 c7d7 c4b5 d7c7 g2g3 a7a6 b5c4 b7b6 b3d2 c7b7 c4b3 b6b5 d1a1 b7d7 h3h4 h7h6 f1e1 a8b8 a4b5 a6b5 e1d1 d7b7 d2f1 b7d7 g1g2 b8a8 a1a8 f8a8 d1b1');
INSERT INTO evaluations VALUES(34027,30,'cp 278',NULL);
INSERT INTO evaluations VALUES(34027,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(34027,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(34027,33,'cp 257',NULL);
INSERT INTO evaluations VALUES(34027,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(34027,35,'cp 276',NULL);
INSERT INTO evaluations VALUES(34027,36,'cp 259','d1e2 f6g8 d2d4 e7h4 e1d1 h4d8 d4d5 f8b4 c3e4 d7d6 e2b5 d8d7 b5b4 d6e5 c2c4 g8f6 e4f6 g7f6 g2g3 b8c6 b4a3 b7b5 f1h3 d7d8 h3c8 a8c8 a3c5 c6e7 g1e2 b5c4 e2c3 d8d6 c5d6 c7d6 g3f4 c8b8 c3e4 e7d5 e4d6 e8e7 d6c4 h8d8 c1d2 d5f4 d1c2 f4d5 a1e1 e7e6 h1f1 d5b4 c2b1');
INSERT INTO evaluations VALUES(34028,30,'cp 302',NULL);
INSERT INTO evaluations VALUES(34028,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(34028,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(34028,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(34028,34,'cp 295',NULL);
INSERT INTO evaluations VALUES(34028,35,'cp 289',NULL);
INSERT INTO evaluations VALUES(34028,36,'cp 284','f6g8 d2d4 e7h4 e1d1 h4d8 d4d5 f8b4 c3e4 d7d6 e2b5 d8d7 b5b4 d6e5 c2c4 b8a6 b4c3 f7f6 g1f3 g8e7 b2b3 c7c6 c1a3 e8g8 a3e7 d7e7 d5d6 e7d8 c4c5 a6c7 f1c4 c7d5 c3d2 c8e6 h1e1 d8d7 a2a4 d5e3 e1e3 e6c4 b3c4 f4e3 d2e3 d7g4 d1d2 g4g2 d2c3 a7a5 e3f2');
INSERT INTO evaluations VALUES(35265,30,'cp 307',NULL);
INSERT INTO evaluations VALUES(35265,31,'cp 283',NULL);
INSERT INTO evaluations VALUES(35265,32,'cp 288',NULL);
INSERT INTO evaluations VALUES(35265,33,'cp 310',NULL);
INSERT INTO evaluations VALUES(35265,34,'cp 289',NULL);
INSERT INTO evaluations VALUES(35265,35,'cp 317',NULL);
INSERT INTO evaluations VALUES(35265,36,'cp 300','d2d4 e7h4 e1d1 h4d8 d4d5 f8b4 c3e4 d7d6 e2b5 d8d7 b5b4 d6e5 c2c4 g8f6 e4f6 g7f6 g2g3 b8c6 b4a3 b7b5 f1h3 d7d8 h3c8 a8c8 a3c5 c6e7 g1e2 b5c4 e2c3 d8d6 c5d6 c7d6 g3f4 c8c5 c3e4 c5d5 d1c2 d5d3 e4f6 e8d8 f4e5 d6e5 h1d1 e7f5 d1d3 c4d3 c2d3 d8e7 f6e4 f7f6 c1e3 f5e3 d3e3 f6f5 e4c5 e7d6 c5d3 d6e6 a1e1');
INSERT INTO evaluations VALUES(38275,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(38275,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(38275,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(38275,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(38275,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(38275,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(38275,36,'cp 78','d1f3 g8f6 f3g3 f6h5 g3h3 h5f6 e4e5 f6d5 c1d2 d5c3 d2c3 d6d5 c3d2 c7c5 h3g3 g7g6 d4c5 b8d7 f1b5 a7a6 b5d7 c8d7 g1f3 h7h5 d2e3 d8a5 c2c3 a5b5 e3g5 h5h4 f3h4 e7g5 g3g5 h8h5 g5g3 b5a4 h4f3 a4e4 e1d2 d7b5 a2a4 b5e2 b2b4 e2f3 g3f3 e4f3');
INSERT INTO evaluations VALUES(44732,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44732,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44732,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44732,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44732,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44732,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44732,36,'cp 0','e7e6 d2d4 c5d4 d1d4 b8c6 d4d1 a7a6 a2a4 d8c7 c4b3 h7h6 g5f3 b7b6 c1f4 c8b7 d1e2 g6g5 f4e3 f6g4 e3d2 g4e5 d2e3');
INSERT INTO evaluations VALUES(45528,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45528,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45528,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45528,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45528,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45528,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45528,36,'cp 0','d2d4 c5d4 d1d4 a7a6 a2a4 b8c6 d4d2 f8g7 e1g1 e8g8 f1d1 d8a5 c4b3 h7h6 g5f3 f8d8 d2f4 g6g5 f4g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(45717,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(45717,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(45717,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(45717,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(45717,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(45717,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(45717,36,'cp 171','e4f6 g7f6 f1d3 c8d7 e1g1 c6c5 f1e1 c5d4 f3d4 b8c6 d3e4 h7h5 d4c6 d7c6 e4c6 b7c6 d1f3 d8d5 f3f6 h8g8 g2g3 f8e7 f6e5 h5h4 c1f4 d5e5 f4e5 f7f6 e5c3 e6e5 g1g2 h4g3 h2g3 e8c8 e1e4 d8d5 a1e1 e7c5 a2a3 c8d7 e4h4');
INSERT INTO evaluations VALUES(45773,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(45773,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(45773,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(45773,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(45773,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(45773,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(45773,36,'cp -36','f8d6 g1f3 e8g8 e1g1 f8e8 d1c2 h7h6 c1f4 d6f4 g3f4 f6e4 c3e4 e8e4 e2e3 c8f5 g1h1 b8d7 c2b3 a8b8 f1g1 e4e6 f3e5 d7e5 f4e5 d8e7 b3c3 e6g6 f2f3 b8e8 a1f1 c6c5 f3f4 e8c8 g2d5 c5d4 c3d4 g6g1 f1g1 c8c2 d5e4 f5e4 d4e4');
INSERT INTO evaluations VALUES(40521,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(40521,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(40521,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(40521,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(40521,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(40521,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(40521,36,'cp -31','g1f3 e8g8 e1g1 f8e8 d1c2 b8d7 f1e1 h7h6 c1f4 d6f4 g3f4 f6e4 c3e4 d5e4 f3e5 d7f6 e2e3 c8f5 a1d1 a7a5 a2a3 e8e7 c2c5 g7g6 c5c2 g8h7 h2h3 d8d6 g1h2 a8g8 e1g1 g8g7 c2c5 d6d8 g2f1 a5a4 f1c4 f6d5 c4d5');
INSERT INTO evaluations VALUES(40520,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(40520,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(40520,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(40520,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(40520,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(40520,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(40520,36,'cp -26','e8g8 e1g1 f8e8 d1c2 b8a6 a2a3 h7h6 c1f4 d6f4 g3f4 f6e4 c3e4 d5e4 f3e5 f7f6 e5c4 d8e7 g1h1 a6c7 c4e3 c7d5 e3d5 c6d5 c2b3 e7f7 f1g1 g8h8 b3g3 c8f5 g2h3 f5h3 g3h3 a8c8 a1c1 f6f5 e2e3 e8g8 h3g3 f7e7 c1c8 g8c8 g3g2');
INSERT INTO evaluations VALUES(36120,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(36120,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(36120,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(36120,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(36120,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(36120,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(36120,36,'cp -25','e1g1 c8f5 f3e5 b8d7 c1f4 d6c7 f1e1 h7h6 e5d7 d8d7 f4c7 d7c7 d1d2 f8e8 f2f3 a8d8 a1c1 c7d7 c3a4 e8e7 a4c5 d7c8 e2e4 d5e4 f3e4 f5h3 g2h3 c8h3 e4e5 f6d5 c5e4 h3g4 e4d6');
INSERT INTO evaluations VALUES(36119,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(36119,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(36119,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(36119,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(36119,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(36119,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(36119,36,'cp -25','f8e8 d1c2 b8a6 a2a3 h7h6 c1f4 d6f4 g3f4 f6e4 c3e4 d5e4 f3e5 f7f6 e5c4 d8e7 c4e3 a6c7 f4f5 g8h8 c2c5 c7d5 c5e7 e8e7 e3d5 c6d5 g2h3 g7g6 f2f3 e4f3 f1f3 e7e2 f5g6 c8h3 f3h3 a8e8');
INSERT INTO evaluations VALUES(46090,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46090,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(46090,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(46090,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(46090,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(46090,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46090,36,'cp 0','f8c5 c4d5 e6d5 a2a3 b8c6 g1f3 e8g8 b2b4 d5d4 c3b5 c5b6 b5d4 c6d4 f3d4 c8g4 f1e2 g4e2 d1e2 b6d4 e3d4 a7a5 b4a5 f8e8 c1e3 d8a5 e2d2 a5a6 d2e2');
INSERT INTO evaluations VALUES(47780,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47780,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47780,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47780,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47780,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47780,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47780,36,'cp 0','h7h6 g5e3 e5d4 d1d4 b8c6 d4d3 d6d5 e4e5 f6e4 b1d2 e4c5 d3c3 c5e4');
INSERT INTO evaluations VALUES(54011,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(54011,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(54011,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(54011,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(54011,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(54011,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(54011,36,'cp 73','g8f6 f1c4 e6d5 c3d5 f8e7 g1e2 f6d5 c4d5 b8a6 e1g1 a6c7 e2f4 e8g8 a2a4 c7d5 f4d5 f8e8 f2f3 c8e6 c2c3 d8d7 c1f4 e7f8 f1f2 f7f5 e4f5 e6f5 f2d2 d7f7 f4g3 a8d8 a4a5 f5e6 h2h3 h7h6 c3c4 e6d5 d2d5 f8e7 d1c2 e7g5 d5d6 d8d6');
INSERT INTO evaluations VALUES(54623,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(54623,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(54623,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(54623,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(54623,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(54623,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(54623,36,'cp 54','g8f6 f1d3 f8d6 e1g1 e8g8 c3e2 d8e7 b2b3 c8d7 d1c2 d7e8 e2f4 b8d7 c1b2 a8d8 d3e2 e8h5 f4h5 f6h5 f3e5 h5f6 a1d1 g8h8 f2f3 g7g5 c4d5 c6d5 e2d3 g5g4 f3g4 d6e5 d4e5 f6g4 b2d4 e7g7 d3e2 f8g8 e2g4 g7g4 g1h1');
INSERT INTO evaluations VALUES(55533,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(55533,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(55533,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55533,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(55533,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(55533,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55533,36,'cp -38','g8f6 b1c3 d6d5 c3d5 f6d5 e4d5 f8b4 b2e5 e8g8 f1e2 f8e8 e5f4 c8g4 f4e3 g4e2 g1e2 d8d5 e1g1 b7b6 c2c3 b4d6 d1b3 d5f5 e2d4 f5d7 b3b5 a7a6 b5d7 b8d7 g2g3 d7e5 e3f4 g7g6 a1c1 e5c4 f4d6 c4d6 f1e1');
INSERT INTO evaluations VALUES(47789,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(47789,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(47789,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(47789,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(47789,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(47789,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(47789,36,'cp -37','b1c3 d6d5 c3d5 f6d5 e4d5 f8b4 b2e5 e8g8 f1e2 f8e8 e5f4 d8d5 g1f3 b4f8 e1g1 c7c5 f1e1 b8c6 c2c3 b7b6 d1b3 d5b3 a2b3 c6a5 a1b1 c8f5 b1b2 f5e6 e2b5 e8e7 b3b4 a5c4 b5c4 e6c4 e1e7 f8e7 b4c5 b6c5 f3e5 c4e6 e5c6 e7f6 b2b1 h7h6');
INSERT INTO evaluations VALUES(2220,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2220,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2220,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2220,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(2220,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2220,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(2220,36,'cp 15','c3d5 f6d8 c2c3 a7a6 e1g1 d7d6 h2h3 c5a7 d3d4 e5d4 c3d4 b7b5 c4b3 c8b7 a1c1 a8c8 a2a4 c6e7 a4b5 a6b5 d1d3 e7d5 b3d5 c7c6 d5b3 c6c5 d4c5 a7c5 b3d5 d8b6 c1c2 b7d5 d3d5 c5b4 d5d4 b4c5 d4d3 b5b4');
INSERT INTO evaluations VALUES(2222,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2222,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2222,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2222,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2222,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2222,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2222,36,'cp 0','f6d8 c2c3 a7a6 d3d4 e5d4 c3d4 c5a7 e1g1 d7d6 h2h3 b7b5 c4b3 c6e7 d5e7 d8e7 b3d5 a8b8 d1d3 c7c5 b2b3 c8d7 f1e1 f8d8 a1b1 b8c8 d5b7 c8b8');
INSERT INTO evaluations VALUES(4232,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4232,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(4232,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(4232,36,'cp 0','c2c3 a7a6 d3d4 e5d4 c3d4 c5a7 e1g1 d7d6 h2h3 b7b5 c4b3 c6e7 d5e7 d8e7 b3d5 a8b8 f1e1 c7c5 a1c1 c8b7 b2b3 b8c8 d5b7 e7b7 f3h4 c5d4 d1g4 c8c1 e1c1 g7g6 h4f5 h6h5 g4h4 d4d3 h4f6 g6f5 f6g5 g8h7 g5h5 h7g7 h5g5 g7h7');
INSERT INTO evaluations VALUES(6428,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(6428,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(6428,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(6428,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(6428,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(6428,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(6428,36,'cp 220','d2d4 e5d4 c3d4 c5b6 d4d5 c6e7 d5d6 c7d6 b1c3 g8f6 c1g5 d6d5 g5f6 g7f6 c3d5 e7d5 c4d5 d7d6 e1g1 e8g8 f3h4 g8h8 a2a4 a7a5 d1h5 d8e7 g1h1 f8g8 h4f5 c8f5 h5f5 e7e5 f5f3 b6c5 a1d1 f6f5 e4f5 g8g5 g2g4 h7h5 f1g1 e5f6 d5b7 h5g4 g1g4 g5g4');
INSERT INTO evaluations VALUES(6446,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(6446,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(6446,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(6446,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(6446,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(6446,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(6446,36,'cp 226','c5b6 f3e5 c6e5 d4e5 g8e7 d1h5 e8g8 c1f4 d7d6 e5d6 c7d6 b1d2 c8e6 c4e6 f7e6 f4g3 d8d7 h5g4 f8f6 g4e2 a8f8 f2f3 e7g6 e1c1 e6e5 c1b1 h7h5 d2c4 d7b5 d1d5 b6c5 h2h4 g6e7 d5d2 f6g6 g3h2 b5a6 h1d1 b7b5 c4e3 c5e3 e2e3');
INSERT INTO evaluations VALUES(6884,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(6884,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(6884,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(6884,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(6884,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(6884,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(6884,36,'cp 111','f2f4 e5f6 f1d3 f6d4 b1c3 c7c6 d3e2 d8b6 c3a4 d4f2 e1f1 b6d4 d1d2 g8f6 e2f3 d7d5 e4d5 f6d5 d2f2 d4a4 f2e1 e8d8 e1d2 a4b4 b2b3 b4d2 c1d2 c8f5 c2c4 d5f6 d2c3 h8e8 h2h3 d8e7 g2g4 f5c2 a1c1 c2e4 f1f2 e7f8 h1e1 e4f3 f2f3');
INSERT INTO evaluations VALUES(8128,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(8128,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(8128,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(8128,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(8128,34,'cp 230',NULL);
INSERT INTO evaluations VALUES(8128,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(8128,36,'cp 233','g5f6 g7f6 c3d5 e8g8 b5c7 a8b8 d1h5 c8d7 e1c1 g8h8 h5h6 c6b4 d5b4 d8c7 b4d5 c7d8 f2f3 d7e6 c1b1 f8g8 g2g4 b7b5 h2h4 a7a5 h6d2 a5a4 d2b4 d8d7 a2a3 g8c8 f1e2 d7b7 d5e7 b7e7 d1d6 e7c7 h1c1 c7e7 c2c3 h7h6 c1d1 h8g7 e2d3');
INSERT INTO evaluations VALUES(8126,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(8126,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(8126,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(8126,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(8126,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(8126,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(8126,36,'cp 240','g7f6 c3d5 e8g8 b5c7 a8b8 d1h5 c8d7 e1c1 g8h8 h5h6 c6b4 d5b4 d8c7 b4d5 c7d8 g2g3 f6f5 d5e7 d8e7 f1h3 f7f6 h3f5 d7f5 e4f5 e7d7 d1d5 d7f5 d5d6 f5g5 h6g5 f6g5 h2h4 g5h4 h1h4 f8f7 b2b3 b8f8 h4e4 f7f2 e4e5 f8c8 d6d2 f2d2 c1d2 c8g8 e5e3 b7b5 a2a3 a7a5 e3c3 g8g5 b3b4');
INSERT INTO evaluations VALUES(8069,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(8069,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(8069,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(8069,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(8069,34,'cp 243',NULL);
INSERT INTO evaluations VALUES(8069,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(8069,36,'cp 237','c3d5 e8g8 b5c7 a8b8 d1h5 g8h8 e1c1 c8d7 h5h6 c6b4 d5b4 d8c7 b4d5 c7d8 g2g3 f6f5 d5e7 d8e7 f1h3 f7f6 h3f5 d7f5 e4f5 e7d7 d1d5 d7f5 d5d6 f5g5 h6g5 f6g5 h2h4 g5g4 d6d2 h8g7 h1e1 f8f5 c1d1 b8f8 e1e2 g7f6 d2d6 f6e7 d6h6 f5f2 h6h7 e7e6 e2f2 f8f2 h7g7 f2g2 g7g4 e6f5 g4g7 f5e4 h4h5 e4e3');
INSERT INTO evaluations VALUES(8068,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(8068,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(8068,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(8068,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(8068,34,'cp 240',NULL);
INSERT INTO evaluations VALUES(8068,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(8068,36,'cp 242','e8g8 b5c7 a8b8 d1h5 g8h8 e1c1 c8d7 h5h6 c6b4 d5b4 d8c7 b4d5 c7d8 g2g3 f6f5 d5e7 d8e7 f1h3 f7f6 h3f5 d7f5 e4f5 e7d7 d1d5 d7f5 d5d6 f5g5 h6g5 f6g5 d6d2 h8g7 h1e1 g7f6 d2d5 b8e8 d5d7 f6g6 e1e2 f8f7 d7f7 g6f7 e2e4 f7e6 c1d2 e8f8 d2e3 f8c8 c2c3 e6f5 h2h4 g5g4');
INSERT INTO evaluations VALUES(10244,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(10244,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(10244,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(10244,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(10244,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(10244,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(10244,36,'cp -3','d7d5 e4d5 f6d5 e1g1 c8f5 c2c3 e8g8 b2b4 c5b6 a2a3 h7h6 c3c4 d5f4 c1f4 b6d4 f4e5 c6e5 a1a2 f8e8 f3e5 d4e5 b1d2 a7a5 b4a5 c7c6 d2f3 e5f6 d3d4 f5e4 f3d2 d8d4 d2e4 d4e4 d1d7 e4e7 e2g4 e7d7 g4d7');
INSERT INTO evaluations VALUES(12410,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12410,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(12410,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(12410,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(12410,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(12410,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(12410,36,'cp 12','c5d4 f3d4 f8b4 d4c6 b7c6 f1d3 e6e5 e1g1 e8g8 c1e3 b4c3 b2c3 d7d6 c4c5 d6d5 e4d5 f6d5 e3d2 d8f6 d1c2 h7h6 d3h7 g8h8 h7e4 c8g4 f1e1 g4h5 h2h3 a8c8 f2f3 f8d8 a1b1 d5f4 d2e3 f4d5 e4f5 h5g6 f5g6 f7g6 b1b7 d5e3 e1e3');
INSERT INTO evaluations VALUES(12473,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(12473,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(12473,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12473,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(12473,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(12473,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(12473,36,'cp 29','f3d4 f8b4 d4c6 b7c6 f1d3 e6e5 e1g1 e8g8 c1e3 b4c3 b2c3 d7d6 h2h3 d8e7 c4c5 d6d5 d1a4 d5e4 d3e4 f6e4 a4e4 e7e6 a1d1 c8a6 f1e1 f7f5 e4c2 e6c4 f2f4 e5f4 d1d4 c4f7 e3f4 a8e8 e1e8 f8e8 c2d2 h7h6 d4d7 f7e6 d7a7 a6b5 a7c7 e6e2 d2e2');
INSERT INTO evaluations VALUES(14088,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(14088,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(14088,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(14088,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(14088,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(14088,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(14088,36,'cp 118','b1c3 b7b6 f1e2 g7g6 c1e3 f8g7 h2h4 e8g8 f3d2 h7h5 f2f3 c7c6 a2a4 c8d7 d2b3 c6c5 g2g4 f6h7 g4g5 f7f5 d1c2 a7a6 a4a5 f5e4 f3e4 b6a5 e1d2 d8c8 b3a5 d7g4 d2c1 g4f3 h1h2 f3g4 c1b1 a8b8');
INSERT INTO evaluations VALUES(14086,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(14086,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(14086,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(14086,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(14086,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(14086,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(14086,36,'cp 116','b7b6 h2h4 g7g6 f1e2 f8g7 c1e3 e8g8 f3d2 h7h5 f2f3 a7a6 a2a4 a8b8 g2g4 c7c6 g4g5 f6e8 b2b4 f7f5 b4b5 c6c5 b5a6 b8a8 d1b3 a8a6 a1b1 e8c7 e1g1 c8d7 b1a1 g8h7 g1h2 e7c8 f3f4 a6a8 e2d3 c7a6');
INSERT INTO evaluations VALUES(14877,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(14877,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(14877,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(14877,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(14877,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(14877,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(14877,36,'cp 46','d2d4 f8d6 a2a3 e8g8 b1d2 f6e4 c2c4 f7f5 b3b4 d8f6 f1e2 a7a6 e1g1 c8d7 c4c5 d6e7 f3e1 f6h6 e1d3 b7b5 d3f4 e7g5 f2f3 g5f4 e3f4 e4f6 g2g3 g7g5 f4g5 h6g5 g1h1 g8h8 f3f4 g5h6 d2f3 f6e4');
INSERT INTO evaluations VALUES(14831,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(14831,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(14831,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(14831,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14831,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(14831,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(14831,36,'cp 50','f8e7 f1e2 e8g8 c2c4 d5c4 e2c4 b7b6 b1d2 c8b7 e1g1 a7a6 a2a3 c6b8 d1e2 b8d7 b3b4 c7c6 e3e4 h7h6 f1d1 c6c5 d4c5 b6c5 b4b5 a6b5 c4b5 d8b6 a3a4 b7a6 d1b1 f8b8 f3e5 a6b5 a4b5 d7e5 a1a8 b8a8 b2e5 f6d7 d2c4 b6b7 b5b6');
INSERT INTO evaluations VALUES(15578,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(15578,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(15578,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(15578,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(15578,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(15578,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(15578,36,'cp -10','d6d5 e4d5 f6d5 f3e5 c6e5 d1e2 f7f6 d3d4 f8b4 c1d2 e8g8 d4e5 b4c3 b2c3 f8e8 e1c1 e8e5 e2b5 d5b6 b5b2 d8e8 f1d3 c8e6 h1e1 b6c4 d3c4 e6c4 e1e5 e8e5 b2b7 c4d5 b7b5 d5g2 b5e5 f6e5 h3h4 a8d8 d2e3 d8d1 c1d1 a7a5 f2f4 g2d5 a2a3 g8f7 f4e5 a5a4 e3f4 f7e6 f4e3 e6e5 e3d4 e5e4');
INSERT INTO evaluations VALUES(16666,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16666,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16666,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16666,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16666,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16666,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(16666,36,'cp 17','e3d4 f8b4 d1e2 d8e7 g1f3 f6e4 c1d2 b4c3 d2c3 e8g8 f3d2 e4c3 e2e7 c6e7 b2c3 b7b6 f1d3 d7d5 e1c1 c8d7 c1b2 f8e8 h1e1 a8d8 e1e5 d5c4 d2c4 g8f8 g2g3 h7h6 h2h4 e7c8 c4e3 e8e5 d4e5');
INSERT INTO evaluations VALUES(21648,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(21648,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(21648,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(21648,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(21648,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(21648,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(21648,36,'cp 103','f8b4 c1d2 g8e7 a2a3 b4a5 c3a4 e8g8 d2a5 c6a5 e2e3 b7b6 d1d2 c7c5 b2b4 a5c6 b4c5 c8g4 f3d4 c6e5 c5b6 a7b6 a4c3 d8d6 c3b5 d6g6 f2f4 e5c6 f1d3 g6f6 e1g1 c6d4 b5d4 g4d7 h2h3 e7f5 d3f5 d7f5 d4f5');
INSERT INTO evaluations VALUES(21808,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(21808,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(21808,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(21808,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(21808,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(21808,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(21808,36,'cp -38','c8g4 a2a3 h7h5 e2g1 a7a5 f1e2 a5a4 h2h3 g4d7 f3h2 g7g6 g1f3 f8e7 c2c3 d7e6 c3d4 e5d4 f3d2 g8f6 b2b4 a4b3 d2b3 f6d7 e1g1 g6g5 f2f4 g5f4 h2f3 h8g8 g1h1 e7d6 b3d4 c6d4 f3d4 d8g5 e2f3 g5h4 d4e6 f7e6 f3h5 e8e7');
INSERT INTO evaluations VALUES(24384,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(24384,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(24384,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(24384,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(24384,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(24384,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(24384,36,'cp 74','e1g1 f8e7 b1c3 e8g8 c4d3 c6c5 a1b1 f8e8 b2b3 f6g4 c3d5 e7h4 h2h3 g4e5 c1b2 c7c6 d5e3 d8e7 b2c3 h4f6 d1d2 e5d3 c2d3 f6d4 f1e1 e7g5 c3d4 c5d4 e3c4 g5d2 c4d2 c6c5 d2c4 e8e6 f2f3 g7g5 b1d1 d7c6');
INSERT INTO evaluations VALUES(24383,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(24383,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(24383,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(24383,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(24383,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(24383,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(24383,36,'cp 62','f8e7 c4d3 c6c5 b1c3 e8g8 a1b1 d7c6 b2b3 f8e8 c1d2 f6d7 c3d5 e7f6 d2a5 c6d5 e4d5 a7a6 f1e1 e8e1 d1e1 f6d4 e1d2 g7g6 d3f1 h7h5 b1e1 d4f6 g2g3 f6d4 f1g2 h5h4 e1e4 h4g3 h2g3 d7e5');
INSERT INTO evaluations VALUES(23990,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(23990,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(23990,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(23990,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(23990,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(23990,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(23990,36,'cp 73','c4d3 e8g8 b1c3 f8e8 h2h3 c6c5 a2a4 d7c6 d3b5 d8d7 f2f3 a7a6 b5c4 a8d8 c1f4 f6h5 f4h2 e7f6 d1d2 f6d4 g1h1 h5f6 c4a6 d4c3 d2c3 c6a4 f1e1 d8b8 a6d3 a4b5 e1d1 b5d3 d1d3 b8b4 b2b3 h7h6 a1a7');
INSERT INTO evaluations VALUES(28245,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(28245,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(28245,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(28245,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(28245,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28245,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(28245,36,'cp 0','g5h4 h7h6 d1b3 b4c3 b3c3 g7g5 h4g3 f6e4 c3c2 e4g3 h2g3 d8f6 e2e3 h6h5 d4d5 e6d5 c4d5 b7d5 c2c7 f6b2 a1d1 b2b4 d1d2 b4b1 e1e2 b1b5 e2d1 b5b1');
INSERT INTO evaluations VALUES(30061,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(30061,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30061,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30061,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30061,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(30061,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(30061,36,'cp 0','b8d7 f3h4 f5e4 f2f3 f6h5 f3e4 h5f4 e2e3 d8h4 g2g3 h4e7 e3f4 d5e4 c3e4 e8c8 d1e2 e7b4 e4c3 c7c5 d4d5 e6d5 e1c1 d5d4 f1h3 d4c3 h3d7 c8b8 e2e5 f8d6 e5d6 b8a8 b2c3 b4c3 c1b1 c3b4 b1a1 b4c3');
INSERT INTO evaluations VALUES(33779,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(33779,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(33779,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(33779,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(33779,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(33779,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(33779,36,'cp 129','d2d4 h7h6 c1e3 c7c6 h2h3 e7c7 c4b3 b7b5 a2a3 a7a5 f3h4 f8e7 e1g1 g7g5 h4f5 c8f5 e4f5 d6d5 d4e5 c7e5 f1e1 e5d6 d1e2 b8d7 e3d4 e8f8 b3d5 a8e8 d4f6 d7f6 c3e4 f6e4 d5e4 e7f6 a1d1 d6e5 c2c3 f8g7 e2h5 h8f8 d1d7 e8e7 d7e7');
INSERT INTO evaluations VALUES(34583,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(34583,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(34583,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(34583,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(34583,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(34583,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(34583,36,'cp -82','g8f6 e1g1 c8g4 b1c3 d5d7 f2f3 g4f5 g5e4 b8c6 d1e2 e8c8 c1e3 c6d4 e2f2 e7e5 a1d1 f5g6 e4f6 g7f6 f3f4 g6d3 d1d3 h8g8 f4e5 f6e5 e3d4 e5d4 c3e4 d7c6 d3d4 d8d4 f2d4 b7b6 d4d3 f8c5 g1h1 g8g4 e4g3 g4b4 b2b3 b4d4 d3f5 c8b7 f5f7 c5d6 f7f5');
INSERT INTO evaluations VALUES(37962,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(37962,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(37962,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(37962,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(37962,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(37962,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(37962,36,'cp -21','b8c6 c4b3 e7e6 d2d3 g8e7 c3e2 e8g8 c2c3 b7b5 c1e3 b5b4 e1g1 a8b8 h2h4 h7h5 d1d2 c6a5 b3c2 b4c3 b2c3 b8b2 f1b1 b2b1 a1b1 d8c7 e2g3 a5c6 c2b3 c8d7 e4e5 d6e5 g3e4 e7f5 e3c5 f8d8 f4e5 c6e5 f3e5');
INSERT INTO evaluations VALUES(40603,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(40603,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(40603,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(40603,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(40603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40603,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40603,36,'cp 0','e8g8 c4d5 e6d5 f1e2 b8c6 e1g1 a7a6 b2b3 c5a7 c1b2 f8e8 a1c1 c8f5 h2h3 d5d4 e3d4 c6d4 f3d4 a7d4 e2f3 f5d3 f1e1 d4f2 g1f2 d8b6 f2g3 b6c7 g3f2 c7c5 f2g3');
INSERT INTO evaluations VALUES(36150,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36150,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36150,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36150,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36150,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36150,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36150,36,'cp 0','c4d5 e6d5 f1e2 b8c6 e1g1 a7a6 b2b3 f8e8 c1b2 c5a7 a1c1 c8f5 h2h3 d5d4 e3d4 c6d4 f3d4 a7d4 e2f3 f5d3 f1e1 d4f2 g1f2 d8b6 f2g3 b6d6 g3f2');
INSERT INTO evaluations VALUES(42446,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(42446,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42446,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(42446,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(42446,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(42446,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42446,36,'cp 0','h7h6 g5h4 b7b6 f2f3 d7d5 c4d5 e6d5 e2e3 d8e7 d1d2 e8g8 c3c4 d5c4 f1c4 f8e8 g1e2 e7e3 d2e3 e8e3 h4f6 g7f6 a1c1 g8f8 e1g1 c7c6 e2g3 c8e6 f1d1 e6c4 c1c4 e3e6 d4d5 c6d5 d1d5 b8a6 g3f5 a8e8 g1f2');
INSERT INTO evaluations VALUES(42345,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42345,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(42345,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(42345,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(42345,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(42345,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(42345,36,'cp 31','g5h4 b7b6 f2f3 d7d6 e2e4 b8c6 f1d3 g7g5 h4f2 c8b7 g1e2 d8e7 h2h4 e8c8 g2g4 d6d5 c4d5 e6d5 e4e5 f6e4 e2g3 e4c3 d1d2 c6e5 d4e5 e7e5 d2e3 h8e8 d3f5 c8b8 e3e5 e8e5 e1d2 d5d4 h1e1 e5e1 a1e1 b7f3 e1e7 c7c5 g3f1 g5h4 f2h4 c3a2 e7f7 d8e8 f1g3 a2c3 f7g7');
INSERT INTO evaluations VALUES(42349,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(42349,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(42349,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(42349,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(42349,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(42349,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(42349,36,'cp 41','b7b6 f2f3 d7d5 c4d5 e6d5 e2e3 e8g8 f1d3 f8e8 h4f2 c7c5 g1e2 c8a6 e2g3 a6d3 d1d3 b8c6 e1g1 d8d7 a1d1 a8d8 f1e1 h6h5 h2h3 c5d4 c3d4 c6e7 e3e4 d5e4 f3e4 h5h4 g3e2 e7g6 e2c3 d8c8 d4d5 g6e5 d3c2 e5g6 c2b3 f6h5 d5d6 h5f4');
INSERT INTO evaluations VALUES(43061,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(43061,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(43061,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43061,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(43061,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(43061,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(43061,36,'cp 83','h7h6 g5f4 d7d6 d1d2 a7a6 f1d3 b8d7 h2h3 b7b6 e1g1 c8b7 a1d1 g6g5 f4g3 e7g6 d2e3 e8g8 f1e1 e6e5 d4e5 d7e5 f3d4 d8f6 d4f5 g6e7 f5g7 f6g7 f2f4 e5d3 e3d3 g5f4 g3f4 a8e8 d3c4 e7g6 f4g3 e8e7 c3d5 b7d5 e4d5 g7f6 e1e7 f6e7');
INSERT INTO evaluations VALUES(47396,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(47396,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(47396,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(47396,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(47396,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(47396,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(47396,36,'cp -11','d8d1 e1d1 b8c6 f2f3 c8d7 c1e3 e8c8 d1e1 f6e8 g1e2 e7e5 a1d1 e8d6 e2g3 g7g6 e3g5 f8e7 g5e7 c6e7 h2h4 d7e6 b2b3 c7c6 c3a4 b7b6 h4h5 c8c7 c2c4 f7f5 c4c5 d6f7 d1d8 h8d8 h5g6 h7g6 h1h7 d8h8 h7g7 f5f4 g3e2 b6b5 a4b2 h8h1');
INSERT INTO evaluations VALUES(33713,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(33713,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(33713,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(33713,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(33713,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(33713,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(33713,36,'cp -16','e1d1 b8c6 f2f3 c8d7 c1e3 e8c8 d1e1 f6e8 g1e2 e7e5 a1d1 e8d6 e2g3 g7g6 e3g5 f8e7 g5e7 c6e7 h2h4 h7h5 b2b3 d7e6 c3b5 d6b5 f1b5 c7c6 d1d8 c8d8 b5d3 e7c8 a2a4 a7a5 e1f2 c8d6 h1a1 d8e7 f2e3 f7f5 g3e2 f5f4 e3f2 d6f7');
INSERT INTO evaluations VALUES(47746,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(47746,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(47746,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(47746,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(47746,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(47746,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(47746,36,'cp 117','f8d6 b1c3 b8c6 c1g5 f7f6 g5d2 c8e6 d1a4 e8g8 e2e3 d8e8 c3b5 a8d8 b5d6 d8d6 d2b4 d6d8 b4f8 c6d4 a4d1 d4f3 d1f3 e8f8 c4c5 f8c5 f3e4 c5d6 e4c2 d6b4 c2c3 b4e4 a1c1 g4e5 h1g1 e6d5 c3d4 e4d4 e3d4 e5c6 f1e2 c6d4 c1c7');
INSERT INTO evaluations VALUES(41758,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(41758,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(41758,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(41758,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(41758,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(41758,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(41758,36,'cp 113','b1c3 b8c6 c1g5 f7f6 g5d2 c8e6 e2e3 d8e7 c3d5 e7f7 f1e2 e8c8 d2c3 c8b8 d1b3 h7h5 h2h3 g4e5 e1c1 g7g5 d1d2 g5g4 f3d4 c6d4 e3d4 e5g6 d5e3 d6f4 h3g4 h5g4 h1h8 d8h8 d4d5 e6c8 c3d4 g4g3 f2g3 f4g3 c4c5 g3f4 d2d1 h8h2');
INSERT INTO evaluations VALUES(41756,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(41756,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(41756,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(41756,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(41756,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(41756,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(41756,36,'cp 102','b8c6 c1g5 f7f6 g5d2 c8e6 e2e3 d8e7 d1b3 e8c8 f1e2 g4e5 f3e5 c6e5 c3d5 e7f7 a1c1 d6c5 d5f4 e6g4 e2g4 e5g4 b3c2 c8b8 f4d3 c5d6 h2h3 g4e5 d3e5 d6e5 e1g1 g7g5 f2f4 g5f4 e3f4 e5d4 g1h2 h8e8 c1e1 f6f5 e1e8 d8e8 d2b4 f7e6');
INSERT INTO evaluations VALUES(49497,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(49497,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(49497,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(49497,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(49497,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(49497,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(49497,36,'cp -104','c3c2 d1c2 d7d5 c4b3 f6e4 b1c3 e4c5 b3d5 b8c6 c2e2 c6b4 c1e3 b4d5 a1d1 c8e6 e3c5 f8c5 e2b5 c7c6 b5c5 d8b6 c5b6 a7b6 g1e2 d5c3 e2c3 a8a5 f2f4 e6a2 e1g1 a2c4 f1e1 g7g6 b2b4 a5a8 b4b5 a8d8 d1b1 d8d3 e1c1 e8d7 b5c6 d7c6');
INSERT INTO evaluations VALUES(51613,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(51613,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(51613,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(51613,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(51613,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(51613,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(51613,36,'cp 165','d2d4 d7d6 f1d3 g8f6 e1g1 g7g6 e4e5 d6e5 d4e5 f6d5 b1a3 b5b4 a3c4 f8g7 c1d2 e8g8 a2a3 c6c5 d3e4 b8c6 d2b4 c8e6 b4c5 a8c8 b2b4 d5c3 d1d8 c6d8 c4d2 c3e4 d2e4 d8c6 f1e1 c6e5 f3e5 g7e5');
INSERT INTO evaluations VALUES(54200,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(54200,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(54200,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(54200,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(54200,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(54200,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(54200,36,'cp -47','g8f6 c2c3 f8e7 f1d3 e8g8 e1g1 b7b6 d1e2 a7a5 b1d2 c8a6 d3a6 b8a6 e3e4 d5e4 d2e4 f6e4 e2e4 a6c7 f4f5 e6f5 e4f5 d8d5 f5d5 c7d5 f1e1 f8e8 c1d2 c5d4 f3d4 e7c5 h2h3 f7f6 e1e8 a8e8 a1e1 c5d4 c3d4 e8e1 d2e1');
INSERT INTO evaluations VALUES(4262,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4262,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4262,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4262,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4262,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4262,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4262,36,'cp 16','b5c6 b7c6 e1g1 d7d5 f3e5 f6e4 d2d4 e4c3 b2c3 d8f6 d4c5 f6e5 d1d4 e5d4 c3d4 c8f5 c2c3 f8e8 c1f4 e8e2 f4c7 a8e8 f2f3 f7f6 c7g3 g8f7 f1e1 f5d3 a2a3 a7a5 a3a4 e2d2 e1e8 f7e8 g3d6 d2e2 a1d1');
INSERT INTO evaluations VALUES(4248,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(4248,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(4248,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(4248,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4248,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(4248,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4248,36,'cp 12','d7c6 d2d3 f8e8 e1g1 f6d7 c1e3 c5d6 f3d2 d7f8 f2f4 e5f4 e3f4 d6f4 f1f4 d8d6 f4f2 c8e6 d2f3 f8g6 f2f1 a8d8 d1d2 b7b6 d2f2 g6e5 f2g3 e5f3 f1f3 d6c5 g1h1 c5b4 c3d1 b4d6 d1e3 c6c5 b2b3 d6g3 f3g3 h7h6 a1f1 a7a5');
INSERT INTO evaluations VALUES(3905,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(3905,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(3905,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3905,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3905,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(3905,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3905,36,'cp 20','d2d3 f6d7 e1g1 f8e8 c1e3 c5d6 f3d2 d7f8 f2f4 e5f4 e3f4 c8e6 d1e1 d6c5 g1h1 f8g6 f4h2 e8f8 e1g3 c5d6 g3f2 f7f6 a1e1 a7a6 h2d6 d8d6 b2b3 a8e8 a2a4 d6b4 c3d1 f6f5 e4f5 f8f5 f2f5 e6f5 e1e8 g8f7 e8e2 g6f4');
INSERT INTO evaluations VALUES(4865,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(4865,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(4865,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(4865,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(4865,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(4865,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(4865,36,'cp 135','c2c3 c5a7 d2d4 e5d4 c3d4 d7d6 h2h3 h7h6 b1c3 b7b5 c4b3 g8e7 a2a4 b5a4 a1a4 e8g8 b3c4 a6a5 c1e3 e7g6 d1d2 c8d7 f1a1 a7b6 e3h6 g7h6 d2h6 d8f6 c3d5 f6g7 h6g5 f8d8 d5f6 g8f8 f6d7 d8d7 c4b5 g6e7 g5g7 f8g7 a4c4 f7f5 b5c6 e7c6 c4c6 f5e4');
INSERT INTO evaluations VALUES(4859,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(4859,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(4859,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(4859,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(4859,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(4859,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(4859,36,'cp 141','c5a7 d2d4 d7d6 d4e5 d6e5 d1d8 c6d8 f3e5 g8f6 f1e1 e8g8 b1d2 b7b5 c4f1 c8b7 e5d3 d8e6 e4e5 f6d7 a2a4 a7b6 a4a5 b6a7 b2b4 c7c5 d2e4 b7e4 e1e4 c5c4 d3e1 f8e8 g2g3 e6c7 e1f3 a8d8 e4e1 c7d5');
INSERT INTO evaluations VALUES(10746,30,'cp 210',NULL);
INSERT INTO evaluations VALUES(10746,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(10746,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(10746,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(10746,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(10746,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(10746,36,'cp 226','c3b5 e8g8 b5d6 c7d6 f1d3 d8a5 c2c3 f8e8 e1g1 a5e5 f2f3 d6d5 d4e5 e8e5 e4d5 e5d5 f1d1 d5c5 c1e3 c5c6 e3d4 d7d5 g1f2 c8d7 d1e1 c6e6 e1e6 d7e6 g2g4 a7a5 a1e1 a5a4 f2e3 h7h6 h2h4 f6e8 e3f4 a4a3 b2b3 e8d6 h4h5 a8c8 d3c2');
INSERT INTO evaluations VALUES(13511,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(13511,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(13511,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(13511,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(13511,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(13511,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(13511,36,'cp -20','e6d5 f1e1 f8e7 c4b5 e8g8 c1f4 c6a5 b5a4 a7a6 c2c3 c8f5 a4c2 a5c6 b1d2 e7d6 f4d6 d8d6 h2h3 h7h6 d3d4 f5c2 d1c2 c5d4 f3d4 c6d4 c3d4 d6b4 d2b3 f8c8 c2d3 b7b6 a1c1 a6a5 a2a3 b4a4 b3d2 c8c1 e1c1 f6h5 d2f1 h5f4');
INSERT INTO evaluations VALUES(13739,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(13739,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(13739,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(13739,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(13739,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(13739,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(13739,36,'cp -16','f1e1 f8e7 c4b5 e8g8 c1f4 c6a5 d3d4 c5d4 f3d4 e7c5 b1c3 a7a6 b5f1 c8g4 d4f3 a5c4 a1b1 b7b5 b2b4 c5a7 a2a4 d8b6 f4g3 g4f3 g2f3 b6c6 d1d3 f8c8 a4b5 a6b5 e1d1 c8d8 c3e2 d8e8 d1e1 g7g6 e2d4 a7d4 d3d4 c4a3');
INSERT INTO evaluations VALUES(16509,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(16509,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(16509,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(16509,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(16509,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(16509,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(16509,36,'cp -50','f8c5 d2d3 d7d5 e4d5 f6d5 d3d4 c5b4 c1d2 b4c3 b2c3 e8g8 f3e5 f8e8 e2e4 d5f6 e4f4 c6e5 d4e5 d8d6 e1c1 d6a3 c1b1 c8e6 c3c4 f6d7 f4g3 a3c5 f2f4 d7b6 g3e3 c5c6 e3f3 c6f3 g2f3 b6c4 f1d3 g7g6 d3e4 c4d2 d1d2 a8d8 h1d1 d8d2 d1d2 b7b6 b1b2');
INSERT INTO evaluations VALUES(19393,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(19393,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(19393,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(19393,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(19393,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(19393,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(19393,36,'cp -24','g7g6 b1c3 f8g7 d2d3 a7a6 g1f3 d7d6 f1e2 g8f6 a1b1 e8g8 h2h3 f6d7 c1e3 c6d4 d1d2 e7e6 h3h4 h7h5 e3g5 d8c7 g5h6 b7b6 e1g1 c8b7 b2b4 f7f5 h6g7 g8g7 d2g5 c7d8 b4c5 d6c5 e4f5 d8g5 f3g5 e6f5 f2f4 b7c6');
INSERT INTO evaluations VALUES(19661,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(19661,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(19661,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(19661,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19661,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(19661,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19661,36,'cp 30','d2d4 c8g4 d4d5 g4f3 d1f3 c6d4 f3d1 e7e5 f1d3 g8f6 e1g1 f6d7 c1e3 a7a6 a1b1 f8g7 d1d2 b7b6 f2f3 e8g8 g1h1 d8e7 b2b3 a8b8 g2g3 g8h8 b3b4 c5b4 b1b4 d7c5 b4b2 h8g8 f1f2 f8c8 d3f1 c8c7 d2d1');
INSERT INTO evaluations VALUES(20193,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(20193,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20193,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(20193,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(20193,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(20193,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20193,36,'cp 30','c8d7 b5a4 g8e7 e1g1 c5d4 c3d4 c6b4 b1c3 d7a4 c3a4 b4c6 d1d3 e7f5 c1d2 h7h5 f1c1 f8e7 a4c5 e7c5 c1c5 f5h4 f3h4 d8h4 d2e3 e8g8 a1c1 f7f6 e5f6 h4f6 a2a3 a7a6 c5c3 f8f7 d3d1 f6g6 d1c2 g6c2 c1c2 a8f8 h2h4 g8h7 f2f3');
INSERT INTO evaluations VALUES(25930,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(25930,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(25930,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(25930,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25930,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25930,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(25930,36,'cp 30','b5a4 c5d4 c3d4 c6b4 a4d7 d8d7 b1c3 g8e7 e1g1 e7f5 c1f4 h7h6 a1c1 f8e7 g2g4 f5h4 f3h4 e7h4 f4g3 h4d8 f2f4 e8g8 f4f5 f7f6 d1b3 b4c6 f5e6 d7e6 g1h1 c6d4 b3b7 a8c8 e5f6 d8f6 b7d5 f8d8 d5e4 e6e4 c3e4 c8c1 f1c1 d4e2 g3c7 d8d7 c1e1 f6e7 e1e2 d7c7');
INSERT INTO evaluations VALUES(22172,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(22172,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(22172,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(22172,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(22172,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(22172,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(22172,36,'cp 175','f3d4 e5d4 d1d4 g8f6 b1c3 f8d6 f1b5 c8d7 e1g1 e8g8 c1g5 d7f5 b5d3 f5d3 d4d3 f8e8 a1d1 h7h6 g5h4 a7a6 f1e1 g7g5 h4g3 d8d7 g3d6 c7d6 h2h3 e8e1 d1e1 a8e8 g1f1 e8e1 f1e1 d7e7 d3e3 e7e3 f2e3 b7b5 a2a3 g8f8 e1e2 f8e7 e2d3');
INSERT INTO evaluations VALUES(22170,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(22170,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(22170,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(22170,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(22170,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(22170,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(22170,36,'cp 161','e5d4 d1d4 g8f6 b1c3 f8d6 f1b5 c8d7 e1g1 e8g8 c1g5 d7f5 b5d3 f5d3 d4d3 h7h6 g5h4 f8e8 a1e1 e8e5 e1e4 g7g5 h4g3 e5e4 c3e4 f6e4 d3e4 d8e8 e4f5 a8d8 b2b3 e8d7 f5d7 d8d7 f1d1 b7b5 g1f1 g8f8 f1e2 d7d8 a2a3 f8e7 e2e3 d6g3 h2g3 d8e8 e3d3 e7d6');
INSERT INTO evaluations VALUES(22162,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(22162,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(22162,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(22162,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(22162,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(22162,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(22162,36,'cp 146','d1d4 g8f6 b1c3 f8d6 f1b5 c8d7 b5d7 d8d7 c1g5 d7f5 g5f6 f5f6 d4f6 g7f6 c3b5 e8c8 b5d6 d8d6 e1c1 h8e8 d1d2 f6f5 b2b3 f5f4 d2d4 f4f3 g2g3 e8e2 d4d2 e2d2 c1d2 d6d5 d2e3 b7b5 a2a3 a7a5 h1c1 d5c5 e3d4 c5f5 c2c4 a5a4 b3a4');
INSERT INTO evaluations VALUES(22161,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(22161,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(22161,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(22161,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(22161,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(22161,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(22161,36,'cp 157','g8f6 b1c3 f8d6 f1b5 c8d7 e1g1 e8g8 c1g5 d7f5 b5d3 f5d3 d4d3 f8e8 d3f3 e8e5 g5h4 d8e7 a1d1 a8e8 h4f6 e7f6 f3f6 g7f6 a2a3 a7a6 g2g3 f6f5 g1g2 g8g7 d1d3 h7h5 f1d1 h5h4 g3h4 g7f6 d3g3 e5e1 d1e1 e8e1 g3g8');
INSERT INTO evaluations VALUES(15791,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(15791,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(15791,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(15791,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(15791,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(15791,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(15791,36,'cp 178','b1c3 f8d6 f1b5 c8d7 c1g5 d8e7 e1d2 d6e5 b5d7 e8d7 d4a4 d7c8 a1e1 e7d6 e1e5 d6e5 g5f4 e5e7 a4c4 h8d8 h1e1 e7d7 d2c1 d8e8 e1d1 b7b5 c4c5 a7a6 d5d6 c8b7 d6c7 d7c6 c5d4 c6g2 b2b3 g2c6 c1b2 a8c8 a2a4 f6e4 a4b5 c6c3 d4c3 e4c3');
INSERT INTO evaluations VALUES(15790,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(15790,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(15790,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(15790,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(15790,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(15790,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(15790,36,'cp 162','f8d6 f1b5 c8d7 e1g1 e8g8 c1g5 d7f5 b5d3 f5d3 d4d3 f8e8 g5h4 e8e5 a1d1 e5h5 h4g3 d6g3 h2g3 h5e5 f1e1 d8d6 e1e5 d6e5 d3e2 e5e2 c3e2 a8d8 c2c4 c7c6 e2c3 c6d5 f2f3 g8f8 c4d5 f8e7 g3g4 h7h6 g1f2 e7d6 f2g3 d6d7 c3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(23034,30,'cp -313',NULL);
INSERT INTO evaluations VALUES(23034,31,'cp -325',NULL);
INSERT INTO evaluations VALUES(23034,32,'cp -325',NULL);
INSERT INTO evaluations VALUES(23034,33,'cp -312',NULL);
INSERT INTO evaluations VALUES(23034,34,'cp -343',NULL);
INSERT INTO evaluations VALUES(23034,35,'cp -348',NULL);
INSERT INTO evaluations VALUES(23034,36,'cp -344','e5f7 e8f7 d1h5 g7g6 h5b5 d7d5 d2d3 f8g7 b1c3 g8f6 e4e5 c7c6 b5b3 h8e8 d3d4 f6h5 c3e4 g7f8 e1g1 f7g8 c1g5 d8b6 e4d6 f8d6 e5d6 h5g7 b3a3 b6d4 f1e1 e8e1 a1e1 c8d7 h2h3 a8f8 e1e7 d4f2 g1h2 f8f7 a3b4 f2f5 b4b7 f7e7 d6e7 g8f7 e7e8r');
INSERT INTO evaluations VALUES(23408,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(23408,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(23408,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(23408,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(23408,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(23408,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(23408,36,'cp 93','d4d5 c6b8 e2e4 d7d6 c2c4 f8g7 h2h3 g8f6 b1c3 e8g8 f1d3 e7e5 c1g5 a7a5 a2a3 b8a6 b2b4 a5b4 a3b4 c8d7 a1b1 c7c5 d5c6 b7c6 e1g1 a6c7 d1d2 c7e6 g5e3 f6e8 b4b5 e6d4 d3e2 d4e2 d2e2 f7f5 f1d1 f5f4 e3c1 g6g5');
INSERT INTO evaluations VALUES(27288,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(27288,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(27288,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(27288,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(27288,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(27288,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(27288,36,'cp 125','e1g1 a7a6 b5c6 d7c6 f3e5 e7e6 e5c6 b7c6 d1f3 d8d7 f1e1 g8e7 b2b3 e7g6 c1b2 f8e7 d2d4 e8g8 a1d1 e7b4 g2g3 h7h6 e1e3 a8d8 a2a3 b4d6 c3a4 d7e7 f3c6 d6a3 b2a3 e7a3 c6c4 g6e7 e3e1 a6a5 c2c3 a3a2 a4c5 a2b2 h2h3');
INSERT INTO evaluations VALUES(27287,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(27287,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(27287,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(27287,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(27287,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(27287,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(27287,36,'cp 120','a7a6 b5c6 d7c6 f3e5 e7e6 e5c6 b7c6 d1f3 d8d7 f1e1 f8e7 d2d4 g8f6 c1f4 a6a5 c3a4 d7d5 f4c7 d5f3 g2f3 f6d5 c7g3 h7h5 c2c4 h5h4 g3e5 d5b4 a1d1 e8c8 a2a3 b4c2 e1e2 f7f6 e2c2 f6e5 d4d5 c8b7 d1e1 e6d5 c4d5 c6d5 e1e5');
INSERT INTO evaluations VALUES(30522,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(30522,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(30522,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(30522,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(30522,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(30522,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(30522,36,'cp 72','g7g6 c1e3 f8g7 d1d2 b8d7 f2f3 h7h5 f1c4 a7a6 c4b3 g8f6 g1e2 c7c6 d5c6 b7c6 e1g1 d7c5 a1d1 e8g8 e2d4 c6b7 a2a4 c5b3 c2b3 g8h7 b3b4 a8c8 h2h3 e7e5 d4c2 b6b5 a4b5 a6b5 d2d6');
INSERT INTO evaluations VALUES(32002,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(32002,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(32002,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(32002,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(32002,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(32002,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(32002,36,'cp -39','h2h3 g4h5 b1c3 b8d7 g2g4 h5g6 d4d5 c7c6 c1f4 g8f6 g4g5 f6e4 d5c6 b7c6 c3e4 g6e4 d1e2 d8a5 f4d2 a5d5 c2c4 e4f3 c4d5 f3e2 d5c6 e2f1 c6d7 e8d7 h1f1 f7f6 f1f3 a8b8 b2b4 h7h6 a1d1 b8b7 g5h6 e7e5 h6g7 f8g7 e1e2 d7e6 f3a3 g7f8 a3a6 f8d6');
INSERT INTO evaluations VALUES(34682,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(34682,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(34682,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(34682,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(34682,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(34682,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(34682,36,'cp -38','g8f6 a2a3 g7g6 f1b5 c7c6 b5c4 f8g7 g1e2 e8g8 e1g1 c8g4 f3g3 b7b5 c4a2 c6c5 d2d3 g4e2 c3e2 e7e6 g3h4 b8d7 c1g5 a5b6 f1e1 a7a5 c2c3 f8e8 e2g3 b5b4 a2c4 b4a3 b2a3 f6d5 a1b1 b6c7 c4d5 e6d5');
INSERT INTO evaluations VALUES(33741,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(33741,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(33741,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(33741,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(33741,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(33741,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(33741,36,'cp -24','a2a3 g7g6 f1b5 c7c6 b5c4 f8g7 g1e2 e8g8 e1g1 c8g4 f3e3 f6d5 e3g3 g4e6 c3e4 d5c7 c4e6 c7e6 g3h4 f8e8 d2d3 a5b6 a1b1 c6c5 c1h6 g7h6 h4h6 b8c6 e2f4 e6f4 h6f4 c6d4 f4e5 d4c2 e5c5 b6c5 e4c5 c2d4');
INSERT INTO evaluations VALUES(35985,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35985,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35985,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(35985,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(35985,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35985,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(35985,36,'cp 0','d1d2 b8c6 d4c6 b7c6 e1g1 a7a5 a1b1 a5a4 h2h3 f8e8 b2b3 a4b3 c2b3 f6e4 c3e4 d6d5 b1d1 c8f5 f1e1 d5c4 d2d8 a8d8 d1d8 e8d8 g5e7 d8a8 e4f6 g7f6 e7f6 c4b3 a2b3 c6c5 e1e5 a8c8 f6e7 f5e6 e7c5 e6b3 c5e3 h7h5 e3d2 c8c2 d2e3');
INSERT INTO evaluations VALUES(35986,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35986,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35986,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35986,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35986,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35986,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35986,36,'cp 0','b8c6 d4c6 b7c6 e1g1 a7a5 h2h3 a5a4 b2b3 a4b3 c2b3 f8e8 b3b4 f6e4 c3e4 g7a1 f1a1 d6d5 d2c3 d5c4 g5e7 d8e7 e4f6 g8f8 f6h7 f8g8 h7f6 g8f8');
INSERT INTO evaluations VALUES(37392,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(37392,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(37392,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(37392,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(37392,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(37392,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(37392,36,'cp -21','g1f3 c7c5 b1c3 c5d4 f3d4 d7d6 f4g3 b8c6 f1e2 c6d4 e3d4 d6d5 e1g1 d5c4 e2c4 c8g4 d1d2 g4d7 f1e1 e7e6 g3d6 f8e8 d4d5 a8c8 c4b3 e6d5 c3d5 d7e6 d5f6 d8f6 d6e5 f6d8 d2d8 c8d8 e5g7 g8g7 b3e6 e8e6 e1e6 f7e6 a1c1 d8d2');
INSERT INTO evaluations VALUES(38269,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(38269,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(38269,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(38269,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(38269,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(38269,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(38269,36,'cp 73','c7c5 d4d5 e7f6 f1d3 g8e7 d1c2 e7g6 e1g1 b8d7 b1c3 b7b6 c1d2 a7a6 a1d1 e8g8 g1h1 d7e5 f3e5 g6e5 d3e2 e5g6 f2f4 f6d4 b2b3 c8d7 d5e6 f7e6 e2d3 d8h4 c3e2 e6e5 f4f5 g6e7 e2d4 c5d4 b3b4 g8h8 a2a4 a8b8 b4b5');
INSERT INTO evaluations VALUES(43262,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(43262,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(43262,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(43262,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(43262,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(43262,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43262,36,'cp 44','e8g8 e1g1 d7d6 f1e1 b8c6 h2h3 e7e5 c2c3 a7a5 a2a4 f8e8 d4d5 c6e7 d3c2 h7h6 b1a3 e8f8 a3b5 f6e8 c3c4 f7f5 e4f5 c8f5 c1e3 d8d7 f3d2 e8f6 b5c3 f5c2 d1c2 e7f5 a1d1 b7b6 d2e4 f6e4 c3e4 f5d4 c2d3 d7f7 c4c5 f7d5 c5b6 c7b6 e3d4 e5d4');
INSERT INTO evaluations VALUES(37464,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(37464,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(37464,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(37464,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(37464,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(37464,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(37464,36,'cp 60','e1g1 d7d6 f1e1 b8c6 c2c3 e7e5 h2h3 a7a5 a2a4 f8e8 d4d5 c6e7 d3c2 h7h6 b1a3 e8f8 a3b5 f6e8 c3c4 f7f5 e4f5 c8f5 c2b3 g6g5 f3d2 c7c6 b5c3 e8f6 d2e4 f6e4 c3e4 f5e4 e1e4 e7f5 b3c2 f5d4 a1a3 d4c2 d1c2 c6d5');
INSERT INTO evaluations VALUES(45488,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(45488,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(45488,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(45488,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(45488,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(45488,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(45488,36,'cp 33','e7e6 g1f3 f8e7 e1g1 b8c6 a2a3 e8g8 c1f4 a7a6 c4a2 b7b5 d4d5 f6d5 c3d5 e6d5 a2d5 c8b7 f3e5 c6e5 d5b7 a8a7 b7e4 e5g6 f4e3 a7d7 d1c2 d7c7 c2e2 e7c5 a1d1 d8f6 e3c5 c7c5 g2g3 f8e8 d1c1 c5c1 f1c1 g6e5 e4g2');
INSERT INTO evaluations VALUES(46374,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46374,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(46374,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46374,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(46374,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(46374,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(46374,36,'cp 34','g1f3 f8e7 e1g1 e8g8 f1e1 b8c6 a2a3 b7b6 d1d3 c8b7 c4a2 h7h6 c1e3 e7d6 c3e4 f6e4 d3e4 f8e8 a1d1 a8c8 e3d2 c6a5 e4g4 d8f6 d4d5 e6d5 d2c3 f6g6 e1e8 c8e8 g4g6 f7g6 a2d5 b7d5 d1d5 a5c4 h2h3 e8c8 d5d1 d6f8');
INSERT INTO evaluations VALUES(46681,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(46681,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46681,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46681,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(46681,36,'cp 0','f8g7 g1f3 c7c5 b1c3 c5d4 e3d4 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 b8c6 c3e2 e7e5 d4e5 d6e5 f4g5 h7h6 g5f6 d8f6 f3f6 g7f6 a1d1 f8d8 d3e4 f6g5 g2g4 a8b8 c2c3 c6a5 b2b3 g8g7 c3c4 a5c6 g1g2 d8d2 d1d2 g5d2 f1d1 b8d8 e4d5 d2g5');
INSERT INTO evaluations VALUES(46905,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(46905,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(46905,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(46905,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(46905,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(46905,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(46905,36,'cp -25','g5e4 e7e5 f1e2 b8c6 e1g1 c8e6 b1d2 f8e7 f1e1 d8d7 c2c3 a7a5 e2f1 e8g8 e4g3 f8d8 h2h3 e7d6 d2f3 d8e8 g3e4 f6e4 d3e4 e8d8 f1b5 f7f6 d1c2 d7f7 a2a4 c6e7 c1e3 e7g6 f3d2 g6f4 e3f4 e5f4 d2f3');
INSERT INTO evaluations VALUES(46910,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(46910,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(46910,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(46910,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(46910,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(46910,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(46910,36,'cp -36','e7e5 f1e2 c8e6 b1d2 f8e7 e1g1 b8c6 f1e1 d8d7 c2c3 a7a5 e4g3 e8g8 e2f1 f8e8 d2f3 e7d6 g3e4 f6e4 d3e4 e8d8 c1e3 d7e7 d1c2 d6c5 c2e2 c5e3 e2e3 d8d6 f1b5 e7f6 e1d1 a8d8 d1d6 d8d6 b5c6 d6c6 e3a7 e6g4 h2h3');
INSERT INTO evaluations VALUES(47989,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(47989,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(47989,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(47989,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47989,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(47989,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(47989,36,'cp 25','e5f3 d6d5 f1d3 f8d6 e1g1 e8g8 f1e1 c8f5 c2c4 c7c6 f3g5 f8e8 d3e4 d5e4 b1c3 b8d7 g5e4 f5e4 c3e4 d6h2 g1f1 h2c7 c1g5 f7f6 g5h4 d7f8 d1d3 d8d7 h4g3 c7g3 e4g3 e8e1 a1e1 a8d8 g3f5 f8e6 d4d5 c6d5 d3d5 d7d5 c4d5');
INSERT INTO evaluations VALUES(48635,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48635,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48635,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(48635,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48635,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48635,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(48635,36,'cp -7','d2d4 f8b4 b1c3 e8g8 f1d3 d5c4 d3c4 c8g4 e1g1 b8c6 c1g5 h7h6 g5h4 b4e7 d4d5 c6e5 c4e2 e5g6 h4g3 e7d6 d1b3 f8e8 b3b7 a8b8 b7a7 b8b2 e2d1 h6h5 g3d6 d8d6 f1e1 e8e1 f3e1 g6f4 a7e3 f6d5 c3d5 f4d5 e3d3 c7c6 d1b3 d6e5');
INSERT INTO evaluations VALUES(52742,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(52742,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(52742,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(52742,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(52742,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52742,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(52742,36,'cp 13','g7g6 d2d4 g8f6 f1g2 f8g7 c4d5 c6d5 e1g1 e8g8 b1c3 f6e4 c1f4 b8c6 a1c1 c8f5 f3e5 e4c3 c1c3 d8b6 e5c6 b7c6 e2e3 b6b2 c3c6 b2a2 c6c5 f8d8 g3g4 f5e6 d1a1 a2a1 f1a1 a7a5 f4c7 d8c8 a1a5 a8a5 c7a5 e6g4 g2d5 g4h3 c5c8 h3c8 d5g2');
INSERT INTO evaluations VALUES(53316,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(53316,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(53316,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(53316,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(53316,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(53316,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(53316,36,'cp -46','b8c6 f1g2 e7e5 e1g1 f8e7 e2e4 d5d4 b1d2 d8c7 f3h4 e7h4 g3h4 g8f6 f2f4 e5f4 d2c4 c8g4 c1f4 g4d1 f4c7 d1c2 e4e5 c2d3 e5f6 d3c4 f6g7 h8g8 f1f5 c4e6 f5c5 g8g7 c7g3 a8d8 a1d1 d4d3 d1d2 c6d4 g1f2 d4c2 c5a5 a7a6');
INSERT INTO evaluations VALUES(19810,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(19810,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(19810,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(19810,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(19810,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(19810,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(19810,36,'cp -26','f1g2 e7e5 e1g1 f8e7 e2e4 d5d4 b1a3 c8e6 a3c4 d8c7 f3e1 h7h5 f2f4 h5h4 g3h4 e7h4 f4e5 e6c4 d3c4 c6e5 e1d3 g8e7 d3e5 c7e5 c1f4 e5e6 e4e5 e6c4 g2b7 a8d8 b7e4 c4e6 d1f3 e7g6 g1h1 e8g8 a2a4 g6e5 f3h5 f7f5');
INSERT INTO evaluations VALUES(54679,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(54679,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(54679,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(54679,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(54679,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(54679,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(54679,36,'cp 195','d2d4 g7g6 b1c3 f8g7 f1c4 b8c6 a2a4 g8h6 h2h4 e7e6 h4h5 g6g5 d4d5 e6d5 d1d5 d8e7 c1e3 h6g4 e1c1 g4e3 f2e3 f6f5 e4f5 g7c3 b2c3 c8d7 g2g4 e8c8 h1e1 h8f8 c1b2 h7h6 e3e4 f8e8 a4a5 a7a6 e4e5 c6e5 f3e5 d6e5 d5f7 e7f7 c4f7');
INSERT INTO evaluations VALUES(2045,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(2045,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(2045,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(2045,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(2045,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(2045,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(2045,36,'cp -38','c6a5 c4b5 g8f6 h5e2 g7g6 d2d3 f8g7 g1f3 e8g8 e1g1 a7a6 b5a4 d7d6 c1e3 b7b5 a4d1 c7c5 b1d2 a5c6 h2h3 c8e6 d1b3 e6b3 d2b3 h7h6 e2d1 e7d7 a2a4 g8h7 d1c2 f6h5 a4b5 a6b5 a1a8 f8a8');
INSERT INTO evaluations VALUES(5641,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(5641,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(5641,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(5641,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(5641,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(5641,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(5641,36,'cp 0','e5f4 d2d4 g7g5 b1c3 h7h6 g2g3 c8h3 g3f4 g5g4 f3e1 e7f6 c1e3 h3f1 c4f1 g8e7 e1g2 h6h5 d1d2 h5h4 f4f5 d6d5 e4e5 f6e5 d4e5 d5d4 c3e4 d4e3 e4f6 e8f8 f6d7 f8e8 d7f6');
INSERT INTO evaluations VALUES(5651,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(5651,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5651,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(5651,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(5651,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(5651,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(5651,36,'cp 0','d2d4 g7g5 b1c3 h7h6 c3d5 g8f6 d1e2 c8g4 c2c3 f6d5 e4d5 c6b8 g2g3 e8f8 c1d2 g4h3 f1f2 b8d7 e2d3 e7f6 c4b3 d7b6 a1e1 h8g8 g3f4 g5f4 g1h1 f6h4 d3h7 h4f2 h7h6 g8g7 h6h8 g7g8');
INSERT INTO evaluations VALUES(5657,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5657,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(5657,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(5657,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(5657,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(5657,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(5657,36,'cp -26','g7g5 b1c3 h7h6 g2g3 c8h3 g3f4 g5g4 f3e1 e7f6 c1e3 h3f1 c4f1 d8d7 e4e5 f6g7 e1d3 h6h5 c3e4 g8e7 c2c3 b7b6 f1g2 d6d5 e4g5 e7f5 e3f2 g7h6 a2a4 c6e7 d1e2 h6g5 f4g5 h8g8 a4a5 g8g5 d3f4 h5h4 a5b6 c7b6 a1e1');
INSERT INTO evaluations VALUES(11542,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(11542,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11542,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(11542,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(11542,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11542,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(11542,36,'cp 0','b7b5 a4b3 f8c5 a2a4 a8b8 e1g1 d7d6 h2h3 e8g8 a4b5 a6b5 d2d3 c6e7 c1g5 e7g6 b1c3 h7h6 g5e3 c5e3 e2e3 g6f4 d3d4 c8h3 g2h3 f4h3 g1g2 h3f4 g2h1 f6g4 e3d2 d8f6 f1g1 f4h5 d2e2 h5f4');
INSERT INTO evaluations VALUES(11711,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11711,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11711,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11711,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11711,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11711,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11711,36,'cp 0','a4b3 f8c5 a2a4 a8b8 e1g1 e8g8 d2d3 d7d6 a4b5 a6b5 h2h3 c6e7 c1e3 c5e3 f2e3 e7g6 b1c3 c7c6 e2f2 d8b6 c3e2 d6d5 e2g3 c8e6 e4d5 c6d5 f3g5 b6c6 f2e2 c6b6');
INSERT INTO evaluations VALUES(11700,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11700,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11700,36,'cp 0','f8c5 a2a4 a8b8 a4b5 a6b5 e1g1 e8g8 d2d3 d7d6 h2h3 c6e7 b1c3 h7h6 c1e3 c5e3 f2e3 c8d7 f1f2 d7e6 b3e6 f7e6 b2b4 e7c6 a1b1 d8e8 f2f1 e8g6 g1h2 g6e8');
INSERT INTO evaluations VALUES(13003,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(13003,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(13003,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13003,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13003,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13003,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13003,36,'cp 0','a2a4 f8g7 e1g1 e8g8 f1e1 h7h6 c1e3 c8g4 b1d2 c6e5 h2h3 g4f3 d2f3 e5c4 d3c4 f6e4 e3h6 g7h6 e1e4 g8g7 d1d5 e7e6 d5b7 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(13160,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(13160,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(13160,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13160,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13160,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(13160,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(13160,36,'cp 12','e7e5 d4e5 c6e5 d1e2 d8e7 c1e3 c8d7 g1f3 e5d3 e2d3 e7e4 d3e2 f8d6 b1d2 e4e6 f3d4 e6e7 e3g5 e8g8 e2e7 d6e7 e1g1 f8e8 a2a4 h7h6 g5e3 a7a6 f1d1 e7f8 a4a5 a8c8 d2f3 f8c5 d4c2 d7f5 e3c5 f5c2 d1d2 c8c5 d2c2 g8f8');
INSERT INTO evaluations VALUES(13799,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(13799,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(13799,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(13799,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(13799,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(13799,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13799,36,'cp 8','d4e5 c6e5 d1e2 f6e4 d3b5 c8d7 g1f3 e5g6 f3d4 f8e7 b5d7 d8d7 b1d2 e4c5 d2b3 e8g8 e1g1 a7a6 b3c5 e7c5 c1e3 a8e8 e2d2 e8e4 a1d1 c5d4 e3d4 f8e8 f1e1 e4e1 d1e1 e8e1 d2e1 f7f6 e1d1 h7h6 a2a4 g6e7 d1e2 e7c6 e2d3 d7e6 g1f1');
INSERT INTO evaluations VALUES(13796,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(13796,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(13796,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(13796,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(13796,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13796,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13796,36,'cp 13','c6e5 d1e2 f6e4 d3b5 c8d7 g1f3 e5g6 f3d4 f8e7 b5d7 d8d7 b1d2 e4c5 d2b3 e8g8 c1e3 b7b6 a1d1 a8d8 e2b5 d7b5 d4b5 d8d7 e1g1 g6e5 f1e1 e5c4 e1e2 f8c8 e3f4 c5e6 f4e5 a7a6 b5d4 e6c5 d4f5 g7g6 b3c5 e7c5 f5h6 g8f8 h6g4 h7h5');
INSERT INTO evaluations VALUES(14024,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(14024,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(14024,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14024,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(14024,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14024,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(14024,36,'cp 13','d1e2 f6e4 d3b5 c8d7 g1f3 e5g6 f3d4 f8e7 b1d2 e4c5 b5d7 d8d7 d2b3 e8g8 c1e3 b7b6 a1d1 a8d8 e2b5 d7b5 d4b5 g6e5 b3c5 e7c5 b2b3 c5e3 f2e3 d8d7 b5d4 f8e8 e1g1 e5g6 d4f5 h7h6 d1d4 g6e7 c3c4 g8f8 f5e7 d7e7 d4d5');
INSERT INTO evaluations VALUES(17000,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(17000,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(17000,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(17000,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(17000,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(17000,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(17000,36,'cp 109','c4f7 e8f7 f3e5 f7g8 d1d4 c7c5 d4c4 d7d5 e4d5 d8d5 e1g1 d5c4 e5c4 b7b5 c4e3 h7h5 h2h3 c8b7 f1d1 h5h4 b1c3 a7a6 c3e2 g7g5 b2b3 h8h6 c1b2 a8e8 e2c1 f6h5 c1d3 h6e6 a2a4 b5b4 d1d2 f8d6 a1d1 g8f7 e3c4 d6c7 b2e5 e6e5');
INSERT INTO evaluations VALUES(21417,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(21417,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(21417,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(21417,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(21417,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(21417,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(21417,36,'cp 12','f8d6 c3c4 d5c4 b1c3 g8f6 f1c4 e8g8 e3e4 e6e5 d4d5 c6a5 c4d3 c7c6 e1g1 h7h6 c1d2 f8e8 f1e1 a7a6 a2a3 c6d5 c3d5 a5c6 d3c4 f6d5 e4d5 c6e7 f3e5 d8c7 d2b4 d6e5 d5d6 e5d6 d1d6 c7c4 e1e7 e8e7 d6e7 c4c6 a1d1');
INSERT INTO evaluations VALUES(24894,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(24894,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24894,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24894,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24894,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24894,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(24894,36,'cp 38','d5c4 h2h3 g4h5 d1a4 e7e6 b1d2 a7a6 a4c4 f8d6 a2a3 d8e7 b2b4 e6e5 c1b2 e5d4 f3d4 h5e2 e1e2 c6d4 b2d4 e8g8 h1d1 a8d8 d4f6 e7f6 d2e4 f6e5 e4d6 d8d6 d1d6 c7d6 a1d1 e5b2 e2f3 b7b5 c4d3 d6d5 g2g3 f8d8 h3h4 b2e5 f3g2 g7g6');
INSERT INTO evaluations VALUES(25733,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(25733,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(25733,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(25733,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(25733,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(25733,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(25733,36,'cp 55','c2c3 g7g6 d2d4 f8g7 c1e3 g8f6 b1d2 e8g8 h2h3 b7b5 a4c2 a6a5 d4e5 c6e5 f3e5 d6e5 a2a4 d8b8 d1c1 f8d8 d2f3 d8e8 f1d1 b8b7 f3d2 g7f8 c2b3 d7c6 a4b5 c6b5 c1c2 e8d8 d2f3 d8d1 a1d1 a5a4 b3a2 b7e4 c2e4');
INSERT INTO evaluations VALUES(30983,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30983,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(30983,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(30983,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(30983,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(30983,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(30983,36,'cp 24','f5e4 h1h3 e7e6 b1c3 e4h7 c1e3 g8e7 f1d3 h7d3 d1d3 c6c5 d4c5 b8d7 f2f4 e7c6 e1c1 f8c5 g1f3 d8b6 e3d2 a7a6 c1b1 e8c8 a2a3 c8b8 h4h5 d8c8 d2c1 b6c7 f3d2 c5a7 d2b3 b8a8 c1d2 h8f8 d2e3 a7e3 d3e3 d7b6 c3e2 b6c4');
INSERT INTO evaluations VALUES(30980,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(30980,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(30980,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(30980,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(30980,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(30980,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(30980,36,'cp 21','h1h3 e7e6 b1c3 e4h7 c1e3 g8e7 h4h5 c6c5 f1d3 c5d4 e3d4 h7d3 d1d3 e7c6 g1e2 c6d4 e2d4 b8c6 f2f4 d8b6 e1c1 c6d4 d3d4 b6d4 d1d4 a8c8 c3e2 b7b5 a2a3 f8c5 d4d1 e8g8 e2d4 c5d4 d1d4 c8c4 h3d3 f8c8 c2c3 a7a5 c1b1 c4d4 d3d4 c8c4 b1c2 c4d4 c3d4 g7g6 h5g6 f7g6 c2d3 g8f7');
INSERT INTO evaluations VALUES(31784,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(31784,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(31784,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(31784,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(31784,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(31784,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(31784,36,'cp 73','f2f3 c7c5 d4c5 d8a5 c1e3 g8f6 e3d4 b8c6 e2e3 e6d7 f1b5 e7e6 a2a3 a5c7 b5c6 d7c6 f3f4 c6d7 b2b4 b7b6 c5b6 a7b6 g1f3 f8e7 e1g1 e8g8 f3e5 f8c8 d1f3 d7e8 f3g3 e7d6 e5g4 f6g4 g3g4 f7f5 g4d1 g8f8 g2g4 f5g4 d1g4 c7f7 g4h3 f7g6 g1h1');
INSERT INTO evaluations VALUES(35311,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(35311,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(35311,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(35311,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(35311,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(35311,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(35311,36,'cp 20','b8c6 g1f3 a7a6 b5c4 b7b5 c4d5 a8b8 d5b3 e7e6 d2d4 f8d6 e1g1 c6a5 f3e5 d6e5 d4e5 c8b7 d1d7 e8d7 f1d1 d7c8 c1g5 g8h6 f2f3 h6f5 g1f2 c7c5 a2a3 h7h6 g5d2 h8d8 c3e2 a5c6 b3a2 g7g5 e2g3 f5g3 f2g3 c6e5 d2c3 e5g6 c3f6 d8g8');
INSERT INTO evaluations VALUES(37468,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37468,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(37468,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(37468,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37468,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(37468,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(37468,36,'cp 0','d2d4 b7b6 c2c4 c7c5 f1e2 c5d4 d1d4 b8a6 b1c3 c8b7 d4d2 e7e6 a1d1 a6c5 d2c1 d8e7 e1g1 d7d5 c4d5 f6d5 c3d5 b7d5 c1a1 g7b2 a1b2 d5f3 e2f3 a8d8 d1d8 f8d8 b2c2 c5d3 f1d1 d3b4 c2b2 d8d1 f3d1 b4d3 b2d4 d3b4 d4b2');
INSERT INTO evaluations VALUES(42390,30,'cp 571',NULL);
INSERT INTO evaluations VALUES(42390,31,'cp 576',NULL);
INSERT INTO evaluations VALUES(42390,32,'cp 571',NULL);
INSERT INTO evaluations VALUES(42390,33,'cp 563',NULL);
INSERT INTO evaluations VALUES(42390,34,'cp 578',NULL);
INSERT INTO evaluations VALUES(42390,35,'cp 578',NULL);
INSERT INTO evaluations VALUES(42390,36,'cp 571','d4e5 d6e5 f3e5 d8e7 e5f3 d7d5 e4e5 c8g4 c1e3 g4f3 d1f3 e7e5 b1c3 b8c6 e1c1 e8c8 f1b5 d5d4 b5c6 b7c6 e3f4 e5c5 c3a4 c5f5 h1e1 h8e8 e1e8 d8e8 d1d4 e8e1 c1d2 e1e4 f4e3 f5f3 g2f3 e4e5 d4c4 e5h5 h2h4 h5f5 e3a7 f5f3 d2e2 f3f5 a7e3 f6d5 a4c3 h7h6 c4c6 d5e3');
INSERT INTO evaluations VALUES(42380,30,'cp 578',NULL);
INSERT INTO evaluations VALUES(42380,31,'cp 578',NULL);
INSERT INTO evaluations VALUES(42380,32,'cp 575',NULL);
INSERT INTO evaluations VALUES(42380,33,'cp 553',NULL);
INSERT INTO evaluations VALUES(42380,34,'cp 571',NULL);
INSERT INTO evaluations VALUES(42380,35,'cp 576',NULL);
INSERT INTO evaluations VALUES(42380,36,'cp 563','d6e5 f3e5 d7d6 e5f3 f6e4 f1d3 e4c5 d3e2 e8g8 c1g5 d8e8 e1g1 b8c6 f1e1 c8e6 e2b5 f7f5 b1c3 c5e4 c3e4 f5e4 f3d4 e6d7 d4c6 b7c6 b5f1 a8b8 g5h4 b8b2 d1c1 b2b8 h4g3 g8h8 c1d2 h7h6 c2c4 d7g4 f2f3 g4f3');
INSERT INTO evaluations VALUES(42725,30,'cp 582',NULL);
INSERT INTO evaluations VALUES(42725,31,'cp 587',NULL);
INSERT INTO evaluations VALUES(42725,32,'cp 575',NULL);
INSERT INTO evaluations VALUES(42725,33,'cp 572',NULL);
INSERT INTO evaluations VALUES(42725,34,'cp 594',NULL);
INSERT INTO evaluations VALUES(42725,35,'cp 567',NULL);
INSERT INTO evaluations VALUES(42725,36,'cp 561','f3e5 d7d6 e5f3 f6e4 f1d3 e4c5 d3e2 e8g8 c1g5 d8e8 e1g1 c5e4 f1e1 e4g5 f3g5 e8d8 g5f3 c8g4 b1c3 b8d7 d1d2 d7f6 h2h3 g4h5 f3h2 h5g6 e2f3 c7c6 c3e4 g6e4 f3e4 d8b6 e4d3 b6b2 a1b1 b2a2 b1b7 a8e8 e1e8 f8e8');
INSERT INTO evaluations VALUES(42719,30,'cp 570',NULL);
INSERT INTO evaluations VALUES(42719,31,'cp 541',NULL);
INSERT INTO evaluations VALUES(42719,32,'cp 559',NULL);
INSERT INTO evaluations VALUES(42719,33,'cp 584',NULL);
INSERT INTO evaluations VALUES(42719,34,'cp 576',NULL);
INSERT INTO evaluations VALUES(42719,35,'cp 555',NULL);
INSERT INTO evaluations VALUES(42719,36,'cp 551','d8e7 e5f3 d7d5 e4e5 b8c6 c1e3 c6e5 f3e5 e7e5 d1d4 e5d4 e3d4 c8f5 b1a3 e8c8 e1c1 h8e8 f2f3 f5g6 d4a7 f6d7 a7f2 d7e5 f1b5 e5c6 b5c6 b7c6 d1d2 d5d4 f2d4 f7f6 a3c4 d8d5 h1d1 c6c5 d4f2 d5d2 d1d2 e8d8 d2d8 c8d8 f2c5 f6f5');
INSERT INTO evaluations VALUES(46165,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46165,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(46165,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46165,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(46165,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(46165,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(46165,36,'cp -17','c5d4 e3d4 f8b4 a1c1 e8g8 g1f3 b7b6 f4g5 c8b7 c4d5 d8d5 g5f6 g7f6 f1e2 b4c3 b2c3 d5a2 e1g1 b8c6 f3d2 a2d5 e2f3 d5f5 f1e1 g8h8 d2e4 a8d8 e4g3 f5g5 d1e2 f6f5 g3h5 d8d7 c1a1 c6e7 f3b7 d7b7 g2g3 f8d8 h5f4');
INSERT INTO evaluations VALUES(46312,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46312,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46312,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46312,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46312,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46312,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46312,36,'cp 0','e3d4 f8e7 c4c5 e8g8 g1f3 b7b6 b2b4 f6e4 a1c1 b6c5 b4c5 d8a5 f4d2 e4d2 d1d2 c8a6 c1b1 a6f1 e1f1 b8c6 g2g3 a8b8 f1g2 a5a6 h2h4 e7f6 h1c1 h7h6 h4h5 a6a5 c1e1 f8d8 e1c1 a7a6 c1e1 d8e8 e1c1');
INSERT INTO evaluations VALUES(46305,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46305,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46305,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46305,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46305,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46305,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46305,36,'cp 0','f8e7 g1f3 e8g8 c4c5 b7b6 b2b4 f6e4 a1c1 a7a5 f1d3 e4c3 c1c3 a5b4 c5c6 d8e8 c6c7 b8c6 c3c1 b4b3 a2b3 c6b4 d3e2 b4a2 c1c2 a2b4');
INSERT INTO evaluations VALUES(47732,30,'cp 417',NULL);
INSERT INTO evaluations VALUES(47732,31,'cp 436',NULL);
INSERT INTO evaluations VALUES(47732,32,'cp 444',NULL);
INSERT INTO evaluations VALUES(47732,33,'cp 435',NULL);
INSERT INTO evaluations VALUES(47732,34,'cp 439',NULL);
INSERT INTO evaluations VALUES(47732,35,'cp 430',NULL);
INSERT INTO evaluations VALUES(47732,36,'cp 443','e5f7 d8e7 d1e2 e7e2 e1e2 e8f7 c4d5 f7g6 h2h4 h6h5 d5e4 g6f7 d2d3 b8d7 c1f4 d7f6 b1d2 f6e4 d2e4 c8f5 e2d2 a8d8 f4c7 d8d7 e4g5 f7g6 c7e5 f8c5 f2f3 h8d8 a1e1 c5f2 e1e2 d7d3 c2d3 d8d3 d2c1 f2e3 e2e3 d3e3 e5c3 e3e7 g5e4 f5e4 f3e4 e7e4');
INSERT INTO evaluations VALUES(47930,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(47930,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47930,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(47930,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(47930,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(47930,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(47930,36,'cp 25','f6e4 d1e2 d8e7 e1g1 e4f6 e2d1 d6d5 c4b5 c7c6 f1e1 c8e6 b5f1 e7d7 b1c3 f8e7 g2g3 b8a6 f1a6 b7a6 d1e2 a8b8 d4e6 f7e6 e2e6 d7e6 e1e6 e8d7 e6e1 e7b4 c1f4 b4d6 f4d6 d7d6 b2b3 h8e8 f2f3 g7g5 a1d1 c6c5 e1e8 b8e8 g1f2 d6c6 d1e1 e8e1 f2e1 c5c4 b3c4 d5c4 c3e4 f6e4 f3e4');
INSERT INTO evaluations VALUES(50129,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(50129,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(50129,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(50129,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(50129,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(50129,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(50129,36,'cp 53','g2g3 d7d6 b1c3 c7c6 f1g2 f8g7 e1g1 e8g8 a1b1 a7a5 c1e3 f6g4 e3d2 g4f6 d4d5 e7e5 d5e6 c8e6 b2b3 b8a6 d2f4 f6h5 f4g5 d8c7 f3d4 a6c5 d4e6 c5e6 g5d2 c7e7 e2e3 e6c5 d1c2 h5f6 b1d1 f6e4 c3e2 e4d2 d1d2 a5a4');
INSERT INTO evaluations VALUES(52726,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(52726,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(52726,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(52726,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(52726,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(52726,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(52726,36,'cp -34','e4e5 d8a5 c2c3 c6c5 b1a3 g8h6 f1b5 b8c6 h5e2 c8g4 g1f3 e7e6 h2h3 g4f3 e2f3 h6f5 e1g1 f8e7 d2d3 a5c7 f3e2 e8g8 b5c6 c7c6 a3c2 b7b5 c2e3 f5e3 c1e3 a7a5 e2g4 g8h8 d3d4 c5c4 f2f4 g7g6 g4e2 f8b8 g2g4 b5b4 f4f5 g6f5 g4f5 b4c3 b2c3 e6f5');
INSERT INTO evaluations VALUES(55775,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(55775,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(55775,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(55775,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(55775,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(55775,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(55775,36,'cp -6','e6d5 d2d4 f8d6 f1d3 g8f6 g1e2 e8g8 e1g1 f8e8 b1d2 c7c5 d4c5 d6c5 d2b3 c5b6 f1e1 b8c6 e2d4 a7a5 c1f4 e8e1 d1e1 c8d7 a2a4 b6d4 b3d4 c6d4 c3d4 d8b6 e1d2 a8e8 f4e3 f6e4 d2c2 h7h6 h2h3 b6d6 f2f3 e4g3 e3f2');
INSERT INTO evaluations VALUES(56277,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(56277,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(56277,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56277,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56277,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56277,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56277,36,'cp 0','d2d4 f8d6 g1f3 g8f6 f1d3 d8e7 d3e2 e8g8 e1g1 c8g4 h2h3 g4h5 f1e1 b8c6 c1g5 a8e8 b1d2 h7h6 g5f6 e7f6 d1b3 c6e7 d2f1 e7g6 f1g3 d6g3 f2g3 c7c6 g3g4 g6f4 f3e5 f6h4 g4h5 h4g3 e2f1 f4h3 g1h1 h3f2 h1g1');
INSERT INTO evaluations VALUES(4381,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4381,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4381,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4381,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4381,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(4381,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(4381,36,'cp -5','c3a4 c5d6 a2a3 f8e8 c2c4 c6d4 c1e3 c7c5 a4c3 a8b8 f1e2 a7a6 e3g5 b7b5 c4b5 d4b5 c3d5 d6e7 d5e7 d8e7 e1g1 h7h6 g5h4 d7d5 e4d5 c8b7 a1c1 b7d5 h4f6 g7f6 d1d2 g8g7 f3h4 d5e6 e2g4 f6f5');
INSERT INTO evaluations VALUES(6571,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(6571,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(6571,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(6571,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(6571,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(6571,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(6571,36,'cp -47','g8f6 c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g1f3 a7a5 f1b5 d7d6 h1f1 h8f8 f3d2 c8e6 e1c1 e8c8 d2b3 e6b3 a2b3 c6d4 b5d3 f8e8 f2f3 c7c6 c3e2 d4e2 d3e2 d6d5 e2d3 d5e4 d3e4 c5d6 g2g3 c8c7 c2c3 b7b5 c1c2');
INSERT INTO evaluations VALUES(7238,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(7238,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(7238,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(7238,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(7238,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(7238,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(7238,36,'cp 163','e8g8 d2d3 b7b5 e1g1 c8g4 f1e1 f8e8 c1g5 h7h6 g5h4 g4h5 a2a3 c6c5 h2h3 g7g5 h4g3 d6f8 d1e2 f8g7 e2e3 b5b4 a3b4 c5b4 c3a4 a6a5 f3e5 f6d7 e5d7 d8d7 e4e5 d7c6 d3d4 f7f5 f2f4 g5f4 g3f4 c6c2');
INSERT INTO evaluations VALUES(2632,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(2632,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(2632,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(2632,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(2632,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(2632,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(2632,36,'cp 177','e1g1 f8e8 d2d3 c8g4 h2h3 g4h5 c1g5 h7h6 g5h4 g7g5 h4g3 d6f8 f1e1 f8g7 c3e2 h5f3 g2f3 a6a5 c2c3 f6h5 a2a4 c6c5 d1c2 f7f5 g1g2 f5f4 g3h2 d8f6 d3d4 a8d8 c2b3 f6f7 b3b7 d8d6 b7a7 d6b6 a1d1 c5d4 e2d4 b6b2');
INSERT INTO evaluations VALUES(7292,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(7292,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(7292,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(7292,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(7292,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(7292,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(7292,36,'cp 184','b5c6 d7c6 f3e5 f6e4 d1e2 d8h4 d2d4 c5d6 g2g3 h4e7 e2e4 f7f6 f1e1 f6e5 d4e5 d6c5 c1e3 c8e6 b1d2 c5e3 e4e3 a8d8 f2f4 e8g8 d2f3 e6g4 e1e2 e7e6 e2f2 e6c4 f3d2 c4d5 e3e4 b7b6 e4a4 a6a5 a1e1 c6c5 d2c4 g4e6 c4e3 d5a2 a4c6');
INSERT INTO evaluations VALUES(10779,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(10779,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(10779,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(10779,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(10779,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(10779,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(10779,36,'cp 84','e3d4 d8e7 f1e2 e7d6 c3b5 d6b6 e4e5 f6e4 a2a4 a7a6 a4a5 b6c6 b5c3 d7d5 e1g1 c5e7 e2d3 e4c3 d4c3 c8e6 f1e1 e8c8 c3d2 c6d7 b2b4 e6f5 d1e2 f5d3 c2d3 c8b8 d3d4 d7b5 e2f3 e7b4 d2b4 b5b4 f3f7 b4d4');
INSERT INTO evaluations VALUES(10778,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(10778,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(10778,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(10778,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(10778,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(10778,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(10778,36,'cp 56','d8e7 f1e2 e7d6 c3b5 d6b6 e4e5 f6e4 d4c5 b6c5 d1d4 c5d4 b5d4 d7d5 e2f3 c8d7 f3e4 d5e4 e1c1 c7c5 d4b3 b7b6 h1e1 d7f5 h2h3 a8d8 b3d2 d8d4 b2b3 e8e7 c2c3 d4d5 d2e4 d5d1 c1d1 h8d8 e4d6 f7f6 d1c1 f5d3 d6c4 d3c4 e5f6 e7f6 b3c4 h7h5 h3h4');
INSERT INTO evaluations VALUES(11374,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(11374,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(11374,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(11374,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11374,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(11374,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11374,36,'cp 52','f8d6 b1d2 f6d7 e1g1 e6e5 d4e5 d7e5 d3c2 e5g6 f1e1 e8g8 f4d6 d8d6 e3e4 c8e6 e4d5 d6d5 c2e4 d5b5 d1c2 e6d5 a2a4 d5e4 e1e4 b5f5 a1e1 h7h6 b2b4 a8d8 g2g3 d8d5 g1g2 c6e5 f3e5 g6e5');
INSERT INTO evaluations VALUES(7164,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(7164,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(7164,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(7164,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(7164,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(7164,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(7164,36,'cp 62','b1d2 f6d7 f4g3 e6e5 e3e4 e5d4 c3d4 c6b4 d3b1 d5e4 d2e4 e8g8 e4d6 c7d6 e1g1 d7f6 f3d2 d6d5 f1e1 f8e8 e1e8 f6e8 h2h3 a6a5 d2f3 c8e6 d1d2 f7f6 a2a3 b4c6 b1d3 e8d6 d2f4 d6e4 a1c1 a5a4 d3e4 d5e4 f4e4');
INSERT INTO evaluations VALUES(14202,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(14202,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(14202,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(14202,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(14202,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14202,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14202,36,'cp 0','d5c4 g1f3 a7a6 d4d5 e7e5 f4g3 c6a5 g3e5 b7b5 g2g3 f8d6 f1g2 e8g8 d1d4 f8e8 e1g1 a5b7 e5d6 b7d6 f3d2 d6f5 d4f4 f6h5 f4f3 h5f6');
INSERT INTO evaluations VALUES(14374,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(14374,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14374,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14374,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14374,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(14374,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14374,36,'cp 9','e2e3 c6a5 f1c4 a5c4 d1a4 c7c6 a4c4 c8e6 c4d3 f6h5 g1e2 g7g6 e1g1 e6f5 e3e4 h5f4 e2f4 f5d7 a1d1 f8g7 e4e5 e8g8 f1e1 d8b6 d4d5 a8d8 d3e2 d7c8 e5e6 f7e6 f4e6 g7c3 b2c3 c8e6 d1b1 b6c5 e2e6 f8f7 e6e2 d8d7 d5c6 c5c6 e2e5');
INSERT INTO evaluations VALUES(14372,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(14372,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(14372,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(14372,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(14372,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(14372,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(14372,36,'cp 6','c6a5 d1a4 c7c6 f1c4 a5c4 a4c4 c8e6 c4d3 f6h5 g1e2 g7g6 e1g1 e6f5 e3e4 h5f4 e2f4 f5d7 a1d1 f8g7 e4e5 e8g8 d3e3 d8b6 f4d3 d7f5 h2h3 a8d8 d3c5 b6c7 f1e1 b7b6 c5a6 c7b7 a6b4 a7a5 b4d3 f5e6 d3f4 g7h6 e3f3 e6c8 f4e2 g8h8 a2a3 b7c7 h3h4');
INSERT INTO evaluations VALUES(14308,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(14308,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(14308,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(14308,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(14308,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(14308,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(14308,36,'cp -32','e7e5 g1f3 e5e4 f3e5 d8d6 f1b5 f6d7 e5c6 b7c6 b5f1 d7b6 b2b3 d6g6 c1d2 f8d6 d2b4 e8g8 b4d6 g6d6 c2c4 d5c4 b3c4 c6c5 b1c3 c8f5 f1e2 f8d8 d4d5 c7c6 e1g1 c6d5 c3d5 d6e5 d1c1 b6d5 c4d5 e5d5 c1c3 a8c8 f1d1 d5c6');
INSERT INTO evaluations VALUES(15118,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(15118,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(15118,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(15118,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(15118,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(15118,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(15118,36,'cp -29','g1f3 e5e4 f3e5 d8d6 f1b5 f6d7 e5c6 b7c6 b5f1 d7b6 b2b3 d6g6 c1d2 f8d6 d2b4 e8g8 b4d6 g6d6 c2c4 d5c4 b3c4 d6g6 b1c3 a8b8 c3e2 c6c5 d4c5 b6c4 d1d4 g6a6 d4c3 c8e6 e2d4 a6a5 c3a5 c4a5 f1e2 a5b3 d4b3 b8b3');
INSERT INTO evaluations VALUES(15245,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(15245,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(15245,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(15245,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(15245,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(15245,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(15245,36,'cp -51','f8d6 h2h3 d5e7 b5a4 e8g8 e1g1 e7g6 f1e1 a7a6 c2c3 c8f5 a4c2 d6e7 b1a3 e7a3 b2a3 h7h6 d3d4 d8f6 f3e5 c6e5 d4e5 g6e5 c1e3 f8d8 e3d4 f5c2 d1c2 f6g6 c2g6 e5g6 a1d1 g6f4 a3a4 f4e6 d1b1 b7b6 a4a5 b6a5 d4e3 d8d3 e1c1 a5a4');
INSERT INTO evaluations VALUES(17296,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(17296,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(17296,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(17296,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(17296,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(17296,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(17296,36,'cp 35','d2d4 d7d5 g1f3 c8f5 c2c4 e7e6 e1g1 f8e7 b1c3 d5c4 d1a4 e8g8 a4c4 f6e4 c1f4 e4c3 b2c3 f5e4 f3e1 c6a5 c4c7 d8c7 f4c7 e4g2 e1g2 b7b6 c7f4 f8c8 f1c1 g7g5 f4d2 a5c4 d2e1 g5g4 a2a4 f7f5 g2e3 c4e3 f2e3 e7g5 g1f2 g8f7');
INSERT INTO evaluations VALUES(18614,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(18614,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(18614,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(18614,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(18614,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(18614,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(18614,36,'cp 17','f8g7 d2d4 c5d4 c3b5 d7d6 b5d4 c6d4 f3d4 g8f6 f1e1 e8g8 c4f1 a7a6 c2c4 c8d7 a2a4 f6g4 h2h3 d8b6 a4a5 b6d4 h3g4 d4d1 e1d1 d7c6 a1a3 c6e4 b2b4 f8d8 c1e3 g7e5 b4b5 a6b5 c4b5 d6d5 g4g5 d5d4 e3d2 e5d6 a3a1 e6e5 a5a6 b7a6 b5a6');
INSERT INTO evaluations VALUES(18838,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(18838,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18838,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18838,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(18838,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(18838,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18838,36,'cp 25','g8f6 b1c3 d7d5 e4d5 e6d5 d2d4 c8e6 c1g5 h7h6 g5f6 d8f6 d4c5 e8c8 c3a4 g7g5 c2c3 d5d4 f3d4 c6d4 c3d4 d8d4 d1c2 f6f4 a4c3 d4d2 c2c1 d2d4 c5c6 b7c6 c3b5 f4c1 f1c1 d4d2 c1c6 c8b8 e2f3 d2b2 b5c7 b2b6 c7e6 f7e6 c6b6 a7b6 a1e1');
INSERT INTO evaluations VALUES(12440,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(12440,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(12440,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(12440,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(12440,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(12440,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(12440,36,'cp 23','b1c3 d7d5 e4d5 e6d5 d2d4 c8e6 h2h3 f8d6 c3b5 c5c4 b5d6 d8d6 f1e1 e8g8 f3e5 f6e4 c1f4 c6e5 f4e5 d6c6 a2a4 f8e8 c2c3 e8e7 f2f3 e4d6 g2g4 f7f6 e5g3 e6f7 d1d2 a8e8 e2f1 g7g5 e1e7 e8e7 a4a5 a7a6');
INSERT INTO evaluations VALUES(12425,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(12425,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(12425,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(12425,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(12425,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12425,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12425,36,'cp 14','d7d5 e4d5 e6d5 d2d4 c8e6 h2h3 f8d6 c3b5 c5c4 b5d6 d8d6 f1e1 e8g8 f3e5 f6e4 c1f4 c6e5 f4e5 d6c6 e5f4 f8e8 f2f3 e4d6 g2g4 f7f6 d1d2 e6f7 e2f1 f7g6 d2h2 d6f7 h3h4 h7h5 f1h3 e8e1 a1e1 a8e8');
INSERT INTO evaluations VALUES(20042,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20042,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(20042,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(20042,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(20042,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20042,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20042,36,'cp 0','a2a4 g8f6 g1f3 a7a6 f1e2 e7e6 e1g1 f8e7 d1e1 e8g8 e1g3 f6h5 g3f2 h5f6');
INSERT INTO evaluations VALUES(23147,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(23147,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(23147,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(23147,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(23147,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(23147,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(23147,36,'cp 116','d1d4 d8f6 d4e3 f8e7 e3g3 c7c6 e2e4 f6g6 f2f3 g8f6 e4e5 f6d5 c3d5 c6d5 f1d3 g6g3 h2g3 d7d6 e5d6 e7d6 g3g4 c8e6 c1e3 h7h6 e1c1 e8e7 d1e1 h8c8 c1b1 d6c5 e3d2 c8e8 d2c3 e7f8 g4g5 d5d4 c3d2 h6g5 h1h7 f8g8 d2g5 f7f6');
INSERT INTO evaluations VALUES(24080,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24080,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24080,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(24080,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(24080,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24080,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(24080,36,'cp 16','c4e6 f7e6 c3a4 c5b6 a4b6 a7b6 a2a4 e8g8 c1d2 c6e7 b2b4 d8d7 b4b5 e7g6 d1e2 d7f7 g2g3 f6d7 g1g2 d7c5 a4a5 b6a5 a1a5 a8a5 d2a5 f8a8 a5d2 g6f8 d2e3 f8d7 e2d2 b7b6 d2c3 a8f8 f3h2 d6d5 c3b4 f7g6 h2g4 d5e4 e3c5 d7c5');
INSERT INTO evaluations VALUES(24079,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24079,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(24079,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24079,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(24079,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(24079,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24079,36,'cp 13','f7e6 c3a4 c5b6 a4b6 a7b6 a2a4 e8g8 c1d2 d8e8 b2b4 f6h5 g1h2 e8f7 c2c3 h5f4 d2e3 f7h5 f3g1 h5f7 g2g3 f4h5 h2g2 f7g6 d1g4 g6f7 g4e2 f7g6 e2d1 h5f6 g1f3 f8f7 f3h4 g6h7 g2h2 g7g5 h4g2 d6d5 f2f3');
INSERT INTO evaluations VALUES(24166,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(24166,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(24166,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(24166,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(24166,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24166,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24166,36,'cp 24','c3a4 c5b6 a4b6 a7b6 a2a4 e8g8 c2c3 d8d7 a1a3 c6e7 b2b4 e7g6 c3c4 d7e7 f3h2 f8f7 a4a5 b6a5 a3a5 f7f8 d1a4 a8a5 b4a5 f8a8 c1e3 e7d7 a4d7 f6d7 f1b1 a8a5 b1b7 g6f4 b7c7 a5a1 h2f1 d7f6 e3f4 e5f4');
INSERT INTO evaluations VALUES(25028,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(25028,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(25028,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(25028,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(25028,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(25028,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(25028,36,'cp 91','f1e2 e7e6 c2c4 d6d5 e1g1 d5c4 f4g3 c6e7 b1d2 e7f5 d2c4 f5g3 h2g3 f8e7 b2b4 e8g8 a2a3 h7h5 f3e5 g4e2 d1e2 g7g6 c4a5 a8b8 a1c1 c7c6 f1d1 g8g7 a5b3 d8d5 b3c5 e7c5 b4c5 f6e4 e2b2 f8h8 f2f3 e4g3 e3e4');
INSERT INTO evaluations VALUES(25718,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(25718,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(25718,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(25718,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(25718,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(25718,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(25718,36,'cp 162','b5e2 g7g6 h2h3 g8e7 b1c3 f8g7 e1g1 e8g8 f1e1 h7h6 c1e3 g6g5 a1c1 f7f5 e4f5 d7f5 d4d5 c6e5 f3e5 g7e5 e2h5 d8d7 e3d4 e5d4 d1d4 e7g6 c3e4 g6e5 d4c3 d7h7 e4g3 a8c8 e1e5 d6e5 c3e5 f5g6 h5g4 h7g7 e5g7 g8g7 g4c8 f8c8');
INSERT INTO evaluations VALUES(26072,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26072,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26072,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26072,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26072,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26072,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26072,36,'cp 0','c2c4 g7g6 c1d2 f8g7 b1c3 c6d4 a4d1 d7g4 f1e2 d4e2 d1e2 a8c8 e1g1 g4f3 g2f3 d8d7 a1d1 d7h3 c3d5 g8f6 d5f4 h3h4 f4g2 h4h5 g2f4');
INSERT INTO evaluations VALUES(33266,30,'cp 530',NULL);
INSERT INTO evaluations VALUES(33266,31,'cp 556',NULL);
INSERT INTO evaluations VALUES(33266,32,'cp 551',NULL);
INSERT INTO evaluations VALUES(33266,33,'cp 531',NULL);
INSERT INTO evaluations VALUES(33266,34,'cp 528',NULL);
INSERT INTO evaluations VALUES(33266,35,'cp 542',NULL);
INSERT INTO evaluations VALUES(33266,36,'cp 562','c3d5 f6d6 c2c3 b4a5 d2d4 f7f6 f3d2 c7c6 d2c4 d6f8 c4a5 c6d5 e4d5 d7d6 f1d3 f8e7 e1g1 e7c7 d1h5 e8f8 a5c4 b7b5 c4e3 b8d7 d3b5 g7g6 h5d1 f8g7 c3c4 h7h5 f2f4 e5e4 b2b4 f6f5 c4c5 a8b8 b5c6 b8b4 c1a3 b4b8 c5d6');
INSERT INTO evaluations VALUES(33260,30,'cp 559',NULL);
INSERT INTO evaluations VALUES(33260,31,'cp 559',NULL);
INSERT INTO evaluations VALUES(33260,32,'cp 561',NULL);
INSERT INTO evaluations VALUES(33260,33,'cp 557',NULL);
INSERT INTO evaluations VALUES(33260,34,'cp 559',NULL);
INSERT INTO evaluations VALUES(33260,35,'cp 562',NULL);
INSERT INTO evaluations VALUES(33260,36,'cp 565','f6d6 c2c3 b4a5 d2d4 f7f6 f3d2 c7c6 d2c4 d6f8 c4a5 c6d5 e4d5 g8e7 a5c4 e7f5 f1d3 b7b5 c4a3 d7d6 d1g4 g7g6 d4e5 d6e5 a3b5 b8a6 b2b4 f8e7 d5d6 e7f7 a2a4 c8e6 f2f4 f5d4 g4h4 d4f5 d3f5 g6f5 f4e5 f6e5 e1g1 h8g8 c1h6 e8d7 a1e1');
INSERT INTO evaluations VALUES(33163,30,'cp 568',NULL);
INSERT INTO evaluations VALUES(33163,31,'cp 558',NULL);
INSERT INTO evaluations VALUES(33163,32,'cp 568',NULL);
INSERT INTO evaluations VALUES(33163,33,'cp 557',NULL);
INSERT INTO evaluations VALUES(33163,34,'cp 579',NULL);
INSERT INTO evaluations VALUES(33163,35,'cp 557',NULL);
INSERT INTO evaluations VALUES(33163,36,'cp 561','c2c3 b4a5 d2d4 f7f6 f3d2 c7c6 d2c4 d6f8 c4a5 c6d5 e4d5 d7d6 f1d3 g8e7 a5c4 e7f5 e1g1 b7b5 c4e3 f5e3 c1e3 f8f7 d3b5 c8d7 b5d3 e8g8 f2f3 d7c8 d1c2 f7h5 c3c4 b8d7 f3f4 e5e4 d3e4 f6f5 e4f3 h5f7 e3f2 d7f6 b2b4 f6e4 a2a4 c8b7 b4b5 a8e8 a4a5 e4f2 c2f2 f7g6');
INSERT INTO evaluations VALUES(42483,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(42483,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(42483,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(42483,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42483,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(42483,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42483,36,'cp 28','b4e7 g1f3 d7d5 b1c3 e8g8 c4d5 e6d5 f1d3 c7c6 e1g1 b8d7 d1c2 e7d6 a1e1 h7h6 e1e2 f8e8 f1e1 a7a6 h2h3 c6c5 d3f5 c5d4 e3d4 e8e2 c3e2 d7f8 f5d3 c8d7 d2f4 d6f4 e2f4 f8e6 f4e6 d7e6 c2b3 d8c7 f3e5 a8c8');
INSERT INTO evaluations VALUES(43226,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(43226,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43226,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43226,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(43226,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43226,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(43226,36,'cp 5','e2e3 e8g8 f1e2 b7b6 c2c4 c8b7 g1f3 c7c5 e1g1 c5d4 f3d4 d7d5 b1d2 b8d7 a1c1 a8c8 b3b4 a7a6 d4f3 c8c7 c4d5 f6d5 b2g7 g8g7 d1b3 g7g8 b3b2 c7c1 f1c1 d8a8 e2f1 b6b5 a2a3 f8c8 c1c2 h7h6 c2c8 a8c8 d2b3 d5b6 f3e5 b7d5 e5d7');
INSERT INTO evaluations VALUES(45834,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(45834,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45834,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(45834,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(45834,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(45834,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(45834,36,'cp 27','b1c3 c8e6 f3g3 d8f6 c4d5 c6d5 g1f3 b8c6 c1g5 f6f5 f3h4 f5g4 g3g4 e6g4 c3d5 a8c8 h2h3 g4e6 d5c3 f7f6 g5e3 e8f7 h4f3 b7b5 a2a3 a7a6 e1c1 h7h5 d3d4 e5d4 f3d4 c6d4 e3d4 f8d6 c3e4 d6f4 c1b1 h8d8 d4c3 e6f5 f2f3 f5g6 c3b4 f4e5 b4a5 d8d1 h1d1 g6e4');
INSERT INTO evaluations VALUES(45833,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45833,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45833,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(45833,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(45833,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45833,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(45833,36,'cp 5','f8b4 g1e2 e8g8 a2a3 b4c3 e2c3 c8e6 f3g3 f7f6 e1g1 b8d7 f1e1 d8e7 d3d4 d5c3 g3c3 e6c4 c3c4 e7f7 c4b4 e5d4 b4d4 d7b6 b2b3 f8d8 d4c5 f7d5 c5c3 d5d4 c3d4 d8d4 c1e3 d4d8 a1d1 b6d5 e3d2 b7b5 d2a5 d8e8 c2c4 b5c4 e1e8 a8e8 b3c4 d5b6 a5b6 a7b6');
INSERT INTO evaluations VALUES(46576,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(46576,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(46576,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(46576,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(46576,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(46576,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(46576,36,'cp 54','g1f3 h7h6 h2h4 g7g6 b1c3 f8g7 f1d3 b8c6 d1e2 e7e6 e3e4 f5e4 c3e4 e8g8 f4d2 e6e5 d4e5 f6g4 e5d6 c7d6 e4c3 f8e8 d2e3 d6d5 e1c1 g4e3 f2e3 d8b6 d3g6 e8e3 e2b5 g7c3 b5b6 a7b6 b2c3 e3c3 d1d5 a8a2 h1e1 g8g7 h4h5 c8g4 c1d2 c3c5 d5c5 b6c5 e1e8 a2a1');
INSERT INTO evaluations VALUES(46577,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(46577,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(46577,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(46577,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(46577,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(46577,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(46577,36,'cp 55','g7g6 f1b5 c8d7 c2c4 f8g7 e1g1 h7h6 b1c3 e8g8 h2h3 d7b5 c3b5 d8d7 b5c3 b8c6 f4h2 e7e6 b2b4 g6g5 d1b3 c6e7 a2a4 a7a6 a1b1 g8h8 c4c5 f6d5 c3d5 e7d5 a4a5 f8c8 b1c1 f5f4 e3f4 g5f4 f1e1');
INSERT INTO evaluations VALUES(46596,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(46596,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(46596,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(46596,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(46596,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(46596,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(46596,36,'cp 124','e7e6 c2c4 f8d6 b1c3 e8g8 a1c1 b8d7 d1c2 d5c4 f1c4 d7b6 c4b3 b6d5 f3e5 d5c3 c2c3 d8e8 e1g1 a7a5 a2a3 b7b6 c3c2 g8h8 c2c6 c8a6 f1d1 f6h5 c6e8 a8e8 f4g5 h5f6 h2h4 f6e4 g5f4 e4f6 e5d3 f6d5 f4e5 a6d3 d1d3');
INSERT INTO evaluations VALUES(47171,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(47171,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(47171,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(47171,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(47171,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(47171,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(47171,36,'cp -69','c7c5 e2e3 c5d4 e3d4 c8f5 f1b5 b8c6 g1f3 e7e6 f3e5 f6d7 e5c6 b7c6 b5d3 f8b4 c1d2 d8a5 c3e2 f5d3 c2d3 a8b8 d2b4 a5b4 d1d2 e8e7 a1c1 h8c8 d2b4 b8b4 e1d2 e7d6 c1c3 a7a5 h1c1 a5a4 h2h3 f7f6 f2f4 a4a3 c1c2 c6c5 d2e3 e6e5 d4c5 d7c5 d3d4');
INSERT INTO evaluations VALUES(50146,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50146,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(50146,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(50146,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(50146,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50146,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(50146,36,'cp 2','c7c6 e2e4 d8a5 c1d2 f8g7 f1d3 d7d6 g1f3 e8g8 e1g1 c8g4 d3e2 b8d7 f1e1 c6d5 c3d5 a5d8 d5f6 g7f6 c2c3 d8c7 a2a4 g4f3 g2f3 a7a6 d2e3 d7e5 d1b3 a8c8 a1d1 b7b5 a4b5 a6b5 e3h6 f6g7 h6e3');
INSERT INTO evaluations VALUES(51475,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(51475,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51475,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(51475,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(51475,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51475,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51475,36,'cp 0','e7e6 b1c3 g8f6 e2e4 f8b4 d5e6 f7e6 e4e5 f6d5 f1d3 d8h4 c1d2 d5c3 b2c3 b4c5 g1h3 c5e7 h3f4 c8b7 g2g3 h4h6 f4g6 h6h3 g6f4');
INSERT INTO evaluations VALUES(52206,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(52206,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(52206,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(52206,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(52206,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(52206,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(52206,36,'cp 187','d7d5 e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 e1g1 g8e7 b1c3 d8d6 c1e3 d6c6 f3d4 c6d7 d1e2 e8f7 a1d1 h7h5 f1e1 a8e8 d4b3 d7g4 f2f3 g4c4 e3a7 c4e2 e1e2 h5h4 h2h3 e5c4 c3d5 e7d5 e2e8 f7e8 d1d5 c4b2 d5b5');
INSERT INTO evaluations VALUES(54641,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(54641,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(54641,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(54641,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(54641,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(54641,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(54641,36,'cp 189','e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 e1g1 f8d6 d1d5 g8e7 d5b7 a8b8 b7e4 e5f3 e4f3 d8c7 h2h3 c7c2 b1a3 c2c6 f3c6 e7c6 a3c4 d6c5 c1f4 b8d8 f1d1 e8f7 c4d6 c5d6 f4d6 f7e6 d6a3 c6e5 b2b3 e5d3 a1b1 d8d5 d1d2 h8b8 g1f1 d5a5 d2d3 a5a3 d3e3 e6f7');
INSERT INTO evaluations VALUES(54636,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(54636,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(54636,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(54636,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(54636,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(54636,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(54636,36,'cp 182','c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 e1g1 g8e7 b1c3 d8d6 f1e1 e8c8 c1e3 a7a6 c3a4 c8b8 c2c3 e7c6 f3e5 f6e5 e3b6 d8d7 b2b4 e5e4 b6e3 d7d8 a1b1 d5d4 c3d4 c6d4 d1h5 f8e7 e1d1 d6f6 a4c5 f6f5 c5d7 f5d7 e3d4 d7e8 h5e2 e8g6 b4b5 a6b5');
INSERT INTO evaluations VALUES(54669,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(54669,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(54669,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(54669,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(54669,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(54669,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(54669,36,'cp 187','c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 e1g1 f8d6 d1d5 g8e7 d5b7 a8b8 b7e4 e5f3 e4f3 d8c7 h2h3 c7c2 b1a3 c2c6 f3c6 e7c6 a3c4 d6c5 c1f4 b8d8 f1d1 e8e7 f4e3 c5e3 c4e3 c6e5 b2b3 e7e6 g1f1 e5d3 d1d2 d8d7 a1d1 h8d8 g2g3 h7h5 h3h4 g7g5 h4g5 f6g5 e3c4 h5h4 f1e2 d3e5 d2d7 d8d7 c4e5 d7d1 e2d1');
INSERT INTO evaluations VALUES(1678,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(1678,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(1678,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(1678,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(1678,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(1678,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(1678,36,'cp 108','d5d8 f1b5 f8d6 e1g1 g8e7 f1e1 e8g8 c3e4 h7h6 c2c4 e7g6 b5a4 c8d7 c1d2 b7b6 d2c3 c6a5 a4d7 d8d7 c3a5 b6a5 d1e2 f8d8 h2h4 a8b8 g2g3 d7c6 a1c1 d6e7 e1d1 e7d6 d1d2 g6e7 h4h5 e7f5 d2d3');
INSERT INTO evaluations VALUES(5656,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(5656,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(5656,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(5656,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(5656,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(5656,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(5656,36,'cp 62','g8f6 c4d3 d6d5 e4e5 f6e4 b1c3 c8f5 e1g1 d8d7 c1e3 e4c3 b2c3 e8g8 d1b1 f5d3 b1d3 d7e6 a1e1 a7a6 e3g5 a8c8 e1b1 b7b6 g5h4 b6b5 a2a4 b5a4 b1a1 e6d7 f1e1 h7h6 h2h3 c8b8 h4e7 c6e7 d3c2 c7c5 a1a4 f8c8 d4c5 d7e6');
INSERT INTO evaluations VALUES(8646,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(8646,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(8646,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(8646,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(8646,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(8646,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(8646,36,'cp 91','c4d3 d6d5 e4e5 f6e4 b1c3 c8f5 e1g1 d8d7 c3e2 c6b4 d3b1 b4c6 c1e3 e8g8 b1c2 c6b4 c2a4 b7b5 a4b3 a7a5 a2a3 a5a4 a3b4 a4b3 d1b3 a8a4 e2f4 f8d8 f4d3 d8a8 a1c1 c7c6 h2h3 h7h6 f1e1 a8b8 d3c5 e7c5 b4c5 a4c4 b3d1 b5b4 f3h4 f5h7 f2f3 e4g5 c1c4 d5c4');
INSERT INTO evaluations VALUES(9004,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(9004,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(9004,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(9004,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(9004,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9004,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9004,36,'cp 0','e6e5 d4e5 c6e5 g1f3 d8e7 f3e5 e7e5 d1a4 c7c6 a4f4 e5e7 f1d3 h7h6 e1g1 e8g8 e3e4 d5e4 d2e4 f6e4 d3e4 e7g5 a2a4 a7a5 f4c7 g5d8 c7f4');
INSERT INTO evaluations VALUES(14350,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(14350,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14350,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(14350,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(14350,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(14350,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(14350,36,'cp 26','d5d4 e5f6 d4c3 f6g7 c3d2 d1d2 f8g7 d2d8 c6d8 c2c3 c8e6 c1e3 h7h5 f1e2 d8c6 e1g1 e8c8 h2h4 e6g4 f1e1 f7f6 g2g3 e7e5 f3d2 g4e2 e1e2 b7b6 a2a4 a7a5 d2e4 c6e7 a1e1 e7f5 e3d2 h8g8 g1g2 f5d6');
INSERT INTO evaluations VALUES(22232,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(22232,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(22232,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(22232,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(22232,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(22232,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(22232,36,'cp 119','g1f3 g8h6 c2c3 c6e7 f1d3 e7f5 d1e2 f6e5 f3e5 d8h4 e2f2 h4f2 e1f2 f5d6 b1d2 f8e7 g2g4 b7b6 h2h4 a7a5 f2g3 e8g8 h1e1 c8a6 d3a6 a8a6 e5c6 f8e8 a2a4 e7f8 h4h5 h6f7 g3g2 a6a8 b2b3 f7d8 c6d8 a8d8 c1a3 d8c8 c3c4 d5c4 b3c4 c7c5');
INSERT INTO evaluations VALUES(22231,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(22231,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(22231,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(22231,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(22231,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(22231,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(22231,36,'cp 135','g8h6 c2c3 c6e7 f1d3 e7f5 d1e2 f6e5 f3e5 d8h4 e2f2 h4f2 e1f2 f8e7 g2g4 f5h4 b1d2 a7a5 f2g3 h6f7 d2f3 h4f3 e5f3 f7d6 h1e1 a5a4 f3g5 e7g5 f4g5 e8g8 c1f4 c8d7 a1c1 f8f7 g5g6 h7g6 d3g6 f7f6 g6b1 a8f8');
INSERT INTO evaluations VALUES(28318,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(28318,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(28318,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(28318,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(28318,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(28318,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(28318,36,'cp 125','d4d5 e8g8 f1e1 h7h6 c1f4 e6d5 c3d5 e7d5 c4d5 b7d5 d1d5 b8a6 d5d2 g8h7 e4e5 a6c5 a1d1 d8e7 h2h4 a7a5 a2a3 a5a4 h4h5 g6h5 f4e3 f8d8 d2d5 h7g8 f3d4 e7e5 d5f3 e5e4 f3h5 e4g6 h5f3 g7d4 e3d4 d8e8 e1e8 a8e8 d4c5 b6c5');
INSERT INTO evaluations VALUES(32491,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(32491,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(32491,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(32491,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(32491,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(32491,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(32491,36,'cp 121','g8f6 a2a3 f8e7 f1d3 c7c5 d4c5 d6c5 e1g1 d7c6 c1f4 b8d7 f3d2 d7f8 f4g3 h7h5 h2h3 f8g6 f2f4 h5h4 g3h2 f6h5 d1e1 d8c7 f4f5 g6f4 d3b5 e6e5 b5c6 b7c6 g1h1 e7f6 d2c4 e8g8 a3a4 a8d8 a4a5 c7d7');
INSERT INTO evaluations VALUES(34041,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(34041,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(34041,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(34041,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(34041,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(34041,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(34041,36,'cp 50','d7d6 e1g1 g7g6 d3d4 f8g7 c2c3 c8e6 c4d3 b8d7 f1e1 e8g8 h2h3 c7c6 b1d2 f8e8 a2a4 e7d8 d4e5 d7e5 f3e5 d6e5 d2f3 d8d6 d1c2 h7h6 c1e3 e8d8 e1d1 d6c7 b2b4 a7a5 d1b1 f6d7 f3d2');
INSERT INTO evaluations VALUES(35086,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(35086,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(35086,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(35086,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(35086,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(35086,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(35086,36,'cp 69','f8c5 c2c3 c5f2 e1f2 f6b6 d2d4 b6b5 a2a4 b5a6 f3e5 g8f6 h1e1 e8g8 b2b4 a6e6 b1d2 d7d5 e4d5 e6d5 d2f3 b8c6 c3c4 d5d6 e5c6 b7c6 c1a3 f8e8 e1e8 f6e8 f3e5 d6f6 d1f3 c8e6 f3f6 e8f6 f2f3 f6d7 e5d7 e6d7 b4b5 a8e8 a3b2');
INSERT INTO evaluations VALUES(33193,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(33193,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(33193,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(33193,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(33193,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(33193,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(33193,36,'cp 93','c2c3 c5f2 e1f2 f6b6 d2d4 b6b5 a2a4 b5b6 f3e5 d7d6 a4a5 b6a6 e5d3 g8f6 h1e1 c8g4 d1c2 e8g8 f2g1 b7b5 c1g5 b8d7 b1d2 a8e8 d3f4 h7h6 g5h4 g7g5 h2h3 g5h4 h3g4 g8h8 d2f3 f8g8 f4d5 a6b7 c2d2 f6g4 f3h4 b5b4 c3b4');
INSERT INTO evaluations VALUES(38850,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(38850,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38850,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(38850,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(38850,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(38850,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(38850,36,'cp 53','b4d2 b1d2 g8e7 g1f3 a7a5 f1b5 c7c6 b5d3 c6c5 d4c5 b8d7 e1g1 d7c5 d2b3 c5d3 d1d3 a5a4 b3d4 e7c6 f1e1 e8g8 c2c3 h7h6 a1b1 c8d7 d3e3 a8c8 a2a3 c6a5 e3f4 a5c4 f4g3 d8c7 h2h4 c8a8 h4h5 c7b6 g3f4 f7f6 e5f6 f8f6 f4g3 a8f8');
INSERT INTO evaluations VALUES(38853,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(38853,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(38853,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(38853,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(38853,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(38853,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(38853,36,'cp 40','b1d2 g8e7 b2b4 b7b6 g1f3 c7c5 c2c3 d8c7 a2a4 c8d7 h2h4 e8g8 f1e2 f8c8 e1g1 h7h6 d1b3 c5d4 c3d4 c7c3 f1c1 c3b3 c1c8 e7c8 d2b3 a7a6 f3e1 a8a7 a4a5 a7b7 a5b6 b7b6 e1c2 c8a7 b3c5 d7b5 e2b5 a7b5 g1f1 b5a7 h4h5 a7c6 c5d3 b6b5 g2g3 a6a5 b4a5 c6a5 d3c5');
INSERT INTO evaluations VALUES(44326,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(44326,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(44326,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(44326,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(44326,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(44326,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(44326,36,'cp 47','b1c3 b8c6 d2d3 f8e7 a2a3 a8b8 c1g5 a7a6 g1f3 e8g8 e2e3 d7d6 e1g1 g8h8 g5f6 e7f6 f3d2 c8e6 a1b1 d8d7 b2b4 c5b4 a3b4 b7b5 c3d5 f6d8 d2e4 c6e7 e4c3 e7d5 g2d5 e6g4 d1d2 f7f5 b1a1 b5c4 d3c4 b8b4 a1a6');
INSERT INTO evaluations VALUES(45785,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(45785,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(45785,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(45785,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(45785,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(45785,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(45785,36,'cp -23','g1f3 f8e7 e1g1 e8g8 e2e3 c8f5 d2d3 h7h6 f3d4 f5h7 e3e4 f8e8 e4d5 c6d5 c1e3 b8c6 d4c6 b7c6 e3d4 f6d7 f1e1 h7f5 c3a4 d8a5 d4c3 e7b4 d3d4 b4c3 a4c3 e8e1 d1e1 a8b8 b2b3 a5a3 a1c1 a3b2 c1d1 f5g6 d1c1 d7f8 h2h3');
INSERT INTO evaluations VALUES(45784,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(45784,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(45784,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(45784,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45784,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(45784,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(45784,36,'cp -27','f8e7 e1g1 e8g8 e2e3 c8f5 d2d3 f8e8 d1c2 h7h6 e3e4 f5h7 c1e3 b8a6 c2b3 f6g4 b3b7 a6b4 e4d5 h7d3 d5c6 d3f1 a1f1 g4e3 f2e3 a8b8 b7d7 e7c5 f3e5 d8d7 e5d7 c5e3 g1h1 b8c8 c3b5 e8e7 b5d6 c8c7 d6f5 e7e6 f1e1 e3g5');
INSERT INTO evaluations VALUES(46189,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46189,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46189,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46189,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46189,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(46189,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46189,36,'cp 0','c4d5 e6d5 b1c3 b8c6 g2g3 c5c4 c1g5 f8e7 f1g2 e8g8 e2e3 c8e6 g5f6 e7f6 f3g1 d8d7 g1e2 f6g5 e1g1 f8d8 b2b3 c6a5 b3c4 a5c4 h2h4 g5h6 d1a4 d7a4 c3a4 b7b6 a4c3 a8c8 a3a4 g7g6 f1b1 a7a6 b1e1 c4b2 e1c1 b2c4');
INSERT INTO evaluations VALUES(55154,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(55154,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(55154,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(55154,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(55154,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(55154,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(55154,36,'cp -15','e7e5 d2d4 e5d4 e3d4 g8f6 h2h3 f8d6 f1d3 e8g8 g1e2 f8e8 e1g1 b8d7 c1f4 d7f8 f4d6 d8d6 d1d2 f8g6 d2b4 d6d8 b1d2 c8f5 d3f5 e8e2 f5d3 a7a5 b4c3 g6f4 f1e1 e2e1 a1e1 c7c6 d3f1 h7h6 c3b3 d8c7 c2c4 d5c4');
INSERT INTO evaluations VALUES(56462,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(56462,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(56462,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(56462,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(56462,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(56462,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(56462,36,'cp -87','h7h5 g1f3 b8c6 f1g2 c8e6 b1c3 f7f6 a2a3 d8d7 h2h4 g8h6 b2b4 h6g4 b4b5 c6d8 a3a4 f8b4 c1d2 e8g8 e1g1 a8c8 a1b1 b4c3 d2c3 c7c5 b5c6 d8c6 f3h2 b7b6 d1e2 g4h2 g1h2 c6e7 c3b4 f8e8 a4a5 e6g4');
INSERT INTO evaluations VALUES(57171,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(57171,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(57171,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(57171,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(57171,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(57171,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(57171,36,'cp -37','e5f4 c1f4 d7d5 e2e4 b8c6 e4d5 d8d5 g1f3 c8g4 f1e2 e8c8 b1c3 d5d7 d1d2 g8f6 e1c1 f8b4 a2a3 b4a5 f4g5 h8e8 h2h3 a5c3 d2c3 e8e2 h3g4 d7g4 g5f6 g7f6 c3f6 g4g2 h1h2 e2c2 c1b1 c2b2 b1a1 g2h2 f3h2 b2h2 f6f4 h2g2 f4f7 c8b8 f7h7 a7a6 h7f7 g2h2');
INSERT INTO evaluations VALUES(57483,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(57483,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(57483,32,'cp 215',NULL);
INSERT INTO evaluations VALUES(57483,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(57483,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(57483,35,'cp 246',NULL);
INSERT INTO evaluations VALUES(57483,36,'cp 236','d8e7 g1f3 d7d6 f1d3 e5d4 d3e2 b8c6 e1g1 c8f5 f1e1 e8c8 f3d4 c6d4 d1d4 c7c5 d4a4 c8b8 b1c3 f5d7 a4b3 e7e6 c3d5 d7c6 d5b4 e6b3 b4c6 b7c6 a2b3 g8f6 c1f4 b8b7 a1a4 d6d5 e2a6 b7b6 f4e3 b6c7 a6f1 d8a8 e3f4 c7b7 g2g3 a8e8 e1a1 e8a8 c2c3 f8e7 b3b4 c5b4 f4e3 b4c3 a4a7 a8a7 a1a7 b7b8 a7e7 c3b2');
INSERT INTO evaluations VALUES(4899,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4899,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(4899,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(4899,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(4899,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(4899,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(4899,36,'cp 28','e1g1 g8f6 f1e1 b7b5 a4b3 d7d6 c2c3 e8g8 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 a2a3 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 a1c1 a8c8 c2d3 f6h5 d3f1 h5f4 g2g3 f4e6 b3b4 d8c7 f1g2 a6a5 c3c4 e6d4 f3d4 e5d4');
INSERT INTO evaluations VALUES(4898,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(4898,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(4898,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(4898,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(4898,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(4898,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(4898,36,'cp 25','g8f6 f1e1 b7b5 a4b3 e8g8 c2c3 d7d6 h2h3 f8e8 f3g5 e8f8 d2d4 c8b7 g5f3 f8e8 b1d2 e7f8 b3c2 g7g6 a2a3 c6b8 a3a4 b8d7 c2d3 c7c6 b2b3 d8c7 d1c2 f6h5 d2f1 f8g7 c1d2 a8c8 d2e3 d6d5 e4d5 e5e4 d3e4 c6d5 e4d3 c7c3');
INSERT INTO evaluations VALUES(9586,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(9586,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(9586,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(9586,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(9586,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9586,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9586,36,'cp 18','b2b3 e8g8 f1d3 d5c4 b3c4 b7b6 e1g1 c8b7 c1b2 c7c5 d1e2 c5d4 e3d4 f6h5 e2e3 d8c7 c3b5 c7f4 b2c1 f4e3 c1e3 h5f6 f1c1 f8d8 d3e2 f6e4 a1b1 e7f8 h2h3 b7c6 f3d2 e4d2 e3d2 d7f6 d2e3 d8d7 b1b3 a8d8 a2a4 f6e4');
INSERT INTO evaluations VALUES(11319,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(11319,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(11319,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(11319,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(11319,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(11319,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(11319,36,'cp 39','c2c3 g7g6 b1d2 e7e5 d4e5 f6g4 g1f3 f8g7 e1g1 g4e5 d3e2 e8g8 d1b3 c6a5 b3c2 e5f3 e2f3 a5c6 a1d1 c6e5 e3e4 c7c6 e4d5 c6d5 f4e5 g7e5 f1e1 d8c7 h2h3 c8e6 d2b3 a8e8 b3d4 e5h2 g1f1 c7c4 c2d3 c4d3 d1d3 h2d6 d4e6 f7e6 g2g3 g8g7');
INSERT INTO evaluations VALUES(12045,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(12045,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(12045,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12045,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(12045,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(12045,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(12045,36,'cp 15','b2b3 g7g6 c1b2 f8g7 e1g1 e8g8 b1c3 f6d7 c3a4 d8a5 b2g7 g8g7 f2f4 d7b6 a4b2 a5c3 d1c1 b6d7 g1h1 d7c5 h2h3 a7a5 f1f3 c3d4 c2c3 c5d3 c3d4 d3c1 a1c1 c8d7 f3c3 h7h5 e4e5 f8b8 e5d6 e7d6 b2c4 a5a4 c4d6 d7e6 c3c6 a4b3 a2b3 b8b3 c6c7 a8a2');
INSERT INTO evaluations VALUES(12456,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(12456,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(12456,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(12456,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(12456,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(12456,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(12456,36,'cp -55','f6e4 c3e4 d7d5 c4d3 d5e4 d3e4 d8d1 e1d1 c8d7 g1f3 f7f5 e4d3 h7h6 h1e1 e8f7 c1f4 f8c5 d1e2 g7g5 f3e5 c6e5 f4e5 h8d8 e2f1 d7c6 a1d1 c5e7 d3c4 d8d1 e1d1 h6h5 b2b3 a8c8 h2h3 g5g4 c4e2 e7f6 e5f4 f7e7 h3g4 h5g4 f4e3 a7a6 f2f3 g4f3 g2f3');
INSERT INTO evaluations VALUES(18957,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18957,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18957,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(18957,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(18957,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18957,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18957,36,'cp 0','f8c5 f1d3 g8f6 c2c4 d7d6 h2h3 c8e6 b1c3 a8c8 e1g1 e8g8 f1e1 a7a6 b2b3 c6d4 c3a4 f6d7 c1e3 c5a7 a4c3 a7c5');
INSERT INTO evaluations VALUES(21005,30,'cp -166',NULL);
INSERT INTO evaluations VALUES(21005,31,'cp -157',NULL);
INSERT INTO evaluations VALUES(21005,32,'cp -174',NULL);
INSERT INTO evaluations VALUES(21005,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(21005,34,'cp -169',NULL);
INSERT INTO evaluations VALUES(21005,35,'cp -159',NULL);
INSERT INTO evaluations VALUES(21005,36,'cp -161','e7e5 e2e3 f8d6 f3f4 e5f4 e3f4 d8e7 d1e2 g8f6 g1f3 e8g8 e2e7 c6e7 f1d3 h7h5 e1g1 c8f5 d3e2 f5c2 a2a4 a7a6 a4a5 c7c5 c1e3 e7f5 e3f2 d6f4 g2g3 f8e8 f1e1 f4e3 b1a3 c2e4 d4c5 a8c8 f2e3 f5e3 f3d2 e4f5 d2b3');
INSERT INTO evaluations VALUES(21684,30,'cp -158',NULL);
INSERT INTO evaluations VALUES(21684,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(21684,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(21684,33,'cp -138',NULL);
INSERT INTO evaluations VALUES(21684,34,'cp -160',NULL);
INSERT INTO evaluations VALUES(21684,35,'cp -146',NULL);
INSERT INTO evaluations VALUES(21684,36,'cp -148','g2g3 d8e7 f1g2 g8f6 a2a3 c8e6 b2b4 h7h5 h2h4 a7a6 g1h3 e7d7 d4e5 c6e5 h3g5 f8d6 c1e3 e8g8 e3f2 f8e8 e1g1 e6f5 b1d2 a6a5 d1b3 c7c6 c3c4 b7b5 c4d5 c6d5 a1d1 a5a4 b3a2 a8c8 e2e4 f5g6 f3f4 e5d3 e4e5 f6g4 a2d5 d3f2 d5d6 d7d6');
INSERT INTO evaluations VALUES(23466,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(23466,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(23466,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(23466,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(23466,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(23466,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(23466,36,'cp -13','e7e5 e2e4 d7d5 d1a4 d8d6 d2d3 g8f6 f1e2 f8e7 b2b4 e8g8 a4c2 a7a5 b4b5 c6b8 b1d2 c7c6 d3d4 b8d7 e1g1 f6e4 d2e4 d6g6 b5c6 b7c6 f3d2 d5e4 c2e4 g6e4 d2e4 e5d4 c3d4 d7b6 e4c5 c8f5 e2f3 b6d5 h2h3 e7f6 c1a3 f8e8');
INSERT INTO evaluations VALUES(24314,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(24314,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(24314,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(24314,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(24314,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(24314,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24314,36,'cp 32','b5c6 d7c6 f1e1 f8e7 d3d4 e5d4 f3d4 c6d7 h2h3 e8g8 d1f3 b7b5 a2a4 b5b4 c3e2 c7c5 d4f5 d7f5 f3f5 f8e8 f5f3 f6d7 e2f4 a6a5 e1d1 e7f6 f4d5 f6e5 c2c3 h7h6 c1f4 d8h4 f4d2 a8d8 d5c7 e8e7 c7b5 d8c8 f3d3 h4f6');
INSERT INTO evaluations VALUES(27513,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27513,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27513,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27513,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(27513,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27513,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27513,36,'cp 0','b5c6 b7c6 e1g1 a7a5 d4c5 f8c5 c3c4 b6c7 b1c3 g8e7 c3a4 c5a7 c4c5 f7f6 f1e1 e8g8 e5f6 g7f6 c1h6 f8f7 f3h4 e6e5 d1h5 c7c8 e1e5 d7g4 e5g5 f6g5 h5g5 g8h8 g5e5 h8g8');
INSERT INTO evaluations VALUES(27510,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27510,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27510,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27510,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(27510,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27510,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(27510,36,'cp 13','b7c6 e1g1 a7a5 c1e3 c5d4 c3d4 c6c5 b2b3 a8c8 b1d2 b6b7 a1c1 g8e7 c1c5 e7c6 c5c2 f8e7 d2b1 e8g8 d1d2 a5a4 f1c1 a4b3 a2b3 e7b4 b1c3 c6a5 c1b1 h7h6 d2d3 b4e7 c3e2 a5c6 e3d2 c6b4 d2b4 b7b4 c2c8 f8c8');
INSERT INTO evaluations VALUES(29582,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29582,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29582,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29582,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29582,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29582,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29582,36,'cp 0','b8d7 c2c4 e7e6 b1c3 f8b4 c1d2 e8g8 f1e2 a7a5 a2a3 b4e7 e1g1 c7c6 f3h4 f5e4 c3e4 f6e4 h4f3 f7f5 d1b3 a8b8 b3a4 b8a8');
INSERT INTO evaluations VALUES(29928,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(29928,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(29928,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(29928,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(29928,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(29928,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(29928,36,'cp 60','f4e5 f6e4 c4d5 e4g5 d5b7 b8d7 d2d4 g4f3 b7f3 g5f3 d1f3 d6e5 e1g1 f7f6 b1c3 f8d6 c3e4 e8g8 e4d6 c7d6 f3d5 g8h8 d5d6 d8c8 c1e3 f8d8 c2c3 d7b6 d6c5 c8c5 d4c5 b6a4 a1d1 h8g8 d1d8 a8d8 f1b1 g8f7 c3c4 a7a5 b2b3 a4c3 b1a1 c3d1 e3c1');
INSERT INTO evaluations VALUES(31453,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(31453,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(31453,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(31453,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(31453,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(31453,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31453,36,'cp 13','b1d2 g8f6 g1f3 h7h6 d1b3 d8c8 a2a4 a7a5 f1e2 b8d7 h2h3 f8e7 e1g1 b7b6 c3c4 e8g8 f1d1 c8b7 a1c1 f8c8 f3e5 d7e5 f4e5 f6d7 e5g3 e7b4 d2b1 d7f6 b1c3 b7e7 g3e5 f6d7 e5h2 d7f6 c3a2 b4d6 h2d6 e7d6');
INSERT INTO evaluations VALUES(31452,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(31452,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(31452,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(31452,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(31452,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31452,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(31452,36,'cp 13','g8f6 g1f3 h7h6 d1b3 d8c8 f1e2 b8d7 f3e5 d7e5 f4e5 f6d7 e5g3 a7a5 a2a4 f8e7 e1g1 d7f6 g3e5 e8g8 f1d1 f6d7 e5g3 d7f6 h2h3 c8d7 a1c1 f6e4 d2e4 f5e4 c3c4 e7d6 g3d6 d7d6 c4d5 e4d5 e2c4 a8b8 b3c3 d5c4 c3c4 b7b5 c4c6 d6c6');
INSERT INTO evaluations VALUES(29256,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(29256,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(29256,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(29256,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29256,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(29256,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(29256,36,'cp 5','g1f3 h7h6 d1b3 d8c8 a2a4 b8d7 h2h3 a7a5 f1e2 f8e7 e1g1 b7b6 c3c4 c8b7 a1c1 e8g8 f1d1 f8d8 f4h2 f5h7 f3e5 d7e5 h2e5 a8c8 d2f3 f6d7 c4d5 c6d5 e5f4 g7g5 f4g3 c8c1 d1c1 d8c8 c1c8 b7c8 f3e5 d7e5 g3e5');
INSERT INTO evaluations VALUES(29257,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(29257,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(29257,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(29257,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(29257,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(29257,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(29257,36,'cp 2','h7h6 d1b3 d8c8 f1e2 b8d7 f3e5 d7e5 f4e5 f8e7 e1g1 e8g8 a2a4 a7a5 a1c1 f6d7 e5g3 b7b6 f1d1 c8b7 c3c4 e7b4 h2h3 d7f6 g3e5 f6d7 e5f4 d7f6 d2b1 f8c8 b1c3 b7e7 f4e5 f6d7 e5g3 d7f6 e2f3 b4d6 g3d6 e7d6');
INSERT INTO evaluations VALUES(32194,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(32194,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(32194,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(32194,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(32194,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(32194,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(32194,36,'cp 94','f3d4 g4e2 d1e2 g7g6 e2b5 b8d7 b5b7 f8g7 d4c6 d8c8 b7c8 a8c8 b1c3 g7c3 b2c3 a7a6 c6d4 g8e7 f2f3 h7h5 e1g1 f7f6 a1b1 d7e5 h2h4 e8d7 f1d1 c8b8 d4b3 d7e6 c1f4 e7c6 b3d4 e6d7 b1b3 e5c4 d4c6 d7c6 d1d4 c4b6');
INSERT INTO evaluations VALUES(32179,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(32179,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(32179,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(32179,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(32179,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(32179,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(32179,36,'cp 89','g4e2 d1e2 g7g6 e2b5 b8d7 b5b7 f8g7 d4c6 d8c8 b7c8 a8c8 b1c3 g7c3 b2c3 c8a8 c6b4 g8e7 f2f3 f7f6 b4d3 e8f7 e1f2 h8b8 g2g4 a7a5 c1f4 a8a6 a1b1 b8b1 h1b1 a6c6 f4d2 d7b6 b1b5 b6c4 f2e2 g6g5 b5b7 e7g6 b7a7 g6e5 d3e5 d6e5');
INSERT INTO evaluations VALUES(32181,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(32181,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(32181,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(32181,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(32181,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(32181,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(32181,36,'cp 103','d1e2 g7g6 e2b5 b8d7 b5b7 f8g7 d4c6 d8c8 b7c8 a8c8 b1c3 g7c3 b2c3 a7a6 f2f3 g8e7 c6b4 d7b6 e1f2 e8d7 c1g5 h8e8 b4d3 c8b8 h1d1 a6a5 d3e5 d7e6 e5g4 b6d7 g5e7 e6e7 a1b1 b8b1 d1b1 d7b6');
INSERT INTO evaluations VALUES(32180,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(32180,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(32180,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(32180,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(32180,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(32180,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(32180,36,'cp 99','g7g6 e2b5 b8d7 b5b7 f8g7 d4c6 d8c8 b7c8 a8c8 c6a7 c8b8 a2a4 g7b2 c1b2 b8b2 e1d2 b2b7 a7c6 g8e7 c6a5 b7a7 a5b3 d7e5 b1c3 e8d7 h1b1 f7f5 f2f4 e5c6 e4e5 d6e5 f4e5 c6e5 c3b5 a7a8 b1e1 e7c6 b3c5 d7c8 c5e6 c8d7 e6c7 a8a5 d2c3 d7e7 a1d1 a5a4');
INSERT INTO evaluations VALUES(34441,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(34441,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(34441,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(34441,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(34441,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34441,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(34441,36,'cp 33','d2d4 b8c6 a2a3 g8h6 f1d3 c5d4 c3d4 h6f5 d3f5 e6f5 b1c3 c8e6 b2b4 f8e7 e1g1 h7h6 h2h4 b6d8 h4h5 d8d7 d1d3 e8g8 c1f4 a7a6 f1c1 c6a7 f3d2 f8c8 d2b3 c8c4 b3c5 d7c6 c3e2 a8c8 d3g3 g8h7 c1c4');
INSERT INTO evaluations VALUES(35006,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(35006,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(35006,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(35006,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(35006,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(35006,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(35006,36,'cp 189','h5e5 f6e5 f3e5 d7d6 e5f3 f7f5 b1c3 f5e4 c3e4 b8c6 c2c3 c8g4 f1e2 d6d5 e4g3 d5d4 d2d3 f8g7 c3c4 c6b4 e1d1 g8f6 a2a3 g4f3 g2f3 b4c6 f3f4 g7h6 h2h4 e8c8 b2b4 d8f8 h4h5 c6d8 a3a4 f6d7 g3e4 d8e6 h5g6 h7g6');
INSERT INTO evaluations VALUES(35002,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(35002,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(35002,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(35002,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(35002,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(35002,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(35002,36,'cp 185','f6e5 f3e5 d7d6 e5f3 f7f5 b1c3 f5e4 c3e4 b8c6 c2c3 c8g4 f1e2 d6d5 e4g3 d5d4 e1g1 g8f6 d2d3 d4c3 b2c3 e8c8 d3d4 f8g7 h2h3 g4e6 f3g5 e6g8 f1e1 h7h6 g5e4 f6e4 g3e4 b7b6 c1e3 c8b8 e2d3 g8f7 a2a4 c6a5 e4d2 h8e8 d3b5');
INSERT INTO evaluations VALUES(35011,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(35011,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(35011,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(35011,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(35011,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(35011,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(35011,36,'cp 196','f3e5 d7d6 e5f3 f7f5 b1c3 f5e4 c3e4 b8c6 c2c3 c8g4 f1e2 d6d5 e4g3 d5d4 e1g1 g8f6 d2d3 d4c3 b2c3 e8c8 d3d4 f8g7 f1e1 h8f8 h2h3 g4d7 c1d2 d7e6 d2g5 d8e8 g5f6 g7f6 g3e4 e6f5 e4f6 f8f6 e2c4 e8e1 a1e1 c6a5 c4b5 f6e6 f3e5 e6b6 b5d3 f5d3');
INSERT INTO evaluations VALUES(35008,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(35008,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(35008,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(35008,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(35008,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(35008,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(35008,36,'cp 201','d7d6 e5f3 f7f5 b1c3 f5e4 c3e4 b8c6 c2c3 c8g4 f1e2 d6d5 e4g3 d5d4 e1g1 g8f6 d2d3 d4c3 b2c3 e8c8 d3d4 f8g7 c1g5 h7h6 g5d2 h8e8 f1e1 g4e6 e2d3 e6f7 a2a4 a7a5 h2h3 e8e1 a1e1 d8e8 d2e3 f6d5 g3e4 d5e3 f2e3 c6e7 f3d2');
INSERT INTO evaluations VALUES(34746,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(34746,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(34746,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(34746,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(34746,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(34746,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(34746,36,'cp 195','e5f3 f7f5 e4e5 d6e5 f3e5 c8e6 d2d4 b8d7 e5d7 e6d7 f1c4 e8c8 e1g1 d8e8 b1d2 g8f6 c2c3 d7c6 c4d3 f8g7 a2a4 f6h5 d2c4 c6e4 d3e4 e8e4 f2f3 e4e6 c1d2 h8e8 f1e1 e6e1 d2e1 h5f4 e1g3 f4e2 g1f2 e2g3 h2g3 g7f6 a1e1 e8e1 f2e1 c8d7 e1d2 f6e7 c4e5 d7d8');
INSERT INTO evaluations VALUES(38003,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(38003,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(38003,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(38003,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(38003,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(38003,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(38003,36,'cp -12','g1f3 g8f6 f1e2 b8c6 e1g1 c8f5 f3h4 f5e6 b1a3 a7a6 c2c3 e8g8 h4f3 a8c8 f1e1 f8e8 h2h3 d8c7 c1g5 c6a5 g5f6 g7f6 a3c2 e6c4 c2e3 c4e2 e1e2 e7e6 f3d2 d6d5 d2f3 h7h5 d1d2 g8g7 a1e1 a5c6 a2a4 c7d8');
INSERT INTO evaluations VALUES(38002,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(38002,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(38002,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(38002,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(38002,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(38002,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(38002,36,'cp -21','g8f6 f1e2 b8c6 e1g1 c8f5 f3h4 f5e6 c2c3 e8g8 b1a3 f8e8 f1e1 a7a6 h4f3 e6d5 a3c2 e7e5 c1g5 h7h6 d4e5 d6e5 g5f6 g7f6 c2e3 d5e6 d1c2 d8e7 e2c4 a8d8 a1d1 d8d1 e1d1 e6c4 e3c4 e5e4 f3d2 f6g5 d2b3 e8d8 d1d8 c6d8');
INSERT INTO evaluations VALUES(38819,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(38819,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(38819,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(38819,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(38819,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(38819,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(38819,36,'cp 90','b2c3 c7c5 c1a3 c5d4 d1d4 g8f6 c4d5 d8a5 a3d6 a5d5 d4d5 f6d5 c3c4 d5f6 g1f3 f6e4 d6a3 b8a6 f3d4 e4c5 d4b5 e8g8 f1e2 c8d7 e1g1 f8d8 f1d1 f7f6 h2h4 d7c6 d1d8 a8d8 b5a7 c6e4 a1d1 d8d1 e2d1 e4d3 d1b3 c5b3 a2b3 a6c7 a3d6 c7e8 a7c8 e6e5 h4h5 g8f7 c4c5 d3a6 g1h1');
INSERT INTO evaluations VALUES(38817,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(38817,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(38817,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(38817,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(38817,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(38817,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(38817,36,'cp 107','g8e7 g1f3 e8g8 f1d3 h7h6 c4d5 e6d5 c3c4 c7c5 e1g1 c5d4 f3d4 b8c6 d4c6 b7c6 d1h5 c8e6 h5e5 d8b8 e5c3 b8c7 c1b2 f7f6 c4d5 e7d5 c3d4 g8h8 a1c1 a8b8 d4e4 e6g8 b2d4 b8d8 h2h3 f8e8 e4f3 d5e7 f3g4 c7a5 d3b1 d8d5');
INSERT INTO evaluations VALUES(44282,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(44282,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(44282,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(44282,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(44282,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(44282,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44282,36,'cp 20','e3d4 d7d5 g1f3 f8b4 c4d5 e6d5 f1b5 b8c6 e1g1 e8g8 c1g5 b4c3 b2c3 h7h6 g5h4 d8d6 h4g3 d6d8 b5c6 b7c6 f3e5 d8b6 d1c2 c8e6 f2f3 f8c8 a2a4 b6a5 c2d2 f6d7 f1e1 d7e5 e1e5 c8e8 e5e3 e6f5 e3e8 a8e8');
INSERT INTO evaluations VALUES(46619,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(46619,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(46619,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(46619,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(46619,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(46619,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(46619,36,'cp 112','c8g4 c1e3 f8g7 d1d2 g4f3 g2f3 e7e6 e1c1 d8e7 f1b5 b8c6 h2h4 h7h5 f3f4 a7a6 b5c6 b7c6 d2d3 a8d8 f4f5 g7d4 e3d4 e6e5 f5g6 e5d4 c3e2 h8g8 e2f4 f7g6 d3b3 e8d7 b3h3 d7e8 h1e1 e7f7 h3e6 f7e6 f4e6 d8d7 c1b1 d4d3 e6c5 d3d2 e1e2 d7d6 d1d2 d6d2 e2d2 b6c4 d2c2');
INSERT INTO evaluations VALUES(47229,30,'cp -184',NULL);
INSERT INTO evaluations VALUES(47229,31,'cp -168',NULL);
INSERT INTO evaluations VALUES(47229,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(47229,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(47229,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(47229,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(47229,36,'cp -178','d5e4 f3e5 b8d7 e5c4 d7b6 c4e3 g7g6 a2a4 a7a5 b1c3 e4d3 f1d3 f8g7 h2h3 c7c6 e1g1 e8g8 f1e1 c8e6 e3g4 f6g4 h3g4 d8c8 g4g5 e6g4 d1d2 f8e8 d3f1 g4f5 d2e2 c8c7 c1e3 f5e6 a1d1 e8d8 e2f3 d8d1 f3d1 a8d8 d1e2 g7c3 b2c3');
INSERT INTO evaluations VALUES(47407,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(47407,31,'cp -174',NULL);
INSERT INTO evaluations VALUES(47407,32,'cp -159',NULL);
INSERT INTO evaluations VALUES(47407,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(47407,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(47407,35,'cp -185',NULL);
INSERT INTO evaluations VALUES(47407,36,'cp -176','f3e5 b8d7 e5c4 d7b6 c4b6 a7b6 b1c3 c8f5 c1f4 e4d3 d1f3 e7e5 f4g5 f5g4 f3b7 h7h6 g5e3 d3c2 f1b5 g4d7 b5d7 f6d7 e1g1 f8c5 a1c1 e8g8 c1c2 c5d4 b7c6 d7c5 a2a3 f8e8 c3b5 a8c8 f1d1 e8e6 c6f3 e6f6 f3g4 c5e6 e3d4 e6d4 b5d4 e5d4');
INSERT INTO evaluations VALUES(49812,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(49812,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(49812,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(49812,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(49812,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(49812,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(49812,36,'cp 152','d1d4 c7c6 f1c4 f8e7 c4d5 c6d5 b1c3 b8c6 d4d5 d7d6 e5d6 d8d6 c1g5 f7f6 e1c1 d6d5 c3d5 e8f7 g5e3 c8f5 d5e7 f7e7 d1d5 f5e6 d5d2 h8d8 h1d1 d8d2 f3d2 a8d8 b2b3 e7f7 d2e4 e6d5 f2f3 d5e4 d1d8 c6d8 f3e4 d8c6 c1b1 f7e7');
INSERT INTO evaluations VALUES(49810,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(49810,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(49810,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(49810,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(49810,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(49810,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(49810,36,'cp 174','c7c6 b1c3 d5c3 d4c3 a7a5 c1d2 f8b4 c3e3 e8g8 a2a3 b4d2 e3d2 b7b5 e1c1 b8a6 d2e3 d8e7 f3d4 f8e8 f2f4 e7f8 f1e2 b5b4 a3a4 a6c5 b2b3 d7d5 d4c6 c8d7 e2b5 e8c8 c6d4 d7b5 d4b5 c5e4 d1d5 e4c3 d5d6 c3b5 a4b5 a5a4 h1d1 a4b3 e3b3 a8a1 c1b2');
INSERT INTO evaluations VALUES(44127,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(44127,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(44127,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(44127,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(44127,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(44127,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(44127,36,'cp 160','f1c4 f8e7 c4d5 c6d5 d4d5 b8c6 b1c3 d7d6 e5d6 c8e6 d6e7 d8d5 c3d5 e6d5 c1e3 d5f3 g2f3 e8e7 e1c1 h8d8 h1e1 d8d1 e1d1 a8d8 d1d8 e7d8 c1d2 a7a6 f3f4 d8d7 c2c4 c6e7 d2d3 d7e6 d3c3 e7f5 a2a4 e6d6 b2b4 g7g6 b4b5 a6b5 a4b5 f5e7 c3b3');
INSERT INTO evaluations VALUES(51008,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(51008,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(51008,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(51008,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(51008,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(51008,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(51008,36,'cp 149','b1c3 f8b4 d1g4 d8f6 e4e5 f6g6 g4g6 f7g6 a2a3 b4c3 b2c3 b7b5 a3a4 b5c4 f1c4 c8b7 f2f3 g8e7 c4d3 a6a5 c1a3 g6g5 g1e2 b7a6 d3b5 e7d5 e1d2 d5b6 e2g3 a6b5 a4b5 c7c6 h1b1 b6c4 d2c1 c4a3 a1a3 c6b5 b1b5 b8c6 g3e4');
INSERT INTO evaluations VALUES(51006,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(51006,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(51006,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(51006,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(51006,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(51006,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(51006,36,'cp 136','d7d6 g1f3 c7c5 d4d5 g7g6 d1a4 c8d7 a4b3 b7b6 a2a4 d8c7 f1d3 f8g7 c1d2 g8e7 e1g1 e8g8 g1h1 d7c8 f3g1 b8d7 f2f4 e6d5 c4d5 f7f5 e4e5 g8h8 g1f3 d6e5 f4e5 d7e5 f3e5 c7e5 d2f4 e5d4 d5d6 e7c6 b3b6 d4d3 b6c6');
INSERT INTO evaluations VALUES(50891,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(50891,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(50891,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(50891,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(50891,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(50891,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(50891,36,'cp 150','g1f3 b8d7 f1e2 c7c5 d4d5 e6d5 d1d5 g8f6 d5d3 f8e7 c1f4 d8c7 e1g1 e8g8 f4g3 d7e5 f3e5 d6e5 f2f4 e7d6 f4f5 d6e7 a2a4 f8d8 c3d5 f6d5 e4d5 e7f6 d3e4 c8d7 e2d3 d8e8 b2b3 a8d8 a1e1 b7b5 g3h4 c7d6 h4f6 d6f6 c4b5');
INSERT INTO evaluations VALUES(51960,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(51960,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(51960,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(51960,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(51960,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51960,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(51960,36,'cp 34','c5d4 f3d4 a7a6 e2e4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 g8e7 a2a4 b5b4 c3a2 a6a5 d3b5 d8c7 c1d2 e7g6 c2c3 b4c3 a2c3 f8e7 d2e3 e8g8 a1c1 c7b7 f2f4 f7f5 b5c6 d7c6 g2g3 b7b3 e3f2 e7b4 e4f5 e6f5');
INSERT INTO evaluations VALUES(52073,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52073,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(52073,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(52073,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52073,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52073,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52073,36,'cp 27','f3d4 a7a6 e2e4 b7b5 f1d3 c8b7 d1e2 b8c6 d4c6 b7c6 e1g1 f8e7 c1f4 b5b4 c3b1 g8f6 f1d1 d8b6 b1d2 b6b7 a2a3 a6a5 a3b4 a5b4 d2c4 a8a1 d1a1 f6e4 d3e4 c6e4 c4d6 e7d6 f4d6 b7c6 e2d2 f7f6 f2f3 c6b6 g1f1 b6b5 f1f2 e4d5 d2b4 b5b4');
INSERT INTO evaluations VALUES(56383,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(56383,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(56383,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(56383,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(56383,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(56383,35,'cp -143',NULL);
INSERT INTO evaluations VALUES(56383,36,'cp -144','d5d4 c3e2 b8c6 d2d3 f7f5 g2g3 g8f6 g1h3 c8e6 c1d2 h7h6 h3f2 d8d7 a2a3 g7g5 e4f5 e6f5 f1g2 f6d5 f2e4 e8c8 d1c1 f5h3 e1g1 h3g2 g1g2 f8e7 b2b4 a7a6 g2h1 h8f8 c1b2 g5g4 f3g4 e7g5 d2g5 h6g5 b4b5 a6b5 b2b5 d5e3 f1f8');
INSERT INTO evaluations VALUES(56575,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(56575,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(56575,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(56575,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(56575,34,'cp -126',NULL);
INSERT INTO evaluations VALUES(56575,35,'cp -134',NULL);
INSERT INTO evaluations VALUES(56575,36,'cp -119','c3e2 b8c6 d2d3 f7f5 g2g3 g8f6 g1h3 c8e6 c1d2 h7h6 h3f2 d8d7 a2a3 g7g5 e4f5 e6f5 f1g2 f6d5 d1c1 f8c5 b2b4 c5b6 e1g1 e8c8 c1b2 d5e3 b4b5 c6e7 d2e3 d4e3 f2e4 f5e4 f3e4 d7g4 a1e1 g4e6 a3a4');
INSERT INTO evaluations VALUES(56690,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(56690,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(56690,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(56690,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(56690,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(56690,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(56690,36,'cp 139','d2d4 b8c6 g1e2 g7g6 e1g1 f8g7 c4b3 e5d4 c2c3 g8h6 c3d4 c8g4 f3d3 g4e2 d3e2 c6d4 b3a4 b7b5 a4b5 d4b5 c1h6 g7h6 e2b5 d8d7 b5c4 e8f8 b1c3 f8g7 a1d1 h8e8 g2g3 g7h8 b2b3 c7c6 h2h4 h6f8 g1g2 d7e6 c4e6 e8e6 d1d3');
INSERT INTO evaluations VALUES(56890,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(56890,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(56890,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(56890,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(56890,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(56890,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(56890,36,'cp 188','d2d4 g5g4 f3d2 d7d5 b1c3 g8f6 e4e5 f6d7 d1g4 c7c5 d4c5 h7h5 g4e2 d7c5 d2b3 c8d7 h2h4 c5a4 c3a4 d7a4 c2c3 a7a6 c1g5 d8b6 g2g3 f8e7 g5e3 b6c7 f2f4 b8d7 f1g2 d7c5 b3d4 c5e4 g2e4 d5e4 e1g1 e7c5 g1h2 c5d4 c3d4 a4b5');
INSERT INTO evaluations VALUES(56888,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(56888,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(56888,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(56888,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(56888,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(56888,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(56888,36,'cp 196','g5g4 f3d2 d7d5 b1c3 g8f6 f1b5 c8d7 b5d7 b8d7 e4e5 f6g8 d1g4 c7c5 d4c5 d8c7 e1g1 h7h5 g4e2 c7e5 e2e5 d7e5 c3b5 e8d7 b2b4 f8g7 c1b2 g8e7 f2f4 e5f3 d2f3 g7b2 a1b1 b2f6 c2c4 d5c4 b5d6 e7d5 d6c4 h8g8 b4b5 d7c7');
INSERT INTO evaluations VALUES(56891,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(56891,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(56891,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(56891,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(56891,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(56891,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(56891,36,'cp 171','f3d2 d7d5 b1c3 g8f6 e4e5 f6d7 d1g4 h7h5 g4e2 c7c5 d4c5 d7c5 d2b3 c5b3 a2b3 c8d7 h2h4 d8b6 c1g5 d7c6 f2f4 b8d7 e1c1 a7a5 c3a4 b6c7 f4f5 f8e7 e2f2 e7g5 h4g5 e8c8 f1e2 e6f5 h1h5 c7e5 c1b1 f5f4 h5h8 d8h8 e2f3 c8b8 a4c3 e5e3');
INSERT INTO evaluations VALUES(57134,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(57134,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(57134,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(57134,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(57134,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(57134,35,'cp -128',NULL);
INSERT INTO evaluations VALUES(57134,36,'cp -125','e5d4 d1d4 b8c6 d4a4 d7d5 f1g2 c8d7 a4b3 f8c5 b1c3 d8h4 g1h3 e8c8 b3d5 h4g4 h3f4 c5d4 h2h3 g4h4 d5f3 g8f6 c1d2 h8e8 f4d3 d7e6 e1c1 d4c3 d2c3 f6d5 f3e4 h4e4 g2e4 d5c3 b2c3 g7g6 e4c6 b7c6 d3b4 e6d5 b4d5 d8d5 d1d5 c6d5 e2e3 e8e4 c1d2');
INSERT INTO evaluations VALUES(12414,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(12414,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(12414,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12414,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(12414,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12414,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(12414,36,'cp 8','e2e3 h7h6 b1d2 f6h5 d4c5 h5f4 e3f4 d8c7 b2b4 b7b6 c5b6 a7b6 f4f5 f8e7 f5e6 d7e6 f1d3 e8g8 e1g1 a8a3 d2b1 a3a8 a2a4 c6e5 f3e5 c7e5 d1e2 e5e2 d3e2 f8d8 a4a5 b6a5 a1a5 c8b7 a5a8 d8a8 e2f3 b7f3 g2f3 a8a1');
INSERT INTO evaluations VALUES(13317,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(13317,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(13317,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(13317,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(13317,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(13317,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(13317,36,'cp 13','d2d4 d6d5 e4d5 d8d5 d4c5 d5c5 c1e3 c5a5 f1c4 e7e6 b1d2 a5c7 e1g1 f6g4 e3g5 h7h6 g5h4 f8d6 h2h3 g4e5 f3e5 c6e5 c4b5 c8d7 b5d7 e5d7 d2e4 d6h2 g1h1 h2f4 g2g3 f4e5 f2f4 e5f6 e4f6 d7f6 c2a4 c7d7 a4d7 e8d7 a1e1 d7e7 f4f5 h8d8 h1g2 d8d3 f5e6 f7e6 g3g4 e7f7 g4g5');
INSERT INTO evaluations VALUES(13296,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(13296,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(13296,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13296,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(13296,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(13296,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13296,36,'cp 24','d6d5 e4d5 d8d5 d4c5 d5c5 c1e3 c5a5 b1d2 a5c7 f1c4 e7e6 a1d1 f8e7 d2e4 f6g4 e3g5 e8g8 e1g1 b7b6 f1e1 c8b7 c4b3 a8d8 d1d8 e7d8 g5d8 c6d8 c2d2 b7e4 e1e4 g4f6 d2f4 c7f4 e4f4');
INSERT INTO evaluations VALUES(14619,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(14619,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(14619,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(14619,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(14619,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(14619,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14619,36,'cp 23','g7g5 d1b3 d8e7 d2d4 f8g7 d4e5 d6e5 a2a4 e8g8 b3c2 f8d8 b1d2 c8e6 h2h3 f6h5 c4e6 e7e6 d2f1 d8d7 f1e3 a8d8 e3f5 c6e7 f5e7 e6e7 c1e3 b7b6 a1d1 h5f6 d1d7 d8d7 a4a5 e7e6 e1d1 g7f8 d1d7 f6d7 f3d2 e6c6 a5b6 c7b6 d2f1 d7f6 f1g3 f8c5 e3c5');
INSERT INTO evaluations VALUES(17711,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(17711,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(17711,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(17711,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(17711,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(17711,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(17711,36,'cp 36','c6b4 g1f3 b4d3 c2d3 e7e6 b1c3 f8d6 f4g3 g8f6 a1c1 e8g8 e1g1 h7h6 e3e4 d6e7 c3a4 c7c6 f1e1 f6h5 g3e5 f7f6 e5g3 c8d7 f3h4 h5g3 h2g3 e7b4 e1e2 d7e8 a2a3 b4d6 e4d5 e6d5 e2e1 e8f7 a4c5 d6c5 d4c5 f8e8 e1e8 f7e8');
INSERT INTO evaluations VALUES(17754,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(17754,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(17754,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(17754,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(17754,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(17754,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(17754,36,'cp 18','b5e2 f7f6 g1f3 g8h6 c1h6 g7h6 e1g1 f8g7 b1d2 e8g8 e5f6 d8f6 c2c3 g8h8 d2b3 e6e5 d4e5 c6e5 d1d5 e5g6 e2d3 c8h3 g2h3 g6f4 d5e4 f4h3 g1g2 h3f4 g2h1 f4d3 e4d3 f6f3 d3f3 f8f3 a1e1 f3f7 b3c5 a8d8 e1d1 d8d1 f1d1 g7f8 c5d3 f8d6 d1e1 f7f8');
INSERT INTO evaluations VALUES(18035,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(18035,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(18035,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(18035,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(18035,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(18035,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(18035,36,'cp 50','d7d6 d2d4 g8f6 h2h3 f8e7 e1g1 e8g8 b1d2 c8d7 f1e1 a6a5 b3c2 f8e8 b2b3 e7f8 c1b2 a5a4 b3a4 b5a4 b2a3 d8c8 d2f1 c8a6 f1e3 a6a5 d4d5 c6e7 a1b1 e7g6 a3b4 a5a7 c3c4 a8b8 g1h2 c7c6 f3d2 c6d5 c4d5 e8c8 a2a3 b8b7');
INSERT INTO evaluations VALUES(17972,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(17972,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(17972,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(17972,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(17972,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(17972,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(17972,36,'cp 48','d2d4 g8f6 e1g1 f8e7 b1d2 e8g8 h2h3 c8d7 f1e1 f8e8 d2f1 a6a5 b3c2 e7f8 f1g3 a5a4 c1d2 d8b8 d2e3 a4a3 b2b3 e5d4 c3d4 c6b4 c2b1 c7c5 e3d2 b8d8 d1c1 d8b6 d2c3 b4c6 d4c5 b6c5 c3f6 c5c1 e1c1 g7f6 b1d3 a8b8 g3h5 f8g7 h5g7 g8g7');
INSERT INTO evaluations VALUES(18879,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(18879,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(18879,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(18879,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(18879,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(18879,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(18879,36,'cp -23','b1c3 f8b4 e1g1 g8e7 c1g5 b4c3 b2c3 h7h6 g5h4 e8g8 d1d6 f7f5 a1d1 d8e8 c4d3 f5e4 d3e4 e7f5 d6d2 d7d5 e4f5 f8f5 h4g3 b7b6 f1e1 c8d7 f3h4 f5f7 h4f3 a8c8 d2d3 f7f5 d3d2 e8f7 h2h4');
INSERT INTO evaluations VALUES(19456,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(19456,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(19456,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(19456,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(19456,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(19456,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(19456,36,'cp -17','e7e6 g1f3 d7d5 e4d5 e6d5 e1g1 g8f6 d3d4 c5d4 f3d4 f8d6 c1e3 e8g8 b1c3 f8e8 e2f3 a7a6 f1e1 h7h6 c3d5 f6d5 f3d5 d6h2 g1h2 d8d5 d1f3 d5f3 d4f3 c8d7 f3d4 c6b4 a1d1 f7f6 f2f3 b4a2 c2c3 a8d8 e3f2 g8f7 e1e8');
INSERT INTO evaluations VALUES(19689,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(19689,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(19689,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(19689,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(19689,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(19689,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(19689,36,'cp 65','f1e2 g8f6 b1c3 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 a7a6 e1g1 c8d7 a2a4 a6a5 d3c2 d7c6 a1d1 f6d7 b2b3 d8c7 c3d5 c6d5 e4d5 d7c5 g2g3 f7f5 g1h1 b7b6 h1g2 g7e5 e3h6 f8f7 f2f3 e5g7');
INSERT INTO evaluations VALUES(19865,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19865,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(19865,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(19865,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(19865,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19865,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19865,36,'cp 0','c8g4 c1f4 e7e6 b1d2 g8f6 e1g1 f8d6 f4d6 d8d6 c2c3 e8g8 f1e1 g4f5 e2b5 d6c7 f3h4 f5e4 h4f3');
INSERT INTO evaluations VALUES(20100,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(20100,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20100,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20100,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20100,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20100,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20100,36,'cp 4','g7g6 f1e2 f8g7 e1g1 d8b6 e3b6 a7b6 a2a3 c8e6 f3g5 e6d7 b1d2 c6d4 e2d3 h7h6 g5f3 d4f3 d2f3 d7c6 c2c3 g8f6 e4e5 d6e5 f3e5 e7e6 c1e3 f6d5 e3d4 e8g8 f1e1 f8d8 d3c2 b6b5 a1d1 d5f4 e5c6 b7c6 d4c5 f4d5 c2e4');
INSERT INTO evaluations VALUES(23404,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(23404,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(23404,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(23404,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(23404,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(23404,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(23404,36,'cp 58','d7d6 d2d4 g8f6 d4d5 c6b8 b1c3 f8g7 h2h3 e8g8 a2a4 e7e6 e1g1 e6d5 e4d5 b8d7 c1g5 h7h6 g5e3 d7c5 d1c1 g8h7 e3d4 f8e8 b2b4 c5e4 c3e4 f6e4 c1b2 g7d4 b2d4 a7a5 a1e1 f7f5 b4b5 c8d7 c4d3 e4c5 e1e8 d7e8');
INSERT INTO evaluations VALUES(23794,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(23794,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(23794,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(23794,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(23794,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(23794,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(23794,36,'cp -40','d1e2 g8e7 b1d2 e7g6 d2b3 c5e7 b3d4 c6d4 b5d7 d8d7 f3d4 e8g8 c2c3 f7f6 d4f3 a8e8 e1g1 e7d8 e5f6 d8f6 c1e3 e6e5 f1d1 e5e4 f3d4 g6e5 e2b5 f8f7 b5d7 f7d7 a1b1 e5c4 d4e2 a7a6 e3d4 f6d4 e2d4');
INSERT INTO evaluations VALUES(25785,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(25785,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(25785,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(25785,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(25785,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(25785,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(25785,36,'cp 62','d2d4 d8f6 d1d3 g4f3 g2f3 e5d4 c3b5 a8b8 b5d4 d6d5 h1g1 c6c5 d4b5 f8d6 c1g5 f6b2 b5d6 c7d6 e1d2 b2b4 d2e2 b4b5 a1b1 b5d3 e2d3 b8b6 b1b6 d5e4 f3e4 a7b6 g1b1 g8f6 b1b6 e8d7 g5f6 g7f6 b6a6 h8b8 a6a7 d7e6 a2a4 b8b4 a4a5 b4d4 d3e3 d4a4 a5a6 h7h5 f2f3 h5h4 a7a8 f6f5 e3f4 f5e4 f3e4 f7f6');
INSERT INTO evaluations VALUES(26336,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(26336,31,'cp 297',NULL);
INSERT INTO evaluations VALUES(26336,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(26336,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(26336,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(26336,35,'cp 272',NULL);
INSERT INTO evaluations VALUES(26336,36,'cp 310','c4f7 e8f7 f3g5 f7e8 d1g4 e7e5 f4e5 c6e5 g4g3 g8f6 d2d3 d8d7 e1g1 f8e7 b1c3 h8f8 f1f4 b7b5 c1d2 h7h6 g5f3 b5b4 c3d1 e5f3 g3f3 e8f7 d1e3 f7g8 e3f5 g8h7 a1f1 a8e8 f3h3 e7d8 f4f3 f6g8 d2f4 f8f6 c2c4 b4c3 b2c3 d8b6 g1h1 b6c7 c3c4 g8e7');
INSERT INTO evaluations VALUES(26220,30,'cp 301',NULL);
INSERT INTO evaluations VALUES(26220,31,'cp 292',NULL);
INSERT INTO evaluations VALUES(26220,32,'cp 281',NULL);
INSERT INTO evaluations VALUES(26220,33,'cp 309',NULL);
INSERT INTO evaluations VALUES(26220,34,'cp 269',NULL);
INSERT INTO evaluations VALUES(26220,35,'cp 305',NULL);
INSERT INTO evaluations VALUES(26220,36,'cp 281','e8d7 f7b3 e7e6 d2d3 g7g5 f4g5 h7h6 h2h3 h6g5 c1e3 d8f6 b1d2 f8e7 c2c3 g4f3 g2f3 a8f8 d1e2 e7d8 e1c1 g8e7 c1b1 d7c7 d1g1 e7g6 e3g5 f6f7 g5e3 g6f4 e3f4 f7f4 g1g7 f8f7 g7f7 f4f7 b3a4 f7f6 a4c6 c7c6 b1c2 d8e7');
INSERT INTO evaluations VALUES(27697,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(27697,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(27697,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(27697,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(27697,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(27697,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(27697,36,'cp 182','e6d7 b3h3 d7e8 e1g1 c6d4 h3h5 e8d7 h5d1 a8d8 d2d3 d7c8 b1d2 h7h5 d2e4 f6f7 c1e3 f8e7 d1d2 c8b8 a2a3 f7f5 e3d4 e5d4 f2f4 f5d5 a1e1 h5h4 h2h3 d5b5 d2c1 a7a6 b2b3 h8f8 b3b4 f8f5 c1b2 f5f8 b2b3 b5d5 b3d5');
INSERT INTO evaluations VALUES(27700,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(27700,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(27700,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(27700,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(27700,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(27700,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(27700,36,'cp 172','b3h3 d7e8 e1g1 h7h5 b1c3 c6d4 c3e4 f6g6 h3d3 a8d8 f1e1 a7a6 c2c3 d4b3 e4d6 g6d6 d3d6 c7d6 a2b3 e8f7 d2d3 f8e7 g1f1 g7g5 a1a5 g5g4 c1e3 e7f6 b3b4 d8d7 b4b5 a6b5 a5b5 d6d5 e1a1 f7g6 a1a7 d5d4 c3d4 e5d4 e3f4 h8c8 b5b6');
INSERT INTO evaluations VALUES(29644,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(29644,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(29644,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29644,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29644,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(29644,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29644,36,'cp 0','g4f3 g2f3 e7e6 d1d3 f8d6 f4d6 d8d6 e1c1 b8c6 e2e4 f6h5 d3e3 e8c8 a2a3 g7g6 c1b1 c8b8 h2h4 h5g7 e3h6 g7h5');
INSERT INTO evaluations VALUES(30928,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(30928,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(30928,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(30928,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(30928,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(30928,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(30928,36,'cp 52','g8f6 c1g5 b8c6 d1d2 a7a6 e1c1 e7e6 c1b1 f8e7 f2f3 d8c7 d4c6 d7c6 h2h4 a8d8 c3e2 h7h6 g5e3 f6d7 e2d4 d6d5 d2c3 d5e4 d4c6 c7c6 c3g7 e7f6 g7g4 e4f3 g2f3 d7e5 d1d8 e8d8 g4g3 c6f3 g3f3 e5f3 f1g2 f3e5 g2b7');
INSERT INTO evaluations VALUES(28875,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(28875,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(28875,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(28875,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(28875,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(28875,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(28875,36,'cp 56','c1g5 b8c6 d1d2 c6d4 d2d4 d8a5 g5d2 a8c8 f1d3 e7e5 d4e3 a7a6 e3g3 a5d8 e1g1 g7g6 f1d1 f8g7 d2g5 h7h6 g5f6 g7f6 d3e2 f6h4 g3d3 e8g8 g2g3 h4e7 c3d5 d7b5 d3f3 b5c6 c2c4 h6h5 h2h4 c6d5 d1d5 c8c5 a1d1 c5d5 d1d5');
INSERT INTO evaluations VALUES(35350,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(35350,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(35350,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(35350,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(35350,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(35350,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(35350,36,'cp 91','c2c3 e7e6 d2d4 f8e7 f1d3 d7d5 e4d5 e6d5 d4c5 c7c5 c1e3 c5a5 e1g1 g8f6 f1e1 e8g8 b1d2 a5c7 h2h3 b8c6 a1c1 h7h6 b2b4 a7a6 a2a4 f8e8 a4a5 e7d6 d1c2 c8d7 e3b6 c7c8 e1e8 d7e8 d2b3 d6c7 c1d1 c7b6 a5b6 c8d8');
INSERT INTO evaluations VALUES(39616,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39616,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(39616,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39616,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(39616,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(39616,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(39616,36,'cp 0','c7c5 g2g3 b8c6 g1e2 d7d6 f1g2 e7e6 c1e3 c6d4 e3d2 d4c6');
INSERT INTO evaluations VALUES(42932,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(42932,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(42932,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(42932,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(42932,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(42932,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(42932,36,'cp 49','b7b6 f1e2 e8g8 e1g1 c8b7 f3g5 c7c5 b2b3 f6e8 g5h3 e7f6 c1b2 c5d4 e3d4 b8c6 d4d5 c6d4 e2d3 g7g6 d1d2 e6e5 f2f3 b7c8 a1e1 d7d6 c3e2 d4e2 d3e2 e8g7 a2a4 d8e7 b2c3 f6h4 g2g3 h4f6 f3f4 e5e4');
INSERT INTO evaluations VALUES(47425,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(47425,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47425,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47425,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(47425,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(47425,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(47425,36,'cp 5','c7c6 a4c4 c8f5 g2g3 e7e6 b1c3 b8d7 f1g2 f8e7 e1g1 e8g8 f1e1 f6e4 c1f4 e4c3 b2c3 f5e4 c4b3 b7b5 f3e5 e4g2 g1g2 d7e5 f4e5 a7a5 a2a4 d8d5 b3d5 e6d5 e2e4 b5b4 e4d5 c6d5 e5c7 f8e8 c3b4 e7b4 e1e5 a8c8 a1c1 b4d2 c1c2 e8e5 d4e5');
INSERT INTO evaluations VALUES(49007,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(49007,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(49007,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(49007,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(49007,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(49007,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(49007,36,'cp 110','d5b6 c4b3 b8c6 f1e1 f8d6 d2d4 e8g8 c1g5 d8d7 d4e5 c6e5 b1d2 e5f3 d1f3 h7h6 g5h4 a7a5 a2a4 d6b4 c2c3 d7d2 e1e2 d2d6 h4g3 d6g6 c3b4 a5b4 b3c2 g6c6 c2e4 c6c4 b2b3 c4c3 f3c3 b4c3 g3c7 b6d7 e4c2 d7f6 e2e5 c8d7 e5c5 f8e8 c5c3');
INSERT INTO evaluations VALUES(49934,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(49934,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(49934,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(49934,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(49934,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(49934,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(49934,36,'cp 48','f8b4 c1d2 d7d6 g1f3 a7a5 d1c2 e8g8 f1g2 d8e7 a2a3 b4c3 d2c3 a5a4 e1g1 f6e4 a1c1 b8d7 f1d1 e4c3 c2c3 c7c6 d1e1 e6e5 e2e4 f5e4 e1e4 d7f6 e4e3 e5e4 f3d2 c8f5 f2f3 d6d5 f3e4 d5e4 c1e1 e7f7 d2e4 f6e4 g2e4 f5e4 e3e4 f7f2 g1h1');
INSERT INTO evaluations VALUES(42914,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(42914,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(42914,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(42914,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(42914,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42914,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(42914,36,'cp 45','c1d2 e8g8 f1g2 d7d6 g1f3 a7a5 d1c2 d8e7 a2a3 b4c3 d2c3 a5a4 e1g1 f6e4 a1d1 e4c3 c2c3 b8c6 f3e1 e6e5 d4e5 e7e5 c3d2 c8e6 g2d5 e6d5 c4d5 c6a5 e1f3 e5e4 f3g5 e4c4 d2d4 c4d4 d1d4 a5b3 d4b4 b7b6 g5e6');
INSERT INTO evaluations VALUES(50598,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(50598,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(50598,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(50598,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50598,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(50598,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(50598,36,'cp 68','g1f3 g8f6 e2e4 b8c6 d2d4 c5d4 f3d4 e7e5 d4f5 d7d6 f5e3 g7g6 f1d3 f8g7 a2a4 e8g8 e1g1 c8d7 a1b1 c6d4 b2b3 a8c8 b3b4 b7b6 e3d5 f6d5 c3d5 d7c6 d5c3 f7f5 e4f5 g6f5 c1e3 g8h8 e3d4 e5d4 c3d5 c6d5');
INSERT INTO evaluations VALUES(54495,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(54495,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(54495,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(54495,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(54495,34,'cp -124',NULL);
INSERT INTO evaluations VALUES(54495,35,'cp -138',NULL);
INSERT INTO evaluations VALUES(54495,36,'cp -133','c4b5 c8d7 b5d7 d8d7 g1f3 b8c6 e1g1 f7f6 d2d4 e8c8 c1e3 g8h6 c2c3 f8d6 g1h1 c5c4 b2b3 h8e8 e3g1 c6a5 f3h4 d6c7 b1a3 e8e4 b3c4 a5c4 a3c4 d5c4 a1b1 c7f4 b1b4 f4c7 h4f3 d8e8 b4c4 e4e2 c4a4 a7a6 f1e1 e2e1 f3e1 h6f5 a4b4');
INSERT INTO evaluations VALUES(54467,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(54467,31,'cp -138',NULL);
INSERT INTO evaluations VALUES(54467,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(54467,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(54467,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(54467,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(54467,36,'cp -122','c8d7 b5d7 d8d7 g1f3 b8c6 c2c3 e8c8 e1g1 f7f6 d2d4 g8h6 c1e3 d8e8 f1e1 h6g4 e3d2 e8e1 d1e1 c5c4 b2b3 c6a5 b3b4 a5c6 b1a3 f8d6 a3c2 c6e7 e1e2 e7f5 a1e1 d6c7 e2e6 d7e6 e1e6 c8d7 e6e1 h7h5 g2g3 c7d6 e1e2 h5h4 f3h4 f5h4 g3h4 h8h4 d2e1 h4h7 e1g3 f6f5 c2e3 g4e3 e2e3');
INSERT INTO evaluations VALUES(55086,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(55086,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(55086,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(55086,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(55086,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(55086,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(55086,36,'cp -41','g8f6 g1f3 c8f5 e2e3 e7e6 c1b2 f8e7 f1e2 e8g8 d2d3 h7h6 e1g1 c7c5 f3e5 f6d7 e5d7 d8d7 b1d2 b8c6 e2h5 b7b5 d2f3 f5h7 a2a4 a7a6 d1e2 a8c8 g1h1 c5c4 b3c4 b5c4 d3c4 d5c4 e2c4 c6b4 f3e5 d7b7');
INSERT INTO evaluations VALUES(55291,30,'cp -237',NULL);
INSERT INTO evaluations VALUES(55291,31,'cp -242',NULL);
INSERT INTO evaluations VALUES(55291,32,'cp -237',NULL);
INSERT INTO evaluations VALUES(55291,33,'cp -232',NULL);
INSERT INTO evaluations VALUES(55291,34,'cp -234',NULL);
INSERT INTO evaluations VALUES(55291,35,'cp -237',NULL);
INSERT INTO evaluations VALUES(55291,36,'cp -231','b8c6 b1c3 g8f6 f1b5 d8d6 d2d4 e4d3 b5d3 c8g4 g1e2 c6b4 d3b5 g4d7 e1g1 d6d1 f1d1 b4c2 b5d7 f6d7 a1b1 e8c8 a2a3 e7e5 d1d2 c2d4 e2d4 e5d4 d2d4 f8c5 d4c4 c7c6 c1g5 f7f6 g5f4 h8e8 b2b4 c5b6 b4b5 c6c5 c3a4 d7e5 c4e4');
INSERT INTO evaluations VALUES(56397,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(56397,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(56397,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(56397,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(56397,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(56397,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(56397,36,'cp -31','g8f6 g1f3 b8c6 f1e2 f8d6 a2a3 a7a5 e4d5 f6d5 d2e4 e8g8 e1g1 c8f5 f3h4 f5e6 h4f3 d6e7 e4g3 f7f5 f1e1 d8d6 e2f1 a8d8 d1e2 e7f6 g3h5 e6f7 h5f6 g7f6 f3h4 c6d4 e2d1 d5e7 c1d2 a5a4 d2c3 f5f4 c3d4 d6d4');
INSERT INTO evaluations VALUES(48841,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(48841,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(48841,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(48841,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(48841,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48841,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(48841,36,'cp -12','g1f3 b8c6 f1e2 f8d6 e1g1 e8g8 f1e1 f8e8 e2f1 a7a5 e4d5 f6d5 h2h3 c8f5 d2c4 a5a4 c2c3 h7h6 c1d2 d5f6 a2a3 f5e6 c4d6 d8d6 c3c4 f6d7 d2c3 c6d4 f3d4 e5d4 c3d2 d7c5 d1f3 c5b3 d2f4 d6b6 a1b1 c7c5 e1e5 f7f6');
INSERT INTO evaluations VALUES(56812,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(56812,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(56812,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(56812,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(56812,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(56812,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(56812,36,'cp 61','d5e4 f3d4 f8c5 c1e3 c5d4 d1d4 d8d4 e3d4 f6f5 b1c3 b8c6 e1c1 c8e6 d4e3 a8d8 d1d8 e8d8 f2f3 g8e7 f3e4 f5e4 f1e2 d8c8 c3e4 c6e5 h1e1 e6d5 e2d1 d5e4 e3d4 e4g2 d4e5 e7g6 e5g7 h8g8 g7d4 b7b6 e1e6 g2d5 e6f6 g8f8 d1g4 c8b7 b2b3');
INSERT INTO evaluations VALUES(56835,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(56835,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(56835,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(56835,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(56835,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(56835,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(56835,36,'cp 66','f3d4 f8c5 c1e3 c5d4 d1d4 d8d4 e3d4 f6f5 b1c3 b8c6 e1c1 c8e6 d4e3 g8e7 f2f3 a8d8 d1d8 e8d8 f3e4 c6e5 f1e2 f5e4 h1d1 e6d7 c3e4 d8c8 d1e1 d7c6 e3d4 e7g6 e2f1 e5g4 e4c5 g4h2 f1c4 h8d8 d4g7 c6g2 c4e6 c8b8 c5d7 b8a8 g7f6 h2f3');
INSERT INTO evaluations VALUES(15947,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(15947,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(15947,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15947,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(15947,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(15947,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(15947,36,'cp 15','g1f3 a7a6 d1c2 d5c4 a2a4 c7c5 f1c4 d8c7 c2e2 f8d6 h2h3 e8g8 e1g1 c5d4 e3d4 b7b6 d4d5 e6d5 c3d5 f6d5 c4d5 c8b7 f1d1 b7d5 d1d5 d7f6 d5d3 f8e8 c1e3 d6c5 e2c2 a8c8 a1c1 c7b8 e3c5 c8c5 d3c3 e8c8 c2e2 b8b7 c3c5 b6c5 a4a5 h7h6 c1c2');
INSERT INTO evaluations VALUES(17546,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(17546,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(17546,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(17546,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(17546,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(17546,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(17546,36,'cp 65','b7c6 f1c4 e7e6 d1e2 f8e7 c1f4 e6e5 f4g3 g8f6 e1g1 c8g4 f2f3 g4d7 g3f2 a6a5 b2b3 e8g8 f1d1 d8c7 d1d2 f8d8 a1d1 f6h5 g2g3 a8b8 g1g2 h5f6 c3a4 h7h6 d2d3 b8b7');
INSERT INTO evaluations VALUES(22573,30,'cp -201',NULL);
INSERT INTO evaluations VALUES(22573,31,'cp -206',NULL);
INSERT INTO evaluations VALUES(22573,32,'cp -205',NULL);
INSERT INTO evaluations VALUES(22573,33,'cp -204',NULL);
INSERT INTO evaluations VALUES(22573,34,'cp -201',NULL);
INSERT INTO evaluations VALUES(22573,35,'cp -210',NULL);
INSERT INTO evaluations VALUES(22573,36,'cp -201','c6e5 d2d4 e5c6 g1f3 g7g6 b1c3 f8g7 c1e3 g8f6 h1g1 e8g8 d1d2 a7a6 g2g4 b7b5 f3g5 c6e7 e1c1 c8b7 f1g2 b7g2 g1g2 b5b4 c3e4 f6e4 g5e4 e7d5 g4g5 a6a5 h2h4 a5a4 h4h5 b4b3 c2b3 a4b3 a2a3 f8e8 h5g6');
INSERT INTO evaluations VALUES(24463,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24463,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(24463,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(24463,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(24463,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24463,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24463,36,'cp 14','e1g1 e7e6 f1e1 f8d6 f3e5 e8g8 c1g5 d6e7 b5c6 d7c6 c3e2 d8b6 e2g3 a8e8 b2b3 f6e4 g5e7 e8e7 g3e4 d5e4 d1g4 c6d5 c2c4 b6d4 g4f4 f7f6 e5f3 d4b6 c4d5 e4f3 e1e6 e7e6 d5e6 b6e6 f4f3 f8e8 g2g3 e6e4 g1g2 e4f3 g2f3 e8d8 a1c1 d8d7 b3b4 g8f7 h2h4 d7d2 c1c7 f7g6 c7b7 d2a2 f3e3');
INSERT INTO evaluations VALUES(24462,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24462,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(24462,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24462,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24462,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24462,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24462,36,'cp 10','d7g4 h2h3 g4f3 d1f3 e7e6 c3e2 d8b6 a2a4 a7a6 b5c6 b6c6 c2c3 f8d6 c1f4 a6a5 f4d6 c6d6 e2f4 e8g8 f4d3 d6c7 f1c1 f8c8 f3e3 f6e8 e3e2 e8d6 g2g3 h7h6 h3h4 c7c6 c1e1 c8b8 g3g4 b7b5 g4g5 h6g5 a4b5 b8b5');
INSERT INTO evaluations VALUES(26672,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(26672,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(26672,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(26672,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(26672,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(26672,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(26672,36,'cp 82','e7e6 b1d2 f8d6 b2b3 g8f6 c1b2 e8g8 h2h3 g4h5 a2a3 h5g6 e1g1 a7a5 f1e1 f6e4 d2e4 g6e4 c4d3 e4d3 d1d3 c6e7 e3e4 d8d7 a1d1 e7g6 g2g3 f7f5 e4f5 e6f5 f3e5 d6e5 d4e5 d7d3 d1d3 c7c6');
INSERT INTO evaluations VALUES(28345,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(28345,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(28345,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(28345,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(28345,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(28345,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(28345,36,'cp 32','d1b3 b8d7 c4d5 f6d5 g5e7 d8e7 f1e2 e8g8 e1g1 f8b8 a2a3 h7h6 h2h3 a7a5 a1c1 d5f6 c3a4 f5e4 f3d2 e4f5 f1e1 e6e5 d4e5 e7e5 d2f3 e5e7 f3d4 f5e4 e1d1 g8h8 a4b6 d7b6 b3b6 e4g6 b6c5 e7c5 c1c5 f6e4');
INSERT INTO evaluations VALUES(31094,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(31094,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(31094,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(31094,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31094,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31094,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31094,36,'cp 0','e7e6 c2c4 h7h6 b1c3 b8d7 f3d2 g8f6 d1b3 d8b6 c4c5 b6b3 a2b3 a7a6 b3b4 e8c8 h2h4 h8g8 d2f3 f8e7 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(31460,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(31460,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(31460,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31460,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31460,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31460,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(31460,36,'cp 0','c2c4 h7h6 e1g1 b8d7 b1c3 g8f6 f3d2 f8e7 e2e4 d5e4 d2e4 f6e4 g2e4 f5e4 c3e4 e8g8 c1f4 d7f6 f1e1 d8a5 f4e5 f6d7 e5f4');
INSERT INTO evaluations VALUES(35271,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(35271,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(35271,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(35271,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(35271,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(35271,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(35271,36,'cp -52','g8f6 h5e2 d7d5 g1f3 b8c6 c2c3 c8e6 g2g3 d5d4 b1d2 e7d7 c3d4 c6d4 f3d4 d7d4 a2a3 f8c5 d2f3 d4d6 b2b4 c5b6 c1b2 f6d7 f1g2 e8g8 e1g1 a7a5 b2c3 f7f6 f1b1 a5b4 c3b4 b6c5 e2c2 c7c6 c2c3 f8b8 a3a4 b7b6 b4c5 d7c5 d3d4');
INSERT INTO evaluations VALUES(34037,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(34037,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(34037,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(34037,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(34037,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(34037,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(34037,36,'cp -55','h5e2 d7d5 g1f3 b8c6 c2c3 c8e6 g2g3 d5d4 b1d2 e7c5 c3d4 c6d4 f3d4 c5d4 a2a3 f8c5 d2f3 d4d6 b2b4 c5b6 c1b2 f6d7 f1g2 a7a5 e1g1 e8g8 b2c3 f8d8 f1b1 a5b4 c3b4 b6c5 f3g5 c5b4 g5e6 d6e6 a3b4 a8a1 b1a1 e6b3 a1c1 d7f8 c1c7 b3b4 c7c4 b4b3 e2c2 d8d3 c2b3');
INSERT INTO evaluations VALUES(44902,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(44902,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(44902,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(44902,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(44902,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(44902,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(44902,36,'cp -46','d5c4 e2e4 b7b5 a2a4 b5b4 c3b1 c8a6 c1e3 e7e6 b1d2 c4c3 b2c3 b4c3 d2b3 f8b4 d1c2 e8g8 f1a6 b8a6 g1e2 c6c5 d4c5 f6d7 e1g1 a8c8 e2c3 a6c5 c3b5 a7a6 b5d4 a6a5 f1c1 d8f6 c2e2 h7h6 b3c5 d7c5 c1c4 c5d7 a1c1 d7b6 c4c8 f8c8 c1c8 b6c8 d4b3');
INSERT INTO evaluations VALUES(45303,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(45303,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(45303,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(45303,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(45303,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(45303,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(45303,36,'cp -10','e7e6 f1e2 f8c5 a2a3 e8g8 e1g1 a7a5 c2c4 d5c4 e2c4 b8d7 b2b3 d8e7 c1b2 b7b6 d1e2 c8b7 f1d1 f8d8 b1d2 f6e8 c4d3 h7h6 h2h3 e8d6 a1c1 e6e5 e3e4 a8c8 d3a6 c8c7 a6b7 d6b7');
INSERT INTO evaluations VALUES(46101,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(46101,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(46101,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46101,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46101,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(46101,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(46101,36,'cp 0','c2c4 f8c5 c4d5 e6d5 f1e2 e8g8 b1c3 a7a6 e1g1 b8c6 b2b3 c5a7 c1b2 f8e8 a1c1 c8f5 e2d3 f5g4 h2h3 g4f3 d1f3 c6e5 f3e2 e5c6');
INSERT INTO evaluations VALUES(46092,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46092,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46092,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46092,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46092,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46092,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(46092,36,'cp -6','f8c5 c4d5 e6d5 f1e2 e8g8 b1c3 b8c6 e1g1 a7a6 b2b3 c5a7 c1b2 f8e8 a1c1 c8f5 e2d3 f5g4 h2h3 g4h5 g2g4 h5g6 d3g6 h7g6 g4g5 f6e4 c3d5 e4g5 c1c4 g5e4 d5c3 e4c5 d1e2 d8e7 f1d1 a8d8 d1d8');
INSERT INTO evaluations VALUES(46197,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46197,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(46197,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(46197,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(46197,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(46197,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(46197,36,'cp -9','f8d6 f4d6 d8d6 e2e3 b7b6 f1e2 e8g8 e1g1 b8c6 b1d2 e6e5 d4e5 c6e5 d1c2 e5f3 d2f3 d6e7 f1d1 f8d8 a2a4 h7h6 a4a5 c8d7 a5b6 a7b6 a1a8 d8a8 f3d2 d7c6 e2f3 a8a2 c2b1 a2a6 c3c4 b6b5 c4d5 c6d5 f3d5 f6d5 d2f1 d5f6 f1g3 g7g6');
INSERT INTO evaluations VALUES(46292,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(46292,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(46292,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(46292,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(46292,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(46292,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(46292,36,'cp -18','f8e7 c4d5 f6d5 g5e7 d5e7 g1f3 b8d7 f1d3 e8g8 e1g1 d7f6 f3e5 b7b6 d1f3 e7d5 c3d5 d8d5 f3d5 f6d5 a1c1 d5f4 f1d1 c8b7 g2g3 f4d5 d3e4 a8c8 e5c4 f8d8 e4g2 f7f6 c4e3 d5e3 f2e3 b7g2 g1g2 g8f7');
INSERT INTO evaluations VALUES(40229,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(40229,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(40229,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(40229,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(40229,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(40229,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(40229,36,'cp -25','c4d5 f6d5 g5e7 d5e7 g1f3 b8d7 f1d3 e8g8 e1g1 a8b8 f1e1 d7f6 f3e5 b7b6 d1d2 c8b7 a1d1 e7d5 c3e4 b8c8 d1c1 c8c1 e1c1 h7h6 a2a3 f6e8 e4c3 d5c3 b2c3 e8f6 d2e3 d8d6 a3a4 f6d5 e3e4 f7f5');
INSERT INTO evaluations VALUES(47002,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47002,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47002,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(47002,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(47002,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47002,36,'cp 0','c7c5 d4c5 e7e5 f4g5 d6d5 e2e3 b8c6 b1c3 c8e6 f1b5 f8c5 g5f6 g7f6 d1f3 a7a6 b5a4 d8b6 f3f6 b6b2 a4c6 b7c6 f6h8 e8e7 h8a8 b2c3 e1e2 c3a1 a8b7 e7f6 b7c6 c5b4 a2a3 b4a3 c6a6 f6g7 g1f3 a1h1 a6a3 h1g2 f3e5 e6h3 a3e7 g2f1 e2f3 f1g2');
INSERT INTO evaluations VALUES(52361,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(52361,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(52361,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(52361,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(52361,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(52361,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52361,36,'cp 15','e7e5 b1c3 d7d5 d2d4 e5d4 e3d4 g8f6 f1d3 c8e6 c4d5 f6d5 g1e2 b8d7 c3d5 e6d5 e1g1 f8e7 e2f4 d7f6 f4d5 f6d5 d1g4 g7g6 c1h6 d8d7 g4f3 e8c8 f3f7 e7f6 f7d7 d8d7 a1e1 f6d4 e1e2 d4g7 h6g7 d7g7 f2f4 g7e7 g2g3 e7e2 d3e2 d5e3 f1e1 c8d7 e2f3 h8e8 g1f2');
INSERT INTO evaluations VALUES(51849,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(51849,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(51849,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(51849,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(51849,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(51849,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(51849,36,'cp 18','d2d4 e5d4 d1d4 g8f6 g1f3 d7d5 f1e2 f8e7 e1g1 e8g8 b1c3 d5c4 d4c4 b8d7 f1d1 d8a5 c4f4 f8d8 h2h3 h7h6 e3e4 d7f8 f4g3 f6h5 g3e5 a5e5 f3e5 h5f6 d1d8 e7d8 e5c4 c8e6 c1e3 d8c7 f2f4 a8d8 a1d1 d8d1 e2d1');
INSERT INTO evaluations VALUES(52831,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(52831,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(52831,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(52831,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(52831,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(52831,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(52831,36,'cp -46','e7e5 c1b2 f7f6 e2e3 b8a6 d1h5 g7g6 h5e2 f8d6 g1f3 c8g4 e3e4 d5d4 c2c3 c6c5 b3b4 d6e7 b4b5 a6c7 a2a4 g8h6 d2d3 e8g8 e1g1 h6f7 e2c2 g4e6 b1d2 b7b6 b2a3 a8c8 a1c1 d8d7 c3d4 c5d4 a3e7 d7e7 d2c4 f7d6 f3d2 d6c4');
INSERT INTO evaluations VALUES(53773,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(53773,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(53773,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(53773,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(53773,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(53773,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(53773,36,'cp -39','c1b2 f7f6 d2d3 g8h6 e2e4 c8g4 g1e2 d5d4 h2h3 g4e6 b2c1 h6f7 f2f4 c6c5 f4f5 e6d7 e1g1 b8c6 a2a4 d8c7 b1a3 e8c8 a3b5 c7b6 b5a3 c8b8 a3c4 b6a6 c1d2 c6b4 h3h4 f8e7 g2f3 h7h5');
INSERT INTO evaluations VALUES(53774,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(53774,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(53774,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(53774,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(53774,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(53774,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(53774,36,'cp -52','f7f6 e2e3 b8a6 f2f4 e5e4 a2a3 g8h6 g1e2 a6c5 b3b4 c5a4 b2c1 a4b6 d2d3 f6f5 e1g1 a7a5 b4a5 a8a5 b1c3 a5a6 h2h3 b6d7 a3a4 d7f6 e2d4 h6f7 c1d2 h7h5 a4a5 h5h4 g3h4 h8h4 c3a4 h4h6 a4b6 c6c5 d4b5 f7d6 b5d6 f8d6 b6c8 d8c8');
INSERT INTO evaluations VALUES(54644,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(54644,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(54644,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(54644,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(54644,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(54644,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(54644,36,'cp -37','c2c4 c8e6 d2d4 e5e4 g1h3 f8d6 b1c3 b8a6 c4d5 c6d5 d1d2 a6c7 f2f3 f6f5 h3g5 e6d7 f3e4 d6e7 g5f3 f5e4 f3e5 g8f6 c3d1 e8g8 e1g1 d7e6 a2a4 f6d7 d1e3 c7a6 f1f8 e7f8 e5d7 d8d7 b2a3 f8a3 a1a3 d7e7 a3a1 e7b4 d2b4 a6b4 a1c1 b4c6 c1c5 c6d4 e3d5 d4b3');
INSERT INTO evaluations VALUES(54614,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(54614,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(54614,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(54614,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(54614,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(54614,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(54614,36,'cp -48','d2d4 e5d4 d1d4 g8f6 c1e3 b8c6 d4a4 f8b4 g1f3 f6g4 e3d2 b4c5 e2e3 e8g8 a4b5 c5b6 f1g2 d5d4 c3d5 d4e3 f2e3 c8d7 d5b6 a7b6 e1g1 c6d4 b5d3 d4f3 g2f3 d7c6 d3d8 f8d8 f3g4 d8d2 f1f2 d2d3 f2e2 a8a4 g4h3 g8f8 h3g2 a4a5 a2a3 a5f5 g2c6');
INSERT INTO evaluations VALUES(56696,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(56696,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(56696,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(56696,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(56696,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(56696,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(56696,36,'cp 118','g7g6 b1a3 b8d7 f1e2 f8g7 h2h4 h7h5 a3c4 g8h6 a2a4 a7a5 e1g1 e8g8 d1c2 h6g4 b2b4 b7b6 b4a5 b6a5 c4e3 d7b6 f1d1 g8h7 c3c4 g7h6 c4c5 h6e3 f2e3 b6d7 a1b1 d6c5 d4c5 d8e7');
INSERT INTO evaluations VALUES(979,30,'cp -155',NULL);
INSERT INTO evaluations VALUES(979,31,'cp -153',NULL);
INSERT INTO evaluations VALUES(979,32,'cp -147',NULL);
INSERT INTO evaluations VALUES(979,33,'cp -143',NULL);
INSERT INTO evaluations VALUES(979,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(979,35,'cp -156',NULL);
INSERT INTO evaluations VALUES(979,36,'cp -150','f8b4 c1d2 d5d6 f1e2 b4d2 d1d2 c8f5 a1d1 e8c8 a2a3 f5g4 e1g1 f7f5 f1e1 d6f6 d2g5 c6d4 g5f6 d4e2 e1e2 g7f6 h2h3 g4f3 g2f3 d8d4 g1f1 h8d8 e2d2 d4h4 d3d4 e5d4 b2b4 b7b6 f1g2 c7c5 g2g3 h4h6 b4c5 b6c5 d1e1 c8c7 e1e7 d8d7');
INSERT INTO evaluations VALUES(1971,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(1971,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(1971,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(1971,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(1971,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(1971,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(1971,36,'cp 35','b5f1 f8c5 d2d3 g5g6 b1d2 d7d5 c3d4 c5b6 d2f3 g8e7 f1e2 d5e4 d3e4 g6e4 e1g1 e8g8 f1e1 e7g6 e2d3 e4d5 d3e4 d5d6 a2a4 b6a5 c1d2 a5d2 d1d2 c8g4 a4a5 c7c6 h2h3 g4f3 e4f3 a8d8 a5a6 b7a6 a1a6 d6d4 d2d4 d8d4 f3c6 f8b8 a6a7');
INSERT INTO evaluations VALUES(2000,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(2000,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(2000,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(2000,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(2000,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(2000,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(2000,36,'cp 24','f8c5 d2d3 g5g6 b1d2 d7d5 g2g3 h7h5 c3d4 c5b6 d2f3 g8e7 f1g2 c8g4 f3e5 g4d1 e5g6 e7g6 e1d1 d5e4 c1e3 e8c8 d3e4 b6d4 e3d4 d8d4 d1e2 h5h4 f2f4 c7c6 a1d1 d4d1 h1d1 h4g3 h2g3 f7f6 d1d2 c8c7 b2b4 b7b6 e2f3 g6e7 d2d1 f6f5');
INSERT INTO evaluations VALUES(2409,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(2409,31,'cp 268',NULL);
INSERT INTO evaluations VALUES(2409,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(2409,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(2409,34,'cp 277',NULL);
INSERT INTO evaluations VALUES(2409,35,'cp 282',NULL);
INSERT INTO evaluations VALUES(2409,36,'cp 268','c4d5 c6d4 f3f7 e7d6 b1a3 d8e7 a3c4 d6d7 f7e7 f8e7 e1d1 c7c6 d5e4 e7f6 d2d3 d7c7 c2c3 d4f5 f2f4 f5d6 c4d6 c7d6 d1c2 e5f4 c1f4 f6e5 f4e5 d6e5 c3c4 e5d6 c2c3 c8e6 h1e1 a7a5 e1e3 a8f8 a2a3 f8f2 a1e1 e6d7 d3d4 b7b6 e4f3 g7g5 b2b4 g5g4 c4c5 b6c5 d4c5 d6c7');
INSERT INTO evaluations VALUES(2408,30,'cp 291',NULL);
INSERT INTO evaluations VALUES(2408,31,'cp 277',NULL);
INSERT INTO evaluations VALUES(2408,32,'cp 285',NULL);
INSERT INTO evaluations VALUES(2408,33,'cp 263',NULL);
INSERT INTO evaluations VALUES(2408,34,'cp 283',NULL);
INSERT INTO evaluations VALUES(2408,35,'cp 265',NULL);
INSERT INTO evaluations VALUES(2408,36,'cp 277','c6d4 f3f7 e7d6 b1a3 d8e7 a3c4 d6d7 f7e7 f8e7 d5e4 e7f6 c2c3 d4e6 e1g1 a7a5 d2d3 e6c5 e4f3 d7e7 c4e5 f6e5 d3d4 e5d6 d4c5 d6c5 f1e1 e7f7 c1e3 c5d6 e3d4 h8d8 e1e3 d6f4 e3e2 f4g5 a1e1 c7c6 h2h3 a5a4 g2g4 c8d7 f3e4 f7g8 d4b6 d8e8 e1d1 d7e6');
INSERT INTO evaluations VALUES(2411,30,'cp 292',NULL);
INSERT INTO evaluations VALUES(2411,31,'cp 284',NULL);
INSERT INTO evaluations VALUES(2411,32,'cp 262',NULL);
INSERT INTO evaluations VALUES(2411,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(2411,34,'cp 277',NULL);
INSERT INTO evaluations VALUES(2411,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(2411,36,'cp 270','f3f7 e7d6 b1a3 d8e7 a3c4 d6d7 f7e7 f8e7 e1d1 e7f6 d2d3 c7c6 d5e4 d7c7 c2c3 d4f5 d1c2 c8e6 f2f3 h7h5 a2a4 a7a5 f3f4 e6c4 d3c4 f5d6 e4g6 c6c5 h1e1 h8h6 f4e5 h6g6 e5d6 c7d7 g2g3 a8f8 e1e2 f6g5 c1g5 g6g5 a1e1 d7d6 c2b3 d6c7 e2e7 c7b6');
INSERT INTO evaluations VALUES(2415,30,'cp 273',NULL);
INSERT INTO evaluations VALUES(2415,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(2415,32,'cp 261',NULL);
INSERT INTO evaluations VALUES(2415,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(2415,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(2415,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(2415,36,'cp 258','e7d6 b1a3 d8e7 a3c4 d6d7 f7e7 f8e7 e1d1 c7c6 d5e4 e7f6 d2d3 d7c7 c2c3 d4f5 h1e1 f5d6 c4d6 c7d6 a2a4 c8e6 d1c2 a7a5 c1d2 h7h5 e1e2 g7g5 f2f3 e6d7 h2h3 a8f8 c3c4 f8a8 d2c3 h8h6 a1b1 h5h4 b1a1 h6h5 c3d2 h5h8 e2e1');
INSERT INTO evaluations VALUES(2421,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(2421,31,'cp 279',NULL);
INSERT INTO evaluations VALUES(2421,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(2421,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(2421,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(2421,35,'cp 271',NULL);
INSERT INTO evaluations VALUES(2421,36,'cp 260','b1a3 d8e7 a3c4 d6d7 f7e7 f8e7 e1d1 c7c6 d5e4 e7f6 d2d3 d7c7 c2c3 d4f5 a2a4 c8e6 h1e1 a7a5 d1c2 f5d6 c4d6 c7d6 c1d2 h7h5 e4g6 a8f8 f2f3 c6c5 e1e2 e6d7 b2b3 h5h4 a1e1 d6c7 h2h3 b7b6 f3f4 h8h6 g6e4 h6h8 e2f2 e5f4 d2f4 c7c8 f4h2 f6d8');
INSERT INTO evaluations VALUES(4344,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(4344,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(4344,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4344,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4344,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4344,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4344,36,'cp 0','c5e3 f2e3 d7d6 a2a4 c6e7 b1c3 e7g6 c3d5 f6e8 d5c3');
INSERT INTO evaluations VALUES(4453,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4453,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(4453,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4453,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(4453,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4453,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4453,36,'cp 7','f2e3 d7d6 a2a4 c6e7 b1c3 c7c6 h2h3 e7g6 d3d4 e5d4 e3d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 f1e1 c8e6 d1d2 d8d7 a4a5 e6d5 e4d5 d7d5 e1e8 a8e8 a1e1 e8e1 d2e1 h7h6 e1b4 d5d7 c2c3 a7a6 f3e5 g6e5 d4e5 d7d5 b4b7');
INSERT INTO evaluations VALUES(4956,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4956,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(4956,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4956,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4956,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4956,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4956,36,'cp -3','c2c3 c5b6 d2d4 e5d4 c3d4 g8e7 b1d2 e8g8 a2a4 a7a5 d2c4 e7g6 a1a3 d6d5 c4b6 c7b6 e4e5 c8f5 a3c3 d8d7 h2h3 b6b5 f3d2 g6f4 d2b3 f4e6 b3c5 e6c5 c3c5 f8b8 d1d2 b5a4 c5a5 a8a5 d2a5 h7h6 a5a4');
INSERT INTO evaluations VALUES(8602,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(8602,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(8602,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(8602,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8602,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(8602,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(8602,36,'cp 25','c4b3 e8g8 c2c3 f8e8 d2d3 h7h6 h2h3 e7f8 b1d2 c6a5 b3a4 c8d7 a4d7 d8d7 d3d4 e5d4 c3d4 a5c6 b2b3 a8d8 a2a3 c6e7 d4d5 c7c6 c1b2 c6d5 b2f6 g7f6 a1c1 d5e4 d2e4 f8g7 e4g3 f6f5 g3h5 e7g6 h5g7 e8e1 d1e1 g8g7 c1d1');
INSERT INTO evaluations VALUES(11148,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(11148,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11148,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11148,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11148,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11148,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(11148,36,'cp 0','c3d5 f6g4 a2a4 e7e6 d5e3 g4e3 c1e3 d6d5 c4a2 f8e7 e1g1 e8g8 c2c3 d5d4 c3d4 c5d4 e3d2 e7d6 a2c4 d8e7 d1e2 e6e5 d2g5 e7d7 g5d2');
INSERT INTO evaluations VALUES(11324,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(11324,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(11324,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(11324,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(11324,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(11324,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(11324,36,'cp 47','c8f5 a2a3 e7e6 c2c4 f8d6 f4g3 e8g8 f1e2 c6e7 e1g1 c7c6 b2b4 e7g6 d1b3 d6g3 h2g3 d8e7 b3b2 h7h5 a3a4 f5g4 b4b5 a6b5 a4b5 h5h4 b5c6 b7c6 a1a8 f8a8 g3h4 g6h4 f1a1 a8a1 b2a1 h4f3 d2f3 d5c4 e2c4 g4h5 a1a8 e7e8');
INSERT INTO evaluations VALUES(24269,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(24269,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(24269,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(24269,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(24269,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(24269,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(24269,36,'cp 53','c2c4 e7e6 f1e2 f8d6 f4g5 e8g8 a2a3 h7h6 g5h4 g7g5 h4g3 g8g7 a1c1 a6a5 e1g1 c6e7 g3e5 c7c6 h2h3 d6e5 f3e5 f6d7 e5d7 d8d7 d2b3 d7c7 e2d3 c7d6 d3f5 e6f5 b3c5 b7b6 c5d3 f8h8 d1c2 d5c4 c2c4');
INSERT INTO evaluations VALUES(16108,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(16108,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(16108,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(16108,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(16108,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(16108,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(16108,36,'cp 136','f6g4 d1d5 d8e7 b1c3 g4e5 f3e5 c6e5 c1f4 e5c4 d5c4 c7c6 e1c1 b7b5 c4d4 e7c5 d4d3 b5b4 c3e2 f8e7 d3g3 e7g5 f4g5 c5g5 f2f4 g5g3 e2g3 a7a5 d1d6 c8b7 g3f5 g7g6 f5e3 c6c5 e3c4 e8e7 h1d1 h8d8 g2g3 a8a6 d6a6 b7a6 c4a5 d8h8 e4e5 a6e2 d1d2 e2f3 d2d6');
INSERT INTO evaluations VALUES(22772,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(22772,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(22772,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(22772,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(22772,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(22772,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(22772,36,'cp -48','g8f6 b1c3 h7h6 f1b5 f8c5 f3e5 c6d4 b5c4 e8g8 d2d3 d7d6 e5f3 c8g4 c1e3 c7c6 d1d2 g4f3 e3d4 c5d4 g2f3 b7b5 c4b3 f6h5 c3d1 d8f6 c2c3 f6f3 h1g1 d4f6 d1e3 h5f4 b3d1 f3h3 d1g4 h3h4 e1c1 d6d5 e4d5 f4d5 e3d5 c6d5');
INSERT INTO evaluations VALUES(16492,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(16492,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(16492,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(16492,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(16492,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(16492,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(16492,36,'cp -48','b1c3 h7h6 f1b5 f8c5 f3e5 c6e5 d2d4 c5d6 f2f4 e5c6 e4e5 d6b4 e5f6 d8f6 a2a3 b4a5 c1e3 a5c3 b2c3 d7d5 d1f3 c8f5 f3d5 e8g8 d5f3 f5c2 e1g1 c6a5 f1f2 f6f5 a3a4 c2e4 f3g3 a7a6 b5e2 f8e8 e2g4 f5f6 f4f5 a5c4 e3f4 c4d6 f4d6 c7d6 a1e1 d6d5 a4a5');
INSERT INTO evaluations VALUES(24527,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(24527,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(24527,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(24527,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(24527,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(24527,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(24527,36,'cp -12','d1e2 d8c7 c1g5 h7h6 g5f6 d7f6 c2c3 f8d6 g2g3 e8g8 a2a4 f8c8 b1a3 a7a6 a3c2 c7a5 f3d2 a5b6 f1b1 b6c6 b1c1 c8c7 d2f3 c6c4 e2c4 c7c4 a4a5 a8c8 c2e1 g7g5 f3d2 c4c7 d2b3 h6h5 e1d3 h5h4 b3c5 d6c5 d3c5 h4h3');
INSERT INTO evaluations VALUES(25364,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(25364,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(25364,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(25364,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(25364,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(25364,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(25364,36,'cp 143','f7e6 g5e6 d8d7 e6f8 h8f8 e1g1 e8c8 b1c3 b7b6 c1e3 c8b7 f2f3 a7a5 d1d2 h7h6 a1d1 d6d5 c3d5 f6d5 e4d5 d7d5 b2b3 c6d4 d1e1 d5c6 f1f2 g7g5 e3d4 e5d4 f2e2 d8e8 h2h3 a5a4 e2e4 e8e4 e1e4 f8f4 e4f4 g5f4 d2e2 a4a3 e2e4');
INSERT INTO evaluations VALUES(31259,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(31259,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(31259,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(31259,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(31259,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(31259,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(31259,36,'cp 63','h2h3 g4f3 f1b5 b8d7 d1f3 a7a6 b5d7 d8d7 e1g1 e7e6 d2d3 f8e7 b1c3 d7c6 c1d2 g8f6 f4f5 e8c8 a1e1 d8e8 f3g3 h8g8 f5e6 f7e6 e4e5 d6e5 c3e4 g8f8 d2c3 c5c4 e4d2 c4d3 c2d3 f6d7 f1f8 e7f8 d2c4 f8c5 g1h1');
INSERT INTO evaluations VALUES(32916,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(32916,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(32916,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(32916,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(32916,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(32916,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(32916,36,'cp 56','g8f6 d2d3 c7c6 e1g1 b8d7 a2a4 d6d5 e4d5 f6d5 c1d2 f7f6 a4a5 e8c8 g2g3 f8c5 f1e1 f6f5 b1c3 d5b4 a1c1 h8e8 g1g2 d7f6 e1e2 c5d6 c4b3 b7b5 a5b6 a7b6 c3a2 b4d5 c1e1 c8d7 a2c3 d5c7 c3d1 c7e6');
INSERT INTO evaluations VALUES(32687,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(32687,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(32687,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(32687,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(32687,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(32687,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(32687,36,'cp 65','d2d3 c7c6 c1d2 f8e7 a2a3 b8d7 b1c3 b7b5 c4a2 e8g8 e1g1 a7a5 c3e2 f8e8 f2f4 e5f4 e2d4 e7d8 g1h1 d8b6 d4f5 d6d5 d2f4 d5e4 f5d6 e8f8 d3e4 b6d4 c2c3 d4e5 a1d1 a5a4 d6f5 e5f4');
INSERT INTO evaluations VALUES(32686,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(32686,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(32686,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(32686,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(32686,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(32686,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(32686,36,'cp 56','c7c6 a2a3 d6d5 e4d5 f6d5 c1d2 b8d7 b2b4 e8c8 e1f1 f7f6 h3h4 f8e7 b1c3 d5c7 c3e4 d7b6 c4a2 b6d5 h4h5 c7e6 e4g3 e6d4 a1c1 h8e8 c2c3 d4e6 g3f5 g7g6 f5e7 e8e7 g2g3 d5c7 d2e3 e7d7 a2b1 c7d5 e3a7 g6g5 a7e3 d5e3 f2e3');
INSERT INTO evaluations VALUES(33150,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(33150,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(33150,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(33150,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(33150,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(33150,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(33150,36,'cp 178','d2d4 c5d4 f3d4 e5d4 e4e5 f6b6 d1f3 b6g6 b1a3 a7a6 c4d3 f7f5 d3f5 g6f7 f3d3 g8e7 f5e4 b8c6 a3c4 b7b5 c4d6 c7d6 e5d6 e8g8 d6e7 f7e7 c1d2 e7f7 d3b3 f7b3 a2b3 c8b7 f1e1 c6d8 e4d3 d8e6 d2b4 f8f6 b4d6 f6f7 f2f3 b7c6 d6e5 e6f4 e5f4 f7f4 e1e7 f4f7 e7e5 f7f4');
INSERT INTO evaluations VALUES(36698,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(36698,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36698,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36698,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36698,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36698,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36698,36,'cp 0','c2c4 c7c5 b1c3 b8c6 c1e3 a8b8 h2h3 a7a6 d3d4 c6d4 f3d4 c5d4 e3d4 c8e6 c3d5 e6d5 c4d5 d8a5 d4c3 a5b6 c3d4');
INSERT INTO evaluations VALUES(36715,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(36715,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36715,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(36715,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(36715,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(36715,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(36715,36,'cp 28','d5c4 f1c4 c7c5 d4d5 h7h6 g5h4 f6e8 c4d3 e8d6 e1g1 c8g4 h2h3 g4f3 d1f3 b8d7 f3e2 c5c4 d3c2 f7f5 a2a4 a8c8 g2g4 d7c5 a1d1 g7c3 b2c3 c5d3 c2d3 c4d3 e2d3 c8c4 h4g3 d8c8 g1h2 c4c3 d3b1 c8d7 g3f4 d6e4 f4h6 f8c8 h6f4');
INSERT INTO evaluations VALUES(38385,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38385,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(38385,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(38385,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(38385,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38385,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(38385,36,'cp 10','f7f5 b1c3 f5e4 c3e4 g8f6 e4f6 e7f6 f1e2 e8g8 g1f3 f8e8 e1g1 c7c5 c1d2 f6f5 d4c5 d6c5 d2c3 b8c6 d1d8 e8d8 f1e1 b7b6 c3g7 g8g7 f3g5 c6d4 e2f3 d4f3 g5f3 g7f6 g1f2 h7h6 h2h4 c8b7 a1d1 b7f3 f2f3 a8b8 d1d8 b8d8');
INSERT INTO evaluations VALUES(38756,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(38756,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(38756,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38756,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38756,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38756,36,'cp 0','b1c3 h4e7 d2d4 g8f6 c1f4 e8g8 f1f2 c7c5 h1e1 c5d4 f3d4 b8c6 f2g1 d8b6 f4e3 b6b2 c3d5 f6d5 c4d5 b2c3 d4c6 b7c6 e3d4 c8g4 d4c3 g4d1 a1d1 c6d5 e4d5 e7g5 c3d2 g5d2 d1d2 f8c8 e1e7 g8f8 e7b7 c8b8 b7c7 b8c8');
INSERT INTO evaluations VALUES(45623,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45623,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(45623,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(45623,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(45623,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(45623,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(45623,36,'cp 30','d2d4 d5c4 a2a4 f8b4 c1d2 a7a5 d1c2 e8g8 d2b4 a5b4 c2c4 d8a5 c4c2 b7b5 e1g1 c8b7 f3e5 b5a4 f1c1 b8d7 e5c4 a5c7 a1a4 c6c5 a4a8 b7a8 g2a8 f8a8 c4d2 a8c8 d4c5 c7c5 c2c5 c8c5 c1c5 d7c5 g1g2 g8f8 d2c4 f8e7 b1d2 f6d5 g2f3 c5d7 d2b3 d5b6 c4b6');
INSERT INTO evaluations VALUES(46549,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(46549,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(46549,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(46549,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(46549,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(46549,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(46549,36,'cp 95','b8d7 a2a4 a7a6 f1c4 g7g6 c1e3 f8g7 d1d2 d6d5 e4d5 d7b6 c4d3 b6d5 e3h6 g7h6 d2h6 e7g8 h6d2 g8f6 c3d5 e6d5 e1g1 e8g8 f1e1 d8d6 f3e5 g8g7 b2b4 f8e8 f2f3 c8e6 c2c3 c7c6 a4a5 f6d7 d3c2 d7e5 d4e5');
INSERT INTO evaluations VALUES(15989,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(15989,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(15989,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(15989,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(15989,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(15989,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(15989,36,'cp 93','a2a4 e7c6 f1c4 f8e7 e1g1 e8g8 f1e1 a7a5 c4f1 c6b4 c1f4 c7c6 f4g3 b7b6 c3b1 d6d5 e4d5 b4d5 c2c3 c8a6 f1a6 a8a6 b1d2 d8a8 d1e2 a8b7 h2h4 h7h6 d2c4 a6a8 c4d6 e7d6 g3d6 f8e8 c3c4 d5f6');
INSERT INTO evaluations VALUES(49758,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(49758,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(49758,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(49758,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(49758,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(49758,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(49758,36,'cp 30','b1c3 d7d5 c1g5 c7c6 e2e3 h7h6 g5h4 f8e7 f1d3 e8g8 g1e2 f8e8 e1g1 b8d7 d1c2 a7a5 c3a4 d7b6 a4b6 d8b6 e2c3 b6d8 h2h3 f6h5 h4e7 e8e7 c3a4 h5f6 f1e1 e7e6 a4c3 c8d7 a1c1 b7b6 b2b3 e6e8 c2d2 a8c8 c3a4 f6e4');
INSERT INTO evaluations VALUES(50233,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(50233,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(50233,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(50233,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(50233,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(50233,35,'cp -110',NULL);
INSERT INTO evaluations VALUES(50233,36,'cp -94','f6e4 f1d3 e4f6 e1g1 f8g7 c2c4 e8g8 b1c3 d7d6 h2h3 e7e6 f1e1 b8d7 c1g5 h7h6 g5e3 b7b6 d1d2 c8b7 e3h6 b7f3 g2f3 f6h7 g1g2 d8h4 h6g7 g8g7 e1e4 h4g5 d2g5 h7g5 e4g4 g5h7 f3f4 h7f6 g4g5 a8d8 g2f3 f6h7 g5g3 d7f6 a1g1');
INSERT INTO evaluations VALUES(51862,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(51862,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(51862,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51862,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(51862,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51862,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(51862,36,'cp 25','g8f6 b1d2 d7d6 c2c3 c8g4 f1e2 g4h5 h2h3 b8d7 g2g4 h5g6 f3h4 d7c5 g4g5 f6d7 h4g6 h7g6 d2f3 c5e6 h3h4 d6d5 e4d5 c6d5 d1b3 d7c5 b3b5 d8d7 d3d4 e5d4 c3d4 c5e4 c1e3 f8d6 b5d7 e8d7 f3e5 d7e7 e1c1 a8c8 c1b1 c8d8 e2f3 d6e5 d4e5 b7b6 f3g2 d5d4 g2e4 d4e3 d1d8 h8d8');
INSERT INTO evaluations VALUES(44112,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44112,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(44112,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44112,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(44112,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(44112,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44112,36,'cp 12','b1d2 d8c7 d3d4 e5d4 e4e5 f6d5 d2e4 d7d6 e5d6 f8d6 d1d4 e8g8 f1d3 d6e7 e1g1 b8d7 f1e1 c7b6 d4c4 b6b4 c1d2 b4c4 d3c4 d7b6 c4d5 c6d5 e4g3 c8e6 g3h5 g7g6 h5f4 e7f6 c2c3 b6c4 f4e6 f7e6 e1e6 c4b2 f3g5 f6g7');
INSERT INTO evaluations VALUES(55265,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(55265,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(55265,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(55265,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(55265,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(55265,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(55265,36,'cp -75','e7e5 f3f4 e5f4 c1f4 g8e7 f4g3 e7g6 g1f3 h7h5 c2c3 c7c5 c3d4 c5d4 f1e2 b8c6 h2h3 h5h4 g3h2 f8d6 e1g1 d6h2 g1h2 c8e6 h2h1 e8g8 b1d2 a8c8 a2a3 d8d6 f3g5 c6e5 d2f3 e5f3 e2f3 c8c5 g5e6 f7e6 d1b3 g6e5 f3e2 f8f1 a1f1 d6b6 b3b6');
INSERT INTO evaluations VALUES(1203,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(1203,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(1203,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(1203,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(1203,34,'cp -143',NULL);
INSERT INTO evaluations VALUES(1203,35,'cp -138',NULL);
INSERT INTO evaluations VALUES(1203,36,'cp -153','d4e4 d1e2 c8f5 b1c3 e4e2 e1e2 e8c8 d3e1 g7g6 d2d3 f8g7 c1e3 b7b6 a1d1 c6c5 e1f3 h7h6 d1d2 f5e6 e2d1 g8e7 d1c1 e6g4 d2e2 e7f5 e3d2 g4f3 g2f3 f5d4 e2e7 d8d7 e7d7 c8d7 f3f4 d4f3 d2e3 h8e8 h2h4 h6h5 c1d1');
INSERT INTO evaluations VALUES(2590,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(2590,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2590,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2590,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2590,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2590,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2590,36,'cp 0','c5b6 c1e3 c8e6 c4e6 f7e6 c3e2 b6e3 f2e3 a6a5 d1d2 c6e7 c2c4 f6d7 a1b1 b7b6 e2c3 e7g6 f3h2 d7f6 d2e2 d8e7 b1c1 f6h7 f1f8 g6f8 c1f1 a5b4 a3b4 h7g5 e2b2 f8d7 c3b5 d7f6 b2e2 g8h7 h2g4 f6g4 e2g4 a8a2 b5c3');
INSERT INTO evaluations VALUES(5141,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(5141,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(5141,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(5141,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(5141,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(5141,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(5141,36,'cp 55','g8f6 e1g1 e8g8 d4c6 d7c6 d1d8 f8d8 a2a4 f6g4 e3g5 d8d4 c4e2 e6e5 a4a5 c8e6 f2f3 g7h6 g5h6 g4h6 f1d1 d4d1 a1d1 g8f8 g2g4 g6g5 c3a4 f8e7 h2h4 g5h4 g1g2 h6g8 a4c5 b7b6 c5d3 e6c4 d1h1 c4d3 e2d3 h7h6 g2f2 b6a5 h1h4 a5a4');
INSERT INTO evaluations VALUES(7810,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(7810,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(7810,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(7810,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7810,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(7810,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(7810,36,'cp 62','e1g1 e8g8 d4c6 d7c6 d1d8 f8d8 f2f3 c8d7 f1d1 a7a5 a2a4 d7e8 g2g4 d8d1 a1d1 f6d7 g4g5 g7f8 g1f2 e6e5 f2e2 b7b6 d1a1 h7h5 b2b3 f8c5 c3d1 c5e3 e2e3 g8f8 d1b2 d7c5 b2d3 c5d3 c2d3 a8d8 f3f4 e5f4 e3f4 b6b5 a4b5 c6b5 c4d5');
INSERT INTO evaluations VALUES(5677,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(5677,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(5677,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(5677,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(5677,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(5677,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(5677,36,'cp 67','g8f6 h2h3 e8g8 b1d2 a7a5 f1b5 c6a7 b5e2 c7c5 e1g1 b7b6 a2a4 c8b7 d1b3 f6d7 f1e1 a8c8 a1c1 e7e6 c1d1 d8e7 e2f1 f8d8 f4g5 f7f6 g5f4 a7c6 e3e4 c5d4 e4d5 c6e5 d5e6 d7c5 b3b6 d4c3 b2c3');
INSERT INTO evaluations VALUES(8726,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(8726,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(8726,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(8726,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(8726,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(8726,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(8726,36,'cp 57','h2h3 e8g8 f1b5 c6b8 b1d2 b7b6 b2b4 c7c6 b5e2 c6c5 a2a4 b8c6 b4b5 c6a5 f3e5 c8b7 h3h4 c5d4 c3d4 f6e4 d2e4 d5e4 e1g1 b7d5 h4h5 g6g5 f4g3 a8c8 a1c1 g7e5 g3e5 c8c1 d1c1 f7f6 e5g3 d8d7 c1a3 f8c8 f1c1 c8c1 a3c1 a5b3 c1c3 e7e6');
INSERT INTO evaluations VALUES(6014,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(6014,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(6014,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(6014,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(6014,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(6014,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(6014,36,'cp -13','a7a6 c1e3 c6d4 d1d2 g8f6 c3d1 e8g8 e1g1 f8e8 c2c3 d4e2 d2e2 c5e3 d1e3 d6d5 e2c2 b7b5 a2a4 c8b7 f3d2 c7c5 a4b5 a6b5 a1a8 d8a8 e4d5 f6d5 e3d5 b7d5 f2f3 f7f6 c3c4 b5c4 d3c4 d5f7 d2e4 a8c6 f1e1 e8a8 h2h3 f7g6 c2f2');
INSERT INTO evaluations VALUES(6164,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(6164,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(6164,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(6164,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(6164,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(6164,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(6164,36,'cp 81','f1d3 g8e7 d1e2 c6b4 d3c4 d7d5 c4b5 c7c6 c2c3 d8a5 a2a3 a5b5 a3b4 b7b6 e2b5 c6b5 b1a3 c8d7 a3c2 a7a5 c1g5 e7c6 g5f6 e8g8 e1e2 f8c8 f6g7 g8g7 e2d2 a5b4 c2b4 a8a4 b4c6 d7c6 h1c1 c8a8 a1a4 b5a4');
INSERT INTO evaluations VALUES(8272,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(8272,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(8272,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8272,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(8272,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(8272,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(8272,36,'cp -6','h2h3 e8g8 c1e3 e7e5 d1d2 c6d4 e1g1 c8d7 a1b1 a7a5 b2b3 f6e8 c3d5 d4e2 d2e2 e8c7 d5b6 a8a6 b6d7 d8d7 a2a4 a6a8 g1h1 b7b6 g2g4 c7e6 f3g5 a8e8 g5e6 f7e6 f2f3 f8f7 f1f2 e8f8 h1g2 d6d5 b1f1');
INSERT INTO evaluations VALUES(8269,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(8269,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(8269,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(8269,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(8269,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(8269,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(8269,36,'cp -11','c8d7 c1e3 e7e5 e1g1 e8g8 a2a3 a7a5 e3d2 c6d4 f3d4 c5d4 c3d5 f6d5 c4d5 f7f5 a1c1 f8f7 e2f3 h7h5 e4f5 d7f5 d1e2 d8h4 b2b4 a5b4 d2b4 g7f8 f3e4 a8a6 c1c2 f8h6 a3a4 f5e4 e2e4 h4e4 d3e4 a6a4');
INSERT INTO evaluations VALUES(9225,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(9225,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(9225,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(9225,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(9225,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(9225,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(9225,36,'cp 58','c4c5 d6e7 a2a3 f6e4 d1c2 e4c3 c2c3 a7a5 c1d2 e8g8 f1d3 a5a4 c3c2 f7f5 h2h3 c8d7 b2b4 a4b3 c2b3 b7b6 c5b6 c7b6 b3b2 c6a5 f3e5 a8c8 a1c1 c8c1 d2c1 d8c7 e5d7 c7d7 b2b6 f8c8 e1e2 a5c4 b6a6');
INSERT INTO evaluations VALUES(10492,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(10492,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(10492,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(10492,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(10492,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(10492,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(10492,36,'cp -94','d7d5 e4d5 d8d5 g1f3 c8g4 h2h3 g4f3 d1f3 d5f3 g2f3 e8c8 c1g5 c6e5 e1c1 e5d3 d1d3 d8d3 c2d3 f6d7 c3d5 b4d6 d3d4 d7b6 d5c3 h8e8 c1c2 g7g6 c2d3 c8d7 h3h4 h7h5 h1c1 c7c6 c1g1 e8e6 a2a3 e6e8 f3f4 b6d5 c3d5 c6d5');
INSERT INTO evaluations VALUES(11024,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(11024,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(11024,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(11024,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(11024,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(11024,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(11024,36,'cp 113','f3g5 d7d5 e4d5 c6a5 d1h5 g7g6 h5e2 e8g8 g5e4 e7f5 e4c5 f5d4 e2d3 b7b6 c5e4 f7f5 e4g3 a5c4 d3c4 d4c2 e1d1 c2a1 b2b3 a7a5 d5d6 f8f7 c4d5 a8b8 d5e5 d8d6 e5d6 c7d6 g3e2 c8b7 f2f3 b6b5 c1b2 b5b4 c3a4 a1b3 a2b3 f7e7 e2f4 b7c6 d2d3 c6b7 b2d4');
INSERT INTO evaluations VALUES(19656,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(19656,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(19656,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19656,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(19656,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(19656,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(19656,36,'cp -16','f8g7 e1g1 e7e6 d2d4 c5d4 c3b5 a7a6 b5d4 c6d4 f3d4 g8f6 e4e5 d6e5 f4e5 f6d5 d1f3 e8g8 c4d5 d8d5 f3d5 e6d5 c1f4 c8d7 a1e1 h7h6 h2h3 g6g5 f4h2 f7f5 e5f6 f8f6 f1f6 g7f6 c2c3 f6d4 c3d4 a8c8 e1e7 c8c1 g1f2 d7c6 e7e1 c1c4 f2e3 c6b5 e1e2 c4a4 b2b3 a4b4 h2d6 b5e2 d6b4 e2b5 e3f2');
INSERT INTO evaluations VALUES(19735,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(19735,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(19735,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(19735,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(19735,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(19735,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(19735,36,'cp 68','b1c3 g8f6 e4d5 f6d5 e1g1 d5c3 b2c3 e7e6 b5c6 b7c6 d2d3 f8d6 f3d2 e8g8 f1e1 e6e5 c3c4 c8f5 d1f3 d8d7 c1a3 f8e8 d2b3 f5g6 a3c5 d6c7 a2a4 f7f5 a1d1 h7h6 a4a5 e8e6 d3d4 e5d4 c5d4 e6e1 d1e1 a8e8 f3c3 e8e1 c3e1 f5f4 e1e2 f4f3 g2f3 d7h3');
INSERT INTO evaluations VALUES(20047,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(20047,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(20047,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(20047,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(20047,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(20047,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(20047,36,'cp 41','d2d4 c5d4 f3d4 c8d7 c2c4 g7g6 b2b3 f8g7 c1b2 g8f6 e1g1 h7h5 d4f3 a8c8 e2e3 e8g8 d1e2 d6d5 c4d5 f6d5 b2g7 g8g7 e2b2 d5f6 f1d1 d8c7 b1c3 d7e6 c3e4 f8d8 e4g5 d8d1 a1d1 e6g4 h2h3 g4f3 g5f3 c6b4');
INSERT INTO evaluations VALUES(22346,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(22346,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(22346,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(22346,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(22346,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(22346,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(22346,36,'cp 204','g8f6 c2c4 c8e6 c4d5 f6d5 b1c3 d5c3 d1d8 a8d8 e5c3 h8g8 f1e2 f8e7 f2f3 e7f6 a1c1 e6a2 c1a1 a2d5 e3e4 d5c6 e2c4 g8f8 a1a7 f6c3 b2c3 f8f6 e1e2 f6d6 e2e3 e8e7 h2h4 d6d2 a7a2 d2a2 c4a2 d8a8 a2c4 c6e8 c4d3 b7b5 h1d1 a8a2 g2g3');
INSERT INTO evaluations VALUES(25535,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(25535,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(25535,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(25535,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(25535,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(25535,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(25535,36,'cp 94','f3d4 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 d8d7 h2h3 a7a5 a2a4 d7e6 c1d2 d6d5 a1e1 d5e4 c3e4 e6d5 d2c3 f6e4 e1e4 f8e8 f1e1 e7c5 g1f1 e8e4 f3e4 c5d6 e1e2 f7f6 b2b3 h7h5 c3e1 g8f7 f4f5 d6c5 c2c4 d5e4 e2e4 c5f8 g2g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(29155,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(29155,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(29155,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(29155,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(29155,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(29155,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(29155,36,'cp 237','g2g4 h5g6 f3e5 f6d7 e5g6 h7g6 c3d5 e7e6 d5e3 b8c6 f1g2 f8d6 c2c3 a8c8 d4d5 e6d5 e3d5 e8g8 e1g1 f8e8 f1e1 e8e1 d1e1 d7b6 d5b6 d8b6 e1e2 b6c7 e2c4 c7b6 c4b3 b6c7 c1g5 d6e7 g5e3 c6a5 b3a4 c8d8 e3f4 e7d6 f4g5 d8d7 a1e1 a5c4 g2d5 c4b2 e1e8 g8h7');
INSERT INTO evaluations VALUES(29153,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(29153,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(29153,32,'cp 247',NULL);
INSERT INTO evaluations VALUES(29153,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(29153,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(29153,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(29153,36,'cp 228','h5g6 f3e5 f6d7 e5g6 h7g6 c3d5 b8c6 d5e3 e7e6 f1g2 f8d6 d1d3 d8c7 c1d2 d6f4 d4d5 e6d5 e1c1 d5d4 e3d5 f4d2 d3d2 c7d6 h1e1 e8f8 d5c3 d6b4 c3e2 b4d2 d1d2 a8e8 e1f1 d4d3 d2d3 d7c5 d3c3 c5a4 c3e3 e8e3 f2e3 a4c5 g2c6 b7c6');
INSERT INTO evaluations VALUES(29190,30,'cp 231',NULL);
INSERT INTO evaluations VALUES(29190,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(29190,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(29190,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(29190,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(29190,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(29190,36,'cp 213','f3e5 f6d7 e5g6 h7g6 c3d5 e7e6 d5e3 f8d6 c2c3 b8c6 f1g2 a8c8 e3c4 d6e7 c1f4 e8g8 d1e2 d7f6 c4e3 e7d6 f4g5 d8a5 g5f6 g7f6 e3c4 a5c7 c4d6 c7d6 h3h4 f8d8 g4g5 d6f4 g5f6 f4f6 g2e4 f6f4 h4h5 g6g5 e2e3');
INSERT INTO evaluations VALUES(34652,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(34652,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(34652,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(34652,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(34652,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(34652,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(34652,36,'cp 63','f1d3 g8f6 g1f3 d4d5 e4c3 d5d8 e1g1 c8g4 h2h3 g4h5 d3e2 b8c6 d2d4 e7e6 f1e1 f8e7 c3b1 e8g8 c2c3 a7a5 b1d2 e7d6 d2c4 h7h6 c4d6 c7d6 a2a4 c6e7 c1f4 a8c8 e2b5 d8c7 c3c4 h5f3 d1f3 e7f5 a1d1 c7b6 d4d5 e6e5 f4h2');
INSERT INTO evaluations VALUES(35458,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(35458,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(35458,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(35458,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(35458,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(35458,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(35458,36,'cp 22','b8d7 f1c4 f7f6 c1d2 c7c6 a2a4 a7a5 f2f4 d8c7 e1c1 g8h6 h2h3 f8d6 f4f5 h8f8 d2h6 g7h6 g2g4 d7b6 c4b3 d6b4 g1e2 c8d7 h3h4 d7e8 h1h3 e8f7 g4g5 f7b3 c2b3 h6g5 h4g5 f8h8 d1h1 f6g5 h3h7 h8h7 h1h7 b6d7 h7g7 a8h8 g7g5 h8h2 g5g1 h2h3 c1c2 h3h2');
INSERT INTO evaluations VALUES(38129,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(38129,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(38129,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(38129,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(38129,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(38129,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(38129,36,'cp -26','b8c6 f1d3 g8e7 e3e4 e8g8 e4d5 e7d5 c3d5 e6d5 e1g1 c8g4 h2h3 g4h5 c2c3 f8e8 g2g4 h5g6 d3g6 h7g6 c1f4 d8f6 f4g3 d5d4 g3h4 f6d6 b2b4 c5b6 d1d3 c6e5 f3e5 d6e5 c3d4 b6d4 a1d1 d4b6 d3f3 e5b5 a2a4 b5b4 d1d7 b4c4 f3b7');
INSERT INTO evaluations VALUES(5564,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(5564,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(5564,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(5564,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(5564,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(5564,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(5564,36,'cp -27','f1d3 g8e7 e1g1 e8g8 e3e4 d5d4 c3e2 e7g6 e2f4 g6e5 f3e5 c6e5 d3e2 b7b6 f4d3 e5g6 d3c5 b6c5 b2b3 e6e5 f2f4 e5f4 c1f4 g6f4 f1f4 d8c7 d1f1 c8e6 f1f2 f8e8 e2d3 e8e7 f2h4 h7h6 a1f1 a8e8 h4f2 f7f6 g1h1 g8h8 a2a4 c7e5 f2g3');
INSERT INTO evaluations VALUES(38193,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38193,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38193,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38193,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38193,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38193,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38193,36,'cp 0','g8f6 e1g1 e8g8 b1c3 h7h6 c1f4 b8c6 d1d2 d5d4 f4h6 d4c3 d2g5 f6e8 g5h5 g7h6 h5h6 e8f6 f3g5 c3b2 a1b1 d8d4 d3h7 g8h8 h7e4 h8g8 e4d3 d4c3 d3h7 g8h8 h7e4 h8g8 e4h7');
INSERT INTO evaluations VALUES(39053,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(39053,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(39053,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(39053,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(39053,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(39053,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(39053,36,'cp 12','f1g2 d7d6 d2d3 g8e7 a1b1 e8g8 e2e4 c7c5 b2b4 b8c6 b4b5 c6d4 g1f3 a7a6 a2a4 c8d7 c1d2 h7h6 h2h3 a6b5 f3d4 b5b4 d4b5 b4c3 b5c3 d7c6 h3h4 f7f5 e1g1 f5f4 g2h3 g8h7 d1g4 b7b6 g3f4 e5f4 c3d5 e7d5');
INSERT INTO evaluations VALUES(41016,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(41016,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(41016,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(41016,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(41016,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(41016,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(41016,36,'cp -67','f6e4 c3e4 e7g5 d1h5 g5f4 g1e2 f4d6 e1c1 e8g8 c1b1 b8d7 c4d3 f7f5 e4d6 d7f6 h5f3 c7d6 e2f4 c8d7 d3c2 d8b6 h1e1 f8e8 f3b3 b6b3 c2b3 d6d5 f4d3 d7c6 f2f3 g8f7 e1e3 c6b5 d3e5 f7f8 e3c3 a8c8 d1c1 c8c3 c1c3 f6d7 c3c7 d7e5 d4e5');
INSERT INTO evaluations VALUES(41416,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(41416,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(41416,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(41416,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41416,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(41416,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(41416,36,'cp 35','c2c4 e8g8 d2d4 c8f5 b2b3 c7c6 c1b2 d6d5 b1c3 f6e4 a1c1 e4c3 b2c3 f5e4 f3e5 e4g2 g1g2 e7e6 d1d2 f8e8 h2h4 b8d7 e5d7 d8d7 h4h5 g6h5 d2f4 d5c4 b3c4 a8c8 c4c5 f7f5 f1d1 h7h6 c1b1 d7d5 f2f3');
INSERT INTO evaluations VALUES(41262,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41262,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(41262,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(41262,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(41262,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41262,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(41262,36,'cp 33','e8g8 d2d4 c8f5 b2b3 c7c6 c1b2 a7a5 b1c3 f6e4 f3h4 e4c3 b2c3 f5c8 d1d2 a5a4 a1d1 b8d7 e2e4 a4b3 a2b3 d8b6 d1b1 e7e5 h4f3 e5d4 f3d4 d7c5 b3b4 c5d7 f1d1 b6c7 h2h3 d7b6 d2d3 a8a3 b1c1 c8d7 d4e2');
INSERT INTO evaluations VALUES(41859,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41859,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41859,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(41859,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(41859,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(41859,36,'cp 0','c7c6 g1f3 e8g8 a2a4 f8e8 c4a2 a7a5 e1g1 h7h6 c1e3 b8a6 d1d2 c8e6 a2e6 e8e6 e3c5 a6c5 d2e3 e6e8 f1e1 d8b6 b2b3 a8d8 a1d1 b6c7 f3h4 c5e6 h4f5 d8d7 d3d4 d6d5 e4d5 e5d4 f5d4 f6d5 c3d5 d7d5');
INSERT INTO evaluations VALUES(43658,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43658,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(43658,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43658,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43658,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(43658,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(43658,36,'cp 11','c7c5 c2c3 d8b6 d1c2 f8d6 b1d2 e8g8 d4c5 b6c5 e1g1 b8d7 b2b4 c5c7 f4d6 c7d6 e3e4 d5e4 d2e4 f6e4 d3e4 d7f6 f1e1 d6c7 c3c4 f6e4 c2e4 h7h6 a2a4 f8d8 b4b5 f7f6 a1d1 c8d7 f3d4 c7c4 e4b7 c4a4 d4c6 d7c6 d1d8 a8d8');
INSERT INTO evaluations VALUES(43567,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(43567,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43567,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43567,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(43567,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43567,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(43567,36,'cp 13','c2c3 d8b6 d1c2 f8d6 d4c5 b6c5 e1g1 b7b6 f1d1 b8d7 f4d6 c5d6 c3c4 c8b7 c4d5 f6d5 b1a3 d7c5 d3f1 d6e7 a3c4 e8g8 f3e5 b6b5 c4a5 a8c8 a1c1 b7a8 e5d3 c5d3 c2c8 d3c1 d1c1 f8c8 c1c8 e7f8 c8f8 g8f8');
INSERT INTO evaluations VALUES(43566,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(43566,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(43566,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(43566,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(43566,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(43566,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43566,36,'cp 10','d8b6 d1c2 f8d6 d4c5 b6c5 e1g1 b8d7 f4d6 c5d6 f1d1 h7h6 c3c4 d5c4 d3c4 d6b8 b1d2 b7b5 c4e2 e8g8 a2a4 b5a4 c2a4 a6a5 b2b3 c8b7 a4b5 b7a6 b5b8 f8b8 e2a6 a8a6 d1c1 b8a8 c1c7 a5a4 a1a4 a6a4 b3a4');
INSERT INTO evaluations VALUES(47573,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(47573,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(47573,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(47573,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(47573,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(47573,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(47573,36,'cp -62','c7c6 g1e2 f8g7 e1g1 e8g8 d2d3 d6d5 e4d5 f6d5 b1c3 c8e6 c3e4 d5c7 f3g3 e6c4 d3c4 d8e7 c1g5 f7f6 g5e3 c7e6 g3h4 b8a6 a1d1 a8e8 c2c3 f6f5 h4e7 e8e7 e4g5 a6c7 g5e6 c7e6 f2f3 f8e8 b2b4 e8a8 a2a4');
INSERT INTO evaluations VALUES(50614,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(50614,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(50614,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(50614,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50614,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50614,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50614,36,'cp 13','d2d3 d7d5 g1f3 e7e6 g2g3 b7b5 f1g2 d5e4 d3e4 d8d1 e1d1 c8b7 a2a4 g8f6 e4e5 f6d5 d1e2 f8e7 a4b5 a6b5 a1a8 b7a8 c1e3 e8g8 b1d2 b8d7 h1a1 a8c6 f3e1 f8a8 a1a8 c6a8 b2b3 d5c3 e2d3 c3d5 g2f3 g7g5 c2c4 b5c4 b3c4 d5b4 d3e2 g5f4 e3f4 b4c6 f3c6');
INSERT INTO evaluations VALUES(52119,30,'cp 312',NULL);
INSERT INTO evaluations VALUES(52119,31,'cp 302',NULL);
INSERT INTO evaluations VALUES(52119,32,'cp 301',NULL);
INSERT INTO evaluations VALUES(52119,33,'cp 316',NULL);
INSERT INTO evaluations VALUES(52119,34,'cp 303',NULL);
INSERT INTO evaluations VALUES(52119,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(52119,36,'cp 325','d4e5 d7d6 g1f3 b8c6 f1b5 f7f6 e5f6 g8f6 e1g1 f8e7 e4e5 d6e5 f3e5 d8d1 f1d1 c8d7 e5d7 e8c8 c1f4 f6d7 b1c3 c6d4 b5d3 d7f6 c3b5 d4b5 d3b5 f6d5 f4g3 e7f6 b5e2 h7h5 e2d3 d5c7 c2c3 h8e8 d3f5 c8b8 d1d8 e8d8 a1e1 h5h4 g3f4 a7a6 f5g4');
INSERT INTO evaluations VALUES(52095,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(52095,31,'cp 310',NULL);
INSERT INTO evaluations VALUES(52095,32,'cp 305',NULL);
INSERT INTO evaluations VALUES(52095,33,'cp 303',NULL);
INSERT INTO evaluations VALUES(52095,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(52095,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(52095,36,'cp 308','d7d6 g1f3 b8c6 f1b5 f7f6 e5f6 g8f6 e1g1 f8e7 e4e5 d6e5 f3e5 e8g8 e5c6 b7c6 d1d8 f8d8 b5c6 a8b8 b1c3 c8f5 f1e1 e7f8 e1e2 c5c4 h2h3 h7h6 g2g4 f5g6 g1g2 f8b4 c1f4 b4c3 b2c3 b8b6 c6f3 b6b2 f4e5 b2c2 e2c2 g6c2 e5d4 d8e8 g2g1 c2e4');
INSERT INTO evaluations VALUES(52596,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(52596,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52596,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52596,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(52596,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52596,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52596,36,'cp 0','e7e6 d2d4 c5d4 d1d4 g8f6 b1c3 b8c6 d4d2 a7a6 a2a4 b7b6 g5f3 c8b7 e1g1 h7h6 d2e2 d8c7 f1d1 g6g5 c4b3 f8g7 c1e3 e8g8 f3d2 c6e5 e3d4 e5c6');
INSERT INTO evaluations VALUES(53472,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53472,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53472,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53472,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53472,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53472,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53472,36,'cp 0','d2d4 c5d4 d1d4 g8f6 b1c3 b8c6 d4d2 a7a6 a2a4 b7b6 b2b3 h7h6 g5f3 c8b7 c1b2 c6a5 c4d3 f8g7 e1g1 e8g8 a1b1 a8c8 f1d1 f8e8 h2h3 a5c6 d3c4 c6a5');
INSERT INTO evaluations VALUES(56781,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(56781,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(56781,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(56781,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(56781,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(56781,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(56781,36,'cp 32','d4e5 g8h6 d1h5 g7g6 h5h3 c7c5 b1c3 c8d7 g1f3 d8b6 f3g5 b8a6 c1d2 e8c8 e1c1 a6c7 c1b1 d7e8 g2g4 f8g7 h3g3 g7f8 h2h4 c8b8 d2c1 f8e7 f1g2 a7a5 h4h5 c5c4 h5g6 h7g6 b1a1 e8c6 h1h3 d5d4 c3e4 e7g5 e4g5 c6g2 g3g2 c7d5 c2c3 d4c3 b2c3');
INSERT INTO evaluations VALUES(57803,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(57803,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(57803,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(57803,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(57803,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(57803,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(57803,36,'cp 98','c1e3 b8a6 e1g1 e7e5 d4e5 d6e5 c4c5 d8d1 f1d1 a6b4 d1d2 b4c6 h2h3 c8e6 f3g5 a8d8 d2d8 c6d8 a1d1 a7a6 b2b4 h7h6 g5e6 d8e6 e2c4 e6d4 a2a4 c7c6 e3d4 e5d4 d1d4 f6g4 d4d6 g4e5 c4b3 g6g5 f2f3');
INSERT INTO evaluations VALUES(6723,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(6723,31,'cp 231',NULL);
INSERT INTO evaluations VALUES(6723,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(6723,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(6723,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(6723,35,'cp 235',NULL);
INSERT INTO evaluations VALUES(6723,36,'cp 237','a5b6 b4b5 c6e7 f3e5 d7d6 e5c4 d6d5 c4b6 a7b6 e4e5 e7g6 f1d3 g8e7 a2a4 c8f5 d3e2 d8d7 g2g4 f5e6 h2h4 f7f5 h4h5 g6f8 g4g5 e6f7 f2f4 f8e6 e1g1 e8g8 b1a3 g7g6 h5h6 a8a5 a3c2 f8a8 c2e3 c7c6 e3g2 c6c5 c1e3 c5c4');
INSERT INTO evaluations VALUES(6312,30,'cp 241',NULL);
INSERT INTO evaluations VALUES(6312,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(6312,32,'cp 259',NULL);
INSERT INTO evaluations VALUES(6312,33,'cp 235',NULL);
INSERT INTO evaluations VALUES(6312,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(6312,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(6312,36,'cp 237','b4b5 c6a5 f3e5 d7d6 e5f3 g8e7 f1d3 c7c5 b5c6 b7c6 e1g1 e8g8 f1e1 c6c5 h2h3 h7h6 b1d2 c5d4 c3d4 f8e8 d4d5 c8d7 c1b2 e7g6 a1c1 a8c8 b2c3 g6e5 f3e5 d6e5 d2f3 f7f6 f3d2 a5b7 d2c4 b6c5 a2a3 b7d6 c3b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(6739,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(6739,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(6739,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(6739,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(6739,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(6739,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(6739,36,'cp 29','d1b3 d8f6 c1g5 f6g6 b1c3 a5c3 b3c3 g8e7 g5e7 e8e7 e4e5 h8e8 a1e1 e7f8 c4d3 g6h6 e1e4 f7f5 e5f6 e8e4 d3e4 h6f6 c3c4 h7h6 f1e1 d7d5 c4d5 c8g4 h2h3 g4f3 e4f3 a8d8 d5c4 a7a6 f3g4 d8d6 e1e3 f6a1 g1h2 a1d4 c4e2 g7g6 e3e8 f8g7 e8c8 d4e5 e2e5 c6e5');
INSERT INTO evaluations VALUES(6736,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(6736,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(6736,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(6736,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(6736,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(6736,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(6736,36,'cp 28','d8f6 c1g5 f6g6 b1c3 a5c3 b3c3 g8e7 g5e7 e8e7 a1e1 h8e8 c4d3 g6h6 e4e5 e7f8 e1e4 f7f5 e5f6 e8e4 d3e4 h6f6 c3c4 h7h6 f1e1 d7d5 c4d5 c8g4 h2h3 g4f3 e4f3 a8d8 d5c4 f6d6 f3h5 g7g6 h5f3 d6b4 c4c1 b4d2 f3c6 d2c1 e1c1 b7c6');
INSERT INTO evaluations VALUES(6863,30,'cp 472',NULL);
INSERT INTO evaluations VALUES(6863,31,'cp 454',NULL);
INSERT INTO evaluations VALUES(6863,32,'cp 470',NULL);
INSERT INTO evaluations VALUES(6863,33,'cp 470',NULL);
INSERT INTO evaluations VALUES(6863,34,'cp 426',NULL);
INSERT INTO evaluations VALUES(6863,35,'cp 471',NULL);
INSERT INTO evaluations VALUES(6863,36,'cp 446','c4f7 e8f8 d2d3 c7c6 f7b3 a7a5 a2a4 d7d5 e4d5 c6d5 c1f4 g8f6 b1a3 c8g4 d1c1 f8g8 f4e5 h7h6 c1f4 d8d7 a3b5 g8h7 e5d4 c5d4 b5d4 f6h5 f4d2 d7d6 d4b5 d6f4 d2f4 h5f4 f1e1 h8e8 f2f3 g4f5 e1e8');
INSERT INTO evaluations VALUES(9626,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(9626,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(9626,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(9626,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(9626,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(9626,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(9626,36,'cp 52','e7e5 c1e3 e8g8 h2h3 e5d4 f3d4 a7a6 e1g1 f8e8 d1c2 d7c5 a1d1 c5d3 c2d3 b7b5 d4c6 d8d7 c6a5 d7e6 c4c5 b5b4 c3d5 f6d5 e4d5 e6f5 c5d6 f5d3 d1d3 c7d6 a5c4 g7f8 d3d2 c8d7 c4b6 a8d8 b6d7 d8d7 f1c1 a6a5 c1c6 b4b3 a2b3 d7b7 c6a6 b7b3 a6a5');
INSERT INTO evaluations VALUES(11870,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11870,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(11870,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(11870,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(11870,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(11870,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11870,36,'cp 21','a8b8 f1c4 f8e7 e1g1 e8g8 g5f4 c8e6 c4b3 f6d7 d1d2 e7f6 f4e3 a7a5 a1d1 d8e7 h2h3 c6c5 b3e6 e7e6 b2b3 f6c3 d2c3 a5a4 f1e1 f7f6 e3d2 d7e5 c3g3 f8e8 f2f4 e5c6 g1h2 a4b3 a2b3 e6f7 d2c3 f7g6 g3d3 b8a8');
INSERT INTO evaluations VALUES(14864,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(14864,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(14864,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(14864,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(14864,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(14864,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(14864,36,'cp 32','c1f4 f8d6 f4g5 d6e7 e2e3 h7h6 g5h4 a7a6 f1d3 e8g8 e1g1 c6b8 f3e5 c7c5 d4c5 e7c5 e3e4 c5e7 e4d5 f6d5 h4e7 d5e7 d1g4 b8d7 e5c4 d8c7 f1d1 d7c5 g4g3 c7g3 h2g3 b7b5 c4a5 c5d3 d1d3');
INSERT INTO evaluations VALUES(16000,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(16000,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(16000,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(16000,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(16000,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(16000,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(16000,36,'cp 51','e7e6 d2d4 f6e4 c3e4 d6d5 d1d3 d5c4 d3c4 d7f6 c1g5 f8e7 g5f6 g7f6 e1c1 d8d5 c4d5 e6d5 e4g3 h7h5 f3h4 e7f8 g3f5 a7a5 h1e1 c8e6 c2c3 c7c6 a2a3 e8d7 c1d2 h8g8 g2g3 g8g5 f5e3 g5g8 h4f5 a8e8 b2b4 g8g5 f5h4 b7b6 b4a5 b6a5 d1a1');
INSERT INTO evaluations VALUES(17483,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(17483,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17483,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(17483,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(17483,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(17483,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(17483,36,'cp 18','b7b5 c4e2 e7e6 a2a4 b5b4 c2c3 g8f6 e4e5 f6d5 d2d4 c5d4 c3d4 c8b7 b1d2 f8e7 d2e4 e8g8 f1e1 c6a5 c1g5 f7f6 e5f6 d5f6 e2d3 a8c8 a1c1 c8c1 d1c1 f6e4 g5e7 d8e7 d3e4 b7e4 e1e4 e7f7 h2h3 a5c6 c1e3 f7f5 f3e5 c6e5 d4e5 a6a5 e3d4');
INSERT INTO evaluations VALUES(18169,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(18169,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(18169,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18169,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18169,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(18169,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(18169,36,'cp 13','c4e2 e7e6 a2a4 b5b4 c2c3 g8f6 e4e5 f6d5 d2d4 c5d4 c3d4 c8b7 b1d2 f8e7 d2e4 e8g8 f1e1 c6a5 c1g5 f7f6 e5f6 d5f6 e4d2 d8b6 e2d3 b7d5 a1c1 b4b3 d3e4 e7b4 g5f6 f8f6 e4d5 e6d5 f3e5 b6d4');
INSERT INTO evaluations VALUES(17574,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(17574,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(17574,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17574,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(17574,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(17574,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(17574,36,'cp 44','d2d4 c5d4 f3d4 g8f6 e4e5 d6e5 d4c6 d8d1 c3d1 b7c6 f4e5 f6g4 c1f4 g7g5 f4g5 h8g8 g5f4 f8g7 d1c3 g7e5 f4e5 g4e5 e1c1 c8f5 g2g3 f7f6 f1e2 a6a5 d1d4 f5g4 e2c4 e7e6 b2b3 e8e7 h1f1 g4f5 f1f5 e6f5 c4g8 a8g8');
INSERT INTO evaluations VALUES(22260,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(22260,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(22260,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(22260,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(22260,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(22260,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(22260,36,'cp 200','d1h5 e8d7 d4e5 c6e5 f4e5 f6e5 h5e5 g8f6 d3f5 d7c6 e5c3 f8c5 b2b4 c8f5 g1f3 f6e4 f3e5 c6d6 b4c5 d6e6 c3d4 b7b6 e5c6 d8d7 d4e5 e6f7 c6d4 f5e6 c5c6 d7e7 f2f3 e4d6 b1c3 h8e8 e1g1 f7f8 e5f4 f8g8 c3b5 d6b5 d4b5 e8c8 f4e5 g8f8 f1e1 e6f7 e5e7 f8e7 g1f2 a7a6 b5d4 e7d6 h2h4 c8e8 a2a3');
INSERT INTO evaluations VALUES(23287,30,'cp -227',NULL);
INSERT INTO evaluations VALUES(23287,31,'cp -225',NULL);
INSERT INTO evaluations VALUES(23287,32,'cp -208',NULL);
INSERT INTO evaluations VALUES(23287,33,'cp -208',NULL);
INSERT INTO evaluations VALUES(23287,34,'cp -223',NULL);
INSERT INTO evaluations VALUES(23287,35,'cp -199',NULL);
INSERT INTO evaluations VALUES(23287,36,'cp -224','f5e4 c4g8 e4f3 g8d5 f3g2 d5g2 d7d6 d2d3 f8e7 e1g1 e8g8 f2f4 c8e6 c3d5 e7h4 c1d2 c6d4 c2c3 d4f5 d5e3 e5f4 e3f5 f8f5 d2f4 d8e7 f4g3 h4g3 h2g3 c7c6 d1a4 f5f1 a1f1 a7a5 f1e1 e7f6 g1h2 f6h6 h2g1 a8f8 a4a5 e6h3');
INSERT INTO evaluations VALUES(29819,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(29819,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(29819,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29819,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(29819,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(29819,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29819,36,'cp 38','h2h3 g4h5 d2d4 b8c6 f1b5 d5c3 b2c3 e7e6 e1g1 a7a6 b5d3 f8d6 a1b1 a8b8 f1e1 e8g8 d1e2 f8e8 c1d2 h5g6 d3g6 h7g6 e2d3 b7b6 f3g5 d8d7 h3h4 d6e7 d3a6 b8a8 a6e2 e7g5 d2g5 a8a2 h4h5 a2a5');
INSERT INTO evaluations VALUES(31806,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(31806,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(31806,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31806,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(31806,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31806,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(31806,36,'cp 24','g1f3 e7e6 c2c4 g8f6 e1g1 c7c6 d1c2 a7a5 d3g6 h7g6 b1d2 f8e7 e3e4 e8g8 f1e1 f8e8 e4e5 f6d7 a1b1 c6c5 c4d5 e6d5 e5e6 f7e6 c2g6 d7f8 g6g4 b8c6 d4c5 e7c5 d2b3 c5b6');
INSERT INTO evaluations VALUES(33730,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(33730,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(33730,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(33730,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(33730,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(33730,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(33730,36,'cp 122','c7c5 d4d5 e7e6 c1d2 e6d5 c3d5 a5d8 d5f6 d8f6 d2c3 f6h6 d1a4 b8c6 f3e5 f8e7 g2g3 e8g8 e5c6 h6c6 a4c6 b7c6 f1g2 a8b8 g2c6 c8e6 e1d2 b8b6 c6e4 a7a5 h1d1 e7f6 c3f6 g7f6 d2c3 f8b8 b2b3 a5a4 b3a4 b6b4 e4d5 b4a4 d5e6 f7e6');
INSERT INTO evaluations VALUES(37609,30,'cp 243',NULL);
INSERT INTO evaluations VALUES(37609,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(37609,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(37609,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(37609,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(37609,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(37609,36,'cp 236','h2h4 h7h6 c1f4 d7d6 d1d2 d8c7 e4e5 d6d5 f1d3 g8e7 e1g1 c7d8 c3d1 e7f5 g2g3 b7b6 d1e3 c6c5 c2c3 a7a5 e3f5 g6f5 f1d1 c8a6 d3a6 b8a6 b2b3 a5a4 f4e3 d8d7 c3c4 a8d8 a1c1 d7c6 c4d5 c6d5');
INSERT INTO evaluations VALUES(37928,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(37928,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(37928,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(37928,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(37928,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(37928,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(37928,36,'cp 23','f7f6 e5f6 g8f6 f1e2 e8g8 e1g1 c8f5 c1f4 b8d7 f1e1 f6h5 f4g5 g7f6 d1d2 f8e8 g5f6 e7f6 h2h3 h5g7 c3d1 f5e4 f3h2 h7h5 c2c4 a8c8 d1c3 d7b6 c4d5 c6d5 e2b5 b6c4 d2e2 e8e6 h2f3');
INSERT INTO evaluations VALUES(38928,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(38928,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(38928,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(38928,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(38928,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(38928,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(38928,36,'cp 27','g8f6 c1g5 e8g8 e2e3 h7h6 g5h4 c8f5 d1b3 b4c3 b3c3 c7c6 f3d2 b8d7 f1e2 f6e4 h4d8 e4c3 b2c3 f8d8 a2a4 c6c5 d2f1 d7b6 a4a5 b6c4 d4c5 a8c8 f1g3 f5e6 a5a6 b7a6 a1a6 c8c5 a6a7 d8b8 e2c4 c5c4 e1g1 c4c3 g3e2 c3c2 e2d4');
INSERT INTO evaluations VALUES(41195,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(41195,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41195,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(41195,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(41195,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(41195,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(41195,36,'cp 33','f6d7 e5d6 c7d6 d1d2 d7f6 g5h6 g7h6 d2h6 d8a5 e1c1 b8d7 f1d3 d7b6 g1e2 c8e6 h1e1 e8c8 a2a3 c8b8 h6d2 h8e8 f2f3 e6c4 c3e4 a5d2 d1d2 f6e4 f3e4 d8c8 e2c3 e7e5 d3e2 c4e2 e1e2 f7f6 d4e5 d6e5 d2d6 e8f8 c3b5 h7h5 h2h3');
INSERT INTO evaluations VALUES(41898,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(41898,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(41898,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41898,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41898,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41898,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41898,36,'cp -10','e8g8 e1g1 f8e8 c1g5 h7h6 g5h4 b8d7 d4b3 c5b6 a2a4 a7a5 g1h1 d7e5 h4f6 d8f6 c3d5 f6h4 d5b6 e5g4 h2h3 c7b6 h1g1 g4f6 d1d6 f6e4 d6d3 c8e6 c4e6 e8e6 c2c3 e6g6 b3d4 a8d8 a1e1 e4c5 d3f3 g6f6 f3e3 c5e6 e3e4 h4e4 e1e4 e6d4 e4d4 d8d4');
INSERT INTO evaluations VALUES(37005,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37005,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37005,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(37005,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(37005,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37005,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37005,36,'cp 0','e1g1 b8d7 g1h1 f8e8 f2f3 d6d5 c4d5 d7b6 c1g5 c7c6 d4c6 b7c6 d5c6 c8a6 f1e1 f6g4 d1d8 g4f2 h1g1 a8d8 c6e8 f2g4 g1h1 g4f2');
INSERT INTO evaluations VALUES(37004,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37004,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37004,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37004,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37004,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37004,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37004,36,'cp 0','b8d7 d4b3 c5b6 a2a4 a7a5 c1g5 f8e8 g1h1 h7h6 g5h4 d7e5 h4f6 d8f6 c3d5 f6h4 d5b6 e5g4 h2h3 c7b6 h1g1 g4e5 c4d5 c8h3 g2h3 h4h3 f2f3 h3g3 g1h1 g3h3');
INSERT INTO evaluations VALUES(42324,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(42324,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(42324,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(42324,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(42324,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(42324,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(42324,36,'cp 181','e4e5 f6g8 c1f4 g8e7 h2h4 d7d6 h4h5 h7h6 e5d6 c7d6 h5g6 e7g6 f4e3 b8c6 g1e2 h6h5 d1d2 d6d5 e1c1 d5c4 e2g3 h5h4 g3e4 f7f5 d4d5 f5e4 d5c6 d8d2 e3d2 e4e3 d2e3 b7c6 c3e4 g7h6 e4g5 a8b8 f1c4');
INSERT INTO evaluations VALUES(42314,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(42314,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(42314,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(42314,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(42314,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(42314,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(42314,36,'cp 163','f6g8 c1f4 g8e7 h2h4 b8c6 h4h5 h7h6 g1f3 d7d6 e5d6 c7d6 h5g6 e7g6 f4e3 g6e7 d1d2 e7f5 e1c1 f5e3 f2e3 d8e7 d2f2 c8d7 f2g3 h8g8 c3b5 e8f8 g3d6 a8d8 d6c7 e6e5 d4d5 d8c8 c7d6 c6d8 a2a4 d7b5 d6e7 f8e7 a4b5');
INSERT INTO evaluations VALUES(44657,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44657,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(44657,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(44657,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44657,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44657,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44657,36,'cp 15','d5b6 d2d3 d7d5 e5d6 e7d6 c4b5 c8d7 b5d7 d8d7 b1c3 b8c6 a2a4 f8e7 e1g1 a8d8 a4a5 b6a8 c3d5 a8c7 d5e7 c6e7 d3d4 e8g8 f1e1 f8e8 d1d3 h7h6 c1e3 e7d5 e3d2 e8e1 a1e1 b7b5 a5b6 a7b6 c2c4 d5f6 h2h3 d6d5 d4c5 b6c5 c4d5');
INSERT INTO evaluations VALUES(44734,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(44734,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(44734,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(44734,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(44734,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(44734,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44734,36,'cp 12','f8g7 e1g1 e8g8 f1e1 b8c6 d2d4 c5d4 f3d4 c6d4 d1d4 c8d7 d4d3 a7a6 a2a4 a8c8 c1g5 d8a5 g5f6 g7f6 c3d5 f6b2 d5e7 g8g7 e7c8 b2a1 e1a1 f8c8 a1b1 a5a4 c4b3 a4b4 b3a2 b4a4 b1b7 a4a2 d3d4 g7g8 b7d7 a2c2 d7d6 c2c7 d6a6 c7c4 d4c4 c8c4 a6a7 c4e4');
INSERT INTO evaluations VALUES(40326,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40326,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40326,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40326,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(40326,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40326,36,'cp 12','e1g1 e8g8 f1e1 a7a6 a2a4 b8c6 d2d4 c6d4 f3d4 c5d4 d1d4 d8c7 c1e3 f6g4 c3d5 c7d8 d4b6 g4e3 e1e3 c8d7 b6d8 f8d8 d5e7 g8f8 e3f3 d7e8 e7d5 b7b5 c4b3 g7b2 a1b1 b5a4 b1b2 a4b3 f3b3 a6a5 d5b6 a5a4 b6a8 a4b3 a8b6 b3c2 b2c2 e8b5 c2c7');
INSERT INTO evaluations VALUES(40325,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40325,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(40325,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40325,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40325,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(40325,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(40325,36,'cp 0','e8g8 f1e1 b8c6 d2d4 c6d4 f3d4 c5d4 d1d4 c8d7 d4d3 d7c6 c1g5 h7h6 g5f4 f6d7 d3e2 g8h7 a1d1 a7a6 c4b3 a6a5 a2a4 d8b6 f4e3 b6b4 e3d2 b4b6 d2e3');
INSERT INTO evaluations VALUES(35928,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(35928,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35928,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35928,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35928,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(35928,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(35928,36,'cp 17','d2d4 c5d4 f3d4 a7a6 c4b3 b8c6 d4c6 b7c6 c1g5 a6a5 f1e1 f6d7 d1d2 d7c5 c3a4 c5a4 b3a4 a8b8 a1d1 c8a6 e4e5 g7e5 f2f4 e5f6 g5f6 e7f6 d2d6 d8d6 d1d6 b8b2 d6c6 a6b7 c6f6 b7d5 f6a6 b2a2 a6a5 g8g7 g1h2 f8c8 a5d5 a2a4');
INSERT INTO evaluations VALUES(46991,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(46991,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(46991,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(46991,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(46991,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(46991,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(46991,36,'cp 32','c7c6 g1f3 e7e5 b1c3 f8e7 e1g1 e8g8 d1c2 c8g4 f1d1 c6c5 h2h3 g4e6 f3g5 b8c6 g5e6 f7e6 d2d3 g8h8 e2e3 d6d5 c2e2 a7a6 b2b3 a8b8 d1f1 d8d7 c1a3 b7b5 a3b2 c6a5 a1e1 a5c6 c3b1 e7d6 b1d2 b8e8 b2c3 d6e7 e1d1');
INSERT INTO evaluations VALUES(49463,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49463,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49463,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(49463,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(49463,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49463,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49463,36,'cp 0','f8e7 e2e3 e8g8 d2d4 b7b6 c2c4 c8b7 f1e2 d7d5 e1g1 c7c5 d4c5 b6c5 b1c3 b8d7 a1c1 a8c8 c4d5 e6d5 e2b5 a7a6 b5d3 f8e8 c3e2 e7d6 d3f5 c8c7 f5d3 c7c8');
INSERT INTO evaluations VALUES(51197,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51197,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(51197,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(51197,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(51197,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(51197,36,'cp -14','c4b3 c8b7 d2d3 g8f6 e1g1 d7d6 f1e1 b8d7 h2h3 d8c7 c1g5 h7h6 g5f4 f8e7 b1d2 e8g8 f4g3 f8e8 f3h2 e7f8 h2f1 a8c8 a2a3 b7c6 g1h1 g7g6 f1e3 f8g7 g3h2 c7b7 b3c2 f6h5 a3a4 d6d5 a4b5 a6b5');
INSERT INTO evaluations VALUES(51214,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(51214,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(51214,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(51214,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(51214,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(51214,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(51214,36,'cp -13','c8b7 d2d3 g8f6 e1g1 d7d6 f1e1 b8d7 h2h3 d8c7 b3c2 f8e7 a2a4 e8g8 b1d2 f8c8 d2f1 e7f8 c1d2 g7g6 d1c1 d7e5 f1h2 e5f3 h2f3 e6e5 d2h6 f8g7 c1e3 c8e8 a4b5 a6b5 c2b3 c5c4 d3c4 a8a1 e1a1 b5c4 h6g7 g8g7');
INSERT INTO evaluations VALUES(30332,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(30332,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(30332,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(30332,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(30332,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(30332,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(30332,36,'cp -11','d2d3 d7d6 e1g1 g8f6 h2h3 d8c7 f1e1 f8e7 c1f4 b8d7 b1d2 e8g8 d1e2 f8c8 a2a4 e7f8 f4g3 c8e8 b3d1 h7h6 d3d4 c5d4 c3d4 e8c8 a4a5 c8e8 e4e5 d6e5 d4e5 f6d5 d2e4 f8b4 e4d6 b4e1 e2e1 c7c6');
INSERT INTO evaluations VALUES(53566,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53566,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(53566,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(53566,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53566,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(53566,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(53566,36,'cp 4','d2d4 d5c4 g1f3 b7b5 f3e5 c8b7 e1g1 g8f6 a2a4 a7a6 b2b3 c4b3 a4b5 a6b5 a1a8 b7a8 c1b2 f8e7 b1d2 e8g8 d1a1 b8d7 e5c6 a8c6 g2c6 d8b6 c6f3 f8c8 d2b3 e7f8 g1g2 b5b4 a1a4 c8c7 f1c1 f6d5 a4a8 c7c1 b3c1 d7f6 c1d3 d5c3 b2c3 b4c3');
INSERT INTO evaluations VALUES(53581,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(53581,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53581,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53581,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(53581,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53581,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53581,36,'cp 0','d5c4 g1f3 b7b5 a2a4 c8b7 e1g1 g8f6 f3e5 a7a6 b2b3 c4b3 c1b2 b8d7 d1b3 d7e5 d4e5 f6d5 a4b5 a6b5 a1a8 b7a8 b1c3 d5c3 b3c3 d8b6 f1c1 a8b7 g2c6 b6c6 c3c6 b7c6 c1c6 e8d7 c6b6 b5b4 b2d4 d7c7 b6a6 c7b7 a6b6');
INSERT INTO evaluations VALUES(54232,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(54232,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(54232,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(54232,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(54232,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(54232,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(54232,36,'cp -176','d5e4 c1e3 g8f6 d4c5 b8d7 b2b4 a7a5 f3e4 a5b4 c3b4 b7b6 f1b5 b6c5 b5c6 c8a6 c6a8 d8a8 b4b5 a6b5 b1c3 b5c4 g1f3 f6e4 d1c2 e4c3 c2c3 a8e4 a1c1 c4d5 e1g1 f8e7 f1d1 e7f6 c3d3 e4d3 d1d3 c5c4 d3d1 c4c3 a2a3 e8g8 g1h1');
INSERT INTO evaluations VALUES(57583,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(57583,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(57583,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(57583,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(57583,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(57583,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(57583,36,'cp 36','e7e6 c1f4 g8f6 g1f3 b7b6 e2e3 f8e7 h2h3 c8b7 f1e2 e8g8 e1g1 g8h8 c3c4 d7d6 b1c3 f6e4 c3e4 b7e4 f3d2 e4b7 e2f3 b7f3 d2f3 b8d7 b2b4 d8e8 f4h2 a8c8 c4c5 d6c5 b4c5 b6c5 a1c1 e7d6 d4c5 d6c5');
INSERT INTO evaluations VALUES(881,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(881,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(881,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(881,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(881,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(881,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(881,36,'cp 64','c2c4 d5h5 f1e2 c8g4 c1e3 e8c8 b1d2 e7e5 d4d5 c6d4 f3d4 e5d4 e2g4 f6g4 e3d4 d8e8 e1f1 g4h2 f1g1 h5d1 a1d1 h2g4 f2f3 g4e5 c4c5 b7b6 c5c6 e5g6 d2e4 g6f4 e4c3 h7h5 g2g3 f4e2 g1g2 e2d4 d1d4 g7g6 d4e4 f8g7 e4e8 h8e8 g3g4 h5g4 h1h7 g7e5 h7f7 e5c3 b2c3 e8e2 g2g3 g4f3 g3f3 e2a2');
INSERT INTO evaluations VALUES(1102,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(1102,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(1102,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(1102,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(1102,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(1102,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(1102,36,'cp 37','c6a5 c4b3 c7c6 e1g1 d7d6 h2h3 a5b3 a2b3 a7a6 f3h2 e7d8 f2f4 e5f4 c1f4 f8e7 h2f3 e8g8 d1e1 d6d5 g1h1 h7h6 a1d1 c8e6 e4e5 f6h7 e1g3 g8h8 c3e2 h7g5 f4g5 e7g5 h1h2 a6a5 f3g5 d8g5 d1a1 g5g3 h2g3 c6c5 d3d4 b7b6 c2c3 f8e8');
INSERT INTO evaluations VALUES(1736,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1736,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1736,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(1736,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1736,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1736,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1736,36,'cp 0','f1b5 c8d7 c2c3 a7a6 b5d3 h7h5 c1e3 f7f6 b1d2 g8h6 h2h3 h6f7 d1c2 a6a5 a2a4 h5h4 f3e1 b7b6 d3e2 c6d8 e2g4 d8e6 e1d3 g7g5 f2f3 f7h6 g4h5 h6f7');
INSERT INTO evaluations VALUES(1695,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1695,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1695,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1695,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1695,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1695,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(1695,36,'cp 0','c8d7 c2c3 a7a6 b5d3 f7f6 c1e3 h7h5 b1d2 g8h6 h2h3 h6f7 d1c2 h5h4 a2a4 a6a5 f3e1 b7b6 d3e2 c6d8 e2g4 d8e6 e1d3 g7g5 d2c4 f7h6 g4h5 h6f7');
INSERT INTO evaluations VALUES(3179,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3179,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(3179,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(3179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3179,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(3179,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(3179,36,'cp -5','e1g1 e7e6 c4a2 b7b6 c1e3 h7h6 d3d4 c8b7 d1d3 c6e7 a1d1 d6d5 e4d5 f6d5 c3d5 e7d5 a2d5 d8d5 d4c5 d5d3 c2d3 g7b2 f3d2 b7a6 c5b6 a7b6 e3h6 f8d8 d2c4 b2d4 h6e3 d4e3 f2e3 a6c4 d3c4 d8d1 f1d1 a8a3 g1f2 a3c3 d1d6 c3c4 d6b6 c4c3 f2f3');
INSERT INTO evaluations VALUES(3178,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(3178,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(3178,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(3178,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(3178,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(3178,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(3178,36,'cp -11','e7e6 c4a2 b7b6 c1e3 h7h6 d3d4 c8b7 d1d3 c6e7 a1d1 d6d5 e4d5 f6d5 c3d5 e7d5 a2d5 d8d5 d4c5 d5d3 c2d3 g7b2 c5b6 a7b6 d1b1 b2g7 b1b6 b7d5 f3d2 a8a3 e3c5 a3d3 c5f8 g7f8 b6b2 f8g7 b2b8 g8h7 d2f3 d5f3 g2f3 d3f3 b8b7 f3f4');
INSERT INTO evaluations VALUES(3965,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(3965,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(3965,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(3965,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(3965,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(3965,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(3965,36,'cp 57','c4d5 e6d5 f1d3 c7c5 e1g1 b4c3 b2c3 c5c4 d3c2 d8a5 f3e5 f6e4 g5e7 d7e5 d4e5 f8e8 e7b4 a5c7 c2e4 d5e4 d1d6 c7d8 d6d8 e8d8 b4d6 f7f6 f1d1 f6e5 d6e5 d8d3 a2a4 b7b6 d1b1 c8d7 e5d4 a8c8 h2h4 h7h6 b1b4 d3d2 a4a5');
INSERT INTO evaluations VALUES(5758,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(5758,31,'cp 269',NULL);
INSERT INTO evaluations VALUES(5758,32,'cp 277',NULL);
INSERT INTO evaluations VALUES(5758,33,'cp 291',NULL);
INSERT INTO evaluations VALUES(5758,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(5758,35,'cp 272',NULL);
INSERT INTO evaluations VALUES(5758,36,'cp 278','e1g1 b4c3 b2c3 g8e7 f3h4 e8g8 d1f3 c8e6 c4d3 d6d5 f3g3 g8h8 e4e5 d8d7 c1a3 f7f5 f2f4 f8f7 a1b1 b7b6 d3b5 a7a6 b5d3 a8g8 g3f2 g7g6 h4f3 h8g7 a3e7 d7e7 d3a6 c6a5 a6b5 g8a8 a2a4 a5c4 b5c4 d5c4');
INSERT INTO evaluations VALUES(5757,30,'cp 266',NULL);
INSERT INTO evaluations VALUES(5757,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(5757,32,'cp 293',NULL);
INSERT INTO evaluations VALUES(5757,33,'cp 287',NULL);
INSERT INTO evaluations VALUES(5757,34,'cp 268',NULL);
INSERT INTO evaluations VALUES(5757,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(5757,36,'cp 280','b4c3 b2c3 g8e7 f3h4 e8g8 d1f3 c8e6 c4d3 e7g6 h4g6 f7g6 f3g3 c6e7 c3c4 a7a6 c1b2 e6d7 f2f4 c7c6 b2a3 d6d5 a1e1 d5e4 e1e4 e7f5 g3g6 f8f6 g6h5 d8a5 a3b2 a5a2 h5e2 f5d6 e4e7 d7f5 d4d5 f5d3 e7g7 g8h8 e2d3 h8g7 b2f6 g7f6 d5c6 d6c4 d3d4 f6f7 c6c7 f7e7 d4g7 e7d6 f1d1 d6c5');
INSERT INTO evaluations VALUES(6592,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(6592,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(6592,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(6592,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(6592,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(6592,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(6592,36,'cp 201','c2c3 b4c5 c1f4 d7d5 e5d6 c5d6 f4d6 c7d6 e2e3 g8f6 b1d2 d6d5 f1d3 c8g4 h2h3 g4h5 d3e2 e8g8 e1g1 f8e8 f3d4 h5e2 d4e2 c6e5 d1b3 a7a6 f1d1 d8c7 e2f4 a8d8 d2f1 e5c4 d1d4 b7b5 a1d1');
INSERT INTO evaluations VALUES(6591,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(6591,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(6591,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(6591,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(6591,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(6591,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(6591,36,'cp 200','b4c5 c1g5 g8e7 b1d2 d7d5 e5d6 d8d6 e2e3 e7g6 d1a4 f7f6 d2e4 d6e7 e4c5 e7c5 g5f4 g6f4 a4f4 c6e5 f1e2 c8d7 a2a3 e8c8 f4d4 e5f3 e2f3 c5a5 d4b4 a5b5 e1c1 b5b4 a3b4 c7c6 h2h4 d7e6 d1d4 e6b3 f3d1 b3e6 d4e4 h8e8');
INSERT INTO evaluations VALUES(6963,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(6963,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(6963,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(6963,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(6963,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(6963,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(6963,36,'cp 69','b1c3 g8f6 e1g1 e7d6 f1e1 a7a6 b5c4 d6c5 c3d5 d7d6 c2c3 c5a7 d2d4 h7h6 h2h3 e8g8 c1e3 f8e8 d1c2 f6d5 c4d5 e5d4 c3d4 c6e7 d5b3 d6d5 a1d1 c7c6 e4e5 c8e6 c2e2 e7f5 e3f4 d8e7 b3c2 c6c5 d4c5 e7c5 e2d2 a7b6 d1c1');
INSERT INTO evaluations VALUES(7895,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(7895,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(7895,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(7895,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(7895,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(7895,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(7895,36,'cp -33','g1f3 a7a6 a2a4 h7h6 g5h4 b7b6 e4e5 f6h7 h4g3 d7d5 e5d6 e7d6 c3e4 d6g3 h2g3 c8b7 d1e2 d8c7 e1g1 e8g8 c2c3 f8d8 e4d2 d8d7 e2e3 a8d8 e3f4 h7f6 f1e1 c7f4 g3f4 f6h5 g2g3 h5f6 f3e5 c6e5 f4e5');
INSERT INTO evaluations VALUES(10293,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(10293,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(10293,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(10293,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(10293,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(10293,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(10293,36,'cp -26','f6e4 c3d4 d7d5 d4c5 d5c4 d1e2 d8d3 f1e1 f7f5 b1d2 e8g8 d2e4 f5e4 e2e4 c8f5 e4h4 a8e8 e1e8 f8e8 c1e3 e8e4 h4g5 e4e7 h2h3 h7h6 g5h5 d3e4 a1c1 f5g6 h5g4 e4g4 h3g4 g6d3 a2a3 g8f7 g1h2 f7e6 f3d4 c6d4 e3d4 e6d5 d4c3 g7g6 h2g3 d5c6 c1e1 e7e1');
INSERT INTO evaluations VALUES(11694,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(11694,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(11694,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(11694,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(11694,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(11694,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(11694,36,'cp 61','c4b3 f8e7 c3d5 e8g8 e1g1 d7d6 c2c3 c8d7 h2h3 f6d5 b3d5 a8b8 b2b4 e7f6 c1e3 c6e7 d5b3 a6a5 a2a3 e7g6 d1d2 c7c6 f1c1 h7h6 b3a2 d7c8 d3d4 c8b7 a2b1 f8e8 c1e1 d6d5 e4d5');
INSERT INTO evaluations VALUES(11703,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(11703,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(11703,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(11703,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(11703,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(11703,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(11703,36,'cp 58','f8e7 c3d5 e8g8 e1g1 d7d6 c2c3 c8d7 h2h3 f6d5 b3d5 a8b8 c1e3 e7f6 b2b4 a6a5 a2a3 c6e7 d5b3 e7g6 d1d2 b8a8 a1b1 a5b4 a3b4 h7h6 b1a1 d7c6 a1a8 d8a8 d2b2 a8d8 f1a1 d6d5 b2c2 d8d7 b3d5 c6d5 e4d5 d7d5 c3c4 b5c4 c2c4 d5d7 a1a7 f8d8 a7c7 d7d3');
INSERT INTO evaluations VALUES(12133,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(12133,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(12133,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12133,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12133,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(12133,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(12133,36,'cp 32','f8c5 d1d2 c6d4 e3d4 a7a6 e1c1 c5d4 d2d4 d8b6 c3e2 b6d4 e2d4 d7c5 g2g4 c8d7 f1e2 f7f6 e5f6 g7f6 e2f3 e8c8 d4e2 d7c6 h2h4 e6e5 g4g5 e5f4 e2f4 f6g5 h4g5 c8b8 b2b3 d8f8 h1h4 c5e4 d1g1 h8g8');
INSERT INTO evaluations VALUES(7734,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(7734,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(7734,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(7734,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(7734,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(7734,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(7734,36,'cp 50','d1d2 c6d4 e3d4 a7a6 h2h4 b7b5 d4c5 d7c5 d2d4 d8e7 a2a3 c8b7 e1c1 a8c8 g2g4 c5e4 c3e4 d5e4 h1h3 c8c5 d4b4 b7d5 g4g5 e8g8 h3c3 f8c8 c3c5 e7c5 b4c5 c8c5 c2c3 e4e3 d1d4 h7h6 f1d3 h6g5 h4g5 g7g6 a3a4');
INSERT INTO evaluations VALUES(18790,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(18790,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18790,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(18790,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(18790,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(18790,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18790,36,'cp 12','d7d5 g1f3 f8d6 f4g3 d8e7 b1d2 g8f6 f1b5 e8g8 b5c6 b7c6 g3d6 e7d6 d1a4 f6d7 e1g1 c5d4 c3d4 c6c5 f1c1 c5d4 a4d4 f7f6 d2e4 d6e7 e4c5 d7c5 d4c5 e7c5 c1c5 f8d8 a1d1 h7h6 b2b3 a7a5 a2a4 e6e5 c5d5 d8d5 d1d5 c8b7');
INSERT INTO evaluations VALUES(18894,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(18894,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(18894,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(18894,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(18894,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(18894,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(18894,36,'cp 17','d4b5 d7d6 g2g3 a7a6 b5c3 c8d7 f1g2 d8c7 b2b3 g8f6 e1g1 f8e7 b1d2 c6e5 c1b2 h7h5 d2f3 d7c6 a1c1 c7b6 d1d2 a8d8 c1d1 e8g8 d2c2 f6d7 e2e4 e5f3 g2f3 h5h4 c2e2 d7e5 f3g2 h4g3 h2g3');
INSERT INTO evaluations VALUES(23303,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(23303,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(23303,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(23303,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(23303,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(23303,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(23303,36,'cp 41','e4f5 e5f4 d2d4 d7d5 f1d3 f8d6 b1c3 g8f6 d1e2 c6e7 e1g1 e8g8 f3g5 f8e8 c1f4 c8f5 f4e5 f5d3 e2d3 e7g6 e5f6 g7f6 g5h3 c7c6 c3e2 e8e6 e2g3 d6g3 d3g3 d8d6 g3d6 e6d6 a1e1 b7b6 h3f2 h7h5 f2d3 g8f7 f1f5 a8h8 a2a4 h5h4 d3e5 g6e5 d4e5 d6e6 e1f1 h8h6 f5f6 h6f6 f1f6 e6f6 e5f6 f7f6 g1f2 f6e6 f2f3');
INSERT INTO evaluations VALUES(23282,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(23282,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(23282,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(23282,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(23282,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(23282,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(23282,36,'cp 71','e5e4 f3e5 g8f6 b1c3 d7d5 f1b5 c8f5 d2d3 e4d3 c2d3 f5d7 d3d4 f8b4 b5c6 d7c6 e1g1 b4a5 f4f5 e8g8 c1f4 a5b6 d1d3 d8d6 a2a3 f6h5 f4e3 a8e8 a1e1 h5f6 g1h1 f6e4 e1c1 a7a6 g2g4 e4c3 d3c3 c6b5 f1f3 c7c5 d4c5 d6e5 c3e5');
INSERT INTO evaluations VALUES(23306,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(23306,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(23306,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(23306,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(23306,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(23306,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(23306,36,'cp 55','f3e5 g8f6 b1c3 d7d5 f1b5 c8f5 e5c6 b7c6 b5c6 e8f7 c6a8 d8a8 c3e2 f8c5 b2b4 c5b4 a1b1 b4a5 h2h3 h8e8 e2g3 f5c8 e1g1 c8a6 f1e1 d5d4 c2c3 a6d3 b1b3 d4c3 d2c3 a8c6 g1h2 a5c3 e1e3 f6d5 e3d3 e4d3 d1d3 h7h6 g3e4 c3b4 a2a3 b4e7 f4f5 c6c1');
INSERT INTO evaluations VALUES(26241,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(26241,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(26241,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(26241,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(26241,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(26241,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(26241,36,'cp -14','e7e6 h2h3 g4f3 d1f3 f8d6 e1g1 d8f6 f3f6 g8f6 c1e3 e8g8 c3a4 f8c8 c2c3 b7b6 b5d3 g7g6 b2b3 a8b8 a1c1 g8g7 f1d1 c6e7 c3c4 e7f5 d3f5 g6f5 c4c5 d6c7 c5b6 a7b6 a4c3 f5f4 e3d2 c7d6 g1f1 b6b5');
INSERT INTO evaluations VALUES(26377,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(26377,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(26377,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(26377,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(26377,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(26377,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(26377,36,'cp -17','h2h3 g4f3 d1f3 f8d6 e1g1 d8f6 f3f6 g8f6 c3a4 a7a6 b5c6 b7c6 f1e1 a8b8 a2a3 a6a5 c1d2 b8a8 f2f3 e8e7 b2b4 a5b4 a3b4 d6g3 e1d1 f6d7 g1f1 h8b8 d1b1 h7h5 a4c5 d7c5 a1a8 b8a8 b4c5 a8a2 b1b7 e7f6');
INSERT INTO evaluations VALUES(29508,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(29508,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(29508,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(29508,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(29508,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(29508,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(29508,36,'cp 21','d1e2 d8d5 g1f3 e7e6 e1g1 b8c6 f1d1 e8c8 c2c4 d5f5 b1c3 f8b4 c1e3 h8g8 c3b5 a7a6 a2a3 b4e7 b5c3 g7g5 a1b1 f6e4 f3d2 e4d2 d1d2 e7f6 b1d1 g5g4 b2b4 e6e5 d4d5 c6d4 e2d3 f5d3 d2d3 f6g7 b4b5 a6b5 c3b5 d4b5 c4b5 e5e4 d3b3 f7f5 g2g3 h7h5 b5b6');
INSERT INTO evaluations VALUES(29507,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(29507,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(29507,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(29507,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(29507,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(29507,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29507,36,'cp 21','d8d5 g1f3 e7e6 c2c4 f8b4 c1d2 b4d2 b1d2 d5d6 e1g1 e8g8 d2e4 f6e4 e2e4 h7h6 e4b7 b8c6 d4d5 e6d5 c4d5 c6b4 b7b5 b4d5 a1d1 c7c6 b5a4 f8e8 f1e1 e8e1 f3e1 a8b8 d1d2 d6b4 a4c2 d5f4 g2g3 f4e6 g1g2 c6c5 e1f3');
INSERT INTO evaluations VALUES(32619,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(32619,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(32619,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(32619,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(32619,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(32619,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(32619,36,'cp 30','g1f3 e7e6 c2c4 f8b4 c1d2 b4d2 b1d2 d5d6 e1g1 e8g8 d2e4 f6e4 e2e4 h7h6 e4b7 b8c6 d4d5 e6d5 c4d5 c6b4 b7b5 b4d5 f1d1 c7c6 b5c4 d6f6 d1d2 f8e8 a1d1 a8d8 h2h3 d8d6 g1h2 d6d7 a2a3 d5b6 c4c2 d7d2 d1d2 b6d5 c2c4 e8a8 c4d4 f6d4');
INSERT INTO evaluations VALUES(32616,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(32616,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(32616,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(32616,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(32616,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(32616,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(32616,36,'cp 25','e7e6 c2c4 f8b4 c1d2 b4d2 b1d2 d5d6 e1g1 e8g8 a1d1 d6b6 d2e4 f6e4 e2e4 b8d7 e4e2 c7c5 d1d2 c5d4 f3d4 f8d8 f1d1 d7f6 h2h3 a7a6 d4f3 d8d2 e2d2 a6a5 b2b3 h7h6 d2d6 b6d6 d1d6 a5a4 b3b4 f6e4 d6b6 a8c8 f3e5');
INSERT INTO evaluations VALUES(31212,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(31212,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(31212,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(31212,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(31212,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(31212,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31212,36,'cp 19','g8f6 e4f3 b8d7 f3e2 d8c7 g1f3 e7e5 e2c4 f8d6 d4e5 d6e5 e1g1 e8g8 f3e5 d7e5 c1f4 c7a5 c4b3 f8e8 d1d2 a5d2 f4d2 f6e4 d2a5 e4c5 f1e1 c5b3 a2b3 e5g6 g1f1 e8e1 a1e1 f7f6 e1d1 g6f8 h2h4 h7h5 f2f3 g8f7 a5b4 a7a5 b4f8 f7f8 d1d7');
INSERT INTO evaluations VALUES(29083,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(29083,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(29083,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(29083,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(29083,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(29083,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(29083,36,'cp 22','e4f3 b8d7 d1e2 d8a5 c2c3 h7h6 g2g3 e7e5 f3g2 f8d6 g1f3 e8g8 e1g1 f8e8 e2c2 e5d4 c3d4 d7b6 a2a3 a5b5 c1d2 f6e4 d2c3 a7a5 a3a4 b5c4 f1c1 e4c3 b2c3 e8e2 f3d2 c4e6 g2f1');
INSERT INTO evaluations VALUES(29082,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(29082,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29082,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29082,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(29082,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(29082,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(29082,36,'cp 33','b8d7 d1e2 d8a5 c2c3 h7h6 g2g3 e7e5 b2b4 a5c7 d4e5 c7e5 c1f4 e5e6 h2h4 f8e7 e1f1 f6d5 f3d5 e6d5 g1f3 e8c8 a1d1 d5e6 e2e6 f7e6 f1e2 e7f6 c3c4 h8e8 b4b5 e6e5 f4e3 c6b5 c4b5 e5e4 f3h2 d7e5 d1d8 f6d8 h1d1 d8b6 d1d2 c8c7 d2c2 c7d6 h4h5 d6e6 e3b6 a7b6');
INSERT INTO evaluations VALUES(36486,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36486,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36486,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36486,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(36486,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36486,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36486,36,'cp 12','h7h6 g5e3 f6g4 e3c1 e7e5 d4e5 g4e5 f3e5 d6e5 c1e3 b8c6 d1d2 g8h7 f2f3 c6d4 e1g1 c8e6 a1d1 b7b6 b2b3 d8d7 c3e2 c7c5 c2c3 d4e2 d2e2 f8d8 e3f2 d7e7 d3c4 a8c8 c4d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(36614,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36614,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36614,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36614,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36614,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36614,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36614,36,'cp 0','c7c5 b1c3 c5d4 e3d4 c8g4 h2h3 g4f3 d1f3 b8c6 c3e2 f8e8 f1d1 d6d5 f3e3 f6h5 f4g5 g7f6 g5h6 f6g7');
INSERT INTO evaluations VALUES(37638,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(37638,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(37638,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(37638,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(37638,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(37638,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(37638,36,'cp 192','h2h4 d7d6 h4h5 g6h5 h1h5 b8c6 b1c3 h7h6 f1e2 g8f6 h5h1 c8d7 e1f1 c6e7 a2a4 a8c8 a4a5 a7a6 f1g1 e7g6 d1b3 c8c7 c1d2 e8g8 h1h6 g7h6 d2h6 e6e5 h6f8 g8f8 g2g3 d7g4 a1d1 c7d7 b3a4 f8g7 d4e5 g6e5 f3e5 d6e5 d1d7 g4d7');
INSERT INTO evaluations VALUES(39165,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(39165,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(39165,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(39165,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(39165,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(39165,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(39165,36,'cp 135','c2c3 b4e7 d2d4 d7d6 f1d3 b8d7 e1g1 e6e5 a2a4 g8f6 a4a5 e8g8 f1e1 f8e8 h2h3 b7b6 a5a6 c7c6 c1e3 b6b5 b1d2 e7f8 d1c2 h7h6 e1d1 g8h8 b2b4 d7b6 d4e5 d6e5 c3c4 c8d7 c4c5');
INSERT INTO evaluations VALUES(39579,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(39579,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39579,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(39579,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(39579,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(39579,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(39579,36,'cp 0','d7d6 g1f3 g8f6 b1d2 f6h5 f4g5 h7h6 g5h4 e8g8 e2e3 c7c5 d4c5 d6c5 h2h3 c8d7 d2e4 d8b6 f3d2 f7f5 d2c4 b6c6 e4g3 h5g3 h4g3 d7e6 f1e2 b8d7 c4a5 c6b6 a5c4');
INSERT INTO evaluations VALUES(38391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38391,36,'cp 0','g1f3 g8f6 b1d2 f6h5 f4g5 h7h6 g5h4 e8g8 e2e3 c7c5 d4c5 d6c5 h2h3 c8d7 d2e4 d8b6 f3d2 f7f5 d2c4 b6c6 e4g3 h5g3 h4g3 d7e6 f1e2 b8d7 c4a5 c6b6 a5c4');
INSERT INTO evaluations VALUES(38390,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38390,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38390,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38390,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38390,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38390,36,'cp 0','g8f6 b1d2 f6h5 f4g5 h7h6 g5h4 e8g8 e2e3 c7c5 d4c5 d6c5 h2h3 c8d7 d2e4 d8b6 f3d2 f7f5 d2c4 b6c6 e4g3 h5g3 h4g3 d7e6 f1e2 b8d7 c4a5 c6b6 a5c4');
INSERT INTO evaluations VALUES(44426,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(44426,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(44426,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(44426,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(44426,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(44426,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(44426,36,'cp 176','e4e5 f6h5 d2d4 c5d4 d1d4 b8c6 d4a4 d7d6 g2g4 h5g7 c1e3 h7h5 e5d6 h5g4 f3e5 e7d6 e1c1 f8e7 f1g2 c8d7 e5d7 d8d7 c3d5 a8d8 d5e7 c6e7 a4a7 d6d5 a7d4 e8g8 c1b1 f8e8 h2h3 e7f5 d4g4 d7c6 e3g5 d8c8 c2c3 e8e5 h3h4 f5d6 g4f4');
INSERT INTO evaluations VALUES(45430,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(45430,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45430,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(45430,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45430,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45430,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45430,36,'cp 19','b8c6 b1d2 d8b6 d2b3 c8g4 h2h3 g4h5 a2a4 h5f3 d1f3 e7e6 f3d1 b6d8 c2c3 f8d6 f4d6 d8d6 f1d3 e8g8 e1g1 e6e5 d4e5 d6e5 f1e1 e5c7 d1d2 f8e8 d3c2 a7a6 a1d1 g7g6 b3d4 c7b6 e1e8 a8e8 d4c6 b7c6 b2b4 b6c7 d1e1 e8e1 d2e1');
INSERT INTO evaluations VALUES(13142,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(13142,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13142,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13142,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(13142,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(13142,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(13142,36,'cp 34','b1d2 a7a6 c2c3 c8f5 f3h4 f5g4 d1b3 c6a5 b3c2 e7e6 h2h3 g4h5 f1d3 f8d6 f4d6 d8d6 d2f3 a5c4 e1g1 e8g8 a2a4 h5f3 h4f3 f8c8 c2e2 d6c7 d3c4 d5c4 a4a5 a8b8 f1e1 b7b6 a5b6 c7b6 f3e5 b6b2 e2b2');
INSERT INTO evaluations VALUES(46983,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(46983,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(46983,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(46983,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(46983,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(46983,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46983,36,'cp 5','c7c5 g1e2 e7e5 g2g3 b8c6 f1g2 a8b8 d2d3 a7a6 a2a4 g7g6 e1g1 f8g7 h2h3 e8g8 c1g5 h7h6 g5e3 g8h7 a1b1 c8e6 d1d2 f6d7 b2b3 c6d4 g1h1 d8a5 f2f4 d4e2 c3e2 a5d2 e3d2 e5f4 g3f4 f7f5 b3b4 f5e4 g2e4 c5b4 b1b4');
INSERT INTO evaluations VALUES(47255,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(47255,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(47255,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(47255,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(47255,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(47255,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(47255,36,'cp -43','c7c5 f1g2 b8c6 g1f3 d8c7 e1g1 e7e5 b1c3 f8e7 e2e4 d5e4 f3g5 c8f5 g5e4 f6e4 c3e4 e8g8 d2d3 a8d8 e4d2 f8e8 f1e1 f7f6 d2c4 c7d7 c4e3 f5e6 g2e4 g7g6 a2a4 b7b6 h2h4 c6b4 d1e2 e7f8 e4g2 h7h5 a4a5 f8g7 a5b6 a7b6 e3c4');
INSERT INTO evaluations VALUES(49691,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(49691,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(49691,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(49691,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(49691,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(49691,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(49691,36,'cp -51','d7d5 d2d3 d5d4 c3e4 f6e4 g2e4 f8d6 e2e3 e8g8 e3d4 e5d4 g1e2 b8c6 e1g1 c8g4 c2c3 f8e8 f2f3 g4e6 c1f4 d6c5 c3c4 d8d7 g3g4 c6e7 e2g3 c7c6 d1d2 a7a5 d2f2 f7f5 g4f5 e7f5 g3f5 e6f5 a1e1 c5b4 e1e2');
INSERT INTO evaluations VALUES(51797,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(51797,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(51797,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(51797,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(51797,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(51797,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(51797,36,'cp 56','f1d3 d7d5 b1d2 a7a5 b2b3 b8d7 g1f3 f7f5 c2c4 g8h6 a2a4 f8d6 c1a3 d6a3 a1a3 h6f7 c4d5 c6d5 e1g1 e8g8 a3a1 g7g5 d3e2 g5g4 f3e1 d7f6 e1d3 c8d7 f2f3 h7h5 a1c1 a8c8 c1c8 d8c8 d1c1 c8c1 f1c1');
INSERT INTO evaluations VALUES(51953,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(51953,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(51953,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(51953,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(51953,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(51953,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(51953,36,'cp -36','c5d4 d1d4 g8f6 b1c3 f8b4 d4d3 d7d5 e4d5 e6d5 c1g5 d8e7 f1e2 d5c4 d3c4 b8c6 g1f3 c8e6 c4h4 e8c8 e1g1 h7h6 g5f6 e7f6 h4f6 g7f6 f1e1 c8b8 a2a3 b4e7 a1c1 f6f5 c3a4 d8d6 a4c5 e6d5 b2b4 h8d8 g1f1 e7f6 e1d1 d5f3 e2f3');
INSERT INTO evaluations VALUES(52054,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(52054,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(52054,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(52054,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(52054,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(52054,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(52054,36,'cp -17','d1d4 b8c6 d4d2 f8c5 b1c3 g8f6 f1e2 d8a5 g1f3 c5b4 e2d3 e8g8 e1g1 d7d6 f1d1 e6e5 f3e1 c6d4 h2h3 b7b6 e1c2 d4c2 d2c2 b4c3 c2c3 a5c3 b2c3 c8e6 a2a4 f8c8 a4a5 b6a5 a1a5 e6c4 d3c4 c8c4 d1d6 c4e4 c1a3 e4e2 a3b4');
INSERT INTO evaluations VALUES(52050,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(52050,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(52050,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(52050,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(52050,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(52050,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(52050,36,'cp -12','b8c6 d4d3 d7d5 c4d5 e6d5 d3d5 d8d5 e4d5 c6b4 b1a3 b4d5 g1f3 f8b4 c1d2 b4d2 f3d2 g8f6 f1c4 e8e7 d2b3 h8d8 e1g1 c8f5 f1d1 a8c8 c4f1 a7a6 b3d4 f5d7 a3c4 g7g6 a1c1 c8c7 f2f3 e7f8 c4e5 d7e8 c1c7');
INSERT INTO evaluations VALUES(18899,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(18899,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(18899,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18899,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(18899,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(18899,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(18899,36,'cp -21','d4d2 g8f6 b1c3 f8c5 f1e2 d8a5 g1f3 c5b4 e2d3 d7d6 e1g1 e8g8 d2e2 b4c3 b2c3 e6e5 a1b1 f6d7 b1b5 a5c7 f1d1 d7c5 d3c2 c8d7 f3h4 c6e7 b5b4 f7f6 h4f5 e7f5 e4f5 d7c6 c1e3 b7b6 e3c5 d6c5 b4b1 c7b7 f2f3 f8d8 c2e4 a8c8 h2h3 c6e4 f3e4 g8h8');
INSERT INTO evaluations VALUES(52388,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(52388,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52388,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(52388,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52388,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(52388,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52388,36,'cp 29','d7d5 e4e5 c6c5 c2c3 c8f5 d2d4 e7e6 c1e3 b8d7 g1f3 c5d4 c3d4 g8e7 b1c3 e7c6 f1d3 f5d3 d1d3 f8e7 h2h4 a8c8 h4h5 h7h6 e1g1 c6a5 f1b1 d8c7 c3b5 c7b6 b2b3 e8g8 b5c3 c8c6 b3b4 a5c4 c3a4 b6d8 a4c5 b7b6 c5a6 f7f5 b4b5 c6c8 a6b4 e7b4');
INSERT INTO evaluations VALUES(52818,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52818,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(52818,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(52818,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(52818,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(52818,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(52818,36,'cp 25','e4e5 c8f5 g1f3 e7e6 d2d4 c6c5 c1e3 c5d4 f3d4 g8e7 b1d2 b8d7 f2f4 g7g5 g2g3 a8c8 d2f3 g5f4 g3f4 f5e4 f1d3 d7c5 d3b5 c5d7 c2c3 h8g8 d1e2 d8a5 b2b4 a5c7 a1c1 a7a6 b5d3 e4d3 e2d3 d7b6 d3h7 b6c4 e3f2');
INSERT INTO evaluations VALUES(2390,30,'cp 252',NULL);
INSERT INTO evaluations VALUES(2390,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(2390,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(2390,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(2390,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(2390,35,'cp 248',NULL);
INSERT INTO evaluations VALUES(2390,36,'cp 244','e5d3 c8e6 b1c3 d8c6 c1e3 e8c8 a2a3 f6f5 e4e5 e6c4 e1c1 g8e7 f2f4 g7g6 h2h4 h7h5 g2g3 f8h6 d1e1 c6d4 e3d4 d8d4 b2b3 c4a6 d3c5 a6f1 h1f1 e7c6 c1b2 h8e8 b3b4 h6g7 e1e3 d4d2 f1d1 d2d1 c3d1 a7a5 c5d3 c6d4 d1c3 a5b4 a3b4');
INSERT INTO evaluations VALUES(4593,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(4593,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(4593,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(4593,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(4593,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(4593,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(4593,36,'cp 25','d7d6 e1g1 c8g4 c1e3 c6d4 e3d4 e5d4 c3d5 f6d5 c4d5 c7c6 d5b3 d8b6 h2h3 g4d7 a3a4 b6c5 d1d2 a7a5 c2c3 c5b6 d2c2 e7f6 a1c1 d4c3 b2c3 b6c7 d3d4 b7b5 f1e1 b5a4 b3a4 c6c5 a4d7 c7d7 e4e5 d6e5 d4e5');
INSERT INTO evaluations VALUES(3395,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3395,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(3395,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3395,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3395,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(3395,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(3395,36,'cp 24','h2h3 c8e6 c3d5 c6d4 f3d4 e5d4 d5f6 e7f6 c4e6 f7e6 c1d2 e6e5 d1g4 a7a5 e1g1 d8e8 f2f4 e5f4 f1f4 f6e5 f4f8 e8f8 a3a4 b7b6 g4f5 a8e8 a1f1 f8f5 f1f5 g7g6 f5f1 e5g7 d2f4 g7e5 g2g3 e8f8 h3h4 f8f7 f1f2 g8g7 f4h6 g7g8 g1g2 f7f2 g2f2 g8f7 b2b3');
INSERT INTO evaluations VALUES(9936,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(9936,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(9936,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(9936,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(9936,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(9936,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(9936,36,'cp 45','d7d6 e1g1 e8g8 c4b3 c8g4 c1e3 g8h8 a1c1 f6g8 c3d5 g8e7 d5b6 a7b6 h2h3 g4f3 d1f3 f7f5 c1d1 f5e4 f3e4 e7f5 d4d5 f5e3 e4e3 c6e5 d1c1 g7g5 c1c3 a8a5 a2a3 a5c5 e3d2 c5c3 d2c3 d8e7 b3d1 h8g8 d1e2 f8f4 f1e1 e7f6 c3c7 f4f2');
INSERT INTO evaluations VALUES(8919,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(8919,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8919,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(8919,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8919,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(8919,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8919,36,'cp 33','e1g1 c8g4 c1e3 e8g8 c4b3 g8h8 a1c1 f6g8 c3d5 g4f3 g2f3 c6a5 b3c2 a5c4 e3f4 c7c6 d5b6 a7b6 c2b1 b6b5 b2b3 d8f6 f4g3 c4a3 e4e5 d6e5 d4e5 f6e6 b1d3 f8d8 d1e2 g8e7 e2e4 g7g6 f1d1 e7f5 e4f4 h8g7 d3f5 d8d1 c1d1 e6f5 f4f5 g6f5 d1d7 g7g6 d7b7');
INSERT INTO evaluations VALUES(8918,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(8918,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(8918,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(8918,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(8918,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(8918,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(8918,36,'cp 35','e8g8 c4b3 c8g4 c1e3 g8h8 g1h1 f6g8 c3d5 g8e7 d5b6 a7b6 h2h3 g4f3 d1f3 f7f5 a1d1 f5e4 f3e4 e7f5 a2a3 d8f6 h1g1 a8e8 e4d3 f6h4 g1h2 h4e7 b3d5 f5e3 f2e3 c6d8 e3e4 c7c6 d5a2 d8e6 f1f8 e8f8 a2e6 e7e6 d4d5 e6e5 h2h1 e5b2 d5c6 b7c6 d3d6');
INSERT INTO evaluations VALUES(3666,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(3666,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(3666,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(3666,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(3666,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(3666,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(3666,36,'cp 33','d4d5 c6e5 f3e5 d6e5 d1f3 c7c6 c1e3 c6d5 c4d5 b6d4 h2h3 c8e6 a1d1 d8b6 e3d4 e5d4 c3e2 a8d8 d1d4 b6b2 f1d1 d8d7 a2a4 b2b6 e2f4 e6d5 f4d5 f6d5 e4d5 b6d6 h3h4 h7h6 g2g3 f8c8 f3d3 c8c5 d3e3 d7d8 g1g2 c5c7 d1d2 c7c5 e3a3 b7b6 a3e3');
INSERT INTO evaluations VALUES(10308,30,'cp -201',NULL);
INSERT INTO evaluations VALUES(10308,31,'cp -200',NULL);
INSERT INTO evaluations VALUES(10308,32,'cp -195',NULL);
INSERT INTO evaluations VALUES(10308,33,'cp -207',NULL);
INSERT INTO evaluations VALUES(10308,34,'cp -206',NULL);
INSERT INTO evaluations VALUES(10308,35,'cp -192',NULL);
INSERT INTO evaluations VALUES(10308,36,'cp -209','f6g4 c1e3 g4e3 f2e3 c5e3 f1c4 d7d6 d1d3 e3h6 e1g1 e8g8 a1d1 c8g4 c4b3 a7a6 a2a4 g4f3 g2f3 c6e5 d3d4 e5g6 g1h1 d8h4 d4g1 h6f4 g1f2 h4f2 f1f2 f8d8 h1g2 c7c6 c3e2 f4e5 c2c3 d8e8 f3f4 e5f6 d1d6 e8e4 d6d7');
INSERT INTO evaluations VALUES(10597,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(10597,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(10597,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(10597,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(10597,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(10597,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(10597,36,'cp 133','e2e3 c5b4 b1c3 b4c3 b2c3 d8e7 d1d5 e7a3 a1d1 f7f6 e5f6 g4f6 d5d2 d7d6 f1d3 e8g8 d2c2 g8h8 c2b3 a3a5 e1g1 c8d7 b3b5 d7g4 c4c5 a5b5 d3b5 d6c5 f4c7 f6e4 b5d3 g4f3 g2f3 e4c3 d1c1 c3a2 c1c5 f8f3 d3e4 f3f6 g1g2 a8f8 c7g3');
INSERT INTO evaluations VALUES(12536,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(12536,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(12536,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(12536,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(12536,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(12536,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(12536,36,'cp 201','f3g5 d7d5 e4d5 c6a5 c4b5 c8d7 d1e2 f8e7 b5d7 f6d7 g5f3 e8g8 d2d3 f8e8 b1c3 a7a6 c1d2 b7b5 a1e1 a5b7 c3e4 b7d6 c2c4 b5c4 d3c4 f7f5 e4d6 e7d6 f3g5 d7f8 g2g3 a6a5 g5h3 d8d7 f2f3 d7a4 g3g4 a4a2 d2c3 a5a4 g4f5 a2b3');
INSERT INTO evaluations VALUES(13634,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(13634,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(13634,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(13634,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(13634,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(13634,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(13634,36,'cp 42','d5c4 f1c4 f8e7 e1g1 e8g8 c1f4 a7a6 c4a2 b7b5 d4d5 e6d5 c3d5 f6d5 a2d5 c8b7 f3e5 c6e5 d5b7 a8a7 b7e4 a7d7 d1h5 e5g6 h5f5 e7f6 a1b1 d8e7 g2g3 d7d8 f4e3 e7e5 b2b4 f8e8 e4c2 e5f5');
INSERT INTO evaluations VALUES(13973,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(13973,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(13973,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(13973,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13973,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(13973,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(13973,36,'cp -11','f8e7 e1g1 e8g8 g5e3 a7a6 a2a4 c8d7 c4d3 a8c8 h2h3 c6b4 a4a5 d8c7 d4f3 d7c6 f1e1 b4d3 c2d3 d6d5 e3b6 c7b8 f3d4 e7b4 d4c6 b7c6 d1a4 c6c5 e4d5 b4c3 b2c3 f6d5 a4a3 d5b6 a5b6 b8b6 a3a6');
INSERT INTO evaluations VALUES(8187,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(8187,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(8187,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(8187,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(8187,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(8187,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(8187,36,'cp -9','e1g1 e8g8 h2h3 c8d7 g5e3 d8c7 c4d3 a7a6 g1h1 c6d4 e3d4 d7c6 d1e2 f6d7 c3d5 c6d5 e4d5 e6e5 d4e3 f7f5 f2f3 e7h4 b2b4 f5f4 e3g1 d7f6 c2c4 f6h5 g1h2 g7g6 a2a3 c7e7 a1d1 h5g3 h2g3 f4g3 d3e4 e7g5');
INSERT INTO evaluations VALUES(8186,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(8186,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(8186,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(8186,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(8186,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(8186,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(8186,36,'cp -12','e8g8 d4c6 b7c6 d1f3 d6d5 c4d3 a7a5 b2b3 c8a6 f1e1 h7h6 g5h4 f8e8 a1c1 d8b8 c3a4 a6b5 c2c4 d5e4 d3e4 f6e4 c4b5 e7h4 e1e4 h4g5 c1c6 b8b5 e4c4 b5d5 a4c3 d5f3 g2f3 e8d8 a2a4 g5e7 g1g2 e7b4 c3e4');
INSERT INTO evaluations VALUES(3130,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(3130,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(3130,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(3130,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(3130,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(3130,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(3130,36,'cp -9','h2h3 c8d7 g5e3 a7a6 c4d3 c6d4 e3d4 d7c6 d1e2 b7b5 a2a3 d8c7 a1d1 a8e8 g1h1 f6d7 f2f3 d7e5 f3f4 e5d7 d1e1 c6b7 e2g4 e7f6 d4f6 d7f6 g4h4 c7d8 h1h2 e6e5 f4f5 d6d5 e4d5 b7d5 d3e4 f6e4 h4d8 e8d8 c3e4 d5c4 f1f2 h7h6 f5f6');
INSERT INTO evaluations VALUES(16032,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16032,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(16032,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(16032,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(16032,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(16032,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(16032,36,'cp 21','e7e6 b1d2 f8e7 h2h3 e8g8 a2a4 a7a5 f1d3 c7c5 e1g1 c5c4 d3e2 b7b6 c2c3 c8b7 b2b3 c4b3 d1b3 d8c8 f1c1 b7a6 c3c4 c8b7 c4d5 f6d5 e2a6 b7a6 f4g5 e7g5 f3g5 d7f6 g5e4 f6e4 d2e4 f8c8 e4d6 c8d8 d6b5 a8c8 e3e4 d5b4 e4e5 b4d5');
INSERT INTO evaluations VALUES(16397,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(16397,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(16397,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(16397,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16397,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(16397,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(16397,36,'cp -1','c6a5 b1c3 f8b4 c1d2 a5c4 d3c4 b4c3 d2c3 f6e4 d1g4 e4c3 g4g7 h8f8 g7e5 d8e7 e5e7 e8e7 b2c3 d7d6 e1c1 c8e6 d1d4 f8g8 g2g4 h7h5 f2f3 g8g5 g1e2 h5g4 f3g4 g5c5 e2f4 e6c4 h1e1 e7d7 a2a4 a8h8 e1e4 c4e6 e4e3 a7a6 c1b2 b7b5 d4e4 c5c4 a4b5 c4e4 e3e4 a6b5 f4e6');
INSERT INTO evaluations VALUES(19241,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(19241,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(19241,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(19241,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(19241,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(19241,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(19241,36,'cp -10','g1f3 g8f6 b1c3 f8g7 d2d3 e8g8 c1e3 d7d6 h2h3 e7e6 c4a2 d6d5 e4d5 f6d5 c3d5 e6d5 e1g1 g7b2 a1b1 b2c3 e3c5 f8e8 c5e3 b7b6 e3d2 c3g7 f1e1 e8e1 d1e1 h7h6 d2c3 c8e6 c3g7 g8g7 e1c3 d5d4 c3a1 e6a2 a1a2 d8d6 a3a4 a8d8 b1b5 d8d7 a2c4 d7c7 f3d2');
INSERT INTO evaluations VALUES(19240,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(19240,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(19240,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(19240,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(19240,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(19240,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(19240,36,'cp -24','g8f6 b1c3 f8g7 d2d3 e8g8 c3d5 f6d5 c4d5 d8c7 h2h4 h7h5 e1g1 e7e6 d5a2 d7d5 f1e1 d5e4 d3e4 b7b5 d1e2 b5b4 e4e5 b4a3 b2a3 c8b7 c2c3 a8d8 c1f4 c6e7 a3a4 b7f3 e2f3 d8d7');
INSERT INTO evaluations VALUES(21410,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(21410,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21410,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(21410,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(21410,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(21410,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21410,36,'cp 17','e2e3 g8f6 a2a3 c6b8 f1d3 c7c5 c2c3 b7b6 f3e5 c8b7 b1d2 f8e7 d1f3 b7a6 d3a6 b8a6 g2g4 f6d7 e1g1 d7e5 f4e5 e8g8 f3g2 f7f5 e5f6 f8f6 f1f6 e7f6 d2f3 h7h6 g1h1 d8e7 g2e2 a6c7 c1d2 a7a6 b2b4 c5b4');
INSERT INTO evaluations VALUES(21764,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(21764,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(21764,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(21764,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(21764,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(21764,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(21764,36,'cp -84','g8f6 g1f3 f8d6 c2c4 d5c4 d3c4 e5e4 f3d4 e8g8 f1e2 d8e7 d1c2 f8e8 d4c6 b7c6 c1d2 f6d7 d2c3 a7a5 b1d2 a5a4 e1c1 d7c5 g2g4 e7h4 h1f1 a8b8 c3d4 h4e7 f1h1 f7f5 f2f4 e4f3 e2f3 c8e6 c1b1 c5b3 d4c3');
INSERT INTO evaluations VALUES(23994,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(23994,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(23994,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(23994,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(23994,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(23994,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(23994,36,'cp -30','c6f3 e8g8 d2d3 f6g4 g5h3 a8c8 b1c3 c8c6 f3e2 c6g6 f2f3 g4h6 c1h6 g6h6 e1g1 d8b6 g1h1 b6b2 c3d5 e7d6 d5e3 b2d4 a1b1 f8e8 c2c4 d6c5 e3d5 d7h3 g2h3 f7f5 b1e1 c5b6 f3f4 e5e4 d3e4 e8e4 e2f3 d4c4 d5e7 g8f8 e7f5 h6e6 e1e4 c4e4 f3e4 e6e4');
INSERT INTO evaluations VALUES(23996,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(23996,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(23996,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(23996,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(23996,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(23996,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(23996,36,'cp -25','e8g8 d2d3 f6g4 g5h3 a8c8 b1c3 c8c6 f3e2 c6g6 f2f3 g4h6 c1h6 g6h6 e1g1 d8b6 g1h1 b6b2 c3d5 e7d6 d5e3 d7h3 g2h3 b2b8 e3f5 h6h5 a1b1 b8d8 f5d6 d8d6 b1b5 a7a6 b5a5 h5f5 e2e4 g7g6 f1g1 d6f6 a5a6 f6a6 e4f5 a6a2 f5f6 a2c2');
INSERT INTO evaluations VALUES(24200,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(24200,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(24200,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24200,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(24200,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24200,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(24200,36,'cp 3','g7g5 d2d4 g5g4 h3g4 f6g4 d4d5 c6b8 a2a4 e8g8 f3h2 g4h2 g1h2 f7f5 b5d7 d8d7 c1h6 f8f6 d1h5 e7f8 h6f8 f6f8 b1d2 d7f7 h5f7 f8f7 a4a5 b8d7 b2b4 d7f6 h2g1 a8f8 a1a2 b7b5 a5b6 a7b6 c3c4 f6e4 d2e4 f5e4 e1e4 f7f4 e4f4 f8f4');
INSERT INTO evaluations VALUES(34792,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(34792,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(34792,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(34792,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(34792,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(34792,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(34792,36,'cp 228','b1c3 g8f6 d4e5 d6e5 d1d2 f8b4 f1b5 b8d7 e1c1 e8g8 f3h4 b4c3 b2c3 g6e4 f2f3 e4d5 d2d5 f6d5 d1d5 c7c6 d5d7 c8d7 b5d3 d7e6 a2a3 f7f6 g5e3 g7g6 g2g4 g8f7 h4g2 b7b5 e3c5 f8h8 g4g5 f6g5 g2e3 h7h5 c3c4 f7f6 c4b5 c6b5 d3b5 a8d8 h1g1 g5g4 g1f1');
INSERT INTO evaluations VALUES(36516,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36516,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(36516,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36516,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(36516,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(36516,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36516,36,'cp 20','g1e2 e7e5 e1g1 c7c6 d2d4 f8e8 e3e4 e5d4 e2d4 b8d7 c1e3 d7c5 f2f3 a7a5 d1c2 a5a4 e3f2 d8a5 h2h3 a5a6 c3e2 b7b5 c4b5 c6b5 c2d2 b5b4 e2f4 a6b7 b2b3 c8d7 f1d1 e8c8 a1c1 a4b3 d4b3 c5e6 c1c8 a8c8');
INSERT INTO evaluations VALUES(36517,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36517,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36517,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36517,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36517,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36517,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(36517,36,'cp 0','e7e5 e1g1 c7c6 d2d4 f8e8 e3e4 e5d4 e2d4 b8d7 c1f4 d7c5 d1c2 c5e4 c3e4 f6e4 d4c6 b7c6 g2e4 a8b8 e4c6 b8b2 c2d1 e8e2 c4c5 c8h3 d1d6 d8c8 c6d5 g7d4 f4e3 d4e3 f2e3 h3f1 a1f1 c8e8 d5f7 e8f7 f1f7 g8f7 d6c7 f7g8 c7d8 g8f7 d8c7');
INSERT INTO evaluations VALUES(36854,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36854,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36854,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(36854,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36854,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(36854,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36854,36,'cp 7','e1g1 c7c6 d2d4 f8e8 d1c2 b8a6 e3e4 e5d4 e2d4 a6c5 c1f4 f6e4 c3e4 c5e4 d4c6 b7c6 g2e4 a8b8 a1b1 g7e5 e4c6 e5f4 c6e8 d8e8 g3f4 c8f5 c2c3 f5b1 f1b1 e8e4 b1e1 e4f4 b2b3 d6d5 c4d5 b8d8 e1e5 f7f6 e5e3 f4g4 g1f1 g4d1 c3e1 d1d5 e3e8 d8e8 e1e8 g8g7 e8e7 g7h6 e7e3 g6g5 h2h3 h6g6 f1g1 d5d1 g1g2 d1d5 g2h2');
INSERT INTO evaluations VALUES(36853,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36853,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36853,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36853,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36853,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36853,36,'cp 30','c7c6 d2d4 f8e8 e3e4 e5d4 e2d4 b8d7 c1e3 d7c5 f2f3 a7a5 d1c2 a5a4 e3f2 d8a5 f1d1 c5e6 a1b1 e6d4 f2d4 c8e6 g2f1 f6d7 c2f2 g7d4 d1d4 d7e5 f3f4 e5g4 f2d2 a5c5 b1d1 f7f5 h2h3 g4f6 e4f5 e6f5 g3g4 f5e6 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(38877,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(38877,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(38877,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38877,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(38877,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(38877,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38877,36,'cp 27','b7b5 c1d2 a7a6 a2a4 c8b7 a4b5 b4c3 d2c3 a6b5 a1a8 b7a8 b2b3 g8f6 b3c4 b5c4 f1c4 e8g8 e1g1 b8d7 c4e2 a8d5 f3e1 d8a8 f2f3 a8a2 d1c2 c7c5 e3e4 d5c4 c2a2 c4a2 g1f2 c5d4 c3d4 e6e5 d4e3 f8b8 e1d3 a2c4 f1a1 c4d3 e2d3 h7h6 a1b1 b8a8 b1c1 d7f8 c1c6 a8a2 f2f1 a2a1 f1e2');
INSERT INTO evaluations VALUES(43885,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(43885,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(43885,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(43885,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(43885,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43885,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(43885,36,'cp 43','d2d4 e7e6 f1g2 c8b7 c2c4 f8b4 c1d2 c7c5 d2b4 c5b4 a2a3 b4a3 b1a3 e8g8 e1g1 a7a5 d1d2 b8a6 f1c1 a6b4 a3b5 b7c6 f3e5 d8e7 e5d3 b4d3 d2d3 f8c8 d3a3 e7a3 a1a3 d7d5 b5d6 c8d8 c4d5 c6d5 g2d5 f6d5 c1c6 d8d7 d6c4 a8d8 c4b6 d5b6 c6b6 d7d4 a3a5');
INSERT INTO evaluations VALUES(44167,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(44167,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(44167,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(44167,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(44167,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(44167,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(44167,36,'cp -12','c5d4 f3d4 h7h6 g5f6 d8f6 b1c3 f8c5 e2e3 b8c6 d4b3 c5a3 d1c2 a3b4 a1c1 b7b6 f1e2 b4c3 c2c3 f6c3 c1c3 e8e7 f2f4 c8b7 e1f2 a7a5 e2f3 a5a4 b3d4 c6d4 e3d4 b7f3 f2f3 h8c8 h2h4 d7d5 b2b3 a4b3 a2b3 d5c4');
INSERT INTO evaluations VALUES(44270,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(44270,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(44270,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(44270,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(44270,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(44270,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(44270,36,'cp -11','f3d4 h7h6 g5f6 d8f6 b1c3 b8c6 e2e3 f8c5 d4b3 c5a3 d1c2 a3b4 a1c1 b7b6 f1e2 c8b7 a2a3 b4c3 c2c3 f6c3 c1c3 g7g5 e1g1 d7d6 b3d4 e8e7 f1d1 h8d8 b2b4 a8c8 f2f3 f7f5 g1f2 c6e5 f3f4 e5f7 e2f3 b7e4 f3e4 f5e4 d4b5 a7a6 b5d4 d8g8');
INSERT INTO evaluations VALUES(44269,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(44269,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(44269,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(44269,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(44269,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(44269,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(44269,36,'cp -11','h7h6 g5f6 d8f6 b1c3 b8c6 e2e3 f8c5 d4b3 c5a3 d1c2 a3b4 a1c1 b4c3 c2c3 f6c3 c1c3 e8e7 f1e2 b7b6 b3d4 c8b7 f2f3 f7f5 e1f2 g7g5 h1d1 c6e5 f3f4 e5f7 c3d3 h8d8 e2f3 b7e4 d3d2 a8c8 b2b3 d7d5 c4d5 d8d5 f3e4');
INSERT INTO evaluations VALUES(44503,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44503,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44503,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44503,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44503,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44503,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44503,36,'cp 0','e7e6 f4d6 f8d6 d1d6 b8a6 b1c3 d8e7 d6g3 a6c5 g3g7 h8g8 g7h6 d7d5 g1f3 e6e5 e2e3 g8g6 f1b5 e8d8 h6h4 g6g4 h4h6');
INSERT INTO evaluations VALUES(44827,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(44827,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(44827,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(44827,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(44827,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(44827,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(44827,36,'cp 82','d2d4 c5d4 f3d4 e7e6 c1e3 a7a6 c4b3 d8c7 f2f3 b8d7 a2a3 d7c5 b3a2 e6e5 d4f5 g7g6 f5g3 c8e6 d1d2 h6h5 g3e2 f8e7 e1g1 e8g8 a3a4 a8c8 a2e6 f7e6 g1h1 f8f7 e3g5 c7c6 f1d1 c6c7 h2h3 c7c6 b2b3 c5d7 a1b1 d7c5 g5f6');
INSERT INTO evaluations VALUES(47230,30,'cp -151',NULL);
INSERT INTO evaluations VALUES(47230,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(47230,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(47230,33,'cp -158',NULL);
INSERT INTO evaluations VALUES(47230,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(47230,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(47230,36,'cp -134','e7e5 c1e3 c8e6 b1d2 b8c6 a2a3 d8d7 g1e2 f8e7 g2g3 h7h6 c2c4 d5d4 e3f2 a7a5 f3f4 d7d8 h2h3 f6d7 d2f3 f7f6 f3h4 e6f7 f1g2 a5a4 h4f5 e7f8 e1g1 d7c5 e2c1 d8d7 b2b4 a4b3 c1b3 c5a4 d1c2 g7g6 f5h4 a4c3 f4f5 g6g5 h4g6 h8g8 c2b2');
INSERT INTO evaluations VALUES(47757,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(47757,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(47757,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(47757,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(47757,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(47757,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(47757,36,'cp -46','f8e7 g5f6 e7f6 b1c3 e8g8 g2g3 b8d7 g1f3 f6e7 f3d2 d7f6 d2c4 c7c6 c4e3 d8b6 a1b1 h7h5 f1e2 g7g6 h2h4 b6d4 e2f3 e7d8 e1g1 d4c5 d1e2 d8a5 f1d1 a5c3 b2c3 c6d5 e3d5 f6d5 d1d5 c5c3 d5d6 c3c7 d6d3 c8e6 b1d1');
INSERT INTO evaluations VALUES(40739,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(40739,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(40739,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(40739,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(40739,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(40739,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(40739,36,'cp -42','g5f6 e7f6 b1c3 b8d7 g1f3 e8g8 g2g3 f6e7 f3d2 c7c6 d2c4 d7f6 c4e3 d8b6 a1b1 g7g6 f1e2 b6d4 d1d3 e7d8 e1g1 d4d3 e2d3 g8g7 f1d1 d8a5 e3c4 a5c3 b2c3 c6d5 c4d6 f8d8 d6c8 a8c8 e4d5 d8d5 b1b7 c8c3 b7a7 e5e4 d3e2 c3c2');
INSERT INTO evaluations VALUES(40733,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(40733,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(40733,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(40733,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(40733,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(40733,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(40733,36,'cp -40','e7f6 b1c3 e8g8 g1f3 b8d7 a2a4 g7g6 h2h4 h7h5 g2g3 f6g7 f1h3 d7f6 h3c8 a8c8 e1g1 c7c6 d1d3 c6d5 c3d5 f6d5 d3d5 c8c2 d5b7 d8b8 b7b8 f8b8 a1c1 b8b2 c1c2 b2c2 f1b1 g7f6 b1b7 a7a5 b7b5 c2c4 b5a5 c4e4 a5a6 f6e7 a6a7 g8f8 f3g5 e7g5 h4g5 f8g7 a4a5 e4a4 a5a6 e5e4 a7e7 a4a6');
INSERT INTO evaluations VALUES(41841,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(41841,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(41841,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(41841,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(41841,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(41841,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(41841,36,'cp -26','b1c3 b8d7 g1f3 e8g8 g2g3 f6e7 f3d2 d7f6 d2c4 c7c6 c4e3 d8b6 a1b1 h7h5 h2h4 g7g6 f1e2 f8d8 a2a4 a7a5 e2c4 f6g4 e3g4 c8g4 d1d2 g8g7 e1g1 d8f8 b2b3 f7f5 e4f5 g6f5 c4e2 b6d4 d2d4 e5d4 e2g4 h5g4');
INSERT INTO evaluations VALUES(49991,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(49991,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(49991,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(49991,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(49991,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(49991,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(49991,36,'cp 51','d2d4 d7d6 f1e2 f8e7 g1h3 e8g8 e1g1 f6e4 c3d5 e4f6 d5e7 d8e7 f2f3 b8c6 d4d5 c6b8 h3f2 a7a5 b2b3 b8d7 c1b2 d7c5 d1d2 b7b6 f2d3 a5a4 d3c5 b6c5 b2c3 h7h5 b3b4 c5b4 c3b4 h5h4 a1b1 f6d7 h2h3 d7c5 g1h2 c8d7 b4a3 f8e8');
INSERT INTO evaluations VALUES(49997,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(49997,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(49997,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(49997,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(49997,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(49997,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(49997,36,'cp 36','d7d6 f1e2 f8e7 g1h3 e8g8 e1g1 f6e4 c3d5 e4f6 d5e7 d8e7 f2f3 a7a5 h3f2 b8c6 d4d5 c6b8 b2b3 b8a6 c1b2 c8d7 d1d2 b7b6 a1e1 e7f7 f2d3 e5e4 d3f2 a6c5 f3f4 f8b8 b2c3 f7g6 f2d1 h7h6 c3d4 b6b5 c4b5 d7b5 e2b5 b8b5 d1c3');
INSERT INTO evaluations VALUES(50096,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(50096,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(50096,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50096,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(50096,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50096,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(50096,36,'cp 6','e7e6 g1f3 b7b6 f1d3 c8b7 e1g1 f8e7 a2a3 c7c5 b2b3 e8g8 c1b2 b8c6 c3e2 f6e4 c2c4 e7f6 d1c2 c5d4 e3d4 c6e7 e2f4 d8e8 d4d5 e6d5 c4d5 g7g5 d5d6 g5f4 d6e7 e8e7 b2f6 e7f6 c2c7 e4c5 d3e2 g8h8 a1d1 a8c8 c7f4');
INSERT INTO evaluations VALUES(49957,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(49957,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(49957,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(49957,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(49957,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(49957,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(49957,36,'cp 23','g1f3 d7d5 c3e2 c7c6 e2f4 f6e4 f3e5 d8f6 d1h5 g7g6 h5e2 f8b4 c2c3 b4e7 g2g3 b8d7 f4d3 d7e5 d3e5 e4c5 c3c4 c5d7 e5d7 c8d7 f1g2 e7d6 c1d2 f6e7 e1c1 a7a5 c4c5 d6c7 d2c3 e8g8 h2h4 e6e5 d4e5');
INSERT INTO evaluations VALUES(49958,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(49958,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(49958,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(49958,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(49958,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(49958,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(49958,36,'cp 14','b7b6 d4d5 f8b4 d5e6 d7e6 c1d2 d8e7 f1c4 b4c3 d2c3 f6e4 c3e5 b8d7 d1e2 c8b7 e1c1 e8c8 h1e1 d7e5 f3e5 e7f6 f2f4 d8d6 c4a6 h8d8 d1d6 d8d6 e1d1 f6e7 d1d6 e4d6 a6b7 c8b7 e5f3 d6e4 a2a3 e7d6 e2d3');
INSERT INTO evaluations VALUES(51108,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51108,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(51108,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(51108,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(51108,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(51108,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(51108,36,'cp 38','c7c5 g1f3 e7e6 d2d4 c5d4 f3d4 b7b5 f1d3 c8b7 e1g1 b8c6 d4c6 b7c6 d1e2 g8e7 c1e3 b5b4 c3b1 e7g6 f2f4 f8e7 b1d2 e8g8 d2c4 d8c7 e3b6 c7b8 e2f2 d7d5 c4a5 b8c8 a5c6 c8c6 e4d5 c6d5 a1e1 d5a2 b2b3 a8e8 f4f5 e6f5 f2f5 e7f6 e1e8 f8e8 g1h1 g8h8');
INSERT INTO evaluations VALUES(51726,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(51726,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(51726,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(51726,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(51726,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(51726,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(51726,36,'cp 236','h2h4 h7h6 c1f4 d7d5 b1c3 g8f6 f1d3 d5e4 c3e4 f6e4 d3e4 b8d7 d1d2 d7f6 e4d3 h6h5 e1c1 f8g7 f4g5 a7a5 c2c3 b7b5 f3e5 d8c7 d1e1 b5b4 c3c4 c6c5 d4d5 e8g8 d5e6 c8e6 e5g6 f8e8 d2f4 c7f4 g6f4 f6g4 f4e6 f7e6');
INSERT INTO evaluations VALUES(52546,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(52546,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(52546,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(52546,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(52546,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(52546,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(52546,36,'cp 175','b1c3 f6f5 c4d5 c6d5 d1a4 c8d7 a4b3 d7c6 c1f4 e7e6 e2e3 f8d6 f1b5 d6f4 b5c6 b8c6 b3b7 g8e7 e3f4 a8b8 b7a6 b8b2 e1g1 d8b6 a6d3 b6a5 d3e3 c6d8 g2g3 e8g8 f3e5 e7g6 e5d7 f8e8 d7c5 d8c6 c5b3 a5b4 f1b1 b2c2 b1c1 c2c1 a1c1 h7h6 c3e2 b4d6 b3c5 g8f7 g1h1 g6e7');
INSERT INTO evaluations VALUES(52542,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(52542,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(52542,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(52542,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(52542,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(52542,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(52542,36,'cp 154','f6f5 c4d5 c6d5 d1a4 c8d7 a4b3 e7e6 b3b7 b8c6 c1f4 c6a5 b7a6 d8b6 a6b6 a7b6 a2a3 g8f6 e2e3 f6e4 h2h4 e4c3 b2c3 a8c8 a1b1 c8c3 e1d2 c3b3 b1b3 a5b3 d2c3 b3a5 f1a6 f8a3 h1b1 a3e7 f3e5 e7d8 f4g5 a5c6 g5d8 c6e5 d8b6 e5f7 b6c7 e8e7');
INSERT INTO evaluations VALUES(54909,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(54909,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(54909,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(54909,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(54909,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(54909,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(54909,36,'cp 96','d2d4 g8f6 b1c3 e7e6 f1d3 a7a6 d1e2 f8e7 e1g1 b8c6 e4e5 f6d5 a2a3 d5c3 b2c3 d6d5 f3d2 c6a5 d2b3 a5c4 a3a4 a6a5 e2g4 g7g6 g4f3 c8d7 d3c4 d5c4 b3d2 d7c6 f3e2 b7b6 d2c4 c6b7 f1d1 b7a6 e2e4 d8d5 e4d5 e6d5 c4e3 e8c8');
INSERT INTO evaluations VALUES(55263,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(55263,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(55263,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(55263,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(55263,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(55263,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(55263,36,'cp 17','g1f3 e7e6 e1g1 g8e7 c2c3 c7c5 b2b4 c5b4 f3d4 e7g6 d3c2 b4c3 d2c3 f8c5 d4b3 d8d1 f1d1 c5e7 a2a4 c8d7 f2f4 e8g8 g2g3 f8c8 c1e3 g6f8 e4e5 d7c6 b3a5 c6d5 c3c4 d5c4 c2e4 c4d5 e4d5 e6d5');
INSERT INTO evaluations VALUES(55262,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55262,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(55262,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(55262,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(55262,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55262,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(55262,36,'cp 13','e7e6 c2c3 c7c5 d3b5 c8d7 b5d7 d8d7 c3d4 c5d4 d2d3 f8d6 b1d2 b8c6 d2c4 g8e7 c1d2 e8g8 e1g1 h7h6 a2a4 e7g6 d1e2 a8c8 g2g3 f7f6 h2h4 g6e5 f3e5 c6e5 c4d6 d7d6 d2f4 c8c5 f4e5 d6e5 a1c1 f8c8 f2f4 e5d6 c1c5 c8c5');
INSERT INTO evaluations VALUES(57375,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(57375,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(57375,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(57375,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(57375,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(57375,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(57375,36,'cp 158','e2e4 d7d5 f1d3 b8c6 a2a3 g8e7 c2c4 d5e4 d3e4 e6e5 d4d5 f6f5 e4c2 e5e4 b1d2 e4f3 d2f3 c6a5 c2a4 e8f7 d1e2 f7g8 b2b4 a5c4 e2c4 d8d5 c4d5 e7d5 c1b2 c8e6 e1c1 d5b6 a4c2 g7g6 h1e1 e6d5 h2h4 c7c6 h4h5 f8h6 c1b1 b6c4 e1e7 c4b2 b1b2 h6g7 b2c1 g6h5 e7b7');
INSERT INTO evaluations VALUES(4452,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4452,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4452,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(4452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4452,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(4452,36,'cp 5','a2a4 d7d6 e1g1 c6e7 b1c3 e7g6 d1e1 c7c6 c4b3 f8e8 f3g5 e8f8 e1g3 d8e7 h2h3 h7h6 g5f3 c8e6 b3e6 e7e6 a4a5 a8d8 f3h4 g6h4 g3h4 d6d5 f1f2 d5e4 c3e4 f6e4 h4e4 d8d7 e4b4 a7a6 b4c5 e6d5 c5d5 c6d5 d3d4');
INSERT INTO evaluations VALUES(5142,30,'cp 342',NULL);
INSERT INTO evaluations VALUES(5142,31,'cp 295',NULL);
INSERT INTO evaluations VALUES(5142,32,'cp 302',NULL);
INSERT INTO evaluations VALUES(5142,33,'cp 298',NULL);
INSERT INTO evaluations VALUES(5142,34,'cp 301',NULL);
INSERT INTO evaluations VALUES(5142,35,'cp 316',NULL);
INSERT INTO evaluations VALUES(5142,36,'cp 336','d4b5 a7a6 b5d6 e8f8 c1e3 b7b5 c4e2 g8e7 d1d2 d8a5 a2a3 b5b4 c3d1 b4a3 c2c3 a5c7 a1a3 h7h5 b2b4 c8b7 e1g1 f8g8 e3c5 c6e5 c3c4 b7c6 d1e3 g8h7 f2f4');
INSERT INTO evaluations VALUES(7460,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(7460,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7460,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(7460,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7460,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(7460,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7460,36,'cp 25','d7d5 f3e5 c6e5 d2d4 d5e4 d4c5 d8e7 b1d2 c8b7 d1e2 e8g8 d2e4 b7e4 c2e4 a8e8 e4f3 e7c5 c1e3 c5e7 a2a4 e7d7 e2d1 d7e6 f1e1 e5f3 d1f3 f6d5 a4b5 a6b5 h2h3 d5e3 e1e3 e6c4 a1a7 e8e3 f3e3 h7h6 g1h2 c4b3 a7c7 b3b2 c7c5 b2b1');
INSERT INTO evaluations VALUES(11418,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11418,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11418,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11418,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11418,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11418,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11418,36,'cp 0','b7b5 d4e5 d6e5 d1d8 c6d8 a4b3 f8d6 c1g5 c7c6 e1c1 d6c7 f3h4 a6a5 g5f6 g7f6 a2a4 b5b4 c3b1 d8b7 b1d2 e8e7 d2c4 c8e6 f2f3 b7c5 g2g3 h7h5 d1d2 h8g8 h1e1 g8g5 d2d1 a8d8 d1d8 c7d8 c4d2 e6h3 d2c4');
INSERT INTO evaluations VALUES(11645,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11645,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11645,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(11645,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11645,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11645,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(11645,36,'cp 0','d4e5 d6e5 d1d8 c6d8 a4b3 f8d6 c1g5 c7c6 e1c1 d6c7 f3h4 a6a5 g5f6 g7f6 a2a4 b5b4 c3b1 d8b7 b1d2 e8e7 f2f3 h8d8 g2g4 d8d4 d2c4 c8e6 c4d2');
INSERT INTO evaluations VALUES(11925,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(11925,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(11925,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(11925,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(11925,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(11925,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(11925,36,'cp 105','g1f3 a8b8 e1g1 f8e7 d2d4 e5d4 f3d4 c6d4 d1d4 e8g8 c1f4 d7d6 d4d3 c8b7 a1d1 b7g2 g1g2 f8e8 h2h4 f6d7 c3d5 h7h6 e2e4 e7f8 f2f3 d7e5 d3e2 e5g6 b2b3 c7c6 h4h5 g6f4 d5f4 b6b5 d1c1 d8c8 e2d2 b5c4');
INSERT INTO evaluations VALUES(17735,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17735,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(17735,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(17735,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(17735,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(17735,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17735,36,'cp 59','d3d4 c8d7 e1g1 g8f6 c1e3 e5d4 f3d4 f8e7 a2a4 e8g8 d4c6 d7c6 f2f3 f6d7 d1d2 d7e5 c4e2 f8e8 a1d1 e7f6 a4a5 e5d7 e2c4 f6g5 e3g5 h6g5 h2h3 d7e5 c4b3 a8b8 f3f4 g5f4 f1f4 b7b5 d1f1 d8g5 c3d5 c6d5 d2d5');
INSERT INTO evaluations VALUES(20561,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(20561,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(20561,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(20561,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(20561,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(20561,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(20561,36,'cp 110','f1c4 f8c5 d2d3 g8f6 g1e2 e8g8 e1g1 c6a5 c1g5 h7h6 g5f6 d8f6 c4d5 c5b6 g1h1 a5c6 d5b3 c8g4 c3d5 f6d6 d5b6 a7b6 h2h3 g4e6 b3e6 d6e6 a2a3 f8d8 e2g3 f7f5 d1h5 c6d4 c2c3 d4b3 a1d1 b3c5 g3f5 c5d3');
INSERT INTO evaluations VALUES(22347,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(22347,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(22347,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(22347,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(22347,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(22347,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(22347,36,'cp 200','f4e5 g8f6 c2c4 c8e6 c4d5 f6d5 a2a3 d8g5 d1d4 a8d8 b1d2 f8e7 d2f3 g5g6 d4a4 c7c6 f1e2 e8g8 e1g1 a7a6 a1c1 d5b6 a4a5 f8f3 e2f3 b6c4 c1c4 e6c4 f1c1 d8d7 e5d4 c4f7 a5c3 e7d6 h2h3 f7e6 h3h4 d6c7 h4h5 g6g5 c1d1 g5e7 h5h6');
INSERT INTO evaluations VALUES(22579,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(22579,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(22579,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(22579,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(22579,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(22579,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(22579,36,'cp -54','d7d5 e4d5 e6d5 c4b3 c6d4 g1f3 c8g4 b1d2 d4b3 a2b3 d8e7 d1e2 f7f6 c2c3 c7c5 h2h3 g4h5 e2e7 g8e7 d3d4 e7c6 d4c5 f8c5 b3b4 c5b6 d2b3 h5f3 g2f3 e8c8 c1e3 h8e8 e1g1 b6e3 f2e3 e8e3 b3c5 e3e5 b4b5 d5d4 b5c6 e5c5 a1a7');
INSERT INTO evaluations VALUES(21371,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(21371,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(21371,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(21371,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(21371,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(21371,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(21371,36,'cp -55','e4d5 e6d5 c4b3 c6d4 b1c3 c7c6 c3e2 d4b3 a2b3 f8d6 c1f4 g8e7 d1d2 e8g8 g1f3 c8g4 f3d4 d6f4 d2f4 g4e2 d4e2 e7g6 f4d2 f8e8 e1g1 d8f6 c2c3 a7a5 f2f4 g6e7 e2g3 e7f5 f1e1 h7h5 e1e8 a8e8 a1e1 e8e1 d2e1 h5h4 g3f5 f6f5');
INSERT INTO evaluations VALUES(23478,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(23478,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23478,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(23478,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(23478,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(23478,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(23478,36,'cp 55','b5a4 b7b5 c2c3 e7e6 c3d4 b5a4 d2d3 c8a6 g1f3 c7c5 d4d5 e6d5 b1c3 d5d4 d1a4 a6b7 c3e2 g8f6 e1g1 f8e7 c1g5 e8g8 b2b4 b7c6 a4c2 c5b4 e2d4 a7a5 a1c1 a8c8 d4f5 h7h6 g5h4 d7d6 c2b2 c6d7 c1c8 d7c8 f5e7 d8e7 a2a3 c8g4 a3b4 g4f3 g2f3 a5b4 b2b4');
INSERT INTO evaluations VALUES(27412,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(27412,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(27412,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(27412,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(27412,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(27412,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(27412,36,'cp 189','d4d5 c6e5 c3b5 a5b6 d2e3 b6a5 c2c3 c7c5 d5c6 g4f3 b2b4 a5d8 d1d8 e8d8 g2f3 e5c6 f1h3 e7e6 e1c1 d8c8 h1e1 a7a5 a2a3 f8e7 e3f4 h8d8 b5c7 d8d1 e1d1 f6h5 c7e6 h5f4 e6f4 c8b8 c1b2 a5b4 a3b4 e7h4 f4d5 h4f2 f3f4 c6d8 d1d2 f2h4 d2e2 a8a6 e2e8 b8a7 b2b3 a6d6 c3c4 h4f2 e8e2 f2b6 e2e7 b6d4 b4b5 d4f2 h3g2 f2h4 e7e8');
INSERT INTO evaluations VALUES(28073,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(28073,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(28073,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(28073,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(28073,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(28073,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(28073,36,'cp 104','c1e3 b8d7 f1e2 g4f3 e2f3 e7e5 d4d5 c6c5 g2g4 g7h6 e3h6 g8h6 h2h4 f7f6 d1d2 h6f7 e1c1 g6g5 c1b1 h7h6 f3e2 d7f8 h4h5 f8d7 e2d3 e8g8 c3e2 a8b8 e2g3 g8h8 g3f5 b7b6 d3c2 a7a6 d2d3 f8e8 b1a1 b8b7');
INSERT INTO evaluations VALUES(29123,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(29123,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(29123,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(29123,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(29123,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(29123,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(29123,36,'cp 124','d1b3 g7g6 g1f3 f8g7 b3b7 b8d7 c3d5 f6d5 b7d5 e8g8 f1e2 f8e8 e1g1 e7e6 d5d6 e6e5 d4d5 a8b8 c1g5 f7f6 g5e3 b8b2 f3d2 g7f8 d6c6 d8c8 g2g4 f5c2 c4c5 d7c5 f1c1 c8c6 d5c6 c2d3 e2f3 e5e4 f3d1 e8c8 e3c5 f8c5 c1c5 b2d2 d1b3 g8g7 a1c1 g7f8 c6c7 f8e7');
INSERT INTO evaluations VALUES(29980,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(29980,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(29980,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(29980,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(29980,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(29980,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(29980,36,'cp 65','d2d3 c7c6 a2a3 b8d7 h3h4 h7h6 h4h5 g6h7 h1g1 a7a5 d1e2 b7b5 c4a2 f8e7 c3d1 b5b4 d1e3 d8c7 c1d2 b4a3 b2a3 d6d5 e4d5 e5e4 d5c6 c7c6 d3e4 f6e4 a2d5 c6f6 a1d1 a8c8 e3c4 e4d2 d1d2 e8g8 g4g5 h6g5 f3g5 f6a1 e2d1 a1c3 g5h7 g8h7');
INSERT INTO evaluations VALUES(29981,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(29981,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(29981,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(29981,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(29981,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(29981,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(29981,36,'cp 74','c7c6 a2a3 b8d7 f3h4 d6d5 h4g6 h7g6 e4d5 c6d5 c4d5 f6d5 c3d5 d8a5 d5c3 f8b4 c1d2 b4c3 d2c3 a5d5 f2f3 d7c5 d1e2 e8g8 e1c1 c5a4 e2e5 a4c3 e5c3 f8c8 c3e1 d5a2 e1b4 c8c6 c1d2 a2d5 c2c4 d5f3 h1f1 f3h3 b4e7 h3h2 d2c3 c6c7 e7e4 a8c8 c3b3');
INSERT INTO evaluations VALUES(31821,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(31821,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(31821,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(31821,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(31821,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31821,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(31821,36,'cp -1','c2c4 e7e6 e2e3 g8f6 a2a3 a7a5 b4b5 b8d7 g1f3 f8d6 f1e2 h7h6 d2d4 d5c4 e1g1 e8g8 b1d2 d7b6 d2c4 b6c4 e2c4 a8c8 d1e2 c7c5 d4c5 d6c5 f1d1 d8e7 c4b3 c5d6 f3e5 d6e5 b2e5 f6e4 a3a4 f8d8 d1d8 c8d8 a1d1 e4c5 d1d8 e7d8');
INSERT INTO evaluations VALUES(39366,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(39366,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(39366,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(39366,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(39366,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(39366,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(39366,36,'cp -6','b4c3 d2c3 d7d6 e2e4 c8e6 g1f3 h7h6 f1d3 d8f6 h2h3 b8d7 d1e2 g8e7 c1e3 g7g5 e1c1 a7a6 b2b3 e8c8 f3d2 f6g7 a3a4 f7f5 e4f5 e7f5 d2e4 f5e3 e2e3 c8b8 a4a5 d8f8 f2f3 f8f4 c1b2 h8f8 d3c2 d7f6 g2g3');
INSERT INTO evaluations VALUES(39383,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(39383,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(39383,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(39383,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(39383,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(39383,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(39383,36,'cp -16','d2c3 d7d6 e2e4 c8e6 g1f3 h7h6 f1d3 d8f6 h2h3 g8e7 d1e2 b8d7 c1e3 g7g5 e1c1 a7a6 b2b3 e8c8 f3d2 f6g7 d3c2 f7f5 e4f5 e7f5 d2e4 f5e3 e2e3 c8b8 f2f3 e6f5 c1b2 h8f8 d1d2 f5e4 c2e4 d7f6 a3a4 h6h5');
INSERT INTO evaluations VALUES(52320,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(52320,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(52320,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52320,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(52320,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(52320,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(52320,36,'cp 84','e7e6 e2e4 f8g7 f1d3 e6d5 e4d5 d8e7 d1e2 e7e2 e1e2 d7d6 b1c3 b8d7 g1f3 d7e5 f3e5 g7e5 h1e1 g8e7 e2f1 b7b6 a2a4 e8g8 c1h6 f8e8 e1e5 d6e5 a4a5 c8a6 c3e4 e8d8 h6g5 f7f5 g5e7 f5e4 e7d8 e4d3 d8g5 a6c4 a5b6 c4d5 a1a7');
INSERT INTO evaluations VALUES(54116,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(54116,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(54116,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(54116,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(54116,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(54116,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(54116,36,'cp -135','e6d5 c4b5 b8d7 b1c3 g8f6 g1f3 d5d4 d1e2 f8e7 c3e4 e8g8 e1g1 f6e4 e2e4 d7f6 e4e1 c8e6 f3e5 d8c7 a2a4 a7a6 b5e2 f6d5 d2d3 a8e8 e2f3 d5b4 e1d1 f7f6 e5c4 b7b5 c4a3 c7d7 a4b5 a6b5 c1d2 e7d6');
INSERT INTO evaluations VALUES(55589,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(55589,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(55589,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55589,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55589,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(55589,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55589,36,'cp 0','e5d4 d1d4 b8c6 d4e3 f8e7 b1c3 g8f6 f1c4 c8e6 c4e6 f7e6 g1f3 d6d5 e3e2 d5e4 c3e4 d8d5 e4g5 h7h6 g5e6 e8f7 e6c7 d5a5 c1d2 a5c7 e2c4 f7g6 f3e5 g6h7 c4d3 h7g8 d3c4');
INSERT INTO evaluations VALUES(56856,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(56856,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(56856,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(56856,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(56856,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(56856,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(56856,36,'cp 70','d2d4 d7d6 d1e2 c7c5 d4c5 d6e5 b1c3 f8c5 e2e5 c5d4 e5g3 g8e7 f1d3 b8c6 g1f3 d4g7 e1g1 e8g8 f1d1 d8a5 g3h4 e7f5 d3f5 a5f5 c1h6 f7f6 h6g7 g8g7 f3d4 c6d4 h4d4 f8f7 c3e4 e6e5 d4e3 f5f4 e3f4 e5f4 f2f3 f6f5 e4d6 f7e7 b2b4 c8e6 c2c4 b7b6 d1d2');
INSERT INTO evaluations VALUES(57218,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(57218,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(57218,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(57218,33,'cp -119',NULL);
INSERT INTO evaluations VALUES(57218,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(57218,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(57218,36,'cp -120','d7d5 d2d4 b8c6 d4e5 c6e5 h3f4 g8f6 c1d2 f8d6 d2c3 c7c6 f4d3 e5g6 g2g3 e8g8 f1g2 f8e8 c3f6 d8f6 b1d2 a7a5 e1g1 g6e5 d3e5 d6e5 c2c3 a5a4 a2a3 f6e7 d1c2 g7g6 e3e4 d5e4 d2e4 c8e6 a1e1 e6b3 c2c1 e5g7 e4d2 b3e6 d2f3 e7f6');
INSERT INTO evaluations VALUES(56474,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(56474,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(56474,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(56474,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(56474,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(56474,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(56474,36,'cp -111','d2d4 b8c6 d4e5 c6e5 h3f4 g8f6 c1d2 f8d6 d2c3 c7c6 f4d3 e5g6 g2g3 e8g8 b1d2 f8e8 c3f6 d8f6 f1g2 a7a5 a2a4 c8f5 e1g1 a8d8 d2b3 d6c7 d1d2 b7b6 b3d4 f5e4 a1e1 h7h5 f2f3 e4d3 d2d3 g6e5 d3f5 f6f5 d4f5 g7g6 f5d4');
INSERT INTO evaluations VALUES(7590,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(7590,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(7590,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(7590,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(7590,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(7590,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(7590,36,'cp 183','f3e5 f6e4 d1e2 d7d5 d2d3 d8d6 e5f3 c8a6 b1d2 e8g8 d2e4 d5e4 e2e4 f8e8 c1f4 d6d7 e4a4 a6b5 a4a5 c5b6 a5c3 a7a5 a2a4 b5a6 a1e1 c6c5 b2b3 a6b7 e1e8 a8e8 f1e1 h7h6 e1e8 d7e8 c3e5 e8d7 f3d2 f7f6 e5e2 d7d5 e2e4 d5e4 d2e4 g7g5 f4d2 g8f7 g1f1');
INSERT INTO evaluations VALUES(7582,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(7582,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(7582,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(7582,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(7582,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(7582,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(7582,36,'cp 189','f6e4 d1e2 d7d5 d2d3 d8f6 e5f3 c8a6 c2c4 e8g8 d3e4 a6c4 e2c2 c4f1 g1f1 c5b6 b1c3 f6g6 c2b1 g6d6 e4e5 d6e6 c1f4 f7f6 e5f6 e6f6 f4g3 a8e8 h2h3 h7h5 h3h4 f6e6 b1d3 e6g4 f3g5 g4f5 d3f5 f8f5 g5f3 c6c5 a1c1 c5c4 c3a4 e8e4 a2a3 c7c6 a4b6 a7b6');
INSERT INTO evaluations VALUES(8959,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(8959,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8959,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(8959,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(8959,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(8959,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(8959,36,'cp 34','e1g1 e8g8 a2a3 b4c3 b2c3 f6e4 c3c4 c6a5 c4c5 a5c4 d3c4 d5c4 f3d2 e4d2 c1d2 b7b6 c5b6 c7b6 d1f3 d8d5 f3g3 f7f5 d2c3 c8b7 f2f3 f8f7 e3e4 d5d7 g3e5 f5e4 f3e4 a8f8 c3b4 f7f1 a1f1 f8f1 g1f1 h7h6 b4d6 b7c6 h2h3 c4c3');
INSERT INTO evaluations VALUES(10900,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(10900,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(10900,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(10900,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(10900,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(10900,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(10900,36,'cp 19','d7d6 a2a3 e8g8 b1c3 c8e6 c3d5 a7a5 e1g1 e6d5 e4d5 c6b8 f1e1 c7c6 d5c6 b7c6 c2c3 b8d7 c4b3 f8e8 c3c4 d8c7 c1d2 a8b8 a1b1 e7f8 b3a4 d7c5 a4c2 g7g6 b2b4 a5b4 a3b4 c5e6 c2a4 f8g7 d2e3 f6d7 f3d2 d6d5 b4b5 c6b5 b1b5 b8b5 c4b5');
INSERT INTO evaluations VALUES(11412,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(11412,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(11412,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(11412,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(11412,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(11412,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(11412,36,'cp 88','b5c6 b7c6 d2d4 f6d7 b1d2 c6c5 d4e5 d7e5 b2b3 e5f3 d1f3 f8e7 c1b2 e8g8 f3g3 f7f6 d2c4 c8e6 a1d1 d8d7 b2c3 a8e8 c4e3 e7d8 h2h3 g8h8 d1d3 d7f7 f2f3 c5c4 d3d2 c4b3 a2b3 e6d7 g1h2 f7g6 g3g6 h7g6 e3d5 h8g8');
INSERT INTO evaluations VALUES(13643,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(13643,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(13643,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(13643,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(13643,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(13643,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(13643,36,'cp 49','e3e4 b7b5 e4e5 f6d7 a2a3 a7a5 f1e2 d7b6 d2f1 b5b4 a3b4 a5b4 f1g3 a8a1 d1a1 c8d7 e1g1 d8a8 g3h5 a8a1 f1a1 h7h6 a1a6 b6c8 g2g4 e8d8 g4g5 d8c7 g5h6 g7h6 h5f6 f8e7 e2d1 c7b7 a6a1 b7c7 f3e1 e7f6');
INSERT INTO evaluations VALUES(14931,30,'cp -367',NULL);
INSERT INTO evaluations VALUES(14931,31,'cp -362',NULL);
INSERT INTO evaluations VALUES(14931,32,'cp -339',NULL);
INSERT INTO evaluations VALUES(14931,33,'cp -378',NULL);
INSERT INTO evaluations VALUES(14931,34,'cp -365',NULL);
INSERT INTO evaluations VALUES(14931,35,'cp -382',NULL);
INSERT INTO evaluations VALUES(14931,36,'cp -355','d5c4 c2c3 d4d3 f6g7 d8e7 c1e3 f8g7 b1d2 c8e6 f3g5 e6d5 d1g4 c6e5 g4h3 h7h6 g5e4 h8g8 e4g3 d5e6 g3f5 e7f6 f5g7 g8g7 h3h6 f6h6 e3h6 g7g2 h6f4 e5g4 e1c1 e8c8 d1e1 g2f2 f4g3 f2g2 h1g1 g2g1 e1g1 b7b6 h2h4 g4f6 b2b3 f6d5 d2c4 d5c3');
INSERT INTO evaluations VALUES(20544,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(20544,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(20544,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(20544,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(20544,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(20544,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(20544,36,'cp -85','d2d3 g8f6 b1d2 f8d6 e2e4 d5d4 b2c1 a7a6 a2a3 c8e6 g2h3 e6h3 g1h3 d8d7 h3g1 b7b5 g1f3 e8g8 e1g1 f8e8 f3h4 d6f8 d1e2 g7g6 e2f3 f8g7 f3g2 a6a5 a1b1 a8b8 h4f3 d7c7 g1h1 g7f8 f3g1 b8c8 a3a4');
INSERT INTO evaluations VALUES(20925,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20925,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(20925,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(20925,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(20925,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(20925,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(20925,36,'cp 12','g5d2 c8f5 c2c4 e7e6 a2a3 d5c4 e2e3 f8d6 f1c4 g8f6 d2c3 e8g8 b1d2 c6e7 e1g1 e7d5 f1e1 d5c3 b2c3 f6e4 d2e4 f5e4 c4d3 e4d3 d1d3 c7c5 g2g3 b7b6 a3a4 d8c7 e1c1 f8c8 g1g2 c5d4 c3d4 c7b7 c1b1 a8b8 e3e4 c8c6 b1e1 b8c8 d4d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(22408,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(22408,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(22408,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(22408,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(22408,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(22408,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22408,36,'cp 87','g7g6 e2e4 f8g7 g1f3 g8f6 f1e2 e7e5 c1e3 e8g8 e1g1 d8e7 d1c2 e5d4 f3d4 f8e8 f2f3 c7c6 e3f2 a7a6 f2h4 d7c5 a1d1 e7c7 h4f2 a6a5 c2d2 a5a4 f1e1 f6h5 e2f1 c8e6 g2g3 c7a5 d4e6 c5e6 f1h3');
INSERT INTO evaluations VALUES(25928,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(25928,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(25928,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(25928,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(25928,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(25928,35,'cp -130',NULL);
INSERT INTO evaluations VALUES(25928,36,'cp -143','c5d4 b5c6 d4c3 c6d7 d8d7 b2c3 a8c8 e1g1 f8c5 c3c4 g8e7 c4d5 d7d5 d1d5 e7d5 c2c4 d5c3 c1d2 c3e4 a1b1 b7b6 d2b4 e8g8 h2h4 c5b4 b1b4 f8d8 f1c1 e4c5 b4b5 h7h6 g1f1 d8d3 f1e2 d3a3 c1c2 c8d8 h4h5 g8h7');
INSERT INTO evaluations VALUES(28399,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(28399,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(28399,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(28399,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(28399,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(28399,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(28399,36,'cp 69','g5f6 e7f6 c4d5 e6d5 d1b3 b8c6 e2e3 c6e7 b3b7 e8g8 c3a4 a8b8 b7a6 d8e8 a4c5 b8b2 f1d3 b2b6 a6a7 f5d3 c5d3 e8b5 a7a3 b6a6 a3c3 f8a8 a2a4 a6a4 a1a4 a8a4 e1g1 g7g6 h2h4 h7h5 f3e5 f6e5 d3e5 c7c6 f1c1 g8g7 c3d2 a4b4');
INSERT INTO evaluations VALUES(30406,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(30406,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(30406,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(30406,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(30406,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(30406,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(30406,36,'cp 55','c2c4 f8d6 f4d6 d8d6 d1b3 b7b6 c4c5 d6d8 c5b6 c7b6 f1e2 g8e7 e1g1 e8g8 f1c1 b8c6 h2h3 h7h6 b1d2 d8d6 c1c3 c6a5 b3a4 a5c6 a4d1 a6a5 a1c1 f8c8 d1b3 c6a7 f3e5 f7f6 e5d3 f5g6 d3f4 g6f7 c3c8 a8c8 c1c8 a7c8 b3a4');
INSERT INTO evaluations VALUES(34097,30,'cp 1018',NULL);
INSERT INTO evaluations VALUES(34097,31,'cp 1055',NULL);
INSERT INTO evaluations VALUES(34097,32,'cp 1055',NULL);
INSERT INTO evaluations VALUES(34097,33,'cp 1055',NULL);
INSERT INTO evaluations VALUES(34097,34,'cp 1055',NULL);
INSERT INTO evaluations VALUES(34097,35,'cp 1048',NULL);
INSERT INTO evaluations VALUES(34097,36,'cp 1063','e4c2 h8g8 d7d5 b1c3 c2f5 g8h7 b8c6 h7h8 f5e5 h8e5 c6e5 e2d4 c7c6 f1e2 g6g5 d4f3 e5f3 e2f3 f8e7 d2d4 f7f5 g2g3 e7f6 c3e2 a7a5 e1d2 e8f7 d2c3 a5a4 h2h3 f7g7 f3g2 a8a5 f2f3 c8d7 c1e3 a4a3 b2b4 a5a6 f3f4 g5f4 e2f4');
INSERT INTO evaluations VALUES(34099,30,'cp 1087',NULL);
INSERT INTO evaluations VALUES(34099,31,'cp 1082',NULL);
INSERT INTO evaluations VALUES(34099,32,'cp 1090',NULL);
INSERT INTO evaluations VALUES(34099,33,'cp 1086',NULL);
INSERT INTO evaluations VALUES(34099,34,'cp 1090',NULL);
INSERT INTO evaluations VALUES(34099,35,'cp 1094',NULL);
INSERT INTO evaluations VALUES(34099,36,'cp 1087','h8g8 d7d5 b1c3 c2f5 g8h7 b8c6 h7h8 f5e5 h8e5 c6e5 e2d4 c7c6 f1e2 f8g7 f2f4 e5c4 d4f3 c4d6 e2d3 e8f8 b2b3 b7b6 f3e5 c8b7 e1g1 d6f5 g2g3 f5d4 c3e2 d4e6 a2a4 a7a5 d3c2 a8e8 g1g2 e6c5 d2d4 c5e4');
INSERT INTO evaluations VALUES(34436,30,'cp -164',NULL);
INSERT INTO evaluations VALUES(34436,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(34436,32,'cp -181',NULL);
INSERT INTO evaluations VALUES(34436,33,'cp -192',NULL);
INSERT INTO evaluations VALUES(34436,34,'cp -223',NULL);
INSERT INTO evaluations VALUES(34436,35,'cp -225',NULL);
INSERT INTO evaluations VALUES(34436,36,'cp -259','b6b2 b1d2 c5c4 a1b1 b2a2 d2c4 d5c4 f3d2 a2a5 f1c4 g8e7 e1g1 b8c6 d2e4 e7g6 g2g3 f8e7 h2h4 e8g8 f2f4 c6b4 h4h5 b7b5 c4b3 g6h8 h5h6 b4d5 e3d2 a5b6 c2c3 h8g6 h6g7 f8d8');
INSERT INTO evaluations VALUES(34473,30,'cp -245',NULL);
INSERT INTO evaluations VALUES(34473,31,'cp -234',NULL);
INSERT INTO evaluations VALUES(34473,32,'cp -203',NULL);
INSERT INTO evaluations VALUES(34473,33,'cp -239',NULL);
INSERT INTO evaluations VALUES(34473,34,'cp -237',NULL);
INSERT INTO evaluations VALUES(34473,35,'cp -228',NULL);
INSERT INTO evaluations VALUES(34473,36,'cp -208','b1d2 c5c4 a1b1 b2a2 d2c4 d5c4 b1a1 a2b2 f1c4 f8b4 e1f1 b2c3 c4d3 c3c7 c2c3 b4e7 h2h4 g8h6 h1h3 h6g4 e3f4 h7h5 f3g5 c8d7 f1g1 d7c6 c3c4 b8d7 d4d5 e7c5 f4g3 d7e5 d5c6 a8d8 c6b7 c7b7 d1a4 b7d7 a4d7 d8d7 d3c2 e8g8 g3e5 g4e5 c2h7 g8h8 a1e1');
INSERT INTO evaluations VALUES(34474,30,'cp -250',NULL);
INSERT INTO evaluations VALUES(34474,31,'cp -242',NULL);
INSERT INTO evaluations VALUES(34474,32,'cp -246',NULL);
INSERT INTO evaluations VALUES(34474,33,'cp -251',NULL);
INSERT INTO evaluations VALUES(34474,34,'cp -244',NULL);
INSERT INTO evaluations VALUES(34474,35,'cp -264',NULL);
INSERT INTO evaluations VALUES(34474,36,'cp -248','c5c4 a1b1 b2a2 d2c4 d5c4 b1a1 a2b2 f1c4 f8b4 e1f1 c8d7 c4d3 d7c6 c2c4 b4c3 a1b1 b2a3 b1b3 a3a5 d1c1 c3b4 c1b2 b4e7 f1e2 b8d7 h1a1 a5d8 b3b7 c6b7 b2b7 a8b8 b7a7 g8h6 c4c5 b8b2 e3d2 e8g8 c5c6 d7b6 a7b7 b2d2 e2d2 b6d5 a1a7 h6f5');
INSERT INTO evaluations VALUES(36343,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36343,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36343,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(36343,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(36343,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36343,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36343,36,'cp 23','b1d2 c7c5 f1d3 f6h5 d4c5 h5f4 e3f4 e7c5 d2b3 c5b6 e1g1 b8c6 f1e1 d8c7 d1d2 c8d7 b3d4 c6d4 c3d4 c7d6 e1e2 g7g6 g2g3 g8g7 d2e3 b6a5 f3e5 h7h5 h2h4 a8c8 g1g2 a5d8 a2a3 d8f6 b2b4 d7a4 a1e1 d6c7 b4b5');
INSERT INTO evaluations VALUES(36342,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36342,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(36342,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(36342,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(36342,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(36342,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36342,36,'cp 16','c7c5 f1d3 f6h5 d4c5 h5f4 e3f4 d8c7 g2g3 a7a5 e1g1 a5a4 d3c2 c8d7 f3e5 d7b5 f1e1 e7c5 d2f3 b8d7 e5d7 b5d7 h2h4 f8c8 a2a3 c7b6 d1d2 d7c6 f3d4 c5d4 c3d4 c6b5 a1c1 c8c6 f4f5 e6f5 c2f5 g7g6 f5h3 c6c1 e1c1');
INSERT INTO evaluations VALUES(36346,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36346,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36346,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(36346,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36346,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(36346,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(36346,36,'cp 9','c7c5 b1d2 f6h5 d4c5 h5f4 e3f4 e7c5 d2b3 c5d6 g2g3 b8c6 e1g1 g7g6 h2h4 h7h5 f1e1 d8b6 f3e5 g8g7 d3c2 b6c7 d1e2 c8d7 e5d7 c7d7 e2d3 a8e8 g1g2 a7a5 a1d1 d7c7 d3e3 e8c8 a2a4 d6e7 b3d4 c6d4');
INSERT INTO evaluations VALUES(35836,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(35836,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(35836,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35836,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(35836,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35836,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(35836,36,'cp 18','b1d2 f6h5 d4c5 h5f4 e3f4 e7c5 e1g1 g7g6 d2b3 c5e7 g2g3 d8b6 h2h4 h7h5 f3e5 b8c6 f1e1 g8g7 a2a4 a7a6 d1c2 b6c7 b3d2 c8d7 d2f3 c6e5 f4e5 b7b5 a4b5 d7b5 c2d2 c7b6 d3b5 a6b5 g1g2');
INSERT INTO evaluations VALUES(38259,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(38259,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(38259,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(38259,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(38259,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(38259,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(38259,36,'cp 71','f3d4 g8f6 c1f4 b8c6 d1d2 c6d4 d2d4 a7a6 e1c1 e8g8 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 c4e6 f7e6 f4c4 b7b5 c4e2 a8e8 h1e1 e7f7 c1b1 c7c6 f2f3 f7c7 g2g3 c6c5 c3e4 c5c4 a2a3 f6e4 e2e4 h7h6 e4g6');
INSERT INTO evaluations VALUES(40065,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40065,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(40065,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(40065,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40065,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40065,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(40065,36,'cp 29','b1c3 h7h6 g5h4 e8g8 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1e2 c5e7 f3e5 a7a6 c3e4 b7b5 c4d3 c8b7 d1c1 a8a7 e4f6 e7f6 h4f6 g7f6 e5f3 b7d5 b2b3 f6f5 f3d4 g8g7 h1d1 a7d7 f2f3 f8d8 e3e4 d5b7');
INSERT INTO evaluations VALUES(40959,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40959,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40959,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40959,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40959,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40959,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40959,36,'cp 0','d5c4 d3c4 c7c5 e1g1 c5d4 e3d4 e8g8 b1c3 b8c6 a2a3 b7b6 d4d5 c6a5 c4a2 f6d5 c3d5 e6d5 a2d5 c8b7 d5b7 a5b7 c1e3 e7f6 e3d4 b7c5 a1c1 g7g6 b2b4 c5e6 d4e5 d8d1 f1d1 f6g7 d1d7 f8d8 c1d1 d8d7 d1d7 g7e5 f3e5');
INSERT INTO evaluations VALUES(41851,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(41851,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(41851,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(41851,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(41851,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(41851,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(41851,36,'cp -88','c5f2 e1f2 f6g4 f2g1 d8g5 c3d5 g5d8 h2h3 g4f6 d2d4 f6d5 c4d5 e5d4 d1d4 e8g8 g1h2 b8d7 h1f1 d7e5 c1f4 d8e7 f4e5 d6e5 d4c3 c7c6 d5c4 e7e8 a2a4 c8e6 c4e6 e8e6 a4a5 f8d8 b2b4 e6e7 a5a6 b7b6 c3c6');
INSERT INTO evaluations VALUES(43107,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(43107,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(43107,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(43107,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(43107,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(43107,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43107,36,'cp 54','e2e3 d7d6 g1f3 e8g8 f1e2 c7c6 g5h4 b8d7 d1b3 d8e8 e1g1 g8h8 a1d1 e7e5 b3c2 e5e4 f3d2 g7h6 h4g3 e8e7 b2b4 g6g5 c4c5 f5f4 c5d6 e7g7 b4b5 f4g3 h2g3 g7g6 b5c6 b7c6 c3e4 f6e4 d2e4 d7f6 e2d3 f6e4 d3e4 g6d6');
INSERT INTO evaluations VALUES(43105,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(43105,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(43105,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(43105,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(43105,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(43105,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43105,36,'cp 40','e8g8 g1f3 d7d6 f1e2 c7c5 d4d5 h7h6 g5f6 g7f6 d1d3 b8d7 e1g1 f6g7 a1e1 a8b8 d3c2 a7a6 a2a4 d8e8 f3d2 e8f7 h2h3 d7f6 e3e4 f6e4 d2e4 f5e4 c3e4 g7d4 g1h1 c8d7 e2d3 f7g7 f2f4 g6g5 f4g5 f8f1 e1f1 b8f8 g5h6 f8f1 d3f1');
INSERT INTO evaluations VALUES(43372,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43372,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43372,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43372,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43372,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(43372,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(43372,36,'cp -22','b7b5 c4d5 a8a7 d2d3 e7e6 d5b3 f8e7 c3e2 d8c7 c2c3 c5c4 b3c2 c4d3 d1d3 e8g8 h2h3 b8d7 e2g3 c8b7 c1f4 a7a8 d3e2 f8c8 a2a3 e6e5 f4d2 a6a5 f1e1 b5b4 a3b4 a5b4 a1a8 c8a8 c2d3 b4c3 d2c3');
INSERT INTO evaluations VALUES(43737,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(43737,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(43737,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(43737,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43737,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(43737,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(43737,36,'cp 0','c4d5 a8a7 d2d3 e7e6 d5b3 f8e7 c3e2 d8c7 c2c3 c5c4 b3c2 c4d3 d1d3 e8g8 e2g3 b8d7 d3e2 c8b7 h2h3 a7a8 c1d2 d7b6 a1d1 a8c8 c2b1 f8e8 f1e1 c7c4 e2e3 c4c7');
INSERT INTO evaluations VALUES(44771,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44771,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(44771,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44771,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44771,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44771,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(44771,36,'cp 0','f1e2 f8g7 e1g1 e8g8 f1e1 b8c6 a2a3 a7a5 a3a4 h7h6 c1f4 e7e5 f4g3 f6h5 b1a3 h5g3 h2g3 c8e6 f3d2 h6h5 e2f3 g7h6 d2c4 a8b8 a3b5 d8e7 c4b6 g8h8 b6d5 e7d8 d1e2 f8g8 a1d1 b8c8 d5e3 d8e7 e3d5');
INSERT INTO evaluations VALUES(44770,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(44770,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44770,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(44770,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(44770,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(44770,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(44770,36,'cp -10','f8g7 e1g1 e8g8 b1d2 b8c6 d3d4 c5d4 c3d4 d8b6 d4d5 c6d4 e2d3 d4f3 d2f3 e7e6 d5e6 c8e6 a2a4 a8c8 a4a5 b6c7 a5a6 b7a6 c1f4 a6a5 d1d2 f6d7 d2a5 c7a5 a1a5 d7c5 d3b1 g7b2 a5a7 e6c4 f1e1 f8d8 h2h4 c5e6 f4g3 d6d5 e4d5 d8d5 b1a2 b2c3 e1b1 c4a2 a7a2');
INSERT INTO evaluations VALUES(40358,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(40358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40358,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(40358,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(40358,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(40358,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(40358,36,'cp 0','e1g1 e8g8 h2h3 e7e5 a2a3 a7a5 c3c4 b8c6 b1c3 c8d7 c1d2 f6e8 c3d5 c6e7 d5c3');
INSERT INTO evaluations VALUES(40357,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(40357,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(40357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40357,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(40357,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(40357,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(40357,36,'cp -14','e8g8 f1e1 b8c6 a2a3 a7a5 a3a4 h7h6 c1f4 e7e5 f4g3 f6h5 f3d2 h5g3 f2g3 c8e6 d2c4 h6h5 b1a3 g7h6 e1f1 c6a7 d1b3 d8e7 b3d1 e7c7 g3g4 h5h4 d1b3 c7e7 b3d1 a7c8 d1e1 a8a6 e2f3 h6f4 e1e2 c8a7');
INSERT INTO evaluations VALUES(35966,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(35966,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(35966,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(35966,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(35966,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(35966,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(35966,36,'cp 0','f1e1 b8c6 a2a3 b7b5 d3d4 d8b6 d4d5 c6a5 e2f1 c8g4 c1f4 g4f3 g2f3 f6h5 f4e3 a5c4 e3c1 c4a5');
INSERT INTO evaluations VALUES(49027,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(49027,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(49027,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(49027,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49027,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(49027,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(49027,36,'cp 41','g1f3 f8d6 e1g1 e8g8 d2d4 e5d4 f3d4 c7c6 f1e1 d6c5 b1c3 d5c3 b2c3 f8e8 c1g5 e8e1 d1e1 d8f8 d4b3 c5a3 b3a5 b8d7 e1d2 h7h6 g5e3 d7e5 a1e1 c8g4 e3d4 e5f3 g2f3 g4f3 d2e3 f3g4 a5b7 f8c8 b7c5 a3c5 d4c5 c8a6 f2f3 g4f5 e3e2 a6c8 c5f2 f5e6 e1d1 e6a2');
INSERT INTO evaluations VALUES(52659,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(52659,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(52659,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(52659,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(52659,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(52659,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(52659,36,'cp 69','g1f3 g8f6 c4d5 c6d5 c1f4 e7e6 e2e3 b8c6 f1d3 f8d6 f3e5 e8g8 e1g1 d6e5 d4e5 f6d7 d1h5 f7f5 e5f6 d7f6 h5e2 e6e5 f4g3 c8e6 a1c1 d8e7 g3h4 a8d8 f2f4 e5e4 d3b1 g8h8 h2h3 a7a6 f1d1 e7e8 e2f2 c6e7 c3e2 e7f5 c1c7 e6d7');
INSERT INTO evaluations VALUES(53424,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(53424,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(53424,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(53424,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(53424,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53424,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53424,36,'cp 0','c8f5 f3h4 f5d7 h4f3');
INSERT INTO evaluations VALUES(31331,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31331,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31331,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31331,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31331,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31331,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31331,36,'cp 0','f1b5 b8c6 e1g1 e7e6 f3d4 g8e7 d4f5 e7f5 c2c3 f8d6 d2d4 h7h5 f1e1 g7g5 c3c4 d6c7 b1c3 d8d6 g2g3 e8c8 b5c6 b7c6 c4c5 d6e7 d1a4 c8b7 a4b4 b7a8 b4a4');
INSERT INTO evaluations VALUES(53615,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(53615,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(53615,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(53615,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(53615,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(53615,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(53615,36,'cp 111','g1f3 g8f6 f1d3 c6c5 e4e5 f6d7 c2c3 a7a5 e1g1 b8c6 d3c2 c5d4 c3d4 b7b5 f1e1 d7b6 d2f1 a5a4 f1g3 c6b4 g3h5 b6c4 c2b1 h7h6 d1e2 d8b6 c1e3 c8d7 e1c1 a8c8 h2h4 e8d8 a2a3 b4c6 b1d3 c6a5 a1b1 b5b4 a3b4 a5b3 c1c2 c4e3 f2e3 c8c2 e2c2 f8b4 h5g7');
INSERT INTO evaluations VALUES(54162,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(54162,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(54162,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(54162,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(54162,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(54162,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(54162,36,'cp 93','c4d5 e6d5 e4d5 a7a6 d2d4 g8f6 f1c4 b7b5 d1e2 f8e7 d4c5 e8g8 d5d6 e7d6 c5d6 b5c4 e1g1 d8d6 e2c4 c8g4 c1f4 d6b6 c4d4 b6b7 b1d2 b8c6 d4a4 g4e6 d2e4 f6e4 a4e4 h7h6 b2b3 a8c8 a1c1 f8d8 e4e3 e6f5 h2h3 f7f6 f4g3');
INSERT INTO evaluations VALUES(54135,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(54135,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(54135,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(54135,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(54135,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(54135,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(54135,36,'cp 87','e6d5 e4d5 a7a6 d2d4 g8f6 f1c4 b7b5 d1e2 f8e7 d4c5 e8g8 d5d6 e7d6 c5d6 b5c4 e1g1 d8d6 e2c4 c8g4 c4d4 d6c7 c1f4 c7b7 b1d2 b8c6 d4a4 f8c8 d2c4 g4f3 c4d6 b7b2 d6c8 a8c8 g2f3 c6d4 g1g2 f6d5 f4g5 h7h6 a1c1 c8f8 a4c4 h6g5 c4d5');
INSERT INTO evaluations VALUES(54532,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(54532,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(54532,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(54532,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(54532,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(54532,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(54532,36,'cp 102','e4d5 a7a6 b1c3 g8f6 d2d4 f6d5 f1d3 d5c3 b2c3 c5d4 e1g1 f8e7 c3d4 e8g8 f1e1 b8c6 a1b1 h7h6 d3e4 e7d6 h2h3 f8e8 b1b3 c8d7 b3b7 a8b8 b7b8 d8b8 c1d2 b8b2 d1b1 b2b1 e4b1 e8e1 d2e1 d7e6 b1e4 c6e7');
INSERT INTO evaluations VALUES(54493,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(54493,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(54493,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(54493,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(54493,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(54493,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(54493,36,'cp 83','a7a6 d2d4 g8f6 f1c4 b7b5 d1e2 f8e7 d4c5 e8g8 d5d6 e7d6 c5d6 b5c4 e1g1 d8d6 e2c4 c8g4 c4d4 d6c7 f3e5 b8c6 e5c6 c7c6 h2h3 g4e6 d4h4 c6e4 h4e4 f6e4 f2f3 e4g3 f1e1 g3f5 b1c3 f8d8 c3e4 h7h6 c1d2 f5d4 d2a5 d8c8 a5c3 d4c2');
INSERT INTO evaluations VALUES(55165,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(55165,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(55165,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(55165,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(55165,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(55165,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(55165,36,'cp -45','f7f6 e5f3 e7e5 d2d4 e5e4 f3d2 f6f5 c2c4 g8f6 b1c3 c7c6 c4d5 c6d5 d2b3 f6h5 c1d2 b8c6 a1c1 a7a6 g2g3 f8e7 e2e3 h5f6 c3e2 e8g8 b3c5 e7d6 e2f4 f8f7 d1b3 d6f4 g3f4 g8h8 f1e2 f7c7 e1g1 b7b6 c5a4 b6b5 a4c5 d8e7');
INSERT INTO evaluations VALUES(3484,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(3484,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(3484,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(3484,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(3484,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(3484,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(3484,36,'cp 73','d4d5 c6a5 b2b3 a8b8 e1g1 a7a6 c1g5 h7h6 g5d2 c7c5 d5c6 a5c6 a1c1 f8e8 c3d5 e7e5 d5f6 d8f6 d2c3 f6e7 d1d2 b7b5 f1d1 b5c4 b3c4 e5e4 c3g7 g8g7 f3d4 e4e3 f2e3 c6d4 d2d4 g7g8 e3e4 c8g4 c1c2 b8c8 d4d6');
INSERT INTO evaluations VALUES(4376,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(4376,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(4376,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(4376,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(4376,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(4376,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(4376,36,'cp 99','f3e5 c6e5 d2d4 c5d6 f2f4 e5c6 e4e5 d6b4 e5f6 d8f6 c3b5 b4a5 c2c3 c6e7 e2d3 c7c6 f4f5 e7d5 d1f3 a7a6 b5a3 b7b5 a3c2 f8e8 c2e3 d5e3 c1e3 b5b4 e3f4 b4c3 f4e5 f6h6 a1e1 c8b7 f5f6 g7g6 b2c3 a5c3 e1b1 b7c8 d3g6 e8e5 f3c3 h6e3 c3e3 e5e3 g6f5 d7d5 f5c8 a8c8');
INSERT INTO evaluations VALUES(6841,30,'cp -417',NULL);
INSERT INTO evaluations VALUES(6841,31,'cp -436',NULL);
INSERT INTO evaluations VALUES(6841,32,'cp -427',NULL);
INSERT INTO evaluations VALUES(6841,33,'cp -448',NULL);
INSERT INTO evaluations VALUES(6841,34,'cp -428',NULL);
INSERT INTO evaluations VALUES(6841,35,'cp -423',NULL);
INSERT INTO evaluations VALUES(6841,36,'cp -436','c5b4 c2c3 e5c4 c3b4 d7d5 e4e5 a7a5 b4b5 a5a4 e1g1 g8e7 b1c3 c4b6 c1g5 h7h6 g5e7 d8e7 f2f4 c8f5 a1c1 h6h5 h2h3 e7b4 f1f2 e8c8 b2b3 a4a3 c3e2 b4b5 e2g3 b5d3 d1d3 f5d3 h3h4 c7c6 f4f5 d8e8 c1e1 d3e4 e1e3 c8c7 f2f1 c7b8 f1f4 g7g6 g3e4 d5e4 f5g6 b6d5');
INSERT INTO evaluations VALUES(6998,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(6998,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(6998,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(6998,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(6998,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(6998,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(6998,36,'cp 100','c4b3 g8f6 f2f4 e5c6 b1c3 e8g8 c1e3 e7b4 e4e5 f6e4 e1g1 b4c3 b2c3 d7d5 d4c6 b7c6 c3c4 c8b7 d1f3 a7a5 a2a4 d8e7 h2h3 a8d8 g1h2 f8e8 f1d1 g8h8 c4d5 c6d5 c2c4 d5c4 d1d8 e7d8 b3c4 e4d6 c4d5 b7d5 f3d5 d6f5 d5d8 e8d8 e3g1 d8d2');
INSERT INTO evaluations VALUES(6999,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(6999,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(6999,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(6999,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(6999,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(6999,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(6999,36,'cp 81','g8f6 b1c3 e8g8 f2f4 e5c6 c1e3 e7b4 e4e5 f6e4 e1g1 e4c3 d1d3 d7d6 b2c3 b4c5 d4c6 c5e3 d3e3 b7c6 a1e1 c8f5 h2h3 d8e7 e5d6 e7e3 e1e3 c7d6 g2g4 f5e6 b3e6 a8e8 e6f7 g8f7 g1f2 e8e3 f2e3 f8b8 f1d1 f7e7 e3d3 b8b2 d1e1 e7d7 e1a1 g7g6 c3c4 b2b8 a2a3 b8f8 a1f1 h7h5 c4c5 d6c5 d3e4 f8e8 e4f3');
INSERT INTO evaluations VALUES(10977,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(10977,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(10977,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(10977,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(10977,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(10977,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(10977,36,'cp 95','b1c3 e8g8 f2f4 e5c6 c1e3 e7b4 e4e5 f6e4 e1g1 e4c3 d1d3 d7d6 b2c3 b4c5 d4c6 c5e3 d3e3 b7c6 a1e1 c8f5 h2h3 d8e7 e5d6 e7e3 e1e3 c7d6 g2g4 f5e6 b3e6 a8e8 e6f7 g8f7 g1f2 e8e3 f2e3 f8b8 f1d1 f7e7 e3d3 b8b2 a2a3 b2a2 d1b1 a2a3 b1b7 e7f8 b7c7 c6c5 f4f5 a7a5 d3c4 a3a2 c7d7 a2c2 d7d6 c2h2 c4c5 h2h3');
INSERT INTO evaluations VALUES(10978,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(10978,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(10978,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(10978,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(10978,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(10978,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(10978,36,'cp 71','e8g8 f2f4 e5c6 c1e3 e7b4 e4e5 f6e4 e1g1 e4c3 d1d3 d7d6 b2c3 b4c5 e5e6 c8e6 d4e6 c5e3 d3e3 f7e6 b3e6 g8h8 a1e1 d8f6 e6d5 c6a5 g2g4 f6h4 e3g3 h4g3 h2g3 c7c6 d5e4 d6d5 e4d3 a8e8 g1g2 h8g8 g4g5 g7g6 g2f3 e8e1 f1e1 g8f7 g3g4 f8e8 e1e8 f7e8');
INSERT INTO evaluations VALUES(4617,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(4617,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(4617,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(4617,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(4617,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(4617,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(4617,36,'cp 69','f2f4 e5c6 c1e3 e7b4 e4e5 f6e4 e1g1 e4c3 d1d3 d7d6 b2c3 b4c5 e5e6 c8e6 d4e6 c5e3 d3e3 f7e6 b3e6 g8h8 a1e1 d8f6 e6d5 c6a5 g2g4 f6h4 e3g3 h4g3 h2g3 c7c6 d5e4 a8e8 e4d3 b7b5 g1f2 g7g6 f1h1 e8e1 h1e1 h8g7 g4g5 a5b7 c3c4 b5c4 d3c4 d6d5 c4d3 b7c5 e1e5 c5d3 c2d3 g7f7 g3g4 f8b8 f4f5 b8b1 f2g2 b1b2 g2g3 b2a2');
INSERT INTO evaluations VALUES(9956,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(9956,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(9956,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(9956,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(9956,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(9956,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(9956,36,'cp -21','d7c6 d2d4 d6b4 e1g1 e5d4 f3d4 b4c3 b2c3 c6c5 d4f5 c8f5 e4f5 d8d1 f1d1 f6e4 c1f4 e4c3 d1d2 e8g8 f4c7 f8e8 g1f1 h7h6 c7a5 c3b5 a2a4 b5d4 a5c3 d4f5 a1b1 b7b6 a4a5 b6a5 d2d5 f5e7 d5d7 e7c6 b1b7 a8d8 d7d8 e8d8 b7c7 c6d4');
INSERT INTO evaluations VALUES(12997,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(12997,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(12997,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(12997,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(12997,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12997,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(12997,36,'cp 5','e1g1 f8g7 d2d4 c6d4 f3d4 c5d4 d1d4 e8g8 f1e1 c8d7 d4d3 a7a6 a2a4 a8c8 c1g5 d8a5 g5f6 g7f6 c3d5 f6b2 d5e7 g8g7 e7c8 b2a1 e1a1 f8c8 c4b3 a5e5 a1d1 d7e6 d3d6 e5e4 b3e6 e4e6 d6e6 f7e6 d1d7 g7h6 g2g4 c8c2 d7b7 c2a2 a4a5 a2a5');
INSERT INTO evaluations VALUES(12996,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12996,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(12996,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(12996,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12996,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12996,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12996,36,'cp 0','f8g7 f1e1 e8g8 d2d4 c6d4 f3d4 c5d4 d1d4 c8d7 d4d3 a7a6 a2a4 a8c8 c1g5 d8a5 g5f6 g7f6 c3d5 f6b2 d5e7 g8g7 e7c8 b2a1 e1a1 f8c8 c4b3 a5e5 a1d1 d7e6 d3d6 e5e4 d6e7 b7b5 d1d7 e4e1 g1h2 e1e5 g2g3 e5f5 e7e6 f5f2 h2h1 f2f1 h1h2');
INSERT INTO evaluations VALUES(8242,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(8242,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8242,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8242,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8242,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8242,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8242,36,'cp 0','d2d4 c5d4 f3d4 e8g8 f1e1 c6d4 d1d4 c8d7 d4d3 a7a6 a2a4 a8c8 c1e3 d8c7 c4b3 d7e6 e3d4 e6b3 c2b3 f8e8 e1e2 c7a5 a1d1 c8c6 e2d2 a5b4 c3a2 b4a5');
INSERT INTO evaluations VALUES(17791,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(17791,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17791,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(17791,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(17791,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(17791,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(17791,36,'cp 37','e1g1 f8e7 c2c3 g8f6 c4b3 e8g8 f1e1 c8e6 b1d2 e6b3 d1b3 d8d7 d2f1 a6a5 a2a4 h7h6 f1g3 f8d8 d3d4 e5d4 f3d4 d8b8 d4f5 e7f8 c1e3 b8e8 e3f4 c6e7 f5e7 f8e7 a1d1 d7c6 h2h3 f6d7 b3c2 d7c5 b2b3');
INSERT INTO evaluations VALUES(19184,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(19184,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(19184,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(19184,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(19184,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(19184,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(19184,36,'cp 115','g8e7 d4e5 d7d6 e5g3 e7g6 h2h4 h7h5 b1c3 f8e7 c1g5 e7g5 h4g5 h5h4 g3e3 c8d7 e1c1 d8b6 e3d2 b6c5 c1b1 c5e5 f1c4 e5f4 d2f4 g6f4 g2g3 f4h3 d1d6 h3g5 f2f4 d7e6 c4e6 g5e6 f4f5 a8d8 d6d8 e6d8 h1h4 h8h4 g3h4 e8e7 b1c2 d8c6');
INSERT INTO evaluations VALUES(19433,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(19433,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(19433,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(19433,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(19433,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(19433,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(19433,36,'cp -50','g7g6 g2g3 f8g7 f1g2 d7d6 g1e2 b7b5 a1b1 a8b8 e1g1 e7e6 c1b2 b5b4 c3a4 e6e5 f2f4 g8e7 f4f5 f7f6 c2c3 c8a6 c3c4 a6b7 g3g4 h7h5 g4h5 h8h5 e2g3 h5h4 g1h1 g6f5 g3f5 e7f5 f1f5 g7h6');
INSERT INTO evaluations VALUES(25261,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(25261,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(25261,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(25261,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(25261,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(25261,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(25261,36,'cp 75','g4h5 f1e1 d8d6 c4b5 h5f3 d1f3 f7f6 b1c3 e8c8 c3d5 d6d5 b5c6 d5c6 f3c6 b7c6 e1e4 f8c5 e4g4 g7g6 g4c4 d8d5 c1e3 c5e3 f2e3 c8d7 a1f1 h8f8 g1f2 d5a5 a2a3 f6f5 f2e2 d7d6 e2d2 h7h5 d2c3 a5c5 h3h4 c5c4 c3c4 a7a5 b2b3 f8b8 e3e4 b8b5 b3b4 d6e6 f1d1 f5f4 d3d4 e5d4');
INSERT INTO evaluations VALUES(25259,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(25259,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(25259,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(25259,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(25259,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(25259,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(25259,36,'cp 69','f1e1 d8d6 c4b5 h5f3 d1f3 f7f6 b1c3 e8c8 c3d5 d6d5 b5c6 d5c6 f3c6 b7c6 e1e4 f8c5 e4g4 d8d7 c1e3 c5b6 g1f1 h7h5 g4a4 c8b7 f1e2 b6d4 a1b1 g7g5 b2b3 d7d5 a4c4 h8g8 b1g1 g5g4 g1f1 f6f5 h3g4 h5g4 e3d4 e5d4');
INSERT INTO evaluations VALUES(26685,30,'cp 375',NULL);
INSERT INTO evaluations VALUES(26685,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(26685,32,'cp 392',NULL);
INSERT INTO evaluations VALUES(26685,33,'cp 386',NULL);
INSERT INTO evaluations VALUES(26685,34,'cp 389',NULL);
INSERT INTO evaluations VALUES(26685,35,'cp 387',NULL);
INSERT INTO evaluations VALUES(26685,36,'cp 393','d4d5 g4f3 g2f3 c6b8 d1e2 g7g6 c1f4 f8g7 c3b5 b8a6 e1c1 e8f8 h2h4 g8f6 f4e5 d8e8 e5c7 f6d5 d1d5 a6c7 b5c7 e8c6 e2c4 c6c4 f1c4 a8c8 d5c5 g7h6 c1b1 f8g7 c4d5 h6f4 c5c4 f4e5 c7b5 c8c4 d5c4 a7a6 b5c3 h8c8 c4d3 f7f5 h4h5 e5b8 c3e2 e7e5 h5h6 g7f8 h1d1 c8c7');
INSERT INTO evaluations VALUES(31608,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(31608,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(31608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31608,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(31608,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(31608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31608,36,'cp -1','e1g1 g8f6 f1e1 f8e7 c2c3 d6d5 d2d3 b8d7 e4d5 f6d5 h2h3 g4h5 d3d4 e5d4 c3d4 e8g8 b1c3 d7b6 a4b3 a7a5 a2a4 f8e8 c1d2 d5b4 c3e4 b4d3 e1e3 h5g6 e3d3 g6e4 d3e3 e4d5 d1c2 e7d6 e3e8 d8e8 a1e1 e8d8 b3d5 b6d5');
INSERT INTO evaluations VALUES(31802,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(31802,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(31802,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(31802,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(31802,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(31802,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(31802,36,'cp 195','b1c3 c7c6 g1f3 e6g4 f1c4 e7e6 c1e3 g8f6 h2h3 g4h5 c4d3 h5g6 d1b3 d8c7 a1c1 b8d7 e1g1 f8e7 f1d1 e8g8 e4e5 f6d5 c3d5 e6d5 d3g6 h7g6 b3d5 d7b6 d5b3 c7d7 a2a4 d7d5 b3c2 d5a2 c1a1 a2d5 a4a5 b6c4 c2c3 a8e8 h3h4 c4e3 f2e3');
INSERT INTO evaluations VALUES(32086,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(32086,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(32086,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(32086,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(32086,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(32086,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(32086,36,'cp 128','h5g6 d2d4 g8f6 d1e2 b8c6 d4d5 c6d4 f3d4 e5d4 c3b5 f8e7 f1g2 e8g8 b5d4 f8e8 e1g1 e7f8 c1g5 c7c6 d5c6 b7c6 d4c6 d8d7 g5f6 g7f6 c6d4 d6d5 f1d1 a8b8 b2b3 d5e4 d4f5 d7c7 f5e3 c7e5 e3c4 e5g5 e2e3 g8g7 e3g5 f6g5 c4e3 f8a3 d1d5 b8c8 a1d1');
INSERT INTO evaluations VALUES(32066,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(32066,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(32066,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(32066,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(32066,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(32066,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(32066,36,'cp 117','d2d4 g8f6 d1e2 b8c6 d4d5 c6d4 f3d4 e5d4 c3b5 f8e7 f1g2 e8g8 b5d4 f8e8 e1g1 e7f8 c1g5 c7c6 d5c6 b7c6 d4c6 d8b6 e4e5 d6d5 g5e3 b6c6 e5f6 c6f6 c2c3 g6e4 f1e1 f6c6 f2f3 e4g6 e2d2 e8e3 e1e3 f8c5 a1e1 a8e8 g1f1 c5e3 e1e3 e8e3 d2e3 h7h5 f1f2');
INSERT INTO evaluations VALUES(32065,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(32065,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(32065,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(32065,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(32065,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(32065,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(32065,36,'cp 135','g8f6 d1e2 b8c6 d4d5 c6d4 f3d4 e5d4 c3b5 f8e7 f1g2 e8g8 b5d4 f8e8 e1g1 e7f8 c1g5 c7c6 d5c6 b7c6 e2a6 g6e4 d4c6 d8c8 a6c8 a8c8 c6d4 e4c2 f1c1 c2e4 c1c8 e8c8 g5f6 e4g2 g1g2 g7f6 a1d1 h7h5 d1d2 f8h6 d4f5 h6d2 f5e7 g8f8 e7c8 h5g4 h3g4 d6d5 g2f3');
INSERT INTO evaluations VALUES(36904,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36904,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36904,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(36904,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(36904,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36904,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36904,36,'cp 0','b1c3 b8c6 c3b5 d6e7 c1f4 a7a6 b5c3 c6b4 f1e1 b4d3 d1d3 f8e8 e1e2 c7c6 a1e1 e7f8 f3e5 c8e6 f4g3 d8b6 f2f4 f6e4 c3e4 d5e4 e2e4 e6f5 d3c4 f5e6');
INSERT INTO evaluations VALUES(36900,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36900,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36900,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36900,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36900,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(36900,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36900,36,'cp 9','b8c6 c3b5 d6e7 c1f4 a7a6 b5c3 c6b4 f1e1 b4d3 d1d3 c7c6 e1e2 f8e8 a1e1 e7f8 f3e5 c8e6 c3a4 d8a5 d3b3 f6e4 e5d3 b7b5 a4c5 e6f5 c2c3 e4c5 d3c5 e8e2 e1e2 a5d8 b3d1 d8f6 f4g3 f6g6 g1h2 h6h5 a2a4 b5a4 d1a4 f8c5 d4c5');
INSERT INTO evaluations VALUES(37655,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(37655,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(37655,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(37655,33,'cp 272',NULL);
INSERT INTO evaluations VALUES(37655,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(37655,35,'cp 288',NULL);
INSERT INTO evaluations VALUES(37655,36,'cp 267','f3e5 d7d5 e4d5 c6d5 c4b5 c8d7 e5d7 b8d7 d2d4 c5e7 d1f3 g8f6 b1c3 a7a6 b5d3 d7b6 c3d1 e8g8 c2c3 b6a4 e1g1 f8e8 b2b3 a4b6 f1e1 d8c7 c1d2 e7d6 g2g3 c7c6 c3c4 f6e4 d2f4 d6f4 f3f4 e4g5 e1e8 a8e8 f4g5 d5c4 d3f1 c4c3');
INSERT INTO evaluations VALUES(39853,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(39853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39853,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(39853,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(39853,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(39853,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(39853,36,'cp -25','f1d3 b8c6 c1f4 h7h6 f4g3 a7a6 a2a3 c5a7 e1g1 d7c5 c3e2 c8d7 b2b4 c5d3 c2d3 c6e7 g3f4 e7g6 d1d2 d5d4 e2d4 d7a4 f4e3 a7d4 f3d4 g6e5 f1c1 e5c6 d4c6 a4c6 e3c5 f7f6 d2e3 e8f7 c5b6 d8d5 f2f3 d5g5 e3g5 h6g5');
INSERT INTO evaluations VALUES(39854,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(39854,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(39854,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(39854,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(39854,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(39854,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(39854,36,'cp -17','a7a6 c1g5 d8c7 d1e2 b8c6 g5f4 b7b5 e1g1 c8b7 f1e1 c5e7 a2a3 d7c5 b2b4 c5d3 e2d3 h7h6 h2h3 e8g8 c3e2 d5d4 e2d4 c6d4 f3d4 f8c8 e1d1 c7c3 f4e3 c8c4 d3c3 c4c3 d1d3 a8c8 d3c3');
INSERT INTO evaluations VALUES(40420,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(40420,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40420,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40420,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(40420,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40420,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(40420,36,'cp 24','h2h3 a7a6 f1d3 b8c6 a2a4 c6b4 d3b1 e8g8 c1d2 a6a5 e1g1 e7e6 f3e5 d8e7 d1b3 f8d8 c3e2 c8d7 d2b4 a5b4 e5d7 d8d7 b1d3 g7f8 f1c1 d7c7 g2g3 a8c8 b3d1 f6e4 c1c7 c8c7 a1c1 e7d8 g1g2 c7c1 e2c1 d8c7 d3e4 d5e4');
INSERT INTO evaluations VALUES(41294,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41294,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41294,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41294,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41294,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41294,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(41294,36,'cp 6','c8g4 b1c3 b8c6 d3e2 g4f3 e2f3 e7e5 d4d5 c6d4 c1g5 c7c5 d5c6 b7c6 b2b4 e8g8 e1g1 d8c7 a1b1 a8b8 d1d2 b8b7 f1c1 f8b8 b4b5 f6d7 f3d1 d7c5 g5e3 h7h5 d1c2 a7a6 a2a4 c7d8 c1d1 a6b5 a4b5 c5e6 b5c6 b7b1 c2b1 d4c6 d2d6');
INSERT INTO evaluations VALUES(42993,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(42993,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(42993,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(42993,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42993,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42993,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42993,36,'cp 16','c1f4 d7d5 e2e3 e8g8 b1c3 b7b6 c4d5 f6d5 c3d5 e6d5 f1d3 c7c5 d1f3 c5d4 e3d4 e7d6 f4e3 c8a6 d3a6 b8a6 g1e2 a6c7 e2c3 d8d7 e1g1 a8d8 g2g3 d6e7 a1c1 h7h6 h2h4 c7e6 f1d1 f8e8 c3e2 d7b5');
INSERT INTO evaluations VALUES(43291,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(43291,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(43291,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(43291,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(43291,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(43291,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(43291,36,'cp 75','d5b6 c4b3 d7d6 d1e2 d6e5 d4e5 c7c5 e1g1 c5c4 b3c4 b6c4 e2c4 b8c6 f1e1 c8e6 c4e2 e6g4 c1f4 d8b6 c2c3 e8g8 b2b3 a8d8 b1d2 g4f5 a2a4 f5d3 e2e3 b6e3 e1e3 a7a5 a1e1 d3c2 e1c1 c2f5 e3e1 d8d7 d2c4 f8d8 h2h3');
INSERT INTO evaluations VALUES(44557,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44557,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(44557,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44557,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44557,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(44557,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(44557,36,'cp -6','c5d4 d1d4 b8c6 d4a4 e7e6 g1f3 d8c7 c3b5 c7b8 f1d3 f8c5 c2c3 a7a6 b5d4 b7b5 a4c2 c8b7 e1g1 h7h6 f1e1 e8g8 h2h3 c6e5 f3e5 b8e5 d4b3 c5b6 c1e3 e5c7 c2e2 e6e5 a1d1 a8c8 e3b6 c7b6');
INSERT INTO evaluations VALUES(44634,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(44634,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44634,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44634,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44634,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(44634,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(44634,36,'cp -2','d1d4 b8c6 d4a4 e7e6 g1f3 d8c7 f1d3 a7a6 c3b5 c7b8 b5d4 c6b4 c1d2 b4d3 c2d3 f8e7 d2b4 e7b4 a4b4 d7d6 e1g1 e8g8 d4e2 a6a5 b4b6 f8e8 h2h3 e6e5 e2g3 a8a6 b6e3 d6d5 f1c1 c8e6 e3d2 a5a4 d3d4 f6e4 g3e4 d5e4 f3e5 f7f6 e5c4 b7b5 c4e3');
INSERT INTO evaluations VALUES(44621,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(44621,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44621,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(44621,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44621,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(44621,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(44621,36,'cp 24','d7d5 f3e5 c8f5 c1f4 b8c6 f1b5 f5d7 e5d7 d8d7 c2c3 e7e6 b1d2 f8d6 f4g5 a7a6 b5a4 d6e7 a2a3 e8g8 a4c2 h7h6 g5h4 c6a5 e1g1 b7b5 h4g3 e7d6 g3e5 d7e7 f2f4 a5c4 d2c4 b5c4 f4f5 e6f5 f1f5 d6e5');
INSERT INTO evaluations VALUES(45858,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(45858,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45858,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45858,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45858,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45858,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45858,36,'cp 0','b7b5 c4b3 f8e7 d2d4 b8d7 f1e1 e8g8 c1g5 d8c7 a2a4 b5b4 c3b1 h7h6 g5h4 e5d4 f3d4 d7c5 b1d2 c8g4 f2f3 g4d7 d2c4 f8e8 c4e3 c5b3 c2b3 d6d5 e3f5 d7f5 d4f5 d5e4 f5e7 e8e7 h4f6 g7f6 f3e4 a8d8 d1g4 g8f8 g4h4 f8g7 h4g4');
INSERT INTO evaluations VALUES(46028,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(46028,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(46028,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(46028,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(46028,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(46028,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(46028,36,'cp 84','f1d3 g7g6 g1f3 f8g7 b1c3 e6d5 e4d5 e8g8 h2h3 f8e8 c1e3 c8f5 e1g1 f5e4 c3e4 f6e4 d1c2 e4f6 a1e1 b8a6 a2a3 d8d7 e3d2 a6c7 e1e8 a8e8 b2b3 b7b5 f1b1 e8b8 b1e1 b8a8 d2a5 a8b8 g2g3 b5c4 b3c4 f6e8 a5c3 d7d8 c3g7');
INSERT INTO evaluations VALUES(47719,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(47719,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(47719,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(47719,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(47719,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(47719,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(47719,36,'cp 26','e1g1 f8e7 f1e1 e8g8 c1f4 c7c5 c2c3 b8c6 h2h3 f8e8 d4c5 e7c5 b1d2 e8e1 d1e1 c8d7 d2f1 a7a5 f4e3 c5e3 e1e3 d8c7 a1e1 a5a4 a2a3 a8e8 e3c5 e8c8 g1h1 b7b6 c5e3 c8e8 e3d2 f6e4 d2c2 c6e5 f3e5 c7e5 d3e4 d5e4');
INSERT INTO evaluations VALUES(47718,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(47718,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(47718,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(47718,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(47718,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(47718,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(47718,36,'cp 31','f8e7 c1f4 e8g8 c2c3 c7c5 b1d2 b8c6 d4c5 e7c5 h2h3 f8e8 f1e1 e8e1 d1e1 c8d7 d2f1 a7a5 f4e3 c5e3 e1e3 d8c7 a1d1 a5a4 a2a3 a8e8 e3c1 c6a5 f3d4 a5c6 d4b5 c7b6 f1e3 c6e5 b5d4 e5c6 d4f5 d7f5 e3f5 c6e7 f5e7 e8e7 c1d2 e7e8');
INSERT INTO evaluations VALUES(40728,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40728,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(40728,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40728,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40728,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40728,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40728,36,'cp 19','h2h3 e8g8 c1f4 c7c5 d4c5 e7c5 b1d2 b8c6 c2c3 f8e8 f1e1 e8e1 d1e1 c8d7 d2f1 a7a5 f4e3 c5e3 e1e3 a5a4 a1e1 d8c7 e3d2 a4a3 b2b3 c6d8 f1g3 d8e6 d3f5 a8e8 e1e3 d7c6 f5e6 e8e6 f3d4 e6e3 d2e3 c6d7 e3d3 c7e5');
INSERT INTO evaluations VALUES(47924,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(47924,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47924,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47924,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47924,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47924,36,'cp 0','d1b3 d8d7 c1g5 b8c6 g5f6 g7f6 g1e2 c6a5 b3b4 a5c6');
INSERT INTO evaluations VALUES(50215,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(50215,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50215,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(50215,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(50215,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50215,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(50215,36,'cp 0','d7d5 c2c4 c7c5 c4d5 f8g7 f1c4 f6d5 d4c5 c8e6 g1f3 d8a5 c1d2 a5c5 d1b3 e8g8 b1a3 b8c6 e1g1 a8d8 f3g5 d5c7 g5e6 c7e6 a1d1 e6g5 d2c3 g7c3 b3c3 c5f5 c4d3 f5e6 d3c4');
INSERT INTO evaluations VALUES(46795,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(46795,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(46795,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46795,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46795,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46795,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46795,36,'cp 8','c2c4 c7c5 c4d5 f8g7 f1c4 f6d5 d4c5 d8a5 c1d2 a5c5 d1b3 c8e6 g1f3 e8g8 b1a3 b8c6 e1g1 a8d8 f3g5 e6c8 a1c1 e7e6 f1d1 c6a5 b3c2 a5c4 a3c4 b7b6 g5f3 d5b4 d2b4 d8d1 c2d1 c5b4 b2b3 c8a6 f3e5 a6c4 e5c4 f8c8 d1d7 b4c5 c1d1 c5c7 d7c7');
INSERT INTO evaluations VALUES(51759,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(51759,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(51759,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51759,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(51759,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(51759,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(51759,36,'cp 50','e1g1 g8f6 f1e1 b8c6 c4b5 d7d6 e4e5 f6d7 e5d6 f8d6 b5c6 b7c6 d2d3 e8g8 b1d2 d6c7 b2b3 f8e8 d2c4 d7b6 f3e5 b6c4 e5c4 d8d5 c1b2 e6e5 c4e3 d5d7 d1h5 f7f6 a1d1 e8d8 b2c3 d7f7 h5f3 c7b6 a2a4 c8d7 e3f1 b6c7 f1g3');
INSERT INTO evaluations VALUES(85,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(85,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(85,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(85,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(85,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(85,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(85,36,'cp 0','c1e3 g8e7 b1c3 c5d6 f2f3 c8d7 a1d1 d6b4 d4e2 e7g6 e1f2 e8c8 g2g3 g6e5 e3d4 b4d6 c3a4 b7b6 d4b6 d7g4 f3g4 c7b6 a4b6 c8c7 b6a4 e5g4 f2f3 g4e5 f3f2');
INSERT INTO evaluations VALUES(84,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(84,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(84,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(84,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(84,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(84,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(84,36,'cp 2','g8e7 b1c3 c5d6 f2f3 c8d7 a1d1 e8c8 d4b3 h7h5 e3c5 d7e6 c5d6 c7d6 h2h4 c8c7 e1f2 b7b5 c3e2 e7g6 b3d4 e6c8 e2g3 g6e5 b2b3 a6a5 g3e2 h8e8 e2f4 g7g6 h1e1 b5b4 d4e2 c8d7 f4d3 f7f5');
INSERT INTO evaluations VALUES(3691,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(3691,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(3691,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(3691,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(3691,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(3691,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3691,36,'cp 24','d8d6 g3g4 g7g6 g4g5 f6g4 d3e2 f7f6 g5f6 e6e5 d4e5 c6e5 f3e5 g4e5 f2f4 e5d7 c3c4 d7f6 b1c3 c8e6 c4d5 f6d5 c3d5 e6d5 e1f2 d6e6 d1d4 c7c6 b2b3 a8d8 d4c3 f8f7 a1d1 h7h5 h1h3 d8e8 d1d2 e6e7 d2d4 h5h4 e2f3 g6g5 f3d5 c6d5 d4d5');
INSERT INTO evaluations VALUES(10283,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(10283,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(10283,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(10283,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(10283,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(10283,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(10283,36,'cp 136','d2d4 c5b6 c1f4 d7d6 d1d3 c6a5 c4b5 c7c6 b5a4 b6c7 b2b4 b7b5 c3b5 c6b5 a4b5 c8d7 b4a5 c7a5 f4d2 e8g8 b5d7 a5d2 f3d2 d8d7 e1g1 a8b8 h2h3 d7a4 a2a3 h7h6 a1b1 b8b1 f1b1 f8e8 b1e1 a4c6 e1e3 c6a4 c2c4 e8b8 e4e5 d6e5 e3e5');
INSERT INTO evaluations VALUES(10290,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(10290,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(10290,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(10290,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(10290,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(10290,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(10290,36,'cp 144','c5b6 c1f4 e8g8 d4d5 f8e8 d5c6 f6e4 e1f1 e4c3 b2c3 d8f6 d1d2 d7c6 f4g5 f6f5 c4d3 f5c5 a1e1 e8e1 f1e1 c8e6 g5f4 e6a2 e1d1 f7f6 h1e1 b6a5 e1e3 g7g6 f4h6 a2f7 d3e2 a8e8 d2d3 e8e3 h6e3 c5d6 d3d6 c7d6 c3c4');
INSERT INTO evaluations VALUES(14271,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(14271,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(14271,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(14271,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(14271,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(14271,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(14271,36,'cp -10','g7g6 c1f4 f8g7 h2h4 e8g8 e2e3 c6b4 d3d2 c8f5 a1c1 a8c8 a2a3 b4c6 c1d1 f6e4 c3e4 d5e4 f3e5 c6e5 d4e5 d8d2 d1d2 h7h5 f1c4 c7c6 a3a4 f5g4 c4e2 g4e6 f2f3 c6c5 c2c4 e4f3 g2f3 f8d8 b2b3 d8d2 e1d2 c8d8 d2c1 b7b6 h1d1 d8d1 c1d1');
INSERT INTO evaluations VALUES(22335,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(22335,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(22335,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(22335,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(22335,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(22335,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(22335,36,'cp -20','d8d4 e5f3 d4e4 d1e2 c8d7 d2d3 e4e2 e1e2 e8c8 h1e1 d8e8 e2f1 e8e1 f1e1 c6c5 b1a3 g8e7 a3c4 e7c6 c1f4 h7h6 f3e5 g7g5 e5d7 c8d7 c4e5 c6e5 f4e5 h8g8 e1d2 f8d6 a1e1 g8e8 e5d6 e8e1 d2e1 c7d6 g2g3 d6d5 c2c4 d7d6 f2f4 g5g4 e1d1 h6h5 d1c2 a7a6 c2d2 b7b5');
INSERT INTO evaluations VALUES(1773,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(1773,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(1773,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(1773,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(1773,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(1773,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(1773,36,'cp -9','e5f3 d4e4 d1e2 c8d7 d2d3 e4e2 e1e2 e8c8 h1e1 d8e8 e2f1 e8e1 f1e1 c6c5 b1d2 g8e7 b2b3 e7c6 c1b2 c6b4 e1d1 h8g8 h2h4 f8e7 a2a3 b4c6 d1e2 g7g6 e2f1 g8e8 a1e1 c6d4 f3d4 c5d4 b2d4 e7h4 d4e5 h7h5 d2f3 h4e7 e5b2 e7d6 f3e5 e8e6 f2f4');
INSERT INTO evaluations VALUES(24613,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24613,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24613,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(24613,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(24613,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24613,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(24613,36,'cp 16','b1d2 e7e6 h2h3 g4h5 d1b3 d8c8 f3h4 c8d7 d4c5 f8c5 g2g4 h5g6 e1c1 a7a6 h4g6 h7g6 f1e2 b7b5 h3h4 c5d6 f4g5 f6h7 c1b1 h7g5 h4g5 e8e7 h1h8 a8h8 e3e4 h8h3 b3c2 d5d4 d2f3 d4d3 d1d3 h3h1 d3d1 h1d1 e2d1 d7c7 a2a4 d6f4 c2d3 c6e5 f3e5');
INSERT INTO evaluations VALUES(24612,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(24612,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(24612,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(24612,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(24612,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(24612,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(24612,36,'cp 20','e7e6 d1b3 d8c8 h2h3 g4h5 f3h4 c8d7 d4c5 f8c5 g2g4 h5g6 h4g6 h7g6 f1e2 a7a6 e1c1 b7b5 h3h4 c5d6 f4g5 f6h7 c1b1 d7a7 d2f3 h7g5 f3g5 d6e7 g5f3 a8d8 b3c2 a7c7 h4h5 g6h5 h1h5 h8h5 g4h5');
INSERT INTO evaluations VALUES(24705,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24705,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(24705,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24705,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24705,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(24705,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(24705,36,'cp 19','f1b5 c5d4 e3d4 f8d6 f4d6 d8d6 e1g1 e8g8 h2h3 g4h5 f1e1 a8c8 a2a4 f6d7 e1e3 a7a6 b5d3 d6c7 d2b3 f8e8 d1c2 h5g6 d3g6 h7g6 c2e2 c7b6 b3d2 c6e7 a4a5 b6b2 a1b1 b2a3 b1b7 d7f6 e2a6 c8a8');
INSERT INTO evaluations VALUES(24643,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(24643,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24643,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(24643,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24643,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(24643,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24643,36,'cp 20','d8c7 h2h3 g4h5 g2g3 e7e6 c1f4 f8d6 f4d6 c7d6 b1d2 e8g8 f1e1 c6e7 a2a4 d6b6 d1c2 h5g6 a4a5 g6d3 c2d3 b6c7 f3e5 a8c8 d3e2 h7h6 h3h4 e7f5 g1g2 f5d6 g3g4 f6e4 f2f3 e4d2 e2d2 d6c4 e5c4 d5c4 d2e3 g8h8');
INSERT INTO evaluations VALUES(24856,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(24856,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(24856,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(24856,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(24856,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(24856,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(24856,36,'cp 43','f1b5 f6d7 e1g1 e7e6 b1d2 f8d6 d1e2 e8g8 e3e4 d5e4 d2e4 h7h6 f1d1 f8e8 h2h3 e6e5 e4d6 c7d6 d4e5 d6e5 c1e3 a7a6 b5a4 d8c7 f3d2 d7f6 d2c4 f5e6 c4b6 a8d8 a4c2 d8d1 a1d1 e8d8 a2a3 c6e7 d1d8 c7d8 f2f3 e7f5');
INSERT INTO evaluations VALUES(25515,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(25515,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(25515,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(25515,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(25515,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(25515,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(25515,36,'cp 21','e1g1 h7h6 d3d4 e5d4 f3d4 c6d4 b5d7 d8d7 d1d4 f8e7 c2c4 e8g8 b1c3 c7c6 c1f4 a7a6 f1e1 a8d8 d4b6 f8e8 a1d1 d7e6 b2b3 d8d7 b6d4 d7d8 f4h2 b7b5 f2f3 b5c4 d4c4 e6c8 g1h1 e7f8 c4d3 d6d5 e4e5 f6d7 d3c2');
INSERT INTO evaluations VALUES(25514,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(25514,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(25514,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(25514,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25514,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(25514,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(25514,36,'cp 28','h7h6 d3d4 c6d4 b5d7 d8d7 f3d4 e5d4 d1d4 f8e7 c2c4 e8g8 b1c3 a7a6 c1e3 c7c6 b2b3 b7b5 a1c1 d7e6 f2f3 b5c4 b3c4 a8b8 c3e2 c6c5 d4c3 b8b4 e2f4 e6d7 f4d3 b4b7 c3a3 d7c6 d3f4 f8b8 f1d1 c6c8 e3d2 g7g5 f4e2 f6h5 d2c3 b7b6');
INSERT INTO evaluations VALUES(31711,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(31711,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(31711,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(31711,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(31711,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(31711,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(31711,36,'cp -35','b8c6 h2h3 g4h5 c1f4 f8d6 f4d6 d8d6 c3b5 d6b8 e1g1 g8f6 c2c3 e8g8 f3e1 h5g6 e1d3 b8d8 b5a3 a8c8 f1e1 h7h6 a3b1 d8b6 d1b3 b6c7 b1d2 g6h7 e2f1 b7b6 d3e5 f6e4 d2e4 c6e5 f1a6 c8b8 e4d2 e5d3 a6d3 h7d3 h3h4');
INSERT INTO evaluations VALUES(35874,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(35874,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35874,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(35874,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35874,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35874,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35874,36,'cp 0','d5c4 e2e4 b7b5 d1e2 b5b4 c3d1 c8a6 f1e1 e7e6 a2a3 b4b3 d1c3 b8d7 c1f4 d8b6 e2d2 c6c5 d4d5 a8d8 c3a4 b6b5 a4c3');
INSERT INTO evaluations VALUES(35887,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35887,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35887,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35887,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35887,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35887,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35887,36,'cp 0','e2e4 b7b5 d1e2 b5b4 c3d1 c8a6 f1e1 e7e6 a2a3 b4b3 d1c3 b8d7 c1f4 d8b6 f4d6 c6c5 d6f8 a8f8 d4d5 f6g4 e2d2 g7d4 c3a4 b6b5 a4c3');
INSERT INTO evaluations VALUES(36163,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(36163,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36163,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36163,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36163,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36163,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36163,36,'cp 0','a2a3 b8c6 b2b4 d5d4 c3b5 c5b6 b5d4 c6d4 f3d4 c8g4 f1e2 g4e2 d1e2 b6d4 e3d4 a7a5 b4a5 f8e8 c1e3 d8a5 e2d2 a5a6 d2e2');
INSERT INTO evaluations VALUES(37423,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(37423,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(37423,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(37423,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(37423,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37423,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37423,36,'cp 0','d7d6 g1f3 b8c6 f1e2 f6d7 d4d5 c6e5 f3e5 d7e5 e1g1 e7e6 d5e6 c8e6 d2f3 d8e7 f3d4 e6d7 e3e4 e5c6 e2f3 c6e5');
INSERT INTO evaluations VALUES(42787,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(42787,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(42787,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(42787,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(42787,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42787,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42787,36,'cp 33','e8g8 g1e2 d7d5 e4d5 f6d5 e1g1 b8c6 c4b3 f8e8 c1b2 d8d6 d1d2 d5f4 e2f4 e5f4 f1e1 c8d7 d3d4 c6a5 c3c4 b7b6 f2f3 e8e1 a1e1 a8e8 e1e8 d7e8 b2c1 d6e7 h2h3 g7g5 c1b2 f7f6 b2c3 a5b3 c2b3 e7e3 d2e3 f4e3 d4d5 g8f7 g2g4 e8d7');
INSERT INTO evaluations VALUES(37276,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(37276,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(37276,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(37276,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(37276,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(37276,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(37276,36,'cp 29','g1e2 d7d5 e4d5 f6d5 e1g1 b8c6 f1e1 c8e6 c4b3 f8e8 c1b2 f7f6 e2g3 d8d7 g3e4 c6a5 c3c4 d5f4 b2c1 f4g6 e4c5 d7c6 c5e6 a5b3 a2b3 c6e6 d3d4 a7a5 f2f4 a5a4 f4e5 f6e5 d4d5 e6b6 g1h1 a4b3 a1a8 e8a8 c2b3 b6d4 c1e3 d4d1');
INSERT INTO evaluations VALUES(48624,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(48624,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(48624,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(48624,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(48624,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(48624,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48624,36,'cp 0','h7h6 g5h4 c7c5 d1e2 c8e6 g1f3 b8c6 c2c3 c5d4 f3d4 c6d4 c3d4 f8b4 b1c3 e8g8 e1g1 f8e8 f1e1 a8c8 e2f3 b4e7 h4g3 e6g4 f3f4 g4d7 f4f3');
INSERT INTO evaluations VALUES(50491,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(50491,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(50491,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(50491,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(50491,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(50491,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(50491,36,'cp -44','g8f6 e4e5 d7d5 c4e2 f6d7 c2c3 c5c4 d2d4 c4d3 d1d3 d8c7 c1f4 b8c6 e1g1 f7f6 c3c4 f6e5 f4g3 e5e4 d3d2 f8d6 c4d5 e6d5 b1c3 e4f3 c3d5 c7a5 d2a5 c6a5 e2f3 d6g3 f2g3 e8g8 d5e7 g8h8 f3g4 d7f6 g4c8 a8c8 e7c8 f8c8 a1e1 a5c6 g3g4 h7h6 h2h3 h8g8 f1f4 f6d7 e1d1 c6e5 f4f2 d7f6');
INSERT INTO evaluations VALUES(54615,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(54615,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(54615,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(54615,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(54615,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(54615,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(54615,36,'cp -106','d5d4 c3b1 b8c6 d2d3 f8d6 g1f3 g8e7 a2a3 c8e6 b1d2 h7h6 b2b4 e8g8 e1g1 e7d5 c1b2 f7f6 d2e4 d6e7 a1c1 a7a5 b4b5 c6a7 a3a4 a8c8 c1c4 b7b6 d1a1 d5b4 c4c8 a7c8 f1c1 d8d7 b2a3 b4a2 a3e7');
INSERT INTO evaluations VALUES(55854,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(55854,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(55854,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55854,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55854,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(55854,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55854,36,'cp -14','b7b6 g1f3 g8f6 c2c4 c8b7 f1d3 d5c4 d3c4 c7c5 e1g1 c5d4 f3d4 b8d7 b2b3 f8c5 c1b2 e8g8 c4e2 a8c8 b1d2 d8e7 e2f3 f6d5 d2c4 f8d8 d1e2 d7f8 g2g3 f8g6 f3g2 c5d4 b2d4 d5f6 g2b7 e7b7 d4f6 g7f6');
INSERT INTO evaluations VALUES(55945,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55945,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55945,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55945,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(55945,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55945,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(55945,36,'cp -5','f8e7 g1f3 g8f6 e1g1 e8g8 c2c4 c7c5 c4d5 e6d5 d3d4 b8c6 b1c3 c5c4 f3e5 c8e6 e5c6 b7c6 b2b3 d8a5 c1d2 a5a6 e2e3 e6f5 b3c4 a6c4 d1a4 c4a4 c3a4 e7a3 d2c1 a3e7');
INSERT INTO evaluations VALUES(56048,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(56048,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(56048,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(56048,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(56048,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(56048,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(56048,36,'cp -104','f8b4 f3e4 b4c3 b2c3 d8h4 e1e2 h4e4 e2f2 g8f6 f1d3 e4c6 d1e1 e8g8 g1f3 b8d7 h1f1 e6e5 d4e5 c6c5 f2g3 f8e8 c1e3 c5f8 e5f6 d7f6 a1d1 g7g6 f3e5 f6h5 g3f2 e8e5 e1d2 c8g4 d1e1 g4f5 f2g1 a8d8 e3g5 f8c5 g5e3 c5c6 e3a7 e5e1 d2e1 f5d3 e1e7 c6d7 e7d7 d8d7 c2d3 d7d3');
INSERT INTO evaluations VALUES(729,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(729,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(729,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(729,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(729,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(729,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(729,36,'cp 43','e1g1 c5d4 c3d4 a7a5 f1e1 f8e7 h2h3 a5a4 a2a3 c6d4 f3d4 b6d4 d2f3 d4b6 d3c2 d7c5 d1d4 c8d7 d4g4 g7g6 a1b1 h7h6 c1e3 a8c8 g4f4 b6b5 f3d4 b5a6 e1c1 a6a5 h3h4 a5d8 d4f3 h6h5 c1e1 d7b5 b1c1 d8d7 e3d2 e8g8 f4h6 c5d3 c2d3 b5d3 f3g5 c8c1 e1c1');
INSERT INTO evaluations VALUES(728,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(728,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(728,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(728,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(728,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(728,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(728,36,'cp 36','c5d4 c3d4 a7a5 h2h3 f8e7 f1e1 a5a4 a1b1 c6b4 d2f1 d7f8 c1g5 b4d3 d1d3 e7g5 f3g5 f8g6 d3a3 b6d8 g5f3 c8d7 e1c1 d8e7 g2g3 e7a3 b2a3 d7c6 h3h4 e8g8 f1e3 f8c8 f3e1 g6e7 f2f4 b7b5 g1f2 h7h5 c1c5 c6d7 e1f3 d7e8 b1c1 g8f8 c5c8 a8c8 c1c8');
INSERT INTO evaluations VALUES(740,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(740,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(740,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(740,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(740,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(740,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(740,36,'cp 43','c3d4 a7a5 f1e1 f8e7 h2h3 a5a4 a2a3 c6d4 f3d4 b6d4 d2f3 d4b6 c1e3 d7c5 a1c1 c8d7 c1c3 b6a5 d3h7 g7g6 e3c5 e7c5 h7g6 f7g6 d1c2 b7b6 c2g6 e8d8 e1c1 h8e8 g6g5 e8e7 g5f6 d7e8 f3g5 d8d7 c1d1 a8c8 g5e6 c5f2 g1f2 e7e6 f6g7 d7d8 g7b7 e8c6 b7f7 a5b5 f2g1 e6e5 f7f6 e5e7');
INSERT INTO evaluations VALUES(5908,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(5908,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(5908,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(5908,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(5908,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(5908,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(5908,36,'cp 38','d8e7 b1c3 g8f6 c1g5 d6b4 d1d2 b4c3 d2c3 e7e4 f1e2 e8g8 g5f6 g7f6 f3d2 e4e7 d2c4 f8d8 c3e3 c8f5 e3e7 c6e7 c4e3 g8g7 e3f5 e7f5 a1d1 f5d4 e2d3 d8d5 e1d2 a8d8 d2c1 f6f5 c2c3 d4b3 a2b3 d5d3 d1d3 d8d3 c1c2 d3d7 b3b4 f7f6 f2f4 g7f7');
INSERT INTO evaluations VALUES(6607,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(6607,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(6607,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(6607,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(6607,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(6607,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(6607,36,'cp -14','b4c3 d2c3 g8f6 g1f3 d8e7 d1e2 e8g8 g2g3 d7d5 f1g2 c8g4 e1g1 a8d8 h2h3 g4f3 g2f3 d5e4 f3e4 f6e4 e2e4 e7e6 e4g2 e6d7 a1e1 f7f6 f2f4 e5f4 f1f4 f8e8 e1e8 d8e8 f4e4 g8f7 g3g4 h7h6 g2f3 e8e4 f3e4 d7e6 g1g2 e6e4 d3e4');
INSERT INTO evaluations VALUES(11978,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11978,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(11978,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(11978,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11978,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11978,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11978,36,'cp 0','d7d5 e1g1 f8e7 d1e2 e8g8 f1e1 d8c7 c1g5 h7h6 g5d2 a7a5 e4d5 c6d5 c3b5 c7c6 e2f3 a5a4 f3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(12067,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12067,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12067,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12067,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12067,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12067,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12067,36,'cp 0','e1g1 f8e7 c1f4 e8g8 d1e2 f8e8 a1e1 a7a5 e4e5 f6d7 c3a4 d7b6 a4b6 d8b6 d3h7 g8h7 e2h5 h7g8 e1e3 g7g6 e3g3 b6d8 g3g6 f7g6 h5g6 g8h8 g6h5 h8g8');
INSERT INTO evaluations VALUES(16460,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(16460,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(16460,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(16460,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(16460,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(16460,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(16460,36,'cp -96','d7d5 c4d5 f6d5 a2a3 d5b6 b2b4 d8d3 g1f3 c8e6 b1c3 f7f6 d1e2 e8c8 e1g1 f8e7 e2d3 d8d3 h2h4 h8d8 f3e1 d3d7 d2d3 c8b8 a1b1 e6f5 c3e4 b6a4 c1b2 a7a6 e4c3 a4c3 b2c3 f5d3 e1d3 d7d3 f1c1');
INSERT INTO evaluations VALUES(17151,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(17151,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(17151,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(17151,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(17151,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(17151,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(17151,36,'cp 75','d2d4 e5d4 f3d4 f8g7 c1e3 e8g8 d4c6 d7c6 b5d3 c8e6 e1g1 f7f5 d1d2 f5e4 c3e4 e7f5 e3g5 d8d5 a1e1 a8e8 b2b3 f5d6 a2a4 a7a6 h2h3 b7b5 g5f4 d6e4 e1e4 e6f5 e4e8 f8e8 d3f5 d5d2 f4d2 g6f5 f1d1 g7f8 d2f4 e8e2 c2c3 e2c2 f4c7');
INSERT INTO evaluations VALUES(17707,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(17707,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(17707,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(17707,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(17707,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(17707,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(17707,36,'cp 25','g8f6 g1f3 c8g4 h2h3 g4f3 d1f3 e7e6 e1c1 f8b4 c3e2 d8e7 g2g4 e8c8 f4g5 b4d6 c1b1 h7h6 g5f6 e7f6 f3f6 g7f6 e2c1 h6h5 f1e2 h5g4 h3g4 c6e7 c2c4 d5c4 e2c4 c7c6 c4e2 f6f5 g4f5 e7f5 e2h5 h8h7 h5g4 d8h8 h1h7 h8h7 g4f5 e6f5 c1d3 c8d7 b1c2 a6a5');
INSERT INTO evaluations VALUES(17938,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(17938,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(17938,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(17938,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(17938,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(17938,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(17938,36,'cp 167','d1d4 d7d6 c1f4 g8e7 e1c1 e7c6 d4e3 c8e6 h2h4 f8e7 e4e5 e8g8 e3g3 g8h8 e5d6 c7d6 f4d6 d8a5 d6e7 c6e7 f1d3 a8c8 h1e1 c8c3 b2c3 e7d5 g3e5 a5a3 c1d2 f8c8 e5e4 a3c3 d2e2 d5f6 e4f4 e6a2 e2f1 a2e6 f1g1 h7h6 d3f5 e6f5 f4f5 c3c2');
INSERT INTO evaluations VALUES(18912,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(18912,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(18912,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18912,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(18912,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(18912,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18912,36,'cp 0','d7d5 b1c3 f8b4 d4c6 b7c6 d1d4 b4f8 c1e3 g8f6 e1c1 f8e7 g2g4 e8g8 g4g5 f6d7 d4d2 a8b8 h2h4 d8a5 d2e1 e7b4 a2a3 b4d6 c3e2 a5a4 e2c3');
INSERT INTO evaluations VALUES(25488,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(25488,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(25488,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(25488,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(25488,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(25488,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(25488,36,'cp 67','c6e7 b5d7 d8d7 d1d3 h7h6 c1e3 f6g4 e3d2 c7c6 d5c6 b7c6 h2h3 g4f6 e1c1 e7g6 h3h4 f6g4 d3e2 h6h5 f3e1 f8e7 f2f3 g4f6 e1d3 d7e6 c1b1 a8b8 b2b3 d6d5 h1e1 d5d4 c3a4 e8g8 g2g3 f8e8 a4b2 c6c5 f3f4 e6g4 f4f5 g4e2');
INSERT INTO evaluations VALUES(29585,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29585,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29585,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29585,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29585,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(29585,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(29585,36,'cp -13','e7e6 c2c4 f8e7 f3e5 h7h6 d1f3 e8g8 c4d5 d8d5 f3d5 e6d5 f1d3 f5d3 e5d3 a7a5 a2a4 b8c6 d2b3 b7b6 c1d2 e7d6 a1c1 c6e7 e1g1 f8e8 h2h3 h6h5 c1c2 g7g5 f1c1 g8g7 d2e1 g5g4 b3d2 g4h3 g2h3 e8g8 d3e5 c7c5');
INSERT INTO evaluations VALUES(30128,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(30128,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(30128,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(30128,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(30128,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(30128,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(30128,36,'cp -15','c2c4 f8e7 f3e5 h7h6 d1f3 e8g8 c4d5 d8d5 f1c4 d5f3 d2f3 b8d7 c4d3 f5d3 e5d3 c7c5 d4c5 e7c5 e1g1 f8c8 c1d2 c5d6 f1d1 c8c7 f3d4 d7e5 d3e5 d6e5 a1c1 c7c1 d1c1 e5d4 e3d4 a7a6 g1f1');
INSERT INTO evaluations VALUES(30766,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(30766,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(30766,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30766,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(30766,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30766,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(30766,36,'cp 16','e7e6 a2a3 g8f6 b1c3 f8d6 g1f3 e8g8 g2g3 e6d5 c4d5 c7c6 c1g5 c6d5 f1g2 b8a6 c3d5 b7d5 g5f6 d8f6 d1d5 f6b2 e1g1 d6e7 f3d4 a6c5 e2e3 a8c8 f1d1 g7g6 a1b1 b2c3 d4b5 c3a5 h2h4 f8d8 d5f3 e7f8 h4h5 c5e6 h5h6 f8h6 b5d6');
INSERT INTO evaluations VALUES(30611,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(30611,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(30611,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(30611,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(30611,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(30611,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(30611,36,'cp 21','a2a3 g8f6 b1c3 f8d6 g1f3 e8g8 g2g3 c7c6 f1g2 c6d5 c4d5 b8a6 c1g5 e6d5 c3d5 b7d5 g5f6 d8f6 d1d5 f6b2 e1g1 b2f6 a1d1 d6e7 f3d4 a6c5 e2e4 a8e8 f1e1 f6g6 f2f4 g6d6 d5d6 e7d6 e4e5 d6b8 d4f5 g7g6 f5e3 g8h8 e3g4 d7d6 g4f6');
INSERT INTO evaluations VALUES(31289,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(31289,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(31289,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(31289,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(31289,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(31289,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(31289,36,'cp 145','d4c5 a7a6 g1f3 e7e6 b2b4 g8e7 f1e2 b8c6 e1g1 g7g6 a2a4 f8g7 c1f4 h7h6 h2h3 g6g5 f4h2 e7g6 f1e1 g5g4 h3g4 f5g4 b1d2 h6h5 d1b3 h5h4 b4b5 a6b5 a4b5 a8a1 e1a1 c6e5 b5b6 e8g8 a1a7 h4h3 a7b7 h3g2 f3e5');
INSERT INTO evaluations VALUES(32417,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(32417,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(32417,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(32417,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(32417,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(32417,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(32417,36,'cp 121','b8c6 f1b5 d8f6 f3g3 f8d6 h2h4 f6g6 g3h3 d6b4 e1g1 g8f6 c3d5 b4d6 d5f6 g6f6 c1e3 e8g8 b5c4 c6a5 c4e2 a5c6 h3g4 c6d4 e2c4 d4e6 c4e6 f6e6 g4e6 f7e6 f1d1 a7a5 d1d3 a5a4 g1f1 b7b5 f1e2 g8f7 a1d1 h7h5 d3c3 f7e7');
INSERT INTO evaluations VALUES(34339,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(34339,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(34339,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(34339,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(34339,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(34339,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(34339,36,'cp 70','d2d4 g8f6 f3e5 g7g6 c1f4 f6h5 f4e3 h5f6 f2f4 f8g7 f1d3 e8g8 e1g1 f6d5 c3d5 c6d5 c2c3 f7f6 e5f3 b8c6 d1d2 c8d7 a1e1 a8c8 a2a4 c6a5 d2f2 a5c4 e3c1 d7g4 h2h3 g4f3 f2f3 e7e5 f4e5 f6e5 f3g4 e5d4 f1f8 c8f8 c3d4 g8h8 g1h1');
INSERT INTO evaluations VALUES(39627,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39627,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(39627,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(39627,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39627,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39627,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(39627,36,'cp 0','c7c5 g1e2 b8c6 e1g1 d7d6 d2d3 e7e6 c2c4 g8e7 b1c3 e8g8 c1e3 b7b6 d3d4 e6e5 d4d5 c6d4 f2f4 c8a6 d1a4 a6c8 f4e5 d6e5 a4d1 c8a6 d1a4');
INSERT INTO evaluations VALUES(37828,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(37828,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37828,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37828,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37828,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37828,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37828,36,'cp 0','g1e2 b8c6 e1g1 g8f6 b1c3 d7d6 c3d5 e8g8 c2c3 e7e5 d5f6 g7f6 d2d3 f6g7 c1e3 c8e6 d1d2 a8c8 f1d1 b7b5 b2b4 c5b4 c3b4 d8d7 a1c1 e6g4 f2f3 g4e6 f3f4 d6d5 e4d5 e6d5 g2d5 d7d5 f4e5');
INSERT INTO evaluations VALUES(37829,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37829,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37829,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37829,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37829,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37829,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37829,36,'cp 0','b8c6 e1g1 g8f6 b1c3 e8g8 c3d5 d7d6 c2c3 e7e5 d5e3 f6g4 e3d5');
INSERT INTO evaluations VALUES(42796,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(42796,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42796,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(42796,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42796,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(42796,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(42796,36,'cp 22','d2c3 d7d6 f2f3 d8e7 c4d3 d6d5 d1e2 b8d7 e4d5 f6d5 g1h3 d7c5 h3f2 e8g8 c3c4 d5b6 e1g1 f7f6 b2b3 c5d3 e2d3 f8d8 d3e4 d8d4 e4e2 c8f5 a3a4 a8d8 a4a5 b6c8 c1e3 d4d7 f2e4 b7b6 e4c3 f5g6 f1d1 d7d1 a1d1 d8d1 e2d1 b6a5 d1d5 g8f8');
INSERT INTO evaluations VALUES(42794,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(42794,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42794,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(42794,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(42794,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42794,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42794,36,'cp 13','d7d6 f2f3 d8e7 c4d3 d6d5 e4d5 f6d5 c3c4 d5b6 b2b3 c8e6 c1b2 b8d7 g1e2 e8c8 e1g1 d7c5 a3a4 c5d3 c2d3 h8e8 a4a5 b6d7 d3d4 e5d4 e2d4 a7a6 d1c2 d7f8 c2f2 e6d7 a1e1 e7g5 b2c3 e8e1 f1e1 d8e8');
INSERT INTO evaluations VALUES(47900,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(47900,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(47900,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(47900,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(47900,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47900,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(47900,36,'cp 33','g1e2 c7c5 d2d3 f8e7 e1g1 b8c6 b1c3 e8g8 f2f4 c8d7 h2h3 c6d4 a2a4 d7c6 g1h2 f8e8 b2b3 h7h6 c1b2 b7b6 e2d4 c5d4 c3e2 a7a6 c2c3 d4c3 e2c3 b6b5 a1c1 a8c8 a4b5 a6b5 c3d5 c6d5 c1c8 d8c8 e4d5 c8c5 f4e5');
INSERT INTO evaluations VALUES(50222,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(50222,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50222,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(50222,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50222,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(50222,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50222,36,'cp 0','d7d5 c2c4 c7c5 d4c5 b8c6 e2e3 f8g7 f1e2 e8g8 e1g1 d5c4 e2c4 d8c7 d1c2 c6b4 c2d1');
INSERT INTO evaluations VALUES(46800,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46800,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46800,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46800,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46800,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(46800,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(46800,36,'cp 8','b2b3 c8f5 e2e3 b8d7 f1d3 f6e4 c1b2 f8g7 f3h4 e7e6 h4f5 e6f5 e1g1 e4d2 d1d2 e8g8 c2c4 d7f6 d2c2 c7c6 g1h1 f8e8 a1e1 f6g4 g2g3 d8e7 h2h3 g4f6 h1g2 h7h5 a2a4');
INSERT INTO evaluations VALUES(50558,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(50558,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50558,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(50558,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(50558,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(50558,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(50558,36,'cp 31','g8f6 f1d3 e6e5 d4b3 d7d5 e4d5 f6d5 e1g1 c8e6 d1f3 d8f6 f3g3 d5e3 g3e3 b8d7 d3e4 a8b8 b1c3 f8e7 a1d1 e8g8 c3d5 f6g5 h2h3 f8d8 f2f4 e5f4 d5e7 g5e7 e3f4 e6b3 c2b3 d7f6 e4d3 b8c8 f1e1 e7c7 f4c7 c8c7');
INSERT INTO evaluations VALUES(43327,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(43327,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43327,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(43327,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43327,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(43327,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(43327,36,'cp 20','f1d3 e6e5 d4b3 d7d5 e4d5 f6d5 e1g1 c8e6 d1f3 b8c6 b1c3 d5e3 f3e3 d8c7 a1d1 e8c8 d3e4 f8d6 e4d5 e6f5 h2h3 c8b8 d1d2 f7f6 d5c6 c7c6 f1d1 f5e6 c3d5 h8e8 c2c4 c6c4 d5f6 g7f6 d2d6 c4c7 d6d3 e6c4 d3d8 e8d8 d1d8 c7d8 e3e4 d8d1 g1h2 c4b3 a2b3 d1d2 e4h7 d2f2');
INSERT INTO evaluations VALUES(53822,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(53822,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53822,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53822,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(53822,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(53822,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(53822,36,'cp 30','g8f6 e2e4 c8e6 d2d3 f8e7 g1f3 a7a6 e1g1 e8g8 h2h3 h7h6 c1e3 b8d7 f3d2 f6h7 f2f4 f7f5 a2a4 e5f4 g3f4 g8h8 c3e2 f5e4 d3e4 d8e8 a4a5 e6g8 e2d4 h7f6 b2b4 a8d8 a1b1 e8f7 d4f5 g8h7 f5e7');
INSERT INTO evaluations VALUES(5122,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(5122,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(5122,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(5122,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(5122,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(5122,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(5122,36,'cp 21','c5d4 f3d4 d7d5 e4d5 c6d4 c3d4 e6d5 f1e1 g8e7 c4b3 c8e6 b1c3 e8g8 c1g5 d8d6 g5e7 d6e7 b3d5 e7b4 d5e6 f7e6 d1e2 g7d4 e2e6 g8h8 c3d1 f8f6 e6e4 f6f4 e4e2 a8f8 e1f1 b4a5 g1h1 f4f7 f2f3 d4g7 e2e3 a5h5 d1f2 h5a5 e3e6 a5e5 e6e5');
INSERT INTO evaluations VALUES(5149,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(5149,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(5149,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(5149,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(5149,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(5149,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(5149,36,'cp 13','c4b3 d7d5 e4d5 e6d5 f1e1 g8e7 f3d4 c6d4 c3d4 e8g8 c1g5 c8e6 b1c3 d8d6 g5e7 d6e7 b3d5 e7b4 d5e6 f7e6 a1b1 g7d4 c3e4 a8d8 d1e2 d8d5 g1h1 b4b6 e1f1 b6b5 e2c2 d4g7 g2g3 b5d3 c2d3 d5d3 f1e1 f8d8 h1g2 d3d7');
INSERT INTO evaluations VALUES(5622,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(5622,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(5622,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(5622,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(5622,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(5622,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(5622,36,'cp 85','d2d4 e5d4 f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 d1f3 d8d7 b1c3 e7f6 f3d3 g8e7 c1d2 e8g8 a1e1 c6c5 b2b3 a8d8 h2h3 h7h6 f2f4 f6d4 g1h2 f8e8 c3e2 d4f6 d2c3 f6c3 e2c3 a7a5 f4f5 f7f6 c3e2 d6d5 e2g3 a5a4 e4e5 f6e5 e1e5');
INSERT INTO evaluations VALUES(5619,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(5619,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(5619,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(5619,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(5619,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(5619,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(5619,36,'cp 74','e5d4 f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 d1f3 d8d7 b1c3 e7f6 c1d2 g8e7 f3d3 e8g8 b2b3 a8d8 f2f4 c6c5 a1e1 e7c6 c3d5 f6d4 g1h1 c6e7 d5e7 d7e7 h2h3 e7d7 h1h2 f8e8 d2e3 d4f6 e1e2 e8e6 f1e1 d8e8 e3f2 d7c6');
INSERT INTO evaluations VALUES(14629,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(14629,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(14629,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(14629,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(14629,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(14629,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(14629,36,'cp 9','e1g1 f8e7 a2a3 d6d5 e4d5 f6d5 e3d2 e8g8 b1c3 d5f6 h2h3 f8e8 d1e2 e7d6 f1e1 c8e6 c4e6 e8e6 c3e4 d8d7 e4f6 e6f6 e2e4 b7b6 a1d1 a8d8 g1f1 a7a5 e4g4 f6e6 d2c3 b6b5 d3d4 e5d4 f3d4 e6e1 f1e1 c6d4 c3d4 d7g4 h3g4');
INSERT INTO evaluations VALUES(14628,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(14628,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(14628,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(14628,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(14628,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(14628,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14628,36,'cp 0','f8e7 a2a3 d6d5 e4d5 f6d5 e3d2 e8g8 f1e1 c8e6 d2c3 e7d6 c3e5 c6e5 f3e5 d8f6 d3d4 d5f4 c4f1 f8d8 b1c3 d6e5 e1e5 f4g6 e5e3 f6d4 d1e2 d4b6 c3a4 b6d4');
INSERT INTO evaluations VALUES(19421,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(19421,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(19421,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(19421,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(19421,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(19421,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(19421,36,'cp 141','d4d5 c6b8 e3e4 d7d6 f1b5 b8d7 a2a4 g7g6 g1f3 a7a6 b5c4 f8g7 e1g1 g8f6 b1c3 e8g8 h2h3 f6e8 d1e2 b7b6 c1f4 d7e5 f3e5 d6e5 f4g3 e8d6 b2b3 f7f6 c3d1 g7h6 d1b2 h6f4 g3f4 e5f4 e4e5 f6e5 e2e5');
INSERT INTO evaluations VALUES(19578,30,'cp -390',NULL);
INSERT INTO evaluations VALUES(19578,31,'cp -400',NULL);
INSERT INTO evaluations VALUES(19578,32,'cp -399',NULL);
INSERT INTO evaluations VALUES(19578,33,'cp -393',NULL);
INSERT INTO evaluations VALUES(19578,34,'cp -389',NULL);
INSERT INTO evaluations VALUES(19578,35,'cp -396',NULL);
INSERT INTO evaluations VALUES(19578,36,'cp -420','d8a5 c2c3 a5b5 a2a4 b5b6 e1g1 e7e6 c3d4 d7d5 b1c3 c8d7 e4d5 g8f6 d5e6 f7e6 d4d5 f8e7 d5e6 d7e6 d1e2 e8g8 f1e1 e6d7 e2e7 f8e8 c1e3 e8e7 e3b6 e7e1 a1e1 a7b6 f2f3 d7c6 g1f2 a8d8 f2e3 f6d5 c3d5 d8e8 e3f2 e8e1 f2e1 c6d5 b2b4 d5b3 a4a5');
INSERT INTO evaluations VALUES(20970,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(20970,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(20970,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(20970,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(20970,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(20970,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(20970,36,'cp -17','d5d4 e3d4 c6d4 f1e2 g8h6 f3d4 d8d4 d2d3 h6f5 e1g1 e7e5 c1d2 f8c5 d2c3 d4d6 b1d2 c5d4 e2f3 e8g8 f1e1 c7c6 f3e4 a7a5 d2f3 f7f6 a2a3 d4c5 e4f5 c8f5 d3d4 e5d4 f3d4 c5d4 d1d4 d6d4 c3d4 g8f7 a1d1');
INSERT INTO evaluations VALUES(21074,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(21074,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(21074,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(21074,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21074,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(21074,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(21074,36,'cp 28','d2d4 g8f6 c1f4 c8f5 b1d2 e7e6 f3h4 d8d7 e2e3 f8d6 h4f5 e6f5 g2g3 e8g8 f1d3 g7g6 d2f3 f8e8 e1f1 h7h5 f1g2 g8g7 d1c2 f6e4 a2a4 c6e7 h2h4 d6f4 e3f4 f7f6 f3g1 g7f7 f2f3 e4d6 b2b3 c7c6 g1h3 b7b5 a4b5');
INSERT INTO evaluations VALUES(21468,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(21468,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(21468,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(21468,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(21468,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(21468,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(21468,36,'cp 48','f7f6 f2f4 f6e5 d4e5 g8h6 f1d3 g7g6 g1f3 f8c5 b1c3 a7a6 d1e2 h6f5 c1d2 c6d4 f3d4 f5d4 e2g4 d8e7 h2h4 c8d7 h4h5 e8c8 h5g6 h7g6 e1c1 e7g7 c3e2 d4f5 c1b1 c5b6 g4g5 d7b5 h1h8 d8h8 d3b5 a6b5 g2g4 f5e3 d2e3 b6e3 d1d3 e3c5 b2b4');
INSERT INTO evaluations VALUES(21529,30,'cp -225',NULL);
INSERT INTO evaluations VALUES(21529,31,'cp -194',NULL);
INSERT INTO evaluations VALUES(21529,32,'cp -199',NULL);
INSERT INTO evaluations VALUES(21529,33,'cp -187',NULL);
INSERT INTO evaluations VALUES(21529,34,'cp -190',NULL);
INSERT INTO evaluations VALUES(21529,35,'cp -210',NULL);
INSERT INTO evaluations VALUES(21529,36,'cp -198','f8d6 d2d4 g8f6 h5d1 e8g8 g1e2 f6g4 b1c3 e5d4 e3d4 d8f6 f2f3 f8e8 e1f1 a7a6 b5a4 b7b5 a4b3 c8e6 f1g1 b5b4 h2h3 g4h2 h1h2 b4c3 b2c3 a6a5 a2a4 c6e7 g2g3 c7c5 c1f4 d6f4 e2f4 c5c4 b3a2 e6d7 d1d2 f6d6 g3g4 a8b8 f4h5 d7a4 a2b1 e7g6 h2f2 b8b2 f3f4');
INSERT INTO evaluations VALUES(5896,30,'cp -234',NULL);
INSERT INTO evaluations VALUES(5896,31,'cp -220',NULL);
INSERT INTO evaluations VALUES(5896,32,'cp -223',NULL);
INSERT INTO evaluations VALUES(5896,33,'cp -220',NULL);
INSERT INTO evaluations VALUES(5896,34,'cp -198',NULL);
INSERT INTO evaluations VALUES(5896,35,'cp -206',NULL);
INSERT INTO evaluations VALUES(5896,36,'cp -218','d2d3 g8f6 h5d1 e8g8 g1e2 c6e7 b1d2 a7a6 b5a4 b7b5 a4b3 c7c5 c2c4 b5c4 d3c4 d5d4 e2g3 d4e3 f2e3 e7f5 g3f5 c8f5 e3e4 f5g4 d1c2 d6c7 h2h3 g4h5 e1g1 h5e2 f1f2 e2d3 c2c3 d8d4 c3d4 e5d4 f2f3 d3e2 f3f5 d4d3 b3a4 a8d8 a4c6');
INSERT INTO evaluations VALUES(22500,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22500,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22500,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(22500,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22500,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22500,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22500,36,'cp 0','d7c6 b1c3 g8f6 f3e5 f6e4 c3e4 d8d4 e1g1 d4e5 d2d4 e5d5 e4c3 d5f5 d4d5 c6d5 d1d5 f5d5 c3d5 f8d6 c1f4 d6f4 a1e1 e8f8 d5f4 c8f5 e1e5 g7g6 f2f3 a8d8 f1e1 d8d4 g2g3 f5c2 e5c5 c2a4 c5c7 d4d7 c7d7 a4d7 e1d1 f8e7 f4d5 e7e6 d5c7 e6e7');
INSERT INTO evaluations VALUES(21316,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21316,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(21316,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(21316,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21316,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21316,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21316,36,'cp 0','b1c3 g8f6 f3e5 f6e4 c3e4 d8d4 e1g1 d4e5 d2d4 e5d5 e4c3 d5f5 d4d5 c6d5 d1e2 f5e6 e2d2 e6c6 d2d5 c6d5 c3d5 f8d6 c1f4 d6f4 a1e1 e8f8 d5f4 c8f5 e1e5 g7g6 f1d1 f8g7 f4h5 g7h7 h5f6 h7g7');
INSERT INTO evaluations VALUES(23120,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(23120,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(23120,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(23120,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(23120,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(23120,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(23120,36,'cp 209','f3e5 g8f6 c1e3 d4e6 b1d2 f8c5 e3c5 e6c5 d3d4 c5e6 c2c3 e8g8 f1c4 d7d6 e5d3 f6e4 c4e6 f8e8 e6f7 g8f7 e1g1 e4d2 d1d2 f7g8 a1e1 c8d7 c3c4 d8f6 d3f4 f6f7 h2h3 a7a5 f4d5 d7c6 d5e3 a5a4 d4d5');
INSERT INTO evaluations VALUES(25332,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(25332,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(25332,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25332,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(25332,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(25332,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(25332,36,'cp 37','d2d3 e5f4 c1f4 f8e7 h2h3 g4h5 d1e1 h5f3 f1f3 c6d4 f3f2 e8g8 b1c3 c7c6 f4e3 d4e6 f2f5 g7g6 c4e6 f7e6 f5f1 d8c7 e3h6 f8f7 e4e5 d6e5 g1h1 e7f8 h6g5 f8g7 g5f6 g7f6 c3e4 a8f8 a2a4 f6d8 f1f7 c7f7 c2c3 f7f4 e1e2 d8e7 b2b4');
INSERT INTO evaluations VALUES(25330,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(25330,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(25330,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(25330,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(25330,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(25330,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(25330,36,'cp 50','e5f4 c1f4 f8e7 h2h3 g4h5 b1c3 c6a5 c4b3 a5b3 a2b3 h5g6 d1e1 c7c6 g1h1 a7a6 c3e2 e8g8 g2g4 d8d7 e1g3 c6c5 g3g2 f8e8 e2g3 b7b5 g3f5 g6f5 g4f5 f6h5 f4g5 g8h8 f1g1 e7g5 f3g5');
INSERT INTO evaluations VALUES(31061,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(31061,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(31061,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31061,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(31061,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(31061,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(31061,36,'cp 40','e7e6 d1b3 b7b6 e2e3 f8d6 f4e5 d6e5 f3e5 g8e7 c4d5 e6d5 f1d3 e8g8 d3f5 e7f5 b1c3 d8d6 e1g1 b8d7 e5f3 f8e8 f1c1 h7h6 b3d1 d7f6 c1c2 c6c5 b2b3 a8c8 a1c1 d6e6 d4c5 b6c5 d1d2 f5d6 h2h3 g8h8 c3e2');
INSERT INTO evaluations VALUES(31440,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(31440,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(31440,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(31440,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(31440,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(31440,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31440,36,'cp 33','d1b3 b7b6 e2e3 g8f6 b1c3 f8d6 f4d6 d8d6 f3h4 f5e4 c4d5 e6d5 c3e4 f6e4 f1d3 g7g6 b3c2 e8g8 e1g1 e4f6 b2b3 f8c8 f1c1 c6c5 h4f3 d6e7 a2a3 b8d7 c2b2 f6e4 c1c2 d7f6 f3e5 c8c7 d3e4 d5e4');
INSERT INTO evaluations VALUES(39776,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(39776,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(39776,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(39776,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(39776,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(39776,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(39776,36,'cp 33','c4d5 e6d5 e2e3 h7h6 g5h4 e8g8 f1d3 c8e6 h2h3 c7c6 e1g1 b8d7 h4g3 e7b4 a1c1 f8e8 d3b1 b4c3 b2c3 f6e4 g3h2 b7b5 c3c4 a8c8 c4c5 e6f5 b1d3 a6a5 f3e1 f5h7 d1c2 d7f8 f2f3 e4f6 h2e5 h7d3 e1d3 f6d7 c1e1 d7e5 d3e5 f7f6 e5d3 f8g6 c2f2');
INSERT INTO evaluations VALUES(39827,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(39827,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(39827,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(39827,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(39827,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(39827,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(39827,36,'cp 48','e2e4 c7c5 d4d5 e6d5 e4d5 d8e7 g1e2 e8g8 f2f3 d7d6 c1d2 b4c3 d2c3 b6b5 e1c1 b8d7 c4b5 a7a6 e2g3 d7b6 c3f6 e7f6 c1b1 f6f4 g3e4 a6b5 f1b5 c8f5 b1a1 f8b8 c2c1 f4c1 d1c1 f5e4 f3e4 b6d5 b5c6 a8a2 a1a2 d5b4 a2b1 b4c6 h1d1 c6d4 c1c3 g7g6 b1a1 h7h6');
INSERT INTO evaluations VALUES(39833,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(39833,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(39833,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(39833,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(39833,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(39833,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(39833,36,'cp 41','c7c5 d4d5 e6d5 e4d5 d8e7 g1e2 d7d6 f2f3 b8d7 c1d2 b4c3 d2c3 b6b5 e1c1 e8g8 e2g3 e7d8 c4b5 d7b6 c3f6 d8f6 g3e4 f6e7 e4c3 a7a6 f1d3 a6b5 d3b5 e7f6 c3e4 f6e7 a2a3 c8f5 b5c6 a8b8 c1b1 c5c4 b1a1 f5g6 h1e1');
INSERT INTO evaluations VALUES(40883,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(40883,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(40883,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(40883,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(40883,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(40883,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(40883,36,'cp 112','e2e4 e8g8 f1d3 e6e5 h2h3 e5d4 f3d4 b8c6 c1e3 f6d7 e1g1 e7f6 d4e2 d7e5 f2f4 e5d3 d1d3 f6c3 e2c3 f7f5 d3d5 g8h8 a1e1 d8f6 d5d2 f5e4 g1h2 f6f7 b2b3 c8d7 c3e4 a8e8 e4c3 h8g8 f4f5 c6e5 f5f6 f7h5 c3d5 e5g4 h2g1 g4f6 d5c7 e8e5 e3a7 d7c6 e1e5 h5e5');
INSERT INTO evaluations VALUES(42512,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(42512,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(42512,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(42512,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42512,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(42512,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(42512,36,'cp 16','d7d5 e4d5 d8d5 g1f3 b7b6 f1d3 c8b7 e1g1 b8d7 f1e1 e8g8 a2a4 d5h5 d3e2 a7a5 b2b3 a8d8 c3b2 h5f5 d1d2 f6h5 d2g5 f5g5 f3g5 h5f6 b2a3 f8e8 e2f3 b7f3 g5f3 f6d5 g1f1 d5c3 g2g3 c3d5 c2c4 d5b4');
INSERT INTO evaluations VALUES(43202,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43202,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(43202,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(43202,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(43202,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43202,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43202,36,'cp 0','d7d5 f4h6 g7h6 d2h6 c8f5 f2f3 f5c2 g2g4 b7b5 a1c1 b5b4 c1c2 b4c3 c2c3 b8d7 g1h3 c7c5 d4c5 d8a5 a2a3 d5d4 b2b4 a5a4 c3c4 a4a3 f1g2 d7e5 c4d4 e5c6 d4c4 c6e5');
INSERT INTO evaluations VALUES(41462,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41462,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41462,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41462,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41462,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41462,36,'cp 0','f4h6 g7h6 d2h6 c8f5 f2f3 f5c2 g2g4 b7b5 a1c1 b5b4 c1c2 b4c3 c2c3 b8d7 g1h3 c7c5 d4c5 d8a5 a2a3 d5d4 b2b4 a5a4 c3c4 a4a3 f1g2 d7e5 c4d4 e5c6 d4c4 c6e5');
INSERT INTO evaluations VALUES(51891,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(51891,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(51891,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51891,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(51891,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(51891,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(51891,36,'cp -8','f8c5 g1f3 g8f6 a2a3 e8g8 e2e3 b7b6 f1e2 c8b7 e1g1 a7a5 b1c3 d7d5 c4d5 f6d5 c1d2 b8d7 c3d5 b7d5 d2c3 c5e7 f3e5 d7e5 c3e5 e7f6 d1d4 f6e5 d4e5 d8f6 e5f6 g7f6 f1c1 f8c8 h2h3 a5a4 e2f3 c8d8 f3d5 d8d5 c1d1 d5d1 a1d1 a8c8 d1d7 c8c2');
INSERT INTO evaluations VALUES(37669,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(37669,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(37669,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(37669,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(37669,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(37669,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37669,36,'cp 0','a2a3 g8f6 e2e3 d7d5 g1f3 e8g8 c4d5 e6d5 f1e2 a7a5 e1g1 b8c6 b1c3 c8f5 d1b3 d8d6 c3a4 c5a7 a4c3');
INSERT INTO evaluations VALUES(53248,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(53248,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(53248,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(53248,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(53248,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(53248,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(53248,36,'cp -7','b8c6 e2e4 c5d4 e4d5 d4c3 d5c6 d8d1 a1d1 b7c6 f4c7 e7e6 d1d8 e8e7 b2b3 g8f6 g1e2 f6d5 c7a5 c6c5 g2g3 c8b7 d8a8 b7a8 f1g2 g7g6 e2c3 f8g7 c3d5 e6d5 a5d2 e7d7 h2h4 h8e8 e1d1 a8c6 h1e1 e8e6 g2h3 f7f5 e1e6 d7e6 h3g2 c6b5 a2a4 b5a6 d2g5 c5c4');
INSERT INTO evaluations VALUES(19773,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19773,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19773,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19773,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19773,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19773,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(19773,36,'cp 0','e2e4 c5d4 e4d5 d4c3 d5c6 d8d1 a1d1 b7c6 f4c7 c8g4 g1f3 a7a5 b2c3 g8f6 f1a6 f6d5 a6b7 e8d7 d1d5 c6d5 b7a8 g4f3 c7a5 f3g2 h1g1 g2e4 e1e2 e7e5 a8b7 f8c5 g1d1 c5d6 b7d5 e4d5 d1d5 d7c6 c3c4 h8a8 d5b5 g7g6 a2a4 f7f6 e2d3 h7h5 c2c3 e5e4 d3e4 d6h2');
INSERT INTO evaluations VALUES(55907,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(55907,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(55907,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55907,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(55907,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(55907,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(55907,36,'cp -39','d5e4 e2e4 g8f6 e4h4 c7c5 d2d4 c5d4 h4d4 c8d7 d4h4 b8c6 f1c4 c6e7 c4d3 e7f5 h4h3 f5d6 h3h4 f6d5 h4d4 d8b6 d4b6 a7b6 c1d2 b6b5 f3e5 d7c6 e1g1 g7g6 f1e1 f8g7 b1c3 e8g8 a2a3 f8d8 c3d5 c6d5 a1d1 d6c4 d2c3 c4e5 c3e5 d5c6');
INSERT INTO evaluations VALUES(56085,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(56085,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(56085,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(56085,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(56085,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(56085,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(56085,36,'cp -20','e2e4 g8f6 e4h4 c7c5 d2d4 c5d4 f3d4 e6e5 f1b5 c8d7 b5d7 b8d7 d4e2 a8c8 e1g1 d8b6 b1c3 f8e7 f1d1 b6e6 h4g3 e8g8 h2h3 a7a6 a2a4 f8e8 a4a5 e6c6 c1e3 e7f8 g3f3 c6f3 g2f3 c8c6 e2g3 h7h6 a1c1 e8c8 g3f5');
INSERT INTO evaluations VALUES(56075,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(56075,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(56075,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(56075,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(56075,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(56075,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(56075,36,'cp -32','g8f6 e4h4 c7c5 d2d4 c5d4 f3d4 e6e5 f1b5 c8d7 b5d7 d8d7 d4b3 b8c6 b1c3 d7f5 c1g5 e8c8 e1g1 f8e7 g5f6 g7f6 h4c4 f5e6 c4e2 f6f5 a1d1 h8g8 d1d8 g8d8 e2f3 e5e4 f3e3 e7b4 f1d1 d8d1 c3d1 b4d6 e3h3 f5f4 h3e6 f7e6 c2c3 c6e5 g2g3 e5f3 g1g2 d6e5');
INSERT INTO evaluations VALUES(56436,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(56436,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(56436,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(56436,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(56436,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(56436,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(56436,36,'cp -74','e5e4 f3d4 f7f5 d2d3 g8f6 d3e4 f5e4 f2f3 e4f3 d4f3 b8c6 f1d3 c8g4 c1d2 f8d6 d1e2 c6e5 e1c1 e8g8 h2h3 g4h5 g2g4 e5d3 c2d3 h5g6 f3h4 d8e8 h4g6 e8g6 c3b5 d6e5 d1f1 a8e8 f1f5 c7c6 b5c3 b7b5 h1f1 b5b4 c3a4 e5d6 c1b1 f6d7 d2c1 f8f5 f1f5 d6c7 e3e4 d5e4 d3e4');
INSERT INTO evaluations VALUES(56255,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(56255,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(56255,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(56255,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(56255,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(56255,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(56255,36,'cp -87','f3g1 f7f5 b2b3 g8f6 c1b2 a7a6 g1h3 f8d6 c3e2 b8d7 e2f4 d8e7 c2c4 h8g8 f4d5 f6d5 c4d5 g7g5 h3g1 d7e5 b2e5 d6e5 a1c1 c8d7 h2h4 e8c8 h4g5 f5f4 g1e2 f4e3 d2e3 g8g5 d1c2 e5d6 a2a3 c8b8 b3b4 g5d5 e2c3');
INSERT INTO evaluations VALUES(6016,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(6016,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(6016,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(6016,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(6016,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(6016,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(6016,36,'cp -11','g8f6 f1e2 e8g8 e1g1 f8e8 d1d2 a7a5 a2a3 c8d7 a1e1 h7h6 e2d1 c5e3 f2e3 c6e7 f3h4 e7g6 h4f5 b7b5 f1f2 g6f8 c3e2 d7f5 f2f5 f8h7 e1f1 g8h8 g1h1 g7g6 f5f2 d8e7 e2g3 h6h5');
INSERT INTO evaluations VALUES(9402,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(9402,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9402,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(9402,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(9402,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(9402,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(9402,36,'cp -11','f1e2 c6d4 d1d2 a7a6 e1g1 e8g8 c3d1 h7h6 c2c3 d4e2 d2e2 c5e3 d1e3 c8e6 f1e1 f8e8 e2c2 c7c5 a2a4 b7b5 h2h3 d8b6 f3h4 d6d5 a4b5 d5d4 c3d4 c5d4 e3c4 e6c4 c2c4 a6b5 c4c2 f6h5 h4f3 a8a1 e1a1 h5f4');
INSERT INTO evaluations VALUES(10270,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(10270,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(10270,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(10270,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(10270,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(10270,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(10270,36,'cp 0','d7d5 f3e5 c6e5 d3d4 c5d6 d4e5 d6e5 c2c4 c7c6 c4d5 f6d5 e2e4 d5b4 d1d8 e8d8 b1d2 b4c2 a1b1 c8e6 d2f3 e6a2 f3e5 f7f6 c1f4 a2b1 e5f7 d8e7 f7h8 b1a2 f1d1 a8h8 e4e5 h8d8 e5f6 g7f6 d1d8 e7d8 g2e4 c2d4 f4e3 f6f5 e4d3 d4f3 g1f1 a2e6 f1e2 f3g1 e2d2 g1f3');
INSERT INTO evaluations VALUES(10892,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(10892,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(10892,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(10892,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(10892,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(10892,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(10892,36,'cp -79','f6e4 c4d5 e4f6 d5c6 d7c6 f3e5 e8g8 d2d4 c6c5 h2h3 c5d4 c3d4 c8e6 b1c3 f6d5 e5d3 f8e8 f1e1 b7b6 d1f3 c7c6 c1d2 e7f8 c3d5 e6d5 e1e8 d8e8 f3g3 e8d7 d2b4 h7h6 b4f8 a8f8 d3f4 f8d8 g3e3 d5c4 b2b3 d7d4 e3d4 d8d4');
INSERT INTO evaluations VALUES(13103,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(13103,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(13103,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(13103,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(13103,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(13103,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(13103,36,'cp 73','d4c5 a7a5 f1b5 d8c7 b2b4 e7e5 c1b2 f8e7 a2a3 e8g8 b5e2 e5e4 f3d4 c7e5 h2h4 e5c7 b1d2 f8d8 d4c6 b7c6 d2b3 f6d7 d1c2 d7e5 b3a5 c8f5 a5b3 e5d3 e2d3 e4d3 c2d1 e7f6 e1g1 c7e7 d1f3 f5e4 f3g4 f6e5 b3d4 e5d4 c3d4 f7f5 g4g3');
INSERT INTO evaluations VALUES(13579,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(13579,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(13579,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(13579,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(13579,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(13579,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(13579,36,'cp -42','f8d6 g1h3 d8c7 e1g1 e8g8 h3f2 f6d7 d2f3 c5c4 d3c2 f7f5 b2b4 c6e7 a2a4 a7a6 f3e5 d7f6 g2g4 c8d7 c1d2 b7b5 a4a5 f5g4 e5g4 d7e8 d1f3 e8g6 c2g6 e7g6 g4f6 g7f6 g1h1 g8h8 f3h3 f8g8');
INSERT INTO evaluations VALUES(8390,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(8390,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(8390,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(8390,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(8390,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(8390,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(8390,36,'cp -28','g1f3 c5d4 c3d4 c6b4 d3b1 b7b6 a2a3 b4c6 d1e2 c8b7 b1d3 e8g8 e1g1 a8c8 f3e5 c8c7 d2b1 f6e4 e5c6 b7c6 c1d2 a7a5 f1c1 f7f5 b1c3 g7g5 d3b5 e4c3 c1c3 c6b5 e2b5 c7c3 b2c3 d8c7 a3a4 h7h6 a1b1 f8b8 b1c1 b8c8 c1b1 g5f4 e3f4 d6f4 d2f4 c7f4');
INSERT INTO evaluations VALUES(14995,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(14995,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(14995,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(14995,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(14995,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(14995,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(14995,36,'cp 83','f6d5 f3e5 c6e5 d4e5 d5c3 d1d8 e8d8 b2c3 f8c5 c1g5 d8e8 f2f3 h7h6 g5h4 a7a5 h4f2 c5f2 e1f2 a5a4 f3f4 e8e7 f1d3 a8a5 h1e1 a5c5 c3c4 c8e6 e1e4 h8d8 a1b1 b7b6 b1b5 a4a3 g2g3 c5b5 c4b5 f7f6 e4a4 f6e5 f4e5 e6a2 a4a3 a2e6 a3a7');
INSERT INTO evaluations VALUES(15334,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15334,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(15334,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(15334,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(15334,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(15334,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(15334,36,'cp 95','f3e5 c6e5 d4e5 f8b4 c1d2 b4c3 b2c3 d8e7 f2f4 c8d7 c3c4 d5b6 f1d3 e7h4 g2g3 h4h3 d1e2 e8c8 e2e3 c8b8 e1c1 h7h5 d2b4 h5h4 b4e7 d8c8 h1g1 h4g3 h2g3 d7f5 e7c5 f5d3 c2d3 c8d8 g3g4 h3e3 c5e3 h8h2 f4f5 h2a2 c1b1');
INSERT INTO evaluations VALUES(15764,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(15764,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(15764,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(15764,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(15764,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(15764,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(15764,36,'cp 96','f8e7 f2f3 e8g8 e1c1 c7c6 g2g4 d8a5 c1b1 f8e8 d4d2 f6d7 h2h4 d7e5 f1e2 b7b5 c3d5 e7d8 d2a5 d8a5 e3d2 a5d8 d5e3 d8b6 d2c1 b6c5 g4g5 c8e6 b2b3 a7a5 h4h5 a5a4 f3f4 c5e3 c1e3 e6g4 h1e1 e5f3 e1f1 f3h2 e2g4 h2g4 e3b6 a4b3 a2b3');
INSERT INTO evaluations VALUES(8692,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(8692,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(8692,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(8692,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(8692,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(8692,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(8692,36,'cp 122','f2f3 e8g8 e1c1 c7c6 g2g4 b7b5 g4g5 f6d7 c1b1 f8e8 h2h4 d8a5 a2a3 d6d5 e3f4 e7c5 d4d2 d5e4 c3e4 a5d2 f4d2 c5f8 d2c3 d7b6 f1d3 b6d5 c3d2 b5b4 a3b4 d5b4 d3c4 a7a5 h4h5 c8f5 h1h4 b4d5 e4g3 f5e6 h5h6 d5e3 h6g7 e3c4 g7f8q g8f8 h4h7');
INSERT INTO evaluations VALUES(8693,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(8693,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(8693,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(8693,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(8693,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(8693,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(8693,36,'cp 114','c7c6 e1c1 d8a5 g2g4 h7h6 h2h4 c8e6 f1c4 e8c8 c1b1 c8b8 g4g5 d6d5 e4d5 h6g5 e3g5 c6d5 c4b3 e7c5 d4d3 d5d4 c3e2 c5b6 b3e6 f7e6 d3g6 h8g8 g5f6 g7f6 g6e4 e6e5 h4h5 a5c5 h5h6 b8a8');
INSERT INTO evaluations VALUES(25504,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(25504,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(25504,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(25504,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(25504,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(25504,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(25504,36,'cp 52','f8e7 b5c6 d7c6 d2d4 e5d4 f3d4 c6d7 c1f4 e8g8 h2h3 c7c6 d1f3 d8b6 a1d1 a8d8 b2b3 f8e8 e1e3 h7h6 e3d3 b6b4 g2g4 d6d5 e4d5 f6d5 c3d5 c6d5 d4f5 d7f5 g4f5 e7h4 c2c3 b4e4 f3e4 e8e4 f4c7 d8d7 d3d5 d7c7');
INSERT INTO evaluations VALUES(24196,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(24196,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(24196,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(24196,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(24196,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(24196,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(24196,36,'cp 50','b5c6 d7c6 d2d4 e5d4 f3d4 c6d7 h2h3 e8g8 c1f4 c7c6 d1f3 d8b6 e4e5 d6e5 f4e5 h7h6 a2a4 a8e8 a4a5 b6a6 d4f5 d7f5 f3f5 a6c4 e1e3 e7d8 a1a4 c4c5 f5f4 f6h5 f4d4 c5d4 e5d4 b7b5 a5b6 a7b6 e3e4 f7f5 e4e3 c6c5 e3e8 f8e8 d4e3 h5f6');
INSERT INTO evaluations VALUES(26398,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26398,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26398,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26398,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26398,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26398,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26398,36,'cp 0','g4f3 e2f3 a8c8 f3e2 c5d4 c3d4 g8e7 b1a3 e7f5 a3c2 h7h5 e2b5 g7g6 c2e3 f5e7 e3c2');
INSERT INTO evaluations VALUES(32114,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(32114,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(32114,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(32114,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(32114,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(32114,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(32114,36,'cp 44','d3c4 c7c6 c1e3 g8f6 b1c3 f8e7 d1e2 b8d7 a2a4 f6g4 e3d2 g4f6 a4a5 d7f8 b2b4 f8e6 e1g1 e8g8 f1d1 d8c8 h2h3 e6d4 e2d3 d4e6 d2e3 f6h5 b4b5 e6f4 d3f1 c6b5 c3b5 f7f5 f3d2 h5f6 b5c3 h7h6 e4f5');
INSERT INTO evaluations VALUES(32113,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(32113,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32113,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(32113,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(32113,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(32113,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(32113,36,'cp 45','c7c6 b1c3 f8e7 c1e3 g8f6 d1e2 b8d7 f3d2 d8c7 d2b3 h7h5 a2a4 e8g8 e1g1 g7g6 a4a5 f6g4 e3d2 a7a6 h2h3 g4f6 d2e3 f6e8 f1d1 e8g7 c3a4 g7e6 e3b6 d7b6 a5b6 c7c8 e2e3 f8d8 b3a5 e6d4 e3d3 d4e6 b2b4 d8d7 c2c3 g8h8 d3f3');
INSERT INTO evaluations VALUES(37168,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(37168,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(37168,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(37168,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(37168,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(37168,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(37168,36,'cp -10','c7c5 e2e3 c5d4 g5f6 d8f6 c3d4 b7b6 f1e2 b8c6 g1f3 c6a5 d1a4 d7d6 e1g1 c8a6 f1c1 f8c8 e2f1 f6d8 h2h3 c8c7 f3d2 a8c8 c1c2 h7h6 a1c1 e6e5 d2b3 a5b3 a4b3 e5d4 b3a4 a6b7 e3d4 b7e4 c2c3');
INSERT INTO evaluations VALUES(39456,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(39456,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(39456,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(39456,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(39456,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(39456,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(39456,36,'cp 134','b8c6 a2a4 d7d6 h2h3 a7a6 e1g1 d8d7 d2d3 g7g5 c1e3 d7g7 e3c5 d6c5 c3d5 g5g4 f3h4 g4h3 g2g3 g8h6 g1h1 c8g4 f2f3 g4c8 c2c3 e8d8 f1g1 h6f7 g3g4 g7h6 h4f5 c8f5 g4f5 f7d6 a4a5 c6e7 b2b4 h3h2 g1g2 e7d5 c4d5 c5b4 c3b4');
INSERT INTO evaluations VALUES(40159,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(40159,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(40159,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(40159,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(40159,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(40159,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(40159,36,'cp 152','f1d3 a7a6 h2h3 h7h5 g1f3 h5h4 a2a3 b8d7 e1g1 f6h5 b2b4 e8f8 a1b1 b7b6 d1c2 g7g6 f1e1 f8g7 d3f1 a6a5 b4a5 a8a5 b1b3 g6g5 c3d1 d7f8 d1e3 f8g6 c2b1 a5a6 f3h2 g6f4 e3f5 c8f5 e4f5 h5f6 b1c2');
INSERT INTO evaluations VALUES(41331,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(41331,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(41331,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(41331,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(41331,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(41331,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(41331,36,'cp -34','c7c5 d4c5 d8a5 b1c3 d6d5 e1g1 a5c5 d1e1 e8g8 h2h3 c8f5 g2g4 f5d3 c2d3 c5b6 a1b1 b8c6 d3d4 e7e6 c1d2 f8c8 e1e2 f6e8 f1d1 e8d6 d2e1 a7a5 c3a4 b6a7 b1c1 c8c7 c1c2 a8c8 d1c1 b7b6 b2b3 c6e7 c2c7 c8c7 c1c7 a7c7 g1g2');
INSERT INTO evaluations VALUES(50255,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(50255,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(50255,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(50255,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50255,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50255,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(50255,36,'cp -15','d7d5 g1f3 c7c5 f1g2 f8g7 e1g1 e8g8 c2c4 d5d4 b3b4 a7a5 b4c5 f6e4 d2d3 e4c5 b1d2 b8c6 b2a3 c5a6 d1b3 a8b8 c4c5 c8e6 d2c4 c6b4 a3b4 a5b4 a2a3 b4a3 b3a3 e6c4 d3c4 b8c8 c5c6 c8c6 f3e5 c6e6');
INSERT INTO evaluations VALUES(51256,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(51256,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(51256,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(51256,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(51256,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(51256,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(51256,36,'cp 77','a4b5 c6b5 c3b5 a6b5 a1a8 c8b7 a8a1 e7e6 g1e2 b7e4 e2c3 f8b4 f1e2 g8f6 e1g1 b4c3 b2c3 e8g8 c1g5 b8d7 e2f3 d8c7 f1e1 e4d5 d1e2 h7h6 g5d2 f8b8 e1b1 d7f8 h2h3 f8g6 a1a6 g6h4 f3d5 f6d5 b1a1 b5b4 a6a7 c7d8 e2c4 b4c3 d2c3 d8g5 g2g3 h4f3 g1g2 f3h4 g2h1 h4f3');
INSERT INTO evaluations VALUES(51254,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(51254,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(51254,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(51254,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(51254,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(51254,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(51254,36,'cp 73','c6b5 c3b5 a6b5 a1a8 c8b7 a8a1 e7e6 g1e2 b7e4 e2c3 f8b4 f1e2 g8f6 e1g1 b4c3 b2c3 e8g8 c1g5 b8d7 e2f3 d8c7 f1e1 c7b7 g5d2 e4f3 d1f3 b7f3 g2f3 f6d5 e1b1 f8b8 g1f1 h7h6 f1e2 d7f6 a1a5 d5c7 h2h4 f6h5 b1a1 c7d5 a5a8 b8a8 a1a8 g8h7');
INSERT INTO evaluations VALUES(51516,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(51516,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(51516,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(51516,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(51516,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(51516,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(51516,36,'cp 80','e2e4 c7c5 g1e2 a7a6 d2d4 c5d4 e2d4 e7e6 f2f3 c8b7 c1e3 g8f6 g2g4 h7h6 d1d2 d7d5 e4e5 f6d7 f3f4 b8c6 e1c1 c6d4 e3d4 f8c5 h2h4 h6h5 g4h5 h8h5 d2e1 d8c7 f1e2 h5h7 h4h5 e8c8 c1b1 c8b8 a2a3 d8h8 h1g1 c5d4 d1d4 d7c5');
INSERT INTO evaluations VALUES(51587,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51587,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(51587,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(51587,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(51587,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51587,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(51587,36,'cp 50','g1f3 e7e6 e2e4 c8b7 d2d4 c5d4 f3d4 g8f6 f2f3 f8b4 c1f4 e8g8 d1d2 d7d5 e4d5 e6d5 e1c1 d5c4 f1c4 d8c8 c4b3 f8d8 d2c2 b8c6 d4c6 c8c6 h1e1 c6c5 f4g3 a8c8 d1d8 c8d8 a2a3 b4a5 c1b1 a5c3 c2c3 c5c3 b2c3 h7h6');
INSERT INTO evaluations VALUES(51586,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(51586,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(51586,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(51586,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(51586,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51586,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(51586,36,'cp 23','c8b7 e2e4 e7e6 d2d4 c5d4 d1d4 b8c6 d4d3 g8f6 f1e2 d8b8 a2a3 f8d6 c1g5 c6e5 d3d2 f6e4 c3e4 b7e4 e1c1 h7h6 g5e3 d6c5 f3e5 b8e5 e3c5 e5c5 d2d7 e8f8 f2f3 e4h7 d1d2 g7g5 d7d6 c5d6 d2d6 f8g7 h1d1 h8c8 d6d7');
INSERT INTO evaluations VALUES(30497,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(30497,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(30497,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(30497,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30497,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(30497,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(30497,36,'cp 28','d2d4 c5d4 f3d4 g8f6 f2f3 b8c6 e2e4 e7e6 c1e3 f8e7 f1e2 e8g8 e1g1 d8b8 g1h1 c6d4 e3d4 f6h5 d1d2 h5f4 d4f2 e7b4 e2d3 b8d6 a1d1 f4d3 d2d3 d6d3 d1d3 b7c6 f1d1 h7h6 c3b5 f7f5 e4f5 c6b5 c4b5 f8f5 g2g3 b4c5 f2c5 f5c5 d3d7');
INSERT INTO evaluations VALUES(52759,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(52759,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(52759,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(52759,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(52759,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(52759,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(52759,36,'cp -30','g8f6 f1d3 c8g4 g1f3 e7e6 h2h3 g4f5 b1c3 f8e7 c1d2 e8g8 g2g4 f5d3 c2d3 c6c5 e1g1 b8c6 c3e2 a8c8 a1c1 c5d4 e2d4 c6d4 f3d4 c8c1 d1c1 d8b6 c1c2 e7c5 d2c3 b6a6 f1a1 f8c8 c2d2 f6d7 a2a4 c5e7 g1h1 d7c5');
INSERT INTO evaluations VALUES(44937,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(44937,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(44937,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(44937,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(44937,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(44937,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(44937,36,'cp -12','f1d3 c8g4 g1f3 e7e6 h2h3 g4f5 b1c3 f8e7 c1d2 e8g8 g2g4 f5d3 c2d3 c6c5 e1g1 b8c6 c3e2 a8c8 a1c1 c5d4 e2d4 c6d4 f3d4 c8c1 d1c1 d8d7 d2c3 d7a4 a2a3 h7h5 g4g5 f6e8 c1d1 a4a6 d1h5 a6d3 h5e2 d3e2 d4e2 e8d6 f1c1');
INSERT INTO evaluations VALUES(44936,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(44936,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(44936,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(44936,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(44936,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(44936,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(44936,36,'cp -27','c8g4 g1f3 e7e6 h2h3 g4f5 b1c3 f8e7 c1d2 e8g8 g2g4 f5d3 c2d3 c6c5 e1g1 b8c6 c3e2 a7a5 f1f2 f6d7 g4g5 b7b5 d4c5 d7c5 d2c3 b5b4 c3d4 d8b6 a1c1 b6b5 d4c5 e7c5 d3d4 c5d6 f3e5 c6e7 e2g3 d6e5 f4e5');
INSERT INTO evaluations VALUES(29043,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(29043,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(29043,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(29043,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(29043,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(29043,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(29043,36,'cp -7','g1f3 e7e6 h2h3 g4h5 c2c4 f8e7 e1g1 e8g8 b1c3 d5c4 d3c4 h5g6 a2a4 c6c5 d4d5 e6d5 c3d5 b8c6 f3e5 c6e5 f4e5 f6d7 e5e6 d7e5 e6f7 g6f7 d5e7 d8e7 c4f7 f8f7 c1d2 e7e6 f1f7 e5f7 d1h5 b7b6 d2c3 a8e8 a4a5 h7h6 e3e4 e6e4 a1e1 e4c6 e1e8 c6e8 h5d5');
INSERT INTO evaluations VALUES(29044,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(29044,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(29044,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(29044,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(29044,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(29044,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(29044,36,'cp -11','e7e6 h2h3 g4f5 e1g1 f8e7 c2c4 d5c4 d3c4 e8g8 g2g4 f5g6 b1c3 b7b5 c4d3 b5b4 c3e2 c6c5 g4g5 f6d7 b2b3 a7a5 c1b2 a5a4 a1c1 a4b3 a2b3 a8a2 c1c2 d8b6 d3g6 h7g6 e2c1 a2a5 c1d3 b8c6 d3c5');
INSERT INTO evaluations VALUES(53739,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(53739,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(53739,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(53739,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(53739,34,'cp -119',NULL);
INSERT INTO evaluations VALUES(53739,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(53739,36,'cp -127','f8d6 d2d4 g8f6 g1e2 d8c7 h2h3 e8g8 e1g1 f8e8 c2c3 c8f5 a4c2 f5g6 b2b3 g6h5 c1e3 b8d7 d1d2 c6c5 e2f4 h5g6 c2g6 h7g6 d4c5 d6c5 f4d5 f6d5 d2d5 c5e3 f2e3 d7f6 d5g5 a8d8 b1a3 f6e4 g5f4 c7f4 e3f4 e4c3 f4f5 g6f5 f1f5 d8d2 a3c4');
INSERT INTO evaluations VALUES(55975,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(55975,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(55975,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55975,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(55975,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(55975,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(55975,36,'cp -65','c7c5 c2c3 d5d4 c3d4 c5d4 g1f3 b8c6 f1b5 c8d7 e1g1 a7a6 b5c6 d7c6 d2d3 g8e7 c1g5 h7h6 g5e7 f8e7 d1e2 e8g8 b1d2 a8c8 a1c1 c6d5 d2e4 d8b6 h2h4 c8c6 c1c6 b7c6 f1c1 a6a5 h4h5 f8a8 g2g4 a5a4 g4g5 h6g5 f3g5');
INSERT INTO evaluations VALUES(54323,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(54323,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(54323,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(54323,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(54323,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(54323,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(54323,36,'cp -62','c2c3 d5d4 c3d4 c5d4 g1f3 b8c6 f1c4 d8c7 c4b5 c8d7 e1g1 c6e5 b5d7 e5d7 f3d4 f8d6 h2h3 c7c4 d4f3 c4d3 b1c3 g8f6 f1e1 e8g8 e1e3 d3a6 d2d4 d7b6 f3e5 a8d8 e3e1 d6e5 e1e5 h7h6 e5e3 b6d5 e3g3 g8h8 c1d2 d8d7 c3d5 f6d5 a1c1');
INSERT INTO evaluations VALUES(9055,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(9055,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(9055,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(9055,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(9055,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(9055,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(9055,36,'cp 209','f1d3 c6e7 h2h4 c7c5 f3e5 e8g8 d3c2 a5c7 h4h5 f6e8 d4c5 f7f6 e5g6 f8f7 g6e7 d8e7 h5h6 g7g6 f4c7 e8c7 b2b4 b7b6 d2b3 c8a6 c5b6 a7b6 a2a4 a6c4 c2d3 a8c8 d1c2 e6e5 a4a5 b6a5 a1a5 f6f5 d3c4 d5c4 b3d2 e5e4');
INSERT INTO evaluations VALUES(9056,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(9056,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(9056,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(9056,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(9056,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(9056,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(9056,36,'cp 209','c6e7 f3e5 e8g8 h2h4 c7c5 h4h5 a5c7 d3c2 f6e8 d4c5 f7f6 e5g6 f8f7 f4c7 d8c7 g6e7 f7e7 h5h6 g7g6 b2b4 f6f5 a2a4 e8f6 a4a5 c8d7 c2b3 d7c6 d2f3 a8d8 d1d4 f6g4 d4d2 a7a6 e1c1 e7e8 f3d4 e6e5 d4f3 e5e4 f3d4');
INSERT INTO evaluations VALUES(12227,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12227,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12227,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12227,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(12227,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12227,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(12227,36,'cp 0','d8a5 g5f6 d7f6 g1f3 g7g6 c4d5 f6d5 d1b3 f8g7 f1c4 e7e6 a2a4 e8g8 e1g1 a8b8 f1c1 f8d8 c3e4 a5c7 a4a5 b7b6 b3a3 g7f8 a3b3');
INSERT INTO evaluations VALUES(12273,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(12273,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(12273,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12273,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12273,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12273,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12273,36,'cp 0','g5h4 f8e7 d4c6 b7c6 e4e5 f6d5 h4e7 d8e7 d1d4 d5c3 d4c3 a7a5 a2a3 f7f6 e1c1 f6e5 c3e5 e8g8 f2f4 e7f6 g2g3 f6e5 f4e5 f8f5 d1e1 c8a6 b2b3 f5f2 f1a6 a8a6 e1d1 a6a7 d1d2 f2f7 d2d6 f7f2');
INSERT INTO evaluations VALUES(12274,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12274,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12274,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12274,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12274,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12274,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12274,36,'cp 0','f8e7 d1d3 e8g8 e1c1 d7d5 e4d5 f6d5 h4e7 d8e7 c3d5 e6d5 d3e3 e7e3 f2e3 f8e8 f1b5 c8d7 d4f3 d7f5 f3d4');
INSERT INTO evaluations VALUES(7891,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7891,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7891,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7891,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7891,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7891,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7891,36,'cp 13','d1d3 d7d5 e1c1 e8g8 e4d5 f6d5 h4e7 d8e7 c3d5 e6d5 a2a3 f8e8 d3d2 a7a6 d2f4 c8d7 f1d3 c6e5 c1b1 b7b5 d4f3 d7g4 d3e2 a8d8 d1d2 e5f3 e2f3 g4f3 f4f3 e7e4 d2d3 e4f3 d3f3 e8e2 h1d1 g7g6 b1c1 g8g7');
INSERT INTO evaluations VALUES(13248,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13248,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13248,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13248,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13248,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13248,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13248,36,'cp 0','g1f3 e7e6 e1g1 f8e7 b1d2 e8g8 f1e1 c6a5 c4b5 c8d7 b5d7 d8d7 a2a4 a5c6 d2c4 d7c7 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(14050,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(14050,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(14050,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(14050,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(14050,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(14050,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(14050,36,'cp 161','f1d3 b7b6 b2b4 c6b8 b1d2 c7c6 a1b1 a7a5 b4b5 c6b5 d3b5 c8d7 b5d3 b6b5 d3b5 d7b5 b1b5 e7f5 g1e2 f8e7 e1g1 e8g8 a2a4 b8c6 d1b3 h7h6 e2f4 a8c8 d2f3 c6a7 b5b7 c8c7 c1d2 c7b7');
INSERT INTO evaluations VALUES(14523,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(14523,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(14523,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(14523,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(14523,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(14523,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(14523,36,'cp -75','f8g7 c2c3 c6a5 c4b5 c7c6 b5a4 e8g8 b1d2 h7h6 g5f6 g7f6 a4c2 c6c5 d2c4 a5c4 d3c4 a7a6 c2b3 f6g7 f3d3 a6a5 b3c2 a5a4 g1f3 c8e6 a1d1 d8b6 d1b1 b6a6 b2b3 f8b8 e1g1 a4b3 a2b3 b7b5 c4b5 a6b5');
INSERT INTO evaluations VALUES(8771,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(8771,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(8771,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(8771,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(8771,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(8771,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(8771,36,'cp -78','g1e2 h7h6 g5f6 g7f6 b1d2 e8g8 e1g1 f6g7 f3g3 g8h7 a1e1 c6a5 a2a4 a5c4 d2c4 f7f5 e4f5 c8f5 f2f4 e5e4 d3d4 d8d7 b2b3 f5e6 c4e3 d6d5 c2c4 c7c6 h2h3 a8d8 c4d5 e6d5 e3g4 h6h5 g4e5 d7d6 f4f5 g6f5 g3g5 g7h6 g5h5');
INSERT INTO evaluations VALUES(21280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21280,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21280,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(21280,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(21280,36,'cp 3','g7g5 h2h3 f8g7 b1d2 g8e7 a2a4 f7f5 a4a5 a7a6 f3h2 f5f4 d1h5 e8d7 b2b4 d8f8 c1b2 d7d8 h5f7 c8d7 h2f3 f8f7 c4f7 h8f8 f7b3 e7g6 e1c1 g7f6 c1c2 h6h5 d1g1 c6e7 d3d4 g5g4 h3g4 h5g4 d4e5 g6e5 f3e5 f6e5 d2c4');
INSERT INTO evaluations VALUES(22919,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(22919,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(22919,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(22919,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(22919,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(22919,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(22919,36,'cp 90','d4d5 c6e5 d1e2 d8e7 b1c3 d7d6 c1d2 e5d7 e1c1 e7e2 g1e2 g8e7 d1e1 d7f6 e2f4 c8d7 c3e4 f6e4 e1e4 e8c8 d2c3 e7g6 f4h5 h8g8 h2h3 g6e7 f1c4 c7c6 d5c6 d7c6 e4g4 d6d5 c4d3 d5d4 c3b4 g7g6 h5f6 e7d5 f6d5 d8d5 b4f8 g8f8 h1e1 c8d7 f2f3');
INSERT INTO evaluations VALUES(24646,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(24646,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24646,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24646,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24646,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(24646,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(24646,36,'cp 18','c2c3 e7e6 h2h3 g4h5 c1f4 f8d6 f4d6 d8d6 b1d2 e8g8 f1e1 d6c7 d1c2 f8c8 a2a3 a7a5 a1c1 c7b8 c2a4 b8c7 f3e5 c7b6 c1c2 b6c7 g2g3 f6d7 e5d7 c7d7 g3g4 h5g6 d3g6 h7g6 d2f3 d7c7 g1g2 c7f4');
INSERT INTO evaluations VALUES(24928,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(24928,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(24928,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24928,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(24928,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(24928,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24928,36,'cp 26','h2h3 g4f3 d2f3 e7e6 f4d2 f6e4 e2e3 f8d6 c2c4 e8g8 a2a3 d5c4 f1c4 e4d2 f3d2 c6e7 d2e4 e7d5 d1c2 a8c8 e1g1 c7c5 c4d5 e6d5 e4d6 d8d6 d4c5 c8c5 c2d2 f8d8 d2d4 b7b6 a1c1 d6c6 d4d2 h7h5 f1d1 c5c1');
INSERT INTO evaluations VALUES(24927,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24927,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(24927,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(24927,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(24927,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24927,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24927,36,'cp 23','g4f3 d2f3 e7e6 f4d2 f8e7 e2e3 e8g8 f1e2 d8d6 a2a3 e6e5 d2b4 d6e6 b4e7 e6e7 e1g1 e5d4 e3d4 e7d6 c2c3 c6d8 e2d3 d8e6 f1e1 g7g6 d1d2 c7c5 d4c5 e6c5 a1d1 d6b6 d3b1 f8e8 f3e5 c5e4 b1e4 d5e4 e5g4 f6g4 h3g4 a8d8 d2f4');
INSERT INTO evaluations VALUES(25326,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(25326,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(25326,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(25326,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(25326,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(25326,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(25326,36,'cp -94','e5f4 c4b5 f6d7 d2d4 g7g5 g2g3 f8g7 b5c6 b7c6 g3f4 e8g8 d1d3 g5f4 c1f4 f7f5 e4e5 g4f3 d3f3 d6e5 d4e5 d7e5 f3g3 d8f6 g1h1 g8h8 a1e1 a8e8 f1g1 e5g6 f4c7 c6c5 g3f3 e8e1 g1e1 f6h4 e1f1');
INSERT INTO evaluations VALUES(25352,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(25352,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(25352,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(25352,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(25352,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(25352,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(25352,36,'cp -75','d2d4 g7g5 c4b5 f6d7 g2g3 f8g7 b5c6 b7c6 g3f4 e8g8 g1h1 f7f5 e4e5 g8h8 d1d3 d6e5 d4e5 g5f4 c1f4 g4f3 f1f3 d7e5 d3d8 a8d8 f3f2 e5g4 f2f1 d8b8 a1b1 f8f7 c3a4 b8b4 a4c5 b4b2 b1b2 g7b2 c5e6');
INSERT INTO evaluations VALUES(25354,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(25354,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(25354,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(25354,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(25354,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(25354,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(25354,36,'cp -71','g7g5 c4b5 a7a6 b5c6 b7c6 h2h3 g4h5 g2g4 f6g4 h3g4 h5g4 d1d3 f8g7 c1d2 e8g8 a1e1 h7h6 c3d1 a8b8 b2b3 f8e8 f3h2 g4h5 d2c3 h5g6 d4d5 c6d5 d3d5 h6h5 d5d2 e8e6 c3g7 g8g7 d1c3 f7f6 d2d3 e6e8 d3a6 b8a8');
INSERT INTO evaluations VALUES(29258,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(29258,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29258,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(29258,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29258,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(29258,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29258,36,'cp 19','d1b3 b7b6 f1e2 a7a5 a2a4 b8d7 h2h3 f6e4 b1d2 f8d6 f4d6 e4d6 c3c4 e8g8 a1c1 f5e4 e1g1 c6c5 f1d1 c5d4 f3d4 d7c5 b3a3 d5c4 d2e4 c5e4 f2f3 e4c5 e2c4 d8h4 c4e2 h4g5 g1h2 f8c8 b2b3');
INSERT INTO evaluations VALUES(30884,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(30884,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(30884,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(30884,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(30884,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(30884,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(30884,36,'cp 101','e4d5 d7b5 c3b5 e6d5 d2d4 f8e7 e1g1 g8f6 c1f4 b8a6 c2c3 e8g8 a2a4 f6e4 h2h3 f8e8 d4c5 e7c5 f3d2 c5d6 b5d6 e4d6 d1f3 d8f6 f4d6 f6d6 f1d1 a6c7 d2c4 d6c6 c4e3 e8e5 d1d4 a8e8 a1d1 a7a6 a4a5 g7g6 e3d5 e5d5 d4d5 c7d5');
INSERT INTO evaluations VALUES(32303,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(32303,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(32303,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(32303,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(32303,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(32303,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(32303,36,'cp -65','g1f3 c7c5 c2c4 g8f6 c4d5 e6d5 f1d3 f5d3 d1d3 c5c4 d3c2 f8d6 e1g1 e8g8 a2a3 b8c6 b1c3 d6c7 c1d2 f8e8 a1e1 a8c8 h2h3 c7a5 c3b5 a5b6 b5c3 c6e7 f3e5 b6a5 e5g4 f6g4 h3g4 a5b6 f4f5 f7f6');
INSERT INTO evaluations VALUES(32304,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(32304,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(32304,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(32304,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(32304,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(32304,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(32304,36,'cp -74','c7c5 c2c4 g8f6 c4d5 e6d5 b1c3 b8c6 a2a3 c5d4 f3d4 f5g4 d1c2 f8d6 f1d3 c6d4 e3d4 e8g8 e1g1 g4d7 c1e3 f6g4 a1e1 f8e8 c2d2 d7c6 g2g3 d8d7 e3f2 g4f6 e1e8 a8e8 f1e1 a7a6 e1e8 f6e8 g1g2 g7g6 h2h3 e8f6 f2e3');
INSERT INTO evaluations VALUES(32479,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(32479,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(32479,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(32479,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(32479,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(32479,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(32479,36,'cp 129','d2d4 b8c6 c2c3 c6a5 c4d3 a5c6 e1g1 g8f6 f1e1 f8e7 h2h3 e8g8 d4d5 c6b8 c3c4 a7a5 c1e3 h7h6 b1c3 b8a6 c4c5 c7c6 c5d6 e7d6 f3d2 c6d5 e4d5 b7b5 d3b5 d7b5 c3b5 f6d5 b5d6 d5e3 e1e3 d8d6 d2c4 d6d1 a1d1 a5a4 c4e5 a6b4 f2f4 b4c2');
INSERT INTO evaluations VALUES(39969,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(39969,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(39969,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(39969,33,'cp 235',NULL);
INSERT INTO evaluations VALUES(39969,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(39969,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(39969,36,'cp 244','h2h4 d7d6 h4h5 g6h5 h1h5 h7h6 c1e3 e6e5 d4e5 c8g4 h5h4 g4f3 g2f3 g7e5 h4h5 e7g6 d1c2 b8d7 e1c1 d8f6 c2d2 e5c3 b2c3 f6f3 d2d6 f3h5 d6d7 e8f8 d7b7 a8e8 b7c6 h5e5 e3d4 e5e4 c6d6 f8g8 c4c5 e8e6 d6d8 e6e8 d8f6 h8h7 f1c4 e4f4 c1c2 f4f6 d4f6 h6h5');
INSERT INTO evaluations VALUES(41199,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(41199,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(41199,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41199,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41199,36,'cp 0','f6d7 g1f3 c7c5 d4c5 d6e5 f4e5 e8g8 e5e6 f7e6 f1c4 d7c5 d1e2 d8b6 e1g1 b8c6 c1e3 b6b2 c3b5 c5e4 g1h1 e4c3 b5c3 b2c3 a1d1 c3a5 h2h3 a5h5 e2d2 h5a5');
INSERT INTO evaluations VALUES(42392,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(42392,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(42392,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(42392,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42392,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(42392,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(42392,36,'cp 22','b8c6 g2g3 d6b4 f1g2 d7d5 e4d5 f6d5 c1d2 d5c3 b2c3 b4a5 e1g1 e8g8 c3c4 a5b6 a2a4 c8g4 h2h3 g4h5 g3g4 h5g6 a4a5 b6c5 f3g5 d8d6 g5e4 g6e4 g2e4 a8d8 d2g5 f7f6 g5e3 b7b6 a5b6 a7b6 d1f3 c6d4 e3d4 c5d4 a1a7 g7g6 f1b1 d4c5');
INSERT INTO evaluations VALUES(10737,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(10737,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(10737,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(10737,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10737,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(10737,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(10737,36,'cp 27','g2g3 d6b4 f1g2 d7d5 e4d5 f6d5 c1d2 d5c3 b2c3 b4a5 e1g1 e8g8 c3c4 a5b6 a2a4 a7a5 d2c3 f7f6 f3d2 c8e6 f1e1 d8d7 d2b3 f8d8 a1b1 h7h6 c4c5 b6a7 d1e2 e6d5 g2d5 d7d5 c3a5 a7c5 a5c7 d8d7 b3c5 d7c7 b1b7 c7b7 c5b7 a8a4 c2c3');
INSERT INTO evaluations VALUES(44850,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(44850,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(44850,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(44850,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(44850,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(44850,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(44850,36,'cp 155','d5c6 b7c6 b5e2 e7e5 d2d3 f8c5 g1f3 b8d7 e1g1 d8c7 b1d2 a7a5 c2c3 e8g8 d3d4 c5a7 f3e5 d7e5 d4e5 c7e5 d2c4 e5e6 c4a5 a7f2 f1f2 a8a5 e2f3 a5a2 a1a2 e6a2 f3c6 a2c4 d1a4 c4c5 a4b5 c5b5 c6b5 f6e4 f2f1 c8e6 b5c6');
INSERT INTO evaluations VALUES(45441,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(45441,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(45441,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(45441,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(45441,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(45441,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(45441,36,'cp -10','c8f5 g1f3 b8c6 f3h4 f5e4 f2f3 e4g6 e2e4 e7e6 h4g6 h7g6 e4e5 f6d7 c1e3 d7e5 d4e5 d5d4 e3d4 d8d4 d1d4 c6d4 e1c1 a8d8 f3f4 f8e7 c3e4 f7f6 f1g2 h8h5 e4d6 e7d6 g2e4 e8e7 d1d4 f6e5 e4g6 e5d4 g6h5 d4d3 h5f3 d6c5 f3b7');
INSERT INTO evaluations VALUES(29136,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(29136,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(29136,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(29136,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(29136,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(29136,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(29136,36,'cp -9','g1f3 b8c6 f3h4 f5e4 f2f3 e4g6 e2e4 e7e6 h4g6 h7g6 e4e5 f6d7 c1e3 d7e5 d4e5 d5d4 e3d4 d8d4 d1d4 c6d4 e1c1 e8c8 f3f4 f8e7 c3e4 f7f6 e5f6 g7f6 e4c3 f6f5 c3b5 e7c5 b5d4 c5d4 f1c4 e6e5 f4e5 d4e5 d1e1 e5c7 c4f7');
INSERT INTO evaluations VALUES(49587,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(49587,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(49587,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(49587,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(49587,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(49587,35,'cp 264',NULL);
INSERT INTO evaluations VALUES(49587,36,'cp 283','d4e5 f6e4 d1d5 e4c5 c1g5 f8e7 h2h4 c7c6 g5e7 d8e7 d5d4 b8a6 b1c3 b7b5 h4h5 a6c7 e2e3 a7a5 e1c1 f7f5 e5f6 g7f6 f3h4 d7d5 e3e4 c5e4 c3e4 d5e4 d1e1 c6c5 d4e3 f6f5 f2f3 e8g8 f3e4 f5f4 e3d2 e7e6 f1d3 e6a2 d2c3 c8e6 c3c5 a2a1 c1d2');
INSERT INTO evaluations VALUES(49510,30,'cp 259',NULL);
INSERT INTO evaluations VALUES(49510,31,'cp 260',NULL);
INSERT INTO evaluations VALUES(49510,32,'cp 268',NULL);
INSERT INTO evaluations VALUES(49510,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(49510,34,'cp 286',NULL);
INSERT INTO evaluations VALUES(49510,35,'cp 291',NULL);
INSERT INTO evaluations VALUES(49510,36,'cp 271','f6e4 d1d5 e4c5 c1g5 f8e7 h2h4 c7c6 g5e7 d8e7 d5d4 b8a6 h4h5 b7b5 b1c3 a6c7 h5h6 g7g6 e1c1 a7a5 e2e3 b5b4 c3b1 a5a4 f1c4 c7e6 c4e6 c5e6 d4b6 e7c5 b6c5 e6c5 h1h4 b4b3 c2b3 a4b3 a2b3 c5b3 c1c2 a8b8 h4f4 b3c5 b1c3');
INSERT INTO evaluations VALUES(56554,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(56554,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(56554,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(56554,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(56554,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(56554,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(56554,36,'cp 7','e5d4 f1c4 f8b4 c1d2 b4d2 d1d2 b8c6 e1g1 g8f6 e3d4 e8g8 h2h3 a7a6 f1e1 c8f5 d2f4 f5b1 a1b1 d8d6 f4d6 c7d6 d4d5 c6b4 a2a3 a8c8 b1c1 b4d5 c4d5 f6d5 c1d1 d5c7 e1e7 d6d5 d1c1 c7e6 c1c8 f8c8 e7b7 h7h6 b7d7');
INSERT INTO evaluations VALUES(56656,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(56656,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(56656,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(56656,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(56656,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(56656,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56656,36,'cp 0','f1c4 f8b4 c1d2 b4d2 d1d2 b8c6 e1g1 g8f6 e3d4 e8g8 h2h3 a7a6 f1e1 d8d6 b1c3 h7h6 a1d1 c8f5 g2g4 f5h7 f3e5 a8d8 e5c6 d6c6 c4f1 c6d6 f2f4 c7c6 f1g2 g8h8 e1e5 f6d7 e5e1');
INSERT INTO evaluations VALUES(56644,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(56644,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(56644,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(56644,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(56644,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56644,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56644,36,'cp 4','f8b4 c1d2 b4d2 d1d2 b8c6 e1g1 g8f6 f3d4 c6d4 d2d4 d8d4 e3d4 c8e6 c4e6 f7e6 b1d2 e8c8 d2f3 d8d5 f1e1 h8e8 e1e2 a7a5 g2g3 c8d7 b2b3 d7d6 a1c1 h7h6 g1g2 c7c6 e2e3 f6d7 c1e1 d7f6 h2h3 f6d7 e3e2 d7f6 e1d1 f6d7 d1c1 e6e5 c1e1 b7b5 d4e5 d7e5');
INSERT INTO evaluations VALUES(9686,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(9686,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(9686,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(9686,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(9686,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(9686,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(9686,36,'cp 40','h2h3 e8g8 a2a4 h7h6 e1g1 d7d6 c1e3 g8h7 a4a5 c6e7 d3d4 e5d4 e3d4 e7c6 d4e3 c8e6 d1e2 a7a6 f3d4 c6d4 e3d4 f6h5 d4e3 d8h4 c4e6 f7e6 e2c4 h4e7 a1d1 h5f4 e3f4 f8f4 c3e2 f4f7 e2d4 e6e5 d4e6 a8e8 e6g7 h7g7 c4c3 e8d8 f2f4');
INSERT INTO evaluations VALUES(11368,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11368,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(11368,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(11368,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(11368,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(11368,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(11368,36,'cp 25','c6b4 e1g1 f8d6 c3e2 e8g8 a2a3 b4d3 c2d3 a6a5 d1d2 b7b6 b2b4 a5b4 f4d6 c7d6 d2b4 c8a6 b4d2 h7h6 f1b1 d8c7 b1b2 f8c8 a3a4 c7d8 h2h3 f6d7 a1a3 c8c6 f3e1 d8c7 a4a5 b6a5 a3a5 c6b6');
INSERT INTO evaluations VALUES(12329,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(12329,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(12329,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(12329,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(12329,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(12329,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(12329,36,'cp -19','f6g4 f1e1 d7d6 e5d6 f8d6 h2h3 g4e5 f3e5 c6e5 c4b5 c8d7 b1c3 e5c6 c3e4 d6e7 d2d3 c6d4 b5d7 d8d7 c1e3 b7b6 e3d4 c5d4 e4d2 a8c8 a2a4 e8g8 d2c4 c8c5 e1e4 f8c8 e4e5 c5e5 c4e5 d7d5 d1e2 c8c7 e5c4 c7c5 a1e1 e7f6 b2b3 h7h6 e2e4 d5e4');
INSERT INTO evaluations VALUES(12300,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(12300,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(12300,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(12300,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(12300,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(12300,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(12300,36,'cp -33','f1e1 d7d6 c4b5 c8d7 e5d6 f8d6 h2h3 g4e5 f3e5 c6e5 a2a4 e5c6 d2d3 e8g8 b1d2 c6d4 b5d7 d8d7 d2c4 b7b6 c1e3 d4c6 c4d6 d7d6 d1g4 a8e8 e3d2 e6e5 g4c4 h7h6 a4a5 c6d4 b2b4 d6g6 g1h1 c5b4 a5b6 g6b6 d2b4');
INSERT INTO evaluations VALUES(17529,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(17529,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(17529,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(17529,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(17529,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(17529,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(17529,36,'cp 12','c2c4 g8f6 b1c3 e7e5 f1d3 f8c5 a2a3 e8g8 g1f3 d7d6 h2h3 c8e6 b2b4 c5a7 e1g1 c6d4 c3d5 e6d5 c4d5 a8c8 c1d2 d8e7 g2g3 c8c7 a1c1 c7c1 d2c1 f8c8 c1d2 d4f3 d1f3 a7d4 f1c1 c8c1 d2c1 g7g6 g1g2 g8g7 c1d2 e7d8');
INSERT INTO evaluations VALUES(18088,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(18088,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(18088,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(18088,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(18088,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(18088,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(18088,36,'cp 148','c6a5 f3e5 a5b3 a2b3 g8e7 e5f3 c8b7 b1c3 d7d5 e4d5 d8d7 d2d4 e8c8 c3e4 e7d5 c2c4 d8e8 e4c3 d5c3 b2c3 h7h5 d4d5 h5h4 h2h3 h8g8 f3d4 g7g5 d1g4 d7g4 h3g4 f8d6 c1a3 d6e5 a3c5 g8h8 f1e1 c8d7 e1e3 e5f4 e3e6 b5c4 b3c4 f4d6 e6e8 h8e8 c5d6');
INSERT INTO evaluations VALUES(18435,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(18435,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(18435,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(18435,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(18435,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(18435,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(18435,36,'cp 163','d2d4 c5d4 f3d4 c8b7 c1f4 g7g6 d4c6 b7c6 d1d4 g8f6 c3b5 c6b5 f1b5 a7a6 b5e2 d7d6 e4e5 d6e5 d4e5 d8d5 e2f3 d5e5 f4e5 e8c8 a2a4 h8g8 e1g1 f6d7 e5g3 e7e6 a4a5 b6b5 f1d1 f8c5 c2c4 b5b4 g3d6 g8e8 a1a4 c5d6 d1d6 d7e5 d6a6 e5f3 g2f3');
INSERT INTO evaluations VALUES(18431,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(18431,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(18431,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(18431,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(18431,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(18431,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(18431,36,'cp 151','c5d4 f3d4 c8b7 c1f4 a7a6 d4c6 b7c6 c3d5 c6d5 d1d5 e7e6 d5b7 d8c8 b7b6 f8c5 b6a5 g8f6 f1d3 c8c6 a5c7 c6c7 f4c7 e8e7 f2f3 d7d5 e1c1 d5e4 d3e4 f6e4 f3e4 a8c8 c7f4 f7f6 e4e5 h7h5 d1d3 c8d8 h1d1 d8d3 d1d3 h8c8');
INSERT INTO evaluations VALUES(19821,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19821,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(19821,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(19821,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19821,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19821,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(19821,36,'cp 0','c2c3 a7a6 f1e2 c5d4 c3d4 c8f5 g1f3 e7e6 b1c3 g8e7 e1g1 f5e4 f3h4 e7g6 h4f3');
INSERT INTO evaluations VALUES(20592,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(20592,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(20592,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(20592,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(20592,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20592,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(20592,36,'cp 37','f8e7 d2d3 g8f6 c1g5 e8g8 g5f6 e7f6 c3d5 g7g6 c2c3 f6g7 a2a4 a8b8 h2h3 a7a6 b2b4 b7b5 a4b5 a6b5 c4b3 g8h8 d1d2 f7f5 d2e3 f5f4 e3e2 g6g5 a1a2 g5g4 h3g4 c5b4 d5b4 c6b4 c3b4 c8g4 a2a7 b8a8 a7a8 d8a8 b3d5');
INSERT INTO evaluations VALUES(24901,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(24901,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(24901,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(24901,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(24901,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(24901,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(24901,36,'cp 39','g1f3 e7e6 e1g1 f8d6 f4g5 h7h6 g5f6 d8f6 c2c4 e8g8 b1c3 d5c4 d3c4 e6e5 d4d5 c6e7 a1c1 f6g6 f1d1 f7f5 c4b3 g6f7 c3b5 e7g6 c1c2 f8d8 f3d2 d8d7 d2c4 d6e7 e3e4 f5e4 b5c3 b7b6 c4e3 a8d8 c3e4 g6f4 c2c6 f4d5 d1d5 d7d5');
INSERT INTO evaluations VALUES(25349,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(25349,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(25349,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(25349,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(25349,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(25349,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(25349,36,'cp 75','h2h3 g4f3 d1f3 c6d4 f3d1 g7g6 c2c3 d4e6 d1b3 h7h5 h3h4 e6c5 b3c2 c5e6 a2a4 f8g7 e1g1 e8g8 b1d2 g7h6 g2h3 f6g4 d2f3 h6c1 a1c1 c7c6 c1d1 d8b6 d3d4 a8e8 f1e1 e5d4 h3g4 h5g4 f3d4 e6d4 d1d4');
INSERT INTO evaluations VALUES(28174,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(28174,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(28174,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(28174,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(28174,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(28174,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(28174,36,'cp 133','f1c4 d7c6 c4d3 g8f6 d1e2 a7a6 e1g1 e8g8 a2a3 b8d7 f1e1 f8e8 e4e5 f6d5 c3e4 d7f8 h2h4 h7h6 c2c4 d5b6 h4h5 d6d5 c4d5 b6d5 c1d2 d8d7 g2g4 c6b5 g1h2 b5d3 e2d3 d7b5 d3c2 b5c6 c2c6 b7c6 a1d1 a6a5 e1g1 f8d7');
INSERT INTO evaluations VALUES(29504,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(29504,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(29504,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(29504,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(29504,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(29504,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(29504,36,'cp 24','g4e2 d1e2 d8d5 g1f3 e7e6 c2c4 f8b4 c1d2 b4d2 b1d2 d5d6 e1g1 e8g8 a1d1 d6b6 d2e4 f6e4 e2e4 b8d7 d1d2 h7h6 f1d1 a7a5 e4e2 c7c6 h2h4 a5a4 g2g3 b6a5 f3e5 a8d8 e5g4 f8e8 d2d3 h6h5 g4e5 d7e5 d4e5 d8d3 d1d3 b7b5 e2h5 a5e1 g1g2 b5c4');
INSERT INTO evaluations VALUES(31461,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31461,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31461,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31461,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31461,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(31461,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31461,36,'cp 3','f8e7 g5e7 d8e7 e2e3 g8f6 f1d3 d5c4 d3c4 e8g8 e1g1 f8d8 a1c1 b8d7 d1b3 b7b5 c4e2 a8c8 f3h4 a7a5 h4f5 e6f5 e2f3 g7g6 f1d1 a5a4 b3c2 g8g7 c3e2 f6d5 a2a3 d7f6 c2c5 d8d6 e2c3 d5c3 c1c3 f6d7 c5b4 d7f6');
INSERT INTO evaluations VALUES(28069,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(28069,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28069,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28069,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28069,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28069,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(28069,36,'cp 0','g5f4 d8b6 d1d2 g8f6 e2e3 e8g8 f1e2 b8d7 e1g1 a7a5 h2h3 f6e4 c3e4 f5e4 a1c1 e7b4 d2d1 b4e7');
INSERT INTO evaluations VALUES(36630,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(36630,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36630,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36630,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(36630,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(36630,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36630,36,'cp 11','f1e2 f8e8 e1g1 b8d7 c2c4 c7c6 a2a3 a7a5 b1c3 e7e5 d4e5 d6e5 c3a4 e5e4 f3d2 d7b6 a4b6 d8b6 d1c2 b6c5 a1d1 c5e7 c4c5 c8f5 b2f6 g7f6 b3b4 a5b4 a3b4 f6g7 d2c4 e7g5 g1h1 f5e6 c4d6 a8a2 c2e4');
INSERT INTO evaluations VALUES(40335,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40335,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(40335,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40335,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(40335,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(40335,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40335,36,'cp 0','c2c3 b8c6 f1e1 e8g8 a2a4 h7h6 c1e3 c8g4 b1d2 c6e5 h2h3 g4f3 d2f3 e5c4 d3c4 f6e4 e3h6 g7h6 e1e4 g8g7 d1d5 e7e6 d5b7 a8b8 b7a7 b8a8 a7b7 a8b8 b7a6 b8a8');
INSERT INTO evaluations VALUES(42195,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(42195,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(42195,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(42195,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(42195,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(42195,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(42195,36,'cp 153','b1d2 e8g8 f1d3 c7c5 d4c5 a5c7 f4g3 d8e7 b2b4 b7b6 c5b6 a7b6 e1g1 c7g3 h2g3 c8a6 d3a6 a8a6 a2a4 f8c8 d1b3 a6a7 f1c1 b8c6 a4a5 c8a8 f3d4 c6d8 c3c4 b6a5 b4b5 e6e5 d4f5 e7e6 b3b1 f6d7 d2f3 a7c7 c1d1 d5c4');
INSERT INTO evaluations VALUES(42194,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(42194,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(42194,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(42194,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(42194,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(42194,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(42194,36,'cp 148','c7c6 f3e5 b8d7 h2h4 a5c7 d1f3 d7e5 d4e5 f6d7 f3g3 g7g6 f1d3 d8e7 e1c1 h7h6 e3e4 a7a6 e4d5 c6d5 c3c4 d7c5 d3c2 b7b5 c4d5 e6d5 d2b3 c5b3 g3b3 c8e6 b3e3 e7b4 c1b1 a8c8 g2g3 c7a5 c2b3 e6f5 b1a1 b4e4 d1c1 e4e3 f4e3 a5c7 f2f4 e8g8 b3d5');
INSERT INTO evaluations VALUES(43546,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(43546,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(43546,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(43546,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(43546,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(43546,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43546,36,'cp 0','b7b5 a2a4 b5b4 c3a2 c8b7 c2c3 b4c3 a2c3 f8e7 e1g1 e8g8 b2b4 b8d7 f2f4 a8c8 a1c1 g7g6 g1h1 e6e5 d4b3 d6d5 b3a5 b7a8 c3d5 f6d5 e4d5 c8c1 e3c1 a8d5 d3g6 d8a8 g6h7 g8h7 d1h5 h7g8 h5g4 g8h7');
INSERT INTO evaluations VALUES(43805,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43805,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43805,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(43805,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43805,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(43805,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(43805,36,'cp 28','h2h4 c8b7 c2c3 f8d6 b1d2 d6f4 e3f4 c7c5 h4h5 e8g8 h5h6 g7g6 f1b5 d8d6 g2g3 c5d4 c3d4 f8c8 d1e2 b8c6 e1f1 c6e7 f1g2 e7f5 b5d3 c8c7 d3f5 e6f5 a1c1 c7c1 h1c1 f6g4 e2b5 g4h6 f3e5 a8c8 c1c8 b7c8 b5c6 d6c6 e5c6 c8d7 c6a7 h6g4 d2f3 f7f6');
INSERT INTO evaluations VALUES(46968,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(46968,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(46968,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(46968,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(46968,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(46968,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(46968,36,'cp -14','e7e5 g1f3 h7h6 g5f6 d8f6 b1c3 g7g6 c4c5 b8d7 c3b5 f6d8 c5d6 c7d6 b5c3 f8g7 e2e3 a7a6 f1c4 b7b5 c4d5 a8b8 d4e5 d6e5 e1g1 e8g8 d1c2 d7b6 d5b3 c8f5 e3e4 f5d7 f1d1 d8e7 h2h3 f8c8 c2d2 d7e6 b3e6 e7e6 b2b3 c8c7 c3d5 b6d5 d2d5 e6d5 d1d5');
INSERT INTO evaluations VALUES(47818,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(47818,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(47818,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(47818,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(47818,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(47818,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(47818,36,'cp 19','d2d3 e5f4 c1f4 d6d5 c4d5 f6d5 e4d5 d8d5 d1d2 c8g4 b1c3 d5e6 g1e2 b8c6 h2h3 g4e2 d2e2 e6e2 c3e2 e8c8 e1c1 f8e7 d1f1 f7f6 f1e1 h8e8 g2g4 c6b4 a2a3 b4d5 h1f1 g7g6 f4d2 e7d6 e2c3 e8e1 f1e1 d5c3 d2c3 d6f4 c1b1 f6f5 g4f5 g6f5');
INSERT INTO evaluations VALUES(47816,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(47816,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(47816,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(47816,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(47816,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(47816,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(47816,36,'cp 11','e5f4 c1f4 d6d5 e4d5 f6d5 c4d5 d8d5 d1d2 c8g4 b1c3 d5e6 g1e2 b8c6 e1c1 e8c8 h2h3 g4e2 d2e2 e6e2 c3e2 f8e7 d1f1 h8e8 f4d2 f7f6 g2g4 e7d6 e2c3 h7h6 c3e4 b7b6 f1f2 d6e5 h1f1 e5d4 f2e2 d4e5 b2b4 a7a5 b4a5 c6a5 d2a5 b6a5 e4c5');
INSERT INTO evaluations VALUES(52725,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(52725,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(52725,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(52725,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(52725,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52725,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(52725,36,'cp 0','e2e3 d8b6 d1c1 c8f5 c2c4 b8d7 g1f3 g8f6 b1c3 e7e6 f1e2 h7h6 g5f4 f6h5 c4c5 b6d8 f4e5 h5f6 e5g3 f6h5');
INSERT INTO evaluations VALUES(52723,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52723,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52723,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52723,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52723,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52723,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52723,36,'cp 0','d8b6 d1c1 c8f5 c2c4 b8d7 g1f3 g8f6 b1c3 e7e6 f1e2 h7h6 g5f4 f6h5 c4c5 b6d8 f4e5 h5f6 e5g3 f6h5');
INSERT INTO evaluations VALUES(57225,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(57225,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(57225,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(57225,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(57225,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(57225,35,'cp -107',NULL);
INSERT INTO evaluations VALUES(57225,36,'cp -88','d7d5 d2d4 b8c6 f1b5 c8d7 d4e5 d8g5 d1e2 g5e5 f3f4 e5e6 c2c4 g8f6 g1f3 a7a6 c4d5 f8b4 c1d2 f6d5 b5c4 b4d2 b1d2 c6b4 d2f1 d7b5 a1d1 e8c8 a2a3 b4c6 e1f2 b5c4 e2c4 f7f6 c4c2 h8e8 d1e1 e6d7 e1d1');
INSERT INTO evaluations VALUES(56481,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(56481,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(56481,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(56481,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(56481,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(56481,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(56481,36,'cp -92','d2d4 b8c6 f1b5 c8d7 d4e5 d8g5 d1e2 g5e5 f3f4 e5e6 c2c4 g8f6 b1c3 d5d4 e3d4 c6d4 e2e6 f7e6 b5d7 e8d7 a1b1 a8e8 g1h3 d7c8 e1g1 f8b4 f1e1 b4c3 b2c3 d4c6 h3g5 e6e5 g5f3 e5f4 c1f4 h8f8 e1e8 f8e8 b1e1 e8e4 e1e4 f6e4 f3g5 e4c3 g5h7');
INSERT INTO evaluations VALUES(4236,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(4236,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(4236,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(4236,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(4236,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(4236,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(4236,36,'cp 38','d7d5 c4b3 f6e4 b1c3 c8g4 c1e3 g4f3 g2f3 e4g5 f3f4 g5e6 d1g4 c6d4 b3d5 c7c6 d5e6 f7e6 a1d1 d8h4 g4h4 d4f3 g1g2 f3h4 g2h3 b6e3 f2e3 h4f5 d1d3 c6c5 b2b3 c5c4 b3c4 a8c8 c3e4 c8c4 e4g5 c4c6 e3e4 h7h6 g5f3 f5e7 f3d4 c6a6 h3g4 e7c6 d4c6 a6c6 d3d6 c6c4 d6e6 c4e4 e6e7');
INSERT INTO evaluations VALUES(4540,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(4540,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(4540,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(4540,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(4540,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(4540,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(4540,36,'cp 175','f3e5 a5c4 d3c4 d7d6 e5d3 c5d4 c3e2 d4b6 e2g3 c8g4 d1d2 f8e8 f1e1 h7h5 d2g5 f6d7 h2h3 d8g5 c1g5 g4e6 b2b3 f7f6 g5e3 b6a5 e1e2 g7g6 a1d1 e6f7 f2f3 a5c3 g3f1 a7a6 d3f4 b7b5 f4d5 f7d5 c4d5 a6a5 e3d4 c3d4 d1d4 b5b4 e2d2 g8f7 g1f2 h5h4 f1e3');
INSERT INTO evaluations VALUES(9593,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(9593,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(9593,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(9593,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(9593,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(9593,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(9593,36,'cp 32','e4f6 e7f6 h2h4 e8g8 d1e2 c7c5 e1c1 c5d4 e2e4 e6e5 g5f6 d8f6 f1b5 f6h6 c1b1 f8e8 f3d4 h6b6 e4e3 a7a6 b5a4 b6b4 a4b3 d7f6 a2a3 b4a5 d4f3 c8e6 b3e6 e8e6 f3g5 e6b6 d1d3 h7h6 g5e4 f6d5 e3d2 a5b5 b2b4 d5f4 d3d8 a8d8 d2d8 g8h7 e4d6 f4e6 d6b5 e6d8 h1d1 a6b5 d1d8 b6g6 d8b8 b7b6 h4h5 g6f6 b1c1 f6f2 b8b6');
INSERT INTO evaluations VALUES(9578,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(9578,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(9578,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(9578,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(9578,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9578,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(9578,36,'cp 26','d7f6 f1d3 c7c5 d4c5 d8a5 c2c3 a5c5 d1e2 c8d7 f3e5 a8d8 e1g1 e7d6 a1e1 c5e5 e2e5 d6e5 e1e5 d7c6 d3c2 h7h6 g5c1 e8g8 c3c4 d8c8 f1d1 b7b5 b2b3 f6d7 e5e1 b5c4 b3c4 f8d8 c1a3 d7b6 c2b3 d8d1 e1d1');
INSERT INTO evaluations VALUES(8703,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(8703,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(8703,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(8703,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(8703,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(8703,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(8703,36,'cp 26','f1d3 c7c5 d4c5 d8a5 c2c3 a5c5 d1e2 c8d7 f3e5 a8d8 e1g1 e7d6 a1e1 c5e5 e2e5 d6e5 e1e5 d7c6 d3e2 h7h6 g5e3 e8g8 c3c4 b7b6 e3c1 c6e4 f2f3 e4d3 g1f2 d3e2 f2e2 f6d7 e5e3 d7b8 b2b3 b8c6 c1b2 c6d4 b2d4 d8d4 e3d3 d4d8 a2a3 d8d3 e2d3');
INSERT INTO evaluations VALUES(8702,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(8702,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(8702,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(8702,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(8702,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(8702,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(8702,36,'cp 25','c7c5 d4c5 d8a5 c2c3 a5c5 d1e2 c8d7 f3e5 a8d8 e1g1 e7d6 a1e1 c5e5 e2e5 d6e5 e1e5 d7c6 d3c2 h7h6 g5c1 b7b5 f2f3 f6d7 e5e2 d7b6 f1d1 e8g8 c2d3 d8d7 e2e1 f8d8 d3e2 d7d1 e1d1 d8d1 e2d1 g8f8 d1e2 e6e5 h2h4 f7f6 b2b3 f8e7 g1f2');
INSERT INTO evaluations VALUES(14857,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14857,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(14857,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(14857,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14857,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(14857,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14857,36,'cp 52','c2c4 f8e7 a2a3 e8g8 b1d2 c6b8 e1g1 b7b6 b2b4 c7c5 d4c5 b6c5 d1b3 b8d7 c1b2 e7d6 b4b5 a7a6 a3a4 c8b7 f1c1 h7h6 b3c2 d6c7 b2a3 c7d6 h2h3 a6b5 a4b5 d8c7 a3b2 f8b8 c1b1 d7b6 a1a8 b8a8 b2f6 g7f6 c4d5 b6d5');
INSERT INTO evaluations VALUES(14782,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(14782,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(14782,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14782,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(14782,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14782,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(14782,36,'cp 39','f8d6 e1g1 e8g8 b2b3 b7b6 c1b2 c8b7 b1c3 c6e7 c3b5 e7g6 b5d6 c7d6 f3d2 a8c8 f1e1 f8e8 a2a4 h7h6 a4a5 d5c4 b3c4 d6d5 c4d5 d8d5 e2f1 d5g5 f2f3 b6b5 d2b3 f6d5 e3e4 d5f4 g2g3 e8d8 a5a6');
INSERT INTO evaluations VALUES(15397,30,'cp -493',NULL);
INSERT INTO evaluations VALUES(15397,31,'cp -481',NULL);
INSERT INTO evaluations VALUES(15397,32,'cp -485',NULL);
INSERT INTO evaluations VALUES(15397,33,'cp -499',NULL);
INSERT INTO evaluations VALUES(15397,34,'cp -502',NULL);
INSERT INTO evaluations VALUES(15397,35,'cp -508',NULL);
INSERT INTO evaluations VALUES(15397,36,'cp -494','f3g1 d8g5 e1f1 g5g6 h2h4 c6d4 d2d3 e4d3 h4h5 g6a6 d1d3 a6d3 c2d3 c8f5 b3a4 c7c6 h1h3 f5h3 g1h3 f7f6 c1e3 e8c8 a1c1 g7g6 c1c4 b7b5 a4b5 d4b5 c4c6 c8b8 c6f6 d8d3 f1e2 d3d6 f6d6 f8d6 h5g6 h7g6 e2d3 b8b7 h3g5 b7c6 g5e4 h8h1 e4d6 c6d6');
INSERT INTO evaluations VALUES(15398,30,'cp -490',NULL);
INSERT INTO evaluations VALUES(15398,31,'cp -503',NULL);
INSERT INTO evaluations VALUES(15398,32,'cp -505',NULL);
INSERT INTO evaluations VALUES(15398,33,'cp -508',NULL);
INSERT INTO evaluations VALUES(15398,34,'cp -505',NULL);
INSERT INTO evaluations VALUES(15398,35,'cp -503',NULL);
INSERT INTO evaluations VALUES(15398,36,'cp -514','d8g5 e1f1 g5g6 h2h4 c6d4 d2d3 e4d3 h4h5 g6a6 d1d3 a6d3 c2d3 c8f5 b3a4 c7c6 h1h3 f5h3 g1h3 f7f6 c1d2 f8c5 h3g1 e8c8 a1e1 g7g6 a4d1 c8b8 a2a4 d4f5 d1c2 g6h5 b2b4 c5e7 d2c3 f5d4 c3d4 d8d4 g1f3 d4g4 g2g3 e7b4 e1e4 g4e4 d3e4 b8c7 f1g2 h8e8 c2b3');
INSERT INTO evaluations VALUES(17154,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(17154,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(17154,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(17154,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(17154,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(17154,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(17154,36,'cp 62','d2d4 e5d4 c3d4 d7d5 e4d5 e7d5 e1g1 f8g7 f1e1 c8e6 c1g5 d8d6 b1d2 d6b4 b5c6 b7c6 d1c1 e8g8 d2e4 a8e8 a2a3 b4b5 g5h6 e6f5 h6g7 g8g7 e4g3 e8e1 f3e1 f5e6 c1d2 d5f6 e1d3 f8b8 b2b4 h7h6 a1e1 e6d5');
INSERT INTO evaluations VALUES(17148,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(17148,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(17148,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17148,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(17148,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(17148,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(17148,36,'cp 53','e5d4 c3d4 d7d5 e4d5 e7d5 e1g1 f8g7 b5c6 b7c6 c1g5 d8d6 f1e1 c8e6 a2a3 e8g8 b1d2 f8e8 d2e4 d6f8 d1a4 e6f5 e4g3 f5d7 h2h3 e8e1 a1e1 f8b8 a4c2 b8b5 f3e5 d7e8 g5d2 b5b6 g3e4 b6d4 e5d3 d5e7 b2b4 e7f5 d3c5 d4d5 e4c3');
INSERT INTO evaluations VALUES(17275,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(17275,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(17275,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(17275,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(17275,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(17275,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(17275,36,'cp -32','d2d3 e7e5 b1d2 d7d5 f1e2 f8d6 g1f3 e8g8 e1g1 f8e8 f1e1 h7h6 e2f1 a7a5 e4d5 f6d5 d2e4 c8f5 d1b3 d5b6 c1e3 f5e6 b3c2 b6d5 e4d6 c7d6 d3d4 e6g4 f3d2 d5e3 f2e3 d6d5 c2b3 c6e7 h2h3 g4f5 e3e4 f5e4 d2e4 d5e4 d4e5');
INSERT INTO evaluations VALUES(17276,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(17276,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(17276,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(17276,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(17276,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(17276,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(17276,36,'cp -33','d7d5 b1d2 e7e5 f1e2 h7h6 g1f3 f8d6 e1g1 e8g8 e4d5 f6d5 d2e4 f8e8 b2b4 c8f5 d1b3 d5f6 e4d6 c7d6 b3b2 d6d5 b4b5 c6a5 c3c4 d5c4 d3c4 d8c7 c4c5 c7c5 c1e3 c5e7 a1c1 b7b6 f1d1 a5b7 c1c6 b7d6');
INSERT INTO evaluations VALUES(17444,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(17444,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(17444,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(17444,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(17444,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(17444,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(17444,36,'cp -13','g8f6 f1d3 d7d6 e1g1 f8e7 a2a4 e8g8 h2h3 c8d7 c1e3 f8e8 f1e1 d8c7 f3d2 d6d5 e4d5 e6d5 e3g5 c6b4 d2f3 h7h6 g5e3 e7c5 e3c5 e8e1 f3e1 c7c5 c3e2 b4c6 d1d2 a8d8 e2g3 d7e6 e1f3 f6e4 d3e4 d5e4 g3e4 d8d2 e4c5 d2c2');
INSERT INTO evaluations VALUES(21416,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21416,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(21416,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(21416,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(21416,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(21416,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(21416,36,'cp 26','c3c4 g8f6 g1f3 d5c4 f1c4 f8d6 b1c3 e8g8 e3e4 e6e5 d4d5 c6e7 e1g1 h7h6 h2h3 a7a6 a2a4 c7c6 d5c6 e7c6 c1e3 d8e7 a1c1 f8e8 c3d5 f6d5 c4d5 c6b4 d5c4 c8e6 d1b3 e6c4 b3c4 a8c8 c4b3 c8c1 f1c1 b4c6 g2g3 c6d4 e3d4 e5d4');
INSERT INTO evaluations VALUES(22563,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(22563,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(22563,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(22563,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(22563,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(22563,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(22563,36,'cp -28','d7c6 d2d3 g8e7 g1e2 e7g6 c1e3 c8e6 b1c3 d8d7 f2f4 e5f4 e2f4 g6f4 e3f4 f8e7 d1f3 h8f8 f3e3 b7b6 e1c1 e8c8 h2h3 c8b7 d3d4 f8e8 c1b1 h7h5 a2a3 e7f6 c3e2 e6c4 h1f1 d7e6 e4e5 f6e7 f1e1 f7f6 e3f3 f6e5 f4e5');
INSERT INTO evaluations VALUES(22079,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(22079,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(22079,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(22079,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(22079,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(22079,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(22079,36,'cp -30','b1c3 g8e7 g1e2 e7g6 d2d3 c8e6 c1e3 d8d7 f2f4 e5f4 e2f4 g6f4 e3f4 f8e7 d1f3 h8f8 h2h3 g7g5 f4d2 g5g4 h3g4 e6g4 f3f2 h7h5 d2h6 f8g8 h1f1 f7f6 c3e2 b7b6 e1c1 e8c8 d1e1 c8b7 h6d2 e7c5 d2e3 c5d6 c1b1 d6e5 d3d4');
INSERT INTO evaluations VALUES(23697,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(23697,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(23697,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(23697,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(23697,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(23697,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(23697,36,'cp 88','g1h1 f6h5 f4e3 h5f6 c4d3 c6b4 d3e2 d6d5 e4e5 f6e4 f3g1 g4e2 g1e2 f7f6 a2a3 e4c3 e2c3 b4c6 e5e6 d8d6 d1f3 d6e6 c3d5 e7d6 c2c4 c6e7 d5e7 d6e7 e3g1 e6c4 a1c1 c4e6 c1c7 b7b6 f3b7 f8e8 f1f3 a8d8 b7a7 e7f8 a7a4 e6d5 a4c2');
INSERT INTO evaluations VALUES(29287,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(29287,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(29287,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(29287,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(29287,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(29287,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(29287,36,'cp 10','d1d3 f8b4 c1d2 b4a5 e1g1 e8g8 f1d1 f8e8 h2h3 b8d7 a2a4 a5c7 e3e4 d5e4 c3e4 e6e5 d2g5 e5d4 e4f6 d7f6 d3f5 d8d7 f3d4 d7f5 d4f5 f6e4 g5e3 c7b6 e3b6 a7b6 f2f3 e4c5 b2b4 c5a4 d1d7 a4b2 a1a8 e8a8 d7b7 b2c4 b7c7 c4e5 f5e7 g8h8 e7c6 e5c6 c7c6');
INSERT INTO evaluations VALUES(34705,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(34705,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(34705,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(34705,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(34705,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(34705,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(34705,36,'cp 152','b1c3 c7c6 c1e3 e7c7 h2h3 b7b5 c4b3 g8f6 a2a3 f8e7 g2g4 f6h7 d4d5 a7a6 d5c6 b8c6 c3d5 c7b7 d1d3 c8e6 e1c1 c6a5 b3a2 a8c8 c1b1 e8g8 h3h4 a5c4 a2c4 c8c4 d5e7 b7e7 d3d6 e7d6 d1d6 c4e4 g4g5 h6g5');
INSERT INTO evaluations VALUES(35569,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(35569,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35569,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35569,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35569,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35569,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(35569,36,'cp 0','a2a4 c8d7 h2h3 b8c6 g5e3 d8c7 d1e2 d6d5 e4d5 e6d5 d4c6 d7c6 e3d4 a8e8 e2f3 f6e4 c3e4 d5e4 d3e4 c6e4 f3e4 e7f6 e4d3 f6d4 d3d4 c7c2 f1c1 c2e2 d4b4 b7b5 a4b5 a6b5 c1e1 e2c4 b4c4 b5c4 e1e8 f8e8 a1a7 g7g6 a7b7 e8e2 b7c7 e2e1 g1h2 e1e2');
INSERT INTO evaluations VALUES(36230,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(36230,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(36230,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(36230,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(36230,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(36230,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(36230,36,'cp 45','d2d4 e5d4 f3d4 c8g4 c4e2 g4e2 d1e2 f6d7 b1c3 b8c6 d4c6 b7c6 c1d2 e7f6 a1d1 a7a5 c3a4 f8e8 b2b3 d8e7 e2d3 d7e5 d3c3 h7h6 c3h3 e5d7 d2c3 f6e5 c3e5 d7e5 h3g3 e5g6 f2f4 c6c5');
INSERT INTO evaluations VALUES(37643,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(37643,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(37643,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(37643,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(37643,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(37643,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(37643,36,'cp 167','d1g4 g7g6 a2a3 b4c3 b2c3 d7d5 g4f4 d5e4 f4e4 d8a5 c1d2 g8f6 e4h4 b8d7 f2f3 b7b6 a3a4 a5h5 h4h5 f6h5 a4a5 b6b5 e1f2 h5f6 g1e2 e8g8 e2c1 h7h5 c1d3 a7a6 a1b1 f8e8 c3c4 e6e5 d4e5 d7e5 d3e5 e8e5');
INSERT INTO evaluations VALUES(37935,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(37935,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(37935,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(37935,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(37935,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(37935,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(37935,36,'cp 90','e2e4 g8f6 f1e2 e8g8 e1g1 e7e5 d4d5 a7a5 c1e3 f6g4 e3g5 f7f6 g5d2 b8a6 h2h3 g4h6 a2a3 h6f7 b2b4 f6f5 d5c6 b7c6 b4b5 a6c5 b5c6 f5e4 f3h2 a8a6 c3d5 a6c6 d2a5 d8a5 d5e7 g8h8 e7c6 a5c7 c6b4 f7g5 e2g4 g5e6 g4e6 c8e6 b4d5 c7d8');
INSERT INTO evaluations VALUES(38821,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(38821,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(38821,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(38821,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(38821,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(38821,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(38821,36,'cp 51','d2c3 g8f6 e2e3 e8g8 g1f3 b7b6 b2b3 d8e7 c3b2 c8b7 f1d3 d5c4 d3c4 b8d7 e1g1 c7c5 d1e2 f8d8 a1c1 h7h6 f1d1 a8c8 h2h3 c5d4 d1d4 f6e4 c1d1 d7c5 d4d8 c8d8 d1d8 e7d8 e2c2 e4d6 c4f1 c5e4 f3e5 f7f6 e5c4 d8d7 c4d6 d7d6');
INSERT INTO evaluations VALUES(38816,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(38816,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(38816,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(38816,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(38816,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(38816,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(38816,36,'cp 47','g8f6 e2e3 b7b6 g1f3 e8g8 b2b3 d8e7 c3b2 c8b7 f1d3 d5c4 d3c4 b8d7 e1g1 f8d8 a1c1 c7c5 d1e2 f6e4 c4b5 a8c8 f1d1 d7f8 b5c4 f8g6 d4c5 e4c5 d1d8 c8d8 c1d1 h7h6 d1d8 e7d8 e2d2 d8d2 f3d2 b7a6 g1f1 a6c4 d2c4 c5d3 f1e2 d3b2');
INSERT INTO evaluations VALUES(42072,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(42072,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(42072,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(42072,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(42072,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(42072,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(42072,36,'cp 55','e2e3 e8g8 g1f3 b7b6 b2b3 c8b7 c4d5 d8d5 f1c4 d5d6 a1c1 b8d7 d1e2 c7c5 d4c5 d7c5 e1g1 f6g4 e2d1 d6e7 d1d4 g4f6 d4h4 f8d8 c3d4 a8c8 f3e5 f6d5 h4e7 d5e7 f1d1 c5d7 h2h3 d7e5 d4e5 e7f5 e5b2 h7h6 d1d8 c8d8 b3b4');
INSERT INTO evaluations VALUES(43095,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(43095,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(43095,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(43095,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(43095,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(43095,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(43095,36,'cp 52','a7a5 f1d3 b8c6 e1g1 d7d6 e3e4 f5e4 c3e4 c8g4 d4d5 g4f3 d1f3 c6e5 f3e2 e5d3 e4f6 g7f6 e2d3 e8g8 c1h6 f8f7 a1e1 f6b2 e1b1 b2g7 h6g7 g8g7 b1b7 d8c8 b7b5 e7e5 c4c5 a8b8 b5a5 c8g4 c5d6 c7d6 a5a6 g4d4 d3d4 e5d4 a6d6 b8b2');
INSERT INTO evaluations VALUES(43152,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43152,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(43152,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(43152,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(43152,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(43152,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(43152,36,'cp 69','d2d4 e8g8 c1e3 d7d6 h2h3 b8a6 g2g4 e7e5 d4d5 a6c5 f2f3 f6e8 d1d2 d8h4 e3f2 h4e7 h3h4 f7f5 g4f5 g6f5 e1c1 f5e4 f2c5 d6c5 c3e4 e8d6 f1d3 b7b5 c4b5 a7a6 b5b6 c5c4 d3c2 a8b8 h4h5 b8b6 h5h6 d6e4 f3e4 g7f6');
INSERT INTO evaluations VALUES(43768,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(43768,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(43768,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(43768,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(43768,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(43768,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(43768,36,'cp 109','c4d5 f6d5 e2e4 d5c3 b2c3 c8b7 f1d3 e7e5 g1f3 b8d7 e1g1 f8d6 d1e2 d8e7 a2a4 e8g8 f1e1 a7a5 c1g5 f7f6 g5h4 g8h8 h4g3 a8e8 d3b5 c7c6 b5c4 e5d4 c3d4 d6b4 f3h4 g7g6 h4f3 b4e1 a1e1 b6b5 a4b5 c6b5 c4b5 e7e4 e2e4 e8e4 b5d7 e4e1 f3e1 b7a6 d7c6 f8d8');
INSERT INTO evaluations VALUES(45903,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(45903,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(45903,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(45903,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(45903,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(45903,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(45903,36,'cp 208','f3g5 d6d5 e4d5 h7h6 g5f3 f8d6 e1g1 e8g8 f1e1 b8d7 b1c3 f8e8 a2a4 d7b6 f3d2 d6c7 d2e4 b6c4 d3c4 c8f5 f2f3 f5g6 c1d2 f6e4 f3e4 d8d7 d1f3 a7a6 b2b3 c7a5 e1f1 f7f6 g2g4 a5b4 g1h1 b4c3');
INSERT INTO evaluations VALUES(45897,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(45897,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(45897,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(45897,33,'cp 216',NULL);
INSERT INTO evaluations VALUES(45897,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(45897,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(45897,36,'cp 207','d6d5 e4d5 h7h6 g5e4 f6d5 b1c3 c8e6 d1f3 d5c7 c1e3 b8c6 e1c1 d8d7 e4c5 f8c5 e3c5 e8c8 f3g3 b7b6 c5a3 f7f5 f2f4 e5f4 g3f4 c6d4 f4f2 h8e8 b2b3 g7g5 h1e1 d7c6 e1e5 f5f4 c3e4 e6c4 e5e8 c7e8 b3c4 e8f6 e4f6 c6f6 a3b2 g5g4 d1e1 f4f3');
INSERT INTO evaluations VALUES(46358,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(46358,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(46358,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(46358,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(46358,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(46358,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(46358,36,'cp 73','e1g1 b8c6 c2c3 f8d6 c1g5 e8g8 b1d2 h7h6 g5h4 f8e8 f1e1 e6e5 f3e5 c6e5 d4e5 e8e5 d1b3 e5e1 a1e1 g7g5 h4g3 d6g3 h2g3 d8b6 b3b6 a7b6 a2a3 c8d7 f2f3 d7a4 e1e7 f6d7 d3f5 a8d8 g3g4 d7c5 e7c7 d5d4');
INSERT INTO evaluations VALUES(46357,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(46357,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(46357,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(46357,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(46357,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46357,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(46357,36,'cp 49','b8c6 f1e1 f8d6 c2c3 e8g8 c1g5 h7h6 g5h4 d8c7 b1d2 f6h5 g2g3 f7f5 f3e5 d6e5 d4e5 g7g6 f2f4 c7b6 g1h1 b6b2 c3c4 c8d7 c4d5 c6b4 a1b1 b2c3 b1b3 c3c5 h4e7 c5e7 d5d6 b4d3 d6e7 d3f2 h1g1 f2d1 e7f8r a8f8 b3a3 d1b2 e1b1 b2a4 b1b7 f8d8 b7a7');
INSERT INTO evaluations VALUES(13677,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(13677,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(13677,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(13677,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(13677,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(13677,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(13677,36,'cp 62','f1e1 f8d6 c2c3 d8c7 c1g5 f6h5 f3e5 g7g6 e5f3 e8g8 b1a3 a7a6 a3c2 c8d7 g2g3 f7f6 g5h6 f8e8 h6d2 c6d8 a1c1 d8f7 b2b3 d7c6 c3c4 c7d7 d3f1 h5g7 c2e3 d6f8 f1g2 g7f5 e3f5 e6f5 h2h4 d5c4 e1e8 a8e8');
INSERT INTO evaluations VALUES(46992,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46992,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46992,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(46992,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46992,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46992,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(46992,36,'cp 0','e7e5 d4e5 f6h5 e2e3 h5f4 e3f4 b8c6 b1c3 c8f5 g1f3 d6e5 f4e5 f8b4 f1e2 b4c3 b2c3 e8g8 f3d4 d8d7 f2f4 a8d8 e1g1 c6d4 d1d4 d7d4 c3d4 d8d4 a1d1 f8d8 d1d4 d8d4 f1d1 d4d1 e2d1 g8f8 g1f2 b7b6 f2e3 f8e7 d1e2 f5c8 g2g4 h7h6 h2h4 c8b7 e2f3 b7c8 f3e2');
INSERT INTO evaluations VALUES(50737,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(50737,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50737,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50737,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(50737,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(50737,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(50737,36,'cp -20','b7b5 c4a2 g8f6 d2d4 b5b4 c3a4 d8a5 a2b3 f6e4 a3b4 a5b4 c2c3 b4b7 d4d5 e6e5 b3c2 e4f6 e1g1 f8e7 c3c4 c8g4 h2h3 g4h5 g2g4 h5g6 a4c3 e8g8 a1a3 a6a5 f3h4 g6c2 d1c2 f6g4 h4f5 g4f6 c1g5 b8a6 g5f6 e7f6 f5d6 b7d7 c3b5');
INSERT INTO evaluations VALUES(51270,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51270,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51270,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(51270,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51270,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51270,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51270,36,'cp 0','c4a2 g8f6 d2d4 c5d4 f3d4 f8e7 d1e2 c8d7 c1e3 b8c6 e1g1 d8c7 g1h1 c7b7 f2f4 e8g8 f4f5 c6d4 e3d4 a8e8 e2f3 e6f5 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(51273,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51273,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(51273,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(51273,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(51273,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(51273,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(51273,36,'cp 0','g8f6 d2d4 c5d4 f3d4 f8e7 d1e2 c8d7 c1e3 b8c6 f2f4 e8g8 e1g1 d8c7 g1h1 a8c8 a1e1 c7b7 d4c6 d7c6 e3d4 f8e8 b2b4 b7d7 e1d1 d7b7');
INSERT INTO evaluations VALUES(50797,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(50797,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(50797,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(50797,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(50797,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(50797,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(50797,36,'cp 80','f8b4 b1c3 c8d7 f1d3 g8e7 d3c2 b8c6 c1d2 e7f5 a2a3 b4e7 c3e2 f5h4 f3h4 e7h4 e1g1 a8c8 g2g3 h4e7 h2h4 c6a5 d2c3 a5c4 c2d3 b7b5 b2b3 c4a3 d1d2 e8g8 c3a5 d8e8 e2f4 b5b4 a5b4 e7b4 d2b4 a3c2 d3c2 c8c2 f1c1 c2c8 b4d6 d7b5 c1c7 c8c7 d6c7 e8d8 a1c1 d8c7');
INSERT INTO evaluations VALUES(54383,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(54383,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(54383,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(54383,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(54383,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(54383,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(54383,36,'cp -25','d5e4 c3e4 g8f6 e4f6 d8f6 g1f3 f8b4 c1d2 b4d2 d1d2 c8d7 f1e2 e8g8 e1g1 d7c6 f3e5 f8d8 d2e3 c6d5 e2b5 a7a6 b5a4 f6e7 a2a3 b7b5 a4b3 f7f6 e5f3 b8c6 f1e1 d5b3 e3b3 d8d6 a3a4 a8b8 a4b5 a6b5 a1c1 e7d7 h2h3 h7h6 e1d1 d6d5 c1c5 c6e7 d1c1');
INSERT INTO evaluations VALUES(1900,30,'cp -445',NULL);
INSERT INTO evaluations VALUES(1900,31,'cp -450',NULL);
INSERT INTO evaluations VALUES(1900,32,'cp -453',NULL);
INSERT INTO evaluations VALUES(1900,33,'cp -452',NULL);
INSERT INTO evaluations VALUES(1900,34,'cp -446',NULL);
INSERT INTO evaluations VALUES(1900,35,'cp -471',NULL);
INSERT INTO evaluations VALUES(1900,36,'cp -475','b4f4 e2e3 f4h6 g2g4 d7d6 g4g5 h6h5 f1e2 d6e5 c2c3 h5g6 d1a4 c8d7 h1g1 c6d4 a4d7 e8d7 f3e5 d7e8 e5g6 d4e2 g1g4 h7g6 e1e2 h8h2 d2f3 h2h5 g4d4 f8e7 a1g1 c7c6 d4d1 a8d8 a2a3 d8d1 g1d1 e7g5 f3g5 h5g5 b2b3');
INSERT INTO evaluations VALUES(1910,30,'cp -447',NULL);
INSERT INTO evaluations VALUES(1910,31,'cp -471',NULL);
INSERT INTO evaluations VALUES(1910,32,'cp -450',NULL);
INSERT INTO evaluations VALUES(1910,33,'cp -458',NULL);
INSERT INTO evaluations VALUES(1910,34,'cp -428',NULL);
INSERT INTO evaluations VALUES(1910,35,'cp -447',NULL);
INSERT INTO evaluations VALUES(1910,36,'cp -447','e2e3 f4h6 c2c3 f7f6 h2h4 f6e5 f1c4 g8e7 c4e2 e7f5 d1a4 f8e7 e1c1 f5d6 f3g5 b7b5 e2b5 d6b5 a4b5 h8f8 f2f4 a8b8 b5d5 a7a6 d5e4 c8b7 e4h7 e5f4 e3f4 h6h7 g5h7 f8f4 d1e1 e8d8 h4h5 d8c8 h5h6 f4g4 h6g7 g4g7 d2e4 c6e5');
INSERT INTO evaluations VALUES(1909,30,'cp -477',NULL);
INSERT INTO evaluations VALUES(1909,31,'cp -469',NULL);
INSERT INTO evaluations VALUES(1909,32,'cp -474',NULL);
INSERT INTO evaluations VALUES(1909,33,'cp -478',NULL);
INSERT INTO evaluations VALUES(1909,34,'cp -457',NULL);
INSERT INTO evaluations VALUES(1909,35,'cp -475',NULL);
INSERT INTO evaluations VALUES(1909,36,'cp -481','f4h6 f1c4 d7d6 e5d6 f8d6 d1e2 c8e6 e1c1 e8c8 h2h4 e6c4 e2c4 h6g6 g2g4 g8f6 g4g5 f6d7 h4h5 g6f5 c4e4 f5e4 d2e4 d6e7 a2a3 f7f5 e4g3 h8f8 g3e2 e7d6 f3h4 f5f4 d1d6 c7d6 e2f4 d8e8 c1b1 h7h6 h4g6 f8f5 g5h6 g7h6 b2b3 c8c7');
INSERT INTO evaluations VALUES(3487,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(3487,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(3487,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(3487,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(3487,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(3487,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(3487,36,'cp 40','d4d5 c6a5 b1a3 c7c5 f1e1 e7e6 e2e4 e6d5 e4d5 a7a6 c1d2 c8f5 d2c3 b7b6 b2b3 f8e8 e1e8 f6e8 c3g7 e8g7 d1d2 d8f6 a1f1 a8e8 a3b1 f5e4 f3g5 e4g2 g1g2 g7f5 b1c3 f5d4 f2f3 f6e5 f1f2 f7f5 c3a4 b6b5 c4b5 a6b5 d2a5 b5a4 a5a4');
INSERT INTO evaluations VALUES(3871,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(3871,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(3871,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(3871,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(3871,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(3871,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(3871,36,'cp 43','c3d4 c5b6 h2h3 f6e4 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 c1g5 f7f6 g5f4 e7g6 f4e3 c8f5 e4c4 b6e3 f2e3 d8d6 a2a4 a8c8 a1c1 g6e5 f3e5 d6e5 d1d4 e5d4 c4d4 c7c6 e3e4 f5d7 a4a5 c6d5 e4d5 c8c5 d4b4 c5a5 b4b7');
INSERT INTO evaluations VALUES(3870,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(3870,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(3870,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(3870,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(3870,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(3870,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(3870,36,'cp 42','c5b6 h2h3 f6e4 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 c8f5 e4e1 b6a5 c1g5 f7f6 d1b3 f8f7 g5f4 a5c3 b2c3 d8d5 c3c4 d5c5 f4e3 c5c6 f3d4 c6b6 d4b5 b6c6 b5a7 c6a6 a7b5 e7g6 a4a5 f7d7 b5d4 f5e4 f2f3 c7c5 d4b5 e4d3 b5c7 d7c7 b3d3 a6c6 d3d5 c6d5 c4d5');
INSERT INTO evaluations VALUES(3667,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(3667,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(3667,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(3667,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(3667,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(3667,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(3667,36,'cp 42','h2h3 f6e4 e1e4 d6d5 c4d5 d8d5 b1c3 d5d8 d4d5 c6e7 a2a4 c8f5 e4e2 e7g6 f3d4 d8f6 c1e3 g6h4 d4f5 h4f5 e3f4 f6d4 d1d4 f5d4 e2e7 d4f5 e7e4 a7a5 g2g4 f5d6 f4d6 c7d6 e4e7 a8b8 a1e1 f7f5 g4g5 b6d8 e7d7 d8g5 e1e6 f8d8 e6d6 g5f6 c3b5');
INSERT INTO evaluations VALUES(5392,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5392,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(5392,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(5392,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5392,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(5392,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5392,36,'cp 0','e2e3 g8f6 f1e2 e8g8 d4f3 b7b6 e1g1 c8b7 a1b1 a8c8 b2b4 e7e6 d1b3 d7d5 c4d5 f6d5 c3d5 d8d5 b3d5 e6d5 f1d1 c6e5 f3d4 e5c4 a2a4 f8d8 e2g4 c8c7 d4b5 c7e7 g4e2 d8a8 e2f3 a8d8');
INSERT INTO evaluations VALUES(5391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5391,36,'cp 0','g8f6 f1e2 e8g8 d4f3 d7d6 e1g1 c8f5 f3g5 a8c8 e3e4 f5d7 c1e3 h7h6 g5f3 f6g4 e3d2 g4e5 a1c1 a7a6 b2b3 e5f3 e2f3 c6d4 d2e3 d4f3 d1f3 b7b5 c3d5 e7e6 e3b6 d8g5 b6e3');
INSERT INTO evaluations VALUES(8040,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8040,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8040,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8040,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8040,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8040,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8040,36,'cp 0','f1e2 e8g8 d4f3 d7d6 e1g1 c8f5 f3g5 a8c8 e3e4 f5d7 c1e3 h7h6 g5f3 f6g4 e3d2 g4e5 a1c1 a7a6 b2b3 e5f3 e2f3 c6d4 d2e3 d4f3 d1f3 b7b5 c3d5 e7e6 e3b6 d8g5 b6e3');
INSERT INTO evaluations VALUES(7405,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(7405,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(7405,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(7405,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(7405,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(7405,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(7405,36,'cp 60','b3c2 c7c5 d2d4 c5d4 c3d4 e8g8 h2h3 e5d4 f3d4 f8e8 b1c3 c8d7 c1f4 a8c8 a1c1 e7f8 c2b1 a5c6 d4c2 c6e5 c2e3 d7e6 f4g5 h7h6 g5h4 e5g6 h4g3 b5b4 c3d5 c8c1 d1c1 e6d5 e4d5 g6e7 b1d3 e7d5 e3d5 f6d5 e1e8 d8e8 d3a6 d5f6 a6f1');
INSERT INTO evaluations VALUES(7406,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(7406,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(7406,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(7406,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(7406,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(7406,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(7406,36,'cp 42','c7c5 d2d4 c5d4 c3d4 e8g8 h2h3 e5d4 f3d4 c8d7 b1c3 f8e8 c3e2 e7f8 e2g3 a8c8 a2a4 a5c6 d4c6 d7c6 a4b5 a6b5 c1d2 d6d5 e4e5 f6e4 d2f4 f8b4 e1f1 b4f8 g3f5 d8d7 c2e4 d5e4 f5d6 f8d6 e5d6 c8d8');
INSERT INTO evaluations VALUES(7512,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(7512,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(7512,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(7512,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(7512,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(7512,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(7512,36,'cp 43','d2d4 c5d4 c3d4 e8g8 h2h3 e5d4 f3d4 c8d7 b1d2 f8e8 d2f1 a8c8 a2a4 a5c6 d4c6 d7c6 a4b5 c6b5 c1d2 d6d5 d2a5 d8d7 e4e5 f6e4 c2e4 d5e4 d1d7 b5d7 a5c3 c8a8 e1e4 e7f8 e4f4 d7e6 f1e3 h7h6 a1d1');
INSERT INTO evaluations VALUES(7510,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(7510,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(7510,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(7510,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(7510,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(7510,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(7510,36,'cp 47','c5d4 c3d4 e8g8 h2h3 e5d4 f3d4 f8e8 b1c3 c8d7 c3e2 e7f8 e2g3 a8c8 b2b3 a5c6 d4c6 d7c6 c1b2 c6b7 a2a4 d6d5 a4b5 a6b5 e4e5 f6e4 g3e4 d5e4 c2e4 b7e4 e1e4 d8d1 a1d1 c8c2 b2d4 e8d8 e4e3 g7g6 e3d3 f8c5 d4c5 d8d3 d1d3 c2c5 f2f4 g8f8 g2g4 c5c2 f4f5 g6f5');
INSERT INTO evaluations VALUES(7699,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(7699,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(7699,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(7699,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(7699,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(7699,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(7699,36,'cp -46','c8d7 c6a6 e8g8 b1c3 h7h6 g5f3 f8e8 e1g1 e7c5 d2d4 c5d4 f3d4 e5d4 c3e2 d7f5 f1d1 d4d3 e2g3 e8e6 a6b5 d8e8 b5f5 e6e1 d1e1 e8e1 g3f1 d3d2 c1d2 e1a1 d2c3 a8d8 c3f6 g7f6 f5f6 d8d1 f6a6 h6h5 a6b5 a1c1 b5e8 g8g7 e8e5 g7h6 e5e2 d1e1 e2d3 c1g5 d3c4');
INSERT INTO evaluations VALUES(8833,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(8833,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(8833,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8833,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(8833,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(8833,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(8833,36,'cp -13','c5e3 f2e3 c6a5 c4b3 e8g8 d1d2 c8d7 e1g1 a5b3 a2b3 a7a5 c3e2 c7c5 f1f2 b7b5 a1f1 a5a4 b3a4 b5a4 c2c4 a4a3 f1a1 a3b2 d2b2 d8b8 b2c1 a8a1 c1a1 b8b7 e2c3 f8a8 f2a2 a8a2 a1a2 b7b4 a2c2 d7e6 c3d5 e6d5');
INSERT INTO evaluations VALUES(8888,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(8888,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(8888,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8888,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(8888,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(8888,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(8888,36,'cp -12','f2e3 c6a5 c4b3 e8g8 e1g1 a5b3 a2b3 a7a5 b3b4 c7c6 d1e1 c8e6 b4a5 a8a5 a1a5 d8a5 c3d5 a5e1 d5f6 g7f6 f1e1 f8a8 g1f2 h6h5 f3h4 g8f8 f2e2 f8e7 e2f3 a8a4 g2g4 h5g4 h3g4 d6d5 h4f5 e6f5 g4f5 d5e4 d3e4 c6c5 e1b1 b7b5 b1d1 a4a2 d1b1 c5c4');
INSERT INTO evaluations VALUES(8885,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(8885,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(8885,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(8885,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(8885,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(8885,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(8885,36,'cp -12','c6a5 c4b3 a5b3 a2b3 e8g8 e1g1 a7a5 d1e1 c7c6 b3b4 c8e6 a1a5 a8a5 b4a5 d8a5 c3d5 a5e1 d5f6 g7f6 f1e1 f8a8 g1f2 g8g7 f3h4 h6h5 c2c4 a8a4 e1d1 g7f8 f2f3 a4b4 d1d2 f8e7 g2g4 h5g4 h3g4 b7b5 c4b5 b4b5 h4f5 e7d7 d2f2 b5b3 f2d2');
INSERT INTO evaluations VALUES(10499,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(10499,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(10499,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(10499,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(10499,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(10499,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(10499,36,'cp -58','f6e4 e1g1 e4c3 b2c3 b4e7 d1h5 e8g8 d4f3 d7d5 f1d1 c8e6 c4d3 g7g6 h5h6 e7f6 f3g5 f6g5 c1g5 f7f6 g5f4 d8d7 d1b1 a8b8 b1e1 b8e8 h2h4 e8e7 e1e3 e6g4 a1e1 e7e3 e1e3 g4f5 d3f5 d7f5 f4c7 f5c2 c7d6 f8d8 e3e1 c2f5 d6c5 b7b6');
INSERT INTO evaluations VALUES(10544,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(10544,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(10544,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(10544,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(10544,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(10544,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(10544,36,'cp 36','b2b4 a5b6 c3a4 d7d6 f1e2 c6e7 e1g1 e7g6 a4b6 a7b6 c2c4 h7h6 a3a4 e8g8 h2h3 f6h7 c1e3 h7g5 f3g5 h6g5 d3d4 g6f4 f1e1 f4e2 d1e2 e5d4 e3d4 f8e8 a4a5 b6a5 e2b2 f7f6 b4a5 d8e7 c4c5 d6c5 d4c3 e7f7');
INSERT INTO evaluations VALUES(10536,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(10536,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(10536,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(10536,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(10536,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(10536,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(10536,36,'cp 30','a5b6 c3a4 d7d6 f1e2 c8g4 c2c3 e8g8 e1g1 h7h6 a4b6 a7b6 h2h3 g4d7 c1e3 c6e7 a3a4 e7g6 c3c4 f6h7 g1h2 c7c5 d1b3 h7g5 f3g5 h6g5 a4a5 b6a5 b4a5 a8a7 g2g3 d7c6 f1b1 a7a5 a1a5 d8a5 e3g5');
INSERT INTO evaluations VALUES(9919,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(9919,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(9919,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(9919,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(9919,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(9919,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(9919,36,'cp 44','c3a4 d7d6 f1e2 e8g8 e1g1 c6e7 c2c4 e7g6 a4b6 a7b6 h2h3 h7h6 a3a4 f6h7 c1e3 f7f5 e4f5 c8f5 d1b3 f5d7 d3d4 e5e4 c4c5 g8h8 f3d2 d8e8 c5d6 g6h4 d6c7 e8g6 g2g3 d7h3 d4d5 a8e8 d5d6 h7g5 b4b5 g5f3 d2f3 h4f3 e2f3 e4f3 f1d1 g6g4 e3b6');
INSERT INTO evaluations VALUES(11289,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(11289,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(11289,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(11289,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(11289,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(11289,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(11289,36,'cp 78','c2c3 e7g6 d2d4 c8d7 d4d5 c6b8 c3c4 f8e7 b1c3 e8g8 a4d7 b8d7 a2a4 a6a5 g2g3 d8c8 b2b3 d7b8 c1e3 f7f5 e4f5 c8f5 f3d2 b8d7 d2e4 f8f7 f2f3 f5h5 d1e2 a8f8 g1g2 d7f6 f1f2 b7b6 e4f6 g7f6 a1e1 f6f5 f3f4 h5e2 e1e2');
INSERT INTO evaluations VALUES(11287,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(11287,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(11287,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(11287,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(11287,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(11287,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(11287,36,'cp 59','c8d7 d2d4 e7g6 d4d5 c6b8 c3c4 h7h6 b1c3 f8e7 b2b4 e8g8 c1e3 d7a4 d1a4 b8d7 a4c2 b7b6 a2a4 a6a5 b4a5 a8a5 c3b5 a5a8 g2g3 d7f6 h2h3 d8d7 g1g2 f6h7 a1a3 e7g5 f3g5 h7g5 e3g5 h6g5 c2e2 f7f5 e4f5');
INSERT INTO evaluations VALUES(13013,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(13013,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13013,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13013,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13013,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13013,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(13013,36,'cp 0','d3d4 c5d4 f3d4 f8g7 d4c2 e8g8 f1e2 f6d7 e1g1 d7c5 f2f3 f7f5 e4f5 c8f5 c2e3 f5d7 e3c2');
INSERT INTO evaluations VALUES(17461,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(17461,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(17461,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(17461,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(17461,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(17461,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(17461,36,'cp 20','f8c5 f1c4 g8f6 c1g5 h7h6 g5f6 d8f6 a2a4 e8g8 c3d5 f6d8 d1d3 g8h8 e1g1 d7d6 c2c3 a6a5 f3d2 c6e7 d2b3 c5a7 d5e7 d8e7 g1h1 e7c7 c4d5 f7f5 b3d2 f5f4 f2f3 c8d7 d2c4 a8c8 b2b3 b7b5 a4b5 d7b5 d5e6 c8b8 f1d1 a7c5');
INSERT INTO evaluations VALUES(22117,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(22117,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(22117,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(22117,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(22117,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(22117,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(22117,36,'cp 135','f3d4 e5d4 c3e2 g8f6 e2d4 f6e4 d1e2 d8e7 e1g1 c7c6 f1e1 d6d5 c4d5 c6d5 e2b5 e7d7 d2d3 f7f5 d3e4 d5e4 e1d1 a7a6 b5b3 f8e7 c1f4 e7f6 c2c3 g7g5 d4f3 d7b5 f4g5 b5b3 a2b3 f6g5 f3g5 e8e7 d1d4 h7h6 g5h3 c8e6 h3f4 h8d8 d4b4 b7b5 f2f3');
INSERT INTO evaluations VALUES(27836,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(27836,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(27836,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(27836,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(27836,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(27836,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(27836,36,'cp 57','f1e2 g4f3 e2f3 f6d7 e1g1 b8c6 c3e2 a7a5 d1d2 a5a4 b2b4 a4b3 a2b3 f8e8 h2h4 a8a1 f1a1 h7h6 g5h6 e7e5 d4d5 c6d4 e2d4 e5d4 h4h5 g7h6 d2h6 d8h4 g2g3 h4h3 a1d1 d7e5 f3g2 h3h5 h6h5 g6h5 f2f4 e5d7 b3b4 e8a8 g2h3 d7f6 h3f5 a8a3 d1d4 a3g3 g1f2');
INSERT INTO evaluations VALUES(27835,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(27835,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(27835,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(27835,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(27835,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(27835,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(27835,36,'cp 42','a7a6 g5e3 f6d7 d1d2 e7e5 h2h4 d7f6 d4e5 d6e5 d2d8 f8d8 c3d5 b8c6 e3g5 f6e4 g5d8 a8d8 a1d1 g4d7 f3g5 e4c5 h4h5 c6d4 h5h6 g7h8 d5c7 h8f6 g5f3 d4c2 e1f1 e5e4 f3e1 c2e1 f1e1 f6b2 c7d5 d8e8 e1f1 d7a4 d1d2 b2e5 g2g3 a4c6');
INSERT INTO evaluations VALUES(31750,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(31750,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(31750,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(31750,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(31750,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(31750,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(31750,36,'cp 37','e7e6 c2c4 f8d6 f4d6 d8d6 d1b3 b7b6 c4d5 e6d5 b1c3 g8e7 c3b5 d6d8 b3a3 a7a6 a1c1 b8c6 b5c3 e8g8 f1e2 c6a5 b2b3 a5b7 f3e5 b7d6 e1g1 f8e8 h2h3 b6b5 e2d3 f5d3 e5d3 a6a5 c3e2 c7c6');
INSERT INTO evaluations VALUES(37731,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(37731,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(37731,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(37731,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(37731,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37731,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(37731,36,'cp 80','e2e4 d7d5 e4e5 g8h6 h2h3 e8g8 c4d5 c6d5 b1c3 f7f6 f1b5 f6e5 d4e5 e7e6 e1g1 c8d7 c1g5 d8e8 b5d3 b8c6 f1e1 h6f7 g5f4 d5d4 c3b5 c6e5 b5c7 e5f3 d1f3 e8d8 f3b7 a8b8 b7a7 e6e5 f4g3 b8b2 d3c4 g8h8 a1c1 d8f6 c7d5 f6d6');
INSERT INTO evaluations VALUES(43167,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(43167,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(43167,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43167,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(43167,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(43167,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(43167,36,'cp 0','g2g3 c7c5 g1f3 d7d5 c4d5 f6d5 c3d5 d8d5 f1g2 e8g8 e1g1 b8c6 c1e3 g7b2 a1b1 b2f6 d1a4 d5d7 e3c5 b7b6 c5b6 c6d4 a4d1 a7b6 f3d4 a8a2 d4c6 d7d6 c6b4 a2a5 b4c6');
INSERT INTO evaluations VALUES(43166,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43166,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43166,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43166,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43166,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43166,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43166,36,'cp 0','c7c5 f1g2 b8c6 g1f3 d7d5 c4d5 f6d5 c3d5 d8d5 e1g1 e8g8 c1e3 g7b2 a1b1 b2f6 d1a4 d5d7 e3c5 b7b6 c5b6 c6d4 a4d1 a7b6 f3d4 a8a2 d4c6 d7d6 c6b4 a2a5 b4c6');
INSERT INTO evaluations VALUES(45489,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(45489,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(45489,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(45489,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(45489,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(45489,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(45489,36,'cp 34','e7e6 e1g1 f8e7 b1c3 b8c6 f1e1 e8g8 a2a3 b7b6 c1f4 c8b7 c4a2 f8e8 f4g3 h7h6 d1e2 d8d7 a1d1 a8d8 d4d5 f6d5 c3d5 e6d5 d1d5 d7c8 e2c4 e7f8 e1e8 d8e8 d5d1 c8f5 h2h3 b6b5 c4g4 f5g4 h3g4 e8d8 d1c1 c6d4 f3d4 d8d4');
INSERT INTO evaluations VALUES(45555,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(45555,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(45555,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(45555,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(45555,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(45555,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(45555,36,'cp 66','d2d4 b8d7 a2a4 f8e7 e1g1 e8g8 f1e1 f8e8 a4a5 d8c7 c4f1 e5d4 f3d4 e7f8 c1f4 a8b8 f4h2 b7b6 a5b6 a7b6 d1d2 b6b5 f2f3 d7e5 c3d1 c8d7 d1e3 c6c5 d4f5 g7g6 f5g3 h6h5 a1d1 c5c4 g1h1 d7h3 g2h3 e5f3 d2g2');
INSERT INTO evaluations VALUES(52915,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(52915,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(52915,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(52915,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52915,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(52915,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(52915,36,'cp 6','g8f6 g1f3 c8f5 d1e2 e7e6 e4f6 d8f6 g2g3 f8b4 f3d2 h7h5 f1g2 h5h4 c2c3 b4e7 g3g4 f6g6 g4f5 g6g2 h1f1 e6e5 d2e4 b8d7 f2f4 g2e2 e1e2 h8h5 f4e5 d7e5 c1e3 f7f6 f1g1 e8f8 a1f1 e5f7 f1f2 f7d6 e4d6 e7d6 f2g2 a8e8 g2g7 d6h2 g1g4 h4h3 g7b7');
INSERT INTO evaluations VALUES(57259,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(57259,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(57259,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(57259,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(57259,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(57259,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(57259,36,'cp -33','c7c6 d2d4 e5d4 c3d4 d7d5 b1c3 g8f6 a4c2 g7g6 c1f4 c8f5 c2c1 f8g7 g1f3 e8g8 e2e3 f6h5 f4e5 g7e5 f3e5 b8d7 e5f3 a7a5 f1e2 h5f6 f3h4 f5e6 h4f3 f6e4 e2d3 e6f5 c1c2 e4c3 b2c3 f5d3 c2d3 f8e8 a2a4 b7b5 a4b5');
INSERT INTO evaluations VALUES(1213,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(1213,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(1213,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(1213,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(1213,34,'cp -110',NULL);
INSERT INTO evaluations VALUES(1213,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(1213,36,'cp -103','c8f5 d2d3 e8c8 d1e2 e4b4 e2e5 g8e7 a2a3 b4g4 h2h3 g4h5 e5f4 f5d7 f4g5 h5g5 f3g5 d7e8 b1c3 b7b6 c1d2 c6c5 g5f3 f7f6 f1e2 e7c6 a1e1 e8h5 e2d1 c6d4 e1e3 f6f5 h3h4 d4f3 g2f3');
INSERT INTO evaluations VALUES(1938,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(1938,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(1938,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(1938,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(1938,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(1938,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(1938,36,'cp 209','f3d4 e5d4 c3d5 f6d8 d1h5 f8d6 e4e5 g7g6 h5f3 d6e7 b5c4 d7d6 f3f4 f7f6 e5d6 e7d6 f4d4 e8f8 d2d3 f8g7 c1f4 g8h6 f4d6 d8d6 e1g1 h6f5 d4f4 d6f4 d5f4 c7c6 f1e1 h7h5 c2c3 g6g5 f4e6 c8e6 e1e6 a8e8 a1e1 e8e6 c4e6 f5e7 e6b3 e7g6 d3d4 h5h4 g2g3 a7a5 a2a4 f6f5');
INSERT INTO evaluations VALUES(1935,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(1935,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(1935,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(1935,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(1935,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(1935,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(1935,36,'cp 203','e5d4 c3d5 f6d8 d1h5 f8d6 e4e5 g7g6 h5f3 d6e7 b5c4 d7d6 f3f4 f7f6 e5d6 e7d6 f4d4 e8f8 d2d3 f8g7 c1f4 g8h6 f4d6 d8d6 e1g1 h6f5 d4f4 d6f4 d5f4 c8d7 f1e1 h8e8 g1f1 c7c6 c2c3 f5d6 c4b3 g6g5 f4e6 d7e6 b3e6 h7h5 d3d4 h5h4 a2a4 a8d8 a4a5 f6f5 e1e5');
INSERT INTO evaluations VALUES(1941,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(1941,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(1941,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(1941,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(1941,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(1941,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(1941,36,'cp 203','c3d5 f6d8 d1h5 f8d6 e4e5 d6e7 b5c4 g7g6 h5g4 c7c6 d5e7 g8e7 g4d4 e7f5 d4c3 e8g8 e1g1 d7d5 e5d6 d8d6 b2b3 f5d4 c1b2 c6c5 f1e1 c8f5 e1e3 a7a6 c4d3 f5d3 c3d3 a8d8 a1e1 d6c6 d3c4 b7b5 c4c3 b5b4 c3c4 d4b5 d2d3 b5d4 e3e4 f8e8');
INSERT INTO evaluations VALUES(1934,30,'cp 223',NULL);
INSERT INTO evaluations VALUES(1934,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(1934,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(1934,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(1934,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(1934,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(1934,36,'cp 203','f6d8 d1h5 f8d6 e4e5 d6e7 b5c4 g7g6 h5g4 c7c6 d5e7 g8e7 g4d4 e8g8 e1g1 e7f5 d4c3 d7d5 e5d6 d8d6 b2b3 f5d4 c1b2 f8d8 a2a4 c8f5 f1e1 a8c8 c4d3 f5d3 c3d3 d6c5 a1c1 d4f5 d3e2 b7b5 d2d3 b5a4 b2f6 d8d7 b3a4 c5d6 e2e5 d6e5 e1e5');
INSERT INTO evaluations VALUES(2614,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(2614,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2614,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2614,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2614,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2614,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2614,36,'cp 0','c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 h7h6 f1f2 c6b8 c3e2 b7b5 e2g3 b8d7 d1e2 c7c5 f3d2 g8h7 f2f1 d8e8 f1f2 e8e7 b2b3 e7e8 a1f1 a8c8 g1h2 e8e7 g3h1 h7h8 h1g3 h8g8 h2g1 c8a8 f1a1 g8h7 a3a4 e7e8 f2f1 d7b6');
INSERT INTO evaluations VALUES(3342,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(3342,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(3342,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(3342,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(3342,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(3342,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(3342,36,'cp 54','f1d3 h7h6 g5h4 f6h5 h4e7 c6e7 h2h3 h5f6 d1c2 c7c6 e1g1 g7g6 f3e5 c8f5 a1d1 d8c7 f2f3 f6d7 d3f5 d7e5 f5d3 e5d3 c2d3 f7f5 b2b4 a7a6 d1c1 g8g7 c3a4 a8e8 a4c5 e7c8 e3e4 f5e4 f3e4 d5e4 c5e4');
INSERT INTO evaluations VALUES(3343,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(3343,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(3343,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(3343,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(3343,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(3343,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(3343,36,'cp 53','h7h6 g5h4 f6h5 h4e7 c6e7 h2h3 h5f6 d1c2 f8e8 f3e5 c7c5 d3b5 c8f5 c2a4 c5d4 a4d4 e8f8 e1g1 a7a6 b5e2 f6e4 a1c1 a8c8 c3e4 d5e4 c1c8 e7c8 d4c3 f5e6 f1d1 d8b6 e5c4 b6c5 d1c1 c5e7');
INSERT INTO evaluations VALUES(11121,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(11121,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(11121,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(11121,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11121,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(11121,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(11121,36,'cp 5','e7e6 d1d2 d8c7 d4c6 b7c6 e1c1 f8e7 d2f4 c7f4 g5f4 d7d5 f2f3 f6d7 c3a4 g7g5 f4e3 c8b7 f1d3 h7h5 h1e1 h8g8 a4b6 a8d8 b6d7 d8d7 c2c3 e7d6 g2g3 e8e7 f3f4 g5f4 e3f4 d6f4 g3f4 d5e4 d3e4 d7d1 e1d1 h5h4');
INSERT INTO evaluations VALUES(11084,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(11084,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(11084,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(11084,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(11084,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(11084,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11084,36,'cp 16','d4c6 b7c6 f1d3 d7d5 e1g1 f8e7 d1f3 e8g8 f3g3 f6h5 g5e7 d8e7 g3e3 h5f6 e4e5 f6d7 c3a4 a6a5 a1c1 f8d8 c2c4 c8a6 b2b3 d5c4 d3c4 a6b5 c4e2 e7b4 f1d1 d7f8 a4c5 b5e2 e3e2 d8d1 e2d1 b4f4 c5d3 f4d4 c1c6 a8d8 c6d6 d8d6 e5d6');
INSERT INTO evaluations VALUES(11980,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(11980,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(11980,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(11980,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(11980,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11980,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(11980,36,'cp 23','c1f4 d7d5 b1d2 g7g6 f4g5 h7h6 g5h4 f8d6 e1g1 e8g8 f1e1 e6e5 e4d5 c6d5 c2c4 g6g5 h4g3 e5e4 d3f1 c8g4 f1e2 d6g3 h2g3 g4f5 c4d5 d8d5 e2c4 d5e5 d2f1 a8d8 d1e2 f5g4 e2c2 a7a5 a1c1 h6h5 f1e3 f8e8 c2c3 e5c3 c1c3');
INSERT INTO evaluations VALUES(14339,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14339,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14339,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14339,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14339,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(14339,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14339,36,'cp 0','f1b5 d8d6 d2d4 c8g4 b5e2 h7h5 a2a3 a7a6 b2b4 e7e5 d4e5 c6e5 c1b2 e8c8 d1d4 e5c6 d4d3 c6e5');
INSERT INTO evaluations VALUES(14172,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(14172,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(14172,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14172,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14172,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(14172,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14172,36,'cp -5','d8d6 d2d4 c8g4 h2h3 g4f3 d1f3 e8c8 c1d2 a7a6 b5c6 d6c6 a2a3 e7e6 e1c1 f8e7 g2g4 h8f8 c1b1 c8b8 f3e2 f6e8 c3a2 e8d6 a2b4 c6e8 b1a2 f7f6 f2f3 d6c4 e3e4 a6a5 b4d3 e6e5 d4e5 f6e5 e4d5');
INSERT INTO evaluations VALUES(16527,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(16527,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(16527,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(16527,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(16527,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(16527,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(16527,36,'cp -67','d7d5 c3d5 f6d5 e4d5 d8d5 g1f3 f7f5 f1e2 c8d7 a2a3 f5f4 e3d2 e8c8 e1g1 h7h5 b2b4 g7g5 f3g5 d5g8 g5f3 h5h4 g1h1 h4h3 g2g3 f8g7 b4b5 c6e7 c2c4 d7f5 d1c2 g8h7 f3g5 f5d3 e2d3 h7d3 c2d3 d8d3 d2b4 e7f5 g5e6 h8e8 e6g7 f5g7 a1e1 e5e4');
INSERT INTO evaluations VALUES(24090,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(24090,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(24090,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(24090,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(24090,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(24090,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(24090,36,'cp 40','h2g3 a7a6 f1e2 e8g8 c3c4 d8e7 e1g1 h7h6 a2a3 f8e8 d1c1 c6b8 b2b4 c7c6 c1b2 b8d7 f1c1 h6h5 a3a4 f6e4 b4b5 a6b5 a4b5 d5c4 e2c4 e4d2 f3d2 e6e5 b5c6 b7c6 a1a8 e8a8 b2b7 a8d8 b7c6 d7f6 d4e5 e7e5');
INSERT INTO evaluations VALUES(24214,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(24214,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(24214,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(24214,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(24214,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(24214,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(24214,36,'cp 44','d4c6 d7c6 b5c6 b7c6 b1c3 e8g8 f2f4 a7a5 d1f3 a5a4 g1h1 f6d7 c1e3 e7f6 e3g1 c6c5 e4e5 d6e5 a1d1 d8c8 c3d5 e5f4 f3f4 c7c6 d5f6 d7f6 g1c5 f8e8 f4f3 h7h6 c5d4 c8g4 f3g4 f6g4 e1e8 a8e8 h2h3 g4f6 d4f6 g7f6 d1d6');
INSERT INTO evaluations VALUES(26908,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(26908,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(26908,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(26908,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(26908,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(26908,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(26908,36,'cp 70','g4f3 d1f3 a7a6 b5c6 b7c6 e1g1 g7g6 d2d4 d8f6 f3f6 g8f6 d4e5 d6e5 f1d1 f6d7 c1e3 f7f5 f2f3 e8c8 d1d3 f8d6 g1f1 d7c5 e3c5 d6c5 d3d8 h8d8 a1e1 c5d6 c3a4 c8d7 f1e2 d7e7 e1d1 h7h5 d1d3 d8b8 c2c3 f5e4 f3e4 b8f8 b2b4 f8f4');
INSERT INTO evaluations VALUES(26920,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(26920,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(26920,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(26920,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(26920,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(26920,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(26920,36,'cp 65','d1f3 a7a6 b5c6 b7c6 e1g1 g7g6 d2d4 d8f6 f3f6 g8f6 d4e5 d6e5 f1d1 f6d7 g1f1 f8d6 c1e3 f7f5 f2f3 d7f6 d1d3 a6a5 b2b3 e8f7 a1e1 f5e4 f3e4 f7e6 c3a4 f6e4 e3h6 e4f6 d3c3 f6d5 c3c6 e6d7 c6c4');
INSERT INTO evaluations VALUES(26919,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(26919,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(26919,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(26919,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(26919,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(26919,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(26919,36,'cp 58','a7a6 b5a4 g7g6 d2d3 f8g7 c3e2 g8f6 c1d2 d8e7 a2a3 f6d7 a4b3 h7h5 e1c1 e8c8 g2g3 d8f8 b3d5 c6d8 f3g2 h5h4 g3g4 d8e6 c1b1 c8b8 h1f1 g6g5 c2c3 d7c5 d2e3 e6f4 e3f4 g5f4 d3d4 c5d7 f1e1 g7h6 d4e5 d7e5');
INSERT INTO evaluations VALUES(29410,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(29410,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(29410,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(29410,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(29410,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(29410,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(29410,36,'cp 155','h2h3 g4c8 f1c4 f8g7 e1g1 e8g8 e4e5 f6e8 c1g5 c7c6 f1e1 d6d5 c4d3 f7f6 g5f4 b8d7 e5f6 e7f6 c3a4 e8c7 c2c4 d7b6 a4b6 a7b6 d1b3 d5c4 d3c4 c7d5 a2a4 g6g5 f4g3 c8f5 h3h4 g5g4 f3h2 g8h8 c4d5 d8d5 b3d5 c6d5');
INSERT INTO evaluations VALUES(31195,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(31195,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31195,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(31195,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(31195,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(31195,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(31195,36,'cp 9','e7e6 c2c3 b8d7 g1e2 g8f6 e2g3 f6e4 g3e4 f8e7 c1f4 e8g8 f1d3 d8a5 e1g1 a5d5 f1e1 f5g6 f3e2 a8d8 a1d1 c6c5 d4c5 d7c5 e4c5 d5c5 d3g6 h7g6 g2g3 c5a5 a2a3 d8d1 e1d1 f8d8 d1d8 e7d8 e2f3 a5b6 f3e2');
INSERT INTO evaluations VALUES(31570,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31570,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(31570,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31570,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(31570,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31570,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31570,36,'cp 28','c2c3 b8d7 g1e2 g8f6 e2g3 f6e4 g3e4 f8e7 c1f4 f5e4 f3e4 d7f6 e4f3 f6d5 f4c1 h7h5 f1c4 g7g6 e1g1 e8g8 g2g3 e7g5 c1g5 d8g5 h2h4 g5g4 f3g4 h5g4 c4e2 d5f6 f2f3 g4f3 e2f3 g8g7 a2a4 a7a5 g1f2 f8d8 g3g4');
INSERT INTO evaluations VALUES(34529,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(34529,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(34529,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(34529,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(34529,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(34529,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(34529,36,'cp 174','d2d4 d6d5 c2c4 d5d8 g1f3 g8f6 b1c3 c7c6 c1e3 c8f5 f3h4 f5c8 h2h3 e7e6 h4f3 b8d7 d1c2 f8b4 f1e2 e8g8 e1g1 d8c7 a2a3 b4e7 c4c5 b7b6 b2b4 a7a5 f1d1 c8a6 e2a6 a8a6 b4b5 a6a8 b5c6 c7c6 f3e5');
INSERT INTO evaluations VALUES(34730,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(34730,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(34730,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(34730,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(34730,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(34730,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(34730,36,'cp 178','b1c3 g8f6 f1e2 b8c6 c1e3 c8d7 e1g1 e7d8 f2f4 f8e7 d1e1 e8g8 a1d1 f8e8 e2f3 f6g4 d4c6 b7c6 e3d4 e7f6 d4f6 d8f6 h2h3 g4h6 d1d3 f6e7 e1d2 a8d8 f1d1 f7f5 e4e5 h6f7 c3e2 g7g5 e5d6 c7d6 f4g5 f7g5 e2f4 d6d5 c2c4');
INSERT INTO evaluations VALUES(34729,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(34729,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(34729,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(34729,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(34729,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(34729,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(34729,36,'cp 167','g8f6 f1e2 b8c6 c1e3 c8d7 e1g1 e7d8 f2f4 f8e7 d1e1 e8g8 a1d1 f8e8 e2f3 f6g4 d4c6 b7c6 e3d4 e7f6 h2h3 f6d4 d1d4 g4h6 d4d3 d8e7 e1d2 d7c8 f1e1 h6f5 c3e2 f5h4 e2g3 h4f3 d3f3 f7f6 g1h2 a7a5 b2b3 a5a4 b3b4 g8h8 f3d3');
INSERT INTO evaluations VALUES(35259,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(35259,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(35259,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(35259,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(35259,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(35259,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(35259,36,'cp 62','e7h4 b1c3 f8b4 f1d3 d7d6 c1d2 b4c3 d2c3 c8g4 g1f3 h4h5 h2h3 g4f3 d1f3 h5f3 g2f3 b8d7 f3f4 f7f6 h1g1 g7g6 c3d2 g8e7 e1e2 e8c8 f4e5 d7e5 f2f3 f6f5 b2b4 d8f8 a1f1 c8b8 c2c4 f8f7 c4c5 e5d3 e2d3');
INSERT INTO evaluations VALUES(37261,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(37261,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(37261,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(37261,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(37261,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(37261,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(37261,36,'cp -19','f1e2 f8e8 e1g1 h7h6 f1e1 b8c6 c1d2 a7a5 h2h3 d7d5 e4d5 f6d5 c3b5 c8f5 a2a3 b4c5 e2f1 f5g6 c2c3 g6h5 d3d4 e5d4 b5d4 c6d4 c3d4 e8e1 d1e1 c5b6 g2g4 h5g6 f3e5 d8f6 e5g6 f6g6 f1c4 d5f6 e1e5');
INSERT INTO evaluations VALUES(38525,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(38525,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(38525,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(38525,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(38525,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(38525,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(38525,36,'cp 79','g1f3 e5d4 f3d4 g8e7 c1e3 b8c6 d1d2 e8g8 f1e2 f7f5 e1g1 c6d4 e3d4 g7d4 d2d4 e7c6 d4d5 g8h8 e4f5 c8f5 d5d2 d8f6 a1d1 g6g5 b2b3 c6e5 f2f4 g5f4 f1f4 f6g7 d1f1 f5g6 c3d5 f8f4 d5f4 a8e8 h2h3 e8f8 d2e3 a7a5 e2f3 c7c6 f3e2 f8e8 f4g6 h7g6 e3g3 h8h7');
INSERT INTO evaluations VALUES(39354,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(39354,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(39354,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(39354,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(39354,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(39354,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(39354,36,'cp 155','c2c3 b4a5 g1f3 g8f6 d2d4 d7d5 e4d5 e5d4 d1e2 d8e7 e2e7 e8e7 b2b4 a5b6 e1g1 h8d8 a2a4 a7a6 f1e1 e7f8 a4a5 b6a7 c1a3 f8g8 c3d4 c7c6 d5d6 d8d6 b1d2 d6d8 d2e4 f6e4 e1e4 c8e6 c4e6 f7e6 e4e6 b8d7');
INSERT INTO evaluations VALUES(42374,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(42374,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(42374,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(42374,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(42374,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(42374,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(42374,36,'cp 218','e4e5 f6d5 f4g5 d5e7 b1c3 d7d5 g5e7 e8e7 c3a4 c8e6 f1d3 e7d7 e1g1 d8e7 c2c3 c7c6 d1e1 g7g5 b2b4 d7c7 a4b6 a7b6 a2a4 b8d7 a4a5 h7h5 e1d2 h8g8 f3e1 h5h4 e1c2 g8f8 c2e3 b6a5 b4a5 h4h3 g2g3');
INSERT INTO evaluations VALUES(48438,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(48438,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(48438,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48438,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(48438,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48438,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(48438,36,'cp 0','f6d7 f1c4 c7c5 g1f3 b8c6 e1g1 c5d4 f3d4 d7e5 d4c6 d8d1 f1d1 b7c6 c4e2 f8e7 c1e3 e8g8 c3a4 e5d7 e2f3 d7e5');
INSERT INTO evaluations VALUES(50080,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(50080,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(50080,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(50080,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(50080,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(50080,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(50080,36,'cp 51','g1f3 g7g6 d2d4 d7d6 f1g2 f8g7 e1g1 c7c6 b1c3 e8g8 a1b1 a7a5 b2b3 b8a6 c1b2 d8c7 d4d5 e7e5 d5e6 c8e6 f3g5 e6c8 g5h3 a6c5 h3f4 c7e7 h2h4 c8e6 d1d2 e6f7 f1d1 a8d8 e2e3 f8e8 c3e2 c5e4 d2a5');
INSERT INTO evaluations VALUES(50508,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(50508,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(50508,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(50508,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(50508,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(50508,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(50508,36,'cp 80','b1c3 b8c6 d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 f1e2 g8e7 e1g1 d7d5 c4d5 c6d5 e4d5 e6d5 e2f3 c8e6 b2b4 c5d6 c1g5 e8g8 c3d5 e6d5 f3d5 d6h2 g1h1 d8d5 g5e7 d5d1 a1d1 f8e8 d1d7 h2e5 f1d1 h7h6 d1d5 e5b2 g2g3 b2f6 e7f6 g7f6 a2a3 a8c8');
INSERT INTO evaluations VALUES(50506,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(50506,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(50506,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(50506,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(50506,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(50506,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(50506,36,'cp 63','d8c7 a2a3 b7b6 f1e2 g8f6 d2d4 c5d4 f3d4 c8b7 f2f3 f8e7 c1e3 e8g8 a1c1 b8c6 d1d2 c6d4 e3d4 f6h5 d4e3 f7f5 e4f5 e6f5 c1d1 a8e8 d2d7 c7e5 e3b6 h5f4 e1g1 f4e2 c3e2 e5b2 d7b7 e7c5 g1h1 b2b6 b7b6 c5b6 e2c3 b6a5 c3d5');
INSERT INTO evaluations VALUES(1107,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(1107,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(1107,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(1107,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(1107,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(1107,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(1107,36,'cp -85','b7b5 c4b3 a7a5 c2c3 a5a4 b3d1 g7g6 b1a3 c8a6 a3c2 f8g7 g1e2 e8g8 e1g1 c6d8 c1g5 d8e6 g5h4 h7h6 f1e1 a6b7 h2h3 e6g5 h4g5 h6g5 e2g3 d7d6 f3e2 b7c8 e2d2 g7h6 c2e3 c7c6 d1g4 f6g4 e3g4 c8g4 h3g4');
INSERT INTO evaluations VALUES(2055,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(2055,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(2055,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(2055,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(2055,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(2055,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(2055,36,'cp 49','g1e2 g8f6 d2d3 c6a5 b1c3 d7d6 c4b3 a5b3 a2b3 a7a6 h2h3 c8e6 d3d4 g7g6 d4e5 d6e5 f3g3 f6d7 c1e3 e8c8 e1c1 h7h5 f2f4 h5h4 g3f2 f8h6 c1b1 h8e8 d1d2 e5f4 e2f4 h6g7 f4e6 e7e6 e3g5 g7c3');
INSERT INTO evaluations VALUES(6582,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(6582,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(6582,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(6582,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(6582,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(6582,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(6582,36,'cp 86','f3e5 c5f2 e1f2 c6e5 d2d4 e5g6 d5c3 d7d6 f1e2 d8h4 f2g1 g8e7 c1e3 a7a6 g2g3 h4h3 e2f1 h3g4 d1g4 c8g4 g1g2 e7c6 c3d5 e8c8 a1e1 d8e8 f1d3 g6e7 h2h3 g4d7 c2c4 e7d5 e4d5 c6b4 d3b1 c7c6 d5c6 b4c6 e3f2 c6b4 e1e8 h8e8 b1h7');
INSERT INTO evaluations VALUES(6610,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(6610,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(6610,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6610,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(6610,36,'cp -8','a7a5 g2g3 g8f6 f1g2 e8g8 e1g1 h7h6 c2c4 d7d6 h2h3 f6d7 d2e3 a5a4 b1a3 b4c5 a3b5 c5e3 f2e3 d7b6 f3h4 c6b8 h4f5 c8f5 e4f5 c7c6 b5c3 b8d7 g1h2 d6d5 c4d5 b6d5 g2d5 c6d5 c3d5 d7f6 d5c3 d8b6 c3a4');
INSERT INTO evaluations VALUES(7039,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(7039,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(7039,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(7039,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(7039,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(7039,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(7039,36,'cp 110','d2d4 d7d6 c1e3 g8f6 d1d2 a7a6 d4d5 c6a7 a2a4 c7c6 h2h4 c6d5 e4d5 d8a5 f3d4 c8d7 d4b3 a5b4 f1e2 h7h5 a4a5 e8g8 e3d4 b7b5 e1g1 d7f5 f1e1 f8e8 e2d3 f5d3 d2d3 b4c4 d3c4 b5c4 b3d2');
INSERT INTO evaluations VALUES(7365,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7365,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7365,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7365,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7365,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7365,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7365,36,'cp 0','b7b5 a4b3 d7d6 a2a4 b5b4 c3d5 c6a5 d5e7 d8e7 d2d4 e8g8 f1e1 a5b3 c2b3 c8g4 d4e5 d6e5 h2h3 f8d8 d1e2 g4f3 e2f3 h7h6 f3e2 e7e6 e2c2 d8d6 c1d2 d6c6 c2d3 a6a5 d2e3 c6d6 d3c2 d6c6');
INSERT INTO evaluations VALUES(7497,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7497,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7497,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7497,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7497,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7497,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7497,36,'cp 0','a4b3 d7d6 a2a4 b5b4 c3d5 c6a5 d5e7 d8e7 d2d4 e8g8 f1e1 a5b3 c2b3 c8g4 d4e5 d6e5 h2h3 f8d8 d1e2 g4f3 e2f3 h7h6 f3e2 e7e6 e2c2 d8d6 c1d2 d6c6 c2d3 a6a5 d2e3 c6d6 d3c2 d6c6');
INSERT INTO evaluations VALUES(9932,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(9932,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(9932,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(9932,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(9932,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(9932,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(9932,36,'cp 161','d4e5 f6g4 c4f7 e8f7 f3g5 f7e8 d1g4 c6e5 g4g3 d8f6 f2f4 e5f7 g5f7 e8f7 c1e3 b6e3 g3e3 d7d5 e4e5 f6b6 e3b6 a7b6 e1g1 c8f5 b1d2 f7e7 h2h3 h7h5 a2a4 h8f8 f1f2 f5d7 b2b3 g7g6 h3h4 d7g4 g2g3 c7c5 a1e1 f8d8 d2f3 d5d4 c3d4 g4f3 f2f3 d8d4 g1f2');
INSERT INTO evaluations VALUES(9913,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(9913,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(9913,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(9913,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(9913,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(9913,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(9913,36,'cp 162','f6g4 c4f7 e8f7 f3g5 f7e8 d1g4 c6e5 g4g3 d8f6 f2f4 e5f7 g5f7 e8f7 c1e3 b6e3 g3e3 d7d5 b1d2 f6b6 e3b6 a7b6 e4e5 c8f5 e1g1 a8a4 h2h3 h7h5 d2f3 f7e7 f3d4 g7g6 b2b3 a4a3 f1f2 c7c6 f2d2 h8a8 g1f2 h5h4 d4f3 a3b3 a2b3 a8a1 f3h4 a1c1 h4f5 g6f5 g2g4 f5g4 h3g4 c1c3 f4f5 c3b3');
INSERT INTO evaluations VALUES(12888,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(12888,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(12888,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(12888,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(12888,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(12888,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(12888,36,'cp 3','e4e5 f6d5 f1c4 d5b6 c4b3 d7d5 e5d6 d8d6 f3d4 c6d4 c3d4 c8e6 e1g1 e6b3 a2b3 e7e6 d1f3 d6d7 f1d1 f8d6 b1c3 e8g8 g2g3 f7f5 c1e3 g8h8 d1d3 h8g8 e3f4 d6f4 f3f4 b6d5 f4e5 a7a6 c3d5 d7d5 e5d5 e6d5');
INSERT INTO evaluations VALUES(12861,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(12861,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12861,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(12861,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(12861,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12861,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(12861,36,'cp 7','f6d5 f1c4 d5b6 c4b3 d7d6 e5d6 d8d6 e1g1 c8e6 f3d4 c6d4 c3d4 e6b3 a2b3 e7e6 d1f3 d6d7 f1d1 f8e7 b1c3 e8g8 d4d5 b6d5 c3d5 e6d5 d1d5 d7e6 c1e3 b7b6 g2g3 e7f6 e3d4 f6d4 d5d4 e6e7 d4e4 e7b7 f3e3 a8d8 e4e7 d8d7 e7e8 d7d8 e8d8');
INSERT INTO evaluations VALUES(13298,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13298,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(13298,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(13298,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13298,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(13298,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13298,36,'cp 0','e4e5 d6e5 f3e5 c6e5 b2e5 a7a6 d1f3 d8d5 f3e2 c8f5 b1c3 d5c6 h1g1 h7h5 h2h3 f6d7 e5f4 f5c2 a1c1 c2g6 g2g3 a8c8 f1g2 c6b6 c3d5 b6d8 d5c3');
INSERT INTO evaluations VALUES(16192,30,'cp 603',NULL);
INSERT INTO evaluations VALUES(16192,31,'cp 574',NULL);
INSERT INTO evaluations VALUES(16192,32,'cp 617',NULL);
INSERT INTO evaluations VALUES(16192,33,'cp 617',NULL);
INSERT INTO evaluations VALUES(16192,34,'cp 617',NULL);
INSERT INTO evaluations VALUES(16192,35,'cp 618',NULL);
INSERT INTO evaluations VALUES(16192,36,'cp 609','d1d5 d8e7 d5e4 d7d6 e1g1 c8d7 b1c3 e8c8 e5d6 e7d6 c4f7 d6b4 e4g6 b4g4 g6g4 d7g4 f7d5 f8d6 h2h3 g4f3 d5f3 c6d4 f3d1 h8e8 c1e3 d6e5 c3e4 c8b8 c2c3 d4f5 e3d2 f5d6 e4d6 e5d6 d1g4 d6h2 g1h2 d8d2 a1e1 e8d8 e1e2 d2e2 g4e2 d8d2 f1d1 d2d1 e2d1 b8c8 d1g4 c8d8 a2a3 d8e7 h2g3 e7f6');
INSERT INTO evaluations VALUES(16189,30,'cp 591',NULL);
INSERT INTO evaluations VALUES(16189,31,'cp 599',NULL);
INSERT INTO evaluations VALUES(16189,32,'cp 599',NULL);
INSERT INTO evaluations VALUES(16189,33,'cp 599',NULL);
INSERT INTO evaluations VALUES(16189,34,'cp 612',NULL);
INSERT INTO evaluations VALUES(16189,35,'cp 599',NULL);
INSERT INTO evaluations VALUES(16189,36,'cp 613','d8e7 d5e4 d7d6 e1g1 c8d7 b1c3 e8c8 e5d6 e7d6 c4f7 d6b4 f7c4 c6a5 c4e2 a5c6 c1e3 f8c5 e4b4 c5b4 c3d5 h8e8 a1d1 b4c5 f1e1 c5e3 d5e3 c6b4 e2c4 b4c6 c4f7 e8f8 f7b3 a7a6 a2a3 c8b8 b3d5 d8e8 b2b4 b8c8 c2c4 d7e6 d5e6 e8e6');
INSERT INTO evaluations VALUES(16736,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(16736,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(16736,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(16736,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(16736,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(16736,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(16736,36,'cp 100','f8e7 d2d4 d7d6 c1h6 g7h6 b1a3 e8g8 a3c2 e5d4 c2d4 c6e5 c4b3 e7g5 f3g5 d8g5 d1d2 a7a5 e1c1 g5d2 d1d2 a5a4 b3c2 e5c4 d2e2 c8d7 f2f4 b7b5 h1e1 f8e8 c2d3 f7f6 b2b3 a4b3 a2b3 c7c5 d4b5 d7b5 b3c4 b5d7 c1d2 d7g4 e2f2 g4e6');
INSERT INTO evaluations VALUES(18585,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(18585,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(18585,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(18585,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(18585,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(18585,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(18585,36,'cp 101','e2e4 g7g6 f1e2 g8f6 d1c2 f8g7 g1f3 e8g8 e1g1 e7e5 d4d5 a7a5 c1e3 f6g4 e3d2 g4f6 h2h3 d7c5 d2e3 d8c7 a1c1 f6e8 f1d1 c7e7 a2a3 f7f5 d5c6 b7c6 b2b4 c5e4 c3e4 f5e4 f3g5 a5b4 a3b4 g7h6 g5e4 h6e3 f2e3 c8f5');
INSERT INTO evaluations VALUES(19651,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(19651,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(19651,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(19651,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(19651,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(19651,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(19651,36,'cp 102','d2d4 f8g7 e1g1 g8f6 d4c5 d6c5 d1e2 e8g8 e4e5 f6d5 f1d1 e7e6 h2h4 c8d7 d1e1 h7h6 c4d3 d8c7 a2a4 a7a6 d3c2 b7b5 h4h5 b5b4 b1d2 b4c3 b2c3 d5c3 e2c4 c6e5 f3e5 g7e5 d2f3 e5g7 h5g6 d7c6 c4g4 c6f3 g2f3 g8h8 c1f4');
INSERT INTO evaluations VALUES(19660,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(19660,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(19660,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(19660,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(19660,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(19660,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(19660,36,'cp 84','g8f6 e1g1 c5d4 c3d4 f8g7 d4d5 c6a5 c4e2 e8g8 b1d2 b7b5 e2b5 d8b6 b5e2 e7e6 b2b4 b6b4 a1b1 b4c3 c1b2 c3c7 b1c1 c7d8 d5e6 c8e6 f3g5 f8e8 g5e6 e8e6 b2d4 e6e8 h2h3 f6e4 d4g7 g8g7 d2e4 e8e4 e2f3');
INSERT INTO evaluations VALUES(21034,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(21034,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21034,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(21034,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21034,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(21034,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(21034,36,'cp 45','g1f3 c8f5 e2e3 c6b4 f1b5 c7c6 b5a4 d8a5 a4b3 g8f6 a2a3 e7e6 e1g1 f8d6 f3e5 b4a6 f2f4 h7h6 c2c4 a5c7 b3c2 f5c2 d1c2 a6b8 b2b4 a7a5 b4a5 a8a5 a3a4 b8d7 c1b2 e8g8 c4c5 d6e7 b2c3 a5a8 a4a5 d7e5');
INSERT INTO evaluations VALUES(21050,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(21050,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(21050,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(21050,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(21050,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(21050,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(21050,36,'cp -15','d5e4 d3e4 d8d1 e1d1 e7e6 b1d2 f8c5 d1e1 g8f6 f1d3 b7b6 a2a3 f6d7 b2b4 c5e7 d2c4 f7f6 e1e2 a7a5 b4a5 b6a5 c1e3 d7c5 h1b1 a5a4 c4d2 e8f7 d3b5 c8d7 f3d4 c6d4 e3d4 h8b8 b5d7 c5d7 d2c4 g7g5 b1b8 a8b8 d4a7 b8b7 a7e3 f7e8 f2f3 h7h5');
INSERT INTO evaluations VALUES(23710,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(23710,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(23710,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(23710,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(23710,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(23710,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(23710,36,'cp -17','h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 c3e2 d6d5 e2d4 c5d4 c4b3 d5e4 d3e4 f6e4 c2c3 d4c5 d1g4 e4f6 g4g3 d8d6 f1d1 f6d5 c1e3 c5e3 f2e3 d6c5 b3d5 c6d5 g3e5 a8e8 e5d5 c5e3 g1h1 e8e7 d1d2 e3f4 d2d4 f4c7 d5d6 c7d6 d4d6 e7e2 b2b3 f8e8 d6d7 e2c2 d7b7 e8e2 a1g1 c2a2');
INSERT INTO evaluations VALUES(23709,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(23709,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(23709,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(23709,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(23709,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(23709,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(23709,36,'cp -15','g4f3 d1f3 c6d4 f3d1 c7c6 c3e2 d6d5 e2d4 c5d4 c4b3 d5e4 d3e4 f6e4 c2c3 d4c5 d1g4 e4f6 g4g3 d8d6 f1d1 f6d5 c1e3 c5e3 f2e3 d6c5 b3d5 c6d5 g3e5 a8e8 e5d5 c5e3 g1h1 e8e7 d1d2 e3f4 d2d4 f4c7 d5d6 c7d6 d4d6 e7e2 b2b3 b7b6 d6d7 a7a5 d7d6 f8b8 h1h2 e2c2 c3c4');
INSERT INTO evaluations VALUES(31574,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(31574,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(31574,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(31574,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31574,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31574,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(31574,36,'cp 19','g8f6 e1g1 f8e7 c2c4 e8g8 f1e1 b8d7 c1d2 a7a5 a1d1 d8c7 g3f5 e7b4 f5e3 h7h6 h2h3 f8d8 d3c2 b4d2 c2d2 c7f4 d2c1 b7b6 b2b3 b6b5 e3g4 f4c1 d1c1 f6h5 g2g3 h5f6 g4f6 d7f6 g1g2');
INSERT INTO evaluations VALUES(32336,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(32336,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(32336,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(32336,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(32336,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(32336,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(32336,36,'cp -99','d1a4 b7b5 a4b5 c7c6 b5a4 g8f6 c4d5 e6d5 b1c3 b8d7 a4c6 a8c8 c6a4 f8d6 c3d5 e8g8 d5f6 d7f6 g1f3 f8e8 d2d3 d8e7 e1g1 e7e2 c1f4 d6f4 a4f4 e2b2 f1b1 b2c3 f4d2 c3a3 d3d4 g4f5 b1b3 a3d6 b3e3 f5e4 a1e1 d6d5 f3h4 d5g5 h4f3 g5f5 f3e5 e4g2 g1g2 h7h6');
INSERT INTO evaluations VALUES(32407,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(32407,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(32407,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(32407,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(32407,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(32407,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(32407,36,'cp -41','d2d4 c7c5 c1f4 b8c6 b1d2 c5d4 f1d3 g8e7 e1g1 e7g6 f4g3 f8e7 d1e1 e8g8 h2h3 g4h5 g3h2 g6h4 e5e6 g8h8 f3e5 c6e5 e1e5 h5g6 e6f7 e7f6 e5e6 f8f7 h2e5 f7e7 e5f6 e7e6 f6d8 a8d8 d2b3 g6d3 c2d3 e6e3 f1f2 h7h6 a1f1 d8e8 b3d4 e3d3 f2f8 e8f8');
INSERT INTO evaluations VALUES(32406,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(32406,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(32406,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(32406,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(32406,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(32406,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(32406,36,'cp -32','c7c5 f1b5 b8c6 b1c3 g8e7 c1f4 g4f3 d1f3 c5d4 b5c6 e7c6 c3d5 c6b4 c2c4 d4c3 a1d1 b4d5 d1d5 d8b6 f4e3 c3b2 e1g1 b6e6 d5b5 f8e7 b5b2 e8g8 b2b7 e6e5 b7a7 a8a7 e3a7 e5a5 a7d4 a5a2 f3g4 f7f6 d4f2 f8b8 h2h3 e7f8 f1d1 h7h6 g4d7 a2c2 d1d2 c2e4 d7d3 e4d3 d2d3 g8f7 g2g4 f7g6');
INSERT INTO evaluations VALUES(37226,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(37226,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37226,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(37226,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(37226,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37226,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(37226,36,'cp 0','d2d3 c7c6 a2a4 d7d6 c1g5 c8g4 f3g3 b8d7 g1f3 g4e6 c4e6 f7e6 e1g1 d8e8 g3h3 a7a5 c3e2 h7h6 g5d2 f8f7 e2g3 d7f8 g3e2');
INSERT INTO evaluations VALUES(37229,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(37229,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(37229,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(37229,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(37229,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37229,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(37229,36,'cp -13','c7c6 a2a4 d7d6 c1g5 c8e6 g5f6 d8f6 f3f6 g7f6 g1e2 f8d8 f2f4 g8f8 h1f1 f8e7 f4e5 d6e5 c4e6 f7e6 e2c1 b8d7 g2g3 h7h5 c1b3 d8g8 b3c5 d7c5 b2b4 c5a6 a1b1 h5h4 c3e2 g8g4 c2c3 a8g8 b1a1 g8h8 e1f2 h4g3 h2g3 h8h2 f2e3');
INSERT INTO evaluations VALUES(38078,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38078,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(38078,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(38078,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(38078,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(38078,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(38078,36,'cp 11','f1d3 h7h6 e1g1 g8e7 b1c3 e8g8 f1e1 d5c4 d3c4 b8d7 h2h3 d7b6 c4b3 b6d5 c1d2 c8e6 c3e4 d8c7 a2a4 a8d8 a4a5 e7g6 e4d6 c7d6 f3e5 g6e5 d4e5 d6e7 d1c1 d5b4 d2b4 e7b4 b3e6 f7e6 a5a6 b7a6 c1c6 b4b2 c6e6 g8h8 f2f3 f8e8 e6a6 b2d4 g1h1 e8e5');
INSERT INTO evaluations VALUES(46682,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46682,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46682,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46682,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46682,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46682,36,'cp 0','c7c5 g1f3 f8g7 f1e2 b8c6 e1g1 f6d5 f4g3 c5d4 e3d4 d8b6 c2c4 d5f6 b1c3 b6b2 c3b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(49160,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(49160,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(49160,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(49160,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(49160,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(49160,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(49160,36,'cp -29','c7c5 f2f4 b8c6 d4c5 d7c5 g1f3 c5d3 c2d3 f8e7 e1g1 e8g8 d3d4 b7b6 f1f2 f7f6 c1e3 c8d7 h2h3 d7e8 c3e2 d8d7 e5f6 e7f6 f3e5 f6e5 d4e5 e8g6 d1d2 a8c8 a1c1 h7h6 e3d4 g6e4 c1c3 c6d4 e2d4 c8c3 d2c3 f8c8 c3g3 c8c4');
INSERT INTO evaluations VALUES(49539,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(49539,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(49539,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(49539,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(49539,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(49539,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(49539,36,'cp -45','f8c5 d2d3 c7c6 b1c3 e8g8 g1f3 f8e8 c4b3 a7a5 e1g1 b7b5 a2a3 h7h6 c1e3 c5f8 f1d1 d7d6 h2h3 b8a6 d3d4 e5d4 f3d4 c8d7 e2f3 a5a4 b3a2 b5b4 a3b4 a6b4 a2b1 d6d5 d4f5 f6e4 c3e4 d5e4 f3g4 d7f5 d1d8 f5g4 d8e8 a8e8 h3g4');
INSERT INTO evaluations VALUES(49688,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(49688,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(49688,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(49688,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(49688,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(49688,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(49688,36,'cp -70','e5e4 f3g1 d7d5 d2d3 b8d7 g1e2 f8c5 e2g3 h7h5 h2h4 e4d3 c2d3 c7c6 d3d4 c5d6 g3f5 d6f8 f5g3 g7g6 f1d3 f8g7 d1c2 e8g8 c3e2 a7a5 f2f3 c6c5 d3b5 d8e7 e1g1 b7b6 c1d2 c8b7 a1c1 f8c8 c2b3 d7f8 e2f4');
INSERT INTO evaluations VALUES(52092,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(52092,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(52092,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(52092,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(52092,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(52092,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(52092,36,'cp 55','b8c6 c1e3 g8f6 b1c3 h7h6 a2a3 b7b6 c5b6 a7b6 c3b5 d7d5 e3d2 c8e6 e2e3 f8e7 d2c3 d8b8 f1e2 f6e4 b2b3 e8g8 c3b2 f8d8 e1g1 e6f5 e2d3 f5e6 d1e1 e4c5 d3e2 e6f5 e1d2 f5g6 f1d1 c5e4 d2e1 e7f6 c2c4 d5c4 d1d8 f6d8 b3c4');
INSERT INTO evaluations VALUES(53004,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(53004,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(53004,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(53004,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(53004,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(53004,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(53004,36,'cp -33','g8f6 f1b5 b8d7 d2d3 a7a6 b5d7 c8d7 g1f3 e7e6 e1g1 f8e7 c2c4 e8g8 b4c5 d6c5 e4e5 f6e8 b1d2 d7c6 d2e4 b7b5 d1e2 d8d7 f1d1 h7h6 c1f4 d7b7 f4e3 c6e4 d3e4 b7e4 d1d7 e7g5 f3g5 h6g5 c4b5 a6b5 e2b5');
INSERT INTO evaluations VALUES(55261,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(55261,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(55261,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(55261,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55261,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(55261,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55261,36,'cp -10','c7c5 g1f3 b8c6 c1f4 e7e6 f1e2 f8d6 f4d6 d8d6 b1a3 g8e7 d1a4 e8g8 c3d4 c6d4 f3d4 c5d4 a3c4 d6c7 g2g3 e6e5 e1g1 c8e6 f1c1 f8c8 a4b4 e7g6 a2a4 g6f8 a4a5 h7h6 h2h4 a8b8 b4d2 f8d7 f2f4 e6c4 c1c4 c7d6 c4c8 b8c8 e2g4');
INSERT INTO evaluations VALUES(55875,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(55875,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(55875,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(55875,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(55875,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(55875,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(55875,36,'cp -50','d5e4 b1d2 g8f6 d1c2 c7c5 d2e4 c5d4 e4f6 g7f6 c3d4 c8d7 f1e2 b8c6 g1f3 a8c8 c2d1 f8b4 c1d2 d8a5 a2a3 b4d2 d1d2 a5d2 e1d2 c6e7 b2b4 d7c6 h1c1 c8d8 d2e1 a7a6 a3a4 e8g8 b4b5 a6b5 a4b5 c6d5 c1c5 g8g7 c5c7 e7f5 a1c1 f5d6 c1c5 b7b6');
INSERT INTO evaluations VALUES(56050,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(56050,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(56050,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(56050,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(56050,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(56050,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(56050,36,'cp -47','b1d2 g8f6 d1c2 c7c5 d2e4 c5d4 e4f6 g7f6 c3d4 c8d7 g1f3 b8c6 f1e2 a8c8 c2d1 c6e7 e1g1 d7c6 a2a4 f8g7 b2b3 e8g8 c1a3 f8e8 a1c1 d8d7 f1e1 e8d8 a4a5 b7b6 a5a6 c6a8 d1d3 a8d5 c1c8 d8c8 e1c1 c8d8 d3d1 e7g6 e2f1 f6f5 f3e5');
INSERT INTO evaluations VALUES(55885,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(55885,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(55885,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(55885,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(55885,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(55885,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(55885,36,'cp -122','d5e4 b1c3 g8f6 d1e2 c8d7 f2f3 b8c6 c3e4 c6d4 e2f2 d7c6 d2d3 a7a5 c1e3 d4f5 a2a3 f5e3 f2e3 f8e7 g1e2 e8g8 e1g1 a5a4 b3a2 c6d5 a2d5 e6d5 e4f6 e7f6 c2c3 d8d7 d3d4 a8e8 e3f2 d7b5 e2g3 g7g6 f3f4 c7c5 f4f5 f6g5');
INSERT INTO evaluations VALUES(56076,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(56076,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(56076,32,'cp -106',NULL);
INSERT INTO evaluations VALUES(56076,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(56076,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(56076,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(56076,36,'cp -113','b1c3 g8f6 d1e2 c8d7 f2f3 b8c6 c3e4 c6d4 e2f2 d7c6 d2d3 a7a5 c1e3 d4b3 a2b3 f6d5 g1e2 b7b6 d3d4 d8d7 c2c4 d5e3 f2e3 f8e7 e2c3 e8g8 e1c1 f8e8 d4d5 e6d5 c3d5 c6d5 d1d5 d7e6 h1d1 a5a4 c1c2 e6g6 g2g3 h7h6 e3d3 a4b3 d3b3');
INSERT INTO evaluations VALUES(56549,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(56549,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(56549,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56549,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56549,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(56549,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(56549,36,'cp 9','d1a4 c7c6 a4c4 c8e6 c4c2 f7f6 g1f3 c6c5 e1g1 b8c6 e2e3 c6b4 c2d1 d8d3 a2a3 b4c6 b1c3 g8e7 b2b4 e7d5 c3d5 e6d5 b4c5 e8c8 c1b2 f8c5 a1c1 c5b6 c1c3 d3a6 d1c2 c8b8 f1b1 h7h6 b1c1 e5e4 f3e1');
INSERT INTO evaluations VALUES(56639,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(56639,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(56639,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(56639,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56639,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56639,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56639,36,'cp -9','c7c6 a4c4 b8a6 g1f3 c8e6 c4e4 f8d6 e1g1 g8f6 e4a4 e8g8 d2d4 e5e4 f3d2 d6b4 d2e4 f6e4 g2e4 d8d4 a4c2 h7h6 c1e3 d4f6 f1d1 b4a5 e3d4 f6e7 e4h7 g8h8 h7f5 a5b6 f5e6 e7e6 b1c3 a6b4 c2a4 c6c5 d4e3 a8d8 a4b3 e6b3 a2b3 f8e8 d1d8 e8d8 a1c1 b4c6 c3a4 f7f6 e3f4');
INSERT INTO evaluations VALUES(1126,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(1126,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(1126,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(1126,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(1126,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(1126,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(1126,36,'cp 187','c7c5 d4f4 f6d5 f4g3 d7d6 e1g1 e7e5 e2b5 c8d7 b5d7 e8d7 g3f3 a8e8 b1c3 d5c3 f3b7 d7e6 b2c3 e6f6 b7a7 f8e7 a7b7 h7h6 c1d2 e8b8 b7f3 e5f5 f3g3 f5c2 d2f4 g7g5 f4d6 e7d6 g3d6 f6g7 d6c5 h8c8 c5e5 f7f6 e5e7 g7g6 a2a4 c8e8 e7a7 c2c3 a4a5 c3e5 a7d7 e8e7 d7d2 b8b2 d2d8 g6g7 a5a6 e7a7');
INSERT INTO evaluations VALUES(763,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(763,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(763,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(763,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(763,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(763,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(763,36,'cp 176','d4f4 f6d5 f4g3 d7d6 e1g1 e7e5 e2b5 c8d7 b5d7 e8d7 g3f3 a8e8 b1c3 d5c3 f3b7 d7e6 b2c3 e6f6 b7a7 f8e7 a7b7 g7g6 c1d2 e5e2 d2f4 e2g4 g2g3 f6g7 a2a4 g6g5 f4d2 g4f5 a4a5 e7f6 d2e3 e8e7 b7c6 f6c3 a1a3 c3e5 a5a6 e7a7 f1b1 h8c8 c6d5 h7h6 d5c4 c8e8 g1g2 f5e6 c4d3 e5f6 b1b7');
INSERT INTO evaluations VALUES(3611,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3611,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(3611,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(3611,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(3611,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(3611,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(3611,36,'cp 21','c3a4 c5b6 a4b6 a7b6 a2a4 c8e6 c4e6 f7e6 c2c3 d8e8 c1e3 f6h5 h2h3 h5f4 e3f4 f8f4 e1e3 e8g6 b2b4 a8f8 d1e2 d6d5 e4d5 e6d5 a1f1 d5d4 c3d4 c6d4 f3d4 f4d4 b4b5 g6d6 e3e5 d4a4 e5e4 a4d4 e4d4 d6d4 f1c1 f8f7 c1c4 d4f6 c4e4');
INSERT INTO evaluations VALUES(5235,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(5235,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(5235,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(5235,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(5235,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(5235,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(5235,36,'cp 13','c4b5 d7d5 d2d3 g8f6 b1c3 c8d7 e4e5 f6g8 a2a4 h7h5 c3e2 d5d4 c2c3 a7a6 b5c6 d7c6 c3d4 h5h4 h2h3 c6f3 g2f3 d8d5 d4c5 d5f3 e2f4 f3d1 f1d1 e7c5 d3d4 c5a7 c1e3 g8e7 a1c1 a8d8 c1c7 a7b8 c7b7 b8e5 g1g2 e7f5 d4d5 f5e3 f2e3 e5f4 e3f4 e6d5');
INSERT INTO evaluations VALUES(5904,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(5904,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(5904,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(5904,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(5904,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(5904,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(5904,36,'cp 19','e1g1 g8e7 c2c4 d5c4 d3c4 e8g8 h2h3 e7f5 c1e3 d8f6 b1c3 f5e3 f2e3 f6h6 d1e2 c6e7 e3e4 c7c6 e2c2 d6c7 a1e1 c8e6 c4e6 h6e6 e4e5 c7b6 c3e4 a8d8 e4d6 e6g6 c2g6 h7g6 d6b7 b6d4 f3d4 d8d4 e5e6 f7e6 e1e6');
INSERT INTO evaluations VALUES(5903,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5903,36,'cp 0','g8e7 f1e1 e8g8 c1g5 f7f6 g5h4 c6b4 d3e2 c8f5 b1a3 b4c6 a3b5 c6b4');
INSERT INTO evaluations VALUES(8720,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(8720,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(8720,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(8720,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(8720,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(8720,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(8720,36,'cp 0','c2c4 d5c4 d3c4 e8g8 h2h3 e7f5 c1e3 d8f6 b1c3 f5e3 f2e3 f6h6 d1e2 c6e7 e3e4 c7c6 e2c2 d6c7 a1d1 g8h8 c3a4 f7f6 a4c5 e7g6 c4e6 g6f4 e6c8 a8c8 c5b7 h6g6 g1h1 f4h5 h1g1');
INSERT INTO evaluations VALUES(6559,30,'cp -553',NULL);
INSERT INTO evaluations VALUES(6559,31,'cp -544',NULL);
INSERT INTO evaluations VALUES(6559,32,'cp -540',NULL);
INSERT INTO evaluations VALUES(6559,33,'cp -526',NULL);
INSERT INTO evaluations VALUES(6559,34,'cp -538',NULL);
INSERT INTO evaluations VALUES(6559,35,'cp -537',NULL);
INSERT INTO evaluations VALUES(6559,36,'cp -538','c5d4 f1d3 d8h4 d1e2 d4c3 b2c3 d7d6 e1g1 g8f6 c1d2 e8g8 f2f4 f8e8 a1e1 c8g4 e2f2 h4f2 f1f2 g4d7 c3c4 a8d8 d2c3 d7c8 f2f3 b7b6 f3g3 g8h8 e4e5 d6e5 f4e5 f6g8 e5e6 f7f6 d3e4 c6e7 c4c5 c8e6 e4d3 e6d7 c5c6 e7c6 g3e3 h7h6 e3e8 d7e8 c3b2');
INSERT INTO evaluations VALUES(6714,30,'cp -517',NULL);
INSERT INTO evaluations VALUES(6714,31,'cp -522',NULL);
INSERT INTO evaluations VALUES(6714,32,'cp -532',NULL);
INSERT INTO evaluations VALUES(6714,33,'cp -513',NULL);
INSERT INTO evaluations VALUES(6714,34,'cp -513',NULL);
INSERT INTO evaluations VALUES(6714,35,'cp -540',NULL);
INSERT INTO evaluations VALUES(6714,36,'cp -525','f1d3 d8h4 d1e2 d4c3 b2c3 d7d6 e1g1 g8f6 f2f4 e8g8 c1d2 f8e8 a1e1 c8f5 e2f3 f5g4 f3f2 h4f2 f1f2 e8e7 h2h3 g4d7 c3c4 a8e8 f2e2 b7b6 e4e5 d6e5 f4e5 c6e5 g1f2 f6h5 g2g4 e5g4 h3g4 e7e2 e1e2 e8e2 d3e2 h5f6 d2f4 d7g4 e2f3 g4f3 f4c7 f3e4 c7e5 e4c2');
INSERT INTO evaluations VALUES(8257,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(8257,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(8257,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(8257,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(8257,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(8257,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(8257,36,'cp -12','e7e6 c4a2 e8g8 c1e3 d6d5 e4d5 f6d5 c3d5 e6d5 e1g1 g7b2 a1b1 b2c3 e3c5 f8e8 a2b3 d8a5 c5e3 c8e6 a3a4 a8c8 f3g5 c6d4 e3d4 c3d4 d1f3 g8g7 g5e6 f7e6 c2c4 e8f8 f3g4 a5c5 g4e6 d4f2 g1h1 c8e8 e6d5 e8e7 d5c5 f2c5 f1f8 g7f8 b3a2 b7b6');
INSERT INTO evaluations VALUES(9087,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(9087,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9087,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(9087,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(9087,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(9087,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(9087,36,'cp 21','e4d5 f6d5 e1g1 e8g8 f1e1 c8e6 c2c3 c5b6 c4b3 f8e8 b1a3 d8d7 f3g5 h7h6 g5e6 e8e6 a3c4 e6f6 c4e3 d5e3 c1e3 a8d8 d1h5 f6f5 h5e2 b6e3 e2e3 d7d6 a1d1 a7a5 b3c4 a5a4 c4b5 a4a3 b2b4 d6f6 e3e2 c6e7 d3d4');
INSERT INTO evaluations VALUES(9074,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(9074,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(9074,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(9074,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(9074,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(9074,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(9074,36,'cp 11','f6d5 e1g1 e8g8 f1e1 c8e6 c2c3 c5b6 c4b3 f8e8 b1a3 d8d7 f3g5 h7h6 g5e6 e8e6 a3c4 e6f6 c1e3 a8d8 c4b6 c7b6 d3d4 e5d4 e3d4 f6g6 d4e3 c6e5 d1h5 d5e3 e1e3 d7c6 g2g3 e5c4 b3c4 c6c4 b2b3 c4c6 a1e1 d8d2 e3e2 d2e2 h5e2 g6e6 e2d2 a7a6 e1e6');
INSERT INTO evaluations VALUES(11855,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(11855,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(11855,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(11855,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(11855,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(11855,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(11855,36,'cp -9','c6c5 d2d3 a7a5 f3h2 a5a4 c1e3 c8b7 h2g4 f8e7 g4f6 e7f6 a1b1 e8g8 b2b4 a4b3 a2b3 f6g5 e1g1 g5e3 f2e3 d8g5 d1e2 b7c8 e2f3 c7c6 b1a1 c8e6 a1a4 a8a4 c3a4 f7f5 e4f5 f8f5 f3e4 f5f7 f1f7 g8f7');
INSERT INTO evaluations VALUES(13043,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(13043,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(13043,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(13043,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(13043,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(13043,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(13043,36,'cp 144','d4d5 c6e5 f3e5 d6e5 f1e2 f8g7 d1a4 c8d7 a4b4 e8g8 b4b7 a7a5 e1g1 a5a4 b7b4 a8b8 b4a3 d8a5 a1b1 a5b4 a3b4 b8b4 e2d3 f8b8 a2a3 b4b7 d3c2 b8a8 c1e3 f6e8 f1d1 e8d6 e3c5 f7f5 c5b4 g7f6 d1c1 f6g5 c1e1 d6c4 c2d3 c4d2 b1d1 d2b3 d3c4');
INSERT INTO evaluations VALUES(15383,30,'cp -463',NULL);
INSERT INTO evaluations VALUES(15383,31,'cp -477',NULL);
INSERT INTO evaluations VALUES(15383,32,'cp -460',NULL);
INSERT INTO evaluations VALUES(15383,33,'cp -498',NULL);
INSERT INTO evaluations VALUES(15383,34,'cp -473',NULL);
INSERT INTO evaluations VALUES(15383,35,'cp -484',NULL);
INSERT INTO evaluations VALUES(15383,36,'cp -485','d5e4 f3g1 d8g5 e1f1 c8g4 d1e1 g5g6 h2h3 g4d7 h3h4 h7h5 c2c3 f8e7 b3c2 f7f5 d2d3 e4d3 c2d3 e8c8 d3c2 h8e8 g1e2 g6f7 e2g3 g7g6 f1g1 a7a6 g3f1 e7f6 f2f3 d7e6 c1g5 f6g5 h4g5 e6a2 c2a4 b7b5 a4c2 c8b7 f1d2 a2d5 e1e3 f7d7 h1h3 f5f4');
INSERT INTO evaluations VALUES(15728,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(15728,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(15728,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(15728,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(15728,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(15728,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(15728,36,'cp 195','f8e7 f1b5 c8d7 b5d7 d8d7 b1c3 h7h6 g5d2 e8g8 h2h3 c7c6 d4d3 f8e8 e1g1 c6d5 e4d5 e7d8 a1e1 d8a5 a2a3 e8e1 f1e1 a8e8 g1f1 a7a6 e1e8 d7e8 b2b4 a5b6 d2f4 e8e7 g2g4 a6a5 d3f5 f6d7 f4g3 d7e5 g3e5');
INSERT INTO evaluations VALUES(8674,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(8674,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(8674,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(8674,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(8674,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(8674,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(8674,36,'cp 199','f1b5 c8d7 b5d7 d8d7 f2f3 e8g8 b1c3 h7h6 g5d2 b7b5 d4d3 f8b8 d2f4 b5b4 c3e2 d7e8 e1g1 c7c6 d5c6 e8c6 e2d4 c6b6 g1h1 b8d8 d4f5 e7f8 f1d1 a7a5 f4e3 b6a6 d3a6 a8a6 a1c1 a6c6 e3d4 d6d5 d4f6 c6f6 f5e3');
INSERT INTO evaluations VALUES(18359,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(18359,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(18359,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(18359,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(18359,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(18359,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(18359,36,'cp 75','g8f6 c2c4 e7e6 e1g1 c8b7 d2d3 d7d5 c4d5 e6d5 d1a4 d8d7 c1f4 f8e7 b1c3 e8g8 a1c1 f8d8 a2a3 f6h5 f4g5 e7g5 f3g5 h5f6 g5f3 c6e7 a4d7 f6d7 b2b4 c7c5 b4c5 b6c5 c3a4 a8c8 g2h3 c5c4 d3c4 d5c4 f1d1 b7f3 e2f3 f7f5 f3f4 d7f6 d1d8 c8d8 c1c4');
INSERT INTO evaluations VALUES(23753,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(23753,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(23753,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(23753,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(23753,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(23753,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(23753,36,'cp 62','b5a4 a7a5 d2d4 f8e8 d4d5 c6b8 c3c4 d7a4 d1a4 c7c6 b1c3 c6d5 c4d5 d8d7 a4b5 e8c8 a2a4 d7b5 c3b5 c8c2 f3d2 b8a6 b5a3 a6b4 a1b1 c2c8 a3c4 f6d7 b2b3 b4c2 e1d1 c8c5 c1a3 c2a3 c4a3 c5c3');
INSERT INTO evaluations VALUES(30308,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30308,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30308,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(30308,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(30308,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30308,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(30308,36,'cp -6','b5e2 e7e6 g1f3 c7c5 e1g1 d8c7 d4c5 f8c5 b2b3 e8g8 c1b2 d7c6 b1d2 c7e7 a2a3 a7a5 c2c3 f8d8 b3b4 c5d6 b4b5 c6e8 d1b1 b8d7 c3c4 d7b6 c4d5 f6d5 d2e4 d6c7 a3a4 a8c8 f1d1 h7h6 d1c1 e6e5');
INSERT INTO evaluations VALUES(37617,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(37617,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37617,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37617,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(37617,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37617,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37617,36,'cp 0','b8c6 c1f4 d7d6 c3b5 e6e5 f4g5 g8e7 b5c3 h7h6 g5e7 c6e7 c3d5 c8d7 a2a4 e7d5 c4d5 d7c6 d5c6 b7c6 e1g1 e8g8 d1e2 d8d7 f3h4 d6d5 g2g3 a8e8 a1e1 g7f6 h4f3 f6g7');
INSERT INTO evaluations VALUES(39025,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(39025,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(39025,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(39025,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(39025,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(39025,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(39025,36,'cp 107','e2e4 d7d5 e4e5 g8e7 c2c3 c7c5 h2h4 b8c6 h4h5 h7h6 h5g6 f7g6 b1a3 g6g5 f4d2 e8g8 f1d3 c5d4 c3d4 d8b6 a3b5 e7f5 a2a4 a7a6 d3f5 e6f5 b5d6 f5f4 d2c3 c6e7 a4a5 b6c6 e1g1 c8g4 c3b4 c6d7 b4c5 e7f5 f1e1 f5d6 e5d6');
INSERT INTO evaluations VALUES(39039,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(39039,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(39039,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(39039,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(39039,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(39039,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(39039,36,'cp 82','d7d5 e4e5 c7c5 d4c5 g8e7 b1c3 b8d7 h2h4 h7h6 f1b5 a7a6 b5d7 c8d7 d1d2 a8c8 c3d1 d8c7 d1e3 c7c5 e1g1 g6g5 h4g5 h6g5 f4g5 e7g6 f1e1 c5c7 e3g4 c7c2 d2d4 c2c4 a1c1 c4d4 c1c8 d7c8 f3d4 e8d7 g4f6 g7f6 g5f6 h8h5 e1e3');
INSERT INTO evaluations VALUES(39594,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(39594,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(39594,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(39594,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39594,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39594,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(39594,36,'cp -11','d7d5 g1f3 g8f6 b2b4 e8g8 e1g1 b8d7 a2a4 f8e8 d3e2 e7e5 f3e5 d7e5 d4e5 e8e5 c1b2 e5e8 c2c4 c8e6 c4d5 f6d5 b2g7 g8g7 d1d4 d8f6 d4f6 g7f6 b4b5 a8c8 b1d2 d5c3 e2f3 e8d8 d2b1 c3b1 f1b1 c7c6 a4a5 c6b5 b1b5');
INSERT INTO evaluations VALUES(38465,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(38465,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(38465,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(38465,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(38465,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(38465,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(38465,36,'cp 0','g1f3 g8f6 e1g1 c7c5 d4c5 e8g8 c2c4 d5c4 d3c4 b8d7 c5c6 b7c6 b1c3 d7b6 c4e2 f6d5 c3e4 a7a5 a2a3 a5a4 f1e1 d8c7 e4c5 c7d6 c5e4');
INSERT INTO evaluations VALUES(38466,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38466,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38466,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38466,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38466,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38466,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38466,36,'cp 0','g8f6 e1g1 e8g8 c2c4 c7c5 d4c5 b8d7 c4d5 d7c5 d3c4 b7b6 b1c3 a7a6 a2a3 c8b7 f1e1 a8c8 c4a2 c5e4 c3e4 f6e4 d1d3 e4c5 d3d1');
INSERT INTO evaluations VALUES(40038,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40038,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(40038,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(40038,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(40038,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(40038,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40038,36,'cp 25','e2e3 e8g8 f1d3 c6c5 d4c5 d5c4 d3c4 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 b5d6 f8d8 d6e4 d8d1 e4f6 g7f6 f1d1 b4e7 h2h3 e6e5 g2g4 a7a5 a2a4 g8f8 h4g3 f8e8 g1f1 e7b4 f1e2 c8e6 c4e6 f7e6 f3d2 b4d2 d1d2 a8c8 f2f4 e5f4 e3f4');
INSERT INTO evaluations VALUES(40624,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(40624,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(40624,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40624,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(40624,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40624,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(40624,36,'cp 20','c4d5 f6d5 c1d2 c5d4 e3d4 c8d7 c3d5 b4d2 d1d2 e6d5 f1e2 b8c6 e1g1 e8g8 f1c1 d7f5 b2b4 a7a6 e2a6 d8b6 a6d3 f5d3 d2d3 b6b4 a1b1 a8a3 d3b5 b4b5 b1b5 a3a2 h2h4 h7h6 b5d5 a2a8 d5b5 f8d8 c1c4 c6a5 c4c7 a5c6 d4d5');
INSERT INTO evaluations VALUES(42447,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42447,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42447,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(42447,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(42447,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(42447,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42447,36,'cp 12','b2c3 c7c5 e2e3 c5d4 g5f6 d8f6 c3d4 b7b6 g1f3 c8b7 f1e2 e8g8 a1c1 b8c6 e1g1 f8c8 h2h3 h7h6 c4c5 c6e7 d1a4 b7c6 a4b4 c6d5 c5b6 a7b6 c1c8 e7c8 f1c1 f6d8 a2a3 c8e7 f3e5 e7c6 b4d6 c6e5 d4e5 d5c6 e2d3 a8c8 f2f4');
INSERT INTO evaluations VALUES(44229,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(44229,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(44229,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(44229,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(44229,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(44229,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(44229,36,'cp 52','d4d5 f6d5 c3d5 e6d5 d1d5 b8c6 e2e4 f8e7 f1c4 e8g8 e1g1 d7d6 f1d1 c8e6 d5d3 e6c4 d3c4 d8b6 c1f4 f8e8 a1b1 e7f8 f3g5 b6c7 g5h3 a7a6 f2f3 c6e5 c4d5 b7b5 f4g3 e5c4 d5d3 c4e5 d3e2 e5g6 c2c4 b5c4 e2c4 c7b6');
INSERT INTO evaluations VALUES(44546,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(44546,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(44546,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(44546,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(44546,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(44546,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(44546,36,'cp -26','d7d5 e2e3 e7e6 d4c5 b8c6 f1d3 f8c5 e1g1 e8g8 e3e4 d5d4 c3e2 e6e5 a2a3 a7a5 h2h3 h7h6 c2c3 d4c3 e2c3 c8e6 c1e3 c5e3 f2e3 a8c8 d1e2 d8e7 a1d1 f8d8 g1h2 g8h8 d3b5 d8d1 f1d1 e6b3 d1d2 e7c5 h2h1');
INSERT INTO evaluations VALUES(45398,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(45398,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45398,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(45398,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45398,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45398,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45398,36,'cp 0','e2e3 e7e6 c2c4 b8c6 c4d5 e6d5 f1b5 f8d6 e1g1 e8g8 d2d4 c5d4 f3d4 d6c7 b5c6 b7c6 d1c2 c8a6 f1c1 c6c5 c2c5 c7h2 g1h2 f6e4 c5c2 a6d3 c2d3 d8h4 h2g1 h4f2 g1h2 f2g3 h2g1');
INSERT INTO evaluations VALUES(46597,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(46597,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(46597,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(46597,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(46597,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(46597,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(46597,36,'cp 110','c2c4 e7e6 e2e3 f8d6 b1c3 e8g8 a1c1 b8d7 d1b3 d8e7 f4d6 c7d6 c4d5 d7b6 d5e6 c8e6 b3c2 f5f4 e3e4 e6f5 f1d3 f6e4 e1g1 e4c3 c2c3 f5d3 c3d3 e7f7 f1e1 b6d5 d3a3 f8d8 e1e2 a7a5 c1e1 h7h6 h2h4 b7b6 e2e6 d5e3 e6d6 e3c2 d6d8 a8d8');
INSERT INTO evaluations VALUES(48211,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48211,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48211,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(48211,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(48211,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(48211,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48211,36,'cp 0','b1d2 c7c5 g1f3 b8d7 f1e2 f8e7 e1g1 e8g8 c2c3 h7h6 g5f4 f6h5 f4e5 h5f6');
INSERT INTO evaluations VALUES(50672,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(50672,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(50672,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(50672,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(50672,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(50672,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(50672,36,'cp 32','d2d4 c5d4 d1d4 g8f6 g1f3 e7e6 e4e5 f6d7 d4f2 d6e5 f4e5 d8b6 f2b6 d7b6 c1e3 b6d5 c3d5 e6d5 e1c1 c8e6 f1e2 b8d7 c2c3 a8c8 e2d3 f8c5 e3c5 c8c5 d3c2 e8e7 h1e1 g7g6 f3d4 h8c8 d1d2 c5c4 a2a3');
INSERT INTO evaluations VALUES(53295,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(53295,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(53295,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(53295,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(53295,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(53295,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(53295,36,'cp -31','g1f3 e7e6 b2b3 f8e7 c1b2 g8f6 f1b5 b8d7 e1g1 a7a6 b5e2 b7b6 f3e5 c8b7 c2c4 e8g8 e2f3 d7e5 f4e5 f6e4 d2d3 e4g5 f3e2 f7f5 e5f6 e7f6 b2f6 f8f6 b1d2 d5d4 e3e4 f6f1 d1f1 d8d6 f1e1 a8f8 e2f3 d6e5');
INSERT INTO evaluations VALUES(56762,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(56762,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(56762,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(56762,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(56762,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(56762,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(56762,36,'cp 107','b1c3 g8f6 c1f4 f8d6 e2e3 e8g8 a1c1 b8c6 a2a3 c6e7 c4c5 d6f4 e3f4 c8d7 f1d3 d7e8 f3e5 f6d7 e5d7 e8d7 e1g1 e7c6 d3c2 b7b6 b2b4 a7a6 d1d2 d8f6 c3e2 b6b5 c1a1 a6a5 h2h3 d7e8 f1c1 g8h8 c1e1 e8h5 f2f3 h5f7');
INSERT INTO evaluations VALUES(57692,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(57692,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(57692,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(57692,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(57692,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(57692,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(57692,36,'cp 23','g8f6 d2d4 f8g7 c2c4 e8g8 f1e2 c7c5 d4d5 e7e6 b1c3 d7d6 e1g1 e6d5 c4d5 c8g4 h2h3 g4f3 e2f3 a7a6 a2a4 b8d7 c1d2 c5c4 f3e2 a8c8 d1c2 f8e8 f1d1 d8e7 d2e1 f6e4 a4a5 e4c3 e1c3 e7e4 a1c1 e4c2 c1c2 g7c3 c2c3 b7b5 a5b6 d7b6');
INSERT INTO evaluations VALUES(4657,30,'cp -421',NULL);
INSERT INTO evaluations VALUES(4657,31,'cp -425',NULL);
INSERT INTO evaluations VALUES(4657,32,'cp -446',NULL);
INSERT INTO evaluations VALUES(4657,33,'cp -446',NULL);
INSERT INTO evaluations VALUES(4657,34,'cp -406',NULL);
INSERT INTO evaluations VALUES(4657,35,'cp -434',NULL);
INSERT INTO evaluations VALUES(4657,36,'cp -413','c6e5 d1h5 f7e6 f2f4 d7d6 f4e5 d6e5 h5h3 e6e7 h3g3 e7f7 g3e5 d8h4 g2g3 h4e7 e5e7 g8e7 c2c3 h8e8 d2d4 c5b6 a2a4 c7c5 a4a5 b6d8 d4c5 e7c6 e1g1 f7g8 c1f4 c6a5 b1d2 c8e6 f4d6 a5b3 d2b3 e6b3 e4e5 a7a5 f1f4 d8g5 a1a3 b3f7 h2h4 g5h6 g1f2 f7e6 f2e3 a8d8 e3d4 h6f4 g3f4');
INSERT INTO evaluations VALUES(4669,30,'cp -437',NULL);
INSERT INTO evaluations VALUES(4669,31,'cp -438',NULL);
INSERT INTO evaluations VALUES(4669,32,'cp -431',NULL);
INSERT INTO evaluations VALUES(4669,33,'cp -430',NULL);
INSERT INTO evaluations VALUES(4669,34,'cp -427',NULL);
INSERT INTO evaluations VALUES(4669,35,'cp -447',NULL);
INSERT INTO evaluations VALUES(4669,36,'cp -451','d1h5 f7e6 f2f4 d7d6 f4e5 d6e5 h5h3 e6e7 h3g3 e7f7 g3e5 c5d4 h1f1 g8f6 e5h5 f7g8 h5h4 c8e6 d2d3 d8d6 b1c3 d6b4 e1d1 d4c3 b2c3 b4c3 f1f6 c3f6 h4f6 g7f6 c1f4 a8c8 d1d2 g8f7 a1f1 h8g8 g2g3 b7b6 f4e5 g8g6 a2a3 c7c5 e5c3 e6h3 f1f4 g6h6 d3d4 c5d4 c3d4 c8d8 d2e3 h3g2');
INSERT INTO evaluations VALUES(4668,30,'cp -463',NULL);
INSERT INTO evaluations VALUES(4668,31,'cp -453',NULL);
INSERT INTO evaluations VALUES(4668,32,'cp -440',NULL);
INSERT INTO evaluations VALUES(4668,33,'cp -465',NULL);
INSERT INTO evaluations VALUES(4668,34,'cp -462',NULL);
INSERT INTO evaluations VALUES(4668,35,'cp -451',NULL);
INSERT INTO evaluations VALUES(4668,36,'cp -451','f7e6 f2f4 d7d6 f4e5 d6e5 h5h3 e6e7 h3g3 e7f7 g3e5 c5d4 h1f1 g8f6 e5h5 f7g8 h5h4 c8e6 d2d3 d8d6 b1d2 f6g4 d2f3 d4f6 h4g3 d6g3 h2g3 b7b6 a2a4 h7h6 a1b1 c7c5 b2b3 a8e8 e1e2 f6e7 c1d2 g8h7 d2c3 h8f8 f3d2 h7g8 f1f8 e8f8 d2f3 e7d6 b3b4 c5c4 c3d4 c4d3 c2d3 f8c8 a4a5');
INSERT INTO evaluations VALUES(5612,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(5612,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(5612,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(5612,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(5612,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(5612,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(5612,36,'cp 184','f1d3 f7f6 e1g1 g8h6 c1h6 g7h6 e5f6 e7f6 d3b5 c8d7 f1e1 d8e7 b5c6 d7c6 f3e5 f6e5 d1h5 e8d7 e1e5 e7g7 b1d2 a8f8 e5e3 h8g8 e3g3 g7f6 h5e2 d7c8 a1e1 c6d7 d2f3 a7a6 e2d2 d7b5 d2e3 b5d7 f3e5 d7b5 g3f3 f6g5 f3f8 g8f8');
INSERT INTO evaluations VALUES(12632,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(12632,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(12632,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(12632,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(12632,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(12632,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(12632,36,'cp 83','d8a5 h2h4 f8c5 d4d2 b7b5 e4e5 b5b4 e5f6 b4c3 b2c3 g7f6 g5f6 h8g8 h1h3 a5c7 d2d3 g8g6 f6d4 c8b7 a1d1 e8c8 h4h5 g6g4 d3h7 f7f5 h7h6 g4e4 f1e2 b7a6 d1d2 a6e2 d2e2 c5d4 c3d4 c7c2 e2e4 c2e4 h6e3 e4b1 e1e2 b1a2 e2f3 a2d5 f3f4 d5g2 e3f3 g2f3 h3f3 d8g8 h5h6 c8d8 f3h3 g8h8 f4g5');
INSERT INTO evaluations VALUES(14481,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(14481,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(14481,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(14481,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(14481,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(14481,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(14481,36,'cp 72','e2e4 e7e5 f1b5 e5d4 f3d4 c8d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 f8e8 c1e3 d8b8 a1b1 f6d7 e3f2 b8b4 b2b3 e7f6 e4e5 d6e5 f3c6 d7b6 c3e4 e8e6 e4f6 g7f6 c6f3 b4f4 f3f4 e5f4 f2d4 a8e8 c2c4 b6d7 f1f4 d7e5 f4f5 c7c6 b1d1');
INSERT INTO evaluations VALUES(15981,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(15981,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(15981,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(15981,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(15981,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(15981,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(15981,36,'cp -37','f8b4 f1d3 e8g8 e1g1 b4c3 b2c3 f8e8 f1e1 h7h6 g5h4 d8e7 a2a4 d7f8 h4g3 c8g4 h2h3 g4f3 d1f3 e7c5 d3f1 f8g6 f3e3 c5e3 e1e3 f6h5 a1d1 h5g3 f2g3 a8d8 e3d3 d8c8 a4a5 g6f8 f1e2 b7b6 a5b6 a7b6');
INSERT INTO evaluations VALUES(17288,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(17288,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(17288,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(17288,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(17288,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(17288,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(17288,36,'cp 25','d2d4 d7d6 c1b2 c8f5 e2e3 c6b4 b1a3 e7e6 c2c3 b4c6 f1d3 d6d5 e1g1 f8e7 a3c2 e8g8 c3c4 a7a5 c2e1 c6b4 d3f5 e6f5 f3e5 b4a6 a2a3 c7c6 e1d3 a6c7 c4c5 f6d7 e5f3 d7f6 b3b4 a5b4 a3b4 a8a1');
INSERT INTO evaluations VALUES(19913,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19913,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19913,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(19913,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(19913,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19913,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19913,36,'cp -7','g7g6 g1f3 f8g7 e1g1 g8f6 f1d1 e8g8 h2h3 d8c7 b5f1 c6b4 f3e1 b7b6 c2c4 c8e6 c1d2 a7a5 a2a3 b4a6 e1f3 f8c8 f3d4 c7d7 d4e6 d7e6 b1c3 a6c5 d1e1 a5a4 a1d1 f6d7 e3g5 g7f6 g5g3 c5b3 d2h6 f6c3 g3c3 e6e5 c3c2 d7c5 h6e3 e5e4 c2e4 c5e4 e3b6 b3c5');
INSERT INTO evaluations VALUES(20123,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(20123,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(20123,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(20123,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(20123,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(20123,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(20123,36,'cp 53','g8f6 b1c3 g7g6 f1e2 f8g7 c1e3 e8g8 d4d3 c8e6 e1g1 f6d7 b2b3 d8a5 e3d2 a5e5 a1c1 d7c5 d3c2 a8c8 f1e1 f7f5 e4f5 e5f5 c2f5 e6f5 h2h3 g8f7 d2e3 f8d8 e1d1 f5d7 h3h4 h7h5 b3b4 c5e6 c3d5 d7a4 d1d2');
INSERT INTO evaluations VALUES(20441,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(20441,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(20441,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(20441,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(20441,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(20441,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(20441,36,'cp -48','g8e7 f1d3 e7g6 d1e2 d8c7 d3g6 h7g6 b1c3 a7a6 e1g1 d5d4 c3b1 f7f6 b1d2 c8d7 f1e1 e8c8 h2h3 g6g5 e2c4 f8e7 c2c3 c8b8 d2b3 d4d3 c1e3 c6e5 c4c7 b8c7 f3e5 f6e5 e3d2 d7c6 e1e5 e7f6 e5e6 c6d5 b3c5 d5e6');
INSERT INTO evaluations VALUES(22819,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(22819,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22819,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(22819,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22819,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(22819,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(22819,36,'cp 5','b1c3 g8f6 g1f3 f8b4 f1b5 e8g8 e1g1 f8e8 f1e1 a7a6 b5c4 b4c5 c3d5 d7d6 c2c3 h7h6 d2d4 c5a7 d4e5 f6d5 c4d5 c6e5 f3e5 d6e5 d1f3 d8e7 e1d1 c7c6 d5c4 c8e6 c4e6 e7e6 c1e3 a7e3 f3e3');
INSERT INTO evaluations VALUES(24764,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(24764,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(24764,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(24764,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(24764,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(24764,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(24764,36,'cp 9','f3e5 c6d7 e1g1 e7e6 c2c4 f8d6 b1c3 e8g8 a2a4 a7a6 e5d7 d8d7 c4c5 d6e7 b2b4 b7b6 a4a5 b6c5 b4c5 f8b8 c1d2 b8b2 a1b1 a8b8 b1b2 b8b2 d1c1 b2b4 c1a3 b4b8 f1a1 f6e4 c3e4 d5e4 a3c3 e7f6 c3c2');
INSERT INTO evaluations VALUES(24765,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(24765,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(24765,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24765,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24765,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24765,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(24765,36,'cp 7','c6d7 e1g1 e7e6 c2c4 d5c4 d1f3 d7b5 f3b7 a8b8 b7f3 d8d5 f3d1 f8d6 b1c3 d5b7 b2b3 c4b3 c3b5 b7b5 a2b3 e8g8 e5c4 b5b3 d1b3 b8b3 a1a7 f8b8 c1a3 d6a3 a7a3 b3b4 c4e5 g7g6 e5c6 b4b3 a3b3 b8b3');
INSERT INTO evaluations VALUES(25374,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(25374,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(25374,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(25374,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(25374,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(25374,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(25374,36,'cp 38','c4b5 f6d7 d2d4 c6d4 b5d7 d8d7 f3d4 e5d4 d1d4 f7f6 f2f4 e6f7 c1e3 f8e7 e1c1 c7c6 g2g4 e8g8 h1g1 b7b5 f4f5 d6d5 e4d5 b5b4 c3e4 d7d5 g4g5 f6g5 e3g5 d5d4 d1d4 f8d8 g5f6 g7g6 f5g6 h7g6 g1d1 d8d4 d1d4 e7f6 e4f6 g8g7 f6e4 a8e8');
INSERT INTO evaluations VALUES(30707,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(30707,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(30707,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(30707,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(30707,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(30707,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(30707,36,'cp 17','d7d5 d2d3 g8f6 c2c4 f8e7 c4d5 e6d5 b1c3 e8g8 c1f4 c7c5 d3d4 b8d7 a1c1 h7h6 f3e5 f8e8 d4c5 d7c5 c3b5 e7f8 e5f3 a8c8 b5a7 c8a8 a7b5 a8a2 d1c2 d8d7 b5c3 a2a8 f3e5 d7e7 f1d1 a8c8 c3d5 b7d5 g2d5 f6d5 d1d5');
INSERT INTO evaluations VALUES(31088,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(31088,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(31088,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(31088,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(31088,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(31088,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(31088,36,'cp -19','g8f6 f1d3 e7e6 d1e2 f5d3 c2d3 c6c5 d4c5 f8c5 d3d4 c5d6 e2b5 d8d7 b5d7 b8d7 c3b5 d6b8 e1e2 a7a6 b5c3 b8d6 f3e1 a8c8 c1d2 f6e4 e1d3 h7h5 h1c1 e8e7 h2h4 c8c7 a2a4 h8c8 a4a5 f7f6 c1c2 e4c3 c2c3 e6e5 c3c7 c8c7 d2b4 d6b4 d3b4 e7e6');
INSERT INTO evaluations VALUES(34238,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(34238,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(34238,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(34238,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(34238,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(34238,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(34238,36,'cp 132','c1d2 a5c7 b1c3 d7d6 e2e4 e7e5 d2e3 b8d7 h2h3 g8f6 g1e2 a7a6 g2g4 h7h6 d4d5 c6d5 c4d5 b7b5 e2g3 d7b6 a1c1 c7d8 b2b3 f8e7 c3e2 f6h7 c1c6 a8b8 f1g2 c8d7 g3f5 d7f5 e4f5 e8g8 e1g1 f8e8 d1d2 e7g5 e2g3 g5e3');
INSERT INTO evaluations VALUES(34237,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(34237,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(34237,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(34237,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(34237,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(34237,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(34237,36,'cp 118','a5c7 b1c3 g8f6 e2e4 d7d6 d2e3 f6g4 e3c1 g4f6 h2h3 e7e5 c1e3 a7a6 d4d5 f8e7 f1d3 b8d7 g1e2 c6d5 c4d5 d7c5 d3c2 e8g8 b2b4 c5d7 a1c1 b7b5 e1g1 d7b6 c2d3 c7b7 f2f4 f6d7 e2g3 b6c4 d3c4 e5f4 e3f4 b5c4 g3f5');
INSERT INTO evaluations VALUES(36771,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(36771,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(36771,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(36771,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(36771,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(36771,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(36771,36,'cp -28','e1g1 c7c5 a2a4 e7e6 c1d2 b7b6 a4a5 c8b7 f3e5 f6e8 f1e1 e8d6 b2b3 d8e7 d1e2 f8c8 a1a4 f7f6 e5f3 b8d7 d3a6 b7a6 e2a6 d7b8 a6f1 b8c6 a5b6 a7b6 e1a1 a8a4 c3a4 e7b7 c2c4 d6e4 f1e2 c6b4');
INSERT INTO evaluations VALUES(36770,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(36770,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(36770,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(36770,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(36770,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(36770,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(36770,36,'cp -22','c7c5 a2a4 b7b6 a4a5 c8a6 a5b6 a7b6 b2b3 d8c8 c1b2 c8b7 d3a6 b8a6 c3b5 f6e4 d1e2 f8d8 h2h3 a6b4 c2c4 a8a1 f1a1 d8a8 a1a8 b7a8 b5c3 e7e6 c4d5 e4c3 b2c3 b4d5 e2d3 g7f6 d3c2 d5c3 c2c3 a8d5 c3c2 c5d4 c2c8 g8g7 f3d4');
INSERT INTO evaluations VALUES(41263,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(41263,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(41263,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(41263,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41263,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(41263,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(41263,36,'cp 39','e1g1 e8g8 d2d4 b8c6 b1c3 e7e5 d4d5 c6b8 e2e4 a7a5 a2a3 b8d7 c1g5 h7h6 g5e3 f6g4 e3d2 d7c5 b2b4 a5b4 a3b4 a8a1 d1a1 c5b3 a1a2 b3d2 f3d2 h6h5 h2h3 g4f6 d2f3 f6h7 c4c5 h7g5 f3g5 d8g5 c3b5 h5h4 g3g4');
INSERT INTO evaluations VALUES(42469,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(42469,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(42469,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(42469,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(42469,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(42469,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(42469,36,'cp 21','d1c2 e8g8 a2a3 b4c3 c2c3 f8e8 b2b4 d7d6 e2e3 e6e5 c1b2 b8d7 d2d3 d7b6 c3c2 c8d7 a1b1 b6a4 b2a1 c7c5 f1e2 c5b4 b1b4 b7b5 c4b5 a8c8 c2d2 d8a5 b4b1 a5d2 e1d2 f6d5 h1c1 c8c1 b1c1 d7b5 c1b1 b5c6 f3h4 g7g6 h2h3 a7a5');
INSERT INTO evaluations VALUES(45490,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(45490,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(45490,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(45490,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(45490,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(45490,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(45490,36,'cp 47','g1f3 e7e6 e1g1 f8e7 b1c3 e8g8 f1e1 b8c6 a2a3 b7b6 c4a2 c8b7 c1f4 f8e8 f4g3 h7h6 c3b5 a8c8 f3e5 c6e5 d4e5 f6e4 b5d6 e7d6 e5d6 d8d7 g3e5 f7f6 e5g3 e4g3 h2g3 c8d8 e1e6 e8e6 d1g4 d7d6 a2e6 g8h8 e6f5 d6c6 f2f3 b7a6');
INSERT INTO evaluations VALUES(51584,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(51584,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(51584,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(51584,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(51584,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(51584,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(51584,36,'cp 40','g1f3 e7e6 e1g1 c8b7 f1e1 d8c7 c2c3 g8f6 e4e5 f6d5 a2a3 a7a6 d2d4 b6b5 c4f1 c5d4 c3d4 f8e7 c1g5 b8c6 g5e7 c6e7 b1d2 e8g8 d2e4 h7h6 d1d2 e7f5 a1c1 c7d8 e4c5 b7c6 f1d3 d7d6 c5e4 d6e5');
INSERT INTO evaluations VALUES(51705,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(51705,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(51705,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(51705,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(51705,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51705,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51705,36,'cp 20','c4b3 f8e7 a2a4 g8f6 a4b5 c6b5 b1c3 c8b7 d2d4 b8d7 c3b5 f6e4 e1g1 e8g8 c1e3 a7a6 b5a3 d8c7 a3c4 a8c8 f1e1 g8h8 c4d2 e4d2 e3d2 e5e4 d2f4 f7f5 f3g5 e7g5 f4g5 f5f4 d1g4 d6d5 g2g3 f4g3');
INSERT INTO evaluations VALUES(7009,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(7009,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(7009,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(7009,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(7009,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(7009,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(7009,36,'cp 55','g8f6 f1d3 c7c6 e1g1 e8g8 b1c3 d6c7 h2h3 d7d6 c1e3 c7a5 a1c1 e7g6 a2a3 a5b6 e3b6 d8b6 d5c6 b7c6 b2b4 b6a6 d1b3 h7h6 f1d1 c8e6 d1d2 a6b6 d3f1 f8d8 c1d1 b6c7 b3c2 a8c8');
INSERT INTO evaluations VALUES(11043,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(11043,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(11043,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(11043,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(11043,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(11043,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(11043,36,'cp 62','f1d3 c7c6 e1g1 e8g8 h2h3 d6c7 b1c3 d7d6 c1e3 e7g6 f1e1 c7b6 e3b6 a7b6 a2a4 h7h6 b2b3 f6h7 a1a2 c6c5 g1h2 g6f4 d3f1 f4g6 g2g3 g6e7 f3h4 f7f5 e4f5 e7f5 h4f5 c8f5 e1e3 d8d7 d1d2 h7g5 h3h4 g5h7 f1g2');
INSERT INTO evaluations VALUES(9521,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(9521,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(9521,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(9521,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(9521,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(9521,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(9521,36,'cp -14','d8e7 b1d2 c8e6 b2b3 f6d7 d2c4 f7f6 c4d6 e7d6 c1e3 e6g4 d1e2 d7c5 h2h3 g4h5 e3c5 d6c5 f1c1 c5e7 e2e3 h5f3 e3f3 e8c8 f3e3 c8b8 f2f4 e5f4 e3f4 b7b6 c1f1 a7a5 f4f5 h7h5 a1e1 d8e8 a2a4 h5h4 f5f2 e7e5 d3d4 e5g5 f2f3 b8b7');
INSERT INTO evaluations VALUES(11420,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(11420,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(11420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11420,36,'cp 0','b7b5 d4e5 d6e5 d1d8 c6d8 a4b3 f6d7 a2a4 b5b4 c1e3 f7f6 b1d2 d7c5 b3c4 c8e6 e3c5 f8c5 f1d1 e8e7 g1f1 g7g6 c4e6 d8e6 d2b3 h8d8 f3d2 c5b6 d2c4 a8c8 h2h4 e6d4 a1c1 d4e6');
INSERT INTO evaluations VALUES(11646,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11646,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11646,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11646,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11646,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11646,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11646,36,'cp 1','d4e5 d6e5 d1d8 c6d8 a4b3 f8d6 a2a4 c8b7 f1e1 d8e6 b3e6 f7e6 b1d2 e8g8 b2b3 h7h6 c1b2 f6d7 b2c3 f8d8 h2h3 b7c6 a4b5 a6b5 a1a8 d8a8 e1a1 a8a1 c3a1 c6b7 g1f1 d7c5 a1e5 c5e4 d2e4 b7e4 e5d6 c7d6 f3d4 e6e5 d4b5 e4c2 b5d6');
INSERT INTO evaluations VALUES(11704,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(11704,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(11704,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(11704,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(11704,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(11704,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(11704,36,'cp -49','f6e4 d1e2 e4c5 f3e5 c6e5 e2e5 d8e7 e5e7 f8e7 b3c2 d7d6 d2d4 c5a4 b2b3 a4b6 c2e4 a8b8 b1a3 f7f5 e4d3 e8f7 e1g1 c8e6 c1f4 b6d5 f4d2 c7c5 a3c2 e7f6 c3c4 d5e7 d4d5 e6d7 a1e1 g7g5 g2g3 e7g6 f2f4 g5f4 d2f4 g6f4 f1f4 b5c4 d3c4 f6c3');
INSERT INTO evaluations VALUES(13645,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(13645,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(13645,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(13645,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(13645,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(13645,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(13645,36,'cp 28','f8d6 b1d2 b7b5 f4d6 d8d6 e1g1 e8g8 f1e1 c8b7 e3e4 d5e4 d2e4 f6e4 c2e4 h7h6 a2a4 a7a6 d4d5 e6d5 d1d5 d6d5 e4d5 c6d8 d5b7 d8b7 e1e5 f7f6 e5h5 g7g6 h5h6 g8g7 h6h4 b7c5 a4b5 a6b5 a1b1 a8a2 f3d4 f8b8 g2g3 c5d3 h4e4 g7f7');
INSERT INTO evaluations VALUES(8404,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(8404,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8404,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(8404,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(8404,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(8404,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(8404,36,'cp 38','f4g3 b7b5 b1d2 e8g8 e3e4 f6e4 d2e4 d5e4 c2e4 c8b7 d1e2 a8b8 e1g1 h7h6 a2a4 a7a6 f1e1 d8d7 a1a2 d7c7 g3d6 c7d6 h2h4 d6c7 e1a1 b8d8 a4b5 a6b5 e2e3 f8e8 h4h5 e6e5 d4d5 c6e7 d5d6 d8d6 e4b7 c7b7 a2a7 b7d5 e3e5 d5e5 f3e5');
INSERT INTO evaluations VALUES(15465,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15465,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(15465,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(15465,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(15465,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(15465,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(15465,36,'cp 30','f8e7 f1b5 c8d7 b5c6 d7c6 d1e2 f6e4 c3e4 c6e4 e2e4 e7g5 d4e5 e8g8 e4b7 a8b8 b7a7 b8b2 f3g5 d8g5 e1g1 g5e5 a7c7 e5e6 a2a4 f8c8 c7a7 g7g6 a4a5 c8c2 a5a6 e6f6 a1a3 c2c3 a7a8 g8g7 a3a1 c3c2 a8d5 c2f2 g2g3 f2f1 a1f1');
INSERT INTO evaluations VALUES(8575,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(8575,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(8575,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(8575,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(8575,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(8575,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(8575,36,'cp 54','f1b5 c8d7 d4d5 c6b8 a2a4 e8g8 g5f6 e7f6 b5d7 b8d7 h2h4 f6e7 f3d2 f7f5 e4f5 f8f5 g2g3 d8e8 d2e4 h7h6 d1d3 f5f8 e1c1 d7f6 b2b3 f6e4 c3e4 a8b8 g3g4 b7b5 a4a5 f8f4 g4g5 h6h5 f2f3 e8f7 h1f1 a7a6 c1b1 b8f8');
INSERT INTO evaluations VALUES(15478,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15478,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15478,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15478,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15478,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15478,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15478,36,'cp 0','d6d5 e4d5 c6d4 c2c3 d4f5 d5d6 f5d6 g5f7 d8d7 f7d6 f8d6 e1g1 b7b5 c4d5 f6d5 d1d5 a8b8 b1d2 c8b7 d5b3 b7c6 a2a4 b5b4 d2c4 c6g2 c4d6 c7d6 g1g2 d7g4 g2h1 g4f3 h1g1 f3g4');
INSERT INTO evaluations VALUES(15083,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15083,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15083,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15083,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15083,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15083,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15083,36,'cp 0','e4d5 c6d4 c2c3 d4f5 d5d6 f5d6 g5f7 d8d7 f7d6 f8d6 e1g1 b7b5 c4d5 f6d5 d1d5 a8b8 a2a4 a7a6 a4b5 c8b7 d5b3 b7g2 g1g2 d7g4 g2h1 g4f3 h1g1 f3g4');
INSERT INTO evaluations VALUES(14982,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14982,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14982,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14982,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14982,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14982,36,'cp 0','c6d4 c2c3 d4f5 d5d6 f5d6 g5f7 d8d7 f7d6 f8d6 e1g1 b7b5 c4d5 f6d5 d1d5 a8b8 a2a4 a7a6 a4b5 c8b7 d5b3 b7g2 g1g2 d7g4 g2h1 g4f3 h1g1 f3g4');
INSERT INTO evaluations VALUES(15533,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(15533,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(15533,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(15533,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(15533,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(15533,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(15533,36,'cp 67','f1g2 f8e7 d2d4 a7a5 b2b3 c8d7 e2e3 e8g8 e1g1 f8e8 h2h3 e7f8 d4d5 c6b4 e3e4 c7c6 a2a3 b4a6 c1e3 a8c8 f3d2 c6d5 e4d5 d7f5 d2e4 h7h6 g3g4 f5g6 d1e2 f6d7 a1b1 a6c5 b3b4 c5e4 g2e4 a5b4 a3b4');
INSERT INTO evaluations VALUES(19333,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19333,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19333,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19333,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19333,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19333,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19333,36,'cp 0','g1f3 d7d6 f1e2 g8f6 e1g1 c8d7 b1c3 f8g7 a4b3 e8g8 f1d1 c6a5 b3b4 a5c6');
INSERT INTO evaluations VALUES(20652,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20652,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(20652,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(20652,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(20652,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(20652,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(20652,36,'cp 12','c8d7 d2d4 c5d4 f3d4 g8f6 c1e3 f8e7 d1e2 e8g8 e1c1 d8c7 g2g4 c6d4 d1d4 d7b5 c3b5 c7c6 b5c3 d6d5 e4d5 f6d5 c3d5 e6d5 d4d1 f8e8 e2d3 a8d8 g4g5 e7c5 e3d2 c6a4 c1b1 a4e4 d2c1 d5d4 a2a3 e4d3 d1d3 g7g6 h1d1 c5b6 c1d2 e8e2');
INSERT INTO evaluations VALUES(25981,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25981,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(25981,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(25981,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(25981,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(25981,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(25981,36,'cp 29','d2d4 c5d4 f3d4 g8f6 c1e3 f8e7 d1e2 e8g8 e1c1 d8c7 g2g4 c6d4 d1d4 d7b5 c3b5 c7c6 b5c3 d6d5 e4d5 f6d5 c3d5 e6d5 d4d1 a8c8 c1b1 f8e8 e2d3 e7c5 e3c5 c6c5 c2c3 c8d8 h1e1 e8e1 d1e1 c5f2 d3e3 f2e3 e1e3 g8f8 a2a3');
INSERT INTO evaluations VALUES(23554,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(23554,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(23554,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(23554,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(23554,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(23554,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(23554,36,'cp -90','a8d8 f3d2 f7f5 e1g1 f5e4 b1c3 e4e3 f2e3 f8b4 d2f3 b4d6 a2a4 a7a6 b2b3 g8f6 c1a3 d6a3 a1a3 c6f3 g2f3 e8e7 f1f2 d8d6 f2g2 g7g6 a3a1 h8d8 a4a5 d6d2 a1e1 d2g2 g1g2 d8d2 e1e2 d2e2 c3e2 f6d5 g2f2 d5b4 c2c3 b4d3 f2g3 e7d7 f3f4 e5f4 e3f4');
INSERT INTO evaluations VALUES(65,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(65,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(65,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(65,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(65,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(65,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(65,36,'cp -87','f3d2 f7f5 e1g1 f5e4 b1c3 e4e3 f2e3 f8b4 d2f3 b4d6 a2a4 g8f6 b2b3 a7a6 c1a3 e8g8 h2h3 d6a3 a1a3 c6f3 f1f3 d8d2 a3a2 b7b6 a2b2 f8d8 f3f5 d8e8 f5f1 h7h6 f1f2 d2d6 g1h2 d6c6 c3b1 f6e4 f2e2 e8d8 b1a3 d8d5 a3c4 g8f7');
INSERT INTO evaluations VALUES(25388,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25388,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25388,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25388,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25388,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25388,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25388,36,'cp 0','h2h3 g4h5 g2g4 h5g6 f3e5 f6e4 e5g6 e4c3 d2c3 h7g6 d1d8 a8d8 e1e2 f7f5 c1f4 f8d6 f4g5 d6e7 g5f4');
INSERT INTO evaluations VALUES(25387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25387,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25387,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25387,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25387,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25387,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25387,36,'cp 0','g4h5 g2g4 h5g6 f3e5 f6e4 d1e2 e4c3 d2c3 d8e7 e1f1 e8c8 e5g6 e7e2 f1e2 h7g6 c1e3 f8d6 h3h4 h8h7 h1h3 c6c5 a2a4 d8h8 a1h1 c8d7 c3c4 b7b6 b2b3 f7f5 e2f3 d6e7 f3g3 e7d6');
INSERT INTO evaluations VALUES(25385,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25385,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25385,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25385,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25385,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25385,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25385,36,'cp 0','g2g4 h5g6 f3e5 f6e4 e5g6 e4c3 d2c3 h7g6 d1d8 a8d8 e1e2 f7f5 c1f4 f8d6 f4g5 d6e7 g5f4');
INSERT INTO evaluations VALUES(28383,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(28383,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(28383,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(28383,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(28383,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(28383,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(28383,36,'cp 112','a2a3 b8c6 c3e2 d8d7 d2d3 a8d8 c4a2 d6d5 e1g1 e8g8 c1e3 g8h8 a1d1 d5d4 e3c1 f6g8 f3g4 g8f6 g4g3 d8e8 c2c3 e8d8 a2c4 a7a6 c4a2 a6a5 f1e1 d7d6 f2f4 f6h5 g3f2 h5f4 e2f4 e5f4 c1f4');
INSERT INTO evaluations VALUES(29953,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(29953,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(29953,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(29953,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(29953,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(29953,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(29953,36,'cp 73','d2d4 b8c6 d4d5 c6d4 c1e3 d4f3 d1f3 f8e7 e1c1 a7a6 c1b1 d8d7 h2h3 h7h5 h1g1 h5h4 e3c1 b7b5 b2b3 f6h7 g2g3 h4g3 f2g3 d7h3 g1h1 h3d7 d1g1 e8c8 f3f7 h7f6 f7g7 h8g8 g7h6 d8f8 h6e3 c8b7 c1b2 d7d8');
INSERT INTO evaluations VALUES(29939,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(29939,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(29939,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(29939,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(29939,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(29939,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(29939,36,'cp 81','b8c6 d4d5 c6d4 c1e3 d4f3 d1f3 f8e7 f3e2 f6d7 e2b5 e7g5 e3g5 d8g5 e1g1 b7b6 b5c6 g5d8 c3b5 e8e7 a2a4 a7a6 b5c3 h8e8 a4a5 d8b8 f1b1 b6b5 b2b4 d7f6 f2f3 b8d8 c3d1 d8d7 c6d7 e7d7 d1e3 h7h5 g1f2 a8b8 c2c4 b5c4 e3c4 b8b5 c4e3');
INSERT INTO evaluations VALUES(30562,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(30562,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(30562,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(30562,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(30562,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(30562,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(30562,36,'cp 119','g1f3 g8f6 f1d3 g7g6 a1b1 c7c5 d4d5 f8g7 e1g1 e8g8 h2h3 a7a6 a2a4 b8d7 c1g5 d8c7 d1d2 a8e8 b1e1 f6h5 g2g4 h5f6 g1g2 e8e7 g5f4 f8e8 f4g3 e6d5 e4d5 e7e1 f1e1 h7h6 d3c2 e8e1 d2e1 b7c8 e1d2');
INSERT INTO evaluations VALUES(30561,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(30561,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(30561,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(30561,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(30561,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(30561,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(30561,36,'cp 108','g8f6 f1d3 g7g6 e1g1 f8g7 f1e1 e8g8 h2h3 b8d7 c1g5 h7h6 g5e3 e6e5 d4d5 a7a5 d1d2 g8h7 a2a3 f6h5 d3c2 b7a6 b2b3 d8e7 g2g3 f7f5 e4f5 g6f5 d2e2 e7e8 g1h1 e5e4 f3d4 a5a4 c3a4 e8g6 a4c3 g7d4 e3d4 f8g8 a1d1 d7e5');
INSERT INTO evaluations VALUES(28643,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(28643,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(28643,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(28643,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(28643,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(28643,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(28643,36,'cp 124','f1d3 g7g6 e1g1 f8g7 h2h3 b8d7 c1g5 h7h6 g5e3 e8g8 d1d2 g8h7 a1d1 e6e5 d3c2 d8e7 d4d5 a7a5 a2a3 b7c8 b2b3 f6h5 c3b5 d7f6 f1e1 c8a6 d1c1 a6c8 c1b1 c8a6 g2g4 h5f4 e3f4 a6b5 c4b5 e5f4 d2f4');
INSERT INTO evaluations VALUES(28644,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(28644,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(28644,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(28644,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(28644,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(28644,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(28644,36,'cp 112','g7g6 e1g1 b8d7 h2h3 f8g7 c1e3 e8g8 d1d2 a7a6 a1d1 b6b5 c4b5 a6b5 b2b4 c7c6 f1e1 f8e8 d3c2 d8e7 c2b3 e7f8 d1c1 d7b6 e4e5 f6d5 c3e4 d5e3 e1e3 d6e5 d4e5 e8d8 d2e1 g7h6 e4d6 f8e7 c1d1 b6d5 b3d5 h6e3 e1e3');
INSERT INTO evaluations VALUES(31325,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(31325,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(31325,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(31325,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(31325,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(31325,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(31325,36,'cp -41','e7e6 d2d4 f8d6 d1e2 b8c6 c2c3 d8a5 e1g1 h7h6 f3e5 a5c7 f1e1 g8e7 c1f4 e8g8 b1d2 c6e5 f4e5 d6e5 e2e5 c7e5 e1e5 f8c8 a2a4 a7a5 b3d1 c8c7 d2b1 f5b1 a1b1 e7f5 d1e2 g7g6 e2d3 f5d6 g2g4 a8c8 g1g2 h6h5');
INSERT INTO evaluations VALUES(31704,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(31704,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(31704,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(31704,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(31704,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(31704,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(31704,36,'cp -34','d2d4 f8d6 e1g1 b8c6 c2c4 d5c4 b3c4 g8e7 b1c3 e8g8 d4d5 e6d5 c3d5 c6a5 d5e7 d6e7 c4e2 e7f6 d1a4 a5c6 a4b5 f5e4 f1d1 d8c8 c1e3 a7a6 b5b3 c8f5 e2d3 e4d3 b3d3 f5d3 d1d3 f8d8 d3d8 a8d8 a1b1 c6e7 e3g5 f6g5 f3g5 h7h6 g5f3');
INSERT INTO evaluations VALUES(39705,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39705,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(39705,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39705,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39705,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39705,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(39705,36,'cp 0','e8g8 a2a4 b8c6 g5e3 c8d7 d4b3 c6a5 d1d2 f8e8 b3a5 d8a5 f1d1 d7c6 f2f3 a5c7 a4a5 d6d5 e3b6 c7b8 e4d5 c6d5 a1a4 e7d6 g2g3 d6g3 h2g3 b8g3 g1f1 g3h3 f1g1');
INSERT INTO evaluations VALUES(35567,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35567,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35567,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35567,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35567,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35567,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35567,36,'cp 0','a2a4 b8c6 g5e3 c8d7 d4b3 c6a5 d1d2 f8e8 f2f3 a8c8 b3a5 d8a5 f1d1 a5c7 e3d4 d7c6 g1h1 h7h6 e2f1 f6d7 d4g1 d7f6');
INSERT INTO evaluations VALUES(41203,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41203,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41203,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41203,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41203,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41203,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41203,36,'cp 0','e8g8 f2f3 c7c5 d4c5 d8a5 d1d2 d6c5 a1d1 b8c6 g1e2 f8d8 c3d5 g8f8 b2b3 c6b4 e1g1 b7b5 g5f6 g7f6 d2f4 b5c4 d5f6 e7f6 f4f6 c8e6 e2f4 f8g8 f4e6 f7e6 f6e6 g8h8 e6f6 h8g8');
INSERT INTO evaluations VALUES(36464,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36464,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36464,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36464,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36464,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36464,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36464,36,'cp 0','f2f3 c7c5 d4c5 d8a5 d1d2 d6c5 a1d1 b8c6 g1e2 f8d8 c3d5 g8f8 b2b3 c6b4 e1g1 b7b5 g5f6 g7f6 d2f4 b5c4 d5f6 e7f6 f4f6 c8e6 e2f4 f8g8 f4e6 f7e6 f6e6 g8h8 e6f6 h8g8');
INSERT INTO evaluations VALUES(41376,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(41376,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(41376,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(41376,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(41376,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(41376,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(41376,36,'cp 46','g1f3 b8d7 e1g1 e8g8 c1e3 e7e5 b1d2 b7b6 f1e1 c8b7 h2h3 f8e8 d4d5 c7c6 c3c4 d7c5 d3c2 c6d5 c4d5 a7a5 a2a3 b7a6 d1b1 a6b5 b2b4 a5b4 a3b4 a8a1 b1a1 c5d3 c2d3 b5d3 e3g5 h7h6 a1c3 h6g5 c3d3 g7h6 d2c4 d8c7 f3d2');
INSERT INTO evaluations VALUES(43793,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(43793,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(43793,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(43793,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(43793,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43793,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(43793,36,'cp 27','f8e7 c4d5 f6d5 g5e7 d8e7 a1c1 e8g8 c3d5 e6d5 e2e3 c8b7 f1d3 c7c5 d1c2 g7g6 d4c5 f8c8 e1g1 b6c5 f1e1 c5c4 d3e2 b8d7 b2b3 c8c7 c2d2 a8c8 c1c2 d7f6 d2d4 a7a5 b3c4 b7a6 e1c1 d5c4 e2c4 a5a4 c4d3 a6d3 d4d3');
INSERT INTO evaluations VALUES(46894,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(46894,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(46894,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(46894,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(46894,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(46894,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(46894,36,'cp -28','c7c5 e2e3 b8c6 h4f6 g7f6 d4c5 e7e6 f1b5 f8c5 g1f3 d8b6 d1d3 a7a6 b5c6 b7c6 e1g1 c5e7 b2b3 c6c5 e3e4 d5e4 c3e4 c8b7 f1d1 e8f8 e4d6 a8d8 d6c4 d8d3 c4b6 d3c3 d1d7 b7c6 d7c7 c6f3 g2f3 h8g8 g1f1 c3c2 a1d1 g8g5 d1d7 c2c1 f1e2 g5e5 e2d3');
INSERT INTO evaluations VALUES(52357,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(52357,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(52357,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(52357,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(52357,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(52357,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(52357,36,'cp -26','a7a5 b4a5 d7d6 c2c4 g8f6 g1f3 e7e5 e2e3 d8a5 d1c2 b8a6 b1c3 g7g6 f1e2 c8f5 d2d3 e5e4 d3e4 f6e4 e1g1 e4c3 c2c3 a5c3 b2c3 h8g8 a2a4 a6c5 a4a5 e8d8 a1a3 d8c7 f3g5 f8g7 c3d4 g7d4 e3d4 c5e6 g5e6 f5e6');
INSERT INTO evaluations VALUES(54990,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54990,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54990,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54990,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(54990,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54990,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54990,36,'cp 0','e7e5 c2c4 g7g6 c1b2 f8g7 g2g3 c7c6 f1g2 g8h6 b1c3 f7f5 e1g1 e8g8 d2d3 b8a6 d1d2 a6c7 e2e3 c7e6 c3e2 a7a5 f3e1 a5a4 a1b1 g6g5 f2f4 a4b3 a2b3 a8a2 e2c3 a2a8 c3e2');
INSERT INTO evaluations VALUES(55631,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55631,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55631,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(55631,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(55631,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55631,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55631,36,'cp 13','c2c4 g7g6 c1b2 f8g7 g2g3 c7c6 f1g2 g8e7 e1g1 e8g8 d1c2 c6c5 b1c3 b8c6 d2d3 h7h6 e2e3 c8e6 a1d1 a7a6 f3d2 d8d7 a2a3 a8d8 c3d5 e6h3 b2c3 h3g2 g1g2 f7f5 e3e4 f5f4 b3b4 e7d5');
INSERT INTO evaluations VALUES(55744,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(55744,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(55744,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(55744,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(55744,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(55744,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(55744,36,'cp 17','d3d4 d6d5 e4e5 c7c5 c2c3 c8d7 g1f3 b8c6 f1e2 c5d4 c3d4 g8e7 b1c3 e7f5 e1g1 a7a6 c1e3 f8e7 a1c1 f5e3 f2e3 a8c8 e2d3 e8g8 g1h1 g7g6 e3e4 d5e4 d3e4 f7f5 e5f6 e7f6 d1d3 g8h8 a2a3 f6g7 d3d2 d7e8 c3e2 d8e7');
INSERT INTO evaluations VALUES(57075,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(57075,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(57075,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(57075,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(57075,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(57075,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(57075,36,'cp -56','c7c5 f3d1 g8f6 g1f3 d7d5 d2d4 b8c6 a2a3 b7b6 b2b3 f8d6 c1b2 e8g8 b1d2 d8c7 f1d3 e6e5 d4e5 c6e5 b2e5 d6e5 f3e5 c7e5 e1g1 h7h6 a3a4 c8b7 a4a5 b7c6 d1e2 f8e8 f1d1 e5c3 d2f1 b6a5 f1g3 a8d8');
INSERT INTO evaluations VALUES(57086,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(57086,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(57086,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(57086,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(57086,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(57086,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(57086,36,'cp -71','c7c6 b5f1 g8f6 d2d4 e5e4 c2c4 d7d5 b1c3 f8d6 g1e2 b8a6 c4d5 a6b4 e2f4 d6f4 e3f4 b4d5 f1c4 c8e6 c4d5 e6d5 e1g1 d8d7 f1e1 e8g8 h2h3 f8e8 c1e3 d5c4 a1c1 c4d3 a2a4 a7a5 d4d5 f6d5 c3d5 d7d5 c1c5 d5a2 d1a1 a2a1 e1a1 g7g6');
INSERT INTO evaluations VALUES(51822,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(51822,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(51822,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(51822,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(51822,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(51822,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(51822,36,'cp -62','b5f1 d7d5 d2d4 e5e4 c2c4 f8d6 b1c3 g8f6 g1e2 b8a6 c4d5 a6b4 e2f4 d6f4 e3f4 b4d5 f1c4 c8e6 c4d5 e6d5 e1g1 e8g8 h2h3 d8d7 c1e3 d5c4 f1e1 c4d3 a2a4 a7a5 d1b3 f8e8 d4d5 f6d5 c3d5 d7d5 b3d5 c6d5 e1c1');
INSERT INTO evaluations VALUES(3563,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3563,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3563,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3563,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3563,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3563,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3563,36,'cp 0','e3b6 a7b6 d3d4 e5d4 f3d4 f6e4 c3e4 d6d5 d4c6 b7c6 c4d3 d5e4 d3e4 d8f6 b2b3 c8f5 e4f5 f6f5 f1e1 f8e8 e1e8 a8e8 d1d2 e8e6 a2a4 e6d6 d2e2 d6g6 g2g3 f5c5 g1g2 g6e6 e2d3 e6d6 d3e4 d6e6');
INSERT INTO evaluations VALUES(3818,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3818,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(3818,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(3818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3818,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3818,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3818,36,'cp 0','c3a4 c5b6 a4b6 a7b6 c4b3 c8g4 c2c3 d6d5 e4d5 c6a5 e1e5 g4f3 g2f3 a5b3 d1b3 f6d7 e5e3 d8g5 g1h1 g5h4 h1g2 h4g5');
INSERT INTO evaluations VALUES(4217,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(4217,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(4217,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(4217,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(4217,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(4217,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(4217,36,'cp 36','c3d5 b4c5 c2c3 a7a5 c1e3 c5e3 f2e3 d7d6 d5f6 d8f6 d3d4 f6e7 d1e1 a5a4 f3h4 c6a5 c4d3 e7g5 h4f3 g5e7 h2h3 c8e6 e1e2 c7c6 c3c4 e6d7 f3d2 f8b8 f1f2 b7b5 c4b5 c6b5 a1c1 d7e8 c1f1 e7g5 b2b4 a4b3 a2b3');
INSERT INTO evaluations VALUES(6922,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(6922,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(6922,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(6922,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(6922,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(6922,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(6922,36,'cp 138','g8e7 a2a4 d7d6 h2h3 c8d7 c3e2 d8c8 c2c3 a7a6 d3d4 c5a7 e2g3 h7h5 c1e3 h5h4 g3e2 c6d8 e1g1 d7e6 c4e6 c8e6 d4d5 e6d7 e3a7 a8a7 a4a5 a7a8 c3c4 e7g6 b2b4 e8g8 a1a3 d8f7 d1d2 d7e7 d2c2 f7g5 c4c5 f8f7 f3g5 f6g5 c5c6 b7c6 c2c6 a8b8 c6a6 b8b4');
INSERT INTO evaluations VALUES(10992,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(10992,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(10992,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(10992,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(10992,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(10992,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(10992,36,'cp 104','a2a4 d7d6 h2h3 c5b4 f3h4 g7g5 d1h5 e8d7 h4f5 e7f5 e4f5 d8e8 c4f7 e8f8 e1g1 b4c3 b2c3 d7d8 a4a5 a7a6 a1b1 f8g7 f7d5 c8f5 c1e3 g7g6 h5e2 f5d7 d3d4 c6a5 c3c4 e5d4 e3d4 a5c6 e2d2 d8c8 f2f4 h8e8 f4f5 d7f5 b1b7 c6e7 f1b1 e7d5 c4d5 g5g4 h3h4 c8d7 d2a5');
INSERT INTO evaluations VALUES(8375,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(8375,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(8375,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8375,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(8375,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(8375,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(8375,36,'cp -12','e1g1 b4c3 e4d5 d8d5 b2c3 e8g8 f3g5 f8d8 d1e2 c8g4 d3h7 f6h7 e2g4 h7f6 g4e2 c6a5 g5f3 e5e4 f3g5 d5c4 e2c4 a5c4 f1e1 a8c8 h2h4 d8e8 c1e3 c4a3 a1d1 a3c2 e1e2 c2e3 e2e3 c8c4 f2f3 g8f8 g5e4 f6e4 f3e4 c4e4 e3e4 e8e4 d1d8 f8e7 d8b8 e4a4 b8b7 e7f6 g1f2 a4a2 f2f3');
INSERT INTO evaluations VALUES(10559,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(10559,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(10559,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(10559,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(10559,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(10559,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(10559,36,'cp 165','c2c3 c5e7 d2d4 d7d6 c4b5 e7f6 b5c6 b7c6 d1a4 e5d4 c3d4 e8g8 a4c6 a8b8 f1e1 d6d5 e4e5 f6e7 c1h6 b8b6 c6c2 b6h6 b1c3 g7g5 c3e2 g5g4 f3d2 f7f6 a1d1 f6e5 d4e5 c7c6 d2f1 g8h8 e2g3');
INSERT INTO evaluations VALUES(10856,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(10856,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(10856,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(10856,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(10856,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(10856,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(10856,36,'cp -50','e7f6 b1c3 c6e7 g2g3 c7c6 f1h3 d8b6 c3a4 b6a5 a4c3 g7g6 e1g1 e8g8 f1e1 f6g7 a2a4 a5c7 d3d4 e5d4 f3d4 d7d5 h3c8 a8c8 d1d3 f8d8 a1d1 a7a6 d4f3 d5e4 d3e4 e7d5 c3d5 c6d5 e4b4 g7f8 b4b3 c7c2 b3b7 c8b8 b7a6 b8b2');
INSERT INTO evaluations VALUES(9990,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(9990,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(9990,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(9990,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(9990,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(9990,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(9990,36,'cp -27','b1c3 c6e7 d1d2 d7d5 e4d5 e8g8 f1e2 e7d5 c3e4 f8e8 e1g1 c7c6 h2h3 f6e7 f3e5 f7f5 c2c4 f5e4 c4d5 e7g5 d2c3 g5f6 d3d4 c6d5 f2f4 c8e6 c3e3 a8c8 a1c1 d8b6 b2b3 f6e7 e2g4 e8f8 c1c8 f8c8 f1f2 c8c7 g2g3 h7h6 g4e6 b6e6 g1g2 e6f6 e5g4 f6d6');
INSERT INTO evaluations VALUES(9989,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(9989,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(9989,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(9989,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(9989,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(9989,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(9989,36,'cp -47','c6e7 d1d2 d7d5 e4d5 e7d5 c3e4 e8g8 f1e2 c7c5 e1g1 d8c7 e4f6 d5f6 f1e1 f8e8 d2c3 c8d7 a2a3 h7h6 b2b4 b7b6 e2f1 e5e4 d3e4 f6e4 c3b3 c5b4 a3b4 a7a5 b3c4 c7c4 f1c4 e4c3 f3e5 d7e6 c4e6 e8e6 b4a5 b6a5 a1a3');
INSERT INTO evaluations VALUES(11497,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(11497,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(11497,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(11497,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(11497,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(11497,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(11497,36,'cp -19','d7c6 b1c3 f8d6 d2d4 d6b4 e1g1 e8g8 d4e5 b4c3 d1d8 f8d8 e5f6 c3f6 c1g5 f6g5 f3g5 f7f6 g5f3 c8g4 f1d1 g4f3 g2f3 g8f7 g1f1 d8d6 f1e2 a8d8 d1d6 c7d6 e2e3 g7g6 a1d1 f7e7 b2b4 d6d5 e4d5 d8d5 d1d5 c6d5 e3d3 e7e6 f3f4 h7h5 a2a4');
INSERT INTO evaluations VALUES(11456,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(11456,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(11456,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(11456,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(11456,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(11456,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(11456,36,'cp -12','b1c3 c8e6 h2h3 f8c5 f3e5 d8d4 e5d3 f6e4 d1e2 e4c3 d2c3 d4c4 d3c5 c4c5 c1f4 e8c8 e1g1 c5f5 f4g3 h8e8 f1e1 f7f6 a1d1 e6f7 d1d8 c8d8 e2d2 f5d7 e1e8 f7e8 d2e2 e8g6 a2a3 d7e7 e2e7 d8e7 g3c7 g6c2 c7b6');
INSERT INTO evaluations VALUES(12680,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12680,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(12680,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(12680,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(12680,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12680,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(12680,36,'cp 18','a7a6 a2a4 b7b6 d2d4 c6d4 f3d4 c5d4 d1d4 e7c6 d4d1 f8c5 c1f4 c8b7 d1d2 d8f6 a1d1 e8g8 f4e3 d7d6 h2h3 f6e7 f2f4 a8d8 d2f2 c6a5 c4e2 a5c6 f1e1 c6b4 e2f1 b7c6 e3c5 b6c5');
INSERT INTO evaluations VALUES(12699,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(12699,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(12699,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(12699,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(12699,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(12699,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(12699,36,'cp 179','d4b5 e7g6 b5d6 f8d6 d1d6 d8e7 d6e7 g6e7 c1f4 f7f5 e1c1 f5e4 c3e4 d7d5 e4d6 e8d7 f4g3 h7h5 h2h4 h8f8 f1e2 e7f5 d6f5 f8f5 c2c4 d7e7 h1e1 e7f7 e2f3 d5c4 d1d6 c6b4 c1d2 f7e7 d6d4 b4c6 d4c4 c8d7 a2a4 a8d8 d2c3 e6e5 c4c5 b7b6 c5c4 d8c8 f3e4 f5f6 b2b3 d7f5');
INSERT INTO evaluations VALUES(14276,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(14276,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(14276,32,'cp -143',NULL);
INSERT INTO evaluations VALUES(14276,33,'cp -157',NULL);
INSERT INTO evaluations VALUES(14276,34,'cp -126',NULL);
INSERT INTO evaluations VALUES(14276,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(14276,36,'cp -142','e7e5 f1d3 f8d6 g1e2 d8e7 b1a3 a7a6 e1g1 e5e4 f3e4 d5e4 d3c2 d6h2 g1h2 f6g4 h2g1 e7h4 f1f4 h4h2 g1f1 h2h1 e2g1 f7f5 f4f2 h7h5 f1e2 g4f2 e2f2 h1h4 f2e2 h4h2 e2f2 h5h4 g1e2 g7g5 d1g1 h2d6 a3c4 d6f6 d4d5 c6e7 g1h2 e7d5 h2e5 c8e6 e2d4 e8e7 c1d2 a8d8');
INSERT INTO evaluations VALUES(15112,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(15112,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(15112,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(15112,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(15112,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(15112,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(15112,36,'cp -122','f1d3 f8d6 g1e2 e5e4 f3e4 d5e4 d3c2 f6g4 h2h3 d8h4 e1d2 g4f6 d1e1 h4g5 e1f2 e8g8 d2d1 b7b6 b1d2 f8e8 d2c4 d6f8 h1f1 c8a6 c2b3 a8d8 g2g4 h7h5 g4h5 g5h5 d1e1 b6b5 c4d2 c6a5 e2g3 h5g6 f1g1 a5b3 a2b3 a6c8 g3e2 g6h6 f2f4 h6f4 e2f4 c7c5 a1a7 c5d4 c3d4 f8d6 d2f1 d6f4 e3f4');
INSERT INTO evaluations VALUES(14803,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(14803,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(14803,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(14803,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(14803,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(14803,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(14803,36,'cp 22','f8d6 e2e3 e8g8 f3e5 f6d7 e5d7 c8d7 f4g3 d8e7 d2f3 h7h6 f1e2 a7a5 e1g1 d6g3 h2g3 f8d8 a2a4 d7e8 d1b3 b7b6 f1d1 e7d6 b3c2 c6e7 e2d3 c7c5 e3e4 d5e4 d3e4 e8c6 d4c5 d6c5 e4c6 c5c6');
INSERT INTO evaluations VALUES(9237,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(9237,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(9237,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(9237,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(9237,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(9237,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(9237,36,'cp 34','e2e3 e8g8 f4g3 b7b6 f1d3 c8b7 e1g1 c6e7 b2b4 e7f5 g3e5 f6e4 d1c2 e4d2 f3d2 d6e5 d4e5 g7g6 e3e4 f5e7 f1e1 c7c5 e4d5 b7d5 c3c4 d5b7 d3e4 b7e4 d2e4 d8c7 e4f6 g8g7 b4b5 a8d8 g2g4 d8d4 c2e2 c7b7 a1d1 f8d8 d1d4');
INSERT INTO evaluations VALUES(15134,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15134,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(15134,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(15134,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(15134,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(15134,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(15134,36,'cp 42','e4d5 f6d5 e1g1 c8e6 d3b5 d5c3 b2c3 d8d6 f1e1 e8c8 b5c6 d6c6 f3e5 c6a4 d2d4 f8d6 c1f4 d6e5 f4e5 a4c6 e1e3 f7f6 e5g3 h7h5 h2h4 g7g6 d1d2 h8e8 a1e1 e6f7 e3e7 e8e7 e1e7 d8d7 e7e1 b7b6');
INSERT INTO evaluations VALUES(15004,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(15004,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(15004,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(15004,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(15004,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(15004,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(15004,36,'cp 38','f6d5 e1g1 c8e6 d3b5 d5c3 b2c3 d8d6 f1e1 e8c8 b5c6 d6c6 f3e5 c6a4 d2d3 f7f6 e5f3 e6g4 c1d2 g4h5 e1e4 a4d7 e4d4 f8d6 d2f4 d7c6 f4d6 d8d6 d4d6 c7d6 c3c4 d6d5 c4d5 c6d5 c2c4 h5f3 g2f3 d5g5 g1h1 h8d8 a1b1 g5f4 h1g2 f4g5 g2f1 g5h5 d3d4 h5h2 d1d3 h2h1 f1e2 d8e8 e2d2');
INSERT INTO evaluations VALUES(15347,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(15347,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(15347,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(15347,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(15347,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(15347,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(15347,36,'cp 35','e1g1 c8e6 d3b5 d5c3 b2c3 d8d5 d1e2 e8c8 c3c4 d5d6 b5c6 d6c6 f3e5 c6e8 f1e1 f8d6 c1b2 e8a4 e2e3 h8e8 e3b3 a4a6 d2d3 f7f6 e5f3 e6d7 d3d4 e8e1 f3e1 c7c5 d4d5 d8e8 e1f3 a6b6 a1e1 e8e1 f3e1 d7f5 b3b6 a7b6 f2f3 f5d7 g2g4 b6b5 e1g2 b7b6 g2e3 b5b4 e3f5 d7f5 g4f5 c8d7 h2h3');
INSERT INTO evaluations VALUES(15716,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(15716,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(15716,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(15716,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(15716,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(15716,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(15716,36,'cp -47','e4d6 f3e5 f8e7 d2d3 e8g8 c1f4 f8e8 b1d2 c8e6 h2h3 e7f8 e5g4 d6f5 g4e3 f5e7 d1f3 f7f6 a2a4 d8d7 f3g3 e7d5 e3d5 c6d5 a4a5 e6f7 d3d4 d7f5 c2c3 f5c2 b2b4 c7c6 g3f3 f7g6 h3h4 c2d3 f3d3 e8e1 g1h2 g6d3 a1e1');
INSERT INTO evaluations VALUES(17608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17608,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(17608,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(17608,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(17608,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(17608,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(17608,36,'cp 18','b7c6 g1f3 c5d4 c3d4 c6c5 e1g1 c5d4 f3d4 c8d7 b1d2 g8e7 d2f3 e7g6 g2g3 f8b4 h2h4 d8b8 d1e2 b4c5 f1e1 h7h6 c1e3 c5b6 d4b3 e8g8 a1c1 g6e7 e3b6 b8b6 e2e3 b6e3 e1e3 f8c8 e3c3 c8c3 c1c3 e7c6 b3d4 a8c8 d4c6 c8c6 c3c6 d7c6 f3d4 c6b5 h4h5');
INSERT INTO evaluations VALUES(18234,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(18234,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(18234,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(18234,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(18234,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(18234,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(18234,36,'cp 10','g1f3 c5d4 c3d4 c8d7 e1g1 c6c5 b1c3 c5d4 c3e2 g8e7 e2d4 e7g6 f1e1 h7h6 g2g3 a8c8 d1d3 a6a5 h2h4 f8c5 c1e3 e8g8 e1c1 f8e8 c1c5 c8c5 d4e6 e8e6 e3c5 g6e5 d3c3 e5f3 c3f3 d7c6 c5d4 d8d6 a2a3 d6d7 a1c1');
INSERT INTO evaluations VALUES(19490,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(19490,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(19490,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19490,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19490,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19490,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19490,36,'cp 0','c1b2 g8f6 e2e3 g7g6 d2d4 c5d4 f3d4 f8g7 f1e2 c6d4 b2d4 d7d5 b1d2 e8g8 e1g1 c8f5 c2c4 a8c8 a1c1 d8a5 d4c3 a5b6 c3d4');
INSERT INTO evaluations VALUES(20837,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(20837,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(20837,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(20837,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(20837,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(20837,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20837,36,'cp 1','g7g5 f4g3 g5g4 f3g1 e7e5 e2e3 d8d6 b1d2 h7h5 h2h4 g4h3 g1h3 c8g4 f1e2 g8h6 e2g4 h6g4 g3h4 e8c8 d4e5 c6e5 d1e2 f8h6 e1c1 h8g8 c1b1 d6d7 b1a1 a7a5 f2f4 e5f7 h3f2 d8e8 d2f1 a5a4 f2g4 d7g4 e2g4 g8g4 h4f6 g4g2 h1h5');
INSERT INTO evaluations VALUES(22340,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(22340,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(22340,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(22340,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(22340,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(22340,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(22340,36,'cp -66','d1e2 c8f5 d2d3 d8e7 d3e4 e7e4 e2e4 f5e4 c2c3 e4f3 g2f3 g8f6 c1g5 e8c8 b1d2 d8d5 h2h4 f6h5 e1c1 f8d6 d1e1 h8f8 e1e4 h5f6 e4e6 c8d7 h1e1 f6h5 e6e4 c6c5 c1c2 h5f6 e4e2 d5f5 g5f6 f5f6 h4h5');
INSERT INTO evaluations VALUES(22341,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(22341,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(22341,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(22341,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(22341,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(22341,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(22341,36,'cp -94','c8f5 d2d3 d8e7 d3e4 e7e4 e2e4 f5e4 c2c3 e4d3 c1e3 f8d6 b1a3 e8c8 e1c1 d3e2 d1d4 e2f3 g2f3 d6a3 d4d8 c8d8 b2a3 g8e7 h1g1 e7f5 h2h4 h7h5 e3g5 d8d7 g1d1 d7c8 d1e1 h8f8 a3a4 b7b6 a4a5 c8d7 e1e2 c6c5 a5b6 a7b6 a2a3');
INSERT INTO evaluations VALUES(26925,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(26925,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(26925,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(26925,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(26925,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(26925,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(26925,36,'cp 76','h2h3 g4d7 d2d3 a7a6 b5a4 e5f4 c1f4 b7b5 a4b3 c6a5 e1g1 g8f6 c2c4 b5c4 b3c4 f8e7 b1c3 e8g8 b2b3 h7h6 g1h1 a5c4 d3c4 d7c6 d1d3 f6h5 f3d4 c6b7 f4d2 e7g5 d4f5 g5d2 d3d2 d8g5 d2g5 h6g5 c3d5 b7d5 e4d5');
INSERT INTO evaluations VALUES(27522,30,'cp -147',NULL);
INSERT INTO evaluations VALUES(27522,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(27522,32,'cp -158',NULL);
INSERT INTO evaluations VALUES(27522,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(27522,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(27522,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(27522,36,'cp -171','c5d4 c3d4 f8b4 b1d2 g8e7 a2a4 e7f5 f1d3 f5e3 f2e3 e8c8 d3b5 f7f6 e1g1 c8b8 g1h1 h8g8 e5f6 g7f6 f3g1 e6e5 d2f3 a7a6 b5c6 d7c6 a1c1 b4a3 c1a1 a3d6 g1e2 c6e8 f3h4 d6b4 e2g3 e8g6 a1c1 b6e6 h1g1 d8c8 c1c8 g8c8 h4g6 h7g6 d4e5 f6e5 g1h1');
INSERT INTO evaluations VALUES(27549,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(27549,31,'cp -160',NULL);
INSERT INTO evaluations VALUES(27549,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(27549,33,'cp -159',NULL);
INSERT INTO evaluations VALUES(27549,34,'cp -171',NULL);
INSERT INTO evaluations VALUES(27549,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(27549,36,'cp -139','c3d4 f8b4 b1d2 g8e7 a2a4 e7f5 f1d3 f5e3 f2e3 e8c8 e1g1 f7f6 d3b5 c8b8 d1e1 f6f5 f3g5 c6d4 e3d4 b6d4 e1f2 d4d2 g5f7 d7c8 a1a2 d2f2 g1f2 b4c3 a2e2 g7g5 b5d3 d8f8 f7h8 f8h8 f1c1 d5d4 h2h4 g5g4 e2c2 c8d7 c2c3 d4c3 c1c3 a7a5 f2e3 d7c6');
INSERT INTO evaluations VALUES(27540,30,'cp -154',NULL);
INSERT INTO evaluations VALUES(27540,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(27540,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(27540,33,'cp -178',NULL);
INSERT INTO evaluations VALUES(27540,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(27540,35,'cp -180',NULL);
INSERT INTO evaluations VALUES(27540,36,'cp -186','f8b4 b1d2 g8e7 a2a4 e7f5 f1d3 f5e3 f2e3 e8c8 e1g1 f7f6 d3b5 c8b8 g1h1 h8g8 e5f6 g7f6 f3g1 e6e5 d2f3 a7a6 b5c6 d7c6 a1c1 b4a3 c1a1 a3d6 f3h4 e5d4 e3d4 b6c7 g2g3 c7e7 f1e1 e7f8 e1f1 c6d7 h4f5 d6a3 f1f2 d8c8 d1d2 a3b4 d2d3 c8c3');
INSERT INTO evaluations VALUES(27308,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(27308,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(27308,32,'cp -155',NULL);
INSERT INTO evaluations VALUES(27308,33,'cp -158',NULL);
INSERT INTO evaluations VALUES(27308,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(27308,35,'cp -166',NULL);
INSERT INTO evaluations VALUES(27308,36,'cp -164','b1d2 g8e7 a2a4 e7f5 f1d3 f5e3 f2e3 f7f6 d3b5 e8c8 e1g1 c8b8 d1e1 f6f5 b5c6 b6c6 e1e2 d8c8 f1d1 c6b6 d1c1 c8c7 c1c7 b6c7 f3e1 c7c3 a1a2 h8c8 e1d3 b4e7 a2b2 d7e8 d2b1 c3c6 b3b4 c6a4 d3c5 e7c5 b4c5 f5f4 b1d2 e8g6 d2b3 f4e3 e2e3 g6e4 h2h4');
INSERT INTO evaluations VALUES(27909,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(27909,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(27909,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(27909,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(27909,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(27909,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(27909,36,'cp 111','h2h4 h7h6 c1f4 g8e7 h4h5 g6g5 f3g5 h6g5 f4g5 g7h6 g5f6 h8g8 d1d3 b8c6 a1d1 d7d5 d3h7 h6g5 h7g8 e7g8 f6d8 a8d8 f1b5 a7a6 b5c6 b7c6 d1d3 g8h6 d3g3 g5f4 g3f3 f4g5 c3e2 c6b5 f3c3 c7c5 f2f4 g5e7 g2g4 e8f8 c3g3 b5e2 e1e2');
INSERT INTO evaluations VALUES(28419,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(28419,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(28419,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(28419,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(28419,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(28419,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(28419,36,'cp 116','e4e5 d6e5 c4f7 e8f8 f7b3 e5d4 d1d4 d8d4 f3d4 c7c5 d4b5 b8c6 h2h3 g4f5 c1e3 c6a5 e3c5 a5b3 c2b3 f6e4 c3e4 f5e4 e1g1 e4d3 b5c7 d3f1 c7a8 f8f7 g1f1 h8a8 a1b1 g7f6 c5e3 a7a6 f1e2 f7e6 e3d2 a8d8 d2c3 f6d4 f2f3 b7b5 b1c1 d4c3 c1c3');
INSERT INTO evaluations VALUES(31256,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(31256,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(31256,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(31256,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(31256,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(31256,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(31256,36,'cp 96','d2d4 b8d7 h2h3 g4h5 d4c5 d7c5 c1e3 c5d7 b1d2 e7e6 g2g4 h5g6 f1g2 f8e7 f3d4 h7h6 d1e2 g8f6 e1c1 a8c8 h3h4 d7e5 g4g5 g6h5 f2f3 f6d7 g5h6 g7g6 c1b1 a7a6 b1a1 d8c7 g2h3 b7b5');
INSERT INTO evaluations VALUES(31250,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(31250,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(31250,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(31250,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(31250,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(31250,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(31250,36,'cp 116','d8c7 d4d5 b8d7 h2h3 g4f3 d1f3 g7g6 f1e2 a7a6 e1g1 f8h6 c1h6 g8h6 f3e3 h6g8 a2a4 g8f6 f2f4 e8c8 a4a5 h8e8 b2b4 e7e6 b4b5 c5c4 b5a6 b7a6 d5e6 e8e6 b1d2 d6d5 e4e5 c7c5 e3d4 f6e8 d2f3 e8c7 f3g5 e6e7 g5h7 c5d4 c3d4');
INSERT INTO evaluations VALUES(31377,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(31377,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(31377,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(31377,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(31377,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(31377,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(31377,36,'cp 33','e1g1 g8h6 c1h6 g7h6 b1d2 b8d7 a2a3 f8g7 b2b4 e8g8 c2c4 a7a5 d1b3 d5c4 d2c4 a5b4 a3b4 a8a1 f1a1 d7b6 c4d6 d8d7 h2h4 b6c8 d6e4 c8e7 e4c5 d7c7 b4b5 c6b5 b3b5 f8b8 a1a7 g7f8 a7b7 b8b7 b5b7 c7b7 c5b7');
INSERT INTO evaluations VALUES(38113,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38113,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38113,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38113,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38113,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(38113,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(38113,36,'cp 15','e6d5 b1c3 g8f6 f1b5 b8c6 c1g5 e8g8 e1g1 c8e6 d1d3 a8c8 a1e1 h7h6 g5h4 c6b4 d3d2 e6f5 e1e5 f5e4 c3e4 d5e4 d2d8 f8d8 h4f6 g7f6 e5e4 b4c2 g2g3 c2d4 f3d4 d8d4 f1e1 d4e4 e1e4 c8d8 e4e2 a7a6 b5a4 g8g7 g1g2 a6a5 e2c2 c5d4 c2c7 d4b2 c7b7');
INSERT INTO evaluations VALUES(41068,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(41068,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41068,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(41068,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(41068,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41068,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41068,36,'cp 4','h7h6 g5h4 g7g6 e2e3 e8g8 g1f3 c8f5 f1d3 f5d3 c2d3 b8d7 f3e5 c7c6 e1g1 f6h5 h4e7 d7e5 d4e5 d8e7 f2f4 a8d8 c3e2 h5g7 a1d1 h6h5 a2a3 a7a5 f1f3 f8e8 h2h3 h5h4 d3c3 g7f5 e2d4 f5g3 c3a5 c6c5 d4b5');
INSERT INTO evaluations VALUES(41254,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(41254,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41254,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41254,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41254,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(41254,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(41254,36,'cp 0','c7c5 g1e2 b8c6 a1b1 e8g8 d2d3 a7a6 e1g1 c8g4 a2a3 a8b8 b2b4 b7b5 c4b5 a6b5 h2h3 g4e2 c3e2 d8b6 c1f4 f6d7 b4c5 b6c5 b1c1 c5b6 f4e3 c6d4 d1d2 d4b3 d2c2 b3d4');
INSERT INTO evaluations VALUES(40343,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40343,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40343,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40343,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40343,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40343,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40343,36,'cp 0','g1e2 b8c6 d2d3 a7a6 a2a3 e8g8 e1g1 c8g4 a1b1 a8b8 b2b4 b7b5 c4b5 a6b5 c1e3 f6d7 f2f3 g4e6 f3f4 c6d4 d1d2 d4b3 d2e1 b3d4');
INSERT INTO evaluations VALUES(42647,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(42647,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(42647,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(42647,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(42647,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(42647,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(42647,36,'cp -14','g1f3 b8c6 c2c4 d7d6 b1c3 e8g8 f1e2 f8e8 e1g1 a7a5 a2a3 c8d7 d1c2 c6d4 f3d4 c5d4 c1e3 c7c6 e2d1 d4b6 c2d2 h7h6 d1f3 d7e6 b2b4 f6d7 b4b5 a5a4 e3b6 d8b6 a1b1 d7c5 f3d1 b6a5 f2f4 e5f4 f1f4');
INSERT INTO evaluations VALUES(43312,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43312,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(43312,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(43312,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43312,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43312,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(43312,36,'cp 0','e4e5 d7d5 e5f6 d5c4 f6g7 f8g7 d2d3 c4d3 c1g5 d8d6 c3e4 d6d5 d1d3 d5d3 c2d3 b8c6 e1c1 e8g8 e4c5 e6e5 c5e4 h7h6 g5e3 f8d8 e3b6 d8d5 e4c3 d5d6 b6c5 d6d8 c5b6');
INSERT INTO evaluations VALUES(44300,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(44300,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(44300,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(44300,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(44300,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(44300,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(44300,36,'cp 96','b1c3 d7d5 e4e5 f6e4 f1d3 e4c3 b2c3 d8c7 c1d2 b8c6 g1h3 c6a5 e1g1 a5c4 d2c1 h7h6 h3f4 c8d7 d1h5 f8e7 a1b1 b7b6 g2g3 g7g6 h5f3 b6b5 h2h4 a7a5 f4g6 f7g6 d3g6 e8d8 f3f7 d7e8 f7g7 h8f8');
INSERT INTO evaluations VALUES(44299,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(44299,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(44299,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(44299,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(44299,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(44299,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(44299,36,'cp 103','d7d5 e4e5 f6e4 f1d3 e4c3 b2c3 d8c7 c1d2 b8c6 g1h3 c6a5 e1g1 a5c4 d2c1 h7h6 h3f4 c8d7 d1h5 f8e7 a1b1 b7b6 g2g3 g7g6 d3g6 f7g6 f4g6 h8h7 g6e7 e8e7 c1g5 h6g5 h5h7 e7d8 h7g8 d7e8 f2f4 c4d2 g8e6 c7c6 e6c6 e8c6 f4g5 d2b1 f1b1 d8e7');
INSERT INTO evaluations VALUES(46698,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(46698,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(46698,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(46698,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(46698,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(46698,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(46698,36,'cp 35','f8g7 e1g1 e8g8 c2c4 c8f5 h2h3 c7c6 c1f4 f8e8 g3g4 f5b1 a1b1 b8d7 d1b3 e7e5 f4h2 d8b6 f1d1 e5d4 f3d4 h7h5 g4h5 f6h5 e2e3 g7e5 h2e5 e8e5 a2a3 a8e8 b3b6 d7b6');
INSERT INTO evaluations VALUES(47376,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(47376,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(47376,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(47376,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(47376,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(47376,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(47376,36,'cp -133','c8f5 f2f3 e7e6 f3e4 f6e4 c3e4 f5e4 g1f3 c7c6 f1d3 d8a5 e3d2 a5d5 c2c4 e4f3 d1f3 d5f3 g2f3 b8a6 e1c1 a8d8 d2e3 g7g6 a2a3 f8g7 d3c2 a6c7 c1b1 e8g8 h2h4 b7b5 e3f4 d8d7 c2e4 b5c4');
INSERT INTO evaluations VALUES(47467,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(47467,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(47467,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(47467,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(47467,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(47467,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(47467,36,'cp 39','d2d4 e7e6 g1f3 c7c5 e2e4 d5b6 b1c3 f8e7 f1b5 c8d7 d4c5 e7c5 e1g1 e8g8 c1f4 d8e7 a2a3 f8d8 b5d3 a7a5 d1e2 d7c6 d3b5 b8d7 b5c6 b7c6 e4e5 h7h6 c3e4 b6d5 a1c1 a5a4 f1d1 c5b6 c1c6 d7c5 e2c2 c5e4');
INSERT INTO evaluations VALUES(47495,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(47495,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(47495,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(47495,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(47495,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(47495,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(47495,36,'cp 62','f1d3 c8f5 f3h4 f5c8 e1g1 e7e6 h4f3 b8d7 c2c4 f7f5 f3e1 f8e7 b1c3 c7c6 c3e2 a7a5 a2a4 e8g8 b2b3 e7d6 c1a3 d6a3 a1a3 d8e7 a3a2 d7f6 e1f3 c8d7 f3e5 d7e8 f2f3 e4d6 d1c1 d6f7 f3f4 c6c5 a2c2 d5c4 d3c4 g8h8 e5d3 c5d4');
INSERT INTO evaluations VALUES(47494,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(47494,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(47494,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(47494,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(47494,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(47494,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(47494,36,'cp 49','c8f5 f3h4 f5c8 e1g1 e7e6 h4f3 b8d7 c2c4 f7f5 f3d2 f8e7 b1c3 c7c6 f2f3 e4f6 c4d5 f6d5 f1e1 e8g8 e3e4 d5f4 d3c2 f4g6 d2b3 d7f6 b3c5 f5e4 c5e4 f6d5 c1d2 g8h8 c3d5 e6d5 e4c5 g6f4 d2f4');
INSERT INTO evaluations VALUES(47955,30,'cp -467',NULL);
INSERT INTO evaluations VALUES(47955,31,'cp -459',NULL);
INSERT INTO evaluations VALUES(47955,32,'cp -451',NULL);
INSERT INTO evaluations VALUES(47955,33,'cp -448',NULL);
INSERT INTO evaluations VALUES(47955,34,'cp -461',NULL);
INSERT INTO evaluations VALUES(47955,35,'cp -475',NULL);
INSERT INTO evaluations VALUES(47955,36,'cp -472','c8g4 f1e2 g4e2 d1e2 c7c5 d2d3 b8c6 f2f4 f8e7 e1g1 d8d7 f4f5 e8c8 c2c3 d8e8 c1f4 e7d8 b1a3 d8b6 g1h1 b6c7 a3b5 a7a6 b5c7 d7c7 a1e1 c5c4 d3c4 c6e5 b2b3 f6d7 b3b4 f7f6 f4e5 e8e5');
INSERT INTO evaluations VALUES(29978,30,'cp -444',NULL);
INSERT INTO evaluations VALUES(29978,31,'cp -438',NULL);
INSERT INTO evaluations VALUES(29978,32,'cp -455',NULL);
INSERT INTO evaluations VALUES(29978,33,'cp -459',NULL);
INSERT INTO evaluations VALUES(29978,34,'cp -454',NULL);
INSERT INTO evaluations VALUES(29978,35,'cp -458',NULL);
INSERT INTO evaluations VALUES(29978,36,'cp -456','f1b5 c7c6 b5e2 g4e2 d1e2 c6c5 d2d3 b8c6 f2f4 f8e7 e1g1 d8d7 c2c4 e8g8 f4f5 c6d4 e2d1 a8e8 a2a4 d7d8 b1c3 g8h8 a1b1 b7b6 g1h1 d4c6 c1f4 f6g8 f1f3 e7g5 f4g5 d8g5 f3g3 g5d8 d1g4 g7g6 c3d5 c6e5 g4f4 f7f6 f5g6 e5g6 f4f2 d8d7');
INSERT INTO evaluations VALUES(49857,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(49857,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(49857,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(49857,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49857,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(49857,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(49857,36,'cp 5','e4c5 b1c3 d7d6 g1f3 d6e5 f3e5 b8d7 e5f3 f8e7 e1g1 e8g8 c3d5 c5d3 d1d3 d7c5 d5e7 d8e7 d3c3 c5e6 f1e1 f7f6 f3d4 e7f7 d4e6 c8e6 a2a4 a7a6 h2h3 f8e8 c1f4 e6d5 f4g3 c7c6 e1e8 a8e8 a1e1 f7d7 e1e8 d7e8 c3a5 e8e4 a5d8 g8f7 d8c7 f7g6');
INSERT INTO evaluations VALUES(49802,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(49802,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(49802,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(49802,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(49802,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(49802,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(49802,36,'cp 9','b1c3 d7d6 g1f3 d6e5 f3e5 b8d7 e5f3 f8e7 e1g1 e8g8 c3d5 c5d3 d1d3 d7c5 d5e7 d8e7 d3c3 b7b6 f1e1 c5e6 f3g5 e7c5 c3c5 e6c5 c1e3 c5a6 c2c3 f7f6 g5e4 f8d8 f2f3 c8e6 a1d1 g8f7 g1f2 d8d1');
INSERT INTO evaluations VALUES(50118,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(50118,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(50118,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(50118,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(50118,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(50118,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(50118,36,'cp 93','g5f6 e7f6 c2c4 d7d5 e2e3 c8e6 c4d5 f8b4 b1c3 d8d5 g1e2 e6f7 a2a3 b4c3 e2c3 d5b3 d1b3 f7b3 a1c1 c7c6 f1d3 g7g6 h2h3 b8d7 e1d2 e8c8 c3e2 h8e8 h3h4 d7b6 h4h5 b6d5 h5g6 h7g6 h1h7 e8e7 h7e7 d5e7 e2f4');
INSERT INTO evaluations VALUES(50067,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(50067,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(50067,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(50067,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(50067,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(50067,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(50067,36,'cp 84','e7f6 e2e3 d7d5 c2c4 c8e6 c4d5 e6d5 b1c3 f8b4 g1e2 b8c6 a2a3 b4c3 b2c3 d5f7 e2f4 d8d6 f1d3 g7g6 e1g1 e8g8 h2h4 a8d8 d1c2 g8h8 a3a4 c6e7 c3c4 g6g5 f4e2 f5f4 h4g5 f6g5 d4d5 d6f6 c2c3 f6c3 e2c3 e7g6 f1b1 b7b6');
INSERT INTO evaluations VALUES(50022,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(50022,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(50022,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(50022,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(50022,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(50022,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(50022,36,'cp 90','e2e3 d7d5 c2c4 c8e6 c4d5 e6d5 b1c3 f8b4 g1e2 b8c6 d1a4 d5f7 e1c1 a7a6 d4d5 b7b5 c3b5 a6b5 a4b5 a8a2 b5c6 d8d7 c6d7 e8d7 c1b1 h8a8 e2d4 a2a1 b1c2 a1d1 f1b5 d7c8 h1d1 f7d5 d4f5 d5g2 f5g7 b4d6 f2f4 a8a5 b5d3 g2c6 g7f5 d6c5 d3c4 c5b4 f5d4 c6e4 c4d3');
INSERT INTO evaluations VALUES(50021,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(50021,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(50021,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(50021,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(50021,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(50021,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(50021,36,'cp 81','d7d5 c2c4 c8e6 c4d5 f8b4 b1c3 e6d5 g1e2 b8c6 a2a3 b4c3 b2c3 e8g8 e2f4 d5f7 f1d3 g7g6 e1g1 c6a5 h2h4 d8d7 d1e2 b7b6 c3c4 a8d8 e2c2 d7d6 f1d1 g8g7 a3a4 f8e8 c4c5 b6c5 d4c5 d6e5 a1b1 d8b8 d3e2 a5b3 c5c6');
INSERT INTO evaluations VALUES(50734,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(50734,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(50734,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(50734,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50734,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(50734,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(50734,36,'cp 49','d2d4 d6d5 e4d5 e6d5 c4b3 c5c4 b3c2 f8d6 e1g1 g8e7 c1g5 f7f6 g5h4 b7b5 b2b3 e8g8 f1e1 c8f5 h4g3 d8d7 g3d6 d7d6 c2f5 e7f5 b1d2 b8c6 b3c4 b5c4 d2f1 d6d7 d1c2 f8e8 e1e8 a8e8 c2a4 a6a5 h2h4 g8f7 h4h5 e8b8 a1e1 f5d6');
INSERT INTO evaluations VALUES(50744,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(50744,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(50744,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(50744,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(50744,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(50744,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50744,36,'cp 28','g8f6 c4d3 b7b5 e1g1 c8b7 b1d2 f8e7 a2a4 c5c4 d3c2 b8d7 f1e1 e8g8 d2f1 f8e8 f1g3 d8c7 h2h3 e7f8 c1g5 b7c6 d1e2 a8b8 a4b5 a6b5 f3h2 b8a8 e2f3 h7h6 g5d2 a8a1 e1a1 e8a8 a1a8 c6a8 h2g4 f6g4 f3g4');
INSERT INTO evaluations VALUES(52735,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(52735,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(52735,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(52735,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(52735,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(52735,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(52735,36,'cp -12','c8g4 e2e3 e7e6 f1e2 g8f6 e1g1 b8d7 d2d3 h7h6 c1b2 f8d6 c4d5 c6d5 h2h3 g4f3 e2f3 e8g8 g2g3 a8c8 a2a3 b7b5 b1c3 d8b6 b3b4 a7a5 b4a5 b6a5 c3e2 a5a8 d1d2 e6e5 f3g2');
INSERT INTO evaluations VALUES(6764,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6764,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(6764,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(6764,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(6764,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(6764,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(6764,36,'cp 47','g8f6 e1g1 e8g8 f1e1 d6c5 c3a4 c5b6 a4b6 a7b6 d2d4 d7d6 h2h3 h7h6 c4f1 c8d7 d4e5 c6e5 f3e5 d6e5 b2b3 d7c6 d1d8 f8d8 f2f3 f6h5 a2a4 h5f4 c1e3 g8f8 f1c4 f7f6 e1d1 f8e7 d1d8');
INSERT INTO evaluations VALUES(9489,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(9489,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(9489,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(9489,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(9489,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(9489,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(9489,36,'cp 89','d1d8 e8d8 f2f3 h8e8 g2g4 a7a5 c1d2 c8e6 c3e2 f6d7 e2f4 c5b4 h2h4 d7e5 f1e2 e6c4 b2b3 c4e2 e1e2 b4d2 a1d1 a5a4 d1d2 d8e7 h1d1 a4b3 a2b3 a8a2 e2e3 h7h6 f4e2 a2b2 e2g3 g7g6 e3f4 e7f6 g4g5 h6g5 h4g5 f6g7 g3e2 b2a2 f4e3 a2b2 f3f4');
INSERT INTO evaluations VALUES(2244,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(2244,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(2244,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(2244,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(2244,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(2244,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(2244,36,'cp 81','e8d8 f2f3 h8e8 h2h4 a7a5 a2a4 f6d7 g2g4 d7e5 e1e2 f7f6 c1e3 c5b4 e2f2 c8d7 c3b1 b7b6 b1d2 d8c8 d2c4 c8b7 c4e5 f6e5 f1d3 e8f8 f2e2 d7e6 h4h5 f8f7 a1f1 b4d6 b2b3 a8f8 g4g5 c6c5 d3b5 c7c6');
INSERT INTO evaluations VALUES(2354,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(2354,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(2354,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(2354,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(2354,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(2354,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(2354,36,'cp 93','f2f3 h8e8 h2h4 a7a5 h4h5 c8e6 c3e2 h7h6 e2f4 e6d7 f1c4 d8e7 c4e2 e7f8 f4d3 c5d6 e1f2 c6c5 c1e3 b7b6 a2a4 d7c6 b2b3 a8d8 g2g4 f6h7 d3f4 h7g5 a1d1 c5c4 e2c4 g5f3 f2f3 e8e4 c4d5 e4e3 f3e3 d6f4 e3f4');
INSERT INTO evaluations VALUES(11916,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(11916,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(11916,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(11916,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(11916,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(11916,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(11916,36,'cp 205','e4e5 f8c5 d4f4 d7d5 e5f6 d5c4 f6g7 h8g8 e1g1 c8e6 b1c3 g8g7 f4f3 g7g6 f1d1 d8c8 d1e1 e8f8 c3e2 c5d6 c1f4 e6g4 f4d6 c7d6 f3d5 g4e2 e1e2 a8b8 a1e1 c8c5 d5f3 d6d5 e2e5 b8d8 h2h4 c5d6 h4h5 g6f6 f3g4 h7h6 g4d4 d6c5 d4h4 c5c6 e1e3 c6d6 c2c3 b6b5 e3e1 d6c6 a2a3 c6b6 e1e2 b6d6 g2g3');
INSERT INTO evaluations VALUES(18835,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(18835,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(18835,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(18835,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(18835,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(18835,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(18835,36,'cp -37','b1c3 g8f6 g2g3 d7d5 e4d5 e6d5 e3g5 f8e7 f1g2 e8g8 g1e2 d5d4 g5f6 e7f6 c3e4 a8b8 e1g1 b7b6 f1e1 f6e5 e4c5 b6c5 g2c6 e5c7 c6g2 c7a5 e1f1 b8b2 d1c1 a5c3 a1b1 b2b1 c1b1 c3d2 b1b8 c8g4 b8d8 f8d8 e2f4 d2f4 g3f4 d8e8 h2h3 g4f5 f1b1 f7f6 b1b7 e8e2 b7a7 e2c2');
INSERT INTO evaluations VALUES(18821,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(18821,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(18821,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(18821,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(18821,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(18821,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(18821,36,'cp -17','g8f6 g2g3 d7d5 e4d5 e6d5 e3g5 d5d4 c3e4 f8e7 g5f6 e7f6 f1g2 e8g8 g1e2 b7b6 e1g1 c8d7 c2c4 d4c3 b2c3 a8c8 f1e1 h7h6 e4d6 c8b8 d6e4 d7g4 h2h3 g4f5 d3d4 c5d4 e4f6 d8f6 e2d4 f5d7 d4c6 d7c6 d1d4 f6g6 a1d1 b8c8 g2c6 g6c6');
INSERT INTO evaluations VALUES(19751,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19751,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(19751,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19751,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(19751,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19751,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19751,36,'cp 0','e2e4 e7e6 e4d5 e6d5 b1c3 g8f6 c3b5 f8c5 b5c7 e8f8 c7a8 f6e4 f4e3 d5d4 e3d2 e4f2 d1h5 d8e7 e1f2 d4d3 f2g3 c5d6 g3f2');
INSERT INTO evaluations VALUES(19924,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19924,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(19924,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19924,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(19924,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19924,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19924,36,'cp 0','c8d7 e1g1 a7a6 b5c6 d7c6 f1e1 g8f6 e4e5 d6e5 f3e5 d8d5 e5c6 d5c6 d2d3 e7e6 b2b3 f6d5 b1d2 f8e7 d2e4 e8g8 c2c4 d5c7 c1b2 f7f6 f2f4 a8d8 d1e2 d8d7 a1d1 f8d8 e2g4 c7e8 g4e2');
INSERT INTO evaluations VALUES(26054,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(26054,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(26054,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(26054,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(26054,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(26054,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26054,36,'cp 9','d2d4 c5d4 f3d4 g8f6 b1c3 g7g6 e1g1 f8g7 c1e3 e8g8 a2a4 a7a6 b5d3 a8c8 d1d2 c6d4 e3d4 d8a5 d2e3 a5b4 c3e2 d7c6 a4a5 f6d7 f1d1 g7d4 e3d4 b4d4 e2d4 d7e5 f2f3 g8g7 d3e2 c8c7 g1f2 e5d7 d4c6 c7c6 c2c3 h7h5 f3f4');
INSERT INTO evaluations VALUES(20096,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(20096,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(20096,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(20096,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(20096,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(20096,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(20096,36,'cp -14','f1b5 c8d7 g1f3 g8f6 c3e3 e7e6 b5d3 f8e7 h2h3 e8g8 e1g1 d8c7 a2a3 c6e5 f3e5 d6e5 c1d2 a7a5 b1c3 f8d8 e3e2 d7c6 d2e3 a5a4 f2f3 h7h6 e2f2 f6d7 f1d1 c7a5 d3f1 a8c8 g1h1 e7c5 c3a2 a5b6 e3c5 d7c5 d1d8 c8d8');
INSERT INTO evaluations VALUES(19912,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(19912,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(19912,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(19912,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(19912,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(19912,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(19912,36,'cp -13','c8d7 g1f3 g8f6 c3e3 e7e6 e1g1 f8e7 b5d3 a7a6 f1d1 d8c7 h2h3 h7h6 b1d2 c6b4 b2b3 b4d3 e3d3 a8c8 c2c4 e8g8 c1a3 f8e8 a1c1 c8d8 e4e5 d6e5 a3e7 e8e7 d3e3 e5e4 d2e4 f6e4 e3e4 e6e5 e4d5 e7e8');
INSERT INTO evaluations VALUES(26036,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(26036,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(26036,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(26036,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(26036,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(26036,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(26036,36,'cp -16','g1f3 g8f6 c3e3 e7e6 e1g1 f8e7 b5d3 d8c7 h2h3 e8g8 c1d2 f8d8 f1e1 c6e5 f3e5 d6e5 a2a4 a7a6 e3e2 d7c6 d2c3 b7b5 a4b5 a6b5 a1a8 c6a8 d3b5 f6e4 b5d3 e4c3 b1c3 g7g6 e2e5 c7e5 e1e5');
INSERT INTO evaluations VALUES(21597,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(21597,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(21597,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21597,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(21597,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(21597,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(21597,36,'cp 0','f8d6 f1d3 g8e7 e1g1 c8f5 f1e1 f5d3 d1d3 d8d7 b1d2 e8g8 d2f1 a8e8 c1d2 c6d8 b2b4 d7f5 d3e2 f5d7');
INSERT INTO evaluations VALUES(5900,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(5900,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(5900,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(5900,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(5900,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(5900,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5900,36,'cp 0','f1d3 g8e7 e1g1 c8f5 f1e1 e8g8 c1g5 d8d7 g5h4 a8e8 h4g3 f5d3 d1d3 e7f5 b1d2 e8e1 a1e1 c6e7 b2b3 c7c6 a2a4 e7g6 a4a5 f5g3 h2g3 f8e8 b3b4 e8e1 f3e1 g6f8 a5a6 b7b6 d2f3 g7g6 g1h1 f8e6 e1c2 h7h5 c2e3');
INSERT INTO evaluations VALUES(5901,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5901,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5901,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5901,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5901,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5901,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5901,36,'cp 6','g8e7 e1g1 c8f5 f1e1 e8g8 b2b3 d8d7 c1a3 a8e8 a3d6 d7d6 a2a4 d6f6 b3b4 h7h6 d1c2 f5d3 c2d3 f6f5 d3f5 e7f5 b1a3 f7f6 a3b5 e8e1 a1e1 f8c8 g2g4 a7a6 b5c7 f5d4 f3d4 c8c7 d4c6 c7c6 e1e7 c6c3 e7b7 d5d4 b7d7 c3c4 b4b5 a6b5');
INSERT INTO evaluations VALUES(8718,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8718,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8718,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(8718,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(8718,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(8718,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(8718,36,'cp 5','f3h4 c8e6 e1g1 d8d7 f1e1 e6g4 h4f3 e8g8 b1d2 a8e8 d2f1 g4f5 d3b5 a7a6 b5a4 b7b5 a4c2 h7h6 c1d2 e7g6 c2f5 d7f5 f1e3 f5h5 g2g3 c6d8 h2h4 f7f5 f3h2 h5d1 a1d1 c7c6 e3g2 e8e1 d1e1 f5f4');
INSERT INTO evaluations VALUES(24187,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24187,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24187,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24187,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24187,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24187,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24187,36,'cp 0','e8g8 a2a4 d7g4 h2h3 g4f3 d1f3 c6d4 f3d1 d4e6 g5h4 g7g6 b5c4 f6h5 h4g3 e7g5 a4a5 c7c6 g3h2 d8f6 c4e6 f7e6 d3d4 e5d4 e4e5 d6e5 c3e4 f6f5 e4d6 f5f6');
INSERT INTO evaluations VALUES(23749,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(23749,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(23749,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23749,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23749,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(23749,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(23749,36,'cp 18','a2a4 h7h6 g5f6 e7f6 c3d5 g7g6 b5c6 b7c6 d5f6 d8f6 d1e1 a7a5 b2b3 c6c5 f3d2 d7e6 f2f3 h6h5 f1f2 f6g5 g1h1 g5h6 e1e2 h6f4 a1f1 h5h4 h2h3 f7f5 f1g1 f8f7 e2e1 g8h7 g1f1 f7f6 d2b1 g6g5 b1c3');
INSERT INTO evaluations VALUES(25281,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(25281,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(25281,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(25281,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(25281,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(25281,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(25281,36,'cp 68','h2h3 g4d7 b5c6 d7c6 c2c4 f8e7 b1c3 e8g8 c1e3 f8e8 d1c2 e7f8 a2a4 h7h6 a4a5 a7a6 f1d1 d8e7 b2b4 c6d7 d1e1 d7e6 c3d5 e7d8 d3d4 e5d4 e3d4 f6d5 e4d5 e6d7 e1e8 d7e8 a1e1 e8d7 e1e3 a8c8 f3d2');
INSERT INTO evaluations VALUES(25280,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(25280,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(25280,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(25280,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(25280,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(25280,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(25280,36,'cp 56','g4d7 b5c6 d7c6 c2c4 f8e7 a2a4 h7h6 b1c3 e8g8 a4a5 f8e8 c1e3 e7f8 f1e1 c6d7 d3d4 e5d4 d1d4 c7c6 e3f4 d7e6 e1d1 d8e7 b2b3 e8d8 f4e3 a7a6 d4d3 a8c8 f3d4 d6d5 e4d5 c6d5 c4c5 c8c5 d4e6 e7e6 e3c5 f8c5 c3a4');
INSERT INTO evaluations VALUES(25309,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(25309,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(25309,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(25309,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(25309,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(25309,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(25309,36,'cp 90','d1b3 c6a5 c4f7 e8e7 b3a4 e7f7 a4a5 g4f3 g2f3 d8d7 c1e3 d6d5 e3g5 f8e7 b1d2 c7c6 f3f4 e7d8 a5b4 a7a5 b4c5 e5f4 g5f4 h8e8 e1c1 f7g8 f4g3 d5e4 d3e4 d7f7 e4e5 f6d7 c5d4 d7b6 c3c4 d8c7 h1g1 a8d8 d4c3 b6a4 c3c2 a4c5 d2e4 c5e4');
INSERT INTO evaluations VALUES(25513,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(25513,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(25513,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(25513,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(25513,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(25513,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(25513,36,'cp 29','e1g1 a7a6 b5a4 h7h6 a4c2 c6e7 b1d2 e7g6 d3d4 b7b5 h2h3 f8e7 f1e1 e8g8 d2f1 c7c5 f1g3 d8c7 g3f5 a6a5 f5e7 g6e7 c1e3 e7g6 f3d2 a5a4 a1c1 c5c4 a2a3 d7c6 d4d5 c6d7 g1h2 g8h8 c1b1 a8e8 d1f3 f8g8 e1g1');
INSERT INTO evaluations VALUES(25970,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(25970,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(25970,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(25970,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(25970,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(25970,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(25970,36,'cp 89','d2d4 g8f6 c1g5 f8e7 d4c5 e8g8 b1c3 f6e4 g5e7 c6e7 b5d7 e4c3 b2c3 d8d7 f1e1 a8c8 d1d4 f8d8 a1b1 h7h6 f3e5 d7c7 g2g3 e7c6 e5c6 c7c6 e1e7 b7b6 c5b6 a7b6 b1b6 c6c3 d4c3 c8c3 a2a4 d5d4 b6b7 c3c2 e7f7 g7g5 g1g2');
INSERT INTO evaluations VALUES(28205,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(28205,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(28205,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(28205,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(28205,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28205,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(28205,36,'cp 43','c4d5 f6d5 g5e7 d8e7 c3d5 e6d5 d1a4 c7c6 f1e2 e8g8 e1g1 f8c8 a1c1 b8d7 f1d1 a7a5 f3e1 c8c7 c1c2 c6c5 d4c5 b6c5 a4a3 d7f6 e2f3 a5a4 e1d3 c5c4 a3e7 c7e7 d3b4 e7d7 c2d2 a4a3 b2a3 a8a3 b4d5 d7d5 f3d5');
INSERT INTO evaluations VALUES(28204,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(28204,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(28204,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(28204,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(28204,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(28204,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(28204,36,'cp 38','f6d5 g5e7 d8e7 c3d5 b7d5 d1a4 b8d7 a1c1 c7c6 f1d3 f7f5 e1g1 e8g8 f1e1 h7h6 a4d1 c6c5 e3e4 f5e4 d3e4 d5e4 e1e4 d7f6 e4e1 a8d8 d1b3 f6d5 d4c5 b6c5 c1c2 f8f5 g2g3 d5b4 c2d2 f5d5 d2d5 d8d5 g1g2');
INSERT INTO evaluations VALUES(31314,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31314,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31314,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(31314,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(31314,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31314,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31314,36,'cp 0','f5d3 d1d3 e7e6 g1f3 b8c6 c3e2 g8f6 e1g1 h7h6 c2c3 f8e7 f1e1 e8g8 a2a4 f6e4 c1f4 d8b6 a1b1 f8c8 e2g3 e4f6 h2h3 a7a5 d3d2 e7f8 d2e3 b6a6 f3e5 c6e5 f4e5 f6d7 e5f4 a6c6');
INSERT INTO evaluations VALUES(31348,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(31348,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31348,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31348,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31348,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(31348,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31348,36,'cp 0','d1d3 e7e6 g1f3 b8c6 c3e2 g8f6 e1g1 h7h6 c2c3 f8e7 f1e1 e8g8 a2a4 f8e8 c1f4 d8b6 h2h3 f6e4 a1b1 b6b3 b1a1 b3b6');
INSERT INTO evaluations VALUES(31347,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(31347,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(31347,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(31347,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(31347,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31347,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31347,36,'cp 0','e7e6 g1f3 b8c6 c3e2 g8f6 e1g1 h7h6 c1f4 f8e7 c2c3 d8b6 d3c2 a8c8 h2h3 f6e4 a2a3 c6a5 c2a4 a5c6');
INSERT INTO evaluations VALUES(31719,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31719,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31719,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31719,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31719,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31719,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31719,36,'cp 0','g1f3 b8c6 c3e2 g8f6 e1g1 h7h6 c2c3 f8e7 f1e1 e8g8 c1f4 d8b6 a1b1 a7a5 f3d2 b6a6 d3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(31718,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31718,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31718,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31718,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31718,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31718,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31718,36,'cp 0','b8c6 c3e2 g8f6 e1g1 d8b6 c2c3 f8d6 a2a4 b6c7 g2g3 f6h5 f3e1 c6a5 d3b5 a5c6');
INSERT INTO evaluations VALUES(26428,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26428,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26428,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26428,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26428,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26428,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26428,36,'cp 0','c3e2 g8f6 e1g1 d8b6 a2a4 c6b4 d3b3 b4c6');
INSERT INTO evaluations VALUES(31528,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(31528,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31528,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31528,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(31528,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(31528,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(31528,36,'cp 12','g8e7 g1f3 f5d3 d1d3 b8d7 h2h4 c6c5 e1f1 f7f6 e5f6 g7f6 c1f4 d8b6 b1d2 e8c8 a1e1 d8e8 d3c2 e7c6 h1h3 e6e5 d4e5 f6e5 f4g3 f8d6 f3g5 d6c7 c3c4 c6b4 c2f5 b6f6 f5f6 d7f6 c4d5 f6d5 g5f7 b4d3 f7h8 d3e1 f1e1 e8h8');
INSERT INTO evaluations VALUES(34857,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(34857,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(34857,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(34857,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(34857,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(34857,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(34857,36,'cp 142','b8d7 g1f3 g8f6 f1d3 f8b4 e1g1 e8g8 f1e1 a5h5 c3e4 b4d2 d1d2 h5h6 d2h6 g7h6 f3e5 b7b6 f2f4 c8b7 e5d7 f6d7 f4f5 e6f5 e4g3 f5f4 g3f5 d7f6 e1f1 f4f3 f5h6 g8h8 a1e1 f6g8 h6g8 f3g2 f1f5 h8g8 e1e7 a8e8 f5g5 g8h8 g5e5 e8e7 e5e7 a7a5 e7c7 h8g7');
INSERT INTO evaluations VALUES(35802,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(35802,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(35802,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35802,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(35802,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(35802,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(35802,36,'cp 37','c8e6 d1e2 e6c4 e2c4 a7a6 c1f4 b7b5 c4d3 b8d7 f1d1 c5c4 d3d2 d7c5 d2e2 d8b6 f3d4 g7g6 f4g5 c5d7 a2a4 b5b4 g5f6 d7f6 c3d5 f6d5 e4d5 e7f6 e2c4 a6a5 a1b1 f6d4 c4d4 b6d4 d1d4 a8c8 b1c1 g8g7 g1f1 f8e8 c2c4 b4c3 c1c3 c8c3 b2c3 e8e5');
INSERT INTO evaluations VALUES(37404,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(37404,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(37404,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(37404,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(37404,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(37404,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(37404,36,'cp -17','d7d5 h2h3 c7c5 g1f3 c5d4 e3d4 b8c6 e1g1 b7b6 c4d5 c6b4 f1e1 c8b7 c1g5 a8c8 a2a3 b4d3 d1d3 b7d5 a1d1 h7h6 g5h4 e7e6 f3e5 g6g5 c3d5 f6d5 h4g3 d5e7 e5c4 d8d5 c4e3 d5d7 d3e2 e7d5 e3d5 d7d5 g3e5 f8e8 e5g7 g8g7');
INSERT INTO evaluations VALUES(36734,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(36734,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(36734,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(36734,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(36734,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(36734,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(36734,36,'cp -20','g1f3 c7c5 e1g1 c5d4 e3d4 b8c6 h2h3 b7b6 c4d5 c6b4 f1e1 c8b7 c1g5 h7h6 g5f6 e7f6 d3c4 b4d5 c3d5 b7d5 c4d5 d8d5 d1b3 d5b3 a2b3 f8d8 e1e7 f6f5 a1a7 a8a7 e7a7 g7d4 a7b7 d4c5 g2g3 d8d1 g1g2 d1b1 b3b4 c5b4 b7b6 b1b2 f3d4 b4c3 b6b2 c3b2 d4c2 g8g7 c2e3');
INSERT INTO evaluations VALUES(44410,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(44410,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(44410,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(44410,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44410,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(44410,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(44410,36,'cp 56','e2e4 d7d6 d2d4 c5d4 f3d4 b8c6 f1e2 c6d4 d1d4 f8g7 c1e3 e8g8 d4d3 c8d7 e1g1 a7a6 a2a4 a6a5 a1b1 d7c6 f2f3 f6d7 d3d2 d7c5 b2b3 a8c8 f1d1 b7b6 c3d5 c5d7 g2g3 f8e8 d5c3 d7c5 e2f1 d8c7 d1c1 e8d8');
INSERT INTO evaluations VALUES(46203,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46203,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46203,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46203,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46203,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46203,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46203,36,'cp 5','g1f3 d8b6 d1c1 c8d7 c2c3 f8d6 f4d6 b6d6 b1d2 e8g8 f1d3 b8c6 d4c5 d6c5 e1g1 e6e5 e3e4 d7e6 b2b4 c5d6 f1e1 a8d8 b4b5 c6a5 c1c2 f8e8 a1d1 d6c7 f3g5 g7g6 e4d5 e6d5 g5e4 d5e4 d2e4 f6e4 e1e4');
INSERT INTO evaluations VALUES(47186,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(47186,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(47186,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(47186,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(47186,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(47186,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47186,36,'cp -37','c8f5 e2e3 e7e6 g1f3 h7h6 b2b4 a7a5 b4a5 a8a5 f1e2 b8d7 c3c4 d8a8 a2a4 f8e7 e1g1 b7b6 c1b2 e8g8 d1b3 c7c6 f1c1 a5a7 h2h3 f8b8 b3d1 b8c8 b2a3 c6c5 a3b2 f5h7 b2c3 c8d8 c3b2 f6e4 c4d5 e4d2 d1d2');
INSERT INTO evaluations VALUES(29575,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(29575,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(29575,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(29575,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(29575,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(29575,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(29575,36,'cp -30','g1f3 h7h6 e2e3 e7e6 b2b4 a7a5 b4a5 f8e7 a2a4 b8d7 f1e2 a8a5 e1g1 e8g8 c3c4 b7b6 c1b2 d8a8 d1b3 c7c6 f1c1 a5a7 h2h3 f8c8 b3d1 a8b7 d1b3 c8a8 b2c3 b7c7 c3b2 f5h7 c4d5 e6d5 b3c3 c6c5 d4c5 b6c5');
INSERT INTO evaluations VALUES(29573,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(29573,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(29573,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(29573,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(29573,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(29573,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(29573,36,'cp -41','h7h6 d1b3 d8c8 c3c4 e7e6 e2e3 a7a5 f1d3 a5a4 b3c2 f5d3 c2d3 f8e7 b2b3 e8g8 c1b2 f8d8 e1g1 b7b6 d3c2 c7c5 a1c1 a4b3 a2b3 b8c6 b2c3 c8b7 c1a1 c5d4 a1a8 d8a8 f3d4 c6b4 c2b1 e7c5 d4b5 b7e7');
INSERT INTO evaluations VALUES(51026,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(51026,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(51026,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(51026,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(51026,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(51026,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(51026,36,'cp 70','e2e4 c7c5 g1f3 b8c6 d2d4 c5d4 f3d4 f8c5 d4c6 b7c6 c1f4 g8e7 d1d2 d7d5 a1d1 e8g8 f1e2 e7g6 e4d5 g6f4 d2f4 c6d5 e1g1 d8d6 f4d6 c5d6 c4d5 f8d8 e2c4 a8b8 d5e6 c8e6 c4e6 f7e6 f1e1 g8f7 c3a4 d6e7 d1d8 b8d8');
INSERT INTO evaluations VALUES(51812,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(51812,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(51812,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(51812,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(51812,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(51812,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(51812,36,'cp 32','b1c3 c6c5 g1f3 a7a6 c1f4 d7d6 g2g3 d8c7 f1g2 b8c6 e1g1 h7h6 a2a4 g8f6 f1e1 f6d7 d3d4 c5d4 f3d4 c6e5 f4e3 f8e7 c3e2 e8g8 b2b3 e5g4 e3d2 d7e5 e2f4 c8d7 c2c3 g4f6');
INSERT INTO evaluations VALUES(51810,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51810,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(51810,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(51810,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(51810,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(51810,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(51810,36,'cp 32','d7d5 d3d4 g8f6 e4e5 f6d7 f2f4 c6c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 d8e7 f1e2 e8g8 e1g1 c6d4 e3d4 c5d4 d2d4 e7c5 a1d1 b7b5 d4c5 d7c5 g1f2 a8b8 b2b4 c5d7 f1e1 f7f6 e5f6 f8f6 e2f3 d7b6 c3e2 b6c4 e2d4 c4a3 d4c6');
INSERT INTO evaluations VALUES(53621,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(53621,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(53621,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53621,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(53621,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(53621,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(53621,36,'cp 32','d3d4 g8f6 e4e5 f6d7 f2f4 c6c5 g1f3 b8c6 c1e3 a7a6 a2a3 c5d4 f3d4 f8c5 d1d2 c5d4 e3d4 c6d4 d2d4 d8b6 c3e2 e8e7 e1c1 b6d4 e2d4 d7c5 g2g3 c8d7 d1e1 a6a5 e1e3 g7g6 f1e2 h7h5 h1d1 h8c8 e3c3 c5e4 c3c8 a8c8 c1b1');
INSERT INTO evaluations VALUES(53241,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(53241,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(53241,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(53241,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(53241,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(53241,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(53241,36,'cp 100','c4d5 d8d5 b1c3 d5d8 d2d4 c5d4 d1d4 c8d7 e2e4 b8c6 d4e3 e7e6 f1e2 f8b4 e1g1 g8f6 a2a3 b4c3 e3c3 e8g8 c3e3 f6g4 e3d2 g4f6 f1d1 e6e5 d2e3 d8e7 h2h3 h7h6 b2b4 d7g4 c1b2 g4f3 e3f3 f8d8 b4b5');
INSERT INTO evaluations VALUES(53234,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(53234,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(53234,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(53234,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(53234,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(53234,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(53234,36,'cp 94','d8d5 b1c3 d5d8 e2e4 a7a6 d2d4 c5d4 d1d4 d8d4 f3d4 e7e6 d4b3 g8f6 f2f3 b8c6 c1e3 e6e5 f1c4 b7b5 c4d5 c8d7 d5c6 d7c6 a1c1 f8e7 c3e2 c6d7 b3c5 e8g8 c5d3 e7d6 e3c5 d6c5 c1c5 f8c8 e1f2 d7e6 c5c8 a8c8 d3e5 e6a2 h1c1 c8c1');
INSERT INTO evaluations VALUES(34370,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(34370,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(34370,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(34370,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(34370,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(34370,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(34370,36,'cp 95','b1c3 d5d8 e2e4 a7a6 d2d4 c5d4 d1d4 d8d4 f3d4 e7e6 d4b3 g8f6 f2f3 e6e5 f1c4 b8c6 c1e3 b7b5 c4d5 c8d7 a1c1 a8c8 e1e2 h7h5 h1d1 h5h4 a2a3 c6d8 b3a5 b5b4 a3b4 f8b4 a5c4 f6d5 c3d5 d7b5 b2b3 b4a3 c1a1 a3f8 e2f2 f7f6 a1c1');
INSERT INTO evaluations VALUES(34364,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(34364,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(34364,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(34364,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(34364,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(34364,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(34364,36,'cp 90','d5d8 e2e4 a7a6 d2d4 c5d4 d1d4 d8d4 f3d4 e7e6 d4b3 g8f6 f2f3 e6e5 f1c4 b8c6 c1e3 b7b5 c4d5 c8d7 a1c1 a8c8 e1e2 h7h5 h1d1 c6d8 b3a5 h5h4 a2a3 b5b4 a3b4 f8b4 a5c4 f6d5 c3d5 d7b5 b2b3 b4a3 c1c3 d8e6 e2e1 b5c4 c3c4 c8c4 b3c4');
INSERT INTO evaluations VALUES(53933,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(53933,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(53933,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(53933,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(53933,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(53933,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(53933,36,'cp 149','g1e2 g8f6 e2c3 f8e7 f1b5 b8d7 a2a4 e8g8 b1a3 f6e8 b5d3 e7g5 c1g5 d8g5 d1f3 a8b8 h2h4 g5f4 f3e2 a7a6 g2g3 f4f6 g3g4 f6f4 g4g5 d7b6 a4a5 b6d7 e2e3 f4e3 f2e3 b7b5 a5b6 d7b6 b2b3 e8c7 e1d2 f7f5 a1f1 f5f4 e3f4 f8f4 f1f4');
INSERT INTO evaluations VALUES(56301,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(56301,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(56301,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(56301,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(56301,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(56301,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(56301,36,'cp 176','b8d7 c1f4 d7e5 f4e5 g8f6 e2e3 f8d6 b1c3 c7c6 e5d6 d8d6 f1d3 c8g4 d1d2 e8g8 e1g1 a8e8 h2h3 g4h5 c3e2 h5e2 d2e2 g7g6 a2a3 h7h5 c2c4 g8g7 a1c1 e8e7 c4d5 f6d5 d3c4 d5f4 e2d2 f4d5 c4d5 c6d5 c1c5 b7b6 c5c1 f8e8 c1c2');
INSERT INTO evaluations VALUES(56447,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(56447,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(56447,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(56447,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(56447,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(56447,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(56447,36,'cp -37','f1g2 g8f6 d2d4 e5d4 c3d4 f8d6 b1c3 e8g8 g1f3 c7c6 e1g1 f8e8 d1c2 b8d7 f1e1 a7a5 a2a3 h7h6 c1f4 d6f4 g3f4 d7b6 e2e3 g7g6 f3e5 c8f5 c2e2 f6h5 g2f3 h5g7 f3g4 b6c8 g1h1 c8d6 e1g1 d8e7 a1c1 g8h7 g4f5 g7f5 e2c2 h7g7 c3e2 e7h4');
INSERT INTO evaluations VALUES(566,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(566,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(566,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(566,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(566,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(566,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(566,36,'cp 10','d7d6 e1g1 e8g8 f1e1 c6e5 b5e2 f6g6 b1d2 e5g4 e2g4 c8g4 f2f3 g4d7 d2f1 a8e8 d4e2 g6g5 e3c5 g5c5 d1d4 c5d4 e2d4 f7f5 e4f5 e7f5 e1e8 f8e8 d4f5 d7f5 g1f2 e8e5 a2a3 a7a5 h2h4 f5e6 a1d1 g8f7 d1d2 b7b6 f1e3 h7h5 g2g4 h5g4 f3g4');
INSERT INTO evaluations VALUES(3301,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(3301,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(3301,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(3301,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(3301,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(3301,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(3301,36,'cp -15','c6a5 c4b3 a7a6 f1e1 b7b5 e3d2 c7c6 c3e2 f8e8 e2g3 e7f8 c2c3 a5b3 a2b3 a6a5 d3d4 c8d7 d1c2 e5d4 c3d4 d8b6 a1d1 d7e6 d1a1 d6d5 e4e5 f6e4 d2e3 e4g3 f2g3 b6a6 g3g4 a5a4 e1f1 a4b3 c2b3 a6a1');
INSERT INTO evaluations VALUES(3938,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(3938,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(3938,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(3938,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(3938,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(3938,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(3938,36,'cp 54','f6e8 d3c2 a7a5 d1e2 d7c5 c1g5 d8d7 a2a3 f7f5 e4f5 g6f5 a1e1 a5a4 g5e3 b7b6 g1h1 d7e7 e2d1 c8d7 g2g3 g7f6 e3h6 f6g7 h6d2 e7f6 d2e3 g7h6 e3c5 b6c5 c2a4 d7a4 c3a4 e5e4 f3h4 f5f4 e1e4');
INSERT INTO evaluations VALUES(4080,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4080,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4080,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(4080,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(4080,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4080,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4080,36,'cp 8','a2a4 b7b6 b1d2 a7a6 f4h2 c8b7 e1g1 f8e8 c2c3 e7e5 d1b3 d8e7 b3a3 f6e4 d2e4 b7e4 f1d1 a6a5 a1c1 e4b7 b2b4 b7c6 b4b5 c6b7 a3b3 h7h6 e2c4 a8d8 f3d2 d8c8 h2g3 h6h5 d1e1 e7f6 c4e2 g7h6 e2f3 e5e4 f3e2');
INSERT INTO evaluations VALUES(10690,30,'cp -161',NULL);
INSERT INTO evaluations VALUES(10690,31,'cp -165',NULL);
INSERT INTO evaluations VALUES(10690,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(10690,33,'cp -158',NULL);
INSERT INTO evaluations VALUES(10690,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(10690,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(10690,36,'cp -175','d7d5 e1g1 b4c3 b2c3 e8g8 c4d3 e4c3 e2b2 c3e4 c1f4 c6e7 h2h3 a7a5 f1e1 a5a4 e1e2 a4a3 b2c2 c7c6 c2b3 e7g6 f4h2 c6c5 d4c5 e4c5 b3c3 c5d3 c3d3 c8d7 d3d5 d7c6 d5d8 f8d8 f3e5 g6e5 h2e5 a8a4 e5c3 d8d5 e2d2 d5d2 c3d2 h7h6 a1c1 c6d5 c1c2 b7b5');
INSERT INTO evaluations VALUES(9186,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(9186,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(9186,32,'cp -174',NULL);
INSERT INTO evaluations VALUES(9186,33,'cp -170',NULL);
INSERT INTO evaluations VALUES(9186,34,'cp -178',NULL);
INSERT INTO evaluations VALUES(9186,35,'cp -172',NULL);
INSERT INTO evaluations VALUES(9186,36,'cp -177','e1g1 b4c3 b2c3 e8g8 c4d3 e4c3 e2b2 c3e4 c1f4 c6e7 h2h3 a7a5 f1e1 a5a4 e1e2 a4a3 b2c2 c7c6 a1b1 e7g6 f4h2 f8e8 f3e5 d8f6 e2e1 g6f8 d3e4 d5e4 h2g3 h7h6 e5c4 f6d4 c4b6 a8a5 b6c8 e8c8 b1b7 d4d3 c2b1 d3b1');
INSERT INTO evaluations VALUES(11119,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(11119,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(11119,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(11119,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(11119,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(11119,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(11119,36,'cp 66','b5c6 d7c6 e1g1 c8g4 h2h3 g4f3 d1f3 e7e5 d2d3 f6d7 f3g3 g7g6 a2a4 f8d6 f2f4 f7f6 f4f5 g6g5 c1e3 d8a5 c3d1 e8c8 b2b3 d8g8 g3f2 h7h5 d1b2 a5d8 g2g3 d6c7 g1g2 c8b8 b2c4 d8e7 a4a5 g5g4');
INSERT INTO evaluations VALUES(11117,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(11117,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(11117,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(11117,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(11117,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(11117,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(11117,36,'cp 72','d7c6 e1g1 c8g4 h2h3 g4f3 d1f3 f6d7 d2d3 e7e5 f3g3 g7g6 f2f4 f8d6 f4f5 d8e7 g3f3 e8c8 f5g6 f7g6 f3g3 e7g7 c1e3 h8f8 f1f8 d7f8 g3g4 c8c7 a1f1 h7h5 g4f3 f8e6 f3f7 g7f7 f1f7 d8d7 f7f6 e6d4 f6g6 d4c2 g1f2 c2e3 f2e3');
INSERT INTO evaluations VALUES(11189,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(11189,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(11189,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(11189,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(11189,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(11189,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(11189,36,'cp 73','e1g1 c8g4 h2h3 g4f3 d1f3 e7e5 d2d3 f6d7 f3g3 g7g6 f2f4 f8d6 f4f5 d8e7 c1e3 h8g8 g3f3 e8c8 b2b3 d6c7 c3e2 f7f6 g1h1 c7a5 g2g4 g6g5 h1g2 g8h8 f1h1 d8g8 e2g3 a5c3 a1b1 h7h6 f3e2 b7b5 e3d2');
INSERT INTO evaluations VALUES(14758,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(14758,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(14758,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(14758,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14758,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(14758,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(14758,36,'cp 64','b1c3 f8e7 e2e3 e8g8 a2a3 h7h6 g5f4 e7d6 f4g3 b7b6 g3h4 d6e7 a1c1 c8b7 h4f6 e7f6 c4d5 e6d5 b2b4 c6e7 f1d3 c7c6 e1g1 d8d6 d1b3 b7c8 h2h3 f8d8 c3e2 a7a5 c1c2 c8d7 f1c1 a5b4 a3b4 g7g6');
INSERT INTO evaluations VALUES(15062,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(15062,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(15062,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(15062,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15062,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(15062,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15062,36,'cp -1','d2d3 c8f5 g1e2 e4d3 c2d3 f8b4 b1d2 f5d3 e2f4 d3e4 e1g1 e4g2 g1g2 d5d4 e3d4 f6d5 f1e1 c6e7 f4d3 e8g8 d3b4 d5b4 b2a3 e7d5 d2c4 a7a5 d1f3 c7c6 a1d1 f8e8 e1e8 d8e8 d1d2 e8d7 d2e2 f7f6 h2h4 h7h6 e2d2 g8h8 a3b4 a5b4 c4e3 a8d8 e3d5 d7d5 h4h5 d5d7 d2d1 d8e8');
INSERT INTO evaluations VALUES(16472,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(16472,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(16472,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(16472,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(16472,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(16472,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(16472,36,'cp 53','d4d5 f8b4 b1d2 c6e7 g1f3 e7g6 h2h4 e5e4 h4h5 e4f3 h5g6 f7g6 g2f3 e8g8 a2a3 b4e7 b2b4 d7d6 d2b3 f6h5 b3d4 e7f6 f1e2 c8d7 b4b5 a7a6 c1b2 d8e7 a3a4 e7f7 f3f4 a6b5 a4b5 d7f5 e2h5 f6d4 a1a8 f8a8 b2d4 g6h5 d1h5 a8a2 h5f7 g8f7 e1g1');
INSERT INTO evaluations VALUES(16770,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(16770,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(16770,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(16770,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(16770,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(16770,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(16770,36,'cp 81','e4c5 b1c3 d7d6 f1b5 a7a6 b5c6 b7c6 d5c6 c8d7 c6d5 c7c6 d5d2 d6d5 e1g1 f8e7 f1e1 d8c7 b2b3 c5e6 f3d4 e6d4 d2d4 e8g8 c3a4 d7f5 c2c3 h7h6 c1f4 c7b7 a4c5 b7b5 b3b4 e7c5 b4c5 b5c4 d4c4 d5c4');
INSERT INTO evaluations VALUES(19346,30,'cp -490',NULL);
INSERT INTO evaluations VALUES(19346,31,'cp -494',NULL);
INSERT INTO evaluations VALUES(19346,32,'cp -502',NULL);
INSERT INTO evaluations VALUES(19346,33,'cp -513',NULL);
INSERT INTO evaluations VALUES(19346,34,'cp -498',NULL);
INSERT INTO evaluations VALUES(19346,35,'cp -524',NULL);
INSERT INTO evaluations VALUES(19346,36,'cp -517','d8a5 c2c3 a5b5 a2a4 b5a5 b2b4 a5h5 b1a3 d4c3 a3b5 a8b8 e1g1 d7d6 a1a3 c8g4 a3c3 b8c8 d1c2 e7e5 b5a3 e8d7 a4a5 g4f3 c3f3 c6d4 c2a4 d7d8 f3h3 h5g4 c1e3 g8f6 f2f3 g4d7 b4b5 d6d5 f1d1 f8a3 a4a3 d7b5 h3h4');
INSERT INTO evaluations VALUES(21471,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(21471,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(21471,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(21471,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(21471,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(21471,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(21471,36,'cp -78','d5d4 c3e2 d4d3 e2c3 c6b4 c2d3 b4d3 f1d3 d8d3 d1e2 d3e2 e1e2 b7b6 d2d3 c8a6 c1e3 e8c8 h1d1 h7h6 h2h4 g8e7 a2a3 e7d5 a1c1 d8d7 b2b4 f8e7 f3d4 d5f4 e3f4 d7d4 g2g3 h8d8 b4b5 a6b7 f4e3 d4g4 a3a4 g4b4 c1c2 g7g5 h4h5 f7f5 e5f6 e7f6 c3e4 b7e4 d3e4');
INSERT INTO evaluations VALUES(22869,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(22869,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(22869,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(22869,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(22869,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(22869,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(22869,36,'cp -50','d2d4 d7d5 d4e5 c6e5 b1d2 c8g4 g1f3 g8f6 f1g2 f8d6 e1g1 e8g8 c2c4 e5c4 d2c4 d5c4 d1c2 f8e8 h2h3 g4h5 c2c4 d8d7 c4e2 d7e6 f1d1 a8d8 b2b3 h7h6 c1b2 e6e7 b2f6 e7f6 a1b1 a7a5 d1d2 c7c6 b1d1 a5a4 g3g4 h5g6 e2c4 a4b3 c4b3');
INSERT INTO evaluations VALUES(22905,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22905,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(22905,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22905,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22905,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22905,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22905,36,'cp 0','g8f6 a2a3 f8e7 g1f3 d7d5 c1e3 c8e6 b1d2 e8g8 f1d3 d5c4 d2c4 d8d5 e1g1 f8d8 d1c2 a7a5 a1d1 d5h5 f1e1 a5a4 d3e2 h5g6 e2d3');
INSERT INTO evaluations VALUES(24852,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(24852,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(24852,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(24852,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(24852,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(24852,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(24852,36,'cp -54','c6b4 e1g1 f5c2 d1d2 c2g6 a2a3 b4c6 f3e5 e7e6 f1d1 f8e7 e5g6 h7g6 e2e4 d5e4 d2c2 e8g8 d4d5 f6d5 g2e4 e7f6 c3d5 e6d5 d1d5 d8e8 c1d2 c6d4 c2d3 d4e6 d2e3 b7b6 h2h4 a8d8 d3c2 c7c5 a1d1 f6d4 d5d8 e8d8 g1g2 d8e7');
INSERT INTO evaluations VALUES(25682,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(25682,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(25682,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(25682,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(25682,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(25682,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(25682,36,'cp 30','d4e5 d8d1 e1d1 e8c8 d1e2 d8e8 c1f4 f7f6 h2h3 f6e5 f4g5 g4h5 g2g4 h7h6 g5h4 h5f7 b1d2 g8f6 f3e1 f8d6 f2f3 f6d7 h4f2 c6c5 b2b3 d7f8 h3h4 f8e6 e1d3 e6f4 d3f4 e5f4 h4h5 d6e5 a1d1 b7b6 d2c4 e5f6 f2h4 f6d4 d1d3 f7c4 b3c4 b6b5 c4b5 a6b5 h1b1 b5b4 h4f2');
INSERT INTO evaluations VALUES(26413,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(26413,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(26413,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(26413,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26413,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26413,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26413,36,'cp 0','f1b5 a8c8 b1d2 d8b6 a2a4 a7a6 b5c6 b6c6 e1g1 f5g6 f1e1 g8e7 d4c5 c6c5 f3d4 c5c7 h2h4 h7h5 d2f3 e7c6 c1g5 f8c5 d4c6 b7c6 b2b4 c5a7 d1e2 c7b7 e2d1');
INSERT INTO evaluations VALUES(26387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26387,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26387,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26387,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26387,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26387,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(26387,36,'cp -12','d8b6 b5a4 f5g4 e1g1 c5d4 c3d4 g8e7 b1d2 g4h5 h2h3 b6a6 d1b3 h7h6 f1e1 a8b8 a4b5 a6b6 a2a3 a7a6 b5e2 b6b3 d2b3 h5f3 e2f3 a6a5 a3a4 e7f5 c1e3 b7b5 e1c1 e8d7 f3d1 b5a4 a1a4 f8b4 c1c2 b8b6 g2g4');
INSERT INTO evaluations VALUES(32205,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32205,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(32205,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(32205,36,'cp -24','f7e6 d2d4 d8f6 b1c3 b8d7 d1d2 e6e5 c3b5 f6d8 d4e5 d6e5 d2d5 f8d6 c1f4 g8e7 d5e6 d7c5 e6c4 e5f4 e4e5 a7a6 e5d6 a6b5 c4b5 c5d7 d6e7 d8e7 b5e2 e7e2 e1e2 e8c8 h1d1 h7h6 e2f2 g7g5 h2h3 d8e8 a2a4 d7c5 a4a5 e8e3 d1d5 c5e4 f2g1 e4f6 d5d4 h8d8 d4d8 c8d8 a5a6 b7a6 a1a6 d8e7');
INSERT INTO evaluations VALUES(32452,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(32452,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(32452,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(32452,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(32452,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(32452,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(32452,36,'cp -35','d2d4 d8f6 b1c3 b8d7 d1d2 e6e5 c3b5 f6d8 d4e5 d6e5 d2d5 f8d6 c1f4 g8e7 d5b3 d7c5 b3c4 e5f4 e4e5 a7a6 e5d6 a6b5 c4b5 c5d7 d6e7 d8e7 b5e2 e7e2 e1e2 e8c8 h1d1 h7h6 e2f2 h8e8 h2h3 g7g5 d1d5 d7f6 d5d8 c8d8 a1d1 d8c8 d1e1 e8e1 f3e1 c8d7 e1d3 d7d6 f2f3 c7c5 d3f2 d6d5 f2g4');
INSERT INTO evaluations VALUES(32449,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(32449,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(32449,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(32449,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(32449,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(32449,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(32449,36,'cp -14','d8f6 b1c3 b8d7 c3b5 e8c8 b5a7 c8b8 a7b5 g7g5 e1g1 c7c6 b5c3 g8e7 a2a4 f8h6 a4a5 g5g4 f3e1 h8g8 g1h1 f4f3 a5a6 h6c1 d1c1 f3g2 e1g2 f6d4 a6b7 b8b7 f1f7 g8g7 c3e2 d4e5 c1e3 d7c5 f7g7 e5g7 e3a3 b7c8 e2c3 c8d7 b2b4 c5b7 c3e2 g7e5 a3a7 d7e8');
INSERT INTO evaluations VALUES(33533,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(33533,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(33533,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(33533,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(33533,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(33533,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(33533,36,'cp 14','f8e7 d1e2 d7d6 c1e3 b8c6 d4c6 b7c6 b2b3 e8g8 e3g5 e6e5 d3c4 h7h6 g5e3 a8b8 a1d1 c8e6 f2f3 e6c4 e2c4 c7b7 e3f2 a6a5 c3a4 f6h5 g2g3 h5f6 c4e2 f6d7 c2c4 b7c7 a4b2 c7b7 b2d3 a5a4');
INSERT INTO evaluations VALUES(33352,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(33352,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(33352,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(33352,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(33352,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(33352,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(33352,36,'cp 25','d1e2 d7d6 a2a4 e8g8 a4a5 c8d7 c1e3 b8c6 d4c6 d7c6 e3b6 c7b8 b6d4 b8c7 h2h3 h7h6 c3a4 d6d5 e4e5 f6e4 c2c4 f7f5 e5f6 e7f6 d4f6 f8f6 d3e4 d5e4 a4c3 c7f4 e2e3 f4e3 f2e3 a8d8 f1f6 g7f6 a1d1 d8d1 c3d1 g8f7 d1c3 f7e7 b2b4 f6f5 c4c5');
INSERT INTO evaluations VALUES(33809,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(33809,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(33809,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(33809,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(33809,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(33809,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(33809,36,'cp -12','e4f6 c1g5 e7e2 f1e2 f8e7 b1c3 b8c6 e1c1 h7h6 g5f6 e7f6 c3d5 f6d8 d3d4 c6e7 d5e7 d8e7 h2h4 h6h5 e2f3 a8b8 g2g3 g7g6 a2a4 e8g8 h1e1 e7d8 c1d2 c8e6 c4e3 f8e8 d1a1 g8g7 a4a5 d8f6 d2d3 c7c6 c2c3 d6d5 f3g2');
INSERT INTO evaluations VALUES(33786,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(33786,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(33786,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(33786,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(33786,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(33786,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(33786,36,'cp -7','c1g5 e7e2 f1e2 f8e7 b1c3 b8c6 h2h4 c6d4 e2d1 e8g8 e1d2 h7h6 g5f6 e7f6 c3d5 f6d8 c2c3 d4f5 a2a4 c7c6 d5e3 f5e3 c4e3 a7a5 d1g4 c8g4 e3g4 h6h5 g4e3 g7g6 d3d4 g8g7 d2d3 d6d5 a1e1 d8c7 g2g3 a8e8 b2b3');
INSERT INTO evaluations VALUES(33784,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(33784,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(33784,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(33784,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(33784,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(33784,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(33784,36,'cp 0','e7e2 f1e2 f8e7 b1c3 b8c6 e1c1 h7h6 g5f6 e7f6 c3d5 f6d8 c2c3 c6e7 d5e7 e8e7 d3d4 c8e6 h1e1 e7f8 c1c2 a7a5 a2a4 h6h5 f2f4 g7g6 c4e3 e6d7 b2b3 f8g7 e2f3 a8b8 g2g3 b7b6 d1a1 h8e8 h2h4 c7c5 e3c4 e8e1 a1e1 c5d4');
INSERT INTO evaluations VALUES(34152,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(34152,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(34152,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34152,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(34152,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(34152,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(34152,36,'cp 3','b7b5 f1d3 g8f6 d1e2 c8b7 c1d2 b8c6 d4c6 b7c6 e4e5 f6d5 c3d5 c6d5 e1g1 c7c6 f2f3 f8e7 f1d1 e8g8 d2b4 e7b4 a3b4 d7d6 e5d6 c6d6 c2c3 d6c7 e2e3 f8d8 d3e2 h7h6 h2h4 a6a5 e3c5 c7c5 b4c5 d8c8 e2b5 c8c5');
INSERT INTO evaluations VALUES(34319,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(34319,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(34319,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(34319,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(34319,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(34319,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(34319,36,'cp 66','g8f6 c1d2 c8f5 g1f3 b8d7 d1e2 a5c7 c4b3 e7e6 f3h4 f5g6 g2g3 a7a5 e1c1 d7b6 a2a3 f8e7 d2f4 e7d6 h4g6 h7g6 f4g5 d6e7 h2h4 a5a4 b3a2 b6d5 a2d5 f6d5 c3e4 f7f6 g5d2 b7b5 h4h5 g6g5 c1b1 c7d7 g3g4 e8f7 h1e1 a8e8 e2f3');
INSERT INTO evaluations VALUES(33584,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(33584,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(33584,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(33584,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(33584,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(33584,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(33584,36,'cp 61','c1d2 c8f5 g1f3 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 b7b5 c4d3 b5b4 c3e4 f6e4 d3e4 d7f6 e4f3 f8e7 g2g4 a8d8 e2c4 f6d5 f3d5 d8d5 d2b4 h8h4 h1h4 e7h4 f2f3 h4e7 b4c3 a7a5 c1b1 e7d6 d1h1');
INSERT INTO evaluations VALUES(33585,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(33585,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(33585,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(33585,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(33585,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(33585,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(33585,36,'cp 70','c8f5 g1f3 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 b7b5 c4b3 d7b6 g2g4 e8c8 g4g5 f6h5 a2a4 a7a6 e2f3 d8d4 d2e3 d4f4 f3e2 c8b7 e3f4 h5f4 e2f3 b6d5 c1b1 f8b4 b3d5 e6d5 c3e2 f4e6 h4h5 g6h5 h1h5 h8f8 a4b5 a6b5 e2d4 b4e7 d4e6 f7e6 f3h3 e6e5 h5h8 f8h8 h3h8 e7g5');
INSERT INTO evaluations VALUES(34630,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(34630,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(34630,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34630,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(34630,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(34630,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(34630,36,'cp 106','e7e6 g1f3 g5d8 d2d4 g8f6 d1d3 f8e7 c1f4 e8g8 e1c1 c7c5 d4c5 d8d3 d1d3 e7c5 f4d6 c5d6 d3d6 b8c6 a2a3 h7h6 e2b5 c8d7 h1d1 d7e8 b5c6 e8c6 f3e5 a8c8 g3g4 c6d5 c3d5 f6d5 c2c3 a7a5 h2h4 a5a4');
INSERT INTO evaluations VALUES(36622,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36622,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36622,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36622,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36622,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36622,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36622,36,'cp -3','f6h5 f4g5 h7h6 g5h4 g6g5 f3e1 h5f4 e3f4 g5h4 c2c3 e7e5 d4e5 d6e5 f4e5 d8d1 e2d1 g7e5 h2h3 e5f6 b1d2 f8d8 e1f3 a7a5 f1e1 a5a4 a2a3 b8d7 d2e4 f6e7 e4g3 h4g3 e1e7 g3f2 g1f2 g8f8 e7e3 a8a6 a1b1 d7f6 d1e2 a6b6 f3d4');
INSERT INTO evaluations VALUES(41809,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41809,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41809,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41809,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41809,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41809,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(41809,36,'cp 14','b7b6 b2b3 c8b7 c2c4 e8g8 c4d5 f6d5 b1d2 b8d7 c1b2 c7c5 d2e4 d6e7 e4g3 a8c8 d3c4 d7f6 a1c1 c5d4 f3d4 a7a6 a2a4 e7c5 g3h5 f6h5 d1h5 d8e7 h5e2 a6a5 d4f3 d5b4 f3e5 f7f6 e5d3 b4d3 c4d3 g8h8 f1d1 f8d8 d3c4');
INSERT INTO evaluations VALUES(43437,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(43437,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(43437,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(43437,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(43437,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43437,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(43437,36,'cp 38','b2b3 f8b4 c1d2 e8g8 f1d3 b8d7 e1g1 d8e7 h2h3 f8d8 d1c2 e6e5 f3e5 d7e5 d4e5 e7e5 c4d5 c8d7 c3e2 b4d6 e2g3 c6d5 d2c3 e5e7 c3f6 e7f6 d3h7 g8f8 h7d3 a8c8 c2d1 g7g6 a1c1 c8c1 d1c1 f8g8 d3e2 d8e8 c1d1 d6g3 f2g3 f6g5');
INSERT INTO evaluations VALUES(44879,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44879,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(44879,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44879,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(44879,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44879,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44879,36,'cp 9','g5f6 e7f6 c4d5 d8d5 b1c3 d5d8 g2g3 b8d7 f1g2 d7b6 e2e3 f8d6 a1c1 g7g6 g1e2 e8g8 e1g1 c8e6 c3e4 a7a5 e4c5 d6c5 d4c5 b6d7 e2d4 d8e7 d1c2 f8d8 d4e6 e7e6 b2b3 f6f5 f1d1 e6e5 d1d4 d7f6 c1d1 d8d4 d1d4 g8g7 h2h3');
INSERT INTO evaluations VALUES(46929,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(46929,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(46929,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46929,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(46929,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(46929,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(46929,36,'cp 0','g1f3 c7c5 d2d3 b8c6 e5d6 e7d6 e1g1 f8e7 a2a4 c6a5 b3a2 e8g8 b1c3 a5c6 h2h3 c8f5 f1e1 d6d5 d3d4 c5c4 a2b1 e7b4 c1d2 d8f6 c3e2 b4d2 d1d2 f8e8 d2f4 e8e4 f4d2');
INSERT INTO evaluations VALUES(46928,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46928,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46928,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46928,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46928,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46928,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46928,36,'cp 0','c7c5 a2a4 b8c6 e5d6 c8f5 d2d4 c5c4 b3a2 e7e6 b1d2 f8d6 d2c4 b6c4 a2c4 c6b4 e1g1 b4c2 c1g5 f7f6 f3h4 f5g6 h4g6 h7g6 d1c2 d6h2 g1h1 h2d6 h1g1');
INSERT INTO evaluations VALUES(47131,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47131,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47131,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47131,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47131,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47131,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47131,36,'cp 2','e7e6 g1f3 b7b6 b2b3 c8b7 c4d5 e6d5 c1b2 b8d7 a1c1 a7a6 c3e2 f8d6 e2d4 c7c5 d4f5 d6f8 f1e2 g7g6 f5g3 f8g7 d2d4 e8g8 c1c2 f8e8 e1g1 h7h5 h2h3 a8c8 d4c5 b6c5 d1a1 d8b6 f3e5 h5h4 e5d7 f6d7 b2g7');
INSERT INTO evaluations VALUES(51784,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(51784,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51784,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(51784,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(51784,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(51784,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(51784,36,'cp 28','d2d4 d7d5 g1f3 g8f6 f1g2 f8e7 e1g1 e8g8 d1c2 d5c4 c2c4 b7b5 c4c2 c8b7 c1g5 b8d7 f3e5 d7e5 d4e5 f6d5 g5e7 d8e7 b1c3 d5c3 c2c3 c6c5 g2b7 e7b7 c3c5 b7e4 f1e1 a8c8 c5b5 c8b8 b5a6 b8b2 a1c1 e4e5 a6a7 b2e2 e1e2 e5e2');
INSERT INTO evaluations VALUES(54793,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(54793,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(54793,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(54793,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(54793,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(54793,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(54793,36,'cp 93','h4g3 c8f5 e2e3 b8c6 h2h4 g5h4 h1h4 d8d7 b1d2 h7h5 f1b5 a7a6 b5c6 d7c6 a1c1 b7b5 h4h5 g8h6 g1f3 f5g4 h5h1 f8g7 h1h4 e8f7 c2c3 a8g8 h4h1 c6d7 d2b3 d7c6 c1a1 h6f5 h1h8 g8h8');
INSERT INTO evaluations VALUES(5452,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(5452,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(5452,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(5452,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(5452,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(5452,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(5452,36,'cp 89','f1b5 c8g4 b5c6 b7c6 b1d2 f7f6 d1a4 g8h6 h2h3 g4d7 d2b3 h6f7 e5f6 e7f6 e1g1 e8g8 b3c5 f8e8 c1f4 g7f8 f1e1 a7a5 a1c1 e8e1 f3e1 f8c5 c1c5 d8b6 e1d3 f7d8 c5c3 d8e6 f4e3 g8f7 a4c2 b6b5 a2a3 a8e8 c2d2 a5a4');
INSERT INTO evaluations VALUES(5449,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(5449,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(5449,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(5449,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(5449,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(5449,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(5449,36,'cp 82','c8g4 b5c6 b7c6 b1d2 f7f6 e1g1 f6e5 d4e5 g8h6 h2h3 g4f3 d2f3 e8g8 c1e3 d8d7 f1e1 h6f7 e3c5 f7d8 f3d4 d8b7 b2b4 a7a5 a2a3 a5b4 a3b4 b7c5 b4c5 a8a1 d1a1 d7b7 a1d1 f8a8 g2g3 b7c7 d1g4 g8h8 g4f4 c7a5 e1e2 a5a1 g1g2 a1a4 e2d2 h8g8 f4g4 g8h8 g4g5 a8b8');
INSERT INTO evaluations VALUES(5859,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(5859,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(5859,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(5859,33,'cp 242',NULL);
INSERT INTO evaluations VALUES(5859,34,'cp 265',NULL);
INSERT INTO evaluations VALUES(5859,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(5859,36,'cp 276','g8e7 h2h4 e8g8 h4h5 h7h6 c4d3 d8d6 e4e5 d6d5 h1h3 f7f5 e1f1 b7b5 f1g1 c8b7 h3g3 g8h8 d1d2 a8d8 a1b1 a7a6 a2a4 f5f4 g3g4 b5b4 g4f4 b7a8 d3e4 d5a5 c3b4 f8f4 b4a5 f4e4 c1a3 c6d4 f3d4 e4d4 d2e2');
INSERT INTO evaluations VALUES(8278,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8278,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8278,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(8278,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(8278,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(8278,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(8278,36,'cp 20','d2d4 e8g8 d4c5 d6c5 c1e3 d8a5 d1b3 a5b4 c3d5 f6d5 c4d5 b4b3 a2b3 c6d4 f3d4 c5d4 e3f4 c8g4 f1e1 f8d8 h2h3 g4f5 e1c1 h7h5 c1c7 d8d7 c7d7 f5d7 f4g5 d4d3 e2d3 g7b2 a1a2 b2f6 g5f6 e7f6 b3b4 d7f5 d5d6 a8d8 a2a7 d8d6 a7b7 d6d3 b4b5 d3d1 g1h2');
INSERT INTO evaluations VALUES(8281,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(8281,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(8281,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(8281,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(8281,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(8281,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(8281,36,'cp 22','e8g8 d4d5 c6a5 f3d2 e7e6 a1b1 e6d5 c4d5 c8f5 e2e4 f5d7 b2b3 f8e8 d2c4 a5c4 b3c4 f6g4 c1d2 g4e5 d1e2 d8c8 c3b5 d7b5 b1b5 b7b6 d2f4 c8d7 b5b3 h7h6 e2c2 a8b8 b3b1 a7a6 a2a4 b6b5 c4b5 a6b5 a4b5 b8b5 b1b5 d7b5');
INSERT INTO evaluations VALUES(11273,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(11273,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(11273,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(11273,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(11273,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(11273,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(11273,36,'cp 72','d1d2 h7h5 e1c1 c8d7 f1d3 f8e7 c1b1 b7b5 d4c6 d7c6 h2h4 d8c7 c3e2 c6d7 h1e1 a6a5 e2g3 b5b4 e3f4 e6e5 f4g5 a5a4 d2b4 d6d5 b4d2 d5d4 f3f4 a8b8 f4e5 e7b4 d2c1 f6g4 g3f5 d7f5 e4f5 b4e1 d1e1 g4f2 c1d2 f2d3 d2d3 e8g8');
INSERT INTO evaluations VALUES(11274,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(11274,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(11274,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(11274,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(11274,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(11274,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(11274,36,'cp 62','h7h5 e1c1 c8d7 c1b1 f8e7 f1d3 b7b5 d4c6 d7c6 c3e2 a8c8 h2h3 h5h4 g2g4 h4g3 e2g3 f6d7 f3f4 e7f6 d2e2 d8a5 f4f5 a5b4 e3c1 d7c5 f5e6 c5a4 e6f7 e8f7 e2e1 b4e1 d1e1 a4c5 g3f5 f6e5 e1g1 b5b4 f5e3 c5d3');
INSERT INTO evaluations VALUES(12158,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12158,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(12158,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(12158,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(12158,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(12158,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(12158,36,'cp 3','f8d6 d1c2 e8g8 e1g1 f8e8 b2b3 b7b6 c1b2 c8b7 e3e4 f6e4 c3e4 d5e4 c2e4 d8e7 f3e5 f7f5 e4e3 d7e5 d4e5 d6b4 a1d1 e8d8 h2h3 d8d1 f1d1 a8d8 d1d8 e7d8 b2c3 b4c3 e3c3 c6c5 e2f3 d8e7 f3b7 e7b7 c3d3 g8f7 d3d6 b7e7 d6e7 f7e7 f2f4 e7e8');
INSERT INTO evaluations VALUES(7749,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(7749,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7749,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7749,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7749,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(7749,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(7749,36,'cp 13','d1c2 e8g8 e1g1 f8e8 b2b3 d8e7 c1b2 b7b6 a2a4 a7a5 e2d3 c8b7 a1e1 h7h6 e3e4 d5e4 c3e4 a8d8 e4d6 e7d6 e1e3 c6c5 d4c5 d7c5 f3e5 c5d3 e3d3 d6e7 f2f3 d8d3 c2d3 e8d8 d3e3 f6d7 f1d1 f7f6 e5d7 d8d7');
INSERT INTO evaluations VALUES(18591,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(18591,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(18591,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(18591,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(18591,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(18591,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(18591,36,'cp 76','f2f4 e7e5 g1f3 e5d4 d1d4 g8f6 c1e3 d6d5 e4e5 f8c5 d4d2 f6e4 c3e4 d5e4 f3d4 d7b6 e1c1 e8g8 d2f2 b6d5 f1c4 d8b6 h2h3 f7f6 e5f6 f8f6 g2g4 c8e6 c4d5 e6d5 a2a3 a8d8 d4f5 c5e3 f5e3 f6d6 h1e1 h7h6 f2g3 d5f7 d1d6 d8d6');
INSERT INTO evaluations VALUES(19197,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(19197,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(19197,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(19197,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(19197,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(19197,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(19197,36,'cp -32','e5f6 g8f6 f1b5 d8b6 a2a4 e7e6 e1g1 f8e7 d2d3 e8g8 f1e1 c6d4 f3d4 c5d4 c1f4 d7d5 c2c4 d4c3 b1c3 a7a6 a4a5 b6d8 b5a4 d8a5 a4d7 a5b4 d7e6 c8e6 e1e6 e7c5 f4g3 a8e8 e6e8 f8e8 h2h3 b4b2 c3a4 b2d4 a4c5 d4c5');
INSERT INTO evaluations VALUES(19195,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(19195,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(19195,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(19195,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(19195,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(19195,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(19195,36,'cp -37','g8f6 f1b5 e7e6 e1g1 d8c7 f1e1 f8e7 b5c6 b7c6 c2c3 e8g8 d2d3 d7d6 d3d4 c5d4 c3d4 c6c5 d4c5 c7c5 b1c3 a8b8 h2h3 a7a5 c1d2 e6e5 a1c1 c8f5 f3h4 f5e6 b2b3 c5d4 h4f3 d4d3 d2g5 d3d1 c3d1 b8c8 d1e3 c8c1 e1c1 f8c8 c1c8 e6c8');
INSERT INTO evaluations VALUES(12708,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(12708,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(12708,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(12708,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(12708,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(12708,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(12708,36,'cp -42','f1b5 e7e6 e1g1 d8c7 f1e1 f8e7 b5c6 b7c6 c2c3 e8g8 d2d3 d7d6 d3d4 c5d4 c3d4 c6c5 d4c5 c7c5 b1c3 a8b8 h2h3 e6e5 c1e3 c5c6 a1c1 c6a8 e3g5 b8b2 g5f6 e7f6 d1d6 c8b7 d6d3 b2b4 d3d6 b4f4 f3e5 f6e5 e1e5 b7g2');
INSERT INTO evaluations VALUES(19597,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(19597,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(19597,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(19597,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(19597,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(19597,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(19597,36,'cp -40','b3c4 e7e6 g1f3 g8f6 c4e2 f8e7 e1g1 e8g8 f1e1 b7b6 d2d3 f8e8 h2h3 c8b7 c1f4 h7h6 a2a4 d7d6 f4g3 f6d7 b1d2 d7f8 a4a5 e6e5 a5b6 a7b6 a1a8 d8a8 g3h4 f7f6 d2c4 b6b5 c4e3 f8e6 h4g3 e6f4 g3f4 e5f4 e3d1 e7f8 f3h4');
INSERT INTO evaluations VALUES(21228,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(21228,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(21228,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(21228,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(21228,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(21228,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(21228,36,'cp -25','f1c4 e7e5 g1f3 d4f3 d1f3 d8f6 f3d3 f8d6 e1g1 g8e7 c3b5 e8g8 c1e3 e7c6 b5d6 c7d6 f1d1 f8d8 a1c1 c8g4 f2f3 g4e6 b2b4 a7a6 c4e6 f7e6 a2a4 d8d7 d3b3 f6f7 b4b5 a6b5 a4b5 c6e7 e3c5 a8d8 c5b6 d8c8 c1c8 e7c8 b6f2 c8e7 f2c5 d6d5 e4d5 e7d5 c5f2 d5f4 f2e3 d7d5 h2h3 h7h6 e3f4 e5f4');
INSERT INTO evaluations VALUES(21224,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(21224,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(21224,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(21224,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(21224,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(21224,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(21224,36,'cp -50','c7c6 g1e2 c8e6 c4d3 g7g6 c1e3 f8g7 e1g1 d4e2 c3e2 d8c7 e2d4 g7d4 e3d4 g8f6 d3e2 a8d8 d1a4 e8g8 a4a7 d8a8 a7b6 c7b6 d4b6 f6d7 b6e3 e6a2 b2b4 a2e6 f1c1 f7f6 b4b5 a8a1 c1a1 c6b5 e2b5 f8c8 b5d7 e6d7 a1c1 d7c6 c1c5 c8d8 f2f3 g8f7');
INSERT INTO evaluations VALUES(22484,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(22484,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(22484,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(22484,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(22484,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(22484,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(22484,36,'cp 114','c1e3 g8f6 b1c3 d7d6 d1d2 f8g7 f2f3 e8g8 e1c1 d6d5 e4d5 f6d5 e3g5 d5c3 g5d8 c3a2 c1b1 f8d8 d2e1 c6d4 f1c4 c8f5 b1a2 f5c2 e1e7 d8f8 d1d4 g7d4 h1c1 c2a4 e7b4 a4c6 c4f7 f8f7 b4d4 a8e8 c1d1 a7a6 a2a1 h7h5 d4c4 g8g7 h2h4 c6b5 c4c3 g7h7 c3b3 h7g7 a1b1 b5e2 d1d2 e2b5 g2g4');
INSERT INTO evaluations VALUES(26243,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26243,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26243,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(26243,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(26243,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(26243,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(26243,36,'cp -4','e7e6 h2h3 g4h5 c1f4 d8f6 g2g3 h5f3 d1f3 g8e7 b1d2 a7a6 b5c6 e7c6 f3b3 e8c8 b3b6 f8d6 f4d6 d8d6 c2c3 f6d8 b6d8 d6d8 a2a4 b7b6 d2f3 f7f6 h3h4 c8d7 f1e1 d8c8 e1e2 h8e8 f3e1 d7d6 e1c2 h7h5 a1e1 d6d7 g1g2 c8b8 b2b3 c6a5 c2b4 a5b3 b4a6');
INSERT INTO evaluations VALUES(26378,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(26378,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(26378,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26378,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(26378,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(26378,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26378,36,'cp 0','h2h3 g4h5 c1f4 d8f6 g2g3 h5f3 d1f3 g8e7 b1d2 a7a6 b5c6 e7c6 f3b3 e8c8 b3b6 f8d6 f4d6 d8d6 c2c3 f6h6 d2b3 e6e5 d4e5 c6e5 b6c5 c8b8 g1g2 h6g6 a1d1 h8d8 d1d2 e5f3 d2e2 f3h4 g2g1 h4f3');
INSERT INTO evaluations VALUES(26581,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(26581,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(26581,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(26581,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(26581,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(26581,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(26581,36,'cp 42','f1b5 e7e6 b5c6 b7c6 d2d3 a7a5 a2a4 f8d6 d1e2 f5g6 b1d2 g8e7 b2b3 c6c5 c1b2 f7f6 e1c1 e7c6 h2h4 h7h5 g2g3 d8d7 h1g1 c6b4 e3e4 e8c8 c2c3 b4c6 c1b1 h8g8 d3d4 c5d4 f3d4 c6d4 e2a6 c8b8 c3d4 g6e4 d2e4 d5e4 a6a5');
INSERT INTO evaluations VALUES(29270,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(29270,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(29270,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(29270,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(29270,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(29270,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(29270,36,'cp 28','d1b3 d8b6 f3e5 b8d7 g2g4 h5g6 e5g6 h7g6 f1g2 g6g5 c4d5 c6d5 c1d2 f8e7 e1g1 b6b3 a2b3 e7b4 f1c1 a7a5 c1c2 e8g8 a1c1 f8d8 d2e1 g7g6 f2f3 f6e8 h3h4 g5h4 e1h4 d8c8 g2f1 g8g7 f1b5 d7f6 h4g3 b4d6 g1g2 d6g3');
INSERT INTO evaluations VALUES(30285,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30285,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30285,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30285,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30285,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30285,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30285,36,'cp 0','d8d7 g1f3 d7d5 d2d4 b8c6 c1e3 e7e6 b1d2 d5f5 c2c3 f8d6 d1b3 c6a5 b3a4 a5c6');
INSERT INTO evaluations VALUES(32713,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32713,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32713,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32713,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32713,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32713,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32713,36,'cp 0','g1f3 d7d5 d2d4 b8c6 c1e3 e7e6 b1d2 d5f5 c2c3 f8d6 d1b3 c6a5 b3a4 a5c6');
INSERT INTO evaluations VALUES(31788,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(31788,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(31788,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(31788,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(31788,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(31788,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(31788,36,'cp 48','d2d4 e6f5 h2h4 h7h6 g2g4 f5e4 f2f3 e4h7 e5e6 f7e6 f1d3 g8f6 d3h7 h8h7 d1d3 c7c5 d4c5 d8a5 c1d2 a5c5 g1h3 b8c6 b1c3 c5c4 d3c4 d5c4 e1c1 e6e5 g4g5 f6h5 g5g6 h7h8 d2e3 e7e6 c3b5 a8c8 b5a7 c6a7 e3a7 f8e7 a7b6 e7h4 h3f2');
INSERT INTO evaluations VALUES(31787,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(31787,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(31787,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(31787,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(31787,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(31787,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(31787,36,'cp 63','e6f5 h2h4 h7h6 g2g4 f5e4 f2f3 e4h7 e5e6 f7e6 f1d3 g8f6 d3h7 h8h7 d1d3 c7c5 d4c5 d8a5 c2c3 e6e5 d3g6 e8d8 b2b4 a5a4 g1e2 b8d7 g4g5 h6g5 c1g5 a7a5 a2a3 e5e4 b1d2 d7e5 g6f5 e5d3 e1f1 a4d7 f3e4 a5b4 c3b4 d7f5 e4f5 d3b4 e2d4 e7e5 d4e6 d8e7 e6c7 a8a5 d2f3 a5c5 a3b4 c5c7');
INSERT INTO evaluations VALUES(35316,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(35316,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(35316,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(35316,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(35316,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(35316,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(35316,36,'cp 118','e7e6 g1f3 g8f6 f3e5 d7d8 c1e3 f8b4 d1f3 e8g8 f1d3 f6d5 e3d2 b8c6 e5c6 b7c6 a2a3 b4d6 e1c1 a8b8 c3a4 d8h4 c2c4 h4d4 d2e3 d4h4 g2g3 h4f6 f3e4 g7g6 c4d5 c6d5 e4h4 f6h4 g3h4 e6e5 h1e1 d5d4 e3h6 f8e8 b2b4 c8g4 d3e2 g4d7 a4c5 d7f5 h6g5 g8g7 h4h5');
INSERT INTO evaluations VALUES(35308,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(35308,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(35308,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(35308,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(35308,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(35308,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(35308,36,'cp 121','g1f3 g8f6 f3e5 d7d8 c1f4 f8b4 f1d3 e8g8 e1g1 b8c6 e5c6 b7c6 f4e5 a8b8 d1e2 f6d7 a1d1 d7e5 d4e5 d8e7 c3e4 f7f5 e5f6 g7f6 e2h5 g8h8 f2f4 e7g7 f1f3 f6f5 e4g5 b4e7 b2b3 e6e5 d3c4 e5f4 g5f7 h8g8 f7d6 g8h8 d6c8 b8c8 f3f4');
INSERT INTO evaluations VALUES(35307,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(35307,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(35307,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(35307,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(35307,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(35307,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(35307,36,'cp 112','g8f6 c1g5 f8e7 f1b5 c7c6 b5d3 d7c7 e1g1 b7b6 d1e2 b8d7 f1e1 e8g8 f3e5 d7e5 g5f4 f6d5 c3d5 c6d5 d4e5 g7g6 f4h6 f8e8 c2c3 a7a5 a1d1 e7f8 h6g5 f8e7 g5f4 c8b7 f4h6 e7f8 h6g5 f8e7 g5f4 e8d8 d3b5 c7c5');
INSERT INTO evaluations VALUES(40085,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40085,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40085,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40085,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40085,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(40085,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(40085,36,'cp 27','e8g8 b2b3 b7b6 c1b2 b8d7 f1e1 c8b7 b1c3 f6e4 e2e3 e4c3 b2c3 a8c8 f3d2 c6c5 c4d5 b7d5 g2d5 e6d5 d1f3 c5d4 c3d4 d7c5 f3e2 d8d7 e1d1 c5e6 d2f3 f8d8 d4b2 h7h6 f3d4 e6d4 b2d4 d7e6 d1d2');
INSERT INTO evaluations VALUES(45120,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(45120,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45120,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45120,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45120,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(45120,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45120,36,'cp 0','d6e5 f3e5 a7a6 f1e2 d8c7 d2d4 e7e6 c1f4 f8d6 c3a4 c5d4 d1d4 f6d5 f4g3 d6e5 d4e5 c7a5 c2c3 e8g8 e2d1 b8d7 e5d4 e6e5 d4d2 d7f6 c3c4 a5d2 e1d2 f6e4 d2e2 d5b4 a4b6 e4g3 h2g3 a8b8 a2a3 b4c6 e2d2 c8e6 d2c3 b8d8 d1f3 f7f5 f3c6');
INSERT INTO evaluations VALUES(45310,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(45310,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45310,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(45310,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(45310,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(45310,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45310,36,'cp -10','f3e5 a7a6 d2d3 d8c7 d1e2 b8c6 e5c6 c7c6 e2e3 g7g6 h2h3 b7b6 f1e2 c8b7 e2f3 c6d7 e1g1 f8g7 a2a4 a8c8 f3b7 d7b7 f1e1 e7e6 c3e4 f6e4 e3e4 b7d7 a1b1 e8g8 b2b3 f8d8 c1g5 d8e8 g5f4 b6b5 a4b5 a6b5 f4e5 f7f6');
INSERT INTO evaluations VALUES(45305,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(45305,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(45305,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(45305,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(45305,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(45305,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(45305,36,'cp -20','a7a6 a2a4 c8e6 d1f3 b8d7 e5c4 e6c4 f1c4 d7e5 f3e2 e5c4 e2c4 e7e6 e1g1 a8c8 d2d3 d8c7 f1e1 f8d6 g2g3 e8g8 c3e4 f6d5 e4d6 c7d6 b2b3 f7f6 e1e2 e6e5 a4a5 f8f7 c1d2 g8h8 a1e1');
INSERT INTO evaluations VALUES(47120,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(47120,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(47120,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(47120,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(47120,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(47120,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(47120,36,'cp -98','c7c5 e2e3 d8b6 b1c3 e7e6 f1b5 c8d7 b5d7 b8d7 g1e2 h7h6 g5f4 f8e7 c3a4 b6c6 a4c5 d7c5 d4c5 e8g8 f4e5 f8c8 c2c3 a7a5 e1g1 a5a4 e5d6 e7d6 c5d6 c6d6 d1d2 f6e8 a1b1 d6c5 d2d4 e8d6 d4c5 c8c5 f1c1 d6c4 g1f2 c5b5 c1c2');
INSERT INTO evaluations VALUES(50139,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(50139,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(50139,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(50139,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(50139,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(50139,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(50139,36,'cp 41','f8g7 e2e3 d7d6 g1f3 e8g8 f1e2 e7e6 b2b4 b8d7 c4c5 h7h6 g5f6 d7f6 e1g1 c8d7 a1c1 d8e7 d1b3 g8h8 f1d1 d7e8 b3a3 a7a6 b4b5 e8b5 e2b5 a6b5 c5d6 e7d6 a3d6 c7d6 c3b5 a8a2');
INSERT INTO evaluations VALUES(50533,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(50533,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50533,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50533,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(50533,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(50533,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50533,36,'cp 0','b7b5 d2d4 c5d4 f3d4 c8b7 f1d3 d7d6 d1e2 b8d7 h2h4 a8c8 c1d2 d7e5 e1c1 g8f6 h4h5 h7h6 c1b1 f8e7 f2f4 e5c4 d2e1 d8b6 e1f2 b6c7 f2e1');
INSERT INTO evaluations VALUES(51220,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51220,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51220,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51220,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(51220,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51220,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(51220,36,'cp 0','d2d4 c5d4 f3d4 c8b7 f1d3 d7d6 d1e2 b8d7 c1d2 g8f6 e1c1 g7g6 f2f3 f8g7 c1b1 d7e5 h2h4 h7h5 d2g5 d8b6 g5e3 b6d8 e3g5');
INSERT INTO evaluations VALUES(50543,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(50543,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(50543,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50543,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50543,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(50543,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50543,36,'cp 0','d7d5 e4d5 e6d5 f1e2 g8f6 d2d4 b8c6 e1g1 c8e6 c1e3 c5c4 f3e5 f8d6 f2f4 c6e7 e2f3 h7h5 h2h3 d6b4 d1e2 d8a5 e3f2 a5c7 c3a2 b4d6 a2c3');
INSERT INTO evaluations VALUES(51352,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(51352,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(51352,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(51352,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(51352,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(51352,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(51352,36,'cp 99','g8f6 c1f4 c8b7 f1d3 a7a6 d1e2 f8e7 a1d1 f6d7 e1g1 b6b5 a2a4 b5b4 c3b1 e8g8 b1d2 d7b6 a4a5 b6a4 d1b1 c7c5 d4d5 e6d5 e4d5 b8d7 b1e1 f8e8 d2c4 b7d5 d3h7 g8h7 e2d3 h7g8 d3d5 d7f6 d5c6 d6d5 c4b6 a4b6 a5b6 a6a5 b6b7');
INSERT INTO evaluations VALUES(52541,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(52541,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(52541,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(52541,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(52541,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(52541,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52541,36,'cp 7','e7f6 f1d3 g8e7 b1d2 b8a6 d2f1 c8f5 f1g3 f5d3 d1d3 e7c8 h2h4 f8d6 g3f5 e8g8 g1e2 f8e8 c1d2 d8d7 a2a3 d6f8 g2g4 c8d6 e1c1 d6e4 d2e1 a6c7 f2f3 e4d6 f5d6 f8d6 e1d2 e8e7 c1b1 a8e8 d1e1 b7b6 g4g5 f6f5 h4h5');
INSERT INTO evaluations VALUES(52574,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(52574,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(52574,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(52574,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(52574,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(52574,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(52574,36,'cp 47','f1e2 g8f6 g1f3 f8g7 e1g1 f6e4 b1c3 e8g8 d1b3 e4c3 b2c3 b7b6 f1e1 c8a6 c1a3 a6c4 e2c4 d5c4 b3c4 c6c5 a1d1 b8c6 h2h4 a8c8 d4c5 d8c7 c5b6 a7b6 h4h5 g7f6 f3d4 c6e5 c4c7 c8c7 d4b5');
INSERT INTO evaluations VALUES(53362,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(53362,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(53362,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(53362,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(53362,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(53362,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(53362,36,'cp 58','c4d5 g8f6 d1a4 d8d7 a4d4 d7d5 b1c3 d5a5 d4e5 a5e5 f3e5 b8c6 e5c6 b7c6 g2g3 e7e6 f1g2 f6d5 c1d2 a8b8 c3a4 f8e7 a1c1 c8a6 h2h4 e8g8 g2f3 f8d8 h4h5 h7h6 e1g1 f7f5 f1d1 g8f7 e2e4 f5e4 f3e4 a6b5 a4c5');
INSERT INTO evaluations VALUES(53850,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(53850,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(53850,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(53850,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(53850,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(53850,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(53850,36,'cp 122','g8f6 e2e4 f8e7 c1f4 d8a5 f1d3 b8d7 e1g1 e6e5 f4e3 e8g8 h2h3 f8e8 d1d2 e7f8 d4d5 c6d5 c4d5 d7c5 d3c2 b7b5 a2a3 b5b4 a3b4 a5b4 c3a4 b4d2 f3d2 c8a6 f1b1 c5d3 a4c3 d3b4 c2a4 e8c8 f2f3 a6d3 b1c1 h7h6');
INSERT INTO evaluations VALUES(45887,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(45887,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(45887,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(45887,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(45887,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(45887,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(45887,36,'cp 141','e2e4 f8e7 c1f4 d8a5 f1d3 b8d7 e1g1 e6e5 f4e3 e8g8 h2h3 a5c7 a2a3 e5d4 f3d4 f8e8 d3c2 e7f8 f1e1 a7a6 a3a4 d7c5 e3f4 h7h6 b2b3 c5e6 f4e3 a6a5 d1d2 e6d4 e3d4 f6d7 f2f4 d7c5 a1d1 c8d7');
INSERT INTO evaluations VALUES(45888,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(45888,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(45888,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(45888,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(45888,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(45888,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(45888,36,'cp 123','f8e7 f1d3 b8d7 h2h3 e6e5 c1e3 e8g8 e1g1 h7h6 d3c2 f8e8 d1e2 d8c7 a2a4 d7f8 d4d5 a7a5 g1h2 f8g6 f3g1 c8d7 g2g3 a8c8 c2d3 e8f8 a1d1 f6h7 b2b3 c8e8 f2f3 f7f5 e4f5 d7f5 h3h4 c7d7 d3f5');
INSERT INTO evaluations VALUES(57346,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(57346,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(57346,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(57346,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(57346,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(57346,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(57346,36,'cp 37','b7b6 g1f3 c8b7 f1e2 f8e7 e1g1 g8f6 b2b3 e8g8 c1b2 f6e4 b1d2 d7d5 c4d5 b7d5 a1c1 b8d7 d2c4 c7c5 c4e5 d7e5 f3e5 c5d4 b2d4 e7a3 c1c2 a8c8 f2f3 c8c2 d1c2 e4d6 e2a6 d8e7 f1d1 d6f7 e5d3 d5b7 a6b7 e7b7');
INSERT INTO evaluations VALUES(57356,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(57356,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(57356,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(57356,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(57356,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(57356,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(57356,36,'cp 17','g8f6 a2a3 b8c6 d4d5 e6d5 c3d5 d7d6 d5f4 c6e5 g2g3 c7c6 f1g2 g7g6 b2b3 f8g7 c1b2 d8e7 e1g1 f6e4 f3d2 e4d2 d1d2 e5c4 b3c4 g7b2 a1b1 b2e5 h2h4 e8g8 f1d1 a7a5 a3a4 g8h8 g2f3 e7f7 d2d3 f7e7 g1g2 f8e8 e2e3');
INSERT INTO evaluations VALUES(57462,30,'cp 393',NULL);
INSERT INTO evaluations VALUES(57462,31,'cp 417',NULL);
INSERT INTO evaluations VALUES(57462,32,'cp 420',NULL);
INSERT INTO evaluations VALUES(57462,33,'cp 438',NULL);
INSERT INTO evaluations VALUES(57462,34,'cp 448',NULL);
INSERT INTO evaluations VALUES(57462,35,'cp 448',NULL);
INSERT INTO evaluations VALUES(57462,36,'cp 447','d1h5 g7g6 e5g6 h7g6 h5h8 e8f7 d2d3 d7d5 b1c3 f8g7 h8h7 b8d7 h7h3 e4d3 f1d3 d7e5 h3g3 e5d3 c2d3 d8e8 e1f1 c7c6 c1g5 g8f6 a1e1 e8f8 c3e2 f6h5 g3f3 f7g8 f3f8 g7f8 f2f3 f8b4 e1d1 c8d7 g2g4 h5g7 h2h4 a8f8 f1g2 g7e6 g5d2 b4c5 h4h5 g6g5 d1e1 f8f7 d3d4 c5d6 h5h6 g8h7 a2a4 h7g8');
INSERT INTO evaluations VALUES(57459,30,'cp 449',NULL);
INSERT INTO evaluations VALUES(57459,31,'cp 437',NULL);
INSERT INTO evaluations VALUES(57459,32,'cp 440',NULL);
INSERT INTO evaluations VALUES(57459,33,'cp 435',NULL);
INSERT INTO evaluations VALUES(57459,34,'cp 461',NULL);
INSERT INTO evaluations VALUES(57459,35,'cp 448',NULL);
INSERT INTO evaluations VALUES(57459,36,'cp 463','g7g6 e5g6 h7g6 h5h8 e8f7 b1c3 d7d5 h8h7 f8g7 d2d3 b8d7 h7h3 e4d3 f1d3 d7e5 h3g3 e5d3 c2d3 d8e8 e1f1 g8f6 c1g5 c8f5 g3c7 e8d7 c7d7 f6d7 c3d5 g7b2 a1d1 d7c5 h2h4 b7b5 g5e3 a8d8 e3c5 d8d5 d3d4 d5d8 h4h5 g6h5 h1h4');
INSERT INTO evaluations VALUES(57537,30,'cp 462',NULL);
INSERT INTO evaluations VALUES(57537,31,'cp 470',NULL);
INSERT INTO evaluations VALUES(57537,32,'cp 472',NULL);
INSERT INTO evaluations VALUES(57537,33,'cp 440',NULL);
INSERT INTO evaluations VALUES(57537,34,'cp 463',NULL);
INSERT INTO evaluations VALUES(57537,35,'cp 470',NULL);
INSERT INTO evaluations VALUES(57537,36,'cp 485','e5g6 h7g6 h5h8 e8f7 d2d3 d7d5 h8h7 f8g7 b1c3 b8d7 h7h3 e4d3 f1d3 d7e5 h3g3 e5d3 c2d3 d8e8 e1f1 g8f6 c1g5 c8f5 h2h4 e8c6 h4h5 f6h5 g3f3 g7c3 b2c3 d5d4 f3c6 b7c6 c3d4 h5g7 g5f4 g7e6 f4e5 g6g5 h1h6 f5d3 f1g1 a8b8 a1e1 b8b1 e1b1 d3b1 a2a3 e6d4 e5d4 b1g6');
INSERT INTO evaluations VALUES(57536,30,'cp 460',NULL);
INSERT INTO evaluations VALUES(57536,31,'cp 484',NULL);
INSERT INTO evaluations VALUES(57536,32,'cp 478',NULL);
INSERT INTO evaluations VALUES(57536,33,'cp 481',NULL);
INSERT INTO evaluations VALUES(57536,34,'cp 471',NULL);
INSERT INTO evaluations VALUES(57536,35,'cp 471',NULL);
INSERT INTO evaluations VALUES(57536,36,'cp 462','h7g6 h5h8 e8f7 d2d3 d7d5 h8h7 f8g7 b1c3 b8d7 h7h3 e4d3 f1d3 d7e5 h3g3 e5d3 c2d3 d8e7 c1e3 d5d4 c3d5 e7e5 g3e5 g7e5 e3f4 c8e6 f4e5 e6d5 a1c1 c7c6 e1g1 g8e7 e5d4 e7f5 d4c3 a7a5 a2a3 b7b5 f1e1 a8e8 e1e8 f7e8 c1e1 e8d7 f2f3 d5b3 e1e4 b5b4 a3b4 a5a4 g1f2 f5d6 e4g4');
INSERT INTO evaluations VALUES(57719,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(57719,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(57719,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(57719,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(57719,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(57719,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(57719,36,'cp 40','g8f6 e2e3 d7d5 c2c4 e7e6 b1c3 f8b4 a1c1 e8g8 g1f3 c7c5 d4c5 f6e4 c4d5 d8d5 d1d5 e6d5 f4e5 b8d7 a2a3 b4c3 e5c3 d7c5 f3d4 c8d7 f2f3 e4c3 c1c3 f8c8 e1d2 c5e6 d4e6 f7e6 f1d3 g8f8 h1c1 c8c3 c1c3 d7c6 b2b4 f8e7 f3f4');
INSERT INTO evaluations VALUES(50294,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(50294,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(50294,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(50294,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(50294,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(50294,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(50294,36,'cp 41','c2c4 e7e6 e2e3 f8b4 b1d2 d7d6 f1d3 c7c5 g1e2 c5d4 e3d4 e8g8 e1g1 e6e5 f4g3 f6h5 d2b3 h5g3 h2g3 e5d4 e2d4 b8d7 d4c2 b4c5 b3c5 d6c5 c2e3 d7f6 d3c2 d8d1 f1d1 c8e6 d1d2 a8d8 a1d1 d8d2 d1d2 g7g6');
INSERT INTO evaluations VALUES(3695,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(3695,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(3695,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(3695,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(3695,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(3695,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(3695,36,'cp 57','b4d6 h2h3 a7a6 a2a3 e6e5 d1c2 e5d4 e3d4 c6e7 f1e1 c8f5 c2d1 e7g6 c4a2 c7c6 f3e5 f6d5 c3d5 c6d5 d1h5 f5e6 a1d1 d8h4 h5h4 g6h4 e5d3 f8c8 d2b4 d6b4 d3b4 c8d8 b4d3 h4g6 d3c5 g6f4 e1e5 d8b8 g1h2 b8c8 a2b3 a8b8 c5e6 f7e6');
INSERT INTO evaluations VALUES(4874,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(4874,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(4874,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(4874,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(4874,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(4874,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(4874,36,'cp 72','f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 d1f3 d7d6 c1f4 c8e6 f4e5 d6e5 f1e2 h7h5 f3e3 h5h4 c3d5 e6d5 e1c1 d8f6 d1d5 g8e7 d5d3 e7g6 g2g3 h4g3 f2g3 g6f8 d3b3 b7b5 e3c3 f8e6 c3c6 e8e7 h1f1 f6h6 c1b1 h8d8 h2h4 h6d2 e2d3 d2g2');
INSERT INTO evaluations VALUES(7468,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(7468,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(7468,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(7468,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(7468,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(7468,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(7468,36,'cp 214','e4e5 f6g4 c3d4 c6d4 f3d4 d8h4 g2g3 h4h5 b3d5 c7c6 d5f3 h5e5 d1e2 e5e2 e1e2 g4h2 h1h2 c5d4 a2a4 b5b4 h2h4 d4f6 h4b4 a6a5 b4e4 e8d8 e4f4 f6g5 b1c3 c8a6 e2d1 g5f4 c1f4 d7d5 d1c2 h7h5 a1h1 h5h4 h1h4 h8h4 g3h4');
INSERT INTO evaluations VALUES(7698,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(7698,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(7698,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(7698,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(7698,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(7698,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(7698,36,'cp -32','a5c6 f3c6 c8d7 c6a6 e8g8 b1c3 h7h6 g5f3 e7c5 e1g1 f8e8 d2d4 c5d4 f3d4 e5d4 c3e2 d7g4 e2g3 f6h5 a6a4 h5g3 h2g3 g4e2 f1e1 d8f6 c1e3 d4e3 e1e2 e3f2 e2f2 f6b2 a1f1 b2b6 g1h2 e8e7 a4c4 b6b7 f2e2 b7c7 c4c7 e7c7 f1f5 a8c8 f5a5 g7g6 a5a4 c8d8 e2e1 d8b8 h2h3');
INSERT INTO evaluations VALUES(7700,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(7700,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(7700,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(7700,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(7700,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(7700,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(7700,36,'cp -41','f3c6 c8d7 c6f3 e8g8 d2d3 f6g4 g5h3 a8c8 b1c3 c8c6 f3e2 c6g6 f2f3 g4h6 c1h6 g6h6 e1g1 d7h3 g2h3 d8b6 g1h1 b6b2 c3d5 e7c5 a1e1 b2a2 c2c4 a2e2 e1e2 f8e8 f1b1 g8f8 h1g2 h6g6 g2h1 c5d4 e2a2 e8a8 f3f4 e5f4 d5f4 g6f6');
INSERT INTO evaluations VALUES(9348,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(9348,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(9348,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(9348,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9348,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9348,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9348,36,'cp -5','g1e2 a7a5 h2h3 c8e6 c4e6 f7e6 c1e3 d8e7 e3c5 d6c5 b1a3 e8c8 f3e3 h8f8 a3c4 f6h5 g2g3 f8f7 e1g1 d8f8 c4d2 b7b6 g1h2 e7d8 e2g1 g7g5 e3e2 h5f6 g1f3 h7h6 f1h1 f6d7 h2g2 a5a4');
INSERT INTO evaluations VALUES(9477,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(9477,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9477,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9477,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9477,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9477,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9477,36,'cp 0','d8e7 a2a3 c5d6 d1e2 c8e6 c1e3 f6g4 e3g5 g4f6');
INSERT INTO evaluations VALUES(13992,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(13992,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(13992,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(13992,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(13992,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(13992,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(13992,36,'cp 63','c1e3 f8e7 f1e2 e8g8 e1g1 c8d7 f2f3 a8c8 d4c6 d7c6 b2b4 h7h6 a1c1 b7b6 a2a3 d8c7 c3d5 c6d5 c4d5 c7d7 d5e6 d7e6 d1d4 c8c1 f1c1 f8c8 c1c8 e6c8 d4c4 c8d7 c4c3 d7e6 e3d4 f6d5 e4d5 e6e2 d4g7 e7f8 g7d4 e2b5');
INSERT INTO evaluations VALUES(14907,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(14907,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(14907,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(14907,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(14907,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(14907,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(14907,36,'cp 34','f8d6 b1d2 e6e5 c4b5 e5e4 f3e5 d6e5 d4e5 f6g4 e1g1 d8h4 b5c6 b7c6 h2h3 g4e5 d1a4 e8g8 a4e4 h4e4 d2e4 e5d3 f1d1 c8a6 b2b4 a8d8 e4g3 c6c5 d1d2 c5c4 c1b2 f8e8 b2c3 a6b7 f2f3 f7f6 e3e4 b7c8 a3a4 a7a6 g3e2 f6f5 e4f5 c8f5 e2d4 f5d7 b4b5 d8a8 a1d1 a6b5 a4b5 g8f7');
INSERT INTO evaluations VALUES(9274,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(9274,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(9274,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(9274,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(9274,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9274,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(9274,36,'cp 22','c4e2 e8g8 b2b4 d8e7 c1b2 e6e5 b4b5 e5e4 f3d2 c6a5 b2c3 b7b6 c3a5 b6a5 d2c4 a7a6 b1c3 a6b5 c3b5 c8d7 c4d6 c7d6 e1g1 f8b8 a1b1 b8b7 d1d2 a5a4 f1c1 a8b8 b5c3 b7b1 c1b1 b8b1 c3b1 e7e6 e2d1 e6c4 b1c3 h7h6 h2h3 g7g6 d2c2 g8g7 c2b2 d6d5');
INSERT INTO evaluations VALUES(15793,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(15793,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(15793,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(15793,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(15793,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(15793,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(15793,36,'cp 77','f8d6 e5f3 d8e7 e1f1 c8g4 c4e2 d4e2 d1e2 e8c8 d2d3 e7d7 h2h3 d8e8 e2d1 g4h5 c1g5 d6e5 f3e5 e8e5 d1d2 f6d5 c3d5 d7d5 g5f4 e5e7 a1e1 h8e8 e1e7 e8e7 f1g1 f7f6 c2c4 d5d4 f4e3 d4d6 d3d4 b7b6 h3h4 e7e4 h1h3 d6e7 h3g3 e4h4 c4c5 e7e4 f2f3 e4b1 g1f2 h4h1 g3g7 b6c5 g7g8 c8b7');
INSERT INTO evaluations VALUES(17369,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(17369,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(17369,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(17369,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(17369,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(17369,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(17369,36,'cp -16','g8f6 f1e1 f6e4 e1e4 d7d5 c4d5 d8d5 b1c3 d5h5 c3a4 b7b5 a4b6 a8b8 b6c8 b8c8 b2b3 c6d4 c1b2 d4f3 d1f3 h5f3 g2f3 f8d6 a3a4 b5b4 g1f1 h8g8 d2d3 e8d7 a1e1 g7g6 e4c4 g8d8 c4h4 h7h5 h4c4 d6e7 f3f4 d7c6 b2c1 d8d4 c1e3 d4c4 d3c4 c8d8 f1e2 e7f6 e2f3 f6d4 e1d1 f7f6 e3d4');
INSERT INTO evaluations VALUES(18911,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(18911,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18911,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(18911,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(18911,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(18911,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(18911,36,'cp 19','f8c5 d4b3 c5e7 c2c4 e7f6 f1e2 g8e7 b1c3 f6c3 b2c3 d7d5 e4d5 e6d5 e1g1 e8g8 c1f4 d5c4 e2c4 d8d1 f1d1 c8e6 c4e6 f7e6 f4g3 g8f7 b3c5 b7b6 c5e4 h7h6 e4d6 f7g8 d1e1 f8f6 a1d1 e7f5 f2f3 f5g3 h2g3 a8d8 d6e4 f6f8 d1d6 d8d6 e4d6 e6e5 g1f2 f8f6 d6e4 f6f8');
INSERT INTO evaluations VALUES(21387,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(21387,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(21387,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(21387,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(21387,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(21387,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(21387,36,'cp 55','g8f6 e2e3 f8e7 b1d2 e8g8 f1d3 a7a5 b2b3 c6b8 e1g1 b7b6 c1b2 c8b7 a1c1 b8d7 c4d5 e6d5 f3e5 c7c5 d3b5 a8c8 d1e2 c5d4 c1c8 d8c8 e5d7 f6d7 f1c1 c8d8 d2f3 d7c5 f3d4 e7f6 e2c2 d8e7');
INSERT INTO evaluations VALUES(23865,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(23865,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(23865,32,'cp -198',NULL);
INSERT INTO evaluations VALUES(23865,33,'cp -165',NULL);
INSERT INTO evaluations VALUES(23865,34,'cp -174',NULL);
INSERT INTO evaluations VALUES(23865,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(23865,36,'cp -200','c6d4 g1h1 c7c6 a2a4 d8f6 c1g5 f6g6 g5e3 g6h5 e3d4 e5d4 c3e2 g4f3 g2f3 h5f3 h1g1 g8f6 b2b4 f3g4 g1h1 c5b4 a1b1 g4f3 h1g1 f3h3 e2f4 h3h6 d1f3 a7a5 f4e2 e8c8 c4f7 f6d7 f3g4 c8b8 f7b3 d7f6 g4f4 g7g5 f4g3 b4c5 g1h1 d8f8 f2f4 g5f4 g3f4 h6h3 f4g3 h3g3 e2g3');
INSERT INTO evaluations VALUES(23895,30,'cp -166',NULL);
INSERT INTO evaluations VALUES(23895,31,'cp -179',NULL);
INSERT INTO evaluations VALUES(23895,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(23895,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(23895,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(23895,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(23895,36,'cp -163','g1h1 c7c6 a2a4 d8f6 c1g5 f6g6 g5e3 g6h5 e3d4 e5d4 c3e2 g4f3 g2f3 h5f3 h1g1 d6d5 c4b3 e8c8 e4d5 c6d5 c2c3 d4c3 e2c3 f3f5 d3d4 c5b4 a1c1 c8b8 d1e2 g8e7 e2e5 f5e5 d4e5 d5d4 c3e4 d4d3 c1c4 e7c6 e4g5 d8d4 g5f7 h8e8 f1d1 d3d2 c4d4 c6d4');
INSERT INTO evaluations VALUES(28595,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(28595,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(28595,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(28595,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(28595,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(28595,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(28595,36,'cp 121','e2e4 a6f1 e1f1 g7g6 g2g3 f8g7 f1g2 e8g8 g1f3 d8b6 h1e1 b8a6 a1b1 f6d7 c1f4 c5c4 d1a4 a6c5 a4c4 b6a6 c4a6 a8a6 e4e5 c5d3 e1e3 d7e5 f3e5 d3e5 b1c1 g7f6 b2b3 a6a3 c1c2 f8a8 h2h3 g6g5 f4e5 f6e5 e3e2 e5c3');
INSERT INTO evaluations VALUES(28593,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(28593,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(28593,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(28593,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(28593,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(28593,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(28593,36,'cp 108','a6f1 e1f1 g7g6 g2g3 f8g7 g1f3 e8g8 f1g2 b8a6 c1g5 d8b6 a1b1 h7h6 g5f6 g7f6 h1e1 f6g7 a2a3 a6c7 h2h4 c7b5 c3b5 b6b5 h4h5 g6g5 e1e3 b5d7 e4e5 d7f5 e5d6 e7d6 b2b4 c5c4 b1c1 g5g4 f3h4');
INSERT INTO evaluations VALUES(28592,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(28592,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(28592,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(28592,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(28592,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(28592,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(28592,36,'cp 123','e1f1 g7g6 g2g3 f8g7 f1g2 e8g8 g1f3 b8a6 c1g5 d8b6 a1b1 h7h6 g5f6 g7f6 h2h4 c5c4 d1e2 a6b4 e2c4 f8c8 c4b3 a8b8 h4h5 g6g5 a2a3 b4d3 b3b6 b8b6 c3d1 b6b3 d1e3 g8f8 h1d1 d3b2 d1c1 c8c1 b1c1 f6c3 e3f5');
INSERT INTO evaluations VALUES(28591,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(28591,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(28591,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(28591,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(28591,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(28591,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(28591,36,'cp 109','g7g6 g2g3 f8g7 g1f3 e8g8 f1g2 d8c8 h1e1 c8b7 d1e2 b8a6 c1g5 f8b8 a1b1 h7h6 g5f6 g7f6 e2d2 f6g7 b2b3 a6c7 b1c1 a8a6 h2h4 h6h5 c1c2 a6b6 d2d1 b6b4 e4e5 d6e5 f3e5 e7e6 d1f3 c7d5 c3d5 e6d5 c2c5');
INSERT INTO evaluations VALUES(30604,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(30604,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(30604,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(30604,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(30604,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(30604,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(30604,36,'cp 140','b1c3 g7g6 c1e3 g8f6 e1g1 f8g7 f1e1 d7d6 d1d2 b8c6 h2h3 c6e7 d3b5 f6d7 a1d1 a7a6 b5d3 d7f6 a2a4 d6d5 e3f4 d5e4 c3e4 b7e4 d3e4 f6e4 e1e4 b6b5 d4d5 d8d5 d2d5 e6d5 d1d5 b5a4 d5d1 g7b2 d1e1 e8g8 e4e7');
INSERT INTO evaluations VALUES(30955,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(30955,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(30955,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(30955,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(30955,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(30955,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(30955,36,'cp 127','g1h3 e7e5 c1e3 b8d7 f1d3 e5d4 e3d4 d8h4 d3g6 f7g6 d1e2 h4e7 e1c1 e8c8 e2d2 g8f6 h3g5 d7c5 h1e1 e7d7 b2b4 c5a6 d4f6 d7d2 d1d2 d8d2 e1e8 c8c7 g5e6 c7d7 e8f8 d7e6 f8h8 d2f2 f6c3 f2g2 g3e2 g2g4 a2a3 c6c5 b4c5 a6c5 c3g7 b7b6 h8a8 h5h4 g7d4 g6g5 a8a7 g4g2 c1d2 e6d5 d4g1 c5e4 d2d3 e4f2 g1f2 g2f2');
INSERT INTO evaluations VALUES(31735,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31735,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31735,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31735,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31735,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31735,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31735,36,'cp 0','b1c3 b8c6 a2a3 a7a6 c1e3 f7f6 g1f3 f6e5 f3e5 c6e5 d4e5 g8e7 g2g4 f5g6 f2f4 h7h5 f4f5 e6f5 g4g5 f5f4 e3f4 d5d4 d1a4 e7c6 f1g2 d4c3 g2c6 b7c6 a4c6 e8f7 c6c4 f7e7 c4c5 e7f7');
INSERT INTO evaluations VALUES(31734,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31734,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31734,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31734,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31734,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31734,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31734,36,'cp 0','b8c6 a2a3 a7a6 c1e3 f7f6 g1f3 f6e5 f3e5 c6e5 d4e5 g8e7 g2g4 f5g6 f2f4 h7h5 f4f5 e6f5 g4g5 f5f4 e3f4 d5d4 d1a4 e7c6 f1g2 d4c3 g2c6 b7c6 a4c6 e8f7 c6c4 f7e7 c4c5 e7f7');
INSERT INTO evaluations VALUES(26437,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26437,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26437,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26437,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26437,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26437,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26437,36,'cp -8','a2a3 a7a6 c1e3 f7f6 g1f3 f6e5 f3e5 c6e5 d4e5 g8e7 g2g4 f5g6 f2f4 h7h5 f4f5 e6f5 g4f5 e7f5 e3f2 d5d4 d1a4 d8d7 a4d7 e8d7 c3e2 a8e8 e1c1 e8e5 e2d4 f5d4 f2d4 e5e8 h1g1 g6f5 d4g7 d7c7 f1d3 f5d3 g7f8 e8f8 d1d3 h8h7 d3d2 c7b8 g1g5 f8f1 c1c2 b8a7 c2b3 f1h1 b3a2');
INSERT INTO evaluations VALUES(32067,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(32067,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(32067,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(32067,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(32067,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(32067,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(32067,36,'cp 13','c7c6 b5e2 b8d7 d2d4 g8f6 c1e3 f8e7 a2a4 e8g8 e1g1 g4h5 f3d2 h5g6 f2f3 b7b6 f3f4 e5f4 e3f4 f8e8 e2d3 d7f8 d1f3 d6d5 e4e5 f6d7 c3e2 f7f6 e5f6 d7f6 e2g3 g6d3 f3d3 f8g6 f4g5 f6d7 g5e7 d8e7 a1d1 a8d8 g3f5');
INSERT INTO evaluations VALUES(31584,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31584,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(31584,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(31584,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(31584,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31584,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31584,36,'cp 24','b5c4 g4h5 h2h3 b7b5 c4b3 g8f6 d2d3 b8d7 g2g4 h5g6 h3h4 h7h6 h4h5 g6h7 g4g5 h6g5 c1g5 h7g6 h5h6 g7h6 d1e2 h6g5 h1h8 f6h5 f3d2 h5f4 e2g4 d8f6 h8h2 e8c8 e1c1 c8c7 c1b1 d7c5 d1g1 f8e7 a2a4 c5b3 d2b3 b5a4 c3a4 d8b8 g4f3 d6d5 e4d5 f4d3 f3f6 e7f6');
INSERT INTO evaluations VALUES(32627,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(32627,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(32627,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(32627,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32627,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(32627,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(32627,36,'cp 82','d2d4 e7e6 c1d2 f8b4 a2a3 b4c3 d2c3 a5b6 d1e2 e8g8 e1c1 b8d7 h1g1 a7a6 f3e5 d7e5 d4e5 f6d5 c4d5 e6d5 d1d5 b6h6 c1b1 h6h2 d5d1 a8d8 e2f3 d8d1 g1d1 f5c8 e5e6 f7e6 f3g4 h2h6 g4g3 e6e5 c3e5 h6g6 g3c3 g6f7 f2f3 c8f5 d1d2 f8c8');
INSERT INTO evaluations VALUES(32629,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(32629,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(32629,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(32629,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(32629,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(32629,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(32629,36,'cp 54','e7e6 c1d2 f8b4 a2a3 b4c3 d2c3 a5b6 d1e2 e8g8 e1c1 b8d7 h1g1 f6d5 g2g4 d5c3 b2c3 f5g6 h2h4 b6c6 c4b5 c6e4 h4h5 a7a6 b5c4 e4e2 c4e2 g6e4 f3d2 e4c6 g4g5 a8d8 d2b3 b7b5 f2f4 c6d5 g1g3 f8e8 e2d3 g7g6 d1e1 g8f8 h5h6 f8e7 e1e3 c7c6 b3a5');
INSERT INTO evaluations VALUES(32697,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(32697,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(32697,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(32697,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32697,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(32697,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(32697,36,'cp 51','c1d2 f8b4 a2a3 b4c3 d2c3 a5b6 d1e2 b8d7 e1c1 e8g8 h1g1 f6d5 g2g4 d5c3 b2c3 f5g6 h2h4 b6c6 c4b5 c6e4 h4h5 a7a6 h5g6 e4f4 c1b2 a6b5 g6h7 g8h8 f3e5 d7e5 e2e5 f4e5 d4e5 h8h7 d1d7 c7c6 d7b7 a8a4 g1g3 g7g5 b7c7 a4f4 f2f3 f4c4');
INSERT INTO evaluations VALUES(33690,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(33690,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(33690,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(33690,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(33690,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(33690,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(33690,36,'cp 38','b1c3 d5d6 g2g3 c8f5 g1f3 b8c6 f1g2 d6d7 d1e2 e7e6 e1g1 f5g4 e2e3 f8e7 h2h3 g4f3 e3f3 e8g8 c3e4 a8d8 e4f6 e7f6 f3e4 d7d4 f1d1 d4e4 g2e4 c6d4 c1d2 d8d7 c2c3 d4f5 g3g4 f5d6 g1f1 d6e4 d3e4 f8d8 f1e2 b7b6 d2e3 d7d1 a1d1 d8d1 e2d1 e6e5 d1e2 f6e7 b2b4 f7f6 a2a4 a7a6');
INSERT INTO evaluations VALUES(33688,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(33688,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(33688,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(33688,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(33688,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(33688,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(33688,36,'cp 26','d5e5 c1e3 f6g4 d1d2 g4e3 f2e3 g7g6 d3d4 e5a5 f1c4 f8g7 c3d5 a5d2 e1d2 e8d8 a1f1 c8e6 c4b3 b8d7 g1f3 h7h6 h2h4 a7a5 d5f4 e6b3 a2b3 e7e5 d4e5 g7e5 f4d3 e5g7 h4h5 h8e8 h5g6 f7g6 f3h4 d7f6 h4g6 a5a4 b3a4 f6e4 d2d1 a8a4');
INSERT INTO evaluations VALUES(35692,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(35692,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(35692,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(35692,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(35692,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(35692,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(35692,36,'cp 62','e2e4 d7d6 g1e2 a7a6 e1g1 e7e5 d1c2 b8d7 a2a4 f8e8 d4d5 a6a5 h2h3 d7c5 c1e3 c6d5 c4d5 c8d7 b2b3 d8c8 c3b5 d7b5 a4b5 c8d7 e2c3 e8c8 a1c1 c5e4 g2e4 c8c3 c2c3 f6e4 c3c4 f7f5 g1h2');
INSERT INTO evaluations VALUES(35696,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(35696,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(35696,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(35696,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(35696,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(35696,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(35696,36,'cp 49','d7d6 g1e2 a7a6 e1g1 e7e5 d4e5 d6e5 c3a4 a6a5 e2c3 b8a6 c1e3 c8g4 f2f3 g4e6 d1e2 d8e7 b2b3 f8e8 e2f2 g7f8 f1d1 e8d8 g2f1 d8d1 a1d1 f6d7 h2h4 f7f6 d1d2 a6c5 h4h5 g6h5 f2h2 e6f7 f1e2 c5a4');
INSERT INTO evaluations VALUES(35914,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(35914,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(35914,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(35914,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(35914,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(35914,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(35914,36,'cp 66','g1e2 a7a6 e1g1 e7e5 d4e5 d6e5 c3a4 a6a5 e2c3 b8a6 c1e3 c8e6 d1e2 d8e7 a1d1 e6g4 f2f3 g4e6 f1e1 f8d8 g2f1 h7h5 e2f2 d8d1 e1d1 g7f8 b2b3 h5h4 g3h4 f6h5 e3g5 f7f6 g5e3 a8d8 d1d8 e7d8 e3b6 d8b8 b6a5 h5f4');
INSERT INTO evaluations VALUES(35913,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(35913,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(35913,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(35913,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(35913,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(35913,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(35913,36,'cp 54','a7a6 e1g1 e7e5 d4e5 d6e5 c3a4 a6a5 e2c3 b8a6 b2b3 d8e7 d1e2 f8d8 c1e3 c8g4 f2f3 g4e6 a1e1 d8e8 g1h1 h7h5 e3g5 e7b4 h2h3 a6c5 a4c5 b4c5 c3a4 c5b4 e1d1 b7b5 c4b5 c6b5 a4b6 a8d8 h1h2 d8d1');
INSERT INTO evaluations VALUES(39331,30,'cp 326',NULL);
INSERT INTO evaluations VALUES(39331,31,'cp 344',NULL);
INSERT INTO evaluations VALUES(39331,32,'cp 369',NULL);
INSERT INTO evaluations VALUES(39331,33,'cp 361',NULL);
INSERT INTO evaluations VALUES(39331,34,'cp 363',NULL);
INSERT INTO evaluations VALUES(39331,35,'cp 360',NULL);
INSERT INTO evaluations VALUES(39331,36,'cp 370','c2c3 b4f8 f3e5 d7d6 e5f3 f8e7 f1d3 b8d7 h2h3 g8f6 e1g1 d7f8 c3c4 f8g6 b1c3 e8g8 f3e1 f8e8 c1e3 b7b6 f2f4 c7c5 d4d5 c8b7 b2b3 a7a6 a1b1 d8c7 e1f3 f6h5 d1d2 e7f6 f4f5 f6c3 d2c3');
INSERT INTO evaluations VALUES(39329,30,'cp 354',NULL);
INSERT INTO evaluations VALUES(39329,31,'cp 348',NULL);
INSERT INTO evaluations VALUES(39329,32,'cp 347',NULL);
INSERT INTO evaluations VALUES(39329,33,'cp 354',NULL);
INSERT INTO evaluations VALUES(39329,34,'cp 365',NULL);
INSERT INTO evaluations VALUES(39329,35,'cp 362',NULL);
INSERT INTO evaluations VALUES(39329,36,'cp 362','b4f8 f3e5 d7d6 e5f3 f8e7 f1d3 g8f6 h2h3 e8g8 e1g1 c7c5 d4d5 f6d7 a2a4 f7f6 f3h4 d7e5 d3c2 g7g5 h4f5 g8h8 f1e1 b8d7 b1d2 f8g8 d2f1 c5c4 f1g3 d7c5 c1e3 c5d3 c2d3 e5d3 e1e2 d3f4 f5h6');
INSERT INTO evaluations VALUES(44916,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44916,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44916,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(44916,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(44916,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44916,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44916,36,'cp 11','f1g2 g7g6 g1f3 f8g7 c4d5 c6d5 b1c3 e8g8 f3e5 c8f5 e1g1 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 f1c1 c3c1 a1c1 d8b6 e2e3 f5d3 h2h4 d3c4 g2f1 b6b2 c1d1 c4a2 a4a7 b2c2 d1e1 h7h5 a7c5 f8c8 c5c2');
INSERT INTO evaluations VALUES(45889,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(45889,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(45889,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(45889,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(45889,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(45889,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(45889,36,'cp 135','g1f3 f8e7 f1d3 b8d7 e1g1 e6e5 h2h3 a7a6 c1e3 b7b5 c4b5 a6b5 a2a4 b5a4 d4e5 d7e5 f3e5 d6e5 c3a4 e8g8 d1c2 c8e6 a4c5 e7c5 e3c5 f8e8 b2b4 d8c7 c5e3 c7b7 c2c3 a8a1 f1a1 h7h6 a1a6 f6e4 c3c6 b7c6 a6c6');
INSERT INTO evaluations VALUES(51534,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(51534,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(51534,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51534,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(51534,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(51534,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(51534,36,'cp 39','a7a6 b1d2 e7e6 g1f3 g8f6 e3e4 c8b7 e4e5 f6d5 a2a4 b5a4 f1d3 d7d6 d2c4 d5b4 d3e2 b7e4 c4e3 d6e5 f3e5 b8c6 e2f3 e4f3 e5f3 f8d6 c2c3 b4d5 d1a4 d8d7 a4c2 e8g8 e3d5 e6d5 f3g5 g7g6 e1g1 f8e8 g5f3 c6e7 h2h3 e7f5 c1d2 e8e2 f1e1');
INSERT INTO evaluations VALUES(51793,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(51793,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51793,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(51793,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(51793,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(51793,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(51793,36,'cp 33','g8f6 e2e3 d7d5 c2c4 f8d6 f4d6 d8d6 b1c3 e8g8 d1c2 d6e7 a1c1 b8d7 h2h3 h7h6 c4d5 e6d5 f1d3 f8e8 e1g1 f6e4 f1e1 d7f6 f3e5 c8f5 c3a4 f5h7 f2f3 e4d6 a4c5 h7d3 c2d3 a7a5 a2a4 a8c8 e3e4 d5e4 f3e4');
INSERT INTO evaluations VALUES(51938,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(51938,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(51938,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(51938,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(51938,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(51938,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(51938,36,'cp 11','d7d5 c4d5 e6d5 f1b5 b8c6 g1f3 g8f6 d2d4 c5d4 e3d4 f8d6 c1g5 e8g8 e1g1 c8g4 g5f6 d8f6 c3d5 f6h6 h2h3 g4f3 d1f3 c6d4 f3d3 d4b5 d3b5 b7b6 a1d1 f8e8 b5d7 h6e6 d5b6 a7b6 d7d6 e6a2 d6d4 a2a5 d4d5 g8f8 d5f3 a5e5 f1e1 e5b2');
INSERT INTO evaluations VALUES(54163,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(54163,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(54163,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54163,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(54163,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(54163,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(54163,36,'cp 21','d2d4 g8f6 g1f3 a7a6 c4d5 e6d5 f1e2 c5d4 e3d4 h7h6 c1f4 b8c6 f3e5 f8d6 e1g1 e8g8 e2f3 c6e7 c3a4 c8f5 d1b3 e7c6 e5c6 b7c6 f4d6 d8d6 b3c3 f8e8 a1c1 a6a5 a4c5 f6e4 f3e4 d5e4 f1e1 d6d5 b2b3');
INSERT INTO evaluations VALUES(55872,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(55872,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(55872,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(55872,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(55872,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(55872,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55872,36,'cp 0','a7a6 g1f3 c7c5 d4c5 f8c5 c2c4 d5c4 d1d8 e8d8 f3e5 d8e7 e5c4 b7b5 c4a5 c8d7 g2g4 c5b6 c1d2 b8c6 f1g2 b6a5 d2a5 c6a5 g2a8 a5b3 a1a2 g8f6 a8g2 b3c1 a2a1 c1b3');
INSERT INTO evaluations VALUES(56864,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(56864,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(56864,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(56864,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(56864,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(56864,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(56864,36,'cp 74','f7f5 b1c3 g8f6 g2g3 f8g7 f1g2 e8g8 e1g1 d7d6 f1e1 f6e4 d1c2 d6d5 b2b3 b7b6 c4d5 e6d5 c1f4 c8b7 f4e5 d8e7 a1c1 g7e5 f3e5 c7c5 e2e3 b8a6 a2a3 a8d8 f2f3 e4d6 c3e2 a6c7 e2f4 c7e6 f4e6 e7e6');
INSERT INTO evaluations VALUES(103,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(103,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(103,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(103,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(103,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(103,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(103,36,'cp 35','d1d2 a7a5 f1d3 g8e7 e1g1 e8g8 d2e2 e7g6 c1e3 c5e3 e2e3 f6f4 e3d4 f4e5 c3e2 e5d4 e2d4 g6f4 f1d1 c6c5 d4f5 g7g6 f5e3 d7d6 d3c4 g8g7 g2g3 f4e6 g1g2 c8b7 f2f3 f8b8 c2c3 a5a4 d1d2');
INSERT INTO evaluations VALUES(5308,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(5308,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(5308,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(5308,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(5308,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(5308,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(5308,36,'cp -14','c2c3 e7e6 c4b5 d8b6 b5a4 g8e7 e1g1 d7d5 e4e5 e8g8 f1e1 c5c4 d2d3 c4d3 d1d3 f7f6 e5f6 g7f6 d3b5 b6b5 a4b5 e6e5 c3c4 e5e4 c4d5 e4f3 d5c6 b7c6 b5f1 e7f5 b1d2 f5d4 d2f3 d4c2 c1h6 c2a1 h6f8 g8f8 e1a1');
INSERT INTO evaluations VALUES(5304,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(5304,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(5304,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(5304,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(5304,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(5304,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(5304,36,'cp -21','e7e6 c4a2 d7d5 e4d5 e6d5 d2d4 c5d4 f3d4 c6d4 c3d4 g8e7 b1c3 e8g8 e1g1 c8e6 c1g5 f8e8 d1d2 d8b6 f1e1 e7f5 a2d5 e6d5 c3d5 b6d6 d5c3 f5d4 a1d1 d6c6 h2h3 h7h5 e1e3 e8e3 f2e3 d4f5 e3e4 g7c3 d2c3 c6e4 d1e1 e4d4 c3d4 f5d4');
INSERT INTO evaluations VALUES(6058,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(6058,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(6058,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(6058,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(6058,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(6058,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(6058,36,'cp 79','c3d5 b4a5 e1g1 g8f6 c2c3 e8g8 b2b4 a5b6 a2a4 a7a6 d2d4 h7h6 d5b6 c7b6 d4d5 c6e7 c4d3 e7g6 f1e1 c8g4 h2h3 g4h5 c3c4 d8d7 g1h2 a6a5 b4a5 a8a5 c1d2 h5f3 d1f3 a5a4 d2h6 f8a8 a1b1 d7c7 h6e3 f6d7 f3g4 d7c5 d3e2 g6f4 e2f1 a4a2 g2g3');
INSERT INTO evaluations VALUES(7655,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7655,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(7655,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(7655,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(7655,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(7655,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(7655,36,'cp -6','f1c4 h7h6 g5f6 d8f6 e1g1 e8g8 d1d3 f8d8 a1d1 c8b7 a2a3 b4a5 b2b4 a5b6 d3h3 b7c8 h3h5 d7d6 d1d3 c8e6 c4e6 f6e6 f1d1 a8b8 h2h3 a7a5 c3e2 d6d5 e2g3 a5b4 a3b4 d5e4 d3d8 b6d8 c2c3 e4e3 f2e3 d8b6 h5f3 b8d8 d1d8 b6d8 g3f5 g8h7 f3g4 d8g5 h3h4');
INSERT INTO evaluations VALUES(7656,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(7656,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(7656,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(7656,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(7656,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(7656,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(7656,36,'cp -33','h7h6 g5f6 d8f6 d1d3 f6d6 d3f3 e8g8 a2a3 b4a5 e1g1 d6d4 c4a2 a5c3 f1d1 d4c5 f3c3 c5c3 b2c3 f8d8 d1d6 g8f8 f2f3 a8b8 a1d1 b8b6 a2c4 c8a6 c4a6 b6a6 d6d7 d8d7 d1d7 a6a3 d7c7 a3c3 c7a7 c3c2 h2h4 g7g6 a7d7 c2c1 g1f2 c6c5 d7c7');
INSERT INTO evaluations VALUES(8775,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(8775,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(8775,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8775,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(8775,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(8775,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(8775,36,'cp 0','a2a4 a7a5 b4b5 c6e7 c1e3 e8g8 e3b6 c7b6 d1b3 e7g6 b1d2 d8e7 a1e1 c8d7 d3d4 a8c8 e1e3 f6g4 e3e1');
INSERT INTO evaluations VALUES(8781,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8781,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8781,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8781,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(8781,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8781,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8781,36,'cp 0','a7a5 b4b5 c6e7 c1e3 e8g8 e3b6 c7b6 b1d2 e7g6 d1b3 d8d7 a1e1 d7g4 g1h1 g4h5 b3a3 f8d8 a3a2 d8f8');
INSERT INTO evaluations VALUES(10171,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(10171,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(10171,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(10171,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(10171,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(10171,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(10171,36,'cp -13','d7d6 f1e2 a7a5 e1g1 e8g8 d2d3 c8g4 f3d2 g4e6 d2f3 c6d4 f3d4 c5d4 c3b5 d4b6 g1h1 c7c6 b5c3 b6d4 f2f4 e5f4 c1f4 f8e8 a1b1 f6d7 d1e1 a5a4 h2h3 d7c5 e1g3 d4e5 b1d1 c5b3 f4e5 d6e5 g3e5 e6c4 e5g3 c4e6 h1h2 b3c5');
INSERT INTO evaluations VALUES(14662,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(14662,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(14662,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(14662,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(14662,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(14662,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(14662,36,'cp 71','f3d4 f8e7 d4c6 b7c6 c4d3 e8g8 f2f4 a7a5 d1f3 a5a4 h2h3 c8a6 c1e3 f8e8 a1d1 f6d7 e4e5 a6d3 d1d3 d6e5 f3c6 e7d6 c3b5 e5f4 b5d6 c7d6 e3f4 d7e5 f4e5 e8e5 c6d6 d8d6 d3d6 f7f6 c2c4 a8b8 f1b1 e5e2 b2b4 a4b3 a2b3');
INSERT INTO evaluations VALUES(15153,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(15153,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(15153,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(15153,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(15153,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(15153,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(15153,36,'cp 38','e4d5 f6d5 d3b5 f8d6 d2d4 e5d4 c1g5 f7f6 f3d4 d5e7 g5e3 e8g8 d4c6 e7c6 b1d2 c8e6 d2e4 a7a6 b5e2 d6e7 e4c5 e6f7 d1d8 a8d8 e2f3 e7c5 e3c5 f8e8 f1d1 d8d1 a1d1 f7a2 b2b3 a6a5 h2h3 a5a4 f3d5 g8h8 d5c6 b7c6 b3a4 e8a8 d1d7 a2d5 d7c7');
INSERT INTO evaluations VALUES(15010,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(15010,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(15010,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(15010,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(15010,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(15010,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(15010,36,'cp 37','f6d5 b1c3 c8e6 f1e1 f7f6 d3f1 d8d7 d2d4 e5d4 c3b5 e8c8 f3d4 d5b6 d4c6 b7c6 d1d7 e6d7 b5d4 c6c5 f1a6 c8b8 d4b3 c5c4 b3d4 f8c5 c1e3 h8e8 c2c3 h7h5 a2a4 b6a4 a6c4 c5b6 c4b3 a4c5 b3c2 c5e6 d4e6 e8e6');
INSERT INTO evaluations VALUES(15354,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(15354,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(15354,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(15354,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(15354,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(15354,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(15354,36,'cp 53','b1c3 c8e6 f1e1 f7f6 d3f1 d8d7 d2d4 e5d4 c3b5 e8c8 f3d4 d5b6 c1e3 e6c4 d4c6 b7c6 d1d7 d8d7 b5d4 c6c5 d4b3 c4f1 g1f1 d7d5 a1d1 d5d1 e1d1 b6a4 e3c1 f8d6 b3a5 h8e8 h2h3 e8e6 d1e1 c8d7 e1e6 d7e6 f1e2 a4b6 e2d3 h7h5 b2b3 a7a6 g2g3 b6d7 a5c4 f6f5 c1e3');
INSERT INTO evaluations VALUES(16802,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(16802,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(16802,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(16802,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(16802,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(16802,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(16802,36,'cp 93','c3d4 d7d6 h2h3 g4h6 c1h6 g7h6 f1b5 f8g7 d4d5 a7a6 b5a4 b7b5 d5c6 b5a4 d1a4 c8e6 b1c3 e8g8 a1d1 d8b8 e1g1 d6e5 b2b3 a6a5 c3d5 g8h8 d1d2 f7f5 f1d1 b8a7 a4b5 a8e8 f3h4 f5f4 b5e2 e6g8 e2e4 g8e6 g1h1 e8d8 e4c2');
INSERT INTO evaluations VALUES(16801,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(16801,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(16801,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(16801,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(16801,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(16801,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(16801,36,'cp 85','d7d6 h2h3 g4h6 c1h6 g7h6 f1b5 f8g7 d4d5 a7a6 b5a4 b7b5 d5c6 b5a4 d1a4 d6e5 b1c3 c8e6 e1g1 e8g8 a1d1 d8b8 b2b3 a6a5 f1e1 b8b4 c3b5 a8a6 d1c1 f8c8 f3e5 e6d5 b5c3 b4a4 c3a4 g7e5 e1e5 d5c6 c1c5 c6a4 b3a4 c7c6 c5a5 a6a5 e5a5');
INSERT INTO evaluations VALUES(17401,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(17401,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(17401,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(17401,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(17401,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(17401,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(17401,36,'cp -59','d7d5 e4d5 e6d5 c4b3 d8e7 d1e2 d5d4 d2d3 c8f5 b3c4 g8f6 c3d4 c5d4 b1d2 e8c8 e2e7 f8e7 f3h4 f5e6 c4e6 f7e6 h4f3 h8e8 e1g1 e6e5 d2c4 e7b4 f1d1 e5e4 d3e4 f6e4 c1f4 b4c5 c4e5 g7g5 e5c6 b7c6 f3g5 d4d3 g5e4 e8e4 a1c1 d8d5 f4e3 c5e3 f2e3 c6c5 b2b3 e4e3');
INSERT INTO evaluations VALUES(17625,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(17625,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(17625,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(17625,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(17625,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(17625,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(17625,36,'cp -43','e4d5 e6d5 c4b3 d8e7 d1e2 d5d4 d2d3 c8f5 b3c4 e8c8 c3d4 c5d4 b1d2 g8f6 e2e7 f8e7 f3h4 f5e6 c4e6 f7e6 h4f3 e6e5 e1g1 h8e8 d2c4 e5e4 d3e4 f6e4 f1e1 e7b4 c1f4 b7b5 a4b5 a6b5 c4d2 e4d2 f4d2 b4d2 e1e8 d8e8 f3d2 c8b7 g1f1 b7b6 d2f3');
INSERT INTO evaluations VALUES(18999,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18999,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18999,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18999,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(18999,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(18999,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(18999,36,'cp -9','f1c4 f8e7 b1c3 d7d6 d2d3 a7a6 a2a4 g8f6 e1g1 e8g8 c1d2 c8e6 f1e1 d8d7 h2h3 a8e8 c3d5 e7d8 d5f6 d8f6 a1b1 e8b8 c4e6 d7e6 b3b4 b7b5 a4b5 a6b5 f3h2 f6e7 h2f1 d6d5 e4d5 e6d5 b4c5 e7c5 d2e3 c5b4 e3d2 b4d2 d1d2 d5d6');
INSERT INTO evaluations VALUES(19688,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(19688,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(19688,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(19688,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(19688,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(19688,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(19688,36,'cp 50','g8f6 f1e2 c6d4 d1d4 f8g7 e1g1 e8g8 d4d3 c8e6 c1e3 f6d7 b2b3 d8a5 e3d4 a8c8 d4g7 g8g7 f2f4 f7f6 a1b1 a7a6 a2a4 a5b4 h2h3 e6f7 e2d1 e7e6 g1h1 g6g5 d3g3 g7h8 f4f5 e6f5 d1c2 b4c5 f1f5');
INSERT INTO evaluations VALUES(22908,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(22908,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(22908,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(22908,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(22908,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22908,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22908,36,'cp 0','e3d4 d7d5 b1c3 g8f6 c4c5 f8c5 d4c5 d5d4 d1e2 c8e6 c3e4 d8e7 e4f6 g7f6 f2f3 c6b4 e2b5 e6d7 b5e2');
INSERT INTO evaluations VALUES(26559,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(26559,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(26559,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(26559,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(26559,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(26559,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(26559,36,'cp 24','g1f3 e7e6 f1b5 f8d6 c3c4 d5c4 b1d2 g8f6 d2c4 e8g8 b5c6 b7c6 e1g1 f5g4 f1e1 f6e4 h2h3 g4h5 d1c2 h5f3 g2f3 d8g5 g1f1 e4f6 b2b3 f6d5 f1e2 d5b4 c2e4 g5e7 c1d2 f7f5 e4b1 b4d5 e1g1 c6c5 d4c5 d6c5 b1c2 e6e5 d2c3 d5c3 c2c3 e5e4 c3e5');
INSERT INTO evaluations VALUES(29628,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(29628,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(29628,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(29628,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(29628,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(29628,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(29628,36,'cp 141','d1b3 e7e6 h2h3 g4f5 b3b7 b8d7 c4c5 f8e7 e2e3 e8g8 g1f3 a8b8 b7a6 f6e4 f1e2 h7h5 c3e4 d5e4 f3d2 e6e5 d2c4 f5e6 e1g1 e7f6 c4a5 e6d5 e2c4 d5a8 c5c6 b8b6 a6a7 a8c6 d4d5 c6b5 c4b5 b6b5 a5c6 d8a8 a2a4 a8a7 c6a7 b5d5 b2b4 f8b8 a7c6 b8b6 b4b5 d7c5 a4a5 b6b5 a5a6 g8h8 a6a7');
INSERT INTO evaluations VALUES(30734,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(30734,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(30734,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(30734,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(30734,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(30734,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(30734,36,'cp 125','e7e6 g1f3 f8g7 d4d5 e6d5 c4d5 b7d5 c3d5 g8f6 c1g5 h7h6 g5h4 g6g5 h4g3 d7d6 f3d4 f6d5 d4f5 g7b2 d1d5 b2a1 d5a8 a1c3 e1e2 c3f6 h1d1 e8g8 f5h6 g8h7 h6g4 f6g7 a8d5 b8d7 f2f3 d8e7 g4e3 d7e5 h2h4 g5h4 d1h1 e5g6 d5h5 h7g8 g3h4');
INSERT INTO evaluations VALUES(34552,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(34552,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(34552,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(34552,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(34552,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(34552,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(34552,36,'cp 220','d2d4 g8f6 g1f3 g7g6 f3g5 e6b6 e2e4 c8g4 d1c2 h7h6 g5f3 f8g7 c1e3 e8g8 h2h3 g4f3 g2f3 c7c5 d4c5 b6a5 e1c1 f6d7 c5c6 b7c6 f3f4 d7c5 e4e5 b8a6 a2a3 e7e6 f1a6 c5a6 c3e4 f8d8 h3h4 h6h5 d1d8 a5d8 h1g1 a8b8 e4d6');
INSERT INTO evaluations VALUES(37995,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(37995,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(37995,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(37995,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(37995,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(37995,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(37995,36,'cp 58','b8c6 c1e3 g7d4 e3d4 c6d4 d1d4 g8f6 f1e2 e8g8 e1g1 c8d7 b2b3 d7c6 a2a3 d8a5 f1d1 h7h5 c3d5 c6d5 e4d5 a5b6 d4b2 f8e8 e2f1 e7e5 d5e6 e8e6 b3b4 a7a5 a1b1 f6e4 d1e1 a8e8 g2g3 a5b4 a3b4 e4f6 e1e6 e8e6 b1d1 b6c7');
INSERT INTO evaluations VALUES(38482,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(38482,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(38482,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(38482,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(38482,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(38482,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(38482,36,'cp 79','g1f3 f7f6 c1f4 g8h6 c2c4 e8g8 b1c3 c7c6 f1e2 g6g5 f4g3 c8e6 e5f6 e7f6 d1b3 d8b6 b3b6 a7b6 c4d5 c6d5 c3b5 b8a6 e1g1 h6f5 a2a3 e6d7 f1d1 f5g3 h2g3 d7b5 e2b5 a6c7 b5d3 f6f5 d1e1 g5g4 f3h4 g7d4 h4f5');
INSERT INTO evaluations VALUES(39362,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39362,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39362,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(39362,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(39362,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(39362,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39362,36,'cp -12','g8f6 g1f3 b4c3 c2c3 d7d6 d2d3 e8g8 e2e3 c8g4 f1e2 c7c5 e1g1 b8c6 h2h3 g4h5 b2b3 d6d5 c4d5 f6d5 c3c2 f8e8 c1b2 b7b6 a1d1 d8d6 d1d2 a8c8 c2c4 f7f6 c4h4 h5g6 a2a3 c6a5 h4a4 a7a6 f1e1');
INSERT INTO evaluations VALUES(42774,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42774,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(42774,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(42774,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42774,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42774,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(42774,36,'cp 0','g1f3 b4c3 c2c3 d7d6 d2d4 e5e4 f3g1 h7h6 e2e3 e8g8 g1e2 d6d5 e2f4 c7c6 c1d2 d5c4 f1c4 b8d7 c3a5 d7b6 b2b3 f6d5 c4d5 b6d5 f4d5 c6d5 a5d8 f8d8 a1c1 c8e6 e1e2 d8c8 c1c8 a8c8 h1c1 e6g4 e2e1 c8c1 d2c1 f7f6 e1d2 g8f7 a2a4 b7b6 d2c3 f7e6 c1a3 g4e2');
INSERT INTO evaluations VALUES(41039,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(41039,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(41039,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(41039,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(41039,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(41039,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(41039,36,'cp 155','e4e5 f6d5 d1g4 g7g6 g1f3 b8d7 c1h6 d7b6 c4d5 e6d5 g4f4 c7c6 e1g1 c8e6 f3g5 d8d7 h6g7 h8g8 g7h6 g8h8 b2b3 e8c8 a2a4 b6a8 f1b1 e6f5 b1c1 f5e6 a4a5 a8c7 c3a4 c7a6 h2h3 d8e8 c1c3 e7b4 c3f3 a6c7 g1h2 e6f5 a1c1 c8b8 g2g4 f7f6 g4f5');
INSERT INTO evaluations VALUES(42221,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(42221,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(42221,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(42221,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(42221,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(42221,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(42221,36,'cp -52','e8g8 f1e2 h7h6 g5d2 f8e8 e1g1 c8f5 e2d3 f5d3 c2d3 b4a5 d1b3 a5b6 a1e1 e8e1 f1e1 b8c6 d2e3 c6e7 c3a4 e7f5 a4c5 d8c8 h2h3 c7c6 f3e5 c8c7 b3c3 f6h7 e3d2 h7f8 a2a4 c7c8 b2b4 a7a6 g2g3 b6c7 g1g2 c7d6 f2f4');
INSERT INTO evaluations VALUES(37101,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(37101,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(37101,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(37101,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(37101,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(37101,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(37101,36,'cp -52','f1e2 h7h6 g5d2 b8c6 h2h3 a7a6 e1g1 b4d6 e2d3 c6b4 f1e1 f8e8 e1e8 d8e8 d3f1 c8f5 a2a3 b4c6 a1c1 a8d8 c3e2 e8e4 e2c3 e4e6 c3a4 e6c8 c2c4 d5c4 c1c4 c8e6 c4c1 f6e4 a4c3 c6e7 d1e2 e4c3 c1c3 e7d5 e2e6 f5e6');
INSERT INTO evaluations VALUES(44303,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(44303,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44303,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(44303,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44303,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(44303,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44303,36,'cp 30','e3d4 d8b6 b2b3 f8d6 f4d6 b6d6 f1e2 e8g8 e1g1 b8c6 c2c4 b7b6 b1c3 c8b7 c3b5 d6f4 d1c1 f4b8 f1d1 c6e7 c1a3 e7f5 a1c1 a7a6 b5c3 b6b5 d4d5 e6d5 e2d3 d5d4 d3f5 d4c3 c1c3');
INSERT INTO evaluations VALUES(44296,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44296,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44296,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44296,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(44296,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(44296,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(44296,36,'cp 14','d8b6 b2b3 f8d6 f4c1 b8c6 f1e2 e8g8 e1g1 d6e7 c2c4 d7d5 b1c3 b6d8 c1b2 b7b6 c4d5 f6d5 c3d5 d8d5 e2c4 d5d6 d4d5 e6d5 d1d5 c8b7 d5d6 e7d6 a1d1 a8d8 f3d4 c6d4 b2d4 h7h6 a2a4 d6e7 d4e3 e7a3 e3f4 a3b4 d1d8');
INSERT INTO evaluations VALUES(45385,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45385,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(45385,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(45385,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45385,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45385,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45385,36,'cp 0','e2e3 g7g6 d4c5 f8g7 f1b5 c8d7 a2a4 e8g8 e1g1 d8c7 b5d7 b8d7 b2b4 a7a5 c2c3 f6e4 a1a3 e7e6 f3d2 e4d2 b1d2 b7b6 c5b6 d7b6 b4b5 g7c3 d2f3 c7c4 f3d4 c3b4 a3a1 b4c3');
INSERT INTO evaluations VALUES(45383,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45383,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45383,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45383,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45383,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45383,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45383,36,'cp 0','e7e6 f1e2 b7b6 c2c4 c8b7 e1g1 f8d6 d4c5 b6c5 d1b3 b7c6 b3c2 e8g8 b1c3 c6b7 c2b3 b7c6');
INSERT INTO evaluations VALUES(46247,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46247,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46247,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46247,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46247,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46247,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46247,36,'cp 0','f1e2 b7b6 c2c4 c8b7 e1g1 f8d6 b1c3 e8g8 d4c5 b6c5 d1b3 b7c6 b3c2 c6b7');
INSERT INTO evaluations VALUES(46641,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(46641,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46641,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(46641,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46641,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46641,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46641,36,'cp -7','f8g7 f2f4 b8c6 a2a4 c8g4 d1d2 e8g8 h2h3 g4e6 c4e6 f7e6 g1e2 e6e5 f4f5 g6f5 e4f5 e7e6 f5e6 c6d4 e1g1 d4e6 e2g3 d8d7 d2d1 g8h8 c1d2 a8e8 c3e4 e6f4 e4f6 f8f6 d1g4 d7g4 h3g4 f6g6 d2f4 e5f4');
INSERT INTO evaluations VALUES(49551,30,'cp -205',NULL);
INSERT INTO evaluations VALUES(49551,31,'cp -228',NULL);
INSERT INTO evaluations VALUES(49551,32,'cp -230',NULL);
INSERT INTO evaluations VALUES(49551,33,'cp -241',NULL);
INSERT INTO evaluations VALUES(49551,34,'cp -231',NULL);
INSERT INTO evaluations VALUES(49551,35,'cp -168',NULL);
INSERT INTO evaluations VALUES(49551,36,'cp -163','f6e4 f3e4 d8h4 e1e2 h4e4 e2f2 f8c5 d2d4 c5d4 f2g3 e4g6 d1g4 g6c2 b1d2 h7h5 g4e4 h5h4 g3f3 c2d1 e4e2 d1a4 g2g4 a4c6 e2e4 b7b5 e4c6 b8c6 f3g2 c8b7 g1f3 c6b4 g2h3 b5c4 f1c4 d4b6 a2a3 b4c2 a1a2 c2e3 c4d3 f7f6 d2e4 b7d5 c1e3 d5a2 e3b6 a7b6 h1c1 c7c6');
INSERT INTO evaluations VALUES(50291,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(50291,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(50291,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(50291,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(50291,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(50291,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(50291,36,'cp 95','b1c3 d7d5 c4d5 f6d5 g1f3 d5c3 b2c3 c7c5 e2e4 e7e6 a1b1 f8e7 f1d3 c5d4 c3d4 e8g8 e1g1 b7b6 c1f4 c8b7 b1c1 b8c6 d3b5 a8c8 d4d5 e6d5 e4d5 c6b4 c1c8 d8c8 d5d6 e7f6 d6d7 c8c5 f4d6 c5b5 d6f8 b4d5 f8d6 d5c3 d1e1 c3e2 g1h1 b5d7 e1e2 d7d6 f1d1');
INSERT INTO evaluations VALUES(50289,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(50289,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(50289,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(50289,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(50289,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(50289,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(50289,36,'cp 50','e7e6 e2e4 f8b4 f1d3 c7c5 d4d5 b4c3 b2c3 d7d6 g1e2 b8d7 e2g3 d7e5 d3e2 e5g6 d5e6 c8e6 f2f4 e6c8 e1g1 e8g8 e2d3 f8e8 a1b1 b7b6 b1b2 d8c7 d1f3 e8e7 d3b1 c8b7 f1e1 a8e8 b2e2 b7a6 b1d3 a6b7 h2h3 g6h4 f3f2 h4g6');
INSERT INTO evaluations VALUES(49236,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(49236,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(49236,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(49236,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(49236,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(49236,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(49236,36,'cp 87','e2e4 f8b4 f1d3 c7c5 a2a3 b4c3 b2c3 d7d6 e4e5 d6e5 d4e5 f6d7 f2f4 b8c6 g1f3 b7b6 e1g1 c8b7 d1e2 d7f8 d3c2 f8g6 f4f5 e6f5 e5e6 e8g8 c2f5 d8f6 e6f7 f8f7 f5e6 a8e8 f3g5 e8e6 g5e6 f6f1 e2f1 f7f1 g1f1 c6a5 f1g1 a5c4 a1a2 b7c6 a2f2 c4d6 e6f4 g6e5 h2h4 d6e4 f2e2 e5c4 h4h5 e4g3 e2e1 g3f5');
INSERT INTO evaluations VALUES(49238,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(49238,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(49238,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(49238,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(49238,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(49238,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(49238,36,'cp 73','f8b4 f1d3 c7c5 a2a3 b4c3 b2c3 d7d6 e4e5 d6e5 d4e5 f6d7 f2f4 b8c6 g1h3 e8g8 e1g1 f7f5 e5f6 d7f6 d1e2 e6e5 h3f2 d8c7 f4e5 c7e5 f2e4 f6e4 f1f8 g8f8 e2e4 e5e4 d3e4 c8e6 c1e3 e6c4 e3c5 f8g8 a1b1 a8e8 e4f3 c4a6 c5f2 e8d8 f2g3 c6a5 h2h3 b7b6 b1e1 a5c4');
INSERT INTO evaluations VALUES(50565,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(50565,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(50565,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(50565,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(50565,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(50565,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(50565,36,'cp 118','b1c3 d7d5 e4e5 b8c6 g1f3 c8d7 f1d3 d8b6 d3c2 c6a5 e1g1 h7h6 f1e1 g8e7 b2b3 b6c7 c1d2 e7c6 a2a3 a8c8 a1a2 c7b6 c3a4 b6d8 a2a1 f8e7 h2h3 b7b5 a4c5 c6d4 f3d4 e7c5 d2a5 d8a5 b3b4 c5d4 b4a5 d4a1 d1a1 c8c2 a1d1');
INSERT INTO evaluations VALUES(50563,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(50563,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(50563,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(50563,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(50563,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(50563,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(50563,36,'cp 119','d7d5 e4e5 b8c6 g1f3 g8h6 c1h6 g7h6 f1d3 f8g7 e1g1 f7f6 d3c2 c8d7 f1e1 f6e5 f3e5 c6e5 d4e5 d8g5 d1d4 a8c8 f2f4 g5h4 e1f1 c8c4 d4d2 e8g8 d2d3 f8f5 c3e2 g7f8 g1h1 f5f7 f1f3 g8h8 d3d1 f8e7 c2d3 c4c8 d1d2 c8g8 a1f1 f7g7 f3h3 h4g4');
INSERT INTO evaluations VALUES(51021,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(51021,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51021,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(51021,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51021,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(51021,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51021,36,'cp -5','d7d6 g1f3 d6e5 f3e5 b8d7 e5d3 g8e7 c2c3 e7f5 f1e2 c7c5 d4c5 d7c5 e1g1 d8c7 c1f4 f8d6 f4d6 c7d6 b1d2 c8d7 d2c4 d6c7 d3c5 c7c5 d1b3 d7c6 f1d1 e8e7 c4d2 h8d8 e2f3 e7f8 f3c6 c5c6 d2f3 c6c7 a2a4 b7b6 d1d8 a8d8 g1h1');
INSERT INTO evaluations VALUES(54925,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54925,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54925,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54925,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54925,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54925,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54925,36,'cp 9','c8f5 e2e3 e7e6 f1d3 g8f6 e1g1 f8e7 d3f5 e6f5 d1d3 d8c8 c3e2 e7d6 b2b3 e8g8 c2c4 c7c6 d3c2 c8e6 a2a4 f8e8 c1a3 d6c7 e2f4 e6e4 f4d3 a7a5 c2d1 b8d7 a1c1 a8d8 a3b2 e4e6 f3e5 f6g4 d3f4 e6e7');
INSERT INTO evaluations VALUES(1365,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(1365,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(1365,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(1365,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(1365,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(1365,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(1365,36,'cp 47','b7b5 e1g1 g8f6 e4e5 f6d5 c3e4 f7f5 e5f6 d5f6 d2d3 g7g6 e4f6 e7f6 d3d4 c8b7 f1e1 e8f7 d4c5 f8c5 c1e3 c5d6 d1d4 h8e8 a1d1 e8e6 d4g4 c6c7 d1d2 h7h5 g4h3 b7f3 h3f3 a8d8 g2g3 d6c5 c2c3 b5b4 d2d3 c5e3 e1e3 b4c3 d3c3 c7b6 f3d5 f7g7 d5d2 e6e3 c3e3 b6c6 e3e7 g7f8');
INSERT INTO evaluations VALUES(1667,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(1667,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(1667,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(1667,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(1667,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(1667,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(1667,36,'cp 88','f3g5 c6d8 f2f4 d8e6 g5e6 c8e6 c4e6 e7e6 f4f5 e6d7 c2c4 f8e7 e1g1 g8f6 d1f3 h7h6 f3g3 h6h5 b1c3 h5h4 g3h3 c7c6 c1g5 f6h7 g5e3 e7g5 e3g5 h7g5 h3g4 f7f6 h2h3 g7g6 b2b4 e8e7 b4b5 g6f5 f1f5 a8g8 g1h1 g5e6');
INSERT INTO evaluations VALUES(1664,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(1664,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(1664,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(1664,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(1664,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(1664,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(1664,36,'cp 78','c6d8 f2f4 d8e6 g5e6 c8e6 c4e6 e7e6 e1g1 c7c6 f4f5 e6d7 g1h1 g8f6 d1e2 f8e7 b1d2 d6d5 d2f3 d5e4 d3e4 d7c7 c1d2 h7h6 a1e1 e7d6 d2c3 e8c8 e2f2 c8b8 f2g3 f6h5 g3g4 h5f4');
INSERT INTO evaluations VALUES(5942,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(5942,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5942,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5942,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5942,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5942,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(5942,36,'cp -1','g8e7 d3d4 e5d4 f3d4 e8g8 c1e3 c5d4 e3d4 f7f5 f2f3 f5e4 c3e4 d6d5 e4g3 c6d4 d1d4 c7c6 b5d3 c8f5 d3f5 e7f5 g3f5 f8f5 e1g1 d8b6 d4b6 a7b6 a2a4 d5d4 f1d1 f5d5 g1f2 b6b5 a4a5 c6c5 h2h4 c5c4 g2g4 a8f8 b2b3 d4d3 c2d3 d5d3 d1d3 c4d3');
INSERT INTO evaluations VALUES(9321,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(9321,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(9321,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(9321,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(9321,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(9321,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(9321,36,'cp -31','c8g4 g5f3 c6a5 c4b3 a5b3 a2b3 e8g8 c1e3 f6d7 h2h3 g4h5 b1d2 a7a5 g1h2 b7b5 d1e1 c7c6 d3d4 c5b6 d4e5 d6e5 e3b6 d8b6 f3h4 f8e8 f2f3 h5g6 e1f2 b6f2 f1f2 d7c5 d2f1 e8d8 f2d2 d8d2 f1d2 c5e6 h2g3 g8f8 h4g6 h7g6 c2c3 e6f4 b3b4');
INSERT INTO evaluations VALUES(9979,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(9979,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(9979,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(9979,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(9979,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(9979,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(9979,36,'cp 51','e8g8 c1d2 d7d5 e2e3 b7b6 a2a3 b4c3 d2c3 a7a5 b2b3 c8a6 c3b2 c6e7 f3d2 a8c8 f1e2 e7g6 e1g1 c7c5 d4c5 b6c5 d2f3 d8b6 f1d1 f8d8 a1c1 f6e4 e2d3 h7h6 h2h3 a6b7 c4d5 e6d5 d3e4 d5e4 c2c3 f7f6');
INSERT INTO evaluations VALUES(12747,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(12747,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12747,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12747,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(12747,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(12747,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(12747,36,'cp 13','c3b3 d5c7 f3g5 c7e6 g5e6 f7e6 f2f4 f8h6 b3h3 h6g7 c1d2 e8g8 b1c3 d7d6 e1c1 d6e5 f4e5 d8c7 d2h6 g7h6 h3h6 c6e5 f1e2 c8d7 h6e3 d7c6 h1e1 a7a6 c1b1 f8f5 g2g4 f5f6 g4g5 f6f5 h2h4 a8f8 e2d3 e5d3 e3e6 g8g7 d1d3 c7f4 c3d5 c6d5 d3d5 f5d5 e6d5');
INSERT INTO evaluations VALUES(18853,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(18853,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(18853,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(18853,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(18853,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(18853,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(18853,36,'cp -12','d7d5 c2c3 f8e7 f1e2 g8h6 e1g1 e8g8 b1a3 b7b6 f4f5 e6f5 e4d5 d8d5 c1h6 g7h6 d1d2 d5d6 a1e1 d6g6 e2d1 e7g5 d2f2 c8d7 a3c4 a8e8 d1a4 e8e1 f1e1 f8e8 e1e8 d7e8 f3h4 g5h4 f2h4 g8g7 h4e1 g6e6 e1g3 e6g6');
INSERT INTO evaluations VALUES(20306,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(20306,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20306,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(20306,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(20306,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(20306,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(20306,36,'cp -7','c2c3 g8h6 f1e2 f8e7 b1d2 e8g8 d2f1 f7f5 f1g3 d8b6 e4e5 d5d4 e1g1 a7a5 a2a4 b6d8 h2h3 b7b6 d1b3 d8d7 c1d2 f8d8 g1h2 a8b8 c3d4 c8a6 f1c1 c5d4 c1c2 h6f7 a1c1 d8c8 g3h5 g8h8 h2g1 e7b4 g1h2 b4d2 c2d2');
INSERT INTO evaluations VALUES(22716,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(22716,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(22716,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(22716,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(22716,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(22716,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(22716,36,'cp -42','b1c3 g8f6 c3d5 c6b4 d5f6 d8f6 a2a3 f6c6 c4f7 e8f7 a3b4 c6g2 d1f3 g2f3 g1f3 d7d6 e3e4 h7h6 d2d3 f8e7 c1e3 c8h3 h1g1 a7a6 e1e2 g7g5 f3e1 h8g8 f2f3 h6h5 b4b5 h3e6 b5a6 a8a6 a1a6 b7a6 b2b3 h5h4 g1g2 g8g6 d3d4 e5d4 e3d4 a6a5');
INSERT INTO evaluations VALUES(27159,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(27159,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(27159,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(27159,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(27159,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(27159,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(27159,36,'cp -1','d7c6 d1e2 f8d6 e1g1 g8f6 f1d1 e8g8 b1c3 d8e8 c1g5 f6h5 f3h4 f7f6 g5e3 h5f4 e3f4 e5f4 h4f5 d6e5 d1d3 e5c3 d3c3 a8d8 f2f3 e8e5 e2c4 g8h8 c4c5 f8e8 c5a7 g7g6 f5h6 h8g7 h6g4 e5b5 a7a3 d8d2 g4f2 b5e2 a1f1 d2c2');
INSERT INTO evaluations VALUES(28936,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(28936,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(28936,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28936,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(28936,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(28936,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(28936,36,'cp -11','c1f4 b8d7 g1f3 g7g6 h2h3 f8g7 e2e3 f6e4 f1d3 e4c3 b2c3 f5d3 d1d3 b7b6 c5b6 a7b6 c3c4 d5c4 d3c4 d8c8 e1g1 e8g8 a1b1 a8a7 f1c1 c6c5 d4c5 b6c5 f4g5 f8e8 f3d4 c8a8 d4b3 h7h6 g5f4 a7a2 b3c5 d7c5 c4c5');
INSERT INTO evaluations VALUES(29479,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(29479,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(29479,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(29479,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(29479,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(29479,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(29479,36,'cp -63','c7c6 a4c2 b7b5 g1f3 b8d7 h2h3 d7b6 e3e4 g7g6 g2g4 a7a5 c1e3 f8g7 f1g2 b5b4 c3e2 e8g8 e1g1 b4b3 c2b1 b3a2 b1c2 c4c3 e2c3 a5a4 c3a2 e6b3 c2d3 b3c4 d3c2 c6c5 f1c1 c4b3 c2d3 c5d4 e3d4 f6g4 a2b4 g7d4');
INSERT INTO evaluations VALUES(28931,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(28931,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(28931,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(28931,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(28931,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(28931,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(28931,36,'cp -45','g1f3 b7b5 a4c2 b8d7 h2h3 d7b6 e3e4 a7a5 g2g4 g7g6 f1g2 f8g7 e1g1 e8g8 c1e3 e6c8 a1d1 d8c7 c3e2 f6d7 f3g5 c8a6 f2f4 b5b4 e4e5 b6d5 g2d5 c6d5 f4f5 c4c3 f1f2 a6e2 c2e2 c3b2 e2b2 f8c8 e3f4 c7c3 e5e6 c3b2 e6f7 g8f8');
INSERT INTO evaluations VALUES(30759,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(30759,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(30759,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(30759,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(30759,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(30759,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(30759,36,'cp 78','b1c3 d7d6 d3d4 f8g7 f1c4 h7h6 d1e2 e7e6 c1f4 a7a6 e1c1 b6b5 c4b3 g8e7 h2h4 b8d7 c1b1 d7b6 d4d5 e6d5 c3d5 e7d5 e4d5 d8e7 f4e3 e7e4 h1e1 e8c8 e2f1 e4f5 e3b6 c7b6 f3d4 g7d4 d1d4 h8e8 f2f3 c8b8 d4e4 e8e5 e4e5 d6e5 f1e2 f5f6 e2e5 f6e5');
INSERT INTO evaluations VALUES(30758,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(30758,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(30758,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(30758,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(30758,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(30758,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(30758,36,'cp 56','g8f6 e4e5 f6d5 c3d5 b7d5 d3d4 c7c5 d4c5 e7e6 f1b5 a7a6 b5e2 b8c6 e1g1 a8b8 c5b6 d8b6 b2b3 f8g7 c1e3 b6a5 d1d2 d5f3 d2a5 c6a5 e2f3 g7e5 a1c1 e5c3 f1d1 d7d5 h2h4 a5c6 h4h5 e8g8 f3e2 d5d4 e3d2');
INSERT INTO evaluations VALUES(31598,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(31598,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(31598,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(31598,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(31598,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(31598,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(31598,36,'cp 114','d8e7 b2b4 a7a6 a2a4 g8f6 e1g1 b8d7 c1b2 e7d8 c4b3 f8e7 a4a5 e8g8 c3d1 d6d5 e4d5 f6d5 b3d5 c6d5 f3d5 e7b4 d1e3 b4c5 b2e5 c5e3 d2e3 d7e5 d5e5 a8c8 f1b1 f8e8 e5f5 d8e7 f5f3 c8c6 b1b6 c6b6 a5b6 h7h6 a1b1 g7g6');
INSERT INTO evaluations VALUES(36112,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(36112,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(36112,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(36112,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(36112,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(36112,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(36112,36,'cp -23','f8e8 f1e1 c8e6 e2d3 b8d7 h2h3 h7h6 g5e3 d8c7 d1c1 d7f8 c3e2 e6d7 b2b3 f8e6 c2c4 c6c5 c1c2 d7c6 d4c5 e6c5 e2d4 c5d3 c2d3 d5c4 d3c4 c6d5 c4c7 d6c7 e1d1 c7b6 d4f5 b6e3 f5e3 d5f3 g2f3 f6h5 e3g2 a8d8 a1c1 d8d1 c1d1');
INSERT INTO evaluations VALUES(36540,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(36540,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36540,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36540,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36540,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36540,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36540,36,'cp 0','e7e5 f1e2 a7a5 e3e4 e5d4 f3d4 f8e8 f2f3 d6d5 c4d5 f6d5 e1g1 d5c3 b2c3 b8d7 c1e3 d7c5 d1d2 d8e7 a1b1 b7b6 f1d1 c8a6 e3g5 g7f6 g5e3');
INSERT INTO evaluations VALUES(37545,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(37545,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(37545,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(37545,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(37545,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(37545,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(37545,36,'cp 100','f1d3 d7d6 e1g1 g8f6 h2h3 e8g8 f1e1 f6d7 c2c3 e7e5 c1e3 b8c6 b1d2 g8h8 d1c2 b7b6 a2a3 c8b7 b2b4 a6a5 a1d1 a5b4 a3b4 h7h6 b4b5 e5d4 c3d4 c6b4 c2b3 b4d3 b3d3 f8e8 d4d5 h8g8 f3d4 d8f6');
INSERT INTO evaluations VALUES(46737,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(46737,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(46737,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(46737,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(46737,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(46737,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(46737,36,'cp -32','b7b6 c5b6 a7b6 c1f4 f6h5 f4g5 h7h6 g5h4 h5g7 h4g3 g7f5 g3f4 f8g7 e2e3 e8g8 g1f3 g6g5 f4e5 f7f6 e5g3 c7c6 f1d3 f5g3 h2g3 e7e5 d1b3 g8h8 d4e5 f6e5 f3e5 g7e5 h1h6 h8g7 h6g6 g7f7 e1c1 b6b5 d3c2 c8e6 f2f4 g5f4 c3e2 b8d7 e2f4');
INSERT INTO evaluations VALUES(55423,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(55423,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(55423,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(55423,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(55423,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(55423,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(55423,36,'cp 97','c4f7 e8f7 f3e5 f7f6 d1f3 f6e5 f3f7 g8f6 b1c3 d6d5 c1f4 e5f4 c3d5 d8d5 e4d5 f8b4 e1f1 c8d7 f7g7 d7b5 f1g1 b8d7 g7g3 f4f5 g3h3 f5g6 h3b3 b4d6 b3b5 a8e8 g1f1 h6h5 a1e1 e8e1 f1e1 g6f7 e1d1 b7b6 h1e1 h8g8 g2g3 a7a5 f2f4 h5h4 b5e2 d7c5 e2f3 h4h3 a2a3 g8g4 e1e3 g4g8');
INSERT INTO evaluations VALUES(1992,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(1992,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(1992,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(1992,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(1992,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(1992,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(1992,36,'cp 11','b1c3 f8b4 c1d2 g8f6 d1e2 e8g8 f2f3 d7d5 a2a3 b4a5 e1c1 d5e4 c3e4 a5d2 e4d2 e5e2 f1e2 f8e8 h1e1 g8f8 e2d3 c8e6 d2c4 a7a5 b2b3 e6c4 d3c4 e8e1 d1e1 c7c6 a3a4 a8d8 e1d1 d8d1 c1d1 f8e7 d1d2 g7g6 d2d3 f6d7 d3e4 f7f6 c4g8 h7h6 e4e3 b7b6 g8c4 d7e5 e3d4 e7d6');
INSERT INTO evaluations VALUES(2938,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(2938,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(2938,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(2938,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(2938,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(2938,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(2938,36,'cp 14','d5c4 d3c4 b7b5 c4e2 a7a6 d1c2 c6c5 d4c5 d7c5 f1d1 d8b6 b2b4 c5d7 a2a4 e7b4 a4b5 b4c3 c2c3 c8b7 c3d4 b6d4 f3d4 a6b5 d4b5 a8a1 d1a1 g7g5 h4g3 f8a8 a1a8 b7a8 g3d6 f6d5 d6a3 g8g7 a3b2 g7f8 b5d4 f8e7 b2a3 e7f6 d4b3 f6g6 a3b2');
INSERT INTO evaluations VALUES(2941,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(2941,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2941,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(2941,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(2941,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(2941,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(2941,36,'cp 12','d3c4 b7b5 c4e2 a7a6 d1c2 c6c5 d4c5 d7c5 f1d1 d8b6 b2b4 c5d7 a2a4 e7b4 a4b5 b4c3 c2c3 c8b7 c3d4 b6d4 f3d4 a6b5 d4b5 a8a1 d1a1 g7g5 h4g3 f8a8 a1a8 b7a8 b5d4 f6e4 g3c7 a8b7 c7a5 e4c5 f2f3 b7a6 e2b5 a6b5 d4b5 c5a6 a5c3 a6c5');
INSERT INTO evaluations VALUES(2940,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(2940,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(2940,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(2940,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(2940,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(2940,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2940,36,'cp 10','b7b5 c4e2 a7a6 d1c2 c6c5 d4c5 d7c5 f1d1 d8b6 b2b4 c5d7 a2a4 e7b4 a4b5 b4c3 c2c3 c8b7 c3d4 b6d4 f3d4 a6b5 e2b5 d7b6 d1c1 a8a1 c1a1 b7d5 a1c1 f8c8 c1c8 b6c8 f2f3 c8d6 e3e4 d6b5 d4b5 d5c6 b5d4 c6d7 h4f6 g7f6 d4b3 g8f8 b3a5 f8e7 g1f2 d7b5 f2e3');
INSERT INTO evaluations VALUES(3404,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(3404,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(3404,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(3404,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(3404,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(3404,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(3404,36,'cp 36','c8g4 d4d5 c6b8 h2h3 g4h5 c4d3 b8d7 c3c4 d7c5 b1c3 c5d3 d1d3 a8c8 b2b3 h5g6 c1d2 f6d7 g2g4 a7a6 a1d1 h7h5 f3h2 e7g5 d3g3 g5h4 g3f3 h4g5 d2g5 d8g5 f3g3 g5h6 h2f3 f7f6 g1g2 d7c5 f3h4 h5g4 h4g6 h6g6');
INSERT INTO evaluations VALUES(23694,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(23694,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(23694,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(23694,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(23694,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(23694,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(23694,36,'cp 17','d4d5 c6b8 h2h3 g4h5 c4d3 b8d7 c3c4 d7c5 b1c3 c5d3 d1d3 h5g6 c1d2 a7a6 a2a4 f6d7 g2g4 d7c5 d3c2 c7c6 a4a5 h7h5 b2b4 c5d7 c2d3 a8c8 g1g2 d7f6 f3h2 c6d5 c3d5 f6d5 d3d5 d8d7 f2f3 c8c7 h2f1 f8c8 e1c1 d7e6 f1e3 e7g5 c1d1');
INSERT INTO evaluations VALUES(4630,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(4630,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(4630,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(4630,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(4630,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(4630,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(4630,36,'cp 144','d2d4 e5d4 c3d4 c5b6 d4d5 c6b8 d5d6 c7d6 b1c3 d6d5 c3d5 e7d5 d1d5 d7d6 c1g5 d8c7 g5f4 f8d8 a1c1 c8e6 d5h5 c7e7 f3g5 h7h6 g5e6 e7e6 b5c4 e6f6 f4d2 b6d4 c4d5 b8c6 c1c3 c6e7 c3f3 e7d5 f3f6 d5f6 h5f3 d6d5 e4d5 d8d5 f3b3 b7b6 h2h3 a8c8 f1d1');
INSERT INTO evaluations VALUES(4628,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(4628,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(4628,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(4628,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(4628,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(4628,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(4628,36,'cp 140','e5d4 c3d4 c5b6 d4d5 c6b4 a2a3 b4a6 b1c3 d7d6 f1e1 c8d7 b2b4 a6b8 b5d7 b8d7 c3a4 f8e8 c1b2 c7c6 d5c6 e7c6 d1d6 b6c7 d6d1 d7e5 f3e5 c6e5 d1e2 d8d3 a4c3 d3e2 e1e2 e5d3 c3d5 c7d6 b2d4 f7f5 f2f3 a8d8 a1d1 f5e4 e2e4 e8e4');
INSERT INTO evaluations VALUES(4632,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(4632,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(4632,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(4632,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(4632,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(4632,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(4632,36,'cp 142','c3d4 c5b6 d4d5 c6b4 a2a3 b4a6 f1e1 d7d6 b1c3 c8d7 b2b4 a6b8 b5d7 b8d7 c3a4 a7a5 a4b6 d7b6 b4b5 c7c5 d5c6 b7c6 b5c6 e7c6 c1g5 d8d7 f3h4 h7h6 g5f4 b6c4 h4f5 g8h7 f5d6 c6e5 d6c4 d7d1 e1d1 e5c4 a3a4 a8d8 g2g3 d8d1 a1d1 c4b2 d1d5 b2a4 d5a5');
INSERT INTO evaluations VALUES(4631,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(4631,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(4631,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(4631,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(4631,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(4631,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(4631,36,'cp 129','c5b6 d4d5 c6b4 a2a3 b4a6 b1c3 d7d6 f1e1 c8g4 c3a4 e7g6 a4b6 a7b6 h2h3 g4f3 d1f3 c7c6 d5c6 b7c6 b5f1 f8e8 f3c3 a6c5 c1e3 g6e5 f2f3 b6b5 a1c1 d8h4 e3c5 d6c5 c3c5 e8e6 c5e3 e6g6 g1h1 a8d8 f3f4 g6g3 e3f2');
INSERT INTO evaluations VALUES(4627,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(4627,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(4627,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(4627,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(4627,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(4627,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(4627,36,'cp 150','d4d5 c6b4 a2a3 b4a6 b1c3 d7d6 b2b4 c7c6 d5c6 b7c6 b5a4 c8g4 c1g5 a8c8 h2h3 g4f3 d1f3 a6c7 g5e7 d8e7 a4c6 e7e5 c6b7 c8b8 b7d5 g8h8 a1d1 f7f5 f3d3 c7d5 e4d5 b8e8 a3a4 e8c8 c3e2 f5f4 f1e1 f4f3 e2d4 e5d5 d4f3 d5d3 d1d3');
INSERT INTO evaluations VALUES(4626,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(4626,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(4626,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(4626,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(4626,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(4626,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(4626,36,'cp 121','c6b8 d5d6 c7d6 b1c3 d6d5 c3d5 e7d5 d1d5 d7d6 c1g5 d8c7 g5f4 b6c5 a1c1 c7b6 b2b4 b6b5 f4d6 c8e6 d5c5 b5c5 d6c5 f8d8 a2a3 b8d7 c5e7 d8e8 e7h4 e8c8 h2h3 d7b6 f3d4 c8c1 f1c1 a8c8 c1a1 c8c3 d4e6 f7e6 f2f3 b6c4 a1d1 c3a3 d1d8 g8f7 d8d7 f7g6 d7b7 a7a5 h4e7 a5b4 e7b4 a3b3 g1f2 b3b2 f2g3');
INSERT INTO evaluations VALUES(5528,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(5528,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(5528,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(5528,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(5528,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(5528,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(5528,36,'cp 94','g8f6 d4d5 c6b8 d1c2 e8g8 h2h3 b8d7 f1d1 d8c7 c2e2 d7c5 b1c3 a7a6 a2a4 c8d7 e4e5 d6e5 f3e5 f6e8 d5d6 e7d6 c3d5 c7d8 e5d7 c5d7 c1d2 e8f6 d5e7 g8h8 d2b4 f6e8 e7d5 d7e5 c4b3 e5c6 b4c3 e8c7 d5b6 a8b8 a4a5 g7c3 b2c3 d8f6 e2d2 f8e8 d2d6 f6d6');
INSERT INTO evaluations VALUES(8322,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(8322,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(8322,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(8322,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(8322,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(8322,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(8322,36,'cp 77','d4d5 c6b8 d1c2 b8d7 h2h3 e8g8 f1d1 d8c7 c2e2 a7a6 a2a4 d7c5 b1c3 c8d7 e4e5 d6e5 f3e5 f6e8 d5d6 e7d6 c3d5 c7d8 e5d7 c5d7 a4a5 a8b8 a1a3 d7e5 c1e3 e5c4 e2c4 e8f6 e3g5 h7h6 d5f6 g7f6 g5h6 f8e8 a3f3 d8e7 h6f4 b8c8 c4b3 e7e2 d1d6 e2b2 b3d1');
INSERT INTO evaluations VALUES(11138,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(11138,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(11138,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(11138,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(11138,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11138,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(11138,36,'cp 26','b5c6 b7c6 f3e5 e7e6 e5c6 d8c7 c6e5 f8d6 e1g1 d6e5 d4e5 c7e5 f1e1 e5f5 c3e2 h7h6 e2d4 f5h5 d1h5 f6h5 b2b3 e8g8 c1a3 f8e8 c2c4 d5c4 b3c4 c8b7 a3d6 h5f6 d4b3 a8c8 f2f3 e8d8 a1d1 b7a8 c4c5 f6d5 b3a5 d5c3 d1d2 c3b5 a2a4 b5d6 d2d6 d8d6');
INSERT INTO evaluations VALUES(11135,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11135,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(11135,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(11135,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(11135,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(11135,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11135,36,'cp 24','b7c6 f3e5 e7e6 e5c6 d8c7 c6e5 f8d6 e1g1 d6e5 d4e5 c7e5 f1e1 e5f5 c3e2 h7h6 e2d4 f5h5 d1h5 f6h5 g2g4 h5f6 f2f3 e8g8 c1f4 c8d7 g1f2 f8e8 b2b3 f6h7 c2c4 d5c4 b3c4 f7f6 f4e3 h7f8 c4c5 e6e5 d4f5 f8g6 f5d6 e8b8 e1b1 g6e7 h2h4 d7e6 a2a4 f6f5 g4f5 e7f5 d6f5 e6f5');
INSERT INTO evaluations VALUES(11783,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(11783,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(11783,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(11783,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(11783,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(11783,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11783,36,'cp 25','f3e5 e7e6 e5c6 d8c7 c6e5 f8d6 d1e2 e8g8 e1g1 c8b7 f1d1 f6e4 c3e4 d5e4 c2c4 a8c8 c1e3 d6e5 d4e5 c7c4 e2c4 c8c4 b2b3 c4c3 a1c1 c3c1 d1c1 f8c8 c1c8 b7c8 g1f1 f7f6 e3c5 c8b7 c5d6 g8f7 f1e2 f6e5 d6e5 b7d5 e2e3');
INSERT INTO evaluations VALUES(14138,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(14138,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(14138,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(14138,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(14138,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(14138,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14138,36,'cp 34','e7e6 b1d2 f8d6 c2c3 e8g8 f1d3 f6d7 d3c2 f8e8 e1g1 e6e5 d4e5 d7e5 f1e1 c8e6 b2b3 e6d7 f3d4 a7a6 d2f1 d8f6 f1g3 c6d4 c3d4 e5g6 f4d6 f6d6 a1c1 c7c6 d1d2 g6h4 d2d3 h4g6 c1d1 e8e7 f2f4 a8e8 g1h2 g6f8 d3d2');
INSERT INTO evaluations VALUES(14553,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(14553,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(14553,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(14553,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(14553,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(14553,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(14553,36,'cp 25','b1d2 f8d6 c2c3 e8g8 f1d3 f6d7 d3c2 f8e8 e1g1 e6e5 d4e5 d7e5 f1e1 c8e6 b2b3 d8f6 f3d4 a7a6 d2f1 e6d7 f1g3 c6d4 c3d4 e5g6 g3h5 f6e7 f4d6 e7d6 d1f3 a6a5 a1c1 c7c6 f3g3 d6g3 h5g3 g6h4 g1h2 g7g6 c2d3 g8g7 c1c5 b7b6 c5c2 a5a4 g3e2 a4b3 a2b3');
INSERT INTO evaluations VALUES(15638,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(15638,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(15638,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(15638,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(15638,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(15638,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(15638,36,'cp 32','f8e7 f1e2 e8g8 d4c6 b7c6 f2f4 d6d5 e4e5 f6d7 e1g1 d7b6 f4f5 b6d7 f5f6 g7f6 e5f6 e7f6 e2d3 d7e5 d1h5 e5d3 c2d3 d8d6 f1f3 d6e5 h5h6 f6g7 h6c6 a8b8 d3d4 e5d6 c6d6 c7d6 f3g3 g8h8 b2b3 c8e6 c3e2 e6f5 e2f4 f5e4 f4h5 e4g6 h5g7 h8g7 a1f1 f8e8 g3h3');
INSERT INTO evaluations VALUES(8652,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(8652,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(8652,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(8652,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(8652,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(8652,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(8652,36,'cp 63','f1b5 c8d7 d4c6 b7c6 b5e2 e8g8 e1g1 f8e8 h2h3 a7a5 e2d3 e7f8 e3g5 f8e7 d1d2 a5a4 a1d1 h7h6 g5e3 c6c5 f2f3 d7c6 b2b3 f6d7 c3a4 c6a4 b3a4 c7c6 f3f4 d8a5 e4e5 a5d2 e3d2 a8a4 e5d6');
INSERT INTO evaluations VALUES(20012,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20012,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20012,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20012,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20012,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20012,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(20012,36,'cp 0','g8f6 f1e2 e7e5 e1g1 c8d7 d1b3 d8c7 f3g5 d7c8 f2f4 h7h6 g5f3 e5f4 c1f4 g7g5 f4g3 f6h5 g3f2 h5f4 b3c2 f8g7 b1a3 c7e7 a1e1 e8g8 f2g3 f4h5 g3f2');
INSERT INTO evaluations VALUES(23093,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(23093,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(23093,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(23093,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(23093,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(23093,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(23093,36,'cp 124','d1f3 g8f6 g2g4 h7h6 h2h4 d7d6 g4g5 h6g5 h4g5 c8g4 f3g2 h8h1 g2h1 f6d7 d2d3 d7c5 b2b4 c5a6 a2a3 c7c6 h1h4 g4d7 f2f4 a6c7 f4f5 d6d5 c4b3 d8e7 c1d2 e8c8 e1c1 d5e4 h4e4 a7a5 c3a4 c7b5 c1b2 b5d6 e4e3 c8b8 b4a5 d7f5 d2b4 f5e6 e3b6 e6b3 b2b3 e7g5');
INSERT INTO evaluations VALUES(23092,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(23092,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(23092,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(23092,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(23092,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(23092,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(23092,36,'cp 102','g8f6 g2g4 h7h6 h2h4 d7d6 g4g5 h6g5 h4g5 c8g4 f3g2 h8h1 g2h1 f6d7 d2d3 d7c5 b2b4 c5a6 a2a3 c7c6 f2f4 a6c7 f4f5 d6d5 e4d5 g4f5 h1f3 d8d7 d5c6 b7c6 c1e3 f5e6 f3e4 f7f5 g5f6 g7f6 b4b5 f6f5 b5c6 f5e4 c6d7 e8d7 c4e6 d7e6 c3e4 c7d5 e3f2 f8e7 e1d2');
INSERT INTO evaluations VALUES(16970,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(16970,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(16970,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(16970,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(16970,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(16970,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(16970,36,'cp 100','g2g4 h7h6 h2h4 d7d6 g4g5 h6g5 h4g5 c8g4 f3g2 h8h1 g2h1 f6d7 d2d3 d7c5 b2b4 c5a6 a1b1 c7c6 h1g2 d8d7 g5g6 g4e6 g6f7 e8f7 c4e6 d7e6 b4b5 a6c7 a2a4 f7g8 c1e3 d6d5 e4d5 c7d5 c3e4 d5b4 b1b2 f8e7 e1d2 a8d8 b5c6 e6c6 g2g4');
INSERT INTO evaluations VALUES(25558,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(25558,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(25558,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(25558,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(25558,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(25558,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(25558,36,'cp 68','b5e2 a7a6 c2c3 f7f6 e5f6 g7f6 e1g1 f8g7 c3c4 e8g8 f1e1 d7e8 b1c3 e8f7 c4d5 e7d5 e2c4 d8d7 c3e4 b7b6 b2b3 f7h5 c1b2 f6f5 e4g5 b6b5 c4d5 e6d5 d1d2 f8e8 f3h4 f5f4 d2f4 c6b4 f4d2 d7g4 b2a3 g7d4 a3b4 g4h4');
INSERT INTO evaluations VALUES(25955,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(25955,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(25955,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(25955,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(25955,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(25955,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(25955,36,'cp 93','b1c3 d8b6 a2a3 g8e7 c3a4 b6c7 c1f4 e7f5 g2g4 f5e7 a1c1 h7h5 h1g1 e7g6 f4d2 c7d8 g4h5 g6h4 g1g4 h4f5 a4c5 h8h5 f1d3 a8c8 b2b4 f5d4 f3d4 c6e5 g4g3 h5h2 d3f1 f8c5 c1c5 c8c5 b4c5 d8f6 d4f3 h2h1 f3e5 f6e5 d2e3');
INSERT INTO evaluations VALUES(29174,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(29174,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(29174,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(29174,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(29174,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29174,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(29174,36,'cp 19','b8c6 c2c3 d8c7 g2g3 e7e6 c1f4 f8d6 f4d6 c7d6 b1d2 e8g8 f1e1 a8c8 d1a4 f6d7 f3e5 h7h5 e5d7 d6d7 d3f1 f7f6 h2h3 g4f5 a4d1 e6e5 d1h5 e5d4 d2f3 d4c3 b2c3 f5e4 e1e3 d7f7 h5f7 g8f7 f3d2 e4g6 h3h4 f7g8 a1e1 g6f7 f1h3 c8c7 d2b3 b7b6');
INSERT INTO evaluations VALUES(30280,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(30280,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(30280,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(30280,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(30280,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(30280,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(30280,36,'cp 62','c2c3 b8c6 d1b3 d8c8 d2d4 f8e7 f1e1 e8g8 d4e5 c6e5 f3e5 d6e5 b3c4 c8d7 b1d2 b7b5 c4e2 e7d6 d2f3 d7c6 c1g5 f6d7 a1c1 h7h6 g5h4 a7a5 f3d2 a5a4 h4g3 f8e8 c1c2 g8h7 b2b3 a4b3 a2b3 a8a3 e1c1 d7c5 b3b4 c5a4');
INSERT INTO evaluations VALUES(35511,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(35511,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(35511,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(35511,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(35511,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(35511,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(35511,36,'cp -82','c3e4 d7d5 e4c3 c7c6 d2d3 b8d7 d1h5 g7g6 h5d1 f8g7 g1f3 h7h6 e1g1 h8f8 c3e2 f7g8 e2g3 b7b6 f1e1 d8f6 c1d2 f6f7 b2b4 c8b7 c2c4 b7a6 a1c1 d5c4 g3e4 c4d3 c1c6 a6c4 e1e3 c4d5 c6d6 a8c8 e3d3 d5e4 d6d7 f7a2 d7g7 g8g7');
INSERT INTO evaluations VALUES(35510,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(35510,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(35510,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(35510,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(35510,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(35510,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(35510,36,'cp -55','d7d5 e4g3 b8c6 d2d3 f8e7 d1h5 f7g8 c1d2 c8e6 g1f3 d8e8 h5e8 a8e8 h2h3 h7h6 e1g1 g8h7 f1e1 e7d6 a1c1 h8f8 c2c4 d5c4 d3c4 e6h3 c4c5 d6e7 f3e5 c6e5 e1e5 h3d7 d2c3 d7c6 c1e1 e7d8 e5e8 f8e8 e1e8 c6e8 f2f3 h7g6 a2a3 h6h5 g3e2 g6f7 g1f2 d8e7 b2b4 e8b5 e2f4 h5h4');
INSERT INTO evaluations VALUES(35501,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(35501,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(35501,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(35501,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(35501,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(35501,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(35501,36,'cp -77','e4c3 c7c6 g1f3 f8d6 d2d4 e5d4 d1d4 d8f6 f3g5 f7g8 c3e2 f6d4 e2d4 b8a6 g5f3 h7h6 e1g1 g8h7 f1e1 a6c7 f3e5 h8f8 a2a4 c6c5 d4b5 c7b5 a4b5 c8f5 c2c3 g7g5 c1d2 a8e8 e5f3 e8e1 f3e1 f8a8 d2e3 a7a5 b5a6 a8a6 a1a6 b7a6 h2h3');
INSERT INTO evaluations VALUES(39707,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39707,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39707,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39707,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39707,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39707,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39707,36,'cp 0','b7b5 f2f3 h7h6 g5e3 b5b4 c3a4 d8a5 b2b3 c8d7 d4f5 e6f5 e3b6 a5e5 b6d4 e5a5');
INSERT INTO evaluations VALUES(40767,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(40767,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(40767,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40767,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40767,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(40767,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40767,36,'cp 41','f2f4 e8g8 g1f3 e5f4 c1f4 c8e6 c4e6 f7e6 d1e2 e6e5 f4e3 d8d7 a2a3 f6g4 e1c1 g4e3 e2e3 b8a6 c1b1 c7c5 d1f1 a6c7 f3d2 b7b5 f1f8 a8f8 h1f1 f8f1 d2f1 h7h6 e3f2 b5b4 a3b4 c5b4 c3d1 d6d5 e4d5 c7d5');
INSERT INTO evaluations VALUES(42740,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(42740,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(42740,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42740,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(42740,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(42740,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(42740,36,'cp -21','d7d6 e1g1 e8g8 f1e1 c7c5 c3e2 b4a5 c2c3 b8c6 d3c2 c8e6 h2h3 b7b5 d2d3 a8c8 e2g3 h7h6 c1e3 d8c7 e3d2 f8e8 f3h4 d6d5 e4d5 e6d5 h4f5 c6e7 f5e7 c7e7 c3c4 a5d2 d1d2 b5c4 d3c4 d5c4 g3f5 e7e6 f5d6 c8d8 a1d1');
INSERT INTO evaluations VALUES(42872,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(42872,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(42872,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(42872,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(42872,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(42872,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(42872,36,'cp 58','b1c3 e4f6 f3e5 e8g8 d2d4 c5e7 f1e1 d7d6 e5f3 c8f5 h2h3 f5g6 d4d5 b8d7 a2a4 a7a6 a4a5 h7h6 c4d3 g6d3 d1d3 f8e8 c1f4 d7f8 d3c4 f8g6 f4g3 d8d7 c4b3 a8b8 a1d1 d7f5 f3d4 f5h5 f2f3 h5g5 c3e4 f6e4');
INSERT INTO evaluations VALUES(43463,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(43463,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(43463,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(43463,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(43463,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43463,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(43463,36,'cp 43','c1g5 c8e6 g5f6 d8f6 b3a5 f6d8 a5c4 g7g6 a2a4 d8c7 c3d5 e6d5 d1d5 b8d7 f1e2 a8d8 d5a5 c7c6 f2f3 f8e7 h2h4 f7f5 e1c1 f5e4 c4e3 e4f3 g2f3 h8f8 c1b1 c6c5 a5d2 c5d4 d2d4 e5d4 d1d4 d7e5 f3f4 e5c6');
INSERT INTO evaluations VALUES(45938,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(45938,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(45938,32,'cp 172',NULL);
INSERT INTO evaluations VALUES(45938,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(45938,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(45938,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(45938,36,'cp 163','h2h3 b8d7 f1e2 f8e7 g1f3 a7a6 a2a4 d7f8 h3h4 d8d7 d1c2 e7d8 a4a5 h7h6 c1d2 d8c7 g2g3 f8h7 c3d1 d7d8 f3h2 b7b6 a5b6 c7b6 e1g1 e8g8 a1a3 a6a5 d1c3 f6e8 e2g4 h7f6 g4c8 d8c8 g1g2 f6g4 f2f3 g4h2 g2h2 e8f6');
INSERT INTO evaluations VALUES(48366,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(48366,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(48366,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(48366,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(48366,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(48366,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(48366,36,'cp -26','b7b6 c2c4 c8b7 g1f3 b8d7 f1d3 f8d6 e1g1 d5c4 d3c4 e8g8 b1d2 e6e5 d4e5 d7e5 f3e5 d6e5 d2f3 f6d7 f3e5 d7e5 d1d8 f8d8 c4e2 b7d5 f2f3 d5c4 e2c4 e5c4 b2b3 c4a5 b3b4 a5b3 a1a2 d8d3 a2c2 c7c5 b4c5 b3c5');
INSERT INTO evaluations VALUES(48691,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(48691,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(48691,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(48691,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(48691,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(48691,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(48691,36,'cp 0','g1f3 f8d6 f1d3 e8g8 e1g1 h7h6 b1c3 b8c6 c3b5 d6e7 c1f4 a7a6 b5c3 c6b4 a2a3 b4d3 d1d3 f8e8 f1e1 c7c6 c3e2 e7f8 e2g3 e8e1 a1e1 a6a5 a3a4 c8e6 c2c3 d8b6 e1e6 f7e6 d3g6 b6b2 f3e5 b2c3 g6f7 g8h7 f7g6');
INSERT INTO evaluations VALUES(49761,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(49761,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(49761,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(49761,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(49761,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(49761,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(49761,36,'cp -103','f8b4 f1d3 d7d5 e4e5 f6e4 d3e4 b4c3 b2c3 d8h4 e1f1 d5e4 d1e1 h4e1 f1e1 c8e6 c1f4 b8d7 g1e2 e6c4 a2a4 f7f6 h1f1 e8c8 e5f6 d7f6 f4g5 d8g8 f1f5 f6d5 e1d2 g7g6 f5e5 g8e8 e2g3 e8e5 d4e5 h8e8 g3e4 e8e5 a1e1 b7b6 h2h3 e5f5 g2g3 c8b7 d2c1');
INSERT INTO evaluations VALUES(42819,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(42819,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(42819,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(42819,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(42819,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(42819,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(42819,36,'cp -107','f1d3 d7d5 e4e5 f6e4 d3e4 b4c3 b2c3 d8h4 e1f1 d5e4 d1e1 h4e1 f1e1 c8e6 c1f4 b8d7 g1e2 e6c4 a2a4 f7f6 a4a5 f6e5 d4e5 e8c8 h1f1 a7a6 f4g3 d7c5 a1b1 e4e3 f1f4 c4d5 b1d1 c5e6 f4b4 d5g2 d1d8 e6d8 b4g4 g7g6 g3h4 g2d5 e2f4 d5e4 h4f6 h8e8 g4g3 e3e2 f4e2 e4c2 e2f4 d8c6');
INSERT INTO evaluations VALUES(50690,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(50690,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(50690,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(50690,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(50690,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(50690,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(50690,36,'cp 49','h2h3 b8c6 d4e3 g7g6 f1e2 f8g7 e1g1 g8f6 c2c4 f6d7 b1c3 d8b6 e3d2 d7c5 c3d5 b6d8 d2c2 a6a5 c1d2 c8d7 d2c3 e8g8 c3g7 g8g7 d5c3 a5a4 c2d2 d8b6 f1d1 d7e6 a1b1 f7f6 d2e3 c5d7 e3b6 d7b6 c3d5 b6d5 c4d5');
INSERT INTO evaluations VALUES(51103,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51103,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(51103,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(51103,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(51103,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(51103,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(51103,36,'cp 23','d7d5 b1d2 c7c5 d4c5 b8c6 g1f3 e7e6 e2e4 f8c5 f1d3 g8f6 e1g1 e8g8 c2c3 f6h5 f4g5 f7f6 g5e3 c5e3 f2e3 g7g6 f3d4 f6f5 e4d5 e6d5 d3f5 d8g5 f5c8 g5e3 g1h1 a8c8 f1e1 e3d3 d2f3 d3d1 a1d1 c6d4 d1d4 h5f6 e1e7 f8f7 e7e6 c8c4');
INSERT INTO evaluations VALUES(52381,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(52381,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(52381,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52381,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(52381,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(52381,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(52381,36,'cp 28','c6c5 d2d4 c5d4 d1d4 g8f6 d4f2 b8c6 c1d2 d7d5 e4d5 f6d5 b1c3 d5c3 d2c3 e7e5 f1d3 c8e6 g1e2 d8c7 f3f4 f8d6 e1c1 e8g8 a2a3 e5f4 f2h4 h7h6 g2g3 c7d8 h4d8 a8d8 e2f4 e6g4 d1e1 f8e8 d3e4 g8f8 h1f1 h6h5 f4g6 f8g8 e4d5 e8e1 f1e1');
INSERT INTO evaluations VALUES(52599,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(52599,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(52599,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(52599,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(52599,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(52599,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52599,36,'cp 36','g1f3 f8g7 d2d4 c5d4 f3d4 g8f6 f2f3 d8b6 c4b5 c8d7 c1e3 b6c7 g2g4 b8c6 d1d2 c6d4 e3d4 d7b5 c3b5 c7c4 d2d3 c4d3 c2d3 e8d7 b5a7 h7h5 g4g5 f6e8 a7b5 a8a5 a2a4 g7d4 b5d4 a5g5 e1f2 g5c5 b2b4 c5c3 f2e3 e7e5 d4e2 c3c2 h1c1 c2b2 c1b1 b2c2 b4b5 f7f5 e4f5 g6f5 d3d4');
INSERT INTO evaluations VALUES(54097,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(54097,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(54097,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(54097,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(54097,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(54097,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(54097,36,'cp 44','e2e4 c5d4 d1d4 d8d4 f3d4 c8d7 f1c4 b8c6 d4b3 g8f6 f2f3 f8b4 e1f2 e8g8 c4e2 f8c8 c1d2 c6a5 b3a5 b4a5 h1c1 d7e8 d2e3 a7a6 a2a3 h7h6 g2g3 e8c6 b2b4 a5d8 a3a4 f6d7 a4a5 d8e7 a1b1 d7e5 f3f4 e5d7 c1d1');
INSERT INTO evaluations VALUES(54099,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(54099,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(54099,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(54099,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(54099,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(54099,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(54099,36,'cp 36','c5d4 d1d4 d8d4 f3d4 c8d7 f1c4 b8c6 d4b3 f8b4 f2f3 g8f6 c4e2 e8g8 e1f2 f8c8 c1d2 c6a5 b3a5 b4a5 h1d1 d7e8 a1c1 a7a6 g2g3 h7h6 d2e3 c8c7 e4e5 a5c3 e5f6 c3e5 c1c7 e5c7 d1c1 c7e5 f6g7 e5g7');
INSERT INTO evaluations VALUES(54452,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(54452,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(54452,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(54452,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(54452,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(54452,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(54452,36,'cp 37','d1d4 d8d4 f3d4 c8d7 f1c4 b8c6 d4b3 f8b4 c1e3 g8f6 f2f3 e8g8 a1c1 f8c8 c4e2 c6e5 b3d2 b4c3 c1c3 c8c3 b2c3 d7a4 e1f2 b7b6 h1b1 a8c8 c3c4 f6d7 f3f4 e5g6 e4e5 d7c5 e3c5 c8c5 g2g3 g6e7 d2b3 a4b3 a2b3 e7c6 b1d1');
INSERT INTO evaluations VALUES(54109,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(54109,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(54109,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(54109,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(54109,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(54109,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(54109,36,'cp -73','b8c6 e2g3 h7h5 f1e2 h5h4 g3f1 h4h3 g2g3 e6e5 c2c3 c8e6 f3g5 e6d7 g5f3 f8e7 c3d4 c5d4 f1d2 g7g5 e1g1 f7f6 f3e1 g8h6 d2c4 d7e6 f2f4 g5f4 g3f4 e5f4 g1h1 b7b5 c4a3 d8d7 c1f4 h6g4 a3b5 c6e5 b5c7 d7c7 e2g4 e6g4 f4e5');
INSERT INTO evaluations VALUES(20214,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(20214,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(20214,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(20214,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(20214,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(20214,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(20214,36,'cp -48','g2g3 e6e5 f1g2 f8e7 e1g1 h7h5 c2c3 a7a5 f3d2 h5h4 g3h4 d4c3 b2c3 d8d3 d2f3 d3c4 f1e1 g8f6 c1g5 e8g8 a2a4 f8d8 d1b1 a8b8 b1b5 c8e6 h4h5 c4d3 h2h3 d8d7 h5h6 b8d8 h6g7 e6c4 b5b2 f6e4 g5e7 c6e7 f3e5');
INSERT INTO evaluations VALUES(54794,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(54794,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(54794,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(54794,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(54794,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(54794,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(54794,36,'cp 195','c8f5 c2c4 e7e6 b1c3 b8c6 c4d5 e6d5 h2h4 f8d6 d1f3 g8e7 h4g5 f6g5 f3h5 e8d7 g3d6 c7d6 h5g5 d8b6 f1b5 a8e8 e1c1 h8g8 g5h4 d7c8 g1e2 c8b8 b5d3 f5d3 d1d3 g8g2 e2f4 g2g8 h4h7 a7a6 d3d2 b8a7 c1b1 e8f8 a2a3');
INSERT INTO evaluations VALUES(54952,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54952,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(54952,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(54952,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(54952,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(54952,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54952,36,'cp 0','c2c4 e7e5 b1c3 g8f6 g1f3 b8c6 b4b5 c6d4 e2e3 d4f3 d1f3 a7a6 b5a6 a8a6 h2h3 g7g6 f1e2 a6a8 d2d4 f8g7 a2a4 e8g8 e1g1 c7c6 a1a3 c8f5 d4d5 e5e4 f3f4 c6d5 c4d5 f6h5 e2h5 g7e5 f4h6 e5g7');
INSERT INTO evaluations VALUES(56223,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(56223,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(56223,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(56223,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(56223,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(56223,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(56223,36,'cp -89','d1e2 d5e4 e2e4 f8e7 e4f4 g8f6 f1c4 e8g8 g1e2 b8c6 c2c3 c8g4 e1g1 c6a5 c4d3 c7c5 b1a3 g4h5 e2g3 h5g6 g3f5 a5c6 a3c2 c5d4 c3d4 f6d5 f4f3 e7f6 g1h1 d5e7 c1e3 e7f5 d3f5 g6f5 f3f5 c6e7 f5d3 a8c8 a1d1 d8d5 e3f4 f8d8 c2e3');
INSERT INTO evaluations VALUES(1094,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(1094,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(1094,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(1094,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(1094,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(1094,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(1094,36,'cp -40','h5e2 c6a5 c4b5 a7a6 b5a4 g7g6 d2d3 f8g7 g1f3 e8g8 e1g1 d7d6 c1e3 b7b5 a4d1 c7c5 a2a4 c8e6 b2b4 c5b4 c3b4 a5c6 e2b2 f6g4 e3d2 d6d5 f3g5 g4f6 g5e6 f7e6 b1c3 c6b4 c3d5 f6d5 e4d5 b4d3 b2a3 e7a3 a1a3 e5e4');
INSERT INTO evaluations VALUES(2440,30,'cp -452',NULL);
INSERT INTO evaluations VALUES(2440,31,'cp -432',NULL);
INSERT INTO evaluations VALUES(2440,32,'cp -440',NULL);
INSERT INTO evaluations VALUES(2440,33,'cp -459',NULL);
INSERT INTO evaluations VALUES(2440,34,'cp -463',NULL);
INSERT INTO evaluations VALUES(2440,35,'cp -455',NULL);
INSERT INTO evaluations VALUES(2440,36,'cp -475','c4e2 d8d5 e1g1 c8e6 b1c3 d5d7 e2f3 a5c6 c3e4 h7h6 d2d3 c6d4 c1d2 f8e7 d2c3 d4f3 d1f3 h8f8 c3e5 f7g8 e5f6 e7f6 e4c5 d7f7 c5e6 f7e6 f3b7 a8b8 b7a7 e6c6 a7e3 f6b2 a1e1 b8e8 e3a7 e8e1 f1e1 c6c2 a7e3');
INSERT INTO evaluations VALUES(3173,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3173,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(3173,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3173,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(3173,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(3173,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3173,36,'cp 0','d1d2 c8g4 f3e1 g4e6 a1d1 e6g4');
INSERT INTO evaluations VALUES(3624,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(3624,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(3624,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(3624,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(3624,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(3624,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(3624,36,'cp -19','a2a4 c6e7 c1e3 c5e3 f2e3 c7c6 b1c3 e7g6 d3d4 c8e6 c4d3 e5d4 e3d4 c6c5 a4a5 a7a6 d1d2 f8e8 h2h3 c5d4 f3d4 g6e5 c3e2 a8c8 e2f4 e6d7 b2b3 c8c5 b3b4 c5c8 d2f2 b7b5 f4d5 f6d5 e4d5 c8c3');
INSERT INTO evaluations VALUES(4918,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(4918,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(4918,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(4918,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(4918,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(4918,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(4918,36,'cp 30','d7d5 d2d4 e5d4 e1g1 g8e7 c3d4 c5b6 e4d5 c6b4 c2b3 b4d5 f1e1 h7h6 b3d5 d8d5 a2a4 b5b4 a4a5 b6a7 d1a4 d5b5 c1d2 c8e6 d2b4 b5a4 a1a4 e7d5 b4a3 e8c8 a3c5 a7c5 d4c5 h8e8 a4d4 g7g5 h2h4 d5f4 d4d8 e8d8 h4g5 h6g5 f3g5 e6d5 b1c3 d5g2');
INSERT INTO evaluations VALUES(5324,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(5324,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5324,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(5324,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(5324,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(5324,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(5324,36,'cp 62','d2d4 d7d6 d4d5 g7c3 b2c3 c6a5 e3e4 e7e5 f1d3 h7h6 f3d2 g8f6 d1e2 f6h5 g2g3 b7b6 d2b3 a5b7 c1e3 c8h3 f2f3 f7f5 e4f5 g6f5 h1g1 d8e7 g3g4 f5g4 f3g4 h5f4 e3f4 e5f4 d3g6 e8d8 e2e7 d8e7 b3d2 a8g8 g6f5 e7d8 d2e4');
INSERT INTO evaluations VALUES(5333,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(5333,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(5333,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(5333,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(5333,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(5333,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(5333,36,'cp 56','d7d6 d4d5 g7c3 b2c3 c6a5 e3e4 e7e5 f3d2 b7b6 f1d3 d8e7 e1g1 g6g5 f2f3 h7h6 d2b3 a5b3 a2b3 a7a5 b3b4 g8f6 g2g3 e8g8 g1g2 g8g7 f1f2 f8g8 f2a2 e7d8 c1e3 c8d7 d1e2 g7h8 g3g4 f6h7 e2d2 h8g7 e3f2 a5a4 d3c2');
INSERT INTO evaluations VALUES(5741,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(5741,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(5741,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(5741,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(5741,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(5741,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(5741,36,'cp 7','c2c3 g8f6 d2d4 c5b6 f1e1 e8g8 a2a4 a7a5 c1e3 e5d4 c3d4 d6d5 e4d5 c6b4 e3f4 f6d5 f4g3 f8e8 e1e8 d8e8 d1b3 c8e6 b1c3 c7c6 a1e1 e8e7 c3e4 d5c7 g3d6 e6c4 b3c4 e7e6 f3e5 c7d5 d6c5 a8d8');
INSERT INTO evaluations VALUES(5732,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(5732,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(5732,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5732,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(5732,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(5732,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(5732,36,'cp 11','c5b6 a2a4 g8f6 f1e1 a7a5 b1a3 e8g8 d2d3 c6e7 c4a2 c7c6 a3c4 b6c7 d3d4 e5d4 c4d6 c7d6 e4e5 d6e5 f3e5 e7f5 c3d4 f6d5 a2b1 f7f6 e5g4 f8e8 a1a3 e8e1 d1e1 g8f7 g4e3 f5d4 e3d5 d8d5');
INSERT INTO evaluations VALUES(8951,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(8951,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(8951,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(8951,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(8951,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(8951,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(8951,36,'cp 59','b4d6 b1d2 f6d7 f4d6 c7d6 e3e4 d5e4 d2e4 d7f6 e4f6 d8f6 e1g1 e8g8 f1e1 c8d7 d3e4 h7h6 d4d5 e6d5 d1d5 f8d8 a1d1 d7e6 d5b5 a8b8 a2a3 e6d7 b5e2 c6e5 f3d4 a7a6 b2b3 d8e8 e2d2 d7c6 e4c2 b8d8 h2h3 d6d5 e1e3 g7g6 f2f3 g8g7');
INSERT INTO evaluations VALUES(9240,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(9240,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(9240,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(9240,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(9240,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(9240,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(9240,36,'cp 60','b1d2 f6d7 f4d6 c7d6 e3e4 d5e4 d2e4 d7f6 e4f6 d8f6 e1g1 e8g8 f1e1 c8d7 d3e4 a8b8 d4d5 e6d5 e4d5 c6e5 e1e3 d7c6 d1d2 f8e8 a1d1 b8d8 h2h3 h7h6 f3h2 c6d5 d2d5 f6e6 d5a5 e5c4 e3e6 c4a5 e6e8 d8e8 g1f1 e8d8 h2f3 d6d5 a2a4 b7b6 h3h4 a5c4');
INSERT INTO evaluations VALUES(10253,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10253,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10253,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10253,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10253,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10253,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10253,36,'cp 0','b1c3 e8g8 g1e2 a7a5 e1g1 f8e8 h2h3 c6d4 c3a4 c5a7 e2d4 a7d4 c2c3 d4a7 c3c4 d7d6 g1h2 c8d7 a4c3 a7d4 c3e2 d4c5 e2c3 c5d4');
INSERT INTO evaluations VALUES(13320,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(13320,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(13320,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(13320,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(13320,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(13320,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(13320,36,'cp -19','e7e5 c3c4 g7g6 b1c3 f8g7 f1e2 e8g8 a2a3 b7b6 e1g1 f6e8 b2b4 e8c7 b4c5 d6c5 c3d5 c7e6 c1d2 a8b8 a3a4 a7a5 d2c3 f7f5 a1b1 c8d7 b1b2 c6d4 d1a1 g8h8 a1d1 d7e8 f1e1 e8c6 b2b1');
INSERT INTO evaluations VALUES(13557,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(13557,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(13557,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(13557,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(13557,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(13557,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13557,36,'cp 25','c4d5 e6d5 e1g1 c5c4 b2b3 c4b3 a2b3 c8f5 f3e5 f8e7 e5c6 b7c6 b1d2 e8g8 c1a3 a7a5 d1c1 h7h5 d2f3 e7a3 c1a3 d8b6 f1c1 f6e4 e2e3 f7f6 f3h4 f5h7 g2e4 h7e4 c1c5 a8a6 a1c1 f8b8 f2f3');
INSERT INTO evaluations VALUES(13517,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(13517,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(13517,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(13517,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(13517,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(13517,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(13517,36,'cp 44','e6d5 e1g1 c5c4 f3e5 f8d6 e5c6 b7c6 b2b3 c4b3 a2b3 c8f5 c1a3 a7a5 a3d6 d8d6 b1d2 e8g8 a1a4 h7h5 h2h4 f8e8 e2e3 f5d3 f1e1 d3b5 a4a2 a8a7 d1c2 d6b4 e1a1 e8a8 d2f3 a5a4 b3a4 a7a4 a2a4 a8a4 a1a4 b5a4 c2c1 b4c4 c1d2 c4c2 d2c2 a4c2');
INSERT INTO evaluations VALUES(16669,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(16669,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(16669,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(16669,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(16669,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(16669,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(16669,36,'cp -63','d7d5 e4d5 f6d5 f1e2 c8e6 c4a4 d5b6 a4e4 d8d5 e4d5 b6d5 e1g1 h7h6 b1d2 e8c8 c2c3 d5f4 e2c4 g7g5 g2g3 e6c4 d2c4 f4d3 a2a4 f7f5 f3e1 d3c5 c1e3 c5b3 a1b1 f8c5 e3c5 b3c5 f2f4 g5g4 e1c2 c5a4 c4e5 c6e5 f4e5 d8d2 c2d4');
INSERT INTO evaluations VALUES(15028,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(15028,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(15028,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(15028,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(15028,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(15028,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(15028,36,'cp -61','e4d5 f6d5 f1e2 c8e6 c4a4 d5b6 a4e4 d8d7 f3g5 d7d5 e1g1 d5e4 g5e4 c6d4 e2d1 e8c8 f1e1 h7h6 b1d2 g7g6 c2c3 d4c6 d1b3 e6d5 e4f6 d5b3 a2b3 f8g7 d2e4 a7a5 c1e3 b6d5 f6d5 d8d5 a1d1 d5d1 e1d1 h8d8 d1d2 d8d2 e4d2');
INSERT INTO evaluations VALUES(14976,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(14976,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(14976,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(14976,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(14976,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(14976,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(14976,36,'cp -52','f6d5 f1e2 c8e6 c4a4 f8c5 e1g1 e8g8 b1c3 d5c3 b2c3 e6d5 a4g4 d8f6 e2d3 d5f3 g4f3 f6f3 g2f3 c6e5 d3e4 c7c6 f1e1 f7f6 c1f4 f6f5 f4e5 f5e4 e1e4 f8f3 e5d4 c5d4 c3d4 a8f8');
INSERT INTO evaluations VALUES(17245,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(17245,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(17245,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(17245,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(17245,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(17245,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(17245,36,'cp 47','e7e5 d4d5 f8b4 b1d2 c6e7 g1f3 b4d2 f3d2 e8g8 f1d3 d7d6 b2b3 c7c6 d5c6 e7c6 e1g1 d6d5 c1b2 c8g4 d3e2 g4e2 d1e2 d8e7 c4d5 f6d5 d2e4 f8d8 a1d1 h7h6 h2h3 d5f6 e4g3 d8d1 f1d1 a8d8 g1h2 d8d1 e2d1 e7d7 d1e1 a7a6 e1c3');
INSERT INTO evaluations VALUES(17293,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(17293,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(17293,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(17293,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(17293,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(17293,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(17293,36,'cp 54','d2d4 e7e6 c2c4 d7d5 b1c3 f8e7 a2a3 e8g8 b2b4 d5c4 f1c4 e7d6 e1g1 e6e5 h2h3 e5d4 e3d4 h7h6 f1e1 c6e7 d1b3 c7c6 f3e5 e7d5 c3d5 c6d5 c4d5 f6d5 b3d5 d6c7 d5f3 d8d4 c1f4 c8e6 a1d1 d4b6 e5d7 e6d7 d1d7 c7f4 f3f4 a8d8 e1e7');
INSERT INTO evaluations VALUES(20767,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(20767,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(20767,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(20767,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(20767,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(20767,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(20767,36,'cp 105','a7a6 f1c4 g8f6 c1g5 f8e7 d1d3 e8g8 c4b3 f6e8 g5e7 d8e7 c3d5 e7g5 d3g3 g5d8 e1c1 c8e6 d1d2 e6d5 b3d5 e8f6 h1d1 d8c7 f2f3 a6a5 a2a3 a8a6 g3f2 f6d5 d2d5 a6c6 c1b1 f8c8 c2c3 g8f8');
INSERT INTO evaluations VALUES(21388,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(21388,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(21388,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(21388,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(21388,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(21388,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(21388,36,'cp 71','g1f3 g8f6 b1c3 f8e7 c1f4 f6e4 e2e3 e4c3 b2c3 e8g8 c4d5 e6d5 f1d3 c6a5 e1g1 c8g4 e3e4 d5e4 d3e4 e7d6 f4g3 d6g3 h2g3 a8c8 d1c2 g7g6 f3e5 g4e6 f2f4 f7f6 e5d3 e6d5 e4d5 d8d5 f4f5 a5c4 a1e1 c8e8 e1e8 f8e8 f5g6 h7g6 f1f6');
INSERT INTO evaluations VALUES(21519,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21519,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(21519,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(21519,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21519,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21519,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21519,36,'cp 0','f8b4 g1f3 g8e7 a2a3 b4c3 b2c3 e8g8 e1g1 a7a6 b5d3 c8f5 c1f4 d8d7 f1e1 f8e8 a3a4 h7h6 a4a5 f5d3 d1d3 e7g6 f4e3 c6e7 a1b1 e7c6');
INSERT INTO evaluations VALUES(21671,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21671,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21671,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21671,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21671,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21671,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21671,36,'cp 0','e5d4 c4d5 d8d5 b1c3 f8b4 e3d4 c8g4 f1e2 g4f3 e2f3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e7d5 e2d3 e8d7 h1c1 a8b8 c3a4 h8e8 g2g3 b4d6 e3d2 e8e6 c1c4 b8e8 a4c5 d6c5 c4c5 e6f6 d2e3 f6e6 e3d2');
INSERT INTO evaluations VALUES(23239,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(23239,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(23239,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(23239,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(23239,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(23239,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(23239,36,'cp 143','f1c4 g8e7 d2d4 d7d5 e4d5 e7d5 e1g1 d5b6 c4b5 a7a6 b5c6 b7c6 d4e5 d8d1 f1d1 c8g4 b1d2 e8c8 d1e1 d8e8 e5f6 e8e1 f3e1 g7f6 b2b3 f8g7 h2h3 g4h5 d2f3 f6f5 c1e3 g7c3 a1c1 c3e1 f3e1 c8b7 e1d3 h8e8 d3c5 b7a7 e3d4 b6d5 a2a3 h5g6 f2f3 f5f4 g1f2');
INSERT INTO evaluations VALUES(23759,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(23759,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(23759,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(23759,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(23759,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(23759,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(23759,36,'cp 50','d4c6 d7c6 b5c6 b7c6 d1d3 f8e8 c1d2 a7a5 a1d1 f6d7 b2b3 e7f6 c3a4 d7e5 d3g3 e5d7 d2c3 h7h6 g3d3 d8b8 h2h3 b8a7 e1e2 f6c3 d3c3 d7e5 g1h2 c6c5 c3d2 e5c6 e2e3 f7f6 f2f4 c6d4');
INSERT INTO evaluations VALUES(23747,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(23747,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(23747,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(23747,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(23747,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(23747,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(23747,36,'cp 44','d7c6 b5c6 b7c6 f2f4 a7a5 d1f3 a5a4 g1h1 f6d7 c1e3 e7f6 e3f2 c6c5 e4e5 d6e5 a1d1 d8c8 c3d5 e5f4 f3f4 c7c6 d5f6 d7f6 f2c5 f8d8 c5d4 c8g4 f4g4 f6g4 h2h3 g4f6 d4f6 g7f6 h1h2 d8d1 e1d1 a8e8 d1d4 e8b8 b2b3 a4b3 c2b3');
INSERT INTO evaluations VALUES(27086,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(27086,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(27086,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(27086,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(27086,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(27086,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(27086,36,'cp 59','c4e2 c7c6 c1e3 f8e7 f3e5 g4e2 e1e2 d6e5 d1d8 e8d8 h1d1 d8c7 b1d2 g8f6 f2f3 f6d7 d2c4 f7f6 a2a4 d7b6 c4d2 b6d7 c2c3 e7c5 a4a5 c5e3 e2e3 d7c5 b2b4 c5e6 h2h4 b7b5 g2g3 g7g5 h4h5 h8g8 d2b3 a8d8 d1d3 d8d3 e3d3 e6g7 h5h6 g8d8 d3c2');
INSERT INTO evaluations VALUES(27088,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(27088,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(27088,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(27088,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(27088,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(27088,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(27088,36,'cp 50','c7c6 f3e5 g4e2 e1e2 d6e5 d1d8 e8d8 a2a4 g8f6 f2f3 d8c7 a4a5 f8e7 c1e3 f6d7 h1d1 d7c5 b1d2 c5e6 g2g3 e7c5 c2c3 c5e3 e2e3 f7f6 h2h4 a8d8 b2b4 g7g5 h4g5 e6g5 d2c4 d8d1 a1d1 g5f7 f3f4 b7b5 c4b2 h7h5 b2d3 h8g8');
INSERT INTO evaluations VALUES(30959,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(30959,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(30959,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(30959,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(30959,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(30959,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(30959,36,'cp 97','h4h5 g6e4 h1h3 e7e6 b1c3 e4h7 c1e3 f8b4 d1d2 d8a5 g1e2 b8d7 a2a3 g8e7 d2d1 b4c3 e2c3 a5b6 c3a4 b6a5 d1d2 a5d2 e1d2 b7b6 f1e2 a8c8 a4c3 c6c5 e2a6 c8d8 a6b5 e8g8 a1g1 c5d4 e3d4 d7c5 f2f4 c5e4 d2e1 g8h8 c3e4');
INSERT INTO evaluations VALUES(30956,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(30956,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(30956,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(30956,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(30956,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(30956,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(30956,36,'cp 125','g6e4 f2f3 e4h7 e5e6 c6c5 d4c5 f7e6 f1d3 h7d3 d1d3 e6e5 c2c4 g8f6 d3g6 e8d7 c4d5 f6d5 b1c3 e7e6 c3d5 e6d5 g6f5 d7e8 f5e5 d8e7 e5e7 f8e7 c1e3 b8c6 g1e2 e8f7 e1c1 a8d8 c1b1 h8e8 e2d4 e7f6 d4f5 d5d4 e3f4 e8e2 d1d3 f6e5 f4e5 c6e5 d3b3 d4d3 h1d1 f7f6 f5d6 e2e3 b3a3 e3f3 a3a7 f3e3 a7b7 f6e6 d6b5 e5c4');
INSERT INTO evaluations VALUES(33083,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(33083,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(33083,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(33083,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(33083,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(33083,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(33083,36,'cp 172','d2d4 c5b6 a2a4 a7a5 d4e5 d6e5 c1g5 f6d6 b1a3 f7f6 d1d6 c7d6 a3b5 e8e7 g5c1 b8c6 c4d5 c6a7 f3d2 a7b5 a4b5 b6c7 f2f3 h7h6 d2c4 c8e6 d5e6 e7e6 c1e3 g8e7 b5b6 c7d8 b2b3 e7c6 e1e2 d8e7 c4a3 h6h5 h1d1 a8b8 h2h4 f6f5 e4f5 e6f5 a3b5 f5e6');
INSERT INTO evaluations VALUES(33088,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(33088,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(33088,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(33088,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(33088,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(33088,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(33088,36,'cp 181','c5b6 a2a4 a7a5 d4e5 d6e5 c1g5 f6d6 b1a3 f7f6 d1d6 c7d6 a3b5 e8e7 g5c1 b8c6 f3d2 c6a7 c4d5 a7b5 a4b5 b6c7 f2f3 c8d7 c3c4 a8b8 d2b1 e7e8 e1e2 g8e7 b1c3 c7d8 h1d1 d8b6 c1d2 h7h6 b2b3 h8f8 d2e3 b6e3 e2e3 b7b6 g2g3 f6f5 e4f5 e7f5 e3f2');
INSERT INTO evaluations VALUES(34863,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(34863,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(34863,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(34863,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(34863,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(34863,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(34863,36,'cp 154','d2d4 f6d8 c1g5 g8f6 d1d3 f8e7 g5f6 e7f6 e1c1 e8g8 h2h4 b8c6 c3e4 c6b4 d3d2 f6e7 a2a3 b4d5 g2g4 e7d6 f3e5 f7f6 e5d3 b7b5 g4g5 a7a5 h1g1 f6g5 g1g5 d6e7 d3e5 e7g5 h4g5 b5b4 a3a4 d5e7 e2d3 d8d4');
INSERT INTO evaluations VALUES(34861,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(34861,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(34861,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(34861,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(34861,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(34861,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(34861,36,'cp 143','f6d8 f3e5 f8d6 e2d3 g8f6 d1f3 c7c5 c1e3 c5d4 e3d4 d6e5 d4e5 b8c6 e5g3 e8g8 e1c1 d8a5 c3e4 f6e4 f3e4 g7g6 a2a3 e6e5 e4e3 c8e6 h2h4 f7f6 h4h5 g6g5 f2f4 g5f4 g3f4 f8f7 g2g4 c6d4 d3h7 g8h7 d1d4 e6g4 e3e4 f6f5');
INSERT INTO evaluations VALUES(35140,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(35140,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(35140,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(35140,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(35140,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(35140,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(35140,36,'cp 71','c2c4 d7d6 g2g3 f8e7 b1c3 f6g6 d2d4 g8f6 f1g2 b8c6 g1f3 e8g8 f3h4 g6h5 d4d5 c6b8 d1c2 b8a6 a2a3 f6e8 g2f3 h5h6 b2c1 e7g5 c1g5 h6g5 e2e4 a6b8 e1g1 g7g6 b3b4 g5e7 f3g2 b8d7 f1e1 d7b6 g2f1 a7a5 c2d2 c8d7 h4f3 f7f6 f1d3 e7d8 a3a4');
INSERT INTO evaluations VALUES(35142,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(35142,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(35142,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(35142,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(35142,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35142,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35142,36,'cp 42','f6g6 d1d4 b8c6 d4e3 f8b4 c2c3 b4e7 b1d2 e7g5 f3g5 f7f6 g5f3 d7d6 f1e2 c8d7 e1g1 g8h6 f3d4 e8g8 e3g3 g6g3 h2g3 f6f5 f1e1 a8e8 d4b5 e8c8 e4f5 h6f5 g3g4 f5e7 e2c4 d6d5 c4b3 f8e8 f2f3 a7a6 b5d4 c6d4');
INSERT INTO evaluations VALUES(35155,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(35155,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(35155,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(35155,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(35155,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(35155,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(35155,36,'cp 44','f1d3 d7d6 g5f4 c8g4 b1d2 b8c6 e1g1 e8c8 f1e1 g6f6 f4g3 c6e5 d3e2 g8e7 f3d4 h7h5 e2g4 e5g4 d2f1 g4e5 h2h3 g7g5 f1e3 g5g4 h3h4 f8h6 e3f5 h8e8 f5e7 f6e7 d4f5 e7e6 c2c3 a7a6 a2a4 d6d5 d1c2 d5e4 f5h6 e6h6');
INSERT INTO evaluations VALUES(42789,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(42789,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(42789,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(42789,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(42789,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42789,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(42789,36,'cp 29','b2c3 e8g8 g1e2 d7d5 e4d5 f6d5 e1g1 b8c6 c4b3 f8e8 h2h3 b7b6 e2g3 c6a5 c3c4 d5e7 b3a4 c8d7 a4d7 d8d7 g3h5 d7d4 c1e3 d4h4 g1h2 e7f5 d1f3 f5d4 e3d4 h4d4 a1e1 e5e4 f3g4 d4d6 f2f4 e4f3 h2g1 d6g6 g4f3 a5c6 h5f4 g6d6');
INSERT INTO evaluations VALUES(43106,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(43106,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(43106,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(43106,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(43106,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43106,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(43106,36,'cp 14','f6e4 e2e3 e4c3 b2c3 c7c5 f1e2 d7d6 e1g1 b8c6 d4c5 d6c5 d1d5 d8d5 c4d5 c6e5 f1c1 b7b6 f3e5 g7e5 f2f3 c8d7 a2a4 e8f7 e2c4 h7h6 g5f4 e5f4 e3f4 h8e8 c1b1 a8b8 g2g3 g6g5 g1f2 g5f4 g3f4 e7e6 a4a5 e6d5 c4d5 d7e6 d5e6 e8e6');
INSERT INTO evaluations VALUES(44043,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(44043,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(44043,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(44043,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44043,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(44043,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(44043,36,'cp 34','g1f3 d7d5 b2b3 b8d7 e1g1 f8d6 d2d4 f6e4 d1c2 a7a5 c1a3 d6a3 b1a3 e8g8 a1c1 d8e7 a3b1 h7h6 f1d1 a5a4 b1d2 e4d2 f3d2 f8e8 e2e4 d5e4 d2e4 d7f6 c2e2 f6e4 e2e4 a4b3 a2b3 c8d7');
INSERT INTO evaluations VALUES(44042,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(44042,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44042,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(44042,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(44042,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(44042,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(44042,36,'cp 33','d7d5 d2d4 f8e7 b2b3 e8g8 e1g1 b7b6 d1c2 b8d7 c1b2 c8b7 f1d1 a8c8 b1c3 c6c5 f3e1 c5d4 d1d4 e7c5 d4d1 b7a8 c4d5 f6d5 c2d2 d5c3 b2c3 a8g2 g1g2 d7f6 d2f4 d8e7 c3f6 e7f6 f4f6 g7f6 d1d7 a7a5 a2a4 f8d8 d7d8 c8d8 e1d3 g8f8 a1c1 d8c8 c1c4 f8e8');
INSERT INTO evaluations VALUES(45945,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(45945,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(45945,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(45945,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(45945,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(45945,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(45945,36,'cp 139','e4d5 b8a6 g1f3 c8f5 f1d3 f5d3 d1d3 f8e7 h2h3 a6b4 d3b1 e8g8 e1g1 f8e8 c1f4 h7h6 a2a3 b4a6 b1d3 a6b8 a1e1 b8d7 g2g4 a7a6 a3a4 d8c7 g1g2 e7f8 b2b3 e8e1 f1e1 a8e8 e1e8 f6e8 c3e4 f8e7 f4d2 d7f8 e4g3 c7d7 g3f5 e7d8');
INSERT INTO evaluations VALUES(47327,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(47327,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(47327,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47327,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47327,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(47327,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(47327,36,'cp 0','e7e6 g1f3 a7a6 a2a4 c7c5 e1g1 b8c6 c3e4 f8e7 c1f4 e8g8 h2h3 b7b6 d1e2 c8b7 c2c3 d8d7 f4g3 a8d8 e4g5 f6d5 e2e4 d5f6 e4e2');
INSERT INTO evaluations VALUES(48463,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48463,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(48463,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(48463,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(48463,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(48463,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48463,36,'cp 14','d3d4 a7a6 a2a4 f8b4 g1f3 c7c5 e1g1 c5d4 c3e2 e8g8 d1d4 d8d4 e2d4 f8d8 f1d1 c8d7 c2c3 b4e7 d4b3 b8c6 c1e3 d7e8 d1d8 a8d8 a4a5 d8c8 f3d2 g8f8 g2g3 f6d5 c4d5 e6d5 b3c5 c6d8 d2b3 e8c6');
INSERT INTO evaluations VALUES(47810,30,'cp -170',NULL);
INSERT INTO evaluations VALUES(47810,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(47810,32,'cp -186',NULL);
INSERT INTO evaluations VALUES(47810,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(47810,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(47810,35,'cp -185',NULL);
INSERT INTO evaluations VALUES(47810,36,'cp -161','b8c6 c2c3 c8g4 f3g3 g4h3 g3h3 f6e4 d2d3 e4c5 h3f3 d8f6 f3f6 g7f6 c1e3 c5e6 b1d2 e8c8 d2f3 d6d5 c4b3 f8c5 e1c1 d5d4 e3h6 c5f8 h6f8 h8f8 b3c4 d8d6 c1c2 d4c3 b2c3 c6a5 c4b3 f8d8 g2g3 a5b3 a2b3 h7h6 b3b4');
INSERT INTO evaluations VALUES(15504,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(15504,31,'cp -171',NULL);
INSERT INTO evaluations VALUES(15504,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(15504,33,'cp -159',NULL);
INSERT INTO evaluations VALUES(15504,34,'cp -170',NULL);
INSERT INTO evaluations VALUES(15504,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(15504,36,'cp -178','c2c3 c8g4 f3g3 g4h3 g3h3 f6e4 d2d3 e4c5 h3f3 d8f6 f3f6 g7f6 c1e3 c6a5 c4b5 c7c6 e3c5 d6c5 b5a4 e5e4 a4c2 e4d3 c2d3 e8c8 e1e2 f8h6 h1d1 a5c4 b2b3 c4d6 e2f1 c8c7 b1a3 h6g7 d1e1 d8d7 a1c1 f6f5 a3c4 d6c4 d3f5 c4e5 f5d7 c7d7 c1d1 d7c7 f2f4 e5g4');
INSERT INTO evaluations VALUES(50476,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(50476,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(50476,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(50476,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(50476,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(50476,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(50476,36,'cp -48','g1f3 d8c7 e1g1 b8c6 f1e1 d7d6 e5d6 f8d6 c2c3 h7h6 d2d4 c5d4 c3d4 g8f6 b1c3 e8g8 c4d3 b7b5 a2a4 b5b4 c3e4 f6e4 d3e4 c8b7 c1e3 c6e7 a1c1 c7b8 e4b7 b8b7 f3d2 a8c8 d1f3 b7d5 d2e4 c8c1 e1c1 e7f5 e4d6 d5d6');
INSERT INTO evaluations VALUES(50475,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(50475,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(50475,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(50475,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(50475,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(50475,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(50475,36,'cp -62','d8c7 e1g1 b8c6 f1e1 f7f6 d2d4 c5d4 e5f6 g8f6 c2c3 b7b5 c4b3 f8d6 c3d4 e8g8 c1e3 c6a5 b1d2 a5b3 d2b3 c8b7 a1c1 c7b8 h2h3 d6b4 e3d2 b4d2 f3d2 b8f4 d1e2 b7c6 b3c5 c6d5 e2e3 f4e3 e1e3 d5a2');
INSERT INTO evaluations VALUES(53002,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(53002,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(53002,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(53002,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(53002,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(53002,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(53002,36,'cp -41','c5b4 d2d4 g8f6 f1d3 e7e5 e1g1 e5d4 f3d4 f8e7 a2a3 b4a3 b1a3 b8c6 d4c6 b7c6 a3c4 e8g8 c1b2 a7a5 a1a4 d6d5 e4d5 c6d5 c4e3 c8e6 d1f3 e7b4 e3f5 e6f5 f3f5 d8d7 b2f6 d7f5 d3f5 g7f6 g2g3 f8e8 f1a1 g8g7 a1d1 e8e5 f5g4 a8d8 g1g2 d5d4 g4f3');
INSERT INTO evaluations VALUES(53117,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(53117,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(53117,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(53117,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(53117,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(53117,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(53117,36,'cp -32','d2d4 g8f6 f1d3 e7e5 a2a3 b4a3 b1a3 f8e7 e1g1 e5d4 f3d4 e8g8 a3c4 b8c6 d4c6 b7c6 c1b2 a7a5 a1a4 d6d5 e4d5 c6d5 c4e3 c8e6 d1f3 e7b4 e3f5 f8e8 f5h6 g8f8 h6f5 e6f5 f3f5 d5d4 g2g3 d8d6 f1d1 a8c8');
INSERT INTO evaluations VALUES(53115,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(53115,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(53115,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(53115,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(53115,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(53115,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(53115,36,'cp -31','g8f6 f1d3 g7g6 h2h3 f8g7 e1g1 e8g8 a2a3 b4a3 c2c4 b8c6 b1c3 a3a2 a1a2 c6b4 a2b2 b4d3 d1d3 b7b6 c1g5 h7h6 g5e3 c8b7 f1b1 a8c8 b2a2 d8c7 f3d2 a7a6 a2a3 f6d7 c3a4 a6a5 a3b3 b7a6 a4b6 d7b6 b3b6 a6c4 d3a3 d6d5 a3a5 d5e4');
INSERT INTO evaluations VALUES(55541,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55541,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(55541,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(55541,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55541,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(55541,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(55541,36,'cp 0','g8f6 d2d3 c7c6 c4b3 d6d5 e4d5 c6d5 d3d4 e5d4 c3d4 f8b4 b1c3 h7h6 g1e2 b8c6 h2h3 e8g8 e1g1 c8f5 c1f4 f8e8 a1c1 a8c8 a2a3 b4a5 b3a4 a7a6 b2b4 a5b6 a4b3 c6e7 b3a4');
INSERT INTO evaluations VALUES(47802,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47802,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47802,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47802,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47802,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47802,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47802,36,'cp 0','d2d3 c7c6 c4b3 d6d5 g1f3 f8d6 e1g1 d5e4 d3e4 e8g8 d1e2 b8d7 f1d1 d8c7 a2a4 b7b6 b1d2 a7a5 d2c4 d6c5 h2h3 f8e8 f3g5 e8e7 g5f3');
INSERT INTO evaluations VALUES(56084,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(56084,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(56084,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(56084,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(56084,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(56084,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(56084,36,'cp 36','d3e4 b8d7 g1f3 b7b6 g2g3 c8b7 d1e2 g8f6 f1g2 f8d6 e1g1 e6e5 f3h4 e8g8 h4f5 f8e8 f5d6 c7d6 c2c4 d7c5 f1d1 a8c8 b2b3 h7h5 c1a3 h5h4 a3c5 b6c5 g3h4 a7a5 d2f1 c8a8 d1d2 a5a4 a1d1 a4b3 a2b3');
INSERT INTO evaluations VALUES(56073,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(56073,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(56073,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(56073,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(56073,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(56073,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(56073,36,'cp 42','b8d7 g1f3 b7b6 g2g3 c8b7 d1e2 f8d6 f1g2 g8f6 e1g1 e6e5 f3h4 e8g8 h4f5 f8e8 d2c4 d6f8 a2a4 a7a5 f1d1 d8b8 b2b3 b7c6 f5e3 b6b5 a4b5 c6b5 c1a3 f8a3 a1a3 d7b6 d1a1 a5a4 e2d3 a4b3 c2b3 a8a3 a1a3 b6c4 b3c4');
INSERT INTO evaluations VALUES(57183,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(57183,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(57183,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(57183,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(57183,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(57183,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(57183,36,'cp -63','d7d5 d2d4 e5d4 e3d4 f8e7 b1c3 g8f6 c1g5 h7h6 g5f6 e7f6 d1d2 e8g8 e1c1 b8c6 g1f3 c8e6 h1g1 e6g4 d2f4 g4f3 g2f3 g8h8 f4g4 c6e7 f1d3 c7c6 c3e2 b7b5 c2c3 d8d6 c1b1 a8d8 f3f4 b5b4 c3b4');
INSERT INTO evaluations VALUES(56419,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(56419,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(56419,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(56419,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(56419,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(56419,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(56419,36,'cp -65','d2d4 e5d4 e3d4 g8f6 b1c3 f8e7 c1g5 h7h6 g5f6 e7f6 d1d2 e8g8 e1c1 f8e8 g1f3 c8g4 f1e2 g4f3 e2f3 c7c6 g2g3 b8d7 h1e1 d8c7 e1e8 a8e8 c3e2 d7b6 b2b3 c7e7 c1b1 b6c8 c2c3 c8d6 e2f4 a7a5 f4d3 b7b6 f3g2 d6e4 d2c2 e7d6 d1e1 g7g6 g2e4 d5e4');
INSERT INTO evaluations VALUES(255,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(255,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(255,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(255,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(255,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(255,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(255,36,'cp 25','g8e7 c3d5 e7d5 e4d5 c6d4 f3d4 e5d4 f1e1 e8d8 d2d3 a7a6 b5a4 b7b5 a4b3 h7h6 c1d2 d7d6 a2a4 a8b8 d1h5 h8e8 e1e8 d8e8 a4a5 e8f8 b3a2 c5a7 b2b4 c8d7 a1e1 b8e8 e1e8 f8e8 f2f4 f6f5 h5f5 d7f5 g1f2 f5d7 h2h3 e8e7 g2g4 h6h5 f2g3 h5g4 h3g4 f7f5 g4g5 d7e8');
INSERT INTO evaluations VALUES(564,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(564,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(564,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(564,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(564,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(564,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(564,36,'cp 35','c3d5 e7d5 e4d5 c6d4 f3d4 e5d4 f1e1 e8d8 a2a4 a7a6 b5c4 d4d3 e1e3 d3c2 d1c2 c5e3 d2e3 d7d6 c1d2 h8e8 c4d3 c8d7 d2c3 f6g5 e3e4 e8e7 a1f1 g5f4 c3d2 f4e5 f1e1 a6a5 d2e3 e5f6 c2b3 b7b6 h2h3 a8c8 e1c1 d8e8 d3a6 c8a8 a6b7 a8b8 c1c7 e7e4');
INSERT INTO evaluations VALUES(1087,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(1087,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(1087,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(1087,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(1087,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(1087,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(1087,36,'cp 70','c3d4 d7d6 f1b5 d6e5 d4e5 c8g4 b5c6 b7c6 e1g1 f6d5 b1d2 e7d7 d2e4 f8e7 h2h3 g4f3 d1f3 e8g8 f1e1 a8e8 c1d2 d7e6 f3g3 f7f5 e4c3 e7c5 a1c1 c5b6 c3a4 f5f4 g3g4 g8h8 c1c4 f8f5 a4c5 b6c5 c4c5 f5e5 e1e5 e6e5');
INSERT INTO evaluations VALUES(1086,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(1086,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(1086,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(1086,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(1086,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(1086,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(1086,36,'cp 54','d7d6 f1b5 d6e5 d4e5 c8g4 e1g1 f6d7 b5c6 b7c6 b1d2 e7e6 h2h3 g4h5 f3d4 h5d1 d4e6 f7e6 f1d1 d7e5 d2e4 a8b8 b2b3 b8b5 c1b2 e5g6 b2d4 b5a5 a1c1 a5d5 g2g3 e6e5 d4a7 f8a3 c1b1 g6e7 d1d5 c6d5 a7c5 a3c5 e4c5 h8f8 b3b4 e5e4 b4b5 e7f5');
INSERT INTO evaluations VALUES(910,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(910,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(910,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(910,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(910,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(910,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(910,36,'cp 64','f1b5 d6e5 d4e5 c8g4 e1g1 f6d7 b5c6 b7c6 b1d2 d7e5 h2h3 e5f3 d2f3 g4e6 f1e1 e7d7 d1c2 f8e7 f3e5 d7d5 c1f4 d5c5 c2e4 e8g8 e5c6 e7f6 a1c1 c5h5 b2b4 e6a2 f4c7 a2d5 c6e7 f6e7 e4e7 h5g6 c7g3 a7a6 e1e5 d5e6 g1h2 a8b8 c1d1 h7h6 e5a5 b8e8 e7c7 e8b8 c7b8 f8b8 g3b8 e6c4 d1d8 g8h7');
INSERT INTO evaluations VALUES(909,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(909,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(909,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(909,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(909,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(909,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(909,36,'cp 55','d6e5 d4e5 c8g4 e1g1 f6d7 b5c6 b7c6 b1d2 d7e5 h2h3 e5f3 d2f3 g4e6 f1e1 e7d7 d1c2 f8e7 f3e5 d7d5 c1f4 e7d6 a1d1 d5c5 c2c5 d6c5 e5c6 c5b6 b2b4 e8g8 a2a4 a7a5 f4d2 e6d7 c6e7 g8h8 b4a5 b6c5 d1c1 c5d6 g2g4 f8b8 e7f5 d6b4 d2b4 b8b4 f5e7 d7e6 c1c7 b4a4 e7c6 h8g8 e1e5 g8h8 e5b5 h7h6');
INSERT INTO evaluations VALUES(5346,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(5346,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(5346,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(5346,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(5346,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(5346,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(5346,36,'cp -28','d7d6 g2g3 e7e5 a2a3 g8e7 f1g2 e8g8 e1g1 a8b8 a1b1 c8g4 h2h3 g4e6 f3g5 e6d7 g5f3 c6d4 f3d4 c5d4 c3e2 d7e6 c2c3 d4c3 e2c3 d6d5 e4d5 e7d5 d1a4 a7a6 c3d5 e6d5 b1d1 f7f5 g2d5 d8d5 a4c4 b8d8 d2e3 d5c4');
INSERT INTO evaluations VALUES(5499,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(5499,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(5499,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(5499,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(5499,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(5499,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(5499,36,'cp -43','a2a3 g8f6 f1e2 e8g8 e1g1 d6d5 c3d5 f6d5 e4d5 d8d5 d2c3 c8f5 c3g7 g8g7 f3d2 d5d7 f1e1 f8e8 h2h3 a8d8 b2b4 d7c7 b4c5 c6d4 a1b1 c7c5 d2b3 d4b3 b1b3 b7b6 e2g4 f5g4 h3g4 e7e5 g2g3 d8d4 c2c3 d4d8 d1e2 d8d7 a3a4 e8d8');
INSERT INTO evaluations VALUES(6951,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(6951,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(6951,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(6951,33,'cp 249',NULL);
INSERT INTO evaluations VALUES(6951,34,'cp 223',NULL);
INSERT INTO evaluations VALUES(6951,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(6951,36,'cp 217','d7d6 c5e3 c3b2 c1b2 g8f6 b1c3 f6g4 e3d2 h8f8 c3d5 g4e5 f3e5 c6e5 e1g1 c8e6 f2f4 e5c4 d2e2 c4b2 f4f5 e6d5 e4d5 e8f7 f5g6 f7g8 e2e6 g8h8 f1f7 f8f7 g6f7 b2d3 a1f1 d3e5 f7f8q d8f8 f1f8 a8f8 h2h4 a7a5 h4h5 f8f7 e6c8 h8g7 c8b7 h7h6 b7b5');
INSERT INTO evaluations VALUES(9847,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(9847,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(9847,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(9847,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(9847,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(9847,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(9847,36,'cp 120','d4e5 f6g4 c4d5 g4e5 f3e5 c6e5 f2f4 c7c6 d5b3 d8h4 g2g3 h4g4 e1g1 g4d1 f1d1 e5g4 e4e5 h6h5 c3e4 d7d5 e5d6 c8f5 c1d2 b4d2 e4d2 e8c8 d2f3 h8h6 d6d7 d8d7 d1d7 c8d7 b3f7 f5c2 a1e1 g4f6 f3d4 c2h7 f7e6 d7d6 d4f3 h7d3 f3g5 d6c7 e6h3 h5h4 e1e7 c7b6 e7g7 h4g3');
INSERT INTO evaluations VALUES(9828,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(9828,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(9828,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(9828,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(9828,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(9828,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(9828,36,'cp 150','f6g4 c4d5 c6e5 f3e5 g4e5 f2f4 c7c6 d5b3 d8h4 g2g3 h4g4 e1g1 g4d1 f1d1 e5g4 e4e5 h6h5 a2a3 b4a5 g1g2 d7d5 e5d6 a5c3 b2c3 c8f5 d6d7 e8d8 h2h3 f5e4 g2g1 g4f6 c1e3 e4d5 f4f5 d5b3 c2b3 f6d7 d1d3 d8c8 a1d1 d7e5 d3d6 h8e8 e3f4 c8c7 f5f6 g7g5 f4g5 h5h4');
INSERT INTO evaluations VALUES(11305,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(11305,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(11305,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(11305,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(11305,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(11305,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(11305,36,'cp 63','b5c6 b7c6 c2c4 e7e6 d1a4 a6a5 e1g1 c8b7 c4d5 e6d5 c1d2 f6d7 b1c3 f8e7 a4b3 d7b6 f1e1 e8g8 c3a4 f8e8 a1c1 e7d6 e3e4 d5e4 f3g5 d8d7 g5e4 b6a4 b3a4 d7f5 e4d6 e8e1 c1e1 c7d6 a4a5 f5a5 d2a5 h7h6 b2b4 b7a6 a2a4 a6c4 e1e7');
INSERT INTO evaluations VALUES(11302,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(11302,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(11302,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11302,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11302,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(11302,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(11302,36,'cp 48','b7c6 c2c4 e7e6 d1a4 a6a5 e1g1 c8b7 c4d5 e6d5 c1d2 f6d7 a4b3 d7b6 b3c2 f8d6 e3e4 d5e4 c2e4 d8e7 e4c2 e8g8 f3g5 f7f5 f1e1 e7d7 g5e6 f8e8 e6c5 e8e1 d2e1 d6c5 d4c5 b6d5 c2b3 b7c8 b1c3 a5a4 c3a4 d7f7 a4c3 d5f4 b3f7 g8f7 b2b3 c8e6 e1d2 f4d3 f2f3 d3c5 g1h1');
INSERT INTO evaluations VALUES(11607,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(11607,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(11607,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(11607,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(11607,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(11607,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(11607,36,'cp 54','c2c4 e7e6 d1a4 d5c4 a4c6 c8d7 c6c4 c7c5 b1c3 c5d4 e3d4 f8d6 e1g1 e8g8 c1g5 h7h6 g5h4 d8c8 c4e2 f6h5 f1c1 c8b7 c3e4 d6f4 e4c5 d7b5 e2d1 b7d5 c1c3 f4d6 h4g3 h5f4 g3f4 d6f4 a2a4 b5c6 c5d3 f4d6 a1c1 c6b7');
INSERT INTO evaluations VALUES(17228,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(17228,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(17228,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(17228,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(17228,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(17228,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(17228,36,'cp 32','g1f3 d7d5 e2e3 f6e4 g5h4 h7h6 c2c4 g7g5 h4g3 e7e6 f3d2 e4g3 h2g3 f8g7 f1e2 e8g8 e1g1 c6e7 b1c3 c7c6 d1c2 c8d7 a1c1 a8c8 b2b4 e6e5 d4e5 g7e5 f1d1 e5g7 d2b3 g8h8 c4d5 c6d5 b3c5');
INSERT INTO evaluations VALUES(17227,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(17227,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(17227,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(17227,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(17227,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(17227,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(17227,36,'cp 16','f6e4 g5f4 d7d5 e2e3 e7e6 h2h3 f8d6 f1d3 d6f4 e3f4 c6b4 e1g1 b4d3 c2d3 e4d6 b1c3 e8g8 c3e2 b7b6 a1c1 c8d7 g2g4 a8c8 b2b4 d6b7 g1g2 a7a5 a2a3 a5b4 a3b4 b7d6 d1c2 d6b5 f1e1 f7f6 c2d2 c8a8');
INSERT INTO evaluations VALUES(23763,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(23763,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(23763,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(23763,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(23763,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(23763,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(23763,36,'cp 129','b1c3 d7g4 h2h3 g4f3 g2f3 a7a6 b5a4 g8h8 f3f4 d6d5 e4e5 f6e4 c3e4 d5e4 e1e4 d8d5 a4c6 b7c6 d1f3 f7f6 g1g2 a6a5 f4f5 c6c5 c1e3 c5d4 e3d4 f6e5 d4e5 e7f6 e5f6 g7f6 a1d1 f8g8 g2h2 d5c6 d1d3 a8d8 d3d8 g8d8 f3g4 c6d6 g4g3 d6g3 h2g3');
INSERT INTO evaluations VALUES(23761,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(23761,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(23761,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(23761,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(23761,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(23761,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(23761,36,'cp 102','d7g4 h2h3 g4f3 g2f3 a7a6 b5a4 g8h8 f3f4 d6d5 e4e5 f6e4 c3e4 d5e4 e1e4 d8d5 a4c6 b7c6 d1f3 f7f6 g1g2 a6a5 e5e6 g7g6 e4e2 f8d8 f3d5 d8d5 c1e3 d5d6 e2c2 d6e6 d4d5 e6e4 d5c6 h8g8 a1c1 e7d6 c2c4 a8e8 b2b3 e4c4 c1c4');
INSERT INTO evaluations VALUES(24239,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(24239,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(24239,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(24239,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(24239,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(24239,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(24239,36,'cp 101','d2d3 h7h6 a2a4 b5b4 a4a5 g7g5 c2c3 f8g7 c1e3 e8g8 b1d2 e7g6 h2h3 d7d6 f1e1 a8b8 d2f1 b7c8 d3d4 c6e7 b3c4 g6f4 f1g3 e7g6 c4f1 b4c3 b2c3 g5g4 h3g4 c8g4 e3f4 e5f4 g3f5 d6d5 f1d3 d5e4 d3e4 f8e8');
INSERT INTO evaluations VALUES(24326,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24326,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24326,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24326,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24326,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24326,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24326,36,'cp 0','b7b5 a4b3 c6a5 h2h3 f8e7 f3h2 e8g8 f2f4 c7c5 f4e5 d6e5 c1g5 a5b3 a2b3 d7e6 g5f6 e7f6 g1h1 f6g5 h2f3 g5f6');
INSERT INTO evaluations VALUES(24349,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24349,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24349,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24349,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24349,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24349,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24349,36,'cp 0','a4b3 c6a5 d3d4 e5d4 c3d5 f8e7 d5e7 d8e7 f1e1 d7g4 d1d4 g4f3 g2f3 e8g8 c1g5 a5b3 a2b3 e7e6 e4e5 f6d7 e5d6 e6g6 d4g4 c7d6 g5f4 g6f6 f4g5');
INSERT INTO evaluations VALUES(24345,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24345,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24345,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24345,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24345,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24345,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24345,36,'cp 0','c6a5 d3d4 e5d4 c3d5 f8e7 d5e7 d8e7 f1e1 d7g4 d1d4 g4f3 g2f3 e8g8 c1g5 a5b3 a2b3 e7e6 e4e5 f6d7 e5d6 e6g6 d4g4 c7d6 g5f4 g6f6 f4g5');
INSERT INTO evaluations VALUES(25801,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(25801,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(25801,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(25801,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(25801,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(25801,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(25801,36,'cp 56','f1e1 g8e7 d2d4 e7g6 c1e3 f8e7 d4e5 d6e5 d1d8 e7d8 f3d2 d8e7 b1c3 g6f4 d2f3 f4g6 c3a4 f7f6 f3d2 g6f8 d2b3 a6b5 b3c5 e8f7 b2b3 h7h5 f2f3 b5a4 c5a4 f8e6 e1d1 h8d8 g1f1 g7g5 a4b2');
INSERT INTO evaluations VALUES(25802,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(25802,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(25802,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(25802,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(25802,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(25802,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(25802,36,'cp 36','g8e7 d2d4 e7g6 c1e3 f8e7 d4e5 d6e5 d1d8 a8d8 a2a3 d8a8 b1c3 f7f6 f3d2 a6c8 a1d1 e8f7 d2b3 a7a5 c3a4 g6f4 h2h4 h7h5 e3c5 g7g5 h4g5 f6g5 b3c1 f4e6 c5e7 f7e7 c1d3 e7f6 g1h2 g5g4 d1d2 c8a6 a4c5 a6d3 c5d3');
INSERT INTO evaluations VALUES(26331,30,'cp 294',NULL);
INSERT INTO evaluations VALUES(26331,31,'cp 281',NULL);
INSERT INTO evaluations VALUES(26331,32,'cp 290',NULL);
INSERT INTO evaluations VALUES(26331,33,'cp 288',NULL);
INSERT INTO evaluations VALUES(26331,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(26331,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(26331,36,'cp 283','c4f7 e8f7 f3g5 f7e8 d1g4 d8c8 g4d1 g8f6 d2d3 h7h6 g5f3 e7e5 b1c3 f8e7 f3h4 c8g4 d1g4 f6g4 h4g6 h8g8 g6e7 e8e7 c3d5 e7f7 f2f4 f7g6 c1d2 g8f8 h2h3 g4f6 d5e3 b7b5 f4e5 d6e5 f1f2 g6h7 a1f1 b5b4 a2a3 b4a3 b2a3');
INSERT INTO evaluations VALUES(26659,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(26659,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(26659,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26659,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(26659,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26659,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(26659,36,'cp 12','f1e2 e7e6 e1g1 g8e7 a2a3 h7h6 b1d2 e7g6 f3d4 g4e2 d4e2 c6e5 d3d4 e5c6 d2f3 f8d6 b2b4 g6h4 e2g3 h4f3 d1f3 c6d4 f3b7 e8g8 c1b2 c7c5 b2d4 c5d4 a1d1 d8b6 b7b6 a7b6 d1d4 d6g3 f2g3 a8a3 d4d6 f8c8 d6b6 c8c2 b6b7 c2e2 b7f7 a3b3');
INSERT INTO evaluations VALUES(27000,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(27000,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(27000,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(27000,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(27000,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(27000,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(27000,36,'cp 109','d4d5 c6b8 f3d2 g4e2 d1e2 b8d7 d2c4 f8e7 c1e3 e7g5 h2h4 g5e3 c4e3 h7h5 b1c3 a7a6 a2a4 d8f6 e2c4 f6d8 a4a5 d8b8 c4e2 g7g6 g2g3 g8f6 f2f3 b7b6 c3d1 e8g8 e2d2 b8a7 b2b3 b6b5 d1f2 a7d4 c2c3 d4d2 e1d2');
INSERT INTO evaluations VALUES(26930,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(26930,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(26930,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(26930,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(26930,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(26930,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(26930,36,'cp 95','c6b8 a2a4 f8e7 e1g1 g8f6 f3d2 g4e2 d1e2 e8g8 b1c3 f6e8 a4a5 g7g6 b2b4 e7g5 a1a3 e8g7 e2b5 d8c8 b5c4 c8d8 c3e2 a7a6 a3c3 g7e8 e2g3 b8d7 d2b3 g5c1 b3c1 f7f5 e4f5 g6f5 f2f4 d7f6');
INSERT INTO evaluations VALUES(27262,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(27262,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(27262,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(27262,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(27262,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(27262,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(27262,36,'cp 106','b5d7 d8d7 b1c3 g8e7 c1e3 e7c6 d4d3 e8c8 e1g1 d8e8 c3d5 c6e5 d3d4 c7c6 d5f4 b7b6 a2a4 f8e7 a1d1 e7d8 b2b3 h7h5 f4d3 d8f6 d3e5 e8e5 e3f4 e5e6 d4d3 h8d8 a4a5 b6a5 d3a6 d7b7 a6a5 e6e4 f4d6 e4e6 a5c5 b7b6 c5f5 g7g6');
INSERT INTO evaluations VALUES(27144,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(27144,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(27144,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(27144,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(27144,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(27144,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(27144,36,'cp 96','d8d7 b1c3 g8e7 c1e3 e7c6 d4c4 f8e7 e1c1 e8g8 c1b1 f8e8 f2f3 e7f8 c3d5 a8c8 c4b5 b7b6 b5a4 a7a5 a4b5 c6e5 b5d7 e5d7 a2a4 g7g6 c2c3 h7h5 h2h3 f8g7 b1c2 d7f6 g2g4 f6d5 d1d5');
INSERT INTO evaluations VALUES(27666,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(27666,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(27666,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(27666,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(27666,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(27666,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(27666,36,'cp 99','b1c3 g8e7 c1e3 e7c6 d4c4 f8e7 e1c1 e8g8 f2f4 f8e8 h1e1 e7f8 e3f2 c6e7 c4b3 c7c6 a2a4 e7c8 f4f5 a7a6 g2g4 b7b5 g4g5 d6d5 e4d5 e8e1 f2e1 c6c5 a4b5 a6b5 b3b5 d7f5 c1b1 c8d6 b5c5 f5g5 c5b6 g5f5 e1g3 f5d7');
INSERT INTO evaluations VALUES(32779,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(32779,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(32779,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(32779,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(32779,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(32779,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(32779,36,'cp 43','g8f6 f3d2 g4e2 d1e2 g7g6 f2f3 f8h6 a2a4 b8d7 b1a3 d7c5 a4a5 f6d7 a3c4 c5e6 d2b3 h6c1 h1c1 e8c8 e2e3 g6g5 h2h4 g5h4 c1h1 d8g8 a1g1 h7h5 h1h4 e6f8 b3c1 f8g6 h4h1 g6e5 c4e5 d7e5 b2b3 h5h4 c1d3 e5d3 c2d3 g8d8');
INSERT INTO evaluations VALUES(32570,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(32570,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(32570,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(32570,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(32570,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(32570,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(32570,36,'cp 37','f3d2 g4e2 d1e2 g7g6 f2f3 f8h6 a2a4 b8a6 d2c4 h6c1 h1c1 e8c8 b1d2 a6c7 a4a5 f6d7 d2b3 h7h5 h2h4 h8g8 e2e3 c7e6 f3f4 f7f5 e4e5 d7c5 b3c5 e6c5 c1d1 d8d1 a1d1 g8d8 d1d8 c8d8 g2g3 c5e4 c4d2 e4d2 e3d2 b7b6 a5b6 a7b6 d2c3 d8d7');
INSERT INTO evaluations VALUES(38504,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(38504,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38504,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(38504,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38504,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(38504,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(38504,36,'cp 48','g8f6 e4f6 e7f6 f1e2 e8g8 e1g1 f8e8 f1e1 c7c6 c1f4 c8f5 d1d2 g7f8 a1d1 f5e4 a2a3 b7b5 d4d5 d8d5 c2c4 b5c4 d2c1 d5f5 e2c4 f5c8 c1d2 e4d5 c4d5 c6d5 d2d5 b8c6 g2g3 e8d8 d5b5 d8d1 e1d1 c8e6 f4e3 c6e7');
INSERT INTO evaluations VALUES(40388,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(40388,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(40388,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(40388,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(40388,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(40388,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(40388,36,'cp 76','f1e2 b8c6 c1e3 e8g8 e1g1 c6d4 e3d4 c8d7 d1c2 d8a5 f1d1 f8c8 c2d3 a7a6 a2a3 d7e6 c3d5 e6d5 e4d5 f6d7 b2b4 a5c7 d4g7 g8g7 a1c1 a6a5 d3d4 g7g8 d4e3 a5b4 a3b4 c7b6 e3b6 d7b6 c4c5 d6c5 b4c5 b6a4 c5c6 b7c6 d5c6 c8c7');
INSERT INTO evaluations VALUES(40387,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(40387,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(40387,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(40387,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(40387,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(40387,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(40387,36,'cp 57','b8c6 c1e3 a7a6 d1d2 e8g8 e1g1 c8d7 f1d1 f6g4 e2g4 d7g4 d4c6 b7c6 f2f3 g4e6 e3d4 g7d4 d2d4 d8b8 c3a4 b8a7 d4a7 a8a7 c4c5 d6d5 e4d5 e6d5 a4b6 d5e6 d1d4 g8g7 b2b3 a6a5 a1c1 h7h5 h2h3 f8b8 d4a4');
INSERT INTO evaluations VALUES(8330,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(8330,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(8330,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(8330,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(8330,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(8330,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(8330,36,'cp 75','c1e3 e8g8 e1g1 a7a6 d1d2 c8d7 a2a3 b7b5 c4b5 c6d4 e3d4 a6b5 f1d1 d8b8 d4f6 g7f6 c3d5 a8a4 d5f6 e7f6 d2f4 b5b4 f4f6 b4a3 b2a3 d7e6 e2f3 b8b3 h2h4 a4a3 h4h5 b3c3 f6c3 a3c3 d1d6 g6h5 f3h5 f8c8 a1a7 c3c6 d6c6 c8c6');
INSERT INTO evaluations VALUES(44821,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(44821,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(44821,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(44821,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(44821,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(44821,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(44821,36,'cp -29','d5c4 e2e3 b7b5 a2a4 b5b4 c3b1 c8a6 h4f6 e7f6 g1f3 f8e7 f1e2 e8g8 e1g1 d8d5 f1e1 f8d8 e3e4 d5d7 d1c2 b4b3 c2c1 e7b4 b1d2 d7b7 e2c4 b4d2 c4a6 b8a6 c1d2 a6b4 a1c1 b4c2 e1d1 b7a6 h2h4 a6a4 d2c3 a4b4 c3b4');
INSERT INTO evaluations VALUES(49476,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(49476,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(49476,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(49476,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(49476,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(49476,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(49476,36,'cp -70','f6d5 e4d5 f8c5 d2d4 e5d4 g1f3 e8g8 f1e2 d8f6 e1g1 h7h6 c1d2 d7d6 c2c3 c8g4 f3d4 g4e2 d4e2 b8d7 e2g3 f8e8 b2b3 a7a5 c3c4 c5d4 a1b1 a5a4 d2h6 a4b3 b1b3 a8a2 b3f3 f6h6 d1d4 a2d2 d4a7 d7e5 f3c3 d2b2');
INSERT INTO evaluations VALUES(49814,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(49814,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(49814,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(49814,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(49814,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(49814,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(49814,36,'cp -66','e4d5 f8c5 g1f3 e8g8 d2d4 e5d4 f1e2 d8f6 e1g1 h7h6 c1d2 d7d6 c2c3 c8g4 c3d4 g4f3 d4c5 f3d5 c5d6 f6d6 d2c3 f8d8 f1e1 b8d7 e2d3 d5c6 d3c2 d7f6 d1d6 d8d6 f2f3 a7a5 g1f2 f6d5 c3e5 d6d7 a1d1 a8d8 c2f5 d7e7 e5c3 e7e1 c3e1 d8e8');
INSERT INTO evaluations VALUES(49805,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(49805,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(49805,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(49805,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(49805,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(49805,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(49805,36,'cp -57','f8c5 d2d4 e5d4 f1e2 d8f6 g1f3 e8g8 e1g1 h7h6 c1d2 d7d6 c2c3 f6f5 f3d4 f5d5 d2e3 c5d4 c3d4 b8d7 d1c2 c7c6 e2f3 d5a5 e3f4 d6d5 a1e1 d7f6 f4e5 f6e4 f3e4 d5e4 c2e4 c8e6 e4h4 f7f6 e5f4 f8e8 h4g3 a5h5 a2a3 e6f7 e1e8 a8e8');
INSERT INTO evaluations VALUES(51515,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(51515,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51515,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(51515,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(51515,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(51515,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(51515,36,'cp 40','d7d5 e2e4 d5e4 c3e4 g8f6 c2c3 c8b7 e4f6 e7f6 d2d4 d8d7 f1d3 f8d6 d1e2 d7e6 e1g1 e8g8 e2c2 g7g6 f1e1 e6d7 f3d2 f8e8 d2e4 b7e4 e1e4 b8c6 g2g3 f6f5 e4e8 a8e8 c1h6 c6e7 a1e1 e7d5 e1e8 d7e8 c2d1 d6f8 h6d2 e8d7 d3c4 c7c6 c4b3 d5f6 d1e2 f8d6 g1g2 h7h5 c3c4');
INSERT INTO evaluations VALUES(53906,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(53906,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(53906,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(53906,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(53906,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(53906,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(53906,36,'cp 79','c2c3 g8f6 c3d4 f8e7 b1c3 e8g8 c4b3 c8g4 d4c5 b8c6 e1g1 d6c5 h2h3 d8d1 f1d1 g4e6 e4e5 f6e8 c3d5 e6d5 b3d5 e8c7 d5c6 b7c6 c1e3 f8d8 d1d8 a8d8 a1c1 c7e6 c1c4 d8d1 g1h2 f7f6 c4a4 d1d7 a4a6 f6e5 a6c6 e5e4 f3d2');
INSERT INTO evaluations VALUES(53905,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(53905,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(53905,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(53905,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(53905,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(53905,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(53905,36,'cp 75','g8f6 c3d4 f8e7 b1c3 e8g8 d4d5 c8g4 h2h3 g4f3 d1f3 b8d7 f3e2 f6e8 e1g1 a7a6 c4d3 e7f6 a2a4 f6d4 e2d1 g7g6 a1a3 a8b8 g1h1 e8c7 a3b3 f8e8 f2f4 b7b5 a4b5 d4c3 b2c3 a6b5 b3b2 d8f6 c3c4 b5c4 b2b8 e8b8 d3c4');
INSERT INTO evaluations VALUES(55730,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(55730,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(55730,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(55730,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(55730,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(55730,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(55730,36,'cp 27','g8f6 c2c4 g7g6 f1e2 f8g7 b2b4 e8g8 e1g1 e6e5 d4e5 f6g4 c1b2 g4e5 f3e5 g7e5 b2e5 d6e5 b1c3 c8e6 d1c1 d8e7 c3d5 e6d5 c4d5 f8d8 f1d1 c7c6 c1c4 c6d5 d1d5 b8c6 a2a3 d8d5 c4d5 a8d8');
INSERT INTO evaluations VALUES(48053,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(48053,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(48053,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(48053,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(48053,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(48053,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(48053,36,'cp 36','c2c4 b7b6 f1d3 c8b7 b1c3 d6d5 e1g1 f8d6 f1e1 e8g8 e3e4 f6e4 c3e4 d5e4 d3e4 b7e4 e1e4 b8d7 c1g5 f7f6 g5h4 f8e8 d1c2 a7a5 a1d1 d7f8 c2e2 d8e7 h4g3 a8d8 e4e3 e7f7 b2b3 h7h6 g3d6 d8d6 g2g3 f7d7 d1d2');
INSERT INTO evaluations VALUES(48045,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(48045,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(48045,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(48045,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(48045,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(48045,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(48045,36,'cp 32','b7b6 f1d3 d6d5 e1g1 c8b7 b1c3 f8d6 f1e1 e8g8 e3e4 f6e4 c3e4 d5e4 d3e4 b7e4 e1e4 b8d7 c1g5 d6e7 g5e7 d8e7 d1e2 e7d6 a1d1 c7c6 e4e3 a8d8 e3d3 d6c7 b2b3 h7h6 g2g3 d7f6 g1g2 f8e8 h2h4 c6c5 d4c5 b6c5 d3d8 e8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(57521,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57521,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57521,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(57521,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57521,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(57521,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(57521,36,'cp 29','d8e7 f3d4 b8c6 e2h5 e8d8 d4c6 d7c6 f1e2 g8f6 h5h3 e7e5 g2g4 h7h5 b1c3 f8d6 h3h4 a7a5 a2a4 c8d7 h2h3 d8c8 g4g5 f6d5 h4e4 d7f5 e4e5 d6e5 c3d5 c6d5 d2d3 c8d7 c2c3 a8e8 c1e3 e5d6 h3h4 g7g6 e1d2 c7c6 a1d1 d7c7 d1f1 h8f8 d3d4 f5e4 f2f3 e4f5 e2d3');
INSERT INTO evaluations VALUES(35240,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(35240,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(35240,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35240,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35240,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(35240,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(35240,36,'cp 23','f3d4 b8c6 e2h5 e8d8 d4c6 d7c6 f1e2 g8f6 h5h3 e7e5 g2g4 h7h5 b1c3 f8d6 h3h4 a7a5 a2a4 c8d7 g4g5 f6d5 h4e4 d7f5 e4e5 d6e5 c3d5 c6d5 d2d3 d8d7 h2h4 a8e8 e1f1 g7g6 e2f3 h8f8 f1g2 e5d4 c1d2 f5g4 f3g4 h5g4 d2e3 d4e3 f2e3 e8e3');
INSERT INTO evaluations VALUES(35239,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(35239,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(35239,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35239,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(35239,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(35239,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(35239,36,'cp 12','b8c6 d4c6 d7c6 e2h5 e8d8 f1e2 g8f6 h5h3 e7e5 g2g4 h7h5 b1c3 f8d6 h3h4 a7a5 a2a4 c8d7 g4g5 f6d5 h4e4 d7f5 e4e5 d6e5 c3d5 c6d5 d2d3 d8d7 c2c3 g7g6 c1e3 h8f8 h2h4 a8e8 e1d2 c7c6 a1e1 e5d6 h1g1 b7b5 a4b5 c6b5 e1a1 a5a4 e2d1 f5h3 g1h1 h3g2 h1e1 d7c6 a1b1');
INSERT INTO evaluations VALUES(3598,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(3598,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(3598,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(3598,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(3598,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(3598,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(3598,36,'cp -33','g7g5 h4g3 c6a5 c4b3 b7b5 b1c3 c8d7 f1e1 f8e8 c3b1 a5b3 a2b3 a7a5 b1d2 c5b6 c2c3 g5g4 h3g4 f6g4 d3d4 d8f6 d2f1 e5d4 c3d4 f6g6 f3h4 g6g7 h4f3 f7f5 e4e5 d7c6 f1e3 g4e3 e1e3 g7g4 d1d3 f5f4 g3f4 g4f4 d3g6 g8h8 a1e1');
INSERT INTO evaluations VALUES(7649,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7649,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7649,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7649,36,'cp 0','e4e5 d8a5 e5f6 b4c3 b2c3 a5c3 d1d2 c3c4 f6g7 h8g8 c1b2 c4e4 d2e3 e4c2 b2f6 c8a6 f2f3 a8b8 e3d2 c2g6 d2d6 g6f6 d6b8 e8e7 b8g8 f6e5 e1f2 e5c5 f2g3 c5g5 g3f2');
INSERT INTO evaluations VALUES(8244,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8244,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(8244,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8244,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8244,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8244,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8244,36,'cp 0','e8g8 c1e3 b7b6 d1d2 c8b7 f1e1 e7e6 e3g5 d8d7 a2a4 a7a6 c4b3 d7c7 c3e2 c6a5 b3a2 a5c6');
INSERT INTO evaluations VALUES(8735,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(8735,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(8735,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(8735,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(8735,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(8735,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(8735,36,'cp 59','b2b3 e8g8 c1b2 f8e8 h2h3 f6e4 c2c4 d5c4 b3c4 e7e5 d4d5 c6b4 f3d2 e4c5 b1c3 b7b6 a1b1 c8d7 d5d6 c7d6 b2a3 a7a5 g2a8 d8a8 g1h2 e5e4 a3b4 a5b4 c3d5 g7c3 b1c1 e4e3 f2e3 a8a2 d5c3 b4c3 c1c3 d7c6 d1c1 a2a7');
INSERT INTO evaluations VALUES(10895,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(10895,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(10895,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(10895,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(10895,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(10895,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(10895,36,'cp -56','d2d3 c6a5 c4b5 a7a6 b5a4 b7b5 a4c2 d7d5 f3e2 c7c5 g1f3 d5e4 d3e4 d8c7 e1g1 e8g8 f1d1 c8e6 c1g5 h7h6 g5f6 e7f6 h2h3 a8b8 b1d2 f8d8 d2f1 d8d1 c2d1 e6c4 e2e3 f6e7 d1b3 a5b3 a2b3 c4b3 a1a6 b3e6');
INSERT INTO evaluations VALUES(10898,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(10898,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(10898,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(10898,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(10898,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(10898,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(10898,36,'cp -55','c6a5 c4b5 a7a6 b5a4 b7b5 a4c2 d7d5 f3e2 c7c5 g1f3 d5e4 d3e4 d8c7 c1g5 e8g8 e1g1 c8e6 b1d2 h7h6 g5f6 e7f6 f1d1 a5c4 d2c4 e6c4 c2d3 c4e6 c3c4 f8d8 h2h3 g7g6 d1c1 b5b4 d3c2 c7c6 c1d1 a6a5 d1d8 a8d8 b2b3 g8g7 a1d1 d8d1 c2d1 c6d6 f3e1');
INSERT INTO evaluations VALUES(18069,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18069,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(18069,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(18069,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(18069,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(18069,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(18069,36,'cp 37','a2b3 d8g5 e5f3 g5g2 h1g1 g2h3 c2c4 c8b7 g1g3 h3h5 b1c3 f7f5 d2d3 f5e4 d3e4 g8f6 c4b5 f8b4 g3g7 b7e4 f3g5 h5d1 e1d1 e4b7 b5a6 a8a6 a1a6 b7a6 g5e4 f6e4 c3e4 b4e7 c1g5 e7g5 g7g5 h8f8 d1d2 f8f4 d2e3 f4h4 e4f6 e8d8 g5h5 h4b4 h5h7 b4b3 e3f4 b3b2 h7d7 d8c8');
INSERT INTO evaluations VALUES(18068,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(18068,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18068,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(18068,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18068,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(18068,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18068,36,'cp 18','d8g5 e5f3 g5g2 h1g1 g2h3 c2c4 c8b7 b1c3 g8f6 g1g3 h3e6 d2d3 g7g6 f3d4 e6e5 c1e3 f6h5 g3g2 h5f4 g2g4 f4e6 d4f3 e5d6 d3d4 f7f5 g4h4 f5f4 e3d2 e6d4 d2f4 d4f3 d1f3 d6f6 f3g3 f8e7 h4g4 f6e6 f4e5 e8g8 c3d5 b7d5 c4d5');
INSERT INTO evaluations VALUES(18567,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(18567,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(18567,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(18567,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(18567,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(18567,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(18567,36,'cp 52','b1d2 g8f6 a2a4 c6c5 c2c3 a7a6 h2h3 b7b6 h3h4 c8b7 f1d3 e7e6 d1f3 f8e7 g2g4 h7h5 g4g5 f6g4 g5g6 e6e5 f4g3 e8g8 g6f7 f8f7 d3h7 g8f8 f3e2 e7f6 g1f3 e5d4 c3d4 c5d4 e3d4 d8e7 h7g6 e7e2 e1e2 a8e8 e2f1');
INSERT INTO evaluations VALUES(19835,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(19835,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(19835,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(19835,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(19835,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(19835,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(19835,36,'cp 32','c4d5 d8d5 g1f3 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5e4 f3e5 c6e5 d4e5 g8e7 f1e2 e8g8 d1a4 d5c6 a4c6 b7c6 e1g1 a7a5 f1e1 e7d5 e2d1 e4e3 f2f3 c8e6 c3d4 f8d8 d4c5 a8b8 b2b3 a5a4 c5d6 a4b3 d1b3 b8b5 a1c1 d8c8 c1c5 b5c5 d6c5');
INSERT INTO evaluations VALUES(22939,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(22939,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(22939,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(22939,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(22939,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(22939,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(22939,36,'cp 13','e3d4 d7d5 b1c3 g8f6 f1b5 f8b4 e1g1 e8g8 c1g5 b4c3 b2c3 h7h6 g5h4 d8d6 b5c6 b7c6 f3e5 f6d7 e5d3 f8e8 f1e1 c8a6 d1d2 c6c5 d4c5 d7c5 h4g3 d6b6 a1b1 b6a5 d3b4 a8d8 d2d4 c5e4 d4a7 e4g3 h2g3');
INSERT INTO evaluations VALUES(23893,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23893,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23893,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23893,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23893,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23893,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23893,36,'cp 0','d2d3 f7f6 h2h3 g4h5 c1e3 c5e3 f2e3 d8e7 d1e1 h5f3 f1f3 a7a5 b1c3 g8h6 e1g3 a5a4 a1f1 e8g8 f3f2 g8h8 g1h1 h6f7 g3f3 e7b4 c3d1 f8d8 h1h2 b4e7 d1c3 e7b4 c3d1');
INSERT INTO evaluations VALUES(23890,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23890,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23890,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23890,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23890,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23890,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23890,36,'cp 0','f7f6 h2h3 g4h5 c1e3 c5e3 f2e3 g8h6 d1e1 d8e7 b1c3 h5f3 f1f3 h6f7 c3e2 e7c5 e2c3');
INSERT INTO evaluations VALUES(27582,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(27582,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(27582,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(27582,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(27582,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(27582,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(27582,36,'cp 40','e7e6 c1g5 d7b5 g5h4 b5c5 d4c5 d6c5 f1d1 g8e7 b1a3 e7c6 a3c4 h7h5 h2h3 f7f6 a2a4 a8d8 d1d8 e8d8 c2c3 f8e7 g1f1 d8c7 h4g3 c7c8 f1e2 b7b6 g3f4 h8d8 g2g4 g7g5 f4h2 h5g4 h3g4 c8b7 h2g3 a7a6 e4e5 f6f5 g4f5 e6f5');
INSERT INTO evaluations VALUES(31050,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(31050,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(31050,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(31050,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(31050,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(31050,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(31050,36,'cp -66','e7e5 e2e3 b8d7 g1f3 e5e4 f3d2 d8g5 b2b4 f8e7 b4b5 e7d8 a2a4 h7h5 d1b3 d8a5 h2h4 g5g6 c1a3 g8f6 a3b4 a5c7 a4a5 a7a6 b5b6 c7b8 b3c2 f6g4 d2b3 d7f6 e1c1 g4h2 c1b2 e8g8 g2g3 h2f1 d1f1 f5g4 f1b1 g4f3 h1c1 f6g4 b2a1 g4h2 c3a4 f8e8 b1b2 g6g4');
INSERT INTO evaluations VALUES(32233,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(32233,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(32233,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(32233,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(32233,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(32233,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(32233,36,'cp 117','g8f6 d4e6 f7e6 g2g3 b8c6 f1g2 g7g6 a2a4 f8g7 a4a5 a7a6 c3e2 d8d7 e2f4 e8g8 c2c3 a8e8 f4d3 c6d8 e1g1 e6e5 f2f3 h7h5 h2h4 d8f7 d3f2 d7c6 g2h3 g7h6 c1h6 f7h6 g1g2 g8g7 d1d2 h6f7');
INSERT INTO evaluations VALUES(30003,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(30003,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(30003,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(30003,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(30003,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(30003,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(30003,36,'cp 134','d4e6 f7e6 d1f3 b8c6 f1b5 d8d7 c1g5 e8c8 e1c1 c8b8 h1e1 c6e5 f3g3 c7c6 b5e2 d7c7 g5e3 h7h5 f2f4 e5d7 h2h3 e6e5 g3f2 a7a6 f4f5 f8e7 a2a3 d7c5 e2c4 c5e4 c3e4 f6e4 e3a7 b8a8 e1e4 d6d5');
INSERT INTO evaluations VALUES(39355,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(39355,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(39355,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(39355,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(39355,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(39355,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(39355,36,'cp -11','d7d6 g1e2 b8c6 a2a3 b4c5 b2b4 c5b6 c3d5 g8f6 d5b6 a7b6 b4b5 c6a5 e2g3 c7c6 b5c6 b7c6 d2d3 c8g4 f1e2 g4e2 g3e2 d6d5 c4d5 c6d5 c1g5 d5e4 g5f6 d8f6 d1a4 f6c6 a4c6 a5c6 d3e4 e8e7 e2c3 h8b8 e1g1 b6b5 f1b1 c6d4 h2h3 e7d6 b1d1 d6c6');
INSERT INTO evaluations VALUES(42714,30,'cp 378',NULL);
INSERT INTO evaluations VALUES(42714,31,'cp 428',NULL);
INSERT INTO evaluations VALUES(42714,32,'cp 434',NULL);
INSERT INTO evaluations VALUES(42714,33,'cp 446',NULL);
INSERT INTO evaluations VALUES(42714,34,'cp 432',NULL);
INSERT INTO evaluations VALUES(42714,35,'cp 426',NULL);
INSERT INTO evaluations VALUES(42714,36,'cp 432','e1f2 f6e4 f2g1 d7d5 c4b3 e8g8 d2d3 e4c5 b1c3 c5b3 a2b3 d5d4 c3e4 a7a5 h2h3 f7f6 e5f3 b7b6 d1e1 b8a6 e1f2 c8b7 c1f4 a6b4 f3d4 d8d7 g1h2 c7c5 d4f3 h7h6 e4d6 b7f3 g2f3 b4c6 f4g3 c6d4 a1f1 d7c6 c2c3');
INSERT INTO evaluations VALUES(42713,30,'cp 421',NULL);
INSERT INTO evaluations VALUES(42713,31,'cp 382',NULL);
INSERT INTO evaluations VALUES(42713,32,'cp 436',NULL);
INSERT INTO evaluations VALUES(42713,33,'cp 407',NULL);
INSERT INTO evaluations VALUES(42713,34,'cp 418',NULL);
INSERT INTO evaluations VALUES(42713,35,'cp 404',NULL);
INSERT INTO evaluations VALUES(42713,36,'cp 409','f6e4 f2g1 d7d5 c4b3 e8g8 d2d3 e4c5 b1c3 c5b3 a2b3 d5d4 c3e4 a7a5 h2h3 a8a6 c1f4 f7f5 e4g5 h7h6 h3h4 b8d7 g5f3 d7f6 h4h5 c7c5 d1e1 f8e8 e1g3 f6d5 f3h4 b7b5 a1e1 d5f4 g3f4 d8g5 f4g5 h6g5 h4g6 a6e6 g1f2 f5f4 e1e2 c8b7 h1e1 e8c8 e5f3 e6e3 c2c4 d4c3 g6e7 e3e7 e2e7');
INSERT INTO evaluations VALUES(42893,30,'cp 420',NULL);
INSERT INTO evaluations VALUES(42893,31,'cp 410',NULL);
INSERT INTO evaluations VALUES(42893,32,'cp 412',NULL);
INSERT INTO evaluations VALUES(42893,33,'cp 431',NULL);
INSERT INTO evaluations VALUES(42893,34,'cp 424',NULL);
INSERT INTO evaluations VALUES(42893,35,'cp 425',NULL);
INSERT INTO evaluations VALUES(42893,36,'cp 440','f2g1 d7d5 c4b3 d8f6 e5f3 f6b6 d2d4 c8e6 b1c3 e4c3 b2c3 b8d7 c1f4 a7a5 a2a4 e8c8 h2h3 h7h6 a1b1 b6c6 d1d3 d8e8 g1h2 g7g5 f3e5 d7e5 f4e5 h8g8 b1f1 f7f5 f1f5 g5g4 f5f2 g4h3 g2h3 e8f8 d3f1 f8f2 f1f2 c8b8 h1g1 g8g1 h2g1');
INSERT INTO evaluations VALUES(43732,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(43732,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(43732,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(43732,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(43732,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(43732,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(43732,36,'cp 143','a2a4 e7e5 d4b3 b5a4 a1a4 f8e7 e2b5 c8d7 b5d7 d8d7 b3a5 e8g8 b2b3 b8c6 a5c4 d7b7 c1e3 c6b4 e1g1 d6d5 e4d5 f6d5 c3d5 b7d5 d1d5 b4d5 c4e5 f8e8 e3d4 e7f6 f1e1 a8c8 c2c4 d5c7 f2f4 c8b8 b3b4 b8c8 d4e3 f6e5 f4e5 e8e5');
INSERT INTO evaluations VALUES(47213,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(47213,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(47213,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(47213,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(47213,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(47213,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(47213,36,'cp -17','c8f5 c2c4 c7c6 c4d5 c6d5 b1c3 e7e6 c1g5 f8e7 g1f3 b8d7 e2e3 a8c8 a1c1 e8g8 f1d3 f5d3 d1d3 e7d6 e1g1 h7h6 g5f6 d8f6 c1c2 d7b6 f1c1 a7a6 c3e2 c8c2 c1c2 f6e7 g2g3 e7d7 b2b3 f8c8 a3a4 d6b4 g1g2 c8c2 d3c2');
INSERT INTO evaluations VALUES(29597,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(29597,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(29597,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(29597,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(29597,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(29597,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(29597,36,'cp -20','c2c4 c7c6 c4d5 c6d5 b1c3 e7e6 g1f3 f8d6 c1g5 b8d7 e2e3 a8c8 f1d3 f5d3 d1d3 d8b6 e1g1 h7h6 g5f6 d7f6 c3a4 b6c6 a4c3 e8g8 f1c1 a7a6 d3d1 c6d7 f3e5 d7e7 e5d3 c8c7 c3e2 f8c8 c1c7 c8c7 a1c1 f6e4 d1a4');
INSERT INTO evaluations VALUES(47868,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(47868,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47868,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47868,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(47868,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47868,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47868,36,'cp 0','c7c5 d2d3 b8c6 f1g2 g7g6 a2a4 f8g7 e1g1 a8b8 c1g5 h7h6 g5f6 g7f6 f3d2 h6h5 f2f4 c8g4 g2f3 g4h3 f3g2');
INSERT INTO evaluations VALUES(49927,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(49927,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(49927,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(49927,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(49927,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49927,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(49927,36,'cp 19','f8b4 f4d2 b7b6 g1h3 e8g8 e2e3 c7c5 d4d5 e6d5 c4d5 b4c3 d2c3 f6e4 h3f4 d7d6 f1b5 d8e7 e1g1 e4c3 b2c3 b8d7 a2a4 d7e5 f4d3 a8b8 d3e5 e7e5 d1d2 f8f7 f2f3 f7e7 a1e1 e5f6 e3e4 f5e4 f3e4 f6e5 e1e3 c8d7 b5d7 e7d7');
INSERT INTO evaluations VALUES(42910,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(42910,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(42910,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(42910,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42910,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(42910,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(42910,36,'cp 3','f4d2 e8g8 e2e3 c7c5 a2a3 b4c3 d2c3 f6e4 g1e2 b7b6 d4d5 e4c3 e2c3 e6e5 d1c2 d7d6 e1c1 b8d7 f2f3 d7f6 f1e2 a7a6 d1g1 e5e4 f3e4 f6e4 c3e4 f5e4 g1f1 c8f5 f1f4 d8e7 h1f1 f5g6 e2g4 f8f4 f1f4 a8f8 g4e6 g8h8 g2g3 f8f4 g3f4 e7h4');
INSERT INTO evaluations VALUES(50056,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(50056,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(50056,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(50056,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(50056,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(50056,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(50056,36,'cp 81','d7d6 g1f3 b8d7 f1e2 a7a6 a2a4 c7c5 d4c5 d7c5 b2b4 c5d7 e1g1 d8c7 c1b2 e7g6 a4a5 f8e7 c3a4 e8g8 b2d4 e7f6 g2g3 g6e7 c2c3 h7h6 f1e1 f6d4 c3d4 e7c6 d1d2 b7b5 a5b6 d7b6');
INSERT INTO evaluations VALUES(46550,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(46550,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(46550,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(46550,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(46550,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(46550,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(46550,36,'cp 104','g1f3 a7a6 f1d3 b8d7 c3e2 g7g6 c1e3 f8g7 d1d2 c7c5 c2c3 e8g8 a1d1 b7b5 e1g1 d8c7 d3b1 d7b6 b2b3 c5c4 h2h4 c4b3 a2b3 a6a5 h4h5 a5a4 b1a2 c8b7 e2g3 a4b3 a2b3 b6c4 b3c4 c7c4 h5g6 e7g6');
INSERT INTO evaluations VALUES(50396,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(50396,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(50396,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(50396,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(50396,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(50396,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(50396,36,'cp -16','d7d6 c5d6 e7d6 g1f3 d6e5 d4e5 f8e7 b1c3 d5c3 d1d8 e7d8 b2c3 b8c6 f1c4 c6a5 c4e2 c8e6 f3d4 e6d5 e1g1 c7c6 d4f5 e8g8 f5d6 d8c7 c1a3 f8d8 f2f4 b7b5 a3c5 c7d6 c5d6 a5c4 d6c5 c4d2');
INSERT INTO evaluations VALUES(47062,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(47062,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(47062,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(47062,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(47062,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(47062,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(47062,36,'cp 0','c5d6 e7d6 g1f3 d6e5 d4e5 d5b4 d1d8 e8d8 b1a3 c8e6 f3g5 b8c6 g5e6 f7e6 f1c4 c6e5 c4e6 b4d3 e1e2 d3c1 h1c1 c7c6 a3c4 d8c7 c4e5 a8e8 e2f1 e8e6 e5f7 h8g8 c1e1 e6e7 f7g5 e7e1 a1e1 h7h6 g5e6 c7d7 e6f8 g8f8 e1d1 d7e6 d1e1');
INSERT INTO evaluations VALUES(52038,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(52038,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(52038,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(52038,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(52038,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(52038,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(52038,36,'cp -36','a3b4 f8b4 c2c3 b4e7 d2d4 d7d6 d1g4 g7g6 f1d3 g8f6 g4e2 e6e5 g1f3 e8g8 h2h3 f8e8 e1g1 e7f8 d4e5 d6e5 f1d1 d8c7 c3c4 c8e6 b1c3 a7a6 c1g5 b8d7 a1c1 a8c8 e2b2 c7b8 c3a4 b7b6 g5f6 d7f6 b2b6 f8a3 c1c2 b8b6');
INSERT INTO evaluations VALUES(52033,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(52033,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(52033,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(52033,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(52033,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(52033,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(52033,36,'cp -34','f8b4 c2c3 b4e7 d2d4 d7d6 d1g4 g7g6 f1d3 g8f6 g4e2 a7a6 c1h6 e6e5 g1f3 e7f8 h6f8 e8f8 d4e5 d6e5 f3e5 c8e6 e1g1 b8d7 e5f3 f8g7 f3d4 a8c8 d4e6 f7e6 d3c4 d7c5 f1d1 d8c7 b1d2 h8d8 e4e5 f6d5 c4d5');
INSERT INTO evaluations VALUES(37695,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(37695,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(37695,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(37695,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(37695,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(37695,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(37695,36,'cp -37','c2c3 b4e7 d2d4 d7d6 d1g4 g7g6 f1d3 g8f6 g4e2 e6e5 g1f3 e8g8 h2h3 b8d7 e1g1 b7b6 f1d1 d8c7 c1h6 f8e8 d4e5 d7e5 f3e5 d6e5 c3c4 e7b4 e2b2 a7a5 b1c3 c8b7 c3d5 b7d5 c4d5 c7d6 d3b5 f6e4 b5e8 a8e8 b2c2 e4c5 g1h1 e5e4');
INSERT INTO evaluations VALUES(37694,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(37694,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(37694,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(37694,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(37694,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(37694,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(37694,36,'cp -35','b4e7 d2d4 d7d6 f2f4 d6d5 e4e5 g8h6 f1d3 b8c6 g1h3 h6f5 e1g1 g7g6 h3g5 h7h5 d1b3 a7a5 c1a3 e7a3 b1a3 c6e7 f1b1 e8g8 h2h3 c8d7 b3c2 d8c7 g2g4 f5g7 a3b5 d7b5 b1b5 b7b6 g1h1 f7f6 g5f3 f6e5 f3e5 f8f4 d3g6 e7g6');
INSERT INTO evaluations VALUES(37716,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(37716,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(37716,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(37716,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(37716,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(37716,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(37716,36,'cp -46','d2d4 d7d6 f2f4 b8c6 f1d3 e6e5 f4e5 d6e5 d4d5 c6b8 g1f3 b8d7 e1g1 g8f6 b1a3 e8g8 d1e2 d8c7 a3b5 c7b8 c3c4 d7c5 d3c2 a7a6 b5c3 c8g4 g1h1 g4h5 c1e3 b8d6 e3g1 f6d7 c3a4 c5a4 c2a4');
INSERT INTO evaluations VALUES(37715,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(37715,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(37715,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(37715,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(37715,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(37715,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(37715,36,'cp -28','d7d6 f2f4 b8c6 f1d3 e6e5 g1f3 e5d4 c3d4 d6d5 e4e5 e7b4 c1d2 g8e7 b1a3 b4d2 d1d2 c8g4 a3b5 g4f3 b5d6 e8f8 g2f3 d8b6 d2e3 b6d4 e3d4 c6d4 e1f2 d4c6 h1b1 a8b8 b1b7 b8b7 d6b7 g7g6 d3b5 f8g7 b5c6 e7c6 b7a5 c6d8 f2g3 a7a6 a1d1 d8e6 d1d5 h8d8 d5d8');
INSERT INTO evaluations VALUES(52994,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(52994,31,'cp -130',NULL);
INSERT INTO evaluations VALUES(52994,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(52994,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(52994,34,'cp -160',NULL);
INSERT INTO evaluations VALUES(52994,35,'cp -156',NULL);
INSERT INTO evaluations VALUES(52994,36,'cp -160','g7g6 f1b5 c8d7 h5e2 d6e5 g1f3 f7f6 b1c3 e7e6 e1g1 d7b5 e2b5 d8d7 a2a4 b8c6 c3e4 g6g5 e4c5 f8c5 b5c5 g8e7 d2d3 e8f7 c1e3 a8c8 c5b5 h8d8 f1d1 h7h5 c2c3 g5g4 f3e1 d7c7 b5c4 e7d5 e3d2 c6a5 c4c7 d5c7 c3c4 a5b3 a1a3');
INSERT INTO evaluations VALUES(52595,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(52595,31,'cp -156',NULL);
INSERT INTO evaluations VALUES(52595,32,'cp -152',NULL);
INSERT INTO evaluations VALUES(52595,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(52595,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(52595,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(52595,36,'cp -133','f1b5 c8d7 h5e2 d6e5 g1f3 f7f6 b1c3 e7e6 c3e4 d7b5 e2b5 d8d7 b5d7 b8d7 d2d3 g8h6 a2a4 h6f7 h2h4 h7h6 b2b3 f8e7 c1b2 d7b8 g2g4 b8c6 e1c1 e8c8 h4h5 d8g8 d1g1 c8d7 c2c3 g6g5 c1c2 g8d8 f3d2 b7b6 e4g3 f7d6 d2c4 d7c7');
INSERT INTO evaluations VALUES(54675,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(54675,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(54675,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(54675,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(54675,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(54675,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(54675,36,'cp 131','e2e4 e7e5 f4e3 b8c6 g1f3 c8g4 d4d5 c6e7 b1c3 a7a6 f1e2 g7g6 d1d3 f8h6 e1c1 h6e3 d3e3 g4f3 e2f3 f6f5 g2g4 f5f4 e3d2 g6g5 h2h4 g8h6 h4g5 h6f7 h1h5 e7g8 f3e2 f7g5 a2a4 d8f6 f2f3 e8c8 c3a2 h7h6 h5h1');
INSERT INTO evaluations VALUES(56650,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(56650,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(56650,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(56650,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(56650,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(56650,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(56650,36,'cp -78','f8c5 c1e3 g8f6 b1c3 f6g4 d4e6 c8e6 d1d8 e8d8 e3c5 b8d7 a1d1 d8c8 c5d4 c7c5 d4g7 h8g8 h2h3 g4f2 e1f2 g8g7 f1e2 a7a6 g2g4 a8b8 a2a4 d7e5 a4a5 b7b5 a5b6 b8b6 c3a4 b6b5 b2b3 g7g6 b3c4 g6f6 f2e1 e5c4 h1f1 f6f1 e2f1 c8b7 d1c1 b5a5 f1c4 a5a4 c4e6 a4e4 e1d2 e4e6 c1c5');
INSERT INTO evaluations VALUES(38919,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(38919,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(38919,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(38919,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(38919,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(38919,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(38919,36,'cp -62','c1e3 g8f6 b1c3 f6g4 d4e6 c8e6 d1d8 e8d8 e3c5 b8d7 a1d1 d8c8 c5d4 c7c5 d4g7 h8g8 h2h3 g4f2 e1f2 g8g7 g2g4 c8c7 f1e2 c7c6 d1d2 d7e5 h1d1 a7a6 a2a4 c6b6 c3d5 b6a5 d1c1 b7b6 d2d1 a8h8 d5f4 b6b5 d1d6 a5a4 d6a6 a4b4 a6b6 b4a5 f4e6 f7e6 b6e6');
INSERT INTO evaluations VALUES(2700,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(2700,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(2700,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(2700,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(2700,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(2700,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(2700,36,'cp 168','a4c6 d7c6 f3e5 f8e8 d2d3 e7b4 c2c3 b4d6 e5f3 c8g4 b1d2 f6d7 h2h3 g4h5 d2c4 c6c5 c4d6 c7d6 g2g4 h5g6 c1g5 f7f6 g5e3 d7e5 f3h4 g6f7 c3c4 b7b5 b2b3 e5c6 d1d2 a8b8 h4f5 f7e6 g1g2 d8d7 a1d1 e6f5 e4f5 c6e5');
INSERT INTO evaluations VALUES(3328,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(3328,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(3328,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(3328,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(3328,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(3328,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(3328,36,'cp 47','f1d3 e7d6 e1g1 d6f4 e3f4 c6e7 f3e5 b7b6 h2h4 c7c5 f1e1 g7g6 g2g3 f6h5 d2f1 a7a5 f1e3 f7f6 e5f3 h5g7 a2a4 a8a7 d1d2 a7c7 a1d1 c5c4 d3c2 c8d7 g1g2 g8h8 g3g4 c7b7 e1h1 h7h5 d1e1 h5g4 e3g4');
INSERT INTO evaluations VALUES(4138,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4138,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4138,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(4138,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4138,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4138,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4138,36,'cp 0','d8f6 c3d5 f6d8 c2c3 a7a6 e1g1 d7d6 c4b3 c5a7 d5e3 c8e6 d1c2 d8e7 f1e1 e7f6 e3d5 f6d8 d5e3');
INSERT INTO evaluations VALUES(9025,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(9025,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(9025,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(9025,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(9025,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(9025,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(9025,36,'cp 85','f6d5 c1d2 d5b6 c4b5 a7a6 b5c6 b7c6 c3e4 d8d5 e1g1 b4d2 e4d2 c6c5 d1c2 c8b7 f1e1 c5d4 c2c7 d5c6 c7c6 b7c6 f3d4 c6d5 a1c1 e8g8 d4c6 h7h6 b2b3 d5c6 c1c6 f8d8 c6d6 b6d5 e1c1 d5b4 a2a4 b4d3 c1c6 d3e5 c6a6 a8c8 g1f1 c8c1 f1e2 d8c8 a4a5 c8c2 a6b6 g8h7 a5a6 c1a1');
INSERT INTO evaluations VALUES(8925,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(8925,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(8925,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(8925,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(8925,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(8925,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(8925,36,'cp 85','c1d2 d5b6 c4b5 c8d7 b5d3 b4f8 d3e4 c6b4 a2a3 b4d5 c3d5 b6d5 e4d5 e6d5 d1b3 d8c8 b3d5 d7c6 d5b3 c8d7 e1g1 f8e7 f1e1 d7e6 b3e6 f7e6 e1c1 e8c8 d2e3 h8g8 c1c3 g7g5 a1c1 h7h5 f3e1 a7a5 e1d3 h5h4 g1f1 g5g4 f1e2 d8d5 d3f4 d5b5 c1c2 c8d7 f2f3 a5a4');
INSERT INTO evaluations VALUES(11371,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(11371,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(11371,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(11371,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(11371,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(11371,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(11371,36,'cp 63','b1d2 f8d6 f1d3 e8g8 f4g3 b7b6 e3e4 d5e4 d2e4 f6e4 d3e4 c8b7 d1e2 a8b8 e1g1 a6a5 e2c2 h7h6 e4h7 g8h8 h7d3 c6e7 f3e5 c7c5 d4c5 d6e5 g3e5 d8d5 f2f4 d5c5 e5d4 c5c6 c2f2 e7g6 d3g6 f7g6 f1e1 h8g8 a1d1');
INSERT INTO evaluations VALUES(11370,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(11370,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(11370,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(11370,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(11370,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(11370,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(11370,36,'cp 50','f8d6 f1d3 c6e7 f3e5 c7c5 h2h4 d8c7 d1f3 c5d4 e3d4 e8g8 h4h5 f6e8 g2g4 f7f6 f4g3 b7b5 e5g6 f8f7 e1g1 d6g3 g6e7 f7e7 f2g3 c8b7 h5h6 g7h6 a1e1 e8g7 e1e3 e6e5 f3f6 e7e6 f6f2 e5e4 d3e2 a8e8 d2b3');
INSERT INTO evaluations VALUES(7163,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(7163,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(7163,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(7163,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(7163,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(7163,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(7163,36,'cp 86','f1d3 f6d7 f4g3 e6e5 e3e4 e5d4 c3d4 d5e4 d2e4 d6b4 e4c3 d7f6 e1g1 e8g8 h2h3 f6h5 g3h2 b4d6 f3e5 h5f6 d3c4 c8e6 c4e6 f7e6 e5c6 b7c6 h2d6 c7d6 d1e2 f8e8 f1e1 f6d5 e2g4 d5f6 g4f4 f6h5 f4d2 h5f6 a1d1 f6d5 c3d5');
INSERT INTO evaluations VALUES(13041,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(13041,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(13041,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(13041,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(13041,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(13041,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(13041,36,'cp 16','c1e3 f8g7 e1g1 c8d7 d1d2 e8g8 f2f3 c6d4 e3d4 d8a5 a1d1 f8d8 e2c4 d7c6 c4b3 a5c7 a2a4 f6d7 f1e1 g7d4 d2d4 d7c5 d1d2 a8c8 e1d1 a7a6 h2h4 c7a5 f3f4 b7b5 a4b5 c5b3 c2b3 c6b5 d4f2');
INSERT INTO evaluations VALUES(13035,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(13035,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(13035,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13035,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13035,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(13035,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13035,36,'cp 7','f8g7 d1d2 e8g8 e1c1 d6d5 f2f3 d5e4 f3e4 c6d4 d2d4 c8e6 d4d8 f8d8 h2h3 g8f8 e2f3 f6d7 c3b5 g7e5 b2b3 d8c8 b5d4 d7c5 f3e2 c5e4 d4e6 f7e6 h1f1 e5f6 d1d7 e4c3 e2d3 c3a2 c1d2 a2c3 d7b7 c8d8 d2c1 c3a2 c1d2');
INSERT INTO evaluations VALUES(14608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14608,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(14608,36,'cp 0','g5f6 d8f6 h2h4 h6h5 d3d4 g7g6 f1b5 a7a6 c3d5 f6d8 d4e5 a6b5 d5f6 e8e7 f6d5 e7d7 d5f6');
INSERT INTO evaluations VALUES(14597,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14597,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14597,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14597,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14597,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14597,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14597,36,'cp 0','d8f6 h2h4 f6d8 d3d4 e5d4 f3d4 f8e7 c3d5 a7a6 d1d2 c6d4 d2d4 e8g8 g2g3 c7c6 d5e7 d8e7 e1c1 c8g4 d1e1 a8d8 e1e3 f8e8 f1g2 c6c5 d4d2 b7b5 h1e1 e7b7 c1b1 g8h8 a2a3 g4e6 f2f4 b5b4 a3b4 c5b4');
INSERT INTO evaluations VALUES(943,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(943,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(943,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(943,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(943,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(943,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(943,36,'cp 0','h2h4 c8e6 c3d5 f6d8 g2g3 h6h5 c2c3 g7g6 d1b3 c6a5 b3a3 a5c6');
INSERT INTO evaluations VALUES(14951,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(14951,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(14951,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(14951,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(14951,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(14951,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(14951,36,'cp 40','d5d4 c3e2 f8d6 c2c3 d4c3 b5c6 b7c6 e2c3 e8g8 h2h3 c8e6 f3g5 e6d7 e1g1 h7h6 g5f3 d7e6 c1e3 f6d7 a1c1 a7a5 b2b3 f7f5 e4f5 f8f5 c3e4 e6d5 f3d2 f5f7 d2c4 d7f6 e4d6 c7d6 d3d4 e5e4 e3f4 d5c4 b3c4 d8d7 f1e1 d7e6 d1c2 a5a4 f4h2 d6d5 c4d5 c6d5 h2e5');
INSERT INTO evaluations VALUES(14958,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(14958,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(14958,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(14958,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14958,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(14958,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14958,36,'cp 39','c3e2 f8d6 e1g1 e8g8 c2c3 d4c3 b5c6 b7c6 e2c3 h7h6 h2h3 c8e6 c1e3 f6d7 a1c1 a7a5 b2b3 f7f5 e4f5 f8f5 c3e4 e6d5 f3d2 f5f7 d2c4 d7f6 e4d6 c7d6 d3d4 e5e4 e3f4 d5c4 b3c4 d8d7 f4g3 a8e8 f1e1 d6d5 g3e5 d7f5 d1a4 f6h5 c4d5 f5f2 g1h1');
INSERT INTO evaluations VALUES(19093,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(19093,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(19093,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(19093,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(19093,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(19093,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(19093,36,'cp -32','g8f6 b1c3 f8b4 f1d3 d7d5 e4d5 d8d5 g1f3 c8g4 d1e2 e8f8 e1g1 d5h5 c3e4 c6d4 e2e3 d4f3 g2f3 g4f3 e4g3 h5d5 d3c4 d5c4 e3f3 h7h5 f3b7 a8e8 c1e3 a7a5 a2a3 b4d6 b7f3 h5h4 g3f5 d6e5 f1c1 c4d5 f3d5 f6d5 e3d4 h8h5 d4e5 h5f5 e5d4 f8g8 c1e1 e8e1 a1e1 h4h3 e1e8 g8h7');
INSERT INTO evaluations VALUES(12616,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(12616,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(12616,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(12616,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(12616,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(12616,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(12616,36,'cp -24','b1c3 f8b4 f1d3 d7d5 e4d5 d8d5 g1f3 c8f5 d1e2 f5e4 d3e4 b4c3 b2c3 d5e4 c1e3 e8g8 e1g1 f6d5 f1e1 d5e3 e2e3 e4e3 e1e3 f8d8 f3d4 c6d4 c3d4 g8f8 a1d1 a8c8 g2g3 d8d7 d4d5 f7f6 d1d2 b7b6 g1g2 f8f7 d5d6 c8c5');
INSERT INTO evaluations VALUES(12615,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(12615,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(12615,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(12615,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(12615,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(12615,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(12615,36,'cp -29','f8b4 f1d3 d7d5 e4d5 d8d5 g1f3 c8f5 d1e2 f5e4 d3e4 b4c3 c1d2 f6e4 d2c3 e8g8 e1g1 e4c3 b2c3 f8e8 e2b2 b7b6 a1d1 d5c5 f1e1 e8e1 f3e1 a8d8 d1d8 c6d8 h2h3 d8e6 e1c2 h7h5 c2e3 c5e5 g2g3 e5e4 b2b5');
INSERT INTO evaluations VALUES(7879,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(7879,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(7879,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(7879,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(7879,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(7879,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(7879,36,'cp -24','f1d3 d7d5 e4d5 d8d5 g1f3 c8f5 d1e2 f5e4 d3e4 b4c3 b2c3 d5e4 c1g5 e4e2 e1e2 f6e4 h1e1 e8g8 a1c1 e4g5 f3g5 a8d8 c1d1 f8e8 e2f1 e8e1 f1e1 d8e8 e1f1 h7h6 g5f3 g8f8 d1d7 e8e7 d7d5 f8e8 f3d4 c6d4 d5d4 e7c7 c3c4 c7c5 d4g4 b7b6 g4g7 c5c4');
INSERT INTO evaluations VALUES(19256,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(19256,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(19256,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(19256,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(19256,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19256,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19256,36,'cp 0','f8g7 f1g2 d7d6 e1g1 c8g4 b2b3 g4f3 g2f3 e7e6 c1b2 g8e7 e2e3 e8g8 d1c2 d6d5 c4d5 e6d5 c3e2 d5d4 c2c5 d4e3 b2g7 e3f2 f1f2 g8g7 e2f4 d8b6 c5g5 b6d4 a1e1 h7h6 f4h5 g7h7 h5f6 h7g7');
INSERT INTO evaluations VALUES(20917,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(20917,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(20917,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(20917,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(20917,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(20917,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(20917,36,'cp 94','d4d5 c6b8 c2c4 f8g7 b1c3 c8g4 f1e2 g7c3 b2c3 g4f3 e2f3 b8d7 e1g1 g8f6 c1h6 h8g8 h6g5 d7c5 f1e1 f6d7 f3g4 d7e5 g4h3 h7h6 g5h6 g6g5 f2f4 g5g4 f4e5 g4h3 e5e6 g8g2 g1h1 f7e6 h6e3 e8d7 e3c5 d6c5 d5e6 d7c8 d1f3 d8h8 a1b1 g2g6 e4e5');
INSERT INTO evaluations VALUES(22551,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(22551,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(22551,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(22551,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(22551,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(22551,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(22551,36,'cp 150','e5d4 d1d4 d7d6 e1g1 c8e6 b1c3 g8f6 f1e1 f8e7 c4e6 f7e6 e4e5 d6e5 d4e5 d8d7 e5e6 d7e6 e1e6 e8f7 e6e2 h8d8 c1f4 d8d7 a1d1 d7d1 c3d1 f6d5 f4d2 b7b5 g2g3 e7f6 d2a5 c7c6 g1f1 h6h5 e2e4 d5b6 b2b3 a8d8 a5b6 a7b6 f1e2');
INSERT INTO evaluations VALUES(22558,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(22558,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(22558,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(22558,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(22558,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(22558,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(22558,36,'cp 162','d1d4 d7d6 b1c3 g8f6 e4e5 f6d7 e5d6 f8d6 d4g7 d8e7 c4e2 d6e5 g7g4 d7f6 g4c4 c8f5 c1f4 e8c8 f4e5 e7e5 e1g1 f5e6 c4b5 e5b5 e2b5 c7c6 b5d3 a7a5 f2f3 f6d5 c3d5 d8d5 g1f2 h6h5 f1e1 b7b5 a1d1 c8c7 h2h3 h5h4 c2c4 d5c5 c4b5');
INSERT INTO evaluations VALUES(22687,30,'cp -607',NULL);
INSERT INTO evaluations VALUES(22687,31,'cp -604',NULL);
INSERT INTO evaluations VALUES(22687,32,'cp -612',NULL);
INSERT INTO evaluations VALUES(22687,33,'cp -595',NULL);
INSERT INTO evaluations VALUES(22687,34,'cp -605',NULL);
INSERT INTO evaluations VALUES(22687,35,'cp -601',NULL);
INSERT INTO evaluations VALUES(22687,36,'cp -605','d8g5 b1c3 g5g6 f2f3 g8e7 c3b5 e8d8 d2d4 a7a6 b5c3 e5d4 c3e2 f7f5 e2f4 g6f6 f1c4 g7g5 f4d3 f8g7 e1g1 d7d6 b2b4 e7g6 d1d2 f5f4 a2a4 h7h5 c1b2 g5g4 c4d5 g4g3 d5b3 h8e8 b4b5 c6e5 d3e5 d6e5 c2c3 g3h2 g1h1 a6b5 c3d4 e5d4 a4b5 a8a1');
INSERT INTO evaluations VALUES(1914,30,'cp -586',NULL);
INSERT INTO evaluations VALUES(1914,31,'cp -578',NULL);
INSERT INTO evaluations VALUES(1914,32,'cp -590',NULL);
INSERT INTO evaluations VALUES(1914,33,'cp -584',NULL);
INSERT INTO evaluations VALUES(1914,34,'cp -598',NULL);
INSERT INTO evaluations VALUES(1914,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(1914,36,'cp -595','a2a3 g8f6 b1c3 g5g6 d2d3 f8c5 c1e3 d7d6 d1d2 c8e6 f1e2 e8c8 e1g1 c5d4 f2f4 g6h6 e3d4 c6d4 e2d1 d6d5 e4d5 e6d5 d2f2 h6f4 f2f4 e5f4 f1f4 d5e6 f4f2 c8b8 c3e4 f6e4 d3e4 d4c6 c2c3 c6e5 d1c2 f7f6 a3a4 h8e8 a4a5 a7a6 b2b3 e6g4');
INSERT INTO evaluations VALUES(26934,30,'cp -409',NULL);
INSERT INTO evaluations VALUES(26934,31,'cp -403',NULL);
INSERT INTO evaluations VALUES(26934,32,'cp -405',NULL);
INSERT INTO evaluations VALUES(26934,33,'cp -395',NULL);
INSERT INTO evaluations VALUES(26934,34,'cp -409',NULL);
INSERT INTO evaluations VALUES(26934,35,'cp -386',NULL);
INSERT INTO evaluations VALUES(26934,36,'cp -420','c6e5 c4e2 g4e2 c3e2 d6d5 d2d4 e5c6 e4e5 d8d7 e1g1 g7g6 a2a4 g8e7 a4a5 e7f5 c2c3 c6d8 c1g5 f8e7 g5f6 e7f6 e5f6 d8e6 f1e1 e8g8 e2g3 f5g3 h2g3 d7d8 d1f3 e6g5 f3f4 g5e4 e1e4 d5e4 a1e1 f8e8 a5a6 b7b6 e1e4 e8e4 f4e4 d8e8 e4f3');
INSERT INTO evaluations VALUES(27085,30,'cp -405',NULL);
INSERT INTO evaluations VALUES(27085,31,'cp -401',NULL);
INSERT INTO evaluations VALUES(27085,32,'cp -389',NULL);
INSERT INTO evaluations VALUES(27085,33,'cp -409',NULL);
INSERT INTO evaluations VALUES(27085,34,'cp -413',NULL);
INSERT INTO evaluations VALUES(27085,35,'cp -403',NULL);
INSERT INTO evaluations VALUES(27085,36,'cp -404','c4b5 c7c6 b5e2 g4e2 d1e2 c6c5 b2b3 f8e7 c1b2 e5c6 f2f4 d8d7 e1c1 e8c8 g2g4 e7f6 g4g5 f6d4 c3b5 d4b2 c1b2 g8e7 e2f2 c8b8 f4f5 a7a6 b5c3 d6d5 f5f6 g7f6 g5f6 d5d4 f2g3 d7c7 f6e7 d4c3 d2c3 c6e7 d1f1 h8f8 f1f6 c7g3 h2g3 e7g6 f6f1 f7f6 h1h7 g6e5');
INSERT INTO evaluations VALUES(27955,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(27955,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(27955,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(27955,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(27955,34,'cp -145',NULL);
INSERT INTO evaluations VALUES(27955,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(27955,36,'cp -143','b4c3 b2c3 b7e4 f3d2 e4f5 d2c4 g8e7 c4e3 f5g6 h2h4 h7h6 c3c4 d7d5 c4d5 e6d5 c1b2 h6h5 c2c4 b8d7 e3d5 e7d5 c4d5 e8g8 f1b5 d7f6 b5c6 f6g4 c6a8 f8e8 e1f1 d8a8 f1g1 a8d5 a1c1 g6e4 f2f3');
INSERT INTO evaluations VALUES(28000,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(28000,31,'cp -128',NULL);
INSERT INTO evaluations VALUES(28000,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(28000,33,'cp -138',NULL);
INSERT INTO evaluations VALUES(28000,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(28000,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(28000,36,'cp -123','b2c3 b7e4 f3d2 e4f5 d2c4 d7d6 a3a4 g8f6 c1g5 b8d7 c4e3 h7h6 g5h4 c7c6 f1a6 d8c7 e3f5 e6f5 f2f3 e8g8 e1g1 d6d5 d1d2 f5f4 f1e1 a8e8 a6d3 e8e1 a1e1 f8e8 e1e8 f6e8 c3c4 d5c4 d3c4 e8f6 c4b3 d7f8 h4f2 f8e6 d2d3 c7d7 g1h1 f6d5');
INSERT INTO evaluations VALUES(27999,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(27999,31,'cp -139',NULL);
INSERT INTO evaluations VALUES(27999,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(27999,33,'cp -129',NULL);
INSERT INTO evaluations VALUES(27999,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(27999,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(27999,36,'cp -124','b7e4 f3d2 e4f5 d2c4 g8f6 c1g5 h7h6 g5h4 e8g8 f1e2 d7d5 c4e3 c7c5 e1g1 f5h7 d4c5 b6c5 e3g4 b8d7 c3c4 d5d4 e2d3 d8c7 g4f6 d7f6 d1f3 f6d7 f1e1 h7d3 c2d3 f8e8 h4g3 e6e5 f3e4 c7d6 a1b1 d6e6 b1b7 f7f5 e4d5 e6d5 c4d5 a8d8 b7a7');
INSERT INTO evaluations VALUES(27893,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(27893,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(27893,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(27893,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(27893,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(27893,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(27893,36,'cp -138','f3d2 e4f5 d2c4 g8e7 c4e3 f5g6 c3c4 d7d5 h2h4 h7h6 c4d5 e6d5 h4h5 g6h7 f1e2 e8g8 e1g1 b8d7 e3g4 e7f5 e2d3 f5d6 c1f4 h7d3 d1d3 d7f6 g4f6 d8f6 f4e5 f6e6 f1e1 f8e8 d3g3 d6f5 g3f4 e6d7 g2g4 f5h4');
INSERT INTO evaluations VALUES(30509,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(30509,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(30509,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(30509,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(30509,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(30509,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(30509,36,'cp 84','b1c3 a7a6 c1g5 d8c7 c3d5 b7d5 e4d5 c7e5 g5e3 e5d5 f1e2 e7e5 d4f5 d5d1 a1d1 b8c6 e2f3 g7g6 f5d6 f8d6 d1d6 g8e7 e3g5 a8a7 g5f6 e8g8 e1g1 e7f5 d6d2 f8e8 f3c6 d7c6 g2g4 f5g7 f1e1 e5e4 f2f3 e4e3 d2d6 e8e6 d6d8 e6e8 e1e3 e8d8 f6d8 a7b7 e3b3 b6b5 g1f2 g7e6 d8f6');
INSERT INTO evaluations VALUES(35514,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(35514,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(35514,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(35514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35514,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(35514,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(35514,36,'cp -12','d1h5 g7g6 h5d5 f7g7 d5e4 f8e7 e4d5 h8e8 f1e2 e7f6 d2d4 b8c6 c2c3 d7d6 b1a3 c8f5 d5f3 f6d4 e1g1 d4f6 a3c4 d6d5 c4e3 f5e4 f3h3 d5d4 e3g4 f6g5 f2f4 g5h4 g2g3 h7h5 h3h4 d8h4 g3h4 h5g4 e2g4 e8h8 f4f5 h8h4 f5f6 g7f8 g4d1 d4c3 b2c3 c6e5 c1g5');
INSERT INTO evaluations VALUES(39823,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39823,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39823,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39823,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(39823,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(39823,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(39823,36,'cp 0','e8g8 e2e3 d7d5 c4d5 e6d5 a1c1 b4e7 f1d3 c7c5 e1g1 c8b7 f3e5 b8c6 d3b1 a8c8 b1f5 c8a8');
INSERT INTO evaluations VALUES(43653,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(43653,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(43653,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(43653,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(43653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43653,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(43653,36,'cp -17','c7c5 d4c5 f8c5 g1f3 b8d7 e3e4 e6e5 f4g5 d8b6 e1g1 d5e4 c3e4 f6e4 d3e4 e8g8 d1d3 h7h6 g5h4 b6b2 f1b1 b2a3 e4b7 a8a7 b7d5 a3d3 c2d3 a6a5 h4g3 f8d8 d5c4 c8a6 c4a6 a7a6 f3e5 d7e5 g3e5 a6e6 d3d4 c5d4 e5d4 d8d4 b1b8 g8h7 b8a8 a5a4 h2h3 e6e2 a8a5 e2b2 a5a8 f7f6 a8a7 h7g6');
INSERT INTO evaluations VALUES(45235,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(45235,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(45235,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(45235,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(45235,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(45235,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(45235,36,'cp 87','c2c4 e7e6 b1c3 e6d5 c4d5 g7g6 e2e4 f8g7 f3d2 e8g8 f1e2 b8d7 e1g1 f8e8 a2a4 g6g5 f1e1 h7h6 d1c2 d7e5 d2f1 e5g6 f1e3 c8d7 c1d2 d8e7 f2f3 f6h5 g2g3 g7e5 a4a5 e5d4 g1h1 g6e5 c3d1');
INSERT INTO evaluations VALUES(49292,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(49292,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(49292,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(49292,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(49292,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(49292,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(49292,36,'cp -44','d8e7 e1d1 d6c4 e5c4 d7d5 f3d5 b8c6 d2d3 c8e6 d5b5 e8c8 c4a5 c6a5 b5a5 e7h4 f2f3 h4d4 a5e1 f8c5 e1e4 d4e4 f3e4 e6g4 d1d2 h8e8 h2h3 g4h5 g2g4 h5g6 b1c3 g6e4 c3e4 e8e4 c2c3 e4f4 d2c2 f4f2 c1d2 d8e8 h1f1 e8e2 f1f2 e2f2 a1e1 c5d6 c2d1 f2h2 e1e3 h2h1 d1c2 d6f4 e3f3 f4d2 c2d2 h1h2 d2c1 f7f6 a2a4 c8d7 d3d4 a7a5 b2b3 d7d6 c1d1');
INSERT INTO evaluations VALUES(34010,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(34010,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(34010,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(34010,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(34010,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(34010,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(34010,36,'cp -56','e1d1 d6c4 e5c4 d7d5 f3d5 b8c6 d2d3 c8e6 d5b5 e8c8 c4a5 c6a5 b5a5 e7f6 f2f3 f6d4 a5e1 f8c5 e1e4 d4e4 f3e4 e6g4 d1d2 h8e8 h2h3 g4h5 g2g4 h5g6 b1c3 g6e4 c3e4 e8e4 c2c3 e4f4 d3d4 c5d6 d2c2 f4f2 c1d2 d6f4 a1d1 d8e8 c2d3 f2f3 d3c4 f4d2 d1d2 f3g3 d2f2 f7f6 a2a4 e8e3 f2f1 g3g2 c4b3 e3e2');
INSERT INTO evaluations VALUES(50507,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(50507,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(50507,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(50507,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(50507,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(50507,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(50507,36,'cp 69','g1f3 b7b6 f1e2 c8b7 e1g1 d7d6 d2d4 c5d4 d1d4 b8d7 f1d1 d8c7 c1g5 g8f6 b2b4 f8e7 g5h4 d7e5 a1c1 e8g8 h4g3 f8d8 f3d2 a8c8 a2a3 h7h6 h2h3 e5g6 d4e3 c7c6 g3h2 b7a8 e2f1 c6b7 g1h1 a6a5 c3b5 a5b4 a3b4 d6d5 c4d5');
INSERT INTO evaluations VALUES(55853,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(55853,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(55853,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(55853,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(55853,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(55853,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(55853,36,'cp -16','g8f6 f1g2 b7b5 b1d2 f8e7 g1e2 c8b7 a2a4 b5b4 c2c4 b4c3 b2c3 e8g8 c3c4 c7c5 e1g1 b8d7 c1a3 a8b8 a1c1 c5d4 a3e7 d8e7 e2d4 d7c5 c4d5 f6d5 d2c4 d5b4 g2b7 e7b7 c4a5 b7b6 a5c4 b6a6 c4e5 f8c8 d1g4 a6a4 c1c4 a4a6 f1c1 b4d3');
INSERT INTO evaluations VALUES(48345,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(48345,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(48345,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(48345,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(48345,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(48345,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(48345,36,'cp -13','f1g2 b7b5 b1d2 c7c5 g1e2 c8b7 e1g1 b8d7 a2a4 b5b4 c2c4 b4c3 b2c3 f8e7 c3c4 e8g8 c1a3 a8c8 f1e1 b7a8 e2f4 d5c4 g2a8 c8a8 d2c4 c5d4 a3e7 d8e7 e3d4 a8c8 c4e3 f8d8 d1e2 d7f8 d4d5 e6d5 e3d5 f6d5 e2e7 d5e7 e1e7 f8g6 f4g6 h7g6 e7e1');
INSERT INTO evaluations VALUES(48346,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48346,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(48346,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(48346,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48346,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(48346,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(48346,36,'cp -14','b7b5 g1e2 b8d7 b1d2 c8b7 e1g1 c7c5 a2a4 a7a6 e3e4 d5e4 e2c3 c5d4 c3e4 d8c8 e4f6 d7f6 g2b7 c8b7 a4b5 a6b5 a1a8 b7a8 d1e2 a8b7 d2b3 f8e7 b3d4 b5b4 f1d1 e8g8 c1f4 f8c8 f4e5 h7h6 d4f3 f6e4 h2h4 e7f6 e5f6 e4f6 f3e5');
INSERT INTO evaluations VALUES(3744,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(3744,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(3744,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(3744,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(3744,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(3744,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3744,36,'cp 13','c8g4 h2h3 g4h5 b2b4 c5e7 b4b5 c6a5 e1e5 d5f6 b1d2 e7d6 e5e1 f8e8 c4b3 e8e1 d1e1 a5b3 a2b3 f6d5 c1b2 d5f4 e1e4 f4d3 b2a3 d6c5 a3c5 d3c5 e4c4 c5e6 a1e1 a7a6 b3b4 h7h6 b5a6 a8a6 c4b5 h5f3 d2f3 d8a8 e1e3 a8b8 b5e2 a6d6 g2g3 c7c6 c3c4 b8d8 g1g2');
INSERT INTO evaluations VALUES(23723,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(23723,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(23723,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(23723,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(23723,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(23723,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(23723,36,'cp 23','h2h3 g4h5 b2b4 c5b6 b4b5 c6a5 e1e5 d5f6 b1d2 f8e8 d1a4 e8e5 f3e5 f6d5 c4d5 d8d5 d3d4 f7f6 e5c4 a5c4 d2c4 a7a6 c4b6 c7b6 b5a6 a8a6 a4b4 b6b5 a2a3 h7h6 c1e3 h5e2 a1e1 e2c4 e3f4 a6e6 e1e6 d5e6 f4g3 e6e4 b4c5 e4c6 g3d6 c6c5 d6c5 g8f7 f2f3 c4d5 g1f2');
INSERT INTO evaluations VALUES(7002,30,'cp 343',NULL);
INSERT INTO evaluations VALUES(7002,31,'cp 332',NULL);
INSERT INTO evaluations VALUES(7002,32,'cp 356',NULL);
INSERT INTO evaluations VALUES(7002,33,'cp 363',NULL);
INSERT INTO evaluations VALUES(7002,34,'cp 354',NULL);
INSERT INTO evaluations VALUES(7002,35,'cp 346',NULL);
INSERT INTO evaluations VALUES(7002,36,'cp 348','d4g7 e7f6 g7g3 h7h5 b1c3 d7d6 c1d2 c8e6 e1c1 h5h4 g3d3 d8d7 d3e3 g8e7 f2f4 e6g4 f1b5 c7c6 b5e2 g4e2 c3e2 h8g8 g2g3 d7c7 e2c3 c7b6 e3f3 e8c8 g3g4 f6d4 f4f5 d4e5 g4g5 c8b8 d2e3 b6a5 c1b1 b8a8');
INSERT INTO evaluations VALUES(6958,30,'cp 350',NULL);
INSERT INTO evaluations VALUES(6958,31,'cp 359',NULL);
INSERT INTO evaluations VALUES(6958,32,'cp 345',NULL);
INSERT INTO evaluations VALUES(6958,33,'cp 340',NULL);
INSERT INTO evaluations VALUES(6958,34,'cp 362',NULL);
INSERT INTO evaluations VALUES(6958,35,'cp 354',NULL);
INSERT INTO evaluations VALUES(6958,36,'cp 332','e7f6 g7g3 d7d6 b1c3 g8e7 c1g5 e7g6 h2h4 h7h6 g5f6 d8f6 e1c1 f6e5 g3e3 e5f4 e3f4 g6f4 g2g3 f4g6 f1e2 c8d7 c3d5 e8c8 d5f6 g6f8 e2f3 a7a6 h1e1 d7e6 e4e5 f8g6 e5d6 d8d6 d1d6 c7d6 b2b3 g6e5 f3g2 b7b5 g2d5 e6d5 f6d5 h6h5 e1e4 e5g6 e4e3 g6e5 f2f4 e5c6 c1b2 c8d7');
INSERT INTO evaluations VALUES(6162,30,'cp 348',NULL);
INSERT INTO evaluations VALUES(6162,31,'cp 350',NULL);
INSERT INTO evaluations VALUES(6162,32,'cp 341',NULL);
INSERT INTO evaluations VALUES(6162,33,'cp 345',NULL);
INSERT INTO evaluations VALUES(6162,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(6162,35,'cp 341',NULL);
INSERT INTO evaluations VALUES(6162,36,'cp 337','g7g3 g8e7 b1c3 d7d6 c1g5 e7g6 h2h4 h7h6 g5f6 d8f6 e1c1 f6e5 g3e3 e5f4 e3f4 g6f4 g2g3 f4g6 f1e2 c8d7 c3d5 e8c8 d5f6 g6f8 e2f3 d7e6 e4e5 f8g6 e5d6 d8d6 d1d6 c7d6 h1d1 g6e5 f3g2 h8d8 a2a4 e5c6 f6d5 c8d7 d5c3 d7e7 b2b3 d8c8 c1b2 a7a6 c3d5 e6d5 g2d5');
INSERT INTO evaluations VALUES(6159,30,'cp 355',NULL);
INSERT INTO evaluations VALUES(6159,31,'cp 349',NULL);
INSERT INTO evaluations VALUES(6159,32,'cp 356',NULL);
INSERT INTO evaluations VALUES(6159,33,'cp 359',NULL);
INSERT INTO evaluations VALUES(6159,34,'cp 353',NULL);
INSERT INTO evaluations VALUES(6159,35,'cp 337',NULL);
INSERT INTO evaluations VALUES(6159,36,'cp 338','g8e7 b1c3 d7d6 c1g5 e7g6 h2h4 h7h6 g5f6 d8f6 e1c1 f6f4 g3f4 g6f4 g2g3 f4g6 f1e2 c8e6 c3b5 e8d8 f2f4 g6e7 e4e5 d6d5 g3g4 e6d7 c2c4 d7b5 c4b5 c7c6 h4h5 d8c7 g4g5 a8f8 e2d3 c6c5 d3c2 d5d4 c2e4 h6g5 f4g5 f8g8 d1g1 c5c4 h5h6 e7g6 e4g6 f7g6');
INSERT INTO evaluations VALUES(11167,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(11167,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(11167,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(11167,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(11167,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(11167,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(11167,36,'cp 35','c6a5 c4d3 e7e5 d4e2 f8e7 e1g1 c8e6 c3a4 f6d7 b2b3 b7b5 a4c3 e8g8 d1d2 d7f6 e2g3 f6g4 g3f5 g4e3 f5e7 d8e7 f2e3 a5b7 b3b4 f8c8 c3d5 e6d5 e4d5 e7e8 a1c1 g7g6 e3e4 e8d8 a2a3 d8b6 g1h1 g8g7 c2c4 b5c4 d3c4 b6d4');
INSERT INTO evaluations VALUES(14892,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14892,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(14892,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(14892,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(14892,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(14892,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14892,36,'cp 55','f8e7 g5f6 g7f6 d1d2 f6f5 e4c3 e6e5 d4d5 e5e4 d5c6 e4f3 d2e3 e8g8 f1d3 e7d6 e3f3 b7c6 e1g1 d8h4 g2g3 h4g4 f3c6 a8b8 c3d5 g4d4 d5e3 f5f4 a1d1 d4g7 e3f5 c8f5 d3f5 b8b2 g1g2 b2a2 d1d5 a2a3 f5d3 a3c3 c6d7 f7f5 d7a4 g7g4 f2f3 g4g7 d3f5');
INSERT INTO evaluations VALUES(8500,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(8500,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(8500,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(8500,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(8500,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(8500,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(8500,36,'cp 68','g5f6 g7f6 d1d2 d8d5 e4c3 d5a5 f1c4 c8d7 d4d5 c6e5 f3e5 f6e5 d5e6 f7e6 e1c1 e8c8 f2f3 h7h5 c1b1 c8b8 d2e2 a7a6 h1e1 h5h4 h2h3 e7d6 d1d2 d7c8 e1d1 a5b6 c3e4 d6a3 c4b3 a3e7 c2c3 d8d2 e4d2 h8d8 d2c4 d8d1 b3d1 b6g1 c4e5 g1h2 b1c2 h2g1');
INSERT INTO evaluations VALUES(18118,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(18118,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(18118,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(18118,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(18118,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(18118,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18118,36,'cp 36','d4d5 c6e7 c4e2 c8b7 c1g5 h7h6 g5e3 e6d5 c3d5 e7d5 e4d5 g8f6 c2c4 b5c4 e2c4 f8d6 e1g1 e8g8 f1e1 f8e8 h2h3 d8c7 d1d3 d6f4 e3f4 c7f4 b2b3 g8f8 g2g3 f4d6 f3h4 e8e1 a1e1 g7g6 d3d2 f8g7 h4g2 a8e8 e1e8 f6e8 g2e3 e8f6 d2b2 g7g8 g1h2 f6e4 c4d3 b7d5');
INSERT INTO evaluations VALUES(19736,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(19736,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(19736,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(19736,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(19736,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(19736,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(19736,36,'cp -32','c8d7 d2d3 e7e6 e1g1 f8d6 c2c4 g8e7 b1c3 e8g8 b5a4 d6c7 c4d5 e6d5 e3e4 c6d4 f3d4 d7a4 d1a4 c5d4 c3d5 e7d5 e4d5 d8d5 a4b3 d5f5 b3b7 a8c8 b7f3 c7b6 g2g4 f5b5 f4f5 f8e8 a2a4 b5b4 f5f6 e8e1');
INSERT INTO evaluations VALUES(26019,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(26019,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(26019,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(26019,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(26019,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(26019,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(26019,36,'cp -12','d2d3 e7e6 e1g1 f8d6 c2c4 g8e7 b1c3 e8g8 b5a4 d6c7 c4d5 e6d5 e3e4 c6d4 a4d7 d8d7 f3d4 c5d4 c3d5 e7d5 e4d5 d7d5 d1b3 d5f5 b3b7 a8c8 b7f3 c7d6 f1f2 f5b5 f4f5 f8e8 c1f4 d6f4 f3f4 b5d3 f2d2 d3e3 f4e3 d4e3 d2e2 e8e5 g2g3');
INSERT INTO evaluations VALUES(20991,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(20991,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(20991,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(20991,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(20991,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(20991,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(20991,36,'cp 4','e7e5 d4e5 c6e5 c1b2 f8d6 g1f3 d8f6 b2e5 d6e5 f3e5 f6e5 d1d4 e5d4 e3d4 c7c6 b1c3 g8e7 c3e2 a7a5 a2a4 e7f5 h2h4 h7h5 c2c3 f5d6 f2f3 c8f5 e1d2 g7g6 e2f4 b7b6 a1e1 e8d7 g2g3 h8e8 e1e8 a8e8 f1e2 d7c7 h1a1 d6b7 f4g2 b7d6');
INSERT INTO evaluations VALUES(27587,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(27587,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(27587,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(27587,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(27587,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(27587,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(27587,36,'cp 61','d4c5 d5d1 e1d1 e8c8 c1d2 e7e5 b2b4 f8e7 h2h3 g4f5 d1c1 e5e4 f3e1 e7f6 e1c2 g8e7 a2a4 e7d5 a1a3 a7a5 b4a5 f6e7 c2d4 f5d7 d4b3 d5c7 d2e3 c7d5 e3d4 c6d4 c3d4 e7g5 c1b2 d7c6 b1c3 d5b4 f1b5');
INSERT INTO evaluations VALUES(29569,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29569,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29569,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(29569,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29569,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(29569,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29569,36,'cp 0','g2g4 f5g6 h3h4 f6g4 e2e4 g4f6 e4e5 f6e4 c3e2 e7e6 h4h5 g6f5 f2f3 f8b4 c2c3 b4e7 f3e4 e7h4 e1d2 h4g5 d2e1 g5h4');
INSERT INTO evaluations VALUES(32043,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(32043,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(32043,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(32043,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(32043,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(32043,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(32043,36,'cp 51','g4h5 d2d4 b8d7 f1e1 g8f6 d1d3 c7c6 b1d2 a7a5 a2a4 d8c7 d4e5 d7e5 f3e5 d6e5 d3g3 h5g6 c4d3 f8b4 c2c3 f6h5 g3e3 b4d6 d2c4 e8g8 d3f1 h5f4 e3g3 f8d8 c1e3 c7e7 c4d6 e7d6 e3b6 d8d7 g1h2 g6h5 g3e3 h5g6 f2f3 f4d3');
INSERT INTO evaluations VALUES(32034,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(32034,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32034,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(32034,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(32034,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(32034,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(32034,36,'cp 47','g2g4 h5g6 d2d4 b8d7 d1e2 h6h5 g4g5 h5h4 b1c3 g8e7 e2e3 c7c6 f1d1 d8c7 c4e2 g6h5 f3e1 e5d4 e3d4 d7c5 f2f4 h5e2 c3e2 f7f5 g5f6 g7f6 g1h2 h8g8 b2b4 c5d7 d4d6 c7d6 d1d6 e7g6 d6d4 a7a5 b4a5 f8c5 d4c4 a8a5 a1b1 a5a2');
INSERT INTO evaluations VALUES(32666,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(32666,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(32666,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(32666,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(32666,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(32666,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(32666,36,'cp 5','f1e2 b8c6 c2c4 d5d6 d2d4 e7e6 e1g1 a8d8 c4c5 d6d5 c1f4 d5e4 f4g5 f8e7 g5e7 e8e7 d1d2 f7f6 a1d1 f5g4 d2e3 e4e3 f2e3 a7a5 a2a3 g4h5 h2h3 g7g5 d1d3 d8b8 g2g4 h5g6 d3c3 h8d8 e2b5 c6a7 b5c4 h7h5 g4h5 g6h5 b2b4 a5b4 a3b4 h5g6');
INSERT INTO evaluations VALUES(36364,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(36364,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(36364,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(36364,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(36364,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(36364,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36364,36,'cp 21','c7c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 a2a3 b4e7 g5f4 c8d7 c4e2 f8d8 b5d6 e7d6 d1d6 f6e4 d6d3 d7e8 f1d1 f7f6 d3d8 a8d8 d1d8 c6d8 f4c7 d8f7 f3e1 e4d6 f2f3 e8b5 g1f2 g8f8 c7a5 b5e2 f2e2 e6e5 e1c2 f8e7 e3e4');
INSERT INTO evaluations VALUES(37020,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(37020,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37020,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(37020,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(37020,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(37020,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(37020,36,'cp 50','c4d5 e6d5 f1d3 h7h6 g5h4 c7c6 g1f3 b4e7 e1g1 f6e4 h4e7 d8e7 f3e5 b8d7 f2f4 d7f6 d3e4 d5e4 d1e1 f8d8 a1d1 c8f5 e1h4 g8f8 h4f2 f8g8 d1d2 f6d5 g2g4 d5c3 b2c3 f5e6 f4f5 e6d5');
INSERT INTO evaluations VALUES(41499,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41499,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(41499,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(41499,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(41499,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(41499,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(41499,36,'cp -8','c7c5 d4c5 e8g8 c2c4 d5c4 d3c4 b8d7 c5c6 d7b6 c6b7 c8b7 c4e2 d8c7 b1c3 f8d8 d1b3 f6e4 f3d4 e4c3 b3c3 c7c3 b2c3 a8c8 c1a3 e7e5 d4b3 b6a4 f1c1 a4c3 e2f1 b7d5 a3b4 e5e4 c1c2 c3a2 c2c8 d8c8 a1a2 d5b3 a2a7');
INSERT INTO evaluations VALUES(44494,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(44494,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(44494,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(44494,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(44494,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44494,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44494,36,'cp 37','g1f3 d7d6 e5d6 e7e6 e1g1 f8d6 d2d4 e8g8 c1g5 d8c7 b1d2 d6e7 g5e7 d5e7 d1e2 b8d7 f1e1 e7d5 a2a3 b7b6 f3e5 c8b7 c4d5 c6d5 e5d7 c7d7 d2f3 a8d8 e2e3 d7c7 f3e5 c7c2 a1c1');
INSERT INTO evaluations VALUES(45661,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45661,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45661,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45661,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45661,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45661,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45661,36,'cp 0','f8b4 a2a3 b4c3 b2c3 d8a5 f3d2 c6c5 f4d6 c5c4 f1e2 a5d8 d6g3 d8a5');
INSERT INTO evaluations VALUES(46984,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(46984,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(46984,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(46984,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46984,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(46984,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(46984,36,'cp 21','b1c3 c7c5 g1e2 e7e5 g2g3 b8c6 f1g2 a7a6 e1g1 a8b8 a2a4 g7g6 d2d3 f8g7 h2h3 e8g8 c1g5 h7h6 g5e3 g8h7 a1b1 b7b6 g1h2 c8d7 b2b3 f6h5 c3d5 d7e6 e2c3 e6d5 c3d5 h5f6 d5c3 d8e7 h2g1 c6b4 d1d2 b8e8 g1h1 f6d7');
INSERT INTO evaluations VALUES(47561,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(47561,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(47561,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(47561,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(47561,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47561,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(47561,36,'cp 42','f1c4 f8g7 g1f3 e8g8 c1e3 f6h5 a2a4 h5f4 c4f1 b8a6 g2g3 f4h5 f1a6 b7a6 d1e2 f7f5 e4f5 c8f5 f3g5 d8b8 e1g1 b8b2 c3e4 b2b7 e2c4 a8e8 a1b1 b7c8 e3a7 h7h6 g5e6 f5e6 b1b8 c8d7 b8e8 f8e8 d5e6 d7e6 c4c7');
INSERT INTO evaluations VALUES(47633,30,'cp 242',NULL);
INSERT INTO evaluations VALUES(47633,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(47633,32,'cp 230',NULL);
INSERT INTO evaluations VALUES(47633,33,'cp 242',NULL);
INSERT INTO evaluations VALUES(47633,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(47633,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(47633,36,'cp 230','d4e5 f6g4 d1d5 d8d7 h2h3 g4e5 f3e5 d6e5 d5e5 d7e7 e5e7 f8e7 b1c3 c7c6 c1e3 e8g8 a2a4 b8d7 c4e2 f8e8 f2f3 e7h4 e3f2 h4f6 a4a5 d7e5 f2g3 e5g6 e1g1 h6h5 a1a4 h5h4 g3c7 c8e6 f3f4 f6c3 b2c3 f7f5 e4e5');
INSERT INTO evaluations VALUES(47605,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(47605,31,'cp 251',NULL);
INSERT INTO evaluations VALUES(47605,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(47605,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(47605,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(47605,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(47605,36,'cp 242','f6g4 d1d5 d8d7 h2h3 g4e5 f3e5 d6e5 d5e5 d7e7 e5e7 f8e7 b1c3 c7c6 c1e3 b8d7 f2f4 e8g8 a2a4 e7b4 e1f2 f8e8 f2f3 b7b6 c3e2 g7g5 h1d1 d7f6 e4e5 g5f4 e3f4 f6d7 c2c3 b4f8 e5e6 f7e6 e2d4 d7f6 d4f5 f6d5 f5h6 g8g7 h6g4 d5f4 f3f4 f8e7 g4e5 e8f8 f4e4 c8b7 c4e6');
INSERT INTO evaluations VALUES(48798,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48798,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48798,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48798,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48798,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48798,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48798,36,'cp 0','c4d5 f6d5 e2e4 d5c3 b2c3 b8c6 f1c4 d8f6 g1f3 f8c5 e1g1 e8g8 d2d3 h7h6 c1e3 c5b6 h2h3 f8d8 c4d5 c6e7 d5a2 e7c6');
INSERT INTO evaluations VALUES(48618,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48618,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48618,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48618,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48618,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48618,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48618,36,'cp 0','f6d5 e2e4 d5c3 b2c3 b8c6 f1c4 d8f6 g1f3 f8c5 e1g1 e8g8 d2d3 h7h6 c1e3 c5b6 h2h3 f8d8 c4d5 c6e7 d5a2 e7c6');
INSERT INTO evaluations VALUES(49028,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49028,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49028,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49028,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49028,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49028,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49028,36,'cp 0','e2e4 d5c3 b2c3 b8c6 f1c4 d8f6 g1f3 f8c5 e1g1 e8g8 d2d3 h7h6 c1e3 c5b6 h2h3 f8d8 c4d5 c6e7 d5a2 e7c6');
INSERT INTO evaluations VALUES(49846,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(49846,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(49846,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(49846,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(49846,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(49846,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(49846,36,'cp 64','g4e5 e2e3 d7d6 f2f4 e5d7 g2g3 g7g6 g1e2 f8g7 e3e4 d7c5 c1e3 e8g8 e3d4 g7d4 e2d4 b8c6 f1g2 f7f5 e1g1 c6d4 d1d4 c5e4 g2e4 f5e4 c3e4 d8e7 a1e1 e7g7 d4d5 g7f7 e4g5 f7d5 c4d5 h7h6 g5e6 f8f7 e1e3 c7c5 d5c6 b7c6 f1e1');
INSERT INTO evaluations VALUES(49875,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(49875,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(49875,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(49875,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(49875,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(49875,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(49875,36,'cp 57','e2e3 d7d6 f2f4 e5d7 f1e2 g7g6 e2f3 f8g7 g1e2 d8h4 g2g3 h4e7 e3e4 e8g8 c1e3 d7b6 b2b3 f7f5 e3d4 f5e4 f3e4 b8c6 d4g7 e7g7 e1g1 c8h3 e4g2 h3g2 g1g2 a8e8 d1d2 g7e7 a1d1 e7e3 f1f2 f8f7 g3g4 e3d2 d1d2 c6b4 a2a3 b4c6 e2g3 b6d7 c3d5 d7f6');
INSERT INTO evaluations VALUES(49874,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(49874,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(49874,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(49874,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(49874,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(49874,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(49874,36,'cp 65','d7d6 f2f4 e5d7 f1e2 g7g6 e2f3 f8g7 g1e2 d8h4 g2g3 h4e7 e3e4 e8g8 c1e3 d7f6 h2h3 c8e6 b2b3 e6d7 e3d4 d7c6 e1g1 b8d7 g3g4 a8e8 e2g3 d7c5 d4c5 d6c5 e4e5 f6d7 f3c6 b7c6 d1f3 f7f6 a1d1 f6e5 f4f5 e5e4 c3e4 d7f6 e4f6 g7f6 f5g6 h7g6 f3c6');
INSERT INTO evaluations VALUES(50876,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(50876,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(50876,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(50876,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(50876,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(50876,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(50876,36,'cp 80','d2d4 e5d4 c2c3 d4d3 c4d3 b8c6 c1e3 g8e7 b1d2 g7g5 d2f1 e7g6 f1g3 c6e5 g3h5 d8e7 f3e5 g6e5 d3e2 c8d7 d1d4 c7c5 d4d2 d7c6 h5g3 e7e6 e1g1 e8c8 f2f4 g5f4 e3f4 h8g8 a1d1 f8e7 g1h1');
INSERT INTO evaluations VALUES(51235,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(51235,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(51235,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(51235,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(51235,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(51235,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(51235,36,'cp 103','c2c4 e7e5 d4f5 d7d5 c4d5 c8f5 e4f5 f8b4 b1c3 g8f6 a2a4 f6d5 c1g5 d8g5 d1d5 a8a7 f1b5 b8d7 d5e5 g5e7 e5e7 e8e7 b5e2 h8c8 e1g1 b4c3 b2c3 c8c3 f1c1 c3c5 g2g4 e7f6 c1d1 f6g5 g1g2 h7h5 g4h5 g5f5 d1d6 a6a5 a1d1 d7e5 e2b5 a7b7');
INSERT INTO evaluations VALUES(54398,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(54398,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(54398,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(54398,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(54398,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(54398,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(54398,36,'cp -13','f8d6 d2d4 g8f6 b1c3 e8g8 e1g1 a7a6 f1e1 h7h6 c1d2 b7b5 c3e2 a6a5 a2a3 b8c6 d2f4 d6f4 e2f4 c8a6 f4d3 f6e4 d3e5 c6e5 f3e5 b5b4 f2f3 e4d6 d1d2 b4a3 a1a3 d8b6 c2c3');
INSERT INTO evaluations VALUES(6753,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6753,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(6753,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(6753,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(6753,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(6753,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(6753,36,'cp 32','g8f6 b5a4 e8g8 e1g1 f8e8 c1g5 a7a6 c2c3 d6e7 b1d2 h7h6 g5h4 b7b5 a4b3 d7d6 a2a4 c8d7 f1e1 f6h5 h4g3 h5g3 h2g3 b5b4 d2f1 a8b8 f1e3 e7f6 b3a2 g7g6 a2d5 c6e7 d5c4 b4c3 b2c3');
INSERT INTO evaluations VALUES(6797,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(6797,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(6797,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(6797,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(6797,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(6797,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(6797,36,'cp 91','d6e5 f2f4 e5f6 f1d3 f6d4 b1c3 c7c6 d3e2 d8b6 c3a4 d4f2 e1f1 b6d4 d1d2 g8e7 e2d3 f2h4 a4c3 e7g6 e4e5 h4e7 h2h4 h7h5 c3e4 d7d5 e5d6 c8f5 d2e3 d4e3 c1e3 f5e4 d3e4 e7d6 g2g3 g6e7 f1g2 d6c7 h1e1 c7b6 a1d1 a8d8 e4d3 b6e3 e1e3 e8f8 g2f2 g7g6');
INSERT INTO evaluations VALUES(7092,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(7092,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(7092,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(7092,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(7092,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(7092,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(7092,36,'cp 94','g2g4 c6d4 d2d4 e8g8 e1c1 f6d7 h2h4 d7e5 f1e2 b7b5 c1b1 d8a5 f3f4 e5c6 d4b6 c8d7 e4e5 a5b6 e3b6 d6d5 c3e4 a8b8 b6e3 b8c8 h4h5 c6b4 h5h6 d5e4 d1d7 g7h6 c2c3 b4d5 e3c1 f8d8 d7d8 c8d8 h1h6 e4e3 e2f3 d5f4 c1e3');
INSERT INTO evaluations VALUES(7087,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(7087,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(7087,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(7087,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(7087,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(7087,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(7087,36,'cp 86','e8g8 e1c1 c6d4 d2d4 f6d7 c1b1 b7b5 g4g5 e7g5 h1g1 e6e5 e3g5 e5d4 g5d8 f8d8 d1d4 d7c5 g1g2 c8h3 g2f2 c5e6 d4d2 b5b4 c3d5 h3f1 f2f1 a6a5 b2b3 g8f8 b1b2 a8b8 a2a4 b4a3 b2a3 d8c8 f1a1 b8b5 a3b2 g7g6 h2h4 c8c6 d2d1 f8g7 a1a4 c6c8 f3f4 c8b8 f4f5 e6c5 a4c4');
INSERT INTO evaluations VALUES(8407,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(8407,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(8407,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(8407,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(8407,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(8407,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(8407,36,'cp 66','f4d6 d8d6 e1g1 f6h5 g2g3 g7g6 f1e1 e8g8 b1d2 c8d7 f3e5 h5g7 d2f3 f7f6 e5d7 d6d7 a1c1 a8e8 c1c2 e8e7 c2e2 f8e8 d3c2 g7f5 a2a3 f5d6 h2h4 d6c4 c2b3 c4d6 b3a4 g8g7 a4c2 c6d8 a3a4 d8f7 d1d3 b7b6 f3d2 f6f5');
INSERT INTO evaluations VALUES(8859,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(8859,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(8859,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(8859,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(8859,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(8859,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(8859,36,'cp -60','g7g5 f3h2 g5g4 g2g3 g4h3 b1c3 c8d7 g1h1 h8g8 c1h6 c6d4 h6e3 c5b6 c3d5 f6d5 c4d5 c7c6 d5c4 d8f6 c2c3 d4e6 e3b6 a7b6 a3a4 e8c8 d1e2 e6c7 e2f3 f6g5 c4f7 g8g7 f3e3 g5e3 f2e3 g7g3 f1f3 g3g2 b2b4 d6d5 c3c4 d5c4 f7c4');
INSERT INTO evaluations VALUES(11012,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(11012,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(11012,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(11012,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(11012,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(11012,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(11012,36,'cp 19','c5b6 a2a4 e8g8 a4a5 b6a5 b5c6 e7c6 f3e5 a5c3 e5c6 d7c6 d2c3 f7f5 c1f4 f5e4 d1d8 f8d8 f4c7 d8d2 f1d1 d2d1 a1d1 c8f5 c7f4 a7a5 d1d4 a5a4 h2h3 g8f7 d4b4 b7b5 c3c4 f5d7 f4e3 a4a3 b2a3 a8a3 c4b5 c6b5');
INSERT INTO evaluations VALUES(12145,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12145,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12145,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12145,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12145,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12145,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12145,36,'cp 0','e6d5 f1d3 f8d6 e1g1 e8g8 d1c2 f8e8 f1e1 h7h6 h2h3 d7b6 c1d2 c8e6 a2a4 b6d7 e1d1 a7a5 d1e1 d8e7 e1c1 a8c8 d2e1 b7b6 d3a6 c8c7 a6d3');
INSERT INTO evaluations VALUES(12192,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12192,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12192,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12192,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12192,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12192,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12192,36,'cp 0','f1d3 f8d6 e1g1 e8g8 d1c2 f8e8 f1e1 d7f8 h2h3 a7a6 c1d2 h7h6 a2a3 c8e6 a1c1 f8d7 e1d1 a8c8 d2e1 d6b8 c2b1 b8c7 b1c2');
INSERT INTO evaluations VALUES(12193,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12193,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12193,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12193,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12193,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12193,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12193,36,'cp 0','f8d6 e1g1 e8g8 d1c2 f8e8 f1e1 a7a6 h2h3 d7f8 c1d2 c8e6 a2a3 h7h6 a1c1 f8d7 e1d1 a8c8 d2e1 d6b8 c2b1 b8c7 b1c2');
INSERT INTO evaluations VALUES(7745,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7745,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7745,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7745,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7745,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7745,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7745,36,'cp 0','e1g1 e8g8 d1c2 f8e8 f1e1 d7b6 h2h3 h7h6 a2a4 a7a5 c1d2 c8d7 a1d1 d8c8 e3e4 d5e4 c3e4 f6e4 d3e4 d7e6 e4h7 g8h8 h7d3 h8g8');
INSERT INTO evaluations VALUES(7744,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(7744,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7744,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7744,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7744,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7744,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7744,36,'cp 0','e8g8 f1e1 f8e8 d1c2 d7f8 h2h3 h7h6 c1d2 c8e6 e1e2 a7a6 a1e1 f8d7 a2a3 a6a5 e1c1 d6c7 c3b5 c7b8 b5c3');
INSERT INTO evaluations VALUES(2925,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2925,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2925,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2925,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2925,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2925,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2925,36,'cp 0','d1c2 f8e8 f1e1 h7h6 c1d2 d7f8 h2h3 a7a6 e1e2 c8e6 a1e1 d8c8 e3e4 e6h3 e4e5 c8g4 f3h4 g4h4 g2h3 d6b8 e5f6 h4h3 f2f4 h3g3 g1h1 g3h4 h1g2 h4g4 g2f2 g4h4');
INSERT INTO evaluations VALUES(16594,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(16594,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(16594,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(16594,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(16594,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(16594,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(16594,36,'cp -72','e5f4 c1f4 d7d5 a2a3 f8c5 c3d5 f6d5 e4d5 d8d5 d1f3 c6e7 f3d5 e7d5 f4g3 c8g4 f1e2 g4e2 g1e2 d5e3 e1d2 e8c8 g3f2 h8e8 f2e3 c5e3 d2c3 g7g6 a1f1 f7f5 g2g3 c7c6 h2h4 a7a5 c3b3 e3d2 e2f4 e8e3 f1f2 d2e1 f2e2 e3e2 f4e2');
INSERT INTO evaluations VALUES(17084,30,'cp 291',NULL);
INSERT INTO evaluations VALUES(17084,31,'cp 293',NULL);
INSERT INTO evaluations VALUES(17084,32,'cp 328',NULL);
INSERT INTO evaluations VALUES(17084,33,'cp 317',NULL);
INSERT INTO evaluations VALUES(17084,34,'cp 316',NULL);
INSERT INTO evaluations VALUES(17084,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(17084,36,'cp 331','g7g6 b1c3 h5g7 d4e4 g7e6 c1e3 f8g7 e1c1 f7f5 e4a4 c7c6 f2f4 b7b5 a4b3 d8e7 e2f3 e8g8 a2a4 a7a6 a4a5 e7f7 c1b1 f8e8 g2g4 f5g4 f3g4 d7d5 e5d6 c8d7 e3b6 e6f8 b3f7 g8f7 f4f5 g7c3 b2c3 g6f5 g4f3 f8g6 h2h4 e8e5 f3h5 a8e8');
INSERT INTO evaluations VALUES(16089,30,'cp 350',NULL);
INSERT INTO evaluations VALUES(16089,31,'cp 346',NULL);
INSERT INTO evaluations VALUES(16089,32,'cp 350',NULL);
INSERT INTO evaluations VALUES(16089,33,'cp 340',NULL);
INSERT INTO evaluations VALUES(16089,34,'cp 338',NULL);
INSERT INTO evaluations VALUES(16089,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(16089,36,'cp 329','b1c3 h5g7 d4e4 g7e6 c1e3 f8g7 e1c1 f7f5 e4a4 e8g8 f2f4 d8e8 g2g3 c7c6 h2h4 b7b5 a4a5 e8d8 a5d8 f8d8 h4h5 d8e8 g3g4 f5g4 h5g6 h7g6 e2g4 d7d5 e5d6 c8d7 a2a4 e6f8 g4d7 f8d7 a4b5 g7c3 b5c6 c3g7 e3d4 g7d4 d1d4 d7f6');
INSERT INTO evaluations VALUES(19250,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(19250,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(19250,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(19250,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(19250,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(19250,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(19250,36,'cp -31','f8g7 b1c3 d7d6 f1e2 c8g4 f3g1 g4d7 g1f3 e7e6 c1f4 d7c8 f4e3 c6d4 d1d2 g8e7 e2d1 e8g8 e3h6 a8b8 h2h4 g7h6 d2h6 f7f6 f3d4 c5d4 c3e2 e6e5 f2f3 c8d7 d1b3 d8a5 h6d2 a5d2 e1d2 g8g7 h4h5 f6f5 h5g6 h7g6 e4f5 e7f5 a1b1');
INSERT INTO evaluations VALUES(20250,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(20250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20250,36,'cp 0','d5c4 d4c5 d8d1 e1d1 g8f6 f1c4 f8c5 b1d2 c8d7 d1e2 e8g8 a1c1 c5e7 h1d1 f8d8 d2b3 a8c8 f3e5 c6e5 f4e5 f6d5 c4d5 d7b5 e2e1 c8c1 d1c1 d8d5 c1c7 d5d7 c7c8 d7d8 c8c7');
INSERT INTO evaluations VALUES(20417,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(20417,31,'cp -155',NULL);
INSERT INTO evaluations VALUES(20417,32,'cp -174',NULL);
INSERT INTO evaluations VALUES(20417,33,'cp -186',NULL);
INSERT INTO evaluations VALUES(20417,34,'cp -189',NULL);
INSERT INTO evaluations VALUES(20417,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(20417,36,'cp -166','c6e5 f2f4 e5c6 d1d2 c8d7 b1c3 g8f6 e1c1 a8c8 f1d3 f8b4 c1b1 c6a5 d4e2 e8g8 h1f1 f8e8 e3d4 a5c6 d4f6 d8f6 a2a3 b4a5 d2e3 f6h4 g2g3 h4d8 g3g4 d8h4 e3g1 a7a6 f4f5 c6e5 h2h3 e5d3 f5e6 f7e6 d1d3 e8f8 d3f3 h4e7 g1a7 f8f3 f1f3 d7e8 e2d4 a5c3 f3c3');
INSERT INTO evaluations VALUES(20780,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(20780,31,'cp -151',NULL);
INSERT INTO evaluations VALUES(20780,32,'cp -162',NULL);
INSERT INTO evaluations VALUES(20780,33,'cp -153',NULL);
INSERT INTO evaluations VALUES(20780,34,'cp -169',NULL);
INSERT INTO evaluations VALUES(20780,35,'cp -157',NULL);
INSERT INTO evaluations VALUES(20780,36,'cp -148','f2f4 e5c6 d1d2 g8f6 b1c3 f8b4 f1d3 b4c5 d3e2 c8d7 e1c1 c6d4 e3d4 d8b6 d4c5 b6c5 a2a3 e8g8 g2g4 a8c8 h2h4 a7a5 g4g5 f6e8 d2d4 e8d6 d4c5 c8c5 e2d3 f8b8 h4h5 b8c8 h1g1 b7b5 g5g6 f7g6 h5g6 h7h6 c3e2 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(22106,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(22106,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(22106,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(22106,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(22106,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(22106,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(22106,36,'cp 112','c4e2 g8f6 f2f4 e5c6 b1c3 g7g6 c1e3 f8g7 d1d3 e8g8 e1c1 f8e8 e2f3 c8g4 h1e1 g4f3 d4f3 d8e7 e3f2 g7h6 d3d2 c6e5 f2h4 e5f3 g2f3 c7c6 c1b1 e7e6 d2d6 f6d5 d6e6 d5c3 b2c3 e8e6 f4f5 g6f5 e1g1 g8f8 e4f5 e6e5');
INSERT INTO evaluations VALUES(25505,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(25505,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(25505,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(25505,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(25505,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(25505,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(25505,36,'cp 9','a7a6 b5c6 d7c6 e1g1 f8e7 f1e1 e8g8 d3d4 f8e8 d1d3 e5d4 f3d4 c6d7 c1f4 b7b5 a1d1 b5b4 c3d5 f6d5 e4d5 d8b8 f4d2 b8b6 c2c4 b4c3 d2c3 e7h4 d3f3 h4f6 d4c2 f6e5 c3e5 e8e5 e1e5 d6e5 c2e3 b6b2 e3c4 b2a2');
INSERT INTO evaluations VALUES(24313,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24313,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24313,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(24313,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(24313,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24313,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24313,36,'cp 16','b5c6 d7c6 e1g1 f8e7 f1e1 e8g8 d3d4 f8e8 d1d3 e7f8 a2a4 h7h6 c1e3 d8c8 d4d5 c6d7 f3d2 f6h5 c3e2 f8e7 a4a5 f7f5 e2g3 h5g3 f2g3 f5e4 d2e4 d7f5 e1f1 c8d7 d3c4 f5g6 e4c5 d6c5 d5d6 g8h7 d6e7 e8e7 e3c5');
INSERT INTO evaluations VALUES(27269,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(27269,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(27269,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(27269,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(27269,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(27269,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(27269,36,'cp 39','d7c6 b1c3 d8d7 h2h4 e8c8 c1f4 d7f7 a2a4 f8b4 f4d2 b4c3 d2c3 f7h5 d1e2 c6e8 c3b4 h5f5 e1g1 e8h5 a1a3 h5f3 a3f3 f5e4 e2e4 d5e4 f3g3 f6f5 f1d1 d8d7 b4f8 g7g6 a4a5 d7f7 f8c5 b7b6 c5a3 g8h6 a5b6 a7b6 g3c3');
INSERT INTO evaluations VALUES(34245,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(34245,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(34245,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(34245,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(34245,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(34245,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(34245,36,'cp 96','g8f6 g1f3 b8c6 a2a3 d7d6 f1e2 c5a5 e1g1 a5d8 c1f4 c8g4 f1e1 g4f3 e2f3 c6e5 f3e2 a7a6 c3d5 f6d7 f4e3 e7e6 d5b4 f8e7 f2f4 e5c6 e2f3 c6b4 a3b4 e8g8 c2c3 a6a5 e4e5 d6e5 f3b7 e5f4 b7a8 f4e3');
INSERT INTO evaluations VALUES(35706,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(35706,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(35706,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(35706,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(35706,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35706,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(35706,36,'cp 44','d4d5 d7d6 h2h3 e7e6 f1d3 b8a6 e1g1 e6d5 c4d5 a6c7 e3e4 b7b5 f1e1 f8e8 a2a3 a7a5 c1f4 c8a6 a1c1 b5b4 c3a4 a6d3 d1d3 a8b8 f4h2 g7f8 e4e5 d6e5 h2e5 f8d6 a4c5 d6e5 e1e5 e8e5 f3e5 d8d5 d3d5 c7d5 e5c6 b8c8');
INSERT INTO evaluations VALUES(37993,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(37993,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(37993,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(37993,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(37993,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(37993,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(37993,36,'cp 56','g8f6 b1c3 e8g8 c4b3 b8c6 h2h3 b7b5 e1g1 b5b4 c3d5 f6d5 b3d5 c8b7 f1e1 a8c8 a2a3 b4a3 a1a3 d8b6 d5c6 c8c6 c1e3 c6c7 a3b3 b6a5 d4d5 f8c8 f3d4 a7a6 e3d2 a5a2 d2c3 a2a4 b3b6 a4a2 d1d3 c7c4');
INSERT INTO evaluations VALUES(40383,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(40383,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(40383,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(40383,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(40383,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(40383,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(40383,36,'cp 41','b1c3 e8g8 c4b3 b8c6 h2h3 b7b5 e1g1 c6a5 e4e5 f6e8 f1e1 e8c7 c1g5 d6e5 f3e5 a5b3 d1b3 c7e6 e5c6 e6d4 c6d4 d8d4 a1d1 d4c4 b3b5 c4b5 c3b5 a7a6 b5c7 a8a7 g5e7 a7c7 e7f8 g7f8 e1e8 g8g7 d1d8');
INSERT INTO evaluations VALUES(40382,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(40382,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40382,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(40382,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(40382,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40382,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40382,36,'cp 46','e8g8 c4b3 b8c6 h2h3 b7b5 e1g1 c6a5 e4e5 f6e8 c3b5 a5b3 a2b3 c8b7 b5a3 d6e5 d4e5 e8c7 a3c4 c7e6 c1e3 d8b8 f3g5 f8d8 d1e2 b7d5 f1d1 e6g5 e3g5 d8d7 e2e3 b8b7 c4a5 b7c7 g5f4 d5e6 d1d7 c7d7 f4h6 a8c8 h6g7 g8g7');
INSERT INTO evaluations VALUES(35998,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(35998,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(35998,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(35998,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(35998,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(35998,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(35998,36,'cp 38','c4b3 b8c6 h2h3 b7b5 e1g1 c6a5 e4e5 f6e8 f1e1 e8c7 e5d6 d8d6 c1g5 c8e6 b3e6 c7e6 c3e4 d6d8 g5h4 e6d4 f3d4 d8d4 h4e7 d4d1 a1d1 f8c8 e4f6 g8h8 g2g4 a5c6 d1d6 a7a5 g1g2 c6e7 e1e7 c8f8 g4g5 h7h6 h3h4 g7f6 g5f6 h8g8 e7e5 b5b4 g2g3 f8c8 h4h5');
INSERT INTO evaluations VALUES(45604,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(45604,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45604,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(45604,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45604,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(45604,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45604,36,'cp 33','e6d5 e2e3 c8f5 d1f3 f5g6 g5f6 d8f6 f3f6 g7f6 h2h4 h7h5 g1e2 b8d7 g2g3 f8d6 e2f4 a7a5 a1c1 e8d8 f2f3 d7f8 f1h3 f6f5 e1f2 f8e6 f4d3 a5a4 h1d1 h8g8 h3g2 g8e8 c3e2 d6e7 d3e5 e7d6 f3f4');
INSERT INTO evaluations VALUES(54251,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54251,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54251,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54251,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54251,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54251,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54251,36,'cp 0','e4d5 e6d5 d3b5 b8c6 d2d4 g8f6 e1g1 f8e7 b1c3 e8g8 d4c5 e7c5 c1g5 a7a6 g5f6 d8f6 c3d5 f6b2 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(55034,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(55034,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(55034,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(55034,33,'cp -178',NULL);
INSERT INTO evaluations VALUES(55034,34,'cp -178',NULL);
INSERT INTO evaluations VALUES(55034,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(55034,36,'cp -161','d5e4 d2d4 e4d3 f1d3 b8c6 g1f3 c8g4 e1g1 c6e5 d3e2 e5f3 g2f3 g4h3 f1e1 d8d6 g1h1 e7e5 b1c3 d6d1 e1d1 c7c6 c3e4 g8f6 c1e3 f6e4 f3e4 f8e7 a2a3 e8g8 b2b4 h7h6 h1g1 h3e6 c4c5 f8d8 d1d8 a8d8 a1c1 f7f6 f2f3 a7a6 c1d1 e6b3 d1d2 d8d2');
INSERT INTO evaluations VALUES(55191,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(55191,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(55191,32,'cp -164',NULL);
INSERT INTO evaluations VALUES(55191,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(55191,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(55191,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(55191,36,'cp -168','d2d4 e4d3 f1d3 b8c6 g1f3 c8g4 e1g1 c6e5 d3e2 e5f3 g2f3 g4h3 f1e1 d8d6 g1h1 d6d1 e1d1 c7c6 b1c3 e7e5 a2a3 f8e7 b2b4 g8f6 c1e3 h3f5 b4b5 e8g8 e2d3 f5e6 b5c6 b7c6 c3e4 f6e4 d3e4 f8c8 f3f4 e5f4 e3f4 e6c4 a1c1 c4e6 e4c6 e7a3 c6a8');
INSERT INTO evaluations VALUES(57135,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(57135,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(57135,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(57135,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(57135,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(57135,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(57135,36,'cp -87','e5d4 d1d4 b8c6 d4e4 f8e7 c1g5 d7d5 g5e7 c6e7 e4f4 g8f6 e2e3 e8g8 h4h5 h7h6 b1c3 c7c5 e1c1 c8e6 g1e2 b7b5 g2g4 b5b4 c3a4 d8a5 b2b3 f6e4 f1g2 a8c8 g2e4 d5e4 f4e4 f8e8 g4g5 h6g5 h5h6 e6f5 e4e5 f7f6 e5g3 c5c4 e2d4 c4b3 a2b3');
INSERT INTO evaluations VALUES(57302,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(57302,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(57302,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(57302,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(57302,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(57302,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(57302,36,'cp -66','d1d4 b8c6 d4d3 d7d5 g1f3 g8f6 a2a3 h7h6 c2c4 d5c4 d3c4 c8e6 c4c2 d8d7 b1c3 f8d6 g2g3 e8g8 f1g2 f8e8 e1g1 a7a6 f1e1 a8d8 b2b4 e6f5 e2e4 f5g4 c1b2 g4f3 g2f3 c6d4 c2d3 d4f3 d3f3 d6e5 g1g2 c7c6 a1d1 d7e6');
INSERT INTO evaluations VALUES(57296,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(57296,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(57296,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(57296,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(57296,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(57296,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(57296,36,'cp -64','b8c6 d4d3 g8f6 a2a3 d7d5 g1f3 f8c5 c1g5 c8e6 e2e3 h7h6 g5f4 e8g8 f1e2 f8e8 e1g1 d8d7 b2b4 e6f5 d3b3 c5d6 f4d6 d7d6 b1c3 a7a6 f1d1 e8d8 b4b5 c6e5 b5a6 b7a6 f3e5 d6e5 e2f3 c7c6 c3e2 c6c5 e2f4 c5c4');
INSERT INTO evaluations VALUES(22928,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(22928,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(22928,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(22928,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(22928,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(22928,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(22928,36,'cp -55','d4a4 g8f6 c2c3 f8c5 c1f4 e8g8 b1d2 f8e8 e2e3 d7d6 f1d3 h7h6 g1f3 c5b6 d2e4 f6g4 e4g3 c6e7 e1g1 e7d5 a1d1 d5f4 e3f4 a7a5 c3c4 c7c6 a4c2 d8f6 d3f5 g7g6 f5c8');
INSERT INTO evaluations VALUES(2873,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(2873,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(2873,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(2873,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(2873,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(2873,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(2873,36,'cp 109','e4e5 f6d5 b1d2 f8e8 d2e4 c5e7 c1d2 d7d6 f2f4 a8b8 b2b3 d5b4 d2b4 b8b4 d1f3 c8f5 e4f2 f5d3 f2d3 b4b6 a1e1 e7f8 e1d1 d6e5 d3e5 d8c8 e5d7 b6a6 d7f8 g8f8 f3d3 f8g8 a2a4 h7h6 f1e1 a6b6 e1e8 c8e8 d3d8 g8f8 d1d7 e8d8');
INSERT INTO evaluations VALUES(3538,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(3538,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(3538,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(3538,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(3538,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(3538,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(3538,36,'cp 41','a1b1 a7a5 b2b3 f6h5 e2e4 f7f5 e4f5 c8f5 h2h3 h5f6 f3g5 h7h6 g5e4 c6d4 e4f6 f8f6 c1e3 c7c6 c3e4 f6f7 f1e1 d8d7 e3d4 e5d4 c4c5 d6c5 e4c5 d7d6 c5e4 d6a3 b1c1 a8f8 e4c5 b7b6 g3g4 f5c8 c5e6');
INSERT INTO evaluations VALUES(8284,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(8284,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(8284,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8284,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(8284,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(8284,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(8284,36,'cp 0','c3d5 e7e5 c2c4 h7h6 a2a3 a7a5 c1d2 e8g8 b2b4 a5b4 a3b4 a8a1 d5f6 d8f6 d1a1 c5b4 a1a4 f6e7 d2b4 c8d7 f1a1 f7f5 e2d1 f5e4 d3e4 g8h7 h2h3 c6b4 a4b4 d7c6 d1c2 g7f6 a1d1 f8d8 d1a1');
INSERT INTO evaluations VALUES(8406,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(8406,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(8406,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(8406,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(8406,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(8406,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(8406,36,'cp 53','f4d6 d8d6 f1d3 e8g8 e1g1 e6e5 f3e5 c6e5 d4e5 d6e5 f1e1 e5c7 d1a4 c8d7 a4d4 a7a6 d2f1 f8e8 e1e3 a8c8 a1e1 c7c5 d4c5 c8c5 e3e8 d7e8 f1d2 c5c6 d2b3 c6b6 f2f3 g8f8 g1f2 e8b5 d3b5 b6b5 e1b1 f6e8 b3d4 b5b6 f2e3 e8d6');
INSERT INTO evaluations VALUES(9543,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(9543,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(9543,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(9543,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(9543,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(9543,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(9543,36,'cp 37','b1c3 e8g8 e1g1 a5c4 d3c4 c8f5 g5f3 f8e8 h2h3 h7h6 f1e1 f6d7 a2a3 a7a5 c1e3 c5f8 f3d2 f5g6 g1h1 a5a4 f2f3 f7f5 d1e2 d8f6 c3b5 a8c8 e3g1 f6f7 b5c3 c8a8 e2d3 d7c5 d3f1 c7c6 a1d1 c5d7 f1d3 d7c5 d3e2 c5d7 e2f1 d7f6 d5c6 b7c6 c4c5 g6h5 d2c4 e5e4');
INSERT INTO evaluations VALUES(10570,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(10570,31,'cp 239',NULL);
INSERT INTO evaluations VALUES(10570,32,'cp 248',NULL);
INSERT INTO evaluations VALUES(10570,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(10570,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(10570,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(10570,36,'cp 262','h2h3 g4f6 f3e5 c6d4 e5d3 d4b5 c3b5 c5e7 e4e5 f6e4 d1e2 d7d5 d3f4 e8g8 c2c4 c7c6 c4d5 c6d5 b5c3 c8e6 f1d1 e7c5 d2d4 e4c3 b2c3 c5b6 c1a3 f8e8 f4e6 e8e6 d1d3 e6c6 a1b1 d8h4 b1b5 h4e4 e2e4 d5e4 d3g3');
INSERT INTO evaluations VALUES(12364,30,'cp -139',NULL);
INSERT INTO evaluations VALUES(12364,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(12364,32,'cp -126',NULL);
INSERT INTO evaluations VALUES(12364,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(12364,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(12364,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(12364,36,'cp -124','d7d5 c4b5 c8d7 f3e2 d5e4 d3e4 a7a6 b5d3 c6e5 b1d2 e5d3 e2d3 f8e7 g1f3 d7c6 d3e2 d8c7 e1g1 e8g8 f1e1 f6d7 b2b3 f8d8 h2h4 h7h6 e4e5 b7b5 c3c4 d7f8 h4h5 c7d7 d2e4 d7d3 c1f4 d3e2 e1e2 b5c4 b3c4 a8b8 e4d6 e7d6 e5d6 f7f6 f4e3 f8d7 a1d1 b8b4 e3d2 b4c4 d2a5');
INSERT INTO evaluations VALUES(13530,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(13530,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(13530,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(13530,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(13530,34,'cp -113',NULL);
INSERT INTO evaluations VALUES(13530,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(13530,36,'cp -126','c4b5 c8d7 f3e2 d5e4 d3e4 a7a6 b5c6 d7c6 b1d2 f8e7 g1f3 d8c7 e1g1 e8g8 f1e1 f8d8 e4e5 f6d7 h2h4 b7b5 b2b3 h7h6 c3c4 a8b8 h4h5 a6a5 c4b5 b8b5 d2c4 a5a4 e2c2 c6f3 g2f3 a4b3 a2b3 d8b8 a1b1 d7b6 c4b6 c7b6');
INSERT INTO evaluations VALUES(18785,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(18785,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(18785,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(18785,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(18785,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(18785,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(18785,36,'cp 15','c5d4 f3d4 g8f6 d4c6 b7c6 e2e4 f8b4 f1d3 e6e5 e1g1 e8g8 c1e3 b4c3 b2c3 d7d6 c4c5 d6d5 e4d5 f6d5 f1e1 d5e3 e1e3 d8e7 d3c2 e7c5 d1h5 g7g6 h5e5 c5e5 e3e5 c8e6 c2b3 e6d5 b3d5 c6d5 e5d5 f8c8 a1c1 c8c4 d5d4 a8c8 d4d3 c8c5 g2g3 a7a5 h2h4 h7h6');
INSERT INTO evaluations VALUES(18900,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(18900,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(18900,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(18900,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18900,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(18900,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(18900,36,'cp 5','f3d4 g8f6 a2a3 f8c5 d4b3 c5b6 c4c5 b6c7 g2g3 e8g8 f1g2 b7b6 e1g1 h7h6 c3b5 c7e5 f2f4 e5b8 b5d4 d8e8 d4c6 d7c6 b3d4 c8d7 d1c2 b8c7 d4f3 e6e5 f3e5 c7e5 f4e5 e8e5 c1f4 e5c5 c2c5 b6c5 f4h6 f8e8 h6d2 e8e2 d2c3 f6d5');
INSERT INTO evaluations VALUES(18893,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(18893,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(18893,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(18893,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18893,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18893,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18893,36,'cp 5','g8f6 d4c6 d7c6 d1d8 e8d8 c1f4 f6d7 e1c1 f7f6 e2e4 e6e5 f4e3 d8c7 h2h4 b7b6 c3a4 d7c5 a4c5 f8c5 e3c5 b6c5 f1e2 c8e6 c1c2 a7a5 c2c3 a8d8 d1d8 h8d8 h1g1 e6f7 g2g4 g7g5 h4g5 f6g5 g1h1 f7g6 f2f3 a5a4 h1h6');
INSERT INTO evaluations VALUES(18985,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(18985,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(18985,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(18985,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(18985,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(18985,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(18985,36,'cp 31','g2g3 g7g6 g1e2 f8g7 f1g2 g8e7 a2a3 d7d6 b2b4 c5b4 a3b4 c6b4 e1g1 a7a5 c1a3 e7c6 d2d4 e5d4 e2d4 c8d7 c3e4 e8g8 e4d6 a8b8 c4c5 b7b6 a1c1 b6c5 d4c6 d7c6 g2c6 b4c6 c1c5 c6b4 d6c4 d8e7 d1d6 e7d6 c4d6 f8d8 d6c4 b8c8 c5c8 d8c8 c4a5 b4d3 f1b1 d3e5');
INSERT INTO evaluations VALUES(22027,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(22027,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(22027,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(22027,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(22027,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(22027,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(22027,36,'cp 44','g8f6 b1c3 f8e7 d4c6 b7c6 f1e2 e8g8 e1g1 d6d5 e4e5 f6d7 f2f4 d7b6 f4f5 b6d7 f5f6 g7f6 e5f6 e7f6 e2d3 d7e5 d1h5 e5d3 c2d3 d8d6 f1f3 d6e5 h5h6 f6g7 h6c6 a8b8 f3g3 d5d4 e3h6 b8b6 c6d5 e5d5 g3g7 g8h8 c3d5 b6h6 g7g5 h6c6 d5f4 f8e8 a1f1 c6c2 h2h4 c2b2');
INSERT INTO evaluations VALUES(15642,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(15642,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(15642,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(15642,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(15642,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(15642,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(15642,36,'cp 63','d4c6 b7c6 b1c3 f8e7 f1e2 a8b8 b2b3 a7a5 d1d2 e8g8 e1g1 d6d5 e2f3 c8g4 f3g4 f6g4 e3f4 g4f6 e4d5 e7b4 d2d3 b4c3 d3c3 f6d5 c3g3 d5f4 g3f4 d8d6 f4e3 f8d8 h2h3 a5a4 b3a4 d6b4 a1d1 d8d5 e3a7 b4b6 a7b6 c7b6 d1d5');
INSERT INTO evaluations VALUES(24757,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24757,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24757,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24757,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(24757,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(24757,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24757,36,'cp 0','c1e3 g4f3 d1f3 c6d4 f3e4 c7c5 b1c3 g7g6 e1c1 f8g7 e3d4 c5d4 c4c5 b6d7 f1b5 d8c7 c3d5 c7e5 e4e5 g7e5 h1e1 f7f6 e1e5 f6e5 d5c7 e8d8 c7a8 d7c5 d1e1 e5e4 b2b4 c5d3 b5d3 e4d3 c1d2 d8d7 a8c7 d7c7 e1e7 c7c6 d2d3 h8f8 e7h7 f8f2 g2g3 f2a2 h7g7 a2h2 g7g6 c6b5 g6g7 b5c6');
INSERT INTO evaluations VALUES(24756,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24756,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24756,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24756,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(24756,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(24756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24756,36,'cp 0','g4f3 d1f3 c6d4 f3e4 c7c5 b1c3 g7g6 e1c1 f8g7 e3d4 c5d4 c4c5 b6d7 f1b5 d8c7 c3d5 c7e5 e4e5 g7e5 h1e1 f7f6 e1e5 f6e5 d5c7 e8d8 c7a8 d7c5 d1e1 e5e4 b2b4 c5d3 b5d3 e4d3 c1d2 d8d7 a8c7 d7c7 e1e7 c7c6 d2d3 h8f8 e7h7 f8f2 d3d4 f2g2 a2a4 c6b6 d4c4 a7a5 b4a5 b6a6 c4b3 g2g1 b3c4');
INSERT INTO evaluations VALUES(26635,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26635,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26635,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(26635,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26635,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26635,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26635,36,'cp 0','g4f3 e2f3 e7e5 e1g1 g8f6 c2c4 e5e4 c4d5 e4f3 d5c6 f3g2 f1f2 b7c6 f2g2 h7h5 b2b3 h5h4 g2g5 d8d7 g5e5 f8e7 c1a3 f6d5 b1c3 e8c8 a3e7 d5c3 d1f3 d8e8 d2c3 e8e7 a1d1 d7e8 e5e7 e8e7 f3c6 e7e3 g1h1 c8b8 c6b5 b8a8 b5c6');
INSERT INTO evaluations VALUES(29828,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(29828,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(29828,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(29828,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(29828,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(29828,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(29828,36,'cp -20','f3g3 b8c6 h2h3 g4d7 g1e2 f8g7 a2a3 e8g8 g3h4 d7e6 c4e6 f7e6 e1g1 a7a5 c1e3 d8e8 b1d2 f6h5 d2f3 h7h6 a1e1 a5a4 h4g4 h5f6 g4g3 e8f7 g1h1 b7b5 f3h2 g8h7 c2c3 a8e8 h2f3 f6h5 g3h4 f7d7 h4g4 d7e7 e1d1 d6d5 f3d2 e7f7');
INSERT INTO evaluations VALUES(29829,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(29829,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(29829,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(29829,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(29829,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(29829,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(29829,36,'cp -35','b8c6 g1f3 f8g7 c1g5 d8d7 h2h3 g4e6 g5f6 g7f6 b1c3 e8c8 e1c1 c8b8 c1b1 d8f8 f3g5 a7a6 g5e6 f7e6 c3e2 d6d5 c4b3 g6g5 e2c3 d5d4 c3e2 h7h5 e2c1 f8g8 f2f3 d7e7 g3e1 g8g7 a2a3 h5h4');
INSERT INTO evaluations VALUES(30265,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(30265,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(30265,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(30265,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(30265,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(30265,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(30265,36,'cp 68','h2h3 g4h5 e1g1 f8e7 c2c4 b8c6 c4d5 d8d5 f1e1 h5f3 d1f3 e4f6 f3d5 f6d5 b1c3 a8d8 d3e4 d5f6 d4d5 c6b4 c1g5 h7h6 g5f6 e7f6 d5d6 d8d6 c3b5 d6b6 e4b7 e8f8 b5c7 g7g6 b7f3 b6b8 a1d1 b8d8 a2a3 b4c2 d1d8 f6d8 e1e8 f8g7 e8h8 g7h8 c7d5 c2d4 f3e4 d4b5 e4d3 b5c7 d5e3 c7e6');
INSERT INTO evaluations VALUES(30263,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(30263,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(30263,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(30263,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(30263,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(30263,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(30263,36,'cp 64','g4f5 e1g1 f8e7 c2c4 d5c4 d3c4 e8g8 b1c3 b8c6 f1e1 e4c3 b2c3 h7h6 c1f4 e7d6 f3e5 c6e5 f4e5 d6e5 e1e5 d8f6 d1e2 f5d7 c4d3 g7g6 e2e3 a8e8 f2f4 c7c5 d4d5 b7b5 a1e1 c5c4 d3b1 f6d6 e3a7 e8e5 f4e5 d6d5 b1e4');
INSERT INTO evaluations VALUES(31326,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(31326,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(31326,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(31326,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(31326,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(31326,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(31326,36,'cp -89','e7e6 d2d4 b8c6 g1e2 g8f6 c2c3 f5e4 f3h3 f8e7 e2g3 e4g6 f2f4 a7a5 f4f5 e6f5 g3f5 e8g8 e1g1 a5a4 b3c2 f8e8 b1d2 a4a3 b2b4 d8c8 d2b3 b7b6 c1g5 f6e4 c2e4 d5e4 g5e7 c6e7 f5d6 c8h3 g2h3');
INSERT INTO evaluations VALUES(31705,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(31705,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(31705,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(31705,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(31705,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(31705,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(31705,36,'cp -78','g1e2 b8c6 d2d4 h7h6 c2c3 f5e4 f3h3 g8f6 e1g1 f8e7 b1d2 e4f5 h3f3 e8g8 e2g3 f5g6 f3e2 d8c7 d2f3 a7a6 f3e5 c6e5 d4e5 f6e4 c1f4 b7b5 a1c1 a8c8 b3c2 e4g3 h2g3 g6c2 e2c2 c7c4 b2b3 c4e4 c2e4 d5e4 f1e1');
INSERT INTO evaluations VALUES(33679,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(33679,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(33679,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(33679,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(33679,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(33679,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(33679,36,'cp 56','d5e4 g1e2 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5d4 d1d4 e8g8 f2f3 e4e7 d4e5 c8e6 e2d4 b8d7 e5e3 f8e8 e1c1 e7c5 d4c2 e6f5 e3c5 d7c5 d1d2 a8d8 d2d8 e8d8 f1e2 d8e8 c2d4 f5d3 e2d3 c5d3 c1d2 d3f4 g2g3 f4e6 h1d1 e6d4 c3d4 b7b6 d4f6 g7f6 d1e1 e8e6 e1e6 f7e6 d2e3 g8f7 b2b4 e6e5 e3e4 f7e6 g3g4 c7c6');
INSERT INTO evaluations VALUES(33709,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(33709,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(33709,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(33709,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(33709,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(33709,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(33709,36,'cp 62','g1e2 e7e5 b1c3 f8b4 c1d2 b4c3 d2c3 e5d4 d1d4 e8g8 f2f3 e4e7 d4e5 c8e6 e2d4 b8d7 e5e3 f8e8 e1c1 e7c5 d4c2 e6f5 e3c5 d7c5 d1d2 a8d8 d2d8 e8d8 f1e2 d8e8 c2d4 f5d3 e2d3 c5d3 c1d2 d3f4 g2g3 f4e6 h1d1 e6d4 c3d4 b7b6 d4f6 g7f6 d1e1 e8e6 e1e6 f7e6 d2d3 c7c5 d3e4 g8g7 e4f4 g7g6 b2b3 g6g7 f4g4 g7g8');
INSERT INTO evaluations VALUES(36765,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(36765,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36765,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36765,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(36765,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36765,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(36765,36,'cp 13','c2c4 b7b6 f1e2 c8b7 e1g1 b8d7 b1c3 c7c5 c3d5 f6d5 c4d5 b7d5 a1c1 a8c8 d1d2 e7e6 f1d1 d8e7 b2a3 f8d8 h2h3 e7e8 d2e1 c5d4 f3d4 d7c5 a3c5 c8c5 c1c5 b6c5 d4f3 d5f3 e2f3 g7f6 f3e2 d8d1 e1d1 e8d8 d1c2 d8d6 g2g3 g8g7 h3h4 d6c7 e2c4 c7d6');
INSERT INTO evaluations VALUES(36725,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(36725,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(36725,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(36725,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(36725,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(36725,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(36725,36,'cp 20','b7b6 c4d5 f6d5 f1c4 c8b7 e1g1 c7c5 d1e2 b8d7 f1d1 e7e6 c4d5 b7d5 e3e4 d5b7 d4c5 g7b2 e2b2 d8f6 b2f6 d7f6 b1d2 f6e4 d2e4 b7e4 f3e5 f8d8 d1d8 a8d8 c5b6 a7b6 a2a4 e4d5 b3b4 d5b7 a4a5 b6a5 b4a5 b7a6 h2h4 d8d4 h4h5 g8g7 a1c1 d4e4 h5g6');
INSERT INTO evaluations VALUES(37752,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(37752,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(37752,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(37752,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(37752,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(37752,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(37752,36,'cp -36','c5d4 g1f3 d7d6 c3d4 d6e5 f3e5 g8h6 f1b5 c8d7 b5d7 b8d7 c1h6 g7h6 d1a4 a7a6 e1g1 h6g7 f1d1 b7b5 a4b3 d7e5 d4e5 d8b6 b1c3 e8g8 c3d5 b6e6 d5c7 e6b3 a2b3 a8a7 c7b5 a7b7 b5a3 g7e5 a3c4 e5f6 a1a6 b7b3 a6b6 b3b6 c4b6 f8b8 b6c4 f6b2 d1d7');
INSERT INTO evaluations VALUES(37846,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(37846,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(37846,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(37846,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(37846,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(37846,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(37846,36,'cp -22','c3d4 d7d6 g1f3 d6e5 f3e5 g8h6 c1h6 g7h6 f1b5 c8d7 d1a4 a7a6 b5d7 b8d7 e1g1 h6g7 f1d1 b7b5 a4b3 d7e5 d4e5 d8b6 b1c3 e8g8 c3d5 b6e6 d5c7 e6b3 a2b3 a8a7 c7b5 a7b7 b5a3 g7e5 a3c4 e5f6 d1d3 f8b8 a1a6 b7b3 d3b3 b8b3 a6a8 g8g7 a8a4 h7h5 g1f1 f6b2 c4b2 b3b2 g2g3 g7f6 f1g2 f6e6 a4a6 e6f5 h2h3');
INSERT INTO evaluations VALUES(37844,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(37844,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(37844,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(37844,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(37844,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(37844,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(37844,36,'cp -30','d7d6 g1f3 d6e5 f3e5 g8h6 c1h6 g7h6 f1b5 c8d7 b5d7 b8d7 d1a4 a7a6 e1g1 h6g7 f1d1 b7b5 a4b3 d7e5 d4e5 d8b6 b1c3 e8g8 c3d5 b6e6 d5c7 e6b3 a2b3 a8a7 c7b5 a7b7 b5a3 g7e5 a3c4 e5f6 d1d3 b7a7 a1a4 f8b8 g1f1 h7h5 g2g3 h5h4 f1e1 g8g7 e1d1 a7b7 d1e2 b7b3 d3b3 b8b3');
INSERT INTO evaluations VALUES(42401,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(42401,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(42401,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(42401,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(42401,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(42401,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(42401,36,'cp 41','f6e4 d2d4 d6e7 f1d3 d7d5 e1g1 c8f5 f1e1 e8g8 c2c4 b8c6 c4d5 d8d5 b1c3 e4c3 b2c3 d5d7 c1f4 e7d6 f4d6 c7d6 d4d5 c6a5 f3d4 f5d3 d1d3 f8e8 h2h4 e8e1 a1e1 a8e8 e1e8 d7e8 g2g3 b7b6 h4h5 h7h6 g1g2 a5b7 d3f3 b7c5 c3c4 g8h8 d4f5 e8e5 f3g4 c5d3');
INSERT INTO evaluations VALUES(42406,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(42406,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(42406,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(42406,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(42406,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(42406,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(42406,36,'cp 44','d2d4 d6e7 f1d3 d7d5 e1g1 b8c6 c2c4 c6b4 d3e2 e8g8 b1c3 c8f5 a2a3 e4c3 b2c3 b4c6 c4d5 d8d5 c1f4 f8e8 f1e1 a8c8 c3c4 d5e4 f4e3 e7f6 a1a2 c6a5 h2h3 c8d8 a2d2 b7b6 e2f1 e4c6 c4c5 b6c5 d4c5');
INSERT INTO evaluations VALUES(43028,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43028,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(43028,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(43028,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(43028,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(43028,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43028,36,'cp 7','e8g8 f2f4 e5f4 e4e5 f6e8 c1f4 c7c6 g1f3 d7d5 e5d6 e8d6 c4b3 b8d7 d3d4 e7h4 f3h4 d8h4 f4g3 h4f6 h1f1 d6f5 d1e2 f5d4 f1f6 d4e2 f6f7 f8f7 e1e2 d7f6 b3f7 g8f7 e2d2 c8f5 a1f1 f7g6 f1e1 g6f7 e1e5 f5g6 d2c1 a8e8 e5e8 f6e8 g3f2 a7a6');
INSERT INTO evaluations VALUES(37322,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37322,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37322,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37322,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37322,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37322,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37322,36,'cp 8','f2f4 e5f4 e4e5 f6e8 c1f4 c7c6 g1f3 d7d5 e5d6 e8d6 c4b3 b8d7 d3d4 e7h4 f3h4 d8h4 f4g3 h4f6 d1d3 d6f5 e1c1 f6g5 d1d2 f5g3 d3g3 g5g3 h2g3 f8e8 d4d5 d7f6 d5c6 b7c6 b3a4 c8b7 h1h4 a8c8 h4b4 e8e7 g3g4 c6c5 b4f4 g8f8 b2b3 e7e3 c1b2 e3e5');
INSERT INTO evaluations VALUES(43550,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(43550,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(43550,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(43550,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(43550,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43550,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(43550,36,'cp 54','d1f3 e6e5 f4e5 d6e5 d4f5 b8c6 f1d3 c8f5 f3f5 d8c8 f5f3 c6b4 c1g5 b4d3 c2d3 f8e7 g5f6 e7f6 e1c1 c8e6 c1b1 a8c8 c3d5 c8c6 d1c1 f6d8 c1c6 b7c6 d5e3 e8g8 e3c4 a6a5 h2h4 d8c7 f3f2 a5a4 f2a7 f8c8 g2g3 h7h6 h1f1 a4a3');
INSERT INTO evaluations VALUES(48276,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(48276,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(48276,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(48276,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(48276,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(48276,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(48276,36,'cp -85','f6e4 g1f3 f8d6 f1d3 e4c3 b2c3 d5c4 d3c4 b8d7 e1g1 e8g8 f1e1 b7b6 c1g5 d6e7 g5e7 d8e7 d4d5 d7c5 e1e3 c8b7 f3d4 e7g5 e3g3 g5f4 d5e6 f7e6 d1e2 b7d5 c4d5 e6d5 g3f3 f4d6 f3e3 d6f6 a1e1 c5e4 d4b5 c7c6 f2f3 c6b5 f3e4 d5e4 e3e4 f6c6');
INSERT INTO evaluations VALUES(48462,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(48462,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(48462,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48462,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48462,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48462,36,'cp 0','c7c5 c1f4 a7a6 a2a4 f8e7 e1g1 b8c6 c3e4 e8g8 h2h3 b7b6 c2c3 c8b7 f1e1 d8d7 f4g3 f6d5 e4d2 d5f6');
INSERT INTO evaluations VALUES(51011,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(51011,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(51011,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(51011,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(51011,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(51011,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(51011,36,'cp 88','g1f3 b7b5 c1e3 d7d6 f1d3 c7c5 e1g1 c5d4 f3d4 g8f6 a2a4 b5b4 c3a2 e6e5 d4f3 b8c6 a4a5 f8e7 c2c3 b4c3 a2c3 e8g8 d1a4 d8e8 e3b6 c8e6 f1d1 e6d7 a4b3 e7d8 b3a3 d7g4 d3e2 a8b8 b6d8 b8d8 d1d6 g4f3 d6d8 e8d8 e2f3 c6d4 c3d5 f6d5 e4d5');
INSERT INTO evaluations VALUES(51730,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(51730,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(51730,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(51730,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(51730,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(51730,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(51730,36,'cp 87','d2d4 a7a6 c4b3 b7b5 d4c5 f8c5 g1f3 c8b7 c1g5 c5e7 h2h4 d7d6 d1e2 h7h6 g5f4 b8d7 e1c1 d8b6 h1e1 a8c8 e2d2 g8f6 e4e5 d6e5 f3e5 d7e5 f4e5 c8d8 d2f4 f6h5 d1d8 e7d8 f4d2 d8f6 e5f6 h5f6 d2f4 e8e7 f2f3 b6d6 f4e3 f6d7 e1d1 d6c7');
INSERT INTO evaluations VALUES(52088,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52088,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52088,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(52088,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(52088,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(52088,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(52088,36,'cp 0','a7a5 c2c3 a5b4 c3b4 b8c6 c1d2 g8f6 e2e3 d7d6 b4b5 c6a5 b1c3 d6c5 g1f3 e5e4 f3g5 c8f5 d1c2 c5c4 a1d1 f8a3 d2c1 d8e7 c1a3 e7a3 f1e2 e8g8 e1g1 a8e8 d1d4 h7h6 g5h3 g7g5 c2d2 e8e5 c3b1 a3a4 b1c3');
INSERT INTO evaluations VALUES(54335,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(54335,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54335,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(54335,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(54335,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(54335,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54335,36,'cp 17','b8c6 c4d5 e6d5 f1e2 g8f6 e1g1 f8d6 b2b3 e8g8 c1b2 f8e8 b1d2 h7h6 a2a3 c6d4 b2d4 f6e4 d4b2 d8e7 d2e4 d5e4 d1d4 d6e5 d4e5 e7e5 b2e5 e8e5 f1d1 c8e6 b3b4 a7a5 h2h3 e5d5 d1d5 e6d5 a1d1 d5b3 d1d6 a5b4 a3b4 a8a1 g1h2');
INSERT INTO evaluations VALUES(55201,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(55201,31,'cp -148',NULL);
INSERT INTO evaluations VALUES(55201,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(55201,33,'cp -159',NULL);
INSERT INTO evaluations VALUES(55201,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(55201,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(55201,36,'cp -145','e7e5 g1e2 b8a6 a2a3 e5d4 e2d4 f7f5 b1c3 g8f6 g2g3 f8d6 d4b5 d6e5 d1d8 e8d8 c1g5 d8e7 e1c1 c7c6 f2f4 e5c3 b5c3 c8e6 f1e2 e7f7 d1d6 a8d8 d6d8 h8d8 h1d1 d8a8 g5h4 h7h6 c1d2 a6c7 d2e3 c7e8 b2b3');
INSERT INTO evaluations VALUES(55995,30,'cp -224',NULL);
INSERT INTO evaluations VALUES(55995,31,'cp -226',NULL);
INSERT INTO evaluations VALUES(55995,32,'cp -219',NULL);
INSERT INTO evaluations VALUES(55995,33,'cp -224',NULL);
INSERT INTO evaluations VALUES(55995,34,'cp -223',NULL);
INSERT INTO evaluations VALUES(55995,35,'cp -230',NULL);
INSERT INTO evaluations VALUES(55995,36,'cp -219','g8f6 b1c3 f8b4 c1d2 d8d4 c3b5 b4d2 d1d2 d4d2 e1d2 e4e3 d2e3 f6d5 e3d2 h7h6 g5f3 a7a6 b5c3 d5c3 d2c3 b7b6 f1d3 c8b7 f3e5 g7g5 a1e1 b7g2 h1g1 g2d5 h2h4 b8d7 h4g5 h6g5 g1g5 f7f6 e5g6 f6g5 g6h8 e8e7 h8g6 e7f6');
INSERT INTO evaluations VALUES(48458,30,'cp -218',NULL);
INSERT INTO evaluations VALUES(48458,31,'cp -233',NULL);
INSERT INTO evaluations VALUES(48458,32,'cp -222',NULL);
INSERT INTO evaluations VALUES(48458,33,'cp -212',NULL);
INSERT INTO evaluations VALUES(48458,34,'cp -212',NULL);
INSERT INTO evaluations VALUES(48458,35,'cp -213',NULL);
INSERT INTO evaluations VALUES(48458,36,'cp -221','b1c3 f8b4 c1d2 d8d4 c3b5 d4d2 d1d2 b4d2 e1d2 b8a6 a1e1 e8e7 g5e4 f6e4 e1e4 b7b6 e4c4 h8d8 d2c1 d8d7 f1e2 c8b7 a2a3 c7c5 c4h4 h7h6 h1e1 a6c7 b5c7 d7c7 e1d1 b7c6 g2g3 a7a5 f2f4 c7d7 d1d7 e7d7 h4h5 f7f6 h5h4 d7e7 c2c3 a8d8 b2b4 c5b4 a3b4');
INSERT INTO evaluations VALUES(48457,30,'cp -214',NULL);
INSERT INTO evaluations VALUES(48457,31,'cp -214',NULL);
INSERT INTO evaluations VALUES(48457,32,'cp -229',NULL);
INSERT INTO evaluations VALUES(48457,33,'cp -214',NULL);
INSERT INTO evaluations VALUES(48457,34,'cp -227',NULL);
INSERT INTO evaluations VALUES(48457,35,'cp -226',NULL);
INSERT INTO evaluations VALUES(48457,36,'cp -220','f8b4 c1d2 d8d4 c3b5 b4d2 d1d2 d4d2 e1d2 b8a6 a1e1 e8e7 g5e4 f6e4 e1e4 b7b6 e4c4 h8d8 d2c1 d8d7 f1e2 c8b7 c4a4 g7g6 h1e1 a6b8 g2g3 a7a6 b5c3 d7d6 e2d3 b8c6 d3e4 a8d8 e4f3 b6b5 a4f4 g6g5 f4g4 f7f6 c3e4 d6d4');
INSERT INTO evaluations VALUES(774,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(774,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(774,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(774,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(774,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(774,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(774,36,'cp 102','d4c5 b6c5 e1g1 e7g6 f1e1 c5b6 b1a3 f8c5 d1e2 c8d7 b2b4 c5e7 a1b1 e8g8 b4b5 c6d8 a3c2 b6c7 c2d4 a8c8 b1b3 f7f6 d3g6 h7g6 c1f4 c7c4 e2c2 f6f5 c2d2 d8f7 h2h4 e7c5 b3b2 c5e7 b2c2 a7a6 f4g5 f7g5 f3g5');
INSERT INTO evaluations VALUES(7381,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(7381,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(7381,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(7381,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(7381,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(7381,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(7381,36,'cp 70','b5c6 e7c6 d2d4 c5d6 b1d2 b7b5 f1e1 a8b8 a2a3 e8g8 b2b4 f7f5 e4f5 f8f5 d2e4 e5d4 c3d4 c8b7 e4g3 d6g3 h2g3 c6e7 c1g5 b7f3 d1d2 f5f7 e1e7 f7e7 g2f3 e7e2 d2e2 d8g5 e2a2 g8f8 a1d1 g5g6 g1g2');
INSERT INTO evaluations VALUES(10157,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(10157,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(10157,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(10157,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(10157,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(10157,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10157,36,'cp -19','d7d6 e1g1 e8g8 d2d3 c8d7 c1g5 h7h6 g5e3 f8e8 d1d2 a7a6 e2d1 d7g4 f3e1 g4d1 a1d1 c6d4 e3d4 c5d4 e1c2 b7b5 b2b3 d4c5 c2e3 c7c6 h2h3 c5e3 d2e3 d6d5 c4d5 c6d5 d3d4 f6e4 c3e4 e5d4 d1d4 e8e4 d4e4 d5e4 e3e4 d8e8 f1e1 e8e4 e1e4');
INSERT INTO evaluations VALUES(9375,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(9375,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(9375,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(9375,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(9375,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(9375,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(9375,36,'cp -7','e1g1 e8g8 d2d3 f8e8 c1e3 c8g4 f3d2 g4e2 d1e2 c6d4 e2d1 f6d7 d2b3 d4b3 d1b3 a8b8 b3c2 c7c6 c2d2 a7a6 b2b3 b7b5 f1d1 c5e3 d2e3 d7f8 c4b5 a6b5 d3d4 d8b6 c3e2 c6c5 d4e5 d6e5 e2c3 f8e6');
INSERT INTO evaluations VALUES(12719,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(12719,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(12719,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(12719,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(12719,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12719,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12719,36,'cp 55','f1g2 f8g7 d2d4 c5d4 f3d4 e8g8 e1g1 d8a5 e2e3 d7d6 c1d2 a5b6 f1e1 c8g4 d1b3 b6c7 h2h3 g4d7 a1c1 a8e8 d4c6 b7c6 b3c2 d7e6 c3e2 c7c8 g1h2 d6d5 c4d5 c6d5 d2c3 c8b8 e2f4 e6f5 c2a4 f5d7 a4b3 b8b3');
INSERT INTO evaluations VALUES(13558,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13558,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(13558,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13558,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(13558,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(13558,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(13558,36,'cp 7','g1f3 d5c4 f1c4 a7a6 e1g1 b7b5 c4a2 f8e7 d1c2 c5d4 e3d4 c8b7 f1d1 c6a5 c2e2 e8g8 f3e5 a8c8 h2h3 b7d5 c3d5 f6d5 c1d2 a5c6 e2e4 d8d6 a1c1 d5f6 e4e1 f6d5 c1c6 c8c6 e5c6 d6c6 d1c1 c6d7 a2d5 e6d5 d2b4 e7b4 e1b4');
INSERT INTO evaluations VALUES(16325,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(16325,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(16325,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(16325,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(16325,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(16325,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(16325,36,'cp -40','f6e4 e1g1 c6d4 d1d4 e4c5 f1e1 c5e6 b1d2 c7c6 b5d3 d7d5 d3f5 d8e7 d4d3 e7d6 d3c3 f8e7 f5e6 c8e6 c3g7 e8c8 d2f3 h8g8 g7h7 e7f6 h7d3 d5d4 g2g3 c6c5 d3e4 d4d3 c1f4 d6d5 c2d3 f6b2 a1b1 b2f6 f3e5 d5e4 e1e4 b7b6 a2a4 f6e5 e4e5 d8d3');
INSERT INTO evaluations VALUES(17104,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(17104,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(17104,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(17104,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(17104,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(17104,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(17104,36,'cp 114','d2d4 e5d4 e1g1 f8e7 f1e1 d7d5 c4b5 c8f5 f3d4 f5d7 b5c6 b7c6 d1e2 e8f8 b1d2 e7d6 d2f3 c6c5 d4b5 d7b5 e2b5 d8d7 b5d3 c5c4 d3d1 d7f5 b2b3 f8f7 b3c4 h8e8 c1e3 d5c4 d1e2 f5b5 f3d2 f7g8 a2a4 b5c6 e2c4 c6c4 d2c4 e8e4 c4d2 e4b4 g2g3 b4b2 e1c1 a7a5 e3d4 b2b4');
INSERT INTO evaluations VALUES(19842,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(19842,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(19842,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(19842,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(19842,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(19842,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(19842,36,'cp 17','g8f6 b1c3 c8f5 e2e3 e7e6 d1b3 f8b4 f1b5 b4c3 b3c3 e8g8 b5c6 a8c8 f3e5 f6g4 e5g4 f5g4 c3b4 c8c6 b4b7 d8c8 b7c8 f8c8 e1g1 a7a5 f2f3 g4f5 b2b3 h7h5 f1e1 f5d3 a1d1 d3g6 e3e4 f7f6 e4e5 g8f7 d1c1 c6c2 h2h4 c8c3 c1c2 c3c2 e5f6');
INSERT INTO evaluations VALUES(13141,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(13141,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(13141,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13141,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(13141,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13141,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(13141,36,'cp 20','b1c3 c8f5 e2e3 e7e6 d1b3 f8b4 f1b5 b4c3 b3c3 e8g8 b5c6 a8c8 f3e5 f6g4 e5g4 f5g4 c3b4 c8c6 b4b7 d8c8 b7c8 f8c8 e1g1 g4e2 f1e1 e2d3 f2f3 h7h5 e1d1 d3g6 d1d2 a7a5 h2h4 f7f6 b2b3 g8f7 a1e1 c6c2 d2c2 c8c2 e1a1 c2c3 a1d1 c3c2 a2a3 c2a2 d1c1 a2a3');
INSERT INTO evaluations VALUES(20313,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(20313,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(20313,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(20313,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(20313,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(20313,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(20313,36,'cp -112','d8b6 b1d2 h7h6 g5e3 c5d4 f3d4 c6e5 f1b5 c8d7 d1e2 a7a6 b5d7 e5d7 e1c1 g8f6 h1e1 f8c5 d2b3 e8g8 g2g4 e6e5 d4f5 c5e3 e2e3 b6e3 f2e3 g8h7 e1g1 g7g6 f5d6 b7b5 b3a5 d7b6 a5c6 b6c4 d6c4 d5c4 g4g5 h6g5 g1g5 a8e8 g5e5 h7g7');
INSERT INTO evaluations VALUES(1454,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(1454,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(1454,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(1454,33,'cp -103',NULL);
INSERT INTO evaluations VALUES(1454,34,'cp -120',NULL);
INSERT INTO evaluations VALUES(1454,35,'cp -131',NULL);
INSERT INTO evaluations VALUES(1454,36,'cp -107','b1d2 h7h6 g5e3 c5d4 f3d4 c6e5 f2f4 e5c6 d2b3 b6c7 d4b5 c7b8 d1f3 g8f6 e1c1 a7a6 b5c3 c8d7 c1b1 b8c7 g2g4 f8b4 e3d2 d5d4 c3e4 f6e4 f3e4 b4d2 d1d2 e8c8 f1g2 f7f5 g4f5 e6f5 e4f3 g7g5 h1d1 h8e8 b3d4 c6d4 d2d4 d7c6 d4d8 e8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(28308,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(28308,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(28308,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(28308,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(28308,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(28308,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(28308,36,'cp 26','e8g8 b1c3 f6e4 c1d2 d7d5 c4d5 e6d5 d2f4 e4c3 b2c3 b8c6 f1e1 c6a5 d1d3 f8e8 f3d2 d8d7 h2h4 e7d6 f4d6 d7d6 d2b3 d6a3 b3a5 a3a5 e2e4 d5e4 g2e4 b7e4 e1e4 a5d5 a1e1 d5d7 e4e8 a8e8 e1e8 d7e8 d4d5 g7g6 c3c4 e8e1 g1g2');
INSERT INTO evaluations VALUES(27786,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(27786,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(27786,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(27786,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(27786,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(27786,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(27786,36,'cp 37','b1c3 d7d5 c4d5 f6d5 d1c2 c7c5 f1d1 b8d7 c3d5 e6d5 b2b3 e7f6 c1b2 f8e8 d4c5 d7c5 e2e3 a8c8 b2d4 f6d4 f3d4 c5e4 c2b2 d8f6 a1c1 h7h5 c1c8 e8c8 h2h4 c8c7 a2a3 g7g6 g2e4 d5e4 b3b4 a7a5 d1c1 c7c1 b2c1 a5b4');
INSERT INTO evaluations VALUES(27785,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(27785,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(27785,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(27785,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(27785,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27785,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(27785,36,'cp 28','d7d5 f3e5 b8d7 c1f4 c7c5 a1c1 a8c8 c4d5 f6d5 c3d5 b7d5 e2e4 d5a8 d4d5 d7e5 f4e5 e7f6 e5c3 e6d5 e4d5 d8d6 d1f3 c8d8 f1d1 f6c3 b2c3 a8b7 c3c4 b7c8 d1e1 d8e8 h2h4 c8d7 f3a3 a7a5 a3b2 e8e1 c1e1 f8b8 e1c1 h7h6 g2f3 d7a4 b2d2 a4d7');
INSERT INTO evaluations VALUES(27760,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(27760,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(27760,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(27760,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(27760,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27760,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(27760,36,'cp 42','f3e5 b8d7 c1f4 c7c5 a1c1 a8c8 c4d5 f6d5 c3d5 b7d5 e2e4 d7e5 e4d5 e5g6 f4e3 c5d4 c1c8 d8c8 e3d4 e6d5 g2d5 e7f6 f1e1 f8d8 d4f6 g7f6 d1h5 d8e8 e1e8 c8e8 h2h4 g8g7 h5g4 e8e1 g1g2 e1e5 g4d7 g6e7 d5b7 a7a5 d7d8 e5c5 b7f3 e7c6 d8c7 c6e5 c7c5 b6c5 f3d5 g7g6');
INSERT INTO evaluations VALUES(27762,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(27762,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27762,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(27762,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(27762,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(27762,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(27762,36,'cp 16','b8d7 c1f4 c7c5 a1c1 a8c8 c4d5 f6d5 c3d5 b7d5 e2e4 d7e5 e4d5 e5g6 f4e3 c5d4 e3d4 c8c1 d1c1 e6d5 f1d1 e7f6 d4f6 d8f6 g2d5 h7h5 c1c6 f6c6 d5c6 f8c8 c6a4 g6f8 f2f4 c8c7 g1f2 g7g6 f2e3 g8g7 e3f3 c7c4 a4b3 c4c7 h2h4 f8d7 d1d6 d7c5 b3d5 c5a4 b2b3 a4c3');
INSERT INTO evaluations VALUES(29090,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(29090,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(29090,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(29090,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29090,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(29090,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(29090,36,'cp 1','e7e6 e1g1 f8e7 d3g6 h7g6 d1e2 e8g8 g3e4 f6e4 e2e4 f8e8 g2g3 d8d5 f1e1 d5e4 e1e4 b8d7 h2h4 a8c8 c1g5 d7f6 e4e2 c6c5 c2c3 c5d4 f3d4 e7f8 d4f3 f6d5 g5e3 a7a6 a1d1 f7f6 g1g2 e8d8 e2d2 d5e3 f2e3 d8d2 d1d2 f8e7');
INSERT INTO evaluations VALUES(30440,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(30440,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(30440,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(30440,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(30440,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(30440,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(30440,36,'cp 205','a4b5 d7b5 b1a3 b5d7 f1c4 e7e6 g1e2 f8b4 e2c3 b8c6 a3c2 a7a5 d4d5 c6e5 c4e2 b4d6 d5e6 f7e6 e1g1 g8e7 f2f4 e5g6 g2g3 e8g8 e2c4 e7c6 c1e3 d8e7 c3a4 a8d8 d1e2 g8h8 b2b3 d7c8 e2g2 e7d7 a1d1 d7e8 e4e5 d6e7 a4c5 d8d1');
INSERT INTO evaluations VALUES(30431,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(30431,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(30431,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(30431,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(30431,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(30431,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(30431,36,'cp 199','d7b5 b1a3 b5d7 f1c4 e7e6 g1e2 f8b4 e2c3 a7a5 a3c2 b8c6 d4d5 c6e5 c4e2 b4d6 d5e6 f7e6 f2f4 e5g6 e4e5 d6e7 e2c4 d8c8 f4f5 g6e5 d1h5 e8f8 f5e6 d7e6 e1g1 g8f6 c4d5 e5f7 h5f3 c7c6 d5e6 c8e6 c2d4 e6g4 f3f2 f8g8 d4c6 e7d6 c3b5 g4c4 b5d6 f7d6 c6a5');
INSERT INTO evaluations VALUES(30772,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(30772,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(30772,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(30772,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(30772,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(30772,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(30772,36,'cp -57','e7e6 h5e2 g8f6 d2d3 f8e7 c4b3 d7d5 e4e5 f6d7 c2c3 c7c5 f2f4 b6b5 g1f3 b8c6 e1g1 e8g8 b1d2 a7a5 b3c2 a5a4 a2a3 d5d4 d2e4 d4c3 b2c3 b5b4 a3b4 c5b4 c2a4 d7c5 e4c5 e7c5 d3d4 c5b6 a4c6 b7c6 a1a8 c6a8 c3b4 b6d4 f3d4 d8d4 c1e3');
INSERT INTO evaluations VALUES(30619,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(30619,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(30619,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(30619,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(30619,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(30619,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(30619,36,'cp -59','h5e2 g8f6 d2d3 f8e7 c4b3 d7d5 e4e5 f6d7 c2c3 c7c5 f2f4 b8c6 g1f3 b6b5 e1g1 e8g8 b1d2 a7a5 b3c2 d8c7 a2a3 a5a4 f1f2 b7a6 e2d1 f7f5 d3d4 d7b6 d2f1 c5d4 f3d4 c6d4 c3d4 a8c8 g2g4 b5b4 g4f5');
INSERT INTO evaluations VALUES(31740,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(31740,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(31740,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(31740,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(31740,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(31740,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(31740,36,'cp 141','c2c4 e7e6 b1c3 f8b4 e2e3 g8e7 f1e2 d8d7 e1g1 b8c6 c4d5 e6d5 a1c1 b4d6 f4d6 d7d6 f3e1 e8g8 e1d3 a7a6 d3c5 f5c8 a2a3 b7b6 c5d3 c8e6 b2b4 g8h8 h2h3 e6g8 d1b3 f8b8 b3b2 c6d8 b4b5 c7c6');
INSERT INTO evaluations VALUES(31991,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(31991,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(31991,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(31991,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(31991,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(31991,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(31991,36,'cp 172','e7e6 d1b3 d8c8 d4d5 e6e5 d5d6 f8d6 c4f7 e8e7 g1f3 g8f6 e1g1 b8c6 c1d2 a7a5 a2a4 h8f8 f7c4 f6e4 c4d5 e4f6 d5c6 b7c6 b1d1 c6c5 d2c3 c8f5 f3d2 f5e6 b3c2 e7f7 d2f3 f7g8 f3g5 e6g4 f2f4 g8h8 f4e5 g4g5 e5f6 g7f6');
INSERT INTO evaluations VALUES(32378,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(32378,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(32378,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(32378,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(32378,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(32378,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(32378,36,'cp 184','b2b4 c7c6 c1d2 f8d6 d1c2 b8d7 e3e4 b7b5 c4b3 a7a5 b4a5 d6c7 g1f3 c7a5 e1g1 a5d2 d4d5 c6d5 e4d5 e6e5 c2d2 g8f6 f1e1 e8g8 f3e5 d7e5 e1e5 d8d6 e5e1 a8c8 b1c1 h7h5 d2d4 c8c1 e1c1 f8e8 g2g3 e8e4 c1c8 e4e8 c8e8 f6e8 g1g2 g7g6 d4d3 d6c5 d5d6 c5c6 g2f1 c6d6 d3b5 e8c7');
INSERT INTO evaluations VALUES(34675,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(34675,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(34675,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(34675,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(34675,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(34675,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(34675,36,'cp 76','e7e5 d4e5 b8c6 e2e4 f8b4 c1d2 a5e5 g1f3 e5e7 d1c2 g8f6 e1c1 c8e6 f1b5 e8g8 a2a3 b4a5 b5c6 b7c6 f3d4 a5b6 d4e6 e7e6 f2f3 f6d7 c1b1 d7e5 c3e2 e5c4 d2c1 a8b8 d1d3 c4e5 d3c3 f8d8 c1f4 e6e7 f4e5 e7e5');
INSERT INTO evaluations VALUES(35141,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(35141,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(35141,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(35141,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(35141,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(35141,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(35141,36,'cp 90','c2c4 f8c5 e2e3 g8e7 b1c3 d7d6 g1e2 a7a5 e1g1 e8g8 d2d4 c5b6 b2b3 b8c6 c1b2 e5d4 e2d4 c6d4 e3d4 f8e8 c3a4 b6a7 f1e1 c8d7 d1d2 e7c6 c4c5 f6f5 a1d1 e8e1 d1e1 a8e8 d4d5 c6b4 e1e8 d7e8 d2e2 g8f8 g2e4 e8b5 e2b5 f5e4');
INSERT INTO evaluations VALUES(35112,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(35112,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(35112,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(35112,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(35112,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(35112,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(35112,36,'cp 91','f8c5 e2e3 d7d6 g1e2 g8e7 b1c3 a7a5 d2d4 c5b6 e1g1 e8g8 c3a4 b6a7 e2c3 b8d7 c3b5 a7b8 b2b3 f6g6 d4e5 d7e5 c1a3 c8g4 f2f3 g4d7 d1d2 e7f5 a1e1 c7c6 a4b6 c6b5 b6a8 b8a7 a8c7 b5c4 c7d5 c4b3 a2b3 d7b5 d2a5 b5f1 g1f1 e5c6 a5d2 f8d8 f3f4');
INSERT INTO evaluations VALUES(35234,30,'cp 246',NULL);
INSERT INTO evaluations VALUES(35234,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(35234,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(35234,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(35234,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(35234,35,'cp 260',NULL);
INSERT INTO evaluations VALUES(35234,36,'cp 249','e5f3 d7d5 d2d3 d5e4 d3e4 b8c6 b1c3 c8e6 f1e2 a8d8 c3d5 e7f7 c1f4 e6d5 e4d5 d8d5 f3d2 f8d6 c2c4 d6f4 c4d5 f7d5 d2f3 g8e7 e1g1 d5d1 f1d1 f4d6 h2h4 e8d8 g1f1 d8c8 g2g3 g7g6 a1b1 a7a6 a2a3 h7h5 b2b4 e7f5 b4b5 a6b5 e2b5 c6e5 f3e5 d6e5');
INSERT INTO evaluations VALUES(35237,30,'cp 221',NULL);
INSERT INTO evaluations VALUES(35237,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(35237,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(35237,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(35237,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(35237,35,'cp 230',NULL);
INSERT INTO evaluations VALUES(35237,36,'cp 225','d7d5 d2d3 d5e4 d3e4 e7e4 f1e2 b8c6 b1c3 f8b4 e1g1 b4c3 b2c3 e4a4 f3d4 c6d4 c3d4 g8e7 f1e1 e8g8 a1b1 a7a6 b1b3 e7g6 b3a3 a4c6 e2f3 c6c4 a3e3 c8d7 f3b7 a8b8 b7e4 f6f5 e4d3 c4a2 d1e2 d7b5 c2c4 a2e2 d3e2 b5d7 e3a3 f8e8 f2f3');
INSERT INTO evaluations VALUES(36896,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(36896,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(36896,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36896,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36896,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(36896,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36896,36,'cp 14','g5h4 c8e6 b1d2 b8d7 c2c3 c7c6 h4g3 d6g3 h2g3 f8e8 a2a4 e6g4 a4a5 a7a6 f1e1 d8c7 d2f1 e8e1 d1e1 g7g6 e1d2 g8g7 f3e1 c6c5 b2b4 c5b4 c3b4 d7f8 b4b5 a6b5 d3b5 g4d7 b5d7 c7d7 f1e3 f8e6');
INSERT INTO evaluations VALUES(36898,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(36898,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36898,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36898,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36898,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36898,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(36898,36,'cp 23','c8e6 f1e1 c7c6 b1c3 b8d7 f3e5 f8e8 e5d7 e6d7 e1e8 d7e8 d1f3 e8d7 h4f6 d8f6 f3f6 g7f6 c3e2 f6f5 f2f4 a7a5 g2g3 g8g7 g1f1 b7b6 a2a4 a8e8 e2g1 c6c5 g1f3 c5c4 d3e2 d7e6 f1f2 e8b8 f3e5 d6e5 f4e5 f7f6 e5f6 g7f6 f2f3 e6f7 a1e1 b6b5 a4b5 b8b5');
INSERT INTO evaluations VALUES(37366,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(37366,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(37366,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(37366,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37366,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(37366,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(37366,36,'cp 0','f1e2 h7h6 g5f4 d7d6 h2h3 b7b6 a2a4 b8c6 f4h2 c8b7 e1g1 f8e8 b1c3 a7a6 d1d2 e7e6 f1e1 c6e7 d2d3 e7c6');
INSERT INTO evaluations VALUES(42583,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(42583,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(42583,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(42583,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(42583,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(42583,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(42583,36,'cp -62','c7c6 d1c2 d7d5 g1f3 d8c7 c1g5 b8d7 e4d5 c6d5 c4d5 f6d5 b1c3 d5c3 b2c3 e8g8 f1e2 b7b6 e1g1 c8b7 a1d1 h7h6 g5h4 a8c8 h4g3 c5d6 f3h4 g7g6 c3c4 f8e8 e2f3 d6f8 f3b7 c7b7 h4f3 f8g7 d3d4');
INSERT INTO evaluations VALUES(43640,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(43640,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43640,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(43640,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(43640,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(43640,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(43640,36,'cp 9','c7c5 d4c5 f8c5 c4d5 f6d5 c3d5 e6d5 a2a3 c5e7 g1f3 e7f6 a1b1 b8c6 f1d3 d8a5 d1d2 c8e6 f4g5 a5d8 g5f6 d8f6 e1g1 e8g8 b2b4 f8d8 f1d1 h7h6 d2b2 d5d4 f3d4 c6d4 e3d4 d8d7 d3e4 e6d5 e4d5 d7d5 b1c1 a8d8 h2h3 d5d4 d1d4 d8d4 b2e2 f6f4');
INSERT INTO evaluations VALUES(44364,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(44364,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(44364,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(44364,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(44364,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(44364,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(44364,36,'cp 121','c1f4 f6h5 e2e3 f8e7 f1e2 d7d5 e1g1 b8d7 e2d3 e8g8 a2a3 g7g5 f3g5 h5f4 e3f4 e7g5 f4g5 d8g5 f1e1 d7f6 d1c1 g5c1 a1c1 g8f7 c3e2 c8d7 e2f4 a7a5 f2f3 b7b6 h2h3 a8e8 e1e5 f8g8 c1e1 h7h5 c4d5 f6d5 d3c4 h5h4 c4d5 c6d5');
INSERT INTO evaluations VALUES(49586,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(49586,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(49586,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(49586,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(49586,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(49586,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(49586,36,'cp 206','d4e5 f6e4 d1d5 e4c5 f1c4 c5e6 b1c3 c7c6 d5d2 d7d6 e5d6 f8d6 c3e4 d6e7 d2d8 e6d8 b2b3 b7b5 c4e2 b8a6 a2a3 f7f5 e4c3 a6c5 e2f3 c8b7 c1b2 e8g8 a3a4 e7f6 e1c1 b5b4 c3e2 f6b2 c1b2 c5e4 d1d7 e4f2');
INSERT INTO evaluations VALUES(50902,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(50902,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(50902,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50902,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(50902,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(50902,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(50902,36,'cp 49','c2c4 g8f6 c4d5 f8b4 b1d2 e6d5 f1d3 e8g8 e1g1 b4d6 f4d6 d8d6 a1c1 f8e8 f1e1 b8d7 d1c2 c7c6 h2h3 a6a5 a2a3 a5a4 f3h2 d7b6 h2f1 c8e6 f1g3 b6d7 c2c3 a8c8 d2f3 c6c5 d4c5 d7c5 c3b4 b7b6 f3d4 e6d7');
INSERT INTO evaluations VALUES(54839,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54839,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54839,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(54839,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(54839,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(54839,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54839,36,'cp 0','c7c5 b1c3 f8g7 f1b5 c8d7 e1g1 b8c6 d2d3 g8f6 d1e1 a8c8 b5c6 d7c6 a2a4 e8g8 e1h4 d8d7 c1d2 d7g4 h4g4 f6g4 h2h3 g4f6 c3b5 c8a8 b5c7 a8b8 c7b5');
INSERT INTO evaluations VALUES(56253,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(56253,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(56253,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(56253,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(56253,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(56253,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(56253,36,'cp -25','c1b2 g8f6 b1c3 f8c5 e2e3 d8e7 g1e2 c7c6 d2d3 c8f5 h2h3 e8g8 e1g1 h7h6 g3g4 f5g6 e2g3 f8e8 d3e4 d5e4 c3e2 b8d7 e2f4 g6h7 d1e2 d7e5 a1d1 a7a5 a2a4 a8d8 d1d8 e8d8 f1d1 d8d1 e2d1 e5f3 g2f3 e4f3 d1f3');
INSERT INTO evaluations VALUES(56254,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(56254,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(56254,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(56254,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(56254,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(56254,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(56254,36,'cp -41','g8f6 d2d3 f8b4 c2c3 b4d6 c3c4 d5c4 b3c4 e4d3 d1d3 d6b4 b1c3 e8g8 d3c2 b8c6 g1f3 c6d4 f3d4 d8d4 c2b3 f6e4 g2e4 d4e4 e1g1 b4c3 b3c3 f7f6 f2f3 e4e6 f1d1 f8e8 e2e4 b7b6 c4c5 c8a6 d1d4 a8d8 c3b3 d8d4 b3e6 e8e6 b2d4');
INSERT INTO evaluations VALUES(48709,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(48709,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(48709,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(48709,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(48709,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(48709,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(48709,36,'cp -19','e2e3 c8f5 g1e2 b8d7 b1c3 c7c6 d2d3 d8a5 e1g1 f8e7 d1d2 a5c7 e2d4 f5g6 c3e2 e8g8 e2f4 d7e5 f2f3 e5f3 d4f3 e4f3 f1f3 f6g4 a1f1 e7f6 b2f6 g4f6 d2c3 b7b6 h2h3 a8e8 g3g4 h7h5 g4g5 f6h7');
INSERT INTO evaluations VALUES(48708,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(48708,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(48708,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48708,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(48708,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(48708,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(48708,36,'cp -10','c8f5 g1e2 b8d7 b1c3 c7c6 d2d3 d8a5 a2a3 d7e5 e2d4 f5g6 d3e4 d5e4 d1d2 a5c7 e1c1 a7a5 f2f3 e5f3 d4f3 e4f3 g2f3 f8c5 c3a4 c5e7 d2d4 e8g8 d4f4 a8c8 f4c7 c8c7 a4b6 f8e8 b6c4 a5a4 b3b4 e7f8 b2e5 c7e7 e5f6 g7f6 d1d4 b7b5 c4d6');
INSERT INTO evaluations VALUES(57072,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(57072,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(57072,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(57072,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(57072,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(57072,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(57072,36,'cp -5','g8f6 f1g2 d7d5 g1f3 c7c5 e1g1 f8e7 b2b3 e8g8 d2d3 b7b5 c2c4 b5c4 d3c4 c8a6 c1b2 b8c6 f3e5 c6e5 b2e5 d8b6 b1c3 a8d8 c3a4 b6b4 e5c3 b4b8 c3a5 d8d7 d1e2 d5c4 b3c4 a6b7 e3e4 b8a8 a4c3 f8b8 f1d1 b7c6 d1d7 f6d7 a1d1 h7h6');
INSERT INTO evaluations VALUES(3186,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(3186,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(3186,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(3186,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(3186,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(3186,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(3186,36,'cp -21','a1b1 f6e8 a2a3 a8b8 c1d2 e8c7 b2b4 c7e6 d1c1 b7b6 d2h6 c8b7 e2d1 f7f6 c1d2 g7h6 d2h6 f8f7 f3h4 c6d4 d1g4 d8f8 h6f8 e6f8 f2f4 f6f5 e4f5 g6f5 g4h5 f7g7 b4c5 d6c5 c3b5 d4b5 c4b5 f8e6 h5f3 b7f3 f1f3');
INSERT INTO evaluations VALUES(4866,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4866,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(4866,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4866,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(4866,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(4866,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(4866,36,'cp 0','d2d3 g8f6 e1g1 e8g8 a2a4 d7d5 e4d5 f6d5 b1c3 d5c3 b2c3 c8g4 h2h3 g4h5 d1e2 c5d6 a1b1 d8d7 c4d5 c6a5 g2g4 h5g4 h3g4 d7g4 g1h1 g4h5 h1g1');
INSERT INTO evaluations VALUES(5043,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(5043,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(5043,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(5043,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(5043,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(5043,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(5043,36,'cp -9','a8b8 e3d4 g8f6 e4e5 f6d5 f1c4 d5c3 d4c3 d7d5 e5d6 g7c3 b2c3 e7d6 e1g1 e8g8 d1d4 c8f5 d4a7 f5c2 a7d4 d6d5 f1c1 c2e4 c4f1 d8a5 c3c4 a5a4 d4d1 b8a8 f2f3 a4d1 c1d1 e4c2 d1c1 c2a4 c4d5');
INSERT INTO evaluations VALUES(16283,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(16283,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(16283,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(16283,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(16283,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(16283,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(16283,36,'cp 14','f8d6 f4g5 d6e7 b1d2 d7d6 g1f3 e8g8 f1d3 e6e5 e1g1 h7h6 g5h4 f6d7 h4e7 d8e7 d1c2 d7f6 e3e4 e5d4 c3d4 c6b4 c2c3 b4d3 c3d3 f8e8 f1e1 c8e6 d4d5 e6d7 h2h3 c7c6 a2a4 c6d5 e4d5 e7d8 e1e8 d8e8 d2c4 d7a4 c4d6 e8d7 d6e4 f6e4 d3e4 a4b5 e4d4');
INSERT INTO evaluations VALUES(16964,30,'cp 540',NULL);
INSERT INTO evaluations VALUES(16964,31,'cp 548',NULL);
INSERT INTO evaluations VALUES(16964,32,'cp 546',NULL);
INSERT INTO evaluations VALUES(16964,33,'cp 540',NULL);
INSERT INTO evaluations VALUES(16964,34,'cp 548',NULL);
INSERT INTO evaluations VALUES(16964,35,'cp 563',NULL);
INSERT INTO evaluations VALUES(16964,36,'cp 561','f3e5 c7c6 b1d2 d7d5 e4d5 f8d6 e5c4 f6d5 c4d6 d8d6 f1e2 e8g8 d2f3 d6g6 e1g1 c8h3 f3h4 g6f6 e2g4 f8e8 g4h3 f6h4 d1f3 h4f6 f3f6 d5f6 c1e3 f6d5 e3d2 c6c5 a2a3 h7h6 f1e1 d5f6 f2f3 a8d8 e1e8 f6e8 d2e3 b7b6 g1f2 g7g5 h3g4 g8f8 a1e1 e8c7 h2h3 f8g7');
INSERT INTO evaluations VALUES(16941,30,'cp 548',NULL);
INSERT INTO evaluations VALUES(16941,31,'cp 537',NULL);
INSERT INTO evaluations VALUES(16941,32,'cp 542',NULL);
INSERT INTO evaluations VALUES(16941,33,'cp 540',NULL);
INSERT INTO evaluations VALUES(16941,34,'cp 548',NULL);
INSERT INTO evaluations VALUES(16941,35,'cp 541',NULL);
INSERT INTO evaluations VALUES(16941,36,'cp 521','d8e7 e5f3 f6e4 c1e3 b7b6 d1d5 c7c6 d5d4 e4c5 c2c4 c5e6 d4d2 c8b7 b1c3 g7g6 f1d3 f8g7 e1c1 e8g8 h1e1 b6b5 e3h6 g7h8 d3c2 f8d8 c4c5 e7c5 c3e4 c5d5 d2d5 c6d5 e4d6 b7c6 f3e5 d8f8 h6f8 a8f8 e5f7 f8f7 d6f7');
INSERT INTO evaluations VALUES(17385,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(17385,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(17385,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(17385,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(17385,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(17385,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(17385,36,'cp 191','d4d5 c6a7 f1d3 g8f6 c3c4 b7b5 b1c3 d7d6 e1g1 f8e7 h2h3 b5c4 d3c4 e8g8 a2a4 e6d5 e4d5 a6a5 c1f4 c8f5 f1e1 a7c8 d1e2 f8e8 f3d2 f5d7 b2b3 e7f8 e2f3 h7h6 e1e8 d8e8 c4d3 c8b6 d2c4 b6c4 b3c4 g7g6 c3b5 d7b5 a4b5 f6d7');
INSERT INTO evaluations VALUES(18647,30,'cp 374',NULL);
INSERT INTO evaluations VALUES(18647,31,'cp 375',NULL);
INSERT INTO evaluations VALUES(18647,32,'cp 379',NULL);
INSERT INTO evaluations VALUES(18647,33,'cp 376',NULL);
INSERT INTO evaluations VALUES(18647,34,'cp 345',NULL);
INSERT INTO evaluations VALUES(18647,35,'cp 401',NULL);
INSERT INTO evaluations VALUES(18647,36,'cp 398','d4d5 c6e7 d1d4 f7f6 b1c3 e6d5 c1f4 g6g5 f4g3 d8b6 c3b5 b6d4 f3d4 e7c6 d4c6 b7c6 b5c7 e8f7 c7a8 d5e4 a1c1 g8h6 f1c4 d7d5 c4e2 f7g6 a8c7 h6f5 c7a6 f5d4 a6c5 d4e2 e1e2 f8c5 c1c5 c8b7 h1d1 h7h5 e2e3 h5h4');
INSERT INTO evaluations VALUES(19470,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(19470,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(19470,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(19470,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19470,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(19470,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19470,36,'cp 9','g1f3 g8f6 b1c3 e7e5 f1b5 d7d6 e1g1 a7a6 b5c6 b7c6 d2d3 g7g6 c3a4 f8g7 c1e3 a8b8 c2c3 e8g8 f1e1 h7h6 b2b3 f8e8 d1c2 c8d7 a1d1 a6a5 a4b2 d8c7 b2a4');
INSERT INTO evaluations VALUES(21463,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(21463,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(21463,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(21463,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(21463,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(21463,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(21463,36,'cp 20','g8e7 c2c3 e7f5 e3f4 f7f6 e5f6 d8f6 f4c7 f6f7 c7g3 f5g3 h2g3 e6e5 f1b5 f8d6 d1f3 e5d4 f3f7 e8f7 g1f3 c8g4 f3d4 a7a6 b5a4 c6d4 c3d4 d6b4 b1c3 a8c8 f2f3 c8c3 b2c3 b4c3 e1e2 c3a1 h1a1 g4e6 a4c2 b7b5 a2a4 e6d7 a4a5 f7e7 e2d2 h7h6 g3g4 h8c8 c2d3 e7d6 a1b1 d7e6 b1e1 e6d7 e1e5');
INSERT INTO evaluations VALUES(26237,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26237,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26237,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(26237,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26237,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26237,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(26237,36,'cp -6','g1f3 a7a6 b5c6 b7c6 h2h3 e7e6 c1f4 f8d6 f4d6 d8d6 d1d2 g8f6 e1g1 c6c5 d4c5 d6c5 c3a4 c5c7 f1c1 f5e4 f3g5 e8g8 c2c4 d5c4 g5e4 f6e4 d2d4 e4d6 a4b6 a8d8 b6c4 d6f5 d4e5 c7d7 c1e1 f5d4 a1c1 f7f6 e5e3 g8h8 b2b3 e6e5');
INSERT INTO evaluations VALUES(28760,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(28760,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(28760,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(28760,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(28760,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(28760,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(28760,36,'cp 93','d7d5 e4e5 f6d7 c2c3 c7c5 e1g1 b8c6 h2h4 h7h6 h4h5 f8e7 a2a3 a7a5 a3a4 c5d4 c3d4 c6b4 d3b5 e8g8 b1a3 d7b8 c1d2 b8c6 f1c1 c6a7 f3h2 a8c8 c1c8 b7c8 a1c1 c8d7 c1c3 a7b5 a4b5 e7g5 h2g4 g8h8 c3g3 g5d2 e2d2 d8g5 d2g5 h6g5 h5h6');
INSERT INTO evaluations VALUES(28855,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(28855,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(28855,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(28855,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(28855,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(28855,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(28855,36,'cp 110','f3e5 d7b5 c3b5 d8b6 b5c3 d5d4 c3e2 f8d6 e5c4 b6c7 d2d3 b8c6 c4d6 c7d6 e1g1 e8g8 c1f4 d6c5 a2a3 h7h6 h2h3 f8e8 f4d2 e8e7 b2b4 c5f5 e2g3 f5d5 c2c4 d4c3 d2c3 e7e6 f1e1 e6e1 d1e1 a8e8 e1d1 c6d4 c3b2 h6h5 a1c1 h5h4 c1c5 d5a2 b2d4 h4g3 f2g3 a2a3');
INSERT INTO evaluations VALUES(29115,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29115,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29115,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29115,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29115,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29115,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29115,36,'cp 0','b8d7 f3e5 a7a6 b5d7 f6d7 e5f7 e8f7 d1f3 f7g6 f3g3 g6f7');
INSERT INTO evaluations VALUES(30813,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(30813,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(30813,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(30813,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(30813,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(30813,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(30813,36,'cp 53','d7d6 g1f3 b8d7 c2c3 g7g6 f1d3 e7e5 e1g1 g8f6 d4e5 d6e5 b2b4 f8g7 d1c2 e8g8 a2a4 a7a6 f1e1 c7c5 b4b5 d8c7 h2h3 f8d8 c1b2 d7f8 d2c4 f6e8 b5a6 b7a6 d3f1 a6b7 a1d1 d8d1 e1d1 f8e6 b2c1');
INSERT INTO evaluations VALUES(32610,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(32610,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(32610,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(32610,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(32610,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(32610,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(32610,36,'cp 97','h2h3 g4h5 g2g4 h5g6 f1g2 b8d7 e1g1 c7c6 f3h4 e7e6 f2f4 e8c8 f4f5 e6f5 h4f5 g6f5 f1f5 d6e6 d1f3 e6e1 f3f1 e1f1 g2f1 h7h6 f1c4 d7b6 c4f7 d8d4 c1e3 d4d8 a2a4 c8b8 a4a5 b6c8 a5a6 b7b6 f5f3 f8d6 f7g6 b8c7 g1h1 c8e7 g6d3 h8f8');
INSERT INTO evaluations VALUES(35774,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(35774,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(35774,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(35774,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(35774,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(35774,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(35774,36,'cp 19','h7h6 g5h4 f8e8 g1e2 b8d7 e1g1 a7a5 a2a3 d7f8 h4g3 f8e6 h2h3 e7f8 g3e5 b7b5 f1d1 f6d7 e5g3 d7b6 e3e4 c8d7 e4d5 b6d5 c3d5 c6d5 c2b3 d8b6 a1c1 b6b7 d3b1 a5a4 b3d3 g7g6 e2f4 e8c8 b1a2 c8c1 d1c1 e6f4 g3f4');
INSERT INTO evaluations VALUES(35727,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(35727,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(35727,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(35727,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(35727,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(35727,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(35727,36,'cp 32','g5h4 f8e8 g1e2 a7a5 e1g1 b8d7 a2a3 d7f8 h2h3 f8e6 h4g3 e7f8 f1e1 e6g5 g3e5 f6d7 e5f4 d7f6 c3a4 g5e4 a1d1 f8d6 f2f3 b7b5 a4c3 d6f4 e2f4 e4d6 e3e4 g7g5 f4e2 b5b4 e4e5 b4c3 e2c3 c8a6 d3a6 a8a6 e5d6 e8e1 d1e1 d8d6 c3a4 f6h5');
INSERT INTO evaluations VALUES(35732,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(35732,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(35732,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(35732,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(35732,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35732,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(35732,36,'cp 25','f8e8 g1e2 b8d7 e1g1 f6h5 h4e7 d8e7 a2a3 d7b6 h2h3 g7g6 b2b4 h5g7 b4b5 c8f5 a3a4 a7a5 e2f4 f5d3 f4d3 b6c4 d3b2 c4d6 b5c6 b7c6 c3e2 e8c8 a1b1 a8b8 b2d3 d6c4 b1b8');
INSERT INTO evaluations VALUES(36940,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36940,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36940,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36940,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36940,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36940,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36940,36,'cp 0','h7h6 f1e1 f8e8 e1e8 d8e8 b1d2 b7b6 d2f1 c7c5 c1e3 b8c6 d1c1 e8f8 f1g3 c8d7 e3f4 a8e8 d3f5 f8e7 f4h6 d6g3 c1g5 g3f2 g1f2 f6e4 f5e4 e7g5 e4h7 g8h7 h6g5 c5d4 a1e1 e8e1 f2e1 d7g4 f3d4 c6d4');
INSERT INTO evaluations VALUES(36903,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36903,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36903,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36903,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36903,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36903,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36903,36,'cp 0','f1e1 f8e8 b1d2 e8e1 d1e1 c8e6 f3e5 b8c6 d2f1 c6e7 f1g3 c7c5 d4c5 d6c5 h2h3 c5d6 e5f3 d8c7 c1e3 d6g3 f2g3 e6f5 d3b5 f5d7 b5d3');
INSERT INTO evaluations VALUES(38147,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(38147,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(38147,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(38147,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(38147,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38147,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38147,36,'cp 11','g8e7 c1d2 c5d4 c3b5 b4d2 d1d2 e8g8 b5d4 b8c6 e1c1 d8c7 d2e3 c8d7 f1d3 c6d4 f3d4 e7c6 f2f4 f8c8 c1b1 c6d4 e3d4 c7c5 d4c5 c8c5 d1d2 g7g6 c2c3 a8c8 d3c2 g8f8 h1d1 c5c7 h2h4 c7c4 d2d4 h7h5 g2g3 c4d4');
INSERT INTO evaluations VALUES(39510,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(39510,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(39510,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(39510,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(39510,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(39510,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(39510,36,'cp 17','d7d6 d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8e8 c2c3 e7d6 f1e1 b8d7 d4b5 d6e5 h2h3 a7a6 b5d4 d7f6 e4f3 c7c5 d4b3 d8c7 c1e3 e5d6 e3g5 e8e1 d1e1 c8f5 g5f6 g7f6 a1d1 a8d8 e1e2 b7b5 c3c4 f5e6 c4b5 a6b5 e2b5');
INSERT INTO evaluations VALUES(39626,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(39626,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(39626,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(39626,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39626,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(39626,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(39626,36,'cp 13','f1g2 c7c5 g1e2 b8c6 e1g1 d7d6 d2d3 e7e6 c2c3 g8e7 d3d4 e8g8 c1g5 c5d4 e2d4 c6d4 c3d4 h7h6 g5e3 d6d5 b1c3 d5e4 g2e4 e7d5 c3d5 e6d5 e4g2 c8e6 d1d2 h6h5 h2h4 a8c8 a1c1 d8d7 f1e1 b7b6 b2b3 e6g4 a2a4 g8h7 e3f4 f8d8 c1c8 d8c8');
INSERT INTO evaluations VALUES(42761,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42761,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42761,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(42761,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42761,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42761,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42761,36,'cp 0','e8g8 f1e2 b4c3 d2c3 d7d6 d1c2 b8d7 e1g1 a7a5 f1d1 b7b6 b2b4 c8b7 f3d2 g8h8 a2a4 g7g6 a1b1 b7c6 b1a1');
INSERT INTO evaluations VALUES(43002,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(43002,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(43002,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(43002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43002,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(43002,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(43002,36,'cp 4','f1d3 d7d5 e1g1 e8g8 b2b3 b7b6 c1b2 c8b7 c2c4 c7c5 d4c5 b6c5 c4d5 e6d5 d1c2 b8c6 a2a3 g7g6 f1e1 a8c8 d3f1 f8e8 b1d2 f6d7 a1c1 e7f6 b2f6 d8f6 c2c3 d5d4 e3d4 e8e1 c1e1 c5d4 c3d3 d7c5 d3c4 c5e6 d2e4');
INSERT INTO evaluations VALUES(43001,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(43001,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43001,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43001,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(43001,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43001,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(43001,36,'cp 0','d7d5 b2b3 e8g8 c1b2 c7c5 e1g1 b8d7 b1d2 b7b6 a2a4 c5d4 e3d4 c8b7 d1e2 a8c8 c2c3 f6h5 g2g3 h5f6 f1c1 h7h6 c1c2 f8e8 d3b5 e8f8');
INSERT INTO evaluations VALUES(43281,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43281,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(43281,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(43281,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(43281,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(43281,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(43281,36,'cp 25','g1f3 b7b6 f1g2 c8b7 c2c4 e7e6 d2d4 e8g8 e1g1 d7d5 b1d2 b8d7 a1c1 c7c5 d4c5 b6c5 e2e3 d8b6 c1c2 a7a5 f3e5 d7e5 b2e5 f8d8 d1a1 f6e8 e5g7 e8g7 d2f3 a8c8 a1f6 g7e8 f6f4 e8d6 f1d1 d6f5 c4d5');
INSERT INTO evaluations VALUES(44647,30,'cp 447',NULL);
INSERT INTO evaluations VALUES(44647,31,'cp 450',NULL);
INSERT INTO evaluations VALUES(44647,32,'cp 448',NULL);
INSERT INTO evaluations VALUES(44647,33,'cp 459',NULL);
INSERT INTO evaluations VALUES(44647,34,'cp 448',NULL);
INSERT INTO evaluations VALUES(44647,35,'cp 440',NULL);
INSERT INTO evaluations VALUES(44647,36,'cp 443','d2d3 e4f2 e1f2 e7e6 g1f3 d7d5 e5d6 f8d6 b1c3 b8c6 c1g5 f7f6 g5e3 e8g8 h1e1 g8h8 c3e4 b7b6 e4d6 d8d6 f2g1 e6e5 d1e2 c6a5 c4b5 c8b7 f3d2 d6c7 e3f2 a5c6 b5c6 c7c6 f2g3 a8d8 e2f2 h8g8 a2a4 c5c4 d2c4 c6c5 f2c5 b6c5 c4e3');
INSERT INTO evaluations VALUES(44645,30,'cp 450',NULL);
INSERT INTO evaluations VALUES(44645,31,'cp 438',NULL);
INSERT INTO evaluations VALUES(44645,32,'cp 459',NULL);
INSERT INTO evaluations VALUES(44645,33,'cp 432',NULL);
INSERT INTO evaluations VALUES(44645,34,'cp 435',NULL);
INSERT INTO evaluations VALUES(44645,35,'cp 455',NULL);
INSERT INTO evaluations VALUES(44645,36,'cp 433','e4f2 e1f2 e7e6 g1f3 d7d5 e5d6 f8d6 b1c3 b8c6 c1g5 f7f6 g5e3 e8g8 h1e1 g8h8 c3e4 b7b6 e4d6 d8d6 f2g1 e6e5 d1e2 c6a5 a2a4 a5c4 d3c4 c8b7 a4a5 a8d8 a5b6 a7b6 b2b3 d6c7 a1d1 d8d1 e1d1 f8e8 h2h3 e8d8 d1d8 c7d8 g1h1');
INSERT INTO evaluations VALUES(45407,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45407,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45407,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(45407,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45407,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45407,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45407,36,'cp 0','f6d5 b1c3 b8c6 g1f3 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 b5e2 f7f6 c3d5 d7d5 e1g1 e7f7 f1d1 d5f5 d1d7 f8e7 c1e3 h8d8 a1d1 a7a5 g1g2 a8c8 d7d8 c8d8 d1d8 e7d8 e2a6 f5g6 g2f1 g6b1 f1g2');
INSERT INTO evaluations VALUES(45512,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(45512,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(45512,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45512,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45512,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45512,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(45512,36,'cp 4','b1c3 b8c6 g1f3 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6c5 b5d6 c3d5 e6d5 e1g1 d8b6 c5d5 b6b7 d5e5 e7d7 c1f4 f7f6 e5d4 b7f3 f4g3 d7e8 g3d6 f8d6 d4d6 e8f7 f1d1 h8e8 d6g3 f3f5 a1c1 a8c8 b2b4 e8e4 c1c8 f5c8 g3b3 f7g6 b4b5 c8e6');
INSERT INTO evaluations VALUES(45510,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45510,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45510,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45510,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45510,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45510,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45510,36,'cp 0','b8c6 g1f3 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 c3d5 d7d5 b5e2 f7f6 e1g1 e7f7 f1d1 d5f5 c1e3 f8e7 a1c1 a6a5 g1g2 f5g6 g2h1 a8c8 c1c8 h8c8 e2b5 c8d8 d1d8 e7d8 b5b7 d8e7 h2h3 g6f5 h1g2 f5g6');
INSERT INTO evaluations VALUES(13205,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13205,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13205,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13205,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13205,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13205,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13205,36,'cp 0','g1f3 c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 b5e2 f7f6 e1g1 d5c3 b2c3 e7f7 c1e3 f8e7 f1c1 d7c6 a4a5 e7c5 g1g2 h8c8 a1b1 a8a7 c3c4 a7c7 e2d3 c5e3 d3e3 c6c5');
INSERT INTO evaluations VALUES(13204,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13204,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13204,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13204,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13204,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13204,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(13204,36,'cp 15','c8g4 d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 b5c4 d5c3 b2c3 d7c8 c4h4 e7e8 e1g1 f8e7 h4g3 e8f8 f1d1 f7f6 a1b1 f8f7 c1e3 h7h5 g3h4 a8b8 b1b8 c8b8 d1d7 b8b1 g1g2 b1g6 h4g3 g6g3 h2g3 h8c8 d7a7 c8c6 a4a5 e6e5 e3b6');
INSERT INTO evaluations VALUES(24656,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24656,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24656,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(24656,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(24656,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(24656,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24656,36,'cp 9','d1b3 g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 b5c4 f7f6 e1g1 d5c3 b2c3 e7f7 c1e3 f8e7 f1d1 d7c8 c4h4 c8c6 g1g2 h7h5 a1b1 g7g5 h4e4 c6e4 f3e4 h8d8 d1d8 e7d8 c3c4 a8c8 c4c5 d8e7 h2h3 f7e8 b1b6 e7c5 b6e6 e8f7');
INSERT INTO evaluations VALUES(24655,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24655,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24655,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24655,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24655,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24655,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24655,36,'cp 12','g4f3 g2f3 e7e6 b3b7 c6d4 f1b5 d4b5 b7c6 e8e7 c6b5 d8d7 a2a4 a7a6 b5c4 f7f6 e1g1 d5c3 b2c3 e7f7 c1e3 d7c8 c4h4 h7h5 a1b1 f8e7 f1d1 f6f5 e3g5 e7g5 h4g5 h8h6 g5e3 h6g6 g1h1 c8c7 b1b6 a8c8 b6a6 c7c3 a6a7 f7g8 a4a5 c3e3 f2e3 c8c2 a5a6 c2a2');
INSERT INTO evaluations VALUES(46148,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46148,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46148,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46148,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(46148,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46148,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(46148,36,'cp 0','h7h6 g5f6 d8f6 f3e5 f8d6 f1b5 b8d7 b5d7 c8d7 e5d7 e8d7 e1g1 d7e7 f1e1 f6g6 b1d2 h8d8 d2f3 e7f8 g2g3 f8g8 a2a4 a8c8 d1e2 a7a6 h2h4 g6e4 f3d2 e4f5 a4a5 f5c2 d2f3 c2e4 f3d2');
INSERT INTO evaluations VALUES(46270,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(46270,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(46270,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(46270,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46270,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(46270,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46270,36,'cp 0','g2g3 f8e7 f1g2 b7b6 e4d5 e6d5 f3e5 d8c7 d2f3 b8d7 e5d7 c8d7 c1f4 e7d6 f4d6 c7d6 d1e2 d7e6 e2e5 d6e5 f3e5 f6d7 e5c6 d7b8 c6e5');
INSERT INTO evaluations VALUES(46271,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(46271,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(46271,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46271,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(46271,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46271,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46271,36,'cp -3','f8e7 f1g2 e8g8 e1g1 b8c6 f1e1 b7b5 e4d5 e6d5 d2b3 c8f5 c1f4 a7a5 a2a4 b5a4 a1a4 f8e8 f3e5 c6e5 f4e5 f5d7 a4a1 a5a4 b3d2 d7g4 d1c1 g4e6 b2b3 f6d7 e5b2 a4b3 a1a8 d8a8 d2b3 a8a7 c1a1 a7a1 e1a1 e8b8');
INSERT INTO evaluations VALUES(47977,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(47977,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(47977,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(47977,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(47977,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(47977,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(47977,36,'cp 139','d2d4 d8e7 b1c3 c7c6 d4d5 h6g4 g1e2 g4f6 e2g3 e7d8 a2a4 f8e7 h2h3 e8g8 e1g1 c6d5 c3d5 f6d5 c4d5 b8c6 c1e3 e7g5 a1d1 c6b4 d5b3 g5e3 f3e3 b4c6 f2f4 c8e6 g3f5 e6f5 e4f5 d8e7 c2c3 f8e8 f4e5 e7e5 e3e5 d6e5');
INSERT INTO evaluations VALUES(49607,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(49607,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(49607,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(49607,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(49607,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(49607,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(49607,36,'cp -105','f6e4 c1b2 d7d5 f3e5 f8d6 e5f3 e8g8 f1e2 c7c5 d2d3 e4f6 e1g1 b8c6 f1e1 f8e8 e2f1 e8e1 d1e1 c8f5 b1d2 h7h6 g2g3 d5d4 d2c4 d8d7 c4d6 d7d6 f1g2 a8d8 e1d2 b7b6 a1e1 f5e6 a2a3 a7a5 f3h4 d6d7 b2c1 f6d5');
INSERT INTO evaluations VALUES(53870,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(53870,31,'cp 247',NULL);
INSERT INTO evaluations VALUES(53870,32,'cp 256',NULL);
INSERT INTO evaluations VALUES(53870,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(53870,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(53870,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(53870,36,'cp 238','f1b5 b8d7 e5d7 c8d7 a2a4 g8f6 b1c3 f8e7 d2d4 c5d4 d1d4 e8g8 e1g1 d7c6 f1d1 a7a6 b5c6 b7c6 c1f4 f8e8 f4d6 e7d6 d4d6 d8b6 b2b3 f6e4 c3e4 e8e4 d1e1 e4e1 a1e1 a8d8 d6f4 b6a5 g1f1 h7h6 f4c4 a5c7 h2h3 c7h2');
INSERT INTO evaluations VALUES(53902,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(53902,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53902,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(53902,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(53902,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(53902,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(53902,36,'cp 43','c8f5 e4f5 b8c6 b1c3 d8d7 f1c4 d7f5 e1g1 e8c8 c3d5 g8e7 c2c3 c6e5 c4b3 f5d3 d5e7 f8e7 b3f7 d3d1 f7e6 c8c7 f1d1 e7f6 e6d5 h8e8 g2g3 e5d7 a2a4 d7b6 d5b3 a7a5 b3c2 h7h6 c2d3 f6g5 c1g5 h6g5 d3b5 e8e5 d1e1 g5g4 e1e5 d6e5');
INSERT INTO evaluations VALUES(31615,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(31615,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(31615,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(31615,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(31615,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(31615,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(31615,36,'cp 28','e4f5 b8c6 b1c3 d8d7 c3d5 e8c8 f1c4 d7f5 e1g1 g8e7 c2c3 c6e5 c4b3 c8b8 c1e3 c5c4 d5e7 f8e7 b3c2 f5g4 d1d5 g4e6 d5d2 e7f6 f1e1 e6d7 a1d1 h8e8 d2d5 d7c6 d5a5 b7b6 a5b4 e8e6 a2a4 e5g4 e3f4');
INSERT INTO evaluations VALUES(31614,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(31614,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(31614,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(31614,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(31614,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(31614,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31614,36,'cp 34','b8c6 b1c3 d8d7 f1c4 d7f5 e1g1 e8c8 c3d5 g8e7 f1e1 c8b8 a1b1 c6e5 d5e3 f5d7 c4f1 d6d5 e3d5 d7d5 d1d5 d8d5 c2c4 d5d8 e1e5 e7c6 e5d5 d8d5 c4d5 c6e5 g2g3 f8d6 g1g2 e5d7 c1d2 h8e8 b2b3 g7g6 h2h3 b8c7 d2a5 c7c8 a5c3 f7f6');
INSERT INTO evaluations VALUES(55663,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(55663,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(55663,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55663,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55663,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55663,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55663,36,'cp 5','b8c6 d4d1 g8f6 b1c3 c8e6 c4e6 f7e6 g1f3 f8e7 e1g1 a7a6 c1e3 e8g8 d1d3 f6g4 c3e2 g4e3 d3e3 d8e8 a1d1 e7f6 e2d4 g8h8 d4c6 e8c6 c2c3 e6e5 f1e1 f8f7 e1e2 b7b5 e2c2 a8f8 e3d3 c6b6 c3c4 a6a5');
INSERT INTO evaluations VALUES(22019,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22019,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22019,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22019,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22019,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(22019,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(22019,36,'cp 12','d4d1 g8f6 b1c3 c8e6 c4e6 f7e6 g1f3 f8e7 e1g1 e8g8 d1d3 d8d7 h2h3 a8e8 c1f4 f6h5 f4h2 h5f4 h2f4 f8f4 c3e2 f4f7 a1d1 e8f8 e2d4 c6d4 f3d4 e7f6 c2c3 f6d4 c3d4 a7a5 d1c1 d6d5 f2f3 c7c6 f1d1 a5a4');
INSERT INTO evaluations VALUES(56920,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(56920,31,'cp 259',NULL);
INSERT INTO evaluations VALUES(56920,32,'cp 238',NULL);
INSERT INTO evaluations VALUES(56920,33,'cp 273',NULL);
INSERT INTO evaluations VALUES(56920,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(56920,35,'cp 260',NULL);
INSERT INTO evaluations VALUES(56920,36,'cp 287','f8g7 d2d4 d7d6 e5c4 b8c6 d4d5 c6e5 b1c3 h7h6 c1e3 g8e7 c4e5 g7e5 h2h4 g5h4 d1d2 e7g6 f2f4 e5g7 e1c1 h6h5 d1e1 a7a6 c1b1 c8g4 e3d4 g7d4 d2d4 d8e7 f1d3 e8c8 d4a7 g6f4 a7a8 c8d7 a8b7 d7e8 e1f1 f4g2 e4e5 g2e3 e5e6 e3f1');
INSERT INTO evaluations VALUES(57016,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57016,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(57016,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(57016,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57016,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57016,36,'cp 0','g8f6 e2e3 b7b6 g1f3 c8b7 f1d3 c7c5 e1g1 b8c6 c1b2 c5d4 e3d4 c6b4 d3e2 f8e7 c2c4 d7d5 a2a3 b4c6 b1c3 e8g8 f1e1 a8c8 e2d3 c6a5 f3e5 d5c4 b3c4 a5c6 e5f3 c6a5');
INSERT INTO evaluations VALUES(49412,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49412,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49412,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49412,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49412,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49412,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49412,36,'cp 0','g1f3 b7b6 c1b2 c8b7 e2e3 d7d5 f1d3 f8d6 c2c4 b8d7 e1g1 e8g8 f3e5 c7c5 e5d7 d8d7 d4c5 d6c5 b2f6 g7f6 b1c3 d5d4 c3e4 b7e4 d3e4 a8d8 d1h5 f6f5 h5g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(49411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49411,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49411,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(49411,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(49411,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49411,36,'cp 0','b7b6 c1b2 c8b7 e2e3 c7c5 c2c4 c5d4 f3d4 b8c6 b1c3 c6d4 d1d4 f8c5 d4f4 d8e7 f1e2 c5d6 f4d4 d6c5');
INSERT INTO evaluations VALUES(4460,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(4460,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4460,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4460,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4460,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4460,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4460,36,'cp -1','f1e2 d7d5 e4d5 f6d5 d1d2 d5c3 d2c3 f8e8 e1g1 c6e7 d3d4 e7d5 c3d2 d5f6 a1d1 e5d4 e3d4 d8d6 f3e5 c8e6 d2g5 d6e7 e2f3 e6d5 d1e1 d5f3 f1f3 h7h6 g5h4 c7c5 c2c3 c5d4 f3g3 g8h7 c3d4 a8d8 e1f1 e7e6 h2h3');
INSERT INTO evaluations VALUES(6291,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(6291,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(6291,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(6291,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(6291,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(6291,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(6291,36,'cp 169','d6e5 f3e5 c6e5 c4e2 d7d6 b1c3 e5c6 e1g1 g8f6 b2b3 e8g8 c1b2 d8e7 c3d5 f6d5 e4d5 f8e8 f1e1 c6e5 d1d4 e7f8 f2f4 e5g4 h2h3 g4f6 e2d3 c8d7 g1h2 a7a5 d4c4 e8e1 a1e1 a8e8 e1f1 c7c5 d5c6 b7c6 c4d4 c6c5');
INSERT INTO evaluations VALUES(6256,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(6256,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(6256,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(6256,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(6256,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(6256,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(6256,36,'cp 163','f3e5 c6e5 c4e2 g8f6 c2c4 e8g8 b1c3 a7a6 e1g1 d7d6 f2f3 b7b5 c4b5 a6b5 c3b5 c8a6 b5c3 a6e2 d1e2 f8e8 f1d1 c7c6 c1e3 d6d5 e4d5 f6d5 c3d5 c6d5 e2f2 d8a5 b2b3 e5c6 h2h3 e8e6 e3b6 a5b5 f2c5 b5e2 a2a4 e6g6 c5f2 e2e6 g1h1 d5d4 b3b4 e6c4 b6c5 c6b4');
INSERT INTO evaluations VALUES(6920,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(6920,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(6920,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(6920,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(6920,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(6920,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(6920,36,'cp 142','d7d6 h2h3 a7a6 a2a4 d8d7 d2d3 g7g5 c3d5 h7h5 c2c3 c5a7 d3d4 c6e7 d4e5 d6e5 d1b3 e7d5 c4d5 g8h6 d5b7 a8b8 b7c8 b8b3 c8d7 e8d7 f1d1 d7e6 d1d5 h8a8 f3d2 b3b8 g1f1 h6f7 d5d3 b8d8 d3d8 a8d8 f1e2 g5g4 h3g4 d8g8 d2b3 g8g4 c1e3 a7e3 e2e3 f7g5 b3c5 e6d6 c5a6 g4g2 a4a5 g5h3');
INSERT INTO evaluations VALUES(6110,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(6110,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(6110,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(6110,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(6110,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(6110,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(6110,36,'cp 151','h2h3 a7a6 a2a4 d8d7 d2d3 g7g5 c3d5 h7h5 c2c3 c5a7 d3d4 c6e7 d5e3 e7c6 b2b4 g5g4 f3h4 g4h3 h4g6 h8h6 g6h4 h6h8 g2g3 e5d4 e3f5 d4c3 h4g6 d7h7 g6h8 h7h8 a1a3 g8e7 a3c3 c6e5 f5h6 e5g4 h6g4 c8g4 c4e2 e8c8 e2g4 h5g4 d1g4 c8b8');
INSERT INTO evaluations VALUES(11766,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(11766,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(11766,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(11766,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(11766,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(11766,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(11766,36,'cp 68','a4b3 h7h6 e1g1 g7g6 d3d4 e5d4 f3d4 f8g7 d4c6 e7c6 b1c3 e8g8 c1e3 d7d6 d1d2 g8h7 a1d1 c8b7 b3d5 d8d7 f2f3 a8e8 f1e1 d7c8 b2b3 c6e7 d5b7 c8b7 c3e2 e7c6 h2h3 c6d8 e3d4 g7d4 d2d4 d8e6 d4e3 e8e7 e2d4 e6d4 e3d4');
INSERT INTO evaluations VALUES(16706,30,'cp -166',NULL);
INSERT INTO evaluations VALUES(16706,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(16706,32,'cp -170',NULL);
INSERT INTO evaluations VALUES(16706,33,'cp -176',NULL);
INSERT INTO evaluations VALUES(16706,34,'cp -177',NULL);
INSERT INTO evaluations VALUES(16706,35,'cp -159',NULL);
INSERT INTO evaluations VALUES(16706,36,'cp -158','f8b4 c2c3 b4c5 c1e3 c5e3 d3e3 e8g8 b1d2 d7d5 e1c1 d5d4 e3d3 f6g4 g1h3 g4e5 d3g3 c8h3 g3h3 d8f6 c3d4 c6d4 h3g3 a8d8 c1b1 c7c5 d2c4 d4f5 g3b3 d8d1 b3d1 f5d4 d1d2 b7b5 c4e5 f6e5 f1d3 c5c4 f2f4 e5c5 d3c2 d4c2 d2c2 c5e3 g2g3');
INSERT INTO evaluations VALUES(18941,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(18941,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(18941,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(18941,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(18941,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18941,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(18941,36,'cp -14','f8c5 f1d3 g8e7 e1g1 e7g6 b1d2 e8g8 d2b3 c5e7 b3d4 c6d4 f3d4 b7b6 c1e3 c8b7 c2c3 a7a6 f2f4 e6e5 d4e2 e5f4 e2f4 b6b5 a4c2 e7g5 c2d2 g6e5 f4d5 g5e3 d2e3 f8e8 a1e1 h7h6 e3g3 b7d5 e4d5 e5d3 e1e8 d8e8');
INSERT INTO evaluations VALUES(19004,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19004,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19004,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(19004,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(19004,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(19004,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19004,36,'cp 12','d7d6 d2d3 g8e7 f2f4 e5f4 c1f4 e7g6 f4e3 f8e7 g1e2 e7f6 e1g1 e8g8 d1d2 a8b8 d3d4 c8g4 c3d5 f6d4 e2d4 c5d4 e3d4 c6d4 d2d4 a7a5 d5e3 g6e5 h2h3 g4e6 e3f5 b7b5 a1d1 b5b4 b2b3 e6f5 e4f5 b8c8 d4d6 d8d6 d1d6');
INSERT INTO evaluations VALUES(20613,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20613,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(20613,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(20613,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(20613,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(20613,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(20613,36,'cp 14','d2d3 g8e7 f2f4 e5f4 c1f4 e7g6 f4e3 f8e7 c3d5 c8e6 g1h3 e8g8 e1g1 d8d7 h3f4 g6f4 d5f4 e7g5 d1d2 a8e8 b2b3 b7b5 c2c3 g5f4 g3f4 e6h3 f4f5 h3g2 d2g2 f7f6 a1d1 b5b4 d3d4 b4c3 d4c5 d7e7 d1d6 e7e4 g2e4 e8e4');
INSERT INTO evaluations VALUES(20616,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20616,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(20616,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(20616,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(20616,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(20616,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(20616,36,'cp 12','g8e7 f2f4 e5f4 c1f4 e7g6 f4e3 f8e7 c3d5 c8e6 g1h3 e8g8 e1g1 d8d7 h3f4 g6f4 d5f4 e7g5 d1d2 g5f4 g3f4 f7f5 b2b3 b7b6 c2c4 f8f6 f1f2 a8f8 a1d1 f6g6 g1h1 g8h8 d1f1 h8g8 a2a3 f5e4 d3e4 e6h3 f4f5 h3g2 f2g2 g6g2 d2g2');
INSERT INTO evaluations VALUES(22345,30,'cp 380',NULL);
INSERT INTO evaluations VALUES(22345,31,'cp 387',NULL);
INSERT INTO evaluations VALUES(22345,32,'cp 375',NULL);
INSERT INTO evaluations VALUES(22345,33,'cp 363',NULL);
INSERT INTO evaluations VALUES(22345,34,'cp 375',NULL);
INSERT INTO evaluations VALUES(22345,35,'cp 398',NULL);
INSERT INTO evaluations VALUES(22345,36,'cp 392','f1b5 e8f7 c1d2 a7a6 b5d3 g7g6 d2c3 g8h6 b1d2 f8g7 e1c1 h8e8 f2f3 d8e7 g2g4 f7g8 d4f2 h6f7 h2h4 c7c5 d5c6 b7c6 h4h5 g6g5 h5h6 g7f8 f3f4 c8e6 f4g5 f6g5 c3f6 e7a7 f6d4 a7e7 d2c4 f7h6 e4e5 d6d5 c4d6 h6g4 f2g3');
INSERT INTO evaluations VALUES(22343,30,'cp 380',NULL);
INSERT INTO evaluations VALUES(22343,31,'cp 396',NULL);
INSERT INTO evaluations VALUES(22343,32,'cp 372',NULL);
INSERT INTO evaluations VALUES(22343,33,'cp 374',NULL);
INSERT INTO evaluations VALUES(22343,34,'cp 371',NULL);
INSERT INTO evaluations VALUES(22343,35,'cp 368',NULL);
INSERT INTO evaluations VALUES(22343,36,'cp 399','c8d7 b1c3 g7g6 b5d7 d8d7 d4b4 a8b8 c1e3 g8e7 b4b3 b7b6 c3e2 f8g7 e2f4 e8g8 f4e6 f8f7 e1c1 a7a5 g2g4 c7c6 h2h4 c6d5 e4d5 b6b5 b3d3 f6f5 g4g5 b8c8 e6g7 f7g7 e3d4 g7f7 h4h5 f5f4 d4f6 g6h5 h1h5 e7g6');
INSERT INTO evaluations VALUES(23331,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(23331,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(23331,32,'cp 250',NULL);
INSERT INTO evaluations VALUES(23331,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(23331,34,'cp 243',NULL);
INSERT INTO evaluations VALUES(23331,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(23331,36,'cp 223','c1e3 g8e7 a2a3 d7d6 f1c4 d6d5 c4b3 e7c6 d4d5 d8d5 b3d5 c8d7 b1c3 h7h5 a1d1 e8c8 h2h4 f8d6 f2f4 c6e7 d5c4 d7g4 d1d3 c7c6 d3d2 b7b5 c4e2 g4e2 e1e2 f6f5 h1d1 d6c7 d2d8 h8d8 d1d8 c7d8');
INSERT INTO evaluations VALUES(23364,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(23364,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(23364,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(23364,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(23364,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(23364,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(23364,36,'cp 176','f3e5 c7c6 b5c4 d7d5 e4d5 e7d5 d2d4 f8d6 e1g1 g8e7 b1c3 d8c7 c3d5 c6d5 c4b5 e7c6 d1h5 e8g8 e5c6 b7c6 b5d3 g7g6 h5h4 d6e7 h4h6 e7f6 c2c3 f6g7 h6h4 c7b6 f1e1 c8a6 d3a6 b6a6 h4e7 a6c4 c1f4 c4b5 e7c7 b5b2 c7c6 a8c8');
INSERT INTO evaluations VALUES(23363,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(23363,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(23363,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(23363,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(23363,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(23363,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(23363,36,'cp 191','c7c6 b5c4 d7d5 e4d5 e7d5 e1g1 f8d6 d2d4 g8e7 b1d2 c8f5 f1e1 e8g8 d2e4 e7g6 e5g6 h7g6 c1g5 d6e7 g5e7 d8e7 e4c3 e7d7 c4d5 c6d5 e1e5 f5e6 d1d2 f7f6 e5e3 f8e8 a1e1 e6f7 c3d1 e8e3 d1e3 a8e8 c2c3 e8e7 e1e2 e7e8 a2a3 g8h7');
INSERT INTO evaluations VALUES(25649,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(25649,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(25649,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(25649,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(25649,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(25649,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(25649,36,'cp 82','c2c4 g8f6 b1c3 d5c4 h2h3 g4f3 d1f3 c6a5 d4d5 c7c6 d5c6 a5c6 f1c4 e7e6 e1g1 f8e7 f3e2 e8g8 a2a3 a8c8 c4d3 d8b6 b2b4 c6d4 e2b2 b6c6 f1c1 d4f5 b4b5 a6b5 f4e5 f8d8 c3b5 c6b6');
INSERT INTO evaluations VALUES(28560,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(28560,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(28560,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(28560,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28560,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(28560,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(28560,36,'cp 19','f3e5 e7e6 e2e3 f6d7 g2g4 f5g6 e5g6 h7g6 e3e4 d7f6 f1g2 f8b4 e4d5 f6d5 f4d2 c7c6 c3e2 b4d2 d1d2 d8c7 h2h3 d5f6 e1c1 b8d7 e2f4 e8c8 f4d3 c6c5 d2e3 c5d4 e3d4 c8b8 d1d2 d7b6 d4e5 c7e5 d3e5 d8d2 c1d2 f6d7 e5d7 b6d7 f2f4 b8c7 a2a3 h8h4 d2e3 d7b6');
INSERT INTO evaluations VALUES(29450,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(29450,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(29450,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(29450,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(29450,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(29450,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(29450,36,'cp 105','g4e2 d1e2 e7e5 g1f3 b8d7 c1e3 c7c6 e1c1 d8c7 h2h3 b7b5 g2g4 h7h6 g4g5 h6g5 f3g5 f8e7 h1g1 d7b6 f2f4 a8d8 e2f3 b6c4 c1b1 b5b4 c3a4 c4e3 f3e3 d8d7 f4e5 d6e5 g5f3 e5d4 f3d4 g7g6 e4e5 f6d5');
INSERT INTO evaluations VALUES(29453,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(29453,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(29453,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(29453,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(29453,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(29453,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(29453,36,'cp 84','g1e2 b8d7 f2f4 e7e5 e1g1 f8e7 c1e3 f6g4 e3d2 g4f6 g1h1 e8g8 d2e3 f6h5 d1d3 h5f4 e2f4 e5f4 e3f4 e7g5 c3d5 g5f4 f1f4 c7c5 a1f1 c5d4 d3d4 d7e5 d4f2 f7f6 a2a3 a7a6 h2h3 g8h8 f4f5 d8e8');
INSERT INTO evaluations VALUES(29926,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(29926,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(29926,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29926,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(29926,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(29926,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29926,36,'cp 2','b8c6 c4b5 e5f4 d2d4 d6d5 e4d5 f6d5 e1g1 f8e7 h2h3 g4h5 d1e2 d5c3 b2c3 e8g8 c1f4 e7d6 e2d2 h5g6 b5d3 d8d7 f3h4 g6d3 c2d3 c6e7 a1e1 b7b5 f4d6 c7d6 c3c4 b5c4 d3c4 a8c8 d2d3 d7c7 a2a4 e7g6 h4g6 h7g6');
INSERT INTO evaluations VALUES(36621,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36621,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36621,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36621,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(36621,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36621,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(36621,36,'cp -17','f6d7 f4h2 e7e5 f1e2 b8c6 c2c3 a7a5 d4e5 d6e5 e1g1 d8e7 e2b5 d7b6 a2a4 c6b8 h2g3 c8d7 d1e2 b8a6 b5d7 b6d7 d2b3 a6c5 b3c5 d7c5 e2c4 g8h8 f1d1 f7f6 d1d5 b7b6 a1d1 e7e8 b2b3 e8c6 e3e4 a8c8 d1e1 f8e8 e1d1 g7f8');
INSERT INTO evaluations VALUES(37510,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(37510,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(37510,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(37510,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(37510,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(37510,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(37510,36,'cp 63','d1d3 d7d6 a2a3 b4c3 d3c3 e6e5 d4e2 g8f6 c3g3 c8e6 e2c3 b8c6 f1d3 f6h5 g3g5 h5f6 c1e3 e8g8 g5g3 c6d4 e1c1 d8b6 f2f4 f6h5 g3f2 h5f4 d3f1 f4g6 h2h4 h7h5 f1e2 g6f4 e3d4 e5d4 d1d4 f4e2 c3e2 a8c8 e2c3 c8c3 b2c3 b6a5');
INSERT INTO evaluations VALUES(38824,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(38824,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(38824,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(38824,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(38824,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(38824,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(38824,36,'cp 57','b1d2 g8f6 g1f3 b7b6 a1c1 e8g8 f1e2 c8b7 e1g1 b8d7 c4d5 f6d5 d1b3 a8c8 d2c4 b7a8 b3a4 a7a5 f1d1 d8e7 c4e5 d7e5 f3e5 e7g5 e5f3 g5e7 a2a3 d5f6 f3e5 c7c5 d4c5 c8c5 c1c5 e7c5 a4d4 f8c8 d4c5');
INSERT INTO evaluations VALUES(38822,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(38822,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(38822,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(38822,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(38822,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(38822,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(38822,36,'cp 57','g8f6 a1c1 e8g8 g1f3 b7b6 f1e2 c8b7 c4c5 c7c6 e1g1 b8d7 b2b4 a7a5 b4b5 a8c8 c5b6 d8b6 d1b3 a5a4 b3a4 c8a8 a4b3 c6b5 e2b5 b7a6 b5d7 f6d7 b3b6 d7b6 f1e1 f8b8 c1a1 a6c4 f3e5 a8a3 e1c1 a3a2 a1a2 c4a2');
INSERT INTO evaluations VALUES(41170,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(41170,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(41170,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(41170,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(41170,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(41170,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(41170,36,'cp 87','g5e3 f6g4 e3c1 g4f6 h2h3 d6d5 e4e5 f6e4 a2a4 c8f5 f1d3 c7c5 d4c5 e8g8 e1g1 b8d7 c1e3 a8c8 c3d5 d7e5 d5b4 e5d3 b4d3 d8d5 f1e1 f8d8 a4a5 e4g5 f3g5 h6g5 a5a6 b7a6 a1a6 f5d3 d1d3 d5d3 c2d3 d8d3 a6a7');
INSERT INTO evaluations VALUES(42629,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42629,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42629,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42629,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42629,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42629,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42629,36,'cp 0','d2d4 e5d4 c3d4 c5b6 b1c3 e8g8 f1c4 c7c6 e4e5 d7d5 c4d3 f8e8 g1e2 f6d7 f3f4 d8h4 g2g3 h4h5 c3a4 b6a5 a4c3');
INSERT INTO evaluations VALUES(42611,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42611,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42611,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42611,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42611,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42611,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42611,36,'cp 0','e5d4 c3d4 c5b6 b1c3 e8g8 g1e2 b8c6 c1g5 h7h6 g5h4 f8e8 e4e5 d7d5 c3a4 b6a5 a4c3 a5b6');
INSERT INTO evaluations VALUES(42685,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42685,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42685,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42685,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42685,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42685,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42685,36,'cp 0','c3d4 c5b6 b1c3 e8g8 g1e2 b8c6 c1g5 h7h6 g5h4 f8e8 e4e5 d7d5 c3a4 b6a5 a4c3 a5b6');
INSERT INTO evaluations VALUES(42681,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42681,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42681,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42681,36,'cp 0','c5b6 b1c3 e8g8 g1e2 b8c6 c1g5 h7h6 g5h4 f8e8 e4e5 d7d5 c3a4 b6a5 a4c3 a5b6');
INSERT INTO evaluations VALUES(46769,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(46769,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(46769,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(46769,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(46769,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46769,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46769,36,'cp 0','f8g7 g1f3 c7c5 e1g1 e8g8 h2h3 d8b6 b1c3 c5c4 d3e2 b6b2 d1d2 b2b6 f3e5 b8c6 e2f3 e7e6 a1b1 b6d8 e5c6 b7c6 e3e4 c8a6 f1e1 a8c8 c3a4 d5e4 f3e4 f8e8 e4f3 a6b5 a4c3 b5a6');
INSERT INTO evaluations VALUES(47826,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(47826,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(47826,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(47826,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(47826,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(47826,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(47826,36,'cp 23','b1c3 g7g6 g2g3 c7c5 g1e2 b8c6 f1g2 f8g7 e1g1 a8b8 a1b1 a7a6 a2a3 b7b5 c4b5 a6b5 b2b4 c5b4 a3b4 e8g8 f2f4 d8b6 g1h1 c8g4 h2h3 g4e2 c3e2 b8d8 d1e1 d6d5 e2c3 d5e4 c1e3 c6d4 d3e4 b6e6 b1d1 e5f4 g3f4 f6h5 c3d5');
INSERT INTO evaluations VALUES(48613,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(48613,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(48613,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(48613,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(48613,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(48613,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(48613,36,'cp 154','f3e5 d8d5 d1e2 f8e7 b1c3 d5d6 b2b3 e8g8 c1b2 b8c6 e5c6 d6c6 e1c1 f8e8 f2f3 a7a6 e2f2 e7c5 d2d4 c5b4 d4d5 c6c5 f2c5 b4c5 f1d3 c5f8 g2g4 b7b5 g4g5 f6d7 h2h4 d7c5 c3e4 c5d3 d1d3 c8f5 d3e3 a6a5 h1e1 a5a4 h4h5');
INSERT INTO evaluations VALUES(49794,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(49794,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(49794,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(49794,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(49794,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(49794,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(49794,36,'cp 69','b8c6 g1e2 c6a5 c4b3 a5b3 a2b3 f7f6 f3g3 d7d5 d2d4 h6f7 e4d5 d8d5 b1c3 d5e6 e1g1 c8d7 g3f3 c7c6 d4e5 f6e5 e2g3 f8e7 c1e3 e6g6 f3e4 g6e4 c3e4 h7h5 e3a7 h5h4 g3e2 d7f5 e2c3 e8g8 f2f3 h4h3 a7e3 h3g2 g1g2 g8h7 f1d1');
INSERT INTO evaluations VALUES(49995,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(49995,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(49995,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(49995,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(49995,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(49995,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(49995,36,'cp 18','g1f3 d7d6 g2g3 a7a5 d3d4 e5e4 f3h4 c7c6 f2f3 g7g6 c1g5 f8e7 f1h3 e8g8 e1g1 b8a6 h4g2 e4f3 e2f3 d6d5 c4c5 f6h5 g5e7 d8e7 f1e1 e7f6 f3f4 a6c7 d1d2 h7h6 e1e5 f6f7 c3a4 h5f6 a4b6 a8b8');
INSERT INTO evaluations VALUES(50318,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(50318,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(50318,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(50318,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(50318,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(50318,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(50318,36,'cp -12','b8c6 c1b2 d7d5 c4d5 f6d5 e2e3 e7e5 a2a3 f8d6 d2d3 d8e7 b1d2 a7a6 b3b4 e8g8 g1f3 c8d7 f1e2 b7b5 d2e4 a6a5 b4a5 a8a5 e4d6 c7d6 e1g1 a5a8 f3d2 d5b6 a1b1 c6a5 b2a1 f8c8 h2h3 b6a4 d3d4 a5c4 d2c4 b5c4');
INSERT INTO evaluations VALUES(50755,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(50755,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(50755,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50755,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50755,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(50755,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50755,36,'cp -1','e1g1 f8e7 d1e2 c8d7 f1d1 b8c6 c1f4 e6e5 f4e3 g8f6 a2a3 a8c8 c3d5 f6d5 c4d5 e8g8 a1c1 g8h8 b2b4 d8e8 h2h3 h7h6 e2d2 f7f5 a3a4 e8g6 e4f5 d7f5 g1h2 f5e6 b4b5 a6b5 a4b5 c6d8 c1c8 e6c8 d1c1 c8e6 c1c7 e6d5 d2d5 g6f7 d5f7');
INSERT INTO evaluations VALUES(50754,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50754,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50754,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50754,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50754,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50754,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50754,36,'cp -9','b7b5 c4e2 f8e7 b2b4 g8f6 a2a4 b5a4 d1a4 b8d7 e4e5 d6e5 f3e5 d8c7 a4c6 c7c6 e5c6 c8b7 c6e7 e8e7 c1e3 h8c8 f1c1 e7f8 h2h3 f8g8 g1h2 h7h6 a1a5 c8e8 e3f4 d7b6 f4d2 e8b8 c1a1 b8c8 e2a6 b7a6 a5a6');
INSERT INTO evaluations VALUES(51560,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(51560,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(51560,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(51560,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(51560,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(51560,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(51560,36,'cp 175','f1d3 d7d6 e1g1 g8f6 d1e2 f8e7 a2a4 b5b4 a4a5 d6d5 b1d2 e8g8 f1e1 d8c7 e4e5 f6d7 d2b3 h7h6 c1f4 c6c5 d4c5 d7c5 f3d4 c8a6 e2g4 g8h8 a1c1 c7d8 b3c5 e7c5 d4b3 a6d3 c2d3 c5e7 f4e3 b8d7 d3d4 a8b8 g4h3 e7g5 c1c6 g5e3');
INSERT INTO evaluations VALUES(51792,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51792,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(51792,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(51792,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(51792,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51792,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(51792,36,'cp 34','g8f6 g1f3 f8e7 b1d2 c6c5 d4c5 e7c5 f1d3 e8g8 e1g1 b8c6 h2h3 d7d5 c2c4 c5d6 f4d6 d8d6 c4d5 d6d5 d1b1 f8d8 d3e2 b7b6 f1d1 c8b7 b2b4 d5d6 b4b5 c6e5 f3e5 d6e5 d2f3 d8d1 b1d1 e5b2 a1c1 b2a2 c1c7');
INSERT INTO evaluations VALUES(44054,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(44054,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(44054,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(44054,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(44054,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(44054,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(44054,36,'cp 35','g1f3 d7d5 c2c4 f8d6 f4d6 d8d6 b1c3 e8g8 a1c1 d6e7 d1b3 b8d7 f1e2 d5c4 b3c4 e6e5 e1g1 e5d4 c4d4 d7b6 h2h3 c8f5 f1d1 f8e8 d4d6 e7d6 d1d6 f6e4 d6d4 e4c3 c1c3 f5e6 c3c2 e8d8 b2b4 d8d4');
INSERT INTO evaluations VALUES(1104,30,'cp -153',NULL);
INSERT INTO evaluations VALUES(1104,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(1104,32,'cp -158',NULL);
INSERT INTO evaluations VALUES(1104,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(1104,34,'cp -143',NULL);
INSERT INTO evaluations VALUES(1104,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(1104,36,'cp -146','d7d5 e4d5 e5e4 f3g3 c6e5 b2b3 e5d3 c4d3 e4d3 e1d1 f6d5 c1a3 e7f6 g1f3 c8e6 a3f8 e8f8 f3d4 h7h5 g3d3 a8e8 b1a3 f6f2 d3f1 e6g4 d1c1 f2f4 f1c4 c7c6 g2g3 f4f2 c4f1 f2f1 h1f1 h8h6 a3c4 c6c5 d4b5 e8e2 b5d6 b7b5 d6b5 e2h2 b5d6 h6f6 c1c2');
INSERT INTO evaluations VALUES(3292,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(3292,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(3292,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(3292,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3292,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(3292,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(3292,36,'cp 18','f6h5 g3d6 d8d6 g2g3 f7f6 c3c4 g7g6 e1g1 b7b6 c4d5 e6d5 d3b5 c8d7 a1c1 h5g7 f1e1 g7f5 d2b1 g8g7 b1c3 f8e8 b5e2 a8c8 c3b5 d6b4 a2a3 b4f8 e2f1 e8e1 f3e1 a7a6 b5c3 c6d4 c3d5 c8c1 d1c1');
INSERT INTO evaluations VALUES(4931,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(4931,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(4931,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(4931,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(4931,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(4931,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(4931,36,'cp 170','e5c4 d6c5 d2d4 c5a7 b1c3 d7d5 c4e5 g8e7 c1g5 c8b7 e4d5 c6d5 d1g4 d8c8 g5e7 c8g4 e5g4 e8e7 e1c1 f7f6 g4e3 e7f7 h1e1 a8d8 c3a4 b7c8 c2c3 h7h5 a4c5 h5h4 b2b4 c7c6 a2a4 a7b8 h2h3 a6a5 c1d2 b8f4 c5d3 f4d6 b4b5 c6b5 a4b5');
INSERT INTO evaluations VALUES(5535,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(5535,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(5535,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(5535,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(5535,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5535,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(5535,36,'cp 18','g8f6 d1d2 e8g8 e1g1 c8d7 h2h3 a7a6 a2a4 a8c8 f1d1 d8c7 d4b3 d7e6 a1a3 c6a5 b3a5 c7a5 e2f3 c8c7 c3d5 a5d2 d1d2 f6d5 e4d5 e6d7 c2c3 b7b5 a4b5 d7b5 f3e2 f8c8 g2g4 c7b7 e2b5 a6b5 a3b3 b7b8 b3b4');
INSERT INTO evaluations VALUES(6019,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(6019,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(6019,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(6019,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(6019,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6019,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6019,36,'cp 22','c2c3 f7f5 e4f5 c8f5 d3d4 c5b6 d4e5 d6e5 d1d8 a8d8 b1d2 g8f6 d2c4 e8g8 e1g1 h7h6 a2a4 f6d5 f1d1 b6c5 c1e3 d5e3 c4e3 c5e3 f2e3 g8f7 b2b4 d8d1 a1d1 f7e6 e2c4 e6e7 c4d5 f8d8 d1d2 f5e6 d5c6 d8d2 f3d2 b7c6 a4a5 e7d6 e3e4');
INSERT INTO evaluations VALUES(8271,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(8271,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(8271,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(8271,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(8271,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(8271,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(8271,36,'cp -22','e8g8 a1b1 a7a6 h2h3 b7b5 a2a3 a8b8 c1e3 b5c4 d3c4 a6a5 f1e1 f6e8 c3b5 e8c7 b2b3 c7e6 d1d2 c8d7 e2f1 f7f6 f1d3 g8h8 d3c2 d8e8 b5c3 e8f7 c3d5 d7e8 c2d3 e6g5 f3g5 f6g5 d5c7 c6d4 c7e8 f7e8');
INSERT INTO evaluations VALUES(8630,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(8630,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(8630,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(8630,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(8630,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(8630,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(8630,36,'cp 21','d3d4 e5d4 f3d4 e8g8 e1g1 c6d4 d1d4 c8e6 c1f4 d6d5 e4d5 f6d5 f4e5 e7d6 a1d1 d5c3 d4c3 d6e5 c3e5 d8e7 e2f3 c7c6 h2h3 a8d8 a2a3 d8d1 f1d1 f8d8 b2b4 d8d1 f3d1 e7d8 d1f3 h7h6 f3e4 b7b6 c2c3 e6d5 e4d5 c6d5 f2f3 a7a5');
INSERT INTO evaluations VALUES(11185,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11185,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11185,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(11185,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11185,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11185,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11185,36,'cp 0','c1f4 e7e6 e4e5 d6e5 d1d8 e8d8 e1c1 d8e8 f3e5 c6e5 f4e5 c8d7 f2f4 a8d8 f1e2 d7c6 d1d8 e8d8 h1d1 d8e8 e2f3 c6f3 g2f3 f8e7 d1d4 f6d7 c3e4 d7e5 f4e5 h7h5 f3f4 h8h6 d4c4 e8d7 c4d4 d7c7 d4c4');
INSERT INTO evaluations VALUES(11253,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(11253,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(11253,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(11253,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(11253,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(11253,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(11253,36,'cp -15','e1g1 b7b5 c4a2 c8b7 f1e1 h7h6 c3e2 g7g5 c1e3 g5g4 f3d2 h6h5 e2f4 f8h6 c2c3 h5h4 d1e2 c6e7 a3a4 h4h3 g2g3 e8f8 a2b3 e6e5 f4d5 f6d5 e4d5 f7f5 e3h6 h8h6 a4b5 a6b5 a1a8 d8a8 f2f3 h6g6 f3g4 g6g4');
INSERT INTO evaluations VALUES(11252,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(11252,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(11252,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(11252,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(11252,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(11252,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(11252,36,'cp -19','b7b5 c4a2 c8b7 f1e1 h7h6 c3e2 g7g5 c1e3 d8d7 a1c1 f6g4 e3d2 g4e5 e2g3 e8c8 f3e5 d6e5 d1h5 f8d6 c2c3 c8b8 b2b4 c5b4 a3b4 d6c7 d2e3 f7f5 e4f5 e6f5 e3c5 d7d3 c1d1');
INSERT INTO evaluations VALUES(11643,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(11643,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(11643,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(11643,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(11643,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(11643,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(11643,36,'cp 6','f8e7 e1g1 e8g8 a2a4 c8d7 b3c2 f8e8 h2h3 h7h6 f1e1 e7f8 b1d2 a8b8 a4b5 a6b5 d2f1 b8a8 c1e3 b5b4 d1e2 d6d5 f1g3 f8d6 e3d2 c6e7 c3b4 e7c6 c2a4 c6b4 a4d7 d8d7 a1a8 e8a8 e4d5 b4c2 e1c1 c2d4 f3d4');
INSERT INTO evaluations VALUES(7397,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(7397,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(7397,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(7397,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(7397,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(7397,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(7397,36,'cp 23','e1g1 e8g8 b3c2 f8e8 b1d2 d6d5 a2a4 a8b8 h2h3 e7f8 a4b5 a6b5 e4d5 f6d5 d2e4 h7h6 f1e1 c8e6 e4g3 f8d6 a1a6 d5b6 g3h5 e6c8 a6a1 c8f5 d3d4 f5g6 c2g6 f7g6 d1b3 g8h8 b3b5 e5d4 b5c6 e8e1 f3e1 g6h5 c3d4 d8f6');
INSERT INTO evaluations VALUES(7396,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(7396,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(7396,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7396,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(7396,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(7396,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(7396,36,'cp 17','e8g8 b3c2 f8e8 b1d2 a6a5 d3d4 e7f8 a2a4 b5b4 f1e1 b4c3 b2c3 e5d4 f3d4 c6e5 h2h3 c8a6 d2f1 d6d5 c1g5 d5e4 c2e4 a8b8 e4c2 d8d6 f1e3 e5d7 g5h4 g7g6 h4g3 d6b6 c2b3 f8g7 e3d5 f6d5 b3d5 e8e1 d1e1 g7d4 c3d4');
INSERT INTO evaluations VALUES(12238,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(12238,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(12238,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(12238,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(12238,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(12238,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(12238,36,'cp 31','f1d3 e7e6 e1g1 f8d6 b1c3 d5c4 d3c4 e8g8 e3e4 e6e5 c1e3 d8e7 h2h3 e5d4 e3d4 d7e5 f3e5 d6e5 f2f4 e5d4 d1d4 b7b5 c4b3 b5b4 e4e5 b4c3 e5f6 e7f6 d4f6 g7f6 b2c3 c8e6 f1e1 f8e8 g1f2 a8d8 a1d1 d8d1 e1d1 e8b8 b3e6 f7e6');
INSERT INTO evaluations VALUES(14272,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14272,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(14272,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(14272,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(14272,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(14272,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(14272,36,'cp -10','g1f3 g7g6 c1f4 f8g7 e2e3 e8g8 f1e2 f6h5 f4g5 h7h6 g5h4 c8f5 d3d2 g6g5 h4g3 e7e6 e2d3 a8c8 d3f5 e6f5 f3e5 c6e5 g3e5 g7e5 d4e5 c7c6 d2d4 f8e8 e1c1 d8e7 d4a7 e7e5 a7b7 f5f4 b7a6 f4e3 f2e3 e5e3 c1b1');
INSERT INTO evaluations VALUES(15540,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(15540,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(15540,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(15540,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(15540,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(15540,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(15540,36,'cp 59','e5d4 e3d4 g7g6 d4d5 c6e7 f1d3 f8g7 c1f4 e8g8 e1g1 f8e8 h2h3 c8f5 d1d2 f5d3 d2d3 a7a6 f1e1 h7h6 a1d1 g6g5 f4d2 d8d7 f3d4 e7g6 b2b3 e8e1 d1e1 a8e8 e1e8 d7e8 d4f5 e8e5 g2g3 f6d7 g1g2 g6e7 f5g7 e5g7 f2f3 e7g6 d3e3');
INSERT INTO evaluations VALUES(15586,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(15586,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(15586,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(15586,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(15586,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(15586,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(15586,36,'cp -12','d6d5 b1d2 a7a5 f1e2 f8c5 e1g1 e8g8 f1e1 h7h6 e2f1 f8e8 e4d5 d8d5 d2c4 a5a4 d1e2 e5e4 d3e4 d5e4 e2d2 e4g6 e1e8 f6e8 f1d3 g6f6 d2f4 f6f4 c1f4 c8e6 c4e3 c6e7 d3c2 c5d6 f4d6 e8d6');
INSERT INTO evaluations VALUES(16595,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(16595,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(16595,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(16595,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(16595,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(16595,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(16595,36,'cp -76','e5f4 c1f4 d7d5 b1d2 f8c5 c2c3 e8g8 d1c2 h7h6 e1c1 f6h5 g2g3 d5e4 d3e4 d8e7 d1e1 c5f2 e1e2 f2b6 e4e5 c8g4 e2e1 a8d8 f1d3 d8d3 c2d3 f8d8 d3e4 h5f4 e4f4 e7e6 c1b1 e6d7 b1c1 c6a5 c1b1 d7d3 b1a1 g4e6 d2e4');
INSERT INTO evaluations VALUES(17935,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(17935,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(17935,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(17935,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(17935,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(17935,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(17935,36,'cp 100','g8f6 f1e1 d7d6 c2c3 f6g4 h2h3 g4e5 d2d3 e5c4 d3c4 d4c3 b1c3 f8e7 d1d3 e8g8 c1f4 e7f6 a1d1 c8e6 c4c5 d6c5 d3g3 d8e7 f4c7 f6c3 b2c3 f8e8 c7d6 e7f6 f2f4 a8d8 d6c5 f6g6 d1d8 e8d8 g3f3 e6a2 f4f5 g6c6 c5d4 b7b5 f3g3 f7f6 e4e5 f6e5');
INSERT INTO evaluations VALUES(20045,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(20045,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(20045,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(20045,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(20045,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20045,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(20045,36,'cp -9','g7g6 d2d3 f8g7 a2a4 g8f6 e1g1 a8b8 c2c3 e8g8 c1e3 b7b6 b1a3 a7a6 d1e1 b6b5 a4b5 a6b5 h2h3 c8d7 e1f2 d8c7 a3b1 b8a8 b1d2 b5b4 d2c4 a8a1 f1a1 b4c3 b2c3 f8d8 f3d2 d8b8 e2f3 d6d5 e4e5 d5c4');
INSERT INTO evaluations VALUES(21720,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(21720,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(21720,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21720,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(21720,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(21720,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(21720,36,'cp -10','e4d5 d8d5 b1c3 d5d6 f1g2 c8g4 h2h3 g4h5 d2d3 f7f5 g3g4 f5g4 f3h2 d6d7 h3g4 h5f7 g2c6 d7c6 d1f3 g8e7 f3c6 e7c6 c1e3 e8c8 h2f3 f8b4 e1e2 f7e6 g4g5 b4c3 b2c3 e5e4 f3d4 e6g4 e2d2 c6d4 c3d4 g4f5 a1c1 c8b8 d3e4 f5e4');
INSERT INTO evaluations VALUES(21578,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(21578,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(21578,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(21578,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(21578,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(21578,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(21578,36,'cp -7','d8d5 b1c3 d5d6 f1g2 c8g4 h2h3 g4h5 g3g4 h5g6 d2d3 e8c8 f3h4 c6d4 h4g6 h7g6 c3e4 d6e7 c1g5 f7f6 g5d2 c7c6 c2c3 d4e6 d1f3 e7c7 h3h4 g8e7 h4h5 e7d5 h5g6 h8h1 g2h1 d5f4 d2f4 e6f4 e1c1 c8b8 c1b1 a7a6 b1a1 b8a8 a1b1 f8e7');
INSERT INTO evaluations VALUES(1729,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(1729,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1729,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1729,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1729,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1729,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1729,36,'cp 0','b1c3 d5d6 f1g2 c8g4 h2h3 g4h5 g3g4 h5g6 d2d3 e8c8 f3h4 c6d4 h4g6 h7g6 c3e4 d6e7 c1g5 f7f6 g5d2 c7c6 c2c3 d4e6 d1f3 e7c7 e1c1 g8e7 h3h4 e7d5 h4h5 g6h5 h1h5 h8h5 g4h5 e6f4 c3c4 d5b4 d2f4 e5f4 g2h3 c8b8 c1b1');
INSERT INTO evaluations VALUES(29046,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(29046,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(29046,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(29046,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(29046,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29046,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(29046,36,'cp 11','h2h3 g4f3 d1f3 e7e6 a2a3 b8d7 f1e2 d8b6 e1g1 f8d6 f1d1 e6e5 d4e5 d6e5 e3e4 d5d4 c3b1 e8g8 b1d2 a8e8 g2g3 e5b8 b2b4 h7h5 d2c4 b6d8 d1d4 h5h4 c1f4 h4g3 f2g3 b7b5 f4b8 d8b8 c4d6 e8e6 d6f5 c6c5 b4c5 d7c5');
INSERT INTO evaluations VALUES(31530,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(31530,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(31530,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(31530,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(31530,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(31530,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(31530,36,'cp 147','d3f5 e6f5 b1d2 g8e7 d2f3 e7g6 d1d3 d8d7 g1e2 c6c5 h2h4 c5c4 d3d1 h7h5 c1g5 b8c6 d1d2 f8e7 g5e7 c6e7 e1g1 f5f4 e2f4 d7g4 f4g6 e7g6 f1e1 g6f4 g2g3 g4f3 d2f4 f3f4 g3f4 g7g6 a2a4 e8d7 g1g2 h8b8 g2f3 a7a5 b2b3 c4b3 c2b3 b7b5 a1c1 b5a4 b3a4');
INSERT INTO evaluations VALUES(31505,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(31505,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(31505,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(31505,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(31505,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(31505,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(31505,36,'cp 144','e6f5 b1d2 g8e7 d2f3 e7g6 d1d3 d8d7 g1e2 c6c5 h2h4 c5c4 d3d1 h7h5 c1g5 b8a6 c2c3 a6c7 d1c2 f8e7 e2g3 e7g5 f3g5 g6e7 g5h3 e8c8 g3e2 e7g6 g2g3 f7f6 e5f6 g7f6 e1c1 b7b6 h3f4 g6f4 e2f4 d8e8 c1b1 c8b7 f4g2');
INSERT INTO evaluations VALUES(31580,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(31580,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(31580,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(31580,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(31580,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(31580,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(31580,36,'cp 158','g1f3 b8a6 b1c3 f8b4 e1g1 b4c3 b2c3 a6c7 d1d3 d8d7 a2a4 g8e7 c1a3 e8g8 a3e7 d7e7 d3f5 c7e6 h2h4 a8c8 f1b1 g7g6 f5f6 f8e8 a4a5 c6c5 b1b5 c8c6 f6e7 e8e7 a1b1 c6c7 g2g3 h7h5 b1b2 c5d4 c3d4 c7c3 f3g5 e6d4 b5b7 e7b7');
INSERT INTO evaluations VALUES(31579,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(31579,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(31579,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(31579,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(31579,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(31579,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(31579,36,'cp 131','b8a6 b1c3 f8b4 e1g1 b4c3 b2c3 a6c7 d1d3 d8d7 f1e1 g8e7 c1a3 c7e6 a3e7 d7e7 d3f5 e8g8 e1e3 a8c8 f5g4 g7g6 h2h4 h7h5 g4g3 b7b6 a1e1 c6c5 e1d1 c5d4 f3d4 c8c4 e3f3 e6d4 c3d4 e7e6 c2c3 e6g4 g3g4 h5g4 f3g3 b6b5');
INSERT INTO evaluations VALUES(31854,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31854,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31854,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31854,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31854,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(31854,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31854,36,'cp 0','g8f6 c2c4 e7e6 g1f3 f8e7 d1b3 e8g8 b3b7 c7c5 f1e2 b8d7 e1g1 c5d4 f3d4 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(33889,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(33889,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(33889,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33889,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33889,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33889,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(33889,36,'cp -5','f1d3 b8c6 d1e2 f6d7 h1d1 a7a6 a2a4 f7f6 d3c4 d7e5 f3e5 c6e5 b1d2 g7g5 c4b3 h8g8 d2c4 e5c4 b3c4 f8d6 b2b3 d6e5 a1a2 b7b6 c1b2 e5b2 a2b2 c8d7 a4a5 b6b5 c4d3 e6e5 e2e3 e8e7 d3e2 d7e6 f2f3 h7h5');
INSERT INTO evaluations VALUES(34114,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(34114,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(34114,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(34114,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(34114,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(34114,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(34114,36,'cp 109','g8f6 g2g3 e7e6 f1g2 c8d7 f3e5 d7c6 e5c6 b8c6 c3e2 f8e7 c2c3 e8g8 e1g1 f8e8 c1f4 d6d7 d1c2 a8d8 a1d1 h7h6 f4c1 f6d5 e2f4 e7g5 f4d3 g5c1 c2c1 b7b6 f1e1 c6e7 d3e5 d7d6 a2a3 a6a5');
INSERT INTO evaluations VALUES(38349,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(38349,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(38349,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38349,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38349,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(38349,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(38349,36,'cp 25','g1f3 g8f6 d2d4 e8g8 c4b3 c8g4 h2h3 g4f3 d1f3 b8c6 f3d1 e7e5 d4e5 d6e5 d1d8 f8d8 c1g5 c6d4 f2f3 h7h6 g5e3 f6h5 e1f2 c7c6 g2g3 g7f8 h1d1 f8c5 c3e2 a7a5 c2c3 d4b3 a2b3 c5e3 f2e3 b7b6 g3g4 h5f6 h3h4 d8d7 d1d7 f6d7 b3b4 a5a4');
INSERT INTO evaluations VALUES(41914,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(41914,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(41914,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(41914,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(41914,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(41914,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(41914,36,'cp 42','e5c4 b4c3 d2c3 f6e4 f1e2 e8g8 e1g1 b8c6 f1e1 c8e6 f2f3 e4c5 a2a4 d8h4 c4e3 a7a5 b2b3 a8e8 e2b5 e6d7 c1a3 h4g5 h2h4 g5h6 a3c5 d6c5 d1d7 e8e3 d7d2 f8e8 e1e3 h6e3 d2e3 e8e3 a1d1 e3c3 d1d7 g8f8 d7c7 c3c2 c7b7');
INSERT INTO evaluations VALUES(42076,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(42076,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(42076,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(42076,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(42076,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(42076,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(42076,36,'cp 51','d2c3 e8g8 g1f3 b7b6 b2b3 c8b7 a1c1 d8e7 f1d3 b8d7 e1g1 c7c5 d1e2 a8c8 f1d1 f8d8 c4d5 c5d4 f3d4 f6d5 c3b2 c8c1 d1c1 d7c5 d3c4 e7g5 d4f3 g5e7 a2a3 a7a6 b3b4 c5a4 b2d4 d8c8 e2e1 b6b5 c4b3 c8c1 e1c1');
INSERT INTO evaluations VALUES(42417,30,'cp 303',NULL);
INSERT INTO evaluations VALUES(42417,31,'cp 311',NULL);
INSERT INTO evaluations VALUES(42417,32,'cp 321',NULL);
INSERT INTO evaluations VALUES(42417,33,'cp 300',NULL);
INSERT INTO evaluations VALUES(42417,34,'cp 311',NULL);
INSERT INTO evaluations VALUES(42417,35,'cp 330',NULL);
INSERT INTO evaluations VALUES(42417,36,'cp 314','d1g4 e4c3 g4g7 c3d5 c2c3 h8f8 c3b4 d8e7 a2a3 f7f6 g7e7 d5e7 e5f6 f8f6 c1g5 f6g6 h2h4 h7h6 g5d2 b8c6 h4h5 g6f6 g1f3 a7a6 f3h2 c6d4 h2g4 f6f7 a1c1 d4b3 g4h6 f7f8 c1d1 b3d2 d1d2 d7d6 f1d3 a6a5 b4b5 c8d7 h6g4 e6e5 f2f3 e8f7 g4e3 f7g7 d3e4 d7b5');
INSERT INTO evaluations VALUES(45432,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(45432,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(45432,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(45432,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45432,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(45432,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(45432,36,'cp 25','c2c3 b8c6 g1f3 a7a6 f1d3 c8g4 h2h3 g4h5 b1d2 e7e6 e1g1 f8d6 f4d6 d8d6 f1e1 a8c8 d1e2 d6f4 a2a4 e8g8 g2g3 f4c7 e2e3 f8e8 g1g2 f6d7 e3g5 h5g6 d3g6 h7g6 h3h4 e6e5 d4e5 c6e5 g5f4 c7b6 f3e5 d7e5');
INSERT INTO evaluations VALUES(45426,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(45426,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(45426,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(45426,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(45426,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(45426,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(45426,36,'cp 26','b8c6 g1f3 a7a6 f1d3 c8g4 h2h3 g4h5 b1d2 e7e6 d1e2 f8d6 f4d6 d8d6 e1g1 d6f4 a1e1 e8g8 g2g3 f4c7 e2e3 c7b6 d2b3 h5f3 e3f3 f8e8 f3f4 a8c8 e1e2 b6c7 f4c7 c8c7 f2f4 b7b6 a2a4 c6e7 g1f2 e8c8 a4a5');
INSERT INTO evaluations VALUES(13136,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(13136,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(13136,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13136,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(13136,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(13136,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(13136,36,'cp 25','g1f3 c8f5 d1b3 d8d7 f3e5 d7c8 f1b5 a7a6 e5c6 b7c6 b5a4 a6a5 c3c4 e7e6 e1g1 f8e7 c4c5 e8g8 b1d2 c8a6 h2h3 f8c8 f1e1 a8a7 b3e3 a7b7 b2b3 a6d3 d2f3 f6d7 a1c1 d3e3 e1e3 f5e4 f3d2 e4f5 e3e1');
INSERT INTO evaluations VALUES(13134,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(13134,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(13134,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(13134,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(13134,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(13134,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(13134,36,'cp 21','c8f5 d1b3 d8c8 f3h4 f5d7 h2h3 h7h6 f4h2 e7e6 h4f3 c6a5 b3c2 a7a6 a2a4 f8e7 b1d2 b7b5 f1d3 a5c4 a4b5 d7b5 d2c4 d5c4 d3e2 e8g8 e1g1 a6a5 f1e1 f6d5 f3e5 e7d6 e2f3 c8c7 f3d5 e6d5 e5g4 b5d7 h2d6 c7d6');
INSERT INTO evaluations VALUES(45710,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(45710,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(45710,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(45710,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(45710,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(45710,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(45710,36,'cp 88','f1c4 c6c5 e1g1 a7a6 c4b3 c5d4 e3d4 f8e7 d1e2 b8c6 f1d1 e8g8 d4d5 e6d5 c3d5 f6d5 b3d5 d8c7 e2e4 e7f6 f3g5 f6g5 c1g5 c7e5 e4e5 c6e5 g5e3 e5c6 a1c1 c8g4 d1d2 g4e6 d5e6 f7e6 d2d7 f8f7 d7d6 a8e8');
INSERT INTO evaluations VALUES(46775,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46775,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46775,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46775,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46775,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46775,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46775,36,'cp 0','c7c5 e2e3 b8c6 f1e2 f8g7 e1g1 e8g8 c2c3 b7b6 b1d2 c8b7 a2a4 e7e6 e2b5 d8e7 b5c6 b7c6 f3e5 f8c8 e5c6 c8c6 d1e2 a7a5 f4h2 e7b7 f1e1 c6c8 d2b1 b7c6 b1d2');
INSERT INTO evaluations VALUES(49145,30,'cp 519',NULL);
INSERT INTO evaluations VALUES(49145,31,'cp 494',NULL);
INSERT INTO evaluations VALUES(49145,32,'cp 526',NULL);
INSERT INTO evaluations VALUES(49145,33,'cp 531',NULL);
INSERT INTO evaluations VALUES(49145,34,'cp 537',NULL);
INSERT INTO evaluations VALUES(49145,35,'cp 555',NULL);
INSERT INTO evaluations VALUES(49145,36,'cp 538','d3e4 f6e5 e2h5 e8d7 h5e5 d8f6 e5d5 d7e8 c1g5 f6d6 b1c3 c7c6 d5b3 b8d7 a1d1 d6c5 g5e3 c5b4 e3d4 b4b3 c2b3 a7a5 f2f4 f8b4 f1e2 b4c5 e1g1 c5d4 d1d4 e8e7 e4e5 g7g6 c3e4 d7b6 e4d6 h8d8 f1d1 c8e6 g1f2 a5a4 d4b4 a4b3 b4b6 b3a2');
INSERT INTO evaluations VALUES(49144,30,'cp 516',NULL);
INSERT INTO evaluations VALUES(49144,31,'cp 534',NULL);
INSERT INTO evaluations VALUES(49144,32,'cp 532',NULL);
INSERT INTO evaluations VALUES(49144,33,'cp 521',NULL);
INSERT INTO evaluations VALUES(49144,34,'cp 525',NULL);
INSERT INTO evaluations VALUES(49144,35,'cp 524',NULL);
INSERT INTO evaluations VALUES(49144,36,'cp 542','f6e5 e2h5 g7g6 h5e5 e8f7 b1c3 f8d6 e5d5 c8e6 d5b7 b8d7 c3b5 h8e8 c1e3 d6e5 f1e2 a8b8 b7a7 f7g8 a1d1 d8c8 a7a5 b8a8 a5d2 d7f6 f2f3 c8b8 b2b3 a8a2 b5c3 a2a8 e3f4 c7c6 f4e5 b8e5 d2e3 e5c7 e3d4 f6d7 e1g1 e8b8 g1h1');
INSERT INTO evaluations VALUES(49148,30,'cp 559',NULL);
INSERT INTO evaluations VALUES(49148,31,'cp 542',NULL);
INSERT INTO evaluations VALUES(49148,32,'cp 532',NULL);
INSERT INTO evaluations VALUES(49148,33,'cp 551',NULL);
INSERT INTO evaluations VALUES(49148,34,'cp 550',NULL);
INSERT INTO evaluations VALUES(49148,35,'cp 540',NULL);
INSERT INTO evaluations VALUES(49148,36,'cp 563','e2h5 e8d7 h5e5 d8f6 e5d5 d7e8 c1g5 f6d6 b1c3 c7c6 d5b3 f8e7 a1d1 d6e5 f2f4 e5c7 g5e7 c7e7 f1e2 c8e6 e2c4 e6c4 b3c4 b8d7 e4e5 a8d8 c3e4 d7b6 d1d8 e8d8 c4e2 d8c8 e1g1 h8d8 c2c4 d8d4 b2b3 d4e4 e2e4 c8c7 e4h7');
INSERT INTO evaluations VALUES(50081,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(50081,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(50081,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(50081,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(50081,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(50081,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50081,36,'cp 29','d7d6 b1c3 e7e5 g1f3 f8e7 b2b4 e8g8 d2d3 a7a6 e1g1 d8e8 a2a4 b8c6 a1b1 a6a5 b4a5 e7d8 e2e3 g8h8 c1a3 h7h6 f3d2 c6a5 f1e1 c8d7 c4c5 d6c5 a3c5 d8e7 c5e7 e8e7 d2f3 e5e4 d3e4 f5e4 f3d2 d7f5 d1c2');
INSERT INTO evaluations VALUES(50166,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(50166,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(50166,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(50166,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(50166,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(50166,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(50166,36,'cp 82','e4e5 d7d5 c4b3 f6d7 d2d4 c7c5 c2c3 b8c6 g1e2 d7b6 e1g1 c5d4 c3d4 f8g7 h2h3 c8d7 c1g5 f7f6 e5f6 e7f6 g5f4 c6a5 b3c2 e8g8 b1c3 a8c8 b2b3 a5c6 a1c1 f8e8 c2d3 g7f8 f4e3 c6b4 d3b1 f6f5 e2f4');
INSERT INTO evaluations VALUES(50538,30,'cp -414',NULL);
INSERT INTO evaluations VALUES(50538,31,'cp -400',NULL);
INSERT INTO evaluations VALUES(50538,32,'cp -401',NULL);
INSERT INTO evaluations VALUES(50538,33,'cp -424',NULL);
INSERT INTO evaluations VALUES(50538,34,'cp -418',NULL);
INSERT INTO evaluations VALUES(50538,35,'cp -419',NULL);
INSERT INTO evaluations VALUES(50538,36,'cp -417','b7b5 a4b3 c5c4 b3c4 b5c4 e1g1 c8b7 d2d3 f8e7 f1e1 c4d3 c2d3 g8f6 c1f4 b8c6 b1c3 e8g8 a1c1 g8h8 d1b3 c6a5 b3a4 a8c8 a2a3 d7d6 b2b4 a5c6 a4b3 c6e5 f3e5 d6e5 f4e5 d8d3 e1d1 d3c4 b3c4 c8c4 c3a4 c4c1 d1c1 b7e4');
INSERT INTO evaluations VALUES(51225,30,'cp -414',NULL);
INSERT INTO evaluations VALUES(51225,31,'cp -414',NULL);
INSERT INTO evaluations VALUES(51225,32,'cp -425',NULL);
INSERT INTO evaluations VALUES(51225,33,'cp -421',NULL);
INSERT INTO evaluations VALUES(51225,34,'cp -399',NULL);
INSERT INTO evaluations VALUES(51225,35,'cp -421',NULL);
INSERT INTO evaluations VALUES(51225,36,'cp -418','a4b3 c5c4 b3c4 b5c4 e1g1 c8b7 b1c3 g8f6 f1e1 f8e7 b2b3 c4b3 a2b3 e8g8 h2h3 g8h8 e4e5 f6g8 d2d4 e7b4 c3a4 f7f6 c2c3 b4e7 e5f6 f8f6 f3g5 f6f5 g5e4 b7e4 e1e4 g8f6 e4e3 h7h6 c3c4 b8c6 d4d5 e6d5');
INSERT INTO evaluations VALUES(51211,30,'cp -437',NULL);
INSERT INTO evaluations VALUES(51211,31,'cp -444',NULL);
INSERT INTO evaluations VALUES(51211,32,'cp -438',NULL);
INSERT INTO evaluations VALUES(51211,33,'cp -421',NULL);
INSERT INTO evaluations VALUES(51211,34,'cp -426',NULL);
INSERT INTO evaluations VALUES(51211,35,'cp -424',NULL);
INSERT INTO evaluations VALUES(51211,36,'cp -440','c5c4 b3c4 b5c4 e1g1 c8b7 d1e2 g8f6 b1c3 d7d5 e4e5 f6d7 b2b3 c4b3 a2b3 f8e7 d2d4 e8g8 e2e3 a6a5 f1e1 b8c6 c3b5 c6b4 e3d2 a8c8 c2c3 b7a6 b5d6 e7d6 e5d6 b4d3 e1e3 d3c1 d2c1 c8b8');
INSERT INTO evaluations VALUES(51179,30,'cp -440',NULL);
INSERT INTO evaluations VALUES(51179,31,'cp -444',NULL);
INSERT INTO evaluations VALUES(51179,32,'cp -427',NULL);
INSERT INTO evaluations VALUES(51179,33,'cp -439',NULL);
INSERT INTO evaluations VALUES(51179,34,'cp -410',NULL);
INSERT INTO evaluations VALUES(51179,35,'cp -416',NULL);
INSERT INTO evaluations VALUES(51179,36,'cp -440','d2d4 c4b3 a2b3 c8b7 d1e2 g8f6 e4e5 f6e4 e1g1 f8e7 c2c4 e8g8 b1c3 e4c3 b2c3 b7f3 g2f3 b5c4 b3c4 g8h8 g1h1 d7d5 c4c5 a6a5 f3f4 b8c6 f4f5 e6f5 f1g1 f7f6 e5e6 a5a4 c1f4 g7g5 e2b5 d8e8 f4d6 a4a3 d6e7 c6e7 b5e8 f8e8');
INSERT INTO evaluations VALUES(50811,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(50811,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(50811,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(50811,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50811,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(50811,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(50811,36,'cp 66','c2c4 e7e6 b1c3 d5c4 f1c4 g8f6 g1f3 b8d7 c4e2 f8d6 e1g1 e8g8 f3e5 f6d5 a1c1 d5f4 e3f4 d8e7 c3e4 d7f6 e4f6 e7f6 g2g3 f6e7 d1b3 a6a5 h2h4 a5a4 b3e3 c8d7 e2f3 d7b5 f1d1 c7c6 g1g2 a4a3 b2b3 a8d8 h4h5');
INSERT INTO evaluations VALUES(51571,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(51571,31,'cp 262',NULL);
INSERT INTO evaluations VALUES(51571,32,'cp 305',NULL);
INSERT INTO evaluations VALUES(51571,33,'cp 298',NULL);
INSERT INTO evaluations VALUES(51571,34,'cp 311',NULL);
INSERT INTO evaluations VALUES(51571,35,'cp 311',NULL);
INSERT INTO evaluations VALUES(51571,36,'cp 310','c4b3 d7d6 c3b5 c6b5 b3d5 b8a6 d5a8 g8e7 a2a4 b5b4 c2c4 b4c3 d2c3 a6c7 a8d5 c7d5 e4d5 c8a6 b2b3 d8a5 b3b4 a5d5 b4b5 d5d1 e1d1 a6b7 c1e3 a7a6 d1c2 e8d7 b5a6 b7a6 h1b1 d7c7 f3d2 e7d5 c3c4 d5e3 f2e3 f6f5 b1f1 g7g6 e3e4 f5f4 g2g3 g6g5 c2d3 f8e7 a4a5 f4g3 h2g3 c7d7');
INSERT INTO evaluations VALUES(51573,30,'cp 321',NULL);
INSERT INTO evaluations VALUES(51573,31,'cp 311',NULL);
INSERT INTO evaluations VALUES(51573,32,'cp 305',NULL);
INSERT INTO evaluations VALUES(51573,33,'cp 293',NULL);
INSERT INTO evaluations VALUES(51573,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(51573,35,'cp 295',NULL);
INSERT INTO evaluations VALUES(51573,36,'cp 287','d7d6 c3b5 c6b5 b3d5 b8d7 d5a8 g8e7 a8d5 e7d5 e4d5 f8e7 e1g1 d7b6 a2a4 b5a4 d2d3 b6d5 a1a4 a7a6 c2c4 d5b6 a4a1 e8g8 d1b3 g8h8 f1e1 f6f5 c1g5 b6d7 g5e7 d8e7 d3d4 e5e4 f3d2 d7f6 b3c2 e7e8 h2h3 e8g6 f2f3 c8b7 f3e4');
INSERT INTO evaluations VALUES(52233,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(52233,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(52233,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(52233,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(52233,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(52233,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(52233,36,'cp 203','d2d4 d7d5 b1c3 c8g4 h2h3 g4h5 f1d3 e7e6 d1e2 f8b4 e4d5 c6d5 e2e6 g8e7 c1d2 b8c6 e1c1 h5f3 g2f3 d8d7 e6g4 d7g4 f3g4 c6d4 a2a3 b4c3 d2c3 d4c6 g4g5 d5d4 c3d2 f6g5 d2g5 h7h6 g5h4 e8g8 h4e7 c6e7 d3c4 g8h8 d1d4 e7c6 d4d2 a8e8 c4d5');
INSERT INTO evaluations VALUES(52231,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(52231,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(52231,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(52231,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(52231,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(52231,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(52231,36,'cp 205','d7d5 b1c3 c8e6 f1d3 e6f7 h2h4 e7e6 d1e2 f8b4 c1d2 g8e7 h4h5 b8d7 h5h6 g7g5 e1g1 a7a6 a1e1 b4a5 e4e5 g5g4 e5f6 g4f3 f6e7 f3e2 e7d8q a8d8 e1e2 a5b6 d2e3 e8g8 c3a4 b6c7 g2g4 b7b5 a4c5 d7c5 d4c5');
INSERT INTO evaluations VALUES(53709,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53709,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53709,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53709,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53709,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(53709,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(53709,36,'cp 0','f8b4 b1c3 g8f6 f1d3 d5c4 d3c4 e8g8 e1g1 c8g4 c1g5 g4f3 d1f3 d8d4 g5f6 d4f6 f3f6 g7f6 c3e4 f6f5 e4g3 f5f4 g3h5 b8d7 h5f4 f8e8 a1d1 d7e5 c4b3 a8d8 f4h5 e8f8 h5f6 g8g7 f6h5 g7g8');
INSERT INTO evaluations VALUES(54914,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(54914,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(54914,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(54914,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(54914,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(54914,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(54914,36,'cp 32','e7e6 c4d5 e6d5 c1f4 c7c6 e2e3 f8d6 f4d6 d8d6 f1d3 g8e7 d1b3 b7b6 g1f3 e8g8 a1c1 c8g4 f3e5 g4e6 e1g1 b8d7 e5d7 e6d7 c3e2 a8c8 c1c2 f8e8 h2h3 h6h5 f1c1 g7g6 a2a4 d7f5 e2f4 h5h4 b3c3 f5d3 f4d3 e7f5');
INSERT INTO evaluations VALUES(55948,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55948,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(55948,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(55948,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(55948,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(55948,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(55948,36,'cp -12','g8f6 g1f3 f8e7 e1g1 c7c5 b2b3 b8c6 c1b2 b7b5 c2c4 b5c4 b3c4 e8g8 d2d3 d8b6 f3e5 c8b7 d1b3 b6c7 b1d2 a8b8 e5c6 b7c6 b3c2 f8d8 b2f6 e7f6 a1b1 b8b1 d2b1 c7d7 b1d2 c6a4 d2b3 d5c4 c2c4 a4b3 a2b3 d7d3 c4d3 d8d3');
INSERT INTO evaluations VALUES(48417,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48417,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(48417,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48417,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48417,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(48417,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(48417,36,'cp 0','g1f3 c7c5 e1g1 f8e7 b2b3 b8c6 c1b2 b7b5 c2c4 b5c4 b3c4 e8g8 d2d3 d8b6 f3e5 c8b7 d1b3 c6a5 b3b6 a7b6 a2a4 a5c6 e5c6 b7c6 b1c3 a8a5 c3b5 f8a8 b5c7 a8a7 c7b5 a7a6 b5c7');
INSERT INTO evaluations VALUES(2610,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(2610,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(2610,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(2610,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(2610,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(2610,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(2610,36,'cp 6','c3d5 h7h6 c2c3 c5a7 d5f6 d8f6 c4b3 c6e7 c1e3 e7g6 d3d4 f8e8 f1e1 e5d4 e3d4 a7d4 d1d4 f6d4 c3d4 c8d7 e1e3 a6a5 b3c2 g6f4 d4d5 c7c6 d5c6 b7c6 c2a4 d6d5 e4d5 f4d5 e3e8 d7e8 a4b3 a5a4');
INSERT INTO evaluations VALUES(4455,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(4455,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(4455,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(4455,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(4455,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(4455,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(4455,36,'cp 63','c1g5 d7d6 f3d2 h7h6 g5h4 a7a6 f2f3 c6b8 d1e2 b7b5 c4d3 b8d7 d2f1 d7c5 f1e3 a8b8 e1g1 c5e6 f1e1 e6f4 e2d2 c8e6 a1d1 f4g6 h4f2 f6d7 d3f1 g8h8 b2b3 g6f4 g2g3 f4h3 f1h3 e6h3');
INSERT INTO evaluations VALUES(4421,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(4421,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(4421,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(4421,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(4421,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(4421,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(4421,36,'cp 58','d7d6 f3d2 h7h6 g5h4 d8e7 f2f3 c8e6 c4d3 c6b8 d2c4 b8d7 c4e3 c7c6 d1d2 b7b5 e1g1 a7a5 a1d1 f8e8 f1e1 d7b6 b2b3 e8d8 d2e2 a5a4 b3b4 a8c8 d1d2 d8e8 e1d1 e8d8 g1h1 g7g5 h4f2 f6h5 e3f5 e6f5 f2b6');
INSERT INTO evaluations VALUES(4620,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(4620,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(4620,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(4620,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(4620,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(4620,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(4620,36,'cp 83','c1f4 d7d6 e1c1 c8e6 c1b1 f6g4 c3d5 e7f6 d4d2 b7b5 c4b3 a7a5 h2h3 g4e5 d5f6 d8f6 f4g5 f6g6 f2f4 h7h6 f4e5 e6b3 a2b3 h6g5 e5d6 c7d6 h1e1 f8d8 e1e3 g6e6 d2d5 e6e5 e3f3 e5d5 d1d5 b5b4 d5g5 d8e8 f3f4 e8e5 f4g4 g7g6 h3h4 a8e8 h4h5 e5g5 g4g5 e8e4 g5a5 e4e2');
INSERT INTO evaluations VALUES(9027,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(9027,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(9027,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(9027,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(9027,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(9027,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(9027,36,'cp 74','h7h6 g5f6 d8f6 e1g1 e8g8 e4e5 f6d8 c3e4 c8d7 a2a3 b4a5 d1d3 c6e7 b2b4 a5b6 c4b3 a7a6 b3c2 e7g6 a3a4 d8e7 h2h4 f8d8 h4h5 g6f8 f1d1 b6a7 d3e3 d7c6 a1b1 f8h7 d1d2 d8d7 e4c5 a7c5 d4c5 d7d2 f3d2 a8d8 b4b5 a6b5 a4b5');
INSERT INTO evaluations VALUES(8792,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(8792,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(8792,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(8792,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(8792,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(8792,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(8792,36,'cp 87','g5f6 d8f6 e1g1 e8g8 d1d3 c6a5 c4b5 a7a6 b5a4 b7b5 a4c2 c8b7 a2a3 b4e7 b2b4 a5c4 c2b3 f6g6 b3c4 b5c4 d3c4 b7e4 c3e4 g6e4 f1e1 e4b7 f3e5 b7d5 a1c1 e7d6 e5c6 d5c4 c1c4 g7g5 h2h4 g5h4 e1e4 h4h3 g2g3 f7f5 e4e6 f5f4 c6e5 f4g3 f2g3 d6e5 d4e5 f8f3 e6h6 f3g3 g1h2');
INSERT INTO evaluations VALUES(8786,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(8786,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(8786,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(8786,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(8786,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(8786,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(8786,36,'cp 83','d8f6 e1g1 b4c3 b2c3 e8g8 e4e5 f6d8 d1e2 a8b8 e2e4 c6e7 c4d3 e7g6 e4g4 g6e7 c3c4 c7c5 g4e4 e7g6 d4c5 c8d7 e4e3 d8e7 a1d1 d7a4 d1d2 a4c6 f3d4 e7c5 d4e6 c5e3 f2e3 f7e6 d3g6 f8f1 g1f1 g8f8 d2f2 f8g8 f2f4 c6e8 g6d3 e8c6 g2g3 b8d8 f1e2 d8c8 h2h4');
INSERT INTO evaluations VALUES(497,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(497,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(497,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(497,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(497,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(497,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(497,36,'cp 85','e1g1 e8g8 d1d3 c6a5 c4b5 a7a6 b5a4 b7b5 a4c2 c8b7 a2a3 b4e7 b2b4 a5c4 c2b3 c4b6 a1c1 f8d8 d3e3 e7f8 h2h3 b6c4 b3c4 b5c4 c3a4 b7c6 a4c5 a6a5 f3e5 c6b5 f1d1 a5b4 a3b4 f6g5 e5c4 g5e3 f2e3 d8b8 e4e5 a8a2 c4d2 f8e7');
INSERT INTO evaluations VALUES(9449,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9449,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9449,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9449,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9449,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(9449,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9449,36,'cp 0','f1e2 e8g8 h2h3 c6e7 d1d2 c7c6 a2a4 a8b8 e1g1 c8d7 d3d4 e7g6 e2d3 d8e7 d3c4 d7e6 c4d3 e6d7');
INSERT INTO evaluations VALUES(10055,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(10055,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(10055,32,'cp -166',NULL);
INSERT INTO evaluations VALUES(10055,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(10055,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(10055,35,'cp -143',NULL);
INSERT INTO evaluations VALUES(10055,36,'cp -138','c6d4 f3d1 h7h6 g5h4 d7d5 e4d5 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 d8d5 g1f3 g7g5 h4g3 e5e4 d3e4 d5e4 d1e2 c8b7 g3c7 e4e2 e1e2 g5g4 f3d4 e8g8 f2f3 f8e8 e2d2 f6d5 c7g3 d5b4 h1e1 e8e1 g3e1 c5d4 c3d4 g4f3 g2f3 b7f3');
INSERT INTO evaluations VALUES(12106,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(12106,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(12106,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(12106,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(12106,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(12106,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(12106,36,'cp -60','c5d4 f3d4 d7e5 f2f4 a7a6 f4e5 a6b5 c3b5 f8e7 e1g1 e8g8 d4f3 c8d7 a2a4 d8b8 e3f4 e7c5 g1h1 c6e7 d1d3 e7f5 c2c3 d7b5 d3b5 c5e3 f4e3 f5e3 f1f2 e3c4 f2e2 b8c7 b2b3 c4a5 e2e3 a8a6 a1f1 h7h6 f3d4 a5c6 b5e2 c6d4 c3d4 a6c6 e3f3 c6c1 e2d2 c1f1 f3f1 f8c8 d2f4 c7d7 f4g4 c8c2');
INSERT INTO evaluations VALUES(12104,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(12104,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(12104,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(12104,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(12104,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(12104,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(12104,36,'cp -64','f3d4 d7e5 f2f4 a7a6 f4e5 a6b5 c3b5 f8e7 e1g1 e8g8 d4f3 c8d7 a2a4 d8b8 e3f4 e7c5 g1h1 c6e7 d1d3 h7h6 c2c3 e7g6 f4g3 a8a5 h2h4 d7b5 a4b5 a5a1 f1a1 g6e7 h4h5 b8e8 f3h4 e8d7 h1h2 f8b8 a1a4 b8c8 d3e2 e7f5 h4f5 e6f5 g3f2 c5f2 e2f2 d7e6');
INSERT INTO evaluations VALUES(13210,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13210,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13210,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13210,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13210,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13210,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13210,36,'cp 0','c8d7 d4d3 g7g6 e1g1 f8g7 h2h3 a7a6 b5c4 b7b5 c4b3 e8g8 f1e1 c6a5 c1g5 a5b3 a2b3 h7h6 g5f6 g7f6 d3e3 f6g7 c3d5 g8h7 d5b6 a8a7 b6d5');
INSERT INTO evaluations VALUES(24472,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24472,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24472,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24472,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24472,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24472,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24472,36,'cp 0','d4d3 g7g6 e1g1 f8g7 c3d5 e8g8 c2c3 a7a6 d5f6 g7f6 b5c6 d7c6 c1h6 f8e8 f1e1 d8a5 h2h3 a8c8 f3d4 c6a4 d3d2 c8d8 d2e2 a5b6 e2d2');
INSERT INTO evaluations VALUES(15136,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(15136,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(15136,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(15136,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(15136,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(15136,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(15136,36,'cp -39','d5d4 c3b1 a7a5 c2c3 f8e7 e1g1 e8g8 c3d4 e5d4 b1a3 f6d7 c1f4 d7c5 f4g3 a5a4 a3b5 c5e6 b5a3 e7d6 d1d2 f8e8 g3d6 d8d6 g2g3 e6c5 f3h4 c8h3 f1e1 d6d7 f2f4 b7b6 h4f3 f7f6 a3c2 a8c8 e2f1 h3f1 e1f1 g8h8');
INSERT INTO evaluations VALUES(15217,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(15217,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(15217,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(15217,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(15217,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(15217,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(15217,36,'cp -50','c3b1 f8e7 e1g1 e8g8 c2c3 a7a5 c3d4 e5d4 b1a3 f6d7 c1f4 d7c5 f4g3 a5a4 a3b5 c5e6 b5a3 c6b4 f3e5 f7f6 e5c4 c8d7 a3c2 b4c2 d1c2 b7b5 c4d2 c7c5 f2f4 d8b6 g1h1 e6d8 a2a3 d8c6 a1e1 a8c8 e2f3 c6a5 f4f5 c5c4');
INSERT INTO evaluations VALUES(17724,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(17724,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(17724,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(17724,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(17724,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(17724,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(17724,36,'cp 46','f7f6 f1e2 f6e5 f3e5 c6e5 d4e5 c8f5 h2h4 e7e6 g2g4 f5e4 f2f3 e4g6 h4h5 g6f7 c2c3 c7c5 b1d2 f8e7 d2b3 d8c7 f3f4 g8h6 c1d2 e8g8 e1g1 a6a5 a2a4 a8d8 g1h2 b7b6 h2h3 f7e8 d2e3 e8c6 d1c2 g8h8 b3d2');
INSERT INTO evaluations VALUES(21047,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(21047,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(21047,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(21047,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(21047,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(21047,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(21047,36,'cp 150','c3d5 f7f5 f1b5 g8f6 d1f3 f6e4 b5c6 b7c6 d5c3 d8d6 c3e4 f5e4 f3e4 c8e6 g1f3 e6d5 e4g4 h7h5 g4h3 h5h4 d2d3 d5e6 g2g4 d6b4 c1d2 e6g4 d2b4 g4h3 h1g1 h3e6 f3e5 h8h5 e5g6 e6d5 b4c3 e7e5 c3e5 e8c8 e1d2 f8c5 d3d4 c5b6 f2f4');
INSERT INTO evaluations VALUES(27033,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(27033,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(27033,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(27033,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(27033,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(27033,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(27033,36,'cp 94','c1f4 g8f6 b1c3 f8e7 e1g1 e8g8 g1h1 f6e4 c3e4 d6d5 f4c7 d8c7 c4d5 a8d8 c2c4 c7b6 a1b1 g4f3 d1f3 b6d4 d5f7 g8h8 f3e2 c6b4 a2a3 d4d3 e2d3 b4d3 g2g3 d3e5 f7e6 g7g6 f1d1 d8d3 e6d5 b7b6 d1d3 e5d3 b2b4 d3f2 h1g2 f2e4 d5e4 f8d8 e4d5 e7f6 b1e1 f6c3');
INSERT INTO evaluations VALUES(27031,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(27031,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(27031,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(27031,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(27031,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(27031,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(27031,36,'cp 61','g8f6 e1g1 f8e7 b1c3 e8g8 g1h1 f6e4 c3e4 d6d5 f4c7 d8c7 c4d5 a8d8 c2c4 c7b6 a1b1 g4f3 d1f3 b6d4 d5f7 g8h8 f3e2 c6b4 a2a3 d4d3 e2d3 b4d3 f7d5 f8f1 b1f1 d3b2 g2g3 b2d3 d5b7 e7a3 f1a1 a3c5 e4c5 d3c5 b7c6 g7g6 a1a7 d8d4 c6d5 d4d2 h1g1 c5d3 a7e7 d3b4 d5e4 d2d4 c4c5');
INSERT INTO evaluations VALUES(29512,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(29512,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(29512,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(29512,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(29512,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(29512,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(29512,36,'cp 56','d8d5 e1g1 b8c6 h2h3 g4f3 e2f3 d5d7 c2c3 e8c8 d1e2 g7g5 d2d4 h8g8 b1a3 h7h5 a3c4 g5g4 f3c6 d7c6 h3h4 c6e4 f1e1 e4e2 e1e2 f6d5 g2g3 f8g7 g1f1 d8d7 a2a4 g8h8 e2e1 b7b6 a4a5 c8b7 c4e5 g7e5 e1e5 e7e6 c1d2 b6b5');
INSERT INTO evaluations VALUES(32252,30,'cp 295',NULL);
INSERT INTO evaluations VALUES(32252,31,'cp 317',NULL);
INSERT INTO evaluations VALUES(32252,32,'cp 353',NULL);
INSERT INTO evaluations VALUES(32252,33,'cp 333',NULL);
INSERT INTO evaluations VALUES(32252,34,'cp 337',NULL);
INSERT INTO evaluations VALUES(32252,35,'cp 332',NULL);
INSERT INTO evaluations VALUES(32252,36,'cp 353','c4e6 f7e6 d1h5 g7g6 e5g6 g8f6 h5h3 h8g8 g6f8 e8f8 b1c3 e6e5 b2b3 d8d7 h3d7 b8d7 e1g1 f8f7 c1b2 a8d8 a1d1 a7a5 f2f4 e4f3 f1f3 c7c5 d1e1 f7e6 c3b5 a5a4 f3f2 f6g4 f2e2 a4b3 a2b3 e6d5 h2h3 d5c6 c2c4 g4h6 e1f1 g8g6 d2d3 d8e8 d3d4 c5d4 b5d4 c6b6');
INSERT INTO evaluations VALUES(32236,30,'cp 354',NULL);
INSERT INTO evaluations VALUES(32236,31,'cp 371',NULL);
INSERT INTO evaluations VALUES(32236,32,'cp 367',NULL);
INSERT INTO evaluations VALUES(32236,33,'cp 374',NULL);
INSERT INTO evaluations VALUES(32236,34,'cp 364',NULL);
INSERT INTO evaluations VALUES(32236,35,'cp 356',NULL);
INSERT INTO evaluations VALUES(32236,36,'cp 370','f7e6 d1h5 g7g6 e5g6 g8f6 h5h3 h8g8 g6f8 e8f8 b1c3 d8d7 b2b3 e6e5 h3d7 b8d7 e1g1 f8f7 c1b2 a8d8 a1d1 a7a5 f2f4 e4f3 f1f3 g8g6 d1e1 f7g8 d2d3 d8a8 h2h3 b7b6 g1h2 a8f8 c3d1 f6d5 f3f8 g8f8 b2e5 d7e5 e1e5 d5b4 e5f5 f8e7 f5h5 b4c2 h5h7 e7d6 h7h5');
INSERT INTO evaluations VALUES(32456,30,'cp 373',NULL);
INSERT INTO evaluations VALUES(32456,31,'cp 368',NULL);
INSERT INTO evaluations VALUES(32456,32,'cp 377',NULL);
INSERT INTO evaluations VALUES(32456,33,'cp 371',NULL);
INSERT INTO evaluations VALUES(32456,34,'cp 357',NULL);
INSERT INTO evaluations VALUES(32456,35,'cp 374',NULL);
INSERT INTO evaluations VALUES(32456,36,'cp 347','d1h5 g7g6 e5g6 g8f6 h5h3 h8g8 g6f8 e8f8 b1c3 b8c6 b2b3 d8d4 c1a3 f8f7 e1c1 a8d8 a3b2 d4e5 f2f3 e5h5 h3h5 f6h5 g2g3 e4f3 h1f1 c6d4 c3e4 f7e7 c2c3 d4b5 f1f3 d8f8 f3f8 g8f8 d1e1 b5d6 c3c4 d6e4 e1e4 f8f1 c1c2 e7f7 e4h4 f7g6 a2a3 b7b6 h4d4 c7c5 d4d7');
INSERT INTO evaluations VALUES(32455,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(32455,31,'cp 367',NULL);
INSERT INTO evaluations VALUES(32455,32,'cp 381',NULL);
INSERT INTO evaluations VALUES(32455,33,'cp 364',NULL);
INSERT INTO evaluations VALUES(32455,34,'cp 372',NULL);
INSERT INTO evaluations VALUES(32455,35,'cp 368',NULL);
INSERT INTO evaluations VALUES(32455,36,'cp 341','g7g6 e5g6 g8f6 h5h3 h8g8 g6f8 e8f8 b1c3 b8c6 b2b3 c6d4 c1a3 f8f7 e1c1 d8d7 a3b2 e6e5 h3d7 f6d7 g2g3 d4f3 c3e4 c7c5 d2d3 f7e6 h2h4 h7h5 d1f1 a8f8 c1d1 b7b5 b2c1 c5c4 b3c4 b5c4 d3c4 d7b6 e4g5 e6d7 c1a3 f8f5 g5f3 f5f3');
INSERT INTO evaluations VALUES(38595,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(38595,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(38595,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(38595,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(38595,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(38595,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(38595,36,'cp 121','c1f4 c8e6 b1d2 g8f6 a2a4 a7a6 c4d3 b8d7 c2c3 d7f8 h2h3 f8g6 f4h2 e8g8 e1g1 d8e7 g1h1 a8e8 d1e1 f6h5 a4a5 b6a7 d3c2 a7b8 g2g4 h5f6 e1g3 c7c5 d4d5 e6c8 f1e1 f6d7 d2c4 d7e5 f3e5 g6e5 c4e5 d6e5');
INSERT INTO evaluations VALUES(39414,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(39414,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(39414,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(39414,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(39414,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(39414,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(39414,36,'cp -51','d8f6 c3b4 f6a1 d1b3 a1f6 b1c3 g8e7 g1f3 d7d5 e4d5 e8g8 c1g5 f6g6 h2h4 h7h6 g5e7 g6g2 h1g1 g2f3 d5d6 c7d6 g1g3 f3f5 c3d5 f5e5 e1f1 b8d7 b3f3 a8b8 d5f6 g8h8 f3g4 f8g8 f6g8 d7b6 g4g7 e5g7 e7f6 c8h3 g3h3 h8g8 f6g7 g8g7 c4b3 b8c8 h3d3 b6c4 h4h5 a7a6 f1e2 c8c6 b3a4');
INSERT INTO evaluations VALUES(40272,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(40272,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(40272,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(40272,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(40272,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(40272,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(40272,36,'cp 46','b1c3 e8g8 f1e2 d5c4 e2c4 f6d5 f4g5 d5c3 b2c3 b8d7 c4d3 d8a5 e1g1 e7e5 d1c2 c6c5 g5h4 e5d4 c3d4 c5d4 f3d4 d7b6 d3e4 f8e8 f1c1 g7d4 e3d4 c8f5 e4f5 a5f5 c2f5 g6f5 c1c7 e8e4 g2g3 e4d4 c7b7 b6d5 a1c1 f5f4 b7d7 a7a5 g1g2 f4g3 h2g3');
INSERT INTO evaluations VALUES(41894,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(41894,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(41894,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(41894,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41894,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(41894,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(41894,36,'cp 15','c3a4 c5b6 a4b6 a7b6 f1e2 h7h6 e1g1 e8g8 b2b4 c8d7 a2a3 c7c5 c1d2 d7a4 d1b1 b8c6 c2c4 b6b5 b4c5 b5c4 c5d6 d8d6 d3c4 f6d7 d2e3 d7c5 e3c5 d6c5 b1b2 f8d8 a1b1 d8d7 f1c1 a8d8 h2h3 d7c7 e2d1 a4d1 c1d1 d8d1 b1d1 c5c4 f3e5');
INSERT INTO evaluations VALUES(44412,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(44412,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44412,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(44412,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(44412,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(44412,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44412,36,'cp 40','f1g2 f8g7 g1f3 d7d5 d2d4 e8g8 d4c5 d8a5 e1g1 d5c4 d1d4 b8c6 d4c4 c8e6 c4b5 a5b5 c3b5 a8d8 c1e3 f6e4 f3d4 c6d4 b5d4 e4c5 f1d1 e6a2 a1a2 c5e6 a2a7 e6d4 d1b1 d4f5 e3c5 d8d2 b2b4 d2e2 a7b7 e2c2 c5e7 f5e7');
INSERT INTO evaluations VALUES(46620,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(46620,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(46620,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(46620,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(46620,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(46620,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(46620,36,'cp 112','h2h3 f8g7 g1f3 e8g8 a2a4 a7a5 c1e3 b8c6 f1e2 f7f5 e4f5 c8f5 e1g1 g8h8 a1c1 b6d5 c3d5 d8d5 e2c4 d5d6 f1e1 e7e6 d1b3 c6d4 f3d4 g7d4 e3d4 d6d4 b3b7 a8b8 b7f3 d4b2 c4b5 c7c6 b5c6 b8c8 f3e3 h8g8');
INSERT INTO evaluations VALUES(50914,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(50914,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(50914,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(50914,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(50914,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(50914,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(50914,36,'cp 40','f8e7 d2d4 e5d4 f3d4 g8f6 b1c3 e8g8 f1e1 c8g4 f2f3 c7c5 d4e2 g4e6 c4d5 f6d5 c3d5 e7h4 e1f1 b8c6 c1f4 c6e5 b2b3 f7f5 e2g3 h4g3 f4g3 f5e4 d5f4 d8d7 f3e4 a8d8 d1d2 e5g6 a1d1 g6f4 f1f4 f8f4 d2f4 c5c4 f4d6 c4b3');
INSERT INTO evaluations VALUES(51896,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(51896,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(51896,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51896,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(51896,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(51896,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(51896,36,'cp 0','b8a6 e3d4 a6c5 e2e3 b7b6 g1f3 c8b7 c2c4 g8f6 f1e2 c5e4 e1g1 d8c7 h2h3 f8c5 a2a3 e8g8 b1d2 a7a5 d2e4 f6e4 a1c1 h7h6 f3d2 c5d4 e3d4 c7f4 d2e4 b7e4 d1b3 f4g5 b3g3 g5g3 f2g3 e4g6 e2f3 a8d8');
INSERT INTO evaluations VALUES(53745,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(53745,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53745,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(53745,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(53745,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(53745,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(53745,36,'cp 30','h5e5 c8e6 e4d5 c6d5 c4b5 b8c6 g1e2 g8f6 e2d4 d8d7 e5e2 f8d6 d4e6 f7e6 d2d4 e8g8 b1d2 c6d4 b5d7 d4e2 d7e6 g8h8 e1e2 a8e8 e2d3 e8e6 d2f3 d6c5 c1e3 c5e3 f2e3 f6g4 h1e1 f8e8 e1e2 e6e3 e2e3 e8e3 d3d2 e3e4 b2b3 h8g8 h2h3 g4e3 a1g1 e3f5 g1f1 f5g3 f1f2 e4e8 d2d3 g3e4 f2f1 a7a6');
INSERT INTO evaluations VALUES(53744,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(53744,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(53744,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(53744,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(53744,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(53744,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(53744,36,'cp 28','c8e6 e4d5 c6d5 c4b5 b8c6 g1f3 g8f6 f3g5 d8d7 e5e2 f8d6 g5e6 f7e6 d2d4 e8g8 b1d2 c6d4 b5d7 d4e2 d7e6 g8h8 e1e2 a8e8 e2d3 e8e6 d2f3 d6c5 c1e3 c5e3 f2e3 f6g4 h1e1 g4f2 d3d2 f2e4 d2e2 e6g6 e1g1 g6b6 b2b3 b6c6 a1c1 f8c8 f3d4 c6h6 g1h1 e4c3 e2e1 c3e4 a2a4 h6g6 g2g3 c8c3 e1e2 g6f6 h1f1');
INSERT INTO evaluations VALUES(56103,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(56103,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(56103,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(56103,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(56103,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(56103,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(56103,36,'cp -52','c7c6 b5e2 f8d6 d2d4 g8e7 g1f3 e8g8 e2d3 h7h6 c3e2 e7g6 e1g1 c8g4 e2g3 d8f6 d3e2 b8d7 h2h3 g4f3 e2f3 d6g3 f2g3 f6d6 g1h2 d7f6 c1d2 f8e8 f1e1 e8e1 d1e1 a8e8 e1f2 h6h5 a1e1 f6e4 f3e4 d5e4');
INSERT INTO evaluations VALUES(53676,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(53676,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(53676,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(53676,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(53676,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(53676,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(53676,36,'cp -57','b5e2 h7h6 d2d4 f8d6 g1f3 g8e7 e2d3 e8g8 c3e2 e7g6 e1g1 c8g4 e2g3 d8f6 d3e2 g6f4 c1f4 f6f4 f1e1 b8d7 g3f1 a8e8 g2g3 f4f6 f1e3 g4e6 c2c3 g7g6 e2f1 e8e7 b2b4 f8e8 b4b5 h6h5 b5c6 b7c6 h2h4 d7b6 f3g5 e6c8 a2a4 a7a5 d1f3 f6f3 g5f3');
INSERT INTO evaluations VALUES(1595,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(1595,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(1595,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(1595,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(1595,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(1595,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(1595,36,'cp 61','c3b5 e8d8 d2d3 g8f6 h2h3 e7e5 g1e2 c8e6 d3d4 d8c8 e1g1 a7a6 d4e5 c6e5 b5d4 e5f3 d4f3 f8d6 e2f4 d6f4 c1f4 e6h3 f4g3 h3g4 f1e1 f6h5 g3h2 h8d8 d1e2 g6c6 e2e7 g4f3 g2f3 b7b6 e7f7 h5f6 e1e7 f6d5 f7f5 c8b7 e7g7 h7h5 h2g3 h5h4 g3h4 d8g8 g7g8 a8g8 h4g3 d5b4 a1d1 b4c2');
INSERT INTO evaluations VALUES(3769,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(3769,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(3769,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(3769,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(3769,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(3769,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(3769,36,'cp 57','c2c4 b7b6 a1c1 c8b7 f1e2 c6e7 c4d5 e6d5 e1g1 f6e4 d2e4 d5e4 f3e5 b7d5 e2c4 d5c4 e5c4 d6g3 h2g3 e7d5 c4e5 f8e8 e5c6 d8d7 b2b4 a7a6 d1b3 d7e6 b3c2 f7f6 a2a4 h7h5 f1d1 e6d7 c2b3 d7e6 b4b5');
INSERT INTO evaluations VALUES(4000,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(4000,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(4000,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(4000,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(4000,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(4000,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(4000,36,'cp 36','f1d3 c7c6 h2h3 f6e4 g5f4 d7f6 f3e5 e7d6 e1g1 d8e7 a2a4 a7a5 d1c2 f8e8 c3e2 g7g6 a1c1 f6h5 f4h2 f7f6 d3e4 d5e4 e5c4 d6h2 g1h2 c8e6 h2g1 f6f5 b2b3 h5f6 e2f4 e6f7 c4e5 f6d5 c2b2 d5f4 e3f4 f7d5 c1c5');
INSERT INTO evaluations VALUES(4003,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4003,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(4003,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(4003,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(4003,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(4003,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(4003,36,'cp 27','c7c6 e1g1 f6e8 g5f4 g7g6 h2h3 d7b6 f1e1 e8g7 d1b3 a7a5 a2a3 f8e8 a1d1 a5a4 b3c2 c8e6 e1e2 e6f5 d3f5 g7f5 e3e4 d5e4 e2e4 b6d5 d1e1 d5f6 e4e2 f5d4 f3d4 d8d4 f4e5 d4d7 c3a4 f6d5 a4c3 e7f8 c3d5 d7d5 e5c3 e8e2 e1e2 d5d7 c2b3 b7b5 b3c2');
INSERT INTO evaluations VALUES(4181,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4181,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4181,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(4181,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(4181,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(4181,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(4181,36,'cp 21','e1g1 d7d6 f1e1 a7a6 b1d2 c8e6 b2b4 c5a7 c4b3 f8e8 a2a4 e6b3 d2b3 c6e7 d1c2 e7g6 c1e3 a7e3 e1e3 d6d5 g2g3 d8d7 g1g2 f6h5 g2h2 h5f6 e3e1 a8d8 b3c5 d7c6 a4a5 f6d7 c5d7 d8d7 d3d4 e5d4');
INSERT INTO evaluations VALUES(5853,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(5853,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(5853,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(5853,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(5853,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(5853,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(5853,36,'cp 202','g1f3 g8f6 e1g1 b4c3 b2c3 f6e4 f1e1 e4d6 c4d3 e8g8 f3g5 g7g6 d1f3 c6e7 h2h4 c8d7 h4h5 h7h6 g5e4 d6e4 d3e4 g6g5 e4b7 a8b8 c1a3 f8e8 a3c5 e7d5 a1b1 d7c8 b7c6 b8b1 e1b1 c8d7 c6b7 a7a5 c3c4 d5f6');
INSERT INTO evaluations VALUES(6124,30,'cp 289',NULL);
INSERT INTO evaluations VALUES(6124,31,'cp 292',NULL);
INSERT INTO evaluations VALUES(6124,32,'cp 278',NULL);
INSERT INTO evaluations VALUES(6124,33,'cp 282',NULL);
INSERT INTO evaluations VALUES(6124,34,'cp 278',NULL);
INSERT INTO evaluations VALUES(6124,35,'cp 292',NULL);
INSERT INTO evaluations VALUES(6124,36,'cp 292','d4e5 d7b6 c4b3 d6e5 d1d8 e7d8 f3e5 c8e6 e1e2 g8f6 b1c3 f6d7 e5d7 b6d7 c1e3 a7a6 b3e6 f7e6 f2f4 d8e7 a1f1 e7c5 e3c5 d7c5 e2e3 c5d7 a2a4 e6e5 f4f5 h7h5 c3d5 e8c8 a4a5 d7f6 f1d1 b7b5 e3f3 h5h4 d5f6 g7f6 d1d8 h8d8');
INSERT INTO evaluations VALUES(6122,30,'cp 271',NULL);
INSERT INTO evaluations VALUES(6122,31,'cp 269',NULL);
INSERT INTO evaluations VALUES(6122,32,'cp 282',NULL);
INSERT INTO evaluations VALUES(6122,33,'cp 284',NULL);
INSERT INTO evaluations VALUES(6122,34,'cp 277',NULL);
INSERT INTO evaluations VALUES(6122,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(6122,36,'cp 270','d7b6 c4b3 d6e5 d1d8 e7d8 f3e5 c8e6 e5d3 g8f6 b3e6 f7e6 b1d2 f6d7 a2a4 a7a5 b2b3 d8f6 a1b1 e8c8 d2f3 d7b8 c1e3 b8c6 e1e2 h7h6 b1d1 b6d7 f3d2 d7e5 d3e5 c6e5 d2f3 e5c6 h2h4 d8d1 h1d1 h8f8 d1d2 b7b6 h4h5 c8b7 f3d4 f6d4 e3d4 c6d4 d2d4');
INSERT INTO evaluations VALUES(6633,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(6633,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(6633,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6633,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(6633,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6633,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(6633,36,'cp 8','g8f6 d2d3 h7h6 e1g1 e8g8 f1e1 d7d6 c4b3 a7a6 h2h3 b7b5 c1e3 f8e8 a2a4 c6a5 b3a2 c7c5 f3d2 b5b4 c3b4 c5b4 d1f3 c8e6 a2d5 a8c8 a1b1 c8c2 b1b4 a5c6 b4c4 c2c4 d3c4 c6d4 e3d4 e5d4 d5e6 f7e6 f3d3 e6e5 c4c5 d6c5 d3a6 g8h8 e1b1');
INSERT INTO evaluations VALUES(10427,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(10427,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(10427,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(10427,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(10427,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(10427,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(10427,36,'cp 24','d2d3 h7h6 e1g1 d7d6 c4b3 e8g8 f1e1 c6a5 c3c4 b7b6 f3d2 a7a6 h2h3 c8e6 c1b2 f6d7 d3d4 e5d4 b2d4 b6b5 e1e3 f7f6 c4b5 a5b3 a2b3 a6b5 d2f1 b5b4 f2f4 e6f7 f1g3 c7c5 a1a8 d8a8 d4b2');
INSERT INTO evaluations VALUES(11614,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(11614,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(11614,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(11614,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(11614,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(11614,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(11614,36,'cp 96','e4d5 f6d5 e1g1 c8b7 f1e1 f8e7 f3e5 c6e5 e1e5 e8g8 d1f3 c7c6 b1c3 e7f6 e5f5 b7c8 f5d5 c6d5 c3d5 c8e6 d5f6 d8f6 f3f6 g7f6 g1f1 a6a5 c2c3 e6b3 a2b3 a5a4 b3b4 f8c8 a1a3 f6f5 c1e3 f7f6 f1e2 g8f7 e2d2 h7h5 e3c5 c8g8 g2g3 f5f4 c3c4 b5c4 d3c4 h5h4');
INSERT INTO evaluations VALUES(11610,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(11610,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(11610,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(11610,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(11610,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(11610,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(11610,36,'cp 62','f6d5 e1g1 c8b7 f1e1 f8e7 f3e5 c6e5 e1e5 e8g8 d1f3 c7c6 b1c3 e7f6 e5f5 b7c8 f5d5 c6d5 c3d5 c8e6 d5f6 d8f6 f3f6 g7f6 g1f1 a6a5 c2c3 e6b3 a2b3 a5a4 b3b4 f8c8 a1a3 f6f5 c1e3 f7f6 f1e2 g8f7 g2g3 h7h5 h2h4');
INSERT INTO evaluations VALUES(11619,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(11619,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(11619,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(11619,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(11619,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(11619,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(11619,36,'cp 70','e1g1 c8b7 f1e1 f8e7 f3e5 c6e5 e1e5 e8g8 d1f3 c7c6 b1c3 e7f6 e5f5 b7c8 f5d5 c6d5 c3d5 c8e6 d5f6 d8f6 f3f6 g7f6 g1f1 a6a5 c2c3 e6b3 a2b3 a5a4 b3b4 f8c8 a1a3 f6f5 c1e3 f7f6 f1e2 g8f7 e2d2 f7e6 a3a1 h7h5 g2g3 h5h4 a1e1 a4a3 b2a3 a8a3 e3c5 e6f7 e1e7 f7g6 c5d4 f5f4 g3f4 a3a1 e7b7 a1h1 b7b5 h1h2 f4f5 g6g5 d4e3 g5h5 b5c5 c8e8');
INSERT INTO evaluations VALUES(11618,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(11618,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(11618,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(11618,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(11618,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(11618,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(11618,36,'cp 74','f8e7 f1e1 e8g8 f3e5 c6e5 e1e5 c8b7 d1f3 c7c6 b1c3 e7f6 e5f5 b7c8 f5d5 c6d5 c3d5 c8e6 d5f6 d8f6 f3f6 g7f6 g1f1 a6a5 c2c3 e6b3 a2b3 a5a4 b3b4 f6f5 c1e3 f8c8 a1a3 f7f6 f1e2 g8f7 e3c5 c8e8 e2d2 a8d8 c3c4 b5c4 d3d4 f7e6 b4b5 d8a8 b5b6 f5f4 a3h3 a8a5 h3h7 a5b5 d2c3 b5b3 c3c4 b3b2');
INSERT INTO evaluations VALUES(12250,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(12250,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(12250,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12250,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(12250,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(12250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12250,36,'cp 14','e7e6 f1d3 f8e7 d1e2 c6c5 e1g1 e8g8 c2c4 c5d4 f3d4 d7c5 f1d1 c5d3 d1d3 d8c7 b2b3 b7b6 d4b5 c7c6 g3f5 e6f5 e2e7 f8e8 e7b4 a7a5 b4c3 f6e4 c3d4 c8a6 c1b2 f7f6 b5c3 e4c3 b2c3 b6b5 c4b5 a6b5 d3g3 g8h8 a2a4 b5e2 a1c1 h7h6 g3e3 e8e3 d4e3 e2a6 c1d1 a6b7 f2f3 c6c7 c3b2 a8d8');
INSERT INTO evaluations VALUES(12186,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(12186,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(12186,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12186,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(12186,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(12186,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12186,36,'cp 12','f1d3 f8e7 d1e2 c6c5 e1g1 e8g8 c2c4 c5d4 f3d4 a7a6 f1d1 d8c7 c1d2 f8d8 d2c3 d7f8 d4f3 f8g6 d3g6 d8d1 a1d1 h7g6 g3e4 b7b5 e4g5 c7c4 e2c4 b5c4 f3e5 f6d5 e5f7 d5c3 b2c3 a8b8 h2h4 e7f6 f7d6 f6c3 d6c4 c3f6 c4d6 g8f8 d6c8 b8c8 g5e6 f8e7');
INSERT INTO evaluations VALUES(12185,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12185,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(12185,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(12185,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12185,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(12185,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(12185,36,'cp 8','c6c5 e1g1 f8e7 d1e2 e8g8 c2c4 c5d4 f3d4 a7a6 b2b3 d8c7 c1b2 f8e8 a1d1 b7b6 g3h5 e7f8 d3e4 c8b7 h5f6 d7f6 e4b7 c7b7 d4f3 a8d8 d1d8 e8d8 f1d1 d8d1 e2d1 f8e7 f3e5 h7h6 d1f3 b7f3 e5f3 f6d7 f3e5 d7e5');
INSERT INTO evaluations VALUES(16123,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(16123,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(16123,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(16123,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(16123,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(16123,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(16123,36,'cp -2','d8e7 e1g1 f6e4 d2d4 c6a5 b3a4 a5c4 a4c4 c7c6 d4e5 d7d5 e5d6 e4d6 c4f4 d6f5 b1d2 c8e6 f1e1 e8c8 f4a4 a7a6 d2e4 g7g5 c1d2 f8g7 g2g4 f5d6 e4c5 e7c7 c5e6 f7e6 a4b3 c7f7 b3e6 f7e6 e1e6 d6c4 d2e3 c4b2 f3d4 b2d3 a1d1 d3c5 e6e7 d8d7 e7d7 c5d7 d4f5');
INSERT INTO evaluations VALUES(1075,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(1075,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(1075,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(1075,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(1075,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(1075,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(1075,36,'cp 0','e1g1 f6e4 d2d4 c6a5 b3a4 a5c4 a4c4 c7c6 d4e5 d7d5 e5d6 e4d6 c4f4 d6f5 b1d2 c8e6 f1e1 e8c8 f4a4 a7a6 d2e4 e7c7 c3c4 f8e7 c4c5 d8d3 c1d2 c8b8 d2a5 c7f4 a5d2');
INSERT INTO evaluations VALUES(17666,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17666,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(17666,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17666,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17666,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(17666,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17666,36,'cp 0','d2d4 c5d4 f3d4 g8f6 d4c6 b7c6 c1f4 d6d5 c4d3 f8e7 e1g1 e8g8 f1e1 a6a5 e4e5 f6d7 d3h7 g8h7 d1h5 h7g8 e1e3 f7f6 e3h3 f6e5 h5h7 g8f7 h7h5');
INSERT INTO evaluations VALUES(17946,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(17946,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(17946,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(17946,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(17946,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(17946,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(17946,36,'cp 132','b5c4 e5d4 e1g1 f8c5 c2c3 d4d3 b2b4 c5b6 d1d3 d7d6 c1e3 b6e3 d3e3 g8h6 b1d2 h6f7 f3d4 e8g8 a2a4 g8h8 d4c6 b7c6 f2f4 d8e7 a4a5 c8e6 c4d3 g7g5 f4g5 f7g5 d2f3 f8g8 f3g5 g8g5 f1f2 e7g7 g2g3 c6c5 e3f3 g5e5 f3f6 g7f6 f2f6 h8g7 f6f3 c5b4 c3b4');
INSERT INTO evaluations VALUES(19863,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19863,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19863,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19863,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19863,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19863,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19863,36,'cp 0','c8f5 c2c4 e7e6 c4c5 a7a6 c1e3 g8e7 b1c3 h7h6 f1e2 g7g6 g2g4 f5e4 c3e4 d5e4 f3d2 c6d4 d1a4 e7c6 d2e4 f8c5 e4c5 b7b5 a4d1 d8d5 e1g1 d5c5 e2f3 d4f3 d1f3 c5d5 f3f6 d5e5 f6f3');
INSERT INTO evaluations VALUES(26286,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26286,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26286,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26286,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26286,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26286,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26286,36,'cp 0','c2c4 e7e6 c4c5 a7a6 b1c3 f5e4 f3g5 e4g6 g5f3');
INSERT INTO evaluations VALUES(21408,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(21408,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(21408,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21408,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(21408,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21408,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(21408,36,'cp 21','g1f3 b7b6 c2c4 g8e7 b1c3 g7g6 d1a4 f8g7 f3e5 c8d7 e5d7 d8d7 f1e2 e8g8 e1g1 d5c4 f1d1 f8d8 a4c4 c6a5 c4a6 c7c6 e2f3 a5b7 c1d2 b7d6 a1c1 h7h5 a6e2 a8c8 d2e1');
INSERT INTO evaluations VALUES(29836,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(29836,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(29836,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(29836,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(29836,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(29836,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(29836,36,'cp 74','h2h3 g4h5 g2g4 h5g6 d2d4 b8d7 d1e2 c7c6 f1d1 d8c7 f3h4 g6h7 h4f5 e5d4 d1d4 g7g5 c1e3 h7f5 e4f5 d7e5 f2f4 g5f4 e3f4 e8c8 f4e5 d6e5 d4d8 c8d8 e2e3 b7b5 c4b3 d8c8 g1g2 f8b4 a2a4 b4c3 e3c3 b5b4 c3b4 h8d8 a1e1 c6c5 b4c3 d8d4 b3f7 c7f7 c3c5 f7c7 c5c7 c8c7 e1e5 d4a4 e5e7 c7d8');
INSERT INTO evaluations VALUES(29835,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(29835,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(29835,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29835,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(29835,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(29835,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(29835,36,'cp 33','g4h5 a2a4 f8e7 d2d3 d8c8 c1e3 a7a5 f1e1 b8c6 c3d5 f6d5 c4d5 c6b8 c2c3 c7c6 d5c4 b8d7 d3d4 e8g8 c4e2 h5g6 e2d3 e5d4 f3d4 f8e8 d3c2 d7f6 e3f4 d6d5 e4d5 g6c2 d1c2 f6d5 f4g3 e7f8 d4f5 e8e1 a1e1 c8d7');
INSERT INTO evaluations VALUES(29831,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(29831,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(29831,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(29831,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(29831,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(29831,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(29831,36,'cp 44','a2a4 f8e7 d2d3 d8c8 c1e3 b8c6 a4a5 c6b4 c4b3 c7c6 g1h1 d6d5 e3d2 d5d4 c3b1 f6d7 b1a3 b4a6 a3c4 c8c7 c2c3 a6c5 b3a2 b7b5 a5b6 a7b6 c3d4 e5d4 g2g4 h5g6 f3d4 c5d3 d4f5 g6f5 g4f5 b6b5 d2a5 c7f4 d1d3 b5c4 a2c4 d7c5 d3e3 f4e3 f2e3 c5e4');
INSERT INTO evaluations VALUES(37100,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(37100,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(37100,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(37100,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(37100,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37100,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(37100,36,'cp 12','e2e3 h7h6 g5h4 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 f6e4 c3a5 d7f6 a1c1 g7g5 h4g3 b7b6 a5a6 c7c5 e1g1 d8e7 d4c5 b6c5 f1d1 h6h5 e2d3 e7e6 a6e6 f7e6 g3e5 f6g4 d3e4 d5e4 h2h3 e4f3 h3g4 h5g4 g2f3 g4f3 a2a3');
INSERT INTO evaluations VALUES(37099,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(37099,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(37099,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(37099,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(37099,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(37099,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37099,36,'cp 12','h7h6 g5h4 c8f5 d1b3 b4c3 b3c3 b8d7 f1e2 f6e4 h4d8 e4c3 d8c7 f8c8 c7a5 c3e2 e1e2 c8c2 f3d2 c2b2 h1c1 g7g5 f2f3 a8e8 a5c3 b2b6 d2b3 e8c8 c3a5 b6c6 c1c6 c8c6 a1c1 c6c4 a2a3 b7b6 a5b4 c4c2 c1c2 f5c2 b3d2');
INSERT INTO evaluations VALUES(37230,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(37230,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(37230,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(37230,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(37230,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37230,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37230,36,'cp -23','c7c6 g1f3 d7d6 a2a4 f8e8 e1g1 h7h6 c1e3 c5b4 c4a2 b4a5 c3e2 b8d7 f3h4 d6d5 h4f5 d7f8 e2g3 a5c7 d1d2 f8e6 e4d5 c6d5 d3d4 e6d4 f5d4 e5d4 d2d4 c7b6 d4d3 b6e3 f2e3 c8e6 a1d1 d8c7 f1f6 c7g3');
INSERT INTO evaluations VALUES(38757,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(38757,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38757,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(38757,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38757,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(38757,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(38757,36,'cp -10','h4e7 c1f4 g8f6 b1c3 e8g8 f1f2 c7c5 h1e1 c5d4 f3d4 b8c6 f2g1 d8b6 f4e3 b6b2 c3d5 f6d5 c4d5 b2c3 d4c6 b7c6 e3d4 c8g4 d4c3 g4d1 a1d1 c6d5 e4d5 e7g5 c3d4 f8c8 c2c3 h7h5 e1e2 c8c4 d1e1 c4a4 g2g3 h5h4 g1g2 g5f6 d4f6');
INSERT INTO evaluations VALUES(40863,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40863,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(40863,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40863,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40863,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40863,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(40863,36,'cp 14','d3d4 d6d5 e1g1 e8g8 f1e1 c8e6 b1d2 h7h6 e2d3 e7d6 f3e5 f8e8 d2f3 c7c5 d4c5 d6c5 c1e3 c5e3 e1e3 b8c6 d1d2 d8b6 b2b3 e6g4 e5g4 f6g4 e3e8 a8e8 a1e1 e8e1 d2e1 g4f6 h2h3 c6b4 e1d2 b4d3 c2d3 b6e6 d2f4 e6b6');
INSERT INTO evaluations VALUES(42105,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(42105,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(42105,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(42105,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42105,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(42105,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(42105,36,'cp 32','h2g3 d8e7 b1d2 e6e5 d4e5 e7e5 g1f3 e5e7 c3c4 c8e6 d1b3 c7c6 c4d5 e6d5 f1c4 e8g8 c4d5 c6d5 e1g1 a7a5 f3d4 b8c6 d2f3 f8c8 f1c1 g7g6 a2a3 h7h5 b3b5 c6d4 f3d4 c8b8 b5b6 a8a6 b6c7 e7c7 c1c7');
INSERT INTO evaluations VALUES(43536,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(43536,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(43536,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(43536,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43536,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(43536,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(43536,36,'cp 32','f8e7 f2f3 h7h5 h2h3 d8c7 a2a3 b7b5 e1c1 h5h4 c1b1 b8d7 h1g1 c8b7 e3f2 a8b8 g2g4 h4g3 g1g3 g7g6 h3h4 b8c8 g3g6 f7g6 d4e6 c7c6 f1h3 d7e5 f2d4 h8h4 d4e5 h4h3 d2g2 h3h6 e5f4 e8f7 f4h6 f7e6 g2g6 e6d7 h6g5 c8g8');
INSERT INTO evaluations VALUES(39736,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(39736,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(39736,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(39736,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(39736,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(39736,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(39736,36,'cp 45','f2f3 h7h5 h2h3 d8c7 h1g1 h5h4 e1c1 b7b5 a2a3 b8d7 g2g4 h4g3 g1g3 g7g6 d2g2 e6e5 d4b3 c8b7 g2f2 f6h5 g3g1 d7c5 c1b1 e8f8 h3h4 c5b3 c2b3 h5f4 g1h1 a8c8 d1c1 c7d8 f2d2 e7h4 e3f4 e5f4 d2f4 h4f6');
INSERT INTO evaluations VALUES(44040,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(44040,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(44040,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(44040,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(44040,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44040,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(44040,36,'cp 38','e2e4 d7d5 e4e5 f6d7 d2d4 d5c4 f1c4 c6c5 g1f3 b8c6 d4d5 e6d5 c1g5 f8e7 g5e7 c6e7 c4d5 e7d5 d1d5 e8g8 e1g1 d7b6 d5c5 c8g4 a1d1 d8c8 c5e3 f8e8 e3f4 b6c4 c3b5 g4f3 f4f3 e8e5 b5d6 c4d6 d1d6 e5f5 f3e4 g7g6 f1d1 c8c5 d6d4 c5c6 e4c6 b7c6 d4d7');
INSERT INTO evaluations VALUES(49502,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(49502,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(49502,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(49502,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(49502,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(49502,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(49502,36,'cp -42','b8c6 d4f4 d8e7 g1f3 d7d6 f1b5 d6e5 f4e5 c8d7 e5e7 f8e7 e1g1 e8c8 b1c3 c6b4 b5d7 d8d7 f3e5 b4c2 a1b1 d7d4 c1g5 a7a6 b1c1 c2b4 a2a3 b4d5 e5f7 h8f8 f7e5 d5c3 c1c3 d4d5 f1e1 e7d6 g5f6 f8f6 e5d3 a6a5 g2g3 h7h6 g1g2 g7g5 e1e8 c8d7 e8b8');
INSERT INTO evaluations VALUES(16353,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(16353,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(16353,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(16353,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(16353,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(16353,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(16353,36,'cp -37','d4f4 d8e7 g1f3 d7d6 f1b5 d6e5 f4e5 c8d7 e5e7 f8e7 e1g1 e8c8 b1c3 a7a6 b5c4 d7g4 c4f7 g4f3 g2f3 c6d4 f7b3 d4f3 g1g2 f3d4 f1e1 d8d7 c1e3 d4b3 a2b3 e7b4 e1d1 b4d6 c3b5 d6e5 d1d7 f6d7 b5c3 h8e8 a1a4 e5f6 e3f4 d7c5 a4c4 c5e6 f4g3 e8d8 g2f1');
INSERT INTO evaluations VALUES(56441,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(56441,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(56441,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(56441,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(56441,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(56441,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(56441,36,'cp -67','c7c6 g2g3 g8f6 f1g2 h7h6 g1e2 f8e7 e3e4 d5e4 c3e4 f6e4 g2e4 e8g8 e1g1 c8g4 c1e3 d8d7 d1d2 b8a6 e4g2 f8e8 a1e1 a6c7 f2f3 g4h3 g2h3 d7h3 e2c3 c7e6 e3f2 e6d4 f2d4 e5d4 c3e4 h3e6 d2f2 e6a2 f2d4');
INSERT INTO evaluations VALUES(1396,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(1396,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(1396,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(1396,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(1396,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(1396,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(1396,36,'cp 20','g1f3 e7e6 c1e3 c5d4 c3d4 d5a5 b1c3 g8f6 f1c4 f8d6 e1g1 e8g8 d4d5 e6d5 c3d5 f6d5 d1d5 a5d5 c4d5 f8d8 f1d1 c8f5 a2a3 d6e7 d1d2 e7f6 a1d1 h7h6 h2h3 a7a5 e3b6 d8e8 g2g4 f5g6 g1g2 a5a4 d5c4 a8c8 g2g3 c6e5 c4b5 e5f3 g3f3');
INSERT INTO evaluations VALUES(5156,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(5156,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(5156,32,'cp 261',NULL);
INSERT INTO evaluations VALUES(5156,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(5156,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(5156,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(5156,36,'cp 281','d4d5 c6e5 c3b5 d7d6 f3e5 g7e5 f2f4 e5f6 d5e6 c8e6 f1e2 d6d5 f4f5 g6f5 e4d5 e6d7 d5d6 a8c8 b5c7 c8c7 d6c7 d8c7 e1g1 f6e5 c1g5 e5h2 g1h1 h2e5 e2f3 h7h6 g5h4 g8e7 a1c1 d7c6 b2b4 e8g8 b4b5 e7g6 b5c6 g6h4 c6b7 c7e7 d1e1 h4g6');
INSERT INTO evaluations VALUES(6722,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(6722,31,'cp 247',NULL);
INSERT INTO evaluations VALUES(6722,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(6722,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(6722,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(6722,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(6722,36,'cp 263','d7d6 a2a4 a5b6 c1f4 d8e7 f1b5 d6e5 f3e5 g8f6 e1g1 e8g8 b5c6 b7c6 a4a5 b6c5 b1d2 f8e8 h2h3 e7e6 f1e1 c5f8 d2f3 f6d7 e5d7 c8d7 f4c7 c6c5 d1c2 d7b5 e1e3 h7h6 c7g3 a8d8 b2b3 b5a6 a1e1 a6b7 f3d2 b7a6 c3c4 a6b7 f2f3 e6a6');
INSERT INTO evaluations VALUES(8501,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(8501,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(8501,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(8501,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(8501,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(8501,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(8501,36,'cp 104','e8g8 e3e4 b7b6 c1f4 c8b7 e4e5 f6d5 c3d5 e6d5 c4e2 d8d7 e2b5 a7a6 b5d3 c6d8 f3g5 g7g6 a1c1 c7c5 f1e1 b7c6 d4c5 e7c5 a2a3 d8e6 d1g4 e6f4 g4f4 a8e8 h2h3 a6a5 f4d2 c5e7');
INSERT INTO evaluations VALUES(3336,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(3336,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(3336,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(3336,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(3336,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(3336,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(3336,36,'cp 142','e3e4 b7b6 c1e3 c6a5 c4d3 c7c5 e4e5 f6d5 d1c2 h7h6 c3d5 e6d5 d4c5 c8g4 c5b6 a7b6 d3e2 a8c8 c2d1 e7c5 e3f4 g4f5 a1c1 d5d4 f3e1 f5e6 e2d3 c5e7 e1f3 c8c1 f4c1 a5c6 a2a3 d8d5 c1f4 f8d8 b2b4 d5b3 f1e1 b3d1');
INSERT INTO evaluations VALUES(10373,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(10373,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(10373,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(10373,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(10373,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(10373,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(10373,36,'cp 34','c2c3 b4e7 b2b4 d7d5 b1d2 e8g8 e1g1 e7d6 e4d5 f6d5 d2e4 a7a6 a2a4 h7h6 f1e1 f8e8 e2f1 c8g4 h2h3 g4h5 c1b2 d5f4 g2g3 f4e6 f1g2 d6f8 d1c2 f7f6 b4b5 a6b5 a4b5 a8a1 b2a1 c6a5 d3d4 e5d4 c2a2 h5f3 g2f3');
INSERT INTO evaluations VALUES(10372,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(10372,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(10372,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(10372,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(10372,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(10372,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(10372,36,'cp 34','b4e7 b2b4 d7d5 b1d2 e8g8 e1g1 e7d6 e4d5 f6d5 d2e4 a7a6 a2a4 h7h6 f1e1 f8e8 e2f1 c8g4 h2h3 g4h5 c1d2 d5f6 e4g3 h5g6 d1c2 d8d7 g3e4 d7e6 f3h4 g6h7 b4b5 c6d8 c3c4 f6e4 d3e4 d6c5 d2e3 c5e3 e1e3');
INSERT INTO evaluations VALUES(11451,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11451,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11451,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11451,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(11451,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(11451,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11451,36,'cp 4','f8d6 b1d2 c8e6 d3d4 f6d7 d4e5 d7e5 f3e5 d6e5 f2f4 e5d4 g1h1 f7f5 d1h5 g7g6 h5e2 d8f6 c2c3 d4a7 e4f5 f6f5 f1e1 e8d7 d2f3 a8e8 c1e3 c6c5 a1d1 d7c8 b2b4 e6d7 e2f2 d7a4 d1d2 h8f8 f3e5 g6g5 e3c5 g5f4 c5a7 e8e5 e1e5');
INSERT INTO evaluations VALUES(12034,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(12034,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(12034,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(12034,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(12034,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(12034,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(12034,36,'cp 49','e4e5 f6d7 f1e1 e7e6 e5d6 f8d6 d2d3 e8g8 b1d2 d6c7 b2b3 e6e5 d2c4 c8a6 f3d2 d8h4 d2e4 h7h6 c1a3 a6c4 b3c4 f7f5 e4c5 d7c5 a3c5 f8f7 g2g3 h4g5 d1e2 a8e8 f2f4 g5g6 a1b1 e5e4 g1h1 g6f6 e2f2 c7a5 e1e2 a5c3 c5a7 f7d7');
INSERT INTO evaluations VALUES(13764,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(13764,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(13764,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(13764,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(13764,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(13764,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(13764,36,'cp -12','e1g1 f8e7 h2h3 e8g8 d2d4 f8e8 f1e1 d8b6 d4c5 e7c5 e1e8 f6e8 d1e2 c8d7 c1e3 c5e3 e2e3 b6e3 f2e3 e8f6 b1d2 a8e8 g1f2 g7g6 b3c2 g8g7 a1d1 c6a5 g2g4 h7h6 f3d4 b7b6 d1e1 f6e4 c2e4 d5e4 f2g3 a5b7 e1d1 f7f6 d2c4 b7c5 a2a3 c5d3');
INSERT INTO evaluations VALUES(13763,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(13763,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(13763,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(13763,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(13763,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(13763,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(13763,36,'cp -12','f8e7 d2d4 e8g8 h2h3 f8e8 c1g5 c8e6 d4c5 e7c5 b1a3 h7h6 g5f6 d8f6 a3b5 a8d8 f1e1 f6g6 b3c2 e6f5 c2f5 g6f5 b5d4 c5d4 f3d4 e8e1 d1e1 c6d4 c3d4 d8c8 a1c1 c8c1 e1c1 g7g5 c1c3 b7b6 a2a3 f5b1 g1h2 b1e4 f2f3 e4e2 h2g3 g8g7 a3a4 b6b5 a4b5 e2b5');
INSERT INTO evaluations VALUES(8122,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(8122,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(8122,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(8122,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(8122,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(8122,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(8122,36,'cp -8','d2d4 e8g8 h2h3 f8e8 c1g5 c8e6 d4c5 e7c5 b1a3 c5a3 b2a3 h7h6 g5f6 d8f6 a1c1 a8c8 f1e1 c6a5 f3d4 a5c4 d1h5 f6g5 h5g5 h6g5 b3c4 c8c4 e1e5 e8c8 d4e6 f7e6 c1e1 c4c3 e5e6 d5d4 e6d6 d4d3 e1e3 c3a3 e3d3 a3d3 d6d3 b7b6 g1h2');
INSERT INTO evaluations VALUES(16239,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(16239,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(16239,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(16239,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(16239,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(16239,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(16239,36,'cp 80','d7d6 e1g1 f8e7 d1e2 e8g8 a2a3 c8d7 f1d1 d5b6 c4a2 d6d5 c1f4 g8h8 f4g3 f7f6 e5f6 g7f6 c2c4 d5c4 b1c3 c6a5 f3e5 f6e5 e2e5 h8g8 e5a5 f8f5 g3e5 e7d6 a2b1 f5h5 h2h3 d6e5 d4e5 d8h4 b1e4 a8f8 a5c5 d7c6 e4c6 b7c6 c5c6 h4f2 g1h1 h5e5');
INSERT INTO evaluations VALUES(16798,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(16798,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(16798,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(16798,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(16798,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(16798,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(16798,36,'cp 63','f8c5 a2a3 a7a5 b1c3 e8g8 c3e4 c5a7 b2b4 f8e8 c1b2 d7d6 f1e2 a5b4 a3b4 c6b4 e1g1 g4e5 f3e5 d6e5 d1b3 b4a6 e2f3 d8h4 a1a5 a7b6 a5e5 e8e5 b2e5 c8g4 e5g3 h4h5 f3g4 h5g4 c4c5 b6c5 b3b7 a8b8 e4f6 g7f6 b7a6');
INSERT INTO evaluations VALUES(17233,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(17233,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(17233,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(17233,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(17233,34,'cp -115',NULL);
INSERT INTO evaluations VALUES(17233,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(17233,36,'cp -115','c6e5 f3c3 e5c4 c3c4 d7d5 e4d5 d8d5 c4d5 f6d5 d2d3 d5b4 e1d1 b7b6 f2f3 e7e5 g1e2 h7h5 h2h4 b4c6 c1e3 c8b7 a2a3 a8d8 b2b4 a7a5 b4b5 c6e7 e2g3 g7g6 b1c3 f7f6 d1e2 e8f7 e2f2 h8g8 a3a4 d8e8 c3e4 e7d5 e3d2 e8d8');
INSERT INTO evaluations VALUES(18475,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(18475,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(18475,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(18475,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(18475,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(18475,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(18475,36,'cp 27','b1c3 g8e7 f1e1 f7f6 b2b3 e7g6 c1a3 e6e5 c3a4 c5c4 a3f8 h8f8 d2d4 c4d3 c2d3 d5e4 d3e4 c8g4 e1e3 d8d1 a1d1 a8d8 d1d8 e8d8 f3e1 d8c7 f2f3 g4c8 g2g3 f8d8 e3d3 d8d3 e1d3 g6f8 d3c5 f8e6 c5e6 c8e6 g1f1 c7d6 f1e2 g7g5 e2d3 h7h5 a4b2 g5g4 f3g4 h5g4 d3c2 d6c5 a2a3');
INSERT INTO evaluations VALUES(18924,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(18924,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18924,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18924,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18924,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(18924,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(18924,36,'cp -20','f3d4 d8b6 d4b3 d7d5 b1c3 d5e4 c3e4 f8e7 c1d2 g8f6 d1f3 c6b4 e4f6 e7f6 e1c1 b4a2 c1b1 a2b4 d2e3 b6c6 f3c6 b4c6 f1b5 e8g8 e3c5 f8d8 b3d2 e6e5 b5c6 b7c6 d2e4 c8e6 e4f6 g7f6 d1d8 a8d8 b1c1 d8d7 h1d1 d7d1 c1d1 a7a5 c5e7 g8g7 e7d6 e5e4 d1e2 a5a4 g2g3 h7h5 d6c5 e6g4 e2e1');
INSERT INTO evaluations VALUES(20699,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20699,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20699,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20699,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20699,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20699,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(20699,36,'cp 0','d3d4 c5d4 f3d4 g8f6 f1e2 f8e7 f2f4 d8b6 c1e3 b6b2 c3b5 e8g8 d4c6 b7c6 a2a3 c6c5 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(21167,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(21167,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(21167,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(21167,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(21167,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(21167,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(21167,36,'cp 68','c6e5 f3e5 d6e5 e2e4 a7a6 f1c4 g8f6 b1d2 b7b5 c4b3 e7e6 d5e6 c8e6 b3e6 f7e6 e1g1 f8d6 d2b3 e8g8 d1e2 d8e8 f2f3 a6a5 c1e3 a5a4 b3c5 f6d7 c5b7 d7f6 b2b4 a4b3 c2b3 e8c6 b7d6 c7d6 f1c1 c6a6 a2a4 b5a4 e2a6 a8a6 a1a4');
INSERT INTO evaluations VALUES(21205,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(21205,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(21205,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(21205,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(21205,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(21205,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(21205,36,'cp 67','f3e5 d6e5 e2e4 a7a6 f1c4 g8f6 b1d2 b7b5 c4b3 e7e6 d5e6 c8e6 b3e6 f7e6 d1e2 f8d6 d2b3 e8g8 e1g1 d8e8 f2f3 a6a5 c1e3 a5a4 b3c5 f6d7 c5b7 d7f6 b2b4 a4b3 c2b3 e8c6 b7d6 c7d6 f1c1 c6a6 a2a4 b5a4 e2a6 a8a6 a1a4 a6a8 c1c6 d6d5 c6e6 d5e4 e6e5 e4f3 g2f3 a8a4 b3a4');
INSERT INTO evaluations VALUES(24276,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(24276,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(24276,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24276,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(24276,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24276,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24276,36,'cp 25','h2h3 g4f3 d1f3 e7e6 e1c1 f8d6 g2g4 d6f4 f3f4 h7h6 f4g3 d8e7 a2a3 e8c8 f2f4 c8b8 c1b1 c6a7 h3h4 a7c8 f1d3 c8d6 g4g5 f6e8 g3f2 g7g6 c3a4 h6g5 h4g5 e8g7 b1a2 g7f5 a4c5 h8h5 h1h5 g6h5 f2e1 b8a7 c2c3 d8h8 e3e4 d6e4 d3e4 d5e4');
INSERT INTO evaluations VALUES(24274,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(24274,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24274,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(24274,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(24274,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24274,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(24274,36,'cp 30','g4f3 d1f3 e7e6 e1c1 f8b4 c3e2 b4d6 g2g4 d6f4 e2f4 d8e7 h3h4 e8c8 g4g5 f6e4 f1d3 e4d6 f3g2 c8b8 h4h5 c6b4 a2a3 b4d3 f4d3 d6e4 d1g1 h7h6 g5h6 g7h6 d3e5 h8g8 g2f3 e4d6 c1b1 g8g5 g1g5 e7g5');
INSERT INTO evaluations VALUES(24861,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(24861,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(24861,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(24861,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(24861,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(24861,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(24861,36,'cp 52','f1b5 f5d7 e1g1 a7a6 b5c6 d7c6 f3e5 c6b5 f1e1 e7e6 a2a4 b5d7 b2b3 a6a5 b1d2 f8d6 c2c4 e8g8 d1c2 d7e8 c1a3 d6a3 a1a3 f6d7 e5d3 f7f6 a3a1 e8g6 c2c3 f8e8 e3e4 c7c6 a1d1 d7f8 d3f4 g6f7 d2f3 g8h8');
INSERT INTO evaluations VALUES(25060,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(25060,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(25060,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(25060,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(25060,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(25060,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(25060,36,'cp 76','h2h3 g4d7 a2a3 f8e7 d1e2 a7a6 c4a2 b7b5 c3d5 f6d5 a2d5 a8b8 c2c3 e7f6 d3d4 e8g8 e1g1 f8e8 a1d1 c6e7 d5a2 e7g6 d4e5 f6e5 f3e5 g6e5 f2f4 e5c4 e3a7 b8b7 a7f2 d8e7 a2b1 f7f5 d1e1 e7f7');
INSERT INTO evaluations VALUES(25058,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(25058,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(25058,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(25058,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(25058,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(25058,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(25058,36,'cp 62','g4d7 a2a3 f8e7 g2g4 f6h7 d1d2 h7g5 f3g5 e7g5 e1c1 d7e6 c4d5 c6d4 d1f1 d4f3 d2d1 e6d5 c3d5 g5e3 d5e3 f3d4 f2f4 d4e6 f4e5 d6e5 h3h4 d8d4 d1e2 d4c5 c1b1 e6f4 e2f2 e8c8 h4h5 f7f6 e3f5 c5f2 f1f2 d8d7 h1f1');
INSERT INTO evaluations VALUES(26134,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(26134,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(26134,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(26134,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(26134,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(26134,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(26134,36,'cp 152','d4f5 d7e5 c1f4 g7g6 f5e3 g8f6 b1c3 f8g7 f4e5 d6e5 e1g1 e8g8 d1e2 d8e8 e3d5 f6d5 c3d5 a8d8 b2b3 g8h8 a2a4 f7f5 f2f3 e8c6 e2c4 f5e4 f3e4 f8f1 a1f1 d8f8 f1f8 g7f8 g1f2 h8g7 c4c3 c6e6 f2e2 b7b6 c3f3');
INSERT INTO evaluations VALUES(26133,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(26133,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(26133,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(26133,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(26133,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(26133,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(26133,36,'cp 149','d7e5 c1f4 g7g6 f5e3 g8f6 f4e5 d6e5 b1c3 f8g7 d1e2 e8g8 e1g1 d8d7 a1d1 d7c6 e3d5 f6d5 d1d5 f8d8 f1d1 d8d5 e4d5 c6d7 e2c4 b7b6 h2h4 a8d8 d5d6 e5e4 c3e4 g7b2 c2c3 g8g7 h4h5 d7g4 f2f3 g4e6 c4d3 g6h5 e4g3');
INSERT INTO evaluations VALUES(27590,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(27590,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(27590,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(27590,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(27590,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(27590,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(27590,36,'cp 26','e7e6 b1c3 d5a5 e1g1 g8f6 c1e3 f8e7 a2a3 f6d5 h2h3 g4f5 c3d5 a5d5 a1c1 e8g8 e2c4 d5d6 d4d5 e6d5 c4d5 f8d8 d5c6 d6d1 f1d1 d8d1 c1d1 b7c6 f3d4 f5e4 f2f3 e4d5 d1c1 a8c8 d4f5 c8c7 f5e7 c7e7 g1f2 e7b7 b2b4 f7f6 e3c5 g8f7 c1d1 b7d7 a3a4 a7a6 a4a5 d5e6');
INSERT INTO evaluations VALUES(27604,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(27604,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(27604,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(27604,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(27604,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(27604,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(27604,36,'cp 24','e1g1 g8f6 b1c3 d5a5 c1e3 f8b4 h2h3 g4h5 d1b3 e8g8 a2a3 b4c3 b2c3 f6d5 c3c4 d5e3 f2e3 a5c7 a1b1 b7b6 b1c1 a8b8 c4c5 b6c5 b3c3 h5f3 e2f3 c6a5 c3c5 c7c5 d4c5 a5b3 c1c3 b3d2 f1e1 f8c8 e1c1 d2f3 g2f3');
INSERT INTO evaluations VALUES(28051,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(28051,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(28051,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(28051,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(28051,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(28051,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(28051,36,'cp 113','d7d6 f1c4 e7e6 d1e2 a7a6 c4b3 b8d7 e1c1 b6b5 h2h4 h7h6 a2a3 d7b6 e4e5 d6d5 f3e1 g8e7 e1d3 b6d7 g2g4 h6h5 f2f3 c7c5 d4c5 d8c7 c1b1 d7c5 h1e1 g7h6 f4h6 c5d3 h6g7 h8g8 e2d3 g8g7 d3d2 e8c8 c3e2 h5g4 f3g4');
INSERT INTO evaluations VALUES(34175,30,'cp 517',NULL);
INSERT INTO evaluations VALUES(34175,31,'cp 525',NULL);
INSERT INTO evaluations VALUES(34175,32,'cp 525',NULL);
INSERT INTO evaluations VALUES(34175,33,'cp 525',NULL);
INSERT INTO evaluations VALUES(34175,34,'cp 521',NULL);
INSERT INTO evaluations VALUES(34175,35,'cp 531',NULL);
INSERT INTO evaluations VALUES(34175,36,'cp 514','f3a8 a7a6 a8f3 e7e5 d4e5 c7e5 b1c3 g8f6 f3g3 e5g3 h2g3 b8d7 g1f3 d7c5 f3d4 c8b7 f2f3 f8d6 d4f5 d6f8 e3e4 c5b3 a1b1 b3c1 b1c1 g7g6 e4e5 g6f5 e5f6 h7h5 c1d1 h8h6 f1e2 b7c6 d1a1 h6f6 a1a6 f6e6 h1h5 b5b4 c3a2 f8d6 a6c6 d6g3 e1d1 e6c6');
INSERT INTO evaluations VALUES(72,30,'cp 508',NULL);
INSERT INTO evaluations VALUES(72,31,'cp 501',NULL);
INSERT INTO evaluations VALUES(72,32,'cp 503',NULL);
INSERT INTO evaluations VALUES(72,33,'cp 515',NULL);
INSERT INTO evaluations VALUES(72,34,'cp 501',NULL);
INSERT INTO evaluations VALUES(72,35,'cp 496',NULL);
INSERT INTO evaluations VALUES(72,36,'cp 514','a7a6 a8f3 g8f6 f3g3 c7b7 g1f3 h7h5 b1d2 h5h4 f3h4 g7g5 h4f3 g5g4 f3e5 f6h5 g3h4 b8c6 e5c6 b7c6 f2f3 f8g7 f3g4 h5f6 h4g5 e8f8 g5c5 c6c5 d4c5 f6g4 d2f3 h8h5 h2h3 g4e5 f3d4 e5d3 f1d3 c4d3 b2b4 h5g5 g2g4 c8b7 e1g1 g7d4 e3d4 g5g6 c1e3');
INSERT INTO evaluations VALUES(38515,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(38515,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(38515,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(38515,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(38515,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(38515,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(38515,36,'cp 55','a7a5 f1d3 b8a6 a2a3 g8f6 g1f3 e8g8 e1g1 c8d7 f1e1 f6e8 a1b1 a6c5 d3c2 f7f5 b2b4 a5b4 a3b4 c5e4 c2e4 f5e4 c1g5 g7f6 c3e4 f6g5 f3g5 h7h6 g5e6 d7e6 d5e6 e8f6 h2h3');
INSERT INTO evaluations VALUES(40473,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(40473,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(40473,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(40473,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(40473,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(40473,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(40473,36,'cp 46','e2e3 e8g8 d1b3 c6c5 d4c5 d8a5 f1e2 d5c4 b3c4 f6d5 e1g1 d5f4 e3f4 a5c5 c4c5 b4c5 f1d1 c5e7 a1c1 b8c6 e2b5 c6b4 f3e5 f8d8 d1d8 e7d8 g2g3 a7a6 b5e2 f7f6 e5c4 g8f8 e2f3 d8e7 c4b6 a8b8 b6c8');
INSERT INTO evaluations VALUES(40472,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(40472,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(40472,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(40472,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(40472,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(40472,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(40472,36,'cp 51','e8g8 d1b3 a7a5 a2a3 b4c3 b3c3 b7b6 f3e5 c8a6 a1c1 d5c4 f1c4 f6d5 c3d3 b6b5 c4b3 d5f4 e3f4 b5b4 b3c4 a6c4 c1c4 d8d6 e1g1 f7f6 e5f3 d6f4 f1e1 f4d6 a3b4 a5b4 h2h4 f8e8 d3c2 a8a5 c4c5 b4b3 c2b3');
INSERT INTO evaluations VALUES(41406,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(41406,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(41406,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(41406,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(41406,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(41406,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(41406,36,'cp 58','d2d4 e8g8 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 f1e2 c7c5 e1g1 c5d4 f3d4 b8c6 e2f3 d5c4 d4c6 b7c6 b2b3 c4a6 e3c5 f8e8 a1b1 c8f5 g2g4 f5e6 d1e2 a6a2 f3c6 a8c8 e2e4 a2a6 c6e8 c8c5 f1d1');
INSERT INTO evaluations VALUES(44633,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(44633,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(44633,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(44633,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44633,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(44633,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44633,36,'cp 24','e3d4 d7d5 f3e5 c8f5 f1b5 b8d7 e1g1 a7a6 b5d3 f5d3 e5d3 e7e6 c1f4 f8e7 c2c3 e8g8 a2a4 f6e8 b1d2 e7d6 f4e5 d6c7 d1e2 a8c8 g2g3 e8d6 e5f4 d6c4 d2c4 d5c4 d3e5 d7e5 d4e5 d8d3 e2d3 c4d3 f1d1 f8d8');
INSERT INTO evaluations VALUES(45855,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(45855,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(45855,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(45855,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(45855,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(45855,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(45855,36,'cp -22','d6d5 c4b3 f8c5 d2d3 h7h6 g5f3 e8g8 e1g1 a7a5 a2a4 f8e8 e4d5 c6d5 f3e5 e8e5 d3d4 c5d4 d1d4 b8c6 d4d1 d5d4 c3b5 c8g4 f2f3 g4e6 c1f4 e5f5 b3e6 f7e6 f4g3 f6h5 d1e2 d8d7 f1e1 e6e5 a1d1 a8d8 b5a3 d7e7 e2b5 h5g3 h2g3 f5g5 a3c4 g5g3 c4e5');
INSERT INTO evaluations VALUES(45797,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(45797,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(45797,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(45797,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(45797,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(45797,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(45797,36,'cp -23','c4b3 f8c5 d2d3 h7h6 g5f3 e8g8 e1g1 a7a5 a2a4 f8e8 e4d5 c6d5 f3e5 e8e5 d3d4 c5d4 d1d4 b8c6 d4d1 d5d4 c3b5 c8g4 f2f3 g4f5 c1f4 e5e7 d1d2 e7d7 d2f2 d4d3 c2d3 c6b4 d3d4 b4d3 f2d2 d3f4 d2f4 f5d3 f1d1 d3e2 d1d2 e2b5 a4b5 d8b6 b3c4 a8e8 a1d1 a5a4 f4g3');
INSERT INTO evaluations VALUES(48312,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48312,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(48312,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48312,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(48312,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(48312,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48312,36,'cp 1','e2e3 f8e7 c1b2 b7b6 f1d3 e8g8 e1g1 c8b7 b1d2 c7c5 d4c5 b6c5 c2c4 d5d4 e3d4 b8c6 a2a3 c6d4 f3d4 c5d4 d2f3 b7f3 d1f3 a7a5 f1d1 e6e5 f3f5 d8c7 d1e1 e7d6 b3b4 a5b4 a3b4 a8a1 b2a1 d6b4 e1e5');
INSERT INTO evaluations VALUES(48814,30,'cp -165',NULL);
INSERT INTO evaluations VALUES(48814,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(48814,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(48814,33,'cp -162',NULL);
INSERT INTO evaluations VALUES(48814,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(48814,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(48814,36,'cp -178','d5e4 b1c3 f8b4 c1e3 a7a5 g1e2 a5a4 b3c4 a4a3 a1b1 a3b2 e1g1 e4d3 d1d3 b4e7 d3d8 e7d8 b1b2 e8g8 f1d1 b8c6 b2b1 c8f5 c4d3 f5d3 c2d3 b7b6 h2h3 d8e7 a2a4 h7h6 d3d4 f8d8 d4e5 c6e5 e3f4 d8d1 b1d1 e7d6 c3b5');
INSERT INTO evaluations VALUES(48941,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(48941,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(48941,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(48941,33,'cp -180',NULL);
INSERT INTO evaluations VALUES(48941,34,'cp -164',NULL);
INSERT INTO evaluations VALUES(48941,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(48941,36,'cp -163','c1e3 f8b4 b1c3 a7a5 g1e2 a5a4 b3c4 a4a3 a1b1 a3b2 e1g1 e4d3 d1d3 b4e7 d3d8 e7d8 b1b2 e8g8 h2h3 b8c6 a2a4 b7b6 c4d3 h7h6 b2b1 c8b7 f1d1 d8e7 e2g3 f8d8 c3b5 d8d7 d3f5 d7d1 b1d1 e7d8 b5c3 g8f8 c3e4 f6e4 f5e4 a8a4');
INSERT INTO evaluations VALUES(50688,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(50688,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(50688,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(50688,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(50688,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(50688,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(50688,36,'cp 117','b1c3 g7g6 g1f3 g8f6 h2h3 f8g7 f1d3 b8d7 e1g1 e7e5 d4d5 d7c5 d3c2 e8g8 c1e3 f6h5 f1e1 h5f4 a2a4 a6a5 e3c5 d6c5 c2d3 g8h8 d3b5 b7b6 c3e2 f4h5 d5d6 a8a7 d1d2 f7f5 e2c3 f5e4 c3e4 f8f3 g2f3 h5f4 b5f1 a7d7 a1d1 d8h4 e1e3 c8b7');
INSERT INTO evaluations VALUES(53991,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(53991,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(53991,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(53991,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(53991,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(53991,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(53991,36,'cp 41','d2d4 c5d4 d1d4 g8f6 g1f3 f8e7 f1e2 b7b6 b2b3 c8b7 e1g1 e8g8 f1d1 b8d7 c1b2 f8e8 a1c1 a7a6 f3d2 a8c8 b3b4 h7h6 d4e3 b7c6 a2a3 e7f8 f2f3 d8c7 g1h1 d7e5 f3f4 e5g6 e3g3 c6a8 f4f5 g6e5 f5e6 e8e6');
INSERT INTO evaluations VALUES(55220,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(55220,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(55220,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(55220,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(55220,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(55220,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(55220,36,'cp -15','a7a6 a3c4 b7b5 c4e3 b8d7 g2g3 c8b7 f1g2 g8f6 e1g1 e7e6 a2a4 f8d6 b2b3 e8g8 c1b2 c7c5 d2d4 h7h6 d4c5 d6c5 f3e5 b7g2 e5d7 d8d7 d1d7 f6d7 e3g2 b5a4 f1d1 d7b6 b3a4 a6a5 b2d4 c5d4 d1d4 f8d8 d4d8 a8d8 g2f4 g8f8 g1g2 d8b8 f4d3 f8e7');
INSERT INTO evaluations VALUES(55479,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(55479,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(55479,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(55479,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(55479,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55479,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(55479,36,'cp 9','g8f6 c1f4 f8d6 f4e5 c8e6 f1d3 e8g8 e1g1 d6e5 f3e5 c7c5 d4c5 d8c7 f1e1 c7c5 d1d2 b8c6 e5c6 b7c6 h2h3 c5d6 a1d1 f8e8 b2b3 a8d8 c3e2 c6c5 d3b5 e6d7 b5d7 d8d7 e2g3 d5d4 e1e8 f6e8 d1e1 e8f6 g3f5 d6c7');
INSERT INTO evaluations VALUES(56906,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(56906,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(56906,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(56906,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(56906,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(56906,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(56906,36,'cp 59','d2d4 e5d4 d1d4 g8f6 b1c3 b8c6 d4e3 d8e7 f1g2 c6b4 e3d2 f8g7 g1h3 e8g8 h3f4 c7c6 e1g1 d7d5 a2a3 d5d4 a3b4 d4c3 b2c3 f6e4 d2e3 f8e8 b4b5 c8f5 a1a7 a8a7 e3a7 e4c3 b5c6 b7c6 a7e7 e8e7 c1e3 g6g5 f4h5 c3e2 g1h1');
INSERT INTO evaluations VALUES(225,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(225,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(225,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(225,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(225,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(225,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(225,36,'cp 51','g8e7 e1g1 e8g8 d4c2 b7b6 b1d2 c6e5 c4e2 e5g6 c2d4 g6f4 e2f3 c8a6 e4e5 f6e5 f3a8 a6f1 a8b7 f8b8 d4f3 e5d6 d1f1 b8b7 f1a6 b7b8 d2c4 d6f6 a6a7 b8e8 e3c5 f6g6 g2g3 b6c5 c4e5 f4h3 g1g2 g6c2 g2h3 d7d6 h3g2 d6e5 a7c7 c2b2 a1e1 b2a2 e1e5');
INSERT INTO evaluations VALUES(1259,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(1259,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(1259,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(1259,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(1259,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(1259,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(1259,36,'cp 18','g8f6 e1g1 d7d6 f2f4 f8e7 g1h1 e8g8 d1e1 c6d4 e3d4 b7b5 e1g3 c8b7 a2a3 a8d8 a1e1 d8d7 e2f3 f8c8 e1d1 d7d8 d1d3 b7c6 h2h3 c7b7 f1e1 h7h6 b2b4 g8h8 c3d1 d6d5 d4f6 e7f6 e4e5 f6e7 d1e3 d5d4');
INSERT INTO evaluations VALUES(661,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(661,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(661,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(661,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(661,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(661,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(661,36,'cp 18','a2a3 f8e7 f2f4 e8g8 e1g1 d7d6 d1e1 c6d4 e3d4 b7b5 e1g3 c8b7 g1h1 a8d8 a1d1 b7c6 e2f3 g8h8 f1e1 h7h6 h2h3 d8c8 d1d3 f8d8 g3f2 d8d7 e1d1 h8g8 f4f5 e6f5 e4f5 c6f3 f2f3 c7b7 f3b7 d7b7 d4f2 b7d7');
INSERT INTO evaluations VALUES(670,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(670,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(670,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(670,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(670,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(670,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(670,36,'cp 21','b7b5 d4c6 d7c6 f2f4 c8b7 e4e5 a8d8 e2d3 f6d7 d1g4 c6c5 e1c1 g7g6 d3e4 b5b4 g4f3 b7e4 c3e4 b4a3 b2a3 c7a5 e4f6 d7f6 d1d8 e8d8 e3d2 a5b5 f3a8 d8c7 e5f6 c5c4 a8a7 b5b7 d2a5 c7c8 a7b7 c8b7 h1d1 f8a3 c1b1 h8c8 d1d7 b7c6 d7f7');
INSERT INTO evaluations VALUES(3591,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3591,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3591,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3591,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3591,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3591,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3591,36,'cp 0','c5b6 c3a4 c6e7 a4b6 a7b6 h2h3 e7g6 f1e1 c7c6 c4b3 b6b5 c2c4 c8e6 c4b5 e6b3 d1b3 c6b5 d3d4 a8c8 c1b2 f8e8 b3d3 d8b6 a1d1 c8c4 f3d2 c4c8');
INSERT INTO evaluations VALUES(3562,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3562,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3562,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3562,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3562,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3562,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3562,36,'cp 0','c3a4 c6e7 f1e1 e7g6 a4b6 a7b6 h2h3 c7c6 c4b3 b6b5 c2c4 c8e6 c4b5 e6b3 d1b3 c6b5 d3d4 a8c8 c1b2 f8e8 b3d3 d8b6 a1d1 g6f4 d3d2 f4g6');
INSERT INTO evaluations VALUES(4054,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(4054,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(4054,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(4054,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(4054,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(4054,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(4054,36,'cp 51','e7e5 e1g1 e5d4 f3d4 f8e8 c1e3 d7c5 d1c2 a7a6 a1d1 c5d3 c2d3 b7b5 f2f3 b5c4 d3c4 c7c5 d4e2 c8e6 c4d3 a8b8 d3c2 d8e7 e3f2 f6d7 e2f4 d7e5 f4e6 f7e6 b2b3 c5c4 b3c4 e7c7 f3f4 e5c4 f4f5 e6f5 e4f5');
INSERT INTO evaluations VALUES(3960,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(3960,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(3960,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(3960,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(3960,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(3960,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(3960,36,'cp 65','e1g1 e5d4 f3d4 f8e8 c1e3 d7c5 d1c2 a7a6 a1d1 c5d3 c2d3 b7b5 d4c6 d8d7 c6a5 d7e6 c4c5 d6c5 e3c5 f6d7 c5d4 g7d4 d3d4 e6b6 d4b6 d7b6 f2f3 c8e6 b2b3 f7f5 a5c6 f5e4 c3e4 g8g7 d1c1 e6f5 f1e1 f5e4 e1e4 e8e4');
INSERT INTO evaluations VALUES(4996,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(4996,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(4996,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(4996,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(4996,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(4996,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(4996,36,'cp 93','d2d4 f7f6 c1e3 d8e7 e1g1 a7a5 c2c4 e5d4 d1d4 e7e6 b1d2 g8e7 c4c5 d6e5 f3e5 f6e5 d4c3 c8a6 f1d1 e8g8 c3b3 e6b3 d2b3 a6c8 a2a4 e7g6 e3d2 g6f4 d2a5 a8b8 a1a3 f4e2 g1h1 f8f2 a5c7 b8b4 b3d2 b4b2 d2f3 c8a6 a4a5 b2b4 f3e5 b4e4 e5d7 e2d4');
INSERT INTO evaluations VALUES(4994,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(4994,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(4994,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(4994,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(4994,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(4994,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(4994,36,'cp 63','f7f6 c1e3 d8e7 e1g1 e7e6 c2c4 e5d4 e3d4 g8e7 c4c5 d6f4 f1e1 e7g6 d4e3 f4e3 e1e3 g6f4 h2h3 e8g8 b1d2 d7d6 c5d6 e6d6 d1b3 c8e6 b3c2 c6c5 e4e5 f6e5 d2e4 e6f5 c2c4 d6d5 c4c5 d5c5');
INSERT INTO evaluations VALUES(5407,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(5407,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(5407,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(5407,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(5407,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(5407,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(5407,36,'cp -22','d4b3 g8f6 b1c3 e8g8 f1e2 d7d6 a2a4 c8e6 e1g1 e6b3 c2b3 e7e6 f4f5 d6d5 f5e6 f7e6 e4d5 f6d5 c3d5 f8f1 e2f1 e6d5 g2g3 g8h8 c1f4 g7b2 a1b1 b2f6 f1g2 d8b6 g1h1 a8d8 d1d3 b6b4 b1c1 c6d4 f4e3 d4f5 e3f2 b4g4');
INSERT INTO evaluations VALUES(12568,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12568,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(12568,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(12568,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12568,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(12568,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12568,36,'cp 16','d7d6 e1g1 g7g6 c2c4 f8g7 a2a3 a7a5 b1c3 e8g8 a1b1 a8b8 c1d2 c8d7 f3e1 c6d4 a3a4 f6e8 c3b5 d4e6 e1c2 f7f5 e4f5 g6f5 f2f4 e5f4 g3f4 d7c6 c2e3 f8f6 g1h1 f6g6 g2d5 e8f6 b5c3 c6d5 c3d5 f6d5 e3d5 g8h8 b2b3 e6d4');
INSERT INTO evaluations VALUES(13875,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(13875,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(13875,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(13875,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(13875,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13875,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(13875,36,'cp 12','e1g1 g7g6 c2c4 a7a6 a2a3 f8g7 b1c3 a8b8 a1b1 e8g8 b2b4 c5b4 a3b4 b7b5 c4b5 a6b5 c1e3 c8e6 d1c2 h7h6 d3d4 e6c4 d4d5 c4f1 g2f1 c6e7 c3b5 b8c8 c2d3 e7d5 e4d5 e5e4 d3d1 e4f3 d1f3 d8d7 b5d4 d7g4 f3g2 g4d7 f1b5 d7c7 d4c6 c7d7');
INSERT INTO evaluations VALUES(13874,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13874,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13874,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13874,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(13874,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(13874,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13874,36,'cp 9','g7g6 a2a3 f8g7 c2c4 a7a6 b1c3 e8g8 c1e3 a8b8 a1b1 b7b5 c4b5 a6b5 b2b4 c5b4 a3b4 c8e6 d1c1 f8e8 c3b5 d8d7 b5c7 d7c7 b4b5 b8c8 b5c6 c7c6 c1c6 c8c6 d3d4 e6a2 b1a1 c6a6 f3g5 e5d4 e3d4 e8c8 f1d1 f6g4 d4g7 g8g7');
INSERT INTO evaluations VALUES(13622,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13622,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(13622,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(13622,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(13622,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13622,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(13622,36,'cp 12','h7h6 g5f6 d8f6 c4d5 e6d5 c3d5 f6e6 d1e2 f8b4 d5b4 c6b4 a1d1 e8g8 e2e6 c8e6 a2a3 b4d5 f1c4 a8c8 c4b3 f8d8 e1g1 e6g4 f1e1 g8f8 d1d3 d5f4 d3e3 g4f3 e3f3 d8d4 g2g3 d4d3 f3d3 f4d3 e1d1 d3b2 d1d7 c8c1 g1g2 b2c4 a3a4');
INSERT INTO evaluations VALUES(13647,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(13647,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(13647,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(13647,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(13647,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(13647,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(13647,36,'cp 24','d3c2 f8d6 b1d2 b7b5 f4d6 d8d6 e1g1 e8g8 f1e1 c8b7 e3e4 d5e4 d2e4 f6e4 c2e4 h7h6 d4d5 e6d5 d1d5 d6d5 e4d5 c6a5 e1e5 a7a6 f3d4 f8e8 a1e1 e8e5 e1e5 a8d8 d5b7 a5b7 g1f1 d8c8 e5e7 b7c5 e7a7 g7g6 g2g3 h6h5 f1e2 c8e8 e2f3');
INSERT INTO evaluations VALUES(17629,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(17629,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(17629,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(17629,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(17629,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(17629,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(17629,36,'cp 24','g8e7 e1g1 e7f5 b1a3 c5d4 c3d4 f8e7 a3c2 e8g8 e2d3 f7f6 c1f4 f6e5 f3e5 c8d7 f1e1 a8c8 e5c6 c8c6 f4e5 d7e8 c2e3 f5e3 e1e3 e7f6 a1c1 c6c1 d1c1 f6e5 e3e5 d8b6 c1e3 b6b2 e5e6 e8b5 d3b5 a6b5 e6e7 b2a2');
INSERT INTO evaluations VALUES(21459,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(21459,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(21459,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(21459,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(21459,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(21459,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(21459,36,'cp -31','d5c4 g1f3 f8b4 c1d2 a7a5 a2a3 b4d2 d1d2 b7b5 b1c3 c8a6 a3a4 b5a4 a1a4 c6b4 a4a5 g8e7 f1e2 b4d3 e2d3 c4d3 e1g1 e8g8 b2b4 a6b7 d2d3 a8b8 c3e4 b7e4 d3e4 b8b4 f1b1 b4b1 e4b1 h7h6 h2h3 c7c6 a5a4 e7f5 b1e4 d8d7 a4c4 f5e7 c4c1 d7a7 h3h4 f8b8 h4h5 a7a2 g2g3');
INSERT INTO evaluations VALUES(27958,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(27958,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(27958,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(27958,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(27958,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(27958,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(27958,36,'cp -27','g8f6 d1e2 c7c5 d4c5 b6c5 c1d2 b4c3 d2c3 c5c4 d3c4 f6e4 e2g4 d8b6 g4g7 h8f8 g1e2 b6f2 e1d1 e4c3 b2c3 b7g2 h1g1 g2f3 g7g3 f2g3 g1g3 f3e2 c4e2 b8c6 a1b1 e8e7 g3h3 f8g8 b1b7 g8g1 d1d2 g1g2 h3d3 a8d8 d3h3 d8b8 b7b8 c6b8 h3h7 b8a6 h7h5 a6c7 d2e3 d7d5 e2d3');
INSERT INTO evaluations VALUES(28260,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(28260,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(28260,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(28260,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(28260,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(28260,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(28260,36,'cp -29','d1e2 c7c5 d4c5 b6c5 g1f3 c5c4 d3c4 b7e4 e1g1 e4f3 e2f3 d7d5 c4d3 e8g8 c1e3 b8d7 g1h1 d8c7 c3b5 c7b7 a2a3 b4c5 b2b4 c5e3 f3e3 a7a6 b5d4 f8c8 a1c1 b7b6 c2c4 d5c4 d3c4 c8c7 e3d3 a6a5 c4b5 a5b4 a3b4 a8c8 c1c7 c8c7 f1d1 g7g6 h2h3');
INSERT INTO evaluations VALUES(28259,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(28259,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(28259,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(28259,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(28259,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(28259,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(28259,36,'cp -32','c7c5 d4c5 b6c5 g1f3 c5c4 d3c4 b7e4 e1g1 b4c3 b2c3 e8g8 c4d3 e4g6 c1e3 f6d5 c3c4 d5e3 e2e3 b8c6 a1b1 h7h6 f3d4 g6d3 c2d3 d8a5 b1b5 a5c3 d4c6 d7c6 b5b3 c3a5 e3f2 a8d8 h2h3 c6c5 b3b5 a5a6 f4f5 d8d3 f5e6 a6e6 b5c5');
INSERT INTO evaluations VALUES(28456,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(28456,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(28456,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(28456,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(28456,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(28456,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(28456,36,'cp -8','e8g8 f1d3 d6g3 h2g3 b8d7 e1g1 f5d3 d1d3 c7c5 b1d2 a8c8 f1c1 d8b6 c3c4 c5d4 d3d4 b6d4 e3d4 d5c4 d2c4 f6e4 f3e5 c8d8 c1c2 d7e5 d4e5 d8d7 c4d6 e4d6 a1d1 g7g6 d1d6 f8d8 d6d7 d8d7 c2c8 g8g7 g3g4 d7d2');
INSERT INTO evaluations VALUES(29008,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29008,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(29008,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(29008,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(29008,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(29008,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(29008,36,'cp 45','b1c3 e7e6 g2g4 h5g6 f3e5 b8d7 h3h4 d5c4 e5g6 h7g6 g4g5 f6d5 f1c4 d7b6 c4e2 d8c7 c1d2 f8e7 f2f4 b6c8 d1c2 c8d6 e1c1 e8c8 e3e4 d5c3 d2c3 d6b5 e2b5 c6b5 c1b1 b5b4 c3e1 e7d6 c2c7 d6c7 e1b4 c7f4 b4c3 f7f5 g5f6 g7f6 d4d5');
INSERT INTO evaluations VALUES(29006,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(29006,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(29006,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(29006,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(29006,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(29006,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29006,36,'cp 26','e7e6 d1b3 d8b6 f3e5 f8e7 c1d2 b8d7 g2g4 h5g6 e5g6 h7g6 f2f3 d5c4 f1c4 f6d5 e1c1 b6b3 c4b3 g6g5 a2a4 e8c8 d1f1 h8h4 c3d5 c6d5 f3f4 d8h8 f4g5 h4h3 h1h3 h8h3 f1f7 e7g5 f7g7 g5e3 d2e3 h3e3 b3c2');
INSERT INTO evaluations VALUES(30436,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(30436,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(30436,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(30436,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(30436,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(30436,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(30436,36,'cp 82','b5b4 f1c4 g8f6 c3d5 f6d5 c4d5 d7c6 d5f3 e7e6 c1d2 c6f3 g1f3 b8d7 a4a5 f8d6 d1a4 a8b8 e3e4 e8g8 e4e5 d6e7 a4c6 b8c8 e1g1 d7b8 c6b7 c7c5 b7a7 c5c4 a7b6 c4c3 b2c3 b4c3 d2e3 b8c6 b6d8 f8d8 f1c1 c6b4 a5a6 f7f6 a6a7 c3c2 g1h1 g8f7 e5f6 g7f6');
INSERT INTO evaluations VALUES(30738,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(30738,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(30738,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(30738,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(30738,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(30738,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(30738,36,'cp 121','b1c3 f7f5 g2g3 g8f6 f1g2 f8g7 e1g1 e8g8 c1g5 b8a6 a2a3 f6e4 c3e4 b7e4 d1d2 c7c6 g5h6 e4f3 g2f3 g7h6 d2h6 a6c7 b2b4 c7e8 f1c1 a8c8 h6d2 e7e6 a3a4 e8f6 c4c5 a7a6 e2e3 f6d5 f3d5 e6d5 c5b6 d8b6');
INSERT INTO evaluations VALUES(35984,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(35984,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(35984,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35984,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35984,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35984,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35984,36,'cp 0','h7h6 g5e3 f6g4 e3c1 b8c6 d4e2 a7a6 a2a4 g4e5 c4b3 e7e6 h2h3 c6a5 b3a2 b7b5 a4b5 a6b5 c3b5 d8b6 b5c3 c8a6 g1h1 e5d7 a1b1 b6b4 c1e3 f8c8 f1e1 d7f6 e2d4 a6c4 d1f3 c4a2 c3a2 b4b7 a2c3 a5c4 e3h6 g7h6 f3f6');
INSERT INTO evaluations VALUES(36526,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36526,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(36526,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36526,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36526,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36526,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(36526,36,'cp 0','b8d7 e2e3 b7b6 f1d3 c8b7 e1g1 f8e8 f4g5 e7e5 d3c2 e5d4 f3d4 d7c5 f1e1 e8e5 g5f4 e5e8');
INSERT INTO evaluations VALUES(4027,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4027,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4027,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4027,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(4027,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(4027,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4027,36,'cp 20','e2e3 b7b6 f1d3 c8b7 e1g1 f8e8 f4g5 e7e5 d3c2 e5d4 e3d4 h7h6 g5f4 f6e4 c3e4 b7e4 f1e1 d7f6 a1c1 d8d7 c2a4 e4c6 e1e8 a8e8 a4c6 d7c6 f4e3 c6d7 d4d5 g6g5 b2b3 g5g4 h3g4 d7g4 e3d4 g4f4 d4b2 f6g4 b2g7 g8g7');
INSERT INTO evaluations VALUES(4024,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4024,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4024,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(4024,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(4024,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(4024,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4024,36,'cp 12','b7b6 f1d3 c8b7 e1g1 f8e8 f4g5 e7e5 d3c2 e5d4 e3d4 h7h6 g5f4 f6e4 c3e4 b7e4 f1e1 e4b7 e1e8 d8e8 a1c1 e8e7 b2b3 d6d5 c4d5 b7d5 c2d3 c7c5 d3b5 a8d8 b5d7 d8d7 d4c5 b6c5 d1e1 e7e1 f3e1 g6g5 f4e3 g7d4 e3d4 c5d4 f2f3');
INSERT INTO evaluations VALUES(37040,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37040,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(37040,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37040,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37040,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(37040,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(37040,36,'cp 0','g1f3 c7c5 a2a3 b4c3 b2c3 b8c6 f1e2 d5c4 e2c4 c6a5 c4e2 c5d4 c3d4 d8d5 c1d2 a5b3 a1b1 b3d2 c2d2 b7b6 f3e5 f6e4 d2b4 d5a2 b1b2 a2a1 b2b1');
INSERT INTO evaluations VALUES(40089,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40089,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40089,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40089,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40089,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(40089,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40089,36,'cp 4','e8g8 f1d3 b7b6 e1g1 c8b7 c4d5 c6d5 d1e2 f6e4 c1d2 b8d7 f1d1 e4d2 d1d2 a7a6 d2c2 b6b5 a3a4 d8a5 e2d1 b5b4 c3e2 a8c8 c2c8 f8c8 e2c1 g7g6 c1b3 a5b6 a1c1 c8c7 a4a5 b6d6 h2h3 c7c1 b3c1');
INSERT INTO evaluations VALUES(35760,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(35760,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(35760,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35760,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(35760,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(35760,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(35760,36,'cp 24','f1d3 b7b6 e1g1 c8b7 c4d5 c6d5 c1d2 a7a6 d1e2 b8c6 f1d1 b6b5 d2e1 c6a5 h2h3 e7d6 c3a2 d8b6 e1b4 f8c8 d1c1 a5c4 b4d6 b6d6 c1c2 b7c6 a1c1 c6d7 a2c3 h7h6 f3d2 a8b8 d2b3 d7e8 b3c5 d6e7');
INSERT INTO evaluations VALUES(41412,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(41412,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(41412,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(41412,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(41412,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(41412,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(41412,36,'cp 44','d2d4 c7c5 d4c5 d8a5 f1d3 a5c5 d1e2 e8g8 c1e3 c5a5 e1g1 c8g4 h2h3 g4f3 e2f3 b8c6 a1b1 f6d7 e3d2 e7e6 g1h1 a8c8 b2b4 a5c7 c3b5 c7b8 c2c4 a7a6 b5a3 b8c7 a3c2 a6a5 a2a3 b7b6 f3f2 a5b4 a3b4 c8a8 d2e3');
INSERT INTO evaluations VALUES(44301,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44301,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(44301,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44301,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(44301,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(44301,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(44301,36,'cp 27','e3d4 b7b6 g1f3 b8c6 a2a4 c8b7 b1a3 f6h5 f4c1 f8d6 f3e5 h5f6 a3c4 d6c7 e5c6 b7c6 c4e5 c6b7 f1d3 e8g8 e1g1 d7d6 e5c4 e6e5 c1g5 e5e4 d3e2 h7h6 g5h4 d6d5 c4e5 d8d6 f2f4 g8h8 h2h3 a8e8 f1e1 a7a6');
INSERT INTO evaluations VALUES(44293,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44293,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(44293,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(44293,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(44293,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(44293,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(44293,36,'cp 19','b7b6 g1f3 b8c6 f1d3 c8b7 e1g1 c6e7 f4g3 e7g6 f1e1 f8e7 f3h4 e8g8 a2a4 f6d5 h4g6 h7g6 h2h3 d5f6 g3h2 d7d6 b1a3 b7c6 a3c2 f6d5 c2e3 d5e3 f2e3 a7a6 d1e2 d8d7 d3a6 c6a4 a6d3 a4c6 e3e4 a8a1');
INSERT INTO evaluations VALUES(47222,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(47222,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(47222,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(47222,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(47222,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(47222,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(47222,36,'cp -72','d5d4 c3d5 f6d5 e4d5 g7g6 f1b5 b8d7 g1e2 f8g7 c2c3 a7a6 b5d7 d8d7 e2d4 g7d4 c3d4 d7f5 f3f5 c8f5 b2b3 e8c8 c1b2 d8d5 f2f3 h7h5 e1g1 f5e6 f1e1 c8d7 b2c3 g6g5 e1e5 f7f6 e5e3 e6f7 a1e1 e7e6 b3b4 h8e8 a2a4 e8g8 g1f2 g5g4 f3g4 g8g4 g2g3 h5h4');
INSERT INTO evaluations VALUES(48109,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(48109,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(48109,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(48109,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(48109,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(48109,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(48109,36,'cp 42','f4e5 b8d7 c2c4 h5f6 b1c3 c7c6 e5g3 f6h5 g3h4 f8e7 h4e7 d8e7 f1d3 e8g8 e1g1 h5f6 a2a4 a7a5 f1e1 d5c4 d3c4 e6e5 c4a2 e5d4 e3d4 e7b4 d1c2 h7h6 h2h3 d7b6 c3e4 f6e4 c2e4 b6d5 a2d5 c6d5 e4d5 c8e6 d5c5 b4c5 d4c5 f8d8 a1d1 a8c8');
INSERT INTO evaluations VALUES(48891,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(48891,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(48891,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(48891,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(48891,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(48891,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48891,36,'cp -29','d2d4 e5e4 c2c4 c7c6 b1c3 f8d6 c1d2 e8g8 c4d5 c6d5 d1b3 d6c7 c3b5 c7b6 d2b4 f8e8 b5d6 e8e6 d6c8 d8c8 b4d2 b8c6 g1e2 c6a5 b3a4 a5c4 e2f4 e6c6 f1c4 c6c4 a4b5 c8d8 a1c1 a7a6 b5b3 b6c7 c1c4 d5c4 b3c4 c7f4 e3f4 a8c8 c4b4');
INSERT INTO evaluations VALUES(49970,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(49970,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(49970,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(49970,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(49970,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49970,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(49970,36,'cp 0','f8b4 d1d2 d7d5 h1g1 b8c6 e2e3 c8d7 f1b5 c6e7 b5d7 d8d7 f3e5 f6e4 e5d7 e4d2 g5e7 b4c3 b2c3 d2c4 d7e5 c4e5 d4e5 e8e7 g2g4 f5f4 e3f4 a8f8 f4f5 h7h5 c3c4 h5g4 g1g4 f8f5 g4g7 e7d8 e1c1 f5e5 f2f4 e5f5 c4d5 e6d5 g7g5 f5f4 g5d5 d8c8 d5d2 f4e4');
INSERT INTO evaluations VALUES(50579,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(50579,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(50579,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(50579,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(50579,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(50579,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(50579,36,'cp 106','g1f3 d7d6 c2c4 g7g6 b1c3 f8g7 f1e2 g8f6 d1c2 b8a6 e1g1 e8g8 h2h3 a6b4 c2a4 f6d7 c1e3 f7f5 e4f5 g6f5 e3g5 d8e8 a2a3 b4a6 a4a5 h7h6 g5h4 e7e5 a1d1 e5e4 f3e1 f5f4 e2h5 e8e6 h5g4 e6g6 f2f3 e4e3 c3e2');
INSERT INTO evaluations VALUES(50578,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(50578,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(50578,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(50578,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(50578,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(50578,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(50578,36,'cp 110','d7d6 c2c4 g7g6 h2h3 g8f6 b1c3 f8g7 c1e3 e8g8 d1c2 b8a6 f1e2 e7e5 d4e5 d6e5 e1g1 d8e7 c3a4 a6b4 c2b1 f6h5 e3c5 e7f6 a2a3 h5f4 f1e1 b4a6 c5f8 g7f8 a4b6 a8b8 b1c2 c8e6 a1d1 a6c5 b2b4 a5b4 a3b4 c5a6 c4c5 a6b4 c2c3 b4a2 c3a3');
INSERT INTO evaluations VALUES(50905,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(50905,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(50905,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(50905,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(50905,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(50905,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(50905,36,'cp 75','c2c3 c7c5 f1d3 b8c6 e1g1 c8d7 f1e1 c5d4 c3d4 d8b6 d3c2 c6b4 c2b3 g8e7 b1c3 e7f5 h2h3 b4c6 c3e2 f8b4 e1f1 b4e7 b3c2 c6b4 c2b1 b4c6 a2a3 c6a5 b1c2 g7g6 f1e1 e8c8 c1d2 d7b5 c2f5 g6f5 d2a5');
INSERT INTO evaluations VALUES(2479,30,'cp -401',NULL);
INSERT INTO evaluations VALUES(2479,31,'cp -404',NULL);
INSERT INTO evaluations VALUES(2479,32,'cp -390',NULL);
INSERT INTO evaluations VALUES(2479,33,'cp -450',NULL);
INSERT INTO evaluations VALUES(2479,34,'cp -410',NULL);
INSERT INTO evaluations VALUES(2479,35,'cp -466',NULL);
INSERT INTO evaluations VALUES(2479,36,'cp -436','g5h3 c8g4 c2c3 d8e8 d1b3 a8d8 h3g1 g4f5 b3d1 f8e7 g1e2 e8g6 e1g1 f5d3 f1e1 d3c2 d1d2 c2e4 e2g3 e4f3 g2f3 g8h7 d2e2 h8f8 c1e3 g6f7 g1g2 e7d6 c3c4 e5e4 f3e4 d5d4 e3d2 d6g3 f2g3 d4d3 e2e3 c6e5 e1f1 f7c4 a1c1');
INSERT INTO evaluations VALUES(2480,30,'cp -477',NULL);
INSERT INTO evaluations VALUES(2480,31,'cp -466',NULL);
INSERT INTO evaluations VALUES(2480,32,'cp -485',NULL);
INSERT INTO evaluations VALUES(2480,33,'cp -481',NULL);
INSERT INTO evaluations VALUES(2480,34,'cp -460',NULL);
INSERT INTO evaluations VALUES(2480,35,'cp -474',NULL);
INSERT INTO evaluations VALUES(2480,36,'cp -458','c8g4 c1e3 d8e8 e1f1 e8h5 d3d4 e5d4 h3f4 h5f7 e3d2 f8d6 h2h3 g4f5 f4e2 a8f8 d2f4 d4d3 c2d3 d6f4 e2f4 f5h7 f4e2 c6e5 e2c1 h7g6 d3d4 e5f3 d1f3 f7f3 g2f3 f8f3 c1b3 g8h7 a1c1 f3f7 c1c3 h8f8 h1h2 f8e8 b3d2 f7f4 d2f3 c7c6 h3h4 g6e4 h2h3 e4f3 h3f3');
INSERT INTO evaluations VALUES(23572,30,'cp -449',NULL);
INSERT INTO evaluations VALUES(23572,31,'cp -454',NULL);
INSERT INTO evaluations VALUES(23572,32,'cp -483',NULL);
INSERT INTO evaluations VALUES(23572,33,'cp -488',NULL);
INSERT INTO evaluations VALUES(23572,34,'cp -477',NULL);
INSERT INTO evaluations VALUES(23572,35,'cp -485',NULL);
INSERT INTO evaluations VALUES(23572,36,'cp -480','c1e3 d8e8 e1f1 e8h5 d3d4 e5d4 h3f4 g4f3 g2f3 h5f5 e3c1 f8d6 d1d3 a8f8 d3f5 f8f5 f4g6 h8h7 h2h4 h6h5 h1g1 f5f3 f1e2 f3f6 f2f4 d4d3 e2d3 g8f7 g6e5 d6e5 f4e5 c6e5 d3e2 h7h8 c1g5 e5f3 g5f6 f3g1 a1g1 g7f6 e2f3 h8e8');
INSERT INTO evaluations VALUES(3863,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(3863,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(3863,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(3863,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(3863,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3863,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(3863,36,'cp 21','c3a4 c5b6 a4b6 a7b6 c2c3 d6d5 d1c2 d8d6 c1g5 d5e4 d3e4 f6h5 f1e1 d6g6 g5c1 g6f6 c1e3 h5f4 e2f1 h7h6 a2a4 f8d8 e1b1 c6e7 b2b4 e7g6 f3e1 c8d7 b1d1 h6h5 a4a5 b6a5 a1a5 a8a5 b4a5 f6g5 h2h3 h5h4 g1h2 d7c6 d1d8 g5d8');
INSERT INTO evaluations VALUES(5161,30,'cp 551',NULL);
INSERT INTO evaluations VALUES(5161,31,'cp 549',NULL);
INSERT INTO evaluations VALUES(5161,32,'cp 561',NULL);
INSERT INTO evaluations VALUES(5161,33,'cp 555',NULL);
INSERT INTO evaluations VALUES(5161,34,'cp 565',NULL);
INSERT INTO evaluations VALUES(5161,35,'cp 589',NULL);
INSERT INTO evaluations VALUES(5161,36,'cp 572','d4b5 c6d4 b5d4 e5d4 e3d4 g8f6 e4e5 f6h5 d4c5 g7e5 c3e4 d8a5 c2c3 f7f6 e4d6 e8d8 d6f7 d8e8 c5b4 a5b4 c3b4 e8f7 g2g3 h8e8 d1d5 f7g7 e1c1 d7d6 f2f4 h5f4 g3f4 e5f4 c1b1 c8g4 d5b7 g7h8 f1b5 a8b8 b7d5 e8e5 d5d6 e5b5 d6f4 g4d1 f4f6 h8g8 h1d1 b5b7 d1e1 g6g5 h2h4 b7g7 f6h6 g5h4 h6d6');
INSERT INTO evaluations VALUES(5160,30,'cp 594',NULL);
INSERT INTO evaluations VALUES(5160,31,'cp 582',NULL);
INSERT INTO evaluations VALUES(5160,32,'cp 577',NULL);
INSERT INTO evaluations VALUES(5160,33,'cp 579',NULL);
INSERT INTO evaluations VALUES(5160,34,'cp 594',NULL);
INSERT INTO evaluations VALUES(5160,35,'cp 578',NULL);
INSERT INTO evaluations VALUES(5160,36,'cp 570','c6d4 b5d4 e5d4 e3d4 g8f6 e4e5 f6h5 d4c5 g7e5 c3e4 d8a5 c2c3 f7f6 e4d6 e8d8 d6f7 d8e8 c5b4 a5b4 c3b4 e8f7 g2g3 f7g7 f1g2 d7d5 d1d5 e5b2 e1g1 b2a1 f1a1 h8e8 a1c1 g7h8 b4b5 a8b8 d5d4 a7a6 g2f3 c8h3 f3h5 e8e6 h5g6 h7g6 d4h4 h8g8 h4h3 e6e7 h3h6 e7g7 b5b6 g6g5 c1e1 b8d8 h2h4 g5h4 h6h4');
INSERT INTO evaluations VALUES(8365,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(8365,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(8365,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(8365,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(8365,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(8365,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(8365,36,'cp -47','a7a6 b5c6 b7c6 d1c2 e8g8 c3e2 d8d6 e1g1 c8g4 b2b3 f8e8 c1b2 c5b6 a1c1 c6c5 e2g3 a8c8 c2d3 c8d8 f3d2 h7h6 h2h3 g4c8 d2f3 a6a5 f1e1 f6e4 g3e4 e8e4 e1d1 e4e8 f3d2 e8e6 d2f3');
INSERT INTO evaluations VALUES(10600,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(10600,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(10600,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(10600,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(10600,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(10600,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(10600,36,'cp 63','g4e5 f3e5 c6e5 f2f4 e5g6 d1d3 e8g8 b1c3 f8e8 c1d2 c7c6 e1c1 d7d5 c4d5 c6d5 c3d5 g6e7 d5c3 d8d3 e2d3 c8g4 d1e1 e8d8 d3e2 g4e6 e3e4 f7f6 e1d1 e7c6 f4f5 e6f7 d2f4 c6d4 e2d3 a8c8 c1b1 a7a6 g2g4 c5d6 f4e3 d6e5 c3d5 g8h8 h1e1');
INSERT INTO evaluations VALUES(10608,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(10608,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(10608,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(10608,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(10608,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(10608,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(10608,36,'cp 51','f3e5 c6e5 f2f4 e5g6 e1g1 e8g8 g1h1 f8e8 b1c3 b7b6 f4f5 g6e5 f5f6 c8b7 f6g7 d8h4 e3e4 c5f2 d1d2 c7c5 d2g5 h4g5 c1g5 f2d4 c3b5 b7e4 a1d1 d4b2 b5d6 e4g6 d6e8 a8e8 g5f6 e8e6 h2h4 h7h5 f6e5 e6e5 e2d3 b2d4 d3g6');
INSERT INTO evaluations VALUES(11028,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(11028,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(11028,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(11028,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(11028,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(11028,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(11028,36,'cp 192','f3g5 d7d5 e4d5 c6a5 d1h5 e7g6 b1c3 f7f6 g5e4 c5e7 c1d2 a5c4 d3c4 e8g8 e1c1 d8e8 c3b5 c8d7 h5e2 d7b5 c4b5 f6f5 e4c3 e7d6 c1b1 e5e4 h1e1 e8d7 a2a4 a8e8 d2e3 a7a6 b5a6 b7a6 e2c4 g6e5 c4b3 d7f7 e3d4');
INSERT INTO evaluations VALUES(11022,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(11022,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(11022,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(11022,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(11022,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(11022,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(11022,36,'cp 188','d7d5 e4d5 e7d5 b1c3 c8e6 g5e6 f7e6 d1g4 d8e7 e1g1 e7f7 c1d2 a8d8 c3e4 c5e7 c4b5 e8g8 b5c6 b7c6 a1e1 f7f5 g4g3 d5f4 f2f3 f5h5 g3g4 h5g4 f3g4 f4d5 f1f8 d8f8 g4g5 e7d6 a2a3 g8f7 g2g3 f7e7 g1g2');
INSERT INTO evaluations VALUES(11140,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(11140,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(11140,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(11140,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(11140,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(11140,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(11140,36,'cp 13','c8g4 h2h3 g4f3 d1f3 e7e6 f1d3 c6b4 d3b1 h7h6 e1g1 f8d6 a2a3 b4c6 b1d3 e8g8 f4d6 d8d6 c3a4 a8c8 a1c1 f8d8 f3e2 e6e5 d4e5 c6e5 c1c8 d8c8 f1d1 b7b5 a4c3 e5c4 a3a4 d6c5 a4b5 a6b5 c3a2 b5b4 b2b3 c4e5 e2d2 d5d4 e3d4 c5d4 d2b4');
INSERT INTO evaluations VALUES(24253,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(24253,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(24253,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24253,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(24253,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24253,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24253,36,'cp 19','h2h3 g4f3 d1f3 e7e6 f1d3 c6b4 d3b1 a8c8 e1g1 f8d6 a2a3 b4c6 c3a4 e8g8 a4c5 d6f4 f3f4 d8e7 f1c1 e6e5 d4e5 c6e5 b2b4 b7b6 c5d3 c8c1 d3c1 f8c8 b1d3 e5d3 c1d3 f6e4 a1c1 e7d7 c1c8 d7c8 d3e5 c8c1 g1h2 c1c7');
INSERT INTO evaluations VALUES(12889,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(12889,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(12889,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(12889,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(12889,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(12889,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(12889,36,'cp -21','d7d5 e4d5 f6d5 c3d5 d8d5 e3b3 c8e6 b3d5 e6d5 c1e3 e7e6 g1e2 c6b4 e2d4 e6e5 f1b5 e8d8 a2a3 e5d4 a3b4 d5g2 h1g1 d4e3 g1g2 d8c7 f2e3 f8b4 c2c3 b4f8 a1d1 a8d8 d1d8 c7d8 g2d2 d8c8 b5c4 f8c5 c4f7 c5e3 f7e6 c8b8 d2d7 h8e8 e6g4 e3g5 e1d1 e8e7 d7e7 g5e7 h2h3 g7g6 g4f3');
INSERT INTO evaluations VALUES(18049,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(18049,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(18049,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(18049,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(18049,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(18049,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(18049,36,'cp 159','c4d5 g8f6 f1e1 f8e7 d5c6 d7c6 f3e5 c8b7 d2d3 c6c5 c2c4 f6d7 e5f3 b5b4 a2a3 a6a5 a3b4 a5b4 a1a8 d8a8 c1f4 a8a2 b2b3 e8g8 b1d2 f8a8 f4c7 d7f8 c7e5 f8e6 h2h4 h7h6 h4h5 a8a3 d1c1 e7f8 e5b2');
INSERT INTO evaluations VALUES(18044,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(18044,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(18044,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(18044,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(18044,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(18044,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(18044,36,'cp 164','g8e7 d2d4 d7d6 f1e1 a8b8 c2c3 h7h6 b1d2 g7g5 d1b3 e7d5 e4d5 c6e7 d4e5 f8g7 h2h4 g5h4 e5d6 c7d6 f3h4 e8g8 d2f3 e7g6 e1e4 g6e5 c1f4 e5d3 f4e3 d3c5 e3c5 d6c5 a1d1 b8b6 a2a4 b6d6 a4b5 a6b5 d1d2 g7f6 c3c4 b5c4 b3c4 f8e8 e4e8 d8e8 d2e2');
INSERT INTO evaluations VALUES(18053,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(18053,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(18053,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(18053,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(18053,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(18053,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(18053,36,'cp 59','g8f6 a2a4 a8b8 b1c3 f8c5 a4b5 a6b5 c1g5 e8g8 e1g1 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 d7d6 d3d4 c5b6 h2h3 b5b4 b3c2 b4c3 b2c3 e5d4 c3d4 c6e7 d5b6 b8b6 d1d3 c7c5 f1d1 c5d4 d3d4 c8e6 a1a7 e7c6 d4d6 d8d6 d1d6');
INSERT INTO evaluations VALUES(19018,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19018,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19018,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19018,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19018,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(19018,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19018,36,'cp 0','d2d3 g8e7 c2c4 g7g6 b1c3 f8g7 f1g2 d7d6 a2a3 a7a5 e1g1 e8g8 a1b1 c8e6 c1d2 b7b6 c3d5 e6c8 d5c3');
INSERT INTO evaluations VALUES(19199,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(19199,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(19199,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(19199,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(19199,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(19199,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(19199,36,'cp 137','c2c3 e7e6 e1g1 g8e7 f1e1 c6e5 c4e2 e7g6 d2d4 e5f3 e2f3 c5d4 c3d4 f8b4 b1c3 e8g8 d1b3 b4a5 c1e3 g6h4 f3d1 g8h8 d4d5 b7b6 g2g3 h4g6 f2f4 c8b7 d1f3 e6d5 e4d5 g6e7 a1c1 a8c8 e3f2 e7f5 e1d1 a5c3 c1c3 c8c3 b3c3');
INSERT INTO evaluations VALUES(20524,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(20524,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(20524,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(20524,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(20524,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(20524,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(20524,36,'cp -23','f1b5 g8f6 b5c6 b7c6 e1g1 f8e7 d2d3 d8c7 f3e5 e8g8 d1f3 c8a6 b2b3 c5c4 d3c4 d5c4 f1d1 c4b3 a2b3 a6b7 c1b2 a8d8 c3e4 c7b6 e4f2 c6c5 f3e2 f8e8 h2h3 a7a6 d1d8 e8d8 a1d1 d8d1 e2d1 h7h6 c2c4 f6e4');
INSERT INTO evaluations VALUES(21218,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(21218,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(21218,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(21218,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(21218,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(21218,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(21218,36,'cp 109','f2f4 e5d7 e2e4 c7c6 d5c6 b7c6 f1c4 e7e6 g1f3 f8c5 d1c2 g8f6 c1d2 a8b8 h2h3 d8c7 g2g4 e8g8 e1c1 c5b4 e4e5 f6d5 f3g5 g7g6 h3h4 b4c3 d2c3 d5e3 c2h2 e3d1 c1d1 d7b6 c4b3 b6d5 h4h5 d5c3 b2c3 f7f6 e5f6 f8f6 h5g6 f6g6 h2h7 c7h7 h1h7');
INSERT INTO evaluations VALUES(23168,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(23168,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(23168,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(23168,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(23168,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(23168,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(23168,36,'cp 153','c2c3 a7a6 b5a4 b7b5 a4c2 d7d6 d2d4 e5d4 c3d4 f8e7 b1c3 g8f6 e1g1 c8b7 f1e1 e8g8 d1e2 b5b4 c3a4 b7c6 e4e5 f6d5 e2d3 g7g6 c1h6 f8e8 a1c1 d8d7 b2b3 e7f8 d3d2 f8h6 d2h6 f7f5 c2d3 d7g7 h6d2 c6a4 b3a4 a8d8 h2h3 d5c3 c1c3 b4c3');
INSERT INTO evaluations VALUES(24937,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24937,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(24937,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24937,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24937,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24937,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(24937,36,'cp 0','e7e6 e1g1 g4h5 b2b3 f8e7 c1b2 a7a5 e2e3 f6e4 a2a3 e8g8 d1d3 h5g6 d3e2 g6h5');
INSERT INTO evaluations VALUES(25222,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25222,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25222,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25222,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25222,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25222,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(25222,36,'cp 0','e1g1 g4h5 b2b3 f8e7 c1b2 a7a5 e2e3 f6e4 c3e2 a5a4 c2c4 a4a3 b2c1 g7g5 c1d2 d8d7 d2e1 h5g6 c4d5 e6d5 f3e5 c6e5 d4e5 e4c5 f2f4 c5d3 e2d4 g5f4 g3f4 d3e1 f1e1 c7c6 d1e2');
INSERT INTO evaluations VALUES(29052,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(29052,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(29052,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(29052,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(29052,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(29052,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(29052,36,'cp 25','h2h3 g4h5 c2c4 e7e6 e1g1 b8d7 c4d5 e6d5 b1c3 f8d6 f3d2 h5g6 e2e4 d5e4 d2e4 f6e4 c3e4 g6e4 g2e4 e8g8 e4f3 d7f6 g1g2 d8b6 d1c2 f6d5 f1e1 f8d8 f3d5 c6d5 c1e3 d8e8 c2f5 a8d8 a1c1 b6b2 f5d5 b2b6');
INSERT INTO evaluations VALUES(29813,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(29813,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(29813,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(29813,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(29813,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(29813,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(29813,36,'cp 155','f3e5 g4h5 d1b3 d5b6 b3h3 h5g6 e5g6 f7g6 e2e3 d8d7 h3f3 b8c6 f1b5 e7e6 c1d2 f8e7 b5c6 b7c6 e1g1 e7f6 d2b4 d7d5 b1d2 d5f3 d2f3 a7a5 b4c5 e8d7 f1c1 g6g5 e3e4 a5a4 e4e5 g5g4 e5f6 g4f3 f6f7 f3g2 g1g2 a8b8 f7f8n h8f8 c5f8 b8f8');
INSERT INTO evaluations VALUES(29900,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(29900,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(29900,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(29900,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(29900,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(29900,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(29900,36,'cp 66','d1b3 d6d5 c4d5 f6d5 f3e5 g4e6 e4d5 e6d5 b3b5 b8c6 e1g1 a7a6 b5e2 d8e7 f1e1 c6e5 e2e5 e7e5 e1e5 d5e6 d2d4 f8d6 e5e1 e8d7 c1e3 h8e8 b1d2 a6a5 g2g3 f7f6 a2a4 b7b6 f2f3 e6f5 g1f2 f5d3 e3f4 d6f4 g3f4 d3c2 d2e4 e8e7 e4g3 e7e1 f2e1 a8e8 e1f2 c2d3');
INSERT INTO evaluations VALUES(33677,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(33677,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(33677,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(33677,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(33677,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(33677,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(33677,36,'cp 34','g1f3 g7g6 d2d4 f8g7 c1e3 e8g8 e1c1 f6d5 c3d5 e6d5 e2c4 c7c6 c4d5 c6d5 f3e1 b8c6 c2c3 c8d7 h2h3 a8c8 f1e2 f8d8 e1c2 c6a5 e3f4 a5c4 g2g4 d8e8 c2e3 c4e3 f4e3 a7a5 c1d2 b7b5 a2a3 b5b4 a3b4 a5b4 c3b4 e7e5 d4e5 g7e5');
INSERT INTO evaluations VALUES(35021,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(35021,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(35021,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(35021,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(35021,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(35021,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(35021,36,'cp 55','f8c5 c2c3 d7d6 d3d4 c5b6 a2a4 a7a6 a4a5 b6a7 c1e3 g8e7 e1g1 e8g8 f1e1 e7g6 c4f1 c8g4 b1d2 g6h4 f1e2 b8c6 f3h4 g4e2 e1e2 f6h4 d4d5 c6e7 e3a7 a8a7 c3c4 a7a8 c4c5 f7f5 a1c1 f5e4 g2g3 h4h5 d2e4 e7f5 c5d6 c7d6 e2e1');
INSERT INTO evaluations VALUES(35365,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(35365,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(35365,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(35365,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(35365,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(35365,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(35365,36,'cp 1','c2c4 h7h6 g5f4 a7a6 b1c3 f8e8 f1e1 b8c6 a1c1 e7f8 h2h3 g7g5 f4e3 f8g7 g2g4 c8d7 f3d2 d7e6 d3d4 d6d5 b2b3 a8d8 d2f3 c6e7 f3e5 f6d7 e5d7 d8d7 c3a4 d7d8 a4c5 e6c8 e2f1 d5c4 b3c4 g7d4 e3d4 d8d4 c5b3');
INSERT INTO evaluations VALUES(38397,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(38397,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38397,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(38397,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(38397,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(38397,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(38397,36,'cp 56','e2e4 g8f6 c1e3 e8g8 d1d2 a7a6 e3h6 b7b5 f1d3 b8c6 d4d5 g7h6 d2h6 f6g4 h6d2 c6e5 f3e5 g4e5 d3e2 e7e6 a2a3 e6d5 d2d5 c7c6 d5d2 a6a5 e1g1 c8g4 f2f3 g4e6 f1d1 e6c4 d2d6 d8b6 g1h1 c4e2 c3e2 e5c4');
INSERT INTO evaluations VALUES(43243,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43243,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(43243,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(43243,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(43243,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(43243,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43243,36,'cp 11','c2c4 d7d6 g1f3 e8g8 b1c3 e7e5 f1e2 a7a5 d4e5 d6e5 f3e5 d8d1 e2d1 f8e8 e5d3 b8a6 d3f4 c7c6 e1g1 f6d7 f2f3 d7c5 a2a3 c8f5 e3e4 g7d4 g1h1 f5e6 f4e6 c5e6 c3e2 d4g7 f3f4 a6c5 e4e5 f7f6 f4f5 g6f5 e5f6 g7f6 f1f5 f6g7 d1c2 a8d8 c1e3');
INSERT INTO evaluations VALUES(44058,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(44058,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(44058,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(44058,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(44058,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(44058,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(44058,36,'cp 139','e4e5 f6g8 d2d4 d7d5 a2a3 c6c5 d4c5 f8c5 b2b4 c5e7 f1d3 a7a5 b4b5 b8d7 c3a4 f7f6 c1b2 f6e5 f3e5 g8f6 e1g1 e8g8 d1e2 d7e5 b2e5 e7d6 c2c4 b7b6 f1e1 d8c7 a1c1 d6e5 e2e5 c7e5 e1e5 d5c4 d3c4 a8b8 e5e1 c8d7 c4e6 d7e6 e1e6');
INSERT INTO evaluations VALUES(44651,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(44651,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(44651,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(44651,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(44651,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(44651,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(44651,36,'cp 138','h2h3 g4h6 c2c3 d7d6 d2d4 c5d4 c3d4 d6e5 f3e5 h6f5 f1b5 b8d7 b1c3 a7a6 b5d7 c8d7 d1b3 e7e6 b3b7 d8c8 b7d7 c8d7 e5d7 e8d7 d4d5 a8c8 d5e6 d7e6 c1d2 f7f6 e1c1 e6f7 c1b1 h7h5 c3e4 f5d6 h1e1 d6e4 e1e4 f8c5 d2e3 c5e3 e4e3 h8e8 d1d7 f7g6');
INSERT INTO evaluations VALUES(44650,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(44650,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(44650,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(44650,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(44650,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(44650,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(44650,36,'cp 133','g4h6 d2d4 c5d4 c2c3 d7d6 c3d4 d6e5 f3e5 h6f5 f1b5 c8d7 b1c3 d7b5 d1b3 e7e6 b3b5 b8d7 e1g1 f5d4 b5a4 f8d6 e5f7 e8f7 a4d4 d7f6 f1d1 d6e7 d4f4 d8c8 c1e3 h8d8 e3d4 c8c6 d4f6 g7f6 d1e1 f6f5 a1c1 a8c8 c3e4 c6d5 e4g5 e7g5 f4g5 c8c1 e1c1 d8d7');
INSERT INTO evaluations VALUES(44640,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(44640,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(44640,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(44640,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(44640,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(44640,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(44640,36,'cp 133','d2d4 c5d4 c2c3 d7d6 c3d4 d6e5 f3e5 h6f5 f1b5 b8d7 b1c3 a7a6 b5d7 c8d7 g2g4 f5d6 e1g1 d7c8 d1b3 h7h5 c3e4 e7e6 b3f3 f7f6 e5g6 h5g4 h3g4 h8g8 e4c3 d6b5 f3h3 b5c3 b2c3 d8d5 g6f4 d5e4 h3h5 g7g6 h5h7 g8g7');
INSERT INTO evaluations VALUES(49169,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49169,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49169,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49169,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49169,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49169,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49169,36,'cp 0','c7c5 c2c3 b8c6 g1f3 f8e7 a2a3 e8g8 e2f4 d8a5 c1e3 c6d4 f3d4 c5d4 e3d4 a5c7 f1d3 d7e5 d3h7 g8h7 d1h5 h7g8 d4e5 e7d6 e5g7 g8g7 h5g4 g7h8 g4h5');
INSERT INTO evaluations VALUES(44010,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44010,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44010,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44010,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44010,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44010,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44010,36,'cp 0','c2c3 b8c6 g1f3 f8e7 a2a3 e8g8 e2f4 d8a5 c1e3 c6d4 f3d4 c5d4 e3d4 a5c7 f1d3 d7e5 d3h7 g8h7 d1h5 h7g8 d4e5 e7d6 e5g7 g8g7 h5g4 g7h8 g4h4 h8g8 h4g3 g8h8');
INSERT INTO evaluations VALUES(44002,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44002,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44002,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44002,36,'cp 0','b8c6 g1f3 f8e7 a2a3 e8g8 e2f4 d8a5 c1e3 c6d4 f3d4 c5d4 e3d4 a5c7 f1d3 d7e5 d3h7 g8h7 d1h5 h7g8 d4e5 e7d6 e5g7 g8g7 h5g4 g7h8 g4h4 h8g8 h4g3 g8h8');
INSERT INTO evaluations VALUES(12125,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12125,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12125,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12125,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12125,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12125,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12125,36,'cp 0','g1f3 f8e7 h2h4 c5d4 c3d4 f7f6 e2f4 d7e5 d4e5 f6e5 f4h3 e8g8 h4h5 e5e4 f3d4 e7c5 d4c6 b7c6 h5h6 d8b6 d1e2 g7g6 a2a3 a8b8 b2b4 c5d4 a1b1 e6e5 c1e3 d4c3 e3d2 c3d4 d2e3');
INSERT INTO evaluations VALUES(12124,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12124,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12124,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12124,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12124,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12124,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12124,36,'cp 0','f8e7 h2h4 c5d4 c3d4 f7f6 e2f4 d7e5 d4e5 f6e5 f4h3 e8g8 h4h5 e5e4 f3d4 e6e5 d4c6 b7c6 h5h6 d8b6 h6g7 f8f7 a2a3 e7c5 d1h5 a8b8 f1e2 c5f2 h3f2 b6f2 e1d1 f2d4 d1e1 d4f2');
INSERT INTO evaluations VALUES(50729,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(50729,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(50729,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(50729,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(50729,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(50729,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(50729,36,'cp -32','b7b5 c4e2 c8b7 d2d3 g8f6 f1e1 b8c6 e2f1 f8e7 c3e2 e8g8 c2c3 a6a5 e2g3 b5b4 a2a3 b4c3 b2c3 b7a6 c1f4 d8d7 d1d2 f8d8 a1b1 a5a4 h2h3 a8b8 b1b8 d8b8 d3d4 a6f1 e1f1 b8b7 d4d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(51268,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(51268,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(51268,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(51268,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(51268,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(51268,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(51268,36,'cp -36','c4e2 c8b7 f1e1 f8e7 d2d3 b8c6 e2f1 g8f6 c3e2 e8g8 c2c3 a6a5 e2g3 b5b4 a2a3 b7a6 c1f4 d8b6 h2h3 h7h6 d1d2 f8c8 a3b4 a5b4 f4h6 g7h6 d2h6 f6h7 g3h5 e7f8 h6f4 b6c7 d3d4 a6f1 a1a8 c8a8 e1f1 b4c3 b2c3 c5d4 c3d4');
INSERT INTO evaluations VALUES(50783,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(50783,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(50783,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(50783,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(50783,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(50783,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(50783,36,'cp 171','e4d5 g8f6 d4c5 f8c5 f1c4 e6d5 c3d5 f6d5 d1d5 d8e7 d5e5 b8c6 e5e7 e8e7 e1g1 h8d8 c2c3 e7f8 c1f4 c8f5 f1d1 a8c8 c4e2 f7f6 h2h4 f5c2 d1d8 c8d8 g1f1 f8f7 a1c1 c2f5 a2a3 c5b6 f4e3 b6e3 f2e3 f5d3 f1f2 d3e2');
INSERT INTO evaluations VALUES(50773,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(50773,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(50773,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(50773,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(50773,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(50773,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(50773,36,'cp 156','g8f6 d4c5 f8c5 f1c4 e6d5 c3d5 e8g8 e1g1 f6d5 d1d5 d8d5 c4d5 b8c6 f1e1 f8d8 c2c4 c5b4 e1d1 c8e6 c1f4 b4e7 g2g3 a8c8 d1d2 e6f5 a1d1 e7b4 d2e2 h7h6 a2a3 b4f8 f3e5 f5e6 g1g2 e6d5 c4d5 c6e5 f4e5 f7f6 e5c3 g8f7');
INSERT INTO evaluations VALUES(51114,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51114,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(51114,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(51114,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(51114,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(51114,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(51114,36,'cp 55','e2e4 e7e6 d2d4 d7d5 b1c3 f8b4 e4e5 c8d7 c1d2 g8e7 a2a3 b4c3 d2c3 b7b6 c3d2 d7b5 f1b5 a6b5 e1g1 b8c6 d1e2 d8d7 d2b4 e7f5 c2c3 c6a5 g2g4 f5e7 b4e7 e8e7 g1h1 h7h6 f1g1 e7d8 g1g3 d8c8 h1g2 a5b3 a1d1 g7g6 h2h4 b3a5 h4h5 g6h5');
INSERT INTO evaluations VALUES(51265,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(51265,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(51265,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(51265,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(51265,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(51265,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(51265,36,'cp 40','a2a4 b5a4 a1a4 f8e7 e1g1 b8d7 d2d4 g8f6 b1c3 d8c7 f1e1 e8g8 h2h3 f8d8 a4a1 h7h6 c1e3 d7f8 f3h4 e5d4 d1d4 d8e8 h4f3 c8e6 b3e6 f8e6 d4b6 a8b8 b6c7 e6c7 a1a5 b8b2 f3d4 c7e6 d4c6');
INSERT INTO evaluations VALUES(53655,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(53655,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(53655,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(53655,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(53655,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(53655,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(53655,36,'cp 59','f1c4 b7b5 c4d3 a7a6 g1f3 c6c5 a2a4 b5b4 c3e4 b8d7 d4c5 c8b7 e4d6 f8d6 c5d6 g8f6 e1g1 d7c5 d3c2 e8g8 d1d4 a8c8 d4b4 b7f3 g2f3 d8d6 f1d1 d6c6 e3e4 e6e5 c1e3 f6h5 e3c5 h5f4 d1d6 c6c5 b4c5 c8c5 c2d3 f4d3 d6d3 f7f5 a1e1 c5c2 b2b3 c2b2 g1g2 h7h5 e4f5 f8f5 e1e4 g8f7 b3b4 b2c2 a4a5');
INSERT INTO evaluations VALUES(53876,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(53876,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53876,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(53876,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(53876,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(53876,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(53876,36,'cp 29','f8e7 b1c3 g8f6 d2d3 e8g8 c1g5 f6e8 g5e7 d8e7 c3d5 e7d8 e1g1 e8f6 f3d2 f6d5 c4d5 b8a6 d5b3 c8e6 c2c3 d8g5 b3e6 f7e6 a2a3 a8d8 f1e1 g5f6 d2f3 a6b8 e1e3 b8c6 d1e2 c6e7 b2b4 e7g6 g2g3 b7b6 b4c5 b6c5 a1b1 h7h6 f3d2 d8b8 b1b8 f8b8');
INSERT INTO evaluations VALUES(55760,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(55760,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(55760,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(55760,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(55760,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(55760,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(55760,36,'cp 40','b8d7 d2d4 g8f6 e1g1 a7a6 b5d7 c8d7 f3e5 c7c5 f1e1 c5d4 d1d4 f8e7 d4h4 e8g8 c1g5 d7b5 a2a4 b5c6 e5c6 b7c6 c3e4 h7h6 g5f6 e7f6 e4f6 d8f6 h4f6 g7f6 e1e4 f8d8 e4c4 d8d6 g1f1 a8d8 g2g3 g8f8 f1e2 f8e7 b2b4');
INSERT INTO evaluations VALUES(55877,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(55877,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(55877,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(55877,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(55877,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(55877,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(55877,36,'cp -46','d5e4 b1d2 g8f6 c2c3 f8e7 g2g3 c7c5 f1g2 c5d4 c3d4 c8d7 g1e2 f6d5 d2e4 d5e3 f2e3 e8g8 e1g1 d7c6 e2f4 b8d7 d1e2 a8c8 a1d1 d8a5 e4c3 c6g2 e2g2 a5b6 f4d3 c8c7 g2f3 h7h6 f1f2 f7f5 c3a4 b6c6 a4c3 c6f3 f2f3');
INSERT INTO evaluations VALUES(56052,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(56052,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(56052,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(56052,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(56052,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(56052,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(56052,36,'cp -36','b1d2 g8f6 c2c3 b7b6 g2g3 c8b7 f1g2 c7c5 d1c2 d8d5 g1e2 c5d4 c3d4 f8b4 e2c3 b4c3 b2c3 e8g8 e1g1 d5h5 d2e4 f6e4 g2e4 b7e4 c2e4 b8d7 c3c4 a8c8 f1c1 d7f6 e4g2 b6b5 c4b5 c8c1 a1c1 h5b5 g2c6 b5h5 c6c5 f6d5 c5a7 d5e3 f2e3');
INSERT INTO evaluations VALUES(2939,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(2939,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(2939,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(2939,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(2939,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(2939,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(2939,36,'cp 28','c4d5 f6d5 h4g3 d5c3 b2c3 b7b6 e1g1 c8b7 d1e2 c6c5 e3e4 c5d4 c3d4 d7f6 a1d1 f8e8 h2h3 a8c8 g3e5 b7c6 d1b1 f6d7 e5f4 d7f6 f1e1 e7a3 e2e3 c6b7 f4g3 c8c3 g3h4 d8c8 e3e2 f6h5');
INSERT INTO evaluations VALUES(6258,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(6258,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(6258,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(6258,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(6258,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(6258,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(6258,36,'cp 46','b4c3 b2c3 d8f6 a2a4 d7d6 d2d4 c8g4 d4d5 c6a5 c4e2 g8e7 f3e1 g4e2 d1e2 e8g8 c3c4 b7b6 c1d2 c7c6 d2a5 b6a5 a1b1 c6d5 c4d5 e7c8 b1b5 c8b6 b5a5 f8c8 a5a6 c8c7 a4a5 b6d7 a6c6 c7c6 d5c6 d7c5');
INSERT INTO evaluations VALUES(6270,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(6270,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(6270,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(6270,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(6270,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(6270,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(6270,36,'cp 54','b2c3 d7d6 d2d4 d8f6 d4d5 c6b8 a2a4 a7a5 c4b5 c8d7 c3c4 d7b5 c4b5 g8e7 a1a3 b8d7 f3d2 e8g8 d2c4 f6g6 a3g3 g6e4 c1h6 e4c4 g3g4 c4d5 g4g7 g8h8 d1g4 f8g8 g4d7 g8g7 h6g7 h8g7 d7e7 d5c4 b5b6 c7b6 e7d6 a8e8 d6b6 c4a4 b6b7 a4c4 b7a7 a5a4 f1a1 c4c2 a7a4 c2a4 a1a4 e8b8 g2g3 b8b1 g1g2 g7g6 h2h4 b1b3 a4a8 f7f6');
INSERT INTO evaluations VALUES(7259,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7259,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(7259,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(7259,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(7259,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(7259,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(7259,36,'cp 11','d2d3 d7d6 c1e3 c5e3 f2e3 c8e6 c3d5 e8g8 d1e1 e6d5 e4d5 c6e7 e3e4 e7g6 g1h1 d8e7 c4b3 a6a5 a3a4 c7c6 d5c6 b7c6 f3h4 g6h4 e1h4 f6d5 h4e1 d5f4 e1c3 e7c7 f1f2 a8b8 a1f1 b8b4 h2h3 f4e6 h1h2 b4d4 g2g3 d6d5 e4d5');
INSERT INTO evaluations VALUES(12361,30,'cp -298',NULL);
INSERT INTO evaluations VALUES(12361,31,'cp -267',NULL);
INSERT INTO evaluations VALUES(12361,32,'cp -284',NULL);
INSERT INTO evaluations VALUES(12361,33,'cp -289',NULL);
INSERT INTO evaluations VALUES(12361,34,'cp -295',NULL);
INSERT INTO evaluations VALUES(12361,35,'cp -279',NULL);
INSERT INTO evaluations VALUES(12361,36,'cp -288','d7d5 e4d5 e6d5 f3e2 f8e7 c4b5 c8f5 d2d3 e8g8 g1f3 d8b6 b5c6 b7c6 e1g1 f8e8 e2d1 e7d6 b1a3 a8d8 a3c2 c5c4 d3d4 h7h6 f1e1 f5e4 c2e3 c6c5 e3f1 c5d4 f3d4 e4d3 f1e3 d6c5 d4b3 c5f8 b3d4 a7a5 h2h3 f6e4 d1f3 e4c5 d4e2 d3e2 e1e2');
INSERT INTO evaluations VALUES(12946,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12946,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(12946,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(12946,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(12946,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12946,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12946,36,'cp 5','d4c5 d8a5 b1c3 a7a6 b5c6 e7c6 e1g1 f8c5 c3e2 e8g8 c2c3 a5b6 a1b1 c8d7 c1f4 a6a5 d1d2 f7f5 e5f6 g7f6 e2g3 f8f7 f4e3 c6e7 c3c4 d7c6 e3c5 b6c5 f3d4 c6d7 c4d5 c5d5 b1d1 e6e5 d4f3 d5d2 f3d2 d7e6 d2e4 e6a2');
INSERT INTO evaluations VALUES(14322,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(14322,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(14322,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(14322,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(14322,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(14322,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(14322,36,'cp -34','e7e5 e4d5 f6d5 g2g3 f8e7 c1d2 e8g8 f1g2 c8e6 e1g1 f7f6 f1e1 d8d7 c3d5 e6d5 c2c3 a8d8 b2b4 a7a6 a2a4 d5f3 d1f3 d7d3 e1d1 d3f3 g2f3 f6f5 f3e2 d8d6 b4b5 c6a5 b5a6 b7a6 d2e3 f8d8 d1d6 d8d6 a1b1 d6c6 b1d1 a5c4 d1d7 c4e3');
INSERT INTO evaluations VALUES(17898,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(17898,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(17898,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17898,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(17898,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(17898,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(17898,36,'cp 44','d7d6 d2d4 e5d4 f3d4 c6d4 d1d4 g8f6 e1g1 f8e7 b1c3 e8g8 a3a4 c8e6 a4a5 e6c4 d4c4 f8e8 h2h3 c7c6 c1e3 f6d7 f1d1 e7f8 e3d4 h7h6 f2f4 d7f6 c4d3 d6d5 e4e5 f6d7 c3a4 c6c5 d4e3 d5d4 e3d4 c5d4 d3d4 d8a5 d4d7 a8d8');
INSERT INTO evaluations VALUES(19838,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(19838,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(19838,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19838,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(19838,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19838,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(19838,36,'cp 12','g8f6 e2e3 c8f5 g1f3 e7e6 d1b3 f8b4 f1b5 b4c3 b3c3 e8g8 b5c6 a8c8 f3e5 f6g4 e5g4 f5g4 c3b4 c8c6 b4b7 d8c8 b7c8 f8c8 e1g1 a7a5 f2f3 g4f5 h2h4 h7h5 b2b3 f7f6 f1f2 g8f7 f2e2 f5g6 e2b2 g6d3 a1e1 c6c2 b2c2 c8c2 e1a1 f7g6 f4d6');
INSERT INTO evaluations VALUES(20015,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(20015,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20015,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(20015,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(20015,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(20015,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(20015,36,'cp 27','d3d4 d6d5 d4c5 d5e4 d1d8 e8d8 c1e3 f7f5 b1a3 e7e5 e1c1 d8e7 g2g3 c8e6 b2b4 a8d8 d1d8 c6d8 f1c4 g7g6 g1h3 f8h6 h3g5 g8f6 c1c2 h6g5 e3g5 d8f7 g5f6 e7f6 c4e6 f6e6 a3b5 h8d8 a2a4 a7a6 b5c7 e6e7 b4b5 a6b5 a4b5 f7g5 c3c4 d8d4 c2c3 g5e6 c7e6 e7e6');
INSERT INTO evaluations VALUES(20817,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(20817,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(20817,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(20817,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(20817,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(20817,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(20817,36,'cp 78','g7g6 c2c4 f7f5 b1c3 g8f6 h2h4 f8g7 h4h5 g6h5 c1g5 e8g8 g5f6 f8f6 f3g5 h5h4 d1h5 h7h6 g5e6 c8e6 d5e6 d8e8 h5f3 f5e4 f3e4 f6e6 h1h4 a8d8 e1c1 e8f7 e4b7 c7c6 b7a7 d6d5 c4d5 c6d5 h4b4 e5e4 d3c2 e6d6');
INSERT INTO evaluations VALUES(22413,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(22413,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(22413,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(22413,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(22413,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(22413,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(22413,36,'cp 24','g1f3 g8f6 b1c3 c7c5 d4d5 d8a5 a2a4 f6e4 d1d3 a5b4 f4c1 f7f5 a4a5 b7b5 a5b6 d7b6 f3d2 a7a5 c3e4 f5e4 d3e4 b4e4 d2e4 b6d5 e4d2 d5b6 b2b3 a5a4 e2e4 a4b3 a1a8 b6a8 c2b3 a8c7 d2c4 g7g6 c1b2 h8g8 f2f4 f8g7 b2g7 g8g7');
INSERT INTO evaluations VALUES(22412,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(22412,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(22412,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(22412,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(22412,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(22412,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(22412,36,'cp 36','g8f6 c2c4 e7e5 d4e5 d6e5 f4e5 d7e5 d1d8 e8d8 f3e5 f8b4 b1d2 f6e4 e5f3 b4a5 a2a3 c7c5 e2e3 d8e7 e1e2 e4d2 f3d2 c8g4 f2f3 g4e6 a1c1 a8d8 d2b3 a5b6 e2f2 a7a5 a3a4 b6c7 b3c5 d8d2 f1e2 e6f5 b2b3 b7b6 c5e4 f5e4 f3e4 h8d8 c1d1 d2b2 d1d8 c7d8');
INSERT INTO evaluations VALUES(22594,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(22594,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(22594,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(22594,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(22594,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(22594,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(22594,36,'cp 62','g8f6 a2a3 g7g6 g1f3 f8g7 f1e2 a7a5 b1c3 e8g8 e1g1 d7d6 a1b1 e6e5 d4d5 c6e7 d1c2 f6d7 b2b4 h7h6 c1b2 b7b6 f3d2 f7f5 c3b5 g6g5 e3e4 f5f4 e2g4 g8h8 b2c3 a5b4 c3b4 e7g8 g4e6 d7c5 e6c8 d8c8');
INSERT INTO evaluations VALUES(16272,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(16272,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(16272,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(16272,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(16272,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(16272,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(16272,36,'cp 67','a2a3 d7d5 g1f3 d5c4 b1d2 c6a5 d1a4 c7c6 d2c4 a5c4 a4c4 a7a5 c4c2 c8d7 c1d2 a8c8 f3e5 c6c5 e5d7 d8d7 a1c1 f8d6 b2b4 a5b4 a3b4 b7b6 f1e2 e8g8 d4c5 b6c5 b4b5 f6d5 c2b2 f8d8 e1g1 d5b6 b2c2');
INSERT INTO evaluations VALUES(25024,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(25024,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(25024,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(25024,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(25024,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(25024,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(25024,36,'cp 57','d2e4 d5e4 h2h3 g4h5 g2g4 h5g6 d4d5 e4f3 d5c6 d8d1 a1d1 b7c6 f4c7 e7e6 e1d2 a7a5 d2c1 h7h5 h1g1 f8e7 g4g5 h5h4 c7e5 h8h5 e5g7 h5g5 f1d3 g6d3 c2d3 f7f6 g7h6 g5g1 d1g1 e8d7 g1g7 e6e5 e3e4 d7e6 h6e3 a5a4 c1c2');
INSERT INTO evaluations VALUES(26959,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(26959,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26959,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(26959,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(26959,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(26959,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(26959,36,'cp 46','g4h5 d2d3 g8f6 a2a4 a7a6 a4a5 d8d7 c1g5 f8e7 g5f6 h5f3 d1f3 e7f6 b1d2 c6d8 c4d5 f6g5 d2c4 d7e7 g2g3 e8g8 h3h4 g5h6 b2b4 a8b8 c4e3 h6e3 f3e3 d8e6 e1g1 c7c6 d5b3 e6c7 g1g2 c7b5 d3d4 e5d4');
INSERT INTO evaluations VALUES(26949,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(26949,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(26949,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(26949,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(26949,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(26949,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(26949,36,'cp 68','a2a4 g8f6 d2d3 a7a6 a4a5 d8d7 c1g5 f8e7 g5f6 g7f6 d1b3 c6d8 f3h4 h5g6 b1d2 f6f5 h4f5 g6f5 e4f5 c7c6 d3d4 d6d5 c4e2 d7f5 d4e5 f5e5 d2f3 e5f6 e1g1 e8g8 c3c4 d5d4 e2d3 d8e6 a1d1 e6f4 b3b7');
INSERT INTO evaluations VALUES(29097,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(29097,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(29097,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(29097,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(29097,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(29097,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(29097,36,'cp 57','e7e6 h2h3 g4f3 e2f3 f8e7 c3e2 e8g8 e1g1 d6d5 e4e5 f6d7 c2c3 c6c5 c1e3 b8c6 f3g4 a7a5 a2a4 a8c8 d1d2 c5d4 c3d4 e7b4 d2d1 c8c7 g2g3 d7b6 b2b3 d8e7 h3h4 f8c8 e2f4 b6d7 d1e2 c6a7 f1c1 c7c3 a1b1 c3c1 b1c1 c8c1 e3c1');
INSERT INTO evaluations VALUES(29347,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(29347,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(29347,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(29347,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(29347,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(29347,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(29347,36,'cp 63','h2h3 g4f3 e2f3 f8e7 e1g1 e8g8 c1e3 a7a5 a2a4 f6d7 d1e2 b8a6 g2g3 e7g5 c3d1 g5e3 d1e3 d6d5 f1d1 d5e4 f3e4 a6c7 c2c4 d8e7 e4g2 f8d8 e3c2 f7f5 g3g4 g7g6 g4f5');
INSERT INTO evaluations VALUES(29842,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(29842,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(29842,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(29842,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(29842,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(29842,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(29842,36,'cp 57','a2a4 g4h5 h2h3 b8d7 g2g4 h5g6 b1c3 g6h7 f3e1 c7c6 e1g2 g7g5 b2b4 f8e7 a4a5 a7a6 d1f3 d7f8 c3a4 f8e6 c1b2 f6d7 c4e6 f7e6 g2e3 h7g6 e3c4 h6h5 g1g2 d8c7 a4b6 d7b6 c4b6 a8d8 b6c4 g6f7 f1h1 c7d7 a1g1 h8g8');
INSERT INTO evaluations VALUES(30820,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30820,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(30820,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(30820,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(30820,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30820,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30820,36,'cp 0','e7e6 c2c3 d7d5 e4e5 c7c5 e1g1 b8c6 d3b5 d5d4 c3d4 c5d4 d2d3 a7a6 b5c6 b7c6 b1d2 f8c5 a2a3 a6a5 d2e4 g8e7 c1g5 h7h6 g5e7 c5e7 a1c1 a8c8 f3d2 e8g8 d1g4 b6b5 g4g3 e7h4 g3h3 h4e7');
INSERT INTO evaluations VALUES(30693,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(30693,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(30693,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(30693,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(30693,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30693,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30693,36,'cp 0','c2c3 d7d5 e4e5 c7c5 e1g1 g8e7 d3b5 b7c6 b1a3 d8d7 b5e2 d5d4 c3d4 c5d4 d2d3 d7d5 d1b3 d5d7');
INSERT INTO evaluations VALUES(31214,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(31214,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31214,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31214,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(31214,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(31214,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31214,36,'cp 23','d3e4 g8f6 e4f3 b8d7 d1e2 g7g6 g1h3 f8g7 e1g1 e8g8 h3f4 f6e8 d4d5 d7e5 d5c6 b7c6 f1e1 e7e6 e1d1 d8c7 f4d3 e5f3 e2f3 e6e5 d3c5 e8f6 c2c4 c7b6 c5b3 b6c7 f3e2 f6d7 e2c2 f7f5 c1e3 e5e4');
INSERT INTO evaluations VALUES(32133,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(32133,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(32133,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(32133,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(32133,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(32133,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(32133,36,'cp 34','c7c5 c3d5 g8f6 d5f6 d8f6 f1e2 f8e7 e1g1 h7h6 f3d2 g4e2 d1e2 b8c6 d2c4 e8g8 c4e3 f6g6 e3f5 f8e8 c1e3 e7f8 c2c4 a8b8 h2h4 g6e6 h4h5 g8h7 e3d2 c6d4 f5d4 c5d4 a2a4 f7f5 f2f4 e5f4 f1f4 f5e4');
INSERT INTO evaluations VALUES(34176,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(34176,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(34176,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(34176,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(34176,34,'cp 230',NULL);
INSERT INTO evaluations VALUES(34176,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(34176,36,'cp 200','c4b3 a7a5 a2a4 b5b4 c3e2 c6c5 c2c3 c8a6 d2d4 c5c4 b3a2 b8c6 e1g1 b4b3 a2b1 f6d6 e2g3 g8e7 d4e5 d6d1 f1d1 e7g6 d1d5 g6e7 c1e3 g7g6 h2h4 f8g7 d5c5 e8g8 h4h5 h7h6 g3f1 f8e8 f1d2 g6h5 d2c4 a6c4 c5c4 e7g6 c4d4');
INSERT INTO evaluations VALUES(34177,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(34177,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(34177,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(34177,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(34177,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(34177,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(34177,36,'cp 248','a7a5 a2a4 b5b4 c3e2 c6c5 c2c3 c8a6 b3d5 b8c6 d2d4 c5d4 c3d4 a8c8 c1g5 f6d6 d1b3 g8h6 a1c1 f8e7 d4e5 c6e5 c1c8 a6c8 g5f4 e5f3 g2f3 d6a6 h1g1 e8g8 e2g3 g8h8 g3h5 e7f6 e4e5 f6h4 b3c2 f8g8 c2e4 h4e7 f4e3 d7d6 e3h6 g7h6 g1g8 h8g8 e5e6 f7e6 d5e6 c8e6 e4e6 g8f8 e6f5 f8g8 f5d5 g8h8');
INSERT INTO evaluations VALUES(34160,30,'cp 210',NULL);
INSERT INTO evaluations VALUES(34160,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(34160,32,'cp 248',NULL);
INSERT INTO evaluations VALUES(34160,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(34160,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(34160,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(34160,36,'cp 207','a2a4 b5b4 c3e2 c6c5 c2c3 c8a6 b3d5 b8c6 d2d4 c5d4 c3d4 a8c8 c1g5 f6d6 d1b3 g8h6 a1c1 f8e7 d4e5 c6e5 c1c8 a6c8 g5f4 e5f3 b3f3 d6f6 e4e5 f6g6 d5e4 g6g4 f3g4 h6g4 h2h3 g7g5 f4g3 g4h6 e1d2 c8a6 h1c1 h8f8 e4h7 f7f5 e5f6 f8f6 e2d4 e7d6 h7d3 d6g3 f2g3 a6d3 c1e1 e8d8 d2d3');
INSERT INTO evaluations VALUES(47771,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47771,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(47771,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(47771,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(47771,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(47771,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(47771,36,'cp 0','d6e5 d1d8 e8d8 g1f3 b8c6 b2b3 a7a5 g2g3 f8b4 c1b2 d8e7 a2a3 b4c3 b2c3 f6e4 c3e5 c6e5 f3e5 e4c5 a1b1 c8f5 b1b2 h8e8 e5d3 f5e4 h1g1 e4d3 e2d3 e7f8 e1d1 a8d8 d1d2 c5e4 d2c1 e4c5');
INSERT INTO evaluations VALUES(48912,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48912,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(48912,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48912,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48912,36,'cp 0','d1d8 e8d8 g1f3 b8c6 c1g5 c8e6 c3d5 h7h6 g5f6 g7f6 e1c1 d8c8 d5f6 e6c4 a2a3 c4b3 d1d2 c6a5 c1b1 c7c6 e2e3 c8c7 d2d7 c7c8');
INSERT INTO evaluations VALUES(35395,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35395,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35395,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35395,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35395,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35395,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35395,36,'cp 0','e8d8 g1f3 b8c6 c1g5 c8e6 c3d5 h7h6 g5f6 g7f6 e1c1 d8c8 d5f6 e6c4 a2a3 c4b3 d1d2 c6a5 e2e3 c7c6 c1b1 c8c7 d2d7 c7c8');
INSERT INTO evaluations VALUES(35427,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35427,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35427,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35427,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35427,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35427,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35427,36,'cp 0','g1f3 b8c6 c1g5 c8e6 c3d5 h7h6 g5f6 g7f6 e1c1 d8c8 d5f6 e6c4 a2a3 c4b3 d1d2 c6a5 e2e3 c7c6 c1b1 c8c7 d2d7 c7c8');
INSERT INTO evaluations VALUES(48007,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(48007,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(48007,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(48007,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(48007,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(48007,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(48007,36,'cp 48','d4e5 c7c6 b1d2 e4d2 d1d2 d6d5 c4d3 b8d7 c2c3 g7g6 e1g1 f8g7 d2f4 e8g8 f1e1 d8c7 h2h4 f7f6 e5f6 c7f4 c1f4 d7f6 a2a4 c8g4 f3d4 g4d7 a4a5 f8e8 a5a6 b7b6 a1d1 c6c5 d4f3 c5c4 d3c2 d7g4 e1e8 a8e8 f4e3 g4h5 d1d4 h5f3 g2f3');
INSERT INTO evaluations VALUES(51417,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(51417,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(51417,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51417,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51417,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51417,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51417,36,'cp 8','c8b7 f1d3 d7d5 c2c4 g8f6 e1g1 d5c4 d3c4 a7a6 b1c3 b6b5 c4b3 c7c5 e3e4 c5d4 f3d4 b8d7 f1e1 d7c5 b3c2 a8c8 c1g5 f8e7 e4e5 f6d5 g5e7 d5e7 d1g4 e7g6 a1d1 d8h4 g4h4 g6h4 c2e4 c5e4 c3e4 b7e4 e1e4 h4g6');
INSERT INTO evaluations VALUES(52372,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(52372,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(52372,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(52372,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52372,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(52372,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52372,36,'cp 10','d7d5 c2c4 g8f6 g1f3 g7g6 c4d5 c6d5 f1g2 f8g7 b1c3 e8g8 e1g1 f6e4 f3e5 c8f5 c1f4 e4c3 b2c3 b8c6 d1b3 c6e5 f4e5 g7e5 d4e5 e7e6 b3b7 d8a5 b7b4 a5a6 b4f4 a6e2 f1e1 e2g4 f4g4 f5g4 c3c4 f8d8 a1c1 a8b8 c4d5 e6d5');
INSERT INTO evaluations VALUES(52942,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(52942,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52942,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(52942,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(52942,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52942,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(52942,36,'cp 7','g8f6 b1c3 b8d7 g1f3 g7g6 c1e3 c6d5 e4d5 f8g7 f1e2 e8g8 a2a4 d8a5 e1g1 d7b6 b2b4 a5b4 a4a5 b6c4 e3d4 b7b5 a1b1 b4a5 c3b5 a7a6 e2c4 a6b5 c4b5 e7e5 d5e6 c8e6 b5e2 f8b8 b1b8 a8b8 f1e1 e6d5 c2c3 b8b2 e2d3 a5a8 h2h3 b2a2');
INSERT INTO evaluations VALUES(57742,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(57742,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(57742,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(57742,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(57742,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(57742,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(57742,36,'cp 213','b1c3 e7e6 h2h3 f8e7 g1f3 d7d6 f1c4 b8d7 a2a4 a7a6 a4a5 c7c5 d4d5 e6e5 e1g1 g8f6 c1d2 f6h5 f1e1 h5f4 c3a4 g7g5 f3h2 d7f6 a4b6 a8b8 c4d3 h8g8 f2f3 e8f8 h2g4 c8g4 f3g4 f6d7 b6c4 f8g7 b2b4 c5b4 d2b4 d8c7 g1h2');
INSERT INTO evaluations VALUES(3663,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3663,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3663,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(3663,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3663,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3663,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(3663,36,'cp 0','a2a4 a7a5 b4b5 c6e7 b1d2 e7g6 c4b3 c8g4 h2h3 g4h5 g1h2 a8c8 g2g4 f6g4 h3g4 h5g4 f1g1 h7h5 d1f1 g6f4 g1g3 c7c6 d2c4 g4f3 g3f3 d8h4 f3h3 f4h3 f1h3 h4f2 h3g2 f2h4 g2h3');
INSERT INTO evaluations VALUES(3665,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3665,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3665,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(3665,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3665,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3665,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3665,36,'cp 0','a7a5 b4b5 c6e7 b1d2 e7g6 c4b3 c8g4 h2h3 g4h5 g1h2 a8c8 g2g4 f6g4 h3g4 h5g4 f1g1 h7h5 d1f1 g6f4 g1g3 c7c6 d2c4 g4f3 g3f3 d8h4 f3h3 f4h3 f1h3 h4f2 h3g2 f2h4 g2h3');
INSERT INTO evaluations VALUES(8296,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(8296,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(8296,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(8296,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8296,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(8296,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(8296,36,'cp 14','d3d4 e8g8 d4d5 c6a5 b1d2 b7b6 c3c4 e7e6 f1e1 f8e8 a1b1 e6d5 c4d5 c8d7 e2f1 c5c4 b2b4 c4b3 a2b3 a5b7 b3b4 a7a5 d1c2 a5b4 b1b4 a8a1 c1b2 a1e1 f3e1 b6b5 e1d3 d8c8 c2b1 b7c5 d3c5 d6c5 b4b5 d7b5 f1b5 f6e4 d2e4 e8e4 b1e4 g7b2 d5d6 c8e6 e4e6 f7e6');
INSERT INTO evaluations VALUES(10660,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(10660,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(10660,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(10660,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(10660,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(10660,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(10660,36,'cp 117','e4d5 e5e4 d5c6 e4f3 c6d7 c8d7 d1f3 e8g8 e1g1 d7c6 f3h3 b7b5 c4b3 b4d6 c2c3 f8e8 d2d4 g7g6 c1h6 a7a5 f1e1 a5a4 b3c2 c6d5 c2d3 c7c6 a2a3 d5e6 h3e3 e6d5 e3d2 d8d7 h2h3 e8e1 a1e1 a8e8 h6g5 g8g7 e1e8 d7e8 d2e3 e8e3 f2e3 f7f5 g5f4 d6f8 g2g4 g7f7 g1f2');
INSERT INTO evaluations VALUES(11027,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(11027,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(11027,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(11027,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(11027,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(11027,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(11027,36,'cp 171','d2d4 e5d4 c3d4 c5b4 b1c3 e8g8 e1g1 c6a5 c4d3 d7d5 e4d5 c8f5 d3f5 e7f5 d1d3 f5e7 f3g5 e7g6 f2f4 h7h6 f4f5 h6g5 f5g6 f7g6 f1f8 d8f8 c1g5 f8f5 d3f5 g6f5 a1c1 c7c6 g1f2 b4c3 c1c3 c6d5 c3c5 b7b6 c5d5 a8c8 f2f3 c8c2 d5f5 c2b2 d4d5 a5b7 f5e5 b2a2 e5e8 g8h7 g5f4');
INSERT INTO evaluations VALUES(11030,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(11030,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(11030,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(11030,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(11030,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(11030,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(11030,36,'cp 147','e5d4 c3d4 c5b4 b1c3 e8g8 e1g1 c6a5 c4d3 d7d5 e4d5 c8f5 d3f5 e7f5 d1d3 f5e7 f3g5 e7g6 f2f4 h7h6 f4f5 h6g5 f5g6 f7g6 f1f8 d8f8 c1g5 f8f5 d3f5 g6f5 a1c1 c7c6 g1f2 a8e8 f2f3 b4f8 c1d1 g7g6 h2h4 c6d5 c3d5 g8f7 g5f4 e8c8 d5e3 f8g7 b2b3 a5c6 e3c4 c8h8 c4d6 f7f8 d6b5 a7a6 d4d5 c6e7 d5d6 a6b5');
INSERT INTO evaluations VALUES(11033,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(11033,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(11033,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(11033,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(11033,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(11033,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(11033,36,'cp 180','c3d4 c5b4 b1c3 c6a5 c4d3 d7d5 e1g1 e8g8 e4d5 c8f5 d3f5 e7f5 d1d3 f5e7 f3g5 e7g6 f2f4 h7h6 f4f5 h6g5 f5g6 f7g6 f1f8 d8f8 c1g5 f8f5 d3f5 g6f5 a1c1 c7c6 h2h4 a8e8 g1f1 a7a6 c3e2 c6d5 c1c7 e8e6 h4h5 e6c6 c7d7 c6d6 d7e7 d6c6 e7e8 g8f7 e8e5 c6c2 e5f5 f7g8');
INSERT INTO evaluations VALUES(11032,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(11032,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(11032,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(11032,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(11032,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(11032,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(11032,36,'cp 162','c5b4 b1c3 c6a5 c4d3 e8g8 e1g1 d7d5 e4d5 e7d5 c1g5 f7f6 g5d2 c8g4 f1e1 a5c6 d3e4 b4c3 b2c3 f6f5 e4c2 d8d6 d1e2 c6a5 h2h3 g4f3 e2f3 a5c4 d2g5 h7h6 g5h4 c4d2 f3d3 d2e4 e1e4 f5e4 d3e4 a8e8 e4h7 g8f7 a1b1 e8e2 c2b3 d6e6 h4g3 e2d2');
INSERT INTO evaluations VALUES(11037,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(11037,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(11037,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(11037,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(11037,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(11037,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(11037,36,'cp 142','b1c3 c6a5 c4d3 e8g8 e1g1 d7d5 e4d5 c8f5 d3f5 e7f5 d1c2 f5e7 f3g5 e7g6 f2f4 h7h6 f4f5 h6g5 f5g6 f7g6 f1f8 d8f8 c2g6 f8d6 g6d6 c7d6 c1g5 a5c4 a1b1 a8e8 g1f2 a7a6 c3e2 b4d2 g5d2 c4d2 b1c1 d2e4 f2g1 e4f6 e2c3 e8e3 c1d1 f6g4 h2h3 g4f6');
INSERT INTO evaluations VALUES(11036,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(11036,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(11036,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(11036,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(11036,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(11036,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(11036,36,'cp 145','c6a5 c4d3 e8g8 e1g1 d7d5 e4d5 c8f5 d3f5 e7f5 d1c2 f5e7 f3g5 e7g6 f2f4 h7h6 f4f5 h6g5 f5g6 f7g6 f1f8 d8f8 c1g5 f8f5 c2f5 g6f5 a1c1 c7c6 d5c6 a5c6 c3b5 a7a6 b5c7 a8c8 c7e6 g8f7 d4d5 c6e7 c1c8 e7c8 e6d8 f7g6 g5f4 b4d6 g2g3 b7b5 g1f2 d6f4 g3f4 c8b6 d5d6 g6f6 d8c6 g7g6 c6e5 f6e6 e5g6 e6d6 g6h4 d6e6 a2a3 a6a5 f2f3 b6d5');
INSERT INTO evaluations VALUES(11620,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(11620,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(11620,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(11620,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(11620,34,'cp 185',NULL);
INSERT INTO evaluations VALUES(11620,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(11620,36,'cp 187','d2d4 c6d4 f3d4 e5d4 d1d4 c8e6 f1e1 d5f6 d4c3 d8d7 a2a4 b5b4 c3f3 c7c6 b1d2 f6d5 d2e4 f8e7 f3g3 e8c8 c1g5 f7f6 g5e3 h8e8 b3c4 c8b7 e4c5 e7c5 e3c5 a6a5 c2c3 d7c7 g3f3 d5f4 c4e6 e8e6 c3b4 e6e1 a1e1 d8d3 f3g4 h7h5 g4f5');
INSERT INTO evaluations VALUES(12276,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(12276,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(12276,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(12276,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(12276,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12276,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12276,36,'cp 30','d7d6 c3d5 f8e7 c2c3 e8g8 c1e3 f6d5 c4d5 c8e6 d5e6 f7e6 d3d4 c5d4 c3d4 e5d4 f3d4 c6d4 e3d4 e7f6 g2g3 d8a5 a2a3 a8d8 d4f6 f8f6 d1b3 b7b6 a1d1 a5c5 b3d3 d6d5 b2b4 c5d6 f1e1 f6f7 e4d5 e6d5 d3d4 d6f6 d4f6');
INSERT INTO evaluations VALUES(13439,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(13439,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(13439,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(13439,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(13439,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(13439,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(13439,36,'cp 41','c3d5 f8e7 c2c3 e8g8 c1e3 f6d5 c4d5 c8e6 d5e6 f7e6 d3d4 c5d4 c3d4 e5d4 f3d4 c6d4 e3d4 e7f6 g2g3 d8d7 d4f6 f8f6 d1d4 e6e5 d4d5 d7f7 d5f7 f6f7 a1c1 d6d5 e4d5 a8d8 f1d1 f7d7 d5d6 d7d6 d1d6 d8d6 c1c7 e5e4 c7b7 d6d1 g1g2 d1d2');
INSERT INTO evaluations VALUES(16684,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(16684,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(16684,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(16684,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(16684,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(16684,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(16684,36,'cp 66','b1c3 f8b4 f2f3 d7d5 e4e5 f6g8 f1b5 c8d7 b5c6 d7c6 g1e2 g8e7 e1g1 e8g8 c1g5 h7h6 g5e7 d8e7 f1e1 e7d7 e2f4 a7a5 d1c2 b4e7 a1d1 f8e8 f4h5 a5a4 a2a3 b7b5 c3e2 b5b4 e2g3 b4b3 c2c3 c6b5 f3f4 c7c5 f4f5 c5d4 c3d4 a8c8');
INSERT INTO evaluations VALUES(17419,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(17419,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(17419,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(17419,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17419,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(17419,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(17419,36,'cp 37','d8c7 d4c6 b7c6 d1e2 g8e7 b1c3 e7g6 f2f4 f8e7 e4e5 c8b7 c3e4 c6c5 c1d2 g6h4 b2b3 h4f5 d2c3 f5d4 c3d4 c5d4 a1d1 f7f5 e4g5 c7b6 g1h1 a8c8 h2h3 e7c5 d1d3 e8g8 f1d1 h7h6 g5f3 b7e4 d3d2 f8d8 a2a4 e4f3');
INSERT INTO evaluations VALUES(20490,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(20490,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(20490,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(20490,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(20490,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(20490,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(20490,36,'cp 55','f8c5 c3d5 g8e7 d5e3 d8d1 e1d1 e8g8 f1c4 e7g6 h1e1 g6e5 f3e5 c6e5 c4b5 a7a6 b5f1 f8d8 c1d2 c5e3 e1e3 e5g4 e3e7 c8e6 f1e2 g4f2 d1e1 f2e4 d2a5 d8e8 e7e8 a8e8 e1f1 e4f6 e2f3 e6d5 f1f2 d5f3 f2f3 f6d5');
INSERT INTO evaluations VALUES(21094,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(21094,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(21094,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(21094,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(21094,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(21094,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(21094,36,'cp -36','c8g4 h2h3 g4f3 d1f3 e7e6 f1b5 g8e7 d2d3 a7a6 b5c6 e7c6 e1g1 f7f5 e5f6 d8f6 f3h5 g7g6 h5g4 f6f5 f1e1 e8c8 e1e6 f5g4 h3g4 d4c3 b1c3 d8d3 c1e3 f8a3 a1d1 d3d1 c3d1 h8d8 b2a3 d8d1 g1h2 d1d8 e3c5 d8d3 e6e8 c8d7 e8h8');
INSERT INTO evaluations VALUES(21362,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21362,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(21362,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(21362,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(21362,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21362,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(21362,36,'cp 32','e6d5 g1f3 c8f5 a2a3 g8f6 c1g5 f8e7 e2e3 f6e4 g5e7 c6e7 f1d3 c7c6 d3e4 f5e4 c3e4 d5e4 f3d2 f7f5 a1c1 d8d5 c1c5 d5f7 g2g3 e8c8 d2c4 d8d5 d1b3 h8d8 c5d5 f7d5 e1g1 c8b8 g1g2 g7g5 f1c1 h7h6 a3a4 b8a8');
INSERT INTO evaluations VALUES(21668,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(21668,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(21668,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21668,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(21668,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21668,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21668,36,'cp 0','e5d4 e3d4 g8f6 c4c5 f8c5 d4c5 d5d4 d1e2 c8e6 c3e4 d8e7 e4f6 g7f6 f2f3 c6b4 e2b5 e6d7 b5e2');
INSERT INTO evaluations VALUES(23156,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(23156,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(23156,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(23156,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(23156,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(23156,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(23156,36,'cp 81','g2f3 d8f6 h1g1 f8b4 b1d2 d7d6 c2c3 b4c5 d2b3 c5b6 c1g5 f6g6 g1g3 f7f6 g5e3 g6f7 b3d4 g8e7 d4b5 e8d8 d1a4 b6e3 f2e3 c8d7 e1c1 a7a6 a4a5 d7b5 f1b5 d8c8 a5a4 g7g5 b5c4 f7g6 c4e6 c8b8 a4d7 e7c6 f3f4 h7h5 e6d5 g6e8 d7g7 h8f8 g7h7 c6e7 d5c4 h5h4 g3g1');
INSERT INTO evaluations VALUES(23988,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(23988,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(23988,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(23988,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(23988,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(23988,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(23988,36,'cp 63','e8g8 c4d3 f8e8 h2h3 c6c5 d1f3 d7c6 d3c4 e8f8 b2b3 f6d7 c1f4 e7f6 a1e1 d7b6 c4a6 f8e8 c3d5 a8b8 f3d3 b6d5 e4d5 c6d7 e1e8 d7e8 d3f3 e8d7 f4d2 g7g6 f1e1 h7h5 a6d3 f6d4 d2a5 g8g7 a2a4 d4e5 a5d2 e5d4');
INSERT INTO evaluations VALUES(23648,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(23648,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(23648,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(23648,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(23648,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(23648,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(23648,36,'cp 75','c4d3 c6c5 a1b1 f8e8 b2b3 d7c6 f1e1 f6d7 c3d5 e7h4 c1d2 d7e5 d2a5 e5d3 d1d3 c6d5 e4d5 d8d7 a5c3 h4g5 h2h3 h7h5 a2a4 g7g6 d3f3 e8e1 b1e1 a8e8 e1e8 d7e8 g1f1 e8e7 a4a5 f7f5 a5a6 g5f6 c3a5 f6g7 f3d1 g7d4 d1d2 d4g7 g2g3');
INSERT INTO evaluations VALUES(24149,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(24149,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(24149,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24149,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(24149,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(24149,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(24149,36,'cp 14','c4b5 e6d7 b5c6 d7c6 d2d4 e5d4 f3d4 e8g8 c1g5 f8e8 f1e1 h7h6 g5h4 d8e7 d4c6 b7c6 c3a4 c5b6 a4b6 a7b6 d1d4 c6c5 d4c3 b6b5 h4f6 e7f6 c3f6 g7f6 a2a3 g8g7 a1d1 g7g6 d1d5 e8e5 f2f3 e5d5 e4d5');
INSERT INTO evaluations VALUES(25645,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(25645,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(25645,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(25645,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(25645,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(25645,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(25645,36,'cp 22','e7e6 b1d2 g8f6 f4g3 f8d6 f1e2 d8e7 f3h4 f5e4 a1c1 e8g8 e1g1 d6g3 h2g3 h7h6 c3c4 e7b4 d2b3 c6a5 b3a5 b4a5 c4d5 e6d5 d1b3 f8b8 h4f3 f6d7 a2a3 c7c6 c1c3 c6c5 d4c5 d7c5 b3a2 c5e6');
INSERT INTO evaluations VALUES(25665,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(25665,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(25665,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(25665,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(25665,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25665,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(25665,36,'cp 34','f1d3 f5d3 d1d3 f8d6 f4g3 h7h5 b1d2 h5h4 g3d6 c7d6 h2h3 e6e5 d4e5 d6e5 e3e4 g8f6 d3c2 d5d4 c3d4 c6d4 f3d4 d8d4 e1g1 e8g8 d2f3 d4e4 c2e4 f6e4 f1e1 e4c5 e1e5 a8c8 f3h4 f8e8 a1e1 e8e5 e1e5 c5d3 e5e2 d3c1 e2e7 b7b5 e7a7 c8c2');
INSERT INTO evaluations VALUES(29480,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(29480,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(29480,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(29480,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(29480,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(29480,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29480,36,'cp -16','g7f6 e2e3 e7e6 h2h4 h8g8 g1e2 h7h6 e2f4 c7c6 f1d3 f5d3 c2d3 f6f5 d1h5 d8f6 e1c1 b8d7 h5f3 f8b4 c1b1 b4c3 b2c3 f6e7 f3e2 e8c8 d1c1 c6c5 c3c4 d5c4 c1c4 b7b5 c4c1 c8b8 d4c5 d7c5');
INSERT INTO evaluations VALUES(35299,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(35299,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(35299,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(35299,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(35299,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(35299,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(35299,36,'cp -67','c3d4 e7e4 c1e3 g8f6 d1d2 e4c6 b1c3 f8b4 g1e2 f6e4 d2d3 e4c3 b2c3 b4a5 d4d5 c6g6 d3d2 d7d6 e2f4 g6f6 e3d4 f6e7 f1e2 e8g8 e1g1 b8d7 f4h5 f7f6 a1e1 e7f7 g1h1 a5b6 h5f4 b6d4 c3d4 d7b6 d2c2 b6d5');
INSERT INTO evaluations VALUES(37472,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(37472,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37472,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(37472,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(37472,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37472,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(37472,36,'cp 23','g1f3 d7d5 c2c4 e7e6 d2d4 b7b6 e1g1 c8b7 b1c3 b8d7 d1c2 c7c5 c4d5 e6d5 a1c1 a8c8 c3b5 b7a6 a2a4 c5d4 c2d1 c8c1 d1c1 d7c5 f3d4 d8d7 c1b1 f6e4 f1d1 f8e8 e2e3 e4g5 d4f3 g5f3 g2f3 a6b5 a4b5 d7b5');
INSERT INTO evaluations VALUES(37471,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(37471,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(37471,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(37471,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(37471,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37471,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(37471,36,'cp 12','b7b6 c2c4 e7e6 e1g1 d7d5 d2d4 c8b7 c4d5 e6d5 b1c3 f8e8 e2e3 b8d7 a1c1 c7c6 c1c2 f6e4 d1c1 a8c8 f1d1 d8e7 g2h3 e4c3 c2c3 c8c7 b2a3 e7f6 h3g2 c6c5 f3e1 c7c8 c3c2 c5d4 d1d4 c8c2 c1c2 f6e6 d4d2 e8c8');
INSERT INTO evaluations VALUES(37502,30,'cp 1351',NULL);
INSERT INTO evaluations VALUES(37502,31,'cp 1341',NULL);
INSERT INTO evaluations VALUES(37502,32,'cp 1341',NULL);
INSERT INTO evaluations VALUES(37502,33,'cp 1348',NULL);
INSERT INTO evaluations VALUES(37502,34,'cp 1351',NULL);
INSERT INTO evaluations VALUES(37502,35,'cp 1348',NULL);
INSERT INTO evaluations VALUES(37502,36,'cp 1355','h8g8 e8e7 g8h7 d7d5 b1c3 d5e4 c3e4 d8h8 h7h8 f6h8 d2d3 e7f8 f2f3 b8d7 g1e2 a7a5 c2c3 a5a4 e1d2 b7b6 h2h4 h8g7 d2c2 c8a6 c1f4 a8a7 d3d4 d7f6 e4f6 g7f6 e2g3 a6f1 a1f1 f6d8 g3e4 d8e7 h4h5 g6h5 h1h5');
INSERT INTO evaluations VALUES(37501,30,'cp 1330',NULL);
INSERT INTO evaluations VALUES(37501,31,'cp 1328',NULL);
INSERT INTO evaluations VALUES(37501,32,'cp 1340',NULL);
INSERT INTO evaluations VALUES(37501,33,'cp 1340',NULL);
INSERT INTO evaluations VALUES(37501,34,'cp 1330',NULL);
INSERT INTO evaluations VALUES(37501,35,'cp 1332',NULL);
INSERT INTO evaluations VALUES(37501,36,'cp 1332','e8e7 g8h7 d7d5 b1c3 d5e4 c3e4 d8h8 h7h8 f6h8 g1e2 a7a5 e2f4 c7c6 c2c3 b8d7 d2d4 d7f6 f2f3 b7b5 a2a3 h8g7 e1f2 c8d7 c1e3 f6d5 f4d5 c6d5 e4c5 d7e8 f1d3 g7f6 h2h4 e7f8 g2g4 f8g8 a1e1 a5a4 e3f4 a8c8 g4g5 f6g7');
INSERT INTO evaluations VALUES(37478,30,'cp 1355',NULL);
INSERT INTO evaluations VALUES(37478,31,'cp 1355',NULL);
INSERT INTO evaluations VALUES(37478,32,'cp 1361',NULL);
INSERT INTO evaluations VALUES(37478,33,'cp 1355',NULL);
INSERT INTO evaluations VALUES(37478,34,'cp 1359',NULL);
INSERT INTO evaluations VALUES(37478,35,'cp 1361',NULL);
INSERT INTO evaluations VALUES(37478,36,'cp 1355','g8h7 d7d5 b1c3 d5e4 c3e4 d8h8 h7h8 f6h8 g1e2 e7f8 c2c3 b8d7 e2f4 d7f6 e4g5 c8d7 f1c4 d7e8 a2a4 a7a5 e1g1 c7c6 d2d4 b7b5 c4b3 b5b4 g2g3 a8b8 b3c4 b8b7 f1e1 f6g4 f2f3 g4f6 g5e4 f6e4 f3e4 b4c3 b2c3');
INSERT INTO evaluations VALUES(37500,30,'cp 1332',NULL);
INSERT INTO evaluations VALUES(37500,31,'cp 1333',NULL);
INSERT INTO evaluations VALUES(37500,32,'cp 1325',NULL);
INSERT INTO evaluations VALUES(37500,33,'cp 1337',NULL);
INSERT INTO evaluations VALUES(37500,34,'cp 1344',NULL);
INSERT INTO evaluations VALUES(37500,35,'cp 1340',NULL);
INSERT INTO evaluations VALUES(37500,36,'cp 1332','d7d5 b1c3 d5e4 c3e4 d8h8 h7h8 f6h8 g1e2 b7b6 d2d4 e7f8 c2c3 c8b7 e4g5 b8c6 e2f4 a8d8 f2f3 d8d7 h2h4 c6e7 f1c4 e7d5 e1f2 a7a5 a2a4 f8g7 g5e4 d5e7 g2g3 g7f8 g3g4 e7c8 b2b3 c8e7 h4h5 g6h5 h1h5');
INSERT INTO evaluations VALUES(38635,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38635,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38635,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(38635,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(38635,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(38635,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38635,36,'cp 0','d2d4 h7h6 c2c4 c7c6 e1g1 g8e7 f1e1 e8g8 b1c3 c8e6 d1b3 d5c4 e2c4 e6c4 b3c4 b8d7 c3e4 d6c7 c1d2 e7d5 a1c1 f8e8 h2h3 a8b8 a2a3 d5f6 e4c5 e8e1 c1e1 c7d6 c4c2 d6c7');
INSERT INTO evaluations VALUES(39351,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(39351,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(39351,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(39351,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(39351,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(39351,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(39351,36,'cp 104','g8f6 c2c3 d7d5 e4d5 b4d6 d2d3 e8g8 c1g5 h7h6 g5f6 d8f6 b1d2 c8f5 e1g1 b8d7 f1e1 a7a6 a2a4 f5g6 d2e4 f6d8 d3d4 g6e4 e1e4 f8e8 d4e5 d7e5 f3e5 e8e5 e4e5 d6e5 g2g3 e5d6 d1e2 g7g6 a1e1 g8g7 g1g2 h6h5 h2h4 d8f6 c4d3 a8h8');
INSERT INTO evaluations VALUES(42758,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(42758,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(42758,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(42758,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(42758,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(42758,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(42758,36,'cp 109','c2c3 d7d5 e4d5 b4d6 d2d3 e8g8 e1g1 h7h6 c1e3 c7c6 d5c6 b8c6 b1d2 c8f5 f1e1 a8c8 a2a4 d6b8 d3d4 f6g4 d4e5 c6e5 f3e5 g4e5 c4f1 e5d3 f1d3 d8d3 d2f3 d3c4 f3d4 f5g6 d1f3 c4c7 g2g3 c7d7 d4b5 a7a6 a1d1 d7h3 b5d6 b8d6 d1d6 f8e8');
INSERT INTO evaluations VALUES(40847,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(40847,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(40847,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(40847,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(40847,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(40847,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(40847,36,'cp 60','e1g1 d6d5 f1e1 e8g8 h2h3 b8d7 c1f4 f8e8 b1d2 d7f8 f4g3 e7d6 f3e5 c7c5 d4c5 d6c5 c2c3 c5d6 d2f3 d8c7 d3c2 c8e6 f3d4 e6d7 e5d7 e8e1 d1e1 c7d7 g3d6 d7d6 d4f5 d6b6 c2b3 f8g6 a1d1 a8e8 e1d2 h7h6 d1e1 e8e1 d2e1 g6f4');
INSERT INTO evaluations VALUES(40846,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(40846,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(40846,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(40846,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(40846,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(40846,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(40846,36,'cp 38','d6d5 h2h3 e8g8 f1e1 b8d7 c1f4 f8e8 b1d2 d7f8 f4g3 e7d6 f3e5 f8g6 d1f3 c7c5 e5g6 e8e1 a1e1 h7g6 d4c5 d6c5 d2b3 c5b6 g3e5 b6c7 f3f4 c7e5 f4e5 d8b6 e5d4 b6d6 c2c3 b7b6 b3d2 c8d7 d2f3 a8e8 a2a3');
INSERT INTO evaluations VALUES(41098,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(41098,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(41098,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(41098,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(41098,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(41098,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(41098,36,'cp 54','f1e1 e8g8 h2h3 b8d7 c1f4 f8e8 b1d2 d7f8 f4g3 e7d6 f3e5 c7c5 d4c5 d6c5 c2c3 c5d6 d2f3 d8c7 d3c2 c8e6 d1d4 f6e4 g3h2 a8d8 a1d1 a7a6 e5d3 d6h2 f3h2 e6d7 c2b3 d7c6 d3b4 f8e6 d4a7 e6f4 h2f3 e4f6 f3d4 e8e1 d1e1 h7h6');
INSERT INTO evaluations VALUES(42426,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(42426,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(42426,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(42426,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(42426,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(42426,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(42426,36,'cp 77','d1g4 e4c3 g4g7 e8e7 b2c3 d7d6 g7g5 e7e8 g5e3 d8f6 g1e2 b8c6 e2f4 e8d8 e1c1 c8d7 g2g3 c6e7 g3g4 h8g8 h1g1 f6g5 f1e2 g5a5 c1b2 a5b6 b2a1 e7d5 f4d5 e6d5 g4g5 g8e8 e3d2 b6c6 d1e1 c6a4 e2d3 e8e1 g1e1');
INSERT INTO evaluations VALUES(44336,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(44336,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(44336,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(44336,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(44336,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(44336,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(44336,36,'cp 62','d7d6 h2h3 f8e7 g1f3 e8g8 f1c4 f8e8 e1g1 e7f8 c1g5 c8f5 a2a4 h7h6 g5h4 b8d7 g2g4 f5h7 c4d3 d8b6 a4a5 b6c7 c3b5 c7b8 h4f6 d7f6 d3h7 f6h7 d1d3 b8d8 g1g2 a7a6 b5c3 g7g6 f3d2 f8g7 d2c4 g7c3 b2c3 h7f6 c4b6');
INSERT INTO evaluations VALUES(45931,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(45931,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(45931,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(45931,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(45931,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(45931,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(45931,36,'cp 64','g1e2 a7a6 a2a4 h7h5 e2g1 g7g6 c1f4 d8e7 f1e2 f8g7 g1f3 e8g8 f3d2 c8f5 e1g1 b8d7 h2h3 f8e8 f4g5 d7e5 f2f4 e5d7 g1h1 e7f8 g5h4 f6h7 d2c4 a8b8 e2d3 f5d3 d1d3 b7b5 a4b5 a6b5 c4a5 b5b4 c3b5 g7b2');
INSERT INTO evaluations VALUES(45338,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(45338,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(45338,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(45338,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45338,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45338,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45338,36,'cp 0','e4d5 f6d5 g1f3 e7e6 d3d4 c5d4 d1d4 b8c6 d4e4 f8e7 e1g1 d5f6 e4h4 f6d5');
INSERT INTO evaluations VALUES(45320,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45320,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45320,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45320,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45320,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45320,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45320,36,'cp 0','f6d5 g1f3 e7e6 d3d4 c5d4 d1d4 b8c6 d4e4 f8e7 e1g1 d5f6 e4h4 f6d5');
INSERT INTO evaluations VALUES(45498,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45498,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45498,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45498,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45498,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45498,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45498,36,'cp 0','g1f3 e7e6 d3d4 c5d4 d1d4 b8c6 d4e4 f8e7 e1g1 d5f6 e4h4 f6d5');
INSERT INTO evaluations VALUES(48526,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(48526,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(48526,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(48526,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(48526,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(48526,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(48526,36,'cp 62','c7c5 g1f3 c5d4 f3d4 b8c6 d4c6 d8d1 c3d1 b7c6 f2f3 f8c5 c1e3 c5e3 d1e3 e8e7 e1c1 f6d7 d1d3 d7b6 d3c3 c8b7 f1e2 h8d8 h2h4 d8d4 h1d1 d4d1 e3d1 e7d6 f3f4 a8d8 d1f2 f7f6 b2b4 b6d7 f2g4 c6c5 e4e5 f6e5');
INSERT INTO evaluations VALUES(49924,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(49924,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(49924,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(49924,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(49924,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(49924,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(49924,36,'cp 54','f8b4 d1c2 b8c6 b2b3 e8g8 c1b2 b7b6 g1f3 c8b7 e1g1 d8e7 d2d4 a8e8 e2e3 b4c3 b2c3 c6d8 d4d5 c7c5 d5c6 b7c6 a1d1 d8f7 f3e5 c6g2 g1g2 e8c8 e5f7 f8f7 c3f6 f7f6 c2d3 c8d8 d3d6 e7e8 b3b4 h7h6 a2a3');
INSERT INTO evaluations VALUES(50053,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(50053,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(50053,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50053,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(50053,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(50053,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(50053,36,'cp 27','d7d5 f1d3 d5e4 d3e4 f7f5 e4f3 b8c6 d4d5 e6d5 c4d5 c6e5 b1c3 e7g6 h2h4 e5f7 f3e2 c8d7 g1f3 f8d6 c1e3 e8g8 d1b3 f7e5 e1c1 e5f3 g2f3 a7a6 c1b1 g6f4 e3f4 d6f4 d5d6 g8h8 d6c7 d8c7 c3d5 d7e6 d5c7 e6b3 a2b3 f4c7');
INSERT INTO evaluations VALUES(50411,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(50411,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(50411,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(50411,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(50411,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(50411,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(50411,36,'cp 121','b4c6 d2d4 d7d6 g1f3 g7g6 e5d6 e7d6 d4d5 c6e5 f3e5 d6e5 h2h4 f8g7 b1c3 e8g8 h4h5 b8d7 c1e3 b7b6 h5g6 f7g6 f1d3 e5e4 d3e4 d8e7 d1c2 d7c5 e1c1 c5e4 c2e4 e7f7 e4h4 g7f6 e3g5 f6g5 h4g5 f7f2 g5h6 f8f7 c3e4 f2g2');
INSERT INTO evaluations VALUES(50306,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(50306,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(50306,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(50306,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(50306,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(50306,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(50306,36,'cp 152','d2d4 d7d6 g1f3 c8g4 e5d6 c7d6 f1e2 e7e5 d4e5 d6e5 d1d8 e8d8 b1c3 b8d7 e1g1 a8c8 f1d1 d8e8 c1e3 f7f5 h2h3 g4h5 g2g4 f5g4 f3h2 f8e7 h3g4 h5f7 c3d5 d7f6 d5e7 e8e7 e3c5 e7e8 h2f3 f6g4 f3g5 g4f6 a1c1 c8d8 d1d8 c6d8');
INSERT INTO evaluations VALUES(50307,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(50307,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(50307,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(50307,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(50307,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(50307,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(50307,36,'cp 131','d7d6 g1f3 c8g4 e5d6 c7d6 f1e2 e7e5 d4e5 d6e5 d1d8 e8d8 b1c3 b8d7 e1g1 a8c8 f1d1 d8e8 c1e3 f8e7 a1c1 f7f6 c3d5 e5e4 f3d4 g4e2 d4e2 f6f5 e2d4 c6d4 e3d4 e7g5 c1c3 e8f7 d4a7 h8d8 c3g3 h7h6 g3b3 c8c4 b3b7 c4c2');
INSERT INTO evaluations VALUES(51694,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(51694,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(51694,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(51694,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(51694,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(51694,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(51694,36,'cp 43','a2a4 f8b4 c1d2 a7a5 a4b5 b4c3 d2c3 c6b5 b2b3 c8b7 b3c4 b5b4 c3b2 g8f6 f1d3 e8g8 e1g1 d8c7 f3d2 b8d7 d3c2 f8d8 f1e1 d7b6 c4c5 b6d7 c2a4 f6e4 a1c1 b7d5 d2f1 e4f6 f1g3 f6e4 g3e4 d5e4 d1g4 d7f6 g4g3 c7g3 h2g3 d8c8 f2f3 e4c6');
INSERT INTO evaluations VALUES(51696,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(51696,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(51696,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(51696,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(51696,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(51696,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(51696,36,'cp 37','f8b4 c1d2 a7a5 a4b5 b4c3 d2c3 c6b5 b2b3 c8b7 b3c4 b5b4 c3b2 g8f6 f1d3 e8g8 e1g1 b8d7 f3d2 d8c7 d3c2 f8d8 f1e1 d7b6 a1c1 a5a4 c4c5 c7c6 e3e4 b6d7 c2d3 d7f8 d2c4 b7a6 b2a1 f8g6 c4b6 a6d3 d1d3 a8b8 d4d5 e6d5 a1f6 c6f6 e4d5 a4a3 d5d6 g6f4');
INSERT INTO evaluations VALUES(37598,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(37598,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(37598,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(37598,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(37598,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(37598,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(37598,36,'cp 30','c1d2 a7a5 a4b5 b4c3 d2c3 c6b5 b2b3 c8b7 b3c4 b5b4 c3b2 g8f6 f1d3 e8g8 e1g1 d8c7 f3d2 b8d7 d3c2 f8d8 f1e1 d7b6 c4c5 b6d7 c2a4 e6e5 d1c2 b7c6 d2b3 c6a4 a1a4 e5d4 b2d4 d7e5 e1a1 e5c6 d4f6 g7f6 h2h3 d8d5 c2e4 a8d8 e4f3 f6f5 b3a5 c6a5 a4a5');
INSERT INTO evaluations VALUES(37597,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(37597,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(37597,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(37597,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(37597,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(37597,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(37597,36,'cp 40','g8f6 a4b5 b4c3 d2c3 c6b5 b2b3 e8g8 b3c4 b5c4 f1c4 c8b7 e1g1 a7a5 f3d2 b8d7 d1b3 d8c7 f1c1 f8c8 c4d3 a5a4 b3a3 f6g4 d2f1 c7d8 c3b4 c8c1 a1c1 h7h6 h2h3 g4f6 f1d2 b7a6 d3e4 a8c8 c1c8 d8c8 a3a4 f6e4');
INSERT INTO evaluations VALUES(52011,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(52011,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(52011,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(52011,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(52011,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(52011,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(52011,36,'cp -25','d7d5 g1f3 b8c6 g2g3 f8e7 f3e5 c6e5 f4e5 e7g5 b1d2 g5h6 f1g2 g8e7 e4d5 e6d5 d2b3 h6c1 a1c1 e8g8 e1g1 d8b6 d3d4 c5c4 b3c5 b6b2 c1b1 b2a2 d1d2 b7b6 b1a1 a2b2 a1b1 c4c3 d2g5 b2c2 g5e7 b6c5 g2d5 c8h3 d5a8 c5d4 a8f3 h3f1 b1f1 c2a2');
INSERT INTO evaluations VALUES(54261,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(54261,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(54261,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(54261,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(54261,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(54261,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(54261,36,'cp -11','g2g3 b8c6 g1f3 g8f6 e4e5 f6g4 c2c3 h7h5 f1g2 g4h6 b1a3 c6e7 f3h4 h6f5 h4f5 e7f5 a3c2 g7g6 e1g1 c8d7 c2e3 f8e7 a2a4 a7a5 g2f3 f5g7 f1f2 d7c6 b2b3 d5d4 c3d4 d8d4 f3c6 b7c6 d1f3 e8g8 e3c2');
INSERT INTO evaluations VALUES(56070,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(56070,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(56070,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(56070,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(56070,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(56070,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(56070,36,'cp 50','d2d4 b8d7 g1f3 g8f6 e4f6 d7f6 c2c3 a7a6 f3e5 c7c5 c1e3 c5d4 e3d4 d8c7 d1a4 f6d7 e5f3 c7b8 a4c2 b8c7 f1e2 d7f6 e1c1 f8d6 e2d3 b7b5 c1b1 c8b7 h1e1 a8d8 f3g5 d6e7 g2g3 b7d5 d3h7 c7a5 b2b3 a5c7 h7g6 f7g6 c2g6 e8d7 g5f7 h8h2 f7d8 e7d8');
INSERT INTO evaluations VALUES(1501,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(1501,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(1501,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(1501,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(1501,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(1501,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(1501,36,'cp -43','g8h6 b1c3 h6f5 c3a4 b6a5 a4c3 f8b4 e1f2 c8d7 c1d2 b4e7 c3e2 a5b6 d1b3 e7d8 d2c3 h7h5 g2g3 a7a5 f1h3 f5h6 h1e1 b6b3 a2b3 d8e7 e1c1 e8g8 f3e1 a8a6 e1d3 f8a8 d3c5 e7c5 d4c5 b7b6 c3d4 b6c5 d4c5');
INSERT INTO evaluations VALUES(5039,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(5039,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(5039,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(5039,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5039,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5039,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5039,36,'cp 8','f1e2 g8f6 b1c3 d7d6 e1g1 e8g8 a1b1 f6d7 d1c2 c6c5 f1d1 c8b7 a2a3 a7a5 b2b3 d8b6 c1e3 b7c6 b3b4 a5b4 a3b4 a8a3 d1d3 g7c3 d3c3 c6e4 c2e4 a3c3 e4e7 c3e3 e7e3 c5b4 e3b6 d7b6 b1b4 b6d7 e2f3 d7e5 f3d5 g8g7 b4b7');
INSERT INTO evaluations VALUES(6984,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(6984,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(6984,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(6984,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(6984,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(6984,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(6984,36,'cp -20','g8f6 d4d5 c6a5 c4d3 c7c6 b1c3 d7d6 e1g1 e8g8 h2h3 c8d7 c1e3 h7h6 a1b1 f8e8 d5c6 b7c6 e3f4 e7f8 e4e5 f6d5 c3d5 c6d5 e5d6 d8f6 d3b5 d7b5 b1b5 a5c4 d6d7 e8d8 f4g3 d8d7 b5d5 d7d5 d1d5 c4b6 d5a5 f6c6 g3e5');
INSERT INTO evaluations VALUES(7035,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(7035,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(7035,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(7035,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(7035,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(7035,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(7035,36,'cp 94','c1e3 d7d6 d1d2 a7a6 f2f3 b7b5 a2a4 b5b4 c3a2 a6a5 f1b5 c8d7 g1e2 g8f6 e1g1 e8g8 b2b3 e7e6 a1d1 c6e7 b5d3 c7c6 a2c1 d7c8 f1e1 f6d7 e3h6 g7h6 d2h6 c6c5 d4c5 d7c5 d3c4 d6d5 e4e5 c5d7');
INSERT INTO evaluations VALUES(7033,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(7033,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(7033,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(7033,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(7033,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(7033,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(7033,36,'cp 86','d7d6 d1d2 a7a6 f2f4 b7b5 g1f3 e7e6 a2a4 b5a4 a1a4 g8f6 e3f2 e8g8 e4e5 f6d7 f1d3 d7b6 a4a1 c6b4 d3e4 b6c4 d2c1 a8b8 f2h4 d8d7 e1g1 c4b6 h2h3 b6d5 e4d5 b4d5 c3d5 e6d5 a1a3 d7c6 h4e7 f8e8 e7g5 h7h6 g5f6');
INSERT INTO evaluations VALUES(5699,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(5699,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(5699,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(5699,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(5699,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(5699,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(5699,36,'cp 97','f2f4 g8f6 g1f3 e8g8 e4e5 f6g4 e3g1 d6e5 d4e5 g4h6 h2h3 c8e6 d1e2 g8h8 a1d1 d8c8 g2g4 f7f6 f1g2 e6g8 g1h2 c8e6 e1g1 a8d8 d1d8 f8d8 h2g3 f6e5 f4e5 e6c4 e2c4 g8c4 f1e1 h6f7 b2b3 c4a6 c3b5 a6b5 c2c4 d8d3 c4b5');
INSERT INTO evaluations VALUES(7573,30,'cp 555',NULL);
INSERT INTO evaluations VALUES(7573,31,'cp 575',NULL);
INSERT INTO evaluations VALUES(7573,32,'cp 586',NULL);
INSERT INTO evaluations VALUES(7573,33,'cp 571',NULL);
INSERT INTO evaluations VALUES(7573,34,'cp 587',NULL);
INSERT INTO evaluations VALUES(7573,35,'cp 582',NULL);
INSERT INTO evaluations VALUES(7573,36,'cp 591','d4f4 c8b7 e5f6 d8f6 f4f6 g7f6 c1e3 c5e3 f2e3 e8c8 e1c1 h8e8 c3d5 e8e5 d5f6 e5f5 f6g4 f5g5 g4f2 d7d5 f2h3 g5g8 h1g1 c7c5 f1e2 c8b8 h3f4 g8e8 g2g3 f7f5 e2b5 e8e7 b5c4 d5d4 f4e6 d8c8 e3d4 c5d4 d1d4 c8c4 d4c4 e7e6');
INSERT INTO evaluations VALUES(8546,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(8546,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(8546,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(8546,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(8546,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(8546,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(8546,36,'cp 45','e5d4 f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 b1c3 e8g8 b2b3 f6d7 c1e3 f8e8 d1d3 d7e5 d3d2 a7a5 a1d1 e7h4 h2h3 d8b8 e3d4 b8b4 g1f1 h7h6 e1e2 e5d7 f2f4 h4f6 d2e3 b4b7 e3g3 f6d4 d1d4 d7c5 e4e5 d6e5 f4e5 c5e6');
INSERT INTO evaluations VALUES(8567,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(8567,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(8567,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(8567,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(8567,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(8567,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8567,36,'cp 51','f3d4 c8d7 d4c6 d7c6 b5c6 b7c6 b1c3 e8g8 f2f4 a7a5 d1f3 a5a4 g1h1 d8c8 h2h3 f8e8 c1d2 f6d7 e1e2 e7f6 e4e5 d6e5 f3c6 d7b6 a1e1 f6h4 e1d1 e8e6 c6f3 e5f4 e2e6 c8e6 d2f4 c7c6 b2b3 h4f6 f4g3 a4b3');
INSERT INTO evaluations VALUES(8565,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(8565,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(8565,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(8565,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(8565,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(8565,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(8565,36,'cp 38','c8d7 d4c6 d7c6 b5c6 b7c6 b1c3 e8g8 b2b3 f6d7 c1e3 f8e8 d1d3 d7e5 d3d2 e7h4 a1d1 a7a5 e3d4 h7h6 c3a4 d8e7 e1e2 h4f6 f2f4 e5g4 a4c3 a8d8 h2h3 f6d4 d2d4 g4f6 d4f2 a5a4 b3a4 f6d7 d1e1 d7c5 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(10225,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10225,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10225,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(10225,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(10225,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10225,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10225,36,'cp 0','f3e5 c6e5 d2d4 d7d5 d4e5 f6e4 f2f3 c5f2 e1d1 d8h4 g2g3 f2g3 h2g3 h4h1 f3e4 c8g4 e2g4 h1f1 d1c2 d5e4 g4e4 e8c8 b1d2 f1f2 a2a4 f2g3 a4a5 g3g6 a1a4 g6e4 d2e4 h7h5 e4g5 h5h4 g5f7 h4h3 f7h8 d8h8 c1f4 h3h2 f4h2 h8h2 c2b3 h2e2 a4g4 e2e5 g4g7 e5a5 g7g8 c8d7 g8g7');
INSERT INTO evaluations VALUES(11073,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(11073,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(11073,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(11073,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11073,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(11073,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11073,36,'cp 0','d2d4 d7d5 e4d5 e6d5 c4e2 f6e4 e1g1 f8e7 c1e3 e4c3 b2c3 c5c4 d1d2 e8g8 a3a4 c8f5 a1b1 d8c7 e3f4 c7d7 h2h3 b7b5 a4b5 a6b5 b1b5 a8a2 e2d1 a2a1 d1e2');
INSERT INTO evaluations VALUES(13734,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(13734,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(13734,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(13734,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(13734,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(13734,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(13734,36,'cp -35','f8e7 c4b5 e8g8 b5c6 b7c6 h2h3 c8f5 b2b3 e7d6 c1a3 d8c8 d2d4 f5h3 f3e5 h3f5 a3c5 d6c5 d4c5 c8c7 b3b4 a7a6 d1d4 f5c2 a2a4 f8e8 e5f3 c2e4 b4b5 e4f3 g2f3 a6b5 a4b5 e8e1 a1e1 c6b5 c3b5 c7d7 b5d6 h7h6 g1g2 d7c6 e1b1 f6d7 b1b5 d7f8 b5b6 f8e6 b6c6 e6d4');
INSERT INTO evaluations VALUES(8115,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(8115,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(8115,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8115,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(8115,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(8115,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8115,36,'cp 0','c4b5 e8g8 b5c6 b7c6 h2h3 e7d6 d2d3 c8e6 c1g5 d8c8 g5f6 g7f6 f3h4 e6h3 d1f3 h3g4 f3f6 c8d8 f6h6 d6e7 g2g3 e7f6 h6f4 h7h5 h4f5 f6g5 f4e5 g5f6');
INSERT INTO evaluations VALUES(8109,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8109,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(8109,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(8109,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(8109,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(8109,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(8109,36,'cp -13','e8g8 b5c6 b7c6 h2h3 c8f5 b2b3 e7d6 c1a3 d8c8 d2d4 f5h3 f3e5 h3f5 a3c5 d6c5 d4c5 c8c7 b3b4 a7a6 e5f3 f5g4 d1d3 g4f3 d3f3 a6a5 a2a3 a8b8 a1b1 h7h6 e1d1 f8d8 c3e2 d8e8 f3f4 c7f4 e2f4 a5b4 a3b4 e8e4');
INSERT INTO evaluations VALUES(3112,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(3112,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(3112,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(3112,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(3112,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(3112,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(3112,36,'cp -23','b5c6 b7c6 h2h3 c8f5 d2d3 f8e8 b2b3 f6d7 c3e2 a7a5 e2g3 f5e6 a2a4 e7d6 f3g5 d8f6 c1d2 d7f8 g3h5 f6f5 h5g3 f5g6 g5e6 f8e6 d1g4 g6g4 h3g4 d6g3 f2g3 f7f6 e1e2 e6d4 e2e8 a8e8 d2a5 e8e2 a5b6 e2c2 a4a5 d4b3');
INSERT INTO evaluations VALUES(14151,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(14151,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(14151,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14151,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14151,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(14151,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14151,36,'cp 0','c8g4 c1e3 g4f3 d1f3 c6d4 f3e4 c7c5 b1c3 g7g6 e1c1 f8g7 e3d4 c5d4 c4c5 b6d7 f1b5 d8c7 c3d5 c7e5 e4e5 g7e5 h1e1 f7f6 e1e5 f6e5 d5c7 e8d8 c7a8 d7c5 d1e1 e5e4 b2b4 c5d3 b5d3 e4d3 c1d2 d8d7 a8c7 d7c7 e1e7 c7c6 d2d3 h8f8 a2a4 f8f2 e7h7 f2g2 d3d4 a7a6 d4c4 g2f2 a4a5 f2f4 c4c3 f4f3 c3c4');
INSERT INTO evaluations VALUES(14840,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14840,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14840,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14840,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(14840,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(14840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14840,36,'cp 10','f1g2 f8e7 e1g1 e8g8 c1f4 b7b6 f1e1 c8b7 e2e4 f6e4 c3e4 d5e4 e1e4 e7d6 f4d2 c6e7 e4e1 c7c5 d4c5 d6c5 d1e2 d8c7 d2f4 c7c6 c2c3 a8d8 a1d1 e7g6 f4g5 d8d1 e1d1 c6c7 g5e3 c5e3 e2e3 h7h6 h2h4 b7d5 a2a3 f8d8 d1d4 g6e7 f3e5 d5g2 g1g2');
INSERT INTO evaluations VALUES(15024,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(15024,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(15024,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(15024,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(15024,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(15024,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(15024,36,'cp -16','f8b4 f1d3 d5c4 d3c4 c8g4 e1g1 e8g8 c1g5 h7h6 g5h4 b4a5 a1c1 f8e8 h2h3 g4f3 d1f3 d8d4 h4f6 d4f6 f3f6 g7f6 c3d5 g8g7 c4a6 a8b8 a6b5 c6d4 b5e8 d4e2 g1h2 e2c1 f1c1 b8e8 c1c5 a5b6 d5b6 c7b6 c5c7 e8e2 c7b7 e2f2 b7a7 f2b2');
INSERT INTO evaluations VALUES(9116,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(9116,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(9116,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(9116,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(9116,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(9116,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(9116,36,'cp -5','f1d3 d5c4 d3c4 e8g8 e1g1 c8g4 c1g5 h7h6 g5h4 b4e7 d4d5 c6e5 c4e2 e5g6 h4g3 e7b4 h2h3 g4h5 c3b5 f6d5 b5c7 d5c7 d1d8 f8d8 g3c7 d8e8 e2d3 e8e7 a1c1 h5f3 g2f3 b4d2 c1c4 b7b5 c4c2 d2f4 c7a5 a7a6 d3g6 f7g6 f1d1');
INSERT INTO evaluations VALUES(15892,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(15892,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(15892,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(15892,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(15892,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(15892,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(15892,36,'cp 127','c3e2 d8e7 e2g3 f7f6 e5f6 g7f6 f1e2 d7b6 c1f4 e7g7 e1g1 f8d6 f4d6 c7d6 a2a4 a7a5 f1e1 e8g8 d1c1 f6f5 g3h5 g7f7 c1g5 f7g6 g5h4 e6e5 d4e5 d6e5 e2b5 e5e4 f3g5 c6e5 g1h1 g6h6 f2f3 e5g6 h5f6 g8g7 f6e8 f8e8 h4h6 g7h6 g5f7 h6g7 b5e8 b6c4 b2b3');
INSERT INTO evaluations VALUES(18423,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(18423,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(18423,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(18423,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(18423,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(18423,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(18423,36,'cp 213','c1f4 e7e6 b1c3 g8e7 e1c1 e7c6 d4d3 f8c5 d3g3 e8g8 h2h4 g8h8 c3a4 c5e7 f4c7 d8e8 a4c3 c8b7 c3b5 f7f6 h4h5 h7h6 c7d6 a8c8 g3a3 e8d8 f2f3 c8b8 c1b1 a7a6 d6e7 d8e7 a3e7 c6e7 b5c7 b8d8 f1a6 b7a6');
INSERT INTO evaluations VALUES(18577,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(18577,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(18577,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(18577,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(18577,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(18577,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(18577,36,'cp 25','g8f6 e4f6 d7f6 c2c3 d8c7 d1f3 c8g4 f3g3 c7g3 h2g3 e7e6 f2f3 g4f5 g1h3 f8d6 g3g4 f5g6 c1f4 e8d7 f4d6 d7d6 a2a4 f6d5 e1d2 a7a5 c4b3 h7h5 g4h5 h8h5 b3d5 e6d5 h3f4 h5h8 a1e1 g6h7 g2g4 g7g5 h1h6 h7g6 h6h8 a8h8');
INSERT INTO evaluations VALUES(12249,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(12249,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(12249,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(12249,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(12249,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(12249,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(12249,36,'cp 32','e4f6 d7f6 c2c3 d8c7 d1f3 c8g4 f3g3 c7g3 h2g3 e7e6 f2f3 g4f5 g3g4 f5g6 g1e2 e8c8 c1g5 f8d6 a2a4 h7h5 g4h5 g6h5 e1d2 d6e7 g5e3 h5g6 e2f4 g6h7 f4d3 h7d3 d2d3 f6d5 e3d2 e7d6 d3c2 a7a5 a1e1 c8c7 c4d3 d8g8 g2g4');
INSERT INTO evaluations VALUES(19755,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(19755,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19755,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(19755,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(19755,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19755,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19755,36,'cp 6','e7e6 a2a3 a7a5 c2c4 f8c5 d1a4 c8d7 c4d5 e6d5 f1b5 g8f6 b1c3 d8b6 a4b3 e8g8 e1g1 d7e6 c3a4 b6a7 b3d3 c5d6 c1d2 f6e4 a1c1 f8d8 f1d1 c6e5 f3e5 d6e5 d2e1 a8c8 d3e2 b7b6 c1c8 d8c8 h2h3 e5f6 b5d3');
INSERT INTO evaluations VALUES(20948,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20948,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(20948,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(20948,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(20948,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20948,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20948,36,'cp 0','c6a5 e2e3 g8f6 f1e2 e7e6 g1f3 f8e7 e1g1 e8g8 a2a4 c7c5 f3e5 f6e8 d2d3 f7f6 e5f3 e6e5 f3d2 b7b6 c2c4 d5d4 e3d4 c5d4 d2b3 a5b3 d1b3 e8c7 e2f3 a8b8 a4a5 c7e6 b2a3 e7a3 b3a3 c8b7 f3b7 b8b7 a5b6 a7b6');
INSERT INTO evaluations VALUES(21234,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21234,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21234,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21234,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21234,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21234,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21234,36,'cp 0','e2e3 e7e6 g1f3 g8f6 f1e2 f8e7 e1g1 e8g8 d2d3 c7c6 a2a4 a7a6 c2c4 d5c4 d3c4 d8d1 f1d1 a5b3 a1a2 a6b5 c4b5 c6b5 e2b5 f8d8 d1d8 e7d8 b1d2 b3c5 f3e5 f6d7 e5c4 d7b8 c4e5');
INSERT INTO evaluations VALUES(21233,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21233,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21233,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21233,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21233,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21233,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21233,36,'cp 0','e7e6 f1e2 g8f6 g1f3 c7c6 e1g1 c6b5 e2b5 c8d7 b2f6 g7f6 b5d7 d8d7 b1c3 a7a6 d2d3 e8c8 c3e2 c8b8 g1h1 h7h5 a1b1 b8a8 d1d2 d7c7 c2c3 f8d6 f1c1 h5h4 h2h3 d8c8 c1c2 d6c5 d2e1 f6f5 e1d2 c8g8 c2b2 c5a3 b2c2');
INSERT INTO evaluations VALUES(21138,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(21138,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(21138,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(21138,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(21138,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(21138,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(21138,36,'cp -20','d4d5 c6e5 f3e4 e7e6 d1d4 f8d6 b1c3 e6d5 c1f4 g8e7 e4d5 e8g8 e1c1 e7f5 d4d2 e5g6 f4d6 f5d6 d1e1 d8f6 g1f3 c8g4 f3g5 g4f5 c3e4 f5e4 g5e4 d6e4 e1e4 f8e8 e4e8 a8e8 g2g3 h7h5 f1b5 e8d8 h1d1 h5h4 d2c3 f6c3 b2c3 h4g3 h2g3');
INSERT INTO evaluations VALUES(22823,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(22823,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(22823,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(22823,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(22823,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(22823,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(22823,36,'cp 15','d2d4 e5d4 e2d4 g8f6 d4c6 b7c6 f1d3 d7d5 b1c3 f8b4 e4d5 e8g8 e1g1 c6d5 h2h3 c7c6 d1f3 f8e8 c1g5 h7h6 g5h4 b4d6 f1e1 c8e6 c3a4 d8a5 h4f6 a5a4 f6c3 a4f4 f3f4 d6f4 c3b4 a7a5 b4c5 f4g5 b2b3 g5f6 a1d1 f6c3 e1e3');
INSERT INTO evaluations VALUES(23496,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(23496,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(23496,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(23496,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(23496,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(23496,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(23496,36,'cp 134','e2e4 c7c6 g1f3 d7d6 h2h3 g7g6 f1d3 g8f6 b1c3 f8g7 e1g1 a6b4 d3e2 a7a5 f1e1 e8g8 c1f4 f8e8 a2a3 b4a6 d1d2 a6c7 a3a4 c8d7 a1d1 b7b5 e2d3 b5b4 c3e2 c7e6 f4h6 d6d5 e4e5 f6e4 d2e3 g7h6 e3h6 d8b6');
INSERT INTO evaluations VALUES(23497,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(23497,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(23497,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(23497,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(23497,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(23497,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(23497,36,'cp 117','c7c6 g1f3 d7d6 h2h3 g8f6 b1c3 a6c7 f1d3 g7g6 c1e3 f8g7 d1d2 e8g8 e3h6 c7e6 h6g7 g8g7 e1g1 b7b5 f1e1 c8b7 d2e3 a7a5 a1d1 b7a6 c3e2 b5b4 d3a6 a8a6 e2g3 d8c7 c2c4 b4c3 b2c3 a6a8 c3c4 d6d5 d1b1 d5c4 b1c1');
INSERT INTO evaluations VALUES(24162,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(24162,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(24162,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(24162,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(24162,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(24162,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(24162,36,'cp 43','h2h3 g4f3 d1f3 f6d7 b1d2 d8e7 d2c4 a7a5 c1d2 a5a4 f3g4 e8g8 b2b4 a4b3 a2b3 f8e8 a1a8 e8a8 b3b4 c5b6 d2h6 f7f6 h6e3 b6e3 c4e3 d7f8 e3f5 e7d7 f2f4 e5f4 g4f4 f8e6 f4g4 g8h8 f5e3 a8a2 g1h2 d7e7 e3f5 e7b4');
INSERT INTO evaluations VALUES(24161,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(24161,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(24161,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(24161,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(24161,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(24161,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(24161,36,'cp 45','g4f3 d1f3 f6d7 b1d2 a7a5 d2c4 d8e7 c1d2 a5a4 b2b4 a4b3 a2b3 e8g8 f3g4 f8e8 c4a5 a8b8 a1b1 d7f8 b3b4 c5d4 h3h4 g8h8 g1h1 f7f6 g2g3 h7h6 b1d1 d4b6 f2f4 e5f4 g3f4 f6f5 e4f5 e7e2 g4e2 e8e2');
INSERT INTO evaluations VALUES(24234,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(24234,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(24234,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24234,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(24234,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(24234,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(24234,36,'cp 29','d2d3 f8e7 e1g1 e8g8 c3d5 c6a5 d5e7 d8e7 f3h4 a5b3 a2b3 d7d5 h4f5 e7e6 f2f3 f6e8 c1e3 e8d6 e3c5 d5e4 f3e4 d6f5 f1f5 f8e8 b3b4 f7f6 d1e1 e6f7 h2h3 e8e6 f5f3 h7h6 e1h4 b7c6 f3g3 g8h7 h4g4 e6e8 g1h1 h7h8');
INSERT INTO evaluations VALUES(24235,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(24235,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(24235,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24235,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(24235,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24235,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(24235,36,'cp 29','f8e7 e1g1 e8g8 c3d5 c6a5 d5e7 d8e7 f3h4 a5b3 h4f5 e7e6 a2b3 d7d5 f2f3 f6e8 c1e3 e8d6 e3c5 d5e4 f3e4 d6f5 f1f5 f8e8 b3b4 f7f6 h2h3 e6c6 a1a5 a8d8 f5f1 c6e6 d1e2 d8d7 a5a3 h7h6 e2f3 g8h7 f1f2 b7c8 a3a1 h7h8 a1f1');
INSERT INTO evaluations VALUES(28477,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(28477,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(28477,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(28477,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(28477,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(28477,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(28477,36,'cp 148','f1e2 c7c6 e1g1 e8g8 d2d4 f8e8 c3b1 h7h6 c2c4 b8d7 c1e3 e6f5 e2d3 f5d3 d1d3 d7c5 d3c2 c5e4 a2a3 b4a5 b2b4 a5c7 b1d2 e4d2 e3d2 a7a5 f1e1 e8e1 a1e1 a5b4 d2b4 c7a5 h2h3 a5b4 a3b4 b7b5 d4d5 c6d5');
INSERT INTO evaluations VALUES(28712,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28712,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28712,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28712,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28712,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28712,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28712,36,'cp 0','g1f3 d7d5 c4d5 e6d5 f1b5 c7c6 b5d3 f8d6 e3e4 d5e4 c3e4 f6e4 d3e4 e8g8 e1g1 h7h6 e4c2 b8d7 d1d3 d7f6 f3e5 d6e5 d4e5 d8d3 c2d3 a8d8 d3c2 f6d5 c1d2 b7c8 a1c1 c6c5 a2a3 f8e8 f2f4 d5e7 d2c3 e7d5');
INSERT INTO evaluations VALUES(30153,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(30153,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(30153,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(30153,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(30153,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(30153,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(30153,36,'cp 39','c7c6 h2h3 g4f5 g2g4 f5g6 f3e5 b8d7 h3h4 d5c4 e5g6 h7g6 g4g5 f6d5 f1c4 f8b4 c1d2 d8e7 a2a3 b4a5 c4e2 d7b6 b2b4 d5c3 d2c3 b6d5 c3d2 a5c7 b4b5 e8g8 b5c6 b7c6 d1c2 a8b8 e2f3 c6c5 c2c5 c7d6 c5c2 d6a3 c2d3 a3b4 e1f1 b4d2 d3d2 d5b6 f1g2 b6d5 d2e2');
INSERT INTO evaluations VALUES(30458,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(30458,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(30458,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(30458,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(30458,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(30458,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(30458,36,'cp 173','d1f3 d8a5 c1d2 g8f6 c3d5 b5b4 d5f6 e7f6 f1c4 d7e6 c4e6 f7e6 g1e2 f8d6 e2f4 d6f4 f3f4 b8a6 a1c1 a8c8 f4e4 a5f5 e4f5 e6f5 c1c4 e8d7 e1e2 h8e8 h1c1 f5f4 h2h4 f4e3 f2e3 g7g6 d2b4 a6b4 c4b4 c8b8 c1c4 a7a5 b4b8 e8b8 c4c5 b8b2 e2f3 f6f5 c5a5');
INSERT INTO evaluations VALUES(33814,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(33814,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(33814,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(33814,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(33814,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(33814,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(33814,36,'cp 70','f7f5 f1e2 g7g6 f3d2 b8d7 d2e4 f5e4 c2c4 f8g7 b1c3 e8g8 e1g1 d7f6 d1d2 e7f7 a1e1 c8f5 e2d1 a8e8 h2h3 a7a6 e3g5 h7h5 d1b3 c7c6 g5f4 d6d5 c4d5 c6d5 f4d6 g8h7 d6f8 g7f8 d2f4 f8h6 f4g3 f7d7');
INSERT INTO evaluations VALUES(35226,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(35226,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(35226,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(35226,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(35226,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(35226,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(35226,36,'cp 53','f3d4 b8c6 d4c6 d7c6 b1d2 f7f6 a2a4 c8d7 d2b3 g8h6 a4a5 e8c8 c1e3 c8b8 f2f3 h6f7 h2h4 f8d6 h4h5 d6g3 e3f2 g3d6 e1c1 b7b6 f2e3 h8e8 f1e2 c6c5 b3d2 d7e6');
INSERT INTO evaluations VALUES(35217,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(35217,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(35217,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(35217,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(35217,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(35217,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(35217,36,'cp 41','b8c6 d4c6 d7c6 b1d2 f7f6 d2b3 f8b4 c1d2 b4d2 e1d2 b7b6 b3d4 c8d7 f1a6 a8d8 d2e3 g8h6 f2f3 c6c5 d4b5 c7c6 b5c3 e8e7 h1d1 h6f7 c3e2 d7e6 b2b3 d8d1 a1d1 h8d8 d1d8 e7d8 e2f4 d8e7 f4e6');
INSERT INTO evaluations VALUES(39587,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(39587,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(39587,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(39587,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(39587,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(39587,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(39587,36,'cp -31','c7c5 c2c3 d7d5 e2e3 g8f6 f1e2 e8g8 g1f3 f6d7 e1g1 b8c6 c3c4 d5c4 b3c4 c5d4 f3d4 d8b6 b2c3 d7c5 d4c6 b7c6 c3g7 g8g7 b1d2 c8f5 d1c1 f8b8 c1c3 f7f6 e2f3 b8d8 d2b3 d8d3 c3a5 c5b3 a2b3 d3b3 a5a4');
INSERT INTO evaluations VALUES(40340,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40340,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(40340,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(40340,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(40340,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40340,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(40340,36,'cp 0','d2d3 b8c6 e1g1 e8g8 a2a3 a8b8 h2h3 e7e6 c4a2 b7b5 c3e2 a7a5 e2g3 b5b4 a3b4 c5b4 c1e3 a5a4 d1d2 c8d7 f4f5 e6f5 e4f5 d6d5 g1h1 b8c8 e3h6 d8c7 h6f4 c7b6 f4h6');
INSERT INTO evaluations VALUES(40341,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40341,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40341,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40341,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40341,36,'cp 11','b8c6 e1g1 e8g8 a2a3 a8b8 h2h3 e7e6 c4a2 b7b5 c3e2 a7a5 e2g3 b5b4 a3b4 c5b4 c2c3 c8a6 f1f2 d8b6 a2b1 c6e7 b1c2 b4c3 b2c3 f6d7 d3d4 f8c8 c1d2 f7f5 e4f5 e7f5 a1b1 b6c7 g3f5 b8b1 d1b1 g6f5');
INSERT INTO evaluations VALUES(8264,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(8264,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(8264,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(8264,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8264,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8264,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8264,36,'cp 0','e1g1 e8g8 d1e1 c8g4 f4f5 g4f3 f1f3 c6d4 f3h3 e7e6 e1h4 f6h5 h4d8 f8d8 c1g5 d8d7 a1e1 g7e5 f5e6 d4e6 g5c1 e6d4 e1f1 d4c2 c4b5 d7d8 b5c4');
INSERT INTO evaluations VALUES(8263,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8263,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8263,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8263,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8263,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8263,36,'cp 0','e8g8 a2a3 a8b8 h2h3 e7e6 c4a2 b7b5 c1e3 a7a5 d1d2 d8c7 d2f2 b5b4 a3b4 a5b4 c3e2 b8a8 f2h4 a8a5 a2b3 a5a1 f1a1 c6a5 b3a4 c8b7 e2g3 f8a8 a1f1 a5c6 a4b3 c6a5');
INSERT INTO evaluations VALUES(3182,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3182,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3182,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3182,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3182,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3182,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3182,36,'cp 0','a2a3 a8b8 c4a2 b7b5 h2h3 e7e6 c1e3 a7a5 d1d2 d8c7 d2f2 b5b4 a3b4 a5b4 c3e2 b8a8 f2h4 a8a5 a2b3 a5a1 f1a1 c6a5 b3a4 c8b7 e2g3 f8a8 a1f1 a5c6 a4b3 c6a5');
INSERT INTO evaluations VALUES(40741,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40741,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40741,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40741,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40741,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40741,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40741,36,'cp 0','c7c6 f1e2 e8g8 e1g1 c6d5 e4d5 d8a5 c1e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(40756,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(40756,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(40756,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(40756,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(40756,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(40756,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(40756,36,'cp -30','h7h6 g5f6 e7f6 c3d5 c8e6 g2g3 e6d5 e4d5 g7g6 h2h4 f6g7 f1h3 e8g8 f3d2 f7f5 g3g4 b8d7 g4f5 g6f5 d2f1 e5e4 f1g3 e4d3 c2d3 d8h4 h3f5 h4f4 d1g4 f4g4 f5g4 d7e5 g4f5 e5f3 e1f1 g7b2 a1d1 b2g7 f5e6 g8h8 g3f5 f3g5 f5h4 g5e6 d5e6');
INSERT INTO evaluations VALUES(45978,30,'cp -153',NULL);
INSERT INTO evaluations VALUES(45978,31,'cp -123',NULL);
INSERT INTO evaluations VALUES(45978,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(45978,33,'cp -147',NULL);
INSERT INTO evaluations VALUES(45978,34,'cp -145',NULL);
INSERT INTO evaluations VALUES(45978,35,'cp -138',NULL);
INSERT INTO evaluations VALUES(45978,36,'cp -148','h7h6 g5f3 f6e4 d4d5 e6e5 c4d3 e4f6 b1c3 b8a6 d3a6 b7a6 f3d2 f8e7 d2c4 e8g8 f2f4 c8g4 d1d3 e5f4 c1f4 g4h5 e1g1 h5g6 d3e2 d8d7 h2h3 a8e8 e2d2 f6h5 f4h2 f7f5 a1e1 f5f4 h2f4 h5f4 f1f4 e7g5 f4f8 e8f8 d2e2 f8e8 e2f2 e8e1 f2e1 g6c2');
INSERT INTO evaluations VALUES(46478,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(46478,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(46478,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(46478,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(46478,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(46478,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(46478,36,'cp 178','c4b5 b8d7 e1g1 a7a6 e5d7 c8d7 b5d7 d8d7 d2d3 f8e7 f1e1 e8g8 b1d2 f8e8 d2f3 d5d4 c1f4 f6h5 f4e5 e7f6 d1d2 a8c8 e1e2 h7h6 a1e1 d7a4 a2a3 f6e5 f3e5 h5f6 e5f3 e8e2 e1e2 a4b5 d2b4 b5b4 a3b4 f6d5 f3d4 d5b4 e2d2 c8d8');
INSERT INTO evaluations VALUES(46382,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(46382,31,'cp 157',NULL);
INSERT INTO evaluations VALUES(46382,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(46382,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(46382,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(46382,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(46382,36,'cp 181','b8d7 e1g1 a7a6 f1e1 f8e7 b5a4 e8g8 d2d4 d7e5 d4e5 f6e4 c1e3 d8c7 e3d4 f8d8 a4b3 e4c5 b1c3 c5b3 c2b3 c8f5 d1f3 c7d7 h2h3 d7e6 e1e2 a8c8 a1d1 e7b4 a2a3 b4c3 d4c3 c8c6 f3g3 d5d4 d1d4 d8d4 c3d4 f5c2 d4e3 c2b3 e2d2');
INSERT INTO evaluations VALUES(49439,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(49439,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49439,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(49439,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49439,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49439,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49439,36,'cp 11','g1f3 d7d5 c1f4 f8d6 e2e3 e8g8 f4d6 d8d6 c2c4 c7c5 c4d5 f6d5 f1c4 c5d4 d1d4 b8c6 d4d2 f8d8 b1c3 d5c3 d2c3 b7b6 c4e2 c8b7 e1g1 d6e7 f1c1 a8c8 c1d1 d8d1 e2d1 g7g6 d1e2 h7h6');
INSERT INTO evaluations VALUES(50176,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(50176,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(50176,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(50176,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50176,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50176,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50176,36,'cp 0','c7c5 g1f3 f8g7 g2g3 b8c6 f1g2 e8g8 e1g1 d7d5 c4d5 f6d5 c3d5 d8d5 c1e3 g7b2 f3d4 d5d7 d4c6 b2a1 d1a1 b7c6 f1c1 a8b8 e3c5 d7e6 c5e3 e6f6 e3d4 f6e6');
INSERT INTO evaluations VALUES(51545,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(51545,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(51545,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(51545,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(51545,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(51545,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(51545,36,'cp 57','e1g1 c8b7 f1e1 d8c7 b1c3 b8c6 c4f1 c7b8 g2g3 g8e7 d2d3 e7g6 h2h4 f8e7 f3h2 f7f6 c3e2 g6e5 c2c3 c5c4 d3c4 e5c4 e2d4 c4d6 f1g2 b8c7 a2a4 e8g8 d4b5 d6b5 a4b5 c6e5 h2g4 a7a5 b5a6 a8a6');
INSERT INTO evaluations VALUES(52678,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(52678,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(52678,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(52678,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(52678,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(52678,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(52678,36,'cp 98','d8a5 e4e5 c8f5 e5d6 e7e6 c1f4 g8f6 f3e5 a5d8 g2g4 f5h7 f1d3 h7d3 d1d3 f8d6 e1c1 b8d7 f4g3 d6b4 h2h4 d7e5 g3e5 b4c3 b2c3 f6g4 e5g7 h8g8 d3h7 e8e7 h1g1 h6h5 f2f3 g4e3 d1e1 e3c4 g7h6 g8g1 e1g1 d8b6 h6g5 e7d6 h7f7 b6b2 c1d1 b2a1 g5c1 a1c3');
INSERT INTO evaluations VALUES(52996,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(52996,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(52996,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(52996,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(52996,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(52996,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(52996,36,'cp -58','g7g6 g1f3 b8c6 f1b5 c8d7 d1e2 g8h6 e5d6 h6f5 e1g1 e7e6 b1a3 f8d6 f3e5 c6e5 f4e5 d6e7 e2d3 h7h5 d3d7 d8d7 b5d7 e8d7 d2d3 b7b6 c1d2 h5h4 a3c4 b6b5 c4e3 c5c4 d3c4 e7c5 g1f2 b5c4 a1d1 g6g5');
INSERT INTO evaluations VALUES(332,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(332,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(332,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(332,33,'cp 257',NULL);
INSERT INTO evaluations VALUES(332,34,'cp 284',NULL);
INSERT INTO evaluations VALUES(332,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(332,36,'cp 262','a3a5 e2e4 g8e7 h2h4 b7b6 f1b5 a7a6 a2a3 a6b5 a3b4 a5a1 b3b1 a1a6 e1g1 e8g8 d2e3 f8d8 d1e2 h7h6 f1d1 e7g6 e2b5 a6b5 c3b5 c6e5 f3e5 g6e5 b5c7 a8a2 e3b6 e5c4 b6d4 a2c2 f2f3 d7d6');
INSERT INTO evaluations VALUES(338,30,'cp 304',NULL);
INSERT INTO evaluations VALUES(338,31,'cp 304',NULL);
INSERT INTO evaluations VALUES(338,32,'cp 310',NULL);
INSERT INTO evaluations VALUES(338,33,'cp 301',NULL);
INSERT INTO evaluations VALUES(338,34,'cp 300',NULL);
INSERT INTO evaluations VALUES(338,35,'cp 316',NULL);
INSERT INTO evaluations VALUES(338,36,'cp 323','e2e4 g8e7 h2h4 a7a6 h4h5 h7h6 f1e2 b4c3 d2c3 a5a4 h1h4 c6d8 d1c1 d8e6 c1b2 g7g5 h4g4 h8g8 e1f1 a4c6 f3d4 e6d4 c3d4 c6e6 g4g3 e7c6 g3f3 e6g4 b3e3 g4h5 f1g1 g5g4 f3f4 b7b5 e3g3 c8b7 f4g4 e8c8 g4g8');
INSERT INTO evaluations VALUES(5551,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(5551,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(5551,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(5551,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(5551,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(5551,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(5551,36,'cp 46','d2d3 g8f6 b1c3 c8g4 c1e3 d8d6 h2h3 g4f3 d1f3 e8g8 a1e1 a8d8 c3b1 f6e8 b1d2 b7b6 f3g4 d6d7 b2b3 g7c3 g4d7 d8d7 f2f4 f7f6 e1d1 e8c7 g2g4 c7b5 g1f2 b5d4 d2b1 c3b2 e3d4 b2d4 f2e2');
INSERT INTO evaluations VALUES(5549,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(5549,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(5549,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(5549,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(5549,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(5549,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(5549,36,'cp 40','g8f6 b1c3 c8g4 h2h3 g4f3 d1f3 e8g8 c1e3 d8d6 a1e1 b7b6 a2a4 a8d8 f3e2 e7e5 a4a5 f8e8 g1h1 d6e7 g2g3 b6b5 f2f3 f6d7 e2f2 a7a6 h3h4 e7d6 c3e2 d6e6 b2b3 h7h6 f2g2 g7f8 g3g4 g8h7');
INSERT INTO evaluations VALUES(10890,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(10890,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(10890,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10890,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(10890,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(10890,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(10890,36,'cp 32','a2a3 e8g8 g1e2 e7c5 e1g1 a7a5 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 d7d6 g1h1 c6e7 d5e7 d8e7 f2f4 c8g4 d1e1 g4e6 c4e6 f7e6 e1g3 e7d7 d3d4 e5d4 c3d4 c5b6 a1d1 f8f7 g3d3 a8f8 g2g3 a5a4 d4d5 e6d5 d3d5 d7h3 f4f5 h3g4 e2c3');
INSERT INTO evaluations VALUES(10908,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(10908,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(10908,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(10908,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(10908,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(10908,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(10908,36,'cp 23','d7d6 g1e2 e8g8 e1g1 c8e6 c3d5 c6b8 d5e7 d8e7 c4e6 f7e6 e2g3 c7c5 b2b4 b7b6 c2c3 b8d7 c1e3 d6d5 e4d5 e6d5 b4c5 d7c5 d3d4 e5d4 e3d4 c5e6 f1e1 e7d7 h2h3 a8e8 a1c1 e6d4 d1d4 d7f7 e1e5 e8e5 d4e5 f8e8 e5d4 f6e4');
INSERT INTO evaluations VALUES(8607,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(8607,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(8607,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(8607,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(8607,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(8607,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(8607,36,'cp 27','g1e2 c8e6 c3d5 c6b8 e1g1 e8g8 a3a4 c7c6 d5e7 d8e7 c4e6 f7e6 b2b4 b8d7 e2g3 f8f7 c1e3 d6d5 c2c3 d7f8 f2f3 f8g6 d1d2 f6d7 a4a5 h7h6 d3d4 a7a6 d2c2 g6f4 g3e2 e7g5 e2f4 e5f4 e3f2 e6e5 a1e1 e5d4');
INSERT INTO evaluations VALUES(12556,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(12556,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(12556,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(12556,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(12556,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(12556,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(12556,36,'cp 32','d2d3 d7d6 g1e2 f8e7 e1g1 c6d4 f2f4 c8d7 a2a4 e8g8 f4f5 f8e8 b2b3 a8c8 h2h3 h7h6 h3h4 b7b6 c1d2 a7a6 e2d4 c5d4 c3d5 f6d5 e4d5 e7f6 g2e4 a6a5 g1h2 c8c7 f1f2 d8e7 b3b4 a5b4 a1b1 e8c8 b1b4 c7c2 b4b6');
INSERT INTO evaluations VALUES(12560,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(12560,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(12560,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(12560,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(12560,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(12560,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(12560,36,'cp 23','f8e7 g1e2 e8g8 f2f4 d7d6 e1g1 c8d7 h2h3 c6d4 a2a4 d7c6 g1h2 f8e8 f1f2 h7h6 b2b3 b7b6 a1a2 a7a6 c3d5 c6d5 e4d5 d4e2 f2e2 e5f4 c1f4 e7f8 e2e8 f6e8 c2c4 g7g6 d1d2 g6g5 f4e3 f8g7 d3d4 g7d4 e3d4');
INSERT INTO evaluations VALUES(13590,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(13590,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13590,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13590,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13590,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13590,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13590,36,'cp 0','g1f3 b7b6 e1g1 c8b7 d1e2 f8e7 b2b3 e8g8 c1b2 a7a5 d3b5 c6a7 b5d3');
INSERT INTO evaluations VALUES(13786,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(13786,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(13786,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(13786,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(13786,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(13786,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(13786,36,'cp -41','f8e7 e2e4 e8g8 e4d5 f6d5 c2c3 d8c7 d2c4 d5b6 d1e2 b6c4 d3c4 c8d7 c1e3 a8d8 f3g5 c7c8 g2d5 b7b6 e2h5 e7g5 e3g5 d8e8 a1e1 d7e6 h5f3 c6a5 b2b3 f7f6 g5c1 e6d5 c4d5 a5b7 c3c4 b7d6 e1d1 b6b5 f3e2 c8d7 c4b5 d6b5');
INSERT INTO evaluations VALUES(8125,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(8125,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(8125,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(8125,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(8125,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(8125,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(8125,36,'cp -26','e2e4 e8g8 e4d5 f6d5 c2c3 d8c7 f1e1 f8d8 d2c4 f7f6 f3h4 c8e6 d1e2 b7b5 c4e3 c7d7 h4f5 e6f5 e3d5 e7f8 c1e3 a8b8 b2b3 a7a5 f2f4 a5a4 f4e5 c6e5 d3d4 c5d4 e3d4 f5e6 d5f4 e6f7 a1d1 d7c7 e2f2 a4a3 d4e5 f6e5 f4d5 f7d5 d1d5 d8d5');
INSERT INTO evaluations VALUES(8124,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(8124,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(8124,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(8124,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(8124,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(8124,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(8124,36,'cp -38','e8g8 e4d5 f6d5 c2c3 d8c7 f1e1 f8d8 a2a4 f7f6 f3h4 g7g5 h4f3 c8g4 a4a5 g8h8 d2c4 a8b8 d1e2 g4h5 g2h3 h5g6 f3d2 b7b5 a5b6 a7b6 d2e4 d5f4 g3f4 g5f4 g1h1 f6f5 e4c5 e7c5 c4e5 c6e5 e2e5 c7e5 e1e5 d8d3');
INSERT INTO evaluations VALUES(16020,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(16020,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(16020,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(16020,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(16020,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(16020,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(16020,36,'cp 187','e4e5 d6e5 d4e5 f6e4 f1c4 e7e6 d1e2 e4c5 e1g1 c7c6 a2a4 a7a5 f1d1 f8e7 b1c3 e8g8 h2h3 d8c7 d1d4 b7b5 a4b5 c8b7 d4g4 g8h8 g4g3 c6b5 c3b5 c7b8 c1f4 d7b6 c4d3 c5d3 e2d3 b7f3 g3f3 b6d5');
INSERT INTO evaluations VALUES(16120,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(16120,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(16120,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(16120,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(16120,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(16120,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(16120,36,'cp -25','d2d3 f8c5 e1g1 e8g8 c1e3 c5b6 e3b6 a7b6 b1c3 d7d6 h2h3 c6d4 b3b4 c7c6 f3d4 e5d4 c3e2 d6d5 e4d5 f6d5 b4b5 f8e8 f1e1 d5c7 b5c6 b7c6 e2g3 c8e6 c4b3 d8g5 b3e6 e8e6 e1e6 c7e6 a2a4 a8a5 d1f3 g5c5');
INSERT INTO evaluations VALUES(16119,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(16119,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(16119,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(16119,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(16119,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(16119,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(16119,36,'cp -26','f8c5 c1e3 c5b6 e3b6 a7b6 b1c3 d7d6 h2h3 e8g8 e1g1 c6d4 b3b4 c7c6 f3d4 e5d4 c3e2 d6d5 e4d5 f6d5 d1d2 c6c5 a2a3 d8d6 f1e1 c8d7 c2c3 d4c3 e2c3 d5f4 d2e3 a8d8 c3e4 d6c7 b4c5 b6c5 e3c5 c7c5 e4c5 d7c6 f2f3 b7b6 c5e4 f4d3 c4d3 d8d3');
INSERT INTO evaluations VALUES(9746,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(9746,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(9746,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(9746,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(9746,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(9746,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(9746,36,'cp -12','c1e3 c5b6 e1g1 e8g8 e3b6 a7b6 b1c3 d7d6 h2h3 c6d4 b3b4 c7c6 f3d4 e5d4 c3e2 d6d5 c4b3 d5e4 d3e4 f6e4 d1d4 d8d4 e2d4 f8d8 d4f3 c8f5 g2g4 f5e6 f1e1 e4d2 f3d2 d8d2 b4b5 e6b3 a2b3 a8a1 e1a1 d2c2 a1a8 g8h7 a8b8 c6b5 b8b7 c2c3 g1g2 f7f6 b7b6 c3b3');
INSERT INTO evaluations VALUES(27294,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(27294,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(27294,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(27294,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(27294,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(27294,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(27294,36,'cp 69','d2d4 g8f6 b1c3 a7a6 b5a4 e7e5 a4c6 d7c6 d1d3 e5d4 f3d4 c6d7 c1f4 f8e7 e1c1 e8g8 f2f3 b7b5 g2g4 f8e8 h2h4 b5b4 c3e2 c7c5 d4f5 d7f5 g4f5 d6d5 e4e5 c5c4 d3d4 d8a5 c1b1 e7c5 d4d2 c4c3 d2d3 f6h5 d3d5 a8d8 d5e4 d8d1 h1d1 e8d8 d1e1 c3b2 f4g5 d8e8 f5f6');
INSERT INTO evaluations VALUES(28542,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(28542,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(28542,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(28542,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28542,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(28542,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(28542,36,'cp 11','f8e7 e1g1 e8g8 d1d2 a6a5 f1d1 a5a4 b3c1 a4a3 b2a3 a8a3 f2f3 d6d5 c3d5 f6d5 e4d5 d8d5 d2d5 e6d5 d1d5 a3e3 a1b1 b8c6 g1f1 e3c3 d5d7 f8d8 d7d8 c6d8 e2d3 g7g6 b1b5 c3c5 b5c5 e7c5 c1b3 c5b6 a2a4 g8f8 a4a5 b6a7 d3c4 f7f5 f1e2 d8c6');
INSERT INTO evaluations VALUES(29000,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(29000,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(29000,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(29000,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29000,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(29000,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(29000,36,'cp 13','d8b6 d1b3 b8d7 g5f6 b6b3 a2b3 d7f6 g1f3 a7a6 c4c5 g7g6 b3b4 a8c8 c3e2 h7h5 e2c1 h5h4 f1d3 f6e4 f3g1 f5d7 f2f3 e4g5 c1e2 f8g7 f3f4 g5e4 e2c3 d7f5 d3e4 f5e4 c3e4 d5e4 e1d2 e7e6 g1e2 f7f5 h1c1 g7f6 d2c2');
INSERT INTO evaluations VALUES(33080,30,'cp 210',NULL);
INSERT INTO evaluations VALUES(33080,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(33080,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(33080,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(33080,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(33080,35,'cp 237',NULL);
INSERT INTO evaluations VALUES(33080,36,'cp 228','c3d5 f6d8 b2b4 c5b6 d2d4 e5d4 e1g1 g8f6 f1e1 e8g8 c1g5 c8e6 g5f6 g7f6 f3d4 b6d4 d1d4 b8d7 d5e3 c7c5 d4c3 c5b4 c3b4 a8c8 c4d5 d8c7 a1d1 b7b6 c2c4 g8h8 b4b2 f8g8 g1h1 c7d8 h2h3 d8e7 e3f5 e6f5 e4f5 e7f8 g2g4 c8e8 e1e8');
INSERT INTO evaluations VALUES(34276,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(34276,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(34276,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(34276,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(34276,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(34276,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(34276,36,'cp 42','b5e2 d7c7 d2d4 g8f6 g1f3 c8f5 f3h4 f5g6 g2g3 e7e6 e2f3 b8d7 h4g6 h7g6 e1g1 f8d6 c1g5 e8g8 d1d3 e6e5 a1d1 e5d4 d3d4 d6e5 d4c4 a8e8 f3g2 e8e6 a2a4 e5d6 g2h3 e6e8 g1g2 d6e5 h3d7 f6d7 f1e1 d7f6 g5f6');
INSERT INTO evaluations VALUES(36237,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(36237,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36237,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36237,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36237,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(36237,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(36237,36,'cp 0','f1d3 h7h6 g5f4 c7c5 d4f3 b8c6 e1g1 c8e6 a2a4 f8e8 f1e1 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(36238,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(36238,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36238,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36238,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36238,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36238,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(36238,36,'cp 0','h7h6 g5e3 f8e8 e1g1 b8d7 d4f5 e7f8 f5g3 c7c6 a2a4 b7b6 f2f4 d7c5 e3d4 c5e6 d4e3');
INSERT INTO evaluations VALUES(38677,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(38677,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(38677,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(38677,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(38677,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(38677,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(38677,36,'cp -28','f7f6 e5d3 g8e7 e2e3 e8g8 f4d6 d8d6 f1e2 e6e5 e1g1 b8c6 c2c3 g8h8 b1d2 a7a5 d4e5 f6e5 c3c4 c8f5 c4d5 d6d5 d1b3 f8d8 b3d5 e7d5 a1d1 d5e3 f2e3 f5d3 e2d3 d8d3 d2c4 e5e4 d1d3 e4d3 f1f7 a8c8 f7d7');
INSERT INTO evaluations VALUES(41471,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(41471,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41471,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41471,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41471,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(41471,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41471,36,'cp 14','c2c3 c7c5 b1d2 d8b6 d1a4 b8c6 a4b5 c5c4 b5b6 a7b6 g2g4 b6b5 g1e2 e7e6 f1g2 c8d7 e1g1 e8g8 e3e4 f6e4 d2e4 d5e4 g2e4 b5b4 a2a3 b4c3 b2c3 c6a5 a1b1 d7c6 e4c6 b7c6 a3a4 f8d8 g1g2 a5b3 f1d1 d8d5');
INSERT INTO evaluations VALUES(45450,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45450,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(45450,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(45450,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(45450,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45450,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(45450,36,'cp 25','b8c6 g1f3 c6b4 e1g1 b4d3 d1d3 e7e6 c1g5 f8e7 f1e1 e8g8 f3e5 c8d7 b1d2 h7h6 g5h4 a8c8 a2a4 f6e8 h4e7 d8e7 c2c3 e8d6 g2g4 d7e8 d3g3 a7a5 h3h4 f7f6 e5d3 e8f7 h4h5 c8c7 e1e3 f6f5 a1e1 f5g4 g3g4 d6f5');
INSERT INTO evaluations VALUES(13166,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(13166,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(13166,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(13166,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(13166,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(13166,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(13166,36,'cp 10','g1f3 c6b4 e1g1 b4d3 d1d3 e7e6 c1g5 f8e7 f1e1 e8g8 f3e5 c8d7 b1d2 h7h6 g5h4 f6e8 h4e7 d8e7 b2b3 e8f6 c2c4 b7b6 a1c1 f8c8 e5d7 e7d7 c4c5 b6c5 d4c5 d7b7 d2f3 f6e4 d3e3 c8c7 f3e5 a7a5 c5c6');
INSERT INTO evaluations VALUES(45754,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45754,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(45754,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(45754,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(45754,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45754,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(45754,36,'cp 17','h7h6 g5h4 g7g6 e2e3 c8f5 f1d3 f5d3 c2d3 b8d7 g1e2 f8g7 f2f3 e8g8 e1g1 f8e8 a1e1 a8c8 h4f2 g7f8 e2f4 b7b5 h2h3 d7b6 d3d1 b5b4 c3a4 b6a4 d1a4 c6c5 d4c5 f8c5 f4d3 c5b6 d3b4 d5d4 e3d4 b6d4 e1e8 f6e8 b4d3 d4f2 d3f2 d8b6 b2b3');
INSERT INTO evaluations VALUES(53583,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53583,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(53583,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(53583,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(53583,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53583,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(53583,36,'cp 14','b8d7 b1d2 b7b6 f1d3 g8f6 g1f3 f8e7 e1g1 e8g8 b2b3 c6c5 c1b2 c8a6 f1e1 a8c8 d3f1 a6b7 c4d5 e6d5 f3e5 f8e8 a1c1 e7f8 f1d3 d8e7 d1e2 c8c7 h2h3 e8d8 f2f4 e7e6 g2g4 f6e4 d2e4 d5e4 d3c4');
INSERT INTO evaluations VALUES(54972,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54972,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54972,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(54972,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54972,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(54972,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(54972,36,'cp -13','g7g6 d2d4 f8g7 c2c3 c7c5 g1f3 g8f6 f1d3 c5d4 c3d4 b8c6 h2h3 d6d5 e4e5 f6e4 e1g1 e8g8 b1d2 c8f5 d1e2 d8b6 e2e3 f8d8 d2e4 d5e4 d3e4 f5e4 e3e4 c6d4 c1e3 b6c6 e4c6 d4c6 a1c1 a7a5 c1c5 a5a4 c5b5 d8d7 f1c1');
INSERT INTO evaluations VALUES(57340,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(57340,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(57340,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(57340,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(57340,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(57340,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(57340,36,'cp 56','a2a3 g8f6 d2d4 g7g6 g1f3 f8g7 b2b4 b7b6 e2e3 e8g8 f1e2 c8b7 e1g1 d7d6 c1b2 b8d7 d1b3 d8e7 a3a4 a7a5 c4c5 a5b4 c5d6 e7d6 c3b5 d6e7 b5c7 a8c8 f1c1 f6d5 c7d5 b7d5 e2c4 d7f6 h2h3 d5c4 c1c4 f6d5 h3h4 c8c4 b3c4 f8a8 c4b3');
INSERT INTO evaluations VALUES(1597,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(1597,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(1597,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(1597,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(1597,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(1597,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(1597,36,'cp 28','c8h3 f3e1 g8f6 e2f3 h3g4 f3c6 b7c6 f2f3 g4c8 d1e2 e7e6 e2c4 c8b7 d2d4 f8d6 e1d3 e8g8 f1f2 f8d8 d3c5 a8b8 b2b3 e6e5 d4e5 d6e5 c5b7 b8b7 f2e2 e5d6 g2g3 h7h5 g1g2 b7b4 c4d3 g6d3 c2d3 f6d5 c1d2 d5b6 a1d1 f7f6 c3e4');
INSERT INTO evaluations VALUES(27610,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(27610,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(27610,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(27610,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(27610,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(27610,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(27610,36,'cp 24','f3e1 g8f6 e2f3 h3g4 f3c6 b7c6 f2f3 g4c8 d1e2 e7e6 e2c4 c8b7 d2d4 f8d6 e1d3 e8g8 f1f2 f8d8 d3c5 b7c8 c3e4 f6e4 c5e4 e6e5 c4c6 c8d7 c6d5 d7e6 d5a5 e5d4 e4d6 c7d6 b2b3 d4d3 c2d3 g6d3 c1a3 d3e3 a5d2 e3b6 a3b2 f7f6 a1d1 e6f7 g1h1 d6d5 h2h3 a7a5');
INSERT INTO evaluations VALUES(3048,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(3048,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(3048,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(3048,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(3048,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(3048,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(3048,36,'cp 15','e1g1 e7e6 b2b3 d7d5 d2d4 c5d4 e3d4 c8d7 a1b1 b7b6 c1a3 c6e7 c4d5 f6d5 c3d5 e6d5 e2f4 a8c8 d1d2 f8e8 f1e1 h7h5 a3e7 e8e7 f4d5 e7e1 b1e1 d7e6 d5f4 d8d4 d2d4 g7d4 f4e6 f7e6 e1e6 c8c2 g2d5 g8f8 g1h1 c2f2 e6g6 h5h4 g3h4 f2a2 g6g5');
INSERT INTO evaluations VALUES(3047,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(3047,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(3047,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(3047,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(3047,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(3047,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(3047,36,'cp 17','e7e6 b2b3 d7d5 d2d4 c5d4 e3d4 c8d7 a1b1 b7b6 c1a3 c6e7 c4d5 e6d5 e2f4 a8c8 f1e1 f8e8 d1d2 e7f5 e1e8 d7e8 d2d3 e8c6 g2h3 f6e4 h3f5 e4c3 f5g6 h7g6 d3c3 c6d7 c3e3 d7f5 b1c1 c8c1 a3c1 d8d7 c1b2 d7b5 f2f3');
INSERT INTO evaluations VALUES(3936,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(3936,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(3936,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(3936,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(3936,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(3936,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(3936,36,'cp 62','c1e3 f6g4 e3g5 f7f6 g5h4 a7a5 f3d2 g4h6 f2f3 d7c5 d1c2 c8d7 e1c1 a5a4 a2a3 h6f7 h4f2 g7h6 c1b1 h6d2 c2d2 c5b3 d2d3 c7c5 g2g4 b3d4 h2h4 d8a5 c3b5 a8c8 f2e1 a5b6 d1c1 d4b5 c4b5 c5c4 d3d1 d7b5');
INSERT INTO evaluations VALUES(3929,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(3929,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(3929,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(3929,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(3929,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(3929,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(3929,36,'cp 64','f6g4 e3g5 f7f6 g5h4 a7a5 f3d2 g4h6 f2f3 d7c5 a2a3 a5a4 d1c2 h6f7 h4f2 g7h6 f2c5 d6c5 e1c1 c8d7 c2d3 f6f5 h2h4 f7d6 c1b1 f5f4 h4h5 g6g5 b1a2 g8h8 c3b5 f8g8 d3c3 h6g7 c3c2 b7b6 d2b1 h7h6 b1c3 g7f6 c3a4 g5g4 f3g4 d7g4 e2g4 g8g4 a4c3');
INSERT INTO evaluations VALUES(7781,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(7781,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(7781,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(7781,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(7781,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(7781,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(7781,36,'cp 38','b2b3 b7b6 c1b2 e8g8 e1g1 c8b7 d1c2 h7h6 a1d1 d8c7 c2e2 a8d8 d3b1 f8e8 e2c2 e7d6 f1e1 d7f8 h2h3 f8g6 c2e2 c7e7 b1g6 f7g6 e3e4 f6h5 b2c1 d6b4 c1d2 g6g5 e2d3 b4c3 d3c3 d5e4 e1e4 c6c5');
INSERT INTO evaluations VALUES(10580,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(10580,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(10580,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(10580,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(10580,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(10580,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(10580,36,'cp 134','c3d4 c5b6 d4d5 c6e7 d1e2 e4f6 d5d6 c7d6 c1g5 e8g8 b1c3 e7g6 c3d5 d8e8 e2e8 f6e8 a1c1 f7f6 d5b6 a7b6 b5c4 g8h8 g5e3 b6b5 c4b5 g6e7 a2a4 e7d5 f1d1 d5e3 f2e3 f6f5 d1d4 g7g6 b2b3 h8g8 f3d2 d6d5 d4d5 d7d6 h2h4 c8e6 d5d4 d6d5 b5e8 a8e8 a4a5 f8f7 d2f3 g8g7');
INSERT INTO evaluations VALUES(10579,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(10579,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(10579,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(10579,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(10579,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(10579,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(10579,36,'cp 140','c5b6 d4d5 c6e7 d1e2 e4f6 d5d6 c7d6 b1c3 e8g8 c1g5 d6d5 a1e1 d5d4 c3e4 e7g6 e4f6 g7f6 g5h6 a7a6 b5a4 d7d5 e2d2 c8g4 f3d4 d8d6 h6f8 a8f8 a4d1 g4c8 d1c2 d6f4 d2f4 g6f4 d4e2 f4e2 e1e2 f8d8 c2b3 g8f8 e2c2 c8e6 f1d1 b6a5 g1f1 a5b6 c2d2');
INSERT INTO evaluations VALUES(11472,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(11472,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11472,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(11472,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11472,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(11472,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11472,36,'cp 0','f8b4 f3d2 b4c3 b2c3 d7d5 c4b3 d5e4 d2e4 f6e4 d3e4 d8e7 b3d5 c8e6 e1g1 e8g8 c3c4 c6d8 d1h5 f7f6 c4c5 c7c6 d5b3 e6b3 a2b3 d8e6 b3b4 f8d8 f1d1 d8d1 a1d1 a6a5 b4a5 a8a5 h5g4 a5a2 e3h6 a2c2 h6e3 e6d4 g4c8 g8h7 e3d4 e5d4 c8h3 h7g6 h3g3 g6h7');
INSERT INTO evaluations VALUES(7276,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7276,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7276,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7276,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(7276,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7276,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7276,36,'cp -1','f3d2 b4c3 b2c3 d7d5 c4b3 d5e4 d2e4 f6e4 d3e4 e8g8 d1d8 f8d8 e1e2 c6a5 h1d1 d8d1 a1d1 a5b3 c2b3 c8g4 f2f3 g4e6 c3c4 f7f6 e2d3 a8d8 d3c2 d8d1 c2d1 g8f7 d1c2 f7e7 e3f2 g7g5 f2c5 e7f7 c2c3 b7b6 c5f2');
INSERT INTO evaluations VALUES(15316,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(15316,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(15316,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(15316,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(15316,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15316,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15316,36,'cp 0','f7f6 e1g1 d5b6 c4b5 f8d6 a2a4 a7a5 c1e3 e8g8 e3b6 c7b6 b1d2 g8h8 c2c3 c8g4 d2e4 d6c5 e4c5 b6c5 e2e3 c6e7 d3d4 c5d4 c3d4 d8d5 d4e5 g4f3 e3f3 d5f3 g2f3 f6e5 a1d1 f8d8 d1d8 a8d8 g1g2 g7g6 f1c1 e7d5 c1e1');
INSERT INTO evaluations VALUES(16461,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16461,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16461,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(16461,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16461,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(16461,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16461,36,'cp 8','b1c3 f8b4 d1c2 b4c3 c2c3 d8e7 f1e2 e8g8 d2d3 d7d5 e1g1 e5e4 d3e4 f6e4 c3b3 d5c4 b3c4 c8f5 c4b5 e4d6 b5c5 f8d8 c1d2 e7e6 d2c3 d6e4 c5a3 h7h6 f1d1 d8d1 e2d1 a8d8 c3e1 b7b6 d1b3 e6g6 a1d1 d8d1 b3d1 e4c5 a3c3 f5e4');
INSERT INTO evaluations VALUES(17033,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(17033,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(17033,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17033,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(17033,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(17033,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(17033,36,'cp 36','f8b4 a2a3 b4c3 d4c3 d7d5 c3e5 c8e6 c4d5 d8d5 e5d5 f6d5 e2e4 d5b6 f2f3 e8c8 c1f4 f7f5 e4f5 e6f5 e1f2 f5d3 a1e1 d3f1 h1f1 h8f8 f4g5 d8e8 e1e8 f8e8 f1d1 b6c4 g5c1 e8d8 d1e1 g7g6 b2b3 c4a5 c1g5 d8d5 h2h4 c7c5 b3b4 a5c6 b4c5 d5c5');
INSERT INTO evaluations VALUES(17437,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(17437,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(17437,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(17437,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(17437,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(17437,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(17437,36,'cp 34','d8c7 c1e3 b7b5 d4c6 c7c6 f1e2 c8b7 e1g1 f8d6 d1d3 g8f6 a1d1 d6e5 e3d4 e5d4 d3d4 e8g8 e2f3 c6c7 d1d3 e6e5 d4e3 a8c8 f1d1 c7c5 f3e2 c5e3 d3e3 g7g6 h2h4 g8g7 d1d6 c8c6 d6c6 b7c6 f2f3 f8c8 g1f2');
INSERT INTO evaluations VALUES(1269,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(1269,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(1269,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(1269,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(1269,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(1269,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(1269,36,'cp 34','c1e3 b7b5 d4c6 c7c6 f1e2 c8b7 e1g1 f8d6 d1d4 g8f6 f1d1 d6b8 a3a4 c6c7 e4e5 b5a4 e3f4 e8g8 a1a4 f6d5 a4c4 c7d8 f4g3 f7f6 c3e4 b8e5 g3e5 f6e5 d4e5 d5f4 e2f3 b7d5 c4c5 a8c8 c5d5 f4d5 e4d6 d8f6 e5f6 d5f6 d6c8 f8c8');
INSERT INTO evaluations VALUES(1268,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(1268,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(1268,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(1268,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(1268,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(1268,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(1268,36,'cp 32','b7b5 d4c6 d7c6 d1g4 e6e5 g4g3 h7h5 e3d4 f7f6 f1e2 e8f7 d4e3 c8e6 h2h4 f8d6 a3a4 g8e7 a4b5 c6b5 e1g1 c7c6 f1d1 d6b4 c3a2 b4d6 d1d2 g7g6 a2c3 d6c7 e2d3 h8d8 a1d1 c7a5 f2f4 e5f4 e3f4 a8c8 g1h2 a5c3');
INSERT INTO evaluations VALUES(18775,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(18775,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(18775,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(18775,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18775,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(18775,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(18775,36,'cp 6','g8f6 d2d4 d7d5 a2a3 d5c4 f1c4 a7a6 e1g1 b7b5 c4a2 c8b7 d1e2 f8e7 f1d1 d8b8 d4d5 e6d5 c3d5 f6d5 a2d5 e8g8 e3e4 b8c7 c1d2 a8d8 a3a4 b5a4 h2h3 c6d4 f3d4 c5d4 a1a4 b7d5 e4d5 d8d5 e2a6 d4d3 g1h1');
INSERT INTO evaluations VALUES(18798,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(18798,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(18798,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(18798,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(18798,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(18798,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(18798,36,'cp 15','c2c4 d7d5 c4d5 e6d5 f1e2 c5d4 f3d4 g8f6 e1g1 f8d6 b2b3 e8g8 c1a3 f6e4 d4c6 b7c6 a3d6 d8d6 e2d3 d6f6 b1d2 e4c5 d2f3 c8b7 d3e2 c5e6 f3d4 c6c5 d4e6 f6e6 a1c1 e6b6 d1d2 f8d8 f1d1 g7g6 h2h3 d5d4 e3d4 c5d4 c1c4');
INSERT INTO evaluations VALUES(20858,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(20858,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(20858,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(20858,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(20858,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(20858,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(20858,36,'cp 105','c2c4 e7e5 d4e5 d5c4 d1d8 e8d8 e5f6 g8f6 f1g2 f8b4 b1c3 h8e8 c1e3 c8e6 e1c1 d8c8 f3d4 c6d4 e3d4 a7a5 e2e4 a8a6 h1e1 b7b5 f2f4 b4c3 d4c3 b5b4 c3d4 e6g4 d1d2 g4e6 d4f6 g7f6 e4e5 c4c3 b2c3');
INSERT INTO evaluations VALUES(26323,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26323,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(26323,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(26323,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26323,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26323,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26323,36,'cp 11','g4f3 d1f3 e7e6 e1g1 a7a6 d2d3 g8f6 c4b3 d8c7 c3e2 c6e5 f3g3 c5c4 d3c4 f6e4 g3e3 e4f6 b3a4 e5d7 b2b3 f8e7 e3g3 e8g8 a4d7 c7d7 c1h6 f6e8 a2a4 a8c8 f1d1 f7f5 h6g5 e7g5 g3g5 e8f6 g5e3 f8d8 e2c3 d7f7 a4a5 d6d5 c3a4 d5c4 a4b6');
INSERT INTO evaluations VALUES(26333,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26333,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(26333,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26333,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(26333,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(26333,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(26333,36,'cp 20','d1f3 e7e6 d2d3 g7g6 c1f4 d8d7 c3e2 g8e7 c4b3 f8g7 c2c3 d6d5 a1d1 e8g8 h3h4 c6a5 h4h5 e6e5 f4g5 f7f6 g5d2 a5b3 a2b3 g6g5 h5h6 g7h8 e1g1 d5d4 c3d4 c5d4 d1c1 e7c6 f3f5 d7f5 e4f5');
INSERT INTO evaluations VALUES(26332,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(26332,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(26332,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26332,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(26332,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26332,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(26332,36,'cp 9','e7e6 d2d3 a7a6 c4b3 c6d4 f3g3 g7g6 c3e2 d4e2 e1e2 g8f6 c2c3 d8c7 h3h4 h7h5 c1f4 a8d8 f2f3 f8g7 d3d4 c5c4 b3c2 e8g8 h1d1 f8e8 e2f1 e6e5 f4e3 f6d7 a2a4 e5d4 e3d4 g7d4 c3d4 d8c8 a4a5 e8e6 f3f4 d7f6');
INSERT INTO evaluations VALUES(31513,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(31513,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(31513,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(31513,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(31513,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(31513,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(31513,36,'cp -50','g8h6 g1f3 c6c5 c1e3 d8b6 b1c3 b8c6 d3b5 f5g6 d4c5 f8c5 e3c5 b6c5 d1d4 c5e7 e1c1 e8g8 d4f2 e7b4 f3d4 c6d4 d1d4 b4c5 d4d2 c5f2 d2f2 d5d4 c3d1 a7a6 b5f1 a8c8 f2d2 c8c7 f1d3 g6d3 d2d3 f8c8 c2c3 d4c3 d1c3 h6g4 d3d2 g7g5 f4g5 g4e5 c1b1 e5c4 d2d3 c4a3 b2a3 c7c3 d3d7 b7b5 h1f1');
INSERT INTO evaluations VALUES(31685,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(31685,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(31685,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(31685,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(31685,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(31685,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(31685,36,'cp -32','b8c6 h2h3 g8f6 f1d3 f5e4 d3e4 d5e4 f3d2 f8e7 d2e4 f6e4 c3e4 d8d4 d1d4 c6d4 a1c1 a8d8 c2c3 d4c6 f4e3 f7f5 e4g5 d8d6 e1e2 e6e5 g5f3 e8f7 g2g3 h8d8 h1d1 d6d1 c1d1 d8d1 e2d1 h7h6 d1e2 f7e6 f3d2 h6h5 f2f3 e6d5 g3g4 f5f4');
INSERT INTO evaluations VALUES(26416,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(26416,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(26416,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(26416,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(26416,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(26416,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(26416,36,'cp -28','h2h3 g8f6 f1d3 f8d6 f4d6 d8d6 c3b5 d6d8 c2c3 e8g8 d3f5 e6f5 e1g1 f6e4 f1e1 f8e8 b5a3 g7g6 f3d2 g8g7 d2e4 f5e4 f2f3 f7f5 d1e2 d8d7 e2f2 e4f3 a3c2 e8e2 e1e2 f3e2 f2e2 a8e8 e2b5');
INSERT INTO evaluations VALUES(34171,30,'cp 435',NULL);
INSERT INTO evaluations VALUES(34171,31,'cp 440',NULL);
INSERT INTO evaluations VALUES(34171,32,'cp 462',NULL);
INSERT INTO evaluations VALUES(34171,33,'cp 467',NULL);
INSERT INTO evaluations VALUES(34171,34,'cp 438',NULL);
INSERT INTO evaluations VALUES(34171,35,'cp 455',NULL);
INSERT INTO evaluations VALUES(34171,36,'cp 467','c3b5 b8a6 g1f3 e5f6 d2d4 g7g6 f3e5 c8b7 e1g1 f8g7 c2c3 g8h6 b5c7 a6c7 d1a4 b7c6 e5c6 e8g8 e2f3 a7a6 c6e5 a8d8 e5d7 f6f3 g2f3 b6b5 a4a5 d8d7 c1f4 c7d5 f4g3 f8a8 a2a4 b5a4 a5a4 d7d8 c3c4 d5f6 f1e1 h6f5 d4d5 d8c8 g3f4 g8h8');
INSERT INTO evaluations VALUES(37790,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37790,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(37790,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(37790,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37790,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(37790,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(37790,36,'cp 0','c5d4 f3d4 b8c6 e2e3 g8f6 d1b3 e8g8 f1e2 b7b6 e1g1 c8b7 d4c6 b7c6 b3c2 d8c7 e3e4 a7a6 a2a4 c7e5 f2f3 f6h5 c1e3 h5f4 c2d2 f4e2 d2e2 f7f5 e4f5 f8f5 a1d1 a8f8 d1d2 f5h5 h2h3 e5g3 e3f2 g3f4 f2e3');
INSERT INTO evaluations VALUES(37857,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37857,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37857,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37857,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37857,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37857,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37857,36,'cp 0','f3d4 b8c6 e2e3 g8f6 d1b3 b7b6 f1e2 e8g8 e1g1 c8b7 d4c6 b7c6 b3c2 d8b8 e3e4 a7a6 a2a4 b8e5 f2f3 f6h5 c1e3 h5f4 c2d2 f4e2 d2e2 f7f5 e4f5 f8f5 a1d1 a8f8 f1e1 f5h5 h2h3 e5g3 e3f2 g3f4 f2e3');
INSERT INTO evaluations VALUES(37852,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37852,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37852,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37852,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37852,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37852,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37852,36,'cp 0','b8c6 e2e3 g8f6 d1b3 b7b6 f1e2 e8g8 e1g1 c8b7 d4c6 b7c6 b3c2 d8b8 e3e4 a7a6 a2a4 b8e5 f2f3 f6h5 c1e3 h5f4 c2d2 f4e2 d2e2 f7f5 e4f5 f8f5 a1d1 a8f8 d1d2 f5h5 h2h3 e5g3 e3f2 g3f4 f2e3');
INSERT INTO evaluations VALUES(38324,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(38324,31,'cp 251',NULL);
INSERT INTO evaluations VALUES(38324,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(38324,33,'cp 264',NULL);
INSERT INTO evaluations VALUES(38324,34,'cp 252',NULL);
INSERT INTO evaluations VALUES(38324,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(38324,36,'cp 268','d1d8 e7d8 f3e5 c8e6 b1c3 g8f6 e5f3 b8d7 f3d4 c7c6 f2f3 d8a5 c1f4 d7b6 d4e6 f7e6 f4d2 e8c8 e1c1 b6d7 h2h4 b7b5 h1h3 a5c7 g2g4 d7e5 h4h5 g7g6 a2a4 a7a6 d2e3 c8b7 d1d8 h8d8 h5g6 h7g6');
INSERT INTO evaluations VALUES(35393,30,'cp 248',NULL);
INSERT INTO evaluations VALUES(35393,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(35393,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(35393,33,'cp 262',NULL);
INSERT INTO evaluations VALUES(35393,34,'cp 273',NULL);
INSERT INTO evaluations VALUES(35393,35,'cp 249',NULL);
INSERT INTO evaluations VALUES(35393,36,'cp 259','e7d8 f3e5 g8f6 f2f3 f6d7 e5d3 b8c6 c1f4 c6d4 b1a3 a7a6 c2c3 d4e6 f4g3 e8g8 a3c2 b7b5 h2h4 c7c6 e1c1 a6a5 c2e3 h7h5 e3f5 d8f6 e4e5 f6d8 f3f4 e6c5 f5d6 c5d3 f1d3 d7c5 d3c2 c8e6 f4f5 e6a2');
INSERT INTO evaluations VALUES(35415,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(35415,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(35415,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(35415,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(35415,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(35415,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(35415,36,'cp 255','f3e5 g8f6 b1c3 c8e6 e5f3 c7c6 f3d4 d8b6 d4e6 f7e6 f2f3 b8d7 c1d2 e8c8 a2a3 b6c7 e1c1 b7b5 f1e2 a7a5 c3a2 d7e5 c1b1 d8d7 a2c1 e5c4 e2c4 b5c4 c1e2 a5a4 e2c3 h8d8 b1c1');
INSERT INTO evaluations VALUES(35414,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(35414,31,'cp 263',NULL);
INSERT INTO evaluations VALUES(35414,32,'cp 245',NULL);
INSERT INTO evaluations VALUES(35414,33,'cp 259',NULL);
INSERT INTO evaluations VALUES(35414,34,'cp 247',NULL);
INSERT INTO evaluations VALUES(35414,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(35414,36,'cp 242','g8f6 f2f3 f6d7 e5d3 b8c6 c1f4 c6d4 b1a3 a7a6 f4e3 d4e6 g2g3 b7b5 e1c1 e8g8 c2c3 d7b6 a3c2 f8e8 f1e2 b6a4 e3f2 d8g5 c1b1 g5f6 c2d4 c8d7 h2h4 a8d8 d4e6 d7e6 d3f4 e6c4 e2c4 b5c4');
INSERT INTO evaluations VALUES(41012,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(41012,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(41012,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(41012,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41012,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(41012,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(41012,36,'cp 33','c2c4 e8g8 d2d4 c7c6 b2b3 b7b6 d1c2 b8d7 f1d1 c8b7 b1c3 b7a6 c2d3 c6c5 c1b2 d8c7 c3b5 a6b5 c4b5 a7a6 b5a6 c5c4 b3c4 d5c4 d3c2 f8c8 d1c1 a8a6 a2a4 e7b4 e2e4 c4c3 b2a3 b4a3 a1a3 b6b5 c2c3 c7c3 c1c3 c8c3 a3c3 a6a4');
INSERT INTO evaluations VALUES(40943,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(40943,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(40943,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40943,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(40943,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40943,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40943,36,'cp 13','e8g8 d2d4 c7c6 b2b3 b7b6 d1c2 b8d7 c1b2 c8b7 b1d2 a8c8 e2e4 c6c5 e4d5 e6d5 f3h4 c5d4 h4f5 f8e8 f1e1 e7f8 e1e8 f6e8 a1c1 d4d3 c2d3 d5c4 d2c4 b7g2 g1g2 b6b5 c4e3 c8c1 b2c1');
INSERT INTO evaluations VALUES(36310,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36310,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(36310,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(36310,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(36310,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36310,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36310,36,'cp 29','d2d4 c7c6 b2b3 b7b6 c1b2 b8d7 d1c2 c8b7 b1c3 b7a6 c2d3 d8c8 e2e3 c6c5 c3b5 d5c4 d3c4 a6b7 d4c5 c8c5 c4c5 e7c5 f3e5 b7g2 g1g2 d7e5 b2e5 f6d5 f1d1 f8c8 e3e4 d5b4 d1d7 c8d8 a1d1 d8d7 d1d7 b4a2 d7a7 a8a7');
INSERT INTO evaluations VALUES(41920,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41920,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(41920,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(41920,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(41920,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(41920,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(41920,36,'cp 54','g2g3 e8g8 f1g2 b7b5 a2a4 c7c6 e1g1 a7a5 c1e3 h7h6 d1e1 f8e8 h2h3 b5a4 a1a4 d6d5 f3d2 d5e4 d2e4 f6e4 a4e4 a5a4 e1d2 a4a3 e4b4 c8e6 c3c4 b8d7 f1a1 e5e4 b4b3 d7e5 g2e4 e6h3 b3a3 a8a3 a1a3 d8f6 e4g2 h3g2');
INSERT INTO evaluations VALUES(42233,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(42233,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42233,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42233,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42233,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42233,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(42233,36,'cp 0','c7c6 e2e3 b4d6 f1d3 e8g8 f3e5 f8e8 f2f4 c6c5 e1g1 b8c6 c3d5 d6e5 f4e5 f6d5 d3h7 g8h7 d1h5 h7g8 h5f7 g8h7 f7h5 h7g8');
INSERT INTO evaluations VALUES(45678,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(45678,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(45678,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(45678,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45678,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(45678,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(45678,36,'cp 39','b1d2 c6c5 c2c3 b8d7 e1g1 f8e7 e3e4 d5e4 d2e4 e8g8 e4f6 d7f6 d4c5 e7c5 c1f4 b7b6 d1c2 c8b7 a1d1 d8c8 f4e5 b7f3 g2f3 c5e7 f1e1 g7g6 c2e2 f8d8 f3f4 c8c5 e5d4 c5h5 e2h5 f6h5 f4f5 g6f5 d3f5 d8d7 f5c2 a8c8 c2a4 d7d8 g1f1 h5f4');
INSERT INTO evaluations VALUES(47102,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(47102,31,'cp -160',NULL);
INSERT INTO evaluations VALUES(47102,32,'cp -180',NULL);
INSERT INTO evaluations VALUES(47102,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(47102,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(47102,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(47102,36,'cp -174','f1c4 c7c6 b1c3 b7b5 c4b3 b5b4 d5c6 d8d1 c3d1 b8c6 a2a3 e7e6 a3b4 f8b4 c2c3 b4c5 g1e2 c6e5 e1g1 a7a5 c1e3 c5e3 d1e3 e8e7 f1d1 c8d7 b3c2 a5a4 d1d4 h8b8 a1a2 d7c6 h2h3 g7g6 c2a4 a8a4 a2a4 c6a4 d4a4 b8b2 a4a7 f6d7 e2d4');
INSERT INTO evaluations VALUES(47684,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(47684,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(47684,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(47684,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47684,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(47684,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47684,36,'cp 20','c7c5 e1g1 b8c6 h2h3 c5c4 d3e2 c8d7 b2b3 b7b5 a2a4 a7a6 c2c3 f8e7 b1d2 e8g8 b3c4 b5c4 d1c2 c6a5 f3e5 d7e8 e3e4 a8c8 e4d5 f6d5 f4e3 a5b3 a1a2 d5e3 f2e3 b3d2 c2d2 d8d5 d2c2 e7g5 e3e4 d5d6 e5c4 d6g3 f1f3 g3c7 g1h1 f7f6');
INSERT INTO evaluations VALUES(52400,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(52400,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52400,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(52400,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(52400,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(52400,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(52400,36,'cp 27','e2e4 g8f6 f2f4 d7d6 b1c3 e7e5 g1f3 b8d7 f1e2 f8e7 e1g1 e5f4 c1f4 d7f8 g1h1 f8g6 f4d2 e8g8 a2a4 c8e6 d1e1 a7a6 a4a5 d6d5 e4e5 f6d7 e1g3 c6c5 d3d4 c5d4 c3a4 d8c7 e2d3 d7e5 f3d4');
INSERT INTO evaluations VALUES(54887,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(54887,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(54887,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(54887,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(54887,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(54887,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(54887,36,'cp 150','e2e4 c7c5 d4d5 g8f6 b1c3 f8g7 f1c4 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 f6e8 f3e2 a7a6 f2f4 b7b5 c4b3 b8d7 f1d1 g7d4 g1h2 b5b4 c3a4 d8a5 e4e5 a5b5 e2c2 e8g7 g2g4 e7e6 b3c4 b5b7 c2g2 e6d5 c4d5');
INSERT INTO evaluations VALUES(1901,30,'cp -698',NULL);
INSERT INTO evaluations VALUES(1901,31,'cp -705',NULL);
INSERT INTO evaluations VALUES(1901,32,'cp -713',NULL);
INSERT INTO evaluations VALUES(1901,33,'cp -712',NULL);
INSERT INTO evaluations VALUES(1901,34,'cp -717',NULL);
INSERT INTO evaluations VALUES(1901,35,'cp -717',NULL);
INSERT INTO evaluations VALUES(1901,36,'cp -736','b4b2 c2c3 b2a1 d2c2 d7d6 e5d6 f8d6 f4d6 c7d6 f3d2 c8e6 d2b3 e6b3 a2b3 a8c8 e2e3 g8f6 f1e2 e8g8 e1g1 d6d5 b1d2 a1a5 f1d1 f8e8 b3b4 a5b6 c2f5 b6c7 f5b1 h7h6 d1c1 c6e5 h2h3 c7d7 b4b5 d7e6 b1b4 e6f5 c3c4 b7b6 c1a1');
INSERT INTO evaluations VALUES(1905,30,'cp -696',NULL);
INSERT INTO evaluations VALUES(1905,31,'cp -721',NULL);
INSERT INTO evaluations VALUES(1905,32,'cp -713',NULL);
INSERT INTO evaluations VALUES(1905,33,'cp -690',NULL);
INSERT INTO evaluations VALUES(1905,34,'cp -698',NULL);
INSERT INTO evaluations VALUES(1905,35,'cp -717',NULL);
INSERT INTO evaluations VALUES(1905,36,'cp -722','d2c1 b2a1 e2e4 a1a2 f1e2 f8b4 c2c3 b4c5 b1d2 g8e7 e1g1 e7g6 f4g3 e8g8 e2c4 a2a3 c1c2 d7d6 c4b5 c8e6 e5d6 c5d6 b5c6 b7c6 e4e5 d6c5 d2e4 a8d8 h2h3 h7h6 f3h4 e6c4 h4g6 f7g6 e4c5 c4f1 c5e6 f1d3');
INSERT INTO evaluations VALUES(2484,30,'cp -450',NULL);
INSERT INTO evaluations VALUES(2484,31,'cp -451',NULL);
INSERT INTO evaluations VALUES(2484,32,'cp -455',NULL);
INSERT INTO evaluations VALUES(2484,33,'cp -467',NULL);
INSERT INTO evaluations VALUES(2484,34,'cp -479',NULL);
INSERT INTO evaluations VALUES(2484,35,'cp -471',NULL);
INSERT INTO evaluations VALUES(2484,36,'cp -478','h7h6 g5h3 c8g4 c2c3 d8e8 e1f1 e8h5 c1e3 a8e8 d3d4 e5d4 h3f4 d4e3 f4h5 e3e2 d1e2 e8e2 f1e2 g4h5 a1d1 h5f7 h1e1 g8h7 h2h4 f8d6 e2f1 h8d8 h4h5 f7h5 d1d5 h5f7 d5d6 d8d6 b2b3 f7g6 c3c4 c6d4 f3d4');
INSERT INTO evaluations VALUES(6372,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(6372,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(6372,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(6372,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(6372,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(6372,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(6372,36,'cp 80','b1d2 g8f6 e2e3 d7d6 f1e2 b4d2 f3d2 e6e5 d4d5 c6e7 e1g1 e8g8 b2b3 f8e8 e2d3 a7a5 c1b2 c8f5 e3e4 f5d7 f1e1 e7g6 d3f1 b7b6 h2h3 d7c8 a2a3 f6d7 b3b4 c7c5 b4b5 d8g5 d2f3 g5h6');
INSERT INTO evaluations VALUES(9650,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(9650,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(9650,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(9650,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(9650,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(9650,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(9650,36,'cp -4','e8f7 f3g5 f7g8 d1g4 d7d5 g4g3 h7h6 g5f3 g8h7 c2c3 d4c3 b1c3 c8f5 f3h4 d8d7 c3d5 d7d5 h4f5 d5e5 g3d3 h8d8 f5d4 h7h8 d4c6 b7c6 d3g3 d8d5 g3e5 d5e5 c1e3 c5e3 f2e3 e5e3 a1c1 a8d8 c1c2 e3e7 f1c1 d8d6 h2h3 e7e6 c2f2 e6e3 f2c2 h8h7 c2c6 d6c6 c1c6 e3e1 g1f2');
INSERT INTO evaluations VALUES(4690,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(4690,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(4690,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(4690,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(4690,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(4690,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4690,36,'cp 0','f3g5 f7g8 d1g4 d7d5 g4g3 c8f5 c2c3 h7h6 g5f3 d4c3 b1c3 g8h7 c1e3 d5d4 a1d1 d4e3 d1d8 e3f2 g1h1 a8d8 f1f2 c5f2 g3f2 h8e8 f3g5 h7g6 g5f3');
INSERT INTO evaluations VALUES(4689,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4689,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4689,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4689,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4689,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4689,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4689,36,'cp 0','f7g8 d1g4 d7d5 g4g3 c8f5 c2c3 h7h6 g5f3 d4c3 b1c3 g8h7 c1e3 d5d4 a1d1 d4e3 d1d8 e3f2 g1h1 a8d8 c3d1 f5d3 f1f2 d3c2 f2f1 c2d1 f3g5 h6g5 g3h3 h7g8 h3e6 g8h7');
INSERT INTO evaluations VALUES(4650,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4650,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4650,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4650,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4650,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4650,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4650,36,'cp 0','d1g4 d7d5 g4g3 h7h6 g5f3 g8h7 c2c3 d4c3 b1c3 c8f5 c1e3 d5d4 a1d1 d4e3 d1d8 e3f2 g1h1 a8d8 c3d1 f5d3 f1f2 d3c2 f2f1 c2d1 f3g5 h6g5 g3h3 h7g8 h3e6 g8h7');
INSERT INTO evaluations VALUES(4649,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4649,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4649,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4649,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4649,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4649,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4649,36,'cp 0','d7d5 g4g3 c8f5 c2c3 d4c3 b1c3 h7h6 g5f3 g8h7 c1e3 d5d4 a1d1 d4e3 d1d8 e3f2 g1h1 a8d8 c3d1 f5d3 f1f2 d3c2 f2f1 c2d1 f3g5 h6g5 g3h3 h7g8 h3e6 g8h7');
INSERT INTO evaluations VALUES(10920,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(10920,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(10920,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(10920,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(10920,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(10920,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(10920,36,'cp 59','g1f3 d7d6 d2d4 a7a5 e1g1 e8g8 b2b3 c8g4 d4d5 c6b4 h2h3 g4h5 a2a3 b4a6 e2e4 c7c6 c1e3 c6d5 c4d5 a8c8 d1d3 a6c5 d3b1 h5g6 f3d2 a5a4 b3a4 f6d7 f1d1 h7h6 b1b5 f7f5 e4f5 g6f5 d2e4 b7b6');
INSERT INTO evaluations VALUES(10919,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(10919,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(10919,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(10919,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(10919,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(10919,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(10919,36,'cp 62','e8g8 e1g1 f8e8 e2e3 e7b4 c3d5 b4f8 d2d4 d7d6 d5f6 d8f6 d4d5 c6b8 e3e4 f8e7 c1g5 f6g6 g5e7 e8e7 d1c2 a7a5 a2a3 b7b6 b2b4 e7e8 f1c1 b8d7 f3d2 g6h5 g2f3 h5g5 h2h4 g5e7 d2f1 g7g6 f1e3 h7h5 f3e2 e8f8 g1h2 d7f6 c4c5 b6c5 b4c5');
INSERT INTO evaluations VALUES(4600,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(4600,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(4600,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(4600,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(4600,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(4600,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(4600,36,'cp 46','e1g1 f8e8 e2e3 e7b4 c3d5 b4f8 d2d4 d7d6 d5f6 d8f6 d4d5 c6b8 e3e4 a7a5 c1e3 b7b6 d1d2 f6e7 b2b3 g7g6 a2a3 b8d7 d2c2 f8g7 a1e1 d7f6 h2h3 f6h5 c2d1 e7f8 f3g5 g7h6 g1h2 f8e7 g5f3 h6e3 e1e3 c8d7 d1d2');
INSERT INTO evaluations VALUES(11315,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(11315,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(11315,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(11315,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(11315,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(11315,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11315,36,'cp 20','c8f5 g1f3 e7e6 f3h4 f5e4 f1e2 h7h6 h4f3 f8d6 e1g1 e4h7 d1b3 a8b8 f4g3 e8g8 f1c1 d8e7 c3c4 c6a5 b3d1 a5c4 d2c4 d5c4 c1c4 h7e4 e2d3 b8c8 d3e4 f6e4 g3e5 e4g5 f3d2 d6e5 d4e5 c7c6 a1c1 f8d8 d1c2');
INSERT INTO evaluations VALUES(24265,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(24265,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(24265,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24265,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(24265,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24265,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(24265,36,'cp 26','g1f3 e7e6 f3h4 f5e4 f1e2 h7h6 h4f3 f8d6 e1g1 e4h7 f4g3 e8g8 c3c4 d8e7 a1c1 d6g3 h2g3 a6a5 a2a3 c6b8 d1b3 c7c6 f3e5 e7c7 e2d3 h7d3 e5d3 b8d7 b3c2 f8e8 a3a4 e6e5 c4d5 f6d5 d4e5 d7e5 d3e5 c7e5 d2c4');
INSERT INTO evaluations VALUES(13574,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(13574,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(13574,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(13574,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(13574,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(13574,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13574,36,'cp 7','a2a3 b7b6 d4c5 f8c5 c2c4 d5c4 d1d8 e8d8 f1c4 d8e7 b1d2 c8b7 e1e2 c5d6 f4d6 e7d6 a1c1 a8c8 c1c2 c6e5 f3e5 d6e5 f2f3 c8c7 h1c1 h8c8 c4d3 c7c2 c1c2 c8c2 d3c2 e5d6 d2c4 d6e7 e3e4 h7h6 c2d3 g7g5 g2g3 f6d7 e2e3 b7a6 c4d2 a6d3');
INSERT INTO evaluations VALUES(15529,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(15529,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(15529,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(15529,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(15529,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15529,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(15529,36,'cp 43','g2g3 g7g6 g1e2 c8e6 f1g2 f8g7 a1b1 d8d7 e1g1 h7h5 h2h4 a7a5 a2a3 e8g8 b2b3 g8h7 c1g5 f6g4 c3d5 f7f6 g5d2 g4h6 b3b4 a5b4 a3b4 c6e7 d1c2 f8f7 d5e7 d7e7 b4b5 h6g4 b1a1 f7f8 f2f3 g4h6 d2e3 c7c6');
INSERT INTO evaluations VALUES(18582,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(18582,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(18582,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(18582,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(18582,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(18582,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(18582,36,'cp 41','f1c4 g8f6 e4g5 e7e6 d1e2 f6d5 d2d4 h7h6 g5e4 f8e7 e1g1 e8g8 f1e1 d8c7 c2c3 f8d8 c4b3 c6c5 b3d5 e6d5 e4g3 e7f8 g3h5 c7b6 c1f4 d7f6 d4c5 f8c5 h5f6 b6f6 f4c7 d8f8 e2d2 c8e6 a2a3 d5d4 f3d4 c5d4 c3d4 e6d5 c7g3');
INSERT INTO evaluations VALUES(18576,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(18576,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(18576,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(18576,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(18576,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(18576,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(18576,36,'cp 55','g8f6 e4g5 e7e6 d1e2 f6d5 d2d4 h7h6 g5e4 f8e7 e1g1 e8g8 f1e1 d8c7 c2c3 f8d8 c4b3 c6c5 b3d5 e6d5 e4c5 e7c5 d4c5 d7c5 h2h3 b7b6 f3d4 c8d7 e2f3 c7d6 c1f4 d6f6 f3g3 d8e8 e1e8 a8e8 f4e3 c5e6 a2a4 e6d4 e3d4 f6g5 g3g5 h6g5');
INSERT INTO evaluations VALUES(19830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19830,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19830,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(19830,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19830,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(19830,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(19830,36,'cp 0','f1b5 c8g4 b5c6 b7c6 d2d3 e7e6 b1d2 g4h5 h2h3 g8e7 e1g1 e7c8 f1e1 h7h6 d2f1 h5g6 f1g3 a7a5 g3e2 f8e7 d1d2 c8b6 e2f4 g6h7 f4h5 h7g6');
INSERT INTO evaluations VALUES(23318,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(23318,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(23318,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(23318,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(23318,34,'cp 232',NULL);
INSERT INTO evaluations VALUES(23318,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(23318,36,'cp 228','b4a6 b1c3 d7d6 c1e3 g8e7 f1a6 b7a6 d1d3 g7g6 f3d2 f8g7 h2h3 f6f5 f2f3 e8g8 e1c1 f5f4 e3f2 f8f7 d2c4 h7h6 b2b3 c8d7 c1b1 g8h7 b1a1 d8c8 d1d2 f7f8 a1a2 g6g5 h1d1 g7f6 a3a4 e7g6 d1h1');
INSERT INTO evaluations VALUES(23344,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(23344,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(23344,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(23344,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(23344,34,'cp 248',NULL);
INSERT INTO evaluations VALUES(23344,35,'cp 249',NULL);
INSERT INTO evaluations VALUES(23344,36,'cp 241','b1c3 d7d6 c1e3 g8e7 f1a6 b7a6 d1d3 g7g6 f3d2 f8g7 h2h3 f6f5 f2f3 f5f4 e3f2 e8g8 e1c1 f8f7 d2c4 c8d7 a3a4 g7f8 b2b3 d7e8 c3a2 d8d7 c1b2 h7h6 h1g1 g6g5 g1h1 d7c8 a2b4 c8b7 c2c3 e8a4 b3a4 a8b8 b2a1');
INSERT INTO evaluations VALUES(23449,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(23449,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(23449,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(23449,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(23449,34,'cp -81',NULL);
INSERT INTO evaluations VALUES(23449,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(23449,36,'cp -68','d7d5 b1c3 d5d4 c3e2 f7f5 d2d3 g8f6 a2a3 f5e4 d3e4 e7e5 f1g2 a7a5 g1f3 f8c5 e1g1 e8g8 c1g5 h7h6 g5f6 d8f6 f3e1 f6e7 e1d3 b7b6 f2f4 c8b7 g1h1 a5a4 c2c3 b7a6 d3e5 c6e5 f4e5 f8f1 g2f1 a6e2 d1e2 e7e5 e2c4 g8h8 c4d5 a8e8');
INSERT INTO evaluations VALUES(25530,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(25530,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(25530,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(25530,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(25530,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(25530,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25530,36,'cp 42','c6d4 b5d7 d8d7 d1d4 f8e7 c2c4 e8g8 b1c3 f8e8 f2f3 a7a6 c1e3 b7b5 a1d1 d7c6 b2b3 b5c4 b3c4 a8b8 d1b1 c6d7 b1b3 c7c6 e1b1 b8b3 a2b3 d7b7 d4b6 b7b6 e3b6 e8b8 b6f2 f6d7 c3e2 e7f6 e2d4 f6d4 f2d4 b8b4 d4f2 d7c5');
INSERT INTO evaluations VALUES(29126,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(29126,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(29126,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(29126,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(29126,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(29126,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(29126,36,'cp 22','b8c6 d1b3 c6a5 b3a4 f5d7 a4c2 a8c8 e2e3 b7b5 a2a3 e7e6 f1d3 a7a6 e1g1 f8e7 h2h3 e8g8 f3e5 h7h6 f1c1 d8b6 b2b3 f8d8 c2b2 b6b7 c1c2 d7e8 d3e2 c8a8 e5d3 d8c8 a1c1 c8c6 d3b4 c6c8 f4g3 f6d7');
INSERT INTO evaluations VALUES(24620,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(24620,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(24620,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(24620,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(24620,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24620,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(24620,36,'cp 34','d1b3 c6a5 b3a4 f5d7 a4c2 e7e6 e2e3 f8b4 f1d3 a8c8 e1g1 h7h6 h2h3 e8g8 f3e5 d7e8 a1c1 b4c3 b2c3 a7a6 c2e2 e8b5 d3b5 a6b5 e2b5 f6e4 f2f3 e4d6 b5b4 b7b6 e5g6 f7g6 f4d6 f8e8 a2a4 a5c4 d6f4 c8a8 b4b5');
INSERT INTO evaluations VALUES(30228,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(30228,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(30228,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(30228,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(30228,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(30228,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(30228,36,'cp 22','d1b3 b8c6 e2e4 f6e4 b3b7 f5d7 b7b3 f8b4 f1d3 a8b8 b3c2 d7g4 d3e4 d5e4 c2e4 d8e7 f3e5 c6e5 d4e5 e7e6 e1g1 e8g8 h2h3 b4c3 b2c3 g4f5 e4f3 f8d8 c1e3 a7a6 f1d1 h7h6 e3d4 a6a5 d4e3 f5g6 d1d4 e6e5 f3c6 d8d4 c3d4 e5d6 a1c1 d6c6 c1c6 b8b1 g1h2');
INSERT INTO evaluations VALUES(31823,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(31823,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(31823,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(31823,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(31823,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(31823,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(31823,36,'cp 84','e7e5 g1f3 b8c6 c4d5 c6b4 e2e4 f5e4 f3e5 e4d5 f1c4 g8f6 e1g1 f8e7 a2a3 b4c6 b1c3 d5c4 b3c4 e8g8 e5c6 b7c6 c4c6 d8d4 c6c7 e7d6 c3b5 d6c7 b5d4 c7b6 d4f5 g7g6 f5d6 f8d8 d6c4 b6d4 c1e3 d4e3 f2e3 f6e4 a1c1 e4d2 c4d2 d8d2 f1f2 d2d3 c1c7');
INSERT INTO evaluations VALUES(31885,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(31885,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31885,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(31885,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(31885,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31885,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(31885,36,'cp 8','c7c6 e2e3 h7h6 g5f4 g8f6 f1d3 f5d3 d1d3 e7e6 e1g1 f8d6 f4d6 d8d6 c2c4 e8g8 b1c3 b8d7 f1d1 a8d8 a2a4 a7a6 h2h3 d8c8 d3e2 d5c4 e2c4 c6c5 a4a5 c8c7 d4c5 d6c5 c4b3 f8b8 d1c1 b7b5 a5b6 b8b6');
INSERT INTO evaluations VALUES(31992,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(31992,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(31992,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(31992,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(31992,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(31992,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(31992,36,'cp 191','a1b1 e7e6 d1b3 d8c8 g1f3 g8f6 e1g1 f8e7 e3e4 e8g8 e4e5 f6d5 c4d5 e6d5 b3d5 f8d8 d5b3 a7a5 f1d1 b8c6 c1e3 a5a4 b3c4 c8e6 c4e6 f7e6 d1c1 h7h6 h2h4 d8c8 g2g3 c6b4 a2a3 b4d5 c1c2 a8a5 g1f1 a5b5 f3d2 b5b6 f1e2 g8f7 b1c1 b6b5 d2b1');
INSERT INTO evaluations VALUES(32991,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(32991,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(32991,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(32991,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(32991,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(32991,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(32991,36,'cp 157','c1g5 f6g6 h2h4 f7f6 g5c1 c5b4 h4h5 g6f7 d3d4 d7d6 d1d3 b7b6 h5h6 c8a6 h6g7 f7g7 d3e3 g8e7 g2g3 a6f1 e1f1 b4c3 b2c3 c6c5 e3d3 h7h5 f3h4 e5d4 a2a4 b8c6 c3d4 c6d4 c2c3 d4c6 d3d6');
INSERT INTO evaluations VALUES(32983,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(32983,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(32983,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(32983,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(32983,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(32983,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(32983,36,'cp 134','f6g6 d1d2 f7f6 g5e3 c5e3 d2e3 d7d6 d3d4 g6h6 e3d3 b8a6 a1d1 a6b4 d3e2 g8e7 a2a3 b4a6 d4d5 c8d7 d5c6 d7c6 d1d6 h6c1 c3d1 a6c5 e2c4 c5a4 f1e2 b7b5 c4b4 a7a5 b4d2 c1d2 f3d2 e8g8 f2f3 a8d8 d6d8 f8d8 b2b4 a5b4 a3b4 c6d7 d1e3');
INSERT INTO evaluations VALUES(32999,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(32999,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(32999,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(32999,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(32999,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(32999,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(32999,36,'cp 138','d1d2 f7f6 g5e3 c5e3 d2e3 d7d6 d3d4 g6h6 e3d3 c8g4 f3d2 b8a6 d3g3 g4h5 f1a6 b7a6 d2c4 a8d8 d4d5 g8e7 d5c6 e7c6 c4e3 h5g6 g3g4 e8g8 c3d5 g6f7 c2c3 h6g6 g4g6 f7g6 f2f3 f6f5 e4f5 g6f5 e3f5 f8f5 e1c1 g8f7 d1d2 c6e7 h1d1 e7d5 d2d5 f5g5 d1d2');
INSERT INTO evaluations VALUES(34842,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(34842,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(34842,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(34842,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(34842,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(34842,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(34842,36,'cp 104','d2d4 e5a5 c1d2 a5b6 c3a4 b6c6 e2c3 c8d7 d1e2 g8f6 e1c1 f8e7 e2e3 c6d6 a4c5 d7c6 f2f3 e8g8 c5e4 d6d7 c1b1 a7a5 g2g4 f6d5 c3d5 e6d5 e4g3 f8e8 h2h4 c6b5 f1b5 d7b5 g3f5 e7f8 e3f4 b8d7 g4g5 a5a4 h4h5 a4a3');
INSERT INTO evaluations VALUES(34839,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(34839,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(34839,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(34839,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(34839,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(34839,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(34839,36,'cp 103','e5a5 c1d2 a5b6 c3a4 b6c6 e2c3 a7a6 f1e2 g8f6 d2f4 f6d5 e2f3 f8d6 f4d6 c6d6 c3e4 d6f4 g2g3 f4h6 c2c4 d5e7 d4d5 e6d5 c4d5 e8g8 d5d6 c7d6 d1d6 h6d6 e4d6 b8c6 f3c6 b7c6 d6c8 a8c8 e1c1 e7d5 h1e1 g7g6 a4c5 f8d8 a2a3');
INSERT INTO evaluations VALUES(36337,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36337,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36337,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(36337,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36337,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36337,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36337,36,'cp 0','g1f3 b8d7 c4d5 e6d5 f1d3 f8e8 e1g1 e7d6 c3b5 d6f8 d1c2 c7c6 b5c3 f8d6 f1e1 d7b6 h2h3 h7h6 c1d2 c8e6 a3a4 a7a5 a1c1 b6d7 c1d1 a8c8 d1c1');
INSERT INTO evaluations VALUES(37369,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(37369,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(37369,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(37369,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(37369,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(37369,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(37369,36,'cp 27','e4e5 f6h5 d2d4 d7d6 h2h3 d6e5 d4e5 d8d1 c3d1 c8e6 c4e2 b8c6 g2g4 c6b4 e2d3 e6d5 e1e2 c7c5 d1c3 c5c4 d3e4 d5e4 c3e4 f7f5 e4g5 f5g4 h3g4 h5f4 c1f4 f8f4 h1h7 f4g4 a1h1 b4d5 g5e6 d5f4 e6f4 g4f4 f3g5 a8f8 h7g7 g8g7 g5e6 g7g8 e6f4 f8f4 e2e3 f4g4 h1d1 g4g5 d1d5 g5h5 e3e4 g8f7');
INSERT INTO evaluations VALUES(39723,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(39723,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39723,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39723,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39723,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39723,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39723,36,'cp 0','a2a4 b8c6 e1g1 e8g8 d4c6 b7c6 a4a5 c8b7 f1e1 c6c5 d1e2 d8c7 e3f4 c7c6 a1d1 a8d8 h2h3 h7h6 f4g3 f8e8 c4b3 c6c7 d1a1 d8c8 b3c4 c7c6 c4b3');
INSERT INTO evaluations VALUES(40997,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(40997,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(40997,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(40997,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(40997,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(40997,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(40997,36,'cp -5','f1d3 c7c5 e1g1 e8g8 b2b3 b7b6 c1b2 b8c6 d4c5 b6c5 c3a4 c8d7 c2c4 c6b4 a4c3 b4d3 d1d3 d5c4 d3c4 d8b8 f1c1 d7c6 c3a4 c6b5 c4h4 f8c8 b2e5 b8b7 a4b2 c8d8 h4g3 d8d7 h2h3 a8d8 b2c4 h7h6 e5b2 b5c4 c1c4');
INSERT INTO evaluations VALUES(40998,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(40998,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(40998,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(40998,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(40998,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(40998,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(40998,36,'cp -10','c7c5 e1g1 c5c4 d3e2 a7a6 e3e4 f6e4 c3e4 d5e4 f3d2 b7b5 d2e4 c8b7 e2f3 e8g8 c1f4 b8d7 f1e1 a6a5 a2a3 d8b6 c2c3 b7d5 d1e2 a8c8 e4d2 d5f3 e2f3 d7f6 d2f1 b6c6 f3c6 c8c6 g2g3 f6d5 f1e3 d5f4 g3f4 e7f6 f4f5');
INSERT INTO evaluations VALUES(42870,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(42870,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(42870,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(42870,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(42870,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(42870,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(42870,36,'cp 76','d7d6 b1c3 e8g8 b2b4 c5b6 a2a4 a7a5 c3d5 b8d7 b4b5 g4f6 d5b6 d7b6 c4b3 h7h6 h2h3 f8e8 f3h2 c7c6 f2f4 c6b5 a4b5 e5f4 c1f4 a5a4 b3a2 a4a3 f4g3 c8e6 a2e6 f7e6 h2g4 f6g4 d1g4 b6d7');
INSERT INTO evaluations VALUES(43705,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(43705,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(43705,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(43705,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(43705,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(43705,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(43705,36,'cp 68','b1c3 f8g7 e2e4 e8g8 a2a4 b8a6 f1e2 e7e6 g1f3 a6b4 e1g1 e6d5 e4d5 d7d6 c1f4 h7h6 e2c4 c8f5 a1c1 g6g5 f4g3 f6e4 f1e1 e4g3 h2g3 d8f6 b2b3 f6g6 e1e2 f5d3 e2e3 d3f5 d1d2 b4c2 e3e2 c2d4 f3d4 c5d4 c3b5 d4d3');
INSERT INTO evaluations VALUES(43704,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(43704,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(43704,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(43704,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(43704,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(43704,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(43704,36,'cp 99','f8g7 e2e4 e8g8 a2a4 b8a6 g1f3 e7e6 f1e2 a6b4 e1g1 e6d5 e4d5 c8b7 e2c4 b7a6 d1b3 d7d6 h2h3 d8d7 c1g5 a6c4 b3c4 f8b8 b2b3 b4a6 a1c1 b8b4 c4d3 c5c4 b3c4 a6c5 d3e2 c5a4 g5f6 g7f6 c3a4 d7a4 c4c5 a4b5 e2b5 b4b5');
INSERT INTO evaluations VALUES(39796,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(39796,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(39796,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(39796,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(39796,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(39796,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(39796,36,'cp 98','e2e4 e8g8 a2a4 e7e6 g1f3 b8a6 f1e2 a6b4 e1g1 e6d5 e4d5 d7d6 c1f4 c8a6 f1e1 d8d7 h2h3 a6e2 e1e2 f8e8 e2e8 f6e8 a4a5 d7f5 f4g3 f5c2 d1c2 b4c2 a1a4 c2b4 c3e4 g7f8 f3d2 b4d5 d2c4 a8a6 f2f3 d5b4 e4c3 d6d5 c4b6 d5d4 c3e4');
INSERT INTO evaluations VALUES(39795,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(39795,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(39795,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(39795,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(39795,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(39795,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(39795,36,'cp 94','e8g8 a2a4 e7e6 g1f3 b8a6 f1e2 a6b4 e1g1 e6d5 e4d5 d7d6 c1f4 c8a6 f1e1 a6e2 e1e2 d8d7 d1d2 c5c4 h2h3 f8b8 f4h6 b4d3 h6g7 g8g7 a1b1 d3c5 d2d4 c5a4 c3e4 d7d8 b1e1 a4c5 f3g5 h7h6 g5f7 c5b3 f7d8 b3d4 d8e6 d4e6 d5e6 f6e4 e2e4 b8b2');
INSERT INTO evaluations VALUES(45908,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(45908,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(45908,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(45908,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(45908,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(45908,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(45908,36,'cp 50','d2d3 f8e7 a2a3 b8c6 c1g5 f6g8 g5d2 g8f6 g1f3 h7h6 e1g1 e8g8 f3e1 c8e6 c3d5 a7a5 a1b1 a5a4 b2b4 a4b3 b1b3 a8a7 a3a4 e6d7 d5b6 a7a6 a4a5 d7e6 b3b5 d6d5 c4d5 f6d5 b6d5 e6d5 g2d5 d8d5 b5b7 e5e4 d3e4 d5e4');
INSERT INTO evaluations VALUES(45909,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(45909,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(45909,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(45909,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(45909,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(45909,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(45909,36,'cp 57','f8e7 c1g5 f6g8 g5d2 g8f6 g1f3 b8c6 a2a3 e8g8 e1g1 a7a5 a1b1 a8b8 f3e1 c8e6 e2e4 c6d4 e1c2 e6g4 f2f3 g4d7 a3a4 h7h5 h2h4 g7g6 g1h2 g8h7 g2h3 d7e6 c3b5 d4b5 a4b5 f6g8 c2e3 b7b6 d2c3 e6h3 h2h3 f7f5 e4f5');
INSERT INTO evaluations VALUES(46900,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46900,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46900,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46900,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(46900,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(46900,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(46900,36,'cp 32','c2c4 e7e6 b1c3 f8e7 b2b3 e8g8 c1b2 c7c5 d4c5 e7c5 c4d5 f6d5 c3d5 e6d5 a2a3 b8c6 f1e2 c8f5 b3b4 c5b6 e1g1 f5e4 a1c1 d8e7 f3d2 e4f5 d2b3 a8d8 e2b5 d5d4 b3d4 c6d4 b2d4 b6d4 e3d4 a7a6');
INSERT INTO evaluations VALUES(46884,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(46884,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46884,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46884,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(46884,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46884,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(46884,36,'cp 25','e7e6 b1c3 f8e7 b2b3 e8g8 c1b2 c7c5 c4d5 c5d4 d1d4 e6d5 f1e2 b8c6 d4d3 c6b4 d3b1 a7a6 e1g1 f8e8 f1d1 c8e6 h2h3 e7d6 f3d4 b4c6 b1d3 e6d7 d4f3 d7e6 a1c1 d8e7 d3b1 a8d8 e2f1 d8c8 c3e2 f6e4 e2f4 d6f4 e3f4 e7c7');
INSERT INTO evaluations VALUES(47675,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(47675,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(47675,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(47675,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(47675,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(47675,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47675,36,'cp 20','b1c3 f8e7 b2b3 e8g8 c1b2 c7c5 d4c5 e7c5 c4d5 f6d5 c3d5 e6d5 a2a3 b8c6 b3b4 c5b6 f1e2 c8f5 e1g1 f5e4 f3d2 e4g6 d2b3 d8g5 a1c1 g6e4 e2f3 a8d8 b3c5 g5e7 f3e4 d5e4 d1c2 b6c5 c2c5 f8e8 c5e7 e8e7 f1d1 e7d7 d1d7 d8d7');
INSERT INTO evaluations VALUES(50872,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(50872,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(50872,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(50872,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(50872,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(50872,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(50872,36,'cp 77','d2d4 b8c6 e1g1 g7g5 c1e3 f8g7 d4e5 d6e5 b1d2 d8f6 d1c2 g8e7 a2a4 e8g8 h2h3 e7g6 f1e1 g6f4 a1d1 c6e7 f3h2 e7g6 e3f4 g5f4 b2b4 a6a5 d2f3 c7c6 d1d2 f8d8 e1d1 d8d2 d1d2 g7f8 c2b3 a5b4');
INSERT INTO evaluations VALUES(50879,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(50879,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(50879,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(50879,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(50879,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(50879,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(50879,36,'cp 68','b8c6 e1g1 g7g5 c1e3 f8g7 d4e5 d6e5 b1d2 d8f6 a2a4 g8e7 f3e1 e7g6 e1d3 g6f4 a4a5 e8g8 f2f3 f4d3 c4d3 f8d8 d3c4 c6e7 d1b3 b7b6 a5b6 c7b6 b3c2 f6c6 c4b3 b6b5 f1d1 g7f8 d2f1 d8d1 c2d1 c8e6 b3e6 c6e6 f1g3 e7g6 g3f5');
INSERT INTO evaluations VALUES(51238,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51238,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51238,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51238,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51238,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51238,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51238,36,'cp 0','c4d5 a8a7 d5f7 e8f7 d1h5 g7g6 h5e5 g8f6 e5b8 a7c7 d2d3 f8g7 b1c3 b5b4 c3e2 d7d5 g1f3 f6d7 b8a8 d7b6 a8b8');
INSERT INTO evaluations VALUES(51250,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51250,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51250,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51250,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51250,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51250,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51250,36,'cp 0','a8a7 d5f7 e8f7 d1h5 g7g6 h5e5 g8f6 e5b8 a7c7 d2d3 d7d5 g1f3 d5e4 d3e4 f8g7 f3g5 f7e8 b1c3 f6d7 b8a8 d7b6 a8b8');
INSERT INTO evaluations VALUES(51606,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(51606,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(51606,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(51606,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(51606,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(51606,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(51606,36,'cp 62','b1c3 d7d5 d2d4 g8f6 e2e4 d5e4 c3e4 f6e4 g2e4 c8b7 g1e2 b8d7 e4g2 g7g6 e1g1 f8g7 h2h4 e7e5 h4h5 g6h5 d4e5 d7e5 e2f4 h5h4 f4h5 e8g8 h5g7 d8d1 f1d1 g8g7 g3h4 f7f6 c1f4 a8e8 f4g3 b7c8 d1d6 c8g4 g3e5 e8e5 d6c6');
INSERT INTO evaluations VALUES(51728,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(51728,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(51728,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(51728,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(51728,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(51728,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(51728,36,'cp 74','f8g7 d2d4 c5d4 f3d4 a7a6 c1e3 b7b5 c4b3 c8b7 d1d2 g8f6 f2f3 b8c6 e1c1 e8g8 h2h4 h7h5 g2g4 c6e5 g4h5 f6h5 c1b1 a6a5 d4b5 a5a4 c3a4 e5f3 d2b4 b7c6 b5c3 a8b8 a4b6 d8f6 b4d6 f6e5 d6e5 f3e5 e3d4 d7d6 b6c4 e5g4 d4g7 g8g7 d1d6 g4f2 d6c6 f2h1');
INSERT INTO evaluations VALUES(52583,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(52583,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(52583,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(52583,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(52583,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(52583,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(52583,36,'cp 113','g1f3 f7f6 f1d3 f8g7 e1g1 f6e5 f3e5 g8h6 c1h6 g7h6 d1f3 c8e6 f3e2 d8d6 b1d2 b8d7 d2f3 d7e5 f3e5 h6f4 a1e1 h8f8 e2d1 e6d7 c2c3 d6f6 e5d7 e8d7 g2g3 f4c7 f2f4 e7e6 e1e3 f6g7 a2a4 a8e8 d1e2 f8f6 f1e1 c7d6');
INSERT INTO evaluations VALUES(52582,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(52582,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(52582,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(52582,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(52582,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(52582,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(52582,36,'cp 118','f8g7 f1d3 f7f6 e1g1 f6e5 f3e5 b8d7 e5f3 g8h6 f1e1 e8g8 c1g5 d7f6 b1d2 h6f7 g5h4 f8e8 c2c4 c8f5 d3f5 g6f5 c4d5 d8d5 d1c2 a8d8 a2a3 e7e6 a1c1 d8d7 b2b4 f6h5 d2c4 h5f4 h4g3 f4g6 h2h3');
INSERT INTO evaluations VALUES(37930,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(37930,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(37930,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(37930,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(37930,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(37930,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(37930,36,'cp 109','f1d3 f7f6 e1g1 f6e5 f3e5 g8h6 c1h6 g7h6 d1f3 c8e6 f3g3 d8d6 f2f4 b8d7 g1h1 e8c8 b1d2 c6c5 g3h4 h6g7 c2c3 c8b8 d2f3 c5d4 c3d4 d8c8 a1e1 g7f6 h4g3 h8f8 h2h3 f6e5 f4e5 d6b4 g3h4 f8f7 e1b1 d7f8 h4e1 b4e1 b1e1 e6f5 d3f5 f7f5');
INSERT INTO evaluations VALUES(53588,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53588,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53588,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53588,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53588,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53588,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53588,36,'cp 0','g8f6 a2a3 f8e7 e2e3 f6h5 f1d3 b8d7 f4e5 h5f6 e5f4');
INSERT INTO evaluations VALUES(45662,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45662,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45662,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45662,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45662,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45662,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45662,36,'cp 0','a2a3 f8e7 e2e3 f6h5 f1d3 b8d7 f4e5 h5f6 e5f4');
INSERT INTO evaluations VALUES(54235,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(54235,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(54235,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(54235,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(54235,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(54235,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(54235,36,'cp -33','d5e4 d3e4 g8f6 e4f3 c5d4 c3d4 f8d6 g1e2 e8g8 b1c3 h7h6 e1g1 b8c6 c1e3 c6e7 e2g3 c8d7 a1c1 d6g3 h2g3 d7c6 f3c6 e7c6 d1f3 d8a5 f1e1 a8d8 e3h6 c6d4 f3f4 a5h5 h6g5 f6g4 g5h4 e6e5 f4g5 h5g5 h4g5 f7f6 f2f3 f6g5 f3g4');
INSERT INTO evaluations VALUES(54285,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(54285,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(54285,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(54285,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(54285,34,'cp -154',NULL);
INSERT INTO evaluations VALUES(54285,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(54285,36,'cp -156','c5d4 f1d3 b8c6 f2f4 g8h6 g1f3 f8b4 c1d2 b4c5 e1g1 c8d7 g1h1 e8g8 c2c3 f7f6 e5f6 f8f6 d1e1 a7a5 c3d4 c6d4 f3d4 c5d4 b1c3 d7c6 a1d1 e6e5 c3b5 c6b5 d3b5 e5e4 b5e2 h6f5 g2g4 f5e3 d2e3 d4e3 e1g3 d5d4 e2c4 g8h8 a2a3 f6h6');
INSERT INTO evaluations VALUES(54429,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(54429,31,'cp -148',NULL);
INSERT INTO evaluations VALUES(54429,32,'cp -143',NULL);
INSERT INTO evaluations VALUES(54429,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(54429,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(54429,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(54429,36,'cp -138','d1d4 b8c6 f1b5 d8a5 b1c3 f8b4 c1d2 g8e7 g1f3 e8g8 b5c6 e7c6 d4e3 f8d8 c3b1 a5c7 e1g1 b4e7 f1e1 b7b5 c2c3 a7a5 f3d4 c8a6 e3f4 c6d4 f4d4 c7b7 a2a3 a8c8 h2h3 g8h8 d2e3 b5b4 a3b4 a5b4 d4d2 b4c3 b1c3 b7b3 a1a6 c8c3');
INSERT INTO evaluations VALUES(54423,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(54423,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(54423,32,'cp -158',NULL);
INSERT INTO evaluations VALUES(54423,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(54423,34,'cp -159',NULL);
INSERT INTO evaluations VALUES(54423,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(54423,36,'cp -148','b8c6 f1b5 d8a5 b1c3 g8e7 g1f3 c8d7 d4d2 c6e5 f3e5 d7b5 c3b5 a5b5 a2a4 b5b6 c1b2 f7f6 a4a5 b6b5 e5d3 e6e5 a5a6 b7b6 d2e2 a8c8 e1g1 e7g6 f2f4 e5e4 f4f5 e8d8 f5g6 e4d3 e2h5 c8c2 b2d4 b5b3 g6h7 b3c4 d4f2 f8d6 h5f7');
INSERT INTO evaluations VALUES(55448,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(55448,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(55448,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(55448,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(55448,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(55448,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(55448,36,'cp 142','b1c3 f8e7 c1e3 g8f6 h2h3 e8g8 f1d3 e6e5 g1f3 b8d7 d3c2 f8e8 e1g1 a7a6 d1e2 c7c6 f1d1 d8c7 a2a4 a6a5 f3d2 d7f8 d4d5 c8d7 g1h2 f8g6 g2g3 c7c8 e2f1 c8d8 f1g2 g8h8 d1h1 a8c8 a1g1 f6h7 h3h4 d7g4 f2f3');
INSERT INTO evaluations VALUES(55446,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(55446,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(55446,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(55446,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(55446,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(55446,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(55446,36,'cp 136','f8e7 c1e3 g8f6 h2h3 e8g8 f1d3 e6e5 g1f3 f8e8 d1e2 e5d4 f3d4 c7c5 d4f3 b8c6 a1d1 e7f8 d3b1 c8d7 e1g1 a7a6 e3f4 d7f5 f1e1 f5g6 g2g4 h6h5 g4g5 f6d7 g1g2 a8b8 f4g3 d8a5 c3d5 b7b5 d5f4 d7e5 c4b5 a5b5 e2b5');
INSERT INTO evaluations VALUES(55527,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(55527,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(55527,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(55527,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(55527,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(55527,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(55527,36,'cp -50','g7g6 h5d1 f8g7 d2d3 g8f6 f1e2 b8c6 f2f4 e5f4 c1f4 c6d4 d1d2 d4e6 f4e3 e8g8 e2f3 c7c6 g1e2 d6d5 e4d5 c6d5 d3d4 b7b5 a2a3 f8e8 e1g1 a7a6 a1e1 c8b7 e3h6 g7h6 d2h6 a8c8 h6d2 f6e4 f3e4');
INSERT INTO evaluations VALUES(3874,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(3874,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(3874,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(3874,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(3874,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(3874,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(3874,36,'cp -126','c6a5 e1g1 a5c4 d3c4 f6d5 c4d5 f7f5 e4f5 c8f5 d1e2 d8f6 c1e3 h7h6 g5e6 f5e6 d5e6 c5e3 f2e3 f6e6 e2b5 f8b8 a2a4 c7c6 b5b4 d6d5 h2h3 e6d7 e3e4 b8e8 f1f2 a8d8 a1f1 e8e7 a4a5 a7a6 b4b3 d7d6 g1h2 d6e6 b3d3 e6g6 f2f5 d5d4 b2b3 d8e8 f1f3 g8h7 f5f8 e8f8');
INSERT INTO evaluations VALUES(9611,30,'cp 399',NULL);
INSERT INTO evaluations VALUES(9611,31,'cp 420',NULL);
INSERT INTO evaluations VALUES(9611,32,'cp 416',NULL);
INSERT INTO evaluations VALUES(9611,33,'cp 399',NULL);
INSERT INTO evaluations VALUES(9611,34,'cp 414',NULL);
INSERT INTO evaluations VALUES(9611,35,'cp 426',NULL);
INSERT INTO evaluations VALUES(9611,36,'cp 409','e4e5 f6h5 c4f7 e8f7 f3g5 f7g8 g5e6 d8e8 e6c7 e8f8 c7a8 d6e5 c1e3 e5d4 e3d4 h5f4 e1g1 d7e5 d4e3 g6g5 a8c7 f8f5 e3f4 g5f4 d1d5 g8f8 f1e1 f4f3 c3e4 f3g2 e1e3 e5g6 d5f5 c8f5 e3f3 g6h4 f3f4 g7e5 c7e6 f8e8 f4h4 f5e6 a1e1 e5b2 e1b1 b2e5 b1b7 e6a2 b7a7 a2b1 a7a8 e8d7 a8h8 e5h8 h4h7 b1c2 e4c5 d7d6 h7h8 c2f5 h2h3');
INSERT INTO evaluations VALUES(12232,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12232,36,'cp 0','d5c4 e2e4 b7b5 a2a4 b5b4 c3a2 a7a5 f1c4 f6e4 d1e2 e4f6 a2b4 a5b4 f3g5 f6d5 c4d5 c6d5 g5f7 e8f7 e2h5 f7f6 h5h4 g7g5 h4g5 f6e6 g5g4 e6f7 g4h5 f7e6');
INSERT INTO evaluations VALUES(12391,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(12391,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(12391,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(12391,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(12391,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(12391,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(12391,36,'cp 1','e4e5 d7d5 c4d3 f6e4 c2c3 c5d4 c3d4 c6b4 d3b5 c8d7 b5d7 d8d7 e1g1 f8e7 f3e1 f7f6 f2f3 e4g5 e5f6 e7f6 c1g5 f6g5 f3f4 g5f6 d1d2 b4c6 e1f3 e8g8 b1c3 a8c8 c3e2 c6e7 a1c1 a7a6 d2d3 d7a4 d3e3 a4d7 e3b3 a6a5 g2g3 e7f5 g3g4 f5h4 f3h4 f6h4 c1c8 f8c8 g1g2 h4e7 a2a3 c8c4');
INSERT INTO evaluations VALUES(15660,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(15660,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(15660,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(15660,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(15660,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(15660,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(15660,36,'cp 208','b1c3 d6d5 c3d5 f6d5 e4d5 d8d5 d2d4 f8d6 f1e2 e8g8 e1g1 c8g4 c1e3 d5f5 f3h4 g4e2 h4f5 e2d1 f1d1 f8e8 c2c4 d6f8 d4d5 c6e5 a1c1 g7g6 f5g3 e5g4 e3f4 f8c5 c1c2 c5d6 f4d6 c7d6 h2h3 g4f6 g1f1 a7a5 g3e2 a5a4 f2f3 h7h5 f1f2 h5h4 e2d4');
INSERT INTO evaluations VALUES(15654,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(15654,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(15654,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(15654,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(15654,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(15654,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(15654,36,'cp 208','d6d5 e4d5 f6d5 c3d5 d8d5 d2d4 f8d6 f1e2 e8g8 e1g1 c8g4 c1e3 d5f5 f3h4 g4e2 h4f5 e2d1 f1d1 c6e7 f5e7 d6e7 d1d3 f7f6 a2a4 g7g5 g1f1 g8f7 a4a5 h7h5 d3b3 b7b6 f1e2 e7d6 h2h3 a8b8 g2g3 f7g6 e2d3 b8e8 a5a6 c7c5 b3b5 h5h4 g3h4 c5d4 h4h5 g6h5');
INSERT INTO evaluations VALUES(20887,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(20887,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(20887,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(20887,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(20887,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(20887,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(20887,36,'cp 121','c8f5 g1f3 e7e6 f1d3 g8e7 c2c3 d8d7 e1g1 c6d8 f3g5 f5d3 d1d3 h7h6 g5h3 g7g5 h3f2 e7f5 f2g4 f8g7 b1d2 e8g8 b2b4 d7e8 a2a4 h6h5 g4e3 f5e3 d3e3 e8g6 f1f8 g7f8 d2f3 g5g4 f3h4 g6g7 c1d2 f8e7 e3f2 e7h4 f2h4 g7g6 a1f1 d8c6');
INSERT INTO evaluations VALUES(21452,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(21452,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(21452,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(21452,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21452,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(21452,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21452,36,'cp 24','d2d4 g8f6 c2c4 f8e7 a2a3 e8g8 b1c3 b7b6 b2b4 a7a5 b4b5 c6a7 f1d3 d5c4 d3c4 c8b7 e1g1 c7c6 b5c6 b7c6 f3e5 a8c8 d1e2 f6e4 e5c6 e4c3 c6d8 c3e2 c4e2 f8d8 f4f5 e6f5 a1b1 c8c3 f1f5 g7g6 f5f1 b6b5 e2g4 f7f5 g4f3');
INSERT INTO evaluations VALUES(22673,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(22673,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(22673,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(22673,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(22673,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(22673,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(22673,36,'cp -102','f8b4 c2c3 d4c3 b1c3 d7d6 e5d6 d8d6 d1d6 c7d6 f1e2 g8f6 e1g1 e8g8 f1d1 f8e8 e2f1 c8e6 c3b5 e8d8 c1f4 d6d5 a1c1 a8c8 a2a3 b4e7 f3d4 c6d4 b5d4 a7a6 f1d3 c8c1 d1c1 d8c8 c1c8 e6c8 f4e3 c8d7 b2b4 f6g4 d3f5 g4e3 f5d7');
INSERT INTO evaluations VALUES(23301,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(23301,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(23301,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(23301,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(23301,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(23301,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(23301,36,'cp 34','e4f5 d8f6 d2d4 e5e4 f3d2 d7d5 d1h5 e8d8 d2c4 f6f5 h5f5 c8f5 c4e3 f5e6 b1d2 g8f6 f2f3 e4f3 d2f3 f8d6 f1d3 c6e7 e1g1 h7h6 c3c4 d5c4 e3c4 e6c4 d3c4 h8e8 f3h4 e8f8 c1d2 c7c6 a1e1 d8c7 a2a3 a8e8 g2g3 g7g5 h4f3 e7d5 f3e5 e8e7 e5g6 e7e1 f1e1 f8e8 e1e8');
INSERT INTO evaluations VALUES(23284,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(23284,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(23284,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(23284,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(23284,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(23284,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(23284,36,'cp 28','d8f6 d2d4 e5e4 f3d2 d7d5 d1h5 e8d8 d2c4 f6f5 h5f5 c8f5 c4e3 f5e6 b1d2 g8f6 f2f3 e4f3 d2f3 f8d6 f1d3 h7h6 e1g1 c6e7 c3c4 c7c6 c1d2 d8d7 f3e5 d7c7 b2b4 h8f8 a2a4 a8d8 a4a5 f6e4 f1f8 d8f8 d2e1 e4g5 a1d1 d5c4 e3c4 e7d5 c4d6');
INSERT INTO evaluations VALUES(26927,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(26927,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(26927,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(26927,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(26927,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(26927,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(26927,36,'cp 31','c6d4 b1d2 g7g6 h2h3 d4f3 d2f3 g4d7 a2a4 f8g7 d1e2 g8f6 e1g1 a7a5 c1g5 e8g8 c4b5 h7h6 g5f6 g7f6 b5d7 d8d7 a1a3 f6d8 f1e1 f7f5 a3b3 b7b6 e2c4 a8c8 b3c3 d8f6 c4d3 g8h8 c3c6 c8e8 d3d2 h8h7 c2c3 d7f7 d2d3');
INSERT INTO evaluations VALUES(27079,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(27079,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(27079,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(27079,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(27079,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(27079,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(27079,36,'cp 27','b1d2 d4f3 d2f3 g8f6 h2h3 g4h5 c1g5 f8e7 g5f6 e7f6 g2g4 h5g6 d1e2 a7a6 e1c1 h7h5 c1b1 f6e7 h1g1 e7g5 c4d3 h5g4 h3g4 h8h6 g1h1 g5f4 d1g1 d8f6 f3e1 e8c8 h1h6 g7h6 e1g2 d8f8 g1h1 f8h8 a2a3 h6h5 g4h5 f6g5 g2f4 e5f4 e2d2');
INSERT INTO evaluations VALUES(30187,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(30187,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(30187,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(30187,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(30187,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(30187,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(30187,36,'cp 85','d3b5 b8d7 b5b7 c7c5 e1g1 f8e7 d4c5 a8b8 b7a7 d7c5 a7a3 e8g8 a3c3 c5a4 c3d4 b8b4 c2c4 d5c4 d4d8 f8d8 f3d4 b4b7 b2b3 c4b3 a2b3 a4c5 b1d2 e7f8 c1a3 e6e5 d4f3 c5b3 d2b3 b7b3 a3f8 d8f8 f3e5 h7h6 h2h3 b3b7 e5c6 f8c8');
INSERT INTO evaluations VALUES(30214,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(30214,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(30214,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(30214,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(30214,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(30214,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(30214,36,'cp 42','b8c6 c3b1 f8d6 c2c3 e8g8 b1d2 g4h5 f1e1 h7h6 h2h3 a7a5 e2f1 c6e7 d2c4 a5a4 a2a3 c7c5 c4d6 d8d6 d4c5 d6c5 d1d4 c5a5 f3e5 f8d8 d4h4 d8d1 e1d1 h5d1 h4b4 a5c7 c1f4 e7d5 b4c4 c7c4 e5c4 d5f4 a1d1');
INSERT INTO evaluations VALUES(30724,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(30724,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(30724,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(30724,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(30724,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(30724,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(30724,36,'cp 74','d7d6 c1e3 f8g7 d1d2 g8f6 g1f3 b8d7 e3h6 g7h6 d2h6 e7e6 e1c1 a7a6 f3d4 d7c5 d1e1 e6e5 d4b3 d8e7 f2f3 e8c8 b3c5 d6c5 h6e3 c8b8 a2a4 h7h5 h2h4 f6e8 c1b1 f7f6 g2g4 b7c8 a4a5 b6b5 g4g5 f6g5 h4g5 e8d6 e3c5 e7g5 e1d1 g5e7 f1e2 b8a8');
INSERT INTO evaluations VALUES(31291,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(31291,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(31291,32,'cp 179',NULL);
INSERT INTO evaluations VALUES(31291,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(31291,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(31291,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(31291,36,'cp 169','d4c5 a7a6 b2b4 e7e6 a2a3 g8e7 f1d3 a6a5 c1b2 a5b4 a3b4 b8c6 b4b5 a8a1 b2a1 f5d3 d1d3 d8a8 d3c3 c6d8 e1g1 a8a4 c3b3 a4a7 b1d2 a7c5 a1d4 c5c7 c2c4 e7f5 c4d5 f5d4 f3d4 c7e5 b3a4 e5d5 b5b6 d8c6 d2b3 f8e7 d4c6 d5c6 a4c6 b7c6 b3d4');
INSERT INTO evaluations VALUES(31397,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(31397,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(31397,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(31397,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(31397,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(31397,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(31397,36,'cp 57','e1g1 b8d7 b1d2 g8e7 d2b3 f5g6 a2a4 e7f5 a4a5 a7a6 c2c4 d5c4 e2c4 f8b4 c1d2 b4d2 d1d2 e8g8 h2h3 a8c8 f1c1 c8c7 c4f1 f5e7 d2b4 g6e4 f3d2 e4d5 b4c3 e7f5 c1e1 d8g5 g2g3 f7f6 a1c1 f6e5 d4e5 g5g6 d2e4');
INSERT INTO evaluations VALUES(31396,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(31396,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(31396,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(31396,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(31396,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(31396,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(31396,36,'cp 45','a7a6 c1e3 g8e7 b1d2 b8d7 d2b3 f5e4 c2c3 e4h7 e3d2 e7g6 c3c4 d5c4 e2c4 f8e7 c4d3 e8g8 b3a5 d8c7 a1c1 d7b6 g2g3 f8d8 h2h4 g6f8 d3h7 g8h7 c1c2 h7g8 a2a3 b6d5 h4h5 f8d7 b2b4 d7b6 a5b3 b6a4 d1e2 c7d7');
INSERT INTO evaluations VALUES(34044,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(34044,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(34044,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(34044,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(34044,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(34044,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(34044,36,'cp -23','c7c6 b1c3 b7b5 c4b3 g7g6 a2a3 f8g7 g1e2 e8g8 d2d4 e5d4 e2d4 d7d6 e1g1 a7a6 c1g5 e7e5 g5f6 e5f6 a1d1 c8b7 f3g3 c6c5 d4e2 c5c4 b3a2 f8d8 b2b3 c4b3 a2b3 b8d7 b3d5 b7d5 c3d5 f6e5 g3h4 g7f8 g1h1 a8c8 f2f4 e5e4');
INSERT INTO evaluations VALUES(34598,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(34598,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(34598,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(34598,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(34598,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(34598,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(34598,36,'cp 52','e7e5 b1c3 d5d6 e2e4 b8c6 d2d4 c6d4 f3d4 d6d4 d1d4 e5d4 c3b5 f8d6 b5d6 c7d6 c1f4 g8f6 a1d1 d6d5 d1d4 f6e4 f2f3 e4f6 f4d6 c8e6 e1f2 e8d7 d6b4 h8c8 f1b5 d7d8 b4c3 h7h5 d4d2 d8e7 c3b4 e7d8 h1e1 a7a6 b4a5 d8e7 b5a4 e7f8 a4b3 c8c6 a5b4 f8g8 b4c3 a8d8 h2h3');
INSERT INTO evaluations VALUES(34625,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(34625,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(34625,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(34625,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(34625,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(34625,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(34625,36,'cp 121','g1f3 g8f6 d2d4 g7g6 c1g5 f5e6 f3e5 c7c6 e2d2 f6d5 f1e2 b8d7 e2g4 e6d6 c3e4 d6c7 g4d7 c8d7 c2c4 d7f5 e4g3 f7f6 c4d5 f6e5 d5c6 c7c6 e1g1 h7h6 a1c1 c6d7 g5e3 f8g7 d4d5 e8f7 b2b3 a8d8 f1d1 h8e8 g3f5 g6f5');
INSERT INTO evaluations VALUES(34962,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(34962,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(34962,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(34962,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(34962,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(34962,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(34962,36,'cp 43','b8c6 d2d4 c6d4 f3d4 e5d4 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 b5d4 g8f6 e2f3 c7c6 a1e1 c8d7 c2c4 h8e8 e1e8 f6e8 d2c3 e8d6 a2a4 d6f5 d4f5 d7f5 a4a5 d8c7 h1e1 a8d8 e1e7 d8d7 e7d7 f5d7 c4c5 f7f6 c3d4 h7h6 b2b4 d7f5 f3e4 f5d7 g2g3 d7g4 e4d3 g4e6 h2h4');
INSERT INTO evaluations VALUES(1737,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(1737,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(1737,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(1737,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(1737,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(1737,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(1737,36,'cp 36','d2d4 c6d4 f3d4 e5d4 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 b5d4 g8f6 e2f3 c7c6 a1e1 h8e8 c2c4 c8d7 e1e8 f6e8 d2c3 a7a5 d4b3 e8d6 a2a4 d7e6 c4c5 d6f5 f3g4 g7g6 g4f5 e6f5 h1e1 f5e6 b3d2 a8c8 d2c4 e6c4 c3c4 c8c7 b2b4 a5b4 c4b4 c7d7 b4a5');
INSERT INTO evaluations VALUES(39035,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(39035,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(39035,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(39035,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(39035,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(39035,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(39035,36,'cp 110','h2h4 d7d5 b1d2 d5e4 d2e4 h7h6 f1d3 g8f6 e4f6 d8f6 g1f3 b8c6 e1g1 e6e5 f3e5 c6e5 d4e5 f6h4 f2f4 e8g8 d1f3 f8d8 a2a4 c8e6 a4a5 h6h5 d3e4 c7c6 a5a6 b7a6 a1a6 a8b8 a6c6 h4g4 c6c7 g4f3 f1f3 b8b2 e3a7');
INSERT INTO evaluations VALUES(39119,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(39119,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(39119,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(39119,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(39119,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(39119,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(39119,36,'cp 36','g8f6 f1b5 a7a6 e4e5 e8g8 e5d6 a6b5 e1g1 b8c6 d4d3 b5b4 c1f4 c7d6 f4d6 f8e8 b1d2 e8e6 a2a3 b4a3 f1d1 h7h6 a1a3 a8a3 d6a3 d7d5 d2f1 d8a5 d3d2 a5d2 d1d2 e6e8 h2h3 c8e6 f3d4 c6d4 d2d4 e6f5 f1e3 f5g6 a3b4 e8e4 d4e4 f6e4 f2f3');
INSERT INTO evaluations VALUES(39591,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39591,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(39591,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(39591,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(39591,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(39591,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39591,36,'cp -5','g8f6 f1e2 b7b6 g1f3 c7c5 e1g1 e8g8 b2b4 d7d6 c1b2 b8d7 a2a3 c8b7 c3c4 a8c8 d1b3 c5d4 b2d4 c8c7 h2h3 e7e6 f1c1 f6e4 b1d2 e4d2 f3d2 d8g5 d2f3 b7f3 e2f3 g7d4 e3d4');
INSERT INTO evaluations VALUES(40613,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(40613,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(40613,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40613,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(40613,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(40613,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(40613,36,'cp -16','c2c4 e8g8 f1e2 d5c4 e1g1 b8c6 e2c4 f6h5 d1d8 f8d8 f4c7 d8d7 c7e5 b7b6 b1c3 c8b7 f1d1 a8d8 d1d7 d8d7 c3e4 c5e7 e5c3 d7c7 a1d1 h5f6 e4f6 e7f6 c3f6 g7f6 c4e2 c6b4 a2a3 b4d5 f3d4 f6f5 e2f3 g8g7 g2g4 f5g4');
INSERT INTO evaluations VALUES(40945,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(40945,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(40945,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40945,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(40945,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(40945,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(40945,36,'cp 17','e8g8 g1f3 b7b6 f1d3 d5c4 d3c4 f6d5 c3d5 e6d5 c4d3 e7b4 f3d2 c8a6 d3a6 b8a6 e1g1 b4e7 d1e2 d8c8 a2a3 c7c5 a1d1 f8d8 d2b1 c5d4 e3d4 e7d6 f4d6 d8d6 d1c1 a6c7 e2c2 d6d7 b1d2 c8b7 d2f3 c7e6 c2b3 g7g6 g2g3 a8d8 f1d1 d7c7');
INSERT INTO evaluations VALUES(36311,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(36311,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36311,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(36311,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(36311,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(36311,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(36311,36,'cp 36','g1f3 b7b6 f1d3 c8a6 c4d5 f6d5 c3d5 d8d5 e1g1 d5b7 a1c1 c7c5 f4g5 f7f6 g5f4 a6d3 d1d3 b8d7 f4g3 f8d8 d3c4 d7f8 d4c5 b6c5 b2b3 b7d5 c4e2 d5d3 e2d3 d8d3 f1d1 d3d1 c1d1 g8f7 g1f1 e6e5 f3d2 a8d8 f1e2 f8e6 d2c4');
INSERT INTO evaluations VALUES(42468,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(42468,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(42468,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(42468,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(42468,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(42468,35,'cp 195',NULL);
INSERT INTO evaluations VALUES(42468,36,'cp 198','e4e5 f6g8 d1g4 g7g6 a2a3 b4c3 d2c3 b8c6 g4g3 f7f6 g1f3 d7d6 e5f6 d8f6 c4c5 e6e5 c1g5 f6e6 e1c1 g8e7 g3h4 d6d5 f3e5 e7f5 h4f4 h7h6 e5c6 h6g5 f4e5 e8g8 e5e6 c8e6 c6e5 g8g7 c5c6 b7c6 e5c6 g5g4 c6b4 f5e7 b4d3 e7f5 b2b3 a8e8 d3c5 f5h4');
INSERT INTO evaluations VALUES(42438,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(42438,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(42438,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(42438,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(42438,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(42438,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(42438,36,'cp 188','f6g8 d1g4 b4f8 d2d4 b8c6 g1f3 d7d5 f1e2 g8e7 e1g1 e7f5 f1d1 h7h5 g4f4 c6e7 e2d3 c7c6 f4d2 d5c4 d3c4 e7d5 d2e2 a7a5 c3e4 f8e7 h2h3 c8d7 c1g5 a5a4 a1c1 e8g8 a2a3 e7g5 f3g5 f7f6 g5f3 d8e7 c4d3 d5f4 e2d2 f4d3');
INSERT INTO evaluations VALUES(43480,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(43480,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(43480,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(43480,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(43480,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(43480,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(43480,36,'cp -7','c8f5 e4f5 d6d5 c1g5 f8b4 e1g1 b4c3 b2c3 e8g8 c3c4 d5c4 d3c4 h7h6 d1d8 f8d8 g5f6 g7f6 a2a4 b8c6 c4d3 a8c8 f1b1 c8c7 b1b6 c6d4 b6f6 d4c2 d3c2 g8g7 f6b6 c7c2 b6b7 c2a2 a1f1 d8d4 b7a7 a2a4 h2h3 h6h5 f1c1 h5h4 c1c7 d4d1 g1h2');
INSERT INTO evaluations VALUES(28537,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(28537,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(28537,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(28537,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(28537,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(28537,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(28537,36,'cp -5','e4f5 d6d5 c1g5 f8b4 e1g1 b4c3 b2c3 e8g8 c3c4 d5c4 d3c4 h7h6 d1d8 f8d8 g5f6 g7f6 a2a4 b8c6 f1b1 d8d4 c4d3 c6b4 c2c3 d4d3 b1b4 a8d8 h2h3 d3c3 b4b7 c3c2 b7b6 h6h5 b6a6 d8d2 a1f1 g8g7 a4a5 c2a2 a6a8 e5e4 a8e8 a2a5 e8e4');
INSERT INTO evaluations VALUES(28533,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(28533,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(28533,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(28533,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(28533,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(28533,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(28533,36,'cp -14','d6d5 c1g5 f8b4 e1g1 b4c3 b2c3 e8g8 c3c4 e5e4 d3e2 d5d4 f2f3 e4e3 c2c3 b8c6 c3d4 d8d4 d1d4 c6d4 e2d3 a8d8 g5e3 d4f3 f1f3 d8d3 e3g5 d3f3 g2f3 f8c8 a1b1 c8c4 b1b7 h7h6 g5f6 g7f6 b7b6 c4a4 b6b2 g8g7 g1f2 a4a5 h2h4 a5f5 b2b4 f5b5 b4a4 a6a5 f2e2 g7f8 e2f2 h6h5 f2e3 b5e5 e3f4');
INSERT INTO evaluations VALUES(45031,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(45031,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(45031,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(45031,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(45031,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(45031,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(45031,36,'cp -123','f3f2 g7g6 g1f3 f8g7 f1c4 e8g8 c4b3 b7b5 e1g1 b8d7 h2h3 c8b7 a2a3 a8c8 c1e3 e7e6 f3g5 d8e7 f2h4 c8e8 a1e1 h7h6 g5f3 f6h5 h4e7 e8e7 e3f2 f8d8 c3e4 c6c5 e4c5 d7c5 d4c5 h5f4 f2e3 g6g5 a3a4 b5b4 e3f4');
INSERT INTO evaluations VALUES(45447,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45447,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(45447,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(45447,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(45447,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(45447,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(45447,36,'cp 22','g1f3 b8c6 c1f4 c8f5 d1b3 d8d7 b1d2 e7e6 f3e5 d7c8 h2h3 a7a6 g2g4 f5g6 e5g6 h7g6 f1e2 b7b5 h3h4 c8d7 a2a3 f8d6 f4e3 d7c7 e1c1 c6a5 b3c2 a5c4 c1b1 e8f8 h4h5 g6h5 g4h5 d6f4 d2c4 b5c4 d1g1');
INSERT INTO evaluations VALUES(45660,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45660,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(45660,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(45660,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(45660,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45660,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(45660,36,'cp 15','a2a3 b8d7 f1d3 a7a6 e3e4 d5e4 c3e4 c6c5 g1f3 b7b5 e4f6 d7f6 d4c5 f8c5 b2b4 c5e7 e1g1 c8b7 d1e2 e8g8 c2c4 b5c4 d3c4 d8b6 f1d1 f8d8 f3e5 a8c8 a1c1 d8d1 c1d1 f6d5 e2h5 g7g6 h5g4 d5f6 g4g3 b6c7 c4d3 f6d5');
INSERT INTO evaluations VALUES(45673,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(45673,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45673,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(45673,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(45673,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(45673,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(45673,36,'cp 42','c6c5 f1e2 f8d6 d4c5 d6f4 e3f4 d8a5 e1g1 a5c5 c2c3 a7a5 f1e1 e8g8 e2d3 a5a4 a2a3 c8d7 a1b1 f8d8 d1e2 b8a6 f3e5 c5b6 d2f3 a6c5 e2e3 d7b5 f4f5 b5d3 e5d3 d5d4 c3d4 c5d3 e3d3 e6f5 d3f5');
INSERT INTO evaluations VALUES(48569,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(48569,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(48569,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(48569,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(48569,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(48569,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48569,36,'cp -14','f8b4 b1c3 f6d5 c1d2 e8g8 g1f3 b4c3 b2c3 f8e8 f1e2 d8e7 c3c4 c8f5 c4d5 f5c2 d2g5 c2d1 g5e7 d1e2 e1e2 e8e7 e2d3 b8d7 h1e1 a8e8 e1e7 e8e7 a1b1 b7b6 b1c1 d7f6 d5d6 c7d6 c1c8 f6e8 f3d2 f7f5 d2c4 g8f7 d4d5 h7h5 a2a4 g7g6 c8c6 e7d7 f2f3 h5h4 h2h3 f7f6');
INSERT INTO evaluations VALUES(50163,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(50163,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50163,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(50163,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50163,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(50163,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50163,36,'cp 0','f6e4 c4f7 e8f7 c3e4 f8g7 d2d4 d7d5 d1f3 f7g8 e4c3 b8c6 c1e3 c6d4 e3d4 g7d4 e1c1 e7e5 f3d5 d8d5 c3d5 d4f2 g1f3 e5e4 f3g5 e4e3 d5c7 e3e2 d1d8 g8g7 c7e8 g7f8 e8d6 f8g7');
INSERT INTO evaluations VALUES(52168,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(52168,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(52168,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(52168,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(52168,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(52168,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(52168,36,'cp -13','g8f6 d5d3 d7d5 e4d5 f6d5 b1c3 d8e7 c1e3 d5e3 f2e3 c5c4 d3c4 e7e3 c4e2 e3e2 f1e2 f8b4 e1c1 b4c3 b2c3 e8g8 e2f3 b8c6 g1e2 c6e5 f3d5 e5g4 d1d2 g4e3 e2f4 e3d5 f4d5 c8e6 d5c7 a8c8 c7e6 f7e6 h1e1 f8f6 d2d7 f6f7 e1d1 c8c3 d7d8 f7f8 d8f8 g8f8 d1d7 e6e5 d7b7');
INSERT INTO evaluations VALUES(44346,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(44346,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(44346,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(44346,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(44346,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(44346,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(44346,36,'cp -12','d5d3 d7d5 e4d5 f6d5 b1c3 d8e7 c1e3 d5e3 f2e3 c5c4 d3c4 e7e3 c4e2 e3e2 g1e2 c8d7 e2d4 b8c6 d4c6 b7c6 e1c1 e8c8 c3e4 d7e8 d1d8 c8d8 f1d3 f8e7 h1f1 f7f6 a2a3 e8d7 b2b4 h7h5 c2c4 h5h4 c1c2 d8c7 c4c5 h4h3 g2g3 h8d8');
INSERT INTO evaluations VALUES(52277,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52277,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52277,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52277,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52277,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52277,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(52277,36,'cp -12','f8g7 b1c3 b8c6 g1e2 a7a6 a2a4 d7d6 g2g3 c8g4 f1g2 g4e2 c3e2 g8f6 e1g1 e8g8 f2f4 a8b8 c1d2 f6d7 d2c3 d8c7 d1b3 b7b6 h2h3 e7e6 g1h1 f8c8 f4f5 d7e5 a1b1 c7e7 h1h2 c8f8 h2h1 e5d7 b3c2 c6d4 e2d4 c5d4');
INSERT INTO evaluations VALUES(37778,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37778,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37778,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37778,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37778,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37778,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37778,36,'cp -17','b1c3 b8c6 g2g3 a7a6 a2a4 e7e6 f1g2 d7d6 g1e2 g8e7 e1g1 e8g8 a1b1 c8d7 h2h3 a8b8 c1e3 d8a5 g1h2 c6d4 e3f4 d4e2 d1e2 a5c7 e2d2 e7c6 f4e3 g8h8 c3e2 b7b5 c4b5 a6b5 a4b5 b8b5 d3d4 c5d4 e2d4 c6d4 e3d4 g7d4 d2d4 h8g8 b2b4 f8b8 b1d1 b5b4 d4d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(37776,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(37776,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(37776,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(37776,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(37776,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(37776,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(37776,36,'cp -6','b8c6 g2g3 a7a6 a2a4 d7d6 f1g2 e7e6 g1e2 g8e7 e1g1 e8g8 a1b1 c8d7 c1f4 d8c7 d1d2 a8b8 f4g5 b7b6 b2b3 b8e8 g5h6 g7h6 d2h6 e6e5 h6d2 f7f5 f2f4 c6b4 c3d5 e7d5 c4d5 f8f7 e2c3 e8f8 f4e5 d6e5 e4f5');
INSERT INTO evaluations VALUES(56970,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(56970,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(56970,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(56970,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(56970,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(56970,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(56970,36,'cp 11','d8e7 d2d3 g8f6 h5e2 b8c6 c1e3 g7g6 g1f3 f8g7 b1c3 d7d6 e1c1 c6a5 a2a3 a5c4 d3c4 c8e6 h2h3 a7a6 g2g4 h6h5 f3g5 h5g4 g5e6 e7e6 c3d5 e8c8 d5f6 g7f6 h3g4 h8h1 d1h1 d8h8 h1h8 f6h8 b2b3 h8f6 e2f3 f6d8 c1d2 e6f6 f3g2 c8d7 g4g5 f6e6 d2c3 d8e7 g2f3 e7d8');
INSERT INTO evaluations VALUES(29,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(29,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(29,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(29,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(29,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(29,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(29,36,'cp -138','g4f3 g2f3 g8e7 d2d3 e7g6 b1d2 f8d6 h2h3 g6h4 b2b3 h8e8 c1b2 d6f4 e1e4 e8e4 d2e4 f7f6 e4f6 g7f6 b2f6 h4f3 f6d8 c8d8 f1g2 f3d4 c2c3 d4e6 a1d1 c6c5 g2f3 d8e7 f3e4 e7f6 h3h4 f4d6 d1e1 e6f4 e1e3 f4g2 e3f3 f6g6 f3h3 g6h5 d3d4 c5d4 c3d4 b7b6 h3h1');
INSERT INTO evaluations VALUES(32,30,'cp -131',NULL);
INSERT INTO evaluations VALUES(32,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(32,32,'cp -141',NULL);
INSERT INTO evaluations VALUES(32,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(32,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(32,35,'cp -128',NULL);
INSERT INTO evaluations VALUES(32,36,'cp -119','g2f3 g8e7 d2d3 e7g6 b1d2 f8d6 h2h3 g6h4 b2b3 h8e8 c1b2 d6f4 e1e4 e8e4 d2e4 f7f6 e4f6 g7f6 b2f6 h4f3 f6d8 c8d8 d3d4 d8e7 c2c3 e7f6 f1e2 f3h4 e2d3 f6f5 a1e1 h4g6 e1e8 h7h5 d3e2 f4d6 e2f3 b7b6 e8c8 g6f4 c3c4 f4e6 c8h8 f5g5 f3e4 g5h4 c4c5');
INSERT INTO evaluations VALUES(31,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(31,31,'cp -138',NULL);
INSERT INTO evaluations VALUES(31,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(31,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(31,34,'cp -124',NULL);
INSERT INTO evaluations VALUES(31,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(31,36,'cp -148','g8e7 d2d3 d8d5 b1c3 d5h5 c1f4 e7g6 f4g3 f8d6 c3e2 h8d8 a2a3 h5c5 a1c1 d6e7 b2b4 c5f5 e2f4 e7g5 f4g6 h7g6 f3f4 g5f4 e1e7 f4d6 g3d6 d8d6 c1e1 d6f6 e1e2 c6c5 h2h4 b7b6 f1g1 g6g5 e7e8 c8b7 h4h5 b7c6 e2e7');
INSERT INTO evaluations VALUES(1533,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(1533,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(1533,32,'cp -143',NULL);
INSERT INTO evaluations VALUES(1533,33,'cp -147',NULL);
INSERT INTO evaluations VALUES(1533,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(1533,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(1533,36,'cp -161','g8e7 a1c1 c3a3 d2b3 e7f5 e3d2 f8e7 f1d3 a7a5 d3f5 e6f5 f3d4 a3a4 d2c3 e8g8 e1g1 c8e6 d4c6 b7c6 b3d4 a4d1 f1d1 e7c5 d4c6 f8c8 c3d4 c5f8 c1c3 h7h6 d1c1 g8h7 h2h3 a5a4 d4b6 f5f4 c6d4 c8c3 c1c3');
INSERT INTO evaluations VALUES(4283,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4283,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4283,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4283,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4283,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4283,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4283,36,'cp 0','c6a5 f3e5 d8e7 e5g4 a5c4 g5f6 g7f6 d3c4 e7e4 d1e2 e4g2 e1d2 f6f5 e2e5 f5g4 e5g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(5358,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(5358,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(5358,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(5358,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(5358,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(5358,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(5358,36,'cp -4','f1g2 g8f6 b1c3 e8g8 e1g1 a8b8 c3d5 d7d6 a2a4 f6d5 e4d5 c6b4 f3e1 e7e6 d5e6 c8e6 c2c3 b4c6 e1f3 e6f5 f1e1 c6e5 f3e5 d6e5 g2e4 f5h3 a4a5 f7f5 e4h1 f5f4 h1e4 g8h8 f2f3 d8c7 d1e2 b7b6 a5b6 a7b6 e2f2 h3e6');
INSERT INTO evaluations VALUES(5875,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(5875,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(5875,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(5875,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(5875,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(5875,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(5875,36,'cp -16','g8f6 f1e2 h7h6 a2a3 b4c3 d2c3 f6e4 e1g1 e8g8 c3d2 d8f6 c2c3 f8e8 d2e3 c8g4 h2h3 g4h5 d1b3 f6d6 a1e1 h5f3 e2f3 c6a5 b3c2 e8e6 a3a4 a8e8 b2b3 d6c6 e1c1 c6b6 f3e4 e6e4 f2f3');
INSERT INTO evaluations VALUES(9114,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(9114,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9114,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(9114,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(9114,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(9114,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(9114,36,'cp -20','f1e2 e8g8 e1g1 h7h6 h2h3 f8e8 a2a3 b4c3 d2c3 f6e4 f1e1 d8f6 e2b5 a7a6 b5c6 f6c6 e1e3 c8f5 f3e5 c6e6 d1h5 f5h7 a1e1 f7f6 e5d3 e6f7 h5d1 b7b6 c3d2 a6a5 h3h4 e4d6 e3e8 a8e8 d2f4 e8e1 d1e1 h7d3 f4d6 d3c2 e1c1 c7d6 c1c2 f7d7 b2b4 g8h8 g1h2 a5a4 h4h5 b6b5');
INSERT INTO evaluations VALUES(6948,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(6948,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(6948,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(6948,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(6948,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(6948,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(6948,36,'cp 210','d5c5 d7d6 c5c3 h7h6 c1b2 g8f6 c3b3 d8e7 f1e1 e7e6 b3c2 e8f7 b1c3 e6c4 a1c1 c8g4 f3d2 c4c5 c2b3 g4e6 c3d5 e6d5 e4d5 c5d5 b3b7 h8e8 b7c7 f7f8 b2f6 g7f6 e1e8 a8e8 d2f1 c6e7 f1g3 f8f7 h2h3 f7f8');
INSERT INTO evaluations VALUES(6942,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(6942,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(6942,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(6942,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(6942,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(6942,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(6942,36,'cp 174','d7d6 c5c3 h7h6 c1b2 g8f6 c3b3 d8e7 f1e1 e7f7 b3a3 c8e6 e4e5 d6e5 b2e5 a8b8 b1c3 f7e7 a3b2 e8f7 c3e4 e6d5 e5f6 g7f6 e4g5 h6g5 e1e7 c6e7 a1e1 b8e8 b2d4 e8d8 d4a7 e7g6 a7c5 d8d7 c5d4 h8d8 d4c3 c7c6 a2a4 g6f4 a4a5 d5f3 c3f3 f4d3 e1e3 d3e5 f3h5 f7g7 h2h3');
INSERT INTO evaluations VALUES(6114,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(6114,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(6114,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(6114,33,'cp 236',NULL);
INSERT INTO evaluations VALUES(6114,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(6114,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(6114,36,'cp 207','c5c3 h7h6 c1b2 g8f6 c3b3 a7a5 b1c3 a5a4 c3a4 d8d7 f1e1 d7f7 b3c2 f7g6 f3h4 g6h5 h4f5 e8d8 c2b3 a8a5 f5g7 h5g5 g7f5 h8f8 f2f4 g5h5 f5g3 h5f7 f4f5 f7b3 a2b3 c6e5 a4c3 a5a1 b2a1 c8d7 c3d1 e5d3 e1f1 d3c5 b3b4 c5d3 a1c3 d6d5 c3f6 f8f6');
INSERT INTO evaluations VALUES(6116,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(6116,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(6116,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(6116,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(6116,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(6116,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(6116,36,'cp 162','h7h6 c1b2 g8f6 c3b3 a7a5 b1c3 a5a4 c3a4 d8e7 f1e1 e7f7 b3c2 f7g6 f3h4 g6h5 h4f5 e8d8 c2b3 h5g6 a4b6 a8a5 b6c8 d8c8 f2f3 h8e8 a1b1 f6d7 b2g7 d7c5 b3c2 b7b6 g7h6 g6g8 e1e2 d6d5 h2h4 d5e4 f3e4 g8g4 h6e3 e8g8 e3c5 a5c5');
INSERT INTO evaluations VALUES(12208,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12208,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(12208,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(12208,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(12208,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(12208,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(12208,36,'cp 20','d8b6 d1b3 f8e7 b3b6 a7b6 f1b5 e8g8 e1e2 f6h5 f4c7 h5f6 c7g3 d7b8 a2a4 c8d7 f3e5 d7b5 a4b5 b8d7 c3c4 f8c8 h1c1 h7h5 c4d5 e6d5 h2h3 d7e5 d4e5 f6e8 e3e4 d5e4 d2c4 e7d8 e2e3 f7f5 e5f6 e8f6 a1a8 c8a8');
INSERT INTO evaluations VALUES(14084,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(14084,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(14084,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(14084,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(14084,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(14084,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(14084,36,'cp 25','g7g6 c4b5 c8d7 c1g5 f8g7 d1d3 f6h5 e1c1 a7a6 b5d7 d8d7 h2h3 h7h6 g5d2 f7f5 g2g4 f5e4 c3e4 h5f6 e4f6 g7f6 d1e1 e8c8 c2c4 c8b8 c1b1 c7c6 d2a5 d8f8 d5c6 e7c6 a5c3 h6h5 f3d2 e5e4 d2e4 f6c3 d3c3');
INSERT INTO evaluations VALUES(14110,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(14110,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(14110,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(14110,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(14110,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(14110,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(14110,36,'cp 68','h2h3 f8g7 g1f3 e8g8 e1g1 c6b4 b1c3 b7b6 a2a3 b4d3 c2d3 c7c5 d4c5 b6c5 f4e5 f6h5 d3d4 c5c4 b2b3 c8a6 e5g7 h5g7 b3b4 f7f6 d1c2 a6b7 c3a4 d8d6 f1c1 f8c8 a4c5 b7c6 c2c3 g7e6 c1c2 a7a6 a3a4 e6c5 d4c5');
INSERT INTO evaluations VALUES(14311,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(14311,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(14311,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(14311,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(14311,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(14311,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(14311,36,'cp 25','c8f5 e2e3 f6e4 c2c4 e7e6 b2b4 f8e7 c1b2 e8g8 b1c3 e4c3 b2c3 a7a5 b4b5 c6a7 c4c5 b7b6 c5c6 f7f6 f1e2 a7c8 e1g1 c8d6 c3e1 h7h6 f3h4 f5h7 a3a4 d6e4 e2d3 e7d6 f2f3 e4g3 e1g3 d6g3 h4g6 h7g6 d3g6 f6f5 f3f4 d8h4');
INSERT INTO evaluations VALUES(14642,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(14642,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(14642,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(14642,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(14642,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(14642,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14642,36,'cp 70','d2d3 e5f4 c1f4 c6a5 d1e1 a5c4 d3c4 f6h5 f4e3 f8e7 b1c3 c7c6 a1d1 e8g8 f3d4 c8g4 d1d3 d8d7 h2h3 g4e6 b2b3 a8e8 e1d2 f7f5 e4f5 e6f5 d4f5 f8f5 f1f5 d7f5 e3a7 f5g6 a7e3 h5g3 e3f4 g3h5 f4h2 e7f6');
INSERT INTO evaluations VALUES(16429,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(16429,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(16429,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(16429,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(16429,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(16429,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(16429,36,'cp -28','f8c5 h2h3 d7d6 d2d3 e8g8 g1e2 a7a6 f1g2 b7b5 b2b3 c6d4 e1g1 a8b8 g1h2 f8e8 c1d2 b5c4 b3c4 c8d7 f2f4 c7c6 e2d4 c5d4 a1c1 e5f4 d2f4 c6c5 c1c2 d7e6 c3e2 d4e5 f4e5 d6e5 e2c3 f6d7 c3d5 h7h6 d1d2');
INSERT INTO evaluations VALUES(10158,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(10158,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(10158,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(10158,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(10158,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(10158,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(10158,36,'cp -12','h2h3 a7a6 d2d3 e8g8 f1g2 b7b5 g1e2 d7d6 e1g1 a8b8 b2b3 c6d4 g1h2 h7h6 c1b2 c7c6 d1d2 d8a5 f2f4 d4e2 d2e2 b5c4 d3c4 c5d4 e2d2 c6c5 f4e5 d6e5 a1d1 c8b7 c3d5 a5a2 b2d4 a2d2 d5f6 g7f6 d1d2 e5d4 f1f6 f8e8 f6f5 b7e4 f5c5 b8b3 g2e4 e8e4 c5d5 e4e1 d5d8 g8h7');
INSERT INTO evaluations VALUES(16587,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(16587,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(16587,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(16587,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(16587,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(16587,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(16587,36,'cp -52','d7d5 b1d2 a7a5 f1e2 c8e6 g1f3 f8c5 f3g5 e6c8 e1g1 e8g8 e4d5 f6d5 g5f3 h7h6 d2e4 c5a7 f1e1 c8f5 c1d2 f5g6 c2c4 d5f4 d2f4 e5f4 e4c3 c6d4 c3b5 d4b5 c4b5 d8f6 d1c1 a7b6 c1c3 a8d8 c3f6 g7f6 d3d4 g6e4 a1c1 g8g7 g1f1 f6f5 e1d1 f8e8 c1c4 e4d5');
INSERT INTO evaluations VALUES(15166,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(15166,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(15166,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(15166,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(15166,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(15166,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(15166,36,'cp -43','b1d2 f8c5 g1f3 a7a5 f1e2 h7h6 e1g1 e8g8 e4d5 f6d5 d2e4 c5b6 c2c4 d5f6 e4g3 d8d6 c1e3 b6e3 f2e3 e5e4 g3e4 f6e4 d3e4 d6g6 g1h1 g6e4 d1d2 e4e7 f3d4 c6e5 d2c3 c8d7 c4c5 f8e8 a1d1 c7c6 e2f3 e7g5 b2b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(18645,30,'cp 275',NULL);
INSERT INTO evaluations VALUES(18645,31,'cp 269',NULL);
INSERT INTO evaluations VALUES(18645,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(18645,33,'cp 287',NULL);
INSERT INTO evaluations VALUES(18645,34,'cp 278',NULL);
INSERT INTO evaluations VALUES(18645,35,'cp 293',NULL);
INSERT INTO evaluations VALUES(18645,36,'cp 263','d4b5 a7a6 b5d6 f8d6 d1d6 d8e7 c1f4 g8f6 f1e2 e7d6 f4d6 e8d8 d6f4 f6e8 e4e5 h7h6 e1c1 f7f6 e5f6 e8f6 f4d6 b7b5 e2f3 c8b7 h1e1 d8e8 a2a4 e8f7 e1e3 h8c8 a4b5 a6b5 c3b5 a8a1 c1d2 a1a5 c2c4 b7a6 e3b3 a6b5 b3b5 a5b5 c4b5 c6d4 b5b6 d4f3 g2f3 f6d5 b6b7');
INSERT INTO evaluations VALUES(20541,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(20541,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(20541,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(20541,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(20541,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(20541,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(20541,36,'cp -66','d8e7 d1e2 e7e2 e1e2 c8g4 a2d5 c6d4 e2f1 d4c2 a1a2 e8c8 d5f7 g4f5 f7c4 a7a6 c4e2 g8e7 b2b3 c2d4 f3d4 d8d4 b1c3 e7c6 f2f3 f5e6 e2d1 c6e5 d1c2 e5d3 c2d3 d4d3 f1e2 d3d4 a2b2 c5c4 b3b4 e6f5 e2d1 f8d6 c3e4 f5e4 f3e4 d4e4 b2c2 b7b5 c1b2');
INSERT INTO evaluations VALUES(24441,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(24441,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24441,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(24441,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(24441,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(24441,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(24441,36,'cp 17','d4c6 d7c6 b5c6 b7c6 c1f4 f8e7 e4e5 d6e5 f4e5 e8g8 d1e2 d8a5 a1d1 f8d8 h2h3 h7h6 f1e1 f6d7 e5d6 e7d6 d1d6 a8b8 e1d1 d7f6 d6d8 b8d8 d1d8 a5d8 a2a3 d8b6 c3d1 b6c5 c2c3 c5g5 d1e3 f6d5 e3c4 g5f5 g1h2 d5b6 c4b6 a7b6 a3a4 f5d5 e2c2');
INSERT INTO evaluations VALUES(33857,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(33857,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(33857,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(33857,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(33857,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(33857,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(33857,36,'cp 166','c4d5 e6d5 c3d5 f6f5 d5c3 f8b4 d1a4 b8c6 e2e3 e8g8 g1e2 c8d7 a2a3 b4c3 e2c3 f8d8 a1d1 f5g6 d1d2 a8c8 a4b3 b7b6 b3d1 c6a5 b2b4 g6c6 d1c2 a5b7 e3e4 a7a5 f1e2 a5b4 a3b4 c6d6 b4b5 c7c5 d4d5 d6e5');
INSERT INTO evaluations VALUES(33855,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(33855,31,'cp 169',NULL);
INSERT INTO evaluations VALUES(33855,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(33855,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(33855,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(33855,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(33855,36,'cp 160','e6d5 c3d5 f6f5 d5c3 f8b4 e2e3 e8g8 a2a3 b4c3 b2c3 f5a5 d1d2 c7c5 g1f3 f8d8 h2h3 b8d7 f1e2 d7b6 e1g1 c8e6 f1c1 a8c8 d2b2 c8c7 e3e4 c5d4 c3d4 b6a4 b2b4 a5b4 a3b4 a4c3 g1f1 d8c8 d4d5 e6d7 e2d3 c3d5 c1c7 d5c7 a1a7 d7c6');
INSERT INTO evaluations VALUES(33864,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(33864,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(33864,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(33864,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(33864,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(33864,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(33864,36,'cp 158','c3d5 f6f5 d5c3 f8b4 e2e3 e8g8 a2a3 b4c3 b2c3 f5a5 d1d2 c7c5 f1d3 b8c6 g1e2 c8e6 e1g1 c5d4 e3d4 a8c8 f1b1 a5c7 e2f4 e6d7 f4d5 c7d6 d5e3 b7b6 d2b2 f8d8 b1d1 c6e7 a3a4 d6f6 d3a6 c8b8 c3c4 e7f5 e3d5 f6d6 a6b5 d7e6');
INSERT INTO evaluations VALUES(34003,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(34003,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(34003,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(34003,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(34003,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(34003,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(34003,36,'cp 16','e8d8 d2d4 d7d6 e1g1 d6e5 d4e5 c8d7 f7d5 e7e5 f1e1 e4f2 d1d2 e5f6 d2f2 f6f2 g1f2 d8c8 c1f4 a7a5 b1c3 a8a6 f2g1 a6b6 d5b3 b6g6 a1d1 f8c5 f4e3 c5e3 e1e3 h8d8 c3d5 b8c6 d5f4 g6f6 f4h5 f6g6 c2c3 d8f8 h5g3 g6g5 d1e1 b7b6 h2h3 h7h6 b3c2 g5c5 c2d3');
INSERT INTO evaluations VALUES(35434,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(35434,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(35434,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35434,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35434,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(35434,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(35434,36,'cp 7','d2d4 d7d6 e1g1 d6e5 d4e5 c8d7 f7d5 e7e5 f1e1 e4f2 e1e5 f2d1 d5b7 d7c6 b7c6 b8c6 e5d5 d8c8 d5d1 f8c5 g1h1 a8b8 b2b3 c5d4 c2c3 d4f6 c1f4 h8e8 b1a3 f6c3 a1c1 c3e5 c1c6 e5f4 g2g3 c8b7 c6c2 f4d6 a3c4 e8e6 d1c1 e6e7 a2a3 b8e8 h1g2 e7e2 g2h3 e2c2 c1c2 d6c5 c4a5 b7b6 h3g2 b6a5 c2c5 a5b6 c5c2 c7c5');
INSERT INTO evaluations VALUES(35433,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(35433,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(35433,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35433,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(35433,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(35433,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(35433,36,'cp 14','d7d6 e1g1 d6e5 d4e5 c8d7 f7d5 e7e5 f1e1 e4f2 d1d2 e5f6 d2f2 f6f2 g1f2 d8c8 c1f4 a7a5 b1c3 a8a6 g2g3 f8d6 f4d6 a6d6 e1e7 g7g5 a1d1 b8c6 d5c6 d6d1 c3d1 d7c6 d1e3 b7b6 e3g4 c6d5 a2a3 h8f8 f2e3 h7h5 g4e5 f8f5 b2b4 d5g2 b4a5 b6a5 e3d4 f5f2 d4c3 h5h4 g3h4 g5h4');
INSERT INTO evaluations VALUES(35419,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(35419,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(35419,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(35419,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(35419,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(35419,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(35419,36,'cp 12','e1g1 d6e5 d4e5 c8d7 f7d5 e7e5 f1e1 e4f2 d1d2 e5f6 d2f2 f6f2 g1f2 d8c8 b1c3 a7a5 c1f4 a8a6 g2g3 f8d6 f4d6 a6d6 e1e7 g7g6 a1d1 b8c6 d5c6 d6d1 c3d1 d7c6 d1e3 a5a4 e3g4 c8d8 e7g7 h8f8 f2e3 f8e8 e3d3 e8e7 g7e7 d8e7 b2b4 a4b3 c2b3 g6g5 g4e5 c6e8 a2a3 e7e6');
INSERT INTO evaluations VALUES(34631,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(34631,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(34631,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(34631,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(34631,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(34631,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(34631,36,'cp 179','d2d4 g5a5 f1c4 c7c6 g1f3 e7e6 e1g1 f8e7 d1e2 g8f6 c1f4 e8g8 f1e1 a5d8 c4d3 b8d7 h2h3 b7b5 c3e4 f6e4 e2e4 d7f6 e4c6 c8d7 c6a6 d8c8 a6c8 f8c8 c2c3 d7e8 f3e5 f6d5 f4d2 a7a5 a2a3 a5a4 d3e4 e7d6 g2g4 d6e5 d4e5');
INSERT INTO evaluations VALUES(35580,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(35580,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(35580,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(35580,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(35580,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(35580,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(35580,36,'cp 77','g2g4 b8c6 e1c1 c6d4 d2d4 f6d7 h2h4 d7e5 f1e2 b7b5 c1b1 e5c6 d4d2 c8b7 d2e1 a8c8 h4h5 h7h6 h1g1 c6e5 e1d2 b5b4 c3a4 d6d5 a4b6 c8c6 f3f4 c6b6 f4e5 d5d4 d2d4 d8d4 e3d4 b6c6 d4e3 c6c8 e2d3 e7g5 e3g5 h6g5');
INSERT INTO evaluations VALUES(36145,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36145,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36145,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(36145,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(36145,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(36145,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(36145,36,'cp -14','d8c7 e1g1 b8d7 b1d2 b7b6 c3c4 c5d4 e3d4 d6g3 h2g3 c8b7 c4d5 b7d5 a1c1 c7b7 d2c4 a8c8 d1e2 b7b8 f1e1 f8d8 a2a3 c8c7 c1c3 d5b7 e1c1 h7h6 c3c2 d7f8 c4e5 d8c8 c2c7 c8c7 c1c7 b8c7 d3a6 b7a6 e2a6 f6d5');
INSERT INTO evaluations VALUES(37661,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(37661,31,'cp 273',NULL);
INSERT INTO evaluations VALUES(37661,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(37661,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(37661,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(37661,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(37661,36,'cp 242','f3e5 d7d6 d2d4 d6e5 d4c5 d8d1 c3d1 c8e6 f1d3 b8d7 c1e3 g8h6 f2f3 f7f5 b2b3 f5e4 d3e4 h6f5 e3f2 e8g8 e4d3 f5d4 c2c3 d4b5 e1g1 a7a5 f1e1 e6f7 d3f5 a8d8 f2h4 d7f6 e1e5 f7g6 f5g6 h7g6 e5e1 f8e8 e1e8 d8e8');
INSERT INTO evaluations VALUES(39010,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(39010,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(39010,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(39010,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(39010,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(39010,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(39010,36,'cp 21','c7c5 c3b5 d7d6 d2d4 c5d4 b5d4 a7a6 g1f3 g8f6 e1g1 e8g8 b2b3 a8a7 f3g5 b7b6 c1b2 d8d7 a1c1 c8b7 g2b7 d7b7 d4c2 h7h6 g5f3 f8d8 d1d3 b8c6 f1d1 b7a8 d1d2 c6e7 b2f6 g7f6 c2b4 a7d7 e2e4 f6g7');
INSERT INTO evaluations VALUES(39131,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(39131,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(39131,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(39131,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(39131,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(39131,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(39131,36,'cp 145','d2d4 c5f8 g1f3 g8e7 f1d3 b8c6 e1g1 e7g6 f1e1 d7d6 d1e2 f8e7 g2g3 e8g8 h2h4 d6e5 d4e5 f7f6 e5f6 e7f6 b1d2 c8d7 d3e4 e6e5 d2c4 d7g4 c2c3 g8h8 c4e3 g4h5 e3d5 f6e7 c1e3 d8d7 d5e7 c6e7');
INSERT INTO evaluations VALUES(39130,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(39130,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(39130,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(39130,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(39130,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(39130,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(39130,36,'cp 119','c5f8 g1f3 d7d5 c2c3 c7c5 f1d3 a7a6 e1g1 b8c6 f1e1 c8d7 d3c2 a8c8 d4c5 f8c5 b1d2 g8e7 d2b3 c5b6 c1e3 e7g6 e3b6 d8b6 d1d2 e8g8 a1d1 f7f5 h2h4 b6d8 b3c5 c8c7 c5d7 c7d7 h4h5 g6h4 f3g5 d8e7 d2f4 h7h6 g5e6 h4g2 g1g2 e7e6 c2b3 g8h8');
INSERT INTO evaluations VALUES(39187,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(39187,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(39187,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(39187,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(39187,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(39187,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39187,36,'cp 0','c7c6 b5e2 d7d5 f3e5 d5e4 e1g1 g8f6 d2d3 e8g8 d3e4 f6e4 d1d8 f8d8 e2c4 e4f2 e5f7 f2h3 g1h1 h3f2');
INSERT INTO evaluations VALUES(45792,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(45792,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(45792,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(45792,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(45792,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(45792,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(45792,36,'cp -24','e1g1 f8d6 d1e2 d8e7 e2e7 e8e7 f1e1 e7d8 d2d4 h8e8 b1d2 a7a5 c2c3 b8d7 a4c2 a5a4 d2f1 h7h6 e1e8 d8e8 a2a3 d7b6 f3e5 c8e6 c1f4 f6d7 e5d3 d6f4 d3f4 b6c4 f1e3 c4e3 f2e3 d7f6 f4e6 f7e6 a1e1 e8e7 e3e4 f6e4 c2e4 d5e4');
INSERT INTO evaluations VALUES(45791,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(45791,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(45791,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(45791,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(45791,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(45791,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(45791,36,'cp -24','f8d6 d1e2 d8e7 e2e7 e8e7 f1e1 e7d8 d2d4 h8e8 e1e8 d8e8 c2c3 c8f5 a4d1 a7a5 a2a4 b8a6 b1d2 a6c7 d2f1 h7h6 f1e3 f5h7 g2g3 c7e6 h2h4 d6c7 e3g2 e8f8 f3e5 f6e4 d1c2 f8g8 c1e3 f7f6 e5d3 g7g5');
INSERT INTO evaluations VALUES(49671,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(49671,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(49671,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(49671,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(49671,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(49671,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(49671,36,'cp -27','g1f3 b8c6 g2g3 f8c5 f1g2 e8g8 e1g1 f8e8 d2d3 h7h6 b1c3 d7d6 h2h3 a7a5 g1h2 a5a4 c1e3 c5e3 e2e3 c8g4 a1e1 g4f3 e3f3 c6d4 f3d1 c7c6 c3e2 d4e2 e1e2 d6d5 e4d5 c6d5 a2a3 d8d7');
INSERT INTO evaluations VALUES(51709,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(51709,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(51709,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(51709,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(51709,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(51709,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(51709,36,'cp 58','g8f6 f3g5 d6d5 e4d5 h7h6 g5e4 f6e4 c3e4 c6d5 e4c3 c8e6 d2d4 b5b4 c3e2 b8c6 d4e5 f8c5 e1g1 e8g8 c1f4 a8c8 h2h3 c5b6 d1d3 d8h4 c2c3 f8d8 a1d1 c6a5 f4e3 b4c3 b2c3 a5c4 b3c2 h4e4 d3e4 d5e4 d1d8 c8d8 e3b6 d8d2 e2d4 a7b6 c2e4 c4e5 d4e6');
INSERT INTO evaluations VALUES(52845,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(52845,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(52845,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(52845,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52845,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52845,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(52845,36,'cp -11','c8f5 e2e3 e7e6 g1f3 a7a5 f1e2 f8b4 c1d2 b4d2 b1d2 b8a6 e1g1 a6b4 a1c1 g8f6 f3e5 f6d7 d2f3 e8g8 e5d3 d8e7 d3b4 e7b4 b2b3 f5e4 f3d2 e4g6 d2f3 b4e7 e2d3 g6h5 h2h3 h7h6 d1e2 h5f3 e2f3 g7g6 f3e2 g8g7 e2b2');
INSERT INTO evaluations VALUES(53915,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53915,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(53915,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(53915,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(53915,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(53915,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(53915,36,'cp -5','g8f6 d5f6 d8f6 f1c4 b8c6 d2d3 f8e7 c2c3 e8g8 d1e2 a8b8 e1g1 f6g6 f3e1 c8e6 e1c2 b8d8 c4e6 g6e6 c2e3 e7g5 e3d5 g5c1 a1c1 c6e7 d5e7 e6e7 f1e1 f8e8 c3c4 e7d7 e2d2 e8e6 b2b4 b7b6 c1d1 c5b4 d2b4');
INSERT INTO evaluations VALUES(2327,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(2327,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(2327,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(2327,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(2327,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(2327,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(2327,36,'cp 31','e1g1 f8c5 f1d1 c8d7 f3d2 a7a5 d2f1 g8e7 c1e3 c5d4 e3d4 c6d4 b1a3 e7c8 c2c3 d4c6 f1e3 c8d6 c4d3 d7e6 a3c4 d6c4 d3c4 d8e7 c4d5 h8d8 g1f1 e7f7 a2a4 e6d5 d1d5 d8d5 e3d5 a8c8 a1d1 h7h5 f1e2 f7e6');
INSERT INTO evaluations VALUES(2328,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(2328,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(2328,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(2328,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(2328,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(2328,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(2328,36,'cp 39','f8c5 f1d1 c8d7 f3d2 a7a5 d2b3 c5b6 c4e6 c6b8 a2a4 g8e7 b1a3 d8c8 e6d7 b8d7 a3c4 h8d8 g1f1 e7c6 f2f3 d7c5 c4b6 c7b6 d1d8 c6d8 c1e3 d8e6 e3c5 b6c5 b3d2 c8c7 d2c4 b7b6 c2c3 c7c6 a1d1 a8d8 d1d5 c6b7 d5d8 e6d8 f1e2 b7c6 e2d3 g7g6 c4e3 d8f7 e3d5 f6f5 d5e7 c6d6');
INSERT INTO evaluations VALUES(3688,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(3688,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(3688,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(3688,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(3688,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(3688,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(3688,36,'cp 36','f3e5 g7g6 b1d2 c6e5 d4e5 f6d7 d2f3 f7f6 e5f6 d7f6 d1a4 d8e7 a4f4 f6e4 f4h4 e7h4 h1h4 g8g7 f1d3 e4d6 f3e5 d6f7 e5f7 f8f7 e3e4 c7c5 f2f3 h7h6 e1f2 c8d7 a1e1 a8e8 e1h1 e8h8 e4e5 f7f8 h4g4 d7e8 f3f4');
INSERT INTO evaluations VALUES(5049,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(5049,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(5049,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(5049,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(5049,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(5049,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(5049,36,'cp -27','f1d3 g8f6 e1g1 d7d5 e4d5 c6d5 f1e1 e8g8 c1f4 f6d7 f4g5 d7e5 d3f1 h7h6 g5f4 e5c6 b1d2 e7e5 f4g3 e5e4 c3c4 c8f5 d1a4 c6d4 c4d5 d8d5 d2b3 d4b3 a2b3 d5d2 a1d1 d2b2 g3f4 b2f6 f4e3 f6e7 h2h3 g7c3 e3h6 f8d8 d1d8 a8d8 e1c1 c3d4 h6e3 d4e3 f2e3');
INSERT INTO evaluations VALUES(5681,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(5681,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(5681,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(5681,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(5681,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(5681,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(5681,36,'cp 52','e1g1 g8f6 c2c4 d5c4 b1a3 a7a5 a3c4 a5a4 f1e1 c8e6 c4a3 e8g8 e2e4 e6g4 a3c2 d8c8 c2e3 g4f3 g2f3 f8d8 d4d5 e7e6 c1d2 c6e7 d2c3 e6d5 e4d5 e7f5 d1d3 a4a3 b2b3 f5e3 e1e3 c8f5 d3f5 g6f5 a1d1 d8d7 e3e2 f6e4 f3e4 f5e4 c3g7 g8g7 e2e4 a8d8');
INSERT INTO evaluations VALUES(5680,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(5680,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(5680,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(5680,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(5680,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(5680,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(5680,36,'cp 43','g8f6 c2c4 d5c4 b1a3 a7a5 a3c4 a5a4 c1f4 e8g8 a1c1 c8e6 e2e3 f6d5 f3g5 d5f4 g3f4 e6d5 g2d5 d8d5 d1f3 e7e6 g5e4 c6e7 c4e5 c7c6 a2a3 f8d8 b2b4 e7f5 e4c5 b7b6 c5d3 g7e5 d3e5 d5f3 e5f3 a8c8 c1c2 f5e7 f1c1 c8c7 g1g2 g8g7');
INSERT INTO evaluations VALUES(5697,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(5697,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(5697,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(5697,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(5697,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(5697,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(5697,36,'cp 72','d4d5 c6e5 f3e5 g7e5 f1e2 c7c6 a2a4 d8a5 b1c3 g8f6 c1d2 e8g8 e1g1 e7e6 g1h1 a5c7 d5e6 c8e6 f2f3 d6d5 c4d5 c6d5 c3b5 c7d8 f3f4 e5b8 e4e5 f6e4 e2f3 a7a6 b5c3 e4c3 b2c3 b8a7 f4f5 e6f5 d2h6 f8e8 f3d5 e8e5 d5b7');
INSERT INTO evaluations VALUES(7187,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7187,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(7187,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(7187,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(7187,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(7187,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(7187,36,'cp 14','c2c3 e8g8 h2h3 h7h6 f1e1 b7b5 a4c2 f8e8 a2a4 c5a7 b1d2 c8e6 d2f1 d6d5 e4d5 d8d5 b2b4 a8d8 f1g3 a7b6 c1d2 d5d6 d1e2 d6d7 a4a5 b6a7 d2e3 a7e3 e2e3 f6d5 e3d2 f7f6 a1d1 d5f4 g3e2 f4g6 d3d4 d7f7 c2e4 c6e7');
INSERT INTO evaluations VALUES(7185,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(7185,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7185,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(7185,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(7185,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(7185,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7185,36,'cp 8','b7b5 a4c2 e8g8 h2h3 h7h6 a2a4 c5a7 a4b5 a6b5 c1e3 c8d7 d3d4 e5d4 c3d4 c6b4 b1d2 b4c2 d1c2 f8e8 f1e1 a7b6 a1c1 a8a4 b2b3 a4a6 c2d3 d8a8 d4d5 b6e3 e1e3 c7c5 d5c6 d7c6 e3e1 c6e4 d2e4 f6e4 d3b5 e4f6 e1e8 f6e8 f3d4 a6a1 c1a1');
INSERT INTO evaluations VALUES(11854,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(11854,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(11854,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(11854,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(11854,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11854,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11854,36,'cp 37','b1c3 f6d7 e1g1 f8e7 d4e5 d6e5 d1e2 e7d6 c1g5 f7f6 g5e3 d7b6 f3h4 g7g6 f2f4 e5f4 e3f4 e8g8 a1d1 d6c5 g1h1 d8e7 f4g3 c8e6 b2b3 c5b4 g3e1 c6c5 h2h3 a8e8 h4f3 a7a5 e4e5 f6e5 f3e5');
INSERT INTO evaluations VALUES(12760,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(12760,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(12760,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(12760,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(12760,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(12760,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(12760,36,'cp 25','f6d5 b1c3 d5c7 b5c6 d7c6 c3e4 b7b6 d2d3 c8e6 e1g1 g7g6 b2b3 f8g7 c1b2 e8g8 f1e1 d8d7 d1e2 a8d8 a1d1 e6d5 c2c4 d5e4 e2e4 f7f5 e4e2 c7e6 d3d4 c5d4 f3d4 e6d4 b2d4 d7c7 f2f4 e7e6 d4f2 d8d1 e2d1 f8d8 d1e2');
INSERT INTO evaluations VALUES(12909,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(12909,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(12909,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(12909,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(12909,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(12909,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(12909,36,'cp 41','b1c3 d5c7 b5c6 d7c6 c3e4 b7b6 d2d3 c8e6 a2a4 g7g6 b2b3 f8g7 e1g1 e8g8 c1b2 e6d5 f1e1 d8d7 h2h3 a8e8 a4a5 c7e6 e4g5 d5f3 g5f3 b6b5 d1d2 e8d8 e1e4 d7c7 d2e2 d8d5 h3h4 e6d4 b2d4 c5d4 h4h5 a7a6 h5g6 h7g6 a1e1');
INSERT INTO evaluations VALUES(12818,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(12818,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(12818,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(12818,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(12818,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(12818,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12818,36,'cp 0','c5d4 e4e5 f6d5 f1c4 d5b6 c4b3 d7d5 e5d6 d8d6 f3d4 c6d4 c3d4 c8e6 e1g1 e6b3 a2b3 g7g6 b1c3 f8g7 d1f3 e8g8 f3b7 f8b8 b7a6 b6c8 a6a4 c8b6');
INSERT INTO evaluations VALUES(12986,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(12986,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(12986,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(12986,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(12986,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(12986,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(12986,36,'cp 50','c2c4 e7f5 f1d3 f8e7 e1g1 e8g8 a2a4 a7a6 b5c3 c8d7 b2b4 a6a5 b4a5 d8a5 c3b5 a5d8 f1e1 f8e8 a1b1 e7f8 c1d2 g7g6 h2h3 b7b6 b1b4 f5d4 b5d4 e5d4 e1e8 d7e8 d1b3 f8g7 b4b6 e8a4');
INSERT INTO evaluations VALUES(12985,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(12985,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(12985,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(12985,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(12985,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(12985,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(12985,36,'cp 39','e7f5 f1d3 f8e7 e1g1 e8g8 a2a4 f5h4 f2f3 a7a6 b5c3 c8f5 a4a5 f5d3 d1d3 f7f5 g1h1 d8e8 c1d2 a8c8 a1e1 f5f4 c3e4 e8g6 e1e2 h4f5 b2b3 g6h5 b3b4 e7h4 f1b1 f5g3 e4g3 f4g3 h2h3 h4g5 d2g5 h5g5 e2e4');
INSERT INTO evaluations VALUES(16381,30,'cp -508',NULL);
INSERT INTO evaluations VALUES(16381,31,'cp -540',NULL);
INSERT INTO evaluations VALUES(16381,32,'cp -525',NULL);
INSERT INTO evaluations VALUES(16381,33,'cp -549',NULL);
INSERT INTO evaluations VALUES(16381,34,'cp -553',NULL);
INSERT INTO evaluations VALUES(16381,35,'cp -548',NULL);
INSERT INTO evaluations VALUES(16381,36,'cp -565','c6d4 c4f7 e8e7 b3c4 b7b5 c4d3 e7f7 g1f3 d4f3 d3f3 d7d5 d2d3 c8g4 f3g3 d5e4 e1g1 e4d3 c2d3 h7h6 a2a4 b5b4 b1d2 g4e2 d2c4 e5e4 c1f4 e4d3 f1c1 c7c5 c4e5 f7g8 c1d1 d8d5 e5d3 e2d1 a1d1 d5f5');
INSERT INTO evaluations VALUES(17341,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(17341,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(17341,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(17341,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(17341,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(17341,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(17341,36,'cp 74','b5c6 b7c6 b2b3 d8c7 c1b2 g8h6 d2d3 f7f6 e4e5 f6e5 f3e5 d7d6 e5c4 h6f5 c4e3 e6e5 e3f5 c8f5 f2f4 f8e7 f4e5 e8g8 b1d2 d6d5 d1h5 c7d7 h5e2 f5g4 d2f3 a6a5 e5e6 d7d6 a1e1 a5a4 e2e3 a4b3 a2b3 a8a2 b2e5 d6e6');
INSERT INTO evaluations VALUES(17328,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(17328,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(17328,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(17328,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(17328,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(17328,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(17328,36,'cp 60','b7c6 b2b3 g8e7 c1a3 e7g6 d2d4 c5d4 a3f8 h8f8 d1d4 f7f6 c2c4 e6e5 d4e3 e8f7 f1d1 d7d6 b1c3 c8e6 d1d2 d8c7 c4c5 f8d8 a1d1 d6c5 e3c5 d8d2 d1d2 a8c8 h2h4 h7h5 d2d6 g6f8 f3d2 f8d7 c5a3 a6a5 c3d1 c8d8 d1e3 d7f8 d2c4 e6c4 e3c4 d8d6 c4d6 f7g8');
INSERT INTO evaluations VALUES(22503,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(22503,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(22503,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(22503,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(22503,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(22503,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(22503,36,'cp 36','a7a6 b5a4 g8f6 d2d4 e5d4 f3d4 f8b4 d4c6 b4c3 b2c3 d7c6 d1d8 e8d8 f2f3 c8e6 a4b3 h8e8 c1e3 a6a5 e1c1 d8c8 a2a4 b7b6 e3d4 c8b7 h2h4 a8d8 h4h5 c6c5 b3e6 e8e6 d4f6 d8d1 h1d1 e6f6 c3c4 g7g6 d1d5 g6h5 d5h5 f6g6 g2g4 g6f6 h5f5 f6f5 e4f5 c7c6 c1d2 f7f6 d2e2 b7c7 e2e3 c7d7 e3f4 d7c7');
INSERT INTO evaluations VALUES(22976,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(22976,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(22976,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(22976,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(22976,34,'cp -119',NULL);
INSERT INTO evaluations VALUES(22976,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(22976,36,'cp -80','c7d6 e5d6 d8f6 d1e2 f6e6 c3b5 e6e2 f1e2 b7b6 b5c7 e8d8 c7a8 f8d6 c1e3 d6c5 e3c5 b6c5 e1g1 c8b7 e2f3 d7d5 c2c4 b7a8 f3d5 a8d5 c4d5 d8d7 a1c1 d7d6 b2b4 c5b4 c1c6 d6d5 c6c7 f7f5 f1b1 a7a5 c7a7 g8f6 a7a5 d5e6 b1b4 f6d5 a5a6 e6f7 b4b7 g6e7 a6a4 h8c8');
INSERT INTO evaluations VALUES(19119,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(19119,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(19119,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(19119,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(19119,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(19119,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(19119,36,'cp -69','e5d6 d8f6 d1e2 f6e6 c3b5 e6e2 f1e2 b7b6 b5c7 e8d8 c7a8 f8d6 c1e3 d6c5 e3c5 b6c5 b2b4 c8b7 e2f3 b7f3 g2f3 d8c8 f4f5 g6h4 b4c5 c8b7 e1c1 g8f6 h1g1 h4f5 g1g5 g7g6 g5f5 g6f5 d1d6 f6e8 d6d7 b7a8 d7f7 a8b8 c1d2 e8c7 d2e3 c7e6 f7f5 h8c8 f3f4 e6c5 e3d4 c5b7 f5h5 c8c2 h5h7 c2a2 h2h4 a2a5 h7h8 b8c7 h4h5 b7d6');
INSERT INTO evaluations VALUES(19118,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(19118,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(19118,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(19118,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(19118,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(19118,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(19118,36,'cp -74','d8f6 d1e2 f6e6 c3b5 e6e2 f1e2 b7b6 b5c7 e8d8 c7a8 f8d6 c1e3 d6c5 e3c5 b6c5 b2b4 c8b7 e2f3 b7f3 g2f3 d8c8 f4f5 g6h4 b4c5 c8b7 e1c1 g8f6 h1g1 h4f5 g1g5 g7g6 g5f5 g6f5 d1d6 f6e8 d6d7 b7a8 d7f7 a8b8 c1b2 e8c7 f7f5 h8e8 f5f7 h7h6 f3f4 b8b7 h2h4 b7c6 f4f5 e8e5 f7f6 c6c5 f6h6 e5f5 h4h5 f5f4 h6h8 f4h4 h5h6 c7b5 h8c8 c5b6');
INSERT INTO evaluations VALUES(26690,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(26690,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(26690,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(26690,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(26690,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(26690,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(26690,36,'cp 81','g8f6 h2h3 g4h5 b1c3 h5f3 e2f3 e7e5 d4d5 c6d4 g2g4 d4f3 d1f3 g7g6 c1e3 f6d7 h1g1 h7h5 e1c1 f8h6 g4g5 h6g7 a2a3 a7a6 f3e2 e8g8 c1b1 b7b5 c3a2 d7c5 f2f3 d8d7 a2b4 d7h3 b4c6 h3d7 b2b3 g8h7');
INSERT INTO evaluations VALUES(29122,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(29122,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(29122,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(29122,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(29122,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(29122,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(29122,36,'cp 40','b8c6 c1g5 e7e6 c4d5 e6d5 f1d3 f5d3 d1d3 f8e7 e1g1 e8g8 a1d1 f6d7 g5e7 c6e7 d3b5 d7b6 f1e1 e7c6 e1e2 d8d6 b5c5 a8d8 g1f1 h7h6 f3e5 d6b8 g2g3 d8d6 d1e1 f8c8 e5c6 b7c6 b2b3 b6d7 c5a5 d7f8 e2e7 f8e6 a5a7 b8a7 e7a7 e6d4');
INSERT INTO evaluations VALUES(29346,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(29346,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(29346,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(29346,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(29346,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(29346,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(29346,36,'cp 92','b7b5 c4b3 b8d7 c3e2 f8e7 a2a4 a7a5 e1g1 e8g8 a4b5 c6b5 c2c3 a5a4 b3a2 f6e8 f3g3 e8f6 f1d1 g7g6 c1h6 f8e8 g3f3 e7f8 h6e3 d7c5 g2g4 c5e6 e2g3 f6d7 g4g5 d7c5 a2d5 d8e7 h3h4 c5b3 a1b1 h7h5 d5a8 e8a8 d3d4 e7b7 d4d5 e6g7');
INSERT INTO evaluations VALUES(29599,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(29599,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(29599,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(29599,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(29599,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(29599,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(29599,36,'cp -31','h2h3 h7h6 g2g4 f5h7 b2b3 a7a6 c1b2 e7e6 f1g2 c7c5 f3e5 f8e7 d2d3 e8g8 e1g1 f6d7 e5d7 d8d7 d1e2 b8c6 b1d2 a8e8 a1e1 f7f6 e3e4 g8h8 d2f3 d7c7 e4d5 e6d5 f4f5 c7g3 b2a3 e7d6 e2f2 g3f2 g1f2 h7g8');
INSERT INTO evaluations VALUES(33459,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(33459,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(33459,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(33459,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(33459,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(33459,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(33459,36,'cp 182','e1g1 e4e7 d2d4 c5b6 c2c4 c7c6 b1c3 d7d5 c1g5 c8e6 f3e5 d5c4 f1e1 e8g8 e2c4 e6c4 e5c4 e7d8 c3e4 b6d4 c4d6 c6c5 d6b7 d8b6 g5f6 d4f6 b7c5 f6e7 d1f3 e7c5 e4c5 b6c6 c5e4 b8d7 a1c1 c6b6 b2b3 a8d8 e4g3 f8e8 e1e8 d8e8 f3g4 d7c5 g3h5 b6g6 g4g6 h7g6 h5f4 e8d8 g2g3 c5e6 f4e6');
INSERT INTO evaluations VALUES(33458,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(33458,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(33458,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(33458,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(33458,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(33458,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(33458,36,'cp 168','b8c6 b1c3 e4e7 d2d4 c5b4 c1g5 b4c3 b2c3 d7d5 c3c4 d5c4 d4d5 c6e5 f1e1 e8g8 e2c4 e5f3 d1f3 e7d6 h2h3 c7c6 g5f4 d6c5 c4b3 f6d5 f4e5 f8d8 b3d5 c6d5 f3g3 c5f8 e5f6 c8e6 f6d8 a8d8 a1d1 f8c5 c2c3 b7b6 g3e5 c5d6 e5d4 h7h6 e1e3 d6c5 d4d2 d8c8 d2d3 c8d8 d1e1 d8c8');
INSERT INTO evaluations VALUES(35597,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(35597,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(35597,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(35597,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(35597,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(35597,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(35597,36,'cp 30','c4d5 e6d5 f1d3 h7h6 g5h4 c8e6 h2h3 c7c6 h4g3 e7b4 e1g1 b8d7 a1c1 f8e8 d3b1 b4c3 b2c3 f6e4 g3h2 b7b5 c3c4 d8e7 c4d5 c6d5 f3d2 e6f5 c1c7 a8c8 d2e4 f5e4 c7a7 e4b1 d1b1 e7e6 b1d3 c8c4 d3a3 e8c8 a3a6 e6a6 a7a6 c4c1');
INSERT INTO evaluations VALUES(36978,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(36978,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36978,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(36978,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36978,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(36978,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(36978,36,'cp 30','b8c6 f1e1 e6e5 d4d5 c6e7 e3e4 h7h6 c1d2 e7g6 a2a4 c7c6 a4a5 c6d5 c3d5 c8e6 d1b3 e6d5 c4d5 f6d5 b3d5 d8e7 d2e3 a8d8 d5b3 b7b6 a5b6 a7b6 g2g3 d8b8 e1d1 d6c5 h2h4 c5e3 b3e3 f8e8 h4h5 g6f8 f3h4 e7e6 e3e2 b8d8 d1d8');
INSERT INTO evaluations VALUES(37029,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(37029,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(37029,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(37029,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(37029,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(37029,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(37029,36,'cp -12','c7c5 g5f6 d8f6 e2e3 d5c4 d1b1 b8d7 f1c4 b7b6 c4d3 h7h6 d3h7 g8h8 h7e4 a8b8 e1g1 c8b7 e4b7 b8b7 a3a4 a7a5 h2h3 f8c8 b1a2 c5d4 c3d4 b7c7 f1d1 f6e7 d4d5 e6d5 a2d5 h8g8 d5d6 e7d6 d1d6 c7c1 a1c1 c8c1 g1h2');
INSERT INTO evaluations VALUES(44893,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44893,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(44893,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44893,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(44893,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(44893,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44893,36,'cp 10','d5c4 e1g1 e7e6 a2a4 b8d7 d1c2 e6e5 b1a3 f8a3 a1a3 e8g8 a4a5 d8e7 c2c4 e5e4 f3g5 f8e8 d2d4 e4d3 c4d3 h7h6 g5f3 d7c5 d3c2 c5e4 a5a6 b7a6 f3d2 c8f5 d2e4 f5e4 g2e4 e7e4 c2e4 e8e4 a3a6 c6c5 c1e3 f6d5 e3c5 a8c8 c5d6 e4e2 a6a7 e2b2');
INSERT INTO evaluations VALUES(45015,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(45015,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45015,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(45015,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(45015,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45015,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(45015,36,'cp 7','e1g1 e7e6 a2a4 b8d7 d1c2 e6e5 b1a3 f8a3 a1a3 e8g8 c2c4 f8e8 a4a5 e5e4 f3g5 d8e7 d2d4 e4d3 c4d3 h7h6 g5f3 d7c5 d3c2 c5e4 a5a6 b7a6 f3d2 c8f5 d2e4 f5e4 g2e4 e7e4 c2e4 e8e4 a3a6 e4e2 a6c6 f6e4 c6c7 g7g5 b2b3 a7a5 f2f3 e4f6');
INSERT INTO evaluations VALUES(46863,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(46863,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(46863,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(46863,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(46863,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46863,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(46863,36,'cp 36','e2e4 d5e4 c3e4 f6f5 e4g3 b8c6 f1b5 d8d6 c2c3 a7a6 b5d3 e7e5 g3f5 c8f5 d3f5 e5d4 d1c2 f8g7 g1e2 a8d8 a1d1 e8g8 f5h7 g8h8 e1g1 f8e8 h7e4 d6c5 e2g3 d4c3 b2c3 c5c3 c2c3 g7c3 d1c1 c3e5 e4c6 b7c6 c1c6 e8e6 c6e6 f7e6 f1c1');
INSERT INTO evaluations VALUES(46864,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46864,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46864,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46864,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(46864,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(46864,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(46864,36,'cp 45','d5e4 c3e4 f6f5 e4g3 b8c6 g1f3 d8d5 c2c3 f5f4 g3e2 c8g4 d1b3 d5b3 a2b3 e8c8 e2f4 g4f3 g2f3 e7e5 d4e5 c6e5 a1a7 c8b8 a7a5 f8d6 f4d3 h8e8 e1d2 e5f3 d2d1 d6f4 a5f5 f3e1 f5f4 e1d3 f1d3 d8d3 d1c2 d3d8 h1d1 e8e2 c2c1 d8d1 c1d1 e2b2');
INSERT INTO evaluations VALUES(48098,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(48098,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(48098,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(48098,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(48098,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(48098,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(48098,36,'cp 34','g1f3 f8e7 d2d4 e8g8 f1g2 b7b6 c3d5 e6d5 c1f4 e7d6 d1d2 b8c6 e1g1 c6e7 f1c1 c7c6 a2a4 a7a5 f3e5 c8b7 e5d3 b7a6 e2e3 a6d3 f4d6 d8d6 d2d3 h7h5 a1b1 f8c8 d3c3 g7g6 b2b4 g8g7 g2h3 a5b4 b1b4 c8c7 b4b6 a8a4');
INSERT INTO evaluations VALUES(49259,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(49259,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49259,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49259,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49259,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49259,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49259,36,'cp 0','b1c3 f8c5 d2d3 d7d6 g1e2 c7c6 a2a3 b7b5 c4a2 a7a5 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(49481,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(49481,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(49481,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(49481,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(49481,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(49481,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(49481,36,'cp -47','f6e4 d1e2 e4f6 e2e5 f8e7 b1c3 e8g8 e5d4 f8e8 c1e3 d7d6 f1e2 b8d7 g1f3 d7b6 d4d2 c8f5 e2b5 b6d7 b5e2 f6e4 c3e4 f5e4 e1g1 d7b6 c2c4 e4f3 g2f3 e7f6 a1e1 a7a5 f3f4 d8d7 g1g2 b6a4 f1g1 f6b2 e1b1 b2f6 b1b7 a4c3 e2d3 d7g4 g2h1 g4f3 g1g2 a5a4 b7c7');
INSERT INTO evaluations VALUES(49804,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(49804,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(49804,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(49804,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(49804,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(49804,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(49804,36,'cp -31','d1e2 e4f6 e2e5 f8e7 e5d4 e8g8 b1c3 f8e8 c1e3 d7d6 f1e2 h7h6 d4d2 f6g4 e2g4 c8g4 h2h3 g4f5 g1e2 f5h7 e1g1 b8d7 e2g3 c7c5 a1e1 d8a5 a2a3 b7b5 d2e2 b5b4 a3b4 c5b4 e2b5 a5d8 b5b4 h7c2 b4d4 c2g6 d4d2 d7e5 e3d4 e7g5 d4e3');
INSERT INTO evaluations VALUES(51776,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(51776,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(51776,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(51776,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(51776,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(51776,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51776,36,'cp -7','d7d5 e4d5 e6d5 c4d3 f8d6 g1f3 g8f6 d1e2 d6e7 e1g1 e8g8 c1g5 h7h6 f1e1 e7d6 g5h4 c8g4 b1d2 b8d7 h2h3 g4f3 d2f3 f8e8 e2e8 d8e8 e1e8 a8e8 h4f6 d7f6 a1e1 e8e1 f3e1 g7g6 c2c3 f6e4 g2g3 f7f5 d3f1 g8f7 e1d3 g6g5 f1e2 f5f4 g3f4 d6f4');
INSERT INTO evaluations VALUES(53557,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53557,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53557,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53557,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53557,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53557,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53557,36,'cp 0','e4d5 e6d5 c4d3 f8d6 g1f3 g8f6 d1e2 d6e7 e1g1 e8g8 c2c3 c8g4 c1f4 f8e8 e2d1 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(53541,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53541,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53541,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53541,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53541,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53541,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53541,36,'cp 0','e6d5 c4d3 f8d6 g1f3 g8e7 e1g1 h7h6 c2c4 e8g8 f1e1 c8g4 h2h3 g4h5 b1d2 b8d7 d1b3 c6c5 c4d5 c5d4 d3c4 d7c5 b3a3 c5e6 a3b3');
INSERT INTO evaluations VALUES(55134,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(55134,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(55134,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(55134,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(55134,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(55134,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(55134,36,'cp -16','e7e5 d2d4 b8c6 d4e5 c6e5 g1f3 f8d6 f3e5 d6e5 e1g1 c7c6 c2c4 g8f6 b1d2 e5c7 d2f3 d8e7 c4d5 f6d5 d1d4 c8f5 e2c4 d5f6 b2b3 f5e4 c1b2 e8g8 a1d1 h7h6 d4c3 f8e8 c4d3 a8d8 d3e4 e7e4 d1d8 e8d8 c3c1 e4e7 f1d1 f6e4 d1d8 c7d8 c1c2 d8c7 f3d2');
INSERT INTO evaluations VALUES(55687,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(55687,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(55687,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(55687,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(55687,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(55687,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(55687,36,'cp -9','g1e2 g8f6 h2h3 c7c6 c1e3 d6d5 e4d5 f6d5 c3d5 d8d5 d1d5 c6d5 e1c1 b8c6 e2c3 c8e6 c3d5 e8c8 f1c4 c6b4 d5b4 d8d1 h1d1 e6c4 a2a3 f8b4 a3b4 b7b6 d1d4 c4e6 h3h4 g7g6 c1d2 c8c7 e3h6 h8c8 g2g4 c7c6 h4h5 g6h5 g4h5 c6b5 h6e3 c8c4 d4c4 b5c4');
INSERT INTO evaluations VALUES(56109,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(56109,31,'cp -133',NULL);
INSERT INTO evaluations VALUES(56109,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(56109,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(56109,34,'cp -138',NULL);
INSERT INTO evaluations VALUES(56109,35,'cp -130',NULL);
INSERT INTO evaluations VALUES(56109,36,'cp -143','c7c6 b5e2 f8d6 d2d4 e5e4 c2c4 g8f6 b1c3 e8g8 c1d2 d5c4 e2c4 b7b5 c4b3 a7a5 g1e2 c8a6 e2g3 f8e8 e1g1 b5b4 c3e2 b8d7 b3a4 a8c8 a2a3 d7b6 a3b4 a5b4 g3f5 b6a4 a1a4 a6b5 f5d6 b5a4 d1a4 d8d6 d2b4 d6d7');
INSERT INTO evaluations VALUES(53681,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(53681,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(53681,32,'cp -123',NULL);
INSERT INTO evaluations VALUES(53681,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(53681,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(53681,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(53681,36,'cp -133','b5e2 g8f6 d2d4 e5e4 c2c4 f8d6 a2a3 e8g8 b1c3 f8e8 c1d2 d5c4 e2c4 b8d7 g1e2 d7b6 c4b3 c8e6 b3e6 e8e6 d1b3 d8d7 h2h3 f6d5 e1c1 a8e8 c1b1 e6f6 h1f1 f6g6 g2g4 f7f5 c3d5 c6d5 d2b4 d6b4 b3b4 f5g4 e2f4 g4h3 f4g6 h7g6');
INSERT INTO evaluations VALUES(57654,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(57654,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(57654,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(57654,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(57654,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(57654,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(57654,36,'cp -36','d7d5 c2c4 e7e6 b1c3 f8g7 g1f3 g8e7 c1e3 e8g8 e3f2 b8d7 c4d5 e6d5 e2e3 f8e8 f1d3 d7b6 e1g1 e7f5 f1e1 c7c6 d1c2 f5d6 a1c1 c8f5 d3f5 d6f5 h2h3 a8c8 b2b3 d8e7 a2a4 f5d6 f3e5 f7f6');
INSERT INTO evaluations VALUES(54854,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(54854,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(54854,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(54854,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(54854,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(54854,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(54854,36,'cp -33','e2e3 f8g7 g1f3 c7c5 f1b5 c8d7 b5e2 g8f6 c2c3 d8c7 e1g1 b8c6 f3e5 h7h5 e5d7 f6d7 d4c5 e7e6 c3c4 d7c5 b1c3 e8g8 c4d5 a8d8 c3b5 c7b8 d5d6 c5e4 d6d7 e4c5 d1d6 b8d6 b5d6 d8d7 f1d1');
INSERT INTO evaluations VALUES(57711,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(57711,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(57711,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(57711,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(57711,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(57711,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(57711,36,'cp 174','g1f3 f8g7 b1c3 c7c6 h2h3 d7d5 e4e5 c8f5 f1d3 f5d3 d1d3 e7e6 c1g5 d8c7 d3d2 g7h6 g5e3 b8d7 e1g1 a7a6 a1e1 c7a5 c3e4 a5d2 e4d6 e8e7 f3d2 h6e3 f2e3 f7f6 d6b7 f6e5 b7a5 a8c8 d4e5 d7e5 e3e4 d5d4 d2f3 e5f3 f1f3');
INSERT INTO evaluations VALUES(57710,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(57710,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(57710,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(57710,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(57710,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(57710,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(57710,36,'cp 156','c7c6 h2h3 d7d5 b1c3 d5e4 c3e4 c8f5 e4g3 b8d7 f1c4 e7e6 e1g1 f8h6 f1e1 h6c1 d1c1 g8e7 c4b3 h5h4 g3e2 e8f8 c2c4 f8g7 c1f4 d7f6 f3h4 d8b8 f4g5 h8h5 g5g3 b8g3 f2g3 b7b5 h4f3 h5h8 e2c3 a8d8 a1d1 b5c4 b3c4 e7d5 g3g4 d5c3 b2c3 f5e4 f3e5');
INSERT INTO evaluations VALUES(390,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(390,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(390,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(390,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(390,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(390,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(390,36,'cp 93','e1g1 c5b6 b2b4 g8f6 a2a4 a7a6 a4a5 b6a7 d2d4 d7d6 d4d5 c6b8 f1e1 e8g8 c4d3 f6h5 c3c4 c8g4 b1d2 h5f4 d3f1 e7f6 h2h3 g4h3 g2h3 f6g6 g1h1 a7f2 e1e2 g6g3 e2f2 g3f2 d1e1 f2e1 f3e1 b8d7 h1h2 f7f5 e1d3 a8e8 d3f2 d7f6 e4f5');
INSERT INTO evaluations VALUES(389,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(389,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(389,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(389,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(389,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(389,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(389,36,'cp 93','g8f6 b2b4 c5b6 d2d4 d7d6 a2a4 a7a6 a4a5 b6a7 d4d5 c6b8 f1e1 e8g8 c4d3 f6g4 a1a2 f7f5 h2h3 g4h6 e4f5 h6f5 c3c4 f5h4 f3h4 e7h4 c1e3 a7e3 e1e3 b8d7 b1c3 d7f6 c3e4 f6e4 d3e4 c8f5 a2e2 a8e8 c4c5 f5e4 e3e4 h4f6');
INSERT INTO evaluations VALUES(612,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(612,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(612,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(612,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(612,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(612,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(612,36,'cp 43','b7b6 a2a3 b4d2 d1d2 d7d6 e1g1 a7a5 b2b4 e5d7 f2f3 e8g8 e3e4 c8b7 f4g5 e7e6 d2c3 f8e8 g5e3 a5b4 a3b4 h7h6 e3d4 e6g6 f1d1 g6g5 d1e1 a8a1 e1a1 e8a8 a1a8 b7a8 d4e3 g5d8 c3a1 a8b7 e3f2 d8a8 a1b2 d7e5 f2d4 c7c5');
INSERT INTO evaluations VALUES(3025,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(3025,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(3025,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(3025,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(3025,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(3025,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(3025,36,'cp -40','c1d2 d7d5 e4d5 f6d5 e1g1 b4c3 b2c3 c8g4 a1b1 f8e8 f1e1 d8c8 h2h3 g4h5 b1b5 d5c7 b5b1 f7f6 g2g4 h5g6 f3h4 g6d3 c2d3 b7b6 d3d4 e5d4 c3d4 e8e1 d2e1 c8e6 b1c1 c7d5 h4f5 g7g6 f5g3 a8d8 g3e2 c6e7 d1d3 f6f5');
INSERT INTO evaluations VALUES(3026,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(3026,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(3026,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(3026,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(3026,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(3026,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(3026,36,'cp -42','d7d5 e4d5 f6d5 e1g1 b4c3 b2c3 c8g4 a1b1 f8e8 f1e1 d8c8 h2h3 g4h5 b1b5 d5c7 b5b1 f7f6 g2g4 h5g6 f3h4 g6d3 c2d3 b7b6 d3d4 e5d4 c3d4 e8e1 d2e1 c8e6 h4f5 a8e8 b1c1 g7g6 f5e3 c7d5 e3d5 e6d5 e1c3 g8f7 c3b2 e8e6 d1b3 d5b3');
INSERT INTO evaluations VALUES(6363,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(6363,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(6363,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(6363,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(6363,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(6363,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(6363,36,'cp 86','a2a3 b4c3 b2c3 e6e5 d4e5 d8h4 f1d3 c6e5 g1f3 e5f3 d1f3 d7d6 c1f4 g8e7 e4e5 e7g6 d3g6 f7g6 e5d6 e8g8 f3d5 g8h8 g2g3 h4f6 e1c1 c7d6 d5d6 c8f5 d6f6 g7f6 d1d6 f8e8 f4e3 h8g7 h1d1 g7f7 c1b2 e8e4 d6d4 a8e8 d4b4 b7b6 c3c4 f5g4 d1d2 h7h5 c4c5 b6c5 b4e4 e8e4');
INSERT INTO evaluations VALUES(6593,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(6593,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(6593,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(6593,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(6593,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(6593,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(6593,36,'cp 214','f3e5 c6e5 b1c3 d7d6 b2b3 c8e6 c1b2 d8f6 d1d2 e8c8 e1c1 f6h6 f2f4 e5c6 c1b1 g8e7 f1e2 h8f8 g2g4 h6h4 h2h3 f8e8 e2f3 f7f6 c3d5 c8b8 d1e1 e6c8 c2c4 e7g6 f3d1 c6e7 b2d4 a7a6 d4f2 h4h6 f2e3 e7d5');
INSERT INTO evaluations VALUES(6590,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(6590,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(6590,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(6590,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(6590,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(6590,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(6590,36,'cp 215','c6e5 b1c3 d7d6 b2b3 g8f6 c1b2 c8d7 f2f4 e5g6 d1d2 d8e7 f1d3 e8c8 e1c1 c8b8 c1b1 h8e8 h1e1 a7a6 a2a4 d7g4 d1c1 g4c8 h2h3 e7f8 b2a3 f6d7 b3b4 h7h5 a4a5 g6e7 b4b5 d7c5 b5b6 d8d7 b6c7 d7c7 d3f1');
INSERT INTO evaluations VALUES(6892,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(6892,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(6892,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(6892,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(6892,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(6892,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(6892,36,'cp 186','b1c3 d7d6 b2b3 c8d7 f2f4 e5g6 d1d2 g8f6 c1b2 d8e7 f1d3 e8c8 e1c1 h8e8 d2f2 c8b8 h2h3 d7c6 h1e1 f6d7 b3b4 b7b6 c1b1 c6b7 g2g3 e7f8 d3b5 a7a6 b5d3 d7f6 a2a3 e8e7 f2f1 d8e8 d3a6 f6e4 a6b7 e4c3 b2c3 e7e1 c3e1');
INSERT INTO evaluations VALUES(9362,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9362,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9362,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9362,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9362,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9362,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9362,36,'cp 0','a7a6 g1f3 e8g8 e1g1 c8e6 c4e6 f7e6 c1e3 f6d7 c3e2 a6a5 e2g3 d8f6 d1e2 c5e3 f2e3 f6g6 g1h2 d7f6 f1f2 f8f7 a1f1 g6h6 h2g1 h6g6');
INSERT INTO evaluations VALUES(10851,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(10851,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(10851,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(10851,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(10851,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(10851,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(10851,36,'cp 69','d7c6 d1d8 e7d8 b1c3 a7a5 a2a4 f6d7 c1f4 d7c5 h2h4 c8e6 c4e2 d8f6 f4c7 f6c3 b2c3 c5e4 c3c4 e8g8 f2f3 f8c8 c7f4 e4c5 e1f2 e6f5 a1a2 c5a6 a2b2 h7h5 f4d6 a6b4 g2g4 h5g4 f3g4 f5e4 h1c1 c8e8 c4c5 e4d5');
INSERT INTO evaluations VALUES(8665,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(8665,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(8665,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(8665,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(8665,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(8665,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(8665,36,'cp 58','d1d8 e7d8 b1c3 a7a5 a2a4 f6d7 c1e3 d8f6 e1c1 d7e5 c4e2 e5g4 e3c5 c8e6 f2f4 f6c3 f4f5 c3b4 c5b4 a5b4 e2g4 h7h5 g4e2 e6d7 b2b3 e8c8 a4a5 h8e8 a5a6 b7a6 h1e1 f7f6 e2a6 c8b8 a6c4 h5h4 d1d3 d7c8 d3d8 e8d8');
INSERT INTO evaluations VALUES(11624,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(11624,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(11624,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(11624,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(11624,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11624,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(11624,36,'cp 40','e4f6 f3e5 c6e5 d3d4 c8e6 d4e5 f6e4 c2c3 f8e7 c1e3 e8g8 b1d2 e4d2 d1d2 a6a5 b3c2 b5b4 a1d1 d8d7 h2h3 b4c3 b2c3 c7c6 f2f4 e6f5 e1e2 f5c2 d2c2 f7f6 e3d4 f6e5 d4e5 e7d6 c3c4 d6e5 f4e5 d7e6 g1h2 a5a4 c4d5 c6d5 d1d4 f8c8 c2d3 c8c4');
INSERT INTO evaluations VALUES(11616,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(11616,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(11616,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(11616,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(11616,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(11616,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11616,36,'cp 31','f3e5 c6e5 d3d4 c8e6 d4e5 f6e4 c1e3 f8e7 c2c3 e8g8 b1d2 e4d2 d1d2 a6a5 b3c2 d8d7 h2h3 c7c6 e3g5 b5b4 g5e7 d7e7 f2f4 f7f5 e5f6 e7f6 c3b4 a5b4 d2b4 e6d7 a2a4 f8b8 b4c5 f6b2 a4a5 b2b4 c5e3 a8a5 a1a5 b4a5');
INSERT INTO evaluations VALUES(11612,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11612,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(11612,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(11612,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(11612,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(11612,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(11612,36,'cp 30','c6e5 d3d4 c8e6 d4e5 f6e4 c1e3 f8e7 c2c3 e8g8 b1d2 e4d2 d1d2 d8d7 h2h3 a6a5 b3c2 c7c6 g2g4 f8e8 a2a4 b5b4 f2f4 f7f5 e5f6 e7f6 c3b4 a5b4 e3c5 f6b2 c2h7 g8h7 d2b2 d5d4 b2d4 b4b3 g1f2 d7d4 c5d4 c6c5 d4c3 e6c4 a4a5 c4a6 a1d1 e8e1 d1e1');
INSERT INTO evaluations VALUES(11674,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(11674,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(11674,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(11674,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(11674,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(11674,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(11674,36,'cp 30','d3d4 c8e6 d4e5 f6e4 c1e3 e4c5 d1f3 f8e7 f3g3 g7g6 b1d2 e8g8 e3h6 f8e8 c2c3 e6f5 d2f3 c7c6 a1d1 d8c8 f3d4 a6a5 g3f4 c5e6 f4c1 e6d4 c3d4 c8e6 e1e2 a8c8 h2h3 a5a4 b3c2 c6c5 c2f5 e6f5 d4c5 c8c5 c1d2');
INSERT INTO evaluations VALUES(12634,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(12634,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(12634,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(12634,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(12634,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(12634,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(12634,36,'cp 71','b1c3 d7d6 c1g5 f8e7 e1c1 e8g8 f2f4 a7a6 c4b3 b7b5 c1b1 c8b7 f4f5 e6f5 e4f5 h7h6 g5h4 d8c7 h1e1 a8e8 a2a3 c7c5 d4f4 e7d8 h4f2 c5c7 f2d4 e8e1 d1e1 f8e8 e1e8 f6e8 c3d5 c7c8 f4f3 c8d7 h2h3 b7c6 f3g4 c6d5 b3d5 e8f6 g4f3 f6d5 f3d5 d8f6 d4f6 g7f6');
INSERT INTO evaluations VALUES(12633,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(12633,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(12633,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(12633,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(12633,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(12633,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(12633,36,'cp 73','d7d6 c1g5 f8e7 e1c1 e8g8 f2f4 a7a6 c4b3 b7b5 h1e1 d8c7 g5f6 e7f6 d4d6 c7d6 d1d6 g7g5 g2g3 g5f4 g3f4 f8d8 e1d1 d8d6 d1d6 f6e7 d6d4 g8f8 e4e5 c8b7 c1d2 a8b8 c3e4 b7e4 d4e4 a6a5 c2c3 h7h5 b3c2 h5h4 d2d3 b5b4 e4c4 b4c3 b2c3 b8d8 d3e2');
INSERT INTO evaluations VALUES(16824,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(16824,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(16824,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(16824,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(16824,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(16824,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(16824,36,'cp -102','e4c3 d2c3 f7f6 b2b4 d7d6 a2a4 a7a6 f1e1 d8d7 c4b3 d7g4 h2h3 g4h5 e1e2 f8e7 f3d4 c8d7 d4e6 d7e6 b3e6 f6f5 e6b3 h5g6 d1d5 c6d8 d5a5 b7b6 a5d5 c7c6 d5d1 d8e6 f2f3 e8g8 a4a5 g8h8 a5b6');
INSERT INTO evaluations VALUES(16840,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(16840,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(16840,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(16840,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(16840,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(16840,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(16840,36,'cp -86','d2c3 f7f6 b2b4 d7d6 a2a4 d8e7 f3d4 e5d4 f1e1 c6e5 c3d4 c8e6 c4f1 e7f7 d4e5 d6e5 c1a3 f8e7 b4b5 e8g8 a3e7 f7e7 d1f3 b7b6 f3c6 e7d6 c6c3 f8d8 e1d1 d6c5 c3c5 b6c5 a4a5 g8f7 d1d8 a8d8 f2f3 d8d4 c2c3');
INSERT INTO evaluations VALUES(16838,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(16838,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(16838,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(16838,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(16838,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(16838,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(16838,36,'cp -82','f7f6 b2b4 d7d6 a2a4 a7a6 h2h3 d8e7 f3d4 e5d4 f1e1 c6e5 c3d4 c8e6 c4f1 e7f7 d4e5 d6e5 d1f3 e6d5 f3c3 f8e7 b4b5 a6a5 c3c7 e7b4 c7f7 d5f7 e1e3 a8c8 f1d3 e8g8 c1a3 b4a3 a1a3 f8d8 d3e4');
INSERT INTO evaluations VALUES(17345,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(17345,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(17345,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(17345,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(17345,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(17345,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(17345,36,'cp -33','c2c3 d7d5 e5d6 f8d6 d2d4 c5d4 f3d4 d8c7 d4c6 c7c6 d1g4 g8e7 b1d2 h7h5 g4f3 c8d7 f3c6 d7c6 e1g1 e8c8 f1e1 d6c7 d2b3 h8e8 b3d4 c6d5 c4d3 e6e5 d4f5 e7f5 d3f5 d5e6 f5c2 h5h4 h2h3 f7f6 c1e3 g7g5 a2a4 e8g8 a1d1 d8d1 c2d1 g5g4 h3g4 e6g4 d1g4 g8g4 b2b3 f6f5 g1h1 e5e4 e1d1');
INSERT INTO evaluations VALUES(17995,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(17995,31,'cp 263',NULL);
INSERT INTO evaluations VALUES(17995,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(17995,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(17995,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(17995,35,'cp 275',NULL);
INSERT INTO evaluations VALUES(17995,36,'cp 280','b1c3 d5e4 d1d8 e8d8 f3g5 d8e8 c1d2 f7f5 e5f6 g8f6 g5e4 c8e6 f2f3 h7h6 e1c1 e8f7 h1e1 a8d8 a2a3 f8e7 e4f6 e7f6 c3e4 f6e7 d2f4 d8d1 e1d1 h8c8 d1d4 g7g5 f4g3 e6d5 d4a4 h6h5 h2h3 f7e6 a4a6 h5h4 g3f2 c8f8 f2e3 d5e4 f3e4 e7d6 e3g5 f8f2');
INSERT INTO evaluations VALUES(19003,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19003,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(19003,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(19003,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19003,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(19003,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19003,36,'cp 16','f1g2 d7d6 d2d3 g8e7 f2f4 e5f4 c1f4 e7g6 f4e3 f8e7 g1e2 c8e6 d1d2 e8g8 e1g1 e7f6 a1b1 f6e5 a2a3 d8d7 b2b4 b7b6 e2f4 g6f4 g3f4 e5c3 d2c3 e6h3 c3d2 h3g2 d2g2 a8e8 f4f5 f7f6 g1h1 g8h8');
INSERT INTO evaluations VALUES(19145,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(19145,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(19145,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(19145,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(19145,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(19145,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(19145,36,'cp 72','b7b6 b1c3 g8e7 c1e3 c8a6 f1e2 e7c6 d4d2 a8c8 e1g1 f8c5 e3g5 f7f6 g5e3 e8g8 a1d1 d7d6 f2f4 d8e7 e3f2 g8h8 b2b3 a6b7 e2g4 c8d8 h2h3 c5b4 d2b2 b4c5 g1h1 e7f7 a2a3 a7a5 f2g1 e6e5 f4f5 h7h5 g4f3 c6d4 g1d4');
INSERT INTO evaluations VALUES(19684,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(19684,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(19684,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(19684,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(19684,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(19684,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(19684,36,'cp 32','b1c3 f8g7 c1e3 g8f6 f2f3 c8d7 d1d2 e8g8 e1c1 a8c8 c4b3 c6d4 e3d4 b7b5 a2a3 a7a5 e4e5 d6e5 d4e5 b5b4 e5f6 g7f6 c3e4 c8c7 e4f6 e7f6 a3a4 d8c8 h1e1 d7f5 c1b1 h7h5 h2h3 c7d7 d2e3 d7d1 e1d1 f8d8 d1d4 d8d4 e3d4 g8g7 d4b6 c8d7 b6a5 d7d1 b1a2 f5c2 b3c2 d1c2 a5b4 c2g2');
INSERT INTO evaluations VALUES(20352,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(20352,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(20352,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(20352,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(20352,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(20352,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(20352,36,'cp -67','g8h6 e1g1 f8e7 c2c3 d5d4 e2b5 d8c7 a2a4 e8g8 b5d3 b7b6 b1a3 g8h8 d3e4 c8b7 d2d3 f7f5 a3b5 c7d7 e4c6 b7c6 c3d4 c5d4 h2h3 e7c5 b5d6 h6f7 d6c4 a7a6 d1e2 b6b5 a4b5 c6b5');
INSERT INTO evaluations VALUES(29128,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(29128,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(29128,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(29128,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(29128,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(29128,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(29128,36,'cp 46','d1b3 b8d7 e2e3 d8b6 b3b6 d7b6 a2a4 e7e6 f2f3 a7a6 a4a5 b6c8 g2g4 f5g6 h2h4 h7h5 g4g5 f6d7 g1e2 c8d6 e2c1 a8c8 f1d3 d6f5 e1e2 f7f6 d3f5 g6f5 c1d3 e8f7 h1b1 f5d3 e2d3 f8b4 a1a4');
INSERT INTO evaluations VALUES(31767,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(31767,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(31767,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(31767,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(31767,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(31767,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(31767,36,'cp 57','g8f6 h2h3 g4h5 d2d3 h7h6 b1c3 e7e5 e1g1 c7c6 f3h4 h5e2 d1e2 g7g6 e2e1 f8e7 h4f3 b8d7 a2a4 e7f8 d3d4 f8g7 d4e5 d6e5 e1e2 e8g8 c1e3 d8c7 f3d2 f8e8 d2c4 g7f8 f2f3 d7c5 b2b3');
INSERT INTO evaluations VALUES(33464,30,'cp 368',NULL);
INSERT INTO evaluations VALUES(33464,31,'cp 359',NULL);
INSERT INTO evaluations VALUES(33464,32,'cp 406',NULL);
INSERT INTO evaluations VALUES(33464,33,'cp 395',NULL);
INSERT INTO evaluations VALUES(33464,34,'cp 406',NULL);
INSERT INTO evaluations VALUES(33464,35,'cp 396',NULL);
INSERT INTO evaluations VALUES(33464,36,'cp 411','b1c3 c7c6 d2d4 c5d4 f3d4 e5d4 e4e5 f6h4 g2g3 h4h3 c1h6 h3h6 c3e4 e8g8 e4d6 b7b5 c4b3 b8a6 d1d4 a6c7 f2f4 c7e6 d4e3 a7a5 a2a4 b5b4 a1d1 c8a6 f1e1 c6c5 b3d5 a8b8 d1d2 b8b6 d5e6 f7e6 e3c5 b6c6 c5a7 f8d8 d2f2 h6g6 f4f5 e6f5 d6f5 g6e8');
INSERT INTO evaluations VALUES(33460,30,'cp 384',NULL);
INSERT INTO evaluations VALUES(33460,31,'cp 400',NULL);
INSERT INTO evaluations VALUES(33460,32,'cp 386',NULL);
INSERT INTO evaluations VALUES(33460,33,'cp 388',NULL);
INSERT INTO evaluations VALUES(33460,34,'cp 398',NULL);
INSERT INTO evaluations VALUES(33460,35,'cp 400',NULL);
INSERT INTO evaluations VALUES(33460,36,'cp 419','c7c6 d2d4 c5d4 f3d4 e5d4 e4e5 f6h4 g2g3 h4h3 c1h6 h3h6 c3e4 e8g8 e4d6 b7b5 c4b3 b8a6 d1d4 a6c7 f2f4 c7e6 d4e3 g8h8 a2a4 b5b4 b3c4 a7a5 e3d3 f7f5 d6f5 h6g6 f5d6 g6d3 c4d3 e6c5 d3c4 c8a6 c4a6 c5a6 a1d1 b4b3 c2b3 a6c5 f4f5 c5b3');
INSERT INTO evaluations VALUES(36447,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(36447,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36447,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36447,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(36447,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(36447,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36447,36,'cp 0','g6g5 h4g3 f6h5 f1e2 e7e6 e1g1 f7f5 h2h3 h5g3 f2g3 b8c6 d4d5 e6d5 d1d5 g8h8 e4f5 c8f5 d5d2 f5h7 g1h2 c6e5 a1d1 h7g6 f3e5 g7e5 c3d5 d8d7 e2d3 g6h5 d3e2');
INSERT INTO evaluations VALUES(40692,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(40692,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(40692,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(40692,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(40692,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(40692,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(40692,36,'cp 55','d2d4 b8c6 h2h3 e5d4 f3d4 c6d4 d1d4 e8g8 b1c3 c7c6 a2a4 d8b6 d4b6 a7b6 c1e3 f6d7 e1d1 f8d8 f2f4 d8e8 g1f2 d7c5 f2f3 g7g5 g2g3 c8e6 c4e6 f7e6 e4e5 d6d5 b2b3 c5d7 c3e2 e7c5 e2d4 c5d4 e3d4 g8f7 h3h4 c6c5 d4f2 e8g8 h4h5');
INSERT INTO evaluations VALUES(43813,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43813,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43813,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43813,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43813,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43813,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43813,36,'cp 0','h7h6 g5f6 d8f6 e2e3 c8b7 g1e2 b8c6 e2g3 f8b4 a1c1 c6e7 g3h5 f6h4 h5g7 e8d8 g7h5 e7f5 h5g3 f5g3 f2g3 h4g5 e1f2 g5f5 f2g1 f5g5');
INSERT INTO evaluations VALUES(45429,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(45429,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45429,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(45429,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(45429,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(45429,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45429,36,'cp 0','d8b6 d1c2 b8c6 b1c3 a7a6 a1c1 c6d4 f3d4 b6d4 e2e4 e7e6 f4e3 d4b4 a2a3 b4a5 e4e5 f6d7 f2f4 d7b8 c2b3 b8d7');
INSERT INTO evaluations VALUES(49112,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(49112,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(49112,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(49112,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(49112,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(49112,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(49112,36,'cp 33','c7c6 f1d3 f8d6 e1g1 f6e4 a2a4 a7a5 c3e2 e8g8 b2b3 d8e7 c1b2 c8d7 f3e5 b8a6 f2f3 e4g5 a1c1 a6b4 c4c5 d6c7 d3b1 g5f7 f3f4 b7b6 c5b6 c7b6 e5f3 g8h8 h2h3 d7e8 d1e1 b6c7 b2a3 f8g8 f3e5 c7d6');
INSERT INTO evaluations VALUES(51430,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(51430,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(51430,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(51430,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(51430,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(51430,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(51430,36,'cp 26','d3d4 c8b7 f1d3 c7c5 c2c3 c5d4 c3d4 b8c6 g1e2 c6b4 b1c3 b4d3 d1d3 f8b4 d4d5 g8e7 e1g1 e8g8 a2a3 b4c3 e2c3 f7f5 d5e6 d7e6 d3d8 a8d8 c1g5 f8e8 e4f5 e6f5 f1d1 d8d1 a1d1 h7h6 g5e7 e8e7 f2f3 b7c6 d1d6 e7c7 g1f2 g8f7 g2g3 c6e8');
INSERT INTO evaluations VALUES(56858,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(56858,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(56858,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(56858,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56858,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(56858,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(56858,36,'cp -12','h5d1 c7c5 d2d4 c5d4 d1d4 g8f6 c1g5 f8g7 b1c3 b8c6 d4d6 h7h6 g5f4 d8b6 d6a3 d7d5 e4d5 f6d5 c3d5 e6d5 e1c1 c8e6 g1f3 e8c8 a3b3 b6b3 a2b3 h8e8 f1b5 e6g4 h1e1 a7a6 b5c6 b7c6 e1e8 d8e8 d1e1 e8e1 f3e1 g4f5 e1f3 c6c5 f4d6 c5c4 d6a3 g6g5 b3c4 d5c4');
INSERT INTO evaluations VALUES(57505,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(57505,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(57505,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(57505,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(57505,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(57505,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(57505,36,'cp 54','f5e4 f3e5 g8f6 f1e2 f8e7 c2c4 e7b4 b1d2 e8g8 e1g1 d7d6 e5g4 c8g4 e2g4 g8h8 d2b3 d6d5 c4c5 b8c6 a2a3 b4a5 g4e2 a7a6 c1f4 d8d7 d1c2 d7e7 a1d1 h7h6 f4e5 a8e8 e5g3 e8d8 h2h3 e7f7 b3a5 c6a5 c2c3');
INSERT INTO evaluations VALUES(57471,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(57471,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(57471,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(57471,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(57471,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(57471,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(57471,36,'cp 64','f3d2 d8h4 g2g3 h4e7 d4d5 e4e3 f2e3 g8f6 b1c3 d7d6 d2e4 f6e4 c3e4 g7g6 f1b5 c8d7 a2a4 a7a6 b5d7 b8d7 e1g1 e8c8 d1g4 f8g7 c1d2 d8f8 a4a5 f8f1 a1f1 h8f8 f1f8 g7f8 c2c4 e7e8 b2b4 f8e7 e4f2 e8f7 g1g2 h7h5 g4e2 g6g5 g3g4 d7f6 g4h5 f6h5 e2g4 c8b8 g4e6 f7e6 d5e6');
INSERT INTO evaluations VALUES(743,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(743,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(743,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(743,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(743,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(743,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(743,36,'cp 80','c2c3 b7b6 e3d2 c5c4 b2b4 c6b4 c3b4 f8b4 d1c2 b4d2 f3d2 b6b5 a4c3 b5b4 c3d1 e8g8 d2f3 f7f6 f1e2 a5b6 d1f2 a7a5 c2d2 c8a6 d2e3 a5a4 e1g1 b4b3 f2d1 f6e5 f4e5 a4a3 d1c3 b3b2 a1b1 a8c8 h2h3 d7b8 f3g5 h7h6 g5e6 b6e6 e2g4 e6g6 g4c8 f8f1 b1f1 a6c8 e3f3');
INSERT INTO evaluations VALUES(742,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(742,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(742,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(742,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(742,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(742,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(742,36,'cp 86','b7b6 e3d2 c5c4 b2b4 c6b4 c3b4 f8b4 d1c2 b4d2 f3d2 b6b5 a4c3 b5b4 c3e2 e8g8 g2g3 a5a3 e1f2 d7b6 h2h4 b6a4 a1b1 c8d7 d2f3 f8c8 f4f5 e6f5 f1h3 a8b8 h3f5 b4b3 a2b3 c4b3 f5h7 g8h8 c2d3 a4b2 b1b2 a3b2 f3g5 d7e6');
INSERT INTO evaluations VALUES(1602,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(1602,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(1602,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(1602,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(1602,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(1602,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(1602,36,'cp -26','e7e5 g1f3 g8f6 c1e3 e5d4 c3d4 f8b4 b1c3 d5a5 d1b3 e8g8 f1c4 a5h5 c4e2 c8e6 b3c2 f6d5 e1g1 a8e8 c3d5 h5d5 a2a3 b4d6 f1e1 h7h6 a1d1 e8d8 e2d3 d5h5 d3e2 e6d5 f3e5 h5h4 f2f4 c6e7 e2c4 d5c4 e5c4 e7d5 e3d2');
INSERT INTO evaluations VALUES(1714,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(1714,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(1714,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(1714,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(1714,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(1714,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(1714,36,'cp -28','g1f3 c8f5 b1d2 e5d4 f1c4 d5c5 e1g1 d4c3 b2c3 e8c8 d1a4 c5a5 a4a5 c6a5 c4f7 a5c6 d2b3 g8f6 b3d4 c6d4 f3d4 f5g6 f7g6 h7g6 d4e6 d8d5 e6f8 h8f8 f1e1 f8e8 e1e8 f6e8 c1e3 d5a5 f2f3 e8d6 g1f2 d6f5 f2e2 f5e3 e2e3');
INSERT INTO evaluations VALUES(4235,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(4235,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(4235,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(4235,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(4235,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(4235,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(4235,36,'cp 217','e4e5 f6e4 d4d5 c6e7 d1e2 f7f5 b5c4 d7d6 b1c3 h7h6 a2a4 g8h8 a4a5 b6c5 c3e4 f5e4 e2e4 c8f5 e4h4 e7g6 h4d8 a8d8 e5e6 f5e4 f3d2 e4c2 a1a2 c2f5 h2h3 g6e7 b2b4 c5b4 g2g4 f5h7 f2f4 c7c6 f4f5 c6d5 c4e2 b4c5 g1g2 e7f5 g4f5 f8f5 f1f5 h7f5 e2g4 f5g4 h3g4 d8f8 d2f3');
INSERT INTO evaluations VALUES(11851,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(11851,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(11851,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(11851,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(11851,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(11851,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(11851,36,'cp -37','h7h6 g5e3 f8e7 b1c3 e8g8 h2h3 f8e8 e1g1 e7f8 f1e1 a7a5 e3d2 g7g6 a1b1 f8g7 a2a3 f6d7 b2b4 d7f8 c3e2 a5b4 a3b4 f8e6 b1a1 c8d7 c2c4 g8h7 a1a5 f7f5 e4f5 g6f5 e2g3 d8f6 g3h5 f6f7 h5g7');
INSERT INTO evaluations VALUES(11880,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(11880,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(11880,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(11880,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(11880,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(11880,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(11880,36,'cp 70','b1c3 a8b8 e1g1 g7g6 b2b3 f8g7 c1b2 e8g8 d1d2 f8e8 a1e1 f6h5 c3d1 g7b2 d1b2 d8f6 b2c4 h5f4 e1e3 f4d3 c2d3 f6d4 d2a5 b8b7 h2h3 c8d7 f1e1 g8g7 a5d2 b7b8 e3e2 c6c5 e1d1 d7c6 c4e3 d4b4 f2f3 b4d4 g1h2');
INSERT INTO evaluations VALUES(14281,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(14281,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(14281,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(14281,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(14281,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(14281,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(14281,36,'cp -43','c1f4 g7g6 e2e3 f8g7 f1d3 e8g8 g1e2 f8e8 f4g5 e7e5 e1g1 h7h6 g5h4 e5d4 e3d4 c6e7 g2g4 c7c5 d4c5 b7b6 c5c6 e7c6 d3b5 c8d7 b5c6 d7c6 e2d4 d8c8 b1a3 f6d7 h4g3 d7e5 d1d2 c8d7 g3e5 e8e5 a1e1 a8e8 e1e5 e8e5 a3c2 e5e7 g1g2 c6b7 h2h3 b7a6');
INSERT INTO evaluations VALUES(16030,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(16030,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(16030,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(16030,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(16030,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(16030,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(16030,36,'cp 25','e2e3 e7e6 c2c4 h7h6 g5h4 f8e7 b1c3 e8g8 d1c2 c7c5 c4d5 c5d4 f3d4 f6d5 h4e7 d5e7 f1e2 d7f6 e1g1 c8d7 c2b3 d8a5 f1d1 a8b8 a2a4 f8c8 b3a3 a5c5 a3c5 c8c5 c3b5 e7c6 a4a5 c6d4 b5d4 b8c8 f2f3 a7a6');
INSERT INTO evaluations VALUES(16029,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(16029,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(16029,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(16029,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(16029,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(16029,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(16029,36,'cp 25','e7e6 c2c4 h7h6 g5h4 f8e7 b1c3 e8g8 d1c2 c7c6 a2a3 a7a6 a1d1 b7b6 f1e2 c8b7 e1g1 a8c8 e3e4 f6e4 h4e7 d8e7 c3e4 d5e4 c2e4 d7f6 e4f4 c6c5 f3e5 a6a5 d1d3 f8d8 f1d1 b7e4 d3e3 e4b7 h2h4 d8d4 d1d4 c5d4 f4d4');
INSERT INTO evaluations VALUES(15944,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(15944,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(15944,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(15944,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(15944,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(15944,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(15944,36,'cp 22','c2c4 h7h6 g5h4 f8e7 b1c3 e8g8 d1c2 c7c5 d4c5 d7c5 c4d5 e6d5 f1d3 c8g4 f3d4 c5e6 e1g1 e6d4 e3d4 g4e6 f1e1 f8e8 h2h3 a8c8 c2b3 f6h5 h4e7 e8e7 c3e2 e7d7 a1c1 d7c7 c1c7 c8c7 e1c1 c7c1 e2c1 d8c7');
INSERT INTO evaluations VALUES(15938,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(15938,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(15938,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(15938,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(15938,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(15938,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(15938,36,'cp 21','h7h6 g5h4 f8e7 b1c3 e8g8 d1c2 c7c6 a2a3 b7b6 c4d5 c6d5 f1d3 c8b7 e1g1 a8c8 f1c1 f6e4 h4e7 d8e7 c2b3 e4c3 c1c3 c8c3 b3c3 f8c8 c3e1 d7f6 a1c1 e7d8 b2b4 f6e4 f3e5 e4d6 a3a4 d8e8 b4b5 c8c1 e1c1 e8c8 c1c8 d6c8 f2f3 c8d6 h2h3');
INSERT INTO evaluations VALUES(17011,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(17011,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(17011,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(17011,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(17011,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(17011,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(17011,36,'cp 139','f3e5 d7d5 e5f3 f8c5 f1e2 e8g8 f3d4 c5d4 c3d5 f6d5 e4d5 d8d5 e1g1 d4f6 e2f3 d5d6 c2c3 c7c6 f1e1 a7a5 d3d4 c8e6 g2g3 d6d7 a2a4 f8e8 c1f4 a8d8 h2h4 h7h6 f3g2 e6g4 d1d2 e8e6 e1e3 e6e3 d2e3 d8e8 g2e4 g4e6 a1e1 e6f5 e3f3 f5g4');
INSERT INTO evaluations VALUES(17305,30,'cp 401',NULL);
INSERT INTO evaluations VALUES(17305,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(17305,32,'cp 398',NULL);
INSERT INTO evaluations VALUES(17305,33,'cp 410',NULL);
INSERT INTO evaluations VALUES(17305,34,'cp 418',NULL);
INSERT INTO evaluations VALUES(17305,35,'cp 429',NULL);
INSERT INTO evaluations VALUES(17305,36,'cp 427','f2f3 e7e6 f3e4 d8h4 e1e2 h4e4 c1e3 c6b4 e2d2 b4d5 e3f2 e4f4 d2e1 b7b6 d1d2 c8b7 g1f3 f4d2 e1d2 f7f6 c2c4 d5f4 b1c3 f6e5 f3e5 f8d6 h1g1 e8g8 f2h4 f8f5 a1e1 f5e5 d4e5 d6c5 h4g3 a8f8 g3f4 f8f4 f1d3 c5g1 e1g1 f4f2 d3e2 f2f4 g2g4 f4f7 h2h4');
INSERT INTO evaluations VALUES(19574,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(19574,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(19574,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(19574,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(19574,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(19574,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(19574,36,'cp -20','g8f6 c2c4 e7e5 b1c3 f8b4 c1d2 e8g8 c3d5 b4c5 f3g5 d7d6 d1c2 g7g6 d5f6 d8f6 g5e4 f6e7 e2e3 c6b4 c2b1 f7f5 e4c5 d6c5 f1e2 f5f4 d2b4 c5b4 e1g1 f4e3 f2e3 c8e6 b1e4 e6f5 e4d5 g8h8 a1d1 a8d8 d5f3 b7b6 d1d8 f8d8 f1d1 d8d1 e2d1 h8g7 e3e4 f5e6 d1e2 a7a5');
INSERT INTO evaluations VALUES(19590,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(19590,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(19590,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(19590,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(19590,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(19590,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(19590,36,'cp -40','e5c4 c3c4 e7e6 g1f3 a7a6 e1g1 b7b5 c4e2 f8e7 a2a4 c8b7 a4b5 a6b5 a1a8 d8a8 b1c3 b5b4 c3b5 a8b8 c2c4 b4c3 d2c3 g8f6 e4e5 f6e4 c3c4 e8g8 c1f4 g7g5 f4e3 b7c6 f1d1 g8g7 b5d6 e7d6 e5d6 f7f6 f3e1 c6b7 e1d3 b8d6');
INSERT INTO evaluations VALUES(19598,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(19598,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(19598,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(19598,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(19598,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(19598,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(19598,36,'cp -37','c3c4 e7e6 g1f3 g8f6 c4e2 b7b5 e1g1 c8b7 d2d3 d7d6 c1g5 f8e7 b1d2 e8g8 g5h4 f6h5 h4e7 d8e7 a2a4 b5b4 g2g3 e6e5 f3h4 g7g6 h4f5 e7e6 a4a5 g6f5 e2h5 f5e4 d2e4 f7f5 e4g5 e6e7 a5a6 b7c6 f2f4 e5f4 f1f4 a8e8 h5h6');
INSERT INTO evaluations VALUES(22566,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(22566,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(22566,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(22566,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(22566,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(22566,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(22566,36,'cp 24','d7d5 e4e5 c8d7 b5e2 f7f6 g1f3 f6e5 f3e5 f8d6 f2f4 g8e7 e1g1 e7f5 c2c3 e8g8 e5d7 d8d7 d1d3 c6e7 b1d2 c7c5 d2f3 c5d4 c3d4 e7c6 c1d2 a8c8 a1c1 d6e7 f3e5 d7e8 d2c3 e7f6 g2g4 f5d6 d3d1');
INSERT INTO evaluations VALUES(24222,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(24222,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(24222,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(24222,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(24222,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(24222,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(24222,36,'cp 57','c1f4 e8g8 f1e1 a7a6 a2a4 f8e8 d4c6 d7c6 a4a5 e7f8 d1f3 g7g6 h2h3 f8g7 a1d1 f6h5 f4e3 d8e7 e3d4 a8d8 d4g7 g8g7 b2b4 h5f6 c4b3 e7e5 f3d3 f6h5 f2f3 h5f4 d3d2 e5g5 g1h2 e8e5 g2g3 f4h5 c3e2 g5d2 d1d2 d8e8 e1d1 h5f6 e2d4 c6d7');
INSERT INTO evaluations VALUES(25737,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(25737,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(25737,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(25737,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(25737,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(25737,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(25737,36,'cp 149','g7g6 b1c3 b7b5 a4b3 h7h6 e1g1 f8g7 a2a3 c6a5 b3a2 g8e7 h2h3 e8g8 f1e1 f8e8 c3e2 c7c5 c1f4 c5d4 e2d4 a5c4 a2c4 b5c4 f4d6 e7c6 e4e5 d8b6 a1c1 b6b2 c1c4 a8c8 e5e6 f7e6 d4c6 d7c6 f3e5 c6d5 c4c8 e8c8 e5g6');
INSERT INTO evaluations VALUES(26801,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(26801,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(26801,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(26801,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(26801,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(26801,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(26801,36,'cp 57','c1f4 f8d6 f4d6 d8d6 e2e3 a7a6 f3h4 f5e4 c4d5 e6d5 d1b3 g8e7 f1e2 e8g8 e1g1 c6d8 c3e4 d5e4 g2g3 c7c6 f2f3 e4f3 h4f3 b7b5 e3e4 c6c5 b3a3 b5b4 a3d3 c5d4 d3d4 d6d4 f3d4 e7c6 d4f3 d8e6 a1d1 a8d8 d1d8');
INSERT INTO evaluations VALUES(28285,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(28285,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(28285,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(28285,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(28285,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(28285,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(28285,36,'cp -20','a3a4 f6e4 d1c2 f7f5 a4a5 e8g8 e2e3 d8f6 f1e2 e4g5 a5b6 a7b6 a1a8 b7a8 d4d5 a8b7 f3g5 f6g5 e1g1 b8a6 e2d3 h7h6 c1a3 d7d6 f2f4 g5f6 e3e4 f5e4 d3e4 e6e5 f4f5 b7c8 g2g4 f6h4 c2g2 c8d7 g1h1 f8a8 a3b4 a6c5 b4c5 b6c5 h2h3 d7a4');
INSERT INTO evaluations VALUES(33472,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(33472,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(33472,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(33472,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(33472,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(33472,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(33472,36,'cp 189','b8c6 b1c3 d7d5 c1g5 c8e6 e4d5 f6d5 g5e7 e8e7 c3d5 e6d5 g1e2 h8e8 e1c1 a8d8 e2c3 e7f8 d3d4 c6e7 c3d5 e7d5 g2g3 c7c6 f1g2 h7h5 c2c3 d5f6 g2f3 f6g4 d1f1 g7g6 h2h3 g4f6 f1e1 h5h4 e1e8 f6e8 h1e1 f8g7 g3h4 e8d6 c1c2 d8h8');
INSERT INTO evaluations VALUES(36759,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(36759,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36759,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(36759,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(36759,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36759,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36759,36,'cp 17','h2h3 c7c5 c2c3 b7b6 e1g1 b8c6 a2a4 e7e6 b1d2 c8b7 d1b1 d8e7 a4a5 a8c8 a5b6 a7b6 b1c2 c8a8 f4g3 f8c8 c2b3 f6d7 e2b5 c6a5 b3d1 d7f6 d1e2 f6e8 b5d3 e8d6 f1e1 a5c6 f3e5 c8d8 d2f3 c6e5 g3e5 g7e5 f3e5');
INSERT INTO evaluations VALUES(36755,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(36755,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36755,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36755,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(36755,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(36755,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(36755,36,'cp 23','c7c5 c2c3 b8c6 e1g1 f6d7 b1d2 f8e8 f4g3 c5d4 e3d4 e7e5 d4e5 d7e5 f1e1 e5f3 e2f3 e8e1 d1e1 c8e6 a1d1 d5d4 f3c6 b7c6 d2f3 c6c5 c3d4 c5d4 g3e5 g7e5 e1e5 d8d5 e5d5 e6d5 f3d4 d5a2 b2b3');
INSERT INTO evaluations VALUES(36844,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(36844,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(36844,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(36844,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(36844,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(36844,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(36844,36,'cp 59','a7a5 f2f3 f6d7 e1g1 b8a6 c1e3 f7f5 e4f5 g6f5 g1h1 g8h8 d3c2 a6b4 c2b1 b4a6 d1d2 d8e7 b1c2 a6b4 c2d1 f5f4 e3f2 b4a6 d1c2 d7c5 a2a3 c8f5 c2f5 f8f5 f2c5 a6c5 b2b4 c5b3 d2c2 f5h5 c2b3 e7h4');
INSERT INTO evaluations VALUES(35615,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(35615,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(35615,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(35615,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(35615,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(35615,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(35615,36,'cp 64','f2f3 f6d7 c1e3 g7h6 e3f2 b8a6 a2a3 d8g5 e1g1 a6c5 c3b5 g5d8 b2b3 c5d3 d1d3 b7b6 b3b4 c8a6 d3c2 h6g5 b4a5 a6b5 c4b5 a8a5 a3a4 a5a8 g1h1 h7h5 e2c1 d7c5 f2c5 d6c5 c1b3 h5h4 a4a5 g5f4 a5a6 d8g5 f1e1 g5h6');
INSERT INTO evaluations VALUES(38292,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(38292,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(38292,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(38292,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(38292,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(38292,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(38292,36,'cp 22','e2e3 g8f6 c2c4 e8g8 b1c3 b7b6 f1d3 c8a6 c4d5 f6d5 c3d5 d8d5 e1g1 d5b7 a1c1 c7c5 d4c5 b6c5 d3a6 b8a6 a2a3 f8d8 d1c2 a8c8 f1d1 a6c7 d1d8 c8d8 f3g5 g7g6 g5e4 c5c4 f4g5 e7g5 e4g5 b7b3 c2b3 c4b3 g5f3 c7d5');
INSERT INTO evaluations VALUES(40028,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(40028,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(40028,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(40028,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(40028,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(40028,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(40028,36,'cp 156','e4e5 f6g8 c1f4 b8c6 d4c6 b7c6 f1c4 g8h6 d1e2 e8g8 e1c1 d8a5 a2a3 a8b8 h1e1 b8b7 c4b3 d7d5 e5d6 h6f5 c3e4 f5d4 e2d2 a5b5 d6e7 b7e7 e4d6 d4b3 c2b3 b5c5 d2c2 c5c2 c1c2 e7e1 d1e1 c8e6 b3b4 h7h5 d6e4 f8e8 c2c1 e8d8 f4e3 d8a8 e3d2 a8d8 g2g3 e6b3 e4c5');
INSERT INTO evaluations VALUES(41771,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(41771,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(41771,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(41771,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(41771,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41771,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(41771,36,'cp 25','f4d6 d8d6 g1f3 e8g8 c4c5 d6e7 b1c3 a7a6 f1e2 b8c6 d1b3 f8d8 a1d1 a8b8 e1g1 b7b6 c5b6 c7b6 d1c1 c6a5 b3c2 b6b5 c3b1 c8d7 c2c5 e7c5 c1c5 f6e4 c5c7 b8c8 f1c1 a5c4 c7c8 d8c8 b2b3 c4d6 c1c8 d6c8 f3e5 d7e8 e5d3 a6a5 d3c5 c8d6 c5e4 d5e4');
INSERT INTO evaluations VALUES(41673,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(41673,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41673,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(41673,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(41673,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(41673,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(41673,36,'cp 12','d8d6 g1f3 e8g8 f1e2 d6e7 b1d2 b7b6 a1c1 b8d7 d1a4 c7c5 c4d5 f6d5 a4a3 c8b7 e1g1 f8c8 e2b5 d5f6 d4c5 d7c5 f3d4 h7h6 b5e2 c8c7 f1d1 a7a6 b2b4 c5e4 c1c7 e7c7 d2e4 f6e4 d1c1 c7e7 d4c6 e7d6');
INSERT INTO evaluations VALUES(33413,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(33413,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(33413,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(33413,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(33413,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(33413,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(33413,36,'cp 12','g1f3 c7c5 c4d5 f6d5 f1c4 d5f6 d1a4 b8d7 e1g1 e8g8 f1d1 b7b6 b1c3 c8b7 c4e2 f8d8 a4a3 d6f8 d4c5 d7c5 d1d8 f8d8 a1d1 d8c7 d1c1 c5d7 f3d2 a7a6 h2h3 b6b5 d2b3 c7b6 a3a5 a8c8 a5b6 d7b6 f2f3 f6d5');
INSERT INTO evaluations VALUES(44564,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(44564,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(44564,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(44564,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(44564,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(44564,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(44564,36,'cp -25','d7d5 e4d5 f6d5 g2g3 d5c3 b2c3 g7g6 f1g2 f8g7 g1e2 b8c6 e1g1 e8g8 c1e3 d8d6 a1b1 f8d8 e3f4 e7e5 f4g5 d8d7 a2a4 a8b8 c3c4 b7b6 e2c3 h7h6 g5e3 c8b7 c3d5 b8d8 f2f4 c6e7 d5e7 d6e7 g2b7 e5f4 e3f4 d7b7 f1e1 e7f6 f4e5 f6f5 e5g7 g8g7');
INSERT INTO evaluations VALUES(47815,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(47815,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(47815,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(47815,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(47815,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(47815,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(47815,36,'cp -29','f6e4 c4f7 e8f7 c3e4 d8h4 g2g3 h4g4 d2d3 g4d1 e1d1 b8c6 f4e5 c8g4 d1e1 c6e5 c1f4 e5c6 e1d2 f8e7 a1f1 f7e8 f4g5 h8f8 f1f8 e7f8 g5e3 e8d7 h2h3 g4f5 g1e2 b7b5 g3g4 f5e4 d3e4 f8e7 h1f1 e7f6 c2c3 a8f8 e2g3 d7e6 b2b3 b5b4');
INSERT INTO evaluations VALUES(49242,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49242,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49242,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49242,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49242,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49242,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49242,36,'cp -2','c7c5 c2c3 d7d5 e2e3 d8b6 d1c2 b8c6 f1e2 c8d7 h4f6 g7f6 e1g1 a8c8 d4c5 b6c5 f1d1 f8g7 b1d2 f6f5 d2f1 c5e7 a1c1 e8g8 f1g3 c6e5 f3d4 a7a6 g3h5 g7h8 c2b3 c8c7 h5f4 f8d8 f4h5 d8c8 h5f4');
INSERT INTO evaluations VALUES(50938,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(50938,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(50938,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(50938,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(50938,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(50938,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(50938,36,'cp 117','b1c3 f8e7 f1d3 e6e5 g1e2 b8c6 d4d5 c6b8 e2g3 g8f6 h2h3 b8d7 d3c2 h7h5 g3f5 e7f8 a2a3 g7g6 f5h4 f6h7 h4f3 f8e7 b2b4 h7g5 e1g1 e8g8 c1e3 g5f3 d1f3 g8g7 f3e2 e7g5 c4c5 g5e3 e2e3 d8e7 f2f4 e5f4 e3f4');
INSERT INTO evaluations VALUES(50936,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(50936,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(50936,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(50936,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(50936,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(50936,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(50936,36,'cp 94','c7c5 d4c5 d6c5 d1d8 e8d8 c1f4 c8d7 e1c1 d8e8 g1f3 g8e7 f3e5 f7f6 e5d7 b8d7 c3a4 a8c8 g2g3 e7c6 f4e3 c6d4 f2f4 g7g5 f1h3 b7b5 c4b5 a6b5 a4c3 b5b4 c3a4 h7h5 c1b1 g5g4 h3f1 h5h4 e3d4 c5d4 f1b5 c8b8 b5d7 e8d7 d1d4 d7e7 h1d1 e7e8 d4d2 h4g3');
INSERT INTO evaluations VALUES(51398,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(51398,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(51398,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(51398,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(51398,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(51398,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(51398,36,'cp 65','d2d4 f8b4 c1d2 b4d2 d1d2 d7d5 f1d3 d5c4 d3c4 c8b7 b1c3 g8f6 d2e3 e8g8 g1f3 b8d7 e4e5 f6d5 c3d5 b7d5 c4d5 e6d5 e1g1 f8e8 f3e1 a8c8 a1d1 a7a5 e3d3 c7c5 f2f4 d7f8 e1c2 f8e6 f4f5 e6d4 c2d4 c5d4 d3d4 d8c7 d1c1 c7d7 c1c8 e8c8 d4b6');
INSERT INTO evaluations VALUES(53044,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(53044,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(53044,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(53044,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(53044,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(53044,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(53044,36,'cp 74','d4c5 d8a5 b1c3 a5c5 e3e4 g8f6 g1f3 b8d7 c1e3 c5c7 c3d5 c7d8 c2c3 e7e5 e3g5 f8e7 g5f6 d7f6 d5f6 e7f6 f1c4 d8c7 c4d5 e8g8 e1g1 c8e6 d5e6 f7e6 d1b3 g8f7 f1d1 c7c6 a2a4 a8c8 b3c2 f7g8 c2e2 a7a6 h2h3 f6e7');
INSERT INTO evaluations VALUES(54712,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(54712,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(54712,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(54712,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(54712,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(54712,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(54712,36,'cp 159','e2e4 d5e4 c3e4 b8c6 g1f3 e7e5 f1b5 g8e7 d1e2 a7a6 b5a4 f6f5 e4g5 e5e4 a4b3 e7d5 e1g1 f8e7 f3e5 c6e5 d4e5 e7g5 c1g5 d8g5 b3d5 g5e7 e2c4 e8f8 a1d1 c7c6 d5f7 c8d7 e5e6 d7e8 f7e8 a8e8 d1d7 e7e6 c4c5 f8g8 d7b7 h7h6 c5c3');
INSERT INTO evaluations VALUES(54718,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(54718,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(54718,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(54718,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(54718,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(54718,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(54718,36,'cp 155','d5e4 c3e4 e7e5 g1f3 b8c6 f1c4 c8f5 e4g3 f5g4 d4d5 c6d4 c2c3 d4f3 g2f3 g4d7 d5d6 f8d6 d1d5 d8e7 d5b7 a8d8 h1g1 e8f8 g3e4 d7e6 c4b3 e6b3 a2b3 e7e6 a1a6 e6d7 e4d6 c7d6 b7a7 d7a7 a6a7 g8e7 e1e2 f8f7 c1e3 d8a8 g1a1 a8a7 a1a7');
INSERT INTO evaluations VALUES(55373,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(55373,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(55373,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(55373,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(55373,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(55373,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(55373,36,'cp 93','g8f6 c1f4 f8g7 e2e3 e8g8 f1e2 d5c4 e2c4 b7b6 c3b5 f6e8 h2h4 h7h6 d4d5 a7a6 b5c3 d8e7 d5e6 c8e6 c4e6 e7e6 f3d4 e6c4 d1f3 e8d6 f4d6 c7d6 f3d5 c4d5 c3d5 g7d4 e3d4 b8d7 e1g1 a8e8 a1c1 g8g7 f1d1 e8e2 c1c7');
INSERT INTO evaluations VALUES(56528,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(56528,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(56528,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(56528,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(56528,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(56528,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(56528,36,'cp -32','d8d1 c3d1 f8b4 d1c3 b8c6 e2e4 c6e5 c1f4 e5d3 f1d3 c4d3 e1c1 b4c3 b2c3 f7f5 d1e1 g8f6 e4f5 e8f7 g1f3 c8f5 e1e5 f5g6 h1e1 h8f8 f3h4 a8d8 c1d2 d8d7 f2f3 f6d5 h4g6 f7g6 f4g3 f8d8 h2h4 b7b5 e1e4 c7c5 c3c4 d5f6 e4f4 b5c4 f4c4 d7d5 e5d5 d8d5');
INSERT INTO evaluations VALUES(34954,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(34954,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(34954,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(34954,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(34954,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(34954,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(34954,36,'cp -11','c3d1 f8b4 d1c3 b8c6 e2e4 c6e5 c1f4 b4d6 f4e5 d6e5 f1c4 g8f6 g1e2 c7c6 f2f4 e5c7 h2h3 b7b5 c4d3 c8e6 e4e5 e8c8 e1c1 f6d5 c3d5 e6d5 d3f5 c8b8 e2c3 d5g2 d1d8 h8d8 h1g1 g7g6 g1g2 g6f5 g2g5 f7f6 g5f5 f6e5 f4e5 d8e8 f5f6 c7e5 f6c6 b8b7 c6c5 a7a6 a2a3 h7h5 c1c2 b7b6 b2b4 e5c3');
INSERT INTO evaluations VALUES(34955,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(34955,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(34955,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(34955,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(34955,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(34955,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(34955,36,'cp -32','f8b4 d1c3 b8c6 e2e4 c6e5 c1f4 b4d6 f4e5 d6e5 g1f3 e5c3 b2c3 g8f6 e4e5 f6e4 f1c4 c8g4 h2h4 h7h5 f3g5 e4g5 h4g5 a8d8 f2f3 g4e6 c4e6 f7e6 e1e2 d8d5 f3f4 d5c5 a1d1 b7b6 g2g4 h8f8 g4h5 f8f4 h5h6 c5e5 e2d3 g7h6 g5h6');
INSERT INTO evaluations VALUES(819,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(819,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(819,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(819,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(819,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(819,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(819,36,'cp 0','c5d4 b2b4 d4c3 b4b5 c3c2 d1c2 c6d4 f3d4 b6d4 c1b2 d4b6 a2a4 f8b4 e2d3 b6a5 f1c1 c8d7 g2g3 g7g6 h2h3 e8g8 c2c7 f8d8 c7b7 a8c8 c1c8 d7c8 b7b8 c8d7 b8b7');
INSERT INTO evaluations VALUES(2061,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(2061,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(2061,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(2061,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(2061,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(2061,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(2061,36,'cp 150','e1g1 d7d6 f3d4 c6d4 d1d4 g8f6 b1c3 e7e5 d4d3 c8e6 c4e6 f7e6 f2f4 e5c5 c1e3 c5c6 f4f5 e6e5 e3g5 f8e7 a1d1 e8g8 g2g3 a8d8 h2h4 c6d7 g5f6 e7f6 c3d5 c7c6 d5e3 b7b5 g1g2 a7a6 b2b3 g8h8 d3e2 d7b7 c2c4 b7a7 f1f3 b5c4 b3c4 a6a5 g2h3 f8f7 e3g4 a7c5 f3f2 a5a4');
INSERT INTO evaluations VALUES(2060,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(2060,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(2060,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(2060,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(2060,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(2060,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(2060,36,'cp 145','d7d6 f3d4 c6d4 d1d4 e7e5 d4d3 g8f6 b1c3 c8e6 c4e6 f7e6 f2f4 e5c5 c1e3 c5c6 f4f5 e6e5 e3g5 f8e7 a1d1 a7a6 h2h3 e8g8 g2g3 c6d7 g5f6 e7f6 g1g2 g8h8 c3d5 a8d8 h3h4 b7b5 g2h3 a6a5 b2b3 c7c6 d5e3 h7h6 c2c4 b5c4 b3c4 d7a7 d3a3 a7c5 d1d3 h6h5 a3c5');
INSERT INTO evaluations VALUES(5443,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(5443,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(5443,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(5443,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(5443,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5443,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5443,36,'cp 0','d1e2 a7a6 f1d1 d8c7 c1f4 g8f6 a1c1 e8g8 h2h3 c6e5 f4e5 d6e5 c4e6 c8e6 c3d5 f6d5 e4d5 c7d6 d5e6 d6e6 e2e5 e6e5 f3e5 f8d8 g1f1 h7h5 g2g3 d8d1 c1d1 e7f6 e5d3 a8c8 f1e2 g7g6 d1d2 f6g7 e2e3 c8e8 e3f3 e8d8 f3e2');
INSERT INTO evaluations VALUES(5442,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5442,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5442,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(5442,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(5442,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(5442,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(5442,36,'cp -17','c8d7 f1d1 a7a6 c1f4 e6e5 f4e3 g8f6 c3d5 f6d5 c4d5 e8g8 a1c1 d8e8 a2a3 a8c8 b2b4 g8h8 h2h3 h7h6 f3d2 f7f5 e4f5 d7f5 d2e4 e8g6 e4g3 f5e6 d5e4 g6f7 e2d2 e6b3 d1e1 d6d5 e4f5 c8d8 e3h6 e5e4 h6f4');
INSERT INTO evaluations VALUES(9198,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(9198,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(9198,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(9198,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(9198,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(9198,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(9198,36,'cp -53','b1c3 c8e6 d2d4 e5d4 e3d4 d8d7 g1f3 e8g8 e1g1 f8e8 c3b5 e6g4 c2c4 a8d8 b5d6 c7d6 a1c1 h7h6 f1e1 e8e1 d1e1 g4h3 e1f1 h3g2 g1g2 d8e8 c1e1 e8e4 c4d5 c6e7 f3d2 e4e1 f1e1 f6d5 e1e4 b7b5 e4f3 b5b4 d2e4 d7c6 a2a3 b4a3');
INSERT INTO evaluations VALUES(12138,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(12138,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(12138,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(12138,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(12138,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(12138,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(12138,36,'cp 50','f3d4 f8c5 d1d2 c6d4 e3d4 a7a6 e1c1 e8g8 d2e3 d8c7 f1d3 c5d4 e3d4 b7b5 c1b1 a8b8 c3e2 a6a5 d4e3 c7b6 e2d4 b5b4 g2g4 c8a6 f4f5 a6d3 c2d3 e6f5 g4f5 b8e8 e5e6 f7e6 f5e6 d7f6 h1e1 a5a4 e6e7 f8f7 d4f5 b6e3');
INSERT INTO evaluations VALUES(14919,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(14919,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(14919,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(14919,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(14919,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(14919,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(14919,36,'cp 84','e2e4 c6a5 f1c4 a5c4 d1a4 f6d7 a4c4 h7h6 c1f4 f8d6 f4g3 d7b6 c4b3 c8d7 e1g1 e8g8 f1e1 a8c8 a1d1 d6g3 h2g3 d8e7 f3e5 f8d8 f2f4 d7e8 e5f3 c7c6 b3c2 e7c7 c2e2 c6c5 d4d5 c5c4 e2c2 e6d5 e4d5 e8d7 d1d4');
INSERT INTO evaluations VALUES(15189,30,'cp -114',NULL);
INSERT INTO evaluations VALUES(15189,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(15189,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(15189,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(15189,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(15189,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(15189,36,'cp -87','e5e4 d3e4 d5e4 f3g5 e4e3 d2e4 e3f2 e4f2 d8d1 f2d1 c8d7 c1d2 h7h6 g5e4 f6e4 g2e4 e8c8 h1f1 f7f6 d1f2 f8c5 f2d3 c5b6 d2c3 h8e8 e4g6 e8e3 d3f4 c6d4 c3d4 b6d4 c2c3 d4c5 e1c1 a7a5 d1d5 b7b6 f1d1 c5d6 g6f5 d7f5 d5f5');
INSERT INTO evaluations VALUES(18946,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(18946,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(18946,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(18946,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(18946,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(18946,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(18946,36,'cp -30','g8f6 b1c3 f8b4 g1e2 f6e4 e1g1 b4c3 e2c3 e4c3 b2c3 e8g8 d1d6 f8e8 b5c4 d8e7 f1d1 e7d6 d1d6 c6a5 c4b5 b7b6 c1a3 a7a6 b5f1 b6b5 a1e1 f7f6 f2f4 a5c4 f1c4 b5c4 f4e5 a8b8 d6d4 e8e5 e1e5 f6e5 d4c4 b8b1 g1f2 c8b7 c4b4 b1b4 c3b4');
INSERT INTO evaluations VALUES(12511,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(12511,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(12511,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(12511,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(12511,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(12511,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(12511,36,'cp -39','b1c3 f8b4 g1e2 f6e4 e1g1 b4c3 e2c3 e4c3 b2c3 e8g8 d1d6 f8e8 b5c4 d8e7 f1d1 e7d6 d1d6 c6a5 c4b5 b7b6 a2a4 h7h6 c1a3 e8d8 a1d1 a5b7 d6d2 d7d6 b5c6 a8b8 a4a5 c8e6 a5b6 a7b6 h2h3 f7f6 c6f3 g8f7 a3b4 b6b5 f3e2 g7g6 b4d6 b7d6 d2d6 d8d6');
INSERT INTO evaluations VALUES(12509,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(12509,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(12509,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(12509,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(12509,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(12509,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(12509,36,'cp -37','f8b4 g1e2 f6e4 e1g1 e4c3 e2c3 b4c3 b2c3 e8g8 d1d6 f8e8 b5c4 d8e7 d6e7 c6e7 f1e1 d7d6 c1a3 e7f5 c4d5 h7h5 h2h3 e8e7 a1b1 c8e6 d5b7 a8d8 b7e4 f7f6 a3b4 e7d7 e4c6 d7f7 a2a4 d8c8 c6b5 d6d5 b5a6 c8c6 a6b5');
INSERT INTO evaluations VALUES(19715,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(19715,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(19715,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(19715,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(19715,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(19715,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(19715,36,'cp -35','c8f5 g1f3 e7e6 c1f4 g8e7 e1g1 d8b6 c2c4 d5c4 b1c3 e7d5 c3d5 e6d5 f1e1 f8e7 b5c6 b6c6 f3e5 c6a6 d1h5 f5e6 e5c4 d5c4 d4d5 e8g8 d5e6 a6c6 h5g4 f7e6 f4e5 e7f6 a1d1 f6e5 e1e5 a8d8 d1d8 f8d8 h2h3 b7b5 e5e6 c6d5 e6a6 d5d1 g4d1 d8d1 g1h2');
INSERT INTO evaluations VALUES(22858,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(22858,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(22858,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(22858,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(22858,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(22858,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(22858,36,'cp -42','d7d5 e2e4 g8f6 e4d5 f6d5 f1e2 d5c3 b2c3 f8d6 a1b1 e8g8 e1g1 h7h6 f3d2 b7b6 e2f3 d8d7 d2e4 d6e7 e4d2 c8b7 d2c4 e7d6 f1e1 a8e8 c4d6 c7d6 a2a4 f7f5 c3c4 c6a5 b1b5 b7c6 c1d2 g8h8 d2a5 c6b5 a4b5 b6a5 f3c6');
INSERT INTO evaluations VALUES(21744,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(21744,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(21744,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(21744,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(21744,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(21744,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(21744,36,'cp -19','e2e4 d5d4 c3e2 c8g4 a2a3 d8f6 e2g3 h7h5 h2h3 g4f3 d1f3 h5h4 g3e2 g7g6 c2c3 f6e7 g2g4 h4g3 c3d4 e5d4 f3g3 g8f6 f2f4 f6h5 g3f3 g6g5 h3h4 g5f4 e2f4 h5f4 c1f4 h8h4 e1c1 h4h1 f3h1 e8c8 h1h5 f7f6 c1b1 c8b8 f1e2 c6e5');
INSERT INTO evaluations VALUES(23663,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(23663,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(23663,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(23663,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(23663,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(23663,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(23663,36,'cp 82','d4c2 d8a5 e3d2 a8c8 a1c1 f6e8 g1h1 f7f5 f2f4 e7e6 e4f5 a5f5 b2b3 c6e7 c2e3 f5a5 e2f3 a5b6 d1e2 e8f6 d2e1 c8e8 e1f2 b6a5 e2d2 d6d5 f1d1 d7c6 c3d5 a5d2 d5f6 g7f6 d1d2 c6f3 g2f3 e8d8 d2d8 f8d8 c1d1 e7c6 d1d8 f6d8');
INSERT INTO evaluations VALUES(24509,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(24509,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(24509,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(24509,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(24509,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(24509,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24509,36,'cp 25','c1f4 f8e7 c2c4 e8g8 f1d1 f8e8 h2h3 a7a5 b2b3 d7f8 g3e4 f8g6 f4g3 h7h6 a2a3 f6e4 d3e4 e7d6 g3d6 d8d6 g2g3 a8d8 h3h4 g6f8 d1d2 f8d7 a3a4 d7f6 e4e2 f6g4 a1d1 d6b4 e2e4 g4f6 e4c2 c6c5 d2d3');
INSERT INTO evaluations VALUES(24507,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24507,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(24507,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24507,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(24507,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24507,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24507,36,'cp 25','f8e7 c2c4 e8g8 f1d1 f8e8 h2h3 a7a5 g3e2 a5a4 e2c3 d8a5 f4g3 a4a3 a1b1 a3b2 b1b2 a8a7 b2b3 b7b6 g3e5 e8a8 g2g4 h7h6 d1b1 a5a6 e5g3 a6c8 a2a4 e7f8 b1d1 f6e8 d3e2 e8d6 f3e5 d7e5');
INSERT INTO evaluations VALUES(26882,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26882,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26882,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26882,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26882,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26882,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26882,36,'cp 0','e1g1 d8d7 c2c4 d5c4 d3c4 e8c8 d1c2 g8f6 b1c3 f8c5 c1g5 d7e6 e2e3 e6c4 f3d2 c4e6 g5f6 g7f6 c3a4 c5b6 f1c1 c6b4 c2e4 b4c6');
INSERT INTO evaluations VALUES(26881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26881,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26881,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(26881,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26881,36,'cp 0','d8d7 c2c4 d5c4 d3c4 d7d1 f1d1 e5e4 f3g5 g4e2 d1e1 e2d3 b1c3 f7f5 f2f3 c6b4 f3e4 b4c2 e1d1 f8c5 g1h1 d3c4 b2b3 h7h6 a1b1 h6g5 b3c4 g8f6 h2h3 f6g4 d1d5 g4f2 h1h2 f2g4');
INSERT INTO evaluations VALUES(28408,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(28408,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(28408,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(28408,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(28408,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(28408,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(28408,36,'cp 148','g5f6 e7f6 f3b3 e8g8 b3b7 c7c6 b1d2 d8b6 b7b6 a7b6 a2a4 b8d7 e1e2 d7c5 b2b3 a8a7 c4d3 f8a8 d2c4 a7a6 a1b1 c5d3 e2d3 b6b5 a4b5 a8d8 d3e2 c6b5 c4e3 f6g5 e3d5 f7f5 f2f3 a6a2 e2d3 d8c8 h2h4 g5f6 d5b4 a2a7 c2c3 h7h5 b1a1 f5e4 f3e4 a7a1 h1a1 f6h4 a1a5 c8b8 b4c6 b8b6 c6e5 h4f6');
INSERT INTO evaluations VALUES(28970,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(28970,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(28970,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(28970,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28970,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(28970,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(28970,36,'cp 36','b1c3 e7e6 f3h4 d5c4 h4f5 e6f5 e2e3 b8d7 f1c4 d7b6 c4b3 f8e7 e1g1 e8g8 f2f3 d8d7 e3e4 a8d8 e4e5 f6d5 d1e2 c6c5 d4c5 e7c5 g1h1 d5c3 b2c3 d8e8 a2a4 a7a5 c1f4 d7c6 e2b5 c6b5 a4b5 e8a8 e5e6 f7e6 b3e6 g8h8 f1e1 a5a4 e6a2 f8e8');
INSERT INTO evaluations VALUES(28960,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(28960,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(28960,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(28960,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(28960,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(28960,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(28960,36,'cp 29','e7e6 f3h4 d5c4 h4f5 e6f5 e2e3 b8d7 f1c4 d7b6 c4b3 d8d7 f2f3 f8e7 e3e4 e8g8 e1g1 a8d8 e4e5 f6d5 g1g2 c6c5 d4c5 e7c5 c3d5 b6d5 d1c2 c5b6 c1d2 b6e3 d2a5 d8c8 c2d1 b7b6 a5b4 e3c5 b3d5 c5b4 d5b3 f8d8 d1d7 d8d7 f1d1 c8d8');
INSERT INTO evaluations VALUES(29243,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(29243,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(29243,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29243,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(29243,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(29243,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(29243,36,'cp 27','f3h4 d5c4 h4f5 e6f5 e2e3 b8d7 f1c4 d7b6 c4b3 a7a5 e1g1 f8e7 f2f3 e8g8 e3e4 c6c5 e4e5 f6d7 c1e3 a5a4 c3a4 b6a4 b3a4 c5d4 e3d4 d7e5 d4e5 d8a5 e5c3 a5a4 d1a4 a8a4 a2a3 e7c5 g1g2 c5d4 a1c1 d4c3 c1c3 f8d8 c3b3 g7g6 b3b7');
INSERT INTO evaluations VALUES(31575,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(31575,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31575,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31575,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(31575,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31575,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(31575,36,'cp 28','d1d3 g8f6 e1g1 f8d6 c2c4 b8d7 c1d2 e8g8 a1d1 a7a5 g3e4 f6e4 d3e4 f8e8 f1e1 d6b4 b2b3 d8c7 h2h3 a8d8 e4e2 h7h6 d2b4 a5b4 d1d2 b7b6 e1d1 c7b7 e2e3 b7c7 h3h4 d7f6 f3e5 c6c5 d4c5 d8d2 d1d2 b6c5 g2g3 e8d8 d2d8 c7d8');
INSERT INTO evaluations VALUES(31915,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31915,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(31915,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31915,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(31915,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(31915,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(31915,36,'cp 14','c7c6 f1g2 g8f6 e1g1 e7e6 h2h3 g4f5 c2c4 f8e7 b1c3 h7h6 f3d2 b8d7 e2e4 d5e4 d2e4 f6e4 c3e4 d7f6 e4c3 d8b6 c4c5 b6a6 d1f3 a8d8 f1d1 e8g8 c1f4 f5c2 d1d2 c2g6 a2a3 g6h5 g3g4 h5g6 b2b4 a6c4');
INSERT INTO evaluations VALUES(31121,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31121,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31121,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(31121,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31121,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(31121,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(31121,36,'cp 16','f1g2 b8d7 e1g1 e7e6 h2h3 g4f5 c2c4 g8f6 b1c3 f8e7 c4d5 e6d5 g3g4 f5g6 f3h4 d8c7 g4g5 f6h5 e2e4 d5e4 d4d5 d7b6 d5c6 b7c6 d1g4 e8g8 h4g6 h7g6 g4e4 b6d5 f1d1 e7d6 c3d5 c6d5 d1d5 a8d8');
INSERT INTO evaluations VALUES(31122,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31122,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31122,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31122,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(31122,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31122,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31122,36,'cp 6','b8d7 c2c4 e7e6 e1g1 g8f6 d1b3 d8b6 c4c5 b6a6 c1f4 b7b6 c5b6 a7b6 f1c1 a8c8 h2h3 g4h5 b1c3 f8e7 g3g4 h5g6 g2f1 e8g8 f3h4 b6b5 h4g6 h7g6 e2e3 a6a7 a2a3 d7b6 b3d1 b6c4 a3a4 b5b4 b2b3 c4a5');
INSERT INTO evaluations VALUES(35098,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35098,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(35098,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35098,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35098,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(35098,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(35098,36,'cp 0','b1c3 c7c6 g1f3 d7d6 d2d4 b8d7 a2a4 g7g6 h5g5 f6g5 f3g5 f8e7 g5f3 f7f5 d4e5 d7e5 f3e5 d6e5 f1c4 g8f6 e3e4 f6e4 c3e4 f5e4 c1h6 e7f8 h6g5 f8e7');
INSERT INTO evaluations VALUES(41429,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41429,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41429,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41429,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41429,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41429,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(41429,36,'cp 0','c7c5 d4c5 d8a5 g5f6 g7f6 a1c1 d5c4 f1c4 e8g8 g1f3 a5c5 d1e2 b8c6 e1g1 c8g4 c4d5 f6c3 f1d1 c5a5 c1c3 a8d8 c3a3 a5c5 a3c3');
INSERT INTO evaluations VALUES(45927,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(45927,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(45927,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(45927,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(45927,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(45927,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(45927,36,'cp 101','b1c3 g7g6 f2f4 f8g7 f1b5 f6d7 a2a4 b8a6 g1f3 a6b4 e1g1 a7a6 b5c4 e8g8 c1e3 d7b6 c4e2 c8g4 d1d2 f8e8 h2h3 g4f3 f1f3 b6d7 e3f2 a8c8 e2c4 d7b6 c4f1 b6d7 g1h1 c5c4 a1e1 d8c7 e1e3 c7a5 f2g1 b7b5 e4e5 d6e5 a4b5 a6b5');
INSERT INTO evaluations VALUES(46878,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(46878,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(46878,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(46878,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(46878,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(46878,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(46878,36,'cp 32','g5f6 e7f6 c4d5 d8d5 e2e3 d5d8 g1f3 f8d6 d1c2 e8g8 b1c3 c7c6 f1d3 f8e8 e1g1 b8d7 a2a3 d7b6 f1e1 d6f8 d3h7 g8h8 h7f5 c8f5 c2f5 h8g8 f5d3 d8d7 e3e4 g7g6 d3c2 a8d8 a1d1 f6f5 e4e5 b6d5 c2b3 d5f4 h2h4 g8g7 g2g3 f4e6 d4d5');
INSERT INTO evaluations VALUES(46875,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(46875,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(46875,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(46875,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46875,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(46875,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(46875,36,'cp 32','e7f6 c4d5 d8d5 b1c3 d5d8 e2e3 f8d6 g1f3 e8g8 d1c2 b8c6 f1c4 c6e7 h2h3 c7c6 e1g1 c8f5 c2e2 f5g6 a1d1 d8c7 e3e4 g6h5 g2g4 h5g6 e4e5 f6e5 d4e5 d6b4 e2e3 a8e8 c4b3 c7b6 d1d7 b6e3 f2e3 e8d8');
INSERT INTO evaluations VALUES(49209,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(49209,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(49209,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(49209,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(49209,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(49209,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(49209,36,'cp 50','f1g2 f8g7 d2d4 e5d4 d1d4 d7d6 g1f3 b8c6 d4e3 d8e7 e1g1 e7e3 c1e3 c8d7 c3d5 f6d5 c4d5 c6b4 f3d4 a7a5 a2a3 b4a6 b2b4 e8g8 a1b1 a5b4 a3b4 f8e8 g2f3 h7h5 f1c1 d7g4 g1g2 g4f3 e2f3 g7d4 e3d4 e8e2 d4e3 a8e8 b4b5 a6c5 e3c5 d6c5 c1c5 e2d2 b1c1 e8e2 c5c7 e2f2 g2g1 f2h2');
INSERT INTO evaluations VALUES(49269,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(49269,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(49269,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(49269,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(49269,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(49269,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(49269,36,'cp 114','f3e5 d7d5 e4d5 d8d5 e5f3 f8d6 c2c4 d5h5 d1e2 d6e7 b1c3 b8c6 c1e3 c8f5 d3d4 f5g6 e2d2 f6e4 d2c1 h5a5 a2a3 e4c3 b2c3 e8g8 f1e2 e7f6 e1g1 f8e8 f1e1 a5f5 c1b2 b7b6 b2b5 c6e7 b5f5 e7f5 e3f4 c7c5 f4e5 f6e7 e2d3 f5h4 d3g6 h4f3 g2f3 f7g6 d4d5 e7f6 e5f6 e8e1 a1e1');
INSERT INTO evaluations VALUES(49247,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(49247,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(49247,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(49247,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(49247,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(49247,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(49247,36,'cp 110','d7d5 e4d5 d8d5 e5f3 f8d6 f1e2 e8g8 d3d4 f8e8 b1c3 d5a5 e1g1 c7c6 f1e1 c8f5 a2a3 b8d7 h2h3 a7a6 e2d3 e8e1 f3e1 a8e8 e1f3 f5d3 d1d3 b7b5 c1e3 c6c5 d4c5 d6c5 e3c5 d7c5 d3d6 b5b4 c3d5 f6d5 d6d5 a5c7');
INSERT INTO evaluations VALUES(50861,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(50861,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(50861,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(50861,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(50861,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(50861,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(50861,36,'cp 90','g2g3 g8f6 f1g2 c7c6 g1f3 c8g4 e1g1 e7e6 h2h3 g4f3 g2f3 f8d6 c3e4 f6e4 f3e4 b8d7 c2c4 e8g8 d1b3 d7f6 e4g2 d8d7 b3c2 a8d8 c1e3 h7h6 f1d1 d7c7 d1d3 d6e7 a1b1 d8d7 a2a3 f8d8 d3b3 c7c8 b1e1');
INSERT INTO evaluations VALUES(51002,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(51002,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(51002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51002,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(51002,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(51002,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(51002,36,'cp 13','d7d5 e4d5 e6d5 c4b3 g8f6 e1g1 f8e7 f1e1 e8g8 c1f4 c8e6 h2h3 f8e8 c2c3 c7c5 b1d2 b8c6 d2f1 d8b6 d4c5 e7c5 d1c2 a6a5 f4e3 c5e3 e1e3 b6c5 a2a4 a8d8 a1d1 d5d4 f3d4 c6d4 c3d4 c5c2 b3c2 f6d5 e3f3');
INSERT INTO evaluations VALUES(51485,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(51485,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(51485,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(51485,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(51485,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(51485,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(51485,36,'cp 27','c8b7 g1f3 c7c5 f1g2 g7g6 d2d4 c5d4 d1d4 g8f6 b1c3 f8g7 e1g1 d7d6 b2b3 b8d7 f1d1 d7c5 c1b2 e8g8 f3e1 f6e4 d4e3 d8d7 c3e4 g7b2 a1b1 b7e4 g2e4 c5e4 b1b2 e4f6 b2d2 a8c8 f2f3 f8e8 g1g2 d7b7 e1d3 a7a6 d1c1');
INSERT INTO evaluations VALUES(30787,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(30787,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(30787,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30787,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(30787,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(30787,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(30787,36,'cp 33','g1f3 c7c5 f1g2 g8f6 e1g1 g7g6 d2d4 c5d4 d1d4 f8g7 b1c3 d7d6 b2b3 b8d7 d4h4 h7h6 f1d1 a8c8 c1b2 f6e4 c3e4 g7b2 a1b1 b2f6 h4f4 e8f8 f4e3 b7e4 e3e4 h6h5 h2h4 f8g7 b1c1 h8e8 f3d4 f6d4 e4d4 d7f6 e2e4 d8c7');
INSERT INTO evaluations VALUES(51772,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(51772,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(51772,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(51772,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(51772,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(51772,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(51772,36,'cp -30','d7d5 c4b3 g8f6 d2d4 f6e4 c3e4 d5e4 c2c3 b8d7 d1e2 c6c5 e2e4 d7f6 e4f4 c5d4 b3a4 c8d7 f4d4 b7b5 a4c2 d7c6 g1f3 a7a6 c1e3 d8c7 a1d1 f8e7 e1g1 c6f3 g2f3 a8c8 e3g5 c7b7 f1e1 h7h6 g5f6 e7f6 d4e4 b7c6 f3f4 c6e4 c2e4 c8c4 a2a3 e8g8 d1d7');
INSERT INTO evaluations VALUES(53554,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(53554,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(53554,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(53554,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(53554,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(53554,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(53554,36,'cp -26','c4b3 g8f6 d2d4 f6e4 c3e4 d5e4 c2c3 b8d7 d1e2 c6c5 e2e4 d7f6 e4f4 c5d4 b3a4 c8d7 f4d4 b7b5 a4c2 d7c6 g1f3 c6f3 d4d8 a8d8 g2f3 f8c5 e1e2 e8e7 h1g1 g7g6 g1d1 d8d1 c2d1 f6d5 d1b3 a7a5 b3d5 e6d5 c1e3 c5d6 a2a4 h8b8 a4b5 b8b5 f3f4 h7h6');
INSERT INTO evaluations VALUES(52860,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(52860,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(52860,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(52860,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(52860,34,'cp -130',NULL);
INSERT INTO evaluations VALUES(52860,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(52860,36,'cp -123','b8d7 b1c3 d7e5 d4e5 d8d1 c3d1 g7g6 h2h3 h7h5 c1f4 c8f5 d1e3 e8c8 f1c4 f8h6 f4h2 e7e6 c4e2 g8e7 e3c4 h8h7 c4d6 c8c7 c2c4 h5h4 e1g1 e4e3 f2f4 h7g7 c4c5 g6g5 f4g5 h6g5 a1d1 f5c2 d1c1 c2g6 b2b4 e7d5');
INSERT INTO evaluations VALUES(53969,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(53969,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(53969,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(53969,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(53969,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(53969,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(53969,36,'cp -45','a7a6 a2a3 b7b5 c4a2 b8c6 d2d3 a8b8 f3d1 g8f6 g1f3 f8e7 e1g1 e8g8 c3e2 d8c7 f1e1 c6e5 e2g3 c8b7 f3e5 d6e5 g3h5 f6h5 d1h5 a6a5 c1d2 e7d6 h5e2 b7c6 a1c1 h7h6 c1d1 a5a4 d1c1 g8h8 b2b3 a4b3 a2b3');
INSERT INTO evaluations VALUES(56611,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(56611,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(56611,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(56611,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(56611,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(56611,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(56611,36,'cp -73','f8c5 d1e2 b8c6 g1f3 g8f6 c2c3 c8e6 c1e3 c5b6 b1d2 e8g8 f3g5 e6d7 g5f3 d8e7 d2c4 h7h6 e3f2 b6f2 e2f2 f6g4 f2h4 g7g5 h4g3 f7f5 e4f5 e5e4 e1c1 e4f3 g3g4 f3f2 g4e2 a8e8 e2f2 d7f5 c4d2 f5g6 f2g3 g6h5 d3d4 h5d1 c1d1 f8f4 f1d3 g8h8 d1c2');
INSERT INTO evaluations VALUES(585,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(585,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(585,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(585,33,'cp 252',NULL);
INSERT INTO evaluations VALUES(585,34,'cp 272',NULL);
INSERT INTO evaluations VALUES(585,35,'cp 248',NULL);
INSERT INTO evaluations VALUES(585,36,'cp 275','c3d4 c6d4 f3d4 e7e5 c1e3 b7b5 c4b3 c8b7 d4f3 b7f3 d1f3 e8g8 b1c3 c5e3 f3e3 e5e3 f2e3 c7c6 e1c1 a8e8 e3e4 e8e5 d1d6 f8e8 b3c2 e5c5 h1d1 a7a5 h2h3 g8f8 d1f1 f8e7 d6d2 e7f8 f1f3 e8e5 d2d1 f8e7 d1f1 e7f8 b2b3 f8e7 a2a4 b5b4');
INSERT INTO evaluations VALUES(583,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(583,31,'cp 253',NULL);
INSERT INTO evaluations VALUES(583,32,'cp 262',NULL);
INSERT INTO evaluations VALUES(583,33,'cp 269',NULL);
INSERT INTO evaluations VALUES(583,34,'cp 261',NULL);
INSERT INTO evaluations VALUES(583,35,'cp 265',NULL);
INSERT INTO evaluations VALUES(583,36,'cp 246','c6d4 f3d4 e7e5 c1e3 b7b5 c4b3 c8b7 d4f3 b7f3 d1f3 e8g8 b1c3 c5e3 f3e3 e5e3 f2e3 c7c6 e1c1 d7d5 h1e1 f8e8 b3c2 a7a5 e3e4 b5b4 c3a4 f6e4 c2e4 d5e4 a4b6 a8b8 b6c4 f7f5 c4a5 b8a8 a5c4 a8a2 c1c2 a2a7 c2b3 a7e7 b3b4 e4e3 b4c3 e3e2 d1d2');
INSERT INTO evaluations VALUES(2811,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(2811,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(2811,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(2811,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(2811,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(2811,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2811,36,'cp 33','f1e1 d7d6 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 b3c2 g7g6 a2a3 c6b8 b2b4 b8d7 c1b2 c7c6 a3a4 f6h5 d2b3 d8c7 b3a5 e5d4 c3d4 d6d5 d1d2 d5e4 f3g5 d7f6 g5e4 f6e4 c2e4 h5f6 e4f3 f6d5 f3d5 c6d5 a4b5 a6b5 a5b7 e8e1 d2e1 c7b7');
INSERT INTO evaluations VALUES(6050,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6050,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6050,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(6050,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(6050,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(6050,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(6050,36,'cp 2','b1c3 g8f6 h2h3 e8g8 f1d3 f8e8 e1g1 c6e5 f3e5 d6e5 d1f3 c7c6 f1d1 d8e7 c3a4 c5d6 c1e3 c8e6 c2c4 f6d7 d3f1 a8d8 a2a3 d7c5 a4c5 d6c5 e3c5 e7c5 b2b4 c5e7 f3e3 b7b6 c4c5 b6c5 e3c5 e7c5 b4c5 d8d4 d1d4 e5d4 f2f3');
INSERT INTO evaluations VALUES(7555,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(7555,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(7555,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(7555,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(7555,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(7555,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(7555,36,'cp 24','d2d4 e5d4 f3d4 c8d7 h2h3 e8g8 f1e1 f8e8 d4f3 d8c8 c1f4 a8b8 e4e5 d6e5 e1e5 c5d6 e5e8 d7e8 f4d6 c7d6 b2b3 d6d5 d1d2 c8c7 a1e1 b8c8 f3e5 c8d8 e5d3 c6c5 d2e3 d5d4 c3d5 d4e3 d5c7 e3f2 g1f2 e8c6 d3c5 d8d2 e1e2 d2e2 f2e2 c6g2 h3h4 h7h6 e2f2');
INSERT INTO evaluations VALUES(7554,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(7554,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7554,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(7554,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7554,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(7554,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(7554,36,'cp 23','e5d4 f3d4 c8d7 h2h3 e8g8 f1e1 h7h6 d4f3 d7e6 e4e5 d6e5 f3e5 e6d5 d1d3 f8e8 c1f4 e8e6 e5f3 c5b6 c3d5 e6e1 a1e1 c6d5 b2b3 f6e4 f4e3 d8d6 c2c4 e4f6 e3d4 d5c4 d3c4 d6d5 c4c3 b6d4 c3d4 d5d4 f3d4 c7c5');
INSERT INTO evaluations VALUES(7560,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(7560,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(7560,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(7560,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7560,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(7560,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(7560,36,'cp 30','f3d4 c8d7 h2h3 e8g8 f1e1 f8e8 d4f3 d8c8 c1f4 a8b8 e4e5 d6e5 e1e5 c5d6 e5e8 d7e8 f4d6 c7d6 b2b3 c8c7 d1d3 h7h6 a1e1 e8d7 e1e7 g8f8 e7e3 b8e8 e3e8 f6e8 f3d2 e8f6 d2e4 f6e4 d3e4 d7e6 e4f3 d6d5 c3e2');
INSERT INTO evaluations VALUES(7559,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(7559,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(7559,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(7559,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(7559,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7559,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7559,36,'cp 13','c8d7 h2h3 e8g8 f1e1 f8e8 d4f3 d8c8 c1f4 a8b8 e4e5 d6e5 e1e5 c5d6 e5e8 d7e8 f4d6 c7d6 b2b3 c8c7 d1d2 d6d5 a1e1 b8c8 f3e5 c7d6 c3d1 d6c7 f2f3 h7h6 e5g4 f6g4 h3g4 c8d8 c2c3 e8d7 d1f2 c7d6 e1d1 d8e8 f2d3 c6c5 d3f4');
INSERT INTO evaluations VALUES(7859,30,'cp -112',NULL);
INSERT INTO evaluations VALUES(7859,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(7859,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(7859,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(7859,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(7859,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(7859,36,'cp -83','d7d5 e4d5 f6d5 a2a3 b4c3 b2c3 c8f5 c3c4 c6d4 b3d4 d5c3 d1d2 e5d4 f1d3 f5d3 d2d3 e8g8 e1g1 d8f6 c1d2 f8d8 a3a4 a8c8 a4a5 f6e5 f1e1 e5c5 d2c3 d4c3 d3e3 h7h6 e3c5 c8c5 a5a6 b7a6 a1a6 d8d2 a6a7 d2c2 e1e7 c2d2 e7f7 d2d1 g1f2 c3c2 f7g7 g8h8 g7h7 h8g8');
INSERT INTO evaluations VALUES(8373,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(8373,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(8373,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(8373,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(8373,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(8373,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(8373,36,'cp -75','e4d5 f6d5 a2a3 b4c3 b2c3 c8f5 c3c4 c6d4 b3d4 d5c3 d1d2 e5d4 f1d3 f5d3 d2d3 e8g8 e1g1 a8c8 c1d2 d8f6 d2c3 d4c3 a1d1 f6b6 d3d4 b6a6 d4c3 c8c4 c3d3 a6c6 f1f2 c4c3 d3d6 c6d6 d1d6 c3a3 f2d2 h7h6 h2h3 a3a1 g1h2 b7b5 d6d7 f8b8 d2e2 a7a5 e2e7 a5a4 e7f7');
INSERT INTO evaluations VALUES(9606,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(9606,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(9606,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(9606,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(9606,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(9606,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(9606,36,'cp 54','e2e4 e7e5 d4d5 a7a5 f3d2 d7c5 d2b3 b7b6 b3c5 b6c5 f1e2 h7h6 g5d2 a8b8 d1c1 e8g8 b2b3 g8h7 e1g1 c8d7 g1h1 c7c6 c1e1 c6d5 c4d5 h7h8 a1c1 f6h7 e2c4 f7f5 e4f5 g6f5 f2f3');
INSERT INTO evaluations VALUES(9607,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(9607,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(9607,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(9607,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(9607,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(9607,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(9607,36,'cp 54','e7e5 d4d5 a7a5 f3d2 h7h6 g5e3 h6h5 h2h3 b7b6 a2a3 e8f8 g2g3 g7h6 e3h6 h8h6 h3h4 f8g7 f1h3 h6h8 f2f3 f6g8 d1c2 d7c5 h3c8 d8c8 e1c1 g8h6 d2b3 c5b3 c2b3 f7f5 b3c2 h8f8 c2e2 f5e4 c3e4');
INSERT INTO evaluations VALUES(14195,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(14195,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14195,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(14195,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(14195,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(14195,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(14195,36,'cp -3','f6e4 g5e3 e7e6 g2g3 f8d6 f1g2 e8g8 c3c4 d6b4 b1d2 e4d6 e1g1 d6f5 e3f4 f5d4 f3d4 c6d4 d2b3 d4b3 d1b3 a7a5 c4d5 e6d5 g2d5 c7c6 d5e4 d8e7 b3c2 h7h6 f1c1 b4d6 e4h7 g8h8 f4d6 e7d6 h7d3 c8e6 c2c5 d6c5 c1c5 f8e8');
INSERT INTO evaluations VALUES(15851,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(15851,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(15851,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(15851,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(15851,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(15851,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(15851,36,'cp -119','d2d4 f8b4 c2c3 b4e7 g5f3 e5e4 f3e5 e8g8 e1g1 f8e8 b1a3 a7a5 f2f3 f7f6 f3e4 f6e5 e4d5 d8d5 d4e5 d5d1 f1d1 e7c5 g1h1 c8e6 a3c4 a5a4 b3b4 e6c4 b4c5 e8e5 c1f4 e5c5 d1d7 c5b5 d7c7 c4f7 c7e7 b5b2 f4e5');
INSERT INTO evaluations VALUES(18252,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(18252,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(18252,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(18252,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(18252,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(18252,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(18252,36,'cp 100','c8d7 b5e2 c6a5 c2c3 a7a6 b2b3 c7c5 e1g1 g8e7 b1a3 e7f5 e2d3 h7h6 a1b1 a5c6 a3c2 c5d4 c3d4 b6b5 a2a4 b5b4 d3f5 e6f5 c2e1 f8e7 e1d3 e8g8 b1a1 d7e6 c1e3 a8c8 d3f4 a6a5');
INSERT INTO evaluations VALUES(20025,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(20025,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(20025,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(20025,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(20025,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(20025,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(20025,36,'cp -23','g8f6 f1e2 g7g6 e1g1 f8g7 c2c3 e8g8 h2h3 e7e5 a2a3 b7b6 b2b4 h7h6 f1e1 c8e6 b4b5 c6e7 c3c4 a7a6 b1c3 a6b5 c3b5 e7c6 b5c3 e6d7 e3d2 a8a6 c3b5 f6e8 a1b1 g8h7 d1b3 d7c8 b5c3 e8c7 c3d5 c7d5');
INSERT INTO evaluations VALUES(13332,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(13332,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(13332,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(13332,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(13332,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(13332,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(13332,36,'cp -12','f1e2 d6d5 e3c5 d5e4 d3e4 d8d1 e2d1 f6e4 c5e3 f7f6 c2c3 e4d6 b1d2 d6f5 e3c5 c8d7 d1c2 e8c8 e1c1 b7b6 c5a3 e7e5 a3f8 h8f8 d2e4 d7e6 c2b3 e6b3 a2b3 d8d1 h1d1 f8d8 d1d8 c6d8 c1c2 c8c7 g2g4 f5e7 f3h4 d8e6');
INSERT INTO evaluations VALUES(20233,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(20233,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(20233,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(20233,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(20233,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(20233,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(20233,36,'cp -28','e6d5 f1b5 g8f6 b5c6 b7c6 e1g1 f8e7 d2d3 e8g8 f3e5 d8c7 d1f3 c8a6 b2b3 a8e8 f1e1 f6d7 e5d7 c7d7 c1b2 c5c4 d3c4 e7c5 g1h1 e8e1 a1e1 f8e8 e1e8 d7e8 f3e2 e8e2 c3e2 d5c4 e2d4 c4b3 a2b3 a6b7 h1g1 f7f6 g1f2 c5b6 c2c4 b7c8 f2e3 g8f7 b2c3 c8d7 e3d3 c6c5 d4e2');
INSERT INTO evaluations VALUES(22357,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(22357,31,'cp 364',NULL);
INSERT INTO evaluations VALUES(22357,32,'cp 355',NULL);
INSERT INTO evaluations VALUES(22357,33,'cp 375',NULL);
INSERT INTO evaluations VALUES(22357,34,'cp 361',NULL);
INSERT INTO evaluations VALUES(22357,35,'cp 374',NULL);
INSERT INTO evaluations VALUES(22357,36,'cp 382','c4f7 e8f7 f3e5 f7f6 b1c3 f8b4 d1d4 b4c3 b2c3 d8e8 e5d7 f6g6 d7e5 g6h7 e1g1 g8f6 f1e1 b7b6 c3c4 h8f8 f2f3 c8e6 c1b2 f6d7 e5d7 e8d7 d4d7 e6d7 c4c5 b6c5 g1f2 h7g8 e1e3 a8e8 b2c3 e8e6 a1d1 d7a4 d1c1 e6g6 g2g3 g6c6 h2h4 c6g6 h4h5');
INSERT INTO evaluations VALUES(24583,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(24583,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(24583,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(24583,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(24583,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(24583,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(24583,36,'cp -16','c1f4 e7e6 d1e2 f6h5 f4g5 d8c7 f3e5 h5f6 g5f6 d7f6 e1g1 f8d6 c2c3 e8g8 b1d2 b7b5 d2f3 c7b7 f1c1 a7a5 g2g3 f6e4 a2a3 f8c8 e5d3 b5b4 c3b4 a5b4 c1c8 a8c8 a3a4 b4b3 f3e1 b7a6 e2d1 a6b6 a4a5 b6d4 d1b3');
INSERT INTO evaluations VALUES(25497,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(25497,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25497,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(25497,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(25497,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(25497,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(25497,36,'cp 45','f6e4 f1e1 e4f6 b5c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 d5e6 b1d2 f8d6 d2c4 d6e5 c4e5 a7a5 e5d3 h7h6 g5f6 g7f6 e1e2 d8c8 d3c5 e6f5 a1e1 h8d8 a2a3 b7b6 c5e4 f5e4 e2e4 c8b7 g2g3 d8d2 e1e2 a8d8 g1g2 a5a4 e4e7 h6h5 e2d2 d8d2 e7f7 d2b2 f7f6');
INSERT INTO evaluations VALUES(25545,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(25545,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(25545,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(25545,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(25545,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(25545,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(25545,36,'cp 43','f1e1 e4f6 b5c6 d7c6 d4e5 d6e5 d1d8 e8d8 f3e5 c6d5 c1g5 c7c6 b1d2 d8c7 e5d3 a8e8 e1e8 f6e8 d3f4 b7b6 a1e1 f8d6 f4d5 c6d5 g5e7 c7d7 d2f3 f7f6 e7d6 d7d6 e1e3 d6d7 g1f1 g7g6 e3e1 e8c7 f3d4 a7a6 e1e3 c7b5 d4b5 a6b5 a2a3 h7h5');
INSERT INTO evaluations VALUES(27015,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(27015,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(27015,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(27015,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(27015,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(27015,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(27015,36,'cp 67','h2h3 g4f3 d1f3 h7h5 h3h4 c6d4 f3d1 g7g6 c2c3 d4e6 d2d4 f8h6 b1a3 h6c1 a1c1 g8f6 a3c4 e5d4 c3d4 e8g8 e4e5 d6e5 d4e5 f6g4 e1g1 d8e7 g2b7 a8d8 d1f3 e6d4 f3e4 f8e8 b7d5 e7d7 e4d4 d7d5 d4d5 d8d5 f1e1 g4e5 c4e5 d5e5 e1e5 e8e5 c1c7 e5e2 c7a7 e2b2');
INSERT INTO evaluations VALUES(28112,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(28112,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(28112,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(28112,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(28112,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(28112,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(28112,36,'cp 45','g4f3 e2f3 e7e5 d4e5 d6e5 d1d8 e8d8 b1c3 g8f6 e1g1 c7c6 c3a4 d8c7 c1d2 b8d7 f1d1 h8d8 a2a3 h7h5 c4c5 b7b6 b2b4 d7f8 a4b2 f8e6 b2c4 d8d4 a1c1 f6d7 g2g3 b6c5 g1g2 c5b4 a3b4 d4d3 f3e2 d3d4 d2e3 d4d1 e2d1');
INSERT INTO evaluations VALUES(28985,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(28985,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(28985,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28985,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(28985,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(28985,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(28985,36,'cp 13','e7e6 e1g1 f5d3 d1d3 f8d6 f4d6 d8d6 c2c4 e8g8 b1c3 b8d7 f1d1 f8d8 b2b3 d6e7 h2h3 h7h6 d3e2 a8c8 e3e4 d5e4 c3e4 f6e4 e2e4 a7a5 e4e2 e7b4 a1c1 d7f6 f3e5 b4b6 c4c5 b6c7 e5c4 f6d5 c4d6 d8d6 c5d6 c7d6');
INSERT INTO evaluations VALUES(31746,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(31746,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(31746,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(31746,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(31746,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(31746,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(31746,36,'cp 113','b1d2 g8f6 f1d3 b8c6 h2h3 g4c8 c2c3 e7e5 e1g1 f8g7 f1e1 e8g8 d2f1 f6h5 c1e3 g8h8 a2a4 b7b6 d1d2 a7a5 f1g3 h5g3 f2g3 c8b7 g3g4 e5d4 c3d4 c6b4 d3b1 c7c5 d4d5 c5c4 e3g5');
INSERT INTO evaluations VALUES(31904,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(31904,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(31904,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(31904,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(31904,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(31904,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(31904,36,'cp -12','e7e6 b1d2 f8d6 f4e3 b8d7 f3f4 g8f6 d2b3 e8g8 f1e2 d7b6 e1g1 b6c8 e2d3 c8e7 c2c3 b7b6 d1e2 g7g6 a1c1 a7a5 g2g3 e7f5 b3d2 f5e3 f2e3 c7c5 a2a4 d8d7 d3b5 d7e7 g1g2 a8d8 e3e4 d5e4 d2e4');
INSERT INTO evaluations VALUES(32340,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(32340,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(32340,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(32340,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(32340,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(32340,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32340,36,'cp 0','b1d2 f8d6 f4e3 b8d7 f3f4 a8c8 g2g3 g8e7 d2f3 e8g8 f1d3 c7c5 d4c5 d6c5 e3c1 e7f5 e1g1 f5d6 g1g2 d7f6 d1e2 f6e4 c2c3 f8e8 c1e3 c5e3 e2e3 e4c5 a1d1 c5d3 e3d3 h7h6 f1e1 b7b5 f3d4 d8b6 e1e2 a7a5');
INSERT INTO evaluations VALUES(35054,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35054,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35054,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35054,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35054,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35054,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35054,36,'cp 13','b7b6 f1d3 e4b7 c2c4 f8b4 b1c3 b4c3 b2c3 d7d6 e1g1 g8f6 a2a4 b8d7 a4a5 a7a6 d1d2 b7b8 f3g5 c8b7 d4d5 e6d5 c4d5 b6b5 e3d4 e8g8 d2c2 h7h6 g5h7 f6h7 d3h7 g8h8 h7e4 d7c5 f1e1 b8d8 c3c4 b5c4');
INSERT INTO evaluations VALUES(37658,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(37658,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(37658,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(37658,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(37658,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(37658,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(37658,36,'cp 72','g1f3 d7d6 d2d4 e5d4 f3d4 b8d7 b2b3 g8f6 f1g2 e8g8 e1g1 f8e8 c1b2 a7a6 c3a4 d7f8 d1c2 c5a7 e2e3 d8a5 d4f3 c8f5 b2c3 f5c2 c3a5 f8d7 a1c1 c2g6 f3h4 g6e4 f1d1 b7b5 a4c3 e4g2 g1g2 b5c4 d1d6 c4b3 a2b3');
INSERT INTO evaluations VALUES(38144,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(38144,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38144,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38144,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38144,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(38144,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(38144,36,'cp 17','e4d5 d8d5 c1d2 b4c3 d2c3 c5d4 c3d4 e6e5 d4e3 g8e7 g1e2 c8f5 d3f5 d5d1 a1d1 e7f5 e3c1 b8c6 e1g1 a8d8 f2f4 d8d1 f1d1 f7f6 c2c4 e8e7 b2b3 e7e6 f4e5 f6e5 d1d5 h8d8 e2c3 f5d4 h2h3 b7b6 d5d8 c6d8 g1f2 d4c2 c3e4');
INSERT INTO evaluations VALUES(38417,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(38417,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38417,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(38417,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(38417,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(38417,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(38417,36,'cp 43','c7c5 c2c3 d6d5 e4e5 g8h6 f2f4 c5d4 c3d4 c8f5 b1c3 f5d3 d1d3 e7e6 e3f2 h6g4 f2g3 g4h6 g1f3 h6f5 g3f2 h7h5 g2g3 g7f8 e1g1 b8c6 h2h3 f8e7 g1g2 f5h6 f1c1 d8d7 c3d1 e8f8 d1e3 h6f5 f2e1 a8c8 e3f5 e6f5 e1d2 f8g7 d3b3 h8d8');
INSERT INTO evaluations VALUES(41484,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(41484,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(41484,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(41484,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41484,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(41484,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(41484,36,'cp 7','b7b6 c2c4 c8b7 f1e2 e8g8 b1c3 c7c5 d4c5 d5c4 c5b6 d8d1 c3d1 c4b3 b6a7 a8a7 a2b3 a7a1 b2a1 b8d7 e1g1 f8b8 d1c3 b7f3 e2f3 b8b3 f1d1 b3a3 c3b5 a3a5 b5d4 d7e5 d4b3 e5f3 g2f3 a5d5 d1c1 h7h5 c1c8 g8h7 a1d4 d5d7 c8c6 f6d5 d4g7');
INSERT INTO evaluations VALUES(41886,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41886,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41886,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41886,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(41886,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(41886,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(41886,36,'cp 29','b8c6 e1g1 e8g8 e2e3 f8e8 d2d3 c8f5 h2h3 a7a5 b2b3 h7h6 a2a3 c6e7 f3h4 f5d7 c1b2 c7c6 d3d4 e5d4 e3d4 c5b6 c3a4 d6d5 f1e1 b6c7 a4c5 d8c8 c5d7 c8d7 d1d3 a8d8 a1d1 e7c8 e1e8 d8e8 h4f3 c8d6 f3e5 d7f5 d3f5 d6f5 c4d5 f6d5 e5c4');
INSERT INTO evaluations VALUES(42066,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(42066,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(42066,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42066,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(42066,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(42066,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(42066,36,'cp 5','b4c3 c2c3 e8g8 g1f3 b8d7 e2e3 c7c5 b2b4 c5d4 c3d4 d5c4 f1c4 d7b6 c4e2 d8d4 f3d4 c8d7 d4b3 b6a4 c1d2 f8c8 e1g1 f6e4 f1c1 e4d2 b3d2 g8f8 e2f3 c8c1 a1c1 a8c8 c1c8 d7c8 g1f1 a4c3 f3e2 c3e2 f1e2 b7b6 e3e4 e6e5 g2g3 f8e7 e2d3 c8b7 d2c4 b7a6 f2f3 a6b5');
INSERT INTO evaluations VALUES(42095,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(42095,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(42095,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(42095,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42095,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(42095,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(42095,36,'cp 14','c2c3 e8g8 c1g5 d5c4 c3c4 b7b6 a1d1 c8a6 c4a4 h7h6 g5h4 d8e7 g1f3 f8d8 e2e3 a6f1 e1f1 c7c5 f1e2 b8d7 f3e5 b6b5 a4b5 d8b8 b5d7 e7d7 e5d7 b8b2 d1d2 b2d2 e2d2 f6d7 h1c1 c5d4 e3d4 d7b6 d2d3 b6d5 h4g3 h6h5 f2f3');
INSERT INTO evaluations VALUES(42091,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(42091,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(42091,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(42091,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(42091,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(42091,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42091,36,'cp 18','e8g8 c1g5 d5c4 c3c4 h7h6 g5f6 d8f6 g1f3 b8c6 c4c3 f8e8 a1d1 e6e5 d4d5 c6b8 e2e3 c8f5 c3c7 b8d7 c7b7 a8b8 b7c6 b8b2 f1c4 f5c2 d1c1 e8d8 e1g1 d7b6 c4a6 c2e4 f1d1 f6c6 d5c6 d8d1 c1d1 e4c6 f3e5 c6d5 a6f1 b2a2 f1b5 a2a3 e5f3 a3c3 f3d4 d5c4');
INSERT INTO evaluations VALUES(37069,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(37069,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(37069,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(37069,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(37069,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(37069,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(37069,36,'cp 15','c1g5 d5c4 c3c4 b7b6 a1d1 c8a6 c4a4 h7h6 g5h4 d8e7 g1f3 f8d8 e2e3 a6f1 h1f1 b8d7 a4c6 a8c8 f3e5 d7e5 d4e5 d8d1 e1d1 e7d7 c6d7 f6d7 f2f4 b6b5 d1e2 c7c5 f1d1 d7b6 d1d8 c8d8 h4d8 g8f8 d8b6 a7b6 h2h4 f8e7 g2g4 g7g6 e2d2 e7d7 e3e4 h6h5 g4h5 g6h5');
INSERT INTO evaluations VALUES(37031,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(37031,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37031,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(37031,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(37031,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(37031,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(37031,36,'cp 15','d5c4 c3c4 b7b6 a1d1 c8a6 c4a4 h7h6 g5h4 d8e7 g1f3 f8d8 e2e3 a6f1 h1f1 b8d7 a4c6 a8c8 f3e5 d7e5 d4e5 d8d1 e1d1 e7d7 c6d7 f6d7 f2f4 b6b5 d1e2 c7c5 f1d1 d7b6 d1d8 c8d8 h4d8 g8f8 d8b6 a7b6 h2h4 f8e7 g2g4 g7g6 e2d2 e7d7 e3e4 h6h5 g4h5 g6h5 d2e3 d7e7');
INSERT INTO evaluations VALUES(42658,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(42658,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(42658,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(42658,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(42658,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(42658,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(42658,36,'cp -66','d7d5 e4d5 f6d5 c3e4 c5e7 d2d4 e5d4 d1d4 e8g8 f1c4 d5b6 d4d8 e7d8 c4d3 f8e8 g1e2 b8c6 c1f4 c8f5 f2f3 b6d5 e1c1 d5f4 e2f4 c6e5 e4c3 e5d3 f4d3 d8h4 d3c5 b7b6 c5e4 a8d8 c3b5 f5e4 f3e4 e8e4 b5a7 f7f5 a7b5 d8d1 h1d1 e4e2 b5c7 e2g2 c7d5');
INSERT INTO evaluations VALUES(43608,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(43608,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(43608,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43608,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(43608,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(43608,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(43608,36,'cp 63','d2d4 e5d4 f3d4 f8e7 e1g1 e8g8 f1e1 b8c6 a2a4 c6d4 d1d4 c8d7 a4a5 d7c6 c1e3 f6d7 d4d2 f8e8 b2b4 e7h4 e3d4 d8g5 d2g5 h4g5 f2f3 d7f8 a1d1 f8e6 d4e3 e6f4 c3d5 c6d5 c4d5 f4d5 d1d5 g5f6 e1d1');
INSERT INTO evaluations VALUES(44098,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(44098,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(44098,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(44098,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44098,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(44098,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(44098,36,'cp 29','g1f3 d7d6 g2g3 g7g6 f1g2 f8g7 e1g1 e8g8 a1b1 a7a5 e2e4 c6c5 b2b3 b8c6 a2a3 c8e6 f3e1 e6g4 f2f3 g4d7 e1c2 f6e8 b3b4 a5b4 a3b4 c5b4 c3d5 e8c7 d5b4 c6b4 c2b4 d7a4 d1e1 c7e6 c1e3 b7b5 e1d2 b5c4 d3c4 e6d4');
INSERT INTO evaluations VALUES(47241,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(47241,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(47241,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(47241,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(47241,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(47241,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(47241,36,'cp -38','c7c5 g2g3 g7g6 f1g2 b8c6 e1g1 f8g7 b1c3 d5d4 c3a4 b7b6 f3e5 c6e5 g2a8 e5g4 a8c6 c8d7 c6d7 d8d7 e2e4 h7h5 d1f3 d7a4 e4e5 e8f8 e5f6 g7f6 b2b3 a4d7 a2a4 f8g7 a4a5 g4e3 a5b6 a7b6 c1e3 d4e3 a1a8 h8a8 f3a8 d7g4 a8g2 f6d4 f1e1 h5h4 h2h3 g4g3 g2g3 h4g3');
INSERT INTO evaluations VALUES(49685,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(49685,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(49685,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(49685,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(49685,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(49685,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(49685,36,'cp 125','b2e5 d7d5 e2e3 f8e7 g1f3 e8g8 f1e2 c7c5 e1g1 b8c6 e5b2 c8f5 d2d3 d5d4 b1d2 f5e6 d2c4 d4e3 f2e3 b7b5 c4e5 c6e5 b2e5 d8b6 d1d2 h7h6 a2a4 a7a6 g1h1 a8d8 a4b5 a6b5 e5g3 f8e8 d2a5 b6b7 h2h3');
INSERT INTO evaluations VALUES(49487,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(49487,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(49487,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(49487,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(49487,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(49487,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(49487,36,'cp 132','f8e7 g1f3 d7d5 e2e3 e8g8 f1e2 c7c5 e5b2 c8f5 e1g1 b8c6 d2d3 d5d4 b1d2 f5e6 d2c4 b7b5 c4e5 c6e5 f3e5 a8c8 e3e4 f6d7 e5f3 f7f5 c2c3 d4c3 b2c3 d7b6 c3e5 f5e4 d3e4 b5b4 a1c1 d8d1 f1d1 c5c4 f3d4');
INSERT INTO evaluations VALUES(51528,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51528,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(51528,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(51528,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(51528,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(51528,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(51528,36,'cp 22','a2a4 e7e6 c2c3 b5a4 e2e3 g8f6 g1f3 c8b7 a1a4 c7c5 b4c5 f8c5 f1e2 e8g8 a4a2 a7a5 e1g1 d7d5 d2d4 c5e7 b1d2 b7a6 e2a6 a8a6 c1a3 e7a3 a2a3 b8d7 d1e2 d8a8 c3c4 a5a4 c4c5 e6e5 d2b1 f8e8 f3e5 d7e5 d4e5');
INSERT INTO evaluations VALUES(55474,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(55474,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(55474,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(55474,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(55474,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(55474,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(55474,36,'cp 75','c2c3 c7c5 f1d3 b8c6 g1f3 d8c7 e1g1 c5d4 c3d4 c6b4 d3b5 c8d7 b5d7 c7d7 b1c3 g8e7 c1d2 b4c6 c3e2 d7d8 b2b4 a7a6 a2a4 e7c8 b4b5 c6a5 e2f4 a5c4 f4h5 a6b5 a4b5 a8a1 d1a1 c8b6 f1c1 d8a8 d2c3 a8a1 c1a1 g7g6 h5f6 e8d8');
INSERT INTO evaluations VALUES(5154,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(5154,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(5154,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(5154,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(5154,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(5154,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(5154,36,'cp 115','g8f6 e1c1 e8g8 e3h6 g7h6 d2h6 d7d5 h2h4 f6g4 d4c6 b7c6 h6d2 a7a5 f2f3 g4e5 d2e3 d8f6 f1e2 c8a6 f3f4 e5d7 e2a6 a8a6 h4h5 d7b6 e3g3 b6c4 b2b3 c4b6 a2a4 d5e4 c3e4 f6a1 c1d2 a1d4 g3d3 d4d3 d2d3 b6d5 g2g3 f8d8 h5g6 h7g6 e4c5 a6a7 d3e2 a7e7');
INSERT INTO evaluations VALUES(6054,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(6054,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(6054,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(6054,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(6054,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(6054,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(6054,36,'cp 124','c2c3 b4a5 b2b4 a5b6 d2d4 e5d4 a2a4 a7a6 c3d4 b6a7 f1c4 c6e7 d5c3 g8f6 e1g1 e8g8 c1g5 e7g6 c3d5 c8e6 d1d3 c7c6 d5f6 g7f6 g5h6 d6d5 e4d5 c6d5 c4b3 f8e8 b4b5 a7b8 b5a6 a8a6 b3c2 e6d7 g2g3 d7g4 f1e1 b8d6 e1e8 d8e8');
INSERT INTO evaluations VALUES(9293,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(9293,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(9293,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9293,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(9293,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9293,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(9293,36,'cp 12','a7a6 b5c6 b7c6 c2c4 c8e6 b1c3 e8g8 c1e3 c5e3 f2e3 f6d7 f1f2 f7f5 d1c2 d7c5 d3d4 f5e4 c3e4 e5d4 e3d4 c5e4 c2e4 d8d7 d4d5 c6d5 c4d5 e6f5 e4c4 f5g6 a1f1 d7b5 f1c1 b5b6 c4d4 a8e8 d4b6 c7b6 c1c7 g6e4');
INSERT INTO evaluations VALUES(11200,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(11200,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(11200,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(11200,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(11200,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11200,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(11200,36,'cp 42','f1d3 d8b6 a1b1 f6h5 f4e5 c6e5 f3e5 h5f6 f2f4 f6d7 e1g1 g7g6 d1e1 f8e7 d2f3 f7f6 e5g4 b6c7 e1g3 f6f5 g4e5 d7e5 f3e5 c8d7 c3c4 d5c4 d3c4 b7b5 c4e2 c5d4 e2f3 a8c8 e3d4 e8g8 h2h4 e7f6 h4h5 c7d6 f1d1');
INSERT INTO evaluations VALUES(11731,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(11731,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(11731,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(11731,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(11731,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(11731,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(11731,36,'cp 54','d7d6 d2d4 f6d7 c1e3 a6a5 c3a4 f8e7 d4e5 d6e5 d1d3 e7d6 f1d1 e8g8 d3c3 d8e7 f3d2 c8a6 c3c6 f8b8 a4c3 a6b7 c6c4 d7f6 e3g5 b7a6 c4c6 a6b7 g5f6 e7f6 c6d7 f6g6 h2h3 a5a4 d2c4 b8d8 d7f5 g6f5 e4f5 d6b4 a2a3 b4c3 b2c3 f7f6 c4e3');
INSERT INTO evaluations VALUES(16568,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(16568,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(16568,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(16568,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(16568,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(16568,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(16568,36,'cp -60','f6e4 e2b5 e4d6 b5c6 d7c6 f3e5 c8e6 f1e1 d6f5 d2d3 d8e7 e5f3 h7h6 c1d2 e8c8 d2c3 c6c5 f3e5 f5d4 b1d2 e7f6 c3d4 c5d4 c2c3 f8d6 c3d4 d6e5 d4e5 f6f5 d2e4 f5e5 d1c2 h8e8 a1c1 e6d5 c2c3 d5e4 e1e4 e5c3 c1c3 e8e4');
INSERT INTO evaluations VALUES(16853,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(16853,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(16853,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(16853,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(16853,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(16853,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(16853,36,'cp -59','e2b5 e4d6 b5c6 d7c6 f3e5 f8e7 f1e1 e8g8 h2h3 f8e8 d2d3 f7f6 e5g4 d6f5 b1c3 e7f8 e1e8 d8e8 c1f4 e8f7 g4e3 f5d4 c3e2 c6c5 f4g3 c8e6 e2f4 e6d7 c2c3 d4e6 d3d4 c7c6 d4c5 e6c5 b2b4 c5e6 d1c2 e6f4 g3f4');
INSERT INTO evaluations VALUES(22634,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(22634,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(22634,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22634,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(22634,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(22634,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(22634,36,'cp 24','d2d4 d7d5 e2e3 g8f6 c2c4 f8e7 a2a3 e8g8 b2b4 d5c4 f1c4 a7a5 b4b5 c6a7 b1c3 c8d7 e1g1 f6d5 c4d5 e6d5 c3d5 a7b5 e3e4 c7c5 a3a4 b5d4 f3d4 c5d4 d1d4 d7c6 c1b2 f7f6 f1d1 c6d5 e4d5 e7d6 h2h3 a8c8 b2a3 f8e8 a3d6 d8d6 a1c1 c8c1 d1c1 e8d8');
INSERT INTO evaluations VALUES(23274,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(23274,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(23274,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(23274,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23274,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(23274,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23274,36,'cp 0','f8e7 g5e3 g8f6 f1c4 c6a5 c4d3 d7d5 b1d2 d5d4 e3g5 a5c6 e1g1 e8g8 d3b5 e7d6 f1e1 d8e7 d2c4 h7h6 g5h4 c8g4 c2c3 d4c3 b2c3 a7a6 b5c6 b7c6 c4e3 g4h5 d1b3 g8h7 e3f5 e7f7 h4f6 f7f6 b3c2 h5g4 f3d2 g4f5 e4f5 f6f5');
INSERT INTO evaluations VALUES(6956,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(6956,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(6956,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6956,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6956,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6956,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6956,36,'cp 12','g5e7 g8e7 f1c4 e7g6 b1c3 d7d6 c3d5 c8g4 d5e3 d8d7 g2g3 c6d4 e3g4 d7g4 f3d2 g4d1 e1d1 h8f8 c2c3 d4f3 d2f3 f8f3 d1e2 f3f6 e2e3 e8d7 a1f1 a8f8 f2f3 c7c6 h2h4 h7h5 b2b4 a7a5 b4a5 f8a8 f3f4 a8a5 c4b3 e5f4 g3f4 a5a3');
INSERT INTO evaluations VALUES(6955,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(6955,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(6955,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(6955,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(6955,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(6955,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(6955,36,'cp 11','g8e7 f1c4 e7g6 b1c3 d7d6 c3d5 c8g4 d5e3 d8d7 g2g3 c6d4 e3g4 d7g4 f3d2 g4d1 e1d1 h8f8 c2c3 d4f3 d2f3 f8f3 d1e2 f3f6 h2h4 e8e7 a1f1 c7c6 e2e3 a7a5 c4e2 a5a4 f2f4 e5f4 g3f4 a8f8 h1g1 g6f4 g1g7 e7e8 e2f3 f8f7 g7f7 e8f7 h4h5 f7e7 f1g1 f4e6 f3g4 e6g7 b2b3 a4b3 a2b3 g7e6 g4e6');
INSERT INTO evaluations VALUES(23750,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23750,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(23750,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23750,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23750,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23750,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23750,36,'cp 0','c6d4 f3d4 e5d4 b5d7 d8d7 c3d4 d6d5 e4e5 f6e8 c1e3 c7c6 b1d2 e8c7 f2f4 f7f5 d2f3 c7e6 d1d3 b7b6 b2b3 a7a5 a2a4 g7g6 g1h1 a8c8 e1c1 c6c5 c1c2 c8c7 a1d1 f8d8 c2c1 d7c6 h3h4 h7h5 c1c2 d8c8 d1c1 c6a8');
INSERT INTO evaluations VALUES(23871,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(23871,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(23871,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(23871,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(23871,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(23871,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23871,36,'cp 44','g4h5 c2c3 a7a6 b1d2 g8e7 a2a4 e8g8 g2g4 h5g6 f3h4 c5a7 d2f3 d8d7 b2b4 g8h8 c1e3 a7e3 f2e3 f7f6 a4a5 c6d8 h4g6 e7g6 d1b3 d8f7 f1f2 a8e8 a1f1 c7c6 b3a2 d7e7 c4b3 e7d7 g1h2 f7h6 c3c4 h6f7 c4c5 d7e7 c5d6 f7d6 h3h4');
INSERT INTO evaluations VALUES(23861,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(23861,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23861,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(23861,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(23861,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(23861,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(23861,36,'cp 30','c2c3 a7a6 b2b4 c5a7 c4b3 g8e7 c1e3 e8g8 b1d2 g8h8 g2g4 h5g6 a2a4 d6d5 e3a7 a8a7 b4b5 d5e4 d3e4 c6a5 f3e5 a6b5 a4b5 a5b3 d2b3 a7a1 b3a1 f7f6 e5g6 e7g6 a1c2 d8e7 d1f3 g6h4 f3e3 f8e8 c2d4 e7e4 e3e4 e8e4');
INSERT INTO evaluations VALUES(23860,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(23860,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(23860,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(23860,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(23860,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(23860,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23860,36,'cp 43','a7a6 b2b4 c5a7 a2a4 g8e7 c4a2 e8g8 c1e3 g8h8 b1d2 f7f6 d1c2 d8d7 b4b5 a6b5 a4b5 c6a5 c2b2 a7e3 f2e3 b7b6 d3d4 a5b7 a2c4 a8a1 b2a1 d6d5 e4d5 b7d6 e3e4 h5f3 f1f3 e7g6 c4f1 f8e8 c3c4 d7e7 f3b3 e5d4 a1d4 d6e4 d2e4');
INSERT INTO evaluations VALUES(24964,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(24964,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(24964,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(24964,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(24964,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(24964,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(24964,36,'cp 45','c2d3 e7e6 e1g1 f8e7 e3e4 e8g8 a2a3 a7a5 c1f4 c6a7 h2h4 b7b6 a1c1 c7c6 f1e1 a8c8 f3e5 f6d7 h4h5 h7h6 d1g4 g8h8 e4d5 d7f6 g4f3 c6d5 c3e2 e7d6 g2g4 c8c1 f4c1 d6e5 d4e5 f6h7 c1e3 a7c6 d3d4');
INSERT INTO evaluations VALUES(27061,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(27061,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(27061,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(27061,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27061,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(27061,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(27061,36,'cp 31','f8e7 e1g1 g8f6 c3d5 e8g8 c1d2 e6g4 c2c3 c6a5 d5e7 d8e7 c4b5 a7a6 b5a4 b7b5 a4c2 c7c5 d1e2 a5c6 h2h3 g4e6 f3g5 f6d7 g5e6 f7e6 c2b3 g8h8 f2f4 e5f4 d2f4 d7e5 f4g3 h7h6 f1f8 a8f8 a1e1');
INSERT INTO evaluations VALUES(29446,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(29446,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(29446,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(29446,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(29446,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(29446,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(29446,36,'cp 47','e2e3 c7c6 b1d2 f6h5 f4g5 h7h6 g5h4 g7g5 h2h3 g4e6 f3g1 h5f6 h4g3 e6f5 h3h4 g5g4 f1d3 f5d3 c2d3 b8a6 e3e4 a6c7 e4e5 f6h5 d1g4 h5g3 g4g3 d8d7 g1e2 e8c8 d2c4 h6h5 g3h3 c7e6 a1d1 f8h6 e1g1 c8b8 e5d6 e7d6 c4e3');
INSERT INTO evaluations VALUES(29445,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(29445,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(29445,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(29445,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(29445,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(29445,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(29445,36,'cp 60','e7e6 h2h3 g4f3 d1f3 d6d5 g2g4 c7c5 b1c3 c5d4 e3d4 f8d6 e1c1 d6f4 f3f4 b8c6 h3h4 d8b8 f4f3 a7a6 g4g5 f6d7 h4h5 b8d6 c3e2 h7h6 g5h6 h8h6 h1g1 d6f8 e2f4 e8c8 c2c3 c8b8 f1d3 b8a8 d3c2 d7f6 c1b1 a8b8');
INSERT INTO evaluations VALUES(29666,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29666,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29666,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29666,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29666,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29666,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(29666,36,'cp -27','g4d7 e3e4 d5e4 f3e4 e7e5 d4e5 f6g4 g1f3 b8c6 c1g5 f8e7 g5f4 e7b4 b1c3 d8e7 d1d2 e7c5 h1f1 g4e5 a2a3 b4c3 d2c3 c5c3 b2c3 f7f6 f4e5 c6e5 f3e5 f6e5 d3c4 h8f8 f1f8 e8f8 e1c1 f8e7 d1d5 a8e8 d5e5 e7f6 e5d5 e8e4 d5d7 e4c4 c1b2 c4c6 b2b3 h7h5 h2h4 g7g6 g2g3 f6e5');
INSERT INTO evaluations VALUES(34197,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(34197,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(34197,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(34197,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(34197,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(34197,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(34197,36,'cp 178','c2c3 c6c5 d4c5 a5c5 b1a3 c5c6 f1d3 f8a3 f3d4 c6c5 d4b5 g8f6 c1e3 c5c6 b5a3 e8g8 e3d4 e6e5 d4e5 d7d5 d3b5 c6e6 e5f6 e6e4 d1e2 g7f6 e1c1 c8g4 f2f3 e4e2 b5e2 g4e6 a3c2 f8d8 f3f4 b8c6 e2f3 a8c8 h1e1 d5d4 f3c6 b7c6 c2d4 e6g4 d4f3 d8d1 e1d1 g4f3 g2f3');
INSERT INTO evaluations VALUES(34430,30,'cp -190',NULL);
INSERT INTO evaluations VALUES(34430,31,'cp -216',NULL);
INSERT INTO evaluations VALUES(34430,32,'cp -208',NULL);
INSERT INTO evaluations VALUES(34430,33,'cp -214',NULL);
INSERT INTO evaluations VALUES(34430,34,'cp -191',NULL);
INSERT INTO evaluations VALUES(34430,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(34430,36,'cp -192','b6b2 b1d2 b2c3 a1c1 c3a5 d4c5 b8d7 g1f3 f8c5 f3d4 c5d4 e3d4 g8e7 d4c3 a5d8 d2f3 a7a5 f1b5 e8g8 c3d4 b7b6 d1e2 h7h6 e1g1 e7f5 h2h3 c8b7 d4b2 d7c5 f3d4 f5d4 b2d4 b7a6 b5a6 c5a6 c1c6 a6b4 c6c3 a8c8 c3c8 d8c8 d4b6 c8c4 e2c4 d5c4 b6a5 b4a2');
INSERT INTO evaluations VALUES(34471,30,'cp -207',NULL);
INSERT INTO evaluations VALUES(34471,31,'cp -216',NULL);
INSERT INTO evaluations VALUES(34471,32,'cp -189',NULL);
INSERT INTO evaluations VALUES(34471,33,'cp -194',NULL);
INSERT INTO evaluations VALUES(34471,34,'cp -214',NULL);
INSERT INTO evaluations VALUES(34471,35,'cp -205',NULL);
INSERT INTO evaluations VALUES(34471,36,'cp -212','b1d2 b2c3 a1c1 c3a5 d4c5 b8d7 g1f3 f8c5 f3d4 c5d4 e3d4 g8e7 d4c3 a5d8 d2f3 a7a5 f1b5 e8g8 c3d4 e7f5 d4b2 b7b6 b2a3 d7c5 e1g1 c8d7 b5d3 c5d3 d1d3 f8e8 f3d4 h7h6 d4f5 e6f5 a3d6 a8c8 d3d5 d7e6 d5b5 d8g5 c1c8 e8c8 b5b6 e6d5 f2f3 c8c2 f1f2');
INSERT INTO evaluations VALUES(34470,30,'cp -204',NULL);
INSERT INTO evaluations VALUES(34470,31,'cp -209',NULL);
INSERT INTO evaluations VALUES(34470,32,'cp -205',NULL);
INSERT INTO evaluations VALUES(34470,33,'cp -198',NULL);
INSERT INTO evaluations VALUES(34470,34,'cp -210',NULL);
INSERT INTO evaluations VALUES(34470,35,'cp -203',NULL);
INSERT INTO evaluations VALUES(34470,36,'cp -203','b2c3 a1c1 c3a5 d4c5 b8d7 g1f3 f8c5 f3d4 c5d4 e3d4 g8e7 d4c3 a5d8 d2f3 a7a5 c3b2 b7b6 f1e2 e8g8 b2a3 d7c5 e1g1 c8a6 a3c5 b6c5 e2a6 a8a6 c1c5 a6c6 c5c6 e7c6 d1d3 d8c7 d3a3 c7b8 a3d6 b8b5 h2h3 h7h6 f1c1 c6b4 c1c7 b5d3');
INSERT INTO evaluations VALUES(34472,30,'cp -212',NULL);
INSERT INTO evaluations VALUES(34472,31,'cp -202',NULL);
INSERT INTO evaluations VALUES(34472,32,'cp -192',NULL);
INSERT INTO evaluations VALUES(34472,33,'cp -212',NULL);
INSERT INTO evaluations VALUES(34472,34,'cp -212',NULL);
INSERT INTO evaluations VALUES(34472,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(34472,36,'cp -186','a1c1 c3a5 d4c5 b8d7 g1f3 f8c5 f3d4 c5d4 e3d4 g8e7 d4c3 a5d8 d2f3 b7b6 c3b4 e8g8 f1e2 d7c5 e1g1 c8a6 b4c5 a6e2 d1e2 b6c5 c1c5 a8c8 f1c1 c8c5 c1c5 d8b6 e2c2 h7h6 h2h3 f8d8 c5c7 e7g6 c7c8 d8c8 c2c8 g6f8 a2a4 b6a5 c8e8 a5c7 f3d4 c7e5');
INSERT INTO evaluations VALUES(36935,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(36935,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(36935,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(36935,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(36935,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(36935,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(36935,36,'cp -30','c7c6 c1g5 f8e8 f1e1 c8e6 e2d3 b8d7 h2h3 h7h6 g5e3 d8c7 c3e2 d7f8 d1c1 e6d7 b2b3 f8e6 c2c4 c6c5 c1c2 d7c6 d4c5 e6c5 e2d4 c5d3 c2d3 d5c4 d3c4 c6d7 d4b5 c7c4 b3c4 d7b5 c4b5 f6e4 e3d4 d6b4 e1d1 a7a6 b5a6 a8a6');
INSERT INTO evaluations VALUES(37074,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(37074,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(37074,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(37074,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(37074,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(37074,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(37074,36,'cp 57','b1c3 c7c5 d4c5 h7h6 d1d4 b8c6 d4h4 d8a5 e1c1 e6e5 f3d2 a5c5 f1e2 c8e6 d2b3 c5b4 g3g4 d5d4 e3d4 e6b3 g4g5 e5d4 g5f6 d4c3 h4b4 c6b4 a2b3 a8c8 f6g7 g8g7 c1b1 c3b2 d1d2 c8d8 h1d1 b4c6 b1b2 d8d2 d1d2 f8d8 d2d8 c6d8 b3b4 d8e6 b2c3 e6c7 c3c4');
INSERT INTO evaluations VALUES(38519,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(38519,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38519,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(38519,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(38519,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(38519,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(38519,36,'cp 62','d2d4 e5d4 f3d4 g8f6 c1e3 e8g8 d1d2 b8c6 e1c1 c6d4 e3d4 c7c6 a2a4 a7a6 f2f3 b7b5 c4b3 c8e6 b3e6 f7e6 c1b1 a8b8 e4e5 d6e5 d4e5 d8d2 d1d2 b8d8 h1d1 d8d2 d1d2 f6d5 e5g7 g8g7 c3e4 f8f5 a4a5 g7f8 e4c5 d5c3 b2c3 f5c5 d2d7 c5c3 b1b2 c3c4 d7a7 c4b4 b2c3 b4c4 c3b3 c4a4 a7a6 c6c5 a6a7 f8e8');
INSERT INTO evaluations VALUES(38610,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(38610,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(38610,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(38610,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(38610,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(38610,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(38610,36,'cp 149','d2d4 e5d4 d1d4 g8f6 c1g5 e8g8 f1e2 f8e8 g5f6 d8f6 d4f6 g7f6 b1c3 a7a6 e1c1 b8d7 f3d4 d7c5 h1e1 c8d7 e2f3 e8e5 g2g3 a8e8 e1e2 f6f5 d1e1 f5f4 c1d2 f4g3 h2g3 e5e2 c3e2 h7h5 c2c3 d7g4 b2b4 c5d7 f3g4 h5g4 d4f5');
INSERT INTO evaluations VALUES(39588,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(39588,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(39588,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(39588,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(39588,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(39588,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(39588,36,'cp -16','d7d5 g1f3 g8f6 a2a4 e8g8 f1e2 c7c5 e1g1 b7b6 a4a5 c8a6 c1d2 e7e6 f3e5 c5d4 e3d4 a6e2 d1e2 b6a5 c3b5 a7a6 d2a5 d8e7 b5c3 e7a7 f1d1 f6d7 e5d7 b8d7 e2d2 f8e8 b2b3 d7f6 h2h3 a7b7');
INSERT INTO evaluations VALUES(41514,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(41514,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(41514,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(41514,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41514,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(41514,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41514,36,'cp 0','e8g8 b1d2 f8e8 e2e4 e7e5 f1e1 a7a5 a2a4 b8c6 c2c3 b7b6 e4d5 f6d5 d2c4 h7h6 c4e3 c8e6 e3c4');
INSERT INTO evaluations VALUES(43155,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43155,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(43155,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(43155,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43155,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43155,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43155,36,'cp 0','d7d6 g2g3 e8g8 f1g2 e7e5 e1g1 a7a5 d2d3 b8a6 d1c2 c7c6 b1d2 c8d7 a2a3 f8e8 b2c3 c6c5 f1e1 d7c6 e2e3 b7b6 a1b1 a8c8 g2h3 c6d7 h3f1 d7c6');
INSERT INTO evaluations VALUES(43886,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43886,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43886,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43886,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(43886,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(43886,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(43886,36,'cp 34','c8b7 c2c4 c7c5 e1g1 e7e6 b1c3 f8e7 d2d4 c5d4 d1d4 d7d6 f1d1 e8g8 c1g5 b8d7 c3b5 d6d5 c4d5 b7d5 a1c1 a8c8 b5c3 d5c6 b2b4 h7h6 g5f4 a7a6 e2e4 b6b5 a2a3 d8b6 d4b6 d7b6 f3e5 f8d8 e5c6 c8c6 d1d8 e7d8 e4e5 f6d5 c3d5 c6c1 f4c1 b6d5 g1f1 d8b6 c1b2');
INSERT INTO evaluations VALUES(45679,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(45679,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(45679,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(45679,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(45679,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(45679,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(45679,36,'cp 26','f8e7 c2c4 e8g8 b2b3 b7b6 c1b2 b8d7 f3d2 c8b7 b1c3 c6c5 d4c5 d7c5 c4d5 e6d5 a1c1 d8d7 d2f3 a8d8 f3d4 f6e4 c3b5 a7a6 b5c3 a6a5 d1c2 d8c8 f1d1 e7f6 e2e3 e4c3 b2c3 c5e4 c2d3 e4c3 c1c3 c8c3 d3c3 g7g6');
INSERT INTO evaluations VALUES(46110,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(46110,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(46110,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(46110,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(46110,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(46110,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(46110,36,'cp -27','d2d4 b8d7 b1d2 f8e7 f1e2 c5d4 c3d4 d8a5 f2f4 f6d5 e1g1 e8g8 g1h1 a5d8 d2e4 b7b6 e2b5 f7f5 b5c6 d7e5 c6a8 c8a6 a8d5 d8d5 d4e5 a6f1 d1d5 e6d5 e4d6 e7d6 e5d6 d5d4 c1d2 f1b5 a2a4 b5c6 d2b4 f8d8 a4a5 g8f7 a5b6 a7b6 a1a7 f7f6');
INSERT INTO evaluations VALUES(46109,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(46109,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(46109,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(46109,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(46109,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(46109,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(46109,36,'cp -23','b8d7 b1d2 c5d4 c3d4 f8e7 f1e2 d8a5 e5c4 a5c7 d2f3 e8g8 e1g1 b7b6 c1d2 c8b7 a1c1 c7d8 f3e5 a8c8 e2f3 b7d5 f3d5 f6d5 d1f3 h7h6 f1d1 d7e5 d4e5 b6b5 c4e3 d5e3 d2e3 c8c1 d1c1 e7g5 c1d1 d8c7');
INSERT INTO evaluations VALUES(46511,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(46511,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(46511,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(46511,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(46511,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(46511,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(46511,36,'cp 130','f8e7 h2h3 b8d7 f1d3 f6e4 e1g1 e8g8 a1c1 g7g5 f4h2 e7d6 c3e2 d8f6 c4c5 d6h2 f3h2 f6g7 d1c2 e4f6 b2b4 g8h8 b4b5 c6b5 d3b5 a7a6 b5d3 d7b8 a2a3 b8c6 c1b1 a8b8 f2f3 c8d7 b1b6 d7e8 f1b1');
INSERT INTO evaluations VALUES(40032,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(40032,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(40032,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(40032,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(40032,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(40032,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(40032,36,'cp 136','f1d3 f6e4 h2h3 b8d7 e1g1 e8g8 a1c1 g7g5 f4h2 e7d6 c3e2 d8e7 h2d6 e4d6 e2g3 g8h8 d1d2 a7a5 f3e5 e7g7 g3h5 g7h6 g2g4 d7f6 h5f6 h6f6 f2f4 d6f7 g1g2 f6e7 f1f2 f7e5 f4e5 f8f7');
INSERT INTO evaluations VALUES(46751,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(46751,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(46751,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(46751,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(46751,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(46751,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(46751,36,'cp 18','f8g7 b1c3 e8g8 c4d5 f6d5 c3d5 d8d5 f4c7 b8a6 f1a6 d5g2 d1f3 g2f3 g1f3 b7a6 a1c1 f7f6 f3d2 c8b7 f2f3 e7e5 d4e5 f6e5 e1e2 e5e4 d2e4 a8e8 e4d2 a6a5 e2f2 a5a4 c7d6 f8f6 d6c5 e8d8 c5d4 f6b6 d4g7 d8d2 f2e1 b7f3 e1d2 f3h1');
INSERT INTO evaluations VALUES(41440,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(41440,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(41440,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(41440,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(41440,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(41440,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(41440,36,'cp 17','b1c3 e8g8 c3d5 f6d5 c4d5 d8d5 f4c7 b8a6 f1a6 d5g2 d1f3 g2f3 g1f3 b7a6 a1c1 f7f6 h1g1 c8b7 e1e2 e7e5 d4e5 a8e8 e5f6 g7f6 f3d2 f6b2 c1b1 f8f7 b1b2 f7c7 d2b3 e8e5 g1c1 c7c1 b3c1 b7d5 c1d3 d5c4 b2b8 g8g7 b8c8 c4d3 e2d3 e5a5 d3e2 a5a2 e2f3 a2a4 c8a8 a4a5 a8a7 g7g8 a7a8 g8g7');
INSERT INTO evaluations VALUES(48260,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48260,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48260,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48260,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48260,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48260,36,'cp 0','d5c4 d3c4 c7c5 g1f3 c5d4 e3d4 f8d6 b1c3 h7h6 e1g1 e8g8 d1e2 b8c6 f1d1 b7b6 f3e5 c8b7 c1f4 g8h8 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(48734,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(48734,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(48734,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(48734,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(48734,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(48734,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(48734,36,'cp -33','f8d6 c1g5 h7h6 g5h4 g7g5 h4g3 b8c6 d3d4 e8g8 f1b5 d8e7 d1e2 e7e2 c3e2 f6e4 b5c6 b7c6 g3d6 e4d6 e1c1 c8g4 h2h4 g8g7 e2g3 a8e8 h4g5 h6g5 d1g1 g4f3 g2f3 f7f6 g3h5 g7f7 f3f4 f8g8 f4g5 f6g5 h5g3 f7g6 g1e1 e8e1 h1e1 g8e8 e1e8 d6e8 g3e2 e8d6 c1d2');
INSERT INTO evaluations VALUES(48802,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(48802,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(48802,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(48802,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(48802,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(48802,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(48802,36,'cp 211','d4e5 f6d7 c4d5 d7e5 b1c3 f8c5 g1f3 e5f3 g2f3 e8g8 c3e4 c7c6 e4c5 d8a5 c1d2 a5c5 d1a4 a7a5 d2c3 c6d5 a4h4 c8f5 h4g5 f5g6 a1d1 c5c7 h2h4 d5d4 d1d4 b8c6 g5f4 c7f4 d4f4 c6b4 h4h5 b4d3 f1d3 g6d3 h1g1 g7g6 g1g5 h7h6 g5g1 a8c8 a2a3 f8d8 h5g6');
INSERT INTO evaluations VALUES(49380,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(49380,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(49380,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(49380,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(49380,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(49380,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(49380,36,'cp -27','d7d5 e2e3 c7c5 d3d4 c5d4 e3d4 f8e7 g1f3 e8g8 c4d5 f6d5 f1d3 b7b6 e1g1 c8b7 f1e1 b8c6 c3d5 d8d5 d3e4 d5d7 c1e3 f8d8 d1e2 e7f6 f3e5 c6e5 d4e5 f6e5 e4h7 g8h7 e2h5 h7g8 h5e5 a8c8 h2h4 d7d5 e5g3 c8c2 a1d1 d5d1 e1d1 d8d1 g1h2 b7d5 g3b8 g8h7 b8a7');
INSERT INTO evaluations VALUES(51884,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(51884,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(51884,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(51884,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(51884,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(51884,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51884,36,'cp 0','d2d4 g8f6 b1c3 d7d5 e4d5 f8d6 d1e2 d8e7 e2e7 e8e7 f1c4 b7b5 c4b3 b5b4 c3e2 f6d5 e1g1 h8e8 c2c4 b4c3 e2c3 c8e6 f1e1 e7f8 f3g5 h7h6 g5e6 f7e6 b3a4 d5b6 a4b3');
INSERT INTO evaluations VALUES(52985,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(52985,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(52985,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52985,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52985,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52985,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52985,36,'cp 0','c8d7 b5d7 b8d7 e1g1 e7e5 c2c3 d4c3 b1c3 g8f6 h2h3 a7a6 c1e3 f8e7 f3h4 g7g6 e3h6 e7f8 h6e3');
INSERT INTO evaluations VALUES(30919,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(30919,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(30919,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30919,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(30919,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(30919,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(30919,36,'cp 6','f3d4 d7b5 d4b5 a7a6 b5d4 g8f6 e1g1 f6e4 f1e1 d6d5 c2c4 b8c6 d4c6 b7c6 b1d2 e4c5 b2b4 c5d3 e1e3 d3c1 d1a4 d8d7 a1c1 g7g6 e3d3 f8h6 c4d5 c6d5 a4d7 e8d7 d3d5 d7e6 d5d4 h6d2 c1c6 e6e5 d4d2 a8c8 d2c2 c8c6 c2c6 h8d8 g2g3 d8d6 c6c7 e5e6 c7a7 f7f6 a7a8 e6d5 g1g2 d5e4');
INSERT INTO evaluations VALUES(53183,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53183,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53183,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(53183,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53183,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53183,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53183,36,'cp 0','b8c6 d1e2 c8g4 b1d2 e7e6 h2h3 g4f3 d2f3 g8e7 h3h4 d8a5 c1d2 a5b5 d2c1');
INSERT INTO evaluations VALUES(56373,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(56373,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(56373,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(56373,33,'cp -150',NULL);
INSERT INTO evaluations VALUES(56373,34,'cp -139',NULL);
INSERT INTO evaluations VALUES(56373,35,'cp -153',NULL);
INSERT INTO evaluations VALUES(56373,36,'cp -140','d5e4 f1b5 c7c6 b5a4 c8f5 g1e2 b8d7 e1g1 g8f6 f1e1 e5d4 e2d4 f5g6 a4c2 f8e7 c1g5 e8g8 b1d2 f8e8 g5f6 d7f6 d2e4 g6e4 c2e4 e7c5 d1d3 d8b6 b2b3 c5d4 c3d4 a8d8 a1d1 e8e6 e1e2 e6d6 d4d5 c6d5 e4f3 g7g6 d3d4 d8d7 h2h3 b6d4 d1d4 a7a6');
INSERT INTO evaluations VALUES(56566,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(56566,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(56566,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(56566,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(56566,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(56566,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(56566,36,'cp -144','f1b5 c7c6 b5a4 c8f5 g1e2 b8d7 a4c2 f5g6 e1g1 g8f6 f1e1 f8d6 e2g3 e8g8 b1d2 e5d4 d2e4 d6g3 h2g3 g6e4 c2e4 d4c3 e4f3 d8a5 b2c3 d7e5 c1e3 a8d8 e3d4 f8e8 d1c1 e8e6 d4e5 e6e5 a1b1 e5e1 c1e1 d8e8 e1d2 a5c7');
INSERT INTO evaluations VALUES(2410,30,'mate 2',NULL);
INSERT INTO evaluations VALUES(2410,31,'mate 2',NULL);
INSERT INTO evaluations VALUES(2410,32,'mate 2',NULL);
INSERT INTO evaluations VALUES(2410,33,'mate 2',NULL);
INSERT INTO evaluations VALUES(2410,34,'mate 2',NULL);
INSERT INTO evaluations VALUES(2410,35,'mate 2',NULL);
INSERT INTO evaluations VALUES(2410,36,'mate 2','f3f7 e7d6 c3e4');
INSERT INTO evaluations VALUES(2414,30,'mate 1',NULL);
INSERT INTO evaluations VALUES(2414,31,'mate 1',NULL);
INSERT INTO evaluations VALUES(2414,32,'mate 1',NULL);
INSERT INTO evaluations VALUES(2414,33,'mate 1',NULL);
INSERT INTO evaluations VALUES(2414,34,'mate 1',NULL);
INSERT INTO evaluations VALUES(2414,35,'mate 1',NULL);
INSERT INTO evaluations VALUES(2414,36,'mate 1','e7d6 c3e4');
INSERT INTO evaluations VALUES(2420,30,'mate 1',NULL);
INSERT INTO evaluations VALUES(2420,31,'mate 1',NULL);
INSERT INTO evaluations VALUES(2420,32,'mate 1',NULL);
INSERT INTO evaluations VALUES(2420,33,'mate 1',NULL);
INSERT INTO evaluations VALUES(2420,34,'mate 1',NULL);
INSERT INTO evaluations VALUES(2420,35,'mate 1',NULL);
INSERT INTO evaluations VALUES(2420,36,'mate 1','c3e4');
INSERT INTO evaluations VALUES(3800,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(3800,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(3800,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(3800,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(3800,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(3800,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(3800,36,'cp -26','e1g1 h7h6 g5f3 c6a5 c4b3 a5b3 a2b3 a7a5 d2d3 c7c6 c1e3 c5e3 f2e3 b7b5 d1e1 c8e6 e1g3 g8h7 g3h4 f6g8 h4d8 f8d8 d3d4 f7f6 c3e2 g8e7 a1a3 e5d4 e3d4 f6f5 e4e5 e7d5 f1e1 d6e5 f3e5 d5b4 e2f4 b4c2 f4e6 c2e1');
INSERT INTO evaluations VALUES(6954,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(6954,31,'cp 246',NULL);
INSERT INTO evaluations VALUES(6954,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(6954,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(6954,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(6954,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(6954,36,'cp 203','h5c5 d7d6 c5e3 c3b2 c1b2 g8f6 b1c3 f6g4 e3d2 g4e5 f3e5 c6e5 c3d5 h8f8 e1g1 c7c6 d5e3 c8e6 a1c1 g6g5 f1d1 e8f7 f2f3 a8c8 d2d6 d8d6 d1d6 e5g6 c1d1 c8e8 g1f2 f7g8 h2h3 h7h5 a2a4 g6f4 a4a5 h5h4 b2e5 f4g6 e5h2 g6f4');
INSERT INTO evaluations VALUES(10053,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10053,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(10053,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(10053,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(10053,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(10053,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10053,36,'cp 0','e8g8 b1d2 c5e7 g1f3 d7d6 a2a4 g8h8 e1g1 f6h5 g5e3 f7f5 e4f5 c8f5 d3d4 d8e8 d4e5 c6e5 f3e5 d6e5 d1f3 e5e4 d2e4 f5g6 f3g4 g6f5');
INSERT INTO evaluations VALUES(11177,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(11177,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(11177,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(11177,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(11177,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11177,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(11177,36,'cp 58','c1e3 e7e5 d4b3 c8e6 d1e2 a8c8 e1c1 c6a5 e2d2 h7h5 c1b1 f8e7 b3a5 d8a5 a2a3 b7b5 f1e2 c8b8 c3a2 a5a4 b1a1 b8c8 e2d3 e8g8 a2c3 a4a5 c3d5 a5d8 d5e7 d8e7 e3g5 c8c5 a1b1 d6d5 e4d5 c5d5 d2e2 e7d7');
INSERT INTO evaluations VALUES(14549,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(14549,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(14549,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(14549,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14549,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(14549,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(14549,36,'cp 42','f1d3 f8d6 b1d2 f6d7 f4d6 c7d6 e3e4 d5e4 d2e4 d7f6 e4f6 d8f6 e1g1 e8g8 d3c2 e6e5 d1d3 g7g6 d3e3 c8g4 d4e5 d6e5 c2e4 g4f3 e4f3 g8g7 a1d1 a8d8 h2h3 f6g5 e3e2 d8d1 f1d1 f8d8 g2g3 g5e7 d1d8 c6d8 f3g2 b7b6 a2a3');
INSERT INTO evaluations VALUES(14550,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(14550,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(14550,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(14550,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(14550,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14550,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(14550,36,'cp 43','f8d6 b1d2 f6d7 d1e2 d8e7 f4d6 c7d6 e3e4 d5e4 d2e4 e8g8 e1g1 e6e5 e4g3 d7f6 f1e1 e7c7 d4e5 d6e5 a1d1 c8e6 f3e5 c7e5 e2e5 c6e5 e1e5 e6a2 g3f5 f8d8 f5d4 g8f8 f2f3 a2d5 g1f2 d8d6 e5e1 a7a6 g2g4 d6b6 d1d2 d5c6 h2h4 a8e8 e1e8 c6e8 d3f5');
INSERT INTO evaluations VALUES(24468,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24468,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24468,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(24468,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24468,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24468,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(24468,36,'cp 21','g7g6 f2f3 a7a6 d4c6 d7c6 b5c6 b7c6 c3a4 d8a5 c2c3 f8g7 a4b6 a8d8 e1g1 e8g8 a2a4 c6c5 b6c4 a5c7 d1e2 a6a5 f1d1 f6d7 e3f2 d8a8 f3f4 d7b6 c4a3 c7b7 a3b5 c5c4 f2b6 b7b6 g1h1 a8c8 g2g3 c8c5 d1d2 f8b8 a1f1 g7f6 f1d1 f6g7 e2f3 b6b7 h2h4 b8c8 h1g2');
INSERT INTO evaluations VALUES(25729,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(25729,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(25729,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(25729,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(25729,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(25729,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(25729,36,'cp 66','e1g1 g7g6 d2d4 f8g7 h2h3 g8e7 d4d5 c6b8 a4c2 e8g8 c3c4 a6a5 a2a3 e7c8 c1e3 d8e8 b1c3 c8b6 e3b6 c7b6 f1e1 e8e7 c2a4 d7a4 c3a4 b8d7 b2b4 h7h5 a4c3 g7h6 d1c2 f7f5 a1b1');
INSERT INTO evaluations VALUES(33201,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33201,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33201,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33201,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33201,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33201,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33201,36,'cp 0','g1f3 b8c6 b1c3 g8e7 c3b5 c5b6 c1e3 b6a5 e3d2 a5b6');
INSERT INTO evaluations VALUES(35704,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(35704,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(35704,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(35704,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(35704,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(35704,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(35704,36,'cp 21','d2d4 c5d4 f3d4 d7d5 c4d5 f6d5 d4b3 d5b6 a2a4 a7a5 b1c3 b8c6 c1f4 c8e6 b3c5 d8c8 c5e6 c8e6 f4c7 b6c4 d1b3 a8c8 b3b7 g7c3 b2c3 c6d8 c7d8 c8d8 g2f3 c4d2 f1d1 e6f6 a1c1 d2f3 e2f3 d8d1 c1d1 f6c3 b7e7 c3f3 d1d8');
INSERT INTO evaluations VALUES(35707,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(35707,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(35707,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(35707,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(35707,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(35707,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(35707,36,'cp 26','c5d4 f3d4 d7d5 c4d5 f6d5 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 f1d1 d8f8 a2a4 b2e2 a1c1 c8e6 f4e3 a7a6 c6f3 e2a2 c3c4 a2a4 c4c5 g7b2 c1b1 a4b4 c5c6 e6f5 e3c5 b4b3 c5e7 f5b1 e7f8 g8f8 d1b1 b3f3 b1b2');
INSERT INTO evaluations VALUES(35717,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(35717,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(35717,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(35717,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(35717,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(35717,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(35717,36,'cp 13','f3d4 d7d5 c4d5 f6d5 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 a1d1 d8f8 c3c4 c8a6 c6b5 a6b5 c4b5 b2b5 d1d7 b5a5 f1b1 a5a2 d7e7 g7f6 e7d7 a2e2 b1b7 a7a5 f4h6 f6g7 h6e3 g7e5 b7a7 e5c3 a7c7 c3g7 g1g2 g7e5 c7c5 e5f6 c5a5 f8e8');
INSERT INTO evaluations VALUES(35714,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(35714,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(35714,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35714,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(35714,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35714,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(35714,36,'cp 15','d7d5 c4d5 f6d5 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 f1d1 d8f8 a2a4 b2e2 a1c1 c8e6 c6d5 e6d5 d1d5 f8c8 d5d7 h7h5 f4e3 c8c3 c1c3 g7c3 d7c7 c3d2 e3d2 e2d2 c7e7 a7a5 e7a7 d2a2 a7a8 g8h7 a8a5 a2a3 g1g2 h7g7 a5a8');
INSERT INTO evaluations VALUES(36352,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36352,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(36352,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(36352,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(36352,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(36352,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36352,36,'cp 5','f1e2 b7b6 e1g1 c8b7 c2c3 c7c5 b1d2 e7d6 e2d3 d6f4 e3f4 b7a6 d3a6 b8a6 f3e5 a6b8 f1e1 d8c7 d2f1 b8c6 f1g3 c6e7 d1d2 f6e8 d4c5 b6c5 c3c4 a8d8 c4d5 d8d5 d2e3 e8f6 a1c1 f8d8 e3f3 e7g6 e5g6 h7g6');
INSERT INTO evaluations VALUES(37727,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(37727,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(37727,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(37727,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(37727,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(37727,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(37727,36,'cp 36','g8f6 e2e4 c6c5 g1e2 b8c6 e1g1 e8g8 a2a3 d7d6 a1b1 a7a6 d2d3 a8b8 b2b4 b7b5 b4c5 d6c5 c1e3 b5c4 b1b8 c6b8 d3c4 f6d7 f2f4 c8b7 d1b1 b7c6 f1d1 d8c7 c3d5 c7d8 e3f2 e7e6 d5e3 d8a5 e4e5 c6g2 g1g2');
INSERT INTO evaluations VALUES(37997,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(37997,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(37997,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(37997,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(37997,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(37997,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(37997,36,'cp 61','f1e2 b8c6 c1e3 c6d4 e3d4 g7d4 d1d4 g8f6 b1c3 d8a5 e1g1 e8g8 b2b3 c8e6 f1e1 a8c8 e2f1 a7a6 a2a4 a5c5 d4d2 c5b4 e1e3 c8c7 d2e1 f8c8 h2h3 e6d7 a1d1 d7c6 c3d5 b4e1 d5f6 g8g7 f6h5 g7h6 e3e1 h6h5 e4e5 d6e5 e1e5 h5h6 a4a5');
INSERT INTO evaluations VALUES(38776,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(38776,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(38776,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(38776,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(38776,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(38776,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(38776,36,'cp 21','b4d6 e2e3 g8f6 f1d3 e8g8 e1g1 b7b6 f1e1 c8a6 d3a6 b8a6 f4g5 d6e7 d1e2 a6b8 c3c4 b8d7 b1c3 h7h6 g5f4 d8c8 c3b5 c7c5 c4d5 f6d5 f4g3 a7a6 e3e4 d5f6 b5c3 c5d4 f3d4 c8b7 a1d1 f8c8 e4e5 f6d5 c3d5 b7d5 e2g4');
INSERT INTO evaluations VALUES(38662,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(38662,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(38662,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(38662,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38662,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38662,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(38662,36,'cp 34','e2e3 g8f6 f1b5 b8d7 e1g1 e8g8 f1e1 d6f4 e3f4 c7c5 b5d3 a7a5 b1d2 d8c7 g2g3 c5c4 d3c2 b7b5 f3e5 b5b4 d2f3 d7b6 d1e2 c8d7 e2e3 d7a4 c2b1 a4c6 e5c6 c7c6 f3e5 c6c7 b1c2 b6d7 h2h4 c7b6 g1g2 f8c8 e3e2 a5a4 g3g4');
INSERT INTO evaluations VALUES(38897,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(38897,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(38897,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(38897,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(38897,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(38897,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(38897,36,'cp 87','e2e3 g8f6 f1c4 e8g8 g1e2 c7c5 e1g1 b8c6 e2g3 d8c7 c1b2 f8d8 d1e2 e6e5 a1d1 c5d4 c3d4 e5d4 e3d4 c7e7 d4d5 e7e2 g3e2 c6e7 b2f6 g7f6 e2g3 c8d7 g3h5 f6f5 d5d6 e7g6 c4d5 a8b8 d1d2 d7e6 d5e6 f7e6 d6d7');
INSERT INTO evaluations VALUES(39270,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39270,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39270,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39270,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(39270,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39270,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(39270,36,'cp -3','g1f3 d7d6 c2c3 g8f6 d2d4 c5b6 f1d3 b8c6 d4d5 c6e7 c3c4 e8g8 b1c3 e7g6 h2h3 c8d7 a3a4 f6h5 a4a5 b6c5 g2g3 c7c6 f3d2 h5f6 d2b3 c5b4 c1d2 b4c3 d2c3 b7b5 a5b6 a7b6 a1a8 d8a8 f2f3 c6d5 c4d5 f6h5 e1f2');
INSERT INTO evaluations VALUES(39269,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39269,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39269,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39269,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39269,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39269,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39269,36,'cp 0','d7d6 c2c3 g8f6 d2d4 c5b6 f1d3 b8c6 d4d5 c6e7 c3c4 e7g6 b1c3 e8g8 c3a4 c8g4 a4b6 a7b6 e1g1 f6h5 h2h3 g4f3 d1f3 h5f4 f3d1 g6h4 g2g3 d8c8 g3h4 c8h3 c1f4 e5f4 f2f3 h3g3 g1h1 g3h4 h1g2 h4g3 g2h1');
INSERT INTO evaluations VALUES(39272,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(39272,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(39272,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(39272,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(39272,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(39272,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(39272,36,'cp -38','d7d6 b1c3 b8c6 g1f3 a7a6 c3d5 g8f6 c2c3 f6d5 e4d5 c6e7 f4e5 d6e5 f3e5 e8g8 d3d4 e7g6 e5f3 c5a7 f1e2 d8d5 e1g1 c8f5 e2d3 f5d3 d1d3 f8e8 b2b3 b7b5 c1e3 c7c5 a1e1 c5d4 e3d4 f7f6 g1h1 d5d7 d3d2 e8e1 f1e1 a8d8');
INSERT INTO evaluations VALUES(40059,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40059,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40059,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(40059,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40059,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(40059,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40059,36,'cp 0','e6d5 e2e3 c8f5 f3h4 f5e6 f1d3 f6e4 g5e7 d8e7 h4f3 e4d6 d1c2 g7g6 h2h3 b8d7 e1g1 e8g8 f1e1 f8e8 b2b4 e7f6 f3d2 f6g5 d2f3');
INSERT INTO evaluations VALUES(41385,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(41385,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(41385,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41385,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(41385,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(41385,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(41385,36,'cp 29','e8g8 e1g1 e7e5 c2c3 d6d5 d4e5 f6e4 b1d2 b8c6 d2e4 d5e4 d3e4 c6e5 f1e1 d8d1 a1d1 e5f3 e4f3 c8e6 a2a4 a7a5 f3d5 e6d5 d1d5 c7c6 d5d2 b7b5 e1a1 f8b8 g2g3 h7h5 g1f1 b5a4 a1a4 b8b5 a4c4 b5b7 c4c6 a8b8 f1e2 b7b2 d2b2');
INSERT INTO evaluations VALUES(36678,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36678,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36678,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(36678,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36678,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(36678,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(36678,36,'cp 33','e1g1 e7e5 d4e5 d6e5 a2a4 d8e7 a4a5 f8d8 d1e2 h7h6 f1e1 c7c5 d3b5 c8d7 f3d2 b8c6 b5c6 d7c6 b1c3 f6e8 d2c4 e8c7 a1d1 c7b5 c3b5 c6b5 d1d5 b7b6 e1d1 d8d5 e4d5 b6a5 b2b3 a5a4 h2h3 a4b3 c2b3 e7d7 e2d2 g8h7 e3c5');
INSERT INTO evaluations VALUES(42048,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42048,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42048,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42048,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(42048,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42048,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42048,36,'cp 0','a2a3 b4c3 b2c3 e8g8 c4d5 e6d5 e2e3 c8f5 g1e2 b8c6 e2g3 f5g6 c3c4 f8e8 e1f2 h7h5 c4d5 d8d5 g3e2 g6h7 e2f4 d5d6 f1c4 c6a5 c4a2 c7c5 h1e1 a8d8 e3e4 c5d4 f4d3 d6h2 c1f4 h2h4 f4g3 h4g5 g3f4');
INSERT INTO evaluations VALUES(42064,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(42064,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42064,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(42064,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(42064,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(42064,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(42064,36,'cp 6','b4c3 b2c3 e8g8 c4d5 e6d5 e2e3 c8f5 g1e2 b8c6 e2g3 f5g6 f1d3 c6a5 e1g1 c7c5 d3g6 h7g6 e3e4 d5e4 f3e4 c5d4 c3d4 d8b6 c1e3 a5c4 e3f2 b6e6 d1e2 a8c8 e4e5 f6d5 g3e4 f7f6 e5f6 d5f6 e4f6 f8f6 e2e6 f6e6 d4d5 e6d6 f1d1 c4b6 h2h3 d6d5 f2b6 d5d1 a1d1');
INSERT INTO evaluations VALUES(43401,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(43401,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(43401,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(43401,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(43401,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(43401,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(43401,36,'cp 67','g7g6 f1e2 b8c6 c1e3 f8g7 d1d2 c8d7 e1g1 e8g8 d4c6 d7c6 f2f3 f6d7 f1d1 d8b8 a1c1 b7b5 b2b4 f8e8 c3d5 c6d5 c4d5 b8b7 c1c2 e8c8 d1c1 d7e5 f3f4 e5c4 e2c4 b5c4 c2c4 c8c4 c1c4 b7b5 c4c1 b5a4 h2h3 a6a5 b4a5 a4e4');
INSERT INTO evaluations VALUES(43634,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43634,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(43634,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(43634,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(43634,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(43634,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(43634,36,'cp 73','d2d4 b8d7 a2a4 e5d4 d1d4 f8e7 f1c4 e8g8 e1g1 d7c5 c1e3 c8e6 a4a5 c7c6 c4e6 c5e6 d4b4 d8b8 a1d1 f8e8 b4c4 e7f8 f1e1 h7h6 c4d3 f6d7 c3a4 b8d8 a4b6 e6c5 e3c5 d7c5 d3d4 a8b8 e4e5 d6e5 d4c3 f8d6 f3e5');
INSERT INTO evaluations VALUES(45094,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(45094,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(45094,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(45094,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(45094,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(45094,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(45094,36,'cp 66','b1c3 d6d5 e2e3 g7g6 f1e2 f8g7 e1g1 e8g8 h2h3 e7e6 b2b3 f6e4 c1b2 d8e7 d1c2 e4c3 b2c3 b7b6 f1c1 c8a6 a2a4 f8d8 e2f1 d8c8 a4a5 b8d7 f3d2 e6e5 a5b6 a7b6 d4e5');
INSERT INTO evaluations VALUES(46051,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(46051,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(46051,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(46051,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(46051,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46051,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(46051,36,'cp 20','b8c6 c1e3 f8e7 e1g1 e8g8 d1e2 e6e5 d4f3 d6d5 e4d5 f6d5 c3d5 d8d5 e3g5 c6d4 f3d4 e7g5 d4b5 d5c5 f1e1 c8e6 b5c3 c5d4 a1b1 h7h6 a2a4 f8d8 e2e5 a8c8 e5e2 g5f6 c3e4 f6e7 b2b3 d4b6');
INSERT INTO evaluations VALUES(49678,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(49678,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(49678,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(49678,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(49678,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(49678,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(49678,36,'cp -55','e5f4 b1c3 d7d5 e4d5 f6d5 f1b5 c7c6 d1e2 c8e6 c3d5 d8d5 b5a4 d5d8 a4b3 d8f6 e1g1 f8d6 f1e1 b8a6 b3e6 f7e6 e2e6 f6e6 e1e6 e8d7 e6e2 h8e8 e2e8 a8e8 a2a3 h7h6 b2b4 g7g5 h2h3 d7e6 c1b2 e6f5 b2d4 d6b8 d4f2 h6h5 f3d4 f5g6 d4b3 a6c7 g1f1 g5g4 a1e1 e8e1 f1e1');
INSERT INTO evaluations VALUES(50839,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(50839,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50839,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(50839,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(50839,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50839,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50839,36,'cp 0','c7c5 c2c4 d5c4 d3c4 c5d4 e3d4 e7e6 g1f3 b7b5 c4d3 c8b7 a2a4 b5a4 d1a4 b7c6 a4c4 c6d5 c4a4');
INSERT INTO evaluations VALUES(51203,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(51203,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(51203,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(51203,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(51203,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(51203,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(51203,36,'cp -12','c4a2 c8b7 g1f3 d7d6 e1g1 g8f6 b1d2 b8d7 f1e1 g7g6 b2b4 f8g7 a1b1 e8g8 c2c4 d8c7 c1b2 c5b4 a3b4 a6a5 b4a5 c7a5 a2b3 b5c4 b3c4 f8b8 b2d4 b7c6 b1b8 a8b8 d1a1 b8a8 a1a5 a8a5 e1a1 a5a1 d4a1 d7c5 d2b3');
INSERT INTO evaluations VALUES(51222,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(51222,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(51222,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(51222,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(51222,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(51222,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(51222,36,'cp -14','g7g6 b1c3 f8g7 g1f3 d8c7 e1g1 d7d6 c1f4 b8c6 d1d2 g8f6 f4h6 e8g8 h6g7 g8g7 h2h3 c8b7 c3e2 a6a5 a1e1 a8d8 f3h2 c5c4 h2g4 f6g4 h3g4 c6e5 f2f3 d6d5 e4d5 b7d5 d3d4 e5d7 c2c3 d7f6 a2b1 h7h6 g1h1 f6d7 f3f4');
INSERT INTO evaluations VALUES(51781,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(51781,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(51781,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(51781,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(51781,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(51781,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(51781,36,'cp -4','d7d5 d2d4 d5c4 g1f3 g8f6 f3e5 b8d7 e5c4 b7b5 c4e5 c8b7 f1g2 d7e5 d4e5 d8d1 c3d1 f6d5 e1g1 e8c8 c1d2 d8d7 a2a4 d5b6 a4b5 c6b5 d2e3 f8b4 e3b6 a7b6 d1c3 b4c3 b2c3 b7g2 g1g2 c8b7 f2f4 h8c8 a1b1 d7d2 g2f2 c8c3 b1b5 c3c2 f1b1 d2e2 f2f3 b7c8 b5b6 e2h2');
INSERT INTO evaluations VALUES(53562,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53562,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53562,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(53562,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53562,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53562,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53562,36,'cp 3','d2d4 d5c4 g1f3 g8f6 f1g2 b8d7 e1g1 b7b5 e2e4 b5b4 e4e5 f6d5 c3e4 d7b6 f1e1 f8e7 c1g5 e8g8 g5e7 d8e7 a1c1 c8a6 g2f1 a6b5 f3d2 c4c3 b2c3 b4c3 e4c3 d5c3 c1c3 f8d8 d2e4 e7b4 a2a3 b4d4 d1d4 d8d4 e4d6 a8d8 d6b5 c6b5');
INSERT INTO evaluations VALUES(53084,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(53084,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(53084,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(53084,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(53084,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(53084,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(53084,36,'cp 22','d2d4 c5d4 e2d4 g8f6 b1c3 a7a6 h2h3 e7e5 d4b3 c8e6 c1e3 b8d7 d1f3 a8c8 f1e2 f8e7 a2a4 d7b8 e1g1 b8c6 c3d5 e6d5 e4d5 c6b4 e3d2 b4d5 c2c4 d5c7 f3b7 c8b8 b7f3 d8d7 d2a5 e8g8 f1d1 c7e6 c4c5 d6d5 e2a6');
INSERT INTO evaluations VALUES(1848,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(1848,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(1848,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(1848,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(1848,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(1848,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(1848,36,'cp 66','c6e5 b1c3 e5f3 g2f3 f8b4 c1d2 g8e7 d1e2 b7b6 h1g1 c8b7 g1g4 e7g6 e1c1 f6e7 a2a3 b4c3 d2c3 e8c8 e2e3 f7f6 h2h4 h8e8 h4h5 g6e5 g4g1 e7f7 b2b3 d7d5 f3f4 e5d7 e3h3 c8b8 e4d5 b7d5 f1b5 a7a6 b5a6 d7c5 a6b5 c7c6 b5d3');
INSERT INTO evaluations VALUES(3856,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(3856,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(3856,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(3856,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(3856,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3856,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3856,36,'cp 13','c6d4 f3h4 c7c6 e2e3 d4f5 h4f3 a7a5 d3d4 e5d4 f3d4 c5d4 e3d4 d6d5 c4d5 f6d5 c3d5 c6d5 f1e1 c8e6 c1d2 f8e8 e1e5 d8d7 d2f4 f7f6 e5e1 h7h6 a1c1 d7b5 d1d2 e6f7 b2b3 e8e1 c1e1 b5b4 d2b4');
INSERT INTO evaluations VALUES(4220,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(4220,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(4220,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(4220,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(4220,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(4220,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(4220,36,'cp -23','b4c3 d2c3 d7d5 e4d5 f6d5 e1g1 c8g4 f1e1 f8e8 h2h3 g4h5 d1e2 d8d6 c4d5 d6d5 g2g4 h5g6 f3h4 e8e6 f2f4 e5f4 e2f3 d5f3 h4f3 e6d6 c3d2 d6d7 d2f4 f7f6 f4d2 c6e7 d2b4 e7d5 f3h4 g6f7 g1f2 a8d8');
INSERT INTO evaluations VALUES(6648,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(6648,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(6648,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6648,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(6648,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(6648,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6648,36,'cp 20','f2e3 d7d5 e4d5 d8d5 b1c3 d5d6 d1d2 g8f6 e1c1 e8g8 f1e2 a8b8 f3g5 b7b5 c1b1 f8d8 h2h4 a7a5 d1f1 h7h6 c3e4 f6e4 g5e4 d6b4 e2f3 f7f5 d2b4 c6b4 e4c5 b4d5 f1e1 d5b6 a2a3');
INSERT INTO evaluations VALUES(6647,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(6647,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(6647,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(6647,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(6647,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(6647,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6647,36,'cp 0','d7d5 e4d5 d8d5 d1d2 d5d6 b1c3 h7h6 d3d4 e5d4 f3d4 c6d4 d2d4 d6d4 e3d4 c8e6 d4d5 e6d7 d5d6 c7c6 f1c4 g8f6 e1g1 e8g8 a1e1 f8e8 c4b3 a8d8 e1e7 e8e7 d6e7 d8e8 c3e4 f6d5 e4d6 e8e7 d6b7 e7e2 f1f2 e2e1 f2f1');
INSERT INTO evaluations VALUES(8729,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(8729,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(8729,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8729,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(8729,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(8729,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(8729,36,'cp 20','f6h5 b1d2 h5f4 e3f4 d8d6 g2g3 f7f6 e1g1 e8g8 c3c4 c8g4 d1b3 c6a5 b3c3 a5c4 d2c4 d5c4 d3c4 g8h8 a1d1 a8d8 d1d2 c7c5 c3e3 c5d4 d2d4 d6c6 c4e2 c6e6 e3d2 e6b6 f1d1 d8d4 d2d4 b6d4 f3d4 g4h3 e2c4 e7e5 f4e5 f6e5 d4e6 h3g4 d1e1 g4e6');
INSERT INTO evaluations VALUES(9278,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(9278,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(9278,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(9278,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(9278,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(9278,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(9278,36,'cp 9','e1g1 b4c3 b2c3 e8g8 f1e1 a7a6 b5c6 b7c6 h2h3 h7h6 d3d4 f6d7 c3c4 e5d4 f3d4 c6c5 d4f5 d7b6 c1b2 c8f5 e4f5 b6a4 a1b1 a4b2 b1b2 d8f6 b2b7 f6f5 b7c7 a8e8 e1e8 f8e8 c7c6 f5e4 c6a6 e4c4 a6d6 c4a2');
INSERT INTO evaluations VALUES(10454,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(10454,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(10454,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(10454,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(10454,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(10454,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(10454,36,'cp -30','f6e4 f3e5 d8f6 e5g4 f6e7 g4e3 e4c3 d1h5 h7h6 f1d3 d7d6 c1d2 c3e4 e3d5 e4g3 d5e7 g3h5 e7c6 b7c6 f2f3 c8e6 a1b1 e8d7 g2g4 h5f6 c2c4 h8b8 e1f2 a7a5 a3a4 d7d8 h2h4 b8b1 h1b1 f6d7 d2c3 f7f6 d3e4 d6d5 c4d5 c6d5 e4c2 d7b6 g4g5 h6g5 h4g5 d8d7 g5f6');
INSERT INTO evaluations VALUES(13147,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13147,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13147,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13147,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13147,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13147,36,'cp 8','g1f3 e7e6 f1d3 f8d6 e1g1 e8g8 h2h3 h7h6 d1e2 c6b4 d3b1 b7b6 c3b5 d6b8 c1d2 b4c6 b5c3 c8b7 a2a3 f8e8 f1d1 d8e7 b1a2 b8d6 a1c1 a8d8 d2e1 d6b8 a2b1 d8c8 b1d3 e6e5 f3e5 c6e5 d4e5');
INSERT INTO evaluations VALUES(16370,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(16370,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(16370,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(16370,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(16370,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(16370,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(16370,36,'cp -63','f8c5 f1c4 e8g8 d2d3 c6a5 c4b3 a5b3 a2b3 h7h6 g5f3 f8e8 c1e3 c5b4 e1g1 b7b6 e3d2 c8b7 f1e1 a7a5 h2h3 d7d6 c3e2 b4d2 d1d2 d6d5 e2g3 d5e4 d3e4 d8e7 g1h2 e7d7 d2d7 f6d7 e1e3 e8e6 a1d1 d7c5');
INSERT INTO evaluations VALUES(18323,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(18323,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(18323,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(18323,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(18323,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(18323,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(18323,36,'cp 78','f8b4 c3d5 b4d6 g1f3 g8e7 d5c3 c8b7 c3b5 e8g8 e1g1 e7c8 e2e3 e5e4 f3e1 f7f5 d2d3 c6e5 b2b3 d8g5 c1b2 e5g4 d1d2 a7a6 b5d6 c8d6 h2h3 g4f6 b2e5 f6e8 a1d1 g5e7 e5a1 e8f6 d2e2 a8e8 d3e4 b7e4 f2f3 e4b7 e1c2 d6f7');
INSERT INTO evaluations VALUES(19185,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(19185,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(19185,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(19185,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(19185,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(19185,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(19185,36,'cp 124','d1d4 g8e7 d4e5 d7d6 e5g3 e7g6 h2h4 h7h5 c1g5 f8e7 b1c3 e7g5 h4g5 h5h4 g3e3 c8d7 f2f4 d8b6 e3b6 a7b6 f4f5 g6e5 a1d1 g7g6 d1d6 g6f5 e4f5 e8c8 c3d5 h8h5 d5b6 c8c7 d6d5 c7b6 d5e5 d7c6 g2g4 h5g5 h1h4 d8d4 f1e2 f7f6 e5e6 g5f5 g4f5 d4h4 e6f6');
INSERT INTO evaluations VALUES(20345,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(20345,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(20345,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(20345,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(20345,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(20345,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(20345,36,'cp -75','g8h6 c3e2 f7f6 d2d4 f8e7 c2c3 e8g8 e5f6 e7f6 c1e3 h6g4 d1d2 g4e3 d2e3 d8a5 e3d2 b7b5 d4c5 b5b4 e2d4 b4c3 b2c3 c8d7 f1e2 a5c5 e1g1 e6e5 f4e5 c6e5 f3e5 f6e5 f1f8 a8f8 a1f1 e5f6 g1h1 f8c8 d4b3 c5d6 b3d4 g8h8 e2d3 d6c5 d4e2');
INSERT INTO evaluations VALUES(20416,30,'cp -206',NULL);
INSERT INTO evaluations VALUES(20416,31,'cp -213',NULL);
INSERT INTO evaluations VALUES(20416,32,'cp -211',NULL);
INSERT INTO evaluations VALUES(20416,33,'cp -204',NULL);
INSERT INTO evaluations VALUES(20416,34,'cp -217',NULL);
INSERT INTO evaluations VALUES(20416,35,'cp -197',NULL);
INSERT INTO evaluations VALUES(20416,36,'cp -213','c6e5 f2f4 e5g6 f1b5 c8d7 d1e2 d7b5 e2b5 d8d7 f4f5 e6f5 a2a4 f5f4 e1g1 e8c8 c1f4 g6f4 b5d7 d8d7 f1f4 f8d6 f4f1 g8f6 f1e1 a7a6 b1d2 g7g6 d2f3 h8e8 e1e8 f6e8 a1e1 e8f6 g1f1 d7c7 g2g3 f6e4 d4c2 c7e7 e1d1 e4f6 c3c4 e7e4 c4d5 e4a4');
INSERT INTO evaluations VALUES(20779,30,'cp -199',NULL);
INSERT INTO evaluations VALUES(20779,31,'cp -192',NULL);
INSERT INTO evaluations VALUES(20779,32,'cp -189',NULL);
INSERT INTO evaluations VALUES(20779,33,'cp -186',NULL);
INSERT INTO evaluations VALUES(20779,34,'cp -176',NULL);
INSERT INTO evaluations VALUES(20779,35,'cp -189',NULL);
INSERT INTO evaluations VALUES(20779,36,'cp -201','f2f4 e5g6 f1b5 c8d7 d1e2 d7b5 d4b5 f8e7 c1e3 d8d7 b1d2 g8f6 e1g1 g6h4 b5d4 e8g8 g2g3 h4f5 d4f5 e6f5 d2f3 f8e8 e2d3 g7g6 e3d4 f6e4 f1e1 a7a6 g1g2 d7c6 e1e2 b7b5 a1d1 c6b7 d4e5 f7f6 e5d4 e7f8 d4g1 a8d8 f3d4 d8c8');
INSERT INTO evaluations VALUES(25500,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(25500,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(25500,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25500,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(25500,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(25500,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(25500,36,'cp 48','c6d4 f3d4 e5d4 b5f1 f8e7 d1d4 e8g8 b1c3 f8e8 c1f4 f6h5 f4e3 e7f6 d4d2 d7c6 f2f3 f6e5 a1d1 a7a6 c3d5 h5f6 c2c4 f6d5 c4d5 c6b5 e3g5 e5d4 d2d4 d8g5 d4b4 b5f1 g1f1 a8b8 b4d2 g5d2 d1d2 g7g6 e1b1 h7h6 b1c1 e8e7 f1e2 f7f5');
INSERT INTO evaluations VALUES(26259,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26259,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(26259,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26259,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26259,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26259,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26259,36,'cp 0','g1f3 d8b6 b1d2 g8f6 d4c5 b6b2 f3d4 e7e5 f4e5 f5g4 a1b1 b2a3 f2f3 c6e5 f3g4 f8c5 d4b5 a3a2 b1a1 a2b2 a1c1 b2a2');
INSERT INTO evaluations VALUES(26260,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26260,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26260,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26260,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26260,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26260,36,'cp 0','d8b6 b1d2 g8f6 d4c5 b6b2 f3d4 f5g4 f2f3 e7e5 d4b5 a8c8 f4g5 g4d7 d2b3 a7a6 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(33317,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(33317,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(33317,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(33317,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(33317,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(33317,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(33317,36,'cp 249','f1d3 g8f6 e1g1 e8g8 f1e1 b8c6 c2c3 f6e8 b1d2 f7f6 d2c4 g8h8 d4d5 c6d8 a2a4 c7c6 c4d6 e8d6 a4a5 d6f7 c3c4 d7d6 h2h3 f7g5 f3h4 c6c5 b2b4 c5b4 c1d2 d8f7 d2b4 f8g8 d3f1 c8d7 h4f5 d7f5 e4f5 e5e4');
INSERT INTO evaluations VALUES(39045,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(39045,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(39045,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(39045,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(39045,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(39045,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(39045,36,'cp 90','d2d4 b7b6 c1e3 c8b7 b1c3 g8e7 f1d3 d7d5 e4e5 e7f5 e3f2 h7h5 b2b4 g7f8 a1b1 a7a5 b4b5 f8b4 b1b3 c7c5 b5c6 b8c6 e1g1 b7a6 c3b5 a5a4 b3b1 b4e7 c2c3 c6a5 d1a4 e8g8 a4d1 a5c4 d1e2 d8d7 g1h1 a8c8');
INSERT INTO evaluations VALUES(40125,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(40125,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(40125,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(40125,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(40125,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40125,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(40125,36,'cp 32','d2d4 e5d4 f3d4 e8g8 e1g1 b7b5 c4d3 f8e8 c3e2 a7a6 e2g3 c6c5 d4f5 c8f5 g3f5 c5c4 d3e2 b8d7 e2f3 d7e5 a2a4 e7f8 c1g5 h7h6 g5h4 b5a4 f5e3 e5f3 g2f3 a8b8 d1d4 f8e7 e3f5 d8d7 d4c4 b8b2 c4a4 d7a4 a1a4 b2c2 f5e7 e8e7 a4a6 e7d7 a6a8 g8h7 h4f6 g7f6 f1d1');
INSERT INTO evaluations VALUES(40188,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(40188,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(40188,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(40188,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(40188,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(40188,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(40188,36,'cp 103','a7a6 h2h3 e6d5 e4d5 c8f5 f1d3 f5d3 d1d3 e8g8 e1g1 f8e8 c1f4 b8d7 f1e1 h7h6 b2b3 d7f8 g2g4 d8d7 f4g3 f8h7 a1d1 e7f8 g1g2 e8e1 d1e1 a8e8 e1b1 g7g6 a2a3 e8c8 a3a4 b7b6 b1e1 c8e8 e1e8 f6e8 f3d2 f8g7 g3f4 h7f8');
INSERT INTO evaluations VALUES(45036,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(45036,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(45036,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(45036,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(45036,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(45036,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(45036,36,'cp -66','d8d4 g1f3 d4d5 c2c4 d5d8 e4f6 g7f6 e1g1 c8g4 b2b4 e7e6 h2h3 g4h5 b4b5 b8d7 d3e2 h5g6 c1f4 f8c5 b5c6 b7c6 d1a4 e8g8 a4c6 d8c8 c6a4 g8g7 f1d1 f8d8 d1d2 d7f8 a1d1 d8d2 f3d2 e6e5 e2g4');
INSERT INTO evaluations VALUES(46443,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(46443,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(46443,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(46443,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(46443,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(46443,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(46443,36,'cp -24','f8e7 e1g1 e8g8 f1e1 b8c6 c1f4 b7b5 c2c3 c8f5 d3d4 a7a5 b3c2 f5c2 d1c2 d8b6 f4e3 b5b4 b1d2 a5a4 h2h3 h7h6 d4c5 e7c5 e3c5 b6c5 a1c1 f8d8 c3c4 d5d4 c2d3 b4b3 a2a3 d8e8 e1e8 a8e8 c1d1 c5b6 d2f1 e8d8 f1g3');
INSERT INTO evaluations VALUES(52162,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(52162,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(52162,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(52162,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(52162,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(52162,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(52162,36,'cp 111','g7g6 e2e4 h7h5 g2g3 d7d6 b1c3 h5h4 f1h3 c8h3 g1h3 f8h6 g3g4 h6c1 d1c1 g8f6 f2f3 e8f8 c1d2 b8d7 e1g1 f8g7 g1h1 a7a6 a2a4 d8b6 d2c2 b6d8 f1g1 d7f8 b2b3 f8h7 a1f1 a8b8 c2d2 d8e7 g1g2 e7d7 h3f2 b7b6');
INSERT INTO evaluations VALUES(52254,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(52254,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(52254,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(52254,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(52254,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(52254,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(52254,36,'cp 62','d7d6 g1f3 g8f6 h2h3 f8g7 a2a4 e8g8 c1e3 b8d7 f1e2 e7e5 d4e5 d7e5 f3e5 d6e5 d1d8 f8d8 e2c4 f6h5 e1g1 h5f4 f1d1 c8e6 c4f1 d8d1 a1d1 g7f8 a4a5 f8b4 a5a6 b7b5 c3e2 f4e2 f1e2 f7f6 g2g3 b4e7 g1g2 g8f7 f2f4 e5f4 g3f4 f6f5 e4e5');
INSERT INTO evaluations VALUES(52589,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(52589,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(52589,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(52589,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(52589,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(52589,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(52589,36,'cp 82','a2a3 f8g7 f2f4 h7h5 g1f3 d6d5 e4e5 c8f5 h2h3 e7e6 c1e3 g7f8 f1d3 f5d3 d1d3 b8d7 e3f2 a7a6 c3d1 g8e7 e1g1 f8h6 d3d2 d8c7 b2b3 a8c8 b3b4 d7b6 d1b2 e7f5 b2d3');
INSERT INTO evaluations VALUES(52534,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(52534,31,'cp -111',NULL);
INSERT INTO evaluations VALUES(52534,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(52534,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(52534,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(52534,35,'cp -165',NULL);
INSERT INTO evaluations VALUES(52534,36,'cp -141','c3b2 c1b2 e7e6 f1d3 d7d6 e1g1 g8f6 e4e5 d6e5 f3e5 f8e7 d1e2 b8d7 f1d1 d7e5 b2e5 c8d7 b1c3 d8a5 a1b1 d7c6 d3b5 e8g8 b5c6 b7c6 b1b7 a8e8 e5d4 c6c5 d4e5 c5c4 g2g3 h7h6 e5d4 a7a6 e2c4 e8c8 c4a4 a5a4 c3a4 f6d5');
INSERT INTO evaluations VALUES(52537,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(52537,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(52537,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(52537,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(52537,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(52537,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(52537,36,'cp -115','c1b2 e7e6 f1d3 g8f6 e1g1 d7d6 e4e5 d6e5 f3e5 b8d7 d1f3 f8c5 f3g3 e8g8 b1d2 d7e5 b2e5 f6h5 g3h3 g7g6 a1d1 f7f6 d2e4 f6e5 e4c5 h5f4 h3e3 d8d4 f1e1 d4e3 f2e3 f4d3 c5d3 e5e4 d3e5 b7b6 d1d6 c8b7 d6d7 b7d5 e1c1 f8d8 d7d8 a8d8');
INSERT INTO evaluations VALUES(52536,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(52536,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(52536,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(52536,33,'cp -124',NULL);
INSERT INTO evaluations VALUES(52536,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(52536,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(52536,36,'cp -126','e7e6 f1d3 d7d6 e1g1 g8f6 d1e2 f8e7 f1d1 e6e5 d3c4 c8d7 b1c3 e8g8 b2a3 b8c6 a3d6 e7d6 d1d6 d8e7 d6d1 d7g4 c3d5 e7d8 e2e3 a8c8 d5b4 d8a5 b4d5 f6d5 c4d5 g4f3 e3f3 c6d4 f3h5 a5c7 f2f4 c7c5 g1h1 e5f4 h5g4 d4b5 g4f4 b5c3 d1d3 c5e7 f4f3 c3d5');
INSERT INTO evaluations VALUES(52605,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(52605,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(52605,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(52605,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52605,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(52605,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52605,36,'cp 0','f8g7 f1e1 b8c6 h2h3 g8f6 d2d4 c5d4 f3d4 c6d4 d1d4 a7a6 e4e5 f6d7 c4f7 e8f7 d4f4 f7g8 f4c4 g8f8 c4f4');
INSERT INTO evaluations VALUES(53489,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(53489,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53489,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53489,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(53489,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(53489,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(53489,36,'cp 39','d3d4 g8f6 d4e5 f6e4 b1d2 e4d2 c1d2 d6d5 c4d3 c6c5 c2c3 f8e7 e1g1 b8c6 h2h3 c8e6 b2b4 d5d4 c3d4 e8g8 b4c5 c6d4 d2c3 d4f3 d1f3 a8c8 a1d1 e7c5 d3e4 d8h4 e4b7 c8d8 c3a5 d8d4 d1d4 h4d4 a5c3 d4h4 b7e4 f8d8');
INSERT INTO evaluations VALUES(53525,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(53525,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(53525,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(53525,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(53525,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(53525,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(53525,36,'cp 108','b8c6 b1c3 g8f6 f1b5 f8d6 d2d3 h7h6 e1g1 e8g8 h2h3 a7a6 b5c6 b7c6 f1e1 f8e8 c3e4 f6e4 d3e4 a6a5 c1d2 a5a4 d1e2 d8e7 d2c3 c8a6 e2d2 e7f6 b2b3 a6b7 a1d1 e8e6 c3b2 f6e7 c2c4 c6c5 b2c3 b7c6 e1e3 a8e8 d2c2 e6g6 d1d2 d6c7');
INSERT INTO evaluations VALUES(55194,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(55194,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(55194,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(55194,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(55194,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(55194,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(55194,36,'cp 37','b8c6 d2d4 e7e5 b1c3 f7f5 c4c5 e5e4 f3d2 a7a6 f1e2 g8f6 d1b3 a8b8 e1g1 g7g6 g2g4 f8h6 d2c4 c8e6 c4d6 e8e7 d4d5 c7d6 d5e6 d6c5 g4f5 g6f5 f1f5 d8g8 g1h1 g8e6 f5c5 e6b3 a2b3 h8d8 a1a4 c6d4 c5c7 e7d6 c7f7 d6e6 e2c4 e6f5 c3e4 f5e4 f7f6 h6e3 c1e3 e4e3');
INSERT INTO evaluations VALUES(56214,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(56214,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(56214,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(56214,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(56214,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(56214,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(56214,36,'cp -12','f8d6 g1f3 h7h6 e1g1 g8f6 c2c4 d5c4 b1d2 e8g8 d2c4 f8e8 f1e1 b8c6 c4e5 d6e5 d4e5 d8d1 e2d1 f6d5 a2a3 c8e6 h2h3 a7a5 c1d2 a5a4 a1c1 e8d8 g1h2 d5e7 d2c3 e6d5 f3d4 c6d4 c3d4 e7c6');
INSERT INTO evaluations VALUES(57277,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(57277,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(57277,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(57277,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(57277,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(57277,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(57277,36,'cp 57','f8b4 d1c2 b4c3 b2c3 f7f5 e2e3 g8f6 c4c5 b7b6 c1a3 b6c5 a3c5 d7d6 c5a3 e8g8 g1h3 b8c6 f1c4 g8h8 e1g1 c8d7 c4e2 d8e8 c3c4 h7h6 h3f4 c6e7 a1b1 g7g5 f4d5 e7d5 c4d5 f6d5 e2c4 d7a4 c2c1 d5b6 d4d5 a8d8 c4e2 h8h7 c1c7 d8d7');
INSERT INTO evaluations VALUES(57420,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(57420,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(57420,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(57420,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(57420,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(57420,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(57420,36,'cp 79','d7d5 d1b3 a7a5 b3d5 d8d5 e4d5 e5d4 g1f3 d4c3 b1c3 f8d6 f3d4 g8e7 d4b5 b8d7 b5d6 c7d6 c1e3 e7f5 e1d2 f5e3 a1e1 e8g8 e1e3 d7b6 f1b5 c8f5 h1e1 f8c8 e3e7 a8b8 f2f3 c8c5 e7e8 b8e8 e1e8 g8f7 e8b8 f5c8 d2d3 b6d5 b5c4 c8e6 b8b7 d5c7 b2b3 e6c4 b3c4 f7e7');
INSERT INTO evaluations VALUES(1753,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(1753,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(1753,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(1753,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(1753,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(1753,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(1753,36,'cp 41','b1c3 f8c5 d1f3 c8e6 c1e3 c5b6 e1c1 f6f3 g2f3 g8f6 h1g1 g7g6 c3e2 b6e3 f2e3 a7a5 a2a4 e8e7 e2d4 h8g8 h2h4 h7h6 d4e6 f7e6 e4e5 f6d7 f3f4 g6g5 h4g5 h6g5 g1h1 g8h8 f1d3 g5f4 e3f4 d7c5 d1g1 c5d3 c2d3 a8g8 h1h8 g8g1 c1c2 g1g2 c2c3 g2e2');
INSERT INTO evaluations VALUES(1752,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(1752,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(1752,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(1752,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(1752,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(1752,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(1752,36,'cp 14','f8c5 d1f3 c8e6 c1e3 c5b4 e1c1 b4c3 b2c3 f6c3 e3d4 c3f3 g2f3 f7f6 f3f4 g8e7 h1g1 h8g8 a2a4 f6f5 f2f3 a8d8 g1g7 g8g7 d4g7 d8d1 c1d1 e7c8 f1d3 c8d6 g7e5 e8d7 d1d2 c6c5 d2e3 b7b6 h2h4 d7e7 e5c3 a7a5 h4h5 e6d7 e4e5 d6b7 h5h6 d7e6 c3a1 b7d8');
INSERT INTO evaluations VALUES(2543,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(2543,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(2543,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(2543,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(2543,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(2543,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(2543,36,'cp 109','c4b3 c6d4 e3d4 b7b5 h2h4 h7h5 e1c1 a6a5 g2g4 e7e5 d4e3 a5a4 b3d5 b5b4 d5a8 b4c3 d2c3 h5g4 e3g5 g4f3 c3f3 c8g4 f3g4 f6g4 g5d8 f8d8 a8c6 g4f2 c6a4 f7f5 e4f5 g6f5 a4c6 f2h1 d1h1 g7h6 c1d1 g8g7 b2b4 f5f4 a2a4 d6d5 a4a5 f4f3 a5a6 h6e3');
INSERT INTO evaluations VALUES(2551,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(2551,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(2551,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(2551,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(2551,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(2551,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(2551,36,'cp 78','c6a5 h2h4 h7h5 e1c1 b7b5 c1b1 a5b3 a2b3 b5b4 c3a4 c8d7 g2g4 d7a4 b3a4 a8c8 g4h5 f6h5 h1g1 d8c7 g1g5 g7f6 g5h5 g6h5 e3h6 e7e6 d2g2 g8h7 h6f8 c8f8 f3f4 f8g8 g2f3 h7h6 f4f5 f6d4 d1d4 g8g1 d4d1 g1d1 f3d1 c7c5 f5e6 f7e6');
INSERT INTO evaluations VALUES(2679,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(2679,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(2679,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2679,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2679,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2679,36,'cp 0','d7d6 a4c6 b7c6 h2h3 f8e8 c1e3 c5e3 f2e3 a6a5 d1e1 a5a4 b2b3 a4b3 a2b3 c8e6 e1f2 a8a1 f1a1 d8b8 c3e2 h7h6 g1h2 c6c5 f2e1 g8h7 e2g3 g7g6 g3e2 h7g7 a1a6 e6c8 a6a1');
INSERT INTO evaluations VALUES(2690,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(2690,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(2690,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(2690,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(2690,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(2690,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(2690,36,'cp 178','e4e5 f6e8 c3d4 c5e7 h2h3 d7d5 b1c3 c8f5 a4b3 f5e6 c3e2 d8d7 b3a4 e6f5 e2f4 f5e4 c1e3 b7b5 a4c2 e4c2 d1c2 c6d8 a2a4 b5a4 a1a4 c7c6 f1a1 e8c7 h3h4 d8e6 f4e6 d7e6 b2b3 e6d7 h4h5 c7b5 a4a6 a8a6 a1a6');
INSERT INTO evaluations VALUES(3984,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(3984,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(3984,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(3984,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3984,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(3984,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(3984,36,'cp 13','f6d5 g5e7 d8e7 a1c1 d5c3 c1c3 c7c5 f1b5 c5d4 d1d4 d7b6 e1g1 f8d8 d4e4 b6d5 c3c2 a7a6 b5e2 c8d7 f3d4 d7b5 d4b5 a6b5 e2b5 d5b4 c2c4 b4a2 b5a4 a8b8 e4e5 a2b4 f1d1 b8c8 c4c8 d8c8 a4d7 e7c5 e5c5');
INSERT INTO evaluations VALUES(3983,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(3983,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(3983,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(3983,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(3983,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3983,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3983,36,'cp 19','g5e7 d8e7 a1c1 d5c3 c1c3 c7c6 d1c2 e6e5 f1e2 e5d4 f3d4 d7b6 e2f3 f8d8 c2e4 e7e4 f3e4 a7a5 e1e2 g7g6 b2b3 g8f8 e4f3 b6d5 c3c4 d5b6 c4c5 b6d7 c5c3 d7e5 h1c1 f8e7 f3e4 f7f5 e4c2 e5f7 h2h4 f7d6 f2f3 a5a4 b3a4');
INSERT INTO evaluations VALUES(3972,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(3972,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(3972,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(3972,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(3972,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(3972,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(3972,36,'cp 14','d8e7 a1c1 d5c3 c1c3 c7c6 d1c2 e6e5 f1e2 e5d4 f3d4 f8d8 e1g1 d7b6 f1d1 b6d5 c3c4 d5b4 d4f5 d8d1 e2d1 e7f6 f5h6 f6h6 c4b4 h6d6 b4d4 d6e7 c2d2 c8e6 h2h3 h7h6 b2b3 c6c5 d4d6 b7b6 d1e2 a8c8 e2a6 c8b8 a6c4 e6c4');
INSERT INTO evaluations VALUES(4063,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(4063,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(4063,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(4063,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(4063,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(4063,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(4063,36,'cp 32','f4e3 f6g4 e3g5 h7h6 g5h4 g4f6 f1e2 e7e5 d1c2 g6g5 h4g3 g5g4 d4e5 f6e4 c2e4 g4f3 g2f3 d7e5 g3e5 g7e5 f3f4 e5c3 b2c3 d8f6 h1g1 g8h8 e1d2 c8f5 e4d4 f6d4 c3d4 c7c6 e2d3 f5d3 d2d3 a8b8 a1b1 b7b5 c4b5 c6b5 d3e4 f8c8 e4d5 b5b4');
INSERT INTO evaluations VALUES(5813,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(5813,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(5813,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(5813,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(5813,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(5813,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(5813,36,'cp 64','g8f6 e2e3 a7a5 a2a3 b4c3 d2c3 e8g8 f1d3 b7b6 d1e2 c8a6 a1c1 a6c4 d3c4 d5c4 e2c4 c6e7 e1g1 a8c8 c4e2 d8d5 f1d1 f6e4 h2h3 f8d8 f3e5 e7g6 e5g6 h7g6 c3e1 e4d6 a3a4 d8d7 e2c2 c7c5 d4c5 c8c5 d1d5 c5c2 c1c2 e6d5');
INSERT INTO evaluations VALUES(6185,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(6185,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(6185,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(6185,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(6185,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(6185,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(6185,36,'cp 208','c1g5 f7f6 g5e3 d7d6 f1c4 c8g4 d4e5 d6e5 h2h3 d8d1 a1d1 g4f3 g2f3 a7a6 c4e6 g8e7 e6d7 e8f7 d7c6 b7c6 e1e2 h8d8 c3a4 g7f8 a4c5 e7g8 c5b7 d8d1 h1d1 f7e8 b7a5 g8e7 b2b3 g6g5 d1h1 e7g6 a5c6 f8d6 h1d1 h7h5 a2a4 e8d7 c6a5 a8b8 c2c3 d7c8 b3b4');
INSERT INTO evaluations VALUES(8445,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(8445,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8445,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8445,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8445,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8445,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8445,36,'cp 8','f1e1 c6a5 c4d5 c7c6 d5f7 e8f7 b2b4 c8e6 b4a5 d8a5 c1e3 h8e8 b1d2 f7g8 a2a4 e7f8 d1b1 a8b8 c2c4 e6f7 d3d4 e5d4 e3d4 f6d7 b1c2 b8d8 a1b1 b7b6 d4c3 a5c5 a4a5 d7e5 a5b6 a7b6 f3e5 d6e5 d2f3 c5c4 b1b6');
INSERT INTO evaluations VALUES(11657,30,'cp 331',NULL);
INSERT INTO evaluations VALUES(11657,31,'cp 330',NULL);
INSERT INTO evaluations VALUES(11657,32,'cp 321',NULL);
INSERT INTO evaluations VALUES(11657,33,'cp 343',NULL);
INSERT INTO evaluations VALUES(11657,34,'cp 330',NULL);
INSERT INTO evaluations VALUES(11657,35,'cp 335',NULL);
INSERT INTO evaluations VALUES(11657,36,'cp 347','e5c6 d8d7 c6a5 d6d5 e4d5 f8b4 a5c6 e8g8 c6b4 d7e7 c3e2 e7b4 e1g1 f6d5 d2d3 c8g4 h2h3 g4e2 d1e2 f8e8 e2d1 a8d8 a2a3 b4d4 f1e1 e8e1 d1e1 h7h6 a1b1 d4a4 e1d1 c7c5 b2b3 a4a5 d1d2 a5c7 c2c4 d5e7 b3b4 c7b6 d2e3 e7f5 e3c5 b6g6 c1f4');
INSERT INTO evaluations VALUES(11608,30,'cp 337',NULL);
INSERT INTO evaluations VALUES(11608,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(11608,32,'cp 325',NULL);
INSERT INTO evaluations VALUES(11608,33,'cp 318',NULL);
INSERT INTO evaluations VALUES(11608,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(11608,35,'cp 323',NULL);
INSERT INTO evaluations VALUES(11608,36,'cp 313','d8d7 c6a5 d6d5 e4d5 f8b4 a5c6 e8g8 c6b4 d7e7 c3e2 e7b4 e1g1 f6d5 b2b3 c8g4 c2c4 d5f4 f2f3 f4e2 d1e2 g4f5 c1b2 f7f6 b2c3 b4b6 d2d4 a6a5 f1d1 f8e8 e2f2 a5a4 b3b4 b6c6 f2f1 c6e6 d1e1 e6b6 a1d1 f5e6 g1h1 c7c6 f1d3 e6f7 d4d5 e8e1 d1e1 c6d5 c4c5 b6c6 h1g1');
INSERT INTO evaluations VALUES(674,30,'cp 322',NULL);
INSERT INTO evaluations VALUES(674,31,'cp 326',NULL);
INSERT INTO evaluations VALUES(674,32,'cp 340',NULL);
INSERT INTO evaluations VALUES(674,33,'cp 346',NULL);
INSERT INTO evaluations VALUES(674,34,'cp 321',NULL);
INSERT INTO evaluations VALUES(674,35,'cp 349',NULL);
INSERT INTO evaluations VALUES(674,36,'cp 350','c6a5 d6d5 e4d5 f8b4 d1e2 e8f8 a5c6 c8b7 e1g1 b4c3 b2c3 b7c6 d5c6 d7c6 c1b2 h7h5 c3c4 a8e8 e2d3 h5h4 h2h3 h8h6 f1e1 f8g8 e1e8 f6e8 d3d5 c6d5 c4d5 c7c6 d5c6 h6c6 a1c1 c6c4 g1f1 c4b4 c1b1 b4c4 c2c3 e8c7 d2d3 c4a4 a2a3 c7e6 c3c4');
INSERT INTO evaluations VALUES(14305,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14305,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14305,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14305,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14305,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14305,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14305,36,'cp 0','c8f5 c1f4 f6h5 f4d2 h5f6 d2f4');
INSERT INTO evaluations VALUES(24868,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24868,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24868,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24868,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24868,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24868,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24868,36,'cp 0','c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(15872,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(15872,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(15872,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(15872,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(15872,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(15872,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(15872,36,'cp -83','c8e6 e1g1 d8d7 b1c3 e8c8 c4d5 e6d5 c3d5 d7d5 d1g4 c8b8 c1e3 h7h5 g4f5 f8e7 e4c3 d5g8 f5g6 a7a6 a1e1 e7d6 a2a3 c6e7 g6e4 g7g5 e4c4 f6f5 c4g8 h8g8 f2f3 h5h4 h2h3 e7c6 e1e2 c6d4 e3d4 e5d4 c3d5 d6a3 d5f6 g8g6 b2a3 g6f6 e2e5 b8c8 f1e1 d8d6 e5e7 f6g6 e1e5');
INSERT INTO evaluations VALUES(16705,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(16705,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(16705,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(16705,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(16705,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(16705,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(16705,36,'cp -39','d7d5 e4d5 c6b4 d3e2 d8e7 e2e7 f8e7 f1b5 c8d7 b5d7 f6d7 b1a3 b4d5 a3c4 d7b6 c4b6 a7b6 e1g1 b6b5 a2a3 c7c6 f1e1 f7f6 c1e3 e8f7 h2h4 h8d8 g2g3 d8d7 a1d1 a8d8 d1d3 c6c5 e1d1 h7h5 g1g2 e7f8 d3d2 g7g6 c2c3 d5e3 f2e3 d7d2 f3d2');
INSERT INTO evaluations VALUES(18355,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(18355,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(18355,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(18355,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(18355,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(18355,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(18355,36,'cp 145','d2d4 e7e6 d4d5 c6e5 g1f3 f8d6 c3b5 d8e7 d1d4 f7f6 c1e3 a7a6 b5d6 e7d6 c2c4 a6a5 d4d2 e5f3 g2f3 d6b4 a1c1 b4d2 e1d2 d7d6 f3f4 g8e7 h2h4 h7h5 f1d3 e8f7 b2b3 h8g8 h1g1 g7g6 e4e5 e6d5 c4d5 e7d5 d3c4 c8e6 c4d5 e6d5 c1c7 f7e6 e5d6 e6d6');
INSERT INTO evaluations VALUES(18571,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(18571,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(18571,32,'cp 190',NULL);
INSERT INTO evaluations VALUES(18571,33,'cp 172',NULL);
INSERT INTO evaluations VALUES(18571,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(18571,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(18571,36,'cp 171','e4d5 g8f6 d5c6 b7c6 g1f3 e7e6 c1f4 c8b7 f1e2 f8e7 f3e5 e8g8 e1g1 a7a5 e2f3 f6d5 e5d7 d8d7 c3d5 c6d5 d1d2 f8c8 f1e1 a5a4 c2c3 b7c6 g2g3 d7b7 h2h4 a8a6 a1b1 h7h6 d2c2 a6b6 f3d1 c6b5 c2d2 b7a6 d1c2 b5c4 a2a3');
INSERT INTO evaluations VALUES(18867,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(18867,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(18867,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(18867,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(18867,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(18867,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(18867,36,'cp -28','d5d1 g8f6 f1d3 d7d5 e4d5 f6d5 g1f3 d5b4 d3e2 d8d1 e2d1 e6e5 a2a3 b4d5 c2c4 d5e7 b2b4 c8e6 b1d2 f7f6 e1g1 e8c8 f1e1 e7f5 c4c5 f8e7 d2e4 h7h5 c1b2 g7g5 b4b5 c6d4 f3d4 f5d4 b5b6 c8b8 a1c1 b8a8 b6a7 a8a7');
INSERT INTO evaluations VALUES(22075,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(22075,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(22075,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(22075,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(22075,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(22075,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(22075,36,'cp 88','c1e3 f8e7 d1d2 a7a6 d4d5 c6e5 d5e6 f7e6 e1c1 g8f6 f2f4 e5g4 g1f3 e8g8 e3g1 b7b5 f1d3 b5b4 c3e2 c7c5 h2h3 g4h6 c2c4 b4c3 e2c3 h6f7 e4e5 f6d5 c3d5 e6d5 d2c2 g8h8 d3h7 d8a5 a2a3 d5d4 c2g6 d6e5 f4e5 c8b7');
INSERT INTO evaluations VALUES(23485,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(23485,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(23485,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(23485,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23485,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(23485,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(23485,36,'cp 13','e5g6 f4g3 e7e5 d5e6 d7e6 d1d8 e8d8 f2f3 f8c5 e2e4 g8f6 h2h4 f6h5 g3f2 c5f2 e1f2 h5f6 b1d2 d8e7 d2c4 h7h5 a2a4 a7a5 g1e2 f6d7 g2g3 b7b6 e2d4 d7e5 b2b3 e5c4 f1c4 g6e5 c4e2 f7f6 h1d1 c8d7');
INSERT INTO evaluations VALUES(24600,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(24600,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24600,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24600,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(24600,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24600,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(24600,36,'cp 41','f3e5 f5d7 c1f4 e7e6 e1g1 f8b4 a2a4 e8g8 f1e1 c6e7 b5d3 e7g6 f4g5 d7e8 e1e3 a8c8 e5g4 b4e7 g4f6 e7f6 g5f6 d8f6 c3e2 e8d7 c2c3 f8e8 d1d2 e6e5 d3c2 e5d4 d2d4 f6d4 e3e8 c8e8 e2d4 g6e5 c2b3 e5d3 b3d5 d3b2 d5b7');
INSERT INTO evaluations VALUES(24608,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(24608,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(24608,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(24608,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(24608,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(24608,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24608,36,'cp 37','f5d7 e1g1 e7e6 f1e1 c6e5 d4e5 d7b5 c3b5 f6d7 b2b3 a7a6 b5d6 f8d6 e5d6 e8g8 c1b2 a8c8 a1c1 c8c6 c2c4 d5c4 c1c4 c6c4 b3c4 d8a5 d1a1 g7g6 h2h3 a5c5 a2a4 c5c4 a4a5 f8d8 b2a3 c4d5 a1c3 d5c6 c3b4 c6b5 b4b5 a6b5 e1c1 d7f6 c1c7 f6e8 c7b7');
INSERT INTO evaluations VALUES(26797,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(26797,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(26797,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(26797,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(26797,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(26797,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(26797,36,'cp 71','c4d5 e6d5 b1c3 g8f6 a1c1 f8e7 f1e2 e8g8 g1f3 f6e4 e1g1 e4c3 c1c3 e7d6 f4d6 d8d6 d1b3 c6d8 a2a4 c7c6 a4a5 f5g4 b3a3 d6a3 c3a3 a8b8 h2h3 g4f3 e2f3 d8e6 h3h4 b8e8 f3e2 g7g6 a5a6');
INSERT INTO evaluations VALUES(32965,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(32965,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(32965,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(32965,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(32965,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(32965,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(32965,36,'cp 12','e3d2 g8f6 f1d3 e8g8 e1g1 b4e7 d1e2 d7d6 g1h1 e6e5 d4b3 c8e6 f2f4 e5f4 d2f4 b8c6 h2h3 f6d7 a1d1 e7f6 d3c4 f6e5 f4e5 d7e5 c4d5 a8e8 e2d2 h7h6 a2a3 c7b6 d5e6 f7e6 f1f8 e8f8 d2d6 f8f6 d6c5 b6c5 b3c5 f6f2 c5b7 f2c2 b7c5');
INSERT INTO evaluations VALUES(32967,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32967,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32967,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32967,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(32967,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(32967,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32967,36,'cp 4','g8f6 f1d3 e8g8 e1g1 b4e7 g1h1 b7b5 f2f4 c8b7 e4e5 f6e8 d1e2 b8c6 d4c6 b7c6 a2a4 b5b4 c3e4 a6a5 c2c3 f7f5 e5f6 e8f6 e4f6 f8f6 f4f5 b4c3 b2c3 a8f8 d2g5 f6f7 g5e7 f7e7 f5e6 f8f1 a1f1 d7e6 d3b5 h7h6 c3c4 c6b7 h2h3 c7g3 e2f2 g3h3 h1g1');
INSERT INTO evaluations VALUES(35007,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(35007,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(35007,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(35007,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(35007,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(35007,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(35007,36,'cp -13','b8c6 d2d3 c6d4 f3d1 b7b5 c4d5 c7c6 c2c3 c6d5 c3d4 e5d4 g1f3 d5e4 d3e4 f8g7 e1g1 g8e7 f1e1 f6b6 c1f4 e8g8 b1d2 f8e8 d2b3 e7c6 e4e5 c8b7 d1d2 a8c8 a1d1 a7a6 h2h3 b7a8 b3d4 c6d4 d2d4 b6d4 f3d4 d7d6 e5e6 f7e6 d4e6 g7b2 f4d6');
INSERT INTO evaluations VALUES(35927,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(35927,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(35927,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(35927,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(35927,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(35927,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(35927,36,'cp 63','d4d5 e7e6 f3d2 h7h6 g5h4 d8b6 d1b3 b6b3 d2b3 e6d5 c4d5 f8e8 b3d2 f6e4 c3e4 f7f5 e1c1 f5e4 h4g3 b8d7 d2c4 g7f8 c4d6 f8d6 g3d6 d7e5 f1b5 c8d7 b5d7 e5d7 h1e1 d7e5 d6e5 e8e5 f2f4 e5e7 c1d2 e7d7 d2e3 a8d8 e3e4 d7e7 e4f3 e7e1 d1e1 d8d5 e1e7 b7b5 f3e3 d5d1 e7a7 c5c4 e3e4 d1d2');
INSERT INTO evaluations VALUES(35968,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(35968,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(35968,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(35968,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(35968,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(35968,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(35968,36,'cp 50','e7e6 f3d2 h7h6 g5h4 d8b6 d1b3 b6b3 d2b3 f8e8 b3d2 a7a6 a2a4 f6h5 f1e2 h5f4 h4g3 g7e5 e1g1 b8d7 d5e6 f4e6 f2f4 e5d4 g1h1 d4g7 c3d5 a8b8 e2g4 e6d4 f4f5 d7e5 g4h3 d4e2 g3e5 g7e5 d2f3 g6f5 a1e1 e2f4 f3e5 d6e5 e4f5 f4d5 c4d5');
INSERT INTO evaluations VALUES(36772,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36772,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36772,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36772,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36772,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36772,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36772,36,'cp 0','a7a5 a2a4 b7b6 f1e1 c8b7 c1f4 e7e6 h2h4 h7h6 b1d2 b8d7 f3e5 d7e5 d4e5 f6d7 d2f3 d8e7 d1d2 g8h7 d2c2 h7g8 a1d1 a8d8 c2d2 g8h7 d2c2');
INSERT INTO evaluations VALUES(37213,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(37213,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(37213,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(37213,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(37213,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(37213,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(37213,36,'cp -37','h7h6 g5h4 c7c6 a2a3 a7a5 c4b3 b7b5 g1e2 c5e7 e1g1 d7d6 f2f3 b8d7 h4f2 d8c7 g1h1 d7c5 b3a2 c5e6 e2g3 b5b4 c3a4 a8b8 g3f5 e7d8 a3b4 a5b4 f5e3 f6h5 c2c3 d8g5 d3d4 b4c3 b2c3 c8a6 a2c4 a6c4 e3c4');
INSERT INTO evaluations VALUES(37143,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(37143,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(37143,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(37143,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(37143,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(37143,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(37143,36,'cp -12','g5h4 c7c6 g1e2 b7b5 c4b3 d7d6 a2a3 b8d7 e1g1 a7a5 g1h1 c5a7 f2f3 d7c5 b3a2 c5e6 d1e1 a7b6 c3d1 c8d7 c2c3 a5a4 d1f2 g7g5 h4g3 c6c5 a1d1 d8e7 a2e6 f7e6 h2h3 g8h7 f2g4 f6g4 f3g4 f8f1 e1f1 a8f8 g3f2 c5c4');
INSERT INTO evaluations VALUES(42071,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(42071,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(42071,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(42071,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(42071,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(42071,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(42071,36,'cp 31','d5c4 e2e3 b7b5 a2a4 c7c6 a4b5 c6b5 b2b3 e8g8 b3c4 b5c4 f1c4 c8b7 e1g1 a7a5 d1e2 b8d7 f1c1 d8b6 a1b1 b6a7 f3e5 b7e4 b1a1 d7e5 d4e5 f6d5 c3d4 a7b7 f2f3 e4g6 c4d5 b7d5 c1c5 d5b7 a1a5 b7b1 e2f1 a8a5 c5a5 b1f1 g1f1 f8c8 f1f2 c8c2 f2g3');
INSERT INTO evaluations VALUES(43380,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(43380,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(43380,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(43380,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(43380,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(43380,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(43380,36,'cp -12','b7b5 c4a2 e7e6 d2d4 c5d4 f3d4 f8e7 d1d3 d8c7 e1g1 e8g8 d3g3 b8c6 d4c6 c7c6 c1h6 f6e8 f1d1 a6a5 a2b3 c8b7 h6e3 b5b4 c3a4 b4a3 a1a3 e8f6 f2f3 d6d5 e4d5 f6d5 b3d5 e6d5 e3d4 f7f6 a3a1 c6c2 a4b6 a8a6 g1h1');
INSERT INTO evaluations VALUES(43739,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43739,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(43739,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(43739,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(43739,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(43739,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(43739,36,'cp -23','c4a2 e7e6 d2d4 c5d4 f3d4 f8e7 d1d3 e8g8 d3g3 c8d7 e1g1 b8c6 f1d1 c6d4 d1d4 a8c8 c1g5 c8c5 g5d2 d8b8 d4d3 f8c8 d2e3 c5h5 e3d4 h5h6 a1d1 d7c6 f2f3 h6g6 g3f2 b8c7 g1h1 c6b7 f2d2 f6d7');
INSERT INTO evaluations VALUES(47798,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(47798,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(47798,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(47798,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(47798,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(47798,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(47798,36,'cp -6','f6e4 c4f7 e8f7 c3e4 d6d5 e4g3 b8c6 d2d3 h7h6 g1e2 f8e7 e1g1 h8f8 d3d4 f7g8 c1e3 f8f7 f1e1 e5d4 e2d4 c6d4 e3d4 c8d7 c2c3 c7c5 d4e5 d7c6 d1d3 d8d7 a1d1 a8e8 f2f4 e7h4 g1h2 h4f6 d3g6 f6e5 e1e5 f7f4 d1d5 c6d5 e5e8 f4f8 e8f8 g8f8 g6d3 f8f7 b2b4 c5b4 c3b4');
INSERT INTO evaluations VALUES(49963,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(49963,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(49963,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(49963,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(49963,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(49963,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(49963,36,'cp 66','c2c4 f8e7 b1c3 f6e4 f1g2 d7d5 c1f4 c7c6 d1c2 b8d7 a1d1 e8g8 e1g1 a7a5 f3e1 g7g5 f4c1 e4c3 c2c3 e7f6 b2b3 d8e7 e1f3 b7b5 c3e3 g5g4 f3e5 c8b7 e5d3 f8e8 c4b5 c6b5 c1d2 b5b4 d3f4 d7f8 h2h4 g4h3 g2h3 b7a6 d1c1 f8g6 f4h5 f6h8 c1c6');
INSERT INTO evaluations VALUES(55128,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(55128,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(55128,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(55128,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(55128,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(55128,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(55128,36,'cp -24','c7c5 e2e4 g8f6 g1f3 b8c6 g2g3 e7e5 f1g2 f8e7 e1g1 e8g8 f1e1 d5e4 d2e4 f6e4 d3e4 d8a5 d1d2 a5d2 c1d2 f8d8 c2c3 f7f6 d2e3 c8g4 h2h3 g4e6 g2f1 c6a5 b2b3 a7a6 f3d2 b7b5 f2f4 e6f7 f4e5 f6e5 d2f3 a5c6');
INSERT INTO evaluations VALUES(55502,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(55502,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(55502,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(55502,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(55502,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(55502,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(55502,36,'cp 83','c1e3 f8e7 g1f3 g8f6 f1d3 c7c6 c2c4 b7b5 d5c6 b5c4 d3c4 b8c6 b1c3 e8g8 e1g1 a8b8 d1e2 f6g4 a1c1 g4e3 f2e3 c6a5 c4d5 c8e6 b2b3 d8b6 c3a4 b6b4 e2a6 e6d5 e4d5 b8b7 a2a3 b4b3 a6a5 e5e4 f3d2 b3e3 g1h1 f8b8 d2c4 e3d3');
INSERT INTO evaluations VALUES(55758,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(55758,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(55758,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(55758,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(55758,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(55758,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(55758,36,'cp -106','c7c6 b5d3 f8d6 e4d5 e6d5 g1f3 g8f6 e1g1 e8g8 f1e1 c6c5 h2h3 b8c6 c2c3 d8b6 b1a3 h7h6 d3f1 a7a6 a3c2 c8f5 c2e3 f5e6 d2d4 c5d4 f3d4 c6d4 d1d4 d6c5 d4f4 a8e8 f1d3 f6e4 d3e4 d5e4 e3f5 c5f2 f4f2 b6f2 g1f2 e6f5');
INSERT INTO evaluations VALUES(53532,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(53532,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(53532,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(53532,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(53532,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(53532,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(53532,36,'cp -83','b5d3 f8d6 e4d5 e6d5 g1f3 g8f6 d3e2 e8g8 e1g1 f8e8 d2d4 b8d7 h2h3 h7h6 f1e1 d7b6 a2a4 c8f5 a4a5 b6c4 e2c4 e8e1 d1e1 d5c4 b1d2 f5c2 d2c4 c2e4 f3e5 d6c7 e1b4 a8b8 c1d2 f6d5 b4b3 f7f6 e5d3 b7b5 a5b6 a7b6 a1e1');
INSERT INTO evaluations VALUES(57568,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(57568,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(57568,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(57568,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(57568,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(57568,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(57568,36,'cp 55','e7e5 c2c4 c7c6 e2e4 c6c5 b1c3 d7d6 c3d5 g8e7 d2d3 e7d5 c4d5 f8e7 f3h4 e8g8 f1g2 g7g6 c1h6 f8f7 e1g1 b8d7 f2f4 e5f4 g3f4 c5c4 d3c4 f6f5 h4f3 f5e4 f3g5 e7g5 f4g5 f7f1 d1f1 d7e5 g2e4 c8g4 a1e1 d8b6 f1f2 e5c4 e4g2 b6f2 g1f2 c4e5');
INSERT INTO evaluations VALUES(57448,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(57448,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(57448,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(57448,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(57448,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(57448,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(57448,36,'cp 66','c2c4 c7c6 e2e4 c6c5 a2a3 b8c6 b1c3 d7d6 b2b4 g8e7 d2d3 g7g6 f1g2 f8g7 e1g1 e8g8 a1b1 g8h8 c3d5 b7b6 h2h3 a8b8 c1d2 h7h6 b4c5 d6c5 a3a4 d8d6 d2c3 c8d7 b1a1 h8h7 f3d2 c6d4 c3d4 c5d4 a4a5 e7d5 e4d5');
INSERT INTO evaluations VALUES(3000,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(3000,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(3000,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(3000,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(3000,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(3000,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(3000,36,'cp 120','d4d5 f8e8 c1f4 d7f8 h2h3 f6h5 f4d2 e7e6 f1e1 a7a6 a2a4 b7b6 d2g5 d8c7 d1d2 e6e5 f3h4 h5f4 g5f4 e5f4 d2f4 c7e7 h4f3 h7h6 f4d2 a8b8 f3h2 f8h7 e2g4 c8g4 h3g4 g7d4 d2h6');
INSERT INTO evaluations VALUES(4028,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4028,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4028,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4028,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4028,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4028,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4028,36,'cp 0','e7e5 f4g5 f8e8 g5h4 h7h6 d1c2 d8e7 c3d5 e7d8');
INSERT INTO evaluations VALUES(4072,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4072,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(4072,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(4072,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4072,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4072,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4072,36,'cp 0','f1e2 f8e8 f4g3 e7e5 d4e5 d6e5 e1g1 f6h5 g3h4 g7f6 h4f6 h5f6 d1c2 d7c5 e3e4 c8g4 f1e1 a7a5 e2f1 a5a4 a1d1 d8e7 h2h3 g4f3 d2f3 g8g7 b2b4 a4b3 a2b3 a8d8 d1a1 d8a8');
INSERT INTO evaluations VALUES(4073,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(4073,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4073,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4073,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4073,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4073,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4073,36,'cp 0','f8e8 e1g1 e7e5 f4g3 f6h5 g3h4 g7f6 h4f6 h5f6 d4e5 d6e5 e3e4 b7b6 d1c2 c8b7 f1e1 d8e7 b2b4 a7a5 a2a3 e8d8 a1d1 g8g7 e2f1 f6e8 d2c4 b7c6 c2b2 c6a4 d1c1 a4c6');
INSERT INTO evaluations VALUES(4929,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(4929,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(4929,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(4929,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(4929,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(4929,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(4929,36,'cp 141','f3e5 d8g5 e5g4 d7d5 h2h3 g5h4 e1g1 c8g4 d1g4 h4g4 h3g4 e8c8 e4d5 c6d5 f1d1 h7h5 g4g5 c5b4 a2a3 b4c3 d2c3 g8e7 d1e1 h8e8 g1f1 e7f5 c1f4 c7c6 a1d1 e8e1 d1e1 c8d7 c3c4 d5c4 e1e4 c4c3 b2c3 g7g6 e4a4 d8e8 a4a6 e8e4 g2g3 h5h4');
INSERT INTO evaluations VALUES(4928,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(4928,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(4928,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(4928,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(4928,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(4928,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(4928,36,'cp 127','d8g5 e5g4 d7d5 h2h3 g5h4 e1g1 c8g4 d1g4 h4g4 h3g4 g8f6 e4d5 e8c8 d5c6 f6g4 c3d1 h8e8 c2c3 d8d3 b2b4 c5a7 d1e3 g4e5 f1e1 e8e6 g1f1 a7e3 e1e3 d3e3 d2e3 e6c6 c1d2 e5c4 d2e1 c6g6 a1d1 h7h5 d1d4 g6g4 d4g4 h5g4');
INSERT INTO evaluations VALUES(6527,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(6527,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6527,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(6527,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(6527,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(6527,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(6527,36,'cp 19','c2c4 g8f6 f1g2 e8g8 e1g1 d7d5 c4d5 f6d5 b1c3 h7h6 c3d5 d8d5 c1d2 f8d8 d1c2 d5d6 a2a3 a7a5 a1c1 c5b6 d2e3 c8g4 e3c5 b6c5 c2c5 a5a4 f1e1 a8a5 c5c2 g4f3 g2f3 c6d4 c2c7 d6c7 c1c7 d4f3 e2f3 a5b5 g1g2 b5b2');
INSERT INTO evaluations VALUES(6538,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(6538,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(6538,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(6538,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(6538,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(6538,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(6538,36,'cp 65','c4f7 e8f7 d1d5 f7f8 d5c5 d8e7 c5e3 g8f6 e4e5 f6d5 e3e4 d5f6 e4e2 f6g4 c1g5 e7e6 g5f4 d7d6 e5d6 c7d6 b1a3 e6e2 e1e2 b7b6 h1d1 c8a6 c3c4 c6a5 d1d4 f8f7 f4d6 h8e8 e2f1 g4f6 f3e5 f7g8 f2f4 a8c8 a1e1 e8d8 e1d1 a5c6 e5c6 c8c6 d6b4 d8d4 d1d4 c6c8 h2h3 c8e8 a3b5 e8e4 d4e4 f6e4 f1g1 h7h6 f4f5 e4f6 g1h2');
INSERT INTO evaluations VALUES(6152,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(6152,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(6152,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(6152,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(6152,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(6152,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(6152,36,'cp 51','e8f7 d1d5 f7f8 d5c5 d8e7 c5e7 g8e7 c3c4 b7b6 c1b2 c6b4 e1d2 e7c6 h1c1 c6a5 d2e3 d7d6 b1d2 c8d7 h2h4 h7h6 f3d4 f8f7 f2f3 h8e8 d4b5 b4a6 b5c3 a8c8 g2g4 a6c5 a1b1 c5a4 h4h5 a4c3 b2c3 a5c6 a2a3');
INSERT INTO evaluations VALUES(4661,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(4661,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(4661,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(4661,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(4661,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(4661,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(4661,36,'cp 76','d1d5 f7f8 d5c5 d8e7 c5e3 g8f6 e4e5 f6d5 e3e4 d5f6 e4e2 f6g4 c1g5 e7e6 b1a3 g4e5 f3e5 e6e5 e2e5 c6e5 a3b5 c7c6 b5c7 a8b8 e1c1 f8f7 h1e1 h7h6 e1e5 h6g5 d1e1 d7d5 e5e7 f7g6 c7e8 c8d7 e8d6 d7f5 e1e5 h8f8 f2f3 b8d8 d6f5 f8f5');
INSERT INTO evaluations VALUES(4664,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(4664,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(4664,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(4664,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(4664,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(4664,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(4664,36,'cp 74','f7f8 d5c5 d8e7 c5e3 g8f6 e4e5 f6d5 e3e4 d5f6 e4e2 f6g4 c1g5 e7e6 b1a3 g4e5 f3e5 e6e5 e2e5 c6e5 a3b5 c7c6 b5c7 a8b8 e1c1 f8f7 h1e1 h7h6 e1e5 h6g5 d1e1 f7g6 e5e7 d7d5 c7e8 c8d7 e8d6 d7f5 e1e5 b8f8 h2h3 f5h3 g2h3 h8h3 d6e8 h3c3 c1b1 f8f2 e7g7 g6h6 e5g5 f2f1 b1b2 c3f3 g5g6 h6h5 g6g1 f1g1 g7g1 f3f4 b2a1 h5h4 e8c7 f4f2 c7e6 f2d2');
INSERT INTO evaluations VALUES(7470,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(7470,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(7470,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(7470,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(7470,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(7470,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(7470,36,'cp 137','d2d4 c5b6 b3d5 e4f6 c1g5 h7h6 g5f6 d8f6 f3e5 e8g8 b1d2 a8b8 d2e4 f6f4 e5c6 d7c6 d5c6 c8g4 d1b1 g4f5 f1e1 f8d8 b1d1 f5e4 c6e4 c7c5 g2g3 f4f6 d4d5 g7g6 g1g2 h6h5 d1d2 b6c7 h2h4 c7d6 a2a3 d8e8 d2g5 g8g7 a1d1 f6e5 f2f4 e5e7 g5e7 e8e7 e4f3 b8e8');
INSERT INTO evaluations VALUES(7469,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(7469,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(7469,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(7469,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(7469,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(7469,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(7469,36,'cp 130','c5e7 a2a4 c8b7 d1e2 d7d5 a4b5 a6b5 a1a8 d8a8 b3d5 e4f6 d5c6 b7c6 d4e5 f6h5 e5e6 c6f3 e2b5 a8c6 b5b8 e7d8 g2f3 c6f3 b8b5 c7c6 e6f7 e8f7 b5c4 f3d5 c4d5 c6d5 f1d1 f7e6 b1a3 d8b6 a3c2 h8a8 c2b4 h5f6 g1f1 b6c5 b4d3 c5d6 c1f4 a8a2 d1e1 e6d7 f4d6 d7d6');
INSERT INTO evaluations VALUES(11642,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(11642,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(11642,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(11642,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(11642,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(11642,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(11642,36,'cp 30','a2a4 c8b7 e1g1 f6e4 d2d4 f8e7 a4b5 a6b5 a1a8 b7a8 d1e2 e4f6 d4e5 d6e5 e2b5 e8g8 b1d2 d8b8 b5b8 f8b8 f1e1 f6d7 b3c2 f7f6 d2c4 c6a7 e1d1 d7f8 c1e3 a7b5 c4a5 g7g6 f3d2 f8d7 d2c4 b5d6 b2b4 d6c4 a5c4');
INSERT INTO evaluations VALUES(12837,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(12837,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(12837,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(12837,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(12837,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(12837,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(12837,36,'cp -19','d7d5 e4d5 f6d5 d3d4 c8f5 e1g1 e7e6 a2a3 d5f6 c2c3 f8d6 d4c5 d6c5 d1d8 a8d8 b1d2 f5d3 e2d3 d8d3 a3a4 a7a6 f1e1 e8g8 g1f1 h7h6 d2c4 f8c8 c1e3 c5e3 e1e3 d3d8 a4a5 g8f8 f3d2 g7g5 g2g3 d8d5 b2b4 c8d8 f1e1 c6e7');
INSERT INTO evaluations VALUES(15128,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(15128,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(15128,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(15128,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(15128,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(15128,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(15128,36,'cp 0','e4d5 f6d5 c1b2 d5c3 b2c3 f8d6 f1b5 c8d7 d2d4 e5d4 f3d4 c6d4 b5d7 d8d7 d1d4 e8c8 e1c1 d6a3 c1b1 h8e8 d4d7 d8d7 d1d7 c8d7 h1e1 e8e1 c3e1 d7e6 c2c3 e6f5 b1c2 a3d6 h2h3 h7h5 c2d3 c7c6 e1d2 d6f4 d2e1');
INSERT INTO evaluations VALUES(14997,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(14997,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(14997,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(14997,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(14997,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14997,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14997,36,'cp 0','f6d5 c1b2 d5c3 b2c3 f8d6 f1b5 c8d7 d2d4 e5d4 f3d4 c6d4 b5d7 d8d7 d1d4 e8c8 e1c1 d6a3 c1b1 h8e8 d4d7 d8d7 d1d7 c8d7 h1e1 e8e1 c3e1 d7e6 c2c3 a3d6 h2h3 e6f5 b1c2 h7h5 c2d3 c7c6 c3c4 g7g5 e1d2 d6f4 d2e1 f4h2 e1d2');
INSERT INTO evaluations VALUES(15340,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(15340,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(15340,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(15340,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(15340,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(15340,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(15340,36,'cp -4','c1b2 d5c3 b2c3 f8d6 f1b5 c8d7 d2d4 e5d4 f3d4 c6d4 b5d7 d8d7 d1d4 e8c8 e1c1 d6a3 c1b1 b7b6 h1e1 f7f6 d4d7 d8d7 d1d7 c8d7 h2h3 h7h5 e1e2 h8g8 f2f3 a3d6 b1c1 f6f5 c1d2 g7g5 g2g4 h5g4 h3g4 g8f8 c3e5 f5g4 f3g4 f8f1 a2a4 d6e5 e2e5 f1f2 d2d3');
INSERT INTO evaluations VALUES(20657,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(20657,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(20657,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(20657,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(20657,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(20657,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(20657,36,'cp -42','e5d6 g8f6 d2d4 c5d4 f3d4 d8d6 c4b5 c8d7 b5c6 d7c6 d4c6 d6c6 e1g1 f8d6 d1e2 a8c8 c2c3 e8g8 b1d2 f8d8 f1e1 a7a6 d2f3 h7h6 c1e3 e6e5 a1d1 d6c7 d1d8 c8d8 e1d1 d8d1 e2d1 c6b5 d1b3 b5e2 e3d2 f6e4 d2e3 b7b5');
INSERT INTO evaluations VALUES(21386,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(21386,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(21386,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(21386,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(21386,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(21386,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(21386,36,'cp 51','g8f6 g1f3 f8e7 b2b4 e8g8 b1c3 c6b8 c4c5 c7c6 f1d3 a7a5 c1b2 b7b6 c3a4 b6c5 b4c5 c8a6 d3a6 a8a6 b2c3 b8d7 e1g1 d8c7 a1b1 a6a8 d1c2 f8b8 b1b8 a8b8 a4b6 d7b6 c3a5 f6d7 c5b6 d7b6 f1b1 h7h6 f3e5 c6c5');
INSERT INTO evaluations VALUES(21782,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(21782,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21782,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(21782,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(21782,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(21782,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(21782,36,'cp 79','g1f3 g8e7 b2b4 e7g6 c1b2 a7a5 b4b5 c6e5 f3e5 g6e5 b2d4 e5c4 e2e3 c8e6 d1c2 c4d6 b1d2 d8g5 c2b2 d6f5 d2f3 g5g6 d4e5 f8e7 f1e2 e8g8 a1c1 f7f6 e5g3 c7c6 e1g1 c6b5 e2b5 f8c8 b5d3 g6f7 g3f4 c8c1 f1c1 a8d8 f3d4 f5d4 e3d4 e7d6 f4d6');
INSERT INTO evaluations VALUES(22461,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(22461,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(22461,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(22461,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(22461,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(22461,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(22461,36,'cp -47','f8g7 d2d3 c6a5 b1c3 d7d6 c1e3 g8f6 a2a4 a5c4 d3c4 f6g4 e3d2 f7f5 f2f3 g4f6 d2g5 h7h6 g5h4 d8e7 g1e2 e7f7 h4f6 g7f6 b2b3 c8e6 a4a5 e8g8 e1g1 f6g7 g1h1 a7a6 c3d5 c7c6 d5b6 a8d8 e2c3 f5f4 d1e2 g8h8');
INSERT INTO evaluations VALUES(25277,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(25277,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(25277,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(25277,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25277,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(25277,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(25277,36,'cp 57','f8e7 d2d4 e8g8 h2h3 g4f3 d1f3 e5d4 b5c6 b7c6 c3d4 d6d5 e4e5 f6e4 b1c3 f7f5 c3e2 d8d7 f3d3 c6c5 f2f3 e7h4 e1f1 c5d4 b2b4 c7c5 b4c5 e4c5 d3d4 h4e7 c1e3 c5e6 d4d3 f5f4 e3f2 a8d8 a1d1 f8e8 g1h2 e7c5 f2c5 e6c5 d3c3 c5e6 e2d4 d7c7 c3c7 e6c7 f1e1 c7e6 h3h4 g8f7 a2a4 e6d4 d1d4');
INSERT INTO evaluations VALUES(24047,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(24047,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(24047,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(24047,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(24047,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(24047,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(24047,36,'cp 58','h2h3 g4f3 d1f3 e8g8 d2d4 e5d4 b5c6 b7c6 c3d4 d6d5 e4e5 f6e4 b1c3 f7f5 c3e2 d8d7 f3d3 c6c5 f2f3 e7h4 e1f1 c5d4 b2b4 c7c5 b4c5 e4c5 d3d4 h4e7 f1d1 c5e6 d4d3 e7c5 c1e3 c5e3 d3e3 a8d8 e2d4 f5f4 e3d3 e6d4 d3d4 d7e6 a2a4 e6b6 a4a5 b6d4 d1d4 f8e8');
INSERT INTO evaluations VALUES(26534,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(26534,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(26534,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(26534,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(26534,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(26534,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(26534,36,'cp 12','e7e6 e2e3 f8d6 f4g3 g8f6 g1f3 e8g8 f1e2 f8e8 e1g1 d6g3 h2g3 e6e5 d4e5 c6e5 d1b3 e5f3 e2f3 a8b8 f1d1 d8e7 c3c4 d5c4 d2c4 h7h5 a1c1 f5g4 b3c3 f6e4 c3c2 g4f5 c2b3 g7g6 f3e4 f5e4 f2f3 e4f5 e3e4 f5e6 b3a3 e7a3 c4a3 c7c6');
INSERT INTO evaluations VALUES(27249,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(27249,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(27249,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(27249,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(27249,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(27249,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(27249,36,'cp 96','c2c4 a7a6 b5a4 g7g6 b1c3 f8g7 c1e3 c6a5 a4d7 d8d7 d1d3 g8f6 e1g1 e8g8 h2h3 b7b5 b2b3 b5c4 b3c4 c7c5 d4d5 e6d5 c4d5 f8e8 a1b1 d7c7 f3d2 f6d7 f1c1 e8b8 b1b8 c7b8 c1b1 b8c7 d3c2 a8b8');
INSERT INTO evaluations VALUES(29393,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(29393,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(29393,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(29393,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(29393,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(29393,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(29393,36,'cp -19','b8c6 h2h3 g4h5 c1f4 h5f3 e2f3 f8d6 c3e2 d6f4 e2f4 d8b6 c2c3 b6b2 c3c4 b2d4 c4d5 d4d1 f1d1 c6e5 d5e6 e5f3 g2f3 e8e7 a1b1 b7b6 a2a4 a8b8 e6f7 e7f7 f4d3 h8e8 b1b3 f7g6 f3f4 e8e4 d3e5 g6f5');
INSERT INTO evaluations VALUES(31037,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(31037,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(31037,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31037,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31037,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31037,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31037,36,'cp 0','e7e6 g1f3 b8d7 e1g1 g8f6 f1e1 f8e7 c4b3 e8g8 f3h4 f8e8 h4g6 h7g6 c2c3 c6c5 g3e4 f6e4 e1e4 d7f6 e4h4 f6d5 h4g4 d5f6');
INSERT INTO evaluations VALUES(31420,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31420,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31420,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(31420,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(31420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31420,36,'cp 0','c2c3 g8f6 g1f3 b8d7 d1e2 d8c7 f3h4 d7b6 c4b3 c6c5 d4c5 f8c5 h4g6 h7g6 c1g5 e8c8 a2a4 a7a6 h2h4 d8d7 a4a5 b6d5 e2f3 c7e5 f3e2');
INSERT INTO evaluations VALUES(32237,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(32237,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(32237,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(32237,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(32237,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(32237,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(32237,36,'cp -40','c4b3 f8e7 d2d4 g8f6 g1f3 e8g8 c1f4 f8e8 e1g1 a7a5 c2c3 e6g4 e2d1 c7c5 b3a4 b8c6 b1d2 c5d4 c3d4 e7b4 a4c6 b7c6 d1c2 d8b6 f4e3 g4h5 a1c1 a8c8 f3e5 b6b5 d2f3 h5f3 e5f3 b4d6 a2a4 b5a6 f3e5 d6e5 d4e5 e8e5');
INSERT INTO evaluations VALUES(32248,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(32248,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(32248,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(32248,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(32248,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(32248,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(32248,36,'cp -49','f8e7 g1f3 a7a5 c2c3 g8f6 e1g1 e8g8 d2d4 f8e8 c1f4 e6g4 e2d1 c7c5 d4c5 e7c5 b1d2 b8c6 h2h3 g4h5 b3c2 c5a7 c2a4 f6d7 a4b5 h7h6 f4g3 d8b6 a2a4 d7c5 f1e1 c5e4 d2e4 d5e4 d1d5 h5g6 f3d2 e4e3 e1e3 e8e3');
INSERT INTO evaluations VALUES(32902,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(32902,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(32902,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(32902,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(32902,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(32902,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(32902,36,'cp 50','b8c6 f1b5 e8c8 b5c6 b7c6 d1d3 c6c5 c1e3 g4f3 g2f3 e7e6 e1c1 g8e7 d3c4 e7d5 d4c5 d5e3 d1d8 c8d8 f2e3 f8c5 c4f4 d8c8 c1b1 h8f8 h1d1 c5e7 f4e4 a5b6 c3d5 e6d5 e4e7 f8d8 d1d3 b6e6 e7c5 a7a6 a2a4 d8d6 a4a5 g7g6 d3b3 c8d7 b3b7');
INSERT INTO evaluations VALUES(34058,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(34058,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(34058,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(34058,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(34058,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(34058,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(34058,36,'cp 185','d2d4 c7c6 e4e5 d7d6 c1f4 d6e5 d4e5 f6d5 c3d5 c6d5 d1d5 b8c6 f1b5 e7c5 e1c1 c5d5 d1d5 a7a6 b5c6 b7c6 d5d3 c8e6 h1d1 e6d5 f3h4 h7h6 h4f5 d5g2 d3g3 g2e4 f5d6 f8d6 e5d6 e8d7 g3g7 e4d5 b2b3 a8e8 d1g1 e8e2 f4g3 h8e8');
INSERT INTO evaluations VALUES(36979,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(36979,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(36979,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(36979,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(36979,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(36979,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(36979,36,'cp 139','e3e4 d6e7 d1e2 b8c6 e4e5 f6d5 h2h4 f7f5 e5f6 d5f6 c4e6 c8e6 e2e6 g8h8 e1g1 e7d6 c1d2 f8e8 e6h3 d8c8 h3c8 a8c8 d2g5 f6g4 a1e1 h7h6 g5c1 g4f6 g2g3 h8g8 g1g2 e8e1 f1e1 c8e8 e1e8 f6e8 g2f1 g8f7 f1e2 e8f6 a2a3 f7e6 e2d3 c6e7 c3b5 a7a6 b5c3');
INSERT INTO evaluations VALUES(38357,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(38357,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38357,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38357,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(38357,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38357,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38357,36,'cp -5','c7c5 e1g1 b8c6 d2d3 e7e6 a2a4 g8e7 c2c3 e8g8 c4b3 a8b8 c1e3 a7a6 b1d2 b7b5 e3f2 b5a4 a1a4 c6a5 e4e5 d6e5 f4e5 d8c7 b3c2 c8b7 d1c1 a5c6 d2c4 e7f5 f1e1 f8d8 c1d2 g7h6 f3g5 c6e7 a4a5 e7c6');
INSERT INTO evaluations VALUES(37964,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37964,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37964,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(37964,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(37964,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37964,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(37964,36,'cp 0','d2d3 e7e6 e1g1 g8e7 a2a4 b8c6 c2c3 e8g8 c4b3 d6d5 e4e5 f7f6 e5f6 f8f6 d1e2 e7f5 b1d2 f6f8 f3e5 c6e5 f4e5 c8d7 b3c2 d7c6 d2f3 d5d4 f3g5 d8d7 c2b3 a8e8 g2g4 f5h4 f1f8 g7f8 g5e4 c6e4 e2e4 f8g7 c1f4 e8f8');
INSERT INTO evaluations VALUES(38358,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(38358,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38358,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(38358,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(38358,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(38358,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(38358,36,'cp 20','g8f6 g1f3 e8g8 c4b3 c8g4 h2h3 g4f3 d1f3 b8c6 f3d1 e7e5 d4e5 d6e5 d1d8 f8d8 c1g5 c6d4 e1c1 d8d7 a2a4 h7h6 g5e3 a7a6 d1d2 b7b5 f2f3 c7c5 c3d5 a8d8 b3a2 c5c4 d5f6 g7f6 e3h6 d4e6 h1d1 d7d2 d1d2 e6f4 d2d8 f6d8 h6f4 e5f4 a4b5 a6b5 e4e5 d8c7');
INSERT INTO evaluations VALUES(40983,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(40983,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(40983,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40983,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(40983,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40983,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(40983,36,'cp 12','g1f3 e8g8 e1g1 f6h5 f4e5 c7c5 c2c4 d5c4 d3c4 c5d4 f3d4 h5f6 d1f3 d8a5 e5g3 f6h5 d4b3 a5g5 b1c3 h5g3 h2g3 b8c6 c4b5 c8d7 f1d1 c6e5 f3e4 d7b5 c3b5 a8b8 b5d4 e5g4 a1c1 g5h5 d4f3 g4f6 e4h4 h5h4 g3h4 f6d5 b3c5 h7h6 c5d3 b8c8');
INSERT INTO evaluations VALUES(41480,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(41480,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(41480,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(41480,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(41480,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(41480,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(41480,36,'cp -23','g1f3 e8g8 e1g1 c7c5 c2c3 d8c7 b1d2 c5d4 c3d4 b8c6 d2b3 f6e4 c1d2 c8f5 a1c1 c7b6 f3e5 c6e5 f4e5 f8c8 d1e2 g7h6 d3e4 c8c1 f1c1 d5e4 h2h3 a8c8 c1c8 f5c8 e2c4 c8e6 c4c5 e6b3 c5b6 a7b6 a2b3 f7f6 g1f1 f6f5');
INSERT INTO evaluations VALUES(41533,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(41533,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(41533,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(41533,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(41533,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(41533,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(41533,36,'cp 74','g1f3 c7c5 d4d5 e8g8 e1g1 e7e6 d5e6 c8e6 c4e6 f7e6 f3g5 d8e7 d1b3 f8e8 c1d2 b8c6 c3e4 f6d5 h2h3 b7b6 g5f3 a8d8 f1d1 d5f6 e4g3 f6d5 a1c1 c6e5 f3e5 g7e5 g3e4 d8d7 d2e1 e8d8 c1c2 d5b4 c2d2 d7d2 d1d2 d8d2 e1d2 e7d7 d2c3');
INSERT INTO evaluations VALUES(41539,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(41539,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(41539,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(41539,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(41539,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(41539,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(41539,36,'cp 158','e8g8 e4e5 f6d7 e5e6 d7b6 e6f7 g8h8 c4b3 c8g4 h2h4 b8c6 h4h5 g6h5 d1d3 c6d4 f3d4 g7d4 f2f3 d4f2 e1e2 d8d3 e2d3 g4f5 c3e4 a8d8 d3e2 f2d4 c1g5 d4b2 a1b1 b2a3 g5h6 f5g6 b1d1 d8a8 e4g5 a3d6 g2g4 h5g4 f3g4 c7c6 b3e6 b6d5 e2f3 a8e8 f3f2 d6c7 h6f8 e8f8 e6d5 c6d5 d1d5');
INSERT INTO evaluations VALUES(36803,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(36803,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(36803,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(36803,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(36803,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(36803,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(36803,36,'cp 175','e4e5 f6d7 e5e6 d7b6 e6f7 g8h8 c4b3 c8g4 h2h4 b8c6 c3e2 d8d6 c1f4 d6b4 f4d2 b4d6 d1c2 g4f5 c2c1 e7e6 f3g5 g7f6 e1g1 f6g5 d2g5 f8f7 f1e1 c6a5 b3d1 a5c6 g5f4 d6b4 d1b3 b6d5 f4g3 a8f8 b3d5 e6d5 a2a3 b4c4 c1g5 h8g8 a1c1 c4b3 h4h5 b3b2 e2f4 c6d4 h5g6 f5g6 f4g6 h7g6');
INSERT INTO evaluations VALUES(46359,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46359,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46359,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(46359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46359,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(46359,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46359,36,'cp 13','a2a3 f8d6 f1d3 e8g8 e1g1 h7h6 h2h3 a7a6 b1c3 b7b5 c1d2 b8c6 e3e4 d5e4 c3e4 f6e4 d3e4 c8b7 f1e1 d8d7 d4d5 e6d5 e4d5 d6h2 g1h2 d7d6 h2g1 d6d5 d2c3 d5d1 a1d1 f8d8 f3d4 c6d4 d1d4 d8d4 c3d4 a8d8 d4c3 f7f6 f2f3 d8d3 e1e8 g8h7');
INSERT INTO evaluations VALUES(46451,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(46451,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(46451,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(46451,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(46451,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(46451,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(46451,36,'cp -46','d1b3 c5c4 b3d1 b8c6 b2b3 f8b4 a2a3 b4c3 d2c3 c4b3 d1b3 e8g8 c1g5 h7h6 g5f6 d8f6 g1f3 c6a5 b3b4 b7b6 e1g1 c8a6 f3d4 f8d8 f1e1 a8c8 e2e4 d5e4 e1e4 a6b7 e4f4 f6e5 g2b7 a5b7 f4f3 b7a5 f3e3 e5g5 a1e1 a5c4 e3e7 c4d2 g1g2 g5f6');
INSERT INTO evaluations VALUES(46659,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(46659,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(46659,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(46659,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46659,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(46659,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(46659,36,'cp 9','f8g7 g1f3 e8g8 d1d2 c8f5 f4h6 e7e5 h6g7 g8g7 d4e5 d6e5 d2d8 f8d8 f3e5 b8a6 e5d3 f5d3 e2d3 d8d7 e1c1 a8d8 f1e2 a6c5 d3d4 c5e4 h1f1 c7c6 f2f3 e4g5 f3f4 g5e6 d4d5 c6d5 f4f5 e6g5 c4d5 f6d5 c3d5 d7d5 d1d5 d8d5 f5g6 g7g6 f1d1');
INSERT INTO evaluations VALUES(47637,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(47637,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(47637,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(47637,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(47637,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(47637,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(47637,36,'cp 37','d3d4 b8d7 f1c4 c7c6 a2a4 f8e7 e1g1 e8g8 h2h3 f8e8 a4a5 e5d4 d1d4 d7c5 f1d1 c8e6 c1e3 e7f8 c4f1 d6d5 e4d5 f6d5 c3d5 e6d5 f3d2 b7b5 c2c4 b5c4 f1c4 d5c4 d4c4 d8d5 c4d5 c6d5 e3c5 f8c5 d2b3 a8c8 d1d5');
INSERT INTO evaluations VALUES(48340,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(48340,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(48340,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(48340,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(48340,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(48340,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(48340,36,'cp -23','b8d7 g1f3 b7b6 e1g1 f8d6 c3c4 d5c4 d3c4 c8b7 b2b3 e8g8 c1b2 d8e7 d1e2 e6e5 f3e5 d7e5 d4e5 d6e5 b1c3 f6e4 f1c1 f8d8 c3e4 e5b2 e2b2 e7e4 c4f1 c7c5 c1d1 h7h6 b2e2 e4e7 d1d8 a8d8 a1d1 b7e4 f2f3 e4c6 d1d8 e7d8 e2c2');
INSERT INTO evaluations VALUES(48672,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48672,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48672,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48672,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(48672,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48672,36,'cp 0','c7c5 d4c5 f8c5 g1f3 d8e7 d1e2 e7e2 e1e2 e8g8 h1e1 f8e8 e2f1 e8e1 f1e1 b8c6 b1d2 c8d7 d2b3 c5b6 c2c3 a8e8 e1f1 f6e4 b3d4 e4c5 d3e2 c5e4');
INSERT INTO evaluations VALUES(49564,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(49564,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(49564,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(49564,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49564,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(49564,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49564,36,'cp 0','f8b4 b1d2 e8g8 a2a3 b4d2 c1d2 f8e8 g1f3 c7c6 d2c3 d7d5 c4d5 c6d5 c3e5 b8c6 e5f4 d8b6 f1g2 b6b2 f4d2 c6d4 f3d4 b2d4 e1g1 c8g4 d2e3 e8e3 f2e3 d4e3 f1f2 a8c8 d1a4 g4e2 a1e1 c8c1 e1c1 e3c1 g2f1 e2d3 a4a7 h7h6 a7b7 d3f1 f2f1 c1a3 b7b8 g8h7 b8e5 a3b4 e5f5 h7g8 f5f3 b4a5 f3d3 a5a2 f1f2 a2a1 g1g2');
INSERT INTO evaluations VALUES(50935,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(50935,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(50935,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(50935,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(50935,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(50935,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(50935,36,'cp 102','e2e4 c7c5 d4c5 d6c5 d1d8 e8d8 g1f3 c8d7 c1f4 d8e8 e1c1 g8e7 f3e5 f7f6 e5d7 b8d7 c3a4 a8c8 g2g3 e7c6 f4e3 c8d8 d1d2 c6d4 e3d4 c5d4 d2d4 d7c5 d4d8 e8d8 a4c3 a6a5 f1e2 f8d6 f2f4 d8e7 h1d1 g7g5');
INSERT INTO evaluations VALUES(51897,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(51897,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(51897,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(51897,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(51897,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(51897,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(51897,36,'cp -13','f8c5 g1f3 g8f6 c2c4 e8g8 a2a3 d7d5 b2b4 c5d6 c1b2 a7a5 b4b5 b8d7 f1e2 d7c5 e1g1 a5a4 b1c3 d8e7 d1c2 e6e5 a1d1 e5e4 f3d2 c8f5 c4d5 f6g4 g2g3 f8c8 e2g4 f5g4 c3e4 g4d1 f1d1 c5e4 c2e4 e7e4');
INSERT INTO evaluations VALUES(37685,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(37685,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(37685,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(37685,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(37685,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(37685,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(37685,36,'cp 0','a2a3 g8f6 g1f3 d7d5 c2c4 e8g8 b2b4 c5d6 c1b2 a7a5 b4b5 b8d7 f1e2 d7c5 e1g1 d8e7 c4d5 e6d5 a3a4 c8e6 b1d2 f8c8 f3d4 e6d7 a1c1 e7e5 d2f3 e5h5 b2a3 f6e4 c1c2 h5g6 c2c1');
INSERT INTO evaluations VALUES(56521,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(56521,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(56521,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(56521,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(56521,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(56521,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(56521,36,'cp 73','f1d3 g8f6 e1g1 f8d6 f1e1 e8g8 b1c3 d6e5 e1e5 b8c6 e5g5 c6d4 g5g3 d4f5 g3f3 f5h4 f3e3 h7h6 d1d2 c7c6 h2h3 f8e8 e3e8 d8e8 d2f4 h4g6 f4g3 e8e5 g3e3 e5e3 c1e3 c8e6 f2f3 a7a6 a1d1 g6e5 d3e2 e5c4 e2c4 e6c4 b2b3');
INSERT INTO evaluations VALUES(3534,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(3534,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(3534,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(3534,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(3534,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(3534,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(3534,36,'cp -50','a2a4 h7h6 g5f6 g7f6 f3e2 g8h8 g1f3 f6g7 a4a5 a7a6 b1d2 f7f5 d2f1 c6e7 c4b3 d8e8 f1e3 f5f4 e3c4 e7c6 c4d2 e8e7 b3d5 c8d7 e1d1 c6d8 d1c2 a8b8 b2b4 c7c6 d5c4 d8e6 c4b3 h8h7 h2h3 c6c5');
INSERT INTO evaluations VALUES(4298,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(4298,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(4298,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(4298,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(4298,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(4298,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(4298,36,'cp -71','h7h6 d2d4 c5b6 g5f3 e5d4 e4e5 d7d5 c4e2 d4d3 e2d3 f6d7 c1f4 d7c5 d3c2 c8g4 b1d2 f8e8 h2h3 g4h5 d1c1 d8d7 f1e1 a8d8 f4g3 c5e6 g1h2 d5d4 c3d4 c6d4 f3d4 e6d4 d2c4 b6c5 c2b1 c5b4 e1e3 h5g6 g3h4 b4e7 h4e7 e8e7 c1c3 c7c5 b1g6 f7g6');
INSERT INTO evaluations VALUES(4148,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(4148,31,'cp -98',NULL);
INSERT INTO evaluations VALUES(4148,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(4148,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(4148,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(4148,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(4148,36,'cp -94','d2d4 c5b6 g5f3 e5d4 e4e5 d7d5 c4e2 d4d3 e2d3 f6d7 c1f4 d7c5 a2a4 a7a5 d3c2 c8g4 b1d2 d5d4 c3d4 c5e6 f4e3 c6d4 c2e4 d8e7 e3d4 e6d4 h2h3 g4h5 g2g4 a8d8 g4h5 d4f3 d1f3 d8d2 f3f5 e7g5 f5g5 h6g5 e4b7 d2b2 b7a6 f8d8 g1g2 g8h7 a1b1 b2b1 f1b1 d8d2 a6c4');
INSERT INTO evaluations VALUES(6419,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(6419,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(6419,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(6419,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(6419,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(6419,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(6419,36,'cp -55','g8f6 d2d3 d7d5 e4d5 f6d5 e1g1 e8g8 c1g5 d8d7 c2c3 d5b6 b3b4 b6c4 d3c4 h7h6 g5h4 e5e4 d1d7 c8d7 f3d2 g7g5 b4c5 g5h4 d2e4 f7f5 e4d2 c6e5 b1a3 d7c6 a3c2 h4h3 f1e1 a8e8 e1e3 c6g2 a1e1 e8d8 c2d4 e5c6 d2f3 f5f4 e3e6 g2f3 d4f3 d8d3 f3d4 f4f3 d4c6 b7c6 e6h6');
INSERT INTO evaluations VALUES(10083,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(10083,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(10083,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(10083,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(10083,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(10083,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(10083,36,'cp -32','d2d3 e8g8 e1g1 d7d5 e4d5 f6d5 c1g5 d8d6 b1d2 d5c3 d2e4 c3d1 e4d6 c7d6 f1d1 h7h6 g5e3 c5b4 d3d4 b4c3 a1c1 e5e4 f3d2 f8e8 h2h3 a7a6 d2f1 c6e7 f1g3 d6d5 c4f1 c8e6 g3e2 c3a5 c2c3 f7f6 a2a4 a5c7 e2f4 e6f7');
INSERT INTO evaluations VALUES(10082,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(10082,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(10082,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(10082,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(10082,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(10082,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(10082,36,'cp -46','e8g8 e1g1 d7d5 e4d5 f6d5 c1g5 d8d7 c2c3 d5b6 b3b4 b6c4 d3c4 h7h6 g5h4 e5e4 d1d7 c8d7 f3d2 g7g5 b4c5 g5h4 d2e4 f7f5 e4d2 c6e5 f2f4 e5d3 d2b3 a8e8 b1d2 f8f7 b3d4 d3c5 d2f3 h4h3 f3e5 f7g7 g2g3 c5e4 f1e1 e4c3 c4c5 a7a6 e5d7 e8e1 a1e1 g7d7 d4f5 c3a2 f5h6 g8f8');
INSERT INTO evaluations VALUES(8491,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(8491,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(8491,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(8491,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(8491,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(8491,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(8491,36,'cp 61','a2a3 f6h5 f4g3 h5g3 h2g3 g7g6 c2c4 e8g8 b1c3 e7f6 d1c2 d5c4 d3c4 e6e5 d4d5 c6e7 c3e4 f6g7 a1d1 c8f5 f3g5 f5e4 g5e4 e7f5 g3g4 f5d6 c4d3 d6e4 d3e4 d8g5 c2e2 f8d8 e2f3 g5e7 e1e2 h7h6 d1c1 d8d7 c1c2 a8d8');
INSERT INTO evaluations VALUES(8838,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(8838,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(8838,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(8838,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8838,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8838,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8838,36,'cp 9','c5b6 a2a4 a7a5 d2d4 e8g8 c1e3 e5d4 c3d4 d6d5 e4d5 c6b4 b1c3 f6d5 c3d5 b4d5 d1d3 c7c6 c4b3 d8f6 b3c2 f6g6 d3g6 f7g6 f3e5 g6g5 c2b3 f8e8 b3d5 c6d5 e1c1 b6a7 g2g4 a7b8 c1c5 b8e5 d4e5 e8e5 c5b5 e5e4 b5d5 c8e6 d5d4 e4d4 e3d4 e6f7 d4e3 a8e8 a1c1 f7d5');
INSERT INTO evaluations VALUES(10279,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(10279,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(10279,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(10279,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(10279,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(10279,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(10279,36,'cp 127','e4e5 d7d5 e5f6 d5c4 f6g7 d8d1 e1d1 h8g8 c3d5 e8d8 c1g5 c6e7 h1e1 c8e6 d5f4 d8d7 g5f6 e7f5 f4e6 f7e6 a1c1 c5e7 f3e5 d7e8 f6e7 f5e7 c1c4 g8g7 g2g3 e7f5 d1e2 a8d8 g3g4 d8d5 f2f4 f5d6 c4c2 c7c6 e2f3 g7g8 c2c3 g8f8 c3d3 d6f7 d3d5 c6d5 f4f5 f7e5 e1e5');
INSERT INTO evaluations VALUES(10048,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(10048,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(10048,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(10048,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(10048,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(10048,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(10048,36,'cp 138','d7d5 e5f6 d5c4 f6g7 d8d1 e1d1 h8g8 c3d5 e8d8 c1g5 c6e7 h1e1 c8e6 d5f4 d8d7 d1c2 g8g7 a1d1 c5d6 f3e5 d7e8 h2h4 e7c6 f4e6 f7e6 e5c6 b7c6 e1e6 e8d7 e6f6 a8e8 g2g3 c6c5 g5e3 d7c6 h4h5 c6b5 f6f3 e8e4 h5h6 g7e7 d1d5 b5c6 d5f5 d6e5 f5f7 e5d4 f7e7 e4e7 g3g4 d4e3 f3e3');
INSERT INTO evaluations VALUES(13266,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13266,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13266,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13266,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13266,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13266,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13266,36,'cp 2','g1f3 e7e6 d2d3 a7a6 a2a3 d6d5 c4a2 d5e4 c3e4 f8e7 c2c3 f6e4 d3e4 d8d1 e1d1 b7b5 c1e3 e8g8 d1e2 a6a5 h1d1 a5a4 g2g4 c8a6 e2f2 b5b4 c3b4 c5b4 a1c1 b4a3 b2a3 f8d8 f4f5 d8d1 c1d1 e6f5');
INSERT INTO evaluations VALUES(13313,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(13313,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(13313,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(13313,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(13313,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(13313,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(13313,36,'cp -23','g7g6 h2h3 e7e5 g2g3 f8g7 f1g2 e8g8 a2a4 c8e6 f3g5 e6d7 e1g1 c6d4 g5f3 d8c8 h3h4 d7g4 e3g5 c8d7 g5f6 g7f6 c3d5 f6g7 d5e3 d4f3 g2f3 g4e6 a4a5 a8e8 c2c3 g7h6 d1a4 d7a4 a1a4 h6e3 f2e3');
INSERT INTO evaluations VALUES(14855,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(14855,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(14855,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(14855,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(14855,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(14855,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(14855,36,'cp 41','c2c4 f8d6 c4c5 d6e7 a2a3 a7a5 b1c3 f6e4 f1d3 f7f5 c3e2 e8g8 a1b1 e6e5 f3e5 e4c5 d3c2 c5e4 e1g1 e7d6 f2f3 e4f6 e5d3 d8e7 d1e1 g7g5 e1f2 b7b6 b2b4 c6d8 b4b5 d8e6 a3a4');
INSERT INTO evaluations VALUES(15305,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(15305,31,'cp -131',NULL);
INSERT INTO evaluations VALUES(15305,32,'cp -131',NULL);
INSERT INTO evaluations VALUES(15305,33,'cp -148',NULL);
INSERT INTO evaluations VALUES(15305,34,'cp -155',NULL);
INSERT INTO evaluations VALUES(15305,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(15305,36,'cp -144','g7g6 h5f3 d8g5 c4d5 c6d4 f3f7 e8d8 e1g1 d4c2 d5e4 c2a1 d2d4 g5e7 f7c4 c8e6 c4d3 e7d7 c1e3 d8e8 d4d5 a8d8 f1d1 c7c6 d5e6 d7d3 d1d3 d8d3 e4d3 e8e7 g1f1 e7e6 f1e2 f8b4 b1d2 b7b5 d2f3 a7a5 f3g5 e6f6 g5e4 f6g7 f2f4 h8e8 f4e5 h7h6 e2d1 e8d8 d1e2 d8d7 e3f2 d7d5 e5e6 b4e7 f2b6');
INSERT INTO evaluations VALUES(14522,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(14522,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(14522,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(14522,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(14522,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(14522,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(14522,36,'cp -125','h5f3 d8g5 c4d5 c6d4 f3f7 e8d8 e1g1 d4c2 d5e4 c2a1 d2d4 g5e7 f7c4 c8e6 c4d3 e7d7 c1e3 d8e8 d4d5 a8d8 f1d1 c7c6 d5e6 d7d3 d1d3 d8d3 e4d3 e8e7 b1d2 e7e6 g1f1 f8b4 f1e2 b7b5 d2f3 b4e7 d3e4 h8d8 e4c6 a1c2 e3a7 e6d6 c6b5 d8a8 a7b6 a8a2 f3d2 c2d4 b6d4 e5d4 e2d3 d6c5 b5c4');
INSERT INTO evaluations VALUES(14521,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(14521,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(14521,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(14521,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(14521,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(14521,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(14521,36,'cp -111','d8g5 c4d5 c6d4 f3f7 e8d8 e1g1 d4c2 d5e4 c2a1 d2d4 g5e7 f7c4 c8e6 c4d3 e7d7 c1e3 d8e8 b1d2 e5d4 e3d4 a8d8 d2f3 f8g7 d4g7 d7g7 d3a3 g7e7 a3a7 c7c6 f1a1 e8f7 e4c6 b7c6 f3e5 f7f6 a7e7 f6e7 e5c6 e7d6 c6d8 h8d8 a2a4 e6d5 a4a5 d6c6 f2f3 c6b5 a5a6 b5b6 a6a7 d8a8');
INSERT INTO evaluations VALUES(21158,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21158,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(21158,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21158,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21158,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(21158,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21158,36,'cp -2','e2e3 c8d7 f1c4 e7e5 a4c2 f8d6 b1c3 f7f5 d2d3 g8f6 f3g5 d8e7 g5f7 c6b4 c2e2 h8f8 f7d6 e7d6 a2a3 b4d5 c3d5 f6d5 e2h5 g7g6 h5h4 d5b6 c1d2 b6c4 d3c4 d6f6 h4f6 f8f6 d2c3 f6c6 c3e5 c6c5 e5c3 c5c4 f2f3 e8c8');
INSERT INTO evaluations VALUES(22031,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(22031,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(22031,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(22031,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(22031,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(22031,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(22031,36,'cp 133','b1c3 g8f6 g1f3 d6d5 e4e5 f6e4 f1d3 c8f5 e1g1 e4c3 b2c3 f5d3 d1d3 f8e7 h2h4 d8d7 f1e1 e8g8 h4h5 h7h6 f3h2 d7e6 h2f1 f7f5 c1f4 c6a5 g2g3 e7g5 f1e3 g5f4 g3f4 a5c4 e3c4 d5c4 d3f3 c7c6 a2a4 b7b5 a4b5');
INSERT INTO evaluations VALUES(23161,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(23161,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(23161,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(23161,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(23161,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(23161,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(23161,36,'cp -15','a5c4 d3c4 g8f6 d1e2 c7c6 b1c3 d7d6 c1e3 d8c7 h2h3 f8e7 e1c1 a7a6 g2g4 b7b5 g4g5 f6d7 f3d2 d7c5 e3c5 d6c5 h3h4 c8e6 d1g1 e8c8 c3d1 d8d4 d1e3 h8d8 g1d1 h7h6 c2c3 d4d7 g5h6 g7h6 e3f5 e7f8 f5e3 b5b4 c3b4 c5b4 d2b3 a6a5 d1d7');
INSERT INTO evaluations VALUES(23113,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(23113,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(23113,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(23113,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(23113,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(23113,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(23113,36,'cp 0','d3c4 d7d6 b1c3 f8e7 c1e3 g8f6 d1e2 c7c6 h2h3 e8g8 e1c1 d8a5 g2g4 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(23110,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23110,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23110,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23110,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23110,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23110,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23110,36,'cp -8','g8f6 f3e5 d8e7 e5g4 f6g4 d1g4 d7d5 g4e2 e7e4 c4d5 c8f5 b1c3 e4e2 e1e2 e8c8 c1e3 a7a6 h1d1 f5c2 d1d2 c2f5 e3f4 d8d7 a1c1 g7g5 f4g3 c8b8 c1d1 f8b4 a2a3 b4c3 b2c3 h8e8 e2f1 b8c8 d5d6 f5e6 d6c7 d7d2 d1d2');
INSERT INTO evaluations VALUES(30553,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(30553,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30553,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(30553,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(30553,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(30553,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30553,36,'cp 30','c2c4 d5c4 b1a3 e7e6 a3c4 g8f6 c4e5 b7d5 d2d4 f8b4 c1d2 b4d2 d1d2 b8d7 e1g1 e8g8 f1d1 a8c8 a1c1 c7c5 d4c5 c8c5 c1c5 d7c5 b2b4 c5e4 d2c2 d8c8 d1c1 c8b7 a2a3 a7a5 e5c6 d5c6 c2c6 b7c6 c1c6 a5b4 a3b4 f8a8 c6b6 a8a1 g2f1');
INSERT INTO evaluations VALUES(34116,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(34116,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(34116,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(34116,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(34116,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(34116,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(34116,36,'cp 93','d2d4 g8f6 g2g3 c8g4 h2h3 g4h5 c1f4 d6b4 g3g4 h5g6 d1c1 b4b6 f1g2 e7e6 e1g1 b8c6 c3a4 b6b4 b2b3 f8d6 f4d2 b4a3 c1d1 d6e7 c2c3 f6e4 d2f4 e8g8 f3d2 e4d2 d1d2 e7d6 f4d6 a3d6 a4c5 b7b6 c5b7 d6d7 f1e1 h7h5 a1d1');
INSERT INTO evaluations VALUES(34911,30,'cp 278',NULL);
INSERT INTO evaluations VALUES(34911,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(34911,32,'cp 286',NULL);
INSERT INTO evaluations VALUES(34911,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(34911,34,'cp 289',NULL);
INSERT INTO evaluations VALUES(34911,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(34911,36,'cp 262','b1c3 d5d8 d2d3 b8c6 d3e4 c8g4 e2e3 g8f6 c1d2 d8d7 e1c1 e8c8 f1b5 g4f3 e3f3 d7e6 b5c6 e6c6 f3f5 c6e6 f5e6 f7e6 f2f3 f8b4 a2a3 b4e7 c3e2 e6e5 c1b1 h8g8 d2g5 a7a5 e2c1 d8d1 h1d1 e7d6 g5d2 a5a4 c1d3 b7b6 d2c3');
INSERT INTO evaluations VALUES(34908,30,'cp 257',NULL);
INSERT INTO evaluations VALUES(34908,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(34908,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(34908,33,'cp 269',NULL);
INSERT INTO evaluations VALUES(34908,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(34908,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(34908,36,'cp 256','d5d8 c3e4 f8e7 d2d4 b8d7 e4g3 d7b6 e2d3 g8f6 f1e2 e8g8 e1g1 f8e8 c2c4 e7d6 f1e1 c8e6 b2b3 c7c6 e2d1 b6d7 c1d2 d8c7 f3g5 d7f8 d1c2 e6g4 g5e4 f6e4 e1e4 e8e4 d3e4 g4e6 a1e1 h7h6 e4d3 a8d8 d3e3 b7b5 g1h1 d8e8 e3d3 b5c4 b3c4 d6g3');
INSERT INTO evaluations VALUES(38155,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(38155,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(38155,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(38155,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38155,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(38155,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(38155,36,'cp 45','f1d3 b8c6 e1g1 b4d6 f1e1 g8e7 c3b5 d6b8 b5a3 h7h6 c2c3 e8g8 a3c2 f7f6 c3c4 e6e5 c4d5 e7d5 d4e5 f6e5 b2b3 c8e6 c1a3 f8e8 d3e4 a7a6 a1c1 b8a7 c2e3 d5e3 d1d8 a8d8 f2e3 e6g4 h2h3');
INSERT INTO evaluations VALUES(43584,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(43584,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(43584,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(43584,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(43584,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(43584,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(43584,36,'cp 74','c4d5 f6d5 g1f3 e7e6 e3e4 d5f6 b1c3 c7c5 c1e3 c5d4 f3d4 e6e5 d4b3 b8c6 d1d8 c6d8 f2f3 c8e6 f1e2 d8c6 e1c1 a8c8 c1b1 f8b4 b3c5 c6d4 c5e6 f7e6 e2d3 f6d7 c3e2 b4c5 e3g5 d4e2 d3e2 c5d4 d1c1 e8f7 g5d2');
INSERT INTO evaluations VALUES(46522,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(46522,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(46522,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(46522,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(46522,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(46522,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(46522,36,'cp 113','e2e3 f8e7 f1d3 f6e4 g5e7 d8e7 e1g1 e8g8 a1c1 e4d6 c3e2 b8d7 d1c2 d7f6 f3e5 c8d7 b2b4 d6f7 a2a3 d7e8 c2c3 f6e4 c3b2 e4f6 e2f4 g7g5 f4e2 g8h8 c4d5 f6d5 d3c4 d5b6 c4a2 f7e5 d4e5 e8g6 e2d4');
INSERT INTO evaluations VALUES(46521,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(46521,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(46521,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(46521,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(46521,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(46521,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(46521,36,'cp 116','f8e7 f1d3 f6e4 g5e7 d8e7 e1g1 e8g8 a1c1 b8d7 c3e2 a7a5 d1c2 d7f6 a2a3 c8d7 f3e5 d7e8 c4c5 f6d7 b2b4 a5b4 a3b4 e4f6 f2f3 e8h5 c1a1 d7e5 d4e5 f6d7 c2c3 h7h6 h2h3 g8h8 c3b2 f8b8 a1a8 b8a8 f1a1 a8a1 b2a1 e7h4 e2f4 d7f8');
INSERT INTO evaluations VALUES(40033,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(40033,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(40033,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(40033,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(40033,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(40033,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(40033,36,'cp 115','a1c1 b8d7 f1d3 f6e4 g5e7 d8e7 e1g1 e8g8 c3e2 a7a5 d1c2 e4d6 f3e5 d7f6 a2a3 d6f7 h2h3 g7g6 c4c5 g6g5 f2f3 f6d7 e5f7 f8f7 c1e1 d7f8 e3e4 c8d7 e4f5 e6f5 f3f4 e7f6 c2d2 g5f4 d2f4 f8g6 f4g3 g8h8');
INSERT INTO evaluations VALUES(47048,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(47048,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47048,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(47048,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47048,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47048,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(47048,36,'cp 0','b1c3 c7c5 g1f3 b8c6 f1c4 e7e6 d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 e1g1 a7a6 a2a4 c8d7 d1e2 a8c8 c4d3 c6b4 a4a5 e6e5 d4f3 d7e6 f3g5 e6d7');
INSERT INTO evaluations VALUES(52144,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52144,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(52144,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52144,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(52144,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(52144,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(52144,36,'cp -4','g8f6 g1f3 f6e4 d1b3 e4d6 c4d5 d8b6 b3d1 b8a6 c3d4 c5d4 f3d4 a6b4 d4e2 b4d5 d1d5 f8e7 e1g1 e8g8 b1c3 d6e8 e2f4 e7d8 d5f3 e8f6 b2b3 b6c6 c3d5 f6d5 f4d5 d7d6 f1d1 c8e6 c1e3 a8c8 d1d2 f8e8 a1d1 a7a6 h2h3 b7b5 f3g3 g8h8 d5f4 d8e7');
INSERT INTO evaluations VALUES(55205,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(55205,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(55205,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(55205,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(55205,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(55205,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(55205,36,'cp 17','e3d4 e7e6 g1f3 b8c6 b1c3 g8e7 d4d5 e6d5 d2d4 c8e6 c4c5 g7g6 b2b4 f8g7 b4b5 c6a5 f1e2 e8g8 e1g1 c7c6 c1f4 e6g4 b5c6 b7c6 f3e5 g4e2 c3e2 f8e8 f1e1 a5c4 e2g3 c4e5 f4e5 g7e5 e1e5 d8a5 e5e2 a5c7 d1d2 a8b8');
INSERT INTO evaluations VALUES(55199,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(55199,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(55199,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(55199,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55199,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(55199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55199,36,'cp 10','e7e6 b1c3 g8e7 g1f3 b8c6 d4d5 e6d5 d2d4 c8e6 d1b3 d8d7 b3b7 a8b8 b7a6 b8b6 a6a4 d5c4 f1e2 b6b4 a4d1 g7g6 f3g5 f8g7 g5e6 f7e6 a2a3 b4b8 e2c4 c6d4 e1g1 e8g8 a1b1 e7f5 c3e4 d7c6 c4d3 b8b3 f1e1 c6d5 e4c3 d5b7 d3c4 b3c3 b2c3 b7b1 c3d4 f5d4 c4e6 d4e6 e1e6');
INSERT INTO evaluations VALUES(55467,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(55467,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(55467,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(55467,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(55467,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(55467,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(55467,36,'cp 36','d1d3 g8f6 e2e4 c7c5 e4d5 e6d5 d4c5 f8c5 c1e3 d8e7 e1c1 c5e3 d3e3 e7e3 f2e3 c8e6 f3d4 b8c6 h2h3 e8e7 f1b5 a8c8 h1f1 h8d8 d4f5 e6f5 f1f5 a7a6 b5f1 d5d4 e3d4 c6d4 f5f4 d4c6 d1d8 c8d8 f1e2 b7b5 a2a3 c6e5 b2b3 e5c6 e2f3');
INSERT INTO evaluations VALUES(55703,30,'cp -224',NULL);
INSERT INTO evaluations VALUES(55703,31,'cp -228',NULL);
INSERT INTO evaluations VALUES(55703,32,'cp -209',NULL);
INSERT INTO evaluations VALUES(55703,33,'cp -211',NULL);
INSERT INTO evaluations VALUES(55703,34,'cp -205',NULL);
INSERT INTO evaluations VALUES(55703,35,'cp -232',NULL);
INSERT INTO evaluations VALUES(55703,36,'cp -240','d6e5 d2d3 e5f4 c1f4 f8d6 f4d6 c7d6 d1g4 d8f6 b1c3 b8c6 g1f3 g8h6 g4g5 f6g5 f3g5 c6d4 e1d2 c8d7 a1e1 e8c8 f1e2 d4e2 e1e2 e6e5 e2f2 f7f6 g5e4 d7c6 h1f1 h6f7 g2g4 d6d5 e4c5 f7d6');
INSERT INTO evaluations VALUES(55993,30,'cp -210',NULL);
INSERT INTO evaluations VALUES(55993,31,'cp -199',NULL);
INSERT INTO evaluations VALUES(55993,32,'cp -220',NULL);
INSERT INTO evaluations VALUES(55993,33,'cp -217',NULL);
INSERT INTO evaluations VALUES(55993,34,'cp -237',NULL);
INSERT INTO evaluations VALUES(55993,35,'cp -204',NULL);
INSERT INTO evaluations VALUES(55993,36,'cp -245','d2d3 e5f4 c1f4 f8d6 f4d6 c7d6 b1c3 g8f6 d1d2 e8g8 e1c1 b8d7 g2g4 f6g4 d3d4 f7f5 g1h3 d7b6 h1g1 g8h8 f1e2 d8h4 h3g5 c8d7 e2g4 f5g4 g5e4 d6d5 e4c5 h7h6 d2g2 f8f2 g2g4 h4g4 g1g4 f2h2');
INSERT INTO evaluations VALUES(55713,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(55713,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(55713,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(55713,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(55713,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(55713,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(55713,36,'cp 49','g1f3 d6d5 c4d5 e6d5 e4d5 g8f6 f1c4 d8e7 d1e2 c8g4 e2e7 f8e7 f3e5 b8d7 e5g4 f6g4 b1c3 g4f6 d2d3 d7b6 e1g1 e8c8 f1e1 h8e8 c4b5 f6d7 c1e3 e7b4 e3d4 e8e1 a1e1 d7f6 a2a3 b4c3 d4c3 b6d5 c3e5 d5b6 e5f6 g7f6 g2g3 d8d5 e1e8 d5d8 g1g2 c7c6 e8d8 c8d8 b5c4 d8e7 g2f3 b6c4');
INSERT INTO evaluations VALUES(57550,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(57550,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(57550,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(57550,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(57550,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(57550,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(57550,36,'cp 122','e7e6 d2d4 b8c6 b1c3 f8b4 a2a3 b4c3 b2c3 g8e7 d1h5 e7g6 h2h4 b7b6 h5e2 g6e7 c4d3 e8g8 h4h5 h7h6 g1f3 d7d6 f3h4 e6e5 h4g6 f8e8 c1e3 d6d5 a1d1 c8e6 d3b5 d8d7 d4e5 f6e5 f2f4 e6g4 b5c6 e7c6 d1d5 e5f4 d5d7');
INSERT INTO evaluations VALUES(57802,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(57802,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(57802,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(57802,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(57802,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(57802,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(57802,36,'cp 63','e7e5 d4d5 a7a5 c1e3 f6g4 e3g5 f7f6 g5d2 b8d7 a2a3 f6f5 f3g5 d7c5 e4f5 g6f5 e2g4 f5g4 f2f3 h7h6 g5e4 c5e4 f3e4 c8d7 c3b5 h6h5 d1e2 a5a4 d2e3 e8f8 f1f8 g7f8 a1f1 d7e8 h2h3 e8b5 e2f2 d8d7 c4b5 g4h3 f2f6 f8g7 f6e6 d7e6 d5e6 c7c6 f1d1 c6b5');
INSERT INTO evaluations VALUES(57809,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(57809,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(57809,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(57809,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(57809,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(57809,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(57809,36,'cp 95','d4d5 a7a5 a2a3 b8d7 c1e3 f6g4 e3g5 f7f6 g5d2 f6f5 e4f5 g6f5 f3g5 d7c5 e2g4 f5g4 f2f3 h7h6 g5e4 c5e4 f3e4 c8d7 c3b5 e8f8 f1f8 g7f8 d1e2 a5a4 a1f1 a8c8 e2d1 h6h5 d2e3 c8a8 d1e2 a8c8 e2c2 d8e8 c2a4 e8g6 a4c2 h5h4 b2b3 f8e7 a3a4 h4h3');
INSERT INTO evaluations VALUES(57808,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(57808,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(57808,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(57808,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(57808,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(57808,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(57808,36,'cp 91','a7a5 a2a3 b8a6 b2b4 a5b4 a3b4 c8g4 c1a3 a6b8 f3d2 g4e2 d1e2 c7c6 b4b5 c6c5 a3b2 a8a1 f1a1 b8d7 c3d1 h7h5 b2c3 g7h6 h2h3 f6h7 a1a7 d8b8 a7a3 b8c7 d2f3 h7g5 f3g5 h6g5 c3a5 b7b6 a5d2 g5d2 e2d2');
INSERT INTO evaluations VALUES(2329,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(2329,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(2329,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(2329,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(2329,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(2329,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(2329,36,'cp 86','c1e3 f8b4 b1c3 b4c3 b2c3 b7b6 e1c1 d8e7 c4d5 a8b8 f2f4 c8e6 c3c4 e7d7 h1e1 e6d5 d1d5 d7c6 f4e5 b8e8 e3d4 f6e5 d4e5 e8e7 e5d4 g8f6 d4f6 g7f6 d5f5 e7e6 c1d2 h8d8 d2c3 d8e8 c3d3 e8d8 f5d5 d8e8 e1e2 e6e5 g2g3 e8e6 a2a3 e6e7 d5e5 f6e5 e2f2 e7d7 d3c3');
INSERT INTO evaluations VALUES(3892,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(3892,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(3892,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(3892,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(3892,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(3892,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(3892,36,'cp -113','c2c3 c6a5 c4d5 c7c6 b2b4 c6d5 e4d5 c8g4 f3g3 g4h5 b4c5 d6c5 c3c4 b7b5 e1g1 h5g6 c4b5 h6f5 g3h3 c5c4 b1c3 c4d3 g5e4 f5d4 c1a3 g6e4 c3e4 d4e2 g1h1 d8d5 h3f3 f8d8 a3e7 e2f4 e7d8 a8d8 a1c1 f7f5 e4d2 d5a2 f1d1 a5b3 d2b3 a2b3 c1b1 b3e6 f3e3 e6g6 e3g3');
INSERT INTO evaluations VALUES(5330,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5330,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5330,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5330,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5330,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5330,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5330,36,'cp 0','e7e6 e2e3 g8e7 b1c3 e8g8 d2d4 c5d4 f3d4 d7d5 c4d5 e7d5 h2h4 h7h5 d1c2 d5b4 c2d1');
INSERT INTO evaluations VALUES(9315,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(9315,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(9315,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(9315,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(9315,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(9315,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(9315,36,'cp -30','e8g8 e1g1 c6a5 c4b3 a5b3 a2b3 f8e8 g5f3 a7a5 d2d3 c8d7 c1e3 c5e3 f2e3 b7b5 d1e1 c7c6 e1g3 d7e6 f3h4 f6h5 g3f2 h7h6 c3e2 g8h7 h4f5 g7g6 f2f3 e8f8 e2g3 h5g3 f5g3 a5a4 b3a4 b5a4 f1f2 d8b8 c2c4');
INSERT INTO evaluations VALUES(14735,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(14735,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(14735,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(14735,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(14735,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(14735,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(14735,36,'cp 21','f8d6 b5d6 c7d6 g1f3 e8g8 c2c3 h7h6 f1e2 f8e8 h2h3 e6e5 f4h2 c8f5 e1g1 f5g6 a1c1 a7a6 f1e1 b7b5 a2a4 d8b6 d1b3 a8b8 a4b5 a6b5 c1a1 g6h5 d4e5 d6e5 b3a3 f6e4 a3a2 e4f6 h2g3 f6e4 a2d5 e4g3 f2g3 b6e3 g1h2');
INSERT INTO evaluations VALUES(9219,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(9219,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(9219,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(9219,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(9219,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(9219,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(9219,36,'cp 11','b5d6 c7d6 g1f3 e8g8 f1e2 f6e4 e1g1 f8e8 e2d3 e4f6 f4g5 h7h6 g5h4 e6e5 d3e2 c8f5 c2c3 g7g5 h4g3 d8b6 b2b3 a8c8 a1c1 e5d4 f3d4 c6d4 d1d4 b6d4 c3d4 f6e4 e2f3 e4d2 f1e1 d2f3 g2f3 e8d8 h2h4 a7a6 g1h2 f7f6 c1c8 d8c8 h4g5 h6g5 g3d6');
INSERT INTO evaluations VALUES(14881,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(14881,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14881,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14881,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(14881,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14881,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(14881,36,'cp 52','d2d4 f8e7 c2c3 e8g8 d1c2 b7b6 b1d2 c8b7 e2e4 d5e4 d2e4 d8c8 f3g5 f6e4 g5e4 c6a5 f1e1 a5c4 c1f4 c4d6 a1d1 f8d8 c2e2 d6e4 g2e4 b7e4 e2e4 e7d6 f4d6 c7d6 e1e3 b6b5 a2a3 a7a5 e4f3 h7h6 h2h4');
INSERT INTO evaluations VALUES(15471,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(15471,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(15471,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15471,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(15471,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15471,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15471,36,'cp -3','c2c4 f8e7 b1c3 e8g8 f1e2 c6d4 g5e3 c7c5 e1g1 c8d7 f3d2 a7a6 a2a4 h7h6 h2h3 b7b6 g1h1 a8b8 f2f4 b8b7 e2f3 e5f4 e3f4 d7e6 c3e2 d4f3 d2f3 f6h5 d1d2 h5f4 e2f4 e7f6 f4d5 f6e5 f3e5 d6e5');
INSERT INTO evaluations VALUES(19840,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19840,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19840,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19840,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19840,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19840,36,'cp 3','c8f5 b1d2 e7e6 g1f3 a7a6 d2b3 g8f6 f1d3 f5d3 d1d3 f8d6 f4d6 d8d6 e1g1 d6c7 b3c5 a6a5 d3e2 e8g8 c5d3 c6e7 f3e5 a8c8 a2a4 e7f5 f1e1 f6e4 e5g4 b7b6 g4e3 f5d6 d3e5 f8e8 e5d3');
INSERT INTO evaluations VALUES(20890,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(20890,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(20890,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(20890,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(20890,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(20890,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(20890,36,'cp 112','f4e5 c8f5 g1f3 e7e6 f1d3 g8e7 c2c3 d8d7 e1g1 c6d8 f3g5 h7h6 g5h3 f5d3 d1d3 g7g5 h3f2 e7f5 f2g4 f8g7 b1d2 e8g8 d2f3 d7f7 c1d2 f7g6 g4f6 g7f6 e5f6 d8c6 g2g4 f5h4 d3g6 h4g6 h2h4 g5h4 g4g5 e6e5 f3e5 g6e5 d4e5 h6g5');
INSERT INTO evaluations VALUES(21216,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(21216,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21216,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21216,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(21216,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(21216,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(21216,36,'cp -17','e5g6 f4g3 g8f6 b1c3 e7e5 f1c4 f8d6 g1f3 a7a6 e1g1 f6h5 a1c1 c8d7 c4e2 h5f4 f1e1 d8f6 e2f1 e8g8 c3a4 b7b6 e1e3 f6e7 a4c3 g8h8 a2a3 f7f6 h2h3 e7e8 g1h2 b6b5 e3e1 f4h5 d1d2 g6f4 g3f4 h5f4 c3e2 f4e2 f1e2 f6f5');
INSERT INTO evaluations VALUES(22414,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(22414,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(22414,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(22414,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(22414,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(22414,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(22414,36,'cp 90','e2e4 e7e5 c2c4 g7g6 b1c3 g8f6 f1e2 f8g7 c1e3 e8g8 e1g1 d8e7 d1c2 f8e8 d4d5 f6g4 e3g5 f7f6 g5d2 f6f5 b2b4 a7a5 a2a3 d7f6 f3g5 a5b4 a3b4 a8a1 f1a1 h7h6 g5e6 c8e6 d5e6 f5e4 c3e4 e7e6 e4f6 g4f6 c2g6 e6f7 g6f7 g8f7 e2d3');
INSERT INTO evaluations VALUES(25490,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25490,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(25490,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(25490,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(25490,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(25490,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(25490,36,'cp 15','d6e5 e1g1 f8d6 c1g5 h7h6 g5h4 a7a6 b5a4 d7e6 d1e2 g7g5 h4g3 f6d7 f1d1 f7f6 a4c6 b7c6 f3d2 h6h5 f2f3 d7f8 d2c4 f8g6 c4d6 c7d6 e2d2 e8e7 g3f2 g6f4 f2e3 h8g8 g1h1 h5h4 h2h3 d8d7 h1h2 a8b8 c3a4 c6c5 b2b3');
INSERT INTO evaluations VALUES(25483,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(25483,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(25483,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(25483,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(25483,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(25483,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(25483,36,'cp 19','e1g1 f8d6 c1g5 h7h6 g5h4 a7a6 b5a4 d7e6 g1h1 e8g8 a4c6 b7c6 b2b3 d6b4 d1e1 a6a5 h2h3 d8d6 h4f6 g7f6 e1e3 d6c5 e3c5 b4c5 c3a4 c5a3 h1g1 a8d8 f1d1 d8d6 d1d3 d6d3 c2d3 f8d8 a1d1 g8f8 g1f1 f8e7 f3h4 h6h5 f1e2 a3b4');
INSERT INTO evaluations VALUES(27885,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(27885,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(27885,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(27885,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(27885,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(27885,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(27885,36,'cp 73','f1e1 b8c6 c1f4 f8e8 h2h3 a7a6 a2a4 e7f8 a4a5 c6b4 c4b3 h7h6 g2g4 a8c8 d4f5 d7e6 b3a4 e6d7 g4g5 d7f5 a4e8 f5h3 e8f7 g8f7 g5f6 d8f6 f4g3 h3e6 c3e2 f6f3 e2d4 f3d1 e1d1 e6g4');
INSERT INTO evaluations VALUES(31299,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(31299,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(31299,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(31299,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(31299,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(31299,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(31299,36,'cp 111','d1b3 d8b6 b1c3 b6b3 a2b3 b8c6 g1f3 e7e6 c4d5 e6d5 c3d5 a8d8 f1c4 f5e4 d5c3 e4f3 g2f3 a7a6 c1e3 c6b4 e1e2 f8e7 h1d1 g8f6 d4d5 b4c2 a1a4 c2e3 f2e3 g7g5 c4d3 f6d5 d3e4 d5c3 b2c3 d8c8 e4b7 c8c3 a4c4 c3c4 b3c4 e7b4 d1d5 e8e7 d5g5 a6a5 g5h5 h8b8 b7d5 b4c5 h5h7 b8b2 e2d3 b2b3 d3e4');
INSERT INTO evaluations VALUES(32117,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(32117,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(32117,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(32117,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(32117,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(32117,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(32117,36,'cp 90','b8d7 b1c3 f8e7 h2h3 g4f3 d1f3 e7f6 f1c4 g8e7 d4d5 e7g6 g2g3 f6g5 e3g5 d8g5 h3h4 g5h6 a2a4 g6e7 a4a5 a7a6 c4e2 e8g8 g3g4 f7f6 f3e3 h6e3 f2e3 g7g5 e1d2 g8g7 b2b4 h7h6 a1f1 f8e8 b4b5 g5h4');
INSERT INTO evaluations VALUES(33072,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(33072,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(33072,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(33072,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(33072,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(33072,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(33072,36,'cp 55','b1c3 g8e7 d3d4 c5b6 a2a4 a7a6 d4e5 d6e5 a4a5 b6a7 h2h3 a7c5 c1e3 c5d6 c4b3 e8g8 f3d2 b8c6 d1f3 f6f3 d2f3 c8d7 f1d1 f8d8 f3e1 e7g6 c3d5 d7e6 c2c3 g6f4 g1f1 g8f8 e1d3 e6d5 e4d5 c6e7 e3f4 e5f4 c3c4 h6h5');
INSERT INTO evaluations VALUES(33071,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(33071,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(33071,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(33071,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(33071,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(33071,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(33071,36,'cp 48','g8e7 d3d4 c5b6 a2a4 a7a6 d4e5 d6e5 a4a5 b6a7 h2h3 a7c5 c1e3 c5d6 c4b3 e8g8 f3d2 b8c6 d1f3 f6f3 d2f3 f8e8 f1d1 d6b4 c3d5 e7d5 b3d5 c6e7 d5b3 e7c6 b3a4 f7f6 c2c3 b4f8 a4b3 c8e6 b3e6 e8e6 g1f1 a8d8 b2b4 d8d1 a1d1 e6d6');
INSERT INTO evaluations VALUES(33903,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(33903,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(33903,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(33903,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(33903,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(33903,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(33903,36,'cp 81','d2d4 h7h6 f1d3 b8c6 c2c3 f8d6 d1e2 c8d7 e1g1 e8g8 b2b4 a7a6 f1e1 c6e7 f3e5 d6e5 d4e5 f6h4 g2g3 h4h3 e2e4 f7f5 e5f6 f8f6 e4h7 g8f7 d3f1 h3h5 f1e2 h5h3 h7e4 d7c6 e2g4 c6e4 g4h3 e4d5 h3f1 e7f5 f2f4 b7b5 a2a4 c7c6 c1e3 f5e3 e1e3 a8b8 e3e1 b8d8 a4a5');
INSERT INTO evaluations VALUES(34183,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(34183,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(34183,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(34183,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(34183,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(34183,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(34183,36,'cp 61','f8b4 e1g1 g8e7 c4b3 d7d6 c3e2 b4c5 g1h1 c8e6 f3d2 g7g5 d3d4 e5d4 f2f4 b8d7 d2f3 e8c8 f3d4 e6g4 c2c3 f6g6 d1c2 d6d5 e2g3 g5f4 c1f4 h6h5 e4d5 e7d5 b3d5 g6c2 d4c2 c6d5 c2e3 d8e8 a1e1 c5e3 f4e3 h5h4 g3f5 g4f5 f1f5 h4h3 g2g3');
INSERT INTO evaluations VALUES(34653,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(34653,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(34653,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(34653,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(34653,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(34653,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(34653,36,'cp 107','b1c3 e7e5 g1f3 d4d6 d2d4 c8g4 f1e2 b8d7 c1e3 a7a6 d4e5 d6d1 a1d1 e8c8 h2h3 g4e6 c3d5 f8c5 e1f2 d8e8 d5f4 g8e7 b2b3 c5e3 f2e3 e7c6 f4d3 e8e7 a2a3 h8e8 g2g4 h7h6 e3d2 b7b6 h1f1 e8h8 a3a4 h8e8');
INSERT INTO evaluations VALUES(34743,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(34743,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(34743,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(34743,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(34743,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(34743,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(34743,36,'cp 58','g8f6 f1c4 c8g4 d1d3 b8d7 e1g1 c7c6 h2h3 g4e6 f1d1 d7c5 d3e2 e6c4 e2c4 b7b5 c4e2 e7c7 c1g5 f8e7 g5f6 e7f6 b2b4 c5e6 a2a4 b5a4 a1a4 e8g8 d1a1 e6d4 e2d3 f8b8 f3d4 e5d4 c3e2 c6c5 d3c4 b8c8 b4c5 c7c5 c4c5 c8c5');
INSERT INTO evaluations VALUES(38182,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(38182,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38182,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(38182,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38182,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(38182,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(38182,36,'cp 52','g8e7 d5e3 d8c7 f1d3 e8g8 g1f3 b8c6 e1g1 c6b4 d3e4 c7f4 e4d5 b4d5 e3d5 f4d6 d5e7 d6e7 c2c3 c8g4 c1g5 e7e6 h2h3 g4h5 f1e1 e6b6 d1c2 f8e8 b2b4 c5f8 e1e8 a8e8 c2d3 b6c6 f3d4 c6d5 g5e3 e8c8 a1c1 c8c4 f2f3 h5g6 d3d2 f7f6 a2a4 h7h6');
INSERT INTO evaluations VALUES(41700,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(41700,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(41700,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(41700,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(41700,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(41700,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(41700,36,'cp 0','e8g8 f1d3 f8e8 c3e2 c7c5 d4c5 d6c5 e1g1 f6e4 c2c3 b8c6 f3d4 c6d4 e2d4 c5d4 c3d4 c8f5 f1e1 d8f6 d3e4 f5e4 c1e3 a8c8 a1c1 c8c1 e3c1 e8c8 f2f3 e4f5 d1b3 f6g6 g1h2 g6d6 h2g1 d6g6');
INSERT INTO evaluations VALUES(36934,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36934,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36934,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36934,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(36934,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(36934,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36934,36,'cp 0','f1d3 f8e8 c3e2 b7b6 e1g1 c7c5 c2c4 c8a6 c4d5 f6d5 d4c5 d6c5 d3a6 b8a6 c1g5 d8d7 a2a3 h7h6 g5h4 c5f8 e2c3 a6c7 d1d3 d7c6 f3d4 c6d7');
INSERT INTO evaluations VALUES(41769,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41769,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41769,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(41769,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(41769,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41769,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(41769,36,'cp 0','c7c5 c2c4 c5d4 f3d4 d6b4 b1c3 d8a5 g5f6 b4c3 b2c3 a5c3 e1e2 g7f6 a1c1 c3a5 c4d5 e8g8 d1b3 e6d5 b3c3 a5d8 f2f3 b8c6 e2f2 c8e6 f1d3 a8c8 d4c6 c8c6 c3d4 d8a5 c1c6 b7c6 d4f6 a5a2 f2g3 a2a3 f6g5 g8h8 g5f6');
INSERT INTO evaluations VALUES(41982,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(41982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41982,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(41982,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41982,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41982,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(41982,36,'cp 0','e6d5 d1a4 b8c6 f1d3 e8g8 g1e2 f8e8 c1d2 b4f8 e1g1 c6b4 d3b1 b4a6 b1d3');
INSERT INTO evaluations VALUES(42572,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(42572,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(42572,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(42572,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(42572,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(42572,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(42572,36,'cp -18','c7c6 g1f3 d7d5 e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 d3d4 e5d4 e1g1 e8g8 b1d2 f8e8 d2b3 c5f8 b3d4 d7b6 c1g5 b6c4 a1b1 h7h6 g5f6 d8f6 c2c3 f6b6 a3a4 g7g6 h2h4 h6h5 f1e1 e8e4 d1d3 e4e1 f3e1 f8g7 g2g3 g7d4 d3d4 b6d4 c3d4');
INSERT INTO evaluations VALUES(45558,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(45558,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45558,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45558,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(45558,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45558,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(45558,36,'cp 5','b7b5 c4b3 f8e7 c3e2 e8g8 c1d2 a7a5 a2a4 b5b4 e2g3 f8e8 c2c3 b4c3 b2c3 b8d7 f1e1 d8c7 d3d4 c8a6 c3c4 e5d4 f3d4 e7f8 d4f5 d7e5 a1c1 e5d3 d2c3 d3c1 d1c1 e8e6 c1b2 a8b8 f5h6 g7h6 c3f6 e6f6 b2f6');
INSERT INTO evaluations VALUES(45922,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(45922,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(45922,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(45922,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(45922,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(45922,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(45922,36,'cp 15','f8e7 g5f4 d8a5 g1f3 f6e4 d1c2 e4c3 c2c3 a5c3 b2c3 b8d7 e2e4 f7f5 e4e5 d6e5 f4e5 e7f6 e5f6 d7f6 c3c4 e8d8 f1d3 h8e8 e1d2 d8c7 h1e1 c8d7 f3e5 f6e4 d3e4 e8e5 e4d3 a8e8 e1e5 e8e5 g2g3 a7a6 a2a4 a6a5 h2h4 h7h6 h4h5 e5e7 a1h1 e7e8 f2f3 e8e7 d3c2 c7b6 h1b1 b6c7');
INSERT INTO evaluations VALUES(40152,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(40152,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(40152,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40152,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(40152,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40152,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40152,36,'cp 29','g5f4 d8a5 g1f3 f6e4 d1c2 e4c3 c2c3 a5c3 b2c3 b8d7 e2e4 f7f5 e4e5 d6e5 f4e5 e7f6 e5f6 d7f6 c3c4 e8d8 f1d3 h8e8 e1d2 c8d7 a1e1 d8c7 f3e5 f6e4 d3e4 e8e5 e4d3 a8e8 e1e5 e8e5 f2f4 e5e7 h1a1 b7b6 a2a4 a7a5 h2h4 g7g6 d3c2 e7e8 a1b1 e8b8 d2e3 c7b7 b1d1 b8e8 e3d2');
INSERT INTO evaluations VALUES(46687,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46687,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(46687,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(46687,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(46687,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(46687,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(46687,36,'cp 21','b8d7 e2e3 f8g7 f1e2 b7b6 e1g1 c8b7 b1d2 a7a6 a2a4 e8g8 f4h2 f8e8 a4a5 g7f8 a5b6 c7b6 d1b1 e7e5 b1a2 e5d4 f3d4 d7c5 f1d1 d8e7 c2c4 f8g7 e2f3 b7f3 d4f3 e7e6 b2b4 c5e4 a2c2 e4d2 d1d2 f6e4 f3d4 e6e7 d2d1');
INSERT INTO evaluations VALUES(50566,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50566,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50566,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(50566,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50566,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(50566,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50566,36,'cp 0','b8c6 d4e3 f8e7 f3d4 d7d5 d4c6 b7c6 e3g3 g8f6 f1d3 e8g8 e1g1 f8e8 c1h6 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(17443,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17443,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17443,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17443,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17443,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17443,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17443,36,'cp 0','d4e3 f8e7 f1d3 d7d5 f3d4 g8f6 d4c6 b7c6 e3g3 e8g8 e1g1 f8e8 c1h6 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(53155,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(53155,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53155,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53155,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(53155,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(53155,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(53155,36,'cp 3','c8d7 b5d7 d8d7 c2c3 e7e6 g1f3 b8c6 e1g1 c5d4 c3d4 g8e7 a2a3 e7f5 b1c3 a8c8 c3e2 f8e7 d1d3 e8g8 c1d2 f7f6 e5f6 e7f6 a1e1 d7f7 d2c3 f5d6 h2h3 a7a5 a3a4 d6e4 e2g3 e4g5 f3g5 f6g5 e1e2 g5f4 f1e1 f4d6 d3e3 c8e8 e3d3');
INSERT INTO evaluations VALUES(30940,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(30940,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(30940,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(30940,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30940,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(30940,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(30940,36,'cp 9','b5d7 d8d7 c2c3 e7e6 g1f3 b8c6 e1g1 c5d4 c3d4 g8e7 b1c3 e7f5 c3e2 f8e7 a2a3 e8g8 d1d3 a7a5 c1d2 f7f6 d2f4 f6e5 f4e5 a5a4 a1e1 f5h4 f3h4 e7h4 g2g3 h4d8 e2c3 a8a6 d3d1 d8a5 g1g2 c6e5 e1e5 a6b6');
INSERT INTO evaluations VALUES(30866,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30866,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(30866,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30866,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(30866,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30866,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30866,36,'cp 0','d8d7 c2c3 e7e6 g1f3 b8c6 e1g1 a8c8 a2a3 c5d4 c3d4 g8e7 b1c3 e7f5 d1d3 f8e7 c3e2 e8g8 c1d2 f7f6 e5f6 e7f6 a1e1 d7f7 d2c3 c8e8 e2f4 f6d8 d3d1 d8f6');
INSERT INTO evaluations VALUES(53596,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(53596,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(53596,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(53596,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53596,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(53596,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53596,36,'cp 12','g8f6 f1d3 c6c5 e1g1 f8e7 d4c5 b8d7 c3a4 d7c5 a4c5 e7c5 b2b3 e8g8 c1b2 d8e7 f3e5 f8d8 d1e2 c8d7 e5d7 d8d7 a1d1 a8d8 c2c4 c5b4 d3c2 g7g6 g2g3 e6e5 g1g2 a7a6 c4d5 f6d5 c2e4 g8g7 e2c2 b7b5 d1d5 d7d5 e4d5');
INSERT INTO evaluations VALUES(45674,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(45674,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(45674,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45674,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(45674,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45674,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45674,36,'cp 10','f1d3 c6c5 e1g1 f8e7 d4c5 b8d7 c3a4 d7c5 a4c5 e7c5 b2b3 e8g8 c1b2 d8e7 f3e5 f8d8 d1f3 c8d7 c2c4 d7e8 c4d5 f6d5 a1c1 c5a3 b2a3 e7a3 f1d1 f7f6 e5c4 a3e7 d3f1 a8c8 e3e4 d5b6 f3e3 b6c4 f1c4 d8d1 c1d1 b7b5 c4e2');
INSERT INTO evaluations VALUES(45675,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(45675,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(45675,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(45675,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(45675,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(45675,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(45675,36,'cp 12','c6c5 e1g1 f8e7 d4c5 b8d7 c3a4 d7c5 a4c5 e7c5 b2b3 e8g8 c1b2 d8e7 f3e5 f8d8 d1e2 c8d7 e5d7 d8d7 a2a3 e6e5 f1d1 e5e4 d3b5 d7d8 b3b4 c5d6 c2c4 a7a6 c4c5 d6h2 g1h2 a6b5 h2g1 e7d7 b2f6 g7f6 e2h5 g8g7 h5h4 d7e6 h4g3 g7h8');
INSERT INTO evaluations VALUES(54999,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(54999,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54999,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(54999,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(54999,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(54999,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(54999,36,'cp 28','e7e5 c2c4 g7g6 d2d4 e5d4 e3d4 f8g7 h2h3 g8e7 f1d3 e8g8 e1g1 b8c6 d4d5 c6e5 f3e5 g7e5 b1d2 e7f5 d2f3 e5g7 a1b1 f8e8 c1g5 g7f6 g5f4 f5d4 f3d4 f6d4 d1d2 d8f6 f1e1 e8e1 b1e1 a7a5 f4e3 d4e3 e1e3');
INSERT INTO evaluations VALUES(56881,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(56881,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(56881,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56881,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(56881,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(56881,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(56881,36,'cp 17','b1c3 d7d5 g1f3 d5e4 c3e4 f8g7 g2g3 c8d7 f1g2 d7c6 f3e5 g7e5 f4e5 f7f5 e5f6 g8f6 d1e2 f6e4 g2e4 c6e4 e2e4 d8d5 e4d5 e6d5 d2d3 b8c6 c1h6 e8c8 e1d2 h8e8 h1f1 c6d4 f1f2 d8d6 c2c3 d4f5 h6f4 d6f6 g3g4 f5g7 a1f1 e8f8 f4e3 f6f2 f1f2 f8f2 e3f2 b7b6 d2e3 g7e6');
INSERT INTO evaluations VALUES(57611,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(57611,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(57611,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(57611,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(57611,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(57611,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(57611,36,'cp 68','e7e5 e2e4 g8f6 e4f5 d7d5 g3g4 b8c6 d2d3 g7g6 b1c3 d5d4 c3e4 g6f5 g4f5 c8f5 c1g5 f8e7 g5f6 e7f6 d1f3 f5e4 f3e4 d8d6 e1c1 e8c8 g1f3 d8f8 f3d2 f6d8 e4e2 f8f4 c1b1 h8f8 d1f1 c6b4 a2a3 b4d5 g2d5 d6d5 f2f3');
INSERT INTO evaluations VALUES(57516,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(57516,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(57516,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(57516,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(57516,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(57516,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(57516,36,'cp 51','e2e4 g8f6 e4f5 b8c6 d2d3 d7d5 g3g4 g7g6 b1c3 h7h6 g4g5 h6g5 c1g5 c8f5 c3d5 f8g7 g1e2 d8d6 d5c3 e8c8 d1d2 f5h3 g2h3 h8h3 e1c1 d8h8 c1b1 f6g4 c3e4 d6e6 a2a3 g4h2 g5e3 b7b6 d2c1 h3h4 e4g5 e6g4 e2g3 h2f3 g3e4 f3g5 e4g5 g4f5 g5e4 f5e6 h1h4 h8h4');
INSERT INTO evaluations VALUES(57679,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(57679,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(57679,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(57679,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(57679,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(57679,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(57679,36,'cp -22','f8g7 d2d4 c7c5 c2c3 g8f6 d4c5 f6e4 c1b2 e4c5 b3b4 c5e4 b1d2 e4d2 d1d2 e8g8 g1f3 d7d5 c3c4 d5c4 b2g7 d8d2 e1d2 g8g7 f1c4 a7a5 b4b5 b8d7 h1c1 d7b6 c4e2 f8d8 d2e1 c8e6 c1c7 d8d7 a1c1 a8d8 a2a3 d7c7 c1c7');
INSERT INTO evaluations VALUES(39634,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(39634,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(39634,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(39634,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(39634,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(39634,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(39634,36,'cp -5','d2d4 c7c5 c2c3 g8f6 g1f3 e8g8 f1e2 b7b6 c1b2 c8b7 e1g1 b8c6 c3c4 c5d4 f3d4 a8c8 b1a3 d7d5 d4c6 c8c6 e2f3 c6d6 d1e2 d8a8 f1d1 d6d7 a1c1 f8c8 h2h3 d5c4 a3c4 b7f3 g2f3 f6d5 a2a4 c8d8 b2g7 g8g7');
INSERT INTO evaluations VALUES(1398,30,'cp 423',NULL);
INSERT INTO evaluations VALUES(1398,31,'cp 412',NULL);
INSERT INTO evaluations VALUES(1398,32,'cp 403',NULL);
INSERT INTO evaluations VALUES(1398,33,'cp 415',NULL);
INSERT INTO evaluations VALUES(1398,34,'cp 428',NULL);
INSERT INTO evaluations VALUES(1398,35,'cp 415',NULL);
INSERT INTO evaluations VALUES(1398,36,'cp 440','b1c3 d5d8 d4d5 c6d4 f3d4 c5d4 f1b5 c8d7 b5d7 d8d7 d1d4 g8f6 c1g5 h7h6 g5f6 g7f6 e1c1 e7e5 d5e6 d7d4 e6f7 e8f7 d1d4 a8e8 d4d7 e8e7 h1d1 h6h5 d1d4 f6f5 c3e2 h5h4 d7e7 f7e7 e2f4 f8h6 c1d1 h6f4 d4f4 e7f6 f4b4 b7b6 g2g3 h4h3 b4h4 h8h4 g3h4 f6g6 d1e2 g6h5 e2f3 h5h4 f3f4 h4h5 c2c4');
INSERT INTO evaluations VALUES(1394,30,'cp 421',NULL);
INSERT INTO evaluations VALUES(1394,31,'cp 408',NULL);
INSERT INTO evaluations VALUES(1394,32,'cp 441',NULL);
INSERT INTO evaluations VALUES(1394,33,'cp 411',NULL);
INSERT INTO evaluations VALUES(1394,34,'cp 428',NULL);
INSERT INTO evaluations VALUES(1394,35,'cp 422',NULL);
INSERT INTO evaluations VALUES(1394,36,'cp 419','d5d8 d4d5 c6d4 f3d4 c5d4 d1d4 g8f6 c1f4 c8d7 e1c1 d8b6 f1e2 b6d4 d1d4 a8c8 h1d1 h7h5 a2a4 a7a6 h2h3 d7f5 a4a5 h5h4 f4e3 e7e5 d5e6 f8c5 e6f7 e8f7 e2c4 f7g6 d4f4 c5e3 f2e3 c8c5 b2b4 c5e5 c4d3 f5d3 d1d3 h8c8 c1b2');
INSERT INTO evaluations VALUES(19764,30,'cp 427',NULL);
INSERT INTO evaluations VALUES(19764,31,'cp 404',NULL);
INSERT INTO evaluations VALUES(19764,32,'cp 425',NULL);
INSERT INTO evaluations VALUES(19764,33,'cp 434',NULL);
INSERT INTO evaluations VALUES(19764,34,'cp 406',NULL);
INSERT INTO evaluations VALUES(19764,35,'cp 431',NULL);
INSERT INTO evaluations VALUES(19764,36,'cp 440','d4d5 c6d4 f3d4 c5d4 d1d4 g8f6 c1f4 c8d7 e1c1 d8b6 f1e2 b6d4 d1d4 a8c8 h1d1 g7g6 d5d6 e7e6 f4e5 f8g7 c3e4 f6e4 e5g7 h8g8 g7e5 f7f6 d4e4 f6e5 e4e5 g8f8 e2f3 b7b6 e5e3 c8c5 d1d2 c5a5 a2a3 e8d8 f3e4 a5h5 h2h3 h5e5 f2f3 f8f4 b2b4');
INSERT INTO evaluations VALUES(2698,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(2698,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(2698,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(2698,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(2698,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(2698,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(2698,36,'cp 27','b7b5 a4b3 d7d6 h2h3 c8b7 d2d4 f8e8 b1d2 e7f8 a2a3 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 c2d3 f6h5 d3f1 h5f4 d1c2 a6a5 a3a4 e5d4 c3d4 d6d5 e4d5 e8e1 a1e1 f4d5 a4b5 c6b5 f1b5 f8b4 c2f5 d7f6');
INSERT INTO evaluations VALUES(2817,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(2817,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(2817,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(2817,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(2817,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(2817,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(2817,36,'cp 39','a4b3 d7d6 h2h3 c8b7 d2d4 f8e8 a2a3 e7f8 b1d2 h7h6 b3c2 c6b8 b2b3 b8d7 c1b2 c7c6 a1c1 a8c8 c2d3 f6h5 d3f1 h5f4 g2g3 f4g6 h3h4 d7f6 c3c4 d8d7 f1g2 d7d8 d1c2 e5d4 f3d4 c6c5 d4f5 g6e5');
INSERT INTO evaluations VALUES(4463,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(4463,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4463,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(4463,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4463,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(4463,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4463,36,'cp 0','c3c4 d7d6 f1e2 f6d7 e1g1 d7c5 c1g5 f7f6 g5e3 b7b6 f3h4 c6e7 g1h1 a7a6 h2h3 c8d7 e2g4 b6b5 a1b1 c5a4 b1b3 a4c5');
INSERT INTO evaluations VALUES(4608,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(4608,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(4608,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(4608,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(4608,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(4608,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(4608,36,'cp 38','f3d4 e7b4 c3d5 f6d5 c4d5 b4c5 c1e3 c6d4 e3d4 d8e7 d1d3 d7d6 a1d1 c5d4 d3d4 c8e6 b2b3 e6d5 d4d5 c7c6 d5d4 a7a6 f1e1 a8d8 c2c4 f8e8 h2h3 h7h6 d4b6 d8d7 g2g3 e7e5 b6e3 e5a5 e3d2 a5d2 d1d2');
INSERT INTO evaluations VALUES(5216,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(5216,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(5216,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(5216,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(5216,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(5216,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(5216,36,'cp 78','f3g5 g8h6 e1g1 e8g8 d1h5 b4c3 b2c3 d8e7 f1d1 c6a5 c4d5 d7d6 h2h3 g8h8 g5f3 f7f5 a1b1 a8b8 c1h6 g7h6 h5h6 f8f6 h6h5 c8d7 e4f5 d7f5 b1b4 b8f8 b4h4 f6g6 d5e4 f5e4 h4e4 e7g7 e4g4 g7h6 h5h6 g6h6 f3g5 b7b6 g5e4 a5c4 e4g3 f8c8 g4g5 c8c7');
INSERT INTO evaluations VALUES(10200,30,'cp -146',NULL);
INSERT INTO evaluations VALUES(10200,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(10200,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(10200,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(10200,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(10200,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(10200,36,'cp -139','d7d5 e4d5 f6d5 a2a3 e8g8 b1d2 f8e8 d2e4 c5f8 f1e2 d5f4 e1g1 b7b6 b3b4 c8b7 f1e1 f4e2 d1e2 c6d4 f3d4 e5d4 e2f3 d8d5 e4d2 d5f3 d2f3 c7c5 b4c5 b6c5 g1f1 e8e1 f3e1 a8e8 b2c1 h7h6 c1f4 g7g5 f4c7 f7f5 e1f3 b7a6');
INSERT INTO evaluations VALUES(10790,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(10790,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(10790,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(10790,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(10790,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(10790,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(10790,36,'cp -80','d7d6 f4e5 d6e5 c1g5 d8e7 d1d3 c8e6 c4e6 d4e6 g5f6 g7f6 b1c3 c7c6 d3f3 h8g8 f3f6 e6f4 f6e7 e8e7 g2g3 f7f6 a1e1 a8d8 g1f2 f4e6 e1d1 h7h5 d1d8 g8d8 f2e3 a7a5 c3d1 h5h4 f1g1 d8h8 d1f2 h8g8 c2c3 a5a4 f2d3 h4g3 h2g3');
INSERT INTO evaluations VALUES(9532,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(9532,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(9532,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(9532,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(9532,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(9532,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(9532,36,'cp -67','f4e5 d6e5 c1g5 d8e7 d1d3 c8e6 c4e6 d4e6 g5f6 g7f6 b1c3 c7c6 d3f3 h8g8 f3f6 e6f4 f6e7 e8e7 g2g3 f7f6 g1f2 f4e6 f2e3 h7h5 c3e2 g8g4 h2h3 g4g6 a1d1 a7a5 d1d2 g6g8 d2d3 e6c5 d3d1 b7b6 f1f5 g8h8 d1f1 a8d8 e2c3 h5h4 g3h4 h8h4');
INSERT INTO evaluations VALUES(9531,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(9531,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(9531,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(9531,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(9531,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(9531,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(9531,36,'cp -65','d6e5 c1g5 d8e7 d1d3 c8e6 c4e6 d4e6 g5f6 g7f6 b1c3 c7c6 d3f3 h8g8 f3f6 e6f4 f6e7 e8e7 g2g3 f7f6 g1f2 f4e6 f2e3 h7h5 a1d1 e6d4 f1f2 g8g4 d1f1 a8f8 c3d1 h5h4 c2c3 h4g3 f2g2 d4b5 d1f2 g4g8 h2g3 b5d6 e3e2 e7e6 f1d1 g8h8 g3g4 f8g8 e2f3');
INSERT INTO evaluations VALUES(9541,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(9541,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(9541,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(9541,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(9541,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(9541,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(9541,36,'cp -62','c1g5 d8e7 d1d3 c8e6 c4e6 d4e6 g5f6 g7f6 b1c3 c7c6 d3f3 h8g8 f3f6 e6f4 f6e7 e8e7 g2g3 f7f6 g1f2 f4e6 a1d1 h7h5 d1d2 g8g4 h2h3 g4g8 c3e2 a7a5 f2e3 g8h8 b2b3 a8d8 d2d8 h8d8 a2a4 e6g5 f1g1 d8g8 h3h4 g5f7 e2c1 b7b5 c1d3 b5a4');
INSERT INTO evaluations VALUES(9540,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(9540,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(9540,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(9540,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(9540,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(9540,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(9540,36,'cp -54','d8e7 d1d3 c8e6 c4e6 d4e6 g5f6 g7f6 b1c3 c7c6 d3f3 h8g8 f3f6 e6f4 f6e7 e8e7 g2g3 f7f6 g1f2 f4e6 a1d1 h7h5 d1d2 a8d8 c3b1 a7a5 f1d1 d8d4 c2c3 d4d2 b1d2 b7b5 d2b3 g8a8 b3c1 e6c5 f2f3 a8h8 h2h4 h8g8 c1d3 c5b7 b2b3 e7e6 d3f2');
INSERT INTO evaluations VALUES(13949,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(13949,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(13949,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(13949,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(13949,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(13949,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(13949,36,'cp -46','f6e4 c3e4 d6d5 d2d3 d5c4 d3c4 d8d1 f1d1 b7b6 e4c3 f8e7 c3b5 e8g8 c1e3 e7f6 a1b1 c8b7 b5d6 b7a6 a2a3 f8d8 g1f2 d8d7 d6b5 d7d1 b1d1 a8d8 d1d8 f6d8 b2b3 f7f6 f3d2 d8e7 d2e4 a6c8 a3a4 g8f8 g2g3 a7a6 b5d6 c8d7 d6b7 g7g6');
INSERT INTO evaluations VALUES(14741,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(14741,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(14741,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(14741,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(14741,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(14741,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(14741,36,'cp 7','c1f4 f6h5 f4c1 h5f6 e2e3 a7a6 f2f4 b7b6 b2b4 f6e4 c3e4 d5e4 d1c2 f7f5 c1d2 c6e7 f1c4 d8d7 g1e2 c8b7 e1g1 b7d5 a2a4 e8f7 f1b1 d5c4 c2c4 e7d5 e2c3 f8e7 c4e2 g7g6 c3d5 e6d5 d2e1 h8b8 e2c2 d7e6 c2a2');
INSERT INTO evaluations VALUES(14828,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14828,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14828,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(14828,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14828,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(14828,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14828,36,'cp 0','a2a3 b7b6 g1f3 f8d6 b2b3 c6e7 d1e2 e8g8 c1b2 c7c5 b1d2 a7a5 a3a4 e7c6 d3b5 c6b4 a1c1 b4a2 c1a1 a2b4');
INSERT INTO evaluations VALUES(18712,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(18712,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(18712,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(18712,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(18712,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(18712,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(18712,36,'cp -42','c2c3 c5d4 c3d4 d7d6 g1f3 d6e5 d4e5 d8d1 e1d1 c8d7 b1c3 g8e7 c3b5 e7d5 f1c4 c6a5 c4d3 a7a6 b5d4 a8c8 d1e2 a5c4 d3e4 f8b4 c1e3 e8e7 a1c1 h8d8 e4d5 e6d5 b2b3 c4e3 f2e3 c8c5 c1c5 b4c5 h1d1 h7h6');
INSERT INTO evaluations VALUES(18710,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(18710,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(18710,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(18710,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(18710,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(18710,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(18710,36,'cp -50','c5d4 c3d4 d7d6 g1f3 d6e5 d4e5 d8d1 e1d1 c8d7 b1c3 g8e7 c3b5 e7d5 f1c4 c6a5 c4d3 a8c8 d1e2 f8e7 c1e3 a7a6 b5d4 a5c4 d3e4 e7c5 e4d5 e6d5 b2b3 c4e3 f2e3 c5a3 h1d1 h7h5 d1d3 h5h4 e2f2 h8h5 h2h3 g7g5 a1d1 g5g4 h3g4 d7g4');
INSERT INTO evaluations VALUES(18871,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(18871,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(18871,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(18871,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(18871,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(18871,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(18871,36,'cp -44','c3d4 d7d6 g1f3 d6e5 d4e5 d8d1 e1d1 g8e7 b1c3 c8d7 c3b5 e7d5 f1c4 c6a5 c4d3 a8c8 d1e2 h7h6 c1e3 a7a6 b5d4 a5c4 d3e4 f8c5 e4d5 e6d5 b2b3 c4e3 f2e3 e8e7 h2h3 c5a3 a1d1 c8c5 d1d3 h8c8 g2g4 g7g6 e2f2 a3b4 h1d1 c5c1 d4e2 c1d1 d3d1');
INSERT INTO evaluations VALUES(18869,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(18869,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(18869,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(18869,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(18869,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(18869,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(18869,36,'cp -44','d7d6 g1f3 d6e5 d4e5 d8d1 e1d1 c8d7 b1c3 g8e7 c3b5 e7d5 f1c4 c6a5 c4d3 a7a6 b5d4 a8c8 d1e2 h7h6 c1e3 a5c4 d3e4 f8c5 e4d5 e6d5 b2b3 c4e3 f2e3 e8e7 a1c1 f7f6 a2a4 c5a3 c1c8 h8c8 h1a1 a3b2 a1d1 b2c3 e5f6 e7f6 d1f1');
INSERT INTO evaluations VALUES(22012,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22012,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22012,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22012,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22012,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22012,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22012,36,'cp 0','c8e6 b1a3 g8f6 c4e6 f7e6 f3g5 e6e5 d1b3 d8d7 a1c1 c6a5 b3c3 a5c6');
INSERT INTO evaluations VALUES(26669,30,'cp 271',NULL);
INSERT INTO evaluations VALUES(26669,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(26669,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(26669,33,'cp 246',NULL);
INSERT INTO evaluations VALUES(26669,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(26669,35,'cp 271',NULL);
INSERT INTO evaluations VALUES(26669,36,'cp 283','f5g4 d4d5 c6e5 d1d4 e5d7 f1c4 c7c6 c1e3 e7e5 d5e6 f7e6 f3g5 f8c5 d4g7 d8f6 g7f6 g8f6 e3c5 d7c5 h2h3 b7b5 c4b5 c6b5 h3g4 b5b4 c3b5 e8g8 f2f3 c5d3 e1e2 d3f4 e2f2 a8d8 a1d1 e6e5 b5d6 f6g4 f3g4 f4h3 f2e3 h3g5 d1d5 f8f4 d5e5 f4g4');
INSERT INTO evaluations VALUES(26682,30,'cp 253',NULL);
INSERT INTO evaluations VALUES(26682,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(26682,32,'cp 298',NULL);
INSERT INTO evaluations VALUES(26682,33,'cp 285',NULL);
INSERT INTO evaluations VALUES(26682,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(26682,35,'cp 298',NULL);
INSERT INTO evaluations VALUES(26682,36,'cp 287','d4d5 c6e5 d1d4 e5f3 g2f3 g4d7 c1e3 b7b5 a2a4 b5b4 c3e2 c4c3 e2f4 g8f6 b2c3 c7c5 d5c6 d7c6 f1b5 c6b5 a4b5 d8d4 c3d4 e7e6 f4d3 f8e7 b5b6 a7a5 e3f4 f6d7 b6b7 a8a7 b7b8q d7b8 f4b8 a7a8 b8e5 f7f6 e5f4 a5a4 e1d2 e6e5 f4e3 e5d4 e3d4 e8g8 h1g1 e7d6');
INSERT INTO evaluations VALUES(31866,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(31866,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(31866,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(31866,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31866,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(31866,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(31866,36,'cp 0','g8f6 c2c4 e7e6 g1f3 b8d7 d1b3 a8b8 b1c3 c7c6 f3h4 f5g4 h2h3 g4h5 c4d5 f6d5 h4f3 h5g6 f1e2 f8d6 e1g1 e8g8 f1d1 f8e8 c1d2 h7h6 d2e1 d6c7 a1c1 a7a5 e2d3 g6h5 d3e2');
INSERT INTO evaluations VALUES(35330,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(35330,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(35330,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(35330,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(35330,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(35330,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(35330,36,'cp 111','g8f6 g1f3 e7e5 c1e3 f6g4 e3d2 b8d7 h2h3 g4f6 d2e3 f8e7 f1d3 d7f8 a1c1 f8g6 e1g1 a7a5 d4d5 e8g8 g1h2 c8d7 b2b3 a8e8 f3g1 c7d8 d1d2 h7h6 g2g3 f6h7 c1d1 c6c5 d2e2 e7g5 f2f4 e5f4 g3f4 g5f6');
INSERT INTO evaluations VALUES(34091,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(34091,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(34091,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(34091,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(34091,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(34091,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(34091,36,'cp 127','h2h3 b8d7 c1e3 e7e5 g1f3 f8e7 a1c1 d7f8 b2b4 f8g6 f1e2 e8g8 e1g1 f6h5 f1e1 h5f4 e2f1 f7f5 a2a4 f5e4 c3e4 g8h8 d4d5 c8f5 e4g3 f5d7 a4a5 c7d8 g3e4 b7b6 a5a6 a8c8 b4b5 c6d5 c4d5 c8c1 e3c1 d7f5 e4g3 f5d7 g1h1 h8g8 h1h2');
INSERT INTO evaluations VALUES(36971,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(36971,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36971,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36971,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(36971,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(36971,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36971,36,'cp 0','f1d3 c7c5 d4c5 d6c5 g1f3 f6h5 c2c4 h5g3 h2g3 g7g6 e1g1 c5e7 a2a3 a7a5 c4d5 e6d5 e3e4 d5e4 d2e4 b8c6 d1e2 c6d4 f3d4 d8d4 d3c4 d4e5 f1e1 c8f5 a1d1 e5e4 e2e4 f5e4 e1e4 a8d8 d1d8 e7d8 g1f1 h7h6 e4e5 d8b6 f2f4');
INSERT INTO evaluations VALUES(36966,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36966,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36966,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36966,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36966,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36966,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36966,36,'cp -1','c7c5 d4c5 d6c5 g1f3 f6h5 c2c4 h5g3 h2g3 g7g6 e1g1 c5e7 a2a3 a7a5 a1c1 e7f6 b2b4 c8d7 c1b1 d5c4 d3c4 b7b5 c4d3 a5b4 a3b4 b8a6 d1e2 f6e7 d3b5 a6b4 d2e4 d7b5 e2b5 a8a5 b5b7 b4d5 f3e5 d8c7 b7c7 d5c7');
INSERT INTO evaluations VALUES(39232,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(39232,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(39232,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(39232,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(39232,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(39232,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(39232,36,'cp -68','g8f6 g1f3 f6e4 d1e2 d7d5 b1c3 f7f5 d2d3 c5f2 e1d1 b8c6 d3e4 d5e4 d1c1 f2b6 f3d2 e8g8 d2c4 d8g5 c1b1 b6d4 h2h4 g5g3 c3b5 f8f7 a2a4 c8e6 h1h3 g3g6 h4h5 g6f6 b5d4 e5d4 h5h6 g7g6 e2e1 a7a5 b2a3 b7b6 b1a2 c6e5 c4e5 f6e5 h3h4 c7c5 g2g3 a8e8 a3b2 e4e3 a1d1');
INSERT INTO evaluations VALUES(41852,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(41852,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(41852,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(41852,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(41852,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(41852,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(41852,36,'cp -44','e8g8 h2h3 c7c6 e1g1 b8d7 c4b3 b7b5 a2a3 c8b7 b1c3 a7a5 g5f3 h7h6 c1e3 f8e8 b3a2 b5b4 c3e2 d6d5 e3c5 d7c5 a3b4 a5b4 e4d5 c6d5 e2g3 d8b6 d3d4 e5d4 d1d4 f6e4 g3f5 b6f6 d4f6 e4f6 f1e1 e8e1 f3e1');
INSERT INTO evaluations VALUES(36987,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(36987,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(36987,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(36987,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(36987,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(36987,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(36987,36,'cp -40','h2h3 c7c6 e1g1 b8d7 c4b3 b7b5 g5f3 f8e8 b1c3 h7h6 a2a3 a7a5 c1e3 c5e3 f2e3 d7c5 b3a2 c8e6 a2e6 e8e6 b2b4 c5d7 d1d2 d7f8 c3e2 e6e8 e2g3 d6d5 e4d5 c6d5 b4a5 a8a5 a1b1 d8c7 f3h4 a5a3 b1b5');
INSERT INTO evaluations VALUES(52203,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(52203,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(52203,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(52203,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(52203,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(52203,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(52203,36,'cp 104','d7d5 e4d5 c6d5 c4b3 b8c6 d2d4 e5e4 b1c3 e4f3 f1e1 f8e7 c3d5 e8f8 d1f3 c6d4 f3d3 d4b3 d3b3 e7d6 c1f4 g7g5 f4g5 f6g5 b3c3 c8f5 c3h8 d6h2 g1h2 d8d5 a1d1 d5f7 h8d4 b7b6 h2g1 a8e8 e1e8 f7e8 c2c4 e8e7 b2b4 f8f7 c4c5 b6c5 b4c5 e7e6 d4c3 e6a2 c5c6 g8e7 c6c7 a2e6 d1d8 e7d5 c3c5 d5e7 d8d6 e6e1 g1h2');
INSERT INTO evaluations VALUES(54640,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(54640,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(54640,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(54640,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(54640,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(54640,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(54640,36,'cp 100','e4d5 c6d5 c4b3 b8c6 d2d4 e5e4 b1c3 e4f3 f1e1 f8e7 c3d5 e8f8 d1f3 c6d4 f3d3 d4b3 d3b3 e7d6 c1f4 g7g5 f4g5 f6g5 b3c3 c8f5 c3h8 d6h2 g1h2 d8d5 a1d1 d5f7 h8d4 b7b6 e1e5 a8e8 h2g1 e8e7 d4d8 f8g7 d1e1 e7e5 d8g5 f7g6 g5g6 h7g6 e1e5 g7f7 e5d5 f7e7 d5d2 f5e6 b2b3 b6b5 a2a4 a7a6 a4b5 a6b5 d2d1');
INSERT INTO evaluations VALUES(54635,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(54635,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(54635,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(54635,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(54635,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(54635,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(54635,36,'cp 114','c6d5 c4b3 b8c6 d2d4 e5e4 b1c3 e4f3 f1e1 f8e7 c3d5 e8f8 d1f3 c6d4 f3c3 d4b3 c3b3 e7d6 c1f4 g7g5 f4g5 f6g5 b3c3 c8f5 c3h8 d6h2 g1h2 d8d5 a1d1 d5f7 h8d4 a8e8 e1e8 f7e8 h2g1 b7b6 d4d5 g8h6 d5d6 f8g7 d6d8 e8d8 d1d8 h6f7 d8a8 f5c2 a8a7 c2d3 b2b3 g7f6 a7b7 b6b5 a2a4 b5a4 b3a4');
INSERT INTO evaluations VALUES(4727,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(4727,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(4727,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(4727,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(4727,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(4727,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(4727,36,'cp 120','d4c6 b7c6 d1d4 b4f8 e1c1 g8f6 e4e5 f6d5 c3d5 c6d5 h2h4 a6a5 h1h3 c8a6 f1a6 a8a6 h3f3 a6c6 d4f4 f7f5 e5f6 d8f6 f4f6 g7f6 f3f6 f8g7 f6f4 g7h6 f4f3 h6e3 f3e3 e8g8 d1d2 c6c4 g2g3 a5a4 b2b3 a4b3 a2b3 c4c7 f2f4 c7a7 d2d4 f8a8 e3e5 a7a1 c1d2 g8h8 e5h5');
INSERT INTO evaluations VALUES(4887,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4887,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(4887,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(4887,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(4887,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(4887,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(4887,36,'cp 25','b5c6 d7c6 d2d4 e5d4 d1d4 f7f6 c3e2 g8e7 c1f4 c8g4 f4d6 d8d6 e1c1 d6d4 d1d4 e8g8 h2h3 c6c5 d4d3 g4f3 d3f3 a8d8 h1d1 d8d1 c1d1 f8d8 f3d3 d8d3 c2d3 g8f7 f2f4 e7c6 d1d2 f7e7 h3h4 b7b5 h4h5 c6d4 e2d4 c5d4 d2e2 c7c5 b2b3 g7g6 h5g6 h7g6 e2f3 a6a5 f3g4 a5a4 b3a4');
INSERT INTO evaluations VALUES(8977,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(8977,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(8977,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(8977,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(8977,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(8977,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(8977,36,'cp 48','e2e3 e8g8 a2a3 b4c3 c2c3 c6e7 b2b3 b7b6 c1b2 c8b7 c3c2 d5c4 f1c4 a8c8 c2e2 e7g6 e1g1 d8e7 f1d1 f8d8 h2h3 f6e4 a1c1 c7c5 d4c5 e4c5 b3b4 d8d1 e2d1 c8d8 d1e2 c5e4 c4b5 h7h6 c1d1 b7d5 f3d2 e4d2 d1d2');
INSERT INTO evaluations VALUES(8975,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(8975,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(8975,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(8975,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(8975,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(8975,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(8975,36,'cp 57','e8g8 a2a3 b4c3 c2c3 c6e7 f1e2 b7b6 e1g1 c8b7 c4d5 e7d5 c3c2 a8c8 b2b4 c7c5 d4c5 b6c5 b4c5 d8e7 c1d2 c8c5 c2b2 f8c8 f1c1 c5c1 a1c1 c8c1 b2c1 f6e4 d2e1 e7c7 c1c7 d5c7 f3d4 e6e5 d4b3 b7d5 b3d2 e4d2 e1d2 e5e4 d2b4');
INSERT INTO evaluations VALUES(3675,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(3675,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(3675,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(3675,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(3675,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(3675,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(3675,36,'cp 58','a2a3 b4c3 c2c3 c6e7 f1e2 b7b6 e1g1 d5c4 c3c4 a7a5 c4c2 c8b7 f1d1 d8b8 b2b3 f8c8 c2b2 b7d5 f3e5 c7c5 d4c5 e7g6 e5g6 h7g6 c5b6 b8b6 c1d2 b6b7 e2f1 b7b3 b2b3 d5b3 d1c1 c8d8 d2e1 a5a4 f2f3 f6d5 e1d2 e6e5 a1b1 d5f4');
INSERT INTO evaluations VALUES(9420,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9420,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9420,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9420,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9420,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9420,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9420,36,'cp 0','a7a5 e1g1 e8g8 h2h3 h7h6 f1e1 c8e6 b1d2 c5a7 c4b3 f8e8 d2c4 f6h5 c4a3 h5f6');
INSERT INTO evaluations VALUES(7159,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7159,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7159,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7159,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7159,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7159,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(7159,36,'cp 16','e1g1 e8g8 h2h3 h7h6 f1e1 c8e6 b1d2 f6h7 c4e6 f7e6 d2b3 c5b6 c1e3 b6e3 e1e3 f8f7 d3d4 h7g5 d1e2 g5f3 e3f3 f7f3 e2f3 d8h4 f3e3 h4g5 e3g5 h6g5 f2f3 g8f7 g2g3 e5d4 b3d4 c6e5 g1g2 c7c6 a1b1');
INSERT INTO evaluations VALUES(11029,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(11029,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(11029,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(11029,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(11029,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(11029,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(11029,36,'cp 140','f3g5 d7d5 e4d5 c6a5 c4b5 c7c6 d5c6 e8g8 c6b7 c8b7 d2d3 h7h6 g5e4 c5b6 b2b4 a5c6 c2c3 a7a5 e4c5 a5b4 c5b7 d8c7 a2a4 c7b7 c3b4 b6d4 a1a3 c6b4 b1c3 b7c7 c1d2 b4d5 c3d5 e7d5 a4a5 d5c3 d2c3 d4c3 a5a6 c3d4 a3b3 f8b8 g2g3 b8d8 b5c4 c7c6 b3b1');
INSERT INTO evaluations VALUES(13005,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13005,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13005,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13005,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13005,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13005,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(13005,36,'cp 16','f1e1 f8g7 c2c3 e8g8 a2a4 h7h6 c1e3 d8b6 b1a3 c8g4 c4b5 a7a6 a3c4 b6c7 b5c6 b7c6 h2h3 g4e6 c4d2 f8b8 b2b3 f6h5 d3d4 c5d4 f3d4 c7d7 b3b4 c6c5 d4e6 d7e6 b4c5 d6c5 d1c2 b8c8 a1b1 h5f6 c3c4 a8b8 b1b8 c8b8 e3c5');
INSERT INTO evaluations VALUES(13303,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(13303,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(13303,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(13303,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(13303,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(13303,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(13303,36,'cp 12','d2d4 c5d4 f3d4 e7e5 d4f3 h7h6 e1g1 f8e7 f1e1 e8g8 e2f1 c8g4 h2h3 g4h5 f1e2 a8c8 a2a3 d8c7 f3h2 h5e2 e1e2 c6d4 e2e1 c7d7 a3a4 c8c6 c1e3 c6c4 e3d4 c4d4 d1e2 f8c8 h2f1 e7d8 f1g3');
INSERT INTO evaluations VALUES(13368,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13368,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13368,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13368,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13368,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13368,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(13368,36,'cp 0','e7e5 d4c6 b7c6 f1d3 f8e7 e1g1 e8g8 d1e2 d8c7 c1d2 c8e6 a1c1 g8h8 g1h1 f6d7 b2b3 d7c5 f2f4 e5f4 d2f4 a8e8 c1d1 c7a5 f4d2 a5c7');
INSERT INTO evaluations VALUES(14067,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(14067,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(14067,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(14067,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(14067,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(14067,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(14067,36,'cp 129','d4d5 a7a6 b5e2 c6b8 c2c4 g7g6 c1e3 f8g7 c4c5 d6c5 e3c5 c8g4 a2a4 b8d7 c5a3 e8g8 e1g1 f8e8 b1d2 e7c8 h2h3 g4f3 e2f3 g7h6 a1c1 c8d6 c1c3 a6a5 d2c4 h6f8 d1c2 d6c4 a3f8 d7f8 c3c4 c7c6 f1d1 c6d5 e4d5 d8d6 c2b3');
INSERT INTO evaluations VALUES(14061,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(14061,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(14061,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(14061,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(14061,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(14061,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(14061,36,'cp 122','a7a6 b5e2 c6b8 c2c4 g7g6 c1e3 b7b6 d1a4 c8d7 a4c2 d7g4 h2h4 f8g7 b1c3 b8d7 e1c1 h7h6 f3e1 g4e2 c2e2 f7f5 f2f3 f5f4 e3f2 e8g8 c1b1 d8e8 e1d3 g8h7 a2a3 h6h5 d3b4 f8f7 b1a1 e7g8 b4d3 g8f6 e2c2 a8b8 c2a4 d7c5 a4e8 f6e8 f2c5');
INSERT INTO evaluations VALUES(11284,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(11284,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(11284,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(11284,33,'cp 103',NULL);
INSERT INTO evaluations VALUES(11284,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(11284,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(11284,36,'cp 128','b5e2 c6b8 c2c4 g7g6 c1e3 f8g7 c4c5 d6c5 e3c5 c8g4 a2a4 e7c8 e1g1 g4f3 e2f3 c8d6 b1d2 e8g8 c5a3 a6a5 f3e2 f8e8 d1c2 b8d7 f1d1 d7f6 e2d3 h7h5 b2b3 c7c6 d5c6 a8c8 a1c1 c8c6 d2c4');
INSERT INTO evaluations VALUES(17135,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(17135,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(17135,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(17135,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(17135,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(17135,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(17135,36,'cp 39','c6a5 c1e3 a5c4 d3c4 d7d6 d1e2 c8e6 e1c1 d8d7 f3d2 e7c6 c3d5 e8c8 b2b3 c8b8 f2f3 f6f5 e4f5 e6f5 e2f2 f8e7 c1b2 h8f8 d5c3 b7b6 d2e4 d7e6 d1d2 f5e4 c3e4 d6d5 c4d5 d8d5 d2d5 e6d5 f2e2 d5e6');
INSERT INTO evaluations VALUES(23568,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(23568,31,'cp 227',NULL);
INSERT INTO evaluations VALUES(23568,32,'cp 238',NULL);
INSERT INTO evaluations VALUES(23568,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(23568,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(23568,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(23568,36,'cp 224','b5c6 d7c6 f3e5 c6e8 b1c3 d8c8 c1f4 f8b4 f4d2 b7b6 c3d5 b4d2 e1d2 c8b7 c2c4 f7f6 e5d3 e8f7 h1e1 a8e8 b2b3 g8e7 a1d1 e7g6 g2g3 a7a5 f2f3 g6e7 d5e7 e8e7 d3f4 g7g5 f4d5 e7e6 d5c3 f7g6 d2c1 f6f5 e4f5 e6e1 d1e1 g6f5');
INSERT INTO evaluations VALUES(23567,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(23567,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(23567,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(23567,33,'cp 229',NULL);
INSERT INTO evaluations VALUES(23567,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(23567,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(23567,36,'cp 236','d7c6 f3e5 c6e8 c1f4 d8c8 b1c3 f8b4 f4d2 b7b6 c3d5 b4d2 e1d2 f7f6 e5d3 c8b7 c2c4 e8f7 a1e1 h7h5 h2h4 c7c6 d5e3 g8e7 e1d1 a8d8 d2c3 e7g6 b2b3 b7c7 e3f5 h8g8 g2g3 f7e6 h1e1 e6d7 f5e3 g8e8 f2f3 g6e5 d3e5 e8e5 c3c2');
INSERT INTO evaluations VALUES(23559,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(23559,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(23559,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(23559,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(23559,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(23559,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(23559,36,'cp 231','f3e5 c6e8 b1c3 d8c8 c1f4 f8b4 f4d2 b7b6 c3d5 b4d2 e1d2 c8b7 c2c4 f7f6 e5d3 e8f7 a1e1 c7c6 d5e3 g8e7 h2h4 a8d8 d2c3 h7h5 e1d1 a7a5 b2b3 b7c7 f2f3 d8d7 d1d2 e7g6 g2g3 h8d8 h1d1 b6b5 c4c5 f7e6 f3f4');
INSERT INTO evaluations VALUES(23558,30,'cp 232',NULL);
INSERT INTO evaluations VALUES(23558,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(23558,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(23558,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(23558,34,'cp 233',NULL);
INSERT INTO evaluations VALUES(23558,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(23558,36,'cp 224','c6e8 c1f4 d8c8 b1c3 f8b4 f4d2 b7b6 c3d5 b4d2 e1d2 f7f6 e5d3 c8b7 c2c4 e8f7 a1e1 h7h5 h2h4 c7c6 d5e3 g8e7 d2c3 b7c7 e1d1 a8d8 b2b3 e7g6 e3f5 h8g8 g2g3 f7e6 h1e1 a7a5 d1d2 g6e5 d3e5 d8d2 c3d2 f6e5 f5e3');
INSERT INTO evaluations VALUES(25340,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(25340,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(25340,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(25340,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(25340,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(25340,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(25340,36,'cp 82','d4d5 g4f3 e2f3 c6d4 c1e3 d4f3 d1f3 f8e7 h2h4 d8d7 e3g5 d7g4 f3e2 g4e2 e1e2 h7h5 g5f6 e7f6 a2a4 g7g6 a4a5 a7a6 c3d1 f6e7 d1e3 e7f8 a1a3 f8h6 e3d1 h8f8 f2f3 e8c8 d1f2 f7f5 a3b3 h6f4 f2h3 f5e4 f3e4 f8f7 b3f3 d8f8');
INSERT INTO evaluations VALUES(25285,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(25285,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(25285,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(25285,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(25285,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(25285,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(25285,36,'cp 64','g4f3 e2f3 c6d4 c1e3 d4f3 d1f3 f8e7 f3e2 f6d7 e2b5 e7g5 e3g5 d8g5 e1g1 b7b6 b5c4 g5d8 c3b5 e8g8 c4c7 d8c7 b5c7 a8c8 c7b5 c8c2 b2b3 f7f5 e4f5 f8f5 b5d6 f5f4 g2g3 f4d4 a1c1 c2c1 f1c1 d4d5 c1c8 d7f8 d6c4 b6b5 c4e3 d5d2 c8b8 d2a2 b8b5 e5e4 b5b4 a2a1 g1g2 a1a3 e3c4 a3a1 b4b8');
INSERT INTO evaluations VALUES(25541,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(25541,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(25541,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(25541,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(25541,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(25541,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(25541,36,'cp 61','f8e7 f1e1 e8g8 c1f4 h7h6 a2a4 a7a6 a4a5 c6b4 h2h3 f8e8 d1d2 c7c6 d4f3 d6d5 c4b3 d5e4 f3e5 d7e6 b3e6 f7e6 d2c1 e7c5 c3e4 f6e4 e1e4 d8d5 e4e1 e8f8 f4g3 a8d8 e1e2 d8e8 c1b1 f8f5 e5f3');
INSERT INTO evaluations VALUES(30019,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(30019,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(30019,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(30019,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(30019,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30019,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(30019,36,'cp -11','h7h6 g5f4 a7a6 f1d3 c7c5 d4c5 f8c5 h2h3 b8c6 e1g1 e8g8 f1e1 a8c8 a2a3 f6d7 d3f1 d8f6 f4g3 c6d4 f3e5 f6d8 e5d7 d8d7 d1d2 f8e8 g3e5 d4c6 e5g3 d5d4 c3e2 b7b5 a1d1 c8d8 e2f4 e6c4 e1e8 d8e8 f4d3 c5d6 d1e1');
INSERT INTO evaluations VALUES(30794,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(30794,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(30794,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(30794,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(30794,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30794,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30794,36,'cp 0','g8f6 f2f3 e7e6 e2e4 f8b4 a2a3 b4c3 b2c3 d7d5 e4e5 f6g8 f1b5 c7c6 b5d3 c6c5 d3b5 b7c6 b5d3');
INSERT INTO evaluations VALUES(36768,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36768,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36768,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36768,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(36768,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36768,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36768,36,'cp 0','b7b6 a2a4 c8b7 c1f4 a7a5 f1e1 f6h5 f4c1 h5f6');
INSERT INTO evaluations VALUES(44418,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(44418,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44418,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(44418,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(44418,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(44418,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44418,36,'cp 15','c5d4 f3d4 d7d5 d4f3 c8e6 e2e3 f8g7 f3g5 e8g8 g5e6 f7e6 g2g3 b8a6 f1h3 a6c5 e1g1 d5c4 d1d4 f6e4 d4d8 a8d8 c3e4 c5e4 h3e6 g8h8 a2a4 b7b6 e6c4 e4d2 c1d2 d8d2 a1a2 g7b2 a4a5 b6a5 a2a5 f8c8 c4e6 c8c7 h2h4 h8g7 h4h5 b2c3 a5a6 g6h5');
INSERT INTO evaluations VALUES(44442,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44442,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(44442,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(44442,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(44442,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44442,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(44442,36,'cp 26','f3d4 d7d5 d4f3 c8e6 e2e3 f8g7 f3g5 e8g8 g5e6 f7e6 f1e2 d5c4 e2c4 d8d1 e1d1 g8f7 d1e2 f8c8 c4d3 b8c6 c1d2 c6e5 h1c1 c8d8 d3c2 e5c4 d2e1 f6d5 c2b3 a8c8 a1b1 g7c3 e1c3 c4a3 b1a1 d5c3 c1c3 c8c3 b2c3 b7b6');
INSERT INTO evaluations VALUES(44437,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(44437,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(44437,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44437,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(44437,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(44437,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(44437,36,'cp 21','d7d5 d4f3 c8e6 e2e3 f8g7 f3d4 d5c4 d4e6 d8d1 e1d1 f7e6 f1c4 e8f7 c1d2 b8d7 d1e2 h8c8 c4d3 d7e5 a1b1 e5d3 e2d3 c8c6 h1c1 a8c8 a2a4 g6g5 h2h3 a7a6 f2f3 h7h5 c3a2 g5g4 c1c6 c8c6 h3g4 h5g4 a2c3 g4f3 g2f3 f6d7 b2b3 c6d6 d3e2 d7c5');
INSERT INTO evaluations VALUES(45714,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(45714,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(45714,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(45714,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(45714,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(45714,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(45714,36,'cp 24','f3e5 f8b4 c1d2 b4e7 e2e3 b7b5 e5c6 b8c6 g2c6 c8d7 c6a8 d8a8 e1g1 e6e5 f2f3 e5d4 e3d4 e8g8 d2a5 a8c6 a2a3 c6a6 d1d2 h7h6 f1e1 d7e6 a5b4 e7b4 d2b4 f6d5 b4d2 b5b4 d2e2 f8c8 b1d2 a6b6 d2e4 b6d4 e2f2 d4f2 g1f2 c4c3 b2c3 b4b3 a1c1 d5b6 c1b1 g8h8 e1e2');
INSERT INTO evaluations VALUES(45690,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(45690,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(45690,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(45690,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(45690,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(45690,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(45690,36,'cp 22','f8b4 c1d2 b4e7 e2e3 b7b5 e5c6 b8c6 g2c6 c8d7 c6a8 d8a8 f2f3 e6e5 e1g1 e5d4 e3d4 e8g8 d2a5 f6d5 d1d2 b5b4 a2a3 b4b3 b1c3 h7h6 f1e1 e7g5 d2f2 d5c3 a5c3 g5f6 e1e3 a7a5 f2e2 a8d5 a1d1 d7f5 e3e8 f8e8 e2e8 g8h7 e8e2 f5c2 d1d2 a5a4 g1f2 c2d3 d2d3 c4d3 e2d3 h7h8');
INSERT INTO evaluations VALUES(47873,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(47873,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(47873,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(47873,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(47873,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(47873,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(47873,36,'cp -1','f8e7 e1g1 c8e6 f3g5 e6g4 d3e2 g4e2 d1e2 b8c6 g5f3 e8g8 e2c4 a7a6 d2d4 e5d4 f3d4 c6d4 c4d4 f8e8 c1f4 d8d7 a1d1 a8d8 h2h3 c7c6 f1e1 b7b5 a2a3 h7h6 f4g3 a6a5 f2f3 d6d5 e4d5 f6d5 c3d5 c6d5');
INSERT INTO evaluations VALUES(49255,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(49255,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(49255,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(49255,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(49255,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(49255,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(49255,36,'cp -23','c7c6 d1e2 d7d5 e4d5 c6d5 c4b5 b8c6 e2e5 f8e7 g1f3 e8g8 b5c6 b7c6 e1g1 e7d6 e5e2 f8e8 e2d1 c6c5 f1e1 e8e1 d1e1 d5d4 b1d2 c8b7 b2b3 d6c7 c1a3 d8d6 d2c4 d6d7 f3e5 d7d5 e5f3 d5h5 e1e7 a8c8 f3e5 f6d5 e7d7 d5f4 d7g4 h5g4');
INSERT INTO evaluations VALUES(52893,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(52893,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(52893,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(52893,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(52893,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(52893,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(52893,36,'cp 193','e2e4 e7e5 d4e5 d8d1 e1d1 f7f6 e5f6 g8f6 f2f3 c8e6 g1e2 e6f7 e2d4 f8c5 c1e3 e8g8 d1e2 b8d7 d4f5 a7a5 e3c5 d7c5 e2e3 g8h8 f1e2 a5a4 h1c1 f6d7 f5d4 a8c8 c3d1 c8e8 d1f2 f7d5 c1c3 b7b6 a1c1 g7g6 g2g3 d5a2 d4c6 a2b3 e2b5');
INSERT INTO evaluations VALUES(52897,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(52897,31,'cp 198',NULL);
INSERT INTO evaluations VALUES(52897,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(52897,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(52897,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(52897,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(52897,36,'cp 199','e7e5 d4e5 d8d1 e1d1 f7f6 e5f6 g8f6 f2f3 f8c5 g1e2 f6d7 c3a4 c5f2 e2f4 b7b5 d1e2 f2h4 a4c3 d7e5 g2g3 h4e7 h2h4 b8a6 c1e3 e5c4 e3d4 e8g8 e2f2 a6b4 a1c1 c4b2 c3b5 f8d8 f1e2 c6b5 d4b2 b4a2 c1c7');
INSERT INTO evaluations VALUES(54723,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(54723,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(54723,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(54723,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(54723,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(54723,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(54723,36,'cp 104','d2d4 c8g4 h2h3 g4f3 e2f3 d8d7 f3f4 e7e6 f1g2 g8e7 e1g1 b8c6 c2c3 a7a5 b2b3 e8f7 f1e1 e7f5 a2a3 g7g6 c1b2 c6e7 c3c4 a8d8 b1d2 e7c6 c4d5 e6d5 d2b1 c6e7 b1c3 f5g7 b3b4 a5b4 a3b4 c7c6 c3a4');
INSERT INTO evaluations VALUES(54717,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(54717,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(54717,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(54717,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(54717,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(54717,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(54717,36,'cp 90','c8g4 h2h3 g4f3 e2f3 e7e6 f1g2 g8e7 e1g1 d8d7 f3f4 c7c6 f1e1 e8f7 b1d2 a7a5 b2b3 h7h5 c2c3 b8a6 d2f3 a6c7 a2a4 e7f5 a1a2 f8d6 a2e2 g7g6 c1d2 b7b6 d1c2 a8e8 h3h4 f5g7 g2h3 g7f5 e1a1 d6e7 e2e1');
INSERT INTO evaluations VALUES(57202,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(57202,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(57202,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(57202,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(57202,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(57202,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(57202,36,'cp -50','e5e4 f3d4 c7c5 d4c2 d7d5 d2d3 f7f5 g2g3 g8f6 f1g2 b8c6 e1g1 c8e6 d3e4 d5e4 d1d8 a8d8 b1d2 c5c4 f2f3 f8c5 e2e3 e8g8 c2d4 e6d5 d4f5 e4f3 g2f3 d5f3 d2f3 f8e8 f5d4 f6g4 d4c6 b7c6 f3d4 g4e5 b2b3 d8d7 g1g2 c5b6 d4f3 e5d3 b3c4 d3c1');
INSERT INTO evaluations VALUES(889,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(889,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(889,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(889,33,'cp 273',NULL);
INSERT INTO evaluations VALUES(889,34,'cp 274',NULL);
INSERT INTO evaluations VALUES(889,35,'cp 257',NULL);
INSERT INTO evaluations VALUES(889,36,'cp 286','f1b5 a5b4 d4d5 a7a6 b5a4 b7b5 c3b5 b4e4 d1e2 a6b5 e2e4 f6e4 a4b5 a8b8 b5c6 c8d7 c6d7 e8d7 b2b3 b8e8 d2e3 e7e6 d5e6 e8e6 g1f3 f8b4 e1f1 d7c8 g2g3 e6c6 a2a3 b4e7 f3d4 c6a6 f1g2 h8d8 d4b5 c7c6 b5d4 e7f6 a1d1 e4c3 d1d3 c3b5 d4b5 c6b5 d3d8 f6d8 e3d4 f7f6 d4b2 d8e7 h1d1 e7a3 b2a3 a6a3');
INSERT INTO evaluations VALUES(2638,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(2638,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(2638,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(2638,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(2638,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(2638,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2638,36,'cp 18','c1e3 c5e3 f2e3 d7d6 d3d4 e5d4 e3d4 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 c6e7 f3e5 f7f6 e5c4 c8e6 d1d3 d8d7 a1e1 e6d5 c4d2 d5e4 d2e4 b7b6 c2c4 e7c6 d4d5 c6e5 d3d4 a8e8 b2b4 a5b4 a3b4 e5g6 c4c5 e8e5 e1d1 f8e8 c5c6 d7f7 e4c3 g8h8 d4c4 e8e7');
INSERT INTO evaluations VALUES(3238,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(3238,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(3238,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(3238,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(3238,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(3238,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(3238,36,'cp 46','f1e2 e7e6 f3f4 c6d4 e3d4 b7b6 e1g1 d8e7 g1h1 c8b7 e2f3 e6e5 f4e5 d6e5 d4f2 e7b4 d1b3 b4b3 a2b3 f8d8 f2g3 a7a5 g3e5 f6d7 e5g7 g8g7 c3d5 b7d5 e4d5 d7c5 a1a3 d8e8 h1g1 f7f5 g2g4 g7f6 g4f5 g6f5 d5d6');
INSERT INTO evaluations VALUES(3237,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(3237,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(3237,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(3237,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(3237,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(3237,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(3237,36,'cp 45','e7e6 f3f4 c6d4 e3d4 b7b6 e1g1 c8b7 e2f3 d8e7 g1h1 e6e5 f4e5 d6e5 d4f2 e7b4 d1b3 b4b3 a2b3 f8d8 c3d5 f6d5 e4d5 d8e8 a1e1 f7f5 g2g4 e5e4 f3g2 b7c8 e1e2 g7f8 f1a1 c8d7 g4f5 d7f5 f2d4 f8b4 h1g1 b4c5 d4c5 b6c5 a1a5 a8b8 a5c5 b8b3 c5c6 b3d3 c6c7');
INSERT INTO evaluations VALUES(4710,30,'cp -315',NULL);
INSERT INTO evaluations VALUES(4710,31,'cp -318',NULL);
INSERT INTO evaluations VALUES(4710,32,'cp -316',NULL);
INSERT INTO evaluations VALUES(4710,33,'cp -322',NULL);
INSERT INTO evaluations VALUES(4710,34,'cp -287',NULL);
INSERT INTO evaluations VALUES(4710,35,'cp -281',NULL);
INSERT INTO evaluations VALUES(4710,36,'cp -286','h7h6 g5e3 d7d6 b1c3 a7a5 e1g1 f7g8 h2h3 c8e6 d1d2 c5e3 f2e3 c6b8 c3e2 b8d7 f1f2 f6h5 d2c3 g8h8 a1f1 d8b8 g1h2 e6g8 b2b3 g8e6 c3d2 b8d8 g2g4 h5f6 e2g3 b7b5 a2a4 b5a4 b3a4 f6h7 d3d4 d8h4 g3f5 e6f5 e4f5');
INSERT INTO evaluations VALUES(4706,30,'cp -300',NULL);
INSERT INTO evaluations VALUES(4706,31,'cp -290',NULL);
INSERT INTO evaluations VALUES(4706,32,'cp -289',NULL);
INSERT INTO evaluations VALUES(4706,33,'cp -284',NULL);
INSERT INTO evaluations VALUES(4706,34,'cp -275',NULL);
INSERT INTO evaluations VALUES(4706,35,'cp -280',NULL);
INSERT INTO evaluations VALUES(4706,36,'cp -271','g5e3 d7d6 e1g1 f7g8 b1c3 c5e3 f2e3 c8e6 d1d2 d8e7 h2h3 c6d8 c3e2 d8f7 e2g3 a8f8 d2a5 a7a6 f1f2 f7g5 g3f5 e7d7 a1f1 e6f5 f2f5 d7c6 c2c4 g8h7 a5d2 b7b5 c4b5 a6b5 b2b4 c6b7 d2c2 h7g8 g1h1 c7c6 c2b3 g8h7 a2a3 b7d7 b3b2 f8g8');
INSERT INTO evaluations VALUES(5012,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(5012,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(5012,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(5012,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(5012,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(5012,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5012,36,'cp 22','d1g4 g8f6 g4g3 e8g8 e4e5 f6e8 f1d3 d7d5 e1g1 g7g6 c1g5 c5e7 g5e7 d8e7 b2b3 a7a5 c3a4 c8a6 d3a6 a8a6 g3e3 e8g7 g2g4 c6c5 h2h3 a6c6 c2c4 c6c8 a4b2 h7h5 a1d1 h5g4 h3g4 f7f5 e5f6 e7f6 b2d3 f6h4 e3g3 h4g3 f2g3 f8f1 g1f1');
INSERT INTO evaluations VALUES(6670,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(6670,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(6670,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(6670,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(6670,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(6670,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(6670,36,'cp -24','b4e7 g1f3 g8f6 f1d3 e8g8 e1g1 f8e8 e3f4 d7d6 h2h3 e7f8 f1e1 h7h6 c3c4 c6e5 f3e5 d6e5 f4f3 c7c6 b1c3 c8e6 d3e2 f8b4 e1d1 f6d7 a2a3 b4c5 c1e3 c5e3 f3e3 d8b6 b2b4 b6e3 f2e3 d7f6 c4c5 a8d8');
INSERT INTO evaluations VALUES(6880,30,'cp 252',NULL);
INSERT INTO evaluations VALUES(6880,31,'cp 258',NULL);
INSERT INTO evaluations VALUES(6880,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(6880,33,'cp 268',NULL);
INSERT INTO evaluations VALUES(6880,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(6880,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(6880,36,'cp 258','g8e7 c1d2 d7d5 e5d6 d8d6 a2a3 b4c5 d1f3 d6g6 f3f4 c7c6 f2f3 g6d6 f4g3 d6g3 h2g3 c8e6 c3e2 f7f6 d2b4 c5b6 e2f4 e6f7 f4h5 c6c5 f1b5 e7c6 b4c3 e8g8 b5c6 b7c6 g3g4 a8d8 h5g3 d8d7 g3f5 f8d8 e1f2 c5c4 f2g3 h7h6 a1e1 f7e6 f3f4 e6f5 e4f5 b6d4 c3d4');
INSERT INTO evaluations VALUES(9347,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(9347,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(9347,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(9347,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(9347,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(9347,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9347,36,'cp -17','c8e6 c4e6 f7e6 c1e3 b7b6 f3e2 c5e3 e2e3 f6d7 b1d2 d8f6 g1f3 h7h6 d3d4 e8g8 e1g1 f8f7 d4e5 c6e5 f3e5 f6e5 d2b3 a8f8 b3d4 g8h8 a1d1 a7a5 f2f3 e5f6 b2b4 g7g5 e3e2 d7e5 g1h1 e5g6 a2a4');
INSERT INTO evaluations VALUES(10273,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(10273,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(10273,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(10273,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(10273,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(10273,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(10273,36,'cp -80','f6e4 a4c5 e4c5 d2d4 e5d4 f3d4 c6e5 c4d5 c7c6 b2b4 c6d5 b4c5 d8a5 d1d2 a5c5 a2a4 e5c4 c1a3 c4d2 a3c5 d7d6 c5d6 d2c4 d6f4 e8g8 e1g1 c8d7 a4a5 a7a6 f1b1 d7c6 h2h4 h7h6 d4c6 b7c6 b1b7 f7f6 b7c7 f8c8 c7b7 c6c5 h4h5 c8e8 g2g3 d5d4 b7c7 a8c8');
INSERT INTO evaluations VALUES(12439,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(12439,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12439,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12439,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12439,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(12439,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12439,36,'cp 0','d7d5 d1e2 f8e7 e1g1 a7a6 a2a4 e8g8 c1f4 h7h6 c2c3 f6h5 f4c1 h5f6');
INSERT INTO evaluations VALUES(13606,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13606,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13606,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13606,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13606,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13606,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13606,36,'cp 0','d1e2 f8e7 e1g1 a7a6 c1f4 e8g8 a2a4 h7h6 f1e1 c5c4 e4e5 c4d3 c2d3 f6d7 d3d4 d8b6 f4e3 c6a5 b1d2 b6b2 e1c1 d7b6 c1b1 b2c3 b1c1');
INSERT INTO evaluations VALUES(16880,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(16880,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(16880,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(16880,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(16880,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(16880,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(16880,36,'cp 23','e1e5 f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 d7d5 a2a4 a7a5 f1d3 g7g6 b1d2 e8g7 d2f3 c8f5 d3f5 g7f5 h2h3 f5g7 e1d2 d8d7 f4e5 f6e7 e5g7 g8g7 f3e5 d7d6 g2g3 a8e8 g1g2 e7f6 e5d3 c7c6 a1e1 e8e6 d3c5 e6e1 d2e1');
INSERT INTO evaluations VALUES(16879,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(16879,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(16879,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(16879,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(16879,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(16879,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(16879,36,'cp 18','f8e7 b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 d7d5 f1d3 g7g6 b1d2 e8g7 d2f3 c8f5 d3f5 g7f5 e1d2 a7a5 a1e1 a5a4 g2g3 c7c6 h2h4 h7h5 f3e5 f5d6 g1g2 d6c4 e5c4 d5c4 d2e2 a4a3 e2c4');
INSERT INTO evaluations VALUES(10967,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10967,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(10967,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(10967,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(10967,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(10967,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(10967,36,'cp 20','b5f1 e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 d7d6 c1e3 c8f5 e1d2 c7c6 b1a3 c6c5 h2h3 h7h6 a1e1 f5g6 f1d3 g6d3 d2d3 a7a6 e3f4 b7b5 a3c2 f6g5 f4h2 d8d7 c2e3 e8f6 a2a4');
INSERT INTO evaluations VALUES(10968,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10968,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(10968,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(10968,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(10968,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(10968,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10968,36,'cp 16','e8g8 d2d4 e7f6 e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 c7c6 b1a3 c6d5 a3c2 b7b6 g2g3 c8b7 f1g2 e8d6 e1d1 d6e4 c1e3 a7a5 d1d3 b7c6 a1d1 f6e7 c2d4 a8c8 d4c6 d7c6 g2e4 d5e4 d3e2 e7d6 a2a3');
INSERT INTO evaluations VALUES(4615,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(4615,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(4615,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(4615,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(4615,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(4615,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4615,36,'cp 15','d2d4 e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 d7d5 b1d2 c8f5 e1e3 c7c6 a1e1 e8d6 f1d3 f5d3 e3d3 f6e7 d3e2 e7f8 f4d6 f8d6 g2g3 d8d7 d2f3 g8f8 e2d3 f8g8 d3e3 g8f8 f3e5 d6e5 d4e5 f8g8 a2a4 a7a5 e3b6 h7h5 g1g2 g7g6 h2h3');
INSERT INTO evaluations VALUES(4614,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(4614,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(4614,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(4614,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(4614,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(4614,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(4614,36,'cp 21','e7f6 e5e1 f8e8 c1f4 e8e1 d1e1 d6e8 c2c3 d7d5 e1e3 c8f5 b1d2 e8d6 a2a4 a7a5 f1d3 f5d3 e3d3 f6e7 g2g3 d8d7 a1e1 d6e8 d3f3 c7c6 h2h4 e7d6 g1g2 h7h6 e1e2 a8d8 f4e5 d6e5 e2e5 e8d6 f3f4');
INSERT INTO evaluations VALUES(4522,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(4522,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(4522,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(4522,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4522,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(4522,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(4522,36,'cp 20','e5e1 f8e8 c2c3 e8e1 d1e1 d6e8 c1f4 d7d5 b1d2 c8f5 e1e3 c7c6 a1e1 e8d6 e3g3 d6e4 d2e4 f5e4 f4e5 f6e5 g3e5 d8d7 e5f4 h7h6 h2h4 a8e8 b2b4 e4g6 e1e5 f7f6 e5e3 e8e3 f4e3 g6f5 a2a4 h6h5 a4a5');
INSERT INTO evaluations VALUES(4523,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(4523,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(4523,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(4523,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(4523,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4523,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(4523,36,'cp 14','f8e8 c2c3 e8e1 d1e1 d6e8 d4d5 d7d6 c1e3 c8d7 h2h3 f6g5 b1d2 d8e7 f2f4 g5h6 g2g4 g7g6 e1f2 h6g7 a1e1 e8f6 e3d4 e7f8 c3c4 a8e8 b2b4 b7b6 f1g2 a7a5 b4b5 e8e1 f2e1 f8e8 e1e8 f6e8 d4g7 g8g7 g1f2 h7h5 g4g5 d7f5 f2e3 h5h4');
INSERT INTO evaluations VALUES(24839,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(24839,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(24839,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(24839,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(24839,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(24839,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(24839,36,'cp 38','e2e3 e7e6 c2c4 f8d6 f4g5 e8g8 b1c3 d6e7 f1d3 d5c4 d3c4 f6e4 g5e7 c6e7 c3e4 f5e4 e1g1 e4f3 d1f3 c7c6 a1c1 g7g6 f1d1 d8b6 f3e2 h7h5 c4d3 f8d8 c1c5 g8g7 d1c1 d8d7 e2c2 a8d8 g2g3 a7a5');
INSERT INTO evaluations VALUES(34521,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(34521,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(34521,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(34521,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(34521,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(34521,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(34521,36,'cp 147','a7a6 d2d4 g8f6 g2g3 c8g4 f1g2 b8d7 h2h3 g4f3 g2f3 c6b6 e1g1 e7e6 c3a4 b6b5 c2c3 f8e7 d1c2 a6a5 c1f4 e7d6 c3c4 b5a6 f4e3 a8d8 a1d1 c7c6 b2b3 e8g8 a4b2 b7b5 b2d3 a5a4 f1e1 f8e8 c4c5 d6c7 b3b4 e6e5');
INSERT INTO evaluations VALUES(35329,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(35329,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(35329,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(35329,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(35329,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(35329,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(35329,36,'cp 130','e2e4 e7e5 c1e3 f8e7 g1f3 g8f6 h2h3 e8g8 f1d3 f8e8 d4d5 a7a5 e1g1 b8a6 a2a3 c8d7 f3d2 h7h6 a1c1 f6h7 d2f3 h7f6 d3b1 a6c5 b2b4 a5b4 a3b4 c5a6 b4b5 a6c5 f3d2 c6d5');
INSERT INTO evaluations VALUES(38359,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(38359,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(38359,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(38359,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38359,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(38359,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(38359,36,'cp 22','g8f6 d1e2 c7c6 g1f3 e8g8 e4e5 f6d5 e1g1 d6e5 d4e5 c8g4 b1d2 b7b5 c4b3 b8d7 h2h3 g4f5 b3d5 c6d5 f1e1 d8b6 d2b3 f5e4 f3g5 d7e5 g5e4 d5e4 e2e4 a7a5 c1e3 b6c6 e3d4 c6e4 e1e4 a5a4 b3c5 e5c6 d4g7 g8g7 c3c4 b5c4 e4c4');
INSERT INTO evaluations VALUES(41246,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(41246,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41246,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(41246,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(41246,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(41246,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(41246,36,'cp 34','d1e2 e8g8 g1f3 c7c6 b1d2 f6h5 e1g1 b7b5 c4b3 a7a5 a2a4 b5a4 a1a4 b8d7 a4a3 c6c5 f1e1 c5d4 c3d4 d7b6 e2d1 a5a4 b3a4 b6a4 a3a4 a8a4 d1a4 h5f4 a4a3 d8c7 d2b1 f4e2 e1e2 c7c1 e2e1 c1c4 b1c3 g7d4 c3d5 f8e8 h2h3 c8b7');
INSERT INTO evaluations VALUES(38403,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(38403,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(38403,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(38403,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(38403,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(38403,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(38403,36,'cp 34','g8f6 c2c4 e8g8 e1g1 c8f5 b2b3 c7c6 c1b2 d6d5 e2e3 b8d7 b1c3 f6e4 d1e2 e4c3 b2c3 a7a5 a1c1 d7b6 f3e5 g7e5 d4e5 d5c4 f1d1 d8c7 b3c4 f5e6 c3a1 f8d8 d1e1 a5a4 h2h3 a4a3 g3g4 c6c5 f2f4 b6c4 c1c4');
INSERT INTO evaluations VALUES(41631,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(41631,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(41631,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(41631,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(41631,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(41631,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(41631,36,'cp 9','e1g1 b8c6 c2c3 a7a5 c4b5 c5a7 b1d2 e8g8 f1e1 f6h7 d2f1 f7f5 e4f5 c8f5 f1g3 d8f6 b5c4 g8h8 c1e3 a7e3 e1e3 f5g6 d1d2 h7g5 f3g5 f6g5 c4d5 g6f7 g3e4 g5g6 d5c6 b7c6 c3c4 c6c5 a1e1 a5a4 e3g3 g6h7 e4c3 f7g6');
INSERT INTO evaluations VALUES(41630,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41630,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41630,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41630,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41630,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41630,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41630,36,'cp 0','c5b6 c2c3 g7g5 a2a4 b8c6 d3d4 g5g4 h3g4 c8g4 b1d2 h8g8 b2b4 g4h3 f3e1 a7a5 b4b5 c6b8 d4e5 d6e5 d1b3 d8d7 d2f3 h3g2 e1g2 g8g2 g1g2 d7g4 g2h2 g4h5 h2g1 h5g4');
INSERT INTO evaluations VALUES(44656,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44656,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(44656,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(44656,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44656,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(44656,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44656,36,'cp 0','d5b6 c4b3 d7d6 e5d6 d8d6 d2d4 c5d4 g1f3 b8c6 e1g1 c8e6 f3d4 c6d4 c3d4 e6b3 a2b3 g7g6 b1c3 f8g7 d1f3 e8g8 f3b7 f8b8 b7e4 d6d4 e4e7 b6d5 c3d5 d4d5 a1a7 a8a7 e7a7 d5b3 a7a3 h7h5 h2h4 g7b2 c1b2 b3b2 a3a4 b8b4 a4e8 g8g7 e8e7 b2d4');
INSERT INTO evaluations VALUES(44448,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(44448,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(44448,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(44448,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(44448,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44448,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(44448,36,'cp 6','c4b3 d7d5 e5d6 d8d6 d2d4 c5d4 g1f3 b8c6 f3d4 c6d4 c3d4 c8e6 e1g1 e6b3 a2b3 g7g6 b1c3 f8g7 d1f3 e8g8 f3b7 f8b8 b7e4 d6d4 e4e7 b6d5 c3d5 d4d5 a1a7 a8a7 e7a7 b8b3 a7c7 h7h5 c7c2 d5b7 h2h3 g7b2 c1b2 b3b2 c2c3 b7b5 c3f6 b2b1 f1b1 b5b1 g1h2');
INSERT INTO evaluations VALUES(44451,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(44451,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44451,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44451,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44451,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44451,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44451,36,'cp 0','d7d6 e5d6 d8d6 d2d4 c5d4 g1f3 b8c6 e1g1 c8e6 f3d4 c6d4 c3d4 e6b3 d1b3 e7e6 f1d1 f8e7 d4d5 e8g8 b1c3 f8d8 d5e6 d6e6 d1d8 a8d8 b3e6 f7e6 c1e3 e7f6 e3b6 a7b6 a1d1 d8a8 d1d7 f6c3 b2c3 a8a2 f2f3 a2c2 d7b7 c2c1 g1f2 c1c2');
INSERT INTO evaluations VALUES(44849,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44849,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44849,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(44849,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(44849,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44849,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44849,36,'cp 10','e7f6 c2c3 f8d6 d1e2 d8e7 e2e7 d6e7 f1e2 a7a5 d2d3 b8a6 c1f4 e8g8 h2h4 a6c5 f4e3 f8e8 e1d2 c8g4 h4h5 a8d8 f3d4 g4e2 d2e2 d8d5 g2g4 e7d6 e2d2 c5e6 a1e1 g7g6 a2a4 e6d4 e3d4 e8e1 d2e1');
INSERT INTO evaluations VALUES(45830,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(45830,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(45830,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(45830,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(45830,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(45830,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(45830,36,'cp 15','c2c3 f8d6 d1e2 d8e7 e2e7 d6e7 a2a4 a7a5 e1d1 b8a6 d1c2 a6c7 d2d4 h7h5 f1d3 g7g6 f3d2 c8e6 h1e1 e8f8 h2h4 f8g7 d2c4 c7d5 c4e3 d5e3 c1e3 e7d6 a1d1 h8e8 c3c4 d6b4 e1e2 e6g4 f2f3 g4d7 b2b3');
INSERT INTO evaluations VALUES(48362,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(48362,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(48362,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(48362,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(48362,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(48362,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(48362,36,'cp -36','c7c5 g1f3 b8d7 d4c5 f8c5 f1d3 e8g8 e1g1 c5e7 e3e4 d5e4 c3e4 f6e4 d3e4 d7c5 d1e2 d8c7 f1e1 f8d8 c1d2 e7f6 c2c3 c5b3 a1d1 b3d2 d1d2 d8d2 e2d2 c8d7 d2e3 a7a6 e1d1 g7g6 d1d2 d7a4 g2g3 a8d8');
INSERT INTO evaluations VALUES(46254,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46254,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(46254,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(46254,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(46254,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(46254,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(46254,36,'cp -32','g1f3 b8d7 d4c5 f8c5 f1d3 c5e7 e1g1 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 d7c5 d1e2 d8c7 c1d2 c8d7 d2c3 a7a6 f1d1 f8d8 h2h3 d7b5 e2e3 e7d6 d1d6 d8d6 c3e5 a8d8 e5d6 c7d6 c2c3 c5e4 e3e4 b5c6 e4e3 h7h6 f3e5 d6d5 e5c6 d5c6 g1h1 b7b6 b2b4');
INSERT INTO evaluations VALUES(46255,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(46255,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(46255,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(46255,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(46255,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(46255,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(46255,36,'cp -38','b8d7 d4c5 f8c5 f1d3 e8g8 e1g1 c5e7 e3e4 f6e4 c3e4 d5e4 d3e4 d7c5 d1e2 d8c7 f1e1 f8d8 c1d2 c5e4 e2e4 f7f6 d2b4 e6e5 a1d1 c8e6 b4e7 c7e7 f3h4 g7g6 h2h3 d8d1 e1d1 a8c8 c2c3 c8c4 h4g6 c4e4 g6e7 g8f7');
INSERT INTO evaluations VALUES(50035,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(50035,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(50035,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(50035,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(50035,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(50035,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(50035,36,'cp 83','c7c6 b1c3 d7d5 c4b3 a7a5 a2a3 a5a4 b3a2 d5d4 c3e2 d8d6 a1b1 c8e6 a2e6 d6e6 b2b3 a4b3 b1b3 e6d7 c2c3 c6c5 f3d2 e7c8 c3d4 e5d4 e2f4 b7b5 d1h5 e8d8 h5d5 d7d5 e4d5 a8a5 a3a4 b5b4 f4e6 d8d7 d2c4 a5a4 e1g1 f8e7 c1f4 b8a6 b3b1 h8g8');
INSERT INTO evaluations VALUES(50206,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(50206,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50206,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(50206,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(50206,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(50206,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(50206,36,'cp 0','f8g7 g1f3 e8g8 e2e3 b7b6 f1e2 c7c5 e1g1 c5d4 f3d4 b8c6 c2c4 c8b7 d4c6 d7c6 d1c2 c6c5 b1c3 e7e6 a1d1 d8e7 d1d2 f8d8 f1d1 f6e8 h2h3 b7c6 e2f3 d8d2 c2d2 c6f3 g2f3 e7g5 g1h2 g5h5 h2g2 h5g5');
INSERT INTO evaluations VALUES(50689,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(50689,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(50689,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(50689,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(50689,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(50689,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50689,36,'cp 9','b8c6 d4d3 g8f6 c3d5 f6d7 c2c4 e7e6 d5c3 d7e5 f3e5 c6b4 d3d2 d6e5 d2d8 e8d8 c1g5 d8c7 e1c1 c8d7 g5e3 b4c6 f2f4 f8b4 c1c2 h8d8 a2a3 b4e7 f4e5 d7e8 f1e2 c6e5 b2b4 b7b6 e3f4 e7f6 d1d8 a8d8 b4b5 c7b7 b5a6 b7a7 h1b1 e5c6');
INSERT INTO evaluations VALUES(17593,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(17593,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(17593,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17593,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(17593,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(17593,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(17593,36,'cp 9','d4d3 g8f6 c3d5 f6d7 c2c4 e7e6 d5c3 d7e5 f3e5 d6e5 d3d8 e8d8 c1e3 d8c7 e1c1 c8d7 f1e2 f7f6 h1f1 f8b4 d1d2 c6d4 e3d4 e5d4 d2d4 b4c3 b2c3 e6e5 d4d2 b7b6 f2f3 a6a5 g2g3 h8d8 f3f4 d7e8 f4e5 f6e5 d2d3 d8d3 e2d3');
INSERT INTO evaluations VALUES(53553,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(53553,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(53553,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(53553,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(53553,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(53553,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53553,36,'cp 13','f8e7 g5e7 d8e7 e2e3 g8f6 b1c3 e8g8 d1c2 d5c4 f1c4 c6c5 e1g1 c5d4 f3d4 c8d7 c4e2 f8c8 f1d1 a7a6 a2a3 c8c7 d4f3 b8c6 e3e4 a8c8 c2d2 e6e5 a1c1 h7h6 h2h3 d7e6 c3d5 e6d5 e4d5 c6d4 c1c7 c8c7 f3d4');
INSERT INTO evaluations VALUES(37877,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(37877,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37877,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(37877,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(37877,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37877,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(37877,36,'cp 13','g5e7 d8e7 e2e3 g8f6 b1c3 e8g8 d1c2 d5c4 f1c4 c6c5 e1g1 c5d4 f3d4 c8d7 c4e2 b8c6 d4f3 a8c8 a1c1 f8d8 h2h3 d7e8 a2a3 h7h6 b2b4 f6d7 c2b2 e7f6 f1d1 d7b6 d1d8 c8d8 b4b5 c6e5 a3a4 d8c8 a4a5 e5f3 e2f3 b6c4 b2a1 c4d6');
INSERT INTO evaluations VALUES(53980,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(53980,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(53980,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(53980,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(53980,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(53980,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(53980,36,'cp -68','b8c6 c2c3 g8f6 d2d3 f8e7 h3f4 e8g8 e1g1 c6e5 f3e2 e5c4 d3c4 b7b6 b1a3 d8c7 f4d3 a7a6 f2f3 h7h6 c1e3 c8d7 f1e1 d7c6 e3f2 f6d7 f2g3 f8d8 a1d1 b6b5 d3f2 a8b8 b2b3 d7f8 c4b5 a6b5 c3c4 b5c4 a3c4 f8g6 f3f4');
INSERT INTO evaluations VALUES(54070,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(54070,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(54070,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(54070,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(54070,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(54070,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(54070,36,'cp 29','e2e4 d5d4 c3b5 b8c6 g1f3 e6e5 b2b4 a7a6 b5a3 g8f6 a3c4 f6e4 c4e5 c6b4 f1c4 c8e6 c4e6 f7e6 a2a3 b4d5 d1d4 d8a5 f3d2 d5f6 d4c4 e4c5 e1g1 f8e7 a1b1 a5c7 c1b2 e8g8 f1e1 b7b5 c4a2 f6d5 c2c4 d5f4 d2f3 e7f6 c4b5 a6b5 b2d4 a8d8 a2d2 c5d3 e5d3 f6d4');
INSERT INTO evaluations VALUES(54347,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(54347,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(54347,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(54347,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(54347,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(54347,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(54347,36,'cp 78','b2b3 c4b3 a2b3 b8c6 f1b5 c8d7 e1g1 g8f6 h2h3 f8e7 c2c3 e8g8 b1d2 d7e8 d1e2 a7a5 b5d3 f6h5 f4h2 f7f5 f1c1 h5f6 d3b5 h7h6 b5c6 b7c6 f3e5 c6c5 e5d3 d8b6 d4c5 e7c5 d3c5 b6c5 h2e5 f6d7 e5d4 c5e7');
INSERT INTO evaluations VALUES(56950,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(56950,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(56950,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(56950,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(56950,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(56950,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(56950,36,'cp -20','d7d5 e4d5 e6d5 c4b3 a7a5 b1c3 c7c6 d1e2 d8e7 a2a4 c8g4 d2d4 g4f3 g2f3 e7e2 c3e2 f8d6 h2h4 h6h5 c2c4 d5c4 b3c4 b8d7 c4a2 g8e7 e1f1 d7b6 h1g1 g7g6 e2c3 e7f5 c1g5 d6e7 d4d5 e7g5 h4g5 b6d5 c3d5 c6d5 a2d5 e8g8 d5b7');
INSERT INTO evaluations VALUES(57437,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(57437,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(57437,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(57437,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(57437,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(57437,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(57437,36,'cp 36','b8c6 f1c4 f8c5 c2c3 c5b6 a2a4 d7d6 d2d4 a7a6 e1g1 c8g4 c4e2 g8e7 d4d5 c6b8 b1d2 e8g8 d2c4 b8d7 c1e3 g4f3 e2f3 b6e3 c4e3 g7g6 f3g2 f6f5 e4f5 e7f5 e3c4 d7f6 d1b3 b7b6 a4a5 b6b5 c4d2 d8e7 f1e1 g8h8 c3c4 f5d4 b3d1 b5b4');
INSERT INTO evaluations VALUES(23241,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(23241,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(23241,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(23241,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(23241,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(23241,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(23241,36,'cp 72','f1g2 d7d5 e4d5 d8d5 b1c3 d5d6 e1g1 c8g4 h2h3 g4h5 d2d4 e8c8 c3b5 d6e7 f1e1 c6d4 b5d4 d8d4 d1d4 e5d4 e1e7 g8e7 f3d4 h5f7 h3h4 e7g6 c1e3 g6e5 g2h3 c8b8 d4e6 h8g8 f2f4 e5c4 a1d1 f8d6 e3d4 b7b6 b2b3 f7h5 g3g4 h5f7 e6g7 g8g7 b3c4 d6f4 d4f6');
INSERT INTO evaluations VALUES(1834,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(1834,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(1834,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(1834,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(1834,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(1834,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(1834,36,'cp 164','a2a3 d7d6 b1c3 f6d8 d4d5 e6d5 e4d5 c6e7 c1e3 g8f6 d1d2 e7f5 e3f4 f8e7 f1d3 f6h5 h2h4 e7f6 e1c1 h5f4 d2f4 f6c3 b2c3 d8f6 d1e1 f5e7 f4f6 g7f6 h4h5 e8f8 c3c4 c8d7 h5h6 c7c6 d5c6 e7c6 f3h4 c6e5 d3f5 d7f5 h4f5 e5c4 f2f4 h8g8');
INSERT INTO evaluations VALUES(4120,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(4120,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(4120,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(4120,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(4120,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(4120,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(4120,36,'cp -51','g1e2 h7h6 g5d2 d7d6 b1c3 c6a5 c3d5 f6d5 d2a5 d5e7 h2h4 h6h5 f3g3 g8h7 f2f3 f7f5 a5d2 f5f4 g3f2 c8d7 a2a4 d8e8 a4a5 e7c6 g2g3 f4g3 f2g3 d7e6 h1g1 e6c4 d3c4 c6d4 e2d4 e5d4 g3g2 d4d3 e1c1 d3c2 c1c2 a8b8 f3f4');
INSERT INTO evaluations VALUES(4332,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(4332,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4332,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4332,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(4332,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4332,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4332,36,'cp 3','d7d5 e4d5 f6d5 d2e4 c5e7 a2a4 g8h8 e1g1 f7f5 e4g3 f5f4 g3e4 d5f6 b2b4 f6e4 d3e4 e7d6 h2h3 d8e8 c1a3 c6e7 f1e1 e8g6 g1h1 a7a6 b4b5 d6a3 a1a3 a6b5 c4b5 g6f6 a3a1 c7c6 b5c4 e7g6');
INSERT INTO evaluations VALUES(5312,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(5312,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(5312,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(5312,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(5312,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(5312,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(5312,36,'cp -13','e7e6 d2d3 g8e7 c2c3 e8g8 a2a4 d7d5 c4a2 b7b6 b1a3 a7a6 e4e5 a8b8 d3d4 c5d4 c3d4 b6b5 a4b5 a6b5 a2b1 b5b4 a3c2 b4b3 c2e3 d8b6 d1d3 c6a5 d3c3 a5c4 e3c4 d5c4 c3c4 c8b7 c4c5 b6c5 d4c5 f8c8 c1d2 b7f3 f1f3 c8c5 b1e4 c5c4 a1a8 b8a8');
INSERT INTO evaluations VALUES(5121,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(5121,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5121,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5121,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5121,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5121,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5121,36,'cp 0','d2d3 g8e7 c2c3 e8g8 c1e3 d7d5 c4b5 b7b6 e3f2 d5d4 a2a4 a7a6 b5c4 a8b8 d1c2 b6b5 a4b5 a6b5 c4b3 b5b4 c3d4 c6d4 f3d4 g7d4 b3c4 e7c6 f2d4 c6d4 c2f2 d8c7 b1d2 c8d7 a1c1 c7d6 c1a1');
INSERT INTO evaluations VALUES(5802,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(5802,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5802,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(5802,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(5802,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(5802,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5802,36,'cp 26','b4c3 d2c3 d5e4 d4d5 f7f6 d5c6 d8d1 a1d1 e4f3 h1g1 b7c6 g2f3 e8f7 d1d8 f7e7 d8d2 e7f7 c3b4 g8h6 f3f4 a8b8 b4a3 h6f5 e1d1 h8e8 d1c1 f5e7 d2d4 e7d5 d4a4 a7a6 a4c4 c8b7 f1d3 g7g6 g1d1 b8d8 c4d4 a6a5 b2b3 b7c8 a3c5 e6e5 f4e5 f6e5 d4a4 d5c3 d3c4 c8e6');
INSERT INTO evaluations VALUES(6268,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(6268,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(6268,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(6268,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(6268,34,'cp 212',NULL);
INSERT INTO evaluations VALUES(6268,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(6268,36,'cp 269','c2c3 b4d6 e1g1 g8e7 f3h4 e5d4 c3d4 d6b4 a2a3 b4a5 b1c3 a5b6 d1f3 d7d5 c3d5 c8e6 f3g3 e6d5 e4d5 c6d4 g3g7 h8g8 g7f6 d8d6 f6d6 c7d6 g2g3 a8c8 b2b3 e7f5 h4f5 d4f5 c1f4 e8e7 a1c1 e7f6 c4d3 c8c1 f1c1 g8d8 b3b4 f5e7 h2h4 e7g6');
INSERT INTO evaluations VALUES(6267,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(6267,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(6267,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(6267,33,'cp 247',NULL);
INSERT INTO evaluations VALUES(6267,34,'cp 247',NULL);
INSERT INTO evaluations VALUES(6267,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(6267,36,'cp 257','b4d6 e1g1 a7a5 a2a3 g8f6 c4d5 d8e7 b1d2 e8g8 f1e1 f8e8 d2c4 b7b5 c4e5 f6d5 e4d5 c6e5 d4e5 d6c5 c1f4 c8b7 d1d3 c7c6 f4e3 c5e3 e1e3 e7c5 d5d6 c5c4 d3f5 a8d8 f5h5 c6c5 f3h4 c4f4 h4f5 f4g5 f5e7 g8f8 h5g5 h6g5');
INSERT INTO evaluations VALUES(6295,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(6295,31,'cp 255',NULL);
INSERT INTO evaluations VALUES(6295,32,'cp 259',NULL);
INSERT INTO evaluations VALUES(6295,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(6295,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(6295,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(6295,36,'cp 229','e1g1 g8e7 f3h4 e8g8 d1h5 e5d4 h4f3 b7b5 c4b3 c6a5 b3c2 d4d3 c2d3 f7f6 h5b5 a5c6 b1d2 d6e5 d2c4 a8b8 b5a4 e7g6 f3e5 f6e5 c1e3 g8h8 f2f3 g6f4 e3f4 e5f4 a1d1 d8f6 g1h1 c8b7 d3c2 d7d6');
INSERT INTO evaluations VALUES(6290,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(6290,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(6290,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(6290,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(6290,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(6290,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(6290,36,'cp 91','c6e7 d1d3 g8f6 b1c3 a7a6 c4b3 b7b5 a2a4 b5b4 c3e2 e7g6 h2h4 a6a5 b3c4 d8e7 e2g3 g6f4 d3f1 g7g6 h4h5 d6c5 h5g6 f7g6 h1h4 d7d6 c1f4 e5f4 h4f4 e8g8 e4e5 f6g4 f4f8 g8f8 e1c1 d6e5 d1e1 c5d6 f3d2 g4f6 g3e4 c8f5 e4f6 e7f6 c4d3');
INSERT INTO evaluations VALUES(7004,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(7004,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(7004,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(7004,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(7004,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(7004,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(7004,36,'cp 84','b1c3 a7a6 d1d3 g8f6 c4b3 e7g6 h2h4 e8g8 h4h5 g6f4 c1f4 e5f4 e4e5 d6e5 f3e5 d8e7 e1c1 e7e5 h1e1 e5d6 c3e4 f6e4 d3e4 b7b5 d1d4 c8b7 e4f4 f8e8 f4d6 e8e1 c1d2 c7d6 d2e1 a8e8 e1d2 g7g5 c2c4 b5c4 b3c4 g8g7 a2a4 f7f5');
INSERT INTO evaluations VALUES(7363,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7363,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7363,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7363,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7363,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7363,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7363,36,'cp 0','d7c6 d2d3 c8g4 h2h3 g4h5 c1e3 d8d7 b1d2 d7e6 f3g5 e6c8 g5f3');
INSERT INTO evaluations VALUES(7128,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7128,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7128,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7128,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(7128,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(7128,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7128,36,'cp 5','d2d3 f6d7 c1e3 e8g8 b1c3 f8e8 h2h3 f7f6 f3h4 e7f8 c3e2 b7b6 d1d2 a6a5 a1d1 d8e7 a2a3 e7f7 f2f4 e5f4 e3f4 d7e5 a3a4 b6b5 d1a1 b5b4 g1h1 c6c5 b2b3 f7d7 h4f5 e5c6 f4e3 c8b7');
INSERT INTO evaluations VALUES(12705,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(12705,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(12705,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(12705,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(12705,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(12705,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(12705,36,'cp 126','c1f4 e7c6 d4d3 d8f6 d3g3 f8b4 f4d2 f6g6 g3e3 d7d6 e1c1 b4c5 e3e1 c8d7 h2h4 e8c8 h4h5 g6f6 f2f4 f6e7 a2a3 c8b8 f1c4 g7g6 c1b1 a7a6 e1f1 d8c8 f1d3 h8g8 d2e3 c5e3 d3e3 g6g5 g2g3 c6a5 c4a2 d7b5 e3d4 a5c6 d4d6 e7d6 d1d6 g5f4');
INSERT INTO evaluations VALUES(16106,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(16106,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(16106,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(16106,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(16106,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(16106,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16106,36,'cp 0','f6e4 d1e2 e4d6 c4d3 e5e4 d3e4 f8e7 e4d3 e8g8 c2c3 b7b6 e1g1 f8e8 d3c2 e7f8 d5e3 a7a5 f1e1 c8a6 e2d1 c6e5 f3e5 e8e5 d2d4 e5e8 e3g4 e8e1 d1e1 d8h4 e1d1 a8e8 g4e5 e8e7 c2b3 h4f6 e5g4 f6g6 g4e5');
INSERT INTO evaluations VALUES(16496,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(16496,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(16496,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(16496,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(16496,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(16496,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(16496,36,'cp -62','c2c4 f8c5 b1c3 d7d6 f1e2 c6d4 a1b1 d4e2 d1e2 f6g4 c3d1 f7f5 b3b4 c5b6 g1f3 e8g8 e1g1 a7a5 b4a5 a8a5 e4f5 c8f5 d1c3 d8a8 h2h3 g4f6 c1e3 b6e3 e2e3 b7b6 f3d2 a8e8 d2e4 e8g6 e4f6 f8f6 c3d5 f6f7');
INSERT INTO evaluations VALUES(17061,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(17061,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(17061,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(17061,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(17061,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(17061,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(17061,36,'cp 73','h2h4 h7h5 a2a3 f8c5 b2b4 c5e7 c4c5 d7d6 f1b5 c8d7 c5c6 b7c6 b5c6 d7c6 d5c6 e8g8 g1e2 d8e8 d1a4 e8d8 c1e3 f6g4 g2g3 g4e3 f2e3 d8c8 e1c1 e7d8 c3d5 c8g4 e2c3 g6e7 d5e7 d8e7 c1b2 e7d8 b4b5 g4g3');
INSERT INTO evaluations VALUES(17838,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(17838,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17838,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(17838,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17838,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17838,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17838,36,'cp 0','c1e3 f8d6 e1g1 g8e7 c2c4 b7b6 c4c5 d6c5 e3c5 b6c5 d1c2 a6a5 f1c1 e7g6 c2c5 g6f4 c5c6 c8d7 c6c2 d7g4 b1c3 g4f3 g2f3 f4h3 g1g2 h3f4');
INSERT INTO evaluations VALUES(18375,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(18375,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(18375,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(18375,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(18375,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(18375,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(18375,36,'cp -78','c8a6 c1e3 g8f6 b1d2 c5d4 e3d4 c6c5 d4e5 f8e7 c2c4 a6b7 d1e2 d7d6 e5g3 f6h5 e1g1 e8g8 f1e1 h5g3 h2g3 d8b6 d2b1 b6b4 b1c3 f8d8 e1d1 e7f6 a2a3 b4b6 d1d2 h7h6 g3g4 a8b8 a1d1 a7a6 d2d3 d8d7 e2d2 b7c6 d3d6 b6b2 d2b2');
INSERT INTO evaluations VALUES(19394,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(19394,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(19394,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(19394,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(19394,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(19394,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(19394,36,'cp -41','g7g6 d2d3 e7e6 b1c3 g8e7 g2g3 d7d5 f1g2 f8g7 e4e5 b7b6 g1f3 e8g8 e1g1 c8a6 d1e2 d8d7 b2b3 a8d8 c1b2 e7f5 a1e1 a6b7 c3b5 a7a6 g3g4 a6b5 g4f5 g6f5 c4b5 c6e7 a2a4 d5d4 f3h4 b7d5 e2h5 d5b3 b2c1 g8h8');
INSERT INTO evaluations VALUES(21114,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(21114,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(21114,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21114,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21114,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21114,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21114,36,'cp 0','e7e5 f4e5 c6e5 d4e5 d8d1 e1d1 c8e6 g1f3 g8h6 h2h3 e8c8 d1c2 f8c5 b1d2 e6f5 c2c3 a7a5 a2a3 b7b5 g2g4 f5d7 b2b4 c5b6 a3a4 a5b4 c3b4 c7c5 b4c3 b5b4 c3c4 d7e6 c4b5 c8b7 a4a5 e6d7 b5c4 d7e6');
INSERT INTO evaluations VALUES(21311,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(21311,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(21311,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(21311,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(21311,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(21311,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(21311,36,'cp 112','f3d4 g8f6 d4c6 b7c6 d1e2 f8e7 c1e3 f6g4 e3d4 g4e5 e1c1 e7f6 c4a6 e8g8 a6c8 d8c8 f2f4 e5g4 h2h3 f6d4 d1d4 g4f6 g2g4 f8e8 h1e1 c6c5 d4d1 a8b8 e2f3 a7a5 e4e5 d6e5 e1e5 e8e5 f4e5 f6h7 f3c6 h7g5 h3h4 g5e6 c3d5 c8b7 b2b3 b7c6 d5e7 g8h7');
INSERT INTO evaluations VALUES(22303,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(22303,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(22303,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(22303,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(22303,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(22303,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(22303,36,'cp 150','a2a4 g7g6 b1c3 f8h6 h2h3 e8f8 c1d2 f8g7 d3d4 a7a6 e1g1 h6d2 d1d2 g8e7 a4a5 h8f8 a1a3 g7h8 f1d1 c8d7 c3e2 e7g8 c4d5 d8e7 c2c4 a8d8 b2b4 e5d4 e2d4 c6d4 f3d4 c7c6 d5g8 h8g8 d1e1 d8e8 d4f3 g6g5');
INSERT INTO evaluations VALUES(24382,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(24382,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(24382,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24382,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(24382,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(24382,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24382,36,'cp 32','g7g6 f2f3 f8g7 c1g5 h7h6 g5e3 e8g8 d1d2 d6d5 e4d5 c6d5 c4b3 a7a5 e1c1 a5a4 b3d5 f6d5 d2d5 a4a3 d5d7 a3b2 c1b1 g7c3 d7c6 d8f6 c6f6 c3f6 e3d4 f6d4 d1d4 c7c5 d4d5 f8c8 b1b2 c8b8 b2c3 a8a2 d5c5 a2a3 c3d2 a3a7 h1e1 b8d8 d2e2 a7e7 e2f1 e7d7 e1e2 g8g7 h2h4 d7d2');
INSERT INTO evaluations VALUES(24391,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(24391,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(24391,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(24391,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(24391,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(24391,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(24391,36,'cp 91','b5a4 f8e7 e1g1 e8g8 d2d3 c6c5 b1c3 d7a4 c3a4 a8c8 g5e4 f6d5 e4g3 f8e8 c1d2 a5c6 c2c3 d5f4 d2f4 e5f4 f3f4 d8d3 b2b3 c8d8 g3f5 e7f6 a4c5 d3c3 c5e4 c3e5 e4f6 e5f6 a1d1 g7g6 f5h6 g8g7 f4f6 g7f6 h6g4 f6g7 g4e3 c6b4 a2a3 b4d3 h2h3 h7h5 b3b4 d3b2 d1d8 e8d8');
INSERT INTO evaluations VALUES(24389,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(24389,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(24389,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(24389,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(24389,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(24389,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(24389,36,'cp 99','f8e7 d2d3 e8g8 e1g1 c6c5 b1c3 d7a4 c3a4 a8c8 g5e4 f6d5 e4g3 f8e8 g3f5 e7f8 f1d1 a5c6 c2c3 e8e6 f5e3 c6e7 e3d5 e7d5 b2b3 d8d7 c1d2 c8e8 a4b2 e6f6 f3e2 f8d6 d1e1 d6c7 g1h1 e8d8 e2f1 d5f4 d2f4 e5f4 f2f3 f6d6');
INSERT INTO evaluations VALUES(24782,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(24782,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(24782,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(24782,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(24782,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(24782,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(24782,36,'cp 75','b1d2 e7e6 h2h3 g4h5 g2g4 h5g6 f3e5 f8d6 e5c6 d8d7 c6a7 c7c6 a7c6 b7c6 b5d3 e8g8 a2a4 d7e7 f2f4 c6c5 e1g1 c5d4 e3d4 g6d3 c2d3 e7c7 d2f3 f8b8 f1f2 d6f4 c1f4 c7f4 g1g2 b8b4 f3e5 f4d4 e5c6 f6g4 d1g4 d4g4 h3g4 b4g4 g2f3 a8a4 a1h1 g8f8 f2g2 g4g2 f3g2');
INSERT INTO evaluations VALUES(25051,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(25051,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(25051,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(25051,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(25051,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(25051,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(25051,36,'cp 34','f1b5 f6d7 f4g3 f8d6 d1e2 e8g8 e3e4 f5h7 e1g1 d6g3 h2g3 c6e7 b5d7 d8d7 f1e1 d5e4 d2e4 b7b6 g3g4 d7d5 e4g3 a8e8 c3c4 d5d8 a1d1 e7g6 e2d2 g6h4 f3h4 d8h4 d2f4 h4e7 a2a3 a7a5 g3e2 c7c5 d4c5 e7c5 d1d7');
INSERT INTO evaluations VALUES(27077,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(27077,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(27077,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(27077,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(27077,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(27077,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(27077,36,'cp 21','c2c4 b5d7 e1g1 g8e7 b1c3 e7g6 c3e2 a7a5 e2g3 f8e7 b2b3 e8g8 c1e3 b7b6 g3f5 g6h4 f3h4 e7h4 d1d3 d7f5 e4f5 d8f6 a1e1 f8e8 g2g3 h4g5 d3e4 a5a4 b3b4 a4a3 e3g5 f6g5 e1e3 a8a4 f1b1 e8a8 e4c2');
INSERT INTO evaluations VALUES(29053,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(29053,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29053,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(29053,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(29053,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(29053,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(29053,36,'cp 25','g1f3 e7e6 e1g1 d5e4 c3e4 f6e4 d3e4 g4h5 c2c3 b8d7 c1f4 d7f6 f1e1 f6e4 e1e4 f8e7 e4e5 h5g6 e5e1 g6h5 a2a4 e8g8 h2h3 a7a5 g2g4 h5g6 f3e5 e7d6 d1b3 d6e5 e1e5 f7f6 e5e1 d8d5 b3d5 e6d5 e1e7 f8f7');
INSERT INTO evaluations VALUES(36945,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(36945,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(36945,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(36945,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36945,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36945,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36945,36,'cp 0','c8g4 c1g5 c7c6 b1d2 h7h6 g5h4 b8d7 h4g3 d6g3 h2g3 f8e8 e1e8 f6e8 d2f1 g7g6 d1d2 g4f3 g2f3 g8g7 f1e3 h6h5 g1g2 e8d6 c2c3 d8f6 a1h1 a8h8 a2a4 a7a5 h1e1 d7b6 d2c2 h8e8 b2b3 e8h8 e1h1 h8e8');
INSERT INTO evaluations VALUES(27880,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27880,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27880,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27880,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27880,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27880,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(27880,36,'cp 3','c1g5 b8d7 b1d2 h7h6 g5h4 c7c6 c2c4 d8c7 d1c2 d5c4 d3c4 f8e8 h2h3 g4f3 d2f3 f6d5 c4d5 c6d5 c2f5 e8e1 a1e1 c7c6 h4g3 d6g3 f2g3 d7f6 e1e7 c6b6 f5e5 b6b2 e5c7 a8f8 c7b7 b2a2 b7a7 a2a7 e7a7 f6e4');
INSERT INTO evaluations VALUES(38058,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(38058,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(38058,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(38058,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(38058,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(38058,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(38058,36,'cp 129','b4f8 a2a3 b7b6 h2h4 c6c5 g1f3 b8c6 c3e2 c8a6 c2c3 d8d7 h4h5 f7f6 g2g3 f6e5 d4e5 g8h6 d2h6 g7h6 f1h3 f8e7 e2f4 c6d8 d1c2 a6b7 e1c1 e7g5 f3h2 d5d4 h1e1 e8g8 c1b1 g8h8 c3d4 c5d4 c2d2 d7f7 d2d4 g5f4 g3f4 f7h5 d4e3 d8c6 h3e6 h5h2 f4f5 c6e5 e3e5 h2e5 e1e5 a8e8');
INSERT INTO evaluations VALUES(40406,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(40406,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(40406,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(40406,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(40406,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(40406,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(40406,36,'cp 46','e4e5 f6h5 f1b5 c8d7 d1e2 d7b5 e2b5 b8d7 d4f3 d6e5 f4e5 a7a6 b5e2 d8b6 c1d2 h5f4 e2e4 f4e6 e1c1 a8c8 c1b1 b6c6 e4c6 c8c6 h1e1 h7h5 c3d5 d7b6 d5b4 c6c5 d2e3 c5b5 c2c3 a6a5 e3b6 a5b4 b6d4 b4c3 d4c3 b5c5 a2a4 c5c4 a4a5 h5h4 b1c2 h8h5 b2b4 c4c6');
INSERT INTO evaluations VALUES(40371,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40371,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(40371,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(40371,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(40371,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(40371,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(40371,36,'cp 42','f6h5 f1b5 c8d7 d1e2 d7b5 e2b5 b8d7 d4f3 d6e5 f4e5 a7a6 b5e2 d8b6 c1d2 b6e6 e1c1 d7e5 h1e1 f7f6 e2e4 e8c8 f3e5 f6e5 e4b4 e6c6 b4e7 h8e8 e7g5 h5f4 g2g3 f4d5 g5h4 d5f6 d2e3 d8d1 e1d1 e8d8 d1e1 d8d7 c1b1 h7h5 h2h3 f6d5 c3d5 d7d5 h4e7 d5d7 e7e8 c8c7');
INSERT INTO evaluations VALUES(41520,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41520,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(41520,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(41520,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41520,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(41520,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(41520,36,'cp 32','f6d7 e5e6 f7e6 d1d2 d7e5 d2d8 e8d8 e1c1 c8d7 c3e4 b7b6 g1h3 d8e8 g5d2 e5f7 d1e1 b8c6 e4g5 c6d8 f1c4 f7d6 c4e6 h7h6 e6d7 e8d7 g5e4 e7e5 f2f3 d8e6 h3f2 a8f8 h2h4 h6h5 f2d3 f8f5');
INSERT INTO evaluations VALUES(43132,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43132,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43132,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43132,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43132,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43132,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43132,36,'cp 6','c7c5 e2e3 d8b6 d1b3 e8g8 b1d2 d7d6 f1c4 h7h6 g5h4 g6g5 h4g3 f6h5 b3c2 b6c7 e1g1 h5g3 h2g3 e7e6 a1c1 b8d7 c4d3 f8d8 c2d1 d6d5 d1e2 c5c4 d3c2 b7b5 f1e1 f7f5 e3e4 d5e4 d2e4 c8b7');
INSERT INTO evaluations VALUES(44271,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(44271,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(44271,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(44271,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(44271,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(44271,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(44271,36,'cp -65','h7h6 g5f6 d8f6 c2c3 b8c6 d4b3 a8b8 a2a4 a7a6 d1d3 b7b5 a4b5 a6b5 d3g3 b5b4 b1d2 f6e5 g3e5 c6e5 f2f4 e5c6 f1d3 b4c3 b2c3 g7g5 g2g3 h6h5 e4e5 g5f4 g3f4 f8h6 h1f1 h8g8 b3d4 c6d4 c3d4 c8b7 d2c4 g8g4 d3e2 g4g2 e2h5 g2h2');
INSERT INTO evaluations VALUES(44776,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44776,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(44776,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(44776,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(44776,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(44776,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(44776,36,'cp 15','e4e5 d6e5 f4e5 f6d7 e5e6 f7e6 f1c4 f8g7 d2d3 e8g8 c4e6 g8h8 e1g1 b8c6 e6d5 d7e5 f3e5 f8f1 d1f1 c6e5 g1h1 d8b6 f1e1 c8f5 c1f4 e5g4 c3d1 a8f8 h2h3 f5d7 f4g3 g4f6 d5f3 d7c6 f3c6 b6c6 d1c3 f8f7 e1e3 h8g8 g3e5 b7b5 a1e1 b5b4 c3e4 f6e4 e3e4 c6e4');
INSERT INTO evaluations VALUES(46578,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(46578,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(46578,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(46578,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(46578,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(46578,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(46578,36,'cp 55','e2e3 h7h6 f4g3 g7g5 c2c4 f8g7 c4c5 e8g8 b1c3 b8c6 f1c4 e7e6 c5d6 c7d6 a1c1 a7a6 d4d5 e6d5 c3d5 c8e6 e1g1 f6e4 f3d2 e4g3 h2g3 g8h8 c4b3 a8b8 d1c2 d8a5 d5c3 d6d5 d2f3 e6g8 f1d1 b8d8 c3e2 a5b4 a2a3 b4e4 f3d4 e4c2 b3c2 c6d4');
INSERT INTO evaluations VALUES(46734,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(46734,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(46734,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(46734,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(46734,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(46734,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(46734,36,'cp 87','f1c4 d5b6 c4b3 f8g7 d1e2 d6e5 d4e5 c7c5 b1c3 c5c4 b3c4 b6c4 e2c4 b8c6 e1g1 c8e6 c4a4 e8g8 f1e1 d8b6 a4h4 h7h5 h2h3 f8d8 g2g4 h5g4 f3g5 c6e5 h4h7 g8f8 g5e6 f7e6 e1e5 g7e5 h7g6 e5g7 c1h6 g7h6 g6h6 f8e8 a1e1 g4h3');
INSERT INTO evaluations VALUES(46732,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46732,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(46732,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(46732,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(46732,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(46732,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(46732,36,'cp 68','d5b6 c4b3 f8g7 d1e2 d6e5 d4e5 c7c5 e1g1 c5c4 b3c4 b6c4 e2c4 b8c6 f1e1 c8e6 c4e2 e8g8 h2h3 e6d5 b1c3 d5f3 e2f3 d8a5 c1f4 a8d8 a2a3 c6d4 f3e4 d8d7 a1d1 f8d8 d1d2 d4e6 d2d7 d8d7 f4g3 g7h6 a3a4 a5b6 e1d1 d7d2 d1d2 h6d2 e4e2 b6d4 c3b5');
INSERT INTO evaluations VALUES(46611,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(46611,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(46611,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(46611,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(46611,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(46611,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(46611,36,'cp 72','c4b3 f8g7 f3g5 e7e6 d1f3 d8e7 g5e4 d6e5 c1g5 e7b4 b1d2 e5d4 a2a3 b4b5 g5f6 e8g8 f6g7 g8g7 f3f6 g7g8 f6d4 b8d7 e1c1 b5e5 d4d3 d7c5 e4c5 e5c5 h2h4 c5f5 d3e2 h7h6 h4h5 g6g5 f2f4 c8d7 f4g5 f5g5 c1b1 d7b5 e2f2 g5f5 f2g3 f5g5 g3g5 h6g5 d2e4 f8d8 e4g5 d8d1 h1d1');
INSERT INTO evaluations VALUES(46616,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(46616,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(46616,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(46616,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(46616,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(46616,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(46616,36,'cp 58','f8g7 e1g1 e8g8 a2a4 d6e5 f3e5 c7c5 d4c5 g7e5 c5b6 d8b6 b1c3 c8e6 a4a5 b6a6 c1h6 f8e8 b3a4 e8c8 f1e1 a6d6 c3b5 d6d1 a1d1 b8c6 c2c3 e6c4 b5a7 a8a7 a4c6 a7a5 c6b7 c8b8 h2h3 b8b7');
INSERT INTO evaluations VALUES(47529,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(47529,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(47529,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(47529,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(47529,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(47529,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(47529,36,'cp 148','g1f3 e4d6 c4d3 c8g4 e3e4 g7g6 c1e3 f8g7 b1d2 e8g8 h2h3 g4f3 d2f3 b8c6 d1b3 e7e5 d4d5 c6e7 e1g1 c7c6 d5c6 e7c6 a1c1 d8a5 a2a3 f8d8 c1c5 a5b6 c5b5 b6c7 b5d5 a8c8 f1c1 c7e7 h3h4 c6d4 f3d4 e5d4 c1c8');
INSERT INTO evaluations VALUES(47530,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(47530,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(47530,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(47530,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(47530,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(47530,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(47530,36,'cp 127','g7g6 c4d3 e4d6 e3e4 c8g4 b1d2 f8g7 h2h3 g4f3 d2f3 e8g8 c1e3 b8c6 a1c1 e7e5 d4d5 c6e7 d1b3 c7c6 d5c6 e7c6 e1g1 a8c8 f1d1 d8e7 g2g3 b7b6 b3a3 f8d8 h3h4 h7h6 g1g2 b6b5 h4h5 g6h5 c1c5 e7b7 c5d5 d6c4 d3c4 b5c4');
INSERT INTO evaluations VALUES(47651,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(47651,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(47651,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(47651,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(47651,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(47651,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(47651,36,'cp 131','e4e5 f6g4 e5d6 f8d6 h2h3 g4f6 f1e1 d6e7 d1d4 b8c6 d4h4 e8g8 b1c3 f6e8 h4g3 e7d6 c1f4 d6f4 g3f4 d8f6 f4g3 c8f5 c3d5 f6d6 g3d6 e8d6 c4b3 a8d8 d5c7 d8d7 c7d5 f5e4 a1d1 e4f3 g2f3 f8d8 d5e3 d6b5 d1d7 d8d7 g1g2 b5c7 g2g3 c6d4 f3f4 d4b3 a2b3 d7d2 g3f3');
INSERT INTO evaluations VALUES(48855,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(48855,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(48855,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(48855,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(48855,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(48855,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(48855,36,'cp -90','f8d6 d2d3 e8g8 g1e2 c7c6 e1g1 f8e8 h2h3 a7a5 c2c4 b8a6 b1c3 a6c7 c3a4 c7e6 a4c3 d5d4 e3d4 e5d4 c3e4 f6e4 d3e4 d6e7 e2f4 a5a4 a1b1 a4b3 a2b3 e6f4 c1f4 c8e6 d1d2 d8d7 g1h2 c6c5 f1e1 b7b5 c4b5 d7b5 e4e5 e7f8 g2a8 e8a8');
INSERT INTO evaluations VALUES(49202,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(49202,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(49202,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(49202,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(49202,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(49202,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(49202,36,'cp 197','g1f3 f8g7 f3e5 e8g8 d2d4 d7d6 e5f3 h7h6 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 c2c3 f8e8 f1e1 d7f6 e4c2 e8e1 d1e1 d8d6 e1e5 d6e5 f3e5 c8e6 a2a4 a8d8 f2f3 f6d5 e5d3 b7b6 a4a5 d8c8 g1f2 c7c5 d4c5 b6c5 d3f4 d5f4 c1f4 e6c4 a1e1');
INSERT INTO evaluations VALUES(49201,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(49201,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(49201,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(49201,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(49201,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(49201,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(49201,36,'cp 190','f8g7 f3e5 e8g8 d2d4 d7d6 e5f3 h7h6 c4d3 c7c5 d4d5 f8e8 f3d2 b8d7 e1g1 a7a6 a2a4 d7e5 d3e2 b7b6 a1a3 g6g5 f2f4 g5f4 f1f4 c8d7 h2h3 e5g6 f4f2 d8e7 e2d3 f6h7 d2c4 b6b5 a4b5 a6b5 a3a8 e8a8 c4b6 a8b8 b6d7 e7d7 g1h1 b5b4 c3e2');
INSERT INTO evaluations VALUES(42327,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(42327,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(42327,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(42327,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(42327,34,'cp 182',NULL);
INSERT INTO evaluations VALUES(42327,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(42327,36,'cp 187','f3e5 e8g8 d2d4 d7d6 e5f3 h7h6 e1g1 f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 b8d7 c2c3 f8e8 f1e1 d7f6 e4c2 e8e1 d1e1 d8d6 e1e5 d6e5 f3e5 c8e6 a2a4 f6d5 f2f4 c7c5 d4c5 g7f8 c5c6 b7c6 g1f2 a8c8 a4a5 c6c5 a1a4 c8b8 e5c6 b8b7 g2g3 d5f6 c6e5');
INSERT INTO evaluations VALUES(49883,30,'cp -284',NULL);
INSERT INTO evaluations VALUES(49883,31,'cp -274',NULL);
INSERT INTO evaluations VALUES(49883,32,'cp -285',NULL);
INSERT INTO evaluations VALUES(49883,33,'cp -271',NULL);
INSERT INTO evaluations VALUES(49883,34,'cp -273',NULL);
INSERT INTO evaluations VALUES(49883,35,'cp -262',NULL);
INSERT INTO evaluations VALUES(49883,36,'cp -282','b1c3 e4c3 d2c3 c7c6 d1e2 d8e7 h2h3 d7d5 c4b3 a7a5 a2a4 b8d7 c1e3 b7b6 e1c1 c8a6 e2d2 e8c8 g1e2 d7c5 e3c5 e7c5 e2g3 g7g6 c3c4 a6c4 g3e4 c5b4 d2b4 f8b4 b3c4 d5c4 c2c3 d8d1 h1d1 b4e7 f2f3 f7f5 g4f5 g6f5');
INSERT INTO evaluations VALUES(50241,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50241,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50241,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50241,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50241,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50241,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50241,36,'cp 0','d7d6 b1c3 c7c5 f1e2 b8c6 c3d5 f8g7 c2c3 e8g8 e1g1 f6d5 e4d5 c6e5 f1e1 b7b6 d3d4 e5f3 e2f3 c5d4 c3d4 c8f5 c1g5 g7f6 g5h6 f6g7');
INSERT INTO evaluations VALUES(50977,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(50977,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(50977,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(50977,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(50977,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(50977,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(50977,36,'cp 66','b7b5 c4b3 c7c5 g1f3 c5d4 f3d4 c8b7 e1g1 b8c6 d4c6 d8d1 f1d1 b7c6 c1e3 g8f6 f2f3 f8e7 c3e2 e8g8 g2g4 h7h6 e2d4 c6d7 h2h4 f8d8 a1c1 f6e8 g4g5 a6a5 d4c6 d7c6 d1d8 a8d8 c1c6 h6g5 h4g5 a5a4 b3c2 e8d6 b2b3 a4b3 c2b3 d8c8 c6c8 d6c8 g5g6');
INSERT INTO evaluations VALUES(51140,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(51140,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(51140,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(51140,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(51140,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(51140,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(51140,36,'cp 80','c4b3 c7c5 g1f3 c5d4 f3d4 c8b7 e1g1 b8c6 d4c6 d8d1 f1d1 b7c6 c1e3 g8f6 f2f3 f8e7 c3e2 e8g8 a1c1 f8c8 e2f4 g8f8 e4e5 f6d5 f4d5 c6d5 c1c8 a8c8 b3d5 e6d5 d1d5 f7f6 e5f6 e7f6 d5d2 c8c1 g1f2 c1b1 b2b3 a6a5 g2g4 a5a4 b3a4 b5a4 g4g5 f6e7 f3f4 a4a3 d2d7 f8e8');
INSERT INTO evaluations VALUES(51143,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(51143,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(51143,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(51143,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(51143,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(51143,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(51143,36,'cp 88','c7c5 g1f3 c5d4 f3d4 c8b7 e1g1 b8c6 d4c6 d8d1 f1d1 b7c6 c1e3 g8f6 f2f3 f8e7 a1c1 e8g8 c3e2 f8c8 e2f4 g8f8 e4e5 f6d5 f4d5 c6d5 c1c8 a8c8 b3d5 e6d5 d1d5 f7f6 e5f6 e7f6 d5d2 f8f7 g1f1 h7h6 f1e2 a6a5 e2d1 a5a4 d2d7 f7e6 d7b7 f6b2 b7b5 b2f6 e3d2 e6f7');
INSERT INTO evaluations VALUES(55158,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(55158,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(55158,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(55158,33,'cp -128',NULL);
INSERT INTO evaluations VALUES(55158,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(55158,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(55158,36,'cp -144','e7e5 c2c3 b8c6 d2d4 e5d4 c3d4 c6b4 c1d2 c8f5 d2b4 f8b4 b1c3 g8f6 e2e3 f6e4 g1e2 c7c6 g2g3 a7a5 f1g2 d8d7 e1g1 e4f6 f2f3 f5e6 e2f4 b4d6 f1e1 e8g8 e3e4 d6f4 g3f4 g7g6 a1c1 g8g7 g1f2 d5e4');
INSERT INTO evaluations VALUES(55862,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(55862,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55862,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(55862,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55862,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55862,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55862,36,'cp 0','g8f6 e2e3 c7c5 b2b3 c5d4 e3d4 f8d6 f1d3 e8g8 e1g1 b8c6 f1e1 a7a5 a2a3 c6e7 c1b2 b7b6 a1c1 e7g6 f3e5 d8e7 e5c6 e7b7 c6e5');
INSERT INTO evaluations VALUES(57100,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(57100,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(57100,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(57100,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(57100,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(57100,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(57100,36,'cp -98','d7d6 e5c4 d6d5 c4e3 c7c6 c2c4 c8e6 b1c3 g8f6 g2g4 d5c4 f1g2 f6g4 e3g4 e6g4 g2e4 f8b4 d4d5 e8g8 h1g1 g4h5 d1d4 h5g6 c1d2 c6d5 d4d5 b8c6 d5d8 a8d8 e4c6 b7c6 g1g4 f8e8 e1c1 b4c5 d2e3 c5e3 f2e3 d8d1 c3d1 g6e4 d1c3 f7f5 g4f4 g7g6 c3e4 e8e4 f4e4 f5e4 c1c2 g8g7 c2c3 g7f6 c3c4');
INSERT INTO evaluations VALUES(57136,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(57136,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(57136,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(57136,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(57136,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(57136,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(57136,36,'cp -51','e5d4 d1d4 b8c6 d4e3 f8e7 c1b2 g8f6 b2f6 g7f6 b1c3 d7d5 e1c1 c8e6 g1f3 f6f5 c1b1 e8g8 h2h3 e7f6 g2g4 f5f4 e3d2 f6g7 e2e4 d8f6 e4d5 a8d8 f1c4 e6d5 g4g5 d5f3 g5f6 d8d2 d1d2 f3h1 f6g7 f8d8 c4d3 g8g7 c3e2 c6e5 e2f4');
INSERT INTO evaluations VALUES(274,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(274,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(274,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(274,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(274,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(274,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(274,36,'cp 31','b1c3 g8e7 d2d3 d7d6 e1g1 h7h6 c1g5 f6g6 h5g6 e7g6 g5e3 c5e3 f2e3 c8e6 b2b3 g6e7 f1f2 e8g8 a2a4 a8d8 c3d5 e6d5 e4d5 c6a5 e3e4 c7c6 d5c6 a5c4 b3c4 b7c6 c4c5 f7f6 c5d6 e7c8 f2f1 c8d6 f3d2 d8d7 h2h3 f8d8 f1b1 c6c5 a4a5');
INSERT INTO evaluations VALUES(308,30,'cp -848',NULL);
INSERT INTO evaluations VALUES(308,31,'cp -859',NULL);
INSERT INTO evaluations VALUES(308,32,'cp -886',NULL);
INSERT INTO evaluations VALUES(308,33,'cp -871',NULL);
INSERT INTO evaluations VALUES(308,34,'cp -871',NULL);
INSERT INTO evaluations VALUES(308,35,'cp -863',NULL);
INSERT INTO evaluations VALUES(308,36,'cp -863','g5g2 h1f1 c6d4 c1e3 d4f3 d1f3 g2f3 e3c5 d7d6 c5e3 f3h5 b1d2 g8e7 f1h1 e8g8 f2f3 c8h3 h1g1 h5h4 g1g3 e7c6 a2a3 h3d7 e3g1 g7g6 e1c1 h4h5 c1b1 a8e8 g3g2 g8g7 d1e1 f7f5 e4f5 h5f5 g1e3 f5h5 e1e2 a7a5 c4d5');
INSERT INTO evaluations VALUES(313,30,'cp -848',NULL);
INSERT INTO evaluations VALUES(313,31,'cp -840',NULL);
INSERT INTO evaluations VALUES(313,32,'cp -836',NULL);
INSERT INTO evaluations VALUES(313,33,'cp -848',NULL);
INSERT INTO evaluations VALUES(313,34,'cp -871',NULL);
INSERT INTO evaluations VALUES(313,35,'cp -875',NULL);
INSERT INTO evaluations VALUES(313,36,'cp -877','h1f1 d7d5 c4b5 d5e4 d3e4 g8f6 b1d2 c8g4 f2f3 g2h2 d1e2 h2g3 e1d1 g4h3 c2c3 h3f1 d2f1 g3g1 b2b4 c5b6 a1b1 h7h5 a2a4 h5h4 a4a5 h4h3 a5b6 h3h2 c1e3 a7b6 d1c2 a8a2 b1b2 a2b2 c2b2 h2h1q e3g1 h1g1 b5c4 e8e7 b2c2 h8a8 b4b5 c6a5 c4a2');
INSERT INTO evaluations VALUES(314,30,'cp -869',NULL);
INSERT INTO evaluations VALUES(314,31,'cp -871',NULL);
INSERT INTO evaluations VALUES(314,32,'cp -868',NULL);
INSERT INTO evaluations VALUES(314,33,'cp -882',NULL);
INSERT INTO evaluations VALUES(314,34,'cp -871',NULL);
INSERT INTO evaluations VALUES(314,35,'cp -882',NULL);
INSERT INTO evaluations VALUES(314,36,'cp -873','d7d5 c4b5 d5e4 d3e4 g8f6 b1d2 c8g4 f2f3 g2h2 d1e2 h2g3 e1d1 g4h3 c2c3 h3f1 d2f1 g3g1 b2b4 c5b6 a2a4 a7a6 b5c6 b7c6 a4a5 a8d8 d1c2 g1f2 f1g3 f2e2 g3e2 b6f2 c1g5 h7h6 a1h1 e8e7 g5d2 h6h5 h1h2 f2a7 h2h3 h5h4 e2c1 e7e6 c1d3 f6d7 c3c4 a7d4');
INSERT INTO evaluations VALUES(3331,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(3331,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(3331,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(3331,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(3331,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(3331,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(3331,36,'cp 86','b1d2 e7d6 h2h4 d8e7 f3e5 d6e5 d4e5 f6d7 d2f3 f7f6 e5f6 d7f6 f3e5 c6e5 f4e5 c8d7 d1b3 f6g4 e5g3 e7c5 d3e2 g4f6 e2f3 b7b5 b3d1 c5b6 e1g1 c7c5 f1e1 b5b4 c3b4 b6b4 b2b3 a7a5 a1c1 f8f7 g3e5 b4h4');
INSERT INTO evaluations VALUES(6350,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(6350,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(6350,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(6350,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(6350,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6350,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6350,36,'cp 25','d7c6 e1g1 c8g4 h2h3 g4h5 c2c3 b4e7 d2d3 g8f6 f1e1 a7a5 d1e2 d8d6 b1d2 d6e6 f3g5 e6c8 e2e3 e8g8 g5f3 c8e6 d3d4 f6d7 d4e5 h5f3 d2f3 d7e5 f3d4 e6c4 e3g3 e7f6 f2f4 e5d3 e4e5 d3e1 e5f6');
INSERT INTO evaluations VALUES(6095,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(6095,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(6095,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(6095,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(6095,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6095,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6095,36,'cp 17','e1g1 f7f6 d2d4 e5d4 f3d4 g8e7 c1e3 e8g8 d1e2 g8h8 b1d2 d8d7 a1d1 b7b6 f1e1 f8e8 a2a3 b4d6 d2f1 a7a6 e2f3 c6c5 d4e2 e7c6 e3f4 c6e5 f4e5 e8e5 f1e3 d7f7 e2c3 c8b7 c3d5 e5e8 e3c4 d6f8 f3g3 b7d5 d1d5 a8d8 d5d8');
INSERT INTO evaluations VALUES(18196,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(18196,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(18196,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(18196,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(18196,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(18196,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(18196,36,'cp 95','e7e6 c2c4 g8e7 b1c3 e7g6 c4d5 c6d5 e4d5 f8e7 d5e6 c8e6 d3d4 e8g8 c1e3 c5d4 f3d4 e6d7 a1c1 f8e8 d1h5 e7f8 f1d1 a8c8 h2h3 h7h6 h5f3 g6h4 f3g3 g8h8 c3d5 c8c1 d1c1 h4f5 d4f5 d7f5');
INSERT INTO evaluations VALUES(18230,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(18230,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(18230,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(18230,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(18230,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(18230,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(18230,36,'cp 82','c2c4 g8e7 b1c3 e7g6 c4d5 c6d5 e4d5 e6d5 d3d4 c5c4 h2h4 c8g4 d1a4 d8d7 f1e1 f8e7 a4d7 e8d7 e1e3 a8b8 c3d5 h8e8 f3d2 g4e6 d5e7 g6e7 d2e4 e7f5 e4c5 d7c6 c5a6 b8d8 e3e4 f5d4 a6b4 c6b5 c1e3 d4e2 g1f1 b5b4 f1e2 g7g6 e4f4 b4b5 a1c1 d8a8 a2a3 a8d8 e2f1 h7h5 e3d4 e6f5');
INSERT INTO evaluations VALUES(18407,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(18407,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(18407,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(18407,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(18407,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(18407,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(18407,36,'cp 35','c5d4 d1d4 g8f6 e4e5 f6d5 b1c3 d8b6 d4b6 a7b6 c3d5 c6d5 c1e3 f8g7 c2c3 f7f6 e5f6 g7f6 a2a4 b6b5 a4a5 d7d6 e3g5 f6g5 f3g5 h7h6 g5f3 e7e5 e1g1 e8e7 f1d1 c8e6 f3e1 b5b4 c3b4 h8b8 e1d3 e5e4 d3f4 b8b4 f4d5 e6d5 d1d5 b4b2 a5a6');
INSERT INTO evaluations VALUES(18417,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(18417,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(18417,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(18417,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(18417,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(18417,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(18417,36,'cp 52','d1d4 f7f6 e1g1 g8h6 c2c4 h6f7 b1c3 e7e6 f1d1 c6c5 d4d3 f8e7 b2b3 c8b7 c1b2 d8c7 d1d2 a7a6 d3e3 e8g8 a1d1 d7d6 f3e1 f8e8 c3e2 b7c6 h2h4 c6d7 e2c3 c7b7 e1d3 d7c6 d3f4 c6d7 c3a4');
INSERT INTO evaluations VALUES(18415,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(18415,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(18415,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(18415,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(18415,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(18415,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(18415,36,'cp 45','f7f6 e1g1 g8h6 c2c4 h6f7 b1c3 e7e5 d4d3 f8e7 f1d1 e8g8 a1b1 a7a5 c1e3 d8e8 e3b6 f7d8 c3b5 d8e6 b5d6 e7d6 d3d6 f8f7 b6c5 c8a6 b2b3 e8b8 f3d2 b8d6 c5d6 a8e8 d2f1 e8d8 f1e3 e6d4 g1h1 g8g7');
INSERT INTO evaluations VALUES(18724,30,'cp -164',NULL);
INSERT INTO evaluations VALUES(18724,31,'cp -148',NULL);
INSERT INTO evaluations VALUES(18724,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(18724,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(18724,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(18724,35,'cp -155',NULL);
INSERT INTO evaluations VALUES(18724,36,'cp -160','g8f6 h5e2 f6e4 e2e4 d7d5 c4d5 d8d5 e4d5 e6d5 d2d4 c5d4 g1f3 d4c3 b1c3 c8e6 c1e3 e8c8 f3d4 c6d4 e3d4 b7b6 e1g1 c8b7 a1c1 h8g8 a2a3 f8c5 f1d1 c5d6 g2g3 d6b8 d1e1 d8c8 c3e2 c8c4 d4c3 g7g6 e2d4 g8e8 c1d1 c4c8 d4e6 f7e6 e1e2 a7a6 g1g2 a6a5 e2e1 b7a6 f2f4');
INSERT INTO evaluations VALUES(12346,30,'cp -147',NULL);
INSERT INTO evaluations VALUES(12346,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(12346,32,'cp -145',NULL);
INSERT INTO evaluations VALUES(12346,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(12346,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(12346,35,'cp -144',NULL);
INSERT INTO evaluations VALUES(12346,36,'cp -153','h5e2 f6e4 e2e4 d7d5 c4d5 d8d5 e4d5 e6d5 d2d4 c5d4 g1f3 d4c3 b1c3 c8e6 c1e3 e8c8 f3d4 c6d4 e3d4 b7b6 a1c1 c8b7 e1g1 h8g8 f1e1 g7g6 a2a3 f8e7 c3b5 d8c8 d4c3 c8c4 h2h3 g8c8 c1d1 e7c5 c3f6 c8g8 b2b3 c4f4 f6d4');
INSERT INTO evaluations VALUES(20999,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(20999,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(20999,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(20999,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(20999,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(20999,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(20999,36,'cp -17','c8f5 e2e4 f5e4 c3e4 d5e4 c2c3 g8f6 d1b3 d8d7 b3b7 a8b8 b7a6 e7e5 f1b5 b8b6 b5c6 b6a6 c6d7 f6d7 g1e2 f7f5 c1e3 f8d6 a2a4 c7c5 e1g1 c5d4 c3d4 e8g8 f1d1 f5f4 e3d2 e5d4 e2d4 d6e5 d2c3 f8f7 d1e1 d7f6 a1c1 h7h6 e1d1 f4f3');
INSERT INTO evaluations VALUES(22403,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(22403,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(22403,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(22403,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(22403,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(22403,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(22403,36,'cp 103','h2h4 g8f6 b1c3 d7b6 c1f4 f8g7 d1d2 c8g4 f3h2 g4d7 f1e2 b6a4 e4e5 f6h5 e2h5 g6h5 h2f3 a4c3 d2c3 d7f5 e1g1 d8d7 f1e1 e8g8 c3d2 f5g4 f3h2 d6e5 f4e5 g7e5 e1e5 f7f6 e5e3 a8d8 h2g4 h5g4');
INSERT INTO evaluations VALUES(22476,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(22476,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22476,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(22476,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(22476,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(22476,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(22476,36,'cp 66','d2d4 f8g7 d4e5 c6e5 f3e5 g7e5 f1c4 d7d6 e1g1 e5g7 f1e1 g8e7 c1g5 h7h6 g5e3 e7c6 b1a3 e8g8 f2f4 f8e8 c4d3 c8e6 d1d2 d6d5 e4e5 d5d4 c3d4 c6d4 d2c3 d4f5 d3f5 e6f5 a1d1 d8e7 a3b5 a8c8 b5a7 c8d8 a7b5 d8d1 e1d1');
INSERT INTO evaluations VALUES(22469,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(22469,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(22469,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(22469,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(22469,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(22469,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(22469,36,'cp 79','f8g7 c1g5 g7f6 h2h4 d7d5 e4d5 d8d5 d4e5 d5d1 e1d1 f6g7 f1c4 g8h6 e5e6 c8e6 h1e1 e8d7 g5h6 g7h6 c4e6 f7e6 d1e2 h8f8 e2f1 f8f5 e1e2 a8d8 b1a3 f5f4 a1e1 e6e5 a3b1 d8f8 b1d2 d7c8 d2e4 b7b6 e4g3 f4a4 a2a3 a4f4 h4h5');
INSERT INTO evaluations VALUES(6186,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(6186,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(6186,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(6186,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(6186,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(6186,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(6186,36,'cp 83','c1g5 g7f6 h2h4 d7d5 e4d5 d8d5 d4e5 d5d1 e1d1 f6g7 f1c4 g8h6 e5e6 c8e6 h1e1 e8d7 c4e6 f7e6 d1c2 h6g4 b1d2 a8f8 d2e4 d7c8 e1e2 h7h6 g5d2 b7b6 h4h5 g6g5 f3e1 g4e5 f2f3 f8d8 e4f2 d8d5 a1d1 h8d8 d2c1 d5d1 f2d1 c6e7');
INSERT INTO evaluations VALUES(23048,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(23048,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(23048,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(23048,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(23048,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(23048,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(23048,36,'cp 90','d4f5 b1c3 f8c5 d1a4 g8f6 f1d3 f5h4 g2g3 h4g6 h2h4 h7h5 c3e4 f6e4 d3e4 e8g8 a4d1 f7f5 e4c2 e5e4 d1h5 g6e5 c1d2 d8e8 h5e2 b7b5 h4h5 b5c4 h5h6 e5d3 c2d3 e4d3 e2h5 e8h5 h1h5 g7h6 a1c1 c8b7 c1c4 c5d6 h5h6 b7d5 c4h4 a7a5 d2c3 d6b4 g1h3 b4c3 b2c3 f8e8 h3g5');
INSERT INTO evaluations VALUES(23060,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(23060,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(23060,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(23060,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(23060,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(23060,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(23060,36,'cp 118','g1f3 f8b4 b1d2 b4d2 f3d2 d7d6 b2b3 g8h6 c1b2 e8g8 g2g3 f5e7 f1g2 e7g6 h2h3 c8d7 d1c2 d8e7 e1c1 b7b6 a2a3 f7f6 d1g1 a8d8 c1b1 d7f5 e3e4 f5d7 h3h4 f6f5 h4h5 g6h8 f2f3 h8f7 b1a1');
INSERT INTO evaluations VALUES(23915,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(23915,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(23915,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(23915,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(23915,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(23915,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(23915,36,'cp 109','c3d4 c5b6 b1c3 g8e7 f1e1 e8g8 h2h3 e7g6 a2a4 a7a6 c1g5 f7f6 b5c4 g8h8 g5c1 b6a5 c1d2 c6e7 c3d5 a5d2 d1d2 e7d5 c4d5 c7c6 d5b3 d7e8 b3c2 a6a5 a1b1 d8c7 b1c1 e8f7 d4d5 c6c5 f3h2 c7b6 h2f1 b6b2 c1b1');
INSERT INTO evaluations VALUES(23914,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(23914,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(23914,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(23914,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(23914,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(23914,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(23914,36,'cp 89','c5b6 b1c3 g8e7 h2h3 e8g8 c1e3 e7g6 b5c4 c6a5 c4d3 a5c6 a1c1 c6e7 f1e1 c7c6 b2b4 d6d5 e4d5 e7d5 c3d5 c6d5 a2a4 a7a5 b4b5 a8c8 d1b3 d7e6 c1c8 e6c8 e3g5 d8d6 d3g6 d6g6 b3d5 h7h6 g5h4 c8h3 h4g3');
INSERT INTO evaluations VALUES(23901,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(23901,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(23901,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(23901,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(23901,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(23901,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(23901,36,'cp 91','b1c3 g8e7 h2h3 e8g8 b5c4 c6a5 c4e2 a5c6 f1e1 e7g6 e2c4 c6e7 c1d2 c7c6 d1b3 h7h6 a1d1 d8c8 c4f1 d6d5 e4d5 c6d5 c3d5 e7d5 b3d5 d7c6 d5h5 f8d8 f3e5 g6e5 d4e5 c8e6 b2b3 a8c8 d2e3 d8d1 h5d1 c8d8 d1h5 d8d5 e3b6 a7b6 f2f4');
INSERT INTO evaluations VALUES(23900,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(23900,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(23900,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(23900,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(23900,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(23900,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(23900,36,'cp 81','g8e7 b5c4 e8g8 h2h3 c6a5 c4e2 a5c6 f1e1 e7g6 c3d5 h7h6 a2a4 b6a5 c1d2 a5d2 d1d2 c6e7 d5c3 c7c6 e2c4 a8b8 e1e3 b8c8 c4b3 d7e6 d4d5 c6d5 f3d4 e7c6 d4c6 b7c6 e4d5 c6d5 c3d5 d8g5 a1d1 e6d5 b3d5 g6f4');
INSERT INTO evaluations VALUES(26350,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(26350,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(26350,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(26350,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(26350,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(26350,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(26350,36,'cp 94','g7g6 d4d5 g4f3 g2f3 c6e5 b1c3 f8g7 e1g1 g8f6 f3f4 e5d7 e2f3 e8g8 c1d2 a7a6 a1c1 d8b6 c1c2 f8c8 d1e2 f6e8 f1c1 b6d8 d2e3 c8c7 f3g4 a8c8 e2d2 g7h6 g4h3 h6g7 b2b3 c7c3 c2c3 g7c3 c1c3 c8c3 d2c3');
INSERT INTO evaluations VALUES(30713,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(30713,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(30713,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(30713,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(30713,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(30713,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(30713,36,'cp 88','e4d5 g8f6 a2a3 f8e7 g2g4 e7c5 f1g2 d8e7 g1e2 f6g4 c3e4 f7f5 e4c5 e7c5 e1g1 e8g8 b2b4 c5c4 f1e1 d7d6 d1d4 c4d4 e2d4 b8d7 c1b2 g4f6 d4f5 a8e8 c2c4 f6h5 e1e8 f8e8 g2f1 d7f6 f5d4 h5f4 a3a4 a7a5 b4a5 b6a5 d4b5 f4e2 g1h1 f6e4 b5c7');
INSERT INTO evaluations VALUES(30710,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(30710,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(30710,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(30710,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(30710,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(30710,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(30710,36,'cp 67','g8f6 d1f3 f8b4 c1d2 e8g8 e1c1 c7c6 f1d3 c6d5 g1e2 b8a6 d3a6 b7a6 d2g5 a6e2 c3e2 b4e7 g5f6 e7f6 d1d5 d8c7 e2g3 a8e8 h1d1 c7c4 c1b1 e8e7 c2c3 f8e8 f3d3 c4d3 d1d3 e8d8 b1c2 g7g6 g3f1 e7e1 f1e3 e1e2 d3d2 e2d2 c2d2 g8f8 d2c2 h7h5 a2a4 f8e7');
INSERT INTO evaluations VALUES(30911,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(30911,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(30911,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(30911,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(30911,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(30911,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(30911,36,'cp -25','b5d7 d8d7 g1f3 d7e6 d1e2 b8c6 b1c3 e8d7 e1g1 e6e2 c3e2 a8e8 f1e1 c6b4 e2f4 e8e1 f3e1 g8f6 c2c3 b4c6 e1c2 f8d6 f4e2 a7a5 g1f1 b7b5 c1g5 f6e8 d3d4 c5c4 c2e3 e8c7 g5f4 g7g6 f4d6 d7d6 b2b3 h8b8 h2h4 h7h5 g2g3 c4b3 a2b3 b5b4 c3c4 d5c4');
INSERT INTO evaluations VALUES(30860,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(30860,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(30860,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(30860,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(30860,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(30860,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(30860,36,'cp -21','d8d7 g1f3 d7e6 d1e2 b8c6 b1c3 e8d7 e1g1 e6e2 c3e2 a8e8 f1e1 c6b4 e2f4 e8e1 f3e1 g8f6 c2c3 b4c6 e1c2 f8d6 f4e2 a7a5 g1f1 b7b5 c1g5 f6e8 a2a3 e8c7 d3d4 c5c4 c2e3 h8b8 h2h4 b5b4 a3b4 a5b4 g5f4 g7g6 f4d6 d7d6 f1e1 c6e7 e1d2 f7f6 e3d1 e7f5 d1e3 f5e3');
INSERT INTO evaluations VALUES(31739,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(31739,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(31739,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(31739,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(31739,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(31739,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(31739,36,'cp 90','b8c6 f4g3 e7e6 a2a3 f8d6 c2c4 c6e7 d1b3 b7b6 b1c3 c7c6 f3h4 f5g6 f1e2 g6f7 b3c2 g8h6 e2d3 d5c4 d3c4 e8g8 a1c1 g8h8 c3e4 d6c7 e1g1 h6f5 h4f5 e7f5 c4d3 f5e7 e4d2 a7a5 f1e1 f7g8 d2f3 c7g3 h2g3 a8c8 b2b4 a5b4');
INSERT INTO evaluations VALUES(33946,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(33946,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(33946,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(33946,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(33946,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(33946,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(33946,36,'cp -9','b1c3 d5a5 e1g1 b8c6 f1e1 c8e6 d2d3 f7f6 a2a3 f8b4 c1d2 b4c3 d2c3 a5d5 d3d4 e5e4 f3d2 f6f5 d1h5 e6f7 h5h4 e8g8 f2f3 e4f3 d2f3 a8d8 f3e5 f8e8 e5f7 d5f7 e1e8 d8e8 a1e1 c6e7 h4f2 e7d5 h2h3 d5c3 b2c3 e8e1 f2e1 h7h6 a3a4 f7c4 e1e5 c4c3');
INSERT INTO evaluations VALUES(33937,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(33937,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(33937,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(33937,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(33937,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(33937,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(33937,36,'cp -3','d5a5 e1g1 b8c6 f1e1 c8e6 d2d3 f7f6 a2a3 f8b4 c1d2 b4c3 d2c3 a5d5 d3d4 e5d4 f3d4 c6d4 d1d4 d5d4 c3d4 e8f7 h2h4 h8d8 d4e3 b7b6 a1d1 d8d1 e1d1 f7e7 f2f3 h7h5 d1d2 a8d8 g1f2 d8d7 d2d7 e7d7 b2b4 c7c6 f2g3 e6f5 c2c3 d7e7 g3f4');
INSERT INTO evaluations VALUES(35812,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(35812,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(35812,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(35812,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(35812,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(35812,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(35812,36,'cp -11','e4e5 d6e5 f3e5 c8d7 c1f4 b8c6 a2a4 d7e8 f1e1 e7d6 e5c6 e8c6 f4d6 d8d6 d1e2 b7b6 e2e5 d6d7 c4b5 a8d8 h2h3 a7a5 b5c6 d7c6 b2b3 f6d5 c3d5 d8d5 e5g3 c6d6 g3d6 d5d6 f2f4 f7f6 e1e4 g8f7 g1f2 f8c8 g2g3 c8d8 e4c4 h7h6 a1e1');
INSERT INTO evaluations VALUES(36468,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(36468,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36468,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36468,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(36468,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36468,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(36468,36,'cp 18','c7c6 e2e3 a7a6 f1d3 b7b5 e1g1 h7h6 g5h4 c8b7 a1b1 b8d7 b2b4 b5c4 d3c4 a6a5 b4b5 d7b6 c4e2 c6b5 h4f6 g7f6 c3b5 b7e4 e2d3 e4d3 d2d3 b6d5 d3b3 e7e6 e3e4 d5c7 b5c7 d8c7 f1c1 c7d7 g2g3 a8c8 a2a4 c8c1 b1c1 f8c8 c1c8 d7c8 g1g2');
INSERT INTO evaluations VALUES(36531,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36531,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(36531,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36531,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36531,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(36531,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36531,36,'cp -12','f6e4 b1d2 e4d2 d1d2 b8d7 a1d1 b7b6 f1e2 c8b7 e1g1 a7a5 f4h2 e7e6 f3e1 f7f5 e1d3 d8e7 d3f4 g8h8 b2b3 e6e5 d4e5 g7e5 f4d5 b7d5 h2e5 e7e5 d2d5 e5d5 d1d5 a8e8 d5d2 d7c5 e2f3 g6g5 f3c6 e8e7 g2g3 f5f4 e3f4 g5f4 c6d5');
INSERT INTO evaluations VALUES(39133,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(39133,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(39133,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(39133,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(39133,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(39133,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(39133,36,'cp 70','d2d4 c5b6 a2a4 b8c6 c2c3 d7d5 c4b5 a7a6 b5c6 b7c6 a4a5 b6a7 e4e5 c6c5 d1g4 e8f8 g1f3 c5d4 c3d4 g8e7 b2b4 h7h5 g4f4 a8b8 e1g1 f8g8 f4d2 e7c6 c1a3 g7g5 b4b5 b8b5 f1c1 c8d7 b1c3 b5a5 c3e2 g5g4 f3g5 c6d4 e2d4 a7d4 d2d4 d8g5 c1c7 d7b5 a3e7 g5f5 e7f6 a5a1 d4a1');
INSERT INTO evaluations VALUES(40168,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(40168,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(40168,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(40168,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(40168,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(40168,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(40168,36,'cp -10','b1d2 h7h6 g5h4 a7a6 a2a4 b8c6 c2c3 e8g8 e1g1 a8b8 f1e1 b7b5 c4b3 b5b4 b3c4 c8b7 a1c1 b4c3 b2c3 f6h5 h4e7 c6e7 d3d4 c5d4 c3d4 d6d5 c4f1 d5e4 d2e4 b7e4 e1e4 h5f6 e4e5 f6d7 e5e1 b8a8 d1d2 d7f6 g2g3 a8b8');
INSERT INTO evaluations VALUES(40256,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(40256,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40256,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(40256,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(40256,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(40256,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(40256,36,'cp 73','c5d4 g4g7 h8g8 g7h7 d8c7 h7d3 c7e5 g1e2 d4c3 d3c3 e5c3 e2c3 e6e5 h2h4 c8f5 c1g5 b8c6 f2f4 d5d4 c3b5 e8d7 f4e5 f5c2 a1c1 d4d3 g2g3 c6e5 f1h3 f7f5 b5d4 a8c8 e1g1 d7d6 h3g2 g8g5 h4g5 c8c4 d4c2 d3c2 f1f2 c4c3 c1c2 c3g3 g1h2 g3a3 f2f1');
INSERT INTO evaluations VALUES(45598,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(45598,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(45598,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(45598,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(45598,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(45598,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(45598,36,'cp -31','b7b6 c5b6 a7b6 g1f3 f8b4 c1d2 e8g8 f1d3 b8d7 e1g1 c8b7 f1e1 h7h6 d1e2 c6c5 a2a3 b4a5 d3b5 d8e7 e1c1 f8c8 d2e1 e6e5 d4e5 d7e5 f3e5 e7e5 f2f3 e5e6 a1b1 d5d4 b2b4 a5b4 a3b4 d4c3 e1c3 f6d5 e3e4 d5c3 c1c3');
INSERT INTO evaluations VALUES(46049,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46049,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(46049,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(46049,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(46049,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46049,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(46049,36,'cp 26','f8e7 c1e3 b8c6 g2g4 d6d5 e4d5 f6d5 c3d5 d8d5 e1g1 c6d4 d1d4 d5d4 e3d4 c8d7 d4g7 h8g8 g7d4 h7h5 f2f3 f7f5 c2c3 h5g4 f3g4 a7a6 a1e1 e8c8 d4b6 d8e8 h2h3 e6e5 e2c4 f5g4 c4g8 e8g8 h3g4 g8g4 g1h2 e7d6');
INSERT INTO evaluations VALUES(40205,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40205,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40205,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(40205,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40205,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(40205,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(40205,36,'cp 16','c1e3 b8c6 g2g4 d6d5 e4d5 f6d5 c3d5 d8d5 e1g1 c6d4 d1d4 d5d4 e3d4 e7f6 a1d1 c8d7 d4f6 g7f6 f2f4 a8c8 c2c4 e8e7 f1e1 e6e5 g1f2 b7b5 c4b5 c8c2 f4e5 d7b5 d1d5 b5c6 d5a5 f6e5 b2b3 e7f6 f2e3 h7h5 g4h5 c2c3 e3d2');
INSERT INTO evaluations VALUES(47426,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(47426,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(47426,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(47426,33,'cp -120',NULL);
INSERT INTO evaluations VALUES(47426,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(47426,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(47426,36,'cp -123','c8d7 a4b3 c7c5 g1f3 b8c6 b1a3 d8c7 a3b5 c7b8 a2a4 e7e6 f1e2 f8e7 d2d3 e6e5 e1g1 e8g8 c2c3 h7h6 b5a3 d7e6 b3c2 b8c7 f1e1 a8d8 e2f1 f6d7 c2e2 g8h8 g2g3 f7f5 f1g2 a7a6 f3d2 e6g8 a3c4');
INSERT INTO evaluations VALUES(30317,30,'cp -120',NULL);
INSERT INTO evaluations VALUES(30317,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(30317,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(30317,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(30317,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(30317,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(30317,36,'cp -128','a4b3 c7c5 g1f3 b8c6 b1a3 d8c7 a3b5 c7b8 a2a4 e7e6 f1e2 f8e7 d2d3 e6e5 e1g1 e8g8 c2c3 h7h6 f1e1 d7e6 b3c2 a7a6 b5a3 b8c7 a3c4 f6d7 b2b3 f7f5 e2f1 a8d8 c1b2 e7f6 g2g3 b7b5 a4b5 a6b5 c4a3 g8h8 f1g2 c7b7 f3d2');
INSERT INTO evaluations VALUES(48411,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48411,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48411,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48411,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(48411,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(48411,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(48411,36,'cp 9','g1f3 c7c5 f1g2 f8e7 c2c4 d5d4 b3b4 e8g8 e1g1 f8e8 d2d3 e7f8 b4b5 h7h6 b1d2 e6e5 a2a4 a7a6 f1e1 a8a7 e2e3 b8d7 d1c2 b7b6 f3h4 c8b7 g2b7 a7b7 e3e4 b7a7 e1f1 d8a8 f2f4 e5f4 g3f4 a6b5 a4b5');
INSERT INTO evaluations VALUES(50466,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(50466,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(50466,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(50466,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(50466,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(50466,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(50466,36,'cp 120','g1f3 d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 g8f6 e4e5 f6e4 c4c5 b7b6 c5b6 e4c3 b2c3 d8b6 f1d3 c6c5 e1g1 h7h6 a1b1 b6c7 d1d2 c8d7 f1c1 e8g8 d3c2 d7b5 g2g4 b8c6 d2e3 b5c4 g4g5 h6h5 g5g6 f7g6 f3g5 c4a2 c2g6 a2b1 c1b1 c7e7 h2h4 g8h8 g6c2');
INSERT INTO evaluations VALUES(50465,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(50465,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(50465,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(50465,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(50465,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(50465,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(50465,36,'cp 111','d7d5 b1c3 f8b4 c1d2 b4c3 d2c3 g8f6 e4e5 f6e4 c4c5 e4c3 b2c3 b7b6 c5b6 d8b6 f1d3 c6c5 e1g1 h7h6 c3c4 c8b7 d4c5 b6c5 a1b1 b7c6 c4d5 c5d5 d1c2 d5d7 f1d1 c6a4 c2c5 d7c6 c5b4 a4d1 b1d1 b8d7 d3e4 a8b8 e4c6 b8b4 d1d7 e8g8 g1f1 b4b2 a2a4');
INSERT INTO evaluations VALUES(52121,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52121,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(52121,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(52121,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52121,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52121,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52121,36,'cp -5','b8c6 b1c3 g8f6 f1c4 d7d6 g1e2 g7g6 d2d3 f8g7 c3d5 f6d5 c4d5 c6e7 d5c4 c8e6 d1d2 e8g8 f2f4 d8d7 a2a4 a7a6 a4a5 e7c6 e1g1 a8e8 f4f5 g6f5 d2g5 f5e4 d3e4 e6c4 b3c4 f7f5 f1f5 c6d4 e2d4 c5d4 a1f1 f8f5');
INSERT INTO evaluations VALUES(54411,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(54411,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(54411,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(54411,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(54411,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(54411,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(54411,36,'cp 79','f8b4 g2g3 g8e7 f1g2 c8d7 e1g1 b4c3 b2c3 d8a5 c1d2 a5a4 h2h4 h7h6 f3e1 b8a6 g2h3 a6c7 e1g2 c7b5 g2e3 e8c8 d1f3 d8f8 f1b1 b5a3 b1b4 a4a5 h3g2 e7c6 b4b2 c8b8 h4h5 b8a8 e3g4 c6e7');
INSERT INTO evaluations VALUES(38171,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(38171,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(38171,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(38171,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(38171,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(38171,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(38171,36,'cp 94','g2g3 c8d7 h2h4 h7h6 f1h3 b4c3 b2c3 d8a5 c1d2 a5a4 e1g1 b8c6 f3e1 e8c8 e1g2 c8b8 d1c1 g8e7 g2e3 f7f6 e5f6 g7f6 h3g2 e6e5 a1b1 d7c8 c1d1 b8a8 f1e1 f6f5 d1h5 h8g8 h5f7');
INSERT INTO evaluations VALUES(54703,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(54703,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(54703,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(54703,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(54703,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(54703,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(54703,36,'cp 30','b8c6 g1f3 c8g4 e2e4 d5e4 c3e4 e7e5 d4e5 d8d1 a1d1 f6e5 f4g3 g8f6 f1d3 e8c8 h2h3 g4h5 e4f6 g7f6 d3f5 c8b8 d1d8 c6d8 g3h4 f8g7 g2g4 h5g6 f3d2 h8f8 d2e4 d8c6 e4c5 f8f7 c5e6 g7f8 e1e2 f8d6 e2e3 c6e7 e3e4 b7b5 c2c3 a7a5 h4g3 a5a4 f2f4 e5f4');
INSERT INTO evaluations VALUES(57451,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(57451,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(57451,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(57451,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(57451,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(57451,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(57451,36,'cp 106','b1c3 d7d6 g1e2 g8h6 e1g1 b8a6 c3d5 c7c6 c4a6 b7a6 d5f4 a8b8 b2c3 h6f7 e2d4 c8d7 f1e1 f7e5 d4f5 e8f7 a1c1 g7g6 f5d6 f7g8 d1d2 d8e7 c1d1 d7g4 f4e2 e7d6 d2d6 f8d6 d1d6 g4e2 e1e2 b8b1 e2e1 b1e1 c3e1 g8f7 f2f4 e5g4 h2h3 g4e3 d6c6 h8e8 c6c7 f7f8');
INSERT INTO evaluations VALUES(57629,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(57629,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(57629,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(57629,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(57629,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(57629,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(57629,36,'cp 49','g8f6 g1f3 d7d6 g2g3 f8g7 f1g2 e8g8 e1g1 c7c6 d1b3 b8a6 c1f4 f6h5 c4c5 g8h8 c5d6 e7d6 f4g5 d8e8 f1e1 h7h6 g5d2 d6d5 e2e3 h5f6 c3e2 f6e4 d2a5 g6g5 a1c1 f8f7 f3e5 f7e7 f2f3 e4d6 a5c3 a6c7');
INSERT INTO evaluations VALUES(4436,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(4436,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4436,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4436,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(4436,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(4436,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4436,36,'cp 0','d7d6 a2a3 b4c3 d2c3 c6e7 d1d2 c8e6 d2e2 c7c5 e1g1 e7g6 f1e1 d8d7 a3a4 a7a6 a4a5 e6c4 d3c4 g6f4 e2d1 d7g4 g2g3 f4h3 g1g2 h3f4');
INSERT INTO evaluations VALUES(3881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3881,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3881,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3881,36,'cp 0','a2a3 b4c3 d2c3 c6e7 e1g1 e7g6 h2h3 h7h6 f1e1 c8e6 f3d2 f8e8 d1f3 d6d5 e4d5 e6d5 f3g3 f6h5 g3g4 h5f6');
INSERT INTO evaluations VALUES(4712,30,'cp -409',NULL);
INSERT INTO evaluations VALUES(4712,31,'cp -404',NULL);
INSERT INTO evaluations VALUES(4712,32,'cp -399',NULL);
INSERT INTO evaluations VALUES(4712,33,'cp -410',NULL);
INSERT INTO evaluations VALUES(4712,34,'cp -406',NULL);
INSERT INTO evaluations VALUES(4712,35,'cp -412',NULL);
INSERT INTO evaluations VALUES(4712,36,'cp -394','d7d5 d2d3 c8g4 d1d2 d5e4 c3e4 f6e4 d3e4 d8d2 c1d2 a7a5 g1h1 b7b6 f2f3 g4e6 f3f4 a8d8 f4e5 f7e8 d2f4 d8d4 b2b3 d4e4 c2c3 c5e3 f4g3 e4g4 f1f3 e3c5 f3f4 g4g5 a1e1 g5g6 e1f1 h7h6 c3c4 e8d7 f1d1 d7c8 g3f2 c5e7 f2g1 c8b7');
INSERT INTO evaluations VALUES(6495,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(6495,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(6495,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(6495,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(6495,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(6495,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(6495,36,'cp -25','g8f6 b1c3 d7d6 c3a4 c5b6 f1e1 e8g8 h2h3 d6d5 e4d5 d8d5 b2b4 f8e8 c1b2 c6b4 a4b6 a7b6 f3e5 b4d3 e5d3 e8e1 d1e1 c8d7 b2f6 g7f6 e1e3 d5g5 e3g3 g5g3 f2g3 a8a3 d3f4 g8f8 g3g4 d7c6 g1h2 h7h6 f4e2 b6b5 e2f4 b5b4 f4d3');
INSERT INTO evaluations VALUES(8102,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(8102,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(8102,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(8102,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(8102,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(8102,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(8102,36,'cp 40','f6h5 f4e3 e7d6 f3e5 g7g6 g2g4 h5f6 d2f3 f6d7 e5c6 b7c6 e3h6 d8c7 e1g1 d6f4 g4g5 f7f5 c3c4 e8f7 f1e1 a8b8 d1c2 h8e8 g1g2 c8b7 c4d5 c6d5 c2c7 f4c7 d3b5 e8e7 b5d7 e7d7 b2b3 c7f4 f3e5 f4e5 e1e5 b8c8 a1e1 d7d6 g2g3 c8c2 e5e2 c2e2 e1e2');
INSERT INTO evaluations VALUES(12287,30,'cp 348',NULL);
INSERT INTO evaluations VALUES(12287,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(12287,32,'cp 339',NULL);
INSERT INTO evaluations VALUES(12287,33,'cp 359',NULL);
INSERT INTO evaluations VALUES(12287,34,'cp 342',NULL);
INSERT INTO evaluations VALUES(12287,35,'cp 332',NULL);
INSERT INTO evaluations VALUES(12287,36,'cp 336','d4e5 d8e7 e5e7 f8e7 e4e5 f6h7 c1e3 b7b6 c3b5 e8g8 e1c1 f8d8 c4d5 a8b8 b5a7 c8a6 a2a4 h7f8 c2c4 f8e6 a7b5 a6b5 c4b5 g8f8 g2g3 d7d6 d5e6 f7e6 e5d6 e7d6 b2b3 f8e7 c1c2 d6c7 d1d8 b8d8 h1d1 d8d5 a4a5 b6a5 d1d5 e6d5 e3c5 e7d7 c2d3');
INSERT INTO evaluations VALUES(12286,30,'cp 340',NULL);
INSERT INTO evaluations VALUES(12286,31,'cp 346',NULL);
INSERT INTO evaluations VALUES(12286,32,'cp 339',NULL);
INSERT INTO evaluations VALUES(12286,33,'cp 335',NULL);
INSERT INTO evaluations VALUES(12286,34,'cp 338',NULL);
INSERT INTO evaluations VALUES(12286,35,'cp 347',NULL);
INSERT INTO evaluations VALUES(12286,36,'cp 361','d8e7 e5e7 f8e7 e4e5 f6h7 c1e3 b7b6 c3b5 e8g8 e1c1 f8d8 c4d5 a8b8 b5a7 c8a6 c2c4 h7f8 a2a4 f8e6 a7b5 a6b5 c4b5 g8f8 d1d2 e6d4 e3d4 c5d4 d2d4 d7d6 e5e6 e7f6 d4e4 f7e6 e4e6 d8e8 h1d1 e8e6 d5e6 b8a8 e6b3 f8e7 g2g3 e7d7 f2f4 a8a5');
INSERT INTO evaluations VALUES(14728,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(14728,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(14728,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(14728,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(14728,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(14728,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(14728,36,'cp 88','g7f6 f1c4 h8g8 e1g1 e6e5 f1e1 c8g4 g1f1 d8d7 d4e5 d7d1 e1d1 g4f3 g2f3 f6e5 c1e3 f8d6 c2c3 c6d8 c4d5 a7a5 c3c4 d6e7 d5e4 c7c6 e4h7 g8h8 h7f5 d8e6 f5e6 f7e6 f1g2 a5a4 h2h3 h8g8 g2h2 g8f8 d1g1 f8f3 g1g8 f3f8 g8g7 f8f7 a1g1 f7g7 g1g7 e8f8');
INSERT INTO evaluations VALUES(15549,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(15549,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(15549,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(15549,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(15549,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(15549,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(15549,36,'cp 41','e5d4 f3d4 f8e7 f1b5 c8d7 h2h3 e8g8 d4c6 d7c6 b5c6 b7c6 e1g1 f6d7 d1d2 e7f6 b2b3 d8b8 a2a3 f8e8 f2f3 a7a5 a1d1 c6c5 d1b1 f6c3 d2c3 d7e5 f1e1 h7h6 f3f4 e5c6 e3f2 b8b7 a3a4 b7b4 c3d3 b4b6 b1d1');
INSERT INTO evaluations VALUES(16850,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(16850,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(16850,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(16850,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(16850,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(16850,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(16850,36,'cp -34','d7d6 d2d4 f8e7 d1d3 e8g8 b1c3 c8g4 d4e5 d6e5 c3d5 f7f5 d5e7 d8e7 e4f5 g4f5 d3d5 e7f7 d5f7 f8f7 c2c3 h7h6 c1e3 a7a5 a2a4 b7b6 f1e1 a8d8 f3d2 f7d7 d2f1 f5c2 f1g3 d7d3 f2f3 d3d6 g3e4 c2e4 f3e4 c6e7 h2h3 e7g6');
INSERT INTO evaluations VALUES(19345,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(19345,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(19345,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19345,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19345,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19345,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19345,36,'cp 0','b7c6 c2c4 e7e5 g2g3 f8b4 c1d2 b4e7 f1g2 g8f6 e1g1 e8g8 d1a4 a7a5 b1c3 h7h6 a1d1 d7d5 a4c6 c8d7 c6b7 a8b8 b7a7 b8a8');
INSERT INTO evaluations VALUES(18440,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(18440,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(18440,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(18440,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18440,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18440,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(18440,36,'cp 0','c2c4 e7e6 g2g3 g8f6 f1g2 f8b4 c1d2 a8b8 e1g1 d7d5 d2b4 b8b4 b2b3 e8g8 d1c2 e6e5 f1d1 c8e6 b1c3 b4b8 c3e4 f6d7 e4d6 d8b6 a1c1 f8d8 e2e3 d7f6 c4c5 b6b4 d1d2 e5e4 c2c3 b4c3 c1c3');
INSERT INTO evaluations VALUES(20732,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(20732,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(20732,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(20732,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(20732,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(20732,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(20732,36,'cp 104','d2d3 d8c7 e4e5 g8e7 c1f4 e7d5 f4g3 c7b6 d1c1 b6d8 e1g1 f7f5 f1d1 d5b6 c2c4 f8e7 d3d4 d8c7 d4c5 e7c5 g3h4 e8g8 b1d2 b6d7 d2b3 b7b6 f3e1 d7e5 b3c5 b6c5 h4g3 h7h6 e1d3 e5d3');
INSERT INTO evaluations VALUES(20915,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(20915,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(20915,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(20915,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(20915,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(20915,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(20915,36,'cp 137','c2c3 f8g7 f2f4 g8h6 f1e2 e8g8 b1d2 f7f6 h2h4 f6e5 f4e5 d8e8 h4h5 c6d8 g1h3 d8e6 d2f3 h6f5 h5g6 e8g6 g2g4 g6g4 h3f2 g4g6 h1g1 g6f7 f2h3 c7c5 d4c5 g8h8 d1d5 a7a5 d5c4 a5a4 h3f4 e6f4 c4f7 f8f7 c1f4');
INSERT INTO evaluations VALUES(22717,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(22717,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(22717,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(22717,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(22717,34,'cp -119',NULL);
INSERT INTO evaluations VALUES(22717,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(22717,36,'cp -98','d7d5 c4e2 e5e4 f3d4 d8g5 g2g3 c8h3 d4c6 b7c6 c2c4 g8f6 d1a4 h3d7 b1c3 f8d6 h2h3 e8g8 a4a5 h7h5 b2b3 d7e6 c1a3 d6a3 a5a3 d5c4 a3a6 c4b3 a2b3 e6b3 a6c6 a7a5 c6b5 g5b5 e2b5 f8b8 b5c6 a8a7 e1g1 b8b6 f1b1 b3c2 b1b6 c7b6 c3b5 a7e7');
INSERT INTO evaluations VALUES(25603,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(25603,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(25603,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(25603,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(25603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25603,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(25603,36,'cp 15','g8f6 d2d4 f6e4 d4e5 e4c5 b3d5 d7d6 e1g1 f8e7 f3d4 c6d4 d5b7 c5b7 c3d4 e8g8 d1f3 d8c8 b1c3 c7c6 c1f4 d6d5 c3e2 f7f6 f3g3 c8d7 a2a4 a6a5 f1c1 b5a4 a1a4 b7d8 c1a1 d7b7 g3c3 e7b4 c3c2 d8e6 f4g3 f6e5');
INSERT INTO evaluations VALUES(24237,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24237,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(24237,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(24237,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24237,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(24237,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24237,36,'cp 22','d2d4 f6e4 d4e5 e4c5 b3d5 d7d6 e1g1 f8e7 f3d4 c6d4 d5b7 c5b7 c3d4 e8g8 d1f3 d8c8 b1c3 c7c6 c1f4 d6d5 c3e2 f7f6 f3d3 f6e5 f4e5 b7d6 b2b3 c8d7 f2f4 a8e8 a1c1 d6f7 h2h3 e7a3 c1d1 a3b4 e2g3 a6a5 g3f5 g7g6');
INSERT INTO evaluations VALUES(26598,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(26598,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(26598,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(26598,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(26598,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(26598,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(26598,36,'cp 86','d1f3 e7e6 c2c4 f8b4 b1c3 g8f6 a2a3 b4c3 b2c3 d8d7 f1d3 d5c4 d3c4 f6d5 e1g1 d5f4 e3f4 e8g8 d4d5 e6d5 c4d5 a8d8 f1d1 d7e7 g2g3 c6a5 d1e1 e7f6 e1e5 b7b6 g1g2 c7c6 d5a2 d8d2 h2h4 h7h6 f3e3 d2a2 a1a2 a5c4 e3e2 c4e5');
INSERT INTO evaluations VALUES(26597,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(26597,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(26597,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(26597,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(26597,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(26597,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(26597,36,'cp 67','e7e6 c2c4 f8b4 b1c3 g8f6 a2a3 b4c3 b2c3 d8d7 f1d3 d5c4 d3c4 f6d5 e1g1 d5f4 e3f4 e8g8 d4d5 e6d5 c4d5 c6a5 d5b7 a5b7 f3b7 d7d2 b7b4 f8d8 g2g3 a7a5 b4c5 d2d6 c5c4 d6d3 c4c6 d3d6 c6d6 d8d6 f1e1 g8f8 c3c4 d6d4 a1c1');
INSERT INTO evaluations VALUES(28990,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(28990,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(28990,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(28990,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(28990,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(28990,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28990,36,'cp 7','e7e6 c2c4 a7a5 h2h3 b8d7 d1b3 d8c8 f1d3 a5a4 b3c2 f5d3 c2d3 f8e7 a1c1 h7h6 e1g1 e8g8 f3e5 d7e5 f4e5 c8d8 f1d1 d8a5 a2a3 a5a6 e3e4 f8d8 d3g3 b7b6 g3f3 a6b7 e5f4 c6c5 e4e5');
INSERT INTO evaluations VALUES(29260,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29260,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29260,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(29260,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29260,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(29260,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29260,36,'cp 12','f3e5 h7h5 f1d3 b8d7 h2h3 d7e5 f4e5 f6d7 d3f5 e6f5 e5f4 d7f6 e1g1 f8d6 f4g5 d6e7 c2c4 d5c4 d2c4 f6d5 g5f4 e8g8 f4h2 g7g6 d1b3 d5b6 f1c1 b6c4 c1c4 d8b6 b3c3 f8d8 h2e5 d8d7 a2a4 a7a5');
INSERT INTO evaluations VALUES(29855,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(29855,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29855,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29855,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29855,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(29855,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(29855,36,'cp 2','f1d3 f5d3 c2d3 f8d6 d1b3 d8c8 f3e5 e8g8 e1g1 f6d7 b1c3 d6e5 d4e5 c7c5 d3d4 b8c6 b3d1 c5d4 e3d4 c8d8 a2a4 d8b6 f4e3 f7f6 e5f6 d7f6 c3e2 a8e8 e2f4 f6e4 a1c1 e4d6 f4g6 f8f6 g6e5 b6b2 e5c6 b7c6 c1c6');
INSERT INTO evaluations VALUES(31293,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(31293,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(31293,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(31293,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(31293,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(31293,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(31293,36,'cp -75','b8c6 f1b5 a8c8 d2d4 e7e6 c1e3 c5c4 e1g1 h7h5 b1c3 f8e7 e3f2 f5g6 a2a3 a7a6 b5c6 c8c6 h2h3 g8h6 d1d2 h6f5 g1h2 c6b6 c3d1 b6b5 a3a4 b5b6 d2c1 d8d7 f2e1 h5h4 e1a5 b6c6 a5e1 c6c8 e1f2 e8g8 a4a5');
INSERT INTO evaluations VALUES(26281,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(26281,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(26281,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(26281,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(26281,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(26281,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(26281,36,'cp -64','f1b5 a8c8 a2a4 e7e6 a4a5 a7a6 b5c6 c8c6 d2d3 g8e7 e1g1 h7h5 h2h3 c6c8 f3h4 g7g6 c1d2 e7c6 h4f5 g6f5 b1c3 f8e7 g1h2 d8c7 d1e1 c6b4 e1d1 c8d8 c3e2 b4c6 b2b3 d5d4 e2g1 e8d7 g1f3 d7c8 d1e2 d8g8 f1g1 c6b4');
INSERT INTO evaluations VALUES(26257,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(26257,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(26257,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(26257,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(26257,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(26257,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(26257,36,'cp -46','a8c8 a2a4 e7e6 a4a5 a7a6 b5c6 c8c6 d2d3 g8e7 e1g1 c6c8 f3h4 h7h6 c1e3 f5h7 b1d2 e7c6 d1g4 c6b4 a1c1 d8a5 f4f5 a5b6 d2b3 e6f5 h4f5 h7f5 g4f5 b6e6 f5f2 b7b6 f2g3 h6h5 e3g5 e6g4 g3e3 d5d4 e3d2 g4e6 d2f2 f8e7 g5e7 e6e7 f2f3 g7g6 b3d2 e8g8 d2c4 e7e6');
INSERT INTO evaluations VALUES(33280,30,'cp -499',NULL);
INSERT INTO evaluations VALUES(33280,31,'cp -509',NULL);
INSERT INTO evaluations VALUES(33280,32,'cp -535',NULL);
INSERT INTO evaluations VALUES(33280,33,'cp -500',NULL);
INSERT INTO evaluations VALUES(33280,34,'cp -517',NULL);
INSERT INTO evaluations VALUES(33280,35,'cp -540',NULL);
INSERT INTO evaluations VALUES(33280,36,'cp -517','h4e4 d1e2 e4h1 e5g6 g8e7 g6h8 b8c6 c2c3 d7d5 d2d3 c8e6 c1f4 h1h2 b1d2 e8c8 d2f3 h2h5 b2b4 c5d6 b4b5 d6f4 b5c6 f4d6 f3d4 h5e2 f1e2 d8h8 c6b7 c8b8 d4e6 f7e6 d3d4 h7h5 a1b1 h5h4 g3h4 e6e5 d4e5 d6e5 h4h5 e5c3 e1f1 h8f8 b1b3 d5d4 f1g1 e7d5 a2a3 f8e8 e2f3 d5f4 h5h6 g7h6');
INSERT INTO evaluations VALUES(33283,30,'cp -524',NULL);
INSERT INTO evaluations VALUES(33283,31,'cp -527',NULL);
INSERT INTO evaluations VALUES(33283,32,'cp -536',NULL);
INSERT INTO evaluations VALUES(33283,33,'cp -521',NULL);
INSERT INTO evaluations VALUES(33283,34,'cp -532',NULL);
INSERT INTO evaluations VALUES(33283,35,'cp -527',NULL);
INSERT INTO evaluations VALUES(33283,36,'cp -535','d1e2 e4h1 e5g6 g8e7 g6h8 b8c6 c2c3 d7d5 d2d3 c8e6 c1f4 h1h2 b1d2 e8c8 b2b4 c5d6 d2f3 h2h5 b4b5 d6f4 b5c6 f4d6 f3d4 h5e2 f1e2 d8h8 c6b7 c8b8 d4e6 f7e6 d3d4 h7h5 e1f1 h8f8 f1g2 h5h4 e2a6 h4g3 f2g3 e7f5 a1e1 f5g3 c3c4 g3f5 c4c5 d6g3 e1e6 g3f2 c5c6 f5e3 e6e3 f2e3 a6d3 e3d4 g2g3 d4e5 g3h4');
INSERT INTO evaluations VALUES(33284,30,'cp -538',NULL);
INSERT INTO evaluations VALUES(33284,31,'cp -540',NULL);
INSERT INTO evaluations VALUES(33284,32,'cp -546',NULL);
INSERT INTO evaluations VALUES(33284,33,'cp -530',NULL);
INSERT INTO evaluations VALUES(33284,34,'cp -542',NULL);
INSERT INTO evaluations VALUES(33284,35,'cp -546',NULL);
INSERT INTO evaluations VALUES(33284,36,'cp -535','e4h1 e5g6 g8e7 g6h8 b8c6 c2c3 d7d5 d2d3 c8e6 b1d2 e8f8 d2b3 c5d6 f2f4 e6h3 c1d2 h1f1 e2f1 h3f1 e1f1 f8g8 h8f7 g8f7 d2e3 b7b6 e3f2 f7e8 b3d2 e8d7 d2f3 e7f5 a1e1 a8e8 e1e8 d7e8 c3c4 f5e7 c4c5 b6c5 a2a3 e8d7 f3g5 h7h5 g5f3 e7f5 h2h3 a7a5 g3g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(33285,30,'cp -513',NULL);
INSERT INTO evaluations VALUES(33285,31,'cp -521',NULL);
INSERT INTO evaluations VALUES(33285,32,'cp -509',NULL);
INSERT INTO evaluations VALUES(33285,33,'cp -511',NULL);
INSERT INTO evaluations VALUES(33285,34,'cp -526',NULL);
INSERT INTO evaluations VALUES(33285,35,'cp -481',NULL);
INSERT INTO evaluations VALUES(33285,36,'cp -498','e5g6 g8e7 g6h8 b8c6 c2c3 d7d5 d2d3 c8e6 c1f4 h1h2 d3d4 c5d6 f4d6 c7d6 b1d2 e8c8 h8f7 e6f7 e2f3 f7g6 f3g2 h2g2 f1g2 d8e8 e1c1 e7g8 g2d5 g8f6 d5f3 d6d5 g3g4 e8f8 d1h1 f6e4 f3e4 d5e4 h1h2 f8f4 d2c4 f4g4 c1d2 g4g1 c4e3 c8b8');
INSERT INTO evaluations VALUES(36982,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(36982,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(36982,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(36982,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(36982,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(36982,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(36982,36,'cp 60','e5f3 d6d5 c4d5 f6d5 e4d5 f8e8 c1e3 d8d5 b1c3 d5c6 e1g1 c5e3 f2e3 e8e3 d1d2 e3e8 f3d4 c6c5 d2f2 c8e6 a1e1 b8c6 d4e6 c5f2 g1f2 f7e6 f2e3 e8f8 e3d2 f8f1 e1f1 a8d8 f1f4 h7h6 d2e3 c6e7 f4c4 d8d7 c4a4 a7a6 c3e4 e7f5 e3f2 f5d6 e4d2 d7d8 f2e2 d8f8 g2g4 g8f7 a4a5');
INSERT INTO evaluations VALUES(37003,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(37003,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(37003,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(37003,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(37003,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37003,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(37003,36,'cp 60','d6d5 c4d5 f6d5 e4d5 f8e8 c1e3 d8d5 b1c3 d5c6 e1g1 c5e3 f2e3 e8e3 d1d2 e3e8 f3d4 c6c5 d2f2 c8e6 c3a4 c5h5 a1e1 b8d7 h2h3 d7f8 a4c3 e6d7 c3e4 f7f6 b2b4 h5f7 c2c4 a8d8 b4b5 f8g6 a2a4 b7b6 e1e3 e8e5 d4f3 e5f5 f1e1 d8e8');
INSERT INTO evaluations VALUES(39560,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39560,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(39560,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(39560,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39560,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39560,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39560,36,'cp 2','e7e5 d2d4 e5d4 e3d4 g8e7 h2h4 d7d5 h4h5 b8c6 c1e3 d5c4 f1c4 e7f5 g1f3 c6d4 f3d4 f5d4 d1d2 c8e6 e1c1 c7c5 h5h6 g7f6 c4e6 f7e6 c3e4 e8g8 c1b1 d8e7 d2c3 a8c8 f2f3 f8d8 h1e1 b7b6 e3d4 d8d4 d1d4 f6d4 c3b3 c8d8 b3a4');
INSERT INTO evaluations VALUES(42308,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(42308,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(42308,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(42308,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(42308,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(42308,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(42308,36,'cp 14','h4g3 f6e4 d1d3 d7d5 f2f3 e4g3 h2g3 d5c4 d3c4 c7c5 a2a3 d8d4 c4d4 c5d4 a3b4 d4c3 b2c3 e8e7 g1h3 f7f6 e2e4 b8d7 h3f2 b7b6 f1b5 a7a5 e1d2 h6h5 c3c4 c8b7 b5d7 e7d7 f2d3 a5b4 d3b4 a8a1 h1a1 d7c7 d2c3 h8d8 b4d3 e6e5');
INSERT INTO evaluations VALUES(42309,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42309,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(42309,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(42309,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(42309,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(42309,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(42309,36,'cp 14','f6e4 d1d3 d7d5 e2e3 b8c6 g1e2 h6h5 h2h4 e4g3 e2g3 g5h4 g3e2 d5c4 d3c4 d8e7 a2a3 b4d6 c3e4 c8d7 a1c1 e8c8 e4d6 c7d6 b2b4 a7a6 a3a4 d6d5 c4c5 e7c5 c1c5 b7b6 c5c1 c8b7 b4b5 a6b5 a4b5 c6e7');
INSERT INTO evaluations VALUES(42306,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(42306,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(42306,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(42306,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(42306,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(42306,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(42306,36,'cp 15','d1c2 b4c3 b2c3 e4g3 h2g3 b7b6 g3g4 c7c5 e2e3 c8b7 g1e2 b8c6 e2g3 c6a5 f2f3 d8c7 e1f2 e8c8 c2a4 c8b8 a1b1 f7f5 g4f5 h6h5 g3h5 e6f5 f1d3 c5d4 c3d4 g5g4 h5f4 h8h1 b1h1 g4f3');
INSERT INTO evaluations VALUES(42307,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(42307,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(42307,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(42307,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42307,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42307,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(42307,36,'cp 20','b4c3 b2c3 e4g3 f2g3 b7b6 e2e3 c8b7 f1d3 b8c6 c2e2 d8f6 g1h3 h6h5 h3f2 f6g7 e3e4 g5g4 e4e5 d7d6 e5d6 c7d6 e1g1 e8c8 e2e3 c6e7 a2a4 h5h4 g3h4 h8h4 d3e4 d8h8 e4b7 c8b7 a4a5 h4h2 a5b6 a7b6 c4c5');
INSERT INTO evaluations VALUES(46285,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(46285,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(46285,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(46285,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(46285,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(46285,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(46285,36,'cp -3','f6e4 d4c5 e4c3 b2c3 b8d7 c3c4 d8a5 d1d2 a5d2 c1d2 f8c5 c4d5 e6d5 f1d3 e8g8 e1g1 d7b8 f1e1 b8c6 c2c3 c8e6 a1b1 b7b6 d2e3 c5e3 e1e3 f8c8 h2h3 c8c7 b1d1 g7g6 d3a6 a8b8 a6b5 b8d8 b5c6 c7c6 f3d4 c6c5 f2f3 d8c8');
INSERT INTO evaluations VALUES(46072,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(46072,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46072,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46072,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(46072,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(46072,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(46072,36,'cp -2','d4c5 e4c3 b2c3 b8d7 c3c4 d8a5 d1d2 a5d2 c1d2 f8c5 c4d5 e6d5 f1d3 e8g8 e1g1 d7b8 f1e1 b8c6 c2c3 f8e8 h2h3 c8d7 a1b1 b7b6 b1d1 h7h6 d3a6 e8e7 d2c1 a8b8 a2a4 d7e6 f3d4 c6a5 d4e6 f7e6 g2g3 e7c7 g1g2 c5e7 c1d2');
INSERT INTO evaluations VALUES(49177,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(49177,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(49177,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(49177,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(49177,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(49177,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(49177,36,'cp 19','c7c5 c2c3 b8c6 c1e3 e7e6 f2f4 c5d4 c3d4 d7b6 b1c3 f8b4 f1d3 b6c4 d1c1 c8d7 g1e2 b4c3 b2c3 f7f5 d3c4 d5c4 d4d5 e6d5 e3c5 e8f7 e1g1 h8e8 h2h3 f7g8 c1e3 d7e6 g1h2 b7b6 c5d6 d8d7 a2a4 g8h8 e3d2 h7h6 a1e1 e6f7 e2d4');
INSERT INTO evaluations VALUES(53278,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(53278,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(53278,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(53278,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(53278,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(53278,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(53278,36,'cp -64','d5e4 b1d2 g8f6 d4c5 b8c6 f1b5 e7e5 b2b4 a7a5 g1e2 f8e7 e1g1 e8g8 d1b3 c8e6 b5c4 e6c4 b3c4 a5b4 c3b4 d8d3 e2c3 f8d8 c3e4 d3c4 e4f6 g7f6 d2c4 c6b4 a1b1 e7c5 c1e3 c5e7 a2a3 b7b5 c4e5 f6e5 a3b4 a8c8 f1c1 c8c1 b1c1 e7b4 g1f1 b4f8 f1e2 d8d5 c1c6');
INSERT INTO evaluations VALUES(53208,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(53208,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(53208,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(53208,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(53208,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(53208,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(53208,36,'cp -59','b1d2 g8f6 d4c5 b8c6 d1a4 d8d5 d2b3 c8d7 f1c4 d5f5 g1e2 c6e5 a4b4 b7b6 b3d4 f5g4 e2f4 e7e6 c4e2 f8c5 b4b3 e5d3 e2d3 e4d3 e1g1 c5d4 c3d4 a8c8 c1d2 f6h5 f4h5 g4h5 f1c1 d7c6 b3d3 e8g8 c1e1 f8d8 d3g3 h5g6 d2c3 c6b7 g3g6 h7g6');
INSERT INTO evaluations VALUES(53311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53311,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53311,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(53311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53311,36,'cp 3','d2d4 e7e6 b1d2 b8d7 g1f3 g8f6 f1d3 b7b6 e1g1 c8b7 d1e2 f8d6 e3e4 d5e4 d2e4 f6e4 d3e4 b7e4 e2e4 e8g8 e4c6 d7f6 c1g5 a8c8 c6a4 c5d4 g5f6 d8f6 f3d4 f6f4 g2g3 f4d2 a4a7 d2b2 f1b1 b2a3 a7a3 d6a3 b1b6');
INSERT INTO evaluations VALUES(53786,30,'cp -131',NULL);
INSERT INTO evaluations VALUES(53786,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(53786,32,'cp -124',NULL);
INSERT INTO evaluations VALUES(53786,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(53786,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(53786,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(53786,36,'cp -138','f8d6 c2c3 c6c5 g1f3 b8c6 b2b4 c5b4 d3b4 c6b4 c3b4 g8f6 f1b5 f6d7 e1g1 e8g8 a1b1 d7b6 a2a4 c8e6 a4a5 a7a6 b5d3 b6c8 d3c2 c8a7 d2d3 a7b5 f3g5 e6d7 f2f4 e5f4 e4e5 d8g5 c1f4 g5g6 e5d6 b5d6 d1d2 d7c6 c2b3 d6f5 d2f2 g8h8');
INSERT INTO evaluations VALUES(53869,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(53869,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(53869,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(53869,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(53869,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(53869,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(53869,36,'cp 28','a7a6 b5c3 g8f6 a2a4 f8e7 b1a3 e8g8 f1c4 c8e6 c4d5 b8c6 e1g1 a8b8 h2h3 c6b4 f1e1 d8d7 a3c4 b7b5 c4e3 f8e8 d5e6 f7e6 a4b5 a6b5 e4e5 d6e5 d1d7 f6d7 a1a5 g8f7 a5b5 b8b5 c3b5 e8a8 b5a3 d7b6 e1d1 f7f8 e3c4 b6c4');
INSERT INTO evaluations VALUES(50713,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(50713,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50713,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50713,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(50713,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(50713,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(50713,36,'cp 40','b5c3 g8f6 g2g3 b8c6 f1g2 a8b8 e1g1 f8e7 c3e2 e8g8 b1c3 f8e8 e2f4 b7b5 c3d5 c8b7 f1e1 f6d5 f4d5 e7f8 h2h4 h7h6 c1e3 c6e5 b2b3 b7c6 d1d2 a6a5 a1d1 a5a4 d2e2 e5d7 e3f4 b5b4 e2d3 d8a5 g3g4');
INSERT INTO evaluations VALUES(50717,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50717,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50717,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50717,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(50717,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(50717,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(50717,36,'cp 33','g8f6 a2a4 f8e7 f1e2 e8g8 e1g1 b8c6 b1a3 c8e6 f1e1 d6d5 e4d5 f6d5 c3d5 e6d5 a3c4 d5e6 c2c3 d8d1 e1d1 f8d8 c1f4 d8d1 a1d1 a8d8 d1d8 c6d8 f4e3 g8f8 g2g3 f8e8 g3g4 d8c6 f2f4 f7f5 g4f5 e6f5 e2f3 f5e6');
INSERT INTO evaluations VALUES(54567,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54567,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(54567,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54567,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(54567,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54567,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54567,36,'cp 0','e2e4 d5e4 d3e4 d8d1 e1d1 b8c6 c1e3 g8f6 f2f3 f8e7 b1d2 e8g8 g2f1 f8d8 d1c1 c8e6 f1h3 e6c8');
INSERT INTO evaluations VALUES(56032,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(56032,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(56032,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(56032,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(56032,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(56032,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(56032,36,'cp 86','c7c5 e2e4 c5d4 f3d4 a7a6 c1e3 g8f6 f1e2 e6e5 d4b3 d8d1 a1d1 f8b4 e1g1 b4c3 b2c3 f6e4 e2f3 f7f5 b3d2 e4d2 d1d2 e5e4 f3h5 g7g6 h5e2 c8e6 f2f3 b8c6 f3e4 e8g8 e4f5 f8f5 f1f5 g6f5 e3f4 a8c8 g1f2 g8f7 e2f3');
INSERT INTO evaluations VALUES(54102,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(54102,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(54102,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(54102,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(54102,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(54102,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(54102,36,'cp 132','e2e4 c5d4 f3d4 a7a6 c1e3 g8f6 f1e2 f8d6 d1d3 d6e7 e1c1 e8g8 c1b1 e6e5 d4f5 d8d3 d1d3 c8f5 e4f5 b8c6 g2g4 h7h6 h2h4 c6d4 e3d4 e5d4 d3d4 a8d8 d4d1 d8d1 h1d1 f8d8 d1d8 e7d8 e2f3 g8f8 g4g5 f6e8 f3b7 h6g5 h4g5 d8g5');
INSERT INTO evaluations VALUES(56492,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(56492,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(56492,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(56492,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(56492,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(56492,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(56492,36,'cp -86','b8c6 e2e3 f7f5 d2d4 e5e4 h3h4 f8d6 g1h3 g8f6 g2f1 c8e6 b2b3 d8e7 c2c4 e8c8 f1e2 h7h6 h4h5 e6f7 c1d2 c6d4 e3d4 c7c5 b1c3 c5d4 c3d5 f7d5 c4d5 e4e3 d2a5 b7b6 d1d4 b6a5 h3f4 e3f2 e1f1 c8b8 f4g6 e7e8 g6h8 d6g3');
INSERT INTO evaluations VALUES(57817,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(57817,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(57817,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(57817,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(57817,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(57817,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(57817,36,'cp 11','c1g5 d8e8 e1c1 b8a6 f3d2 h7h6 g5h4 c7c6 d2b3 a6c7 f2f3 c7e6 d1d2 e6f4 h1g1 f4e2 c3e2 c8e6 b3a5 a8b8 g1d1 e8c8 c4c5 b7b6 a5b3 g6g5 h4f2 g5g4 c1b1 f6h7 f2e3 g4f3 e2g3 f3g2 d2g2 g8h8 g3f5 c8g8 d1d6 h7f6 b3d2 e6f5 e4f5 b6c5');
INSERT INTO evaluations VALUES(57815,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57815,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57815,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57815,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(57815,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57815,36,'cp 0','d8e8 c3d5 f6d5 c4d5 c7c6 e2c4 c6d5 c4d5 b8d7 f3d2 d7c5 e1g1 c8e6 f1d1 e6d5 e4d5 h7h6 g5e3 c5a4 d2e4 a4b2 d1b1 b2c4 b1b7 f7f5 e3h6 g7h6 e4f6 g8f8 f6d7 f8g8');
INSERT INTO evaluations VALUES(35354,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35354,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35354,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35354,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35354,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35354,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(35354,36,'cp 9','e1c1 b8a6 f3d2 h7h6 g5h4 c7c6 d2b3 a6c7 f2f3 c7e6 h4f2 g7f8 g2g3 h6h5 h2h4 f6d7 c3a4 b7b6 f2e3 e8d8 a4c3 c8a6 d1d2 d7c5 b3c5 f8c5 e3c5 e6c5 h1d1 c5e6 d2d8 a8d8 d1d8 e6d8 c1d2 d8e6 e2f1 e6d4');
INSERT INTO evaluations VALUES(2930,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(2930,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(2930,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(2930,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(2930,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(2930,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(2930,36,'cp 27','e3e4 d5c4 d3c4 e6e5 c1e3 d8e7 h2h3 e5d4 e3d4 d7e5 f3e5 d6e5 f2f4 e5d4 d1d4 b7b5 c4b3 b5b4 e4e5 b4c3 e5f6 e7f6 d4f6 g7f6 b2c3 c8e6 f1d1 f8d8 g1f2 h7h5 f2e3 h5h4 b3e6 f7e6 a1b1 d8d1 b1d1 a8b8 d1d6 g8f7 d6d7 f7g6 d7a7 b8b1');
INSERT INTO evaluations VALUES(2932,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(2932,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(2932,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(2932,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(2932,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(2932,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(2932,36,'cp 21','d5c4 d3c4 e6e5 c1e3 d8e7 h2h3 e5d4 e3d4 d7e5 f3e5 d6e5 f2f4 e5d4 d1d4 b7b5 c4d3 c6c5 d4e5 e7e5 f4e5 f6d7 e5e6 f7e6 d3b5 d7e5 a1c1 a8b8 b2b3 c5c4 b5c4 e5c4 f1f8 g8f8 b3c4 c8d7 c3b5 f8e7 b5d4 e7f6 c1f1 f6e7 d4f3 b8b2 f1f2 b2b4 f3e5 d7e8 a2a3 b4b3');
INSERT INTO evaluations VALUES(3899,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(3899,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(3899,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(3899,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(3899,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(3899,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(3899,36,'cp 80','f2f4 c5d4 d1f3 f8e8 d2d3 d4e5 f4e5 f6g4 f3g3 g4e5 c1f4 d8d4 e1c1 f7f6 c3e2 d4d7 h1e1 b7b5 c1b1 a7a5 a2a3 d7f7 e2d4 b5b4 f4e5 e8e5 a3a4 e5g5 g3f2 c8g4 d4f3 g4f3 g2f3 b4b3 f3f4 b3c2 f2c2 g5g4 c2f2');
INSERT INTO evaluations VALUES(6002,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6002,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(6002,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(6002,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(6002,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(6002,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(6002,36,'cp 22','f7f5 e1g1 g8f6 e2e3 c5b6 d2d4 e5e4 f3d2 e8g8 f2f3 e4f3 d2f3 c6e7 b2b4 c7c6 a2a4 d6d5 b4b5 d5c4 d1e2 e7d5 e2c4 b6a5 c3d5 c6d5 c4d3 f8e8 c1a3 c8e6 a1c1 a8c8 c1c8 e6c8 f3e5 c8e6 d3c2 a5c7 f1c1 c7b8 g2f1');
INSERT INTO evaluations VALUES(8991,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(8991,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(8991,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(8991,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(8991,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(8991,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(8991,36,'cp 92','e8g8 b2b3 b7b6 f1d3 a7a5 a1c1 c8a6 e1g1 d8e7 f1e1 a5a4 d1c2 h7h6 h2h3 c6b4 c3b4 e7b4 c4d5 a6d3 c2d3 f6d5 e3e4 d5f4 d3e3 f4g6 c1c7 f8c8 c7c8 a8c8 b3a4 b4a4 a2a3 c8a8 e1b1 a4a3 b1b6 a3a5 b6c6 a5a1 g1h2 a8f8 d4d5 e6d5 e4d5');
INSERT INTO evaluations VALUES(3681,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(3681,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(3681,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(3681,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(3681,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(3681,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(3681,36,'cp 98','b2b3 b7b6 f1d3 c6e7 a1c1 d5c4 d3c4 f6e4 c3b2 c8b7 e1g1 e7g6 b2a3 c7c5 d4c5 e4c5 d1d8 f8d8 f1d1 d8d1 c1d1 h7h6 g1f1 a8c8 c4e2 e6e5 f3e1 g6f8 e1d3 c5d3 e2d3 f8e6 f2f3 e5e4 f3e4 e6c5 a3c5 c8c5 d3c4 b7e4 d1d7 b6b5 c4f7 g8f8 d7a7');
INSERT INTO evaluations VALUES(10362,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10362,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10362,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(10362,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(10362,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(10362,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(10362,36,'cp 19','a2a3 b4c3 d2c3 d7d5 d3b5 e8g8 e1g1 d5e4 b5c6 d8d1 f1d1 e4f3 c6f3 c8g4 f3g4 f6g4 h2h3 g4f6 a3a4 a7a5 c1e3 f8d8 c3c4 f6e4 b2b3 f7f5 f2f3 e4f6 e3d2 g8f7 d2c3 d8e8 g2g4 f7g6 g1g2 h7h5 d1e1 h5g4 h3g4 e5e4 g4f5 g6f5');
INSERT INTO evaluations VALUES(10713,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(10713,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(10713,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(10713,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(10713,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(10713,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(10713,36,'cp 75','e1g1 a7a6 b5a4 e8g8 f1e1 f8e8 d2d4 h7h6 b1d2 d6f8 a4c2 d7d6 h2h3 c8d7 d2f1 a6a5 a2a4 e5d4 f3d4 d6d5 d4c6 b7c6 e4e5 f6e4 f2f3 e4c5 f1g3 d8h4 g1h2 h6h5 f3f4 d7g4 d1d2 a8b8 b2b4 a5b4 c3b4 c5e6 d2d3 f8b4 d3h7 g8f8');
INSERT INTO evaluations VALUES(10712,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(10712,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(10712,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(10712,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(10712,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(10712,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(10712,36,'cp 56','a7a6 b5a4 e8g8 f1e1 h7h6 d2d4 f8e8 b1d2 d6f8 a4c6 b7c6 f3e5 c6c5 e5d3 c5d4 c3d4 a6a5 b2b3 c8b7 f2f3 d7d5 e4e5 f6d7 c1b2 c7c5 a1c1 a8c8 f3f4 b7a6 d1f3 c5d4 c1c8 d8c8 e1c1 c8b7 b2d4 e8c8 f3e3 a6d3 c1c8 b7c8 e3d3 c8c1 g1f2 d7c5 d4c5');
INSERT INTO evaluations VALUES(10728,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(10728,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(10728,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(10728,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(10728,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(10728,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(10728,36,'cp 21','e8g8 a2a3 d6c5 c1g5 h7h6 g5f6 d8f6 b2b4 c5d6 c3d5 f6d8 e1g1 c6d4 c2c3 d4f3 d1f3 c7c6 d5e3 a7a5 b4b5 d6c5 f1d1 c5e3 f3e3 d7d5 e4d5 c6d5 c4a2 f8e8 d1e1 d8d6 c3c4 d6a3 c4c5 a3b2 e1b1 b2d4 e3d4 e5d4 c5c6 b7c6 b5c6 c8e6');
INSERT INTO evaluations VALUES(4520,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(4520,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(4520,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(4520,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(4520,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(4520,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(4520,36,'cp 29','h2h3 d6b4 e1g1 b4c3 b2c3 h7h6 a2a4 d7d6 d1e2 a7a6 a4a5 c8e6 c4e6 f7e6 c1e3 d8e8 f1b1 a8b8 g1h2 g8h7 c3c4 e8h5 f3g1 h5f7 c2c3 f6h5 b1b2 h5f4 e2d2 f7g6 e3f4 e5f4 g1f3 e6e5');
INSERT INTO evaluations VALUES(11069,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(11069,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(11069,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(11069,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(11069,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(11069,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(11069,36,'cp 19','d2d4 c5d4 f3d4 d8c7 d4c6 b7c6 d1e2 f8e7 e4e5 f6d5 c4d3 e8g8 c3d5 c6d5 c1f4 c8b7 f1d1 a6a5 a1c1 f7f5 e5f6 c7f4 f6e7 f8f7 c2c3 f4f6 b2b4 f6e7 e2d2 a8f8 d1f1 a5a4 a2a3 f7f4 f2f3 h7h6 g1h1 e6e5 c3c4 e5e4 f3e4 f4f1 c1f1 d5e4 f1f8 g8f8 d3e2 e4e3');
INSERT INTO evaluations VALUES(12190,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(12190,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(12190,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12190,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(12190,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(12190,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(12190,36,'cp 37','e2e3 h7h6 g5h4 f8e7 d1c2 e8g8 f1d3 f8e8 e1g1 f6e4 d3e4 e7h4 e4h7 g8h8 f3h4 d8h4 h7d3 h4e7 a2a3 d7b6 a1e1 c8d7 c3e2 h8g8 e2f4 e7f6 f2f3 a8c8 b2b3 f6e7 a3a4 c6c5 d4c5 c8c5 c2d2 e8c8 a4a5');
INSERT INTO evaluations VALUES(12189,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(12189,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(12189,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(12189,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(12189,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(12189,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(12189,36,'cp 31','h7h6 g5h4 f8e7 d1c2 e8g8 f1d3 f8e8 e1g1 f6e4 d3e4 e7h4 e4h7 g8h8 f3h4 d8h4 h7d3 h4e7 f1e1 d7f6 a2a3 c8d7 b2b4 a7a5 c2b2 b7b5 h2h3 a5b4 a3b4 e7d6 a1a5 a8a5 b4a5 e8a8 e1a1 h8g8 a5a6 a8a7 b2c2 g7g6');
INSERT INTO evaluations VALUES(12284,30,'cp 421',NULL);
INSERT INTO evaluations VALUES(12284,31,'cp 433',NULL);
INSERT INTO evaluations VALUES(12284,32,'cp 459',NULL);
INSERT INTO evaluations VALUES(12284,33,'cp 428',NULL);
INSERT INTO evaluations VALUES(12284,34,'cp 404',NULL);
INSERT INTO evaluations VALUES(12284,35,'cp 416',NULL);
INSERT INTO evaluations VALUES(12284,36,'cp 460','d1b3 d7d5 e4d5 c6a5 b3a4 c8d7 c4b5 f8d6 b5d7 f6d7 c3b5 d6b8 c1d2 a7a6 b5d4 b7b5 a4a5 d8a5 d2a5 e5d4 f1e1 e8f8 a1c1 f8g8 c1c6 g8h7 d5d6 a8a7 f3d4 d7f6 g1f1 a7d7 d4f5 h7g6 f5e7 g6h7 e7c8 h6h5 e1d1 f6g4 c8b6 d7d8 d6d7 g4h2 f1g1 h2g4');
INSERT INTO evaluations VALUES(13587,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13587,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(13587,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13587,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13587,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13587,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13587,36,'cp 0','a2a3 f8e7 d4c5 a7a5 b2b3 e7c5 c1b2 e8g8 b3b4 c5d6 b4b5 c6e5 c3c4 e5f3 d1f3 d8e7 c4d5 e6d5 b2f6 g7f6 f1d3 d6a3 f3h5 f6f5 a1a3 e7a3 h5g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(13626,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(13626,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(13626,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(13626,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(13626,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(13626,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(13626,36,'cp 6','c4b5 c8d7 e4d5 e6d5 e1g1 f8e7 c1e3 e8g8 b5a4 e7b4 a4b3 b4c3 b2c3 f8e8 f1e1 h7h6 h2h3 a8c8 d4c6 b7c6 d1d3 d7e6 c3c4 c6c5 c4d5 f6d5 a1d1 d5e3 d3e3 d8c7 b3e6 e8e6 e3c3 e6e1 c3e1 c8d8 d1d8 c7d8 c2c4 d8c7 g2g3');
INSERT INTO evaluations VALUES(16831,30,'cp -182',NULL);
INSERT INTO evaluations VALUES(16831,31,'cp -173',NULL);
INSERT INTO evaluations VALUES(16831,32,'cp -177',NULL);
INSERT INTO evaluations VALUES(16831,33,'cp -182',NULL);
INSERT INTO evaluations VALUES(16831,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(16831,35,'cp -171',NULL);
INSERT INTO evaluations VALUES(16831,36,'cp -165','e4f6 e1g1 f8e7 f1e1 d7d6 b1d2 e8g8 b2b4 f8e8 c4b3 h7h6 a2a4 e7f8 a4a5 a7a6 d2e4 c8e6 b3a4 f6e4 d3e4 d8f6 c1e3 e8d8 f3d2 f6g6 a4c2 f7f5 f2f4 d6d5 e4f5 e6f5 c2f5 g6f5 f4e5 c6e5 e3d4 f8d6 d4e5 d6e5 d2f3');
INSERT INTO evaluations VALUES(16420,30,'cp -152',NULL);
INSERT INTO evaluations VALUES(16420,31,'cp -163',NULL);
INSERT INTO evaluations VALUES(16420,32,'cp -164',NULL);
INSERT INTO evaluations VALUES(16420,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(16420,34,'cp -153',NULL);
INSERT INTO evaluations VALUES(16420,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(16420,36,'cp -165','e1g1 f8e7 f1e1 d7d6 b2b4 h7h6 b1d2 e8g8 c4b3 f8e8 a2a4 e7f8 d2e4 c8e6 b3e6 e8e6 b4b5 c6e7 e4f6 e6f6 d3d4 e5d4 d1d4 c7c5 b5c6 b7c6 a4a5 c6c5 d4a4 e7g6 c3c4 a8b8 a4c6 g6e7 c6a4 d6d5 a5a6 d5c4 a4c4');
INSERT INTO evaluations VALUES(17340,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(17340,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(17340,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(17340,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(17340,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17340,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(17340,36,'cp 43','b5c6 b7c6 e1g1 g8e7 h2h4 d7d5 f1e1 d5e4 e1e4 e7f5 b1a3 f7f6 d1e2 f5d6 e4e3 d8d7 b2b3 f8e7 c1b2 d6f5 e3e4 f5d6 e4a4 a6a5 d2d4 c5d4 c3d4 c8a6 a3c4 e8g8 a1e1 d6c4 b3c4 c6c5 e2e6 d7e6 e1e6');
INSERT INTO evaluations VALUES(17327,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(17327,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(17327,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(17327,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(17327,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(17327,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(17327,36,'cp 26','b7c6 e1g1 d7d5 d2d3 g8e7 c1e3 e7g6 b1d2 f8e7 h2h4 h7h5 e4e5 a6a5 g2g3 d5d4 c3d4 c5d4 e3d4 c6c5 d4e3 d8d3 a1c1 e8g8 e3c5 e7c5 c1c5 c8b7 c5c3 d3f5 d1c2 f5c2 c3c2 a5a4 f1c1 f8b8 b2b3 a4b3 a2b3 b7f3 d2f3 b8b3 c2c8 a8c8 c1c8 g6f8');
INSERT INTO evaluations VALUES(18146,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(18146,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(18146,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(18146,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(18146,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(18146,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(18146,36,'cp 32','e1g1 g8e7 h2h4 d7d5 f1e1 d5e4 e1e4 d8d3 e4e3 d3d8 b1a3 e7f5 e3e1 f8e7 a3c4 e8g8 b2b3 f7f6 d2d4 c5d4 f3d4 f5d4 c3d4 c6c5 d4d5 d8d5 d1d5 e6d5 c4b6 a8a7 b6d5 g8f7 e1e7 a7e7 d5e7 f7e7 c1e3 e7d6 h4h5 c8e6 a1c1');
INSERT INTO evaluations VALUES(18145,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(18145,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18145,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18145,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(18145,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(18145,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(18145,36,'cp 20','g8e7 d2d4 e7g6 c1g5 f7f6 g5e3 c5d4 c3d4 d7d5 b1d2 f8e7 h2h4 d5e4 d2e4 d8d5 d1c2 e8g8 a1c1 f8d8 e4g3 g6f8 f3d2 c8b7 c2b3 d5b3 d2b3 a6a5 g3e4 f8d7 f1d1 a5a4 b3c5 d7c5 e4c5 b7c8 c1c4 c8a6 c5a6 a8a6 g2g3 d8d5');
INSERT INTO evaluations VALUES(18188,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(18188,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(18188,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18188,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(18188,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(18188,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(18188,36,'cp -28','g8f6 d1e2 d8b6 d4f3 d7d5 e4e5 f6d7 e5e6 f7e6 e2e6 d7f6 e6e5 e7e6 e1g1 f8e7 c1f4 e8g8 b1d2 c6c5 f1e1 f6h5 e5h5 f8f4 h5e8 e7f8 c2c4 f4f6 e1e2 c8b7 e8a4 f8d6 a1e1 a8b8 f3g5 h7h6 g5e6 d5d4 a4b3 b6c6 f2f3 b8e8 b3c2');
INSERT INTO evaluations VALUES(18354,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(18354,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(18354,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(18354,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(18354,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(18354,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(18354,36,'cp 125','e7e5 f1c4 f8c5 e1g1 g8f6 c3d5 e8g8 c2c3 c6a5 c4d3 c5d6 f1e1 f6d5 e4d5 c7c6 d5c6 a5c6 d3e4 d6c7 c3c4 c8b7 b2b4 a8b8 d2d4 c6b4 e4b7 b8b7 d4e5 b4c6 d1d3 h7h6 c1f4 f8e8 a1d1 d8e7 d3d7 e7d7 d1d7 e8e7 d7e7 c6e7');
INSERT INTO evaluations VALUES(18931,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(18931,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(18931,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(18931,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(18931,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(18931,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(18931,36,'cp -28','c3g3 g8f6 f1d3 h7h5 h2h4 d7d5 e4d5 f8d6 c1f4 d6f4 g3f4 f6d5 f4g5 d8f6 g5f6 d5f6 b1a3 e8e7 g1f3 h8d8 e1g1 c8d7 c2c3 c6a5 f1e1 d7a4 d3e2 a8c8 a3b5 a4b5 e2b5 c8c5 b5f1 a5c4 e1e2 g7g6 a1e1 c4d6 f3d4 d6f5 d4f3 b7b5');
INSERT INTO evaluations VALUES(21918,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(21918,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21918,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21918,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21918,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21918,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21918,36,'cp 4','g7g6 b1c3 f8g7 g2g3 a7a5 f1g2 g8e7 e1g1 e8g8 a1b1 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 d8d7 c1e3 d4e6 d1d2 c7c6 c3a4 d7d8 h3h4 f7f5 g2h3 g8h8 a4b6 a8a6 b6a4');
INSERT INTO evaluations VALUES(22964,30,'cp -181',NULL);
INSERT INTO evaluations VALUES(22964,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(22964,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(22964,33,'cp -168',NULL);
INSERT INTO evaluations VALUES(22964,34,'cp -177',NULL);
INSERT INTO evaluations VALUES(22964,35,'cp -186',NULL);
INSERT INTO evaluations VALUES(22964,36,'cp -154','g7g5 h2h4 g5g4 f3g1 g8f6 c1f4 d7d5 b1c3 d5d4 c3b5 f8b4 f4d2 a7a6 d2b4 c6b4 b5a3 c8e6 g1e2 d8d6 d1d2 b4c6 e2f4 e8c8 f1e2 h8g8 e1g1 c6e5 f4e6 d6e6 f1f5 e5f3 e2f3 g4f3 f5f3 f6g4 d2f4 g4e3 f3e3 d4e3 a3c4 g8g4 f4e5 d8g8 e5e6 f7e6 c4e3 g4h4 a1f1 g8g6 f1f7');
INSERT INTO evaluations VALUES(29218,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(29218,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(29218,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(29218,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(29218,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(29218,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(29218,36,'cp -12','f8e7 b2b4 e8g8 f1e2 b7b6 c1d2 a7a5 a2a3 b8d7 e1g1 a5b4 a3b4 d8c7 d1b3 c7b7 c5b6 b7b6 b4b5 f6e4 c3e4 f5e4 d2b4 a8a1 f1a1 e7b4 b3b4 e4f3 g2f3 c6b5 b4e7 d7f6 a1b1 b6d8 e7d8 f8d8 h2h4 b5b4 b1b4 g7g6 g1g2 g8g7 b4b7 d8c8 b7a7');
INSERT INTO evaluations VALUES(30607,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(30607,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(30607,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(30607,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(30607,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(30607,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(30607,36,'cp -13','a7a6 b5e2 c7c5 d2d4 c5d4 f3d4 d8c7 f2f4 b8c6 d4c6 b7c6 d1d3 g8f6 a2a3 f8e7 c1e3 d7d6 e3d4 e8g8 e1g1 c7b7 e2f3 f6d7 a1d1 a8d8 g1h1 f8e8 f3h5 g7g6 h5f3 e6e5 d4e3 d7c5 d3e2 e7f6 c3d5 c6d5 d1d5 e5f4');
INSERT INTO evaluations VALUES(32864,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(32864,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(32864,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(32864,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(32864,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(32864,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(32864,36,'cp 105','g4d7 d2d4 e5a5 c1d2 g8f6 e2f4 a5b6 c3d5 f6d5 f4d5 b6e6 d5e3 e6d6 c2c3 e7e5 d4d5 c7c6 d1b3 d6c7 d5c6 d7c6 e1c1 b8d7 e3d5 c7d8 d2e3 f8c5 e3c5 d7c5 b3b4 d8d6 b4h4 d6d8 h4f2 c5d7 d5b4 d8f6 f1d3 e8g8 d3c2 a7a5 b4c6 f6c6');
INSERT INTO evaluations VALUES(34712,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(34712,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(34712,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(34712,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34712,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34712,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34712,36,'cp 0','g8f6 h5e2 g7g6 g1f3 f8g7 b1c3 h7h6 a2a4 e8g8 e1g1 b8c6 h2h3 g8h7 c1e3 c8e6 a4a5 e6d7 c3b5 e7d8 b5c3 d8e7');
INSERT INTO evaluations VALUES(33778,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33778,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33778,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33778,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33778,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33778,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33778,36,'cp 3','h5e2 g7g6 b1c3 f8g7 g1f3 e8g8 e1g1 c8e6 a2a4 f8e8 c1e3 c7c6 c4e6 e7e6 f3g5 e6e7 f2f4 e5f4 e3f4 d6d5 a1e1 h7h6 g5f3 d5e4 c3e4 f6e4 e2e4 e7e4 e1e4 b8d7 f1e1 e8e4 e1e4 d7f6 e4e2 f6d5 f4d2 f7f5 c2c4');
INSERT INTO evaluations VALUES(36873,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(36873,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(36873,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(36873,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(36873,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(36873,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(36873,36,'cp 74','d4d5 a7a5 c1g5 h7h6 g5d2 b8a6 f2f3 f6h5 d2e3 a6b4 a2a3 b4d3 d1d3 f7f5 e1c1 d8e7 e2g3 h5g3 h2g3 h6h5 c1b1 b7b6 d1g1 a5a4 e3d2 c8d7 b1a1 e7f7 d2e3 f5e4 c3e4 d7f5 g3g4 f5e4 d3e4');
INSERT INTO evaluations VALUES(39068,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(39068,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(39068,32,'cp -190',NULL);
INSERT INTO evaluations VALUES(39068,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(39068,34,'cp -182',NULL);
INSERT INTO evaluations VALUES(39068,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(39068,36,'cp -168','h7h6 d2d4 b8c6 c2c3 d7d6 d1b3 d8d7 c4d3 g8f6 c1d2 e8g8 d2e1 c6e7 h2h3 c7c5 b1d2 d7c7 e1f2 b7b6 a1e1 c8e6 d3c4 e6c4 d2c4 c5d4 f2d4 e7c6 d4f6 g7f6 e1d1 a8d8 d1d5 f8e8 f1e1 f6g7 e1d1 e8e4 c4d6');
INSERT INTO evaluations VALUES(38983,30,'cp -192',NULL);
INSERT INTO evaluations VALUES(38983,31,'cp -186',NULL);
INSERT INTO evaluations VALUES(38983,32,'cp -166',NULL);
INSERT INTO evaluations VALUES(38983,33,'cp -182',NULL);
INSERT INTO evaluations VALUES(38983,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(38983,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(38983,36,'cp -184','d2d4 b8c6 c2c3 d7d6 a2a4 g8f6 b1d2 e8g8 f1e1 f8e8 h2h3 d6d5 e4d5 e8e1 d1e1 f6d5 d2e4 c8f5 c1d2 c6a5 c4a2 d5e3 e4g5 h6g5 d2e3 f4e3 e1e3 a5c6 f3g5 f5g6 e3g3 d8d6 g3g4 c6d4 c3d4 d6d4 g4d4 g7d4 g1h1 d4b2 a1f1 a8f8 f1e1 f8d8 e1f1');
INSERT INTO evaluations VALUES(38984,30,'cp -200',NULL);
INSERT INTO evaluations VALUES(38984,31,'cp -168',NULL);
INSERT INTO evaluations VALUES(38984,32,'cp -185',NULL);
INSERT INTO evaluations VALUES(38984,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(38984,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(38984,35,'cp -189',NULL);
INSERT INTO evaluations VALUES(38984,36,'cp -186','d7d6 c2c3 b8c6 a2a4 g8f6 b1d2 e8g8 f1e1 c6e7 a4a5 e7g6 h2h3 f8e8 c4d3 c8e6 b2b4 c7c6 c1b2 g5g4 h3g4 f6g4 d2f1 h6h5 d1d2 b7b5 a5b6 a7b6 a1a8 d8a8 d2e2 c6c5 d3b5 e8b8 e2d2 a8b7 c3c4 c5d4 f3d4 e6d7 b5d7 b7d7');
INSERT INTO evaluations VALUES(46720,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(46720,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46720,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46720,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(46720,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46720,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46720,36,'cp 0','c7c5 d2d4 c5d4 f3d4 f8g7 c1e3 e8g8 d1d2 d6d5 e4d5 f6d5 e1c1 d5e3 d2e3 d8a5 e3e7 b8c6 d4b3 a5f5 e7h4 g7f6 h4g3 f6e5 g3h4');
INSERT INTO evaluations VALUES(50673,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50673,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(50673,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(50673,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50673,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(50673,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50673,36,'cp 22','b1c3 b8c6 d2d4 c5d4 f3d4 g8f6 c1e3 g7g6 d4c6 b7c6 e4e5 f6g4 e3g1 f8g7 d1d2 d6e5 d2d8 e8d8 h2h3 g4f6 f4e5 f6d7 e5e6 g7c3 b2c3 f7e6 e1c1 d8c7 g1h2 e6e5 f1c4 h8f8 h2g3 f8f6 h1e1 c8b7 e1e5 d7e5 g3e5 c7b6 e5f6 e7f6 d1d7 b6c5 d7h7 c5c4 h7b7');
INSERT INTO evaluations VALUES(50671,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(50671,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(50671,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(50671,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(50671,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(50671,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(50671,36,'cp 28','b8c6 d2d4 c5d4 f3d4 g8f6 f1c4 e7e6 c4b3 f8e7 c1e3 e8g8 d1f3 d8c7 f4f5 c6d4 e3d4 e6e5 d4f2 b7b5 a2a3 c8b7 e1c1 b5b4 a3b4 a6a5 b4b5 a5a4 b5b6 c7b8 b3d5 a4a3 b2b3 f8c8 c1b1 h7h6 d5b7 b8b7 b1a1 e7f8 g2g4 f8e7 h1e1');
INSERT INTO evaluations VALUES(51349,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(51349,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(51349,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(51349,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(51349,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(51349,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(51349,36,'cp 109','c8b7 f1e2 g8f6 d1c2 g7g6 c1e3 f8g7 g1f3 e8g8 a1d1 d8e7 e1g1 c7c5 d4d5 e6d5 c4d5 f8e8 e2b5 b8d7 a2a4 a7a6 b5c6 b7c6 d5c6 d7e5 f3e5 e7e5 g2g3 a8c8 e3f4 e5h5 c6c7 c8c7 d1d6 c7d7 d6b6 f6g4 f2f3 g7d4 g1h1 g4e5 f4e5 h5e5 b6a6');
INSERT INTO evaluations VALUES(54036,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54036,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54036,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54036,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54036,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54036,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54036,36,'cp 0','c8d7 b5e2 d8c7 e1g1 g8f6 b1d2 c5d4 e3d4 f8d6 c2c4 d5c4 d2c4 e8g8 c1g5 f6d5 a1c1 b8c6 e2d3 h7h6 g5d2 d5f6 f1e1 f8d8 c4e5 d7e8 a2a3 a8c8 b2b4 c7b6 e5c4 b6c7');
INSERT INTO evaluations VALUES(30881,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30881,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30881,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30881,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30881,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30881,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30881,36,'cp 0','b5e2 g8f6 e1g1 d8c7 b2b3 c5d4 e3d4 b7b5 b1d2 f8e7 f3e5 e8g8 e5d7 c7d7 c1b2 a7a5 c2c4 b5c4 b3c4 b8c6 c4c5 a8b8 d1c2 c6b4 c2b1 b4c6');
INSERT INTO evaluations VALUES(54328,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(54328,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(54328,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(54328,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(54328,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(54328,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(54328,36,'cp 60','d1a4 c8d7 a4d4 e6d5 d4d5 g8f6 d5d1 f8b4 c1d2 b4c3 d2c3 e8g8 g2g3 f6e4 f1g2 e4c3 b2c3 d8c7 a1c1 d7c6 g1f3 f8d8 d1c2 b8a6 e1g1 a6c5 f3d4 c6g2 g1g2 g7g6 h2h4 h7h5 f1d1 b7b6 d4f3 c7c6 g2g1 d8d1 c1d1 a8c8 d1d4 c6b7');
INSERT INTO evaluations VALUES(54352,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(54352,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(54352,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(54352,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(54352,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(54352,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(54352,36,'cp 96','d1e2 f8e7 g2g3 b8c6 f1g2 g8f6 e1g1 e8g8 h2h4 h7h6 f1d1 e7b4 e4e5 f6h7 e2e3 c6e7 a2a4 c8d7 g2h3 d8e8 c3b5 e8d8 b2b3 a7a6 b5d6 d7c6 c1a3 b4a3 a1a3 e7c8 d6c8 a8c8 b3b4 d8e7 d1b1 c8a8 a3a1 b7b5 a4a5 c6e8 c2c3 f7f5 e5f6 f8f6');
INSERT INTO evaluations VALUES(55278,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(55278,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(55278,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(55278,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(55278,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(55278,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(55278,36,'cp 86','e2g3 d3c2 d1c2 e7e6 g1f3 c7c5 b2b4 c5b4 d2d4 g8f6 c1b2 b4b3 c2b3 d8b6 f1c4 b6b3 a2b3 f8b4 e1e2 e8g8 f3e5 f8d8 h1c1 a7a5 c4b5 f6e8 b5e8 d8e8 b2c3 b4c3 c1c3 f7f6 a1c1 b8a6 e5d3 b7b6 e2e3 c8d7 f2f3 a8c8 g3e2 c8c3 e2c3');
INSERT INTO evaluations VALUES(55795,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(55795,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(55795,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(55795,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(55795,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(55795,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(55795,36,'cp -107','d5d4 c3e2 e6e5 a2a3 g8e7 d2d3 e7g6 g2g3 a7a5 f2f4 a5a4 h2h4 f8d6 f1h3 h7h5 h3c8 d8c8 f4f5 g6f8 g1h3 f8d7 c1d2 f7f6 h3f2 d7c5 e2c1 b7b6 g3g4 h5g4 d1g4 c8d7 c1e2 b8a6 e1c1 e8c8');
INSERT INTO evaluations VALUES(2956,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(2956,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(2956,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(2956,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(2956,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(2956,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(2956,36,'cp 21','c4d5 e6d5 f1d3 f6e4 g5f4 d7f6 e1g1 c8f5 f3e5 f8e8 a2a4 a7a5 d1c2 e4c3 b2c3 f5d3 c2d3 f6e4 f2f3 e4d6 e3e4 d8c8 h2h3 c8e6 e4d5 c6d5 f1e1 e7f8 g2g4 a8c8 g1g2 d6c4 e5c4 d5c4 e1e6 c4d3 e6e8 c8e8');
INSERT INTO evaluations VALUES(3023,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(3023,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(3023,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(3023,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(3023,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(3023,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(3023,36,'cp -30','c3d5 f6d5 e4d5 e5e4 f3d4 b4d2 d1d2 c6d4 d2d4 d7d6 e1c1 f8e8 f1e2 a7a6 c1b1 b7b5 d4e3 c8d7 g2g4 e8e5 h2h4 d8e7 g4g5 a8f8 a2a3 f7f6 h1g1 f6g5 h4g5 g7g6 e2g4 d7g4 g1g4 f8f3 e3d4 a6a5 g4g1');
INSERT INTO evaluations VALUES(3378,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(3378,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(3378,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(3378,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(3378,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(3378,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(3378,36,'cp 13','a2a3 c8e6 c3d5 d8d7 e1g1 c6d8 d5e7 d7e7 c4e6 d8e6 f1e1 a7a5 d3d4 e5d4 e3d4 e6d4 d1d4 a8d8 d4c4 c7c6 e4e5 d6e5 e1e5 e7c7 a1e1 f8e8 g2g3 h7h6 e5e8 f6e8 c4c3 e8f6 g1g2 d8e8 h2h3 e8e1 c3e1 c7d8 c2c4 a5a4');
INSERT INTO evaluations VALUES(4043,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(4043,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(4043,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(4043,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(4043,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(4043,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(4043,36,'cp 72','h2h3 e7e5 d4d5 d7c5 f3d2 a7a5 f1e2 f6e8 d1c2 f7f5 e1c1 b7b6 d1g1 c5e6 d5e6 f5f4 e3b6 c8e6 c4c5 c7b6 e2c4 e6c4 d2c4 b6c5 h3h4 d8e7 c3d5 e7f7 c1b1 a5a4 h4h5 g6h5 g2g4 f4g3 f2g3 g8h8');
INSERT INTO evaluations VALUES(5935,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(5935,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(5935,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(5935,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(5935,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(5935,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(5935,36,'cp 22','e5d4 f3d4 g8e7 c1e3 e8g8 d1d2 c5d4 e3d4 c6d4 d2d4 c7c6 b5e2 d6d5 e1c1 c8e6 e2f3 d5e4 d4d8 a8d8 c3e4 b7b6 e4g5 e6d5 h1e1 f8e8 f3d5 e7d5 e1e8 d8e8 g5f3 g7g6 f3d4 d5e7 a2a4 h7h5 d4f3 e7d5');
INSERT INTO evaluations VALUES(5956,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(5956,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(5956,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(5956,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(5956,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(5956,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(5956,36,'cp 34','f3d4 g8e7 c1e3 e8g8 d1d2 c6d4 e3d4 a7a6 b5e2 c5d4 d2d4 b7b5 a1d1 a8b8 a2a3 d8e8 e1g1 f7f5 f1e1 e8f7 e4f5 e7f5 d4f4 c8b7 e2d3 h7h6 f2f3 f7f6 g1f2 b8d8 d3f5 f6f5 f4f5 f8f5 e1e6 f5f7 h2h3 g8f8 d1e1 b7c6');
INSERT INTO evaluations VALUES(5953,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(5953,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(5953,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(5953,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(5953,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(5953,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(5953,36,'cp 23','g8e7 c1e3 e8g8 d1d2 c5d4 e3d4 c6d4 d2d4 c7c6 b5e2 d6d5 e1c1 c8e6 e2f3 d5e4 c3e4 d8d4 d1d4 e6d5 h1d1 e7f5 d4b4 b7b5 e4g5 a7a5 b4f4 f5e7 f4d4 a5a4 a2a3 a8e8 g5e4 e7g6 g2g3 f7f5 e4g5 d5f3 g5f3 f5f4 g3g4 g6e5 f3e5');
INSERT INTO evaluations VALUES(8557,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(8557,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(8557,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(8557,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(8557,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(8557,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(8557,36,'cp 40','b5a4 e8g8 e1g1 a7a6 h2h3 f8e8 f1e1 e7f8 c1g5 h7h6 g5h4 b7b5 a4b3 c6a5 b3c2 c7c5 b1d2 g7g6 a2a4 c8b7 d3d4 e5d4 c3d4 c5d4 a4b5 a6b5 f3d4 g6g5 h4g3 a5c4 a1a8 b7a8 d2c4 b5c4 e4e5 d6e5 e1e5 f8g7 d4f5 d8d1 c2d1 e8e5');
INSERT INTO evaluations VALUES(8559,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(8559,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(8559,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(8559,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8559,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(8559,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(8559,36,'cp 39','c2c3 e8g8 e1g1 a7a6 b5a4 b7b5 a4c2 f8e8 f1e1 d6d5 a2a4 d5e4 d3e4 d8d1 e1d1 b5b4 c1e3 b4c3 b2c3 c6a5 f3e5 e7d6 f2f4 d6e5 f4e5 a5c4 e3f4 c4e5 b1d2 c8b7 a1b1 b7c6 b1b4 a6a5 b4d4 h7h6 f4g3 c6b7 d2b3 e5c6 d4d2 c6e5 b3c5 b7c6 d2d4 a8b8');
INSERT INTO evaluations VALUES(8555,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(8555,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(8555,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(8555,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(8555,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(8555,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(8555,36,'cp 17','e8g8 e1g1 f8e8 b1d2 e7f8 f1e1 a7a6 b5a4 b7b5 a4c2 c8b7 d3d4 g7g6 a2a3 c6b8 d2b3 e5d4 c1g5 b8d7 c3d4 h7h6 b3a5 h6g5 a5b7 d8b8 b7a5 c7c5 f3g5 f8h6 g5f3 c5d4 a5b3 b8c7 f3d4 d7e5 a3a4 a8c8 a4b5 a6b5');
INSERT INTO evaluations VALUES(10159,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(10159,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(10159,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(10159,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(10159,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(10159,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(10159,36,'cp -17','a7a6 h2h3 b7b5 d2d3 d7d6 g1e2 a8b8 e1g1 e8g8 b2b3 c6d4 g1h2 h7h6 f2f4 b5c4 b3c4 e5f4 c1f4 f6d7 d1d2 d7e5 a1c1 d4e2 c3e2 c5b4 d2c2 c7c5 c1b1 c8e6 f4e3 b4a5 e2f4 b8b1 f1b1 g7g5');
INSERT INTO evaluations VALUES(14107,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(14107,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(14107,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(14107,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14107,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(14107,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(14107,36,'cp 72','h2h3 f8g7 g1f3 e8g8 b1d2 f8e8 d1b3 c6a5 b3a3 a5c6 f1b5 c8d7 b5e2 a7a5 e1g1 a5a4 c3c4 e7e6 f1c1 f6e4 c4d5 e6d5 d2e4 e8e4 f4g3 g7f8 a3c3 f8b4 c3c2 e4e8 a2a3 d7f5 c2d1 b4d6 g3d6 d8d6 e2b5 f5d7 c1c3 c6e7 b5d7 d6d7 f3e5');
INSERT INTO evaluations VALUES(14704,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(14704,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(14704,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(14704,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(14704,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(14704,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(14704,36,'cp 92','c8e6 b5c7 e8d7 f2f3 g7g5 c7e6 f7e6 f4c1 e4d6 g1h3 h7h6 h3f2 d8b6 c2c3 d7c7 e2e4 d5e4 f3e4 c6e5 f1e2 b6c6 e1g1 a8d8 d1c2 d6c4 e2c4 e5c4 b2b3 c4e5 c1b2 f8c5 g1h1 c5f2 c2f2 e5d3 f2e2 d3b2 e2b2 c6e4');
INSERT INTO evaluations VALUES(15391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15391,36,'cp 0','h7h6 a2a4 f8d6 f3d2 d8e7 e1g1 e8g8 a4a5 c8e6 d1f3 a7a6 c4d5 f7f5 d5c6 b7c6 f3c6 e6d7 c6d5 d7e6');
INSERT INTO evaluations VALUES(17124,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(17124,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(17124,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(17124,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(17124,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(17124,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(17124,36,'cp 212','b5c6 d7c6 d2d3 e4f6 f3e5 f8e7 d1e2 c6c5 e5c4 a7a5 c1g5 a8a6 b1c3 a5a4 g5f6 a6f6 e2e5 c8e6 c3b5 e8d7 b2b3 c7c6 b5d6 b7b5 a1d1 d8c7 d3d4 b5c4 d4c5 e7d6 d1d6 d7c8 d6e6 c7e5 e6e5 h8d8 e5e7 a4b3 a2b3 c4b3 c2b3 d8d7 e7e8 d7d8 b3b4 f6f7 g1f1 f7d7 e8d8 d7d8 e1e7 d8d7 e7d7 c8d7 f1e2 d7e6 e2d3 e6f6');
INSERT INTO evaluations VALUES(17094,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(17094,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(17094,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(17094,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(17094,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(17094,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(17094,36,'cp 206','d7c6 d2d3 e4f6 f3e5 f8e7 d1e2 c6c5 e5c4 a7a5 c1g5 a8a6 b1c3 c8e6 e2f3 d8c8 e1e2 e8g8 a1e1 h7h6 g5d2 a5a4 c3a4 e6d5 f3h3 f6e4 a4c3 d5c4 d3c4 e7f6 c3d5 e4d2 e2d2 a6e6 d2e2 e6e2 e1e2 c8d7 h3d3 g8f7 b2b3 c7c6 d5f4 d7d3 f4d3 f8d8 g1f1 f6c3 e2e3 g7g5');
INSERT INTO evaluations VALUES(17394,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(17394,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17394,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(17394,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(17394,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(17394,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(17394,36,'cp 36','d2d4 c5d4 f3d4 d8c7 c1e3 b7b5 d4c6 c7c6 f1e2 c8b7 d1d3 c6c7 g2g4 a8c8 h1f1 g8e7 e1c1 e7c6 e3b6 c6e5 d3d4 f8c5 b6c5 c7c5 f2f4 c5d4 d1d4 e5c6 d4d2 c6e7 d2d6 c8c6 f1d1 c6d6 d1d6 b7c6 e2d3 e7g6 c3e2 h7h6 e4e5 e8e7');
INSERT INTO evaluations VALUES(19837,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(19837,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(19837,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(19837,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(19837,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(19837,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(19837,36,'cp 116','e4d5 d8d5 c1e3 e7e6 b1c3 f8b4 d1g4 d5e5 e1c1 b4c3 b2c3 g8e7 e3f4 e5f6 d4c6 e7c6 f4d6 f6c3 c1b1 c8d7 f1e2 h7h5 g4g5 c3f6 g5b5 a8d8 b5b7 f6f2 d6c7 e8g8 c7d8 c6d8 b7f3 f2b6 b1a1 d7c6 f3h5 c6a4 h5h4 a4c2 e2d3 c2d3 d1d3 b6c7 h4g3');
INSERT INTO evaluations VALUES(20689,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(20689,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(20689,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(20689,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(20689,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(20689,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20689,36,'cp 11','c5d4 f3d4 g8f6 f1e1 f8e7 c4f1 e8g8 c2c4 d6d5 e4d5 e6d5 b1c3 c6d4 d1d4 d5c4 d4d8 e7d8 f1c4 c8d7 c1f4 a8c8 c4b3 f8e8 e1e8 d7e8 a1e1 e8c6 f4e3 a7a6 e1d1 d8e7 f2f3 g8f8 h2h3 h7h6 c3e2 c8d8 d1d8 e7d8 g1f2 f6d5 e3d4 d8c7');
INSERT INTO evaluations VALUES(20718,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20718,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(20718,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(20718,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(20718,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(20718,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20718,36,'cp 11','f3d4 g8f6 b1c3 f8e7 c1e3 e8g8 c4b3 c8d7 d1e2 a7a6 f2f4 a8c8 a1d1 c6a5 e4e5 a5b3 a2b3 f6d5 c3d5 e6d5 e2f3 d6e5 f4e5 e7c5 c2c3 d8c7 e3f4 c7c6 g1h1 c5d4 d1d4 d7e6 f1d1 c6b6 d4b4 b6c6 h2h3 c6c7');
INSERT INTO evaluations VALUES(21399,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(21399,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(21399,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(21399,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(21399,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(21399,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(21399,36,'cp 15','f8d6 f1b5 d6f4 e3f4 d8d6 g1e2 g8f6 c2c3 e8g8 d2f3 c6e7 h2h4 c7c5 b5d3 b7b6 h4h5 d6c7 f3e5 f6e4 d1c2 f7f6 e5g4 c5d4 f2f3 e6e5 c3d4 c7c2 d3c2 c8g4 f3g4 a8c8 c2b3 g8h8 a1d1 e5f4 e2f4 f6f5 f4d5 e7d5 b3d5 e4g3 h1h3 c8e8 e1d2 g3e4 d5e4 e8e4 g4g5 e4d4 d2c1 d4d1 c1d1 f8e8 h3b3 g7g6');
INSERT INTO evaluations VALUES(21646,30,'cp -147',NULL);
INSERT INTO evaluations VALUES(21646,31,'cp -168',NULL);
INSERT INTO evaluations VALUES(21646,32,'cp -199',NULL);
INSERT INTO evaluations VALUES(21646,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(21646,34,'cp -191',NULL);
INSERT INTO evaluations VALUES(21646,35,'cp -191',NULL);
INSERT INTO evaluations VALUES(21646,36,'cp -196','f1b5 g7g5 e1g1 h7h6 d2d4 g8e7 c2c3 f8g7 b5d3 e7f5 b2b4 e8g8 b4b5 c6e7 b1a3 a7a6 b5a6 b7b6 d1e2 c7c5 a3b5 c5c4 d3f5 c8f5 a6a7 f5d3 e2f2 d3f1 f2f1 f8e8 b5d6 a8a7 d6e8 d8e8 g2g3 f4g3 h2g3 e8c8 f3h2 c8f5 f1f5 e7f5');
INSERT INTO evaluations VALUES(21661,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(21661,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(21661,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(21661,33,'cp -120',NULL);
INSERT INTO evaluations VALUES(21661,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(21661,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(21661,36,'cp -115','f1b5 f8d6 b5c6 b7c6 c1b2 g8f6 h5g5 d8e7 g1e2 e8g8 g5h4 a7a5 e1g1 c8a6 d2d3 a5a4 e2g3 e7e6 f1e1 f6d7 e3e4 d6e7 h4h5 g7g6 h5h6 d5d4 f2f4 e5f4 h6f4 c6c5 b1d2 e7d6 f4h4 f7f6 d2f3 d7e5 f3e5 f6e5 h4h6 d6e7 h6d2 f8f4 e1f1 f4f1 a1f1 a4b3');
INSERT INTO evaluations VALUES(21527,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(21527,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(21527,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(21527,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(21527,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(21527,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(21527,36,'cp -114','f8d6 c1b2 g8f6 b5c6 b7c6 h5g5 d8e7 g1e2 h7h6 g5h4 e8g8 b1c3 c8d7 f2f3 f8e8 e1g1 c6c5 e3e4 d5d4 c3d1 f6d5 h4e7 d5e7 d2d3 e7c6 b2a3 a7a5 d1b2 a8a6 b2c4 a5a4 f3f4 d7g4 f1e1 g4e2 f4e5 c6e5 e1e2 e5c4 b3c4');
INSERT INTO evaluations VALUES(5895,30,'cp -113',NULL);
INSERT INTO evaluations VALUES(5895,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(5895,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(5895,33,'cp -126',NULL);
INSERT INTO evaluations VALUES(5895,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(5895,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(5895,36,'cp -128','b5c6 b7c6 c1b2 g8f6 h5g5 d8e7 g1e2 h7h6 g5h4 e8g8 b1c3 c8d7 f2f3 a7a5 e1g1 c6c5 a2a4 c5c4 f3f4 c4b3 c2b3 f8b8 f4e5 e7e5 c3d1 e5g5 b2f6 g7f6 h4f2 b8b3 e2d4 b3b4 d1c3 g5e5 f2g3 e5g3 h2g3');
INSERT INTO evaluations VALUES(21971,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(21971,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(21971,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(21971,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21971,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(21971,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(21971,36,'cp 15','d3b5 c8d7 b1c3 c6d4 b5d7 d8d7 d2d3 g7g6 f3d4 e5d4 c3d5 f8g7 c1d2 c7c6 d5f4 g8e7 e1g1 e8g8 h2h3 d6d5 d1g4 d7g4 h3g4 d5e4 d3e4 c6c5 c2c4 e7c6 f4d3 f8e8 a1e1 c6e5 d3e5 g7e5 f2f4 e5c7 g1f2 f7f6 f2f3 b7b6 a2a4 e8e7 g2g3 e7e8 f1h1 e8e7');
INSERT INTO evaluations VALUES(22763,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(22763,31,'cp -97',NULL);
INSERT INTO evaluations VALUES(22763,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(22763,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(22763,34,'cp -95',NULL);
INSERT INTO evaluations VALUES(22763,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(22763,36,'cp -107','c6d4 f2f4 d7d5 f4e5 d5e4 c3e4 c8g4 e4d6 d8d6 e5d6 g4d1 e1d1 f8d6 c2c3 d4e6 g1f3 g8f6 d1c2 e8g8 f1c4 f8e8 h1d1 a7a5 c1e3 e6f4 e3d4 f4d5 d1e1 a5a4 a1d1 e8e1 d1e1 g8f8 g2g4 c7c6 h2h3 g7g6 d4f2 a8e8');
INSERT INTO evaluations VALUES(25596,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(25596,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(25596,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(25596,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(25596,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(25596,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(25596,36,'cp -34','g8f6 e1g1 f8d6 a3a4 d8b6 f1e1 c6e5 c1f4 e5f3 d1f3 d6f4 f3f4 d7d6 c3b1 d6d5 e4d5 b7d5 a2d5 f6d5 f4e5 e8g8 b1d2 f8d8 a4b5 a6b5 a1a8 d8a8 d2f3 h7h6 e5e4 a8d8 e4e5 c5c4 d3c4 b5c4 g2g3 b6b4');
INSERT INTO evaluations VALUES(24228,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(24228,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(24228,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(24228,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(24228,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(24228,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(24228,36,'cp -38','e1g1 f8d6 a3a4 d8b6 f1e1 c6d4 h2h3 d4f3 d1f3 e8g8 c1f4 d6f4 f3f4 b6c6 c3d1 b5a4 a2b1 d7d5 c2c3 c6b6 b1c2 b7c6 e4d5 f6d5 f4h4 a4a3 a1a3 h7h6 c2a4 f8b8 h4g3 c6a4 a3a4 d5e7 g3e3 e7f5 e3e5 a6a5');
INSERT INTO evaluations VALUES(26668,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(26668,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(26668,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(26668,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(26668,34,'cp 177',NULL);
INSERT INTO evaluations VALUES(26668,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(26668,36,'cp 165','d4d5 c6b8 f1c4 e7e5 f3e5 f8d6 c4d3 f5d3 e5d3 g8f6 b1c3 e8g8 e1g1 f8e8 f1e1 c7c6 e3e4 b8d7 c1f4 d6f4 d3f4 d8b6 d1d2 h7h6 a1d1 a8d8 d2c2 d7e5 h2h3 b6c5 c2e2 c5b4 a2a3 b4c4 e2e3 c4b3 e3a7 b3b2 a7c5 e5d7 c5b4 b2b4 a3b4 d7e5 f2f3 c6d5');
INSERT INTO evaluations VALUES(26745,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(26745,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(26745,32,'cp 207',NULL);
INSERT INTO evaluations VALUES(26745,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(26745,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(26745,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(26745,36,'cp 218','d1b3 c6a5 c4f7 e8e7 b3a3 e7f7 a3a5 e5d4 c3d4 g8f6 b1c3 c7c6 a5a4 d8d7 c1e3 g4e6 e1g1 f8e7 f3e1 h8e8 f2f3 f7g8 a4c2 e7d8 e1d3 d7f7 h2h3 d8a5 d3f4 e6c4 f1f2 c4a6 c3e2 f7c4 c2d1 a5b6 b2b3');
INSERT INTO evaluations VALUES(26798,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(26798,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(26798,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(26798,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(26798,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(26798,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(26798,36,'cp -42','c6b4 b1a3 c7c6 f1e2 d5c4 e2c4 b4d5 f4e5 f8b4 e1f1 b4a3 b2a3 g8f6 c4d3 d8e7 e5f6 d5f6 d3f5 e6f5 d1b3 a8c8 g2g3 e8g8 f1g2 c6c5 a1c1 b7b6 h1d1 e7e4 d4c5 c8c5 c1c5 b6c5 b3d3 c5c4 d3e4 f6e4 f3d4 g7g6 f2f3 e4f6 d4c2 g8g7 g2f2 f8c8 d1d6 c8c5');
INSERT INTO evaluations VALUES(26945,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(26945,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(26945,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(26945,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(26945,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(26945,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(26945,36,'cp 39','h5g6 b1c3 c6a5 d1e2 g8f6 a2a3 a5c4 d3c4 c7c6 c1d2 f6d7 e1c1 f8e7 d2e3 d8c7 f3d2 h7h5 c1b1 d7f8 f2f3 f8e6 d2f1 h5h4 e3f2 a7a6 f1e3 b7b5 e3f5 g6f5 g4f5 e6f4 e2f1 e7g5 c4c5 d6c5 f2c5 a8d8 f1f2 g5e7');
INSERT INTO evaluations VALUES(27177,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27177,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(27177,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(27177,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(27177,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27177,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(27177,36,'cp 14','b5e2 d5d4 c2c3 g8e7 d2d3 e7f5 b1d2 a7a5 h2h4 f8e7 h4h5 f7f6 g2g4 f5h6 e5f6 g7f6 f3d4 c6d4 c3d4 d7c6 e2f3 c6f3 d1f3 d8d4 d2b3 d4g4 f3b7 g4b4 b7b4 e7b4 c1d2 h6f5 d2b4 a5b4 e1d2 h8g8 h1c1 g8g2 d2e2 a8a7 a2a3 b4a3 a1a3 a7a3 b2a3 g2h2 c1c7 h2h5 e2d2 h5h2');
INSERT INTO evaluations VALUES(28037,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(28037,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(28037,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(28037,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(28037,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(28037,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(28037,36,'cp 117','h2h4 h7h5 g1h3 g8f6 f2f3 d7d6 d1d2 b8d7 h3g5 c7c5 e1c1 d8c7 a2a3 c5d4 e3d4 a7a6 c3d5 b7d5 e4d5 b6b5 g2g4 c7b7 c1b1 a8b8 f1e2 b5b4 d2b4 b7b4 a3b4 b8b4 c2c3 b4b8 e2a6 e8g8 h1e1 h5g4 f3g4 d7e5 a6e2 f6d5 d4e5 g7e5 d1d5');
INSERT INTO evaluations VALUES(28059,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(28059,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(28059,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(28059,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(28059,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(28059,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(28059,36,'cp -17','d8d7 d2d4 c5d4 b1d2 g8f6 f3d4 e8g8 e1g1 b8c6 d2f3 f8e8 c2c3 e7c5 e2d3 c6d4 f3d4 e8e4 f2f3 c5d4 c3d4 e4e6 c1d2 a8e8 a1e1 h7h6 e1e6 e8e6 f1e1 f6e8 e1e6 d7e6 d3e3 e6e3 d2e3 f7f6 g1f1 g7g5 h2h3 e8d6 f1e2');
INSERT INTO evaluations VALUES(32542,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(32542,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(32542,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(32542,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(32542,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(32542,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(32542,36,'cp -12','d2d4 c5d4 f3d4 g8f6 b1d2 b8c6 d2f3 e8g8 e1g1 e7c5 e2d3 a8e8 c2c3 c6d4 f3d4 e8e4 f2f3 c5d4 c3d4 e4e6 c1d2 f8e8 a1e1 h7h6 e1e6 e8e6 f1e1 e6e1 d2e1 f6e8 d3b3 f7f6 g1f2 a7a6 h2h3 g8f7 f2g1 h6h5 e1c3 g7g5 a2a3');
INSERT INTO evaluations VALUES(30489,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(30489,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(30489,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(30489,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(30489,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(30489,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(30489,36,'cp 102','g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 f8e7 b1a3 c5c4 d3c2 a7a6 c2b1 b8c6 a3c2 b6b5 a2a3 a6a5 c1d2 d7b6 e2d1 b6a4 d1c1 e8d7 f1e1 h7h6 c2e3 a8b8 b1c2 h6h5 b2b3 c4b3 c2b3 h5h4 b3c2');
INSERT INTO evaluations VALUES(33438,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(33438,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(33438,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(33438,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(33438,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(33438,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(33438,36,'cp 10','a7a6 f1d3 c7c5 d4c5 d7c5 e1g1 c8d7 d1d2 e8g8 c3e2 c5d3 c2d3 b8c6 a2a3 f8c8 h2h3 h7h6 e2d4 a6a5 f1e1 a5a4 a1b1 e7d8 d2f4 c6d4 f4d4 b7b5 e1c1 c8c7 b1a1 a8c8 c1c7 d8c7');
INSERT INTO evaluations VALUES(33338,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(33338,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(33338,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(33338,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(33338,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(33338,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(33338,36,'cp 24','f1d3 c7c5 d4c5 d7c5 d1d2 e8g8 e1g1 b8c6 f1e1 b7b5 c3e2 b5b4 e2d4 c6d4 f3d4 a6a5 e1e3 c8b7 e3f3 b7a6 f3g3 f7f5 e5f6 f8f6 a1e1 a6d3 c2d3 a8f8 g3f3 e6e5 f3f6 e7f6 d4f3 e5e4 d3e4 d5e4 d2d5 g8h8 d5c5 e4f3 c5a5 f6b2');
INSERT INTO evaluations VALUES(34966,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(34966,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(34966,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(34966,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(34966,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(34966,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(34966,36,'cp 91','d1e2 e5e2 g1e2 g8f6 c1f4 c7c6 e1c1 f8e7 e2g3 c8e6 c3e4 f6e4 g3e4 b8d7 e4d6 e7d6 d1d6 d7f6 f1e2 e8g8 h1e1 f8e8 f4e3 h7h6 b2b3 f6d5 e3d2 a8d8 d6d8 e8d8 e1d1 d5e7 a2a4 b7b6 e2a6 e7f5 a4a5 c6c5 f2f3 f5d4 d2e3');
INSERT INTO evaluations VALUES(38236,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(38236,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(38236,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(38236,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(38236,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(38236,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(38236,36,'cp 54','d2d4 g8f6 f1d3 b8d7 e1g1 e8g8 d1c2 f8e8 f1e1 e7f8 c1e3 c7c6 b1d2 a7a5 h2h3 h7h6 b2b3 b7b5 a2a4 c8a6 e1d1 d8c7 a4b5 c6b5 d1c1 a5a4 b3a4 b5a4 d3a6 e5d4 e3d4 a8a6 a1a4 a6a4 c2a4 c7b7 c1a1 f6e4 d2e4 e8e4');
INSERT INTO evaluations VALUES(38230,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(38230,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(38230,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(38230,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(38230,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(38230,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(38230,36,'cp 62','b8d7 f1d3 g8f6 e1g1 e8g8 d1c2 f8e8 h2h3 e7f8 f1e1 a7a5 c1e3 c7c6 b1d2 a5a4 b2b3 a4b3 a2b3 a8a1 e1a1 h7h6 a1d1 d8c7 d1e1 b7b5 e1a1 d7b6 d4e5 d6e5 c3c4 b5c4 d2c4 b6c4 d3c4 c8e6 a1a6 c7b7 a6a7 b7b8 c4e6 e8e6');
INSERT INTO evaluations VALUES(39029,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(39029,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(39029,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(39029,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(39029,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(39029,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(39029,36,'cp 82','d7d5 h2h4 h7h6 e4e5 c8d7 g1f3 g8e7 b2b4 b7b6 d1d2 e7f5 c3d1 c7c5 c2c3 f5e3 d1e3 c5d4 c3d4 b6b5 f1d3 a7a5 e1g1 a5b4 a1c1 h6h5 d2b4 b8c6 b4b1 c6e7 d3b5 e8g8 b5d7 d8d7');
INSERT INTO evaluations VALUES(41628,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(41628,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(41628,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(41628,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(41628,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(41628,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41628,36,'cp 16','c3a4 c5b6 c2c3 e8g8 e1g1 b8c6 a4b6 a7b6 c4b3 c8e6 b3e6 f7e6 a2a4 d8d7 c1e3 f8f7 b2b4 a8f8 h2h3 c6e7 c3c4 e7g6 f3h2 d6d5 h2g4 d7e7 g4f6 e7f6 c4d5 e6d5 e4d5 f8d8 a4a5 d8d5 a5b6 c7b6 d1b3 d5d7 b3c4 g8h7');
INSERT INTO evaluations VALUES(42751,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(42751,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(42751,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(42751,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42751,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42751,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(42751,36,'cp -12','d7d6 a2a3 b4c3 b2c3 e8g8 d2d3 c8e6 c4e6 f7e6 f3h3 d8e8 g1e2 f6h5 e1g1 b8d7 f2f3 h5f4 c1f4 e5f4 g2g3 f4g3 e2g3 e6e5 c3c4 b7b6 g3f5 d7f6 f3f4 f6h5 f4e5 d6e5 h3g4 f8f6 g1h1 g8h8 g4h3');
INSERT INTO evaluations VALUES(43020,30,'cp -149',NULL);
INSERT INTO evaluations VALUES(43020,31,'cp -155',NULL);
INSERT INTO evaluations VALUES(43020,32,'cp -160',NULL);
INSERT INTO evaluations VALUES(43020,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(43020,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(43020,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(43020,36,'cp -147','b8c6 e5f4 e8g8 d2d3 d7d5 f4d2 f8e8 b1c3 b7b5 c3b5 e7c5 b5c3 a7a5 g1e2 a5a4 a2a3 h7h6 d3d4 c6d4 e2d4 d5e4 d4e2 e4f3 g2f3 d8e7 d2f4 c8b7 c1d2 a8d8 e1c1 d8d2 d1d2 c5e3 f4f5 e8d8 c3b1 b7a6 f5a5 a6e2 f1e2 e3d2 b1d2 e7e2 a5c3 f6d5');
INSERT INTO evaluations VALUES(10872,30,'cp -157',NULL);
INSERT INTO evaluations VALUES(10872,31,'cp -173',NULL);
INSERT INTO evaluations VALUES(10872,32,'cp -180',NULL);
INSERT INTO evaluations VALUES(10872,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(10872,34,'cp -180',NULL);
INSERT INTO evaluations VALUES(10872,35,'cp -190',NULL);
INSERT INTO evaluations VALUES(10872,36,'cp -174','e5f4 e8g8 d2d3 d7d5 f4d2 f8e8 d2f2 c6b4 b1c3 e7c5 f2d2 b7b5 a2a3 b4c6 g1e2 c5b6 b2b3 a7a5 e2f4 c6d4 f4e2 c8a6 e2d4 b6d4 c1b2 c7c5 f1e2 d5e4 d3e4 d4f2 e1d1 f2d4 d1c1 b5b4 a3b4 c5b4 e2a6');
INSERT INTO evaluations VALUES(44948,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(44948,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(44948,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(44948,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(44948,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(44948,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(44948,36,'cp 14','c1f4 c8f5 f3h4 f5b1 a1b1 e7e6 e2e3 f8d6 f1d3 e8g8 e1g1 d8c7 f4d6 c7d6 c3c4 b8d7 b1c1 f8c8 h4f3 c6c5 d4c5 d7c5 d3e2 c5e4 c4d5 d6d5 d1d5 f6d5 f1d1 g7g6 a2a3 d5f6 f3d4 c8c1 d1c1 a8d8 g2g3 e6e5 d4b5');
INSERT INTO evaluations VALUES(44925,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(44925,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(44925,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44925,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(44925,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(44925,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(44925,36,'cp 18','c8f5 f3h4 f5b1 a1b1 e7e6 e2e3 f8d6 f1d3 d6f4 e3f4 e8g8 e1g1 g7g6 g2g3 b8d7 g1g2 c6c5 h4f3 c5d4 f3d4 d8c7 d1d2 a7a6 b1e1 a8c8 h2h4 f8e8 d3c2 h7h5 e1e2 d7b6 f1e1 b6c4 d2c1 b7b5 f2f3 c7b6 a2a3');
INSERT INTO evaluations VALUES(28983,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(28983,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(28983,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(28983,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(28983,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(28983,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(28983,36,'cp 17','f3h4 f5b1 a1b1 e7e6 e2e3 f8d6 f1d3 d6f4 e3f4 d8c7 d1d2 e8g8 e1g1 c6c5 d4c5 c7c5 f1e1 b8c6 h4f3 g7g6 g2g3 f8d8 d2e2 a8c8 a2a3 a7a6 b1a1 g8g7 h2h3 h7h6 g1g2 d5d4 c3c4 a6a5 a1c1 f6d7 c1d1 b7b6 h3h4');
INSERT INTO evaluations VALUES(46323,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(46323,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(46323,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(46323,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(46323,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(46323,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(46323,36,'cp 33','d3c2 f8d6 g1f3 d6f4 e3f4 d8b6 d1c1 e8g8 b1d2 b6c7 f3e5 f6e8 c1d1 b8c6 d2f3 f7f6 e5c6 b7c6 d1d2 a7a5 e1g1 a5a4 b2b3 a4b3 a2b3 a8a1 f1a1 c4b3 c2b3 e8d6 b3c2 c8d7 d2d3 g7g6 h2h4 f8b8 g2g4');
INSERT INTO evaluations VALUES(46320,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(46320,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(46320,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(46320,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(46320,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(46320,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46320,36,'cp 12','f8d6 g1f3 b7b5 b1d2 c8b7 e1g1 e8g8 f4d6 d8d6 f3e5 b8d7 f2f4 f6e4 a2a4 a7a6 d1h5 g7g6 h5e2 e4d2 e2d2 b7c6 a4b5 a6b5 h2h4 f7f5 a1a8 f8a8 g2g4 d7e5 f4e5 d6e7 d2g2 g8h8 g2h2 b5b4 c3b4 e7b4 h4h5 a8g8');
INSERT INTO evaluations VALUES(49514,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(49514,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(49514,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(49514,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(49514,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(49514,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(49514,36,'cp -25','f8b4 g5d2 b4d2 d1d2 e5d4 d2d4 d7d5 c4d5 b8c6 d4e3 c6e7 b1c3 e8g8 a1d1 e7d5 c3d5 f6d5 e3c5 c7c6 e2e3 d8f6 c5d4 c8e6 a2a3 f6d4 d1d4 d5f6 d4d2 a8d8 f1d3 e6d5 g1f3 f6e4 d3e4 d5e4 h1g1 e4d5 e1e2 d5f3 g2f3 d8d2 e2d2 f8d8 d2e2 d8d5 e3e4 d5c5 g1d1');
INSERT INTO evaluations VALUES(49545,30,'cp -229',NULL);
INSERT INTO evaluations VALUES(49545,31,'cp -202',NULL);
INSERT INTO evaluations VALUES(49545,32,'cp -211',NULL);
INSERT INTO evaluations VALUES(49545,33,'cp -230',NULL);
INSERT INTO evaluations VALUES(49545,34,'cp -204',NULL);
INSERT INTO evaluations VALUES(49545,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(49545,36,'cp -182','d7d5 c4b3 a7a5 a2a3 e5e4 f3d1 f8d6 d2d3 e8g8 g1e2 d8e7 b1c3 c7c6 e2g3 b8a6 b3a2 a6c5 d3d4 c5a6 f2f3 e4f3 d1f3 b7b5 e1g1 b5b4 c3a4 c8g4 f3f2 d6g3 f2g3 g4e2 f1e1 e2b5 g3e5 e7e5 d4e5');
INSERT INTO evaluations VALUES(48788,30,'cp -194',NULL);
INSERT INTO evaluations VALUES(48788,31,'cp -190',NULL);
INSERT INTO evaluations VALUES(48788,32,'cp -200',NULL);
INSERT INTO evaluations VALUES(48788,33,'cp -192',NULL);
INSERT INTO evaluations VALUES(48788,34,'cp -210',NULL);
INSERT INTO evaluations VALUES(48788,35,'cp -212',NULL);
INSERT INTO evaluations VALUES(48788,36,'cp -216','c4b3 a7a5 a2a3 f8d6 f3d1 e8g8 d2d3 b8a6 g1e2 a6c5 b3a2 b7b5 f2f3 e5e4 d3e4 d5e4 f3f4 c8a6 e1g1 c5a4 c2c4 a4b6 b1c3 b5c4 e2g3 f8e8 a2b1 a6b7 d1e2 g7g6 b1c2 d8e7 c1d2 h7h5 a1d1 b7c6 e2e1 h5h4 g3e2 h4h3');
INSERT INTO evaluations VALUES(49763,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49763,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49763,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49763,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(49763,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(49763,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(49763,36,'cp 0','e4e5 f6d5 d1f3 c7c6 b1c3 d5c3 b2c3 d7d6 f3f4 d8a5 f4g3 d6e5 f1c4 e5d4 g1e2 d4c3 h1f1 c8e6 c4e6 f7e6 g3e3 e6e5 e3h3 f8e7 c1g5 e7g5 h3c8 a5d8 c8e6 d8e7 e6c8');
INSERT INTO evaluations VALUES(49501,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(49501,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(49501,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49501,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(49501,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(49501,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49501,36,'cp -8','f6d5 d1f3 c7c6 b1c3 d5c3 b2c3 d7d6 c1f4 d6e5 f4e5 b8d7 e5f4 f8a3 f3e3 a3e7 f1d3 d7b6 g1e2 b6d5 e3f3 d5f4 e2f4 g7g6 e1g1 e8g8 a1b1 e7d6 a2a4 a8b8 f3g3 c8d7 h2h4 d8f6 f1f3 f6g7 g1h1 b7b6 b1e1 b8e8 e1f1');
INSERT INTO evaluations VALUES(50677,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(50677,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(50677,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(50677,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(50677,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(50677,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(50677,36,'cp -14','e7e6 e1g1 g8f6 e4e5 d6e5 d1d8 e8d8 f3e5 d8e8 c4e2 b8c6 e5c4 c8d7 c1e3 c6b4 f1c1 b4d5 c3d5 f6d5 e3d4 h8g8 e2f3 a8c8 f3d5 e6d5 c1e1 d7e6 c4b6 c8c6 b6d5 f8d6 d4c3 e8d7 a1d1 c6c5 d5e3 f7f6 d1d3 g8d8 e1d1 d7c7 a2a3 d6e7 d3d8 e7d8');
INSERT INTO evaluations VALUES(52761,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(52761,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(52761,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(52761,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(52761,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(52761,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(52761,36,'cp -7','e2e3 g8f6 f1d3 c8g4 g1f3 e7e6 h2h3 g4f5 e1g1 f5g6 c2c4 b8d7 d3g6 h7g6 d1b3 d5c4 b3b7 d8c8 b7c8 a8c8 c1d2 c6c5 f1c1 c5d4 f3d4 f6e4 d2e1 e6e5 f4e5 d7e5 b1d2 e4d2 e1d2 f8c5 g1f1 f7f6 d4f3 e5f3 g2f3 c5b6');
INSERT INTO evaluations VALUES(53482,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(53482,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(53482,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(53482,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(53482,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(53482,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(53482,36,'cp 102','c2c3 c6c5 f1d3 b8c6 e1g1 c8d7 f1e1 a7a6 a2a3 c5c4 d3c2 d8c7 b1d2 c6a5 a1b1 g8e7 d2f1 e8c8 c1f4 c8b8 f1e3 e7c8 h2h4 f8e7 f4g3 c8b6 h4h5 d8g8 g3h4 a5c6 h4e7 c6e7 g2g3 b8a8 e3g2 g7g5 h5g6 f7g6');
INSERT INTO evaluations VALUES(56608,30,'cp -132',NULL);
INSERT INTO evaluations VALUES(56608,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(56608,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(56608,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(56608,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(56608,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(56608,36,'cp -150','b8c6 f1b5 c8d7 e1g1 a7a6 b5a4 f7f5 e4g3 h7h5 d2d4 h5h4 a4c6 d7c6 g3f5 d8d7 d1g4 g8e7 d4d5 c6d5 b1c3 d5c6 f5e3 h4h3 g2g3 d7g4 e3g4 h8h5 f1e1 e7g6 c3e4 e8c8 f2f3 h5f5 e1f1 f8e7 c1e3 g6f8 a1e1 f8e6 c2c3 a6a5 f1f2 b7b6 g1f1 a5a4 f1g1 f5f8');
INSERT INTO evaluations VALUES(56642,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(56642,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(56642,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(56642,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(56642,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56642,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(56642,36,'cp -9','d8d4 e3d4 g8f6 g1f3 b8c6 e1g1 c8e6 c4e6 f7e6 b1c3 e8c8 c1e3 h7h6 h2h4 f8d6 a1d1 h8e8 g2g3 e6e5 d4e5 c6e5 f3e5 d6e5 g1g2 a7a6 f1e1 b7b5 g2f3 e5d4 e3d4 e8e1 d1e1 d8d4 e1e7 d4d2 e7g7');
INSERT INTO evaluations VALUES(34972,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(34972,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(34972,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34972,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34972,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34972,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34972,36,'cp -1','e3d4 g8f6 g1f3 b8c6 e1g1 c8e6 c4e6 f7e6 b1c3 e8c8 c1e3 h7h6 a1e1 f8b4 a2a3 b4a5 e1c1 a5c3 c1c3 f6d5 c3c4 g7g5 f1c1 h8g8 h2h3 d8d7 c4c2 a7a6 e3d2 g8d8 c1d1 c8b8 d2c3 d5f4 g2g3');
INSERT INTO evaluations VALUES(34971,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34971,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34971,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34971,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34971,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(34971,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(34971,36,'cp 0','g8f6 g1f3 b8c6 e1g1 c8e6 c4e6 f7e6 c1d2 e8c8 d2c3 f8b4 f1d1 b4c3 b1c3 h8e8 g1f1 d8d6 d1d2 e6e5 d4e5 d6d2 f3d2 e8e5 a1e1 e5e1 f1e1 c6d4 d2e4 f6e4 c3e4 b7b6 e1d2 c8d7 f2f4 c7c5 d2e3 d7e6 g2g4');
INSERT INTO evaluations VALUES(57240,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(57240,31,'cp -142',NULL);
INSERT INTO evaluations VALUES(57240,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(57240,33,'cp -150',NULL);
INSERT INTO evaluations VALUES(57240,34,'cp -166',NULL);
INSERT INTO evaluations VALUES(57240,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(57240,36,'cp -171','f8a3 b1a3 d8e7 a3b1 d7d5 e2e3 c7c5 f1b5 e8f8 b5e2 b8c6 e2b5 c8d7 d2d3 g8f6 b1d2 h7h5 b5c6 d7c6 g1e2 h5h4 h2h3 a7a6 a2a3 a8d8 d1c1 h8h6 e1g1 g7g5 f2f3 f6h5 e3e4 h5g3 f1e1');
INSERT INTO evaluations VALUES(39398,30,'cp -168',NULL);
INSERT INTO evaluations VALUES(39398,31,'cp -153',NULL);
INSERT INTO evaluations VALUES(39398,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(39398,33,'cp -147',NULL);
INSERT INTO evaluations VALUES(39398,34,'cp -141',NULL);
INSERT INTO evaluations VALUES(39398,35,'cp -141',NULL);
INSERT INTO evaluations VALUES(39398,36,'cp -150','b1a3 d8e7 a3b1 d7d5 e2e3 c7c5 f1b5 e8f8 c2c4 a7a6 b5a4 h7h5 d2d3 g8f6 g1f3 e5e4 d3e4 d5e4 f3g5 b8d7 a4d7 c8d7 b1c3 h5h4 c3d5 f6d5 d1d5 d7c6 d5f5 g7g6 f5f4 a8d8 a1d1 d8d1 e1d1 f8g7 g5h3 f7f6 d1c2 g6g5 f4g4 c6d7');
INSERT INTO evaluations VALUES(39397,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(39397,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(39397,32,'cp -152',NULL);
INSERT INTO evaluations VALUES(39397,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(39397,34,'cp -159',NULL);
INSERT INTO evaluations VALUES(39397,35,'cp -144',NULL);
INSERT INTO evaluations VALUES(39397,36,'cp -162','d7d5 c2c3 g8f6 e2e3 e8g8 d2d4 f6e4 d4e5 e4c3 d1c2 c3e4 f1d3 c8f5 g1f3 b8c6 c2b2 d8e7 a3c2 f5g4 d3e2 f8e8 a1d1 c6e5 b2e5 e7e5 f3e5 g4e2 d1d5 e2a6 d5a5 c7c5 e5c4 a6c4 b3c4 e8d8 f2f3 d8d2 f3e4 d2c2 a5c5 c2g2 c5c7 g2a2');
INSERT INTO evaluations VALUES(57407,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(57407,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(57407,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(57407,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(57407,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(57407,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(57407,36,'cp 77','b8c6 a2a3 f8c5 b2b4 c5d4 a1a2 d7d6 b4b5 c6e7 f3d4 e5d4 d2d3 c7c5 f1e2 a7a6 e1g1 e7g6 f2f4 g8e7 a3a4 a6b5 c4b5 e8g8 b1d2 d6d5 d1c2 b7b6 a4a5 c8d7 e4d5 d7b5 a5b6 d8b6 a2a8 f8a8 d5d6 e7d5 c2b3 b5c6 e2f3 b6b3 d2b3 a8d8 b3a5 c6a8 a5c4 d5c3 f3a8 d8a8 d6d7');
INSERT INTO evaluations VALUES(4302,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(4302,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(4302,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(4302,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(4302,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(4302,35,'cp -108',NULL);
INSERT INTO evaluations VALUES(4302,36,'cp -90','c6a5 c4b3 a5b3 a2b3 d7d5 e4d5 f6d5 e1g1 f7f6 g5e4 c5e7 b1a3 c8f5 a3c4 d8d7 c4e3 d5e3 c1e3 a7a5 f2f4 e5f4 e3f4 f5g6 d1f3 f8e8 e4c3 e7c5 g1h1 c7c6 f4h2 b7b5 h2g1 c5d6 f1e1 d6c7 c3e4 c7e5 e4c5 d7d5 f3d5 c6d5 d3d4 e5d6 c2c3 e8e1 a1e1');
INSERT INTO evaluations VALUES(5178,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5178,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(5178,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(5178,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(5178,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(5178,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(5178,36,'cp 8','d4b3 c5b4 c1d2 a7a5 b1c3 a5a4 a2a3 b4e7 b3c1 g8f6 f1d3 e8g8 e1g1 c6e5 h2h3 d7d6 c3a4 e5d3 c1d3 f6e4 d2e3 d6d5 c4d5 e6d5 d3f4 e7g5 a4c3 e4c3 b2c3 g5f4 e3f4 f8e8 d1d2 c8f5 f1e1 d8f6 e1e8 a8e8 a3a4 f6e6 f2f3 e6c6 a4a5 e8c8 a1d1 f5e6 f4e3 c6c3 d2c3 c8c3');
INSERT INTO evaluations VALUES(6192,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(6192,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(6192,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(6192,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(6192,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(6192,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(6192,36,'cp 254','b4b5 c6e7 f3e5 d7d5 e4d5 g8f6 c1a3 e8g8 e1g1 f8e8 a3e7 e8e7 d2d4 f6d5 d1f3 c8e6 a2a4 b6a5 f1c1 a8c8 b1d2 c7c5 b5c6 b7c6 d2e4 e7e8 h2h3 f7f6 e5g6 d8c7 h3h4 d5f4 f3f4 c7f4 g6f4 e6c4 e4d6 a5c7 d6e8 c7f4');
INSERT INTO evaluations VALUES(6528,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(6528,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(6528,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6528,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6528,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6528,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6528,36,'cp 12','d7d5 e1g1 e5e4 f3e1 c8g4 b1c3 f7f5 c3a4 c5e7 f2f3 g4h5 d3e4 d5e4 d1d8 a8d8 f3e4 f5e4 g2e4 h5e2 f1f2 e2c4 e4c6 b7c6 b2b3 c4b5 e1d3 b5d3 c2d3 d8d3 c1f4 g8f6 f4c7 e8g8 a1c1 f6g4 f2f8 e7f8 c7f4 f8d6 a4c5 d3d5 c5e6');
INSERT INTO evaluations VALUES(5872,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(5872,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5872,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(5872,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(5872,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(5872,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(5872,36,'cp 11','e1g1 e5e4 f3e1 f7f5 c2c4 d5c4 d1a4 c4d3 e2d3 d8d4 a4b3 d4b4 d3e4 f5e4 b1c3 g8f6 c3e4 f6e4 e1d3 b4b3 a2b3 c6b4 g2e4 b4d3 e4d3 c8e6 f1e1 e8c8 d3h7 h8h7 e1e6 d8d1 g1g2 c5f2 g2f2 h7h2 f2f3 h2c2 e6e7 c2c1 a1c1 d1c1 e7g7 c1d1 g7h7 d1f1 f3g4');
INSERT INTO evaluations VALUES(5871,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(5871,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5871,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(5871,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(5871,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(5871,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(5871,36,'cp 8','e5e4 f3e1 f7f5 c2c4 d5c4 d1a4 c4d3 e2d3 d8d4 a4b3 d4b4 d3e4 b4b3 a2b3 f5e4 b1c3 c8f5 c3e4 c5b6 c1e3 b6e3 f2e3 f5e4 g2e4 g8f6 e4c6 b7c6 e1f3 e8g8 f3e5 f8e8 e5c4 e8e6 a1a2 f6g4 f1e1 g4e5 e1e2 e5c4 b3c4 a7a5 g1f2 e6f6 f2g2');
INSERT INTO evaluations VALUES(7293,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(7293,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(7293,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(7293,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(7293,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(7293,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(7293,36,'cp 147','b5c6 d7c6 f3e5 d8e7 e5f3 c8g4 h2h3 g4h5 c1e3 e8c8 d1e2 c5e3 e2e3 h5f3 e3f3 g7g5 b1d2 h7h5 f3e3 f6d7 g2g3 c8b8 g1g2 h8g8 d2c4 f7f6 f2f4 g5f4 e3f4 g8g5 c4e3 d8g8 e3f5 e7b4 b2b3 b4c3 f4f2 d7e5');
INSERT INTO evaluations VALUES(7286,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(7286,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(7286,32,'cp 147',NULL);
INSERT INTO evaluations VALUES(7286,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(7286,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(7286,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(7286,36,'cp 131','d7c6 f3e5 d8e7 e5f3 c8g4 h2h3 g4e6 d1e1 e8c8 b1c3 h8g8 c3a4 c5a7 c1e3 a7e3 e1e3 b7b6 f3g5 h7h6 g5e6 f7e6 a4c3 g7g5 a2a4 g5g4 h3h4 c8b7 a4a5 g4g3 f2g3 f6g4 e3e2 b6b5 e4e5 e7c5 g1h1 c5e5');
INSERT INTO evaluations VALUES(7205,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(7205,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(7205,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(7205,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(7205,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(7205,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(7205,36,'cp 156','f3e5 d8e7 e5f3 c8g4 h2h3 g4e6 f3g5 e8c8 b1d2 f6d7 g5e6 f7e6 d1h5 g7g5 d2b3 h7h6 b3c5 e7c5 h5e2 e6e5 c1e3 c5e7 g2g3 d8g8 g1g2 b7b6 d3d4 c8b7 a1d1 e5d4 e3d4 d7e5 f1e1 g8f8 d4e5 e7e5 c2c3 h6h5 f2f3 a6a5');
INSERT INTO evaluations VALUES(7195,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(7195,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(7195,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(7195,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(7195,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(7195,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(7195,36,'cp 174','d8e7 e5f3 c8g4 h2h3 g4h5 d1e2 e8c8 c1e3 c5e3 e2e3 h5f3 e3f3 c8b8 f3e3 f6h5 b1c3 g7g5 c3e2 h8g8 e2g3 h5f4 a1b1 h7h5 g3f5 e7f6 a2a4 g8g6 b2b4 f4d5 e3e1 d5e7 f5e7 f6e7 b4b5 a6b5 a4b5 g5g4 h3g4 h5g4');
INSERT INTO evaluations VALUES(11762,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(11762,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(11762,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(11762,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(11762,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(11762,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(11762,36,'cp 41','d7d5 e1g1 d5d4 a2a4 a8b8 a4b5 a6b5 c2b3 e7g6 d2d3 f8e7 b3d5 b8b6 b1a3 d8d7 a3c2 e8g8 d5c6 b6c6 c3d4 e5d4 c2d4 c6d6 d4f5 d6d3 f5e7 g6e7 d1e2 e7c6 c1f4 c8b7 h2h3 f7f6 f1c1 f8d8 c1c5 d3d1 a1d1 d7d1 e2d1 d8d1 g1h2 c6d4 f3d4 d1d4');
INSERT INTO evaluations VALUES(12775,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(12775,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(12775,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(12775,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(12775,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(12775,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(12775,36,'cp -84','e7e6 c4b3 d7d5 e4d5 c6d4 f3d1 e6d5 b3a4 c8d7 a4d7 d8d7 c3e2 d4c6 g1f3 d5d4 d2d3 f8d6 e1g1 e8g8 e2g3 a7a5 c1d2 a5a4 c2c3 a4a3 d1b3 a3b2 b3b2 c6e5 f3e5 d6e5 a1e1 f8e8 c3c4 e5d6 e1e8 a8e8 b2b3');
INSERT INTO evaluations VALUES(13164,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(13164,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(13164,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(13164,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(13164,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(13164,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(13164,36,'cp 13','c8g4 c2c3 e7e6 c1f4 f6h5 f4e3 h5f6 h2h3 g4h5 e3f4 f8d6 f4d6 d8d6 f1e1 d6c7 b1d2 e8g8 d1e2 c7b6 d2b3 a8e8 e2e3 h5f3 e3f3 e6e5 d4e5 e8e5 f3f4 f8e8 e1e5 c6e5 d3f1 e5c4 f1c4 e8e4 f4f5 e4c4 a1e1 g7g6 f5e5 c4e4 e1e4 f6e4 b3d4 b6b2');
INSERT INTO evaluations VALUES(14115,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(14115,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(14115,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(14115,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(14115,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(14115,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(14115,36,'cp 30','b2b3 f8g7 f1e2 e8g8 c1b2 f6e4 e1g1 e4c3 b2c3 f8e8 d1d2 c8g4 f1d1 e7e6 a2a4 c6e7 c3b2 e7f5 c2c4 g4f3 e2f3 c7c6 g2g3 a7a5 d2c2 h7h5 f3g2 d8e7 b2a3 e7d8 a1c1 g7f8 a3b2 f8g7 h2h4 d8e7 c1a1 e7d8 c2d3 d8b6');
INSERT INTO evaluations VALUES(14441,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(14441,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(14441,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(14441,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(14441,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(14441,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(14441,36,'cp 61','h2h4 f7f6 b1d2 e4d2 d1d2 e7e5 d4e5 f6e5 f4g3 a7a6 f1e2 c8e6 g1f3 d8f6 e1g1 h7h6 a1c1 f8e7 c3c4 e8c8 c4d5 d8d5 d2c2 d5c5 c2b1 e6f5 e2d3 h8f8 c1c5 e7c5 f1c1 c5e7 b1c2 f5d3 c2d3 f6f5 d3c4 e7d6 f3d2 g7g5 d2e4 f5f7 h4g5 h6g5 c4f7 f8f7');
INSERT INTO evaluations VALUES(15052,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(15052,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(15052,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(15052,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(15052,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(15052,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(15052,36,'cp 15','e1g1 f8d6 c1g5 c8e6 c2c3 d8d7 b1d2 f6e4 d2e4 d5e4 f3d2 f7f5 d2c4 h7h6 c4d6 c7d6 g5e3 d6d5 f2f3 e8g8 f3e4 f5e4 d1b3 c6e7 e3f4 e7f5 a1e1 f5h4 b3b5 d7b5 e2b5 h4f3 g2f3 f8f4 f3e4 f4g4 g1f2 d5e4 b5e2 a8f8 f2e3 g4g2 f1f8 g8f8 e3e4 g2h2 d4d5');
INSERT INTO evaluations VALUES(15051,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(15051,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(15051,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(15051,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(15051,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(15051,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(15051,36,'cp 23','f8d6 c1g5 c8e6 c2c3 d8d7 b1d2 f6e4 d2e4 d5e4 f3d2 f7f5 d2c4 h7h6 c4d6 c7d6 g5e3 d6d5 f2f3 e8g8 d1d2 a8d8 b2b3 c6e7 f3e4 f5e4 f1f8 d8f8 a1f1 f8f1 e2f1 a7a5 e3f4 g8h7 h2h3 d7c6 f1e2 a5a4 b3a4 c6a4 d2b2 e7g6 f4g3');
INSERT INTO evaluations VALUES(9196,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(9196,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(9196,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(9196,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(9196,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(9196,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(9196,36,'cp 29','c1g5 c8e6 c2c3 d8d7 b1d2 f6e4 d2e4 d5e4 f3d2 f7f5 d2c4 h7h6 c4d6 c7d6 g5e3 d6d5 f2f3 e8g8 d1d2 a8d8 a1d1 c6e7 f3e4 f5e4 f1f8 d8f8 d1f1 f8f1 e2f1 g8h7 e3f4 a7a5 f4b8 e7g6 h2h3 a5a4 b8g3 d7c6 g1h2 g6e7 g3e5 e7g6 e5b8 g6e7 f1e2 e6d7');
INSERT INTO evaluations VALUES(16647,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(16647,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(16647,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(16647,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(16647,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(16647,35,'cp -111',NULL);
INSERT INTO evaluations VALUES(16647,36,'cp -89','g7g5 b1c3 h7h6 h2h4 h8g8 h4g5 h6g5 d1e2 d7d6 c4b5 c8d7 c1d2 d8e7 e2f2 e8c8 b5c6 d7c6 f2a7 d6d5 f3d4 e7c5 d4c6 c5c6 a7a8 c8d7 a8a5 f8c5 h1h6 d8a8 e4d5 f6d5 h6c6 a8a5 c3d5 d7c6 d2a5 c6d5 a5c7 c5d4 e1f1 d5c6 c7a5');
INSERT INTO evaluations VALUES(16698,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(16698,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(16698,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(16698,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(16698,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(16698,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(16698,36,'cp 40','e4e5 f6g4 d1e2 d7d6 e5d6 c8e6 d6c7 d8c7 f3g5 c7d6 g5e6 d6e6 e2e6 f7e6 f1e2 g4f6 c1e3 f8b4 e1g1 a8c8 a1c1 b4c3 b2c3 e8f7 c3c4 h8d8 f1d1 d8d1 c1d1 b7b6 g2g3 c6a5 d1c1 a5b7 h2h4 b7c5 e3f4 h7h6 g1g2 f7e7 f4e3 c5d7 c1d1 e6e5');
INSERT INTO evaluations VALUES(17667,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(17667,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(17667,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(17667,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(17667,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(17667,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(17667,36,'cp -8','g8f6 e1g1 f8e7 f1e1 b7b5 c4a2 c6e5 f3e5 d6e5 d1f3 e8g8 f3g3 d8c7 a3a4 c8d7 c1h6 f6e8 a4b5 a6b5 a2b3 a8b8 h2h3 g8h8 h6e3 e8f6 a1a6 b8a8 e1a1 a8a6 a1a6 d7c6 g3f3 c7b7 a6a1 f8c8 c3b1 c5c4 d3c4 b5c4');
INSERT INTO evaluations VALUES(18797,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(18797,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(18797,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(18797,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(18797,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(18797,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(18797,36,'cp 15','e2e4 d7d5 e4d5 d8d5 f1e2 g8f6 e1g1 c5d4 f3d4 c8d7 e2f3 d5d6 d4c6 d7c6 d1d6 f8d6 f3c6 b7c6 b1d2 h7h6 a2a4 f6d5 a4a5 f7f6 d2c4 d6c7 c1e3 d5e3 c4e3 a8d8 g2g3 e8f7');
INSERT INTO evaluations VALUES(20752,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(20752,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(20752,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(20752,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(20752,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(20752,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(20752,36,'cp 150','g8f6 c1f4 f8e7 f1b5 c8d7 b5d7 d8d7 d1d3 e8g8 e1c1 d7e6 h1e1 a8d8 c1b1 a7a6 f4g5 f6g4 g5e7 e6e7 c3d5 e7e6 d3g3 g4f6 f2f3 f6d5 d1d5 f7f5 d5f5 f8f5 e4f5 e6f5 e1e7 f5f6 e7b7 d8e8 a2a3 f6d4 b7d7 e8e2 g3g4 d4g4 f3g4 e2g2 d7d6 g2h2 d6a6 g8f7 a6a7 f7f6');
INSERT INTO evaluations VALUES(21182,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(21182,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(21182,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(21182,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(21182,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(21182,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(21182,36,'cp 36','e7e5 c2c4 e5e4 f3g1 g8f6 b1c3 g7g6 d1c2 d8e7 c1d2 f8g7 c3d5 f6d5 c4d5 c6b8 g1e2 e8g8 a1c1 c7c6 e2f4 g7h6 d5c6 h6f4 c6b7 e7b7 e3f4 c8e6 c2c7 b7b2 c7d6 b8d7 d4d5 e6f5 d6b4 b2b4 d2b4 f8c8 e1d2 d7b6 d5d6 c8c1');
INSERT INTO evaluations VALUES(26524,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(26524,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(26524,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(26524,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(26524,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(26524,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26524,36,'cp 25','e7e6 a2a3 g8f6 c1g5 f8e7 e2e3 e8g8 f1d3 d5c4 d3c4 h7h6 g5f6 e7f6 c4d3 f5d3 d1d3 c6e5 f3e5 f6e5 d3b5 e5f6 e1g1 a8b8 f1d1 c7c6 b5a4 a7a5 g2g3 f6e7 a1c1 d8b6 a4c2 f8d8 h2h4 d8d7 c3a4 b6c7 a4c5');
INSERT INTO evaluations VALUES(29846,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(29846,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(29846,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(29846,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(29846,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(29846,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(29846,36,'cp 113','c1e3 g7g6 b1c3 f8g7 c3e2 b8c6 a2a3 a8b8 g2g4 d8e7 f3g2 b7b5 c4a2 b5b4 a2b3 e8g8 h3h4 g8h7 h4h5 g6g5 e1g1 b4a3 a1a3 a7a5 e2g3 f6g4 b3a4 g4e3 f2e3 c6b4 a4b3 h7h8 g2e2 b4c6 b3a4 c6a7 g3f5 e7d8 a4b3 a7c6 e2g4 c6b4 a3a5');
INSERT INTO evaluations VALUES(29962,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29962,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29962,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29962,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29962,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29962,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29962,36,'cp 0','d2d4 g7g5 d1d3 h7h6 e4e5 d6e5 c4f7 e8f7 f3e5 f7g8 d3b3 g8h7 b3d3');
INSERT INTO evaluations VALUES(30001,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(30001,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(30001,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(30001,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(30001,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(30001,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(30001,36,'cp 127','d2d4 e5d4 f3d4 b8c6 f1b5 e6d7 e1g1 f8e7 d4c6 d7c6 b5c6 b7c6 f2f4 e8g8 d1f3 a7a5 c1e3 a5a4 a1d1 d8c8 e4e5 f6d5 c3d5 c6d5 f3d5 c8e6 d5e6 f7e6 d1d2 h7h5 e5d6 c7d6 g2g4 d6d5 e3d4 h5g4 h3g4 e7f6 g1f2 f6e7 f2e3 g8f7 f1f2 a8b8');
INSERT INTO evaluations VALUES(30649,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(30649,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(30649,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(30649,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30649,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(30649,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(30649,36,'cp 23','g8f6 d4d5 f8d6 g1f3 e8g8 c1g5 h7h6 g5f6 d8f6 g2g3 a7a5 f1g2 b8a6 e1g1 a6c5 a1c1 a5a4 f3d2 e6d5 c4d5 f8e8 d2e4 c5e4 g2e4 b7a6 f1e1 a6c4 e4d3 c4d3 d1d3 d6e5 f2f4 e5c3 c1c3 c7c5 e2e4 d7d6 g1h1 f6e7');
INSERT INTO evaluations VALUES(30742,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(30742,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(30742,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(30742,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(30742,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(30742,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(30742,36,'cp 65','h2h4 e7e6 b1c3 d7d6 e3e4 a7a6 f4g5 f8e7 d1d2 h7h6 g5e3 g8f6 d4d5 f6g4 e1c1 e6e5 g2g3 b8d7 f1h3 d7f6 f2f3 g4e3 d2e3 b6b5 c1b1 b7c8 h3c8 d8c8 g1h3 c8b7 g3g4 a6a5 h3f2 f6d7 e3e2 b5b4 c3a4 b7a6');
INSERT INTO evaluations VALUES(30873,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(30873,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(30873,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(30873,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(30873,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(30873,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(30873,36,'cp 37','d2d3 b8c6 b5c4 f8e7 f3g5 g8h6 h2h4 e8g8 c3d5 c6a5 f2f3 g8h8 c1d2 f7f6 g5h3 h6g8 d2a5 d8a5 d1d2 e7d8 d2a5 d8a5 e1e2 b7b5 c4b3 f8c8 d5e3 g8e7 g2g4 a8b8 a2a4 c5c4 d3c4 b5c4 e3c4 a5c7 h1d1 d7e6 h4h5 e6c4 b3c4 b8b2 c4b3 e7c6');
INSERT INTO evaluations VALUES(32845,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(32845,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(32845,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(32845,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(32845,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(32845,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(32845,36,'cp 87','b1c3 d5a5 c1d2 e7e6 f1c4 c7c6 g2g4 f5g6 g1f3 b8d7 c3d5 a5d8 d5f4 g8f6 d1e2 f6g4 f4e6 f7e6 c4e6 d8e7 e6g4 e7e2 e1e2 g6c2 h1e1 d7f6 e2f1 e8d8 g4e6 f8d6 f3e5 c2g6 a1c1 g6h5 f1g2 h8e8 e5c4 d6c7 d2g5');
INSERT INTO evaluations VALUES(32843,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(32843,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(32843,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(32843,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(32843,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(32843,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(32843,36,'cp 89','d5a5 d1f3 b8c6 f1b5 f5d7 d4d5 c6e5 b5d7 e5d7 g1e2 g8f6 c1f4 e8c8 e1c1 e7e5 d5e6 f7e6 e2d4 e6e5 d4e6 d8e8 e6f8 h8f8 f4g3 d7b6 h1e1 b6c4 a2a3 e5e4 c3e4 a5b5 b2b3 f6e4 f3g4 b5f5 g4f5 f8f5 b3c4 f5f8 f2f3 e4g3 h2g3 b7b6 e1e8 f8e8 c1d2 e8d8 d2e1 d8d1 e1d1 c8d7 g3g4 d7d6');
INSERT INTO evaluations VALUES(37412,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(37412,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(37412,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(37412,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(37412,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(37412,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(37412,36,'cp 38','d7d6 c1e3 c7c5 g1e2 d8a5 e2c1 c5d4 c1b3 a5h5 b3d4 b8c6 d1d2 c6d4 e3d4 c8e6 f1e2 h5a5 a1b1 a8c8 b2b3 f6d7 d4g7 g8g7 e1g1 a7a6 f1c1 g6g5 g2g3 f7f6 c1c2 c8c7 f3f4 b7b5 c4b5 a5b6 g1h1 a6b5 c3b5 c7c2 d2c2 b6e3 b1f1');
INSERT INTO evaluations VALUES(37690,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(37690,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(37690,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(37690,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(37690,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37690,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(37690,36,'cp 105','b1c3 b8c6 g1f3 d7d5 e4e5 f7f6 f1d3 f6e5 f3e5 g8f6 c1e3 e8g8 a1c1 c8d7 e1g1 a8c8 d3b1 b4d6 f2f4 d6c7 d1f3 c7b6 c1d1 b6a5 c3e2 c6e7 f3h3 e7f5 g2g4 f5e3 h3e3 d7b5 f1f3 b5a4 b2b3 a4b5 g4g5 f6e4 b1e4 d5e4');
INSERT INTO evaluations VALUES(37688,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(37688,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(37688,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(37688,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(37688,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(37688,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(37688,36,'cp 73','g8e7 f1d3 e8g8 g1f3 b8c6 e1g1 d7d6 a2a3 b4a5 c3e2 a5b6 b2b4 a7a6 a1b1 e7g6 c1g5 c6e7 a3a4 e6e5 h2h3 h7h6 g5e3 e5d4 e2d4 g6e5 a4a5 b6a7 d3e2 e7g6 d4f5 a7e3 f2e3 c8f5 e4f5 g6h4 f3h4 d8h4 d1d6');
INSERT INTO evaluations VALUES(38365,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(38365,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(38365,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(38365,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(38365,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(38365,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(38365,36,'cp 44','d2d4 g8f6 f1g2 e8g8 g1f3 c7c5 d4d5 e7e6 e1g1 e6d5 c4d5 f8e8 c1f4 c8f5 f3h4 f5g4 h2h3 f6h5 f4c1 g4d7 a2a4 h5f6 c1f4 b8a6 h4f3 f6h5 f4c1 c5c4 f3d2 a8c8 d2e4 d8e7 g3g4 f7f5 e4d6 e7d6 g4h5 a6c5 c3b5 d6e5');
INSERT INTO evaluations VALUES(40401,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(40401,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(40401,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40401,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40401,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40401,36,'cp 0','e8g8 h2h3 b8c6 d4c6 b7c6 c1g5 d8a5 d1d2 a8b8 a1b1 f8e8 b2b3 h7h6 c3d5 a5d2 d5f6 g7f6 g5d2 g8g7 f2f4 c6c5 a2a4 c8e6 d3b5 e8c8 b5a6 c8c6 a6b5');
INSERT INTO evaluations VALUES(36017,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36017,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36017,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36017,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36017,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36017,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36017,36,'cp 0','h2h3 b8c6 c1e3 d6d5 e4d5 c6b4 d5d6 d8d6 d3c4 f6d5 c3d5 b4d5 d4b5 d5e3 d1d6 e7d6 f2e3 g7b2 a1d1 c8e6 c4e6 f7e6 d1d6 a8c8 f1f8 g8f8 b5a7 c8c2 d6e6 b2c1 g1f1 f8f7 e6e4 c2a2 a7b5 a2b2 b5d6 f7f6 d6c4 b2c2 c4d6');
INSERT INTO evaluations VALUES(41850,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(41850,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(41850,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(41850,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41850,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(41850,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(41850,36,'cp 0','g1f3 h7h6 g5h4 d8e7 b1d2 g7g5 h4g3 f6h5 c2c3 b8c6 c4b5 e8g8 b5c6 b7c6 d3d4 e5d4 f3d4 h5g7 e1g1 f7f5 d4f5 c8f5 e4f5 g7f5 b2b4 c5b6 a2a4 e7f7 d2f3 a7a5 d1c2 f7h7 c2a2 h7f7');
INSERT INTO evaluations VALUES(41847,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41847,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(41847,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41847,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41847,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41847,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41847,36,'cp 0','h7h6 g5h4 d8e7 b1d2 g7g5 h4g3 f6h5 c2c3 b8c6 c4d5 a7a5 d5c6 b7c6 d3d4 e5d4 f3d4 h5g3 h2g3 e7f6 d1a4 c8d7 f2f3 c5d4 a4d4 f6d4 c3d4 d7e6 e1f2 f7f5 e4f5 e6f5 g3g4 f5g6 a1e1 e8d7 e1e3 g6f7 b2b3 a5a4 d2e4 a4b3');
INSERT INTO evaluations VALUES(43369,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(43369,31,'cp -91',NULL);
INSERT INTO evaluations VALUES(43369,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(43369,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(43369,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(43369,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(43369,36,'cp -59','e7e6 c4b3 h7h6 g5f3 b7b5 d2d3 d6d5 e4d5 e6d5 a2a3 f8e7 e1g1 e8g8 c1f4 b8c6 h2h3 f8e8 c3e2 c5c4 b3a2 e7c5 c2c3 c5b6 a2b1 a8a7 d3c4 d5c4 d1d8 e8d8 b1c2 a7e7 a1e1 c8b7 e2g3 f6d5 f4d2 c6a5 e1e7 d5e7 d2e3 b6e3 f2e3 b7f3');
INSERT INTO evaluations VALUES(44983,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(44983,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44983,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(44983,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44983,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44983,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(44983,36,'cp 0','b8d7 e1g1 e7e5 c2c4 d5c4 d3c4 f8e7 d1c2 a7a5 b1c3 e8g8 b2b3 f8e8 c1b2 e7f8 a1d1 d8c7 f3g5 h7h6 g5e4 f6e4 g2e4 d7f6 e4g2 c8e6 c3a4 a8d8 e2e3 d8d1 f1d1 e6g4 d1c1 g4h5 h2h3 e8d8 c2c3 d8e8');
INSERT INTO evaluations VALUES(45980,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(45980,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(45980,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(45980,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(45980,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(45980,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(45980,36,'cp -41','h5e2 f8e7 g1f3 d6d5 c4b3 b7b5 c2c3 e8g8 e1g1 b8c6 e4e5 f6d7 c1f4 a7a5 a2a3 d5d4 c3c4 c8a6 c4b5 a6b5 a3a4 b5a6 b1a3 d7b6 f3d2 g8h8 f4g3 a8b8 a3c4 c6b4 a1c1 d8d7 c4b6 b8b6 b3c4 f8b8 c4a6 b6a6 d2c4 b4d5 c4d6');
INSERT INTO evaluations VALUES(46461,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(46461,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(46461,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(46461,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(46461,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(46461,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(46461,36,'cp -25','g1e2 b8c6 d2d3 f8e7 e1g1 e8g8 b1c3 c8e6 f1e1 e7d6 e2f4 d8d7 f4d5 f6d5 a2d5 e6d5 c3d5 d6h2 g1h2 d7d5 b2b4 c5b4 a3b4 d5d4 c1a3 f8e8 h2g1 e8e1 d1e1 b7b5 a1d1 h7h6 e1e4 d4d7 d1e1 a7a5 c2c3 a5b4 a3b4');
INSERT INTO evaluations VALUES(47166,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(47166,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(47166,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(47166,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(47166,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(47166,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(47166,36,'cp -37','c1f4 c7c5 e2e3 b8c6 f1b5 c8d7 a2a4 a7a6 b5c6 d7c6 c2c3 e7e6 h4h5 h7h6 g1f3 a8c8 f3e5 f8e7 e1g1 e8g8 b1d2 d8b6 a1a2 a6a5 f1e1 f8d8 f4g3 f6h7 d1g4 c6e8 g3h4 e7h4 g4h4 f7f6 e5g6 e8g6 h5g6 h7g5 h4g3');
INSERT INTO evaluations VALUES(48104,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(48104,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(48104,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(48104,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(48104,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(48104,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(48104,36,'cp 99','f1c4 c7c5 g1f3 b8c6 e1g1 c5d4 f3d4 d5c3 b2c3 c8d7 f1e1 a7a6 a1b1 c6d4 c3d4 a8c8 d1e2 b7b5 c4b3 d7c6 c1e3 d8d7 e1d1 c6e4 b1a1 d7b7 d4d5 e4d5 e2g4 h7h5 g4g5 b7d7 b3d5 e6d5 a1c1 c8c4 c1c4 b5c4 d1b1');
INSERT INTO evaluations VALUES(50260,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50260,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(50260,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(50260,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(50260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50260,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50260,36,'cp 0','e2e4 d7d6 b1c3 c7c5 f1e2 b8c6 c3d5 f8g7 c2c3 e8g8 e1g1 f6d5 e4d5 c6e5 f1e1 b7b6 d3d4 e5f3 e2f3 c5d4 c3d4 c8f5 c1g5 g7f6 g5h6 f6g7');
INSERT INTO evaluations VALUES(50767,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(50767,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(50767,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(50767,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(50767,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(50767,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(50767,36,'cp 127','b8c6 f1d3 f8e7 e1g1 g8f6 h2h3 e8g8 d1e2 c8d7 c1f4 f8e8 a1d1 d8c7 d3b1 e6e5 f4g3 a8c8 g3h4 e5d4 f3d4 c6d4 d1d4 d7e6 f1d1 b7b5 h4g3 e6c4 b1d3 c4d3 e2d3 c7b7 f2f3 b5b4 c3e2 d6d5 e4e5 f6d7 d4d5 d7b6 d5d4 b6a4');
INSERT INTO evaluations VALUES(17689,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(17689,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(17689,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(17689,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(17689,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(17689,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(17689,36,'cp 124','f1d3 f8e7 e1g1 g8f6 a2a3 c8d7 d3c2 a8c8 f1e1 e6e5 d4d5 c6a5 h2h3 e8g8 c2d3 d8c7 a1b1 f8e8 c3a2 e7d8 b2b3 b7b5 c1e3 c7b8 f3d2 d8b6 d2f1 b6e3 f1e3 a5b7 b3b4 a6a5 d1e2 a5b4 a2b4');
INSERT INTO evaluations VALUES(52100,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(52100,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(52100,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(52100,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(52100,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(52100,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(52100,36,'cp 16','g8f6 d1b3 d8e7 g1f3 b8c6 b1a3 d7d6 d2d4 e5d4 e1g1 c8g4 c3d4 g4f3 b3f3 c6d4 f3d3 f6d7 c1e3 d7e5 d3c3 e5c4 c3c4 e7e6 e3d4 e6c4 a3c4 c5d4 f1d1 b7b5 c4a5 g7g6 d1d4 a8c8 d4d5 f8g7 a5b7 e8g8 d5d2 c8c4 b7d6 c4b4 a1b1 g7e5 a2a3 b4b3 d2d5 e5d6 d5d6 b5b4 d6a6 f8d8 a6a7 d8d2 a7a8 g8g7 f2f3 d2b2');
INSERT INTO evaluations VALUES(44312,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(44312,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(44312,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44312,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(44312,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(44312,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44312,36,'cp 27','d1b3 d8e7 g1f3 b8c6 b1a3 d7d6 d2d4 e5d4 e1g1 c8g4 c3d4 g4f3 b3f3 c6d4 f3d3 e8c8 c1g5 e7e4 d3e4 f6e4 g5d8 c8d8 c4d5 e4f6 d5b7 d8c7 b7a6 c7b6 a6d3 d6d5 f1d1 d4c6 a3c2 c5c4 d3e2 f8e7 e2f3 h8d8 g1f1 c6b4 c2b4 e7b4 a1c1 b6c5');
INSERT INTO evaluations VALUES(52980,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(52980,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(52980,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(52980,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(52980,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(52980,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(52980,36,'cp 34','g1f3 d6d5 e4e5 c8f5 d2d4 e7e6 f1e2 h7h5 e1g1 g8h6 h2h3 c6c5 c1e3 b8c6 b1d2 d8b6 c2c4 c5d4 c4d5 e6d5 f3d4 c6d4 d1a4 f5d7 a4d4 h6f5 d4b6 a7b6 e3f2 a8a4 d2b3 a4f4 a1c1 g7g5 c1c7 f8g7 c7b7 g5g4 f2b6 f4f1 e2f1 g7e5');
INSERT INTO evaluations VALUES(52979,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(52979,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(52979,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(52979,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(52979,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(52979,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52979,36,'cp 27','d6d5 d2d3 d5e4 d3e4 d8d1 e1d1 c8g4 d1e2 g8f6 e4e5 f6d5 h2h3 g4f3 e2f3 b8d7 h3h4 e7e6 g2g3 h7h5 h1h2 a7a5 a2a4 f7f5 b1d2 d7c5 d2c4 f8e7 h2e2 b7b6 c1e3 e8c8 e3c5 e7c5 a1d1 c8c7 e2d2 g7g6 d2d3 d5b4 d3d8 h8d8 d1d8');
INSERT INTO evaluations VALUES(53511,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(53511,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53511,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(53511,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(53511,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(53511,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(53511,36,'cp 31','b8c6 h2h3 g8e7 f3h2 g7g6 e1g1 f8g7 h2g4 h6h5 g4h2 e8g8 h2f3 c6d4 f3g5 e7c6 a2a3 d4e6 h3h4 e6g5 h4g5 a8b8 a3a4 c6e7 c3d5 e7d5 c4d5 c8d7 c2c3 d7c6 d5a2 d6d5 c1e3 d5d4 c3d4 c5d4 e3d2 a7a5 b2b4');
INSERT INTO evaluations VALUES(55988,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(55988,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(55988,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55988,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(55988,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(55988,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55988,36,'cp 0','d1e2 b8c6 a2a3 a7a5 g1f3 f7f5 e5f6 g8f6 c1g5 f8d6 b1d2 e6e5 g2g3 e8g8 f1g2 c8g4 h2h3 g4e6 f3h4 d8d7 g5f6 g7f6 d2e4 g8h8 e2h5 d7g7 h4f5 g7d7');
INSERT INTO evaluations VALUES(56775,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(56775,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(56775,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(56775,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(56775,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(56775,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(56775,36,'cp -62','d8h4 e1e2 c7c5 g1f3 h4d8 e2f2 b8c6 c2c3 g8h6 c1h6 g7h6 f1b5 d8b6 a2a4 f8e7 h1f1 e8g8 f2g1 a7a6 b5c6 b6b2 b1d2 b2c3 c6d5 e6d5 d4c5 e7c5 g1h1 c3d3 a1b1 g8h8 d2b3 d3d1 f1d1 c5b4 b3d4 a6a5 d1c1 c8d7 c1c7 d7a4 e5e6 a4e8 c7b7 e8g6');
INSERT INTO evaluations VALUES(622,30,'cp -1162',NULL);
INSERT INTO evaluations VALUES(622,31,'cp -1177',NULL);
INSERT INTO evaluations VALUES(622,32,'cp -1198',NULL);
INSERT INTO evaluations VALUES(622,33,'cp -1225',NULL);
INSERT INTO evaluations VALUES(622,34,'cp -1247',NULL);
INSERT INTO evaluations VALUES(622,35,'cp -1254',NULL);
INSERT INTO evaluations VALUES(622,36,'cp -1281','c6g2 h1f1 c8g4 f2f3 g4f3 d1f3 d4f3 e1d1 f6g4 h8f7 g4h2 c2c3 h2f1 b2b3 g2h1 d1c2 b7b5 f7d6 c7d6 c4b5 e8f7 b5c4 f7f6 c4a6 f3e1 c2b2 d6d5 a2a4 h1e4 b2a2 a8b8 b3b4 e4c2 c1b2 c2a4 b1a3 f1d2 a1e1 a4a6 e1d1 a6e2 d1h1 d2e4');
INSERT INTO evaluations VALUES(623,30,'cp -1281',NULL);
INSERT INTO evaluations VALUES(623,31,'cp -1274',NULL);
INSERT INTO evaluations VALUES(623,32,'cp -1275',NULL);
INSERT INTO evaluations VALUES(623,33,'cp -1311',NULL);
INSERT INTO evaluations VALUES(623,34,'cp -1315',NULL);
INSERT INTO evaluations VALUES(623,35,'cp -1336',NULL);
INSERT INTO evaluations VALUES(623,36,'cp -1345','h1f1 c8g4 f2f3 g4f3 d1f3 d4f3 e1d1 f6g4 h8f7 g4h2 b2b3 h2f1 c2c3 g2h1 d1c2 b7b5 f7d6 c7d6 c4b5 e8f7 c2b2 f3e1 b5a6 f1g3 b1a3 h1c6 a6b5 c6e4 a1b1 a7a6 b5a4 e1d3 b2a1 e4e1 a4c6 a8b8 a3c2 e1f2 c6d5 f7g6 d5c4 d3c1 b1c1 f2d2 a1b2 g3e4 c1g1 g6f6 g1f1 f6e7 f1f7 e7e8 f7f3');
INSERT INTO evaluations VALUES(624,30,'cp -1364',NULL);
INSERT INTO evaluations VALUES(624,31,'cp -1353',NULL);
INSERT INTO evaluations VALUES(624,32,'cp -1361',NULL);
INSERT INTO evaluations VALUES(624,33,'cp -1348',NULL);
INSERT INTO evaluations VALUES(624,34,'cp -1382',NULL);
INSERT INTO evaluations VALUES(624,35,'cp -1388',NULL);
INSERT INTO evaluations VALUES(624,36,'cp -1413','c8g4 d1e2 g4e2 c4e2 d4e2 e1e2 g2e4 e2d1 e4f3 d1e1 f6d5 d2d3 d5b4 c1e3 b4c2 e1d2 c2e3 b1c3 e3f1 a1f1 e8c8 f1d1 f3f2 d2c1 f8c5 c1b1 d8h8 c3e4 f2e2 e4c3 e2h2 d1e1 h2g3 e1d1 h8d8 d1d2 c5e3 d2d1 g3g2 c3b5 g2f2 b5c3 d8d4 c3e4 f2e2 e4c3 d4d3 d1h1 e2g2 h1h7 d3d2 h7h8 c8d7 a2a4 d2b2 b1a1');
INSERT INTO evaluations VALUES(1622,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(1622,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(1622,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(1622,33,'cp 208',NULL);
INSERT INTO evaluations VALUES(1622,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(1622,35,'cp 199',NULL);
INSERT INTO evaluations VALUES(1622,36,'cp 211','d2d4 e5a5 d4d5 c6d8 c1d2 g8f6 e2f4 a5b6 d1e2 e7e6 e1c1 f8d6 d5e6 c8e6 g2g4 e8g8 g4g5 d6f4 d2f4 f6e8 h2h4 e8d6 f4d6 c7d6 f1g2 a8c8 c3d5 e6d5 g2d5 d8c6 e2d2 g8h8 a2a3 c6e5 c1b1 e5c4 d5c4 c8c4');
INSERT INTO evaluations VALUES(1618,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(1618,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(1618,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(1618,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(1618,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(1618,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(1618,36,'cp 217','e5a5 d4d5 c6d8 c1d2 c7c6 e2f4 a5c7 f1c4 g7g6 d1e2 f8g7 e1c1 g7c3 d2c3 c7f4 c1b1 c6d5 c4b5 e8f8 c3h8 f7f6 d1d5 d8f7 h1d1 c8f5 d5d4 f4e5 g2g4 e5e2 b5e2 e7e5 d4b4 f5e6 b4b7 e5e4 b1c1 g6g5 h2h3 f7h8 d1d6 a8e8 b7a7 h8g6 e2b5 e8e7 d6a6 g6f4 a7e7 g8e7 a6a8 f8g7');
INSERT INTO evaluations VALUES(3285,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(3285,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(3285,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(3285,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(3285,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(3285,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(3285,36,'cp 12','b1d2 d8e7 d4c5 d6c5 e3e4 f6h5 d1e2 f7f5 e4d5 e6d5 e2e7 c5e7 g3c7 c8d7 e1g1 a8c8 c7e5 c6e5 f3e5 d7a4 f1e1 e7d6 g2g3 g7g6 e5f3 f8e8 f3d4 a7a6 d3c2 e8e1 a1e1 a4d7 f2f4 g8f8 g1g2 h5f6 d2f3 c8e8 f3e5');
INSERT INTO evaluations VALUES(4769,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(4769,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(4769,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(4769,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(4769,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(4769,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(4769,36,'cp 100','d2d4 e5d4 c3d4 c5a7 b1c3 b7b5 a4b3 c8g4 c1e3 c6a5 h2h3 g4f3 g2f3 g8e7 h3h4 e8g8 h4h5 a5b3 a2b3 f7f5 d1d2 d8d7 h1g1 d7e6 d4d5 a7e3 d2e3 e6f6 e1c1 f5e4 h5h6 g7g6 c3e4 f6f3 e3d4');
INSERT INTO evaluations VALUES(4767,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(4767,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(4767,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(4767,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(4767,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(4767,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(4767,36,'cp 118','e5d4 c3d4 c5a7 b1c3 b7b5 a4b3 c6a5 c1g5 d8d7 b3c2 g8e7 e1g1 e8g8 f1e1 c8b7 a2a4 b5b4 g5e7 d7e7 c3d5 b7d5 e4d5 e7f6 d1d2 b4b3 c2d3 a7b6 a1c1 a8e8 e1e8 f8e8 h2h4 g7g6 d3a6 e8e4 h4h5 g6h5 c1e1 e4e1 d2e1 f6f5 e1e2 f5d5');
INSERT INTO evaluations VALUES(4774,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(4774,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(4774,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(4774,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(4774,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(4774,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(4774,36,'cp 118','c3d4 c5a7 b1c3 b7b5 a4b3 c6a5 c1g5 f7f6 g5e3 g8e7 e1g1 a5b3 d1b3 d6d5 e4d5 c8b7 a2a4 a7b6 e3f4 b7d5 c3d5 d8d5 b3a3 e8f7 a4a5 b6a7 f4c7 a8c8 a1c1 h8e8 c7b6 a7b8 b6c5 e7f5 a3d3 e8e4 g2g3 c8e8 f1e1 e4e1 c1e1 e8e1 f3e1 b8d6 c5d6 f5d6');
INSERT INTO evaluations VALUES(4773,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(4773,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(4773,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(4773,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(4773,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(4773,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(4773,36,'cp 106','c5a7 b1c3 b7b5 a4b3 h7h6 e1g1 g8e7 d4d5 c6b8 f1e1 e8g8 h2h3 c7c5 e4e5 c5c4 b3c2 b5b4 c3e2 d6e5 f3e5 e7d5 c2e4 c8e6 e2f4 d5f4 c1f4 b8d7 e4a8 d7e5 f4e5 d8a8 a2a3 b4b3 d1e2 a8c6 a1d1 a7c5 e5c3 f8c8 d1d2 c5f8 e2e5');
INSERT INTO evaluations VALUES(6629,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(6629,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(6629,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(6629,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6629,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6629,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6629,36,'cp 0','b4c3 b2c3 d7d6 g1f3 f7f5 d2d4 g8f6 d4e5 d6e5 d1d8 e8d8 a1b1 h8e8 e1g1 d8e7 f3g5 e8d8 g2c6 b7c6 g5f3 e7e6 f3g5');
INSERT INTO evaluations VALUES(6806,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(6806,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(6806,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(6806,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(6806,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(6806,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(6806,36,'cp 189','e4e5 d8e7 f2f4 d7d6 e1g1 d6e5 f4e5 e7e5 c4f7 e8e7 d4e5 f6e5 f7b3 c8e6 f1e1 e6b3 e1e5 b3e6 b1c3 g8f6 c3b5 a8d8 c1e3 e7f7 b5c7 e6c4 e5c5 c4d5 c7d5 f6d5 e3d4 d8d7 a1f1 f7g6 f1e1 h8d8 c5c4 b7b5 c4c6 g6f7 d4c5 a7a5 a2a3 a5a4 g1f2 h7h5 h2h3 d5e7 c5e7 d7e7 e1e7 f7e7 f2e3 d8d1 c6b6');
INSERT INTO evaluations VALUES(6805,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(6805,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(6805,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(6805,33,'cp 170',NULL);
INSERT INTO evaluations VALUES(6805,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(6805,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(6805,36,'cp 159','d8e7 f2f4 d7d6 e1g1 d6e5 f4e5 e7e5 c4f7 e8e7 d4e5 f6e5 f7b3 c8e6 f1e1 e6b3 e1e5 b3e6 b1c3 g8f6 c3b5 a8d8 c1g5 e7f7 b5c7 e6g4 a1f1 d8d1 f1d1 g4d1 c2c4 h8d8 c7d5 d8d7 g5f6 g7f6 e5f5 d7d6 d5c3 d1c2 f5f2 c2g6 h2h3 f7e6 f2e2 e6f7 c3d5 b7b5 e2e7 f7f8 e7a7 b5c4 d5e3');
INSERT INTO evaluations VALUES(403,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(403,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(403,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(403,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(403,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(403,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(403,36,'cp 189','f2f4 d7d6 e1g1 d6e5 f4e5 e7e5 c4f7 e8e7 d4e5 f6e5 f7b3 c8e6 f1e1 e6b3 e1e5 b3e6 b1c3 g8f6 c3b5 a8d8 c1f4 d8d5 e5e6 e7e6 b5c7 e6f5 c7d5 f6d5 f4g3 d5f6 a1f1 f5g6 f1d1 b7b5 c2c3 h8e8 g1f1 e8e4 g3f2 a7a5 d1d4 g6f5 d4d6 b5b4 f2d4 b4c3 b2c3 e4h4 h2h3 h4e4');
INSERT INTO evaluations VALUES(6887,30,'cp 280',NULL);
INSERT INTO evaluations VALUES(6887,31,'cp 266',NULL);
INSERT INTO evaluations VALUES(6887,32,'cp 254',NULL);
INSERT INTO evaluations VALUES(6887,33,'cp 259',NULL);
INSERT INTO evaluations VALUES(6887,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(6887,35,'cp 272',NULL);
INSERT INTO evaluations VALUES(6887,36,'cp 264','d4e5 g8e7 c1d2 d7d5 e5d6 d8d6 a2a3 b4c5 d1f3 c8d7 e1c1 d6f6 d2f4 e8c8 f3g3 h7h5 h2h4 d7c6 d1d8 h8d8 f1c4 d8d4 c4d3 d4d7 f4e3 c5e3 g3e3 c8b8 g2g3 a7a6 h1e1 e7c8 d3c4 d7e7 c4d5 c8b6 d5b3 e7e5 f2f4');
INSERT INTO evaluations VALUES(7622,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(7622,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(7622,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(7622,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(7622,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(7622,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(7622,36,'cp 86','d2c3 e8g8 f1d3 f8e8 f2f3 f6d5 c3d2 d8h4 g2g3 h4h3 d1e2 a7a5 e1c1 d5b4 a2a4 d7d6 d2b4 a5b4 b2b3 h3h5 c1b1 c8e6 h1e1 g7g6 e2f2 h5c5 f2d2 g8g7 f3f4 c5c3 d2c3 b4c3 e1e3 e8b8 b1c1 c6c5 f4f5 g6f5 e4f5 e6d5 f5f6 g7f6 d3f1 c7c6 e3c3 f6e7 d1e1 e7d7 f1d3 d5g2 d3h7');
INSERT INTO evaluations VALUES(7619,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(7619,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(7619,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(7619,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(7619,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(7619,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(7619,36,'cp 57','e8g8 f1d3 f8e8 f2f3 f6d5 c3d2 d8h4 g2g3 h4h5 f3f4 h5h3 d1f3 a7a5 a2a3 d5f6 e1c1 d7d5 e4e5 c8g4 f3f1 f6d7 f1h3 g4h3 d1e1 h7h5 d3e2 g7g6 e2f3 a5a4 e1e3 h3e6 h2h3 c6c5 g3g4 d5d4 f3a8');
INSERT INTO evaluations VALUES(11344,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(11344,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(11344,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(11344,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(11344,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(11344,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(11344,36,'cp 52','d3d4 e5d4 f3d4 c6d4 e3d4 f8e7 d1d3 e8g8 e1c1 b7b5 c4d5 f6d5 c3d5 c7c5 d4e3 c8b7 f2f3 b7d5 d3d5 f8e8 g2g4 e7f8 h2h4 d8f6 d5f5 f6e6 c1b1 a6a5 h4h5 a5a4 f5e6 f7e6 g4g5 d6d5 h1g1 h6g5 g1g5 f8e7 g5g6 d5d4 e3d2');
INSERT INTO evaluations VALUES(12152,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(12152,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(12152,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(12152,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(12152,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(12152,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(12152,36,'cp 40','g1f3 f6h5 f4g5 f8e7 g5e7 d8e7 f1d3 e8g8 d1c2 h5f6 c4d5 e6d5 e1g1 f8e8 a1e1 h7h6 f3d2 c6c5 d4c5 d7c5 e1c1 b7b6 d2f3 c8b7 f3d4 a8c8 c3e2 f6e4 c2d1 c5d3 c1c8 b7c8 d1d3 e4c5 d3a3 c8a6 f1d1 e8c8 d4f5 e7f6');
INSERT INTO evaluations VALUES(13053,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13053,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13053,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(13053,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(13053,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13053,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13053,36,'cp 0','b7c6 b1c3 e7e6 e1g1 d6d5 c4d3 f8e7 f1e1 e8g8 d1e2 f8e8 c1g5 h7h6 g5f4 f6d7 e2g4 d7f6');
INSERT INTO evaluations VALUES(13847,30,'cp 362',NULL);
INSERT INTO evaluations VALUES(13847,31,'cp 346',NULL);
INSERT INTO evaluations VALUES(13847,32,'cp 339',NULL);
INSERT INTO evaluations VALUES(13847,33,'cp 361',NULL);
INSERT INTO evaluations VALUES(13847,34,'cp 377',NULL);
INSERT INTO evaluations VALUES(13847,35,'cp 381',NULL);
INSERT INTO evaluations VALUES(13847,36,'cp 361','f3g5 c8e6 g5e6 f7e6 c4e6 c6d4 e6c4 d8d7 c3d5 f6d5 c4d5 d4e6 c2c3 e8c8 b2b4 e6c7 d2d3 c7d5 e4d5 c8b8 c3c4 b8a8 c1e3 g7g5 e3g5 d8e8 a1b1 c5b4 g5e3 h7h5 d1d2 b7b6 b1b4 f8e7 g1h1 h5h4 h2h3 h8f8 f1b1 f8g8');
INSERT INTO evaluations VALUES(13876,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(13876,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(13876,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(13876,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(13876,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(13876,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(13876,36,'cp 77','f1b5 c8e6 e1g1 f8e7 e4e5 d6e5 f3e5 d8b6 d1f3 a8c8 b2b3 e8g8 e5c6 b7c6 b5d3 f8e8 c1d2 e6g4 f3g3 c5c4 b3c4 g4h5 a1b1 b6c5 f1e1 e7d6 g3g5 e8e1 b1e1 c5g5 d2g5 d6b4 g5d2 c8d8 c3e4 b4d2 e4d2 g8f8 d2e4 f6e4 e1e4 h5g6 e4e3 g6d3 c2d3');
INSERT INTO evaluations VALUES(13830,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(13830,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(13830,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(13830,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(13830,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(13830,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(13830,36,'cp 67','c8e6 e1g1 f8e7 e4e5 d6e5 f3e5 d8b6 d1f3 a8c8 b2b3 e8g8 e5c6 b7c6 b5d3 f8e8 c1d2 e6g4 f3g3 g4h5 f1e1 h5g6 c3a4 b6b7 d3g6 h7g6 c2c4 b7d7 a1d1 c8d8 g3f3 e7f8 e1e8 d7e8 d1e1 e8d7 d2a5 d8e8 e1e8 f6e8 g2g3 e8c7 a5c7 d7c7 g1g2 c7d6 a4b2 d6d2 b2d3 d2c3 h2h3 f8d6');
INSERT INTO evaluations VALUES(14241,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(14241,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(14241,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(14241,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(14241,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(14241,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(14241,36,'cp -44','c8f5 e2e3 e7e6 f1d3 c6b4 d3f5 e6f5 c1d2 b4c6 e1g1 f8b4 c3b1 e8g8 d2b4 c6b4 f1e1 c7c6 c2c3 b4a6 b1d2 a6c7 a2a4 c7e8 c3c4 d5c4 d2c4 a7a5 d1b3 e8d6 a1c1 d6c4 c1c4 d8e7 f3e5 f8e8 h2h3 f6e4 g1h2 g7g6 b3c2 e7e6');
INSERT INTO evaluations VALUES(24841,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(24841,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(24841,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(24841,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(24841,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(24841,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(24841,36,'cp -34','e2e3 c6b4 f1d3 e7e6 d3f5 e6f5 a2a3 b4c6 d1d3 d8c8 d3b5 a7a6 b5b3 c8d8 c1d2 b7b5 b3a2 f8e7 f3e5 d8d6 h2h3 h7h5 e1g1 d6e6 b2b4 e8g8 d2e1 f8d8 a2b3 c6e5 f4e5 f6e4 c3e4 f5e4 e1g3 h5h4 g3f4 d8f8 c2c3 c7c6 b3c2 f7f6 a3a4 f6e5 f4e5 e7g5');
INSERT INTO evaluations VALUES(14386,30,'cp -304',NULL);
INSERT INTO evaluations VALUES(14386,31,'cp -326',NULL);
INSERT INTO evaluations VALUES(14386,32,'cp -317',NULL);
INSERT INTO evaluations VALUES(14386,33,'cp -313',NULL);
INSERT INTO evaluations VALUES(14386,34,'cp -333',NULL);
INSERT INTO evaluations VALUES(14386,35,'cp -325',NULL);
INSERT INTO evaluations VALUES(14386,36,'cp -326','c6d4 e3d4 d8d4 f1b5 f6d7 f3e2 d4h4 g2g3 h4h6 g1f3 e7e6 h1f1 f8d6 g3g4 e8g8 g4g5 h6h5 e1c1 a7a6 b5d3 d7c5 h2h4 b7b5 e2e3 c8b7 c3e4 c5d3 c2d3 h5g4 f3d2 g4h4 c1b1 a8e8 a2a3 e8d8 d1e1 b7d5 f1h1 h4g4 e1g1');
INSERT INTO evaluations VALUES(17553,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(17553,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(17553,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(17553,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(17553,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(17553,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(17553,36,'cp -28','b7b5 a4b3 c8g4 c2c3 e7e6 h2h3 g4h5 d1e2 g8f6 c1g5 h7h6 g5f6 h5f3 f6d8 f3e2 e1e2 e8d8 b1d2 f8d6 e2d3 d8e7 a2a4 b5b4 h1e1 b4c3 b2c3 e7f6 a1b1 g7g6 b3c2 c6a5 g2g3 h8c8 c2b3 a8b8 b3c2 f6g7 b1b8');
INSERT INTO evaluations VALUES(18194,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(18194,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(18194,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(18194,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(18194,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(18194,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(18194,36,'cp -25','a4b3 c8g4 c2c3 e7e6 h2h3 g4h5 d1e2 g8f6 c1g5 f8e7 b1d2 e8g8 e2e3 h5g6 e1g1 h7h6 g5f4 d8b6 f1c1 f8c8 b3c2 g6c2 c1c2 a6a5 a1c1 e7f8 f3e5 c6e5 e3e5 f6e8 e5e2 e8d6 d2f3 d6c4 h3h4 b6b7 a2a3 f8e7 h4h5 a8a6 g2g4 b5b4 c3b4');
INSERT INTO evaluations VALUES(18193,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(18193,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(18193,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(18193,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(18193,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(18193,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(18193,36,'cp -22','c8g4 c2c3 e7e6 h2h3 g4h5 d1e2 g8f6 c1g5 f8e7 b1d2 e8g8 e2e3 h7h6 g5f4 h5g6 e1g1 d8b6 f1c1 f8c8 b3c2 g6c2 c1c2 a6a5 a1c1 e7f8 f3e5 c6e5 f4e5 f6e8 h3h4 e8d6 e3e2 c8c6 g2g3 a8c8 e5d6 f8d6');
INSERT INTO evaluations VALUES(17805,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(17805,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(17805,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(17805,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(17805,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(17805,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(17805,36,'cp -17','b7b5 a4b3 c6a5 e1g1 g8f6 c1d2 a5b3 a2b3 f8e7 c2c4 c8g4 b1c3 c7c6 h2h3 g4h5 c3e2 h5f3 g2f3 f6h5 e2g3 h5g3 f2g3 e8g8 g1h2 a6a5 d1e1 b5b4 e1e2 d6d5 d2e3 d8d7 f3f4 e5f4 g3f4 d5e4 d3e4 f8d8 e3b6 d7d3 e2g4');
INSERT INTO evaluations VALUES(17985,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(17985,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(17985,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(17985,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(17985,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(17985,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(17985,36,'cp -9','a4b3 c6a5 c1d2 a5b3 a2b3 g8f6 c2c4 c8d7 e1g1 f8e7 d2a5 e8g8 f1e1 d8c8 b1c3 c8b7 h2h3 b5b4 c3d5 f6d5 e4d5 f7f5 d1d2 c7c5 d3d4 e5e4 d4c5 d6c5 f3e5 e7d6 f2f4 e4f3 e5f3 a8e8 g1h1 h7h6 a1d1 e8e1 d1e1 f8e8 e1e8 d7e8 d2e1 e8f7');
INSERT INTO evaluations VALUES(17975,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(17975,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(17975,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(17975,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(17975,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(17975,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17975,36,'cp -13','c6a5 c1d2 a5b3 a2b3 f8e7 c2c4 c8d7 d2a5 g8f6 e1g1 e8g8 f1e1 d8b8 d3d4 e5d4 d1d4 f8e8 e4e5 d6e5 f3e5 c7c5 d4d3 b5c4 b3c4 b8b2 b1c3 b2b7 h2h3 b7c8 e5d7 c8d7 d3d7 f6d7 c3d5 e7d6 e1e8 a8e8 a1d1 d7e5 d5f6 g7f6 d1d6 e5c4 d6a6 c4a5 a6a5 e8e1 g1h2 g8g7 a5c5');
INSERT INTO evaluations VALUES(21043,30,'cp -217',NULL);
INSERT INTO evaluations VALUES(21043,31,'cp -233',NULL);
INSERT INTO evaluations VALUES(21043,32,'cp -223',NULL);
INSERT INTO evaluations VALUES(21043,33,'cp -226',NULL);
INSERT INTO evaluations VALUES(21043,34,'cp -224',NULL);
INSERT INTO evaluations VALUES(21043,35,'cp -231',NULL);
INSERT INTO evaluations VALUES(21043,36,'cp -225','d5e4 d4d5 c6b4 f3g5 d8d5 d1d5 b4d5 f1c4 g8f6 b1c3 c7c6 c3e4 c8f5 c4d3 f5g6 c1d2 h7h6 e4f6 e7f6 d3g6 f7g6 g5e4 e8f7 e1c1 f6f5 e4c3 d5f6 d2e3 f8b4 e3d4 h8e8 g2g3 b7b6 a2a3 b4e7 a3a4 g6g5 f2f3 h6h5 h1e1 e7b4 e1e8 a8e8 d4f6 f7f6 d1d7 e8e1 c1d2 e1b1 d7a7 b1b2');
INSERT INTO evaluations VALUES(21137,30,'cp -214',NULL);
INSERT INTO evaluations VALUES(21137,31,'cp -222',NULL);
INSERT INTO evaluations VALUES(21137,32,'cp -208',NULL);
INSERT INTO evaluations VALUES(21137,33,'cp -213',NULL);
INSERT INTO evaluations VALUES(21137,34,'cp -206',NULL);
INSERT INTO evaluations VALUES(21137,35,'cp -212',NULL);
INSERT INTO evaluations VALUES(21137,36,'cp -206','d4d5 c6b4 f3g5 d8d5 d1d5 b4d5 f1c4 g8f6 b1c3 c7c6 g5e4 d5c3 e4c3 c8f5 c4b3 e7e6 f2f3 f8c5 c1d2 e8c8 e1c1 h7h5 a2a4 c5b6 h2h4 a7a6 d2g5 d8d1 h1d1 b6c7 d1e1 f5g6 c3e2 f6d5 g2g3 h8e8 g5d2 a6a5 b3c4 g6h7');
INSERT INTO evaluations VALUES(21342,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(21342,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(21342,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(21342,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(21342,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(21342,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(21342,36,'cp 9','g8f6 g1f3 f8d6 f4g3 e8g8 b1d2 c6e7 b5d3 b7b6 e3e4 f6e4 d2e4 d5e4 d3e4 a8b8 c2c3 c8a6 d1d2 a6b7 e4b7 b8b7 e1g1 d8d7 d2e2 h7h6 f1e1 b7b8 e2a6 b8a8 a2a4 f8d8 a4a5 b6b5 f3e5 d6e5 e1e5 e7d5 h2h4 d8c8 h4h5 c8d8');
INSERT INTO evaluations VALUES(21534,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(21534,31,'cp -137',NULL);
INSERT INTO evaluations VALUES(21534,32,'cp -177',NULL);
INSERT INTO evaluations VALUES(21534,33,'cp -156',NULL);
INSERT INTO evaluations VALUES(21534,34,'cp -179',NULL);
INSERT INTO evaluations VALUES(21534,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(21534,36,'cp -172','d5e4 f3e5 d8g5 d1a4 g5g2 b5c6 b7c6 a4c6 e8d8 h1f1 c8h3 c6a8 d8e7 e1d1 g2f1 d1c2 g8h6 a8c6 f7f6 c6c7 e7e6 d2d4 f6e5 c7e5 e6f7 c1h6 f1d3 c2c1 g7h6 e5h5 f7g7 h5e5 g7g6 e5h8 d3f1 c1c2 f1f2 b1d2 e4e3 a1g1 g6h5 g1d1 e3d2 h8e5 h3f5 c2b3 h5g6 e5e8 g6g5 e8e5 h6h5 h2h3 g5h4 d1d2 f2d2 e5f5 f8h6 d4d5 d2f4 f5e6 h6g5 b3c2 f4c4 b2b3 c4f4');
INSERT INTO evaluations VALUES(21548,30,'cp -182',NULL);
INSERT INTO evaluations VALUES(21548,31,'cp -212',NULL);
INSERT INTO evaluations VALUES(21548,32,'cp -191',NULL);
INSERT INTO evaluations VALUES(21548,33,'cp -199',NULL);
INSERT INTO evaluations VALUES(21548,34,'cp -212',NULL);
INSERT INTO evaluations VALUES(21548,35,'cp -190',NULL);
INSERT INTO evaluations VALUES(21548,36,'cp -210','f3e5 d8g5 d1a4 g5g2 b5c6 b7c6 a4c6 e8d8 e1d1 f8d6 h1e1 g8f6 d2d4 d6e5 c1e3 a8b8 d4e5 f6g4 c6d5 d8e8 d5c5 c8f5 d1c1 g2h2 e3g5 h2e5 c5a7 e5b5 b2b4 b5b6 a7b6 b8b6 b1d2 g4f2 e1f1 b6g6 g5e3 f2d3 c1c2 f5h3 f1h1 g6g3 e3d4 h3f5 h1f1 f5g6 d2c4');
INSERT INTO evaluations VALUES(23956,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(23956,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(23956,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(23956,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(23956,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(23956,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(23956,36,'cp 96','d6d5 e4d5 f6d5 b1c3 d5c3 b2c3 e8g8 a4c2 f8e8 d4d5 c6a5 f3e5 e7d6 c1f4 g7g6 d1d4 f7f6 e5f3 e8e1 a1e1 d8f8 f3d2 d6c5 d4d3 d7f5 d2e4 c5d6 f4d6 c7d6 g2g4 f5e4 d3e4 b7b5 h2h4 a5c4 h4h5 f8f7 h5g6 h7g6 f2f4 f6f5 g4f5 g6f5 e4e6 f7e6');
INSERT INTO evaluations VALUES(25367,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(25367,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(25367,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(25367,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(25367,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25367,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(25367,36,'cp 12','f7e6 e1g1 d8d7 f1e1 h7h6 d3d4 e5d4 f3d4 f8e7 c3e2 e6e5 d4c6 d7c6 e2g3 e8g8 d1d3 a7a5 g3f5 f8f7 c2c4 a5a4 f2f3 f6d7 f5e7 f7e7 c1e3 c6a6 h2h3 e7f7 a1d1 b7b6 d1d2 d7c5 e3c5 b6c5 e1d1 a6b6 g1h1');
INSERT INTO evaluations VALUES(25960,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(25960,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(25960,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(25960,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(25960,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(25960,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(25960,36,'cp -64','a4f4 g8e7 f1d3 e7g6 d3g6 h7g6 e1g1 a8c8 f1e1 f7f6 f4g4 e8f7 c2c4 d5c4 a2a3 h8h5 b1c3 c6e5 e1e5 f6e5 c1g5 d8c7 a1e1 f8d6 f3h4 h5h4 g5h4 d7c6 g4c4 c6d5 c4e2 c7c6 f2f3 d6a3 b2a3 c6c3 e2e5 c3e5 e1e5 c8c4 h4g3');
INSERT INTO evaluations VALUES(26683,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(26683,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(26683,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(26683,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(26683,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(26683,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(26683,36,'cp 26','e7e6 b1c3 f8b4 d1c2 b7b5 a2a3 b4a5 d4d5 e6d5 e1c1 a5c3 d1d5 d8f6 c2c3 f6c3 b2c3 a7a6 d5g5 g4d7 f3d4 g8f6 d4c6 d7c6 f2f3 c6d7 g5g7 h7h6 e3d4 e8e7 g2g4 d7e6 g4g5 f6h5 g5h6 h5g7 d4g7 h8h7 h2h4 a8h8 c1d2 h8d8 d2c2 d8h8 h4h5 h7h6 g7h8 h6h8 c2d2 a6a5 h5h6 f7f5 h6h7 e7f7 d2e3 f7g7');
INSERT INTO evaluations VALUES(27101,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(27101,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(27101,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(27101,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(27101,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(27101,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(27101,36,'cp 82','c7c6 c1e3 g8f6 e1c1 f6g4 c4e6 f7e6 f2f4 d8e7 h1e1 e7f7 h2h3 g4e3 d4e3 f8e7 c1b1 e8g8 c3e2 d6d5 g2g4 a7a5 a2a3 a5a4 e1f1 b7b5 e4d5 e6d5 e2d4 e7c5 f4f5 a8e8 e3c3 c5d4 d1d4 f7d7 d4d1 d7d6 g4g5 e8e5 c3g3 d5d4 h3h4 c6c5 f5f6 d6d5 f6f7 f8f7 f1f7 g8f7');
INSERT INTO evaluations VALUES(28324,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(28324,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(28324,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(28324,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(28324,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(28324,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(28324,36,'cp 118','d1d2 d7d6 e1c1 b8d7 h1e1 a7a6 e3h6 g7h6 d2h6 e7g8 h6f4 d8f6 f4e3 f6g7 e4e5 d6d5 c3e2 e8c8 c2c3 g8e7 g2g4 c8b8 c1b1 h7h6 f3d2 c7c5 f2f4 c5c4 d3c2 h6h5 h2h3 h5g4 h3g4 b8a7');
INSERT INTO evaluations VALUES(28325,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(28325,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(28325,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(28325,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(28325,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(28325,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(28325,36,'cp 127','d7d6 e1c1 b8d7 d4d5 e6d5 e3h6 e8g8 e4d5 d7f6 h6g7 g8g7 h2h4 e7d5 h4h5 d5c3 h5g6 f7g6 d2h6 g7g8 b2c3 d8e7 f3g5 e7g7 h6g7 g8g7 g5e6 g7f7 e6f8 a8f8 f2f3 f7g7 d1e1 f8f7 c1b2 h7h5 d3c4 b7d5 c4b3 b6b5 h1h4 d5b3 a2b3 g6g5 h4d4 g7g6 b2b1 a7a6 c3c4 f6d7 c4b5');
INSERT INTO evaluations VALUES(28718,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28718,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(28718,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28718,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28718,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28718,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28718,36,'cp 0','d7d5 e1g1 f8d6 b2b3 e8g8 b1c3 a7a6 c1b2 b8d7 a2a4 d6b4 c3a2 b4d6');
INSERT INTO evaluations VALUES(28753,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28753,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(28753,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(28753,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(28753,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28753,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28753,36,'cp 14','c7c5 c2c4 c5d4 d1d4 b8c6 d4f4 f8c5 b2b3 e8g8 b1c3 c6e7 c1b2 e7g6 f4d2 d8e7 e2e3 d7d5 c4d5 f8d8 h2h4 f6d5 c3d5 d8d5 d2e2 a8d8 f1d1 d5d1 a1d1 d8d1 e2d1 f7f6 h4h5 g6f8 d1e2 g8f7 f3e1 b7g2 g1g2 e7b7 e3e4 f8d7 e1d3 c5d6 e2g4');
INSERT INTO evaluations VALUES(29068,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29068,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(29068,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(29068,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(29068,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(29068,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(29068,36,'cp 12','g1f3 b8d7 f1d3 f5d3 d1d3 e7e6 e1g1 f8e7 c3e4 e8g8 c2c4 h7h6 f1d1 d8c7 c1e3 f6g4 e3d2 a8d8 h2h3 g4f6 d2e3 c6c5 e4f6 e7f6 d3e2 c5d4 e3d4 f6d4 d1d4 d7e5 d4d8 e5f3 e2f3 f8d8 b2b3 d8d2 g2g3 a7a6 f3e3 d2b2 a1d1 b2c2');
INSERT INTO evaluations VALUES(29067,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(29067,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(29067,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(29067,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29067,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29067,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29067,36,'cp 0','e7e6 f1d3 f5d3 d1d3 f8e7 c3e4 f6e4 d3e4 e8g8 e4e2 d8d5 a2a4 b8d7 e1g1 d5f5 f1d1 h7h6 c1e3 f8d8 e2d3 f5d3 d1d3 f7f6 h2h4 g7g5 g2g4 d7b6 g1f1 b6d5 f1e2 b7b6 e3d2 g8f7 h4h5 d5b4 d2b4 e7b4 c2c3 b4d6');
INSERT INTO evaluations VALUES(29625,30,'cp -152',NULL);
INSERT INTO evaluations VALUES(29625,31,'cp -155',NULL);
INSERT INTO evaluations VALUES(29625,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(29625,33,'cp -123',NULL);
INSERT INTO evaluations VALUES(29625,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(29625,35,'cp -154',NULL);
INSERT INTO evaluations VALUES(29625,36,'cp -141','c4d5 c7c6 h2h3 g4f5 d5c6 b8c6 d2d3 a8c8 g1f3 e7e5 e2e4 f5g6 c1e3 f8d6 e1g1 g6h5 b1d2 e8g8 d1a4 f8e8 f1c1 d6b8 a4b3 d8d7 a2a4 h7h6 c1c5 b8d6 c5c3 c8d8 a1c1 d6b8 b3b5 a7a6 b5b3 b8a7 e3a7');
INSERT INTO evaluations VALUES(30666,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30666,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(30666,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(30666,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30666,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30666,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30666,36,'cp 0','f8b4 g1e2 g8f6 e4e5 f6d5 e3d2 d5c3 e2c3 b4c3 b2c3 d8h4 d2e3 b8c6 f1d3 f7f6 g2g3 h4h3 d3f1 h3f5 f1d3');
INSERT INTO evaluations VALUES(27945,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(27945,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(27945,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(27945,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(27945,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(27945,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(27945,36,'cp 20','g1e2 g8f6 e4e5 f6d5 e3d2 d5c3 e2c3 b4c3 d2c3 d8h4 d1e2 b8c6 e1c1 e8c8 c1b1 d7d6 g2g3 h4h6 f1g2 d6e5 d4e5 d8d1 h1d1 h8d8 h2h4 d8d1 e2d1 h6g6 d1e2 c6e7 g2h3 a7a5 b1c1 h7h5 a2a3 b7e4 b2b3 g6h6 c3d2 h6g6 d2e1 e4d5 b3b4 a5b4 a3b4 c8b8 c1b2 g6e4 e2e4 d5e4');
INSERT INTO evaluations VALUES(32878,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(32878,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(32878,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(32878,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(32878,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(32878,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(32878,36,'cp 76','c3d5 e8d7 d5e3 f5g6 g1f3 b8c6 d2d4 e7e6 h2h4 h7h5 a2a3 f8d6 b2b4 a8d8 c2c4 d7c8 c1b2 g8f6 b4b5 c6a5 c4c5 d6f4 a3a4 h8e8 a1a3 c7c6 b2c3 f4c7 e1g1 f6d5 e3d5 e6d5 f3e5 c7e5 d4e5 a5c4');
INSERT INTO evaluations VALUES(35091,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(35091,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(35091,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(35091,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(35091,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(35091,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(35091,36,'cp 155','f6e5 f1d3 g8f6 g1f3 e5e7 b1c3 d7d6 c1g5 c7c6 d1d2 h7h6 g5f4 c8g4 f3d4 g7g5 f4e3 c6c5 d4e2 b8c6 h2h4 g5h4 f2f3 g4e6 e2f4 e8c8 e1c1 f8g7 d2f2 h4h3 g2h3 c6b4 a2a3 b4d3 d1d3 h8g8 d3d1 e7c7 h1g1 c7c6 c1b1 g7h8 g1g8 d8g8 f4e6');
INSERT INTO evaluations VALUES(35249,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(35249,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(35249,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(35249,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(35249,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(35249,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(35249,36,'cp -21','g8f6 b1c3 d7d6 g1e2 f8g7 e1g1 c7c6 a2a4 e8g8 d2d3 c8e6 c4e6 e7e6 f3g3 f6h5 g3h4 b8a6 f2f4 h5f4 e2f4 e5f4 c1f4 a6b4 h4f2 a8e8 f2d2 d6d5 a1e1 e6d7 h2h3 c6c5 a4a5 d5e4 c3e4 f7f5 e4d6 e8e1 f1e1');
INSERT INTO evaluations VALUES(34007,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(34007,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(34007,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(34007,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(34007,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(34007,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(34007,36,'cp -12','b1c3 f8g7 g1e2 e8g8 d2d3 d7d6 e1g1 c7c6 a2a4 c8e6 c4e6 e7e6 f3g3 b8a6 f2f4 e5f4 e2f4 e6e7 c1d2 a6b4 c3e2 a7a5 c2c3 b4a6 h2h3 a6c5 d2e3 f8e8 g1h1 f6d7 e3g1 c5b3 a1a3 d7c5 e2d4 b3d4 g1d4 g7d4 c3d4');
INSERT INTO evaluations VALUES(35531,30,'cp -317',NULL);
INSERT INTO evaluations VALUES(35531,31,'cp -316',NULL);
INSERT INTO evaluations VALUES(35531,32,'cp -337',NULL);
INSERT INTO evaluations VALUES(35531,33,'cp -341',NULL);
INSERT INTO evaluations VALUES(35531,34,'cp -288',NULL);
INSERT INTO evaluations VALUES(35531,35,'cp -303',NULL);
INSERT INTO evaluations VALUES(35531,36,'cp -316','f3e5 f7e8 d1g4 g8f6 g4f4 d7d6 e5f3 h8g8 e1g1 d8e7 b1c3 e7g7 f4g3 g7g3 h2g3 b8c6 d2d4 c8g4 c1g5 f6h5 c3d5 a8c8 g5h4 g4f3 g2f3 h5g3 h4g3 g8g3 g1f2 g3g7 c2c3 g7f7 f1h1 c6e7 d5e3 e7g6 a1g1 e8d7 g1g5 f8e7 g5h5 c8h8 e3f5 g6f4 h5h2 h7h5 h1g1 h8f8 f2e3 f4d5 e3e2');
INSERT INTO evaluations VALUES(35530,30,'cp -312',NULL);
INSERT INTO evaluations VALUES(35530,31,'cp -318',NULL);
INSERT INTO evaluations VALUES(35530,32,'cp -325',NULL);
INSERT INTO evaluations VALUES(35530,33,'cp -325',NULL);
INSERT INTO evaluations VALUES(35530,34,'cp -344',NULL);
INSERT INTO evaluations VALUES(35530,35,'cp -336',NULL);
INSERT INTO evaluations VALUES(35530,36,'cp -324','f7e8 d1g4 g8f6 g4f4 d7d6 e5f3 h8g8 e1g1 d8e7 d2d4 e7g7 f4g3 g7g3 h2g3 f6g4 b1c3 c8e6 c1g5 e6c4 f1e1 b8c6 a1d1 h7h6 g5h4 e8d7 b2b3 c4f7 e4e5 d6d5 c3e2 a8e8 e2f4 f8b4 e1e2 b7b5 a2a4 b5a4 b3a4 b4c3 d1d3 c3a5 e5e6 f7e6 f4e6 e8e6 e2e6 d7e6');
INSERT INTO evaluations VALUES(57546,30,'cp -333',NULL);
INSERT INTO evaluations VALUES(57546,31,'cp -334',NULL);
INSERT INTO evaluations VALUES(57546,32,'cp -324',NULL);
INSERT INTO evaluations VALUES(57546,33,'cp -317',NULL);
INSERT INTO evaluations VALUES(57546,34,'cp -325',NULL);
INSERT INTO evaluations VALUES(57546,35,'cp -299',NULL);
INSERT INTO evaluations VALUES(57546,36,'cp -300','d1g4 g8f6 g4f4 d7d6 e5f3 h8g8 e1g1 d8e7 b1c3 e7g7 f4g3 g7g3 h2g3 b8c6 d2d4 c8g4 c1g5 f6h5 c3d5 a8c8 g5h4 g4f3 g2f3 h5g3 h4g3 g8g3 g1f2 g3g7 c2c3 g7f7 f1h1 c6e7 d5e3 e7g6 e3f5 d6d5 f2e3 g6e7 f5h6 f8h6 h1h6 e8d7 a1h1 c8f8');
INSERT INTO evaluations VALUES(57545,30,'cp -336',NULL);
INSERT INTO evaluations VALUES(57545,31,'cp -337',NULL);
INSERT INTO evaluations VALUES(57545,32,'cp -343',NULL);
INSERT INTO evaluations VALUES(57545,33,'cp -344',NULL);
INSERT INTO evaluations VALUES(57545,34,'cp -336',NULL);
INSERT INTO evaluations VALUES(57545,35,'cp -331',NULL);
INSERT INTO evaluations VALUES(57545,36,'cp -348','g8f6 g4f4 d7d6 e5f3 h8g8 e1g1 d8e7 d2d4 e7g7 f4g5 b8c6 b1c3 c8g4 d4d5 g4f3 g5g7 f8g7 d5c6 f3h5 e4e5 f6g4 c6b7 a8b8 e5d6 c7d6 c1f4 g4e5 a1e1 e8d7 c3d5 g8e8 b2b3 b8b7 c2c4 h5g6 f4g3 e5d3 e1e8 d7e8 g3d6 g7d4 g1h1 d3f2 f1f2 d4f2');
INSERT INTO evaluations VALUES(50030,30,'cp -323',NULL);
INSERT INTO evaluations VALUES(50030,31,'cp -320',NULL);
INSERT INTO evaluations VALUES(50030,32,'cp -348',NULL);
INSERT INTO evaluations VALUES(50030,33,'cp -333',NULL);
INSERT INTO evaluations VALUES(50030,34,'cp -326',NULL);
INSERT INTO evaluations VALUES(50030,35,'cp -332',NULL);
INSERT INTO evaluations VALUES(50030,36,'cp -311','g4f4 d7d6 e5f3 h8g8 e1g1 d8e7 d2d3 e7g7 f4g5 b8c6 b1c3 c8e6 g1h1 f8e7 g5g7 g8g7 c3e2 f6h5 f3d4 e6g4 c1e3 c6b4 a1e1 c7c5 e3d2 a7a5 a2a3 g4e2 d4e2 b4c2 e1c1 c2d4 e2d4 c5d4 c1c7 b7b6 c7c6 e8d7 c6b6');
INSERT INTO evaluations VALUES(50029,30,'cp -329',NULL);
INSERT INTO evaluations VALUES(50029,31,'cp -325',NULL);
INSERT INTO evaluations VALUES(50029,32,'cp -340',NULL);
INSERT INTO evaluations VALUES(50029,33,'cp -326',NULL);
INSERT INTO evaluations VALUES(50029,34,'cp -330',NULL);
INSERT INTO evaluations VALUES(50029,35,'cp -328',NULL);
INSERT INTO evaluations VALUES(50029,36,'cp -336','d7d6 e5f3 h8g8 e1g1 d8e7 d2d3 b8c6 b1c3 e7g7 f4g5 c8e6 g1h1 f8e7 g5g7 g8g7 c3e2 f6h5 f3d4 e6g4 c1e3 g7g6 a1e1 c6d4 e2d4 e8d7 b2b4 a8g8 b4b5 g6f6 f1f6 e7f6 e1f1 a7a6 b5a6 b7a6 d4f5 g8g6 f1f2 g4f5 f2f5');
INSERT INTO evaluations VALUES(49149,30,'cp -328',NULL);
INSERT INTO evaluations VALUES(49149,31,'cp -325',NULL);
INSERT INTO evaluations VALUES(49149,32,'cp -336',NULL);
INSERT INTO evaluations VALUES(49149,33,'cp -330',NULL);
INSERT INTO evaluations VALUES(49149,34,'cp -329',NULL);
INSERT INTO evaluations VALUES(49149,35,'cp -344',NULL);
INSERT INTO evaluations VALUES(49149,36,'cp -322','e5f3 h8g8 e1g1 d8e7 d2d3 b8c6 b1c3 e7g7 f4g5 c8e6 g5g7 f8g7 f3g5 e8d7 g5e6 d7e6 c3e2 a8f8 c1g5 d6d5 g5f6 g7f6 e4d5 e6d5 c2c3 c6e5 d3d4 e5g4 e2f4 d5d6 f1f3 f6g5 a1f1 f8e8 f4h5 e8e2 h5g3 e2b2 g3e4 d6c6 e4g5 g8g5');
INSERT INTO evaluations VALUES(36513,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(36513,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(36513,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(36513,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36513,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36513,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(36513,36,'cp 0','c7c5 e1g1 b8c6 d2d3 a8b8 a2a3 a7a6 a1b1 b7b5 c4b5 a6b5 b2b4 c8g4 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 b1a1 a3b2 a1b1');
INSERT INTO evaluations VALUES(35936,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35936,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35936,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35936,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35936,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35936,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35936,36,'cp 0','e1g1 b8c6 a2a3 a7a6 d2d3 c8g4 a1b1 a8b8 b2b4 b7b5 c4b5 a6b5 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 b1a1 a3b2 a1b1');
INSERT INTO evaluations VALUES(35935,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35935,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35935,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35935,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35935,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35935,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35935,36,'cp 0','b8c6 a1b1 a8b8 a2a3 a7a6 b2b4 b7b5 c4b5 a6b5 d2d3 c8g4 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 b1a1 a3b2 a1b1');
INSERT INTO evaluations VALUES(3183,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3183,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3183,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3183,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3183,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3183,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3183,36,'cp 0','d2d3 a8b8 a1b1 a7a6 a2a3 c8g4 b2b4 b7b5 c4b5 a6b5 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 e4e5 f6e8 b1a1 a3b2 a1b1 b2a2 e5e6 f7e6 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(3188,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3188,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3188,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3188,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3188,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3188,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3188,36,'cp 0','a8b8 a1b1 a7a6 a2a3 c8g4 b2b4 b7b5 c4b5 a6b5 h2h3 g4e2 c3e2 d8b6 c1f4 f8c8 b4c5 b6c5 d3d4 c5a3 e4e5 f6e8 b1a1 a3b2 a1b1');
INSERT INTO evaluations VALUES(37024,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37024,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37024,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37024,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37024,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37024,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37024,36,'cp 0','c7c5 c4d5 c5d4 f3d4 d8d5 g5f6 g7f6 e2e3 b8c6 d4c6 d5c6 f1d3 f6f5 e1c1 b4c3 b2c3 e6e5 d3f5 c8f5 c2f5 c6c3 c1b1 a8e8 f5g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(37566,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(37566,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(37566,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(37566,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(37566,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(37566,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(37566,36,'cp 90','d2d4 d7d6 b1c3 e7e6 c1f4 g8e7 h2h4 h7h6 e1g1 e8g8 f1e1 b8d7 d1e2 a7a6 a1d1 c8b7 c4d3 b6b5 e2d2 g8h7 h4h5 g6g5 e4e5 e7f5 f3g5 h6g5 f4g5 f7f6 g5h4 b5b4 e5d6 b4c3 d2c3 h7h8 d6c7 d8c8');
INSERT INTO evaluations VALUES(40676,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(40676,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40676,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(40676,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(40676,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(40676,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40676,36,'cp 13','c7c5 a2a4 e8g8 d2d3 c8e6 f1e1 b8c6 c3d5 e6d5 e4d5 c6b4 f3h4 f6d5 h4f5 d5f6 c2c3 b4c6 d1f3 g8h7 a4a5 a7a6 f3g3 f6h5 g3g4 g7g6 c1h6 h5g7 f5e7 h7h6 e7c6 b7c6 g4f3 d8c7 c4a2 g7f5 d3d4 c5d4 a2b1 c7d7 b1f5 d7f5 f3c6 d4c3 b2c3');
INSERT INTO evaluations VALUES(40687,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40687,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(40687,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(40687,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40687,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40687,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40687,36,'cp 0','b8c6 f1e1 g7g5 d3d4 h8g8 d4e5 d6e5 d1d8 e7d8 c4b5 f6d7 b1d2 d8e7 d2f1 h6h5 b5c6 b7c6 f3d2 g5g4 h3g4 h5g4 d2c4 f7f6 c4a5 g4g3 f2g3 e7b4 c1d2 b4c5 d2e3 c5b4');
INSERT INTO evaluations VALUES(41525,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(41525,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(41525,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(41525,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(41525,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(41525,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(41525,36,'cp 38','d4e5 f6d7 e5e6 f7e6 g1f3 e8g8 d1d2 d7f6 e1c1 d8d2 g5d2 b8c6 f1c4 f6e8 h1e1 e8d6 c4b3 h7h6 h2h4 a8b8 b3e6 c8e6 e1e6 g8f7 e6e1 b8d8 c3d5 d6c4 d2c3 g7c3 d5c3 d8d1 c3d1 f7g7 d1e3 c4b6 e1d1 e7e5 d1e1');
INSERT INTO evaluations VALUES(41797,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(41797,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41797,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41797,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41797,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(41797,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(41797,36,'cp 0','e2e3 d6f4 e3f4 b7b6 h2h4 c7c5 h4h5 h7h6 g1f3 e8g8 f3e5 f6d7 e5f3');
INSERT INTO evaluations VALUES(44860,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(44860,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(44860,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(44860,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(44860,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(44860,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(44860,36,'cp 103','d5c6 b8c6 g1f3 c8g4 e1g1 d8c7 h2h3 g4h5 c4e2 e7e5 d2d3 a8d8 b1c3 f8e7 f3h4 h5e2 c3e2 e8g8 h4f5 f8e8 f1e1 f6d5 e2c3 d5c3 b2c3 e7f8 d1g4 g7g6 c1h6 f8h6 f5h6 g8g7 h6f5 g7h8 f5g3 e8f8 a1d1 f7f5 g4h4 a7a6');
INSERT INTO evaluations VALUES(44839,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(44839,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(44839,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(44839,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(44839,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(44839,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(44839,36,'cp 95','b8c6 g1f3 c8g4 h2h3 g4h5 c4e2 d8c7 e1g1 e7e5 c2c3 f8c5 d2d4 e5d4 f3d4 h5g6 c1e3 e8g8 b1d2 f8e8 f1e1 c5d4 c3d4 a8d8 d2b3 c7b6 a1c1 g6e4 e2f3 e4d5 f3d5 d8d5 d1e2 h7h6 e2c4 c6e7 e3d2 e8c8 c4d3 c8c1 e1c1 d5d8');
INSERT INTO evaluations VALUES(13183,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(13183,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(13183,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(13183,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(13183,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(13183,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(13183,36,'cp 105','g1f3 c8g4 h2h3 g4h5 c4e2 d8c7 c2c3 e7e5 e1g1 e5e4 f3d4 h5e2 d1e2 c6e5 d4b5 c7c5 b2b4 c5e7 d2d4 e4d3 e2e3 a8d8 f1e1 e5c4 b5c7 e8d7 e3d3 d7c8 d3f5 e7d7 f5d7 d8d7 c7b5 a7a6 b5a3 c4e5 c1g5 h7h6 g5f6 g7f6 a3c2 h8g8 g1f1 f8e7 c2d4 e5d3');
INSERT INTO evaluations VALUES(13182,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(13182,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(13182,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(13182,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(13182,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(13182,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(13182,36,'cp 83','c8g4 h2h3 g4h5 c4e2 d8c7 e1g1 e7e5 d2d3 a8d8 b1c3 h5g6 c1e3 f8e7 f3h4 e8g8 e2f3 f8e8 a2a3 a7a6 g2g3 c6d4 e3d4 d8d4 h4g6 h7g6 d1e2 e7c5 c3e4 f6e4 f3e4 f7f5 c2c3 d4d8 e4g2 g8h7 h3h4 e5e4 d3d4 c5a7');
INSERT INTO evaluations VALUES(45614,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(45614,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(45614,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(45614,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(45614,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(45614,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(45614,36,'cp -81','d5c4 g5f6 d8f6 g2g3 b8d7 e2e3 e6e5 f1c4 e5d4 d1d4 f8e7 d4f4 d7c5 g1f3 c8h3 h1g1 h7h5 e1c1 e8g8 f3e5 a7a5 c4a2 g7g5 f4f6 e7f6 e5d3 c5d3 d1d3 f8e8 d3d2 g8f8 a2b1 a8d8 d2d8 e8d8 g1d1 d8d1 c3d1 h5h4 f2f4 b7b5 f4g5 f6g5 g3h4');
INSERT INTO evaluations VALUES(49184,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(49184,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(49184,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(49184,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(49184,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(49184,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(49184,36,'cp -94','e8f7 c3e4 d7d5 e4g3 b8c6 d2d3 f8e7 d1h5 f7g8 c1d2 c8e6 g1f3 d8e8 h5e8 a8e8 h2h3 h7h6 e1g1 g8h7 f1e1 e7d6 a1c1 h8f8 c2c4 d5c4 d3c4 e6h3 c4c5 d6e7 f3e5 c6e5 e1e5 h3d7 d2c3 e7g5 c1c2 h7g6 c2e2 g5f6 e5e3 e8e3 e2e3 f6c3 e3c3 c7c6');
INSERT INTO evaluations VALUES(50979,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(50979,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(50979,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(50979,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(50979,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(50979,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(50979,36,'cp 121','g1f3 g8f6 b1c3 c7c5 d4d5 e6d5 c3d5 f8e7 e1g1 f6d5 c4d5 e8g8 c1e3 c8e6 d5e6 f7e6 d1b3 d8c8 a2a4 b8c6 a4a5 h7h6 f1c1 g8h7 h2h3 a8b8 e3c5 e7c5 c1c5 c6d4 c5c8 d4b3 c8b8 f8b8 a1a3 b3c1 f3d4 e6e5 d4f3 b8d8');
INSERT INTO evaluations VALUES(50978,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(50978,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(50978,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(50978,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(50978,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(50978,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(50978,36,'cp 114','g8f6 b1c3 c7c5 d4d5 e6d5 c3d5 f8e7 e1g1 f6d5 c4d5 e8g8 c1e3 c8e6 d5e6 f7e6 a1c1 d8d1 f1d1 b7b6 f3e5 a8a7 g1f1 a7b7 g2g3 f8c8 e5c4 b8c6 f2f4 c8d8 d1d8 e7d8 c1d1 d8e7 d1d2 c6a5 c4a5 b6a5 f1e2 a5a4 e2f3 g8f7 g3g4 h7h6 f3e2 c5c4 d2c2');
INSERT INTO evaluations VALUES(52729,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(52729,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(52729,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(52729,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(52729,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(52729,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(52729,36,'cp -6','g1f3 g8f6 e2e3 c8g4 c2c4 b8d7 c4d5 c6d5 a2a3 e7e6 f1e2 f8d6 h2h3 g4f3 e2f3 e8g8 e1g1 a8c8 d2d3 d6e5 d1b3 e5b2 b3b2 d8c7 a1a2 c7e5 b2b3 c8c7 a2c2 f8c8 f1c1 h7h6 d3d4 e5d6 c2c7 c8c7 b1c3 a7a6');
INSERT INTO evaluations VALUES(53584,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(53584,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(53584,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(53584,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(53584,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53584,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(53584,36,'cp 16','b1d2 g8f6 g1f3 b8d7 e2e3 f8e7 b2b3 b7b6 c1b2 e8g8 f1d3 c8a6 a1c1 c6c5 e1g1 f8e8 d1e2 a8c8 c4d5 a6d3 e2d3 e6d5 c1c2 h7h6 d3a6 c8c7 f1c1 d8b8 a3a4 e7f8 a6e2 e8c8 d2f1 f6e4 f3d2 e4d2 c2d2 c5d4 c1c7');
INSERT INTO evaluations VALUES(55506,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(55506,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55506,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(55506,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(55506,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(55506,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(55506,36,'cp 31','e2e4 f8e7 a2a3 f7f5 e4f5 c8f5 g1e2 f5g6 e2g3 g8f6 c2c4 e8g8 b1c3 b8d7 f1d3 d8e8 e1g1 a7a5 d3e2 d7c5 c1e3 c5e4 c3b5 e8d7 g3e4 g6e4 b5c3 e4g6 a1c1 f8f7 d1d2 a5a4 c3b5 h7h6 f1e1 f6g4 e2g4 d7g4');
INSERT INTO evaluations VALUES(636,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(636,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(636,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(636,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(636,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(636,35,'cp 225',NULL);
INSERT INTO evaluations VALUES(636,36,'cp 216','d2d4 c6c5 f2f3 e7d6 c1e3 c5d4 d1d4 d6d4 e3d4 c8e6 e5d3 e8c8 c3e2 f6g8 d4c3 g8e7 e1c1 e7g6 h2h4 h7h5 a2a3 e6c4 a3a4 f7f6 b2b3 c4f7 g2g3 f8d6 c1b2 h8e8 e2f4 g6f4 g3f4 d6e7 f4f5 g7g6 f5g6 f7g6 a4a5');
INSERT INTO evaluations VALUES(634,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(634,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(634,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(634,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(634,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(634,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(634,36,'cp 209','c6c5 f2f3 e7d6 c1e3 c5d4 d1d4 d6d4 e3d4 c8e6 e5d3 e8c8 c3e2 f8e7 d4e3 f6d7 e2f4 d7f8 h2h4 f7f5 f4e6 f8e6 e4f5 e6d4 e3d4 d8d4 h4h5 h8e8 e1f2 c7c5 h5h6 c5c4 h6g7 c4d3 c2d3 d4d3 a1c1 c8d8 h1h7 d3d2 f2f1 e7f6 g2g4');
INSERT INTO evaluations VALUES(637,30,'cp 238',NULL);
INSERT INTO evaluations VALUES(637,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(637,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(637,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(637,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(637,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(637,36,'cp 216','f2f3 e7d6 c1e3 c5d4 d1d4 d6d4 e3d4 c8e6 e5d3 e8c8 c3e2 f6d7 d3f4 d7c5 h2h4 e6c4 h4h5 h8g8 a1d1 c5e6 f4e6 c4e6 b2b3 c7c5 d4e3 d8d1 e1d1 f8d6 e2f4 d6f4 e3f4 g8e8 g2g4 f7f6 d1e2 c5c4 f4e3 c4b3 a2b3');
INSERT INTO evaluations VALUES(803,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(803,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(803,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(803,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(803,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(803,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(803,36,'cp -27','b3c2 c8f5 c2c1 e7e6 b1d2 f8e7 f3h4 f5e4 h4f3 e4g6 f3h4 g6h5 h2h3 f6d7 h4f3 e8g8 f1e2 f8c8 e1g1 h5g6 e2d1 b6d8 d1c2 b7b5 c2g6 h7g6 a2a3 a7a5 e3e4 b5b4 f1e1 d7f6 c1d1 d5e4 d2e4 d8d5 f3d2 f6e4 d2e4 g6g5');
INSERT INTO evaluations VALUES(805,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(805,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(805,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(805,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(805,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(805,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(805,36,'cp -30','c8f5 c2c1 e7e6 b1d2 f8e7 f3h4 f5e4 h4f3 e4g6 h2h3 b6d8 f3h4 g6e4 h4f3 e8g8 d2e4 f6e4 f3d2 e4d2 c1d2 b7b5 f1e2 e7d6 f4d6 d8d6 e1g1 f8b8 e2f3 f7f5 f1b1 h7h6 d2c2 b8b6 b2b3 a8b8 b3c4 b5c4 b1b6 b8b6 a2a3');
INSERT INTO evaluations VALUES(27396,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(27396,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(27396,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(27396,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(27396,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(27396,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(27396,36,'cp -24','c2c1 e7e6 b1d2 f8e7 f3h4 f5e4 h4f3 e4g6 f3h4 g6h5 h2h3 f6d7 h4f3 e8g8 f1e2 f8c8 e1g1 h5g6 e2d1 b6d8 d1c2 b7b5 c2g6 h7g6 e3e4 b5b4 a2a4 c6a5 c1c2 a5b3 d2b3 c4b3 c2b3 d5e4 f3d2 d7f6 f4e5 b4c3');
INSERT INTO evaluations VALUES(27395,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(27395,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(27395,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(27395,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(27395,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(27395,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(27395,36,'cp -28','e7e6 b1d2 f8e7 f3h4 f5e4 f1e2 h7h6 h4f3 e4h7 b2b3 c4b3 a2b3 f6h5 f4g3 e8g8 c1b2 a7a5 e1g1 h5g3 h2g3 f8c8 a1a2 b6c7 f1a1 c6b8 a1c1 h7f5 f3e1 b7b6 e1d3 b8d7 c3c4 f5d3 e2d3 e7b4 d2b1 g7g6 b1c3 g8g7');
INSERT INTO evaluations VALUES(4885,30,'cp 239',NULL);
INSERT INTO evaluations VALUES(4885,31,'cp 229',NULL);
INSERT INTO evaluations VALUES(4885,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(4885,33,'cp 252',NULL);
INSERT INTO evaluations VALUES(4885,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(4885,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(4885,36,'cp 262','b1c3 b7b5 c4d3 d7d6 a2a4 c8d7 e1g1 g8e7 d4d5 c6e5 f3e5 d6e5 a4b5 b4c3 b2c3 a6b5 a1a8 d8a8 d1b3 e8g8 c1a3 f8e8 a3e7 e8e7 d3b5 h7h6 b5d7 e7d7 c3c4 d7d8 c4c5 a8a7 b3c3 d8a8 c5c6 a7d4 c3d4 e5d4 f2f4 g8f8 f1d1 a8a4 g1f2 a4a2 f2f3 a2a3 f3g4 a3a2 e4e5 a2g2 g4f5 g2h2');
INSERT INTO evaluations VALUES(4883,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(4883,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(4883,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(4883,33,'cp 249',NULL);
INSERT INTO evaluations VALUES(4883,34,'cp 234',NULL);
INSERT INTO evaluations VALUES(4883,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(4883,36,'cp 220','b7b5 c4d3 d7d6 a2a4 b5a4 e1g1 a8b8 d1a4 c8d7 a4c2 g8e7 a1a6 e8g8 c3a2 b4a5 d4d5 c6b4 a2b4 a5b4 c1e3 d7g4 f3d2 g4d7 f1a1 e7g6 d3f1 g6e5 d2c4 e5c4 f1c4 f7f5 c4d3 f5f4 e3d4 f4f3 g2g3 d7b5 d3b5 b8b5 c2c4 b4c5 c4b5 c5d4');
INSERT INTO evaluations VALUES(5930,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(5930,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(5930,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(5930,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(5930,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(5930,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(5930,36,'cp 125','a7a6 b5d3 c8g4 c1e3 d6d5 h2h3 g4h5 e1g1 g8e7 a1c1 e8g8 e4e5 f7f6 e5f6 f8f6 g2g4 h5g6 f3e5 d8d6 f2f4 a8e8 c3e2 g6d3 d1d3 f6f8 a2a3 b4a5 b2b4 a5b6 e2g3 e7g6 g3f5 d6e6 c1e1 g6e7 f5e7 e6e7 f4f5 e7d6');
INSERT INTO evaluations VALUES(8603,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8603,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8603,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(8603,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(8603,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8603,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8603,36,'cp 0','f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 d8d6 e1g1 f7f5 e4c6 b7c6 d2d3 e5e4 d3e4 f5e4 c1g5 d6g6 g5e7 e4f3 d1f3 e8e7 a1e1 e7d8 e1e4 c8f5 e4f4 f5d7 f1e1 d8c8 f4f8 h8f8 f3f8 c8b7 f8b4 b7a6 b4a4 a6b7');
INSERT INTO evaluations VALUES(13454,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(13454,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(13454,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(13454,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(13454,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(13454,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(13454,36,'cp 51','c1h6 g7h6 f1b5 d8b6 d1e2 f8g7 e1g1 e8g8 d4c5 b6c5 b1d2 c5b6 d2b3 c8d7 b3d4 c6d4 f3d4 d7b5 d4b5 f7f6 e5f6 f8f6 a1d1 f6f7 a2a4 a8d8 b5d4 e6e5 d4f3 a7a5 h2h3 e5e4 f3h2 b6a6 e2a6 b7a6 f1e1 f7f8 d1d2 d5d4 c3d4 d8d4 d2d4 g7d4 h2g4 h6h5 e1e4 d4b2 g4e3 b2g7');
INSERT INTO evaluations VALUES(13451,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(13451,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(13451,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(13451,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(13451,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(13451,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(13451,36,'cp 50','g7h6 f1b5 d8b6 d1e2 f8g7 e1g1 e8g8 d4c5 b6c5 b1d2 c5b6 d2b3 c8d7 b3d4 c6d4 f3d4 d7b5 d4b5 f7f6 e5f6 f8f6 a2a4 f6f7 e2e3 b6e3 f2e3 a7a5 f1f3 a8f8 g1f2 f7f6 f2e2 g8f7 f3f6 g7f6 e3e4 f8g8 g2g3 g8d8 a1d1 f6e5 e4d5 d8d5 d1d5 e6d5 b5d4 f7f6');
INSERT INTO evaluations VALUES(16462,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(16462,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(16462,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(16462,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(16462,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(16462,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(16462,36,'cp 23','f1g2 d7d5 c4d5 f6d5 e1g1 d5b6 d2d3 f8e7 b1c3 e8g8 c1e3 f8e8 a1c1 h7h6 d3d4 e5d4 f3d4 c6d4 e3d4 c7c6 e2e3 c8f5 c3e2 e7f6 d1b3 d8e7 d4f6 e7f6 f1d1 f5e4 g2e4 e8e4 c1c5 a8d8 d1d8 f6d8 e2f4');
INSERT INTO evaluations VALUES(17530,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(17530,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(17530,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17530,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(17530,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17530,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17530,36,'cp 0','g1f3 e7e6 f1d3 f8c5 e1g1 g8e7 c2c4 d7d6 b1c3 c8d7 a4d1 e8g8 f1e1 c6e5 c1e3 d8b6 e3c5 e5f3 d1f3 d6c5 a1b1 a8d8 f3e3 b6a5 b1d1 d7c6 d3f1 a5b4 b2b3 b4b6 d1b1 b6a5 b1d1 a5b6');
INSERT INTO evaluations VALUES(18107,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(18107,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(18107,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(18107,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(18107,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(18107,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(18107,36,'cp -66','c8b7 d1e2 a8c8 f1d1 f8c5 c1g5 c5e7 g5f4 c6a5 e4e5 a5b3 a2b3 e7c5 c3e4 g8e7 f3g5 e7g6 f4d2 b5b4 e4c5 c8c5 g5e4 b7e4 e2e4 a6a5 d2e3 c5e5 e4b7 e8g8 d1d7 d8h4 h2h3 h7h6 a1c1 e5f5 c1c4 h4f6 e3d4 e6e5 d4e3 g6f4');
INSERT INTO evaluations VALUES(25593,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(25593,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(25593,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(25593,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(25593,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(25593,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(25593,36,'cp -58','d1e2 a8c8 f1d1 f8c5 c1f4 c6a5 f3e5 g8f6 f4g5 a5b3 a2b3 c8c7 e5d3 c5e7 e4e5 f6d5 c3d5 b7d5 g5e7 d8e7 a1a6 e8g8 b3b4 e7g5 f2f3 d5b3 d1d2 b3c4 a6d6 h7h6 e2e1 c4d3 d6d3 f8c8 h2h4 g5f5 d3d7 c7d7 d2d7 f5f4 h4h5 c8c2 d7d8 g8h7 b2b3 g7g6 h5g6 h7g6 g1h1 c2b2 d8d1 b2b3');
INSERT INTO evaluations VALUES(25592,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(25592,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(25592,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(25592,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(25592,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(25592,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(25592,36,'cp -62','a8c8 f1d1 f8c5 c1f4 c6a5 f3e5 g8f6 f4g5 a5b3 a2b3 c8c7 e5d3 c5e7 e4e5 f6d5 c3d5 b7d5 g5e7 d8e7 a1a6 e8g8 b3b4 e7g5 f2f3 d5b3 d1d2 b3c4 a6a3 h7h6 a3c3 c7a7 e2e1 f8a8 h2h4 g5f5 e1e3 a7a1 g1h2 a1f1 e3e4 f5e4 f3e4 c4d3 d2d3 a8a2 b2b3 f1e1 d3d7 e1e4');
INSERT INTO evaluations VALUES(19338,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(19338,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(19338,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19338,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(19338,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(19338,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19338,36,'cp 10','d7d6 b1c3 g8f6 d3d4 c5d4 f3d4 e7e5 d4e2 c8e6 c4b3 d8d7 e2g3 c6d4 e1g1 f8e7 d1d3 e8g8 c1e3 d4b3 a2b3 d6d5 e4d5 f6d5 c3d5 e6d5 f1d1 f8d8 a1a5 d5c6 d3d7 d8d7 d1d7 c6d7 a5e5 e7f6 e5a5 a7a6 c2c3 d7e6 b3b4 a8d8 f2f3 d8d1 g1f2');
INSERT INTO evaluations VALUES(19705,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(19705,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(19705,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(19705,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(19705,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19705,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(19705,36,'cp 28','e1g1 g8f6 f1e1 c8g4 c4b5 e7e6 h2h3 g4h5 c1f4 e6e5 f4h2 f8e7 c2c3 e8g8 b1d2 f6h7 g2g4 h5g6 h2g3 c6a5 d3d4 e5d4 c3d4 c5d4 f3d4 e7h4 g3h4 d8h4 b5f1 h4f6 d4b5 a5c6 b5c3 h7g5 c3d5 f6b2 e1e3 g5e6 f2f4 c6d4 f4f5 d4c2 f5e6 c2a1');
INSERT INTO evaluations VALUES(22488,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(22488,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(22488,32,'cp -164',NULL);
INSERT INTO evaluations VALUES(22488,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(22488,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(22488,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(22488,36,'cp -192','h7h6 d2d4 f8g7 c2c3 d7d6 d1b3 d8d7 c4d3 g8f6 b1d2 e8g8 f1e1 f6h5 d2f1 d7d8 h2h3 c6e7 b3d1 e7g6 b2b4 a7a5 b4b5 f8e8 a2a4 c7c5 b5c6 b7c6 c1a3 c6c5 a1b1 c5d4 c3d4 g5g4 h3g4 c8g4 e4e5 d6e5 d3g6 f7g6 d4e5 d8d1 b1d1 h5g3 f1g3 f4g3');
INSERT INTO evaluations VALUES(25288,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(25288,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(25288,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(25288,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(25288,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(25288,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(25288,36,'cp -25','c6d4 c3d5 c7c6 d5f6 g7f6 g5e3 d6d5 e3d4 d5c4 d4e3 c4d3 d1d3 d8d3 c2d3 h8g8 h2h3 g4e6 f3h4 a7a5 e1e2 a5a4 g2g4 h7h5 h4f5 a8a5 e2f3 g8h8 h1b1 e8d7 b2b4 a4b3 a2b3 a5b5 a1a8 b5b3 b1b3 h5g4 h3g4 e6b3 f3g3 b3e6 f2f4 d7c7 g3f3');
INSERT INTO evaluations VALUES(25619,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(25619,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(25619,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(25619,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(25619,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(25619,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(25619,36,'cp 52','b5c6 d7c6 d2d4 c5d4 f3d4 a8c8 c1g5 h7h6 g5h4 g7g6 d4c6 c8c6 c3d5 f8g7 c2c3 g8f6 h4f6 g7f6 d1b3 d8d7 f2f4 e8g8 a1d1 b7b5 g1h1 h6h5 d5b4 c6c4 e4e5 f6g7 b3a3 h5h4 h2h3 d7b7 e5d6 e7d6 b4a6');
INSERT INTO evaluations VALUES(25611,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(25611,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(25611,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(25611,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25611,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(25611,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(25611,36,'cp 62','d7c6 d2d4 c5d4 f3d4 a8c8 c1g5 h7h6 g5h4 g7g5 h4g3 f8g7 d4f5 g7c3 b2c3 c6e4 d1d4 e4f5 d4h8 e8f8 f2f4 g5g4 f1e1 f7f6 e1e3 f8f7 a1e1 e7e5 h2h3 g4h3 g3h4 e5f4 e3e7 d8e7 e1e7 f7e7 h8g7 e7e6 g2h3 f5c2 g7b7 g8e7 b7a6 c8c3 a6e2 e6f7 e2h5 c2g6 h5h6 e7f5 h6f4');
INSERT INTO evaluations VALUES(25628,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(25628,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(25628,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(25628,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(25628,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(25628,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(25628,36,'cp 55','d2d4 c5d4 f3d4 a8c8 c1g5 h7h6 g5h4 g7g5 h4g3 f8g7 f1e1 c6d7 c3d5 e7e5 d4f5 d7f5 e4f5 g8f6 d5e3 e8g8 c2c4 b7b5 b2b3 f8e8 d1d3 h6h5 h2h3 b5c4 b3c4 c8c6 a1d1 d8a5 d3e2 a5a4 g3h2 e8b8 g1h1 a4a3 h2g1 b8b2 d1d2 b2d2 e2d2 g5g4 h3g4 h5g4 e1d1');
INSERT INTO evaluations VALUES(26674,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(26674,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(26674,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(26674,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(26674,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(26674,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26674,36,'cp 20','e7e6 d4d5 e6d5 c4d5 d8d7 h2h3 g4f3 d1f3 g8f6 e1g1 f8d6 b1c3 f6d5 c3d5 e8g8 f1d1 f7f5 e4f5 d7f5 f3f5 f8f5 c1e3 h7h6 g2g3 a7a5 a1c1 a5a4 d5b6 a8a5 g3g4 f5f7 b6c4 a5b5 b2b3 a4b3 c4d6 c7d6 a2b3');
INSERT INTO evaluations VALUES(30096,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(30096,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30096,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30096,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(30096,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(30096,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30096,36,'cp 0','c7c5 e1g1 b8c6 b1d2 d8b6 d4c5 f8c5 d2b3 e8g8 a2a4 a7a5 b3c5 b6c5 c2c3 f6e4 h2h3 c5b6 e2b5 c6a7 d1d4 b6d8 b5e2 a7c6 d4d1');
INSERT INTO evaluations VALUES(30350,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(30350,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(30350,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(30350,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(30350,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(30350,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(30350,36,'cp 133','c7c5 c2c3 g8f6 e4e5 f6d5 d4c5 f8c5 a2a4 b5b4 b1d2 f7f5 e5f6 d5f6 f3g5 e8g8 g5h7 f6h7 d3h7 g8h7 d1h5 h7g8 h5c5 b8c6 d2c4 f8f5 c5e3 d8c7 e3b6 c7b6 c4b6 a8d8 f1d1 d7d5 c1e3 a6a5 c3b4 d5d4 e3d2 c6b4 d2b4 a5b4');
INSERT INTO evaluations VALUES(31765,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(31765,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(31765,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(31765,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(31765,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(31765,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(31765,36,'cp 63','g4f3 d1f3 c7c5 f1b5 b8d7 b2b3 e7e6 c1b2 a7a6 b5d7 d8d7 b1c3 g8e7 e1c1 e7c6 c3e2 e6e5 e2c3 f8e7 c3d5 a8d8 h3h4 e8g8 g2g4 c6b4 d5b4 c5b4 d2d4 f7f6 c1b1 a6a5 f3e2 d7c6 d4d5 c6d7 b2c1 d8c8 c1e3 b7b5 d1c1 a5a4 f2f3 h7h6');
INSERT INTO evaluations VALUES(31769,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(31769,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(31769,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(31769,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(31769,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(31769,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(31769,36,'cp 65','d1f3 c7c5 f1b5 b8d7 b2b3 e7e6 c1b2 a7a6 b5d7 d8d7 b1c3 g8e7 e1c1 e7c6 c3e2 e6e5 e2c3 f8e7 c3d5 a8b8 f3g4 d7g4 h3g4 c6b4 d5b4 c5b4 d2d4 f7f6 f2f3 b8c8 c1b1 h7h6 a2a3 b4a3 b2a3 e8d7 d1d2 c8c6 a3b4 b7b6 c2c4 e5d4 d2d4 a6a5 b4a3 h8c8 f3f4');
INSERT INTO evaluations VALUES(31768,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(31768,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(31768,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(31768,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(31768,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(31768,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(31768,36,'cp 64','c7c5 f1b5 b8d7 b2b3 a7a6 b5d7 d8d7 c1b2 e7e6 b1c3 g8e7 e1c1 e7c6 c3e2 e6e5 e2c3 f8e7 c3d5 a8b8 f3g4 d7g4 h3g4 c6b4 d5b4 c5b4 d2d4 f7f6 f2f3 h7h6 c1d2 b7b5 d2d3 a6a5 c2c3 b4c3 b2c3 e7d8 d1c1 d8b6 g4g5 f6g5 d4e5 d6e5 c3e5 b8d8 d3e2 e8g8');
INSERT INTO evaluations VALUES(32892,30,'cp 320',NULL);
INSERT INTO evaluations VALUES(32892,31,'cp 328',NULL);
INSERT INTO evaluations VALUES(32892,32,'cp 328',NULL);
INSERT INTO evaluations VALUES(32892,33,'cp 337',NULL);
INSERT INTO evaluations VALUES(32892,34,'cp 300',NULL);
INSERT INTO evaluations VALUES(32892,35,'cp 326',NULL);
INSERT INTO evaluations VALUES(32892,36,'cp 314','c1d2 c7c6 e2e4 f5c8 g1f3 a5d8 f1c4 e7e6 e1g1 f8e7 e4e5 b8d7 c3e4 d7b6 c4b3 a7a5 f1e1 g8h6 d2h6 g7h6 a1c1 a5a4 b3c2 e8d7 a2a3 d7c7 e4c3 c7b8 c2a4 h6h5 a4c2 h5h4 h2h3 h7h5 c2e4 h8g8 d1b3 b6d5 c3e2 d8d7 g1h1 g8g7 e4d5 d7d5 b3c2');
INSERT INTO evaluations VALUES(32891,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(32891,31,'cp 306',NULL);
INSERT INTO evaluations VALUES(32891,32,'cp 310',NULL);
INSERT INTO evaluations VALUES(32891,33,'cp 292',NULL);
INSERT INTO evaluations VALUES(32891,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(32891,35,'cp 331',NULL);
INSERT INTO evaluations VALUES(32891,36,'cp 313','c7c6 e2e4 f5c8 g1f3 a5d8 f1c4 e7e6 e1g1 f8e7 f1e1 g8f6 a1c1 e8g8 a2a3 b8d7 c4a2 d8b6 e4e5 f6d5 a2b1 f8d8 d2g5 d8e8 d1d2 e7g5 f3g5 d7f8 c3e4 b6d8 e4d6 e8e7 g5e4 f7f5 e5f6 d5f6 d2g5 f6e4 d6e4 g8h8 h2h4');
INSERT INTO evaluations VALUES(33789,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(33789,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(33789,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(33789,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(33789,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(33789,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(33789,36,'cp 140','b8c6 d4c6 b7c6 e1g1 c8b7 f1e1 f6d7 c1e3 g7g6 e3d4 f7f6 d1d2 f8g7 b2b4 e8g8 a1b1 a7a6 f2f4 g8h8 d4f2 a8b8 d3c4 g6g5 f4f5 d7e5 c4b3 f8d8 b1d1 e5g4 f2d4 c6c5 b4c5 d6c5 c3d5');
INSERT INTO evaluations VALUES(34461,30,'cp 335',NULL);
INSERT INTO evaluations VALUES(34461,31,'cp 336',NULL);
INSERT INTO evaluations VALUES(34461,32,'cp 313',NULL);
INSERT INTO evaluations VALUES(34461,33,'cp 340',NULL);
INSERT INTO evaluations VALUES(34461,34,'cp 321',NULL);
INSERT INTO evaluations VALUES(34461,35,'cp 351',NULL);
INSERT INTO evaluations VALUES(34461,36,'cp 335','c1e3 b8d7 d1d2 f6d8 f2f3 h7h5 e1c1 d7b6 h2h3 g8f6 a2a4 c8d7 f1d3 h5h4 h1e1 f8e7 e3g5 e8g8 a4a5 b6c8 e4e5 f6h5 g5e7 d8e7 f3f4 f8e8 e5e6 f7e6 d3g6 h5g3 g6e8 e7e8 d4f3 e8f8 f3g5 f8f6 d2e3 g3f5 e3e4');
INSERT INTO evaluations VALUES(34622,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(34622,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(34622,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(34622,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(34622,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(34622,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(34622,36,'cp 129','g7g6 d2d4 f8g7 f1c4 c8e6 c4d3 f5a5 f3g5 b8c6 g5e6 f7e6 e1g1 e8c8 c3e4 g7d4 e4g5 g8f6 g5f7 c6e5 f7h8 e5d3 c2d3 d8h8 d1e2 a5b6 c1f4 f6d7 a1b1 h8d8 h2h3 e6e5 f4e3 d7c5 e3d4 e5d4 b1d1 b6e6 f1e1 e6e2');
INSERT INTO evaluations VALUES(34623,30,'cp 269',NULL);
INSERT INTO evaluations VALUES(34623,31,'cp 269',NULL);
INSERT INTO evaluations VALUES(34623,32,'cp 268',NULL);
INSERT INTO evaluations VALUES(34623,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(34623,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(34623,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(34623,36,'cp 273','e2e4 f5a5 d2d4 g8f6 c1d2 c7c6 d1c2 a5d8 g1f3 c8g4 d2e3 g4f3 g2f3 e7e6 e1c1 f8e7 c1b1 e8g8 h1g1 g8h8 h2h4 b8d7 f3f4 d7b6 f1e2 e7b4 h4h5 h7h6 e2d3 b4c3 c2c3 b6d7 g1g3 f8g8 d1g1 a8c8 d3c2 d8e7');
INSERT INTO evaluations VALUES(34621,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(34621,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(34621,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(34621,33,'cp 251',NULL);
INSERT INTO evaluations VALUES(34621,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(34621,35,'cp 245',NULL);
INSERT INTO evaluations VALUES(34621,36,'cp 247','f5d7 g1f3 e7e6 d2d4 f8b4 f1d3 b8c6 e1g1 g8e7 d1b3 e8g8 f1d1 b4c3 b2c3 c6a5 b3c2 b7b6 c3c4 h7h6 c1b2 e7g6 a1c1 c7c5 d4c5 c8b7 c2c3 f7f6 d3f1 d7c7 f3d4 c7e5 c5b6 a7b6 c3g3 b7e4 g3e5 g6e5 d4e6');
INSERT INTO evaluations VALUES(34926,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34926,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(34926,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(34926,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(34926,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(34926,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(34926,36,'cp 22','d5d4 d1d4 e5d4 c3b5 f8b4 c1d2 b4d2 e1d2 e8d8 b5d4 g8f6 a1d1 f6e4 d2e3 e4d6 d4b5 h8e8 e3f3 c8f5 c2c4 b8d7 b5d6 c7d6 f3g3 d7f6 f2f3 f5e6 b2b3 d8e7 g1e2 d6d5 e2f4 e8d8 d1e1 a8c8 f1d3 d5c4 d3c4 d8d6 e1e2 c8c5 g3f2 f6h5 f4e6 f7e6 g2g4');
INSERT INTO evaluations VALUES(36511,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36511,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36511,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36511,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36511,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36511,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36511,36,'cp 11','f6e4 c4f7 f8f7 c3e4 d6d5 e4g5 f7f8 c2c3 h7h6 g5h3 c8h3 g2h3 c7c5 d4c5 e7e5 f4e5 b8c6 e1g1 d8d7 c1e3 d7h3 d1d5 g8h8 d5e4 g6g5 a1e1 g5g4 f3d4 c6e5 e3f4 a8e8 e4g2 e5d3 g2h3 g4h3 e1e3 e8e3 f4e3 d3b2 f1f8 g7f8 c5c6 b7c6 d4c6 b2d1 e3d4 h8g8 c6e5');
INSERT INTO evaluations VALUES(36707,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36707,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(36707,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36707,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(36707,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36707,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(36707,36,'cp 12','c4f7 g8f7 c3e4 f7g8 e1g1 b8d7 e4g5 d7b6 d1d3 d8e8 f1e1 e8c6 g5e6 c8e6 e1e6 a8e8 a2a4 c6d5 e6e2 a7a6 a4a5 b6d7 c2c4 d5f5 d3f5 g6f5 c1e3 e7e5 d4e5 d6e5 e2d2 e5f4 e3f4 d7e5 f3e5 g7e5 a1f1 e5f4 f1f4 e8e4 f4e4 f5e4 d2d7 e4e3');
INSERT INTO evaluations VALUES(36582,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(36582,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(36582,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(36582,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(36582,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(36582,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(36582,36,'cp -104','f6g4 d1d2 g4e3 d2e3 b8c6 g1e2 e7e5 d4d5 c6d4 e1g1 f7f5 e3d2 d8h4 e4f5 d4e2 d3e2 g6f5 f2f4 c8d7 a1d1 g8h8 c3b5 d7b5 g2g3 h4e7 c4b5 e5f4 g3f4 f8g8 g1h1 e7e4 e2f3 e4c4 f1e1 c4a2 e1e7 a2b2 d2b2 g7b2 e7c7 g8g7');
INSERT INTO evaluations VALUES(39153,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(39153,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(39153,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(39153,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(39153,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(39153,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(39153,36,'cp 67','c2c3 b4e7 e3e4 c7c5 f1d3 b8c6 g1f3 d7d5 e4e5 c8d7 e1g1 c5d4 c3d4 d8b6 d3e2 g8h6 c1h6 g7h6 d1d2 f7f6 e5f6 e7f6 f1d1 e8g8 b1c3 f6g7 c3a4 b6c7 a1c1 c7f4 d2f4 f8f4 e2b5 e6e5 d4e5 d7g4 a4c3 g4f3 g2f3 c6e5 d1d5 e5f3 g1g2 f3h4 g2f1 g8h8');
INSERT INTO evaluations VALUES(39151,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(39151,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(39151,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(39151,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(39151,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(39151,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(39151,36,'cp 59','b4e7 e3e4 c7c5 f1d3 b8c6 g1f3 d7d5 e4e5 c8d7 d1e2 c5d4 f3d4 d8c7 d4c6 b7c6 e1g1 g8h6 c1h6 g7h6 b1d2 e8g8 c3c4 g8h8 a1c1 f8g8 g2g3 g8g7 d2b3 a8b8 e2e3 c7b6 e3b6 a7b6 c4d5 e6d5 b3d4 e7c5 d3f5 c5d4 f5d7 d4b2 c1c6');
INSERT INTO evaluations VALUES(39290,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(39290,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(39290,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(39290,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(39290,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(39290,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(39290,36,'cp 128','c4f7 e8f8 d1h5 g8f6 h5c5 f8f7 g1e2 h8e8 e1g1 d7d6 c5f2 f6e4 f2f4 e4f6 b1c3 f7g8 d2d3 c8g4 e2g3 b8c6 h2h3 g4e6 b2b3 d8d7 c1b2 e8f8 f4h4 a8e8 a1e1 c6e5 c3e4 e5g6 e4f6 f8f6 h4g5 f6f1 e1f1 d7e7 g5e3 c7c5 g3h5 e7h4');
INSERT INTO evaluations VALUES(39778,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(39778,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(39778,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(39778,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(39778,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(39778,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(39778,36,'cp 48','e2e3 e8g8 c4c5 f6h5 f4e5 f7f6 e5g3 g7g6 f1d3 b8d7 a2a3 c7c6 b2b4 e6e5 d4e5 h5g3 h2g3 d7e5 d1c2 g8g7 e1g1 c8e6 a1d1 d8c7 e3e4 d5e4 f3e5 f6f5 f2f4 e4d3 d1d3 a8d8 f1d1 d8d3 d1d3 f8d8 g1h2 d8d3 c2d3');
INSERT INTO evaluations VALUES(39777,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(39777,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(39777,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(39777,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(39777,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(39777,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(39777,36,'cp 25','e8g8 c4d5 f6d5 c3d5 e6d5 f1d3 e7d6 f4d6 d8d6 d1c2 g7g6 c2c3 c7c6 e1g1 b8d7 a2a4 d7f6 b2b4 c8e6 h2h3 f8c8 f3e5 f6e8 f2f3 f7f6 e5g4 e6g4 h3g4 e8c7 a4a5 d6e7 f1e1 c8e8 g1h2');
INSERT INTO evaluations VALUES(35599,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(35599,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(35599,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35599,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(35599,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(35599,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(35599,36,'cp 35','c4d5 f6d5 c3d5 e6d5 f1d3 e7d6 f4d6 d8d6 d1c2 g7g6 c2c3 c7c6 e1g1 a6a5 f3d2 b8d7 a2a4 d7f6 a1b1 c8e6 b2b4 a5b4 b1b4 f8b8 f1b1 f6d7 b4b7 b8b7 b1b7 a8a4 d2b3 g8g7 h2h3 a4a3 d3c2 h7h5 e3e4 d6f4 g2g3 f4g5');
INSERT INTO evaluations VALUES(39993,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(39993,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(39993,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(39993,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(39993,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(39993,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(39993,36,'cp 56','e4e5 f6g4 c4f7 e8f7 f3g5 f7g8 d1g4 b8c6 g4c4 e7e6 d2d3 c6e5 c4c5 h7h6 d3d4 b7b6 c5b5 e5c6 g5f3 c6d4 f3d4 g7d4 c1e3 d4e3 f2e3 c8b7 b5d3 d8h4 g2g3 h4h5 e1g1 h8h7 f1f6 h7g7 a1f1 a8c8 f1f2 h5g4 f2f4 g4h5 a2a4 b7c6 c3b5 c6b5 a4b5 h5d5 f4d4 d5g5 f6f4 g5e5 d4d7 g7d7 d3d7 e5e3 g1g2 c8f8');
INSERT INTO evaluations VALUES(44393,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(44393,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44393,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(44393,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44393,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(44393,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(44393,36,'cp 35','e1g1 a7a6 f1e1 b8c6 c4f1 d7d5 e4d5 e7d5 d2d4 d5f6 c1e3 c5d4 f3d4 c8d7 c2c4 d8c7 b1c3 f8e7 g2g3 c6d4 e3d4 d7c6 d4e5 c7d8 d1e2 e8g8 a2a3 f6d7 e5d4 e7f6 a1d1 d8a5 f1g2 a8d8 g2c6 b7c6 d4f6 d7f6 d1d8 f8d8');
INSERT INTO evaluations VALUES(44932,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(44932,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(44932,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(44932,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(44932,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(44932,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44932,36,'cp 0','c8f5 e2e3 e7e6 f1e2 f8d6 f4d6 d8d6 b1d2 e8g8 c2c4 a7a5 d1b3 d6e7 a2a3 h7h6 e1g1 f8c8 f1c1 b8d7 c4d5 e6d5 b3b7 a8b8 b7a6 b8a8 a6b7');
INSERT INTO evaluations VALUES(28991,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(28991,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(28991,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(28991,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(28991,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(28991,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28991,36,'cp 14','e2e3 e7e6 f1d3 f5d3 c2d3 d8b6 d1d2 c6c5 b1c3 c5d4 f3d4 f8e7 a1c1 e8g8 e1g1 f6d7 f1d1 b8c6 d4c6 b7c6 d3d4 f8c8 b2b3 c6c5 d4c5 d7c5 c1c2 c5d7 d1c1 e7a3 c3a4 b6b7 c2c8 a8c8 c1c8 b7c8 d2a5 c8c1 g1h2');
INSERT INTO evaluations VALUES(28988,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(28988,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(28988,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(28988,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28988,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(28988,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(28988,36,'cp 13','e7e6 f1d3 f8d6 f4d6 d8d6 e1g1 e8g8 c2c4 d5c4 d3c4 c6c5 d1b3 b8c6 d4c5 d6c5 b1d2 f8d8 b3b7 a8b8 b7c7 b8b2 f1c1 d8d7 c4f1 b2c2 c7c8 d7d8 c1c2 c5c2 c8b7 f6e4 d2e4 f5e4 b7c7 e4f3 g2f3 d8d2 c7c8 d2d8');
INSERT INTO evaluations VALUES(29259,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(29259,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(29259,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(29259,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(29259,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(29259,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(29259,36,'cp 10','f1d3 f8d6 f4d6 d8d6 e1g1 b8d7 c2c4 d5c4 d3f5 e6f5 d1c2 g7g6 c2c4 e8g8 c4b3 a7a5 b1c3 f8e8 f1d1 d6b4 f3d2 g8g7 a2a3 b4b3 d2b3 h7h5 a3a4 h5h4 a1a3 e8e7 d1c1 f6e4 c3e4 f5e4 b3d2 f7f5 a3b3');
INSERT INTO evaluations VALUES(47146,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(47146,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(47146,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(47146,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(47146,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(47146,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(47146,36,'cp -82','c7c5 c4d5 d8d5 e2e3 c5d4 d1d4 d5d4 e3d4 b8c6 f1b5 c8d7 g1e2 a7a6 b5a4 c6b4 a4b3 b4d3 e1d2 d3c1 h1c1 g7g6 d2e1 e7e6 b1d2 a6a5 a2a4 d7c6 d2c4 f6d5 e1f2 h7h5 c4e5 d5b4 f3f4 f8g7 c1c3 g7f6 e5c6 b7c6 f2e3 h5h4 a1d1 f6e7 b3c4 e8g8');
INSERT INTO evaluations VALUES(47468,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(47468,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(47468,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(47468,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(47468,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(47468,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(47468,36,'cp 20','e7e5 b1c3 d5b6 g1f3 b8c6 e1g1 f8e7 a2a3 c8e6 b2b4 a7a5 b4b5 c6d4 d2d3 e8g8 c1b2 c7c6 f3e5 e7f6 e2e3 f6e5 e3d4 e5d4 b5c6 b7c6 d1d2 a8b8 g2c6 b6d7 c3b5 d4b2 d2b2 d8b6 a1c1 d7c5 b2d4 c5b3 d4b6 b8b6');
INSERT INTO evaluations VALUES(49023,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(49023,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(49023,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(49023,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(49023,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(49023,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(49023,36,'cp 124','f3e5 f8d6 e5c4 e8g8 f1e2 b8c6 b1c3 d6b4 c1d2 f8e8 c3d5 d8d5 e1g1 c6d4 c4e3 d4e2 d1e2 b4d6 e3d5 e8e2 a1e1 e2e1 f1e1 c8e6 d5c3 e6d7 f2f3 f7f6 c3e4 d6e7 g1f2 g8f7 d2f4 a8c8 g2g4 a7a5 h2h4 h7h6 f4g3 b7b5 c2c3 d7e6 e1a1 a5a4 d3d4');
INSERT INTO evaluations VALUES(48994,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(48994,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(48994,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(48994,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(48994,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(48994,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(48994,36,'cp 103','c7c5 f1e2 d8c7 e5c4 f8e7 e1g1 e8g8 b1c3 d5c3 b2c3 c8e6 c4e3 b8c6 e2f3 a8e8 c1d2 b7b6 f3d5 e7d6 d1h5 c7d7 a1e1 e6d5 e3d5 c5c4 d3d4 c6e7 d5e3 d7c6 f2f4 f7f5 g2g3 b6b5 h5f3 c6d7 g1h1 a7a6 f3g2 e7c6');
INSERT INTO evaluations VALUES(49603,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(49603,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(49603,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(49603,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(49603,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(49603,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(49603,36,'cp -113','d7d5 e4d5 f6d5 g2g3 f8c5 f1g2 e8g8 b1c3 d5c3 b2c3 b8c6 g1f3 c8g4 e1g1 d8d7 a1b1 c5b6 d1e1 f8e8 c1e3 h7h6 f3d2 g4h3 g2h3 d7h3 e1e2 c6d4 e3d4 e5d4 e2f3 d4c3 d2e4 e8e6 b1b5 a8e8 b5f5 e6g6 h4h5 g6g4 f5f7 h3h5');
INSERT INTO evaluations VALUES(50892,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(50892,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(50892,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(50892,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(50892,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(50892,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(50892,36,'cp 137','b1c3 b8d7 f1e2 c7c5 d4d5 e6d5 d1d5 g8f6 d5d3 f8e7 c1f4 d8c7 a1d1 d7e5 d3c2 e8g8 e1g1 c8e6 b2b3 e5g6 f4e3 e6d7 h2h3 a8e8 f3h2 d7c6 f2f4 e7d8 e2d3 g6e7 c3d5 e7d5 c4d5');
INSERT INTO evaluations VALUES(52224,30,'cp 406',NULL);
INSERT INTO evaluations VALUES(52224,31,'cp 422',NULL);
INSERT INTO evaluations VALUES(52224,32,'cp 423',NULL);
INSERT INTO evaluations VALUES(52224,33,'cp 420',NULL);
INSERT INTO evaluations VALUES(52224,34,'cp 413',NULL);
INSERT INTO evaluations VALUES(52224,35,'cp 428',NULL);
INSERT INTO evaluations VALUES(52224,36,'cp 431','f3e5 d8e7 e5f3 e7e4 c4e2 e4f5 e1g1 d7d5 d2d4 f8d6 c2c4 g8e7 c4d5 e8g8 b1c3 b8a6 d4c5 d6c5 a2a3 e7g6 b2b4 c5d6 f3d4 f5d7 d4e6 f8f7 c3b5 a6c7 e6c7 d6c7 d5d6 c7b8 e2c4 a7a6 c4f7 d7f7 b5c7 b8c7 d6c7 f7c7 d1d5 c7f7 d5d8 g6f8');
INSERT INTO evaluations VALUES(53604,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(53604,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(53604,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(53604,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(53604,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(53604,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(53604,36,'cp 34','f8e7 e1g1 g8f6 c2c4 b8d7 d1c2 b7b6 b2b3 e8g8 f1d1 c8a6 b1d2 c6c5 e2e4 d5c4 e4e5 f6d5 b3c4 d5b4 c2c3 a6b7 d4c5 e7c5 a2a3 b4a6 c3c2 f7f5 f3g5 b7g2 g5e6 d8e7 e6f8 g2b7 f8d7 e7d7 d2f3');
INSERT INTO evaluations VALUES(53760,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(53760,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(53760,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(53760,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(53760,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(53760,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(53760,36,'cp 182','d4e5 f8b4 d1d4 d8b6 d4b6 a7b6 e4d5 b4c3 b2c3 b8d7 d5c6 b7c6 f2f4 g8e7 c3c4 a8a4 f1d3 d7c5 g1e2 e8g8 e2c3 c5d3 c2d3 a4a8 c3e4 c8f5 a2a4 b6b5 e4c5 f8d8 c1a3 b5c4 d3c4 e7g6 e1g1 d8d2 f1f2 d2c2 f2c2');
INSERT INTO evaluations VALUES(53743,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(53743,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(53743,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(53743,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(53743,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(53743,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(53743,36,'cp 207','f8b4 d1d4 d8b6 d4b6 b4c3 b2c3 a7b6 e4d5 b8d7 d5c6 b7c6 f2f4 g8e7 c3c4 a8a4 g1f3 e8g8 a2a3 d7c5 c1e3 c8e6 f3d2 c5d7 e1f2 f8a8 f1d3 a4a3 a1a3 a8a3 h1b1 e6f5 b1b3 a3b3 d2b3 c6c5 d3f5 e7f5 g2g4 f5e7 b3d2 g7g5 f4f5 d7e5 e3g5 f7f6');
INSERT INTO evaluations VALUES(54750,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(54750,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(54750,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(54750,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(54750,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(54750,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(54750,36,'cp 48','e7e6 f1e2 g8f6 g1f3 a7a5 e1g1 f8d6 b2b3 e8g8 c1b2 c8d7 d1c1 d7e8 c4c5 d6e7 f3e5 b8d7 e5d3 g7g5 b1d2 g8h8 a2a3 f8g8 b3b4 c7c6 b4a5 d8a5 b2c3 a5c7 c1b2 f6e4 a3a4 e7f6 d2e4 f5e4');
INSERT INTO evaluations VALUES(56760,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(56760,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(56760,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(56760,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(56760,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(56760,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(56760,36,'cp 67','g1f3 g8f6 f1e2 f8e7 b2b3 c7c6 e1g1 e8g8 c1b2 f6e4 b1d2 b8d7 f3e1 e7d6 d2f3 d8e7 f3e5 d7e5 d4e5 d6b4 c4d5 e6d5 d1c2 b4a3 b2a3 e7a3 e1d3 a7a5 a1c1 a3e7 a2a4 e4g5 h2h3 g5e6 f2f4 g8h8 d3c5 g7g5 g1h1 e7c5 c2c5 e6c5 c1c5 g5f4');
INSERT INTO evaluations VALUES(56761,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(56761,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(56761,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(56761,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(56761,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(56761,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(56761,36,'cp 67','c7c6 f1e2 g8f6 e1g1 f8e7 b2b3 e8g8 c1a3 c8d7 a3e7 d8e7 b1d2 d7e8 d1c2 e8g6 c4d5 e6d5 e2d3 f6e4 f3e5 a7a5 a1c1 g6e8 d3e2 e4d6 a2a4 d6e4 f2f4 e4d2 c2d2 b8d7 c1c2 g8h8 d2c3 d7f6 e2d3 g7g6 f1c1');
INSERT INTO evaluations VALUES(54625,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(54625,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(54625,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(54625,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(54625,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(54625,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(54625,36,'cp 62','f1e2 g8f6 e1g1 f8e7 b2b3 e8g8 d1c2 c8d7 c1b2 a7a5 b1c3 e7d6 c3a4 d7e8 a4c5 d8e7 c5d3 e8h5 f3e5 h5e2 c2e2 f6d7 e5f3 d7f6 a2a4 b8d7 f3e5 f8d8 f2f3 d7f8 e2c2 f8g6 g1h1 d6c7 b2c3 f6d7 f3f4');
INSERT INTO evaluations VALUES(56680,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(56680,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(56680,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(56680,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(56680,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(56680,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(56680,36,'cp 109','g7g6 a2a4 a7a6 h2h4 h7h5 a4a5 f8h6 g2g3 d8e7 g1h3 c8g4 f2f3 h6c1 a1c1 g4c8 d1d2 g8h6 b2b4 b8d7 f1d3 e7g7 c3e2 e8g8 e1g1 b7b6 a5b6 c7b6 g1g2 b6b5 c2c4 b5c4 d3c4 a8b8 c4d3 a6a5 b4a5 d7c5 h3f2 f6f5 g2h1 c8d7 e4f5');
INSERT INTO evaluations VALUES(796,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(796,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(796,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(796,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(796,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(796,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(796,36,'cp -33','c5c4 b3c2 c8f5 c2c1 f6h5 f4g3 h5g3 h2g3 g7g6 f1e2 f8g7 b1d2 h7h5 e1g1 e8g8 f1e1 e7e6 f3h4 f5g4 f2f3 g4f5 h4f5 e6f5 f3f4 b6c7 b2b3 b7b5 e2f3 c6e7 a2a4 a7a6 g1f2 f8c8 a4b5 a6b5 b3c4 b5c4 c1c2 a8a1 e1a1 c8b8');
INSERT INTO evaluations VALUES(1603,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(1603,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(1603,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(1603,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(1603,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(1603,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(1603,36,'cp 6','e7e5 g1f3 c8g4 b1c3 f8b4 f1e2 e5d4 e1g1 d5h5 f3d4 b4d6 h2h3 g4e2 c3e2 g8e7 d4c6 e7c6 e2d4 h5d1 a1d1 c6d4 d1d4 e8c8 b2b3 a7a6 f1d1 d6e5 d4d5 g7g6 g1f1 d8d5 d1d5 e5f6 f1e2 h8d8 d5d8 c8d8 e2d3 d8d7 g2g4 f6e7 d3e4 d7e6 e3f4 c7c6 f2f3');
INSERT INTO evaluations VALUES(1715,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(1715,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(1715,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(1715,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(1715,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(1715,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(1715,36,'cp 0','g1f3 e5d4 f3d4 f8b4 c2c3 b4e7 b1d2 g8f6 d4c6 d5c6 d1f3 c6f3 d2f3 f6d5 e3d4 e8g8 e1c1 c8e6 f1d3 f8e8 d3c2 c7c6 h1e1 a8d8 a2a4 a7a5 d4e5 e6g4 e1e4 g4f5 e4e1');
INSERT INTO evaluations VALUES(2901,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(2901,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(2901,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(2901,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(2901,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(2901,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(2901,36,'cp -66','e1c1 d8c7 e3h6 g7h6 d2h6 d5e4 f3e4 c8e6 f1e2 f8b8 h6e3 a7a5 a2a3 a5a4 d1d4 c7b6 c3d1 a8a5 h1f1 c6c5 d4d2 b6c7 c2c4 a5a7 d1c3 c7h2 e4e5 f6d7 g2g3 h2h3 c1b1 d7b6 c3d5 a7d7 e3c5 h3g3 d5e7 g8g7');
INSERT INTO evaluations VALUES(2899,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(2899,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(2899,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(2899,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(2899,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(2899,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(2899,36,'cp -73','d8c7 e3h6 g7h6 d2h6 d5e4 f3e4 a8b8 f1e2 c8e6 h6e3 c6c5 h2h4 h7h5 c3d5 e6d5 e4d5 b8b4 a2a3 b4e4 e3f2 f8d8 e2f3 e4f4 f2e3 c7d6 c1b1 d8b8 b1a1 b8b7 d1e1 b7d7 c2c3 a7a5 a1b1 g8g7 e1e2 f6g4 e3d2 g4h6 e2e4 f4e4 f3e4 d6f6 b1a1');
INSERT INTO evaluations VALUES(5413,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5413,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5413,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(5413,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5413,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(5413,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5413,36,'cp 5','d7d6 c1e3 c8e6 d1d2 g8f6 f2f3 d6d5 e4d5 f6d5 c3d5 d8d5 d2d5 e6d5 e1c1 e8c8 f1b5 a7a6 b5e2 d8d6 e3f4 e7e5 f4e3 h8e8 e2a6 b7a6 c2c4 c6d4 c4d5 d4b3 a2b3 e8d8 c1c2 c8b7 d1d2 d6d5 h1d1 d8c8 c2b1 d5d2 e3d2 c8d8 b1c2 d8d6 d1a1 f7f5 d2c3 g7f6 g2g3 d6c6 a1a4 c6e6 a4b4 b7a8');
INSERT INTO evaluations VALUES(7538,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(7538,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(7538,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(7538,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(7538,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(7538,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(7538,36,'cp 42','f1d3 h7h6 g5d2 e8g8 e1g1 a7a5 h2h3 a5a4 b2b4 a4b3 c2b3 f6d7 a2a4 e7f6 b3b4 f6d4 a1a3 d7e5 c3e2 d4b2 a3b3 e5d3 b3d3 b2f6 d1c2 c8a6 d3e3 a6e2 e3e2 c6c5 b4b5 f6d4 g1h2 d8d7 f2f4 f8e8');
INSERT INTO evaluations VALUES(7653,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7653,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7653,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(7653,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7653,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7653,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7653,36,'cp 0','e1g1 e8g8 d1f3 c5d4 c3e2 d4c5');
INSERT INTO evaluations VALUES(7652,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7652,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7652,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7652,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7652,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7652,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7652,36,'cp -5','c5d4 d1f3 e8g8 h2h3 d7d6 c4b3 c8e6 c3e2 d4b6 e2g3 h7h6 f1e1 a7a5 b3a4 a8c8 c2c3 b6c5 a4c2 g8h7 g3f5 f8e8 a2a3 e6f5 f3f5 g7g6 f5f3 d6d5 f3e2 f6d7 c1d2 d8f6 a1d1 c5b6 b2b4');
INSERT INTO evaluations VALUES(8438,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8438,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8438,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8438,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8438,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8438,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8438,36,'cp -4','c6a5 c4b3 c7c6 h2h3 e8g8 d3d4 d8c7 a2a3 a5b3 c2b3 e5d4 f3d4 f8d8 e3f4 c7b6 f1e1 c6c5 d4e2 c8d7 e2g3 d8e8 b3b4 d7e6 d1e2 c5b4 f4e3 b6c7 c3b5 c7d7 a1d1 d6d5 e4e5 f6h7 e3a7 b4a3 b2a3');
INSERT INTO evaluations VALUES(8439,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(8439,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(8439,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(8439,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(8439,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(8439,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(8439,36,'cp 32','a2a3 e8g8 h2h3 c8e6 c3d5 d8d7 e1g1 e7d8 c4b3 f6h7 c2c4 f7f5 e4f5 e6f5 d3d4 e5e4 f3d2 d8g5 d5c3 h7f6 b3a4 d7f7 f2f4 e4f3 e3g5 h6g5 a4c6 b7c6 d1f3 f5d7 a1e1 g5g4 h3g4 f6g4 f3f7 f8f7 f1f7 g8f7 d2e4');
INSERT INTO evaluations VALUES(9875,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(9875,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(9875,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(9875,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(9875,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(9875,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(9875,36,'cp -8','c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 d7d5 b3b4 h8f8 b1c3 f7g8 c3e4 d5e4 f3e5 c8e6 e1g1 d8d6 b4c3 f8f5 e5c4 e6c4 c3c4 f5d5 a1c1 c7c6 c4b3 b7b6 b3c4 b6b5 c4c6 d6c6 c1c6 d5d4 b2b3 d4d2 a2a4 d2b2 c6e6 b2b3 a4b5 b3b5 e6e4 b5b2 e4g4');
INSERT INTO evaluations VALUES(9874,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(9874,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(9874,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(9874,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(9874,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(9874,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(9874,36,'cp -17','f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 d7d5 b3b4 h8f8 e1g1 f7g8 b1d2 a7a5 b4b3 a5a4 b3e3 e4d2 f3d2 c8g4 e3g3 g4h5 a1e1 d8g5 g3g5 h6g5 e1e5 f8f4 e5d5 h5f7 d5d7 f7a2 d2f3 g5g4 f3e5 a2e6 d7c7 e6d5 e5d7 f4d4 d7b6 a8d8 b6d5 d4d5 c7b7');
INSERT INTO evaluations VALUES(12435,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12435,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(12435,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(12435,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12435,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12435,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12435,36,'cp 0','f1e2 f8e7 c1f4 d7d5 b1d2 f6h5 f4e3 h5f6');
INSERT INTO evaluations VALUES(12434,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12434,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12434,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12434,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12434,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12434,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12434,36,'cp 0','d8c7 e1g1 d7d5 e4d5 f6d5 d3d4 c5d4 c3d4 f8e7 b1c3 e8g8 f1e1 d5f6 c1e3 f6g4 e2d3 g4e3 e1e3 f8d8 a1c1 c7b6 c3a4 b6c7');
INSERT INTO evaluations VALUES(13216,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13216,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13216,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13216,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13216,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13216,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13216,36,'cp 0','e7e5 b5c4 f8e7 d2d3 h7h6 a2a4 e8g8 c1d2 c8e6 c3d5 d8d7 a4a5 e7d8 c4b3 f6h7 b3a4 f7f5 b2b4 f5e4 d3e4 h7f6 d5f6 d8f6 b4c5 d6c5 d1e2 d7c7 a4c6 c7c6 d2c3 e6g4 e2c4 g4e6');
INSERT INTO evaluations VALUES(13829,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13829,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13829,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13829,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13829,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(13829,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13829,36,'cp 0','b5c4 f8e7 d2d3 h7h6 f1e1 e8g8 a2a4 c8e6 h2h3 d8d7 c1d2 e7d8 f3h2 a8c8 c3d5 e6d5 e4d5 c6e7 d1f3 e7g6 d2c3 g6f4 c3d2');
INSERT INTO evaluations VALUES(14242,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(14242,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(14242,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(14242,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(14242,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(14242,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(14242,36,'cp -36','e2e3 c8f5 f1d3 e7e6 g1f3 c6b4 d3f5 e6f5 c1d2 b4c6 e1g1 f8b4 c3a4 e8g8 d2b4 c6b4 d1e2 d8e7 f3e5 f6e4 a2a3 b4c6 a4c3 e4c3 b2c3 f7f6 e5c6 b7c6 c3c4 f8e8 f1e1 a8b8 c4c5 b8b2 e2d3 e7e4 d3e4 e8e4');
INSERT INTO evaluations VALUES(14240,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(14240,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(14240,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(14240,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(14240,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(14240,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(14240,36,'cp -34','c8f5 f1d3 e7e6 g1f3 c6b4 f3e5 f8d6 c1d2 b4d3 c2d3 h7h6 e1g1 e8g8 a1c1 c7c6 a2a3 f5h7 c3a4 f6d7 d2b4 d6b4 e5d7 d8d7 a3b4 b7b6 c1a1 f7f6 d1d2 f8e8 a4c3 e6e5 f4e5 f6e5 c3e2 d7e6 e2c3 e5d4 e3d4 e6d7');
INSERT INTO evaluations VALUES(24840,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(24840,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(24840,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(24840,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(24840,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(24840,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(24840,36,'cp -25','g1f3 c6b4 f1d3 e7e6 f3e5 f8d6 a2a3 b4d3 c2d3 e8g8 e1g1 c7c5 c3b5 d6e7 d4c5 e7c5 b2b4 c5b6 a3a4 a7a6 b5d4 f8e8 c1b2 f6d7 d4f5 e6f5 d1f3 a8c8 a1c1 h7h6 h2h3 d7e5 f4e5 c8c1 f1c1 f7f6 a4a5 b6a7');
INSERT INTO evaluations VALUES(16199,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(16199,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(16199,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(16199,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(16199,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(16199,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(16199,36,'cp 70','f3e5 d7d5 c4b5 a7a6 b5c6 b7c6 e1g1 f8d6 f2f3 d6e5 d4e5 e4c5 c1e3 d8e7 b2b4 c5b7 f3f4 e8g8 e3d4 c8f5 d1h5 f5d3 f1d1 d3g6 h5f3 g6f5 b1d2 c6c5 b4c5 b7c5 f3d5 c5e6 d5f3 e7h4 d1f1 f8d8 d2e4 h4g4 f3g4 f5g4 d4e3 d8d3 f1e1 g4f5 e4g3 g7g6 a1d1 a8d8 d1d3 d8d3 g3f5');
INSERT INTO evaluations VALUES(17248,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(17248,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(17248,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(17248,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(17248,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(17248,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(17248,36,'cp 22','e7e6 a2a3 d7d5 g1f3 f8d6 f4g5 d6e7 e2e3 h7h6 g5h4 a7a6 f1d3 b7b6 e1g1 e8g8 c3e2 c8b7 c2c3 a6a5 a1c1 c6b8 f3e5 b8d7 f2f4 f6e4 h4e7 d8e7 d3e4 d5e4 e2g3 f8d8 e5d7 d8d7 f4f5 e7g5 f1f4 e6f5 g3f5 a8e8 d1g4 g5g4');
INSERT INTO evaluations VALUES(16282,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(16282,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(16282,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(16282,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(16282,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(16282,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(16282,36,'cp 25','a2a3 d7d5 g1f3 f8d6 f4g5 d6e7 e2e3 a7a6 f1d3 h7h6 g5h4 b7b6 d1e2 c8b7 h2h3 c6b8 e3e4 d5e4 c3e4 b8d7 h4g3 f6e4 d3e4 b7e4 e2e4 e8g8 a1d1 e7d6 e1g1 d6g3 f2g3 c7c5 f3e5 d7f6 e4e3 d8c7 d4c5 c7c5 e3c5');
INSERT INTO evaluations VALUES(17988,30,'cp -367',NULL);
INSERT INTO evaluations VALUES(17988,31,'cp -363',NULL);
INSERT INTO evaluations VALUES(17988,32,'cp -356',NULL);
INSERT INTO evaluations VALUES(17988,33,'cp -346',NULL);
INSERT INTO evaluations VALUES(17988,34,'cp -367',NULL);
INSERT INTO evaluations VALUES(17988,35,'cp -367',NULL);
INSERT INTO evaluations VALUES(17988,36,'cp -373','c6d4 d1d4 c7c5 d4c3 c5c4 b3c4 b5c4 c3c4 f8e7 c4d3 c8e6 e1g1 d6d5 e4d5 d8d5 d3d5 e6d5 f1e1 g8f6 b1d2 a8d8 d2f3 d5f3 g2f3 e8d7 c1g5 h8e8 e1e3 e7d6 g5f6 g7f6 a1d1 e8e6 d1d4 d7e7 d4a4 d8c8 c2c3 c8c6 e3e6 e7e6 g1g2 e6e7');
INSERT INTO evaluations VALUES(18025,30,'cp -348',NULL);
INSERT INTO evaluations VALUES(18025,31,'cp -329',NULL);
INSERT INTO evaluations VALUES(18025,32,'cp -328',NULL);
INSERT INTO evaluations VALUES(18025,33,'cp -357',NULL);
INSERT INTO evaluations VALUES(18025,34,'cp -348',NULL);
INSERT INTO evaluations VALUES(18025,35,'cp -362',NULL);
INSERT INTO evaluations VALUES(18025,36,'cp -355','d1d4 c7c5 d4c3 c5c4 b3c4 b5c4 c3c4 f8e7 c4d3 c8e6 c2c4 d8c7 b1d2 g8f6 e1g1 f6g4 d3g3 e8g8 h2h3 g4e5 f2f4 e6c4 d2c4 e5c4 b2b3 e7f6 a1b1 c4b6 f1d1 c7c2 g3d3 c2d3 d1d3 f8e8 c1e3 b6c8 e3f2 h7h6 b1d1 e8e4');
INSERT INTO evaluations VALUES(18854,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(18854,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(18854,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(18854,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(18854,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(18854,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(18854,36,'cp -10','g1f3 d7d5 f1e2 f8d6 e1g1 g8e7 g2g3 f7f6 c2c3 e8g8 c1e3 b7b6 a2a3 c8b7 b2b4 d8c7 b1d2 c6d8 e3f2 d8f7 f1e1 a8d8 e2f1 d5e4 d2e4 e6e5 f4e5 d6e5 d1e2 e7f5 b4c5 b6c5 f1g2 e5d6 d3d4 c5d4');
INSERT INTO evaluations VALUES(20276,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20276,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20276,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20276,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(20276,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(20276,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(20276,36,'cp -8','g8f6 e1g1 c5d4 c3d4 f8d6 b1c3 e8g8 f1e1 h7h6 e2e4 f6e4 c3e4 d5e4 e1e4 c8d7 f3e5 d7e8 c1d2 a8c8 d2c3 c6e7 e4e1 b7b6 a1c1 e8b5 d1h5 d6e5 h5e5 e7d5 g2d5 e6d5 e5f5 a7a5 a2a3 c8c4 c1d1 f8e8');
INSERT INTO evaluations VALUES(21546,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(21546,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(21546,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(21546,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21546,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(21546,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(21546,36,'cp -3','d8d1 e1d1 c8d7 c2c3 h7h5 a2a4 h5h4 b1d2 f7f6 d1c2 f8c5 h1f1 a7a6 b5e2 c6d8 d2c4 d8e6 f3e1 c5a7 e1d3 g8h6 c1e3 a7e3 c4e3 a6a5 e3d5 e8c8 d5e7 c8b8 e7g6 h8h7 f1e1 c7c6 g6h4 h6f7 h4f3 g7g5 a1d1 g5g4 f3d2 h7h2');
INSERT INTO evaluations VALUES(1696,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(1696,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(1696,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(1696,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1696,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1696,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1696,36,'cp -7','e1d1 c8d7 c2c3 h7h5 a2a4 h5h4 b1d2 f7f6 d1c2 g8e7 d2b3 e7c8 c1e3 b7b6 b5a6 c8d6 f3d2 c6b8 a6d3 a7a5 d2c4 d7e6 c4d6 f8d6 f2f3 b8d7 d3b5 e8e7 h1d1 d6c5 b5d7 c5e3 d7e6 e7e6 b3d2 e3d2 c2d2 a8d8 d2c2 d8d1 a1d1 g7g6 b2b4 e6e7 b4a5 b6a5');
INSERT INTO evaluations VALUES(21776,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(21776,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(21776,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(21776,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(21776,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(21776,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(21776,36,'cp 31','g8e7 g2g3 e7g6 f1g2 c8g4 b1d2 d8d7 h2h3 g4f5 a2a3 f7f6 e5f6 g7f6 d1a4 e8c8 e1c1 g6f4 g3f4 d7e8 f3e1 f8d6 g2c6 e8c6 a4c6 b7c6 e1d3 d8e8 d1e1 c6c5 c1d1 c8b7 h1g1 f5h3 b2b4 c5b4 a3b4 a7a5 b4a5 h3d7 d2b3 d7a4 d1c2 c7c5');
INSERT INTO evaluations VALUES(25979,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(25979,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(25979,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(25979,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(25979,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(25979,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(25979,36,'cp 116','b1c3 f8e7 e1g1 g8f6 d4d5 e6d5 e4d5 c6b8 d1b3 e8g8 f1e1 b8a6 c1e3 f6g4 e3d4 e7f6 h2h3 f6d4 f3d4 a6c5 b3d1 g4f6 b2b4 c5a6 b5a6 b7a6 d4c6 d8b6 c6e7 g8h8 c3e4 f6g8 e7c6 d7c6 d5c6 b6c6 e4d6 g8f6 d1d4 h8g8 d4c5 c6b6 a1c1 a8d8 d6c4 b6c5 b4c5');
INSERT INTO evaluations VALUES(26263,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(26263,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(26263,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(26263,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(26263,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(26263,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26263,36,'cp 26','f1b5 g8f6 h2h3 g4d7 c2c4 d5c4 b3c4 e7e6 e1g1 f8d6 b1c3 e8g8 b5a4 a7a6 d1e2 c6e5 a1b1 a8b8 f3e5 d6e5 a4d7 d8d7 f1d1 f8d8 d2d4 c5d4 e3d4 e5d6 b2c1 h7h6 c1e3 b7b5 c4b5 a6b5 b1b5 d7c6 b5b8 d8b8');
INSERT INTO evaluations VALUES(26980,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(26980,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(26980,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(26980,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(26980,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(26980,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(26980,36,'cp -44','e5f4 d2d4 g8f6 e1g1 f8e7 d1c2 d6d5 e4d5 f6d5 c2b3 c6a5 b3a4 c7c6 c4d5 d8d5 c1f4 a5c4 a4b3 b7b6 f1e1 g4f3 g2f3 e8d7 b1d2 c4d2 b3d5 c6d5 f4d2 f7f6 d2f4 g7g5 f4g3 h7h5 h2h4 a8g8 g1f2 e7d6 e1g1 d6g3 f2g3 d7e6 a1e1 e6f5 h4g5 g8g5 g3f2 g5g1');
INSERT INTO evaluations VALUES(27005,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(27005,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(27005,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(27005,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(27005,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(27005,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(27005,36,'cp 105','d4d5 c6d4 f3g4 h7h5 g4h3 g7g5 c2c3 g5g4 h3g4 h5g4 c3d4 d8d7 d4e5 d6e5 b1c3 e8c8 d1d3 f8b4 c1g5 f7f6 g5d2 f6f5 e1c1 g8f6 e4f5 h8h5 d3c4 b4c3 d2c3 d7f5 d5d6 c7c6 f2f3 g4f3 g2f3 f6d7 h1g1 h5h2 c3a5 d8f8 g1g7 f5f3 c4e6 f3f4 c1b1 f4f5 e6f5 f8f5');
INSERT INTO evaluations VALUES(26932,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(26932,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(26932,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(26932,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(26932,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(26932,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(26932,36,'cp 103','c6d4 f3g4 h7h5 g4h3 g7g5 c2c3 g5g4 h3g4 h5g4 c3d4 d8d7 d4e5 d6e5 b1c3 e8c8 f2f4 f8d6 f4f5 g4g3 h2h3 d6c5 d1f3 c5f2 e1e2 d7d6 c1g5 g8f6 h3h4 d8g8 a1c1 f6h7 f3d3 h7g5 h4g5 d6d8 d3d2 h8h1 c1h1 g8g5 d5d6 f2d4 c3b5 g5g4 d2d3 c7d6 b5d4 e5d4');
INSERT INTO evaluations VALUES(27082,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(27082,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(27082,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(27082,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(27082,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(27082,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(27082,36,'cp 112','f3g4 h7h5 g4h3 g7g5 c2c3 g5g4 h3g4 h5g4 c3d4 d8d7 d4e5 d6e5 b1c3 e8c8 d1d3 f7f5 e4f5 g8e7 c1g5 h8h5 d3e3 c8b8 h2h4 e7f5 e3e4 g4g3 e1c1 f8h6 f2f4 e5f4 g5d8 d7d8 c1b1 a7a6 d1e1 d8g8 a2a3 f5e3 e1e2 g8f8 c3d1 f8f5 e4f5 e3f5 h1e1 h5h4');
INSERT INTO evaluations VALUES(28700,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(28700,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28700,36,'cp 0','d7d5 e4d5 e6d5 c4b3 f8d6 c3b5 e8g8 e1g1 c7c5 b5d6 d8d6 f1e1 b8c6 d3d4 f8e8 c2c3 e8e1 f3e1 c5d4 c3d4 a8e8 c1e3 c6a5 e1d3 b7a6 h2h3 a6d3 d1d3 a5b3 d3b3 e8c8 b3a4 d6d7 a4d7 f6d7 g2g3 d7f6 a1c1 c8c1 e3c1');
INSERT INTO evaluations VALUES(29373,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(29373,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(29373,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(29373,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(29373,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(29373,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(29373,36,'cp 98','d1b3 b8d7 c4c5 d8c8 c1f4 f8e7 h2h3 e8g8 a1c1 f5e4 f1e2 d7b8 b3d1 b8c6 a2a3 e4f3 e2f3 e7d8 b2b4 a7a6 e1g1 d8c7 f4e3 h7h6 c1b1 f8d8 a3a4 b7b6 b4b5 c6e7 d1d3 e7f5 c3e2 a6b5 a4b5 f5e3 d3e3 a8a2 c5c6 c8a8 f1c1 d8c8 g2g3');
INSERT INTO evaluations VALUES(29489,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(29489,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(29489,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(29489,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(29489,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(29489,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(29489,36,'cp 125','g1f3 e7e6 d1a4 b8d7 d5e6 f7e6 d2d4 c7c5 b1c3 f8d6 d4c5 d6c5 f1e2 e8g8 e1g1 a7a5 a4b3 d8c7 c3b5 c7b8 f3d4 a5a4 b3d1 f8d8 d4f5 e6f5 g2g3 d7e5 d1c2 f6e4 b5c3 e4c3 b2c3 b8c8 c1g5 d8e8 a1d1 g8h8 g5f4 c8e6 f4e5 e6e5 e2f3 a8d8 d1d8');
INSERT INTO evaluations VALUES(30391,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(30391,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(30391,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(30391,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(30391,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(30391,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30391,36,'cp 12','e2e3 e7e6 f1d3 f5g6 e3e4 g8f6 e4d5 f6d5 c3d5 d8d5 g1f3 f8d6 f4d6 c7d6 e1g1 g6h5 d3e2 e8g8 c2c4 d5f5 d4d5 e6d5 d1b3 h5f3 e2f3 d5c4 b3c4 b8c6 f3c6 b7c6 c4c6 f8d8 h2h3 d6d5 f1d1 f5e6 c6a4 h7h6 d1d2 a6a5 a1d1 e6b6 d2d5 d8d5 d1d5');
INSERT INTO evaluations VALUES(30390,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(30390,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(30390,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(30390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30390,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(30390,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(30390,36,'cp 0','e7e6 f1d3 f5g6 e3e4 g8f6 e4d5 f6d5 c3d5 d8d5 g1f3 f8d6 f4d6 c7d6 e1g1 g6h5 d3e2 e8g8 c2c4 d5f5 d1d2 f5f6 a1d1 b8c6 f1e1 a8c8 f3g5 h5e2 g5e4 f6d8 d2e2 d6d5 e4c3 d8d7 c4d5 e6d5 e2f3 c6e7 f3e2');
INSERT INTO evaluations VALUES(30401,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30401,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30401,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30401,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30401,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30401,36,'cp 0','f1d3 f5g6 h2h4 g8f6 h4h5 g6d3 d1d3 h7h6 e3e4 c7c5 g1e2 c5d4 e2d4 f8b4 e4d5 f6d5 f4e5 e8g8 e5g7 g8g7 d4e6 f7e6 d3g6 g7h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(33359,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(33359,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(33359,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(33359,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(33359,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(33359,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(33359,36,'cp 136','c3b5 e8d8 c2c4 b8c6 d4c5 e7c5 c4d5 c5d5 d1c1 d8e7 f1e2 a7a6 c1g5 e7f8 a1d1 h7h6 g5f4 g7g5 f4a4 d5c5 b5d6 d7b6 a4g4 c5b4 g4b4 c6b4 h2h4 g5g4 f3d4 b6d7 d4e6 f7e6 d6c8 f8e8 d1d7 e8d7 c8b6 d7e7 b6a8 h8a8 e2g4 b4d3 e1e2 d3e5 g4h5 a8c8 h1d1 a6a5 b2b3 b7b5 d1d2 b5b4 f2f4 e5d7 h5f3');
INSERT INTO evaluations VALUES(33356,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(33356,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(33356,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(33356,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(33356,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(33356,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(33356,36,'cp 137','e8d8 c2c4 b8c6 d4c5 e7c5 c4d5 c5d5 d1c1 d8e7 f1e2 a7a6 c1g5 e7f8 a1d1 h7h6 g5e3 d5c5 b5d4 f8g8 d4c6 c5e3 c6e7 g8f8 e7g6 f7g6 f2e3 f8e7 e1g1 g6g5 d1d4 b7b5 f1c1 a8a7 g1f2 h8f8 f2e1 f8d8 c1c6 d7f8 d4d6 a7d7 f3d4 c8b7 c6a6 b7a6 d6a6 d7d4 e3d4 d8d4 a6a7 d4d7 a7d7 f8d7');
INSERT INTO evaluations VALUES(37144,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(37144,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(37144,32,'cp 199',NULL);
INSERT INTO evaluations VALUES(37144,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(37144,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(37144,35,'cp 236',NULL);
INSERT INTO evaluations VALUES(37144,36,'cp 219','e5f7 d8e7 f7e5 g8h8 e1g1 d7d6 e5f3 c8g4 c2c3 e7e4 b1d2 e4g6 c4e2 c7c5 h2h3 g4f3 e2f3 d6d5 d4c5 b6c5 d2b3 c5b6 f1e1 b8c6 c1e3 b6c7 b3d4 f8e8 d1b3 c6a5 b3b4 c7b6 a2a4 a8d8 b4b5 a5c4 a4a5 b6c7 b5b7 c7b8 a5a6 c4e3 e1e3 e8e3 f2e3');
INSERT INTO evaluations VALUES(37724,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(37724,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(37724,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(37724,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(37724,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(37724,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(37724,36,'cp 47','b1c3 d7d6 e1g1 c6c5 d2d4 c5d4 f3d4 g8f6 h2h3 e8g8 c1e3 b8c6 d1d2 d8b6 d4b3 b6c7 a1d1 c6e5 c4e2 a7a6 a2a4 b7b6 f1e1 c8b7 f2f3 a8c8 b3d4 c7c5 e2f1 e5c4 f1c4 c5c4 e3h6');
INSERT INTO evaluations VALUES(37723,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(37723,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(37723,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(37723,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(37723,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(37723,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(37723,36,'cp 47','d7d6 d2d4 g8f6 c4b3 b7b5 e4e5 d6e5 f3e5 e8g8 e1g1 a7a5 a2a3 c8b7 f1e1 b8d7 d1f3 d7e5 d4e5 f6d5 b3d5 c6d5 c3b5 d8b6 f3d3 b7a6 a3a4 a8b8 d3d5 f8d8 d5f3 a6b5 a4b5 b6b5 e5e6 f7e6 c2c3 d8d5 f3g4 b5d7 h2h4 b8d8 g4e6 d7e6 e1e6 d5d1 g1h2 g7f6 e6e4 d8a8 g2g3 g8f7 b2b3 a8d8 h2g2');
INSERT INTO evaluations VALUES(37933,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(37933,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(37933,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(37933,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(37933,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(37933,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(37933,36,'cp 72','e1g1 g8f6 f1e1 d6d5 e4d5 c6d5 c4f1 e8g8 c2c3 b8c6 h2h3 d8c7 f3e5 c8f5 g2g4 f5d7 e5d3 h7h5 c1f4 c7b6 f1e2 h5g4 h3g4 c6d8 a2a4 d8e6 f4h2 a7a5 b1a3 f6e4 a3c2 d7c6 c2e3 a8e8 f2f3 e4d6 g1g2');
INSERT INTO evaluations VALUES(38571,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(38571,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(38571,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(38571,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(38571,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(38571,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(38571,36,'cp 152','h2h4 d5e4 c3e4 h7h6 f1d3 b8d7 d1e2 g8f6 c2c3 f6e4 d3e4 c7c5 d4c5 d7c5 e2b5 c5d7 c1e3 a7a6 b5b4 a6a5 b4d6 a5a4 a2a3 d8e7 d6e7 e8e7 e1c1 h8d8 e4c2 b7b6 c2e4 a8a7 f3d4 g7d4 d1d4 d7c5 d4d8 e7d8 e4c2 c8b7 h1g1 h6h5 g2g4 h5g4 g1g4 b7f3 g4f4');
INSERT INTO evaluations VALUES(38979,30,'cp 851',NULL);
INSERT INTO evaluations VALUES(38979,31,'cp 855',NULL);
INSERT INTO evaluations VALUES(38979,32,'cp 851',NULL);
INSERT INTO evaluations VALUES(38979,33,'cp 855',NULL);
INSERT INTO evaluations VALUES(38979,34,'cp 863',NULL);
INSERT INTO evaluations VALUES(38979,35,'cp 863',NULL);
INSERT INTO evaluations VALUES(38979,36,'cp 863','h5e5 e8f8 e5h8 d8f6 h8f6 g8f6 b1c3 b8c6 g1e2 c6a5 d2d4 a5c4 d4c5 b7b6 b2b3 c4a5 c1f4 b6c5 f4c7 a5b7 e1c1 f6e8 c7f4 h7h5 e4e5 c5c4 b3b4 e8c7 c3e4 a7a5 e4f6 b7d8 d1d7 c8d7 f6d7 f8g7 d7b6 a5b4 b6a8 c7a8 c2c3 b4c3 e2c3');
INSERT INTO evaluations VALUES(38978,30,'cp 851',NULL);
INSERT INTO evaluations VALUES(38978,31,'cp 840',NULL);
INSERT INTO evaluations VALUES(38978,32,'cp 840',NULL);
INSERT INTO evaluations VALUES(38978,33,'cp 848',NULL);
INSERT INTO evaluations VALUES(38978,34,'cp 840',NULL);
INSERT INTO evaluations VALUES(38978,35,'cp 844',NULL);
INSERT INTO evaluations VALUES(38978,36,'cp 848','e8f8 e5h8 d8f6 h8f6 g8f6 b1c3 b8c6 g1e2 f8g7 c4b3 d7d6 d2d3 c8g4 f2f3 g4e6 c1g5 h7h6 g5f6 g7f6 e1d2 h6h5 a1f1 a7a5 f3f4 c6d4 b3e6 f7e6 e4e5 f6e7 c3e4 d4e2 d2e2 c5b6 a2a4 d6d5 e4g5 a8f8 h2h3 h5h4 c2c3 c7c6 d3d4');
INSERT INTO evaluations VALUES(42075,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(42075,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(42075,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(42075,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(42075,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(42075,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(42075,36,'cp 104','c4d5 e6d5 c1a3 b8c6 e2e3 c6e7 f3e5 e8g8 f1d3 f8e8 e1g1 e7g6 f2f4 b7b6 d3b5 c8b7 b5e8 d8e8 e5g4 f6g4 d1g4 e8e3 f1f2 h7h5 g4h5 g6f4 h5f3 e3f3 f2f3 f4e6 a1e1 b7a6 e1e5 a6c4 f3f2 a8e8 h2h4 f7f6 e5e3 g8f7 g2g4 c7c5 d4c5 b6c5 g4g5 e6g5 e3e8 g5h3 g1g2 h3f2');
INSERT INTO evaluations VALUES(41986,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(41986,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(41986,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(41986,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(41986,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(41986,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(41986,36,'cp 103','d8d5 e2e3 b7b6 f1e2 c8b7 e1g1 c7c5 c3c4 d5c6 d1b3 b8d7 a2a4 a7a5 f1d1 e8g8 c1b2 a8d8 d1e1 c6c7 f3d2 d8c8 d2f1 f8d8 h2h3 f6e4 a1c1 h7h6 e1d1 d7f6 e2f3 e4g5 f3b7 c7b7');
INSERT INTO evaluations VALUES(42553,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(42553,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(42553,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(42553,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(42553,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42553,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(42553,36,'cp 19','d2d4 e5d4 e4e5 d7d5 c4b5 c8d7 b5d7 f6d7 c3d4 c5b4 b1c3 c7c5 g1f3 c5d4 d1d4 b8c6 d4g4 d5d4 a2a3 b4f8 f3d4 d7e5 d4c6 e5g4 c6d8 a8d8 e1g1 f7f6 c1f4 f8d6 f4d6 d8d6 a1d1 d6d1 f1d1 e8e7 c3b5 a7a6 b5d6 g7g6 d6b7 h8b8 b7c5 b8b2 h2h3 b2c2 d1d7 e7e8 h3g4 c2c5 d7h7');
INSERT INTO evaluations VALUES(46836,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(46836,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(46836,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(46836,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(46836,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(46836,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(46836,36,'cp 37','d5c3 b2c3 c7c5 e2e4 f8g7 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 f1d3 c8g4 d2a5 c6a5 h2h3 g4f3 g2f3 e7e6 e4e5 f8d8 d3e4 a8c8 f3f4 d8d7 e1g1 b7b6 f1d1 a5c4 g1f1 g7f8 f1e2 f8e7 a2a4 c4e3 f2e3 c8c3 e2f3 f7f5 e5f6 e7f6 e4d3 d7c7');
INSERT INTO evaluations VALUES(47007,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47007,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47007,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47007,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47007,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47007,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47007,36,'cp 0','g1f3 d6d5 e2e3 c8f5 c2c4 e7e6 d1b3 a7a5 a2a3 d8c8 b1c3 c7c6 f3h4 f5e4 h4f3');
INSERT INTO evaluations VALUES(47006,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47006,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47006,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47006,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47006,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47006,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47006,36,'cp -13','d6d5 e2e3 c8f5 c2c4 e7e6 d1b3 a7a5 a2a3 d8c8 b1c3 c7c6 f3h4 f5e4 h4f3 e4g6 f3e5 b8d7 e5g6 h7g6 c1d2 f8e7 f1e2 d5c4 b3c4 b7b5 c4d3 b5b4 c3e4 e8g8 e4f6 g7f6 a3b4 a5b4 a1a8 c8a8 e1g1 f8b8 d2e1');
INSERT INTO evaluations VALUES(49082,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(49082,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(49082,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(49082,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(49082,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(49082,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(49082,36,'cp 131','b8d7 e5c4 f8e7 d2d4 e8g8 f1d3 c7c5 e1g1 d7f6 c4e5 c8e6 f1e1 a8c8 a1b1 b7b6 c1g5 c5c4 d3e2 f6e4 g5d2 e6d5 e2f3 f7f5 b1b5 d5b7 f3e4 b7e4 d2f4 e7d6 f2f3 e4b7 d1d2 d8c7 a2a4 c8e8 f4g3');
INSERT INTO evaluations VALUES(51548,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(51548,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(51548,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(51548,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(51548,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(51548,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(51548,36,'cp 94','c8b7 f1d3 g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 f8e7 b1a3 c5c4 d3c2 a7a6 c1d2 b8c6 c2b1 b6b5 a3c2 d7b6 e2d1 e8d7 d1c1 h7h6 h2h3 a6a5 a2a3 d7c8 f3h2 c8b8 c2e3 d8c7 b1c2 h6h5 f2f4 g7g6');
INSERT INTO evaluations VALUES(30488,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(30488,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(30488,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(30488,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(30488,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(30488,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(30488,36,'cp 109','f1d3 g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 b8c6 h2h4 f8e7 h4h5 h7h6 c1f4 c5c4 d3c2 b6b5 a2a3 a7a5 f3h2 b5b4 e2g4 e7g5 f4g3 b4b3 c2d1 g5c1 g4g7 h8f8 d1b3 c1b2 a1a2 b2c3 b3a4 c3d4 h2f3 d4c5 g7h6 f8g8 g3h4 d8b6 b1c3');
INSERT INTO evaluations VALUES(51566,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(51566,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(51566,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(51566,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(51566,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(51566,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(51566,36,'cp 4','g8f6 g1f3 f6e4 e1g1 d7d5 f3e5 f8d6 d2d4 e8g8 b1d2 e4d2 c1d2 d8c7 f1e1 b8d7 d2f4 d7e5 d4e5 d6c5 c2c3 c8f5 b3c2 f5c2 d1c2 a7a5 f4e3 c5b6 a2a4 f8e8 e3b6 c7b6 a4b5 c6b5 a1d1');
INSERT INTO evaluations VALUES(52132,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(52132,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(52132,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(52132,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(52132,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(52132,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52132,36,'cp 35','f1c4 g8f6 f3b3 d8e7 d2d3 b8c6 g1e2 c6a5 b3c3 a5c4 c3c4 d7d6 c4b3 c8e6 c2c4 g7g6 b3b5 e7d7 a2a4 a7a6 b5d7 f6d7 b1c3 f8e7 h2h4 h7h5 c3d5 e6d5 e4d5 f7f5 e2c3 a8b8 g2g3 e8g8 a4a5 g8g7 c1g5 e7g5');
INSERT INTO evaluations VALUES(53184,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53184,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53184,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(53184,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(53184,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(53184,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53184,36,'cp 37','b1d2 b8c6 f2f4 g8h6 g1f3 e7e6 d2e4 f8e7 h2h3 h6f5 g2g4 f5h4 d1e2 h4f3 e2f3 c6b4 f3e2 c8d7 f1g2 d7a4 b2b3 a4c6 e1g1 d8c7 f4f5 b4d5 f5f6 g7f6 e5f6 e7d6 e4d6 c7d6 c1d2 e8c8 a1e1 d8g8 e2e5 c8d7 g1h2 h7h6 g2d5 d6d5 e5d5 c6d5 e1e5 g8g6 e5h5 d7d6 a2a4 b7b6 a4a5');
INSERT INTO evaluations VALUES(53639,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(53639,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(53639,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(53639,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(53639,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(53639,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(53639,36,'cp 15','c6c5 d2d4 g8h6 c2c3 c5d4 c3d4 h6f5 b1a3 f8a3 b2a3 b8c6 a1b1 f5h4 f3h4 d8h4 g2g3 h4e7 c1b2 c8d7 f1d3 a8c8 d1d2 e8g8 e1g1 b7b6 b1c1 g7g6 h2h4 c6a5 b2c3 a5c6 c3b4 c6b4 a3b4 c8c1 f1c1 f8c8 c1c8 d7c8');
INSERT INTO evaluations VALUES(54098,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(54098,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(54098,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(54098,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(54098,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(54098,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(54098,36,'cp 21','f1c4 a7a6 e1g1 g8f6 c4e2 c5d4 f3d4 f8c5 b2b3 d8e7 c1b2 e8g8 b1d2 e6e5 d4f3 b8c6 d1c2 c8g4 a1c1 c5a7 f1d1 a8c8 c2b1 f8d8 d2e4 f6e4 b1e4 g4h5 e4f5 h5g6 f5h3 a7b6 f3h4 g6e4 h4f5');
INSERT INTO evaluations VALUES(54926,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(54926,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(54926,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(54926,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(54926,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(54926,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(54926,36,'cp 50','e2e4 d5e4 c3e4 c8f5 e4g3 f5g6 g1f3 b8d7 f1d3 g6d3 d1d3 g8f6 d3b3 a8b8 e1g1 e7e6 c2c4 f8d6 d4d5 e8g8 d5e6 f7e6 c1e3 b7b6 a1d1 d8e8 b3c2 b8d8 h2h3 d7e5 f3e5 d6e5 d1d8 e8d8 g3e2 f8e8 f1d1 d8e7 e2f4 e5f4 e3f4 e6e5 f4e3 e7f7 c2c3 f7g6 c3b3');
INSERT INTO evaluations VALUES(54931,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(54931,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(54931,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(54931,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(54931,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(54931,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(54931,36,'cp 53','d5e4 c3e4 c8f5 e4g3 f5h7 g1f3 g8f6 f1d3 h7d3 d1d3 c7c6 c1f4 e7e6 e1c1 b8d7 g3e4 f6e4 d3e4 d7f6 e4e2 f8e7 h1e1 e8g8 c1b1 d8d5 f3e5 b7b5 g2g4 f6d7 e1g1 d7e5 d4e5 d5c5 f4d2 f8d8 f2f4 a7a5 f4f5 e7g5 d2g5 d8d1 g1d1 h6g5 f5e6');
INSERT INTO evaluations VALUES(56123,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(56123,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(56123,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(56123,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(56123,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(56123,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(56123,36,'cp 40','g8f6 b1c3 f6d5 e4d5 c7c6 d2d4 c6d5 g1e2 b8c6 e1g1 e5d4 e2d4 f8e7 c1e3 e8g8 d1h5 c8e6 a1d1 c6d4 e3d4 d8d7 h2h3 f8e8 d4e3 e8d8 h5f3 e7b4 e3d4 f7f6 c3e2 b4d6 e2f4 d6f4 f3f4 d8e8 f1e1 e6f5 c2c3 f5e4');
INSERT INTO evaluations VALUES(1793,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(1793,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(1793,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(1793,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(1793,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(1793,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(1793,36,'cp 32','d7c6 d1e2 c8e6 e1g1 e8c8 b1d2 e6c4 e2c4 d8e8 f2f4 f6e6 c4e6 e8e6 g2g3 g7g5 g1g2 g5f4 g3f4 f7f5 e4e5 g8e7 g2f3 h8g8 d2c4 e7g6 f3e2 g6h4 c1d2 e6g6 c4e3 f8c5 a2a4 h6h5 f1f2 g6g1 a1g1 g8g1 a4a5 h4g6');
INSERT INTO evaluations VALUES(6263,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(6263,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(6263,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(6263,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(6263,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(6263,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(6263,36,'cp 156','c2c3 b4d6 d3d4 g8f6 e1g1 e8g8 f1e1 f8e8 d1b3 e8e7 c4d3 a7a6 b1d2 b7b5 d3b1 c8b7 a2a3 e7e8 b1a2 d8e7 h2h3 c6a5 b3d1 a8c8 a3a4 c8a8 b2b4 a5c4 a2c4 b5c4 d2c4 e5d4 e4e5 e7e6 d1d4 b7f3 g2f3 d6f8 c1f4 f6h5 f4h2 e6h3');
INSERT INTO evaluations VALUES(6262,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(6262,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(6262,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(6262,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(6262,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(6262,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(6262,36,'cp 141','b4d6 d3d4 g8f6 e1g1 e8g8 f1e1 a7a6 a2a3 f8e8 d1b3 d8e7 b3c2 b7b5 c4a2 a6a5 b1d2 a5a4 d2f1 c8b7 f1g3 c6a5 f3e5 d6e5 g3f5 e7f8 d4e5 e8e5 f2f3 b7d5 c1f4 e5f5 e4f5 d5a2 a1a2 a5b3 c2f2 f6d5 f4g3 c7c6 h2h3 f7f6 g1h2 a8d8 f2c2 d5b6 c2e4 f8f7 g3d6');
INSERT INTO evaluations VALUES(11825,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(11825,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(11825,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(11825,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(11825,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(11825,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(11825,36,'cp -87','e5e4 f3e5 c6d5 d2d4 f8d6 e1g1 e8g8 c1g5 c7c6 e5c6 d8b6 c6e5 f6e8 c3d5 b6b7 c2c4 f7f6 g5f4 f6e5 d4e5 c8e6 f4g3 d6c5 b2b4 c5b6 a2a4 a8d8 a4a5 b6c7 d1c2 e6d5 c4d5 d8d5 c2e4 d5b5 e4c4 b7d5 c4d5 b5d5 f2f4 c7b8 a1c1 e8c7 g3e1 d5b5 g2g3 f8d8 g1f2 d8d3 f2g2 c7d5 g2h3 d5e7 c1c5 b5c5 b4c5 d3d5');
INSERT INTO evaluations VALUES(12209,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(12209,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12209,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(12209,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(12209,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12209,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(12209,36,'cp 25','b1d2 f8e7 h2h3 d8b6 d1b3 e8g8 a2a4 b6b3 d2b3 f6e4 h3h4 f7f6 b3d2 e4d2 f3d2 c5d4 e3d4 e6e5 f4e3 d7b8 d4e5 f6e5 g2g3 b8c6 f1g2 c8e6 e1g1 a8d8 f1e1 h7h6 a4a5 d8c8 d2b3');
INSERT INTO evaluations VALUES(13524,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(13524,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(13524,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(13524,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(13524,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(13524,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(13524,36,'cp 82','c4d5 e6d5 g5f6 g7f6 e2e3 c5c4 a2a3 c8e6 g2g3 b7b5 f1g2 c6e7 e1g1 a8b8 f3h4 f8g7 d1h5 e8g8 g3g4 d8d7 h2h3 a7a5 f1d1 d7d8 h4f5 e6f5 g4f5 d8d7 g2f3 g8h8 f3g4 e7g8 g1h1 b5b4 a3b4 a5b4 c3a4 d7e7 a4c5 g7h6 g4f3 b8d8 a1a6 c4c3 a6c6 c3b2');
INSERT INTO evaluations VALUES(16351,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(16351,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(16351,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16351,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16351,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16351,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(16351,36,'cp 17','f6e4 f1d3 d7d5 e5d6 e4d6 e1g1 f8e7 c1f4 c6b4 f4d6 d8d6 d3c4 b4d5 d1d3 c8e6 b1d2 e8g8 a1e1 c7c6 d2e4 d6d8 c4b3 d5f4 d3d1 e6b3 a2b3 a7a5 c2c3 f4g6 f3d2 d8d7 d1c2 b7b5 d2f3 d7c7 e4c5');
INSERT INTO evaluations VALUES(17911,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17911,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17911,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(17911,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(17911,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17911,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(17911,36,'cp 36','f1c4 d7d6 b1c3 f8e7 d2d4 c6d4 f3d4 e5d4 d1d4 g8f6 c1f4 b7b5 c4d5 f6d5 c3d5 e8g8 e1c1 c8b7 f4d2 c7c5 d4d3 e7f6 d5f6 d8f6 f2f3 f8d8 h1e1 b7c6 d2c3 f6e6 d3d2 f7f6 b2b3 e6f7 d2f4 f7e6 h2h4 a6a5 g2g4 a5a4 b3a4');
INSERT INTO evaluations VALUES(18868,30,'cp -158',NULL);
INSERT INTO evaluations VALUES(18868,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(18868,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(18868,33,'cp -166',NULL);
INSERT INTO evaluations VALUES(18868,34,'cp -156',NULL);
INSERT INTO evaluations VALUES(18868,35,'cp -171',NULL);
INSERT INTO evaluations VALUES(18868,36,'cp -158','d8a5 b1c3 a5e5 c1e3 g8f6 f1e2 d7d5 d1d2 e5b8 f2f4 f8b4 a2a3 f6e4 d4c6 b7c6 d2d4 b4c3 b2c3 b8b2 e1g1 b2c3 e2d3 c3d4 e3d4 e8g8 d3e4 d5e4 d4c5 f8d8 f1d1 d8d5 c2c4 d5d1 a1d1 c8a6 d1d4 f7f5 d4d6 a6c4 d6c6 c4d5 c6c7 a7a5 c7a7 a8a7 c5a7');
INSERT INTO evaluations VALUES(19625,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(19625,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(19625,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(19625,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(19625,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(19625,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(19625,36,'cp 89','f1c4 h7h6 d5c6 e7c6 d1e2 g8f6 e1g1 f8e7 c3d5 c8e6 f1d1 e8g8 h2h3 a8c8 c2c3 a7a6 d5f6 e7f6 c4b3 d8d7 f3h2 f6e7 h2f1 e7d8 e2d2 d8c7 b3d5 f7f5 e4f5 f8f5 f1e3 f5f7 d2d3 e6d5');
INSERT INTO evaluations VALUES(23127,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(23127,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(23127,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(23127,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(23127,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(23127,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(23127,36,'cp 83','d8e7 e1g1 g7g6 h5d1 f8g7 c2c3 c7c6 d2d3 d7d6 f1e1 d4c3 b2c3 c8e6 c4b3 e7d7 d3d4 e6b3 a2b3 g8e7 h2h4 h7h5 c1f4 e8g8 g2g3 f8e8 b3b4 e7c8 b1d2 b7b5 a1a6 c8b6 d2f3 d7b7 a6a1 b6c4 f3d2 c4b6 d1b3 b7e7 d2f3');
INSERT INTO evaluations VALUES(24085,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(24085,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(24085,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(24085,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(24085,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(24085,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(24085,36,'cp -25','g7g6 d1f3 e8g8 e1e6 f7e6 f6f7 g8h8 g5e6 d8e7 c1g5 e7f7 e6c5 f7f3 g2f3 f8f3 c2c3 a8e8 b1d2 f3f5 c5e4 h8g7 c3d4 c6d4 g1g2 e8e5 h2h4 h7h6 g5e3 d4c2 a1c1 c2e3 f2e3 b7b5 b2b3 c4b3 a2b3 c7c5 c1a1 e5e7 g2g3 e7d7 d2f3 d7d3 a1a7 g7f8 a7a8 f8e7 e4d2 d3e3 g3f2 e3c3 f2e2 f5f4 a8a7 e7f6 a7b7 c5c4 b7b5 c4b3 b5b3 c3b3');
INSERT INTO evaluations VALUES(25609,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(25609,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(25609,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(25609,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(25609,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(25609,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(25609,36,'cp 14','b5a4 b7b5 a4b3 a6a5 a2a3 c5c4 b3c2 g8e7 f3g5 h7h6 g5h7 b5b4 b1d2 b4b3 c2b1 d8c7 e1g1 e7g6 h7f8 h8f8 f2f4 c6e7 d2f3 f8g8 d1e1 e8c8 b1g6 e7g6 g1h1 f7f5 e5f6 g7f6 f4f5 e6f5 f3h4 c8b7 h4f5 d7f5 f1f5');
INSERT INTO evaluations VALUES(25610,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(25610,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(25610,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(25610,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(25610,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(25610,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(25610,36,'cp 21','b7b5 a4b3 a6a5 e1g1 a5a4 b3c2 c5d4 c3d4 c6b4 b1a3 b4c2 a3c2 g8e7 c2b4 e7f5 c1d2 a8c8 a1c1 c8c4 b4c2 f8e7 b2b3 c4c6 h2h3 e8g8 g2g4 f5h4 f3h4 e7h4 d2b4 h4g5 f2f4 g5e7 d1d3 f7f5 d3g3 f5g4 h3g4 c6c8 b4e7 d8e7 c2e3 b5b4');
INSERT INTO evaluations VALUES(27550,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(27550,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(27550,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(27550,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(27550,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(27550,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(27550,36,'cp 46','c3d4 g8e7 b1c3 e7f5 c3a4 b6a5 c1d2 f8b4 d2b4 a5b4 a2a3 b4e7 a1c1 g7g5 a4c5 g5g4 f3e1 c6d4 c5d7 d4e2 d1e2 e8d7 e2g4 e7h4 g4e2 f5d4 e2d3 a8c8 c1d1 d4c6 e1f3 h4g4 h2h3 g4g6 d3e3 d7c7 d1d2 c7b8 f1c1 g6e4 e3e4 d5e4');
INSERT INTO evaluations VALUES(27850,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(27850,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(27850,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(27850,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(27850,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(27850,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(27850,36,'cp 26','g4f3 e2f3 f6d7 c1e3 b8c6 d4d5 c6a5 f3e2 g7c3 b2c3 e7e5 d5e6 f7e6 e1g1 f8f7 f2f4 d8e7 e2d3 a8f8 d1d2 d7c5 e3c5 d6c5 f1f2 e6e5 f4f5 b7b6 d2h6 e7f6 a1f1 a5b7 f2f3 g6f5 h6f6 f7f6 e4f5 b7d6 f1e1 e5e4 d3e4 d6e4 e1e4 f6f5 f3f5');
INSERT INTO evaluations VALUES(27857,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(27857,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(27857,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(27857,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(27857,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(27857,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(27857,36,'cp 43','e2f3 f6d7 c1e3 b8c6 d4d5 c6a5 f3e2 g7c3 b2c3 e7e5 e3h6 d8e7 h3h4 d7c5 f2f3 f8b8 d1d2 f7f6 g2g4 e7f7 e1f2 g8h8 a1g1 b8g8 f2g2 b7b6 g2h2 a5b7 e2d1 a7a5 d1c2 a5a4 g1g2 c5d7 d2e2 f7e7 h1g1');
INSERT INTO evaluations VALUES(27856,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(27856,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27856,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(27856,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(27856,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(27856,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(27856,36,'cp 51','f6d7 c1e3 b8c6 d4d5 c6a5 f3e2 g7c3 b2c3 e7e5 h3h4 b7b6 e3h6 d7c5 h6f8 d8f8 d1c2 h7h5 f2f3 g8g7 e1c1 f8e7 d1g1 a5b7 c2d2 c5d7 c1b1 b7c5 g2g3 a8f8 e2d1 f8h8 d2g5 d7f6 b1a1 e7d8 g5e3 a7a5 g1f1 a5a4 a1b2');
INSERT INTO evaluations VALUES(29192,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(29192,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(29192,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(29192,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(29192,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(29192,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(29192,36,'cp 75','d1b3 d8c8 c4d5 e6d5 a1c1 a7a5 f3e5 b8d7 a2a3 f8e7 e5d7 f5d7 c3a4 a8a7 f1d3 e8g8 b3b6 a7a8 e1g1 d7e6 b6b3 f6d7 d3b1 h6h5 h2h3 g7g6 b1d3 e7f6 b3c2 f6g7 f4g5 f8e8 f2f4 b7b5 a4c3 b5b4 c3e2');
INSERT INTO evaluations VALUES(29474,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(29474,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(29474,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(29474,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(29474,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(29474,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(29474,36,'cp 222','e2e4 c7c6 f3g5 e6g4 f2f3 g4h5 f1c4 h7h6 g5h3 g7g5 g2g4 h5g6 h3f2 e7e5 d4e5 d8d1 c3d1 f6d7 e5e6 d7e5 e6f7 g6f7 c4f7 e8f7 e1e2 c6c5 d1e3 b8c6 e3f5 c6d4 f5d4 c5d4 c1d2 a8c8 a1c1 f8e7 f2d3 e5g6 c1c8 h8c8 h1c1 c8c1 d3c1');
INSERT INTO evaluations VALUES(29990,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(29990,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(29990,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(29990,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(29990,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(29990,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(29990,36,'cp 86','h2h3 g4d7 c1e3 b8c6 d1d2 d8f6 e1c1 f8e7 f3d4 e8c8 g2g4 c6d4 e3d4 f6g6 f1d3 f7f5 g4f5 g6f7 h1g1 e7f6 d2f4 h8f8 b2b3 d8e8 c1b2 c8b8 d3c4 f7e7 g1e1 e7e1 d1e1 e8e1 f4g4 e1e7 c4e6 d7e6 f5e6 f6d4 c3d4 f8f2');
INSERT INTO evaluations VALUES(30316,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(30316,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(30316,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(30316,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(30316,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(30316,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(30316,36,'cp 48','g1f3 e7e6 a2a3 b8c6 b1d2 a7a5 b2b3 c6a7 e2e3 a8c8 c1b2 f8e7 c4d3 d7b5 d2c4 a5a4 f1e2 c7c5 e1g1 e8g8 f1c1 h7h6 a1b1 a4b3 d3b3 b5a6 a3a4 f6d7 b2a3 b7b5 a4b5 a7b5 b3a4 b5c7 d4c5 d7c5 a4a5 c7d5');
INSERT INTO evaluations VALUES(30315,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(30315,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(30315,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(30315,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(30315,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(30315,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(30315,36,'cp 44','e7e6 a2a3 f8d6 g2g3 d7c6 f1g2 b8d7 e1g1 e8g8 b1c3 f6d5 f1e1 d7b6 c4d3 d5c3 b2c3 f7f5 c3c4 c6e4 d3b3 c7c5 c1b2 a8c8 a3a4 b6d7 e2e3 d8b6 b3b5 b6c6 a4a5 h7h6 b5c6 c8c6 b2a3 f8d8 f3d2 e4g2 g1g2 c5d4 e3d4 d6a3 a1a3 e6e5 d4e5 d7c5');
INSERT INTO evaluations VALUES(33384,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(33384,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(33384,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(33384,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(33384,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(33384,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(33384,36,'cp -7','d2c3 b8c6 c1d2 c8d7 e1c1 h7h6 d1e1 e8c8 f3d4 h8e8 d4c6 d7c6 f1c4 d6d5 c4d3 b7b6 h2h4 c8b7 f2f3 a7a5 h4h5 e7f6 e1e8 d8e8 h1e1 b6b5 a2a3 e8e1 d2e1 f6e5 d3f5 c6e8 e1f2 e5f4 c1d1 b7c6 f5d3 a5a4');
INSERT INTO evaluations VALUES(33383,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(33383,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(33383,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(33383,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(33383,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(33383,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(33383,36,'cp -12','b8c6 c1d2 c8d7 e1c1 h7h6 d1e1 e8c8 f3d4 d8e8 d4c6 d7c6 f2f3 h6h5 c3c4 h5h4 h2h3 h8h5 f1d3 c8d7 c1d1 g7g6 b2b3 e7g5 f3f4 e8e1 h1e1 g5e7 e1e2 e7f6 d3e4 a7a6 e2e1 b7b5 e4c6 d7c6 e1e8 b5c4 b3c4 c6d7 e8b8');
INSERT INTO evaluations VALUES(34084,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(34084,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(34084,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(34084,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(34084,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(34084,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(34084,36,'cp 103','b8d7 c4d3 e7e6 d1e2 b7b5 e1g1 c8b7 a2a3 a7a6 b2b4 g7g6 c3e4 f6e4 d3e4 f8g7 c2c4 e8g8 c4c5 a6a5 h2h4 e6e5 d4e5 d7e5 d2f4 f8e8 a1d1 c7e7 f1e1 e5c4 e2c2 c4a3 c2c1 a3c4 f4g5 g7f6');
INSERT INTO evaluations VALUES(35075,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(35075,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(35075,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(35075,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(35075,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(35075,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(35075,36,'cp 92','g1f3 f6d8 c2c4 d7d5 f1d3 c7c5 e1g1 g8f6 d1e2 b8c6 b1c3 c5d4 e3d4 d5c4 d3c4 f8e7 f1d1 e8g8 d4d5 e6d5 c3d5 f6d5 c4d5 d8e8 d1e1 c8g4 h2h3 g4f3 e2f3 e8d7 c1e3 a8d8 a1d1 e7f6 d5c6 d7c6 f3c6 b7c6 b2b3 a7a5 d1c1 a5a4 c1c6 a4b3 a2b3 f8e8 b3b4');
INSERT INTO evaluations VALUES(35076,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(35076,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(35076,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(35076,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(35076,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(35076,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(35076,36,'cp 90','d7d5 c2c4 f6d8 b1c3 g8f6 f1d3 c7c5 e1g1 d5c4 d3c4 c5d4 e3d4 f8e7 d1e2 e8g8 f1d1 f6d5 c1d2 b8c6 a1c1 e7d6 c4b3 c6e7 e2d3 d6f4 c3d5 e7d5 d2f4 d5f4 d3e4 f4d5 b3d5 e6d5 e4f4 c8d7 h2h4 d7c6 c1c5 h7h6 d1e1 f7f6');
INSERT INTO evaluations VALUES(37215,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(37215,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(37215,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(37215,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(37215,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(37215,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(37215,36,'cp -18','c5e7 b1c3 c7c6 a2a4 b8a6 c4a6 b7a6 g1e2 h7h6 g5f6 e7f6 e1g1 f8e8 c3d1 d7d5 d1e3 c8b7 a1d1 a8b8 b2b3 g7g6 g1h1 a6a5 e2g1 f6g7 f3g3 b7c8 d1e1 b8b4 f2f3 b4b7 g3f2 a7a6 g1e2 h6h5 g2g4 h5g4 f3g4');
INSERT INTO evaluations VALUES(37321,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(37321,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(37321,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(37321,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(37321,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(37321,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(37321,36,'cp -4','b1c3 c7c6 a2a4 b8a6 c4a6 b7a6 g1e2 h7h6 g5c1 d7d5 h2h3 f8e8 f3g3 g8h7 e1g1 a8b8 g1h1 b8b7 f2f3 e7d6 g3h4 a6a5 c3d1 c8e6 b2b3 f6d7 h4d8 e8d8 d3d4 f7f6 d1e3 d7f8 e3f5 e6f5 e4f5 d8b8 c1a3 d6a3 a1a3');
INSERT INTO evaluations VALUES(38798,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(38798,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(38798,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(38798,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(38798,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(38798,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(38798,36,'cp 99','c2c3 b4e7 f1d3 c7c5 e4e5 b8c6 g1e2 c5d4 c3d4 f7f6 d2f3 e7b4 c1d2 b4d2 d1d2 d8a5 d2a5 c6a5 h2h4 a5c6 e1g1 c8d7 a1c1 a8c8 f1e1 g8h6 c1c3 c6b4 c3c8 d7c8 e2f4 c8d7 d3b1 e8f7 a2a3 b4c6 b2b4 a7a6 b1d3 f6e5 d4e5 f7e7 g2g3 h6f7');
INSERT INTO evaluations VALUES(38797,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(38797,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(38797,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(38797,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(38797,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(38797,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(38797,36,'cp 95','b4e7 e4e5 c7c5 f1d3 b8c6 g1e2 c5d4 c3d4 f7f6 d2f3 e7b4 c1d2 b4d2 d1d2 d8a5 d2a5 c6a5 a1c1 c8d7 e1d2 a5c6 h1e1 g8e7 g2g3 e8g8 e5f6 f8f6 e2f4 a8f8 h2h4 d7e8 f3e5 c6d4 c1c7 e7c6 d2c3 e8g6 c7b7 c6e5 e1e5 d4f3 e5e6 d5d4 c3b4 f6e6 f4e6 g6d3 e6f8');
INSERT INTO evaluations VALUES(39342,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(39342,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(39342,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(39342,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(39342,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(39342,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(39342,36,'cp 0','a7a5 g1f3 b8c6 c2c3 b4c5 d2e3 c5e3 f2e3 d7d5 e4d5 d8d5 e3e4 d5d6 f1e2 g8f6 b1a3 e8g8 a3c4 d6e7 e1g1 a5a4 h2h3 h7h6 d1e1 c8e6 c4e5 c6e5 f3e5 f6e4 d3d4 e4f6 e2d3 c7c5 e1h4 c5d4 c3d4 f8d8 a1e1 e7b4 f1f6 g7f6 e1e3 b4d4 h4h6 f6e5 d3h7 g8h8 h7b1 h8g8');
INSERT INTO evaluations VALUES(41971,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(41971,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(41971,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(41971,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(41971,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(41971,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(41971,36,'cp -58','f6e7 f1d3 c7c5 d4c5 b8d7 g1f3 d7c5 e1g1 c8d7 a2a3 h7h6 c3e2 d8b6 a1b1 c5d3 c2d3 a8c8 d1d2 e8g8 g2g3 c8c7 b1c1 f8c8 c1c7 c8c7 g1g2 d7a4 f1c1 c7c6 e2d4 c6c1 d2c1 a7a5 c1c8 e7f8 c8c3 a4d1 c3d2 d1f3 d4f3');
INSERT INTO evaluations VALUES(40898,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(40898,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(40898,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(40898,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(40898,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(40898,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(40898,36,'cp -54','f1d3 c7c5 d4c5 b8d7 g1f3 d7c5 e1g1 c8d7 a2a3 d8b6 d1d2 h7h6 a1b1 e8g8 c3e2 c5d3 c2d3 f8c8 g2g3 c8c7 g1g2 a8c8 b1c1 c7c1 f1c1 c8c1 d2c1 d7b5 c1c3 a7a5 h2h4 e7c5 d3d4 c5f8 e2f4 a5a4 c3d2 b6c7 g3g4 c7c4');
INSERT INTO evaluations VALUES(42110,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(42110,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(42110,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(42110,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(42110,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(42110,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(42110,36,'cp 48','d8e7 g1f3 c7c5 c2c3 c8d7 f3e5 d7c6 f1d3 b8d7 d2f3 h7h6 d1e2 a7a5 d3c2 d7e5 f3e5 f6d7 e5d7 c6d7 e1c1 e8c8 d4c5 e7c5 e2f3 f7f5 g3g4 f5g4 f3g4 g7g5 g4d4 c5d4 c3d4 h6h5 f2f4 g5f4 e3f4');
INSERT INTO evaluations VALUES(44257,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(44257,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44257,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44257,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44257,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(44257,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44257,36,'cp 0','e6d5 c4d5 h7h6 g5c1 b7b5 e2e4 c5c4 d5d6 d8a5 c1d2 b5b4 f1c4 f8d6 d2e3 d6e5 d1b3 e8g8 g1f3 f6e4 c4d5 e4c5 e3c5 a5c5 f3e5 c5c1 e1e2 c8a6 e2f3 c1h1 d5f7 f8f7 b3f7 g8h8 f7e8 h8h7 f3g3 a6b7 e8g6 h7g8 g6f7 g8h8 f7f8 h8h7 f8f5 h7g8');
INSERT INTO evaluations VALUES(44347,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44347,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44347,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44347,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44347,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44347,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44347,36,'cp 0','c4d5 h7h6 g5c1 b7b5 e2e4 c5c4 d5d6 d8a5 c1d2 b5b4 f1c4 f8d6 d2e3 d6e5 d1b3 e8g8 g1f3 f6e4 c4d5 e4c5 e3c5 a5c5 f3e5 c5c1 e1e2 c8a6 e2f3 c1h1 d5f7 f8f7 b3f7 g8h8 b1c3 h1a1 f7e8 h8h7 e8g6 h7g8 g6f7');
INSERT INTO evaluations VALUES(44333,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44333,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44333,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44333,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44333,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44333,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44333,36,'cp 0','h7h6 g5c1 b7b5 e2e4 c5c4 d5d6 d8a5 c1d2 b5b4 f1c4 f8d6 d2e3 d6e5 d1b3 e8g8 g1f3 f6e4 c4d5 e4c5 e3c5 a5c5 f3e5 c5c1 e1e2 c8a6 e2f3 c1h1 d5f7 f8f7 b3f7 g8h8 e5g6 h8h7 g6f8 h7h8');
INSERT INTO evaluations VALUES(45877,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(45877,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(45877,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(45877,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(45877,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(45877,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(45877,36,'cp 35','d3d4 b8d7 f1d3 f8e7 e1g1 e8g8 f1e1 f8e8 h2h3 h7h6 a2a4 e5d4 f3d4 e7f8 d4b3 d7e5 d3f1 e5g6 f2f4 c8d7 d1f3 f6d5 f3f2 d5c3 b2c3 f7f5 e4f5 e8e1 f2e1 g6h4 e1f2 d7f5 b3d4 d8f6');
INSERT INTO evaluations VALUES(46601,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(46601,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(46601,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(46601,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(46601,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(46601,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(46601,36,'cp 60','f1e2 e7e6 e1g1 f8e7 c2c4 c7c6 d1c2 e8g8 b2b3 c8d7 f3e1 d7e8 e1d3 b8d7 a2a4 a7a5 b1d2 e8g6 d2f3 f6e4 c1a3 e7a3 a1a3 g6h5 c2b2 d8e7 c4c5 e4f6 b3b4 h5f3 e2f3 a5b4 b2b4 e6e5 d3e5 d7e5 d4e5');
INSERT INTO evaluations VALUES(49448,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(49448,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49448,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(49448,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49448,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49448,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49448,36,'cp -5','d7d5 e4e5 f6d7 f2f4 f8e7 g2g3 c7c5 a2a4 e8g8 g1f3 b8c6 f1g2 a7a6 e1g1 a8b8 d1e2 b7b5 a4b5 a6b5 f4f5 b5b4 f5e6 f7e6 c3d1 c6d4 f3d4 f8f1 g2f1 c5d4 f1h3 d7f8 h3g4 d8c7 d1f2 f8g6 c1f4 g6f4 g3f4 c8d7 a1a6');
INSERT INTO evaluations VALUES(50006,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(50006,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(50006,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(50006,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(50006,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(50006,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(50006,36,'cp 145','f8b4 d1d3 d7d5 a2a3 b4c3 d3c3 b8c6 g1f3 c8e6 b2b4 e8g8 c1g5 d8d6 g5f6 f8f6 e2e3 a8f8 f1e2 c6e7 c3c5 d6c5 b4c5 b7b6 c5b6 a7b6 h2h4 e6f5 e1g1 f5e4 f1c1 c7c5 c2c4 d5d4 e3d4 c5d4 f3d4 f6f2');
INSERT INTO evaluations VALUES(50043,30,'cp 683',NULL);
INSERT INTO evaluations VALUES(50043,31,'cp 690',NULL);
INSERT INTO evaluations VALUES(50043,32,'cp 688',NULL);
INSERT INTO evaluations VALUES(50043,33,'cp 698',NULL);
INSERT INTO evaluations VALUES(50043,34,'cp 701',NULL);
INSERT INTO evaluations VALUES(50043,35,'cp 740',NULL);
INSERT INTO evaluations VALUES(50043,36,'cp 751','e5h8 d7d5 d2d4 c8e6 c2c4 d5e4 d4d5 e7d5 c4d5 d8d5 h8h7 b8c6 b1c3 d5f5 h7c7 c6d4 c1e3 d4c2 e1d2 c2a1 f1b5 e6d7 b5d7 f5d7 c7d7 e8d7 c3e4 f8g7 h1a1 g7b2 a1b1 b2e5 b1b7 d7c6 b7e7 e5h2 g2g3 h2g3 f2g3 a8d8 d2e2 a7a5 e7a7 d8d5 e2f3 d5h5 e4c3 c6d6 a7a6 d6e5 a6a5 e5e6 a5h5 g6h5');
INSERT INTO evaluations VALUES(43306,30,'cp 744',NULL);
INSERT INTO evaluations VALUES(43306,31,'cp 732',NULL);
INSERT INTO evaluations VALUES(43306,32,'cp 732',NULL);
INSERT INTO evaluations VALUES(43306,33,'cp 725',NULL);
INSERT INTO evaluations VALUES(43306,34,'cp 725',NULL);
INSERT INTO evaluations VALUES(43306,35,'cp 732',NULL);
INSERT INTO evaluations VALUES(43306,36,'cp 713','d7d5 d2d4 d8d6 f1e2 c8e6 c2c4 d5c4 b1c3 b8d7 c1g5 e8c8 h8h7 d8e8 a1d1 c8b8 e1g1 d7b6 a2a3 a7a6 h2h3 e6d7 f1e1 d7e6 e2f3 e7c8 h3h4 d6d7 h7d7 e6d7 h4h5 b6a4 c3a4 d7a4 e4e5 f8g7 h5h6 g7h6 g5h6 a4d1 e1d1 e8h8');
INSERT INTO evaluations VALUES(50072,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(50072,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(50072,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(50072,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(50072,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(50072,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(50072,36,'cp 88','d7d5 f1d3 c7c5 b2b3 e7e6 f5e6 f8e7 g1f3 e8g8 d1e2 b8c6 e1g1 f6h5 g2g3 g7g6 f3e5 c6d4 e2e1 c8e6 c1b2 c5c4 b3c4 d5c4 c3e2 c4d3 e2d4 e6h3 e1e4 h5f6 e4d3 d8d5 d3b3 h3f1 g1f1 d5b3 a2b3 f6e4 f2f4 e7f6 d4e6 f6e5 b2e5 e4d2 f1e2');
INSERT INTO evaluations VALUES(46586,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(46586,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(46586,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(46586,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(46586,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(46586,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(46586,36,'cp 120','f1d3 c7c5 b2b3 e7e5 f5e6 f8e7 d1e2 b8c6 g1f3 e8g8 e1g1 f6h5 g2g3 g7g6 c1b2 h5g7 f3e5 c8e6 e5c6 b7c6 c3a4 e6h3 f1e1 e7f6 b2f6 d8f6 f2f4 g6g5 a4c5 g5f4 g3f4 f6f4 e2e3 f4h4 e3g3 f8f4 g3h4 f4h4 d3f1 h3f5 c2c3 g7h5 e1e5 a8f8 a1e1 h4g4 g1h1 g4h4 f1g2 h5f4 e5e8 h4g4');
INSERT INTO evaluations VALUES(52323,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52323,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52323,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52323,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52323,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52323,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52323,36,'cp -4','b8c6 g1f3 f8g7 e4e5 g8h6 c1f4 e8g8 e1g1 d7d6 e5d6 e7d6 f1e1 c8g4 h2h3 g4f3 d1f3 h6f5 a1d1 a8c8 c4d5 d8d7 g2g4 f5d4 f3g3 f8e8 f4d6 h7h6 d6c5 b7b6 e1e8 d7e8 c5d4 c6d4 g3f4 d4e6 f4e3 c8d8 d5e6');
INSERT INTO evaluations VALUES(19305,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19305,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(19305,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(19305,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(19305,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(19305,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19305,36,'cp -1','g1f3 f8g7 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f1e1 c8g4 h2h3 g4f3 d1f3 h6f5 a1d1 d8b6 f3e4 g7d4 e4c2 b6b4 c3d5 b4b2 c2b2 d4b2 g2g4 f5d4 g1g2 d4e6 f4d6 f8d8 d5c7 e6c7 d6c7 d8d1 e1d1 a8e8 d1b1 b2e5 b1b7 e8e7 c4d5 e7c7 b7c7 e5c7 d5c6 g8f8 c6e4 f8e7 a2a4 f7f6 g2f3 c7a5 f3g2 f6f5');
INSERT INTO evaluations VALUES(52637,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(52637,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(52637,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(52637,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(52637,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(52637,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(52637,36,'cp 55','f8g7 b1c3 b8c6 c1e3 g7d4 e3d4 c6d4 d1d4 g8f6 f1e2 e8g8 e1g1 c8e6 b2b3 d8a5 a1b1 a8c8 f1d1 a7a6 a2a4 a5c5 d4d2 g8g7 e2f1 c5b4 d2b2 g7g8 d1e1 c8c7 h2h3 f8c8 g1h1 f6d7 f2f4 f7f6 c3d5 e6d5');
INSERT INTO evaluations VALUES(53279,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(53279,31,'cp -122',NULL);
INSERT INTO evaluations VALUES(53279,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(53279,33,'cp -108',NULL);
INSERT INTO evaluations VALUES(53279,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(53279,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(53279,36,'cp -111','d5e4 f3g5 c5d4 f1c4 e7e6 g5e4 g8f6 b1d2 b8c6 e1g1 c8d7 f1e1 f8e7 e4f6 e7f6 d2e4 e8g8 d1h5 f6e7 c4d3 f7f5 e4g5 e7g5 c1g5 d8c7 d3c4 a8e8 a1c1 f8f7 g5d2 c6e5 c4b3 e5c4 d2f4 c7f4 b3c4 f4c7 h5e2 f7e7 c1d1 d7c6');
INSERT INTO evaluations VALUES(53209,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(53209,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(53209,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(53209,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(53209,34,'cp -110',NULL);
INSERT INTO evaluations VALUES(53209,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(53209,36,'cp -135','f3g5 g8f6 f1c4 e7e6 b1d2 c5d4 g5e4 f8e7 e4f6 e7f6 d2e4 e8g8 e1g1 f6e7 f1e1 b8c6 c1d2 c6e5 c4b3 a7a5 d1h5 e5g6 e4g5 e7g5 h5g5 d8g5 d2g5 h7h6 g5d2 c8d7 c2c3 d4c3 d2c3 d7c6 f2f3 f8c8 a1d1 g6e7 d1d6 c6d5 b3c2 d5a2 d6d7');
INSERT INTO evaluations VALUES(53347,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(53347,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(53347,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(53347,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(53347,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(53347,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(53347,36,'cp 176','d1d4 e7e6 e2e4 g8f6 b1c3 e6d5 e4d5 b8d7 f1b5 f8d6 g1f3 e8g8 e1g1 d8b6 d4h4 a7a6 c1e3 d6c5 b5d7 c8d7 e3c5 b6c5 f1d1 a8c8 f3e5 c5d6 d1e1 h7h6 a2a3 d6b6 h4b4 b6b4 a3b4 f8d8 f2f3 d7b5 a1d1 f6d7 e5d7 d8d7 g1f2');
INSERT INTO evaluations VALUES(53345,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(53345,31,'cp 174',NULL);
INSERT INTO evaluations VALUES(53345,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(53345,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(53345,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(53345,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(53345,36,'cp 159','e7e6 e2e4 e6d5 e4d5 g8f6 b1c3 f8d6 f1b5 c8d7 c1g5 d7b5 g5f6 g7f6 c3b5 d6e5 d4d2 d8d7 b5c3 d7g4 d2e3 g4g2 e1c1 g2g5 e3g5 f6g5 g1f3 e5f4 c1b1 b8a6 h1e1 e8f8 c3e2 a8d8 e2f4 g5f4 d5d6 a6c5 h2h4 c5e6 f3g5 e6g5 h4g5 f4f3 e1e7');
INSERT INTO evaluations VALUES(53801,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(53801,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53801,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(53801,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53801,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53801,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(53801,36,'cp 13','f8e7 h2h3 f7f5 e4f5 c8f5 g1f3 g8f6 f1c4 b7b5 c4d3 f5d3 d1d3 d8a5 f3d2 e8g8 d5c6 b5b4 d2b3 a5c7 c3d5 f6d5 d3d5 g8h8 e1g1 c7c6 f1d1 c6d5 d1d5 h8g8 c1d2 b8c6 a2a3 b4a3 a1a3 f8c8 b3c1 g8f7 b2b3 a8b8 c2c4 f7e6 c1e2 h7h6 d2c3 c6b4 d5a5 d6d5 a5a7 d5d4');
INSERT INTO evaluations VALUES(55240,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(55240,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(55240,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(55240,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(55240,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(55240,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(55240,36,'cp 82','f1e2 e7e6 g1f3 f8d6 c2c4 b7b6 c1e3 g8f6 b1c3 b8d7 d1c2 c8b7 e1c1 f6g4 e3g5 d6e7 g5e7 d8e7 e2d3 b7f3 g2f3 g4f6 c3b5 e8g8 b5c7 a8c8 c7b5 a7a6 b5c3 b6b5 c4c5 d7b8 d3e4 b8c6 e4c6 c8c6 h1e1 f8d8');
INSERT INTO evaluations VALUES(55333,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(55333,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(55333,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(55333,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(55333,34,'cp 250',NULL);
INSERT INTO evaluations VALUES(55333,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(55333,36,'cp 272','d4e5 b8c6 f1b5 f8g7 b1c3 c8e6 c1g5 f7f6 e5f6 g8f6 d1d2 e8g8 e1c1 d8e7 h1e1 e7f7 h2h3 c6e5 f3d4 e6a2 f2f4 e5c4 b5c4 a2c4 e4e5 d6e5 f4e5 f6e8 e5e6 f7f2 d4f3 f2d2 f3d2 g7c3 b2c3 c4d5 g2g4 e8d6 e6e7 f8f7 d2b3 d5f3 d1d6 c7d6 e7e8q a8e8 e1e8 g8g7');
INSERT INTO evaluations VALUES(55319,30,'cp 255',NULL);
INSERT INTO evaluations VALUES(55319,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(55319,32,'cp 257',NULL);
INSERT INTO evaluations VALUES(55319,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(55319,34,'cp 237',NULL);
INSERT INTO evaluations VALUES(55319,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(55319,36,'cp 264','b8c6 f1b5 f8g7 b1c3 c8e6 c1g5 f7f6 e5f6 g8f6 d1d2 e8g8 e1c1 c6e5 f3e5 d6e5 f2f3 c7c6 d2d8 f8d8 d1d8 a8d8 b5e2 b7b5 c1b1 a7a6 b2b3 g8f7 c3d1 g7f8 d1f2 f8b4 f2d3 b4c3 h1d1 d8d7 a2a3');
INSERT INTO evaluations VALUES(55676,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(55676,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(55676,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(55676,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(55676,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(55676,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(55676,36,'cp 107','c1f4 f8e7 g1f3 g8f6 b1c3 e8g8 f1d3 b8c6 a2a3 c8g4 f4e3 g4h5 d1d2 d6d5 e4e5 f6g4 e1c1 g4e3 d2e3 b7b5 c3b5 c6a5 b5c3 h5g6 e5e6 a8b8 e6f7 f8f7 d3g6 h7g6 e3e6 e7g5 c1b1 a5c4 f3e5 b8b2 b1a1 c4e5 a1b2');
INSERT INTO evaluations VALUES(55674,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(55674,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(55674,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(55674,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(55674,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(55674,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(55674,36,'cp 82','f8e7 b1c3 g8f6 g1f3 e8g8 f1d3 b8a6 a2a3 c7c5 d4d5 a6b8 f3d2 b8d7 e1g1 f6e8 d2c4 e7f6 d1h5 g7g6 h5f3 d7e5 c4e5 f6e5 f3g3 d8e7 a3a4 c8d7 c3d1 e5f4 g3f4 e7e5 f4e5 d6e5 a4a5 e8d6');
INSERT INTO evaluations VALUES(55825,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(55825,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(55825,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(55825,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(55825,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(55825,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55825,36,'cp 0','f8d6 e2e3 g8f6 g1f3 e8g8 f4d6 d8d6 b1d2 b7b6 f1e2 c7c5 e1g1 b8c6 a2a3 e6e5 d4e5 c6e5 c2c4 c8e6 d1c2 e5c4 d2c4 d5c4 e2c4 e6c4 c2c4 d6e7 f1d1 h7h6 d1d3 a8d8 a1d1 d8d3 d1d3 f8d8 f3d2 d8d3 c4d3 e7e6 d3c2');
INSERT INTO evaluations VALUES(55950,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(55950,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(55950,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55950,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(55950,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(55950,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(55950,36,'cp -55','d2d4 g8f6 f3d1 c7c5 g1f3 b8c6 f1e2 b7b6 e1g1 f8d6 d4c5 b6c5 c2c4 e8g8 c4d5 e6d5 b2b3 c8b7 b1d2 f6e4 c1b2 d8e7 a2a3 a8d8 a1c1 f7f5 b3b4 c5c4 b4b5 c6e5 b2e5 d6e5 f3e5 e7e5 d1c2 e4d2 c2d2 f5f4 d2c3 e5c3 c1c3 f4e3 f2e3');
INSERT INTO evaluations VALUES(55982,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55982,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55982,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55982,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(55982,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55982,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55982,36,'cp 0','d2d3 b8c6 d1e2 b7b6 b1d2 g8e7 a2a3 c8b7 f2f4 e7f5 d2e4 h7h5 g1f3 c6e7 e1g1 d8d7 c1d2 e7d5 c2c3 e8c8 c3d4 c8b8 b2b4 d5e7 d2c3 e7d5');
INSERT INTO evaluations VALUES(1093,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(1093,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(1093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1093,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1093,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1093,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(1093,36,'cp 6','h5d1 d7d6 g1f3 g7g6 d2d3 h7h6 e1g1 f8g7 a2a4 e8g8 a4a5 c8g4 h2h3 g4f3 d1f3 c6d4 f3d1 c7c6 c4a2 b7b5 c1e3 d4e6 f2f3 g8h7 a2e6 e7e6 d3d4 a7a6 d4e5 d6e5 d1e2 a8d8 e3f2 f6h5 a1d1 f8e8');
INSERT INTO evaluations VALUES(1217,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(1217,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(1217,32,'cp -141',NULL);
INSERT INTO evaluations VALUES(1217,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(1217,34,'cp -120',NULL);
INSERT INTO evaluations VALUES(1217,35,'cp -146',NULL);
INSERT INTO evaluations VALUES(1217,36,'cp -133','e1e2 c8f5 d3e1 c6c5 d2d3 e8c8 c1e3 b7b6 b1c3 g7g6 e3g5 d8e8 e2d2 h7h6 g5e3 g8e7 e1f3 f8g7 a1e1 g6g5 d2c1 f5d7 e3d2 e7g6 e1e8 d7e8 h2h3 e8c6 c3d1 c6f3 g2f3 g7e5 h1e1 c8b7 d2c3 h8e8 c1b1 f7f5 c3e5 g6e5');
INSERT INTO evaluations VALUES(1216,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(1216,31,'cp -144',NULL);
INSERT INTO evaluations VALUES(1216,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(1216,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(1216,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(1216,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(1216,36,'cp -115','c8f5 d3e1 g8e7 d2d3 c6c5 b2b3 e7c6 e1f3 f8e7 c1b2 e8g8 b1d2 a8e8 e2d1 f7f6 h1e1 b7b6 a2a3 g8f7 h2h3 h7h5 d2e4 g7g5 e4g3 f5g6 h3h4 g5g4 f3d2 f6f5 e1h1 e8d8 g3e2 b6b5 a3a4 f8e8 a4b5 a6b5');
INSERT INTO evaluations VALUES(1852,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(1852,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(1852,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(1852,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(1852,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(1852,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(1852,36,'cp -13','h7h6 d2d3 g8e7 b1d2 d7d6 b2b4 c8e6 b4b5 c6b8 c4e6 f6e6 g1e2 b8d7 e1g1 g7g6 h5f3 a7a6 b5a6 a8a6 a1b1 b7b6 d3d4 f8g7 d4d5 e6f6 f3h3 g6g5 c3c4 f6g6 e2c3 g5g4 h3d3 e8g8 d3e2 h6h5 f2f3 a6a8 f3g4 h5g4 a2a4 g7h6 d2b3 h6c1 b1c1 f7f5');
INSERT INTO evaluations VALUES(4066,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(4066,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(4066,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(4066,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(4066,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(4066,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(4066,36,'cp 114','e4e5 f6e8 c4c5 c7c6 c1e3 d6e5 d4e5 e8c7 f1c4 e7e6 d1e2 c7d5 c3e4 d5e3 e2e3 d8a5 e1f2 b7b6 h2h4 d7c5 e4c5 a5c5 e3c5 b6c5 h4h5 f8d8 g2g4 c8b7 f4f5 g6f5 g4f5 e6f5 h5h6 g7f8 f3g5');
INSERT INTO evaluations VALUES(6765,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(6765,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(6765,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(6765,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(6765,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(6765,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(6765,36,'cp 46','d2d4 g8f6 e1g1 e8g8 f1e1 h7h6 b1d2 f8e8 c4b3 b7b5 a2a4 b5b4 d2c4 b4c3 d4e5 c6e5 c4e5 d6e5 f3e5 e8e5 f2f4 e5e8 e4e5 a8b8 b3c2 f6h7 c1e3 h7f8 d1d3 a7a5 d3c3 d7d6 c3a5 d6e5 a1d1 c8d7 a5c3 e5f4 e3f4 e8e1 d1e1 f8e6 f4g3 b8c8 c2b3 c7c5');
INSERT INTO evaluations VALUES(6770,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(6770,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(6770,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(6770,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(6770,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(6770,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6770,36,'cp 16','g8f6 e1g1 e8g8 f1e1 h7h6 b1d2 f8e8 d1b3 d8e7 a2a4 c6a5 b3a2 f6h7 c4f1 b7b6 d2c4 a5c4 f1c4 e8f8 a4a5 a8b8 a2e2 c8b7 d4e5 d6e5 f3e5 e7e5 f2f4 e5c5 c1e3 c5e7 e4e5 d7d6 a5a6 b7a8 e2g4 d6e5 f4e5 e7e5 e3b6 h7f6 g4h4 e5c3 b6a7 b8b2 c4f1 b2b4');
INSERT INTO evaluations VALUES(7816,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(7816,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7816,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7816,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7816,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7816,36,'cp 0','b7c6 e4e5 f6e4 e1g1 e4c3 d1g4 h8g8 c1g5 d8c7 b2c3 b4c3 a1d1 h7h6 g5h6 c3e5 f2f4 e5c3 f1f3 c3f6 h6g5 f6g5 f4g5 d7d5 c4e2 g8h8 h2h3 c7b8 f3e3 b8b6 g4g3 b6b8');
INSERT INTO evaluations VALUES(8839,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(8839,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(8839,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(8839,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(8839,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(8839,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(8839,36,'cp 28','d2d4 c5b6 a2a4 a7a5 c1e3 f6g4 b1d2 e8g8 h2h3 g4e3 f2e3 c6e7 c4b3 e5d4 e3d4 c7c6 d1e2 e7g6 e1f1 b6c7 a1e1 c8e6 b3c2 f8e8 f1f2 c6c5 d4c5 g6e5 f3e5 d6e5 e1f1 d8e7 c2b3 e7c5 g1h1 e6b3');
INSERT INTO evaluations VALUES(8863,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(8863,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(8863,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(8863,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(8863,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(8863,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(8863,36,'cp 38','c5b6 a2a4 a7a5 c1e3 c8g4 d1b3 e8g8 b1d2 e5d4 f3d4 g4h5 f2f3 c6e5 e3f2 f8e8 c4f1 h5g6 d2c4 e5c4 f1c4 f6d7 g1h1 d7e5 c4f1 e5d7 f2g1 d7c5 b3d1 d6d5 b2b4 c5a6 e4d5 a5b4 a4a5 b6c5 c3b4 a6b4 e1e8 d8e8');
INSERT INTO evaluations VALUES(8956,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(8956,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8956,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(8956,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(8956,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(8956,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(8956,36,'cp 27','f1d3 e8g8 a2a3 b4d6 e3e4 e6e5 c3d5 f6d5 e4d5 c6d4 f3d4 e5d4 e1g1 d8h4 g2g3 h4g4 f1e1 g4d1 a1d1 c8g4 d1b1 f8e8 g1g2 h7h6 h2h3 g4d7 b1d1 a8d8 g3g4 c7c6 d5c6 d7c6 f2f3 g7g5 d2a5 b7b6 a5b4 d6b4 e1e8 c6e8 a3b4 g8f8 g2g3');
INSERT INTO evaluations VALUES(8960,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(8960,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(8960,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(8960,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8960,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(8960,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(8960,36,'cp 23','e8g8 a2a3 b4d6 e3e4 e6e5 c3d5 f6d5 e4d5 c6d4 f3d4 e5d4 e1g1 d8h4 g2g3 h4g4 d2b4 f8e8 d1g4 c8g4 f2f3 g4d7 f1e1 d6e5 g1g2 b7b6 b4d2 h7h5 g2f2 e5d6 d2f4 d6f4 g3f4 c7c5 d5c6 d7c6 e1e5 g8f8');
INSERT INTO evaluations VALUES(3672,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(3672,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(3672,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(3672,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(3672,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(3672,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3672,36,'cp 26','a2a3 b4d6 e3e4 e6e5 c3d5 f6d5 e4d5 c6d4 f3d4 e5d4 e1g1 c8d7 f1e1 d8f6 d1h5 h7h6 h2h4 f8e8 d2g5 e8e5 g2g3 e5d5 g5f6 d5h5 f6d4 a8d8 d3e4 b7b6 a1d1 d7g4 d1d2 d6e5 c2c3 e5f6 e4g2 g4e6 b2b4 a7a5 g2f3 h5f5 f3c6 e6d5');
INSERT INTO evaluations VALUES(12472,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(12472,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(12472,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(12472,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(12472,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(12472,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(12472,36,'cp 32','e3d4 f8b4 f1d3 d7d5 c4d5 e6d5 e1g1 e8g8 h2h3 h7h6 c1f4 c8e6 a1c1 a8c8 a2a3 b4c3 c1c3 f6e4 c3b3 c6a5 b3b4 f8e8 f3e5 d8f6 f4g3 f6d8 f1e1 e4g3 f2g3 f7f6 e5g6 d8d6 e1e3 e6f7 e3e8 c8e8 g6f4 d6c6 g1h2');
INSERT INTO evaluations VALUES(14948,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(14948,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(14948,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(14948,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(14948,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(14948,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14948,36,'cp 35','f8b4 f3e5 e8g8 e1g1 b4c3 b5c6 b7c6 b2c3 f6e4 f2f3 e4d6 e5c6 d8d7 c6b4 d6c4 c1f4 a7a5 b4d3 d7c6 h2h4 f8e8 f1e1 c8d7 e1e8 a8e8 d1f1 h7h6 a1e1 d7e6 d3c5 e6f5 f1f2 f5c2 g2g4 c2h7 f4c7 a5a4 c7f4 a4a3 f2g3 f7f6');
INSERT INTO evaluations VALUES(18270,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(18270,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(18270,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(18270,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(18270,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(18270,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(18270,36,'cp -67','f7f5 d1a4 g8f6 d2d4 f5e4 d4e5 e4f3 e5f6 d8d7 g2f3 g7f6 c1e3 a8b8 b2b3 c6c5 a4h4 d7h3 h4e4 h3e6 c3c4 e6e4 f3e4 f6f5 h1g1 e8f7 b1c3 f8g7 a1c1 f5e4 c3e4 a7a5 g1g3 a5a4 e1d2 a4b3 a2b3 b8b3 c1g1 h8g8 e4g5 f7e8 g5e6 c8e6 g3g7 b3b2 d2c1 g8g7 g1g7');
INSERT INTO evaluations VALUES(19706,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(19706,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(19706,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(19706,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(19706,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(19706,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(19706,36,'cp 117','c2c3 e7e6 d2d4 g8f6 d4d5 e6d5 e4d5 c6e5 f1e1 f8e7 f3e5 d6e5 c1f4 e5f4 d5d6 c8e6 d1a4 e8f8 d6e7 d8e7 c4e6 f7e6 a4f4 f8f7 b1d2 h8d8 d2f3 d8d5 h2h3 f7g8 e1e3 a8d8 a1e1 d5d1 f4c4 d8d5 c4f1 d1e1 f1e1 e7d7 g1h2 d5d6 e1e2 f6d5 f3e5 d7c7 e3g3 d5f4 e2f3');
INSERT INTO evaluations VALUES(20080,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(20080,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(20080,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(20080,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(20080,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(20080,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(20080,36,'cp 106','e7e6 g1f3 g8f6 f1d3 f8e7 e1g1 a7a6 c1e3 e8g8 a1c1 f6g4 e3d2 b7b5 h2h3 g4f6 c3e2 c8b7 e2g3 a8c8 d1e2 c6a5 c1c8 b7c8 e4e5 f6d5 e2e4 g7g6 d2h6 f8e8 h3h4 d6e5 d4e5 d5b4 d3e2 c8b7 e4f4 b7f3');
INSERT INTO evaluations VALUES(20728,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(20728,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(20728,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(20728,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(20728,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(20728,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(20728,36,'cp 137','g1f3 g8f6 f1d3 a7a6 e1g1 f8e7 h2h3 b7b5 a2a4 b5b4 c3e2 e8g8 e2g3 c8b7 c1e3 f8e8 a1c1 a8c8 f3d2 e7f8 f1e1 c6e7 c1c8 b7c8 d1e2 a6a5 e1c1 c8d7 b2b3 e7c6 g3h5 f6h5 e2h5 g7g6');
INSERT INTO evaluations VALUES(21816,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(21816,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(21816,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(21816,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(21816,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(21816,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(21816,36,'cp 55','f1b5 f7f5 e4c3 f8d6 d1e2 g8e7 f3e5 e8g8 e5f3 a7a6 b5c4 g8h8 d2d4 b7b5 c4b3 b5b4 c3a4 f5f4 c2c3 e7g6 e1g1 c8d7 f1e1 g6h4 b3d5 b4c3 a4c3 a8b8 e2d1 d8f6 d5c6 d7c6 d4d5 c6e8 c3e4 h4f3 d1f3 f6e5 f3c3 e5d5 e4d6 d5d6 b2b3 e8c6 c1b2 d6g6 f2f3 b8e8 e1e8 f8e8 a1c1 e8e2 c3g7 g6g7 b2g7 h8g7 c1c6 e2a2 c6c7 g7g6');
INSERT INTO evaluations VALUES(21545,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(21545,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(21545,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(21545,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(21545,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21545,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(21545,36,'cp 23','f7f5 e4c3 f8d6 d1e2 g8e7 f3e5 e8g8 e5f3 a7a6 b5c4 g8h8 d2d4 b7b5 c4b3 b5b4 c3a4 f5f4 c2c3 e7g6 e1g1 c8d7 c1d2 c6a5 b3c2 d7b5 e2d1 a5c4 c2g6 h7g6 d2c1 g6g5 h2h3 g5g4 h3g4 d8d7 b2b3 b5a4 f3h4 a4b5 h4g6 h8g8 b3c4 b5c4 f1e1 f8e8 c1f4 e8e1 d1e1 a8e8 g6e5 d6e5 d4e5 d7g4');
INSERT INTO evaluations VALUES(22001,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(22001,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(22001,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(22001,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(22001,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(22001,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(22001,36,'cp 50','c2c4 f7f5 b1c3 g7g6 e1g1 f8g7 a1b1 a7a5 d2d3 h7h6 a2a3 g6g5 e2e3 g8f6 b2b4 a5b4 a3b4 e8g8 b4b5 c6e7 c1b2 c8e6 d1c2 d8d7 b1a1 f6h7 a1a8 f8a8 f1a1 a8a1 b2a1 g5g4 f3d2 b7b6 c2a4 h7g5 f2f4 g4f3 d2f3 g8h8 f3h4');
INSERT INTO evaluations VALUES(23091,30,'cp -637',NULL);
INSERT INTO evaluations VALUES(23091,31,'cp -642',NULL);
INSERT INTO evaluations VALUES(23091,32,'cp -634',NULL);
INSERT INTO evaluations VALUES(23091,33,'cp -648',NULL);
INSERT INTO evaluations VALUES(23091,34,'cp -628',NULL);
INSERT INTO evaluations VALUES(23091,35,'cp -655',NULL);
INSERT INTO evaluations VALUES(23091,36,'cp -667','d8g5 c4f1 g8f6 d2d3 g5g6 c2c3 d4c6 b1d2 d7d5 d2f3 d5e4 f3h4 g6g4 d1g4 f6g4 f1e2 f8e7 g2g3 e4d3 e2d3 g7g6 e1g1 c8e6 d3e4 e6c4 f1e1 e7c5 c1e3 g4e3 f2e3 e8c8 b2b4 c5e7 e4c6 b7c6 h4f3 f7f6 e3e4');
INSERT INTO evaluations VALUES(1987,30,'cp -649',NULL);
INSERT INTO evaluations VALUES(1987,31,'cp -650',NULL);
INSERT INTO evaluations VALUES(1987,32,'cp -648',NULL);
INSERT INTO evaluations VALUES(1987,33,'cp -661',NULL);
INSERT INTO evaluations VALUES(1987,34,'cp -636',NULL);
INSERT INTO evaluations VALUES(1987,35,'cp -648',NULL);
INSERT INTO evaluations VALUES(1987,36,'cp -627','c4f1 g8f6 c2c3 d4c6 d2d3 g5g6 b1d2 d7d5 d2f3 d5e4 f3h4 g6g4 d1g4 f6g4 h2h3 g4f6 c1e3 f8e7 d3e4 f6e4 h4f3 e4d6 a2a4 e8g8 f1e2 c8e6 e1g1 c6a5 b2b4 a5c4 e2c4 e6c4 f1e1 f7f6 e3c5 g8f7 f3d2 c4e6 d2f3 f8d8 f3d2');
INSERT INTO evaluations VALUES(23456,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(23456,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(23456,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(23456,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(23456,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(23456,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(23456,36,'cp -104','d2d4 d7d5 h2h3 h7h5 g4g5 c8f5 g1f3 e7e6 f1g2 g8e7 f3h4 d8d7 c2c3 e8c8 b1d2 e6e5 d2b3 f5h7 c1e3 d7e8 d4e5 c6e5 e3a7 f7f6 a7d4 e7c6 e1g1 f6g5 h4f3 g5g4 f3e5 c6e5 d4e5 e8e5 d1d4 f8d6 d4e5 d6e5 a1e1 h7f5 h3g4 h5g4');
INSERT INTO evaluations VALUES(24359,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(24359,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(24359,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24359,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(24359,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(24359,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(24359,36,'cp 0','h2h3 g4h5 d2d4 e5d4 d1d4 h5f3 g2f3 f8e7 f1e1 e8g8 c1d2 f6h5 f3f4 g7g6 a1d1 d8d7 g1h2 f8e8 d4d3 a8d8 d2c1 e7f8 c3e2 d7e6 e2d4 e6d7');
INSERT INTO evaluations VALUES(24358,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(24358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24358,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(24358,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(24358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24358,36,'cp 8','g4h5 d2d4 e5d4 d1d4 h5f3 g2f3 f8e7 d4a4 d8d7 f1d1 c6c5 a1b1 d7a4 c3a4 f6d7 a4c3 d7b6 c1e3 a7a5 a2a3 e8g8 c3b5 f8c8 g1g2 f7f5 e4f5 a5a4 b2b4 a4b3 c2b3 d6d5 b3b4 d5d4 b4c5 e7c5');
INSERT INTO evaluations VALUES(24355,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(24355,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(24355,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(24355,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(24355,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(24355,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(24355,36,'cp 0','d2d4 e5d4 d1d4 h5f3 g2f3 f8e7 d4a4 d8d7 f1d1 c6c5 a1b1 d7a4 c3a4 a7a5 g1g2 f6d7 d1d5 d7f6 d5d3 f6d7');
INSERT INTO evaluations VALUES(24963,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(24963,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24963,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(24963,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(24963,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24963,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24963,36,'cp 17','e7e6 e1g1 f8e7 b2b3 a7a6 c1b2 e8g8 c3b1 f8e8 f1d1 a6a5 a2a3 c6b8 c2c4 b8d7 b1d2 c7c6 a1c1 e7d6 e3e4 f6e4 d2e4 d5e4 d3e4 d8e7 a3a4 a8d8 d1d3 d6a3 e4e2 h7h6 g2g3 a3b2 e2b2 c6c5 c1d1');
INSERT INTO evaluations VALUES(25227,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25227,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(25227,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(25227,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(25227,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(25227,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(25227,36,'cp 19','e1g1 f8e7 f1e1 e7b4 c1d2 e8g8 a2a3 b4c3 d2c3 a7a5 c3d2 f6e4 c2c4 d8d7 e1c1 f7f6 d2e1 f8b8 a3a4 e4d6 b2b3 b8e8 h2h3 d7f7 d3c2 a8d8 c1d1 g8h8 f3h2 c6b4 e1b4 a5b4 d1c1 c7c6 a4a5 d5c4 b3c4 e6e5 d4e5 e8e5 h2f3');
INSERT INTO evaluations VALUES(26018,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(26018,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(26018,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(26018,33,'cp 136',NULL);
INSERT INTO evaluations VALUES(26018,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(26018,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(26018,36,'cp 141','g1f3 a7a6 b5a4 e7e5 e1g1 f8d6 d2d4 e5d4 f3d4 d8c7 a4c6 d7c6 d4c6 c7c6 c1g5 f7f6 g5e3 e8c8 b1d2 g8e7 d1g4 c8b8 g4c4 d6c7 c4c6 e7c6 c2c3 c6e5 f1d1 h7h5 h2h3 g7g5 d2e4 d8d1 a1d1 e5c4 e4f6 c4b2 d1d7 b2c4');
INSERT INTO evaluations VALUES(26017,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(26017,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(26017,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(26017,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(26017,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(26017,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(26017,36,'cp 144','a7a6 b5a4 e7e5 e1g1 f8d6 f1e1 g8e7 a4c6 d7c6 f3e5 d6e5 e1e5 e8g8 d2d4 f8e8 b1c3 d8d7 f2f3 a8d8 c1f4 f7f6 e5e2 b7b5 e2d2 a6a5 f4g3 e7d5 c3d5 c6d5 g3f2 a5a4 h2h3 b5b4 c2c3 b4c3 b2c3 d5c4 d1c2 d8c8 a1e1 e8e1 f2e1 c8e8 e1g3 a4a3');
INSERT INTO evaluations VALUES(27140,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(27140,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(27140,32,'cp 231',NULL);
INSERT INTO evaluations VALUES(27140,33,'cp 238',NULL);
INSERT INTO evaluations VALUES(27140,34,'cp 232',NULL);
INSERT INTO evaluations VALUES(27140,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(27140,36,'cp 223','d8d7 f3d4 e5d4 d1d4 g8f6 b1c3 f8e7 e1g1 e8g8 h2h3 f8e8 f1e1 a7a6 c1e3 a8c8 a1d1 b7b5 a2a3 c7c6 d5c6 d7c6 e3f4 a6a5 e1e2 c6b7 d4d3 b5b4 c3b5 b4a3 b2a3 e7f8 f4d6 e8e4 d6f8 c8f8 e2d2');
INSERT INTO evaluations VALUES(27659,30,'cp 244',NULL);
INSERT INTO evaluations VALUES(27659,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(27659,32,'cp 228',NULL);
INSERT INTO evaluations VALUES(27659,33,'cp 231',NULL);
INSERT INTO evaluations VALUES(27659,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(27659,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(27659,36,'cp 242','f3d4 e5d4 d1d4 g8f6 e1g1 f8e7 h2h3 e8g8 c2c4 d7e8 c1e3 f6d7 d4d3 e7f6 d3c2 e8e7 b1d2 a8e8 g2g3 g7g5 a1e1 f6g7 b2b3 e7f6 g1g2 f6e5 f1h1 e5b2 c2b2 g7b2 e1b1');
INSERT INTO evaluations VALUES(27658,30,'cp 248',NULL);
INSERT INTO evaluations VALUES(27658,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(27658,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(27658,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(27658,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(27658,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(27658,36,'cp 237','e5d4 d1d4 g8f6 c2c4 f8e7 h2h3 e8g8 e1g1 d7e8 d4d3 f6d7 c1e3 f7f5 e4f5 e8h5 b1c3 h5f5 d3f5 f8f5 f1e1 e7f6 c3e4 f6e5 e1e2 h7h6 a1e1 f5f7 a2a4 a8e8 f2f3 b7b6 b2b3 a7a6 e3d2 d7c5');
INSERT INTO evaluations VALUES(28710,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28710,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28710,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(28710,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(28710,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28710,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(28710,36,'cp 0','f8b4 f3d2 e8g8 a2a3 b4e7 b1c3 d7d5 f1e2 c7c5 d4c5 b6c5 c4d5 f6d5 d1b3 d8b6 e2f3 f8d8 b3b6 a7b6 d2c4 b8d7 e1c1 d5c3 f3b7 c3d1 h1d1 a8a4 f4c7 a4c4 c1b1 d8e8 d1d7 e7f6 b7c6 e8c8 c6b7');
INSERT INTO evaluations VALUES(29369,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(29369,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(29369,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(29369,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(29369,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(29369,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(29369,36,'cp 35','f4b8 d8b8 f1b5 e8d8 e1g1 a7a6 b5e2 f8d6 d4c5 d6c5 f3h4 f5g6 c2c4 d8e7 a1c1 d5c4 d2c4 g6e4 c4d2 h8c8 d2e4 f6e4 d1d3 e4f6 h4f3 b8d6 d3b1 e7f8 a2a3 d6b6 f3e5 c8c7 c1c2 c5e7 f1c1 c7c2 b1c2 a8d8 g2g3 g7g6 b2b4 f8g7 g1g2');
INSERT INTO evaluations VALUES(32580,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(32580,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(32580,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(32580,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(32580,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(32580,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(32580,36,'cp 26','b8d7 h2h3 g4f5 f3h4 f5g6 e1g1 e7e6 h4g6 h7g6 d2f4 f8e7 a2a3 e8g8 f1e1 a7a6 e2f3 a5b6 d1d3 f8e8 a1d1 a8d8 b2b4 f6d5 f4h2 d5c3 d3c3 d7f6 h2e5 f6d5 c3b3 e7f6 d1d3 f6e5 d4e5 d5e7 e1d1 d8d3 d1d3 e8d8 g2g3 d8d3');
INSERT INTO evaluations VALUES(33322,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(33322,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(33322,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(33322,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(33322,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(33322,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(33322,36,'cp -20','d7d6 f3g5 g7g6 h5h4 c8e6 c4e6 f7e6 d2d3 b8c6 g5f3 e7h4 f3h4 g8f6 h4f3 h7h6 c1e3 c5b6 h2h4 e8d7 b1c3 c6d4 e1d2 a8f8 h1h3 f6g4 h3g3 h6h5 c3a4 b6a5 c2c3 g4e3 f2e3 d4f3 g2f3 b7b5 b2b4 b5a4 b4a5 f8b8 a1c1 h8f8 a5a6 b8b5 d3d4 b5b2 c1c2 b2b1');
INSERT INTO evaluations VALUES(35543,30,'mate 10',NULL);
INSERT INTO evaluations VALUES(35543,31,'mate 10',NULL);
INSERT INTO evaluations VALUES(35543,32,'mate 10',NULL);
INSERT INTO evaluations VALUES(35543,33,'mate 10',NULL);
INSERT INTO evaluations VALUES(35543,34,'mate 10',NULL);
INSERT INTO evaluations VALUES(35543,35,'mate 10',NULL);
INSERT INTO evaluations VALUES(35543,36,'mate 10','g7g5 h2h4 f8b4 c2c3 d8e7 c1g5 h6g7 g5e7 g8e7 f5f7 g7h6 f7f6 e7g6 h4h5 b4c3 b2c3 d7d6 h5g6 c8h3 h1h3');
INSERT INTO evaluations VALUES(35547,30,'mate 10',NULL);
INSERT INTO evaluations VALUES(35547,31,'mate 10',NULL);
INSERT INTO evaluations VALUES(35547,32,'mate 10',NULL);
INSERT INTO evaluations VALUES(35547,33,'mate 10',NULL);
INSERT INTO evaluations VALUES(35547,34,'mate 10',NULL);
INSERT INTO evaluations VALUES(35547,35,'mate 10',NULL);
INSERT INTO evaluations VALUES(35547,36,'mate 10','h2h4 f8b4 c2c3 d8e7 c1g5 h6g7 g5e7 g8e7 f5f7 g7h6 f7f6 e7g6 h4h5 b4c3 b2c3 d7d6 h5g6 c8h3 h1h3');
INSERT INTO evaluations VALUES(35546,30,'mate 9',NULL);
INSERT INTO evaluations VALUES(35546,31,'mate 9',NULL);
INSERT INTO evaluations VALUES(35546,32,'mate 9',NULL);
INSERT INTO evaluations VALUES(35546,33,'mate 9',NULL);
INSERT INTO evaluations VALUES(35546,34,'mate 9',NULL);
INSERT INTO evaluations VALUES(35546,35,'mate 9',NULL);
INSERT INTO evaluations VALUES(35546,36,'mate 9','f8b4 c2c3 d8e7 c1g5 h6g7 g5e7 g8e7 f5f7 g7h6 f7f6 e7g6 h4h5 b4c3 b2c3 d7d6 h5g6 c8h3 h1h3');
INSERT INTO evaluations VALUES(36743,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(36743,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(36743,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(36743,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(36743,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(36743,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(36743,36,'cp -24','c7c5 e1g1 c8g4 d3e2 c5d4 e3d4 b8c6 h2h3 g4f3 e2f3 d8b6 c3e2 f8e8 d1d3 f6e4 f4e3 a8c8 a2a3 f7f5 f1c1 c8d8 f3e4 f5e4 d3d2 e7e5 c1d1 b6c7 d4e5 c6e5 b2b3 a7a6 a1b1 c7f7 a3a4 e5c6 e3b6 d8d7 e2d4 g7e5 c2c3 e5f6 d4c6');
INSERT INTO evaluations VALUES(41577,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41577,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(41577,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41577,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(41577,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(41577,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41577,36,'cp 38','h4g3 f6e4 a2a3 b4c3 b2c3 h6h5 h2h4 g5g4 f3d2 e4g3 f2g3 d8d6 e1f2 f7f5 d1e1 b8d7 f2g1 c7c6 a3a4 d7f8 e2e3 f8g6 a4a5 c8d7 c4c5 d6c7 c3c4 e8g8 g1h2 g8g7 f1d3 a7a6 a1b1 g6h4 h2g1 h4g6 h1h5 f8h8 h5h8 g6h8 d2f1 h8f7 c4d5 c6d5');
INSERT INTO evaluations VALUES(41578,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41578,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41578,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(41578,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(41578,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41578,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(41578,36,'cp 40','f6e4 a2a3 b4c3 b2c3 h6h5 h2h4 g5g4 f3d2 e4g3 f2g3 d8d6 e1f2 f7f5 d1e1 b8d7 f2g1 c7c6 c4d5 e6d5 c3c4 d7f8 e2e3 f8g6 f1d3 e8g8 g1h2 g8g7 a3a4 c8e6 a4a5 d6c7 c4c5 a7a6 d2f1 g6h4 h2g1 h4g6 h1h5 f8h8 h5h8 a8h8 a1b1 c7e7 b1b6 e6c8 e1c3 h8e8 c3a3 e7c7');
INSERT INTO evaluations VALUES(41575,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(41575,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(41575,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(41575,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41575,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(41575,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(41575,36,'cp 26','a2a3 b4c3 b2c3 h6h5 h2h4 e4g3 f2g3 g5g4 f3d2 d8d6 e1f2 f7f5 d1e1 b7b6 f2g1 c8b7 c4d5 e6d5 e2e3 b8c6 g1h2 e8g8 c3c4 f5f4 e3f4 c6d4 c4d5 f8e8 d2e4 d6g6 f1d3 b7d5 f4f5 d4f5 h1f1 g6e6 e1f2 d5e4 d3e4 e6e4 f2f5 e4f5 f1f5 a7a5 f5h5');
INSERT INTO evaluations VALUES(41987,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(41987,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(41987,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(41987,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41987,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41987,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41987,36,'cp 8','e6d5 e2e3 c8f5 c3c4 c7c6 g1f3 e8g8 a3a4 b8d7 a4a5 f6e4 c4d5 c6d5 f1d3 d8f6 e1g1 f8c8 c1d2 e4d6 d3f5 f6f5 d2b4 d6c4 d1b1 f5h5 f3d2 b7b5 a5b6 c4d2 b4d2 d7b6 b1d3 b6c4 f1c1 h5g5 a1a6 c8b8 d2e1 b8b2 d3c3');
INSERT INTO evaluations VALUES(42259,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(42259,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42259,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42259,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(42259,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(42259,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(42259,36,'cp 31','e2e3 c8f5 g1e2 e8g8 e2f4 c7c5 f2f3 h7h5 h2h3 b8c6 g2g4 f5g6 f1g2 c6e7 e1g1 a8c8 f1f2 h5g4 h3g4 c5d4 c3d4 g6h7 a3a4 e7g6 g2f1 f8e8 a4a5 e8e7 f4g2 e7c7 c1d2 g6f8 g2f4 f8g6 f4d3 g6h4 f2h2 f6d7 d3f4 c7c2');
INSERT INTO evaluations VALUES(42904,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(42904,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(42904,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(42904,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(42904,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(42904,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(42904,36,'cp 168','d1d5 e4d6 d5e5 d8e7 e5e7 e8e7 c4d5 d6e8 c1g5 e7f8 f3d4 h7h6 g5f4 b8c6 e1c1 c6d4 d1d4 d7d6 h2h4 a8b8 h1e1 h8g8 h4h5 c8d7 d5f3 b7b6 d4c4 b8c8 f3b7 d7b5 c4b4 c8b8 b7f3 b5d7 b4d4 b8c8 c3c4 g7g5 h5g6 g8g6 e1h1 f8g7 f4e3 e8f6 d4f4');
INSERT INTO evaluations VALUES(43160,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(43160,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43160,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(43160,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43160,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43160,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43160,36,'cp 0','c7c5 d2d4 c5d4 f3d4 b8c6 d1b3 b7b6 f1e2 c8b7 e1g1 e8g8 f1d1 e7e6 d4b5 f6e8 c3e4 d7d5 c4d5 e6d5 e4g3 c6a5 b3a4 a5c6');
INSERT INTO evaluations VALUES(40000,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40000,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40000,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40000,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40000,36,'cp 10','d2d4 c5d4 f3d4 b8c6 d1b3 e8g8 f1e2 b7b6 e1g1 c8b7 f1d1 d8b8 c1d2 e7e6 d4c6 b7c6 a1c1 f8d8 c3b5 f6e4 d2e1 a7a5 b5d4 e4c5 b3c2 b8e5 b2b3 c6e4 c2d2 e4b7 f2f3 e5g5 e1f2 d7d5 c4d5 d8d5 d2c2');
INSERT INTO evaluations VALUES(43302,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(43302,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(43302,32,'cp 176',NULL);
INSERT INTO evaluations VALUES(43302,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(43302,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(43302,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(43302,36,'cp 136','c1f4 c7c5 d4c5 g7b2 b1d2 b2g7 f1c4 b8c6 h2h3 e7e6 a1b1 e8g8 e1g1 b7b6 d1e2 f7f5 f1d1 f5e4 e2e4 h6f7 e4e3 b6c5 d2e4 d7d5 e4c5 d8f6 c4b3 a7a5 f4g3 c6e7 g3h4 g6g5 f3g5 a5a4 g5f3 e7f5 h4f6 f5e3 f6g7 e3d1 g7f8 a4b3 b1d1 b3c2');
INSERT INTO evaluations VALUES(44704,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(44704,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(44704,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(44704,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(44704,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(44704,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(44704,36,'cp 58','g5f6 e7f6 c4d5 c6d5 e2e3 b8c6 g1e2 f8b4 a1c1 c8e6 e2f4 e8g8 f1e2 d8b6 e2d3 f8c8 e1g1 b4c3 b2c3 c6e7 f4e6 b6e6 d1b3 c8c7 c3c4 a8c8 c4c5 g8g7 g2g3 b7b6 d3a6 c8a8 c5b6 c7c1 f1c1 a7b6 a6b7 a8b8 c1c7 h7h5');
INSERT INTO evaluations VALUES(45033,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45033,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45033,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(45033,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45033,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45033,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45033,36,'cp 0','g7g6 g1f3 f8g7 f1c4 c6c5 d4d5 e8g8 e1g1 b8d7 c4e2 d7b8');
INSERT INTO evaluations VALUES(45039,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(45039,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(45039,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(45039,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(45039,34,'cp 191',NULL);
INSERT INTO evaluations VALUES(45039,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(45039,36,'cp 210','f1e2 e7e6 g1f3 c6c5 c1e3 c5d4 f3d4 a7a6 e4e5 f6d5 c3d5 d8d5 e1g1 d5e5 a1c1 b8d7 d4f3 e5d6 d1c2 f8e7 f1d1 d6b8 f3g5 e7g5 e3g5 h7h6 g5h4 e8g8 h4e7 f8e8 e7d6 b8a7 c2c3 e6e5 e2c4 a7b6 c4b3 d7f6 d6e5 f6g4 e5d4 b6g6 h2h3 g4f6 d4f6 g6f6 c3f6 g7f6');
INSERT INTO evaluations VALUES(45531,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(45531,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(45531,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(45531,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(45531,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(45531,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(45531,36,'cp -43','e1g1 f8g7 c2c3 e8g8 f1e1 b8c6 h2h3 c6a5 c4b3 a5b3 a2b3 b7b5 b3b4 c8b7 c1e3 d8c7 b4c5 d6c5 g5f3 f8d8 d1e2 f6d7 b1d2 e6e5 b2b4 a7a5 d2b3 c5b4 c3b4 a5a4 e1c1 c7b8 b3a5 d8c8 c1c8 b8c8 a1c1');
INSERT INTO evaluations VALUES(45530,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(45530,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(45530,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(45530,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(45530,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(45530,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(45530,36,'cp -34','b8c6 g5f3 f8g7 c2c3 e8g8 f1e1 c6a5 c4b3 a5b3 a2b3 e6e5 d3d4 e5d4 c3d4 c5d4 f3d4 f8e8 b1c3 d6d5 e4e5 f6g4 f2f4 f7f6 e5e6 f6f5 h2h3 d8b6 c3e2 g4f6 c1e3 b6d8 d4f3 c8e6 a1a7 a8a7 e3a7 b7b6 d1d4 f6d7 d4d2');
INSERT INTO evaluations VALUES(46728,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(46728,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(46728,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(46728,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46728,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(46728,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46728,36,'cp 7','c7c5 e4e5 d6e5 f4e5 f6d7 f1c4 f8g7 e5e6 f7e6 d2d3 e8g8 c4e6 g8h8 e1g1 d7e5 e6c8 e5f3 f1f3 d8c8 c3e4 f8f3 d1f3 c8f5 c1f4 b8c6 c2c3 a8f8 a1f1 f5d5 f3h3 f8f5 e4g5 h7h6 g5e4 c5c4 f4h6 c4d3 f1f5 g6f5 h6g7 h8g7 e4d2 g7f7 h3h5 f7g7');
INSERT INTO evaluations VALUES(46940,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(46940,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(46940,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(46940,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(46940,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(46940,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(46940,36,'cp 54','c4b3 b8c6 g1f3 c8f5 a2a4 c6a5 b3a2 c7c5 e1g1 e7e6 d2d4 f8e7 c1e3 c5c4 f3e5 e8g8 g2g4 f5g6 f2f4 a5c6 a4a5 b6d5 c3d5 d8d5 a2c4 d5e4 f1e1 e4c2 d4d5 c6e5 d1c2 e5f3 g1f2 g6c2 f2f3 f8c8 c4a2 e6d5 a2d5 e7f6 d5b7 f6b2 b7c8');
INSERT INTO evaluations VALUES(46950,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46950,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(46950,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(46950,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46950,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(46950,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(46950,36,'cp 22','e7e6 d1g4 b8d7 g1f3 d7f6 g4h4 f8e7 d2d4 e8g8 c3e4 a7a5 a2a4 b6d7 c2c3 f6e4 h4e4 c7c5 b3c2 d7f6 e4h4 c5d4 f3d4 e6e5 d4f5 c8f5 c2f5 d8c7 f5c2 h7h6 h4g3 g8h8 g3h3 h8g8 c1e3 e7c5 e3c5 c7c5 e1c1 c5f2');
INSERT INTO evaluations VALUES(47421,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(47421,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(47421,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(47421,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(47421,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(47421,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(47421,36,'cp 23','e7e6 d3d4 f8e7 c1f4 e8g8 f1e2 c7c5 d4c5 e7c5 e1g1 b8c6 a2a3 a7a6 c3a4 c5a7 d1d8 f8d8 f4c7 d8e8 a4b6 a7b6 c7b6 f6d5 b6c5 e8d8 a1d1 b7b6 c5e3 d5e3 f2e3 c8b7 b2b4 g8f8 d1d8 c6d8 f1d1 b7f3 e2f3');
INSERT INTO evaluations VALUES(48543,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(48543,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(48543,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(48543,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(48543,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(48543,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(48543,36,'cp 17','d3d4 f8e7 f1e2 e8g8 e1g1 c7c5 d4c5 e7c5 c1f4 b8c6 d1e1 d8e7 a1d1 f8d8 e2d3 c6b4 d3c4 d8d1 e1d1 a7a6 a2a3 b4c6 c4d3 h7h6 b2b4 c5a7 f4d6 e7d8 c3e4 b7b5 f1e1 c8b7 e4c5 d8b6 f3e5 c6d4 d3h7 f6h7 d1d4 a8d8 e1d1');
INSERT INTO evaluations VALUES(48877,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(48877,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(48877,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(48877,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(48877,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(48877,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(48877,36,'cp -67','d3d4 e5e4 c2c4 c7c6 b1c3 f8d6 d1c2 e8g8 c4d5 c6d5 c3b5 d6b4 c1d2 b4d2 c2d2 b8c6 g1e2 c8d7 b5c3 c6e7 a1c1 b7b5 a2a3 d8b6 e2g3 g7g6 f1e2 h7h5 h3h4 a8c8 e1g1 g8g7 c1c2 a7a5 f1c1 b5b4 c3a2 c8c2 c1c2 b4a3 b2a3');
INSERT INTO evaluations VALUES(49067,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(49067,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(49067,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(49067,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(49067,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(49067,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(49067,36,'cp 22','b1d2 e4d2 c1d2 f8e7 f1d3 c7c5 c2c3 b8c6 e1g1 e8g8 f1e1 c8e6 d1e2 d8d7 a1d1 e6g4 h2h3 g4h5 d3b1 d7e6 b2b4 b7b6 a2a3 a8d8 d2f4 h7h6 b1a2 h5g6 a2b3 c5b4 a3b4 g6e4 f3d4 c6d4');
INSERT INTO evaluations VALUES(52612,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(52612,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(52612,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(52612,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(52612,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(52612,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(52612,36,'cp 61','d2d4 c5d4 f3d4 b8c6 f1e2 f8g7 c1e3 c6d4 e3d4 g7d4 d1d4 g8f6 b1c3 e8g8 e1g1 d8b6 d4b6 a7b6 a2a4 h7h5 b2b3 g8g7 f1d1 c8e6 f2f4 f8c8 d1d4 f6d7 b3b4 c8c6 g1f2 a8c8 c3d5 e6d5 e4d5 c6c7 g2g3 f7f5 f2e3 d7f6 e3d2 f6g4 e2g4 h5g4');
INSERT INTO evaluations VALUES(53096,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(53096,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(53096,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(53096,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(53096,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53096,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53096,36,'cp 0','e2e4 g7g6 b1c3 f8g7 f1b5 c8d7 b5c4 b8c6 e1g1 c6a5 c4e2 a5c6');
INSERT INTO evaluations VALUES(54181,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54181,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(54181,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54181,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(54181,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54181,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54181,36,'cp 0','d5c4 f1c4 a7a6 g1f3 g8f6 c4e2 b7b5 e1g1 b8d7 d4c5 f8c5 b2b4 c5e7 c1b2 c8b7 b1d2 e8g8 e2d3 d7c5 d3c2 c5d7');
INSERT INTO evaluations VALUES(54540,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(54540,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(54540,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(54540,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(54540,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(54540,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(54540,36,'cp -27','g8f6 d2d4 c5c4 b3c2 f8d6 h2h3 e8g8 g1e2 b7b5 e1g1 h7h6 c1f4 b8c6 b1d2 a7a5 f4d6 d8d6 e2g3 f8e8 a2a3 c8d7 d2f3 g7g6 d1d2 g8g7 f1e1 e8e1 a1e1 b5b4 a3b4 a5b4 c2b1 a8e8 e1e8 f6e8 d2e3 c6e7 e3d2 e8c7 f3e5 b4b3 g3f1 c7b5 e5d7 d6d7');
INSERT INTO evaluations VALUES(46449,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(46449,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(46449,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(46449,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(46449,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(46449,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(46449,36,'cp -19','d2d4 c5c4 b3c2 f8d6 g1e2 e8g8 h2h3 h7h6 e1g1 f8e8 e2g3 b7b5 a2a3 a7a5 d1f3 b8c6 c1f4 d6f4 f3f4 c8d7 b1d2 d8b8 f4f3 b5b4 c2f5 b8d6 f5d7 d6d7 g3f5 g8h7 a1e1 e8e6 f3g3 a8g8 e1e6 f7e6 f5e3 d7f7 f1e1 f7g6 g3g6 h7g6');
INSERT INTO evaluations VALUES(55963,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(55963,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(55963,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(55963,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(55963,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(55963,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(55963,36,'cp 60','c2c3 c7c5 g1f3 b8c6 f1d3 g8e7 e1g1 c5d4 c3d4 e7f5 c1e3 f8e7 b1c3 e8g8 d1d2 f5e3 f2e3 c8d7 e3e4 d5e4 d3e4 f7f5 e5f6 e7f6 a1d1 g8h8 g1h1 a8c8 f3g5 f6g5 f1f8 d8f8 d2g5 f8f6 g5e3 d7e8 e4f3 e8g6 c3e4 g6e4 f3e4 c8d8');
INSERT INTO evaluations VALUES(57032,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(57032,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(57032,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(57032,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(57032,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(57032,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(57032,36,'cp -34','d7d5 e4e5 c7c5 c2c3 d5d4 d2d3 g8e7 g1f3 b8c6 h4h5 h7h6 g2g3 d4c3 b2c3 b7b5 f1g2 c8b7 e1g1 d8d7 d1e2 a8d8 d3d4 e7d5 e2b5 c6e5 b5d7 e5d7 c1a3 f8e7 f1c1 d5f6 b1d2 c5d4 a3e7 e8e7 c3d4 d8c8 d2b3 c8c1 a1c1 h8c8 c1c8 b7c8');
INSERT INTO evaluations VALUES(55884,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(55884,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(55884,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(55884,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(55884,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(55884,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(55884,36,'cp -27','e4e5 c7c5 c2c3 b8c6 d2d4 g8h6 g1f3 d8b6 f1d3 c8d7 e1g1 c5d4 d1e2 h6g4 c1f4 d4c3 b2c3 f7f6 c3c4 f6e5 c4d5 e6d5 b1c3 c6d4 e2d1 d4f3 d1f3 e5e4 f3g3 b6g6 c3d5 e4d3 d5c7 e8f7 c7a8 f8e7 a8c7');
INSERT INTO evaluations VALUES(1529,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(1529,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1529,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1529,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(1529,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(1529,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(1529,36,'cp -7','a4a5 b6c7 c3d4 c6a5 f1b5 a5c6 b1c3 c8d7 e1g1 g8e7 b5d3 c6b4 c3b5 d7b5 d3b5 e7c6 c1d2 f8e7 d2b4 e7b4 d1a4 b4e7 f1c1 e8g8 b5c6 b7c6 c1c6 c7b7 b2b3 f8b8 g2g3 a7a5 a1c1 e7f8 c1c3 b7b5 a4b5 b8b5 g1g2 h7h6 c6c7 a8b8 c7a7 b8b7 a7a8 b5b3 c3b3 b7b3');
INSERT INTO evaluations VALUES(2729,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2729,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2729,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2729,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(2729,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(2729,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2729,36,'cp 0','c6a5 b3c2 c7c5 b1d2 f8e8 h2h3 a5c6 a2a4 c8d7 d2f1 h7h6 f1g3 a8c8 a4b5 a6b5 c2b3 c6a5 b3c2');
INSERT INTO evaluations VALUES(2740,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2740,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2740,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2740,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2740,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2740,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2740,36,'cp 0','b3c2 c7c5 b1d2 a5c6 a2a4 c8d7 d2f1 h7h6 f1e3 f8e8 c2b3 c6a5 b3a2 b5a4 f3d2 d7c6 d1f3 a8b8 h2h3 e7f8 e3g4 f8e7');
INSERT INTO evaluations VALUES(2743,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2743,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2743,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(2743,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2743,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2743,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(2743,36,'cp 0','c7c5 b1d2 a5c6 a2a4 b5b4 h2h3 f8e8 d2c4 c8e6 f3g5 e6c4 d3c4 b4c3 b2c3 c6a5 d1d3 g7g6 g5f3 a8b8 f3h2 f6h5 g2g3 d8c8 h3h4 c8h3 e1e3 h5f6 e3e1 f6h5');
INSERT INTO evaluations VALUES(2829,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2829,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2829,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2829,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(2829,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2829,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2829,36,'cp 0','b1d2 a5c6 a2a4 c8d7 d2f1 h7h6 f1g3 f8e8 h2h3 a8c8 a4b5 a6b5 c1e3 e7f8 c2b3 c6a5 b3c2');
INSERT INTO evaluations VALUES(2830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2830,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2830,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2830,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2830,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2830,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2830,36,'cp 0','a5c6 a2a4 c8d7 d2f1 h7h6 f1e3 f8e8 a4b5 a6b5 a1a8 d8a8 h2h3 d7e6 c2b3 e6b3 d1b3 e8b8 b3d1 a8a2 f3h4 g7g6 d1f3 f6h7 f3g4 h7f6');
INSERT INTO evaluations VALUES(8553,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(8553,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(8553,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(8553,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(8553,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(8553,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(8553,36,'cp 70','d2d4 e8g8 f1e1 c8d7 b5a4 a7a5 h2h3 f8e8 d4d5 c6b8 c3c4 b8a6 a4d7 d8d7 b1c3 c7c6 c1g5 a6c5 b2b3 h7h6 g5f6 e7f6 a2a3 g7g6 d1c2 f6g7 a1b1 d7e7 b3b4 a5b4 a3b4 c5d7 d5c6 b7c6 b4b5 a8c8 f3d2 h6h5');
INSERT INTO evaluations VALUES(8545,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(8545,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(8545,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(8545,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(8545,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(8545,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(8545,36,'cp 51','e8g8 f1e1 c8d7 b5f1 d7g4 d4d5 c6b8 b2b4 b8d7 h2h3 g4h5 b1d2 c7c6 c3c4 a7a5 d5c6 b7c6 b4b5 d7c5 c1a3 h5f3 d1f3 c5e6 f3e3 f6d7 d2f3 c6c5 g2g3 e6d4 e3d3 d7b6 f3d4 c5d4 h3h4 d8c7 e1c1 f8e8 f1h3 e7f8 a1b1 a8b8');
INSERT INTO evaluations VALUES(14014,30,'cp 362',NULL);
INSERT INTO evaluations VALUES(14014,31,'cp 367',NULL);
INSERT INTO evaluations VALUES(14014,32,'cp 367',NULL);
INSERT INTO evaluations VALUES(14014,33,'cp 363',NULL);
INSERT INTO evaluations VALUES(14014,34,'cp 383',NULL);
INSERT INTO evaluations VALUES(14014,35,'cp 355',NULL);
INSERT INTO evaluations VALUES(14014,36,'cp 369','d5c6 b7c6 c1f4 d8b6 a1d1 c8d7 f1c4 f6g4 f4d6 f8d6 d4b6 a7b6 d1d6 e8e7 d6d2 g4e5 c4b3 g7g5 f2f3 b6b5 h2h4 h7h6 e1f2 d7e6 h4g5 h6g5 h1h8 a8h8 b3e6 f7e6 b2b3 b5b4 c3a4 e5d7 f2e3 c6c5 c2c3 b4c3 a4c3 h8h1 d2c2 d7e5');
INSERT INTO evaluations VALUES(17287,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17287,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(17287,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(17287,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17287,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17287,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17287,36,'cp 0','e7e5 g1f3 e5e4 f3d4 f8c5 d2d3 c6d4 e3d4 c5b6 b1d2 e8g8 d3e4 f6e4 d2e4 f8e8 f1e2 e8e4 e1g1 d7d5 e2d3 e4e6 d1d2 a7a5 a2a4 c7c6 a1e1 d8f6 g2g3 g7g6 c2c3 f6f3 d3e2 f3f6');
INSERT INTO evaluations VALUES(18458,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(18458,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(18458,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(18458,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(18458,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(18458,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(18458,36,'cp -26','e7e5 e1g1 f8e7 b1c3 g8f6 f3d2 d8c7 f1e1 e8g8 d2c4 c8e6 f2f4 e6g4 d1d2 e5f4 d2f4 g4e6 c4d6 c7d6 f4d6 e7d6 e4e5 d6e7 e5f6 e7f6 c1f4 c5c4 d3c4 f6d4 f4e3 d4c3 b2c3 e6c4 a2a3 a7a5 h2h3 f7f6 a1d1 f8e8 d1d6 e8e6 d6e6 c4e6 e3d4');
INSERT INTO evaluations VALUES(20576,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(20576,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(20576,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(20576,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(20576,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(20576,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(20576,36,'cp 87','g8f6 d2d4 c5d4 c3d4 e5d4 f3d4 c8d7 b1c3 f8e7 c1e3 e8g8 a1c1 a8c8 d4c6 b7c6 b5a6 c8b8 b2b3 d7e6 a6d3 f6g4 e3f4 e7g5 d1d2 h7h6 d3e2 g5f4 d2f4 g4f6 f1d1 d6d5 e2f3 d8a5 f4d2 d5e4 c3e4 a5d2 e4f6 g7f6');
INSERT INTO evaluations VALUES(26459,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(26459,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(26459,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26459,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(26459,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(26459,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(26459,36,'cp 39','g7g5 f4g3 e7e6 f1d3 f5d3 d1d3 f8d6 b1d2 g5g4 f3g1 f6f5 g1e2 d8d7 h2h4 g8f6 e1g1 c6e7 c3c4 d6g3 f2g3 e8g8 a1e1 b7b6 e1c1 g8g7 c4d5 e6d5 d3c2 c7c6 e2f4 a7a5 g1h2 f8e8 a2a3 e7g6 c2c6 d7c6 c1c6 g6f4 e3f4');
INSERT INTO evaluations VALUES(26868,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(26868,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(26868,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(26868,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(26868,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(26868,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(26868,36,'cp 62','h2h3 g4e6 e1g1 f8d6 f1e1 g8e7 b1c3 e8g8 c3b5 e7g6 b5d6 d8d6 a2a4 e6f5 c2c3 a7a5 f3h2 f8e8 h2f1 h7h6 f1g3 f5d7 b2b3 g6h4 d1d2 c6e7 d2f4 e7f5 f4d6 f5d6 c1f4 h4g6 f4d2 f7f5 f2f3 f5f4 g3h5 e8f8 g1f2 a8e8 e2d3 e8e1 a1e1 b7b6');
INSERT INTO evaluations VALUES(29070,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(29070,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(29070,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29070,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29070,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29070,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(29070,36,'cp 6','g2g4 f5g6 h2h4 h7h5 g4g5 f6d5 g1e2 b8d7 b1c3 d5c3 e2c3 d8a5 d1e2 e7e6 c1d2 e8c8 e1c1 d7b6 d2f4 a5b4 a2a3 b4a5 c3b1 a5f5 f4g3 b6d5 h1e1 f8d6 g3d6 d8d6 f3e4 f5e4 e2e4 g6e4 e1e4 f7f5 g5f6 g7f6 b1c3 d5c3 b2c3 h8d8 c3c4');
INSERT INTO evaluations VALUES(29970,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(29970,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(29970,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(29970,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(29970,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(29970,35,'cp 203',NULL);
INSERT INTO evaluations VALUES(29970,36,'cp 212','d1e2 b8d7 c1g5 h7h6 g5e3 c7c6 e1c1 d8c7 h1e1 e8c8 c1b1 c8b8 f2f3 a7a6 a2a4 d7b6 d4b3 f8e7 a4a5 b6c8 e2c4 h8f8 e3f2 f6d7 f3f4 g7g5 g2g3 h6h5 c4e2 h5h4 c3a4 c6c5 f4g5 e7g5 g3h4 g5f6 e2f3 c7c6 a4c3 f6c3 f3c3 d8e8 c3d3');
INSERT INTO evaluations VALUES(29969,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(29969,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(29969,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(29969,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(29969,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(29969,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(29969,36,'cp 182','b8c6 c1e3 f8e7 d4f5 g7g6 f5e7 d8e7 e3g5 e7e5 f2f4 e5e6 e1g1 e8c8 f4f5 e6e5 e2b5 e5d4 g1h1 f6e4 g5d8 h8d8 c3e4 d4e4 f5g6 f7g6 b5b3 c8b8 f1f7 h7h5 b3c3 d8c8 a1f1 a7a6 h2h3 g6g5 c3d3 e4e6 d3b3 e6b3 a2b3 g5g4 f7f5 g4h3 g2h3 h5h4');
INSERT INTO evaluations VALUES(30464,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(30464,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(30464,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(30464,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(30464,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(30464,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(30464,36,'cp 114','c6b5 b2b3 e7e6 c1d2 g8f6 b3c4 b5c4 g1f3 f8d6 f1c4 e8g8 e1g1 b8d7 d2a5 d8e7 d1e2 d6b4 a5b4 e7b4 f1b1 b4e7 c4a6 a8b8 b1c1 e7d6 e2a2 h7h6 h2h3 f6d5 a6e2 b8a8 a2a5 f8b8 f3d2 d6b4 d2e4 b4b7 e2f3 b7b6 e4c5 d7c5 c1c5 b6a5');
INSERT INTO evaluations VALUES(30471,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(30471,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(30471,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(30471,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(30471,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(30471,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(30471,36,'cp 137','b2b3 e7e6 c1d2 f8d6 b3c4 b5c4 f1c4 g8f6 g1f3 e8g8 e1g1 b8d7 c4a6 a8b8 h2h3 d6b4 d1c2 b4d2 f3d2 h7h6 f1c1 f6d5 a6c4 a7a5 c4d5 e6d5 c2a2 b8b5 d2b1 d7f6 b1c3 b5b4 c1d1 a5a4 c3a4 d8a5 a4c3 a5a2 a1a2 f8b8');
INSERT INTO evaluations VALUES(33407,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(33407,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(33407,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(33407,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(33407,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(33407,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(33407,36,'cp 19','e8g8 d1b3 c7c5 c4d5 e6d5 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 c8d7 f2f3 f8c8 d1b1 c8c5 b1b7 c5c3 e1d2 c3a3 b7b2 a8c8 f1e2 f6e8 e2b5 e8f6 h1b1 d7b5 d4b5 a3a4 h2h3 c8c5 b1c1 c5c1 d2c1 f6d7 b5c3 a4c4 c1d2 d5d4 e3d4');
INSERT INTO evaluations VALUES(35376,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35376,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(35376,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(35376,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(35376,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(35376,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(35376,36,'cp 22','d1b3 c7c5 c4d5 e6d5 d4c5 b4c3 b3c3 f6c3 b2c3 b8d7 a1d1 d7f6 f3d4 c8d7 d4b3 f8c8 f2f3 d7a4 f1e2 g7g6 h2h4 a4b3 a2b3 c8c5 e1d2 a8c8 d1a1 a7a6 h1c1 h6h5 a1a4 g8g7 e2d3 c8c7 c1c2 c5c6 a4a5 b7b5 d3e2 f6d7');
INSERT INTO evaluations VALUES(34450,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(34450,31,'cp 260',NULL);
INSERT INTO evaluations VALUES(34450,32,'cp 270',NULL);
INSERT INTO evaluations VALUES(34450,33,'cp 273',NULL);
INSERT INTO evaluations VALUES(34450,34,'cp 265',NULL);
INSERT INTO evaluations VALUES(34450,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(34450,36,'cp 249','c1g5 f6d6 f3e5 d6d1 a1d1 f7f6 g5h4 g8h6 e5d3 c8e6 f2f3 b8d7 h4f2 a7a5 f1e2 f8b4 f2d4 h6f7 d3f4 f7d8 e1f2 e6f7 h2h4 d8e6 f4e6 f7e6 a2a3 b4e7 f3f4 b7b5 g2g4 e8c8 g4g5 c8c7 e2f3 d7b6 f4f5 e6f7 d4e3 d8d1 h1d1 b6c4');
INSERT INTO evaluations VALUES(34449,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(34449,31,'cp 254',NULL);
INSERT INTO evaluations VALUES(34449,32,'cp 273',NULL);
INSERT INTO evaluations VALUES(34449,33,'cp 246',NULL);
INSERT INTO evaluations VALUES(34449,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(34449,35,'cp 243',NULL);
INSERT INTO evaluations VALUES(34449,36,'cp 253','f6d6 f3e5 d6d1 a1d1 f7f6 g5h4 g8h6 e5d3 c8e6 f2f3 b8d7 h4f2 a7a5 h2h4 f8e7 b2b3 e7b4 e1d2 h6f7 a2a4 b4d6 d2c1 e8c8 c3e2 d7e5 d3b2 d6b4 d1d8 h8d8 e2d4 e6d7 h4h5 f7d6 f1e2 c8c7 h1d1 d6f7 g2g4 b4d6 b2c4 e5c4 e2c4');
INSERT INTO evaluations VALUES(34464,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(34464,31,'cp 258',NULL);
INSERT INTO evaluations VALUES(34464,32,'cp 271',NULL);
INSERT INTO evaluations VALUES(34464,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(34464,34,'cp 266',NULL);
INSERT INTO evaluations VALUES(34464,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(34464,36,'cp 247','f3e5 d6d1 a1d1 f7f6 g5h4 g8h6 e5d3 c8e6 f2f3 b8d7 h4f2 a7a5 f1e2 f8b4 f2d4 b4d6 d4e3 d6b4 e3d2 h6f7 d3f4 f7d8 a2a3 b4e7 e1f2 e6f7 d2e3 d8e6 f4e6 f7e6 f3f4 e8c8 h2h4 b7b5 e2f3 e7c5 h4h5 c5e3 f2e3');
INSERT INTO evaluations VALUES(36299,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(36299,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(36299,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(36299,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(36299,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(36299,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(36299,36,'cp -15','d5c4 d3c4 h7h6 g5f6 e7f6 g1f3 c7c5 c3e4 b8d7 e1g1 c5d4 e4f6 d7f6 f3d4 d8e7 h2h3 c8d7 d1b3 a8c8 f1c1 c8c7 c4b5 f8c8 c1c7 c8c7 a1d1 e6e5 b5d7 e5d4 d7b5 a7a6 b5e2 d4e3 b3e3 e7e3 d1d8 g8h7 f2e3 h7g6 e2d3 g6g5 g2g3 f6d7 d8a8');
INSERT INTO evaluations VALUES(37063,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(37063,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(37063,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(37063,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(37063,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(37063,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(37063,36,'cp 23','b7b6 g1f3 c8a6 b2b3 c7c5 a1c1 b8d7 f1e2 a8c8 c3b2 c5d4 e3d4 c8c7 e1g1 d8e7 f1e1 f8c8 e2f1 h7h6 f3d2 d7f8 a3a4 f8g6 a4a5 a6b7 c1a1 b7a8 d2f3 f6e4 a5a6 e7b4 a1a4 b4d6 f3e5 g6e5 d4e5');
INSERT INTO evaluations VALUES(37560,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(37560,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(37560,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(37560,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(37560,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(37560,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(37560,36,'cp 136','d1g4 d8f6 c1d2 f6g6 g4f4 b4c3 d2c3 c8a6 g1e2 d7d6 f2f3 b8d7 h2h4 h7h5 a2a4 a6e2 e1e2 g8e7 f4g5 d6d5 g5g6 e7g6 a4a5 d5e4 f3e4 e8g8 e2e3 d7f6 f1e2 a8c8 e2f3 c7c5 d4c5 c8c5 c3b4 c5c2 b4f8 g8f8 a5b6 a7b6');
INSERT INTO evaluations VALUES(40509,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(40509,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(40509,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(40509,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(40509,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(40509,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(40509,36,'cp -26','c8f5 e2d3 f5d3 d1d3 e8g8 e1g1 b8d7 a1e1 h7h6 g5f6 d7f6 g2g3 g7g6 e1e2 d8d7 f3e5 d7c7 f1e1 g8g7 e5f3 a8e8 g1g2 c7c8 e2e8 f8e8 e1e8 c8e8 c3d1 d6c7 d1e3 h6h5 c2c3 f6g4 e3f1 e8e4 d3e4 d5e4 f3g5');
INSERT INTO evaluations VALUES(42350,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(42350,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(42350,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(42350,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(42350,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(42350,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(42350,36,'cp -46','d8f6 d1f3 f6e7 b1d2 d7d6 g1e2 b8d7 c2c3 d7f6 d3d4 c5b6 e1c1 c8d7 f3g3 e8c8 f2f3 e5d4 e2d4 f6h5 g3f2 g7g6 h1e1 c8b8 c1b1 h5f4 d2b3 h6h5 f2d2 e7f6 g2g3 f4e6 f3f4 e6d4 c3d4 h5h4 d2e2 h4g3 h2g3 h8e8 e4e5 f6e7 c4d3 d7e6');
INSERT INTO evaluations VALUES(33448,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(33448,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(33448,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(33448,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(33448,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(33448,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(33448,36,'cp -51','d1f3 f6e7 g1e2 d7d6 b1d2 b8d7 c2c3 d7f6 d3d4 c5b6 e1c1 c8d7 f3g3 e8c8 f2f3 e5d4 e2d4 f6h5 g3f2 h8e8 h1e1 g7g6 f3f4 c7c6 c4e2 h5f6 e2f3 b6c7 c1b1 c8b8 h2h3 h6h5 a2a3');
INSERT INTO evaluations VALUES(44214,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44214,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(44214,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44214,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44214,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(44214,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44214,36,'cp 0','d7d5 g1f3 d5c4 f1c4 a7a6 e1g1 b7b5 c4e2 b8d7 d4c5 f8c5 b2b4 c5e7 c1b2 e8g8 b1d2 c8b7 a1c1 a8c8 c1c8 d8c8 d1a1 c8a8 f1c1 b7d5 h2h3 h7h6 e2d3 f8c8 c1c8 a8c8 b2d4 c8a8 d3e2 e7d6 d4b2 d5b7 f3e1 b7d5');
INSERT INTO evaluations VALUES(44377,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(44377,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(44377,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(44377,33,'cp -99',NULL);
INSERT INTO evaluations VALUES(44377,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(44377,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(44377,36,'cp -107','d7d5 e2e4 d5d4 e4e5 d4c3 d1d8 e8d8 e5f6 g7f6 b2c3 b8c6 g1e2 b7b6 e2g3 c8a6 c1f4 e6e5 f4c1 c6a5 g3h5 d8e7 f2f4 h7h6 g2g4 h8g8 f1g2 a8d8 h2h3 a6c4 e1f2 c4f7 h5g3 f7g6 f4e5 f6e5 g4g5 f8g7 h1e1 d8d3 c1f4 h6g5 f4g5 e7d7 a1d1 d7c7');
INSERT INTO evaluations VALUES(46527,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(46527,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(46527,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(46527,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(46527,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(46527,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(46527,36,'cp -87','e2e4 d5d4 e4e5 d4c3 d1d8 e8d8 e5f6 g7f6 b2c3 b8c6 g1e2 b7b6 e2g3 c8a6 f1e2 d8e7 c1f4 a8d8 e1g1 c6a5 f1e1 e6e5 f4e3 a6c4 f2f4 c4e2 e1e2 d8d5 a1f1 e7e6 g3e4 f8e7 f4e5 f6e5 e3g5 h8f8 f1f8 e7f8 e4f6 d5d3 f6h7 f8g7 h2h4 d3c3 h4h5 c3d3');
INSERT INTO evaluations VALUES(47728,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(47728,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(47728,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(47728,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(47728,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(47728,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(47728,36,'cp -51','f8b4 e4d5 d8d5 g1f3 e8g8 f1e2 d5a5 d1d2 c8f5 e1g1 b8c6 a2a3 b4c3 b2c3 f5e4 f1b1 f8e8 a3a4 a5d5 c3c4 d5a5 d2a5 c6a5 e2d3 e4f3 g2f3 a8d8 c2c3 c7c5 d4d5 b7b6 h2h4 g8f8 g1g2 h6h5 f3f4 g7g6 f4f5 g6f5 b1e1 e8e5 g2f3');
INSERT INTO evaluations VALUES(48369,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48369,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(48369,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48369,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48369,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48369,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48369,36,'cp 0','f8d6 c2c4 c7c6 b1d2 e8g8 e2e3 b8d7 f1d3 f8e8 e1g1 e6e5 d4e5 d7e5 f3e5 e8e5 c4d5 c8h3 g2h3 e5d5 d3c2 d5g5 g1h1 d8c8 d1f3 g5h5 h1g1 h5g5');
INSERT INTO evaluations VALUES(50295,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(50295,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(50295,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50295,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50295,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50295,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50295,36,'cp 0','d7d5 g1f3 c7c5 e2e3 c5d4 e3d4 g7g5 h4g3 b8c6 c2c3 f6e4 b1d2 e4g3 h2g3 g5g4 f3e5 c6e5 d4e5 h6h5 f1b5 c8d7 b5d7 d8d7 d2b3 f8g7 d1d4 d7f5 d4d5 f5e5 d5e5 g7e5 e1c1 e7e6 h1e1 e5c7 b3c5 c7b6 e1e5 b6c7 e5e1');
INSERT INTO evaluations VALUES(51014,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(51014,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(51014,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(51014,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(51014,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(51014,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(51014,36,'cp 97','d2d4 b7b5 c1e3 d7d6 f1d3 c7c5 d4c5 d6c5 a2a4 b5b4 c3b1 b8c6 b1d2 g8f6 d2c4 f8e7 f3d2 e6e5 a4a5 c8e6 c4b6 a8b8 d2c4 e8g8 e1g1 f6g4 d1e2 g4e3 f2e3 e7g5 b6d5 e6d5 e4d5 d8d5 e2g4 g7g6 a1d1 e5e4 g4e4 d5e4 d3e4');
INSERT INTO evaluations VALUES(51796,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(51796,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(51796,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(51796,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(51796,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(51796,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(51796,36,'cp 77','e2e4 d7d5 g1f3 f8b4 e4e5 g8e7 a2a3 b4c3 b2c3 d8a5 a3a4 c6c5 d1d2 h7h6 f1b5 c8d7 b5e2 b8c6 c1a3 c5d4 c3d4 a5d2 e1d2 b7b6 h1b1 c6a5 c2c3 f7f6 f3h4 h8g8 g2g3 f6e5 a3e7 e8e7 h4g6 e7d6 g6e5 a8f8 d2e3 f8c8 e2a6 c8c3 e3d2');
INSERT INTO evaluations VALUES(52867,30,'cp -136',NULL);
INSERT INTO evaluations VALUES(52867,31,'cp -116',NULL);
INSERT INTO evaluations VALUES(52867,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(52867,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(52867,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(52867,35,'cp -125',NULL);
INSERT INTO evaluations VALUES(52867,36,'cp -112','e7e5 c2c3 g8f6 g1f3 f8d6 d2d3 e8g8 g2g3 c8g4 f1e2 g4e6 b4c2 c6c5 c1g5 b8d7 e1g1 b7b5 g5d2 h7h6 a2a4 b5a4 a1a4 a7a5 c3c4 d7b8 f3h4 b8c6 c2a3 e6h3 f1e1 f8e8 a3b5 d6f8 a4a1 c6b4 h4g2 g7g5 f2f4 d8d7 g1h1 e5f4 g3f4 g8h8');
INSERT INTO evaluations VALUES(53782,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(53782,31,'cp -112',NULL);
INSERT INTO evaluations VALUES(53782,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(53782,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(53782,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(53782,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(53782,36,'cp -101','c2c3 g8f6 g1f3 f8d6 d2d3 e8g8 g2g3 c8g4 b4c2 c6c5 f1e2 g4d7 e1g1 b8c6 a2a4 a7a6 c2a3 h7h6 c1d2 d7h3 a3c4 d4c3 d2c3 h3f1 e2f1 d8c7 a4a5 a8d8 f1h3 f8e8 c4b6 c6d4 c3d4 e5d4');
INSERT INTO evaluations VALUES(53888,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(53888,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53888,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(53888,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(53888,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(53888,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(53888,36,'cp 45','f1g2 b8c6 a2a3 a7a5 d2d3 g7g6 c3d5 h7h6 c1d2 g8f6 d5f6 d8f6 e2e3 f8g7 d1b3 f6d8 g1e2 c8d7 e2c3 e8g8 e1g1 a8b8 a1b1 d7e6 c3d5 b7b6 b3c2 c6e7 b2b4 a5b4 a3b4 e7d5 c4d5 e6d7 b4c5 b6c5 b1b8 d8b8 f1b1');
INSERT INTO evaluations VALUES(57664,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(57664,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(57664,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(57664,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(57664,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(57664,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(57664,36,'cp 13','d7d5 e4d5 f8g7 f1c4 g8f6 b1c3 c7c6 d5c6 b8c6 g1e2 e8g8 d2d4 d8b6 c4b3 e7e5 d4d5 c6d4 e1g1 f8e8 e2g3 h7h6 c3e4 f6e4 g3e4 f7f5 e4g3 f5f4 g3e4 c8f5 f1e1 b6d8 d5d6 g8h7 c2c3 d4b3 a2b3 d8h4 f2f3');
INSERT INTO evaluations VALUES(291,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(291,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(291,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(291,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(291,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(291,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(291,36,'cp 4','b1c3 g8e7 f5e3 d7d6 f1e2 c8d7 e1g1 e8c8 a1b1 f6e5 e2d3 c6b4 f1e1 b4d3 d1d3 f7f5 b2b4 c5d4 c1d2 f5e4 c3e4 h8e8 a2a4 d6d5 e4c3 e5f6 c3b5 d4b6 a4a5 a7a6 d2c3 d7b5 c3f6 b5d3 a5b6 g7f6 c2d3 c7b6 b4b5 a6b5 b1b5 e7c6 b5b6 c8c7');
INSERT INTO evaluations VALUES(290,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(290,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(290,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(290,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(290,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(290,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(290,36,'cp 19','g8e7 f5e3 d7d6 f1e2 c8d7 e1g1 e8c8 a1b1 f6e5 e2d3 f7f5 e4f5 c5e3 c1e3 d7f5 f1e1 e5f6 f2f3 h8e8 c3e4 f6g6 e3f2 d6d5 e4c5 f5d3 c5d3 d5d4 d1d2 e7d5 e1e8 d8e8 b1e1 b7b6 g1f1 g6f7 e1e8 f7e8 d2e2 e8e2 f1e2 c8d7 h2h3 d7d6');
INSERT INTO evaluations VALUES(1719,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(1719,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(1719,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1719,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1719,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(1719,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(1719,36,'cp 7','d5e6 f3e5 c6e5 f4e5 e6e5 d1e2 e5e2 f1e2 c8d7 d2d4 e8c8 h1f1 f7f6 c1f4 g8e7 e1c1 h7h5 e2d3 d7e8 c3e2 e8g6 c1d2 h5h4 c2c3 g6d3 d2d3 e7g6 d1e1 f8e7 f4d2 e7d6 g2g3 h4g3 h2g3 c7c6 f1h1 d8e8 e1f1 h8h1');
INSERT INTO evaluations VALUES(1688,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(1688,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(1688,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(1688,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(1688,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1688,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1688,36,'cp 17','f1e2 f8c5 f3e5 c6e5 d2d4 e5f3 g2f3 c5b6 a2a4 a7a5 d1d3 g8e7 c1e3 c8d7 e1c1 e8c8 h1e1 e7f5 e3g1 e6d6 c1b1 h8e8 d3d2 c8b8 e2c4 e8e1 d2e1 f7f6 c3d5 d8e8 e1c3 d7a4 d5b6 d6b6 d4d5 b6b4 c3b4 a5b4 b2b3 a4d7 b1b2 f5e3 d1d2');
INSERT INTO evaluations VALUES(4592,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(4592,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(4592,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(4592,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(4592,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(4592,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(4592,36,'cp 41','d2d3 a7a6 c1e3 f6g4 e3d2 e7c5 e1g1 g4f6 d2g5 h7h6 g5h4 c5e7 h2h3 d7d6 h4g3 c8e6 c3d5 e6d5 e4d5 c6d4 f3d4 e5d4 g3f4 f6d7 f4d2 e7f6 d1h5 f6g5 f2f4 g5f6 h5g4 b7b5 c4b3 a6a5 a1e1');
INSERT INTO evaluations VALUES(6823,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(6823,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(6823,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(6823,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(6823,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(6823,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(6823,36,'cp 28','d2d4 c7c6 b5a4 c5d6 d4e5 d6e5 e1g1 g8f6 b1d2 e8g8 d2f3 f8e8 f1e1 d7d6 f3e5 d6e5 d1d8 e8d8 f2f3 c8e6 a4b3 e6b3 a2b3 f6d7 c1e3 a7a6 e1d1 f7f6 g2g3 d7f8 g1f2 g8f7 f3f4 e5f4 g3f4 d8e8 f2f3 a8d8 f4f5 f8d7 e3d4 d7e5 d4e5 d8d1 a1d1 e8e5');
INSERT INTO evaluations VALUES(6822,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(6822,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(6822,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(6822,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(6822,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(6822,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(6822,36,'cp 25','c7c6 b5a4 c5d6 d4e5 d6e5 e1g1 g8f6 b1d2 e8g8 d2f3 f8e8 f1e1 d7d6 f3e5 d6e5 d1d8 e8d8 f2f3 c8e6 a4b3 e6b3 a2b3 f6d7 g1f2 f7f6 c1e3 a7a6 g2g3 d7f8 f3f4 f8e6 f2f3 g8f7 e1d1 d8d1 a1d1 e5f4 g3f4 a8d8 d1d8 e6d8');
INSERT INTO evaluations VALUES(6938,30,'cp 246',NULL);
INSERT INTO evaluations VALUES(6938,31,'cp 242',NULL);
INSERT INTO evaluations VALUES(6938,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(6938,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(6938,34,'cp 227',NULL);
INSERT INTO evaluations VALUES(6938,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(6938,36,'cp 263','c3d5 c6a5 c4e2 a5c6 f3h4 b4a5 e1g1 c6e7 b2b4 e7d5 b4a5 d5f4 d2d4 d7d6 c1f4 e5f4 e2c4 e8f8 f1e1 g7g5 h4f5 h7h5 d1f3 h8h7 a5a6 b7b6 a1b1 c7c6 c4d3 d8c7 h2h4 g8h6 c2c3 h6f7 h4g5 f6g5 e4e5');
INSERT INTO evaluations VALUES(8830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8830,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8830,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8830,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8830,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(8830,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(8830,36,'cp -8','c6a5 c4b3 e8g8 h2h3 c7c6 d3d4 a5b3 a2b3 e5d4 f3d4 f8e8 d4f3 a7a5 c1f4 c5b4 d1d3 c8e6 f4d2 f6d7 c3e2 d7e5 f3e5 d6e5 d3d8 e8d8 d2b4 a5b4 a1a8 d8a8 e2c1 a8d8 g1f1 c6c5 f1e2 d8d4 e2e3 f7f5 e4f5 e6f5 c1d3 f5d3');
INSERT INTO evaluations VALUES(10735,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(10735,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(10735,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(10735,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(10735,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(10735,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(10735,36,'cp 136','f1d3 e8g8 e1g1 f8e8 f1e1 h7h6 b1d2 e5d4 c3d4 d6f8 h2h3 d7d6 a2a3 c6e7 d2f1 c7c5 d4d5 e7g6 f1g3 a7a6 a3a4 f6d7 c1d2 d7e5 f3e5 d6e5 a4a5 g6f4 d3c4 c8d7 a1c1 f8d6 g1h2 g7g6 d1f3 d7b5 c4b5 a6b5');
INSERT INTO evaluations VALUES(13169,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(13169,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13169,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(13169,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(13169,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(13169,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(13169,36,'cp 17','f1d3 c6b4 e1g1 b4d3 d1d3 e7e6 f3e5 f8e7 f1e1 c8d7 c1g5 e8g8 b1d2 h7h6 g5h4 f6e8 h4e7 d8e7 b2b3 e8f6 d3g3 f8c8 c2c4 d5c4 b3c4 e7d8 d2e4 d7c6 e4f6 d8f6 e1e3 c8d8 g3g4 f6g5 g4g5 h6g5');
INSERT INTO evaluations VALUES(14278,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(14278,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(14278,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(14278,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(14278,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(14278,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(14278,36,'cp 0','d8d6 c3c4 e7e5 c4d5 c6d4 e2e4 c8g4 f1d3 c7c6 h2h3 g4f3 d2f3 d4f3 d1f3 c6d5 e1g1 d5d4 d3b5 f6d7 f3g3 e8c8 c1g5 f8e7 g5d2 c8b8 b5d7 d8d7 g3g7 e7f6 g7h6 f6e7');
INSERT INTO evaluations VALUES(14289,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(14289,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14289,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14289,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14289,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14289,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14289,36,'cp 0','g1f3 c8f5 a2a3 e7e6 b1d2 h7h5 c2c4 a7a5 b2b3 f6e4 c1b2 h5h4 f3h4 h8h4 g3h4 e4f2 e1f2 d8h4 f2g1 h4g5 g1f2');
INSERT INTO evaluations VALUES(14297,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14297,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14297,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14297,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14297,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14297,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14297,36,'cp 0','c8f5 a2a3 e7e6 b1d2 f8e7 b2b3 h7h6 c1b2 a7a5 f1d3 f5d3 c2d3 e8g8 d1e2 c6a7 a1c1 f6d7 e1g1 c7c5 d4c5 e7c5 g1g2 a8c8 e3e4 f8e8 d3d4 c5b6 h2h4 c8c1 f1c1 d8a8 d2f1 d7f6 e4e5');
INSERT INTO evaluations VALUES(15079,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(15079,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(15079,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(15079,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(15079,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(15079,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(15079,36,'cp -94','c8e6 f3d2 d8d7 c2c4 d5c4 g5f6 g7f6 d2c4 e8c8 d1a4 e6d5 g2d5 d7d5 h1f1 d5e6 b1c3 a7a6 c4e3 f6f5 e1c1 h7h5 e3c2 c6d4 e2e3 b7b5 a4a5 d4c2 c1c2 c8b7 c2b1 f8c5 h2h4 h8g8 a2a4 c5b6 a5b4 c7c5 b4a3 b5b4');
INSERT INTO evaluations VALUES(16946,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(16946,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16946,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16946,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16946,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16946,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16946,36,'cp -4','f8e7 f1d3 e8g8 b1c3 e7d6 e5g5 h7h6 g5h4 d6e5 e1g1 f8e8 f1e1 c7c5 c3d5 f6d5 h4d8 e8d8 e4d5 d7d6 a2a4 b7b6 c2c3 a7a6 g2g3 c8d7 c1e3 b6b5 a4b5 a6b5 a1a8 d8a8 e3c5 e5c3 e1e7 d7h3 e7a7 a8a7 c5a7 c3b2 d3b5 b2a3 f2f3 g8f8 a7e3 g7g5 g1f2 f8e7 e3d4');
INSERT INTO evaluations VALUES(10974,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10974,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10974,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10974,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10974,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10974,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10974,36,'cp 0','f1d3 e8g8 b1c3 e7d6 e5g5 f8e8 e1g1 d6e5 f2f4 h7h6 g5h4 e5d4 g1h1 d7d6 c1d2 d6d5 e4d5 f6d5 a1e1 d5c3 e1e8 d8e8 f1e1 e8f8 d2c3 d4c3 b2c3 c8d7 e1e7 d7c6 f4f5 a8e8 f5f6 g7f6 h4g4 f8g7 e7e8 c6e8 g4c8 g7f8 c8g4');
INSERT INTO evaluations VALUES(17034,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(17034,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(17034,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(17034,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(17034,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(17034,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(17034,36,'cp 44','d1d4 f8b4 a2a3 b4c3 d4c3 e8g8 c1g5 f8e8 e2e3 d7d6 g5f6 d8f6 c3f6 g7f6 e1c1 e8e5 d1d4 c8d7 f1d3 f6f5 c1d2 a7a5 b2b3 g8g7 h1c1 b7b6 d2e2 a8e8 d4f4 e8d8 f4h4 d8e8 e2f1 d7c6 d3e2 c6e4 b3b4 h7h6 f2f3 e4b7');
INSERT INTO evaluations VALUES(17479,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(17479,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(17479,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(17479,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(17479,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(17479,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(17479,36,'cp 14','e7e6 e1g1 d7d5 e4d5 e6d5 c4b3 f8d6 d2d4 c5c4 f1e1 g8e7 b3c2 d8c7 b1d2 e8g8 d2f1 c8f5 c1g5 f5c2 d1c2 b7b5 e1e2 a6a5 a1e1 b5b4 c2a4 e7f5 f3h4 f5h4 g5h4 c7d7 h4g3 h7h6 f1e3 f8d8 g3d6 d7d6 e3f5');
INSERT INTO evaluations VALUES(17506,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(17506,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(17506,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(17506,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(17506,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(17506,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(17506,36,'cp 7','d2d4 c5d4 f3d4 e7e6 b1c3 c6d4 d1d4 b7b5 c1e3 c8b7 e1c1 g8f6 c1b1 a8c8 e4e5 f6d5 c3d5 b7d5 d4d2 f8c5 f1d3 e8g8 h1e1 d8b6 b2b3 a6a5 f4f5 e6f5 d3b5 d5e4 b5d3 e4d3 d2d3 c5e3 d3e3 b6c7 c2c4 a5a4 d1d4 a4b3');
INSERT INTO evaluations VALUES(18427,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(18427,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(18427,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(18427,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(18427,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(18427,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(18427,36,'cp 99','e1g1 e7e6 f1e1 f8e7 c2c3 g8f6 e4e5 f6d5 d2d4 d7d6 c4b5 c8b7 d1a4 d8d7 e5d6 a7a6 f3e5 a6b5 a4a8 b7a8 e5d7 e7d6 d7e5 c5d4 e5f3 e8e7 f3d4 c6d4 c3d4 a8c6 b1a3 d6a3 b2a3 h8a8 e1e2 e7d7 f2f3 h7h5 g1f2 f7f6 c1b2 d5f4 d4d5 f4d5 e2c2 b5b4 a3b4 d5b4 c2d2 b4d5');
INSERT INTO evaluations VALUES(20436,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(20436,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(20436,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(20436,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(20436,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(20436,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(20436,36,'cp 0','c3d4 d8b6 d1d2 c8d7 b1c3 a8c8 a1c1 c6a5 f1d3 a5c4 d3c4 c8c4 g1e2 g8e7 e1g1 e7f5 e2g3 f8b4 g3f5 e6f5 d2d3 b4c3 c1c3 c4c3 d3c3 e8g8 f1c1 d7e6 h2h4 h7h6 c3b3 b6b3 a2b3 f8c8 h4h5 c8c1 e3c1 g8h7 f2f4 h7g8 g1f2 e6d7 f2e3 d7e6 c1d2 e6d7 d2b4');
INSERT INTO evaluations VALUES(21721,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(21721,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(21721,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(21721,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(21721,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21721,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(21721,36,'cp 52','e4d5 d8d5 b1c3 d5e6 f1b5 a7a6 b5c6 e6c6 e1g1 f7f6 d2d4 c8g4 d4e5 g4f3 d1f3 c6f3 g2f3 f6e5 f1e1 e8c8 e1e5 f8e7 e5e1 e7h4 c3e4 d8f8 e4g3 g8e7 e1e4 e7g6 f3f4 f8e8 e4e8 h8e8 c1e3 h4e7 a1d1 e7d6 g3e2 e8f8 d1d4 g6h4 d4e4 h4f3 g1h1 g7g6 b2b4 b7b5 e2g3 h7h5 c2c3');
INSERT INTO evaluations VALUES(21579,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(21579,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(21579,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(21579,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(21579,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(21579,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(21579,36,'cp 61','d8d5 b1c3 d5e6 f1b5 a7a6 b5c6 e6c6 e1g1 f7f6 d2d4 c8g4 d4e5 g4f3 d1f3 c6f3 g2f3 f6e5 f1e1 e8c8 e1e5 f8e7 e5e1 e7h4 c3e4 g8f6 c1f4 f6e4 e1e4 h4f6 c2c3 d8d3 g1g2 h8d8 a3a4 b7b6 a1e1 a6a5 e4c4 d8d7 f4e3 d7d6 f3f4 c8d7 c4e4 d3d1 e1d1 d6d1 g2f3 d1b1');
INSERT INTO evaluations VALUES(1731,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(1731,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(1731,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(1731,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(1731,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(1731,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(1731,36,'cp 48','b1c3 d5e6 f1b5 a7a6 b5c6 e6c6 e1g1 f7f6 d2d4 c8g4 d4e5 g4f3 d1f3 c6f3 g2f3 f6e5 f1e1 e8c8 e1e5 f8e7 e5e1 e7h4 c3e4 g8f6 c1f4 f6e4 e1e4 h4f6 c2c3 d8d3 g1g2 h8d8 a1e1 b7b6 a3a4 c8b7 f4e3 d3d1 g2f1 d8d5 e3d4 d1e1 f1e1 f6d4 e4d4 d5d4 c3d4 b7c6 e1d2');
INSERT INTO evaluations VALUES(1730,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(1730,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(1730,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(1730,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(1730,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(1730,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(1730,36,'cp 59','d5e6 f1b5 a7a6 b5c6 e6c6 e1g1 f7f6 d2d4 c8g4 d4e5 g4f3 d1f3 c6f3 g2f3 f6e5 f1e1 e8c8 e1e5 f8e7 e5e1 e7h4 c3e4 g8f6 c1f4 f6e4 e1e4 h4f6 c2c3 d8d3 g1g2 h8d8 a1e1 b7b6 a3a4 c8b7 f4e5 f6e5 e4e5 d3d2 e1e2 a6a5 f3f4');
INSERT INTO evaluations VALUES(22502,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(22502,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(22502,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(22502,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(22502,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(22502,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(22502,36,'cp 110','a2a3 d7d6 b1c3 a7a6 f1d3 g8f6 h2h3 e6e5 c1e3 e5d4 f3d4 c6d4 e3d4 f8e7 f2f4 b7b5 d4f2 e8g8 d1e2 c8b7 e1c1 f6d7 h1e1 f8e8 c1b1 e7f6 e4e5 d6e5 d3f5 b7c6 e2g4 e5f4 e1e8 d8e8 d1d7 f6c3 b2c3');
INSERT INTO evaluations VALUES(29212,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29212,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(29212,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(29212,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(29212,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(29212,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(29212,36,'cp 25','f3g3 b8d7 c1d2 h7h5 h2h4 f7f6 g1e2 d8c7 e1c1 e8c8 a2a3 d7b6 c4b3 e6f5 c3e4 b6d7 f2f4 e5f4 e2f4 d7c5 e4c5 f8c5 h1f1 d5f4 d2f4 c7d7 d1e1 f5g4 f4e3 c5d6 g3f2 a7a6 e3c5 c8b8 c5b4 h8e8 b4d6 d7d6');
INSERT INTO evaluations VALUES(29550,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(29550,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(29550,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(29550,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(29550,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(29550,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(29550,36,'cp -68','e7e6 g1f3 f8e7 d3f5 e6f5 d1d3 d8c8 b2b3 f6e4 c1a3 e7a3 b1a3 b8d7 c2c4 e8g8 e1g1 d7f6 a1c1 f8e8 a3b1 c7c6 h2h3 g7g6 f3e5 c8d8 f1e1 a7a5 a2a4 a8c8 d3d1 c8c7 d1c2 c7e7 c2b2 g8h8 b1c3 e4c3 b2c3 f6e4');
INSERT INTO evaluations VALUES(29695,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(29695,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(29695,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(29695,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(29695,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(29695,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(29695,36,'cp 13','c7c6 h2h3 g4h5 c2c4 e7e6 d2d4 b8d7 c4d5 e6d5 b1c3 f8d6 f3d2 e8g8 d1b3 d7b6 f2f4 h7h6 e2e4 d5e4 g3g4 h5g6 f4f5 g6h7 d2e4 d6c7 c1e3 f6e4 c3e4 b6d5 e3f2 d5f4 a1e1 a8b8 e4c5 g8h8 c5b7');
INSERT INTO evaluations VALUES(30370,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(30370,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(30370,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(30370,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(30370,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(30370,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(30370,36,'cp 12','b5e2 e7e5 a2a4 g8f6 e2c4 f8e7 d2d3 h7h6 e1g1 e8g8 c1d2 b8c6 f1e1 a8b8 h2h3 g8h8 c4b3 c6a5 b3a2 a5c6 a2c4 h8h7 c3d5 f6d5 c4d5 e7f6 c2c3 c6e7 d5b3 e7c6 d2e3 d7e6 f3d2 f6g5 b3e6 f7e6 e1f1');
INSERT INTO evaluations VALUES(31941,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(31941,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(31941,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(31941,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(31941,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(31941,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(31941,36,'cp 176','g2g4 h5g6 h2h4 e7e6 b1d2 f7f6 h4h5 g6c2 d1c2 f6e5 d4e5 g8h6 f1h3 b8c6 d2f3 f8b4 e1f1 d8d7 f1g2 b4a5 c1f4 e8c8 e2e3 a5b6 a1c1 d8f8 f3g5 f8e8 a2a3 h6f7 g5f3 h7h6 b2b4 c8b8 g2h2 e8f8 h1d1 f7g5 f3h4 c6d4 d1d4 g5h3 h2h3 b6d4 h4g6 d4b6');
INSERT INTO evaluations VALUES(31940,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(31940,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(31940,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(31940,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(31940,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(31940,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(31940,36,'cp 167','h5g6 h2h4 e7e6 b1d2 f7f6 h4h5 g6c2 d1c2 f6e5 d4e5 g8h6 f1h3 b8c6 d2f3 f8b4 e1f1 d8d7 f1g2 b4a5 c1f4 e8c8 e2e3 d8f8 a1c1 a5b6 a2a3 h6f7 b2b4 h7h6 g2h2 f7g5 f3h4 g5h3 h2h3 f8f7 h4g6 h8d8 h1d1 d5d4 f2f3 d4d3 c2c4 c6e7 h3g2 d3d2 c1c2 e7g6');
INSERT INTO evaluations VALUES(31807,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(31807,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(31807,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(31807,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(31807,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(31807,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(31807,36,'cp 171','h2h4 e7e6 b1d2 f7f6 h4h5 g6c2 d1c2 f6e5 d4e5 g8h6 f1h3 b8c6 d2f3 f8b4 e1f1 d8d7 f1g2 b4a5 c1f4 a5b6 f4g3 h6f7 g4g5 e8c8 c2d2 d8f8 b2b4 f7e5 f3e5 c6e5 g3e5 f8f2 g2g3 h8f8 h1h2 f2f5 h3f5 f8f5 e5f4 e6e5 f4e3 b6e3 d2e3');
INSERT INTO evaluations VALUES(32039,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(32039,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(32039,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(32039,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(32039,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(32039,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(32039,36,'cp 160','d2d4 g4h5 b1d2 b8d7 d1b3 g8f6 b3b7 f8e7 d4e5 d6e5 b7b3 e8g8 b3c2 a7a5 e1g1 c7c6 f1e1 e7c5 h2h3 f6e8 c4e2 d8f6 d2c4 e8d6 c4d6 c5d6 e1d1 d7c5 c1e3 f8d8 a1c1 c5e6 d1d2 d6c5 d2d8 a8d8 e3c5 e6c5 c1d1 h5f3 e2f3');
INSERT INTO evaluations VALUES(32051,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(32051,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(32051,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(32051,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(32051,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(32051,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(32051,36,'cp 144','b8d7 d1b3 g4h5 b3b7 f8e7 d4e5 d6e5 b1d2 g8f6 b7b3 e8g8 b3c2 c7c6 e1g1 a7a5 h2h3 h5g6 f1e1 d8c7 f3h4 g6h7 h4f5 e7c5 f5g3 f8d8 d2f3 c5f8 f3h4 d7b6 c4e2 c6c5 c1e3 f6d5 h4f5 d5e3 f5e3 c5c4 b2b3 c4b3 a2b3');
INSERT INTO evaluations VALUES(33835,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33835,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(33835,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(33835,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(33835,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(33835,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(33835,36,'cp 11','c1f4 c8f5 b1d2 b8c6 c2c3 e8c8 d2e4 f5e4 f3g5 e4g6 e2e7 f8e7 g5f3 d8e8 e1c1 g6e4 h2h4 c6d8 f3e5 f7f6 e5g4 d8e6 f4g3 h7h5 g4e3 c7c6 d1e1 e7d8 c1d2 c8d7 b2b4 a7a5 b4b5 a5a4 b5c6 b7c6 f1d3 e4d3 d2d3 d8c7 c3c4 c7g3 f2g3 d5c4 e3c4 e6c7 c4b6 d7d6');
INSERT INTO evaluations VALUES(34224,30,'cp 259',NULL);
INSERT INTO evaluations VALUES(34224,31,'cp 279',NULL);
INSERT INTO evaluations VALUES(34224,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(34224,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(34224,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(34224,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(34224,36,'cp 248','b1c3 a5e5 g1f3 e5c7 d1d4 d7d6 c1f4 c7b6 e1c1 c8g4 d4d3 b8d7 f4d6 a8d8 d6f8 d7f8 d3d8 b6d8 d1d8 e8d8 f3d4 f7f6 f2f3 g4d7 f1d3 d8c7 h2h4 f8g6 c1d2 g6e5 a2a3 g8e7 d2e3 a7a5 b2b3 e7g6 a3a4 e5d3 c2d3');
INSERT INTO evaluations VALUES(34223,30,'cp 249',NULL);
INSERT INTO evaluations VALUES(34223,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(34223,32,'cp 251',NULL);
INSERT INTO evaluations VALUES(34223,33,'cp 252',NULL);
INSERT INTO evaluations VALUES(34223,34,'cp 243',NULL);
INSERT INTO evaluations VALUES(34223,35,'cp 246',NULL);
INSERT INTO evaluations VALUES(34223,36,'cp 260','a5e5 g1f3 e5c7 d1d4 d7d6 c1f4 c8g4 f3d2 g4e6 d2c4 e6c4 f1c4 b8d7 c4b3 g8f6 e1c1 f6h5 f4d6 f8d6 d4d6 c7d6 d1d6 e8e7 h1d1 h5f6 f2f4 h8d8 e4e5 f6e8 d6d4 d7c5 d4d8 c5b3 a2b3 a8d8 d1d8 e7d8 c1d2 f7f6 c3e2 f6e5 f4e5 d8d7 d2e3 e8c7 e2d4 g7g6 b3b4 b7b6 g2g3');
INSERT INTO evaluations VALUES(34220,30,'cp 267',NULL);
INSERT INTO evaluations VALUES(34220,31,'cp 288',NULL);
INSERT INTO evaluations VALUES(34220,32,'cp 272',NULL);
INSERT INTO evaluations VALUES(34220,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(34220,34,'cp 274',NULL);
INSERT INTO evaluations VALUES(34220,35,'cp 259',NULL);
INSERT INTO evaluations VALUES(34220,36,'cp 270','g1f3 e5c7 d1d4 d7d6 c1f4 c7b6 e1c1 c8g4 d4d2 b8d7 f4d6 f8d6 d2d6 e8c8 d6g3 g8h6 h2h3 g4f3 g2f3 b6c7 f3f4 d7c5 f1c4 h8e8 g3e3 c5d7 e3f3 g7g6 f3g3 d7f6 d1d8 c7d8 h1d1 d8e7 f2f3 e8d8 d1d8 e7d8 c3e2 f6h5 g3g5');
INSERT INTO evaluations VALUES(34219,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(34219,31,'cp 257',NULL);
INSERT INTO evaluations VALUES(34219,32,'cp 271',NULL);
INSERT INTO evaluations VALUES(34219,33,'cp 261',NULL);
INSERT INTO evaluations VALUES(34219,34,'cp 278',NULL);
INSERT INTO evaluations VALUES(34219,35,'cp 253',NULL);
INSERT INTO evaluations VALUES(34219,36,'cp 257','e5c7 d1d4 d7d6 c1f4 c7b6 e1c1 c8g4 d4d3 b8d7 f4d6 a8d8 d6f8 d7f8 d3d8 b6d8 d1d8 e8d8 f3d4 f7f6 f2f3 g4d7 f1d3 g8e7 c3e2 c6c5 d4f5 d7f5 e4f5 g7g6 h1d1 d8c7 e2c3 g6f5 c3b5 c7b8 b5d6 f8g6 d6f5 e7f5 d3f5 g6h4 f5e4 h4g2 d1d7 b8c8 d7b7');
INSERT INTO evaluations VALUES(34547,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(34547,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(34547,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(34547,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(34547,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(34547,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(34547,36,'cp 85','f1e2 e6g6 d2d4 g6g2 e2f3 g2g6 b1c3 g8f6 g1e2 g6f5 e2g3 f5e6 f3e2 g7g6 h2h4 f8g7 h4h5 e8g8 h1h4 b8d7 c1f4 c7c6 d1d2 b7b5 f4h6 g7h6 d2h6 g6g5 d4d5 c6d5 h6g5 g8h8 c4d5 e6b6 h4b4 a7a6 a2a4 b6c5 g5f4 d7b6 a1c1 b6d5 c3d5 c5d5 a4b5');
INSERT INTO evaluations VALUES(34548,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(34548,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(34548,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(34548,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(34548,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(34548,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(34548,36,'cp 55','e6g6 d2d4 g8f6 b1c3 g6g2 e2f3 g2g6 g1e2 g6f5 e2g3 f5e6 f3e2 g7g6 h2h4 f8g7 h4h5 e8g8 c1f4 e6b6 d1d2 b8c6 f4e3 b6b4 e1c1 e7e5 d4e5 c6e5 a2a3 b4d6 d2d6 c7d6 h5h6 g7h8 d1d6 c8g4 e3d4 g4e2 g3e2 f8e8 c1b1 e5c4 d6f6 h8f6');
INSERT INTO evaluations VALUES(40634,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(40634,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40634,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40634,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(40634,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40634,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(40634,36,'cp 21','f1d3 e8g8 e1g1 b8c6 a2a3 b4c3 b2c3 b7b6 a3a4 f8e8 c1a3 c8b7 d1e2 a8c8 h2h3 f6e4 d3e4 d5e4 f3d2 c6a5 c3c4 b7a6 f1c1 d8d7 d2e4 a5c4 e4d6 c4a3 c1c8 e8c8 e2a6 c8c7 d6b5 a3b5 a6b5 g8f8 b5d7 c7d7 a1c1 d7d5 c1c7');
INSERT INTO evaluations VALUES(44791,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44791,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44791,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(44791,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(44791,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44791,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44791,36,'cp 10','f8g7 c1e3 e8g8 f1e2 b8c6 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 a1d1 g8g7 b2b3 b7b6 f2f4 c8e6 c2c4 h7h5 f1e1 g7f6 g1f2 f8h8 g2g4 h5g4 e2g4 a8d8 d1d8 h8d8 g4e6 f7e6 f2e2 e6e5 f4e5 f6e5 e2e3 e5f6');
INSERT INTO evaluations VALUES(40398,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(40398,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40398,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(40398,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(40398,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(40398,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40398,36,'cp 7','c1e3 e8g8 f1e2 b8c6 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 a1d1 g8g7 b2b3 h7h5 c2c4 b7b6 d1d4 c8f5 f2f4 a8d8 f1d1 d8d4 d1d4 g7f6 g1f2 e7e5 f4e5 f6e5 f2e3 f8c8 d4d5 e5f6 e3d4 f6e7 e2f3 f5e6 d5e5 c8d8 d4c3 h5h4');
INSERT INTO evaluations VALUES(40392,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40392,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40392,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(40392,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40392,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(40392,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(40392,36,'cp 7','e8g8 f1e2 b8c6 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 e2f3 c8e6 b2b3 a8c8 c2c4 c8c7 g1f1 h7h5 d1d4 e6f5 a1d1 b7b6 f1e1 c7c5 e1d2 f8c8 a2a4 c5c7 g2g4 h5g4 h3g4 f5e6 d2c2');
INSERT INTO evaluations VALUES(36007,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(36007,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36007,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(36007,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36007,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(36007,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36007,36,'cp 11','f1e2 b8c6 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 e2f3 a8b8 c2c4 b7b6 a1c1 g7f6 b2b4 c8e6 c4c5 b6c5 b4c5 f8c8 a2a4 b8b4 a4a5 c8c7 c5c6 b4b5 c1b1 b5b1 d1b1 f6e5 g1h2 e5d6 b1b7 e6c4');
INSERT INTO evaluations VALUES(45857,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45857,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45857,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45857,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45857,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45857,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45857,36,'cp 0','b7b5 c4b3 f8e7 d2d4 b8d7 e1g1 a7a5 a2a3 e8g8 c1e3 c8a6 f1e1 f8e8 f3g5 e8f8');
INSERT INTO evaluations VALUES(46179,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46179,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(46179,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(46179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46179,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46179,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46179,36,'cp 0','c5d4 e3d4 f8e7 e1g1 e8g8 b1c3 b8c6 f3e5 d5c4 e5c6 b7c6 e2c4 d8c7 d1f3 e7d6 c1g5 d6h2 g1h1 h2d6 g5f6 g7f6 c3e4 f6f5 f3h5 f5e4 h5g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(46269,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(46269,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(46269,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(46269,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(46269,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(46269,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(46269,36,'cp -11','f1e2 b8c6 e4d5 e6d5 d3d4 c5d4 f3d4 f8b4 e1g1 e8g8 c1g5 h7h6 g5h4 f8e8 e2b5 b4c3 b2c3 d8d6 f2f3 c8d7 a1b1 c6d4 d1d4 d7b5 b1b5 b7b6 h4f6 d6f6 d4f6 g7f6 b5d5 a8c8 d5d6 c8c3 d6f6 g8g7 f6d6 c3c2');
INSERT INTO evaluations VALUES(47314,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(47314,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(47314,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(47314,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(47314,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(47314,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(47314,36,'cp -73','a7a6 e2e3 b7b5 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 b1c3 c7c6 f1e2 b8d7 e1g1 e7e6 f3e5 f8b4 e5d7 d8d7 g5f6 g7f6 e2h5 c6c5 d1g4 b4c3 b2c3 c5d4 e3d4 e8f8 f1a1 a8c6 g4f4 f8e7 f4g4 h8d8 g4g7 d8f8 g7g4 d7b7 h2h3 b5b4 c3b4 c4c3 g4g3 c3c2 a1c1');
INSERT INTO evaluations VALUES(48364,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(48364,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(48364,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(48364,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(48364,34,'cp -74',NULL);
INSERT INTO evaluations VALUES(48364,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(48364,36,'cp -59','c7c5 e2e3 b8d7 d4c5 f8c5 g1f3 b7b6 c3a4 c5e7 c2c4 a7a6 c4d5 f6d5 c1d2 c8b7 a4c3 d5f6 f1e2 a8c8 e1g1 e8g8 a1c1 h7h6 b2b4 e7d6 d1b3 d8e7 f1d1 c8c7 c3b1 f8c8 c1c7 c8c7 d1c1 b6b5 c1c7 d6c7 b3c2 d7b6 d2e1 g8h8 f3d4');
INSERT INTO evaluations VALUES(46257,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(46257,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(46257,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(46257,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(46257,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(46257,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(46257,36,'cp -66','e2e3 b8d7 d4c5 f8c5 g1f3 b7b6 c3a4 c5e7 c2c4 a7a6 c4d5 f6d5 c1d2 c8b7 a4c3 d5f6 f1e2 e8g8 e1g1 a8c8 a1c1 e7d6 c3a2 d7c5 d2b4 d8e7 a2c3 f8d8 d1c2 g7g6 f1d1 h7h6 c2b1 a6a5 b4c5 d6c5 d1d8 c8d8 c1d1 d8d1 b1d1');
INSERT INTO evaluations VALUES(48770,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(48770,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(48770,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(48770,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48770,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(48770,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(48770,36,'cp 16','f6e4 b1d2 e4d2 c1d2 f8e7 c2c3 c7c5 f1d3 b8c6 e1g1 e8g8 f1e1 c8e6 d1e2 d8d7 a1d1 e6g4 h2h3 g4h5 g2g4 h5g6 f3g5 e7g5 d2g5 f8e8 d3g6 h7g6 f2f4 d5d4 e2c4 b7b6 g1h2 a8c8 c4e2 d7d5 e2g2 d5a2');
INSERT INTO evaluations VALUES(49066,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(49066,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(49066,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(49066,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(49066,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(49066,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(49066,36,'cp 26','b1d2 e4d2 c1d2 f8e7 c2c3 c7c5 f1d3 b8c6 e1g1 e8g8 f1e1 c8e6 d1e2 d8d7 a1d1 e6g4 h2h3 g4h5 a2a3 d7e6 b2b4 h7h6 d2f4 a8d8 f4g3 a7a6 d3b1 h5g6 b1a2 g6e4 f3h4 g7g5 h4f3 c5c4 f3d4 c6d4 d1d4 e4f5 f2f3');
INSERT INTO evaluations VALUES(48938,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(48938,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(48938,32,'cp -76',NULL);
INSERT INTO evaluations VALUES(48938,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(48938,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(48938,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(48938,36,'cp -86','g1f3 f8c5 d1e2 b8c6 c2c3 c8e6 c1e3 c5b6 b1a3 d8e7 f3g5 e6d7 g5f3 b6e3 e2e3 f6g4 e3d2 f7f5 e4f5 d7f5 d2g5 e7c5 e1c1 e8g8 d3d4 e5d4 f3d4 c6d4 d1d4 h7h6 g5d2 a8e8 f1c4 g8h8 a3c2 c5e7 c4b3 f5c2 d4g4 c2f5');
INSERT INTO evaluations VALUES(48937,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(48937,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(48937,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(48937,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(48937,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(48937,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(48937,36,'cp -70','f8c5 d1e2 b8c6 c2c3 c8e6 c1e3 c5b6 b1a3 d8e7 a3c2 e8c8 f3g5 e6d7 e1c1 c8b8 g2g3 f6g4 e3b6 a7b6 g5f3 f7f5 e4f5 d7f5 f1g2 h8e8 h2h3 g4f6 f3g5 h7h6 g5e4 f6e4 g2e4 f5e4 e2e4 e7e6 c1b1 e8f8 e4g4 e6g4 h3g4');
INSERT INTO evaluations VALUES(51178,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(51178,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(51178,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51178,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(51178,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(51178,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51178,36,'cp 25','c4d3 c8b7 d1e2 g8f6 e1g1 b8c6 d4c6 b7c6 b1d2 f8e7 d2b3 d8c7 c1d2 e8g8 b3a5 d7d5 a5c6 c7c6 e4d5 f6d5 c2c4 b5c4 d3c4 e7f6 a1b1 c6a4 c4b3 a4b5 e2e4 f8d8 f1c1 a8c8 c1c8 d8c8 b1c1 c8c1 d2c1 b5b4 e4d3 b4e1 d3f1 e1b4 g2g3 h7h6 f1e2');
INSERT INTO evaluations VALUES(53990,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(53990,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(53990,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53990,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(53990,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(53990,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53990,36,'cp 16','e6e5 d2d3 b8c6 a2a3 a7a5 f1e2 f8e7 e1g1 g8f6 f3e1 e8g8 f2f4 e5f4 c1f4 f6e8 e2g4 c8e6 g4e6 f7e6 d1g4 d8d7 e1f3 e8f6 g4h3 e6e5 h3d7 f6d7 f4d2 f8f7 f1f2 a8f8 a1f1 c6d4 c3d5 e7d8 a3a4 d4b3 d2e3 b3d4 h2h3 d4f3 f2f3 f7f3 f1f3');
INSERT INTO evaluations VALUES(54209,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(54209,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54209,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54209,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(54209,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54209,36,'cp 0','b8c6 f1g2 g8f6 e1g1 c5d4 c3d4 f8d6 c1g5 e8g8 b1c3 h7h6 g5f6 d8f6 a1c1 f6d8 a2a3 d8b6 c3a4 b6b5 b2b4 f8d8 a4c3 b5b6 c3a4');
INSERT INTO evaluations VALUES(20275,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20275,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20275,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20275,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20275,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20275,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(20275,36,'cp 0','f1g2 g8f6 e1g1 c5d4 c3d4 f8d6 c1g5 h7h6 g5f6 d8f6 a2a3 e8g8 b1c3 f8d8 a1c1 f6e7 f1e1 f7f5 c3b5 d6b8 b5c3');
INSERT INTO evaluations VALUES(56465,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(56465,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(56465,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(56465,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(56465,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(56465,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(56465,36,'cp -86','g8f6 b1d2 c7c6 g2g3 f8d6 f1g2 e8g8 e1g1 a7a5 e3e4 f8e8 h2h3 a5a4 e2c3 d5d4 c3e2 b7b5 c2c3 d4c3 e2c3 h7h6 d2f3 c8a6 c1e3 c6c5 f3h2 d6f8 h2g4 f6g4 d1g4 b5b4 c3d5 b8d7 f1d1 d7f6 d5f6 d8f6 a1c1 e8d8 e3c5 a6d3 c5f8');
INSERT INTO evaluations VALUES(50,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(50,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(50,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(50,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50,36,'cp 43','c6d4 e3d4 b7b5 a2a3 a7a5 e4e5 d6e5 d4e5 b5b4 e5f6 g7f6 c3e4 c8c7 e4f6 e7f6 a3a4 d8c8 h1e1 g8g7 c1b1 h7h5 d2f2 d7c6 d1d4 f8d8 d4d8 c8d8 f2b6 d8d7 b6a5 c7a7 a5b4 c6a4 b3c4 d7d4 b4c3 d4c3 b2c3 a7c7 c4b3 a4b3 c2b3 c7c3 b1b2');
INSERT INTO evaluations VALUES(767,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(767,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(767,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(767,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(767,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(767,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(767,36,'cp 0','c5d4 c3d4 e7f5 c1b2 c8d7 h2h4 h7h5 f1e2 f8e7 g2g3 c6a5 b1d2 a5c6');
INSERT INTO evaluations VALUES(779,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(779,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(779,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(779,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(779,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(779,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(779,36,'cp 0','c3d4 e7f5 c1b2 c8d7 h2h4 h7h5 f1e2 f8e7 g2g3 c6a5 b1d2 a5c6');
INSERT INTO evaluations VALUES(1236,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(1236,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(1236,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(1236,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(1236,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(1236,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(1236,36,'cp 99','g8e7 e1g1 b7b5 a4b3 c8b7 c3d5 e7d5 b3d5 a8b8 c2c3 h7h6 d2d4 f8e7 f1e1 d7d6 a2a4 e8g8 c1e3 e5d4 c3d4 f8e8 a4b5 a6b5 d1b3 c6d8 d5b7 d8b7 a1a7 e7f8 b3d5 f6e6 d5c6 b7d8 c6c7 b8c8 c7d7 e6d7 a7d7 e8e4 d4d5 e4e7 d7e7 f8e7 e3d4 e7f6 d4f6 g7f6');
INSERT INTO evaluations VALUES(2314,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(2314,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(2314,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(2314,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(2314,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(2314,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(2314,36,'cp 61','f3g5 c6e5 c1f4 h7h6 g5e4 e5g6 f4e3 f7f5 e4c5 g8f6 b1c3 c7c6 e1c1 d8c7 f1c4 f8d6 e3d4 g6h4 g2g3 h4f3 d4f6 g7f6 c5e6 c8e6 c4e6 f5f4 c3e4 a8d8 e4d6 d8d6 d1d6 c7d6 e6g4 f3e5 g4h5 f4g3 h2g3 f6f5 h5e2 d6e7 c1d2 e5g4 f2f3 g4e5 h1h5 e7f6 d2e3 f6g6 h5h4 h6h5 a2a4');
INSERT INTO evaluations VALUES(2312,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(2312,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(2312,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(2312,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(2312,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(2312,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(2312,36,'cp 63','c6e5 c1f4 h7h6 g5e4 e5g6 f4e3 f7f5 e4c5 g8f6 b1c3 c7c6 e1c1 d8c7 f1c4 f8d6 h1e1 b7b6 c5d3 f5f4 e3d4 c8f5 c3e2 f4f3 g2f3 a8e8 c4f7 e8e7 f7g6 f5g6 e2f4 e7e1 d1e1 g6d3 f4d3 c6c5 d4e5 d6e5 e1e5 c7d6 c2c4 h8f8 f3f4 f6d7 e5e3 d7f6 c1d2 f6h5 e3e4 g7g5 f4g5 h6g5 d2e3 h5f6');
INSERT INTO evaluations VALUES(2321,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(2321,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(2321,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(2321,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(2321,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(2321,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(2321,36,'cp 83','c1f4 h7h6 g5e4 e5g6 f4d2 c7c6 b1c3 f7f5 e4g3 g8f6 f1d3 g6h4 e1g1 f8d6 c3e2 d6e5 c2c3 h8e8 f1e1 e8e7 h2h3 c8d7 e2d4 f5f4 d4f5 d7f5 g3f5 h4f5 d3f5 g7g5 g1f1 d8c7 c3c4 a8d8 d2c3 d8e8 c3b4 e5d6 c4c5 e7e1 a1e1');
INSERT INTO evaluations VALUES(7883,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(7883,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7883,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7883,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7883,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7883,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7883,36,'cp 0','e8e7 d6c8 a8c8 f1d3 h8e8 c1d2 b4c3 d2c3 d7d5 e4d5 f6d5 d1h5 d5c3 b2c3 g7g6 h5h6 e5e4 d3c4 c6d4 h6h4 e7f8 h4h6');
INSERT INTO evaluations VALUES(7885,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(7885,31,'cp 203',NULL);
INSERT INTO evaluations VALUES(7885,32,'cp 188',NULL);
INSERT INTO evaluations VALUES(7885,33,'cp 204',NULL);
INSERT INTO evaluations VALUES(7885,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(7885,35,'cp 213',NULL);
INSERT INTO evaluations VALUES(7885,36,'cp 182','d8e7 d6e7 e8e7 c1g5 h7h6 c3d5 e7d8 g5f6 g7f6 d5e3 d7d6 a1d1 d8c7 e3d5 c7d8 d5f6 c8e6 f6d5 a8b8 d5e3 d8e7 a2a4 a7a6 f2f3 b7b5 a4b5 a6b5 e1f2 b5b4 f1a6 c6d4 d1a1 b8a8 a6d3 d4c6 h2h4 h8b8 g2g4 b4b3 c2c3 a8a2 a1b1 c6a5 g4g5 h6g5 h4g5');
INSERT INTO evaluations VALUES(465,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(465,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(465,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(465,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(465,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(465,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(465,36,'cp 201','d6e7 e8e7 c1g5 h7h6 g5f6 g7f6 c3d5 e7d8 d5f6 d7d6 e1c1 c8e6 f6d5 h6h5 d5e3 d8e7 c1b1 a7a6 f1c4 a8d8 c4d5 c6b4 d5b3 a6a5 d1d2 b7b5 h1d1 h5h4 f2f3 b4c6 b3d5 c6b4 c2c3 b4d5 e3d5 e6d5 d2d5 d8g8 d1d2 h4h3 g2g3 b5b4');
INSERT INTO evaluations VALUES(458,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(458,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(458,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(458,33,'cp 197',NULL);
INSERT INTO evaluations VALUES(458,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(458,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(458,36,'cp 199','e8e7 c1g5 h7h6 g5f6 g7f6 c3d5 e7d8 d5f6 d7d6 e1c1 c8e6 f6d5 h6h5 d5e3 d8e7 c1b1 a7a6 f1c4 a8c8 c4d5 b7b5 c2c3 a6a5 d1d3 h5h4 h1d1 h8d8 g2g3 h4g3 f2g3 b5b4 h2h4 b4c3 d3c3 c6d4 d1c1 c8c3 c1c3 d4b5 c3c6 e7d7');
INSERT INTO evaluations VALUES(9513,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(9513,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9513,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(9513,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9513,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(9513,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9513,36,'cp -8','d2d4 f6e4 d1f3 e4f6 b1c3 e8g8 c1f4 d6b4 e1c1 b4c3 f3c3 c8e6 h1e1 a7a5 a2a4 f8e8 f4d2 f6d5 c3g3 f7f6 e5d3 b7b6 d3f4 d5f4 d2f4 d8d5 f4c7 b6b5 e1e3 b5a4 d1e1 e6f7 e3e8 a8e8 e1e8 f7e8 g3e3 e8f7 f2f3');
INSERT INTO evaluations VALUES(10085,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(10085,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(10085,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(10085,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(10085,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(10085,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(10085,36,'cp -69','d7d6 h2h3 e8g8 e1g1 c6d4 c3a4 f6e4 c2c3 d4e6 d2d4 e5d4 c3d4 d6d5 a4c5 e4c5 d4c5 d5c4 b3c4 b7b6 c1a3 b6c5 d1d5 a8b8 d5e5 f7f6 e5e3 d8d7 f3d2 f8e8 d2b3 c8b7 b3c5 e6c5 e3c5 b7c6 c5a7 b8a8');
INSERT INTO evaluations VALUES(10195,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(10195,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(10195,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(10195,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(10195,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(10195,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(10195,36,'cp -57','c5b6 f1c4 e5d4 e4e5 d7d5 c4e2 f6e4 c3d4 f7f6 e5f6 d8f6 b1c3 b6d4 c3e4 d5e4 f3d4 c6d4 c1g5 f6b6 e1g1 e8g8 a1c1 d4e2 d1e2 b6g6 g5h4 c7c6 f1e1 c8e6 e2e4 g6e4 e1e4 e6a2 f2f3 a7a5 e4e7 f8f7 c1e1 b7b5 e7e8 f7f8 e8f8 g8f8 e1e7 a8c8 e7a7 a5a4 h4g3 a2f7 a7b7 h7h5 g3e5 c8d8 e5c3 h5h4');
INSERT INTO evaluations VALUES(11164,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(11164,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(11164,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(11164,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(11164,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11164,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(11164,36,'cp 53','d1d2 e7e6 e1c1 c8d7 f2f4 f8e7 d4f3 b7b5 g5f6 g7f6 c1b1 d8b6 f4f5 e8c8 a2a3 c8b8 f3d4 d8c8 f1e2 c8c7 h2h4 h8c8 d4c6 d7c6 d2d4 b6d4 d1d4 e6f5 e4f5 c6g2 h1g1 g2c6 g1g7 d6d5 c3d5 c6d5 d4d5 c7c2');
INSERT INTO evaluations VALUES(13933,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(13933,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(13933,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(13933,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(13933,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(13933,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(13933,36,'cp -75','f6e4 c4e6 c8e6 e1e4 d6d5 e4e1 d5d4 d2d3 f8e7 c3d4 c5d4 c1f4 e8g8 b1d2 d8b6 d2c4 e6c4 d3c4 b6b2 f3e5 f8e8 e1e2 b2b6 a1b1 b6a6 e5c6 a6c6 d1d4 e7c5 d4d3 e8e2 d3e2 a8e8 e2f3 c6f3 g2f3 b7b6 b1d1 f7f6 g1f1 g8f7 d1d7 e8e7 d7d8 f7e6 f4e3 e6f5 e3c5 b6c5 f1g2 f5f4 d8b8 f6f5 b8g8');
INSERT INTO evaluations VALUES(14601,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(14601,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(14601,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(14601,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(14601,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(14601,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(14601,36,'cp -29','c6a5 d5f6 d8f6 f3d2 f8e7 d1h5 a5c4 d2c4 c7c6 c1d2 c8e6 a2a4 a8d8 e1g1 e8g8 b2b3 f8e8 a1e1 f6h4 h5f3 b7b6 h2h3 e7g5 d2b4 c6c5 b4c3 f7f6 c4e3 g5e3 f2e3 e8f8 f3e2 h4g5 g1h1 g5g6 c3d2 g8h7 d2c3 g6f7 e2f2 f8e8 c3b2');
INSERT INTO evaluations VALUES(18093,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(18093,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(18093,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(18093,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(18093,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(18093,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(18093,36,'cp 172','a4b3 c6a5 f3e5 a5b3 a2b3 g8e7 e5d3 d7d5 e4e5 h7h5 b1c3 e7g6 d1e2 c7c6 e5f6 e8f7 f6g7 f8g7 c3d1 d8e7 e2e7 f7e7 b3b4 h8g8 c2c3 g6h4 f2f3 c8f5 d3e1 d5d4 d1f2 g7f6 f2e4 f5e4 f3e4 g8g4 g2g3 h4g6 g1f2 g4e4 d2d3 e4e6 c1d2 h5h4 e1f3 h4g3 h2g3 g6e5 f3e5 e6e5');
INSERT INTO evaluations VALUES(18655,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(18655,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(18655,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(18655,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(18655,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(18655,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(18655,36,'cp 54','d7d6 c3d5 g8f6 c2c3 f8e7 d2d4 e5d4 c3d4 f6d5 c4d5 e8g8 c1e3 d8b6 a1b1 e7f6 d4c5 d6c5 b1c1 f6e7 d1e2 c6b4 d5c4 b6g6 f1e1 c8h3 e2f1 e7d6 f3d2 a8d8 f2f4 h3e6 f4f5 g6h5 h2h3 e6c4 d2c4 b7b6');
INSERT INTO evaluations VALUES(20138,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(20138,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(20138,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(20138,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(20138,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(20138,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(20138,36,'cp 44','c3d5 g8f6 c2c3 f8e7 d2d4 c5d4 c3d4 e8g8 d5e7 d8e7 d4d5 c6b4 a2a3 b4a6 f1e1 a6c5 f3d2 c8d7 a3a4 f8c8 b2b3 a7a6 a4a5 d7b5 c1a3 e7d8 d1f3 b7b6 a3c5 c8c5 a5b6 d8b6 c4b5 c5b5 d2c4 b6c7 e1c1 c7d8 h2h3 a6a5 c1c3 a8b8 f3e3');
INSERT INTO evaluations VALUES(20859,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(20859,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(20859,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(20859,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(20859,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(20859,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(20859,36,'cp 47','e7e5 d4e5 f6e5 e2e4 g8f6 e4d5 d8d5 d1d5 f6d5 f3e5 c6e5 g2d5 c7c6 d5g2 f8c5 b1c3 c8f5 c1f4 e5g6 f4d2 e8g8 e1c1 c5f2 c3e4 f5g4 d1f1 f2d4 d2b4 f8f1 h1f1 a8d8 b4c5 d4c5 e4c5 b7b6 c5b7 d8f8 f1f8 g8f8 g2c6 g6e5 c6e4 f8e7 c1d2 g4e6 b2b3 e5g4 e4h7 g4h2');
INSERT INTO evaluations VALUES(22268,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(22268,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(22268,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(22268,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(22268,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(22268,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(22268,36,'cp 45','d4e5 f6e5 e2e4 g8f6 e4d5 d8d5 d1d5 f6d5 e1g1 c8e6 a2a3 f8e7 f1e1 e8c8 b1d2 e7f6 d2e4 h7h6 b2b3 d5c3 e4f6 g7f6 c1b2 c3b5 a1d1 d8d1 e1d1 e6f5 c2c4 b5d6 d1e1 h8d8 f3d2 f5g6 b2c3 c6d4 c3d4 e5d4');
INSERT INTO evaluations VALUES(22240,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(22240,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22240,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(22240,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(22240,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(22240,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(22240,36,'cp 15','d2d4 e5e4 c2c4 a7a6 b5c6 b7c6 g1e2 f8b4 b2c3 b4d6 d1d2 a8b8 c3a5 g8h6 b1c3 e8g8 c4c5 d6e7 f2f3 e4f3 g2f3 h6f5 e2f4 g7g5 f4g2 f5g7 h2h4 c8f5 e1c1 h7h5 g2e1 g5g4 d2h2 b8b7 f3g4 h5g4 e1g2 d8d7 g2f4 g8h7 d1e1 f8g8 c1b2');
INSERT INTO evaluations VALUES(23563,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(23563,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(23563,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(23563,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(23563,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(23563,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(23563,36,'cp 140','c4e6 f7e6 a2a4 g8h6 c1e3 h6f7 b1d2 g7g5 e1e2 f8e7 d2c4 d8e8 a4a5 a7a6 h2h3 f7d6 c4d2 h7h5 a1a4 a8d8 d2f3 d6f7 a4c4 d8d7 c4c3 h8g8 c3b3 g5g4 h3g4 h5g4 f3d2 f7d6 h1h5 e7f6');
INSERT INTO evaluations VALUES(23562,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(23562,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(23562,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(23562,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(23562,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(23562,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(23562,36,'cp 145','f7e6 b1d2 g8h6 h2h4 a7a5 d2c4 h6f7 a2a4 a8a6 c1g5 d8c8 e1c1 f8c5 f2f3 a6c6 b2b3 b7b6 d1d3 h7h6 g5d2 c5e7 h4h5 h8d8 d3d8 c8d8 c2c3 e7f6 c1c2 f7d6 c4b2 d6f7 b2d3 d8e7 d2c1 e7e8 h1h3 f6e7 c1e3');
INSERT INTO evaluations VALUES(23565,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(23565,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(23565,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(23565,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(23565,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(23565,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(23565,36,'cp 144','a2a4 g8h6 b1d2 h6f7 h2h4 a7a5 d2c4 a8a6 c1g5 d8c8 e1c1 f8c5 f2f3 a6c6 b2b3 b7b6 c2c3 h7h6 g5d2 h8g8 h1h3 g7g6 c1b2 c5d6 d2e3 d6f8 e3c1 f8g7 b2c2 g7f6 g2g3 g8d8 d1f1 d8g8 h3h2 h6h5 c1e3 g8d8 h2d2 d8d2 c2d2');
INSERT INTO evaluations VALUES(24597,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(24597,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(24597,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(24597,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(24597,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(24597,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(24597,36,'cp -4','e7e6 f1d1 d8c7 c1e3 f8e7 e4e5 d6e5 d4e5 e7d8 d1d3 e8g8 e5c7 d8c7 f3d4 c6d7 d4b5 c7e5 e3d4 e5d4 d3d4 f8c8 a2a4 d7c6 a1d1 g8f8 d1d2 f8e7 f2f3 h7h6 b2b3 a7a6 b5d6 c8d8 d6c4 d8d4 d2d4 a8c8 a4a5 f6d7 d4d2 c8d8');
INSERT INTO evaluations VALUES(25716,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(25716,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(25716,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(25716,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(25716,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(25716,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(25716,36,'cp 107','b5c6 b7c6 e1g1 g8f6 d1f3 f6g4 f3e2 f8e7 d4c6 d7c6 e2g4 e8g8 c3d5 f7f5 d5e7 d8e7 e4f5 e7e5 c1d2 h7h5 g4c4 c6d5 c4c3 e5c3 d2c3 f8f5 f2f3 a8e8 f1f2 d5e6 h2h3 f5d5 f2e2 g8f7 a1e1 e6d7 e2e8 d7e8 c3g7 d5d2 g7c3 d2c2');
INSERT INTO evaluations VALUES(25703,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(25703,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(25703,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(25703,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(25703,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(25703,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(25703,36,'cp 117','b7c6 e1g1 g8f6 d1f3 f6g4 f3e2 f8e7 d4c6 d7c6 e2g4 e8g8 c3d5 f7f5 d5e7 d8e7 e4f5 e7e5 c1d2 e5f5 g4f5 f8f5 f2f3 a8e8 b2b3 h7h5 a1e1 f5d5 e1d1 e8e2 f1f2 e2e8 g1f1 g8f7 f2e2 g7g5 e2e8 f7e8 h2h3 e8d7 f1f2 d7c8 a2a3');
INSERT INTO evaluations VALUES(26746,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(26746,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(26746,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(26746,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(26746,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(26746,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(26746,36,'cp 151','c2c3 d8d7 d4d5 c6e7 h2h3 g4f3 d1f3 e7g6 a2a4 a7a6 a4a5 f8e7 c4b3 g8f6 b1d2 e8g8 d2c4 f6h7 b3a4 d7c8 a4c2 e7g5 c4e3 h7f6 c3c4 c7c6 f3d1 g5f4 a1a3 c8d7 b2b4 c6d5 c4d5 a8c8 c2a4 d7d8 g2g3 f4g5 g1h2');
INSERT INTO evaluations VALUES(26744,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(26744,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(26744,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(26744,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(26744,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(26744,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(26744,36,'cp 128','d8d7 d4d5 c6e7 h2h3 g4f3 d1f3 g7g6 a2a4 f7f5 a4a5 g8f6 a5a6 b7a6 b1d2 f8g7 c4a6 e8g8 f3d3 a8b8 a6c4 b8a8 c4b5 d7d8 b5a6 a8b8 a6c4 b8a8 b2b4 g8h8 a1a5 h6h5 d3c2 f5f4 d2f3 g7h6 h3h4 f6h7 a5a6');
INSERT INTO evaluations VALUES(26812,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(26812,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(26812,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(26812,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(26812,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(26812,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(26812,36,'cp 32','d3f5 e6f5 g1e2 d8d7 b1d2 f8e7 h2h3 g7g5 f4e5 c6e5 d4e5 e8c8 d2f3 g5g4 h3g4 f5g4 f3h4 e7h4 h1h4 h7h5 d1d4 g8e7 e1c1 e7c6 d4f4 d7e6 g2g3 e6e5 f4e5 c6e5 d1h1 e5c4 h4h5 h8h5 h1h5 c7c6 e2f4 c4d6 h5h7 d8e8 a2a3');
INSERT INTO evaluations VALUES(26795,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(26795,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(26795,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(26795,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(26795,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(26795,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(26795,36,'cp 29','e6f5 d1d3 d8d7 a2a3 g8f6 g1e2 f6e4 h2h4 f8d6 c2c4 d5c4 d3c4 e8g8 b1c3 c6e7 a1c1 a8c8 c4b3 b7b6 c1c2 e4c3 c2c3 d6f4 e2f4 f8d8 e1g1 e7g6 f4g6 h7g6 f1c1 f5f4 b3c2 f4e3 f2e3 d7e6');
INSERT INTO evaluations VALUES(26861,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(26861,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(26861,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(26861,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(26861,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(26861,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(26861,36,'cp 42','g1e2 d8d7 a2a3 g8f6 d1d3 f6e4 h2h4 f8d6 f4d6 e4d6 b1d2 e8g8 d2f3 a8d8 e2f4 c6e7 f3e5 d7c8 d3c3 d6e4 c3a5 a7a6 f2f3 f7f6 e5d3 b7b6 a5a4 e4d6 a4b3 f8e8 h4h5 d6c4 e1c1 h7h6 d1e1');
INSERT INTO evaluations VALUES(30618,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(30618,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(30618,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(30618,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30618,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(30618,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(30618,36,'cp 15','d2d4 c7c5 c2c3 b8c6 f1d3 c5d4 c3d4 c6b4 d3e2 b4d5 b1c3 d5c3 b2c3 a8c8 c1d2 g8e7 e1g1 e7g6 a2a4 f8e7 a4a5 e8g8 a5b6 a7b6 e2a6 b7a6 a1a6 c8a8 a6a8 d8a8 d1b3 a8c6 f1b1 f8b8 b3b5 c6b5 b1b5 f7f6 e5f6 e7f6 g1f1 g8f7 g2g3 g6e7 h2h4 d7d6 h4h5');
INSERT INTO evaluations VALUES(32235,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(32235,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(32235,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(32235,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(32235,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(32235,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(32235,36,'cp 125','f3d4 g8f6 d4e6 f7e6 c1e3 b8c6 g2g3 d8d7 d1e2 h7h5 e1c1 e8c8 c1b1 c8b8 f1h3 d7e8 f2f4 a7a6 h3g2 f6g4 e3g1 f8e7 e2c4 e7f6 h2h3 f6c3 h3g4 c3a5 g4g5 e6e5 f4f5 e8e7 g1e3 a5b6 c4b3 b8a7 a2a3 b6e3 b3e3 c6d4');
INSERT INTO evaluations VALUES(38436,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38436,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(38436,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(38436,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(38436,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(38436,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(38436,36,'cp 13','d3d4 g8f6 f1d3 e7e5 d4e5 d6e5 e1g1 d8e7 b1d2 f6h5 d2c4 b8c6 c2c3 h5f4 c4e3 c8e6 d3c4 e6c4 e3c4 e8g8 d1b3 b7b6 c1f4 e5f4 e4e5 a8d8 a1e1 e7e6 e1e4 d8d5 e4f4 c6e5 f3e5 g7e5 f1e1 f8d8 c4e5 d5e5 b3e6');
INSERT INTO evaluations VALUES(38618,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(38618,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(38618,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(38618,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(38618,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(38618,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(38618,36,'cp -18','g8f6 f1d3 d8e7 d3e2 c8g4 e1g1 e8g8 h2h3 f8e8 c1e3 g4h5 b1d2 c7c6 f1e1 e7c7 e2d3 b8d7 d1c2 h5g6 c3c4 g6d3 c2d3 d5c4 d2c4 d7b6 c4d6 c7d6 e3g5 h7h6 g5f6 d6f6 e1e5 f6d6 a1e1 e8e5 e1e5 d6b4 d3c3 b6d5 c3b4 d5b4');
INSERT INTO evaluations VALUES(39003,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(39003,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(39003,32,'cp -133',NULL);
INSERT INTO evaluations VALUES(39003,33,'cp -146',NULL);
INSERT INTO evaluations VALUES(39003,34,'cp -130',NULL);
INSERT INTO evaluations VALUES(39003,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(39003,36,'cp -151','b8c6 d2d3 c6d4 f3d1 d7d5 c4b3 g8e7 f2f3 e8g8 h3f2 c7c5 e1g1 b7b5 c1e3 d8b6 c2c3 d4c6 d1c1 a7a5 a2a4 c8a6 a4b5 a6b5 b1a3 f8b8 e4d5 e6d5 a3b5 b6b5 a1a3 g8h8 f1e1 e7f5 b3d5 f5e3 c1e3');
INSERT INTO evaluations VALUES(40471,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(40471,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(40471,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(40471,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(40471,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(40471,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(40471,36,'cp 42','d1b3 b4c3 b3c3 d5c4 f1c4 f6d5 c4d5 c6d5 g1e2 e8g8 e1g1 b8c6 f1c1 c8d7 f4d6 f8e8 e2f4 a8c8 a2a4 b7b6 b2b3 d8g5 c3b2 h7h6 b3b4 c6e7 b4b5 c8c1 a1c1 e8c8 d6e7 c8c1 b2c1 g5e7 f4d3 e7d8 g1h1 d7e8 h2h3 f7f6 d3f4 e8f7 c1c2');
INSERT INTO evaluations VALUES(43497,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(43497,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(43497,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(43497,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(43497,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(43497,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(43497,36,'cp -42','b7b5 c4a2 b8d7 e1g1 c8b7 f1e1 f8e7 c3e2 h7h6 c2c3 c5c4 e2g3 c4d3 d1d3 d8c7 h2h3 a8c8 a2b1 e8g8 d3e2 f8e8 c1d2 e6e5 a3a4 b5a4 a1a4 e7f8 a4a1 d6d5 b2b4 d7b6 b1d3 b6c4 e4d5 c4d2 e2d2 b7d5');
INSERT INTO evaluations VALUES(43757,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(43757,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(43757,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(43757,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(43757,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(43757,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(43757,36,'cp -20','c4a2 b8d7 e1g1 c8b7 c3e2 d8c7 f1e1 d6d5 e4d5 f6d5 c2c4 d5f6 e2c3 b5c4 a2c4 h7h6 c1d2 f8e7 d1e2 e8g8 a1c1 e7d6 h2h3 f8b8 c1b1 d6f8 c3e4 f6d5 b1c1 a6a5 e4g3 c7b6 b2b3 a5a4 b3a4 a8a4 f3e5 d7e5');
INSERT INTO evaluations VALUES(43758,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(43758,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(43758,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(43758,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(43758,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(43758,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(43758,36,'cp -37','b8d7 e1g1 c8b7 c3e2 d8c7 f1e1 g7g6 e2g3 f8g7 c1f4 e8g8 d1d2 d7e5 f3e5 d6e5 f4h6 f6g4 h6g7 g8g7 h2h3 g4f6 d2e3 f8d8 g3f1 a8c8 a1d1 h7h6 f1d2 f6h5 g2g3 d8d7 a2b3 h5f6 a3a4 c8d8 e3e2 c7d6 d2f1 d6d4 a4b5');
INSERT INTO evaluations VALUES(45166,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(45166,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(45166,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(45166,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(45166,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(45166,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(45166,36,'cp 54','f1g2 b8c6 g1f3 e7e5 d2d3 f8e7 e1g1 h7h6 a2a3 e8g8 a1b1 a7a6 f3d2 c8e6 c3d5 a8b8 e2e3 d8d7 b2b3 e7d8 c1b2 e6g4 f2f3 g4h5 d1e2 f8e8 e2f2 h5g6 e3e4 f6d5 c4d5 c6d4 b2d4 c5d4 f3f4 f7f5 e4f5 d7f5 d2e4');
INSERT INTO evaluations VALUES(46590,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(46590,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(46590,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(46590,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(46590,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(46590,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(46590,36,'cp 209','c4d5 e7e6 d5e6 c8e6 g1f3 b8c6 g2g3 f8b4 f1g2 h7h6 e1g1 e8g8 c1f4 d8e7 d1c2 g7g5 f4e5 f6g4 a2a3 b4c3 c2c3 a8d8 b2b4 g4e5 d4e5 a7a6 a1d1 e6d5 d1d2 d5f3 d2d8 f8d8 c3f3 e7e5 f3b3 g8h8 g2c6 b7c6');
INSERT INTO evaluations VALUES(48943,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48943,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48943,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48943,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48943,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48943,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48943,36,'cp 0','d3e4 d8d1 c3d1 b8c6 f1d3 c8e6 a2a3 a7a5 c1e3 f6g4 e3d2 f8c5 h2h3 g4f6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(48931,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48931,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48931,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48931,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(48931,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48931,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48931,36,'cp 0','d8d1 c3d1 b8c6 f1d3 h7h6 c1e3 c8e6 d3b5 e6d7 b5d3');
INSERT INTO evaluations VALUES(33916,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33916,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33916,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33916,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33916,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33916,36,'cp 0','c3d1 b8c6 f1d3 f6d7 d3b5 f7f6 b5c6 b7c6 c1e3 a7a5 a2a4 c8a6 b2b3 f8b4 f3d2 d7c5 f2f3 h7h5 d1b2 c5e6 e1c1 a6e2 d1e1 e2a6');
INSERT INTO evaluations VALUES(49598,30,'cp -452',NULL);
INSERT INTO evaluations VALUES(49598,31,'cp -466',NULL);
INSERT INTO evaluations VALUES(49598,32,'cp -465',NULL);
INSERT INTO evaluations VALUES(49598,33,'cp -479',NULL);
INSERT INTO evaluations VALUES(49598,34,'cp -486',NULL);
INSERT INTO evaluations VALUES(49598,35,'cp -490',NULL);
INSERT INTO evaluations VALUES(49598,36,'cp -487','f6e4 g1h3 d8h4 e1e2 b8c6 c2c3 d7d5 d2d3 f8e7 d1e1 e4c5 e1h4 e7h4 b1d2 h7h5 h1g1 a7a5 d2b3 h5g4 f3g4 h4e7 h3f2 h8h2 b3c5 e7c5 f1g2 c8e6 c1e3 c5d6 g2f3 e8c8 g1g2 h2h8 g2g1 a5a4 g1h1 c6e7 h1h8 d8h8 a1h1 h8h1 f3h1 b7b6 h1f3 f7f6 e3d2 e7g6 e2d1');
INSERT INTO evaluations VALUES(50079,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(50079,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(50079,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(50079,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(50079,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(50079,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(50079,36,'cp 75','g2g3 g7g6 b1c3 d7d6 d2d4 f8g7 f1g2 c7c6 e1g1 e8g8 b2b4 e7e5 d4e5 d6e5 e2e4 d8d1 f1d1 c8e6 g2f1 f5e4 f3g5 e6g4 d1e1 b8d7 g5e4 g4f5 c1a3 f5e4 c3e4 f6e4 e1e4 f8f3 a3b2 d7f6 e4e2 e5e4 a1e1 a8d8 c4c5 f3d3');
INSERT INTO evaluations VALUES(50987,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(50987,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(50987,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(50987,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(50987,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(50987,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(50987,36,'cp 101','g2g3 b8c6 f1g2 f8e7 e1g1 g8f6 f1e1 a8b8 c1e3 e8g8 d1e2 h7h6 a1d1 e7b4 f3e5 c6e7 e3d2 b7b5 h2h4 b4c3 d2c3 c8b7 g2b7 b8b7 c3b4 f8e8 b4c5 e7f5 b2b3 a6a5 d1d3 b5b4 e5c6 d8a8 e1d1');
INSERT INTO evaluations VALUES(51007,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(51007,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(51007,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(51007,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(51007,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(51007,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(51007,36,'cp 58','d7d5 b1c3 d5e4 c3e4 g8f6 f1d3 f8b4 c1d2 b4d2 d1d2 b8d7 h2h4 b7b6 e4f6 d7f6 e1c1 c8b7 d4d5 d8e7 h1e1 e8c8 d2c3 h8e8 f3e5 d8d6 g2g3 e8d8 f2f4 f6d7 d5e6 f7e6 d3c2 d6d1 c2d1 d7e5 c3e5 c8b8 e5e6 e7c5 d1c2 d8d4 e6e8 b8a7');
INSERT INTO evaluations VALUES(51381,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(51381,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(51381,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(51381,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(51381,34,'cp -112',NULL);
INSERT INTO evaluations VALUES(51381,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(51381,36,'cp -119','a2a4 g8f6 a4b5 f6e4 g1f3 f8b4 c1d2 e4d2 f3d2 e8g8 f1c4 b8d7 e1g1 g8h8 b1c3 f7f5 d5d6 d7b6 d6c7 d8c7 d1b3 b4c5 c3d5 c7d6 a1d1 c8b7 d5b6 c5b6 c4e2 b6d4 d2c4 d6c7 d1c1 a8c8 f1d1 c7c5 c4e3 c5c1 d1c1 c8c1 e2f1');
INSERT INTO evaluations VALUES(51382,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(51382,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(51382,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(51382,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(51382,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(51382,35,'cp -114',NULL);
INSERT INTO evaluations VALUES(51382,36,'cp -126','g8f6 g1f3 f6e4 a4b5 f8b4 c1d2 e4d2 f3d2 e8g8 f1c4 b8d7 b1c3 f7f5 d5d6 g8h8 e1g1 d7b6 d6c7 d8c7 d1b3 b4c5 c3d5 c7d6 d5b6 c5b6 c4d5 c8e6 d5e6 d6d2 a1d1 d2h6 e6d7 e5e4 b3e6 h6e6 d7e6 g7g6 e6c4 a8e8 d1d6 f5f4 b2b3 e4e3');
INSERT INTO evaluations VALUES(51385,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(51385,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(51385,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(51385,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(51385,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(51385,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(51385,36,'cp 152','c1g5 f8e7 h2h4 h7h6 g5d2 e7f8 f1d3 f8g7 d1e2 d7d6 b1c3 g8e7 e4e5 c8b7 c3e4 d8d7 c2c3 b8c6 e4f6 g7f6 e5f6 e7d5 c3c4 d5f6 d4d5 c6b8 d2c3 d7e7 d5e6 b8c6 h4h5 g6g5 e6f7 e8f7 d3g6 f7g7 f3d4 e7e2 e1e2 c6d4 c3d4 a8e8 g6e8 h8e8 e2f1');
INSERT INTO evaluations VALUES(53065,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(53065,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(53065,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(53065,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(53065,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(53065,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(53065,36,'cp 1','g8f6 h2h3 g7g6 c2c3 f8g7 d3c2 e7e5 d2d4 c5d4 c3d4 e5d4 f3d4 b8c6 e1g1 e8g8 d4c6 b7c6 b1c3 d6d5 f1e1 f6e4 c3e4 d5e4 c2e4 c8e6 d1a4 d8d4 a4d4 g7d4 e1d1 d4f6 c1e3 a8b8 e4c6 b8b2 e3a7 b2a2 a7d4 a2a1');
INSERT INTO evaluations VALUES(53100,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(53100,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(53100,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(53100,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(53100,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(53100,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(53100,36,'cp 11','e2e3 g8f6 f1b5 b8d7 b5e2 g7g6 g1f3 b7b6 d2d4 c5d4 f3d4 c8b7 e1g1 f8g7 c2c4 e8g8 b1c3 d6d5 c4d5 f6d5 c3d5 b7d5 a1c1 d7c5 b2a3 a8c8 d4b5 a7a6 b5d4 c8c7 e2f3 d8a8 b3b4 d5f3 d1f3 a8f3 g2f3');
INSERT INTO evaluations VALUES(53098,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(53098,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(53098,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(53098,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(53098,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53098,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(53098,36,'cp 5','g8f6 f1b5 b8d7 b5e2 b7b6 g1f3 c8b7 c2c4 g7g6 d2d4 e7e6 b1c3 f8g7 e1g1 e8g8 a1c1 a7a6 d1c2 d8e7 f1d1 f8d8 h2h3 a8c8 b2a1 c5d4 f3d4 d6d5 c4d5 e6d5 c2d2 b6b5 e2f1 b5b4 c3e2 f6e4 d2e1 a6a5 e2f4 g8h8');
INSERT INTO evaluations VALUES(54664,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(54664,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(54664,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(54664,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(54664,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(54664,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(54664,36,'cp 81','f1d3 d8c7 f2f4 b8c6 g1f3 g8h6 e1g1 c8d7 g1h1 a7a6 a2a4 c5c4 d3c2 f8e7 d1e1 f6f5 b1d2 e8g8 f3g5 b7b5 d2f3 b5b4 e1h4 c7b6 h4h5 b4b3 c2d1 a6a5 c1d2 c6d8 h2h3 d7e8');
INSERT INTO evaluations VALUES(54811,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(54811,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(54811,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(54811,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(54811,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(54811,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(54811,36,'cp 48','d2d4 g8f6 e2e4 f8g7 f1e2 e8g8 c1e3 b8c6 d4d5 c6e5 f2f4 e5g4 e3d2 g4h6 h2h3 e7e6 d5e6 f7e6 g2g4 b7b5 c4b5 c8b7 d1c2 h6f7 g1f3 c7c6 b5c6 b7c6 f3d4 f6g4 e1c1 g7d4 h3g4 a8c8 e2c4 d6d5 e4d5 c6d5 c4d5 e6d5 c1b1 d8b6');
INSERT INTO evaluations VALUES(77,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(77,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(77,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(77,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(77,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(77,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(77,36,'cp 62','d6d2 g8e7 b1c3 e8g8 b2b3 d7d6 c1b2 c8e6 e1c1 a8c8 c3d5 f6g6 f1d3 f7f5 d5e7 c6e7 f2f3 f5e4 f3e4 e7c6 h1f1 f8f1 d1f1 h7h6 b2c3 b7b5 c1b2 g6e8 d3e2 e8d7 f1d1 c8d8 a2a3 d7c7 b2b1 c7b6 c3b2 b6c5 h2h4 a6a5 g2g4');
INSERT INTO evaluations VALUES(976,30,'cp -599',NULL);
INSERT INTO evaluations VALUES(976,31,'cp -611',NULL);
INSERT INTO evaluations VALUES(976,32,'cp -603',NULL);
INSERT INTO evaluations VALUES(976,33,'cp -599',NULL);
INSERT INTO evaluations VALUES(976,34,'cp -606',NULL);
INSERT INTO evaluations VALUES(976,35,'cp -601',NULL);
INSERT INTO evaluations VALUES(976,36,'cp -589','d5f3 g5f3 e5e4 f3g5 c6d4 b1a3 f8a3 b2a3 d4c2 e1d1 c2a1 g5e4 e8g8 c1b2 f8e8 f2f3 c8e6 b2a1 e6a2 h2h4 a8d8 h1e1 a2d5 h4h5 h7h6 a1c3 f7f5 e4c5 b7b6 e1e8 d8e8 c5d7 d5b3 d1c1 e8e2 d7b6 a7b6 f3f4 e2g2 c3e5 c7c5 c1b2 g8f7 b2c3 g2g3 c3b2');
INSERT INTO evaluations VALUES(1000,30,'cp -604',NULL);
INSERT INTO evaluations VALUES(1000,31,'cp -610',NULL);
INSERT INTO evaluations VALUES(1000,32,'cp -585',NULL);
INSERT INTO evaluations VALUES(1000,33,'cp -586',NULL);
INSERT INTO evaluations VALUES(1000,34,'cp -603',NULL);
INSERT INTO evaluations VALUES(1000,35,'cp -598',NULL);
INSERT INTO evaluations VALUES(1000,36,'cp -594','g5f3 e5e4 f3g5 c6d4 b1a3 f8a3 b2a3 d4c2 e1d1 c2a1 g5e4 c8f5 f2f3 e8c8 c1b2 d8d3 h1e1 f7f6 a3a4 h8d8 e1e2 a1b3 a2b3 d3b3 d1c2 b3b6 b2c3 f5g6 d2d4 b6c6 c2b3 g6f7 b3b2 d8d5 e4f6 g7f6 e2e7 f7g6 h2h4 c6c4 b2b3 c4d4 c3d4');
INSERT INTO evaluations VALUES(1001,30,'cp -625',NULL);
INSERT INTO evaluations VALUES(1001,31,'cp -621',NULL);
INSERT INTO evaluations VALUES(1001,32,'cp -623',NULL);
INSERT INTO evaluations VALUES(1001,33,'cp -617',NULL);
INSERT INTO evaluations VALUES(1001,34,'cp -625',NULL);
INSERT INTO evaluations VALUES(1001,35,'cp -629',NULL);
INSERT INTO evaluations VALUES(1001,36,'cp -648','e5e4 f3g5 c6d4 b1a3 f8a3 b2a3 d4c2 e1d1 c2a1 c1b2 c8g4 d1c1 g4f5 g2g4 f5g4 h1g1 h7h5 h2h3 g4f5 b2a1 h8h6 a1g7 h6g6 g7c3 e8c8 h3h4 g6b6 c3b4 f5g6 g5f7 g6f7 g1g5 b6f6 b4e7 f6f2 e7d8 c8d8 g5a5 f2f1 c1b2 f1f3 a5a7 f3f2 a7a8 d8d7 b2c1 f7a2');
INSERT INTO evaluations VALUES(1378,30,'cp 405',NULL);
INSERT INTO evaluations VALUES(1378,31,'cp 411',NULL);
INSERT INTO evaluations VALUES(1378,32,'cp 431',NULL);
INSERT INTO evaluations VALUES(1378,33,'cp 399',NULL);
INSERT INTO evaluations VALUES(1378,34,'cp 399',NULL);
INSERT INTO evaluations VALUES(1378,35,'cp 375',NULL);
INSERT INTO evaluations VALUES(1378,36,'cp 398','f1e2 c6e5 e1g1 d7d6 f1e1 e7c7 b1a3 f7f6 b2b4 c5b4 a3b5 c7a5 a2a3 a7a6 a3b4 a5a1 b5c7 e8f7 c7a8 g7g6 f3d4 a1c3 d4e6 c8e6 c1d2 c3c8 a8b6 c8d8 d5e6 f7g7 d2e3 g8e7 f2f4 e5c6 b6d7 e7f5 e3b6 d8c8 e2d3 d6d5 d3f5 f8b4 e1f1 c6e7 f5h3 c8c4 d1f3');
INSERT INTO evaluations VALUES(1377,30,'cp 409',NULL);
INSERT INTO evaluations VALUES(1377,31,'cp 418',NULL);
INSERT INTO evaluations VALUES(1377,32,'cp 403',NULL);
INSERT INTO evaluations VALUES(1377,33,'cp 429',NULL);
INSERT INTO evaluations VALUES(1377,34,'cp 435',NULL);
INSERT INTO evaluations VALUES(1377,35,'cp 423',NULL);
INSERT INTO evaluations VALUES(1377,36,'cp 425','c6e5 e1g1 d7d6 f1e1 e7c7 b1a3 f7f6 b2b4 c5b4 a3b5 c7d8 a2a3 b4b3 c2b3 a7a6 b5c3 f8e7 f3d4 f6f5 e2h5 g7g6 f2f4 d8c7 f4e5 d6e5 c1b2 g8f6 a1c1 c7d6 h5f3 e8g8 c3a4 e5d4 e1e7 d6e7 d1d4 f5f4 a4b6 a8b8 d5d6 e7e3 d4e3 f4e3 b2f6 c8e6 f6d4 f8f3 g2f3 b8d8 d4e3 e6b3 c1c8');
INSERT INTO evaluations VALUES(1383,30,'cp 441',NULL);
INSERT INTO evaluations VALUES(1383,31,'cp 429',NULL);
INSERT INTO evaluations VALUES(1383,32,'cp 430',NULL);
INSERT INTO evaluations VALUES(1383,33,'cp 442',NULL);
INSERT INTO evaluations VALUES(1383,34,'cp 438',NULL);
INSERT INTO evaluations VALUES(1383,35,'cp 424',NULL);
INSERT INTO evaluations VALUES(1383,36,'cp 439','e1g1 d7d6 f1e1 e7c7 b1a3 c8d7 a3b5 c7b6 a2a4 e5f3 e2f3 f8e7 b2b4 d7b5 b4c5 b6c5 c1e3 c5c8 a4b5 g8f6 e3a7 c8d8 d1e2 e8f8 c2c4 f6d7 f3g4 e7f6 a1a3 a8c8 e2d3 d7e5 e1e5 f6e5 g4c8 d8c8 h2h3 h7h5 a7e3 g7g6 e3d4 f8g7 d4e5 d6e5 d3c3 c8c5 c3e5 f7f6 e5e3 c5c4 a3c3');
INSERT INTO evaluations VALUES(4538,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(4538,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(4538,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(4538,33,'cp 199',NULL);
INSERT INTO evaluations VALUES(4538,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(4538,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(4538,36,'cp 194','e1c1 f6e8 h2h4 d7d6 e4e5 c8e6 f1d3 d8c8 e5d6 c7d6 h1e1 a7a6 a2a3 c8c5 d3e4 a8b8 e4f3 b7b5 c3e4 c5c8 e1e3 h7h6 e4c3 c8c7 g2g4 c7c5 d4c5 d6c5 c3d5 e6d5 d1d5 c5c4 e3e7 b8b6 g4g5 h6g5 h4g5 b6e6');
INSERT INTO evaluations VALUES(4537,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(4537,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(4537,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(4537,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(4537,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(4537,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(4537,36,'cp 192','d7d6 e4e5 f6e8 h2h4 d8d7 h4h5 d7g4 e5d6 e8d6 c3d5 g4d4 d1d4 d6f5 d4f4 c7c6 f4f5 f8e8 f5g5 c6d5 g5d5 c8e6 d5d4 g7g6 f1b5 e8d8 h1d1 d8d4 d1d4 g8g7 b5e2 a8c8 b2b3 b7b6 c1d2 c8c5 h5g6 h7g6 g2g3 g7f6 d4d8 c5f5 f2f4 f5c5 d8a8 c5c7 c2c4 g6g5 d2c3 g5f4 g3f4');
INSERT INTO evaluations VALUES(3925,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(3925,31,'cp 202',NULL);
INSERT INTO evaluations VALUES(3925,32,'cp 196',NULL);
INSERT INTO evaluations VALUES(3925,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(3925,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(3925,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(3925,36,'cp 218','e4e5 f6e8 h2h4 d8d7 h4h5 d7g4 e5d6 g4d4 d1d4 e8d6 c3d5 d6f5 d4f4 c7c6 f4f5 f8e8 f5g5 c6d5 g5d5 c8e6 d5d4 g7g6 f1b5 e8d8 h1d1 d8d4 d1d4 g8g7 b5e2 b7b6 c1d2 a8c8 b2b3 c8c5 h5g6 h7g6 d4d8 c5f5 f2f3 g7f6 e2d3 f5g5 g2g4 g5c5 d2e3 c5c7 a2a3 g6g5 e3d4');
INSERT INTO evaluations VALUES(6508,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(6508,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(6508,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(6508,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(6508,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6508,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(6508,36,'cp 18','g1f3 d7d6 c2c3 c5b6 e1g1 g8f6 d2d3 a7a5 f1e1 e8g8 c1e3 b6e3 e1e3 h7h6 h2h3 b7b5 d3d4 f8e8 b1d2 c8d7 a2a4 b5a4 a1a4 e5d4 c3d4 c6b4 a4a3 a5a4 d4d5 a8b8 f3d4 b4a6 d2c4 a6c5 d1c2 f6e4 g2e4 c5e4 e3e4');
INSERT INTO evaluations VALUES(7239,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(7239,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(7239,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(7239,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(7239,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(7239,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(7239,36,'cp 177','e8g8 e1g1 f6d7 c1g5 f7f6 g5h4 d7e5 f3e5 d6e5 b1c3 e5d4 a2a4 d8d7 h4g3 b7b5 d1b1 a6a5 c3e2 d4b6 b2b4 c8b7 e2f4 a5b4 b1b4 b6d4 c2c3 d4e5 d3d4 e5d6 b4b1 d7f7 f4d3 c6c5 g3d6 c7d6 d4c5 d6c5 b1b5 b7a6');
INSERT INTO evaluations VALUES(2633,30,'cp 189',NULL);
INSERT INTO evaluations VALUES(2633,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(2633,32,'cp 194',NULL);
INSERT INTO evaluations VALUES(2633,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(2633,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(2633,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(2633,36,'cp 195','e1g1 f6d7 b1c3 f7f5 c1g5 d8e8 e4f5 d7e5 g5f4 e5f3 d1f3 f8f5 f3e3 e8e3 f2e3 d6c5 d3d4 c5b6 c3a4 f5f7 a4b6 c7b6 e3e4 c8g4 f4e5 f7f1 a1f1 a8f8 f1e1 g4h5 a2a4 b6b5 a4a5 h5g6 b2b4 g6f7 e5g3 f7e6 e1b1 f8f7 c2c3 e6c4');
INSERT INTO evaluations VALUES(8410,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(8410,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(8410,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(8410,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(8410,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(8410,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(8410,36,'cp 21','e8g8 f1e2 d6e7 f3e5 a7a5 e2d3 c6e5 g3e5 b7b6 d1e2 c8b7 a2a4 d8c8 e5f6 e7f6 d2f3 f6e7 f3e5 e7d6 f2f4 b7a6 d3a6 c8a6 e2a6 a8a6 e1d2 h7h5 h2h3 a6a7 g2g4 d6e5 f4e5 h5h4 d2c2 a7c7 g4g5 f8c8 a1b1');
INSERT INTO evaluations VALUES(3291,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(3291,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(3291,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3291,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(3291,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(3291,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3291,36,'cp 33','f1e2 d6e7 e1g1 f6h5 f1e1 c8d7 e2d3 h5g3 h2g3 g7g6 a2a4 a8c8 b2b4 e7f6 a1c1 c8c7 d2b3 b7b6 b4b5 c6e7 b3d2 e7f5 f3e5 f5d6 d2f3 f6g7 c3c4 d5c4 d3c4 g7h6 c1c2 d6c4 e5c4 d7e8 c4e5 c7c2');
INSERT INTO evaluations VALUES(8725,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(8725,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(8725,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(8725,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(8725,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(8725,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(8725,36,'cp 38','f6h5 f4g5 h7h6 g5h4 d8d6 c3c4 e8g8 c4d5 d6d5 f1c4 d5d7 e1g1 e7e5 d4e5 c6e5 f3e5 g7e5 d1c2 d7c6 a1c1 c8f5 c2b3 c6b6 h4e7 b6b3 c4b3 f8e8 e7a3 c7c6 d2c4 e5c7 a3d6 c7b6 d6b4 b6c7 b4c3 f5e6 f1d1 a8d8 h2h3 d8d1 c1d1 e8d8 d1d8 c7d8 c4d2');
INSERT INTO evaluations VALUES(9980,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(9980,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(9980,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(9980,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(9980,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(9980,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(9980,36,'cp 47','e2e3 e8g8 d1c2 a7a5 g1f3 b7b6 f1e2 c8a6 e1g1 b4c3 d2c3 c6b4 c2b1 b4d5 c3e1 d5e7 b2b3 e7g6 b1c2 d7d6 e1c3 a6b7 f3d2 c7c5 d4c5 b6c5 a1d1 d8c7 c2b2 c7e7 e2d3 e6e5 f2f3 f6h5 d2e4 f7f5');
INSERT INTO evaluations VALUES(9974,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(9974,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(9974,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(9974,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(9974,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(9974,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(9974,36,'cp 45','e8g8 d1c2 a7a5 g1f3 d7d5 f1e2 b4c3 d2c3 d5c4 e2c4 c6b4 c2e2 b4d5 c3d2 f6e4 e1g1 b7b6 f1d1 e4d2 d1d2 c8b7 a1d1 a8c8 f3e5 d8g5 g2g3 g5e7 e3e4 d5f6 f2f3 f8d8 c4a6 c7c5 a6b7 e7b7 d4c5 d8d2 e2d2 b6c5 d2a5 b7b2');
INSERT INTO evaluations VALUES(10006,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(10006,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(10006,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(10006,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(10006,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(10006,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(10006,36,'cp 43','b2c3 d7d6 d2d4 e8g8 f1e1 c8g4 a1b1 a8b8 c1g5 d8e7 h2h3 g4f3 d1f3 e5d4 b5c6 b7c6 c3d4 b8b1 e1b1 e7e4 g5f6 e4f3 g2f3 g7f6 b1b7 f8c8 b7a7 c8b8 a7c7 b8b4 c7c6 b4d4 g1f1 d6d5 c6c5 d4a4 c5d5 a4a2 d5c5 a2a3 f1e2');
INSERT INTO evaluations VALUES(10003,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(10003,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(10003,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(10003,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(10003,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(10003,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(10003,36,'cp 40','d7d6 d2d4 e8g8 f1e1 c8g4 a1b1 a8b8 c1g5 d8e7 h2h3 g4f3 d1f3 e5d4 b5c6 b7c6 b1b8 f8b8 c3d4 h7h6 g5h4 e7e6 f3a3 g7g5 h4g3 f6h5 g1h2 h5g3 f2g3 e6c4 e4e5 c4d4 e5e6 f7e6 a3a6 g5g4 a6c6 g4h3 c6c7 b8f8');
INSERT INTO evaluations VALUES(10032,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(10032,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(10032,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(10032,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(10032,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(10032,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10032,36,'cp 5','f6e4 d1e1 e4f6 b5c6 d7c6 e1e5 d8e7 f1e1 e7e5 e1e5 c5e7 e5e2 f6d5 d2d3 f7f6 b1c3 e8f7 c3d5 c6d5 c1f4 c7c6 a1e1 e7f8 f3d4 c8d7 a2a4 h7h5 h3h4 a8c8 c2c3 c6c5 d4e6 c8e8 e6f8 e8e2 e1e2 h8f8 a4a5 b7b6 e2e1 f8e8 e1e8 d7e8 a5b6 a7b6 f4d6 f7e6 d6c7 d5d4 c7b6');
INSERT INTO evaluations VALUES(11489,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(11489,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(11489,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(11489,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(11489,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(11489,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(11489,36,'cp 131','a4c6 b7c6 f3e5 f8c5 d3d4 c5b6 b1c3 e8g8 e1g1 f8e8 f1e1 d7d5 e4d5 c6d5 c1f4 c8f5 c3a4 f6g4 e5f3 d8d7 a4b6 e8e1 f3e1 c7b6 h2h3 g4f6 f4e5 d7c6 c2c3 f6d7 e1d3 a8e8 d3b4 c6e6 b4a6 d7e5 d4e5 e6e5 a6b4 d5d4 d1d4');
INSERT INTO evaluations VALUES(15077,30,'cp -139',NULL);
INSERT INTO evaluations VALUES(15077,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(15077,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(15077,33,'cp -129',NULL);
INSERT INTO evaluations VALUES(15077,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(15077,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(15077,36,'cp -117','d5e4 g5f6 d8f6 d3e4 h7h5 f1c4 g7g5 d1e2 c8d7 f3d2 c6d8 d2f1 d8e6 c4e6 f6e6 f1e3 g5g4 b1d2 f8h6 b2b3 e8c8 d2c4 c8b8 e3f5 d7b5 f5h6 h8h6 a1d1 h6h8 h2h4 b5a6 d1d8 h8d8 e1g1 b7b6');
INSERT INTO evaluations VALUES(15206,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(15206,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(15206,32,'cp -114',NULL);
INSERT INTO evaluations VALUES(15206,33,'cp -127',NULL);
INSERT INTO evaluations VALUES(15206,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(15206,35,'cp -118',NULL);
INSERT INTO evaluations VALUES(15206,36,'cp -112','g5f6 d8f6 d3e4 g7g5 f1c4 c8d7 f3d2 g5g4 d2f1 e8c8 f1e3 h7h5 b1d2 f8h6 d1e2 c8b8 d2b3 f6g6 a1d1 h5h4 e1g1 g4g3 e3f5 h6f4 b3c5 g3f2 f1f2 d7f5 d1d8 h8d8 e4f5 g6h6 f2f3 d8d2 e2e4 a7a5 f5f6 h6f6 f3d3 f6g5 d3d2 f4d2');
INSERT INTO evaluations VALUES(17991,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(17991,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(17991,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17991,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(17991,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(17991,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(17991,36,'cp 14','g1f3 c6c5 c2c3 c5d4 c3d4 c7c5 e1g1 c5d4 f3d4 g8e7 b1d2 e7g6 f1e1 c8d7 d2f3 h7h6 g2g3 a8c8 d1d3 d8b6 a2a3 d7b5 d3d1 f8c5 c1e3 b6b7 d4b5 b7b5 e3c5 c8c5 f3d4 b5d7 d1d3 a6a5 h2h4 e8g8 h4h5 g6e7 e1c1');
INSERT INTO evaluations VALUES(17990,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(17990,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(17990,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(17990,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(17990,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(17990,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(17990,36,'cp 8','c6c5 c2c3 c8d7 e1g1 c5d4 c3d4 c7c5 b1c3 c5d4 c3e2 g8e7 e2d4 e7g6 f1e1 h7h6 g2g3 a8c8 d1d3 f8b4 c1d2 b4d2 d3d2 e8g8 b2b3 d8b6 e1c1 g6e7 g1g2 e7c6 d2e3 b6a5 a2a4 c6b4 e3d2 a5b6 a4a5 b6b7 h2h4');
INSERT INTO evaluations VALUES(20573,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20573,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20573,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(20573,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20573,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20573,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20573,36,'cp 0','a7a6 b5c4 b7b5 c4d5 g8e7 d2d3 c8b7 f3g5 e7d5 c3d5 h7h6 d1h5 g7g6 h5f3 h6g5 d5f6 e8e7 c1g5 d8c8 f6h7 e7e8 h7f6');
INSERT INTO evaluations VALUES(22292,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(22292,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(22292,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(22292,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(22292,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(22292,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(22292,36,'cp 111','c6e7 f3h4 g7g6 f2f4 f8h6 f4e5 d6e5 h4f3 c8d7 f1d3 h6c1 d1c1 h7h5 c1e3 e7c8 e1c1 g8h6 f3h4 g6g5 h4f3 d8e7 h2h3 c8b6 d1f1 e8c8 c1b1 h6f7 e3f2 f7d6 f3d2 h8f8 a2a4 c8b8 a4a5 b6a4 c3a4 d7a4 a5a6 b7b6');
INSERT INTO evaluations VALUES(22440,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(22440,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(22440,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(22440,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(22440,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22440,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22440,36,'cp 6','f8g7 d2d3 h7h6 b5c6 d7c6 c1e3 b7b6 d3d4 e5d4 e3d4 g8f6 d1d2 c8g4 e1c1 g4f3 g2f3 d8e7 d4f6 g7f6 f3f4 a8d8 d2d8 e7d8 d1d8 e8d8 h2h4 f6c3 b2c3 d8e7 c1d2 h8d8 d2e3 h6h5 h1g1 a7a5 a2a4 e7e6 g1b1 e6f6 f2f3 c6c5 e3e2 d8a8 b1g1');
INSERT INTO evaluations VALUES(6171,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(6171,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6171,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(6171,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6171,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6171,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6171,36,'cp 0','d2d3 h7h6 c1e3 c6d4 b5c4 g8e7 e1g1 e8g8 c3d5 e7d5 c4d5 c7c6 e3d4 e5d4 d5b3 g8h7 a2a4 d8c7 f3h4 g7f6 h4f3');
INSERT INTO evaluations VALUES(28733,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(28733,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(28733,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(28733,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(28733,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(28733,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28733,36,'cp 5','f6h5 f4g5 f8e7 h2h4 c7c5 b1d2 c5d4 g5e7 d8e7 c3d4 b8c6 f1b5 e8g8 e1g1 h5f6 a2a3 e7d8 b5d3 c6e7 e3e4 h7h6 a1c1 f6g4 f1e1 a8c8 c1c8 d8c8 e4e5 f7f6 d2e4 c8c7 e4d6 b7d5 d3e4 f6e5 d4e5 d5e4');
INSERT INTO evaluations VALUES(29178,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(29178,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29178,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(29178,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(29178,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(29178,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29178,36,'cp 12','e7e6 c1f4 b8c6 c2c3 f8d6 f4d6 d8d6 b1d2 e8g8 f1e1 g4h5 e2d3 d6f4 h2h3 f8e8 d3b5 f6d7 d1b3 e8c8 a2a4 f4c7 b5d3 a7a6 b3a3 h5g6 d3g6 h7g6 h3h4 e6e5 d4e5 c6e5 a3b3 e5d3 e1e3 d7c5 b3c2 d3f4');
INSERT INTO evaluations VALUES(29845,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(29845,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(29845,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(29845,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(29845,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(29845,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(29845,36,'cp 102','b8c6 c1e3 c6a5 c4b3 f8e7 f3g3 f6h5 g3g4 g7g6 g2g3 a5b3 a2b3 c7c6 e1c1 d8d7 g4f3 e7g5 d3d4 d7e7 h3h4 g5e3 f3e3 e8f8 f2f4 h5f6 d4d5 c6c5 d1f1 f8g7 f4f5 a7a6 h1g1 h6h5 f1f3 f6g4 e3d2 b7b5 g1f1');
INSERT INTO evaluations VALUES(25065,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(25065,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(25065,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(25065,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(25065,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(25065,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(25065,36,'cp 100','c1e3 c6a5 e1c1 a5c4 d3c4 c7c6 d1d2 d8d7 h1d1 d7e6 b2b3 f8e7 c1b2 f6d7 f3g3 e8g8 a2a4 f8d8 g3g4 d7f6 g4f3 f6e8 c3e2 e7g5 e3g5 h6g5 f3g4 e6g4 h3g4 d8d7 f2f3 a8d8 a4a5 g8f8 e2c3 e8f6');
INSERT INTO evaluations VALUES(25068,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(25068,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(25068,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(25068,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(25068,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(25068,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(25068,36,'cp 97','c6a5 e1c1 a5c4 d3c4 c7c6 d1d2 d8d7 b2b3 f8e7 h1d1 d7e6 c1b2 f6d7 f3g3 e8g8 a2a4 f8d8 g3g4 b7b6 a4a5 e6g4 h3g4 b6a5 f2f3 d7b6 d1a1 d8b8 b2c1 b6d7 d2d1 a5a4 a1a4 a7a5 c1b2 g8f8 a4a1 d7b6 e3g1 b6d7 c3a4');
INSERT INTO evaluations VALUES(31910,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(31910,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(31910,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(31910,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(31910,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(31910,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(31910,36,'cp 80','f2f3 g4c8 e2e4 e7e6 f1d3 g8f6 c1g5 f8e7 e4e5 f6d7 g5e7 d8e7 f3f4 a7a6 d1g4 g7g6 g1f3 c7c5 a2a3 b8c6 g4g3 c6d4 f3d4 c5d4 c3e2 d7c5 e2d4 c8d7 e1g1 e8g8 a1e1 c5d3 c2d3 e7c5 g3f2 c5b6 g2g4 d7b5 f2e3 f8c8 h2h3');
INSERT INTO evaluations VALUES(32765,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(32765,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32765,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32765,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32765,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32765,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32765,36,'cp 0','g8f6 d1e2 g7g6 b2b3 f8g7 c1b2 e8g8 c2c4 d7g4 f2f3 g4g5 d4c2 g5a5 b2c3 a5c5 c3d4 c5a5');
INSERT INTO evaluations VALUES(33018,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(33018,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(33018,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(33018,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(33018,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(33018,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(33018,36,'cp 96','c3e2 f6g6 d3d4 c8g4 f3d2 g4e2 c4e2 g8f6 e2d3 b8d7 c2c3 e8g8 f1e1 f8e8 d2c4 g6h5 d1h5 f6h5 c1d2 h5f6 c4a5 a8b8 a2a4 e7d8 a5b3 a7a5 g2g3 d8c7 g1g2 d6d5 e4d5 f6d5 c3c4 d5f6 b3a5 c7a5 d2a5 e5d4 a5c7');
INSERT INTO evaluations VALUES(33061,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(33061,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(33061,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(33061,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(33061,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(33061,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(33061,36,'cp 145','d2d4 g8f6 c1f4 e8g8 f4c7 b8c6 e1g1 e6d7 c7g3 f8d8 e2b5 f6h5 g3e5 a7a6 b5c6 d7c6 e5g7 h5g7 h2h3 c8e6 d1d2 b7b5 a2a3 a6a5 f3e5 c6b7 c3e2 g7f5 c2c3 b5b4 a3b4 a5b4 a1a8 b7a8 e2f4 b4c3 b2c3 e6b3 f1e1 f7f6 e5d3 b3c4');
INSERT INTO evaluations VALUES(33059,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(33059,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(33059,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(33059,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(33059,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(33059,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(33059,36,'cp 143','g8f6 c1f4 e8g8 f4c7 b8c6 d1d2 f6d5 c3d5 e6d5 c2c3 f8e8 c7f4 e7e5 d4e5 d5d2 f4d2 c6e5 f3e5 e8e5 d2f4 e5e8 f4e3 c8f5 a1d1 f5e6 a2a3 e6b3 d1d7 b7b6 e1g1 h7h5 f1e1 g7f6 e2b5 e8d8 d7b7 b3d5 b7c7 f6e5');
INSERT INTO evaluations VALUES(35177,30,'cp -200',NULL);
INSERT INTO evaluations VALUES(35177,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(35177,32,'cp -196',NULL);
INSERT INTO evaluations VALUES(35177,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(35177,34,'cp -209',NULL);
INSERT INTO evaluations VALUES(35177,35,'cp -207',NULL);
INSERT INTO evaluations VALUES(35177,36,'cp -196','e4e2 c4e2 b8c6 b1c3 f8c5 c3d5 c5b6 a2a4 g8f6 d5f6 g7f6 d2d3 h7h5 c1d2 d7d5 a4a5 b6c5 c2c3 a7a6 b2b4 c5d6 e1g1 c8g4 h2h3 g4d7 f1b1 h8g8 b4b5 a6b5 b1b5 d7h3 f3h4 h3g4 e2g4 h5g4 h4f5 e8d7 b5d5 c6e7 f5e7');
INSERT INTO evaluations VALUES(37154,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(37154,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(37154,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(37154,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(37154,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(37154,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(37154,36,'cp 45','d6b4 e1g1 d7d6 c3e2 b4c5 c2c3 c5b6 e2g3 h7h6 c4b3 c8e6 h2h3 f8e8 b3e6 f7e6 a2a3 c7c5 b2b4 b8c6 a1b1 c5b4 a3b4 e8f8 b4b5 c6e7 d1b3 d8d7 b1a1 e7g6 c3c4 a7a5 c1e3 b6e3 f2e3');
INSERT INTO evaluations VALUES(38154,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38154,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(38154,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38154,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(38154,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(38154,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(38154,36,'cp 58','c1f4 g8f6 e2e3 d8a5 f3d2 e8g8 a1c1 c8d7 a2a3 b4c3 c1c3 f8c8 c3c8 d7c8 f1d3 c8d7 d1b3 d7c6 f2f3 b8d7 b3b4 a5d8 f4g3 a7a5 b4b3 f6h5 g3f2 e6e5 d4e5 d7e5 d3e2 e5g6 b3d3 a8c8 e1g1 d5d4 f1c1 d4e3 d3d8 c8d8 f2e3 h5f4 e2f1');
INSERT INTO evaluations VALUES(38668,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(38668,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(38668,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(38668,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(38668,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(38668,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(38668,36,'cp 70','e2e4 d6b4 e4d5 e6d5 f1d3 b8c6 e1g1 g8e7 c3e2 b4d6 e2g3 c6b4 d3b5 b4c6 f1e1 e8g8 b5d3 e7g6 c2c3 c6e7 a2a4 c7c6 h2h4 g6f4 c1f4 d6f4 d1c2 g7g6 h4h5 g8g7 a4a5 e7g8 a5a6 b7b5 f3e5 d8f6 e1e2 g8h6 a1e1 c8g4 e5g4');
INSERT INTO evaluations VALUES(38675,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(38675,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(38675,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(38675,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(38675,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(38675,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(38675,36,'cp 68','d6b4 e4d5 e6d5 f1d3 b8c6 e1g1 g8e7 c3e2 b4d6 e2g3 c6b4 d3e2 e8g8 c2c3 b4c6 e2d3 e7g6 f1e1 c6e7 a2a4 c7c6 f3e5 c6c5 e5g6 h7g6 d4c5 d6c5 c1g5 d8c7 d1d2 f8e8 d3b5 c8d7 g5f4 c5d6 f4d6 c7d6 g3e4 d6c7 e4g5 d7c6 b5d3 c7d6 h2h4');
INSERT INTO evaluations VALUES(49409,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(49409,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(49409,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(49409,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(49409,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(49409,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(49409,36,'cp -49','c7c5 c2c3 c5d4 c3d4 b7b5 e2e3 b5b4 f1d3 c8a6 a2a4 f8e7 d3a6 b8a6 d1d3 d8b6 b1d2 b6b7 d2b3 a6c7 c1d2 e8g8 a1c1 a7a5 e1g1 c7d5 f3e5 d5b6 d3b5 f8b8 b3c5 b7d5 d2e1 d5a2 e1h4 e7c5 d4c5 b6a4');
INSERT INTO evaluations VALUES(50583,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(50583,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(50583,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(50583,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(50583,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(50583,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(50583,36,'cp 76','b1d2 d7d5 f1d3 d5e4 d2e4 g8f6 e4f6 e7f6 g1e2 f8d6 c2c4 e8g8 d1c2 d6b4 c1d2 b4d2 c2d2 f6f5 e1g1 g7g6 f1e1 b8d7 e2f4 d7f6 d3c2 a6a5 e1e5 d8b6 d2c3 f8d8 a1d1 b6b4 h2h4 c8d7 c4c5 b4c3 b2c3 d8e8 c3c4 h7h6 f4d3 e8e5 d3e5 d7e6 c2b3 f6e4');
INSERT INTO evaluations VALUES(50613,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(50613,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50613,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(50613,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(50613,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(50613,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(50613,36,'cp 4','b8c6 d2d4 c5d4 f3d4 e7e6 b1c3 c6d4 d1d4 b7b5 c1e3 c8b7 e1c1 a8c8 c1b1 g8f6 e4e5 f6d5 c3d5 b7d5 d4d2 f8e7 f1d3 d7d6 e5d6 d8d6 h1f1 d6b4 f4f5 b4d2 d1d2 e8g8 g2g3 e7f6 c2c3 h7h5 d3c2 d5c4 f1e1');
INSERT INTO evaluations VALUES(51592,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(51592,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(51592,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(51592,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(51592,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(51592,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(51592,36,'cp 62','c8b7 b1d2 e7e6 f1d3 c5d4 c3d4 g8f6 g1f3 b7a6 d3a6 b8a6 e1g1 d8c8 e4e5 f6d5 d2e4 f8e7 c1g5 e8g8 h2h4 c8d8 f1e1 h7h6 g5e7 d5e7 h4h5 a6c7 f3h4 c7e8 e4c3 e8c7 g2g3 a8c8 c3e4 e7f5 h4f5 e6f5');
INSERT INTO evaluations VALUES(52154,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(52154,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(52154,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(52154,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(52154,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(52154,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(52154,36,'cp 115','f1c4 c3b2 c1b2 d7d6 d1b3 g8h6 e1g1 b8c6 b2c3 d8b6 b3a4 a7a6 c4d5 b6c7 f1e1 b7b5 e4e5 b5a4 e5d6 e8d8 d6c7 d8c7 c3a5 c6a5 d5a8 a5c4 b1c3 c4b6 c3d5 b6d5 a8d5 f8d6 f3e5 c8e6 d5c4 h8b8 c4e6 f7e6 f2f4');
INSERT INTO evaluations VALUES(52142,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(52142,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(52142,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(52142,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(52142,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(52142,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(52142,36,'cp 89','b8c6 b1c3 d7d6 c1f4 f8e7 d1b3 c6a5 c4b5 e8f8 b3c2 c8g4 e1c1 a7a6 b5e2 a5c6 e4e5 g4f3 e2f3 e7g5 c2f5 g5f4 f5f4 c6e5 f3b7 a8b8 c3e4 c5c4 e4d6 d8f6 f4f6 g8f6 d1d4 h7h5 c1b1 h8h6 f2f4 e5d3 d6c4 d3f4 b7f3 g7g5 h1e1 f4e6 d4d6 g5g4 f3d1 e6c5 d1c2 h5h4 e1d1 h4h3');
INSERT INTO evaluations VALUES(53255,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(53255,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(53255,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(53255,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(53255,34,'cp -98',NULL);
INSERT INTO evaluations VALUES(53255,35,'cp -110',NULL);
INSERT INTO evaluations VALUES(53255,36,'cp -76','c5d4 f3d4 g8f6 e2e3 b8c6 f1b5 d8c7 e1g1 a7a6 b5e2 e7e5 f4e5 c7e5 c2c4 f8c5 d4f3 e5h5 f3d4 h5g5 c4d5 f6d5 g1h1 e8g8 d4f3 g5d8 e3e4 d5b4 b1c3 c8e6 c1f4 d8d1 f1d1 a8d8 a2a3 b4c2 d1d8 f8d8 a1d1 d8d1 e2d1 c2d4 c3d5 d4f3 g2f3');
INSERT INTO evaluations VALUES(53752,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(53752,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(53752,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(53752,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(53752,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(53752,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(53752,36,'cp -73','c4d5 c6d5 d2d4 e5e4 f2f3 f8e7 b1c3 e7h4 e1f1 b8c6 f3e4 d5e4 g4g5 h4g5 c1g5 d8g5 g2e4 g8f6 g1f3 g5h5 d4d5 c6e7 h1g1 e8g8 d5d6 e7g6 e4g6 c8h3 f1f2 h7g6 g1g5 h5h6 d1d4 a8e8 a1d1 h3d7 h2h4 d7c6 g5e5 e8d8 d1d2 f8e8 f2g1 e8e5 f3e5');
INSERT INTO evaluations VALUES(53701,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(53701,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(53701,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(53701,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(53701,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(53701,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(53701,36,'cp -95','c6d5 d2d4 e5e4 f2f3 b8c6 b1c3 f8e7 f3e4 d5e4 c1f4 d8d4 d1d4 c6d4 a1d1 d4e6 f4c1 h7h5 c3e4 h5g4 h2h3 g4h3 e4d6 e7d6 d1d6 h3h2 g1f3 c8d7 h1h2 h8h2 f3h2 a8c8 c1d2 c8c2 d2c3 g8f6 c3f6 g7f6 d6d2 c2c1 e1f2');
INSERT INTO evaluations VALUES(54607,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(54607,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(54607,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(54607,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(54607,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(54607,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(54607,36,'cp -76','d2d4 e5e4 f2f3 b8c6 b1c3 f8e7 c1f4 e7h4 f4g3 e4f3 g2f3 g8h6 f3d5 c8g4 g1f3 g4f3 d5f3 e8g8 f3c6 b7c6 d1d3 d8g5 c3e4 h4g3 h2g3 g5g6 e1c1 h6g4 d1f1 f8e8 f1f4 h7h5 e4f2 g6d3 f2d3 g4f6 d3e5 c6c5 e2e3 a8c8 c1b1 f6d5 f4f5 f7f6 f5h5 f6e5 e3e4');
INSERT INTO evaluations VALUES(53987,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(53987,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(53987,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(53987,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(53987,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(53987,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53987,36,'cp -12','a7a6 c4b3 d6d5 d2d3 d5e4 d3e4 d8d1 e1d1 b7b5 c2c3 c8b7 b3c2 g8f6 a2a4 f8e7 a4b5 a6b5 a1a8 b7a8 h1e1 e8g8 b1a3 f8d8 c1d2 a8c6 c3c4 b5c4 d1c1 f6d7 a3c4 c6b5 c4a3 b5a6 d2a5 d8c8 e1d1 c5c4 a5c3 d7c5');
INSERT INTO evaluations VALUES(50742,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50742,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50742,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50742,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50742,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50742,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50742,36,'cp 0','c4b3 d6d5 d2d3 d5e4 d3e4 d8d1 e1d1 b7b5 c2c3 c8b7 b3c2 g8f6 a2a4 f8e7 a4b5 a6b5 a1a8 b7a8 b1d2 e8g8 b2b3 c5c4 b3c4 b5c4 h1e1 b8d7 d2c4 f8c8 f3d2 f6g4 d1e2 g4h2 c1b2 a8c6 c4a5 c6a8');
INSERT INTO evaluations VALUES(54743,30,'cp 270',NULL);
INSERT INTO evaluations VALUES(54743,31,'cp 269',NULL);
INSERT INTO evaluations VALUES(54743,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(54743,33,'cp 275',NULL);
INSERT INTO evaluations VALUES(54743,34,'cp 269',NULL);
INSERT INTO evaluations VALUES(54743,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(54743,36,'cp 262','d2d4 e7e6 f1c4 f8d6 e1g1 g8e7 d1e2 b8c6 f1e1 e8g8 c4e6 g8h8 e6c8 a8c8 c3e4 d8d7 c2c3 a7a6 c1d2 e7g6 e2d3 f8e8 a1d1 c8d8 d3b1 b7b5 h2h3 c6e7 b2b3 e7f5 c3c4 d6f8 d4d5 b5c4 b3c4 c7c6 e4c3 c6d5 e1e8');
INSERT INTO evaluations VALUES(54741,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(54741,31,'cp 271',NULL);
INSERT INTO evaluations VALUES(54741,32,'cp 268',NULL);
INSERT INTO evaluations VALUES(54741,33,'cp 258',NULL);
INSERT INTO evaluations VALUES(54741,34,'cp 256',NULL);
INSERT INTO evaluations VALUES(54741,35,'cp 273',NULL);
INSERT INTO evaluations VALUES(54741,36,'cp 257','e7e6 f1c4 f8d6 e1g1 b8c6 d1e2 g8e7 c1d2 e8g8 c4e6 g8h8 e6c8 a8c8 c3e4 d8d7 f1e1 f8e8 e2d3 b7b6 e1e2 e7f5 g1f1 c6b4 d2b4 d6b4 a2a3 b4f8 a1d1 f5d6 e4d6 f8d6 h2h3 e8e2 d3e2');
INSERT INTO evaluations VALUES(55195,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(55195,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(55195,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(55195,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(55195,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(55195,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(55195,36,'cp 42','f2e3 b8c6 b1c3 e7e5 d2d4 f7f5 d4d5 e5e4 d5c6 d8d1 e1d1 e4f3 g2f3 b7c6 f1d3 g8f6 d1c2 f8d6 h2h4 f6h5 c1d2 g7g6 a1e1 e8g8 a2a4 a8b8 b2b3 f8e8 a4a5 g8f7 a5a6 c8a6 e1a1 a6b7 e3e4 c6c5 e4f5 g6f5 d3f5 b7f3');
INSERT INTO evaluations VALUES(5096,30,'cp -165',NULL);
INSERT INTO evaluations VALUES(5096,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(5096,32,'cp -162',NULL);
INSERT INTO evaluations VALUES(5096,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(5096,34,'cp -179',NULL);
INSERT INTO evaluations VALUES(5096,35,'cp -178',NULL);
INSERT INTO evaluations VALUES(5096,36,'cp -175','g8f6 f3e3 d7d5 c4b3 e8g8 e1g1 a7a5 b3a4 b7b5 e3c5 b5a4 c5c6 c8d7 c6c5 h7h6 e4e5 h6g5 e5f6 g7f6 c5d6 a8c8 c2c3 e6e5 d6d5 d7f5 d5d8 f8d8 c1e3 c8b8 b2b3 g8g7 b3a4 f5d3 f1d1 b8b2 b1a3 g5g4 a3b5 d8d7 c3c4 b2b4 c4c5 b4a4');
INSERT INTO evaluations VALUES(7808,30,'cp -206',NULL);
INSERT INTO evaluations VALUES(7808,31,'cp -186',NULL);
INSERT INTO evaluations VALUES(7808,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(7808,33,'cp -181',NULL);
INSERT INTO evaluations VALUES(7808,34,'cp -180',NULL);
INSERT INTO evaluations VALUES(7808,35,'cp -186',NULL);
INSERT INTO evaluations VALUES(7808,36,'cp -193','f3e3 e8g8 c4b3 d7d5 e1g1 a7a5 c2c3 a5a4 b3d1 d5d4 e3e2 d4c3 b2c3 b7b5 c1e3 h7h6 g5f3 f6d7 a2a3 c8a6 f1e1 g8h7 d1c2 d8c7 h2h4 f7f5 h4h5 f5f4 h5g6 h7h8 e3c1 d7e5 b1d2 b5b4 a3b4 c5b4 a1a4 b4c3 d2c4 e5f3 e2f3 c6d4 f3d1');
INSERT INTO evaluations VALUES(5823,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(5823,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(5823,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(5823,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(5823,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(5823,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(5823,36,'cp 37','g8f6 g1f3 d8d6 f1b5 h7h6 e1g1 c8d7 b1d2 a7a6 b5a4 e8g8 c3c4 a6a5 a1c1 d6b4 b2b3 f8d8 d2b1 b4d6 b1c3 c6b8 f3e5 d7a4 c3a4 b8d7 e5d3 d5c4 b3c4 b7b6 d1b3 c7c5 d4c5 d7c5 d3c5');
INSERT INTO evaluations VALUES(9005,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(9005,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(9005,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(9005,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(9005,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(9005,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(9005,36,'cp 44','g1f3 d8d6 f1b5 h7h6 e1g1 c8d7 b1d2 a7a6 b5d3 e6e5 e3e4 e8g8 e4d5 d6d5 c3c4 d5d6 c4c5 d6e7 f1e1 d7g4 d1b1 e5e4 d3e4 f6e4 e1e4 e7d7 b1c1 a8d8 d4d5 g4f3 d5c6 d7d2 g2f3 b7c6 e4e7 d2d3 g1g2 f8e8 c1e1 g8f8 e7e8 d8e8 e1a5 e8e7 b2b3 d3d4 a1f1 f8g8 a5a6 d4c5 f1h1 c5c2 a6a3');
INSERT INTO evaluations VALUES(6056,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(6056,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(6056,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(6056,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(6056,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(6056,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(6056,36,'cp 48','d4e5 b4c3 b2c3 d6e5 d1d8 e8d8 a2a4 f7f6 a4a5 g8e7 c1e3 a8b8 f3d2 b7b6 d2b3 c8d7 e1d2 e7c8 c3c4 c8d6 f2f3 d7e6 d2c3 d8e7 c4c5 b6c5 b3c5 h8d8 a5a6 h7h5 f1d3 c6b4 h1b1 b4d3 c2d3 d6b5 c3d2 h5h4');
INSERT INTO evaluations VALUES(6301,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(6301,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(6301,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(6301,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(6301,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(6301,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(6301,36,'cp 100','b1c3 d7d6 c1e3 g8f6 h2h3 e8g8 f2f4 e5g6 d1d3 c5d4 d3d4 c8e6 h1f1 e6b3 a2b3 f8e8 e1c1 a7a5 g2g4 b7b5 e4e5 d6e5 f4e5 g6e5 d4d8 a8d8 d1d8 e8d8 e3f4 d8e8 c3b5 f6d5 f4g3 f7f6 f1d1 c7c6 b5d4 d5b4 c2c3 b4d3 c1b1 d3c5 b1c2 e5g6 g3c7 c5e6 c7d6 e6d4 d1d4 g6e5');
INSERT INTO evaluations VALUES(7585,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(7585,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(7585,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(7585,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(7585,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(7585,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(7585,36,'cp -18','d8e7 d1e2 f6d5 c3e4 c5d4 f2f4 f7f5 e5f6 d5f6 e4f6 d4f6 e2e7 e8e7 f1d3 d7d6 a1b1 a7a5 e1f2 a5a4 h1e1 e7f7 c1d2 c8e6 a2a3 c6c5 d2c3 a8b8 d3e4 e6d7 e4d3 h7h5 d3c4 f7g6 c4d3 d7f5 c3f6 g7f6 d3f5 g6f5 e1e7 f5f4 e7c7 h8e8 c7f7');
INSERT INTO evaluations VALUES(446,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(446,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(446,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(446,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(446,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(446,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(446,36,'cp -29','d1e2 f6d5 c3e4 c5d4 f2f4 f7f5 e5f6 d5f6 e4f6 d4f6 e2e7 e8e7 f1d3 d7d5 a1b1 a7a5 a2a4 c6c5 b2b3 e7d6 c1d2 f6d4 e1d1 h8e8 h2h3 c8d7 d3h7 c5c4 h7g6 c4c3 g6e8 a8e8 d2e1 d7f5 h1f1 f5e4 e1f2 c7c5 f2d4 c5d4 f1e1 d6c5');
INSERT INTO evaluations VALUES(13624,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13624,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(13624,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13624,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(13624,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(13624,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(13624,36,'cp 14','e2e3 h7h6 g5h4 f8e7 f1d3 f6e4 h4e7 e4c3 d1c2 d8e7 c2c3 c8d7 e1g1 e7b4 c3c2 e8g8 a1c1 a8c8 c2e2 b4d6 e2d2 c8c7 c1c3 f8c8 f1c1 c6e7 f3e5 d7e8 d3b1 d6b6 e5d3 c7c3 c1c3 c8c3 d2c3 b6d6');
INSERT INTO evaluations VALUES(13623,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(13623,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(13623,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(13623,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(13623,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(13623,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13623,36,'cp 26','h7h6 g5h4 f8e7 f1d3 f6e4 h4e7 e4c3 b2c3 c6e7 e1g1 e8g8 f3e5 d8a5 c3c4 e7c6 e5c6 b7c6 c4c5 c8a6 d3a6 a5a6 d1c2 a8e8 a1b1 e6e5 d4e5 e8e5 b1b4 f8e8 c2a4 a6a4 b4a4 e5e7 f1b1 e7c7 b1b4 g7g5 a4a6 h6h5 g2g3 g8g7 b4a4 e8b8 a6a7 b8b1 g1g2 c7a7 a4a7');
INSERT INTO evaluations VALUES(14044,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(14044,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(14044,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(14044,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(14044,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(14044,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(14044,36,'cp 59','h7h6 g5d2 e7f5 c2c3 c6b8 b1a3 f8a3 d1a4 b8d7 a4a3 a7a5 h2h4 b7b6 a3a4 c8a6 f1a6 a8a6 h4h5 d8c8 g2g4 f5e7 e1c1 a6a8 a4b5 c7c5 c1b1 a5a4 d1g1 a8a5 b5e2 c8a6 e2a6 a5a6 a2a3 a6a8 f3e1 b6b5 e1d3');
INSERT INTO evaluations VALUES(14037,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(14037,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(14037,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(14037,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(14037,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(14037,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(14037,36,'cp 82','g5c1 f7f6 f1d3 f6e5 f3e5 c6e5 d4e5 d8d7 d1h5 g7g6 d3g6 e7g6 h5g6 d7f7 g6f7 e8f7 b1c3 b7b6 f2f3 c8b7 c3e2 h8g8 g2g3 a7a5 e1f2 a5a4 h2h4 b7a6 h1e1 a6e2 e1e2 h6h5 c1g5 f8c5 f2g2 c5e7 f3f4 a4a3 b2b3 b6b5 c2c3 c7c5 a1d1 a8c8 c3c4 b5c4 b3c4 d5d4 d1b1 e7g5 h4g5');
INSERT INTO evaluations VALUES(14170,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(14170,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14170,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14170,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14170,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14170,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14170,36,'cp 0','e7e6 d1e2 c8d7 d2d3 f8c5 b2b3 c6e7 b5d7 d8d7 f3e5 d7d8 b1d2 a7a5 a2a4 e7f5 d2f3 h7h5 e1g1 f6g4 d3d4 c5d6 c1d2 e8g8 e5g4 h5g4 f3g5 f5h6 e2d3 h6f5');
INSERT INTO evaluations VALUES(16046,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(16046,31,'cp -66',NULL);
INSERT INTO evaluations VALUES(16046,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(16046,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(16046,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(16046,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(16046,36,'cp -50','g1e2 f8g7 d2d3 e8g8 b5c6 b7c6 b1c3 d7d5 c1g5 c8e6 c3a4 h7h6 g5f6 g7f6 e1g1 h6h5 h2h3 a7a5 a4c5 e6c8 a2a3 f6e7 b2b4 a8b8 d3d4 d5e4 f3c3 e5d4 e2d4 d8d5 c5b3 d5g5 d4c6 a5b4 c6e7 g5e7 a3b4 b8b4');
INSERT INTO evaluations VALUES(21636,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(21636,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(21636,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(21636,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(21636,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21636,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(21636,36,'cp 61','f8b4 b1d2 d5c4 e2e3 d8d5 f1e2 c8f5 e1g1 b4d2 c1d2 g8e7 d2c3 a8d8 d1a4 f5d3 f1d1 e8g8 f3e1 c6e5 c3e5 d5e5 e1d3 c4d3 e2d3 g7g6 d1d2 e7c8 a1d1 a7a5 g2g3 c8b6 a4c2 f8e8 d3f1 d8d2 d1d2');
INSERT INTO evaluations VALUES(23427,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(23427,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(23427,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23427,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(23427,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23427,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23427,36,'cp 1','e7e5 a2a3 g8f6 b2b4 d7d5 c4d5 f6d5 c1b2 f8d6 d1c2 e8g8 g1f3 d8e7 f1d3 g8h8 b4b5 c6d8 b1c3 d5c3 c2c3 f8e8 a3a4 f7f6 d3e4 d8e6 e1g1 e6c5 e4c2 e5e4 f3d4 e7e5 f2f4 e4f3 d4f3 e5c3 b2c3 c5e4 c2e4 e8e4 f3d4 c8d7 a4a5');
INSERT INTO evaluations VALUES(23838,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(23838,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(23838,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23838,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(23838,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(23838,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(23838,36,'cp 21','h2h3 g4f3 d1f3 g8f6 b1c3 a7a6 b5c6 b7c6 d2d3 e8g8 e1g1 a6a5 a2a4 d6b4 c3e2 d8e7 a1d1 f8e8 e2g3 g7g6 e3e4 h7h5 b2c1 e7d6 g3e2 a8d8 c1g5 e8e6 g1h1 d8e8 f3e3 b4c5 e3g3 d5e4 g5f6 e6f6 d3e4 d6e7');
INSERT INTO evaluations VALUES(23837,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(23837,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(23837,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(23837,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(23837,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(23837,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(23837,36,'cp 18','g4f3 d1f3 g8f6 b1c3 a7a6 b5c6 b7c6 e1g1 e8g8 d2d3 a6a5 a2a4 f8e8 e3e4 d6b4 c3e2 d8d6 e2g3 g7g6 g1h1 h7h5 g3e2 e8e6 e2g1 d5e4 d3e4 d6d2 a1c1 d2f4 f3f4 e5f4 f2f3 h5h4 g1e2 f6h5 c1d1 h5g3 e2g3 f4g3');
INSERT INTO evaluations VALUES(23840,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(23840,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(23840,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(23840,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(23840,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(23840,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(23840,36,'cp 21','d1f3 g8f6 b1c3 a7a6 b5c6 b7c6 d2d3 a6a5 e1g1 e8g8 a2a4 d6b4 e3e4 f8e8 c3e2 g7g6 a1d1 d8e7 g2g4 h7h5 g1g2 b4d6 b2c1 g8g7 c1g5 h5g4 h3g4 e7e6 g5f6 e6f6 f3f6 g7f6 f1h1 e8h8 f2f3 h8h1 d1h1 f6e6');
INSERT INTO evaluations VALUES(23839,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(23839,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(23839,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(23839,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23839,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(23839,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(23839,36,'cp 25','g8f6 b1c3 a7a6 b5c6 b7c6 d2d3 e8g8 e1g1 f8e8 e3e4 d6c5 c3e2 d8d6 g1h1 c5a3 b2a3 d6a3 e2g3 a6a5 f3e3 a5a4 h1h2 e8e6 f2f4 e5f4 e3f4 d5e4 g3e4 a3e7 e4f6 e6f6 f4g3 a4b3 a2b3 a8a1 f1a1 h7h6 a1e1 e7d7 e1e4 c6c5 e4g4 g7g6 g4e4 d7d6 g3d6');
INSERT INTO evaluations VALUES(24100,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24100,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24100,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(24100,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(24100,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(24100,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(24100,36,'cp 28','b1c3 a7a6 b5c6 b7c6 d2d3 e8g8 e1g1 a6a5 a2a4 d6b4 c3a2 b4d6 e3e4 f8e8 a2c3 d6b4 c3e2 g7g6 a1d1 d8d6 f3e3 f6h5 e2g3 h5g3 e3g3 e8e6 b2c1 a8e8 c1g5 f7f6 g5e3 d6e7 g1h1 b4a3 e3d2 a3b4 d2b4 e7b4');
INSERT INTO evaluations VALUES(24332,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(24332,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(24332,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(24332,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(24332,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(24332,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(24332,36,'cp 34','h7h6 b1d2 b7b5 a4c2 c6e7 d3d4 e7g6 h2h3 f8e7 f1e1 c7c5 d2f1 e8g8 f1g3 d8c7 g3f5 f8e8 f5e7 e8e7 d4d5 e7e8 c1e3 a6a5 f3d2 a5a4 c3c4 e8b8 c2d3 b5c4 d2c4 d7b5 d1c2 f6d7 d3f1 d7b6 b2b3 a4b3 a2b3 b5c4 b3c4');
INSERT INTO evaluations VALUES(24949,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(24949,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(24949,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(24949,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(24949,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24949,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(24949,36,'cp 37','e7e6 b1d2 f8e7 c2c4 h7h6 a2a3 a7a5 b2b3 g4f5 c1b2 e8g8 h2h3 f6e4 d2e4 f5e4 b2c3 e7f6 d1c1 b7b6 f3d2 e4f5 c4d5 e6d5 b3b4 d8d7 c1b2 a5b4 a3b4 c6e7 f1c1 a8a1 c1a1 f8b8 d2f3 c7c6 b4b5 d7e6 a1c1 c6b5 b2b5 b8c8 c3d2 f5e4 c1c8 e7c8');
INSERT INTO evaluations VALUES(25223,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(25223,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(25223,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(25223,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(25223,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(25223,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(25223,36,'cp 37','b1d2 f8e7 c2c4 e8g8 b2b3 c6b8 c1b2 b8d7 c4c5 a7a5 h2h3 g4f5 g2g4 f5e4 g4g5 f6e8 d2e4 d5e4 f3d2 e7g5 d2e4 c7c6 e2c4 g5h6 e4g3 d7f6 e3e4 f6d5 e4d5 e6d5 g3f5 d8g5 d1g4 d5c4 b3c4 e8f6 g4g5');
INSERT INTO evaluations VALUES(26298,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26298,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26298,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26298,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26298,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26298,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26298,36,'cp 0','d8c7 h2h3 g4h5 d1b3 a8d8 g2g4 h5g6 c1f4 c7f4 b3b7 c6b8 e2b5 b8d7 b5d7 d8d7 b7c8 d7d8 c8c6 d8d7');
INSERT INTO evaluations VALUES(29042,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(29042,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(29042,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(29042,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(29042,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(29042,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(29042,36,'cp 40','e7e6 h2h3 g4f3 d1f3 f8d6 g2g4 d6f4 f3f4 b8d7 e1c1 a8c8 c1b1 e8g8 g4g5 f6h5 f4d6 g7g6 f1e2 h5g7 h3h4 c6c5 h4h5 d8c7 d6c7 c8c7 c2c3 c5d4 e3d4 d7b6 h5h6 g7f5 d2f3 b6c8 e2d3 c8d6 f3e5 b7b5 f2f4');
INSERT INTO evaluations VALUES(29293,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(29293,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(29293,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(29293,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(29293,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(29293,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(29293,36,'cp 61','h2h3 g4f3 d1f3 f8d6 g2g4 d6f4 f3f4 b8d7 e1c1 d8e7 h3h4 e8c8 c1b1 e6e5 f4g3 h7h6 f1d3 g7g6 c2c4 h8e8 h4h5 g6h5 c4d5 c6d5 g4h5 c8b8 g3h4 a7a6 d1c1 b8a8 d3f5 e7d6 d2b3 e5d4 h4d4 d6e5 d4e5 d7e5 b3d4 f6e4 c1c2');
INSERT INTO evaluations VALUES(31546,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(31546,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(31546,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(31546,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(31546,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(31546,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(31546,36,'cp -17','c6c5 c3e2 b8c6 c2c3 f5e4 c1e3 c5d4 c3d4 g8h6 e2g3 f8e7 g1f3 h6f5 g3f5 e4f5 f1d3 f5d3 d1d3 d8b6 e1g1 e8g8 h2h4 h7h6 e3d2 f8c8 f1b1 a7a5 h4h5 c8c7 b1c1 a5a4 c1c2 a8c8 a1c1 c6a7 c2c7 c8c7 c1c7');
INSERT INTO evaluations VALUES(31677,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(31677,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(31677,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(31677,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(31677,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(31677,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(31677,36,'cp -12','c3e2 f5e4 c2c3 b8c6 c1e3 c5d4 c3d4 g8h6 e2g3 f8e7 g1f3 h6f5 g3f5 e4f5 f1d3 f5d3 d1d3 d8b6 e1g1 e8g8 b2b4 a7a6 f1b1 f8c8 h2h4 h7h6 h4h5 c8c7 f3d2 b6b5 d3b5 a6b5 b1c1 a8c8 d2b3 c6d8 c1c7 c8c7 a1c1 c7c4 b3c5');
INSERT INTO evaluations VALUES(32210,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(32210,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(32210,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32210,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(32210,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(32210,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(32210,36,'cp 40','c4b3 g8f6 g1e2 d6d5 e4d5 f6d5 e1g1 b8c6 d2d4 d5c3 b2c3 e6b3 a2b3 f8d6 d4d5 c6e7 e2g3 e8g8 g3e4 e7g6 c3c4 b7b6 d1h5 a7a5 h5f5 g6h4 f5g5 d8g5 c1g5 h4f5 g2g4 f5d4 e4d6 c7d6 g5e7 f8c8 e7d6 d4c2 a1a2');
INSERT INTO evaluations VALUES(32524,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(32524,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(32524,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(32524,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(32524,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(32524,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(32524,36,'cp -11','b5d7 d8d7 e4d5 d7d5 d1f3 d5e6 g1e2 b8c6 e1g1 e8c8 d2d4 g8f6 c2c3 e6g4 f3g4 f6g4 h2h3 g4f6 b1d2 e7e5 d2f3 e5d4 e2d4 c6d4 f3d4 f8c5 d4b3 f6e4 b3c5 e4c5 f1e1 c5a4 a1b1 h8e8 c1e3 a7a6 g1f1 f7f6 e1e2 b7b5 b1e1 e8e4 a2a3 a6a5');
INSERT INTO evaluations VALUES(32466,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(32466,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(32466,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(32466,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(32466,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(32466,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(32466,36,'cp -8','d8d7 e4d5 d7d5 d1f3 d5e6 g1e2 b8c6 e1g1 e8c8 d2d4 g8f6 c2c3 e6d7 c1f4 e7e5 d4e5 f6g4 f3g3 d7f5 b1d2 g4e5 d2b3 f8d6 e2d4 c6d4 b3d4 e5f3 g3f3 f5f4 f3f4 d6f4 d4f3 h8e8 f1e1 g7g6 g2g3 e8e1 a1e1 d8d3 g1f1 f4d6 f3g5 d3d2 g5f7 d2b2 f7d6 c7d6');
INSERT INTO evaluations VALUES(32956,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(32956,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(32956,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(32956,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(32956,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(32956,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(32956,36,'cp -6','e4d5 d7d5 d1f3 d5e6 g1e2 b8c6 e1g1 e8c8 d2d4 g8f6 c2c3 e6g4 b1d2 e7e5 f3g4 f6g4 h2h3 g4f6 d2f3 e5d4 e2d4 c6d4 f3d4 f8c5 d4b3 c5b6 c1g5 d8d6 a1e1 h8d8 e1e7 d8d7 e7e2 a7a5 f1e1 h7h6 g5f4 d6d5 a2a4 h6h5 g2g3 d7d8 g1f1 h5h4');
INSERT INTO evaluations VALUES(34215,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(34215,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(34215,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(34215,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(34215,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(34215,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(34215,36,'cp 170','f6d4 f3d4 d7d6 c1f4 c8e6 e1c1 b8d7 f1e2 g8f6 d4e6 f7e6 f4d6 f8d6 d1d6 e8e7 h1d1 d7e5 h2h4 h7h6 h4h5 a8g8 f2f4 e5f7 d6d2 e6e5 g2g3 g8d8 a2a4 b7b6 b2b4 a7a5 b4b5 d8d2 d1d2 c6b5 e2b5');
INSERT INTO evaluations VALUES(34222,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(34222,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(34222,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(34222,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(34222,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(34222,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(34222,36,'cp 171','f3d4 f8c5 c1e3 g8f6 e4e5 f6g4 c3e4 d7d6 e4d6 c5d6 e5d6 g4e3 f2e3 b8d7 f1e2 e8g8 e1c1 f8e8 d4f5 g8f8 f5e7 d7f6 e2f3 c8e6 a2a4 a7a5 h2h4 e8d8 e3e4 f6d7 b2b3 d7e5 c1b2 f7f6 c2c4 d8d6 e7g6 h7g6 d1d6 f8e7 d6d2');
INSERT INTO evaluations VALUES(34221,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(34221,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(34221,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(34221,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(34221,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(34221,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(34221,36,'cp 172','f8c5 c1e3 d7d6 e1c1 g8f6 f2f3 e8g8 g2g4 f8e8 e3f4 d6d5 e4d5 f6d5 c3d5 c6d5 f1b5 c8d7 b5d7 b8d7 d4b5 a8c8 d1d5 d7b6 d5d3 b6c4 b5c3 c4e5 f4e5 e8e5 h1d1 f7f5 a2a3 c5b6 h2h3 e5e3 d3e3 b6e3 c1b1 f5g4');
INSERT INTO evaluations VALUES(35753,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35753,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(35753,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(35753,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(35753,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(35753,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35753,36,'cp 18','a7a5 f1e1 b8d7 f3d2 e6e5 d4e5 d7e5 c4d5 f6d5 c3d5 c6d5 d2f3 e5f3 g2f3 d5d4 d1d3 e7c5 d3b5 d8e7 c1f4 a8a7 e2e3 c8d7 b5d3 d4e3 f4e3 b7b6 g1g2 d7e6 a1c1 a7c7 e3c5 c7c5 d3e3 h7h6 c1c5 e7c5 e3c5');
INSERT INTO evaluations VALUES(37958,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(37958,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(37958,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(37958,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(37958,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(37958,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(37958,36,'cp 54','d2d4 c5d4 c3d4 g8f6 b1c3 e8g8 c4b3 b8c6 h2h3 b7b5 e1g1 e7e6 c3b5 f6e4 d4d5 c6e7 d5e6 c8e6 b3e6 f7e6 d1e2 d6d5 b5c3 g7c3 b2c3 a8c8 f3g5 e4g5 c1g5 d8d7 f1e1 c8c3 e2e6 d7e6 e1e6 e7c6 a1d1 d5d4 g2g4 c3c4 e6d6 c4c2 d1d2 c2d2 g5d2 f8c8');
INSERT INTO evaluations VALUES(37966,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(37966,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(37966,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(37966,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(37966,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(37966,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(37966,36,'cp 43','c5d4 c3d4 g8f6 b1c3 e8g8 c4b3 b8c6 h2h3 b7b5 e1g1 b5b4 c3d5 f6d5 b3d5 c8b7 f1e1 a8c8 a2a3 b4a3 b2a3 d8a5 a1b1 b7a8 c1e3 e7e6 d5b3 a5a3 b1a1 a3b4 a1a4 b4b5 b3c4 b5b2 d4d5 c6b4 c4b3 e6d5 e4d5 b4c2 b3c2');
INSERT INTO evaluations VALUES(38162,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(38162,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(38162,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(38162,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(38162,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(38162,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(38162,36,'cp 81','d8c7 f2f4 g8e7 f1d3 b8d7 g1e2 f7f6 e5f6 d7f6 c1e3 a7a6 e1g1 e8g8 e3d4 c8d7 e2g3 f8f7 b4b5 a6a5 d1e2 a5a4 h2h3 a4a3 a1a2 e7g6 e2e3 a8a4 f1f2 g6e7 a2a1 e7f5 g3f5 e6f5 e3e5 c7e5 f4e5');
INSERT INTO evaluations VALUES(33328,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(33328,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(33328,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(33328,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(33328,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(33328,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(33328,36,'cp 78','f2f4 b8d7 f1d3 g8e7 g1e2 f7f6 e5f6 d7f6 c1e3 a7a6 e1g1 e8g8 e3d4 c8d7 d4e5 c7b6 e2d4 a8c8 g1h1 e7g6 d4b3 d7b5 e5d4 b6c7 b3c5 b5d3 c2d3 c8e8 d1e1 c7c6 e1d2 f6d7 a1e1 d7c5 d4c5 f8f7 h1g1 e6e5 f4e5 f7f1 e1f1 g6e5');
INSERT INTO evaluations VALUES(38862,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(38862,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(38862,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(38862,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(38862,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(38862,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(38862,36,'cp 16','a2a3 b4c3 e2c3 b8c6 d4d5 e6d5 d1d5 c8e6 d5e4 g8f6 e4a4 d8d4 f1b5 d4a4 b5a4 e8c8 a4c6 b7c6 c1e3 e6f5 e1g1 a7a5 a1c1 h8e8 f1e1 f5g6 f2f3 f6d5 c3d5 d8d5 g1f2 c6c5 h2h4 a5a4 e1e2 f7f6 e3f4 d5d4 e2e8 g6e8 c1e1 e8g6 f2g3 g6c2 e1e7');
INSERT INTO evaluations VALUES(38865,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(38865,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(38865,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(38865,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(38865,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(38865,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(38865,36,'cp 20','b4c3 e2c3 b8c6 f1b5 g8e7 c1e3 e8g8 d1d2 e6e5 d4e5 d8d2 e3d2 c6e5 c3e4 c8f5 f2f3 a7a6 b5e2 e7c6 e1c1 c6d4 e2f1 f5e4 f3e4 a8d8 d2c3 e5c6 c3d4 c6d4 f1c4 b7b5 c4a2 d4e2 c1b1 e2f4 g2g3 f4e6 b1c1 g7g6 d1d8 f8d8 h1f1 d8d4 a2e6 f7e6');
INSERT INTO evaluations VALUES(38908,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(38908,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(38908,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(38908,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38908,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38908,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(38908,36,'cp 11','e2c3 b8c6 d4d5 e6d5 d1d5 c8e6 d5e4 g8f6 e4a4 d8d4 f1b5 d4a4 b5a4 e8c8 a4c6 b7c6 c1e3 h8e8 e1g1 a7a5 a1c1 f6d5 e3d4 g7g6 f2f3 c6c5 d4e5 e6f5 f1e1 d5c3 e5c3 e8e1 c1e1 a5a4 e1e7 d8d7 e7e5 f5c2 e5c5 c2b3');
INSERT INTO evaluations VALUES(38906,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(38906,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(38906,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(38906,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(38906,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(38906,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(38906,36,'cp 8','b8c6 d4d5 e6d5 d1d5 c8e6 d5e4 g8f6 e4a4 d8d4 f1b5 d4a4 b5a4 e8c8 a4c6 b7c6 c1e3 e6f5 e1g1 a7a5 a1c1 f6d5 c3d5 d8d5 h2h3 h8e8 f1e1 c8b7 f2f3 a5a4 g2g4 f5g6 g1f2 h7h5 f2g3 h5g4 f3g4 d5e5 g3f2 c6c5 h3h4 f7f6 h4h5 g6f7 e3f4 e5e1 c1e1');
INSERT INTO evaluations VALUES(39017,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(39017,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(39017,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(39017,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(39017,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(39017,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(39017,36,'cp 33','g8f6 b2b4 e8g8 f1e2 b7b6 e1g1 c8b7 b1c3 d7d6 d1c2 d8e7 c1a3 b8d7 f1d1 a7a6 a3b2 e6e5 a1b1 e5d4 f3d4 d7e5 b4b5 a6b5 c4b5 f8c8 b1c1 f6d7 d4c6 b7c6 b5c6 d7c5 c3d5 e7e6 f2f4 e5c6 b2g7 g8g7 e2f3');
INSERT INTO evaluations VALUES(42322,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(42322,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(42322,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(42322,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(42322,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(42322,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(42322,36,'cp 46','b2b4 b7b6 f1e2 e8g8 b1c3 d8e7 e1g1 c8b7 d1b3 d7d6 f3d2 f8c8 c1a3 b8d7 a1c1 a7a6 h2h3 c8b8 f1d1 c7c5 b4c5 b6c5 b3c2 b8c8 d2b3 c5d4 e3d4 a8b8 c2d3 d7b6 b3a5 e7c7 a5b7 b8b7 c4c5 d6c5 d4c5 b6d5');
INSERT INTO evaluations VALUES(42644,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(42644,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(42644,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(42644,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(42644,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(42644,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(42644,36,'cp -23','f3g3 d7d6 f1e2 b8c6 g1f3 h7h6 c1e3 f6h5 g3h4 d8h4 f3h4 h5f4 e3f4 e5f4 c2c3 g7g5 h4f5 d6d5 b1d2 d5e4 d3e4 c8e6 f5g7 e8e7 a2a4 a7a6 b2b4 c5d6 g7f5 e7f6 e1c1 c6e5 f5d6 c7d6 h2h3 h6h5 c1b2 a6a5 b4b5 g5g4 d2b3 f4f3 g2f3 g4f3 e2f1 h8g8');
INSERT INTO evaluations VALUES(43969,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(43969,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(43969,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(43969,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(43969,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(43969,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(43969,36,'cp -23','f8e7 e1g1 e8g8 a2a3 b8d7 h2h3 a7a6 d2d3 d7c5 d3d4 e5d4 f3d4 c8d7 d1f3 a6a5 e4e5 c5b3 c2b3 d6e5 d4c6 d7c6 f3c6 b5b4 f1d1 d8c8 c6c8 a8c8 a3b4 a5b4 c3a4 f6e4 c1e3 f7f5 a4b6 c8c2 b6c4 f5f4 e3b6 f4f3 c4e5 f3g2 e5d3 e7f6 g1g2');
INSERT INTO evaluations VALUES(45965,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(45965,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(45965,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(45965,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(45965,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(45965,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(45965,36,'cp -52','d6e5 f3e5 a7a6 a2a4 f8d6 d2d4 d8c7 e5g4 f6g4 d1g4 c5d4 g4d4 d6e5 d4h4 c8d7 e1g1 d7c6 c1e3 e8g8 a1d1 b8d7 f1e1 a8c8 e3d4 e5f6 d4f6 d7f6 d1d4 h7h6 c4f1 c7b6 d4b4 b6a5 b4d4 c8d8 e1d1 e6e5 d4d6 d8d6 d1d6 a5c5');
INSERT INTO evaluations VALUES(46111,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(46111,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(46111,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(46111,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(46111,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(46111,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(46111,36,'cp -45','f3e5 a7a6 a2a4 d8c7 d1e2 f8d6 f2f4 e8g8 e1g1 f6d7 e5d7 c8d7 d2d3 b8c6 c3e4 c6d4 e2f2 b7b5 c4b3 d7c6 c1e3 c6e4 d3e4 d4b3 c2b3 a8b8 a1c1 b5a4 b3a4 f8c8 c1c4 c7b7 e4e5 d6f8 f1c1 b7b2 f2b2 b8b2 e3c5');
INSERT INTO evaluations VALUES(46104,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(46104,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(46104,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(46104,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(46104,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(46104,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(46104,36,'cp -40','a7a6 a2a4 d8c7 d1e2 f8d6 f2f4 e8g8 e1g1 b7b6 b2b3 c8b7 c1b2 b8c6 e5c6 b7c6 e2f2 f6h5 c3e2 b6b5 a4b5 a6b5 a1a8 f8a8 c4d3 a8a2 b2e5 c5c4 b3c4 b5c4 e5d6 c7d6 e2c3 c4d3 c3a2 d3c2 f2b6 h7h6 a2b4 h5f4 b6c6 f4e2 g1f2 d6c6 b4c6 c2c1q f1c1 e2c1');
INSERT INTO evaluations VALUES(47014,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(47014,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47014,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47014,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47014,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47014,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(47014,36,'cp 0','e7e5 c2c4 d6d5 b1c3 d5c4 d3c4 e5d4 e3d4 f8d6 g1f3 b8c6 c1g5 e8g8 e1g1 h7h6 g5h4 c8g4 f1e1 g8h8 a2a3 g4f3 d1f3 c6d4 f3d3 g7g5 a1d1 c7c5 h4g3 d8c7 c3b5 d4b5 g3d6 c7d6 d3d6 b5d6 d1d6 h8g7 e1e7 a8e8 e7b7');
INSERT INTO evaluations VALUES(50923,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(50923,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(50923,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(50923,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(50923,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(50923,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(50923,36,'cp 77','d2d4 b8c6 f1c4 h7h6 d1a4 d8d7 a4c2 f8e7 e1g1 g8f6 f1e1 e8g8 h2h3 g8h8 d4d5 c6b8 c4e2 f6h7 e2d3 a6a5 c1e3 h7g5 f3h2 d7d8 c2e2 b8d7 b1d2 g5h7 h2f3 h7g5 a2a4 g5f3 d2f3');
INSERT INTO evaluations VALUES(50918,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(50918,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(50918,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(50918,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(50918,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(50918,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(50918,36,'cp 71','b8c6 f1c4 h7h6 d1a4 d8d7 a4c2 f8e7 e1g1 g8f6 f1e1 d7g4 c4d5 e8g8 d5c6 b7c6 d4e5 f6h5 e5d6 c7d6 f3d4 g4g6 e4e5 f7f5 e5f6 g6c2 d4c2 e7f6 b1a3 d6d5 c1e3 a8b8 c2b4 a6a5 b4c6 b8b2 c6d4 b2b6');
INSERT INTO evaluations VALUES(17798,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(17798,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(17798,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(17798,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(17798,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(17798,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(17798,36,'cp 62','f1c4 h7h6 e1g1 g7g5 c1e3 f8g7 d4e5 d6e5 d1d8 c6d8 b1d2 c8e6 f1d1 g8e7 a2a4 e6c4 d2c4 f7f6 h2h4 d8e6 h4h5 e8f7 g1f1 h8d8 a4a5 d8d1 a1d1 g7f8 f1e2 e7c6 b2b4 a8d8 d1d8 c6d8 f3d2 e6f4 e3f4 g5f4 d2b3 d8e6');
INSERT INTO evaluations VALUES(17795,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(17795,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(17795,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(17795,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(17795,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(17795,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(17795,36,'cp 71','h7h6 e1g1 g7g5 c1e3 f8g7 d4e5 d6e5 b1d2 d8f6 a2a4 g8e7 a4a5 e7g6 d1b3 e8g8 g2g3 f8d8 g1g2 a8b8 b3c2 g7f8 b2b4 b7b5 c4b3 c8e6 b3e6 f6e6 h2h3 d8d7 f1d1 b8d8 d2f1 d7d6 f3h2 d6d1 a1d1 d8d1 c2d1 e6c4');
INSERT INTO evaluations VALUES(51815,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(51815,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(51815,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(51815,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(51815,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(51815,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(51815,36,'cp 7','d7d5 b1c3 d5d4 c3b1 f8d6 d2d3 c6c5 a2a4 b8c6 b1a3 d6c7 c1d2 e6e5 g2g3 c8g4 f4f5 c7a5 c2c3 d8d7 a3c4 e8c8 e1f2 g4f3 d1f3 h7h5 f1e2 g8f6 h2h3 h5h4 g3g4 d7c7 f2g2 f6d7 g4g5 d7b6 c4b6 a7b6');
INSERT INTO evaluations VALUES(53625,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(53625,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(53625,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(53625,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(53625,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53625,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53625,36,'cp 23','b1c3 d5d4 c3b1 f8d6 g2g3 c6c5 d2d3 e6e5 b1a3 d6c7 f1g2 e5f4 c1f4 g8e7 e1g1 b8c6 e4e5 e7g6 a3c4 e8g8 f3g5 g6f4 g3f4 f7f6 e5f6 g7f6 g5e4 c6b4 e4c5 b4d5 d1f3 d5f4 c5b7 c8b7 f3b7 a8b8 b7e4 f6f5 e4f3 g8h8 g1h1 f4g6 f3h5');
INSERT INTO evaluations VALUES(53703,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(53703,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(53703,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(53703,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(53703,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(53703,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(53703,36,'cp -53','c6d5 d2d4 e5d4 d1d4 g8f6 d4a4 b8c6 c1e3 f8b4 g1f3 e8g8 f3d4 d8e7 f1g2 f6g4 d4c6 b7c6 e3c1 c8d7 e1g1 b4c3 b2c3 e7e2 c1f4 g4e5 f4e5 e2e5 c3c4 d7e6 c4d5 c6d5 f1e1 e5d6 a1d1 a8d8 a4a7 d5d4 d1d3 e6f5 d3d2 h7h6 e1d1 d4d3 a7e3');
INSERT INTO evaluations VALUES(57025,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57025,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(57025,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57025,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57025,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57025,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57025,36,'cp 0','g8f6 g1f3 c7c5 e2e3 d7d5 a2a3 b7b6 c2c4 c5d4 f3d4 b8d7 c4d5 f6d5 f1b5 c8b7 e1g1 a7a6 b5c6 b7c6 d4c6 d8c7 c6d4 f8e7 d4f5 e7f8');
INSERT INTO evaluations VALUES(57090,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(57090,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(57090,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(57090,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(57090,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(57090,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(57090,36,'cp 44','g8f6 b1c3 c7c6 g1e2 b8a6 g2g4 d8e7 f1g2 e7e5 d1d4 e5d4 e2d4 f6d5 c3e4 a6c5 e4g3 d5c7 e3e4 d7d5 e4d5 c7d5 g2d5 c6d5 f2f3 g7g6 c1e3 f8g7 e1c1 c5e6 g3e2 h7h5 e2c3 h5g4 f3g4 g7d4 e3d4 e6d4 d1d4 c8e6 c3d5 e8c8');
INSERT INTO evaluations VALUES(49478,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(49478,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(49478,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(49478,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(49478,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(49478,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(49478,36,'cp 65','g1e2 f8d6 b1c3 d6e5 e2g3 e5c3 b2c3 d7d6 c1b2 b8d7 a2a4 e8g8 a4a5 d7e5 f1e2 h7h5 c3c4 h5h4 g3f1 c7c6 f2f4 e4f3 g2f3 f8e8 d1d4 c8f5 d4h4 c6d5 h1g1 f5c2 h4g5 c2g6 f3f4 e5c4 e2c4 d5c4 f4f5 c4c3 b2c3 a8c8 f5g6 c8c3');
INSERT INTO evaluations VALUES(816,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(816,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(816,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(816,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(816,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(816,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(816,36,'cp 0','c5d4 c1h6 g7h6 c3d4 c8d7 g2g3 a8c8 f1h3 a7a5 e1g1 a5b4 a3b4 c6b4 b1a3 c8c3 a3b1');
INSERT INTO evaluations VALUES(824,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(824,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(824,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(824,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(824,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(824,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(824,36,'cp 0','b4b5 c6a5 c3d4 c8d7 b1c3 h6f5 a1b1 b6c7 c1d2 a5c4 f1d3 a8c8 b1b3 c4a5 b3b1');
INSERT INTO evaluations VALUES(3619,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(3619,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(3619,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(3619,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(3619,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(3619,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3619,36,'cp 10','f1e1 a7a6 d2f1 c8e6 c4e6 f7e6 f1g3 d8e7 c1e3 c5e3 e1e3 f6d7 h2h3 f8f7 d1b3 a8b8 a1f1 b7b5 b3c2 b8f8 g3e2 e7f6 f3h2 f6h4 e3g3 c6e7 d3d4 h4h5 b2b3 e7g6 c3c4 g6f4 e2f4 e5f4 g3d3 b5b4 h2f3 a6a5 c2e2');
INSERT INTO evaluations VALUES(3620,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(3620,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(3620,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(3620,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(3620,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3620,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3620,36,'cp 0','a7a6 d2f1 c8e6 c4e6 f7e6 c1e3 c5e3 e1e3 f6d7 a1c1 d8f6 f1g3 f8f7 d3d4 a8f8 c1c2 c6e7 g3h5 f6g6 h5g3');
INSERT INTO evaluations VALUES(3644,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(3644,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(3644,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(3644,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(3644,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(3644,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(3644,36,'cp -22','g7g5 h4g3 c6a5 h2h4 g5g4 f3d2 a5c4 d5f6 d8f6 d2c4 b7b5 c4e3 c5e3 f2e3 a7a5 h1f1 f6e7 e1f2 c8e6 a2a4 b5b4 f2g1 c7c5 d1e2 f7f6 f1f2 f8f7 g1h2 g8h7 d3d4 a8c8 c2c4 c8a8 a1f1 h6h5 d4e5 d6e5');
INSERT INTO evaluations VALUES(3523,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(3523,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(3523,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(3523,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(3523,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(3523,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(3523,36,'cp -19','h4g3 c6a5 h2h4 g5g4 f3d2 a5c4 d5f6 d8f6 d2c4 b7b5 c4e3 c5e3 f2e3 a7a5 h4h5 f6g5 d1d2 c8e6 e1c1 f7f6 a2a3 g8h7 g3h4 g5h5 h4f6 h5g6 f6h4 g4g3 d2e1 f8f2 d1d2 a8f8 d2f2 g3f2 h4f2 g6g2 f2h4 g2f3 b2b4 a5b4 a3b4 e6g4 c1b2 f8f7 h1g1 f3h3 e1g3 h3g3');
INSERT INTO evaluations VALUES(5329,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5329,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5329,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5329,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(5329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5329,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5329,36,'cp 0','g8f6 c1g5 h7h6 g5f6 g7f6 e2e3 f6g7 g1e2 a8b8 e1g1 d7d6 d1d2 a7a6 b2b3 e8g8 a1c1 d8c7 c3d5 c7d8 d5c3');
INSERT INTO evaluations VALUES(8021,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8021,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8021,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8021,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8021,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8021,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8021,36,'cp 0','a1b1 e8g8 a2a3 a7a6 b2b4 b7b5 c4b5 c5b4 a3b4 a6b5 g1f3 d7d6 e1g1 c8e6 f3g5 e6d7 c3b5 a8b8 b5a3 c6b4 c1d2 b4c6 d2c3 b8b1 d1b1 d8a8 b1a1 f8b8 g5f3 d7e6 f3g5');
INSERT INTO evaluations VALUES(5355,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(5355,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(5355,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(5355,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(5355,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(5355,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(5355,36,'cp 44','c1e3 d7d6 d3d4 c5d4 c3d4 c8g4 b1c3 g8f6 h2h3 g4f3 d1f3 e8g8 a1d1 d8c7 f1d3 e7e5 d4e5 d6e5 e1g1 a8d8 c3b5 c7e7 b5a7 c6d4 e3d4 e5d4 a7b5 d8e8 d1c1 f6d7 f1d1 d7e5 f3e2 e5d3 e2d3 e7e4 d3e4 e8e4');
INSERT INTO evaluations VALUES(6318,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(6318,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(6318,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(6318,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(6318,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(6318,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(6318,36,'cp 95','a2a4 a7a6 d2d4 e5d4 c3d4 d7d6 a4a5 b6a7 e1g1 g8f6 d4d5 c6e5 f3e5 d6e5 b1d2 e8g8 c1b2 f8e8 d1c2 f6h5 g1h1 h5f4 g2g3 f4g6 f2f4 c8h3 f1f3 h3g4 f3b3 e5f4 g3f4 g6f4 b3g3 d8h4 c2c3 f7f6 a1f1');
INSERT INTO evaluations VALUES(8620,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(8620,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(8620,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(8620,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(8620,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(8620,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8620,36,'cp 60','d4d5 c6b8 b1c3 b8d7 d1e2 e8g8 a2a4 a7a5 c1e3 f6g4 e3d2 d7b6 c4b3 c7c6 a1d1 c6d5 b3d5 b6d5 c3d5 c8e6 d5e7 d8e7 h2h3 g4f6 d2g5 a8d8 b2b3 h7h6 g5f6 e7f6 d1d3 f6e7 f1e1 f7f5 e4f5 e6f5 d3d5 e7c7 c2c4 f5e6 d5d3 f8f4 f3d2');
INSERT INTO evaluations VALUES(10147,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10147,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10147,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10147,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10147,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10147,36,'cp 0','e8g8 g1f3 f6g4 d1e2 c5f2 e1f1 c6d4 f3d4 f2d4 h2h3 g4f6 c3b5 d7d5 e4d5 d4b6 b5c3 e5f4 c1f4 b6d4 c3b5 d4b6');
INSERT INTO evaluations VALUES(10561,30,'cp -251',NULL);
INSERT INTO evaluations VALUES(10561,31,'cp -246',NULL);
INSERT INTO evaluations VALUES(10561,32,'cp -248',NULL);
INSERT INTO evaluations VALUES(10561,33,'cp -255',NULL);
INSERT INTO evaluations VALUES(10561,34,'cp -239',NULL);
INSERT INTO evaluations VALUES(10561,35,'cp -253',NULL);
INSERT INTO evaluations VALUES(10561,36,'cp -238','c6e5 f3e2 d8e7 e1g1 e5c4 e2c4 h6g4 g5f3 d7d6 b2b4 c5b6 b1d2 e8g8 a2a4 a7a5 b4b5 f8e8 c1b2 g4e5 c4b3 c8e6 c2c4 d4c3 b3c3 f7f6 f1e1 a8d8 c3c2 e7f7 b2d4 d8c8 h2h3 h7h6 f3e5 f6e5 d4b6 c7b6 c2b2 e8d8 a1c1 f7f6');
INSERT INTO evaluations VALUES(14154,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(14154,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(14154,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(14154,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(14154,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(14154,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(14154,36,'cp 191','d4d5 c6b8 c1g5 f7f6 g5e3 e7e5 a1c1 f8e7 g1f3 e8g8 f1e2 g8h8 e1g1 c8d7 d1b3 e7d6 c3b5 d7b5 e2b5 d8e7 f3h4 g7g6 g2g3 a7a6 b5d3 a6a5 g1g2 b8a6 e3b6 c7b6 d3a6 a8a6 f1e1 h8g7 h4f3 a5a4 b3d3 a6a8 h2h4 d6c5 e1h1 f6f5 f3g5 f5e4');
INSERT INTO evaluations VALUES(14404,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(14404,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(14404,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(14404,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(14404,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(14404,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(14404,36,'cp 135','c1h6 g7h6 c2c3 c8f5 f1d3 d8d7 g1e2 h8g8 d3f5 d7f5 e2g3 f5g6 e1g1 e8c8 b1d2 e7e6 f2f4 f7f6 e5f6 g6f6 d1e2 f8d6 a1e1 d8e8 d2f3 g8g4 g3h5 f6g6 h2h3 g4g2 e2g2 g6h5 f3e5 c6e7 g2f3 h5f5 g1h1 e8g8 f1g1');
INSERT INTO evaluations VALUES(14544,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(14544,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14544,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14544,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14544,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(14544,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(14544,36,'cp 0','g7g5 h4g3 f6e4 a2a3 h6h5 c4d5 e6d5 c3b5 h5h4 g3c7 d8e7 c7e5 f7f6 b5c7 e8f7 c7d5 e7d8 d1b3 c6a5 b3a2 f6e5 d5c3 c8e6 f3e5 f7e7 e5g6 e7d7 g6e5');
INSERT INTO evaluations VALUES(14715,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(14715,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(14715,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(14715,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(14715,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(14715,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(14715,36,'cp 56','h2h4 f8d6 b1d2 e4d2 d1d2 f7f6 e1c1 d6f4 e3f4 d8d6 h4h5 c8d7 d1e1 e8c8 g2g3 c8b8 f1d3 d8e8 e1e3 c6e7 d2c2 e7f5 e3e2 a7a6 c2b3 b8a7 a2a3 d6b6 b3b4 f5d6 h1e1 a6a5 b4c5 e8e7 c5b6 a7b6 c1d2 h8e8 d3h7');
INSERT INTO evaluations VALUES(16400,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(16400,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(16400,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(16400,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(16400,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(16400,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(16400,36,'cp -60','d7d5 e4d5 f6d5 g1f3 c8e6 d1e2 d8d7 b1d2 e8c8 e1c1 f7f6 h1e1 e6f7 d3d4 c6a5 c4b5 c7c6 b5d3 e5d4 f3d4 d5e3 f2e3 c8b8 d2b3 a5b3 a2b3 f8d6 d3c4 h8e8 c4f7 d7f7 h2h3 g7g6 c1b1 f7c7 g2g4 d6g3 e1h1 g3e5');
INSERT INTO evaluations VALUES(17745,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17745,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17745,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(17745,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(17745,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(17745,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(17745,36,'cp 13','b7c6 g1f3 c6c5 c2c3 c5d4 c3d4 c8d7 e1g1 c7c5 b1c3 c5d4 c3e2 g8e7 e2d4 e7g6 g2g3 d8b8 h2h4 h7h6 f1e1 f8b4 e1e2 e8g8 h4h5 g6e7 e2c2 b8b6 c1e3 f8c8 a1c1 b6b7 e3d2 a6a5 a2a3 b4d2 d1d2 c8c2 c1c2 e7c6 c2c1 c6d4');
INSERT INTO evaluations VALUES(18133,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(18133,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(18133,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(18133,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(18133,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(18133,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(18133,36,'cp -42','g8f6 f2f4 d7d5 e4e5 d5d4 c3e2 f6d5 e2g3 c6e7 g1f3 e7f5 d1e2 f8e7 e1g1 d8d7 g3f5 e6f5 c1d2 d7c6 h2h3 e8g8 g2g4 c8e6 e2g2 a8d8 a1e1 d5e3 d2e3 f5g4 a2e6 g4f3 f1f3 f7e6 e3d2 f8f5 f3g3 c6g2 g3g2 g7g6 g1h2');
INSERT INTO evaluations VALUES(11773,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(11773,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(11773,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(11773,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(11773,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(11773,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(11773,36,'cp -20','f2f4 d7d5 e4e5 d5d4 c3e2 f6d5 e2g3 c6e7 c1d2 e7f5 d1f3 c8b7 g3f5 e6f5 f3g3 g7g6 g1f3 d8b6 e1g1 a8c8 g3f2 f8e7 h2h3 d5c7 a3a4 e8g8 a4b5 a6b5 g1h2 c7d5 f1e1 c8a8 f2g3 g8g7 a2d5 b7d5 f3h4 b6e6 h4f3 a8a1 e1a1');
INSERT INTO evaluations VALUES(19907,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19907,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(19907,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19907,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(19907,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19907,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19907,36,'cp 0','c8d7 b5c6 b7c6 g1e2 g8f6 e1g1 h7h6 d4d3 d8c7 b2b3 e7e5 d3g3 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(26031,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26031,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26031,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26031,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26031,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26031,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26031,36,'cp 0','b5c6 b7c6 g1e2 g8f6 e1g1 e7e6 d4d3 h7h6 b2b3 d8c7 f1d1 a8d8 d3g3 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(26011,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26011,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26011,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26011,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26011,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26011,36,'cp 0','b7c6 g1e2 g8f6 e1g1 h7h6 d4d3 d8c7 h2h3 a8d8 b2b3 e7e6 d3g3 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(20719,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(20719,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(20719,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(20719,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(20719,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(20719,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(20719,36,'cp 74','f8e7 f1e2 g8f6 e1g1 e8g8 c1e3 c8d7 g1h1 c6d4 d1d4 a7a6 b2b4 b7b5 c4b5 a6b5 c3b5 f6g4 a2a4 e7f6 e4e5 g4e3 e5f6 e3c2 d4b2 c2a1 f6g7 f8e8 b5d6 d7a4 d6e8 a4e8 e2f3 a8a4 f1a1 a4a1 b2a1 d8d2 h1g1 d2b4 h2h3 b4e7 a1e5 e7a3 g1h2 a3b4 e5g3');
INSERT INTO evaluations VALUES(21070,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(21070,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(21070,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(21070,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(21070,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(21070,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(21070,36,'cp -29','e7e5 f1b5 f8d6 d2d3 a7a6 b5c6 b7c6 e3e4 g8e7 d3d4 f7f6 e1g1 e8g8 e4d5 c6d5 d4e5 d6e5 f3e5 f6e5 f1e1 d8d6 c1g5 c7c6 c3e4 d6c7 g5e7 c7e7 e4g3 a8b8 b2b3 e5e4 d1d2 f8f6 f2f3 e7c5 d2e3 c5e3 e1e3 e4f3 e3f3 f6f3 g2f3 g8f7 a1d1 c8h3');
INSERT INTO evaluations VALUES(23296,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(23296,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(23296,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(23296,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(23296,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(23296,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(23296,36,'cp 43','b1c3 g7g6 a1b1 a7a5 a2a3 f8g7 b2b4 a5b4 a3b4 d7d6 b4b5 c6e7 d2d3 g8f6 g1f3 h7h6 c1b2 e8g8 e1g1 c8e6 b1a1 d8d7 d1c2 g6g5 a1a8 f8a8 f1a1 a8a1 b2a1 c7c6 h2h4 g5h4 f3h4 f5f4 b5c6 b7c6 c3e4 f4g3 e4g3');
INSERT INTO evaluations VALUES(24484,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(24484,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(24484,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(24484,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(24484,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(24484,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(24484,36,'cp 48','c1e3 a8c8 c4b3 g7g6 f2f3 f8g7 d1d3 c6e5 d3d2 h7h5 e1g1 e5c4 b3c4 c8c4 a2a4 e8g8 a4a5 a7a6 f1d1 d8b8 d4e2 c4c8 g1h1 f8e8 h2h3 d7c6 e3h6 g7h8 e2d4 f6d7 h6e3 h8g7 b2b3 b8c7 e3h6');
INSERT INTO evaluations VALUES(25969,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25969,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25969,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25969,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25969,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(25969,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25969,36,'cp 12','e1g1 f8d6 c2c4 g8e7 b1c3 d5d4 c3e4 d8c7 e4d6 c7d6 f3d2 f7f5 a2a3 e8g8 b2b4 b7b6 f2f4 h7h6 d2f3 a7a6 b5a4 e7g6 g2g3 f8e8 c1d2 g8h7 f1e1 e8e1 f3e1 a8b8 e1g2 c6d8 a4d7 d6d7 d1h5 g6e7 a1f1 d8c6 h5f7');
INSERT INTO evaluations VALUES(26135,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(26135,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(26135,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(26135,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(26135,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(26135,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(26135,36,'cp 128','d4f5 d7c5 b1c3 g7g6 f5e3 g8f6 f2f3 c5e6 e3d5 f6d5 c3d5 f8e7 d1d3 e6c7 d5e7 d8e7 c1e3 b7b6 e1g1 e8g8 f1d1 f8d8 c2c4 c7e6 b2b3 a8c8 g1h1 h7h5 d3e2 e7f6 h2h3 c8c6 e2f2 e6f4 d1d2 d8c8 a1d1');
INSERT INTO evaluations VALUES(26136,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(26136,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(26136,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(26136,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(26136,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(26136,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(26136,36,'cp 117','d7c5 b1c3 g7g6 f5e3 g8f6 f2f3 c5e6 e3d5 f6d5 c3d5 f8e7 c1e3 e7g5 d1d2 g5e3 d5e3 d8c7 e1c1 e8c8 e3d5 c7d7 h2h4 h7h5 d2e3 c8b8 d1d2 f7f5 c1b1 h8f8 h1d1 f5e4 e3e4 d7f7 a2a4 f7f5 d5e7 f5e4 f3e4');
INSERT INTO evaluations VALUES(28018,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(28018,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(28018,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(28018,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(28018,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(28018,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(28018,36,'cp 129','h2h4 c7c5 d4d5 h7h5 g1h3 d7d6 a2a4 b8d7 f2f3 a7a6 c1g5 b7c8 e1g1 d7e5 c4e2 c8h3 g2h3 d8c8 g1h2 e5d7 g5e3 g8f6 f3f4 e8g8 e2c4 b6b5 a4b5 a6b5 a1a8 c8a8 c4b5 g8h7 b5d3 e7e6 d5e6 f7e6 d1e2 a8c6');
INSERT INTO evaluations VALUES(28136,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(28136,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(28136,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(28136,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(28136,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(28136,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(28136,36,'cp 20','e1g1 g8e7 f1e1 h7h6 c2c4 d5c4 d3c4 e8g8 h2h3 g4h5 b1c3 e7c6 c3e4 b8d7 e4d6 c7d6 d4d5 c6e5 c4e2 e5f3 e2f3 h5f3 d1f3 d8f6 f3g3 f8e8 c1d2 d7b6 a1d1 b6d5 g3d3 d5c7 d2c3 f6g5 d3d6 e8e1 c3e1 c7e6 d6b4 b7b6 e1c3 a8d8 d1d8 g5d8');
INSERT INTO evaluations VALUES(28135,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(28135,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(28135,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(28135,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28135,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(28135,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28135,36,'cp 26','g8e7 f1e1 h7h6 c2c4 d5c4 d3c4 g4h5 a2a3 b8d7 h2h3 e8g8 c4a2 e7c6 g2g4 h5g6 b1c3 f8e8 e1e8 d8e8 c1e3 d7f6 f3h4 g6h7 d1f3 e8d7 a2b1 a8d8 b1f5 h7f5 h4f5 d6e7 e3f4 f6d5 f4g3 d5c3 b2c3 e7f8 a3a4 b7b6 h3h4 c6e7 f5e3 c7c6 g1g2 e7d5 e3d5 d7d5 f3d5');
INSERT INTO evaluations VALUES(28415,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(28415,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(28415,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(28415,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(28415,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28415,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(28415,36,'cp 12','h2h3 g4h5 f1e1 h7h6 c2c4 d5c4 d3c4 e8g8 b1c3 e7c6 g2g4 h5g6 g4g5 d8d7 g1g2 h6g5 c1g5 g6h5 c3e4 d6b4 e1e3 c6d4 f3e5 h5d1 e5d7 d1f3 e3f3 b8d7 f3d3 c7c5 a2a3 d7e5 a3b4 e5c4 e4c5 d4c6 b4b5 c6e5 d3d4 b7b6 c5d7 e5d7 d4d7 c4b2 d7c7 b2d3 a1d1 f7f6 d1d3');
INSERT INTO evaluations VALUES(29653,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(29653,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(29653,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(29653,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(29653,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(29653,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(29653,36,'cp -36','e7e6 h2h3 g4f5 c2c4 c7c6 b1c3 f8b4 c1d2 b4c3 d2c3 f6e4 g2g4 f5g6 f1d3 b8d7 e1g1 e4c3 b2c3 e8g8 a1b1 d8c7 b1b2 d7f6 g4g5 f6e4 f3e5 e4c3 d1b3 g6d3 e5d3 c3e4 b3b7 c7b7 b2b7');
INSERT INTO evaluations VALUES(30170,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(30170,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(30170,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(30170,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(30170,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(30170,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(30170,36,'cp -30','h2h3 g4f5 c2c4 c7c5 c4d5 e6d5 g2g4 f5b1 a1b1 b8c6 f1g2 f8d6 f3e5 h7h6 c1d2 e8g8 d2c3 d8b6 d1b3 b6b3 a2b3 c5d4 e3d4 f6e4 e1g1 f7f6 e5d3 f6f5 b3b4 c6e7 b1a1 f8c8 f1c1 c8c7 b4b5 e7g6 g4f5 g6f4 d3f4 d6f4');
INSERT INTO evaluations VALUES(31287,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(31287,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(31287,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(31287,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(31287,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31287,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31287,36,'cp 16','e7e6 f1e2 b8c6 e1g1 c5d4 c3d4 g8e7 b1c3 h7h6 a2a3 e7c8 c1e3 c8b6 e2d3 f5d3 d1d3 f8e7 f3d2 e8g8 f2f4 d8d7 h2h3 f7f6 c3e2 a8c8 b2b3 a7a5 e5f6 e7f6 d2f3 a5a4 f3e5 f6e5 d4e5 d5d4 e2d4 c6d4 d3d4 b6d5 d4a4');
INSERT INTO evaluations VALUES(31989,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(31989,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(31989,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(31989,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(31989,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(31989,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(31989,36,'cp 61','e7e6 h2h3 g4f3 d1f3 b8c6 c4a6 d8c8 f3c6 b7c6 a6c8 a8c8 c3e4 g8f6 d2d3 c8b8 b2b3 f8b4 c1d2 h7h5 h3h4 e8e7 e4g5 b4a3 e1e2 f6d5 g5e4 f7f6 g2g3 a3d6 e2d1 c6c5 c2c3 a7a5 d1c2 e7d7 a2a4 b8a8 a1e1');
INSERT INTO evaluations VALUES(32377,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(32377,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(32377,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(32377,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(32377,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(32377,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(32377,36,'cp 61','h2h3 g4f3 d1f3 b8c6 c4a6 d8c8 f3c6 b7c6 a6c8 a8c8 c3e4 h7h5 d2d3 c8b8 c1e3 a7a5 e1c1 h5h4 e3f4 e8d7 h1e1 f8e7 f4e5 f7f6 e5c3 h8h5 a2a4 g8h6 b2b3 h5f5 d3d4 f5d5 c1b2 h6f5 d1d3 b8d8');
INSERT INTO evaluations VALUES(34489,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(34489,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(34489,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(34489,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(34489,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(34489,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(34489,36,'cp 35','b1c3 d5d8 g1f3 f8e7 f1b5 c8d7 b5d7 b8d7 d4d5 e6d5 d1d5 d7c5 e1g1 g8f6 d5e5 e8g8 f1d1 c5d7 e5g3 d8b6 a1b1 f8e8 c1e3 d7c5 f3d4 c5e4 d4f5 e4g3 f5e7 e8e7 e3b6 a7b6 h2g3 a8c8 d1d6 e7e6 d6d4 g7g6 b1d1 e6c6 a2a3');
INSERT INTO evaluations VALUES(34487,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(34487,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(34487,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(34487,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(34487,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(34487,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(34487,36,'cp 34','d5d6 g2g3 g8f6 f1g2 f8e7 g1e2 e8g8 e1g1 f8d8 a2a3 b8c6 c1f4 d6d7 b2b4 h7h6 a1c1 e7d6 d4d5 e6d5 c3d5 f6d5 d1d5 d7e7 d5f3 c6e5 f4e5 d6e5 e2f4 e7f6 f1d1 d8d1 c1d1 a8b8 f4h5 f6f3 d1d8 g8h7 g2f3 c8e6 d8b8 e5b8');
INSERT INTO evaluations VALUES(34777,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(34777,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(34777,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(34777,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(34777,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(34777,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(34777,36,'cp 48','b1c3 g8e7 d2d3 c8e6 c1e3 b8c6 e1c1 f6f3 g1f3 g7g6 h2h4 h7h5 c4e6 f7e6 c3e2 e8d7 a2a4 a7a5 c2c3 f8g7 c1c2 a8f8 d1f1 b7b6 f3g5 g7h6 f2f4 e5f4 e3f4 h6g5 f4g5 c6e5 e2f4 h8g8 d3d4 e5g4 f4d3');
INSERT INTO evaluations VALUES(35258,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(35258,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(35258,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(35258,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(35258,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(35258,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(35258,36,'cp 167','d2d4 d7d5 b1c3 d5e4 c3e4 g8f6 f1d3 b8d7 c2c3 f6e4 d3e4 d7f6 e4c2 c7c5 c1f4 c5d4 f3d4 e7c5 d1d3 f8e7 e1g1 e8g8 f1e1 c5h5 f4e5 f6d7 e5g3 d7f6 h2h3 f6d5 a2a3 f8d8 a1d1 c8d7 g3e5 e7f6 d3g3 f6e5 e1e5');
INSERT INTO evaluations VALUES(36696,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(36696,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(36696,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(36696,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(36696,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(36696,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(36696,36,'cp -20','c7c5 b1c3 b8c6 a2a4 b7b6 c3d5 f6d5 e4d5 c6b4 f3e1 c8b7 c2c3 b4a6 e1c2 a6c7 c2e3 a8b8 a1b1 e7e6 d5e6 f7e6 d3d4 b7g2 e3g2 d8d7 c1e3 c7d5 g2f4 d5e3 f2e3 g7h6 b2b4 c5d4 c3d4 h6f4 f1f4 f8f4');
INSERT INTO evaluations VALUES(35967,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(35967,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(35967,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(35967,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(35967,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(35967,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(35967,36,'cp -18','a2a4 b8c6 b1c3 a8b8 c3d5 f6d5 e4d5 c6b4 f3e1 e7e6 d5e6 c8e6 c2c3 b4d5 e1c2 b7b6 c2e3 d5e7 f1e1 h7h6 h2h4 d8d7 a4a5 b6b5 d3d4 c5d4 c3d4 f8e8 d4d5 e6h3 a5a6 h3g2 g1g2 b5b4 a1b1 h6h5 c1d2 e7f5 e3f5 d7f5');
INSERT INTO evaluations VALUES(37287,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(37287,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(37287,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(37287,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(37287,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(37287,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(37287,36,'cp 48','f1e2 b7b6 f3g5 c8b7 e1g1 c7c5 e2f3 b7f3 d1f3 b8a6 f3e2 a6b4 b2b3 b4c6 f1d1 h7h6 g5f3 c5d4 f3d4 c6d4 d1d4 e7c5 d4d2 d8e8 c1b2 a8d8 a1d1 d7d6 a2a3 a7a5 c3a4 e8e7 b2f6 e7f6 d2d3 f6e7 d1d2 c5a3 a4b6');
INSERT INTO evaluations VALUES(37427,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(37427,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(37427,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(37427,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(37427,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(37427,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(37427,36,'cp 9','e2e3 d7d6 h2h3 b7b6 f1c4 c8b7 e1g1 b8d7 a2a4 a7a6 f4h2 f8e8 b1d2 e7e5 d4e5 d7e5 f3e5 d6e5 d1c2 d8e7 f1d1 a6a5 d1e1 e8d8 e3e4 e7c5 b2b3 c5e7 d2f3 f6e8 a1d1 d8d1 e1d1 e8d6 c4d5 b7d5 d1d5');
INSERT INTO evaluations VALUES(38064,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(38064,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(38064,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(38064,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(38064,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(38064,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(38064,36,'cp 209','g1f3 g8e7 e1g1 b8d7 d1c2 e8g8 c1g5 h7h6 g5h4 b7b5 c4d3 e6e5 c3e2 b4d6 d4e5 d7e5 f3e5 d6e5 a1d1 c8g4 f2f4 g7g5 f4e5 g5h4 h2h3 g4e2 d3e2 d8b6 g1h1 b6e3 e2g4 a8d8 d1d6 e7g6 d6d8 f8d8 e5e6 g6e5 e6f7 g8f8 g4h5 d8d3 b2b4');
INSERT INTO evaluations VALUES(38063,30,'cp 194',NULL);
INSERT INTO evaluations VALUES(38063,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(38063,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(38063,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(38063,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(38063,35,'cp 207',NULL);
INSERT INTO evaluations VALUES(38063,36,'cp 182','g8e7 c1g5 b7b5 c4d3 h7h6 g5e3 b8d7 e1g1 e8g8 c3e2 c8b7 a2a3 b4d6 a1c1 a8c8 e2g3 d7b6 f1e1 e7g6 d1e2 d6b8 e2d2 a7a6 b2b4 a6a5 b4a5 b6a4 c1c2 d8e7 g3h5 c6c5');
INSERT INTO evaluations VALUES(38332,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(38332,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(38332,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(38332,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(38332,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(38332,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(38332,36,'cp 42','c7c6 b5d3 g8f6 h2h3 b8d7 e1g1 e8g8 a2a4 e7e5 d4e5 d7e5 f3e5 d6e5 c1e3 d8e7 b1d2 f8d8 d1f3 c8e6 f1d1 b7b6 f3e2 f6h5 e2e1 h7h6 a4a5 h5f4 d3f1 c6c5 g1h2 a8b8 g2g3 f4h5 d2b1 h5f6 d1d8 e7d8 a5b6 b8b6 e3c1 d8d4');
INSERT INTO evaluations VALUES(37931,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(37931,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(37931,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(37931,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(37931,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(37931,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(37931,36,'cp 41','b5d3 b8d7 e1g1 g8f6 f1e1 e7e5 d4e5 d7e5 f3e5 d6e5 a2a4 e8g8 h2h3 f6h5 d3f1 d8f6 b1d2 f8d8 g2g3 c8e6 d1e2 g7h6 d2f3 h6c1 a1c1 f6e7 e2e3 f7f6 a4a5 h5g7 e3c3 e6f7 a5a6 g7e6 a6b7 e7b7 f1c4 a7a5 c1d1 b7b4');
INSERT INTO evaluations VALUES(39874,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39874,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39874,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(39874,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(39874,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39874,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(39874,36,'cp 14','e5e4 f3h4 d7d5 c4d5 c6d5 d2d3 f6g4 e1g1 g7g5 d3d4 c5e7 h2h3 g4f2 f1f2 g5h4 g3g4 c8e6 c1f4 h7h5 d1b3 b8c6 g2e4 d5e4 d4d5 d8b6 b3b6 a7b6 d5e6 f7e6 f2g2 h5g4 g2g4 e7f6 f4e3 f6c3 b2c3 a8a3 g4e4 e6e5 a1b1 a3c3 b1b6 h8h7 b6b1 h7e7 g1f2 c3c2 e4h4 e7f7 f2e1 c2a2');
INSERT INTO evaluations VALUES(41372,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(41372,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41372,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(41372,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(41372,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(41372,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(41372,36,'cp 3','e8g8 f1d3 c7c5 e1g1 c5d4 c3d4 e7e5 d4e5 d6e5 d2c4 b8c6 c4e5 c6e5 f3e5 f8e8 e5c4 c8e6 c4e3 a8c8 d1e2 d8d4 f1d1 e8d8 e3c2 d4b6 c2e1 f6d7 c1e3 b6b4 h2h3 a7a6 a1c1 c8c1 d1c1 b4b2');
INSERT INTO evaluations VALUES(36670,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(36670,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(36670,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(36670,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36670,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(36670,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(36670,36,'cp 29','f1d3 c7c5 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 a2a4 b8a6 d2c4 b7b6 f1e1 a6c5 c1e3 c8a6 e3c5 b6c5 c4a3 a6d3 d1d3 g7h6 f3d2 h6d2 d3d2 a8b8 a3c4 b8b4 b2b3 f6h5 g1h1 d8c7 d2c2 f7f5 e4f5 g6f5 a1d1 h5f4 g2g3 f4g6 h2h4 f5f4 h1g2 f4f3 g2g1 c7f7');
INSERT INTO evaluations VALUES(36672,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(36672,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36672,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36672,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(36672,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36672,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(36672,36,'cp 15','c7c5 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 a2a4 b8a6 d3a6 a8a6 f1e1 f6d7 d2c4 d7c5 c1g5 d8c7 d1c2 f7f5 f3d2 f5e4 d2e4 h7h6 e4c5 c7c5 g5e3 c5d5 c4b6 d5f7 b6c8 a6c6 c2d2 f8c8 d2a5 e5e4 a5b5 c6c2 e1e2 d6d5 e2c2 c8c2 a1d1');
INSERT INTO evaluations VALUES(41957,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(41957,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(41957,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(41957,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(41957,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(41957,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(41957,36,'cp 163','e1g1 e4c3 b2c3 b4e7 e3e4 e8g8 c1f4 b7b6 d1d3 a7a5 a1d1 c8a6 c4a6 b8a6 f1e1 a5a4 d3e2 d8e8 h2h4 b6b5 d4d5 e7c5 c3c4 b5b4 e4e5 h7h6 f3d2 f7f6 e5f6 f8f6 g2g3 f6f4 g3f4 e8g6 g1h1 e6d5 c4d5 a8f8 e2e6 g6f7 d2c4 c5f2 e6f7 f8f7 e1e5 f2h4');
INSERT INTO evaluations VALUES(43611,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(43611,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(43611,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(43611,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(43611,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43611,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(43611,36,'cp 12','f8e7 h2h3 e8g8 f1e1 c7c5 a2a4 b8c6 c1d2 a8b8 f3h2 f6h7 c3d5 b7b5 d5e7 c6e7 a4b5 a6b5 c4a2 h7f6 f2f4 e5f4 d2f4 c8e6 a2e6 f7e6 f4g3 e7g6 h2g4 g8h7 g4e3 d8c7 e1f1 b8a8 d1e2 c7d7 e2f2 a8a1 f1a1 g6e5');
INSERT INTO evaluations VALUES(39769,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(39769,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(39769,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(39769,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(39769,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(39769,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(39769,36,'cp 14','c1e3 e8g8 c3d5 f6d5 c4d5 b8d7 a2a4 f8e8 a4a5 d7f6 d5c4 e7f8 f1e1 c7c6 h2h3 d6d5 e3b6 d8d6 c4b3 f6d7 b6e3 d5e4 d3e4 d7c5 e3c5 d6c5 e1e3 c8e6 b3e6 e8e6 d1d7 e6e7 d7d3 e7e6 b2b3 a8e8 d3d7 e6e7');
INSERT INTO evaluations VALUES(44695,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(44695,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(44695,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(44695,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(44695,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(44695,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(44695,36,'cp 48','g1f3 b8c6 b1c3 e7f5 f1b5 c8d7 e1g1 d8b6 b5a4 f8e7 c3e2 a8c8 a2a3 c6a5 a4c2 d7b5 f1e1 b5e2 e1e2 e8g8 b2b4 a5c6 c2f5 e6f5 d1d3 a7a5 e2b2 a5b4 a3b4 g7g6 c1h6 f8e8 h2h4 c6d8 a1a5 d8e6 d3b5 b6b5');
INSERT INTO evaluations VALUES(44694,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(44694,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(44694,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(44694,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(44694,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(44694,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(44694,36,'cp 30','b8c6 f1b5 c8d7 b1c3 d8b6 e1g1 e7f5 b5a4 f8e7 c3e2 e8g8 a4c2 a8c8 a2a3 f5h4 f3h4 e7h4 c1e3 c6a5 e2f4 f7f5 e5f6 h4f6 d1d3 g7g6 f4g6 f8f7 g6e5 f6e5 d4e5 b6b2 f1c1 b2e5 e3a7 c8c3 a7d4 c3d3 d4e5 d3d2');
INSERT INTO evaluations VALUES(51518,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(51518,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(51518,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(51518,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(51518,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(51518,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51518,36,'cp 9','e2e4 c7c5 g1e2 c8b7 c2c4 b8c6 b1c3 g8f6 g2g3 g7g6 f1g2 f8g7 e1g1 e8g8 b2b3 e7e6 f2f4 d7d6 c1b2 a8c8 d1d2 a7a6 c3d1 b6b5 d1e3 d8a5 d2a5 c6a5 g3g4 b5c4 b3c4 f8d8 g4g5 f6g4 b2g7 g4e3 g7f6 e3f1 f6d8');
INSERT INTO evaluations VALUES(52925,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52925,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52925,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52925,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(52925,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52925,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52925,36,'cp 0','d3e4 g8f6 b1c3 b8d7 d2d4 f6e4 c3e4 d7f6 e4c5 e7e5 f3e5 f8c5 d4c5 d8a5 c1d2 a5c5 e1g1 e8g8 e5d3 c5c4 b2b3 c4d4 d2e3 d4c3 e3d2');
INSERT INTO evaluations VALUES(52912,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52912,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52912,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52912,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52912,36,'cp 0','g8f6 b1c3 b8d7 d2d4 f6e4 c3e4 d7f6 e4c5 e7e5 f3e5 f8c5 d4c5 d8a5 c1d2 a5c5 e1g1 e8g8 e5d3 c5c4 d1c1 c8f5 b2b3 c4d4 d2e3 d4c3 e3d2');
INSERT INTO evaluations VALUES(45045,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45045,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45045,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45045,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45045,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45045,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45045,36,'cp 0','b1c3 b8d7 d2d4 f6e4 c3e4 d7f6 e4c5 e7e5 f3e5 f8c5 d4c5 d8a5 c1d2 a5c5 e1g1 e8g8 e5d3 c5c4 b2b3 c4h4 d1c1 h7h6 f2f3 c8f5 d2e1 h4g5 e1d2');
INSERT INTO evaluations VALUES(53517,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(53517,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(53517,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(53517,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(53517,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(53517,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(53517,36,'cp 103','d4b3 g8f6 c1f4 b8c6 b1c3 a7a6 d1e2 d8c7 e1c1 c6e5 c3d5 f6d5 c4d5 f8e7 h2h4 c8g4 f2f3 g4e6 b3d2 e8c8 d2c4 e7f6 g2g4 c8b8 c1b1 e5c4 d5c4 e6c4 e2c4 h6h5 d1d5 h5g4 f3g4 f6e5 f4e5 d6e5 g4g5 d8d5');
INSERT INTO evaluations VALUES(53963,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53963,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(53963,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53963,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(53963,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(53963,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(53963,36,'cp -5','d2d4 c5d4 e5d6 d8d6 d1d4 b8c6 d4d6 f8d6 c1e3 g8f6 b1c3 d6b4 e3d2 e8g8 f1d3 f8d8 e1c1 b4c5 h1f1 c8d7 c3e4 f6e4 d3e4 a8c8 g2g4 f7f6 h2h3 d7e8 d2e3 c5e3 f2e3 h7h6 f3d4 g8f7 d4c6 b7c6 b2b3 d8d1 f1d1 f7e7 c1b2 a7a5 a2a3 c6c5');
INSERT INTO evaluations VALUES(54821,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(54821,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(54821,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(54821,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(54821,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54821,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(54821,36,'cp 28','c2c4 g8f6 b1c3 f8g7 f1e2 e8g8 g1f3 e7e6 e1g1 c7c5 d4d5 e6d5 c4d5 c8g4 e3e4 a7a6 a2a4 f8e8 h2h3 g4f3 e2f3 b8d7 a4a5 b7b5 a5b6 d8b6 d1c2 a8b8 f1d1 c5c4 a1a4 e8c8 f3e2 a6a5 a4c4 c8c4 e2c4');
INSERT INTO evaluations VALUES(54813,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(54813,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(54813,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(54813,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(54813,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(54813,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(54813,36,'cp 22','e7e5 g1f3 d8e7 b1c3 f8g7 e3e4 g8f6 d4d5 e8g8 c1g5 h7h6 g5h4 e7e8 f3d2 f6h5 d1c2 a7a5 f2f3 b8a6 a2a3 f7f5 h4f2 b7b6 e1c1 a6c5 g2g3 h5f6 f2c5 b6c5 f1d3 a8b8 c1b1 c8d7 b1a1 f6h7 e4f5 g6f5 d1g1 h7g5 g3g4 f5g4 f3g4');
INSERT INTO evaluations VALUES(55497,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(55497,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(55497,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(55497,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(55497,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(55497,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(55497,36,'cp 83','b7b6 f1e2 a7a5 h2h3 g7g6 a2a3 h7h5 b1c3 b8d7 a1b1 f8h6 b2b4 e8f8 c1h6 g8h6 d1d2 f8g7 e1g1 h8f8 b1e1 h6g8 e1a1 g8f6 d2g5 f8g8 f1e1 d8e8 a1c1 g8f8 c3b5 e8d8 e2f1 c8a6 f3h4 a6b5 h4f5 g7g8 c4b5 a5b4');
INSERT INTO evaluations VALUES(55728,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(55728,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(55728,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(55728,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(55728,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(55728,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(55728,36,'cp 59','e2e4 b8d7 d1e2 a7a6 a2a4 g8f6 g2g4 h7h6 h2h4 c7c5 d4c5 d7c5 f1g2 d8b6 b2b3 g7g6 g4g5 h6g5 h4g5 h8h1 g2h1 f6h5 e2e3 c8d7 g1e2 d7c6 c1a3 c5d7 e3d2 e8c8 e2d4 e6e5 d4e2 d7c5 d2e3 c8b8 e1c1 b6c7 h1f3 f8e7 f3h5 g6h5');
INSERT INTO evaluations VALUES(56673,30,'cp 487',NULL);
INSERT INTO evaluations VALUES(56673,31,'cp 482',NULL);
INSERT INTO evaluations VALUES(56673,32,'cp 465',NULL);
INSERT INTO evaluations VALUES(56673,33,'cp 482',NULL);
INSERT INTO evaluations VALUES(56673,34,'cp 495',NULL);
INSERT INTO evaluations VALUES(56673,35,'cp 487',NULL);
INSERT INTO evaluations VALUES(56673,36,'cp 486','f3h4 g8e7 d4e5 f6e5 d1h5 e8d7 f2f4 c7c6 f4e5 d8a5 b1c3 a5e5 h5f7 e5f6 f7f6 g7f6 c1f4 b7b5 c4f7 d7c7 e1c1 c7b7 d1d6 b8d7 h1d1 d7b6 d6d8 b5b4 c3e2 c6c5 f7h5 h8g8 d8f8 g8f8 f4d6 f8g8 d6e7 c8g4 h5g4 g8g4 h4f5 g4g2 e2f4 g2h2 f4e6 b6c4 e6c5 b7b6 c5d7 b6b7 e7b4 a7a5 f5d6 c4d6 b4d6');
INSERT INTO evaluations VALUES(2689,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(2689,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(2689,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(2689,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(2689,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(2689,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(2689,36,'cp 169','c5e7 e4e5 f6e8 h2h3 d7d6 b1c3 c8f5 a4c6 b7c6 f1e1 a6a5 d1a4 a8a6 c1d2 h7h6 a1c1 d8a8 c3e2 a8b7 e2g3 f5h7 a4a3 a6a8 a3c3 c6c5 d4c5 d6c5 d2e3 b7b5 e3c5 b5c5 c3c5');
INSERT INTO evaluations VALUES(4705,30,'cp -300',NULL);
INSERT INTO evaluations VALUES(4705,31,'cp -277',NULL);
INSERT INTO evaluations VALUES(4705,32,'cp -276',NULL);
INSERT INTO evaluations VALUES(4705,33,'cp -264',NULL);
INSERT INTO evaluations VALUES(4705,34,'cp -264',NULL);
INSERT INTO evaluations VALUES(4705,35,'cp -270',NULL);
INSERT INTO evaluations VALUES(4705,36,'cp -265','h2h3 c8e6 b1c3 a7a5 c1e3 f7g8 c3e2 f6h5 a2a3 h7h6 d1d2 c5e3 f2e3 c6b8 b2b4 b7b6 c2c4 b8d7 g1h2 d8g5 e2c3 e6h3 g2h3 g5g3 h2h1 g3h3 h1g1 h5g3 d2g2 h3g2 g1g2 g3f1 a1f1 a5b4 a3b4 a8a3 f1c1 a3b3 c3d5 b3d3 d5c7');
INSERT INTO evaluations VALUES(4785,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4785,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(4785,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4785,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(4785,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4785,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4785,36,'cp 21','f3e5 d8d4 e5g4 c8g4 d1g4 d4f2 e1d1 g8f6 g4f3 f2h4 h1f1 e8g8 f3f5 c5e7 h2h3 f6h5 f1f3 a8d8 b1c3 g7g6 f5g4 h4g4 h3g4 h5g7 d1e2 g7e6 c1e3 e7g5 e3g5 e6g5 f3f2 d8d7 e2e3 f8e8 a2a4 c6c5 a4a5 e8e5 a1f1 c7c6 b2b3 d7e7 c3e2 g5e6 f2f6');
INSERT INTO evaluations VALUES(4777,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(4777,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(4777,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4777,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(4777,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4777,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4777,36,'cp 18','d8d4 e5g4 c8g4 d1g4 d4f2 e1d1 g8f6 g4f3 f2h4 h1f1 e8g8 f3f5 c5e7 h2h3 f6h5 f1f3 a8d8 b1c3 g7g6 f5g4 h4g4 h3g4 h5g7 d1e2 g7e6 c1e3 d8d7 a2a4 e7d6 c3b1 c6c5 b1d2 b7b5 f3f1 f8d8 a4b5 a6b5 a1a5 c7c6 g4g5 d6e5');
INSERT INTO evaluations VALUES(7498,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(7498,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(7498,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(7498,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(7498,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(7498,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(7498,36,'cp 58','a4c2 d7d5 d2d4 c8g4 e4d5 d8d5 c2b3 d5d6 d4e5 c6e5 d1d6 c7d6 f3e5 d6e5 f1e1 f6d7 c1e3 e8g8 b1d2 a8c8 h2h3 g4f5 d2f3 f8d8 a2a4 b5b4 c3b4 e7b4 e1d1 h7h6 a1c1 c8c1 d1c1 d8b8 b3c4 a6a5 g2g4 f5e4 f3d2 e4g6 c4b5 d7f6 d2c4 f6d5 c4e5 d5e3 f2e3');
INSERT INTO evaluations VALUES(7496,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(7496,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(7496,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(7496,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(7496,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(7496,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(7496,36,'cp 42','d7d5 d2d4 f6e4 f3e5 c6e5 d4e5 e8g8 b1d2 e4d2 d1d2 c8e6 d2d3 g7g6 c1h6 f8e8 d3d2 d8d7 d2f4 e7f8 h6f8 e8f8 a1d1 a8d8 f4g5 f8e8 f2f4 e6f5 c2f5 d7f5 g5f5 g6f5 d1d3 f7f6 d3g3 g8f7 g3h3 f7g8 f1f3 f6e5 f4e5');
INSERT INTO evaluations VALUES(7857,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(7857,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(7857,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(7857,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(7857,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(7857,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(7857,36,'cp -46','h7h6 g5f6 d8f6 b3d2 c6e7 f1c4 b4c3 b2c3 e8g8 d1f3 f6c6 a2a4 d7d6 a4a5 c8e6 c4e6 f7e6 f3e3 c6c7 e1g1 f8f7 f1d1 a8f8 f2f3 e7c6 d2b3 f7d7 c3c4 c6b4 e3d2 b4a6 a1a4 c7c6 d1a1 a6c5 b3c5 d6c5 d2f2 d7d4 f2g3 c6c7');
INSERT INTO evaluations VALUES(7811,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(7811,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(7811,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(7811,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(7811,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(7811,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(7811,36,'cp -36','g5f6 d8f6 b3d2 c6e7 f1c4 b4c3 b2c3 e8g8 d1f3 f6c6 e1g1 d7d6 a2a4 c8e6 c4e6 f7e6 f3e3 c6c7 a4a5 f8f7 f1d1 d6d5 d2f3 a8f8 d1d2 e7g6 e4d5 e6d5 d2d5 g6f4 d5c5 c7d8 c5e5 f4g2 e3d4 d8d4 f3d4 f7f2 a5a6');
INSERT INTO evaluations VALUES(9517,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(9517,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(9517,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(9517,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(9517,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(9517,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(9517,36,'cp -30','c8g4 h2h3 g4h5 d2d4 h5f3 d1f3 e5d4 c3e2 c6c5 c1f4 d6f4 e2f4 e8g8 e4e5 f6d7 a1e1 d8g5 g2g3 c7c6 e5e6 d7e5 f3e4 e5g6 h3h4 g5e7 e6f7 e7f7 f4e6 f8e8 f2f4 g6f8 f4f5 e8e7 e4g4 f8e6 e1e6 a8f8 f1f4 f8e8 e6e7 e8e7 f5f6 e7e6 f6g7 f7g6 g4f3 g6g7');
INSERT INTO evaluations VALUES(11146,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11146,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(11146,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(11146,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(11146,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(11146,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(11146,36,'cp 21','d2d4 c6d4 f3d4 c5d4 d1d4 e7e6 c1e3 f8e7 h2h3 e8g8 a2a4 d8c7 c4b3 c8d7 f2f4 a8d8 d4d3 d7c6 e3d4 c7a5 a1d1 a5b4 d3e3 f6d7 g1h2 h7h6 e3g3 e7f6 c3a2 b4a5 f4f5 f6d4 d1d4 a5e5 f5e6 e5d4 e6e7 d8e8 e7f8q e8f8 a2c3 d7c5 f1d1');
INSERT INTO evaluations VALUES(12945,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(12945,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12945,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12945,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(12945,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(12945,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(12945,36,'cp 0','b5a4 c8d7 g1f3 d8a5 d4c5 c6e5 a4d7 e5d7 b2b4 a5c7 b1a3 a7a6 c3c4 d5c4 a3c4 e7d5 a2a3 b7b6 c4b6 d7b6 c5b6 c7b6 c1b2 f8e7 b2g7 h8g8 g7d4 b6b5 g2g3 d5b4 a3b4 e7b4 f3d2 a8d8 d4e3 b4d2 e3d2 b5e5 e1f1 e5d5 d1a4 d8d7 f1g1 d5d2 a4a6 g8g5 h2h4 g5g3 f2g3 d2e3 g1g2 e3e4 g2g1');
INSERT INTO evaluations VALUES(14890,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(14890,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(14890,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14890,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(14890,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(14890,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(14890,36,'cp 46','e2e3 c6a5 f3e5 c7c6 e5c4 a5c4 f1c4 f8e7 g5h4 e8g8 e1g1 c8d7 d1e2 a8c8 f1d1 f6d5 h4g3 d5c3 b2c3 d8a5 a2a4 a5c3 g3e5 f8d8 a1b1 b7b5 a4b5 c6b5 c4b5 d7b5 b1b5 d8d7 d1b1 f7f6 e5g3 c3c4 e2f1 f6f5 b5b8 c4f1 g1f1 c8b8 b1b8 g8f7 b8a8');
INSERT INTO evaluations VALUES(14889,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(14889,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(14889,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14889,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(14889,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(14889,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14889,36,'cp 45','c6a5 f3e5 c7c6 e5c4 a5c4 f1c4 f8e7 e1g1 f6d5 g5e7 d8e7 a2a4 e7b4 d1d3 a7a5 f2f4 e8g8 e3e4 d5e3 d3e3 b4c4 h2h3 b7b6 a1d1 c4b4 f4f5 f7f6 d4d5 c6d5 d1d4 b4b2 f5e6 c8e6 e4d5 e6d7 g1h2 a8e8 e3g3 f8f7 d5d6 d7c6');
INSERT INTO evaluations VALUES(15550,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(15550,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(15550,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(15550,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(15550,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(15550,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(15550,36,'cp 37','e5d4 f3d4 f6e4 c3e4 d8e7 f2f3 d6d5 f1b5 c8d7 e1g1 d5e4 b5c6 b7c6 f1e1 e8c8 e1e4 e7b4 d4b3 d7f5 d1d8 c8d8 e4b4 f8b4 c2c3 b4d6 c1e3 a7a6 b3c5 h8e8 g1f2 d6c5 e3c5 d8c8 b2b3 a6a5 g2g4 f5e6 c5d4 g7g6 a1e1 e8d8 e1e5 d8d5 e5d5 c6d5 f2g3 e6d7 g3f4 d7b5 f4g5');
INSERT INTO evaluations VALUES(15648,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(15648,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(15648,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(15648,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(15648,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15648,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(15648,36,'cp 48','f3d4 f6e4 c3e4 d8e7 f1d3 c6d4 e1g1 d6d5 d3c4 d5e4 d1d4 c8e6 d4e4 e8c8 c1e3 e6c4 e4c4 e7e6 c4e6 f7e6 a1e1 d8d5 c2c4 d5d3 e3g5 h7h6 g5c1 c8d7 e1d1 d3d1 f1d1 d7c8 g1f1 f8c5 c1d2 g7g5 f2f3 h8f8 b2b4 c5e7 f1e2 a7a6 a2a4 b7b6 d2c3 f8f4 e2d3 e7d6 d1e1 c8d7 e1e4 h6h5 e4f4');
INSERT INTO evaluations VALUES(15639,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(15639,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(15639,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(15639,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(15639,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(15639,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(15639,36,'cp 25','f6e4 c3e4 d8e7 f2f3 d6d5 f1b5 c8d7 b5c6 b7c6 e1g1 d5e4 f1e1 e8c8 e1e4 e7b4 d4b3 b4b7 c1g5 f7f6 g5f4 d7f5 e4d4 d8d4 d1d4 b7b6 f4e3 b6d4 b3d4 f5d7 g1f2 f8d6 d4b3 d7f5 c2c3 a7a6 b3c5 a6a5 a2a4 d6c5 e3c5 h8d8 b2b4 d8d3 a1c1 d3d2 f2g3 d2a2 b4a5 a2a4 c5f8 g7g6 f8e7 a4a5 e7f6 f5e6 c1e1 c8d7 f6d4');
INSERT INTO evaluations VALUES(17749,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17749,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(17749,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17749,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17749,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17749,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(17749,36,'cp 0','b7c6 d2d4 f7f6 c1e3 g8e7 h2h4 a6a5 h4h5 f6f5 h5h6 g7h6 d4e5 f5e4 f3h4 h8g8 d1h5 e7g6 b1d2 d6d5 f2f4 e4f3 d2f3 d8d7 e3h6 f8h6 h5h6 d7g4 h4g6 g4g3 e1e2 g3g2 e2e3 h7g6 a1g1 g2b2 g1g6 b2c3 e3f4 g8f8 g6f6 f8g8 f6e6 c8e6 h6e6 e8f8 e6f6 f8e8 f6e6');
INSERT INTO evaluations VALUES(18012,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18012,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(18012,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(18012,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(18012,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(18012,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(18012,36,'cp 0','d2d4 f7f6 c1e3 g8e7 h2h4 e5d4 c3d4 d6d5 b1d2 d5e4 d2e4 e7d5 e1g1 c8f5 e4g3 f5g4 d1c2 d8d7 e3d2 f8e7 f1c1 e8g8 c2c6 f8b8 c6d7 g4d7 c1c2 e7d6 a1c1 b8e8 g1f1 d7b5 f1g1');
INSERT INTO evaluations VALUES(18005,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18005,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(18005,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(18005,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(18005,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18005,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(18005,36,'cp 0','f7f6 c1e3 g8e7 h2h4 e5d4 f3d4 c6c5 d4e2 d8d7 b1d2 c8b7 b2b4 e7g6 h4h5 g6e5 e2f4 a6a5 b4a5 f8e7 e1g1 b7c6 d1b3 c6a4 b3b7 a4c6');
INSERT INTO evaluations VALUES(18335,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(18335,31,'cp 244',NULL);
INSERT INTO evaluations VALUES(18335,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(18335,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(18335,34,'cp 243',NULL);
INSERT INTO evaluations VALUES(18335,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(18335,36,'cp 239','c4b5 f8c5 c2c3 d8f6 e1g1 g8e7 f1e1 a7a6 b5f1 d7d6 d2d4 e5d4 c1g5 f6g6 g5e7 c6e7 c3d4 c5b6 b1d2 c8g4 d1a4 g4d7 a4c2 e8g8 d2c4 d7g4 c4b6 c7b6 c2d3 e7c6 f3d2 d6d5 e4d5 c6b4 d3c4 b4c2 f1d3 g4f5 d3c2');
INSERT INTO evaluations VALUES(18334,30,'cp 264',NULL);
INSERT INTO evaluations VALUES(18334,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(18334,32,'cp 230',NULL);
INSERT INTO evaluations VALUES(18334,33,'cp 252',NULL);
INSERT INTO evaluations VALUES(18334,34,'cp 235',NULL);
INSERT INTO evaluations VALUES(18334,35,'cp 240',NULL);
INSERT INTO evaluations VALUES(18334,36,'cp 234','a8b8 b5a4 g8e7 e1g1 g7g6 f1e1 f8g7 c2c3 e8g8 d2d4 e5d4 c3d4 d7d6 h2h3 h7h6 b1c3 g6g5 c1e3 b8b2 c3e2 f7f5 e4f5 c8f5 a4b3 g8h7 d4d5 b2b3 d1b3 g7a1 e1a1 c6e5 f3e5 d6e5 e3c5 d8d7 c5e7 d7e7 a1c1 f8f7 e2g3 f5g6 b3e3 e7d6 e3a7 d6d5 c1c7');
INSERT INTO evaluations VALUES(21419,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(21419,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(21419,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(21419,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(21419,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(21419,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(21419,36,'cp 19','g1f3 g8f6 b1d2 c6b4 d3e2 c7c5 c2c3 b4c6 e1g1 a7a6 a2a3 b7b5 e2d3 c8b7 d1e2 c5c4 d3c2 c6a5 e3e4 d5e4 d2e4 a5b3 a1b1 b3c1 b1c1 f8e7 c1e1 e8g8 f3e5 d8c7 g2g3 f6e4 c2e4 e7d6 e5g4 a6a5 e4b7 c7b7 g4e3 d6c7 e2h5 f8d8 f2f4');
INSERT INTO evaluations VALUES(21641,30,'cp 290',NULL);
INSERT INTO evaluations VALUES(21641,31,'cp 289',NULL);
INSERT INTO evaluations VALUES(21641,32,'cp 295',NULL);
INSERT INTO evaluations VALUES(21641,33,'cp 294',NULL);
INSERT INTO evaluations VALUES(21641,34,'cp 289',NULL);
INSERT INTO evaluations VALUES(21641,35,'cp 286',NULL);
INSERT INTO evaluations VALUES(21641,36,'cp 299','e4d5 c6e5 d1e2 d8e7 b1c3 c8g4 f2f3 g4f5 c1f4 f7f6 e1c1 e8c8 g2g4 f5g6 g1h3 h7h5 f4g3 h5g4 f3g4 g6f7 c1b1 g8h6 h3f2 c8b8 h2h4 e7b4 g3e5 f6e5 f1h3 f7e8 g4g5 h6f7 h3g4 b4b6 b1a1 f8b4 c3e4 f7d6 h4h5 b6b5 e2b5 e8b5 a1b1 a7a6');
INSERT INTO evaluations VALUES(21564,30,'cp 307',NULL);
INSERT INTO evaluations VALUES(21564,31,'cp 304',NULL);
INSERT INTO evaluations VALUES(21564,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(21564,33,'cp 284',NULL);
INSERT INTO evaluations VALUES(21564,34,'cp 299',NULL);
INSERT INTO evaluations VALUES(21564,35,'cp 266',NULL);
INSERT INTO evaluations VALUES(21564,36,'cp 283','c6e5 d1e2 d8e7 b1c3 c8g4 f2f3 g4f5 c1f4 f7f6 e1c1 e8c8 g2g4 f5g6 f4g3 h7h5 g1h3 g6f7 h3f4 h5g4 f3g4 g7g5 f4h5 c8b8 h5f6 g8f6 e2e5 e7e5 g3e5 h8h6 f1e2 d8e8 e5g3 f8b4 e2b5 e8c8 h1f1 f7g8 h2h4 g5h4 g3f4 h6h7 d5d6 c7d6 d1d6 b4d6 f4d6 b8a8 f1f6');
INSERT INTO evaluations VALUES(22166,30,'cp 282',NULL);
INSERT INTO evaluations VALUES(22166,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(22166,32,'cp 298',NULL);
INSERT INTO evaluations VALUES(22166,33,'cp 281',NULL);
INSERT INTO evaluations VALUES(22166,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(22166,35,'cp 300',NULL);
INSERT INTO evaluations VALUES(22166,36,'cp 290','d1e2 d8e7 b1c3 c8g4 f2f3 g4f5 c1f4 f7f6 e1c1 e8c8 g2g4 f5g6 g1h3 h7h5 f4g3 h5g4 f3g4 g6f7 h3f4 e7b4 c1b1 g8e7 h2h4 g7g5 f4d3 e5d3 e2d3 f7d5 d3e2 b4b6 g3f2 b6c6 c3d5 e7d5 f1g2 f8g7 g2f3 d5f4 e2e3 c6b6 e3e4 d8d1 h1d1 b6b2 b1b2 f6f5 f2d4 f5e4 d4g7 h8e8 d1e1 e4f3 e1e8 c8d7 e8e1 g5h4 g7c3');
INSERT INTO evaluations VALUES(21818,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(21818,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21818,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(21818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21818,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(21818,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21818,36,'cp 0','d8d1 c3d1 g8f6 f1d3 c8e6 a2a3 h7h6 d3b5 e6d7 b5d3');
INSERT INTO evaluations VALUES(26232,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(26232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26232,36,'cp -3','g8f6 f1e1 g7g6 c2c4 f8g7 b1c3 e8g8 d4d3 f6d7 c1e3 d7c5 e3c5 d6c5 d3e3 e7e6 e4e5 c6f3 e3f3 d8b6 b2b3 a8d8 f3e3 d8d4 c3b5 d4d7 a1d1 f8d8 d1d7 d8d7 g2g3 a7a6 b5c3 b6a5 g1g2 a5b6');
INSERT INTO evaluations VALUES(24598,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24598,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(24598,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(24598,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24598,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24598,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24598,36,'cp 0','b1c3 e7e6 f1d1 f8e7 e4e5 d6e5 f3e5 d8d4 d1d4 a8c8 d4c4 f6d5 c3d5 e6d5 c4g4 e7f6 e5f3 d5d4 f3d4 h7h5 g4f4 f6e5 f4h4 e5f6');
INSERT INTO evaluations VALUES(29100,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(29100,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(29100,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(29100,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(29100,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29100,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(29100,36,'cp 14','g4f3 d1f3 b8c6 c4b5 a8c8 e1g1 e7e6 b5c6 c8c6 d2d3 f8e7 c1d2 e8g8 a2a4 a7a6 a4a5 h7h6 d2e3 f6d7 f3e2 e7f6 f2f4 b7b5 a5b6 c6b6 a1a2 f6d4 c3d1 d8c7 c2c3 d4f6 d1f2 f6e7 c3c4 f8b8 f1a1 e7f6');
INSERT INTO evaluations VALUES(29618,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(29618,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(29618,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(29618,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(29618,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(29618,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(29618,36,'cp 21','e2e3 e7e6 c2c4 f8e7 c4d5 f6d5 g5e7 d8e7 f1e2 e8g8 h2h3 g4h5 b1c3 d5c3 b2c3 c7c5 a2a4 b8d7 e1g1 a8c8 a4a5 c8c7 d1b3 f8c8 b3a3 d7f6 g2g4 h5g6 c3c4 c5d4 e3d4 e7a3 a1a3 g6e4 f3d2 h7h6 a3b3');
INSERT INTO evaluations VALUES(31131,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(31131,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(31131,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(31131,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(31131,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31131,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(31131,36,'cp 14','e7e6 e1g1 f8d6 b1d2 b8d7 f3e1 g4e2 d1e2 g8e7 d3d4 d6f4 c2c3 e8g8 e1d3 f4d2 c1d2 d5e4 e2e4 e7g6 f1e1 f8e8 d2f4 d7f6 e4f3 d8d5 f3d5 f6d5 f4d2 e8d8 a2a4 a7a5 g1f1 h7h6 g2g3 b7b6 f1e2 d8d7 f2f3 f7f6 h2h4 g6e7 h4h5');
INSERT INTO evaluations VALUES(31473,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(31473,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(31473,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(31473,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(31473,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(31473,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31473,36,'cp 31','e1g1 f8d6 b1d2 b8d7 d3d4 g8e7 f3e1 g4e2 d1e2 d6f4 c2c3 e8g8 e1d3 e7g6 f1e1 f4d2 c1d2 d5e4 e2e4 f8e8 d2f4 g6f4 e4f4 d8b6 e1e2 c6c5 d4c5 d7c5 d3c5 b6c5 a1d1 a8d8 e2d2 c5e7 h2h3 a7a5 a2a3 h7h6 d2d6 b7b5 d6d8 e8d8');
INSERT INTO evaluations VALUES(31617,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(31617,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(31617,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(31617,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(31617,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(31617,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(31617,36,'cp 75','f8e7 h2h3 g4e6 c2c4 e7f6 d1c2 g8e7 b1c3 b8c6 c1e3 c6d4 e3d4 c5d4 c3b5 e8g8 e1g1 e7c6 a1d1 d8b6 c2d2 c6e5 f3d4 e5c4 d2c2 f6d4 b5d4 b6b2 c2b3 b2b3 a2b3 c4a5 b3b4 a5c6 d4e6 f7e6 d1d6 c6b4 d6e6 a7a5 e6b6 f8b8 e2c4 g8h8');
INSERT INTO evaluations VALUES(32755,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(32755,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(32755,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(32755,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(32755,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(32755,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(32755,36,'cp -36','c2c3 g8e7 d2d4 e7f5 e1g1 b8c6 a2a4 a8c8 b1a3 c5d4 c3d4 f8b4 a3c2 e8g8 c2b4 c6b4 c1d2 b4c6 d2c3 c8c7 h2h4 h7h5 a4a5 a7a6 f3g5 g7g6 g2g3 f8c8 d1d3 c6e7 f1c1 c7c4 d3d2 d7b5 g5f3');
INSERT INTO evaluations VALUES(33097,30,'cp 338',NULL);
INSERT INTO evaluations VALUES(33097,31,'cp 354',NULL);
INSERT INTO evaluations VALUES(33097,32,'cp 316',NULL);
INSERT INTO evaluations VALUES(33097,33,'cp 322',NULL);
INSERT INTO evaluations VALUES(33097,34,'cp 328',NULL);
INSERT INTO evaluations VALUES(33097,35,'cp 313',NULL);
INSERT INTO evaluations VALUES(33097,36,'cp 333','d2d4 c5b6 a2a4 a7a6 a4a5 b6a7 c2c3 h7h6 c1e3 f6d8 b1d2 g8e7 f1d3 e8g8 e1g1 e7g6 f1e1 g8h8 d1b3 b8c6 d2f1 a8b8 f1g3 c8g4 f3d2 c6e7 h2h3 g4d7 b3c2 c7c5 d2f3 c5d4 c3d4');
INSERT INTO evaluations VALUES(33096,30,'cp 330',NULL);
INSERT INTO evaluations VALUES(33096,31,'cp 333',NULL);
INSERT INTO evaluations VALUES(33096,32,'cp 326',NULL);
INSERT INTO evaluations VALUES(33096,33,'cp 320',NULL);
INSERT INTO evaluations VALUES(33096,34,'cp 340',NULL);
INSERT INTO evaluations VALUES(33096,35,'cp 325',NULL);
INSERT INTO evaluations VALUES(33096,36,'cp 319','c5b6 a2a4 a7a6 c2c3 h7h6 c1e3 f6e7 b1d2 g8f6 f1d3 c7c5 e1g1 e8g8 d4d5 c8g4 h2h3 g4h5 a4a5 b6a7 d2c4 f6e4 d3e4 e7e4 c4d6 e4e7 e3f4 b8d7 f1e1 e7f6 f4g3 a7b8 d6e4 f6g6 g3b8 f8b8 e4g3 b7b5 g3h5 g6h5 d5d6');
INSERT INTO evaluations VALUES(33653,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(33653,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(33653,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(33653,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(33653,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(33653,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(33653,36,'cp 26','b8c6 d2d4 c8g4 d4d5 g4f3 g2f3 d6b4 d5c6 b4c4 c6b7 a8b8 d1e2 c4b4 e2a6 b4b7 a6a4 c7c6 c1f4 b8d8 e1g1 e7e6 f1d1 d8d1 a1d1 f8e7 c3e4 e8g8 e4f6 e7f6 b2b3 f8d8 d1d8 f6d8 a4d4 d8b6 d4d6 h7h6 g1g2 b7a6 d6c6 a6a2 c6a8 g8h7 a8e4 f7f5 e4c6 a2b1 f4c7 b6c7 c6c7 b1d1');
INSERT INTO evaluations VALUES(34481,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(34481,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(34481,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(34481,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(34481,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(34481,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(34481,36,'cp 103','b1c3 d5d8 e2e4 c5d4 f3d4 a7a6 c1e3 g8f6 f1c4 f8d6 e1g1 e8g8 a1c1 b7b5 c4b3 c8b7 b3e6 f7e6 d4e6 d8d7 e6f8 d6f8 f2f3 d7e6 a2a4 b5b4 c3e2 b8c6 d1c2 g8h8 f1d1 a8e8 e2f4 e6f7 a4a5 c6a5 c2c7 f7c7 c1c7');
INSERT INTO evaluations VALUES(34476,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(34476,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(34476,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(34476,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(34476,34,'cp 120',NULL);
INSERT INTO evaluations VALUES(34476,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(34476,36,'cp 126','d5d8 e2e4 c5d4 f3d4 a7a6 c1e3 g8f6 f1e2 f8d6 e1g1 e8g8 d1d3 d6c7 a1d1 e6e5 d4b3 d8d3 d1d3 b8c6 c3d5 c7d8 e2f3 a8b8 d3c3 c8e6 f1d1 f8e8 b3c5 c6d4 e3d4 e5d4 c3b3 f6d5 e4d5 e6f5');
INSERT INTO evaluations VALUES(34522,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(34522,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(34522,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(34522,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(34522,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(34522,35,'cp 209',NULL);
INSERT INTO evaluations VALUES(34522,36,'cp 199','c6b6 d2d4 g8f6 e1g1 c7c6 c1f4 b8d7 c3a4 b6d8 c2c4 e7e6 f1e1 f6h5 f4e3 h5f6 a4c3 f8e7 h2h3 d7f8 e2d3 f8g6 d3g6 h7g6 f3e5 e8g8 d1f3 d8b6 a1b1 f6d7 c3e4 b6a5 e3g5 d7e5 d4e5 e7g5 e4g5 a5a2 f3c3 a2a6 b2b3 a6b6 c3g3');
INSERT INTO evaluations VALUES(35332,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(35332,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(35332,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(35332,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(35332,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(35332,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(35332,36,'cp 119','b1c3 e7e5 c1e3 b8d7 h2h3 g8f6 g1f3 f8e7 f1d3 d7f8 a2a4 f8g6 a4a5 e8g8 e1g1 c8d7 d4d5 c6c5 g1h2 b7b6 a5a6 c7c8 f3g1 g6f4 d3c2 f6e8 g1e2 e7g5 e2f4 e5f4 e3d2 c8d8 h2g1 g7g6 c3e2 f7f5 e4f5 d7f5 c2f5 f8f5');
INSERT INTO evaluations VALUES(36668,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(36668,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36668,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(36668,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(36668,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(36668,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(36668,36,'cp 32','g1f3 b8c6 h2h3 e7e5 d4e5 c6e5 f3e5 d6e5 a2a4 f6h5 d1f3 c8e6 e1g1 a7a5 e3c5 f8e8 f1d1 d8f6 f3f6 h5f6 c3b5 a8c8 f2f3 g7f8 c5f2 f6d7 b2b3 f8c5 d3f1 c5f2 g1f2 g8f8 f1c4 d7c5 c4e6 e8e6 d1d2 f8e7 a1d1');
INSERT INTO evaluations VALUES(40348,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(40348,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(40348,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(40348,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(40348,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(40348,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(40348,36,'cp 65','f1e2 c5d4 f3d4 b8c6 c1e3 e8g8 e1g1 c6d4 e3d4 c8d7 d1c2 d8a5 f1d1 f8c8 c2d3 a7a6 d3e3 f6g4 e3g3 g7d4 d1d4 g4f6 a1d1 c8c7 f2f4 d7e6 e4e5 d6e5 f4e5 f6d7 g1h1 a8c8 b2b3 d7e5 c3d5 e5c6 d4d3');
INSERT INTO evaluations VALUES(41081,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41081,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(41081,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(41081,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(41081,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(41081,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41081,36,'cp 0','c7c6 d1c2 g7g6 e2e3 c8f5 f1d3 f5d3 c2d3 e8g8 e1g1 f8e8 d3c2 b8d7 h2h3 a7a5 f1d1 d8b6 f3e1 f6h5 f4h2 h5g7 e1d3 b6a6 a1c1 d7f6 h2e5 f6d7 e5h2');
INSERT INTO evaluations VALUES(40116,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(40116,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(40116,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40116,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(40116,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40116,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40116,36,'cp 9','d1c2 g7g6 e2e3 c8f5 f1d3 f5d3 c2d3 e8g8 d3c2 f8e8 e1g1 b8d7 h2h3 a7a5 a1d1 e7b4 f3e1 d7b6 e1d3 b6c4 d3c5 d8e7 b2b3 c4b6 f4g5 b4c5 d4c5 b6d7 c3a4 e7e4 c2c1 e4e6 f2f3 f6h5 g5h6 h5g7 f1e1');
INSERT INTO evaluations VALUES(43144,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(43144,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(43144,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(43144,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(43144,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(43144,35,'cp 146',NULL);
INSERT INTO evaluations VALUES(43144,36,'cp 142','e4e5 f6g4 d2d4 d7d6 h2h3 g4h6 f2f4 a7a6 g2g4 b7b5 c4d5 c7c6 d5g2 f7f6 g1f3 d6e5 d4e5 d8d1 c3d1 h6f7 e5f6 e7f6 e1g1 e8g8 a2a4 b5a4 a1a4 f6f5 g4f5 c8f5 d1e3 f5d7 c2c3 a8a7 a4a5 f8e8 e3g4 a7c7 g4e5 d7f5 f3d4 f5d7 e5d7');
INSERT INTO evaluations VALUES(44084,30,'cp -278',NULL);
INSERT INTO evaluations VALUES(44084,31,'cp -295',NULL);
INSERT INTO evaluations VALUES(44084,32,'cp -285',NULL);
INSERT INTO evaluations VALUES(44084,33,'cp -291',NULL);
INSERT INTO evaluations VALUES(44084,34,'cp -288',NULL);
INSERT INTO evaluations VALUES(44084,35,'cp -290',NULL);
INSERT INTO evaluations VALUES(44084,36,'cp -297','d7d5 c4d3 f8d6 b1c3 e8g8 e1g1 d5d4 c3e2 b8d7 e2g3 d7c5 f3e2 c8g4 f2f3 g4h3 g2h3 c5e6 e2e1 e6f4 d3c4 b7b5 c4e2 g7g6 g1h1 d8d7 d2d3 g8h8 a2a4 a7a6 h3h4 d7h3 e1f2 h3h4 e2d1 f4e6 g3e2 h4f2 f1f2 f6h5 f2f1 a6a5 a4b5 c6b5 f3f4 h5f4 e2f4 e6f4');
INSERT INTO evaluations VALUES(45808,30,'cp -303',NULL);
INSERT INTO evaluations VALUES(45808,31,'cp -286',NULL);
INSERT INTO evaluations VALUES(45808,32,'cp -286',NULL);
INSERT INTO evaluations VALUES(45808,33,'cp -280',NULL);
INSERT INTO evaluations VALUES(45808,34,'cp -279',NULL);
INSERT INTO evaluations VALUES(45808,35,'cp -279',NULL);
INSERT INTO evaluations VALUES(45808,36,'cp -275','c4d3 f8d6 b1c3 e8g8 e1g1 d5d4 c3e2 b8d7 e2g3 d7c5 f3e2 c8g4 f2f3 g4h3 g2h3 c5e6 e2e1 e6f4 d3e2 g7g6 d2d3 f4h3 g1h1 h3f4 f1g1 g8h8 e2f1 d8e7 c1d2 f8g8 a2a4 a8d8 g3e2 f6h5 e2g3 h5g7 g3e2 g7e6 e2f4 e6f4 a4a5 g8g7 e1f2 d8g8 d2f4 e5f4 f1h3 d6e5');
INSERT INTO evaluations VALUES(44086,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(44086,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(44086,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(44086,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(44086,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(44086,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(44086,36,'cp -84','d7d5 e4d5 c6d5 c4b5 b8c6 d2d4 e5e4 g1e2 f8e7 e1g1 e8g8 h2h3 h7h6 f2f3 a7a6 b5a4 b7b5 a4b3 c6a5 c1f4 a5b3 c2b3 f8e8 f4e5 e4f3 f1f3 f6e4 c3e4 d5e4 f3g3 e7g5 e5g7 d8b6 b3b4 c8b7 h3h4 g5h4 d1b3 h4g3 b3g3 b6g6');
INSERT INTO evaluations VALUES(45810,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(45810,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(45810,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(45810,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(45810,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(45810,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(45810,36,'cp -69','e4d5 c6d5 c4b5 b8c6 d2d4 e5e4 g1e2 f8e7 e1g1 e8g8 h2h3 h7h6 f2f3 a7a6 b5a4 b7b5 a4b3 c6a5 c1f4 e4f3 f1f3 a5b3 c2b3 f6e4 a1c1 c8e6 c3e4 d5e4 f3g3 e7h4 g3c3 f8e8 f4c7 d8d7 e2f4 h4g5 d4d5 g5f4 d5e6 d7d1 c1d1 f4g5 e6f7 g8f7 d1f1 f7g6 c3c6 g6h7');
INSERT INTO evaluations VALUES(45743,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(45743,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(45743,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(45743,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(45743,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(45743,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(45743,36,'cp -81','c6d5 c4b5 b8c6 d2d4 e5d4 d1d4 a7a6 b5c6 b7c6 g1f3 c6c5 d4d3 d5d4 c3e2 f8e7 b2b4 e8g8 b4c5 e7c5 e1g1 c8b7 e2g3 f8e8 c1f4 a6a5 f1e1 e8e1 a1e1 b7a6 d3b3 d4d3 f4e5 f6g4 g3e4 c5b6 c2d3 a5a4 b3a4');
INSERT INTO evaluations VALUES(46491,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(46491,31,'cp -71',NULL);
INSERT INTO evaluations VALUES(46491,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(46491,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(46491,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(46491,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(46491,36,'cp -76','c4b5 b8c6 d2d4 e5e4 g1e2 f8e7 e1g1 e8g8 f2f3 e4f3 f1f3 c8g4 f3f1 f8e8 b5a4 a7a6 d1d3 b7b5 a4b3 c6a5 e2g3 g4e6 g3f5 a5b3 c2b3 f6e4 f5e7 e8e7 c1f4 d8d7 h2h3 a8e8 a1c1 e6f5 c3e4 e7e4 d3c3 d7e6 f1f2 e6g6 g1h2 h7h6 b3b4 e8c8 c3g3 c8c1 g3g6 f5g6');
INSERT INTO evaluations VALUES(47689,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(47689,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(47689,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(47689,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(47689,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(47689,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(47689,36,'cp -25','b8d7 e2e3 a7a6 f3e5 c7c5 f1e2 d8c7 e1g1 f8d6 f2f4 b7b5 a2a3 e8g8 g1h1 c8b7 e2d3 c5c4 d3e2 b7c6 e5d7 f6d7 f4f5 d6h2 e2g4 h2g3 h4g3 c7g3 f5e6 d7f6 e6f7 g8f7 d1f3 g3f3 g4f3 f7e7 h1g1 a8e8 b2b3 e7d6 b3c4 b5c4');
INSERT INTO evaluations VALUES(48505,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(48505,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(48505,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(48505,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(48505,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(48505,35,'cp -77',NULL);
INSERT INTO evaluations VALUES(48505,36,'cp -77','f8b4 g1e2 f6d5 e1c1 d5e3 f3e3 b8d7 a2a3 b4d6 g2g4 d8h4 c3e4 d7f6 e4f6 g7f6 c1b1 c8d7 d4d5 h4g4 f1h3 g4c4 d5e6 d7e6 h3e6 c4e6 e3f2 e8c8 e2c3 d6e5 f2a7 e6a6 a7a6 b7a6 c3d5 d8d7 d1d3 h7h5 c2c4 f6f5 c4c5 e5g7');
INSERT INTO evaluations VALUES(42137,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(42137,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(42137,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(42137,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(42137,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(42137,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(42137,36,'cp -88','g1e2 f6d5 e1c1 d5e3 f3e3 b8d7 a2a3 b4d6 c3e4 d7f6 e2c3 e8g8 e4f6 d8f6 g2g3 e6e5 c3d5 f6f5 f1d3 f5h5 d3e2 e5d4 d1d4 h5g6 d5e7 d6e7 e3e7 g6h6 d4f4 h6e6 e7e6 c8e6 e2c4 a8e8 c4e6 e8e6 h1d1 f8e8 f4d4 g8f8 d4d7 e8e7');
INSERT INTO evaluations VALUES(48592,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(48592,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(48592,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(48592,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(48592,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(48592,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(48592,36,'cp 144','b1c3 f8d6 d2d3 e8g8 g1e2 c7c6 e1g1 g8h8 d5c6 b8c6 c3e4 d6e7 e4f6 e7f6 e2c3 c8f5 a2a4 h7h6 c1e3 a8c8 d1f3 f5h7 f3d5 c6b4 d5d8 f6d8 c4b3 a7a6 a1d1 d8a5 e3d2 b4c6 c3d5 a5d2 d1d2 f8d8');
INSERT INTO evaluations VALUES(48741,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(48741,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(48741,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(48741,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(48741,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48741,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48741,36,'cp -27','f8d6 e1g1 e8g8 c2c4 b8c6 b1c3 d5d4 c3b5 d6e7 f1e1 a7a6 b5a3 h7h6 h2h3 c6b4 a3c2 b4c2 d1c2 c7c5 c1f4 e7d6 f4d6 d8d6 f3d2 b7b6 e2f3 a8b8 d2e4 f6e4 f3e4 c8d7 b2b3 a6a5 e1e2 g7g6 c2d2 g8g7 a1e1 b8e8 a2a4 e8e7 e4d5 e7e2 e1e2 f8e8');
INSERT INTO evaluations VALUES(41728,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(41728,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(41728,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(41728,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(41728,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(41728,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(41728,36,'cp -16','e1g1 b8c6 f1e1 e8g8 b2b3 h7h6 b1d2 f8e8 c1b2 a7a5 a2a3 c8f5 d2f1 d5d4 f1g3 f5g6 f3h4 g6h7 h4f5 d6e5 e2f3 f6d5 f5d4 c6d4 e1e5 e8e5 b2d4 e5e6 d4b2 d8g5 c2c4 d5f4 f3b7 a8b8');
INSERT INTO evaluations VALUES(41727,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(41727,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(41727,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(41727,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(41727,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(41727,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(41727,36,'cp -22','b8c6 b2b3 e8g8 f1e1 f8e8 c1b2 h7h6 a2a3 c8f5 b1d2 a7a5 d2f1 d5d4 f3d4 c6d4 b2d4 d6h2 f1h2 d8d4 d1d2 b7b6 h2f1 f5g4 e2g4 f6g4 a1d1 a5a4 b3b4 c7c5 c2c3 d4d6 d3d4 c5b4 c3b4 d6d5 e1e8 a8e8 d1e1');
INSERT INTO evaluations VALUES(48822,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(48822,31,'cp -118',NULL);
INSERT INTO evaluations VALUES(48822,32,'cp -112',NULL);
INSERT INTO evaluations VALUES(48822,33,'cp -143',NULL);
INSERT INTO evaluations VALUES(48822,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(48822,35,'cp -163',NULL);
INSERT INTO evaluations VALUES(48822,36,'cp -136','d5d4 c3e2 d4d3 c2d3 d8d3 e2c3 c8g4 g1f3 b8c6 h2h3 c6b4 h3g4 b4c2 d1c2 d3c2 e1g1 c2d3 f3e5 d3d6 d2d4 d6d4 c1f4 f8b4 c3b5 d4b6 a2a4 e8g8 g4g5 f6h5 f4e3 b6e6 e5d3 b4a5 b5d4 e6e8 d3c5 a5b6 d4f5 g7g6 f5h6 g8g7 h6g4 b6c5 e3c5 e8e6 c5d4 f7f6 g2f3 a8d8 d4c3');
INSERT INTO evaluations VALUES(48921,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(48921,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(48921,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(48921,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(48921,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(48921,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(48921,36,'cp -119','c3e2 d4d3 c2d3 d8d3 e2c3 c8g4 g1f3 b8c6 h2h3 g4h5 g3g4 h5g6 g2f1 d3d6 d2d3 f6d7 a2a3 d7c5 b2b4 c5e6 h3h4 f7f6 h4h5 g6f7 d1a4 a7a6 c1e3 e8c8 c3d5 c8b8 a1c1 c6e7 d5c3 e7c8 a4c2 d6d7 h5h6 c8b6 h6g7 f8g7 g4g5 f6g5 e3g5 e6d4 f3d4 e5d4 c3e2');
INSERT INTO evaluations VALUES(49257,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(49257,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(49257,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(49257,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(49257,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(49257,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(49257,36,'cp 50','d7d6 d2d4 b8d7 f1e1 f8e7 c4b3 e8g8 c2c4 f6h7 b1c3 h7g5 c1e3 a7a5 h2h3 e7f6 f3g5 f6g5 a1b1 b7b6 c3d5 c8b7 b3a4 g5e3 e1e3 g8h8 d1d2 e5d4 d2d4 d7c5 a4c2 c5e6 d4c3 d8g5 b1d1 g5e5 b2b3 a8e8 a2a3 b7d5 c4d5 e6f4');
INSERT INTO evaluations VALUES(49555,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(49555,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(49555,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(49555,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(49555,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(49555,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(49555,36,'cp -60','c7c6 g1f3 e5e4 f3d4 d7d5 c4d5 f8c5 e2e3 c5d4 e3d4 e8g8 f1e2 f6d5 e1g1 c8f5 c3d5 d8d5 d2d3 e4d3 e2d3 f5d3 d1d3 b8a6 d3g3 f8e8 c1d2 f7f6 a1e1 d5d7 h2h4 a6c7 h4h5 g8f7 g3d3 g7g6 d3f3 c7e6 f3h3 a8d8 h5g6 h7g6 d2e3 d7d5 h3h7 e6g7');
INSERT INTO evaluations VALUES(51262,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(51262,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(51262,32,'cp 222',NULL);
INSERT INTO evaluations VALUES(51262,33,'cp 196',NULL);
INSERT INTO evaluations VALUES(51262,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(51262,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(51262,36,'cp 218','c3b5 e7e6 b5a3 f8a3 a1a3 g8f6 f1c4 e8g8 g1f3 d8c7 c4d3 a6a5 c1d2 b8c6 e1g1 c8a6 d2c3 f8b8 d3a6 a8a6 d1e2 c7b7 f3e1 b7b5 e1d3 c6b4 c3b4 a5b4 a3a6 b5a6 f1c1 h7h6 h2h3 f6e4 e2d1 a6b5 d3e5 b4b3 f2f3');
INSERT INTO evaluations VALUES(51260,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(51260,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(51260,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(51260,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(51260,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(51260,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(51260,36,'cp 200','e7e6 b5a3 f8a3 a1a3 g8f6 f1c4 e8g8 g1f3 d8c7 c4d3 a6a5 c1d2 b8c6 e1g1 c8a6 d3a6 a8a6 d2c3 f6d5 f3g5 f8a8 d1c2 d5f6 f1a1 c6b4 c2e2 b4d5 g5f3 h7h6 c3e1 a5a4 h2h3 c7c6 f3e5 c6b7 e5d3 f6e4 e2c2');
INSERT INTO evaluations VALUES(51396,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(51396,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(51396,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(51396,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(51396,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(51396,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(51396,36,'cp 12','d7d5 c4d3 d5e4 d3e4 c7c6 e4f3 c8b7 g1e2 b8d7 c1f4 g8e7 b1c3 e7g6 f4g3 f8e7 e2f4 g6f4 g3f4 d7f6 f4e5 e8g8 d1d2 f6d5 c3d5 c6d5 c2c3 b7a6 f3e2 e7g5 d2c2 a6e2 c2e2 a8c8 e1g1 g5f6 e5g3 d8d7 f1e1 d7c6 h2h4 g8h8');
INSERT INTO evaluations VALUES(52024,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(52024,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(52024,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(52024,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(52024,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(52024,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(52024,36,'cp 37','e2e4 d7d5 e4d5 e6d5 g1e2 d5d4 e1g1 f8e7 c2c3 b8c6 c3d4 c5d4 d1a4 e7f6 d2d3 g8e7 b1d2 e8g8 e2f4 f6e5 d2e4 c8f5 f1e1 e5f4 c1f4 e7g6 a4b5 g6f4 g3f4 f5e4 e1e4 d8d7 a1e1 a7a6 b5c5 f8e8 e4e8 a8e8 e1e8 d7e8 g2c6 e8c6 c5d4 c6g6 g1h1 g6c6 d4e4');
INSERT INTO evaluations VALUES(356,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(356,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(356,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(356,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(356,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(356,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(356,36,'cp 44','d1d4 c6h6 b1c3 g8e7 e1c1 e7c6 d4a4 d7d6 c3d5 e8d8 f1b5 c8d7 e4e5 c6e5 b5d7 e5d7 h1f1 h8f8 a4a5 b7b6 a5c3 a8c8 g2g4 a7a5 h2h4 c7c6 g4g5 h6e6 d5f4 e6a2 c3d4 f8e8 f4h5 g7g6 h5f6 e8e7 d4d6 a2a1 c1d2 a1b2 d2e1 e7e3 e1f2');
INSERT INTO evaluations VALUES(2613,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2613,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2613,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2613,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2613,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2613,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2613,36,'cp 0','d2d3 c8e6 c4e6 f7e6 c1e3 f6d7 c3e2 a6a5 e2g3 a5a4 d1e2 b7b5 e3c5 d7c5 c2c3 d8f6 a1d1 c5b3 d3d4 a8b8 g3h5 f6g6 h5g3');
INSERT INTO evaluations VALUES(4011,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(4011,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(4011,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(4011,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(4011,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(4011,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(4011,36,'cp 35','h7h6 g5h4 g6g5 h4g3 f6h5 d1c2 e7e6 f3d2 h5g3 h2g3 a7a6 e1g1 d8e7 f1e1 b7b6 a1d1 c8b7 a2a3 c7c6 e2f3 a8b8 b2b4 f8d8 a3a4 d7f6 b4b5 a6b5 a4b5 c6b5 c3b5 d6d5 c4d5 b7d5 d1c1 g5g4 f3d5 f6d5 e1f1 e7d7 c2b3 d8c8 d2c4');
INSERT INTO evaluations VALUES(4212,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(4212,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(4212,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(4212,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(4212,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(4212,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(4212,36,'cp 156','c3d4 c5b6 d4d5 c6b4 e4e5 f6e8 c1d2 a7a5 a2a3 b4a6 b1c3 d7d6 e5e6 d8e7 c3a4 b6a7 d2a5 f7e6 d5e6 a6c5 a4c5 a7c5 d1d2 f8f3 g2f3 d6d5 c4b3 c7c6 a5b4 c8e6 b4c5 e7c5 f1e1 c5e7 b3c2 e8d6 d2e3 a8e8 f3f4 e7d8 e3d3');
INSERT INTO evaluations VALUES(4210,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(4210,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(4210,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(4210,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(4210,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(4210,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(4210,36,'cp 146','c5b6 d4d5 c6e7 e4e5 f6e8 d5d6 c7d6 e5d6 e7g6 b1c3 b6a5 d1d3 a5c3 d3g6 d8f6 g6h5 c3b2 c1b2 f6b2 f3e5 e8d6 c4b3 a7a5 f1e1 a5a4 b3d5 b2d2 e5g6 d2g5 h5g5 h6g5 g6f8 g8f8 a1b1 a8a5 e1d1 f8e7 d5f3 f7f6 d1d4 d6f7 f3b7 c8b7 b1b7 f7e5 b7c7 e7e6 f2f3 d7d5 c7g7');
INSERT INTO evaluations VALUES(4136,30,'cp 172',NULL);
INSERT INTO evaluations VALUES(4136,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(4136,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(4136,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(4136,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(4136,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(4136,36,'cp 185','d4d5 c6e7 e4e5 f6g4 h2h3 g4f2 f1f2 b6f2 g1f2 d7d6 b1c3 d6e5 f2g1 f8e8 f3e5 d8d6 e5d3 c8h3 d1f3 h3f5 c1e3 f5d3 c4d3 e7g6 f3f2 g6e5 d3f5 d6e7 e3d4 a8d8 a1f1 c7c6 d4a7 c6d5 f5c2 d8d7 a7c5 e7g5 c5d4 e5c6 c2f5 d7e7 d4c5 e7e5 f5d7 e8d8 f2f7 g8h8 d7c6');
INSERT INTO evaluations VALUES(5628,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(5628,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(5628,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(5628,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(5628,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(5628,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(5628,36,'cp 30','c8g4 h2h3 g4h5 f1e1 g8f6 g2g4 h5g6 d2d4 h7h5 g4g5 f6h7 h3h4 d8d7 c4f1 f7f6 c3d5 e8g8 f1g2 f6g5 h4g5 e7d8 c2c3 d7f7 d1e2 a7a5 a2a4 h5h4 f3h4 g6h5 g5g6 h5e2 g6f7 f8f7 e1e2 d8h4 d4e5 c6e5 c1e3 e5g4 a1d1 g4e3 d5e3 h7g5 e3f5');
INSERT INTO evaluations VALUES(5873,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(5873,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(5873,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(5873,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(5873,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(5873,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(5873,36,'cp 42','c1g5 b4e7 g5f4 g8f6 a2a3 e8g8 d1c2 f6h5 f4d2 c8g4 e2e3 f8e8 f1d3 g4f3 g2f3 e7a3 a1a3 c6d4 c2a4 d4f3 e1d1 f3e5 c3e2 c7c5 d3b5 e8e7 d1c1 a7a6 d2a5 d8d6 a5c3 a8d8 c3e5 e7e5 b5d3 c5c4 d3c2 h5f6 c1b1');
INSERT INTO evaluations VALUES(6875,30,'cp 459',NULL);
INSERT INTO evaluations VALUES(6875,31,'cp 437',NULL);
INSERT INTO evaluations VALUES(6875,32,'cp 451',NULL);
INSERT INTO evaluations VALUES(6875,33,'cp 459',NULL);
INSERT INTO evaluations VALUES(6875,34,'cp 450',NULL);
INSERT INTO evaluations VALUES(6875,35,'cp 441',NULL);
INSERT INTO evaluations VALUES(6875,36,'cp 455','c2c3 d4b3 a2b3 c5e3 f2e3 d8h4 e1d2 g8f6 f1d3 f6e4 d3e4 h4e4 d1f3 e4g6 f3g3 d7d6 g3g6 h7g6 g1f3 e8e7 c3c4 c8g4 h2h3 g4f3 g2f3 a7a6 b1c3 f7f5 a1g1 e7f6 g1g3 g6g5 h1g1 h8h5 h3h4 h5h4 g3g5 a8g8 g5g6 f6f7 c3e2 a6a5 f3f4 h4h2 g6e6 g7g5 g1g5 g8g5 f4g5 f7g7 e6e7 g7g6 e7c7');
INSERT INTO evaluations VALUES(6873,30,'cp 446',NULL);
INSERT INTO evaluations VALUES(6873,31,'cp 448',NULL);
INSERT INTO evaluations VALUES(6873,32,'cp 442',NULL);
INSERT INTO evaluations VALUES(6873,33,'cp 426',NULL);
INSERT INTO evaluations VALUES(6873,34,'cp 424',NULL);
INSERT INTO evaluations VALUES(6873,35,'cp 425',NULL);
INSERT INTO evaluations VALUES(6873,36,'cp 448','d4b3 a2b3 c5e3 f2e3 d8h4 e1d2 g8f6 f1d3 f6e4 d3e4 h4e4 d1f3 e4g6 f3g3 d7d6 g1f3 g6g3 h2g3 c7c6 d5c6 b7c6 d2c1 c8e6 b3b4 e8d7 b1d2 h7h5 e3e4 e6g4 f3e1 f7f5 e4f5 g4f5 c3c4 h8b8 h1h5 f5g6 h5h1');
INSERT INTO evaluations VALUES(7423,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(7423,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(7423,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(7423,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(7423,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(7423,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(7423,36,'cp 117','c5a7 b1c3 e8g8 c1g5 h7h6 g5f6 d8f6 b3d5 c8d7 a1c1 a8b8 c3e2 b8b6 d5c6 b6c6 c1c6 d7c6 d1c2 c6a8 c2c7 a7b8 c7c2 b8a7 f1d1 a7b6 e2g3 f6f4 d1e1 a8b7 h2h3 g7g6 e1e3 f8c8 c2d2 g8g7 g3e2 f4f6 d2d3 f6e6 b2b3 g7g8 a2a4 e6d7 e2c3 f7f5 e4f5 d7f5 d3f5 g6f5 a4b5');
INSERT INTO evaluations VALUES(11888,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(11888,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(11888,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(11888,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(11888,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(11888,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(11888,36,'cp 0','d4e5 d8d5 e1g1 c8b7 g5f3 f6d7 b2b3 e8c8 f1e1 d5d3 b1a3 b7f3 d1f3 d3f3 g2f3 c4b3 a2b3 d7c5 a3c4 c5d3 e1e2 f8c5 c1b2 h8e8 b2c3 f7f6 g1f1 f6e5 c4e5 e8e5 c3e5 d3e5 e2e5 c5d4 e5e1 d4a1 e1a1 c8b7 f1e2 d8e8 e2d3 e8d8');
INSERT INTO evaluations VALUES(11887,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11887,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11887,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11887,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11887,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11887,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11887,36,'cp 0','d8d5 e1g1 c8b7 g5f3 f6d7 b2b3 e8c8 f1e1 d5d3 b1a3 b7f3 d1f3 d3f3 g2f3 c4b3 a3c4 d7c5 a2b3 c5d3 e1e2 f8c5 c1b2 h8e8 b2c3 f7f6 g1f1 f6e5 c4e5 e8e5 c3e5 d3e5 e2e5 c5d4 e5e1 d4a1 e1a1 c8b7 f1e2 d8e8 e2d1 e8f8 d1e2');
INSERT INTO evaluations VALUES(17963,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(17963,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(17963,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(17963,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(17963,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(17963,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(17963,36,'cp 113','d4d5 c6b8 g1f3 d7d6 c2c4 e7e5 b1c3 c8g4 a2a4 g7g6 a4a5 g4f3 d1f3 f8h6 c1h6 g8h6 f3h3 h6g8 b2b4 b8d7 f1e2 d8g5 h3g3 h7h6 g3f3 h6h5 h2h4 g5e7 f3e3 g8f6 c3a4 f6g4 e3a3 e8g8 f2f3 g4f6 c4c5 f6e8 a4c3 e8g7 c5c6 b7c6 d5c6');
INSERT INTO evaluations VALUES(18082,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(18082,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(18082,32,'cp 163',NULL);
INSERT INTO evaluations VALUES(18082,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(18082,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(18082,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(18082,36,'cp 153','a4b3 c8b7 c2c3 d8f6 c3d4 f8b4 b1c3 f6d4 e1g1 g8e7 c1e3 d4f6 c3e2 e8g8 a1c1 b4a5 e2g3 d7d5 d1h5 d5e4 d3e4 a8e8 f1d1 a5b6 e3b6 c7b6 e4e5 f6f4 c1c7 b7c8 g3e2 f4f5 h5f5 e7f5 e2f4 g7g6 f4d5 e8e5 g2g4 f5h4 f2f4 e5e2 g1f1 e2h2 d5e7 g8g7 e7c8 h4f3');
INSERT INTO evaluations VALUES(19594,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(19594,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19594,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(19594,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(19594,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(19594,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(19594,36,'cp 0','a7a6 a2a4 g8f6 g1e2 e7e6 e1g1 d7d5 e4d5 e6d5 c4a2 d4e2 d1e2 f8e7 f1e1 c8e6 d2d3 h7h6 c1f4 e7d6 f4d6 d8d6 e2e5 d6e5 e1e5 e8d7 a1e1 d7d6 f2f4 f6d7 e5e2 d7f6');
INSERT INTO evaluations VALUES(20933,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(20933,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20933,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20933,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20933,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20933,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20933,36,'cp 0','g7g5 h4g3 h6h5 f3g5 e7e5 g5f3 e5e4 f3e5 c6e7 e2e3 h5h4 g3f4 f7f6 f1b5 c7c6 b5e2 f6e5 e2h5 e8d7 h5g4 d7e8 g4h5');
INSERT INTO evaluations VALUES(20920,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20920,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20920,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20920,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20920,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20920,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20920,36,'cp 0','h4g3 h6h5 f3g5 e7e5 g5f3 e5e4 f3e5 c6e7 e2e3 h5h4 g3f4 f7f6 f1b5 c7c6 b5e2 f6e5 e2h5 e8d7 h5g4 d7e8 g4h5');
INSERT INTO evaluations VALUES(20923,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20923,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20923,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20923,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20923,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20923,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20923,36,'cp 0','h6h5 f3g5 e7e5 g5f3 e5e4 f3e5 c6e7 e2e3 h5h4 g3f4 f7f6 f1b5 c7c6 b5e2 f6e5 e2h5 e8d7 h5g4 d7e8 g4h5');
INSERT INTO evaluations VALUES(20934,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(20934,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(20934,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(20934,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(20934,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(20934,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(20934,36,'cp 45','e2e3 g8f6 f1b5 g7g5 h4g3 f6e4 b1d2 e4g3 h2g3 f8g7 g1f3 e8g8 e1g1 c6b8 c2c3 c8f5 f3e1 e7e5 d4e5 g7e5 b5d3 f5d3 e1d3 e5g7 d1f3 b8c6 d2b3 b7b6 a1d1 d8d7 d3b4 c6b4 c3b4 c7c6 b4b5 g7b2 b5c6 d7c6 d1d5');
INSERT INTO evaluations VALUES(21369,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(21369,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(21369,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21369,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(21369,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(21369,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21369,36,'cp 23','f8e7 g5e7 d8e7 e2e3 g8f6 a2a3 e8g8 b1c3 f8d8 d1c2 g7g6 c4d5 e6d5 f1e2 c6b8 b2b4 c7c6 e1g1 c8g4 h2h3 g4f3 e2f3 a7a6 c2b3 b8d7 a3a4 b7b5 a4a5 d8e8 f1c1 a8c8 f3e2 c8c7 b3b2');
INSERT INTO evaluations VALUES(5608,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(5608,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(5608,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(5608,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(5608,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(5608,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(5608,36,'cp 47','g5e7 d8e7 b1c3 g8f6 e2e3 e8g8 a2a3 f8d8 d1c2 c6b8 c4d5 e6d5 h2h3 b8d7 f1e2 c7c6 e2d3 d8e8 e1g1 f6e4 f1e1 d7f6 f3e5 e4c3 b2c3 c8e6 c2b2 c6c5 a3a4 a8c8 e1b1 c5c4 d3c2 c8c7 a4a5 g7g6');
INSERT INTO evaluations VALUES(22746,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(22746,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(22746,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(22746,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(22746,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(22746,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(22746,36,'cp 15','b2b4 g8f6 c1b2 d7d5 c4d5 d8d5 e2e3 d5e6 b1c3 f8d6 a1c1 a7a5 b4b5 c6d4 c3a4 d4f5 d1c2 e6e7 g1f3 e8g8 f1d3 e5e4 b2f6 e7f6 d3e4 d6a3 c1b1 g7g6 e1g1 f5d6 e4d3 c8e6 f3d4 a3b4 d4e6 f6e6 d3e2 a8d8 d2d4 d6e4 e2c4 e6e7 b1c1 c7c6 b5c6 b7c6 c4d3 e4d2 f1d1');
INSERT INTO evaluations VALUES(24797,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(24797,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(24797,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(24797,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(24797,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(24797,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(24797,36,'cp 52','e7e6 e2e3 f8b4 a1c1 f6e4 c4d5 e6d5 f1d3 e8g8 e1g1 b4c3 b2c3 c6a5 f3g5 f8e8 f2f3 e4g5 d3f5 a5c4 f1e1 b7b5 d1e2 g7g6 f5d3 g5e6 f4g3 c7c5 e2f2 c5d4 c3d4 d8b6 e3e4 b6d4 f2d4 e6d4 e4d5 e8e1 g3e1 a8d8 d3c4 b5c4 c1c4 d8d5');
INSERT INTO evaluations VALUES(25137,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(25137,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(25137,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(25137,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(25137,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(25137,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(25137,36,'cp 50','e2e3 f8b4 a1c1 f6e4 c4d5 e6d5 f1d3 e8g8 e1g1 b4c3 b2c3 c6a5 d1c2 f8e8 f1e1 a8c8 f3d2 f5g6 c3c4 e4d2 c2d2 a5c4 d3c4 d5c4 c1c4 d8d7 d4d5 c7c5 e1c1 b7b6 f2f3 f7f5 h2h3 h7h6 d5d6 c8d8 c4c3');
INSERT INTO evaluations VALUES(24885,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(24885,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(24885,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(24885,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(24885,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24885,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(24885,36,'cp 71','g2g4 h5g6 f1b5 f6e4 f3e5 e4c3 b2c3 a7a6 b5a4 b7b5 e5c6 d8d6 c6e5 b5a4 c3c4 d5c4 d1f3 c7c6 e5g6 h7g6 a1b1 e7e6 b1b6 d6d5 e1e2 f8e7 c1d2 c6c5 d4c5 e7c5 b6c6 c5e7 f3d5 e6d5 d2c3 e8d7 c6b6 h8b8 b6b8 a8b8 c3g7 f7f6 h3h4 d7e6 h4h5 b8b2 h5g6 b2a2 f2f4 a2c2 e2f3 a4a3 f4f5 e6d6 e3e4');
INSERT INTO evaluations VALUES(24884,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24884,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(24884,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(24884,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(24884,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(24884,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(24884,36,'cp 41','h5g6 f1b5 f6e4 f3e5 e4c3 b2c3 a7a6 b5a4 b7b5 e5c6 d8d6 c6e5 b5a4 c3c4 d5c4 c2c3 d6d5 d1a4 d5b5 a4c4 f7f6 c4b5 a6b5 e5g6 h7g6 c3c4 b5c4 a2a4 e7e5 d4e5 f8b4 e1e2 b4c3 a1a2 c3e5 c1b2 e5b2 a2b2 c4c3 b2a2 a8a5 e2d3 f6f5 g4g5 f5f4 h3h4 f4e3 f2e3 e8d7 d3c3 d7c6 c3d4 h8h5 d4e4 c6d7 e4d3 d7c8 d3e2');
INSERT INTO evaluations VALUES(25009,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(25009,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(25009,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(25009,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(25009,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(25009,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(25009,36,'cp 60','f1b5 f6e4 f3e5 e4c3 b2c3 a7a6 b5a4 b7b5 e5c6 d8d6 c6e5 b5a4 c3c4 d5c4 c2c3 d6d5 d1a4 d5b5 a4c4 f7f6 c4b5 a6b5 e5g6 h7g6 c3c4 b5c4 a2a4 e7e5 d4e5 f8b4 e1e2 b4c3 a1a2 c3e5 c1b2 e5b2 a2b2 c4c3 b2b3 c3c2 e2d2 e8d7 b3a3 a8a5 d2c2 h8a8 h1d1 d7e6');
INSERT INTO evaluations VALUES(26876,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(26876,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(26876,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(26876,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(26876,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(26876,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(26876,36,'cp 87','d1b3 g4f3 g2f3 d5c4 f1c4 d8d7 d4e5 c6a5 c4f7 d7f7 b3b5 c7c6 b5a5 f7f3 h1g1 g8h6 b1d2 f3f5 e3e4 f5f4 e5e6 g7g6 a5g5 f4g5 g1g5 e8c8 d2b3 d8e8 b3d4 f8g7 c1e3 h8f8 e1e2 g7d4 e3d4 e8e6 f2f3 h6f7 g5a5 a7a6 d4c3 f8d8 h2h4 d8d5 h4h5 f7d6 e2f2 d5a5 c3a5 g6h5 a1h1 e6e5 a5b4 c6c5 b4c3 e5g5 f2e3 b7b5 e3f4 d6f7 c3f6 g5g6');
INSERT INTO evaluations VALUES(32642,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(32642,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(32642,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(32642,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(32642,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(32642,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(32642,36,'cp 33','e7e6 h2h3 g4h5 d2d4 c7c6 f1e1 f8e7 g2g4 h5g6 c4b3 b8d7 c3e2 a5c7 c1f4 e7d6 d1d2 e8g8 f4d6 c7d6 d2f4 d6b8 f4b8 f8b8 e2f4 d7f8 f3e5 f6d5 b3d5 c6d5 c2c3 b8e8 a2a4 a7a5 e5d3 g6d3 f4d3 h7h6 a1d1 f8d7 g1g2 a8c8');
INSERT INTO evaluations VALUES(32699,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(32699,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(32699,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(32699,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(32699,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(32699,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(32699,36,'cp 27','h2h3 g4h5 d2d4 b8c6 c4b5 f8b4 d1d3 e8g8 b5c6 b7c6 f3e5 c6c5 c1e3 b4c3 b2c3 c5d4 c3d4 h5g6 e5g6 h7g6 c2c4 f8b8 f1b1 b8b1 d3b1 a5c3 b1c1 c3c1 a1c1 a8b8 e3f4 f6e8 d4d5 b8b2 a2a3 b2a2 c1c3 e6d5 c4d5 e8f6 c3c7 f6d5 c7c8 g8h7 f4d6 a2a1 g1h2 g6g5 c8a8 f7f6 a8a7 d5f4 d6c5 h7g8 a7b7');
INSERT INTO evaluations VALUES(33580,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(33580,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(33580,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(33580,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(33580,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(33580,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(33580,36,'cp 75','d2d4 c8f5 c1d2 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 c4b3 a7a5 g2g3 d7b6 h4g6 h7g6 a2a4 f8b4 d2f4 c7d7 c3e4 f6e4 e2e4 b6d5 f4e5 b4d6 c2c4 d5f6 e4e3 d7c7 f2f4 d6e5 d4e5 f6d7 b3c2 c7b6 e3b6 d7b6');
INSERT INTO evaluations VALUES(33633,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(33633,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(33633,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(33633,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(33633,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(33633,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(33633,36,'cp -11','f8e7 d2c3 b8c6 c1d2 c8d7 e1c1 h7h6 f3d4 e8c8 d4c6 d7c6 f2f3 h6h5 h2h4 d8e8 d1e1 h8f8 b2b3 e7f6 f1d3 c6d7 c3c4 f6e5 e1e2 g7g6 a2a4 e5g3 e2e8 f8e8 c1d1 a7a6 d2e1 g3e5 e1d2 d7e6 d2g5 c8d7 a4a5');
INSERT INTO evaluations VALUES(34403,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(34403,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(34403,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(34403,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(34403,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(34403,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(34403,36,'cp 146','d2d4 g8f6 h2h3 h7h6 c1e3 e7c7 c4b3 b7b5 a2a3 b8d7 f3h4 d7b6 e1g1 f8e7 d1e2 a8b8 f2f4 b6c4 b3c4 b5c4 f4e5 d6e5 e2c4 b8b2 d4e5 c7e5 e3d4 e5e6 c4d3 e8g8 e4e5 f6h5 h4f5 b2b7 c3e4 e6g6 f5d6 b7c7 e4c5 a7a5 a1e1 g6d3 c2d3');
INSERT INTO evaluations VALUES(34404,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(34404,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(34404,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(34404,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(34404,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(34404,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(34404,36,'cp 127','g8f6 c1e3 h7h6 h2h3 b7b5 c4b3 e7c7 a2a3 f8e7 g2g4 f6h7 d1d2 a7a5 d4d5 a5a4 b3a2 c7b7 d2d3 e8g8 d5c6 b8c6 d3b5 b7b5 c3b5 a8b8 b5c3 b8b2 c3a4 b2b8 a2d5 c8b7 a4b6 h7f6 e1g1 c6a5 d5b7 b8b7 f3d2 f8b8 a1b1 e7d8 b6d5 b8c8 f1c1 b7b1 c1b1 f6d5 e4d5 a5c4 d2c4');
INSERT INTO evaluations VALUES(36644,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36644,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36644,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36644,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36644,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(36644,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(36644,36,'cp 26','c2c4 c8f5 b1c3 f6e4 c3e4 f5e4 d4d5 e4f3 e2f3 a7a5 a1b1 a5a4 g2g3 b8d7 d1e2 d7c5 c1d2 d8e8 h2h4 b7b6 g1g2 h7h5 d2b4 f7f5 e2c2 g7f6 f3e2 f6g7 b4a3 g7e5 a3c5 b6c5 f2f4 e5f6 e3e4 f5e4 c2e4 a8b8 f4f5');
INSERT INTO evaluations VALUES(39104,30,'cp 226',NULL);
INSERT INTO evaluations VALUES(39104,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(39104,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(39104,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(39104,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(39104,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(39104,36,'cp 228','g1f3 b8c6 f1d3 g8f6 e1g1 e8g8 f1e1 f6e8 h2h3 f7f6 b1d2 d6e7 d2c4 d7d6 b2b4 g8h8 c4e3 c6b8 d3c2 c7c6 a2a4 e8c7 d1d3 g7g6 a4a5 b8d7 c2b3 c7b5 d4d5 b5c7 c3c4 c6d5');
INSERT INTO evaluations VALUES(39103,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(39103,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(39103,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(39103,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(39103,34,'cp 236',NULL);
INSERT INTO evaluations VALUES(39103,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(39103,36,'cp 228','b8c6 f1d3 g8f6 e1g1 e8g8 f1e1 f6e8 a2a3 f7f6 b2b4 g8h8 b1d2 d6e7 d2c4 d7d6 c4e3 g7g6 a3a4 e8g7 a4a5 a7a6 e3d5 c8e6 h2h3 e6f7 d5e7 c6e7 d4d5 d8d7 c3c4 g6g5 b4b5 e7g6 b5a6 b7a6 c4c5 g6f4 d3f1 f7h5 g2g4 h5g4 h3g4');
INSERT INTO evaluations VALUES(39185,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(39185,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(39185,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(39185,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(39185,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(39185,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(39185,36,'cp 161','d7d6 f1d3 b8d7 b2b4 a5b6 b1d2 e6e5 d2c4 g8f6 e1g1 e8g8 c1g5 f8e8 f1e1 d8e7 a2a4 c7c6 c4b6 d7b6 a4a5 b6d7 h2h3 h7h6 g5h4 d7f8 h4g3 f6h5 g3h2 f8g6 f3e5 d6e5 d1h5 e5d4 c3d4 e7b4 h5c5 b4c5 d4c5 c8e6');
INSERT INTO evaluations VALUES(39459,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(39459,31,'cp 180',NULL);
INSERT INTO evaluations VALUES(39459,32,'cp 208',NULL);
INSERT INTO evaluations VALUES(39459,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(39459,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(39459,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(39459,36,'cp 199','d7d6 d2d4 e5d4 f3d4 b8c6 d4c6 b7c6 d1h5 g7g6 h5f3 e8f8 b1c3 f8g7 c3a4 c5b6 f3c3 d8e8 b2b4 e8e5 c3d2 e5d4 d2e2 g8h6 f1d1 d4e5 c1b2 e5e7 a4b6 a7b6 a2a4 h6f7 f2f4 d6d5 e4d5 e7e2 c4e2 c6d5 c2c4 d5c4 e2c4 f7d6 c4d5 c8g4 d5a8 g4d1 a8c6');
INSERT INTO evaluations VALUES(40252,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(40252,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(40252,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40252,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40252,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40252,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40252,36,'cp 3','c8g4 f1e1 h7h6 h2h3 g4h5 g2g4 h5g6 c3e2 e8g8 c1f4 f8e8 f3h4 d8c7 f4d6 c7d6 e2g3 b8d7 d3g6 f7g6 e1e3 g6g5 d1e2 d7f6 g3f5 e7f5 h4f5 e8e3 e2e3 d6d7 a1e1 a8e8 e3d2 e8e1 d2e1 f6e4 g1g2 g7g6 f5g3 e4g3 f2g3 g8f7 a2a4 b7b6 c2c3 d7e7 e1e7 f7e7 g2f2 a7a5 b2b4 a5b4 c3b4');
INSERT INTO evaluations VALUES(41825,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(41825,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(41825,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(41825,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(41825,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(41825,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41825,36,'cp 47','c2c4 e8g8 d2d4 b7b6 c4d5 e6d5 b1c3 c8b7 d1b3 f8e8 c1f4 d6f4 g3f4 b8d7 e2e3 c7c5 f1d1 c5d4 d1d4 a8c8 b3d1 a7a6 f3e1 d7c5 a1c1 c5e6 d4d2 c8c7 e1f3 d8c8 c3e2 c7c1 d1c1 c8c1 e2c1 e8c8 c1b3 c8c4 f3d4 e6d4');
INSERT INTO evaluations VALUES(45038,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(45038,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(45038,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(45038,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(45038,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(45038,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(45038,36,'cp 175','c8g4 d1b3 d8b6 f3e5 b6b3 a2b3 g4c8 f1e2 e7e6 c1f4 f8b4 f2f3 f6d7 e5c4 b7b5 c4d6 e8e7 e4e5 a7a5 e1f2 d7b6 h2h4 c8d7 h4h5 h7h6 h1c1 b8a6 g2g4 a6c7 c3a2 b6d5 a2b4 a5b4 f4d2 a8a1 c1a1 h8a8 a1a8 c7a8 e2d3 e7f8 d6e4 a8c7 f2g3');
INSERT INTO evaluations VALUES(29084,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(29084,31,'cp 183',NULL);
INSERT INTO evaluations VALUES(29084,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(29084,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(29084,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(29084,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(29084,36,'cp 187','d1b3 d8b6 f3e5 b6b3 a2b3 g4c8 f1e2 e7e6 c1f4 f8b4 f2f3 f6d7 e5c4 b7b5 c4d6 e8e7 e4e5 d7b6 e1f2 c8d7 d6e4 h7h6 h1c1 a7a5 c3a2 b6d5 a2b4 d5f4 e4d6 f7f6 d6b7 b8a6 b4c6 d7c6 c1c6 h8c8 a1c1 c8c6 c1c6 f4e2 f2e2 a5a4 b7d6 e7d7 c6b6 a6c7 d6b5 c7b5');
INSERT INTO evaluations VALUES(48468,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(48468,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(48468,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(48468,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(48468,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(48468,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(48468,36,'cp 70','g1f3 f8e7 e1g1 a7a6 a2a4 b8c6 c1f4 e8g8 f1e1 e7d6 f3e5 c8d7 d1d3 d7e8 a1d1 d6e5 f4e5 c6e5 d4e5 d8d3 c4d3 f6d7 d3e4 a8b8 f2f4 d7c5 b2b3 c5e4 e1e4 b7b5 e4d4 f7f6 e5f6 g7f6 a4b5 a6b5 c3e4');
INSERT INTO evaluations VALUES(48648,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(48648,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(48648,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(48648,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(48648,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(48648,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(48648,36,'cp 17','e2e3 f8d6 f4d6 d8d6 d1c2 c8g4 f1d3 e8g8 h2h3 g4h5 g1e2 h5g6 e1g1 g6d3 c2d3 a7a5 d3b3 b8a6 b1c3 c7c6 e2f4 f8b8 f1e1 a6c7 f4d3 d6e7 b3c2 c7e6 a1d1 b8e8 d1c1 g7g6 a2a3 f6h5 b2b4 a5b4 a3b4 e7h4 b4b5');
INSERT INTO evaluations VALUES(48645,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(48645,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(48645,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(48645,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(48645,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48645,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(48645,36,'cp 10','c8f5 f1d3 f5d3 d1d3 c7c6 h2h3 f8d6 g1e2 e8g8 d3b3 d6f4 e2f4 d8e7 b3a3 e7d8 e1g1 a7a5 a3b3 d8e7 b1d2 f6e8 a2a4 e8d6 b3c3 b8d7 f1c1 d7f6 f4d3 f6e4 d2e4 d6e4 c3c2 g7g6 a1b1 e4d6 b2b4 g8g7 d3c5 a5b4 b1b4');
INSERT INTO evaluations VALUES(50643,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(50643,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(50643,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(50643,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(50643,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(50643,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(50643,36,'cp 77','e2e4 b7b5 a4b5 c6b5 c3b5 a6b5 a1a8 c8b7 a8a1 e7e6 g1e2 b7e4 e2c3 f8b4 f1e2 g8f6 e1g1 b4c3 b2c3 e8g8 c1g5 b8d7 e2f3 d8c7 g5d2 h7h6 f1e1 e4d5 d1e2 f8b8 e1b1 d7f8 f3d5 f6d5 e2f3 f8g6 h2h3 g6h4 f3g3 c7g3 f2g3 h4f5 g1f2 b8b7 a1a6 b5b4');
INSERT INTO evaluations VALUES(50644,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(50644,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(50644,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(50644,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(50644,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(50644,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(50644,36,'cp 86','b7b5 a4b5 c6b5 c3b5 a6b5 a1a8 c8b7 a8a1 e7e6 g1e2 b7e4 e2c3 f8b4 f1e2 g8f6 e1g1 b4c3 b2c3 e8g8 c1g5 h7h6 g5f6 d8f6 a1a5 b5b4 c3b4 b8c6 a5c5 c6b4 c5c4 b4d5 e2d3 d5b6 c4c1 f6d4 d3e4 d4e4 d1b3 b6d5 g2g3 e4f5 b3b7 f5d3 f1d1');
INSERT INTO evaluations VALUES(52587,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(52587,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(52587,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(52587,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(52587,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(52587,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(52587,36,'cp 75','b1c3 f8g7 c1e3 g8f6 g1f3 e8g8 h2h3 e7e5 d4d5 a7a5 f1e2 b8a6 e1g1 c8d7 f1e1 a5a4 b2b4 a4b3 a2b3 c6d5 c3d5 f6d5 c4d5 a6c5 a1a8 d8a8 d1b1 f8c8 e1c1 b7b5 e2f1 g7f8 f3d2 f7f5 b3b4 c5e4 d2e4 f5e4');
INSERT INTO evaluations VALUES(52585,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(52585,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(52585,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(52585,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(52585,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(52585,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(52585,36,'cp 87','g8f6 f1e2 f8g7 g1f3 e8g8 e1g1 e7e5 d4d5 b8a6 c1e3 f6g4 e3g5 f7f6 g5h4 c6c5 f3e1 g4h6 g2g4 c8d7 g1h1 b7b6 e1g2 a6c7 a2a4 h6f7 h4g3 a8b8 e2d3 a7a6 d1e2 c7e8 h2h4 f6f5 g4f5 g6f5 e4f5 f7h6 g2e3 h6f5 e3f5');
INSERT INTO evaluations VALUES(52838,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(52838,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(52838,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(52838,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(52838,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(52838,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(52838,36,'cp 24','c2c4 g8f6 b1c3 b8d7 d2d4 e7e6 f1d3 d5c4 d3c4 b7b5 c4d3 b5b4 c3e4 f6e4 d3e4 d8b6 e1g1 c8a6 f1e1 f8e7 a2a3 d7f6 e4d3 a6d3 d1d3 a7a5 a3b4 e7b4 c1d2 c6c5 d3c2 e8g8 d4c5 b4c5 e1c1 c5b4 h2h3 f8b8 c2a4 b8c8 d2b4 b6b4');
INSERT INTO evaluations VALUES(56780,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(56780,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(56780,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(56780,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(56780,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(56780,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(56780,36,'cp 135','c2c3 c7c5 a2a3 c8d7 b2b4 c5c4 a3a4 b8a6 h2h4 a6c7 b1a3 a7a6 g1h3 d7c6 f1e2 d8d7 a3b1 g8e7 h4h5 e7c8 h5h6 g7g6 c1g5 f8e7 b1d2 e7g5 h3g5 c8b6 a4a5 c6a4 d1b1 b6c8 b1a2 a4c6 d2f3 c8a7 a2d2 a7b5 g2g3 e8c8 e2d1 h8g8 e1f1 d7e7 f1g2 c8b8');
INSERT INTO evaluations VALUES(995,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(995,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(995,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(995,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(995,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(995,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(995,36,'cp -63','d2d3 c8f5 a2a3 e8c8 e1g1 e5e4 d3e4 d5e4 e2d3 e4d5 d3f5 d5f5 d1e2 f8c5 c1e3 c5e3 f2e3 f5f6 c2c3 f6g6 a1d1 h8e8 d1d8 c6d8 f1e1 g6e4 f3g5 e4e7 g5f3 d8c6 h2h3 a7a6 e2f2 c8b8 f3d4 c6e5 d4f5 e7f6 f5d4 f6f2 g1f2 g7g6');
INSERT INTO evaluations VALUES(2756,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2756,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2756,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2756,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2756,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2756,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2756,36,'cp 0','c5b6 d2d4 f8e8 f3g5 e8e7 g5f3');
INSERT INTO evaluations VALUES(3257,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(3257,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(3257,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(3257,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(3257,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3257,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3257,36,'cp 0','d6d5 e1c1 d5e4 f3e4 c8d7 c1b1 d8c8 h2h3 f8d8 d2c1 d7e8 d4b3 f6d7 b3d4');
INSERT INTO evaluations VALUES(6184,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6184,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(6184,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(6184,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(6184,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(6184,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(6184,36,'cp 38','e2e4 d7d6 d2d3 c6d4 g1e2 c8g4 h2h3 d4e2 c3e2 g4e6 c1e3 g8e7 h3h4 c7c5 h4h5 e7c6 e2c3 h7h6 h5g6 f7g6 c3d5 h6h5 d1d2 d8d7 a2a4 e8g8 b2b4 c6b4 d5b4 c5b4 d2b4');
INSERT INTO evaluations VALUES(6862,30,'cp 441',NULL);
INSERT INTO evaluations VALUES(6862,31,'cp 435',NULL);
INSERT INTO evaluations VALUES(6862,32,'cp 418',NULL);
INSERT INTO evaluations VALUES(6862,33,'cp 432',NULL);
INSERT INTO evaluations VALUES(6862,34,'cp 472',NULL);
INSERT INTO evaluations VALUES(6862,35,'cp 466',NULL);
INSERT INTO evaluations VALUES(6862,36,'cp 475','c4f7 e8e7 f7c4 d7d6 c1g5 g8f6 f2f4 d8e8 e1g1 h7h6 g5h4 e7f8 b1d2 b7b5 c4b3 a7a5 h4f6 g7f6 a2a4 b5a4 b3a4 c7c6 d1e1 c8d7 a4b3 h6h5 d2f3 h5h4 f4f5 e8e7 f3h4 d7e8 h4f3 d6d5 e4e5 f6e5 f3e5 c5d6 e5g6 e8g6 f5g6 f8g8 e1e7');
INSERT INTO evaluations VALUES(7015,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(7015,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(7015,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(7015,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(7015,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(7015,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(7015,36,'cp 75','c6e5 f3e5 g7e5 f1d3 d7d6 e1g1 g8f6 a2a4 e8g8 a4a5 f6d7 b1c3 e5g7 c1e3 d7c5 f2f3 c7c6 d3e2 c6d5 d1d5 g7c3 b2c3 c8d7 d5d2 f8e8 f1d1 d7c6 c3c4 d8c7 e4e5 c5d7 e5d6 c7d6 d2d4 d6d4 d1d4');
INSERT INTO evaluations VALUES(7040,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(7040,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(7040,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(7040,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(7040,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(7040,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(7040,36,'cp 86','f3e5 g7e5 c2c4 d7d6 f1d3 g8f6 b1c3 e7e6 e1g1 e6d5 c4d5 e8g8 c1g5 h7h6 g5h4 e5f4 f2f3 c7c6 d5c6 b7c6 c3e2 f4e3 g1h1 g6g5 h4e1 f8e8 e1c3 d6d5 e4e5 f6h5 d3h7 g8h7 d1d3 h7g8 d3e3 c8a6 a1e1 a6e2 e3e2 h5f4 e2d2 d5d4 c3d4');
INSERT INTO evaluations VALUES(9387,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(9387,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9387,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(9387,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(9387,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(9387,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(9387,36,'cp 0','c6d4 f3d4 c5d4 e1g1 e8g8 c1g5 d4c3 b2c3 h7h6 g5e3 f8e8 d1d2 a7a5 f1b1 a8b8 e3a7 b8a8');
INSERT INTO evaluations VALUES(12453,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(12453,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(12453,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(12453,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(12453,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(12453,35,'cp -95',NULL);
INSERT INTO evaluations VALUES(12453,36,'cp -58','h7h6 g5d2 d7d5 f1d3 f8b4 g1e2 e8g8 a2a3 b4d6 h2h3 f6e4 c3e4 d5e4 d3e4 f7f5 e4c6 b7c6 d2c3 e6e5 d1d2 d6c5 d2d8 f8d8 e1g1 c8a6 f1e1 e5e4 e2g3 g7g6 g3f1 d8d7 f1e3 g8f7 a1d1 a8d8 d1d7 d8d7 e1d1 d7d1 e3d1');
INSERT INTO evaluations VALUES(12821,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(12821,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(12821,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(12821,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(12821,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(12821,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(12821,36,'cp 13','e4e5 f6d5 g1f3 d7d6 f1b5 c8d7 e5d6 e7e6 b1c3 f8d6 c3d5 e6d5 d1e2 d8e7 d2d4 a7a6 b5c6 b7c6 b2a3 d7e6 c2c3 e8g8 e1g1 c5d4 a3d6 e7d6 c3d4 e6g4 e2e5 d6b4 e5e1 b4e1 f1e1 g4f3 g2f3 f8b8 a1c1 b8b6 e1e7 a6a5 e7c7 a5a4 c1c6 b6c6 c7c6');
INSERT INTO evaluations VALUES(16044,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(16044,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(16044,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(16044,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(16044,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(16044,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(16044,36,'cp 133','b5c6 d7c6 f3e5 f8g7 e5f3 b7b6 e1g1 c6c5 d2d3 e8g8 a2a4 a7a5 c1e3 f6h5 f1e1 c8b7 d3d4 c5d4 e3d4 f8e8 h2h3 h5f6 d1d3 b7a6 d3e3 a6b7 a1d1 d8e7 e4e5 f6d5 c3d5 b7d5 d4c3 d5c6 f3g5 h7h6 g5e4 g7e5 c3e5 e7e5 e3h6');
INSERT INTO evaluations VALUES(16198,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(16198,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(16198,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(16198,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(16198,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(16198,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(16198,36,'cp -78','d7d5 c4b5 f8c5 d2d4 e5d4 c3d4 c5b6 a2a4 a7a5 b5c6 b7c6 b1c3 e8g8 c3e4 d5e4 e1e4 c6c5 c1e3 c8b7 e4g4 f7f5 g4g6 f5f4 d4c5 f4e3 c5b6 e3f2 g1f2 d8h4 f2g1 h4e4 g6g3 e4e3 g1h1 c7b6 d1d4 e3d4 f3d4 f8f7 g3g6 a8d8 d4f3 f7f6 g6f6 g7f6 h1g1 b7f3 g2f3 d8d3 f3f4 g8f7 a1c1 f7g6');
INSERT INTO evaluations VALUES(14596,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(14596,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(14596,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(14596,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(14596,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(14596,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(14596,36,'cp -46','c4b5 f8c5 d2d4 e5d4 c3d4 c5b6 a2a4 a7a5 b5c6 b7c6 b1c3 e8g8 c3e4 d5e4 e1e4 c8f5 e4e1 f5g6 c1f4 f7f6 f3h4 g6h7 d1b3 d8d5 b3d5 c6d5 h4f3 g7g5 f4e3 f8e8 f3d2 h7g6 d2b3 g8f7 e1c1 h6h5 h2h3 g6f5 c1c6 f5d7 c6c3 a8d8 b3c5 d7c8');
INSERT INTO evaluations VALUES(18324,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(18324,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(18324,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(18324,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(18324,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(18324,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(18324,36,'cp 103','g1f3 a8b8 e1g1 f8e7 b2b3 g8f6 c1b2 d7d6 d2d4 e5d4 f3d4 c6d4 d1d4 c8b7 b1c3 b7g2 g1g2 e8g8 c3d5 f6e8 d4e3 e7g5 e3f3 g5f6 b2c3 f6c3 d5c3 e8f6 a1d1 f8e8 d1d2 f6d7 f1d1 d8e7 f3f4 d7c5');
INSERT INTO evaluations VALUES(20310,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20310,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(20310,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20310,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20310,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20310,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20310,36,'cp 0','g8f6 e2e4 d5e4 d3e4 f6e4 b1a3 f8e7 d1d8 e7d8 f1d1 d8e7 c1e3 e8g8 f3d2 e4d2 d1d2 e6e5 a3c4 c8e6 g2c6 b7c6 c4e5 f8c8 e5d7 c8d8 d7e5');
INSERT INTO evaluations VALUES(21804,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(21804,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(21804,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(21804,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(21804,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(21804,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(21804,36,'cp -21','c3e2 c8g4 a2a3 d8f6 e2g3 h7h5 h2h3 g4f3 d1f3 h5h4 g3e2 g7g6 f3f6 g8f6 e2g1 f8h6 g1f3 f6d7 c1h6 h8h6 g2g3 h4g3 f2g3 f7f6 f1e2 e8e7 f3h4 a7a5 e2g4 c6d8 e1g1 d8e6 f1f2 a5a4 a1f1 a8h8 b2b4 a4b3 g4e6 e7e6 c2b3');
INSERT INTO evaluations VALUES(21968,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21968,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21968,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21968,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21968,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21968,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21968,36,'cp 0','f7f5 b1d2 g8f6 f1e2 f8e7 d1b3 f5e4 d3e4 c6a5 b3a4 a5c6');
INSERT INTO evaluations VALUES(24866,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(24866,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(24866,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(24866,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(24866,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(24866,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(24866,36,'cp -21','e7e6 f1d3 f8d6 c3b5 f5g6 e1g1 a7a6 b5d6 c7d6 b2b3 a8c8 c1d2 g6d3 c2d3 e6e5 d2c3 e5e4 f3d2 c6e7 c3b2 e4d3 d2f3 e8g8 d1d3 d8d7 f1c1 d7e6 a3a4 h7h5 f3g5 e6g4 g5f3 g4e4 f3e1 e4e6 c1c8 f8c8 a1c1 c8c1');
INSERT INTO evaluations VALUES(25181,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(25181,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(25181,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(25181,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(25181,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(25181,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(25181,36,'cp -11','f1d3 f8d6 c3b5 f5d3 c2d3 a7a6 b5d6 c7d6 e1g1 e8g8 c1d2 h7h6 h2h3 f8e8 d2c3 a8c8 f1e1 e6e5 d4e5 d6e5 a1c1 d8d7 d1b3 f6h7 e1d1 d5d4 e3d4 e5d4 c3d2 h7f8 d1e1 g8h7 c1c5 e8e1 f3e1 c8d8 c5c1 f8e6 e1f3 a6a5 b3b5');
INSERT INTO evaluations VALUES(25840,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(25840,31,'cp 348',NULL);
INSERT INTO evaluations VALUES(25840,32,'cp 360',NULL);
INSERT INTO evaluations VALUES(25840,33,'cp 367',NULL);
INSERT INTO evaluations VALUES(25840,34,'cp 370',NULL);
INSERT INTO evaluations VALUES(25840,35,'cp 369',NULL);
INSERT INTO evaluations VALUES(25840,36,'cp 361','e7e6 f3g3 g8f6 f1e2 f6e4 g3e5 f8b4 b1d2 e4f6 e2f3 a7a5 f3a8 d8a8 g1e2 a5a4 e1g1 e8g8 e3e4 f8c8 e5f4 a8c6 f4g3 f6e4 d2e4 c6e4 e2c3 e4d3 f1d1 d3g3 h2g3 b4f8 d4d5 e6d5 c3d5 d7c6 c1f4 c8e8 f4e3 b5b4 d5b6 c4c3 b2c3 b4c3 b6a4');
INSERT INTO evaluations VALUES(28385,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(28385,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(28385,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(28385,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(28385,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(28385,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(28385,36,'cp 77','b8c6 c1e3 c6a5 e1c1 c7c6 g2g4 e8g8 a2a4 a5c4 d3c4 d8a5 h3h4 f6d7 e3g5 a5d8 g5e7 d8e7 f3d3 d7f6 f2f3 f8d8 c3e2 d8d7 e2g3 a8d8 d3b3 e7e6 h1e1 g7g6 h4h5 f6e8 c1b1 e8g7 e1h1 e6e7 b3e3 g7e6 g3e2 b7b6 e3f2 d6d5 c4d5 c6d5 h5g6 f7g6');
INSERT INTO evaluations VALUES(24057,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(24057,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(24057,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(24057,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(24057,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(24057,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(24057,36,'cp 71','c1e3 c6a5 e1c1 c7c6 g2g4 e8g8 a2a4 a5c4 d3c4 d8a5 h3h4 f6d7 e3g5 a5d8 g5e7 d8e7 f3d3 d7f6 f2f3 f8d8 c3e2 f6e8 e2g3 g7g6 g3f1 e8g7 f1e3 g7e6 d3d2 e7f6 d1f1 a7a6 a4a5 a8b8 b2b3 b7b5 a5b6 b8b6 c1b2 b6b7');
INSERT INTO evaluations VALUES(31532,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(31532,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(31532,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31532,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(31532,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31532,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31532,36,'cp 0','d8b6 d1c1 b8d7 f1e2 f7f6 e5f6 g8f6 c2c4 f8e7 h2h3 e8g8 b1d2 f6e4 e1g1 h7h6 d2e4 f5e4 f3d2 e4f5 d2f3');
INSERT INTO evaluations VALUES(31928,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(31928,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(31928,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31928,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(31928,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(31928,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31928,36,'cp 41','h2h3 g4f3 d1f3 d5e4 f3e4 c7c6 d2d4 g8f6 e4h4 b8d7 f1d3 e7e6 c2c3 e6e5 e1g1 e5d4 h4d4 f8e7 d3c2 e8g8 b1d2 f8e8 a2a4 a7a5 d2e4 f6e4 d4e4 d7f6 e4f3 d8d5 f3d5 f6d5 g2g3 e7f6 f1d1 e8e2');
INSERT INTO evaluations VALUES(37109,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37109,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37109,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37109,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37109,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37109,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37109,36,'cp 0','a2a3 b4e7 d2g5 h7h6 g5h4 c7c6 e2e3 c8f5 h4f6 e7f6 f1d3 f5d3 d1d3 b8d7 e1g1 a7a5 f1e1 f8e8 e1e2 d7f8 a1c1 f8e6 g2g3 f6e7 g1g2 e7d6 h2h4 d8d7 c3a4 d7d8');
INSERT INTO evaluations VALUES(37899,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(37899,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(37899,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(37899,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(37899,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(37899,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(37899,36,'cp 121','f1d3 g8f6 e1g1 b8d7 e4e5 d6e5 d4e5 f6d5 c3e4 b7b5 a2a4 b5b4 e4d6 e7d6 e5d6 h7h6 c2c3 b4c3 b2c3 d5c3 d1c2 c3d5 c1a3 c6c5 f1b1 d8a5 b1b5 a5d8 b5b3 e8g8 a1d1 a7a5 d3h7 g8h8 h7e4 d7f6 e4d5 f6d5 c2c5 c8d7 f3e5 d7a4');
INSERT INTO evaluations VALUES(37932,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(37932,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37932,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(37932,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(37932,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(37932,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(37932,36,'cp 59','g8f6 c4b3 e8g8 e1g1 d8c7 e4e5 d6e5 d4e5 f8d8 d1e2 f6d5 f1d1 c8e6 c3d5 c6d5 c1f4 a7a5 a2a4 b8a6 h2h4 a6c5 f3d4 a8c8 h4h5 g6g5 d4e6 f7e6 f4g5 c5b3 c2b3 c7e5 e2g4 e5f5 g4g3 g7e5 f2f4 e5g7 h5h6 g7f8 d1e1');
INSERT INTO evaluations VALUES(38060,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(38060,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(38060,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(38060,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(38060,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(38060,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(38060,36,'cp 161','c2c3 b4f8 f1d3 b7b6 b1d2 c8a6 d3a6 b8a6 d2f1 a6c7 f1g3 a7a5 c1e3 a5a4 e1g1 h7h5 e3g5 f8e7 h2h4 g8h6 a1c1 b6b5 d1c2 g7g6 g3e2 h6f5 g5e7 d8e7 g2g3 c6c5 f3g5 e8g8 f1e1 a8a7 e2f4 c5c4 c2d1');
INSERT INTO evaluations VALUES(38059,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(38059,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(38059,32,'cp 148',NULL);
INSERT INTO evaluations VALUES(38059,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(38059,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(38059,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(38059,36,'cp 142','b4f8 f1d3 b7b6 b1d2 c8a6 d3a6 b8a6 d2f1 a6c7 f1g3 a7a5 c1e3 h7h5 h2h4 f8e7 e3g5 g7g6 a1c1 b6b5 e1g1 g8h6 g3e2 h6f5 g2g3 c7a6 d1d3 e7g5 f3g5 d8d7 b2b3 c6c5 d3f3 a8c8 f1d1 c5c4 e2f4 a5a4 b3c4 c8c4 d1e1 a6c7');
INSERT INTO evaluations VALUES(38209,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(38209,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38209,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(38209,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(38209,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(38209,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(38209,36,'cp 16','e1g1 g8f6 a2a4 e8g8 b1c3 c8e6 h2h3 b8c6 f1e1 c6b4 c4e6 f7e6 c3e2 c7c5 c2c3 b4c6 b2b4 c5b4 c3b4 c6b4 d1b3 a7a5 c1a3 g8h7 a3b4 a5b4 b3e6 f6d7 d3d4 e5d4 e2d4 e7f6 e6f5 h7h8 d4e6 f6a1 e6f8 d7f8 e1a1 d8d7 f3d4 d7f5 d4f5 f8e6 f5d6');
INSERT INTO evaluations VALUES(39016,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(39016,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(39016,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(39016,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(39016,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(39016,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(39016,36,'cp 40','g1f3 g8f6 b2b4 e8g8 f1e2 d7d6 e1g1 b7b6 b1c3 c8b7 d1b3 d8e7 f3d2 b8d7 a2a4 e6e5 c1b2 a7a5 b4b5 f8e8 e2f3 b7f3 d2f3 e7e6 h2h3 e5e4 f3d2 h7h5 c3d5 f6d5 c4d5 e6f5 a1c1 a8c8 c1c6 d7f6');
INSERT INTO evaluations VALUES(39440,30,'cp 241',NULL);
INSERT INTO evaluations VALUES(39440,31,'cp 257',NULL);
INSERT INTO evaluations VALUES(39440,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(39440,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(39440,34,'cp 241',NULL);
INSERT INTO evaluations VALUES(39440,35,'cp 243',NULL);
INSERT INTO evaluations VALUES(39440,36,'cp 225','d4e5 b8c6 g1f3 d7d6 b1a3 d6e5 d1d8 e8d8 a3c4 f7f6 b2b4 a5b6 a2a4 a7a6 c4b6 c7b6 f1c4 d8c7 c1e3 c8g4 f3d2 g8e7 f2f3 g4d7 a4a5 e7c8 e1g1 b6a5 b4a5 b7b6 c4d5 b6a5 f3f4 c8d6 d2b3 a8b8 b3a5 b8b2 a5c6 d7c6 f4e5 d6e4');
INSERT INTO evaluations VALUES(41773,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(41773,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41773,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(41773,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(41773,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(41773,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41773,36,'cp 12','c4c5 d6e7 b2b4 b7b6 a2a3 e8g8 g1f3 c7c6 c1b2 a7a5 f1e2 c8a6 e1g1 d8c7 d1c2 b8d7 f1b1 h7h6 e2a6 a8a6 c3e2 f8a8 f3e1 f6g4 g2g3 g4f6 e1d3 c7b7 e2c1 b6b5 f2f3 e7d8 g1g2 d8c7 b2c3 a6a7 b4a5 c7a5');
INSERT INTO evaluations VALUES(44688,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(44688,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(44688,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(44688,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(44688,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44688,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44688,36,'cp -4','d2d4 e7f5 f1d3 f5d4 f3d4 c5d4 e1g1 d8b6 b1d2 c8d7 d2f3 a7a5 d1e2 b8a6 e2f2 a6b4 f3d4 f8c5 c1e3 e8g8 a2a3 b4d3 c2d3 f7f6 a1c1 f8f7 e5f6 f7f6 b2b4 a5b4 a3b4 b6b4 c1b1 b4a4 b1b7 e6e5 d4f3 a4c6 e3c5 c6b7 f3e5 d7e8 h2h3 a8c8 c5d4');
INSERT INTO evaluations VALUES(48762,30,'cp 809',NULL);
INSERT INTO evaluations VALUES(48762,31,'cp 801',NULL);
INSERT INTO evaluations VALUES(48762,32,'cp 801',NULL);
INSERT INTO evaluations VALUES(48762,33,'cp 801',NULL);
INSERT INTO evaluations VALUES(48762,34,'cp 817',NULL);
INSERT INTO evaluations VALUES(48762,35,'cp 819',NULL);
INSERT INTO evaluations VALUES(48762,36,'cp 809','e5c6 d8e7 c6e7 f8e7 b1c3 c7c6 c1e3 e8g8 e1c1 c8e6 d3d4 b8a6 f2f3 h7h5 c1b1 e7d6 a2a3 e6f5 e2d2 a6c7 h2h3 a8d8 g2g4 f5g6 c3e2 c7e6 e2c1 f6d7 h1g1 d7b6 f3f4 h5g4 h3g4 b6c4 f1c4 d5c4');
INSERT INTO evaluations VALUES(47745,30,'cp 790',NULL);
INSERT INTO evaluations VALUES(47745,31,'cp 778',NULL);
INSERT INTO evaluations VALUES(47745,32,'cp 778',NULL);
INSERT INTO evaluations VALUES(47745,33,'cp 778',NULL);
INSERT INTO evaluations VALUES(47745,34,'cp 771',NULL);
INSERT INTO evaluations VALUES(47745,35,'cp 778',NULL);
INSERT INTO evaluations VALUES(47745,36,'cp 780','d8e7 c6e7 f8e7 b1c3 c7c6 c1e3 e8g8 e1c1 b7b5 f2f3 b5b4 c3a4 b8d7 e2f2 a7a5 g2g4 c8a6 g4g5 f6h5 f1h3 a6b5 h3d7 b5a4 f3f4 g7g6 c1b1 a4b5 h2h4 a5a4 d7g4 e7d6 g4h5 g6h5 e3d4 b4b3 d4f6 b3c2 f2c2 a4a3 b2b3 d6f4');
INSERT INTO evaluations VALUES(50305,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50305,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50305,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(50305,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(50305,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50305,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(50305,36,'cp 11','a7a5 d2d4 d7d6 g1f3 c8g4 e5d6 e7d6 b1c3 b8a6 h2h3 g4h5 f1e2 f8e7 e1g1 e8g8 f1e1 a6b4 c1f4 e7f6 c3a2 b4c6 g2g4 h5g6 e2f1 d8d7 g4g5 f6e7 d4d5 c6e5 f3e5 d6e5 f4e5 b6a4 a2c3 a4c3 e5c3 e7g5 d1g4 d7g4 h3g4 g5f6 c3f6 g7f6 f1g2');
INSERT INTO evaluations VALUES(43696,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(43696,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(43696,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(43696,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(43696,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(43696,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43696,36,'cp 12','g1f3 d7d6 d2d4 c8g4 e5d6 e7d6 f1e2 f8e7 h2h3 g4h5 e1g1 e8g8 c1f4 b8a6 b1c3 a6b4 f1e1 e7f6 c3a2 b4a2 a1a2 f8e8 b2b3 c7c6 a2d2 h7h6 g2g4 h5g6 e2d3 e8e1 d1e1 g6d3 d2d3 d6d5 c4c5 b6d7 f4d6 b7b6 d3e3 b6c5 d4c5');
INSERT INTO evaluations VALUES(50360,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(50360,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(50360,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(50360,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(50360,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50360,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(50360,36,'cp 29','g4g5 f6h5 e2e4 g7g6 b1c3 c7c6 f1e2 h5f4 d2d4 f4e2 g1e2 d7d5 h2h4 d5e4 c3e4 c8f5 e2g3 f5e4 g3e4 h7h5 e1g1 f8g7 c2c3 d8d5 f1e1 e8g8 c1f4 b8d7 d1b3 d5f5 f4g3 b7b5 a2a4 a7a5 g1g2 d7b6 e4c5 b6a4 c5a4 b5a4 a1a4 e7e6 b3c4 f5b5 c4b5 c6b5');
INSERT INTO evaluations VALUES(51105,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(51105,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(51105,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(51105,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(51105,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(51105,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51105,36,'cp 25','g8f6 g2g3 b7b5 f1g2 c8b7 a2a4 e7e6 e1g1 f8e7 a4b5 a6b5 a1a8 b7a8 c1g5 b5b4 d1d3 a8e4 d3b3 d7d5 c2c4 b4c3 b2c3 e8g8 b1d2 c7c5 f1a1 c5d4 c3d4 e4f3 d2f3 b8c6 b3b5 d8d7 g5f6 g7f6 e2e3 f8b8');
INSERT INTO evaluations VALUES(54889,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(54889,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(54889,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(54889,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(54889,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(54889,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(54889,36,'cp 150','f1c4 f8g7 g1f3 b8c6 d1b3 e7e6 e1g1 c6d4 f3d4 d8d4 f1d1 d4c5 c1e3 c5e7 c4e2 g8f6 b3c3 c8d7 e3c5 f6h5 c3e3 e7f6 b1c3 h5f4 e2f1 a7a6 a1c1 g6g5 c5a3 f6e5 g2g3 f4g6 c3e2 d7c6 e2d4 e5e4 d4c6 e4e3 f2e3 b7c6 f1g2 g6e5 c1c6 a8d8 d1d8 e8d8 c6a6 d8c8');
INSERT INTO evaluations VALUES(54882,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(54882,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(54882,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(54882,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(54882,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(54882,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(54882,36,'cp 179','b8c6 g1f3 f8g7 d1b3 e7e6 e1g1 g8e7 b1c3 e8g8 f1d1 c6a5 b3a4 a5c4 a4c4 c8d7 c1f4 a8c8 a1c1 d8e8 h2h3 b7b6 b2b4 f7f5 f4e5 g7h6 e4f5 f8f5 c1c2 f5f3 g2f3 e8f7 b4b5 f7f3 c4d3 f3d3 d1d3 h6g7 c2c1 g7h6 c1a1 e7d5 a2a4 c7c6 c3e4 c6b5');
INSERT INTO evaluations VALUES(56461,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(56461,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(56461,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(56461,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(56461,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(56461,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(56461,36,'cp -19','e2e4 g8f6 g1f3 b8c6 f1e2 f8d6 e1g1 e8g8 f1e1 a7a5 e4d5 f6d5 d2c4 f8e8 c2c3 h7h6 h2h3 c8f5 e2f1 a5a4 c1d2 d8f6 c4d6 c7d6 a1c1 d5c7 d2e3 d6d5 d3d4 e5e4 f3e5 c6e5 d4e5 f6e5 c3c4 d5c4 c1c4 c7e6');
INSERT INTO evaluations VALUES(56852,30,'cp -332',NULL);
INSERT INTO evaluations VALUES(56852,31,'cp -293',NULL);
INSERT INTO evaluations VALUES(56852,32,'cp -277',NULL);
INSERT INTO evaluations VALUES(56852,33,'cp -325',NULL);
INSERT INTO evaluations VALUES(56852,34,'cp -325',NULL);
INSERT INTO evaluations VALUES(56852,35,'cp -325',NULL);
INSERT INTO evaluations VALUES(56852,36,'cp -336','e8f7 f3e5 f7e8 d1g4 g8f6 g4f4 d7d6 e5f3 h8g8 e1g1 d8e7 d2d4 e7g7 f4g5 b8c6 b1c3 c8g4 d4d5 g4f3 g5g7 f8g7 d5c6 f3h5 e4e5 f6g4 c6b7 a8b8 e5d6 c7d6 c3e4 e8d7 c1f4 g4e5 f4e5 g7e5 e4f6 e5f6 f1f6 d7c6 a1e1 b8b7 b2b3 g8e8 e1e8 h5e8 g1f2 e8g6 f6f4 b7f7 f4f7 g6f7 f2e3 f7g6 c2c3 c6d5 e3f3 g6e4 f3f4 e4b1 f4g5 b1a2');
INSERT INTO evaluations VALUES(57333,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(57333,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(57333,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(57333,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(57333,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(57333,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(57333,36,'cp -82','g8f6 d2d3 d7d5 g1f3 f8d6 g2g3 e8g8 f1g2 h7h6 e1g1 f8e8 b1d2 e5e4 f3d4 c7c5 d4b5 e4d3 c2d3 d6f8 e2e4 b8a6 b5c3 c8g4 f2f3 g4c8 f3f4 c5c4 e4e5 f6g4 d2f3 c4d3 d1d3 f8c5 g1h1 d5d4 c3a4 c8e6 f4f5 a6b4');
INSERT INTO evaluations VALUES(57700,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(57700,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(57700,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(57700,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(57700,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(57700,35,'cp 191',NULL);
INSERT INTO evaluations VALUES(57700,36,'cp 173','e7e6 b1c3 a7a6 d2d4 b7b5 c4b3 d7d6 h2h4 c7c5 c1g5 f8e7 d4c5 d6c5 d1d8 e7d8 g5d8 e8d8 a2a4 c5c4 b3a2 b8c6 a4b5 a6b5 c3b5 d8e7 g1e2 c6b4 b5a3 a8a3 b2a3 b4c2 e1d2 c2a1 h1a1 g8f6 f2f3 c8a6 d2c3 f6d7 e2d4 h8c8 a2b1 h7h5 b1c2');
INSERT INTO evaluations VALUES(139,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(139,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(139,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(139,33,'cp 160',NULL);
INSERT INTO evaluations VALUES(139,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(139,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(139,36,'cp 128','f1d3 c8d7 c1e3 h7h6 e1g1 b4c3 b2c3 g8e7 f3d2 e7f5 e3f4 b6d8 d2b3 b7b6 a2a4 a8c8 f4c1 e8g8 c1a3 f8e8 d3a6 c8c7 a3d6 f5d6 e5d6 c7c8 a6c8 d7c8 d1d3 d8d6 a4a5 e6e5 d4e5 c6e5 d3g3 c8f5 a5b6 d6b6 b3d4 f5e4 f2f3');
INSERT INTO evaluations VALUES(140,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(140,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(140,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(140,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(140,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(140,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(140,36,'cp 158','c8d7 c1e3 g8e7 e1g1 b4c3 b2c3 h7h6 a1b1 b6c7 e3c1 c6a5 c1a3 d7c6 f3d2 e8g8 f1e1 a5c4 d2c4 d5c4 d3c4 f8c8 c4d3 e7d5 a3d6 c7d8 d1d2 b7b5 d6c5 d8g5 d2g5 h6g5 e1c1 a7a6 d3f1 g7g6 a2a4 g8g7 a4a5 g5g4');
INSERT INTO evaluations VALUES(712,30,'cp -553',NULL);
INSERT INTO evaluations VALUES(712,31,'cp -532',NULL);
INSERT INTO evaluations VALUES(712,32,'cp -557',NULL);
INSERT INTO evaluations VALUES(712,33,'cp -565',NULL);
INSERT INTO evaluations VALUES(712,34,'cp -563',NULL);
INSERT INTO evaluations VALUES(712,35,'cp -540',NULL);
INSERT INTO evaluations VALUES(712,36,'cp -525','e5c6 d4d5 d2d3 d5c6 e1g1 f8b4 c1d2 b4c5 b1c3 e8g8 c3e4 f6e4 d1f3 f7f5 d3e4 c8a6 e4e5 c6f3 g2f3 a6f1 g1f1 f8d8 d2a5 c5b6 a5b6 a7b6 a2a4 d8c8 b2b4 c8c4 a4a5 b6a5 b4b5 g8f7 b5b6 c4c2 b6b7 a8b8');
INSERT INTO evaluations VALUES(709,30,'cp -543',NULL);
INSERT INTO evaluations VALUES(709,31,'cp -540',NULL);
INSERT INTO evaluations VALUES(709,32,'cp -569',NULL);
INSERT INTO evaluations VALUES(709,33,'cp -572',NULL);
INSERT INTO evaluations VALUES(709,34,'cp -573',NULL);
INSERT INTO evaluations VALUES(709,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(709,36,'cp -585','d4d5 c6d4 d5d4 d2d3 c8g4 d1d2 f8c5 b1c3 e4d3 e1g1 d3c2 f1e1 g4e6 d2c2 e8g8 c1e3 d4e5 e3h6 e5b8 h6g7 g8g7 c2d2 g7h7 d2c2 h7h8 c3e4 c5d4 c2d2 b8b2 d2h6 f6h7 a1b1 d4g7 b1b2 g7h6 a2a4 h6g7 b2b4 f8d8 e4g3 a7a5 b4b5 g7c3 e1b1 h7f6');
INSERT INTO evaluations VALUES(3767,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(3767,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(3767,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(3767,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(3767,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(3767,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(3767,36,'cp 29','a2a3 d8e8 f3e5 f6e4 e5c6 e8c6 b1d2 b7b6 d2e4 d5e4 d3e2 f8d8 g3h4 d8f8 a1c1 e6e5 c2c3 c8e6 h4g3 f8d8 e1g1 f7f5 d1c2 f5f4 d4e5 f4g3 e5d6 g3f2 f1f2 d8d6 c1d1 d6d1 c2d1 e6f7 d1d4 a7a5 f2f4 a8e8 b2b4 c6d5 d4d5 f7d5 b4a5');
INSERT INTO evaluations VALUES(4394,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(4394,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(4394,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(4394,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(4394,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(4394,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(4394,36,'cp -37','a7a5 g2g4 a8b8 c1g5 h7h6 g5h4 b7b5 c3d5 g7g5 h4g3 b5b4 a3b4 a5b4 d5e3 c5e3 f2e3 d7d6 f1e2 g8g7 e1d2 f8h8 h3h4 f6g4 f3g5 h6h5 g5f3 h8h6 f3h2 b4b3 c2c3 g4h2 h1h2 h6g6 d1g1 c8g4 e2g4 h5g4 h4h5 g6g5');
INSERT INTO evaluations VALUES(5203,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5203,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5203,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5203,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5203,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5203,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5203,36,'cp 0','g8f6 b1c3 e8g8 a2a3 d7d6 d1d3 c8e6 c1g5 a8c8 g5f6 d8f6 c3d5 f6g6 e1g1 a7a6 c2c3 g8h8 a3a4 c5a7 a1d1 g6h6 d3d2 h6h5 c4e2 h5g6 d2d3 g6h6 d3d2');
INSERT INTO evaluations VALUES(7854,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(7854,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(7854,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7854,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(7854,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7854,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7854,36,'cp 0','b1c3 d7d6 d1d3 a7a6 c1g5 b7b5 c4d5 c6b4 d3e2 b4d5 c3d5 d8a5 g5d2 a5d8');
INSERT INTO evaluations VALUES(9801,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(9801,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(9801,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(9801,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(9801,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(9801,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(9801,36,'cp 23','c5b6 c3d5 e8g8 c2c3 d7d6 a2a4 c8e6 d5b6 a7b6 d1c2 b6b5 c1e3 a8a4 a1a4 b5a4 c2a4 d8a8 f1a1 a8a4 a1a4 d6d5 e4d5 f6d5 a4a3 f7f6 b2b4 c6e7 e3d2 g8f7 c3c4 d5b6 a3a7 f8b8 b4b5 b6c8 a7a3 c8d6 d2b4 e7g6 a3a7 g6f4 e2f1');
INSERT INTO evaluations VALUES(10167,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(10167,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(10167,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(10167,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(10167,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(10167,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(10167,36,'cp -13','d7d6 h2h3 h7h6 a2a3 a7a5 c1e3 e8g8 e1g1 f8e8 b1c3 c8d7 a1c1 f6h7 d1d2 h7f8 d3d4 c6d4 f3d4 e5d4 e3d4 f8e6 d4c5 e6c5 e2d1 c5e4 c3e4 e8e4 d1f3 e4e8 f3b7 a8b8 b7d5 a5a4 c1c3 c7c6 d5f3 d8b6 d2d6 b6b2 d6d4 b8d8');
INSERT INTO evaluations VALUES(9381,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(9381,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(9381,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(9381,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(9381,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(9381,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9381,36,'cp -9','h2h3 h7h6 a2a3 a7a5 c1e3 c8d7 e1g1 e8g8 b1c3 c6d4 e3d4 e5d4 c3b5 d7b5 c4b5 a5a4 d1c1 d8d7 c1c4 d7e6 f3d2 f6d7 c4c2 d7b6 d2c4 e6d7 b2b4 a4b3 c2b3 b6c4 b3c4 a8a3 a1a3 c5a3 e2g4 d7e7 g2g3 f8a8 c4d5');
INSERT INTO evaluations VALUES(16302,30,'cp -373',NULL);
INSERT INTO evaluations VALUES(16302,31,'cp -366',NULL);
INSERT INTO evaluations VALUES(16302,32,'cp -369',NULL);
INSERT INTO evaluations VALUES(16302,33,'cp -362',NULL);
INSERT INTO evaluations VALUES(16302,34,'cp -365',NULL);
INSERT INTO evaluations VALUES(16302,35,'cp -371',NULL);
INSERT INTO evaluations VALUES(16302,36,'cp -368','h7h6 g5h4 g7g5 h4g3 f6e4 g3f3 e4d6 b5c6 d7c6 d2d3 d6f5 b1c3 f5d4 f3d1 c8e6 c1e3 d8e7 d1d2 e8c8 e1c1 f8g7 f2f3 f7f5 d2f2 g7f6 c1b1 b7b6 g1e2 c8b7 d1e1 e7f7 g2g4 f5f4 e3d2 c6c5 b2b3 a7a5 c3e4 a5a4 e2d4 e5d4 e4f6 f7f6 f2e2 h8e8 e2e4 b7b8');
INSERT INTO evaluations VALUES(16099,30,'cp -375',NULL);
INSERT INTO evaluations VALUES(16099,31,'cp -400',NULL);
INSERT INTO evaluations VALUES(16099,32,'cp -386',NULL);
INSERT INTO evaluations VALUES(16099,33,'cp -378',NULL);
INSERT INTO evaluations VALUES(16099,34,'cp -388',NULL);
INSERT INTO evaluations VALUES(16099,35,'cp -388',NULL);
INSERT INTO evaluations VALUES(16099,36,'cp -395','g5g3 f6e4 g3f3 e4g5 f3e3 g5e6 b5c6 d7c6 d2d3 f8d6 b1c3 e6d4 e3d2 e8g8 c3e4 d6e7 e4g3 f7f5 g1e2 e7d6 f2f3 c6c5 e2d4 c5d4 d2f2 b7b6 f3f4 d8e7 c1d2 e5f4 g3e2 f8e8 e1d1 c8b7 e2d4 e7g5 d4f3 g5h5');
INSERT INTO evaluations VALUES(17180,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(17180,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(17180,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(17180,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(17180,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17180,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(17180,36,'cp 11','d7d5 e4d5 c6a5 d2d3 f7f6 g5e4 f6f5 e4c3 c7c6 c1g5 b7b5 d1h5 g7g6 h5h4 b5c4 g5f6 c4d3 f6h8 d3c2 b1a3 c8a6 h8e5 e7d5 h4d8 a8d8 a3c2 a5c4 c3d5 d8d5 e5d4 c6c5 d4c3 c4e5 c3e5 d5e5 e1d1 e5d5 d1c1 f8h6 c2e3 d5d4 h1e1 a6b7 g2g3 b7e4 f2f4 e8f7 e1d1 g6g5 g3g4 d4d1 c1d1 f5g4 f4g5 e4f3 d1e1');
INSERT INTO evaluations VALUES(14057,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(14057,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(14057,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(14057,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(14057,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(14057,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(14057,36,'cp 28','e4d5 c6a5 d2d3 f7f6 g5e4 f6f5 e4c3 c7c6 a2a3 c6d5 c4a2 c8e6 f2f4 e7g6 e1g1 a5c6 f4e5 g6e5 d3d4 e5g4 c3e2 f8d6 h2h3 g4f6 e2f4 e6f7 b1c3 c6e7 d1f3 d6b8 f4d3 e8g8 c1g5 f6e4 g5e7 d8e7 c3d5 e7d8 f3f5 e4g3 f5f2 g3f1 a1f1');
INSERT INTO evaluations VALUES(14056,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(14056,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(14056,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(14056,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(14056,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(14056,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(14056,36,'cp 22','c6a5 d2d3 f7f6 g5e4 f6f5 e4c3 c7c6 c1g5 b7b5 d1h5 g7g6 h5h4 b5c4 g5f6 c4d3 f6h8 d3c2 b1a3 c8a6 h8e5 e7d5 h4d8 a8d8 a3c2 a5c4 c3d5 d8d5 e5d4 c6c5 d4c3 c4e5 c3e5 d5e5 e1d1 e5d5 d1c1 f8h6 c2e3 d5d4 h1e1 e8f7 g2g3 d4c4 c1b1 c4d4 a2a4 d4d2 a1a3 d2f2 e1d1 f7e6 e3d5 a6b7 d5c7 e6f6');
INSERT INTO evaluations VALUES(17915,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(17915,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(17915,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(17915,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(17915,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(17915,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(17915,36,'cp 11','f8c5 c2c3 g8f6 d2d3 d7d6 e1g1 e8g8 h2h3 h7h6 c4b3 f8e8 f1e1 c8e6 b3e6 f7e6 c1e3 c5e3 e1e3 e8f8 b1d2 f6d7 d1b3 d7c5 b3c2 d8f6 b2b4 c5d7 a1f1 a6a5 b4b5 c6e7 d3d4 e7g6 f1e1 f6e7 g2g3 a8d8 e1d1');
INSERT INTO evaluations VALUES(18391,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(18391,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(18391,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(18391,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(18391,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(18391,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(18391,36,'cp 175','c4b3 g8f6 e4e5 f6e4 d2d3 e4c5 d1f3 c5b3 a2b3 d7d5 e5d6 d8d6 b1a3 d6c5 b3b4 c5b6 c2c4 d4c3 b2c3 c8e6 f3g3 h7h5 a3c2 h5h4 g3e5 e8c8 c1e3 b6b8 e5b8 c8b8 a1a7 d8d3 f1a1 d3c3 a7a8 b8c7 a1a7 c7d6');
INSERT INTO evaluations VALUES(27247,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(27247,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(27247,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(27247,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(27247,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(27247,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(27247,36,'cp 21','a4b3 g8f6 f1c4 d8e7 e1g1 c6a5 b3d3 a5c4 d3c4 d7e6 c4d4 e7d7 c2c4 d7c6 b1d2 f6g4 b2b3 f8e7 d4d3 g4e5 d3c2 e5f3 d2f3 e7f6 a1b1 e8g8 c1b2 f6b2 b1b2 a7a5 a2a4 a8e8 h2h3 e6d7 f1e1 c6c5 c2d3 f7f5 e4e5 d7c6 e5d6 e8e1 f3e1 c5d6 d3d6');
INSERT INTO evaluations VALUES(28692,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28692,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28692,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(28692,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28692,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28692,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28692,36,'cp 0','b1c3 h7h6 g5h4 f8e7 e2e3 e8g8 f1e2 f6e4 c3e4 b7e4 h4e7 d8e7 e1g1 c7c5 a1c1 d7d6 e2d3 e4f3 d1f3 b8d7 f3b7 f8b8 b7f3');
INSERT INTO evaluations VALUES(29716,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(29716,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(29716,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(29716,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(29716,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(29716,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(29716,36,'cp 70','f1c4 e7e6 b1c3 c7c6 e1g1 f8e7 f1e1 f6e4 c4d3 e4c3 b2c3 f5d3 d1d3 b8d7 a1b1 e8g8 c3c4 b7b6 e3e4 e6e5 d4e5 d8c7 c1b2 d7c5 d3e2 c5e6 f3d4 e7c5 d4b3 a8d8 e1d1 c7e7 g2g3 h7h6 g1g2 e7g5 h2h4 e6f4 g2f1 g5e7 g3f4 e7h4');
INSERT INTO evaluations VALUES(31315,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31315,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31315,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31315,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31315,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31315,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(31315,36,'cp 0','e7e6 d1b3 d8c7 f1e2 g8f6 b1a3 f8e7 f3e5 h7h6 c1f4 c7c8 a3c2 e8g8 c2e3 f5h7 e2d3 b8c6 d3h7 g8h7 e5d3 f6h5 f4g3 e7g5 e3g4 f7f5 f2f4 g5f4 g3f4 h5f4 d3f4 f5g4 e1g1 f8f5 b3d1 e6e5 d4e5 f5e5 f4d3 e5e4 d3f4 e4e5');
INSERT INTO evaluations VALUES(31697,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(31697,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31697,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31697,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31697,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(31697,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(31697,36,'cp 13','c1f4 b8c6 b1d2 a7a6 d2b3 g8f6 f1d3 f5d3 d1d3 f8d6 f4d6 d8d6 b3c5 e8g8 e1g1 d6c7 d3e2 a8e8 f1e1 e6e5 d4e5 c6e5 f3e5 c7c5 e2d3 h7h6 g2g3 c5a5 g1g2 e8e6 e5f3 e6e4 e1e2 f8e8 f3d4 e4e2 d4e2 f6e4 e2d4 a5a4 h2h4 a4d7 d3f3');
INSERT INTO evaluations VALUES(31395,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(31395,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31395,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(31395,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31395,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31395,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(31395,36,'cp 28','f1e2 b8d7 e1g1 g8e7 b1c3 f5h7 e2d3 h7d3 c2d3 e7f5 c3e2 a7a5 a2a3 a5a4 c1d2 d8b6 d1c2 f8e7 g2g4 f5h4 f3h4 e7h4 f2f4 h4e7 d2e3 e8g8 a1c1 f7f6 e5f6 f8f6 e2c3 b6b3 c2b3 a4b3 f4f5');
INSERT INTO evaluations VALUES(34148,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(34148,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(34148,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(34148,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(34148,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(34148,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(34148,36,'cp 63','e1g1 b8c6 d4c6 b7c6 b2b3 g8e7 c1b2 e7g6 b1d2 f8d6 d1h5 d6e5 b2e5 c7e5 h5e2 e5c7 e2e3 e8g8 a1e1 e6e5 d2c4 f8d8 e1d1 d7d6 f1e1 c8e6 d1d2 g6e7 d3f1 f7f6 e3c3 d6d5 c4a5 h7h5 e4d5 e6d5 e1c1');
INSERT INTO evaluations VALUES(34147,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(34147,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(34147,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(34147,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(34147,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(34147,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(34147,36,'cp 61','b7b6 c2c4 f8d6 d4f3 b8c6 b1c3 g8f6 h2h3 e8g8 c1e3 d6c5 d1d2 d7d6 a2a3 c6a5 a1d1 c5e3 d2e3 c8b7 f3d2 a5c6 f2f4 a8d8 c3a4 f6d7 b2b4 h7h6 a4c3 c6e7 d1c1 f8e8 d2b3 b7a8 f1d1 e7g6 b3d4 a8b7 d4e2 d7f6 c3a4 f6d7');
INSERT INTO evaluations VALUES(34170,30,'cp 362',NULL);
INSERT INTO evaluations VALUES(34170,31,'cp 412',NULL);
INSERT INTO evaluations VALUES(34170,32,'cp 401',NULL);
INSERT INTO evaluations VALUES(34170,33,'cp 409',NULL);
INSERT INTO evaluations VALUES(34170,34,'cp 420',NULL);
INSERT INTO evaluations VALUES(34170,35,'cp 435',NULL);
INSERT INTO evaluations VALUES(34170,36,'cp 421','e5d6 d2d4 c7c6 e1g1 g7g6 f3e5 f8g7 c1f4 g8h6 d1d2 h6f5 e5c6 d6c6 e2b5 f5d4 b5c6 b8c6 f4h6 e8g8 h6g7 g8g7 c3e2 e7e5 e2d4 e5d4 c2c3 f8d8 c3d4 c6d4 d2e3 c8g4 f2f3 g4f5 f1d1 d4c2 e3e5 g7g8 a1c1 a8c8 d1d8 c8d8 e5e7 d8c8 e7a7');
INSERT INTO evaluations VALUES(34411,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(34411,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(34411,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(34411,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(34411,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(34411,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(34411,36,'cp 50','c8g4 b2b4 a7a6 a2a4 b8d7 c1e3 f6g6 g1h1 g8f6 b4b5 f8e7 b5c6 b7c6 h2h3 g4e6 c4e6 f7e6 d3d4 g6h5 d4e5 d7e5 f3e5 h5e5 d1d3 e8g8 a1e1 e5h5 f2f4 e6e5 a4a5 h5g6 f4f5 g6f7 e3g1 d6d5 e1b1 d5d4');
INSERT INTO evaluations VALUES(37359,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37359,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37359,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37359,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37359,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(37359,36,'cp 0','c7c5 b1c3 c5d4 d1d4 b8c6 d4d2 h7h6 g5f6 g7f6 f1e2 d8a5 a1c1 f6g7 e1g1 d7d6 f1d1 f8d8 h2h3 c8d7 d2c2 a8c8 d1d5 a5b6 d5d2 b6a5');
INSERT INTO evaluations VALUES(38097,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(38097,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(38097,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38097,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(38097,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(38097,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(38097,36,'cp 35','b1c3 f7f5 f3e5 g8f6 f2f4 e8g8 f1e2 b8d7 e1g1 d6b4 d1b3 b4c3 b2c3 f6e4 e2d3 d7e5 f4e5 b7b6 c4d5 e6d5 d3e4 f5e4 f1f8 d8f8 c1a3 f8f7 a1f1 c8f5 a3d6 h7h6 b3a4 f7d7 c3c4 d5c4 a4c4 f5e6 c4c2 e6d5 g1h1 g8h8 h2h3');
INSERT INTO evaluations VALUES(39988,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(39988,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(39988,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(39988,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(39988,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(39988,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(39988,36,'cp 190','e4e5 f6g8 c1f4 d7d6 e5d6 e7d6 c3e4 d6d5 e4d6 e8f8 d1d2 g8f6 f1d3 f6h5 g1f3 d8e7 e1f1 c8e6 a1e1 h5f4 d2f4 e7f6 f4e3 b8d7 h2h4 h7h5 f3g5 f8g8 g5e6 f7e6 e3e6 f6e6 e1e6 d7f8 e6e2 g7d4 d6b7 h8h7 b7a5 a8c8 g2g3 d4f6 f1g2 h7e7 e2e7 f6e7 a2a3 g8g7 h1d1 f8d7 b2b4 d7e5 f2f4 e5d3');
INSERT INTO evaluations VALUES(40283,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40283,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40283,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40283,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(40283,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(40283,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40283,36,'cp 24','c2c4 e8g8 c4d5 c6d5 b1c3 b8c6 f3e5 c8f5 e5c6 b7c6 c1f4 f6d7 a1c1 e7e5 d4e5 d7e5 b2b3 h7h5 d1d2 f8e8 f1d1 a8c8 f4g5 d8d7 c3a4 d7b7 a4c5 b7b5 a2a4 b5b6 g5e3 e5g4 e3f4 g4e5 a4a5 b6b5 h2h3 e5d7 c5d7 f5d7');
INSERT INTO evaluations VALUES(40279,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40279,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(40279,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(40279,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(40279,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(40279,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40279,36,'cp 13','e8g8 c4d5 c6d5 b1c3 b8c6 c1f4 f6e4 f3e5 e4c3 b2c3 d8a5 d1b3 c6e5 d4e5 c8e6 b3b7 a8b8 b7e7 f8e8 e7h4 a5c3 g2f3 g7e5 a1c1 c3b2 f4e5 b2e5 f1d1 b8b2 h4a4 e8e7 a4a5 e5f6 c1b1 g8g7 b1b2 f6b2 d1d2 b2c1 g1g2 h7h5 h2h4 c1c7 a5c7 e7c7');
INSERT INTO evaluations VALUES(44111,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(44111,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(44111,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(44111,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(44111,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(44111,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(44111,36,'cp 26','d3d4 d8a5 f2f3 d7d6 c1e3 c8e6 d1d2 f8e7 a2a3 e5d4 e3d4 e8g8 g1e2 b8d7 e2f4 b7b5 f1e2 a5c7 e1g1 a7a5 g1h1 f8e8 d4g1 h7h6 h2h3 b5b4 f4e6 f7e6 a3b4 a5b4 c3a2 d6d5 a2b4 a8a1 f1a1 d5e4 f3f4 e7c5 g1c5');
INSERT INTO evaluations VALUES(44504,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(44504,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(44504,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(44504,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(44504,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(44504,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(44504,36,'cp 34','e7e6 e2e4 b8c6 c1e3 d8a5 a2a3 f6e4 g1e2 d7d5 c5d6 e4d6 e2d4 f8e7 d4c6 b7c6 e3f4 d6e4 d1d4 e4f6 f4d6 c8b7 e1c1 a8d8 d4e5 a5e5 d6e5 e8g8 f1d3 c6c5 f2f3 f6d5 c3d5 b7d5 h1e1 h7h5 d3e4 f8e8 e5f4 h5h4 f4c7 d8d7 c7e5 d5e4 e1e4 d7d1 c1d1');
INSERT INTO evaluations VALUES(44152,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(44152,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(44152,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(44152,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(44152,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(44152,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(44152,36,'cp 37','e2e4 b8c6 c1e3 d8a5 a2a3 f6e4 g1e2 d7d5 c5d6 e4d6 e2d4 f8e7 d4c6 b7c6 e3f4 d6e4 d1f3 f7f5 f1c4 e4c3 b2c3 a5c5 f3d3 e6e5 f4e3 c5d6 e1g1 d6d3 c2d3 f5f4 e3d2 e7d6 f1e1 e8d8 d3d4 h8e8 d4e5 d6e5 c4d3 g7g6 g2g3 e5d6 e1e8 d8e8 d2f4 d6f4 g3f4 a8b8 a1b1 b8b1 d3b1 c8h3 b1e4 c6c5 f2f3 e8e7 e4c6 e7f6 g1f2 h3e6 c6e4 e6d7 f2g3');
INSERT INTO evaluations VALUES(45666,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(45666,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(45666,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(45666,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45666,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(45666,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(45666,36,'cp 30','c6c5 b1d2 d8b6 d1c2 f6h5 f4e5 b8c6 a2a4 c5d4 a4a5 b6d8 e3d4 a7a6 f1d3 c8d7 e1g1 c6e5 f3e5 h5f4 e5d7 d8d7 g2g3 f4d3 c2d3 f8d6 b2b4 a8c8 f1e1 e8g8 e1e3 c8c7 d2b3 c7c4 h2h4 d7c7 b3c5 e6e5 d4e5 d6c5 b4c5 c4c5 g1g2 f8c8 a1e1');
INSERT INTO evaluations VALUES(46719,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(46719,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(46719,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46719,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(46719,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46719,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(46719,36,'cp 27','e7e5 d4e5 d6e5 g1f3 c8g4 b1c3 b8c6 e1g1 e5f4 c1f4 f6h5 f4e3 f8g7 h2h3 g7d4 c3d5 d4e3 d5e3 g4f3 d1f3 e8g8 f3f2 d8g5 h3h4 g5g3 e3d5 g8g7 a1d1 a8d8 f2g3 h5g3 f1e1 g3f5 e4f5 d8d5 d3e4 d5d1 e1d1 c6e5 g1f2 c7c6 f5g6 h7g6 e4f3 f8e8 b2b3');
INSERT INTO evaluations VALUES(48320,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(48320,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(48320,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(48320,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(48320,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(48320,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(48320,36,'cp -52','b8d7 c1f4 f8b4 f3d2 e8g8 e2e3 f8e8 c3b5 e6e5 d4e5 d7e5 c2c3 b4f8 d2f3 f6d7 f1e2 a7a6 f3e5 d7e5 b5d4 c7c5 d4f3 e5c6 e1g1 h7h6 a2a3 c8f5 b2b4 a8c8 b4c5 c6a5 f3d4 f5e4 c5c6 a5c6 e2g4 c8a8 g4f3 d8f6 f4g3 e4f3 d1f3 f6f3 d4f3');
INSERT INTO evaluations VALUES(49344,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(49344,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(49344,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(49344,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(49344,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(49344,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(49344,36,'cp -69','b7b6 c5b6 a7b6 e2e3 f6e4 b1d2 f8b4 g1f3 d7d5 a2a3 e8g8 a1b1 b4d6 d2e4 d5e4 f3d2 c8b7 d1c2 d8h4 d2c4 d6e7 f1e2 b8d7 c1d2 b7d5 b1c1 c7c6 e1g1 b6b5 c4a5 f8c8 d2b4 e7b4 a3b4 h4e7');
INSERT INTO evaluations VALUES(49904,30,'cp 250',NULL);
INSERT INTO evaluations VALUES(49904,31,'cp 251',NULL);
INSERT INTO evaluations VALUES(49904,32,'cp 245',NULL);
INSERT INTO evaluations VALUES(49904,33,'cp 250',NULL);
INSERT INTO evaluations VALUES(49904,34,'cp 245',NULL);
INSERT INTO evaluations VALUES(49904,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(49904,36,'cp 249','h2h3 g4f6 f3e5 f8b4 e5d3 b4c3 d2c3 d8e7 e4e5 d7d6 d1e2 d6e5 d3e5 c8e6 c1g5 b8d7 e5d7 e7d7 e2e3 f6d5 e3d4 f7f6 g5d2 d5e7 d4d7 e6d7 e1c1 e8c8 h3h4 h7h5 f1d3 h8e8 h1e1 d7g4 f2f3 g4d7 d2f4 e7d5 f4h2 d5e3 d1d2 e3f5 e1e8 d8e8');
INSERT INTO evaluations VALUES(50434,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(50434,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(50434,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(50434,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(50434,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(50434,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(50434,36,'cp 36','f4g5 h7h6 g5h4 h5f6 e2e3 e7e6 f1d3 f8e7 g1f3 d7d5 e1g1 b8d7 c2c4 c7c5 b1c3 d5c4 d3c4 e8g8 f1e1 c5d4 e3d4 f6d5 h4g3 d7f6 f3e5 c8d7 c4d5 f6d5 c3d5 e6d5 d1b3 d7c6 e5c6 b7c6 a1d1 e7f6');
INSERT INTO evaluations VALUES(51436,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51436,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(51436,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(51436,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(51436,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(51436,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51436,36,'cp 35','f1g2 c8b7 c2c4 c7c5 b1c3 g8f6 e1g1 f8e7 d2d4 c5d4 d1d4 e8g8 f1d1 d7d6 e2e4 b8d7 b2b3 a8c8 c1a3 d7c5 f3e1 a7a6 d4e3 d8c7 a1c1 f8e8 e1d3 f6d7 f2f4 c5d3 d1d3 b7a8 c1d1 d7c5 d3d2 c7b8 e4e5 d6e5');
INSERT INTO evaluations VALUES(51437,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(51437,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(51437,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(51437,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(51437,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(51437,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(51437,36,'cp 32','c8b7 c2c4 c7c5 b1c3 g8f6 e1g1 f8e7 f1e1 d7d5 c4d5 f6d5 c3d5 d8d5 d2d4 e8g8 e2e4 d5h5 c1f4 c5d4 f3d4 h5d1 e1d1 b8c6 d4c6 b7c6 e4e5 c6g2 g1g2 a8d8 f4e3 f7f6 d1d8 f8d8 a1c1 d8d7 e5f6 g7f6 b2b3 g8f8 h2h3 h7h5 g2f3 f8f7 c1c8');
INSERT INTO evaluations VALUES(51783,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(51783,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(51783,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(51783,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(51783,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51783,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(51783,36,'cp 33','e2e4 d7d5 e4e5 d5d4 g2g3 g8e7 d2d3 b8d7 d1e2 e7g6 f3d4 d7e5 d4c2 c6c5 f2f4 e5c6 f1g2 f8e7 e1g1 a8b8 b1c3 e8g8 c1e3 c6d4 e2f2 d4c2 f2c2 b7b6 a1d1 c8b7 d3d4 c5d4 e3d4 d8c8 b2b3 b7g2 g1g2 f8d8');
INSERT INTO evaluations VALUES(51778,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(51778,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(51778,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(51778,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(51778,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(51778,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(51778,36,'cp 31','d7d5 e4e5 d5d4 d2d3 g8e7 d1e2 e7g6 g2g3 b8d7 f3d4 d7e5 d4c2 c6c5 f2f4 e5c6 f1g2 f8e7 e1g1 e8g8 c1e3 d8d7 b1c3 b7b6 a1d1 c8b7 d3d4 c5d4 c2d4 c6d4 e3d4 b7g2 e2g2 d7c7 f4f5 e6f5 b2b3 e7c5 d4c5 c7c5 g2f2 c5f2 f1f2 f5f4 g3f4 a8d8 c3d5');
INSERT INTO evaluations VALUES(53559,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(53559,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53559,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(53559,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(53559,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(53559,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(53559,36,'cp 30','e4e5 c6c5 c4d5 e6d5 f1b5 b8c6 d2d4 c5d4 f3d4 f8b4 b1c3 g8e7 c1g5 c8d7 e1g1 c6d4 g5e7 d8e7 b5d7 e7d7 d1d4 b4c3 b2c3 b7b6 a1d1 a8c8 f2f4 c8c4 d4d2 e8g8 f4f5 f8e8 e5e6 d7c6 d2d5 c6d5 d1d5 f7e6 f5e6 c4c3');
INSERT INTO evaluations VALUES(52351,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(52351,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(52351,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(52351,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(52351,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(52351,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(52351,36,'cp -90','c6d5 d1d5 g8f6 d5b3 d7d5 g1f3 b8c6 c1f4 e7e6 e2e3 f6e4 c2c4 g7g5 f4g3 d8a5 b1c3 f8g7 a1c1 e8g8 c4d5 e6d5 f1e2 f8d8 b3b5 a5b5 e2b5 e4g3 h2g3 g5g4 f3d4 c6d4 e3d4 g7d4 b5d3 c8e6 h1h7 a8c8 h7h5 g8g7 c1c2 d8h8 h5h8 c8h8 c3b5 d4b6 b5d6');
INSERT INTO evaluations VALUES(52498,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(52498,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(52498,32,'cp -91',NULL);
INSERT INTO evaluations VALUES(52498,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(52498,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(52498,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(52498,36,'cp -90','d1d5 g8f6 d5b3 d7d5 g1f3 a7a6 c1f4 b8d7 b1d2 d7c5 b3a3 e7e6 e2e3 c8d7 f1e2 b7b5 b2b4 c5a4 c2c3 f8e7 e1g1 e8g8 f4e5 d7e8 a1c1 e7d6 e2d1 a4b6 e5d4 a6a5 a3b2 d8c7 d4f6 g7f6 c3c4 d5c4');
INSERT INTO evaluations VALUES(52497,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(52497,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(52497,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(52497,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(52497,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(52497,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(52497,36,'cp -99','g8f6 d5b3 b8c6 g1f3 d7d5 c1f4 e7e6 b1d2 f6d7 f4g5 f7f6 g5e3 a7a5 a2a4 f8c5 e3c5 d7c5 b3a3 b7b6 c2c3 e8g8 d2b3 c5e4 e2e3 e6e5 f1e2 c8e6 e1g1 h7h6 a1d1 f6f5 b3d2 e4c5 b2b4 a5b4 c3b4 a8a4');
INSERT INTO evaluations VALUES(54506,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(54506,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(54506,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(54506,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(54506,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(54506,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(54506,36,'cp 13','c5d4 d1d4 g8f6 c4d5 d8d5 b1c3 d5e6 c1e3 b8c6 f1b5 f8e7 e1g1 e8g8 d4a4 c8d7 f3d4 c6d4 e3d4 a7a6 b5d7 e6d7 a4d7 f6d7 c3d5 e7c5 f1d1 f8d8 d4c5 d7c5 a1c1 c5e6 f2f4 g7g6 d5b6 d8d1 c1d1 a8e8 f4f5 e6g5 h2h4 g5e4 f5g6 h7g6');
INSERT INTO evaluations VALUES(55917,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(55917,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(55917,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(55917,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(55917,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(55917,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(55917,36,'cp 39','e4d5 d8d5 b1c3 d5a5 d2d4 c8d7 c1d2 b8c6 g1f3 e8c8 e1c1 g8f6 a2a3 a7a6 e2c4 d7e8 f1e2 a5b6 d2e3 f6d5 c3d5 e6d5 c4c3 f8d6 g2g3 c8b8 c1b1 f7f6 f3d2 c6a5 e2d3 e8g6 b2b4 b6c6 c3c6 a5c6 d2b3 g6d3');
INSERT INTO evaluations VALUES(55785,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(55785,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(55785,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(55785,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(55785,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(55785,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(55785,36,'cp 37','d8d5 g1f3 g8f6 b1c3 d5a5 d2d4 f8b4 c1d2 e8g8 a2a3 b7b6 e2e3 c8a6 a1c1 b4c3 d2c3 a5f5 f1a6 b8a6 c3d2 c7c5 e1g1 f5e4 d4c5 a6c5 b2b4 c5d7 c2c4 a8c8 e3e4 f6e4 d2e3 f8d8 f1d1 h7h6 h2h3 e4f6');
INSERT INTO evaluations VALUES(34836,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(34836,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(34836,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(34836,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(34836,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(34836,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(34836,36,'cp 31','b1c3 d5a5 d2d4 g8f6 c1d2 b8c6 g1f3 c8d7 a2a3 a7a6 g2g3 e8c8 f1g2 a5b6 c3a4 b6a7 d2e3 f6d5 a4c3 d5e3 f2e3 f7f6 e1c1 c8b8 d1d3 f8d6 h1d1 e6e5 d4e5 c6e5 f3e5 f6e5 e2h5 g7g6 h5g5 a7b6 c3e4 d7e6 e4d6 c7d6 g2d5 e6d5 d3d5 h8e8');
INSERT INTO evaluations VALUES(34830,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(34830,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(34830,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(34830,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(34830,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(34830,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(34830,36,'cp 39','d5f5 g1f3 c8d7 d2d4 g8f6 a2a3 d7c6 f3e5 f8d6 e5c6 b8c6 c1e3 f5a5 e2d3 c6e7 e1c1 c7c6 e3d2 a5f5 f2f3 f5d3 f1d3 e8c8 c1b1 d6c7 c3e2 e7f5 c2c3 h7h5 d2g5 c8d7 d3f5');
INSERT INTO evaluations VALUES(56483,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(56483,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(56483,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(56483,33,'cp 167',NULL);
INSERT INTO evaluations VALUES(56483,34,'cp 157',NULL);
INSERT INTO evaluations VALUES(56483,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(56483,36,'cp 162','f3e5 f8d6 d2d4 b8d7 e5d7 c8d7 f1g2 c7c6 e1g1 g8f6 d1d3 d7e6 c1g5 e8g8 b1d2 h7h6 g5f6 d8f6 e2e4 d5e4 d2e4 f6d8 e4d6 d8d6 c2c4 d6b4 b2b3 f8d8 a2a3 b4b6 d4d5 d8d7 f1e1 c6d5 c4d5 a8c8 a1d1 c8d8');
INSERT INTO evaluations VALUES(56799,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(56799,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(56799,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(56799,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(56799,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(56799,35,'cp 167',NULL);
INSERT INTO evaluations VALUES(56799,36,'cp 156','e2e4 f8b4 f1d3 g8e7 e1g1 e8g8 c3e2 b8c6 c2c3 b4d6 f1e1 a7a5 a2a3 g8h8 d1c2 d8e8 e4d5 e6d5 c1f4 c8g4 f4d6 c7d6 f3d2 g4d7 b2b4 a5b4 a3b4 a8a1 e1a1 e7g6 h2h3 f6f5 c2b2 f5f4 d2f3 d7f5 e2c1 h7h6 d3f5 f8f5');
INSERT INTO evaluations VALUES(56801,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(56801,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(56801,32,'cp 138',NULL);
INSERT INTO evaluations VALUES(56801,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(56801,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(56801,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(56801,36,'cp 140','f8b4 f1d3 g8e7 e1g1 e8g8 d1e2 b4a5 f1e1 c7c5 e4d5 e7d5 c1d2 b8c6 c3d5 e6d5 d4c5 c8g4 c2c3 c6e5 d2f4 f8e8 e2d1 a5c7 f4g3 d8d7 d3e2 e5f3 e2f3 e8e1 d1e1 g4f3 g2f3 b7b6 g3c7 d7c7 c5b6 c7b6 e1d2 a8d8 a1d1 f6f5 h2h4 b6g6 d2g5 d8d6 g5g3 d6b6 d1d5');
INSERT INTO evaluations VALUES(975,30,'cp -888',NULL);
INSERT INTO evaluations VALUES(975,31,'cp -888',NULL);
INSERT INTO evaluations VALUES(975,32,'cp -888',NULL);
INSERT INTO evaluations VALUES(975,33,'cp -898',NULL);
INSERT INTO evaluations VALUES(975,34,'cp -913',NULL);
INSERT INTO evaluations VALUES(975,35,'cp -901',NULL);
INSERT INTO evaluations VALUES(975,36,'cp -882','d5g2 d1f3 g2g5 c3e2 g5g6 d2d3 c6b4 e1d2 c8d7 a2a3 d7c6 f3e3 c6h1 e3e5 e8d7 a3b4 f8b4 c2c3 b4d6 e5e3 a8e8 e3h3 f7f5 e2g3 h1f3 d2c2 g6g4 h3g4 f5g4 c1e3 h7h5 c3c4 h5h4 g3f1 b7b6 d3d4 f3e4 c2c3 e4g2 c4c5 d6h2 c3d3 g2f1 a1f1 h4h3 e3d2 b6c5 d4c5');
INSERT INTO evaluations VALUES(999,30,'cp -898',NULL);
INSERT INTO evaluations VALUES(999,31,'cp -894',NULL);
INSERT INTO evaluations VALUES(999,32,'cp -894',NULL);
INSERT INTO evaluations VALUES(999,33,'cp -901',NULL);
INSERT INTO evaluations VALUES(999,34,'cp -898',NULL);
INSERT INTO evaluations VALUES(999,35,'cp -888',NULL);
INSERT INTO evaluations VALUES(999,36,'cp -894','d1f3 g2g5 c3e2 g5g6 d2d3 c6b4 e1d2 c8d7 a2a3 d7c6 f3e3 c6h1 a3b4 f8b4 c2c3 b4d6 b2b3 e8g8 c1b2 h1c6 c3c4 f8e8 a1g1 g6h5 h2h3 f7f6 e2c3 e5e4 d3d4 a8d8 d2c2 h5f5 c3e2 f5f3 g1f1 f6f5 e3f3 e4f3');
INSERT INTO evaluations VALUES(1156,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(1156,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(1156,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(1156,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(1156,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(1156,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(1156,36,'cp 94','b5c6 b6c6 e1g1 e7e6 f1e1 c6c7 d2d4 c5d4 d1d4 g8e7 a2a4 d7d6 a4a5 c8d7 c3a4 d7a4 d4a4 c7c6 a4b3 e7g6 a1a4 f8e7 a4b4 a8b8 b4b6 c6c7 c1e3 e8g8 c2c3 f8d8 e1d1 g6e5 f3e5 d6e5 h2h3 d8d1 b3d1 b8d8 d1b3 e7c5 b6b7 c7a5');
INSERT INTO evaluations VALUES(1155,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(1155,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(1155,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(1155,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(1155,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(1155,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(1155,36,'cp 81','b6c6 e1g1 e7e6 f1e1 c6c7 d2d4 c5d4 d1d4 g8e7 a2a4 d7d6 c1g5 f7f6 g5e3 c8d7 a4a5 e7c6 d4c4 a8c8 e3b6 c7b8 c4b3 f8e7 c3e2 e8f7 f3d4 e7d8 e1d1 d8b6 a5b6 c6d4 d1d4 c8c5 a1d1 c5b5 b3d3 b5b2 d4d6 d7c6 e2f4');
INSERT INTO evaluations VALUES(1153,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(1153,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(1153,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(1153,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(1153,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(1153,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(1153,36,'cp 105','e1g1 e7e6 f1e1 c6c7 d2d4 c5d4 f3d4 f8c5 d4f5 e8f8 f5g3 b7b5 e4e5 d7d5 e5d6 c5d6 a2a4 b5b4 c3e4 d6e7 c1g5 c8b7 d1g4 h7h5 g5e7 g8e7 g3h5 g7g6 h5g3 b7e4 g4e4 e7d5 h2h3 a8c8 e4d4 f8g8 g3e4 h8h5 c2c3 b4c3');
INSERT INTO evaluations VALUES(1152,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(1152,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(1152,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(1152,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(1152,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(1152,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(1152,36,'cp 109','d7d6 d2d4 c8g4 d4c5 c6c5 h2h3 g4f3 d1f3 e7e6 c1e3 c5c7 e3d4 a8c8 a1d1 g8e7 f3e3 b7b5 a2a4 b5b4 c3e2 e7c6 c2c3 c6d4 c3d4 f8e7 d4d5 e8g8 d5e6 f7e6 e2d4 c7d7 e3b3 d6d5 b3d3 d7d6 e4d5 e6d5 d4f5 d6f6 d3d5 g8h8 f5e7 f6e7 d5a5 c8c2 a5a6');
INSERT INTO evaluations VALUES(2646,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(2646,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(2646,32,'cp 106',NULL);
INSERT INTO evaluations VALUES(2646,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(2646,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(2646,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(2646,36,'cp 117','g2g4 d7d5 e4d5 c6d4 g4g5 h6g5 c1g5 d8d6 f3d2 b7b5 c4a2 c5b6 h1g1 c8f5 g5f6 d6f6 c3e4 f5e4 d2e4 f6f4 g1g3 f4h6 d5d6 c7d6 c2c3 d6d5 a2d5 a8d8 d5a2 d4e6 a2e6 f7e6 d1e2 f8f4 g3f3 f4f8 f3f8 d8f8 e2g4 f8f4 g4g2 h6h5 e1d2');
INSERT INTO evaluations VALUES(2802,30,'cp 401',NULL);
INSERT INTO evaluations VALUES(2802,31,'cp 420',NULL);
INSERT INTO evaluations VALUES(2802,32,'cp 397',NULL);
INSERT INTO evaluations VALUES(2802,33,'cp 409',NULL);
INSERT INTO evaluations VALUES(2802,34,'cp 433',NULL);
INSERT INTO evaluations VALUES(2802,35,'cp 405',NULL);
INSERT INTO evaluations VALUES(2802,36,'cp 428','e4e5 b6d4 e5f6 d4f6 b1c3 c6a5 b3c2 c8b7 d1d3 g7g6 c1h6 a5c4 b2b3 c4a3 h6f8 d8f8 a1c1 a8d8 c3d5 f6b2 d3d2 b7d5 d2d5 b2c1 f1c1 c7c6 d5g5 a3c2 c1c2 d8e8 h2h4 e8e6 c2d2 d7d5 g5f4 e6e4 f4g3 f8e7 g1h2 g8g7');
INSERT INTO evaluations VALUES(2801,30,'cp 435',NULL);
INSERT INTO evaluations VALUES(2801,31,'cp 424',NULL);
INSERT INTO evaluations VALUES(2801,32,'cp 400',NULL);
INSERT INTO evaluations VALUES(2801,33,'cp 431',NULL);
INSERT INTO evaluations VALUES(2801,34,'cp 406',NULL);
INSERT INTO evaluations VALUES(2801,35,'cp 420',NULL);
INSERT INTO evaluations VALUES(2801,36,'cp 428','b6d4 e5f6 d4f6 b1c3 c6a5 b3c2 g7g6 f1e1 c8b7 c1h6 f8e8 e1e8 d8e8 d1d3 d7d5 a1e1 e8d7 b2b4 d7c6 c3d5 c6d5 d3e3 d5e6 e3f4 e6d6 f4d6 c7d6 b4a5 b7d5 c2b3 d5c4 h6e3 a8e8 e1c1 d6d5 h2h4 f6b2 c1d1 d5d4 e3d4 b2d4 d1d4 e8e2 g1f1 e2b2 b3c4 b5c4 d4c4 b2a2');
INSERT INTO evaluations VALUES(5627,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(5627,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(5627,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5627,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(5627,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(5627,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(5627,36,'cp 24','c6a5 c4d3 c7c5 c3d5 g8f6 d5e7 d8e7 c2c3 a5c6 d3c2 h7h6 d2d3 c8d7 a2a3 a7a6 b2b4 e8g8 c1e3 b7b6 e1g1 d6d5 f1e1 d5d4 c3d4 c5d4 e3d2 b6b5 a1c1 d7e6 c2b3 f8c8 f3h4 e7d7 f2f4 e6b3 d1b3');
INSERT INTO evaluations VALUES(8528,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(8528,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(8528,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(8528,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(8528,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(8528,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(8528,36,'cp -37','e7g5 c4d5 e8g8 c1g5 d8g5 f3e3 g5f6 b1c3 c6d4 e1c1 c7c6 d5f3 c8d7 c1b1 a8e8 c3e4 f6d8 h2h4 d8a5 c2c3 d4f3 g2f3 g8h8 b2b3 a5a3 e3c5 a3c5 e4c5 d7c8 c5e4 f7f5 e4d6 e8e6 d6c8 f8c8');
INSERT INTO evaluations VALUES(9179,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(9179,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(9179,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(9179,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(9179,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(9179,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(9179,36,'cp -39','c4d5 e8g8 c1g5 d8g5 f3e3 g5f6 b1a3 c6d4 c2c3 f6d8 d5c4 d4f5 e3f3 c7c6 c4b3 a7a5 e1g1 a5a4 b3c2 f5h4 f3g3 c8e6 a3c4 e6d5 c4e3 h4g6 f1e1 d5e6 e3c4 e6c4 d3c4 d8d2 g3d3 d2d3 c2d3 f7f5 c4c5 e5e4 d3c2 g6e5 f2f3 f8d8 a1d1 d8d5 d1d5 c6d5');
INSERT INTO evaluations VALUES(9519,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(9519,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(9519,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(9519,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(9519,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(9519,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(9519,36,'cp -20','d8e7 c1e3 c8e6 d1e2 h7h6 a2a3 f6d7 f3d2 a7a5 h2h3 h8g8 d2f3 e8c8 e1c1 g7g5 d3d4 e5d4 f3d4 d7b6 d4e6 e7e6 e3b6 c7b6 c1b1 b6b5 e2e3 c8b8 d1d3 d6e5 h1d1 d8d3 e3d3 g5g4 h3g4 e6g4');
INSERT INTO evaluations VALUES(9881,30,'cp 288',NULL);
INSERT INTO evaluations VALUES(9881,31,'cp 292',NULL);
INSERT INTO evaluations VALUES(9881,32,'cp 266',NULL);
INSERT INTO evaluations VALUES(9881,33,'cp 300',NULL);
INSERT INTO evaluations VALUES(9881,34,'cp 279',NULL);
INSERT INTO evaluations VALUES(9881,35,'cp 296',NULL);
INSERT INTO evaluations VALUES(9881,36,'cp 305','d2d4 e5d4 h2h3 g4e5 f3e5 c6e5 c3d4 e5c4 d4c5 e8g8 d1d3 c4e5 d3g3 e5g6 f2f4 d8h4 g3h4 g6h4 f4f5 b7b6 c1f4 b6c5 f4c7 d7d5 c7g3 g7g5 f5g6 h4g6 e4d5 f7f5 g3f2 c5c4 b1c3 f5f4 f2c5 f8d8 a1d1 c8f5 d5d6 f4f3 g2g4 f5d7 c3d5 a8b8 d1d2 d7c6 d5e7 g6e7 e1e7 d8d7 g1f2 d7e7 d6e7 g8f7');
INSERT INTO evaluations VALUES(9885,30,'cp 315',NULL);
INSERT INTO evaluations VALUES(9885,31,'cp 307',NULL);
INSERT INTO evaluations VALUES(9885,32,'cp 305',NULL);
INSERT INTO evaluations VALUES(9885,33,'cp 306',NULL);
INSERT INTO evaluations VALUES(9885,34,'cp 287',NULL);
INSERT INTO evaluations VALUES(9885,35,'cp 306',NULL);
INSERT INTO evaluations VALUES(9885,36,'cp 322','e5d4 h2h3 g4f2 g1f2 d4c3 f2g3 c3b2 c1b2 e8g8 b1c3 d7d6 a1c1 c8e6 c3d5 c6e5 c4b3 b7b6 g3h2 a7a5 a2a4 c7c6 d5f4 e6b3 d1b3 d8f6 b2e5 d6e5 f4d3 c5b4 d3b4 a5b4 b3b4 c6c5 b4c4 f8d8 e1f1 d8d6 c1d1 f6e6 d1d5 d6d5 e4d5');
INSERT INTO evaluations VALUES(9887,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(9887,31,'cp 321',NULL);
INSERT INTO evaluations VALUES(9887,32,'cp 332',NULL);
INSERT INTO evaluations VALUES(9887,33,'cp 310',NULL);
INSERT INTO evaluations VALUES(9887,34,'cp 308',NULL);
INSERT INTO evaluations VALUES(9887,35,'cp 310',NULL);
INSERT INTO evaluations VALUES(9887,36,'cp 286','h2h3 g4e5 f3e5 c6e5 c3d4 e5c4 d4c5 e8g8 d1d5 c4a5 e1e3 a5c6 e3g3 d8h4 b2b3 c6e7 d5d2 g8h7 c1b2 f8g8 g3f3 h4h5 b1c3 b7b6 a1e1 f7f6 c3d5 e7d5 e4d5 g8e8 d2d3 h7g8 e1e8 h5e8 f3g3 g8h8 d3f3 e8f8 d5d6 c8a6 d6c7 b6c5 f3d5 a8e8 d5d7');
INSERT INTO evaluations VALUES(11416,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(11416,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(11416,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(11416,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(11416,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(11416,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(11416,36,'cp 189','f3g5 d6d5 c3d5 h7h6 g5f3 f8d6 a2a4 f6d5 c4d5 e8g8 e1g1 c6e7 d5b3 e7g6 c1e3 a6a5 f3d2 b7b6 d2c4 c8e6 c2c3 c7c5 c4a3 d8e7 b3e6 e7e6 f2f3 f7f5 a3c4 d6c7 d1e2 g6f4 e3f4 e5f4 f1e1 a8d8 e4e5 f8e8 a1d1 g7g6 d3d4 c5d4');
INSERT INTO evaluations VALUES(11884,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(11884,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(11884,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(11884,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(11884,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(11884,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(11884,36,'cp 2','b5c4 d4e5 d8d5 e1g1 c8b7 g5f3 f6d7 b2b3 e8c8 f1e1 d5d3 b1a3 b7f3 d1f3 d3f3 g2f3 c4b3 a2b3 d7c5 a3c4 c5d3 e1e2 f8c5 c1b2 h8e8 b2c3 f7f6 g1f1 f6e5 c4e5 d3e5 c3e5 e8e5 e2e5 c5d4 e5e1 d4a1 e1a1 d8d2 a1a7 d2d3 a7a8 c8d7 a8g8 g7g6 g8g7 d7e6 g7c7 d3b3 f1g2 h7h6 c7c6 e6f5');
INSERT INTO evaluations VALUES(12935,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(12935,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(12935,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(12935,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(12935,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(12935,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(12935,36,'cp 120','d1d4 d7d6 c1g5 c8d7 e1c1 d8a5 f2f4 h7h6 g5f6 g7f6 f1c4 e7e6 a2a3 f8e7 c1b1 e8c8 d4d3 c8b8 c3e2 d8c8 e2d4 a5b6 c4b3 a7a5 a3a4 e6e5 d4e2 d7e6 e2c3 c8c5 c3d5 e6d5 b3d5 e5f4');
INSERT INTO evaluations VALUES(13686,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(13686,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(13686,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(13686,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(13686,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(13686,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(13686,36,'cp 35','f8e7 f1e1 e8g8 a2a3 b7b6 c4a2 c8b7 c1f4 f8e8 f4g3 h7h6 a1c1 d8d7 d1e2 f6h5 e1d1 a8d8 e2b5 h5g3 h2g3 e7f8 b5a4 c6e7 a4a7 b7f3 a7d7 d8d7 g2f3 e8d8 d4d5 e7d5 c3d5 e6d5 f3f4 d5d4 d1d3 g7g6 g1g2 g8g7 b2b4 g6g5 a2b3 g5f4');
INSERT INTO evaluations VALUES(8106,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(8106,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(8106,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(8106,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(8106,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(8106,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(8106,36,'cp 30','f1e1 e8g8 a2a3 b7b6 d1d3 c8b7 c4a2 f8e8 c1f4 e7d6 f4g5 h7h6 g5h4 g7g5 h4g3 d6g3 h2g3 g5g4 f3h2 d8d4 d3e2 d4c5 h2g4 f6g4 e2g4 c5g5 g4g5 h6g5 c3e4 e8d8 a1d1 c6d4 e4g5 d4c2 d1d8 a8d8 e1c1 d8d2 a2b1 c2d4 c1c7 b7d5 b1e4 d5e4 g5e4 d2b2 c7a7 b2c2');
INSERT INTO evaluations VALUES(18208,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(18208,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(18208,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(18208,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(18208,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(18208,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(18208,36,'cp 0','e7e5 f1e1 f8e7 c3e2 g8f6 e2g3 e8g8 c2c3 f8e8 d2d4 c5d4 c3d4 e5d4 f3d4 e7f8 d4c6 d8b6 c6d4 c8b7 c1e3 f6e4 d4f5 b6c6 d1g4 e4f2 f5h6 g8h8 h6f7 h8g8 f7h6 g8h8');
INSERT INTO evaluations VALUES(18823,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(18823,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18823,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(18823,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18823,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18823,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18823,36,'cp 0','d7d5 e4d5 e6d5 e1g1 f8d6 d2d4 c5d4 f3d4 g8f6 c1e3 a7a6 b1d2 e8g8 f1e1 f8e8 h2h3 e8e3 f2e3 d6g3 e1f1 d8e7 f1f3 g3c7 d1b3 e7d6 f3f4 d6e7 f4f3');
INSERT INTO evaluations VALUES(20094,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(20094,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(20094,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(20094,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(20094,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(20094,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(20094,36,'cp 54','e1g1 g8f6 c3c4 g7g6 b1c3 f8g7 c1f4 c8g4 a1c1 e8g8 h2h3 g4f3 d1f3 f6d7 f3e2 d7c5 d3b1 g7c3 c1c3 e7e5 f4e3 c5e6 e2d2 d8f6 c3d3 f8d8 g2g3 a8c8 f2f4 e5f4 g3f4 e6c5 d3a3 a7a5 d2g2 f6e7 f4f5 g8h8 f5g6 f7g6');
INSERT INTO evaluations VALUES(20093,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(20093,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(20093,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(20093,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(20093,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(20093,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(20093,36,'cp 57','g7g6 d1b3 g8f6 f1d1 f6d7 d3c4 d7e5 f3e5 c6e5 c4e2 f8g7 c3c4 d8b6 c1e3 b6b3 a2b3 e5g4 e3c1 a7a6 b1c3 c8e6 f2f4 g4f6 c1e3 e8g8 h2h3 f6d7 g2g4 g7c3 b2c3 f7f6 h3h4 a8c8 g1f2 d7c5 e3c5 c8c5 f2e3 g8g7 g4g5 f6g5 h4g5 a6a5');
INSERT INTO evaluations VALUES(20179,30,'cp 169',NULL);
INSERT INTO evaluations VALUES(20179,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(20179,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(20179,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(20179,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(20179,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(20179,36,'cp 159','e4d5 e6d5 d4c5 c8e6 c1e3 g8e7 g1f3 e7f5 d1e2 f5e3 e2e3 f8e7 f3d4 d8d7 b2b4 e7f6 a1d1 e8g8 e1g1 d7c8 b5c6 b7c6 h2h3 f8e8 e3d3 c8c7 a2a3 h7h6 f1e1 a7a5 d1b1 e6d7 e1e8 a8e8 c3e2 g7g6 d4f3 d7f5 d3d2 f5e4 e2d4 f6g7 b1e1 e8d8 b4a5');
INSERT INTO evaluations VALUES(21141,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(21141,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(21141,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(21141,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(21141,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(21141,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(21141,36,'cp 103','e2e3 d4e3 f2e3 g8f6 f1b5 c8d7 e1g1 g7g6 d2d4 f8g7 b1d2 e8g8 d1e2 a7a6 b5c6 d7c6 f3e5 d8d6 c2c4 f6e4 e5c6 e4d2 c6e7 d6e7 e2d2 f8e8 a1e1 f7f5 b2c3 g7h6 f1f3 a8d8 d4d5 c7c6 c3b4 e7e5 f3h3 h6g7 d5d6 c6c5 d6d7 e8e7');
INSERT INTO evaluations VALUES(21140,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(21140,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(21140,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(21140,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(21140,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(21140,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(21140,36,'cp 113','d4e3 f2e3 g8f6 f1b5 c8d7 d1e2 a7a6 b5c6 d7c6 b1c3 d8d6 e1c1 e8c8 d2d4 f6e4 f3e5 e4c3 b2c3 c6e8 d4d5 e7e6 e5c4 d6e7 d5d6 c7d6 c3a5 e8c6 c4e5 c6e8 c1b1 c8b8 e5c4 e8c6 a5d8 e7d8 e2d2 f8e7 c4d6 d8c7 e3e4 f7f6 e4e5 f6e5 d6f7 h8f8 f7g5 f8f6 h1f1 e5e4 f1f6 e7f6 g5e6');
INSERT INTO evaluations VALUES(21833,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(21833,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(21833,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(21833,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(21833,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(21833,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(21833,36,'cp 46','f8b4 b1c3 g8f6 e1g1 e8g8 h2h3 h7h6 c1f4 a7a6 f1e1 b7b5 c4b3 c6a5 b3c2 c8b7 a2a3 b4c3 b2c3 f8e8 f4e5 a5c4 d1d3 c4e5 f3e5 c7c5 e5g4 f6e4 f2f3 c5c4 d3e3 e4g5 e3e8 d8e8 e1e8 a8e8 g1f2 g5e6 h3h4 e6f4 g4e3 f4d5 c2e4 d5e3 e4b7 e3c2 a1c1 c2a3 b7a6');
INSERT INTO evaluations VALUES(22121,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(22121,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(22121,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(22121,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(22121,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(22121,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(22121,36,'cp 93','f3d4 e5d4 d1d4 g8f6 b1c3 f8e7 c1f4 e8g8 e1c1 c8e6 f2f3 f6d7 c3d5 f8e8 d5e7 d8e7 h1e1 e6c4 d4c4 d7c5 c1b1 h7h6 b2b3 a7a6 c4c3 a8d8 f4e3 c5e6 g2g3 c7c6 a2a4 d6d5 e3b6 e6c7 b6c7 e7c7 e4d5 e8e1');
INSERT INTO evaluations VALUES(22116,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(22116,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(22116,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(22116,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(22116,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(22116,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(22116,36,'cp 80','e5d4 d1d4 g8f6 b1c3 f8e7 c1g5 c7c6 e1c1 d8b6 g5f6 g7f6 d4d2 h7h5 f2f4 c8g4 d1e1 b6c5 c4e2 g4d7 h1f1 a7a6 a2a3 e8c8 c3a4 c5a7 f4f5 b7b5 a4c3 h5h4 g2g4 h4g3 h2g3 h8g8 d2f4 a7c5 c1b1 c8b8');
INSERT INTO evaluations VALUES(22135,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(22135,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22135,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(22135,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(22135,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(22135,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(22135,36,'cp 86','d1d4 g8f6 b1c3 f8e7 c1f4 e8g8 e1c1 c8e6 f2f3 f6d7 c3d5 f8e8 c1b1 a7a6 d5e7 d8e7 h1e1 e6c4 d4c4 d7c5 f4e3 h7h6 b2b3 e7h4 h2h3 c5e6 c4c3 h4g3 e1e2 g3e5 c3e5 d6e5 e2d2 e8d8 b1b2 d8d2 d1d2');
INSERT INTO evaluations VALUES(26426,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26426,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(26426,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(26426,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(26426,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(26426,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(26426,36,'cp 20','h2h3 g4h5 c1f4 f8d6 f4d6 d8d6 b1d2 g8e7 e1g1 e8g8 f1e1 h5g6 d3b5 h7h6 d2b3 a7a6 b5d3 g6d3 d1d3 d6c7 b3c5 e7f5 d3e2 a8c8 c5d3 f5d6 h3h4 a6a5 a2a4 f8e8 f3d2 e8e7 a1d1 c8e8 e2f3 d6f5 h4h5 b7b6 e1e2');
INSERT INTO evaluations VALUES(26893,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(26893,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(26893,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(26893,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(26893,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(26893,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(26893,36,'cp 82','f8d6 h2h4 d8e7 c1e3 e8c8 d1e2 g8f6 b1d2 g7g6 e1c1 f6h5 a2a3 d6f4 f1h3 c8b8 h3g4 f4e3 e2e3 h5f6 e3g5 d8d4 g4h3 h8d8 h3f1 f6h5 g5e7 c6e7 f1c4 f7f6 d2b3 d4d1 h1d1 d8d1 c1d1 e7c6 d1d2 c6e5 c4e2 h5f4 b3c5 b8c8 c5d3');
INSERT INTO evaluations VALUES(23848,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(23848,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(23848,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(23848,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(23848,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(23848,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(23848,36,'cp 88','h2h4 d8e7 c1e3 e8c8 d1e2 g8f6 b1d2 g7g6 e1c1 f6h5 a2a3 c8b8 e2c4 d6f4 f1d3 c6e5 c4b3 c7c6 c1b1 d8d7 d2c4 h8d8 c4e5 f4e5 d1g1 h5f4 d3c4 b8a8 a3a4 e5b8 a4a5 h7h5 b3a4 a7a6 c4a6 b7a6 a4c6 d7b7 c6a6 b7a7 e3a7 b8a7 b1a1');
INSERT INTO evaluations VALUES(27354,30,'cp 821',NULL);
INSERT INTO evaluations VALUES(27354,31,'cp 817',NULL);
INSERT INTO evaluations VALUES(27354,32,'cp 821',NULL);
INSERT INTO evaluations VALUES(27354,33,'cp 825',NULL);
INSERT INTO evaluations VALUES(27354,34,'cp 817',NULL);
INSERT INTO evaluations VALUES(27354,35,'cp 828',NULL);
INSERT INTO evaluations VALUES(27354,36,'cp 832','f3c6 e6d7 c6a8 g5d8 a8d8 e8d8 d2d3 h7h5 c1e3 a7a6 h2h4 f7f5 b1d2 f8d6 f2f3 d7e6 e1e2 h8e8 b2b3 d6b4 d2c4 b4d6 a1f1 d8c8 a2a3 c8d7 a3a4 e6d5 f1a1 d5e6 h1f1 e6d5 a1d1 d7e6 e3d2 d6e7 f1h1 e5e4 f3e4 f5e4 d3d4 e7f6 d2a5 e6d7 c4e3 d5e6 d4d5 e6g4 e3g4 h5g4');
INSERT INTO evaluations VALUES(27353,30,'cp 816',NULL);
INSERT INTO evaluations VALUES(27353,31,'cp 813',NULL);
INSERT INTO evaluations VALUES(27353,32,'cp 813',NULL);
INSERT INTO evaluations VALUES(27353,33,'cp 801',NULL);
INSERT INTO evaluations VALUES(27353,34,'cp 817',NULL);
INSERT INTO evaluations VALUES(27353,35,'cp 816',NULL);
INSERT INTO evaluations VALUES(27353,36,'cp 817','e6d7 c6a8 g5d8 a8d8 e8d8 d2d3 a7a6 c1e3 f8b4 e1e2 f7f5 c2c3 b4e7 b1d2 g7g5 d2f3 e5e4 d3e4 f5e4 f3e5 d7b5 e2d2 d8c8 a2a4 b5e8 d2c2 e8h5 h2h3 e7d6 g2g4 h5e8 e5c4 h7h5 b2b3 e8f7 e3g5 f7c4 b3c4 h5g4 h3g4 h8h1 a1h1');
INSERT INTO evaluations VALUES(27355,30,'cp 832',NULL);
INSERT INTO evaluations VALUES(27355,31,'cp 828',NULL);
INSERT INTO evaluations VALUES(27355,32,'cp 832',NULL);
INSERT INTO evaluations VALUES(27355,33,'cp 828',NULL);
INSERT INTO evaluations VALUES(27355,34,'cp 832',NULL);
INSERT INTO evaluations VALUES(27355,35,'cp 835',NULL);
INSERT INTO evaluations VALUES(27355,36,'cp 840','c6a8 g5d8 a8d8 e8d8 d2d3 a7a6 c1e3 f8d6 b1d2 d7c6 f2f3 d8d7 e1e2 h7h5 h2h4 h8g8 b2b3 d7e6 d2c4 f7f6 e3d2 g8b8 a2a3 d6c5 d2b4 c5d4 a1b1 d4a7 b4a5 e6d7 a5d2 a7c5 d2e3 c5d6 a3a4 c6d5 b1g1 f6f5 g1d1 b8e8 d1a1');
INSERT INTO evaluations VALUES(70,30,'cp 830',NULL);
INSERT INTO evaluations VALUES(70,31,'cp 838',NULL);
INSERT INTO evaluations VALUES(70,32,'cp 830',NULL);
INSERT INTO evaluations VALUES(70,33,'cp 832',NULL);
INSERT INTO evaluations VALUES(70,34,'cp 825',NULL);
INSERT INTO evaluations VALUES(70,35,'cp 812',NULL);
INSERT INTO evaluations VALUES(70,36,'cp 816','g5d8 a8d8 e8d8 d2d3 d7c6 f2f3 d8d7 c1e3 a7a6 e1e2 h7h5 h2h4 g7g6 b1d2 d7e6 b2b3 f8d6 d2c4 h8b8 a2a4 c6d5 e3d2 d6f8 d2a5 b8c8 a5c3 f7f6 b3b4 d5c6 c3d2 c8b8 c2c3 f8g7 d2e3 c6d7 h1b1 g7f8 c4d2 d7c6');
INSERT INTO evaluations VALUES(28309,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(28309,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(28309,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(28309,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(28309,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28309,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(28309,36,'cp 27','e1g1 e8g8 f1e1 d7d5 c4d5 e6d5 b1c3 b8d7 c1f4 c7c5 d4c5 d7c5 c3b5 f6e4 a1c1 c5e6 f4e5 a8c8 c1c8 d8c8 e2e3 c8d7 b5c3 f8d8 h2h4 e7b4 d1b3 b4f8 e1c1 d8c8 b3d1 h7h6 c1c2 d7e8 f3d4 e4c3 c2c3');
INSERT INTO evaluations VALUES(29796,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(29796,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(29796,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(29796,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(29796,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(29796,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(29796,36,'cp 38','e7e6 d1b3 f8b4 c1d2 b8c6 e2e3 e8g8 a2a3 b4c3 d2c3 a7a5 b3b7 d8d6 b7b3 f8b8 b3c4 c6b4 a3b4 a5b4 a1a8 b8a8 c3d2 a8a1 e1e2 a1b1 g2g4 f5g4 f1g2 b1b2 h1a1 h7h6 e2e1 g4f3 g2f3 d6h2 a1a8 g8h7 f3d5 h2g1 e1e2 e6d5 c4d5 g1c1 d5e4 g7g6 e4d3');
INSERT INTO evaluations VALUES(30045,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(30045,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(30045,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(30045,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(30045,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(30045,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(30045,36,'cp 62','d1b3 f8b4 a2a3 b4c3 b2c3 e8g8 c1g5 d8d6 c3c4 d5b6 g5d2 f8c8 c4c5 d6d5 b3d5 b6d5 f3h4 f5c2 a1c1 c2a4 e2e3 b8c6 h4f3 a7a6 f1d3 b7b6 d3c2 a4c2 c1c2 a6a5 c5b6 c7b6 e1e2 c6e7 h1c1 f7f6 e3e4 c8c2 c1c2');
INSERT INTO evaluations VALUES(32923,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(32923,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(32923,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(32923,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(32923,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(32923,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(32923,36,'cp 190','f3e2 g8e7 c3b5 b8a6 d2d4 e5d4 c2c3 e8c8 e1g1 d6d5 e4d5 e7d5 b5d4 a6c5 e2h5 f6g6 h5f3 d5f6 f1e1 f8d6 d4b5 g6g4 f3g4 f6g4 b5d6 d8d6 h2h3 g4h6 c1f4 d6d7 e1e2 h8d8 f4g5 f7f6 g5e3 b7b6 e3c5 d7d1 a1d1 d8d1 g1h2 b6c5 e2e7 h6f5');
INSERT INTO evaluations VALUES(36800,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36800,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36800,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(36800,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36800,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36800,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36800,36,'cp 0','c7c5 d4d5 f6e8 a2a4 e8d6 c4e2 c8g4 e1g1 b8d7 e3e4 g4f3 e2f3 a7a6 a4a5 g7d4 c3e2 d4g7');
INSERT INTO evaluations VALUES(36024,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36024,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36024,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36024,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36024,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36024,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36024,36,'cp 0','d4d5 c8g4 a2a4 a7a6 e3e4 b8d7 e1g1 f6e8 c4e2 g4f3 e2f3 e8d6 a4a5 g7d4 c3e2 d4g7');
INSERT INTO evaluations VALUES(37052,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(37052,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37052,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37052,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37052,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(37052,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37052,36,'cp 0','b4c3 b2c3 c7c5 f1e2 d5c4 e2c4 d8c7 c4e2 b7b6 e1g1 f8d8 c1b2 c8b7 f1e1 b8c6 a1c1 c6a5 c3c4 h7h6 f3d2 a5c6 d2f3');
INSERT INTO evaluations VALUES(37071,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37071,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37071,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37071,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37071,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37071,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37071,36,'cp 0','b2c3 c7c5 f1e2 d5c4 e2c4 d8c7 c4e2 b7b6 c1b2 c8b7 e1g1 f8d8 f1e1 b8c6 a1c1 c6a5 c3c4 h7h6 f3d2 a5c6 d2f3');
INSERT INTO evaluations VALUES(38367,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(38367,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(38367,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(38367,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(38367,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(38367,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(38367,36,'cp 48','d2d4 b8c6 g1f3 e7e5 d4d5 c6d4 e1g1 d4f3 e2f3 f7f5 b1c3 g8f6 f3f4 e5e4 c1e3 e8g8 d1d2 f6g4 e3d4 g7d4 d2d4 d8f6 d4d2 f6g7 c3b5 a7a6 b5d4 g7h6 h2h3 g4e5 d2c3 e5d3 f2f3 c8d7 f3e4 f5e4 g2e4 d3c5 e4g2 d7h3 f1e1');
INSERT INTO evaluations VALUES(38379,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(38379,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(38379,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(38379,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38379,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(38379,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(38379,36,'cp 43','b8c6 g1f3 e7e5 d4d5 c6d4 f3d4 e5d4 b1d2 g8e7 h2h4 h7h6 d2f3 c7c5 d5c6 e7c6 c1f4 c8e6 e1g1 e8g8 d1d2 d8a5 f4d6 f8e8 b2b4 a5a3 d2d3 a3a4 d3b3 a4b3 a2b3 e8d8 d6c7 d8d7 b4b5 d7c7 b5c6 b7c6');
INSERT INTO evaluations VALUES(40328,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40328,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(40328,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40328,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40328,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40328,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40328,36,'cp 0','e8g8 e1g1 b8c6 c1e3 b7b6 f1e1 c8b7 d1d2 a7a6 a2a4 e7e6 c4b3 d8c7 a1d1 a8d8 e3g5 d8d7 g5h6 c6e5 f3h2 e5c6');
INSERT INTO evaluations VALUES(42373,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(42373,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(42373,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(42373,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(42373,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(42373,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(42373,36,'cp 141','d7d5 f4g5 d5e4 c3e4 e8g8 f1d3 c8g4 c2c3 b8d7 e1g1 c7c5 d1c2 c5d4 f3d4 b6d4 c3d4 a8c8 g5f6 d7f6 e4f6 g8h8 c2f2 d8f6 f2f6 g7f6 f1f6 h8g7 f6f4 g4e6 g1f2 f8d8 a2a3 h7h6 a1d1 e6d5 d1d2 c8c1 d3e2 c1h1 h2h3 a7a5 d2d1 h1d1 e2d1 d8d6 f4f5');
INSERT INTO evaluations VALUES(45271,30,'cp 397',NULL);
INSERT INTO evaluations VALUES(45271,31,'cp 401',NULL);
INSERT INTO evaluations VALUES(45271,32,'cp 401',NULL);
INSERT INTO evaluations VALUES(45271,33,'cp 412',NULL);
INSERT INTO evaluations VALUES(45271,34,'cp 405',NULL);
INSERT INTO evaluations VALUES(45271,35,'cp 410',NULL);
INSERT INTO evaluations VALUES(45271,36,'cp 421','d1a4 b8c6 a4e4 e7e5 e4e3 f8e7 d2d4 e5d4 c3d4 e8g8 d4c5 f8e8 e1g1 e7f6 e3d2 d6c5 d2d8 e8d8 b1c3 h7h6 c3e4 f6e7 c1e3 b7b6 a1d1 c8e6 e4c3 e7f6 a2a3 g8f8 e2b5 d8d1 f1d1 c6a5 f3d2 a8d8 b5e2 a5c6 d2f3 c6d4 e3d4 c5d4 c3e4 e6b3 d1d2 f6e7 f3d4 b3d5 e4c3 d5b7 d4b3 d8d2');
INSERT INTO evaluations VALUES(45272,30,'cp 416',NULL);
INSERT INTO evaluations VALUES(45272,31,'cp 420',NULL);
INSERT INTO evaluations VALUES(45272,32,'cp 425',NULL);
INSERT INTO evaluations VALUES(45272,33,'cp 417',NULL);
INSERT INTO evaluations VALUES(45272,34,'cp 409',NULL);
INSERT INTO evaluations VALUES(45272,35,'cp 416',NULL);
INSERT INTO evaluations VALUES(45272,36,'cp 406','b8c6 a4e4 e7e5 e4e3 f8e7 d2d4 e5d4 c3d4 e8g8 d4c5 c6b4 b1a3 d6c5 e1g1 c8e6 c1d2 b4d5 e3e4 a7a6 a1d1 b7b5 a3b1 d5f6 e4c2 d8c8 b1c3 h7h6 c2c1 f8e8 f1e1 e7f8 e2f1 c5c4 f3d4 e6g4 e1e8 c8e8');
INSERT INTO evaluations VALUES(13407,30,'cp 422',NULL);
INSERT INTO evaluations VALUES(13407,31,'cp 414',NULL);
INSERT INTO evaluations VALUES(13407,32,'cp 420',NULL);
INSERT INTO evaluations VALUES(13407,33,'cp 398',NULL);
INSERT INTO evaluations VALUES(13407,34,'cp 405',NULL);
INSERT INTO evaluations VALUES(13407,35,'cp 427',NULL);
INSERT INTO evaluations VALUES(13407,36,'cp 417','a4e4 e7e5 e4e3 f7f5 d2d4 d8f6 d4d5 c6e7 c3c4 f5f4 e3b3 e7f5 b1d2 f8e7 b3b5 e8f7 d2e4 f6g6 e4c5 h8e8 c5e6 f7g8 e1g1 c8e6 d5e6 e5e4 b5d5 e4f3 e2f3 g6f6 c1d2 f5d4 f3d1 d4c6 d2c3 f6g5 f1e1 g5d5 c4d5 c6e5 d1a4 e8f8 c3e5 d6e5 e1e5 f4f3 g2f3 f8f3');
INSERT INTO evaluations VALUES(13406,30,'cp 397',NULL);
INSERT INTO evaluations VALUES(13406,31,'cp 405',NULL);
INSERT INTO evaluations VALUES(13406,32,'cp 409',NULL);
INSERT INTO evaluations VALUES(13406,33,'cp 428',NULL);
INSERT INTO evaluations VALUES(13406,34,'cp 426',NULL);
INSERT INTO evaluations VALUES(13406,35,'cp 436',NULL);
INSERT INTO evaluations VALUES(13406,36,'cp 428','e7e5 e2c4 c8e6 c4e6 f7e6 e4e2 d8f6 e1g1 h7h6 d2d3 d6d5 c1e3 f8d6 a2a3 e8g8 g1h1 f6g6 b2b4 a8d8 b4c5 d6c7 f3g1 d5d4 e3d2 e5e4 e2e4 g6e4 d3e4 c7e5 g1f3 d4c3 d2c3 e5c3 b1c3 c6d4 f3d4');
INSERT INTO evaluations VALUES(49314,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(49314,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(49314,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(49314,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(49314,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(49314,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(49314,36,'cp 26','e6d5 d2d4 d7d6 g1f3 d6e5 f3e5 f8d6 f1d3 e8g8 e1g1 b8c6 d1h5 g7g6 e5c6 b7c6 h5h6 d8f6 c2c3 f8e8 c1g5 f6e6 f2f3 c8d7 h6h4 f7f6 g5f4 d6f4 h4f4 e6d6 f4d6 c7d6 b2b4 e8e3 f1d1 g8f7 g1f2 a8e8 a1c1 d7f5 d3f1 h7h5 b4b5 f5d7 b5c6 d7c6 d1e1 e3e1 c1e1');
INSERT INTO evaluations VALUES(49816,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(49816,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(49816,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(49816,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(49816,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(49816,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(49816,36,'cp 29','g1f3 d7d6 d2d4 d6e5 f3e5 f8d6 f1d3 e8g8 e1g1 b8c6 d1h5 g7g6 e5c6 b7c6 h5h6 d8f6 c2c3 f8e8 c1g5 f6e6 h2h3 a7a5 g5d2 e6d7 a1e1 e8e1 f1e1 c8a6 d3a6 a8a6 h6h4 a6a8 d2h6 a8e8 e1e8 d7e8 g2g3 d6f8 h6f8 g8f8 g1f1 f8g7 h4f4 e8e7 b2b3 e7a3 f4e5 g7g8 e5c7 a3a2 c7c6 a2b3 c6a8 g8g7 a8a5');
INSERT INTO evaluations VALUES(49764,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(49764,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(49764,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(49764,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(49764,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(49764,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(49764,36,'cp 35','d4d5 c7c6 g1e2 b8a6 g2g4 d8e7 f1g2 e7e5 d1d4 e5d4 e2d4 f6d5 c3e4 a6c5 e4g3 d5c7 e3e4 d7d5 e4d5 c7d5 g2d5 c6d5 f2f3 c5e6 c1e3 e6d4 e3d4 f7f6 e1f2 e8f7 a1d1 f8d6 c2c3 c8d7 d4e3 d7c6 h2h4 d6g3 f2g3 a8e8 g3f2 h7h5 g4g5 a7a6 b2b3 e8e6 g5f6 e6f6');
INSERT INTO evaluations VALUES(49830,30,'cp -518',NULL);
INSERT INTO evaluations VALUES(49830,31,'cp -535',NULL);
INSERT INTO evaluations VALUES(49830,32,'cp -532',NULL);
INSERT INTO evaluations VALUES(49830,33,'cp -532',NULL);
INSERT INTO evaluations VALUES(49830,34,'cp -553',NULL);
INSERT INTO evaluations VALUES(49830,35,'cp -550',NULL);
INSERT INTO evaluations VALUES(49830,36,'cp -576','d8h4 g2g3 e4g3 h2g3 h4g3 e1e2 g3e5 e2f2 f8c5 d2d4 c5d4 f2g2 d4b2 c1b2 e5b2 d1e1 e8d8 b1d2 b8c6 f1d3 h7h6 a1b1 b2e5 e1f2 d7d6 f3f4 e5d5 d3e4 d5d4 h1e1 h8e8 e1e3 c8d7 b1e1 f7f5 d2b3 d4f6 e4f3 e8e3 f2e3 g7g5 e3d2 g5f4 d2d5 f6g7 g2f1 a8b8 f3h5 b7b6 d5f7 g7f7 h5f7 c6e5 f7h5 c7c6 e1e2');
INSERT INTO evaluations VALUES(50450,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(50450,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(50450,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(50450,33,'cp -154',NULL);
INSERT INTO evaluations VALUES(50450,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(50450,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(50450,36,'cp -150','d7d5 b1c3 e4c3 d2c3 e7e6 c3c4 d5d4 c1d2 c7c5 g2g3 b8c6 f1g2 f8e7 e1g1 e8g8 c2c3 e7f6 c3d4 c5d4 a1e1 e6e5 b2b4 c8e6 e1c1 a7a6 f1e1 d8b6 b4b5 a6b5 c4b5 c6e7 f3e5 a8a2 e5f3 e7f5');
INSERT INTO evaluations VALUES(47506,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(47506,31,'cp -139',NULL);
INSERT INTO evaluations VALUES(47506,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(47506,33,'cp -137',NULL);
INSERT INTO evaluations VALUES(47506,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(47506,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(47506,36,'cp -140','b1c3 e4f6 d2d4 a7a6 f3e5 b8c6 e5c6 b7c6 f2f3 e7e6 c3a4 a6a5 e2f2 f8b4 c2c3 b4d6 c1e3 e8g8 e1c1 f8e8 h2h4 f6d7 g2g4 e6e5 f1d3 e5d4 c3d4 c8a6 d3a6 a8a6 c1b1 d7b6 a4c5 a6a8 b1a1 d6c5 d4c5');
INSERT INTO evaluations VALUES(51004,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(51004,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(51004,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(51004,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(51004,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(51004,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(51004,36,'cp 144','e2e4 f8b4 d1g4 d8f6 e4e5 f6g6 g4g6 f7g6 a2a3 b4c3 b2c3 b7b5 a3a4 b5c4 f1c4 g8e7 g1e2 b8c6 a4a5 d7d6 f2f4 e8g8 e1g1 a8b8 e2g3 f8d8 a1a3 g8f7 g3e2 f7g8 f1e1 c6a7 c1e3 a7c6 e2g3 d6d5 c4d3 e7f5 e3c1 f5g3 h2g3 g8f7 e1e2 h6h5 c1e3 c8b7');
INSERT INTO evaluations VALUES(53420,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(53420,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(53420,32,'cp -98',NULL);
INSERT INTO evaluations VALUES(53420,33,'cp -101',NULL);
INSERT INTO evaluations VALUES(53420,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(53420,35,'cp -90',NULL);
INSERT INTO evaluations VALUES(53420,36,'cp -87','b8c6 d3b5 c8f5 g1f3 a8c8 e1g1 e7e6 d2d4 f8d6 f1e1 g8f6 b5c6 c8c6 c2c3 e8g8 f3e5 c6c8 c1f4 f6e8 f4g3 d6b8 a2a4 e8d6 a4a5 f8e8 b1d2 f7f6 e5f3 h7h6 d2b3 d6c4 g3b8 c8b8 b3c5 c4b2');
INSERT INTO evaluations VALUES(53748,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53748,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(53748,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(53748,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53748,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53748,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53748,36,'cp 0','e4d5 g8f6 b1c3 e5e4 f3e2 c6d5 c4d5 f6d5 e2b5 b8c6 b5d5 d8d5 c3d5 f8d6 g1e2 e8g8 b2b3 c8e6 d5c3 c6b4 e2d4 d6e5 d4e6 f7e6 a1b1 a8c8 c1b2 b4c2 e1e2 c2d4 e2e1 d4f5 e1e2');
INSERT INTO evaluations VALUES(53693,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53693,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53693,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53693,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53693,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53693,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53693,36,'cp 0','g8f6 b1c3 e5e4 f3e2 c6d5 c4d5 f6d5 e2b5 b8c6 b5d5 d8d5 c3d5 f8d6 g1e2 e8g8 b2b3 c8e6 d5c3 c6b4 e2d4 d6e5 d4e6 f7e6 a1b1 b4c2 e1e2 c2d4 e2e1 a8c8 c1b2 d4c2 e1e2 c2d4');
INSERT INTO evaluations VALUES(45742,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45742,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45742,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45742,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45742,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45742,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45742,36,'cp 0','b1c3 e5e4 f3e2 c6d5 c4d5 f6d5 e2b5 b8c6 b5d5 d8d5 c3d5 f8d6 g1e2 e8g8 b2b3 d6e5 a1b1 c8e6 d5f4 c6b4 f4e6 f7e6 e2c3 a8c8 c1b2 b4c2 e1e2 c2d4 e2e1 d4f5 e1e2');
INSERT INTO evaluations VALUES(57122,30,'cp -156',NULL);
INSERT INTO evaluations VALUES(57122,31,'cp -143',NULL);
INSERT INTO evaluations VALUES(57122,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(57122,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(57122,34,'cp -146',NULL);
INSERT INTO evaluations VALUES(57122,35,'cp -175',NULL);
INSERT INTO evaluations VALUES(57122,36,'cp -169','d7d5 c4d5 d8d5 g1f3 c8g4 b1c3 d5a5 f3g5 b8c6 d1b3 c6d8 d2d3 h7h6 c1d2 a5b6 c3d5 b6b3 a2b3 e8d7 g5f3 g4f3 e2f3 d8e6 f3f4 e5f4 f1h3 f8d6 d5f4 g8e7 f4h5 h8g8 d2c3 f7f5 e1d2 c7c6 b3b4 a7a6 d2c2 a8f8 h1e1 d6h2 e1e6 d7e6');
INSERT INTO evaluations VALUES(57203,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(57203,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(57203,32,'cp -134',NULL);
INSERT INTO evaluations VALUES(57203,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(57203,34,'cp -158',NULL);
INSERT INTO evaluations VALUES(57203,35,'cp -161',NULL);
INSERT INTO evaluations VALUES(57203,36,'cp -158','d7d5 g2g3 c8e6 d2d4 b8c6 g1h3 d8d7 h3f2 h7h5 b1a3 f7f6 d1a4 a7a6 d4e5 f6e5 c1g5 g8f6 e1c1 f8c5 f2d3 c5d6 e2e4 d5d4 c3d4 c6d4 a4d7 f6d7 f3f4 e6g4 d1d2 d4f3 d2c2 f3g5 f4g5 h5h4 g3h4 b7b5 a3b1 h8h4 b1c3 e8c8 c1b1');
INSERT INTO evaluations VALUES(56446,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(56446,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(56446,32,'cp -141',NULL);
INSERT INTO evaluations VALUES(56446,33,'cp -154',NULL);
INSERT INTO evaluations VALUES(56446,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(56446,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(56446,36,'cp -157','g2g3 b8c6 d2d4 g8f6 g1h3 d8e7 a2a3 c8e6 f1g2 e8c8 b2b4 h7h5 c1g5 e7d7 h3f2 f8d6 d1a4 h5h4 d4e5 d6e5 g5h4 h8h4 g3h4 e5d6 b1d2 c8b8 f2d3 f6h5 e1g1 h5f4 d3f4 d6f4 b4b5 f4e3 g1h1 c6e7 d2b3 e7f5 b3d4 d7e7 d4e6 f7e6 a4g4');
INSERT INTO evaluations VALUES(1653,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(1653,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(1653,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(1653,33,'cp 225',NULL);
INSERT INTO evaluations VALUES(1653,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(1653,35,'cp 232',NULL);
INSERT INTO evaluations VALUES(1653,36,'cp 227','d2d4 e7e6 c3b5 d6d8 c1f4 f8b4 c2c3 b4a5 b5a3 g8f6 a3c4 a5b6 e2d3 f6d5 f4g3 c6e7 a2a4 a7a5 h2h3 c8d7 e1g1 e8g8 f1e1 e7f5 g3h2 f5h4 f3e5 d7e8 d1c2 h4g6 a1d1 d8g5 c4b6 d5b6 b2b3 b6d5 h2g3 e8c6 e5c6 b7c6');
INSERT INTO evaluations VALUES(2387,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(2387,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(2387,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(2387,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(2387,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(2387,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(2387,36,'cp 15','d8e6 c1e3 g8e7 b1c3 e6d4 e3d4 e7c6 f1b5 c8d7 e1c1 e8c8 e5e6 d7e6 b5c6 b7c6 c3e4 d8d5 h2h4 f8e7 d4g7 d5d1 c1d1 h8d8 d1e2 e6d5 e4f6 e7f6 g7f6 d8d6 f6g5 d5g2 h1g1 g2e4 c2c3 d6d5 g5e3 e4g6 g1g5 a7a5 e2f3 c8d7 g5d5 c6d5 f3f4 d7e6 e3d4 f7f6 d4e3 a5a4');
INSERT INTO evaluations VALUES(4194,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4194,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4194,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(4194,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(4194,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(4194,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(4194,36,'cp 15','d2d3 d7d6 c3a4 c5b6 e1g1 c8e6 c4e6 f7e6 a4b6 a7b6 f1e1 c6e7 c2c3 e7g6 d3d4 f6d7 d4e5 d7e5 f3e5 g6e5 f2f4 e5d7 c1e3 d8e8 d1d3 d7c5 e3c5 b6c5 e1f1 b7b5 a1e1 e8d7 f1f2 c5c4 d3f3 f8f7 f3g4 a8f8 e1e2 d6d5 g2g3 d7c6 e4d5');
INSERT INTO evaluations VALUES(4933,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(4933,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(4933,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(4933,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(4933,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(4933,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(4933,36,'cp 45','g8f6 d2d4 d7d6 d4e5 d6e5 d1d8 e7d8 a2a4 a8b8 b1c3 e8g8 a4b5 a6b5 c1e3 d8e7 c3d5 e7d6 d5f6 g7f6 f1d1 c8e6 b3e6 f7e6 c2c3 b8a8 b2b4 g8f7 g1f1 h7h5 g2g3 f7e7 f3d2 f8h8 f1e2 e7d7 d2f3 c6e7 h2h3 a8a1 d1a1');
INSERT INTO evaluations VALUES(5254,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(5254,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(5254,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(5254,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(5254,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(5254,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(5254,36,'cp 0','d7c6 d1d8 e8d8 b1c3 e7e5 b2b3 g8f6 f1e2 c8e6 e1g1 f6d7 f1d1 g7f8 c1e3 h7h5 h2h3 f8h6 e3c5 h6f8');
INSERT INTO evaluations VALUES(6700,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6700,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(6700,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6700,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(6700,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(6700,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6700,36,'cp 0','d7d6 f3g5 c6e5 c4b3 c8g4 b3f7 e5f7 d1g4 g8f6 g4e2 d8e7 e1g1 e8g8 f2f3 a8e8 g5f7 f8f7 c3c4 b7b6 c1b2 h7h5 a1d1 h5h4 b2f6 e7f6 h2h3 f6g5 d1d5 e8e5 a2a4 g5g3 f1d1 a7a5 d1d3 g3f4 d3d1');
INSERT INTO evaluations VALUES(7434,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(7434,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7434,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7434,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7434,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7434,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7434,36,'cp 0','d7d6 a2a4 a8b8 a4b5 a6b5 d3d4 c5b6 b1a3 e5d4 c3d4 c6a5 e4e5 d6e5 d4e5 d8d1 b3d1 f6d5 a3b5 e8g8 b5d4 c7c5 d4e2 f8d8 e2f4 c8f5 f3h4 f5c8');
INSERT INTO evaluations VALUES(7888,30,'cp 201',NULL);
INSERT INTO evaluations VALUES(7888,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(7888,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(7888,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(7888,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(7888,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(7888,36,'cp 215','d1d6 d8e7 d6e7 e8e7 c1g5 h7h6 c3d5 e7d8 g5f6 g7f6 d5f6 d7d6 f6d5 c8e6 d5e3 d8e7 c2c3 h8b8 f1c4 e6c4 e3c4 b7b5 c4e3 e7e6 e1c1 b5b4 e3f5 b8d8 c3b4 c6b4 a2a3 b4c6 c1b1 c6e7 f5h6 d8h8 h6g4 d6d5 e4d5 e7d5 g4e3 d5e3 f2e3');
INSERT INTO evaluations VALUES(10406,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(10406,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(10406,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(10406,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(10406,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(10406,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(10406,36,'cp -20','b4d2 d1d2 e8g8 h2h3 d7d6 g2g3 f6h5 f1g2 f7f5 e4f5 c8f5 e1g1 d8e8 f3h4 f5d7 b1c3 c6d4 c3e2 d4e2 d2e2 d7c6 a1e1 e8f7 e2d2 a8e8 b2b3 c6g2 h4g2 c7c5 e1e3 b7b6 b3b4 h7h6 b4c5 b6c5 g1h2 g8h8 f2f4 e5f4 g2f4 e8e3 d2e3 h5f4 f1f4');
INSERT INTO evaluations VALUES(11231,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(11231,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(11231,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(11231,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(11231,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(11231,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(11231,36,'cp 97','d1d2 f8e7 f2f3 b7b5 c3d5 e8g8 f1e2 f6d5 e4d5 c6d4 b3d4 e5d4 e3d4 f8e8 e1g1 e7g5 f3f4 g5f6 e2f3 c8f5 c2c3 g7g6 a1e1 f6d4 d2d4 h7h5 e1e8 d8e8 h2h3 e8e7 g2g4 f5d7 g1g2 a8c8 d4d2 e7f6 g2g3 h5h4 g3h2 c8c4 f3e2 c4a4 a2a3 a4e4 e2d3 e4e7 f1f2 e7e8 f4f5');
INSERT INTO evaluations VALUES(12476,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12476,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(12476,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(12476,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(12476,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12476,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(12476,36,'cp -14','f8b4 a2a3 b4c3 b2c3 d7d5 d4c6 b7c6 f1d3 e8g8 e1g1 c8a6 c1f4 f8e8 d1e2 a6d3 c2d3 e6e5 f4e3 d8d6 e2f2 h7h6 a3a4 f6d7 f2a2 a7a5 f1b1 a8b8 a2c2 d7c5 h2h3 c5e6 b1b8 e8b8 a1b1 b8b1 c2b1 d5d4 c3d4 e6d4 b1b7');
INSERT INTO evaluations VALUES(7834,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7834,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7834,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7834,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7834,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7834,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(7834,36,'cp 0','a2a3 b4c3 b2c3 d7d5 d4c6 b7c6 f1d3 e8g8 e1g1 d8b6 g1h1 c8a6 c1f4 c6c5 d3a6 b6a6 f4g5 a6c4 g5f6 g7f6 d1e1 a8d8 e1e3 d8b8 e4e5 f6f5 e3g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(15191,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(15191,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(15191,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(15191,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(15191,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(15191,35,'cp -68',NULL);
INSERT INTO evaluations VALUES(15191,36,'cp -68','b1d2 f8d6 e3e4 d5e4 d2e4 f6e4 d3e4 c8e6 c2c3 e8g8 g1f3 a7a5 a2a4 d8e7 d1e2 d6c5 f3g5 e6d7 e1g1 f7f6 g5f3 d7e6 c1e3 c6d8 f3h4 c5e3 e2e3 g7g6 f2f4 e6c4 f1e1 d8f7 h2h3 e5f4 g3f4 a8e8 h4f3 f7d6');
INSERT INTO evaluations VALUES(16846,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(16846,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(16846,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(16846,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(16846,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(16846,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16846,36,'cp -33','e5d4 f3d4 d7d5 d4c6 b7c6 e4d5 c6d5 b1c3 c8b7 d1e2 d8e7 e2e7 f8e7 c1f4 e8c8 e1c1 a7a6 f2f3 h8e8 h1e1 d5d4 c3e4 h7h5 c2c3 d4c3 d1d8 c8d8 e4c3 g7g6 h2h3 f7f6 c1c2 b7c6 e1e6 d8d7 e6e1 g6g5 f4h2');
INSERT INTO evaluations VALUES(16869,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(16869,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(16869,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(16869,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(16869,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(16869,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(16869,36,'cp -19','f3d4 d7d5 d4c6 b7c6 e4d5 c6d5 b1c3 d5d4 c3e2 f8c5 c2c3 d4d3 d1a4 d8d7 a4e4 d7e6 e4e6 c8e6 e2f4 e6f5 c1e3 c5e3 f2e3 d3d2 e1d2 e8c8 d2e1 h8e8 a1d1 e8e3 e1f2 d8e8 d1d5 f5e4 d5d1 e4b7 h1e1 e3e1 d1e1 e8e1 f2e1 c8d7 g2g3 g7g5 f4h3 f7f6 e1d2 d7d6 b2b3 c7c5 h3f2 f6f5 d2e3 b7d5 f2d3');
INSERT INTO evaluations VALUES(16863,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(16863,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(16863,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(16863,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(16863,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(16863,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(16863,36,'cp -30','d7d5 d4c6 b7c6 e4d5 c6d5 b1c3 d5d4 c3e2 f8c5 c2c3 d4c3 d1a4 c8d7 a4e4 c5e7 e2c3 e8g8 e1g1 a8b8 c1f4 b8b2 f4c7 d8c7 c3d5 c7b7 d5e7 g8h8 e4b7 b2b7 f1d1 g7g6 f2f3 f8b8 e7d5 b7b2 d5f4 d7f5 g2g4 f5b1 d1d7 b8b7 d7b7 b2b7 h2h4 h8g7 a2a3');
INSERT INTO evaluations VALUES(17555,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(17555,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(17555,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(17555,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(17555,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(17555,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(17555,36,'cp 119','b5c6 b7c6 d4c5 d8a5 b1d2 a5c5 e1g1 g8f6 b2b4 c5b4 a1b1 b4c5 c2c4 e7e5 d1a4 f8e7 c1a3 c5a7 c4c5 d6c5 f3e5 e8g8 e5c6 a7c7 f1c1 f8e8 c6e7 c7e7 a3c5 e7d8 c1c2 c8e6 f2f3 a8c8 d2f1 d8d3 b1c1 f6d7 c5d4 d7f8 c2d2 e6d7 c1c8 e8c8 a4b4 d3b5 b4b5 d7b5 g1f2 f7f6 f1e3 f8e6 f2g3 g8f7 e3d5 e6d4 d2d4 c8c2');
INSERT INTO evaluations VALUES(17540,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(17540,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(17540,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(17540,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(17540,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(17540,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(17540,36,'cp 101','b7c6 d4c5 d8a5 b1d2 a5c5 e1g1 g8f6 b2b4 c5b4 a1b1 b4c5 c2c4 e7e5 d1a4 f8e7 c1a3 c5a7 c4c5 d6c5 f3e5 e8g8 e5c6 a7c7 f1c1 f8e8 c6e7 c7e7 a3c5 e7d8 c5e3 c8d7 a4d4 d7e6 d4d8 e8d8 b1b2 a8c8 c1c8 d8c8 f2f3 f6d7 e3d4 h7h6 d2b3 f7f5 e4f5 e6f5 g2g4 f5g6 g1f2');
INSERT INTO evaluations VALUES(18207,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(18207,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(18207,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(18207,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(18207,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(18207,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(18207,36,'cp 125','d4c5 d8a5 b1d2 a5c5 e1g1 g8f6 b2b4 c5b4 a1b1 b4c5 c2c4 e7e5 d1a4 f8e7 c1a3 c5a7 c4c5 d6c5 f3e5 e8g8 e5c6 a7c7 f1c1 f8e8 c6e7 c7e7 a3c5 e7d8 c1c2 c8e6 f2f3 a8c8 a4a3 h7h6 d2f1 d8c7 f1e3 c7e5 b1c1 e8d8 c5d6 d8d6 c2c8 e6c8 c1c8 g8h7 a3c3 e5c3 c8c3 d6d2');
INSERT INTO evaluations VALUES(18203,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(18203,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(18203,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(18203,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(18203,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(18203,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(18203,36,'cp 112','d8a5 b1d2 a5c5 e1g1 g8f6 b2b4 c5b4 a1b1 b4c5 c2c4 e7e5 d1a4 f8e7 c1a3 c5a7 c4c5 d6c5 f3e5 e8g8 e5c6 a7c7 f1c1 f8e8 c6e7 c7e7 a3c5 e7d8 c1c2 c8e6 f2f3 a8c8 d2f1 d8d3 b1c1 f6d7 c5f2 c8c2 c1c2 h7h6 c2d2 d3b5 a4d1 d7b6 h2h3 b6c4 d2d4 c4b2 d1a1 b2d3 f2g3 e8c8 a1d1 c8c3');
INSERT INTO evaluations VALUES(18154,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18154,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(18154,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(18154,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(18154,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(18154,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(18154,36,'cp 10','e1g1 d7d5 c4d3 g8f6 d1e2 d8c7 e4e5 f6d7 f1e1 a6a5 h2h4 d7c5 e2f3 c5d3 c2d3 f8c5 c1f4 e8g8 b1d2 c8a6 f3g3 g8h8 d2b3 c5e7 d3d4 h7h6 e1c1 f8b8 g3e3 a5a4 b3c5 a6c4 b2b3 a4b3 a2b3 e7c5');
INSERT INTO evaluations VALUES(18153,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(18153,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18153,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18153,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18153,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(18153,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18153,36,'cp 0','d7d5 c4d3 g8f6 d1e2 d8c7 e4e5 f6d7 f1e1 a6a5 c1f4 c7b6 b1d2 b6b2 a1b1 b2a2 b1a1 a2b2 d2b3 a5a4 e2g4 c6c5 f4c1 b2c3 c1d2 c3b2 d2c1');
INSERT INTO evaluations VALUES(18935,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18935,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18935,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(18935,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18935,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18935,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(18935,36,'cp 0','f8e7 f1d3 d7d5 e4e5 f7f6 e3e2 f6e5 f3e5 c6e5 e2e5 g8f6 c1f4 e8g8 b1d2 d8b6 e1g1 f6d7 e5h5 f8f4 h5h7 g8f8 h7h8 f8f7 h8h5 f7g8');
INSERT INTO evaluations VALUES(19639,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(19639,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(19639,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(19639,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(19639,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(19639,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(19639,36,'cp 59','d2d4 c5d4 d1d4 g8f6 e4e5 d6e5 d4d8 e8d8 b5c6 b7c6 f3e5 d8e8 c1d2 f6d7 e5d7 c8d7 e1c1 d7f5 d2e3 h7h5 h1e1 h5h4 d1d4 h8h5 d4c4 f5d7 c3e4 d7e6 c4a4 a7a5 f2f3 f7f6 b2b3 e8f7 h2h3 e6d5 e3d2');
INSERT INTO evaluations VALUES(20611,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(20611,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(20611,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(20611,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(20611,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(20611,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(20611,36,'cp 126','d4d5 c6e7 a2a4 g7g6 h2h4 g8f6 f3d2 f8h6 f1e2 e8g8 h4h5 g8g7 d2c4 e7g8 c1h6 g8h6 b1d2 h6g8 d1c2 d8e7 c4e3 h7h6 g2g3 c8d7 e1c1 e7e8 d2c4 e8e7 d1f1 a8b8 h5g6 f7g6 f2f3 b7b5 c4a5 b5a4 a5c6 d7c6 d5c6');
INSERT INTO evaluations VALUES(21438,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21438,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(21438,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(21438,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(21438,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(21438,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(21438,36,'cp 21','e6e5 d4e5 d5e4 d1e2 c6e5 d2e4 d8e7 c1g5 f7f6 g5f4 c8d7 e1c1 e8c8 g1f3 e7e6 f4e5 f6e5 e2c4 g8f6 c4e6 d7e6 d1d8 c8d8 e4g5 e6a2 f3e5 d8c8 g5f7 h8g8 g2g3 f8c5 f2f4 a2f7 e5f7 g8e8 f1c4 e8e7 h1d1 c7c6 b2b4 c5e3 c1c2 c8c7 d1e1');
INSERT INTO evaluations VALUES(21702,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(21702,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21702,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(21702,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(21702,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21702,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(21702,36,'cp 36','d4e5 d5e4 d1e2 c8f5 d2e4 d8d5 e4g3 f5e6 g1h3 e6h3 g2h3 d5e5 e2e5 c6e5 c1e3 e8c8 e3a7 g8f6 f1e2 e5d3 e2d3 d8d3 e1e2 d3d6 a7d4 d6c6 h1d1 f8c5 d4c5 c6c5 d1d4 h8e8 e2f1 c5c6 a1d1 c6a6 a2a4 g7g6 g3e2 f6h5 b2b4 a6a4');
INSERT INTO evaluations VALUES(22195,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(22195,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(22195,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(22195,33,'cp 209',NULL);
INSERT INTO evaluations VALUES(22195,34,'cp 196',NULL);
INSERT INTO evaluations VALUES(22195,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(22195,36,'cp 205','c4b5 c7c6 d5c6 b7c6 b5a4 f8d6 e1g1 g8f6 b1c3 e8g8 f1e1 h7h6 h2h3 f8e8 d2d4 e5d4 e1e8 f6e8 f3d4 c8b7 d1e2 d8f6 d4f3 d6f8 f3e5 a8d8 e5g4 f6g6 c1f4 e8d6 g4e3 d6f5 e3f5 g6f5 f4g3');
INSERT INTO evaluations VALUES(22196,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(22196,31,'cp 224',NULL);
INSERT INTO evaluations VALUES(22196,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(22196,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(22196,34,'cp 223',NULL);
INSERT INTO evaluations VALUES(22196,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(22196,36,'cp 222','c7c6 d5c6 b7c6 b5a4 f8d6 e1g1 g8f6 b1c3 e8g8 f1e1 d8c7 h2h3 f8e8 f3g5 d6e7 d2d3 c8f5 c3e4 f5g6 c1d2 f6e4 g5e4 a5b7 d1f3 a8c8 d2c3 b7c5 a4b3 c8d8 f3e3 g6e4 d3e4 e7d6 a1d1 a7a5 e3g5 c5e6 b3e6 e8e6');
INSERT INTO evaluations VALUES(26041,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(26041,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(26041,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(26041,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(26041,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(26041,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(26041,36,'cp -41','a7a6 b5e2 g8f6 c2c3 g7g6 e1g1 f8g7 a4c2 b7b5 f1d1 e8g8 c1f4 d8b6 h2h3 a8c8 b1d2 e7e5 f4g5 c6d4 f3d4 e5d4 c2d3 d4c3 b2c3 f8e8 e2f3 d7c6 d3d6 f6d7 d6g3 d7c5 a1c1 c5e4 d2e4');
INSERT INTO evaluations VALUES(25618,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(25618,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(25618,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(25618,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(25618,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(25618,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(25618,36,'cp -32','b5e2 g8f6 c2c3 g7g6 a4c2 f8g7 e1g1 e8g8 c1f4 d8c7 b1d2 a8c8 d2c4 d7e6 c4e3 b7b5 f1d1 f6h5 f4g5 h7h6 g5h4 c7b7 c2d2 f8e8 a2a4 b5b4 c3b4 c6b4 f3d4 b7e4 d4e6 e4h4 e2h5 f7e6');
INSERT INTO evaluations VALUES(29262,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(29262,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(29262,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29262,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(29262,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(29262,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(29262,36,'cp 17','f3h4 f5g4 b1d2 g4h5 c2c4 f8e7 h2h3 b8d7 g3g4 h5g6 h4g6 h7g6 d1c2 e8g8 f1d1 e7d6 e2e4 d5e4 d2e4 f6e4 c2e4 d7f6 e4f3 d8c7 c1e3 a8d8 a1c1 d6h2 g1h1 h2d6 b2b3 f8e8 d4d5 e6d5 c4d5 c6c5 e3g5 b7b6 d1e1 c7d7 c1d1 d8c8 e1e8 c8e8 g5f6');
INSERT INTO evaluations VALUES(30668,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30668,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(30668,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(30668,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(30668,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(30668,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(30668,36,'cp 39','g8f6 c1e3 d7d5 e4e5 f6g8 f1d3 c7c5 f3f4 g8h6 e3f2 f8e7 h2h3 b8c6 g1f3 e8g8 e1g1 f7f6 e5f6 f8f6 d1d2 e7d6 f3e5 h6f5 c3e2 f5d4 e2d4 c6d4 f2d4 c5d4 e5g4 f6f8 a1e1 d8d7 g4e5 d7a4 g2g4');
INSERT INTO evaluations VALUES(30908,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(30908,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(30908,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(30908,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(30908,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(30908,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(30908,36,'cp 225','d1e2 d8e7 b5d7 b8d7 c3d5 e7e2 e1e2 f8d6 h1e1 f7f6 d5c3 g8e7 c3e4 d6c7 d2d4 c5d4 f3d4 c7e5 e1d1 f6f5 e4g5 e5d4 d1d4 e7c6 d4d1 d7f6 g5e6 e8f7 e6g5 f7g8 g5f3 h7h6 f3d4 c6d4 d1d4 g7g5 e2f1 g8f7 c1e3 h8e8');
INSERT INTO evaluations VALUES(30909,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(30909,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(30909,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(30909,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(30909,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(30909,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(30909,36,'cp 216','d8e7 b5d7 b8d7 c3d5 e7e2 e1e2 f8d6 d2d4 g8e7 d5e7 d6e7 c1e3 e8g8 h1d1 c5d4 d1d4 d7f6 d4d2 a8c8 e2f1 f6e4 d2e2 e7c5 a1d1 c5e3 e2e3 f8e8 c2c3 e4c5 e3e8 c8e8 d1d4 e8c8 f1e2 g7g6 g2g3 b7b5 d4b4 c8e8 e2f1 c5d3');
INSERT INTO evaluations VALUES(31199,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(31199,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31199,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31199,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(31199,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(31199,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31199,36,'cp 23','g8f6 d1e2 e7e6 e4f6 d8f6 g2g3 f8b4 f3d2 b8d7 f1g2 e8g8 e1g1 h7h6 a2a4 a7a5 g3g4 f5h7 f2f4 f6h4 d2c4 b4c5 c1e3 c5e3 c4e3 d7b6 e2f2 h4f2 f1f2 a8d8 b2b3 d8d4 a1f1 f7f5 g4g5 h6g5 f4g5 f5f4 e3g4 b6d5 g2d5 e6d5');
INSERT INTO evaluations VALUES(32620,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(32620,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(32620,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(32620,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(32620,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(32620,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(32620,36,'cp 47','d5f5 d2d3 b8c6 h2h3 g4f3 e2f3 f5d7 e1g1 e7e6 c3e2 a8d8 d3d4 f8e7 c1f4 f6d5 c2c3 e8g8 f1e1 d5f4 e2f4 e7d6 f4d3 b7b6 d1e2 c6e7 a1d1 h7h6 g2g3 c7c6 h3h4 d7c7 d1d2 c6c5 d4c5 d6c5 d3c5 c7c5 g1g2 d8d2');
INSERT INTO evaluations VALUES(39454,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(39454,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(39454,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(39454,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(39454,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(39454,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(39454,36,'cp 52','c1d2 g8f6 e2e3 e8g8 f1d3 c7c5 a2a3 b4c3 d2c3 c5d4 e3d4 f6e4 g1e2 e4c3 e2c3 b8c6 d4d5 c6d4 e1g1 d8f6 c3e2 d4e2 d1e2 e6e5 c4c5 d7d6 b2b4 e5e4 d3b5 a7a6 b5a4 f6e5 c5c6 b7c6 a4c6 a8a7 e2e3 a7e7 f2f3 e4f3 e3e5 e7e5 f1f3 g7g6 a1c1 g8g7 h2h3');
INSERT INTO evaluations VALUES(41485,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(41485,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(41485,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41485,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(41485,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41485,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(41485,36,'cp 12','e2e3 b7b6 f1e2 e8g8 c2c4 c8b7 e1g1 b8d7 b1c3 c7c5 c3d5 f6d5 c4d5 b7d5 a1c1 a8c8 d1d2 e7e6 f1d1 d8e7 h2h3 f8d8 b2a3 d5b7 d4c5 d7c5 a3c5 b6c5 d2a5 b7d5 f3d2 g7f8 e2f3 d8d7 e3e4 d5c6 d2c4 c8d8 d1d7 d8d7');
INSERT INTO evaluations VALUES(43434,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(43434,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(43434,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(43434,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(43434,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43434,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43434,36,'cp 11','h2h3 e7e6 f1d3 f8d6 c3e2 d8c7 c1g5 c8d7 e1g1 d7c6 d1c1 b8d7 g5f4 c6b5 f4d6 c7d6 e2f4 b5d3 f4d3 e8g8 c1e3 d7b6 c2c3 a8c8 e3e2 d6c7 f3e5 f6e4 f1e1 a6a5 a2a3 f7f6 e5g4 f8f7 g4e3 c8e8');
INSERT INTO evaluations VALUES(43496,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(43496,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(43496,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(43496,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(43496,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43496,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(43496,36,'cp -18','d2d4 b7b5 c4a2 f8e7 d1e2 b5b4 a3b4 c5b4 c3a4 c8b7 d4d5 e6d5 e4d5 d8a5 a2b3 b7d5 c1g5 e8g8 e1g1 d5f3 e2f3 b8d7 g5f4 a5c7 f1e1 a8e8 a1d1 a6a5 h2h3 d7e5 f3e2 h7h6 c2c3 e5g6 f4e3 d6d5 c3c4 d5c4 e2c4 c7c4 b3c4 e8d8 e3d4 f6d5 a4b6 d5b6');
INSERT INTO evaluations VALUES(44463,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(44463,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(44463,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44463,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(44463,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(44463,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(44463,36,'cp 10','d7d5 e4e5 d5d4 e5f6 d4c3 f6g7 c3d2 c1d2 f8g7 d1c2 d8b6 d2c3 g7c3 c2c3 h8g8 f1d3 b8d7 g2g3 d7f6 g1f3 c8h3 f3e5 e8c8 e1c1 b6f2 d1d2 f2c5 e5f7 f6e4 d3e4 d8d2 c1d2 c5f2 d2c1 f2f7 e4c2 g8d8 h1e1 h7h5 a2a3 h3g4 c3e3 f7c4 e3a7 c4d5 a7a8 c8c7 e1e7 d8d7 e7d7 g4d7 a8a4 d5h1 c2d1 h1h2 a4a5 c7c8 a5h5 h2g3');
INSERT INTO evaluations VALUES(49680,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(49680,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(49680,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(49680,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(49680,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(49680,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(49680,36,'cp 76','g4g5 f6h5 d2d4 g7g6 d4e5 b8c6 b1c3 c6e5 e2e4 f8g7 f2f4 e5c6 g1e2 d7d6 d1d3 h7h6 h2h4 c8g4 c1d2 h6g5 h4g5 d8d7 e1c1 e8c8 g2f3 g4f3 d3f3 f7f6 g5f6 g7f6 h1g1 d8g8 f3d3 d7f7 c3d5 c8b8 c1b1 g8e8 d2c1 f6g7 e2c3 a7a6 a2a3 h5f6');
INSERT INTO evaluations VALUES(49519,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(49519,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(49519,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(49519,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(49519,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(49519,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(49519,36,'cp 42','f6h5 d2d4 g7g6 d4e5 f8g7 e2e4 b8c6 b1c3 c6e5 g1e2 d7d6 f2f4 e5c6 c1e3 h7h6 h2h4 c8g4 d1d2 d8d7 e1c1 e8c8 d1e1 h8e8 a2a3 f7f5 e4f5 g4f5 c3d5 c8b8 e2c3 a7a6 g2f3 d7f7 h1g1 c6a5 f3h5 g7c3 d5c3 g6h5');
INSERT INTO evaluations VALUES(50523,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(50523,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(50523,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50523,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(50523,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(50523,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(50523,36,'cp 22','b7b5 d2d4 c5d4 d1d4 d8c7 e1g1 b8c6 d4e3 f8e7 f1d1 d7d6 f3d4 c6d4 e3d4 g8f6 a2a4 e6e5 c3d5 f6d5 d4d5 c8b7 d5d3 e8g8 a4b5 a6b5 a1a8 f8a8 c1d2 a8c8 d1a1 c8a8 a1a8 b7a8 h2h3 a8c6 d2b4 h7h6 e2g4 g8f8 g2g3 g7g6 h3h4 h6h5 g4d1');
INSERT INTO evaluations VALUES(50766,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50766,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(50766,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(50766,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(50766,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50766,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(50766,36,'cp 21','g8f6 f2f4 b8d7 a2a4 f8e7 e1g1 d7c5 c1e3 e8g8 g1h1 h7h6 a4a5 c5d3 c2d3 c8d7 d1f3 f6h7 h2h3 a8c8 d4c2 d7c6 f3d1 h7f6 c2d4 c6d7 d1e2 f6h7 e2f3 e6e5 f4e5 d6e5 d4f5 d7f5 f3f5 e7g5 e3b6 d8d3 c3d5 d3b3');
INSERT INTO evaluations VALUES(53306,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(53306,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(53306,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(53306,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(53306,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(53306,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(53306,36,'cp -49','b8c6 g1f3 g8f6 f1g2 d8c7 e1g1 e7e5 b1c3 h7h6 d2d3 c8e6 c3b5 c7b8 c2c4 d5d4 e2e3 f8e7 e3d4 c5d4 b2a3 e7a3 b5a3 e8g8 d1d2 b8d6 b3b4 c6b4 a3b5 d6c5 f1b1 b4c6 b5c7 e5e4 c7e6 f7e6 d3e4 a8d8 f3e1 c6e5 e1d3 e5d3 d2d3 f6g4');
INSERT INTO evaluations VALUES(19808,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(19808,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(19808,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(19808,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(19808,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(19808,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(19808,36,'cp -42','g1f3 d8c7 b1c3 g8f6 f1g2 e7e5 e1g1 f8e7 e2e4 d5e4 f3g5 c8f5 g5e4 f6e4 c3e4 e8g8 e4c3 c7d7 c3d5 e7d6 d2d3 f7f6 c2c4 a8e8 b2c3 f5e6 d1c2 b7b6 a1e1 c6e7 d5e7 e8e7 e1e3 f8e8 f1e1 e6f7 c2e2 h7h6 h2h3 f6f5');
INSERT INTO evaluations VALUES(54246,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54246,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(54246,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(54246,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(54246,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(54246,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54246,36,'cp 17','e4d5 e6d5 d2d4 c5d4 d1d4 g8f6 c1g5 b8c6 d4a4 f8b4 g1e2 c8g4 f1g2 g4e2 e1e2 b4c3 b2c3 e8g8 h1e1 h7h6 g5f6 d8f6 e2f1 f6c3 a1b1 a8d8 e1d1 c6e5 a4d4 c3d4 d1d4 b7b6 b1b5 d8d7 d4d5 d7d5 g2d5');
INSERT INTO evaluations VALUES(54142,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54142,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(54142,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54142,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(54142,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(54142,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(54142,36,'cp 9','e6d5 d2d4 c5d4 d1d4 g8f6 c1g5 b8c6 d4a4 f8b4 g1e2 c8g4 f1g2 g4e2 e1e2 b4c3 b2c3 e8g8 h1e1 h7h6 g5f6 d8f6 e2f1 f6c3 a1d1 d5d4 g2c6 b7c6 d1d3 c3c5 d3d4 f8d8 e1e4 d8d4 e4d4 a8b8 a4c4 b8b1 f1g2 c5c4 d4c4 b1a1 c4a4 a1c1 a4a7 c1c2 g2f3 c6c5 a7a8 g8h7 a2a4 c5c4 h2h4 c4c3');
INSERT INTO evaluations VALUES(54547,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(54547,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(54547,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(54547,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(54547,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(54547,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54547,36,'cp 12','d2d4 c5d4 d1d4 g8f6 c1g5 b8c6 d4a4 f8b4 g1e2 c8g4 f1g2 g4e2 e1e2 b4c3 b2c3 e8g8 h1e1 h7h6 g5f6 d8f6 e2f1 a8d8 a1d1 f6c3 d1d5 f8e8 e1d1 d8d5 g2d5 e8e7 d5c6 b7c6 a4b3 c3e5 b3c4 g8h7 d1d4 g7g6 f1g2 h7g7 c4d3 e5a5 a2a4 a5c5 d4c4 c5d5 d3d5 c6d5 c4d4 e7e2 d4d5');
INSERT INTO evaluations VALUES(55851,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55851,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(55851,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55851,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55851,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(55851,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55851,36,'cp -21','e2e4 d5e4 d3e4 g8f6 e4h4 b8c6 g1f3 c6e7 f1d3 e7f5 h4f4 f8d6 f4d2 e8g8 d2e2 f8e8 e1g1 b7b6 c1g5 h7h6 g5f6 d8f6 c2c3 c8b7 f1e1 f5h4 f3h4 f6h4 g2g3 h4d8 d3e4 b7e4 e2e4 c7c5 b1d2 c5d4 e4d4 d8c7 d4e4 a8b8 a1d1');
INSERT INTO evaluations VALUES(278,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(278,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(278,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(278,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(278,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(278,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(278,36,'cp 35','g8e7 d1e2 d7d6 c3d5 e7d5 e4d5 c6d4 f3d4 c5d4 c2c3 d4b6 a2a4 a7a6 d3d4 e8g8 a4a5 b6a7 d4e5 f6g6 e1g1 f8e8 e5e6 f7e6 c1e3 a7e3 e2e3 c8d7 f1e1 e6e5 c4d3 d7f5 d3f5 g6f5 a1a4 a8b8 h2h3 e8f8 f2f3 f5f7 c3c4 h7h6 e3c3 f7d7 a4b4 b7b6 a5b6 b8b6');
INSERT INTO evaluations VALUES(4014,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(4014,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(4014,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(4014,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(4014,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(4014,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(4014,36,'cp 28','a7a6 e1g1 h7h6 g5h4 b7b6 d1c2 c8b7 a1d1 e7e6 h2h3 c7c5 d3e2 d8c7 h4g3 f8d8 f1e1 a8c8 a2a3 c5d4 e3d4 e6e5 c2b1 e5d4 d1d4 d7e5 f3e5 d6e5 d4d1 c7e7 e2f1 d8d1 b1d1 f6h5 g3h2 e5e4 d1b3 g7d4 c3d5 b7d5 c4d5');
INSERT INTO evaluations VALUES(6403,30,'cp 354',NULL);
INSERT INTO evaluations VALUES(6403,31,'cp 374',NULL);
INSERT INTO evaluations VALUES(6403,32,'cp 350',NULL);
INSERT INTO evaluations VALUES(6403,33,'cp 352',NULL);
INSERT INTO evaluations VALUES(6403,34,'cp 370',NULL);
INSERT INTO evaluations VALUES(6403,35,'cp 382',NULL);
INSERT INTO evaluations VALUES(6403,36,'cp 379','f3g5 g8h6 f1e1 e8f8 d1f3 d8e7 c1f4 d7d6 e5e6 e7f6 b1d2 c6e5 f4e5 d6e5 f3g3 c5d6 d2e4 f6e7 e6f7 h6f5 g3h3 h7h5 h3d3 g7g6 c2c3 f8g7 c3d4 f5d4 e4d6 e7d6 d3g3 b7b5 c4b3 e5e4 g3d6 c7d6 b3d5 c8f5 d5a8 h8a8 a1d1 d4c2');
INSERT INTO evaluations VALUES(6402,30,'cp 389',NULL);
INSERT INTO evaluations VALUES(6402,31,'cp 375',NULL);
INSERT INTO evaluations VALUES(6402,32,'cp 371',NULL);
INSERT INTO evaluations VALUES(6402,33,'cp 363',NULL);
INSERT INTO evaluations VALUES(6402,34,'cp 348',NULL);
INSERT INTO evaluations VALUES(6402,35,'cp 341',NULL);
INSERT INTO evaluations VALUES(6402,36,'cp 381','g8h6 f1e1 e8f8 d1f3 d8e7 c1f4 d7d6 e5e6 e7f6 b1d2 c6e5 f4e5 d6e5 f3g3 c5d6 d2e4 f6e7 e6f7 h6f5 g3h3 h7h5 a1d1 g7g6 h3b3 f8g7 c4e6 b7b6 c2c3 h8f8 e6c8 a8c8 g5e6 g7h6 c3d4 e5d4 e4d6 e7d6 b3c4 c7c5 e6f8 c8f8');
INSERT INTO evaluations VALUES(6455,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(6455,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(6455,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(6455,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(6455,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(6455,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(6455,36,'cp -54','a7a5 d2d3 d7d6 c1e3 g8e7 d3d4 e5d4 f3d4 c8e6 b1c3 e8g8 f1e2 c6d4 e3d4 e7c6 d4c5 d6c5 e1g1 c6d4 c3b5 b7b6 b5d4 d8d4 d1d4 c5d4 f2f4 c7c5 e2d3 g7g6 a1e1 f8e8 h2h3 e6d7 e4e5 g8g7 g2g4 f7f6 d3e4 a8d8 e4d5');
INSERT INTO evaluations VALUES(8096,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(8096,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(8096,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8096,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(8096,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(8096,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(8096,36,'cp 8','c4c5 f6e4 g5e7 d8e7 f1b5 e4c3 b2c3 c8d7 e1g1 e8g8 f1e1 c6a5 b5d3 b7b6 f3e5 a5c6 e5c6 d7c6 c5b6 a7b6 d1b1 h7h6 b1b6 c6a4 e1b1 f8c8 b6b2 a4e8 b2d2 e7a3 b1c1 a3a5 d3c2 e8a4 a1b1 a4c2 c1c2 c8c4 b1b4 a8c8 a2a4 c4b4 c3b4 c8c2 d2c2 a5b4');
INSERT INTO evaluations VALUES(8611,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(8611,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(8611,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(8611,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(8611,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(8611,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(8611,36,'cp 24','e1g1 e8g8 d2d3 e5f4 c1f4 c6a5 c4b3 c8d7 g1h1 b7b5 d1e2 a5b3 a2b3 a7a5 h2h3 h7h6 f4h2 b5b4 c3d1 c7c5 e4e5 d6e5 f3e5 a8a6 d1e3 d7e6 e3c4 f8e8 e2f2 e7f8 h1g1 d8c7 e5f3 c7b7 a1a5 a6a5 c4a5 b7a6');
INSERT INTO evaluations VALUES(8610,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(8610,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8610,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(8610,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(8610,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(8610,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(8610,36,'cp 13','c8g4 c4e2 e5f4 d2d3 d6d5 c1f4 e8g8 c3d5 f6d5 e4d5 d8d5 c2c3 e7d6 f4d6 d5d6 d3d4 c6e7 e2c4 g4e6 c4e6 d6e6 f1e1 e6d7 f3e5 d7d6 d1f3 e7d5 e1e2 a8e8 a1e1 c7c6 a2a4 f7f6 e5d3 e8e2 e1e2 d5c7 a4a5 f8e8 e2e8 c7e8 a5a6 b7a6 f3e4 e8c7 d3c5 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(24059,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(24059,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(24059,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(24059,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(24059,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(24059,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(24059,36,'cp 31','c4e2 e5f4 d2d4 d6d5 c3d5 f6d5 e4d5 d8d5 c1f4 g4f3 e2f3 d5d4 g1h1 e8c8 d1e2 h8e8 f3c6 b7c6 e2a6 c8b8 a6c6 d4b6 c6c4 e7f6 f1f3 b8c8 c4f7 b6c6 f7h5 e8e2 h5h7 f6b2 a1b1 c6c2 h7h3 d8d7 b1f1 c2e4 f4e3 b2f6 e3a7 e2b2 f3f2 b2f2 a7f2 e4e2');
INSERT INTO evaluations VALUES(8684,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(8684,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(8684,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(8684,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(8684,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(8684,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(8684,36,'cp 25','h7h6 g5f4 a7a6 a2a4 c8e6 b1c3 e8g8 a1d1 d8d7 c4e2 a8d8 d4e3 d7c6 e2f3 f6g4 f3g4 e6g4 c3d5 c6d7 d1d4 g4e6 d5e7 d7e7 h2h3 d8e8 f1e1 f7f6 f2f3 e7f7 e3c3 e8c8 f4g3 f8e8 e1d1 g8h8 g1h2 c8d8 b2b3 e8e7');
INSERT INTO evaluations VALUES(9668,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(9668,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(9668,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(9668,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(9668,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(9668,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(9668,36,'cp 55','g1e2 d7d5 c1g5 d5c4 f3f6 d8f6 g5f6 h8g8 d3c4 c8e6 b1d2 g6g5 h2h4 g5h4 f6h4 c6b4 e1c1 g8g2 d1g1 g2g1 h1g1 b4a2 c1b1 a2b4 g1g8 c5f8 g8h8 e8d7 e2c3 a8e8 h8h7 f8e7 h4g3 e7d6 c3d5 d7c8 d2f3 b4c6 d5f6 e8d8 b2b3 a7a6 h7h5 c8b8 f3e5 d6e5 g3e5 c6e5 h5e5');
INSERT INTO evaluations VALUES(13793,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(13793,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13793,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(13793,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(13793,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(13793,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(13793,36,'cp 0','f1d3 c8g4 e1g1 f8e7 d3e2 e8g8 c1f4 f8e8 h2h3 g4h5 a1c1 h7h6 f3e5 h5e2 c3e2 d8b6 e5c6 b7c6 c1c2 a8c8 e2c3 e7f8 f4e5 f6d7 e5f4');
INSERT INTO evaluations VALUES(16033,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(16033,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(16033,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(16033,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(16033,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(16033,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(16033,36,'cp 33','e2e3 e7e6 c2c4 f8e7 c4d5 f6d5 f4g3 e8g8 e3e4 d5f6 b1c3 c7c5 d4d5 e6d5 e4d5 f6h5 f1d3 h5g3 h2g3 d7f6 e1g1 a7a6 a2a4 d8c7 d1b3 c8g4 f1e1 f8e8 f3e5 e7f8 e5c4 a8d8 e1e8 f6e8 a1e1 e8d6 c3e4 d6e4 e1e4 g4d7 d5d6 c7c6');
INSERT INTO evaluations VALUES(16478,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(16478,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(16478,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(16478,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(16478,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(16478,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(16478,36,'cp -25','e5e4 f3e5 c6e7 d4d5 c7c6 d5d6 e7f5 e5c4 b7b5 g2g4 b5c4 g4f5 d8b6 f1g2 b6c5 c3e4 f6e4 g2e4 f8d6 e4g2 e8g8 e3e4 d6e5 e1g1 c4c3 a1b1 d7d5 b2c3 c8a6 f1e1 e5c3 c1d2 d5e4 d2c3 c5c3 g2e4 c3f6 d1f3 a6c4 e4c6 a8c8 c6e4 h7h6 g1g2 f8e8');
INSERT INTO evaluations VALUES(16894,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(16894,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(16894,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16894,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(16894,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16894,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(16894,36,'cp 8','c7c6 b5c4 d4f3 d1f3 f8c5 c4b3 e8g8 d2d3 d7d6 c1e3 c8g4 f3g3 c5e3 g3e3 d8b6 e3b6 a7b6 a2a4 g4e6 a1d1 e6b3 c2b3 d6d5 e4d5 f6d5 d3d4 d5c3 b2c3 e5d4 d1d4 f8d8 f1d1 d8d4 d1d4 b6b5 g1f1 b5a4 d4a4 a8a4 b3a4 f7f6 f1e2 g8f7 e2d3 f7e7 d3d4 e7d6 g2g4 h7h6');
INSERT INTO evaluations VALUES(17999,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(17999,31,'cp 178',NULL);
INSERT INTO evaluations VALUES(17999,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(17999,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(17999,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(17999,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(17999,36,'cp 183','f1e1 g8f6 e4d5 e5e4 d2d3 c6d5 d3e4 d5e4 d1d8 e8d8 f3g5 d8e8 g5e4 f6e4 e1e4 c8e6 b2b3 h8g8 c1a3 a8d8 e4e1 f8a3 b1a3 e8e7 g1f1 e7d7 a1c1 d7c8 a3b1 h7h5 b1c3 h5h4 h2h3 g7g5 f2f3 e6f5 c3e4 f5e4 e1e4 d8d2');
INSERT INTO evaluations VALUES(18390,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(18390,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(18390,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(18390,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(18390,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(18390,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(18390,36,'cp 146','c4b3 f8b4 c1d2 b4d2 b1d2 d7d6 d1h5 d8e7 a2a4 g8f6 h5h4 e8g8 e1g1 a7a5 f2f4 b5a4 b3a4 c6c5 f4f5 h7h6 g2g4 d6d5 f1e1 d5e4 d2e4 a8a7 e4f6 e7f6 h4f6 g7f6 a4c6 f8d8 b2b3 g8f8 a1a4 h6h5 h2h3 a7a6 c6b5 a6a8 b5c4 c8d7');
INSERT INTO evaluations VALUES(22887,30,'cp -189',NULL);
INSERT INTO evaluations VALUES(22887,31,'cp -204',NULL);
INSERT INTO evaluations VALUES(22887,32,'cp -190',NULL);
INSERT INTO evaluations VALUES(22887,33,'cp -198',NULL);
INSERT INTO evaluations VALUES(22887,34,'cp -213',NULL);
INSERT INTO evaluations VALUES(22887,35,'cp -201',NULL);
INSERT INTO evaluations VALUES(22887,36,'cp -205','f8e7 g5f3 g8f6 f3d4 f6e4 d4c6 d7c6 d1d8 e7d8 f2f3 e4d6 c1f4 d8f6 b1c3 c8f5 e1c1 e8c8 h2h4 h7h5 c3e2 d8e8 b2b3 f5g6 f4g5 f6g5 h4g5 e8e5 f3f4 e5e8 e2d4 g6e4 d4f3 c6c5 f3e5 f7f6 e5f3 e4f3 g2f3 e8e3 f1g2 c5c4 b3c4 d6c4 g2h3 c8b8 d1d7 f6g5 f4g5');
INSERT INTO evaluations VALUES(23330,30,'cp 208',NULL);
INSERT INTO evaluations VALUES(23330,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(23330,32,'cp 223',NULL);
INSERT INTO evaluations VALUES(23330,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(23330,34,'cp 203',NULL);
INSERT INTO evaluations VALUES(23330,35,'cp 194',NULL);
INSERT INTO evaluations VALUES(23330,36,'cp 199','g8e7 d4d3 e7c6 d3g3 c6b4 c3b5 d7d6 g3b3 b4c6 c1e3 c6a5 b3c3 a5c6 c3c4 c8g4 b5c3 a7a6 c3d5 g4h5 h2h4 f8e7 c4c3 e8g8 c3b3 h5f7 b3b7 f7d5 e4d5 c6e5 e1c1 e5g4 e3d4 f6f5 b7b3 e7h4 g2g3 h4f6');
INSERT INTO evaluations VALUES(23442,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(23442,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(23442,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(23442,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(23442,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(23442,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(23442,36,'cp -45','e7e5 b1c3 g8f6 g1f3 f8c5 f1e2 e8g8 e1g1 d7d6 d2d3 a7a6 c1b2 h7h6 d1d2 c8e6 c3d1 c6e7 f1e1 e7g6 h2h3 f8e8 e2f1 b7b5 a2a3 c5b6 c2c4 b5c4 d3c4 a8b8 b3b4 c7c6 d2c2 b6a7 g2g3 d8d7');
INSERT INTO evaluations VALUES(24582,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(24582,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(24582,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(24582,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(24582,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(24582,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(24582,36,'cp -13','e7e6 d1e2 d8c7 c1g5 f8d6 g5f6 d7f6 c2c3 e8g8 g2g3 h7h6 f1c1 f8e8 a2a4 a8c8 b1d2 a7a5 f3e1 b7b6 e1d3 c7b7 a1a2 f6e4 a2a1 c8b8 h2h4 e8c8 c1c2 e4f6 d2f3 b6b5 a4b5 b7b5 d3e5 b5e2 c2e2');
INSERT INTO evaluations VALUES(27853,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(27853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27853,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(27853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27853,36,'cp 0','g4f3 d1f3 f6d7 d4d5 g7c3 b2c3 d7e5 f3g3 e5d3 g3d3 b8d7 e1g1 d7c5 d3e2 e7e5 d5e6 c5e6 f2f4 f8e8 c1e3 b7b6 e2f3 e6c5 f4f5 d8d7 e3c5 b6c5 e4e5 d6e5 a1d1 d7e7 f3c6 a8d8 f5f6 e7f8 d1d5 e5e4 d5d8 e8d8 c6e4');
INSERT INTO evaluations VALUES(27859,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(27859,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27859,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27859,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(27859,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(27859,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(27859,36,'cp 0','d1f3 f6d7 d4d5 g7c3 b2c3 d7e5 f3g3 e5d3 g3d3 b8d7 e1g1 d7c5 d3e2 e7e5 d5e6 c5e6 c1e3 f8e8 f2f4 b7b6 e2f3 e6c5 e3c5 b6c5 a1d1 d8e7 e4e5 d6e5 f4e5 a8b8 f3f2 a7a5 d1d5 a5a4 f2c5 e7c5 d5c5 e8e7 c5a5 b8b2 f1f2 b2b1 f2f1');
INSERT INTO evaluations VALUES(27858,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27858,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(27858,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27858,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27858,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27858,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(27858,36,'cp -16','f6d7 d4d5 g7c3 b2c3 d7e5 f3e2 e5d3 e2d3 b8d7 e1g1 d7c5 d3e2 e7e5 d5e6 c5e6 c1e3 d8e7 a1e1 b7b6 f2f4 f7f6 a2a4 a7a5 e2g4 e6c5 e3c5 b6c5 e1e3 a8e8 f1e1 e7f7 g4e2 e8e6 e1b1 f7e7 f4f5 e6e5 b1b7 g8g7 h3h4 e7f7');
INSERT INTO evaluations VALUES(28398,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(28398,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(28398,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28398,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(28398,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(28398,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28398,36,'cp 7','b8d7 d1b3 e8g8 c4d5 f6d5 g5e7 d8e7 f1e2 d5c3 b3c3 c7c5 e1g1 c5d4 f3d4 f5e4 a1c1 d7b6 f1d1 h7h6 c3a5 b6d5 a5c5 e7f6 c5a3 f8d8 h2h3 a7a6 a3b3 b7b5 f2f3 f6g5 d4e6 f7e6 f3e4 d5e3 d1d8 a8d8 b3e6 g8h7');
INSERT INTO evaluations VALUES(32110,30,'cp 225',NULL);
INSERT INTO evaluations VALUES(32110,31,'cp 248',NULL);
INSERT INTO evaluations VALUES(32110,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(32110,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(32110,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(32110,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(32110,36,'cp 224','d1f3 d8d7 f3b3 b7b6 d4e5 d6e5 b1c3 a7a6 c3d5 b8c6 b3g3 g8f6 c1e3 f6e4 g3f3 e4d6 e1c1 e8c8 c4a6 c8b8 h1e1 d7f5 e3b6 c7b6 f3c3 d8c8 a6c8 f5c8 f2f4 d6b5 c3c4 b5d6 c4b3 c6d4 b3d3 c8c6 e1e5 d4e6 e5e1 d6c4 f4f5 e6c7 d5c7 b8c7');
INSERT INTO evaluations VALUES(32109,30,'cp 202',NULL);
INSERT INTO evaluations VALUES(32109,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(32109,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(32109,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(32109,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(32109,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(32109,36,'cp 205','d8d7 f3b3 b7b6 d4e5 d6e5 b1c3 a7a6 c3d5 b8c6 b3g3 g8f6 c1e3 f6e4 g3f3 e4d6 e1c1 e8c8 c4a6 c8b8 h1e1 e5e4 f3e2 d7e6 e3f4 f8e7 c1b1 g7g5 d5e7 c6b4 e7c6 b4c6 f4d6 d8d6 d1d6 e6d6 e2e4 h8d8 a2a3 h7h5 b1a2 h5h4 h2h3 d6d5 e4d5 d8d5 e1e8 d5d8 e8d8 c6d8 a6e2 d8e6 e2c4 e6d4 c4f7 d4c2');
INSERT INTO evaluations VALUES(33611,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(33611,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(33611,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(33611,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(33611,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(33611,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(33611,36,'cp 24','d2d4 b8c6 d4c5 d5c5 c2c3 g7g6 c1e3 c5a5 f1c4 f8g7 e1g1 e8g8 d1e2 a5c7 f1d1 c8g4 c4d5 c6e5 h2h3 g4f3 d5f3 e5f3 e2f3 a8d8 a2a4 b7b6 d1d8 f8d8 a1d1 d8d1 f3d1 g7f6 d1d5 g8g7 g2g3 e7e6 d5b5 h7h5 h3h4 c7d6 b5c4 d6d1 g1h2');
INSERT INTO evaluations VALUES(33669,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(33669,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(33669,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(33669,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(33669,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(33669,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(33669,36,'cp 112','a7a6 g1f3 e6d6 e1g1 e7e6 f3e5 b8c6 e5c6 d6c6 e2f3 c6d6 c3e2 f8e7 c2c4 c7c5 c1f4 d6b6 d4c5 b6c5 d1b3 e8g8 e2c3 e6e5 f4e3 c5b4 f1d1 b4b3 a2b3 a8b8 c3d5 f6d5 d1d5 c8e6 d5e5 e7f6');
INSERT INTO evaluations VALUES(35990,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(35990,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(35990,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(35990,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(35990,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(35990,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(35990,36,'cp -69','f6g4 c3d1 b8c6 d4c6 b7c6 e3d4 g4f6 d1c3 a8b8 b2b3 c6c5 d4f6 g7f6 a1d1 c8b7 e1g1 f6c3 d2c3 b7e4 f1e1 e4f5 h2h3 f5c8 c3e3 b8b7 d1d2 b7d7 c4b5 d7c7 c2c4 a7a6 b5a4 c8b7 e1d1 d8a8 f2f3 b7c6 a4c6 a8c6 d2d3');
INSERT INTO evaluations VALUES(37513,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(37513,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(37513,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(37513,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(37513,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(37513,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(37513,36,'cp 78','c1e3 d7d6 d1d2 g8f6 f2f3 h7h5 h2h3 b8c6 e1c1 c8d7 f1d3 b7b5 g2g4 b5b4 d4c6 d7c6 c3e2 a6a5 c1b1 a8c8 e3d4 e7e5 d4e3 h5g4 h3g4 h8h1 d1h1 d6d5 e3h6 g7h8 h6g5 h8g7 e2g3 d5e4 g3e4 c6e4 f3e4 d8c7 g5e3 f6g4 d3b5 e8f8');
INSERT INTO evaluations VALUES(37512,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(37512,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(37512,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(37512,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(37512,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(37512,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(37512,36,'cp 75','d7d6 d1d2 g8f6 f2f3 h7h5 h2h3 b8d7 f1e2 b7b6 e1c1 c8b7 c1b1 b6b5 g2g4 a8c8 a2a3 d7c5 h1e1 d8c7 e2f1 h5g4 h3g4 f6d7 c3d5 b7d5 e4d5 c5a4 e3g5 e8g8 c2c3 d7b6 d4c6 b6d5 c6e7 d5e7 e1e7 c7c6 f1e2 f8e8 d2d6 c6d6 d1d6 g7c3');
INSERT INTO evaluations VALUES(37967,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(37967,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(37967,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(37967,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(37967,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37967,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(37967,36,'cp 77','d2d4 c5d4 f3d4 b8c6 c1e3 g7d4 e3d4 c6d4 d1d4 g8f6 f1e2 e8g8 e1g1 d8b6 d4b6 a7b6 a2a4 h7h5 b2b3 c8d7 f2f3 g8g7 h2h4 d7e6 g1f2 f6d7 f2e3 d7c5 f1b1 f8c8 e3d2 f7f5 e4f5 e6f5 b1b2 f5e6 a1d1 g7f7 d2e3 c5d7');
INSERT INTO evaluations VALUES(37969,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(37969,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(37969,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(37969,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(37969,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37969,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(37969,36,'cp 52','c5d4 f3d4 b8c6 c1e3 c6d4 e3d4 g7d4 d1d4 g8f6 f1e2 e8g8 e1g1 d8a5 c3d5 f6d5 e4d5 e7e5 d5e6 c8e6 f1d1 a8c8 b2b3 f8d8 e2f3 d8d7 f3g4 e6g4 d4g4 d7d8 g4d4 a5c5 d4d5 c5d5 d1d5 g8f8 a1d1 f8e7 g1f1 c8c5 f1e2 c5d5 d1d5 e7d7 e2d3 d8c8 g2g3');
INSERT INTO evaluations VALUES(38194,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38194,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38194,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38194,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38194,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38194,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38194,36,'cp 0','b1c3 g8f6 g1f3 e8g8 e1g1 b8c6 c1g5 h7h6 g5f6 d8f6 c3d5 f6b2 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(40105,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40105,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(40105,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40105,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(40105,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(40105,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(40105,36,'cp 24','h7h6 g5h4 e8g8 f1d3 f8e8 g1e2 b8d7 e1g1 a7a5 h2h3 d7f8 h4g3 e7d6 g3d6 d8d6 a2a3 f8e6 e2g3 c8d7 c3a4 d6c7 a1c1 b7b6 c2b3 a8b8 a4c3 c7d8 f1d1 b6b5 b3c2 d8e7 c3e2');
INSERT INTO evaluations VALUES(41278,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(41278,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41278,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41278,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(41278,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41278,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41278,36,'cp 0','e8g8 g1e2 c7c5 e1g1 e7e6 d4d5 b8d7 d5e6 f7e6 e3e4 d7e5 c1g5 e5f7 g5c1');
INSERT INTO evaluations VALUES(41330,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(41330,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(41330,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(41330,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(41330,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(41330,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(41330,36,'cp -5','e3e4 e8g8 g1f3 c7c5 d4c5 d8a5 b1c3 a5c5 d1e2 b8c6 a2a3 e7e5 f4e5 d6e5 h2h3 c6d4 e2f2 c8e6 c1e3 f6h5 e1g1 f7f5 e4f5 e6f5 c3e4 c5c7 a1e1 f5e4 d3e4 d4f3 g2f3 h5f6 f2h4 f6e4 h4e4 a8c8 f1f2 f8d8 c2c3 b7b6 e3g5 d8d7 f2d2 d7d2 g5d2');
INSERT INTO evaluations VALUES(44062,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(44062,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(44062,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(44062,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(44062,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(44062,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(44062,36,'cp -36','f8b4 c1d2 d7d6 d5c6 b7c6 g1f3 e8g8 e2e3 c8f5 f1e2 a7a5 e1g1 b8d7 a2a3 b4c3 d2c3 f6e4 c3e1 a5a4 f3d2 d8e7 d2e4 f5e4 f2f3 e4g6 e1b4 d7c5 d1d2 f8d8 a1d1 h7h6 g1h1 a8b8 b4c5 d6c5 d2c1 e7f6 d1d8 f6d8 f1d1 d8f6 e3e4 b8b3 e2f1');
INSERT INTO evaluations VALUES(39878,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(39878,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(39878,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(39878,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(39878,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(39878,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(39878,36,'cp -42','c1d2 d7d6 d5c6 b7c6 e2e3 a7a5 g1f3 e8g8 f1e2 c8f5 e1g1 b8d7 a2a3 b4c3 d2c3 f6e4 c3e1 a5a4 f3d2 d7c5 d2e4 f5e4 f2f3 e4g6 e1b4 d8f6 d1d2 f8d8 e2d1 g6d3 d1e2 d3e2 d2e2 c5b3 a1d1 f6e6 b4c3 h7h6 f3f4 b3c5 f4e5 d6e5 d1d8 a8d8 f1d1 d8b8 c3b4');
INSERT INTO evaluations VALUES(44398,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(44398,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(44398,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(44398,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(44398,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(44398,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(44398,36,'cp 80','c1g5 f6e4 c3e4 f5e4 f2f3 d7d5 d1d2 b8d7 c4d5 h7h6 g5h4 c6d5 e2e3 d8b6 g1e2 e7e5 a2a3 f8d6 h4g3 e5d4 g3d6 e4f3 d2b4 f3e2 f1e2 b6b4 a3b4 d4e3 a1c1 d7f6 c1c7 c8d7 c7b7 f6e4 d6f4 e8g8 f4e3 e4f6 b7a7 a8a7 e3a7');
INSERT INTO evaluations VALUES(44778,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(44778,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(44778,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(44778,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(44778,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(44778,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(44778,36,'cp 77','f8g7 e2e4 e8g8 h2h3 e7e6 f1d3 e6d5 e4d5 f8e8 c1e3 a7a6 e1g1 b8d7 d1d2 b7b5 c4b5 d7b6 b5a6 c8a6 d3a6 a8a6 b2b3 d8a8 a1d1 f6e4 c3e4 e8e4 d2d3 e4e8 d3b5 e8b8 b5c6 b8c8 c6d6 g7f8 d6f4 b6d5 f4c4 d5e3 f2e3 a6a2');
INSERT INTO evaluations VALUES(45221,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45221,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(45221,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(45221,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45221,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(45221,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45221,36,'cp 0','b8c6 a2a4 e7e6 f1e2 f8e7 e1g1 a7a6 d1e1 e8g8 b1c3 a8b8 e1g3 f6h5 g3f2 h5f6');
INSERT INTO evaluations VALUES(45353,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(45353,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(45353,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(45353,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(45353,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(45353,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(45353,36,'cp -25','b8c6 e2e3 c8g4 f1e2 e7e6 e1g1 a8c8 h2h3 c5d4 e3d4 g4f3 e2f3 f8d6 f4d6 d8d6 c3b5 d6b8 a2a4 a7a6 b5a3 b7b5 a4b5 a6b5 b2b4 c6b4 d1e2 b8b6 c2c3 b4c6 e2b5 b6b5 a3b5 e8g8 c3c4 d5c4 b5d6 c6d4 d6c8 d4f3 g2f3 f8c8');
INSERT INTO evaluations VALUES(13088,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(13088,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(13088,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(13088,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(13088,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(13088,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(13088,36,'cp -19','e2e3 c8g4 f1b5 e7e6 h2h3 g4f3 d1f3 c5d4 e3d4 d8b6 f3d3 f8b4 e1g1 e8g8 b5c6 b4c3 d3c3 f8c8 c3d2 b7c6 c2c3 f6e4 d2e2 c6c5 f1d1 c5c4 d1e1 a7a5 a2a4 c8e8 a1d1 a8c8 d1a1 f7f6 a1d1 b6c6 f2f3');
INSERT INTO evaluations VALUES(13087,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(13087,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(13087,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(13087,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(13087,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(13087,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(13087,36,'cp -17','c8g4 f1b5 e7e6 h2h3 c5d4 e3d4 g4f3 d1f3 d8b6 f3d3 f8b4 e1g1 e8g8 b5c6 b4c3 d3c3 f8c8 c3d2 b7c6 c2c3 c6c5 f2f3 c5d4 c3d4 a7a5 f4e3 b6a6 f1c1 a5a4 c1c8 a8c8 a1c1 c8c1 d2c1 h7h6 e3f4 a6c4 c1c4 d5c4 f4d6 f6d5');
INSERT INTO evaluations VALUES(49715,30,'cp -300',NULL);
INSERT INTO evaluations VALUES(49715,31,'cp -302',NULL);
INSERT INTO evaluations VALUES(49715,32,'cp -294',NULL);
INSERT INTO evaluations VALUES(49715,33,'cp -299',NULL);
INSERT INTO evaluations VALUES(49715,34,'cp -303',NULL);
INSERT INTO evaluations VALUES(49715,35,'cp -316',NULL);
INSERT INTO evaluations VALUES(49715,36,'cp -291','b8c6 d2d3 d7d5 f5g5 c8e6 e4d5 d8d5 b1c3 d5a5 a2a3 e8c8 f1e2 f6d5 g5d2 f8b4 e2f3 c6d4 g1e2 d4f3 g2f3 b4e7 c3d5 a5d5 d2e3 d5c6 e3a7 c6c2 a7a8 c8d7 a8b7 c2d3');
INSERT INTO evaluations VALUES(52161,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(52161,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(52161,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(52161,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(52161,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(52161,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(52161,36,'cp 125','g7g6 b1c3 d7d6 f1e2 h7h5 d1a4 c8d7 a4b3 d7c8 g1f3 f8h6 c1h6 g8h6 e1c1 e8f8 h2h3 f8g7 d1g1 b8d7 g2g3 a7a6 c1b1 d7f8 f3d2 d8g5 d2f1 f8h7 f2f3 f7f5 h3h4 g5f6 b1a1 h8f8 b3c2 a8b8 f1d2 c8d7 e4f5 h6f5 c3e4');
INSERT INTO evaluations VALUES(53077,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(53077,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53077,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(53077,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53077,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53077,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(53077,36,'cp 11','d2d4 c5d4 d1d4 b8c6 d4f2 g8f6 b1c3 g7g6 f1c4 f8g7 g1e2 c8e6 c4e6 f7e6 c1e3 e8g8 f2h4 b7b5 e1g1 d8c8 b2b3 f6h5 a1d1 a7a6 h4e1 c8d7 g1h1 a8c8 g2g4 h5f6 e2f4 e6e5 f4d5 g8h8 c3e2 f6d5');
INSERT INTO evaluations VALUES(53937,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(53937,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(53937,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(53937,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(53937,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(53937,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(53937,36,'cp 80','e4d5 g7g6 b1c3 f8g7 f1d3 d8e7 d1e2 e7e2 e1e2 b8d7 g1f3 d7e5 f3e5 g7e5 h1e1 g8e7 c1g5 f7f6 g5d2 g6g5 c3b5 e8d7 a1b1 e7g6 e2f1 a7a6 b5c3 g6f4 d3e4 d7c7 b2b4 c5b4 b1b4 b7b6 e4f3 c8d7 e1b1');
INSERT INTO evaluations VALUES(54447,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(54447,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(54447,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(54447,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(54447,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(54447,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(54447,36,'cp 43','e3d4 g8f6 e1g1 f8e7 f1e1 e8g8 b1c3 b8c6 a2a3 b7b6 c4a2 c8b7 c1f4 f8e8 f4g3 a8c8 d1d3 e7d6 g3h4 c6e7 h4f6 g7f6 d4d5 e7d5 c3d5 b7d5 a2d5 e6d5 f3d4 d6e5 d4f5 g8h8 h2h3 e5b2 a1d1 e8e1 d1e1 d8d7');
INSERT INTO evaluations VALUES(54597,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(54597,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(54597,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(54597,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(54597,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(54597,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(54597,36,'cp -32','c4b3 b8c6 g1e2 f8c5 b1c3 g8f6 e1g1 d5d4 c3e4 f6e4 d3e4 e8g8 c2c3 a7a5 c3d4 c6d4 e2d4 c5d4 b3d5 c8e6 d5e6 f7e6 d1e2 a5a4 c1e3 d8b6 e3d4 e5d4 h2h3 b6a6 f1e1 a6e2 e1e2 a4a3 b2a3 a8a3 a1b1 d4d3');
INSERT INTO evaluations VALUES(717,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(717,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(717,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(717,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(717,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(717,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(717,36,'cp 42','b6d4 d2f3 d4b6 a2a3 f8e7 c1e3 e7c5 e3d2 e8g8 a1c1 f7f5 c1c3 c5e7 d1c2 b6d8 f1c1 d7b6 f3d4 c8d7 c2b3 a8b8 d2e3 d8e8 h2h3 e7d8 f2f4 f8f7 b3b4 b6c4 d3c4 d5c4 c3c4 d8b6 b4b3 b8c8 c4c8 d7c8 d4b5 b6e3 b3e3');
INSERT INTO evaluations VALUES(752,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(752,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(752,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(752,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(752,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(752,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(752,36,'cp 42','d2f3 d4b6 a1b1 f8e7 f1e1 d7c5 c1e3 b6d8 d3c2 c8d7 d1d4 h7h5 b1c1 c5e4 e1d1 d7c6 d4d3 d8b8 a2a3 a7a6 b2b4 c6b5 d3d4 e8g8 c2e4 d5e4 d4e4 b5c6');
INSERT INTO evaluations VALUES(751,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(751,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(751,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(751,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(751,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(751,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(751,36,'cp 38','d4b6 a2a3 f8e7 d3c2 d7b8 c1e3 b6c7 a1c1 b8c6 h2h4 c8d7 c2b1 a7a5 f1e1 h7h6 h4h5 a5a4 c1c3 c7b8 d1e2 c6a5 e3f4 b8d8 f3d2 a8c8 e2g4 e7g5 c3g3 d8e7 f4g5 e7g5 g4b4 a5c6 g3g5 c6b4 g5g7 e8f8 g7g4 b4c6 d2f3 f8e8 f3d4');
INSERT INTO evaluations VALUES(719,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(719,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(719,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(719,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(719,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(719,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(719,36,'cp 34','a2a3 f8e7 c1e3 e7c5 e3d2 e8g8 a1c1 f7f5 c1c3 c5e7 d1c2 b6d8 d2e3 d7e5 f3e5 d5d4 d3f5 f8f5 e3d4 d8d4 c3c8 f5f8 c8f8 a8f8 e5f3 d4c5 c2e2 c5d5 f1d1 d5a2 e2e3 e7f6 b2b4 a7a6 d1e1 h7h6 h2h4 f8d8 h4h5 a2d5 e3e6 d5e6 e1e6 d8d3 a3a4 f6c3 b4b5 a6b5 a4b5 g8f7 e6e4 d3d1 g1h2');
INSERT INTO evaluations VALUES(3199,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(3199,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(3199,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(3199,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(3199,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3199,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3199,36,'cp -8','a2a4 a8b8 c3d5 f6d5 e4d5 c6b4 f3e1 e7e6 d5e6 c8e6 c2c3 b4c6 e1f3 c6e5 c1e3 e6g4 d3d4 e5c4 e3f4 c5d4 c3d4 b8c8 d1b3 c4a5 b3b4 c8c4 b4a3 d6d5 f3e5 g7e5 d4e5 d5d4 a1c1 d4d3 b2b4 c4f4 g3f4 d3d2 c1d1 g4d1 f1d1 a5c6 b4b5 c6a5 a3b4');
INSERT INTO evaluations VALUES(5164,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(5164,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(5164,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(5164,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(5164,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(5164,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(5164,36,'cp 25','b7c6 f1d3 d7d5 e1g1 g8f6 f1e1 e8g8 c3a4 c5e7 e4e5 f6d7 c1f4 c6c5 c2c4 d5d4 d3e4 a8b8 d1g4 f8e8 b2b3 c8b7 e4b7 b8b7 a1d1 b7b8 d1d3 g7g6 d3h3 d8a5 g4e2 a5a6 f4d2 a6b7 f2f4 f7f5 e5f6 e7f6 a4b2 b7c7');
INSERT INTO evaluations VALUES(5632,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(5632,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(5632,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(5632,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(5632,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(5632,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(5632,36,'cp 13','g8f6 d2d3 e8g8 a2a4 a7a5 f1e1 c8e6 b1d2 e6c4 d2c4 h7h6 c1d2 f8e8 d1b3 d8d7 c4e3 c6d8 e3f5 d8e6 d3d4 e7f8 f5g3 c7c6 h2h3 d7c7 b3c2 b7b5 b2b3 a8c8 a4b5 c6b5 d4d5 e6c5 d2e3 g7g6 e3c5');
INSERT INTO evaluations VALUES(8588,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(8588,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(8588,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(8588,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(8588,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(8588,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(8588,36,'cp 32','d2d3 e8g8 a2a4 a7a5 d1b3 c8g4 b1d2 f6d7 h2h3 d7c5 b3d1 g4e6 f1e1 e7g5 c4e6 c5e6 d2c4 g5f4 c1f4 e6f4 c4e3 d8d7 d3d4 e5d4 e3d5 f4d5 e4d5 c6e7 d1d4 e7g6 b2b3 h7h6 c3c4 a8e8 d4c3 e8e1 f3e1 b7b6 e1f3 f8d8 a1e1');
INSERT INTO evaluations VALUES(8437,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(8437,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8437,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8437,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8437,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8437,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(8437,36,'cp -12','c6a5 a2a3 a5c4 d3c4 c8e6 d1d3 c7c6 e1c1 d8a5 f3d2 a8d8 f2f4 e5f4 e3f4 e8g8 d2b3 a5a6 b3d4 e6c4 d3e3 a6b6 d4f5 b6e3 f4e3 f8e8 b2b3 c4e6 f5e7 e8e7 e3a7 d8a8 a7d4 a8a3 c1b2 a3a5 d4f6 g7f6 d1d6 a5g5 d6d2 h6h5 h1f1 f6f5 d2d4 b7b5 b2c1 g5g2 e4f5');
INSERT INTO evaluations VALUES(9601,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(9601,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(9601,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(9601,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(9601,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(9601,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(9601,36,'cp -3','a2a4 c7c6 f1c4 h7h6 g5e3 e8g8 e1g1 e7b4 d1d3 d8c7 h2h3 d7c5 d3e2 a7a5 f3h4 f8e8 a1d1 c8e6 c4e6 c5e6 e2f3 e6d4 e3d4 e5d4 d1d4 c7e5 h4f5 a8d8 f1d1 d8d4 f5d4 b4c3 f3c3 f6e4 c3f3 e4f6 c2c3 f6d5 d4f5 d5f4 f3e3 e5e3');
INSERT INTO evaluations VALUES(10572,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(10572,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(10572,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(10572,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(10572,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(10572,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(10572,36,'cp 100','e1g1 a7a6 b5a4 b7b5 a4b3 e8g8 c3d5 h7h6 c2c3 d7d6 h2h3 g4f6 d3d4 c5a7 a2a4 b5b4 d5b4 c6b4 d4e5 f6e4 c3b4 c8b7 c1f4 a8b8 b4b5 a6b5 a4b5 a7c5 b3c2 f8e8 a1a4 d6e5 d1d8 b8d8 f3e5 g7g5 c2e4 b7e4 a4e4 g5f4');
INSERT INTO evaluations VALUES(10571,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(10571,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(10571,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(10571,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(10571,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(10571,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(10571,36,'cp 101','a7a6 b5a4 b7b5 a4b3 e8g8 c3d5 h7h6 c2c3 d7d6 h2h3 g4f6 d3d4 c5a7 a2a4 b5b4 d5b4 c6b4 d4e5 f6e4 c3b4 a8b8 b3c2 c8b7 b4b5 d6e5 d1d8 f8d8 b5a6 b7a8 b2b3 e4c5 c1e3 a8f3 g2f3 d8d6 b3b4 b8b4 f1d1 g8f8 d1d6 c7d6 a4a5 f8e7 g1g2 b4b2');
INSERT INTO evaluations VALUES(12309,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(12309,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(12309,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12309,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(12309,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(12309,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(12309,36,'cp 25','d7d6 d1e2 f8e7 e1g1 e8g8 e2e4 g8h8 c4b3 f7f5 e4e2 d5c7 e5d6 d8d6 b1c3 e7f6 f1d1 c6d4 f3d4 f6d4 e2f3 e6e5 c3e2 c7e6 c1e3 c8d7 b3e6 d6e6 e3d4 e5d4 e2d4 e6b6 a1c1 a8d8 h2h3 b6b2 c1c7 d7b5 c7b7 d8d4 d1d4 b2d4 b7b5');
INSERT INTO evaluations VALUES(13902,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(13902,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(13902,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(13902,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(13902,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(13902,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(13902,36,'cp 43','e1g1 f8e7 d1e2 e8g8 e2e4 g8h8 c4b3 f7f5 e4e2 d5c7 e5d6 d8d6 b1c3 e7f6 f1d1 c6d4 f3d4 f6d4 e2f3 e6e5 c3e2 c7e6 c1e3 c8d7 b3e6 d7c6 f3h3 d6e6 e3d4 e5d4 e2d4 e6f6 d4c6 f6c6 h3b3 f8d8 h2h3 h7h6 a1c1');
INSERT INTO evaluations VALUES(13901,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(13901,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13901,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(13901,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(13901,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(13901,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(13901,36,'cp 24','f8e7 d1e2 e8g8 e2e4 c8d7 c4b3 d6e5 d4e5 d8b6 b1c3 b6b4 e4d3 b4g4 h2h3 g4h5 c3d5 e6d5 b3d5 c6e5 f3e5 h5e5 d5b7 a8d8 d3e3 e7d6 e3e5 d6e5 f1e1 e5f6 c1e3 f6b2 a1b1 b2a3 b1b3 a3d6 b7d5 a7a5 g2g4 d8b8 e3a7 b8b3 d5b3 f8c8');
INSERT INTO evaluations VALUES(12323,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(12323,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(12323,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(12323,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(12323,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(12323,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(12323,36,'cp 86','b5c6 b7c6 e4e5 f6d5 d2d3 d7d6 b1d2 d8c7 d2c4 d6e5 f3e5 f8e7 c1d2 e8g8 d2a5 c7b7 f2f4 f7f6 e5g4 b7d7 b2b3 c8a6 d1f3 e7d8 a5d8 a8d8 c4d2 d5b4 g4e3 f8e8 f1e1 d7d4 f3f2 b4d5 d2e4 d5f4 f2f4 f6f5 f4h4 f5e4 h4e4 d4e4 d3e4 e8f8');
INSERT INTO evaluations VALUES(12702,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(12702,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(12702,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(12702,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(12702,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(12702,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(12702,36,'cp 67','b5c4 d7d5 c4b3 d8b6 d4d1 d5e4 b1c3 e7f5 e1g1 f8e7 c3e4 a7a5 c2c3 e8g8 c1f4 a5a4 b3a4 b6b2 a4b3 f5h4 f4g5 h4g6 g5e7 g6e7 f1e1 c8f5 e4d6 f5e6 d6b7 b2c3 b3e6 f7e6 d1d7 e7f5 d7e6 g8h8 e6e5 c3d2 a1d1 d2a2 b7c5 f5h4 e5g3 h4g6 h2h3 f8e8 c5d3 e8e1 d1e1');
INSERT INTO evaluations VALUES(12703,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(12703,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(12703,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(12703,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(12703,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(12703,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(12703,36,'cp 72','d7d5 c4b3 d8b6 d4d1 d5e4 b1c3 e7f5 e1g1 f8e7 c3e4 e8g8 c2c3 a7a5 c1f4 a5a4 b3a4 b6b2 a4b3 b2a3 f1e1 a3a5 a2a4 f5h4 e4d6 e7d6 f4d6 a5g5 d6g3 h4f5 d1f3 c8e6 b3e6 f7e6 g3f4 g5g6 f3e4 g6f6 f4e5 f6h4 e4h4 f5h4 a4a5 f8f5 e5c7 a8c8 c7b6 g8f7');
INSERT INTO evaluations VALUES(17133,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(17133,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(17133,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(17133,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(17133,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(17133,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(17133,36,'cp 155','c6a5 c4e2 c7c5 c3d5 e7d5 e4d5 f8d6 c2c3 b7b6 d2d4 c5d4 c3d4 e5e4 f3h4 g7g6 e2h5 d8c7 d1g4 c8a6 g4e4 e8d8 f1e1 d8c8 c1h6 c8b7 a1c1 c7d8 h5e2 a6e2 e4e2 h8e8 e2e8 d8e8 e1e8 a8e8 g1f1 b6b5 b2b3 b7b6 g2g3 a5b7 h6g7 d6e7 h4g2 b7d6 g2e3 f6f5 f1e2 e8c8 c1c8 d6c8 g7h6 e7d6 h6g5');
INSERT INTO evaluations VALUES(17884,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(17884,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(17884,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(17884,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(17884,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(17884,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(17884,36,'cp 120','c2c3 g7g6 f1d3 f8g7 e1g1 d7d6 f1e1 e6e5 h2h3 h7h6 c1e3 g8f6 b1d2 e8g8 d2f1 d6d5 e4d5 e5d4 d5c6 d4e3 f1e3 b7c6 d1c2 f6d5 a1d1 d8f6 d3e4 d5e7 a2a4 c8e6 e3g4 e6g4 h3g4 f8d8 g2g3 d8d1 e1d1');
INSERT INTO evaluations VALUES(18094,30,'cp 652',NULL);
INSERT INTO evaluations VALUES(18094,31,'cp 651',NULL);
INSERT INTO evaluations VALUES(18094,32,'cp 664',NULL);
INSERT INTO evaluations VALUES(18094,33,'cp 668',NULL);
INSERT INTO evaluations VALUES(18094,34,'cp 664',NULL);
INSERT INTO evaluations VALUES(18094,35,'cp 675',NULL);
INSERT INTO evaluations VALUES(18094,36,'cp 687','d1h5 g7g6 e5g6 h7g6 h5h8 e8f7 h2h4 f8g7 h8h7 g8e7 h4h5 d8g8 h5g6 e7g6 h7h5 d7d5 d2d3 g8e8 b1c3 e8e5 h5e5 f6e5 c1h6 g7f6 h6d2 f7g7 e1c1 c8e6 h1h6 a8h8 h6h8 g6h8 d1h1 f6e7 c3e2 h8g6 e2g3 d5e4 g3e4 g7g8 f2f3 e6f7 c1b1');
INSERT INTO evaluations VALUES(18774,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(18774,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18774,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(18774,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(18774,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(18774,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(18774,36,'cp 27','d2d4 d7d5 g1f3 g8f6 a2a3 d5c4 f1c4 a7a6 e1g1 b7b5 c4a2 f8e7 d1c2 c5d4 e3d4 c8b7 f1d1 c6a5 c2e2 e8g8 f3e5 a8c8 c1f4 b7d5 c3d5 f6d5 f4g3 a5c6 e5g4 d8d7 g4e3 e7f6 e3d5 e6d5 g3e5 f8e8 f2f4 c6e7 e2f3 c8c2 a1c1 c2c1 d1c1 e7f5 e5f6');
INSERT INTO evaluations VALUES(18784,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(18784,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(18784,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(18784,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(18784,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(18784,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(18784,36,'cp 20','d7d5 g1f3 g8f6 a2a3 d5c4 f1c4 a7a6 e1g1 b7b5 c4a2 f8e7 d1c2 c5d4 e3d4 c8b7 f1d1 c6a5 c2e2 e8g8 f3e5 a8c8 h2h3 b7d5 c3d5 f6d5 c1d2 a5c6 e2e4 d5f6 e5c6 f6e4 c6d8 e7d8 d2b4 f8e8 d1e1 e4f6 d4d5 e6d5 e1e8 f6e8 a2d5 a6a5 a1e1 e8f6');
INSERT INTO evaluations VALUES(18859,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(18859,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(18859,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(18859,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(18859,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(18859,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(18859,36,'cp 33','c2c4 d7d5 c4d5 e6d5 d2d4 g8f6 e1g1 c5c4 b1c3 f8e7 b2b3 c4b3 d1b3 e8g8 c1f4 c6a5 b3b1 a7a6 f3g5 g7g6 e2e4 b7b5 h2h4 h7h6 e4e5 h6g5 e5f6 e7f6 h4g5 f6g5 f4g5 d8g5 b1b4 a5c6 b4c5 c8e6 c3d5 a8c8 d5f4 g5c5 d4c5 c8c7 f4e6 f7e6 a1c1 f8d8');
INSERT INTO evaluations VALUES(18888,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(18888,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(18888,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(18888,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(18888,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(18888,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(18888,36,'cp -50','d7d5 e4d5 e6d5 c4b3 d8e7 d1e2 d4d3 e2e7 g8e7 e1g1 c8f5 c1e3 f5g6 f1d1 e8c8 b1a3 e7f5 d1d3 f5e3 d3e3 f8c5 e3e2 d5d4 c3d4 c5d4 f3d4 c6d4 e2e3 h8e8 b3a4 e8e3 f2e3 d4e6 a1c1 c8b8 a3c4 f7f6 c1d1 d8d1 a4d1 e6c5 b2b4 c5d3');
INSERT INTO evaluations VALUES(20388,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(20388,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(20388,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(20388,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(20388,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(20388,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(20388,36,'cp 37','f1d3 f8d6 f4d6 d8d6 a2a3 g8e7 e1g1 c8d7 f1e1 e8g8 b1d2 a8c8 c2c3 f7f6 c3c4 g8h8 a1c1 f8d8 c4d5 e6d5 c1c3 e7f5 d2b3 b7b6 h2h3 g7g6 d3f5 d7f5 c3e3 f5d7 b3d2 d8e8 d2f1 h8g7 d1e2 e8e3 f1e3');
INSERT INTO evaluations VALUES(21765,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21765,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(21765,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(21765,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(21765,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(21765,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(21765,36,'cp -3','b2b4 g8f6 g1f3 f8d6 c1b2 d8e7 b1d2 a7a5 b4b5 c6b8 c2c4 c7c6 f1e2 e8g8 e1g1 b8d7 d1c2 a5a4 f1c1 f8e8 c4c5 d6c7 a1b1 c6b5 b2a1 d7b8 b1b5 b8c6 e2f1 f6d7 g2g3 d7f8 a1c3 h7h6 f1g2 e8d8 e3e4 d5d4 c3b4 c6b4');
INSERT INTO evaluations VALUES(22655,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(22655,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(22655,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(22655,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(22655,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(22655,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(22655,36,'cp -1','g8f6 d2d3 f6d5 e4d5 c6e7 d3d4 d7d6 c2c4 e7f5 d4e5 d6e5 f1d3 c7c6 g1f3 f8b4 c1d2 b4d2 d1d2 c6d5 f3e5 e8g8 e1g1 d5c4 e5c4 c8e6 f1d1 d8f6 d2f4 a8c8 c4e3 f5e3 f4f6 g7f6 f2e3 c8c7 g1f2 g8g7 d1d2 f8d8 b2b3 c7d7 a1d1 f6f5 b3b4 b7b6 f2g3');
INSERT INTO evaluations VALUES(16332,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(16332,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16332,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(16332,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16332,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(16332,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(16332,36,'cp 0','d2d3 f6d5 e4d5 c6e7 d3d4 d7d6 c2c4 e7f5 d4e5 d6e5 f1d3 c7c6 g1f3 f8b4 c1d2 b4d2 d1d2 c6d5 f3e5 e8g8 e1g1 d5c4 e5c4 c8e6 d3f5 e6f5 d2f4 f5e6 f1d1 d8e7 a2a3 e6c4 f4c4 b7b6 d1d4 h7h6 a1d1 a8d8 d4d8 f8d8 d1d8 e7d8 h2h3 d8d2 c4c8 g8h7 c8f5 h7h8 f5c8 h8h7');
INSERT INTO evaluations VALUES(24677,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(24677,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(24677,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(24677,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(24677,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(24677,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(24677,36,'cp 238','d4d5 c6e5 f3e5 d6e5 f1b5 g4d7 b5d7 d8d7 f2f4 d7g4 d1a4 g4d7 a4b3 e5f4 c1f4 e7e5 f4e5 f8d6 e5f6 g7f6 e1c1 e8c8 c3b5 c8b8 c1b1 f6f5 b5d6 d7d6 e4f5 h8e8 g2g4 e8e2 b3g3 a7a5 g3d6 d8d6 h2h3 a5a4 h1e1 d6d5 e1e2 d5d1 b1c2 d1h1 e2e3 b8a7 a2a3 b7b5 e3e7 a7b6 e7f7 h1h3');
INSERT INTO evaluations VALUES(25679,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25679,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25679,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25679,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25679,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25679,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25679,36,'cp 0','f8d6 c1e3 c6c5 c2c3 f7f6 b1d2 g8e7 d1b3 h5f7 b3b7 a8b8 b7a6 b8a8 a6b7 a8b8');
INSERT INTO evaluations VALUES(23780,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23780,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23780,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23780,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(23780,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23780,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23780,36,'cp 0','c1e3 c6c5 b1d2 f7f6 c2c3 g8e7 d1b3 h5f7 b3b7 a8b8 b7a6 b8a8 a6b5 c7c6 b5b7 a8b8 b7a6 b8a8');
INSERT INTO evaluations VALUES(29324,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(29324,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(29324,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(29324,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(29324,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(29324,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(29324,36,'cp 16','h2h4 f6h5 f3e5 h5g3 f2g3 b8d7 e5d7 d8d7 h4h5 g6e4 d1e2 f7f5 c1f4 e8c8 e1c1 f8d6 h1e1 g7g6 h5g6 d6f4 g3f4 h7g6 c4d3 d7d6 d3e4 d6f4 e2e3 f4e3 e1e3 f5e4 e3e4 h8h2 e4g4 d8f8 g4g6 f8f2 d1g1 a7a5 g6e6 f2g2 g1g2 h2g2 e6e7 c8b8 c2c4 g2h2');
INSERT INTO evaluations VALUES(31300,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(31300,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31300,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(31300,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(31300,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(31300,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(31300,36,'cp 19','b8c6 e2e4 f5e4 c3e4 e7e5 d4e5 d5e4 f1c4 d8a5 e1f1 c6e5 d1d5 a5d5 c4d5 e5c6 d5e4 g8f6 e4c6 b7c6 a1e1 f8e7 f4g5 f6d5 g5e7 d5e7 g2g3 f7f6 f1g2 e8f7 g1f3 h8d8 e1e2 a8b8 h1e1 b8b7 b2b3 h7h5 e1c1 b7b4 f3d2 d8d5 d2e4 b4d4 e4c5 d4d2');
INSERT INTO evaluations VALUES(32244,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(32244,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(32244,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(32244,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(32244,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(32244,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(32244,36,'cp -21','b8c6 c2c3 d8d7 b1d2 h7h6 d2b3 e8c8 c1e3 f8d6 e2b5 g8e7 b3c5 d6c5 b5c5 e6g4 f3e5 c6e5 d4e5 b7b6 c5b5 c7c5 b5d7 g4d7 f2f4 c8b7 f1e2 d7e6 e2f3 h6h5 h2h3 e7f5 e3f2 a7a5 a2a4 h5h4 e1c1 h8e8');
INSERT INTO evaluations VALUES(37187,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37187,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37187,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(37187,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37187,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(37187,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37187,36,'cp 0','d7d5 e4e5 f6e4 f1d3 c7c5 g1e2 c5d4 e2d4 b8d7 d3e4 d5e4 c2e4 f7f5 e4f4 d8a5 e1g1 a5e5 d4f3 e5f4 c1f4 b4c3 b2c3 f8e8 f3d4 d7c5 d4b5 e8d8 b5c7 a8b8 c7b5');
INSERT INTO evaluations VALUES(37054,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(37054,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(37054,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37054,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37054,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37054,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37054,36,'cp 0','e4e5 f6e4 f1d3 c7c5 g1f3 c5d4 f3d4 b8d7 d3e4 d5e4 c2e4 d8c7 e1g1 c7e5 e4h4 d7f6 d4b5 e5f5 a2a3 b4e7 f1d1 a7a6 b5d6 f5g6 h4g5 f8d8 c4c5 e7d6 c5d6 c8d7 g5g6 h7g6 c1g5 d7c6 d1d2 d8d7 a1c1 f6e8 g5e7 f7f6 f2f3 g8f7 h2h4 e6e5');
INSERT INTO evaluations VALUES(37075,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37075,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37075,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37075,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37075,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37075,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37075,36,'cp 0','f6e4 f1d3 c7c5 g1f3 c5d4 f3d4 b8d7 d3e4 d5e4 c2e4 d8c7 e1g1 c7e5 e4h4 d7f6 d4b5 e5f5 f1d1 c8d7 a2a3 b4c5 b5d6 c5d6 d1d6 d7c6 h4f4 f5c2 f4d2 c2g6 d2g5 g6c2');
INSERT INTO evaluations VALUES(37011,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37011,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37011,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37011,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37011,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37011,36,'cp 0','g1f3 c7c5 f1d3 c5d4 f3d4 b8d7 d3e4 d5e4 c2e4 d8c7 e1g1 c7e5 e4h4 d7f6 d4b5 e5f5 f1d1 c8d7 a2a3 b4c5 b5d6 c5d6 d1d6 d7c6 h4f4 f5c2 f4d2 c2g6 d2g5 g6c2');
INSERT INTO evaluations VALUES(38093,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(38093,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(38093,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(38093,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(38093,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(38093,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(38093,36,'cp 66','f4d6 d8d6 e2e4 d5e4 c3e4 d6b4 d1d2 b4d2 e1d2 b8d7 f1d3 g8f6 e4f6 g7f6 g1e2 b7b6 b2b4 a7a5 a2a3 c8b7 h1b1 e8e7 g2g3 a5b4 a3b4 c6c5 a1a8 b7a8 d4c5 b6c5 b4b5 a8b7 b1a1 h8a8 a1a8 b7a8 e2c3 f6f5 d3e2 a8b7');
INSERT INTO evaluations VALUES(38073,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(38073,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38073,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(38073,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(38073,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(38073,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(38073,36,'cp 52','d8d6 e2e4 d5e4 c3e4 d6b4 d1d2 b4d2 e1d2 b8d7 f1d3 g8f6 f2f4 f6e4 d3e4 c6c5 b2b3 c5d4 g1e2 d4d3 e2c1 a8b8 c1d3 b7b6 h1e1 g7g6 g2g4 h7h5 g4g5 c8b7 a1d1 a7a5 e4b7 b8b7 d2c2 e8e7 d3e5 b7c7 d1d7 c7d7 e5d7');
INSERT INTO evaluations VALUES(33036,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(33036,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(33036,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(33036,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(33036,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(33036,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(33036,36,'cp 73','e2e4 d5e4 c3e4 d6b4 d1d2 b4d2 e1d2 b8d7 f1d3 g8f6 f2f4 f6e4 d3e4 c6c5 g1f3 d7f6 h1e1 c5d4 f3d4 f6e4 e1e4 b7b6 e4e5 c8b7 g2g3 e8c8 d2c3 h7h5 f4f5 e6f5 d4f5 h5h4 f5g7 h4g3 h2g3 h8h2 g7f5 d8d2 b2b3 d2c2 c3b4 c2a2 f5d6 c8c7');
INSERT INTO evaluations VALUES(39765,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(39765,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(39765,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(39765,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(39765,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(39765,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(39765,36,'cp 19','f2f3 e8g8 c1e3 b8c6 d4c6 b7c6 d1e1 f6d7 h2h3 d7c5 a1d1 c5d3 c2d3 c6c5 d3d4 c5d4 e3d4 g7d4 d1d4 d8b6 e1f2 c8e6 f3f4 f7f6 f4f5 e6f7 f1e1 a8b8 c3d5 f7d5 e4d5 b6b2 e1e7 b2b1 g1h2 b1f5 f2f5');
INSERT INTO evaluations VALUES(40926,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40926,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(40926,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40926,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40926,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40926,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(40926,36,'cp 0','c4d5 e6d5 d1c2 h7h6 g5h4 f6e4 h4e7 d8e7 b1c3 e4c3 c2c3 c8f5 e2e3 c7c6 g1e2 e8g8 e2f4 b8d7 c3b4 e7f6 f1e2 a7a5 b4b7 f8b8 b7c7 b8c8 c7b7');
INSERT INTO evaluations VALUES(41332,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(41332,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(41332,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(41332,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(41332,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(41332,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(41332,36,'cp -26','b1c3 d6d5 f1e2 c7c5 d4c5 d8a5 e1g1 a5c5 c3a4 c5c7 b2b3 c8d7 c1b2 e8g8 a1c1 d7a4 b3a4 f8d8 c2c4 b8c6 d1b3 c6a5 b3b5 a7a6 b5b4 a5c4 e2c4 d5c4 c1c4 c7d7 b4b3 b7b5 a4b5 a6b5 c4b4 d7a7 b2f6 g7f6 b4b5 d8c8 f3g5 f6g5 f4g5 a7a2');
INSERT INTO evaluations VALUES(44743,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44743,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44743,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44743,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44743,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44743,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44743,36,'cp 0','b8c6 f1e1 f8g7 c2c3 e8g8 a2a4 h7h6 c1e3 c8g4 b1d2 c6e5 h2h3 g4f3 d2f3 e5c4 d3c4 f6e4 e3h6 g7h6 e1e4 g8g7 d1d5 e7e6 d5b7 a8b8 b7a7 b8a8 a7b7 a8b8');
INSERT INTO evaluations VALUES(45417,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(45417,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(45417,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(45417,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(45417,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45417,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(45417,36,'cp 10','c4d5 f6d5 b1c3 b8c6 f1c4 d5b6 c4b5 g7g6 e1g1 f8g7 d2d4 e8g8 b5c6 b7c6 h2h3 c8e6 f1e1 a8c8 c1f4 b6d5 f4e5 d5c3 b2c3 c6c5 e5g7 g8g7 d4c5 d8d1 a1d1 c8c5 f3d4 c5c3 d4e6 f7e6 d1d7 g7f7 d7a7 f8c8 a2a4 c3c2 a7b7 f7f6');
INSERT INTO evaluations VALUES(45409,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(45409,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45409,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(45409,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(45409,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(45409,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(45409,36,'cp 9','f6d5 b1c3 b8c6 f1c4 e7e6 e1g1 f8d6 d2d4 c6e7 c3d5 e6d5 c4b5 c8d7 b5d3 e8g8 f1e1 f8e8 c1d2 d8b6 d2c3 e7g6 g2g3 d7g4 d1a4 e8e1 f3e1 g6e7 e1g2 e7f5 a1e1 g7g6 g2e3 f5e3 e1e3 g4e6 a2a3 g8g7');
INSERT INTO evaluations VALUES(45515,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(45515,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(45515,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(45515,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45515,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45515,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45515,36,'cp 2','b1c3 b8c6 f1b5 e7e6 e1g1 f8d6 d2d4 e8g8 b5d3 h7h6 f1e1 d5f6 c3e4 d6c7 a2a3 e6e5 d4e5 f6g4 c1f4 g4e5 f3e5 c6e5 d3e2 c8f5 e4c3 f5e6 f4g3 e5c6 e2f3 c7g3 h2g3 d8d1 a1d1 a8d8 f3e4 e6b3 d1d8 f8d8');
INSERT INTO evaluations VALUES(46058,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(46058,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(46058,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(46058,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(46058,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(46058,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(46058,36,'cp 120','f1d3 f8e7 b1c3 b8c6 e1g1 e8g8 h2h3 d8b6 c3a4 b6d8 c1e3 d6d5 e4e5 f6e4 a4c3 f7f5 e5f6 e4f6 d1e2 e7d6 a1e1 c8d7 d3b1 d8e8 e2d2 a8c8 e3g5 c6a5 b2b3 d6b4 e1c1 a5c6 f3e5 c6e5 d4e5 b4c3 c1c3 c8c3 d2c3 f6e4');
INSERT INTO evaluations VALUES(48081,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48081,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48081,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48081,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48081,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48081,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48081,36,'cp 0','b1c3 d5b6 c4b3 a7a5 a2a3 f8e7 g1e2 c7c5 d2d3 e8g8 f3g3 e7h4 g3e3 h4e7');
INSERT INTO evaluations VALUES(48077,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48077,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48077,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48077,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48077,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48077,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48077,36,'cp 0','d5b6 c4b3 a7a5 a2a3 f8e7 g1e2 c7c5 d2d3 e8g8 f3g3 e7h4 g3e3 h4e7');
INSERT INTO evaluations VALUES(49601,30,'cp -101',NULL);
INSERT INTO evaluations VALUES(49601,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(49601,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(49601,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(49601,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(49601,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(49601,36,'cp -80','g4g3 f8e7 b1c3 e8g8 d2d3 d7d5 e4d5 f6d5 g1f3 d5c3 b2c3 b8d7 f1e2 e7d6 g3g5 d8e8 f3d2 f7f5 e1g1 d7b6 c3c4 b6a4 g5h4 d6e7 h4h3 e7b4 d2b3 e5e4 d3e4 f5e4 h3e3 a4c3 c1b2');
INSERT INTO evaluations VALUES(51113,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(51113,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(51113,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(51113,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(51113,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(51113,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(51113,36,'cp 24','c2c4 e7e6 g1f3 g8f6 b2b3 d7d5 b1c3 c7c5 c1b2 b8c6 c4d5 e6d5 d2d4 c8g4 f1e2 g4f3 e2f3 c5d4 e3d4 f8b4 e1g1 b4c3 b2c3 e8g8 f1e1 d8d7 d1d3 f8e8 a1c1 e8e1 c1e1 a8e8 e1e8 d7e8 d3e3 e8e6 g1f1 f6e4 c3e1');
INSERT INTO evaluations VALUES(51661,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(51661,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(51661,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(51661,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(51661,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(51661,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(51661,36,'cp -106','g8f6 a4b5 c8g4 f3g3 c6b5 f1e2 g4e2 g1e2 d8d7 b1d2 e7e6 e1g1 b8c6 b2b3 c4b3 c1b2 a7a5 e2f4 f6d5 f4d5 e6d5 d2b3 f7f6 e3e4 f8b4 f1c1 e8g8 b3c5 d7f7 g3d6 c6e7 d6d7 f8d8 d7b5 d5e4 b5e2 a8b8 c5a6 b8b6 a6b4 a5b4 e2e4 e7d5 c1c5');
INSERT INTO evaluations VALUES(52724,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52724,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(52724,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52724,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(52724,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52724,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(52724,36,'cp 0','d8b6 e2e3 b6b2 b1d2 e7e6 f1d3 f8e7 h2h4 h7h6 g5f4 c6c5 d4c5 g8f6 e1g1 e8g8 f3d4 e7c5 d4b5 b8c6 a1b1 b2a2 d2b3 c5e7 b5c3 a2a3 c3b5');
INSERT INTO evaluations VALUES(53691,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(53691,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(53691,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(53691,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(53691,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(53691,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(53691,36,'cp -31','c6d5 c4b3 b8c6 g1e2 f8c5 e1g1 g8e7 b1c3 e8g8 c3a4 c5d6 a4c3 d6b8 d3d4 c8e6 d4e5 d5d4 c3e4 e6b3 a2b3 b8e5 c1f4 d8d5 f4e5 d5e5 e2g3 a8d8 f1e1 d4d3 e4c3 d3c2 d1c2 e5c5 g1h1 f7f5 a1d1 e7d5 c2d2 d5c3');
INSERT INTO evaluations VALUES(53828,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(53828,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(53828,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(53828,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(53828,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(53828,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(53828,36,'cp 48','f8e7 d2d4 b8d7 a2a4 g8f6 f1c4 e5d4 d1d4 e8g8 e1g1 d7c5 f1d1 c8e6 c1f4 a7a5 f4h2 f8e8 f3g5 d6d5 e4d5 e6d5 d4f4 c5e6 g5e6 f7e6 f4e3 e7d6 h2d6 d8d6 b2b3 d6f8 d1d4 a8d8 d4d3 e6e5 a1d1 e5e4 d3d4 d5c4 d4c4 d8d1 c3d1');
INSERT INTO evaluations VALUES(54511,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(54511,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(54511,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(54511,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(54511,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(54511,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(54511,36,'cp 0','c5d4 e3d4 b8c6 f1b5 g8f6 g1f3 f8d6 c1g5 e8g8 e1g1 c8g4 b5e2 d6e7 f1e1 h7h6 g5h4 f8e8 h2h3 g4f3 e2f3 d8b6 h4f6 e7f6 e1e8 a8e8 c3d5 b6d4 d1d4 f6d4 a1c1 e8d8 d5c3 d4f6 c1d1 d8d1 c3d1 g8f8 g1f1 f8e8 f3c6 b7c6');
INSERT INTO evaluations VALUES(56776,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(56776,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(56776,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(56776,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(56776,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(56776,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(56776,36,'cp 47','b8c6 f1d3 g8h6 c1h6 g7h6 e1g1 c8d7 c2c3 a7a5 b1d2 d8e7 a2a3 e7g7 d1e2 f8c5 g1h1 e8g8 a1e1 a5a4 f3g1 c6e7 g1h3 g8h8 f2f4 e7f5 d3f5 f8f5 f1f3 d7e8 f3g3 e8g6 d2f3 g7e7 f3d4 c5d4 c3d4 c7c5 e2e3 c5c4');
INSERT INTO evaluations VALUES(22216,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(22216,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(22216,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(22216,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(22216,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(22216,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(22216,36,'cp 81','f1d3 g8h6 c1h6 g7h6 e1g1 c8d7 c2c3 a7a5 b1d2 f8c5 a2a3 e8g8 b2b4 a5b4 a3b4 a8a1 d1a1 c5b6 a1b1 d8e7 b4b5 c6d8 c3c4 d7e8 b1b3 e8f7 g2g3 f7h5 c4d5 h5f3 d2f3 f8f3 d3h7 e7h7 b3f3 h7f7 f3f7 g8f7');
INSERT INTO evaluations VALUES(5126,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5126,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5126,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(5126,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5126,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5126,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5126,36,'cp 0','c1e3 c6d4 h2h4 h7h5 f1e2 d7d6 e3g5 d8b6 e1g1 g8e7 a1b1 e7c6 c3b5 d4f3 e2f3 c6e5 b2b4 c8d7 b4c5 d6c5 f3e2 a7a6 b5c3 b6a5 g5d2 a5d8 d2g5');
INSERT INTO evaluations VALUES(5916,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(5916,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(5916,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(5916,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(5916,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(5916,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(5916,36,'cp 21','d1d2 g8f6 f3e5 c8d7 e1c1 f6h5 g2g3 h5f4 g3f4 d8h4 e2e3 d6b4 e5f3 b4c3 f3h4 c3d2 c1d2 f7f6 h4f3 c6e7 d1c1 e8f7 h2h4 h7h5 f1e2 a7a5 c2c4 h8c8 c4c5 b7b5 f3e1 g7g6 e1d3 c8b8 b2b3 c7c6 a2a3 a5a4');
INSERT INTO evaluations VALUES(7689,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(7689,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7689,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7689,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7689,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7689,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7689,36,'cp 0','e4e5 f6d5 c3d5 c6d5 d2c3 e7e6 d1d2 d7d6 e5d6 g7c3 d2c3 e8g8 e1c1 d8d6 c3f6 c8a6 f1a6 d6a6 h2h4 a6a2 h4h5 a8b8 h5g6 f7g6 f6e6 f8f7 b2b3 a2a1 c1d2 a1a5 d2c1 a5a1');
INSERT INTO evaluations VALUES(8615,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(8615,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(8615,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(8615,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(8615,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(8615,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(8615,36,'cp 25','f4e5 d6e5 d2d3 c6a5 c4b3 a5b3 a2b3 c8e6 f3g5 d8d6 b1c3 a7a5 d1e1 e8g8 e1g3 g8h8 g5e6 f7e6 c1e3 b7b6 h2h3 f6d7 g3g4 d6c6 g1h2 f8f1 a1f1 a8f8 f1f3 f8f3 g4f3 e7c5 e3c5 c6c5 f3f7');
INSERT INTO evaluations VALUES(9082,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(9082,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(9082,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(9082,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(9082,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(9082,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(9082,36,'cp -39','e4d5 c8g4 d1e1 f6d5 d2d4 c5d4 b1c3 d4c3 b2c3 d8d7 f2f3 g4e6 c1a3 e8c8 a1d1 c6a5 c4b3 d7c6 e1e5 a5b3 c2b3 c6b6 g1h1 d8e8 e5d4 b6d4 d1d4 d5c3 a3b2 c3a2 g5e6 e8e6 d4a4 e6a6 a4a6 b7a6 f1a1 a2b4 b2g7 h8g8 a1a4 b4d5 g7d4 c8b7 h1g1 g8g6 g1f2 g6c6 g2g4 d5f4 d4e3');
INSERT INTO evaluations VALUES(9069,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(9069,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(9069,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(9069,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(9069,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(9069,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(9069,36,'cp -35','c8g4 d1e1 f6d5 d2d4 c5d4 b1c3 d4c3 b2c3 d8d7 f2f3 g4e6 c1a3 e8c8 a1d1 d7e8 g5e6 f7e6 d1b1 h7h5 c4d3 e8d7 f1f2 g7g5 c3c4 d5f4 d3e4 h8g8 a3c1 d7d1 c1f4 d1e1 b1e1 e5f4 e4c6 b7c6 e1e6 c8b7 f2e2 d8d1 e2e1 g8d8 g1f2 d1d4 e1e4 d4e4 e6e4 d8d2 e4e2 d2d4 e2e4 c6c5 e4d4 c5d4 f2e2 b7c6 e2d3 c6c5 h2h3');
INSERT INTO evaluations VALUES(9424,30,'cp 249',NULL);
INSERT INTO evaluations VALUES(9424,31,'cp 271',NULL);
INSERT INTO evaluations VALUES(9424,32,'cp 284',NULL);
INSERT INTO evaluations VALUES(9424,33,'cp 294',NULL);
INSERT INTO evaluations VALUES(9424,34,'cp 300',NULL);
INSERT INTO evaluations VALUES(9424,35,'cp 288',NULL);
INSERT INTO evaluations VALUES(9424,36,'cp 301','d4c6 b7c6 e3c5 d6c5 d1d8 e8d8 e1c1 d8e7 c3a4 f6d7 h2h4 a8b8 a2a3 h7h5 d1e1 c5c4 f1c4 d7b6 c4b3 b6a4 b3a4 c6c5 a4b3 c8a6 h1h3 c5c4 b3a4 h8h6 h3g3 a6c8 g3g5 h6b6 g5c5 c8e6 c5c7 e7d6 c7a7 b6b2 a4b3 b8b3 c2b3 b2f2 b3b4 f2a2 a7a5 a2a1 c1d2 c4c3 d2e2 e6g4 e2f2 a1a2 f2e3 a2g2 a5d5 d6e7 d5c5');
INSERT INTO evaluations VALUES(11271,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(11271,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(11271,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(11271,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(11271,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(11271,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(11271,36,'cp 58','d1d2 f8e7 e1c1 c6d4 d2d4 e6e5 d4d2 c8e6 f2f3 d8c7 c1b1 a8c8 g2g4 e8g8 g4g5 f6h5 a2a3 b7b5 d1c1 c7b7 h1d1 h5f4 e3f4 e5f4 h2h4 c8c5 c3d5 e6d5 e4d5 f7f6 g5f6 e7f6 e2d3 g8h8 d2f4 b7d5 d1e1 a6a5 f4d2');
INSERT INTO evaluations VALUES(15482,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15482,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15482,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15482,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15482,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15482,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15482,36,'cp 0','g1e2 c8g4 f3e3 c6b4 c4b3 g4e6 b3a4 e6d7 a4b3');
INSERT INTO evaluations VALUES(15784,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(15784,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(15784,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(15784,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(15784,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(15784,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(15784,36,'cp 35','e4d3 c7c5 c1e3 d8c7 d1e2 a7a6 e1g1 b7b6 c2c4 c5d4 e3d4 c8b7 f3e5 a8d8 d4c3 f8c5 a1d1 a6a5 a2a3 c5d4 c3d4 d8d4 e2e3 d4d8 b2b4 e8g8 f1e1 a5b4 a3b4 f6d7 e5d7 d8d7 d3f1 f8d8 d1d7 d8d7 e1c1 g7g6 h2h3 h7h5 b4b5 c7d6 c4c5 b6c5 c1c5');
INSERT INTO evaluations VALUES(16069,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16069,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16069,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16069,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(16069,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(16069,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16069,36,'cp 17','g2g3 f8g7 f1g2 d7d6 c1e3 a7a5 e1g1 e8g8 h2h3 c8e6 d3d4 e5d4 f3d4 c6d4 e3d4 c7c6 a2a4 d6d5 e4d5 f6d5 c3d5 e6d5 d4g7 d5g2 g7f8 g2f1 d1d8 a8d8 f8e7 d8d7 g1f1 d7e7 a1d1 g8g7 d1d4 h7h5 d4d6 b7b5 d6c6 b5a4');
INSERT INTO evaluations VALUES(16405,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(16405,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(16405,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(16405,33,'cp -84',NULL);
INSERT INTO evaluations VALUES(16405,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(16405,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(16405,36,'cp -73','b1c3 c6d4 a2a3 c7c6 d2d3 d7d5 c4a2 f8c5 g1e2 e8g8 e2d4 c5d4 c3e2 d4c5 c2c3 f8e8 e2g3 c8e6 d1e2 f6d7 c1e3 d8b6 e3f2 d5e4 g3e4 e6a2 a1a2 c5f2 e2f2 e8d8 e1g1 f7f5 e4d6 b6f2 g1f2 d7c5 d6f5 d8d3');
INSERT INTO evaluations VALUES(17870,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(17870,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(17870,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(17870,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(17870,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(17870,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(17870,36,'cp 168','c3d4 f8b4 b1c3 g8f6 e4e5 f6e4 e1g1 b4c3 b2c3 d7d5 e5d6 e8g8 d6c7 d8c7 f1e1 e4c3 d1c2 c6a5 c1a3 c7c4 a3f8 c3d5 c2e4 c8e6 f8d6 a8d8 d6e5 d8c8 h2h3 c4c6 e4h4 a5c4 h4g3 f7f6 e5f4 g7g5 f4d2 c6d7 a2a4 g8g7 g1h2 e6f7');
INSERT INTO evaluations VALUES(17868,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(17868,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(17868,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(17868,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(17868,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(17868,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(17868,36,'cp 126','b7b5 c4b3 f8b4 b1c3 g8e7 e1g1 e8g8 d1d3 b4c3 b2c3 d7d5 f3h4 c6a5 b3c2 e7g6 h4g6 f7g6 d3g3 d5e4 c2e4 c8b7 e4g6 d8d6 f2f3 b7d5 f1e1 f8f6 g6c2 d6g3 h2g3 f6c6 c2d1 a8d8 e1e2 a5c4 a2a4 d5f7 a4a5 c4d6 c1d2 d6c4 d2e1 c6e6 g1f2 d8c8 e2e6 f7e6 g3g4 c7c5 d1c2 c5d4 c3d4');
INSERT INTO evaluations VALUES(17925,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(17925,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(17925,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(17925,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(17925,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(17925,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(17925,36,'cp 39','b5a4 f8c5 e1g1 g8e7 d2d3 e8g8 b1d2 c5b6 g1h1 d7d5 f2f4 d4c3 b2c3 b6a5 c1b2 b7b5 a4c2 d5e4 d3e4 f8e8 f4f5 e7c6 a2a4 d8f6 c2b3 c6e5 a4b5 a6b5 d1e2 c8b7 e2b5 b7c6 b5e2 a5b6 a1e1 f6d6 d2f3 e5f3 e2f3');
INSERT INTO evaluations VALUES(17939,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(17939,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(17939,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(17939,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(17939,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(17939,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(17939,36,'cp 51','f8c5 e1g1 g8e7 d2d3 c5a7 b1d2 e8g8 a4b3 d7d5 f1e1 c7c6 d2f3 d4c3 b2c3 h7h6 f3e5 a7b8 e4d5 e7d5 c3c4 b8e5 e1e5 d5f6 c1a3 f8e8 e5e8 f6e8 a3b2 d8g5 d1f3 c8g4 f3e3 e8f6 h2h3 g4f5 a1e1 g5e3');
INSERT INTO evaluations VALUES(18100,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(18100,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(18100,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(18100,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(18100,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(18100,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(18100,36,'cp -8','c4d3 g8e7 c1e3 e7g6 a2a4 b5a4 c3a4 a8b8 d1e2 f8e7 d3a6 c8a6 e2a6 e8g8 f1d1 c6b4 a6e2 d7d6 g2g3 h7h6 h2h4 b4c6 a4c5 d8c8 c5d3 e7f6 a1a4 c8b7 d1c1 f8c8 a4c4 g6e5 d3e5 f6e5 b2b4 c6b4 f3e5 d6e5 c4c8 b8c8 c1c8 b7c8');
INSERT INTO evaluations VALUES(20600,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(20600,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(20600,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(20600,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(20600,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(20600,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(20600,36,'cp 90','c2c3 c8g4 h2h3 g4h5 d2d3 g8f6 b1d2 f8e7 f1e1 e8g8 a2a4 a7a6 d2f1 g8h8 f1g3 h5g6 f3h4 f6e8 h4f5 e7g5 c1g5 d8g5 f5e3 a8b8 e1f1 g5d8 c4b3 c6a5 b3a2 a5c6 d1g4 c6e7 g4e2 b7b5 h3h4 h7h6 f2f4 e5f4 f1f4 b5a4');
INSERT INTO evaluations VALUES(27862,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(27862,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27862,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(27862,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27862,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(27862,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(27862,36,'cp 75','f1e2 c7c5 d4d5 e7e6 e4e5 d6e5 f4e5 f6d7 c1g5 d8c8 e1g1 g4f3 g2f3 g7e5 f3f4 e5d4 g1h1 d7b6 a2a4 e6d5 c4d5 c8d7 e2b5 d7d6 d1f3 d4c3 b2c3 d6d5 a4a5 c5c4 a5b6 d5b5 b6a7 b5c6 f4f5 c6f3 f1f3 b8c6 g5e3 f8e8 f5f6 e8e5 f3g3 e5h5 g3g4');
INSERT INTO evaluations VALUES(27861,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(27861,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27861,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(27861,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(27861,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(27861,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(27861,36,'cp 28','c7c5 d4d5 e7e6 e1g1 e6d5 e4d5 f6h5 h2h3 g4f3 f1f3 g7d4 g1h1 f7f5 f3d3 h5g7 e2f3 d4c3 d3c3 b8d7 c1d2 f8e8 c3b3 b7b6 a2a4 a7a5 h1h2 d7f6 d1f1 d8c7 a1e1 e8e1 f1e1 a8e8 b3e3 g8f7 e3e8 f6e8 e1h4 c7d7 d2c3 d7a4');
INSERT INTO evaluations VALUES(28009,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(28009,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(28009,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(28009,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(28009,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(28009,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(28009,36,'cp 91','c1e3 d7d6 d1d2 g8f6 g1f3 b8d7 f3d4 a7a6 f2f3 d7e5 e1c1 h7h5 a2a4 f6d7 f3f4 c7c5 d5c6 e5c6 f1c4 b6b5 c4d5 d8c7 d4c6 b7c6 e3d4 g7d4 d2d4 e8g8 f4f5 a8c8 f5g6 c6d5 d4d5 d7f6 g6f7 f8f7 d5b3 c7c4 b3c4 c8c4 a4b5 a6b5 h1e1 b5b4');
INSERT INTO evaluations VALUES(28424,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(28424,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(28424,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(28424,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(28424,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28424,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(28424,36,'cp 45','g4f3 g2f3 c7c5 d4c5 d6c5 d1d8 e8d8 c1e3 f6d7 e1c1 g7c3 b2c3 d8c7 e3f4 e7e5 f4e3 b8c6 f3f4 b7b6 f4e5 h8e8 f2f4 f7f6 f1g2 f6e5 f4f5 d7f6 e3g5 f6h5 d1d2 e8f8 h1f1 f8f7 h3h4 a8f8 g2h3 g6f5 e4f5');
INSERT INTO evaluations VALUES(28429,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(28429,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(28429,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(28429,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(28429,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(28429,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(28429,36,'cp 43','g2f3 b8d7 c1e3 c7c5 d4d5 f6h5 d1d2 e8g8 f3f4 g7c3 d2c3 e7e5 d5e6 f7e6 c3d2 d8f6 d2d6 f6b2 a1d1 b2c3 d6d2 c3d2 d1d2 d7b6 h1g1 b6a4 d2d3 a4b6 e3c5 f8f4 d3e3 a8d8 f1e2 h5f6 f2f3 d8c8 c5d4 f6d7 e3c3 f4h4 g1g3 g8f7 c3a3 e6e5 d4c3 a7a6 a3b3 c8g8 b3b1 b6a4 b1b7 a4c3 b7d7 f7e6');
INSERT INTO evaluations VALUES(29238,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(29238,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(29238,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29238,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(29238,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(29238,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29238,36,'cp 0','b8d7 f1d3 f5g6 e1g1 f8e7 b2b3 g6h5 c1b2 e8g8 d3e2 h5g6 e2d3');
INSERT INTO evaluations VALUES(29691,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(29691,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(29691,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(29691,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(29691,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(29691,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(29691,36,'cp 20','f2f3 g4c8 d3d4 c7c5 c3d5 f6d5 e4d5 d8d5 c1e3 c5d4 d1d4 d5d4 e3d4 b8c6 d4c3 e7e5 e1c1 f7f6 f1b5 f8c5 g1e2 c8e6 f3f4 e8g8 f4e5 f6e5 e2g3 c6b4 c3b4 c5b4 a2a3 b4c5 g3e4 c5e7 b5d3 a8d8 h1e1 h7h5');
INSERT INTO evaluations VALUES(29690,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(29690,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(29690,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(29690,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(29690,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(29690,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(29690,36,'cp 14','g4c8 d3d4 c7c5 e4d5 f6d5 c3d5 d8d5 c1e3 c5d4 d1d4 d5d4 e3d4 b8c6 d4e3 c8f5 e1c1 a8c8 g1e2 a7a6 g2g4 f5e6 a2a3 g7g6 h2h3 h7h5 f1g2 f8g7 c1b1 g7e5 d1d2 h5g4 f3g4 f7f6 e2f4 e5f4 e3f4 e8f7 h1d1 h8d8 f4e3 d8d2');
INSERT INTO evaluations VALUES(31107,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(31107,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(31107,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(31107,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(31107,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(31107,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(31107,36,'cp 121','d1b3 e7e5 d4e5 g8e7 g1f3 d5c4 b3b7 b8d7 b7a6 g4e6 f3g5 e6d5 a6a4 h7h6 e5e6 f7e6 g5f3 d5f3 e2f3 d7b6 a4c2 e7d5 c2g6 e8d7 c1d2 d8f6 g6c2 d5b4 c2e4 f6f5 e1c1 d7c7 d2f4 c7b7 g2g4 f5e4 f3e4 f8c5 a2a3 b4d3 f1d3 c4d3 d1d3 c5f2 c1c2 f2c5 h1d1');
INSERT INTO evaluations VALUES(37793,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37793,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37793,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37793,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37793,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37793,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37793,36,'cp 0','c5d4 e3d4 d7d5 c4d5 g8f6 f1c4 f6d5 d1b3 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5 a7a6 b5c7');
INSERT INTO evaluations VALUES(37858,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37858,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37858,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37858,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37858,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37858,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37858,36,'cp 0','e3d4 d7d5 c4d5 g8f6 f1c4 f6d5 d1b3 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(37855,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37855,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37855,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37855,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37855,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37855,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37855,36,'cp 0','d7d5 c4d5 g8f6 f1c4 f6d5 d1b3 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(37827,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(37827,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(37827,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(37827,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(37827,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37827,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37827,36,'cp -10','b8c6 c2c3 e7e5 d2d3 g8f6 c1g5 h7h6 g5f6 g7f6 h2h4 d7d5 e4d5 d8d5 e1g1 c8g4 f3e5 d5g2 g1g2 g4d1 e5c6 d1g4 f1e1 g4e6 c6e5 e8c8 b1d2 g6g5 h4g5 h6g5 g3g4 c8c7 a2a4 b7b6 d2f3 e6d5 g2g3 d8e8 a1d1');
INSERT INTO evaluations VALUES(38992,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(38992,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(38992,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(38992,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(38992,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(38992,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(38992,36,'cp 82','g8e7 f1d3 h7h6 g5e3 d7d5 e4e5 c7c5 c2c3 b8c6 e1g1 c8d7 b1d2 g6g5 h2h3 e7f5 d1e2 c5d4 c3d4 c6b4 d3b1 d8b6 f1c1 d7b5 e2d1 b4d3 b1d3 b5d3 c1c3 b6a6 a2a4 e8g8 b2b4 f5e3 f2e3 d3e2 b4b5 e2d1 b5a6 d1f3 a6b7');
INSERT INTO evaluations VALUES(43062,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(43062,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(43062,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(43062,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(43062,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(43062,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(43062,36,'cp 79','f1d3 h7h6 g5e3 d7d5 e4e5 c7c5 c2c3 c8d7 b1d2 b8c6 e1g1 e8g8 f1e1 c5d4 c3d4 c6b4 d3b1 e7f5 d2b3 f5e3 e1e3 d8e7 h2h4 h6h5 b3c5 d7e8 e3c3 b4c6 d1e1 f7f6 c5d3 a7a5 e1e3 f6e5 d3e5 c6e5 f3e5 g7e5 e3e5 e7h4 e5e6 e8f7');
INSERT INTO evaluations VALUES(39057,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(39057,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(39057,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(39057,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(39057,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(39057,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(39057,36,'cp 107','d4e5 b8c6 b1c3 c6e5 f2f4 e5c6 g1f3 d7d6 f1d3 g8f6 e1g1 e8g8 c1e3 c6b4 d3b1 f8e8 a2a3 b4c6 h2h3 b7b6 d1e2 f6d7 b1c2 g7c3 b2c3 c8a6 a1d1 c6a5 c2d3 d8e7 f4f5 a6b7 f5g6 f7g6 e3g5 e7g7 c4c5');
INSERT INTO evaluations VALUES(40265,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40265,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40265,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40265,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(40265,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(40265,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(40265,36,'cp 13','c6d5 h2h3 a7a6 f1d3 e8g8 e1g1 b8c6 c1d2 b7b5 d1b3 c8b7 f1c1 c6a5 b3d1 a5c4 c1c2 c4d2 d1d2 f6e4 d2d1 d8d7 a1c1 a8c8 d3f1 d7d6 a2a3 e7e6 b2b4 d6e7 c3e4 d5e4 f3d2 e6e5 c2c8 f8c8 c1c8 b7c8');
INSERT INTO evaluations VALUES(41565,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41565,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(41565,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(41565,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(41565,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(41565,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41565,36,'cp 33','c7c5 g1f3 d8a5 d1d2 b8c6 a1b1 a7a6 b1c1 c5d4 c3d4 a5d2 e1d2 e7e6 f1d3 e8g8 c1c5 c8d7 h1b1 b7b5 b1c1 a8c8 f3e5 c6e5 d4e5 c8c5 c1c5 f8c8 f2f4 g6g5 g2g3 g5f4 g3f4 f7f6 c5c8 d7c8 e3d4 f6e5 d4e5 g7f8 d2c3 g8f7 c3d4 c8b7 e5b8 f8a3 b8c7 b7c6 c7a5 c6b7 a5b6');
INSERT INTO evaluations VALUES(40431,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40431,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(40431,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(40431,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(40431,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(40431,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(40431,36,'cp 30','g1f3 d8a5 d1d2 e8g8 a1c1 c5d4 c3d4 a5d2 f3d2 e7e6 d2b3 c8d7 f1c4 b8c6 e1g1 f8c8 f1d1 d7e8 d4d5 c6e5 c4e2 e6d5 e4d5 g7f8 f2f4 e5d7 e2f3 c8c1 d1c1 b7b6 g1f2 a8d8 c1c7 d7f6 c7a7 f6d5 f3d5 d8d5 e3b6 d5d7 a7d7');
INSERT INTO evaluations VALUES(42361,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(42361,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(42361,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(42361,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(42361,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(42361,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42361,36,'cp 3','g1f3 d7d6 c2c3 b8c6 b2b4 c5b6 a2a4 a7a5 b4b5 c6e7 b1d2 e8g8 e1g1 e7g6 f1e1 f8e8 d1b3 c8e6 c4e6 e8e6 b3c2 f6h5 d2c4 g6h4 f3h4 d8h4 c4b6 c7b6 c1e3 e6g6 g1h1 a8f8 e1f1 f8c8 c3c4 h5f4 e3f4');
INSERT INTO evaluations VALUES(42358,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42358,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42358,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42358,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42358,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42358,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42358,36,'cp 0','d7d6 c2c3 a7a5 e1g1 b8c6 b1d2 e8g8 a2a4 c8e6 f1e1 d8b8 f3h4 b8d8');
INSERT INTO evaluations VALUES(44638,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(44638,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(44638,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(44638,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(44638,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(44638,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(44638,36,'cp 52','e4e5 f6g8 c1g5 b8c6 g1f3 f7f6 e5f6 e7f6 g5f4 f8b4 f1d3 d7d5 e1g1 b4c3 b2c3 g8e7 d1b3 e8g8 h2h3 c8f5 d3f5 e7f5 f1d1 f5e7 b3b7 c6a5 b7b5 a7a6 b5e2 a5c4 h3h4 f8e8 e2e6 g8h8 h4h5 e7g6 e6g4 g6e5 g4f5 d8c8 f5c8 a8c8 d1d5 e5f3 g2f3 c4e5 f4e5 e8e5 d5e5 f6e5');
INSERT INTO evaluations VALUES(45875,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(45875,31,'cp 258',NULL);
INSERT INTO evaluations VALUES(45875,32,'cp 270',NULL);
INSERT INTO evaluations VALUES(45875,33,'cp 279',NULL);
INSERT INTO evaluations VALUES(45875,34,'cp 264',NULL);
INSERT INTO evaluations VALUES(45875,35,'cp 265',NULL);
INSERT INTO evaluations VALUES(45875,36,'cp 248','d4e5 f6g4 f1c4 f8e7 e5d6 d8d6 d1d6 e7d6 h2h3 g4e5 f3e5 d6e5 c1d2 b8d7 e1c1 e8g8 f2f4 e5d4 c4e2 a7a5 e2f3 d7b6 c3e2 d4c5 d2c3 c8e6 c1b1 g7g6 g2g4 e6c4 b2b3 c4e2 f3e2 f8e8 e2f3 a5a4 e4e5 a8d8 d1d8 e8d8 f4f5 a4b3');
INSERT INTO evaluations VALUES(45853,30,'cp 264',NULL);
INSERT INTO evaluations VALUES(45853,31,'cp 263',NULL);
INSERT INTO evaluations VALUES(45853,32,'cp 251',NULL);
INSERT INTO evaluations VALUES(45853,33,'cp 265',NULL);
INSERT INTO evaluations VALUES(45853,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(45853,35,'cp 270',NULL);
INSERT INTO evaluations VALUES(45853,36,'cp 262','f6g4 f1c4 f8e7 e5d6 d8d6 d1d6 e7d6 h2h3 g4e5 f3e5 d6e5 c1d2 b8d7 f2f4 b7b5 c4e2 e5d4 e4e5 d7b6 e1c1 c8e6 c3e4 e8g8 c2c3 e6d5 e2d3 f7f5 e4g3 d4c5 b2b4 c5e7 g3f5 f8e8 d2e3 e7f8 e3d4 d5a2 d4b6 a7b6 d3e4 a8a3 c1b2');
INSERT INTO evaluations VALUES(46120,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(46120,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(46120,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(46120,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(46120,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(46120,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46120,36,'cp 18','f1c4 a7a6 a2a4 f8e7 e1g1 b8c6 d4c5 d8d1 f1d1 e7c5 c1d2 b7b6 c4d3 c5e7 c3e4 c8b7 e4f6 e7f6 a1c1 e8g8 d3e4 a8c8 b2b4 f8d8 d2e1 d8d1 c1d1 c8c7 d1c1 c7c8 c1b1 b6b5 a4b5 a6b5 h2h3 b7a6 b1d1 c6e7 d1d7 h7h6 d7a7');
INSERT INTO evaluations VALUES(46438,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(46438,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46438,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(46438,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46438,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46438,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(46438,36,'cp 18','e1g1 b8c6 c2c4 f8e7 b1c3 e8g8 c1e3 d5c4 e2c4 c6d4 f3d4 c5d4 e3d4 a7a6 c4b3 c8e6 b3e6 f7e6 g2g3 d8d7 d4f6 d7d1 f1d1 e7f6 c3e4 a8d8 e4f6 g7f6 a1c1 d8d1 c1d1 f8f7 d1d6 e6e5 g1g2 g8g7 a2a4 f7c7 d6b6 h7h5 h2h3 g7g6 g2f3 g6f5 a4a5 c7d7 g3g4 h5g4 h3g4 f5g5');
INSERT INTO evaluations VALUES(46436,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(46436,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(46436,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(46436,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46436,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46436,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(46436,36,'cp 27','b8c6 c2c4 f8e7 b1c3 e8g8 c1e3 c5d4 f3d4 c6d4 e3d4 d5c4 e2c4 a7a6 c4b3 b7b5 d1d3 c8b7 f2f3 e7c5 a1d1 c5d4 d3d4 d8a5 f1e1 a8d8 d4c5 d8c8 c5d6 b7c6 h2h3 a5b6 d6d4 b6c7 d4f2 h7h6 f2c5 c7b8 c5e5 b8e5 e1e5 f8d8 d1d8 c8d8 e5e7');
INSERT INTO evaluations VALUES(13755,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(13755,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(13755,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13755,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(13755,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(13755,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(13755,36,'cp 12','c2c4 f8e7 b1c3 e8g8 c1e3 d5c4 e2c4 c5d4 f3d4 c8d7 d4f3 d8c8 f1e1 d7g4 c3d5 f6d5 c4d5 c8f5 d5c6 b7c6 h2h3 g4f3 d1f3 f5f3 g2f3 e7f6 a1b1 f8d8 g1g2 f6d4 e3d4 d8d4 e1d1 d4d1 b1d1 a8b8 d1d6 a7a6 d6c6');
INSERT INTO evaluations VALUES(46662,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(46662,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(46662,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(46662,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(46662,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(46662,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(46662,36,'cp 21','b1c3 f8g7 g1f3 c8f5 e2e3 f6e4 c3e4 f5e4 f1d3 e4f3 d1f3 c7c6 e1c1 b8d7 h2h4 e7e5 f4h2 d8h4 h2e5 d7e5 h1h4 e5f3 g2f3 h7h5 h4h3 g7f6 f3f4 h5h4 c1d2 a8d8 d1h1 e8f8 d2e2 f8g7 b2b4 d6d5 c4c5 b7b6 e2f3 b6c5 b4c5 d8c8');
INSERT INTO evaluations VALUES(48547,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(48547,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48547,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48547,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(48547,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(48547,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(48547,36,'cp -29','c7c6 b5d3 f8b4 a2a3 b4a5 h2h3 e8g8 g1e2 b8d7 e1g1 f8e8 b2b4 a5b6 c3a4 b6c7 c1f4 d7f8 f4c7 d8c7 e2g3 a7a5 a4c5 b7b6 c5b3 a5a4 b3d2 c8a6 a1c1 f8g6 d3a6 a8a6 f1e1 e8e1 d1e1 a6a8 d2f3 a8e8');
INSERT INTO evaluations VALUES(45720,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(45720,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(45720,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(45720,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45720,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(45720,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(45720,36,'cp -12','b5d3 f8b4 g1e2 e8g8 e1g1 f8e8 e2g3 b8d7 f1e1 e8e1 d1e1 d7f8 c1g5 b4e7 h2h3 g7g6 g5h6 f8e6 c3e2 d8b6 a1b1 c8d7 c2c3 e7d6 e1d2 a8e8 g3f1 c6c5 d4c5 d6c5 h6e3 c5e3 f1e3 d5d4 e2d4 e6d4 c3d4 b6d4');
INSERT INTO evaluations VALUES(48782,30,'cp 173',NULL);
INSERT INTO evaluations VALUES(48782,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(48782,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(48782,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(48782,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(48782,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(48782,36,'cp 165','e4d5 f8d6 b1c3 e8g8 e1g1 c8f5 a2a4 h7h6 f3d2 b8d7 a4a5 a7a6 d2e4 d6b4 d1e2 f8e8 e4g3 f5g6 c3e4 g8h8 e2f3 f6h7 c2c3 b4f8 b2b4 d7f6 e4c5 f8c5 b4c5 e5e4 d3e4 f6e4 g3e4 e8e4');
INSERT INTO evaluations VALUES(48588,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(48588,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(48588,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(48588,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(48588,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(48588,35,'cp 158',NULL);
INSERT INTO evaluations VALUES(48588,36,'cp 149','f8d6 b1c3 e8g8 e1g1 c8f5 a2a4 a7a6 d1e2 f8e8 f3g5 d6b4 c3e4 f5g6 c2c3 b4f8 f1e1 h7h6 e4f6 d8f6 g5e4 f6d8 g2g4 b8d7 c1e3 f8d6 e2f3 e8f8 f3g2 g8h8 a4a5 f7f5 e4d6 c7d6');
INSERT INTO evaluations VALUES(52163,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(52163,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(52163,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(52163,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(52163,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(52163,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(52163,36,'cp 150','h2h4 d7d6 b1c3 f8e7 g2g3 g8f6 e2e4 e8g8 g1f3 b8a6 f1e2 c8d7 f3h2 a6c7 a2a4 c7a6 h2f1 a6b4 f1e3 g7g6 e3g2 h7h5 c1h6 f8e8 f2f3 f6h7 g3g4 g6g5 h4g5 h5g4 f3g4 e7g5 h6g5 d8g5 d1d2 g5d2 e1d2 g8g7 g2e3 e8h8 h1h5 a8f8 g4g5');
INSERT INTO evaluations VALUES(52776,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(52776,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(52776,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(52776,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(52776,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52776,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(52776,36,'cp 3','c8f5 c1f4 e7e6 e2e3 f8d6 f4d6 d8d6 f1e2 g8f6 e1g1 e8g8 c2c4 d5c4 e2c4 c6c5 d1b3 b8c6 d4c5 d6c5 c4e2 a7a6 b1d2 c5e7 f3h4 f8d8 f1d1 a8b8 a1c1 h7h6 h4f5 e6f5 d2c4 b7b5 c4d2 c6b4 a2a3 b4d5 e2f3 g7g6 b3c2 g8g7 b2b3');
INSERT INTO evaluations VALUES(53240,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(53240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53240,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(53240,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(53240,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(53240,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53240,36,'cp 4','d5d4 b2b4 g8f6 b4b5 b8d7 e3d4 c5d4 f3d4 d7c5 d4b3 e7e5 b1c3 c8e6 d1c2 d8d7 f1e2 e6f5 b3c5 f8c5 c2b3 e8g8 c1a3 c5a3 b3a3 a7a6 b5b6 d7c6 e1g1 f8d8 a1d1 f5c2 d1c1 c2d3 e2d3 d8d3 c4c5 d3d2 a3a4 c6a4 c3a4 d2a2');
INSERT INTO evaluations VALUES(53914,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(53914,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(53914,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(53914,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(53914,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(53914,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53914,36,'cp 9','d2d4 d6d5 f1b5 b8c6 g1f3 g8f6 e1g1 f8e7 d4c5 e7c5 b1c3 e8g8 c1g5 c8e6 d1d3 a8c8 a1e1 c5e7 a2a3 h7h6 g5h4 d8b6 h2h3 f8e8 e1e2 f6e4 c3d5 e6d5 d3d5 e7h4 b5c4 h4f6 e2e4 e8e4 d5e4 b6b2 f1d1 c6e5 c4b3 e5f3 g2f3 b2e5 d1d7 e5e4 b3f7 g8f8 f3e4');
INSERT INTO evaluations VALUES(54329,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(54329,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(54329,32,'cp 219',NULL);
INSERT INTO evaluations VALUES(54329,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(54329,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(54329,35,'cp 216',NULL);
INSERT INTO evaluations VALUES(54329,36,'cp 226','f1b5 c8d7 d1d4 e6d5 b5d7 d8d7 d4d5 g8f6 d5d7 b8d7 c1e3 f6g4 e3d4 e8c8 e1c1 f8c5 g1h3 c5d4 d1d4 d7f6 d4f4 h7h6 h3g1 a7a6 g1e2 h8e8 e2g3 e8e6 h2h3 g4e5 g3f5 g7g6 f5e3 g6g5 f4f5 e5g6 h1d1 d8d1 c1d1 g6e7 f5c5 c8b8 c5c4 e7g6 c4d4 h6h5 d1e2 e6e5 e2f1 g6e7 c3e2');
INSERT INTO evaluations VALUES(55494,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(55494,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(55494,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(55494,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(55494,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(55494,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(55494,36,'cp -21','a7a6 a2a4 a6b5 a4b5 a8a1 b2a1 c7c6 e2e3 g8f6 g1f3 d8a5 a1c3 a5a4 c3b2 c8f5 b1c3 a4c2 d1a1 f8e7 f1e2 c2b3 a1a8 e8g8 a8b7 b3b2 b7e7 c6b5 e1g1 b5b4 c3b5 f5d7 e7d6 b8c6 e2c4 e5e4 f3d4 f8c8 c4b3 b2d2 f1d1');
INSERT INTO evaluations VALUES(56963,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(56963,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(56963,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56963,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(56963,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(56963,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(56963,36,'cp 24','d7d5 d2d4 f8b4 e4e5 g8e7 d1d2 b7b6 a2a3 b4c3 d2c3 c8a6 f1a6 b8a6 c3d3 a6b8 c1e3 d8c8 h2h4 c7c5 h4h5 c5d4 f3d4 b8c6 f2f4 c6d4 e3d4 c8c4 e1c1 a8c8 d3c4 c8c4 b2b3 c4c7 g2g4 e8g8 c2c3 e7c6 h1h3 c6d4 d1d4 f8c8 c1b2 f7f6 e5f6 g7f6');
INSERT INTO evaluations VALUES(4076,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4076,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(4076,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4076,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4076,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4076,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(4076,36,'cp 0','e1g1 d8e8 e3e4 e7e5 f4c1 e5d4 c3d4 f6e4 d1c2 d7f6 f1e1 d6d5 c2c7 e8d8 c7d8 f8d8 h2h3 c8f5 c1f4 f6h5 f4c1');
INSERT INTO evaluations VALUES(4829,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(4829,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4829,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(4829,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4829,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4829,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(4829,36,'cp 10','d7d6 c1e3 g8f6 c3d5 e8g8 e3c5 d6c5 d5e3 b7b5 c4a2 d8d6 e1g1 f8d8 f3h4 c6d4 h2h3 g8f8 c2c3 d4e6 e3d5 f6d5 a2d5 c7c6 h4f5 d6c7 d5e6 c8e6 d1f3 f8g8 a1d1 f7f6 f3g4 e6f5 g4f5 c7d7 f5h5 d7d6 h5e2 d8e8 e2g4 a8d8 f1e1 d6e6 g4g3');
INSERT INTO evaluations VALUES(4756,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4756,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(4756,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(4756,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(4756,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4756,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(4756,36,'cp 19','c1e3 g8f6 c3d5 c5e3 f2e3 c8e6 e1g1 e8g8 d5f6 d8f6 f3d4 f6g6 c4e6 f7e6 d4c6 b7c6 d1e2 c6c5 b2b3 f8f1 a1f1 a8f8 f1f8 g8f8 g1f2 g6f6 f2e1 f6h4 e2f2 h4f2 e1f2 f8e8 f2e1 e8d8 e1d2 g7g5 d2c3 g5g4 b3b4 c5b4 a3b4 d8e7');
INSERT INTO evaluations VALUES(4913,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(4913,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(4913,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4913,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(4913,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(4913,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(4913,36,'cp 14','d7d6 c2c3 g8f6 e1g1 e8g8 c1g5 h7h6 g5h4 g7g5 h4g3 c5b6 a2a4 a8b8 a4b5 a6b5 b1a3 c6e7 f1e1 e7g6 d3d4 d8e7 a3c2 c8b7 f3d2 b8a8 a1a8 f8a8 c2e3 b7c8 e3f5 c8f5 e4f5 g6f4 d1f3 a8b8 g3f4 g5f4 g2g3');
INSERT INTO evaluations VALUES(6790,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(6790,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(6790,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(6790,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(6790,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(6790,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(6790,36,'cp 125','e4e5 d6c5 b2b4 c5f8 d1g4 c7c6 b5a4 d7d5 g4d4 g8h6 b1c3 b7b5 a4b3 a7a5 b4a5 d8a5 a2a4 b5b4 c3e2 h6f5 d4f4 f8e7 d2d4 e8g8 e2g3 f5h4 f4e3 c8e6 c1b2 h4f5 g3f5 e6f5 f1c1 a5b6 c2c4 d5c4 b3c4 f5e6 h2h3 e6c4');
INSERT INTO evaluations VALUES(7119,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(7119,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(7119,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(7119,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(7119,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(7119,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(7119,36,'cp 25','e8g8 h2h3 b7b5 a4c2 c8b7 b1d2 f8e8 d3d4 e7f8 f1e1 g7g6 a2a3 c6b8 b2b4 b8d7 c1b2 c7c6 a3a4 d8c7 c2b3 d6d5 e4d5 c6d5 d4e5 d7e5 a4b5 a6b5 e1e5 e8e5 a1a8 b7a8 d1a1 a8c6 f3e5 c7e5 d2f3 e5c7');
INSERT INTO evaluations VALUES(2574,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(2574,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(2574,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(2574,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(2574,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(2574,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(2574,36,'cp 31','f1e1 f8e8 h2h3 b7b5 a4c2 e7f8 c1g5 c6b8 b1d2 b8d7 d3d4 h7h6 g5h4 c8b7 b2b3 c7c6 a2a4 a6a5 c2d3 d8b6 h4g3 b7a6 d1c2 f6h5 g3h2 h5f6 a1b1 b6c7 e1c1 c7b6 d3f1 b6a7 g1h1 a8b8 a4b5 c6b5');
INSERT INTO evaluations VALUES(2575,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(2575,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(2575,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(2575,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(2575,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(2575,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(2575,36,'cp 25','f8e8 b1d2 e7f8 d2f1 h7h6 f1g3 b7b5 a4b3 c8d7 d3d4 a6a5 c1d2 a5a4 b3c2 d7g4 c2d3 d8b8 d4d5 c6e7 h2h3 g4d7 b2b3 e7g6 c3c4 g6f4 d3f1 a4b3 a2b3 b8b7 a1a8 e8a8 c4b5 d7b5 f1b5 b7b5 d2f4 e5f4 g3f5 c7c5 d5c6');
INSERT INTO evaluations VALUES(7952,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(7952,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(7952,32,'cp 193',NULL);
INSERT INTO evaluations VALUES(7952,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(7952,34,'cp 186',NULL);
INSERT INTO evaluations VALUES(7952,35,'cp 196',NULL);
INSERT INTO evaluations VALUES(7952,36,'cp 198','e4e5 f6d5 g5e7 d5e7 e1c1 e8g8 f2f4 b7b6 h1g1 e7f5 d4f2 f5e7 g2g4 c8b7 c3b5 f7f6 e5f6 f8f6 f2g3 d8f8 d1d7 b7c6 d7d1 a8d8 d1d8 f8d8 b5a7 c6d5 a7b5 e7g6 f4f5 e6f5 g4f5 f6f5 f1d3 f5f3 g3c7 d8c7 b5c7 d5f7');
INSERT INTO evaluations VALUES(9814,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9814,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9814,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9814,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9814,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9814,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9814,36,'cp 0','a7a5 e1g1 d7d6 f1e1 e8g8 b1d2 c8e6 h2h3 d8e7 c4b5 c6a7 b5c4');
INSERT INTO evaluations VALUES(7247,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(7247,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7247,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7247,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7247,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7247,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7247,36,'cp 0','e1g1 d7d6 f1e1 e8g8 h2h3 c8e6 b1d2 d8e7 b2b3 a8d8 c4b5 c6a7 b5c4');
INSERT INTO evaluations VALUES(10519,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(10519,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(10519,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(10519,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(10519,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(10519,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(10519,36,'cp -77','f6e4 d2e2 d8e7 e1g1 e4f6 e2d3 e8g8 f1e1 e7d8 d4d5 c6a5 d5d6 a5c4 d3c4 c7d6 b1c3 b7b6 c4f4 c8a6 f3d4 f8e8 d4f5 e8e6 f5d6 d8c7 e1d1 a8f8 a1c1 c7b8 f4d2 h7h6 h2h3 a6b7 d6b7 b8b7 c3b5 b7b8 b2b3 e6e5 b5d6 e5d5 d2f4 f6h5 f4f3 d5d6 f3h5 g8h7 d1d6 b8d6');
INSERT INTO evaluations VALUES(10702,30,'cp -73',NULL);
INSERT INTO evaluations VALUES(10702,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(10702,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(10702,33,'cp -52',NULL);
INSERT INTO evaluations VALUES(10702,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(10702,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(10702,36,'cp -67','d2e2 d8e7 e1g1 e4f6 e2d3 e8g8 f1e1 e7d8 d4d5 c6a5 d5d6 b7b6 b1c3 a5c4 d3c4 c7d6 c4f4 c8b7 f3d4 b7c6 h2h3 f8e8 f4d6 d8b8 e1e8 f6e8 d6b8 a8b8 a1d1 c6b7 d4f5 b8d8 f2f3 g8f8 g1f2 e8c7 f5d6 b7c6 h3h4 c7e8 c3b5 e8d6');
INSERT INTO evaluations VALUES(13260,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(13260,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(13260,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(13260,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(13260,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(13260,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(13260,36,'cp -13','e7e6 c4a2 b7b6 g1f3 f8e7 e1g1 c8b7 h2h3 e8g8 f1e1 d8c7 c3b5 c7d7 b5c3 a7a6 c3e2 c6e5 e2g3 e5f3 d1f3 d6d5 c1f4 d5e4 g3e4 f6d5 f4h2 a8d8 c2c3 h7h6 a1d1 d7c6 f3e2 d5f6 a2b1 f6e4');
INSERT INTO evaluations VALUES(13944,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(13944,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(13944,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(13944,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(13944,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(13944,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(13944,36,'cp -13','g1f3 f8e7 e1g1 a7a6 f1e1 e8g8 c4a2 b7b5 c3e2 c8b7 c2c3 c6e5 e2g3 d8b6 c1f4 e5f3 d1f3 a6a5 f3d1 b5b4 h2h3 f8d8 a1c1 a8c8 f4e3 b7c6 a3b4 a5b4 a2b1 c8a8 d3d4 b6b7 d1f3 a8a1 b1d3 a1a2 c3b4');
INSERT INTO evaluations VALUES(15532,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(15532,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(15532,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(15532,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(15532,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(15532,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(15532,36,'cp 40','d2d4 e5d4 e3d4 f8e7 h2h3 e8g8 f1d3 d6d5 e1g1 c6b4 c4c5 b4d3 d1d3 f8e8 f1e1 c7c6 b2b4 c8d7 f3e5 f6e4 c3e4 d5e4 d3e4 e7f8 e4d3 f7f6 e5f3 e8e1 f3e1 d7e6 a2a4 d8d7 c1d2 a8d8 e1f3 d7d5 b4b5');
INSERT INTO evaluations VALUES(15833,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(15833,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(15833,32,'cp -125',NULL);
INSERT INTO evaluations VALUES(15833,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(15833,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(15833,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(15833,36,'cp -76','h7h6 g5f3 e5e4 f3e5 a5c4 b3c4 f8d6 d2d4 e4d3 e5d3 e8g8 e1g1 c8g4 f2f3 g4f5 c1e3 b7b5 b1d2 b5c4 d2c4 f6d5 e3f2 d6e7 d3e5 e7f6 f3f4 d5b6 c4e3 d8d1 f1d1 f5e6 c2c4 f6e5 f4e5 f8d8 a1c1 g8h7 d1d8 a8d8 e3d5 d8d7 a2a4 b6c4 d5f4 c4b2 a4a5 a7a6 h2h4 d7d1 c1d1 b2d1 f2c5');
INSERT INTO evaluations VALUES(16055,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(16055,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(16055,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(16055,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(16055,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(16055,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(16055,36,'cp 40','f8g7 h2h3 d7d6 a2a4 h7h6 c1e3 e8g8 a4a5 c8d7 e1g1 d8e7 f1e1 a7a6 c3e2 d7e6 c4e6 e7e6 c2c4 g8h7 d1b3 e6c8 e3d2 f6d7 b3c2 f7f5 b2b4 f5e4 d3e4 c8e8 c2d3 e8f7 a1d1 a8e8 e2c3');
INSERT INTO evaluations VALUES(16806,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(16806,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(16806,32,'cp -105',NULL);
INSERT INTO evaluations VALUES(16806,33,'cp -133',NULL);
INSERT INTO evaluations VALUES(16806,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(16806,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(16806,36,'cp -137','f8c5 c1e3 c5e3 f2e3 d8e7 d1d5 e4g5 b1d2 e8g8 f1c4 g5f3 d2f3 d7d6 e1c1 d6e5 d5e4 g7g6 c4d5 c8f5 e4h4 e7h4 f3h4 f5g4 h4f3 a8e8 h2h3 g4f3 g2f3 e8d8 h3h4 d8d6 h4h5 f8d8 e3e4 g8g7 h5g6 h7g6 d5c6 d6d1 h1d1');
INSERT INTO evaluations VALUES(20111,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(20111,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(20111,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20111,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(20111,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(20111,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(20111,36,'cp -13','b1c3 a7a6 a2a4 g8f6 d2d4 c5d4 c4d4 c8d7 c1g5 e7e6 g1e2 h7h6 g5h4 d7c6 a1d1 a8c8 e1g1 f8e7 h4f6 e7f6 d4d6 d8b6 d6a3 a6a5 d1d3 b6b4 a3b3 e8g8 f1d1 f8d8 h2h3 d8d3 d1d3 c8d8 d3d8 f6d8 g1f1 b4b3 c2b3 g8f8');
INSERT INTO evaluations VALUES(20513,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(20513,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(20513,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(20513,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(20513,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20513,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(20513,36,'cp 9','c5d4 f3d4 g8f6 f1b5 c8d7 e1g1 f8d6 d4f3 d7g4 b2b3 e8g8 c1b2 a7a6 b5e2 f8e8 h2h3 g4h5 d1d2 d8d7 f3d4 h5g6 a1d1 d6c7 d4c6 b7c6 e2f3 a8c8 c3e2 f6e4 d2c2 c6c5 f3e4 g6e4');
INSERT INTO evaluations VALUES(21912,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(21912,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(21912,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(21912,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(21912,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(21912,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(21912,36,'cp 20','d2d3 d8f6 f4e5 c6e5 g1f3 e5f3 d1f3 f6f3 g2f3 g7g6 c1e3 f8g7 b1c3 c8e6 c4b3 e6b3 a2b3 e8d7 e1d2 f7f5 b3b4 f5e4 f3e4 g8f6 b4b5 h8f8 h2h3 a7a6 b5a6 a8a6 a1a6 b7a6 h1a1 f6h5 a1a6 g7c3 d2c3 f8f3 e3d4 h5f4 a6a7 f3h3 c3c4');
INSERT INTO evaluations VALUES(24871,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(24871,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(24871,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(24871,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(24871,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(24871,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(24871,36,'cp 52','c2c4 e7e6 b2b4 c6b8 c4c5 c7c6 f1e2 f8e7 b1d2 f6e4 c1b2 e4d2 f3d2 b8d7 d2b3 e8g8 d1d2 d8c7 e1g1 h7h6 a3a4 a7a6 b2c3 e6e5 b4b5 a6b5 a4b5 e5d4 c3d4 c6b5 d2c3 a8a1 f1a1 f8c8 a1a7 d7c5 b3c5');
INSERT INTO evaluations VALUES(24806,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(24806,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(24806,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(24806,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(24806,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(24806,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(24806,36,'cp 36','e7e6 b2b4 c6b8 c4c5 c7c6 f1e2 f8e7 c1b2 b8d7 e1g1 e8g8 f3h4 f6e4 h4f5 e6f5 b1c3 a7a6 a3a4 g7g6 b4b5 d7f6 d1c2 a6b5 a4b5 d8d7 b5b6 h7h5 a1a7 a8a7 b6a7 f8a8 f1a1 g8g7 f2f3 e4c3 b2c3 e7d8');
INSERT INTO evaluations VALUES(25146,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(25146,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(25146,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(25146,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(25146,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(25146,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(25146,36,'cp 42','b1d2 f8e7 b2b4 e8g8 c4c5 c6b8 c1b2 b8d7 d2b3 c7c6 f1d3 f5d3 d1d3 b7b6 e1g1 a7a5 d3c2 d8b8 b2c3 f6e4 c5b6 a5b4 c3b4 e7b4 a3b4 b8b6 b3c5 e4c5 b4c5 b6c7 f1b1 h7h6 h2h3 e6e5 d4e5 d7e5 f3d4 e5c4 c2c3 f8b8 b1b8 a8b8 d4f5 c7e5 c3e5 c4e5 a1a7');
INSERT INTO evaluations VALUES(27937,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(27937,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(27937,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(27937,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(27937,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(27937,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(27937,36,'cp 67','g1e2 g8f6 a2a3 b4c5 e2g3 e8g8 f1e2 f8e8 e1g1 e6d5 e4d5 c5f8 c1f4 a7a6 d5d6 c7d6 g3f5 b6b5 f4d6 e8e6 d1d2 b8c6 e2f3 c6a5 f3b7 a5b7 d6f8 d8f8 c3d5 b7d6 f5e7 g8h8 d5f6 f8e7 f6d5 e6e2 d2f4 e7e5 f4e5 e2e5');
INSERT INTO evaluations VALUES(28328,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(28328,31,'cp 214',NULL);
INSERT INTO evaluations VALUES(28328,32,'cp 226',NULL);
INSERT INTO evaluations VALUES(28328,33,'cp 229',NULL);
INSERT INTO evaluations VALUES(28328,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(28328,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(28328,36,'cp 240','e4e5 f6g8 h2h4 h7h6 d1e2 d7d5 c4b5 c7c6 b5d3 e7e6 h4h5 g6g5 e1g1 g8e7 c1d2 a7a6 c3d1 c6c5 c2c3 a6a5 d1e3 b8c6 f1e1 d8c8 d3b5 b7a6 a2a4 c8b7 e3g4 e7f5 d2e3 c5d4 f3d4 a6b5 e2b5 f5e3 f2e3 a8c8 e3e4 d5e4 e1e4');
INSERT INTO evaluations VALUES(28327,30,'cp 243',NULL);
INSERT INTO evaluations VALUES(28327,31,'cp 226',NULL);
INSERT INTO evaluations VALUES(28327,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(28327,33,'cp 234',NULL);
INSERT INTO evaluations VALUES(28327,34,'cp 245',NULL);
INSERT INTO evaluations VALUES(28327,35,'cp 248',NULL);
INSERT INTO evaluations VALUES(28327,36,'cp 220','f6g8 h2h4 h7h6 d1e2 d7d5 c4b5 c7c6 b5d3 e7e6 h4h5 g6g5 e1g1 g8e7 c1d2 b8d7 a1e1 b6b5 c3d1 c6c5 c2c3 b5b4 d1e3 a7a5 e3g4 e7c6 d2e3 c5d4 c3d4 d8b6 e1c1 b7a6 f3e1 b4b3 a2a3 a6d3 e2d3 b6a6 f2f4 a6d3 e1d3 g5f4');
INSERT INTO evaluations VALUES(29415,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(29415,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29415,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(29415,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(29415,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(29415,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(29415,36,'cp 1','e7e6 f1d3 f5d3 c2d3 f8d6 f3e5 f6d7 e1g1 d6e5 d4e5 c7c5 d1g4 g7g5 f4g3 b8c6 f2f4 h6h5 g4d1 g5f4 g3f4 c6e5 b1c3 e5g6 f4h2 a7a6 g1h1 d8g5 d1e2 a8c8 e2f2 f7f5 e3e4 d5e4 d3e4 f5f4 a1e1');
INSERT INTO evaluations VALUES(30987,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(30987,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(30987,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30987,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(30987,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(30987,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(30987,36,'cp 30','f5d3 d1d3 e7e6 e1g1 c6c5 c1e3 c5d4 e3d4 b8c6 c2c4 d5c4 d3c4 g8e7 f1d1 e7f5 b1c3 d8c8 d4c5 f8c5 c4c5 f5e7 c3e4 e8g8 e4d6 c8b8 f3d4 b8c7 f2f4 a8d8 d4c6 e7c6 b2b4 d8d7 a2a4 b7b6 c5c4 c6e7 c4c7 d7c7 a1c1 c7c1 d1c1 e7d5 c1c4');
INSERT INTO evaluations VALUES(30978,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(30978,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(30978,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(30978,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(30978,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(30978,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(30978,36,'cp 36','d1d3 e7e6 e1g1 c6c5 c2c3 b8c6 a2a3 a7a6 h2h3 c5d4 c3d4 g8e7 b1c3 c6a5 c3e2 d8c7 b2b4 a5c4 c1d2 e7g6 h3h4 f8e7 g2g3 e7d8 a1c1 c7d7 h4h5 g6e7 c1c2 a8c8 f1c1 e7c6 e2f4 c6e7 g1g2 e8g8 g3g4');
INSERT INTO evaluations VALUES(30977,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(30977,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(30977,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(30977,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(30977,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(30977,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(30977,36,'cp 21','e7e6 e1g1 c6c5 c2c3 b8c6 a2a3 a7a6 h2h3 c5d4 c3d4 g8e7 b1c3 c6a5 c3e2 d8c7 b2b4 a5c4 e2f4 g7g6 d3e2 f8g7 f4d3 e8g8 a3a4 f8c8 c1d2 e7f5 g2g4 f5e7 a1c1 c7d8 g1g2 e7c6');
INSERT INTO evaluations VALUES(31355,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(31355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31355,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31355,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(31355,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(31355,36,'cp -24','h2h3 g4f3 d1f3 e7e6 b2b3 b8c6 b1c3 a7a6 c1b2 h7h5 g2g3 g8f6 f1d3 c6b4 e1c1 d8a5 c1b1 a8d8 f4f5 d5d4 c3a4 b4d3 c2d3 a5d5 f3d5 f6d5 b2d4 b7b5 a4c3 d5e7 d4b6 d8b8 b6f2 e7f5 g3g4 f5d6 c3e4 h5g4 h3g4 d6e4');
INSERT INTO evaluations VALUES(31454,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(31454,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(31454,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31454,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31454,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31454,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31454,36,'cp 9','d8b6 d1b3 g8f6 b1d2 h7h5 a2a4 f6e4 h2h4 b8d7 a4a5 b6b3 d2b3 f8d6 f4d6 e4d6 b3c5 f7f6 f3d2 e8e7 f1e2 b7b6 c5d7 e7d7 c3c4 b6b5 c4c5 d6c4 b2b3 c4d2 e1d2 a7a6 e2d3 f5d3 d2d3 e6e5 f2f3 d7c7 e3e4');
INSERT INTO evaluations VALUES(32787,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(32787,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(32787,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(32787,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(32787,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(32787,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(32787,36,'cp 151','g2g4 f5e6 f1g2 a5c7 g4g5 b8d7 d1e2 d7b6 g1f3 e6c4 e2e5 c7e5 f3e5 f7f6 e5c4 b6c4 d4d5 e8c8 e1c1 e7e5 d5c6 b7c6 g2h3 c8c7 d2e3 c4e3 d1d8 c7d8 f2e3 f6g5 h1d1 d8e8 h3d7 e8f7 d7c6 g8f6 d1d8 g5g4 d8a8 f7g6 b2b4 h7h5 a2a3 h5h4 c1b2 g6f5 a8a7 g4g3 h2g3 h4h3 c6h1 e5e4 b2b3 f8d6 a7g7 d6e5');
INSERT INTO evaluations VALUES(34498,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(34498,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(34498,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(34498,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(34498,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(34498,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(34498,36,'cp -3','g8e7 c2c3 h7h5 h2h4 c7c6 a2a4 a7a5 b1a3 c8f5 g1e2 b8d7 c1g5 f7f6 g5f4 e8f7 e2g3 f5g6 e1c1 d7b6 b2b3 e7f5 a3c2 f5g3 f4g3 b6c8 f2f3 f8d6 g3d6 c8d6 c2e3 d6f5 e3f5 g6f5 c1d2 f7e6 f1d3 f5d3 d1e1 e6f7 d2d3 g7g6');
INSERT INTO evaluations VALUES(34578,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(34578,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(34578,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(34578,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(34578,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(34578,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(34578,36,'cp 125','e7e5 d2d4 e5d4 f3d4 g8f6 f1e2 f8d6 c1e3 c5a5 d4b5 d6b4 a2a3 b4c3 b5c3 e8g8 b2b4 a5e5 d1d4 b8c6 d4e5 c6e5 e1c1 c7c6 h1e1 c8f5 e3d4 f8e8 h2h3 e5g6 a3a4 g6f4 e2f3 a7a6 e1e8 a8e8 d4e3 f4e6 a4a5 h7h5 d1d6 h5h4 c1b2 f5g6');
INSERT INTO evaluations VALUES(36472,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(36472,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(36472,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(36472,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(36472,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(36472,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36472,36,'cp 35','b1c3 c7c5 d4d5 h7h6 g5h4 g6g5 h4g3 c8f5 f1d3 f5d3 d1d3 f6h5 e1g1 f7f5 h2h3 d8e8 c3b5 e8c8 g3h2 b8d7 b5c3 c8e8 c3e2 b7b6 a1b1 e8f7 b2b3 a7a6 f3d2 f5f4 e3f4 g5f4 g2g4 f4f3 g4h5 f3e2 d3e2 d7f6 f2f4 f7h5 e2h5 f6h5 g1h1');
INSERT INTO evaluations VALUES(39110,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(39110,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(39110,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39110,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(39110,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39110,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(39110,36,'cp 37','b8c6 f1e2 g8f6 e1g1 e8g8 c1e3 f6g4 e3d2 d6c5 b1c3 g4f6 c3a4 c5e7 c2c4 d7d6 a4c3 a7a5 h2h3 h7h6 d2e3 f6h7 d3d4 h7g5 d4e5 g5f3 e2f3 d6e5 c3d5 a5a4 f3g4 e7g5 e3c5 f8e8 b2b4 a4b3 g4c8 d8c8 a2b3 a8a1');
INSERT INTO evaluations VALUES(6779,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(6779,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(6779,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(6779,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(6779,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(6779,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(6779,36,'cp 37','f1e2 g8f6 c1e3 f6g4 e3d2 e8g8 e1g1 f8e8 h2h3 g4f6 d2e3 b7b5 b1d2 h7h6 c2c3 d6f8 b2b4 d7d5 a2a4 d5d4 c3d4 e5d4 e3f4 c6b4 a4b5 c7c6 b5c6 b4c6 d1c1 c8d7 c1b2 a7a5 f3d4 a5a4 d4c6 d7c6 b2c3 c6d7 f4e3 a4a3');
INSERT INTO evaluations VALUES(39675,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(39675,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(39675,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(39675,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(39675,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(39675,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(39675,36,'cp 33','c1e3 c8e6 d1d2 e8g8 e1c1 b8d7 g2g4 b7b5 g4g5 f6h5 c1b1 d7b6 b3a5 a8c8 a2a3 g7g6 h2h4 h5g3 h1g1 g3h5 c3d5 b6d5 e4d5 e6d5 d2d5 d8a5 f1d3 a5c7 c2c4 b5c4 d1c1 c7d7 d3c4 h5f4 e3f4 c8c5 d5e4 e5f4 g1d1 f8c8 e4f4 d6d5 c4a6 c5c1 d1c1 c8c1 f4c1 d7f5 b1a2 f5f3');
INSERT INTO evaluations VALUES(41384,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(41384,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(41384,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(41384,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(41384,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(41384,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(41384,36,'cp 33','g1f3 e7e5 c2c3 e8g8 e1g1 d6d5 d4e5 f6e4 b1d2 c8f5 d3e4 d5e4 f3d4 b8c6 d4f5 g6f5 f2f4 e4f3 d2f3 c6e5 f3e5 d8d1 a1d1 g7e5 f1f5 f7f6 e3f4 e5f4 f5f4 a8e8 f4g4 g8f7 d1d7 e8e7 d7e7 f7e7 g1f2 e7f7 g4d4 f7e6 d4h4 f8f7 h4e4 e6f5 e4d4 f5e6 d4d8 c7c5 a2a3 b7b5 d8a8');
INSERT INTO evaluations VALUES(41796,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(41796,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(41796,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(41796,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(41796,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(41796,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(41796,36,'cp 7','f4g5 d6e7 e2e3 c7c5 c2c3 b8d7 f1d3 d8c7 g5f4 e7d6 f4d6 c7d6 d1e2 e8g8 e3e4 d5e4 d2e4 f6e4 e2e4 d7f6 e4e5 f8d8 e1c1 c5d4 e5d4 d6d4 f3d4 g8f8 h1e1 c8d7 g2g3 a8c8 f2f4 c8c7 d3b5 g7g6 a2a4 d8c8 b5d7 c7d7 d4b5 a7a6 d1d7');
INSERT INTO evaluations VALUES(42620,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42620,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42620,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42620,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(42620,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(42620,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42620,36,'cp 0','f1c4 c7c6 d2d3 b7b5 c4b3 e8g8 c1g5 c5e7 g1e2 h7h6 g5e3 d7d6 a2a3 b8d7 h2h3 d7c5 d3d4 c5b3 c2b3 c8e6 b3b4 a7a5 e1g1 a5b4 a3b4 f8e8 a1a8 d8a8 f1d1 e6b3 d1c1 b3e6');
INSERT INTO evaluations VALUES(44266,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(44266,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(44266,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(44266,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(44266,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(44266,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(44266,36,'cp 85','c2c4 g7g6 b1c3 f8g7 e2e4 e8g8 h2h3 f8e8 c1e3 e6d5 e4d5 b7b5 c4b5 d7d6 f1c4 b8d7 e1g1 d7b6 f3d2 c8b7 e3g5 h7h6 g5h4 d8d7 d1f3 d7f5 f3f5 g6f5 h4f6 g7f6 a2a4 f6c3 b2c3 b7d5 c4d5 b6d5 d2c4');
INSERT INTO evaluations VALUES(46896,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(46896,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46896,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46896,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(46896,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(46896,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(46896,36,'cp -16','c7c5 e2e3 b8c6 h4f6 e7f6 b1c3 c5c4 a2a3 d8a5 g2g3 c8e6 f1g2 f8d6 e1g1 f6f5 c3e2 e8g8 f3d2 c6e7 a3a4 a5c7 c2c3 g7g5 b2b3 c4b3 d1b3 f8c8 a4a5 f5f4 g3f4 g5f4 e3f4 g8h8 f1e1 d6f4 e2f4 c7f4');
INSERT INTO evaluations VALUES(47985,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(47985,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(47985,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(47985,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(47985,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47985,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(47985,36,'cp 35','g1f3 g7g6 f1c4 d5e7 e1g1 f8g7 h2h3 e8g8 f1e1 h7h6 b1c3 b8d7 a2a4 d7f6 c1e3 g6g5 a4a5 f8e8 a5a6 b7a6 c4a6 c8a6 a1a6 d8c8 d1a1 c8f5 a6a7 a8a7 a1a7 f5c2 a7c7 c2b2 e3d2 e7d5 c3d5 e8e1 d2e1 f6d5 c7d8 g8h7 d8d6 b2a2 d6d7 d5f4 d7g4 a2a1');
INSERT INTO evaluations VALUES(47984,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(47984,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(47984,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(47984,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(47984,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(47984,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(47984,36,'cp 40','g7g6 f1c4 d5e7 h2h3 f8g7 a2a4 e8g8 e1g1 b8d7 f1e1 d7f6 a4a5 c8d7 b1c3 h7h6 c1f4 a7a6 d1d2 g6g5 f4g3 e7f5 g3h2 f8e8 e1e8 d8e8 a1e1 e8f8 c4d3 f5e7 c3e4 f6e4 d3e4 d7c6 d4d5 c6d7 c2c3 e7g6 e4c2 a8e8 f3d4');
INSERT INTO evaluations VALUES(50379,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(50379,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(50379,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(50379,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(50379,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(50379,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(50379,36,'cp -30','e7e5 d2d4 e5d4 e3d4 d7d5 g1f3 f8d6 f1g2 d8e7 c1e3 f6g4 d1d3 g4e3 d3e3 b8d7 e3e7 d6e7 e1g1 d7b6 b1d2 e8g8 a2a4 a7a5 f1e1 e7d6 d2b3 c7c6 f3e5 f8e8 c2c3 c8f5 b3c5 d6c5 d4c5 b6d7 e5d7 f5d7');
INSERT INTO evaluations VALUES(50655,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(50655,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(50655,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(50655,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(50655,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(50655,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(50655,36,'cp -15','g8f6 b1c3 b8c6 c4a2 b7b5 e1g1 e7e6 d2d4 c6d4 f3d4 c5d4 d1d4 f8e7 c1g5 c8b7 f2f4 e8g8 f4f5 e6f5 f1f5 b7c8 f5f3 c8e6 a2e6 f7e6 a1d1 a8c8 g5f6 f8f6 f3f6 e7f6 d4d6 d8d6 d1d6 a6a5 d6e6 f6c3 b2c3 c8c3 e4e5 c3a3');
INSERT INTO evaluations VALUES(51193,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(51193,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(51193,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(51193,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(51193,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(51193,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(51193,36,'cp -48','b8c6 e1g1 c8b7 d1e2 a8c8 f1d1 f8c5 c1g5 g8e7 a1c1 d8b6 c3d5 e6d5 e4d5 e8g8 d5c6 e7c6 g5e3 c5e3 e2e3 b6e3 f2e3 f8d8 f3g5 c6e5 c1c8 b7c8 g5e4 d7d6 e4d6 c8e6 b3e6 f7e6 d1d4 e5c6 d4d2 g8f8 d6e4 d8d2 e4d2 c6e5 b2b3 f8e7 d2e4 e5d3 a2a3 e6e5 g1f1 e7e6 e4g5 e6f5 g5h7 d3c1');
INSERT INTO evaluations VALUES(52250,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(52250,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(52250,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(52250,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(52250,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(52250,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(52250,36,'cp 56','d7d5 c4d5 c6d5 c1f4 b8c6 e2e3 g8f6 g1f3 f6h5 f4e5 f7f6 e5g3 e7e6 f1d3 c8d7 e1g1 f8e7 e3e4 d5e4 d3e4 f6f5 e4c6 d7c6 g3e5 e8g8 f1e1 d8d7 h2h3 a8c8 a1c1 a7a6 d1e2 e7f6 e1d1 c8e8 e2e3 f6d8 b2b4 f8f7 f3g5');
INSERT INTO evaluations VALUES(54590,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(54590,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(54590,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(54590,33,'cp 164',NULL);
INSERT INTO evaluations VALUES(54590,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(54590,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(54590,36,'cp 188','f1b5 c8d7 d1e2 f8e7 c3d5 d7c6 e2e5 f7f6 b5c6 b8c6 e5e4 d8d7 f3d4 c6d4 e4d4 a8d8 c2c4 e7b4 c1d2 b4d2 d4d2 g8e7 d2e2 e8f7 e1c1 e7d5 d1d5 d7c7 h1d1 d8d5 d1d5 h8e8 e2g4 e8e7 g2g3 c7b6 g4f3 g7g6 b2b3 b6e6 d5d3 e6h3 c1b2 h3f1 d3d6 f1e2 f3e2 e7e2 b2c3 e2a2 d6d7 f7e6 d7b7 a2f2');
INSERT INTO evaluations VALUES(673,30,'cp 300',NULL);
INSERT INTO evaluations VALUES(673,31,'cp 304',NULL);
INSERT INTO evaluations VALUES(673,32,'cp 289',NULL);
INSERT INTO evaluations VALUES(673,33,'cp 282',NULL);
INSERT INTO evaluations VALUES(673,34,'cp 303',NULL);
INSERT INTO evaluations VALUES(673,35,'cp 298',NULL);
INSERT INTO evaluations VALUES(673,36,'cp 300','d2d4 g7g6 e1g1 f8g7 e5f3 e8g8 e4e5 f6e8 f1e1 h7h6 b2b3 f7f6 c1b2 g6g5 c3e4 f6f5 e4c5 g5g4 f3d2 f5f4 c5d3 h6h5 d2c4 e7e6 b2a3 f8f5 g2g3 g7h6 a3c1 f4g3 f2g3 e8g7 d3c5 e6g6 c1h6 g6h6 c5e4 a6a5 e4f6 g8h8 d4d5 d7d6 c4e3 f5f3 d1d2 c8b7');
INSERT INTO evaluations VALUES(672,30,'cp 285',NULL);
INSERT INTO evaluations VALUES(672,31,'cp 310',NULL);
INSERT INTO evaluations VALUES(672,32,'cp 306',NULL);
INSERT INTO evaluations VALUES(672,33,'cp 301',NULL);
INSERT INTO evaluations VALUES(672,34,'cp 307',NULL);
INSERT INTO evaluations VALUES(672,35,'cp 284',NULL);
INSERT INTO evaluations VALUES(672,36,'cp 334','g7g6 e1g1 f8g7 e5f3 e8g8 e4e5 f6d5 c3e4 f7f5 e4c5 h7h6 c2c4 d5b6 b2b3 g6g5 c5d3 d7d5 e5d6 c7d6 f1e1 e7d8 c4c5 d6c5 c1a3 f8e8 e1e8 d8e8 a3c5 b6d5 d1c2 c8e6 a1e1 g5g4 f3e5 f5f4 c5a3 e6f5 c2c4 g8h7 e5c6 e8g8 d3b4 d5b4 a3b4 g8c4 b3c4 g7f6 d4d5 f4f3 c6e7 f6e7 e1e7 h7g6 e7e1 g6f7');
INSERT INTO evaluations VALUES(4079,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4079,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(4079,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(4079,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(4079,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(4079,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(4079,36,'cp 22','b7b6 a2a4 c8b7 b1d2 a7a6 c2c3 f8e8 f4h2 e7e5 d1b3 d8e7 f1d1 a6a5 d4e5 d7e5 f3e5 d6e5 h2g3 e8d8 g3h4 h7h6 b3c2 e7c5 a1c1 a8c8 b2b3 c5e7 c1b1 e5e4 e2c4 g6g5 h4g3 c7c5 c4b5 b7c6 b5c6 c8c6');
INSERT INTO evaluations VALUES(9414,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(9414,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9414,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(9414,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(9414,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(9414,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(9414,36,'cp -36','a7a5 f1e2 e8g8 e1g1 c8d7 f1e1 f8e8 e2f1 h7h6 c3e2 a5a4 e2g3 d6d5 e4d5 f6d5 c1d2 c5b6 c2c4 d5f4 d3d4 b6d4 f3d4 c6d4 d2f4 e5f4 d1d4 f4g3 f2g3 d7c6 d4f4 d8g5 f4g5 h6g5 f1d3 a8d8 e1e8 d8e8 a1d1 g8f8 g1f2 e8e5 d1d2');
INSERT INTO evaluations VALUES(10987,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(10987,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(10987,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(10987,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(10987,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(10987,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(10987,36,'cp 65','d1d4 c7c6 a2a4 d7d6 c1f4 e8g8 a1d1 d8a5 b2b4 a5h5 e1g1 d6d5 e4d5 e7b4 d5c6 b4c5 d4e5 h5e5 f4e5 b7c6 e5f6 g7f6 c3e4 c5e7 e4d6 c8e6 c4e6 f7e6 f1e1 e6e5 d6f5 e7b4 e1e4 f8d8 d1d8 a8d8 g1f1 b4c5 e4c4 d8d5 f1e2 g8f7 g2g4 f7e6 c4c3 e6d7');
INSERT INTO evaluations VALUES(11050,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(11050,31,'cp 209',NULL);
INSERT INTO evaluations VALUES(11050,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(11050,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(11050,34,'cp 201',NULL);
INSERT INTO evaluations VALUES(11050,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(11050,36,'cp 201','e4e5 f6g8 c1f4 d7d6 f1b5 a7a6 b5c6 b7c6 d1d2 f7f6 d2e2 g6g5 e5f6 g8f6 f4g5 e8g8 e1c1 h7h6 g5f6 f8f6 f3h4 f6f7 g2g3 c8h3 f2f4 d8c8 d1d3 a8b8 h1e1 c8b7 b2b3 b7b4 c1b2 h3d7 b2a1 c6c5');
INSERT INTO evaluations VALUES(13523,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(13523,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(13523,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(13523,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(13523,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13523,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(13523,36,'cp 17','c5d4 f3d4 c6d4 d1d4 d5c4 d4d8 e8d8 f1c4 f8d6 c3e4 d6e5 e1c1 d8e7 f2f4 e5c7 g5h4 h7h6 e4f6 g7f6 h1f1 h8d8 g2g3 b7b5 d1d8 c7d8 c4b5 c8b7 f1d1 a8c8 c1b1 b7e4 b1a1 d8a5 a2a3 c8c2 g3g4 e4d5 e3e4 c2h2 e4d5 h2h4 d5d6 e7d8');
INSERT INTO evaluations VALUES(13915,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(13915,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(13915,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(13915,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(13915,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(13915,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(13915,36,'cp -82','a7a6 f1e1 f8e7 d2d3 h7h6 g5f3 b7b5 c4b3 d6d5 a2a3 d5e4 c3e4 f6e4 d3e4 c5c4 d1d8 e7d8 b3a2 d8b6 b2b3 c4b3 c2b3 c8b7 c1e3 b6e3 e1e3 e8e7 a1c1 h8d8 b3b4 a8c8 e3c3 a6a5 b4a5 c6a5 c3c8 b7c8 f3e5 d8d2 f2f4 c8a6 c1c7 e7e8 c7a7 d2a2 a7a6 a2a3');
INSERT INTO evaluations VALUES(14119,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(14119,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(14119,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(14119,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14119,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(14119,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(14119,36,'cp 55','e1g1 f8g7 b2b3 e8g8 c1b2 f8e8 b1d2 c8f5 c2c4 e7e6 h2h3 h7h5 f3e5 f6d7 f2f4 d7e5 d4e5 d8e7 a1c1 a8d8 d1e1 d5d4 d2e4 d4d3 b2c3 d3e2 e1e2 c6d4 e2e3 c7c5 f1d1 b7b6 d1d2 f5e4 g2e4 d4f5 e4f5 e6f5 c1d1 d8d2 e3d2');
INSERT INTO evaluations VALUES(14118,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(14118,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(14118,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(14118,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(14118,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(14118,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(14118,36,'cp 59','f8g7 b2b3 e8g8 c1b2 f8e8 b1d2 c8f5 c2c4 e7e6 h2h3 h7h5 f3e5 f6d7 f2f4 d7e5 d4e5 d8e7 d1e1 a8d8 e1f2 a7a5 c4d5 e6d5 d2f3 f7f6 e5f6 g7f6 b2f6 e7f6 a1d1 f5e4 d1d2 c6e7 f3e5 e4g2 g1g2 e7c6 e5f3 d5d4 f1d1 e8e3 g2h2');
INSERT INTO evaluations VALUES(14937,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(14937,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(14937,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(14937,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(14937,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(14937,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(14937,36,'cp 18','d8d5 c2c4 f8b4 c1d2 b4d2 d1d2 d5d6 e1g1 e8g8 b5c6 b7c6 f1e1 f8e8 d3d4 e5d4 e1e8 f6e8 d2d4 c8g4 b1d2 g4f3 d2f3 c6c5 d4c3 a8d8 a1e1 g7g6 f3g5 e8g7 b2b3 d6d2 c3d2 d8d2 g2g3 h7h6 g5e4 d2a2 e4c5 g8f8 h2h3 g7f5 e1d1 f5d6 g1f1 a2a3 f1g2 f8e7 d1d5 e7f6');
INSERT INTO evaluations VALUES(954,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(954,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(954,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(954,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(954,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(954,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(954,36,'cp 20','c2c4 f8b4 c1d2 b4d2 d1d2 d5d6 e1g1 e8g8 b5c6 b7c6 f1e1 f8e8 d3d4 e5d4 e1e8 f6e8 d2d4 c8g4 b1d2 g4f3 d2f3 c6c5 d4c3 g7g6 a1e1 e8g7 f3g5 a8d8 b2b3 d6d2 c3d2 d8d2 g2g3 h7h6 g5e4 d2a2 e4c5 g8f8 h2h3 g7f5 g1f1 h6h5 e1d1 f5d6 d1d5 a2b2 d5e5 b2d2 f1g2 d2d1 g2f3');
INSERT INTO evaluations VALUES(19809,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(19809,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(19809,32,'cp -101',NULL);
INSERT INTO evaluations VALUES(19809,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(19809,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(19809,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(19809,36,'cp -74','e7e5 d2d3 g8f6 b1d2 f8e7 e2e4 d5d4 b2c1 c8e6 a2a4 a7a6 g2h3 e6h3 g1h3 d8c8 h3g5 b7b5 e1g1 e8g8 a4b5 a6b5 a1a8 c8a8 f2f4 e5f4 g3f4 h7h6 g5h3 a8a2 f4f5 f8a8 d2f3 f6g4 h3f4 c5c4 b3c4 b5c4 f4d5 e7d6 d3c4 d6c5 g1h1 a8a4 f5f6 a4c4 f3h4');
INSERT INTO evaluations VALUES(20264,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(20264,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20264,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(20264,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(20264,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20264,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20264,36,'cp 0','g8f6 b1d2 f8d6 d4c5 d6c5 g1f3 e8g8 e1g1 f8e8 e3e4 e6e5 f4g5 d5d4 c3d4 c5d4 f3d4 c6d4 g5e3 c8g4 f2f3 d4f5 d1e2 f5d4');
INSERT INTO evaluations VALUES(21374,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(21374,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(21374,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(21374,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(21374,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(21374,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(21374,36,'cp 46','g8f6 e2e3 f8d6 f4d6 d8d6 g1f3 e8g8 c4c5 d6e7 f1b5 c8d7 e1g1 a7a6 b5c6 d7c6 f3e5 f8b8 f1e1 c6e8 a1c1 f6d7 e5d3 d7f6 b2b4 e8b5 d3b2 f6e4 a2a4 e4c3 c1c3 b5c6 c3c1 e6e5 b2d3 e5e4 d3f4 e7d7');
INSERT INTO evaluations VALUES(21659,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(21659,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(21659,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(21659,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(21659,34,'cp 200',NULL);
INSERT INTO evaluations VALUES(21659,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(21659,36,'cp 193','c4d5 d8f6 d5c6 b7c6 d2d3 a7a5 g1e2 f8d6 c1g5 f6g6 h5g6 h7g6 a2a4 f7f6 g5e3 g6g5 b1d2 c8e6 e2c1 g8e7 c1b3 e7g6 b3c5 e6f7 g2g3 g6f8 f2f3 f8e6 c5e6 f7e6 e1e2 d6b4 h2h4 g5g4 e2f2 e8f7');
INSERT INTO evaluations VALUES(21553,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(21553,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(21553,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(21553,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(21553,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(21553,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(21553,36,'cp 180','d8f6 d5c6 b7c6 d2d3 a7a5 b1d2 f8d6 g1e2 c8e6 b2b3 g7g6 h5f3 f6f3 d2f3 a5a4 c1b2 f7f6 d3d4 e5d4 f3d4 e6d7 d4f3 d7g4 f3d2 a4b3 f2f3 g4e6 e2d4 e6d7 c2b3 g8e7 d2c4 d6b4 e1f2 c6c5 d4e2 e8f7 e4e5');
INSERT INTO evaluations VALUES(22781,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(22781,31,'cp -46',NULL);
INSERT INTO evaluations VALUES(22781,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(22781,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(22781,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(22781,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(22781,36,'cp -75','g8f6 f1d3 c6d4 f3g3 d7d6 g1e2 h7h5 e2d4 e5d4 c3d5 h5h4 d5f6 d8f6 g3f3 f6f3 g2f3 g7g6 d3c4 c7c6 d2d3 h8h5 h1g1 d6d5 c4b3 a7a5 a2a4 c8e6 c2c3 f8c5 c1d2 e8e7 b3d1 h5h8 e1e2 h4h3 c3d4 c5d4 d2c3 d4c3 b2c3 b7b5 e2e3 b5b4 c3b4 a5b4 a4a5');
INSERT INTO evaluations VALUES(27342,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(27342,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(27342,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(27342,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(27342,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(27342,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(27342,36,'cp 37','h7h6 g5h4 d7g4 h4g3 e8g8 h2h3 g4g5 a1d1 f8e8 f1e1 a7a6 a2a4 e7f8 e1e3 g5h5 b2b3 a8c8 d4d3 h5a5 d3e2 e8e6 c3d5 f6d5 d1d5 a5b4 e2d3 c8e8 d5d4 b4b6 g1f1 g7g6 d4c4 b6a5 f2f4 b7b5 a4b5 a6b5 c4c6');
INSERT INTO evaluations VALUES(27931,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(27931,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(27931,32,'cp 157',NULL);
INSERT INTO evaluations VALUES(27931,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(27931,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(27931,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(27931,36,'cp 150','e1g1 b4c3 b2c3 g8e7 a2a4 a7a5 f3e5 d7d6 e5g4 b8d7 d1e2 h6h5 g4e3 d7f6 f2f3 d8d7 h2h4 e8g8 c3c4 e7g6 e2e1 d7d8 c1b2 f8e8 e1f2 f6h7 g2g3 e6e5 c2c3 g8h8 g1h2 e8g8 d3c2 d8d7 f3f4 e5f4 g3f4 h7f6 f2g3 f6g4 e3g4 h5g4');
INSERT INTO evaluations VALUES(27930,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(27930,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(27930,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(27930,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(27930,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(27930,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(27930,36,'cp 140','b4c3 b2c3 g8e7 a2a4 a7a5 f3e5 d7d6 e5g4 b8d7 d1e2 d6d5 f2f3 e8g8 c3c4 d5c4 d3c4 c7c5 c2c3 d8c8 c1f4 e7g6 f4g3 b7a6 g4e3 a6c4 e2c4 c5d4 c3d4 c8c4 e3c4 g6e7 f1c1 f8c8 g1f2 c8c6 c4d6 f7f5 a1b1 d7f6 c1c6 e7c6 e4f5 e6f5 d6f5 f6d5 b1b5');
INSERT INTO evaluations VALUES(28937,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(28937,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(28937,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(28937,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(28937,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(28937,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(28937,36,'cp -9','g1f3 g7g6 f1d3 f5d3 d1d3 a7a5 e1g1 f8g7 c1d2 e8g8 d2c3 b7b6 c5b6 d8b6 b1d2 b8d7 a2a4 f8c8 f1c1 c6c5 b2b3 f6e8 c1c2 c5d4 c3d4 g7d4 f3d4 c8c2 d3c2 d7c5 a1c1 a8c8 h2h3 e8d6 c2c3 b6a6 c3a1 c5d3 c1c8 a6c8 d2f3 c8c1 a1c1 d3c1');
INSERT INTO evaluations VALUES(28938,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(28938,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(28938,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(28938,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(28938,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(28938,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(28938,36,'cp -23','g7g6 b2b4 f8g7 c1b2 a7a5 a2a3 e8g8 b1d2 b8a6 d1b3 b7b6 f1a6 a8a6 e1g1 f6d7 c5b6 d8b6 f1c1 f8a8 h2h3 a5b4 a3b4 a6a1 b2a1 b6b5 f3e5 d7e5 d4e5 e7e6 a1d4 g7f8 d4c5 f8c5 b4c5 b5e2 b3b2 f5d3 b2c3 a8a2');
INSERT INTO evaluations VALUES(29652,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(29652,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(29652,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(29652,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(29652,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(29652,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(29652,36,'cp -50','g1f3 e7e6 h2h3 g4f5 c2c4 c7c5 c4d5 e6d5 g2g4 f5b1 a1b1 c5d4 f1b5 b8c6 e1g1 a7a6 b5c6 b7c6 d1d4 f8d6 b2b3 e8g8 c1b2 f8e8 d4d3 d6c5 f3e5 d8b6 f1e1 a8d8 b1c1 f6e4 c1c2 f7f6 e5f3 b6b7');
INSERT INTO evaluations VALUES(34186,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(34186,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(34186,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(34186,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(34186,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(34186,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(34186,36,'cp 51','f8b4 d3d4 d7d6 e1g1 g8e7 a2a3 b4a5 b2b4 a5c7 a3a4 e8g8 a4a5 e5d4 d1d4 b8d7 c1e3 f8d8 d4f6 d7f6 f1e1 d6d5 e3c5 f6e4 c5e7 d8e8 c3e4 d5e4 e7c5 e4f3 e2f3 e8e1 a1e1 c8e6 a5a6 b7a6 f3c6 a8b8 e1a1 c7b6 h2h3 b6c5 b4c5 b8b2');
INSERT INTO evaluations VALUES(34262,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(34262,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(34262,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(34262,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(34262,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(34262,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(34262,36,'cp 174','c1d2 a5c7 d4d5 d7d6 c2c4 e7e5 f1e2 g8f6 b1c3 f8e7 g1f3 c7d8 h2h3 e8g8 d1c2 h7h6 e1g1 a7a6 g1h1 b8d7 a2a4 f6h7 a4a5 e7g5 f3g5 h6g5 c3d1 d7f6 d1e3 c8d7 b2b4 c5b4 d2b4');
INSERT INTO evaluations VALUES(34263,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(34263,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(34263,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(34263,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(34263,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(34263,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(34263,36,'cp 160','a5c7 d4d5 g8f6 b1c3 d7d6 f1c4 e7e6 g1e2 f8e7 e2f4 e6e5 f4e2 h7h5 e2g3 a7a6 a2a4 h5h4 g3f1 f6h5 g2g3 b8d7 f1e3 g7g6 c4d3 c7d8 d3e2 h5f6 h2h3 f6h7 e2g4 e7g5 d1e2 e8f8 g3h4 g5h4 e2f3 d8f6 e3c4 f6f3 g4f3 h7g5 d2g5 h4g5 c4d6');
INSERT INTO evaluations VALUES(34946,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(34946,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(34946,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(34946,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(34946,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(34946,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(34946,36,'cp 41','g1f3 d5e6 b1c3 g8f6 d2d4 f8d6 d4e5 d6e5 d1b3 e6b3 a2b3 e5d6 c3b5 b8c6 c1d2 f6e4 b5d6 c7d6 d2c3 e8g8 f3d4 c8e6 b3b4 e4c3 b2c3 a8c8 e1d2 a7a6 f1e2 c8c7 a1a3 c6d4 e3d4 g7g6 h2h4 e6c4 h1e1 c4e2 e1e2 c7c4');
INSERT INTO evaluations VALUES(38639,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(38639,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(38639,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(38639,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(38639,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(38639,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(38639,36,'cp -39','c2c4 c7c6 g1f3 b8d7 d2d3 g8f6 g2g3 e8g8 c4c5 d6c7 f1g2 b7b6 c5b6 a7b6 e1g1 f8e8 b1d2 c7d6 f1e1 d6f8 d1c2 c8b7 e2e4 d5d4 e1b1 b6b5 b2c1 c6c5 b4c5 d7c5 b1b5 f6d7 d2b3 a8c8 c1g5 f7f6 b3c5 d7c5 b5b7 c5b7 c2b3 g8h8');
INSERT INTO evaluations VALUES(40788,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(40788,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40788,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(40788,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(40788,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(40788,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(40788,36,'cp 30','c7c5 e1g1 b8c6 d2d3 h7h6 b2b3 c8e6 e2e4 a7a6 f3e1 e8g8 e1c2 e6g4 f2f3 g4d7 f3f4 b7b5 c2e3 c6d4 c1b2 a8b8 f1f2 f8e8 f4f5 a6a5 h2h3 d7c6 g1h2 b5c4 e3c4 f6d7 c4e3 e7f6 c3d5 f6g5 b2c3');
INSERT INTO evaluations VALUES(42360,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(42360,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(42360,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(42360,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(42360,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(42360,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(42360,36,'cp 31','f3g3 b8c6 b1c3 f6h5 g3f3 h5f6 g1e2 d7d6 h2h3 c6d4 e2d4 c5d4 c3e2 d4c5 e1g1 c5b6 c1e3 e8g8 a1d1 c8e6 e3b6 a7b6 c4e6 f7e6 d3d4 e5d4 e2d4 d8d7 f3b3 f6e4 d4e6 f8f7 f2f3 e4g3 f1e1 d7a4 b3a4 a8a4 a2a3');
INSERT INTO evaluations VALUES(42479,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42479,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(42479,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(42479,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(42479,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42479,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42479,36,'cp 0','e8g8 f1g2 d7d5 g1f3 d5c4 a2a3 b4c3 b2c3 c7c5 e1g1 c5d4 d1d4 d8d4 f3d4 f6d5 d4b5 b8a6 a3a4 c8d7 c1a3 f8d8 b5d6 a8b8 f1c1 d7a4 d6c4 a4b5 c4d6 b5c6 d6c4');
INSERT INTO evaluations VALUES(43703,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(43703,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43703,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43703,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43703,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43703,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43703,36,'cp 0','c8b7 e3e4 c7c5 d4c5 f8c5 e4e5 f6d5 d3e4 d5c3 d1d8 e8d8 e4b7 a8a7 b2c3 a7b7 a2a4 d8e7 a4b5 a6b5 e1e2 h8c8 a1a8 c8d8 c1g5 f7f6 g5d2 b8c6 a8a6 c6b8 a6a1 b8c6');
INSERT INTO evaluations VALUES(28525,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(28525,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28525,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28525,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28525,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(28525,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(28525,36,'cp -6','e3e4 c7c5 d4c5 f8c5 e4e5 f6d5 d3e4 c5b4 e1g1 b4c3 b2c3 b8d7 f3g5 d7e5 d1d4 d8f6 e4d5 b7d5 f1e1 a8d8 d4e5 f6e5 e1e5 f7f6 g5e6 f6e5 e6d8 e8d8 a2a4 d8e7 a4b5 a6b5 f2f3 h8a8 a1a8 d5a8 g1f2 e7d6 c1a3 d6d5 a3f8 g7g6 f2e3 d5c4 f8b4 a8c6 b4a5 g6g5 g2g3 c4b3 f3f4 g5f4 g3f4 e5f4 e3f4 b3c4 f4g5 h7h6 g5h6');
INSERT INTO evaluations VALUES(44221,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(44221,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(44221,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44221,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(44221,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(44221,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44221,36,'cp 15','e2e3 d7d5 g1f3 d8b6 d1c2 f6h5 f4e5 b8c6 h2h3 c8d7 e5h2 a8c8 c2b3 f8e7 b3b6 a7b6 b1d2 e8g8 f1b5 h5f6 a2a4 f8d8 e1e2 f6e8 f3e5 c6e5 d4e5 e8c7 b5d7 d8d7 h1b1 h7h6 g2g4 c7a6 h2g3 c5c4 h3h4');
INSERT INTO evaluations VALUES(46384,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(46384,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(46384,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(46384,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(46384,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(46384,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(46384,36,'cp -96','b8c6 g1f3 f8d6 e1g1 e8g8 c1g5 c8e6 b5a4 h7h6 g5f6 d8f6 a4b3 c6e7 d1e2 a8e8 f1e1 d6b8 a2a4 a7a6 a4a5 b8c7 g1f1 f6f5 f3e5 e7c6 e5c6 b7c6 f1g1 e6d7 e2f3 f5g6 d3d4 e8e1 a1e1 c7a5 e1a1 a5c7 a1e1 a6a5 g2g3 d7g4 f3e3 g4h3');
INSERT INTO evaluations VALUES(13694,30,'cp -99',NULL);
INSERT INTO evaluations VALUES(13694,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(13694,32,'cp -111',NULL);
INSERT INTO evaluations VALUES(13694,33,'cp -116',NULL);
INSERT INTO evaluations VALUES(13694,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(13694,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(13694,36,'cp -94','g1f3 f8d6 e1g1 e8g8 c1g5 c8e6 b5a4 h7h6 g5f6 d8f6 a4b3 c6e7 f1e1 a8d8 d1e2 f8e8 c3b5 d6b8 f3e5 e7f5 c2c3 d5d4 b5d4 e6b3 e5g4 e8e2 g4f6 g7f6 d4e2 b3d5 b2b3 f5h4 f2f3 d5f3 g2f3 h4f3 g1f2 f3e1 a1e1 d8d3 e2g3 b8g3 h2g3 d3c3 e1e7 g8g7 e7b7 c3c2 f2e3 c2a2 b3b4 a2a3 e3d4');
INSERT INTO evaluations VALUES(47246,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(47246,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(47246,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(47246,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(47246,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(47246,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(47246,36,'cp -21','g7g6 f1g2 f8g7 e1g1 c7c5 d2d3 e8g8 f3e5 f6g4 e5g4 c8g4 b1c3 d5d4 c3e4 d8b6 a1b1 b8a6 c1d2 c5c4 h2h3 g4e6 e4g5 c4d3 e2d3 e6d7 g3g4 a8c8 f4f5 h7h6 g5e4 a6c7 a2a4 d7a4 g4g5 c7d5 g5h6 a4c2 h6g7 c2d1 g7f8q c8f8 b1d1');
INSERT INTO evaluations VALUES(48868,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(48868,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(48868,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(48868,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(48868,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(48868,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(48868,36,'cp -51','c7c6 g1f3 f8d6 e1g1 e8g8 b1d2 f8e8 e2e4 h7h6 d1c2 b8d7 f1e1 a7a5 a2a4 d8c7 b2b3 d5e4 d3e4 b7b5 g2f1 c7b6 c1a3 d6b8 f1h3 c8a6 d2f1 b8a7 h3d7 f6d7 f1e3 b5b4 e1d1 d7c5 c3b4 a5b4 a3b2 f7f6 e3f5 b6c7');
INSERT INTO evaluations VALUES(50781,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50781,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(50781,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(50781,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(50781,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(50781,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(50781,36,'cp 24','f8d6 f4g3 g8f6 b1d2 d8b6 d1b3 b6c7 d4c5 c7c5 f1d3 e8g8 e1g1 d6g3 h2g3 h7h6 c3c4 b8c6 c4d5 c5d5 f1d1 d5b3 d2b3 b7b6 b3d2 f8d8 d3e2 a6a5 a1c1 c8b7 d2c4 b7a6 d1d8 a8d8 c4a5 b6a5 e2a6 c6b4 a6e2 b4a2');
INSERT INTO evaluations VALUES(51633,30,'cp -148',NULL);
INSERT INTO evaluations VALUES(51633,31,'cp -153',NULL);
INSERT INTO evaluations VALUES(51633,32,'cp -165',NULL);
INSERT INTO evaluations VALUES(51633,33,'cp -152',NULL);
INSERT INTO evaluations VALUES(51633,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(51633,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(51633,36,'cp -151','a7a5 d1a4 e7e5 e2e3 e5d4 e3d4 g8f6 g1f3 f8e7 f1e2 e8g8 a2a3 b6b5 a4c2 a5b4 e1g1 f6e4 b1d2 c8f5 e2d3 b4a3 a1a3 a8a3 c1a3 e4d2 c2d2 f5d3 d2d3 b8a6 f3e5 d8c7 e5f3 c7c8 f1e1 e7f6 a3c1 f8e8 e1e8 c8e8 c1f4 e8e4 d3d2 e4b1 f3e1');
INSERT INTO evaluations VALUES(51252,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(51252,31,'cp -166',NULL);
INSERT INTO evaluations VALUES(51252,32,'cp -163',NULL);
INSERT INTO evaluations VALUES(51252,33,'cp -165',NULL);
INSERT INTO evaluations VALUES(51252,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(51252,35,'cp -159',NULL);
INSERT INTO evaluations VALUES(51252,36,'cp -155','d1a4 e7e5 e2e3 e5d4 e3d4 f8e7 f1e2 g8f6 g1f3 e8g8 a2a3 c8f5 c5b6 d8b6 e1g1 b8d7 b4b5 c6b5 a4b5 b6c7 b5b2 f8c8 c1d2 d7b6 b1c3 b6c4 e2c4 c7c4 b2b5 c4b5 c3b5 f6e4 a3a4 c8c2 d2e3 a8c8 f1c1 e7b4 f3e5 e4c3 c1c2 f5c2 b5c3 b4c3');
INSERT INTO evaluations VALUES(52630,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(52630,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(52630,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(52630,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(52630,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(52630,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(52630,36,'cp 62','e7e6 e2e4 f8g7 g1f3 e6d5 c4d5 g8f6 f3d2 b8d7 f1e2 e8g8 e1g1 f8e8 a2a4 d7e5 d2b1 a7a6 h2h3 a8b8 f2f4 e5d7 b1d2 c5c4 e2c4 d7c5 e4e5 d6e5 f4e5 e8e5 d2f3 e5f5 g2g4 b7b5 a4b5 a6b5 g4f5 b5c4 c1e3 f6e4');
INSERT INTO evaluations VALUES(53931,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(53931,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(53931,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(53931,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(53931,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(53931,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(53931,36,'cp 124','f8e7 a2a4 g8f6 b1c3 e8g8 f3d2 b8a6 f1a6 b7a6 d2c4 f6d7 d1d3 d7b6 c4a5 c5c4 d3f3 d8e8 a5c6 c8d7 c6e7 e8e7 a4a5 b6c8 f3e2 d7b5 f2f4 f7f5 e1g1 f5e4 c1e3 e5f4 f1f4 f8f4 e3f4 e7d7 e2e4 c8e7 e4e6 d7e6 d5e6 d6d5 a1d1 a8d8');
INSERT INTO evaluations VALUES(55007,30,'cp -156',NULL);
INSERT INTO evaluations VALUES(55007,31,'cp -134',NULL);
INSERT INTO evaluations VALUES(55007,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(55007,33,'cp -130',NULL);
INSERT INTO evaluations VALUES(55007,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(55007,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(55007,36,'cp -129','c7c6 b5a4 g8f6 g1e2 c8g4 c2c3 b8d7 a4c2 e7e5 b1d2 e5d4 c3d4 f8d6 c2e4 g4e2 e1e2 e8g8 e4f3 f8e8 e2f1 d8b6 d2c4 b6a6 f3e2 c6c5 c4e5 a6b6 e5d7 f6d7 c1e3 b6b2 a1c1 a8d8 d1b3 b2b3 a2b3 c5d4 e3d4 d6a3 c1c2 d8c8 c2c8 e8c8 g2g3 a3c5 d4b2 c5b4');
INSERT INTO evaluations VALUES(52695,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(52695,31,'cp -120',NULL);
INSERT INTO evaluations VALUES(52695,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(52695,33,'cp -115',NULL);
INSERT INTO evaluations VALUES(52695,34,'cp -110',NULL);
INSERT INTO evaluations VALUES(52695,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(52695,36,'cp -104','b5a4 g8f6 g1e2 c8g4 c2c3 b8d7 b1d2 e7e5 a4c2 e5d4 c3d4 f8d6 c2e4 g4e2 e1e2 f6e4 d2e4 e8g8 h1e1 d6b4 c1d2 b4d2 e4d2 d8b6 e2f1 b6b2 d2c4 b2b5 d1b3 b5a6 f1g1 c6c5 d4d5 b7b5 c4e3 c5c4 b3b4 d7e5 e1b1 f8b8 a2a4 e5d3 b4c3 d3f4 c3e5 f4e2 g1f1');
INSERT INTO evaluations VALUES(55376,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(55376,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(55376,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(55376,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(55376,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(55376,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(55376,36,'cp 89','c1f4 f8g7 b1c3 g8f6 e2e3 e8g8 a1c1 a7a6 f1e2 d5c4 e2c4 b7b6 e1g1 c8b7 c4b3 h7h6 h2h3 b8d7 d1e2 c7c5 d4c5 d7c5 f1d1 f6d5 b3d5 e6d5 f4e5 g7e5 f3e5 d8e7 e5f3 f8d8 e2d2 e7f6 f3d4 a8c8 f2f3 g8h7 b2b4 c5e6 d4e6 f6e6 c3e2 e6e7');
INSERT INTO evaluations VALUES(55801,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(55801,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(55801,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(55801,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(55801,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(55801,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(55801,36,'cp -79','d5d4 c3a4 c7c5 g2g3 b8c6 a2a3 a8b8 f1g2 f8d6 e1g1 g8e7 b2b4 c5b4 c4c5 d6c7 a3b4 c6b4 e2e3 e6e5 e3d4 e5d4 c1b2 b4c6 f1e1 c8g4 h2h3 g4h5 b2a3 e8g8 a4b2 d8d5 b2d3 a7a5 g3g4 h5g6 f3e5 d5d8 f2f4 g6d3');
INSERT INTO evaluations VALUES(56770,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(56770,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(56770,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(56770,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(56770,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(56770,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(56770,36,'cp 66','c1f4 g8f6 e2e3 a7a6 f1d3 c7c5 d4c5 f8c5 e1g1 b8c6 c3e2 d8e7 e2d4 e8g8 c2c4 c8d7 a1c1 c6b4 d3b1 d5c4 c1c4 b4d5 f4e5 b7b5 c4c1 a8c8 a2a3 c5d6 b1a2 c8c1 d1c1 d6e5 f3e5 e7d6 e5d3 f8c8 c1a1 e6e5 d4f3 e5e4');
INSERT INTO evaluations VALUES(56896,30,'cp 840',NULL);
INSERT INTO evaluations VALUES(56896,31,'cp 839',NULL);
INSERT INTO evaluations VALUES(56896,32,'cp 840',NULL);
INSERT INTO evaluations VALUES(56896,33,'cp 844',NULL);
INSERT INTO evaluations VALUES(56896,34,'cp 848',NULL);
INSERT INTO evaluations VALUES(56896,35,'cp 844',NULL);
INSERT INTO evaluations VALUES(56896,36,'cp 836','h5e5 d8e7 e5h8 g8f6 b2b3 d7d6 c1b2 b8d7 g1f3 a7a6 b1c3 b7b5 f3g5 b5b4 c3e2 h7h6 g5h7 f6h7 h8h7 d7e5 e2f4 c8b7 d2d4 e5d7 f1c4 d7f6 h7h8 d6d5 c4e2 f6e4 e1g1 e7f6 h8f6 e4f6 f2f3 a6a5 f4d3 a5a4 b3a4 a8a4 d3c5 a4a7 g1f2 f6d7 c5b7 a7b7 a1d1 f8e7 c2c3 e8f8 c3b4 b7b4');
INSERT INTO evaluations VALUES(56894,30,'cp 841',NULL);
INSERT INTO evaluations VALUES(56894,31,'cp 840',NULL);
INSERT INTO evaluations VALUES(56894,32,'cp 832',NULL);
INSERT INTO evaluations VALUES(56894,33,'cp 835',NULL);
INSERT INTO evaluations VALUES(56894,34,'cp 840',NULL);
INSERT INTO evaluations VALUES(56894,35,'cp 832',NULL);
INSERT INTO evaluations VALUES(56894,36,'cp 835','g8e7 e5h8 e7c6 g1f3 d7d5 b1c3 c8g4 a2a3 g4f3 g2f3 d5d4 c3e4 b8d7 h8h7 d8e7 f1c4 d7e5 c4e2 e8c8 d2d3 d4e3 f2e3 e5d7 e1g1 c8b8 d3d4 d7f6 h7h4 f8g7 e4f6 g7f6 h4e4 e7f8 f3f4 f8g7 c2c3 c6e7 a3a4 e7f5 b2b4 c7c6 a4a5');
INSERT INTO evaluations VALUES(50046,30,'cp 844',NULL);
INSERT INTO evaluations VALUES(50046,31,'cp 848',NULL);
INSERT INTO evaluations VALUES(50046,32,'cp 848',NULL);
INSERT INTO evaluations VALUES(50046,33,'cp 848',NULL);
INSERT INTO evaluations VALUES(50046,34,'cp 850',NULL);
INSERT INTO evaluations VALUES(50046,35,'cp 848',NULL);
INSERT INTO evaluations VALUES(50046,36,'cp 848','e5h8 d7d5 g1f3 h7h6 b2b3 c8e6 f3d4 e7f5 d4e6 f7e6 h8h7 g6g5 h7g6 e8d7 c2c4 d8e8 g6e8 d7e8 c4d5 e6d5 b1c3 c7c6 g2g4 f5d6 h2h4 b8d7 c1b2 a8c8 d2d4 c8c7 f1e2 f8e7 h4h5 d7f8 e1c1 f8e6 a2a4');
INSERT INTO evaluations VALUES(960,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(960,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(960,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(960,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(960,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(960,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(960,36,'cp 0','d5c6 b7c6 f3f6 g7f6 d2d3 f6f5 b1c3 f7f6 f2f4 c8e6 c1e3 f8d6 e1g1 h7h5 a1e1 h5h4 h2h3 a7a5 b2b3 d6b4 e3d2 e5e4 f1f2 b4c5 d2e3 c5b4');
INSERT INTO evaluations VALUES(945,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(945,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(945,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(945,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(945,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(945,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(945,36,'cp -8','b7c6 f3f6 g7f6 d2d3 f6f5 b1c3 h7h5 f2f4 f7f6 c1e3 h5h4 e1g1 h4h3 g2g3 c8e6 c3a4 e8f7 a4c5 e6c8 b2b3 f8d6 a1e1 a7a5 c5a4 c8e6 f1f2 a8e8 f4e5 f6e5 a4c5 e6c8 c5e4 h8g8 g1f1 f7g6 e4d6 c7d6 e3b6 a5a4 b3a4 g8g7');
INSERT INTO evaluations VALUES(1898,30,'cp -435',NULL);
INSERT INTO evaluations VALUES(1898,31,'cp -424',NULL);
INSERT INTO evaluations VALUES(1898,32,'cp -466',NULL);
INSERT INTO evaluations VALUES(1898,33,'cp -444',NULL);
INSERT INTO evaluations VALUES(1898,34,'cp -448',NULL);
INSERT INTO evaluations VALUES(1898,35,'cp -474',NULL);
INSERT INTO evaluations VALUES(1898,36,'cp -464','b4f4 e2e3 f4h6 b1d2 f7f6 e5e6 g8e7 d1b3 d7e6 e1c1 e7d5 h2h4 f8e7 g2g4 c8d7 b3b7 a8b8 b7a6 e8g8 a6d3 g8h8 f1e2 d7e8 d3c2 a7a5 d1g1 b8d8 h4h5 d5b6 d2c4 b6d7 c4d2 d7c5 d2b3 c5b3 a2b3 c6e5 f3e5 f6e5 g4g5');
INSERT INTO evaluations VALUES(1908,30,'cp -454',NULL);
INSERT INTO evaluations VALUES(1908,31,'cp -449',NULL);
INSERT INTO evaluations VALUES(1908,32,'cp -448',NULL);
INSERT INTO evaluations VALUES(1908,33,'cp -455',NULL);
INSERT INTO evaluations VALUES(1908,34,'cp -477',NULL);
INSERT INTO evaluations VALUES(1908,35,'cp -437',NULL);
INSERT INTO evaluations VALUES(1908,36,'cp -472','e2e3 f4f5 f1d3 f5h5 b1d2 c6e5 f3e5 h5e5 d2f3 e5e7 d3c2 d7d6 e3e4 c8g4 e1g1 g4f3 d1f3 c7c6 f3g3 g7g6 a1d1 f8g7 f2f4 g8h6 f4f5 e8g8 g3d6 e7d6 d1d6 h6g4 h2h3 g4e3 c2d3 g8h8 d6d7');
INSERT INTO evaluations VALUES(4200,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(4200,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(4200,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(4200,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(4200,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(4200,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4200,36,'cp 45','b1c3 d7d6 g2g4 f6h7 g4g5 h7g5 f3g5 h6g5 c3d5 c6d4 h1g1 d4e6 h3h4 c7c6 d5e3 e6f4 h4g5 c8e6 c4e6 f7e6 d1g4 d6d5 c1d2 d8b6 e3d1 a8e8 g1h1 d5e4 d3e4 c5d4 a1b1 b6a6 d1e3 f8f7 g5g6 f7f6');
INSERT INTO evaluations VALUES(4579,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(4579,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4579,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4579,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4579,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(4579,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4579,36,'cp 0','e7b4 e1g1 b4c3 b2c3 h7h6 c4b5 d7d6 c3c4 c6e7 c4c5 e7g6 c5d6 c7d6 b5a4 c8e6 c2c4 f6g4 e3d2 f7f5 e4f5 e6f5 h2h3 g4f6 f1e1 g8h8 a4c2 d8d7 c2a4');
INSERT INTO evaluations VALUES(4973,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(4973,31,'cp 386',NULL);
INSERT INTO evaluations VALUES(4973,32,'cp 386',NULL);
INSERT INTO evaluations VALUES(4973,33,'cp 413',NULL);
INSERT INTO evaluations VALUES(4973,34,'cp 406',NULL);
INSERT INTO evaluations VALUES(4973,35,'cp 451',NULL);
INSERT INTO evaluations VALUES(4973,36,'cp 444','d4d5 d8f6 d5a8 f6f2 e1d1 g8e7 h1f1 f2d4 b1d2 e8g8 c2c3 d4e5 d2f3 e5h5 a8a7 d7d5 e4d5 h5g4 a7a4 e7d5 c1d2 c7c6 f1e1 b6b5 c4b5 c6b5 a4g4 c8g4 a2a4 f8d8 d1c2 b5a4 a1a4 g4f5 c2c1 h7h5 c3c4 d5b6 a4a5 d8c8 b2b3 f5e6 f3g5 c5f2 e1e6 f7e6 c1c2 b6c4 b3c4 e6e5 a5b5 f2d4 c2b3 h5h4');
INSERT INTO evaluations VALUES(4975,30,'cp 428',NULL);
INSERT INTO evaluations VALUES(4975,31,'cp 443',NULL);
INSERT INTO evaluations VALUES(4975,32,'cp 430',NULL);
INSERT INTO evaluations VALUES(4975,33,'cp 459',NULL);
INSERT INTO evaluations VALUES(4975,34,'cp 440',NULL);
INSERT INTO evaluations VALUES(4975,35,'cp 438',NULL);
INSERT INTO evaluations VALUES(4975,36,'cp 418','d8f6 d5a8 f6f2 e1d1 g8e7 h1f1 f2d4 b1d2 e8g8 c2c3 d4e5 d2f3 e5h5 a8a7 d7d5 e4d5 h5g4 a7a4 e7d5 c1d2 c7c6 f1e1 c8e6 d1c1 g4g2 f3g5 c5e3 d2e3 d5e3 g5e6 b6b5 e6f4 g2f2 f4d3 f2e1 d3e1 b5a4 c4e2 e3d5 e2f3 f8d8 f3d5 d8d5 c1c2 d5f5 c3c4 f5f2 c2c3 f2h2');
INSERT INTO evaluations VALUES(5395,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(5395,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(5395,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(5395,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(5395,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(5395,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(5395,36,'cp 111','d7d6 d4d5 c6e5 f3e5 g7e5 f1d3 g8f6 e1g1 c8g4 c3e2 g4e2 d1e2 d8c8 h2h3 f6h5 e2f3 e8g8 a1b1 e7e6 d5e6 f7e6 f3g4 h5f4 f1d1 d6d5 b2b3 f4d3 d1d3 e5g7 c1b2 g7b2 b1b2 d5c4 b3c4 c8c6 g4e2 c6c4 b2b7');
INSERT INTO evaluations VALUES(5534,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(5534,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(5534,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(5534,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(5534,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(5534,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(5534,36,'cp 110','d4d5 c6e5 f3e5 g7e5 f1d3 g8f6 e1g1 c8g4 c3e2 g4e2 d1e2 d8c8 h2h3 f6h5 e2f3 e8g8 a1b1 c8c7 c1h6 e5g7 h6e3 h5f6 f1e1 f6d7 b2b3 a7a6 a2a4 b7b6 f3e2 a8e8 d3e4 d7c5 e4f3 e7e5 d5e6 e8e6');
INSERT INTO evaluations VALUES(6477,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(6477,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(6477,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(6477,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(6477,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(6477,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(6477,36,'cp -53','g8f6 f1e2 e8g8 e1g1 f8e8 f1e1 d7d6 d2d3 a7a6 c1b2 c8e6 d1d2 c6d4 f3d4 e5d4 c3d1 d6d5 e4e5 f6d7 f2f4 f7f6 e2h5 g7g6 h5f3 f6e5 f4e5 d8h4 b3b4 c5a7 d2f2 h4f2 d1f2 c7c6 a2a4 e6f5 g2g4 d7e5 e1e5 e8e5 g4f5');
INSERT INTO evaluations VALUES(7123,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(7123,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(7123,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(7123,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(7123,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(7123,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(7123,36,'cp 30','c2c3 e8g8 h2h3 f8e8 c3c4 f6d7 c1e3 e7f6 b1c3 c6d4 e3d4 e5d4 c3e2 c7c5 e2f4 a8b8 a4d7 c8d7 a2a4 b7b5 a4b5 a6b5 b2b3 d7c6 a1a6 c6b7 a6a7 d8b6 a7a1 b8a8 d1c2 f6e7 a1a8 e8a8 f4d5');
INSERT INTO evaluations VALUES(10221,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(10221,31,'cp -198',NULL);
INSERT INTO evaluations VALUES(10221,32,'cp -171',NULL);
INSERT INTO evaluations VALUES(10221,33,'cp -194',NULL);
INSERT INTO evaluations VALUES(10221,34,'cp -201',NULL);
INSERT INTO evaluations VALUES(10221,35,'cp -185',NULL);
INSERT INTO evaluations VALUES(10221,36,'cp -202','f6g4 c1e3 c5e3 f2e3 g4e3 d1d2 c6b4 a1c1 e3f1 h1f1 d7d6 d3d4 b4c6 d4e5 c8h3 f1f2 c6e5 f3e5 d6e5 d2d5 d8d5 c3d5 a8c8 c1d1 h7h5 f2d2 h8h6 d5e3 h6a6 a2a3 a6a4 c2c4 h3e6 e3f5 g7g6 f5e3 e8e7 e3d5 e6d5 d2d5');
INSERT INTO evaluations VALUES(10774,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(10774,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(10774,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(10774,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(10774,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(10774,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(10774,36,'cp 98','f3e5 d8e7 c4f7 e8f8 e5d3 f8f7 e4e5 d7d6 e5f6 e7e2 f6g7 h8g8 d3c5 d6c5 d1e2 d4e2 g1h1 g8g7 c1e3 c8d7 b1d2 a8g8 g2g3 g8e8 h2h4 e2d4 e3d4 c5d4 f1e1 d7f5 c2c3 d4c3 b2c3 g7g6 e1e8 f7e8 d2b3 g6a6 f2f3 a6a4 b3c5 a4c4 a1e1 e8f8');
INSERT INTO evaluations VALUES(10761,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(10761,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(10761,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(10761,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(10761,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(10761,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(10761,36,'cp 81','d8e7 c4f7 e8f8 e5d3 f8f7 e4e5 d7d6 e5f6 e7f6 d3c5 d6c5 c2c3 d4f5 d1d5 f6e6 d5c5 h8e8 c5c7 f7g8 b1a3 c8d7 c1d2 e6d5 c7c4 d5c4 a3c4 d7b5 b2b3 b5c4 b3c4 f5d6 f1d1 d6c4 d2f4 b7b5 g1f1 a7a6 d1d4 c4b6 d4d6 b6c8 d6c6 c8e7 c6c5 a8c8 f4e3 c8c5 e3c5 e7c6 c5e3 e8c8 a1e1');
INSERT INTO evaluations VALUES(12262,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(12262,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(12262,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12262,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(12262,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(12262,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(12262,36,'cp 35','d2d4 c5d4 d1d4 e7e5 d4d3 h7h6 a2a4 a7a6 b5c4 d7c5 d3e2 d8c7 e1g1 f8e7 b2b3 c8e6 f3d2 e8g8 a4a5 a8c8 c1b2 e6g4 f2f3 g4h5 f1d1 c5e6 g1h1 e6f4 e2e3 f8d8 d2f1 h5g6 d1d2 d6d5 c3d5 f4d5 e4d5');
INSERT INTO evaluations VALUES(13877,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(13877,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(13877,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(13877,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(13877,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(13877,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(13877,36,'cp 125','g5f6 g7f6 d4f5 d6d5 d1d5 c8f5 e4f5 f8b4 e1c1 b4c3 b2c3 d8d5 d1d5 a8d8 f1c4 h8g8 g2g3 e8e7 h1e1 b7b6 d5d8 g8d8 c4d3 e7f8 e1e4 c6a5 e4a4 f8g7 c1d2 d8d7 d2e2 d7c7 c3c4 c7c8 e2d2 c8c7 d2e3 h7h5 e3d2 c7c8 h2h4 g7f8');
INSERT INTO evaluations VALUES(14132,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(14132,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(14132,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(14132,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(14132,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(14132,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(14132,36,'cp 45','b1d2 e7e6 f1d3 f8d6 g1f3 f6d7 d1e2 d8e7 f4d6 c7d6 e3e4 d5e4 d2e4 e8g8 e1g1 e6e5 e4g3 d7f6 a1e1 e7d8 d4e5 c6e5 f3d4 f8e8 d3c2 c8d7 c2b3 d6d5 e2d2 e5c4 e1e8 d8e8 b3c4 d5c4 h2h3 b7b5 d4f5');
INSERT INTO evaluations VALUES(14551,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(14551,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(14551,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(14551,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(14551,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(14551,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(14551,36,'cp 55','b1d2 c6b4 d3b5 c7c6 b5e2 c6c5 e1g1 b4c6 c2c3 f6h5 f4e5 f8e7 d4c5 e7c5 a1c1 h5f6 e5g3 c5d6 g3h4 e8g8 c3c4 d6e7 d1c2 c8d7 a2a3 a8c8 f1d1 d5c4 d2c4 f6d5 h4g3 c6a5 f3e5 b7b5 c4a5 c8c2 c1c2 d8a5 e5d7');
INSERT INTO evaluations VALUES(14552,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(14552,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(14552,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14552,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14552,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(14552,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(14552,36,'cp 46','c6b4 d3e2 c7c5 e1g1 b4c6 c2c3 f8d6 d4c5 d6c5 a1c1 e8g8 c3c4 c5b6 a2a3 d8e7 f4g3 e6e5 c4d5 f6d5 d2c4 f8d8 d1c2 e5e4 f3e5 c6e5 g3e5 c8f5 c2b3 a8c8 c4b6 a7b6 e5d4 d5f4 e3f4 d8d4 c1c8 f5c8 f1c1 c8d7 b3b6 d4d2');
INSERT INTO evaluations VALUES(14602,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(14602,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(14602,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14602,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14602,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(14602,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(14602,36,'cp 67','c6e7 a2a4 g7g6 h2h3 f8g7 c1e3 c8d7 f3h2 a7a6 a4a5 f6h5 d1d2 f7f5 f2f3 f5e4 c3e4 d7b5 c4b5 a6b5 h2g4 e7f5 e1g1 h5f4 e3f4 e5f4 d2f4 h6h5 f4d2 e8g8 g4f2 g7b2 a1e1 b2d4 c2c3 d4g7 e4g5 g7h6 f2e4 a8a5 g2g4 h6g5');
INSERT INTO evaluations VALUES(14040,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(14040,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(14040,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(14040,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(14040,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(14040,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(14040,36,'cp 56','h2h3 g7g6 a2a4 f8g7 f3h2 c8d7 a4a5 a7a6 c1e3 f6h5 d1d2 f7f5 g2g4 f5g4 h3g4 h5f4 f2f3 g6g5 d2f2 f4h3 f2f1 h3f4 e1d1 c7c5 f1f2 e8g8 h2f1 f4g6 f2g3 g6h4 f1d2 e7g6 c4e2 g6f4 e2f1 f8f7 e3f2 d7b5 g3g1');
INSERT INTO evaluations VALUES(15689,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15689,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(15689,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(15689,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(15689,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(15689,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(15689,36,'cp 82','c2c4 f6h5 f4g5 f8e7 g5e7 d8e7 b1c3 h5f6 f1e2 e8g8 e1g1 e6e5 h2h3 a7a5 d1c2 e5e4 f3d2 c8f5 a2a3 f8e8 f1e1 c6b8 b2b4 b8d7 d2b3 a5b4 a3b4 b7b6 a1a8 e8a8 e1b1 h7h5 b3d2 c7c6 c2b2');
INSERT INTO evaluations VALUES(17203,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(17203,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(17203,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(17203,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(17203,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(17203,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(17203,36,'cp 65','e7c6 d4e3 d7d6 b1c3 c8e6 c3d5 c6e5 c4e2 c7c6 d5f4 e6c4 e2c4 e5c4 e3c3 d8a5 c3a5 c4a5 b2b3 e8c8 e1e2 b7b6 c1b2 a5b7 h2h4 h8g8 f2f3 b7c5 a1d1 c5e6 f4e6 f7e6 h4h5 f8e7 f3f4 e7f6 b2c1 e6e5 f4f5 d6d5 e2f3 d8d7 g2g4 h7h6 c2c4 d5d4 b3b4 g8e8 f3f2');
INSERT INTO evaluations VALUES(16859,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(16859,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(16859,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(16859,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(16859,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(16859,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(16859,36,'cp 64','d4e3 d7d6 b1c3 c8e6 c3d5 c6e5 c4e2 c7c6 d5f4 e6c4 e2c4 e5c4 e3c3 d8a5 c3a5 c4a5 b2b3 e8c8 e1e2 b7b6 c1b2 d8e8 e2f3 a5b7 b2d4 b7d8 g2g3 h8g8 h2h4 g7g6 a1d1 h7h5 d4e3 d8e6 c2c4 f7f6 f4e6 e8e6 a2a4');
INSERT INTO evaluations VALUES(17890,30,'cp -567',NULL);
INSERT INTO evaluations VALUES(17890,31,'cp -575',NULL);
INSERT INTO evaluations VALUES(17890,32,'cp -561',NULL);
INSERT INTO evaluations VALUES(17890,33,'cp -578',NULL);
INSERT INTO evaluations VALUES(17890,34,'cp -558',NULL);
INSERT INTO evaluations VALUES(17890,35,'cp -559',NULL);
INSERT INTO evaluations VALUES(17890,36,'cp -565','d8g5 e1g1 g5h4 d2d3 f8c5 b1d2 d7d6 c2c3 c8g4 d2f3 h4h5 h2h3 g4f3 d1f3 h5f3 g2f3 g8e7 b2b4 c5b6 a2a4 a6a5 b4b5 c6b8 f3f4 b8d7 c4b3 f7f5 f4e5 f5e4 d3d4 d6e5 c1g5 e5d4 c3d4 b6d4 a1d1 d7e5 g5e7 e5f3 g1g2 e8e7 b3d5 h8f8 d5e4');
INSERT INTO evaluations VALUES(1241,30,'cp -548',NULL);
INSERT INTO evaluations VALUES(1241,31,'cp -542',NULL);
INSERT INTO evaluations VALUES(1241,32,'cp -545',NULL);
INSERT INTO evaluations VALUES(1241,33,'cp -568',NULL);
INSERT INTO evaluations VALUES(1241,34,'cp -567',NULL);
INSERT INTO evaluations VALUES(1241,35,'cp -563',NULL);
INSERT INTO evaluations VALUES(1241,36,'cp -566','e1g1 g5h4 d2d3 f8c5 b1d2 g8f6 h2h3 c5a7 d2f3 h4h5 c2c3 d7d6 a2a4 h7h6 c1e3 a7e3 f2e3 e8g8 a4a5 g8h8 f1f2 h5g6 g1h2 c6e7 d1b3 a8b8 a1f1 e7g8 b3b4 c8d7 b4b3 b7b5 a5b6 c7b6 c4a6 b6b5 b3b4 f8d8');
INSERT INTO evaluations VALUES(18653,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(18653,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(18653,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(18653,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(18653,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(18653,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(18653,36,'cp 19','e1g1 g8f6 c4b5 d8c7 f1e1 d7d6 e4e5 f6d7 c1f4 f8e7 b1d2 e8g8 e5d6 e7d6 f4d6 c7d6 d2e4 d6c7 b5c6 c7c6 d3d4 c6b6 a2a4 c5d4 d1d4 b6d4 f3d4 f8d8 a1d1 d7f8 d4c6 d8d1 c6e7 g8h7 e1d1 c8d7 e4c5 d7e8 b2b3 b7b6 c5a6 f8g6 a6c7 a8b8 e7g6 h7g6');
INSERT INTO evaluations VALUES(19658,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(19658,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19658,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19658,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19658,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19658,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19658,36,'cp 0','e1g1 f8g7 c2c3 g8f6 e4e5 d6e5 f4e5 f6d5 d2d4 c5d4 c3d4 e8g8 b1c3 d5b6 c4b3 c8g4 g1h1 d8d7 c1g5 c6a5 b3c2 a5c6');
INSERT INTO evaluations VALUES(19657,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19657,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19657,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(19657,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19657,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19657,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19657,36,'cp 0','f8g7 d2d3 e7e6 c2c3 g8e7 a2a4 e8g8 b1a3 b7b6 c4a2 d8d7 c1e3 d6d5 e3f2 c8b7 f2h4 b7a6 a3b5 a6b7');
INSERT INTO evaluations VALUES(5478,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(5478,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5478,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5478,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5478,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5478,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5478,36,'cp 2','d2d3 e7e6 a2a4 g8e7 c2c3 b7b6 c4b3 e8g8 d1e2 d8d7 b1a3 a8b8 c1e3 a7a6 a3c4 b6b5 c4d2 c8b7 f1c1 b5b4 b3a2 b7a8 a2c4 a8b7 h2h3 d7c7 c4b3 f8e8 e2f2 c6a5 b3a2 b4c3 b2c3');
INSERT INTO evaluations VALUES(19945,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19945,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19945,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19945,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19945,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19945,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19945,36,'cp 0','g8f6 g1f3 e7e6 e1g1 f8e7 f1e1 e8g8 b1d2 c6a5 c4b5 c8d7 b5d7 d8d7 d2c4 a5c6 a2a4 d7c7 c1f4 f6h5 f4d2 a8d8 c4a3 h5f6 d2f4 f6h5');
INSERT INTO evaluations VALUES(20880,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(20880,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(20880,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(20880,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(20880,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(20880,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(20880,36,'cp 137','f1b5 c8d7 g1f3 e7e6 c2c4 f8b4 b1c3 g8f6 a1c1 a7a6 b5c6 d7c6 f3e5 e8g8 e5c6 b7c6 e1g1 b4d6 c4c5 d6e7 d1a4 d8d7 a4a5 e7d8 c3e2 d7c8 c1c2 f6d7 h2h3 g7g5 f4h2 c8b7 e2c1 b7b5 a5c3 d8f6 c1d3 a6a5 h2e5 f6e5');
INSERT INTO evaluations VALUES(25595,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(25595,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(25595,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(25595,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(25595,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(25595,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(25595,36,'cp -28','d2d3 f8d6 e1g1 g8f6 a3a4 d8b6 f1e1 c6e5 c1f4 e5f3 d1f3 e8g8 a4a5 b6c7 f4d6 c7d6 f3g3 d6d4 a2b3 b7c6 h2h3 d4b4 g3e3 a8c8 e4e5 f6h5 c3e4 c6e4 e3e4 b4d2 e4e3 d2e3 f2e3 f7f5 c2c3 g7g6 a1d1 f5f4 e3f4 f8f4 g1h2');
INSERT INTO evaluations VALUES(26909,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(26909,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(26909,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(26909,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(26909,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(26909,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(26909,36,'cp 32','a7a6 b5c6 b7c6 h2h3 g4h5 d2d4 e5d4 d1d4 h5f3 g2f3 g8e7 f3f4 c6c5 d4d3 g7g6 f4f5 e7c6 c3d5 f8g7 g1h1 c6b4 d5b4 c5b4 a2a3 b4a3 a1a3 e8g8 b2b3 f8e8 a3a4 c7c6 f5g6 h7g6 f2f4 d6d5 e4e5 c6c5 f4f5 d5d4 e5e6 d8d5 h1h2 g6f5 e6f7 d5f7 f1f5');
INSERT INTO evaluations VALUES(27179,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(27179,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(27179,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(27179,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(27179,34,'cp 131',NULL);
INSERT INTO evaluations VALUES(27179,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(27179,36,'cp 158','e1g1 a7a6 b5d3 g8f6 f1e1 c6b4 d3e4 d7c6 c1g5 f8e7 g5f6 e7f6 e4c6 b4c6 d4d5 c6b4 d5e6 d8d1 a1d1 f7e6 e1e6 e8f7 e6e2 h8d8 d1e1 b4c6 g1f1 h7h6 e2e4 f6c3 b2c3 a8c8 g2g3 d8d5 h2h4 c8d8 f1g2 d5a5 a2a4 a5c5 e1e3 d8d7 g3g4');
INSERT INTO evaluations VALUES(27178,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(27178,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(27178,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(27178,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(27178,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(27178,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(27178,36,'cp 132','f8d6 f1e1 g8e7 c3e4 e8g8 a2a3 e7g6 b5f1 c6e7 g2g3 e7f5 h2h4 f8e8 c2c4 d6e7 c1d2 d7c6 d2c3 h7h6 a1c1 f5d6 e4d6 e7d6 b2b4 a7a6 f1g2 a8c8 h4h5 g6f8 d4d5 e6d5 e1e8 d8e8 c4d5 c6d7 d1d4 f7f6');
INSERT INTO evaluations VALUES(32178,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(32178,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(32178,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(32178,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(32178,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(32178,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(32178,36,'cp 63','d8d7 e1g1 g7g6 c2c3 b8c6 c4b5 f8g7 c3d4 a7a6 b5c6 b7c6 c1e3 g8e7 b1c3 e8g8 a1d1 d6d5 e3f4 a8e8 f4e5 f7f6 e5f4 d5e4 c3e4 d7f5 e4c3 e7d5 c3d5 c6d5 d1e1 f5d7 f3b3 c7c6 e1e8 f8e8 f1d1');
INSERT INTO evaluations VALUES(35032,30,'cp 223',NULL);
INSERT INTO evaluations VALUES(35032,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(35032,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(35032,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(35032,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(35032,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(35032,36,'cp 187','c1e3 e4c6 b1c3 f8b4 g1e2 d7d5 a2a3 b4a5 d1d2 c6d7 g2g4 f7f6 f2f4 b8c6 f1g2 d7f7 b2b4 a5b6 e1g1 g8e7 a3a4 a7a6 a4a5 b6a7 e5f6 f7f6 f4f5 e6f5 c3d5 e7d5 g2d5 c8e6 g4g5 f6f7 d5e6 f7e6 b4b5 c6e7 e2f4');
INSERT INTO evaluations VALUES(35492,30,'cp -238',NULL);
INSERT INTO evaluations VALUES(35492,31,'cp -244',NULL);
INSERT INTO evaluations VALUES(35492,32,'cp -240',NULL);
INSERT INTO evaluations VALUES(35492,33,'cp -225',NULL);
INSERT INTO evaluations VALUES(35492,34,'cp -250',NULL);
INSERT INTO evaluations VALUES(35492,35,'cp -256',NULL);
INSERT INTO evaluations VALUES(35492,36,'cp -246','f8d6 e1g1 h8f8 d2d4 f7g8 b1c3 c8f5 c3b5 g8h8 c1g5 a7a6 b5d6 d8d6 g5h4 b7b5 h4g3 d6b6 c4b3 f5e4 a2a4 b8d7 d5d6 c7d6 a4b5 a6b5 a1a8 f8a8 f2f3 e4d5 b3d5 f6d5 f1e1 a8d8 h2h3 d7f6 g1h2 b5b4 c2c4 b4c3 b2c3 b6c6 d1b3 c6c3');
INSERT INTO evaluations VALUES(37114,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(37114,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(37114,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(37114,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(37114,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(37114,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(37114,36,'cp -41','c7c5 e1g1 c5c4 d3e2 f8e8 a2a3 b4c3 d2c3 f6e4 f3e5 b8c6 c3e1 a7a5 e2f3 e4g5 e1c3 c8f5 a1c1 c6e7 f1e1 b7b5 f3h5 e7g6 h5g4 f5e6 e5g6 h7g6 e1e5 e6g4 d1g4 g5e4 e5e8 d8e8 c1e1 e8f8 g4f3');
INSERT INTO evaluations VALUES(38047,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(38047,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(38047,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(38047,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(38047,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(38047,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(38047,36,'cp 36','g8f6 e2e3 e8g8 f1d3 d5c4 d3c4 b7b6 e1g1 c8b7 a1c1 b8d7 d1e2 c6c5 a2a3 c5d4 a3b4 d4c3 d2c3 f6e4 c3d4 d8e7 f3e5 d7e5 d4e5 a8d8 f1d1 e7g5 e5f4 g5e7 b4b5 e6e5 f4g3 g8h8 e2c2 h7h6 f2f3 e4g3');
INSERT INTO evaluations VALUES(40480,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40480,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(40480,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40480,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40480,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(40480,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(40480,36,'cp 28','e2e3 e8g8 f1d3 b7b6 c4d5 e6d5 e1g1 a7a5 f3e5 c8a6 d3a6 a8a6 a1c1 b4d6 d2e1 b6b5 f2f4 b5b4 c3a4 f6e4 d1h5 d8e7 c1c2 g7g6 h5h3 f8e8 e5d3 b8d7 e1h4 f7f6 f1c1 d7b6 a4c5 d6c5 d3c5 e4c5');
INSERT INTO evaluations VALUES(40078,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40078,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40078,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(40078,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(40078,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40078,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40078,36,'cp 20','b7b6 e1g1 b8d7 c4d5 c6d5 d1a4 a7a6 b2b4 e8g8 b4b5 c8b7 b5a6 b7a6 e2a6 d7b8 a4b5 a8a6 a2a4 d8c8 c1d2 b8d7 f1c1 c8b7 b5b3 f8c8 h2h3 a6a8 c3b5 c8c1 d2c1 h7h6 c1a3 e7a3 b3a3 f6e4 f3e1 d7f6 e1d3 a8c8');
INSERT INTO evaluations VALUES(43081,30,'cp 136',NULL);
INSERT INTO evaluations VALUES(43081,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(43081,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(43081,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(43081,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(43081,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(43081,36,'cp 124','d7d5 d1d2 c8d7 f1d3 d5e4 c3e4 d7c6 e3h6 e8g8 e1c1 b8d7 h6g7 g8g7 h2h4 h7h5 f3g5 c6e4 g5e4 e7g8 g2g4 d7f6 g4h5 f6h5 d3e2 d8d5 d2e3 g8f6 e4f6 h5f6 h4h5 d5e4 h5g6 f8h8 e2d3 e4e3 f2e3 f7g6 h1g1 h8h6 g1g5 a8h8 d1g1 g7f7 d3g6 f7e7 c2c4 h6h1 g6d3');
INSERT INTO evaluations VALUES(46267,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(46267,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(46267,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(46267,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(46267,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(46267,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(46267,36,'cp 33','e4e5 f6d7 c2c3 c5d4 c3d4 b8c6 d2f3 f7f6 e5f6 d7f6 g1e2 f8d6 e1g1 e8g8 c1f4 d6f4 e2f4 f6e4 f4e2 c8d7 d3e4 d5e4 f3d2 d8h4 d1b3 a8d8 a1c1 d7e8 b3e6 e8f7 e6e4 h4f6 e4e3 c6d4 e2d4 f6d4 e3d4 d8d4 d2f3 d4e4 c1c7 e4e2 c7b7 f7d5 b7a7 e2b2 f3e5 b2e2 e5d7 f8f5 a7c7 d5a2 c7c8 g8f7');
INSERT INTO evaluations VALUES(46329,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46329,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46329,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(46329,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46329,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(46329,36,'cp 0','d3e2 b7b5 a2a4 b5b4 b2b3 c4b3 c3b4 f8b4 b1d2 b3b2 c1b2 e8g8 e1g1 c8a6 e2a6 b8a6 b2a3 d8d6 a3b4 a6b4 a4a5 f8c8 d1a4 b4d3 a4b5 c8c3 b5b7 c3c8');
INSERT INTO evaluations VALUES(47276,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(47276,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(47276,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(47276,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(47276,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(47276,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(47276,36,'cp -21','g1f3 c8g4 d2d4 b8d7 f1g2 c7c6 e1g1 e7e5 d4e5 d7e5 b1d2 f8e7 h2h3 g4h5 b2b3 e5f3 g2f3 h5f3 d2f3 f6e4 c1b2 e7f6 b2f6 d8f6 g1g2 e8c8 d1d4 f6d4 e3d4 f7f6 f1e1 h8e8 e1e3 e4d6 a1e1 e8e3 e1e3 c8d7 a2a4 h7h6 c2c3 a7a5 f3h4 g7g5');
INSERT INTO evaluations VALUES(47583,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(47583,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(47583,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(47583,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47583,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(47583,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(47583,36,'cp 32','d8e7 b1c3 c7c6 a2a4 h6g4 g1e2 g4f6 e1g1 f8g7 f3g3 c8e6 c4e6 f7e6 g3h3 e8g8 c1g5 b8d7 c3b1 a7a5 b1d2 d6d5 g1h1 e7d6 g5e3 b7b5 f2f4 e5f4 e2f4 f8e8 d2f3 e6e5 f4e2 e8f8 f3g5 b5a4 a1a4');
INSERT INTO evaluations VALUES(48754,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(48754,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(48754,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(48754,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(48754,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(48754,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(48754,36,'cp -24','c4b3 f8c5 e1g1 a7a5 b1c3 c7c6 f1d1 c5d4 f3f4 d4a7 d2d4 e8g8 f4f3 f8e8 c1g5 c8e6 g5f6 d8f6 f3f6 g7f6 e5f3 b8d7 c3e2 e6g4 d1e1 f6f5 c2c3 a7b8 e2c1 b8d6 c1d3 g4f3 e1e8 a8e8 g2f3 d7f8 a1e1 f8e6 b3c2 h7h5 d3e5 f5f4 e5d3 g8g7');
INSERT INTO evaluations VALUES(48761,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(48761,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(48761,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(48761,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(48761,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(48761,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(48761,36,'cp -29','f8c5 e1g1 a7a5 c2c3 a5a4 b3d1 d5d4 b2b4 a4b3 d1b3 e8g8 c1b2 a8a6 c3d4 c5d4 b2d4 d8d4 f3e3 d4d8 d2d4 c8e6 b1d2 e6b3 d2b3 b8d7 e5d7 f6d7 e3f3 d8a8 a2a4 a6a4 a1a4 a8a4 f3b7 f8b8 b3c5 b8b7 c5a4 b7b4 a4c5 b4d4 c5d7');
INSERT INTO evaluations VALUES(48772,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(48772,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48772,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(48772,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(48772,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(48772,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(48772,36,'cp -29','d5e4 d3e4 b8d7 g1f3 h7h6 g5f6 d8f6 c3d5 f6d8 d5e3 f8b4 c2c3 b4c5 d1d2 a7a5 f1c4 d8f6 b2b4 c5e3 d2e3 e8g8 e1g1 d7b6 c4e2 c8g4 f1d1 f8d8 d1d8 f6d8 e3c5 g4f3 e2f3 a5b4 c5e5 b4c3 e5c3 c7c6 f3g4 d8c7 a1d1');
INSERT INTO evaluations VALUES(49205,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(49205,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(49205,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(49205,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(49205,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(49205,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(49205,36,'cp 85','g1f3 d8e7 d3d4 b8c6 d4d5 c6a5 c4d3 f8g7 c2c4 b7b6 e1g1 e8g8 c1g5 h7h6 g5e3 d7d6 b1c3 f6h7 f3d2 a5b7 d3c2 b7c5 f2f3 a7a5 a2a3 f7f5 e4f5 c8f5 g2g4 f5d3 c2d3 c5d3 d1c2 d3f4 e3f4 f8f4 g1h1 a8f8 c2g6');
INSERT INTO evaluations VALUES(49797,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(49797,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(49797,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(49797,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(49797,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(49797,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(49797,36,'cp 75','f1c4 b8c6 g1e2 d7d6 d2d3 c8e6 c1h6 g7h6 c4e6 f7e6 f3h3 d8f6 f2f4 e8c8 e1g1 c8b8 f4f5 e6f5 f1f5 f6g6 b1c3 c6e7 f5f2 d6d5 a1f1 d5e4 f2f6 g6g8 d3e4 e7c8 f6f7 g8g6 e2g3 f8c5 g1h1 h8f8 f7f8 d8f8 f1f8 c5f8');
INSERT INTO evaluations VALUES(51477,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(51477,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(51477,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(51477,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(51477,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(51477,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(51477,36,'cp -30','c7c5 f1c4 e7e6 h5e2 c8b7 d2d3 g8e7 g1f3 b8c6 b1c3 a7a6 c1e3 b6b5 c4b3 d8b6 c3b1 c6d4 e3d4 c5d4 b1d2 e7g6 g2g3 f8d6 a2a4 e8g8 a4b5 a6b5 e1g1 d6e7 h2h4 h7h6 f1e1 a8a1');
INSERT INTO evaluations VALUES(51581,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(51581,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(51581,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(51581,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(51581,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(51581,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(51581,36,'cp -23','f1c4 e7e6 h5e2 b8c6 g1f3 g8e7 e1g1 c8b7 f1e1 a7a6 c2c3 d7d5 e4d5 e7d5 d2d4 c5d4 c3d4 b6b5 c4b3 f8b4 c1d2 b4d2 b1d2 d5f6 a2a4 e8g8 a4b5 c6d4 f3d4 d8d4 b5a6 b7a6 b3c4 a6b7 a1a8 b7a8 d2f1 h7h6 f1e3 a8e4 e3g4 f6g4 e2g4 d4c4 g4e4');
INSERT INTO evaluations VALUES(51580,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(51580,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(51580,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(51580,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(51580,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(51580,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(51580,36,'cp -26','e7e6 h5e2 g8f6 g1f3 b8c6 e4e5 f6d5 c4d5 e6d5 b1c3 d7d6 c3b5 c8e6 d2d4 a7a6 c1g5 d8d7 b5d6 f8d6 e5d6 d7d6 e2d2 e8g8 g5f4 d6d7 e1g1 f7f6 h2h3 h7h6 c2c3 c5c4 a1e1 e6f5 e1e3 f8e8 e3e8 a8e8 f1e1 e8e1 f3e1 c6e7 f2f3 f5b1 a2a3 e7g6 f4g3 a6a5 g3f2');
INSERT INTO evaluations VALUES(51617,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(51617,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(51617,32,'cp 227',NULL);
INSERT INTO evaluations VALUES(51617,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(51617,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(51617,35,'cp 193',NULL);
INSERT INTO evaluations VALUES(51617,36,'cp 199','f1b5 c8b7 d2d3 e7e6 e1g1 f8e7 c2c4 d8c7 b1c3 a7a6 b5a4 b8c6 f1e1 g8f6 h2h3 e8g8 c3d5 e6d5 e4d5 a8e8 d5c6 b7c6 a4c6 c7c6 f3e5 c6c8 b2b3 h7h6 c1d2 d7d6 e5f3 d6d5 d1c2 d5d4 d2f4 c8c6 c2d1 e7d6 f4e5 d6e5 e1e5 e8e5 f3e5 c6b7 e5f3');
INSERT INTO evaluations VALUES(51614,30,'cp 213',NULL);
INSERT INTO evaluations VALUES(51614,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(51614,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(51614,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(51614,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(51614,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(51614,36,'cp 188','c8b7 d2d3 e7e6 e1g1 f8e7 c2c4 d8c7 b1c3 g8f6 c1g5 a7a6 b5a4 b8c6 a4c6 b7c6 f1e1 h7h6 g5h4 d7d6 d3d4 c5d4 f3d4 c6b7 b2b3 e8g8 d1d2 g7g5 h4g3 f6h5 a1c1 f8d8 e1d1 a8c8 d4e2 h5g3 e2g3 c7a5 c3d5 a5d2 d5e7 g8f8 d1d2 f8e7');
INSERT INTO evaluations VALUES(52199,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(52199,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(52199,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(52199,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(52199,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(52199,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(52199,36,'cp 137','f6e5 b1c3 d8c7 f1c4 g8f6 a2a3 a7a5 c1e3 b8a6 e1g1 f8c5 f3g5 b7b5 c4a2 d7d6 e3c5 a6c5 g5f7 e8g8 d1d6 c7d6 f7d6 c8e6 f2f3 e6a2 c3a2 f8d8 d6f5 g8f7 f5e3 g7g6 a1d1 d8d4 a2c1 f7e7 g1f2 a8d8 c1e2 d4d1 f1d1 d8d1 e3d1 f6d7 e2c1 a5a4');
INSERT INTO evaluations VALUES(53513,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(53513,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(53513,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(53513,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(53513,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(53513,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(53513,36,'cp 75','c2c3 g8f6 d3d4 d8c7 d1c2 c8g4 b1d2 e5d4 c3d4 c5d4 f3d4 b8c6 d4c6 b7c6 e1g1 f8e7 f1e1 g4h5 d2b3 h5g6 c1f4 c7b6 f4e3 c6c5 c2e2 e8g8 f2f3 f6d7 e3d2 a7a5 d2c3 e7f6 a1d1 f6c3 b2c3 a8b8 a2a4 d7e5 b3d2 b6b2 c4b5 c5c4 h2h3 b2c3 d2c4 e5c4 e2c4');
INSERT INTO evaluations VALUES(53961,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(53961,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(53961,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(53961,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(53961,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(53961,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(53961,36,'cp -52','b8c6 e5d6 d8d6 e1g1 g8f6 c2c3 f8e7 d2d4 c5d4 f3d4 e8g8 d4c6 d6c6 b1d2 b7b5 c4d3 c8b7 d1f3 f8d8 f3c6 b7c6 d3e2 a7a5 a2a3 f6d5 d2f3 c6e8 f1d1 h7h6 c1d2 e7f6 f3d4 a8b8 d4b3 a5a4 b3d4 g7g5 e2f3 f6g7 a1c1 b8c8 f3e2');
INSERT INTO evaluations VALUES(54911,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(54911,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(54911,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(54911,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(54911,34,'cp 165',NULL);
INSERT INTO evaluations VALUES(54911,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(54911,36,'cp 160','g1f3 g8f6 b1c3 b8c6 c1g5 g7g6 d1d2 d6d5 e4d5 f6d5 f1c4 d5c3 b2c3 c8f5 e1g1 f7f6 g5e3 h5h4 f1e1 e7e5 a1b1 e5e4 e3f4 g6g5 d4d5 c6a5 c4b5 e8f7 f3d4 d8d5 d4f5 d5f5 f4c7 a5c6 d2e2 f8e7 e2e4');
INSERT INTO evaluations VALUES(54910,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(54910,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(54910,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(54910,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(54910,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(54910,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(54910,36,'cp 159','g8f6 b1c3 b8c6 c1g5 g7g6 d1e2 d6d5 e4d5 f6d5 c3e4 f8g7 c2c4 d5b6 d4d5 c6e5 f3e5 g7e5 e1c1 c8f5 e4g3 d8d6 g3f5 g6f5 c1b1 f5f4 e2c2 h8g8 g5h4 b6d7 g2g3 e8c8 f1h3 c8b8 h1e1 e5f6 h3d7 d6d7 h4f6 e7f6 c4c5 g8g5 c2b3 b8a8 c5c6 b7c6 d5c6 d7f5 b1a1 d8d1 e1d1');
INSERT INTO evaluations VALUES(56952,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(56952,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(56952,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(56952,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(56952,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(56952,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(56952,36,'cp 29','g8f6 g2g3 f8b4 c1d2 b4e7 f1g2 e8g8 e1g1 d7d5 d1c2 b8d7 d2f4 c7c6 f1d1 a7a5 f3e5 f6h5 f4e3 h5f6 b1d2 d7e5 d4e5 f6d7 c4d5 c6d5 e3d4 d7b8 a1c1 b8c6 d4c3 d8b6 e2e3 c6b4');
INSERT INTO evaluations VALUES(49237,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(49237,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(49237,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(49237,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(49237,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(49237,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(49237,36,'cp 41','g2g3 f8b4 c1d2 b4e7 f1g2 e8g8 d1c2 d7d5 e1g1 b8d7 b2b4 c7c6 c4c5 f6e4 b1c3 e4d2 c2d2 a7a5 a2a3 b7b6 d2c2 c8a6 f1b1 d8c7 e2e4 b6c5 b4c5 a6c4 f3d2 c4a6 b1e1 e7f6 e4e5 f6e7 d2f3 f8b8 g2f1 a6f1 g1f1 b8b7 e1b1 a8b8');
INSERT INTO evaluations VALUES(1786,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(1786,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(1786,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(1786,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(1786,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(1786,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(1786,36,'cp -55','g8f6 d2d3 c6d4 c4b3 a7a5 g1e2 a5a4 e2d4 e5d4 b3c4 f8g7 a2a3 e8g8 b1d2 c7c6 h2h3 b7b5 c4a2 d7d6 f2f4 c8e6 d2f3 c6c5 e1g1 f6d7 c1d2 a8c8 g2g4 e6a2 a1a2 c5c4 d2b4 d7c5 b4c5 c8c5 h3h4 c4d3 c2d3 c5c2 f1f2 f8c8 a2a1');
INSERT INTO evaluations VALUES(1043,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(1043,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(1043,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(1043,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(1043,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(1043,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(1043,36,'cp -57','d2d3 c6d4 c4b3 a7a5 g1e2 a5a4 e2d4 e5d4 b3c4 f8g7 a2a3 e8g8 b1d2 c7c6 h2h3 d7d6 e1g1 b7b5 c4a2 c8e6 f2f4 e6a2 a1a2 c6c5 d2f3 c5c4 c1d2 c4d3 c2d3 f6d7 d2b4 d7c5 f1d1 a8c8 a2a1 f7f5 e4e5 d6e5 f4e5 f8e8 f3d4 g7e5 d4b5');
INSERT INTO evaluations VALUES(3411,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(3411,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(3411,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(3411,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(3411,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(3411,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(3411,36,'cp 36','c6e5 c4b3 c7c5 d4f3 c5c4 f3e5 c4b3 e5c4 b3c2 d1c2 c8e6 c4e3 a8c8 c2d3 f8e8 c1d2 a7a6 f1d1 h7h6 b2b3 d8d7 d2e1 b7b5 f2f3 d7b7 e1f2 b5b4 c3e2 a6a5 d3d2 e7f8 e2f4 c8c3 e3d5 f6d5 f4d5 e6d5 d2d5 b7d5');
INSERT INTO evaluations VALUES(3428,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(3428,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(3428,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(3428,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(3428,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(3428,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(3428,36,'cp 28','c4b3 c7c5 d4f3 c5c4 f3e5 c4b3 e5c4 b3c2 d1c2 c8e6 c4e3 a8c8 c2d3 f8e8 f1d1 e7f8 c1d2 a7a6 a1c1 b7b5 b2b3 d8d7 f2f3 d7b7 d2e1 b5b4 c3d5 f6d5 e3d5 a6a5 c1c8 e8c8 g1h2 b7c6 d5f4 e6d7 d3e2 h7h6 e1f2 d7e6 f4d5 e6d5 d1d5 a5a4 f2e1 a4b3');
INSERT INTO evaluations VALUES(3429,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(3429,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(3429,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(3429,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(3429,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(3429,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(3429,36,'cp 25','c7c5 d4f3 c5c4 f3e5 c4b3 e5c4 b3c2 d1c2 c8e6 c4e3 a8c8 c2d3 f8e8 f1d1 e7f8 e3f1 d8a5 c1e3 a7a6 f1g3 f6d7 g3f5 d7e5 d3b1 e6f5 e4f5 e5c4 c3d5 e8e5 b2b4 a5a4 d1d4 c4e3 f2e3 c8c2 d4g4');
INSERT INTO evaluations VALUES(4386,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(4386,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(4386,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(4386,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(4386,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(4386,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(4386,36,'cp -6','d7d5 e4d5 f6d5 b2b4 c5d6 b4b5 c6e7 c3c4 d5f6 b1d2 e7g6 f1e1 f8e8 e2f1 b7b6 a2a4 c8b7 a4a5 b6a5 a1a5 a7a6 b5a6 b7a6 d1c2 d6b4 a5a4 c7c5 c1a3 b4a3 a4a3 a6b7 e1a1 h7h6 g2g3 d8c7 f1g2 c7d6 c2b3 b7c6 a3a8 c6a8 h2h3 a8c6 a1a7');
INSERT INTO evaluations VALUES(4597,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(4597,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(4597,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(4597,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(4597,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(4597,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(4597,36,'cp 20','b1c3 d7d6 h2h3 c8e6 c3d5 a7a5 e1g1 e6d5 e4d5 c6b8 f1e1 c7c6 d5c6 b7c6 c4a2 b8d7 c2c3 f8e8 d1b3 e8f8 a2b1 a8b8 b3d1 d8c7 d3d4 f8e8 d4e5 d7e5 f3e5 d6e5 d1c2 e5e4 c1e3 e7d6 b1a2 d6f4 e3f4 c7f4 a1d1 c6c5');
INSERT INTO evaluations VALUES(5731,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5731,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(5731,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(5731,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(5731,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(5731,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(5731,36,'cp -13','c5e3 f2e3 c6a5 d1e2 g8f6 e1c1 d8e7 d3d4 a5c4 e2c4 c7c6 c4d3 c8e6 h2h3 a8c8 g2g4 b7b5 d4e5 d6e5 f3e5 b5b4 c3e2 e8g8 g4g5 h6g5 h1g1 f6d7 e5d7 f8d8 d3d6 d8d7 d6e7 d7e7 g1g5 e6a2 d1g1 g7g6 e4e5');
INSERT INTO evaluations VALUES(5755,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(5755,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(5755,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(5755,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(5755,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(5755,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5755,36,'cp 0','f2e3 c6a5 d1e2 a5c4 d3c4 g8f6 e1c1 d8e7 h1g1 a7a6 h2h3 c8e6 c3d5 e7d8 d5c3');
INSERT INTO evaluations VALUES(5754,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(5754,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(5754,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(5754,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(5754,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(5754,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(5754,36,'cp -23','c6a5 d1e2 a5c4 d3c4 g8f6 e1c1 d8e7 h1g1 a7a6 h2h3 c8e6 c3d5 e7d8 d5f6 d8f6 b2b3 e8c8 d1f1 f6e7 f3d2 e7h4 e2d3 h6h5 d2b1 g7g5 b1c3 g5g4 h3g4 h4g4 c1b2 h5h4 a2a4 c8b8 a4a5 b8a7');
INSERT INTO evaluations VALUES(7288,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(7288,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(7288,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7288,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7288,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7288,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7288,36,'cp 0','d7c6 f3e5 f6e4 d1e2 d8d5 e5d3 e8g8 f1e1 c8f5 d3c5 e4c5 b1c3 d5d6 d2d3 c5e6 f2f3 b7b6 e2f2 c6c5 b2b3 d6d4 c1b2 f8e8 a2a4 f5g6 c3d1 d4d7 d1c3');
INSERT INTO evaluations VALUES(7207,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7207,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7207,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7207,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7207,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7207,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(7207,36,'cp 0','f3e5 f6e4 d1e2 d8d5 e5d3 e8g8 f1e1 c8f5 d3c5 e4c5 d2d3 c5e6 c1e3 d5b5 b2b3 b5e5 b1d2 f8e8 a2a4 e6d4 e2d1 e5d5 e3d4 d5d4 d2f3 d4d6 d1d2 c6c5 d2g5 f5d7 f3d2 f7f6 g5h5 b7b6 h5f3 d7c6 f3g4 c6d7');
INSERT INTO evaluations VALUES(7197,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7197,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7197,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7197,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(7197,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7197,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(7197,36,'cp 6','f6e4 d1e2 d8d5 e5d3 e8g8 f1e1 c8f5 d3c5 e4c5 d2d3 c5e6 c1e3 d5d7 b1c3 a8e8 e2f3 f5g6 c3e2 e8e7 a2a4 a6a5 e2f4 f8e8 f4e6 e7e6 e3d2 b7b6 e1e6 e8e6 a1e1 e6e1 d2e1 f7f6 h2h4 g6f5 f3f4 h7h6 f2f3 f5g6 e1g3 c6c5 f4c7 d7a4 c7b6 a4c2');
INSERT INTO evaluations VALUES(8020,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8020,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8020,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8020,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8020,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8020,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8020,36,'cp 0','d7d5 c3d5 f6d5 c4d5 d8d5 e1g1 e8g8 c1e3 g7b2 a1b1 b2f6 f3d4 d5a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(10005,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(10005,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(10005,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(10005,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(10005,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10005,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(10005,36,'cp 10','d7d6 f3d2 e8g8 e1g1 c8e6 h2h3 a7a6 b5c6 b7c6 c3c4 f6d7 d1e1 d8f6 a1b1 a8b8 f2f3 c6c5 f1f2 h7h6 b1b3 f6g6 g1h2 f7f5 b3b8 d7b8 f3f4 f5e4 d2e4 e5f4 c1f4 b8c6 f4g3 f8f2 g3f2 e6f7 e1e2 g6e6 f2g3 c6d4 e2f2 d4f5');
INSERT INTO evaluations VALUES(12651,30,'cp 161',NULL);
INSERT INTO evaluations VALUES(12651,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(12651,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(12651,33,'cp 181',NULL);
INSERT INTO evaluations VALUES(12651,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(12651,35,'cp 175',NULL);
INSERT INTO evaluations VALUES(12651,36,'cp 174','f8e7 e4e5 f6g4 e3e4 d7d6 e5d6 d8d6 f1b5 e8f8 c1f4 d6e6 b5c4 e6e4 c3e4 g4f6 e4f6 e7f6 e1c1 g7g5 f4d6 f8g7 d6c5 c8f5 c5e3 b7b6 d1d6 a8c8 c4b3 h8e8 c2c3 h7h5 g2g3 c8d8 d6d8 e8d8 h1d1 h5h4 d1d8 f6d8 b3d5 f5d7 c1d2 g7g6 e3d4 d8c7');
INSERT INTO evaluations VALUES(15566,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(15566,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(15566,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(15566,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(15566,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(15566,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(15566,36,'cp 18','f1b5 g7g6 c1b2 f8g7 d2d4 e5d4 f3d4 c8d7 d4c6 b7c6 b5d3 d6d5 e1g1 e8g8 f2f3 d8e7 d1e1 d5e4 c3e4 f6e4 b2g7 e7c5 g1h1 g8g7 d3e4 a8d8 a1c1 d7e6 e1g3 c5d6 f3f4 e6d5 e4d3 f8e8 c1e1 c6c5 h2h3 a7a5 g3f2 e8e1');
INSERT INTO evaluations VALUES(16738,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16738,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(16738,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(16738,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(16738,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16738,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16738,36,'cp 0','c6d4 f3d1 c7c6 c2c3 d7d5 c4d5 c6d5 c3d4 d5e4 d3e4 d8d4 d1d4 e5d4 g1f3 f8c5 b1d2 c8e6 d2b3 e6b3 c1h6 g7h6 a2b3 a8d8 a1c1 c5b6 e1d2 d4d3 c1c3 b6f2 h1f1 f2b6 c3d3 d8d3 d2d3 e8e7 f3h4 h8d8 d3c2 d8c8 c2b1 c8g8 h4f5 e7e8 f5d6 e8e7');
INSERT INTO evaluations VALUES(17068,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17068,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17068,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17068,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17068,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17068,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17068,36,'cp 3','f3d1 c7c6 c2c3 d7d5 c4d5 c6d5 c3d4 d5e4 d3e4 d8d4 d1d4 e5d4 g1f3 f8c5 b1d2 c8e6 d2b3 e6b3 c1h6 g7h6 a2b3 a8d8 a1c1 c5b6 e1d2 d4d3 c1c3 b6f2 c3d3 d8d3 d2d3 e8e7 h1c1 f2b6 c1c4 e7e6 f3h4 h8g8 g2g3 g8d8 d3c2 h6h5 h4f5');
INSERT INTO evaluations VALUES(16889,30,'cp 292',NULL);
INSERT INTO evaluations VALUES(16889,31,'cp 313',NULL);
INSERT INTO evaluations VALUES(16889,32,'cp 343',NULL);
INSERT INTO evaluations VALUES(16889,33,'cp 296',NULL);
INSERT INTO evaluations VALUES(16889,34,'cp 292',NULL);
INSERT INTO evaluations VALUES(16889,35,'cp 335',NULL);
INSERT INTO evaluations VALUES(16889,36,'cp 287','d2d4 f6d5 d4e5 c7c6 b1d2 f8e7 d2e4 e8g8 b5d3 f7f6 c2c4 d5b4 e5f6 e7f6 c4c5 d7d5 c5d6 b4d3 d1d3 h7h6 d3b3 g8h7 c1f4 f6b2 a1d1 f8f4 b3b2 f4f7 e4c5 b7b6 d6d7 c8d7 b2b3 d8f6 c5d7 f6f2 g1h1 f2f5 g2g4 f5f3 b3f3 f7f3 e1e7 f3f2 d7e5 f2a2 d1d7 a8g8 d7a7 a2a7 e7a7 g8d8 e5c6 d8d2');
INSERT INTO evaluations VALUES(17023,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(17023,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(17023,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(17023,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(17023,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(17023,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(17023,36,'cp 132','e1g1 d7d5 e4d5 f8e7 c3d4 c8g4 f3g3 e8g8 c1f4 d8d7 f4e5 f6e4 g3e3 e4d6 c4b3 f7f6 e5f4 g7g5 f4g3 g8h8 b1c3 d6f5 e3d3 f5g3 h2g3 e7d6 d3d2 a7a5 a1e1 d7g7 a2a3 f6f5 b3d1 g4d1 d2d1 a8e8 d1d2 g7g6 e1e8');
INSERT INTO evaluations VALUES(17715,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(17715,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(17715,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(17715,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(17715,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(17715,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(17715,36,'cp 63','e2e3 g8f6 c2c4 c8f5 b1c3 e7e6 a2a3 f8d6 f4g5 d6e7 f1d3 d5c4 d3c4 e8g8 e1g1 f6e4 g5e7 d8e7 a1c1 f8d8 f1e1 h7h6 h2h3 c6a7 c3e4 f5e4 c4f1 c7c6 f3e5 a7b5 f2f3 f7f6 e5c4 e4g6 d1b3 a8b8 c4a5 g6f7 b3c2');
INSERT INTO evaluations VALUES(18313,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(18313,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(18313,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(18313,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(18313,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(18313,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(18313,36,'cp 79','f8d6 d2d4 c6d4 f3d4 e5d4 d1d4 d8e7 d4g7 d6e5 g7g4 g8f6 g4e2 e5c3 b2c3 f6e4 e1g1 c8b7 f2f3 e4c5 e2e7 e8e7 f1e1 c5e6 c1f4 a8g8 f4g3 a7a6 b5e2 d7d6 g3h4 f7f6 e2c4 d6d5 c4b3 e7f7 h4g3 g8d8 c3c4 h8e8 a1d1 d5d4 d1b1 b7c6 c4c5 c6d5 c5b6 c7b6 b3d5 d8d5 b1b6');
INSERT INTO evaluations VALUES(18318,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(18318,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(18318,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(18318,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(18318,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(18318,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(18318,36,'cp 94','h7h6 e1g1 c8b7 d3d4 e5d4 c2c3 d4d3 f1e1 g7g5 c4d3 g8e7 c1e3 f8g7 b1d2 d7d6 a2a4 d8d7 a4a5 c6a5 e3d4 h8g8 b2b4 a5c6 d4g7 g8g7 d2f1 g5g4 f3d4 c6e5 d3b5 c7c6 b5e2 e8f8 f1g3 c6c5 e2b5 d7c7 d4f5 e7f5 g3f5 g7g6');
INSERT INTO evaluations VALUES(18397,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(18397,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(18397,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(18397,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(18397,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(18397,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(18397,36,'cp 12','e1g1 f7f6 f3h4 g8e7 f2f4 e5f4 e4e5 d7d5 e5f6 e7g6 d2d3 d8f6 h4g6 h7g6 c1f4 f8e7 b1d2 e8g8 d1e2 c8f5 a1e1 a8e8 f4e5 f6e6 d2f3 f5g4 e2d2 e6d7 e5g3 g4f3 f1f3 f8f3 g2f3 e7f6 e1e8 d7e8 b2b3 e8e7 h2h3 f6d4 g1g2 e7f7 g3f2 f7f6 d2e1 d4f2 e1f2 f6c3');
INSERT INTO evaluations VALUES(20268,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(20268,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(20268,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(20268,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(20268,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(20268,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(20268,36,'cp -28','f8e7 b1d2 f7f5 f1b5 g8f6 a2a4 d8c7 e1g1 e8g8 f3e5 c8d7 b2b3 f8d8 b5c6 d7c6 c1a3 c6e8 d1e2 b7b6 f1c1 a8c8 h2h3 c7b7 c1c2 e7d6 c3c4 e8h5 e2f2 d6e5 d4e5 f6e4 d2e4 d5e4 c2d2 b7c7 a1c1 d8d2');
INSERT INTO evaluations VALUES(21514,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(21514,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(21514,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(21514,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(21514,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(21514,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(21514,36,'cp 78','c2c3 g8f6 e4e5 f6d7 c3d4 d7b6 g2g3 c6e7 f1g2 c8d7 e1g1 d7c6 d2d3 e7f5 a2a3 d8d7 c1e3 b6d5 d1d2 e8c8 a1c1 c8b8 f3g5 h7h6 g5e4 f8e7 c1c2 c6a4 b2b3 a4b3 c2b2 b3a4 f1b1 a4c6 e2c3 b7b6 c3d5 c6d5');
INSERT INTO evaluations VALUES(23830,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(23830,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(23830,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(23830,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(23830,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(23830,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(23830,36,'cp 56','c4b5 a7a6 b5a4 b6a7 d4d5 b7b5 d5c6 b5a4 h2h3 g4f3 d1f3 g8e7 c1g5 f7f6 g5e3 a7e3 f3e3 e8g8 b1c3 e7c6 c3a4 f6f5 e4f5 f8f5 a1c1 c6e5 e3e4 f5f8 b2b3 e5g6 a4c3 d8h4 e4e6 g8h8 c3d5 a8e8 e6g4 h4g4 h3g4 e8e2 d5c7 g6f4 c1d1 d6d5 c7a6 e2a2 a6c5 a2b2 g2g3 f4e2 g1g2 e2c3 d1d3');
INSERT INTO evaluations VALUES(24559,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(24559,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24559,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(24559,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(24559,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24559,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(24559,36,'cp 21','f3h4 f5e6 c4d5 c6d5 h4f3 f6e4 d1b3 d8b6 g5f4 f7f6 g2g3 g7g5 f4e3 f8g7 f1g2 a8c8 e1g1 e4c3 b2c3 b6b3 a2b3 c8c3 a1a7 c3b3 f3d2 b3b6 f1b1 e8f7 b1b6 d7b6 a7b7 b6c8 b7b5 h8d8 b5c5 c8b6 h2h4 g5h4 g3h4 b6d7');
INSERT INTO evaluations VALUES(24561,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(24561,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24561,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24561,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(24561,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(24561,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24561,36,'cp 31','c1f4 e7e6 e1g1 f8e7 c2c4 e8g8 f1d1 d8b6 a1b1 a7a5 g3e4 f6e4 d3e4 f8e8 e4c2 c6c5 d4d5 e6d5 d1d5 d7f6 d5d2 e7f8 f4e5 a5a4 h2h3 b6c6 b1d1 h7h6 d2e2 b7b6 d1e1 a8d8 e5f6 e8e2 e1e2 g7f6 e2e3 b6b5 a2a3');
INSERT INTO evaluations VALUES(25634,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(25634,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(25634,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(25634,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(25634,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(25634,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(25634,36,'cp 63','d2d3 g4h5 h2h3 e7e6 b1d2 g8f6 a2a4 f6d7 a4a5 e6e5 d1e1 f7f6 b2b3 d8c7 e1e3 h5f7 d2c4 f8e7 f3d2 d7f8 f2f4 e5f4 e3f4 c7f4 f1f4 a8d8 f4f2 f8e6 d2f3 d8d7 f3h4 e6d4 c4b6 d7d8 c1b2 e8g8 a1f1 e7d6 g2g4 d6c7 h4f5 d4e6 b6c4');
INSERT INTO evaluations VALUES(25631,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(25631,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(25631,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(25631,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(25631,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(25631,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(25631,36,'cp 57','g4h5 h2h3 e7e6 b1d2 g8f6 a2a4 f6d7 a4a5 f7f6 d1e1 e6e5 b2b3 d8c7 e1e3 h5f7 d2c4 f8e7 f3d2 d7f8 f2f4 e5f4 e3f4 c7f4 f1f4 f8e6 f4f2 a8d8 d2f3 e6d4 c1e3 f7e6 f3h4 e8g8 g2g4 g7g6 h4f3 d4f3 f2f3 f6f5 e3h6');
INSERT INTO evaluations VALUES(26205,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(26205,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(26205,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(26205,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(26205,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(26205,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26205,36,'cp 0','e7e6 h2h3 f8e7 e1g1 g8f6 c2c4 e8g8 b1c3 a7a6 c1e3 d7e5 a2a4 d8c7 a1c1 a8d8 b2b3 c7b6 c1b1 b6a5 b1c1');
INSERT INTO evaluations VALUES(26650,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26650,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26650,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26650,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26650,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(26650,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26650,36,'cp -1','e7e6 c1e3 f7f6 f2f4 f8b4 g1f3 f6e5 f4e5 b4c3 b2c3 g8e7 f1d3 d8d7 e1g1 h7h6 f3h4 f5d3 d1d3 h8g8 a2a4 e8c8 a4a5 g7g5 h4f3 d8f8 f3d2 f8f1 d3f1 e7f5 e3f2 g8f8 d2b3 d7f7 c3c4 f5e7 c4d5 e7d5 b3c5 f7f5 c2c4 d5f4 a5a6 b7b6 c5b7 f5d3 f1d3');
INSERT INTO evaluations VALUES(26847,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(26847,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26847,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26847,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26847,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26847,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26847,36,'cp 0','c1e3 f7f6 f2f4 f8b4 g1f3 g8e7 f1d3 f6e5 f4e5 e8g8 e1g1 b4c3 b2c3 d8e8 d1d2 h7h6 a1b1 b7b6 f3h4 f5d3 f1f8 g8f8 d2d3 f8g8 g2g4 a8c8 b1f1 c6d8 e3h6 g7h6 f1f6 d8f7 f6e6 e8d7 e6g6 e7g6 d3g6 g8h8 g6f6 h8g8');
INSERT INTO evaluations VALUES(29886,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(29886,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(29886,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(29886,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(29886,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(29886,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(29886,36,'cp 52','h5g6 d2d4 f6e4 c3e4 d6d5 e4c3 d5c4 f3e5 f8b4 d1f3 b8d7 e5g6 h7g6 e1g1 c7c6 f1e1 e8f8 g1g2 d7b6 c1d2 f8g8 e1e2 d8d4 a1e1 d4d7 g2f1 b4c3 d2c3 b6d5 e2e4 d5c3 f3c3 b7b5 a2a4 d7d5 a4b5');
INSERT INTO evaluations VALUES(29891,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(29891,31,'cp 90',NULL);
INSERT INTO evaluations VALUES(29891,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(29891,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(29891,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(29891,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(29891,36,'cp 107','a2a3 f8e7 g2g4 h5g6 b1c3 c7c6 c4a2 b8d7 h3h4 h7h5 g4g5 f6g4 d1e2 f7f6 g5f6 g7f6 h1g1 d8c7 c1d2 e8c8 e1c1 g6e8 f3e1 d7b6 f2f3 g4h6 f3f4 h6g4 a2e6 e8d7 e6d7 d8d7 e1f3 d6d5 f4e5 f6e5 e4d5 b6d5 c3e4 d5f4 e2f1 f4e6 f3g5 h8f8 g5e6 f8f1 e6c7 f1g1 d1g1 c8c7');
INSERT INTO evaluations VALUES(31154,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(31154,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(31154,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(31154,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(31154,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(31154,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(31154,36,'cp -23','d8b6 b2b3 e7e6 b1c3 f8a3 c3a4 b6c7 f4c1 a3c1 a1c1 b8d7 f1d3 g8e7 g1f3 c6c5 a4c5 d7c5 d4c5 c7c5 d3f5 e7f5 c2c4 d5c4 c1c4 c5b6 e1g1 a8d8 d1c2 h7h6 f1c1 f5e7 b3b4 e8g8 g2g3 d8c8 c2b3 c8c4 c1c4 f8d8 h2h4 a7a5 g1g2 d8d5 h4h5 a5b4 b3b4');
INSERT INTO evaluations VALUES(32193,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(32193,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(32193,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(32193,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(32193,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(32193,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(32193,36,'cp 134','d1b3 g4f3 g2f3 g7g6 c3d4 f8g7 c1e3 b8c6 b3a4 g8e7 b1c3 e8g8 e1c1 a7a6 h2h4 b7b5 a4c2 b5b4 c3a4 h7h5 f3f4 b4b3 a2b3 c6b4 c2d2 a8b8 f4f5 e7c6 f1c4 c6a5 f5g6 a5c4 g6f7 f8f7 b3c4 b4a2 c1b1 d8e8 a4c3 a2c3 d2c3 c7c5 h1g1 f7b7 e3c1 c5d4 c3d4 g8h7 g1g7 b7g7 f2f3 e8g6 d4d6 g6d6 d1d6');
INSERT INTO evaluations VALUES(33641,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33641,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(33641,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(33641,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(33641,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33641,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33641,36,'cp -6','e2e3 c7c5 g1e2 c5d4 e2c3 d5e5 d1d4 f8g7 c3d5 e8f8 d4e5 g7e5 e3e4 b8c6 c1h6 e5g7 h6f4 c8e6 e1c1 a8d8 f1c4 h7h5 f4e3 h5h4 h1e1 g7e5 f2f4 e5d6 c1b1 f7f6 a2a3 f8f7 c4e2 f7g7 d1c1 h4h3 g2g3');
INSERT INTO evaluations VALUES(34575,30,'cp 800',NULL);
INSERT INTO evaluations VALUES(34575,31,'cp 794',NULL);
INSERT INTO evaluations VALUES(34575,32,'cp 794',NULL);
INSERT INTO evaluations VALUES(34575,33,'cp 794',NULL);
INSERT INTO evaluations VALUES(34575,34,'cp 794',NULL);
INSERT INTO evaluations VALUES(34575,35,'cp 794',NULL);
INSERT INTO evaluations VALUES(34575,36,'cp 797','c3b5 b8a6 d1e2 g8f6 a2a4 f6d5 g1f3 c8g4 b5d4 e8c8 e2c4 f7f6 d4c6 g4f3 c6d8 f3g2 h1g1 c8d8 g1g2 c7c6 e1f1 e7e6 c4g4 e6e5 d2d4 h7h5 g4e4 e5d4 e4d4 f8c5 d4c4 c5f8 c2c3 a6c7 b2b4 d8c8 c1b2 c8b8 h2h4 c7e6');
INSERT INTO evaluations VALUES(34574,30,'cp 782',NULL);
INSERT INTO evaluations VALUES(34574,31,'cp 786',NULL);
INSERT INTO evaluations VALUES(34574,32,'cp 781',NULL);
INSERT INTO evaluations VALUES(34574,33,'cp 778',NULL);
INSERT INTO evaluations VALUES(34574,34,'cp 778',NULL);
INSERT INTO evaluations VALUES(34574,35,'cp 778',NULL);
INSERT INTO evaluations VALUES(34574,36,'cp 771','b8a6 d2d4 g8f6 d1e2 c8g4 e2e5 e8c8 c2c4 g4e6 e5e2 c8b8 h2h3 c7c6 b5a3 d8d4 g1f3 d4d6 e1g1 g7g6 c1f4 f8g7 a1c1 a6c7 a3c2 f6h5 f4d6 e7d6 c2d4 h5f4 e2e4 g7d4 f3d4');
INSERT INTO evaluations VALUES(35391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35391,36,'cp 0','f8d8 c1g5 d8e8 e1c1 b8a6 f3e1 a6c5 f2f3 c7c6 g5e3 c5e6 e1c2 a7a5 b2b3 f6d7 h2h4 h7h5 c3a4 g7f8 g2g3 e6c5 a4c3 c5e6');
INSERT INTO evaluations VALUES(36269,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(36269,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(36269,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(36269,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(36269,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(36269,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(36269,36,'cp 55','d1d2 b8c6 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 h1e1 a8d8 c4e6 f7e6 f4e3 d8d1 c3d1 f8e8 g2g3 e7d6 e3e2 d6d5 a2a3 d5g2 e2c4 f6d5 h2h4 h7h6 c4e2 d5f6 d1e3 g2c6 e3g4 f6g4 e2g4 e6e5 f2f3');
INSERT INTO evaluations VALUES(36270,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(36270,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(36270,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(36270,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(36270,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(36270,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(36270,36,'cp 54','b8c6 e1c1 c6d4 d2d4 a7a6 e4e5 d6e5 d4e5 e7d6 e5d4 d6f4 d4f4 d8e7 f1c4 c8e6 h1e1 a8d8 c4e6 f7e6 g2g3 d8d1 c3d1 f8e8 f4e3 e7d6 d1c3 e6e5 e3e2 d6c6 f2f3 f6d5 c3e4 d5f6 e2d3 f6e4 e1e4 c6h6 d3d2 h6d2 c1d2 g8f7 d2e3 f7e6 e4c4 c7c6 c4c5');
INSERT INTO evaluations VALUES(36625,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(36625,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(36625,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(36625,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(36625,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(36625,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(36625,36,'cp -5','f6h5 f4g5 h7h6 g5h4 f7f5 c2c3 g6g5 h4g3 e7e6 a2a4 b8c6 a4a5 c8d7 e1g1 g8h8 f3e1 g5g4 e1d3 e6e5 d4e5 c6e5 a5a6 b7b6 d3f4 d8e8 f4h5 e8h5 f2f4 e5g6 e2d3 a8e8 d1e2 d7c6 f1f2 g7f6 d2f1 f6h4');
INSERT INTO evaluations VALUES(38343,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(38343,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(38343,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(38343,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(38343,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(38343,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(38343,36,'cp -12','e5d6 d8d6 d2d4 b8c6 f1b5 c8d7 c2c3 e7e5 d4e5 d6d1 e1d1 g8h6 c1g5 h6g4 g5h4 c6e5 b5d7 e8d7 b1d2 a8e8 h2h3 e5f3 d2f3 g4h6 h1e1 e8e1 d1e1 h6f5 e1f1 f5h4 f3h4 h8e8 h4f3 g7f6 a1d1 d7c6 g2g3 b7b5');
INSERT INTO evaluations VALUES(39074,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(39074,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(39074,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(39074,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(39074,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(39074,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(39074,36,'cp 140','c7c5 c2c3 g8f6 d2d4 b8c6 d4e5 d6e5 f3e5 c6e5 c4e2 d7d6 f2f3 e8g8 c3c4 c8e6 b2b3 a7a6 c1b2 b7b5 c4b5 a6b5 e2b5 c5c4 b2e5 c4b3 e5d4 b3b2 d4b2 d8b6 b2d4 b6b5 b1c3 b5g5 d1d3 f8d8 f1d1 d6d5 d4f6 g7f6 e4d5');
INSERT INTO evaluations VALUES(40228,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(40228,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(40228,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(40228,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(40228,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(40228,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(40228,36,'cp -22','e8g8 a1c1 h7h6 g5h4 b8c6 c4c5 f6e4 h4e7 d8e7 f1b5 e4c3 b2c3 c8d7 e1g1 f8c8 c3c4 d5c4 d1d2 c8d8 b5c4 e7f6 c1b1 c6e7 b1b7 d7c6 b7b3 c6f3 b3f3 d8d4 d2c1 f6e5 f1e1 e5g5 c4a6 g5c1 e1c1 e7c6');
INSERT INTO evaluations VALUES(35840,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(35840,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(35840,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(35840,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(35840,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(35840,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(35840,36,'cp -18','c4d5 f6d5 g5e7 d5e7 f1d3 b8d7 e1g1 d7f6 f3e5 b7b6 d1f3 e7d5 a2a3 c8b7 f1e1 a8c8 a1c1 d8e7 c3d5 b7d5 f3f4 d5b7 c1c8 f8c8 e1c1 c8c1 f4c1 e7d6 h2h3 h7h6 c1e3 d6d5 f2f3 d5d8 d3c2 d8c7 e3d3 g7g6 h3h4 f6d5 d3e4');
INSERT INTO evaluations VALUES(42453,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(42453,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(42453,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42453,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(42453,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(42453,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(42453,36,'cp 49','b1d2 d7d5 c2c3 b4d6 f4g5 d6e7 e2e3 c7c5 f1d3 b8d7 e1g1 c5c4 d3c2 b7b5 f3e5 d7e5 d4e5 f6d7 g5e7 d8e7 f2f4 a7a5 d1f3 e8g8 f3h3 f7f5 a2a3 d7c5 d2f3 a8b8 f3d4 b8b6 g2g4 g7g6 h3g3 g8h8 h2h4 f5g4 g3g4 b5b4');
INSERT INTO evaluations VALUES(42873,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(42873,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(42873,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(42873,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(42873,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(42873,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(42873,36,'cp 66','e5d4 f3d4 e8g8 d1g4 c5d4 g4g3 d7d6 c3d5 d4f6 d5f6 d8f6 e1g1 c8e6 c4e2 a7a5 f1d1 b8c6 c1d2 f6g6 g3g6 h7g6 a2a4 f7f6 h2h4 f6f5 e4f5 e6f5 c2c3 f5c2 d1c1 c2b3 c1e1 c6e5 d2g5 b3c4 e2d1 a8e8 d1c2 b7b6');
INSERT INTO evaluations VALUES(42880,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(42880,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(42880,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(42880,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(42880,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(42880,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(42880,36,'cp 73','f3d4 d7d5 e4d5 e8g8 e1g1 g4h2 c3e4 c5d4 c1g5 f7f6 g5f4 d4e5 f4h2 e5h2 g1h2 f6f5 e4d2 d8d6 h2g1 b8d7 d2f3 f8e8 f1e1 e8e1 d1e1 d7f6 c4d3 b7b6 c2c4 c8d7 e1c3 a8e8 b2b4 f6e4 c3b3 d6f6 a1e1 h7h6 d3e4 e8e4 e1e3 e4e3 b3e3 f5f4 e3a3 d7g4 c4c5 g4f3 a3f3 f6a1 g1h2');
INSERT INTO evaluations VALUES(42879,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(42879,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(42879,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(42879,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(42879,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(42879,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(42879,36,'cp 42','d7d5 e4d5 e8g8 e1g1 g4h2 c3e4 c5d4 c1g5 f7f6 g5f4 d4e5 f4h2 e5h2 g1h2 f6f5 e4d2 d8d6 h2g1 b8d7 d2f3 f8e8 f1e1 e8e1 d1e1 d7f6 c4d3 c8d7 c2c4 a8e8 e1a5 d6b6 a5b6 a7b6 a2a4 g8f8 b2b4 e8a8 f3e5 g7g6 d3c2 d7e8 b4b5 f8e7 f2f4 f6d7 g1f2 e7d6 a1h1 d7e5 f4e5 d6e5 h1h7');
INSERT INTO evaluations VALUES(44035,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(44035,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(44035,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44035,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44035,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(44035,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(44035,36,'cp 19','h7h6 g5h4 f8e7 e2e3 e8g8 g1f3 d7d5 f1d3 d5c4 d3c4 c6c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 c4e2 a7a6 b5d4 c8d7 d4c6 d7c6 f3e5 a8c8 h4f6 g7f6 e5c6 c8c6 a2a3 b4d6 e2f3 c6b6 b2b4 f8d8 g2g3 f6f5 d1d4 g8f8 f1d1 f8e7 d4c4 d8a8 d1c1 a6a5 b4a5 a8a5');
INSERT INTO evaluations VALUES(44401,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(44401,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(44401,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(44401,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(44401,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(44401,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(44401,36,'cp 49','d7d5 c4d5 c6d5 c1f4 b8c6 e2e3 f6h5 f4g5 f8g7 g5h4 g7f6 h4f6 h5f6 h2h3 e8g8 f1b5 c8d7 e1g1 a8c8 b5d3 c6b4 d3e2 e7e6 a1c1 d8e7 a2a3 b4c6 b2b4 f8d8 d1b3 c8c7 f3d2 d8c8 b3b2 b7b6');
INSERT INTO evaluations VALUES(44421,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(44421,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44421,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44421,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44421,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44421,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44421,36,'cp 0','c5d4 e3d4 d7d5 c4d5 f8g7 f1c4 f6d5 d1b3 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(44797,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(44797,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(44797,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(44797,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(44797,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(44797,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(44797,36,'cp 58','b1c3 f8g7 c1e3 e8g8 d1d2 b8c6 e1c1 d6d5 e4d5 f6d5 d4c6 b7c6 e3d4 g7d4 d2d4 d8b6 c3a4 b6c7 f1c4 f8d8 a4c5 c8f5 c4b3 d5f4 d4c4 d8d5 g2g4 f5e6 c5e6 f4e6 d1d5 c6d5 c4c7 e6c7 h1d1 e7e6 d1d4 g6g5 c2c4 d5c4 d4c4');
INSERT INTO evaluations VALUES(45112,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(45112,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45112,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45112,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(45112,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(45112,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(45112,36,'cp -9','c8d7 b5d7 d8d7 d2d3 e7e6 e1g1 f8e7 c1f4 e8g8 b1d2 b8c6 a2a4 b7b6 h2h3 f8d8 d1e2 h7h6 d2c4 d7b7 e4e5 f6d5 f4g3 d6e5 f3e5 c6e5 g3e5 d8d7 f1d1 a8d8 e2g4 d5f6 g4e2 f6h7 d3d4 h7f8 d4c5 e7c5 d1d7 f8d7');
INSERT INTO evaluations VALUES(28870,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(28870,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(28870,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(28870,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(28870,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(28870,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(28870,36,'cp -7','b5d7 d8d7 d2d3 e7e6 a2a4 b8c6 c1f4 f8e7 e1g1 e8g8 b1d2 b7b6 f1e1 f6h5 f4e3 h7h6 d1e2 f8d8 d2f1 h5f6 f1g3 f6g4 e3d2 a8c8 g3h5 g4f6 h5g3 f6h7 g3h5 c6e5 d3d4 e5f3 e2f3 e7g5 d2g5 h6g5');
INSERT INTO evaluations VALUES(28844,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28844,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28844,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(28844,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28844,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28844,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28844,36,'cp -13','d8d7 d2d3 e7e6 c1f4 f8e7 e1g1 b8c6 a2a4 e8g8 b1d2 b7b6 f1e1 f8d8 d2c4 h7h6 d1e2 d7b7 h2h4 a7a6 c4e3 d6d5 e4e5 f6h7 e3g4 b6b5 f4e3 g8h8 g4h2 b5a4 a1a4 d8b8 e1a1 b7b2 e2b2 b8b2 a4a6');
INSERT INTO evaluations VALUES(46553,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(46553,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(46553,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(46553,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(46553,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(46553,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(46553,36,'cp 98','g1f3 a7a6 d1c2 d5c4 e2e3 b8d7 f1c4 b7b5 c4e2 c8b7 e1g1 e7d5 e3e4 d5c3 b2c3 f8e7 c3c4 c7c5 d4d5 e6d5 e4d5 e8g8 f1d1 b5b4 c1f4 e7f6 a1b1 h7h6 e2d3 a6a5 d1e1 f6c3 e1e3 b7a6 f4g3 f8e8 e3e8 d8e8 f3d2 d7e5 d2e4 a6c4 d3c4 e5c4 e4c3 b4c3 c2c3 e8e4 b1e1');
INSERT INTO evaluations VALUES(46552,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(46552,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(46552,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(46552,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(46552,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(46552,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(46552,36,'cp 87','a7a6 d1c2 d5c4 e2e3 b8d7 f1c4 b7b5 c4e2 c8b7 e1g1 e7d5 e3e4 d5c3 b2c3 f8e7 c3c4 c7c5 c4b5 e8g8 f1d1 a6b5 e2b5 d7f6 d4c5 b7e4 d1d8 e4c2 d8a8 f8a8 a2a4 e7c5 c1e3 c5e3 f2e3 c2e4 a4a5 g8f8');
INSERT INTO evaluations VALUES(48410,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(48410,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48410,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48410,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48410,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48410,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(48410,36,'cp 0','g1f3 c7c5 e1g1 f8e7 c1b2 b8c6 f3e5 c6e5 b2e5 e8g8 d2d3 f6d7 e5b2 d8c7 a2a4 b7b6 e2e4 c8b7 d1e2 d5d4 f2f4 a7a6 b2c1 b6b5 b1d2 a8c8 h2h4 c8b8 d2f3 b8c8');
INSERT INTO evaluations VALUES(50075,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(50075,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(50075,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50075,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(50075,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(50075,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50075,36,'cp 38','b7b6 g2g3 c8b7 b2b3 e7e6 f1g2 f8b4 c1b2 e8g8 e1g1 b4c3 b2c3 c7c5 b3b4 c5b4 c3b4 d7d6 a2a4 b8c6 d1b3 a8b8 d2d3 c6b4 b3b4 a7a5 b4b5 e6e5 a1b1 d8c8 b5b3 h7h6 b3b6 b7f3 b6b8 c8b8 b1b8 f8b8 g2f3 b8b2 f1a1 e5e4');
INSERT INTO evaluations VALUES(50389,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(50389,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50389,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(50389,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(50389,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50389,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50389,36,'cp 0','d2d4 d7d5 g1f3 e7e6 c2c4 d5c4 e2e3 b7b5 a3a4 c7c6 a4b5 c6b5 b2b3 f8e7 b3c4 b5c4 f1c4 e8g8 e1g1 a7a5 c1a3 b8c6 b1c3 c8b7 d1e2 e7a3 a1a3 c6b4 f1c1 a8c8 f3d2 d8c7 h2h3 c7c6 e2f3 c6c7');
INSERT INTO evaluations VALUES(52318,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(52318,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(52318,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(52318,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(52318,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(52318,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(52318,36,'cp 7','b8c6 c2c4 f8g7 b1c3 d7d6 a2a3 e7e6 a1b1 g8e7 b2b4 e8g8 c3a4 b7b6 c1b2 a8b8 b2g7 g8g7 e1g1 c5b4 a3b4 b6b5 c4b5 b8b5 a4c3 b5b4 b1b4 c6b4 d1a1 a7a5 c3d5 e6e5 d5e7 d8e7 a1a5 b4c6 a5c3 c8d7 f1b1 f8c8 c3e3 h7h6 h2h3 d6d5 g3g4 e7d6 g4g5 h6g5 f3g5');
INSERT INTO evaluations VALUES(52836,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(52836,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(52836,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(52836,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(52836,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(52836,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(52836,36,'cp -12','e7e5 g1f3 f8d6 e1g1 g8f6 b1c3 a7a5 e2e4 d5e4 c3e4 f6e4 d3e4 e8g8 c1e3 d8c7 d1d3 f8e8 a2a4 b8a6 f1d1 d6c5 d3e2 c8g4 h2h3 g4e6 f3g5 e6c8 g2f1 h7h6 g5f3 c8e6 b2b3 c5b4 e3d2 a8d8 d2b4 a6b4 d1d8 e8d8 a1d1 d8d1');
INSERT INTO evaluations VALUES(53823,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(53823,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(53823,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(53823,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(53823,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(53823,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(53823,36,'cp 36','g1f3 g8f6 e1g1 f8e7 b1c3 e8g8 d1c2 c8g4 f1d1 c6c5 h2h3 g4e6 f3g5 b8c6 g5e6 f7e6 d2d3 d8d7 d1f1 g8h8 a2a3 h7h6 e2e3 d6d5 a1b1 a8b8 c1d2 b8d8 b2b3 a7a6 d2c1 h8g8 c3a4 d7d6 c2e2 b7b5 a4c3 d6d7 h3h4 d8b8');
INSERT INTO evaluations VALUES(57176,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(57176,31,'cp -147',NULL);
INSERT INTO evaluations VALUES(57176,32,'cp -132',NULL);
INSERT INTO evaluations VALUES(57176,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(57176,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(57176,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(57176,36,'cp -110','e5e4 f3d4 c7c5 d4f5 d7d5 d2d3 d5d4 f5g3 c8g4 g3e4 b8c6 c2c3 d8d7 b1d2 f8e7 h1g1 c6e5 d1b3 g4e6 c3c4 f7f5 e4g5 e7g5 g1g5 g8f6 b3b5 d7b5 c4b5 e8g8 b2b3 f8e8 c1b2 b7b6 f1g2 h7h6 g2a8 h6g5 a8g2 g5g4 d2c4');
INSERT INTO evaluations VALUES(348,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(348,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(348,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(348,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(348,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(348,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(348,36,'cp 98','b1c3 g8f6 e4e5 c6f3 g2f3 f6h5 c3d5 e8d8 c1g5 f7f6 e5f6 g7f6 d5f6 h7h6 h2h4 c5e7 f6h5 h6g5 h4g5 e7g5 f1d3 c7c6 e1e2 h8e8 e2f1 e8f8 h1g1 g5e7 f3f4 d7d5 a1e1 e7c5 g1g5 c8h3 f1e2 d8c7 e2f3 h3d7 d3g6 f8h8 g6f5 a8e8 e1e5 h8f8 e5e8 d7e8 h5g7 c7d6 f5g4 e8f7 g7f5 d6e6 f5h6 e6f6 h6f7 f8f7 g5g8 f6e7');
INSERT INTO evaluations VALUES(344,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(344,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(344,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(344,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(344,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(344,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(344,36,'cp 101','g8f6 e4e5 c6f3 g2f3 f6h5 c3d5 e8d8 h1g1 c7c6 c1g5 f7f6 e5f6 g7f6 d5f6 h7h6 g5h4 c5e7 f6h5 e7h4 f1h3 d8c7 g1g4 h4g5 f3f4 d7d5 g4g3 c8h3 g3h3 g5e7 e1e2 e7d6 a1g1 a8f8 h3h4 a7a5 g1g6 a5a4 e2d3 f8f7 h4g4 a4a3 b2b3 d6c5 h5g7 c5f2 f4f5 c7b6 h2h4');
INSERT INTO evaluations VALUES(3364,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(3364,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(3364,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(3364,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(3364,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(3364,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(3364,36,'cp -23','c6a5 c4b3 c7c6 d3d4 h7h6 g5h4 f6h5 h4e7 d8e7 h2h3 b7b6 f1e1 e7f6 f3e5 d6e5 d1h5 e5d4 e4e5 f6f5 h5f5 c8f5 c3e4 f8d8 g2g4 f5e4 e1e4 c6c5 e5e6 a5b3 a2b3 f7e6 e4e6 d4d3 c2d3 d8d3 b3b4 c5b4 e6b6 d3b3 a1a4 a7a5 b6b5 b3b2 a4a5 a8a5 b5a5');
INSERT INTO evaluations VALUES(5052,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5052,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(5052,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5052,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(5052,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(5052,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5052,36,'cp 0','b1c3 a8b8 e3d4 g8f6 e4e5 f6d5 f1c4 d5c3 d4c3 d7d5 e5d6 g7c3 b2c3 d8d6 d1d6 e7d6 e1c1 d6d5 h1e1 c8e6 c4e2 e8e7 d1d4 h8e8 d4a4 e7d6 a4a7 b8a8 a7a8 e8a8 c1b2 a8b8 b2c1');
INSERT INTO evaluations VALUES(7037,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(7037,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(7037,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(7037,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(7037,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(7037,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(7037,36,'cp 80','d7d6 d1d2 g8f6 b1c3 f6g4 e3g5 a7a6 h2h3 h7h6 g5h4 g4f6 e1c1 b7b5 f1d3 f6d7 h1e1 e8g8 c3d5 g6g5 f3g5 h6g5 d2g5 f7f6 g5h5 e7e5 d4e5 c6e5 f2f4 e5d3 d1d3 c7c6 c1b1 c8b7 d5c3 d8e7 h5d1 a8d8 e4e5 d6e5');
INSERT INTO evaluations VALUES(5703,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(5703,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(5703,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(5703,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(5703,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(5703,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(5703,36,'cp 95','b1c3 g8f6 d1d2 e8g8 d4d5 c6b8 e3h6 c7c6 e1c1 c8g4 h2h4 c6d5 e4d5 b8d7 h6g7 g8g7 d1e1 h7h5 f3d4 d7b6 f2f3 g4d7 f3f4 b6d5 c3d5 f6d5 f4f5 d5f6 f1d3 f6g4 d3e2 g4f2 h1f1 f2e4 d2e3 d6d5 f5g6 f7g6 e2d3 f8f1');
INSERT INTO evaluations VALUES(8543,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(8543,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(8543,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(8543,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(8543,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8543,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8543,36,'cp -7','b1c3 e8g8 e1g1 h7h6 g5f6 e7f6 c3d5 a7a6 b5a4 c8e6 a4b3 c6a5 d5f6 d8f6 b3e6 f7e6 d1d2 a5c6 c2c3 f6g6 g1h1 f8f7 f3g1 a8f8 f2f3 c6e7 d3d4 g6f6 d4e5 f6e5 a1e1 e7g6 g1h3 e5f6');
INSERT INTO evaluations VALUES(8542,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(8542,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(8542,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(8542,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8542,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(8542,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8542,36,'cp -5','e8g8 e1g1 h7h6 g5f6 e7f6 c3d5 a7a6 b5a4 c8e6 d5e3 c6e7 f1e1 c7c6 a4b3 e7g6 b3e6 f7e6 g2g3 d8d7 h2h4 f8f7 f3h2 a8f8 d1g4 g8h7 h4h5 g6h8 e1f1 f6g5 g4e2 g5e3 e2e3 c6c5 a2a4 b7b5 a4b5 a6b5 c2c3');
INSERT INTO evaluations VALUES(8730,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(8730,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(8730,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(8730,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(8730,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(8730,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(8730,36,'cp 80','h2h3 c6b4 e1g1 e8g8 b1c3 b7b6 a2a3 b4d3 c2d3 c7c5 d4c5 b6c5 f4e5 c8e6 c3a4 c5c4 f3d4 e6d7 a4c5 a8c8 b2b4 c4b3 d4b3 d7b5 a1c1 f6d7 e5g7 g8g7 a3a4 d7c5 c1c5 c8c5 b3c5 d8d6 d1a1 g7g8 f1c1 f8c8 d3d4 b5c4 c1d1 a7a5 d1c1 e7e5');
INSERT INTO evaluations VALUES(8813,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(8813,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(8813,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(8813,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(8813,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(8813,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(8813,36,'cp 4','c5b6 b1d2 e8g8 a2a4 a7a5 b4b5 c6e7 h2h3 e7g6 f1e1 f8e8 d3d4 e5d4 c3d4 d6d5 e4d5 e8e1 d1e1 f6d5 d2b3 c8e6 c1d2 g6h4 f3h4 d8h4 d2a5 b6a5 b3a5 h4d4 a1c1 d4b6 c4d5 e6d5 a5c4 b6e6 e1e6 f7e6 c4e3 c7c6 e3d5 e6d5 b5c6 b7c6 c1c5 a8a4');
INSERT INTO evaluations VALUES(9798,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9798,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9798,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9798,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9798,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9798,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9798,36,'cp 0','d7d5 e4d5 f6d5 e1g1 e8g8 f3e5 c6e5 d3d4 d5b4 d4e5 b4d3 b1a3 f8e8 a3b5 c8f5 e5e6 f5e6 b5c7 d3f2 d1d8 a8d8 c7e6 e8e6 b3e6 f2d3 g1h1 d3f2');
INSERT INTO evaluations VALUES(10188,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(10188,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(10188,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(10188,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(10188,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(10188,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(10188,36,'cp 30','e1g1 d7d6 b1c3 e8g8 e2e3 f8e8 d2d3 c5b4 c3e2 a7a5 a2a3 b4c5 b2b3 c5b6 d3d4 h7h6 c1b2 c8f5 h2h3 d8e7 a1c1 f5h7 f1e1 e5e4 f3h2 c6b8 e2f4 c7c6 h2g4 b8d7 g4f6 d7f6 g1h2');
INSERT INTO evaluations VALUES(11002,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(11002,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(11002,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(11002,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(11002,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(11002,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(11002,36,'cp 50','a7a6 b5c4 b7b5 c4b3 e8g8 d4d5 c6a5 c1d2 a5b3 d1b3 c7c5 d2c3 g7c3 b1c3 d7d6 h2h3 f7f6 a2a4 b5b4 c3e2 g8h8 a1d1 g6g5 f3d2 e7g6 d2c4 a6a5 e2g3 c8a6 g3f5 a6c4 b3c4 g6e5 c4e2 c5c4 e2e3 a8c8 e3d4 c4c3 b2c3 c8c3');
INSERT INTO evaluations VALUES(11457,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(11457,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(11457,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(11457,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11457,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11457,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11457,36,'cp 0','c8g4 d2d3 f8d6 h2h3 g4h5 b1d2 f6d7 d2c4 f7f6 c2c3 e8g8 d3d4 a6a5 a2a4 f8e8 c4d6 c7d6 d1b3 d6d5 b3b7 d5e4 f3h4 c6c5 c1e3 a8b8 b7a7 b8a8');
INSERT INTO evaluations VALUES(13067,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(13067,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(13067,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(13067,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(13067,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(13067,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(13067,36,'cp 146','g1f3 d8b6 b5a4 c8g4 e1g1 e7e6 d2d4 b6a6 b1c3 f8b4 d1e2 a6a5 a4c6 b7c6 c3e4 a5d5 e4g3 b4e7 f1d1 g4f3 g2f3 d5d6 b2b3 e8g8 c2c4 a7a5 a2a4 d6c7 d1d3 f8d8 c1e3 f6e8 a1d1 a8b8 f3f4 g7g6 h2h3 e8g7 e2f3 c7b7 e3d2');
INSERT INTO evaluations VALUES(13391,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13391,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13391,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13391,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13391,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13391,36,'cp 0','g1f3 g7g6 f1e2 f8g7 e1g1 e8g8 f1d1 c8g4 c3d5 f6d7 c2c3 a8c8 h2h3 g4f3 e3f3 e7e6 d5e3 c6e5 f3f4 d8c7 f4h4 g7f6 h4f4');
INSERT INTO evaluations VALUES(13717,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13717,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13717,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(13717,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13717,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(13717,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13717,36,'cp -8','e1g1 f8e7 d2d4 e8g8 d4c5 e7c5 c1g5 c8g4 g5f6 d8f6 d1d5 g4f3 d5f3 f6f3 g2f3 f8d8 b5c6 b7c6 a2a4 d8d3 g1g2 a8e8 b1a3 e8e5 f3f4 e5f5 f1d1 d3d1 a1d1 c5a3 b2a3 g7g6 d1d4 f5c5 c3c4 g8g7 g2g3 c5a5 d4d6 a5a4 d6c6 a4a3 g3g4 a3c3 c6a6 h7h5 g4g5');
INSERT INTO evaluations VALUES(13716,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(13716,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(13716,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(13716,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(13716,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(13716,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(13716,36,'cp -6','f8d6 d2d4 e8g8 d4c5 d6c5 c1g5 c8g4 g5f6 d8f6 d1d5 g4f3 d5f3 f6f3 g2f3 f8d8 b5c6 b7c6 a2a4 d8d3 g1g2 a8e8 b1a3 e8e5 h2h4 e5f5 a1d1 d3d1 f1d1 c5a3 b2a3 g7g6 d1d4 f5c5 c3c4 g8g7 g2g3 c5a5 d4d6 a5a4 d6c6 a4a3 c4c5 a3c3 c6c7 a7a5 c5c6 h7h5 g3f4');
INSERT INTO evaluations VALUES(14482,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(14482,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(14482,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(14482,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(14482,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(14482,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(14482,36,'cp 55','c2c4 g7g6 d4d5 c6b8 b1c3 f8g7 e3e4 e8g8 f1d3 c8g4 h2h3 g4f3 d1f3 b8a6 e1g1 f6d7 d3c2 a6b4 c2b1 a7a5 f3e2 e7e6 f2f4 b4a6 f1f3 a6c5 b1c2 g7c3 f3c3 e6e5 f4f5 d8h4 g1h1 g8g7 c1d2');
INSERT INTO evaluations VALUES(14470,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(14470,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(14470,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(14470,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(14470,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14470,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(14470,36,'cp 43','g7g6 b1c3 f8g7 d4d5 c6b8 e3e4 e8g8 f1d3 c7c6 e1g1 c6d5 e4d5 e7e6 d5e6 c8e6 h2h3 d6d5 f3d4 e6c8 c1e3 b8c6 a1c1 c6d4 e3d4 d5c4 d3c4 a7a6 d4c5 d8d1 f1d1 f8e8 c3a4 b7b5 a4b6 a8b8 b6c8 b8c8 c4b3 f6e4 c5e3 c8c1');
INSERT INTO evaluations VALUES(15658,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(15658,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(15658,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(15658,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(15658,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(15658,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(15658,36,'cp 15','b1c3 f8e7 c1d2 e8g8 e1c1 a7a6 c3d5 f6g4 e3e1 f8e8 c1b1 g4f6 f1c4 c8e6 e1e3 e7f8 d5f6 d8f6 d2c3 f6e7 c4e6 e7e6 f3g5 e6d7 g5h3 b7b5 b2b3 a6a5 h3f4 b5b4 c3b2 a5a4 f4d5 a4b3');
INSERT INTO evaluations VALUES(15652,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(15652,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(15652,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(15652,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(15652,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15652,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(15652,36,'cp 22','g7g6 b2b3 f6g4 e3e2 f8g7 c1b2 e8g8 e1c1 a7a5 h2h3 g4f6 e4e5 f6e8 e2e3 c8e6 h3h4 c6e5 f3e5 g7e5 h4h5 g6g5 g2g3 h7h6 f2f4 e5g7 f1d3 c7c6 f4g5 d8g5 e3g5 h6g5 h5h6 g7e5 c3e4 f7f6 b2e5 d6e5 e4c5 e6d5 h1f1 b7b5 c5d7 f8f7 d7b6 a8d8 b6d5 d8d5');
INSERT INTO evaluations VALUES(15810,30,'cp 366',NULL);
INSERT INTO evaluations VALUES(15810,31,'cp 348',NULL);
INSERT INTO evaluations VALUES(15810,32,'cp 344',NULL);
INSERT INTO evaluations VALUES(15810,33,'cp 372',NULL);
INSERT INTO evaluations VALUES(15810,34,'cp 390',NULL);
INSERT INTO evaluations VALUES(15810,35,'cp 367',NULL);
INSERT INTO evaluations VALUES(15810,36,'cp 372','f3e5 f8c5 c4b3 c5b6 d2d4 e8g8 c1g5 c7c6 e1g1 c6d5 a2a3 b4c6 c3d5 c6d4 g5f6 g7f6 e5g4 c8g4 d1g4 g8h8 a1d1 f8g8 g4h3 g8g5 g1h1 d4b3 h3b3 d8d6 f2f4 g5g8 c2c4 d6e6 b3f3 a8e8 b2b3 e6c6 b3b4 c6c4 d5f6 e8e3 f3b7 g8g6');
INSERT INTO evaluations VALUES(15792,30,'cp 373',NULL);
INSERT INTO evaluations VALUES(15792,31,'cp 378',NULL);
INSERT INTO evaluations VALUES(15792,32,'cp 359',NULL);
INSERT INTO evaluations VALUES(15792,33,'cp 348',NULL);
INSERT INTO evaluations VALUES(15792,34,'cp 369',NULL);
INSERT INTO evaluations VALUES(15792,35,'cp 370',NULL);
INSERT INTO evaluations VALUES(15792,36,'cp 385','f8d6 d2d4 e8g8 a2a3 c7c5 d4c5 b4c2 d1c2 d6e5 e1g1 d8c7 b2b4 e5h2 g1h1 c7e5 c2e2 f8e8 e2e5 h2e5 c1d2 c8d7 a1c1 a8c8 f2f3 g8f8 g2g4 c8d8 f1d1 a7a6 d1e1 h7h6 c1d1 g7g5 h1g2 h6h5 g4h5 f6h5 d5d6 g5g4 c3d5 g4f3 g2f3 d7c6 d2h6 e5g7 h6g7 h5g7 f3f2 e8e1 d1e1 g7e6 d5e7');
INSERT INTO evaluations VALUES(16975,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(16975,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(16975,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(16975,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(16975,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(16975,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(16975,36,'cp 62','d7d6 d2d4 f8e7 c1e3 e8g8 h2h3 c8e6 c4e6 f7e6 d4e5 d6e5 b1d2 d8d3 f3d1 f6e4 d2e4 d3e4 d1d7 e7d6 d7e6 g8h8 f1d1 b7b6 d1d4 e4c2 d4d2 c2e4 c3c4 h7h6 a1d1 e4h4 c4c5 b6c5 e6d5 h4e7 d1c1 e7f6 c1c4 a8d8 d5e4 f6e6 d2d5 a7a5 e3c5 d6c5 d5c5 d8d1 g1h2 f8f2');
INSERT INTO evaluations VALUES(17368,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(17368,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(17368,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(17368,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(17368,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(17368,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(17368,36,'cp -15','c4a2 d7d5 b1c3 g8f6 g1f3 f8e7 e1g1 e8g8 c1f4 b7b5 f1e1 d5e4 c3e4 c8b7 c2c3 f6e4 d3e4 d8d1 a1d1 f8d8 h2h3 g8f8 a2b1 d8d1 e1d1 a8d8 b1c2 f7f6 f4e3 d8d1 c2d1 c5c4 a3a4 e6e5 f3d2 b7c8 a4b5 a6b5 g1f1 c8e6 f1e2');
INSERT INTO evaluations VALUES(19999,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(19999,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(19999,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(19999,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(19999,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(19999,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(19999,36,'cp 6','e7e5 e2c4 h7h6 e1g1 g8f6 d2d3 f8e7 a2a4 e8g8 c1d2 c8e6 c3d5 d8d7 c2c3 e7d8 f3e1 c6a5 d5f6 d8f6 c4e6 d7e6 c3c4 a5c6 e1c2 f6d8 c2e3 d8g5 e3f5 g5d2 d1d2 a8d8 a1e1 e6g6 f5e3 g8h7 f2f3 d8d7 a4a5 c6d4');
INSERT INTO evaluations VALUES(22291,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(22291,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(22291,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(22291,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(22291,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(22291,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(22291,36,'cp 90','c6e7 c4b3 h7h5 h2h3 g7g6 c2c4 f8h6 b1c3 e8f8 f3d2 f8g7 b3c2 c7c5 b2b4 c5b4 c3e2 c8d7 a2a3 a7a5 e1g1 b7b5 c2d3 b5c4 d2c4 h6c1 d1c1 d7b5 a3b4 a5b4 a1b1 a8c8 b1b4 b5c4 d3c4 d8c7 c4d3 c7c1 f1c1');
INSERT INTO evaluations VALUES(20796,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(20796,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(20796,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(20796,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(20796,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(20796,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(20796,36,'cp 124','c4b5 e8f7 b1c3 h7h5 e1g1 g7g5 a2a4 a7a6 b5d3 e7g6 b2b4 g5g4 f3e1 f8h6 a4a5 g6f4 c3e2 g8e7 e2f4 e5f4 f2f3 e7g6 d3e2 d8e7 a1a3 h8e8 e1d3 c8d7 c2c4 f6f5 f3g4 h5g4 e4f5 e7e2 f5g6 f7g6 c1f4 e2d1 f1d1 h6f4 d3f4 g6f5 g2g3 e8e4');
INSERT INTO evaluations VALUES(22967,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22967,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(22967,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(22967,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22967,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22967,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22967,36,'cp 0','g8e7 c1f4 d7d5 e4a4 e7g6 e1c1 g6f4 a4f4 f8d6 f4d2 c8e6 e2e3 d8f6 c3d5 e6d5 d2d5 e8c8 d5f3 f6e6 f3d5');
INSERT INTO evaluations VALUES(23189,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(23189,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(23189,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(23189,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(23189,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(23189,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(23189,36,'cp 74','f1d3 g7g6 g1f3 f8g7 d1e2 f6e5 d4e5 d7d6 e5d6 c7d6 c2c4 g8e7 b1c3 e8g8 c1g5 h7h6 g5e7 c6e7 e1g1 d8b6 f1e1 g7c3 b2c3 e6e5 f3d2 g8g7 a1d1 c8f5 d3f5 g6f5 d2b3 e7g6 c4c5 d6c5 d1d7 f8f7 d7f7 g7f7 e2h5 b6f6 b3c5 f7g7 c5b7');
INSERT INTO evaluations VALUES(24651,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(24651,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(24651,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(24651,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24651,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(24651,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(24651,36,'cp 45','f1e2 e7e5 d2d3 f8b4 c1d2 d8e7 e1g1 a8d8 d1c2 e8g8 b1c3 g4f3 e2f3 c6d4 c2d1 e7d7 f3d5 b4c3 b2c3 d4e6 d5e6 f7e6 d2e3 d7d3 d1b3 b7b6 h2h3 h7h6 a1d1 d3e4 c4c5 f6d5 d1e1 e4d3 c5b6 a7b6 c3c4 d5e3 e1e3 d3d4 b3c2 g8h8 f1e1 d4d2');
INSERT INTO evaluations VALUES(24652,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(24652,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(24652,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24652,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24652,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24652,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(24652,36,'cp 37','e7e5 d2d3 f8b4 c1d2 b4d2 b1d2 g4f5 d1c2 e8g8 e1g1 c6b4 c2c3 f5d3 c3b4 d3e2 f1e1 e2f3 d2f3 e5e4 f3d2 f8e8 d2f1 a7a5 b4c3 d8b6 a1d1 b6b4 c3c2 a8d8 d1d8 e8d8 e1d1 d8d1 c2d1 h7h5 b2b3 g7g6 h2h3 g8g7 f1e3 h5h4');
INSERT INTO evaluations VALUES(25276,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(25276,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(25276,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(25276,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(25276,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(25276,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(25276,36,'cp 91','d2d4 g4d7 b5d3 h7h6 b1d2 f8e7 a2a3 a7a5 d2c4 b7b5 c4e3 e8g8 b2b3 f8e8 d1c2 e7f8 c1b2 b5b4 c3b4 a5b4 d4d5 c6e7 a3b4 c7c6 e3c4 a8a1 f1a1 d7g4 c4e3 c6d5 e3g4 f6g4 d3b5 d5e4 b5e8 e4f3 e8a4 f3g2');
INSERT INTO evaluations VALUES(26777,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(26777,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(26777,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(26777,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(26777,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(26777,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(26777,36,'cp 99','d1d3 g8f6 b1c3 f8e7 f2f3 f6d7 c1e3 e7f6 e1c1 f6d4 e3d4 d8e7 h2h4 d7e5 d3e3 e8g8 g2g4 b7b5 h1g1 a7a5 h4h5 f7f6 h5h6 g7g5 g1e1 b5b4 c3d5 c6d5 e4d5 e7f7 d4e5 f8e8 e3d2 e8e5 e1e5 d6e5 c1b1 a8d8 b2b3');
INSERT INTO evaluations VALUES(26842,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(26842,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(26842,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(26842,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(26842,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(26842,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(26842,36,'cp 65','e1g1 a8b8 b1d2 g8f6 f1e1 f8d6 e2e4 d5e4 d2e4 f6e4 e1e4 g4h5 c2c3 e8g8 b2b4 h7h6 d1b3 d8d7 e4e1 f8e8 c1e3 b7b6 f3d2 h5g6 d2c4 b6b5 c4d6 c7d6 a2a4 a7a6 d4d5 e6d5 a4b5 a6b5 b3d5 c6e7 d5d1 e7f5 d1d2 f5e3 e1e3 e8e3');
INSERT INTO evaluations VALUES(26841,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(26841,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(26841,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(26841,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(26841,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(26841,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(26841,36,'cp 56','a8b8 f1e1 g8f6 c2c3 f8d6 b1d2 h7h6 e2e4 f6e4 d2e4 d5e4 e1e4 g4f5 e4e1 e8g8 b2b4 c6e7 a2a4 c7c6 c1d2 b7b6 c3c4 a7a5 c4c5 b6c5 d4c5 d6c7 b4a5 c7a5 d2a5 d8a5 d1d6 a5a7 f3d4 f8d8 d6e5 f7f6 e5f4 e7d5 d4c6 d5f4 c6a7 f4g2 g1g2');
INSERT INTO evaluations VALUES(28168,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(28168,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(28168,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(28168,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(28168,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(28168,35,'cp 145',NULL);
INSERT INTO evaluations VALUES(28168,36,'cp 150','g1e2 c7c5 a2a3 b4c3 b2c3 g8f6 f2f3 e8g8 e3e4 f5g6 e1g1 b8c6 c4a2 d8e7 d1b3 a8c8 c1e3 b7b6 b3a4 e6e5 d4c5 b6c5 e2g3 c6d8 a1b1 d8e6 a4a6 c8c7 a2c4 h7h5 f1d1 e6f4 a6d6 h5h4 d6e7 c7e7 g3f1 e7c7 b1b5 f8c8 d1d6');
INSERT INTO evaluations VALUES(30091,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(30091,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30091,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(30091,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(30091,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(30091,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(30091,36,'cp 34','d3f5 e6f5 d1d3 d8d7 g1e2 b8a6 d3b3 c7c6 b1d2 f8e7 c2c4 e8g8 e1g1 d5c4 d2c4 a6b4 a1c1 f8e8 f1d1 b4d5 c4e5 d7c8 f4g5 f6e4 g5e7 e8e7 e2c3 e4f6 a2a4 c8e6 a4a5 g7g6 a5a6 a8b8 e5d3 g8g7 c3d5 e6d5 b3d5 f6d5 a6b7 e7b7');
INSERT INTO evaluations VALUES(30057,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(30057,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(30057,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(30057,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30057,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(30057,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(30057,36,'cp 30','e6f5 d1d3 d8d7 g1e2 b8a6 e1g1 f8e7 a2a3 c7c6 b1d2 f6e4 a1d1 e8g8 c2c4 e4d2 d1d2 d5c4 d3c4 a6c5 e2c3 d7e6 d4d5 c6d5 c3d5 f8d8 d5e7 e6e7 f4d6 d8d6 c4c5 d6d7 c5e7 d7e7 f1c1 g7g6 g2g3 a7a5 g1g2 g8g7 h2h4 e7e4 d2d5 a5a4 c1c7');
INSERT INTO evaluations VALUES(30232,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30232,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(30232,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(30232,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(30232,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(30232,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(30232,36,'cp 28','d1d3 d8d7 g1e2 b8a6 b1d2 f8e7 c2c3 e8g8 h2h3 g7g6 f2f3 f6h5 e1c1 c7c5 g2g4 h5f4 e2f4 e7g5 g4f5 g5f4 e3f4 c5d4 f5g6 f7g6 d3d4 a6c7 d2c4 d7f5 c4d6 f5f4 d4f4 f8f4 d6b7 f4f3 h3h4 a8e8 d1d2 e8e5 b7a5 f3f6 c1b1 e5f5 a5b3');
INSERT INTO evaluations VALUES(31114,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(31114,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(31114,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(31114,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(31114,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(31114,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(31114,36,'cp 36','e7e6 c2c4 b8d7 b1d2 g8f6 f1e2 f6h5 f4g5 f8e7 g5e7 d8e7 d1b3 a8b8 h2h3 g4f3 e2f3 h5f6 a1c1 e8g8 e1g1 a7a5 b3c3 b8a8 b2b3 f8e8 c1c2 h7h5 c2c1 e6e5 c4d5 f6d5 f3d5 c6d5 d4e5 d7e5 c3d4 a8d8 f1d1');
INSERT INTO evaluations VALUES(31468,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(31468,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(31468,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(31468,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(31468,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(31468,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(31468,36,'cp 38','h2h3 g4h5 c2c4 f8d6 f4d6 d8d6 d1b3 d6c7 c4d5 e6d5 b1c3 b8d7 g2g4 h5g6 a1c1 c7b6 f1e2 g8e7 f3h4 f7f6 f2f3 b6b3 a2b3 a7a5 e1f2 e8f7 h4g6 e7g6 h3h4 h7h5 g4g5 g6e7 e3e4 f6g5 h4g5 d5e4 e2c4 f7g6');
INSERT INTO evaluations VALUES(32035,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(32035,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(32035,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(32035,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(32035,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(32035,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(32035,36,'cp 117','g4f3 d1f3 d8d7 c3d5 c7c6 d5e3 g7g6 b2b4 f8g7 c1b2 d7e7 a2a4 g8f6 b4b5 b8d7 b2a3 d7f8 e1g1 f8e6 c4e6 e7e6 f1b1 e8g8 a4a5 c6c5 d2d3 h6h5 c2c4 g8h8 a3c1 h8h7 c1d2 g7h6 e3d5 f6d5 e4d5');
INSERT INTO evaluations VALUES(32046,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(32046,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(32046,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(32046,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(32046,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(32046,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(32046,36,'cp 118','d1f3 g8f6 a2a3 g7g6 d2d3 f8g7 c1e3 b8c6 c3e2 d8d7 c2c3 a7a5 c4a2 a8b8 d3d4 a5a4 e1g1 c6a5 c3c4 e8g8 c4c5 a5b3 a2b3 a4b3 c5d6 c7d6 e2c3 d7e7 a1d1 b7b5 d4e5 d6e5 c3d5 f6d5 d1d5');
INSERT INTO evaluations VALUES(32045,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(32045,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(32045,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(32045,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(32045,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(32045,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(32045,36,'cp 109','g8f6 a2a3 g7g6 d2d3 f8g7 c3e2 b8c6 c4a2 a8b8 c1e3 d8d7 e1g1 a7a5 c2c3 c6d8 f3g3 b7b5 g3h4 f6h5 d3d4 d8e6 d4e5 d6e5 a1d1 d7e7 h4e7 e8e7 e2c1 h8d8 c1b3 d8d1 f1d1 b8a8 b3c5 a8d8 d1d8 e6d8 g2g3 h5f6 b2b4 a5b4 c3b4 f6d7 h3h4');
INSERT INTO evaluations VALUES(29843,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(29843,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(29843,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(29843,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(29843,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(29843,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(29843,36,'cp 139','a2a3 b8c6 c3e2 g7g6 d2d3 f8g7 c4a2 d8d7 c1e3 a8b8 c2c3 a7a5 e1g1 c6d8 g2g4 b7b5 d3d4 d8e6 f3g2 b5b4 c3b4 a5b4 a2e6 d7e6 d4d5 e6c8 f1c1 b4a3 b2a3 c8a6 g2f3 e8g8 c1c6 a6d3 e2g3 b8b1 a1b1 d3b1 g1g2 b1b7 a3a4 f8a8');
INSERT INTO evaluations VALUES(32752,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(32752,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(32752,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(32752,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(32752,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(32752,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(32752,36,'cp -85','c5d4 f3d4 b8c6 d4c6 d7c6 c2c3 f8c5 e1g1 g8e7 b1d2 e8g8 d2f3 h7h6 f1e1 f8c8 c1e3 a7a5 e3c5 c6c5 d1d4 b7b5 a2a3 c5c7 d4d2 a8b8 a1c1 e7f5 g2g4 f5e7 h2h4 c7b6 f3d4 e7c6 g1g2 c6d4 c3d4');
INSERT INTO evaluations VALUES(34270,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(34270,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(34270,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(34270,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(34270,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(34270,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(34270,36,'cp 107','g8f6 c4b3 d7d6 d2d4 c8e6 e1g1 b8d7 h2h3 e6b3 a2b3 e7d8 d1e2 d8c7 f1d1 a7a6 b3b4 a8d8 c1g5 f8e7 d4d5 c6d5 g5f6 d7f6 c3d5 f6d5 d1d5 e8g8 c2c3 g7g6 f3d2 d8c8 b4b5 a6b5 e2b5 c8a8 a1d1 c7c6 d2c4 c6b5 d5b5');
INSERT INTO evaluations VALUES(34593,30,'cp -568',NULL);
INSERT INTO evaluations VALUES(34593,31,'cp -572',NULL);
INSERT INTO evaluations VALUES(34593,32,'cp -597',NULL);
INSERT INTO evaluations VALUES(34593,33,'cp -574',NULL);
INSERT INTO evaluations VALUES(34593,34,'cp -586',NULL);
INSERT INTO evaluations VALUES(34593,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(34593,36,'cp -594','d5e5 g1e2 e5a1 e2c3 e7e5 d1f3 b8c6 f1d3 a7a6 e1g1 c8e6 b1a3 f8b4 c3a4 b4a3 c1a3 a1d4 a4c5 e6g4 c2c3 g4f3 c3d4 f3d5 c5b7 c6d4 b7c5 g8e7 f2f4 e7c6 f4e5 e8g8 a3b2 a6a5 f1e1 d4e6 c5e6 f7e6 d3b5 c6b4 a2a3 f8d8 g1h1 b4c2 e1f1 g8h8 b2c3 c2a3');
INSERT INTO evaluations VALUES(39308,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(39308,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(39308,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(39308,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(39308,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(39308,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(39308,36,'cp 20','e2e3 c7c6 d2d4 e5d4 e3d4 c5e7 d5c3 g8f6 g1f3 d7d5 f1d3 d5c4 d3c4 b8d7 e1g1 d7b6 c4b3 b6d5 f1e1 e8g8 d1d3 c8e6 c1g5 a7a5 f3e5 d5c3 b2c3 e6b3 a2b3 f8e8 h2h3 h7h6 g5f4 f6d5 f4d2 e7d6 c3c4 d6e5');
INSERT INTO evaluations VALUES(41626,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41626,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(41626,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(41626,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41626,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41626,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41626,36,'cp 0','c7c6 g2g4 a7a5 a2a4 d8b6 d1e2 b8a6 g4g5 h6g5 f3g5 c8g4 c4f7 e8e7 f2f3 g4d7 c1d2 a8f8 f7c4 f6h5 e1c1 c5d4 e2h2 h5f4 c1b1 h8h6 h3h4 f8h8 d2e1 h6h5 g5f7 h8f8 f7g5');
INSERT INTO evaluations VALUES(42499,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(42499,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(42499,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(42499,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(42499,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(42499,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(42499,36,'cp 53','b2c3 b7b6 f1d3 c8b7 f2f3 d7d6 e3e4 f6d7 g1h3 d8h4 h3f2 h4e7 e1g1 b8c6 c1a3 f7f6 f3f4 e6e5 f4e5 f6e5 c4c5 d6c5 d4d5 c6d8 f2h3 a7a6 d1g4 d7f6 g4g3 f6h5 g3f3 h5f6 c3c4 b7c8 a1e1 c8h3 f3h3 e8g8 a3c1 d8f7 d3c2 f7d6 c1g5 g8h8 c2d3 a8e8 a2a4');
INSERT INTO evaluations VALUES(42495,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(42495,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(42495,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(42495,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(42495,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(42495,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(42495,36,'cp 75','b7b6 f1d3 c8b7 f2f3 d7d6 g1e2 b8c6 e3e4 c6a5 e1g1 f6d7 a1b1 d8e7 c1e3 b7a6 d1a4 e8g8 e2c1 a6b7 c1b3 a5b3 a2b3 a7a5 a4a2 e6e5 b1e1 f8e8 h2h3 f7f6 g1h2 g8h8 a2d2 a5a4 b3b4 e5d4 c3d4 a4a3');
INSERT INTO evaluations VALUES(43250,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(43250,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43250,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(43250,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(43250,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(43250,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43250,36,'cp 32','e2e4 d7d6 f1d3 e8g8 e1g1 c7c5 c2c3 c5d4 c3d4 e7e5 d4d5 a7a5 f1e1 b8a6 d3a6 a8a6 a2a4 f6d7 d2c4 f7f5 c1g5 d8c7 a1c1 d7c5 g5e3 f5e4 f3d2 c5d3 d2e4 d3c1 d1c1 c7b8 e3b6 c8f5 e4d6 a6b6 d6f5 b6f6 f5g7 g8g7 c4e5 f6f2 e5g4');
INSERT INTO evaluations VALUES(43265,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43265,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(43265,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(43265,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(43265,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(43265,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(43265,36,'cp 15','e8g8 f1d3 c7c5 c2c3 d7d6 e1g1 c5d4 c3d4 e7e5 d4d5 a7a5 a2a4 b7b6 f1e1 b8a6 d2c4 a6c5 c1g5 c8a6 a1c1 d8b8 b2b3 f8c8 h2h3 b6b5 a4b5 a6b5 c4d6 b8d6 d3b5 c5e4 g5f6 c8c1 d1c1 e4f6 c1c6 a8d8 f3e5 f6d5 c6d6 d8d6 e5c4 d6e6 e1e6 f7e6 c4a5');
INSERT INTO evaluations VALUES(46854,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46854,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46854,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(46854,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(46854,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46854,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46854,36,'cp 7','e2e3 c7c5 g1e2 h7h6 g5f4 f8g7 d1d2 e7e5 f4e5 g7e5 d4e5 d8d2 e1d2 b8c6 e2g3 c8f5 d2c3 e8c8 f1e2 c6e5 a1d1 b7b6 b2b4 c5b4 c3b4 c8c7 g3f5 g6f5 b4c3 h8g8 g2g3 g8e8 a2a4 e5c6 d1b1 e8e5 b1b5 e5c5 h1b1 d8d5 b5c5');
INSERT INTO evaluations VALUES(46853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46853,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(46853,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(46853,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(46853,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(46853,36,'cp -25','h7h6 g5f4 f8g7 g1e2 g6g5 f4e5 g7e5 d4e5 d8d1 a1d1 b8c6 h2h4 g5g4 e2d4 c6e5 d4b5 a8b8 b5c7 e8f8 f1e2 c8f5 d1d4 f7f6 c7b5 f8f7 b5c3 b8c8 b2b3 a7a6 e1g1 h8g8 g1h2 b7b5 c3d5 e5d3 h2g3 f5e6 e2d3 e4d3');
INSERT INTO evaluations VALUES(50699,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(50699,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50699,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(50699,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50699,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(50699,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(50699,36,'cp 34','c1g5 b8c6 e2e3 c8f5 a2a3 f7f6 g5h4 g8h6 h4g3 e7e6 f1d3 f8e7 g1e2 e8g8 e2f4 f5d3 d1d3 d8d7 c3a4 a8b8 f4e2 b8e8 a1c1 h6f5 e1g1 g7g5 e2c3 f5g3 h2g3 f6f5 b2b4 e7d6 a4c5 d7e7 g1h1');
INSERT INTO evaluations VALUES(50723,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50723,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(50723,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(50723,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(50723,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(50723,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(50723,36,'cp 24','f8e7 f1c4 g8f6 c1g5 b8d7 a2a4 h7h6 g5e3 d8c7 f3d2 d7c5 e1g1 c8e6 c3d5 f6d5 e4d5 e6d7 a4a5 e8g8 c4e2 b7b5 a5b6 c7b6 c2c3 a6a5 d2c4 b6c7 b2b4 a5b4 c3b4 a8a1 d1a1 c5b3 a1b2 b3d4 e3d4 e5d4 b4b5 c7c5 b5b6 d7b5 c4a3 b5e2 b2e2 c5a3 e2e7');
INSERT INTO evaluations VALUES(52090,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(52090,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52090,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52090,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52090,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52090,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52090,36,'cp 0','f8c5 g1f3 b8c6 a2a3 d7d6 c3a4 c8f5 e2e3 a7a6 a4c5 d6c5 d1d8 a8d8 c1d2 g8e7 e1c1 b7b5 d2e1 d8d1 c1d1 e7c8 f3d2 c8d6 d2b3 d6b7 d1c1 e8e7 f1e2 h8c8 b3d2 b7d6 d2b3');
INSERT INTO evaluations VALUES(37708,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37708,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37708,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37708,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37708,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37708,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(37708,36,'cp 15','g1f3 d7d6 a2a3 b8c6 b2b4 c5b6 c3a4 g8f6 e2e3 e8g8 c1b2 a7a5 b4b5 c6b8 c2c4 d8c7 d1c2 b8d7 f1e2 d7c5 a4b6 c7b6 e1g1 h7h6 f3d2 c8e6 a3a4 b6c7 f2f4 e6g4 e2d3 a8c8 h2h3 c5d3 c2d3 g4e6 f1d1 f6d7 d2f3');
INSERT INTO evaluations VALUES(37707,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(37707,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(37707,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37707,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(37707,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(37707,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(37707,36,'cp 2','b8c6 a2a3 d7d6 b2b4 c5b6 c3a4 g8f6 e2e3 e8g8 c1b2 a7a5 b4b5 c6b8 f1e2 b6c7 c2c4 b8d7 a1c1 d8e7 f3d2 f8d8 e1g1 d7c5 a4c5 d6c5 d1c2 c8f5 e3e4 f5g6 f1e1 d8d7 b2c3 a8d8 e2f3 g6h5 f3h5 f6h5 g2g3 a5a4 g1g2');
INSERT INTO evaluations VALUES(5206,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(5206,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(5206,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(5206,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(5206,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(5206,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(5206,36,'cp 9','a2a3 d7d6 b2b4 c5b6 e2e3 g8f6 c3a4 e8g8 c1b2 a7a5 b4b5 c6b8 c2c4 d8c7 a4b6 c7b6 d1c2 b8d7 f1e2 d7c5 e1g1 b6c7 f3d2 c8e6 a3a4 a8c8 f1d1 f8d8 b2a3 f6d7 a1c1 f7f5 d2b3 c5b3 c2b3 d7c5 a3c5 c7c5 b3c2 b7b6 d1d2 c5b4 c1d1');
INSERT INTO evaluations VALUES(52744,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52744,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52744,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52744,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52744,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52744,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52744,36,'cp 0','e7e5 d2d4 e5e4 a2a3 f8e7 b1c3 f7f5 c4d5 c6d5 h2h4 b8c6 g1h3 c8e6 h3g5 e6c8');
INSERT INTO evaluations VALUES(53646,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(53646,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(53646,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(53646,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(53646,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(53646,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(53646,36,'cp 98','g8f6 e1g1 f8d6 d1e2 b8d7 f1e1 e8g8 c3e4 f6e4 e2e4 d7f6 e4h4 f6d5 c1g5 d6e7 c4d5 e7g5 f3g5 h7h6 d5b3 h6g5 h4e4 a7a5 a2a4 d8f6 e1e3 f6f4 e4f4 g5f4 e3f3 g7g5 a1e1 f8d8 c2c3 a8a6 e1e5 f7f6 b3e6 c8e6 e5e6 a6b6');
INSERT INTO evaluations VALUES(45702,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(45702,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(45702,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(45702,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(45702,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(45702,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(45702,36,'cp 109','e1g1 f8e7 f1e1 e8g8 a2a4 a7a5 d1e2 b7b6 c3e4 c8a6 c4a6 a8a6 e4f6 e7f6 c1f4 a6a7 h2h4 a7b7 g2g4 f6e7 g4g5 e7d6 f4e5 d8c7 h4h5 b8d7 e5d6 c7d6 e2e4 d6d5 h5h6 d5e4 e1e4 g7h6 g5h6 g8h8 e4h4 f8g8 g1f1');
INSERT INTO evaluations VALUES(45701,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(45701,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(45701,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(45701,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(45701,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(45701,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(45701,36,'cp 105','f8d6 d1e2 b8d7 c1g5 e8g8 f1e1 d8c7 c4d3 h7h6 g5h4 f8e8 f3e5 b7b6 h4g3 c8b7 a2a3 a7a6 a1d1 b6b5 e5d7 f6d7 g3d6 c7d6 c3e4 d6c7 e2e3 a8d8 b2b4 b7c8 h2h3 g8f8 e4c5 d7c5 d4c5 e6e5 d3e4 f7f5 d1d8 e8d8');
INSERT INTO evaluations VALUES(54426,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(54426,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54426,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54426,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54426,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54426,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(54426,36,'cp 0','g1f3 g8h6 b1d2 b8c6 d2b3 a7a5 f1d3 d8b6 a2a4 f8b4 c1d2 h6f5 d2b4 b6b4 d1d2 f5e3 h1g1 b4b6 e1e2 c8d7 b3d4 c6d4 f3d4 e3c4 d3c4 d5c4 g1d1 e8g8 d4f5 d7c6 f5e7 g8h8 d2d4 b6c7 d4d6 c7b6');
INSERT INTO evaluations VALUES(56259,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(56259,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(56259,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(56259,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(56259,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(56259,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(56259,36,'cp 0','b8d7 d2d3 f7f5 e2f4 d7e5 d3e4 d5e4 d1d8 e8d8 c1d2 g8f6 d2c3 f8d6 b1d2 c7c6 e1c1 d8c7 f2f3 e4f3 d2f3 e5f3 c3f6 g7f6 g2f3 h8e8 d1d3 a7a5 d3c3 d6e5 c3d3');
INSERT INTO evaluations VALUES(707,30,'cp -539',NULL);
INSERT INTO evaluations VALUES(707,31,'cp -559',NULL);
INSERT INTO evaluations VALUES(707,32,'cp -534',NULL);
INSERT INTO evaluations VALUES(707,33,'cp -537',NULL);
INSERT INTO evaluations VALUES(707,34,'cp -555',NULL);
INSERT INTO evaluations VALUES(707,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(707,36,'cp -575','a5c6 e5c6 d4d5 e1g1 d5c6 d2d3 f8b4 c2c3 b4d6 b1d2 e4d3 f1e1 c8e6 d1f3 c6f3 d2f3 e8g8 c1e3 f8b8 a1d1 e6f5 b2b3 f5h7 c3c4 b8d8 c4c5 d6f8 b3b4 f6d5 e3d2 a7a5 b4b5 f8c5 f3e5 c5b6 a2a3 b6c7 e5c6 d8e8 e1e8 a8e8 h2h3');
INSERT INTO evaluations VALUES(916,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(916,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(916,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(916,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(916,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(916,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(916,36,'cp 61','a2a4 c8e6 c1g5 c6b4 e1g1 h7h6 g5f6 e7f6 c4e6 f6e6 d3d4 f8e7 a4a5 e8g8 d1e2 c7c6 f1d1 e5d4 f3d4 e6e5 a5a6 b4a6 d4c6 b7c6 e2a6 e5c5 a1a4 c5b6 a6d3 e7f6 b2b3 f8d8 g2g3 f6c3 d3c3 d6d5 e4d5 d8d5 c3f3 d5d1 f3d1');
INSERT INTO evaluations VALUES(1675,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(1675,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(1675,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(1675,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(1675,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(1675,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(1675,36,'cp -38','d5c6 e1g1 c8g4 f3e5 g4d1 e5c6 d1c2 f1e1 g8e7 c6e7 f8e7 d2d4 a8d8 c1f4 c2b1 a1b1 d8d7 b1c1 c7c6 f4b8 a7a5 b8a7 f7f6 g1f1 e8f7 a7c5 h8d8 c5e7 d7e7 e1e7 f7e7 c1c4 e7d7 f1e2 d8e8 e2d2 e8e4 g2g3 b7b6 a2a4 f6f5 d2d3');
INSERT INTO evaluations VALUES(1757,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(1757,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(1757,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(1757,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(1757,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(1757,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(1757,36,'cp -32','e1g1 c8g4 f3e5 g4d1 e5c6 d1c2 f1e1 g8e7 c6e7 f8e7 d2d4 a8d8 c1f4 d8d4 f4e5 d4e4 b1c3 e4e1 a1e1 e8d7 c3d5 e7d8 e5c7 d8c7 d5c7 d7c7 e1c1 h8c8 c1c2 c7b8 c2c8 b8c8 f2f4 c8d7 g1f2 d7e6 f2e3 e6d5 g2g4 a7a5 e3d3 a5a4 d3e3 b7b6 h2h3 b6b5 e3d3 h7h6 h3h4');
INSERT INTO evaluations VALUES(1706,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(1706,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(1706,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(1706,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(1706,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(1706,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(1706,36,'cp -30','d5d1 e1d1 c8g4 c1e3 c6e5 b1d2 g8f6 h2h3 e5f3 d2f3 e8c8 d1c1 g4h5 f1d3 f8e7 h1d1 a7a6 d3f5 c8b8 d1d8 h8d8 g2g4 h5g6 f5g6 h7g6 c2c4 d8h8 f3g1 f6d7 c1c2 d7e5 e3d4 e7f6 g4g5 e5c6 d4f6 g7f6 g5f6 b8c8 a1e1 c8d7');
INSERT INTO evaluations VALUES(1733,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(1733,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(1733,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(1733,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(1733,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(1733,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(1733,36,'cp -26','e1d1 c8g4 c1e3 c6e5 b1d2 e8c8 d1c1 e5c6 h2h3 g4h5 g2g4 h5g6 f3h4 g8e7 h4g6 h7g6 d2c4 e7d5 c2c3 d5e3 f2e3 d8e8 c1c2 f8c5 a1e1 c6e5 c4e5 e8e5 f1d3 e5e7 d3c4 c7c6 b2b4 c5b6 e3e4 f7f6 e4e5 f6e5 c4d3 g6g5 e1e4 c8c7');
INSERT INTO evaluations VALUES(1732,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(1732,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(1732,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(1732,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(1732,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(1732,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(1732,36,'cp -23','c8g4 c1e3 c6e5 b1d2 e8c8 d1c1 e5c6 h2h3 g4h5 g2g4 h5g6 f3h4 g8f6 c2c3 f6d5 h4g6 h7g6 c1c2 d5e3 f2e3 d8e8 d2c4 f8c5 a1e1 c6e5 c4e5 e8e5 f1d3 e5e7 d3c4 c7c6 b2b4 c5b6 h3h4 a7a5 b4a5 b6a5 h4h5 g6h5 h1h5 h8h5 g4h5');
INSERT INTO evaluations VALUES(27646,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(27646,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(27646,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(27646,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(27646,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(27646,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(27646,36,'cp -21','c1e3 c6e5 b1d2 g8e7 h2h3 g4f3 d2f3 e8c8 d1e1 e5f3 g2f3 e7d5 f1c4 d5e3 f2e3 f8c5 e3e4 h8f8 a1d1 d8e8 e1d2 e8e5 c2c3 c7c6 h3h4 f7f5 e4f5 f8f5 h1e1 e5e1 d1e1 f5f3 c4d3 g7g6 e1e8 c8c7 e8h8 f3f7 d2c2 c7d7 h8a8 c5e7 a8a7');
INSERT INTO evaluations VALUES(3254,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(3254,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(3254,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3254,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(3254,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3254,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3254,36,'cp 12','d1d2 d6d5 e1c1 c6d4 d2d4 f6e4 d4d5 e4d6 e3d4 c8e6 d5g5 d6e4 g5f4 e4c3 d4c3 d8b8 f4b8 a8b8 c3g7 g8g7 d1d4 b8c8 b2b3 c8c5 e2d3 f8c8 c1b2 c5g5 g2g3 g5h5 c2c4 h5h3 d3e4 b7b6 e4d5 e6d5 d4d5 a7a6 d5d7');
INSERT INTO evaluations VALUES(4563,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(4563,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(4563,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4563,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4563,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(4563,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(4563,36,'cp 0','h7h6 g5d2 e7b4 e1g1 d7d6 a2a3 b4c3 d2c3 c8e6 c4e6 f7e6 d1e2 d8e8 a1e1 e8h5 b2b4 a7a6 f3d2 h5g6 d2f3');
INSERT INTO evaluations VALUES(6417,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(6417,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(6417,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(6417,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(6417,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(6417,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(6417,36,'cp 15','e8f8 g4f3 g8f6 g1e2 c6d4 e2d4 e5d4 c3d5 c7c6 d5f6 d8f6 f3h5 d7d6 d2d3 a7a5 a2a3 f6g6 h5g6 h7g6 a3a4 h8h5 f2f3 f7f6 h2h3 g6g5 g2g4 h5h4 e1f2 f8e7 c1d2 c5b4 c2c3 b4c5 f2g2 c8e6 c4e6 e7e6 c3d4 c5d4 d2c3 d4e5');
INSERT INTO evaluations VALUES(6919,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(6919,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(6919,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(6919,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(6919,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6919,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(6919,36,'cp 24','g5e3 c5e3 f2e3 d7d5 e4d5 d8d5 b1c3 d5f7 d3d4 e5d4 e3d4 c8e6 d1d2 e8c8 e1c1 g8e7 f1b5 e6g4 d4d5 c8b8 h1e1 f7h5 a2a3 g4f3 g2f3 h5f3 e1g1 c6e5 g1g7 e7f5 g7g2 f3h3 b5f1 h3h4 g2f2 f5d6 f2f4 h4g5 c3e4 d6e4 f4e4 h8e8 d2g5 f6g5 f1h3 e5g6 h3e6 d8d6 d1e1 a7a6 c2c4');
INSERT INTO evaluations VALUES(9428,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(9428,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(9428,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9428,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9428,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9428,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(9428,36,'cp 0','f6e4 d2d3 e4f6 c1g5 c8e6 c2c3 e6d5 c4d5 h7h6 g5f6 d8f6 d5c6 b7c6 d3d4 e5d4 f1e1 e8f8 c3d4 c5b6 d1c2 c6c5 d4c5 b6c5 f3e5 c5f2 c2f2 d6e5 f2c5 f8g8 e1e5 g8h7 c5c2 f6g6 c2g6 h7g6 e5e7 h8d8 a1c1 a8c8 e7c7 c8c7 c1c7 d8d1 g1f2 d1b1 c7a7 b1b2 f2g1 b2b1 g1f2');
INSERT INTO evaluations VALUES(12038,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12038,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12038,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(12038,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12038,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12038,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12038,36,'cp 0','d8b6 a1b1 f6g4 d1f3 g7g6 h2h3 b6c5 h3g4 c5g5 c3d5 c6d5 f3c3 g5g4 c3c6 e8d8 c6a8 g4e4 f1e2 e4g2 h1h4 f8g7 a8a7 d5d4 a7a5 d8d7 e2f1 g2f3 f1e2 f3g2');
INSERT INTO evaluations VALUES(14896,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(14896,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(14896,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(14896,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(14896,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(14896,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(14896,36,'cp 90','f8d6 f4g3 f6e4 d1c2 e4g3 h2g3 c6b4 c2e2 b4d5 e1g1 c8d7 b1d2 e8g8 a2a3 d6e7 a1c1 d5f6 e3e4 a7a5 f1e1 a8c8 e4e5 f6d5 d2e4 b7b6 c4d3 d7a4 d3b1 c7c6 e2d2 c6c5 d4c5 b6c5 g3g4 a4c6 e4c5');
INSERT INTO evaluations VALUES(16494,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(16494,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(16494,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(16494,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(16494,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(16494,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(16494,36,'cp -91','f6e4 d3e4 d7d5 e4d3 e5e4 d3b5 d8f6 b1c3 e4f3 d1f3 f6f3 g2f3 c8e6 d2d4 e8d7 c3e2 a7a6 b5c6 d7c6 h2h4 e6f5 e2g3 f5d7 g3e2 a6a5 a2a4 a8e8 c1e3 d7f5 e1d2 h7h5 e2f4 f8d6 f4d3 g7g6 a1g1 b7b6 e3g5 c6d7 g5f6 h8g8 f6e5 f5d3 e5d6 d7d6 d2d3');
INSERT INTO evaluations VALUES(18280,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(18280,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(18280,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(18280,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(18280,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(18280,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(18280,36,'cp 48','g7g6 e1g1 f8g7 b1c3 g8f6 c1f4 a7a5 d1d2 e8g8 a1e1 c8a6 b2b3 a6d3 c2d3 d8b8 e4e5 d6e5 f4e5 b8b4 d2b2 f8d8 e1d1 f6g4 a2a3 b4e7 e5g7 e7d6 g2g3 g8g7 f1e1 g4f6 b2c2 d6c5 g1g2 a8b8 h2h3');
INSERT INTO evaluations VALUES(18525,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(18525,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(18525,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(18525,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(18525,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(18525,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(18525,36,'cp 62','c6c5 c3e2 c5d4 e2d4 g8e7 f1d3 e7c6 c1f4 d7c5 c2c3 c5d3 d1d3 c8d7 e1g1 h7h6 h2h3 a7a6 f1e1 f8e7 d3d2 a8c8 d4b3 a6a5 b3d4 c6d4 f3d4 e7c5 e1e3 g7g5 f4g3 d8b6 a1d1 c5d4 c3d4 b6b4 e3c3 c8c3 b2c3 b4b5 d2e3 e8g8');
INSERT INTO evaluations VALUES(20454,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(20454,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(20454,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(20454,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(20454,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(20454,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(20454,36,'cp 71','f8d6 b5c6 b7c6 b1c3 g8f6 e1g1 e8g8 f1e1 h7h6 d2d3 c8g4 c3e4 f6e4 d3e4 f7f5 d1d3 d6c7 d3d8 f8d8 e4f5 g4f5 c2c3 a8b8 h2h3 a7a5 f3e5 d8e8 c1f4 c7e5 f4e5 b8b2 a2a4 b2d2 e5c7 e8e1 a1e1 d2a2 e1e5 f5c2 e5a5 a2a4 a5c5 c2e4 f2f3');
INSERT INTO evaluations VALUES(21533,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21533,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(21533,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(21533,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21533,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(21533,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(21533,36,'cp 13','d5d4 c3e2 d8d6 e1g1 g8e7 d2d3 c8g4 f3e1 f7f6 h2h3 g4d7 c2c3 a7a6 b5c4 d4c3 b2c3 c6a5 c4b3 a5b3 a2b3 f6f5 e1f3 f5e4 d3e4 d6d1 f1d1 e7c6 b3b4 f8d6 b4b5 c6d8 c3c4 d7e6 f3d2 e6f7 c1a3 d6a3 a1a3 d8e6 b5a6');
INSERT INTO evaluations VALUES(24936,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(24936,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(24936,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(24936,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(24936,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(24936,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(24936,36,'cp 12','f3e5 g4f5 f1g2 e7e6 e1g1 f8e7 e5c6 b7c6 c3a4 e8g8 c1f4 e7d6 d1d2 f6e4 d2e3 f8e8 c2c4 h7h6 a1c1 a8b8 f1d1 a7a5 f4e5 d6b4 c4d5 e6d5 g2e4 f5e4 e3f4 d8d7 f2f3 e4g6 e5c7 b8b5 a2a3 b4f8 e2e3 e8e6');
INSERT INTO evaluations VALUES(25589,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(25589,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(25589,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(25589,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(25589,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(25589,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(25589,36,'cp 29','c2c4 e7e6 d1b3 d5c4 f1c4 d7b6 e1g1 b6c4 b3c4 f8d6 f4g3 e8g8 f1c1 f6e4 b1d2 h7h6 a2a3 a7a6 c4e2 d8e7 d2e4 f5e4 f3e5 a8d8 f2f3 e4h7 b2b4 d6e5 g3e5 c7c6 e2c4 f8e8 c1c3 f7f6 e5g3 h7g6 e3e4 d8d7 a1e1 g8h8');
INSERT INTO evaluations VALUES(25726,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25726,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25726,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(25726,36,'cp 0','b7b5 a4b3 c6a5 e1g1 g8f6 d3d4 e5d4 c3d5 f8e7 d5e7 d8e7 f1e1 d7g4 d1d4 a5b3 a2b3 g4f3 g2f3 e8g8 c1g5 e7e6 e4e5 f6d7 e5d6 e6g6 d4g4 c7d6 g5f4 g6f6 f4g5');
INSERT INTO evaluations VALUES(25764,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25764,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25764,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25764,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25764,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25764,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25764,36,'cp 0','a4b3 c6a5 e1g1 g8f6 d3d4 e5d4 c3d5 f8e7 d5e7 d8e7 f1e1 d7g4 d1d4 g4f3 g2f3 e8g8 c1g5 a5b3 a2b3 e7e6 e4e5 f6d7 e5d6 e6g6 d4g4 c7d6 g5f4 g6f6 f4g5');
INSERT INTO evaluations VALUES(25758,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25758,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25758,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25758,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25758,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25758,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25758,36,'cp 0','c6a5 e1g1 g8f6 d3d4 e5d4 c3d5 f8e7 d5e7 d8e7 f1e1 a5b3 a2b3 d7g4 d1d4 g4f3 g2f3 e8g8 c1g5 e7e6 e4e5 f6d7 e5d6 e6g6 d4g4 c7d6 g5f4 g6f6 f4g5');
INSERT INTO evaluations VALUES(25881,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(25881,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(25881,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(25881,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(25881,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(25881,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(25881,36,'cp 124','g1f3 a8b8 e1g1 g8f6 e2e4 f8e7 a2a3 a7a5 c3d5 e8g8 d2d3 d7d6 c1e3 b8a8 h2h3 f8e8 d1c2 e7f8 a1c1 h7h6 b2b3 c6b8 d3d4 c7c6 d5f6 d8f6 f1d1 b8d7 d4e5 d7e5 f3d4 d6d5 f2f4 e5d7 c4d5 c6d5');
INSERT INTO evaluations VALUES(25880,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(25880,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(25880,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(25880,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(25880,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(25880,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(25880,36,'cp 116','a8b8 e1g1 g8f6 e2e4 f8e7 d2d3 e8g8 a2a3 a7a5 c3d5 d7d6 c1e3 b7c8 h2h3 h7h6 a1b1 b8a8 d5c3 c8e6 b2b4 a5b4 a3b4 d8d7 g1h2 a8a3 d1c2 f8a8 b4b5 c6d8 e3c1 a3a5 c3d5 e6d5 e4d5 e7f8 c1b2 a5a2 d3d4 e5e4 f3d2');
INSERT INTO evaluations VALUES(26102,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(26102,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(26102,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(26102,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(26102,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(26102,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26102,36,'cp 36','e1g1 g8e7 b1d2 c6c5 h2h3 g4f3 d2f3 c5d4 d1d4 d8b6 d4b6 d7b6 c1e3 e7c6 e2b5 f8e7 e3b6 a7b6 c2c3 g7g5 g2g3 e8c8 g1g2 h7h5 f1e1 c6a7 a2a4 c8c7 b5d3 a7c6 f3d4 c6d4 c3d4 d8c8 d3b5 c7b8 e1e2 c8c7');
INSERT INTO evaluations VALUES(26101,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(26101,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(26101,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(26101,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(26101,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(26101,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(26101,36,'cp 26','g8e7 b1d2 c6c5 c2c4 e7c6 c4d5 e6d5 h2h3 g4e6 f1e1 f8e7 d4c5 e8g8 d2b3 d7c5 b3c5 e7c5 e2d3 d8b6 e1e2 a7a5 a2a3 a5a4 c1f4 f8c8 a1c1 c6d4 f3d4 c5d4 c1c8 e6c8 d1c2 h7h6 c2d2 c8d7 d3b1 d7b5 e2e1 b5c4 e1d1 d4c5 d1c1 c4b3 c1e1');
INSERT INTO evaluations VALUES(30665,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(30665,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(30665,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(30665,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(30665,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(30665,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30665,36,'cp 16','d7d5 e1g1 b8d7 f3d2 g8f6 c2c4 f8e7 b1c3 e8g8 e2e4 d5e4 d2e4 f6e4 c3e4 c7c5 d4d5 d7f6 e4f6 e7f6 d1b3 e6d5 c4d5 f6d4 c1e3 d4e3 b3e3 d8f6 e3c3 a8d8 c3f6 g7f6 f1d1 d8d6 a2a4 f8d8 a4a5 g8g7 h2h4 h7h6 a5b6');
INSERT INTO evaluations VALUES(34605,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34605,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(34605,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(34605,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(34605,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34605,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34605,36,'cp -18','a7a6 b1c3 d5a5 g1f3 b8c6 h2h3 c8d7 d2d4 e8c8 c1d2 g8f6 c3d1 a5h5 d1e3 e7e5 f3e5 h5e2 f1e2 c6e5 d4e5 f6e4 a1d1 f8c5 e2f3 d7c6 f3e4 c6e4 e1g1 c5e3 d2e3 e4c2 d1c1 c2a4 b2b3 a4c6 f2f3 h8e8');
INSERT INTO evaluations VALUES(35164,30,'cp 348',NULL);
INSERT INTO evaluations VALUES(35164,31,'cp 345',NULL);
INSERT INTO evaluations VALUES(35164,32,'cp 344',NULL);
INSERT INTO evaluations VALUES(35164,33,'cp 332',NULL);
INSERT INTO evaluations VALUES(35164,34,'cp 339',NULL);
INSERT INTO evaluations VALUES(35164,35,'cp 332',NULL);
INSERT INTO evaluations VALUES(35164,36,'cp 356','g1f3 h4d8 e2e4 b8c6 b1c3 d7d6 f1b5 c8g4 c1e3 d8d7 e5d6 f8d6 h2h3 g4e6 a2a3 a7a6 e4e5 d6e7 d1d7 e6d7 b5c4 d7e6 c4e6 f7e6 c3e2 g8h6 e3h6 g7h6 e2f4 e8f7 e1e2 a8d8 a1d1 e7g5 d1d8 h8d8 f4d3 c6d4 f3d4 d8d4 c2c3 d4a4 h1d1 a4c4');
INSERT INTO evaluations VALUES(35162,30,'cp 332',NULL);
INSERT INTO evaluations VALUES(35162,31,'cp 343',NULL);
INSERT INTO evaluations VALUES(35162,32,'cp 328',NULL);
INSERT INTO evaluations VALUES(35162,33,'cp 351',NULL);
INSERT INTO evaluations VALUES(35162,34,'cp 351',NULL);
INSERT INTO evaluations VALUES(35162,35,'cp 360',NULL);
INSERT INTO evaluations VALUES(35162,36,'cp 336','h4d8 b1c3 b8c6 e2e4 d7d6 f1b5 c8g4 c1f4 g8e7 h2h3 g4e6 f3d4 d6e5 f4e5 e6d7 e5g3 a7a6 b5e2 e7g6 d1d2 f8e7 c3d5 e8g8 g3c7 e7g5 d2d3 d8c8 e1g1 c6b4 d5b4 c8c7 b4d5 c7d6 d4f3 g5h6 a1d1 a8d8 f1e1 g6f4 d5f4 d6f4 e4e5 f4f5 d3f5 d7f5');
INSERT INTO evaluations VALUES(35693,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35693,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(35693,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(35693,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(35693,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(35693,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(35693,36,'cp 22','d7d5 c4d5 c6d5 b1c3 f6e4 f3e5 c8f5 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 d1a4 e7e6 a4c6 a8c8 c6a4 c8c3 f1c1 c3c1 a1c1 d8b6 e2e3 b6b2 a4c6 a7a5 a2a4 h7h5 h2h4 e6e5 d4e5 f8c8 c6c8 f5c8 c1c8 g8h7 c8c7 b2a1 g1h2 g7e5 c7f7 h7h8 g2d5 e5f4 f7f4 a1f1 d5f7 h8g7 f7e6 g7h8 f4f7 f1e1');
INSERT INTO evaluations VALUES(36458,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(36458,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(36458,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(36458,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(36458,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(36458,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(36458,36,'cp -61','d6e5 d4e5 d8d1 c3d1 f6e4 c1e3 f8d8 f1e2 g6g5 g1f3 b8c6 f4g5 c6e5 e3f4 e5f3 g2f3 e4d6 d1c3 c8e6 c3d5 e6d5 c4d5 g7b2 a1b1 b2c3 e1f2 d6f5 b1b7 d8d5 h1d1 d5a5 b7b5 c3d4 f2e1 a5b5 e2b5 a8b8 b5d3 f5e3 d1d2 e7e5 f4e3 d4e3 d2e2 e3f4 h2h4 b8d8');
INSERT INTO evaluations VALUES(36794,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(36794,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(36794,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(36794,33,'cp -85',NULL);
INSERT INTO evaluations VALUES(36794,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(36794,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(36794,36,'cp -62','d4e5 d8d1 c3d1 f6e4 c1e3 f8d8 f1e2 g6g5 f4g5 b8c6 g1f3 c6e5 e3f4 e5f3 g2f3 e4d6 d1c3 c8e6 c3d5 g7b2 a1b1 e6d5 c4d5 b2c3 e1f2 d6f5 h1c1 c3d4 f2g2 b7b6 c1c7 d8d5 g2h3 e7e6 e2c4 d5d8 c4a6 d4e3 f4e5 d8d2 a6b7 a8d8');
INSERT INTO evaluations VALUES(36997,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(36997,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(36997,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(36997,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(36997,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(36997,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(36997,36,'cp 23','c2c3 h7h6 f1e1 a7a6 b1d2 b7b5 c4b3 c5a7 a2a4 c7c5 d2f1 b8c6 a4b5 a6b5 c1e3 c6a5 b3c2 c8e6 h2h3 a7b6 f1g3 b5b4 c2a4 a5b3 a4b3 a8a1 d1a1 e6b3 e1c1 b4c3 b2c3 b3e6 c1b1 b6c7 c3c4 d8d7');
INSERT INTO evaluations VALUES(37680,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(37680,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(37680,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(37680,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(37680,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(37680,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(37680,36,'cp 66','d2d4 c5b6 c2c3 b8c6 b1d2 g8e7 g1f3 e8g8 d2c4 d7d5 c4b6 a7b6 e4e5 d8c7 c1d2 f7f6 e5f6 f8f6 d2g5 f6f8 f1e2 h7h6 g5h4 e7f5 h4g3 e6e5 d4e5 c8e6 e1g1 f5g3 h2g3 c6e5 d1d4 e5c6 d4d2 e6f7 f1b1 f8d8 b1d1 c7e7 e2d3');
INSERT INTO evaluations VALUES(37675,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(37675,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(37675,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(37675,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(37675,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(37675,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(37675,36,'cp 50','c5b6 c2c3 b8c6 b1d2 g8e7 g1f3 e8g8 f1d3 d7d5 e4e5 f7f6 e1g1 e7g6 c1a3 f8f7 a3d6 b6c7 d6c7 f7c7 f1e1 c8d7 e5f6 d8f6 d3g6 f6g6 d2f1 a8c8 e1e3 c6e7 f3e5 g6e8 d1d2 e7f5 e3f3 f5h4 f3h3 h4g6 e5g6 e8g6 a1e1 c8f8 h3g3 g6f6');
INSERT INTO evaluations VALUES(37641,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(37641,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(37641,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(37641,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(37641,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(37641,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(37641,36,'cp 60','c2c3 b8c6 b1d2 g8e7 g1f3 e8g8 f1d3 d7d5 e4e5 f7f6 e1g1 e7g6 c1a3 f8f7 a3d6 b6c7 d6c7 f7c7 e5f6 g6f4 f6f7 c7f7 d1c2 f4d3 c2d3 d8f6 f1e1 c8d7 a2a4 f6f5 d3f5 e6f5 a4a5 f7e7 e1e7 c6e7 f3e5 d7e6');
INSERT INTO evaluations VALUES(37840,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(37840,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(37840,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(37840,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(37840,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(37840,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(37840,36,'cp 91','e2e4 e7e6 b1c3 d7d6 f1d3 g8f6 g1f3 e6d5 e4d5 e8g8 h2h3 f8e8 c1e3 b8d7 e1g1 e8e3 f2e3 d8e7 d1d2 f6h5 g2g4 h5g3 f1f2 h7h5 g1g2 h5g4 h3g4 g7c3 d2c3 g3e4 d3e4 e7e4 g2g3 d7e5 f3e5 d6e5 f2h2 c8g4 a1f1 f7f5 c3d2 e4c4 f1e1 a8d8');
INSERT INTO evaluations VALUES(39732,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39732,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39732,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39732,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39732,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39732,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39732,36,'cp 0','e8g8 a2a4 b8c6 c1e3 c8d7 h2h3 c6a5 f2f4 a5b3 c2b3 a8c8 e4e5 f6d5 c3d5 e6d5 d1f3 d6e5 f4e5 e7c5 a1c1 c5d4 e3d4 c8c1 f1c1 f8e8 f3d5 d7e6 d5d8 e8d8 d4e3 h7h6 g1f2 e6b3 c1c7 b3a4 c7b7 a4b5 b7c7 a6a5 c7c5 d8b8 c5c7 a5a4');
INSERT INTO evaluations VALUES(35576,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35576,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35576,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35576,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35576,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35576,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35576,36,'cp 0','a2a4 b8c6 c1e3 c8d7 h2h3 c6a5 f2f4 d8c7 d1f3 a8c8 g2g4 a5b3 c2b3 e6e5 d4e2 e5f4 e2f4 h7h6 f4d5 f6d5 c3d5 c7d8 e3b6 d8e8 d5c7 e8d8');
INSERT INTO evaluations VALUES(39976,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(39976,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39976,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39976,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(39976,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(39976,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(39976,36,'cp 17','d7d5 c4d5 c6d5 d2d4 e8g8 f3e5 c8f5 e1g1 f6e4 c1f4 e4c3 b2c3 b8c6 e5c6 b7c6 e2e3 f8e8 a1c1 d8a5 c3c4 a5a2 c4d5 c6d5 c1c5 f5e4 d1a1 a2a1 f1a1 e4g2 g1g2 e7e5 d4e5 g7e5 a1d1 e5f4 g3f4 a8d8 d1d5 d8d5 c5d5 e8c8 d5a5 c8c2 a5a7');
INSERT INTO evaluations VALUES(40126,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(40126,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(40126,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(40126,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(40126,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(40126,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(40126,36,'cp 32','d2d4 b7b5 c4d3 e5d4 f3d4 e8g8 c3e2 f8e8 e2g3 e7f8 c1g5 h7h6 g5f6 d8f6 c2c3 b5b4 d3c2 c8a6 f1e1 b4c3 b2c3 g7g6 c2a4 e8c8 a1b1 a6c4 a2a3 c8c7 g3f1 c6c5 d4b5 c4b5 a4b5 f6c3');
INSERT INTO evaluations VALUES(40132,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(40132,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(40132,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(40132,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40132,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(40132,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(40132,36,'cp 42','b7b5 c4d3 e5d4 f3d4 e8g8 c3e2 f8e8 e2g3 a7a5 a2a4 b5a4 a1a4 e7f8 c1g5 h7h6 g5f6 d8f6 c2c3 f6d8 d1f3 c8a6 d3a6 a8a6 f1a1 g7g6 f3e2 d6d5 d4c2 d8b6 b2b4 b8d7 a4a5 a6a5 b4a5 b6b2 e2d1 d7f6 a5a6');
INSERT INTO evaluations VALUES(41529,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(41529,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(41529,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(41529,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(41529,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(41529,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(41529,36,'cp 18','d4e5 d8d1 e1d1 f6d7 a2a4 b8a6 g1f3 f7f6 f1a6 b7a6 c3d5 e8d8 e5f6 e7f6 a1a3 c8b7 a3d3 d7c5 d3d4 f6f5 f3e5 g7e5 f4e5 d8c8 d5e7 c8b8 h1e1 c5e6 d4d2 h8e8 e7d5 b7d5 d2d5 b8b7 c2c4 a8d8 d1c2 c7c6 d5d1 f5f4 g2g3 e6d4 c2c3 d4e2 e1e2 d8d1 c1f4 c6c5 e5e6 b7c6 e6e7 c6d7 f4g5');
INSERT INTO evaluations VALUES(43036,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(43036,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(43036,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(43036,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(43036,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(43036,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(43036,36,'cp 37','f1g2 d7d6 e2e3 c7c6 d2d4 c8e6 b2b3 e5d4 d1d4 d6d5 c4d5 f6d5 c3d5 e6d5 g2d5 c6d5 c1b2 e7f6 d4d2 b8c6 g1f3 f6b2 d2b2 e8g8 e1g1 d8e7 f3d4 a8c8 f1d1 f8d8 a1c1 h7h5 d4c6 b7c6 h2h4 c8c7 c1c3 g7g6 b2c2 c6c5 c3d3 e7e6 c2d2 c5c4 b3c4');
INSERT INTO evaluations VALUES(44233,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(44233,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(44233,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44233,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44233,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44233,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(44233,36,'cp 0','d7d5 g1f3 b7b6 b1d2 f8e7 e1g1 c8b7 d1e2 b8c6 b2b3 a7a5 d3b5 e8g8 c1b2 a8c8 a1d1 d8c7 d1c1 f8d8 f1d1 c6a7 b5d3 a7c6');
INSERT INTO evaluations VALUES(47591,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(47591,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47591,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(47591,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(47591,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(47591,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(47591,36,'cp 36','b2b3 c7c5 d4c5 f8c5 c1b2 e8g8 g1f3 d5c4 f1c4 a7a6 c4d3 b7b5 a2a4 b5a4 a1a4 c8b7 h2h3 b8d7 e1g1 a6a5 d3e4 d8b8 e4b7 b8b7 a4c4 c5b4 f3d2 f8c8 c4c8 b7c8 d2c4 c8b7 d1a1 d7b6 c4b6 b7b6');
INSERT INTO evaluations VALUES(48651,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(48651,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(48651,32,'cp -100',NULL);
INSERT INTO evaluations VALUES(48651,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(48651,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(48651,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(48651,36,'cp -97','g1f3 f8e7 c1e3 f6g4 e3g1 e8g8 f1e2 f8e8 a2a3 e7d6 d1d2 c7c6 b1c3 c8f5 h2h3 g4f6 g1h2 b8d7 e1g1 f6e4 c3e4 f5e4 f3e5 d7f6 c2c3 c6c5 g1h1 c5c4 f1e1 e8e7 a3a4 e4f5 g2g4 f5e4 e2f3 e4f3 e5f3 f6e4');
INSERT INTO evaluations VALUES(50657,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(50657,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(50657,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(50657,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(50657,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(50657,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(50657,36,'cp 21','c5d4 f3d4 g8f6 e1g1 g7g6 b1c3 f8g7 f1e1 e8g8 c3d5 b8d7 c2c3 f8e8 a2a4 d7c5 d5f6 g7f6 c1h6 c8d7 a4a5 a8c8 c4f1 d8c7 b2b4 c5e6 a1c1 b7b5 a5b6 c7b6 d4e6 d7e6 h6e3 b6c6 e3d4 f6g5 c1a1 e6c4 h2h3 h7h6 e4e5 c4f1 e1f1');
INSERT INTO evaluations VALUES(50680,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(50680,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(50680,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50680,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(50680,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(50680,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(50680,36,'cp 30','d1d4 b8c6 d4e3 g8f6 e1g1 e7e6 c4e2 b7b5 c2c4 b5c4 e2c4 f8e7 b1c3 f6g4 e3e2 e8g8 f1d1 c6e5 f3e5 g4e5 c1f4 d8b6 c3a4 b6c6 c4b3 c6b5 e2e3 a8b8 a1c1 c8d7 a4c3 b5b7 e3g3 f8c8 f4e5 d6e5 g3e5 e7f6');
INSERT INTO evaluations VALUES(50679,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(50679,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50679,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(50679,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(50679,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(50679,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(50679,36,'cp 9','g8f6 e1g1 c8g4 d4e3 e7e6 c4e2 b8d7 h2h3 g4f3 e2f3 d7e5 b1d2 d8c7 c2c4 e5f3 e3f3 f8e7 f1d1 e8g8 b2b3 f8d8 c1b2 h7h6 a1c1 a8c8 b2c3 f6d7 f3e2 e7g5 c3d4 c7a5 c1c2 d6d5 d4e3 g5f6 e4d5 e6d5 c2c1 a5a2');
INSERT INTO evaluations VALUES(50782,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(50782,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(50782,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50782,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(50782,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(50782,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(50782,36,'cp 19','c2c3 f8d6 f4g3 g8f6 b1d2 d8b6 d1c2 e8g8 f1e2 h7h6 e1g1 d6g3 h2g3 c8d7 c2b3 b6a7 d4c5 f8c8 c3c4 c8c5 f1c1 b8c6 b3a3 d5c4 c1c4 c5a5 a3d3 a5d5 d3c3 d7e8 a2a3 c6a5 c4c7 a7b6 e3e4 d5d7 c7d7 e8d7 e4e5 f6d5');
INSERT INTO evaluations VALUES(50952,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(50952,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(50952,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(50952,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(50952,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(50952,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(50952,36,'cp -27','g8f6 e2e3 c7c5 d4c5 f8c5 f1d3 b8d7 e3e4 d8b6 e1g1 b6b2 d1e1 c5b4 f4d2 d5e4 c3e4 b4d2 e4d6 e8e7 f3d2 b2b4 d6f5 e7f8 a1b1 b4f4 f5g3 g7g6 d2e4 f8g7 e4f6 d7f6 e1c3 b7b5 b1b4 f4d6 g3e4 d6e7 a2a4 c8b7 a4b5 h8c8 c3f6 e7f6 e4f6 g7f6 f1b1 a6b5 b4b5 b7a6 b5a5 a6d3 a5a8 c8a8');
INSERT INTO evaluations VALUES(51157,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51157,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51157,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(51157,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(51157,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(51157,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(51157,36,'cp 27','a4b3 d7d6 d2d3 g8f6 f2f4 b8d7 g1f3 e5f4 e1g1 d7c5 c1f4 f8e7 f3d4 c8g4 d1e1 g4d7 b1c3 c5b3 a2b3 c7c5 d4e2 e8g8 e2g3 d7e6 h2h3 h7h6 e1f2 d6d5 g1h1 d5d4 c3e2 a6a5 f4d2 f6d7 g3f5 e6f5 f2f5 a5a4 b3a4 a8a4');
INSERT INTO evaluations VALUES(51156,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(51156,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51156,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(51156,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(51156,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(51156,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(51156,36,'cp 20','d7d6 d2d3 g8f6 a2a4 c8b7 a4b5 a6b5 a1a8 b7a8 g1f3 f8e7 b1c3 b5b4 c3e2 b8d7 e1g1 e8g8 e2g3 d6d5 e4d5 a8d5 b3d5 f6d5 f1e1 e7d6 f3d2 d7b6 d2c4 b6c4 d3c4 d5f6 h2h3 h7h6 g3f5 f8e8 f5d6 c7d6 c1d2 d8a5 c2c3 b4c3 d2c3');
INSERT INTO evaluations VALUES(52448,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(52448,31,'cp -64',NULL);
INSERT INTO evaluations VALUES(52448,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(52448,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(52448,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(52448,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(52448,36,'cp -61','d7d5 f1b5 c8d7 b5d7 d8d7 d2d3 b8c6 d1e2 g8f6 f2f3 c6d4 e2f2 e7e5 e1g1 h7h6 b1c3 f8e7 c3d5 f6d5 e4d5 d7d5 c2c3 d4c6 f3f4 e8g8 f2g3 a8d8 c1e3 d5d3 f4e5 c5c4 h3f2 d3g6 g3g6 f7g6 f2e4 f8f1 a1f1 c6e5 e3a7');
INSERT INTO evaluations VALUES(52837,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(52837,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(52837,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(52837,33,'cp -57',NULL);
INSERT INTO evaluations VALUES(52837,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(52837,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(52837,36,'cp -60','e7e5 d2d4 e5e4 c2c4 f7f5 g1h3 g8f6 b1c3 f8d6 c4d5 c6d5 h3f4 d6f4 e3f4 b8c6 c1e3 e8g8 e1g1 c6a5 f1e1 h7h5 g2f1 c8e6 a1c1 a8c8 c3a4 a5c4 a4c5 d8e7 d1a4 b7b6 c5e6 e7e6 a4a7 h5h4 a7b7 c4b2 c1c8');
INSERT INTO evaluations VALUES(53777,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(53777,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(53777,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(53777,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(53777,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(53777,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(53777,36,'cp -50','d2d4 e5e4 c2c4 b8a6 c4d5 c6d5 f2f3 f7f5 f3e4 d5e4 g2f1 g8f6 g1h3 f8d6 f1c4 a6c7 b1c3 d8e7 e1g1 c8d7 h3f4 d6f4 f1f4 b7b5 c4b3 b5b4 c3a4 c7d5 f4f2 e8g8 a4c5 g8h8 c1d2 a7a5 a2a3 a8c8 a3b4 a5b4 b3d5 f6d5');
INSERT INTO evaluations VALUES(52914,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(52914,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(52914,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(52914,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(52914,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(52914,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(52914,36,'cp -73','g8f6 e4f3 e7e5 d2d3 c8e6 b1c3 d8c7 g1e2 f8e7 e1g1 e8g8 h2h3 b8d7 c3e4 f8e8 e2g3 a8d8 f1e1 f6e4 d3e4 d7f8 d1e2 f8g6 f3g4 g6f4 e2f3 e6g4 h3g4 g7g6 c1e3 f4e6 a1d1 e6d4 e3d4 d8d4 d1d4');
INSERT INTO evaluations VALUES(45047,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(45047,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(45047,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(45047,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(45047,34,'cp -56',NULL);
INSERT INTO evaluations VALUES(45047,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(45047,36,'cp -60','e4f3 e7e5 d2d3 f8e7 g1e2 c8e6 b1c3 d8c7 e1g1 e8g8 h2h3 b8d7 c3e4 a8d8 e2g3 f6d5 e4g5 d7c5 f3d5 c6d5 d3d4 e5d4 g5f3 d4d3 c2d3 c7b6 f3d4 f8e8 g3h5 g7g6 h5f4 e7f6 d4e6 c5e6 f4e6 e8e6 a1b1 b6a6');
INSERT INTO evaluations VALUES(53387,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(53387,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53387,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(53387,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(53387,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53387,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(53387,36,'cp 0','b8c6 g1f3 c8f5 c2c4 e7e6 c4c5 a7a6 c1e3 g8f6 b1c3 f8e7 f1e2 e8g8 a1c1 d8c7 e1g1 a8d8 f1e1 f6e4 e2d3 h7h6 a2a3 e4c3 b2c3 f5d3 d1d3 e6e5 d4e5 c6e5 f3e5 c7e5 c3c4 d5c4 d3c4 e5d5 c4g4 d5d7 g4d7 d8d7 g1f1 e7f6 e1d1');
INSERT INTO evaluations VALUES(53656,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(53656,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53656,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(53656,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(53656,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(53656,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(53656,36,'cp 36','b7b5 a2a4 f8b4 c1d2 a7a5 a4b5 b4c3 d2c3 c6b5 b2b3 c8b7 c3b2 g8e7 b3c4 b5c4 f1c4 e8g8 e1g1 b8c6 e3e4 d8b6 d1e2 b7a6 c4a6 b6a6 e2a6 a8a6 b2a3 f8b8 f1d1 c6b4 f3e5 b8b7 a1c1 h7h6 e5c4 b7b5 g1f1 e7c6 c4d6 b5b8');
INSERT INTO evaluations VALUES(56702,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(56702,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(56702,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(56702,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(56702,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(56702,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(56702,36,'cp 156','d3d4 c8e6 c2c4 g7g6 b1c3 f8h6 f1e2 h6c1 a1c1 g8h6 e1g1 h6f7 b2b4 e8g8 c4c5 a7a5 c5d6 c7d6 b4b5 b8d7 d4d5 e6g4 f3e5 g4e2 e5f7 e2d1 f7d8 a8d8 f1d1 d7c5 f2f3 f6f5 h2h3 b7b6 a2a4 g8g7 h3h4 d8d7 e4f5 f8f5 g1f2');
INSERT INTO evaluations VALUES(57274,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(57274,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(57274,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(57274,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(57274,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(57274,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(57274,36,'cp 158','b1c3 g8f6 d1c2 b8c6 f1g2 d7d5 c4d5 c6b4 c2b1 c8f5 c3e4 f5e4 g2e4 f6e4 b1e4 f8e7 g1f3 e8g8 e1g1 f8e8 e4c4 d8d5 c4d5 b4d5 f1d1 f7f5 g1f1 e7f6 d2d4 h7h6 d1d3 g7g5 c1d2 g5g4 f3e5 f6e5 d4e5 e8e5 d2h6 c7c6 a1d1 a8e8 e2e3 b7b6 h6g5 g8f7 d1c1');
INSERT INTO evaluations VALUES(554,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(554,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(554,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(554,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(554,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(554,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(554,36,'cp 93','f1e2 c5b6 e1g1 c6d4 c3d4 d7d5 a2a4 c7c6 e4d5 e8g8 b1c3 h6f5 e2d3 f5e3 f2e3 f6h6 d1f3 c6c5 a4a5 b6c7 h2h3 c5d4 c3b5 c7d8 b5d4 h6d6 f3e4 g7g6 d4b5 d6b8 e4d4 a7a6 b5c3 d8c7 d3c2 c7e5 d4d2 f7f5 c2b3 b8d6 g1h1 c8d7 a1d1 e5f6');
INSERT INTO evaluations VALUES(823,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(823,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(823,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(823,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(823,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(823,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(823,36,'cp -18','h6f5 c1b2 c8d7 g2g4 f5e7 b1c3 h7h5 c3a4 b6d8 g4g5 e7g6 f1d3 b7b5 a4c5 f8c5 d4c5 e8g8 e1g1 a7a5 d1d2 a8b8 f1e1 d8c7 a1d1 d5d4 d3e4 f8d8 d2e2 a5b4 a3b4 d4d3 e2d3 c6b4 d3d4 b4c6 d4d6 c7c8 f3d4 c6e5 c5c6 d7e8');
INSERT INTO evaluations VALUES(1651,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(1651,31,'cp 201',NULL);
INSERT INTO evaluations VALUES(1651,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(1651,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(1651,34,'cp 218',NULL);
INSERT INTO evaluations VALUES(1651,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(1651,36,'cp 198','d4d5 c6e5 c1f4 e5d3 f1d3 d6f4 d3b5 e8d8 g1e2 f4d6 e1g1 g8f6 f1e1 a7a6 b5a4 e7e5 d5e6 c8e6 e2d4 d8c8 d4e6 d6d1 a1d1 f7e6 e1e6 f8c5 g2g4 h8f8 h2h3 h7h6 g1g2 b7b5 a4b3 c8b7 f2f4 a8e8 g2f3 c7c6 d1e1 e8d8 h3h4 d8d2 e1d1 d2f2 f3g3 b7c7 c3e2 f8e8 e6e8 f6e8 e2d4 c5d4');
INSERT INTO evaluations VALUES(4311,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(4311,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(4311,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(4311,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(4311,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(4311,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(4311,36,'cp -12','c5b6 e1g1 d7d6 d2d3 c6d4 f3d4 b6d4 d1e1 c7c6 a1b1 a7a5 h2h3 a5b4 a3b4 d6d5 c4b3 c8e6 e4d5 f6d5 c3d5 e6d5 b3d5 d8d5 c1e3 b7b5 e3d4 d5d4 c2c3 d4d5 b1c1 f7f6 e1e3 f8d8 f1d1 a8a2 e3g3 e5e4 d3e4 d5d1 c1d1 d8d1 g1h2');
INSERT INTO evaluations VALUES(4240,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(4240,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(4240,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(4240,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(4240,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(4240,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(4240,36,'cp -8','e1g1 d7d6 d2d3 c6d4 f3d4 b6d4 d1e1 c7c6 a1b1 b7b5 c4b3 c8e6 b3e6 f7e6 c1e3 d4e3 e1e3 d8b6 a3a4 b6e3 f2e3 a7a6 b1b3 d6d5 f1a1 f6g4 c3d1 d5d4 e3d4 e5d4 a4b5 c6b5 b3a3 a8c8 c2c3 d4c3 a3c3 c8c3 d1c3 f8c8');
INSERT INTO evaluations VALUES(5220,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(5220,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(5220,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(5220,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(5220,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(5220,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(5220,36,'cp 72','d2d4 c5d4 f3d4 d8e7 e1g1 g8f6 f1e1 c6d4 d1d4 d6c5 d4d1 d7d6 c1e3 c5e3 e1e3 c8d7 d1d4 e8g8 a1d1 f8c8 c4f1 b7b5 d4d6 e7d6 d1d6 b5b4 c3d1 c8c2 e4e5 f6d5 d6d7 d5e3 f2e3 a8c8 d7a7 g7g6 a2a3 c2d2 d1f2 b4a3 b2a3 c8c1 g2g3 c1e1 a7c7 e1e3 f2g4');
INSERT INTO evaluations VALUES(5357,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(5357,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(5357,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(5357,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(5357,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(5357,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(5357,36,'cp -7','c2c3 g8f6 e1g1 e8g8 f1e1 e7e5 c3c4 d7d6 b1c3 a7a5 a2a3 h7h6 c1d2 c8d7 a1b1 f6e8 c3d5 e8c7 h2h3 c7e6 b2b4 c5b4 a3b4 a5b4 d2b4 c6b4 b1b4 d7c6 d1c2 f7f5 e1b1 f5e4 d3e4 e6c5 e2f1 f8f7 d5c3 a8a3 b4b2 g8h7 c3d5');
INSERT INTO evaluations VALUES(5354,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(5354,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(5354,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(5354,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(5354,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(5354,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(5354,36,'cp -15','g8f6 e1g1 e8g8 f1e1 e7e5 c3c4 d7d6 b1c3 a7a5 a2a3 h7h6 c1d2 c8d7 a1b1 f6e8 c3d5 e8c7 h2h3 c7e6 b2b4 c5b4 a3b4 c6b4 d2b4 a5b4 d1d2 d7c6 d2b4 g8h7 b1a1 a8a1 e1a1 f7f5 b4d2 h6h5 d2d1 f5e4 d3e4 g7h6 a1b1 f8f7 d1d3 e6c5');
INSERT INTO evaluations VALUES(8028,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(8028,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(8028,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(8028,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(8028,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(8028,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(8028,36,'cp -7','e1g1 e8g8 f1e1 e7e5 c3c4 d7d6 b1c3 b7b6 a2a3 f6e8 b2b4 e8c7 c3d5 c7d5 c4d5 c6d4 b4c5 b6c5 a1b1 f7f5 a3a4 h7h5 f3d4 c5d4 c1d2 d8f6 f2f3 g8h7 b1c1 f5f4 c1c7 f8f7 c7f7 f6f7');
INSERT INTO evaluations VALUES(8027,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(8027,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(8027,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(8027,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(8027,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(8027,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(8027,36,'cp -17','e8g8 f1e1 e7e5 c3c4 f6e8 b1c3 h7h6 c3d5 a7a5 a1b1 d7d6 a2a3 e8c7 c1d2 c8d7 g2g3 c7d5 e4d5 c6e7 b2b4 a5b4 a3b4 b7b6 d1b3 e7f5 e2f1 g8h7 b4c5 b6c5 d2c3 a8b8 b3a3 f5d4 f3d4 b8b1 e1b1 e5d4');
INSERT INTO evaluations VALUES(7788,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(7788,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(7788,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(7788,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(7788,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(7788,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(7788,36,'cp 13','h7h6 g5h4 b7b5 c4d3 e8g8 e1g1 c8b7 a2a3 a7a5 d1c2 b5b4 h4f6 d7f6 c3a4 f6d7 d3e4 d8c7 f3d2 b7a6 f1c1 a6b5 e4h7 g8h8 h7d3 c7d8 d2f3 b4a3 b2a3 a8c8 c1d1 h8g8 d3e4 d7f6 a4c5 f6e4 c2e4 d8d5 e4c2 d5c4 c2c4 b5c4');
INSERT INTO evaluations VALUES(9013,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(9013,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(9013,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(9013,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(9013,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(9013,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(9013,36,'cp 25','h2g3 d8e7 c2c4 d5c4 d3c4 e8g8 b1c3 f8d8 e1g1 h7h6 a2a3 e6e5 f1e1 c8e6 c4e6 e7e6 d1a4 a7a6 a1d1 d8d7 d4e5 c6e5 d1d7 e5d7 f3d4 e6b6 a4c2 c7c6 c3a4 b6d8 e1d1 d8e7 a4c3 d7e5 c3e2');
INSERT INTO evaluations VALUES(9825,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(9825,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(9825,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9825,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9825,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(9825,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(9825,36,'cp 5','d7d6 d2d3 f6d5 c4d5 e8g8 c2c3 a7a5 a2a4 d8f6 d5b3 c6e7 h2h3 e7g6 c1e3 g6h4 f3h4 c5e3 f2e3 f6h4 d1c2 c8e6 f1f2 h4g5 f2f3 g5e7 b3e6 f7e6 a1f1 f8f3 f1f3 e7d7 c2b3 b7b6 d3d4 a8f8 f3f8 g8f8 d4d5 f8f7 b3c4 e6d5 e4d5 d7e7');
INSERT INTO evaluations VALUES(12715,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12715,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12715,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(12715,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12715,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12715,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12715,36,'cp 0','e4e5 f6g4 c4f7 e8f7 f3g5 f7g8 d1g4 c6e5 g4e4 e5c6 b1c3 e7e6 e1g1 f8g7 c1e3 d7d5 e4f4 d8f8 f4f8 g7f8 e3f4 h7h6 g5f3 c8d7 f4g3 g6g5 f1e1 a8c8 a2a4 h8h7 c3b5 g5g4 f3e5 a7a6 b5c3 c6e5 g3e5 f8e7 a4a5 e7d8 d3d4 c5d4 e5d4 d8c7');
INSERT INTO evaluations VALUES(15203,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(15203,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(15203,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(15203,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(15203,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(15203,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(15203,36,'cp -27','c3d5 d8d6 d5f6 g7f6 g5d2 f6f5 e4f5 c8f5 g2g3 e8c8 f1g2 d6d7 e1g1 d8e8 f1e1 f7f6 f3h4 f5g4 d1b1 e8d8 c2c4 d4c3 b2c3 d7d3 b1b3 d3a6 d2e3 c6a5 b3a4 g4d7 a4c2 a5c4 a1b1 c4e3 f2e3 c7c6 e1d1 d7e6 d1d8 c8d8 c2e4');
INSERT INTO evaluations VALUES(14970,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(14970,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(14970,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(14970,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(14970,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(14970,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(14970,36,'cp -45','d8d6 d5f6 g7f6 g5d2 f6f5 e4f5 c8f5 g2g3 d6d7 f1g2 e8c8 e1g1 f5g4 d1e1 f8d6 a2a4 f7f6 f3h4 h8g8 a1b1 g4h3 g2h3 d7h3 e1e4 c6e7 e4g2 h3d7 a4a5 a7a6 f2f4 c8b8 f4e5 f6e5 b1e1 d8e8 g2e4 h7h5 e4f3 g8f8');
INSERT INTO evaluations VALUES(16976,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(16976,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(16976,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(16976,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(16976,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(16976,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(16976,36,'cp 10','f8c5 d3d4 d7d5 e4d5 e5d4 e1g1 e8g8 h2h3 h7h6 f1d1 d4c3 b1c3 c8d7 a2a4 f8e8 c1f4 c5d6 a4a5 a8b8 c4d3 a7a6 d1e1 b7b5 e1e8 d7e8 f4d2 b5b4 c3e4 e8b5 e4f6 d8f6 f3f6 g7f6 d3f5 b8d8 a1c1 d6e5 b2b3 d8d5 d2b4 g8g7 b4c5 d5d2');
INSERT INTO evaluations VALUES(18429,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(18429,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(18429,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(18429,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(18429,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(18429,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(18429,36,'cp 37','g7g6 g1f3 c8b7 e1g1 f8g7 e2e3 c6a5 d1e2 g8f6 d2d4 d7d5 c4d5 f6d5 c1d2 c5d4 f3d4 e8g8 f1d1 g7d4 c3d5 b7d5 d2a5 d5g2 d1d4 d8c7 g1g2 b6a5 b2b3 f8d8 d4c4 c7d7 h2h4 h7h5 a1c1 a8c8 c4c5 c8c5 c1c5 a5a4 b3a4 d7a4 c5c7 d8d1 c7e7');
INSERT INTO evaluations VALUES(21714,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(21714,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(21714,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21714,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(21714,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(21714,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(21714,36,'cp 12','e4d5 d8d5 b1c3 d5d6 d1e2 c8g4 h2h3 g4f3 e2f3 g8e7 d3c4 d6g6 d2d3 e7f5 c3e2 f8c5 c1d2 e8c8 e1c1 c8b8 h3h4 c6d4 e2d4 f5d4 h4h5 d4f3 h5g6 f3d2 c4f7 h7g6 f7g6 d2c4 h1h8 d8h8 d3c4 c5f2 d1f1 f2e3 c1b1 c7c6 g6e4 e3f4 c2c3 b8c7');
INSERT INTO evaluations VALUES(25090,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(25090,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(25090,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(25090,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(25090,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(25090,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(25090,36,'cp 70','f1e1 f8e7 d2d4 e5d4 f3d4 e8g8 d4f5 f6d7 c3d5 e7f6 d5f6 d8f6 c1d2 f8e8 d2c3 f6g6 f2f3 f7f6 f5e3 g6f7 b2b3 d7e5 d1d2 e5g6 c3b2 g6e7 c2c4 f6f5 e3d5 e7d5 e4d5 c6d7 d2a5 b7b6 e1e8 d7e8 a5c3 e8d7 a1e1 a8e8 e1e8 d7e8 c3d4');
INSERT INTO evaluations VALUES(27849,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(27849,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(27849,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(27849,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(27849,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(27849,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(27849,36,'cp 91','f1e2 a7a5 f3d2 g4e2 d1e2 f6d7 e1g1 b8c6 d2f3 e7e5 a1d1 e5d4 f3d4 c6d4 e3d4 g7d4 d1d4 d8g5 c3d5 g5d8 e2d2 d7c5 d5c3 d8e7 b2b3 f7f6 c3d5 e7g7 f2f3 b7b6 h2h4 f6f5 e4f5 f8f5 d4g4 a8e8');
INSERT INTO evaluations VALUES(27990,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(27990,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(27990,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(27990,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(27990,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(27990,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(27990,36,'cp 129','d7d6 g1h3 g8f6 d1e2 e8g8 e1g1 e6e5 f2f4 f8e8 h3f2 b8d7 e2f3 d8e7 f1e1 h7h6 h2h4 e5f4 c1f4 d7f8 a2a4 a7a5 e1e3 f8g6 f4g3 h6h5 f3f5 b7c8 f5b5 c8d7 b5b2 f6g4 f2g4 d7g4 a1e1 g6h4 b2b5 h4g6 e4e5 d6e5 d3g6 f7g6');
INSERT INTO evaluations VALUES(28829,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(28829,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(28829,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28829,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(28829,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(28829,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(28829,36,'cp 27','d7d5 d2d4 b8d7 f3e5 c7c5 c2c4 c5d4 d1d4 e7e6 c1g5 f8c5 d4f4 d7e5 f4e5 c5d6 e5d4 d6e7 c4d5 e8g8 b1c3 h7h6 g5f6 e7f6 d4d3 f6c3 d3c3 b7d5 a1d1 d8c8 g2d5 c8c3 b2c3 e6d5 d1d5 f8e8 e2e3 e8c8 d5d3 c8c5 f1c1 a8c8 a2a3 c8c7 d3d8 g8h7 d8a8 c5c3');
INSERT INTO evaluations VALUES(28925,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(28925,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(28925,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(28925,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(28925,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(28925,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(28925,36,'cp 105','h2h3 g4f3 d1f3 b8c6 d4d5 c6e5 f3b3 g7g6 f1e2 f8g7 b1a3 e8g8 e1g1 a8c8 c1e3 f7f5 f2f3 f5f4 e3f2 d8e8 a1d1 e8f7 f1e1 b6d7 a3b5 a7a5 b3c2 g6g5 b2b3 f7g6 e2d3 e5d3 c2d3 g6d3 d1d3 g7e5 d3d1 b7b6 f2d4');
INSERT INTO evaluations VALUES(31907,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31907,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(31907,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31907,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(31907,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(31907,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(31907,36,'cp 49','g4f5 c2c4 e7e6 b1c3 g8f6 g1f3 f8d6 f4d6 d8d6 e2e3 e8g8 d1b3 b7b6 f1e2 h7h6 e1g1 a7a6 f3e5 b8c6 e5c6 d6c6 c4d5 e6d5 e2f3 a8d8 a1c1 c6b7 c3e2 c7c6 e2f4 f5e4 f3e4 f6e4');
INSERT INTO evaluations VALUES(35926,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(35926,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(35926,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(35926,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(35926,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35926,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(35926,36,'cp 31','d4d5 h7h6 g5h4 c8f5 f1d3 f5d3 d1d3 g6g5 h4g3 f6h5 e1g1 f7f5 h2h3 b8d7 g3h2 d8e8 a1b1 a7a6 a2a4 b7b6 b2b4 g7c3 d3c3 h5f6 b4c5 f6e4 c3e1 b6c5 f3d2 e4d2 e1d2 a8b8 b1b8 e8b8 d2c2 b8e8 f1b1 f5f4 e3f4 g5f4 b1b7 d7e5 f2f3');
INSERT INTO evaluations VALUES(38453,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38453,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38453,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38453,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38453,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38453,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38453,36,'cp 0','g8f6 e2e3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(41017,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(41017,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41017,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41017,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(41017,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(41017,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(41017,36,'cp 36','f1c4 c7c5 d4c5 e8g8 g1f3 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 g5f4 c8d7 f4d6 b4d6 b5d6 a8b8 a2a3 f8d8 b2b4 g8f8 d1d2 f8e7 f1d1 a7a6 c4f1 d7e8 h2h3 h7h6 b4b5 c6a5 d6e8 d8e8 d2c2');
INSERT INTO evaluations VALUES(41014,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(41014,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(41014,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(41014,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(41014,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(41014,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(41014,36,'cp 19','c7c5 d4c5 e8g8 d1e2 b8d7 g1f3 h7h6 g5f4 d7c5 e1g1 a7a6 a2a4 c8d7 f1d1 d8e8 a4a5 d7c6 f3d4 c5e4 d4c6 e8c6 c3e4 f6e4 c4d3 a8c8 e2f3 e4c5 d3c2 c6f3 g2f3 e7f6 a1b1 f8d8 b2b4 c5d7 b1c1 d7e5 d1d8 c8d8');
INSERT INTO evaluations VALUES(41453,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41453,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(41453,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(41453,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(41453,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41453,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(41453,36,'cp 37','c4d5 f6d5 f1g2 d5b6 e1g1 b8c6 e2e3 e8g8 b1c3 a7a5 d4d5 g7c3 d5c6 d8d1 f1d1 c3f6 f3d4 b7c6 d4c6 c8d7 a1b1 d7c6 g2c6 a8d8 d1d8 f8d8 g1f1 b6d5 c6d5 d8d5 f1e2 a5a4 c1d2 d5c5 b2b4 a4b3 a2b3 c5c2 e2d3 c2a2 f2f3 h7h6');
INSERT INTO evaluations VALUES(41694,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(41694,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(41694,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(41694,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(41694,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(41694,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(41694,36,'cp -84','f1b5 c7c6 d1e2 d6e7 b5d3 c6c5 d4c5 e8g8 c1e3 f8e8 f3e5 b8d7 e1g1 e7c5 e3c5 d7c5 e2f2 f6e4 f2d4 c5e6 d4e3 f7f6 e5f3 d8d6 f3h4 e4c5 d3b5 e8d8 f4f5 e6g5 e3d4 g5e4 b2b4 d6b6 b5d3 c5a4 d4b6 a7b6 f1e1 e4d6 a2a3 a4b2 b1c3 b2d3 c2d3 c8f5 h4f5 d6f5');
INSERT INTO evaluations VALUES(42326,30,'cp 210',NULL);
INSERT INTO evaluations VALUES(42326,31,'cp 220',NULL);
INSERT INTO evaluations VALUES(42326,32,'cp 216',NULL);
INSERT INTO evaluations VALUES(42326,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(42326,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(42326,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(42326,36,'cp 194','e4e5 f6g8 h2h4 h7h6 c1f4 b8c6 d1d2 d7d5 a2a3 b7b6 b2b4 c8b7 f1b5 d8e7 c3e2 e8c8 f4g3 e7d7 a3a4 g8e7 c2c3 a7a6 b5d3 c8b8 g3f4 d8g8 a1b1 b8a7 e2g3 g7f8 d3e2 e7f5 g3f5 g6f5 e1g1 c6e7 h4h5 d7a4 b4b5 a6a5 g2g3 a4a3');
INSERT INTO evaluations VALUES(48999,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(48999,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(48999,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(48999,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48999,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(48999,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(48999,36,'cp -17','b8c6 c1g5 d8d6 g1e2 c8e6 b1c3 d5c3 e2c3 d6d7 c4e6 d7e6 a2a3 h7h6 g5e3 e8c8 e1c1 g7g6 h5e2 f7f5 h1e1 f8g7 c1b1 h6h5 e3g5 g7f6 g5f6 e6f6 e2e3 h5h4 f2f4 e5f4 e3f4 g6g5 f4f2 a7a6 g2g4 f5f4 e1e4 h4h3 d3d4');
INSERT INTO evaluations VALUES(15306,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(15306,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(15306,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(15306,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(15306,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(15306,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(15306,36,'cp -12','c1g5 d8d6 b1c3 c8e6 g1e2 d5c3 e2c3 e6c4 d3c4 d6g6 h5g6 h7g6 e1c1 f7f6 g5e3 g6g5 f2f3 c6d8 c3e4 d8e6 c2c3 b7b6 a2a4 a7a5 d1d5 e6f4 e3f4 g5f4 h2h3 c7c6 d5d2 a8d8 h1d1 d8d2 d1d2 f6f5 e4d6 f8d6 d2d6 h8h6 d6h6 g7h6 b2b4 c6c5 b4a5 b6a5 c1c2 e8f7');
INSERT INTO evaluations VALUES(53398,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(53398,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(53398,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(53398,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(53398,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(53398,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53398,36,'cp 32','c3d4 e7e6 b1c3 f8d6 e2e4 d5e4 c3e4 d6e7 f1d3 g8f6 e1g1 c8d7 e4c3 e8g8 f1e1 b8c6 a2a3 d8c7 c1e3 f8d8 a1c1 c7a5 d3b1 a8c8 d1d3 a5a6 d3a6 b7a6 h2h3 a6a5 b1d3 h7h6 g2g4 e7d6 c3b5');
INSERT INTO evaluations VALUES(53595,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(53595,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(53595,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(53595,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(53595,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(53595,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(53595,36,'cp 25','g1f3 g8f6 f1d3 c6c5 e1g1 f8e7 d4c5 b8d7 b2b4 a7a5 c3b5 a5b4 b5d6 e7d6 c5d6 d7b8 c1b2 d8d6 d1c1 b8c6 a2a3 e6e5 d3b5 e5e4 f3e5 b4a3 e5c6 b7c6 b2a3 d6c7 b5e2 c6c5 c2c4 c8a6 a3c5 c7c5 c4d5 c5c1 f1c1 f6d5 e2a6 e8g8 a6c4 d5b6');
INSERT INTO evaluations VALUES(53620,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(53620,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(53620,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(53620,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(53620,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(53620,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53620,36,'cp -7','d5d4 c3b1 c6c5 g2g3 b8c6 f1h3 f8d6 a2a4 g8e7 b1a3 e8g8 c1d2 d6c7 e1g1 a7a6 d1e2 b7b6 f3h4 a8b8 f2f4 b6b5 e2h5 f7f5 e4f5 e6f5 a1e1 b5a4 a3c4 c6b4 d2b4 c5b4 h5g5 e7d5 h4f5 d8g5 f4g5 a4a3 b2a3');
INSERT INTO evaluations VALUES(54720,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(54720,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(54720,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(54720,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(54720,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(54720,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(54720,36,'cp 40','d2d4 b8c6 e2e3 e7e5 f4e5 f6e5 d4e5 c8g4 f1b5 f8c5 b1c3 g8e7 c3a4 c5b6 a4b6 c7b6 h2h3 g4h5 e1g1 e8g8 b5e2 d8c7 c1d2 c6e5 d2c3 e5f3 e2f3 h5f3 f1f3 c7c5 c3d4 c5c6 d1e2 f8f3 e2f3 a8f8');
INSERT INTO evaluations VALUES(55301,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(55301,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(55301,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(55301,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(55301,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(55301,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(55301,36,'cp 37','c7c5 f1g2 b8c6 d2d3 g8f6 g2c6 b7c6 g1f3 f6d5 c3e4 g7g6 e1g1 f8g7 f3d2 e8g8 f1e1 a7a5 a2a4 d5f6 e4f6 e7f6 d2c4 c8e6 c1e3 f8e8 d1f3 d8d5 f3d5 c6d5 c4d6 e8b8 d6b5 b8c8 b2b3 g7f8 e3f4 c8c6 h2h4 d5d4 b5a3 f8d6 f4d6 c6d6');
INSERT INTO evaluations VALUES(55442,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(55442,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(55442,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(55442,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(55442,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(55442,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(55442,36,'cp 29','d6d5 e4d5 e6d5 c4d3 f8d6 e1g1 c7c6 b1c3 g8e7 c3e2 e8g8 c1f4 c8f5 f1e1 f8e8 h2h4 d6f4 e2f4 f5g4 e1e3 b8d7 d1d2 g4f3 e3f3 d7f6 a1e1 d8d7 c2c3 e7f5 e1e8 a8e8 f4d5 d7d5 f3f5 d5a2 f5e5 a2a1 g1h2 a1a4 e5e8 f6e8');
INSERT INTO evaluations VALUES(55570,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(55570,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(55570,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55570,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(55570,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(55570,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(55570,36,'cp 0','e5d4 d1d4 b8c6 d4d1 f7f5 e4f5 c8f5 b1c3 d8d7 g1f3 g7g6 f1e2 f8g7 c1e3 e8c8 e1g1 g8e7 d1a4 c8b8 a1c1 f5g4 f1e1 e7f5 e3g5 g4f3 e2f3 d8e8 g5d2 a7a6 f3d5 h7h6 c3e4 c6e5 a4d7 e5d7 d2c3 g7c3');
INSERT INTO evaluations VALUES(56408,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(56408,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(56408,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(56408,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(56408,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(56408,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(56408,36,'cp 24','e4d5 d8d5 b1c3 d5e6 g1f3 c8d7 f1e2 b8c6 d2d4 c6d4 f3d4 e5d4 d1d4 g8e7 c1f4 e7f5 d4d2 e8c8 e1c1 f8d6 g2g4 d6f4 d2f4 f5d6 e2f3 h8e8 f4d4 e6h6 c1b1 c8b8 h2h4 d7c8 c3d5 c7c6 d5b4 h6f6 d4f6 g7f6 b1c1 d6c4 b4d3 b8c7 d1e1 a7a5 h1g1 e8e1 g1e1 c8e6 b2b4 c4a3 b4a5 a3c4 f3e2 c4e5');
INSERT INTO evaluations VALUES(56173,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(56173,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(56173,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(56173,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(56173,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(56173,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(56173,36,'cp 32','d8d5 b1c3 d5e6 g1f3 c8d7 f1e2 b8c6 d2d4 c6d4 f3d4 e5d4 d1d4 g8e7 c1f4 e7f5 d4d3 e8c8 g2g4 d7c6 g4f5 e6e7 f5f6 e7e6 d3e3 e6e3 f2e3 g7g5 c3b5 c6b5 f4g5 h8g8 h1g1 b5e2 e1e2 f8d6 h2h4 d6h2 g1h1 h2e5 c2c3 d8e8 a1f1 e8e6 e2d3 h7h6 g5f4 e5f4 f1f4');
INSERT INTO evaluations VALUES(34953,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(34953,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(34953,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(34953,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(34953,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(34953,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(34953,36,'cp 35','b1c3 d5e6 g1f3 c8d7 f1e2 b8c6 d2d4 c6d4 f3d4 e5d4 d1d4 g8e7 c1f4 e7f5 d4d2 e8c8 e1c1 f8d6 g2g4 d6f4 d2f4 f5d6 e2f3 h8e8 f4d4 e6h6 c1b1 c8b8 h2h4 a7a6 d4b4 d6b5 g4g5 h6g6 c3b5 a6b5 h1e1 d7c6 e1e8 d8e8 b4c3 c6f3 c3f3 f7f6 f3d5 f6g5 h4g5 g6c6 d5c6 b7c6 d1d7 e8e2');
INSERT INTO evaluations VALUES(56674,30,'cp 217',NULL);
INSERT INTO evaluations VALUES(56674,31,'cp 260',NULL);
INSERT INTO evaluations VALUES(56674,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(56674,33,'cp 245',NULL);
INSERT INTO evaluations VALUES(56674,34,'cp 228',NULL);
INSERT INTO evaluations VALUES(56674,35,'cp 245',NULL);
INSERT INTO evaluations VALUES(56674,36,'cp 263','c4f7 e8e7 f7h5 g8f6 f3h4 e7d7 h4g6 d8e8 g6h8 f6h5 d1g4 d7c6 g4g6 c8d7 b1c3 b7b6 g2g4 h5f6 c1h6 c6b7 h6g5 b8c6 g6e8 a8e8 f2f3 c6d4 e1f2 f8e7 g5f6 e7f6 h8g6 d7e6 h2h4 e6f7 h4h5 f6g5 g6h4 d4c2 a1d1 c2e3 d1g1 d6d5 e4d5 e3d5 c3d5');
INSERT INTO evaluations VALUES(56807,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(56807,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(56807,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(56807,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(56807,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(56807,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(56807,36,'cp 25','d2d4 c7c5 f1d3 g8h6 c2c3 b8c6 g1f3 f8e7 e1g1 e8g8 g1h1 b7b6 d1c2 h6f5 e5f6 e7f6 g2g4 c5c4 d3e2 f5d6 e2d1 b6b5 a2a3 d8a5 c2g2 b5b4 c1e3 f6e7 f1g1 a8b8 d1c2 d6e4 f4f5 b4c3 b2c3 e7d6 c2e4 d5e4 f3g5 e6e5 b1d2 e5d4');
INSERT INTO evaluations VALUES(57780,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(57780,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(57780,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57780,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(57780,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57780,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57780,36,'cp 9','d7d6 c1f4 c8g4 h2h3 d6e5 f4e5 g4f3 d1f3 g7e5 d4e5 c7c6 f3e3 e8g7 f1d3 b8d7 e1c1 d8b6 e3b6 d7b6 b2b3 f8d8 d3c2 b6d7 f2f4 a7a5 c3e2 d7c5 d1d8 a8d8 h1d1 d8d1 c1d1 f7f6 e5f6 e7f6 a2a3 g8f7 g2g4 f6f5 d1d2 c5e6 b3b4 a5b4 a3b4 f7f6 d2e3 g6g5 c2b1');
INSERT INTO evaluations VALUES(57777,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(57777,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(57777,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(57777,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(57777,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(57777,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(57777,36,'cp 0','c1f4 c8g4 h2h3 d6e5 f4e5 g4f3 d1f3 b8c6 e1c1 c6e5 d4e5 d8b8 f3e3 c7c6 d1e1 f7f6 e5f6 e7f6 c3e4 e8d6 e4d6 b8d6 c4c5 d6d5 e3e6 d5e6 e1e6 f8e8 f1c4 e8e6 c4e6 g8f8 h1d1 a8e8 e6b3 e8e5 d1d7 e5e7 d7d8 e7e8');
INSERT INTO evaluations VALUES(3920,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(3920,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(3920,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3920,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(3920,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(3920,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(3920,36,'cp 27','c3a4 b7b5 f3d4 b5c4 d4f5 c5b6 f5g3 c4d3 d1d3 h7h6 a4b6 a7b6 c2c4 d8e8 b2b3 f6d7 d3c2 d7c5 a2a4 c8d7 c1e3 c5b7 f2f4 e5f4 f1f4 b7c5 a1f1 f7f6 e3c1 d7e6 c1b2 c5d7 g3f5 e6f5 e4f5 e8f7 g1h2 d7e5 c2e4 f8e8');
INSERT INTO evaluations VALUES(5337,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(5337,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(5337,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(5337,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(5337,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(5337,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(5337,36,'cp -47','a7a6 g2g3 e7e5 d2d3 d7d6 f1g2 g8e7 e1g1 b7b5 f3d2 e8g8 c3d5 a8b8 a1b1 c6d4 c2c3 e7d5 e4d5 d4f5 b3b4 c5b4 c3b4 d8b6 d2b3 c8d7 b3a5 b8c8 b1c1 g7h6 c1c2 f5e7 f1e1 c8c2 d1c2 f8c8 c2e2 h6g7 e2d2 h7h5 e1c1');
INSERT INTO evaluations VALUES(5777,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(5777,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(5777,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(5777,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(5777,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(5777,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5777,36,'cp 30','g8f6 e1g1 e8g8 d4d5 c6e5 f3e5 d6e5 d1e2 c8g4 e2d3 c7c6 c1e3 c6d5 c4d5 g4e6 f1d1 e6d5 e4d5 b6e3 d3e3 d8d6 a1c1 f8c8 d1e1 d6b4 h2h3 h7h6 c1b1 c8e8 a2a3 b4d6 b1d1 a7a6 c3e4 f6e4 e3e4 a8d8 e1e3');
INSERT INTO evaluations VALUES(8328,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(8328,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(8328,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(8328,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(8328,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(8328,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(8328,36,'cp 57','e8g8 f1e2 e7e6 f3f4 c6d4 e3d4 b7b6 e1g1 d8e7 e2f3 c8b7 d1d2 e6e5 f4e5 d6e5 d4e3 f8d8 d2e2 f6e8 g2g3 g7f8 a1c1 e7b4 g1h1 f8c5 e3g5 f7f6 g5h6 c5d4 f3g4 f6f5 g4f3 e8d6 e4f5 e5e4 f3g2 d6f5 h6g5 d8e8 g3g4 f5d6 c3d5 b7d5 c4d5 d4e5 b2b3 a8c8 c1c6');
INSERT INTO evaluations VALUES(9731,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(9731,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(9731,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(9731,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(9731,34,'cp 153',NULL);
INSERT INTO evaluations VALUES(9731,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(9731,36,'cp 136','b5c6 d7c6 f3e5 d8e7 e5f3 c8g4 c1e3 e8c8 d1e2 f6d7 b1d2 g7g5 a1e1 c5d6 h2h3 g4h5 d2c4 f7f6 e3d4 c6c5 d4c3 d7e5 c4d6 c7d6 c3e5 f6e5 e2e3 h5f3 e3f3 h8f8 f3h5 e7e6 h5g4 e6g4 h3g4 c8c7 f2f3 b7b5 b2b3 a7a5');
INSERT INTO evaluations VALUES(9720,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(9720,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(9720,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(9720,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(9720,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(9720,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(9720,36,'cp 134','d7c6 f3e5 d8e7 e5f3 c8g4 c1e3 e8c8 b1d2 g7g5 d1e1 h8e8 a2a3 c8b8 b2b4 c5d6 f3d4 d6e5 c2c3 f6h5 d2c4 f7f6 f2f3 g4c8 g1h1 h5f4 e1d2 e8g8 c4a5 b8a8 a5b3 h6h5 g2g3 f4g6 a3a4 h5h4 g3g4 e5f4 a4a5 f4e3 d2e3');
INSERT INTO evaluations VALUES(10076,30,'cp -165',NULL);
INSERT INTO evaluations VALUES(10076,31,'cp -197',NULL);
INSERT INTO evaluations VALUES(10076,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(10076,33,'cp -157',NULL);
INSERT INTO evaluations VALUES(10076,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(10076,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(10076,36,'cp -149','c5b6 f3e2 a7a5 b4b5 c6e7 d2d3 c7c6 b5c6 b7c6 g1f3 e7g6 e2c2 e8g8 e1g1 d7d5 c4b3 d8c7 b3a4 d5e4 d3e4 c8a6 c3c4 b6d4 f3d4 e5d4 f2f4 f8d8 e4e5 d4d3 c2d2 f6g4 h2h3 g4h6 f4f5 g6h4 d2f2 c7e5 c1b2 e5e2 b1d2 h6f5 a1e1 e2f2 f1f2 f7f6 c4c5 f5g3 e1e7 h4g6 a4b3 g8h8');
INSERT INTO evaluations VALUES(9921,30,'cp -179',NULL);
INSERT INTO evaluations VALUES(9921,31,'cp -178',NULL);
INSERT INTO evaluations VALUES(9921,32,'cp -172',NULL);
INSERT INTO evaluations VALUES(9921,33,'cp -170',NULL);
INSERT INTO evaluations VALUES(9921,34,'cp -163',NULL);
INSERT INTO evaluations VALUES(9921,35,'cp -164',NULL);
INSERT INTO evaluations VALUES(9921,36,'cp -167','f3e2 e8g8 d2d3 a7a5 b4b5 c6e7 g1f3 e7g6 e1g1 c7c6 a2a4 d7d5 c4b3 f8e8 c1e3 c8e6 b1d2 c6b5 a4b5 d5e4 d3e4 e6b3 d2b3 a5a4 b3d2 b6e3 e2e3 d8c7 g2g3 g6f8 g1g2 f8e6 a1a3 h7h6 e3e2 e6c5 f1d1 e8d8 d1a1 b7b6');
INSERT INTO evaluations VALUES(10313,30,'cp -68',NULL);
INSERT INTO evaluations VALUES(10313,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(10313,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(10313,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(10313,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(10313,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(10313,36,'cp -76','b1c3 e8g8 c1g5 h7h6 g5f6 d8f6 d1d2 d7d5 e1c1 d5e4 c3e4 f6e5 e4c5 e5c5 d2c3 c5c3 b2c3 c8e6 a2a3 a8d8 f1d3 c6e5 g1e2 e6c4 e2f4 c4d3 f4d3 e5c4 a3a4 f8e8 d3b2 c4a5 d1d8 e8d8 h1d1 d8d1 c1d1 g8f8 d1d2 f8e7 c3c4 e7d6 d2c3 h6h5 g2g3 a5c6 b2d3 a7a6 d3f4');
INSERT INTO evaluations VALUES(10550,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(10550,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10550,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(10550,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10550,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10550,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10550,36,'cp 0','c6a5 c4b3 a5b3 a2b3 d7d6 b1c3 a7a5 g5f3 h6g8 d3d4 e5d4 f3d4 g8e7 e1g1 e8g8 c1g5 h7h6 g5h4 d8d7 f1e1 c7c6 c3a4 c5a7 a4c3');
INSERT INTO evaluations VALUES(11011,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(11011,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(11011,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(11011,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(11011,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(11011,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(11011,36,'cp 36','e5d4 c3d4 c5b4 c1d2 b4d2 d1d2 d7d5 e4d5 e7d5 b5c6 b7c6 e1g1 e8g8 b1c3 f7f6 a1c1 c8e6 h2h3 f8e8 f1e1 e6f7 c3e4 d8d7 c1c5 d5b6 e4c3 e8e1 d2e1 b6d5 e1d2 d7d6 c5a5 d5c3 d2c3 f7d5 b2b4 d5f3 c3f3 d6d4 f3c6');
INSERT INTO evaluations VALUES(11021,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(11021,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11021,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11021,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(11021,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(11021,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(11021,36,'cp 38','c3d4 c5b4 c1d2 b4d2 d1d2 d7d5 e4d5 e7d5 b5c6 b7c6 e1g1 e8g8 b1c3 f8e8 a1c1 c8e6 f1e1 f7f6 h2h3 d5c3 d2c3 e6d5 f3d2 a7a5 d2c4 e8e1 c1e1 d8d7 a2a3 a5a4 c3c2 d5f7 c4e3 f7b3 c2c5 a8e8 e1c1 e8e4 c5a7 e4d4 a7a8 d7d8 a8c6');
INSERT INTO evaluations VALUES(11020,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(11020,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(11020,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(11020,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(11020,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(11020,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(11020,36,'cp 36','c5b4 c1d2 b4d2 d1d2 d7d5 e4d5 e7d5 b5c6 b7c6 e1g1 e8g8 b1c3 c8e6 f3e5 d5e7 a1d1 f8e8 h2h3 d8d6 f1e1 f7f6 e5g4 e7g6 g4e3 a8d8 d2c2 d6b4 a2a3 b4b3 c2c1 e6f7 c3e2 g6e7 e2g3 g8f8 e3f5 e7f5 g3f5 b3b5 f5e3 d8d6 c1c3 e8d8 b2b4 b5b6');
INSERT INTO evaluations VALUES(11010,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(11010,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(11010,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(11010,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(11010,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(11010,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(11010,36,'cp 37','c1d2 b4d2 d1d2 d7d5 e4d5 e7d5 b5c6 b7c6 e1g1 e8g8 b1c3 f7f6 h2h3 f8e8 f1e1 c8e6 c3e4 e6f7 a1c1 d8d7 a2a3 d5e7 e4c5 d7d6 d2c2 e7g6 e1e3 f7d5 c5e4 d6d7 c1e1 d7f7 c2d2 a8d8 d2c3 d8a8 b2b3 a8d8 e4c5');
INSERT INTO evaluations VALUES(12647,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(12647,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(12647,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(12647,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(12647,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(12647,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(12647,36,'cp 160','d4d3 d7d5 e4d5 f8d6 d3e2 d8e7 e2e7 e8e7 a2a4 c8f5 b1a3 f5e4 e1g1 e4d5 f1e1 d5e6 a4a5 a8d8 g2g3 a7a6 c4e6 f7e6 a3c4 d6c7 b2b3 d8d5 c1a3 e7f7 e1e2 f6d7 a1e1 h8e8 a3b2 b7b5 a5b6 d7b6 c4e3 d5d8 e1a1 d8a8 a1a5 b6d7 a5a4 e8b8 e3c4');
INSERT INTO evaluations VALUES(14280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14280,36,'cp 0','h7h6 f1g2 c8f5 c3c4 e7e6 e1g1 f8d6 b2b3 e8g8 c1b2 f8e8 d1c1 f5h7 b1c3 c6b8 f3d2 b8c6');
INSERT INTO evaluations VALUES(14338,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(14338,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(14338,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(14338,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(14338,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(14338,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(14338,36,'cp -85','e7e5 e2e3 c8f5 d2d3 d8d7 b1d2 e8c8 a2a3 f8d6 b3b4 h7h5 h2h3 c8b8 g1e2 h8e8 d2b3 d5d4 d1d2 d4e3 f2e3 f5e6 e1c1 e6b3 c2b3 d7e6 g2c6 b7c6 e3e4 e6b3 d2c2 b3c2 c1c2 f6d7 g3g4 h5g4 h3g4 c6c5 c2b3 b8b7 d1f1 c5b4 a3b4');
INSERT INTO evaluations VALUES(20990,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(20990,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(20990,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20990,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20990,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(20990,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(20990,36,'cp 16','c2c4 e7e5 d4e5 f8b4 c1d2 d5c4 g1f3 g8h6 e2e3 h6g4 d1a4 b4d2 b1d2 c4c3 b2c3 e8g8 a4e4 d8e7 h4h5 c6e5 f3e5 e7e5 e4e5 g4e5 h5h6 f8d8 d2e4 c8f5 e4g3 f5g6 f2f4 e5d3 f1d3 d8d3 f4f5 d3e3 e1f2 e3g3 f2g3 g6f5 g3f2 f5g6 h6g7 g8g7 h1h4');
INSERT INTO evaluations VALUES(26630,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(26630,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(26630,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(26630,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(26630,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(26630,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(26630,36,'cp 97','e2f3 d8d7 c2c4 e7e6 b1c3 f8b4 e1g1 g8f6 d1b3 b4c3 b3c3 h7h5 b2b4 e8g8 a2a4 a7a6 c1d2 c6e7 c3c2 c7c6 f3e2 f8e8 f1c1 d5c4 c2c4 e7f5 h2h3 a8d8 d2e1 h5h4 e2f3 f6d5 c4b3 d7e7 c1c5 e7g5 b3c2 f5d6');
INSERT INTO evaluations VALUES(26627,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(26627,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(26627,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(26627,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(26627,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(26627,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(26627,36,'cp 81','e7e6 c2c4 d8d7 b1c3 g8f6 e1g1 f8b4 d1b3 b4c3 b3c3 h7h5 b2b4 e8g8 a2a4 a7a6 c1d2 c6e7 c3c2 c7c6 f3e2 f8e8 f1c1 d5c4 e2c4 a8d8 d2e1 e7d5 c2b2 f6e4 b4b5 c6b5 a4b5 e4d6 c4d5 e6d5 b5a6 b7a6 a1a6 d8a8 a6a1 h5h4');
INSERT INTO evaluations VALUES(27992,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(27992,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(27992,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(27992,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(27992,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(27992,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(27992,36,'cp 127','b2c3 d7d6 g1h3 g8f6 d1e2 e6e5 f2f4 e8g8 e1g1 f8e8 h3f2 b8d7 e2f3 h7h6 h2h4 e5f4 c1f4 e8e6 a1e1 d8e7 e1e3 a8e8 f1e1 f6h7 f3h3 b7c8 d3c2 d7f8 f4g3 h7f6 h3f5 e7d7 e4e5 d6e5 g3e5 f8g6');
INSERT INTO evaluations VALUES(29890,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(29890,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(29890,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(29890,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(29890,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(29890,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(29890,36,'cp 102','c7c6 a2a3 f8e7 c4a2 b8a6 d1e2 h5g6 f3h4 a6c5 c1e3 d8c7 e1c1 f6e4 c3e4 e7h4 e4c5 d6c5 e3c5 h4f6 f2f4 e8c8 e2g4 c7d7 g4d7 d8d7 f4e5 f6e5 c5a7 e5f4 c1b1 h8e8 h1f1 f4d6 d1e1 d7e7 e1e7 e8e7 b1c1 e7e2 a2f7 e2g2 a7b6 g6f7');
INSERT INTO evaluations VALUES(30554,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(30554,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(30554,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(30554,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(30554,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(30554,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(30554,36,'cp 95','e7e6 c3e2 c7c5 c2c3 b8c6 g1f3 b7a6 h2h4 h7h5 c1g5 d8d7 g2g3 c5d4 e2d4 a6f1 e1f1 g8e7 f1g2 a8c8 d1e2 e7g6 h1d1 c6d4 d1d4 f8c5 d4d3 d7a4 d3d2 c8c7 a1d1 a4c6 d1e1 c6a4 f3d4 c5d4 d2d4 a4c6 e1d1 g6e7 g5e7');
INSERT INTO evaluations VALUES(30580,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(30580,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(30580,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(30580,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(30580,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(30580,35,'cp 92',NULL);
INSERT INTO evaluations VALUES(30580,36,'cp 104','c3e2 c7c5 c2c3 b8c6 g1f3 b7a6 h2h4 h7h5 g2g3 d8d7 c1g5 a8c8 e2f4 a6f1 e1f1 g8e7 f1g2 e7f5 a1c1 f8e7 d1d3 a7a5 d3c2 d7b7 c2d2 c8a8 d2d3 b7d7 h1e1 a5a4 c1b1 a8c8 d3d1 c6a5 g5e7 f5e7 b2b3 e7f5 d1d3 a4b3 a2b3 a5c6');
INSERT INTO evaluations VALUES(31851,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(31851,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(31851,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(31851,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(31851,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(31851,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(31851,36,'cp -28','e7e6 g1f3 h7h6 e2e3 g8f6 f1e2 f8d6 b2b3 c7c6 c1b2 b8d7 c3c4 a7a5 e1g1 e8g8 a2a3 f6e4 d2e4 f5e4 c4c5 d6c7 b3b4 d8f6 f3d2 e4f5 d2f3 g7g5 b4b5 d7c5 b5b6 c7b8 d4c5 f6b2 e2d3 a5a4 d3f5');
INSERT INTO evaluations VALUES(31862,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(31862,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(31862,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(31862,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(31862,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(31862,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(31862,36,'cp 0','c2c4 e7e6 b1c3 c7c5 c4d5 c5d4 d1a4 b8d7 a4d4 e6d5 c3d5 g8e7 d5e7 f8e7 d4g7 e7f6 g7h6 a8c8 g2g3 c8c6 h6d2 d8c8 e2e4 f5e6 d2e3 e8g8 f1b5 c6c2 b5d3 c2c6 d3b5');
INSERT INTO evaluations VALUES(35038,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(35038,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(35038,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(35038,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(35038,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(35038,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(35038,36,'cp 94','b1c3 d7d5 g1f3 h4e7 c4b3 g8f6 d1e2 f6e4 c3e4 d5e4 e2e4 e7b4 e4e2 c8d7 c2c3 b4a5 f3e5 f8d6 e5c4 a5g5 c4d6 c7d6 e1g1 d7b5 c3c4 b5c6 f2f4 g5f6 f4f5 f6d4 e2f2 d4f2 f1f2 e6f5 f2f5 e8g8 d2d4 c6e4 f5f4 f8e8 c1d2 b8c6 a1e1 e4g6 e1e8 a8e8 d2c3 g8f8 g1f2 e8c8 f2e2 a7a6');
INSERT INTO evaluations VALUES(36012,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(36012,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(36012,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36012,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(36012,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(36012,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36012,36,'cp 15','b8c6 e1g1 c8d7 f2f3 c6d4 e3d4 a7a6 a2a4 d8a5 f1d1 f8e8 c3d5 a5d2 d5f6 g7f6 d1d2 f6d4 d2d4 e8c8 c2c3 b7b5 a4b5 d7b5 g1f2 a8b8 d4d2 h7h5 f2e3 g8g7 e3d4 e7e5 d4e3 b8b6 e2d3 h5h4 a1a5 g7f6 g2g3 c8h8 a5a1 b5d3 e3d3 h4g3 h2g3');
INSERT INTO evaluations VALUES(36394,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(36394,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(36394,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(36394,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(36394,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(36394,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(36394,36,'cp 12','h7h6 g5h4 b8c6 b1d2 f6h5 h4e7 c6e7 g2g3 h5f6 c2c3 c8f5 f3e5 f8e8 f1e1 c7c6 g1g2 d8c7 d1c2 f5d3 e5d3 e7f5 e1e5 f5d6 a1e1 f6d7 e5e8 a8e8 e1e8 d6e8 d2f3 d7f8 c2e2 f8e6 h2h4 e8d6 e2e1 d6c4 b2b3 c4d6 h4h5 c7d7');
INSERT INTO evaluations VALUES(38305,30,'cp -137',NULL);
INSERT INTO evaluations VALUES(38305,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(38305,32,'cp -157',NULL);
INSERT INTO evaluations VALUES(38305,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(38305,34,'cp -147',NULL);
INSERT INTO evaluations VALUES(38305,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(38305,36,'cp -128','c7c6 b5d3 g8f6 g1f3 e8g8 e1g1 c6c5 e2d1 b8c6 h2h3 h7h6 c2c3 c8e6 f1e1 f8e8 d3f1 d8b6 d2d3 e7d6 b1a3 a7a6 a3c2 d5d4 c3d4 c5d4 c2a3 b6d8 a3c4 d6c7 a2a4 f6d5 c1d2 d8f6 a4a5 a8c8 d1b3 f6g6 f3h4 g6h5 h4f3');
INSERT INTO evaluations VALUES(37882,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(37882,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(37882,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(37882,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(37882,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(37882,35,'cp -132',NULL);
INSERT INTO evaluations VALUES(37882,36,'cp -146','b5d3 g8f6 g1f3 e8g8 e1g1 c6c5 e2d1 b8c6 h2h3 h7h6 c2c3 f8e8 f1e1 d8b6 d3f1 c8f5 d2d3 e7d6 e1e8 a8e8 b1a3 a7a6 a3c2 d5d4 c3d4 c5d4 c1d2 f5e6 b2b4 f6d5 a2a3 d6b8 a1c1 b8a7 d1e1 e6d7 e1d1 d5c3 d2c3 d4c3 d3d4 d7f5 c2e3');
INSERT INTO evaluations VALUES(38345,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(38345,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38345,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(38345,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(38345,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38345,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(38345,36,'cp 24','c7c5 c2c3 d8b6 b1d2 c5d4 c3d4 g7d4 d2c4 d4f2 e1e2 b6c5 c4d6 c5d6 d1d6 e7d6 e2f2 b8d7 f1b5 h7h6 g5f4 g8f6 h1e1 a7a6 b5d7 c8d7 f2g1 e8g8 a1d1 f8d8 f4h6 f6e8 h6d2 d7a4 d1c1 a8c8 d2c3 a4b5 f3d4 b5d7 a2a3 g8g7 d4b5 g7g8');
INSERT INTO evaluations VALUES(38898,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(38898,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(38898,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(38898,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(38898,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(38898,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(38898,36,'cp 112','g8f6 e4e5 f6e4 d1a4 b8c6 c1b2 f7f5 e5f6 g7f6 f1c4 c8d7 a4c2 e4d6 c4d3 d8e7 e1g1 e7g7 b2a3 e8c8 a3d6 c7d6 a1b1 c8b8 f1e1 c6a5 c3c4 d7c6 d4d5 e6d5 f3d4 d5c4 d4c6 a5c6 d3c4 d6d5 c4a6 c6a5 c2a4 g7c7 a4d4 b8a8 a6f1 a8b8 g2g3');
INSERT INTO evaluations VALUES(39608,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(39608,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(39608,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(39608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39608,36,'cp 0','c7c5 c2c3 d8b6 g1f3 b8c6 d4c5 b6b2 b1d2 b2a3 f4e3 g7c3 a1b1 g8f6 b1b3 a3a5 d1c2 c3b4 a2a3 b4d2 e3d2 a5a4 c2b2 e8g8 b3e3 c6a5 d2b4 f6g4 e3c3 a5c6 f1c4 a7a5 c4b3 a4b5 b3c4');
INSERT INTO evaluations VALUES(40645,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(40645,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(40645,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(40645,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(40645,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(40645,35,'cp -79',NULL);
INSERT INTO evaluations VALUES(40645,36,'cp -87','d5d4 g5f6 d8f6 c3d5 f6d8 e2e4 d4e3 f2e3 e8g8 a1c1 b8d7 b2b4 c5d6 f1c4 d7e5 e1g1 c8g4 g1h1 e5f3 g2f3 d8h4 d5f6 h4f6 f3g4 f6e5 d1e2 g7g6 c4b3 a8e8 b4b5 e8e7 c1d1 d6c5 e3e4 c5d4 e2d3 e5e4 d3e4 e7e4 g4g5 g8g7');
INSERT INTO evaluations VALUES(41490,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41490,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41490,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41490,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41490,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(41490,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41490,36,'cp 0','e8g8 e1g1 c8f5 f3h4 f5c8');
INSERT INTO evaluations VALUES(36769,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36769,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36769,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36769,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36769,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36769,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36769,36,'cp 0','e1g1 c8f5 f3h4 f5c8');
INSERT INTO evaluations VALUES(45216,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(45216,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(45216,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(45216,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45216,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(45216,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(45216,36,'cp 8','g7g6 d2d4 c5d4 f3d4 f8g7 f2f3 e8g8 c1e3 b8c6 d1d2 d6d5 e1c1 d5e4 d4c6 d8d2 d1d2 b7c6 c3e4 f6d5 e3c5 g7h6 c2c4 a8b8 c4d5 c6d5 e4c3 h6d2 c1d2 b8b2 d2c1 b2b7 c3d5 c8e6 f1a6 e6d5 a6b7 d5b7 c5e7 f8c8 c1b2 b7d5 h1e1 c8e8 e7b4 e8e1 b4e1');
INSERT INTO evaluations VALUES(46656,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(46656,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(46656,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(46656,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(46656,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(46656,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(46656,36,'cp 37','g1f3 f8g7 d2d4 e8g8 f1g2 c7c5 d4d5 e7e6 e1g1 f8e8 c1f4 e6e5 f4g5 h7h6 g5f6 d8f6 e2e4 f6e7 g1h1 a7a6 a2a4 b7b6 d1d3 b8d7 g2h3 h6h5 a1e1 g7f6 f3d2 g8g7 f2f4 h5h4 f4f5 e8h8 g3g4 f6g5 d2f3 g5f4 b2b3 a8b8');
INSERT INTO evaluations VALUES(47384,30,'cp -186',NULL);
INSERT INTO evaluations VALUES(47384,31,'cp -177',NULL);
INSERT INTO evaluations VALUES(47384,32,'cp -162',NULL);
INSERT INTO evaluations VALUES(47384,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(47384,34,'cp -163',NULL);
INSERT INTO evaluations VALUES(47384,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(47384,36,'cp -171','b1c3 c8f5 h2h3 e7e6 g2g4 f5g6 g1e2 c7c5 g4g5 f6d5 d4c5 f8c5 f1g2 b8c6 c3e4 d8b6 e4c5 b6c5 g2d5 e6d5 c1e3 c5b4 c2c3 b4b2 e1g1 g6h5 d1d2 b2d2 e3d2 e8g8 f2f3 c6e5 e2d4 f7f6 d2f4 e5c4 g5f6 f8f6 f4g5 f6g6');
INSERT INTO evaluations VALUES(47379,30,'cp -176',NULL);
INSERT INTO evaluations VALUES(47379,31,'cp -170',NULL);
INSERT INTO evaluations VALUES(47379,32,'cp -175',NULL);
INSERT INTO evaluations VALUES(47379,33,'cp -177',NULL);
INSERT INTO evaluations VALUES(47379,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(47379,35,'cp -174',NULL);
INSERT INTO evaluations VALUES(47379,36,'cp -175','c8f5 h2h3 h7h6 c1e3 e7e6 g2g4 f5h7 f1g2 b8c6 g1e2 f8d6 e1g1 e8g8 e2g3 d6g3 f2g3 f6d5 c3d5 e6d5 d1d2 d8d7 f1f2 c6d8 a2a4 h7g6 g2f1 f7f6 c2c4 c7c6 a1c1 d8e6 h3h4 a8d8 c4d5 d7d5 d2b4 d5d6 b4d6');
INSERT INTO evaluations VALUES(51579,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(51579,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(51579,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(51579,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(51579,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(51579,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(51579,36,'cp 168','c5b6 d8b6 g1f3 g8f6 b2b3 g7g6 e2e3 f8g7 c1b2 e8g8 f1e2 f8d8 e1g1 d7d5 d1c1 b8c6 b1d2 c8f5 c2c4 a8c8 c4d5 c6b4 d2c4 b6c5 f1d1 b4c2 a1b1 c2b4 b2d4 c5c7 d4e5 c7b7 b1b2 d8d5 d1d5 b7d5 c1f1');
INSERT INTO evaluations VALUES(51576,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(51576,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(51576,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(51576,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(51576,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(51576,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(51576,36,'cp 163','d8b6 g1f3 g8f6 b2b3 g7g6 e2e3 f8g7 c1b2 e8g8 b1d2 b8c6 c2c4 f8d8 f1e2 a7a5 e1g1 a5a4 d1c1 d7d6 f3d4 c8d7 d4b5 c6e5 b2c3 d8c8 c3d4 b6b7 f2f4 e5c6 e2f3 c6d4 f3b7 d4e2 g1h1 e2c1 b7c8');
INSERT INTO evaluations VALUES(51852,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(51852,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(51852,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(51852,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(51852,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51852,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(51852,36,'cp 39','d7d5 d4e5 d5d4 c3e4 d8a5 c1d2 a5e5 e4g3 c6c5 g1f3 e5c7 e2e3 d4e3 d2e3 g8f6 f1d3 f8e7 h2h3 e8g8 e1g1 f8d8 d1e2 b8c6 f1e1 e7d6 g3e4 f6e4 d3e4 h7h6 a1d1 c8e6 e3d2 d6f8 d2c3 d8d1 e1d1 a8e8 b2b3 e8d8 d1d8 c6d8 c3b2 d8c6 e2d2 f8d6 b2c3 c7e7');
INSERT INTO evaluations VALUES(53757,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(53757,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(53757,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(53757,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(53757,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(53757,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(53757,36,'cp 52','d4e5 d5d4 c3e4 d8a5 c1d2 a5e5 e4g3 e5d6 g1f3 g8f6 d1c2 c6c5 e1c1 d6b6 e2e3 f8e7 f1d3 b8c6 e3d4 c5d4 h1e1 e8g8 a2a3 h7h6 g3f5 c8f5 d3f5 a8d8 f3e5 f8e8 e5d3 g7g6 f5h3 e7f8 e1e8 d8e8 d1e1 e8e1 d2e1 b6c7 f2f4 c6a5 c4c5 f8g7 g2g3 a5c6');
INSERT INTO evaluations VALUES(53887,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(53887,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(53887,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(53887,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(53887,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(53887,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(53887,36,'cp 74','e2e3 f7f5 h2h4 g8f6 d2d4 e5e4 f3g5 f8e7 d4c5 d6c5 d1d8 e7d8 b2b3 h7h6 g5h3 d8c7 c1b2 b8c6 e1c1 c7e5 c3d5 e8f7 h3f4 e5b2 c1b2 f6d5 d1d5 h8d8 d5d8 c6d8 f1e2 c8e6 f2f3 g7g5 h4g5 h6g5 f4h3 f7f6 f3e4 d8c6 e4f5 e6f5 h1f1 a8d8');
INSERT INTO evaluations VALUES(53886,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(53886,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(53886,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(53886,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(53886,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(53886,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(53886,36,'cp 57','f7f5 h2h4 g8f6 d2d4 e5e4 f3g5 b8c6 c3d5 c5d4 e3d4 f8e7 d5f4 d8b6 c1e3 b6b4 d1d2 h7h6 g5h3 e8f7 h4h5 b4d2 e1d2 h8d8 a1c1 a7a5 f1e2 d6d5 c4d5 c6b4 c1c7 f6e8 c7c3 e8f6 e2c4 e7d6 a2a3 b7b5 c4b5 b4d5 f4d5 f6d5 b5c6 d5e3 c6a8 e3g2 h1g1 g2h4 a8d5 f7f8 d2c2');
INSERT INTO evaluations VALUES(55032,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(55032,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(55032,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(55032,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(55032,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(55032,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(55032,36,'cp -62','e7e5 e2e3 g8f6 c1b2 f8d6 c2c4 c7c6 g1f3 b8d7 d2d3 e8g8 f1e2 f8e8 d1c2 a7a5 c4d5 c6d5 b4a5 d8a5 b1c3 d7c5 d3d4 e5d4 f3d4 d6e5 e1g1 e5d4 e3d4 c5e4 c3e4 d5e4 f1d1 c8g4 e2g4 f6g4 h2h3');
INSERT INTO evaluations VALUES(57118,30,'cp -156',NULL);
INSERT INTO evaluations VALUES(57118,31,'cp -171',NULL);
INSERT INTO evaluations VALUES(57118,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(57118,33,'cp -167',NULL);
INSERT INTO evaluations VALUES(57118,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(57118,35,'cp -155',NULL);
INSERT INTO evaluations VALUES(57118,36,'cp -155','f8b4 g1f3 d7d6 e2e3 c7c5 c1b2 g8f6 a2a3 b4a5 b1c3 e8g8 f1e2 f8e8 e1g1 b8c6 d2d3 a5c3 b2c3 d6d5 c4d5 f6d5 c3b2 b7b6 h2h3 d8e7 f1e1 c8b7 e2f1 e7d7 d1c2 a8d8 a1d1 f7f6 f1e2 h7h6 f3h4');
INSERT INTO evaluations VALUES(10868,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(10868,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(10868,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(10868,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(10868,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(10868,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(10868,36,'cp -72','e5b5 c6d4 b5d3 c7c5 c2c3 d4c6 g1f3 d7d5 e4d5 f6d5 f1e2 e8g8 e1g1 c8e6 f1e1 d8d7 e2f1 a8d8 b1d2 e6f5 d3c4 f8e8 d2e4 c6a5 c4b5 d7c7 e4g3 a7a6 b5e2 f5g6 c1g5 f7f6 g5d2 e7f8 e2d1 e8e1 d1e1 d8e8 e1c1 g6f7 c3c4 d5b4 a2a3 b4c6 c1c3 g8h8');
INSERT INTO evaluations VALUES(11003,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(11003,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(11003,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(11003,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(11003,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(11003,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(11003,36,'cp 76','c3d4 a7a6 b5c4 b7b5 c4b3 c6a5 c1d2 a5b3 d1b3 d7d5 e4e5 e8g8 d2b4 f8e8 f1c1 c8d7 b1d2 c7c6 b4c5 a6a5 c1e1 e7f5 a2a3 h7h5 h2h3 g7h6 d2f1 h6f4 b3c3 d7e6 b2b4 f4h6 f1d2 a5a4 d2f1 g8h7 a1d1 h6f4 c3d3 d8d7');
INSERT INTO evaluations VALUES(13028,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(13028,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13028,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13028,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13028,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13028,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13028,36,'cp 0','f6g4 d4c6 b7c6 e3f4 f8g7 e1g1 e8g8 d1d2 g4e5 c4b3 a8b8 h2h3 c6c5 a1b1 g8h8 f1e1 d8a5 f4g5 f7f6 g5e3 g6g5 d2e2 c8a6 e2d2');
INSERT INTO evaluations VALUES(15313,30,'cp -179',NULL);
INSERT INTO evaluations VALUES(15313,31,'cp -185',NULL);
INSERT INTO evaluations VALUES(15313,32,'cp -198',NULL);
INSERT INTO evaluations VALUES(15313,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(15313,34,'cp -182',NULL);
INSERT INTO evaluations VALUES(15313,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(15313,36,'cp -196','c8e6 f3e2 d8d7 g1f3 e8c8 e1g1 f7f6 d3d4 e6g4 h2h3 g4f3 e2f3 e5d4 c4b5 c8b8 b5c6 d7c6 c3d4 d5b4 b1d2 c6f3 d2f3 b4c2 a1b1 c2d4 f3d4 d8d4 c1e3 d4a4 a2a3 f8d6 f1d1 h8e8 b1c1 a4e4 d1d2 b7b6 g2g3 b8b7 g1g2 a7a5 c1d1 h7h5 h3h4 a5a4 g2f3');
INSERT INTO evaluations VALUES(25107,30,'cp -189',NULL);
INSERT INTO evaluations VALUES(25107,31,'cp -190',NULL);
INSERT INTO evaluations VALUES(25107,32,'cp -210',NULL);
INSERT INTO evaluations VALUES(25107,33,'cp -206',NULL);
INSERT INTO evaluations VALUES(25107,34,'cp -182',NULL);
INSERT INTO evaluations VALUES(25107,35,'cp -195',NULL);
INSERT INTO evaluations VALUES(25107,36,'cp -155','f3e2 f7f6 g1f3 g7g5 h2h3 d8d7 c1e3 e8c8 b1d2 c6a5 c4b3 a5b3 a2b3 a7a6 d3d4 d5f4 e3f4 e5f4 e1c1 h7h5 e2d3 h8h6 b3b4 g5g4 f3h4 f6f5 d2c4 d7a4 h3g4 f5g4 c4e5 e6b3 d1e1 a6a5 b4b5 c8b8 h4g6 a4a2 g6f4');
INSERT INTO evaluations VALUES(15681,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15681,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(15681,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(15681,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(15681,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(15681,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(15681,36,'cp 50','c1e3 f8e7 f1d3 c8e6 d1e2 d8d7 e1c1 e7f6 h1e1 e8c8 d3b5 c8b8 f3g5 f6g5 e3g5 d8e8 e2d2 a7a6 b5f1 c6e7 b2b3 h7h6 g5f4 e7f5 c1b2 g7g5 f4e3 f5e3 e1e3 f7f6 d2d4 d7g7 h2h4 e6d7 d1e1 e8e3');
INSERT INTO evaluations VALUES(16459,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16459,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(16459,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(16459,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(16459,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(16459,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(16459,36,'cp -13','d7d5 c4d5 f6d5 b1c3 c8e6 g1f3 f8e7 e1g1 e8g8 a2a3 d8d7 c1d2 a8d8 c3d5 e6d5 d2c3 d7e6 b2b4 e7f6 f1e1 c6d4 c3d4 e5d4 d1a4 a7a6 a4c2 c7c6 f3d2 h7h6 a1b1 f8e8 d2e4 f6e7 e4c5 e7c5 g2d5 d8d5 b4c5');
INSERT INTO evaluations VALUES(18786,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(18786,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(18786,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(18786,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(18786,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(18786,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(18786,36,'cp 125','d4d5 c6e7 b1c3 d7d6 e3e4 a7a6 a2a4 g7g6 f2f4 f8g7 g1f3 g8f6 f1e2 e8g8 e1g1 a8b8 h2h3 e6d5 c4d5 b7b5 a4b5 a6b5 e4e5 d6e5 d5d6 e7f5 f4e5 f6d7 c1g5 f7f6 g5f4 d7e5 d1d5 e5f7 d6d7 d8d7 d5d7 c8d7 f4b8 f8b8 a1a7');
INSERT INTO evaluations VALUES(19476,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(19476,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(19476,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(19476,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(19476,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(19476,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(19476,36,'cp -39','g8f6 f2f4 g7g6 f1e2 f8g7 b1c3 e8g8 e2f3 c6d4 c3e2 d4f3 g1f3 d7d6 e1g1 e7e6 h2h3 b7b5 e2g3 f8e8 c1e3 f6d7 d1d2 f7f5 a1b1 c8b7 b2b4 c5c4 b1e1 d7f6 e4f5 e6f5 d3c4 b5c4 e3d4 d8c7 c2c3 e8e1 f1e1 b7f3 g2f3 a8e8');
INSERT INTO evaluations VALUES(20556,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(20556,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(20556,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(20556,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(20556,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(20556,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(20556,36,'cp -14','d2d4 e5e4 f3e5 c8e6 e1g1 g8f6 b1c3 a8c8 f2f3 c6d4 b3a4 d4c6 f3e4 f8c5 g1h1 e8g8 e4d5 f6d5 e5c6 b7c6 c3e4 c5e7 d1e2 d5f6 a4b3 f6e4 e2e4 e6d5 e4f4 e7f6 c2c3 d8e7 c1e3 f8e8 a1e1 e7b7 f4g3 f6e5 g3f2 d5b3 a2b3 b7b3 e3a7 f7f6 h1g1 e5d6 e1e8 c8e8 a7d4 g8f7 h2h3');
INSERT INTO evaluations VALUES(20949,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(20949,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(20949,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(20949,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(20949,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(20949,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(20949,36,'cp -57','e7e5 e2e3 a7a6 b1c3 c8e6 f1e2 f8e7 d4e5 c6e5 g1f3 e5f3 e2f3 g8f6 d1c2 e8g8 e1g1 e7c5 c3d5 e6d5 c2c5 d5f3 g2f3 d8d7 c5g5 f8e8 e3e4 e8e6 g1h1 d7d3 c1e3 f6e8 a1e1 d3c2 e3d4 c2c4 g5d5 c4d5');
INSERT INTO evaluations VALUES(21552,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(21552,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(21552,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(21552,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(21552,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(21552,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(21552,36,'cp -43','e2e3 g7g6 b1c3 g8e7 f1b5 c8d7 g1e2 f8g7 e1g1 e5d4 e3d4 e8g8 c1g5 h7h6 g5f4 a7a6 b5c6 d7c6 f4e5 f7f6 e5g3 d8d7 e2f4 g6g5 f4h5 g7h8 f2f4 f6f5 d1d2 g5g4 g3f2 e7g6 f1e1 a8e8 a1d1 e8e1 d1e1 f8e8 e1e8 d7e8 h5g3 e8d7 f2e3');
INSERT INTO evaluations VALUES(21551,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(21551,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(21551,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(21551,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(21551,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(21551,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(21551,36,'cp -56','a7a6 b1c3 c8f5 d1a4 e5d4 e3d4 g7g6 g1f3 f8g7 f1e2 g8e7 e1g1 e8g8 c1f4 f5e6 e2d3 b7b5 c5b6 c7b6 c3e2 e6d7 a4d1 d7g4 f3e5 c6e5 f4e5 e7c6 e5g7 g4e2 d3e2 g8g7 e2f3 d8d6 a1c1 f8d8 d1d2 d6f6 f1d1 h7h5 c1c3 d8d6 g2g3 a8e8 g1g2 h5h4');
INSERT INTO evaluations VALUES(21284,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(21284,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(21284,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(21284,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(21284,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(21284,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(21284,36,'cp 7','g8f6 e1g1 f8e7 a2a4 e8g8 c2c3 d6d5 e4d5 f6d5 b1d2 c8f5 f1e1 a7a6 d2f1 f5g6 a4a5 d8d7 e3d2 e7f6 d1b3 f8d8 d2e3 g8h8 f1g3 f6e7 e3d2 f7f6 h2h3 e7f8 g3e4 f8e7 d3d4 a8b8 d4e5 c6e5 f3e5 f6e5 e4g3');
INSERT INTO evaluations VALUES(21603,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(21603,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(21603,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(21603,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(21603,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(21603,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(21603,36,'cp -11','g8f6 b1c3 f8d6 c3b5 d6e7 f1d3 a7a6 b5c3 e8g8 h2h3 e7d6 c1d2 h7h6 e1g1 c8e6 f1e1 f8e8 a1c1 c6e7 d1c2 c7c6 e3e4 f6e4 c3e4 d5e4 d3e4 e7d5 a2a3 d8f6 e4h7 g8h8 h7e4 e8e7 f3e5 d6e5 d4e5 f6e5 d2c3');
INSERT INTO evaluations VALUES(22133,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(22133,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(22133,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(22133,33,'cp 178',NULL);
INSERT INTO evaluations VALUES(22133,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(22133,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(22133,36,'cp 164','g8f6 c2c3 d6d5 e4d5 f6d5 c3d4 c7c6 d1b3 f8d6 b1c3 e8g8 c4d5 c6d5 b3d5 d8b6 a1b1 c8e6 d5e4 e6d7 d4d5 a8e8 c1e3 b6c7 e4h4 f7f5 f1e1 h7h6 b1c1 c7b8 a2a3 b7b5 h2h3 a7a6 h4d4 g8h7 d4a7 b8a7 e3a7 e8e1 c1e1 b5b4 a3b4 d6b4 e1a1 a6a5');
INSERT INTO evaluations VALUES(15761,30,'cp 150',NULL);
INSERT INTO evaluations VALUES(15761,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(15761,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(15761,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(15761,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(15761,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(15761,36,'cp 147','c2c3 d6d5 e4d5 f6d5 c3d4 c7c6 b1c3 f8d6 d1b3 e8g8 c4d5 c6d5 b3d5 d8b6 a1b1 f8e8 c1e3 e8e6 f1c1 e6g6 c3e4 c8d7 b2b4 a7a6 e4d6 b6d6 d5d6 g6d6 c1c7 d7c6 f2f3 a8d8 g1f2 d6d7 c7d7 d8d7 f2e2 h7h6 g2g3');
INSERT INTO evaluations VALUES(22283,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(22283,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(22283,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(22283,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(22283,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(22283,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(22283,36,'cp 93','c4b5 f8d6 d2d4 e5e4 f3e5 d6e5 b5c6 b7c6 d1h5 e8f8 d4e5 d8e8 h5e8 f8e8 c1e3 c8f5 b1d2 g8e7 e1g1 e7g6 f2f4 e4f3 f1f3 f5c2 e3c5 c2e4 d2e4 d5e4 f3f5 e8d7 f5f7 d7e6 f7g7 g6e5 g7c7 e5d3 c5e3 e6d5 a1f1 h8f8 b2b3 f8f1 g1f1 a7a5 c7d7 d5e5 e3d4 e5e6 d7h7 a8f8 f1e2 d3f4 e2e3');
INSERT INTO evaluations VALUES(22273,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(22273,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(22273,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(22273,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(22273,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(22273,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(22273,36,'cp 81','f8d6 d2d4 e5e4 f3e5 d6e5 b5c6 b7c6 d1h5 e8f8 d4e5 d8e8 h5e8 f8e8 c1e3 c8f5 b1d2 g8e7 f2f3 e4f3 d2f3 f5g4 e1g1 h8f8 f3d4 e7f5 d4e6 e8e7 e6f8 f5e3 f8h7 e3f1 a1f1 c6c5 h7g5 g4e6 b2b3 a7a5 f1f4 a8b8 f4a4 b8b5 g5e6 e7e6 a4g4 e6e5 g4g7 e5d4 g7c7 d4c3 c7c8');
INSERT INTO evaluations VALUES(22300,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(22300,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(22300,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(22300,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(22300,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(22300,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(22300,36,'cp 117','c6a5 c4g8 h8g8 d3d4 c7c5 d4d5 a7a6 h2h4 h7h5 b2b3 g7g6 d1d3 b7b5 e1g1 f8e7 c1e3 e8f7 c3e2 g8g7 a2a4 b5b4 c2c3 a8b8 a1b1 b4c3 d3c3 b8b4 e2g3 f7g8 c3c2 b4b8 e3h6 g7f7 h6d2 c8d7 b3b4 c5b4 b1b4 b8b4');
INSERT INTO evaluations VALUES(26731,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(26731,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(26731,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(26731,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(26731,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(26731,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(26731,36,'cp 115','h2h3 g4e6 c4b3 g8f6 d2d4 e6d7 c1e3 f8e7 d1d3 e8g8 a2a3 e5d4 f3d4 c6d4 e3d4 d7e6 a1d1 e6b3 c2b3 c7c6 d3g3 d8c7 b3b4 a8d8 b4b5 f8e8 b2b4 d8d7 f2f4 c7d8 f1e1 c6c5 d4f6 e7f6 c3d5 c5b4 a3b4');
INSERT INTO evaluations VALUES(26730,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(26730,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(26730,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(26730,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(26730,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(26730,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(26730,36,'cp 84','g4e6 c4b3 g8f6 d2d4 e6d7 c1e3 f8e7 a2a4 e8g8 d1e2 f8e8 a1d1 e5d4 f3d4 e7f8 e2f3 a7a6 d4c6 b7c6 e3d4 d8e7 f1e1 f6h7 a4a5 a8b8 f3e2 h7g5 e2a6 g5h3 g2h3 e7g5 g1f1 g5h5 e1e3 c6c5 a6e2 h5e2 c3e2 c5d4 e2d4 c7c5');
INSERT INTO evaluations VALUES(28047,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(28047,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(28047,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(28047,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(28047,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(28047,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(28047,36,'cp 90','d7d6 a2a4 b8d7 a4a5 a7a6 c2c4 g8f6 f1e1 e8g8 c1d2 e7e6 a5b6 c7b6 b1c3 e6e5 d4d5 d7c5 d3c2 a6a5 d1e2 a8b8 c3b5 b7a6 a1a3 a6b5 c4b5 b8c8 g2g3 c5d7 c2d3 d7c5 e1a1 f6d7 b2b4 a5b4 d2b4');
INSERT INTO evaluations VALUES(28488,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(28488,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(28488,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(28488,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(28488,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28488,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28488,36,'cp 12','e8g8 d1b3 c7c5 d4c5 b4c5 c4d5 e6d5 f1e2 b8c6 e1g1 a8c8 b3b7 c6b4 c3b5 c8b8 b7c7 d8c7 b5c7 f8c8 c7b5 f5d3 b5d4 c5d4 f3d4 d3e2 d4e2 c8c2 d2b4 b8b4 f1c1 b4b2 c1c2 b2c2 e2d4 c2b2 d4c6 f6e4 f2f3 e4c3 c6a7 b2a2 a1a2 c3a2 a7b5 a2b4');
INSERT INTO evaluations VALUES(28847,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(28847,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(28847,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(28847,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(28847,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(28847,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(28847,36,'cp 99','b5d7 d8d7 d4e2 h7h6 c1e3 b8a6 e1g1 a6c7 d1d3 f8e7 e2g3 d6d5 e4d5 c7d5 a1d1 d5c3 d3c3 d7e6 f1e1 e8g8 e3h6 f8d8 h6d2 d8d5 c3f3 a8d8 d2c3 d5d1 e1d1 d8d1 f3d1 e6a2 c3e5 a2e6 e5d4 a7a5 c2c3 b7b5 d1e2 b5b4 e2e6 f7e6 g3e2 f6d5 c3b4 e7b4 g2g3 a5a4 h2h3 d5c7 g1g2');
INSERT INTO evaluations VALUES(28838,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(28838,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(28838,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(28838,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(28838,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(28838,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(28838,36,'cp 120','d8d7 d4e2 h7h6 c1e3 b8c6 d1d2 f8e7 f2f3 e8g8 h2h4 d6d5 c3d5 f6d5 e4d5 c6d4 e1c1 d7d5 c1b1 d5c6 e2d4 e5d4 e3d4 f8d8 d2f4 a8c8 f4e4 c6e4 f3e4 a7a6 c2c3 b7b5 b1c2 b5b4 h4h5 d8e8 h1f1 b4c3 b2c3 e7f8 c2d3 e8d8');
INSERT INTO evaluations VALUES(32556,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(32556,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(32556,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(32556,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(32556,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(32556,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(32556,36,'cp 88','d4e2 h7h6 c1e3 b8c6 d1d2 f8e7 f2f3 e8g8 h2h4 b7b5 g2g4 b5b4 c3d5 f6d5 e4d5 c6d4 e3d4 e5d4 e1c1 d7b5 e2d4 b5d5 c1b1 e7f6 g4g5 d5d4 d2d4 f6d4 d1d4 a7a5 g5h6 g7h6 d4d6 g8h7 d6f6 h7g7 f6f5 g7g6 f5d5 f8d8 d5b5 a8a7 b1c1 d8e8 h4h5 g6h7 h1d1 e8e2 b2b3 e2h2 d1d5');
INSERT INTO evaluations VALUES(29010,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(29010,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(29010,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(29010,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(29010,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(29010,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(29010,36,'cp 35','d5c4 e2e3 e7e6 f1c4 f8d6 h2h3 g4f3 d1f3 d6f4 f3f4 f6d5 f4f3 e8g8 e1g1 b8d7 f1d1 d8e7 c3e2 f8d8 a1c1 d5f6 e2g3 g7g6 f3f4 d7b6 c4b3 a7a5 a2a3 a5a4 b3a2 a8c8 e3e4 c6c5 d4c5 d8d1 c1d1 c8c5 e4e5 f6d5 f4h6');
INSERT INTO evaluations VALUES(30021,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(30021,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(30021,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30021,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(30021,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(30021,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(30021,36,'cp 22','g1e2 h7h6 g5e3 f8b4 a2a3 b4a5 e2g3 e8g8 f1d3 f8e8 e1g1 c7c6 c3e2 b8d7 f2f3 d7f8 c2c3 a5c7 d1d2 f8g6 a1e1 a7a5 d3b1 d8b8 b1g6 f7g6 e3f4 e6f7 f4c7 b8c7 e2f4 f6d7 f4d3 d7b6 f3f4 e8e1 f1e1 b6c4 d2f2 a8e8');
INSERT INTO evaluations VALUES(31198,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(31198,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(31198,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(31198,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(31198,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(31198,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31198,36,'cp 24','e7e6 e1g1 g8f6 f1e1 b8d7 e4d6 f8d6 d3f5 h7h6 g2g3 e8g8 f5h3 a7a5 d2d3 a5a4 a1b1 b7b5 h3g2 d8b6 c1e3 c6c5 c2c4 a8a7 a2a3 f6g4 e3d2 g4e5 f3e5 d6e5 d2c3 e5f6 c4b5 b6b5 d3d4 f6d4 c3d4 c5d4');
INSERT INTO evaluations VALUES(31572,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(31572,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31572,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(31572,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(31572,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(31572,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31572,36,'cp 30','e1g1 g8f6 f1e1 b8d7 e4d6 f8d6 d3f5 e8g8 f5h3 a7a5 g2g3 a5a4 c2c3 h7h6 h3g2 d8c7 d2d3 f8d8 a1b1 c6c5 d1e2 a8b8 a2a3 b7b5 d3d4 f6d5 d4c5 d6c5 c1d2 c7b6 b1d1 b8c8 g2f1 d5f6 d2e3 c5e3 e2e3 d7c5 d1d4 d8d4 f3d4');
INSERT INTO evaluations VALUES(31571,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(31571,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(31571,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(31571,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(31571,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31571,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(31571,36,'cp 21','g8f6 f1e1 b8d7 e4d6 f8d6 d3f5 a7a5 f5h3 a5a4 c2c3 e8g8 g2g3 h7h6 h3g2 d8b6 d2d4 f8d8 f3d2 d6f8 d1f3 b6a6 g2f1 a6a7 d2e4 f6e4 e1e4 c6c5 d4c5 f8c5 e4e2 a7b6 c1f4 e6e5 f4e3 c5e3 f3e3 b6e3');
INSERT INTO evaluations VALUES(36974,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(36974,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(36974,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(36974,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(36974,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(36974,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(36974,36,'cp 41','c2c4 b7b6 c4d5 e6d5 b1c3 c8b7 d1b3 f8e8 f1d1 b8d7 c1f4 d6f4 g3f4 c7c5 a1c1 h7h6 d4c5 d7c5 b3a3 a7a5 e2e3 a8c8 b2b4 a5b4 a3b4 f6e4 c3e4 c5e4 c1c8 b7c8 b4b2 c8e6 h2h3 d8a8 f3d4 e6d7 b2b6 a8a2');
INSERT INTO evaluations VALUES(36960,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(36960,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36960,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(36960,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(36960,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(36960,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(36960,36,'cp 21','b7b6 b1c3 c8b7 c4d5 e6d5 c3b5 d6e7 c1f4 b8a6 e2e3 c7c6 b5c3 a6c7 d1c2 c7e6 f4e5 f8e8 a1d1 d8d7 e5f6 e7f6 f3e5 f6e5 d4e5 d7e7 f2f4 a8d8 h2h3 a7a5 g1h2 b7a6 f1e1 e6c5 b2b3 h7h6');
INSERT INTO evaluations VALUES(36991,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(36991,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(36991,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(36991,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(36991,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(36991,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(36991,36,'cp -12','c7c6 e1g1 h7h6 a2a4 a7a5 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1d2 b8d7 b2b3 d8b6 f1f2 f8f7 a1f1 a8f8 f3h4 f6e8 c3d1 f7f2 d1f2 e8f6 h4f3 d6d5 f2g4 f6g4 h3g4 d5e4 d3e4 d7c5 d2d6 c5e4 d6e6 g8h8 e6e5 b6e3 g1h2');
INSERT INTO evaluations VALUES(37432,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(37432,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(37432,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(37432,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(37432,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(37432,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37432,36,'cp 13','g1f3 d7d6 h2h3 c7c5 e1g1 d8b6 b1d2 c5d4 e3d4 f6d5 f4g5 h7h6 d2c4 b6c7 g5d2 a7a6 f1e1 b8c6 e2f1 e7e6 a2a4 c8d7 c4e3 d5e3 d2e3 c6e7 c2c4 d7c6 e3d2 a8d8 f1d3 d8d7 d2c3 f8d8 d1d2');
INSERT INTO evaluations VALUES(39930,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(39930,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(39930,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(39930,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(39930,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(39930,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(39930,36,'cp -5','e8g8 d1c2 b8c6 e2e4 d7d5 e4e5 f6e4 f1d3 c5d4 d3e4 d5e4 c2e4 d4c3 f3g5 g7g6 c1e3 d8a5 g5f3 f7f6 e5f6 f8f6 e1g1 e6e5 f1c1 f6d6 c4c5 c8f5 e4c4 f5e6 c4h4 d6d7 h4f6 e6b3 f3e5 c6e5 f6e5 c3c2 e3d4 d7d4 e5d4 a8d8 d4e3 d8d1 c1d1 c2d1q a1d1 b3d1 e3e8 g8g7 e8d7 g7h6 d7d1 a5c5 d1d2 c5g5 f2f4 g5h5 d2d8');
INSERT INTO evaluations VALUES(40682,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(40682,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40682,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(40682,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40682,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(40682,36,'cp 10','e8g8 e1g1 c7c6 c4b3 b7b5 h2h3 f8e8 c3e2 a7a5 a2a4 b8d7 e2g3 e7f8 a4b5 c6b5 c2c4 b5c4 b3c4 c8a6 f3d2 a6c4 d3c4 d8b8 b2b3 d7c5 e3c5 d6c5 g3f5 e8d8 d1e2 a5a4 b3a4 b8b4 f1d1 a8a4 a1a4 b4a4 f5e3 d8d4 e3d5 f6d5');
INSERT INTO evaluations VALUES(36183,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36183,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(36183,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36183,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36183,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(36183,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36183,36,'cp 12','h2h3 c7c6 c4b3 d6d5 e4d5 c6d5 d3d4 e5e4 f3e5 b8c6 f2f4 e4f3 d1f3 c8e6 e1g1 a8c8 e5d3 c6a5 d3c5 d8b6 a1e1 a5b3 a2b3 c8e8 f1f2 b6c6 b3b4 c6b6 b4b5 a7a6 f3g3 g8h7 e3f4 a6b5 g3d3 h7g8 d3b5 b6b5 c3b5 e7c5 d4c5 f6e4 b5d6 e4f2');
INSERT INTO evaluations VALUES(36181,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(36181,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(36181,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(36181,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(36181,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36181,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36181,36,'cp -1','c8e6 e1g1 c7c5 e3d2 b8c6 a2a4 f6e4 c3e4 d6d5 c4b5 c6d4 e4c3 a7a6 f3e5 a6b5 c3b5 d4b5 a4b5 d8b6 a1a8 f8a8 c2c4 d5c4 d3c4 e7f6 d2c3 f6e5 c3e5 f7f6 d1d6 b6d6 e5d6 e6c4 f1e1 b7b6 d6c7 a8a2 c7b6 a2b2 b6c5 b2b5 c5d4 g8f7 e1d1 b5g5 d4e3 g5g6 g1h2');
INSERT INTO evaluations VALUES(41933,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(41933,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(41933,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(41933,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(41933,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41933,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(41933,36,'cp 33','c1d2 d5c3 b2c3 b4e7 f1d3 c7c5 g1f3 b8c6 e1g1 e8g8 d1e2 b7b6 a1d1 c8b7 f1e1 a8c8 h2h3 d8d7 d2c1 f8d8 h3h4 e7d6 d3b1 c6e7 e3e4 c5d4 c3d4 d6b4 c1d2 b4d2 e2d2 h7h6 a2a3 e7g6 h4h5 g6f8');
INSERT INTO evaluations VALUES(41932,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41932,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(41932,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(41932,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(41932,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(41932,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(41932,36,'cp 24','e8g8 a2a3 b4e7 g1f3 b7b6 d1c2 c7c5 f1d3 h7h6 e1g1 c8b7 d4c5 b6c5 c3d5 b7d5 d2c3 b8d7 d3h7 g8h8 h7e4 d8b6 a1d1 d5e4 c2e4 d7f6 e4c2 b6a6 b2b3 f6d5 c3b2 a8b8 d1c1 a6b7 f3d2 d5b6 h2h3 h8g8 f1d1 f8c8');
INSERT INTO evaluations VALUES(42721,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(42721,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(42721,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(42721,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(42721,34,'cp -157',NULL);
INSERT INTO evaluations VALUES(42721,35,'cp -122',NULL);
INSERT INTO evaluations VALUES(42721,36,'cp -114','d7d5 e5f6 d8f6 c4b5 c7c6 g1e2 c6b5 d1b3 b8c6 e1g1 f6h6 c3d5 b4d6 h2h3 a7a6 b3d3 c8e6 d5f6 e8e7 f6e4 h8d8 e4d6 d8d6 d3f3 e6d5 f3g3 a8d8 e2f4 f7f6 b2a3 b5b4 a3c1 g7g5 f4d5 d6d5 f2f4 g5f4 f1f4 d5d1 g1h2 h6g6 g3f2 d8d4 c1b2 d4f4 f2f4 d1a1 b2a1 e7d7 a1f6 d7e6 f6g5 c6e5 f4g3 e6d5 g5e3');
INSERT INTO evaluations VALUES(43754,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(43754,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(43754,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(43754,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(43754,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(43754,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(43754,36,'cp 19','d1d2 b8d7 a2a3 h7h5 f1e2 c8b7 e1c1 f8e7 c1b1 a8c8 h1g1 g7g6 h2h3 h5h4 e3f2 d7c5 d2e1 f6h5 d4b3 c5d7 e1d2 h5g3 e2d3 d8c7 g1e1 e8g8 c3e2 g3h5 g2g4 h4g3 e2g3 h5g3 f2g3 g8g7 f3f4 d7c5 g3h2 c5d3 c2d3 f8h8');
INSERT INTO evaluations VALUES(43755,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43755,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(43755,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(43755,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(43755,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(43755,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(43755,36,'cp 34','b8d7 a2a3 h7h5 e1c1 c8b7 c1b1 f8e7 f1e2 a8c8 h1g1 d8c7 e3g5 c7b6 h2h3 h5h4 d4b3 b6c7 d2e3 h8h5 g5f4 d7e5 g2g4 h4g3 f4g3 h5h3 g3e5 d6e5 g1g7 h3h8 e3a7 f6d7 d1g1 e7f6 g7g8 e8e7 g8c8 h8c8 g1c1');
INSERT INTO evaluations VALUES(44168,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(44168,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(44168,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44168,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(44168,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(44168,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44168,36,'cp 4','d4d5 e6d5 c4d5 h7h6 g5c1 b7b5 e2e4 c5c4 d5d6 d8a5 c1d2 b5b4 f1c4 f8d6 d2e3 d6e5 d1b3 e8g8 g1f3 e5c7 e1g1 c8b7 a2a3 b7e4 b1d2 e4f3 d2f3 b8c6 a3b4 a5b4 b3b4 c6b4 e3c5 f8b8 c5a7 b8c8 c4b5 b4c6 a7c5 a8a1 f1a1 c8b8 b5c6 d7c6 c5d4 f6d5 a1c1 c7b6 c1c2 b6d4');
INSERT INTO evaluations VALUES(46811,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(46811,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(46811,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(46811,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(46811,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46811,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(46811,36,'cp 12','f8g7 c2c4 c7c6 b2b3 e8g8 c1b2 f8e8 d2d4 c8f5 b1c3 b8d7 a1c1 a8c8 f1e1 d7b6 a2a4 e7e6 f3d2 b6d7 e2e4 d5e4 d2e4 f6e4 c3e4 f5e4 e1e4 c6c5 e4e2 d8b6 d4d5 e6d5 e2e8 c8e8 b2g7 g8g7 c4d5');
INSERT INTO evaluations VALUES(48080,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(48080,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(48080,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(48080,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(48080,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(48080,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(48080,36,'cp 28','d2d4 f8e7 e1g1 e8g8 c4e2 c7c5 c2c4 d5f6 b1c3 c5d4 f3d4 e6e5 d4b5 a7a6 d1d8 e7d8 b5d6 b8c6 f1d1 c6d4 e2d3 d8e7 d6c8 a8c8 a1b1 f8d8 c1e3 c8c7 a2a3 c7d7 d3f1 d4b3 d1d7 d8d7 g2g3 f6g4 e3b6 e7d8 b6d8 d7d8 h2h3 g4f6');
INSERT INTO evaluations VALUES(48082,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(48082,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(48082,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(48082,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(48082,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(48082,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(48082,36,'cp 24','c7c5 e1g1 c5d4 f3d4 f8c5 d4b3 c5b6 c4d5 d8d5 d1d5 e6d5 b1c3 c8e6 f1d1 b8c6 c3d5 a8d8 c2c4 c6e5 c1g5 f7f6 d5b6 d8d1 a1d1 a7b6 g5f4 e5c4 b3d4 e8f7 d4e6 f7e6 b2b3 c4e5 a2a4 h8c8 f4e3 e5d7 f2f3 c8c2');
INSERT INTO evaluations VALUES(50177,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50177,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50177,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(50177,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(50177,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(50177,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(50177,36,'cp 12','g2g3 f8g7 b1c3 c7c5 f1g2 b8c6 g1f3 d7d5 e1g1 e8g8 c3d5 f6d5 c4d5 d8d5 c1e3 g7b2 f3d4 d5d7 d4c6 b2a1 d1a1 b7c6 f1c1 a8b8 e3c5 f8e8 h2h4 d7c7 a2a4 c8d7 c5e3 e7e5 c1c4 a7a5 a1a3 c7c8 g1h2 d7e6 c4c6 c8d7');
INSERT INTO evaluations VALUES(51050,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(51050,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(51050,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51050,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51050,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51050,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51050,36,'cp 0','g8f6 g1e2 d7d6 b1c3 f8e7 c4b3 e8g8 d2d3 c7c6 e1g1 b7b5 a2a3 a6a5 c1e3 b8d7 h2h3 f8e8 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(43692,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43692,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43692,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43692,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43692,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43692,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43692,36,'cp 0','b1c3 d7d6 g1e2 f8e7 e1g1 b7b5 c4b3 e8g8 d2d3 c7c6 a2a3 b8d7 c1e3 d7c5 b3a2 f6g4 e3c1 g4f6');
INSERT INTO evaluations VALUES(53392,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(53392,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(53392,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(53392,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(53392,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(53392,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(53392,36,'cp 19','c3d4 g8f6 b1c3 b8c6 g1f3 c8f5 e2e3 e7e6 d1b3 f8b4 f3e5 b4c3 b2c3 e8g8 e5c6 b7c6 b3a3 a7a5 f1e2 d8b6 e1g1 f8c8 a1c1 c6c5 d4c5 b6c5 a3c5 c8c5 c3c4 d5c4 c1c4 c5c4 e2c4 f6d5 f4e5 a5a4 f1c1 a8c8');
INSERT INTO evaluations VALUES(55834,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55834,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(55834,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55834,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55834,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55834,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55834,36,'cp 0','g8f6 f1d3 c7c5 g1f3 b7b6 c1b2 b8c6 d4c5 f8c5 a2a3 e8g8 e1g1 c8b7 b3b4 c5d6 b1d2 d8c7 b2f6 g7f6 f3d4 c6d4 d1g4 g8h8 g4h4 d4f5 d3f5 e6f5 h4f6 h8g8 f6g5 g8h8');
INSERT INTO evaluations VALUES(48309,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48309,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48309,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(48309,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(48309,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(48309,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(48309,36,'cp 14','g1f3 b7b6 f1b5 c7c6 b5d3 c6c5 c1b2 b8c6 a2a3 f8d6 d4c5 b6c5 e1g1 c8b7 c2c4 e8g8 b1d2 a7a5 c4d5 e6d5 f1e1 f8e8 d1c2 h7h6 a1d1 a8c8 h2h3 c8c7 c2c1 c7c8 d3f5 c8b8 f5b1 d5d4 e3d4 c6d4 e1e8 d8e8 f3d4');
INSERT INTO evaluations VALUES(57097,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(57097,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(57097,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(57097,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(57097,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(57097,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(57097,36,'cp 52','g1f3 e5e4 f3d4 c7c5 d4b5 d7d6 b5c3 f7f5 f2f3 e4f3 e2f3 f8e7 a2a4 g8f6 f1e2 b8d7 e1g1 e8g8 f1e1 f6e8 f3f4 b7b6 e2f3 e7h4 g2g3 h4f6 d1d3 a7a6 f3g2 a8b8 c1e3 f6c3 b1c3 d7f6 b2b3 e8c7 e3f2 b6b5 a4b5 a6b5 c3d1 b5b4');
INSERT INTO evaluations VALUES(932,30,'cp 644',NULL);
INSERT INTO evaluations VALUES(932,31,'cp 648',NULL);
INSERT INTO evaluations VALUES(932,32,'cp 658',NULL);
INSERT INTO evaluations VALUES(932,33,'cp 662',NULL);
INSERT INTO evaluations VALUES(932,34,'cp 650',NULL);
INSERT INTO evaluations VALUES(932,35,'cp 649',NULL);
INSERT INTO evaluations VALUES(932,36,'cp 678','c4d5 f8c5 d2d4 c5d4 e1g1 h8f8 c2c3 d4b6 f7g5 c8f5 g2g4 f5g6 g5e6 b6f2 g1g2 f6e6 d5e6 g6e4 g2h3 a8d8 d1d8 c6d8 e6f5 f2c5 b1d2 e4d5 b2b4 c5e7 h3g3 g7g6 f5c2 f8f1 d2f1 d8f7 c1e3 f7d6 e3a7 b7b6 f1e3 d5e4 c2b3 e7g5 a1e1 e8d7 a2a4 g5f4 g3f2 e4c6 e3d5 f4h2 c3c4 c6d5 c4d5');
INSERT INTO evaluations VALUES(5033,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(5033,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(5033,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(5033,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(5033,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(5033,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(5033,36,'cp -39','d7d6 d1e2 g8f6 e4e5 d6e5 e2e5 e8g8 e1g1 a7a5 f1e1 f6d5 e5g3 a5a4 b1d2 c8f5 a2a3 e7e5 d2e4 h7h6 f2f3 d8e7 g3f2 f8d8 h2h3 g8h7 c1e3 d5e3 f2e3 f5e4 f3e4 g7f6 e3e2 h6h5 e1f1 h7g7 g1h1 a8b8 e2f2 h5h4 a1e1 c6c5');
INSERT INTO evaluations VALUES(8253,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(8253,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(8253,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8253,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8253,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8253,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(8253,36,'cp 0','a2a4 e8g8 f1e1 d6d5 e4d5 f6d5 a4a5 e7e6 a5a6 d8c7 c1g5 b7a6 b1d2 a8b8 d2e4 b8b2 d1c1 b2b6 c4d5 e6d5 g5f4 c7e7 f4g5');
INSERT INTO evaluations VALUES(9574,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(9574,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(9574,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(9574,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(9574,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(9574,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(9574,36,'cp 53','e8g8 a2a4 e5d4 f3d4 f8e8 f1e1 d7e5 c4a2 h7h6 h2h3 e5c6 c1f4 e7f8 d4c6 b7c6 e4e5 d6e5 d1d8 e8d8 f4e5 c8f5 e1e2 f8d6 a1d1 d8d7 a2b3 a8d8 d1e1 g8f8 g2g4 f5h7 f2f4 d6e5 f4e5 f6d5 c3d5 c6d5 e5e6 f7e6 e2e6');
INSERT INTO evaluations VALUES(3981,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(3981,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(3981,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(3981,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(3981,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(3981,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(3981,36,'cp 52','a2a4 c7c6 h2h3 h7h6 a4a5 d8c7 f1e1 f8e8 c4f1 d7f8 d4d5 c8d7 f3d2 e8b8 d2c4 c6d5 e4d5 f8g6 d1d3 b7b5 a5b6 a7b6 a1a8 b8a8 c1e3 a8b8 c4a3 c7d8 e1a1 g6h4 a3b5 h4f5 e3d2 b8a8 a1a8 d8a8 d3c4 a8a1 b2b3 f6e8 c3a4 e7d8 c4d3 d7b5 d3b5 e8f6 a4c3 f5d4 b5d3 a1a5 d2e3');
INSERT INTO evaluations VALUES(10654,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(10654,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(10654,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(10654,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(10654,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(10654,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(10654,36,'cp 32','d1e2 e4d6 d5b4 c6b4 e2e5 d8e7 e5e7 e8e7 c4b3 b7b6 e1g1 c8b7 f1e1 e7d8 a2a3 b4c6 c2c3 c6e7 f3d4 f7f6 d2d3 c7c5 d4e2 d6f5 e2g3 h7h5 b3c2 f5g3 f2g3 e7d5 c1d2 a7a5 g1f2 a5a4 h2h4 b7c6 d3d4 c5d4 c3d4 a8c8 c2g6 d5e7');
INSERT INTO evaluations VALUES(11096,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(11096,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(11096,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(11096,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(11096,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(11096,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(11096,36,'cp 88','d4c6 d7c6 d1d8 e8d8 e4e5 f6d5 c3e4 b7b5 a2a4 c8d7 h2h4 b5b4 h4h5 d8c7 c1g5 h7h6 g5h4 c6c5 b2b3 d7c6 e1c1 a6a5 d1d2 d5b6 e2f3 c5c4 h1d1 c6d5 d2d4 c4b3 c2b3 d5b3');
INSERT INTO evaluations VALUES(11966,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11966,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(11966,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(11966,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11966,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11966,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11966,36,'cp 0','c6d5 f1d3 d8c7 c1f4 a8b8 e1g1 b8b2 d1f3 b2b6 f1b1 g7g6 b1b6 c7b6 c2c4 d5c4 a1b1 b6a6 d3c4 a6c4 b1c1 c4a6 f3a8 e8d8 f4d2 f8a3 d2g5 a3e7 g5d2');
INSERT INTO evaluations VALUES(12000,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12000,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12000,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12000,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12000,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12000,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12000,36,'cp 0','f1d3 d8c7 c1f4 a8b8 e1g1 b8b2 d1f3 b2b6 a1b1 h7h6 f3g3 g7g6 f4e3 b6b1 f1b1 f8g7 f2f4 e8g8 g3f2 c7a5 e3c5 c8a6 d3a6 a5a6 c5f8 a6a2 b1b8 a2a1 f2f1 a1d4 f1f2 d4d1 f2f1');
INSERT INTO evaluations VALUES(18248,30,'cp 149',NULL);
INSERT INTO evaluations VALUES(18248,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(18248,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(18248,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(18248,34,'cp 118',NULL);
INSERT INTO evaluations VALUES(18248,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(18248,36,'cp 175','a2a4 c8g4 d4d5 c6e5 d1d4 e5f3 g2f3 g4f3 h1g1 g8f6 b1d2 f3h5 a4b5 e7e6 d4c4 d8d6 c4c6 d6c6 d5c6 f8e7 f2f3 e8g8 f1e2 h5g6 d2b3 f6h5 c1e3 e7d6 e2c4 d6h2 g1h1 h2g3 e1d2 g3e5 b3c5 e5b2 a1a2 b2e5 c5d7 f8d8 d2c2 e5f4 e3a7 f4g3 a7d4 g3d6 a2a8 d8a8');
INSERT INTO evaluations VALUES(18485,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(18485,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(18485,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(18485,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(18485,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(18485,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(18485,36,'cp -21','c4d3 g8f6 d1e2 f8e7 e1g1 e8g8 b1c3 f6d7 c1f4 d7c5 a1d1 c5d3 e2d3 f7f6 f1e1 a7a5 c3a4 e6e5 e4d5 c6d5 d3d5 d8d5 d1d5 c8e6 d5d2 e6a2 f4e3 a2e6 a4b6 a8a6 b6d5 e6d5 d2d5 g8f7 d5b5 f8c8 e1c1 a5a4 c2c3 a4a3 b2a3 e7a3 c1a1 c8c3 b5b7 f7g6 g2g3');
INSERT INTO evaluations VALUES(21152,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(21152,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(21152,32,'cp 206',NULL);
INSERT INTO evaluations VALUES(21152,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(21152,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(21152,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(21152,36,'cp 199','d2d4 e7e6 e2e4 f8b4 a2a3 b4a5 b2b4 a5b6 f1b5 c8d7 e1g1 g8f6 b5c4 e8g8 d1d3 h7h6 h2h3 c6e7 f1d1 c7c6 a3a4 b6c7 f3e5 d8b8 c1e3 c7d6 e5d7 f6d7 b4b5 d6f4 b5c6 e7c6 d4d5 e6d5');
INSERT INTO evaluations VALUES(21127,30,'cp 224',NULL);
INSERT INTO evaluations VALUES(21127,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(21127,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(21127,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(21127,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(21127,35,'cp 220',NULL);
INSERT INTO evaluations VALUES(21127,36,'cp 193','e7e6 e2e4 f8b4 a2a3 b4a5 b2b4 a5b6 f1b5 c8d7 e1g1 g8f6 b5a4 e8g8 a4c2 d7e8 d4d5 e6d5 c3d5 f6d7 a3a4 a7a5 b4a5 b6c5 e4e5 f7f6 d1d3 e8g6 d3c4 g6c2 d5e7 g8h8 e7c6 c2d3 c4d3 b7c6 d3c2 f8e8 e5f6 d8f6 c1b2 f6g6 c2g6');
INSERT INTO evaluations VALUES(21506,30,'cp 197',NULL);
INSERT INTO evaluations VALUES(21506,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(21506,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(21506,33,'cp 204',NULL);
INSERT INTO evaluations VALUES(21506,34,'cp 233',NULL);
INSERT INTO evaluations VALUES(21506,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(21506,36,'cp 193','e2e4 f8b4 a2a3 b4a5 b2b4 a5b6 f1b5 c8d7 e1g1 g8f6 b5a4 e8g8 a4c2 d7e8 d4d5 c6e7 c1g5 e6d5 g5f6 g7f6 c3d5 e7g6 d1c1 c7c6 d5b6 a7b6 c1h6 d8e7 f3d4 e8d7 f2f4 f6f5 d4f5 d7f5 e4f5 e7h4 h6g5 f7f6 g5h4 g6h4 a3a4 f8d8 a1d1');
INSERT INTO evaluations VALUES(22837,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(22837,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(22837,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(22837,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(22837,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(22837,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(22837,36,'cp -26','f4e5 c6e5 g1f3 e5f3 d1f3 d7d5 b2b3 g8f6 c1b2 f8e7 f1d3 c8g4 f3g3 e8g8 b1c3 g4h5 g3h3 h5g6 d3g6 f7g6 e1c1 f8f7 h1f1 c7c6 f1f3 e7d6 d1f1 d8e7 g2g4 a8f8 g4g5 f6d7 f3f7 f8f7 f1f7 e7f7 c1b1 f7f5');
INSERT INTO evaluations VALUES(22902,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(22902,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(22902,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(22902,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(22902,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(22902,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(22902,36,'cp -49','f8e7 e3b3 g8f6 b1c3 a7a5 g1f3 a5a4 b3d1 e7b4 c1d2 a4a3 b2a3 b4a3 e2e3 e8g8 d1b3 f8e8 f1e2 b7b6 e1g1 c8b7 f1d1 c6e5 f3e5 e8e5 d2e1 b7c6 c3d5 f6e4 b3b1 a3d6 a2a4 c6a4 d1d4 f7f5 f2f3 e4c5 e1b4 d8e8');
INSERT INTO evaluations VALUES(6992,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(6992,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(6992,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(6992,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(6992,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(6992,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(6992,36,'cp -36','e3b3 g8f6 b1c3 a7a5 g1f3 a5a4 b3d1 e7b4 c1d2 a4a3 b2a3 b4a3 d1b3 e8g8 e2e3 f8e8 f1e2 b7b6 e1g1 c8b7 f1d1 c6e5 c3d5 f6d5 c4d5 a3d6 a2a4 b7a6 f3d4 a6e2 d4e2 d8h4 h2h3 h4e4 e2d4 e5d3 b3c2 d6e5 a1a3 e5d4 c2d3 e4d3 a3d3 d4c5');
INSERT INTO evaluations VALUES(25772,30,'cp 238',NULL);
INSERT INTO evaluations VALUES(25772,31,'cp 232',NULL);
INSERT INTO evaluations VALUES(25772,32,'cp 257',NULL);
INSERT INTO evaluations VALUES(25772,33,'cp 250',NULL);
INSERT INTO evaluations VALUES(25772,34,'cp 265',NULL);
INSERT INTO evaluations VALUES(25772,35,'cp 269',NULL);
INSERT INTO evaluations VALUES(25772,36,'cp 266','h2h3 g4e6 f3d4 g8f6 e1g1 d8b6 f1e1 f8e7 c2c4 e8g8 b1c3 e6d7 a1b1 f8e8 c1e3 b6b7 e3f4 b7b6 d1d3 a8d8 b2b3 h7h6 e1d1 e7f8 d3f3 b6b7 d1e1 b7b6 d4c2 b6a5 e1d1 d7c8 c2d4 a5b6 b3b4 c8d7 f4g3 c6c5 d4c2 c5b4 g3h4 f8e7 h4f6 e7f6');
INSERT INTO evaluations VALUES(27634,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27634,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27634,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(27634,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27634,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27634,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27634,36,'cp 0','e8c8 c1e3 g8f6 a2a3 a5f5 e1g1 e7e5 d4d5 e5e4 f3d4 c6d4 e2g4 f6g4 d1d4 f8d6 h2h3 g4h2 d4a7 h2f3 g1h1 f5e5 g2g3 e5h5 h1g2 f3h4 g2h2 h4f3');
INSERT INTO evaluations VALUES(37,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37,36,'cp 0','c1e3 g8f6 a2a3 a5f5 e1g1 e7e5 d4d5 e5e4 f3d4 c6d4 e2g4 f6g4 d1d4 f8d6 h2h3 g4h2 d4a7 h2f3 g1h1 f5e5 g2g3 e5h5 h1g2 f3h4 g2h2 h4f3');
INSERT INTO evaluations VALUES(36,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36,36,'cp 0','g8f6 a2a3 a5f5 e1g1 e7e5 d4d5 e5e4 f3d4 c6d4 e2g4 f6g4 d1d4 f8d6 h2h3 g4h2 d4a7 h2f3 g1h1 f5e5 g2g3 e5h5 h1g2 f3h4 g2h2 h4f3');
INSERT INTO evaluations VALUES(16,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16,36,'cp 0','a2a3 a5f5 e1g1 e7e5 d4d5 e5e4 f3d4 c6d4 e2g4 f6g4 d1d4 f8d6 h2h3 g4h2 f1e1 h2f3 g2f3 f5h3 f3f4 f7f5 d4a7 g7g5 a7a8 c8d7 a8a4 d7c8');
INSERT INTO evaluations VALUES(29105,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(29105,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(29105,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(29105,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(29105,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(29105,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(29105,36,'cp 137','e4e5 d6d5 c4b3 f6g8 h2h3 g4f5 c2c4 d5c4 d3c4 d8d1 b3d1 f5d7 b1c3 e7e6 d1c2 g8e7 c1f4 d7c6 e1c1 e7g6 f4g3 c6f3 g2f3 b8c6 f3f4 a8d8 h3h4 a7a6 d1d8 e8d8 h4h5 g6e7 c3a4 e7f5 c2f5 e6f5 g3h4 d8c7 a4c3 f8e7 h4e7 c6e7 c3d5 c7d7 h1d1 e7d5 c4d5');
INSERT INTO evaluations VALUES(31338,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(31338,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(31338,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(31338,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(31338,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(31338,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(31338,36,'cp 3','g8f6 c4d5 f6d5 h2h3 g4f3 d1f3 e7e6 f1b5 b8c6 e1g1 a7a6 b5c6 b7c6 b1d2 f8e7 f1d1 e8g8 d2c4 e7f6 b2b3 c6c5 c1b2 c5d4 b2d4 f6d4 d1d4 d8f6 f3f6 g7f6 a1d1 a8c8 d1d3 c8c7 g1h2 h7h6 c4d6 c7c1 d3d2 c1c3 d2d1 f8d8 d6c4');
INSERT INTO evaluations VALUES(31599,30,'cp 135',NULL);
INSERT INTO evaluations VALUES(31599,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(31599,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(31599,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(31599,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(31599,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(31599,36,'cp 136','d1f3 d8e7 a2a4 b8d7 e1g1 g8f6 b2b4 a7a6 c1b2 e7d8 d2d3 g7g6 b4b5 f8g7 b5c6 b7c6 b2a3 d8c7 c3d1 e8g8 d1e3 a6a5 a1b1 f8d8 g2g4 d7f8 h3h4 h7h5 g4g5 f6e8 b1b3 d8b8 f1b1 b8b3 b1b3 c7d7 g1g2 e8c7 b3b7 f8e6 c4e6 d7e6');
INSERT INTO evaluations VALUES(32029,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(32029,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(32029,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(32029,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(32029,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(32029,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(32029,36,'cp 96','b8c6 e5d6 f8d6 c4b5 g8f6 b5c6 b7c6 h2h3 g4f3 d1f3 e8g8 e1g1 f8e8 b1c3 e8e6 c1f4 d8e7 e4e5 d6e5 f4e5 e6e5 f3c6 a8b8 a1b1 b8b6 c6f3 f6e4 c3e4 e5e4 f3c3 e4e2 b1e1 b6e6 e1e2 e6e2 a2a4 a7a6 c3c6 e7b4 c6c7 b4a4 c7d8 a4e8 d8e8 e2e8');
INSERT INTO evaluations VALUES(32033,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(32033,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(32033,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(32033,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(32033,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(32033,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(32033,36,'cp 122','b1c3 h5g6 d1e2 c7c6 c1e3 g8f6 g2g4 f8e7 e1c1 b8d7 f3h4 b7b5 c4b3 d7c5 h4f5 c5b3 a2b3 e8g8 g4g5 h6g5 d1g1 a7a5 f5e7 d8e7 h3h4 g5g4 h4h5 g6h7 e3g5 e7e6 c1b1 a5a4 b3b4 g8h8 f2f3 a4a3 b2b3 g4f3 e2f3');
INSERT INTO evaluations VALUES(32207,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(32207,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(32207,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(32207,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(32207,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(32207,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32207,36,'cp 0','c7c6 b5e2 g8f6 b1c3 c6c5 f3g5 b8c6 g5e6 f7e6 d2d3 d8d7 a2a4 f8e7 e1g1 e8g8 g2g3 c6d4 c1g5 f6e8 g5d2 e8f6');
INSERT INTO evaluations VALUES(31610,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31610,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31610,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31610,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31610,36,'cp 0','b5e2 g8f6 b1c3 c6c5 f3g5 b8c6 g5e6 f7e6 d2d3 d8d7 e1g1 f8e7 a2a4 e8g8 g2g3 c6d4 c1g5 f6e8 g5d2 e8f6');
INSERT INTO evaluations VALUES(32601,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(32601,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(32601,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(32601,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(32601,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(32601,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(32601,36,'cp 39','h2h3 g4f5 c1f4 f8e7 f1e1 b8d7 c3b1 e8g8 b1d2 a5a4 b2b3 a4a5 d2c4 a5d8 c4d6 f6d5 d6f5 d5f4 f5e7 d8e7 e2f1 f8d8 g2g3 f4g6 d1d2 c6c5 d4c5 e7c5 a1d1 c5b6 c2c4 d7f6 d2b2 d8d1 e1d1 a8d8 d1d8 b6d8 b3b4 h7h6');
INSERT INTO evaluations VALUES(32702,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(32702,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(32702,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(32702,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(32702,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(32702,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(32702,36,'cp 37','h2h3 g4h5 d2d4 b8d7 c1d2 c7c6 c3e4 a5b6 e4f6 d7f6 c2c3 f8d6 f3e5 h5e2 d1e2 b6b5 e2b5 c6b5 a2a4 b5a4 a1a4 f6d5 f1a1 a7a6 e5d3 e8g8 a4a5 f7f6 g1f1 a8c8 g2g4 f8e8 h3h4 d6c7 a5a2 a6a5 h4h5 b7b5 a2a5 c7a5 a1a5 c8b8 g4g5 g8f7 f1e2');
INSERT INTO evaluations VALUES(34684,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34684,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(34684,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(34684,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(34684,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(34684,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(34684,36,'cp 27','e7e5 f1g2 g8f6 g1f3 b8c6 d2d3 c6d4 e1g1 c8g4 h2h3 g4f3 g2f3 c7c6 f3g2 a8d8 c3b1 a5c7 b1d2 d4e6 f1e1 f8c5 d2b3 c5b6 a2a4 a7a5 d1e2 f6d7 c1e3 b6e3 e2e3 e8g8 a1d1 h7h6 d3d4 e5d4 b3d4 d7c5 d4e6 c5e6 e3c3 d8d1 e1d1 f8b8 c3a3');
INSERT INTO evaluations VALUES(36521,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(36521,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(36521,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(36521,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(36521,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(36521,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(36521,36,'cp 5','a1b1 a7a6 c1d2 c7c6 g1f3 b7b5 b2b4 d6d5 c4d5 c6d5 e1g1 c8b7 e2e3 b8d7 f3d4 e7e5 d4b3 a8a7 e3e4 d5d4 c3e2 b7a8 b1c1 a7c7 c1c7 d8c7 d1b1 f8c8 f1c1 c7b8 g2h3 c8c1 e2c1 b8a7 h3g2 d7b6 b3c5 f6d7 g1h1 d7c5');
INSERT INTO evaluations VALUES(43127,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(43127,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(43127,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(43127,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(43127,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(43127,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(43127,36,'cp -32','f6e4 g5h4 c7c5 f1d3 d8a5 e1f1 e4f6 b1c3 c5d4 e3d4 e8g8 g1f3 d7d6 d1d2 c8g4 c3d5 a5d8 d5f6 e7f6 h2h3 g4f3 g2f3 d8b6 d4d5 f8e8 f1g2 a7a5 a1b1 b8d7 b2b3 d7c5 h1d1 h7h6 h4g3 f6f5 g3f4 g8h7 h3h4 h6h5 f4g5 b6b4 d3f1 g7e5 g2g1 b7b6 g5e3 b4d2 d1d2');
INSERT INTO evaluations VALUES(45731,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(45731,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(45731,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(45731,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(45731,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(45731,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(45731,36,'cp 5','f8b4 g1f3 c6d5 e4d5 e8g8 e1g1 e5e4 f3d4 b8d7 d4e6 f7e6 d5e6 g8h8 e6d7 c8d7 c1g5 d8c7 g5f6 g7f6 c4d5 a8e8 g1h1 d7f5 d1d4 b4c5 d4c4 c5b6 c4c7 b6c7 a1e1 c7e5 d5b7 f5g6 c3e4 e5b2 e4g3 e8b8 b7e4 g6f7 e1d1 b2e5 e4f5 f7a2');
INSERT INTO evaluations VALUES(40499,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40499,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40499,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40499,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40499,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(40499,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40499,36,'cp 0','g1f3 c6d5 e4d5 e8g8 e1g1 e5e4 f3d4 b8d7 d4e6 f7e6 d5e6 g8h8 e6d7 c8d7 c1g5 b4c3 b2c3 d8c7 g5f6 f8f6 d1d4 f6d6 c4d5 d7c6 c3c4 c7a5 f1d1 a8d8 d4e4 h7h6 g2g3 c6d5 c4d5 d6d5 d1e1 d5d2 e4g6 a5c5 e1e8 d8e8 g6e8 h8h7 e8e4 h7g8 e4e8');
INSERT INTO evaluations VALUES(47073,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(47073,31,'cp -167',NULL);
INSERT INTO evaluations VALUES(47073,32,'cp -171',NULL);
INSERT INTO evaluations VALUES(47073,33,'cp -169',NULL);
INSERT INTO evaluations VALUES(47073,34,'cp -163',NULL);
INSERT INTO evaluations VALUES(47073,35,'cp -160',NULL);
INSERT INTO evaluations VALUES(47073,36,'cp -171','e4f6 e1g1 c7c6 d3d4 c8g4 h2h3 g4h5 c4e2 e7e6 f1e1 d6d5 f3e5 h5e2 d1e2 f8e7 c1f4 e8g8 b1d2 b8d7 a1d1 f8e8 c2c3 d7f8 a2a4 f6d7 d2f3 d7e5 f3e5 e7d6 e2g4 d8c7 g4g3 f7f6 e5d3 d6f4 g3f4 c7f4 d3f4 e6e5');
INSERT INTO evaluations VALUES(46981,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(46981,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(46981,32,'cp -164',NULL);
INSERT INTO evaluations VALUES(46981,33,'cp -163',NULL);
INSERT INTO evaluations VALUES(46981,34,'cp -165',NULL);
INSERT INTO evaluations VALUES(46981,35,'cp -168',NULL);
INSERT INTO evaluations VALUES(46981,36,'cp -164','d3d4 d6d5 c4e2 c8g4 e1g1 c7c6 c1f4 g4f3 e2f3 e7e6 b1d2 f8d6 f4e3 b8d7 c2c4 e8g8 g2g3 h7h6 d1b3 a8b8 a1d1 d6c7 b3c2 d7b6 b2b3 b6c8 f3e2 c8d6 e2d3 d8d7 d2f3 d5c4 b3c4 d6f5 e3f4 c7f4');
INSERT INTO evaluations VALUES(47422,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(47422,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(47422,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(47422,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(47422,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(47422,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(47422,36,'cp 27','g1f3 e7e6 d3d4 f8e7 c1f4 e8g8 f1e2 c7c5 d4c5 e7c5 e1g1 b8c6 a2a3 b7b6 d1e1 c8b7 b2b4 c5e7 c3b5 d8c8 a1d1 f8d8 d1d8 c8d8 e1c1 d8f8 f1d1 a8d8 f4c7 d8d1 c1d1 f6e8 c7g3 e7f6 d1d7 f8e7 d7e7 f6e7 f3e5 a7a6 e5c6 b7c6 b5c7 e8c7 g3c7 b6b5 c7b6 e7f6 c2c4 b5c4');
INSERT INTO evaluations VALUES(48122,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(48122,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(48122,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(48122,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(48122,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(48122,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(48122,36,'cp 43','c3e2 e4g5 f3g5 d8g5 d2d4 g5d8 a2a3 c7c5 c2c3 f8e7 e2f4 c8d7 f1d3 c5d4 c3d4 d8a5 e1f1 b8c6 c1e3 e7g5 b2b4 a5b6 f4h5 g5e3 f2e3 e8c8 f1e1 c8b8 h1f1 f7f6 e5f6 g7g6 h5f4 h8f8 d1f3 f8f6 f3g3 b8a8 f4e2 c6b4 f1f6 b4d3 e1f1');
INSERT INTO evaluations VALUES(48878,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(48878,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(48878,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(48878,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(48878,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(48878,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(48878,36,'cp -78','g1f3 f8d6 c2c4 c7c6 b2b3 e8g8 c1b2 d8e7 b1d2 a7a5 e1g1 c8f5 f1e1 f8d8 a2a3 b8d7 d2f1 f6e8 f1d2 f5g6 c4d5 c6d5 h2h3 e5e4 d3e4 d5e4 f3d4 d7e5 b3b4 a5b4 a3b4');
INSERT INTO evaluations VALUES(49484,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(49484,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(49484,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(49484,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(49484,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(49484,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(49484,36,'cp -5','a2a3 f8c5 c2c4 d7d6 e2e3 c7c6 g1e2 a7a5 b1c3 b8a6 e2g3 h7h5 h2h4 c5b6 c1d2 a6c5 b2b4 a5b4 a3b4 a8a1 d1a1 c5b3 a1b2 b3d2 b2d2 c6c5 b4c5 b6c5 c3a4 c5a7 f1e2 e8g8 e1g1 f6g4 e2g4 h5g4 g3e4 d8h4 e4d6 g4g3 f2g3 h4g3 c4c5');
INSERT INTO evaluations VALUES(53111,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(53111,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(53111,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(53111,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(53111,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(53111,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(53111,36,'cp 71','e7e6 e2e4 g8f6 f1c4 f8e7 g1e2 e6d5 c3d5 f6d5 c4d5 e8g8 e1g1 b8a6 c2c3 a6c7 e2f4 f8e8 a2a4 c7d5 f4d5 c8e6 f2f3 e6d5 d1d5 d8b6 f1d1 c5c4 g1f1 b6b3 c1e3 b3b2 e3d4 b7b6 d5c4 b2c2 c4d3 c2d3 d1d3 e8d8 a4a5 b6a5 a1a5');
INSERT INTO evaluations VALUES(54977,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(54977,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(54977,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(54977,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54977,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(54977,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54977,36,'cp 0','c7c5 f1b5 c8d7 b5d7 d8d7 c2c4 e7e6 b1c3 g8f6 g1e2 b8c6 d2d4 c5d4 e2d4 a8c8 d4e2 a7a6 e1g1 f8e7 a2a4 e8g8 c1e3 c6b4 d1b3 b4d3 b3c2 d3b4');
INSERT INTO evaluations VALUES(55208,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(55208,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(55208,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(55208,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(55208,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(55208,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(55208,36,'cp -29','b8c6 d2d3 e7e5 f1g2 f8c5 e1g1 a7a5 e2e3 g8f6 e3d4 c5d4 b1c3 e8g8 f1e1 f8e8 c3b5 d4b6 c1e3 b6e3 e1e3 c8g4 h2h3 g4f3 d1f3 d8d7 a1d1 a8d8 f3e2 h7h5 d3d4 e5d4 e3e8 d8e8 e2d2 h5h4 b5d4 c6d4 d2d4 d7d4 d1d4 h4g3');
INSERT INTO evaluations VALUES(57312,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(57312,31,'cp -161',NULL);
INSERT INTO evaluations VALUES(57312,32,'cp -148',NULL);
INSERT INTO evaluations VALUES(57312,33,'cp -148',NULL);
INSERT INTO evaluations VALUES(57312,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(57312,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(57312,36,'cp -144','b8c6 g1f3 d7d5 f3d4 d5e4 f1b5 c8d7 b1c3 f8b4 c1e3 d8h4 e3f2 h4e7 d1d2 g8f6 a2a3 b4c3 d2c3 f6d5 c3c4 d5b6 c4e2 c6d4 f2d4 e8c8 e1c1 f7f6 b5d7 b6d7 h1e1 h8e8 e2g4 e7f7 g4h3 g7g6 g2g4');
INSERT INTO evaluations VALUES(22945,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(22945,31,'cp -145',NULL);
INSERT INTO evaluations VALUES(22945,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(22945,33,'cp -147',NULL);
INSERT INTO evaluations VALUES(22945,34,'cp -151',NULL);
INSERT INTO evaluations VALUES(22945,35,'cp -149',NULL);
INSERT INTO evaluations VALUES(22945,36,'cp -163','g1f3 d7d5 f3d4 d5e4 f1b5 c8d7 c1e3 c6b4 b1c3 d7b5 d4b5 c7c6 b5d4 g8f6 d1e2 b4d5 c3d5 d8d5 a2a3 e8c8 e1c1 f8d6 c2c4 d5h5 e2h5 f6h5 g2g3 g7g6 h1e1 a7a6 b2b4 h8e8 d4b3 f7f5 e3b6 d6c7 b6c7 d8d1 c1d1 c8c7');
INSERT INTO evaluations VALUES(2725,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2725,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(2725,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(2725,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(2725,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2725,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2725,36,'cp 0','c8b7 f1e1 c6a5 b3c2 c7c5 b1d2 f8e8 d2f1 h7h6 f1g3 e7f8 d3d4 e5d4 c3d4 c5d4 f3d4 d8c7 c1f4 a5c4 b2b3 c4e5 a1c1 e5g6 f4d2 c7d8 d4f3 f6d7 f3d4');
INSERT INTO evaluations VALUES(5336,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(5336,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(5336,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(5336,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(5336,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(5336,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(5336,36,'cp -34','c1b2 a7a6 g2g3 g8f6 d2d3 d7d6 f1g2 e8g8 e1g1 b7b5 d1d2 e7e5 c3d5 a6a5 a2a4 b5b4 d5e3 g7h6 d2d1 f6g4 e3c4 c8e6 h2h3 g4f6 f3h2 h6g7 h2g4 f6g4 h3g4 h7h6 c4e3 a8b8 a1b1 g8h7 f2f4 e5f4 g3f4 c6d4 d1d2 d8h4 d2f2');
INSERT INTO evaluations VALUES(8524,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(8524,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8524,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(8524,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8524,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8524,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8524,36,'cp 0','c2c4 e8g8 c4d5 f6d5 b1c3 c8e6 c1d2 d8d7 a2a3 a8d8 a1c1 h7h6 c3d5 e6d5 b2b4 a7a6 d2c3 d7e6 d1c2 e7f6 c2b2 d8d7 f1e1 f8e8 c1b1 e8d8 b1c1');
INSERT INTO evaluations VALUES(9603,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(9603,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(9603,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(9603,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(9603,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9603,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(9603,36,'cp 0','a2a4 c7c6 d1e2 e8g8 e1g1 d8c7 f3e1 d7c5 e1d3 c8g4 f2f3 g4e6 d3c5 e7c5 g1h1 e6c4 e2c4 b7b6 c3d1 a7a5 d1f2 h7h6 f2d3 f6d7 d3c5 d7c5 c1e3 c5e6 h2h3 a8b8 f1d1 f8d8 b2b3 g8h7 h1g1 b8c8 c4e2 e6d4 e2c4');
INSERT INTO evaluations VALUES(10333,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(10333,31,'cp 192',NULL);
INSERT INTO evaluations VALUES(10333,32,'cp 210',NULL);
INSERT INTO evaluations VALUES(10333,33,'cp 226',NULL);
INSERT INTO evaluations VALUES(10333,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(10333,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(10333,36,'cp 212','b2c3 f6e4 d1d5 f7f5 f1c4 d8e7 e1g1 e7e6 d5b5 e6g6 a2a4 a7a6 b5b2 b7b6 a4a5 b6b5 c4d5 a8b8 e5e6 c6e7 e6d7 c8d7 f3e5 g6f6 d5f7 e8d8 c3c4 d7e6 f7e6 f6e6 f2f3 e4f6 c1g5 e7c6 e5c6 e6c6 b2e5 d8c8 e5f5 c6d7 f5c5 d7d6 c5d6 c7d6 g5f4 b5c4 f4d6 b8b5 a1b1 f6d5 b1b5 a6b5 a5a6');
INSERT INTO evaluations VALUES(10332,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(10332,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(10332,32,'cp 237',NULL);
INSERT INTO evaluations VALUES(10332,33,'cp 222',NULL);
INSERT INTO evaluations VALUES(10332,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(10332,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(10332,36,'cp 194','f6e4 d1d5 f7f5 f1c4 d8e7 e1g1 e7e6 d5b5 a7a6 b5b3 e6g6 c4d5 e4c5 b3a3 c5e6 f3d4 e6d4 c3d4 c6d4 a3e3 c7c6 e3d4 c6d5 c1a3 g6e6 a3d6 e8f7 a2a4 b7b6 d4b6 a6a5 f1b1 h7h6 h2h3 c8a6 b6a5 h8c8 a5d2 c8c6 a4a5 a8c8 d2d4 c6c2 d4a7 a6d3 b1b7 f7g8 b7d7');
INSERT INTO evaluations VALUES(10675,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(10675,31,'cp 238',NULL);
INSERT INTO evaluations VALUES(10675,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(10675,33,'cp 240',NULL);
INSERT INTO evaluations VALUES(10675,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(10675,35,'cp 219',NULL);
INSERT INTO evaluations VALUES(10675,36,'cp 214','d1d5 f7f5 f1c4 d8e7 e1g1 e7e6 d5b5 a7a6 b5b3 e6g6 c4d5 e4c5 b3a3 c5e6 f3d4 c6d4 c3d4 e6d4 a3e3 c7c6 e3d4 c6d5 c1a3 g6e6 a3d6 b7b5 a2a4 b5b4 d4b4 a6a5 b4b6 e8f7 f1b1 c8a6 b6a5 h8c8 a5d2 d5d4 a4a5 e6d5 f2f3 h7h6 h2h3 c8c4 b1b6 f7g8 g1h2 c4c8 a1d1 g8h7 d2f2 c8c4 f2g3 d5f7');
INSERT INTO evaluations VALUES(13569,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(13569,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(13569,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(13569,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(13569,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(13569,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(13569,36,'cp -12','e2e3 f8d6 f4d6 d8d6 f1e2 e8g8 e1g1 b7b6 b1d2 e6e5 d4e5 c6e5 c3c4 c8e6 d1c2 e5f3 d2f3 h7h6 f1d1 f8d8 a2a4 d6c7 a4a5 d5c4 e2c4 e6c4 c2c4 d8d1 a1d1 b6a5 d1c1 c7b7 c1a1 b7b2 a1a5 b2b1 g1h2 b1b4 a5c5');
INSERT INTO evaluations VALUES(14505,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(14505,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(14505,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(14505,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(14505,34,'cp -137',NULL);
INSERT INTO evaluations VALUES(14505,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(14505,36,'cp -124','c1f4 c7c6 h2h3 g7g6 f4e5 d4e6 d1d8 e6d8 g1f3 f8g7 a2a4 e8g8 e5g3 g7h6 a4a5 f6h5 g3h4 f8e8 e1f2 f7f6 g2g4 h5f4 e4e5 f6e5 f3e5 c8e6 e5f3 h6g7 a1e1 b7b5 a5b6 a7b6 f2g3 f4d5 f3g5 e6d7 f1c4 e7e6 g5f3');
INSERT INTO evaluations VALUES(15440,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(15440,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(15440,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15440,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(15440,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(15440,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(15440,36,'cp 0','a7a6 b5e2 g7g5 h2h3 e5e4 f3h2 f8g7 d2d3 e4d3 e2d3 h7h5 d1d2 c8e6 b1c3 d8e7 h2f3 g5g4 h3g4 h5g4 h1h8 g7h8 f3d4 c6d4 e3d4 e6f5 e1f1 f5d3 d2d3 e8c8 a1e1 e7d7 c3e2 f6h5 d3h7 h5f6');
INSERT INTO evaluations VALUES(16582,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(16582,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(16582,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(16582,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(16582,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(16582,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(16582,36,'cp -64','d7d5 d2d3 c8e6 g1f3 d5d4 c3e2 a7a5 e2g3 f8e7 f1e2 f6d7 h3h4 h7h6 h4h5 a5a4 f3h2 e7g5 e2g4 g5c1 d1c1 e8g8 c1d2 a8a5 g4f5 a5a6 f5e6 f7e6 h2f3 a6b6 a1b1 g8h7 d2e2 f8f4 b2b4 a4b3 b1b3 d7c5 b3b6 c7b6');
INSERT INTO evaluations VALUES(17857,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(17857,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(17857,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17857,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(17857,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(17857,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17857,36,'cp 0','f8c5 d2d3 g8f6 c1e3 d7d6 e1g1 e8g8 e3c5 d6c5 c3d5 d8d6 d5e3 b7b5 c4a2 f8d8 a3a4 a8b8 h2h3 c8e6 a4b5 a6b5 a2e6 d6e6 d1e2 f6h5 e3d5 h5f6');
INSERT INTO evaluations VALUES(4828,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(4828,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4828,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(4828,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(4828,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(4828,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(4828,36,'cp 7','e1g1 g8f6 d2d3 d7d6 c1e3 e8g8 c3d5 f6d5 c4d5 c6e7 d5b3 c5e3 f2e3 e7g6 d1e1 d8e8 e1g3 c8e6 b3e6 e8e6 f1f2 f7f5 e4f5 f8f5 a1f1 a8f8 g3g4 f5f6 g4e6 f6e6 f3d2 f8f2 f1f2 g6f8 e3e4 f8d7 b2b4 e6f6 d2c4 g8f7 c4e3 f6f2 g1f2 d7f8 h2h3 f8e6');
INSERT INTO evaluations VALUES(4827,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(4827,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(4827,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(4827,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(4827,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(4827,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4827,36,'cp 5','g8f6 d2d3 d7d6 c1e3 e8g8 c3d5 f6d5 c4d5 c6e7 d5b3 c5e3 f2e3 e7g6 d1e1 d8e8 e1g3 c8e6 b3e6 e8e6 f1f2 f7f5 e4f5 f8f5 g3g4 a8f8 a1f1 f5f6 g4e6 f6e6 f3d2 f8f2 f1f2 g6f8 e3e4 f8d7 d2c4 e6f6 b2b4 g7g6 c4e3 g8g7 f2f1 f6f1 g1f1');
INSERT INTO evaluations VALUES(19156,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(19156,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(19156,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(19156,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(19156,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(19156,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(19156,36,'cp 187','d2d3 a7a6 b5a4 b7b5 a4b3 g8f6 f2f4 d7d6 a2a4 c8e6 b3e6 f7e6 f4e5 d6e5 d1f3 f8b4 f3h3 e8g8 a4b5 d8d6 a1a6 a8a6 b5a6 d6a6 c1d2 b4e7 b1a3 e7a3 b2a3 f8f7 d2b4 h7h6 h3f3 a6c6 f1f2 c6b6 f3d1 b6c6 h2h3 g8h7 d1e2 h7h8 g1h2 h8g8 f2f1 f7c7 f1c1 c6b6 c2c4 d4c3 c1c3 c7c3');
INSERT INTO evaluations VALUES(19155,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(19155,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(19155,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(19155,33,'cp 162',NULL);
INSERT INTO evaluations VALUES(19155,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(19155,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(19155,36,'cp 178','a7a6 b5a4 b7b5 a4b3 g8f6 f2f4 d7d6 a2a4 b5b4 b1d2 f8e7 f4e5 d6e5 d2c4 d8c7 d1e1 c8e6 e1g3 a8c8 g3g7 h8g8 g7h6 g8g6 h6h4 g6g4 h4e1 e8f8 c1d2 f8g7 d2b4 e7b4 e1b4 f6e4 a1e1 e6c4 b3c4 e4d6 c4d5 c7c5 b4c5 c8c5 e1e5 c5c2 b2b4 g4g6 a4a5 c2b2 e5e7 b2b4 e7d7');
INSERT INTO evaluations VALUES(20425,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(20425,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(20425,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(20425,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(20425,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(20425,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(20425,36,'cp 24','f7f6 g1f3 f6e5 d4e5 g8h6 f1b5 h6f5 b5d3 f8e7 d3f5 e6f5 e1g1 d5d4 d1b3 g7g5 f1d1 g5g4 f3d4 c6d4 b3a4 b7b5 a4d4 d8d4 d1d4 c8e6 b1c3 a7a6 c1e3 e8f7 c3e2 a8c8 e2f4 h8d8 d4d8 e7d8 a1d1 d8c7 e3d4 c7d8 h2h3 d8g5 f4e6 f7e6 h3g4 f5g4 d4c3');
INSERT INTO evaluations VALUES(21469,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(21469,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(21469,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(21469,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(21469,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(21469,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(21469,36,'cp -20','d5d4 a2a3 b7b6 f1c4 g8h6 d2d3 h6f5 d1e2 h7h5 b1d2 c6e7 d2e4 c8b7 e1g1 d8d7 e2f2 c7c5 a3a4 e7d5 c4b5 b7c6 c2c4 d5e3 f1e1 f8e7 f2e2 a8d8 b2b3 d7c7 b5c6 c7c6 a4a5 b6b5 c1e3 d4e3 g2g3');
INSERT INTO evaluations VALUES(23154,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(23154,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(23154,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(23154,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(23154,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(23154,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(23154,36,'cp 66','d8f6 f3e3 f8b4 c2c3 b4e7 f1d3 f6h6 e3h6 g8h6 f2f3 h6g8 c3c4 d7d6 b1c3 c7c6 c1e3 g8f6 a1d1 e8g8 e1g1 c8e6 h2h3 f6d7 b2b3 f7f6 g2g4 a7a6 c3e2 e6f7 e2d4 g7g6 g1f2 d7c5 d3c2 f8e8 h3h4');
INSERT INTO evaluations VALUES(23328,30,'cp 215',NULL);
INSERT INTO evaluations VALUES(23328,31,'cp 215',NULL);
INSERT INTO evaluations VALUES(23328,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(23328,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(23328,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(23328,35,'cp 218',NULL);
INSERT INTO evaluations VALUES(23328,36,'cp 196','d1d4 g8e7 b1c3 e7c6 d4e3 f8d6 c1d2 b7b6 e1c1 c8b7 f2f4 c6a5 c4d5 d8e7 e4e5 f6e5 d5b7 a5b7 c3d5 e7f7 f4e5 d6c5 e3b3 e8g8 h1f1 f7e6 f1f8 c5f8 d5c7 e6b3 a2b3 a8c8 c7b5 a7a6 b5d4 b7c5 d2e3 b6b5 c2c3 c8e8 d4f3 c5b3 c1c2 b3a5 d1d7 a5c6 e3f4 e8e7 d7d1 b5b4 c3b4 c6b4 c2b3 b4c6');
INSERT INTO evaluations VALUES(25519,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(25519,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(25519,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(25519,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(25519,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(25519,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(25519,36,'cp 26','f8e7 a2a4 e8g8 c2c3 h7h6 d2d4 e5d4 c3d4 c6b4 b1c3 c7c6 b5c4 d6d5 e4d5 f6d5 c1d2 d5c7 d2f4 d7f5 a4a5 e7d6 f4d6 d8d6 c3e4 f5e4 e1e4 b4d5 f3e5 c7e6 d1c2 a8e8 a1d1 e8e7 a5a6 b7a6 c2d3');
INSERT INTO evaluations VALUES(26639,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(26639,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(26639,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(26639,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(26639,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(26639,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(26639,36,'cp 68','d2d4 e7e6 e1g1 a8b8 c2c3 f8d6 b1d2 g8f6 f1e1 e8g8 e2e4 f6e4 d2e4 d5e4 e1e4 g4f5 e4e1 h7h6 b2b4 c6e7 a2a4 c7c6 c1d2 b7b6 c3c4 a7a5 b4b5 b8c8 d1b3 c6c5 d4c5 d6c5 e1d1 d8d3 d2c3 d3c2 b3c2 f5c2 d1d2 c2e4 a1d1 e7g6 f3d4 e4g2');
INSERT INTO evaluations VALUES(29187,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(29187,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(29187,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(29187,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(29187,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(29187,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(29187,36,'cp -39','h2h3 g4h5 d2d4 b8c6 c1e3 e7e6 g2g4 h5g6 f3h4 f8d6 h4g6 h7g6 b1d2 e8g8 c2c3 f8e8 e1g1 d8c7 d1f3 b7b5 a2a4 a7a6 g1g2 c6a5 a4b5 a6b5 b3c2 b5b4 f1c1 b4c3 b2c3 a5c6 c2d3 e6e5 d3b5 e8c8 a1a8 c8a8 b5c6');
INSERT INTO evaluations VALUES(29186,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(29186,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(29186,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(29186,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(29186,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(29186,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(29186,36,'cp -51','g4h5 d2d4 b8c6 g2g4 h5g6 c1e3 e7e6 f3h4 f8d6 h4g6 h7g6 b1d2 e8g8 c2c3 f8e8 e1g1 d8c7 d1f3 b7b5 a2a4 a7a6 a4b5 a6b5 h3h4 a8a1 f1a1 c6a5 b3d1 b5b4 a1c1 b4c3 b2c3 a5c6 h4h5 g6h5 g4h5 e6e5 d2f1 e5d4 c3d4 c6d4 f3h3 d4c6 h5h6');
INSERT INTO evaluations VALUES(30364,30,'cp 127',NULL);
INSERT INTO evaluations VALUES(30364,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(30364,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(30364,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(30364,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(30364,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(30364,36,'cp 111','f1d3 d7d6 g1f3 b8d7 f3g5 e7e6 e1g1 d8f6 c1e3 h7h6 g5h3 e6e5 d4d5 f6d8 c3e2 g8f6 e2g3 h6h5 f2f3 d7c5 a1c1 h5h4 g3e2 c7c6 c2c4 c5d3 d1d3 c6d5 c4d5 f8e7 e2c3 e8g8 b2b4 f6h5 d3d2');
INSERT INTO evaluations VALUES(31189,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(31189,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(31189,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(31189,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(31189,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(31189,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(31189,36,'cp 75','f3e5 h7h5 f1d3 d8d4 e5f7 e8f7 d3f5 d4d1 e1d1 g7g6 f5d3 g8f6 h1e1 b8d7 c3e4 f6e4 e1e4 e7e6 d1e2 d7f6 e4c4 f8h6 c1h6 h8h6 a2a4 h6h8 c4b4 h8b8 a4a5 f6d5 b4c4 a7a6 h2h4 b8d8 e2f1 f7f6 c4e4 d8d6 a1e1 a8d8');
INSERT INTO evaluations VALUES(31470,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(31470,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(31470,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(31470,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(31470,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(31470,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31470,36,'cp 26','f1d3 g8f6 e1g1 f6e4 c3e4 d5e4 d3e4 b8d7 h2h3 g4h5 g2g4 h5g6 e4g6 h7g6 g1g2 b7b5 b2b3 f8d6 c2c4 e8g8 f1e1 d7f6 c1g5 d8c7 d1d3 b5c4 b3c4 d6b4 e1e2 a8d8 a2a3 b4d6 a1d1 c6c5 d4d5 e6d5 c4d5 c5c4 d3c3 d6e7 d1c1 d8d5 c3c4 c7c4 c1c4 e7a3');
INSERT INTO evaluations VALUES(35827,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(35827,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(35827,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(35827,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(35827,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(35827,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35827,36,'cp 39','d4c5 e7c5 d1c2 c5b4 f1e2 d5c4 e1g1 b8c6 e2c4 b4c3 c2c3 f6d5 c4d5 e6d5 f3d4 c6d4 c3d4 c8e6 f1c1 d8b6 d4b6 a7b6 a2a3 f8c8 f4c7 a8a6 c7e5 c8c1 a1c1 b6b5 f2f3 a6c6 e5c3 h7h5 g1f2 f7f6 c1d1 h5h4 d1d2 c6c4 c3d4');
INSERT INTO evaluations VALUES(37056,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(37056,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(37056,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(37056,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(37056,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(37056,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(37056,36,'cp 43','b7b6 b2b3 d8e7 f1d3 c8b7 e1g1 f8d8 d1e2 c7c5 a1c1 b8d7 f1d1 d5c4 d3c4 f6e4 c3a1 a8c8 d4c5 d7c5 d1d8 e7d8 e2b2 e4f6 c4e2 c5d3 c1c8 b7c8 b2c3 c8a6 c3d4 d8d4 f3d4 g8f8 d4c6 d3b4 e2a6 b4a6');
INSERT INTO evaluations VALUES(37636,30,'cp 425',NULL);
INSERT INTO evaluations VALUES(37636,31,'cp 404',NULL);
INSERT INTO evaluations VALUES(37636,32,'cp 426',NULL);
INSERT INTO evaluations VALUES(37636,33,'cp 411',NULL);
INSERT INTO evaluations VALUES(37636,34,'cp 429',NULL);
INSERT INTO evaluations VALUES(37636,35,'cp 428',NULL);
INSERT INTO evaluations VALUES(37636,36,'cp 421','d4b5 d7d5 c1f4 b8a6 e4d5 e6e5 d1e2 d8e7 d5d6 e7e6 f4g3 g8h6 e2e4 e8g8 e1c1 h6f5 f1c4 a6c5 c4e6 c5e4 c3e4 c8e6 b5c7 e6d7 c7a8 f8a8 h1e1 a8c8 e4c3 f5g3 h2g3 g8f8 f2f4 e5f4 g3f4 c8d8 c3e4 b7b6 c2c3 h7h6 c3c4 d7c6 g2g3 f7f5 e4c3 g7f6 c3d5 d8d6 d5f6');
INSERT INTO evaluations VALUES(40023,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40023,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40023,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40023,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40023,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(40023,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40023,36,'cp 0','d7d5 d1b3 e8g8 c4d5 b8d7 f1c4 d7b6 e1g1 b6c4 b3c4 b7b6 c1g5 c8b7 g5f6 e7f6 f1e1 d8d6 h2h4 f8d8 h4h5 d6d7 c4b3 g7f8 c3e4 f8g7 e4c3');
INSERT INTO evaluations VALUES(40417,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40417,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(40417,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40417,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40417,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40417,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40417,36,'cp 0','d1b3 e8g8 c4d5 b8d7 a2a4 a7a5 f1e2 d7b6 e1g1 f6d5 c3d5 b6d5 c1d2 d8d6 f1e1 d5b6 h2h4 c8e6 b3b5 e6d7 b5b3');
INSERT INTO evaluations VALUES(41522,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(41522,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41522,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41522,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41522,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41522,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41522,36,'cp 0','f6d7 g1f3 c7c5 d4c5 e8g8 e5e6 f7e6 f1c4 d7c5 d1e2 d8b6 e1g1 b8c6 c1e3 b6b2 c3b5 c5e4 g1h1 b2b4 a1b1 b4a5 e2d3 e4f6 e3d2 a5d8 d3d8 c6d8 b5c7 a8b8 c7a6 b8a8 a6c7');
INSERT INTO evaluations VALUES(43261,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(43261,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(43261,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(43261,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(43261,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(43261,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(43261,36,'cp 50','f1d3 e8g8 g1f3 d7d6 e1g1 b8d7 c1e3 e7e5 b1d2 b7b6 a2a4 e5d4 c3d4 c7c5 e3f4 d6d5 e4e5 f6h5 f4g5 f7f6 e5f6 g7f6 g5h6 f6g7 h6g7 h5g7 f1e1 d8f6 d3b5 a7a6 b5f1 c8b7 a4a5 c5c4 a5b6 f6b6 h2h3 b6b2');
INSERT INTO evaluations VALUES(43618,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43618,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(43618,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(43618,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(43618,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(43618,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(43618,36,'cp 71','f8e7 c1f4 e8g8 f1e1 f8e8 a2a4 e7f8 f4g3 b8c6 d4c6 b7c6 e4e5 d6e5 e1e5 d8d1 a1d1 e8e5 g3e5 c8f5 e5f6 g7f6 c4d3 f5g6 g1f1 a8d8 h2h3 f8c5 f1e2 d8e8 e2f3 f6f5 a4a5 g8g7 c3a4 c5b4 d3a6 b4a5 a4c5');
INSERT INTO evaluations VALUES(39770,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(39770,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(39770,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(39770,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(39770,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(39770,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(39770,36,'cp 73','f1e1 e8g8 c1f4 f8e8 a2a4 e7f8 f4g3 c7c6 d4f3 b8d7 g3d6 d7b6 d6f8 d8d1 e1d1 b6c4 f8c5 b7b6 b2b3 b6c5 b3c4 c8g4 h2h3 g4h5 d1e1 h5f3 g2f3 a8b8 c3d1 b8b4 a4a5 f6d7 d1e3 d7e5 g1g2 e5c4 g2g3 e8d8 f3f4 c4e3 e1e3');
INSERT INTO evaluations VALUES(44976,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(44976,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(44976,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(44976,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(44976,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(44976,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(44976,36,'cp 84','e4e5 f6g8 d3d4 c8f5 f1e2 e7e6 e1g1 b8d7 a2a4 a7a5 c2c3 h7h6 b1d2 g8e7 d2b3 f5g6 f3h4 g6e4 f2f3 e4h7 f3f4 e7f5 h4f5 h7f5 g2g4 f5e4 b3d2 e4h7 f1f2 d8b6 d2f1 f7f6 e2d3 e8c8 d3h7 h8h7 f1g3 h7h8');
INSERT INTO evaluations VALUES(46193,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(46193,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(46193,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(46193,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(46193,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(46193,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(46193,36,'cp -7','f8d6 f4d6 d8d6 f1e2 e8g8 g1f3 b7b6 e1g1 b8c6 b1d2 f8d8 f1e1 c8b7 e2b5 a7a6 b5a4 f6d7 a1c1 e6e5 e3e4 e5d4 c3d4 c6d4 f3d4 c5d4 a4d7 d6d7 e4e5 d4d3 d1f3 a8c8 c1c8 b7c8 f3d3 d5d4 d2c4 d7c7 c4d2 c8e6 d2f3 e6a2 f3d4');
INSERT INTO evaluations VALUES(47375,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(47375,31,'cp -113',NULL);
INSERT INTO evaluations VALUES(47375,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(47375,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(47375,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(47375,35,'cp -110',NULL);
INSERT INTO evaluations VALUES(47375,36,'cp -98','e7e5 g1f3 e5d4 d1d4 d8d4 f3d4 f6e4 b1d2 f8c5 d4b5 e4d2 c1d2 b8a6 e1c1 e8g8 b5c3 c8g4 d1e1 g4e6 f1a6 b7a6 d2f4 c5b6 c3a4 a8c8 e1e2 h7h6 h1e1 f8d8 f4e5 e6g4 a4b6 a7b6 e2e3 c7c5 h2h3 g4f5 g2g4 f5g6 e5c3 b6b5 c3a5 d8d7');
INSERT INTO evaluations VALUES(50185,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(50185,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(50185,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(50185,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50185,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(50185,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50185,36,'cp 0','f8g7 c1b2 c7c5 d4c5 b8a6 g2g3 d8a5 b1d2 a6c5 f1g2 e8g8 g1f3 d7d6 e1g1 c8d7 f3d4 a7a6 d4c2 a5c7 a1c1 d7c6 g2h3 c6d7');
INSERT INTO evaluations VALUES(43178,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43178,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43178,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43178,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43178,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43178,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43178,36,'cp 0','c1b2 c7c5 d4c5 b8a6 g2g3 e8g8 g1f3 a6c5 f1g2 d7d5 c4d5 d8d5 e1g1 d5h5 b1d2 f8d8 d1c2 c8e6 b3b4 c5a6 c2a4 e6d7 a4b3 d7e6');
INSERT INTO evaluations VALUES(43179,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43179,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43179,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43179,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43179,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43179,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43179,36,'cp 0','c7c5 d4c5 b8a6 g2g3 e8g8 f1g2 d8a5 b1d2 a6c5 g1f3 d7d6 e1g1 c8d7 f3d4 a7a6 d4c2 a5c7 a1c1 d7c6 g2h3 c6d7');
INSERT INTO evaluations VALUES(50205,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(50205,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(50205,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50205,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50205,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(50205,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(50205,36,'cp 14','c1b2 f8g7 g1f3 e8g8 e2e3 b7b6 f1e2 c7c5 c2c4 c8b7 e1g1 c5d4 f3d4 d7d5 b1d2 b8d7 a1c1 a8c8 b3b4 a7a6 d4f3 d5c4 c1c4 c8c4 d2c4 b7d5 a2a4 d8c7 c4d2 c7b7 d1b1 f8c8 f1c1 c8c1 b1c1 a6a5 b4a5 b6a5');
INSERT INTO evaluations VALUES(51478,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51478,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51478,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(51478,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(51478,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(51478,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51478,36,'cp 3','c1b2 e7e6 b4b5 g8f6 g1f3 f8e7 a2a4 e8g8 g2g3 a7a6 c2c4 a6b5 a4b5 a8a1 b2a1 c7c6 f1g2 c6b5 c4b5 d7d5 e1g1 c8b7 d2d3 b8d7 b1d2 d8c7 a1b2 f8a8 d1b1 e7f8 f1c1 c7d6 c1c2 a8a4 c2c1');
INSERT INTO evaluations VALUES(51479,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(51479,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51479,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(51479,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51479,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51479,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51479,36,'cp 2','e7e6 b4b5 g8f6 g1f3 f8e7 a2a4 d7d5 c2c4 e8g8 e2e3 c8b7 f1e2 b8d7 e1g1 c7c5 d2d3 a7a6 d1c2 h7h6 b1d2 d8c7 f1c1 f8c8 b2c3 a6a5 c2b2 e7d6 h2h3 c8e8 c1d1 a8d8 a1c1 e6e5 c4d5 f6d5 d2c4 d5c3 b2c3');
INSERT INTO evaluations VALUES(51860,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51860,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(51860,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(51860,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(51860,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(51860,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51860,36,'cp -5','d7d5 e4d5 c6d5 d3d4 e5e4 f2f3 g8f6 f3e4 f6e4 c3e4 d5e4 f1c4 f8d6 d1h5 d8c7 c4b3 e8g8 g1e2 c8e6 c2c3 e6b3 a2b3 b8c6 e1g1 g7g6 h5h4 f7f5 c1h6 f8f7 e2f4 d6f4 h4f4 c7f4 h6f4 f7d7 h2h3 a8e8 b3b4 g8g7 g2g4 e4e3 g4f5');
INSERT INTO evaluations VALUES(53765,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(53765,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(53765,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53765,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53765,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53765,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53765,36,'cp 0','d3d4 d5e4 d4e5 d8d1 c3d1 c8e6 d1c3 b8d7 c1f4 f7f6 c3e4 f6e5 f4g3 g8f6 e4d2 f8c5 g1f3 e8c8 f1d3 f6h5 g3h4 d8e8 e1g1 h5f4 d2e4 e6d5 f1e1 c5b4 e1e3 f4d3 c2d3 e8e6 e4g5 e6e8');
INSERT INTO evaluations VALUES(54897,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(54897,31,'cp 108',NULL);
INSERT INTO evaluations VALUES(54897,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(54897,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(54897,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(54897,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(54897,36,'cp 96','f4e5 g8f6 h2h4 h8g8 e5f6 e7f6 c2c4 d5c4 h4g5 f6g5 f1c4 b8d7 b1c3 d7b6 c4b3 d8e7 g1f3 c8e6 b3e6 f7e6 d1b3 e8c8 c3e4 f8g7 a2a4 a7a5 a1c1 c8b8 c1c5 d8d5 e1g1 g5g4 f3d2 h6h5 f1c1 d5c5 e4c5');
INSERT INTO evaluations VALUES(55338,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(55338,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(55338,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(55338,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(55338,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(55338,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(55338,36,'cp 77','d2d4 e5d4 d1d4 g8f6 c1g5 f8g7 b1c3 h7h6 g5e3 e8g8 d4d2 f8e8 f1d3 d6d5 e4e5 f6e4 c3e4 d5e4 d3e4 d8d2 f3d2 e8e5 e1c1 b8c6 h1e1 c8e6 a2a3 a8e8 e3f4 e5b5 c2c3 e8e7 d2f3 b5a5 e4c6 b7c6 f3d4 g7d4 d1d4 g6g5 d4d8 g8h7 f4h2 h6h5 d8a8 a7a6 f2f3');
INSERT INTO evaluations VALUES(55369,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(55369,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(55369,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(55369,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(55369,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(55369,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(55369,36,'cp 153','h2h4 g8f6 g2g4 h7h5 g4h5 f6h5 c1g5 f8e7 d1d2 a7a6 f1g2 b7b5 e4e5 d6d5 e1c1 c8b7 g1f3 b8d7 g2h3 d7b6 c3e2 b6c4 d2d3 b5b4 h3g4 a6a5 d1g1 b7a6 d3d1 c4b6 f3d2 a6e2 g4e2 d8d7 d2b3 a5a4 b3c5 e7c5 d4c5');
INSERT INTO evaluations VALUES(56573,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(56573,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(56573,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(56573,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(56573,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(56573,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(56573,36,'cp 50','g1f3 f8d6 c2c3 c7c5 c3d4 e5d4 e2e3 d4e3 c1e3 g8e7 e1g1 e8g8 b1c3 e7f5 e3g5 f7f6 g5d2 c8e6 f1e1 e6f7 g2h3 f5d4 f3d4 c5d4 c3e4 b8c6 h3g2 f8e8 a2a3 d6f8 a1c1 d8d7 b2b4 a7a5 b4a5 c6a5 d2b4 a5b3 c1b1 f8b4');
INSERT INTO evaluations VALUES(56572,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(56572,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(56572,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(56572,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(56572,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(56572,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(56572,36,'cp 29','f8d6 e1g1 c7c5 e2e3 g8e7 e3d4 c5d4 a2a3 e8g8 f1e1 b8c6 b2b4 a7a6 c2c3 c8g4 h2h3 g4e6 c3d4 e5d4 b1d2 h7h6 c1b2 d8d7 d2e4 a8d8 h3h4 e7f5 f3d2 d6e7 d2c4 b7b5 c4d2 a6a5 d2b3 a5b4 e4c5 e7c5');
INSERT INTO evaluations VALUES(4385,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(4385,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4385,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(4385,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4385,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(4385,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4385,36,'cp 0','b2b4 c5b6 f1e2 d7d5 b1d2 f8e8 e1g1 d5d4 c3d4 c6d4 f3d4 b6d4 a1b1 c7c5 b4c5 d4c5 d2f3 b7b6 c1b2 c5d6 d3d4 e5d4 e2b5 c8d7 b5c4 d7e6 c4b5');
INSERT INTO evaluations VALUES(5665,30,'cp 287',NULL);
INSERT INTO evaluations VALUES(5665,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(5665,32,'cp 286',NULL);
INSERT INTO evaluations VALUES(5665,33,'cp 300',NULL);
INSERT INTO evaluations VALUES(5665,34,'cp 288',NULL);
INSERT INTO evaluations VALUES(5665,35,'cp 292',NULL);
INSERT INTO evaluations VALUES(5665,36,'cp 303','f3e5 d6e5 d1h5 g7g6 h5e5 g8f6 c1h6 d8d6 e5d6 c7d6 b1c3 c8e6 c4b5 f6d7 e1c1 a7a6 b5e2 g6g5 f2f4 g5f4 h6f4 d7e5 g2g3 h8g8 c3d5 a8c8 h1e1 e7g5 d5e3 c8c6 h2h4 g5f6 e3f5 e8d7 f5d4 c6b6 b2b3 g8c8 d1d2 e5g6 f4e3 b6b4 d4f5 f6e5 e1d1 b4e4 f5d6 e4e3 d6c4 d7c7');
INSERT INTO evaluations VALUES(5664,30,'cp 261',NULL);
INSERT INTO evaluations VALUES(5664,31,'cp 305',NULL);
INSERT INTO evaluations VALUES(5664,32,'cp 305',NULL);
INSERT INTO evaluations VALUES(5664,33,'cp 294',NULL);
INSERT INTO evaluations VALUES(5664,34,'cp 288',NULL);
INSERT INTO evaluations VALUES(5664,35,'cp 306',NULL);
INSERT INTO evaluations VALUES(5664,36,'cp 290','d6e5 d1h5 g7g6 h5e5 g8f6 c1h6 d8d6 e5d6 c7d6 b1c3 c8e6 c4b5 f6d7 e1c1 a7a6 b5e2 g6g5 f2f4 g5f4 h6f4 h8g8 g2g3 d7e5 c3d5 a8c8 h1e1 e7g5 d5e3 c8c6 h2h4 g5f6 e3f5 e8d7 c1b1 g8e8 b2b3 e8g8 e2f1 b7b5 f1h3 h7h5 e1e2 e6f5 e4f5 d7e7 e2d2 g8c8 h3g2 c6b6');
INSERT INTO evaluations VALUES(5672,30,'cp 317',NULL);
INSERT INTO evaluations VALUES(5672,31,'cp 305',NULL);
INSERT INTO evaluations VALUES(5672,32,'cp 312',NULL);
INSERT INTO evaluations VALUES(5672,33,'cp 306',NULL);
INSERT INTO evaluations VALUES(5672,34,'cp 312',NULL);
INSERT INTO evaluations VALUES(5672,35,'cp 312',NULL);
INSERT INTO evaluations VALUES(5672,36,'cp 306','d1h5 g7g6 h5e5 g8f6 c1h6 d8d6 e5d6 c7d6 b1c3 c8e6 c4b5 f6d7 e1c1 a7a6 b5e2 g6g5 f2f4 g5f4 h6f4 d7e5 g2g3 e6h3 c3d5 a8c8 d1d2 h7h5 h1d1 h3g4 h2h4 g4e2 d2e2 c8c6 a2a4 f7f6 b2b3 e7d8 c1b2 e8f7 c2c3 e5g6 f4e3 g6e5 e2d2 h8g8');
INSERT INTO evaluations VALUES(6831,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(6831,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(6831,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(6831,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(6831,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(6831,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(6831,36,'cp 71','d1h5 c5b4 c2c3 g8f6 h5h4 d4c3 b2c3 b4e7 e4e5 f6d5 h4g3 e8f8 b5c4 c7c6 e1g1 h7h5 h2h3 d7d6 f2f4 c8f5 d3d4 h8h6 c4d3 d8d7 g3f3 f8g8 b1d2 h6g6 d3e4 h5h4 f3d3 f5e4 d2e4 d7f5 f1f3 d5b6 a2a4 d6d5 e4f2 f5d3 f2d3 g6g3');
INSERT INTO evaluations VALUES(7052,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(7052,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(7052,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(7052,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(7052,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(7052,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(7052,36,'cp 95','b1c3 d8e7 d6e7 c6e7 c1g5 b7b5 f2f3 c8b7 e1c1 h7h6 g5h4 d7d5 h4f6 g7f6 c3d5 b7d5 e4d5 e7f5 f1d3 f5d6 c2c3 e8d7 c1c2 h8c8 a2a4 c8c5 b2b4 c5c7 a4b5 a6b5 d1a1 c7a7 a1a7 a8a7 c2b3 f6f5 h1e1 f7f6 f3f4 e5e4 d3e2 a7a8 g2g3 d7c7 e1f1 h6h5 b3b2 c7d7 f1d1 d7c7 d1c1');
INSERT INTO evaluations VALUES(9492,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(9492,31,'cp -132',NULL);
INSERT INTO evaluations VALUES(9492,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(9492,33,'cp -118',NULL);
INSERT INTO evaluations VALUES(9492,34,'cp -133',NULL);
INSERT INTO evaluations VALUES(9492,35,'cp -139',NULL);
INSERT INTO evaluations VALUES(9492,36,'cp -113','f6g4 c1e3 g4e3 f2e3 c5e3 d1f3 e3g5 g2g3 e8g8 f1g2 d8d4 c3d1 f7f5 c2c3 d4b6 f3f2 f5e4 f2b6 c7b6 g2e4 c8g4 h2h4 g5h6 d1f2 h6d2 e1d2 f8f2 d2e3 f2b2 a2a4 a8f8 h1e1 g4f5 e4f5 f8f5 e3d4 f5f3 e1e3 f3e3 d4e3 g8f7 c3c4 f7f6 e3e4 b2b3 a4a5 b6b5 c4b5 b3b5 d3d4 f6e6 a5a6 b7b6 g3g4 e6d6');
INSERT INTO evaluations VALUES(9510,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(9510,31,'cp -107',NULL);
INSERT INTO evaluations VALUES(9510,32,'cp -130',NULL);
INSERT INTO evaluations VALUES(9510,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(9510,34,'cp -113',NULL);
INSERT INTO evaluations VALUES(9510,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(9510,36,'cp -123','c1e3 g4e3 f2e3 c5e3 d1f3 e3g5 g2g3 e8g8 f1g2 d8d4 c3d1 f7f5 c2c3 d4b6 f3f2 f5e4 f2b6 a7b6 g2e4 c8g4 a2a4 g5h6 h2h3 g4f5 e1e2 f5e4 d3e4 a8d8 d1e3 f8e8 h1e1 e8e4 e2f3 e4e5 e3g4 e5f5 f3e4 f5h5 g4h6 h5h6 e4f4 h6h3 e1e7 d8f8 f4g4 h3h2 e7c7 h7h5 g4g5 f8f3 a1g1 f3f7 c7f7 g8f7 b2b3 f7e7 g1e1 e7f7');
INSERT INTO evaluations VALUES(9511,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(9511,31,'cp -125',NULL);
INSERT INTO evaluations VALUES(9511,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(9511,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(9511,34,'cp -116',NULL);
INSERT INTO evaluations VALUES(9511,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(9511,36,'cp -136','g4e3 f2e3 c5e3 d1f3 e3g5 g2g3 e8g8 f1g2 d8d4 c3d1 f7f5 c2c3 d4b6 f3f2 f5e4 f2b6 a7b6 g2e4 c8g4 a2a4 g5h6 h2h3 g4f5 e1e2 f5e4 d3e4 a8d8 d1e3 f8e8 e3f5 h6g5 a1d1 e8e4 e2f3 e4e5 g3g4 g7g6 d1d8 g5d8 h1d1 d8f6 f5e3 g8f7 d1d7 e5e7 d7e7 f7e7 f3e4 e7e6 e3c4 b6b5 a4b5 c6b5 c4a5');
INSERT INTO evaluations VALUES(10355,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(10355,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(10355,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(10355,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(10355,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(10355,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(10355,36,'cp -26','f1d3 e8g8 g1e2 d7d5 c3d5 f6d5 e4d5 d8d5 e1g1 e5e4 d3c4 d5d2 e2g3 d2d1 a1d1 f8e8 c4d5 c8g4 d5c6 b7c6 d1d4 a7a5 h2h3 g4e6 d4e4 f7f6 a2a3 b4c5 f1e1 g8f7 h3h4 a5a4 b3b4 c5f8 b2d4 a8d8 c2c3 d8d5 h4h5 f6f5 e4e5 g7g6 h5g6 h7g6 f2f3');
INSERT INTO evaluations VALUES(10862,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(10862,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(10862,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(10862,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(10862,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(10862,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(10862,36,'cp 147','b5c6 b7c6 f3e5 e8g8 e1g1 a7a5 f1e1 c8b7 b1c3 a5a4 a1b1 f8e8 c1d2 d7d5 d1f3 e7d6 e5g4 f6g4 f3g4 b7c8 g4f3 c8e6 h2h3 d5d4 c3e2 c6c5 b1a1 f7f6 d2f4 e6f7 f4d6 c7d6 e2f4 d8a5 e1b1 a5d2 b1c1 f7e6 f3g3');
INSERT INTO evaluations VALUES(11375,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(11375,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(11375,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(11375,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(11375,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(11375,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(11375,36,'cp 77','c2c3 f8d6 b1d2 f6d7 f4g3 e6e5 e3e4 e5d4 c3d4 d5e4 d3e4 e8g8 e4c6 b7c6 e1g1 f8e8 a1c1 a6a5 d1c2 d7f6 c2c6 c8e6 a2a3 e6d5 c6c3 f6e4 d2e4 e8e4 f1e1 e4e1 c1e1 a5a4 f3e5 h7h6 e5c4 d5e6 g3d6 c7d6 c4e3 a8c8 c3d3');
INSERT INTO evaluations VALUES(13379,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13379,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13379,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13379,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13379,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13379,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13379,36,'cp 0','g7g6 h2h3 c6d4 d1d4 f8g7 f1d3 e8g8 e1g1 c8d7 d4f2 d7c6 a2a4 d8b6 f2b6 a7b6 a1a3 f8c8 b2b4 c6e8 e4e5 d6e5 f4e5 f6d7 c3d5 g7f8 d3b5 c8c2 b5d7 e8d7 d5b6 d7c6 b6a8 c2g2 g1h1 g2d2 h1g1 d2g2');
INSERT INTO evaluations VALUES(13683,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(13683,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13683,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(13683,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(13683,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(13683,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(13683,36,'cp 24','e3d4 f6h5 f4e3 f8d6 e1g1 h5f4 e3f4 d6f4 g2g3 f4d6 b1d2 c8d7 a2a3 a8c8 a1c1 e8g8 f1e1 d7e8 d1e2 g8h8 b2b4 c6e7 h2h4 h7h6 d3b1 e7g8 c3c4 d5c4 d2c4 d6b8 c4e5 c8c1 e1c1 g8f6 b4b5 f6d5');
INSERT INTO evaluations VALUES(14633,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14633,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14633,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14633,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14633,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14633,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14633,36,'cp 0','f6e4 c3e4 d6d5 c4d3 d5e4 d3e4 f8d6 e1g1 e8g8 f1e1 c6d4 b2b4 d4f3 e4f3 c7c6 c1b2 d8g5 h3h4 g5f6 e1e3 a7a5 b4a5 a8a5 g2g3 c8e6 d2d4 e5d4 b2d4 d6e5 d4e5 a5e5 d1d4 e5e3 d4f6 g7f6 f2e3 f8d8 a1b1 d8d2 c2c3 d2d3 b1b7 d3c3');
INSERT INTO evaluations VALUES(19719,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(19719,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(19719,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(19719,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(19719,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(19719,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(19719,36,'cp -57','e7e6 d2d4 f8d6 d1g4 d8f6 c3e2 g8e7 c1f4 h7h5 g4f3 d6f4 f3f4 f6f4 e2f4 h5h4 g1f3 f7f6 c2c3 a7a6 b5a4 b7b5 a4b3 g7g5 f4d3 g5g4 f3d2 a6a5 a2a4 b5b4 f2f3 e6e5 f3g4 c8g4 d4e5 f6e5 c3c4 d5d4 d3c5 h4h3 g2h3 h8h3 e1g1 g4e2 d2e4 e2f1 e4d6 e8d8 a1f1');
INSERT INTO evaluations VALUES(20851,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(20851,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(20851,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(20851,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(20851,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(20851,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(20851,36,'cp 39','c8g4 h2h3 g4h5 f1e2 e7e6 c3a4 f6f5 c2c4 g8f6 c1d2 f6e4 a4c3 d8d7 a1c1 f8b4 e1g1 e8c8 c4d5 e6d5 f3e5 h5e2 d1e2 c6e5 d4e5 b4e7 b2b4 c8b8 c3e4 f5e4 d2c3 g7g5 e5e6 d7e6 c3h8 d8h8 b4b5 h7h5 e2d1 e7d6 d1a4 g5g4 h3h4');
INSERT INTO evaluations VALUES(21475,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(21475,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(21475,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(21475,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(21475,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(21475,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(21475,36,'cp 100','d2d4 f7f6 c2c3 f6e5 f3e5 c6e5 d4e5 d8h4 b1d2 g8h6 d2f3 h4e4 f1e2 f8e7 e1g1 e8g8 h2h3 e4g6 e2d3 g6h5 b2b4 c8d7 a2a4 c7c5 b4c5 e7c5 a4a5 a8c8 d1e2 b7b5 a5b6 c5b6 c1e3 c8c3 e3b6 a7b6 a1a7 f8f3 e2f3 h5f3 g2f3');
INSERT INTO evaluations VALUES(21983,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(21983,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(21983,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(21983,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(21983,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(21983,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(21983,36,'cp 3','c6d4 f3d4 e5d4 c3e2 d6d5 d2d3 c7c5 g2g3 d5e4 f1g2 e4d3 d1d3 g8e7 c2c3 e7c6 c3d4 c5d4 g2c6 b7c6 d3e4 c8e6 e4c6 e6d7 c6e4 f8e7 c1g5 f7f6 g5f4 e8g8 e1g1 d7h3 f1c1 f8e8 e4d4 e7a3 a1a3 e8e2 d4c4 e2e6 a3e3 d8d7 e3e6 h3e6 c4c6 d7c6');
INSERT INTO evaluations VALUES(22153,30,'cp 181',NULL);
INSERT INTO evaluations VALUES(22153,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(22153,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(22153,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(22153,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(22153,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(22153,36,'cp 185','b7b6 b2b4 a5b7 f1b5 c8d7 b5a6 a8b8 e1g1 f8e7 c2c4 g8f6 b1c3 e8g8 d1e2 c7c6 d5c6 d7c6 c3d5 c6d7 a2a4 f6e8 a4a5 b6a5 b4b5 e8f6 a6b7 b8b7 d2a5 d8e8 a5b4 f6d5 e4d5 f7f6 f3d2 f6f5 a1a6 b7b5 b4d6 b5b6 d6e7 b6a6 e7f8 g8f8');
INSERT INTO evaluations VALUES(18251,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(18251,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(18251,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(18251,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(18251,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(18251,35,'cp 202',NULL);
INSERT INTO evaluations VALUES(18251,36,'cp 212','b2b4 a5b7 f1b5 c8d7 b5a6 d7c8 c2c4 g8f6 b1c3 f8e7 a2a3 e8g8 e1g1 f6d7 a6b5 d7f6 a1c1 g8h8 b5c6 a7a6 h2h3 a8b8 d2e3 c8d7 d1a4 b8a8 a4c2 a8b8 f3e1 d8e8 c6d7 f6d7 e1d3 f7f5 e4f5 f8f5 c3e4 f5f8 f2f4');
INSERT INTO evaluations VALUES(24918,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(24918,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24918,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(24918,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(24918,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(24918,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(24918,36,'cp 48','f3e5 c6e5 d4e5 f6h5 f4e3 g7g6 d1b3 g4c8 g2g3 c7c6 f1g2 d8c7 f2f4 c8e6 e3f2 f7f6 e5f6 e7f6 e1g1 f6f5 b1d2 h5f6 e2e4 f5e4 d2e4 f6e4 g2e4 c7d7 e4f3 f8d6 f1e1 e8g8 e1e6 d7e6 b3b7 f8c8 a1e1 e6f7 b7a6 f7d7 c3c4');
INSERT INTO evaluations VALUES(25905,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(25905,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(25905,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(25905,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(25905,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(25905,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(25905,36,'cp -82','b7c6 e1g1 c6c5 d4f3 g8e7 c2c4 d7c6 c4d5 e7d5 c1g5 f8e7 g5e7 d8e7 d1e2 a8b8 f1c1 e8g8 b1c3 d5f4 e2e3 f4g6 b2b3 b8b4 f3d2 b4d4 c3e2 d4d7 d2f1 c6b5 e2f4 g6f4 e3f4 d7d4 f4e3 f8d8 h2h3 b5c6 f2f3 h7h6 c1c3 e7b7 c3c5');
INSERT INTO evaluations VALUES(25822,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(25822,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(25822,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(25822,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(25822,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(25822,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(25822,36,'cp -86','e1g1 c6c5 d4f3 g8e7 c2c4 d7c6 c4d5 e7d5 b1d2 f8e7 d2c4 e8g8 c1e3 d8c7 d1e2 h7h6 b2b3 a8d8 f1c1 d5e3 e2e3 d8d5 h2h3 f8d8 e3e2 a7a5 e2c2 d5d7 c1d1 c6f3 d1d7 d8d7 g2f3 c7d8 a1c1');
INSERT INTO evaluations VALUES(25821,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(25821,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(25821,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(25821,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(25821,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(25821,35,'cp -99',NULL);
INSERT INTO evaluations VALUES(25821,36,'cp -91','c6c5 d4f3 g8e7 c2c4 d7c6 c4d5 e7d5 b1a3 f8e7 c1e3 e8g8 a3c4 d8c7 d1e2 d5e3 e2e3 f8d8 f1d1 d8d5 d1d2 a8d8 b2b3 h7h6 h2h3 d5d2 f3d2 e7g5 e3e2 g5f4 c4d6 f4e5 e2e5 c7d6 e5d6 d8d6 d2c4 d6d4 c4e5 c6b7');
INSERT INTO evaluations VALUES(25828,30,'cp -98',NULL);
INSERT INTO evaluations VALUES(25828,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(25828,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(25828,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(25828,34,'cp -85',NULL);
INSERT INTO evaluations VALUES(25828,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(25828,36,'cp -79','d4f3 g8e7 c2c4 d7c6 c4d5 e7d5 b1a3 f8e7 a3c4 e8g8 c1e3 d8c7 d1e2 a8d8 b2b3 h7h6 a1c1 c6b5 f1d1 d5b4 h2h3 b5a6 a2a3 b4c6 e2c2 d8d1 c1d1 f8d8 d1d2 d8d2 c2d2 a6c4 b3c4 c6e5 f3e5 c7e5 d2d7 e5a1 g1h2');
INSERT INTO evaluations VALUES(25830,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(25830,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(25830,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(25830,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(25830,34,'cp -99',NULL);
INSERT INTO evaluations VALUES(25830,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(25830,36,'cp -81','g8e7 c2c4 d7c6 c4d5 e7d5 b1a3 f8e7 c1e3 e8g8 a3c4 d8b8 d1c1 d5b6 f3d2 b8c7 c1c3 b6d5 c3a3 d5e3 a3e3 a8d8 b2b3 d8d4 d2f3 d4e4 e3c3 c6d5 c4e3 d5b7 a1d1 b7a8 d1d2 f8d8 d2d8 e7d8 e3c4');
INSERT INTO evaluations VALUES(25976,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(25976,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(25976,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(25976,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(25976,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(25976,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(25976,36,'cp 0','c6d4 b5d7 d8d7 d1d4 g8f6 a2a4 f8e7 f1d1 d7c6 c1g5 e8g8 g5f6 e7f6 d4d6 c6c4 d6a3 f8d8 a3b3 c4c6 g2g3 h7h6 b3b5 c6c7 b5b3');
INSERT INTO evaluations VALUES(27673,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(27673,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(27673,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(27673,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(27673,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(27673,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(27673,36,'cp 30','d1d7 e8d7 e1e2 d7e6 c1e3 g7g6 g2g4 f7f5 g4f5 g6f5 e4f5 e6f5 h1g1 g8f6 b1d2 f8d6 g1g5 f5e6 a1g1 a7a6 g5g3 h8g8 e3g5 f6h5 g3g4 a8f8 d2e4 g8g6 b2b3 f8g8 h2h3 d6e7 g5e7 e6e7 e2e3 h5f4 g4g6 f4d5 e3e2 h7g6');
INSERT INTO evaluations VALUES(27650,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(27650,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(27650,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(27650,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(27650,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(27650,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(27650,36,'cp 20','e8d7 e1e2 g7g6 b1c3 d7e6 a2a4 g8f6 f2f3 f8b4 c3d1 b4e7 a4a5 f6e8 d1e3 f7f5 e3d5 e7d6 c1e3 c7c6 d5c3 d6e7 h1d1 h7h5 d1d3 e8d6 b2b3 f5e4 c3e4 d6e4 f3e4 a7a6 a1f1 a8d8 c2c3 d8c8 d3d1 h8g8 f1f3 c8d8');
INSERT INTO evaluations VALUES(27702,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(27702,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(27702,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(27702,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(27702,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(27702,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(27702,36,'cp 31','a2a4 g7g6 a4a5 d7e6 e1e2 f7f5 b1c3 g8f6 a1a4 f8d6 a5a6 b7b6 a4c4 f5f4 c3b5 h8c8 g2g3 g6g5 h1d1 d6c5 b2b4 c7c6 b5c7 c8c7 b4c5 b6c5 c1a3 c7d7 d1d3 d7d3 c2d3 f6d7 g3f4 g5f4 a3c5 d7c5 c4c5 e6d6 d3d4 e5d4 c5h5');
INSERT INTO evaluations VALUES(28676,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(28676,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(28676,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(28676,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(28676,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(28676,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(28676,36,'cp 93','c3d5 b7d5 c1f4 d7d6 f1b5 b8d7 c2c4 d5b7 d1a4 f8e7 e1g1 a7a6 b5c6 b7c6 a4c6 a8c8 d4d5 e8g8 d5e6 f7e6 f4g3 d8e8 a1d1 g8h8 c6e4 d7e5 f3e5 d6e5 g3e5 e8g6 e4e2 e7f6 b2b3 f6e5 e2e5 g6f6 e5g3 c8d8 d1d8 f6d8 h2h3');
INSERT INTO evaluations VALUES(28674,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(28674,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(28674,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(28674,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(28674,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(28674,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(28674,36,'cp 76','b7d5 c1f4 b8c6 h2h4 h7h6 h1h3 d7d6 a2a3 a7a6 f1d3 g7g6 d1e2 f8g7 a1d1 b6b5 e1f1 d8d7 h3g3 d6e5 f3e5 g7e5 f4e5 c6e5 e2e5 e8c8 a3a4 g6g5 h4g5 h6g5 a4b5 a6b5 e5g5 d8g8 g5g8 h8g8 g3g8 c8b7');
INSERT INTO evaluations VALUES(33046,30,'cp 303',NULL);
INSERT INTO evaluations VALUES(33046,31,'cp 301',NULL);
INSERT INTO evaluations VALUES(33046,32,'cp 311',NULL);
INSERT INTO evaluations VALUES(33046,33,'cp 305',NULL);
INSERT INTO evaluations VALUES(33046,34,'cp 301',NULL);
INSERT INTO evaluations VALUES(33046,35,'cp 322',NULL);
INSERT INTO evaluations VALUES(33046,36,'cp 318','c5d6 g1e2 b8c6 c1f4 c8e6 e1c1 d6f4 e2f4 e8c8 f1c4 e6d7 h1e1 g8h6 d5c3 h8e8 e1e8 d8e8 h2h3 c6e5 c4d5 e8e7 d5b3 e5g6 f4d5 e7e8 d5e3 d7c6 b3d5 g6e7 d5c6 b7c6 d1d4 e7f5 e3f5 h6f5 d4f4 e8e5 c1d2 h7h5');
INSERT INTO evaluations VALUES(33037,30,'cp 307',NULL);
INSERT INTO evaluations VALUES(33037,31,'cp 300',NULL);
INSERT INTO evaluations VALUES(33037,32,'cp 313',NULL);
INSERT INTO evaluations VALUES(33037,33,'cp 326',NULL);
INSERT INTO evaluations VALUES(33037,34,'cp 336',NULL);
INSERT INTO evaluations VALUES(33037,35,'cp 325',NULL);
INSERT INTO evaluations VALUES(33037,36,'cp 314','g1e2 c8e6 c1f4 b8c6 e1c1 d6f4 e2f4 e8c8 f1c4 e6d7 h1e1 g8h6 d5c3 h8e8 e1e8 d8e8 c3b5 e8e7 f4d5 e7e6 h2h3 c6e5 c4e2 d7b5 e2b5 h6f5 d5f4 e6e7 a2a4 e5c6 b2b3 e7e4 g2g3 g7g5 f4d5 e4e5 b5c6 b7c6 d5b4 e5e2 b4c6 e2f2 c6a7 c8b7');
INSERT INTO evaluations VALUES(34393,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(34393,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(34393,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(34393,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(34393,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(34393,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(34393,36,'cp 53','d2d4 e7e6 g1f3 g8f6 f1d3 f8e7 e1g1 e8g8 a2a3 c8d7 f1e1 b8c6 c1f4 f8c8 d3c2 d7e8 h2h3 c8d8 b2b4 a5b6 f4e3 a8c8 d1d3 h7h6 a1d1 e7f8 d4d5 b6c7 c3b5 c6b4 b5c7 b4d3 c7e8 d3e1 e8f6 g7f6 f3e1 d8d5 d1d5 e6d5');
INSERT INTO evaluations VALUES(34499,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(34499,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(34499,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(34499,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(34499,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(34499,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(34499,36,'cp -18','g8e7 b1c3 h7h5 d2d4 c7c6 c3d1 b8a6 c2c3 f7f6 h2h4 a6c7 c1f4 c7e6 f4d6 g7g6 d1e3 e6g7 f1d3 e7f5 d6f8 e8f8 d3f5 g7f5 e3f5 c8f5 e1d2 a7a5 f3g1 f8f7 g1e2 g6g5 a2a4 a8e8 f2f3 f5g6 e2c1 b7b6 c1d3');
INSERT INTO evaluations VALUES(34528,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(34528,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(34528,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(34528,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(34528,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(34528,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(34528,36,'cp 126','g8f6 f1c4 c7c6 d5c6 b8c6 g1f3 c8g4 e1g1 a7a6 f1e1 g4h5 c4e2 e7e6 d2d3 d6c7 f3h4 h5e2 c3e2 h7h6 c2c3 f8d6 e2g3 e8g8 h4f3 f8d8 d1e2 b7b5 g3e4 f6e4 e2e4 b5b4 c3c4 d6c5 c1e3 c5e3 f2e3');
INSERT INTO evaluations VALUES(35647,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(35647,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(35647,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(35647,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(35647,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(35647,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(35647,36,'cp 86','c1e3 e7e5 e1g1 b8c6 d4d5 c6e7 f3d2 f6d7 b2b4 a7a5 a2a3 f7f5 f2f3 g8h8 d2b3 a5b4 a3b4 a8a1 d1a1 e7g8 b3d2 f5e4 d2e4 d7f6 e4f6 g8f6 a1c1 d8e7 e3g5 e7e8 c1d2 c8d7 f1e1 e8f7 g5e3 f6h5 e2f1 h5f4 h2h3 h7h6 c4c5 b6c5 b4c5 d6c5 e3c5');
INSERT INTO evaluations VALUES(36676,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(36676,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(36676,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(36676,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(36676,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(36676,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(36676,36,'cp 55','e1g1 b8d7 a2a4 e7e5 d4e5 d7e5 f3e5 d6e5 a4a5 d8e7 b1c3 c8e6 b2b3 f8d8 c1a3 e7e8 d1f3 b7b6 f1d1 a7a6 a5b6 c7b6 d3c4 b6b5 c4e6 e8e6 c3d5 f6d5 e4d5 e6f5 c2c4 f5f3 g2f3 a8c8 a3e7 d8d7 d5d6 b5c4 b3c4');
INSERT INTO evaluations VALUES(36675,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(36675,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(36675,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(36675,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(36675,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(36675,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(36675,36,'cp 70','b8d7 a2a4 e7e5 d4e5 d7e5 f3e5 d6e5 a4a5 c8e6 d1f3 d8e7 b2b3 f8d8 c1a3 e7e8 b1c3 a7a6 c3a4 f6h5 a4c5 e6c8 f1d1 g7f8 d3c4 d8d1 f3d1 h5f4 d1f3 b7b5 a5b6 c7b6 c5d3 f4d3 a3f8 e8f8 f3d3 c8b7 a1d1 f8e7 d3e2');
INSERT INTO evaluations VALUES(36923,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(36923,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(36923,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(36923,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(36923,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(36923,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(36923,36,'cp 175','c3d5 b8d7 f1e2 h7h6 g5h4 g7g5 d5f6 d7f6 h4g3 c7c5 g3d6 d8d6 d4c5 d6c5 d1d4 c5c6 e1g1 c8f5 c2c3 f8e8 f1e1 f5e4 d4d2 a8d8 f3d4 c6b6 h2h3 e4g6 e2f3 f6e4 f3e4 g6e4 a2a4 g8h7 a1d1 e4g6 b2b4 b6f6 e1e8 d8e8 a4a5');
INSERT INTO evaluations VALUES(36936,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(36936,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(36936,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(36936,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(36936,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(36936,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(36936,36,'cp -23','c3b5 d6e7 c1f4 c7c6 f4c7 d8e8 b5c3 c8f5 e1g1 e8c8 c7f4 f8e8 e2d3 a7a5 a2a4 b8a6 d3f5 c8f5 c3e2 f6e4 c2c3 e7d6 f4e5 f7f6 e5d6 e4d6 f1e1 a6c7 b2b3 f5d7 e2g3 c7e6 d1d2 g7g6 h2h4 e6g7 e1e8 a8e8 h4h5');
INSERT INTO evaluations VALUES(39159,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(39159,31,'cp 191',NULL);
INSERT INTO evaluations VALUES(39159,32,'cp 184',NULL);
INSERT INTO evaluations VALUES(39159,33,'cp 176',NULL);
INSERT INTO evaluations VALUES(39159,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(39159,35,'cp 165',NULL);
INSERT INTO evaluations VALUES(39159,36,'cp 187','c2c3 b4e7 f1d3 d7d5 e4e5 c7c5 e1g1 c5c4 d3c2 b8c6 b2b3 b7b5 f1e1 a7a6 h2h3 c8d7 c1f4 g7g5 f4e3 h7h5 f3h2 d8c7 b1d2 c7a5 b3c4 b5c4 d2f3 a5c3 e3d2 c3a3 d2g5 e7g5 f3g5 g8e7 h2f3 c6b4 a1b1 b4c2 d1c2 a3d3 c2c1 a8c8');
INSERT INTO evaluations VALUES(39157,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(39157,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(39157,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(39157,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(39157,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(39157,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(39157,36,'cp 163','b4e7 f1d3 d7d5 b1d2 f7f5 e4e5 g8h6 b2b4 h6f7 e1g1 b7b6 a2a4 a7a5 b4b5 b8d7 c3c4 c8b7 d1c2 c7c6 c1a3 c6b5 c4b5 a8c8 c2b2 e7a3 a1a3 e8g8 a3c3 g7g5 c3c8 b7c8 f1c1 g5g4 f3e1 f7g5 f2f4 g4f3 e1f3 g5e4');
INSERT INTO evaluations VALUES(39260,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(39260,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(39260,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(39260,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(39260,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(39260,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(39260,36,'cp -38','d7d5 e4d5 e5e4 f3g5 d8g5 d2d4 g5d5 d4c5 d5d1 e1d1 b8c6 h2h3 c8d7 c1d2 g8e7 d1c1 e8c8 b1c3 c6d4 b2b4 d4f3 d2e3 e7f5 a2a4 f5e3 f2e3 h7h5 f1e2 f3e5 c1b2 d7e6 a1d1 d8d1 c3d1 f7f5 h3h4 a7a5 b4a5 e6c4 e2c4');
INSERT INTO evaluations VALUES(39536,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(39536,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(39536,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(39536,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(39536,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(39536,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(39536,36,'cp 12','c2c4 g8f6 b1c3 d7d6 e2e4 e7e6 g1f3 e8g8 f1d3 b8a6 e1g1 a6c5 d3c2 f8e8 h2h3 e6d5 e4d5 c7c6 c1g5 c6d5 c3d5 c8e6 d5f6 g7f6 g5f6 d8f6 d1d6 f6b2 c2g6 h7g6 d6c5 b2f6 a1e1 e6h3 e1e8 a8e8 c5a7 h3d7 f1e1 e8e1 f3e1 f6a1');
INSERT INTO evaluations VALUES(39534,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(39534,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39534,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39534,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(39534,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39534,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39534,36,'cp 0','d7d6 e2e4 g8f6 b1c3 e8g8 g1f3 e7e6 f1d3 e6d5 c4d5 f8e8 e1g1 b8d7 d3c2 d7c5 f1e1 a7a5 a2a4 c7c6 h2h3 c6d5 e4d5 e8e1 d1e1 c8d7 a1b1 d8b6 e1d1 a8e8 c1e3 b6b4 e3d2 b4b6');
INSERT INTO evaluations VALUES(42078,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(42078,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(42078,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(42078,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(42078,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(42078,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(42078,36,'cp 42','d2c3 d5c4 e2e3 b7b5 a2a4 c7c6 a4b5 c6b5 b2b3 e8g8 b3c4 b5c4 f1c4 c8b7 c3b4 f8e8 e1g1 a7a5 b4a3 b8d7 d1e2 d8b6 f1c1 e8c8 a1b1 b6a7 f3d2 h7h6 h2h3 a5a4 c4b5 c8c1 b1c1 a7a5 f2f3 a5a7 b5d7 f6d7 c1c7 d7b6 e2b5 a8c8 c7c5 a7a6 c5c8 b7c8 b5e8 g8h7 e8f7 b6d5');
INSERT INTO evaluations VALUES(43552,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(43552,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(43552,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(43552,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(43552,34,'cp 123',NULL);
INSERT INTO evaluations VALUES(43552,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(43552,36,'cp 128','f1d3 f8e7 e1g1 b8c6 f1e1 e8g8 c1f4 c6d4 f3d4 e6e5 f4e3 e5d4 e3d4 b7b5 h2h3 c8e6 d1d2 f6d7 a1d1 e7f6 d4f6 d8f6 b2b3 d7c5 d3b1 a8d8 f2f4 b5b4 c3e2 a6a5 e2g3 a5a4 b3a4 e6d7 e4e5 d6e5 f4e5 f6h4');
INSERT INTO evaluations VALUES(43553,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(43553,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(43553,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(43553,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(43553,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(43553,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(43553,36,'cp 109','f8e7 e1g1 b8c6 c1e3 e8g8 a1c1 f6g4 e3d2 b7b5 c3e2 c8b7 e2g3 g4f6 d1e2 a8c8 c1d1 d8b6 d2e3 b6c7 h2h3 c6b8 f1e1 b8d7 f3h2 f8e8 a2a3 g7g6 d3b1 d7b6 e3d2 e7f8 e4e5 c7c6 f2f3 d6e5 d4e5 f6d5');
INSERT INTO evaluations VALUES(39749,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(39749,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(39749,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(39749,33,'cp 124',NULL);
INSERT INTO evaluations VALUES(39749,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(39749,35,'cp 110',NULL);
INSERT INTO evaluations VALUES(39749,36,'cp 105','e1g1 b8c6 a2a3 c8d7 c1f4 e8g8 a1c1 c6d4 f3d4 e6e5 f4e3 e5d4 e3d4 a8c8 f2f4 f8e8 d1f3 d7e6 h2h3 f6d7 d3b1 e6b3 f1f2 e7f6 f2d2 f6d4 d2d4 d8e7 c1f1 d7c5 f1f2 c5e6 d4d2 e6c5 g1h2 b7b5 f3g3 b3c4 d2d4 g8h8 f2f3 e7f6 g3f2 c5b3 d4d1 h7h6 b1c2');
INSERT INTO evaluations VALUES(39748,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(39748,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(39748,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(39748,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(39748,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(39748,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(39748,36,'cp 101','b8c6 a2a3 c8d7 c1f4 e8g8 a1c1 c6d4 f3d4 e6e5 f4e3 e5d4 e3d4 d7e6 f2f4 a8c8 h2h3 f6d7 d3b1 b7b5 g1h2 d7c5 b1c2 e7f6 f1f3 f6d4 d1d4 e6b3 f3g3 g7g6 d4f2 b3c2 c1c2 f8e8 c2e2 a6a5 g3f3 b5b4 a3b4 a5b4');
INSERT INTO evaluations VALUES(45939,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(45939,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(45939,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(45939,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(45939,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(45939,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(45939,36,'cp 151','b1c3 h7h5 h2h4 g7g6 g2g3 d8e7 g1h3 b8a6 f1g2 f6h7 f2f4 f8g7 h3f2 c8d7 f4f5 a6c7 a2a4 e8c8 g2h3 g6f5 e4f5 e5e4 c3e4 g7d4 e1f1 c7e8 c1f4 h7f6 d1f3 a7a5 e4g5 d4b2 a1b1 b2d4 f1g2');
INSERT INTO evaluations VALUES(46019,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46019,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46019,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46019,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(46019,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(46019,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46019,36,'cp 0','d3d4 c5d4 f3d4 f8e7 f1e2 a7a6 c1e3 e8g8 a2a4 b7b6 e2f3 f6d7 f3g4 d7f6 g4f3');
INSERT INTO evaluations VALUES(47685,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(47685,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(47685,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(47685,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(47685,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(47685,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(47685,36,'cp 12','f8d6 a2a3 b8c6 b1c3 a7a6 f1d3 c6d4 f3d4 e6e5 f4g3 e5d4 e3d4 d6g3 f2g3 d8e7 d1e2 e7e2 c3e2 c8d7 g3g4 d7b5 e1d2 e8g8 h1e1 b5d3 d2d3 f8e8 e2c3 c7c6 b2b4 g8f8 a3a4 a6a5 b4b5 e8e1 a1e1 a8e8 e1b1 f6e4 c3e4 e8e4 b5c6 b7c6');
INSERT INTO evaluations VALUES(49406,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(49406,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(49406,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(49406,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(49406,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(49406,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(49406,36,'cp -9','f6h5 f4e3 f8e7 h2h4 d7d5 g2g4 h5f6 g4g5 f6g4 e3f4 h7h6 g1h3 h6g5 h4g5 b8c6 d2f3 e6e5 d4e5 c8e6 e2e3 d8d7 d1e2 d5d4 f3d4 c6d4 e3d4 d7d4 c2c3 d4a4 e2b5 a4b5 f1b5 c7c6 b5d3 g4e5 f4e5 h8h3 h1h3 e6h3 e5f4 h3e6 e1c1 e8c8');
INSERT INTO evaluations VALUES(49867,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(49867,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(49867,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(49867,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(49867,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(49867,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(49867,36,'cp 76','h2h3 g4h6 d2d4 d7d6 d1e2 c8e6 c1f4 d6e5 e2e5 b8c6 f1b5 h6f5 e5c7 d8c7 f4c7 a7a6 b5c6 b7c6 b1d2 a8c8 c7f4 h7h5 e1g1 f8e7 a1e1 e8g8 c2c3 c6c5 g2g4 h5g4 h3g4 f5h4 d4d5 e6d5 e1e7 h4g6 f4d6 g6e7 d6e7 f8e8 e7d6 e8e6 d6f4 c8e8 f1f2 a6a5 a2a3');
INSERT INTO evaluations VALUES(52025,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(52025,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(52025,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(52025,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(52025,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(52025,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(52025,36,'cp -56','d7d5 e4d5 e6d5 c4e2 g8f6 g1f3 f8d6 d2d4 b8c6 d4c5 d6c5 e1g1 e8g8 b1d2 c5b6 c2c3 a7a6 d2b3 f8e8 f1e1 c8g4 c1f4 h7h6 h2h3 f6e4 b3d4 g4h5 f4e3 h5g6 e2f1 b6c7 g2g3 d8d7 d4e2 g6f5 h3h4 f5g4 f1g2 a8d8 e2f4 c7f4 e3f4');
INSERT INTO evaluations VALUES(52133,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(52133,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(52133,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(52133,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(52133,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(52133,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(52133,36,'cp 17','d2d3 d7d6 f1g2 b8c6 b1c3 g8e7 f2f4 e5f4 c1f4 e7g6 f4e3 f8e7 d1d2 e7f6 g1e2 e8g8 e1g1 c8e6 f1f2 b7b5 a2a3 a8b8 a1f1 b5b4 a3b4 c5b4 c3d1 d6d5 f2f6 g7f6 e2d4 c6d4 e3d4 g6e5 d4a7 d5e4 a7b8');
INSERT INTO evaluations VALUES(52567,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(52567,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(52567,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(52567,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(52567,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(52567,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(52567,36,'cp 114','d4e5 c8g4 b1d2 e7e6 h2h3 g4h5 f1e2 g8h6 e1g1 d8d7 g2g4 h5f7 c2c4 f8e7 c4d5 c6d5 d2b3 b8c6 f3d4 e8g8 f2f4 c6d4 b3d4 e7c5 c1e3 f7g6 a1c1 c5b6 d1d2 g6e4 a2a4 b6d4 d2d4 h6f7 d4b4 a7a6 a4a5 f8c8 c1c3 c8c3 b2c3 a8c8 g1h2 g8h8 f1c1');
INSERT INTO evaluations VALUES(52575,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(52575,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(52575,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(52575,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(52575,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(52575,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(52575,36,'cp 40','g8f6 f1e2 f8g7 b1c3 e8g8 e1g1 f8e8 a2a4 a7a5 c4d5 c6d5 c1d2 b8c6 d1b3 a8b8 h2h3 e7e6 f1c1 c8d7 d2e1 d8e7 f3e5 e8c8 e5d3 b7b6 d3e5 d7e8 e5c6 e8c6 c3b5 f6e8 f2f3 c6b5 e2b5 e8d6 e1g3');
INSERT INTO evaluations VALUES(52588,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(52588,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(52588,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(52588,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(52588,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(52588,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(52588,36,'cp 66','f8g7 h2h3 g8f6 c1e3 e8g8 a2a4 b8d7 f1e2 e7e5 d4e5 d7e5 e1g1 d8c7 d1d2 e5f3 e2f3 c8e6 e3f4 c7a5 f1d1 f8d8 a1b1 f6e8 b2b4 a5c7 f4g5 d8d7 c3e2 e8f6 d2f4 d6d5 g5f6 c7f4 e2f4 g7f6 e4d5 c6d5 f4d5 e6d5 d1d5 d7d5 f3d5');
INSERT INTO evaluations VALUES(52928,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(52928,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(52928,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(52928,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(52928,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(52928,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(52928,36,'cp 54','d3e4 d8c7 g1f3 b8d7 a2a4 e7e5 f1d3 d7c5 d2c4 g8f6 e1g1 c8e6 f3g5 e6g4 d3e2 a8d8 d1e1 g4e2 e1e2 h7h6 g5f3 f8d6 f3h4 e8g8 h4f5 c5e4 c4d6 e4d6 e2e5 f6e8 f5e7 g8h7 f1e1 d8d7 e7f5 d6f5 e5f5 h7h8 c1e3 e8f6 f5f3');
INSERT INTO evaluations VALUES(52920,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(52920,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(52920,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(52920,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(52920,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(52920,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(52920,36,'cp 36','b8d7 g1f3 d8c7 a2a4 e7e5 d2c4 g8f6 f1d3 d7c5 e1g1 c8e6 f3g5 f8e7 g5e6 c5e6 c1d2 e8g8 d1f3 f6d7 f1d1 f8d8 b2b4 e7g5 d2g5 e6g5 f3e3 g5e6 c2c3 b7b6 d3f1 d7f6 g2g3 d8d1 a1d1 a8d8 d1d8 e6d8 f2f3 h7h6 h2h4 d8e6 g1g2');
INSERT INTO evaluations VALUES(54936,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(54936,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(54936,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(54936,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(54936,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(54936,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(54936,36,'cp 159','e2e4 b7b5 a2a4 g8f6 a4b5 f6e4 b1c3 c8b7 f1c4 e7e6 e1g1 b8d7 f1e1 e4c3 b2c3 f8d6 c4d3 e8g8 d3c2 f8e8 d1d3 f7f5 c1b2 d7f6 c3c4 b7e4 d3d1 c7c6 b5c6 d8b6 c2b3 a8b8 b2a3 d6a3 a1a3 e4f3 d1f3 b6d4 c6c7 b8c8 b3a4 e8e7 e1d1');
INSERT INTO evaluations VALUES(56271,30,'cp -85',NULL);
INSERT INTO evaluations VALUES(56271,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(56271,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(56271,33,'cp -72',NULL);
INSERT INTO evaluations VALUES(56271,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(56271,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(56271,36,'cp -74','d5d4 c3b1 f8e7 g1f3 g8f6 f1g2 c7c5 e1g1 b8c6 e2e3 c8f5 e3d4 c5d4 d2d3 e8g8 f1e1 f8e8 c1g5 d8b6 b1a3 e7a3 b2a3 f6d7 a1b1 b6a6 d1b3 e8e1 f3e1 a8e8 g2d5 f5e6 g5d2 d7c5 b3c4 e6d5 c4d5 c5d3 e1d3 a6d3 b1b7');
INSERT INTO evaluations VALUES(56956,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(56956,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(56956,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56956,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(56956,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(56956,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(56956,36,'cp 24','g8f6 e2e3 f6h5 g2g3 d7d6 h2h4 a7a6 f1g2 c7c5 b1c3 c5d4 e3d4 h5f4 g3f4 d6d5 d1d3 b8c6 e1c1 f8b4 a2a3 b4c3 d3c3 c8d7 g2h3 d8c7 f3e5 a8c8 h1g1 c6e5 c3c7 c8c7 f4e5 e8g8 d1e1 g8h8');
INSERT INTO evaluations VALUES(49241,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(49241,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(49241,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(49241,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(49241,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(49241,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(49241,36,'cp 23','c2c4 f8b4 b1d2 d7d6 a2a3 b4d2 f4d2 b7b6 e2e3 c8b7 f1e2 a7a5 d1c2 b7e4 e2d3 e4d3 c2d3 e8g8 e1g1 f8e8 e3e4 b8d7 a1e1 d6d5 c4d5 e6d5 e4e5 f6e4 d2e3 c7c5 d4c5 d7c5 d3b5 d8d7 b5d7 c5d7 e1d1 d7e5 f3e5 e8e5 e3b6');
INSERT INTO evaluations VALUES(120,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(120,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(120,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(120,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(120,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(120,35,'cp 147',NULL);
INSERT INTO evaluations VALUES(120,36,'cp 151','c3d5 b6a5 e3d2 a5d8 d2c3 c6d4 c3d4 g8f6 d5f6 e7f6 f1c4 e8g8 d1d3 d7d6 e1c1 d8e7 f2f3 c8e6 c1b1 f8d8 h1e1 e7c7 c4e6 f7e6 h2h4 b7b6 g2g4 a8c8 d3e2 c7c6 d4c3 d6d5 e4d5 d8d5 d1d5 c6d5 e2e6 d5e6 e1e6 g8f7 e6d6 c8c7 a2a4 h7h5 a4a5 b6a5 d6a6 a5a4 a6a4 h5g4 f3g4 g7f8 b1c1 f8c5 c3d2 f6f5 g4f5');
INSERT INTO evaluations VALUES(1665,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(1665,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(1665,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(1665,33,'cp 142',NULL);
INSERT INTO evaluations VALUES(1665,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(1665,35,'cp 136',NULL);
INSERT INTO evaluations VALUES(1665,36,'cp 131','d2d4 c8g4 c4b5 e5d4 d1d4 e7d7 d4d3 g4f3 d3f3 f8e7 c3d5 a7a6 b5a4 b7b5 a4b3 g8f6 e4e5 d6e5 c1e3 e8g8 d5f6 e7f6 b3d5 c6d4 e3d4 a8d8 d4c5 d7d5 f3d5 d8d5 c5f8 g8f8 c2c3 f8e7 e1e2 d5c5 g2g3 e5e4 h1e1 f6e5 f2f3 e4f3 e2f3 e7d7 e1e4 f7f5 e4e2 e5d6');
INSERT INTO evaluations VALUES(1704,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(1704,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1704,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(1704,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(1704,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(1704,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1704,36,'cp 0','c3d4 c8g4 f1e2 g4f3 e2f3 f8b4 b1c3 d5c4 f3c6 b7c6 d1e2 c4e2 e1e2 g8e7 c1e3 e8d7 h1c1 h7h5 c3a4 e7d5 e3d2 a8b8 c1c2 h8e8 e2f3 e8e6 a1c1 b8e8 d2b4 e6f6 f3g3 f6g6 g3f3');
INSERT INTO evaluations VALUES(4359,30,'cp -91',NULL);
INSERT INTO evaluations VALUES(4359,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(4359,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(4359,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(4359,34,'cp -83',NULL);
INSERT INTO evaluations VALUES(4359,35,'cp -76',NULL);
INSERT INTO evaluations VALUES(4359,36,'cp -86','e5d4 c3a4 c5d6 c2c3 d4c3 a4c3 h7h6 f1e1 d6c5 c1e3 c5e3 e1e3 f6g4 e3e1 d7d6 h2h3 g4e5 f3e5 c6e5 c4f1 f7f5 e4f5 c8f5 d1d2 f5h7 a1d1 g8h8 f2f4 e5d7 g2g4 h7g8 f1g2 a8b8 f4f5 a7a6 b2b3 f8e8 e1e8 d8e8 g4g5 h6g5');
INSERT INTO evaluations VALUES(4403,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(4403,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(4403,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(4403,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(4403,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(4403,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(4403,36,'cp -61','c3a4 c5d6 c2c3 d4c3 a4c3 h7h6 h2h3 a7a6 f1e1 b7b5 c4d5 f6d5 e4d5 c6e7 c3e4 c8b7 e4d6 c7d6 c1f4 e7d5 f4d6 f8e8 e1e8 d8e8 f3d4 e8e4 d1g4 e4g4 h3g4 a8c8 a1d1 c8c4 d1d2 f7f6 f2f3 c4c1 g1h2 c1b1 d4f5 b7c6 f5e7 d5e7 d6e7 a6a5 a2a3 g8f7 e7c5 f7g6 c5e7 b1c1');
INSERT INTO evaluations VALUES(4855,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(4855,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(4855,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(4855,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(4855,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(4855,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4855,36,'cp 9','d7d6 c3d5 g8f6 d2d3 f6d5 c4d5 e8g8 c2c3 c5a7 d5b3 d8f6 a2a4 c8g4 h2h3 g4f3 d1f3 f6f3 g2f3 g8h8 b3d5 a6a5 f3f4 c6e7 d5b7 a8b8 b7a6 e5f4 c1f4 f7f5 a1b1 f5e4 f4h2 e4d3 a6d3 e7d5 d3b5 d5f4 h2f4 f8f4 b2b4 b8f8 b4a5');
INSERT INTO evaluations VALUES(4760,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(4760,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(4760,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(4760,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(4760,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(4760,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(4760,36,'cp 16','c3d5 g8e7 c2c3 e7d5 c4d5 d8f6 d2d3 e8g8 a2a4 h7h6 d5b3 f6g6 c1e3 a6a5 f1e1 f8d8 g1h1 c5e3 e1e3 c8e6 h2h3 g6f6 b3e6 f6e6 d1c2 c6e7 d3d4 e7g6 h1g1 g6f4 c3c4 e6f6 a1f1');
INSERT INTO evaluations VALUES(5670,30,'cp 322',NULL);
INSERT INTO evaluations VALUES(5670,31,'cp 329',NULL);
INSERT INTO evaluations VALUES(5670,32,'cp 352',NULL);
INSERT INTO evaluations VALUES(5670,33,'cp 340',NULL);
INSERT INTO evaluations VALUES(5670,34,'cp 319',NULL);
INSERT INTO evaluations VALUES(5670,35,'cp 348',NULL);
INSERT INTO evaluations VALUES(5670,36,'cp 351','d4g7 e7f6 g7g3 c8e6 c1d2 g8e7 f2f4 h8g8 g3f2 e7c6 f1b5 d8e7 e1c1 f6c3 d2c3 e6d7 h1e1 e8c8 b2b3 a7a6 b5c4 c8b8 c4f1 e7f8 g2g3 d8e8 c1b2 d7e6 e4e5 d6e5 f4e5 h7h5 f1g2 e8d8 d1d8 c6d8 g2e4 g8h8 e1d1 f8e7 e4g2 h5h4 g3h4');
INSERT INTO evaluations VALUES(10530,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(10530,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(10530,32,'cp -57',NULL);
INSERT INTO evaluations VALUES(10530,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(10530,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(10530,35,'cp -53',NULL);
INSERT INTO evaluations VALUES(10530,36,'cp -47','d2c3 e8g8 f2f3 d7d5 d1d2 d5d4 c3b4 c6b4 d2b4 a7a5 b4a3 f6d7 e1c1 d8f6 c1b1 b7b6 c2c3 d4c3 a3c3 f6c3 b2c3 d7e5 g1e2 f7f5 e2g3 f5e4 g3e4 c8f5 d1d4 a8d8 f1e2 c7c5 d4d8 f8d8 b1c2 d8f8 h1f1 g7g5 c2d2 f5e4 f3e4 f8f1 e2f1 g8g7 f1e2 g7f6');
INSERT INTO evaluations VALUES(10529,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(10529,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(10529,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(10529,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(10529,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(10529,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(10529,36,'cp -53','e8g8 f2f3 d7d5 d1d2 d5d4 c3b4 c6b4 d2b4 a7a5 b4d2 a5a4 g1e2 c7c5 e2f4 d8c7 c2c4 c8e6 f1e2 f6d7 e1g1 c7d6 b2b3 d6b6 d2b2 f7f5 e4f5 e6f5 e2d3 b6d6 d3f5 f8f5 f4d5 a8e8 b3a4 d7e5 b2b7 e5c4 b7c7 d6c7 d5c7');
INSERT INTO evaluations VALUES(11615,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(11615,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(11615,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(11615,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(11615,34,'cp 174',NULL);
INSERT INTO evaluations VALUES(11615,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(11615,36,'cp 173','e4d5 f6d5 d2d4 c8b7 f3e5 c6e5 d4e5 f8c5 b1d2 d8e7 d2e4 e8c8 e4c5 e7c5 d1g4 c8b8 c2c4 h7h5 g4e2 b5c4 e2c4 c5c4 b3c4 h8e8 c1g5 f7f6 e5f6 g7f6 g5h6 h5h4 h2h3 e8g8 f2f3 g8g6 h6c1 d5b4 b2b3 b7d5 f1d1 d8g8 g2g4 d5f3 c4g8 f3d1 g8c4 d1c2 c1a3 b4c6');
INSERT INTO evaluations VALUES(11611,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(11611,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(11611,32,'cp 177',NULL);
INSERT INTO evaluations VALUES(11611,33,'cp 177',NULL);
INSERT INTO evaluations VALUES(11611,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(11611,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(11611,36,'cp 188','f6d5 d2d4 c8b7 f3e5 c6e5 d4e5 f8c5 b1d2 d8e7 d2e4 e8c8 d1g4 c8b8 e4c5 e7c5 c2c4 h7h5 g4e2 b5c4 e2c4 c5c4 b3c4 h8e8 c1g5 f7f6 e5f6 g7f6 g5h6 e8h8 h6c1 d5b4 b2b3 b4d3 h2h3 d3e5 c4e2 h8g8 f2f3 g8e8 f1f2 e5d3 e2d3 d8d3 g1h2 e8e6 c1h6 h5h4 f2c2 e6c6 c2c6 b7c6');
INSERT INTO evaluations VALUES(11627,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(11627,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(11627,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(11627,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(11627,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(11627,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(11627,36,'cp -40','c6d4 f3d4 e5d4 a2a4 a8b8 d1d4 c7c5 d4d1 c5c4 b3a2 f8e7 b1c3 e8g8 c1e3 c8d7 e3d4 f8e8 h2h3 b5b4 c3e2 b8c8 e2g3 e7f8 f1e1 a6a5 d1f3 d7a4 a2c4 c8c4 d4f6 d8f6 f3f6 g7f6 a1a4 c4c2 a4a5 e8e5 a5a8 d6d5');
INSERT INTO evaluations VALUES(15477,30,'cp 521',NULL);
INSERT INTO evaluations VALUES(15477,31,'cp 536',NULL);
INSERT INTO evaluations VALUES(15477,32,'cp 530',NULL);
INSERT INTO evaluations VALUES(15477,33,'cp 532',NULL);
INSERT INTO evaluations VALUES(15477,34,'cp 550',NULL);
INSERT INTO evaluations VALUES(15477,35,'cp 555',NULL);
INSERT INTO evaluations VALUES(15477,36,'cp 575','g5f7 d8e7 f7h8 c8g4 f2f3 g4e6 d2d3 d6d5 e4d5 f6d5 c4d5 e6d5 e1g1 e8c8 b1c3 e7c5 g1h1 f8e7 d1e2 d5e6 c3e4 c5d5 c1g5 d8h8 g5e7 c6e7 f3f4 e7c6 f4f5 c6d4 e2f2 h8f8 f5f6 g7f6 e4f6 d5d8 c2c3 d4c6 a1d1 d8f6 f2f6 f8f6 f1f6');
INSERT INTO evaluations VALUES(14511,30,'cp 541',NULL);
INSERT INTO evaluations VALUES(14511,31,'cp 567',NULL);
INSERT INTO evaluations VALUES(14511,32,'cp 555',NULL);
INSERT INTO evaluations VALUES(14511,33,'cp 563',NULL);
INSERT INTO evaluations VALUES(14511,34,'cp 557',NULL);
INSERT INTO evaluations VALUES(14511,35,'cp 547',NULL);
INSERT INTO evaluations VALUES(14511,36,'cp 575','d8e7 f7h8 c8e6 d2d3 e6c4 d3c4 e8c8 b1c3 e7e6 c3d5 f8e7 e1g1 d8h8 b2b4 h7h5 d1d3 h5h4 h2h3 c8b8 b4b5 c6a5 c1d2 b7b6 d2a5 b6a5 a1d1 g7g5 f2f3 f6d5 d3d5 e6d5 d1d5 b8b7 f1d1 b7b6 c4c5 d6c5 d5e5 e7d6 e5f5 h8e8');
INSERT INTO evaluations VALUES(907,30,'cp 561',NULL);
INSERT INTO evaluations VALUES(907,31,'cp 573',NULL);
INSERT INTO evaluations VALUES(907,32,'cp 560',NULL);
INSERT INTO evaluations VALUES(907,33,'cp 567',NULL);
INSERT INTO evaluations VALUES(907,34,'cp 575',NULL);
INSERT INTO evaluations VALUES(907,35,'cp 576',NULL);
INSERT INTO evaluations VALUES(907,36,'cp 578','f7h8 c8g4 f2f3 g4e6 d2d3 d6d5 e4d5 f6d5 c4d5 e6d5 e1g1 e8c8 b1c3 e7c5 g1h1 f8e7 d1e2 d8h8 c1e3 c6d4 c3d5 d4e2 e3c5 e7c5 a1e1 e2f4 d5f4 e5f4 e1e4 c5e3 h2h4 h7h5 b2b3 g7g6 h1h2 c7c5 h2h3 a7a5 g2g3 c8c7 g3f4 e3d4 e4e6 c5c4 b3c4');
INSERT INTO evaluations VALUES(620,30,'cp 578',NULL);
INSERT INTO evaluations VALUES(620,31,'cp 568',NULL);
INSERT INTO evaluations VALUES(620,32,'cp 563',NULL);
INSERT INTO evaluations VALUES(620,33,'cp 559',NULL);
INSERT INTO evaluations VALUES(620,34,'cp 555',NULL);
INSERT INTO evaluations VALUES(620,35,'cp 553',NULL);
INSERT INTO evaluations VALUES(620,36,'cp 569','c8g4 f2f3 g4e6 d2d3 d6d5 c4d5 f6d5 e4d5 e6d5 e1g1 e8c8 b1c3 e7c5 g1h1 f8e7 d1e2 d5e6 c3e4 c5b6 c1e3 b6b2 a2a4 e6b3 f1c1 b3a2 h8f7 a2f7 c1b1 c6d4 e2f2 b2c2 e3d4 c2f2 d4f2 b7b6 f2g3 d8d3 g3e5 g7g6 a4a5 d3d5 a5b6 a7b6 e5g3');
INSERT INTO evaluations VALUES(15752,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(15752,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(15752,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(15752,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(15752,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(15752,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(15752,36,'cp 103','f3d4 e5d4 c3e2 d6d5 e4d5 f8c5 c1g5 h7h6 g5f6 d8f6 e1g1 c8g4 c4b5 e8f8 f2f3 g4f5 e2g3 c5d6 d1d2 g7g6 g3e4 f5e4 f3e4 f6e5 g2g3 f8g7 d2f4 a8f8 f4e5 d6e5 g1g2 a7a6 b5c4 f7f6 g2h3 e5d6 a1e1 h6h5 f1f2 b7b5 c4b3 a6a5 a2a4 f8b8');
INSERT INTO evaluations VALUES(15748,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(15748,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(15748,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(15748,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(15748,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(15748,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(15748,36,'cp 90','e5d4 c3e2 d6d5 e4d5 f8c5 e1g1 f6d5 c4d5 d8d5 e2f4 d5f5 f1e1 e8f8 f4h5 c8e6 c1f4 f7f6 f4c7 a8c8 c7g3 e6d7 e1e4 f8f7 b2b4 c5b6 a2a4 a7a5 b4a5 b6a5 a1b1 d7c6 e4d4 h8d8 d4c4 d8d7 h5f4 a5c7 d1h5 f5h5 f4h5 c7a5 h5f4 g7g5');
INSERT INTO evaluations VALUES(16584,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(16584,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(16584,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(16584,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(16584,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(16584,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(16584,36,'cp -49','d7d5 c1g5 d5d4 c3d5 d8d6 g5f6 g7f6 d1f3 f6f5 e4f5 c8d7 g2g4 e8c8 f1g2 c6a5 b2b4 d7c6 h2h4 h7h6 b4a5 d6c5 e1f1 c6d5 f3d1 h8g8 g2d5 c5d5 d1f3 g8g4 f3g4 d5h1 f5f6 c8b8 g4f3 h1f3 g1f3 d8d5 a5a6 b8c8 a6b7 c8b7 a1e1 f8d6 f3h2 a7a5 h2g4 d6a3 g4h6');
INSERT INTO evaluations VALUES(22214,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(22214,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(22214,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(22214,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(22214,34,'cp 195',NULL);
INSERT INTO evaluations VALUES(22214,35,'cp 187',NULL);
INSERT INTO evaluations VALUES(22214,36,'cp 184','d1g4 g8h6 c1h6 g7h6 g1h3 d8e7 c2c3 e7f7 f1e2 h8g8 g4h5 f7h5 e2h5 e8e7 e1g1 c6d8 b1a3 c8d7 h3f4 f8g7 h5e2 c7c5 e2d3 g8f8 d3h7 d7e8 f1f3 c5d4 c3d4 h6h5 a1f1 g7h6 g2g3 a8c8 h7d3 f8g8 a3c2 h5h4 g1f2 h4g3 h2g3');
INSERT INTO evaluations VALUES(22685,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(22685,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(22685,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(22685,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(22685,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(22685,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(22685,36,'cp 83','g1f3 d7d6 e5d6 f8d6 d2d4 g8f6 c1g5 e8g8 g2g3 c8f5 f1g2 c6b4 b1a3 h7h6 g5f6 d8f6 e1g1 f6e6 f3d2 f5g4 g2f3 g4f3 f1f3 c7c5 c2c3 b4c6 a3b5 c5d4 c3d4 d6b4 e2e3 b4d2 d1d2 a8d8 d2f2 c6b4 b5c3 f7f5 a1f1 b4d5 f3f5 f8f5 f2f5 e6f5 f1f5');
INSERT INTO evaluations VALUES(22683,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(22683,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(22683,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(22683,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(22683,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(22683,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(22683,36,'cp 64','d7d6 e5d6 f8d6 d2d4 g8f6 c1g5 e8g8 g2g3 c8f5 f1g2 c6b4 b1a3 h7h6 g5f6 d8f6 e1g1 f6e6 f3d2 f5g4 g2f3 g4f3 f1f3 c7c5 c2c3 b4c6 a3b5 d6b8 d2b3 c5d4 b5d4 e6e8 d4c6 b7c6 d1d3 b8c7 d3c4 c7b6 g1g2 a8d8 e2e4 d8d6 a1e1 d6e6 a2a4 e6e5 c4d3 e8e7 b3d4 e7c5');
INSERT INTO evaluations VALUES(21891,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(21891,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(21891,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(21891,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(21891,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(21891,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(21891,36,'cp 91','e5d6 f8d6 b1c3 d8e7 e2e3 g8f6 d2d4 f6g4 d1e2 e8g8 c1d2 f8e8 e3e4 c6d4 f3d4 e7h4 g2g3 d6g3 h2g3 h4g3 e1d1 g4f2 d1c1 c8g4 d4f5 g4f5 d2e1 g3f4 e2d2 f4d2 c1d2 a8d8 f1d3 f2h1 e4f5 d8d4 c3b5 d4g4 b5c7 g4g2 d2c3 e8c8');
INSERT INTO evaluations VALUES(23232,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(23232,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(23232,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(23232,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(23232,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(23232,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(23232,36,'cp 114','d7d5 e4d5 d8d5 c2c4 d5a5 e3d2 f8b4 d4d5 b4d2 b1d2 c6d8 a2a3 c7c6 b2b4 a5c7 d2e4 g8h6 c4c5 e8g8 d5d6 c7f7 f1d3 d8e6 d3c4 f7g6 e4g3 c8d7 d1b3 a8e8 e1g1 g8h8 c4d3 g6f7 f1e1 e6f4 b3c3 f4d3 c3d3 f7g6 d3g6 h7g6 a3a4 h8g8 a4a5 a7a6 e1d1 d7e6');
INSERT INTO evaluations VALUES(23393,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(23393,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(23393,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(23393,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(23393,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(23393,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(23393,36,'cp 68','b1c3 e7e6 d4d5 e6d5 c3d5 d7d6 g1f3 g8f6 c2c4 h7h6 g2g3 g7g6 h2h4 f8g7 f1g2 f6e4 f3d2 e4f6 a1c1 e8g8 e1g1 f8e8 b2b4 f6d5 g2d5 g8h7 a2a3 a7a5 b4b5 c6d4 e2e3 d4e6 d1c2 d8e7 d2f3 e6c5 f1d1 c5e4');
INSERT INTO evaluations VALUES(23991,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(23991,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(23991,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(23991,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(23991,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(23991,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(23991,36,'cp 35','f1e1 e8g8 h2h3 f8e8 c1f4 c6c5 d4f3 d8b8 b2b3 b8b4 d1d3 d7c6 a1d1 a8d8 f3d2 f6h5 f4e3 h5f6 c3d5 f6d5 e4d5 c6d7 d2c4 e7f6 e3d2 b4b8 d3f3 f6d4 a2a4 e8e1 d1e1 d8e8 e1e8 b8e8 g2g4 h7h6 g1g2 e8d8 c4a5 d4f6 a5c6 d7c6 d5c6 d8e7');
INSERT INTO evaluations VALUES(33592,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(33592,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(33592,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(33592,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(33592,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(33592,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(33592,36,'cp 39','a5b6 e4f6 e7f6 g1f3 c8e6 f1d3 c6c5 c2c3 f8d6 e1g1 b8c6 d3e4 c5d4 e4c6 b6c6 f3d4 c6a6 f1e1 e8g8 d4e6 f7e6 d1b3 f8e8 d2e3 a6c6 e1d1 g8f7 a2a4 a7a6 b3b6 d6e5 d1d3 a8c8 a1d1 c6b6 e3b6 c8c4 a4a5 e5c7 d1d2 e6e5 d3d7 e8e7 d2d1 c7b6 a5b6');
INSERT INTO evaluations VALUES(36375,30,'cp 164',NULL);
INSERT INTO evaluations VALUES(36375,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(36375,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(36375,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(36375,34,'cp 154',NULL);
INSERT INTO evaluations VALUES(36375,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(36375,36,'cp 153','d1e2 c7c5 d4c5 e7c5 e1g1 b8c6 e4e5 f6d5 f1d1 c6e7 c4d3 h7h6 e2e4 e7g6 c3a4 d8e7 h2h4 f8d8 a4c5 e7c5 e4g4 g8h8 d3g6 f7g6 g4g6 c5f8 c1e3 c8d7 a1c1 d8c8 f3d4 d5e7 g6g4 c8c1 d1c1 a8c8 c1d1 f8f7 h4h5 e7d5 e3d2');
INSERT INTO evaluations VALUES(36395,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(36395,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(36395,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(36395,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(36395,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(36395,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(36395,36,'cp 13','e1g1 b8c6 b1d2 h7h6 g5h4 f6h5 h4e7 c6e7 g2g3 h5f6 c2c3 c8f5 f3e5 f8e8 f1e1 f5d3 e5d3 e7f5 e1e8 d8e8 d2f3 c7c6 d1c2 e8d8 g1g2 d8c7 a2a4 f5d6 f3e5 d6c4 a1e1 a8e8 e1e2 b7b6 h2h3 c6c5 e2e1 c5d4 c3d4');
INSERT INTO evaluations VALUES(38099,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(38099,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(38099,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(38099,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(38099,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(38099,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(38099,36,'cp 39','f4g3 g8f6 b1d2 c6c5 d4c5 d6c5 c2c4 e8g8 a2a3 c5e7 a1c1 f6h5 g3e5 d5c4 f1c4 b8d7 e1g1 d7e5 f3e5 h5f6 d2f3 d8b6 d1b3 c8d7 b3b6 a7b6 e5d7 f6d7 c4b5 d7c5 g2g3 e7f6 c1c2 f8c8 f1c1 c8d8 b5e2 c5b3 c1d1 d8d1 e2d1 a8d8');
INSERT INTO evaluations VALUES(38100,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(38100,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(38100,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(38100,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(38100,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(38100,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(38100,36,'cp 43','g8f6 b1d2 c6c5 d4c5 d6c5 c2c4 e8g8 a2a3 c5e7 a1c1 f6h5 g3e5 d5c4 d2c4 h5f6 d1d8 f8d8 c4a5 b7b6 e5b8 b6a5 b8c7 d8d5 f1e2 c8b7 c7e5 d5c5 c1c5 e7c5 e1g1 c5b6 f1c1 a8c8 c1c8 b7c8 f3d2 c8b7 g2g3 f6e4 e2f3 e4d2 f3b7 d2c4');
INSERT INTO evaluations VALUES(40558,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(40558,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(40558,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(40558,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(40558,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(40558,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(40558,36,'cp 43','b1d2 c6c5 c2c3 d8c7 d4c5 c7c5 f1d3 c5b6 d1c2 e8g8 c3c4 d6g3 h2g3 h7h6 e1g1 b8c6 a2a3 a7a5 f1c1 c8d7 b2b3 b6c5 c2b2 c5e7 a1b1 f8d8 b3b4 a5b4 a3b4 a8a4 b4b5 c6b4 d3f1 b4a2 c1c2 a2b4 c4d5 f6d5 c2c1 d7e8');
INSERT INTO evaluations VALUES(39358,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(39358,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(39358,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(39358,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(39358,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(39358,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(39358,36,'cp -20','b4c3 d2c3 d7d6 e3e4 g8e7 f1d3 e8g8 g1e2 f7f5 f2f4 b8d7 e4f5 d7c5 e2g3 c5d3 d1d3 c8f5 g3f5 e7f5 e1g1 c7c6 f4e5 d6e5 d3d8 a8d8 c1g5 d8e8 f1e1 a7a5 a1d1 e5e4 d1d7 a5a4 d7b7 e4e3 c4c5 h7h6 g5f4 e8e4 e1f1');
INSERT INTO evaluations VALUES(39379,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(39379,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39379,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(39379,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(39379,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39379,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39379,36,'cp 0','b2c3 d7d6 e3e4 f7f5 e4f5 c8f5 d2d3 g8f6 g1e2 f5g6 e2g3 b8d7 f1e2 e8g8 e1g1 h7h6 e2f3 a8b8 c1e3 a7a5 e3a7 b8a8 a7e3');
INSERT INTO evaluations VALUES(39670,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(39670,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(39670,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(39670,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(39670,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(39670,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(39670,36,'cp 37','c1g5 b8d7 a2a4 h7h6 g5e3 d8c7 f3d2 d7c5 e1g1 c8e6 c3d5 e6d5 e4d5 e8g8 a4a5 c5d7 d1f3 b7b5 a5b6 d7b6 b2b3 b6c4 d2c4 f8b8 e3d2 b8b5 c4e3 c7b7 a1a4 e7d8 f1d1 a8c8 a4c4 d8b6 c4c6 b6e3 d2e3 b5b4');
INSERT INTO evaluations VALUES(39663,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(39663,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(39663,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(39663,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(39663,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(39663,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(39663,36,'cp 37','b8d7 a2a4 h7h6 g5e3 d7b6 f3d2 c8e6 c4b3 e8g8 e1g1 a8c8 d1e2 b6d7 f1d1 d7c5 e3c5 c8c5 e2d3 c5c6 d2c4 e6g4 f2f3 g4e6 c4e3 c6c5 g1h1 d8b6 e3f5 e6f5 e4f5 d6d5 c3d5 f6d5 b3d5 b6b2 c2c4 c5c7 h2h3 e7g5 a1b1 b2a2 b1b7 c7b7 d5b7 a2a4');
INSERT INTO evaluations VALUES(39744,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(39744,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(39744,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(39744,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(39744,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(39744,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(39744,36,'cp 20','g2g4 d6d5 e4d5 f6d5 c3d5 d8d5 e1g1 c8d7 c2c4 d5d6 d3e4 b8c6 d4c6 d7c6 d1d6 e7d6 e4c6 b7c6 a1d1 e8c8 g1g2 f7f5 e3g5 d8d7 h2h3 e6e5 d1d2 d6b4 d2d7 c8d7 g4f5 h8f8 f1d1 d7c7 g2g3 f8f5 g5e3 g7g6 b2b3 e5e4 d1d4');
INSERT INTO evaluations VALUES(40361,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(40361,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40361,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40361,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(40361,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40361,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40361,36,'cp 0','e4e5 f6d5 e5d6 d8d6 e1g1 e8g8 f1e1 b8c6 b1d2 d6c7 d2b3 f8d8 b3c5 c6b4 a2a3 c7c5 a3b4 d5b4 e1e2 c8f5 c1e3 c5c7 e3f4 c7b6 f4e3');
INSERT INTO evaluations VALUES(43995,30,'cp -117',NULL);
INSERT INTO evaluations VALUES(43995,31,'cp -123',NULL);
INSERT INTO evaluations VALUES(43995,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(43995,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(43995,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(43995,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(43995,36,'cp -75','c5d4 c3b5 b8c6 c1f4 f8c5 e1g1 a7a6 b5d6 c5d6 e5d6 e8g8 f1e1 h7h6 c2c4 d4c3 b2c3 e6e5 f3e5 d7e5 f4e5 c6e5 e1e5 d8d6 e5d5 d6c7 d5d4 c7c3 a1c1 c3a3 d3c4 a3e7 h2h3 b7b5 c4b3 c8f5 c1c6 a6a5 g2g4 f5e4 c6b6 a8d8 b6b5 e4c6 b5a5 d8d4 d1d4');
INSERT INTO evaluations VALUES(46114,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(46114,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(46114,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(46114,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(46114,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(46114,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(46114,36,'cp 10','g1f3 a7a6 a2a4 b8c6 e1g1 f8e7 d4c5 d8d1 f1d1 e7c5 c1d2 e8e7 a4a5 c5b4 c3a4 c8d7 a4b6 a8b8 f3d4 b4d2 b6d7 f6d7 d4c6 b7c6 d1d2 d7c5 d2c2 h8d8 c4e2 c5b3 a1a3 b8b4 f2f3 d8d5 c2c6 b3a5 c6a6 b4b2 a6a7 e7f6 a3a5');
INSERT INTO evaluations VALUES(46694,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(46694,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(46694,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(46694,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(46694,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(46694,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(46694,36,'cp 16','f8g7 e3e4 e8g8 e1g1 c7c5 c2c3 c5d4 c3d4 b8c6 h2h3 e7e5 d4d5 c6b4 c1g5 h7h6 g5f6 d8f6 d3b5 h6h5 b1d2 a7a6 b5e2 a6a5 a2a3 b4a6 e2a6 a8a6 a1c1 f6d8 d1e2 g7h6 c1c2 a5a4 f1c1 c8d7 c2c7 b7b5 c1c3 a6a8 c7b7');
INSERT INTO evaluations VALUES(46825,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(46825,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(46825,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(46825,33,'cp 132',NULL);
INSERT INTO evaluations VALUES(46825,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(46825,35,'cp 166',NULL);
INSERT INTO evaluations VALUES(46825,36,'cp 160','e2e4 c7c5 d4d5 f8g7 f1c4 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 f6e8 f3e2 b8d7 f2f4 e8d6 c4d3 g7d4 g1h1 f7f6 c1d2 a8c8 a1d1 a7a6 e2f3 g8h8 b2b3 g6g5 c3e2 g5f4 e2d4 d7e5 f3e2 c5d4 d2f4 f8g8 d3b1 d8b6 e2f2 d6b5 a2a4 b5d6 f2d4 b6d4');
INSERT INTO evaluations VALUES(46826,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(46826,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(46826,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(46826,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(46826,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(46826,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(46826,36,'cp 140','c7c5 d4d5 f8g7 f1c4 e8g8 e1g1 c8g4 h2h3 g4f3 d1f3 f6e8 c4d3 b8d7 f3e2 a7a6 a2a4 e7e5 d5e6 f7e6 d3c4 d8e7 c1e3 g7d4 a1d1 d4e3 e2e3 d7e5 c4a2 e5c6 a4a5 c6a5 c3a4 c5c4 a4b6 a8d8 d1d8 e7d8 b6c4 a5c4 a2c4 e8g7 e3b3 d8d4 c4e6 g7e6 b3e6 f8f7');
INSERT INTO evaluations VALUES(47385,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(47385,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(47385,32,'cp -113',NULL);
INSERT INTO evaluations VALUES(47385,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(47385,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(47385,35,'cp -115',NULL);
INSERT INTO evaluations VALUES(47385,36,'cp -117','c8f5 f1c4 c7c6 g1e2 e7e6 e2g3 b8d7 e1g1 f8b4 d1e2 d7b6 c4b3 a7a5 a2a4 f5g6 f4e5 b6d7 e5f4 f6d5 f4d2 d7f6 f1e1 b4c3 b2c3 h7h5 h2h3 h5h4 b3d5 e6d5 g3f1 g6h5 e2e3 f6d7 c3c4 e8g8 c4d5 c6d5 e3f4 d7b6 f1e3');
INSERT INTO evaluations VALUES(47589,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(47589,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(47589,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(47589,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(47589,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(47589,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(47589,36,'cp 68','f8g7 e2e3 e8g8 a1c1 h7h6 g5h4 b8c6 c4d5 e6d5 h2h3 c6e7 f1d3 c7c6 g2g4 f6e8 d1c2 e8d6 f3e5 g6g5 h4g3 f7f6 d3h7 g8h8 e5g6 e7g6 h7g6 f6f5 g3d6 d8d6 g4f5 d6f6 h3h4 g5g4 c3e2 c8f5 g6f5 f6f5 c2f5 f8f5 h1g1 h6h5 e2g3');
INSERT INTO evaluations VALUES(49630,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(49630,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(49630,32,'cp -102',NULL);
INSERT INTO evaluations VALUES(49630,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(49630,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(49630,35,'cp -106',NULL);
INSERT INTO evaluations VALUES(49630,36,'cp -77','d7d5 d2d3 c7c5 b1d2 b8c6 g1f3 h7h6 e4d5 f6d5 d2c4 f8d6 g2g3 e8g8 f1g2 c8f5 e1g1 d8d7 c2b3 d5b6 c4d6 d7d6 f3e1 f5e6 b3b5 e6d5 c1e3 b6d7 g2d5 d6d5 a1d1 a7a6 b5b3 d5b3 a2b3 f8e8 f2f4 a8d8 e1f3 e5f4');
INSERT INTO evaluations VALUES(51305,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(51305,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(51305,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(51305,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(51305,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(51305,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(51305,36,'cp 75','e7e6 b1d2 c7c5 c2c3 d8d7 h2h4 h7h5 d2f3 c8a6 f1a6 b8a6 g1e2 g8e7 e1g1 e7g6 e2f4 g6f4 c1f4 a8c8 d1e2 a6b8 f4g5 b8c6 a1d1 c5c4 g2g3 b6b5 b2b4 c6e7 d1a1 e7f5 f1b1 f8e7 g5e7 f5e7 g1g2 e7f5');
INSERT INTO evaluations VALUES(52510,30,'cp 223',NULL);
INSERT INTO evaluations VALUES(52510,31,'cp 234',NULL);
INSERT INTO evaluations VALUES(52510,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(52510,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(52510,34,'cp 246',NULL);
INSERT INTO evaluations VALUES(52510,35,'cp 233',NULL);
INSERT INTO evaluations VALUES(52510,36,'cp 240','e2e4 e7e6 f1c4 d7d5 e4d5 e6d5 c4b3 g8f6 g1f3 h7h6 c1f4 f8d6 f4e5 e8g8 e1g1 c8e6 f1e1 b8c6 b1d2 b7b5 a2a3 c6a5 b3c2 a5c4 d1c1 f8e8 e5d6 c4d6 a3a4 b5a4 c2a4 e6d7 a4d7 e8e1 c1e1 d8d7 f3e5 d7c7 h2h3 a7a5');
INSERT INTO evaluations VALUES(1310,30,'cp -84',NULL);
INSERT INTO evaluations VALUES(1310,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(1310,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(1310,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(1310,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(1310,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(1310,36,'cp -87','c5d4 f3d4 g8e7 c2c4 c6c5 d4f3 c8b7 c4d5 e7d5 b1a3 b6a6 f3g5 d5e7 d1b3 e7g6 b3b5 a6b5 a3b5 h7h6 g5h3 a8d8 b5d6 f8d6 e5d6 g6h4 f2f3 h4f5 f1d1 e6e5 d1e1 f7f6 c1e3 f5e3 e1e3 d8d6 e3a3 d6a6 a3c3 e8e7 c3c5 h8c8 c5c8 b7c8 h3f2');
INSERT INTO evaluations VALUES(1315,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(1315,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(1315,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(1315,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(1315,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(1315,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(1315,36,'cp -83','f3d4 c6c5 d4f3 g8e7 c2c4 c8b7 c4d5 b7d5 b1c3 d5c6 d1e2 e7f5 b2b3 h7h6 f3d2 f5d4 e2h5 b6b7 f2f3 f8e7 d2c4 d4f5 c1b2 a8d8 c3e4 c6e4 f3e4 b7e4 a1e1 e4h4 h5h4 f5h4 c4d6 e7d6 e5d6 d8d6 b2a3 d6a6 a3c5 a6a2 e1d1 a2g2 g1h1');
INSERT INTO evaluations VALUES(1312,30,'cp -83',NULL);
INSERT INTO evaluations VALUES(1312,31,'cp -99',NULL);
INSERT INTO evaluations VALUES(1312,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(1312,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(1312,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(1312,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(1312,36,'cp -104','c6c5 d4f3 g8e7 c2c4 c8b7 c4d5 e7d5 b1a3 b6a6 f3g5 a8d8 d1h5 d8d7 g5h7 a6c6 f2f3 c5c4 h5h4 b7a6 a1b1 d5e7 h7f6 g7f6 h4h8 e7g6 h8g8 c6c5 g1h1 c4c3 f1g1 c3c2 a3c2 c5c2 c1h6 f6e5 b1c1 c2f5 h6f8 g6f8 g1d1 d7d1 c1d1 f5c2 d1d8 e8d8 g8f8 d8c7 f8f7 c7d6');
INSERT INTO evaluations VALUES(2068,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(2068,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(2068,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(2068,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(2068,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(2068,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(2068,36,'cp 56','b1c3 g8f6 c1d2 e7c5 e3e2 f8e7 g1f3 d7d5 e4e5 f6e4 c3d5 c5d5 c2c4 d5e6 e2e4 e6g6 e4g6 h7g6 d2c3 g6g5 e1c1 g5g4 f3d4 c6d4 c3d4 c8e6 f1e2 e8c8 d4e3 d8d1 e2d1 g4g3 f2g3 e6c4 a2a3 c4e6 d1f3 h8d8 h1d1');
INSERT INTO evaluations VALUES(2066,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(2066,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(2066,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(2066,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(2066,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(2066,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(2066,36,'cp 62','g8f6 c1d2 e7c5 e3e2 f8e7 g1f3 d7d6 e1c1 c8e6 d2e3 c5a5 a2a3 d6d5 e3d2 d5e4 c3e4 a5d5 e4c3 d5f5 f3g5 e8c8 g5e6 f7e6 e2e1 a7a6 f2f3 c8b8 c1b1 e6e5 d2c1 f5g6 d1d2 d8d2 e1d2 e5e4 f3e4 f6e4 c3e4 g6e4 f1d3');
INSERT INTO evaluations VALUES(1112,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(1112,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(1112,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(1112,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(1112,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(1112,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(1112,36,'cp 70','c1d2 e7c5 e3e2 f8e7 g1f3 d7d6 h2h3 c8e6 a2a3 f6d7 e1c1 d7b6 d2e3 c5a5 e2d3 b6c4 c3d5 c4e3 d3e3 e8g8 d5e7 c6e7 f3d4 e6d7 f1e2 a5b6 h1e1 f8e8 e2g4 a8d8 g4d7 d8d7 f2f4 h7h6 e3c3 d6d5 d4b3 b6c6 c3c6 e7c6 e4e5 d7d8 b3c5');
INSERT INTO evaluations VALUES(4943,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(4943,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(4943,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(4943,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(4943,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(4943,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(4943,36,'cp 46','b1c3 g8f6 b2b3 e8g8 e1g1 a7a5 h2h3 f8e8 c1e3 e7f8 f1e1 c8b7 d1f3 d6d5 e4d5 c6d5 a1d1 c7c6 c3a4 f8b4 c2c3 b4d6 c4d3 f6d7 e3f4 d8f6 f4d6 f6d6 c3c4 h7h6 d3f5 d7f6 f3c3 d6b4 c3b4 a5b4 e1e8 f6e8 c4d5 c6d5 f5d3 e8d6');
INSERT INTO evaluations VALUES(5034,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(5034,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(5034,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(5034,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(5034,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(5034,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(5034,36,'cp -25','g8f6 b1c3 e8g8 e1g1 d7d6 a1b1 a7a5 b2b3 f6d7 f3g3 d7b6 c4e2 f7f5 c1b2 f5f4 g3f3 b6d7 c3a4 d7e5 b2e5 g7e5 a4b2 d8c7 f1d1 c7a7 b2d3 e5d4 b1c1 c8d7 h2h3 e7e5 c2c3 d4c5 d3c5 a7c5 a2a4 d7e6');
INSERT INTO evaluations VALUES(5163,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5163,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5163,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5163,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(5163,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5163,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5163,36,'cp 0','b7c6 e1g1 d7d5 c4d3 g8f6 b1c3 e8g8 c1f4 c5e7 f1e1 a7a5 e4e5 f6d7 c3a4 c8a6 d3h7 g8h7 d1h5 h7g8 e1e3 f7f6 e3h3 f6e5 h5h7 g8f7 h7h5');
INSERT INTO evaluations VALUES(5011,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5011,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5011,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5011,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5011,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5011,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5011,36,'cp 0','e1g1 d7d5 c4d3 g8f6 b1c3 e8g8 c1f4 c5e7 f1e1 a7a5 e4d5 c6d5 a2a4 e7b4 f4d2 b4c5 d1f3 c8b7 a1d1 d8b6 c3b5 b7c6 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(5010,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5010,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5010,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5010,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5010,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(5010,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(5010,36,'cp -13','d7d5 c4d3 g8f6 b1c3 e8g8 c1f4 c5e7 f1e1 f6d7 d1h5 d5d4 c3a4 e6e5 f4d2 a7a5 b2b3 c8a6 h5e2 d8c8 d3a6 a8a6 a4b2 e7b4 a2a4 b4d2 e2d2 c8c7 f2f3 f8a8 d2f2 d7c5 b2c4 f7f6 a1d1 c5e6 h2h3');
INSERT INTO evaluations VALUES(6420,30,'cp -165',NULL);
INSERT INTO evaluations VALUES(6420,31,'cp -152',NULL);
INSERT INTO evaluations VALUES(6420,32,'cp -168',NULL);
INSERT INTO evaluations VALUES(6420,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(6420,34,'cp -173',NULL);
INSERT INTO evaluations VALUES(6420,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(6420,36,'cp -172','d8f6 a2a4 d7d6 a4a5 f6g6 e1g1 g8f6 d2d4 e5d4 c3d4 c5d4 f3d4 c6d4 b3d3 d4e6 b1c3 e6c5 d3e2 e8g8 f2f3 c8e6 c1e3 f6d7 c3d5 e6d5 c4d5 a8c8 g1h1 h7h6 a1c1 a7a6 e2d2 c8e8 f1d1 e8e7 e3d4 c7c6 d5a2 f8e8 a2b1');
INSERT INTO evaluations VALUES(8765,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(8765,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(8765,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(8765,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(8765,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(8765,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(8765,36,'cp -44','a7a6 c2c3 c8d7 b2b4 c5a7 b1d2 f6h5 f3h2 h5f4 g3f4 g5f4 d1h5 d8e7 a2a4 h8h7 b4b5 c6a5 b5a6 e8c8 a6b7 c8b8 c4b5 c7c6 b5a6 h7g7 h2f3 g7g2 h1f1 f7f6 a1b1 d6d5 h5h6 d7e6 e1e2 g2g7 f1g1 g7h7 h6g6 e6h3 e2e1 h3d7 e4d5 c6d5 b1b5 d7b5 a6b5 e5e4 d3e4');
INSERT INTO evaluations VALUES(8780,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(8780,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(8780,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(8780,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(8780,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(8780,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8780,36,'cp 2','a2a4 e8g8 e3b6 a7b6 h2h3 c8e6 c4b5 c6d4 f3d4 e5d4 c3e2 c7c6 b5c4 d6d5 e4d5 e6d5 b2b3 f8e8 e2g3 d8d7 f1e1 c6c5 d1d2 d5c6 d2f4 f6d5 c4d5 d7d5 f2f3 e8e6 f4f5 a8e8 e1e6 d5e6 f5e6 e8e6 a4a5 b6a5');
INSERT INTO evaluations VALUES(9515,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(9515,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(9515,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(9515,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(9515,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(9515,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(9515,36,'cp -20','d6b4 d1d3 e5d4 f3d4 b4c3 b2c3 c6c5 d4e2 d8d3 c2d3 c5c4 e4e5 c4d3 e5f6 d3e2 f6g7 h8g8 c1h6 a7a5 e1e2 a8a6 h6f4 a6c6 e2d2 g8g7 g2g3 f7f6 h1e1 e8d8 d2c2 c8g4 c2b2 g7d7 h2h4 b7b6 a1c1 c6c4 e1e3 d7d2 c1c2 d2d5 f2f3 g4f5 c2e2 d8c8 e3e8 c8b7');
INSERT INTO evaluations VALUES(10093,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10093,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10093,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10093,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10093,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10093,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(10093,36,'cp 0','d2d3 e8g8 a2a3 c6d4 g1f3 c7c6 c4a2 d7d6 c3a4 d8a5 a4c3');
INSERT INTO evaluations VALUES(10651,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(10651,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(10651,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(10651,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(10651,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(10651,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(10651,36,'cp 47','d8e7 b1c3 c6e5 f3e5 g4e5 d5d2 c7c6 f1e2 d7d6 e1g1 a7a6 c3a4 c5a7 b2b3 c8f5 c1a3 a8d8 a1d1 b7b5 a4b2 e8g8 b2d3 f5d3 e2d3 a7c5 d2a5 b5c4 b3c4 c5a3 a5a3 e7e6 d3e2 e5c4 a3a6 c4e5 d1c1 g7g6 a2a4 d8b8 a4a5 f8e8');
INSERT INTO evaluations VALUES(587,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(587,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(587,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(587,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(587,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(587,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(587,36,'cp 36','f1e2 g4e5 f3e5 c6e5 e1g1 c7c6 d5d2 d7d6 b1c3 a7a6 b2b3 c8f5 c3a4 e8g8 a4c5 d6c5 c1b2 f8d8 d2c3 f7f6 a1d1 b7b5 h2h3 b5b4 c3c1 f5d3 e2d3 e5d3 c1c2 d3b2 c2b2 d8d7 b2e2 a8d8 d1d7 e7d7 e2h5 d7d6 h5f5');
INSERT INTO evaluations VALUES(11692,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(11692,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(11692,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(11692,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(11692,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(11692,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(11692,36,'cp 124','c4d5 b5b4 c3a4 f6d5 e4d5 c6e7 f3e5 d7d6 e5c6 e7c6 d5c6 d8f6 a2a3 a6a5 d2d4 f8e7 d4d5 e8g8 c2c3 f6g6 a3b4 a5b4 c3b4 c8f5 a1a3 f5c2 d1e2 e7h4 e2c4 a8a4 c4h4 a4a3 b2a3 c2e4 f2f3 e4d5 h4e7 d5c6 e7c7 c6b5 f1e1');
INSERT INTO evaluations VALUES(12491,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(12491,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(12491,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(12491,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(12491,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(12491,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(12491,36,'cp 24','d7d5 f1d3 f6h5 f4e3 f8d6 b1d2 h5f4 e3f4 d6f4 e1g1 e8g8 g2g3 f4d6 a2a3 c8d7 d1e2 a8c8 f1e1 g7g6 a1d1 d8b6 d2f1 c6e7 h2h4 h7h5 f1e3 d7a4 d1a1 b6c7 e3g2 g8g7 a1c1 e7g8 e2d2 g8f6 f3e5');
INSERT INTO evaluations VALUES(12695,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(12695,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(12695,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(12695,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(12695,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(12695,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(12695,36,'cp 39','d7d5 e4d5 e7d5 c4d5 e6d5 c1e3 f8e7 d1f3 e7f6 e1c1 c8e6 d4e6 f7e6 f3h5 g7g6 h5h3 d8e7 e3h6 f6g5 c1b1 g5h6 h3h6 e8c8 h1e1 d5d4 c3e4 c8b8 h6d2 h8e8 f2f3 a7a6 e4f2 e7c7 h2h4 e6e5 f2d3 c6a5 b2b3 a5c6 e1e4 b8a7 d1e1 e8e7 a2a3 a7a8 d2e2 d8e8');
INSERT INTO evaluations VALUES(15369,30,'cp -451',NULL);
INSERT INTO evaluations VALUES(15369,31,'cp -457',NULL);
INSERT INTO evaluations VALUES(15369,32,'cp -450',NULL);
INSERT INTO evaluations VALUES(15369,33,'cp -431',NULL);
INSERT INTO evaluations VALUES(15369,34,'cp -433',NULL);
INSERT INTO evaluations VALUES(15369,35,'cp -436',NULL);
INSERT INTO evaluations VALUES(15369,36,'cp -461','c6e5 c4b5 c7c6 e4c3 c6b5 d1e2 d8e7 c3d5 e7d6 d5c3 b5b4 c3b5 d6e7 a2a3 a7a6 a3b4 a8b8 b5d4 e5g6 c2c3 e7e2 d4e2 f8d6 d2d3 f7f6 c1e3 c8d7 e1g1 e8f7 e2d4 b8d8 d4b3 d6c7 d3d4 h8e8 b3c5 d7c8 c3c4 g6e7 d4d5 e7f5 e3d2 f5d4 a1e1 h7h5 e1e8 d8e8');
INSERT INTO evaluations VALUES(15852,30,'cp -459',NULL);
INSERT INTO evaluations VALUES(15852,31,'cp -433',NULL);
INSERT INTO evaluations VALUES(15852,32,'cp -457',NULL);
INSERT INTO evaluations VALUES(15852,33,'cp -455',NULL);
INSERT INTO evaluations VALUES(15852,34,'cp -444',NULL);
INSERT INTO evaluations VALUES(15852,35,'cp -451',NULL);
INSERT INTO evaluations VALUES(15852,36,'cp -432','c4b5 c7c6 e4c3 c6b5 d1e2 d8e7 c3d5 e7d6 d5c3 b5b4 c3b5 d6e7 a2a3 a7a6 a3b4 a8b8 b5d4 e5g6 c2c3 e7e2 d4e2 f8d6 d2d3 c8d7 c1e3 f7f6 e1g1 e8f7 e2d4 d6c7 d4b3 b8d8 d3d4 d7c8 f1d1 g6e7 c3c4 b7b5 d1c1 b5c4 c1c4 c7d6 b3c5 d6c5 b4c5 h8e8 c4b4 e7d5 e3d2 d5b4 d2b4');
INSERT INTO evaluations VALUES(15409,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(15409,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(15409,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(15409,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(15409,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(15409,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(15409,36,'cp 35','b2c3 f8c5 c1b2 e8g8 d2d4 c5d6 e1g1 f8e8 e2e3 c8g4 h2h3 g4h5 d1a4 d8d7 f3d2 h5g6 d2c4 a8d8 f1d1 e5e4 c4d6 c7d6 c3c4 g6h5 g3g4 h5g6 d4d5 c6e5 a4d7 d8d7 b2e5 e8e5 g2f1 f7f5 f1e2 b7b6 g1g2 d7c7 d1c1 g6f7');
INSERT INTO evaluations VALUES(15492,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(15492,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(15492,32,'cp -59',NULL);
INSERT INTO evaluations VALUES(15492,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(15492,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(15492,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(15492,36,'cp -62','c8g4 f3g3 d8d7 h2h3 g4e6 c4e6 f7e6 g1f3 e8c8 b2b4 h7h6 c1e3 g7g5 b1d2 f8e7 d2b3 f6h5 g3h2 h5f4 e3f4 e5f4 d3d4 h6h5 b4b5 c6b8 a2a4 a7a6 c3c4 g5g4 f3d2 g4g3 f2g3 f4g3 h2g3 h8g8 g3f3 d8f8 f3e2 d6d5 c4d5 e6d5 e4e5');
INSERT INTO evaluations VALUES(25310,30,'cp -64',NULL);
INSERT INTO evaluations VALUES(25310,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(25310,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(25310,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(25310,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(25310,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(25310,36,'cp -45','f3g3 d8d7 h2h3 g4e6 c4e6 d7e6 c1g5 d6d5 b1d2 e8c8 g5f6 g7f6 g3g4 h7h5 g4e6 f7e6 g1f3 f8h6 g2g3 c6e7 a2a4 c7c5 a4a5 h8f8 e1e2 c8b8 h1b1 d8c8 b1h1 f8g8 h1g1 a7a6 g1e1 g8d8 e1d1 e7g6 d1e1 f6f5 h3h4 h6d2');
INSERT INTO evaluations VALUES(25311,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(25311,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(25311,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(25311,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(25311,34,'cp -89',NULL);
INSERT INTO evaluations VALUES(25311,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(25311,36,'cp -85','d8d7 c4b3 f8e7 h2h3 g4e6 g3e3 e8g8 b3e6 d7e6 e3e2 d6d5 g1f3 h7h6 e1g1 f8e8 f1e1 a8d8 b2b4 e7f8 e2c2 c6e7 c1e3 b7b6 b1d2 e7g6 d2f1 c7c5 b4c5 f8c5 e3c5 b6c5 f1g3 c5c4 e4d5 c4d3 d5e6');
INSERT INTO evaluations VALUES(16091,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(16091,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(16091,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(16091,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(16091,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(16091,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(16091,36,'cp -20','b7b5 c4d5 c7c6 c2c3 c6d5 c3d4 e5d4 g1f3 f8g7 e1g1 e8g8 f3d4 d5e4 d3e4 c8b7 d4b5 b7e4 b1c3 e4c6 b5d4 h6f5 d4f5 g6f5 c1e3 d8h4 d1d6 f8e8 d6g3 h4g3 h2g3 a8b8 a1b1 a7a5 f1d1 b8b4 a2a3 b4b3 e3d4 g7d4 d1d4 g8g7 d4d2 e8b8');
INSERT INTO evaluations VALUES(17931,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(17931,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(17931,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(17931,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(17931,34,'cp 134',NULL);
INSERT INTO evaluations VALUES(17931,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(17931,36,'cp 152','f3d4 e5d4 e1g1 f8c5 d2d3 c7c6 b1d2 d7d6 f2f4 g8f6 f4f5 d8e7 d2f3 c8d7 c1g5 h7h6 g5h4 e8c8 d1e1 c5a7 e1a5 a7b8 a1e1 d8g8 a4b3 c6c5 a5d2 d7c6 c2c3 d4c3 d2c3 g8e8 b3c4 b7b5 c4b3 b8c7 h4f6 e7f6 c3f6 g7f6');
INSERT INTO evaluations VALUES(17930,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(17930,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(17930,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(17930,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(17930,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(17930,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(17930,36,'cp 145','e5d4 e1g1 f8c5 d2d3 c7c6 b1d2 g8e7 f2f4 d7d5 f4f5 a6a5 d1h5 b7b5 a4b3 a5a4 f5f6 g7f6 e4d5 e7d5 d2e4 c5e7 b3d5 d8d5 h5d5 c6d5 e4f6 e7f6 f1f6 h8g8 c1d2 h7h5 f6f4 g8g4 a1e1 c8e6 f4f2 e8d7 e1e5 a8h8 h2h3 g4g6 g1h2 d7d6 e5e2 d6c5 f2f4 b5b4');
INSERT INTO evaluations VALUES(17942,30,'cp 144',NULL);
INSERT INTO evaluations VALUES(17942,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(17942,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(17942,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(17942,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(17942,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(17942,36,'cp 120','e1g1 g8f6 c2c3 f8e7 a4c2 d7d5 e4e5 f6g4 c3d4 g4h6 b1c3 h6f5 c3e2 c7c6 b2b3 g7g6 c1b2 f5g7 h2h3 c8e6 b2c3 a6a5 a2a3 h7h5 e2f4 e6f5 c2f5 g7f5 d1c2 f5g7 a3a4 d8d7 a1c1 d7f5 c2f5 g7f5');
INSERT INTO evaluations VALUES(18297,30,'cp 851',NULL);
INSERT INTO evaluations VALUES(18297,31,'cp 855',NULL);
INSERT INTO evaluations VALUES(18297,32,'cp 863',NULL);
INSERT INTO evaluations VALUES(18297,33,'cp 867',NULL);
INSERT INTO evaluations VALUES(18297,34,'cp 867',NULL);
INSERT INTO evaluations VALUES(18297,35,'cp 871',NULL);
INSERT INTO evaluations VALUES(18297,36,'cp 875','d4d5 d8f6 d5a8 f6d8 c1f4 f8c5 a8a7 c7c6 a7b8 d7d6 b1c3 c8e6 b8d8 e8d8 c4e6 f7e6 e4e5 d6d5 c3a4 g8e7 a4c5 b6c5 a2a4 d8c7 e1e2 h8f8 f4g3 d5d4 h2h4 c5c4 a4a5 e7d5 a5a6 f8a8 e2f3 d5b4 f3e4 a8a6 a1a6 b4a6 e4d4 a6b4 d4c4 b4c2');
INSERT INTO evaluations VALUES(18296,30,'cp 873',NULL);
INSERT INTO evaluations VALUES(18296,31,'cp 861',NULL);
INSERT INTO evaluations VALUES(18296,32,'cp 865',NULL);
INSERT INTO evaluations VALUES(18296,33,'cp 875',NULL);
INSERT INTO evaluations VALUES(18296,34,'cp 875',NULL);
INSERT INTO evaluations VALUES(18296,35,'cp 861',NULL);
INSERT INTO evaluations VALUES(18296,36,'cp 871','d8f6 d5a8 f6d8 c1f4 f8c5 a8a7 d7d6 c4b5 e8f8 b1c3 g8e7 e1c1 c8e6 e4e5 d6d5 c3e4 f8g8 e4c5 b6c5 a7c5 c7c6 b5e2 e7g6 f4e3 d8a8 f2f4 e6f5 g2g4 f5e4 h1f1 a8a2 c5c6 e4g2 f1g1 g2h3 d1d5 a2a1 c1d2 a1b2 c6c3');
INSERT INTO evaluations VALUES(18467,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18467,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18467,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18467,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(18467,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(18467,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18467,36,'cp 0','g8f6 b1c3 g7g6 b2b3 f8g7 c1b2 e8g8 e1g1 f8e8 f1e1 d8a5 c3a4 a5g5 d1e2 f6h5 g2g3 h5f4 e2e3 e7e5 a1d1 f4h3 g1g2 h3f4');
INSERT INTO evaluations VALUES(18701,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(18701,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(18701,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(18701,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(18701,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(18701,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(18701,36,'cp 122','c1e3 g8f6 f1d3 d7d6 c5d6 f8d6 h2h3 b7b6 b1d2 c8b7 g1f3 e8g8 e1g1 c6e5 f3e5 d6e5 a1b1 e5d6 d1e2 d8c7 a2a4 a8c8 b2b3 f8d8 b3b4 d6h2 g1h1 h2e5 a4a5 b6a5 b4a5 e5c3 a5a6 b7c6 f1d1 h7h6 h1g1 c3e5 c2c4 e5h2 g1h1 h2f4 c4c5 f6d7 d3c4 f4e3 e2e3 d7e5 c4f1');
INSERT INTO evaluations VALUES(19337,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(19337,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(19337,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(19337,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(19337,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(19337,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(19337,36,'cp 15','e7e6 d1e2 g8f6 e4e5 d7d5 c4b3 f6d7 d2d4 a7a5 b3a4 d7b6 a4b5 c5d4 c3d4 f8b4 b1c3 c8d7 e1g1 b6c4 b5c4 d5c4 f1d1 c6e7 f3d2 e8g8 e2g4 b7b5 d2e4 g8h8 g4h5 b4c3 b2c3 d7c6 e4d6');
INSERT INTO evaluations VALUES(18652,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(18652,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(18652,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18652,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18652,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(18652,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(18652,36,'cp 12','d1e2 g8f6 c4b3 d7d5 d2d3 f8e7 e1g1 c8d7 e4e5 f6h7 c1e3 a7a5 d3d4 h7g5 f3e1 a5a4 b3d1 c5d4 c3d4 e8g8 e1d3 d8b6 b1c3 a4a3 c3a4 a3b2 a1b1 b6c7 d3c5 e7c5 a4c5 a8a2 d1b3 a2a5 e2g4 a5c5 d4c5 c7e5 b3c2');
INSERT INTO evaluations VALUES(19385,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(19385,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(19385,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(19385,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(19385,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(19385,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(19385,36,'cp -82','g8f6 b1c3 g7g6 d2d3 f8g7 g1e2 d7d6 a1b1 f6d7 c1b2 h7h5 h2h3 h5h4 c3d5 e8g8 b2g7 g8g7 d1d2 b7b6 e2c3 e7e6 d5e3 f7f5 f1e2 c8b7 e1g1 c6d4 f2f4 d8e7 c3b5 a8e8 b3b4 a7a6 b5d4 c5d4 e3c2 e6e5 e4f5 g6f5 g1h1');
INSERT INTO evaluations VALUES(21472,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(21472,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(21472,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(21472,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(21472,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(21472,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(21472,36,'cp 86','d2d4 f7f6 f1b5 c8d7 g1f3 d8e7 b5c6 d7c6 c3e2 e7f7 a2a4 g8e7 c2c3 c6d7 e2f4 e7c6 f4d3 f7h5 b2b4 a7a6 e1g1 f6e5 d3e5 c6e5 f3e5 h5d1 f1d1 f8d6 c1f4 h8f8 f4g3 g7g5 d1e1 h7h5 h2h4 g5h4 g3h4 f8f5 h4g3 d6e5 g3e5 e8c8 f2f3');
INSERT INTO evaluations VALUES(22752,30,'cp -126',NULL);
INSERT INTO evaluations VALUES(22752,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(22752,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(22752,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(22752,34,'cp -117',NULL);
INSERT INTO evaluations VALUES(22752,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(22752,36,'cp -138','c6d4 g1f3 d4f3 e2f3 d7d6 d1e2 d8e7 g2g3 h7h5 b1c3 c7c6 e1c1 c8e6 e2e3 g8f6 c1b1 e7c7 f3f4 e5f4 g3f4 f6g4 e3f3 e8c8 f4f5 e6d7 c3e4 f7f6 f1h3 d6d5 e4g3 g4e5 f3e2 f8b4 g3h5 b4d6');
INSERT INTO evaluations VALUES(23197,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(23197,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(23197,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(23197,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(23197,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(23197,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(23197,36,'cp 31','d7d5 d1e2 d5e4 d3e4 f8d6 a2a3 g8e7 b5c4 c8g4 c1e3 f6f5 b1c3 f5f4 h2h3 g4f3 e2f3 d8d7 e3d2 c6d4 f3d3 d6c5 e1c1 e8c8 f2f3 c8b8 d2e1 d7e8 d3f1 e7c8 c1b1 c8b6 h3h4 e8c6 c4d5 c6f6 d5a2 a7a6 h4h5 g7g5 h5g6 h7g6 h1h8 d8h8 e1f2');
INSERT INTO evaluations VALUES(24818,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(24818,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(24818,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(24818,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24818,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(24818,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24818,36,'cp 0','e7e6 b1d2 f8d6 e2e3 e8g8 g2g4 f5g6 f4d6 c7d6 f3h4 g6e4 h4g2 d8b6 d1b3 b6c7 b3d1');
INSERT INTO evaluations VALUES(25156,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(25156,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25156,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(25156,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25156,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25156,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(25156,36,'cp 0','e2e3 f8d6 b1d2 e8g8 g2g4 f5g6 f4d6 c7d6 f3h4 g6e4 h4g2 d8b6 d1b3 b6c7 b3d1');
INSERT INTO evaluations VALUES(29469,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(29469,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(29469,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(29469,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(29469,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(29469,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(29469,36,'cp 71','b1c3 d5c4 g1f3 c7c6 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 e3e4 h7h6 g5h3 c6c5 d4d5 e6d5 e4d5 c8h3 g2h3 e8g8 g1h1 f6e8 f1g1 g8h8 c1e3 b8d7 d1d2 e7d6 c3e4 d8h4 d2g2 d6f4 e3f2 h4h5 d5d6 d7e5 c4e2 a8d8 g1d1 e5d7 f2g3 f4g3 h2g3');
INSERT INTO evaluations VALUES(29466,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(29466,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(29466,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(29466,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(29466,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(29466,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(29466,36,'cp 70','c7c6 g1f3 d5c4 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 d1c2 e8g8 f1d1 b8d7 g1h1 h7h6 g5e4 f6e4 c3e4 a7a5 c1d2 d7f6 a1c1 f6e4 c2e4 c8d7 d2e1 e7d6 e1h4 d8e8 a2a4 a8c8 e4d3 g8h8 d3e2 c6c5 d4c5 d6c5');
INSERT INTO evaluations VALUES(28926,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(28926,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(28926,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(28926,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(28926,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(28926,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(28926,36,'cp 67','g1f3 d5c4 f3g5 e6g4 f2f3 g4c8 f1c4 e7e6 e1g1 f8e7 c3e4 f6e4 g5e4 e8g8 d1c2 b7b6 f1d1 c8b7 c1d2 b8d7 b2b4 d7f6 c4d3 f6d5 a1b1 d8c7 e4g3 h7h6 e3e4 d5f4 d3f1 e7g5 a2a4 f8d8 g1h1 f4g6 g3h5 c7e7');
INSERT INTO evaluations VALUES(31433,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(31433,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31433,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31433,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31433,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(31433,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31433,36,'cp 0','e6d5 c1f4 g8f6 e2e3 b8d7 f1e2 d8b6 d1c1 h7h6 e1g1 f8e7 f3e1 e8g8 e2d3 f5d3 e1d3 f6h5 f4e5 h5f6');
INSERT INTO evaluations VALUES(38338,30,'cp -158',NULL);
INSERT INTO evaluations VALUES(38338,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(38338,32,'cp -144',NULL);
INSERT INTO evaluations VALUES(38338,33,'cp -142',NULL);
INSERT INTO evaluations VALUES(38338,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(38338,35,'cp -136',NULL);
INSERT INTO evaluations VALUES(38338,36,'cp -133','d6e5 f4e5 g7e5 d4e5 d8d1 e1d1 b8c6 f2f4 f7f6 g1f3 c8g4 f1e2 e8c8 d1c1 f6e5 f4e5 g4f3 e2f3 c6e5 b1d2 e5f3 d2f3 g8f6 h1e1 h8e8 f3e5 d8d5 a2a4 f6d7 e5d7 c8d7 a1a3 e7e5 a3f3 d7e6 e1f1 e8e7 f3f6 e6d7 f6f7 d7d8 b2b3 a7a6 c2c4 d5d7 f7e7 d8e7');
INSERT INTO evaluations VALUES(38487,30,'cp -123',NULL);
INSERT INTO evaluations VALUES(38487,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(38487,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(38487,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(38487,34,'cp -125',NULL);
INSERT INTO evaluations VALUES(38487,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(38487,36,'cp -141','f4e5 g7e5 d4e5 d8d1 e1d1 b8c6 f2f4 f7f6 g1f3 c8g4 f1e2 e8c8 d1c1 f6e5 f4e5 g4f3 e2f3 c6e5 f3d1 g8f6 h1e1 f6d7 a2a4 d8f8 a1a3 f8f2 a3g3 h8f8 b1d2 f8f4 b2b3 c8d8 d2f3 e5f3 d1f3 c7c6 g3h3 d7f6 a4a5 a7a6 h3h6 f4d4 e1e2 f2e2 f3e2 d8c7 e2c4');
INSERT INTO evaluations VALUES(38483,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(38483,31,'cp -121',NULL);
INSERT INTO evaluations VALUES(38483,32,'cp -126',NULL);
INSERT INTO evaluations VALUES(38483,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(38483,34,'cp -126',NULL);
INSERT INTO evaluations VALUES(38483,35,'cp -135',NULL);
INSERT INTO evaluations VALUES(38483,36,'cp -135','g7e5 d4e5 d8d1 e1d1 b8c6 f2f4 f7f6 g1f3 c8g4 f1e2 e8c8 d1c1 f6e5 f4e5 g4f3 e2f3 c6e5 b1d2 g8f6 h1e1 e5f3 d2f3 h8e8 f3e5 d8d6 b2b3 f6d7 e5c4 d6f6 e1e2 f6f1 c1b2 f1a1 b2a1 e7e5 a1b2 e8e6 c4d2 d7f6 c2c4 b7b6 b2c3 c8d7 d2f3 e5e4 f3g5 e6e7 c3d4');
INSERT INTO evaluations VALUES(39427,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39427,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39427,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(39427,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(39427,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(39427,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39427,36,'cp 0','b2c3 d7d6 g1f3 b8c6 f3g5 c6e5 c4b3 h7h6 f2f4 h6g5 f4e5 d8e7 d1d5 e7e5 d5f7 e8d8 f7f8 d8d7 b3g8 e5e4 e1f2 d7c6 g8d5 e4d5 f8h8 d5f5 f2e1 f5e5 e1f2');
INSERT INTO evaluations VALUES(39425,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39425,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39425,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39425,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39425,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39425,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39425,36,'cp 0','b8c6 g1f3 d7d6 f3g5 c6e5 c4b3 h7h6 f2f4 h6g5 f4e5 d8e7 d1d5 e7e5 d5f7 e8d8 f7f8 d8d7 b3g8 e5e4 e1f2 d7c6 g8d5 e4d5 f8h8 d5f5 f2e1 f5e4 e1f2');
INSERT INTO evaluations VALUES(39848,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(39848,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(39848,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39848,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39848,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(39848,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(39848,36,'cp 6','e2e4 e8g8 e4e5 f6e8 g1f3 d7d6 c4b5 a7a6 a2a4 c8g4 b1d2 d6e5 f1c4 g7f6 e1g1 e5e4 b2f6 e8f6 h2h3 a6b5 c4b5 g4f5 f3e5 f6d5 d2c4 d5b4 d1d8 f8d8 c4b6 a8a7 b5c4 d8d6 c4f7 g8g7 b6c4 d6d4 f7e8 b4d3 e8b5 f5e6 e5d3 e4d3 f1d1 b8a6 a4a5 a6b4 d1d2 e6f5');
INSERT INTO evaluations VALUES(42230,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(42230,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(42230,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(42230,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(42230,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(42230,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(42230,36,'cp -12','e8g8 f1d3 c7c5 g1e2 c5c4 d3c2 b8c6 e1g1 f8e8 e2f4 b4a5 h2h3 a7a6 a2a3 c6e7 f4h5 f6h5 d1h5 g7g6 h5h4 c8f5 c2f5 e7f5 h4d8 e8d8 f1b1 a5c3 d2c3 b7b5 a3a4 g8g7 g2g4 f5d6 c3b4 d6e4 f2f3');
INSERT INTO evaluations VALUES(37106,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(37106,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37106,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37106,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37106,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(37106,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37106,36,'cp -12','d1c2 f8e8 a2a3 b4c3 d2c3 f6e4 g1f3 c8f5 f1d3 b8d7 e1g1 f5g6 a3a4 c7c6 c3b4 e8e6 a4a5 g6f5 f1c1 d8f6 f3d2 a7a6 f2f3 e4d6 b4d6 f5d3 c2d3 e6d6 d2f1 d6e6 d3b3 a8b8 f1g3 g7g6 g3e2 e6e8 b3a3');
INSERT INTO evaluations VALUES(42689,30,'cp -312',NULL);
INSERT INTO evaluations VALUES(42689,31,'cp -309',NULL);
INSERT INTO evaluations VALUES(42689,32,'cp -292',NULL);
INSERT INTO evaluations VALUES(42689,33,'cp -308',NULL);
INSERT INTO evaluations VALUES(42689,34,'cp -325',NULL);
INSERT INTO evaluations VALUES(42689,35,'cp -317',NULL);
INSERT INTO evaluations VALUES(42689,36,'cp -309','c5f2 e1f2 f6g4 f2e1 d8g5 d1d2 g5e3 d2e3 g4e3 b1a3 d7d6 e1d2 e3g4 a3b5 b8a6 g1h3 f7f5 b5c3 f5f4 b2b3 h7h6 c3d1 a6c5 h3f2 a7a5 g2g3 g7g5 f2g4 c8g4 f1g2 e8e7 g3f4 g5f4 d1f2 h8g8 h1g1 a8f8 g2e4 c5e4 f2e4 g4f5 e4f2 c7c6');
INSERT INTO evaluations VALUES(42991,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(42991,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(42991,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(42991,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(42991,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(42991,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(42991,36,'cp 13','d7d5 c1f4 b7b6 c4d5 f6d5 c3d5 e6d5 e2e3 c7c5 d1f3 e8g8 f1d3 c8e6 g1e2 b8c6 e1g1 d8d7 f3g3 e6f5 d3f5 d7f5 d4c5 b6c5 f4d6 f5e6 d6e7 e6e7 e2f4 a8d8 a1c1 c5c4 h2h4 e7e5 g3g5 f7f6 g5h5 e5h5 f4h5');
INSERT INTO evaluations VALUES(43331,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43331,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43331,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43331,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(43331,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43331,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(43331,36,'cp 0','f8b4 f1c4 d7d6 e1g1 b4c3 b2c3 d8c7 d1d3 b8d7 f1d1 d7c5 d3d6 c7d6 d1d6 f6e4 d6d1 f7f6 d1e1 c8d7 f3e5 f6e5 f2f3 d7b5 c4b5 a6b5 f3e4 e8d7 c1e3 d7c6 e3c5 c6c5 e1f1 h8f8 f1f8 a8f8 a1d1 f8a8 d1d5 c5b6 d5d6 b6c7 d6d5');
INSERT INTO evaluations VALUES(39657,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39657,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39657,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39657,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39657,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39657,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39657,36,'cp 0','f1c4 d7d6 e1g1 b4c3 b2c3 d8c7 d1d3 b8d7 f1d1 b7b5 c4b3 d7c5 d3d6 c7d6 d1d6 f6e4 d6b6 f7f6 b3d5 c8b7 b6b7 e8c8 c1e3 d8d5 b7g7 e4c3 g7a7 h8d8 g2g3 d5d1 a1d1 d8d1 g1g2 c5e4 a7a6 c3d5 a6a8 c8d7 a8a7 d7e6 a7a6 e6e7 a6a7');
INSERT INTO evaluations VALUES(45682,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(45682,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(45682,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(45682,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45682,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(45682,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(45682,36,'cp 33','e4e5 f6d7 f3f4 c6c5 g1f3 b8c6 c1e3 a7a6 d1d2 b7b5 a2a3 d8a5 d4c5 f8c5 e3c5 d7c5 f3d4 a5b6 d4c6 b6c6 d2d4 e8g8 h2h4 c8b7 f1d3 a8c8 h4h5 h7h6 c3e2 a6a5 e1g1 b5b4 a3b4 a5b4 f4f5 c5e4 f5e6 f7e6 f1f8 c8f8');
INSERT INTO evaluations VALUES(46834,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46834,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(46834,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(46834,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(46834,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(46834,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(46834,36,'cp -13','d8d5 e2e3 c7c5 g1e2 c5d4 e2c3 d5e5 d1d4 f8g7 c3d5 e8f8 d4e5 g7e5 e3e4 b8c6 c1h6 e5g7 h6f4 c8e6 e1c1 a8d8 f1c4 f7f6 f2f3 f6f5 c1b1 g7e5 f4e3 h7h5 c4b5 f5e4 f3e4 f8g7 b5a4 d8c8 b2b3 h8d8 a4c6 c8c6 d5e7 c6d6 d1d6 e5d6');
INSERT INTO evaluations VALUES(48924,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(48924,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(48924,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(48924,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(48924,34,'cp -57',NULL);
INSERT INTO evaluations VALUES(48924,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(48924,36,'cp -62','f8d6 e2g3 c7c5 c1g5 e8g8 g3h5 d6e7 h5f6 e7f6 g5c1 b7b6 f1e2 b8c6 a2a4 d8d6 e1g1 g7g6 f3d2 d6e7 e2g4 c8a6 g2g3 a8e8 f2f4 e5f4 g3f4 f6g7 a4a5 c6a5 f4f5 a6c8 d1f3 g6f5 g4f5 c8f5 f3f5 e7e6 f5e6 f7e6 d2f3 a5c6 c1d2');
INSERT INTO evaluations VALUES(51086,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(51086,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(51086,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(51086,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(51086,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(51086,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(51086,36,'cp 80','b1c3 e7e6 e2e4 f8b4 a2a3 b4c3 b2c3 d7d5 c4d5 e6d5 e4d5 g8e7 d2d4 e8g8 f1d3 c8f5 c3c4 c7c6 g1f3 c6d5 c4c5 b8c6 e1g1 d8d7 f1e1 f5d3 d1d3 e7f5 c1f4 f8e8 a1d1 e8e1 d1e1 g7g6 h2h3 a8e8 f3e5 c6e5 f4e5 d7c6');
INSERT INTO evaluations VALUES(53180,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(53180,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53180,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53180,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53180,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(53180,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(53180,36,'cp 13','d2d3 c8d7 b1d2 d7c6 d2e4 e7e6 f1e2 f8e7 g1f3 b8d7 e1g1 g8h6 d1e1 h6f5 g2g4 c6e4 d3e4 f5h4 f3h4 e7h4 e1d2 h4e7 c2c3 e8g8 a2a4 a7a6 g1h1 d4c3 d2c3 c5c4 c3c4 a8c8 c4d3 d7c5 d3e3 b7b5 a4b5 a6b5 e2b5');
INSERT INTO evaluations VALUES(55270,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(55270,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(55270,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(55270,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(55270,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(55270,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(55270,36,'cp 42','f1c4 e7e6 d2d3 b8c6 f3g3 c6a5 b1d2 g8e7 g1e2 e7c6 e1g1 d8d7 c2c3 d4c3 e2c3 f7f6 d2b3 a5c4 d3c4 f8b4 c3b5 e8g8 c1e3 e6e5 a2a3 b4a5 b5a7 c6a7 b3a5 b7b6 a5b3 a7c6 f1d1 d7e6 b3d2 f8d8 b2b4 c6d4 e3d4 d8d4 g3e3 c8a6 b4b5');
INSERT INTO evaluations VALUES(55190,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(55190,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(55190,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(55190,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(55190,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(55190,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(55190,36,'cp 27','e7e6 d2d3 b8c6 b1d2 g8f6 c4b5 e6e5 g1e2 c8d7 b5c6 b7c6 d2c4 f8d6 c1g5 h7h6 g5f6 d8f6 f3f6 g7f6 e2g3 e8c8 e1g1 h6h5 h2h4 d7e6 g3f5 c8d7 b2b3 a7a6 g2g3 d6b4 f5g7 d8g8 g7e6 f7e6 g1h1 b4c3');
INSERT INTO evaluations VALUES(55492,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55492,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55492,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55492,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(55492,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55492,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55492,36,'cp 0','c7c6 b5e2 b8d7 g1f3 b7b5 d2d4 g8f6 e1g1 c8b7 f1e1 f8e7 e2d3 e8g8 a2a4 a7a6 a4b5 a6b5 a1a8 d8a8 d4e5 d7e5 f3e5 d6e5 d1f3 f8d8 f3g3 a8b8 c1h6 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(53797,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53797,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53797,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53797,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53797,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53797,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53797,36,'cp 0','b5e2 b8d7 g1f3 g8f6 e1g1 b7b5 d2d4 c8b7 f1e1 f8e7 e2d3 e8g8 a2a4 a7a6 a4b5 a6b5 a1a8 d8a8 d4e5 d7e5 f3e5 d6e5 d1f3 f8d8 f3g3 a8b8 h2h3 f6h5 g3f3 h5f6');
INSERT INTO evaluations VALUES(56139,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(56139,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(56139,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(56139,33,'cp 171',NULL);
INSERT INTO evaluations VALUES(56139,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(56139,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(56139,36,'cp 188','d1d4 b8d7 b1c3 g8f6 f1b5 f8d6 c1g5 d8e7 g1e2 d6e5 d4d2 a7a6 b5d7 c8d7 e1c1 e8c8 a2a3 h7h6 g5f4 c8b8 h1e1 h8e8 f4e5 e7e5 e2g3 e5d6 e1e8 d8e8 f2f3 e8d8 a3a4 d8e8 d2d4 h6h5 b2b4 d6e5 d5d6 e5d6 d4d6 c7d6 d1d6');
INSERT INTO evaluations VALUES(56136,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(56136,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(56136,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(56136,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(56136,34,'cp 184',NULL);
INSERT INTO evaluations VALUES(56136,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(56136,36,'cp 187','b8d7 b1c3 g8f6 f1b5 f8d6 c1g5 d8e7 g1e2 d6e5 d4d2 a7a6 b5d7 c8d7 e1c1 h7h6 g5f4 e8c8 e2g3 h8e8 f2f3 c8b8 f4e5 e7e5 h1e1 e5d6 e1e8 f6e8 d2d4 d7c8 d1d2 b7b6 b2b3 c8b7 g3f5 d6h2 f5g7 e8g7 d4g7 h2f4 g7d4 f4g5 c1b2');
INSERT INTO evaluations VALUES(56904,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(56904,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(56904,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(56904,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(56904,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(56904,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(56904,36,'cp 16','f8g7 f1g2 g8e7 d2d3 d7d6 a1b1 e8g8 e2e4 c7c5 a2a3 a7a5 g1e2 b8c6 e1g1 a8b8 c3b5 c8e6 e2c3 c6d4 b5d4 c5d4 c3b5 e7c6 f2f4 d8d7 b2b4 e5f4 g3f4 f7f5 e4f5 e6f5 b4a5 c6a5 a3a4 g8h8');
INSERT INTO evaluations VALUES(3469,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(3469,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(3469,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(3469,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(3469,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(3469,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(3469,36,'cp 38','h2h3 f6d7 f4h2 e7e5 d3e2 b7b6 e1g1 c8b7 a2a4 d8e7 b1a3 f8c8 b2b4 c6d8 f1e1 a7a6 d4e5 d7e5 f3e5 g7e5 h2e5 e7e5 d1b3 d8e6 a3c2 e5f6 e2g4 c8f8 f2f4 e6g7 c2d4 f8e8 a4a5 h7h5 g4f3 a8d8 a5b6');
INSERT INTO evaluations VALUES(7492,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(7492,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(7492,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(7492,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(7492,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(7492,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(7492,36,'cp 15','d7d6 a2a3 c6a5 b3a2 e8g8 e1g1 c8e6 b2b4 e6a2 a1a2 a5c6 c1g5 f6e8 g5e3 e8f6 f1e1 h7h6 h2h3 d8d7 c3d5 f6d5 e4d5 c6d8 c2c4 f7f5 d1b3 d8f7 a3a4 f5f4 e3d2 f7g5 f3g5 e7g5 a4b5 a6b5 a2a8 f8a8 e1c1 d7f7 c4b5');
INSERT INTO evaluations VALUES(7392,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(7392,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(7392,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(7392,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(7392,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(7392,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(7392,36,'cp 27','a2a3 c6a5 b3a2 c8e6 b2b4 e6a2 c3a2 a5c6 c2c4 c6d4 c1e3 c7c5 a2c3 e8g8 e1g1 h7h6 g1h1 d4c6 a1b1 d8d7 c3d5 a8b8 b4c5 d6c5 b1c1 b5c4 d5f6 e7f6 e3c5 f8d8 d3c4 d7e6 d1c2 c6d4 c5d4 e5d4 c2d3');
INSERT INTO evaluations VALUES(10593,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(10593,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(10593,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(10593,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(10593,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(10593,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(10593,36,'cp 188','c4d5 e4f2 e1f2 d4c3 f2g3 c3b2 c1b2 d7d6 h2h3 c8e6 d5c6 b7c6 d1c2 e8g8 b1d2 e6d5 d2e4 c5b6 g3h2 d6e5 b2e5 f7f6 e5g3 d8e7 h1e1 e7b4 e4c3 d5f7 e1e4 b4c5 e4h4 f7g6 c2b3 g6f7 b3b2 f8e8 a1d1 a8d8 d1d8 e8d8 c3e4 c5d5');
INSERT INTO evaluations VALUES(11629,30,'cp -142',NULL);
INSERT INTO evaluations VALUES(11629,31,'cp -129',NULL);
INSERT INTO evaluations VALUES(11629,32,'cp -150',NULL);
INSERT INTO evaluations VALUES(11629,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(11629,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(11629,35,'cp -146',NULL);
INSERT INTO evaluations VALUES(11629,36,'cp -132','c6d4 f3d4 e5d4 c3d5 c8b7 c2c3 d4c3 e1g1 f8e7 d5e7 d8e7 f1e1 e7e5 b2c3 e8g8 c1d2 f8e8 f2f3 h7h6 d1c1 f6d7 c3c4 c7c5 d2c3 e5e7 c3a5 d7e5 c1d2 e7f8 c4b5 a6b5 a5c7 c5c4 c7d6 e8d8 a1d1 c4b3 d6f8 d8d2 d1d2');
INSERT INTO evaluations VALUES(11990,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(11990,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(11990,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(11990,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(11990,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(11990,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(11990,36,'cp -5','d1d3 f8b4 c1d2 a8b8 b1c3 b4c5 b2b3 c5d4 f1e2 e8g8 e1g1 d7d6 d2e3 c6c5 e3d2 c8b7 e2f3 f6e8 a1e1 e8c7 c3e2 c7e6 f3g4 f8e8 g4e6 e8e6 e2d4 e5d4 f2f3 a7a5 a2a3 h7h5 f1f2 h5h4 h2h3 b7c6 a3a4 e6g6');
INSERT INTO evaluations VALUES(12319,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(12319,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(12319,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(12319,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(12319,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(12319,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(12319,36,'cp 41','c6d4 e4e5 f6g8 f1e1 a7a6 b5f1 d7d6 d2d3 d4f3 d1f3 d6d5 c3e2 c8d7 a2a4 g8e7 d3d4 e7g6 c2c3 f8e7 g2g3 e8g8 h2h4 f7f6 e5f6 e7f6 h4h5 g6e7 e2f4 c5d4 e1e6 d8c8 f1h3 d4c3 b2c3 f6c3 e6e7 d7h3 a1b1 c3f6 e7b7 h3f5 c1a3 f5b1 b7b1 c8f5');
INSERT INTO evaluations VALUES(12507,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(12507,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(12507,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(12507,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(12507,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(12507,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(12507,36,'cp -21','a7a6 g1f3 d7d6 c1f4 f8e7 e4e5 d6e5 d1d8 e7d8 f3e5 c6e5 f4e5 c8d7 e1c1 d7c6 f1e2 a8c8 h2h4 h7h5 f2f3 d8c7 e5c7 c8c7 d1d4 e8e7 c3d1 f6d5 e2c4 d5b6 h1e1 h8c8 c4d3 g7g6 d1c3 c7d7 d4d7 b6d7 b2b4 f7f5 c3e2 d7b6');
INSERT INTO evaluations VALUES(12785,30,'cp -130',NULL);
INSERT INTO evaluations VALUES(12785,31,'cp -148',NULL);
INSERT INTO evaluations VALUES(12785,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(12785,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(12785,34,'cp -159',NULL);
INSERT INTO evaluations VALUES(12785,35,'cp -141',NULL);
INSERT INTO evaluations VALUES(12785,36,'cp -155','f6e4 f1e1 e4f6 d2d4 c5d4 f3d4 e7e6 c4f1 f8c5 d4b3 c5b6 c2c4 d7d5 b1c3 d5c4 f1c4 d8d1 e1d1 e8g8 c1g5 h7h6 g5h4 f8d8 h4g3 c8d7 c3a4 c6e7 a4b6 a7b6 f2f3 a8a4 c4d3 d7c6 g3e5 e7g6');
INSERT INTO evaluations VALUES(14630,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(14630,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(14630,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(14630,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(14630,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(14630,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(14630,36,'cp 46','d2d4 f8e7 b1c3 e8g8 f1e1 e5d4 f3d4 c6d4 d1d4 c7c6 a2a4 f8e8 e1d1 d8b6 c1f4 e8d8 d4d3 a7a5 d3f3 c8e6 c4e6 f7e6 e4e5 d6e5 f4e5 f6d7 e5f4 d7c5 b2b3 e7f6 f3e3 d8d1 a1d1 b6b4 f4e5 f6e5 e3e5 a8e8 e5h5');
INSERT INTO evaluations VALUES(19087,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(19087,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19087,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19087,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19087,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19087,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19087,36,'cp 0','g8f6 b1c3 f8b4 c1d2 d7d6 f1b5 a7a5 c3d5 b4d2 f3d2 e8g8 d5f6 d8f6 d2c4 f6g6 e1g1 c6d4 f2f3 f7f5 c2c3 d4b5 a4b5 f5e4 f3e4 c8h3 b5d5 h3e6 d5d3 f8f1 a1f1 d6d5 c4e5 g6e4 d3d4 e4d4 c3d4 a8c8 a2a3 a5a4 h2h3 c8c2 f1f2 c2c1 f2f1');
INSERT INTO evaluations VALUES(12610,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(12610,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12610,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12610,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12610,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12610,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12610,36,'cp 0','b1c3 f8b4 c1d2 e8g8 f3e5 a7a5 a2a3 b4c3 d2c3 f8e8 f2f4 d7d5 e4d5 d8d5 a1d1 d5c5 c3d4 c6d4 a4d4 c5c2 f1c4 c8e6 e1g1 e6c4 d4c4 c2c4 e5c4 b7b5 c4d6 e8e3 d6f5 e3e6 f5d6');
INSERT INTO evaluations VALUES(12608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12608,36,'cp 0','f8b4 c1d2 e8g8 f3e5 a7a5 a2a3 b4c3 d2c3 f8e8 f2f4 d7d5 e4d5 d8d5 a1d1 d5c5 c3d4 c6d4 a4d4 c5c2 f1c4 c8e6 c4e6 e8e6 e1g1 c2c7 d1c1 c7e7 f1d1 h7h6 d4c5 e7e8 c5c7 e6e7 c7d6 e7e6 d6d4');
INSERT INTO evaluations VALUES(19282,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(19282,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(19282,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(19282,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(19282,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(19282,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(19282,36,'cp -11','g8f6 b1c3 f8g7 f1g2 d7d6 e1g1 e8g8 a2a4 a8b8 c3d5 f6d5 e4d5 c6b4 f3e1 e7e6 d5e6 c8e6 c2c3 b4c6 e1f3 d8d7 c1e3 e6g4 a4a5 b7b6 a5b6 a7b6 d1d2 c6e5 f3e5 d6e5 e3g5 g4h3 g2h3 d7h3 a1a7 h7h6 g5e7 f8e8 f1e1 h3f5 d2e2 f5e6');
INSERT INTO evaluations VALUES(21658,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(21658,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(21658,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(21658,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(21658,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(21658,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(21658,36,'cp -80','f7f6 g5c1 c8e6 f1e2 d8d7 e1g1 e8c8 c2c3 g7g5 b1d2 d5e4 d3e4 g8e7 d2b3 d7e8 d1c2 e7g6 f1d1 d8d1 c2d1 h7h5 f3e1 g6f4 c1f4 g5f4 e1d3 c6d8 b3c5 e6c4 b2b4 d8e6 d3f4 c4e2 f4e2 f8c5 b4c5 e8c6 d1d5 e6c5 e2g3 c6d5 e4d5 h5h4 g3f5 c8d7 c3c4 h4h3');
INSERT INTO evaluations VALUES(22255,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(22255,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(22255,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(22255,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(22255,34,'cp -60',NULL);
INSERT INTO evaluations VALUES(22255,35,'cp -103',NULL);
INSERT INTO evaluations VALUES(22255,36,'cp -95','g5c1 c8e6 f1e2 d8d7 e1g1 e8c8 c2c3 g7g5 b2b4 c6e7 b4b5 c8b8 f3d2 e7g6 e4d5 e6d5 c3c4 d5e6 d2b3 g6f4 c1f4 g5f4 b1d2 f8b4 a2a3 b4d2 b3c5 d7d6 c5e6 d6e6 d1d2 g8h6 e2f3 h6f5 g1h1 f5d4 f3e4 h8g8 d2a5 f6f5 e4d5 d8d5 c4d5');
INSERT INTO evaluations VALUES(21868,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(21868,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(21868,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21868,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(21868,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(21868,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(21868,36,'cp 15','f8e7 b1c3 g8f6 g1e2 a7a6 b5c6 b7c6 e1g1 e8g8 h2h3 f8e8 f1e1 h7h6 b2b3 d6d5 e4e5 f6h7 d1d3 c8b7 c3a4 e7b4 c2c3 b4f8 f2f4 f7f6 c1e3 f6e5 f4e5 e8e5 a4c5 b7c8 e3d4 f8c5 d4c5 c8d7');
INSERT INTO evaluations VALUES(22377,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(22377,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(22377,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(22377,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(22377,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(22377,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(22377,36,'cp 143','b1c3 a7a6 c1f4 f8e7 f2f3 g8f6 d1d2 d7e5 e1c1 e8g8 g2g4 b7b5 h2h4 b5b4 c3d5 f6d5 f4e5 d6e5 d4c6 d8e8 e4d5 e7c5 d1e1 c5d6 h4h5 c8d7 h5h6 g7g6 c6b4 e8e7 c2c3 a6a5 b4c6 e7f6 d2e3 d7c6 d5c6 f6e6 c1b1 f8d8 b1a1 a5a4');
INSERT INTO evaluations VALUES(22376,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(22376,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(22376,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(22376,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(22376,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(22376,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(22376,36,'cp 166','c7c6 c1e3 g8f6 d1e2 g7g6 e1c1 f8g7 e2d2 e8g8 e3h6 g7h6 d2h6 f8e8 f2f3 d8e7 g2g4 d7e5 h2h3 c6c5 d4b5 e5f3 f1g2 f3d4 b5d4 c5d4 d1d4 c8d7 h6d2 d7c6 h1d1 a8d8 g2f3 e7e5 g4g5 e5g3 g5f6 g3f3 d2g5 f3f2 c1b1 g8h8 h3h4');
INSERT INTO evaluations VALUES(23412,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(23412,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(23412,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(23412,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(23412,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(23412,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(23412,36,'cp 113','d2d4 d7d6 c1e3 g8f6 f2f3 e7e5 g1e2 e5d4 e2d4 f8g7 d1d2 c6d4 e3d4 c7c6 g2g4 b7b5 e1c1 e8g8 g4g5 f6h5 d4g7 h5g7 d2d6 d8b6 h2h4 c8e6 d6d4 b6a5 a2a3 a8d8 d4c5 d8d1 c3d1 a5c7 c1b1 c7f4 f1d3 f4f3 h1f1 f3g3 c5c6 g7h5 c6b5 g3h4 b5a5');
INSERT INTO evaluations VALUES(23913,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(23913,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(23913,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(23913,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(23913,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(23913,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(23913,36,'cp 37','d4c6 b7c6 b5d3 g8f6 h2h3 e8g8 b1c3 a7a5 f1e1 d8e7 c1f4 c5d4 d1d2 a5a4 f4e3 e7e5 e3d4 e5d4 a1d1 f8b8 b2b3 a4b3 c2b3 h7h6 d2c2 d4e5 c3b1 d7e6 b1d2 f6d7 d2f3 e5c5 c2d2 d7e5 f3e5 c5e5');
INSERT INTO evaluations VALUES(23903,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(23903,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(23903,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(23903,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(23903,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(23903,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(23903,36,'cp 24','b7c6 b5d3 g8f6 b1c3 a7a5 h2h3 e8g8 f1e1 f8e8 c1e3 c5e3 e1e3 d8b8 b2b3 b8b4 d1e1 h7h6 a2a3 b4b6 c3a4 b6d4 a1d1 d4e5 e1c3 e5c3 a4c3 g7g5 c3e2 f6h5 d1f1 c6c5 a3a4 a8b8 f2f4 b8b4 g1h2 h5f4');
INSERT INTO evaluations VALUES(25053,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(25053,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(25053,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(25053,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(25053,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(25053,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(25053,36,'cp 82','g2g4 h5g6 a2a4 d8d7 g1g2 c6a5 c4a2 f8e7 f3h4 g6h7 b2b4 a5c6 b4b5 c6b4 a2b3 a7a5 h4f5 h7f5 e4f5 d6d5 c3e2 d5d4 e2g3 e7d6 c1d2 e8c8 d1f3 d7e8 a1e1 c8b8 g2h1 b7b6 f3g2 e8d7 g3e4 f6e4');
INSERT INTO evaluations VALUES(26931,30,'cp 525',NULL);
INSERT INTO evaluations VALUES(26931,31,'cp 530',NULL);
INSERT INTO evaluations VALUES(26931,32,'cp 505',NULL);
INSERT INTO evaluations VALUES(26931,33,'cp 515',NULL);
INSERT INTO evaluations VALUES(26931,34,'cp 521',NULL);
INSERT INTO evaluations VALUES(26931,35,'cp 529',NULL);
INSERT INTO evaluations VALUES(26931,36,'cp 515','d5c6 b7c6 b1c3 g8f6 h2h3 g4c8 f1c4 f8e7 d1e2 e8g8 e1g1 a7a5 c1e3 g8h8 a1d1 f6d7 f1e1 d8e8 c4d3 c8b7 g1h1 a5a4 e3c1 d7c5 f3g1 f7f5 e4f5 c5d3 e2d3 c6c5 c3d5 e7h4 e1f1 e8f7 f2f3 f7d5 d3d5 b7d5 d1d5 f8f5 d5d1');
INSERT INTO evaluations VALUES(26758,30,'cp 501',NULL);
INSERT INTO evaluations VALUES(26758,31,'cp 511',NULL);
INSERT INTO evaluations VALUES(26758,32,'cp 493',NULL);
INSERT INTO evaluations VALUES(26758,33,'cp 485',NULL);
INSERT INTO evaluations VALUES(26758,34,'cp 497',NULL);
INSERT INTO evaluations VALUES(26758,35,'cp 511',NULL);
INSERT INTO evaluations VALUES(26758,36,'cp 510','b7c6 h2h3 g4c8 f1c4 f8e7 b1c3 g8f6 d1e2 e8g8 c1e3 d6d5 e1c1 e7b4 f3e5 c8b7 e4d5 c6d5 c4d5 f6d5 c3d5 b7d5 e5c6 d5c6 d1d8 a8d8 e2c4 c6g2 h1g1 b4a5 g1g2 d8d6 e3f4 d6g6 f4g3 f8d8 b2b3 c7c6 f2f4 a5c7 c4a4 a7a5 a4c4 d8d6');
INSERT INTO evaluations VALUES(28458,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(28458,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(28458,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28458,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28458,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(28458,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28458,36,'cp 0','e8g8 c2c4 c7c6 f1e2 a7a5 g3d6 d8d6 e1g1 d6e7 a2a3 b8d7 d1b3 a5a4 b3b7 f8b8 b7c7 f6e8 c7c6 b8c8 c6b5 c8b8');
INSERT INTO evaluations VALUES(30258,30,'cp 192',NULL);
INSERT INTO evaluations VALUES(30258,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(30258,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(30258,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(30258,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(30258,35,'cp 181',NULL);
INSERT INTO evaluations VALUES(30258,36,'cp 189','h2h3 g4h5 g2g4 h5g6 f3e5 b8c6 d1e2 c6e5 e2e5 d8e7 e5e7 d5e7 b1c3 h7h5 h1g1 e7c6 c1e3 h5g4 h3g4 e8c8 e1c1 c6e5 c4d5 f8b4 d5g2 d8e8 d1e1 e5g4 g2b7 c8b7 g1g4 b4e7 a2a3 a7a6 b2b4 e7d6 e1g1 d6h2 g1g2 h8h3 g4c4 h2e5 g2g1 h3f3 g1h1 e5c3');
INSERT INTO evaluations VALUES(30257,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(30257,31,'cp 189',NULL);
INSERT INTO evaluations VALUES(30257,32,'cp 182',NULL);
INSERT INTO evaluations VALUES(30257,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(30257,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(30257,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(30257,36,'cp 190','g4h5 g2g4 h5g6 f3e5 b8c6 d1e2 c6e5 e2e5 d8e7 e5e7 d5e7 b1c3 h7h5 h1g1 e7c6 c1e3 e8c8 e1c1 h5g4 h3g4 f8d6 c4d5 d6h2 g1g2 h2d6 c3e4 d6e7 d5c6 b7c6 g2g1 a7a5 a2a3 h8h3 g1h1 d8h8 h1h3 h8h3 c1d2 f7f6 d1e1 a5a4 e4c3 h3h4 c3a4 h4g4 a4c5');
INSERT INTO evaluations VALUES(30363,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(30363,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(30363,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(30363,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(30363,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(30363,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(30363,36,'cp 107','e7e6 f1d3 d7d6 e1g1 f8e7 d1e2 g8f6 f1e1 f6d7 d4d5 e6e5 a2a3 e8g8 b2b4 c7c6 d5c6 b7c6 c3d1 c6b7 a1b1 g7g6 c1h6 f8e8 d1e3 d7f6 f3d2 e7f8 h6f8 e8f8 c2c4 b8d7 e3d5');
INSERT INTO evaluations VALUES(30354,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(30354,31,'cp 116',NULL);
INSERT INTO evaluations VALUES(30354,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(30354,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(30354,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(30354,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(30354,36,'cp 122','f1d3 d7d6 e1g1 f8e7 d1e2 g8f6 f1e1 e8g8 e4e5 d6e5 d4e5 f6d7 c3e4 b8c6 e4g5 g7g6 d3e4 d7c5 c2c3 a6a5 e4c2 b7a6 e2e3 a6d3 e1d1 d3c2 d1d8 f8d8 f3d4 c6d4 c3d4 d8d4 e3d4 a8d8 d4g4 h7h5 g4f3 d8d1 f3d1 c2d1 c1e3 d1g4 a1c1 g4f5 g5f3 f5e4');
INSERT INTO evaluations VALUES(31365,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31365,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31365,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31365,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31365,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31365,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31365,36,'cp 0','e7e6 b1c3 b8c6 a2a3 g8e7 c1e3 a8c8 c3b5 e7g8 b5c3');
INSERT INTO evaluations VALUES(32707,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(32707,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(32707,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(32707,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(32707,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(32707,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(32707,36,'cp 18','c2c4 d5d6 d2d4 b8c6 d4d5 g4f3 e2f3 c6d4 e1g1 f8e7 c1e3 d4f3 d1f3 e8g8 f3f5 g7g6 f5d3 b7b6 f1d1 d6f6 a1c1 a7a5 c4c5 b6c5 e3c5 e7d6 b2b3 f6f5 c1c4 f5d3 d1d3 a5a4 b3b4 f8b8 d3c3 f7f5 g2g4 g8f7 g4f5 g6f5');
INSERT INTO evaluations VALUES(33538,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(33538,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(33538,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(33538,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(33538,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(33538,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(33538,36,'cp 42','f2f4 d7d5 e4e5 f6e4 c3e4 d5e4 c1e3 b8c6 d4c6 c7c6 d1d4 f8e7 f1e2 e8g8 d4b6 c8d7 e1c1 a8c8 c2c3 c6b6 e3b6 d7a4 d1d4 a4c6 h2h4 c6d5 d4d1 e4e3 b6e3 d5g2 h1g1 g2c6 h4h5 c6a4 d1f1 f8d8 h5h6 g7g6 e3b6 d8d7 f4f5');
INSERT INTO evaluations VALUES(33718,30,'cp -206',NULL);
INSERT INTO evaluations VALUES(33718,31,'cp -213',NULL);
INSERT INTO evaluations VALUES(33718,32,'cp -185',NULL);
INSERT INTO evaluations VALUES(33718,33,'cp -182',NULL);
INSERT INTO evaluations VALUES(33718,34,'cp -193',NULL);
INSERT INTO evaluations VALUES(33718,35,'cp -182',NULL);
INSERT INTO evaluations VALUES(33718,36,'cp -176','f3f2 g4f5 g1f3 f6g4 f2d2 g4e3 d2e3 e7e6 c3b5 f8b4 c2c3 b4a5 f1d3 f5f6 e1g1 a5b6 b5d4 f6h6 e3h6 g7h6 f3e5 b8d7 e5c4 c7c6 c4d6 e8e7 d6c4 h8d8 c4b6 a7b6 a1e1 d7f6 a2a3 h6h5 f1f3 c6c5 d4f5 e7f8 f5g3 h5h4 f3f6');
INSERT INTO evaluations VALUES(34937,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(34937,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(34937,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(34937,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(34937,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(34937,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(34937,36,'cp 32','d5d8 g1f3 f8d6 d2d4 e5d4 f3d4 g8f6 d4b5 d6c5 d1d8 e8d8 f1e2 c7c6 b5d4 c5d4 e3d4 c8e6 e1g1 h8e8 f2f3 a7a5 c3e4 b8a6 c1g5 d8e7 a2a3 e8d8 f1e1 h7h6 g5h4 d8d4 e2a6 a8a6 h4f2 d4c4 e4c5 a6a8 c5b7 c4c2 b7c5 c2b2 a1c1 a5a4 f2g3');
INSERT INTO evaluations VALUES(56626,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(56626,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(56626,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(56626,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(56626,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(56626,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(56626,36,'cp 40','g1f3 f8d6 d2d4 e5d4 f3d4 g8f6 d4b5 d6e7 d1d8 e7d8 e3e4 a7a6 b5d4 c7c5 d4b3 b7b6 c1e3 d8c7 e1c1 e8g8 h2h3 f8e8 g2g4 h7h6 h1g1 c8b7 b3d2 b6b5 h3h4 b5b4 c3a4 f6e4 d2e4 b7e4 a4c5 e4f3 f1g2 f3g2 g1g2 b8c6 g4g5 h6h5 g5g6 c6e5 g6f7 e5f7 e3d4');
INSERT INTO evaluations VALUES(35082,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(35082,31,'cp 184',NULL);
INSERT INTO evaluations VALUES(35082,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(35082,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(35082,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(35082,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(35082,36,'cp 170','d2d4 f8b4 g1e2 b8c6 c1e3 d7d6 d1d2 b4a5 a2a3 c8d7 e1c1 f6e7 g2g4 g8f6 f2f3 e8c8 h2h4 c8b8 c1b1 d6d5 e4e5 f6g8 f3f4 e7e8 e2g3 f7f6 f1e2 h7h6 d1f1 g8e7 h4h5 b8a8 f1e1 a8b8 e5f6 g7f6 g4g5 h6g5 f4g5 f6g5 e3g5');
INSERT INTO evaluations VALUES(35283,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(35283,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(35283,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(35283,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(35283,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(35283,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(35283,36,'cp 32','g7g6 c2c4 d7d6 b1c3 f8g7 g2g3 c8g4 c3d5 e7d8 c1g5 g8f6 h2h3 g4e6 d1b3 e6c8 d3d4 h7h6 d5f6 g7f6 g5e3 b8c6 a1d1 e5d4 f3d4 c6d4 e3d4 f6d4 d1d4 d8f6 d4d2 h6h5 f1g2 h5h4 c4c5 d6c5 b3b5 f6c6 b5c6 b7c6 e4e5 e8g8 g2c6');
INSERT INTO evaluations VALUES(35987,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(35987,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(35987,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(35987,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(35987,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(35987,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(35987,36,'cp 17','g5h6 g7h6 d2h6 b8c6 e1c1 d8b6 h6d2 b6d4 d2d4 c6d4 d1d4 c8e6 b2b3 a8c8 c1b2 g8g7 f2f3 h7h5 f1e2 h5h4 h1d1 c8c5 c3a4 c5g5 e2f1 e6d7 a4c3 f8c8 a2a4 a7a6 c3d5 f6d5 e4d5 c8h8 d1e1');
INSERT INTO evaluations VALUES(36022,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36022,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36022,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(36022,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(36022,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36022,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(36022,36,'cp 12','g7h6 d2h6 b8c6 e1c1 d8a5 a2a3 a5b6 h6d2 b6d4 d2d4 c6d4 d1d4 c8e6 f1d3 f8c8 h1e1 c8c5 d4b4 a8b8 h2h3 g8f8 b4b5 c5b5 c3b5 a7a6 b5c7 e6d7 c7d5 f6d5 e4d5 b8c8 b2b4 e7e5 d5e6 f7e6 f2f4 f8f7 c1b2 f7e7 g2g4');
INSERT INTO evaluations VALUES(37507,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(37507,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(37507,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(37507,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(37507,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(37507,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(37507,36,'cp 108','c1e3 c5b4 d1g4 g8f6 g4g7 h8g8 g7h6 e6e5 d4f5 g8g6 f5g7 e8e7 g7f5 e7e8 h6h3 f6e4 h3h7 e4f6 h7h8 g6g8 h8h4 g8g4 h4h3 d7d5 f2f3 g4g8 e1c1 b4c3 f5d6 e8e7 d6c8 d8c8 h3c8 g8c8 b2c3 c8c3 e3d2 c3c7 h2h4 b8c6 h4h5 c6d4 f1d3 a8c8 c2c3 e5e4 h1e1 f6h5 f3e4 c7c3 d2c3 c8c3 c1b1 h5f4');
INSERT INTO evaluations VALUES(37506,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(37506,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(37506,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(37506,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(37506,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(37506,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(37506,36,'cp 130','d7d6 d1d2 g8f6 e1c1 d8e7 h1g1 b7b5 g2g4 c8b7 d4b3 c5e3 d2e3 b8c6 g4g5 f6d7 e3g3 e8c8 h2h4 d7b6 g3e3 b6d7 b3d4 d7e5 f1e2 c6d4 d1d4 e7c7 g1d1 e5c6 d4d2 h7h6 g5g6 f7g6 e3g3 g6g5 h4g5 d6d5 e4d5 c7g3 f2g3 e6d5 g5h6 h8h6 c3d5 c8b8');
INSERT INTO evaluations VALUES(38135,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(38135,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(38135,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(38135,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38135,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(38135,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(38135,36,'cp 16','g8e7 b1c3 c5b6 f2f4 c8d7 d1d2 b8c6 e1c1 c6d4 e3d4 h7h5 c1b1 d7c6 d4f2 b6a5 d2d3 e7f5 c3b5 a5b6 b5d4 b6d4 f2d4 f5d4 d3d4 d8b6 d4b6 a7b6 b2b3 h8g8 d1d4 g7g5 g2g3 g5f4 g3f4 g8g4 f1d3 g4h4 a2a4 e8e7 b1b2');
INSERT INTO evaluations VALUES(38513,30,'cp -198',NULL);
INSERT INTO evaluations VALUES(38513,31,'cp -196',NULL);
INSERT INTO evaluations VALUES(38513,32,'cp -197',NULL);
INSERT INTO evaluations VALUES(38513,33,'cp -205',NULL);
INSERT INTO evaluations VALUES(38513,34,'cp -216',NULL);
INSERT INTO evaluations VALUES(38513,35,'cp -192',NULL);
INSERT INTO evaluations VALUES(38513,36,'cp -167','g8f6 b1d2 b8c6 a2a4 e8g8 a4a5 c6e7 f1e1 e7g6 d1c2 d6d5 e4d5 f8e8 e1e8 d8e8 d5d6 c7d6 c2g6 d6d5 g6d3 d5c4 d2c4 f6e4 c1f4 g5f4 a1e1 f7f5 c4d6 e8c6 d6e4 f5e4 d3e4 c6e4 e1e4 c8d7 e4f4 a8f8 f4e4 d7c6 e4e3 h6h5 f3e5 c6e8 h2h4 g7f6 g2g3 g8h7 e3e1');
INSERT INTO evaluations VALUES(39773,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(39773,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(39773,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(39773,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(39773,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(39773,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(39773,36,'cp 56','c1f4 e8g8 e1g1 f8e8 a2a4 b8c6 d4c6 b7c6 f1e1 c8g4 d1d2 a6a5 b2b3 g4e6 c4d3 f6d7 a1d1 d7c5 d3f1 d8b8 f2f3 b8b4 c3e2 b4d2 d1d2 g7g5 f4e3 h7h5 e2d4 e6d7 f1c4 e7f6 d4e2 d7e6 e3c5 e6c4 c5d4 f6d4 d2d4 c4e2 e1e2');
INSERT INTO evaluations VALUES(40748,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(40748,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(40748,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(40748,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(40748,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(40748,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(40748,36,'cp 121','e5f4 b8c6 c2c3 d6d5 b1d2 e8g8 f1e2 f8e8 e4e5 f6d7 d2f1 f7f6 e5f6 d7f6 f4d2 f6e4 d2c2 e7h4 f1g3 h4g3 h2g3 c8f5 c2d1 d8f6 g1f3 e8e6 c1f4 a8e8 f4e5 c6e5 d4e5 f6d8 e1g1 c7c5 d1c1 f5g4 f1e1 e6h6 f3h4 g4e2 e1e2');
INSERT INTO evaluations VALUES(40986,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(40986,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(40986,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(40986,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(40986,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(40986,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(40986,36,'cp 14','g1f3 e8g8 f1d3 c7c5 c2c3 f6h5 d4c5 h5f4 e3f4 e7c5 e1g1 g7g6 d2b3 c5e7 f1e1 b8c6 g2g3 a7a5 a2a4 c8d7 d3b5 f8e8 d1e2 d8b6 a1d1 h7h5 h2h4 a8c8 d1d2 g8g7 g1g2 e8d8 e2d1 b6c7 d2e2 b7b6 b3d4 e7f6 d4c6 d7c6 b5c6 c7c6');
INSERT INTO evaluations VALUES(42015,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(42015,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(42015,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(42015,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(42015,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(42015,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(42015,36,'cp -59','b4c3 b2c3 c7c5 g2g3 d8a5 f4d2 a5a4 d4c5 b8c6 f1g2 e8g8 e1g1 e6e5 d2e3 c8f5 f3h4 f5c2 d1d2 c6a5 a1c1 c2e4 g2e4 a4e4 d2d3 e4g4 d3f5 a5c4 e3g5 g4f5 h4f5 f6e4 g5e7 f8e8 e7d6 e4d6 f5d6 c4d6 c5d6 e8d8 c3c4 d8d6 c4d5 d6d5 c1c7 b7b6 e2e4');
INSERT INTO evaluations VALUES(42098,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(42098,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(42098,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(42098,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(42098,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(42098,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(42098,36,'cp -40','b2c3 c7c5 d4c5 d8a5 d1d3 e8g8 f3d2 a5c5 d3g3 f6h5 g3e3 c5e3 f4e3 e6e5 e3c5 f8d8 e1c1 c8e6 e2e4 d5e4 d2e4 d8d1 c1d1 b8d7 c5d6 f7f5 e4g5 e6a2 d1c1 h5f6 f1b5 a7a6 b5a4 a8c8 h1d1 e5e4 a4d7 f6d7');
INSERT INTO evaluations VALUES(42097,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(42097,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(42097,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(42097,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(42097,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(42097,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(42097,36,'cp -38','c7c5 d4c5 d8a5 d1d3 e8g8 f3d2 a5c5 d3g3 f6h5 g3e3 c5e3 f4e3 h5f6 c3c4 b8c6 e3g5 d5d4 g5f6 g7f6 g2g4 a8b8 f1g2 c6a5 a1b1 b7b6 c4c5 c8d7 c5b6 b8b6 b1b6 a7b6 e1g1 f8c8 g2e4 d7a4 e4d3 e6e5 f1b1 c8c6 b1b2 g8g7 h2h3 c6c3 b2b6 a4c2 d3c2 c3c2');
INSERT INTO evaluations VALUES(42783,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(42783,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42783,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(42783,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42783,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42783,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(42783,36,'cp 9','c1d2 b4d2 d1d2 e5d4 d2d4 b8c6 d4d2 d7d5 b1c3 d5c4 d2d8 e8d8 f1c4 d8e7 g1f3 h8d8 a2a3 c6a5 c4a2 c7c5 b2b4 c5b4 a3b4 a5c6 b4b5 c6b4 f3d4 c8d7 a2b3 a7a6 b5a6 a8a6 e1e2 b4c6 c3b5 c6e5 b5c7 a6a1 h1a1 e5c6 f2f3 c6d4 e3d4');
INSERT INTO evaluations VALUES(42782,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42782,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42782,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(42782,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42782,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42782,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42782,36,'cp -5','b4d2 d1d2 e5d4 d2d4 b8c6 d4d2 d7d5 b1c3 d5c4 d2d8 e8d8 f1c4 d8e7 a1d1 h8d8 d1d8 c6d8 f2f4 a7a6 g1f3 b7b5 c3d5 f6d5 c4d5 c7c6 d5e4 g7g6 e1f2 c8e6 a2a3 f7f5 e4c2 e6d5 h2h4 a6a5 h4h5 d8e6 h5g6 h7g6 h1h7 e7d6 h7h6 a5a4');
INSERT INTO evaluations VALUES(43120,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43120,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(43120,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(43120,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(43120,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(43120,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(43120,36,'cp 0','d7d6 g2g3 a7a5 g1e2 a5a4 f1g2 e8g8 e1g1 a4a3 b2b3 c7c5 d5c6 b7c6 a1b1 f6d7 c3a4 d7b6 a4b6 d8b6 c1e3 b6a5 e3d2 a5a7 d2e3');
INSERT INTO evaluations VALUES(43223,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(43223,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(43223,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(43223,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(43223,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(43223,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(43223,36,'cp -4','c7c5 c2c3 e8g8 f1e2 b7b6 c1b2 c8b7 e1g1 b8c6 c3c4 c5d4 f3d4 c6d4 b2d4 a8c8 b1a3 d7d5 c4d5 f6d5 d4g7 g8g7 d1d4 d5f6 d4b2 d8d7 f1d1 d7f5 a1c1 c8c1 d1c1 f8c8 c1c8 f5c8 a3c4 g7g8 f2f3 c8c5 g1f2 b7d5 c4e5 c5c7 f2g1 d5e6 g1f2');
INSERT INTO evaluations VALUES(44332,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(44332,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(44332,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(44332,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(44332,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(44332,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44332,36,'cp 14','f6d5 c4d5 d7d6 g1f3 f8e7 e2e4 c8g4 f1d3 e8g8 e1g1 e7f6 h2h3 g4f3 d1f3 b8d7 f3e2 a7a6 c1e3 d8a5 a2a3 d7e5 e3d2 a5c7 d2c3 a8e8 a1e1 e5d3 e2d3 f6c3 d3c3 b7b5 e1e3 f7f6 f2f4 c5c4 f1e1 c7c5 g1h2 f6f5');
INSERT INTO evaluations VALUES(44360,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(44360,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(44360,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(44360,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(44360,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(44360,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44360,36,'cp 18','c4d5 d7d6 g1f3 f8e7 e2e4 e8g8 f1d3 c8g4 e1g1 e7f6 h2h3 g4f3 d1f3 b8d7 f3e2 a7a6 c1f4 d8e7 a1e1 b7b5 b2b3 f6c3 f4d2 e7f6 d3b1 g7g6 f2f4 f6d4 d2e3 d4g7 e1d1 f7f5 e4f5 g6f5 g1h1 a8e8 e2f3 c3d4 e3d4 c5d4');
INSERT INTO evaluations VALUES(44876,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(44876,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(44876,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(44876,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(44876,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(44876,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(44876,36,'cp -8','c6d5 g1f3 b8c6 d2d4 c8f5 b1c3 a8c8 d1b3 c6a5 b3b5 f5d7 b5d3 e7e6 e1g1 f8b4 f3e5 h7h6 c1d2 a5c4 e5c4 d5c4 d3c2 d7c6 a2a3 b4a5 g2c6 c8c6 e2e4 d8d4 f1d1 a5c3 d2c3 d4e4 c2e4 f6e4 c3g7 h8g8 g7h6');
INSERT INTO evaluations VALUES(45974,30,'cp -198',NULL);
INSERT INTO evaluations VALUES(45974,31,'cp -187',NULL);
INSERT INTO evaluations VALUES(45974,32,'cp -194',NULL);
INSERT INTO evaluations VALUES(45974,33,'cp -197',NULL);
INSERT INTO evaluations VALUES(45974,34,'cp -202',NULL);
INSERT INTO evaluations VALUES(45974,35,'cp -194',NULL);
INSERT INTO evaluations VALUES(45974,36,'cp -194','h7h6 g5h3 a7a6 d2d3 d6d5 c4b3 b8c6 f3e3 f8d6 b1d2 e8g8 c2c3 d5d4 e3e2 e6e5 e1g1 c8g4 f2f3 g4h3 g2h3 f6h5 d2b1 d8d7 e2g2 h5f4 c1f4 e5f4 b3d5 d6e5 g2g4 d7c7 g1h1 f8d8 d5c6 c7c6 g4f5 d4c3 f5e5 d8d3 e5f4 c3b2');
INSERT INTO evaluations VALUES(46143,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46143,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(46143,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46143,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46143,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(46143,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46143,36,'cp 8','c5d4 f3d4 f8b4 b1c3 d8a5 g5f6 b4c3 b2c3 a5c3 e1e2 g7f6 a1c1 c3a5 c4d5 e8g8 d5e6 c8e6 d4e6 a5a2 e2e1 a2a5 d1d2 a5d2 e1d2 f7e6 c1c7 f8d8 d2c2 b8a6 f1a6 d8c8 c7c8 a8c8 c2d3 b7a6 h1b1 a6a5 b1b5 c8d8 d3e2 a5a4 b5a5 d8a8 g2g4 e6e5 e2f3 a4a3 a5a3 a7a5 f3e4');
INSERT INTO evaluations VALUES(46353,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(46353,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46353,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(46353,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(46353,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46353,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(46353,36,'cp -2','f8d6 g1f3 b8c6 e1g1 e8g8 c1g5 h7h6 g5f6 d8f6 a1c1 f8d8 a2a3 f6e7 f1e1 e7f8 c1c2 c8d7 e2e4 d5e4 c3e4 a8c8 c2d2 d7e8 d4d5 e6d5 d2d5 d6e7 e4c3 e7f6 d1b3 b7b6 c3e4 f6e7 d5d8 c8d8 b3c4 e8d7 e4c3 e7f6 b2b4 c6e7 c3e4 f6b2');
INSERT INTO evaluations VALUES(47419,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(47419,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(47419,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(47419,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(47419,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(47419,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(47419,36,'cp 39','e7e5 g1f3 b8c6 b2b4 f8d6 b4b5 c6d4 d1a4 d4f3 g2f3 h7h6 d2d3 e8g8 e1g1 a7a6 c1a3 c8h3 a3d6 d8d6 f1b1 a6b5 a4b5 f8b8 a2a4 c7c6 b5b4 d6b4 b1b4 a8a5 f3g2 h3d7 a1b1 b7b5 a4b5 c6b5 f2f4 e5f4 g3f4 b8b6');
INSERT INTO evaluations VALUES(48158,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(48158,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(48158,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(48158,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(48158,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(48158,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(48158,36,'cp -41','c7c6 b5d3 f8e7 e1g1 e8g8 f1e1 c6c5 b2b3 a7a6 e4d5 e6d5 c3e2 b8c6 e2g3 f8e8 c1b2 c8d7 d3f5 d5d4 c2c3 d7f5 g3f5 e7f8 c3d4 e8e1 d1e1 g7g6 f5e3 c5d4 e3c2 f8c5 d2d3 d8d5 e1d2 a8d8 g1f1 f6h5 d2e2 h5f4 e2e4 f4e6');
INSERT INTO evaluations VALUES(45594,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(45594,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(45594,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(45594,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(45594,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(45594,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(45594,36,'cp -34','b5d3 f8e7 e1g1 e8g8 b2b3 c6c5 f1e1 a7a6 e4d5 e6d5 c3e2 b8c6 e2g3 f8e8 c1b2 c8d7 d3f5 d5d4 c2c3 d7f5 g3f5 e7f8 c3d4 g7g6 f5g3 c5d4 e1e8 f6e8 d2d3 f8g7 d1d2 d8d5 a1e1 a8d8 h2h4 e8f6 g3e4 d5f5 b2a3 f6e4 e1e4 h7h6 d2e2 d8c8 e4e8 c8e8 e2e8 g8h7');
INSERT INTO evaluations VALUES(48661,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(48661,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(48661,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(48661,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(48661,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(48661,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(48661,36,'cp 0','f8d6 c1g5 c7c6 d1f3 b8d7 f3e3 d6e7 g1f3 e8g8 f1d3 h7h6 g5f4 f8e8 e1g1 e7f8 e3c1 a7a5 a2a4 f6e4 c3d1 d7f6 d1e3 f8d6 f4d6 e4d6 c2c3 d6e4 f1e1 f6h5 f3e5 d8g5 e5f3');
INSERT INTO evaluations VALUES(48823,30,'cp -108',NULL);
INSERT INTO evaluations VALUES(48823,31,'cp -106',NULL);
INSERT INTO evaluations VALUES(48823,32,'cp -121',NULL);
INSERT INTO evaluations VALUES(48823,33,'cp -112',NULL);
INSERT INTO evaluations VALUES(48823,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(48823,35,'cp -100',NULL);
INSERT INTO evaluations VALUES(48823,36,'cp -114','d5e4 d3e4 f6e4 c3e4 b8c6 d2d3 f8e7 e1g1 f7f6 f1e1 e8g8 c2c3 d8d5 d3d4 c8g4 h2h3 g4h5 g2g4 h5g6 e4g3 e5d4 f3d4 c6d4 d1d4 d5d4 c3d4 e7d6 c1d2 f8d8 a1c1 a7a5 g3e4 d6f8 d2c3 b7b6 f2f3 g6f7 a2a3');
INSERT INTO evaluations VALUES(48887,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(48887,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(48887,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(48887,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(48887,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(48887,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(48887,36,'cp -69','d2d3 b8c6 a2a3 c8e6 b2b3 f8d6 c1b2 e8g8 b1d2 a7a5 g1f3 f6d7 e2e4 d5d4 e1g1 d7c5 g1h2 f7f6 f3h4 d8d7 d2c4 b7b5 c4d6 c7d6 f2f4 b5b4 h4f5 f8c8 d1d2 c8d8 a3b4 a5b4 g3g4 g8h8 d2f2 a8a1 f1a1 g7g6');
INSERT INTO evaluations VALUES(48874,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(48874,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(48874,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(48874,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(48874,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(48874,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(48874,36,'cp -59','b8c6 a2a3 c8e6 b2b3 f8d6 c1b2 a7a5 g1f3 h7h6 e2e3 d8d7 d3d4 e5d4 f3d4 c6d4 b2d4 d6e7 b1c3 e8g8 c3e2 f8e8 e2f4 e6f5 e1g1 f6e4 f4h5 e7f8 g3g4 f5h7 f2f3 e4d6 f3f4 c7c6 f4f5 d6b5 c2c3 b5d4 d1d4');
INSERT INTO evaluations VALUES(51255,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(51255,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(51255,32,'cp 221',NULL);
INSERT INTO evaluations VALUES(51255,33,'cp 221',NULL);
INSERT INTO evaluations VALUES(51255,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(51255,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(51255,36,'cp 212','a4b5 c6b5 c3b5 e7e6 b5a3 g8f6 a3c4 f6e4 g1f3 f8b4 f3d2 e8g8 f1d3 c8b7 e1g1 e4d6 c4d6 b4d6 d2c4 b8c6 d3e4 d8c7 c4d6 c7d6 c1d2 a8b8 d1a4 h7h6 f1d1 c6e7 d2b4 d6d8 e4b7 b8b7 b4c3 d8d6 a4a6');
INSERT INTO evaluations VALUES(51253,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(51253,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(51253,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(51253,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(51253,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(51253,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(51253,36,'cp 187','g8f6 f1c4 e7e6 g1f3 f8e7 e1g1 e8g8 d1e2 c6b5 c4b5 c8b7 b5a4 a6a5 f1d1 e7b4 a4c2 b8d7 e3e4 d8c7 c1d2 f8c8 d1c1 c7b6 c2d3 h7h6 h2h3 b7a6 d3a6 b6a6 e2a6 a8a6 e4e5 f6e8 c3e4 c8c1 a1c1 b4d2 f3d2 a6b6');
INSERT INTO evaluations VALUES(53466,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(53466,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(53466,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(53466,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(53466,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(53466,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(53466,36,'cp 32','e7e6 d2d4 f8e7 f1c4 g8f6 e1g1 e8g8 f1e1 a7a6 d4d5 e6d5 c3d5 f6d5 d1d5 b8c6 d5e4 e7f6 c1d2 c8d7 a1d1 f8e8 e4f4 d7e6 d2c3 d8b8 f4b8 a8b8 c4e6 f7e6 d1d7 f6c3 b2c3 h7h6 g2g3 e8d8 d7c7 b8c8');
INSERT INTO evaluations VALUES(54479,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54479,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(54479,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(54479,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(54479,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(54479,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(54479,36,'cp 31','f8c5 d1c2 d8b6 e2e3 g8f6 f1e2 b8c6 a2a3 c5d6 b2b4 a7a5 b4b5 c6e5 b1d2 e8g8 c1b2 e5f3 d2f3 f6e4 e1g1 c8e6 f1c1 f8c8 c2d1 h7h6 a3a4 a8b8 e2f1 d6b4 h2h3 b4c3 b2c3 c8c3 c1c3 e4c3 d1e1 c3e4 f3d4 b8c8 a1c1 c8c1 e1c1');
INSERT INTO evaluations VALUES(54622,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(54622,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(54622,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(54622,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(54622,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(54622,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(54622,36,'cp 56','f1e2 g8f6 d1c2 a7a5 c1d2 f8d6 c4c5 d6c7 f2f4 b8d7 g1f3 e8g8 f3e5 f6e4 c3e4 f5e4 e1g1 d7e5 f4e5 c8d7 f1f8 d8f8 c2b3 a8b8 a1f1 f8e7 b3a3 b8a8 b2b4 a5a4 e2d1 h7h6 d2e1 e7g5 g1h1 d7e8 e1g3');
INSERT INTO evaluations VALUES(55940,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(55940,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(55940,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(55940,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55940,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(55940,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(55940,36,'cp -40','g8f6 d2d4 c7c5 d4c5 f8c5 f1d3 b8c6 e1g1 e8g8 h2h3 e6e5 e3e4 d5d4 c3e2 h7h6 a2a3 a7a5 c2c3 c8e6 c3d4 c5d4 f3d4 c6d4 c1e3 d4b3 a1b1 d8e7 d1e1 b3c5 e1c3 a8c8 e3c5 c8c5 c3d2 f8d8 d2e3 f6h5 f1c1 c5c1 b1c1');
INSERT INTO evaluations VALUES(48413,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(48413,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(48413,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(48413,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(48413,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(48413,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(48413,36,'cp -27','d2d4 c7c5 d4c5 b8c6 f1d3 f8c5 e1g1 e6e5 e3e4 d5d4 c3e2 e8g8 h2h3 c8e6 e2g3 h7h6 a2a3 f8e8 b2b4 c5d6 c1b2 f6d7 c2c3 d4c3 b2c3 a7a5 b4a5 d8c7 d3b5 e8d8 b5c6 c7c6 c3b4 d7c5 d1c2 c5b3 c2c6 b7c6 a1d1 c6c5 b4c3 b3a5');
INSERT INTO evaluations VALUES(1408,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(1408,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(1408,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(1408,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(1408,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(1408,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(1408,36,'cp 52','c3d4 e7e6 b1c3 d5d8 f1d3 g8f6 a2a3 f8e7 e1g1 e8g8 f1e1 b7b6 c1g5 h7h6 g5h4 c8b7 d3b1 g7g6 d1d3 f6h5 h4g3 h5g3 h2g3 e7f6 e1e6 c6a5 e6e1 b7f3 d3f3 f6d4 b1e4 a8c8 a1d1 a5c4 e4b7 c8c5');
INSERT INTO evaluations VALUES(2533,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(2533,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(2533,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(2533,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(2533,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(2533,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(2533,36,'cp -19','c4a2 b7b5 c3e2 d8c7 c2c3 f8d8 f1e1 c5c4 e2g3 c4d3 d1d3 c8b7 d3e2 c6e5 f3d4 h7h6 f2f4 e5d7 a2b3 e7f8 e1f1 a8c8 b3c2 g7g6 g1h1 d8e8 c1e3 e6e5 d4b3 f8g7 a1d1 d6d5 f4f5 g6g5 g3h5 f6h5 e2h5 d7f6');
INSERT INTO evaluations VALUES(3403,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(3403,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(3403,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(3403,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(3403,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(3403,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(3403,36,'cp 71','d4d5 c6b8 a2a4 b8d7 d1e2 a7a5 c1e3 f6g4 e3d2 d7b6 c4b3 c7c6 h2h3 g4f6 d2e3 c6d5 e4d5 c8f5 a1e1 a8c8 f3d2 c8c3 e3b6 d8b6 b2c3 f5g6 e2e3 b6e3 e1e3 e7d8 e3e1 d8b6 d2c4 b6c7 f2f3 f6d5 c4e5 d6e5 b3d5 g6c2 d5b7 f8d8 e1a1 c2d3 f1f2');
INSERT INTO evaluations VALUES(4762,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(4762,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(4762,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(4762,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(4762,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(4762,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(4762,36,'cp 134','d2d4 c5a7 d4e5 d6e5 d1d8 c6d8 f3e5 g8f6 f1e1 e8g8 c1f4 b7b5 c4d5 c8b7 d5b7 d8b7 b1d2 a7b6 f4e3 b6e3 e1e3 b7c5 e5c6 f8e8 a1e1 f6g4 e3g3 g4e5 c6e5 e8e5 b2b4 c5a4 f2f4 e5e6 d2b3 e6c6 e1e3 a8d8 e4e5 d8d1 g1f2 g7g6 g3f3 g8f8 f2g3 f8e7 g3h3 c6c4 g2g3 d1b1 h3g4');
INSERT INTO evaluations VALUES(4765,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(4765,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(4765,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(4765,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(4765,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(4765,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(4765,36,'cp 132','c5a7 d4e5 d6e5 d1d8 c6d8 f3e5 g8f6 f1e1 e8g8 c1f4 b7b5 c4d5 c8b7 d5b7 d8b7 b1d2 a7b6 f4e3 b6e3 e1e3 a8d8 e3e2 f8e8 e5c6 d8d6 c6d4 c7c5 d4f3 f6d5 a2a4 d5f4 e2e3 c5c4 a4b5 a6b5 g2g3 f4d3 b2b3 c4b3 d2b3 b7c5 b3c5 d3c5 a1a5 d6b6 e4e5 e8d8');
INSERT INTO evaluations VALUES(5636,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(5636,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5636,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(5636,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5636,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5636,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(5636,36,'cp 0','c6a5 c4d5 g8f6 b2b4 a5c6 d5c6 b7c6 a2a4 e8g8 b4b5 c6b5 a4b5 c8d7 c2c4 d8c8 b1c3 a7a6 c1e3 a6b5 c4b5 c8b7 a1b1 c7c6 b5c6 b7c6 b1c1 c6b7 c1b1');
INSERT INTO evaluations VALUES(5725,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(5725,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(5725,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5725,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5725,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(5725,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5725,36,'cp -10','g8f6 f1e1 a7a5 d2d3 e8g8 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 d1e2 f8f7 e1f1 d8d7 f3h2 a8f8 h2g4 f6g4 e2g4 c6b4 g4e2 b4a6 f1f7 f8f7 d3d4 d7e7 a2a3 e5d4 e3d4 e7f6 e4e5 d6e5 e2e5 f6e5 d4e5 a6c5 a1f1');
INSERT INTO evaluations VALUES(5796,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(5796,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(5796,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(5796,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(5796,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(5796,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(5796,36,'cp 31','e2e3 g8f6 f1d3 e8g8 e1g1 b4d6 f4g5 h7h6 g5h4 d6e7 a2a3 b7b6 h4g3 c8b7 f1e1 c6a5 e3e4 f6e4 c3e4 d5e4 d3e4 b7e4 e1e4 a5c4 d1e2 d8d5 f3e5 c4e5 g3e5 a8c8 a1e1 e7d6 c2c4 d5c6 b2b4 d6e5 e4e5');
INSERT INTO evaluations VALUES(8494,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(8494,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(8494,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(8494,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(8494,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(8494,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(8494,36,'cp 62','c2c3 e7d6 f1d3 e8g8 f3e5 f6d7 d2f3 f7f6 e5c6 b7c6 d1c2 h7h6 f3h4 f6f5 h4g6 f8f7 h2h4 d8f6 h4h5 e6e5 d4e5 d7e5 g6e5 d6e5 e1g1 c8e6 a1d1 f7e7 f4e5 f6e5 c3c4 a8e8 c4d5 c6d5 f1e1 a7a5 b2b3 e6f7 d3f5 f7h5');
INSERT INTO evaluations VALUES(9039,30,'cp 184',NULL);
INSERT INTO evaluations VALUES(9039,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(9039,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(9039,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(9039,34,'cp 199',NULL);
INSERT INTO evaluations VALUES(9039,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(9039,36,'cp 183','b4c3 b2c3 f6e4 f1e1 e4d6 c4d3 a7a6 a2a4 c8d7 d4d5 c6e7 d5e6 d7e6 d1c2 d8d7 f3e5 d7c8 c3c4 c7c5 e5f3 h7h6 c1f4 c8c7 a1d1 e8c8 f3d2 g7g5 f4e5 e7c6 e5h8 d8h8 c2c3 h8d8 h2h3 c6d4 d2b3 e6f5 d3f5 d6f5 d1d2 c8b8 b3d4 f5d4');
INSERT INTO evaluations VALUES(9052,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(9052,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(9052,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(9052,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(9052,34,'cp 189',NULL);
INSERT INTO evaluations VALUES(9052,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(9052,36,'cp 179','b2c3 f6e4 f1e1 e4d6 c4d3 e8g8 f3g5 g7g6 d1g4 h7h5 g4h4 d8f6 h4g3 f6g7 c1a3 b7b6 g5e4 c8b7 g3h4 d6e4 a3f8 g7f6 h4f6 e4f6 f8h6 a7a5 a1c1 a5a4 h6g5 f6e8 g5f4 h5h4 c1b1 a4a3 d3e4 c6a5 e4b7 a5b7 c3c4');
INSERT INTO evaluations VALUES(9050,30,'cp 165',NULL);
INSERT INTO evaluations VALUES(9050,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(9050,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(9050,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(9050,34,'cp 180',NULL);
INSERT INTO evaluations VALUES(9050,35,'cp 210',NULL);
INSERT INTO evaluations VALUES(9050,36,'cp 193','f6e4 f1e1 e4d6 c4d3 c8d7 a1b1 c6e7 c3c4 d7c6 f3g5 d8d7 c1b2 h7h6 d4d5 e6d5 b2g7 d5c4 g7f6 h6g5 e1e7 d7e7 f6e7 e8e7 d1e2 e7f8 d3c4 d6c4 e2c4 h8h6 c4c3 a8d8 h2h3 d8d5 c3a3 f8g7 b1e1 a7a5 a3e7 h6d6 e7c7 g7g6');
INSERT INTO evaluations VALUES(8934,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(8934,31,'cp 205',NULL);
INSERT INTO evaluations VALUES(8934,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(8934,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(8934,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(8934,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(8934,36,'cp 205','f1e1 e4d6 c4d3 c6e7 c3c4 f7f6 d1c2 h7h6 f3h4 e8g8 d4d5 e6e5 f2f4 f8e8 c1a3 c8d7 c4c5 e5e4 c5d6 e4d3 d6e7 d3c2 e7d8r a8d8 a3e7 d8c8 a1c1 g8f7 e7b4 c7c5 e1e8 c8e8 b4c5 b7b6 c5f2 e8e2 h4f3 d7f5 f3d4 e2d2 f2e3 d2d4 e3d4');
INSERT INTO evaluations VALUES(11254,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(11254,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11254,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(11254,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(11254,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(11254,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(11254,36,'cp 0','d2d4 c5d4 f3d4 f8e7 c1e3 e8g8 h2h3 c8d7 c4d3 d8c7 f2f4 c6d4 e3d4 d7c6 d1e2 c7a5 e2f3 f6d7 g1h1 e7f6 c3e2 a5b4 c2c3 b4b2 d4f6 d7f6 f1b1 b2d2 b1d1 d2b2');
INSERT INTO evaluations VALUES(14053,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14053,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14053,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14053,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14053,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14053,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14053,36,'cp 0','a7a6 b5c6 e7c6 e5g3 h7h5 h2h4 d5d4 g1e2 f8d6 g3d6 d8d6 e3d4 c6d4 e2d4 d6d4 c2c3 d4f6 d1f3 f6f3 g2f3 c8f5 b1a3 e8c8 a3c4 h8e8 c4e3 f5d3 e1c1 g7g6 h1h3 d8d6 d1e1 d6f6 e3c2 e8e1 c2e1 d3f1 h3h1 f1e2 h1h3');
INSERT INTO evaluations VALUES(11282,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11282,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11282,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11282,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11282,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11282,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11282,36,'cp 0','b5c6 e7c6 e5g3 h7h5 h2h4 d5d4 g1e2 f8d6 g3d6 d8d6 e3d4 c6d4 e2d4 d6d4 c2c3 d4f4 d1f3 f4f3 g2f3 c8f5 b1a3 e8c8 a3c4 h8e8 c4e3 f5d3 e1c1 g7g6 h1h3 e8e6 h3g3 a6a5 d1e1 a5a4 c1b2 e6d6 f3f4 d6f6 e3g2 f6b6 g2e3');
INSERT INTO evaluations VALUES(11280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11280,36,'cp 0','e7c6 e5g3 h7h5 h2h4 d5d4 g1e2 f8d6 g3d6 d8d6 e3d4 c6d4 e2d4 d6d4 c2c3 d4f4 d1f3 f4f3 g2f3 c8f5 b1a3 e8c8 a3c4 f5d3 e1c1 d3c4 b3c4 d8d3 h1h3 h8e8 c1c2 d3d6 d2d4 e8e2 d1d2 e2e1 c4c5 d6f6 d2d1 e1e2 d1d2');
INSERT INTO evaluations VALUES(11401,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11401,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11401,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11401,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11401,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11401,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11401,36,'cp 0','e5g3 h7h5 h2h4 d5d4 g1e2 f8d6 g3d6 d8d6 e3d4 c6d4 e2d4 d6d4 c2c3 d4f4 d1f3 f4f3 g2f3 c8f5 b1a3 e8c8 a3c4 f5d3 e1c1 d3c4 b3c4 d8d3 h1h3 h8e8 c1c2 d3d6 d2d4 e8e2 d1d2 e2e1 c4c5 d6f6 d2d1 e1e2 d1d2');
INSERT INTO evaluations VALUES(22590,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(22590,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(22590,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(22590,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(22590,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(22590,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(22590,36,'cp 49','d2d4 g8f6 a2a3 d7d5 b1c3 f8e7 e2e3 e8g8 b2b4 d5c4 f1c4 e7d6 e1g1 e6e5 h2h3 e5d4 e3d4 h7h6 f1e1 c6e7 b4b5 c7c6 b5c6 b7c6 f3e5 d6e5 d4e5 d8d1 e1d1 f6d5 c3e4 c8f5 e4d6 f5e6 f2f4 e7f5 c4d3 f5d6');
INSERT INTO evaluations VALUES(23371,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(23371,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(23371,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(23371,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(23371,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(23371,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(23371,36,'cp 70','g1f3 e7g6 h2h4 f8b4 b1d2 e5e4 h4h5 e4f3 h5g6 f7g6 e2f3 g8f6 a2a3 d8e7 d1e2 b4d6 e2e7 d6e7 f1d3 e8g8 d2e4 f6e4 f3e4 e7f6 a1b1 d7d6 e1e2 b7b6 f2f3 a7a5 b2b3 f6e5 e2f2 e5d4 f2g3 g6g5 a3a4 h7h6 c1b2 d4b2 b1b2 c8d7 g3f2');
INSERT INTO evaluations VALUES(29705,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(29705,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(29705,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(29705,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(29705,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(29705,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(29705,36,'cp 43','e7e6 d4d5 f8b4 d5e6 d8d1 e1d1 f5e6 h1e1 e8g8 c4e6 f7e6 c1d2 b4c5 d1e2 b8c6 e2f1 e6e5 c3e4 c5b6 c2c3 a8d8 d2g5 c6a5 a2a4 a7a6 g5f6 g7f6 a1d1 d8d1 e1d1 a5b3 d1d7 f6f5 e4g5 b3c5 d7h7 c5a4 c3c4 a4b2');
INSERT INTO evaluations VALUES(30197,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30197,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(30197,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(30197,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(30197,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(30197,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(30197,36,'cp 43','d4d5 f8b4 d5e6 d8d1 e1d1 f5e6 h1e1 b4c3 b2c3 b8c6 c4e6 f7e6 e1e6 e8d7 e6e3 a8e8 d1e2 f6d5 e3e8 h8e8 e2d3 d7c8 c1g5 h7h6 g5e3 d5e3 f2e3 e8d8 d3e2 d8d5 f3d4 d5c5 a1f1 c6d4 e3d4 c5c3 e2d2 c3c4 d2d3 c4a4 f1f8 c8d7 f8f7 d7c6 f7g7 a4a3 d3d2 a3a2 g7h7 a7a5');
INSERT INTO evaluations VALUES(30090,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(30090,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(30090,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(30090,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(30090,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(30090,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(30090,36,'cp -12','f5d3 d1d3 c7c5 g1f3 d8b6 d3e2 f8e7 a2a4 b8d7 a4a5 b6c6 b1d2 e8g8 e1g1 f8c8 h2h3 a7a6 d4c5 d7c5 f3d4 c6e8 c3c4 e7f8 d4f3 d5c4 d2c4 c5b3 c4b6 b3a1 f1a1 f6e4 f3e5 f8d6 b6a8 c8a8 a1c1');
INSERT INTO evaluations VALUES(30159,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(30159,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(30159,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(30159,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(30159,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(30159,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(30159,36,'cp 0','d1d3 c7c5 g1f3 b8d7 d3e2 f8e7 e1g1 e8g8 a2a4 h7h6 f3e5 c5d4 e3d4 d7e5 f4e5 f6d7 e5f4 d8b6 a4a5 b6c6 f1c1 f8c8 b1d2 b7b5 g2g3 a7a6 b2b4 c6b7 a1b1 c8c6 d2b3 c6c4 b3d2');
INSERT INTO evaluations VALUES(30158,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(30158,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(30158,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(30158,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(30158,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(30158,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(30158,36,'cp -11','c7c5 g1f3 d8b6 d3e2 f8d6 f4d6 b6d6 b1d2 e8g8 e1g1 d6b6 f1b1 b8d7 c3c4 c5d4 f3d4 f8c8 c4d5 f6d5 d2e4 d7c5 e4c5 c8c5 b1c1 h7h6 c1c5 b6c5 e2d2 a8c8 d4b3 c5b4 a1c1 c8c1 d2c1 b4a4 b3c5 a4c6 a2a3 c6c7 c5d3 c7c1 d3c1 g8f8 g1f1 f8e7 c1d3');
INSERT INTO evaluations VALUES(30565,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(30565,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(30565,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(30565,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(30565,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(30565,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(30565,36,'cp 91','g8f6 c1f4 f8e7 d1e2 a7a6 e1g1 c7c5 d4d5 e6d5 c3d5 e8g8 a1d1 b7d5 e4d5 b6b5 b2b3 b8d7 f1e1 f8e8 c2c4 b5c4 b3c4 f6h5 f4c1 g7g6 e2c2 e7f6 e1e8 d8e8 g2g4 h5g7 c1f4 d7e5 f3e5 f6e5 d1e1 e8d7 f4e5 d6e5 h2h3 h7h5 e1e4 h5g4 h3g4 g7e8 f2f3 d7e7 f3f4 e8f6');
INSERT INTO evaluations VALUES(28646,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(28646,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(28646,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(28646,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(28646,34,'cp 114',NULL);
INSERT INTO evaluations VALUES(28646,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(28646,36,'cp 112','d1e2 f8e7 a2a3 a7a6 e4e5 f6d5 c3e4 b6b5 e1g1 b8d7 f1e1 d6e5 d4e5 c7c5 b2b3 b5b4 e4d6 e7d6 e5d6 e8g8 e2e4 d7f6 e4e5 a8c8 f3g5 d5c3 f2f3 a6a5 a3b4 c5b4 g5e4 b7e4 f3e4 c8c6 c1f4 a5a4 f4g3 f6d7 e5d4 a4b3 c2b3');
INSERT INTO evaluations VALUES(31140,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(31140,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(31140,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31140,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(31140,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31140,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(31140,36,'cp 17','e5e6 g8h6 d2d4 d8d6 e6f7 h6f7 f1e2 e7e6 f3g5 h5g6 e1g1 f8e7 g5f7 g6f7 b1d2 e8g8 e2d3 e7f6 c2c3 e6e5 d4e5 f6e5 d2f3 b8d7 f1e1 a8e8 f3e5 d7e5 c1f4 d6f6 f4e5 e8e5 e1e5 f6e5 d1c2 h7h6 c2d2 a7a5 a1e1 e5f6 a2a4 f7h5');
INSERT INTO evaluations VALUES(32517,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(32517,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(32517,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(32517,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(32517,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(32517,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(32517,36,'cp -38','b5e2 g8f6 g1f3 f8d6 d2d4 h7h6 e1g1 e8g8 c2c4 d5c4 e2c4 b8c6 c1e3 a7a6 a2a3 c6e7 b1c3 b7b5 c4d3 d7f5 d3c2 a8c8 f1e1 c7c6 h2h3 f8e8 c2f5 e7f5 d1d3 d8d7 e3d2 c8d8 e1e8 d8e8 a1e1 e8e1 d2e1');
INSERT INTO evaluations VALUES(33008,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(33008,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(33008,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(33008,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(33008,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(33008,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(33008,36,'cp 52','g8f6 d2e3 c5b4 a2a3 b4a5 f3d2 b8c6 d2c4 a5c7 e3g5 c7b8 a3a4 d7d6 c4e3 e8g8 d1d2 a7a6 f1d3 b7b5 g5f6 g7f6 e3d5 b5a4 d5f6 e7f6 e1g1 c8d7 a1a4 c6e5 a4a3 e5d3 c2d3 b8b4 f1a1 f8b8 h2h3 b4b2 d2b2 b8b2 a3a6 a8a6 a1a6 b2d2');
INSERT INTO evaluations VALUES(33397,30,'cp -43',NULL);
INSERT INTO evaluations VALUES(33397,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(33397,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(33397,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(33397,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(33397,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(33397,36,'cp -25','e2e3 c7c5 d1c2 b8c6 f1c4 d5f5 c2f5 c8f5 d4c5 e8g8 e1g1 f5e4 f1d1 a8d8 d1d8 f8d8 g1f1 e4f5 f3g5 f5d3 c4d3 d8d3 a1b1 d3d5 g5f3 d5c5 f3e1 e7e5 f2f3 f7f5 b2b3 e5e4 c1b2 g7b2 b1b2 c5c1 b2c2 c1a1 f1f2 a7a5 f3e4 f5e4');
INSERT INTO evaluations VALUES(33396,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(33396,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(33396,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(33396,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(33396,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(33396,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(33396,36,'cp -24','c7c5 d1c2 b8c6 c2c5 d5c5 d4c5 e8g8 f1c4 c8f5 e1g1 a8c8 h2h3 c6b4 f3d4 c8c5 b2b3 c5c7 c1d2 b4c6 a1c1 f8d8 f1e1 h7h5 d4f5 d8d2 f5g7 g8g7 e1e2 d2d7 e2e1 a7a6 c4f1 c6b4 a2a4 b4a2 c1c7 d7c7 e1d1 a2c3 d1d4');
INSERT INTO evaluations VALUES(34168,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(34168,31,'cp 206',NULL);
INSERT INTO evaluations VALUES(34168,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(34168,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(34168,34,'cp 210',NULL);
INSERT INTO evaluations VALUES(34168,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(34168,36,'cp 198','b1c3 d5d8 g1f3 g8f6 e2e4 e7e6 f1d3 c8b7 e1g1 a7a6 d1e2 b8d7 e4e5 f6d5 c3d5 b7d5 d3a6 f8e7 a6c4 d5c4 e2c4 e8g8 c4c6 a8a7 c1e3 d8c8 a2a4 f8d8 f1c1 c8a8 h2h4 d7f8 h4h5 h7h6 f3h2 e7g5 f2f4 a8c6 c1c6');
INSERT INTO evaluations VALUES(34166,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(34166,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(34166,32,'cp 185',NULL);
INSERT INTO evaluations VALUES(34166,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(34166,34,'cp 181',NULL);
INSERT INTO evaluations VALUES(34166,35,'cp 206',NULL);
INSERT INTO evaluations VALUES(34166,36,'cp 205','d5d8 e2e4 e7e6 g1f3 g8f6 f1d3 c8b7 e1g1 a7a6 f1e1 f8e7 d1e2 b8d7 c1g5 h7h6 g5f4 g7g5 f4g3 h6h5 f3g5 h5h4 g3f4 f6h5 g5h3 h5f4 h3f4 e6e5 d4e5 d7e5 a1d1 e7d6 d3c2 d8d7 f4d5 e8c8 f2f4 h4h3 d5e3 e5g6 g2g3 g6f4 g3f4 d7e6 g1h1 d6f4 d1d8 h8d8');
INSERT INTO evaluations VALUES(51311,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(51311,31,'cp 208',NULL);
INSERT INTO evaluations VALUES(51311,32,'cp 197',NULL);
INSERT INTO evaluations VALUES(51311,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(51311,34,'cp 207',NULL);
INSERT INTO evaluations VALUES(51311,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(51311,36,'cp 193','e2e4 e7e6 g1f3 g8f6 f1d3 c8b7 e1g1 f8e7 d1e2 b8d7 f1e1 a7a6 a2a3 b6b5 c1f4 e8g8 b2b4 f8e8 h2h4 d7f8 a1d1 f8g6 f4d2 d8d7 e4e5 f6d5 c3d5 b7d5 h4h5 g6h4 f3h4 e7h4 e2g4 h4d8 e1e3 f7f5 e5f6 d8f6 e3g3 e8f8 d2f4 d7f7 f4e5 f6e5 d4e5');
INSERT INTO evaluations VALUES(34318,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(34318,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(34318,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(34318,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34318,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(34318,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(34318,36,'cp 73','d2d4 g8f6 g1f3 c8f5 c1d2 b8d7 d1e2 a5c7 e1c1 e7e6 f3h4 f5g6 h4g6 h7g6 h2h4 f8d6 d2g5 b7b5 c4b3 e8g8 c3e4 a7a5 c2c3 f6e4 e2e4 d7f6 g5f6 g7f6 c1b1 a5a4 b3c2 a4a3 h4h5 f6f5 e4e2 g6g5 g2g4 a3b2 g4f5 e6f5 c2f5 g8h8');
INSERT INTO evaluations VALUES(35202,30,'cp 1255',NULL);
INSERT INTO evaluations VALUES(35202,31,'cp 1254',NULL);
INSERT INTO evaluations VALUES(35202,32,'cp 1251',NULL);
INSERT INTO evaluations VALUES(35202,33,'cp 1248',NULL);
INSERT INTO evaluations VALUES(35202,34,'cp 1248',NULL);
INSERT INTO evaluations VALUES(35202,35,'cp 1248',NULL);
INSERT INTO evaluations VALUES(35202,36,'cp 1255','f3h4 g8f6 b1c3 f8c5 h4f5 e8g8 f5e3 d7d6 g2g3 b8c6 f1g2 c5e3 d2e3 c8e6 e1g1 a8c8 b2b3 a7a6 a2a4 f6d7 f2f3 f7f6 c1a3 h7h5 d1e2 g7g6 a1b1 f8f7 e2d2 c8d8 f1e1 f7e7 h2h3 e6f7 g3g4 h5g4');
INSERT INTO evaluations VALUES(35193,30,'cp 1206',NULL);
INSERT INTO evaluations VALUES(35193,31,'cp 1201',NULL);
INSERT INTO evaluations VALUES(35193,32,'cp 1201',NULL);
INSERT INTO evaluations VALUES(35193,33,'cp 1201',NULL);
INSERT INTO evaluations VALUES(35193,34,'cp 1201',NULL);
INSERT INTO evaluations VALUES(35193,35,'cp 1205',NULL);
INSERT INTO evaluations VALUES(35193,36,'cp 1201','g8f6 b1c3 b8c6 h4f5 d7d6 f5e3 c8e6 g2g3 g7g6 f1g2 h7h5 h2h3 f8h6 d2d3 e8c8 a1b1 c6d4 e1g1 h5h4 g3g4 c8b8 a2a4 h6g5 c1d2 a7a6 d1c1 h8e8 b2b4 e8e7 c1b2 e7d7 f2f3 d8e8 b1e1 e8e7 a4a5 g5e3 d2e3');
INSERT INTO evaluations VALUES(36423,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(36423,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(36423,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(36423,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(36423,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(36423,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(36423,36,'cp 38','c2c4 e4f6 b1c3 b8c6 h2h3 c6b4 c4c5 b4d3 d1d3 c7c6 c1f4 f8e8 f1e1 c8e6 e1e3 f6h5 f4h2 e7g5 f3g5 d8g5 e3e5 g5g6 d3g6 h7g6 a2a4 b7b6 c5b6 a7b6 e5e3 e8c8 g2g4 h5f6 h2f4 c6c5 d4c5');
INSERT INTO evaluations VALUES(36422,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(36422,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(36422,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(36422,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(36422,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(36422,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(36422,36,'cp 34','e4f6 b1c3 b8c6 h2h3 c6b4 c4c5 b4d3 d1d3 c7c6 c1f4 f8e8 a1e1 c8d7 f3e5 d7e6 e1e2 f6d7 e5f3 d7f8 f1e1 f8g6 f4g3 d8d7 e2e3 b7b5 a2a3 e7f8 d3e2 h7h6 f3e5 g6e5 e3e5 a7a6 e5e3 a6a5 e2f3 b5b4');
INSERT INTO evaluations VALUES(36695,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(36695,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(36695,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(36695,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(36695,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(36695,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(36695,36,'cp 74','b1c3 b8c6 e1g1 c8g4 e4e5 f6d7 c1e3 c6b4 d3e4 c7c5 d4c5 d7c5 e3c5 d6c5 h2h3 d8d1 a1d1 g4f3 f1f3 f8d8 f3f1 a8b8 a2a3 b4c6 c3b5 c6d4 b5d4 c5d4 g1f2 e7e6 f2f3 b8c8 e4d3 g7f8 g2g4 f8e7 h3h4 g8g7 h4h5 g6h5 g4h5 a7a5 h5h6 g7h6');
INSERT INTO evaluations VALUES(37135,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(37135,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(37135,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37135,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(37135,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(37135,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37135,36,'cp 0','g5h4 c7c5 e2e3 c5d4 e3d4 d8a5 f1d3 d7d5 h4f6 b4c3 b2c3 d5c4 d3b1 a5c3 f3d2 g7f6 d1g4 g8h8 g4h4 h8g7 h4g4');
INSERT INTO evaluations VALUES(37138,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37138,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37138,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37138,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37138,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37138,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37138,36,'cp 6','c7c5 e2e3 c5d4 f3d4 d7d5 c4d5 d8a5 h4f6 b4c3 b2c3 a5c3 e1e2 g7f6 d5e6 c8e6 d4e6 c3c4 e2f3 c4e6 d1d4 b8d7 d4g4 e6g4 f3g4 a8d8 a1c1 d7e5 g4g3 d8d2 f2f4 e5d3 f1d3 d2d3 h1e1 f8e8 g3f3 b7b6 e1e2 f6f5 c1c7 d3a3 h2h3 h6h5 g2g4 f5g4 h3g4 h5g4 f3g4 a3e3 e2e3 e8e3 c7a7 g8g7 a7a6 e3c3 a2a4 c3a3 a6b6 a3a4');
INSERT INTO evaluations VALUES(38714,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(38714,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(38714,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(38714,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38714,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(38714,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(38714,36,'cp 21','b8c6 f1g2 f7f5 e1g1 g8f6 e2e3 c5b6 d2d3 a7a5 a1b1 e8g8 a2a3 d8e8 b2b4 h7h6 b4b5 c6e7 a3a4 c8e6 c1a3 e7g6 f3d2 f5f4 g2b7 f4e3 f2e3 b6e3 g1h1 f6g4 f1f8 e8f8 c3e4 a8b8 b7d5 g4f2 e4f2 e6d5 c4d5 e3f2 d2e4 f2b6 d1h5');
INSERT INTO evaluations VALUES(6001,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(6001,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(6001,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(6001,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(6001,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(6001,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(6001,36,'cp 31','f1g2 g8f6 e1g1 e8g8 e2e3 c5b6 d2d3 c8e6 b2b3 h7h6 c1b2 d8c8 d3d4 e6g4 d4d5 c6e7 d1d3 c8d7 f3h4 e7f5 h4f5 g4f5 e3e4 f5h3 c3a4 d7g4 a4b6 h3g2 g1g2 a7b6 f2f3 g4g5 d3e2');
INSERT INTO evaluations VALUES(39243,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(39243,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39243,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39243,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(39243,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39243,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39243,36,'cp 0','b8c6 g1f3 d7d5 d1a4 d5e4 d3e4 c8d7 a4c2 a7a5 f1b5 g8e7 e1g1 e7g6 b1d2 e8g8 d2c4 f7f6 h2h3 d8e7 c4e3 d7e6 f1d1 e7f7 e3f5 g8h8 b2b3 c6e7 c1e3 c5e3 f5e3 g6f4 b5f1 a5a4 c3c4 f7h5 e3d5');
INSERT INTO evaluations VALUES(42895,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(42895,31,'cp 286',NULL);
INSERT INTO evaluations VALUES(42895,32,'cp 279',NULL);
INSERT INTO evaluations VALUES(42895,33,'cp 301',NULL);
INSERT INTO evaluations VALUES(42895,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(42895,35,'cp 304',NULL);
INSERT INTO evaluations VALUES(42895,36,'cp 280','d1g4 d7d6 g4g7 d8f6 g7f6 e4f6 e5d3 c8f5 f2f3 h7h5 e1f2 b8d7 c1d2 f6d5 a1c1 d5e7 h2h4 f7f6 d3f4 d7e5 c3c4 e8f7 b2b3 e5g6 f4d5 e7d5 c4d5 a7a5 f1e2 a5a4 b3b4 g6e5 a2a3 h8e8 g2g3 f5d7 h1e1 d7f5 d2c3 f5d7 c3b2 e8g8 c2c4');
INSERT INTO evaluations VALUES(45175,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(45175,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(45175,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(45175,33,'cp 193',NULL);
INSERT INTO evaluations VALUES(45175,34,'cp 193',NULL);
INSERT INTO evaluations VALUES(45175,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(45175,36,'cp 187','d4c5 a7a6 c1e3 f6g4 e3g5 d8a5 f1e2 h7h5 h2h3 g4f6 g5f6 e7f6 c5d6 c8e6 a2a3 b8c6 g1f3 a5d8 d6d7 d8d7 c3d5 f8c5 b2b4 c5a7 c2c4 f6f5 d1c2 e8g8 a1d1 f5e4 c2e4 e6d5 d1d5 d7e8 e4c2 a8d8 d5d8 e8d8 c4c5 a6a5 b4b5 c6d4 f3d4 d8d4 b5b6 a7b8 e1g1 f8c8 f1d1 c8c5 d1d4 c5c2 d4d8 g8h7 d8b8');
INSERT INTO evaluations VALUES(45113,30,'cp 190',NULL);
INSERT INTO evaluations VALUES(45113,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(45113,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(45113,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(45113,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(45113,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(45113,36,'cp 206','a7a6 c1e3 f6g4 e3g5 d8a5 f1e2 h7h5 h2h3 g4f6 g5f6 e7f6 c5d6 c8e6 g1f3 b8c6 a2a3 a5d8 d1d3 d8d6 c3d5 e6d5 e4d5 c6e5 f3e5 d6e5 e1c1 f8d6 g2g3 h5h4 f2f4 e5e7 e2g4 e8g8 g3h4 d6f4 c1b1 e7d6 h1f1 f8e8 h4h5 a8d8 g4f5 b7b5 d3f3 f4e5 c2c3');
INSERT INTO evaluations VALUES(45595,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(45595,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(45595,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(45595,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(45595,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(45595,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(45595,36,'cp -16','b7b6 b2b4 a7a5 a2a3 e6e5 e2e3 h7h6 g5h4 e5d4 e3d4 f8e7 f1d3 f6h5 d1h5 e7h4 c3a4 a5b4 a4b6 a8a3 a1a3 b4a3 g1f3 h4f6 e1g1 b8d7 f1e1 e8f8 f3e5 d7e5 d4e5 f6e7 b6c8 d8c8 h5f3 c8e8 e5e6 e7f6 d3b1 f7e6 f3a3');
INSERT INTO evaluations VALUES(46198,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(46198,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46198,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(46198,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(46198,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46198,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46198,36,'cp -9','c5d4 c3d4 d8b6 d1b3 b6b3 d2b3 b8c6 a2a3 f6e4 e2e3 f8d6 f4d6 e4d6 f1d3 f7f6 a1c1 c8d7 b3c5 e8e7 c5d7 e7d7 h2h4 a8c8 e1e2 h7h5 c1c5 b7b6 c5c1 c6e7 d3a6 c8c1 h1c1 h8b8 a6d3 b8c8 c1c8 e7c8 f3e1');
INSERT INTO evaluations VALUES(46398,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(46398,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(46398,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(46398,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(46398,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46398,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(46398,36,'cp -5','f8c5 e2e3 e8g8 a2a3 a7a5 h2h3 b8c6 f1b5 d5d4 f3d4 c6d4 e3d4 c5d4 e1g1 d8b6 d1e2 f8d8 c3a4 b6a7 f1e1 c8d7 b5d7 d8d7 c1e3 d4e3 e2e3 a7e3 e1e3 d7d4 b2b3 a8d8 a4b6 g8f8 a1c1 d4d3 e3d3 d8d3');
INSERT INTO evaluations VALUES(46615,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(46615,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(46615,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(46615,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(46615,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(46615,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(46615,36,'cp 77','e5d6 c7d6 b1c3 f8g7 c1e3 e8g8 a1c1 c8g4 f1e2 e7e6 h2h3 g4f3 e2f3 b8c6 b2b3 d6d5 c4d5 b6d5 c3d5 e6d5 e1g1 f8e8 d1d3 c6e7 g2g4 h7h6 d3b5 a8b8 f1d1 a7a6 b5d3 b8c8 c1c8 d8c8 d3d2 g8h7 d1c1 c8d7');
INSERT INTO evaluations VALUES(47923,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(47923,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(47923,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(47923,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(47923,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(47923,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(47923,36,'cp -35','f8e7 h2h3 e8g8 e1g1 b8c6 c1e3 a7a6 a1c1 h7h6 f1e1 c8e6 c3d5 f8e8 e3d2 e7f8 d2c3 f6d7 c4d3 e6d5 e4d5 e8e1 d1e1 c6e5 f3e5 d7e5 c3e5 d6e5 e1e5 f8d6 e5e4 g7g6 c1c2 d8e8 g2g3 e8e4 d3e4 g8g7 h3h4 a6a5 g1g2 a5a4 h4h5 a4a3 h5g6');
INSERT INTO evaluations VALUES(40817,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(40817,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(40817,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(40817,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(40817,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(40817,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(40817,36,'cp -28','h2h3 e8g8 e1g1 b8c6 c1e3 h7h6 a1c1 a7a6 f1e1 c8e6 c3d5 f8e8 e3d2 f6d7 d2c3 e7f8 c4d3 e6d5 e4d5 e8e1 d1e1 c6e5 f3e5 d6e5 c3e5 d7e5 e1e5 f8d6 e5d4 d8d7 c1e1 a8e8 e1e8 d7e8 g2g3 g7g6 g1g2 e8e1 d4e3 e1b4 e3e8 g8g7 b2b3 b4c5');
INSERT INTO evaluations VALUES(48525,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(48525,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(48525,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(48525,33,'cp 166',NULL);
INSERT INTO evaluations VALUES(48525,34,'cp 143',NULL);
INSERT INTO evaluations VALUES(48525,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(48525,36,'cp 113','b1c3 c7c5 c1f4 c5d4 f3d4 a7a6 f1e2 f8c5 d4b3 c5e7 d1d8 e7d8 e1g1 b8c6 f1d1 e8g8 f4e3 d8e7 a2a3 f8d8 d1d8 e7d8 f2f3 b7b5 a1d1 d8e7 a3a4 b5a4 c3a4 a8b8 b3c5 a6a5 g1f1 e6e5 c5a6 c8a6 e2a6 c6d4 e3d4 e5d4 d1d4 g8f8');
INSERT INTO evaluations VALUES(48958,30,'cp -103',NULL);
INSERT INTO evaluations VALUES(48958,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(48958,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(48958,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(48958,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(48958,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(48958,36,'cp -94','c7c6 b5e2 f8d6 e5f3 e8g8 d2d3 e4f6 e1g1 c8f5 d3d4 f8e8 c2c4 b8d7 b1c3 d5c4 e2c4 d7b6 c4d3 f5g4 c1g5 h7h6 g5f6 d8f6 c3e4 f6f4 g2g3 f4f3 e4d6 f3d1 a1d1 e8e6 d3f5 e6d6 f5g4 a8d8 f1e1 d6d4 d1d4 d8d4 e1e8 g8h7');
INSERT INTO evaluations VALUES(45831,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(45831,31,'cp -95',NULL);
INSERT INTO evaluations VALUES(45831,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(45831,33,'cp -97',NULL);
INSERT INTO evaluations VALUES(45831,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(45831,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(45831,36,'cp -79','b5e2 f8d6 e5f3 e4g5 e1g1 g5f3 e2f3 e8g8 d2d4 c8f5 f1e1 b8d7 b1d2 h7h6 d2f1 f5h7 a2a4 d7f6 a4a5 d8d7 f1g3 a7a6 c2c3 d7c7 g3f1 a8d8 g2g3 f8e8 e1e8 d8e8 f3g2 c7d7 f1e3 d7e6 d1b3 h7d3 b3b7 c6c5 g2d5 f6d5 b7d5');
INSERT INTO evaluations VALUES(49541,30,'cp -95',NULL);
INSERT INTO evaluations VALUES(49541,31,'cp -81',NULL);
INSERT INTO evaluations VALUES(49541,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(49541,33,'cp -109',NULL);
INSERT INTO evaluations VALUES(49541,34,'cp -126',NULL);
INSERT INTO evaluations VALUES(49541,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(49541,36,'cp -104','f6e4 d2d3 d8h4 g2g3 e4g3 g1f3 h4h3 h1g1 g3f5 c4f7 e8d8 f4e5 f8e7 d1e2 h8f8 f7d5 h7h6 b1c3 b8c6 e2g2 h3h5 g2g4 h5g4 g1g4 d7d6 e5d6 c7d6 g4g1 g7g5 h2h4 g5g4 g1g4 f5e3 c1e3 c8g4 f3h2 c6b4 e1d2 b4d5 c3d5 g4e6 d5e7 d8e7 a1e1 h6h5 h2f1 f8f7 f1g3 e6g4 e1g1 e7d7 c2c4 a8f8 g3e4 d7c7 a2a3 f8e8');
INSERT INTO evaluations VALUES(50408,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(50408,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50408,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(50408,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50408,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(50408,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(50408,36,'cp -17','e7e6 b1c3 d5b4 f3d1 c7c5 g1f3 d7d6 a2a3 b4c6 e5d6 f8d6 d2d4 c5d4 f3d4 d6e5 d4b3 e5c3 b2c3 d8c7 d1h5 b8d7 c3c4 e8g8 f1d3 f7f5 e1g1 d7f6 h5f3 c7f7 c1b2 e6e5 d3f5 c8f5 f3f5 f7c4 b3d2 c4d5 a1d1 a8d8 f2f4 e5f4 f1f4 b7b5 b2f6 g7f6 f5e4 d5e4 f4e4 g8g7');
INSERT INTO evaluations VALUES(49329,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49329,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49329,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49329,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(49329,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49329,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(49329,36,'cp -10','b1c3 d5b4 f3d1 c7c5 g1f3 d7d6 e5d6 f8d6 a2a3 b4c6 g2g3 d6e7 f1g2 e8g8 e1g1 b7b6 d2d4 c8a6 c1e3 a6f1 d1f1 c6d4 e3d4 c5d4 a1d1 b8c6 f3d4 c6d4 g2a8 d8a8 d1d4 f8d8 d4d8 a8d8 f1e2 d8c8 c3e4 c8d7 h2h3 d7c8 c2c3 c8d7 e2c4 b6b5');
INSERT INTO evaluations VALUES(51001,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(51001,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51001,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(51001,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(51001,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(51001,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(51001,36,'cp 16','d1e2 f8e7 e4e5 d7d5 c4b3 c7c5 c2c3 c8d7 d2d4 d8b6 c1e3 b8c6 d4c5 e7c5 e3c5 b6c5 b3c2 c5b5 c2d3 b5b6 e1g1 g8e7 b2b4 e8g8 b1d2 e7g6 d3g6 f7g6 d2b3 b6c7 b3c5 g6g5 e2e3 g5g4 f3d4 c6e5 c5e6 d7e6 d4e6');
INSERT INTO evaluations VALUES(51084,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(51084,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(51084,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(51084,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(51084,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(51084,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(51084,36,'cp 16','b4b5 c7c6 e2e3 g8f6 g1f3 e7e6 b5c6 b7c6 c2c4 d7d5 f1e2 c8a6 c1b2 f8b4 e1g1 e8g8 b1c3 b8d7 c4d5 c6d5 e2a6 a8a6 a2a3 b4c5 a3a4 d8b8 c3b5 f8c8 a1b1 c5b4 b1c1 c8c1 d1c1 b8b7 c1a1 a6c6 f1b1 c6c2 b2c3 b4c3 b5c3');
INSERT INTO evaluations VALUES(51098,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(51098,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(51098,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(51098,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(51098,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(51098,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(51098,36,'cp 0','c7c6 e2e3 e7e6 b5c6 b7c6 c2c4 g8f6 g1f3 f8e7 f1e2 d7d5 e1g1 c6c5 d2d4 e8g8 c1b2 c5d4 f3d4 a5a4 c4d5 f6d5 b1a3 c8b7 a1c1 d5b4 e2f3 b7f3 d1f3 d8d5 f3g4 d5g5 g4e4 g5d5');
INSERT INTO evaluations VALUES(51302,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(51302,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(51302,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(51302,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(51302,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(51302,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(51302,36,'cp 26','f1g2 c8b7 d2d3 g8f6 e1g1 e7e6 c2c4 f8e7 c4d5 f6d5 d3d4 c7c5 d4c5 b6c5 d1b3 d8c8 c1g5 d5f6 b1d2 e8g8 a1c1 h7h6 g5f6 e7f6 b3c2 f6e7 f3e5 b7g2 g1g2 b8c6 e5c6 c8c6 g2g1 f8d8 d2c4 d8d4 b2b3 a8d8 e2e3 d4d5 f1d1 d5d1 c1d1 d8d1 c2d1 c6b7');
INSERT INTO evaluations VALUES(51303,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(51303,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(51303,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51303,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(51303,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(51303,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(51303,36,'cp 22','c8b7 c2c4 d5c4 b1a3 b8d7 e1g1 e7e6 a3c4 f8e7 a1b1 c7c5 c4e5 e7f6 d1a4 f6e5 f3e5 b7g2 g1g2 d8c7 e5f3 g8f6 b2b4 e8g8 c1b2 c5b4 f1c1 c7b7 a4b4 f8c8 d2d3 h7h6 e2e4 c8c1 b1c1 a8c8 c1c8 b7c8 b4e7 c8c2 e7d8 g8h7 b2f6 d7f6 f3e5');
INSERT INTO evaluations VALUES(52608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52608,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52608,36,'cp 2','b8c6 e1g1 f8g7 d2d3 e7e6 c2c3 g8e7 c1e3 e8g8 b1d2 c6a5 d1e2 a5c4 d3c4 b7b6 e4e5 c8b7 e3f2 h7h6 a1e1 a8b8 d2e4 b7e4 e2e4 d6e5 f4e5 d8c7 f2h4 b6b5 c4b5 g6g5 h4g3 b8b5 b2b3 e7g6 e4c4 c7b7 c4e2 b5b6 f3d2');
INSERT INTO evaluations VALUES(52926,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(52926,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(52926,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(52926,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(52926,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(52926,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(52926,36,'cp -55','d3e4 g8f6 e4f3 e7e5 b1c3 c8e6 g1e2 f8c5 d2d3 e8g8 e1g1 b8d7 c1e3 c5e3 f2e3 f6e8 d1d2 f7f5 a1d1 d8c7 e3e4 f5f4 d3d4 e8d6 b2b3 a8d8 c3a4 d6f7 d4e5 f7e5 e2d4 e5f3 d4f3 g8h8 d2f2 b7b6 c2c4');
INSERT INTO evaluations VALUES(54687,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(54687,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(54687,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(54687,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(54687,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(54687,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(54687,36,'cp 59','g8f6 c2c4 g7g6 b1c3 f8g7 g1f3 b8c6 f1e2 e8g8 f4g3 e7e6 e1g1 g8h8 a1c1 d8e7 h2h3 a7a6 g3h2 b7b6 d1a4 c8b7 b2b4 c6d8 c4c5 d8f7 a4b3 g6g5 c5d6 c7d6 b4b5 g5g4 h3g4 f5g4');
INSERT INTO evaluations VALUES(54950,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(54950,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54950,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54950,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(54950,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(54950,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(54950,36,'cp 0','h7h6 g5h4 c7c5 d4c5 g7g5 h4g3 d8a5 c2c3 a5c5 e2e4 g8f6 b1d2 c8d7 h2h4 g5g4 g1e2 b8c6 g3f4 c6e5 e2g3 c5c7 d2b3 e7e6 d1d4 d7c6 f1e2 e5d7 d4d3 d7e5');
INSERT INTO evaluations VALUES(55049,30,'cp -208',NULL);
INSERT INTO evaluations VALUES(55049,31,'cp -214',NULL);
INSERT INTO evaluations VALUES(55049,32,'cp -207',NULL);
INSERT INTO evaluations VALUES(55049,33,'cp -217',NULL);
INSERT INTO evaluations VALUES(55049,34,'cp -203',NULL);
INSERT INTO evaluations VALUES(55049,35,'cp -198',NULL);
INSERT INTO evaluations VALUES(55049,36,'cp -207','c8g4 c2c4 e7e6 c4d5 e6d5 b1c3 f8b4 g1h3 g8f6 h1g1 d8d7 h3f4 e8g8 f4d3 b4c3 b2c3 g4f5 d1d2 f5g6 f2f3 f8e8 e1f2 a7a5 d2g5 h7h6 g5g2 f6h5 d3e5 e8e5 d4e5 b8c6 g2h3 d7e7 c1h6 c6e5 h6e3 a8e8 e3d4 e5c4');
INSERT INTO evaluations VALUES(56786,30,'cp 276',NULL);
INSERT INTO evaluations VALUES(56786,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(56786,32,'cp 289',NULL);
INSERT INTO evaluations VALUES(56786,33,'cp 280',NULL);
INSERT INTO evaluations VALUES(56786,34,'cp 288',NULL);
INSERT INTO evaluations VALUES(56786,35,'cp 293',NULL);
INSERT INTO evaluations VALUES(56786,36,'cp 298','g1f3 g8f6 f1c4 b8c6 d1e2 f8d6 d2d4 e8g8 e1g1 a7a6 b1d2 d8e8 f1e1 b7b5 c4b3 g8h8 c2c3 c8b7 b3c2 e8h5 d2e4 f6e4 c2e4 a8e8 c1d2 b7a8 a1c1 e6e5 e4c6 a8c6 f3e5 h5e2 e1e2 h8g8 c1e1');
INSERT INTO evaluations VALUES(245,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(245,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(245,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(245,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(245,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(245,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(245,36,'cp -22','h4e4 f1e2 b4c3 b2c3 e4c6 e1g1 g8f6 c1b2 e8g8 c3c4 f8e8 f1e1 d7d5 b2f6 c6f6 c4d5 c8d7 d1d2 b7b6 c2c4 g7g6 a1c1 a8d8 h2h3 h7h5 c4c5 d7c6 c5b6 c7b6 e2f3 e8e1 c1e1 d8d6 e1c1 c6d7 h3h4 b6b5 c1c7 f6h4 c7a7 h4f6 a7a8 g8h7');
INSERT INTO evaluations VALUES(246,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(246,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(246,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(246,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(246,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(246,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(246,36,'cp -23','f1e2 b4c3 b2c3 e4c6 e1g1 g8f6 c1b2 e8g8 c3c4 f8e8 f1e1 d7d5 b2f6 c6f6 c4d5 c8d7 c2c4 e8e5 d1d2 a8e8 e2f1 b7b6 e1e5 e8e5 d2c3 f6d6 g2g3 e5e8 a1c1 a7a5 h2h4 h7h5 c3d4 d6a3 d4d2 g7g6 c1e1 a5a4 e1e8 d7e8 f1g2 a3c5 d2d3 e8d7');
INSERT INTO evaluations VALUES(463,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(463,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(463,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(463,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(463,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(463,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(463,36,'cp 39','e1g1 d7d6 h2h3 c8d7 f1e1 e8g8 c4b3 a8c8 a2a4 c6d4 e3d4 d7c6 c3d5 c6d5 e4d5 f8e8 c2c3 f6d7 d4e3 c8c7 g2g3 d7e5 g1g2 g7f6 e3h6 f6g7 h6g5 a5a6 h3h4 h7h5 e1e4 e5g4 e4b4 g4e5 d1d2 e5d3');
INSERT INTO evaluations VALUES(462,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(462,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(462,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(462,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(462,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(462,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(462,36,'cp 31','d7d6 h2h3 c8d7 f1e1 e8g8 c4b3 a8c8 a2a4 f8d8 d4f3 a5h5 d1d2 h7h6 f3d4 c6a5 a1d1 a5b3 c2b3 g8h7 f2f3 h5a5 d4c2 a5c7 d1c1 c7b8 a4a5 d7e6 c2d4 d6d5 d4e6 f7e6 e4d5 f6d5 c3d5 c8c1 e1c1');
INSERT INTO evaluations VALUES(1380,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(1380,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(1380,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(1380,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(1380,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(1380,35,'cp 128',NULL);
INSERT INTO evaluations VALUES(1380,36,'cp 113','e7e5 f4e5 d7d6 e5f4 a7a6 b1c3 c8e6 f1e2 f8e7 e1c1 e8c8 a2a4 g7g5 f4e3 e7f6 h2h4 g5g4 c3d5 f6e5 d5b6 c8c7 b6c4 e6c4 e2c4 g8f6 f2f4 g4f3 g2f3 f6h5 h1g1 h5f4 g1g5 f4g6 g5h5 g6f4 h5f5 f7f6');
INSERT INTO evaluations VALUES(1340,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(1340,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(1340,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(1340,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(1340,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(1340,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(1340,36,'cp 141','f4e5 d7d6 e5f4 a7a6 b1c3 c8e6 e1c1 e8c8 f1e2 f8e7 a2a4 g7g5 f4e3 e7f6 c3d5 e6d5 d1d5 g8e7 d5d1 h7h5 c2c3 c8c7 e2c4 h8f8 c1c2 d8e8 f2f3 e7f5 e3c5 d6c5 e4f5 e8e5 d1d2 h5h4 h1d1 f8d8 c4f7 d8d2 d1d2 e5f5 f7e6 f5e5');
INSERT INTO evaluations VALUES(1336,30,'cp 133',NULL);
INSERT INTO evaluations VALUES(1336,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(1336,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(1336,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(1336,34,'cp 141',NULL);
INSERT INTO evaluations VALUES(1336,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(1336,36,'cp 132','d7d6 e5f4 a7a6 b1c3 c8e6 f1e2 f8e7 e1c1 e8c8 a2a4 g7g5 f4e3 g5g4 c3d5 h7h5 h2h3 g4h3 g2h3 e7f8 h3h4 f8h6 h1g1 e6d5 d1d5 h6e3 f2e3 g8e7 d5h5 e7c6 h5h8 d8h8 h4h5 c8d8 g1g5 d8e7 c1d2 e7f6 g5f5 f6e7 e2c4 c6e5 d2c3 a6a5 c4d5');
INSERT INTO evaluations VALUES(3273,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(3273,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(3273,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(3273,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(3273,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(3273,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(3273,36,'cp 94','c8e6 c1b1 f6d7 c3d5 h7h5 d4g7 g8g7 d5f4 d8c7 f1b5 d7f6 f4e6 f7e6 b5d3 e6e5 h2h4 f8f7 d2e2 a8c8 g2g4 e7e6 c2c3 c7e7 a2a3 a7a6 d3c2 b7b5 e2d2 h5g4 f3g4 f6g4 h4h5 g6g5');
INSERT INTO evaluations VALUES(23685,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(23685,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(23685,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(23685,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(23685,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(23685,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(23685,36,'cp 138','c1b1 f6d7 c3d5 h7h5 h2h4 a8c8 d4g7 g8g7 d2d4 d7f6 d4a7 e6d5 e4d5 c8c5 c2c4 b7b5 c4b5 f6d5 f1d3 d5b4 h1e1 e7e6 e1e6 b4d3 d1d3 d8c8 e6e1 f8e8 e1d1 c8f5 a7a4 c5b5 g2g4 f5e5 b2b3 e8b8 g4h5 b5b4 a4c6 e5h5 c6c7 h5c5 c7d6 c5d6 d3d6 b4h4 b1c2 f7f5');
INSERT INTO evaluations VALUES(23684,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(23684,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(23684,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(23684,33,'cp 122',NULL);
INSERT INTO evaluations VALUES(23684,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(23684,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(23684,36,'cp 125','f6d7 h2h4 g7d4 d2d4 h7h5 c3d5 d7e5 f3f4 e5g4 f1e2 e6d5 e2g4 d5c6 g4f3 d8a5 d4d3 g8g7 h1g1 f7f5 g1e1 a8e8 a2a3 a5c5 g2g3 f8f6 e4f5 c5f5 d3f5 g6f5 f3h5 e8h8 h5e2 f6g6 e1g1 c6e4 h4h5 g6h6 g3g4 f5g4 e2g4 d6d5 g4f3 g7f7 f3e4 d5e4');
INSERT INTO evaluations VALUES(3780,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(3780,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(3780,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(3780,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(3780,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(3780,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(3780,36,'cp 22','c5b6 a2a4 a7a5 d2d3 c6e7 d3d4 e7g6 c1e3 f6e4 d4e5 b6e3 e1e3 d6d5 b1d2 e4c5 b2b4 a5b4 c3b4 c5e6 d2b3 c8d7 b5d7 d8d7 a4a5 b7b6 a5b6 c7b6 b3d4 a8a1 d1a1 h7h6 g1h2 g6e7 e3a3 e7c6 d4c6 d7c6 f3d4 e6d4 a1d4');
INSERT INTO evaluations VALUES(5007,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(5007,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(5007,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(5007,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(5007,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(5007,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(5007,36,'cp 56','f1e1 d7d6 b1c3 g8f6 c1f4 e6e5 f4e3 e8g8 d1d2 d8e7 e3g5 c8e6 c4b3 f8d8 h2h3 a8b8 a1d1 d8d7 d2c1 e6b3 c2b3 e7e6 c3a4 b8c8 d1d3 g7f8 d3c3 d7c7 e1e2 f8e7 g5f6 e7f6 e2c2 c6c5 c1d2 c8d8 d2e3 f6e7');
INSERT INTO evaluations VALUES(6879,30,'cp 611',NULL);
INSERT INTO evaluations VALUES(6879,31,'cp 594',NULL);
INSERT INTO evaluations VALUES(6879,32,'cp 611',NULL);
INSERT INTO evaluations VALUES(6879,33,'cp 613',NULL);
INSERT INTO evaluations VALUES(6879,34,'cp 612',NULL);
INSERT INTO evaluations VALUES(6879,35,'cp 617',NULL);
INSERT INTO evaluations VALUES(6879,36,'cp 621','g1f3 d4c5 f3e5 g8f6 e5d3 c5b6 b1c3 d7d6 f1e2 e8g8 e2f3 f8e8 e1g1 c7c5 c1f4 c5c4 d3b4 b6a5 b4c2 a5c3 b2c3 f6e4 f1e1 c8f5 d1d4 d8f6 c2e3 f6d4 c3d4 f5d7 e3c4 a8c8 c4d6 e4d6 f4d6 e8e1 a1e1 b7b6 d6f4 c8c2');
INSERT INTO evaluations VALUES(6878,30,'cp 590',NULL);
INSERT INTO evaluations VALUES(6878,31,'cp 615',NULL);
INSERT INTO evaluations VALUES(6878,32,'cp 605',NULL);
INSERT INTO evaluations VALUES(6878,33,'cp 611',NULL);
INSERT INTO evaluations VALUES(6878,34,'cp 614',NULL);
INSERT INTO evaluations VALUES(6878,35,'cp 605',NULL);
INSERT INTO evaluations VALUES(6878,36,'cp 576','d4c5 f3e5 g8f6 e5d3 c5b6 b1c3 d7d6 f1e2 c7c6 c1f4 e8g8 e1g1 h7h6 d5c6 f8e8 e2f3 b7c6 e4e5 d6e5 d3e5 c8b7 d1d8 a8d8 e5c4 b7a6 c4b6 a6f1 g1f1 a7b6 f3c6 d8d4 c6e8 d4f4 e8a4 f6d5 c3d5 f4a4 f1e2 g8f8 e2f3 a4d4 d5b6 d4b4');
INSERT INTO evaluations VALUES(9044,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(9044,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(9044,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(9044,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(9044,34,'cp 205',NULL);
INSERT INTO evaluations VALUES(9044,35,'cp 205',NULL);
INSERT INTO evaluations VALUES(9044,36,'cp 217','c6e7 f1e1 b7b6 e3e4 c8b7 c4b5 f6d7 b5a4 e8g8 c3c4 h7h6 c1a3 c7c5 d4d5 e7g6 d1b3 d8e7 a1d1 a8d8 h2h3 e6e5 a3b2 b7c8 b3c3 f7f6 b2c1 d7b8 f3h2 f6f5 e4f5 c8f5 g2g4 f5d7 a4c2 e7f6 c2g6 f6g6 c3e5');
INSERT INTO evaluations VALUES(9250,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(9250,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(9250,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(9250,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(9250,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(9250,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(9250,36,'cp 46','e3e4 e6e5 c1g5 c8g4 d3b5 e8g8 c3d5 c6d4 g5f6 g7f6 b5e2 g4e6 f3d4 e5d4 d1d2 c7c6 d5f4 d6f4 d2f4 d8a5 f4d2 a5d2 e1d2 a8c8 b2b3 f8e8 a2a4 c8c7 h1e1 c6c5 a4a5 g8g7 e2d3 h7h5 h2h4 g7f8 d2e2 c5c4 b3c4 e6c4 e2f3 b7b5 a5b6 a7b6 f3f4 b6b5 a1a6 c4d3');
INSERT INTO evaluations VALUES(10475,30,'cp -401',NULL);
INSERT INTO evaluations VALUES(10475,31,'cp -444',NULL);
INSERT INTO evaluations VALUES(10475,32,'cp -380',NULL);
INSERT INTO evaluations VALUES(10475,33,'cp -437',NULL);
INSERT INTO evaluations VALUES(10475,34,'cp -432',NULL);
INSERT INTO evaluations VALUES(10475,35,'cp -439',NULL);
INSERT INTO evaluations VALUES(10475,36,'cp -451','c6d4 e4e5 f6d5 d1g4 d5c3 g4d4 c3a2 c1d2 c7c5 d4g4 b4d2 e1d2 a2b4 g4g7 h8f8 d3h7 d7d5 h2h4 c5c4 h4h5 d8b6 g7f6 b6f6 e5f6 f8h8 c2c3 h8h7 c3b4 h7h6 g2g4 c8g4 a1g1 g4c8 g1g7 e8f8 h1g1 h6f6 g7g8 f8e7 g1e1 e7d7 e1e5 d7c6 f2f3 f6f3 g8d8 f3f2 d2c3 c6c7 e5e8 f2h2 d8d5');
INSERT INTO evaluations VALUES(10792,30,'cp -443',NULL);
INSERT INTO evaluations VALUES(10792,31,'cp -451',NULL);
INSERT INTO evaluations VALUES(10792,32,'cp -456',NULL);
INSERT INTO evaluations VALUES(10792,33,'cp -429',NULL);
INSERT INTO evaluations VALUES(10792,34,'cp -434',NULL);
INSERT INTO evaluations VALUES(10792,35,'cp -445',NULL);
INSERT INTO evaluations VALUES(10792,36,'cp -450','e4e5 f6d5 d1g4 d5c3 g4d4 c3a2 c1d2 c7c5 d4g4 b4d2 e1d2 a2b4 g4g7 h8f8 d3h7 d7d5 f2f4 d8h4 h1f1 c8e6 c2c3 e8c8 c3b4 f8h8 a1a7 c8b8 a7a1 h8h7 g7g3 c5b4 g3h4 h7h4 h2h3 e6f5 f1f2 b8c7 a1e1 c7b6 g2g4 f5e4 f4f5 h4h3 e1e3 h3h1 e5e6 f7e6 f5e6 d8e8');
INSERT INTO evaluations VALUES(11363,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(11363,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(11363,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(11363,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(11363,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(11363,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(11363,36,'cp 50','a2a3 f8e7 d1c2 e8g8 b2b3 c6b8 f1d3 c7c5 c4d5 c5d4 f3d4 e6d5 c1b2 f8e8 e1g1 b8d7 d4f5 d7e5 f5e7 d8e7 a1d1 c8e6 c3e2 a8c8 c2d2 e5d3 d2d3 e6g4 h2h3 g4h5 d1c1 h5e2 d3e2 c8c1 f1c1 d5d4 b2d4 e7a3 c1c7 a3b3 d4f6 b3b1 g1h2 g7f6 e2f3 b1e4 f3f6');
INSERT INTO evaluations VALUES(11364,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(11364,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(11364,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(11364,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(11364,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(11364,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(11364,36,'cp 53','f8e7 d1c2 e8g8 b2b4 d5c4 f1c4 e7d6 c1b2 e6e5 f3e5 c6e5 d4e5 d6e5 f2f4 e5d6 e1g1 c7c6 h2h3 a6a5 b4b5 c8d7 c4d3 c6b5 c3e4 f6e4 d3e4 a8c8 c2d3 d6c5 a1d1 d7e6 e4h7 g8h8 d3c3 c5d4 d1d4 c8c3 d4d8 f8d8 b2c3 h8h7 c3a5 d8a8 a5b4 h7g8 f4f5 e6d5');
INSERT INTO evaluations VALUES(12403,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(12403,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(12403,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(12403,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(12403,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(12403,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(12403,36,'cp 0','d7d5 c4d5 e6d5 c1g5 f8e7 d1c1 d5d4 g5f6 e7f6 c3d5 f6e5 c1c5 e5d6 c5c1 e8g8 e1f1 f8e8 h2h4 h7h6 c1d2 c8d7 h4h5 a8c8 h1h4 d7e6 d5f4 e6d7');
INSERT INTO evaluations VALUES(14133,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(14133,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(14133,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(14133,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(14133,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(14133,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(14133,36,'cp 44','e7e6 f1b5 f8d6 b1d2 e8g8 e1g1 c6e7 f3e5 a7a5 f1e1 f6d7 b5d3 b7b6 e5d7 c8d7 d1e2 f7f5 c3c4 d5c4 d3c4 g8h8 d2f3 e7g6 f4g3 d8e7 c4b5 d7b5 e2b5 e7e8 b5e8 a8e8 g3d6 c7d6 e1c1 e8c8 d4d5 e6d5 f3d4');
INSERT INTO evaluations VALUES(14243,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(14243,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(14243,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(14243,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(14243,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(14243,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(14243,36,'cp -77','c8f5 g1f3 e7e6 b1d2 f8d6 f1e2 e8g8 e1g1 h7h6 b2b4 c6e7 a2a4 b7b6 b4b5 a8c8 c3c4 f5h7 d1b3 c7c5 c1b2 e7f5 e2d3 h7g6 d3f5 g6f5 a1c1 d8e7 c4d5 e6d5 d4c5 b6c5 b2f6 g7f6 b3d5 e7e3 g1h1 e3d3 d5c4 d3c4 c1c4 f5d3');
INSERT INTO evaluations VALUES(14740,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14740,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14740,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14740,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14740,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14740,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14740,36,'cp 0','b7b6 b2b4 f6e4 c1b2 b6c5 b4c5 a8b8 d1c2 c6b4 c2d1');
INSERT INTO evaluations VALUES(11815,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11815,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11815,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11815,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11815,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11815,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11815,36,'cp -6','b2b4 f6e4 c1b2 a8b8 a2a3 b6c5 d4c5 a7a5 c3e4 d5e4 d1a4 e4f3 a4c6 c8d7 c6f3 a5b4 g2g3 b4a3 b2a3 d8g5 f1g2 f8c5 e1g1 e8g8 h2h4 g5e5 a3c5 e5c5 f1c1 c5d6 f3e3 d7b5 a1a7 c7c5 c1c5 b5e2 c5c7 e2g4 g2f3 g4f5 h4h5 d6d3 e3d3 f5d3 g1g2 h7h6 c7e7 b8b2 a7c7 b2d2 g2g1 g7g5');
INSERT INTO evaluations VALUES(14862,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14862,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(14862,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14862,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14862,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(14862,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(14862,36,'cp 4','f8d6 c3b5 e8g8 c2c4 f8e8 b5d6 c7d6 c4d5 e6d5 f1e2 c6a5 c1d2 f6e4 d2a5 d8a5 b2b4 a5b6 d1b3 c8e6 a1c1 e8c8 e1g1 c8c7 a3a4 a7a5 c1c7 b6c7 b4a5 a8a5 e2d3 g7g6 f1b1 c7c6 d3b5 c6c7 b3d1 a5b5 a4b5 e4c3 d1e1 c3b1 e1b1 c7c3 h2h3');
INSERT INTO evaluations VALUES(9262,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(9262,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(9262,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(9262,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(9262,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9262,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(9262,36,'cp 7','c3b5 e8g8 c2c4 d5c4 f1c4 e6e5 d4e5 d6e5 d1d8 f8d8 c4e2 f6e8 f3e5 c6e5 e3e4 c8e6 c1f4 e5d3 e2d3 d8d3 b5c7 e8c7 f4c7 a8c8 c7e5 c8c4 e1e2 d3b3 h1d1 c4e4 e2f1 e6c4 f1g1 f7f6 d1d8 g8f7 e5c3 c4b5 f2f3 e4e8 d8e8 f7e8 g1f2 b5c6 h2h4 a7a5 c3a5 b3b2 f2g3 b2b3 a1d1');
INSERT INTO evaluations VALUES(15600,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(15600,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(15600,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(15600,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(15600,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(15600,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(15600,36,'cp 48','b1c3 g7g6 d2d4 f8g7 d4e5 c6e5 f3e5 d6e5 d1d8 e8d8 f2f4 c7c6 h1f1 c8e6 f4f5 e6c4 f1h1 h7h5 b2b3 c4a6 g2f1 a6f1 h1f1 d8e8 c1g5 f6d7 f5g6 f7g6 c3d1 g7f8 e1e2 f8e7 g5e3 h5h4 g3g4 e7c5 h2h3 c5e3 d1e3 e8e7 a2a4 d7c5');
INSERT INTO evaluations VALUES(16627,30,'cp -541',NULL);
INSERT INTO evaluations VALUES(16627,31,'cp -512',NULL);
INSERT INTO evaluations VALUES(16627,32,'cp -518',NULL);
INSERT INTO evaluations VALUES(16627,33,'cp -534',NULL);
INSERT INTO evaluations VALUES(16627,34,'cp -578',NULL);
INSERT INTO evaluations VALUES(16627,35,'cp -582',NULL);
INSERT INTO evaluations VALUES(16627,36,'cp -580','d7d5 f5f4 f8d6 f4h4 c6d4 e4e5 d6e5 f1d3 d8e7 e1f1 d4f5 h4g5 e8g8 b1c3 h7h6 g5d2 f8e8 d2e1 e7d8 g1f3 e5c3 e1c3 f6e4 c3e1 d8f6 f1g1 c8d7 e1d1 f6b6 d1f1 c7c5 c2c3 a8d8 h2h3 b6f6 h3h4 d5d4 c3d4 f5d4 f3d4 c5d4 d3e4 e8e4 c1d2 d8e8 a1e1 e4e1 d2e1');
INSERT INTO evaluations VALUES(15017,30,'cp -555',NULL);
INSERT INTO evaluations VALUES(15017,31,'cp -567',NULL);
INSERT INTO evaluations VALUES(15017,32,'cp -578',NULL);
INSERT INTO evaluations VALUES(15017,33,'cp -567',NULL);
INSERT INTO evaluations VALUES(15017,34,'cp -573',NULL);
INSERT INTO evaluations VALUES(15017,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(15017,36,'cp -573','f5f3 c8g4 f3d3 f6e4 f1e2 f8c5 e2g4 e4f2 d3e2 e8f8 c1e3 c5e3 e2e3 f2g4 e3c5 f8g8 g1f3 d8e8 e1d2 b7b6 c5a3 d5d4 h1e1 e8d7 a3d3 h7h5 h2h3 g4e3 f3d4 c6e5 d3e4 e5c4 d2c1 a8e8 e4d3 h8h6 b1c3 c7c6 d4f5 e3f5 d3d7 e8e1 c3d1 f5e3 d7d8 g8h7');
INSERT INTO evaluations VALUES(17762,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(17762,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(17762,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(17762,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(17762,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(17762,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(17762,36,'cp 58','f8d6 f1d3 g8f6 b1d2 f6d7 f4g3 e6e5 e3e4 e5d4 c3d4 d5e4 d3e4 e8g8 e4c6 b7c6 e1g1 c6c5 g3h4 f7f6 f1e1 f8e8 e1e8 d8e8 h4g3 d6f8 a1c1 c5d4 c1c7 d7c5 f3d4 c8b7 b2b3 a8c8 c7c8 b7c8 h2h3');
INSERT INTO evaluations VALUES(19158,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(19158,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(19158,32,'cp 192',NULL);
INSERT INTO evaluations VALUES(19158,33,'cp 183',NULL);
INSERT INTO evaluations VALUES(19158,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(19158,35,'cp 183',NULL);
INSERT INTO evaluations VALUES(19158,36,'cp 193','e1g1 a7a6 b5a4 b7b5 a4b3 d7d6 d1f3 a8a7 a2a4 b5b4 b1d2 g8f6 d2c4 c8e6 c1g5 f8e7 g5f6 e7f6 f3g3 a7d7 a4a5 e8g8 c4b6 d7e7 f2f4 g7g6 f4f5 e6b3 c2b3 f6e5 g3g4 f7f6 f5g6 g8h8 g4h4 e7g7 f1f5 g7g6 g2g3 d8e7 a1f1 e7g7 b6c4 g6g4 h4h5');
INSERT INTO evaluations VALUES(19770,30,'cp 303',NULL);
INSERT INTO evaluations VALUES(19770,31,'cp 332',NULL);
INSERT INTO evaluations VALUES(19770,32,'cp 298',NULL);
INSERT INTO evaluations VALUES(19770,33,'cp 295',NULL);
INSERT INTO evaluations VALUES(19770,34,'cp 296',NULL);
INSERT INTO evaluations VALUES(19770,35,'cp 307',NULL);
INSERT INTO evaluations VALUES(19770,36,'cp 304','c4d5 d8d5 b1c3 d5d8 d4d5 c6e5 f2f4 e5d7 e3e4 g7g6 g1f3 f8g7 a2a4 g8h6 h2h3 e8g8 c1e3 f7f6 d1d2 a8b8 f1d3 b7b6 e1g1 h6f7 c3b5 b8a8 d5d6 c8b7 d3c4 g8h8 f1d1 b7c6 d2c2 c6b7 c4e6 f7d6 b5d6 e7d6');
INSERT INTO evaluations VALUES(20373,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(20373,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(20373,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(20373,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(20373,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(20373,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(20373,36,'cp 36','c4c5 c8d7 f1b5 b7b6 b5c6 d7c6 f3e5 d8c8 b2b4 b6c5 b4c5 g8e7 e1g1 f7f6 e5d3 e7f5 d3f4 e8f7 f1e1 c6d7 d1h5 f7g8 c1e3 c8e8 h5d1 h7h5 b1c3 g7g5 f4d3 a8b8 f2f3 e8f7 e3f2 g8g7 a2a4 f8e7 d1d2 h5h4 c3b5');
INSERT INTO evaluations VALUES(20818,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(20818,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(20818,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(20818,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(20818,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(20818,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(20818,36,'cp 42','g7g6 c2c4 f7f5 b1c3 g8f6 f1d3 f8g7 e1g1 e8g8 b2b4 a7a5 b4a5 a8a5 a2a4 b7b6 f1e1 f5e4 c3e4 f6e4 d3e4 e7f5 c1d2 a5a7 a1a3 f5d4 f3d4 e5d4 d1c1 c8d7 a4a5 b6a5 c4c5 a7a8 a3a5 a8a5 d2a5 g7e5 c5c6 d7f5 c1d2 d8c8 f2f4 f5e4 e1e4');
INSERT INTO evaluations VALUES(23396,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(23396,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(23396,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(23396,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(23396,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(23396,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(23396,36,'cp 22','f5e4 c3e4 g8f6 e4f6 e7f6 f1e2 d7d5 e1g1 f8d6 d2d4 e8g8 c2c4 d5c4 e2c4 g8h8 h2h3 c8f5 f1e1 f8e8 c1d2 e8e1 d1e1 d8d7 e1e3 a8e8 e3b3 a7a5 a2a3 a5a4 b3b5 e8b8 a1e1 c6e7 d2b4 d7b5 c4b5 d6b4 a3b4 e7c6 b5a4 c6b4 a4b3 c7c6 e1e7 g7g5 h3h4 g5g4 f3d2 b4d3 b3c2 f5g6 c2d3 g6d3');
INSERT INTO evaluations VALUES(23401,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(23401,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(23401,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(23401,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(23401,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(23401,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(23401,36,'cp 21','c3e4 g8f6 e4g3 e7e5 d2d4 e5d4 f3d4 f8c5 c1e3 c5d4 e3d4 e8g8 f1e2 c6d4 d1d4 d7d5 e1g1 d8d6 f1e1 c8e6 a2a4 c7c5 d4d2 a8e8 b2b3 e6f7 e2b5 e8e1 a1e1 d6c7 d2g5 h7h6 g5d2 f8d8 h2h3 a7a6 b5d3 d8e8 e1e8 f6e8 c2c4 d5c4 d3c4 f7c4');
INSERT INTO evaluations VALUES(23400,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(23400,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(23400,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23400,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(23400,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(23400,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(23400,36,'cp 12','g8f6 e4g3 e7e5 d2d4 e5d4 f3d4 f8c5 c1e3 c5d4 e3d4 e8g8 f1e2 c6d4 d1d4 d7d5 e1g1 c7c6 a1e1 d8b6 d4c3 c8d7 a2a3 c6c5 e2d3 b6c7 h2h3 a8e8 b2b4 b7b6 b4c5 b6c5 e1e8 f8e8 f1b1 g8f7 d3e2 d5d4 c3d3 f7f8 e2f3 c7e5');
INSERT INTO evaluations VALUES(24151,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(24151,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(24151,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(24151,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(24151,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24151,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24151,36,'cp 25','c4e6 f7e6 c3a4 c5b6 c2c3 e8g8 b2b4 c6e7 d1b3 d8d7 a4b6 a7b6 a2a4 h7h6 c1e3 f6g4 e3d2 g4f6 h2h3 f6h5 b3d1 e7g6 d2e3 h5f6 a4a5 b6b5 f3h2 f8f7 h2g4 d7c6 g4f6 f7f6 d1b3 d6d5 f2f3 a8d8 b3c2 d5d4 e3d2 f6f7 g1h2 d4c3 d2c3');
INSERT INTO evaluations VALUES(25877,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(25877,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(25877,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(25877,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(25877,34,'cp 155',NULL);
INSERT INTO evaluations VALUES(25877,35,'cp 151',NULL);
INSERT INTO evaluations VALUES(25877,36,'cp 149','e1g1 f8c5 c3d5 g8f6 f1e1 c6a5 c4f1 e8g8 d5f6 d8f6 c2c3 c5d6 d2d4 a5c6 d4d5 c6d8 c1g5 f6g6 g5d2 g6h5 f1e2 h5g6 c3c4 f7f6 e2f1 g6e8 b2b4 d8f7 d1b3 c7c5 a2a3 e8e7 f3h4 g7g6 g2g3 b7c8 h4f3 c8b7 f1h3 a8e8 h3g2 b7c8 f3h4');
INSERT INTO evaluations VALUES(26698,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(26698,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(26698,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(26698,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(26698,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(26698,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(26698,36,'cp 58','a1c1 e7e6 c3b5 f8d6 b5d6 c7d6 c2c3 b4c6 d1b3 d8e7 e2e3 g8f6 f3h4 f5g6 f4g5 a7a6 g5f6 g7f6 b3d1 e7c7 f1e2 e8g8 e1g1 c6e7 g2g3 g8h8 d1d2 b7b5 g1h1 c7d7 a2a3 f8c8 h4g2 c8g8 g3g4 a8c8');
INSERT INTO evaluations VALUES(26697,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(26697,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(26697,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(26697,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(26697,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(26697,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(26697,36,'cp 56','e7e6 c3b5 f8d6 b5d6 c7d6 c2c3 b4c6 d1b3 d8e7 e2e3 g8f6 f3h4 f5g6 f4g5 a7a6 h4g6 h7g6 f2f3 e7d7 f1d3 f6h7 g5h4 h7f6 h4f2 e8g8 e1g1 b7b5 f2h4 c6a5 b3c2 e6e5 c2f2 a8c8 h4f6 g7f6 e3e4 g8g7 e4d5 e5d4 f2d4 a5c4 b2b3 c4b2 c3c4');
INSERT INTO evaluations VALUES(27523,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(27523,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(27523,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(27523,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(27523,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(27523,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(27523,36,'cp -10','f1e2 c5d4 c3d4 a8c8 e1g1 g8e7 g1h1 a7a6 c1b2 e7f5 b1c3 b6a7 c3a4 b7b5 a4c5 f8c5 d4c5 a7c5 a1c1 c5e7 e2d3 e8g8 d1d2 f8d8 d3b1 b5b4 h2h3 a6a5 h1g1 f5h4 f3h4 e7h4 c1c5 a5a4 b3a4 c6e7 b2d4 d7a4 f2f4 c8c5 d4c5 d8c8 c5b4 a4b5 f1c1 c8c1 d2c1 h7h6');
INSERT INTO evaluations VALUES(32564,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(32564,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(32564,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(32564,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(32564,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32564,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32564,36,'cp 0','g7g6 c1g5 f8g7 e1g1 b8c6 d4c6 b7c6 b2b3 d7b7 d1e1 e8g8 a1d1 f8e8 f2f3 f6d7 g1h1 b7b4 g5d2 b4b7');
INSERT INTO evaluations VALUES(32558,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32558,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32558,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32558,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32558,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32558,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32558,36,'cp 0','e1g1 f8g7 c1g5 b8c6 d4c6 b7c6 b2b3 d7b7 d1e1 f6d7 a1d1 d7c5 g5e3 b7b4 e3d2 e8g8 f2f3 b4a3 d2c1 a3a6 a2a4 c5e6 c1e3 a6b7 f3f4 b7b4 c3d5 b4b7');
INSERT INTO evaluations VALUES(33112,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(33112,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(33112,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(33112,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(33112,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33112,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33112,36,'cp -13','d5d4 g1f3 d4d8 d1e2 b8c6 e1c1 c8d7 f3e5 c6e5 e2e5 g8f6 c3e4 f6e4 d2b4 d8g5 e5g5 e4g5 d1e1 h7h5 b4c3 f7f6 f1c4 e8c8 h2h4 g5f7 c4e6 d7e6 e1e6 d8e8 e6e8 h8e8 h1e1 e8e1 c3e1 c8d7 c1d2 d7e6 d2e3 f7d6 e3d3 c7c5 f2f3 g7g5 b2b3 d6f5 h4g5 f6g5');
INSERT INTO evaluations VALUES(33282,30,'cp 255',NULL);
INSERT INTO evaluations VALUES(33282,31,'cp 249',NULL);
INSERT INTO evaluations VALUES(33282,32,'cp 243',NULL);
INSERT INTO evaluations VALUES(33282,33,'cp 257',NULL);
INSERT INTO evaluations VALUES(33282,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(33282,35,'cp 239',NULL);
INSERT INTO evaluations VALUES(33282,36,'cp 258','b8c6 e5f3 h4d8 c2c3 c5e7 d2d4 d7d5 e4d5 d8d5 c3c4 d5f5 c1d2 g8f6 d4d5 c6b8 d2b4 e8g8 b4e7 f8e8 b1c3 f5d7 e1c1 e8e7 e2d2 d7d6 f3d4 c7c6 f2f3 c6d5 d4b5 d6c5 c4d5 b8a6 a2a3 c8d7 b2b4 c5b6 d5d6 e7e3 d2d4 a6b8 d4b6 a7b6 c1b2 b8c6 f1c4 g8h8 c4f7 c6e5');
INSERT INTO evaluations VALUES(35333,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(35333,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(35333,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(35333,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(35333,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(35333,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(35333,36,'cp 64','e7e5 h2h3 g8f6 a2a4 f8e7 f1c4 e8g8 e1g1 h7h6 c1e3 b8d7 c4a2 f8e8 d4d5 c6d5 a2d5 d7b6 d5b3 c8e6 f3d2 a7a6 d1e2 b6d7 f1d1 c7c6 a4a5 e6b3 c2b3 d6d5 c3d5 f6d5 a1c1 d5f4 e3f4 c6b5 e2b5 a6b5 d2f3 d7f6 f4e3 a8a5 c1c7 a5a2 c7b7 a2b2 b7b5 f6e4');
INSERT INTO evaluations VALUES(35340,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(35340,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(35340,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(35340,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(35340,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(35340,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(35340,36,'cp 62','a2a4 f8e7 f1c4 g8f6 h2h3 b8d7 e1g1 e8g8 f1e1 f8e8 c4a2 h7h6 c1e3 d7f8 d4d5 c8d7 f3d2 f8g6 d1f3 e8f8 a4a5 b7b5 a5b6 a7b6 a2c4 c7b7 f3e2 c6d5 c4d5 f6d5 c3d5 b6b5 c2c4 d7c6 c4b5 c6b5 d2c4 f7f5 d5e7 g6e7 c4d6');
INSERT INTO evaluations VALUES(36539,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36539,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36539,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36539,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36539,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36539,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(36539,36,'cp 0','f8e8 g1e2 e7e5 h2h3 c7c6 e1g1 e5e4 g3g4 d6d5 e2g3 h7h6 f2f3 e4f3 d1f3 c8e6 c4d5 f6d5 c3e2 d8f6 f3f6 d5f6 e3e4 b8a6 e4e5 f6h7 b2b3 f7f6 e2f4 a8d8 c1a3 f6e5 f4g6 e5d4 g6e7 g8h8 e7g6');
INSERT INTO evaluations VALUES(43084,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(43084,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(43084,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(43084,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(43084,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(43084,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(43084,36,'cp 95','f1d3 d7d6 a2a4 b8d7 e1g1 e8g8 a4a5 c7c5 f1e1 e7c6 c1g5 d8c7 b1a3 c5d4 a3b5 c7b8 b5d4 h7h6 g5h4 e6e5 d4c2 b7b5 c2e3 d7f6 h4f6 g7f6 h2h4 a7a6 b2b4 c8e6 h4h5 f6g7 d1d2 g6h5 f3h4 b8d8 h4f5 e6f5 e3f5 d8g5 d2g5 h6g5');
INSERT INTO evaluations VALUES(46327,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(46327,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(46327,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(46327,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(46327,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(46327,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46327,36,'cp 0','b7b5 b1d2 c8b7 e1g1 b8d7 f1e1 f8d6 e3e4 d5e4 d2e4 b7e4 c2e4 f6e4 e1e4 e8g8 c1f4 d6f4 e4f4 d7f6 d1d2 f6d5 f4g4 d5f6');
INSERT INTO evaluations VALUES(48950,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(48950,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(48950,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(48950,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(48950,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(48950,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(48950,36,'cp 29','f1c4 f8d6 d1e2 d8e7 e2e7 d6e7 e1g1 e8g8 f1e1 e7d6 c1g5 b8d7 b1c3 c7c6 c4b3 h7h6 g5h4 a7a5 h2h3 g7g5 h4g3 d6g3 f2g3 d7b6 a2a4 b6d5 b3d5 f6d5 c3d5 c6d5 e1e7 a8a6 a1c1 a6b6 c1c5 b6b2 c5a5 c8e6 a5b5 b2a2 a4a5 f8c8 b5b7 a2a5 f3e5');
INSERT INTO evaluations VALUES(49467,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(49467,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(49467,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(49467,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(49467,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49467,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(49467,36,'cp 0','c2c4 f8e7 d2d4 e8g8 b1c3 d7d5 b2b3 c7c5 f1d3 b7b6 e1g1 c5d4 e3d4 c8b7 f1e1 b8c6 c1b2 c6b4 d3f1 b4c6');
INSERT INTO evaluations VALUES(51072,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(51072,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(51072,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(51072,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(51072,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(51072,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(51072,36,'cp 113','b1c3 d7d6 g1f3 b8c6 e2e4 e7e5 d4d5 c6e7 f1e2 g8f6 a2a4 e7g6 h2h4 g6f4 e2f1 f4h5 f3h2 g7g6 a4a5 f8g7 g2g3 e8g8 f1e2 b7b6 a5b6 c7b6 e1g1 c8d7 a1a3 d8e7 c1e3 f8b8 b2b4 b6b5 c4b5 a6b5 a3a8 b8a8');
INSERT INTO evaluations VALUES(51071,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(51071,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(51071,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(51071,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(51071,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(51071,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(51071,36,'cp 109','d7d6 g1f3 b8c6 e2e4 e7e5 d4d5 c6e7 f1e2 g8f6 a2a4 e7g6 h2h4 g6f4 e2f1 c8g4 g2g3 f4h5 f1e2 d8d7 f3h2 g4e2 d1e2 g7g6 c1e3 f8g7 a4a5 e8g8 e1g1 d7e8 c4c5 f6d7 c5c6 b7c6 d5c6 d7f6 e2c4 e8c8 b2b4 c8h3');
INSERT INTO evaluations VALUES(51451,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(51451,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(51451,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(51451,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(51451,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(51451,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(51451,36,'cp 87','f2f4 b8c6 f4e5 c6e5 b1c3 f8b4 c1d2 d7d6 g1f3 c8g4 f1e2 g4f3 e2f3 d8h4 g2g3 h4h3 a2a3 b4c3 d2c3 g8e7 f3e2 h3g2 e1d2 c7c5 d1f1 g2f1 a1f1 e7c6 f1f2 e8e7 h1f1 f7f6 b2b4 c6d4 e2d1 h8e8 c3b2 d4e6 b4b5 e6d4');
INSERT INTO evaluations VALUES(52308,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(52308,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(52308,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(52308,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(52308,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(52308,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(52308,36,'cp -7','b8c6 f1g2 f8g7 e1g1 d7d6 d2d3 g8f6 a2a4 a8b8 b1a3 e8g8 a3c4 h7h6 c2c3 e7e5 c4e3 c8e6 f3h4 f6e8 e3d5 e6d5 e4d5 c6e7 c3c4 f7f5 c1d2 g6g5 h4f3 a7a5 d2c3 e7g6 f3d2 b7b6 h2h3 g6e7 d1e2 d8d7 b2b3 e8f6');
INSERT INTO evaluations VALUES(52719,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(52719,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(52719,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(52719,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(52719,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(52719,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(52719,36,'cp 82','d5c6 b8c6 e2e3 e7e5 d2d3 f8d6 f1e2 g8f6 b1c3 e8g8 g1f3 d8e7 e1g1 f8d8 c3e4 f6e4 d3e4 c8e6 c1d2 a7a5 d1a4 e6d7 a4c2 d8c8 d2c3 a5a4 f1d1 d7e6 c2d3 c8d8 d3b1 h7h6 e2b5 e6g4 a2a3 g4e6 b5c6 b7c6 c3e5');
INSERT INTO evaluations VALUES(52691,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(52691,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(52691,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(52691,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(52691,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(52691,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(52691,36,'cp 89','b8c6 e2e3 e7e5 d2d3 f8d6 f1e2 d8e7 b1c3 g8f6 g1f3 e8g8 e1g1 c8f5 e3e4 f5e6 c1e3 f8d8 a1c1 h7h6 h2h3 d6c7 a2a3 a8c8 f1e1 c7b8 d1a4 f6h7 e1d1 e7d6 d3d4 c6d4 e3d4 e5d4 d1d4 d6e7 d4d8 c8d8');
INSERT INTO evaluations VALUES(19900,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(19900,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(19900,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(19900,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(19900,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(19900,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(19900,36,'cp 78','e2e3 e7e5 b1c3 g8f6 f1b5 f8b4 g1e2 e8g8 e1g1 c8e6 d2d4 e5d4 e3d4 a7a6 b5c6 b7c6 c1g5 h7h6 g5f6 d8f6 d1a4 a6a5 a4c6 a8c8 c6a4 f6g5 f1c1 f8e8 c3e4 g5e7 c1c8 e6c8 e4g3 c8g4');
INSERT INTO evaluations VALUES(55829,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(55829,31,'cp -84',NULL);
INSERT INTO evaluations VALUES(55829,32,'cp -82',NULL);
INSERT INTO evaluations VALUES(55829,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(55829,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(55829,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(55829,36,'cp -72','c7c5 e2e3 a7a6 g1f3 g8f6 a2a4 d8c7 f1d3 b7b6 e1g1 f8e7 d1e1 b8c6 b2b3 c6b4 f3e5 e8g8 c1a3 c5d4 e3d4 b4d3 c2d3 e7a3 a1a3 f6e8 e1d2 f7f6 e5f3 c7e7 a3a1 a6a5 f1c1 c8a6 c1c2 e8d6 a1c1 a8c8 c3e2');
INSERT INTO evaluations VALUES(56602,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(56602,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(56602,32,'cp -63',NULL);
INSERT INTO evaluations VALUES(56602,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(56602,34,'cp -77',NULL);
INSERT INTO evaluations VALUES(56602,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(56602,36,'cp -47','g8f6 g1f3 f6e4 f4e5 f8c5 e2f4 e8g8 f1c4 b8c6 f4d3 c5e7 e1g1 c6a5 b2b3 a5c4 b3c4 e4g5 f3g5 e7g5 d1f3 c8e6 d3f4 g5f4 f3f4 e6c4 d2d3 c4e6 c1a3 f8e8 a3c5 d8d7 c5d4 b7b6 f1f3 h7h6 f3g3 g8h7 a1c1 a8d8 d4e3 d7c6');
INSERT INTO evaluations VALUES(56686,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(56686,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(56686,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(56686,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(56686,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(56686,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(56686,36,'cp 155','b8c6 a2a4 g7g6 h2h3 f8h6 b1c3 e8f8 c1d2 f8g7 d3d4 a7a6 e1g1 h6d2 d1d2 g8e7 c3e2 h8f8 f1e1 a6a5 a1d1 e7g8 e2c3 g7h8 c3d5 c8e6 b2b3 e6f7 d4e5 c6e5 c4e2 f7d5 e4d5 e5f3 e2f3 f6f5 e1e6 d8d7');
INSERT INTO evaluations VALUES(57773,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(57773,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(57773,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(57773,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(57773,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(57773,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(57773,36,'cp 145','g1f3 b8c6 g7h7 d7d5 d2d3 e4f6 h7h4 c8g4 f1e2 d8d7 h2h3 g4f3 e2f3 e8c8 h4f4 d8e8 e1f1 d7e6 b1c3 e7d6 f4e3 e6d7 e3h6 d6e5 c3e2 f6e4 d3e4 d5e4 h6d2 d7d2 f3g4 f7f5 c1d2 f5g4 h3g4 e5b2 a1b1 e4e3 d2e3 e8e3 b1b2 e3e4 g2g3 e4g4 e2f4 c6e5 b2b3 b7b6 f4d5 g4e4 d5e3 e5g6');
INSERT INTO evaluations VALUES(3207,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(3207,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(3207,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(3207,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3207,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(3207,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(3207,36,'cp -17','h2h3 e7e5 c3c4 a7a6 b1c3 a8b8 a2a4 b7b6 f3e1 f6e8 e1c2 e8c7 c3d5 g8h8 g1h2 a6a5 d5c7 d8c7 c2a3 c7d7 a3b5 f7f5 e4f5 g6f5 c1g5 c8b7 d1h5 c6b4 g2b7 b8b7 a1d1 b4c2 b2b3 c2d4 b5d4 c5d4');
INSERT INTO evaluations VALUES(4587,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(4587,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(4587,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(4587,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(4587,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(4587,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(4587,36,'cp 15','d7d6 a2a4 c8e6 c4e6 f7e6 c2c3 a7a5 b1d2 h7h6 d2f1 d8e8 d1b3 e8d7 h2h3 b7b6 c1d2 f6h5 f1h2 h5f4 d3d4 f4d3 e1e3 d3f4 a1d1 e5d4 c3d4 e7f6 d2c3 c6b4 h2g4 d7f7 c3b4 a5b4 g4f6 f7f6 b3b4');
INSERT INTO evaluations VALUES(3388,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3388,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(3388,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3388,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(3388,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(3388,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(3388,36,'cp 40','a2a3 c8e6 b1c3 e6c4 d3c4 f8e8 a3a4 a7a5 b2b3 c6b8 d1e2 b8a6 c1e3 f6g4 e3d2 g4f6 a1d1 c7c6 d2e3 d8c8 h2h3 c8e6 f3h4 g7g6 g1h2 a6b4 h4f3 f6h5 g2g3 h5g7 e1f1 f7f6 e3h6 e7f8 f3e1 f6f5 e1g2 e6f7 f2f4 e5f4 f1f4');
INSERT INTO evaluations VALUES(7174,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7174,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7174,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(7174,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7174,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7174,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7174,36,'cp 0','e1g1 e8g8 d2d3 c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 h7h6 c3e2 c6e7 e2g3 e7g6 f3h2 d8e7 g3h5 f6h7 f1f8 g6f8 h5g3 e7g5 g3f1 g5e7');
INSERT INTO evaluations VALUES(7173,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7173,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(7173,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7173,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7173,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7173,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7173,36,'cp 0','e8g8 d2d3 c8e6 c4e6 f7e6 c1e3 c5e3 f2e3 h7h6 c3e2 c6e7 e2g3 d8d7 c2c4 e7g6 f1f2 b7b5 d1c2 c7c5 f3d2 f8f7 a1f1 a8f8 c2b3 g8h7 b3c2');
INSERT INTO evaluations VALUES(7917,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(7917,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(7917,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(7917,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(7917,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(7917,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(7917,36,'cp 44','d2d4 d7d5 c4d5 e6d5 d4c5 e7c5 a2a3 f6e4 c3e4 d5e4 d1d8 c6d8 f3e5 f7f5 f2f3 e4f3 e5f3 c5b6 b2b4 e8g8 c1b2 a7a5 b4b5 a5a4 e1c1 d8f7 f3e5 f7e5 b2e5 h7h6 e5d4 b6d8 g2d5 g8h7 h1f1 d8g5 c1c2 f8e8 e2e4 a8a5 c2c3 c8d7 b5b6 e8c8 c3b2 a5b5 b2a1');
INSERT INTO evaluations VALUES(9152,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(9152,31,'cp 239',NULL);
INSERT INTO evaluations VALUES(9152,32,'cp 278',NULL);
INSERT INTO evaluations VALUES(9152,33,'cp 253',NULL);
INSERT INTO evaluations VALUES(9152,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(9152,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(9152,36,'cp 258','d1b3 c8e6 b3b7 d5e7 c4e6 f7e6 f3g5 a8b8 b7a6 d8d5 d2d3 e7f5 a6c4 c5e7 g5e4 a7a5 b1d2 e8g8 f1e1 h7h6 b2b3 b8d8 c1b2 f5h4 a1d1 f8f7 d2f1 d8f8 f1e3 d5d7 f2f3 h4g6 g1h1 g6f4 c4b5 f8d8 e3g4 d7d5 b5a4 f7f8 g4e3 d5d7 a4c4 f4d5 b2c1 e7h4 e1f1');
INSERT INTO evaluations VALUES(10007,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(10007,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(10007,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(10007,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(10007,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10007,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(10007,36,'cp 12','b2c3 d7d6 h2h3 h7h6 d1e2 c8d7 e1g1 e8g8 a2a4 f8e8 b5c4 b7b6 f3h2 a7a6 e2d1 d7e6 c4e6 f7e6 f2f4 e5f4 c1f4 e8f8 d1e2 d8e7 f4g3 a8e8 h2f3 e6e5 g3h4 g8h8 h4f6 f8f6 a1e1 f6g6');
INSERT INTO evaluations VALUES(10045,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(10045,31,'cp 230',NULL);
INSERT INTO evaluations VALUES(10045,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(10045,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(10045,34,'cp 219',NULL);
INSERT INTO evaluations VALUES(10045,35,'cp 234',NULL);
INSERT INTO evaluations VALUES(10045,36,'cp 217','e5c6 d7c6 h2h3 c8e6 d2d3 d8e7 g2g3 e8c8 f1g2 f6d7 d1e2 c5d4 e1g1 g7g5 c1e3 c6c5 c3b5 d4e3 e2e3 c7c6 b5c3 f7f6 c3e2 d8e8 f2f4 h8g8 b2b3 g5f4 f1f4 c8b8 a1f1 a7a6 f4f2 b8a8 e2f4 e6f7 g1h1 d7e5');
INSERT INTO evaluations VALUES(9959,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(9959,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(9959,32,'cp 214',NULL);
INSERT INTO evaluations VALUES(9959,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(9959,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(9959,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(9959,36,'cp 207','d7c6 h2h3 c8e6 d2d3 d8e7 g2g3 e8c8 d1e2 f6d7 f1g2 c5d4 c1e3 c6c5 e1g1 f7f6 c3b5 d4e3 e2e3 a7a6 b5c3 c8b8 f2f4 h8e8 a1e1 h7h6 g1h2 g7g5 a2a4 h6h5 a4a5 h5h4 g3g4 g5f4 e3f4 e6f7');
INSERT INTO evaluations VALUES(10416,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(10416,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10416,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10416,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10416,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10416,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(10416,36,'cp 0','e8g8 e2e3 d7d5 c4d5 f6d5 c3d5 b4d2 d1d2 d8d5 a2a3 f8d8 f1e2 d8d6 a1c1 d6g6 f3h4 g6h6 h4f3');
INSERT INTO evaluations VALUES(4444,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(4444,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4444,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(4444,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(4444,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(4444,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(4444,36,'cp 11','g2g3 h7h6 f1g2 f8e8 e1g1 a7a5 a2a3 b4c3 d2c3 d7d5 c4d5 f6d5 f1e1 d5c3 b2c3 a5a4 d3d4 e5d4 c3d4 c8e6 e2e4 e6g4 d4d5 c6e5 h2h3 e5f3 g2f3 g4f3 d1f3 c7c6 d5c6 b7c6 a1c1 d8d7 g1g2 d7d6 c1c3 c6c5 e4e5');
INSERT INTO evaluations VALUES(10606,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(10606,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(10606,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(10606,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(10606,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(10606,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(10606,36,'cp 51','b1c3 e8g8 f1e2 d7d6 b2b3 e5f3 e2f3 c6e5 c1b2 c7c6 e1g1 f8e8 c3e4 d8e7 e4c5 d6c5 d1e2 c8f5 b2e5 e7e5 a1d1 a7a5 f3g4 f5g4 h3g4 a5a4 d1d7 e8e7 d7d2 a4b3 a2b3 h7h6 f1d1 e5g5 g2g3 f7f5 d2d8 a8d8 d1d8 g8h7 e2c2 g5g4');
INSERT INTO evaluations VALUES(10639,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(10639,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(10639,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(10639,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(10639,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(10639,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(10639,36,'cp 65','e1g1 d7d6 h2h3 g4f6 c2c3 a7a5 d3d4 c5b6 f1e1 e8g8 c1e3 d6d5 f3e5 f6e4 b1d2 e4d2 d1d2 c6e5 d4e5 c7c6 g1h2 b6e3 e1e3 c8e6 a1e1 d8b6 f2f4 a8d8 e1d1 d5d4 c3d4 e6a2 f4f5 a2b3 d1e1 h7h6 e5e6 f7e6 f5e6');
INSERT INTO evaluations VALUES(10638,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(10638,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(10638,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(10638,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(10638,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(10638,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(10638,36,'cp 55','d7d6 c2c3 a7a5 d3d4 c5a7 f1e1 e8g8 h2h3 g4f6 c1e3 c8d7 b1d2 b7b5 a2a4 b5a4 a1a4 f8e8 a4a3 d8b8 b2b3 e5d4 f3d4 h7h6 a3a1 a7d4 e3d4 c6d4 c3d4 c7c6 g1h2 b8b4 a1a4 b4b8 d2c4 d6d5 e4d5 e8e1 d1e1 b8b3 a4a5');
INSERT INTO evaluations VALUES(14829,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(14829,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(14829,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(14829,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14829,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(14829,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(14829,36,'cp -1','c6b4 d3b5 c7c6 b5e2 c6c5 c2c3 b4c6 f3e5 d8c7 b1d2 f8d6 a2a4 e8g8 e2d3 f6d7 e1g1 f7f6 d1h5 f6f5 h5e2 d7f6 b2b3 c5d4 c3d4 c8d7 d2f3 f8c8 c1d2 d7e8 f1c1 c7e7 e5c6 e8c6 d2e1 c6e8 a4a5 a7a6 f3e5 c8c1 a1c1 f6e4');
INSERT INTO evaluations VALUES(15541,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(15541,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(15541,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(15541,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(15541,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(15541,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(15541,36,'cp 118','d4d5 c6b8 e2e4 b8d7 f1d3 g7g6 h2h3 f8g7 c1g5 e8g8 d3c2 a7a5 a2a3 d8e8 g2g4 d7c5 d1e2 c8d7 g5e3 c7c6 e1c1 a8b8 f3d2 b7b5 e3c5 d6c5 d5c6 d7c6 c4b5 c6b5 c3b5 e8b5 e2b5 b8b5 d2c4 g7h6 c1b1 f8b8 c2a4 f6e4 a4b5');
INSERT INTO evaluations VALUES(17167,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(17167,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(17167,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(17167,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(17167,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(17167,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(17167,36,'cp 48','e7c6 b1c3 f8c5 d1f3 d7d6 f3g3 e8g8 c1d2 f7f5 f1c4 g8h8 e1c1 f5e4 c3e4 c5d4 h1e1 c8f5 d2c3 d4c3 e4c3 d8d7 a2a3 h7h6 c4d5 a7a6 f2f3 a8e8 e1e8 f8e8 g3f2 f5h7 f2d2 h7g8 g2g3 b7b5 f3f4 b5b4 a3b4 c6b4 d5g8 h8g8');
INSERT INTO evaluations VALUES(16871,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(16871,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(16871,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(16871,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(16871,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(16871,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(16871,36,'cp 60','b1c3 f8c5 d1f3 e8g8 f3g3 d7d6 c1d2 d8e8 e1c1 f7f5 e4f5 c8f5 f1c4 g8h8 h1e1 e8g6 g3g6 f5g6 f2f3 c5f2 e1f1 f2d4 d1e1 a7a5 c4b3 d4c5 b3d5 a8e8 e1e8 f8e8 f1e1 e8e1 d2e1 c6b4 d5b3 c5d4 c3b5 d4b6 a2a4 b4a6 e1d2');
INSERT INTO evaluations VALUES(16870,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(16870,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(16870,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(16870,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(16870,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(16870,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(16870,36,'cp 59','f8c5 d1f3 e8g8 f3g3 c5d4 c3d5 d7d6 c2c3 d4f6 h2h4 c8e6 c1e3 f8e8 e1c1 e6d5 e4d5 c6e5 h4h5 d8d7 h5h6 d7g4 g3h2 g7g6 d1d4 g4d7 f1e2 f6h8 h2f4 c7c5 d5c6 b7c6 h1d1 d6d5 f4g3 a7a5 f2f4');
INSERT INTO evaluations VALUES(18260,30,'cp -141',NULL);
INSERT INTO evaluations VALUES(18260,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(18260,32,'cp -127',NULL);
INSERT INTO evaluations VALUES(18260,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(18260,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(18260,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(18260,36,'cp -103','e5e4 d1e2 c6d5 d2d3 c8a6 c2c4 g8f6 d3e4 f8b4 c1d2 a6c4 e2d1 b4c5 d1a4 d8d7 a4d7 f6d7 b1c3 e8c8 c3d5 h8e8 d5c3 f7f5 d2e3 c5e3 f2e3 d7c5 e1f2 c5d3 f2g3 f5e4 f3d4 d8d6 h2h3 d3b2 a1b1 b2d3 h1f1 h7h6 g3h2 d6b6 d4b3 c4b3');
INSERT INTO evaluations VALUES(19646,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(19646,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19646,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19646,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(19646,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(19646,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(19646,36,'cp 13','f8g7 f1e1 a7a6 a2a4 g8f6 d2d4 c5d4 f3d4 e8g8 h2h3 c6d4 d1d4 c8d7 d4d3 a8c8 c1e3 d8c7 c4d5 d7e6 e3d4 e6d5 e4d5 c7c4 a1d1 c4d3 d1d3 f8e8 g2g4 f6d7 d4g7 g8g7 d3e3 g7f8 g4g5 d7c5 b2b3');
INSERT INTO evaluations VALUES(21013,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21013,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(21013,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(21013,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21013,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21013,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21013,36,'cp 0','g1f3 c8f5 c2c4 e7e6 c4d5 c6b4 d1a4 c7c6 d5c6 b7b5 a4b5 b4c2 e1d1 c2a1 e3f4 f5b1 c6c7 d8d7 b5b8 d7c8 b8b5');
INSERT INTO evaluations VALUES(21014,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21014,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21014,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21014,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21014,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21014,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21014,36,'cp 0','c8f5 c2c4 e7e6 c4d5 c6b4 d1a4 c7c6 d5c6 b4c2 e1d1 b7b5 a4b5 c2a1 c6c7 d8d7 b5b8 d7c8 b8b5');
INSERT INTO evaluations VALUES(21766,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(21766,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(21766,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(21766,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(21766,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(21766,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(21766,36,'cp -22','b2b4 g8f6 c1b2 f8d6 c2c4 e8g8 c4d5 f6d5 b1c3 d5c3 b2c3 f8e8 b4b5 c6d4 g1f3 d4f5 g2g4 f5h4 f3h4 d8h4 d1c2 e5e4 a3a4 a7a6 c2b2 d6f8 f1e2 a6b5 e2b5 e8d8 b5e2 b7b6 a4a5 c8a6 a5b6 a6e2 a1a8 d8a8 e1e2 c7b6 b2b6');
INSERT INTO evaluations VALUES(22526,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(22526,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(22526,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(22526,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(22526,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(22526,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(22526,36,'cp 60','d2d4 e5d4 c3d4 d7d5 e4d5 d8d5 b1c3 f8b4 f1e2 g8e7 a2a3 b4c3 b2c3 c6a5 e1g1 e8g8 f1e1 e7g6 d1a4 b7b6 c3c4 d5f5 a4b4 c8d7 h2h3 f8e8 b4c3 f5f6 c1b2 a8d8 e2f1 c7c5 d4d5 f6c3 b2c3 e8e1 a1e1 a5b7 f3e5 g6e5 c3e5');
INSERT INTO evaluations VALUES(22523,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(22523,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(22523,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(22523,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(22523,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(22523,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(22523,36,'cp 63','e5d4 c3d4 d7d5 e4d5 d8d5 b1c3 f8b4 f1e2 g8e7 a2a3 b4c3 b2c3 c6a5 e1g1 e8g8 c1f4 c8e6 f1e1 d5c6 a1b1 a5c4 f3e5 c4e5 f4e5 a8d8 e2f3 c6c3 b1b7 c3a3 b7c7 f8e8 h2h3 a3a5 d1a1 a5a1 e1a1 f7f6 a1a7 f6e5 c7e7 e8e7');
INSERT INTO evaluations VALUES(22545,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(22545,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(22545,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(22545,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(22545,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(22545,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(22545,36,'cp 81','c3d4 d7d5 e4d5 d8d5 b1c3 f8b4 f1e2 g8e7 a2a3 b4c3 b2c3 c6a5 c1f4 c7c6 e1g1 e8g8 f1e1 c8e6 e2d3 a8e8 f3e5 e6f5 d3f1 e7g6 c3c4 d5d8 e5g6 f5g6 f4e3 b7b6 a1c1 e8e7 d1a4 g6f5 a4b4 e7d7 c4c5 b6b5 a3a4 a7a6 c1a1 a5c4 f1c4 b5c4 b4c4');
INSERT INTO evaluations VALUES(23279,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(23279,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(23279,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(23279,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(23279,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(23279,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(23279,36,'cp 53','f1c4 f8c5 c2c3 d7d6 d2d4 c5b6 h1f1 g8f6 d4e5 d6e5 d1d8 c6d8 f3e5 f6e4 b1d2 e4d6 c4b3 c8e6 d2c4 b6c5 c4d6 c5d6 b3a4 c7c6 c1f4 e6d5 e5g6 h7g6 f4d6 d8f7 d6g3 d5g2 f1f2 g2h3 e1c1 e8g8 a4c2 g6g5 a2a4 a8e8 c2g6 e8e7 a4a5 f8e8');
INSERT INTO evaluations VALUES(23276,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(23276,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(23276,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(23276,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(23276,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(23276,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(23276,36,'cp 54','f8c5 c2c3 d7d6 d2d4 c5b6 h1f1 g8f6 d4e5 d6e5 d1d8 c6d8 f3e5 f6e4 a2a4 c8e6 a4a5 b6c5 b1d2 e4f6 e1d1 e8g8 d1c2 f8e8 c4e6 d8e6 d2f3 c5d6 f1e1 e6c5 c1e3 c5e4 e5c4 d6c5 e3c5 e4c5 a1d1 e8e1 d1e1 a8e8 c4e5 h7h6 b2b4 c5e4 e5d3 e4d6 e1e8 d6e8');
INSERT INTO evaluations VALUES(24075,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(24075,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(24075,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(24075,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(24075,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(24075,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24075,36,'cp 23','h2h3 g4h5 c3d5 f6d5 c4d5 e8g8 g2g4 h5g6 b2b4 c5b6 c1e3 d8e7 d5b3 c6d8 a3a4 b6e3 f2e3 d8e6 a4a5 c7c6 d1e1 g8h8 e1g3 f7f6 a5a6 b7b6 f3h4 g6h7 h4f5 h7f5 f1f5 e6c7 h3h4 d6d5 e4d5 c7d5 a1f1 d5c3 g3f3');
INSERT INTO evaluations VALUES(24074,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(24074,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(24074,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(24074,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24074,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24074,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(24074,36,'cp 15','g4h5 c3d5 f6d5 c4d5 e8g8 g2g4 h5g6 d1e2 d8e7 c1e3 c5b6 g1h1 c6d8 f1g1 c7c6 d5b3 d8e6 a1f1 g8h8 b3e6 e7e6 g1g3 f7f6 f3h4 g6h7 h4f5 d6d5 f1g1 e6d7 h3h4 b6e3 e2e3 h7f5 g4f5 f8g8 g3g6');
INSERT INTO evaluations VALUES(24070,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(24070,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24070,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24070,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(24070,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(24070,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24070,36,'cp 13','c1e3 c6d4 g2g4 f6g4 f3d4 e5d4 h3g4 h5g6 e3d2 d4c3 d2c3 h6h5 g1g2 d8e7 d1f3 h5g4 f3f4 h8h3 c3g7 e7h4 g7f6 h4h5 f2f3 g4f3 f1f3 h3f3 f4f3 e8d7 f3h3 h5h3 g2h3 c7c6 b2b4 c5b6 a1f1 a8g8 f6h4 d7e8 h4g3 e8e7 a3a4 g6h5 a4a5 b6d4 b4b5');
INSERT INTO evaluations VALUES(28664,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(28664,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(28664,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28664,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(28664,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(28664,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(28664,36,'cp 35','f3e5 c7c5 c2c4 b8d7 c1g5 f8e7 b1c3 c5d4 d1d4 e8g8 c4d5 d7e5 d4e5 h7h6 g5h4 e6d5 c3d5 b7d5 f1d1 d8e8 h4f6 e7f6 e5d5 e8e2 d1d2 e2g4 g2f3 g4g5 d5g5 f6g5 d2d7 f8d8 a1d1 a8c8 d7a7 d8d1 f3d1 c8c1 a7d7 c1a1 g1g2 a1a2 d1b3 a2b2 b3f7 g8h7 f7d5 g5f6');
INSERT INTO evaluations VALUES(31884,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(31884,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(31884,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(31884,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(31884,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(31884,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31884,36,'cp 19','f7f6 g5f4 d8d7 f4g3 b8c6 g1f3 g7g5 c2c4 c6b4 b1a3 e7e6 f1e2 c7c6 e1g1 h7h5 h2h4 d7h7 c4d5 e6d5 d1a4 g8h6 a1c1 a7a6 a3b1 f5c2 c1c2 h7c2 a4c2 b4c2 h4g5 h5h4 g3f4 f6g5 f4g5 f8e7');
INSERT INTO evaluations VALUES(32106,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(32106,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(32106,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(32106,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(32106,34,'cp 101',NULL);
INSERT INTO evaluations VALUES(32106,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(32106,36,'cp 88','d1f3 b8c6 d2d3 g8f6 c4b5 e5f4 c1f4 f6d7 f3f2 f8e7 d3d4 a7a6 b5a4 e7g5 c2c3 g5f4 f2f4 d8f6 f4g3 f6g6 g3f3 g6f6 f3d3 d7b6 a4c2 e8c8 b1d2 c8b8 e1c1 f6e7 d3g3 f7f6 d1e1 h8g8 c1b1 g8e8 h1f1 e7d7 h2h3 c6a5 c2d3');
INSERT INTO evaluations VALUES(32105,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(32105,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(32105,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(32105,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(32105,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(32105,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(32105,36,'cp 90','b8c6 d2d3 g8f6 c4b5 e5f4 c1f4 f6d7 f3f2 f8e7 d3d4 e7g5 c2c3 g5f4 f2f4 d8f6 f4e3 a7a6 b5d3 e8c8 b1d2 c8b8 e1c1 f6e7 h1e1 h7h5 c1b1 d8e8 e3f3 c6d8 e1f1 h5h4 a2a3 f7f6 d3c2 d8f7 d1e1 f7g5');
INSERT INTO evaluations VALUES(33982,30,'cp 319',NULL);
INSERT INTO evaluations VALUES(33982,31,'cp 324',NULL);
INSERT INTO evaluations VALUES(33982,32,'cp 310',NULL);
INSERT INTO evaluations VALUES(33982,33,'cp 337',NULL);
INSERT INTO evaluations VALUES(33982,34,'cp 332',NULL);
INSERT INTO evaluations VALUES(33982,35,'cp 343',NULL);
INSERT INTO evaluations VALUES(33982,36,'cp 325','b1c3 d7d6 d2d4 c8e6 c4d5 c7c6 d4e5 f6g6 d5e6 g6e6 e5d6 f8d6 c1g5 b8d7 d1d2 d6e7 e1c1 e7g5 d2g5 e8g8 f3d4 e6c4 h1e1 d7c5 d4b3 c5b3 a2b3 c4e6 f2f4 f7f6 g5c5 h6f7 h2h3 f8d8 d1d8 f7d8 f4f5 e6e5 c5e5 f6e5');
INSERT INTO evaluations VALUES(35035,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(35035,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(35035,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(35035,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(35035,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(35035,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(35035,36,'cp 126','d2d4 d7d5 g1f3 h4d8 f1d3 c7c5 c2c3 b8c6 e1g1 c5d4 c3d4 c6b4 d3e2 g8e7 b1c3 e7f5 a2a3 b4c6 b2b4 a7a6 b4b5 c6a5 e2d3 f8e7 a3a4 e8g8 g2g4 f5h4 f3h4 e7h4 d1c2 h7h6 g1h1 a5c4 f2f4 a6b5 g4g5 h6g5 f4g5 h4g5 f1g1 g5c1');
INSERT INTO evaluations VALUES(36485,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(36485,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(36485,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(36485,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(36485,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(36485,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(36485,36,'cp 141','e1c1 c7c6 h2h4 b7b5 f2f3 c8e6 g5h6 b5b4 c3b1 d8a5 h6g7 g8g7 h4h5 b8d7 a2a3 a8b8 h5g6 f7g6 b2b3 g7g8 g1h3 e6h3 h1h3 f8f7 g2g4 d7f8 f1d3 a5b6 a3b4 b6b4 d2b4 b8b4 b1c3 d6d5 g4g5 f6h5 e4d5');
INSERT INTO evaluations VALUES(37456,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37456,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37456,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37456,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37456,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37456,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37456,36,'cp 0','d7d5 b1d2 c7c5 d4c5 b8d7 b2b4 a7a5 c2c3 a5b4 c3b4 f6e4 f3d4 e4c3 d1e1 e7e5 d4b3 e5e4 b3d4 d7e5 c1b2 c3a4 b2c1');
INSERT INTO evaluations VALUES(37705,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(37705,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(37705,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(37705,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(37705,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(37705,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(37705,36,'cp -11','g8f6 f1c4 d7d6 e1g1 h7h6 b1c3 c8e6 c4e6 f7e6 d1d3 e8g8 c1e3 c5e3 d3e3 b8c6 a1d1 d8e7 c3e2 a8b8 c2c4 b7b5 c4b5 b8b5 b2b3 d6d5 d1c1 e7d6 e2g3 d5d4 e3e2 b5c5 e2a6 f8b8 h2h3 g8h7 c1c5 d6c5 g3e2 b8b6 a6c8');
INSERT INTO evaluations VALUES(40005,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(40005,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(40005,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(40005,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40005,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(40005,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(40005,36,'cp 12','d7d5 d2d4 c5d4 f3d4 e8g8 c4d5 f6d5 b1c3 d5c3 b2c3 b8c6 d4c6 b7c6 d1d8 f8d8 g2c6 a8b8 c1f4 b8b2 f1d1 d8f8 a2a4 b2e2 a1c1 e2a2 d1d2 a2d2 f4d2 c8e6 c6b5 f8c8 d2e3 a7a5 c3c4 g7b2 c1b1 b2a3 b5a6 c8c7 b1b8 g8g7 e3b6 c7d7 b6a5');
INSERT INTO evaluations VALUES(40068,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(40068,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(40068,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(40068,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(40068,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(40068,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(40068,36,'cp 60','e4e5 f6e4 c3e4 d5e4 g5e7 d8e7 g1e2 c6c5 e2c3 b8c6 c3e4 c6d4 e4d6 e8f8 d1d2 b7b6 h2h4 c8b7 h4h5 a8d8 e1c1 b7c6 f2f4 d8d6 e5d6 e7d6 h1h3 f8e7 d2f2 h8d8 d1d2 f7f5 h3a3 a7a5 a3e3 e7f7 g2g3 c6b7 h5h6 g7h6 f1g2 b7g2 f2g2 a5a4 c1b1');
INSERT INTO evaluations VALUES(40092,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(40092,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(40092,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(40092,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(40092,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(40092,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(40092,36,'cp 65','f6e4 g5e7 d8e7 c4d5 c6d5 f1d3 e8g8 d3e4 d5e4 g1e2 b8c6 e1g1 e7h4 d1a4 c8d7 a4c2 c6b4 c2b3 b4c6 b3b7 f8d8 a1b1 a8b8 b7a6 d7e8 h2h3 c6d4 e2d4 d8d4 a6a7 b8d8 a7c5 h7h6 a2a3 d4d3 b2b4 h4f4 b4b5');
INSERT INTO evaluations VALUES(42315,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(42315,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(42315,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(42315,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(42315,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(42315,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(42315,36,'cp 43','h7h6 g5f4 g6g5 f4d2 d7d6 g2g3 b8c6 e2e3 e6e5 d4d5 e5e4 d5c6 e4f3 d1f3 a8b8 h2h4 g5g4 f3d1 b7c6 b2b3 d6d5 c4d5 f6d5 a1c1 d5b4 f1e2 e8g8 e1g1 b4d3 e2d3 d8d3 c3a4 c8a6 f1e1 f8d8 a4c5 d3d2 c5a6 d2a2 a6b8 d8d1 e1d1 c6c5');
INSERT INTO evaluations VALUES(42905,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(42905,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(42905,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(42905,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(42905,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(42905,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(42905,36,'cp 87','d1d5 e4f6 d5e5 d8e7 e5e7 e8e7 f3d4 d7d6 f2f3 b8c6 e1f2 a7a6 f1d3 c8e6 d4c6 b7c6 h1e1 a6a5 b2b4 h8b8 c1d2 h7h6 g2g4 e7f8 f2g3 f6d7 f3f4 d7f6 f4f5 e6d7 c3c4 a5b4 a3b4 c6c5 b4c5 a8a1');
INSERT INTO evaluations VALUES(45196,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45196,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45196,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45196,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45196,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(45196,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(45196,36,'cp -4','b8c6 f1d3 e7e5 d3c2 d6d5 e4d5 d8d5 d1e2 c8e6 d2d3 e8c8 e1g1 h7h6 f1d1 h8g8 c2b3 d5d6 b1a3 g7g5 a3c4 e6c4 b3c4 e5e4 f3h2 c6e5 d3d4 g5g4 h3g4 e5c4 e2c4 d6c7 c4e2 c5d4 d1d4 d8d4 c3d4 f8d6 g2g3 d6g3 f2g3 c7g3 e2g2 g3d3 g2f2 f6g4 h2g4 g8g4 g1h2');
INSERT INTO evaluations VALUES(46427,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(46427,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(46427,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(46427,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(46427,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(46427,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46427,36,'cp 23','b8c6 f1g2 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 h7h5 f2f3 h5h4 g3g4 d8a5 d1c2 a5a2 e2e3 e8b8 c5d4 a7a5 c2f2 h4h3 g2h3 a2b3 c1a1 a5a4 f1c1 f6h7 h3f1 h7g5 h2h4 g5e6');
INSERT INTO evaluations VALUES(13748,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(13748,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(13748,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(13748,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(13748,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(13748,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(13748,36,'cp 38','f1g2 c5d4 f3d4 f8c5 d4c6 b7c6 e1g1 e8g8 c3a4 c5d6 c1e3 f8e8 a1c1 c8d7 a4c5 d6c5 e3c5 f6e4 c5d4 a7a5 f2f3 e4d6 d1d2 a5a4 c1c2 d6c4 d2c1 a8b8 e2e4 h7h6 f1d1 d8e7 g2f1 c4e5 c1f4 e7g5 d4e5 g5e5 f4e5 e8e5 e4d5 e5d5 d1d5');
INSERT INTO evaluations VALUES(46474,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46474,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46474,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46474,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(46474,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46474,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(46474,36,'cp 0','c1g5 f8b4 d1a4 b8c6 g1h3 h7h6 g5f6 d8f6 e1g1 f6d6 f2f3 e4f3 f1f3 e8g8 a2a3 c8h3 g2h3 b4c3 f3c3 a8e8 e2e3 e8e7 h3g2 f8e8 a4b3 e7d7 g2h3 d7e7');
INSERT INTO evaluations VALUES(46467,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(46467,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(46467,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46467,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(46467,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(46467,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46467,36,'cp 0','b8d7 d1b3 f8d6 b3b5 e8g8 c3d5 a7a6 b5a4 f8e8 a2a3 b7b5 a4c2 h7h6 d5f6 d7f6 g5f6 d8f6 e2e3 c8d7 g1e2 a8c8 e2c3 b5b4 a3b4 d7b5 g2e4 e8e4 c2e4 c8e8 e4d5 d6b4 h1g1 e8c8 f2f3 b5c6 d5c4 f6f3 e1d2 b4c3 b2c3 c8b8 a1f1 f3e4 c4a2 c6d5 a2c2 e4c2 d2c2 d5e4 c2c1 b8b1 c1d2 b1b2 d2e1 b2b1');
INSERT INTO evaluations VALUES(49513,30,'cp -194',NULL);
INSERT INTO evaluations VALUES(49513,31,'cp -199',NULL);
INSERT INTO evaluations VALUES(49513,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(49513,33,'cp -187',NULL);
INSERT INTO evaluations VALUES(49513,34,'cp -207',NULL);
INSERT INTO evaluations VALUES(49513,35,'cp -233',NULL);
INSERT INTO evaluations VALUES(49513,36,'cp -228','f8e7 g1f3 e8g8 e5a5 d7d5 e2e3 b8c6 a5d2 f6e4 d2d1 d8d6 f1e2 d6g6 e1g1 f8d8 b1d2 e7d6 f3h4 g6h6 g2g3 c8h3 h4g2 a7a5 d2f3 c6e7 c2c4 c7c6 a2a3 d8e8 f3d2 a8d8 c4d5 c6d5 d2e4 d5e4 c1d2 d6c7');
INSERT INTO evaluations VALUES(43023,30,'cp -240',NULL);
INSERT INTO evaluations VALUES(43023,31,'cp -217',NULL);
INSERT INTO evaluations VALUES(43023,32,'cp -205',NULL);
INSERT INTO evaluations VALUES(43023,33,'cp -210',NULL);
INSERT INTO evaluations VALUES(43023,34,'cp -213',NULL);
INSERT INTO evaluations VALUES(43023,35,'cp -201',NULL);
INSERT INTO evaluations VALUES(43023,36,'cp -218','c1g5 e8g8 b1c3 d7d5 e2e3 f8e8 e5g3 c7c6 f1d3 d5d4 g5f6 e7f6 c3e4 d4e3 e1c1 f6b2 c1b2 e8e4 d3e4 d8d1 g3e3 b8d7 e4d3 d7f6 f2f3 c8d7 e3e2 d1e2 g1e2 g8f8 e2g3 a8d8 g3e4 f6e4 d3e4 h7h6 b2c3 d8e8 c3d4 b7b6 h1e1 e8d8 d4e5');
INSERT INTO evaluations VALUES(43011,30,'cp -199',NULL);
INSERT INTO evaluations VALUES(43011,31,'cp -214',NULL);
INSERT INTO evaluations VALUES(43011,32,'cp -215',NULL);
INSERT INTO evaluations VALUES(43011,33,'cp -242',NULL);
INSERT INTO evaluations VALUES(43011,34,'cp -230',NULL);
INSERT INTO evaluations VALUES(43011,35,'cp -217',NULL);
INSERT INTO evaluations VALUES(43011,36,'cp -232','e8g8 b1c3 d7d5 e2e3 f8e8 e5g3 c7c6 f1d3 h7h6 g5f6 e7f6 g1e2 f6h4 g3f4 h4g5 f4f3 b8d7 e1g1 d7e5 f3g3 h6h5 h2h3 h5h4 g3h2 g5f6 a1d1 e5d3 c2d3 f6e5 f2f4 e5c3 b2c3 e8e3 e2d4 e3g3 f4f5 c8d7 d4e2 d8g5 g1h1 g3e3 e2d4 c6c5 d4f3 g5f5 h2c7 e3d3');
INSERT INTO evaluations VALUES(50913,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(50913,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(50913,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(50913,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(50913,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(50913,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(50913,36,'cp 37','g8f6 e1g1 f8e7 f1e1 b8c6 d2d4 e5d4 f3d4 c6d4 d1d4 e8g8 b1c3 b7b5 c4f1 a8b8 c1f4 b5b4 c3e2 f6d7 c2c3 e7f6 d4d2 d7c5 e2d4 b4c3 b2c3 c8b7 e4e5 d6e5 f4e5 f8e8 e5f6 d8f6');
INSERT INTO evaluations VALUES(51698,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51698,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51698,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(51698,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51698,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51698,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51698,36,'cp 0','b5b4 c3b1 c8a6 c1g5 d8a5 b1d2 c4c3 d2b3 a5b6 f1d3 c3b2 a1b1 h7h6 g5e3 g8f6 b1b2 b8d7 b2c2 f8e7 f3e5 a8c8 e5c4 a6c4 c2c4 b6b8 c4c1 e8g8 e1g1 f8d8 f2f4 d7b6 c1a1 b6d7');
INSERT INTO evaluations VALUES(51785,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(51785,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51785,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(51785,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(51785,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(51785,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(51785,36,'cp 13','d7d5 g1f3 d5c4 b1a3 f8a3 b2a3 g8f6 c1b2 b8d7 a1c1 d8b6 c1c2 b6a6 d1c1 b7b5 d2d3 c4d3 c2c6 a6a5 c1c3 a5c3 c6c3 c8b7 e2d3 b7d5 e1g1 e8g8 c3c7 f8d8 f3d4 d7b6 f1c1 f6e8 c7e7 d5g2 g1g2 a7a6 c1c6 b6a4 b2a1 h7h6');
INSERT INTO evaluations VALUES(52148,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52148,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(52148,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(52148,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(52148,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52148,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52148,36,'cp -9','c5d4 g1f3 f8c5 f3d4 g8f6 b1c3 e8g8 f1e2 d7d5 e4d5 f6d5 e1g1 d5c3 b2c3 b8c6 d4c6 b7c6 d1a4 d8e7 e2f3 a8b8 f3c6 f8d8 c6e4 e7f6 a4c2 g7g6 f1e1 c8e6 c1e3 d8c8 e4f3 c8c7 c2a4 c5e3 e1e3 c7c3 a4a7 c3e3 a7b8 g8g7');
INSERT INTO evaluations VALUES(53373,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(53373,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(53373,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(53373,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(53373,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(53373,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(53373,36,'cp -55','g8f6 d4f3 b8c6 c2c4 e7e5 c4d5 f6d5 e3e4 f8b4 c1d2 b4d2 d1d2 d5c7 d2d8 e8d8 b1c3 d8e7 c3b5 c7e6 e1c1 c8d7 b5c3 h8c8 c1b1 c8d8 f1c4 f7f6 d1d2 e6f4 h1c1 d7e6 c3d5 f4d5 c4d5 e6d5 d2d5 a8c8 c1c3 g7g5 a2a3 h7h5 d5d8 c8d8 b1c1 h5h4');
INSERT INTO evaluations VALUES(55468,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(55468,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(55468,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(55468,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(55468,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(55468,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(55468,36,'cp 26','b2b3 g8f6 c1b2 b8d7 f1d3 b7b6 e1g1 c8b7 c2c4 f8d6 c4d5 e6d5 f3e5 e8g8 b1d2 c7c5 e5d7 f6d7 a1c1 d8e7 f1e1 f8e8 d2f1 c5d4 b2d4 d7c5 f1g3 c5d3 d1d3 e7g5 e1d1 a8c8 c1c8 e8c8 g3f5 d6f8 h2h3');
INSERT INTO evaluations VALUES(55699,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(55699,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55699,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(55699,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(55699,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(55699,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(55699,36,'cp 23','c7c6 b5e2 g8f6 b1c3 b7b5 a2a3 b8d7 d2d4 e6e5 e1g1 f8e7 d4d5 c6d5 c3d5 a7a6 d5e7 d8e7 c2c4 a8b8 f3h4 d7c5 c4b5 a6b5 f2f3 e8g8 c1e3 c8e6 a1c1 f6e4 f3e4 e7h4 d1d6 c5b3 c1c3 b5b4 a3b4 h4e4 e2f3 e4b4 d6e5 b8d8 e3g5 b4b6 g1h1 f7f6 g5e3 f6e5 e3b6 d8d7 b6g1');
INSERT INTO evaluations VALUES(57057,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(57057,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(57057,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(57057,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(57057,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(57057,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(57057,36,'cp -36','d2d4 c7c5 d4d5 e6d5 e2e3 g8f6 g1e2 f8d6 b1c3 d6e5 g2g3 b8c6 c3d5 f6d5 d1d5 d7d6 f1g2 c8e6 d5d3 e8g8 e1g1 c6b4 d3d1 b4a2 c1d2 e6c4 a1b1 d8b6 f1e1 a2b4 e2f4 c4a2 d2b4 a2b1 f4d5 b6d8 f2f4 c5b4 f4e5');
INSERT INTO evaluations VALUES(57015,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(57015,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(57015,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(57015,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(57015,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(57015,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(57015,36,'cp -23','c7c5 d4d5 e6d5 e2e3 d5d4 e3d4 c5d4 g1f3 f8b4 c1d2 b4c5 c2c3 g8e7 c3d4 c5b6 b1c3 d7d5 g2g3 b8c6 f1g2 e8g8 c3e2 c8f5 e1g1 f8e8 d2c3 h7h6 f1e1 e7g6 a2a3 f5e4 f3h2 a8c8 h4h5 g6f8 e2f4 f8e6 g2e4 e6f4 e4f3 f4e6');
INSERT INTO evaluations VALUES(1089,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(1089,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1089,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1089,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1089,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1089,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(1089,36,'cp 0','h7h6 g5f3 c6a5 c4b3 g7g6 c2c4 f8g7 e1g1 e8g8 b1c3 c7c5 h2h3 d7d6 c1d2 a5c6 a2a3 a7a5 b3a4 e7d8 c3d5 g8h7 a1b1 h7g8 f1e1 g8h7 e1f1');
INSERT INTO evaluations VALUES(1372,30,'cp 207',NULL);
INSERT INTO evaluations VALUES(1372,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(1372,32,'cp 195',NULL);
INSERT INTO evaluations VALUES(1372,33,'cp 211',NULL);
INSERT INTO evaluations VALUES(1372,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(1372,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(1372,36,'cp 198','c1d2 e7e5 d4c4 a5b6 b1c3 g8f6 e1c1 f8e7 d2e3 b6d8 f2f4 f6g4 e3c5 e7c5 c4c5 d8e7 c5c7 e7d8 c7d6 d8e7 d6d2 e7c5 d2e2 g4f6 f4e5 c5e5 e2b5 e5b5 c3b5 e8e7 f1c4 f6g4 b5c7 a8b8 c7d5 e7d8 h1f1');
INSERT INTO evaluations VALUES(1373,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(1373,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(1373,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(1373,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(1373,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(1373,35,'cp 223',NULL);
INSERT INTO evaluations VALUES(1373,36,'cp 225','e7e5 d4c4 a5c5 b1c3 c5c4 f1c4 a8b8 c3b5 a7a6 b5c7 e8d8 c7d5 f8c5 f2f4 g8f6 d5f6 g7f6 f4e5 f6e5 c4f7 b7b5 e1c1 c8b7 d1e1 d8e7 h1f1 h7h6 g2g4 b8f8 f7g6 b7c8 f1f8 e7f8 g6f5 f8g7 h2h4 d7d6 e1f1 h8f8 b2b4 c5d4 c2c3 d4b6 c1c2 c8b7 f1d1');
INSERT INTO evaluations VALUES(3630,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(3630,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3630,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3630,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(3630,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3630,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(3630,36,'cp 9','c2c3 a7a6 c4b3 c8e6 b3e6 f7e6 b2b4 c5b6 c1e3 f8f7 b1d2 d8d7 c3c4 b6e3 f2e3 a8f8 f1f2 d7e7 d1c2 g7g5 f3h2 h6h5 a1f1 g5g4 h3g4 f6g4 f2f7 f8f7 f1f7 g8f7 d2f1 e7g5 b4b5 c6e7 h2g4 h5g4 c2f2 f7g8');
INSERT INTO evaluations VALUES(4096,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(4096,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(4096,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(4096,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(4096,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(4096,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(4096,36,'cp 39','e1g1 f6g4 e3d2 e7e5 d2g5 g4f6 d1d2 c7c6 f1e1 a7a5 b2b3 e5d4 f3d4 d8b6 d3f1 f8e8 a1d1 d7c5 f2f3 a5a4 g5e3 b6a5 d4e2 f6d7 d1b1 a4b3 a2b3 c5e6 e1d1 a5b4 e3f2 d7c5 h2h3 e8d8 d2e3 b7b5 f3f4 c8b7 f4f5');
INSERT INTO evaluations VALUES(4095,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(4095,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(4095,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(4095,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(4095,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(4095,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(4095,36,'cp 45','f6g4 e3d2 e7e5 d2g5 g4f6 d1d2 c7c6 f1e1 a7a5 h2h3 a5a4 a1b1 f8e8 b2b4 a4b3 a2b3 e5d4 f3d4 d7c5 g5h6 c5e6 h6g7 g8g7 d4e6 c8e6 c3e2 d8b6 b3b4 f6d7 d2c3 g7g8 e2d4 e6a2 b1d1 d6d5 d4f3 h7h6 c3b2 d5e4 d3e4');
INSERT INTO evaluations VALUES(4990,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(4990,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(4990,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(4990,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(4990,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(4990,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(4990,36,'cp 9','g8f6 b1c3 d7d6 c1e3 c5b4 e4e5 f6d5 f1b5 c8d7 b5c4 e8g8 c4d5 c6d5 f3d5 d7e6 d5d4 c7c5 d4d6 d8a5 e1g1 f8d8 d6e7 d8d7 e7h4 b4c3 b2c3 a5c3 a1d1 c3c2 d1c1 c2b2 c1c5 a8d8 h4a4 b2a2 a4a2 e6a2 f1a1 a2e6');
INSERT INTO evaluations VALUES(6130,30,'cp 557',NULL);
INSERT INTO evaluations VALUES(6130,31,'cp 548',NULL);
INSERT INTO evaluations VALUES(6130,32,'cp 542',NULL);
INSERT INTO evaluations VALUES(6130,33,'cp 562',NULL);
INSERT INTO evaluations VALUES(6130,34,'cp 572',NULL);
INSERT INTO evaluations VALUES(6130,35,'cp 571',NULL);
INSERT INTO evaluations VALUES(6130,36,'cp 575','c4f7 e8f8 f3g5 d7b6 a2a4 d8d7 f7e6 d7c6 a4a5 h7h6 d4d5 c6e8 e6c8 h6g5 c8b7 b6d7 b7a8 e8a8 c1g5 g7b2 a1a4 b2f6 g5e3 c7c6 e1g1 g8h6 d5c6 a8c6 a4b4 f8g7 f2f4 h8c8 h2h3 c6c2 b1d2 c8c3 d1c2 c3c2');
INSERT INTO evaluations VALUES(6131,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(6131,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(6131,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(6131,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(6131,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(6131,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(6131,36,'cp 101','e2e4 g8f6 f1e2 e7e5 c1e3 e8g8 e1g1 d8e7 d1c2 e5d4 f3d4 f8e8 f2f3 c7c6 a1d1 a7a6 e3g5 d7c5 f1e1 c5e6 d4e6 c8e6 c2d2 a8d8 a2a4 e7c7 g1h1 d8d7 g5f4 c7b6 f4d6 d7d6 d2d6 b6b2 c3b1 f6d7 d6d2 b2f6 a4a5 d7e5 d2c2 f6e7 b1d2');
INSERT INTO evaluations VALUES(6829,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(6829,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(6829,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(6829,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(6829,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(6829,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(6829,36,'cp 21','e1g1 c7c6 b5a4 d7d5 b2b4 c5b6 b4b5 g8e7 b5c6 e8g8 e4d5 d8d5 c1a3 b6c5 d1b3 c5a3 b1a3 d4c3 f1e1 d5b3 a2b3 e7c6 d2c3 c8e6 a3b5 a8c8 b5d6 c8c7 e1d1 a7a5 d6b5 c7c8 f2f3 f8d8 d1d6 g7g6 a1d1 d8d6 b5d6 c8c7 c3c4');
INSERT INTO evaluations VALUES(6828,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(6828,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(6828,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(6828,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(6828,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(6828,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(6828,36,'cp 20','c7c6 b5a4 g8e7 d2d3 d7d5 b1d2 c5b6 e4d5 e7d5 f1e1 c8e6 d2e4 d4c3 b2c3 e8g8 h2h3 b6c7 d1f3 h7h6 d3d4 f8e8 c1d2 d5b6 a4b3 e6d5 f3d3 b6d7 a1d1 e8e4 e1e4 d5b3 a2b3 d7c5 d3f3 c5e4 f3e4 d8f6 c3c4 f6g6 e4g6 f7g6');
INSERT INTO evaluations VALUES(7027,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(7027,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(7027,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(7027,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(7027,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(7027,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(7027,36,'cp 94','d4d5 c6e5 f3e5 g7e5 e2e4 d7d6 f1d3 g8f6 b1c3 e7e6 e1g1 f6d7 d1e1 e8g8 f2f4 e5g7 c1e3 b7b6 e1d2 d7c5 d3c2 e6d5 c3d5 c7c6 d5c3 c8a6 f1d1 a6c4 d2d6 a8c8 e4e5 d8d6 d1d6 f8d8 e3c5 b6c5 a1d1 d8d6 d1d6 c4e6 c2b3 e6b3');
INSERT INTO evaluations VALUES(7214,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(7214,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(7214,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(7214,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(7214,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(7214,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(7214,36,'cp -41','h7h6 g5h4 d8e7 d1e2 c8e6 e2e3 g7g5 h4g3 e6g4 h2h3 g4h5 h3h4 d6c5 e3e2 e8c8 g3e5 g5h4 e5f6 e7f6 e4e5 f6e7 e2e4 h5f3 g2f3 c5d4 e1c1 d4f2 f3f4 f2g3 d3d4 f7f6 a2a3 f6e5 f4e5 e7e6 c1b1 c8b8 d1f1 h4h3 e4f3');
INSERT INTO evaluations VALUES(7157,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(7157,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(7157,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(7157,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(7157,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(7157,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(7157,36,'cp -41','g5d2 c8e6 d1e2 d8e7 c3d1 f6d7 d1e3 e8c8 h2h4 c6c5 d2c3 f7f6 e3f5 e7f7 h4h5 d7b8 f3h4 b8c6 e1g1 c8b8 f5d6 c7d6 f2f4 h8g8 h4g6 f7d7 e2f2 e6g4 f4e5 d6e5 f2c5 d8c8 c5b6 g4h5 g6h4 h5g4 f1f2 g4e6 a1f1 h6h5 a2a4 g7g6 a4a5 g6g5');
INSERT INTO evaluations VALUES(7858,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(7858,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(7858,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(7858,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(7858,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(7858,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(7858,36,'cp -44','b4c3 b2c3 d8a5 f1d3 a5c3 f3d2 d7d6 e1g1 c8g4 d1b1 c6a5 d2b3 a5c4 d3c4 c3c4 b3a5 c4c7 g5f6 g7f6 b1b7 a8c8 b7b4 c7c3 a1b1 e8e7 f1c1 g4d7 b1b3 c3b4 b3b4 f6f5 e4f5 d7f5 a5c4 h8g8 c4e3 e7e6 e3f5 e6f5 b4b7');
INSERT INTO evaluations VALUES(9609,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(9609,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(9609,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(9609,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(9609,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(9609,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(9609,36,'cp 98','h2h3 e7e5 f3e5 h7h6 g5f6 d7f6 e5f3 e8g8 f1d3 c7c5 d4c5 d6c5 e1g1 c8e6 d1e2 d8b6 a1b1 f6h5 f1e1 h5f4 e2e3 f4d3 c2d3 f8e8 a2a3 b6a6 e4e5 a8d8 b1d1 e6b3 d1d2 b3d5 c3e4 d5e4 e3e4 a6b6 e4e3 b6c7');
INSERT INTO evaluations VALUES(13083,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13083,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(13083,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(13083,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(13083,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(13083,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(13083,36,'cp 10','c5d4 e3d4 c8g4 c1g5 f6e4 c4d5 e4g5 d5c6 b7c6 f1e2 g5f3 e2f3 g4f3 d1f3 a8c8 a1d1 e7e6 f3g3 d8d6 d1d3 h7h5 g3g5 d6d8 g5e5 h5h4 e1g1 d8f6 f1d1 f8e7 e5a5 h4h3 g2g3 c8b8 b2b3 e7b4 a5a7 e8g8 a7c7 b4c3 d3c3 b8d8 c7c6 d8d4 d1d4 f6d4 c6f3 d4e5');
INSERT INTO evaluations VALUES(17130,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(17130,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(17130,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(17130,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(17130,34,'cp 133',NULL);
INSERT INTO evaluations VALUES(17130,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(17130,36,'cp 133','d2d4 c6d4 f3d4 e5d4 d1d4 e7g6 c1e3 b7b6 f1d1 g6e5 b1c3 c7c6 b5e2 f8c5 d4d2 c5e3 d2e3 b6b5 e2h5 g7g6 h5e2 e8g8 e3d4 d8b6 d4d6 g8g7 b2b4 e5f7 d6e7 b6d8 e7d8 f8d8 a2a4 b5a4 a1a4 a7a6 f2f4 d7d6 d1a1 g6g5 g2g3 f6f5 e2a6 g5f4 a6c8 a8a4 a1a4');
INSERT INTO evaluations VALUES(17316,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(17316,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(17316,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(17316,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(17316,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(17316,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(17316,36,'cp 21','h8g8 b1c3 f8c5 c3d5 d7d6 e1g1 e8f7 c2c3 c6e7 d2d4 e7d5 d1b3 c5b6 a2a4 a7a5 c1e3 g8e8 d4e5 b6e3 f2e3 d6e5 e4d5 f7g8 e3e4 a8a6 c3c4 a6b6 b3c2 c7c6 b2b3 b6b4 a1d1 c6d5 e4d5 d8d6 d1e1 c8d7 c2c3 b7b6 h2h3 d6c5 g1h1');
INSERT INTO evaluations VALUES(23502,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(23502,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(23502,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(23502,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(23502,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(23502,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(23502,36,'cp 23','b1c3 f8c5 e1g1 e8f7 c3d5 d7d6 a2a4 g8e8 c2c3 a7a6 d2d4 c5a7 d4e5 d6e5 a4a5 c8e6 c1e3 a7e3 f2e3 f7g8 c3c4 c6e7 d1b3 c7c6 d5b6 a8b8 f1d1 d8c7 b3c2 b8d8 c4c5 e7c8 b2b4 d8d1 a1d1 c8b6 a5b6');
INSERT INTO evaluations VALUES(19693,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(19693,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(19693,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(19693,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(19693,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(19693,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(19693,36,'cp 69','b1c3 g8f6 f2f3 f8g7 d1d2 e8g8 g2g4 c6d4 e3d4 d8a5 h2h4 h7h5 g4g5 f6e8 d4g7 e8g7 c3d5 a5d8 e1c1 c8e6 d5e3 d8c7 c1b1 a8c8 f1d3 f8d8 c2c4 b7b5 c4b5 d6d5 d1c1 c7e5 c1c8 d8c8 e4d5 e6d5 e3d5 e5d5 h1d1 d5f3 d2e2');
INSERT INTO evaluations VALUES(20823,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(20823,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(20823,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(20823,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(20823,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(20823,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(20823,36,'cp -9','e7f6 f1e2 c8f5 c2c3 d8d7 e1g1 f8d6 a2a4 g8e7 d2d4 e8g8 f3h4 f5e6 b1a3 a7a6 b2b4 e7g6 h4g6 h7g6 a3c2 f8e8 c2e1 e6f5 e1d3 g6g5 a4a5 b7b6 f1e1 b6b5 e2f1 e8e1 d3e1 a8e8 h2h3 c6e7 e1d3 c7c6 d3c5 d6c5 d4c5');
INSERT INTO evaluations VALUES(22329,30,'cp -244',NULL);
INSERT INTO evaluations VALUES(22329,31,'cp -265',NULL);
INSERT INTO evaluations VALUES(22329,32,'cp -264',NULL);
INSERT INTO evaluations VALUES(22329,33,'cp -265',NULL);
INSERT INTO evaluations VALUES(22329,34,'cp -272',NULL);
INSERT INTO evaluations VALUES(22329,35,'cp -259',NULL);
INSERT INTO evaluations VALUES(22329,36,'cp -258','f6e5 d1h5 e8d7 h5e5 d8f6 e5h5 b7b6 d2d3 f6e6 f2f4 g7g6 h5f3 f8g7 b1c3 c8b7 c1d2 a8f8 e1c1 g8e7 g2g4 d7c8 c1b1 c6c5 h1f1 e6d7 c3e2 h8g8 f3g3 e7c6 h2h3 f8e8 d2c3 c6d4 e2d4 c5d4 c3d2 b7c6 f4f5 g8f8 d1e1');
INSERT INTO evaluations VALUES(22337,30,'cp -269',NULL);
INSERT INTO evaluations VALUES(22337,31,'cp -271',NULL);
INSERT INTO evaluations VALUES(22337,32,'cp -256',NULL);
INSERT INTO evaluations VALUES(22337,33,'cp -259',NULL);
INSERT INTO evaluations VALUES(22337,34,'cp -256',NULL);
INSERT INTO evaluations VALUES(22337,35,'cp -273',NULL);
INSERT INTO evaluations VALUES(22337,36,'cp -272','d1h5 e8d7 h5e5 d8f6 e5h5 b7b6 d2d3 f6e6 f2f4 g7g6 h5f3 f8g7 b1c3 c8b7 c1d2 a8f8 e1c1 d7c8 g2g4 c6c5 h1f1 e6d7 c3e2 h7h5 h2h3 h5g4 h3g4 g8h6 f1g1 h6f7 f3g3 c5c4 d2c3 f7d6 c3g7 d7g7 c1b1 c4d3 c2d3 g7d7 f4f5');
INSERT INTO evaluations VALUES(22336,30,'cp -269',NULL);
INSERT INTO evaluations VALUES(22336,31,'cp -257',NULL);
INSERT INTO evaluations VALUES(22336,32,'cp -251',NULL);
INSERT INTO evaluations VALUES(22336,33,'cp -260',NULL);
INSERT INTO evaluations VALUES(22336,34,'cp -275',NULL);
INSERT INTO evaluations VALUES(22336,35,'cp -271',NULL);
INSERT INTO evaluations VALUES(22336,36,'cp -277','e8d7 h5e5 d8f6 e5h5 b7b6 d2d3 f6e6 f2f4 g7g6 h5f3 f8g7 b1c3 c8b7 c1d2 a8f8 e1c1 d7c8 g2g4 g8e7 f3g3 e6d7 c1b1 c6c5 h2h3 h8g8 h1f1 e7c6 g3h4 g6g5 h4h7 g8h8 h7g6 g5f4 f1f4 c6e5 f4f8 h8f8 g6h7 f8h8');
INSERT INTO evaluations VALUES(22374,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22374,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(22374,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(22374,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(22374,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(22374,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22374,36,'cp 84','d2d4 e5d4 d1d4 g8f6 b1c3 c7c6 c1e3 d6d5 e4e5 f6g4 e3g1 d8b6 d4b6 a7b6 h2h3 g4h6 g2g4 h6g8 g1d4 h7h5 h1g1 f8c5 e1c1 g8e7 d1e1 c5d4 f3d4 h5g4 h3g4 g7g6 f4f5 g6f5 e5e6 d7c5 g4f5 f7e6 b2b4 c5e4 c3e4 d5e4');
INSERT INTO evaluations VALUES(24298,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(24298,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(24298,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(24298,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(24298,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(24298,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(24298,36,'cp 43','c2c3 f8e7 h2h3 g4h5 d2d3 b7b5 a4b3 e8g8 b1d2 c6a5 b3c2 c7c5 d2f1 f8e8 f1g3 h5g6 f3h4 d6d5 e4d5 d8d5 c1g5 a8c8 a2a3 c5c4 g3f5 c4d3 d1d3 e7f8 d3d5 f6d5 a1d1 d5b6 g5c1 g6f5 h4f5 c8c7 f5e3 g7g6 e3d5');
INSERT INTO evaluations VALUES(24296,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(24296,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(24296,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(24296,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(24296,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(24296,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(24296,36,'cp 43','f8e7 h2h3 g4h5 d2d3 e8g8 b1d2 b7b5 a4b3 c6a5 b3c2 c7c5 d2f1 a5c6 f1g3 h5g6 f3h4 d6d5 e4d5 f6d5 h4g6 h7g6 d1g4 d8c8 g3e4 f8d8 g4g3 f7f5 e4g5 f5f4 g3h4 e7g5 h4g5 c8f5 g5g4 f5g4 h3g4 d5f6');
INSERT INTO evaluations VALUES(23944,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(23944,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(23944,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(23944,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(23944,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(23944,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(23944,36,'cp 54','h2h3 g4h5 d2d3 e8g8 b1d2 h7h6 d2f1 f6h7 g2g4 h5g6 f1g3 e7g5 d3d4 g5c1 a1c1 h7g5 f3g5 d8g5 a4c6 b7c6 d4e5 d6e5 d1e2 a8d8 c1d1 a6a5 g1g2 g8h7 d1d8 f8d8 e1d1 d8d1 e2d1 f7f6 b2b4 a5b4 c3b4 c6c5 b4b5 g5f4 d1f3 c5c4');
INSERT INTO evaluations VALUES(25123,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(25123,31,'cp 147',NULL);
INSERT INTO evaluations VALUES(25123,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(25123,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(25123,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(25123,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(25123,36,'cp 142','d2d4 f8b4 a2a3 b4a5 f1e2 f6e4 d1d3 e4c3 b2c3 d8d7 e1g1 e8g8 a1b1 e6f5 d3d2 a7a6 e2d3 b7b6 f1e1 f8e8 c1b2 b6b5 b1a1 c6e7 f3e5 d7d5 a3a4 f5d3 e5d3 d5c6 d3e5 c6d5 a4b5 a6b5 e5d3 d5c6 d3f4 f7f6');
INSERT INTO evaluations VALUES(25121,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(25121,31,'cp 140',NULL);
INSERT INTO evaluations VALUES(25121,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(25121,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(25121,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(25121,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(25121,36,'cp 135','f8b4 a2a3 b4a5 f1e2 f6e4 d1d3 e4c3 b2c3 c6e7 e1g1 e8g8 c1d2 e6f5 d3b5 b7b6 f1e1 f5c2 f3e5 a7a6 b5b2 c2e4 e2c4 e4d5 c4d5 e7d5 e5c4 d8f6 e1e5 f6c6 c4a5 b6a5 b2b3 f8d8 c3c4 d5b6 e5c5 c6d6 d2c3 a5a4 b3c2 d8e8');
INSERT INTO evaluations VALUES(26278,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(26278,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(26278,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(26278,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(26278,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(26278,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(26278,36,'cp 0','d4c5 c6e5 f1b5 f5d7 d1d5 d7b5 d5e5 f7f6 e5e6 b5d7 e6e2 e7e5 b2b4 d7f5 e2b5 d8d7 b5d7 f5d7 b1d2 b7b6 c5b6 a7b6 d2c4 d7c6 f2f3 b6b5 c4a5 c6d7 g1e2 f8d6 a5b7 d6c7 b7c5 g8e7 c5d7 e8d7 c1e3 e7d5 e3c5 d7c6 e1g1 a8a3');
INSERT INTO evaluations VALUES(28902,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(28902,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(28902,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(28902,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(28902,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(28902,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(28902,36,'cp 60','e2e3 e7e6 d1b3 d8c8 a1c1 a7a5 f3e5 b8d7 e5d7 c8d7 c4c5 d7c8 c3a4 f6d7 f4e5 f7f6 e5g3 f8e7 f2f3 e8g8 e3e4 f5h7 f1d3 f6f5 e4e5 c8c7 e1g1 g7g5 b3c2 f8f7 g3f2 h6h5 a2a3 f7g7 b2b4 f5f4 d3h7 g7h7');
INSERT INTO evaluations VALUES(28901,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(28901,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(28901,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(28901,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(28901,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(28901,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(28901,36,'cp 53','e7e6 d1b3 d8c8 a1c1 b8d7 c4d5 f6d5 f4g3 a7a5 f1e2 f8e7 e1g1 d5c3 b2c3 a5a4 b3b2 e8g8 f1d1 f8d8 c3c4 c6c5 h2h3 a4a3 b2b3 e7f6 e2d3 f5d3 d1d3 c8c6 d4d5 e6d5 d3d5 c6b6 c1b1 b6b3 b1b3 b7b6 d5d6');
INSERT INTO evaluations VALUES(29868,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(29868,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(29868,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(29868,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(29868,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(29868,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(29868,36,'cp 99','c1e3 b8c6 e1g1 f8e7 e2b5 e8g8 b1d2 g4h5 b5c6 b7c6 d1b3 a8b8 b3c2 f7f5 e4f5 e6f5 f3e5 b6d5 e5c6 d5b4 c2b3 h5f7 c6d8 f7b3 d2b3 f8d8 f1c1 e7d6 b3a5 d8e8 a5c4 b4d5 e3d2 e8e4 d2c3 b8e8 a2a3 d5c3 b2c3 e4e2');
INSERT INTO evaluations VALUES(30642,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(30642,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(30642,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30642,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30642,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30642,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30642,36,'cp 0','g8f6 g1f3 d7d5 c4d5 e6d5 f1b5 c7c6 b5d3 f8d6 e1g1 b8d7 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 e4c2 e8g8 d1d3 d7f6 f3e5 d6e5 d4e5 d8d3 c2d3 a8d8 f1d1 f6d5 c1d2 c6c5 h2h3 d5c7 d2e3 c7d5');
INSERT INTO evaluations VALUES(33048,30,'cp 301',NULL);
INSERT INTO evaluations VALUES(33048,31,'cp 302',NULL);
INSERT INTO evaluations VALUES(33048,32,'cp 293',NULL);
INSERT INTO evaluations VALUES(33048,33,'cp 302',NULL);
INSERT INTO evaluations VALUES(33048,34,'cp 295',NULL);
INSERT INTO evaluations VALUES(33048,35,'cp 294',NULL);
INSERT INTO evaluations VALUES(33048,36,'cp 309','c3d5 c5d6 g1e2 c8e6 c1f4 b8c6 f4d6 e8c8 d5b6 a7b6 e1c1 g8e7 e2c3 e7f5 d6f4 d8d1 c3d1 h8d8 d1c3 c6b4 a2a3 b4d5 c3d5 e6d5 f2f3 d5c6 f1d3 g7g6 h1d1 d8e8 c1d2 f5h4 d3f1 h4f5 d1e1 e8d8 d2c1 d8d4 f4g5 d4d5 f1c4 d5c5 c4d3 c5d5 h2h4 b6b5 b2b3');
INSERT INTO evaluations VALUES(35270,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(35270,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(35270,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(35270,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(35270,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(35270,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35270,36,'cp 0','g8f6 h5d1 c7c6 f1e2 d7d5 e4d5 f6d5 d2d4 d5c3 b2c3 b8d7 g1f3 e5e4 f3d2 d7b6 e1g1 b6a4 c3c4 g7g6 d2e4 e7e4 e2f3 e4e7 c2c3 a4c3 d1d2 c3a4 d2d1');
INSERT INTO evaluations VALUES(34036,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(34036,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(34036,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(34036,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(34036,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34036,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(34036,36,'cp 0','h5d1 g7g6 f1c4 c7c6 g1f3 b7b5 c4b3 d7d6 d2d4 f8g7 e1g1 e8g8 c1g5 b8d7 d1d2 e5d4 f3d4 c8b7 f1e1 a7a5 a2a3 b5b4 c3a4 b4a3 a1a3 f8e8 a4c3 e7f8 b3a4 d7e5 a4b3');
INSERT INTO evaluations VALUES(35831,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(35831,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(35831,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(35831,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(35831,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(35831,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(35831,36,'cp 22','c4d5 f6d5 e1g1 b8c6 a2a3 b7b6 d1c2 d5c3 c2c3 d8d6 d4c5 d6c5 c3c5 e7c5 b2b4 c5e7 c1b2 c8b7 f1d1 f8d8 a1c1 a8c8 d3f1 e7f6 d1d8 c8d8 b2f6 g7f6 f1e2 d8c8 f3d2 g8g7 f2f3 f6f5 e2b5 c6e7 c1c8 b7c8 d2c4');
INSERT INTO evaluations VALUES(40931,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(40931,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40931,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(40931,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40931,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40931,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(40931,36,'cp 11','b1d2 c7c5 e2e3 b8d7 f1d3 d8c7 g5f4 e7d6 f4d6 c7d6 d1e2 e8g8 e3e4 d5e4 d2e4 f6e4 e2e4 d7f6 e4e5 f8d8 e5d6 d8d6 e1c1 c5d4 f3d4 g8f8 d3e2 c8d7 h1e1 a8c8 f2f4 b7b6 e2f3 h7h6 h2h4 h6h5 d1d2 c8c5 a2a3 f8e7 g2g3 g7g6 e1d1 d7a4');
INSERT INTO evaluations VALUES(41225,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(41225,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41225,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(41225,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(41225,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41225,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41225,36,'cp 0','e8g8 d2d3 a7a6 a2a4 c7c5 e1g1 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(36494,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36494,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36494,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36494,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36494,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36494,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36494,36,'cp 0','d2d3 a7a6 a2a4 c7c5 e1g1 b7b6 c1g5 h7h6 g5h4 b8c6 f3h2 e7e6 f2f4 f6e4 h4d8 g7d4 f1f2 e4f2 d1f3 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(43780,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(43780,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(43780,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43780,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(43780,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(43780,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(43780,36,'cp 53','e4e5 f6d5 c3b5 d5b6 f1e2 d8d7 b5c3 e7e6 g1f3 b8c6 e1g1 c8b7 c1e3 f8e7 a2a3 a7a5 d1c2 a5a4 h2h4 c6a5 f3g5 b6d5 g5e4 e8g8 a1d1 h7h6 e2h5 d5c3 e4c3 a8b8 c2e2 b7a8 f1e1 b8b3 e3c1');
INSERT INTO evaluations VALUES(44814,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(44814,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(44814,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(44814,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(44814,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(44814,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(44814,36,'cp 33','f1b5 c8d7 b5e2 f8g7 g1f3 d7g4 a1b1 c5d4 c3d4 g4f3 e2f3 g7d4 e1g1 e8g8 e4e5 b8c6 c1h6 f8e8 e5e6 f7e6 d1b3 d8d6 b1d1 e8d8 f1e1 d6c5 b3e6 g8h8 d1c1 c5d6 e6d6 d8d6 g2g3 a8d8 f3c6 d6c6 c1c6');
INSERT INTO evaluations VALUES(45254,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(45254,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(45254,32,'cp -77',NULL);
INSERT INTO evaluations VALUES(45254,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(45254,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(45254,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(45254,36,'cp -79','f6e4 f1d3 e4f6 e1g1 g7g6 f1e1 f8g7 b1c3 a7a6 h2h3 e8g8 a2a4 b7b6 d4d5 c8b7 d3c4 b8d7 c1g5 a8c8 d1e2 h7h6 g5f6 g7f6 c4a6 b7a6 e2a6 c8c3 b2c3 f6c3 a1c1 c3e1 c1e1 d7f6 a6b7 f8e8 f3d4 d8d7 b7d7 f6d7');
INSERT INTO evaluations VALUES(45284,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(45284,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(45284,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(45284,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(45284,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(45284,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(45284,36,'cp -75','f1d3 e4f6 e1g1 g7g6 f1e1 f8g7 h2h3 a7a6 b1c3 e8g8 a2a4 e7e6 c1g5 b8c6 d4d5 c6b4 d5e6 c8e6 d3f1 d6d5 f3d4 h7h6 g5e3 f8e8 d1d2 g8h7 a4a5 b4c6 d4c6 b7c6 b2b4 d8b8 e3b6 f6d7 c3a4 d7b6 a4b6 g7a1');
INSERT INTO evaluations VALUES(46003,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(46003,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(46003,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(46003,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46003,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46003,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(46003,36,'cp -11','a7a6 d2d4 b7b5 c4a2 c5d4 f3d4 f8e7 d1d3 d8c7 e1g1 b8c6 d4c6 c7c6 f2f3 e8g8 c1e3 c8b7 a1d1 c6c7 e3d4 b7c6 g1h1 f8c8 d3e2 c7b7 b2b4 a6a5 a2b3 h7h6 e2f2 a5a4 b3a2 c6e8 f2e2 a8b8 e4e5 d6e5');
INSERT INTO evaluations VALUES(48303,30,'cp -74',NULL);
INSERT INTO evaluations VALUES(48303,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(48303,32,'cp -65',NULL);
INSERT INTO evaluations VALUES(48303,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(48303,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(48303,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(48303,36,'cp -57','e2e3 c7c5 a2a3 a7a6 g1f3 b7b5 f1d3 c5c4 d3e2 f8d6 f3e5 e8g8 e1g1 c8b7 g2g4 b8c6 g4g5 f6e8 e2g4 d6c7 c1d2 c6e7 d1e2 e8d6 e2g2 a6a5 c3e2 e7f5 g4f5 d6f5 e2g3 c7e5 g3f5 e5c7 f5g3 f7f6 h2h4 g8h8 g2h3 d8d7 c2c3 c7d6');
INSERT INTO evaluations VALUES(48301,30,'cp -75',NULL);
INSERT INTO evaluations VALUES(48301,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(48301,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(48301,33,'cp -66',NULL);
INSERT INTO evaluations VALUES(48301,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(48301,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(48301,36,'cp -66','c7c5 f1d3 a7a6 g1f3 b7b5 a2a3 c5c4 d3e2 c8b7 f3e5 f8d6 e1g1 e8g8 g2g4 f6e8 g4g5 f7f6 g5f6 e8f6 c1d2 b8c6 e2f3 c6e7 c3e2 a6a5 e2g3 b5b4 a3b4 a5b4 e5c4 a8a1 d1a1 d5c4 f3b7 e7d5 a1a4 c4c3 b2c3 b4c3 b7d5');
INSERT INTO evaluations VALUES(48386,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(48386,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(48386,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(48386,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(48386,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(48386,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(48386,36,'cp -10','e4e5 f6d7 f3f4 c7c5 g1f3 b8c6 f1b5 d8b6 b5c6 b7c6 d2d3 f8e7 e1g1 e8g8 b2b3 c5c4 d3d4 a7a5 f1f2 c8a6 c3a4 b6c7 c1e3 f7f6 e5f6 d7f6 f3e5 f6e4 f2f3 e7d6 b3c4 d5c4 h2h3 a6b5 a4c5 e4c5 d4c5 d6e5 f4e5 f8f3');
INSERT INTO evaluations VALUES(49961,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(49961,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(49961,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(49961,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(49961,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(49961,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(49961,36,'cp 32','b7b6 e1g1 f8e7 c2c4 c8b7 b1c3 e8g8 b2b3 f6e4 c1b2 c7c5 d1e2 e4c3 b2c3 e7f6 a1d1 b7f3 e2f3 b8c6 d3e2 d8e7 c3b2 c6b4 a2a3 b4c6 d1d3 a7a5 d4d5 f6b2 d5c6 d7c6 f3c6 f8c8 c6d6 e7d6 d3d6 b2a3 d6b6');
INSERT INTO evaluations VALUES(50669,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50669,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(50669,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(50669,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(50669,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(50669,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(50669,36,'cp 26','d2d4 g8f6 d4c5 f6e4 c5d6 d8b6 d1e2 e4d6 b1c3 c8e6 b2b3 b8c6 c3a4 b6a5 c1d2 a5c7 c2c4 a8d8 a1d1 e6f5 g2g4 f5c2 d1c1 h7h5 f1g2 h5g4 h3g4 h8h1 g2h1 c2g6 f3h4 g6h7 a4c3 c7c8 c3d5 c6d4 e2d1 e7e6 d2a5 d4c2 c1c2 h7c2 d1e2 c2d3 e2d3 e6d5 a5d8');
INSERT INTO evaluations VALUES(50757,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(50757,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(50757,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(50757,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(50757,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(50757,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(50757,36,'cp 37','g8f6 c4d3 f8e7 c2c4 b8c6 d4c6 b7c6 b1c3 e8g8 c1f4 a6a5 d1c2 e6e5 f4e3 c8e6 b2b3 f6d7 a1d1 d8b8 h2h3 d7c5 e3c5 d6c5 d3e2 b8b4 d1d3 f8d8 f1d1 d8d4 e2g4 e6g4 h3g4 a8d8 g2g3 e7g5 g1g2 h7h6 c3a4 b4a3 a4b6');
INSERT INTO evaluations VALUES(43523,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43523,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(43523,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(43523,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(43523,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(43523,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(43523,36,'cp 48','c4d3 f8e7 c2c4 b8c6 d4c6 b7c6 b1c3 e8g8 c1f4 a6a5 b2b3 e6e5 f4e3 f6g4 e3d2 g4f6 d3e2 f6d7 e2g4 d7c5 d2e3 e7g5 e3c5 d6c5 g4c8 a8c8 g2g3 d8e7 d1c2 g7g6 f1d1 h7h5 d1d3 c8d8 a1d1 h5h4 d3d8 f8d8 d1d8 e7d8 g1g2');
INSERT INTO evaluations VALUES(51402,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(51402,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(51402,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(51402,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(51402,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(51402,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(51402,36,'cp 28','g2g3 c7c5 f1g2 c8b7 b1c3 g8f6 e1g1 f8e7 d2d4 c5d4 f3d4 b7g2 g1g2 d8c8 d1d3 b8c6 b2b3 e8g8 d4c6 c8c6 d3f3 f8d8 f1d1 a7a6 f3c6 d7c6 c1f4 b6b5 g2f3 h7h6 d1d8 a8d8 a1d1 d8d1 c3d1 b5c4 b3c4 f6d7 f4c7 e7b4 f3e4 f7f6 f2f4 b4c5 e2e3');
INSERT INTO evaluations VALUES(53808,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53808,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53808,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53808,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(53808,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53808,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53808,36,'cp 0','g8f6 d2d4 e5d4 f3d4 f6e4 d1e2 d8e7 e1g1 e4c5 c1e3 d6d5 c4b3 c5b3 a2b3 g7g6 e2d2 f8g7 e3g5 e7d6 g5f4 d6e7');
INSERT INTO evaluations VALUES(53958,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(53958,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53958,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(53958,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(53958,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53958,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(53958,36,'cp 10','c8d7 b5d7 b8d7 e1g1 g8f6 d2d3 f8e7 d1e2 a7a6 a2a4 d7b8 e4e5 d6e5 f3e5 d8c7 c1f4 e7d6 f4g3 e8g8 a4a5 b8c6 e5c6 c7c6 g3d6 c6d6 a1a4 f8d8 f1e1 a8c8 e2e5 d6e5 e1e5 f6d7 e5e3 d7f6 a4c4 d8d4 g1f1 g8f8 g2g3 h7h6');
INSERT INTO evaluations VALUES(30875,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(30875,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(30875,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30875,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(30875,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(30875,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30875,36,'cp 3','b5e2 b8c6 d2d4 c5d4 f3d4 g8f6 c1e3 f8e7 f2f4 e8g8 e1g1 c6d4 e3d4 d7c6 d1d3 b7b5 a2a3 d8b8 d3g3 a7a5 b2b4 b8b7 e2f3 a5b4 a3b4 a8a1 f1a1 f8a8 a1d1 a8c8 h2h3 h7h6 g1h1 b7c7 e4e5 d6e5 d4e5 c7b7');
INSERT INTO evaluations VALUES(54530,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(54530,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(54530,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(54530,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(54530,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(54530,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(54530,36,'cp 25','b8c6 f1g2 g8f6 e1g1 c5c4 b1c3 f8e7 b2b3 c4b3 a2b3 e7b4 d1d3 e8g8 c1g5 b4e7 f3e5 c6b4 d3d1 c8f5 g5f6 e7f6 e2e4 d5e4 c3e4 f5e4 g2e4 f6e5 d4e5 b4c6 a1c1 d8d1 f1d1 a8d8 d1d8 f8d8 e4c6 b7c6 c1c6 d8b8');
INSERT INTO evaluations VALUES(55023,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(55023,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(55023,32,'cp -115',NULL);
INSERT INTO evaluations VALUES(55023,33,'cp -121',NULL);
INSERT INTO evaluations VALUES(55023,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(55023,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(55023,36,'cp -103','e7e5 e2e3 b8c6 d2d4 f8d6 d4e5 c6e5 g1f3 c7c6 f3e5 d6e5 b1d2 d8e7 d2f3 e5b2 c1b2 e7b4 f3d2 b4b2 c2c4 d5c4 a1b1 b2c3 f1c4 g8h6 b1b3 c3a5 d1c2 e8g8 c4d3 g8h8 d3h7 b7b6 h7e4 c8e6 c2c3 e6b3 c3a5 b6a5 a2b3 a8b8 e1e2 c6c5 f2f3 f8d8 g2g4');
INSERT INTO evaluations VALUES(837,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(837,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(837,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(837,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(837,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(837,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(837,36,'cp -22','f7f6 e5f6 g7f6 f1d3 f5e3 f2e3 f8h6 d1e2 a7a5 b4a5 a8a5 b1c3 b6b3 e1g1 a5a3 a1a3 b3a3 c3b5 a3e7 e3e4 c6b4 b5c3 d5e4 d3b5 b4c6 b5c6 b7c6 c3e4 e8g8 g1h1 g8h8 f1a1 e6e5 e4g3 h6g7 a1e1 c6c5 d4e5 f6e5 g3h5 c8b7 h5g7 e7g7 e2e5 c5c4 e5g7 h8g7 e1e3 b7f3 g2f3 f8g8 h1g1 g7h8 g1f2');
INSERT INTO evaluations VALUES(3602,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(3602,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(3602,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(3602,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(3602,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(3602,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(3602,36,'cp -36','d2d3 c6a5 c4b3 a5b3 a2b3 a7a5 h2h3 f8e8 c1e3 c5e3 f2e3 c8e6 d1e1 c7c6 e1g3 b7b5 f3h4 f6h5 g3f2 g8h7 h4f3 h7h8 f3h4 e8g8 c3e2 h8h7 h4f5 g7g6 f5h4 d8e7 h4f3 f7f5 e4f5 g6f5 f3h4 g8f8 g1h2 b5b4');
INSERT INTO evaluations VALUES(4154,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4154,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4154,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4154,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4154,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4154,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4154,36,'cp 0','c5b6 d2d3 a7a6 c1e3 d7d6 c3d5 f6d5 c4d5 d8f6 f3d2 b6d4 c2c3 d4c3 a1c1 c3b2 c1b1 b2c3');
INSERT INTO evaluations VALUES(4129,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4129,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(4129,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4129,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4129,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4129,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4129,36,'cp 0','d2d3 a7a6 c1e3 d7d6 c3d5 f6d5 c4d5 d8f6 f3d2 b6d4 c2c3 d4c3 a1c1 c3b2 c1b1 b2c3');
INSERT INTO evaluations VALUES(4161,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(4161,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(4161,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(4161,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(4161,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(4161,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(4161,36,'cp 16','c2c3 d7d6 e1g1 a7a5 a2a4 c5a7 b1a3 g8h8 a3c2 c8e6 b2b3 d8e7 f1e1 a8e8 a1b1 a7b6 h4g3 e6c4 b3c4 e7d7 h2h3 f6h5 g3h2 h5f6 b1b5 e8e7 e1f1 d7e6 f3h4 g7g6 h4f3 c6b8 c2e3 b6e3');
INSERT INTO evaluations VALUES(5860,30,'cp 237',NULL);
INSERT INTO evaluations VALUES(5860,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(5860,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(5860,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(5860,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(5860,35,'cp 250',NULL);
INSERT INTO evaluations VALUES(5860,36,'cp 250','b2c3 g8e7 e1g1 e8g8 f1e1 b7b6 c4d3 h7h6 h2h4 c6a5 h4h5 c7c5 c1f4 c8b7 d1d2 f7f6 d2e2 d8d7 a1d1 a8d8 d4c5 b6c5 d3b5 d7c8 f4d6 f8f7 f3h4 a7a6 d6e7 d8d1 e1d1 f7e7 h4g6 e7f7 b5a4 b7c6 a4c2 c6b5 e2g4 a5c4 e4e5 c4e5 g6e5 f6e5 d1d6 f7f6 g4e4 c8c7 e4e5 c5c4');
INSERT INTO evaluations VALUES(6229,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(6229,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(6229,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(6229,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(6229,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(6229,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6229,36,'cp 3','d7d6 b1c3 g8f6 d2d3 a7a6 c3d5 f6d5 c4d5 e8g8 c1e3 c5e3 f2e3 c6e7 d5a2 e7g6 d1e1 d8e7 e1g3 c8e6 a2e6 e7e6 f1f2 f7f5 e4f5 f8f5 a1f1 a8f8 g3g4 f5f6 g4e6 f6e6 f3d2 f8f2 g1f2 e6f6 f2e2 g8f7 c2c4 f6f1 e2f1 f7e6 h2h3 b7b5 c4b5 a6b5 b2b4');
INSERT INTO evaluations VALUES(6769,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(6769,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(6769,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(6769,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(6769,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(6769,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(6769,36,'cp 109','d6c5 d1e2 e5f4 c2c3 c6a5 c4b5 a7a6 b5a4 c5e7 d2d4 b7b5 a4c2 e7h4 e1d1 h4e7 c1f4 g8f6 b1d2 e8g8 d4d5 d7d6 a2a4 c8g4 b2b4 a5c4 d2c4 b5c4 d1c1 a6a5 b4b5 f6d7 c1b2 d7b6 h1g1 e7f6 e2d2 f8e8 f3d4 f6d4 c3d4 g4h5');
INSERT INTO evaluations VALUES(8073,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(8073,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(8073,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(8073,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(8073,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(8073,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(8073,36,'cp 48','d4c5 d8a5 a1c1 d5c4 f1c4 a5c5 d1e2 e8g8 e1g1 c8d7 c4d3 f8d8 c3e4 c5b6 e4f6 e7f6 g5f6 g7f6 b2b3 a8c8 f1d1 c6e7 c1c8 d7c8 f3d2 f6f5 d2f1 e7g6 f1g3 g6e5 g3h5 g8f8 h5f4 e5d3 d1d3 d8d3 e2d3 f8e7 d3c4 h7h6');
INSERT INTO evaluations VALUES(8168,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(8168,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(8168,32,'cp -38',NULL);
INSERT INTO evaluations VALUES(8168,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(8168,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(8168,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(8168,36,'cp -28','a7a6 e1g1 b7b5 c4a2 c8b7 f1e1 d8c7 c1f4 e8g8 d1e2 f6d7 c3d1 d7e5 c2c3 e5f3 e2f3 c6e5 f3e2 e5g6 f4g3 g8h8 d1e3 e7g5 a1d1 f7f5 e4f5 e6f5 e3d5 c7d7 e2h5 g5h6 d3d4 f5f4');
INSERT INTO evaluations VALUES(7075,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(7075,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(7075,32,'cp -35',NULL);
INSERT INTO evaluations VALUES(7075,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(7075,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(7075,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(7075,36,'cp -27','c4a2 b7b5 e1g1 c8b7 f1e1 e8g8 c1f4 d8c7 d1e2 h7h6 f4g3 c6e5 f3d2 g7g6 g3h2 f6h5 e2e3 g8h7 f2f4 e5c6 d2f3 e7f6 e4e5 f6g7 c3e4 d6e5 f4e5 c6e5 f3e5 g7e5 h2e5 c7e5 e4c5');
INSERT INTO evaluations VALUES(9758,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(9758,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9758,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9758,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9758,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9758,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9758,36,'cp 0','a7a6 a2a4 e8g8 h2h3 d7d6 d2e3 c5b4 e1g1 b4c3 b2c3 d6d5 e4d5 f6d5 e3d2 f8e8 f1e1 d5f6 d2e3 c8e6 c4e6 e8e6 a4a5 f6d5 e3d2 d5f6');
INSERT INTO evaluations VALUES(10321,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(10321,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(10321,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(10321,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(10321,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(10321,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(10321,36,'cp -37','f6d5 e4d5 c6d4 f3d2 d7d6 c2c3 d4f5 g2g3 a7a5 f1g2 b7b5 d2e4 c5b6 e1g1 e8g8 g1h1 a8b8 d1h5 f7f6 g3g4 f5e7 f2f4 d8e8 h5e8 f8e8 f4e5 f6e5 h2h3 e7d5 e4d6 c7d6 g2d5 c8e6 d5e6 e8e6 c1d2 d6d5 a1e1 b8e8 h1g2 h7h6 h3h4 e5e4 d3e4 e6e4 g4g5 e4e2 g2h3 h6g5 h4g5 e2e1 d2e1');
INSERT INTO evaluations VALUES(10556,30,'cp 269',NULL);
INSERT INTO evaluations VALUES(10556,31,'cp 294',NULL);
INSERT INTO evaluations VALUES(10556,32,'cp 282',NULL);
INSERT INTO evaluations VALUES(10556,33,'cp 298',NULL);
INSERT INTO evaluations VALUES(10556,34,'cp 290',NULL);
INSERT INTO evaluations VALUES(10556,35,'cp 287',NULL);
INSERT INTO evaluations VALUES(10556,36,'cp 282','d2d4 e5d4 c3d4 c5b4 b1c3 e8g8 e1g1 c6e7 c1h6 g7h6 f3e5 b4c3 b2c3 d7d6 e5d3 g8h8 d1d2 e7g8 a1e1 a7a5 e1e3 c8e6 d4d5 e6d7 f1e1 a5a4 d3f4 f8e8 c4e2 d8g5 e3g3 g5e5 f4h5 e5e4 h2h3 e8e5 c3c4 a8e8 h5g7 e8f8 f2f4 e5e7 d2c3 f7f6 g7h5 e4e2 e1e2 e7e2');
INSERT INTO evaluations VALUES(10560,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(10560,31,'cp 283',NULL);
INSERT INTO evaluations VALUES(10560,32,'cp 278',NULL);
INSERT INTO evaluations VALUES(10560,33,'cp 289',NULL);
INSERT INTO evaluations VALUES(10560,34,'cp 272',NULL);
INSERT INTO evaluations VALUES(10560,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(10560,36,'cp 262','e5d4 c3d4 c5b4 b1c3 e8g8 e1g1 c6e7 c1h6 g7h6 f3e5 b4c3 b2c3 d7d6 e5d3 g8h8 d1d2 d6d5 c4d5 e7d5 e4d5 d8d5 d2h6 d5d6 h6h5 d6g6 h5e5 f7f6 e5e3 f8g8 d3f4 g6f5 f1e1 c8d7 c3c4 a8e8 e3d2 e8e1 a1e1 g8e8 e1e8 d7e8 h2h3 e8f7 d4d5 h8g7 h3h4 f5e5 g2g3 e5e4 c4c5 e4e5 g1h2 e5e7');
INSERT INTO evaluations VALUES(10565,30,'cp 285',NULL);
INSERT INTO evaluations VALUES(10565,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(10565,32,'cp 272',NULL);
INSERT INTO evaluations VALUES(10565,33,'cp 287',NULL);
INSERT INTO evaluations VALUES(10565,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(10565,35,'cp 290',NULL);
INSERT INTO evaluations VALUES(10565,36,'cp 294','c3d4 c5b4 b1c3 e8g8 e1g1 c6e7 c1h6 g7h6 f3e5 b4c3 b2c3 d7d6 e5d3 g8h8 d1d2 e7g8 a1e1 a7a5 e1e3 h6h5 f2f4 c8e6 d4d5 e6d7 d3f2 f7f6 c4e2 h5h4 f4f5 c7c6 c3c4 f8e8 e3h3 c6d5 c4d5 d8e7 d2f4 b7b5 g1h1');
INSERT INTO evaluations VALUES(10564,30,'cp 267',NULL);
INSERT INTO evaluations VALUES(10564,31,'cp 266',NULL);
INSERT INTO evaluations VALUES(10564,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(10564,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(10564,34,'cp 287',NULL);
INSERT INTO evaluations VALUES(10564,35,'cp 278',NULL);
INSERT INTO evaluations VALUES(10564,36,'cp 274','c5b4 b1c3 e8g8 e1g1 c6e7 c1h6 g7h6 f3e5 b4c3 b2c3 d7d6 e5d3 g8h8 d1d2 e7g8 a1e1 a7a5 e1e3 h6h5 h2h3 c8d7 c4d5 c7c6 d5b3 a5a4 b3d1 b7b5 e4e5 h5h4 f1e1 d7f5 d1c2 f5g6 d3b4 g6c2 d2c2 c6c5 b4d3 d6e5 d4c5 f7f6 c3c4 b5c4 c2c4 g8e7 c4h4 e7f5');
INSERT INTO evaluations VALUES(13640,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(13640,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(13640,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(13640,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(13640,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(13640,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(13640,36,'cp 49','e3d4 d5c4 f1c4 f8d6 e1g1 e8g8 c1g5 d6e7 f1e1 f6d5 c4d5 e6d5 g5e7 c6e7 d1b3 a8b8 h2h3 c8e6 c3e2 d8d6 e2g3 e7g6 e1e3 f8c8 a1e1 c8c7 g3e2 b8f8 f3e5 a7a6 e5d3 b7b6 e2c3 h7h6 c3e2 e6f5 d3e5 f8c8');
INSERT INTO evaluations VALUES(14822,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(14822,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(14822,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(14822,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(14822,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(14822,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(14822,36,'cp 24','f6e4 h2h3 f8d6 f1d3 e8g8 f4d6 c7d6 e1g1 e4f6 c2c4 d5c4 d3c4 d8b6 b1c3 c6e7 b2b4 c8d7 d1b3 d7c6 f3d2 e7d5 c3d5 c6d5 f1c1 f8c8 a3a4 a7a6 b3b2 d5c6 c4d3 f6d5 b4b5 a6b5 a4b5 c6e8 c1c8 a8c8 d3e2');
INSERT INTO evaluations VALUES(15251,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(15251,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(15251,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(15251,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(15251,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(15251,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(15251,36,'cp 19','f8d6 e1e4 d5e4 e5c6 b7c6 b5c6 c8d7 c6a8 d8a8 b2b3 d6e5 b1c3 e8g8 c1b2 f8e8 d1f1 a8c6 c3d1 c6c2 b2e5 e8e5 d1e3 c2c6 d2d4 e4d3 f1d3 d7e8 a1d1 f7f6 h2h3 h7h5 d3b1 e8f7 d1d4 c6c5 b1b2 a7a5 b2d2 e5e8 e3d5 f7e6 d5f4 e6f7 d4d7 c5e5');
INSERT INTO evaluations VALUES(15523,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(15523,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(15523,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(15523,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(15523,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(15523,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(15523,36,'cp 34','c8g4 d4e5 d6e5 b1d2 f8d6 h2h3 g4h5 c2c3 d8d7 c4e2 e8g8 a2a4 a7a5 f3e5 c6e5 e2h5 f6h5 d1h5 a8e8 h5e2 b7b6 d2b3 e5g6 b3d4 d6c5 f1d1 d7e7 d4f5 e7e4 e2e4 e8e4 c1e3 c5e3 f5e3 g6f4 g1f1 g7g5');
INSERT INTO evaluations VALUES(19361,30,'cp -94',NULL);
INSERT INTO evaluations VALUES(19361,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(19361,32,'cp -80',NULL);
INSERT INTO evaluations VALUES(19361,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(19361,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(19361,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(19361,36,'cp -77','e7e5 c5c4 g8f6 b1c3 f8b4 c1g5 e8g8 g1e2 d7d5 e4d5 d8d5 c4d5 f6d5 a2a3 d5c3 e2c3 b4c3 b2c3 c8e6 f1d3 f7f6 g5c1 c6a5 e1g1 e6c4 f1d1 b7b6 d3f5 f8d8 c1e3 g7g6 f5g4 g8f7 d1d8 a8d8 a1d1 d8d1 g4d1 f6f5 f2f4 e5e4');
INSERT INTO evaluations VALUES(20077,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(20077,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20077,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(20077,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(20077,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20077,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20077,36,'cp 15','g8f6 b1c3 e7e5 d4f3 f8e7 c1g5 e8g8 g5f6 e7f6 c3d5 c8e6 e1g1 a8c8 c2c4 g7g6 d1d2 g8g7 f1d1 e6d5 c4d5 c6b8 g2g3 b8d7 b2b4 f6e7 g1g2 d7f6 e2d3 d8d7 a1c1 h7h6 d2e2 c8c1 d1c1 e7d8 d3b5 d7g4 c1c4 d8b6');
INSERT INTO evaluations VALUES(13378,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(13378,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(13378,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(13378,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(13378,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(13378,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(13378,36,'cp 17','b1c3 e7e5 d4f3 f8e7 c1g5 c8e6 g5f6 e7f6 e1g1 a8c8 c3d5 e8g8 c2c4 g7g6 d1d2 f6g7 a1d1 d8d7 f1e1 g8h8 d5c3 c8d8 c3b5 d7e7 b5d6 f7f5 c4c5 f5e4 f3g5 g7h6 g5f7 e6f7 d2h6 f7a2 d6e4 c6d4 e2f1 b7b6 c5b6 a7b6 f2f3 e7c7 h2h4 a2b3 d1d2');
INSERT INTO evaluations VALUES(20813,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(20813,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(20813,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(20813,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(20813,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(20813,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(20813,36,'cp 128','b1c3 f7f5 f1d3 g8f6 c1g5 f5f4 g5f6 g7f6 d1h5 e7g6 d3c2 c8d7 g1f3 d8e7 c2a4 d7a4 c3a4 e7f7 a4c3 a7a6 e1c1 f8g7 c1b1 a8b8 g2g3 e8g8 c3e2 b7b5 g3f4 e5f4 c4b5 f7e8 e2c3 a6b5 h5f5 b5b4 c3e2');
INSERT INTO evaluations VALUES(20811,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(20811,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(20811,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(20811,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(20811,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(20811,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(20811,36,'cp 132','f7f5 f1d3 g7g6 h2h4 f8h6 c1g5 h6g7 h4h5 h7h6 g5e7 g8e7 h5g6 e8g8 g1e2 d8e8 d1d2 e8g6 e1c1 g6g5 d2g5 h6g5 e4f5 g5g4 f5f6 f8f6 c3e4 f6h6 c1c2 h6h1 d1h1 c8f5 e4g5 a8f8 d3f5 e7f5 g5e4 f5d4 c2d3 d4e2 d3e2 f8f4 e2d3 c7c6 e4d6 f4d4 d3e2 c6d5 c4d5 d4d5');
INSERT INTO evaluations VALUES(21053,30,'cp -44',NULL);
INSERT INTO evaluations VALUES(21053,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(21053,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(21053,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(21053,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(21053,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(21053,36,'cp -48','d2d4 e7e6 c2c3 d5e4 b1d2 e6e5 d4e5 c6e5 d1a4 c8d7 a4e4 d8e7 d2c4 e5g6 e4e7 f8e7 g1e2 e8c8 c1e3 g8h6 e2d4 g6h4 e3h6 g7h6 e1f2 e7c5 c4e3 h4f5 e3f5 d7f5 a1d1 f5e6 a2a3 a7a5 f1d3 c5d4 c3d4 d8d4 d3f5 h8d8 d1d4');
INSERT INTO evaluations VALUES(24930,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(24930,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(24930,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(24930,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(24930,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(24930,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(24930,36,'cp -16','c6b4 e1g1 e7e6 c1d2 b4d3 c2d3 f8e7 h2h3 g4f3 d1f3 e8g8 d2c3 c7c5 b1d2 b7b6 a1c1 a8c8 g1h2 c5d4 c3d4 d8d7 f3e2 c8c7 d2f3 f8c8 c1c7 d7c7 d4c3 f6e8 f1c1 c7b7 e2d2 a7a5 a2a3 e8d6 c3e5 f7f6 e5d4 c8c1');
INSERT INTO evaluations VALUES(26096,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26096,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(26096,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26096,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(26096,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(26096,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(26096,36,'cp 21','f1d3 f5d3 d1d3 g8f6 e1g1 f8e7 e3e4 d5e4 c3e4 e8g8 f1d1 c6c5 e4f6 d7f6 b2b3 d8c7 d3e2 f8d8 c1b2 a7a5 h2h3 a5a4 d4c5 c7c5 b2d4 c5c6 d1d3 e7c5 a1d1 a4b3 a2b3 c5d4 f3d4 c6c7 d4b5 c7a5 g2g3 h7h6 b5d6');
INSERT INTO evaluations VALUES(27834,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(27834,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(27834,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(27834,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(27834,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(27834,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(27834,36,'cp 65','f1e2 g4f5 e1g1 h7h6 g5h4 a7a6 h2h3 g6g5 h4g3 f6e4 c3e4 f5e4 a1c1 b8d7 c4c5 f7f5 c5d6 e7d6 f3d2 e4d5 e2f3 d5f3 d1b3 f8f7 d2f3 a8b8 c1c3 d7f8 f1c1 c7c6 a2a3 b8a8 g3h2 d8e7 f3d2 d6d5');
INSERT INTO evaluations VALUES(27833,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(27833,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(27833,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(27833,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(27833,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(27833,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(27833,36,'cp 72','g4f5 e1g1 a7a5 e2d3 f5d3 d1d3 h7h6 g5h4 c7c6 a1d1 b8d7 b2b3 g6g5 h4g3 f6h5 e3e4 h5g3 h2g3 e7e6 f1e1 f8e8 f3h2 d8e7 f2f4 a8d8 d3f3 e8f8 e1e2 f8e8 d1d3 e6e5 f4e5 d6e5 d4d5 d7c5 d3e3 g7f8 f3f5 c5d7 g3g4 d7f6');
INSERT INTO evaluations VALUES(28008,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(28008,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(28008,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(28008,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(28008,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(28008,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(28008,36,'cp 32','d7d6 h2h4 h7h5 f1c4 c7c6 d1e2 g8f6 c1g5 e8g8 e1g1 a7a6 a2a4 c6d5 c3d5 b8d7 a1d1 f6d5 e4d5 f8e8 b2b3 d8c7 g5e3 e8c8 f1e1 d7f6 e3d4 b6b5 e2e7 c7e7 e1e7 a8b8 c4f1 c8c2 d4f6 g7f6 e7d7 c2b2 d7d6 g8g7 a4b5 a6b5');
INSERT INTO evaluations VALUES(28053,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(28053,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(28053,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(28053,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(28053,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(28053,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(28053,36,'cp 110','e7e6 f1d3 b8c6 c1e3 g8f6 a2a3 c6e7 d1e2 e8g8 e3d2 c7c5 e4e5 f6e8 e1c1 a8c8 h2h4 h7h5 h1e1 b7f3 e2f3 c5d4 c3e4 d7d6 g2g4 d6e5 g4h5 f7f5 e4g5 e5e4 d3e4 f5e4 f3e4 d8d5 e4e6 d5e6 g5e6 d4d3 e6g7 e8g7 e1e7');
INSERT INTO evaluations VALUES(27929,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(27929,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(27929,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(27929,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(27929,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(27929,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(27929,36,'cp 110','f1d3 b8c6 c1e3 g8f6 a2a3 c6e7 d1e2 e8g8 e4e5 f6d5 c3d5 e7d5 h2h4 d5e3 e2e3 h7h5 e1c1 f7f5 f3g5 g7h6 e3e2 h6g5 h4g5 d8e7 d3c4 b7e4 d4d5 b6b5 c4b3 a8b8 d5d6 e7d8 e2f2 c7d6 d1d6 d8c7 h1d1');
INSERT INTO evaluations VALUES(28130,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(28130,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(28130,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(28130,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(28130,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(28130,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(28130,36,'cp 79','c2c3 b8d7 h2h3 g4f3 d1f3 e7e5 c1e3 e5d4 e3d4 g7d4 c3d4 d8f6 f3f2 d6d5 e4e5 f6b6 f2d2 a7a5 b1c3 g8e7 e1c1 d7f8 g2g4 e8c8 c1b1 h7h5 f1g2 h5g4 h3g4 h8h1 d1h1 f7f5 g4g5 c7c6 a2a3 f8e6 h1h7 c8d7 g2d5 b6d4 d2d4');
INSERT INTO evaluations VALUES(28148,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(28148,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(28148,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(28148,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(28148,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(28148,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(28148,36,'cp 25','c3a4 c5b6 d2d3 g8f6 a4b6 a7b6 h2h3 g4e6 c4e6 f7e6 c2c3 e8g8 e1g1 d8e8 c1e3 h7h6 a2a4 b8c6 b2b4 e8g6 g1h2 g6f7 d1c2 f7h5 c2b3 h5f7 h2g1 f6h5 g1h1 f7g6 f3h4 g6f6 g2g3 f6f7 c3c4 h5f6 b4b5 c6d4 e3d4 e5d4');
INSERT INTO evaluations VALUES(28377,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(28377,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(28377,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(28377,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(28377,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(28377,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(28377,36,'cp 24','h2h3 g4h5 a2a4 d8d7 a4a5 e8g8 c3d5 f6d5 c4d5 c7c6 d5b3 h5g6 f1e1 b8a6 c2c3 e7f6 d2d3 a6c5 b3c2 c5e6 c1e3 d6d5 c2b3 e6c7 f3h2 f6e7 h2g4 e7d6 e4d5 c7d5 d3d4 d5e3 e1e3 e5e4 g4e5 d6e5 d4e5 d7c7');
INSERT INTO evaluations VALUES(28376,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(28376,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(28376,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28376,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(28376,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(28376,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(28376,36,'cp 20','g4h5 g2g4 h5g6 d2d3 c7c6 a2a4 b8a6 f3h4 a6c7 d1f3 e8g8 h4f5 d6d5 c4b3 c7e6 e4d5 c6d5 c3d5 f6d5 f3d5 g6f5 g4f5 e6d4 c1e3 d4b3 d5b3 b7b6 a4a5 b6a5 b3b5 a8c8 a1a5 c8c2 a5a7 e7g5 e3c5 d8d5 b2b4 f8d8 b5b7 d5b7 a7b7 d8d3 b7a7');
INSERT INTO evaluations VALUES(28374,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(28374,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(28374,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(28374,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(28374,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(28374,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(28374,36,'cp 33','g2g4 h5g6 d2d3 c7c6 a2a4 b8a6 f3h4 a6c7 h4f5 e8g8 f5e7 d8e7 d1e1 d6d5 e4d5 f6d5 f2f4 f7f5 c1d2 a8e8 e1g3 g6f7 c4d5 f7d5 a1e1 e7c5 g3f2 c5f2 f1f2 f5g4 h3g4 e5f4 e1e8 f8e8 d2f4 c7e6 f4d6 e8d8 c3d5 c6d5 d6g3 d8c8 a4a5 h7h6 c2c3');
INSERT INTO evaluations VALUES(29085,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(29085,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(29085,32,'cp 121',NULL);
INSERT INTO evaluations VALUES(29085,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(29085,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(29085,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(29085,36,'cp 110','f1e2 g4e2 d1e2 d8d4 g1f3 d4b6 e4e5 f6d5 c3e4 e7e6 a2a3 b8d7 c2c4 f7f5 e5f6 d5f6 e4f6 g7f6 e2e6 f8e7 c1e3 b6a5 e3d2 a5h5 e1g1 h5f7 a1e1 f7e6 e1e6 e8f7 f3g5 f7f8 e6e2 a8e8 g5e4 h8g8 d2c3 g8g4 g1h1 b7b6 h2h3 g4h4 e2e3 f8f7 b2b4 h7h6 f1d1 e8d8 c4c5 b6c5 e4d6 e7d6 d1d6 h4f4 d6c6 c5b4 c3b4');
INSERT INTO evaluations VALUES(31373,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(31373,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(31373,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31373,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(31373,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(31373,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(31373,36,'cp 27','g6h7 f1d3 h7d3 d1d3 g8f6 c1d2 f8e7 e1c1 e8g8 g3e4 b8d7 h1g1 f6e4 d3e4 d7f6 e4e2 d8d5 c1b1 d5b5 e2b5 c6b5 g2g4 a8d8 c2c3 d8d5 d2e3 e7d6 a2a3 a7a5 f3d2 d6h2 g1g2 h2c7 d1g1 g8h7 f2f3 d5d7 d2e4 f6d5 e3d2 c7f4 g4g5 f4d2 g2d2 h6g5');
INSERT INTO evaluations VALUES(31455,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(31455,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(31455,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(31455,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(31455,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31455,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(31455,36,'cp 15','g8f6 e1g1 f8d6 f4d6 d8d6 c2c4 f5d3 d1d3 e8g8 f1d1 b8d7 b1c3 f8e8 b2b3 a8d8 e3e4 d5e4 c3e4 f6e4 d3e4 h7h6 e4e3 d6a3 d1d2 a3a5 a1d1 e6e5 d4e5 d7e5 e3e5 a5e5 f3e5 d8d2 d1d2 e8e5 d2d8 g8h7 g2g3 e5e2 d8d7 e2a2 d7b7');
INSERT INTO evaluations VALUES(31957,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(31957,31,'cp 120',NULL);
INSERT INTO evaluations VALUES(31957,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(31957,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(31957,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(31957,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(31957,36,'cp 124','d1f3 e7e6 f3b7 b8d7 d2d3 f8d6 g2g3 g8f6 b7g2 e8g8 g1f3 d7b6 c4b3 a7a5 a2a4 b6d5 f3h4 f5g4 c1d2 g4h5 h2h3 a8b8 g3g4 h5g6 h4g6 f7g6 e1g1 d6f4 c3d5 f4d2 d5e3 b8b3 c2b3 d2e3 f2e3 d8d3 a1e1 d3b3 g2c6 g8h8');
INSERT INTO evaluations VALUES(32267,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(32267,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(32267,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(32267,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(32267,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(32267,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(32267,36,'cp -23','f8b4 c2c3 b4a5 a2a4 a7a6 b5a3 c7c6 e2e3 g8f6 f1d3 f5d3 d1d3 b8d7 a3b1 e8g8 b1d2 c6c5 g1f3 a5c7 e1g1 c5c4 d3c2 c7f4 e3f4 d8c7 f3e5 b7b5 a4b5 a6b5 b2b4 c4b3 c2b3 a8a1 f1a1 h7h6 g2g3 f8c8 a1c1 c7b7 c1a1 f6e4 d2e4 d5e4 e5d7 b7d7');
INSERT INTO evaluations VALUES(34309,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(34309,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(34309,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(34309,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(34309,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(34309,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(34309,36,'cp 40','g8f6 d2d3 b8d7 c1f4 d7b6 f4d2 g7g6 e1c1 f8g7 h2h3 b6d5 h1e1 d5c3 d2c3 e8g8 c3d2 f8e8 f3e5 c8e6 f2f4 f6d7 e5d7 e6d7 e2f3 e7e5 a2a4 e5f4 d2f4 d7e6 a4a5 e6d5 f3d5 c6d5 c1d2 f7f5 e1e8 a8e8 d1b1');
INSERT INTO evaluations VALUES(39943,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39943,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39943,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(39943,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39943,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39943,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(39943,36,'cp 15','e6e5 d4f5 e8g8 d1f3 d7d5 e4d5 b4c3 d2c3 c8f5 f3f5 d8d5 f5d3 b8c6 d3d5 f6d5 c3d2 d5b4 e1c1 b4a2 c1b1 a2b4 f1c4 a8c8 c2c3 c6a5 c4e2 b4d5 h1e1 f8d8 e2g4 c8c6 g4f3 c6d6 e1e5 a5c4 e5d5 d6d5 f3d5 d8d5 b1c2 g8f8 d2c1 d5d1 c2d1 f8e7 b2b3 c4e5 c1e3 a7a5 d1c2 e7e6');
INSERT INTO evaluations VALUES(40088,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(40088,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(40088,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(40088,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(40088,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(40088,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40088,36,'cp 16','g1f3 b7b6 f1d3 c8b7 c4d5 c6d5 c1d2 e8g8 e1g1 a7a6 f3e5 f6e4 d2e1 b8d7 e5d7 d8d7 d1c2 e4c3 e1c3 h7h6 c3b4 f8c8 c2e2 e7b4 a3b4 a6a5 b4a5 b6a5 h2h3 d7d6 e2d2 d6d8 d3e2 d8b6 f1c1 c8c1 a1c1 b7a6 e2a6 b6a6');
INSERT INTO evaluations VALUES(40234,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(40234,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(40234,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(40234,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(40234,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(40234,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(40234,36,'cp 30','f1d3 b8c6 e1g1 c6b4 d3e2 b4c6 f1e1 f6e4 g5e7 e4c3 b2c3 c6e7 f3e5 e8g8 e2d3 d8c7 d1h5 g7g6 h5h4 f7f6 e5g4 g8g7 h4h6 g7g8 h2h3 c8d7 a1b1 a8c8 a2a4 b7b6 e1e3 e6e5 d4e5 d7g4 e5f6 e7f5 d3f5 g4f5 e3e7');
INSERT INTO evaluations VALUES(40235,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(40235,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(40235,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(40235,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(40235,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(40235,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40235,36,'cp 21','b8c6 e1g1 c6b4 d3e2 b4c6 f1e1 e8g8 e2d3 c6b4 d3f1 c8d7 a2a3 b4c6 f1d3 a8c8 h2h3 a7a6 c3e2 c6a5 c2c3 a5c4 d1c1 f8e8 e2g3 h7h6 g5f6 e7f6 d3c4 d5c4 g3h5 d7c6 c1f4 f6g5 f4e5 f7f6 e5g3 g7g6 h5f4 g5f4 g3f4 g8g7 e1e3');
INSERT INTO evaluations VALUES(40269,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(40269,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(40269,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(40269,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(40269,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(40269,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(40269,36,'cp 52','g1f3 e8g8 f1e2 d5c4 e2c4 b8d7 e1g1 f6h5 f4g5 h7h6 g5h4 g6g5 h4g3 h5g3 h2g3 e7e5 f1e1 e5d4 e3d4 d7b6 c4b3 c8g4 a2a4 g4f3 d1f3 d8f6 a4a5 b6d7 a1d1 f6f3 g2f3 a8d8 e1e7 g7f6 e7e3 f8e8 e3e8 d8e8 c3e4 f6e7 d4d5 d7c5 b3c4 c5e4 f3e4');
INSERT INTO evaluations VALUES(40799,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40799,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40799,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40799,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40799,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40799,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40799,36,'cp 0','g2g3 c7c5 f1g2 b8c6 b1c3 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(41002,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(41002,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(41002,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41002,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41002,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41002,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41002,36,'cp 0','c7c5 d4c5 e8g8 e1g1 a7a5 c2c3 a5a4 b2b3 a4b3 d2b3 b8a6 c1b2 a6c5 b3c5 e7c5 c3c4 b7b6 d1e2 c8b7 a2a4 d8e7 f3e5 f8d8 f1d1 d5c4 e5c4 b7d5 h2h3 h7h6 c4e5 f6e4 e2c2 e7b7 c2e2');
INSERT INTO evaluations VALUES(41149,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(41149,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41149,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(41149,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(41149,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(41149,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(41149,36,'cp 0','c2c4 c7c6 c1g5 e8g8 e1g1 f8e8 b1d2 c8e6 c4c5 d6c7 g5h4 b8d7 f1e1 d7f8 b2b4 f8g6 h4g3 g6f4 e2f1 e7g6 b4b5 e6d7 e1e8 d8e8 d1b3 c6b5 a1e1 g6e7 b3e3 e7g6');
INSERT INTO evaluations VALUES(41772,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(41772,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(41772,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41772,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(41772,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(41772,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(41772,36,'cp 12','f4d6 d8d6 b1c3 e8g8 e2e3 d5c4 f1c4 c7c5 e1g1 b8c6 d4c5 d6c5 c4e2 f8d8 d1b3 b7b6 a1d1 c8b7 d1d8 a8d8 f1d1 d8d1 b3d1 c5e7 h2h3 a7a6 f3d2 c6e5 d1a4 b6b5 a4d4 e7c7 c3e4 f6e4 d2e4 e5g6 e4d6 b7d5');
INSERT INTO evaluations VALUES(42854,30,'cp 434',NULL);
INSERT INTO evaluations VALUES(42854,31,'cp 452',NULL);
INSERT INTO evaluations VALUES(42854,32,'cp 456',NULL);
INSERT INTO evaluations VALUES(42854,33,'cp 480',NULL);
INSERT INTO evaluations VALUES(42854,34,'cp 493',NULL);
INSERT INTO evaluations VALUES(42854,35,'cp 520',NULL);
INSERT INTO evaluations VALUES(42854,36,'cp 499','d2d4 c5d4 c1h6 e8g8 f3g3 d8f6 h6c1 d7d6 g1f3 b8c6 e1g1 f6g6 g3g6 h7g6 c4b3 c8d7 f3d4 c6d4 c2c3 d4e2 g1h1 e2c1 f1c1 b7b5 b1d2 a7a5 h1g1 f8b8 d2f1 a5a4 b3d1 f7f5 f2f3 b5b4 c3b4 b8b4 b2b3');
INSERT INTO evaluations VALUES(43392,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(43392,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(43392,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(43392,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(43392,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(43392,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(43392,36,'cp 28','d2d4 c5d4 d1d4 b8c6 d4f2 e7e6 h2h3 f8e7 f1d3 e8g8 c1e3 b7b5 e1g1 c6b4 a2a3 b4d3 c2d3 c8b7 f3d4 a8c8 c3e2 f8e8 e2g3 d8d7 g1h2 a6a5 e3d2 e7d8 a1e1 h7h6 e4e5 d6e5 f4e5 d8b6 d2c3 f6d5 g3h5 b5b4 a3b4');
INSERT INTO evaluations VALUES(44533,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44533,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44533,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44533,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44533,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44533,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(44533,36,'cp 0','g7g6 d2d4 c5d4 e3d4 d7d5 c4d5 f6d5 f1c4 f8g7 d1b3 e7e6 c1g5 d8b6 b1c3 h7h6 g5d2 d5e7 b3b6 a7b6 c3b5 e8g8 b5c7 a8a7 c7b5');
INSERT INTO evaluations VALUES(45996,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(45996,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(45996,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(45996,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(45996,34,'cp -106',NULL);
INSERT INTO evaluations VALUES(45996,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(45996,36,'cp -62','b8c6 g1e2 d6d5 c4b3 f8e7 a2a4 e8g8 e1g1 c8d7 h2h3 c6a5 b1d2 d7c6 b3a2 c6a4 e4e5 f6d7 a2b1 a4b5 f3g3 a5c6 d2f3 d8c7 c1h6 e7f6 e5f6 c7g3 e2g3 g7h6 g3h5 a7a5 f1e1 c6d4 f3h2 a5a4 h2g4 d4f5 c2c3 b5c6');
INSERT INTO evaluations VALUES(13938,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(13938,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(13938,32,'cp -70',NULL);
INSERT INTO evaluations VALUES(13938,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(13938,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(13938,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(13938,36,'cp -75','g1e2 d6d5 c4b3 f8e7 b1c3 d5d4 c3b1 h7h5 h2h4 f6g4 f3g3 b7b5 a2a4 c5c4 a4b5 c4b3 b5c6 b3c2 b1a3 e8g8 a3c2 e7d6 e2f4 f7f5 e4e5 g4e5 e1g1 e5c6 g3f3 d6f4 c1f4 c8b7 f3g3 d8d5 f1e1 a7a5 c2a3 c6b4 e1e5 d5b3');
INSERT INTO evaluations VALUES(46770,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(46770,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46770,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46770,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46770,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46770,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(46770,36,'cp 0','c7c5 g1f3 b8c6 f1e2 f8g7 e1g1 c5d4 e3d4 d8b6 f4c1 e8g8 c2c3 c8f5 b1d2 b6c7 f3h4 f5d7 h4f3');
INSERT INTO evaluations VALUES(48773,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(48773,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(48773,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(48773,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(48773,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(48773,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(48773,36,'cp -25','b8c6 b1c3 f8b4 f3d2 d5d4 c3d5 b4e7 d5e7 d8e7 f1e2 h7h6 g5f6 e7f6 e2g4 c8g4 d1g4 a7a5 h2h4 e8g8 e1g1 a5a4 g4d7 f6h4 d7c7 h4e7 c7e7 c6e7 f1c1 a8a6 g1f1 f7f6 f1e2 h6h5 g2g3 h5h4 c2c4 h4g3 f2g3 a6b6 a1b1');
INSERT INTO evaluations VALUES(15078,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(15078,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(15078,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(15078,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(15078,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(15078,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(15078,36,'cp -30','b1c3 f8b4 f3d2 d5d4 c3d5 b4e7 d5e7 d8e7 f1e2 h7h6 g5h4 c8e6 c2c3 e8c8 a1c1 g7g5 h4g3 f6d7 c3d4 c6d4 d2f3 d4f3 g2f3 f7f5 e4f5 e6f5 d3d4 e7b4 d1d2 b4d2 e1d2 h8e8 h2h4 d7f6 c1c4 c7c6 h4g5 h6g5 c4a4 d8d4 a4d4 e5d4');
INSERT INTO evaluations VALUES(51092,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(51092,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(51092,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(51092,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(51092,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(51092,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(51092,36,'cp 106','c7c6 d2d4 g7g6 g1f3 f8g7 c1e3 d7d5 e4d5 c6d5 f1b5 b8c6 f3e5 c8d7 d1e2 g8f6 e1c1 e8g8 e5d7 d8d7 c1b1 f8c8 f2f3 d7d6 g2g4 c6a7 e3c1 c8c3 b2c3 a7b5 e2b5 d6c7 h2h4 c7c3 d1d3 c3c4 b5c4 d5c4');
INSERT INTO evaluations VALUES(52336,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52336,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52336,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52336,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52336,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(52336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52336,36,'cp 0','d4c3 b1c3 f8g7 f1c4 b8c6 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4d6 f8e8 d6f4 c8g4 c4d5 a8c8 d1b3 c6a5 b3b5 a7a6 b5d3 a5c6 d3d2 g4f3 d5f3 h6f5 d2d8 e8d8 a1d1 f5d4 f3e4 d8d7 e4c6 c8c6 f1e1 c6c8 g1f1 h7h6 f4e3 b7b5');
INSERT INTO evaluations VALUES(52341,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52341,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52341,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52341,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52341,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52341,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52341,36,'cp 0','b1c3 b8c6 f1c4 f8g7 e4e5 g8h6 e1g1 e8g8 c1f4 d7d6 e5d6 e7d6 f4d6 f8e8 d6f4 c8g4 c4d5 a8c8 d1b3 c6a5 b3d1');
INSERT INTO evaluations VALUES(52951,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(52951,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(52951,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(52951,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(52951,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(52951,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(52951,36,'cp 10','g8f6 c4d3 e7e5 g1e2 d6d5 d4e5 f6g4 h2h3 d5e4 h3g4 e4d3 c2d3 c8g4 d3d4 d8d7 b1c3 b8a6 f2f3 g4e6 e2f4 h7h5 c1e3 h5h4 d1a4 f8e7 f4e6 d7e6 e1c1 a6b4 a2a3 b4d5 c3d5 e6d5 c1b1 a7a6 a4c2 e8c8 f3f4 c8b8 f4f5 b8a8 c2e2');
INSERT INTO evaluations VALUES(54441,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(54441,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(54441,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(54441,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54441,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(54441,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(54441,36,'cp -2','f1c4 g8e7 e1g1 a7a6 f1e1 e7f5 c4f1 f8e7 d2d3 c8d7 b1d2 d7c6 d2e4 e8g8 c1f4 b8d7 g2g4 f5h4 f3h4 e7h4 f1g2 h4e7 a1c1 e7b4 e1e2 b4e7');
INSERT INTO evaluations VALUES(679,30,'cp 878',NULL);
INSERT INTO evaluations VALUES(679,31,'cp 881',NULL);
INSERT INTO evaluations VALUES(679,32,'cp 901',NULL);
INSERT INTO evaluations VALUES(679,33,'cp 896',NULL);
INSERT INTO evaluations VALUES(679,34,'cp 919',NULL);
INSERT INTO evaluations VALUES(679,35,'cp 919',NULL);
INSERT INTO evaluations VALUES(679,36,'cp 922','f3c6 e7d7 c6a8 f8d6 d2d3 e8f7 b1d2 h7h6 a8f3 f7g8 f3e4 g8f7 b2b3 d7f5 e4f5 c8f5 e1e2 d6b4 d2e4 a7a5 a2a4 f7e6 h2h4 h8d8 h4h5 f5e4 d3e4 d8d4 c1e3 d4d6 a1a2 d6a6 g2g3 e6e7 h1d1 a6c6 d1d5 b4c3 e2d3 e7f6');
INSERT INTO evaluations VALUES(678,30,'cp 905',NULL);
INSERT INTO evaluations VALUES(678,31,'cp 896',NULL);
INSERT INTO evaluations VALUES(678,32,'cp 896',NULL);
INSERT INTO evaluations VALUES(678,33,'cp 888',NULL);
INSERT INTO evaluations VALUES(678,34,'cp 901',NULL);
INSERT INTO evaluations VALUES(678,35,'cp 890',NULL);
INSERT INTO evaluations VALUES(678,36,'cp 894','e7d7 c6a8 f8d6 d2d3 e8f7 b1d2 h7h6 a8f3 d7f5 f3f5 c8f5 e1e2 h6h5 h2h4 h8h6 d2e4 h6g6 c1g5 d6f8 a2a4 g6b6 b2b3 a7a6 g5d2 f7e6 f2f3 c7c5 a4a5 b6c6 g2g4 h5g4 a1g1 e6f7 f3g4 f5d7 g4g5 f7e6 h4h5');
INSERT INTO evaluations VALUES(1397,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(1397,31,'cp 255',NULL);
INSERT INTO evaluations VALUES(1397,32,'cp 263',NULL);
INSERT INTO evaluations VALUES(1397,33,'cp 262',NULL);
INSERT INTO evaluations VALUES(1397,34,'cp 254',NULL);
INSERT INTO evaluations VALUES(1397,35,'cp 277',NULL);
INSERT INTO evaluations VALUES(1397,36,'cp 259','b1c3 d5d6 d4d5 c6d4 e2e4 c8d7 f3d4 c5d4 c3b5 d6c5 d1d4 c5d4 b5d4 g8f6 f1d3 e7e5 d5e6 f7e6 f2f3 f8d6 c1e3 e8e7 e1e2 h8c8 d4b3 a7a5 a2a4 c8c6 e3d2 b7b6 d2c3 d6b4 c3b4 a5b4 a4a5 b6a5 b3a5 c6c7 a5c4 a8c8');
INSERT INTO evaluations VALUES(1393,30,'cp 263',NULL);
INSERT INTO evaluations VALUES(1393,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(1393,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(1393,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(1393,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(1393,35,'cp 269',NULL);
INSERT INTO evaluations VALUES(1393,36,'cp 270','d5d6 d4d5 c6d4 e2e4 c8d7 f3d4 c5d4 c3b5 d6c5 d1d4 c5d4 b5d4 e7e6 d5e6 f7e6 f1d3 g8f6 e1e2 f8d6 c1d2 e8g8 a1c1 a8c8 g2g3 c8c1 h1c1 e6e5 d4f5 f6e4 d3e4 d7f5 e4b7 f8b8 b7d5 g8f8 c1c6 b8b6 f2f3 b6c6 d5c6 f8e7 c6e4 f5e6');
INSERT INTO evaluations VALUES(2026,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(2026,31,'cp -59',NULL);
INSERT INTO evaluations VALUES(2026,32,'cp -66',NULL);
INSERT INTO evaluations VALUES(2026,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(2026,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(2026,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(2026,36,'cp -72','e7e4 f1e2 d7c6 e1g1 c8e6 b1c3 e4h4 g2g3 h4d8 c3e4 f8e7 c1e3 g8f6 e4c5 e7c5 e3c5 b7b6 c5a3 d8d1 f1d1 c6c5 b2b3 e8g8 a3b2 a8d8 b2e5 e6f5 e5c7 d8d1 a1d1 f5c2 d1d2 c2f5 e2a6 f6e4 d2d5 f5e6 d5e5 e4f6 f2f3 f6d5 c7d6');
INSERT INTO evaluations VALUES(2261,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(2261,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(2261,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(2261,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(2261,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(2261,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(2261,36,'cp 187','e8d8 c4f7 f8b4 b1d2 d8e7 f7b3 g8f6 f2f3 g7g5 a2a3 b4d6 d2c4 g5g4 c1d2 c8e6 e1c1 a8f8 h1f1 e6c4 b3c4 g4f3 g2f3 f6h5 b2b4 a7a6 c1b2 h5f4 f1g1 b7b5 c4a2 f4e6 g1g6 f8f6 g6f6 e7f6 a2d5 h8g8 h2h4 g8g2 d2h6 e6d4');
INSERT INTO evaluations VALUES(2310,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(2310,31,'cp 211',NULL);
INSERT INTO evaluations VALUES(2310,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(2310,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(2310,34,'cp 211',NULL);
INSERT INTO evaluations VALUES(2310,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(2310,36,'cp 197','c4f7 g8f6 f2f3 f8b4 b1d2 d8e7 f7b3 g7g5 a2a3 b4d6 d2c4 g5g4 c1e3 c8e6 e1c1 b7b6 h1f1 g4f3 g2f3 a8f8 e3d2 e6c4 b3c4 f6h5 d2c3 h5f4 f1g1 h8h7 c3e1 f4h3 g1g6 f8f3 g6e6 e7d7 e6e5 h3f2 d1d5 f3f4');
INSERT INTO evaluations VALUES(2308,30,'cp 199',NULL);
INSERT INTO evaluations VALUES(2308,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(2308,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(2308,33,'cp 203',NULL);
INSERT INTO evaluations VALUES(2308,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(2308,35,'cp 192',NULL);
INSERT INTO evaluations VALUES(2308,36,'cp 217','f8b4 b1d2 g8f6 a2a3 b4d6 f2f3 d8e7 f7g6 c8e6 d2f1 d6c5 c1e3 c5d4 e3d4 e5d4 e1c1 c7c5 f1g3 h8d8 h1e1 b7b5 g3f5 e7f8 f5h4 a8c8 g2g3 a7a5 a3a4 c8b8 f3f4 f6g4 a4b5 b8b5');
INSERT INTO evaluations VALUES(9795,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9795,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9795,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(9795,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(9795,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(9795,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(9795,36,'cp -14','e5d4 c3d4 c5b4 c1d2 f6e4 d2b4 c6b4 c4f7 e8f7 d1b3 d7d5 b3b4 h8f8 e1g1 f7g8 b1d2 a7a5 b4b3 a5a4 b3e3 e4d2 f3d2 f8e8 e3c3 c8g4 f1e1 d8f6 d2f1 g4h5 h2h3 c7c6 a1c1 e8f8 c3d2 a4a3 b2b4 h5g6 e1e3 g8h7 c1c3 a8a4 e3f3 f6d6 f3f8 d6f8');
INSERT INTO evaluations VALUES(9823,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(9823,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(9823,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(9823,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(9823,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(9823,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(9823,36,'cp 0','e4e5 d7d5 c4b5 f6e4 c3d4 c5b6 b1c3 c8g4 h2h3 g4h5 d1a4 e8g8 b5c6 b7c6 a4c6 b6a5 f3d2 e4c3 b2c3 f7f6 d2b3 h5e8 c6e6 e8f7 e6c6 f7e8');
INSERT INTO evaluations VALUES(10421,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(10421,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10421,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10421,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10421,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10421,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10421,36,'cp 7','b4c3 b2c3 d7d6 e2e4 f6d7 d2d3 d7c5 d3d4 c5d7 g1f3 e8g8 e1g1 f8e8 c1e3 d7f6 f3d2 f6g4 d1e2 b7b6 f2f4 c6a5 f4f5 g4e3 e2e3 f7f6 f1f2 g8h8 g3g4 a7a6 g2f1 c8d7 f1d3 d7a4 e3g3 d8e7');
INSERT INTO evaluations VALUES(10453,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10453,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10453,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(10453,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(10453,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(10453,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(10453,36,'cp 13','b2c3 e8g8 e2e4 c6a5 d2d3 c7c6 g1f3 d7d5 c4d5 c6d5 e4d5 d8d5 c3c4 d5d6 e1g1 f8e8 c1b2 c8f5 f1e1 a5c6 d1b3 a8d8 a1d1 b7b6 d3d4 c6d4 f3d4 e5d4 e1e8 d8e8 b2d4 e8d8 g2f3 d6e6 d4c3 d8e8 d1e1 e6d7 e1e8 d7e8 c4c5 f5e6 b3c2 b6c5 c3f6 g7f6 c2c5 e6h3 c5e3 e8e3 f2e3 g8g7 f3e4 h7h6');
INSERT INTO evaluations VALUES(10452,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(10452,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(10452,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(10452,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10452,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10452,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(10452,36,'cp 0','e8g8 e2e4 c6a5 d2d3 c7c6 g1f3 f8e8 e1g1 h7h6 h2h3 d7d6 a3a4 c8e6 a1b1 a8c8 c4c5 b7b6 c5d6 d8d6 f1e1 c8d8 d3d4 f6d7 c1e3 d6a3 e3c1');
INSERT INTO evaluations VALUES(12612,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12612,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12612,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12612,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12612,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12612,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12612,36,'cp 1','b1c3 d7d5 e4d5 c8f5 d5c6 b7c6 d1f3 d8d7 f1c4 f8c5 c1g5 f5c2 g5f6 g7f6 e1g1 e8g8 a1c1 c2g6 c3e4 g6e4 f3e4 c5d4 c4d3 f6f5 e4f5 d7f5 d3f5 d4b2 c1c6 b2d4 g2g3 a8b8 c6c2 f8d8 g1g2 b8b2 f1c1 b2c2 c1c2 g8g7 c2d2 d4b6 d2d8 b6d8 g2f3 h7h5 h2h4');
INSERT INTO evaluations VALUES(13878,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(13878,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(13878,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13878,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(13878,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13878,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(13878,36,'cp 25','f8e7 d1d2 h7h6 g5f6 e7f6 g2g3 a7a5 a2a4 e8g8 h2h4 c6b4 e1c1 d8b6 f1h3 c8e6 h3e6 f7e6 d2d6 b6f2 d6e6 g8h8 b3d2 f2g3 e6h3 g3f2 h3h2 f2f4 d1f1 f4h2 h1h2 a8c8 c1b1 b7b6 d2f3 f6e7 h2h1 c8c5 h4h5 f8f4');
INSERT INTO evaluations VALUES(8141,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(8141,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(8141,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(8141,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(8141,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(8141,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(8141,36,'cp 13','g5f6 e7f6 f1c4 a7a6 e1g1 e8g8 a2a4 c8e6 c3d5 a8c8 c2c3 f6g5 d1e2 c6e7 d5e7 d8e7 c4e6 e7e6 b3d2 g5d2 e2d2 c8c6 f1d1 f7f5 e4f5 f8f5 d2e3 e6b3 d1d2 d6d5 f2f3 d5d4 e3e4 d4c3 b2c3 f5f8 e4e5 b3c3 e5c3 c6c3 a1b1 c3c7 b1d1 f8c8 h2h4 c7c5');
INSERT INTO evaluations VALUES(17944,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17944,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17944,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17944,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17944,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17944,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17944,36,'cp 0','d7c6 d2d4 e5d4 d1d4 d8d4 f3d4 c8d7 h2h4 e8c8 f2f3 h7h5 d4e2 d7e6 c1e3 f8d6 e1f2 e6f7 a1d1 g8e7 e3f4 d6c5 f4e3');
INSERT INTO evaluations VALUES(21133,30,'cp -163',NULL);
INSERT INTO evaluations VALUES(21133,31,'cp -189',NULL);
INSERT INTO evaluations VALUES(21133,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(21133,33,'cp -169',NULL);
INSERT INTO evaluations VALUES(21133,34,'cp -185',NULL);
INSERT INTO evaluations VALUES(21133,35,'cp -199',NULL);
INSERT INTO evaluations VALUES(21133,36,'cp -174','e7e5 g1f3 c6d4 f1c4 f8c5 b1c3 c8e6 c4e6 d4e6 f3e5 g8f6 e5d3 c5b6 h2h4 d8d4 d1f3 b6a5 c1d2 a5c3 d2c3 d4e4 f3e4 f6e4 c3e5 e8c8 e1c1 d8d3 d1d3 e4f2 h1h3 f2d3 h3d3 f7f6 e5h2 h8d8 d3e3 c8d7 h4h5 b7b6 b2b4 g7g5 h5g6 h7g6');
INSERT INTO evaluations VALUES(22441,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(22441,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(22441,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(22441,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(22441,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(22441,35,'cp 61',NULL);
INSERT INTO evaluations VALUES(22441,36,'cp 63','f8g7 c2c3 a7a6 b5a4 d7d6 d2d4 c8d7 c1e3 g8f6 b1d2 e8g8 h2h3 b7b5 a4c2 a6a5 d4e5 c6e5 f3e5 d6e5 a2a4 d8b8 d1b1 d7e6 f1e1 f8d8 b2b4 c7c6 d2b3 a5b4 c3b4 a8a4 b3c5 a4a8 c5e6 a8a1 b1a1 f7e6');
INSERT INTO evaluations VALUES(6173,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(6173,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(6173,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(6173,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(6173,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(6173,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(6173,36,'cp 61','c2c3 a7a6 b5a4 d7d6 d2d4 c8d7 c1e3 g8f6 b1d2 e8g8 h2h3 a6a5 f1e1 d8e7 a4c2 a5a4 e3g5 b7b5 d2f1 h7h6 g5h4 f8e8 f1e3 e7d8 d4e5 c6e5 f3e5 e8e5 d1f3 e5e6 a2a3 d7c6 a1d1 c6e4 c2e4 e6e4');
INSERT INTO evaluations VALUES(6172,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(6172,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(6172,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(6172,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(6172,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(6172,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(6172,36,'cp 70','a7a6 b5a4 d7d6 d2d4 c8d7 c1e3 g8f6 b1d2 e8g8 h2h3 b7b5 a4c2 a6a5 d4e5 c6e5 f3e5 d6e5 a2a4 d8b8 d1b1 f8d8 f1d1 b5a4 c2a4 d7b5 b1a2 h7h6 a4b5 b8b5 a2a4 b5b7 d1b1 g7f8 a4c2 g6g5 b2b4 a5b4 a1a8 d8a8 c3b4');
INSERT INTO evaluations VALUES(22527,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(22527,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(22527,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(22527,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(22527,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(22527,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22527,36,'cp 12','g8f6 e1g1 d7d5 e4d5 d8d5 a2a3 f8e7 c2c4 d5d6 b2b4 e8g8 b1d2 a7a5 c1b2 a5b4 a3b4 a8a1 d1a1 d6b4 b2c3 b4d6 f3e5 c6e5 c3e5 d6d8 d2f3 f8e8 d3d4 e7d6 a1b2 d6e5 f3e5 c7c5 e5f3 c5d4 f3d4 d8c7 h2h3 c8d7 d4f3 g8h8');
INSERT INTO evaluations VALUES(23253,30,'cp 204',NULL);
INSERT INTO evaluations VALUES(23253,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(23253,32,'cp 212',NULL);
INSERT INTO evaluations VALUES(23253,33,'cp 191',NULL);
INSERT INTO evaluations VALUES(23253,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(23253,35,'cp 214',NULL);
INSERT INTO evaluations VALUES(23253,36,'cp 210','g8e7 a2a3 c6d4 d1d4 e7c6 d4e3 f8d6 c1d2 d6e5 f1c4 d7d6 c4d5 e5c3 e3c3 c6e7 d5c4 e7c6 f2f4 d8e7 c4d5 c8d7 e1g1 e8c8 a3a4 e7e8 b2b4 c6e7 d5b3 f6f5 a1e1 e8g6 c3d4 b7b6 d2c3 h8e8 d4g7 g6g7 c3g7 f5e4 b3f7 d7f5 g7f6 c8d7 f7e8 d8e8 f6e7 d7e7 c2c4');
INSERT INTO evaluations VALUES(24611,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(24611,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(24611,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(24611,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(24611,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(24611,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(24611,36,'cp 88','d4c5 f6e4 f3d4 d8d7 f1d3 g7g6 d3e4 f5e4 f2f3 e4b1 a1b1 f8g7 f4g3 e8g8 e1g1 f8c8 b2b4 c6d8 d1d2 b7b6 d4b3 b6c5 b3c5 d7c6 g3h4 e7e6 e3e4 d8b7 e4d5 e6d5 h4f2 b7c5 f2c5 a7a5 a2a3 c6b5 f1d1 b5c4 g1h1 a5b4 c3b4 a8a3');
INSERT INTO evaluations VALUES(24683,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24683,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24683,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24683,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24683,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24683,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(24683,36,'cp 0','b5c3 d8b6 a1b1 a6b4 b1c1 e7e6 f1e2 b4a6 c1b1 a6b4');
INSERT INTO evaluations VALUES(24684,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24684,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24684,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24684,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24684,36,'cp 0','d8b6 a1b1 a6b4 b1c1 e7e6 f1e2 b4a6 c1b1 a6b4');
INSERT INTO evaluations VALUES(25426,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(25426,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25426,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(25426,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(25426,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(25426,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(25426,36,'cp -12','f8e7 d3d4 e5d4 f3d4 d8d7 d4c6 d7c6 d1d3 h7h6 g5e3 e8g8 e1g1 f6d7 c3e2 d7e5 d3b3 c6e4 b3e6 g8h7 e6b3 a8e8 e2g3 e4c4 f1d1 b7b6 e3d4 c4b3 a2b3 e5c6 d4c3 a7a5 d1e1 e7h4 e1e8 f8e8 a1d1 e8f8 f2f3 h4f6 c3f6 f8f6 g1f2');
INSERT INTO evaluations VALUES(24037,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(24037,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(24037,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(24037,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24037,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(24037,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(24037,36,'cp -15','d3d4 e5d4 f3d4 d8d7 d4c6 d7c6 d1d3 h7h6 g5d2 e8c8 e1c1 c8b8 f2f3 f6d7 d3b5 g7g5 b5c6 b7c6 b2b3 d8f8 h2h3 d7e5 c3a4 e5g6 d2c3 h8g8 a4b2 h6h5 b2c4 h5h4 c4a5 c6c5 c1b2 b8c8 a5c6 c8b7 c6a5 b7b6 a5c4 b6c6 c4a5 c6d7 a5c4');
INSERT INTO evaluations VALUES(26329,30,'cp 274',NULL);
INSERT INTO evaluations VALUES(26329,31,'cp 275',NULL);
INSERT INTO evaluations VALUES(26329,32,'cp 294',NULL);
INSERT INTO evaluations VALUES(26329,33,'cp 299',NULL);
INSERT INTO evaluations VALUES(26329,34,'cp 285',NULL);
INSERT INTO evaluations VALUES(26329,35,'cp 301',NULL);
INSERT INTO evaluations VALUES(26329,36,'cp 279','c4f7 e8f7 f3g5 f7e8 d1g4 d8c8 g4h4 g8f6 f2f4 g7g6 e1g1 f8g7 b1c3 h7h6 g5f3 c8e6 h4e1 e8d7 a2a3 f6h5 a1b1 a8f8 c3d5 g6g5 f4g5 h6g5 c1g5 c6d4 f3d4 g7d4 g5e3 f8f1 e1f1 d4e3 d5e3 e6h6 e3f5 h6g5 d3d4 e7e6 f1b5 d7c7');
INSERT INTO evaluations VALUES(28439,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(28439,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(28439,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(28439,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(28439,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(28439,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(28439,36,'cp 119','d1b3 g4f3 g2f3 d5c4 b3b7 b8d7 b7c6 e8g8 f3f4 d7b6 f1g2 e7e6 b2b3 c4b3 a2b3 b6d5 e1g1 d8d6 c6c4 f8b8 c1a3 d6d7 f1c1 a7a5 c3a4 d7b5 a1b1 b5e8 c4c6 g7f8 c6e8 b8e8 a3f8 g8f8 h2h3 e8b8 a4c5 f8e7 c5d3');
INSERT INTO evaluations VALUES(29249,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(29249,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(29249,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(29249,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(29249,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(29249,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(29249,36,'cp 4','d1b3 d8b6 c4c5 b6b3 a2b3 h7h6 b3b4 a7a6 b1c3 f8e7 e1g1 e8g8 h2h4 b8d7 a1a3 f8e8 f1d1 a8d8 c1e3 f6e4 c3e4 f5e4 g2h3 e7f6 f3d2 e4h7 f2f4 f6e7 e3f2 d7f6 f2e1 h7f5 h3g2 f6e4 d2e4 f5e4');
INSERT INTO evaluations VALUES(29467,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(29467,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(29467,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(29467,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(29467,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(29467,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(29467,36,'cp 86','c7c6 d1b3 d8b6 c4c5 b6c7 c1f4 c7c8 f3g5 e6d7 e2e4 h7h6 g5f3 f6e4 c3e4 d5e4 f3e5 d7e6 f1c4 e6c4 e5c4 b8d7 e1g1 g7g5 f4e5 d7e5 c4e5 e7e6 a1e1 c8c7 e5c4 e8c8 e1e4 c8b8 g1h1 h6h5 f1d1 h5h4 h2h3 d8d5 b3f3 d5d8 h1g1 f8g7');
INSERT INTO evaluations VALUES(28927,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(28927,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(28927,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(28927,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(28927,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(28927,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(28927,36,'cp 82','d1b3 e6c8 c1f4 e7e6 e2e3 f8d6 f4d6 d8d6 f1d3 e8g8 e1g1 b8d7 a1d1 b7b6 f1e1 f8d8 h2h3 c8b7 e3e4 d5e4 c4c5 d6c7 c3e4 f6e4 d3e4 d7f6 d1c1 b6c5 c1c5 f6d7 c5c3 a8c8 b3c2 d7f6 e1c1 h7h6 e4c6 b7c6 c3c6 c7c6 c2c6 c8c6');
INSERT INTO evaluations VALUES(30174,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(30174,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30174,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30174,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30174,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30174,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(30174,36,'cp 0','c2c4 c7c5 c4d5 f6d5 d3b5 b8d7 e1g1 c5d4 d1d4 g4f3 g2f3 a7a6 b5e2 f8d6 f3f4 d8h4 b1c3 d7f6 c3d5 f6g4 e2g4 h4g4 g1h1 g4f3 h1g1 f3g4');
INSERT INTO evaluations VALUES(30633,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(30633,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(30633,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(30633,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(30633,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(30633,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(30633,36,'cp 34','c4d3 c7c5 c2c3 b8c6 g1e2 f8d6 e1g1 g8e7 b1a3 c5d4 c3d4 d6a3 b2a3 e8g8 c1e3 d7d5 e4e5 e7f5 a1c1 a8c8 d3f5 e6f5 e2f4 b7a6 f1e1 a6c4 d1h5 d8d7 h5f3 c6e7 h2h4 h7h6 e3d2 e7c6 e5e6 f7e6 f4e6');
INSERT INTO evaluations VALUES(32166,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(32166,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32166,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32166,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32166,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32166,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(32166,36,'cp 1','e5f4 d2d4 g4f3 e2f3 d8h4 e1f1 g8e7 b1c3 b8c6 c3e2 g7g5 g2g3 h4h3 f3g2 h3h5 d1d3 f4g3 h2g3 h5g6 c2c3 h7h6 g3g4 e8c8 e2g3 c8b8 c1d2 d6d5 a1e1 f8g7 f1g1 d5e4 d3e4 g6e4 g3e4 e7g6 h1h3 g6h4 g2h1 h8f8 e4g3 c6e5 g3f5 h4f5 g4f5');
INSERT INTO evaluations VALUES(33707,30,'cp -93',NULL);
INSERT INTO evaluations VALUES(33707,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(33707,32,'cp -95',NULL);
INSERT INTO evaluations VALUES(33707,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(33707,34,'cp -114',NULL);
INSERT INTO evaluations VALUES(33707,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(33707,36,'cp -121','c7c6 a2a3 b4a5 f3g3 c8f5 e3f4 b8d7 f4c7 f6h5 c7a5 h5g3 h2g3 b7b6 g1f3 b6a5 f1a6 d7c5 a6e2 a8c8 f3e5 g7g6 e2f3 f8g7 e5c6 g7c3 b2c3 f5d7 c6a5 h7h5 h1h4 e8g8 h4b4 c8c7 a5b7 c5e6 b7a5 d7c8 b4c4 c7c4 a5c4 e6g5 c4e3 g5f3 g2f3 c8e6 e3d5 g8g7 d5e7');
INSERT INTO evaluations VALUES(34535,30,'cp 122',NULL);
INSERT INTO evaluations VALUES(34535,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(34535,32,'cp 108',NULL);
INSERT INTO evaluations VALUES(34535,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(34535,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(34535,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34535,36,'cp 86','c2c4 e7e6 c4c5 d6d8 b1c3 g8f6 c1f4 f8e7 c3b5 b8a6 a2a3 e8g8 b2b4 f6h5 f4d2 h5f6 b5c3 a6b8 e2e3 b7b6 f1d3 c7c6 e1g1 c8a6 d3a6 b8a6 d1a4 a6b8 f1c1 d8c7 c5b6 c7b6 a4c2 f8c8 c3a4 b6c7 a4b2 a7a5 b4a5 c6c5 d4c5 b8c6 b2d3 c6a5');
INSERT INTO evaluations VALUES(35042,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(35042,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(35042,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(35042,33,'cp 99',NULL);
INSERT INTO evaluations VALUES(35042,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(35042,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(35042,36,'cp 110','h4d8 a2a3 d7d6 c2c4 g7g6 f1e2 b8d7 g1f3 f8g7 b1c3 g8f6 e1g1 e8g8 d1c2 d8e7 f4g3 b7b6 f1e1 c8b7 a1d1 f6h5 g3h4 g7f6 h4f6 h5f6 e3e4 c7c5 e2f1 e6e5 d4e5 d7e5 f3e5 d6e5 c3d5 b7d5 c4d5');
INSERT INTO evaluations VALUES(36981,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(36981,31,'cp 128',NULL);
INSERT INTO evaluations VALUES(36981,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(36981,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(36981,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(36981,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(36981,36,'cp 149','e5f3 c5b4 c3d5 f6d5 e4d5 f8e8 c4e2 d8e7 c2c3 b4a5 d2d4 b8d7 a2a4 c7c6 c1g5 d7f6 g5f6 e7f6 e1g1 g7g6 f1e1 c8d7 d5c6 b7c6 d1d3 g8g7 b2b4 a5c7 a4a5 d7g4 f3d2 g4f5 d3g3 c7b8 e2d3 d6d5 g3f3 e8e1 a1e1 f6g5 d3f5 g5d2');
INSERT INTO evaluations VALUES(37002,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(37002,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(37002,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(37002,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(37002,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(37002,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(37002,36,'cp 131','c8g4 e1g1 f8e8 d2d4 c5b6 h2h3 g4f3 g2f3 b8d7 g1h2 c7c5 d4c5 d7c5 f1g1 c5e6 c4e6 f7e6 c1h6 e8e7 d1e2 d8e8 g1g2 e8h5 h6f4 e7f7 a1g1 e6e5 f4e3 b6e3 f2e3 h5h4 e2e1 h4e1 g1e1 a8d8 h3h4');
INSERT INTO evaluations VALUES(37233,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(37233,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(37233,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(37233,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(37233,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(37233,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(37233,36,'cp 12','d7d6 c2c3 h7h6 f1e1 c7c6 h2h3 c5b6 b1d2 b8d7 c4b3 f8e8 d3d4 d7f8 b3c2 b6c7 d2f1 d6d5 f3e5 f6e4 d1h5 c8e6 c1f4 d8f6 h5f3 g7g5 f4h2 f6f3 e5f3 c7h2 g1h2 f7f5 h3h4 e4f2 h4g5 h6g5 f3g5 f2e4 g5e6 f8e6');
INSERT INTO evaluations VALUES(37614,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(37614,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(37614,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(37614,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(37614,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(37614,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(37614,36,'cp 21','b8c6 d2d4 c5d4 c3b5 d7d6 b5d4 c6d4 f3d4 g8f6 f1e1 e8g8 c4f1 a7a6 c2c4 c8d7 a2a4 f6g4 h2h3 d8b6 a4a5 b6d4 h3g4 d4d1 e1d1 d7c6 f2f3 f8d8 a1b1 d6d5 e4d5 e6d5 c1g5 d8d7 g5e3 d5c4 d1d7 c6d7 f1c4 a8c8 c4d5 d7c6 d5c6 c8c6 b2b4 h7h5 g4h5');
INSERT INTO evaluations VALUES(38266,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(38266,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(38266,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(38266,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(38266,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(38266,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(38266,36,'cp 106','e2e4 c7c5 d4d5 e7f6 g1e2 g8e7 e2g3 e8g8 f1d3 b8d7 e1g1 f6c3 b2c3 e6e5 a2a4 e7g6 g3f5 d7f6 g2g3 f8e8 f5e3 c8h3 f1e1 b7b6 f2f3 h7h5 d3f1 h3f1 e1f1 g6f8 h2h4 g7g6 d1c2 g8g7 e3d1 d8c7 g1g2 f8d7 d1f2');
INSERT INTO evaluations VALUES(39365,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(39365,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(39365,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(39365,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(39365,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(39365,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(39365,36,'cp 10','c1d2 b4d2 d1d2 g8f6 g1f3 d7d6 e2e3 e8g8 f1e2 c7c6 e1g1 a7a5 h2h3 c8f5 b1c3 f8e8 a2a3 b8a6 a1d1 h7h6 f1e1 a6c7 e2f1 f5g6 e3e4 c6c5 d1b1 c7e6 g2g3 g6h5 f1g2 f6d7 f3h4 e6d4 c3b5 d4b3 d2c2 b3d4 b5d4');
INSERT INTO evaluations VALUES(39364,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(39364,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(39364,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(39364,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(39364,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39364,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39364,36,'cp 0','b4d2 d1d2 g8f6 g1f3 d7d6 e2e3 e8g8 f1e2 c8f5 b1c3 a7a5 e1g1 c7c6 a2a3 b8a6 a1b1 d8e7 b2b4 f5g6 h2h3 f8d8 b4a5 e5e4 d3e4 f6e4 c3e4 g6e4 b1b2 a6c5 f3d4 c5a4 b2b3 a4c5');
INSERT INTO evaluations VALUES(40974,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40974,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(40974,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40974,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(40974,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(40974,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40974,36,'cp 12','b1c3 e8g8 c1f4 b7b6 e2e3 c7c5 d4c5 b6c5 f1e2 c8a6 a1c1 a6c4 e2c4 d5c4 f3e5 f6d5 d1f3 e7f6 e5c4 f6c3 b2c3 d5f4 e3f4 d8d5 f3d5 e6d5 c4e3 f8d8 e1g1 b8c6 f1d1 c6e7 g2g3 f7f6 c1b1 d5d4 c3d4 c5d4 b1b7 d4e3 d1d8 a8d8');
INSERT INTO evaluations VALUES(43747,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43747,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43747,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43747,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(43747,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43747,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43747,36,'cp 0','f8e7 d1f3 d8b6 e1c1 b8d7 c1b1 e8g8 f3e2 b6c7 g2g4 d7c5 a2a3 c8b7 f2f3 c5b3 c2b3 b5b4 c3a4 b4a3 b2a3 f8b8 b1a2 b7c8 d1c1 c7b7 c1c6 c8d7 c6b6 b7a7 d4c6 d7c6 b6c6 a7b7 c6b6 b7a7');
INSERT INTO evaluations VALUES(50838,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(50838,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(50838,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(50838,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(50838,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(50838,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(50838,36,'cp 51','e2e4 d5e4 c3e4 c8f5 f1d3 e7e6 g1f3 g8f6 c1g5 f8e7 e4f6 e7f6 d3f5 f6g5 d1e2 g5e7 f5e4 c7c6 e1c1 d8c7 c2c3 b8d7 c1b1 e8c8 g2g3 c8b8 f3d2 d7b6 h1e1 h7h5 h2h4 c6c5 d1c1 g7g5 h4g5 e7g5 d4c5 c7c5 d2b3');
INSERT INTO evaluations VALUES(50843,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(50843,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(50843,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(50843,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(50843,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(50843,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(50843,36,'cp 31','d5e4 c3e4 c8f5 f1d3 e7e6 g1f3 g8f6 c1g5 f8e7 e4f6 e7f6 d3f5 f6g5 d1e2 g5h6 c2c3 c7c6 f3e5 d8e7 f5e4 b8d7 e5c4 e8g8 e1g1 a8d8 g2g3 g7g6 a2a4 h6g7 a1d1 h7h5 a4a5 d7f6 e4f3 f6d5 h2h4 e7c7 d1d3 c6c5 d4c5');
INSERT INTO evaluations VALUES(50863,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(50863,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(50863,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(50863,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(50863,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(50863,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(50863,36,'cp 48','c3e4 g8f6 e4f6 e7f6 f1d3 b8c6 c2c3 f8d6 g1e2 c6e7 d1c2 h7h6 e1g1 e8g8 f1e1 c7c6 c1f4 c8e6 d3h7 g8h8 h7e4 f8e8 h2h3 e7d5 f4d6 d8d6 b2b3 a8d8 a1d1 d5f4 c3c4 h8g8 e2f4 d6f4 e1e3 f4c7 e4h7 g8f8 c2d2 c7d6 h7e4');
INSERT INTO evaluations VALUES(50862,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(50862,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(50862,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(50862,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(50862,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(50862,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(50862,36,'cp 36','c8f5 f1d3 e7e6 g1f3 g8f6 c1g5 f8e7 e4f6 e7f6 d3f5 f6g5 d1e2 g5h6 c2c3 c7c6 f3e5 d8e7 f5e4 b8d7 e5c4 e8g8 e1g1 g7g6 g2g3 d7f6 e4g2 a8d8 a2a4 h6g7 f1e1 h7h5 a4a5 f8e8 a1d1 f6d5 h2h4 e7c7 e2f3 d5e7');
INSERT INTO evaluations VALUES(50906,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(50906,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(50906,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(50906,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(50906,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(50906,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(50906,36,'cp 76','d2d4 e5d4 c2c3 d4d3 e1g1 b8c6 c1f4 f8e7 c4d3 g8f6 h2h3 e8g8 b1d2 f6d7 f3d4 c6e5 d3c2 b7b6 f4g3 c8a6 f2f4 e5d3 a1b1 d3c5 d1g4 a6f1 b1f1 d7f6 g4f3 c5e6 d4f5 g8h8 f1d1 f6d7');
INSERT INTO evaluations VALUES(54376,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54376,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(54376,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(54376,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(54376,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(54376,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(54376,36,'cp 9','f1d3 f8d6 g1f3 g8f6 f3e5 b8d7 f2f4 f6e4 e1g1 e8g8 c1d2 f7f5 b1c3 d7e5 f4e5 d6e7 c3e4 f5e4 f1f8 e7f8 d3e2 c8d7 a2a4 a8c8 h2h3 f8e7 g1h2 h7h6 a4a5 g8h7 d1b3 e7h4 a1f1 b7b6 a5b6 d8b6 b3b6 a7b6 d2c3 d7e8 f1f8');
INSERT INTO evaluations VALUES(54477,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54477,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54477,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54477,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54477,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54477,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54477,36,'cp 0','f8c5 g1f3 g8f6 e1g1 e8g8 b1c3 h7h6 h2h3 b8c6 a2a3 a7a6 c1f4 f8e8 f1e1 e8e1 d1e1 f6h5 f4h2 c8e6 e1d2 a8c8 a1e1 h5f6 b2b4 c5b6 h2f4 d8f8 f3e5 f6h5 f4h2 h5f6');
INSERT INTO evaluations VALUES(54906,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(54906,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(54906,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(54906,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(54906,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(54906,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(54906,36,'cp 106','e2e4 g8f6 b1c3 e7e6 f1d3 b8c6 c1e3 f6g4 e3d2 e6e5 d3b5 c8d7 b5c6 b7c6 h2h3 e5d4 f3d4 g4e5 d1e2 c6c5 d4f3 e5g6 h3h4 f8e7 h4h5 g6f8 e1c1 f8e6 c3d5 e8g8 h1g1 c7c6 d5e7 d8e7 d2c3 f7f6 f3h4 e6f4 e2e3 f4h5');
INSERT INTO evaluations VALUES(56610,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(56610,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(56610,32,'cp 129',NULL);
INSERT INTO evaluations VALUES(56610,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(56610,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(56610,35,'cp 153',NULL);
INSERT INTO evaluations VALUES(56610,36,'cp 160','g1f3 c8g4 d1b3 g4f3 g2f3 f8d6 f1c4 b8c6 c4f7 e8f8 c3e2 c6a5 b3c3 f8f7 c3a5 g8e7 h1g1 g7g6 d2d3 e7c6 a5b5 d8f6 c1d2 f6f3 b5b7 c6e7 b7b3 f7g7 g1g3 f3h1 e2g1 h1h2 d2e3 h8f8 b3c3 a8b8 e1c1 f8f2 g3h3 h2g2 e3f2 g2f2 c1b1');
INSERT INTO evaluations VALUES(56609,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(56609,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(56609,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(56609,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(56609,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(56609,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(56609,36,'cp 133','b8d7 d2d4 f8d6 c1g5 g8e7 f1e2 e5d4 f3d4 e8g8 d1c2 h7h6 g5h4 d7f6 h4f6 g7f6 e1c1 c7c6 g2g3 d8c7 d1d2 d6c5 h1d1 c8h3 d2d3 f8d8 c2d2 c5d4 d3d4 d8d4 d2d4 g8g7 f2f4 e7g6 d4c5 c7b6 c5b6 a7b6 d1d6 g6f8 a2a4 a8e8 c1d2 e8e7 g3g4 f8g6 d2e3 f6f5 g4f5 h3f5 b2b3 f5d7');
INSERT INTO evaluations VALUES(718,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(718,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(718,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(718,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(718,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(718,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(718,36,'cp 62','f3d4 b6d4 d2f3 d4b6 a2a3 f8e7 c1e3 e7c5 e3d2 e8g8 a1c1 f7f5 c1c3 c5e7 d1c2 b6d8 f1c1 d7b6 f3d4 c8d7 c2b3 f8f7 c3c7 e7f8 d2e3 d7a4 b3c3 a4d7 c7b7 b6a4 c3d2 a4c5 d4c6 d7c6 b7f7 g8f7 e3c5 f8c5 c1c5 d8b6 c5c2 a8c8');
INSERT INTO evaluations VALUES(1135,30,'cp 168',NULL);
INSERT INTO evaluations VALUES(1135,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(1135,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(1135,33,'cp 161',NULL);
INSERT INTO evaluations VALUES(1135,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(1135,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(1135,36,'cp 187','b1c3 c7c6 c4b3 e7b4 d4d3 b4d6 d3e2 d6e5 f2f4 e5h5 e4e5 h5e2 e1e2 f6g8 c3e4 a7a5 a2a4 d7d5 e5d6 c8f5 e2f3 f5e4 f3e4 f8d6 h1e1 f7f6 e4f3 e8d7 c1e3 d7c7 a1d1 g8h6 g2g4 h8e8 e3d4 e8e1 d1e1 b7b6 d4c3 d6c5 h2h3 a8d8 g4g5 f6g5');
INSERT INTO evaluations VALUES(1134,30,'cp 185',NULL);
INSERT INTO evaluations VALUES(1134,31,'cp 199',NULL);
INSERT INTO evaluations VALUES(1134,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(1134,33,'cp 194',NULL);
INSERT INTO evaluations VALUES(1134,34,'cp 188',NULL);
INSERT INTO evaluations VALUES(1134,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(1134,36,'cp 191','c7c6 c4b3 e7b4 d4d3 d7d6 c1f4 c8g4 a2a3 b4a5 e1g1 a8d8 a1e1 f8e7 d3g3 g4e6 b3e6 f7e6 g3g7 h8f8 f4e3 e8d7 f2f3 a5h5 g7g3 d7c8 e3a7 f6d7 a7e3 d7e5 g3g7 e7f6 g7h6 h5h6 e3h6 f8g8 g2g3 e5c4');
INSERT INTO evaluations VALUES(1298,30,'cp 251',NULL);
INSERT INTO evaluations VALUES(1298,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(1298,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(1298,33,'cp 262',NULL);
INSERT INTO evaluations VALUES(1298,34,'cp 251',NULL);
INSERT INTO evaluations VALUES(1298,35,'cp 263',NULL);
INSERT INTO evaluations VALUES(1298,36,'cp 247','c6d4 g7g6 e1g1 f8g7 d4e2 c8b7 f2f3 d6d5 b1c3 g8e7 d2d4 e8g8 d1d3 d5e4 c3e4 f8e8 c2c3 d7b5 e2f4 e7d5 e4c5 b7c6 f4d5 c6d5 b2b3 b5d3 c5d3 d5c6 c1f4 c6b5 d3c5 b5f1 g1f1 e8e7 a1d1 a8e8 c5a6');
INSERT INTO evaluations VALUES(1299,30,'cp 268',NULL);
INSERT INTO evaluations VALUES(1299,31,'cp 264',NULL);
INSERT INTO evaluations VALUES(1299,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(1299,33,'cp 259',NULL);
INSERT INTO evaluations VALUES(1299,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(1299,35,'cp 268',NULL);
INSERT INTO evaluations VALUES(1299,36,'cp 262','g7g6 e1g1 f8g7 d4e2 c8b7 d2d3 d6d5 e4d5 d7d5 e2f4 d5d7 b1d2 g8e7 d2b3 b7c6 b3a5 e8g8 a5c6 d7c6 c2c3 e7f5 d3d4 a6a5 f4d3 f5h4 d1g4 h4f5 f1e1 a8e8 e1e8 f8e8 g4d1 a5a4 c1d2 a4a3 b2a3 e8a8 d3b4 c6e4 d1f1 a8a3 a1e1');
INSERT INTO evaluations VALUES(2123,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(2123,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(2123,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(2123,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(2123,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(2123,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(2123,36,'cp 8','d7c6 d2d4 c5d4 d1d4 g8f6 b1c3 c8g4 d4e5 c7d6 e5d6 e7d6 f3d4 e8c8 f2f3 g4h5 c1g5 d6d5 e4d5 f8b4 e1c1 b4c3 b2c3 d8d5 g5f6 g7f6 d1e1 h8d8 e1e7 d5c5 c1b2 c5e5 h1e1 e5e1 e7e1 c6c5 d4b3 b7b6 a2a4 c8c7 a4a5 d8d5 a5b6 a7b6');
INSERT INTO evaluations VALUES(3200,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3200,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(3200,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3200,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(3200,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3200,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3200,36,'cp 0','f2f4 a8b8 h2h3 f6d7 a2a4 a7a6 c3d5 b7b5 a4b5 a6b5 d5e3 b5b4 g3g4 h7h6 c2c3 c6a5 a1a2 e7e6 f4f5 b4c3 b2c3 a5b3 c1b2 b3a5');
INSERT INTO evaluations VALUES(3216,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3216,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(3216,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3216,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3216,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3216,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3216,36,'cp 0','a8b8 h2h3 e7e6 a2a4 a7a6 c1e3 b7b5 a4b5 a6b5 e4e5 f6d5 e3f2 d5c3 e2c3 b5b4 c3e4 d6e5 e4c5 d8c7 d1f3 c6e7 f4e5 g7e5 d3d4 e5d6 f3f6 e7f5 g3g4 d6e7 f6e5 e7d6');
INSERT INTO evaluations VALUES(4019,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(4019,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(4019,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(4019,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(4019,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(4019,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(4019,36,'cp 45','h7h6 g5d2 e7e5 d4d5 f6h5 e1g1 h5f4 d3c2 d7b6 b2b3 c8g4 h2h3 g4h5 f1e1 c7c6 g1h2 f7f5 e4f5 g6f5 d2f4 e5f4 d1d2 b6d7 f3d4 d8f6 c3e2 h5e2 d4e2 f4f3 g2f3 d7e5 d2f4 c6d5 f4g3 f5f4 g3g2 d5c4 b3c4 g8h8 e2d4 f6h4');
INSERT INTO evaluations VALUES(4619,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(4619,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(4619,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(4619,33,'cp 185',NULL);
INSERT INTO evaluations VALUES(4619,34,'cp 192',NULL);
INSERT INTO evaluations VALUES(4619,35,'cp 204',NULL);
INSERT INTO evaluations VALUES(4619,36,'cp 171','e4e5 f6e8 g5e7 d8e7 c3d5 e7d8 e1c1 d7d6 e5d6 c7d6 a2a3 c8e6 d5f4 a7a6 f1e2 a8c8 d1d2 b7b5 h1e1 e6d7 f4d5 d7e6 e2f3 a6a5 h2h3 e6d5 f3d5 e8f6 d5b7 c8c7 b7f3 b5b4 a3b4 a5b4 d4b4 c7c5 e1e3 d8c7 b4d4');
INSERT INTO evaluations VALUES(5129,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5129,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(5129,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5129,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5129,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5129,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5129,36,'cp 0','g8e7 e2e3 e8g8 d2d4 c5d4 f3d4 d7d5 c4d5 e7d5 a1b1 c6d4 e3d4 b7b6 c3d5 e6d5 c1f4 c8e6 b1c1 d8d7 b2b3 a8c8 d1d2 h7h5 h2h4 e6f5 f1e1 f8e8 e1e8 d7e8 c1c8 e8c8 g2d5 c8d7 d5c4 b6b5 c4e2 d7d4 d2d4 g7d4 g1h1');
INSERT INTO evaluations VALUES(7976,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(7976,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(7976,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(7976,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7976,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7976,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7976,36,'cp 0','e2e3 e8g8 d2d4 c5d4 f3d4 d7d5 c4d5 e7d5 a1b1 c6d4 e3d4 b7b6 c3d5 e6d5 c1f4 c8e6 b1c1 a8c8 d1d2 h7h5 c1c8 d8c8 b2b3 c8d7 f1c1 f8c8 c1c8 d7c8 h2h4 c8f8 g1h2 f8a3 g2f1 a3e7 f4e3 e6g4 f1g2 g4e6');
INSERT INTO evaluations VALUES(7975,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7975,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7975,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7975,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7975,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7975,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(7975,36,'cp 0','e8g8 d2d4 c5d4 f3d4 d7d5 c4d5 e7d5 a1b1 c6d4 e3d4 d8b6 c1g5 h7h6 c3d5 e6d5 g5e7 f8e8 e7c5 b6c6 b1c1 c8e6 c5f8 c6b5 f8g7 g8g7 d1d2 b5a6 a2a3 a6d6 f1e1 b7b6 c1c3 a8c8 c3c8 e8c8 e1c1 c8e8 c1e1');
INSERT INTO evaluations VALUES(7821,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(7821,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7821,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7821,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7821,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7821,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(7821,36,'cp 0','e1g1 e8g8 f1e1 e6e5 d4b3 d7d6 c1d2 c8g4 b5e2 g4e6 e2d3 a8c8 a2a3 b4c3 d2c3 d8c7 d1e2 f8d8 h2h3 a7a6 a3a4 f6d7 e2e3 d7f6');
INSERT INTO evaluations VALUES(8629,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(8629,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(8629,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(8629,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(8629,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(8629,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(8629,36,'cp 12','d6d5 e4d5 f6d5 d3d4 e5d4 f3d4 d5c3 b2c3 e8g8 d4c6 b7c6 e2f3 c8a6 f1e1 e7f6 c1a3 f6c3 a3f8 d8f8 d1d7 c3a1 e1a1 f8c5 a1d1 a6c4 a2a4 c4e6 d7c7 g7g6 c7c6 c5c6 f3c6 a8c8 c6d5 e6d5 d1d5 c8c2 g1f1 c2c7 d5a5 h7h5 f1e2 g8g7 a5a6 c7e7 e2f3');
INSERT INTO evaluations VALUES(12593,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(12593,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(12593,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(12593,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(12593,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(12593,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(12593,36,'cp 37','f1c4 c6d4 e3d3 f8b4 c1g5 b4c3 d3c3 d7d5 g5f6 d8f6 c4d5 c8e6 c3d3 e8g8 g1f3 e6d5 e4d5 f8d8 c2c4 b7b5 c4b5 d8d5 f3d4 d5d4 d3e2 a8d8 e1g1 d4d2 e2c4 d2b2 a2a4 f6b6 a1b1 b2d2 c4c3 g7g6 a4a5 b6d4 c3d4 e5d4 b5b6 a7b6 a5b6 d4d3 b6b7 d8b8 b1b3 d2e2 f1a1 g8f8 g1f1');
INSERT INTO evaluations VALUES(14512,30,'cp -471',NULL);
INSERT INTO evaluations VALUES(14512,31,'cp -478',NULL);
INSERT INTO evaluations VALUES(14512,32,'cp -464',NULL);
INSERT INTO evaluations VALUES(14512,33,'cp -487',NULL);
INSERT INTO evaluations VALUES(14512,34,'cp -475',NULL);
INSERT INTO evaluations VALUES(14512,35,'cp -493',NULL);
INSERT INTO evaluations VALUES(14512,36,'cp -502','e8f7 c4e2 d8d5 e1g1 c8e6 d2d3 d5d7 b1c3 f8e7 e2h5 f7g8 h5f3 a5c6 c3e4 h7h6 c1e3 a8d8 d1e2 g8h7 c2c3 f6e4 d3e4 d7c8 a1d1 h8f8 f3h5 h7h8 b2b4 a7a5 d1d8 c6d8 b4a5 d8c6 h5g6 c6a5 a2a3 a5c6');
INSERT INTO evaluations VALUES(16466,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(16466,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(16466,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(16466,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(16466,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16466,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16466,36,'cp -12','d7d5 c4d5 f6d5 d1c2 f7f5 e2e4 f5e4 f3e5 c6e5 c2e4 f8d6 e4d5 c7c6 d5b3 d6c7 d2d4 d8d4 c1e3 d4g4 f1e2 g4e6 b3b4 a7a5 b4e4 e8g8 f2f4 e5f7 e1c1 f7d6 e4e6 c8e6 g2g3 e6b3 d1e1 a8d8 e2d1 b3d1 e1d1 f8e8 h1e1 g8f7 c1c2');
INSERT INTO evaluations VALUES(16702,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(16702,31,'cp -136',NULL);
INSERT INTO evaluations VALUES(16702,32,'cp -137',NULL);
INSERT INTO evaluations VALUES(16702,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(16702,34,'cp -149',NULL);
INSERT INTO evaluations VALUES(16702,35,'cp -116',NULL);
INSERT INTO evaluations VALUES(16702,36,'cp -141','c3c4 d7d5 e4d5 f6d5 g1f3 f8e7 f1e2 c8e6 c4a4 d5b6 a4f4 e7d6 f4e4 d8d7 c1e3 e8c8 e3b6 e6f5 e4a4 c7b6 b1c3 c6b4 a4d7 d8d7 e1g1 f5c2 a1c1 c2g6 c1d1 h8d8 a2a3 b4c6 f3h4 d6e7 d1d7 d8d7 h4g6 h7g6 f1d1 c6d4 e2d3 e7f6 c3e4 c8c7 g1f1 d4c6 e4f6 g7f6');
INSERT INTO evaluations VALUES(19508,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(19508,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(19508,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(19508,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(19508,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(19508,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(19508,36,'cp -37','e7e6 b1c3 g8f6 d2d3 d7d5 c4b5 f8e7 g1f3 e8g8 b5c6 b7c6 e1g1 f6d7 f1e1 f7f6 a4a5 f8e8 b2b3 e7d6 a5a6 a8b8 c1a3 d6f8 d1e2 d5d4 c3a4 e6e5 f3d2 c8a6 a3c1 d8a5 d2f1 a6b5 c1d2 a5c7 a4b2 a7a6 b2c4 g7g6 d2a5 c7b7');
INSERT INTO evaluations VALUES(18862,30,'cp -22',NULL);
INSERT INTO evaluations VALUES(18862,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(18862,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(18862,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(18862,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(18862,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(18862,36,'cp -20','b1c3 f8e7 g1f3 g8f6 c4b5 e8g8 e1g1 c6d4 e4e5 d4f3 d1f3 f6e8 f3d1 a7a6 b5e2 d7d6 d2d4 c5d4 d1d4 d6e5 d4e5 e7d6 e5h5 f7f5 f1d1 c8d7 e2c4 d8c7 b2b3 a8d8 c1a3 d6a3 a1a3 d8c8');
INSERT INTO evaluations VALUES(18860,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(18860,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(18860,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(18860,33,'cp -19',NULL);
INSERT INTO evaluations VALUES(18860,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(18860,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(18860,36,'cp -31','g8f6 d2d3 d7d5 c4b5 f8e7 g1f3 e8g8 b5c6 b7c6 e1g1 f6d7 f1e1 f7f6 a4a5 e7d6 b2b3 d8c7 h2h3 d7e5 c3a4 a8b8 a5a6 b8b5 f3e5 d6e5 c1a3 f8d8 a4c5 e5c3 e1e2 c3a1 d1a1 d5d4 e4e5 f6e5 c5e4 c7f7');
INSERT INTO evaluations VALUES(12446,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(12446,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(12446,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(12446,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(12446,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(12446,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(12446,36,'cp -12','d2d3 a7a6 g1f3 d7d5 c4b3 d5e4 c3e4 f8e7 e4f6 e7f6 c2c3 b7b6 c1f4 c8b7 e1g1 e8g8 d1e2 c6e7 a1d1 a8a7 f3e5 e7g6 e5g6 h7g6 f4e3 b7d5 b3d5 d8d5 f1e1 a7d7 d3d4 c5c4 e3f4 f8d8 f4e5 b6b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(20161,30,'cp -125',NULL);
INSERT INTO evaluations VALUES(20161,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(20161,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(20161,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(20161,34,'cp -129',NULL);
INSERT INTO evaluations VALUES(20161,35,'cp -123',NULL);
INSERT INTO evaluations VALUES(20161,36,'cp -135','a7a6 b5c6 b7c6 g1f3 g8h6 c2c4 h6f5 c4d5 e6d5 d1c2 f5d4 f3d4 c6c5 d4f3 f8e7 e1g1 e8g8 b1c3 d5d4 c3a4 c8b7 b2b3 f8e8 f1e1 e7f8 c2f5 h7h6 a1c1 a8c8 a4b2 d8d5 b2d3 c5c4 c1c4 c8c4 b3c4 d5c4');
INSERT INTO evaluations VALUES(21346,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(21346,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(21346,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(21346,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(21346,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(21346,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(21346,36,'cp 34','d5e4 c3e4 g8f6 e4f6 g7f6 d2d4 d8d5 d1e2 h8g8 c2c4 d5h5 e1g1 h5g4 g2g3 a7a6 b5a4 c8d7 d4d5 c6e5 a4d7 e5d7 g1g2 e8c8 h2h3 g4f5 f3d4 f5g6 d5e6 d7e5 e6f7 g6f7 e2e4 f7c4 c1e3 c4d3 e4d3 e5d3 a1b1 d8d7 f1d1 d3b4 a2a3 b4d5 e3d2 f8c5 d4e6 c5d6 d1e1 f6f5 b1c1 c7c6 c1d1');
INSERT INTO evaluations VALUES(21448,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21448,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21448,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(21448,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21448,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21448,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(21448,36,'cp 0','d5e4 d3e4 d8d1 c3d1 g8f6 d1c3 c8d7 a2a3 f8c5 h2h3 e8c8 f1d3 d7e8 b2b4 c5e7 c3e2 f6d7 c1b2 d7b6 b4b5 c6a5 b2g7 h8g8 g7c3 a5b3 c2b3 d8d3 f3e5 d3d8 g2g3 h7h5 a3a4 f7f6 a4a5 f6e5 a5b6 c7b6 a1a7 e8b5 a7a8 c8d7 a8d8 d7d8 c3e5');
INSERT INTO evaluations VALUES(22326,30,'cp 331',NULL);
INSERT INTO evaluations VALUES(22326,31,'cp 309',NULL);
INSERT INTO evaluations VALUES(22326,32,'cp 326',NULL);
INSERT INTO evaluations VALUES(22326,33,'cp 327',NULL);
INSERT INTO evaluations VALUES(22326,34,'cp 316',NULL);
INSERT INTO evaluations VALUES(22326,35,'cp 341',NULL);
INSERT INTO evaluations VALUES(22326,36,'cp 336','f1b5 g8f6 e4f5 c8f5 d1e2 d8e7 d4d5 a7a6 b5a4 b7b5 c3b5 a6b5 a4b5 f5d7 d5c6 d7g4 h2h3 g4h5 g2g4 h5f7 e1g1 h7h6 f1e1 g7g6 a2a4 f8g7 a4a5 e8g8 a1a3 e5e4 g4g5 h6g5 f3g5 d6d5 c2c4 d5d4 g5e4 f6e4 e2e4 e7e4 e1e4');
INSERT INTO evaluations VALUES(24053,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(24053,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(24053,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(24053,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(24053,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(24053,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(24053,36,'cp 3','h2h3 g4h5 a2a4 d8d7 c1e3 c6b4 d3d4 e5d4 e3d4 e8g8 c3d5 b4d5 e4d5 c7c5 d5c6 b7c6 f1e1 d6d5 c4e2 f8e8 f3e5 h5e2 e5d7 e2d1 d7f6 e7f6 e1d1 f6d4 d1d4 a8b8 b2b4 e8e2 b4b5 e2c2 b5c6 c2c6 d4d5 g7g6 d5d7 b8b2 d7a7 g8g7 a7a8');
INSERT INTO evaluations VALUES(24052,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(24052,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(24052,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24052,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(24052,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(24052,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(24052,36,'cp 4','g4h5 a2a4 d8d7 c1e3 c6b4 d3d4 e5d4 e3d4 e8g8 c3d5 b4d5 e4d5 c7c5 d5c6 b7c6 f1e1 d6d5 c4e2 f8e8 f3e5 h5e2 e5d7 e2d1 d7f6 e7f6 e1d1 f6d4 d1d4 a8b8 b2b4 e8e2 b4b5 e2c2 b5c6 c2c6 d4d5 g7g6 d5d7 a7a6');
INSERT INTO evaluations VALUES(24051,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(24051,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24051,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(24051,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(24051,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(24051,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(24051,36,'cp 8','c3d5 c6a5 c4b5 c7c6 d5e7 d8e7 b2b4 c6b5 b4a5 e8g8 a1b1 a7a6 a2a4 b5a4 c1a3 a8c8 c2c4 c8c6 b1b4 h7h6 g1h2 h5f3 d1f3 f8b8 b4a4 b7b5 a5b6 b8b6 f1a1 f6d7 g2g3 d7c5 a3c5 d6c5 h2g2 e7b7 a1a2 b6b3 a4a5 b7d7 a2a3 b3a3 a5a3 c6b6 h3h4');
INSERT INTO evaluations VALUES(24989,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(24989,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(24989,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(24989,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(24989,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(24989,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(24989,36,'cp 63','e7e6 b1c3 f8d6 h2h3 g4h5 f1e1 e8g8 g2g4 h5g6 e3e4 e6e5 d4d5 c6e7 c1g5 f6d7 c4f1 h7h6 g5e3 d6b4 f3d2 e7c8 a2a3 b4c5 e3c5 d7c5 b2b4 c5d7 a1c1 c8d6 d2b3 b7b6 c3b5 a7a5 b5c7 a8c8 c7b5 c8c1 d1c1 a5b4');
INSERT INTO evaluations VALUES(25238,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(25238,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(25238,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(25238,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(25238,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(25238,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(25238,36,'cp 64','h2h3 g4h5 f1e1 f8d6 b1c3 e8g8 g2g4 h5g6 e3e4 e6e5 d4d5 c6e7 c1g5 f6d7 a2a3 h7h6 g5e3 a7a6 c4f1 d7f6 d1d3 g8h8 f3d2 c7c6 d5c6 e7c6 a1d1 d6c7 d3c4 a8c8 d2f3 c7d6 c4d3 d6c7 f3h4 g6h7 h4f5 h7f5 e4f5');
INSERT INTO evaluations VALUES(25704,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(25704,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(25704,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(25704,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(25704,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(25704,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(25704,36,'cp 73','b5e2 g8e7 c2c3 f7f6 e5f6 g7f6 e1g1 e7g6 c3c4 d8e7 b1c3 e8c8 a1b1 d5c4 b2b3 e6e5 b3c4 d7f5 c3d5 e7f7 b1b2 e5e4 d1b3 c6a5 b3c2 d8d5 c4d5 f7d5 f3e1 d5d4 c1e3 d4e5 c2d2 a5c6 f2f4 e5a5 d2a5 c6a5 e1c2 f8d6 g2g3 h8g8 e3d2 a5c6');
INSERT INTO evaluations VALUES(26174,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(26174,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(26174,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(26174,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(26174,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(26174,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(26174,36,'cp 39','c2c3 g8f6 d1e2 f8h6 d2d4 h6c1 f1c1 e8g8 e4e5 f6h5 g2g3 h5g7 d4c5 d7c5 c1d1 d8c7 e5d6 e7d6 c3c4 f8e8 e2c2 c5e4 b1c3 e4c3 c2c3 a8c8 a1c1 g7f5 b2b3 e8e2 d1d2 c8e8 c3d3 e2d2 d3d2 c7c6 d2d3 b7b5 g3g4 f5g7 f3d4 b5c4 b3c4 c6e4 d3e4 e8e4');
INSERT INTO evaluations VALUES(26173,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(26173,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(26173,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(26173,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(26173,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(26173,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(26173,36,'cp 38','g8f6 d1e2 f8h6 d2d4 h6c1 f1c1 e8g8 g2g3 c5d4 c3d4 a8c8 b1c3 d8a5 a2a3 c8c6 c1e1 g8g7 h2h4 e7e5 a1d1 f8c8 g1g2 a7a6 d1d3 b7b5 e1d1 b5b4 a3b4 a5b4 d1d2 c8e8 e2e3 h7h6 d4e5 d7e5 d3d4 b4c5 d2d1 e5f3 e3f3 c6b6 d1d2 e8e6');
INSERT INTO evaluations VALUES(26863,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(26863,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(26863,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(26863,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(26863,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(26863,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(26863,36,'cp 33','h2h3 g4f3 d1f3 c6d4 f3d3 d4e6 c3d5 f8c5 d5f4 d8d4 d3d4 e6d4 f1d3 g8f6 c1d2 d4c6 f2f3 e8g8 e1c1 c6e5 h1e1 e5d3 f4d3 c5d6 c2c4 d6g3 e1e3 f8e8 c1c2 h7h5 a2a4 h5h4 a4a5 a8d8 b2b4 c7c6 e3e8 d8e8 d2c3 f6d7 c2b3 f7f6 d1d2 e8d8 b4b5 c6b5 c4b5 g3d6');
INSERT INTO evaluations VALUES(28762,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(28762,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(28762,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(28762,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(28762,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(28762,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(28762,36,'cp 70','d7d6 e1g1 b8d7 c2c3 f8e7 f1e1 e8g8 d2f1 c7c5 f1g3 a7a6 d1e2 b6b5 e4e5 f6d5 e2e4 g7g6 c1h6 c5d4 e5d6 e7d6 e4d4 e6e5 f3e5 d6e5 e1e5 d7e5 h6f8 f7f6 f8c5 d5c3 d3c2 d8d4 c5d4 c3a4 c2e4 b7e4 g3e4 e5d7 a1e1 g8f7 h2h3 a8d8 e4d6 f7g8 e1e7 d7c5');
INSERT INTO evaluations VALUES(29588,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29588,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29588,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29588,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29588,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29588,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29588,36,'cp 0','c2c4 c7c6 f3h4 f5e4 e2e3 e7e6 a2a3 f8e7 h4f3 e4f5 f3h4');
INSERT INTO evaluations VALUES(30743,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(30743,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(30743,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(30743,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(30743,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(30743,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(30743,36,'cp 32','d7d6 c2c4 g8f6 b1c3 f8g7 d1c2 b8d7 f1e2 f6h5 f4g5 h7h6 g5h4 g6g5 h4g3 a7a6 d4d5 h5g3 h2g3 c7c6 a1d1 c6d5 c4d5 a8c8 f3d4 b6b5 c2b3 d8a5 a2a3 g7d4 d1d4 d7f6 e1g1 e8f8 a3a4 f8g7 a4b5 c8c5 e3e4 a6b5 e2b5');
INSERT INTO evaluations VALUES(30831,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(30831,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(30831,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(30831,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(30831,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(30831,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(30831,36,'cp -8','e7e6 d1e2 d7d5 c1d2 b8c6 g1f3 f8c5 e4d5 d8d5 b1c3 d5d7 c3e4 c6d4 f3d4 c5d4 c2c3 d4f6 e1c1 f6e7 h1g1 g8f6 e4f6 e7f6 g2g4 e8c8 d3d4 h7h5 g4g5 f6e7 g5g6 f7f5 f1g2 b7g2 g1g2 e7f6 c1b1 c8b8 d1e1 d7c6 g2g3');
INSERT INTO evaluations VALUES(30702,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(30702,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(30702,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30702,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30702,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(30702,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30702,36,'cp 0','d1e2 d7d5 c1d2 b8c6 g1f3 f8c5 e4d5 d8d5 b1c3 d5d7 c3e4 c6d4 e4c5 b6c5 f3d4 c5d4 e2e5 f7f6 e5c5 g8e7 f1e2 h7h5 e1g1 e7f5 a2a4 f5h4 f1f2 h4f5');
INSERT INTO evaluations VALUES(32494,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32494,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32494,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32494,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(32494,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(32494,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32494,36,'cp 10','b5e2 g8f6 g1f3 c7c5 e1g1 d8c7 d4c5 f8c5 b2b3 e8g8 c1b2 f8d8 b1d2 d7e8 c2c4 c7e7 d1c2 b8d7 a1c1 a8c8 c2b1 h7h6 c4d5 f6d5 d2e4 c5a3 c1c8 d8c8 f1c1 c8c1 b2c1 a3c1 b1c1 d7f6 e4f6 d5f6 f3d4 g8h8 d4f3');
INSERT INTO evaluations VALUES(34360,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(34360,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(34360,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(34360,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(34360,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(34360,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(34360,36,'cp 86','g1e2 g8f6 b1c3 d5d8 c1g5 e7e6 d4d5 f8e7 d5e6 d8d1 a1d1 c8e6 c3b5 b8a6 e2f4 e6a2 b5c3 a2e6 g5f6 g7f6 c3d5 e8f8 f1c4 a6b4 f4e6 f7e6 d5c7 b7b5 c7a8 b5c4 d1d7 f8f7 d7a7 h8b8 e1d2 b4c6 a7a4 b8b2 a4c4 e7d6 h1d1 c6d4');
INSERT INTO evaluations VALUES(35236,30,'cp 243',NULL);
INSERT INTO evaluations VALUES(35236,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(35236,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(35236,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(35236,34,'cp 267',NULL);
INSERT INTO evaluations VALUES(35236,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(35236,36,'cp 238','c1e3 d7d6 d1d2 c8e6 b1c3 e7d8 e1c1 b8d7 d4d5 e6f7 f3h4 g8e7 g2g3 g7g6 f2f4 c7c6 d5c6 e7c6 h4f3 d8a5 f4e5 d7e5 f3e5 f6e5 a2a3 f8e7 f1h3 e8g8 c1b1 c6d4 e3d4 e5d4 c3d5 e7d8 d2d4 f7d5 d4d5 a5d5 d1d5 f8f2 d5d6 d8f6 h1d1');
INSERT INTO evaluations VALUES(38264,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(38264,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(38264,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(38264,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38264,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(38264,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(38264,36,'cp 17','d6d5 e4d5 e6d5 c4d3 g8f6 e1g1 e8g8 f1e1 c7c5 d4c5 e7c5 b1c3 c8e6 a2a3 a7a6 h2h3 f8e8 b2b4 c5a7 c1e3 a7e3 e1e3 b8c6 c3e2 d8c7 d1d2 e6d7 e3e8 a8e8 a1e1 h7h6 e2c3 c6e5 f3e5 e8e5 a3a4 e5e1 d2e1 c7c6 a4a5 c6c7');
INSERT INTO evaluations VALUES(39027,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(39027,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(39027,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(39027,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(39027,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(39027,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(39027,36,'cp 49','b1c3 d7d5 e3e4 d5e4 c3e4 b8d7 f1b5 a7a6 b5d3 g8f6 c1g5 e8g8 e1g1 h7h6 g5h4 b7b5 c2c3 c8b7 a2a4 b5b4 h4f6 d7f6 e4c5 b7f3 d1f3 f6d5 d3e4 c7c6 a4a5 b4c3 b2c3 e6e5 f1d1 e5d4 c3d4 f8e8 g2g3 d5b4 c5b3 d8e7 e4c6 b4c6 f3c6 a8b8');
INSERT INTO evaluations VALUES(39961,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(39961,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(39961,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(39961,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(39961,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(39961,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(39961,36,'cp 57','e2e4 c6c5 d2d3 d7d6 f2f4 a7a6 g1f3 b8c6 e1g1 b7b5 b2b3 a8b8 c3d5 b5c4 b3c4 c8g4 d5e3 e8g8 c1d2 g4f3 g2f3 f6d7 a1b1 b8b1 d1b1 d8a8 b1e1 f8e8 e3f5 c6d4 h2h4 d4f3 f1f3 d7f6 g1g2 a8b7');
INSERT INTO evaluations VALUES(42648,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(42648,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(42648,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(42648,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42648,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(42648,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(42648,36,'cp -9','g1f3 d7d6 e1g1 e8g8 c2c3 a7a5 d3d4 c5a7 d4e5 d6e5 b1d2 d8e7 d2c4 b8d7 d1c2 a7c5 c1e3 c5e3 c4e3 b7b6 e3f5 e7c5 a1d1 c8b7 e2d3 g7g6 f5e3 f8d8 f1e1 g8g7 h2h3 b6b5 a2a4 b5a4 c2a4 b7c6 a4c2 a5a4 d3f1');
INSERT INTO evaluations VALUES(43102,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(43102,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(43102,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43102,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(43102,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(43102,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(43102,36,'cp 49','c2c4 d7d6 b1c3 e8g8 c4c5 g8h8 d1b3 d6c5 d4c5 c7c6 f3e5 d8a5 e1c1 b8d7 e5f7 f8f7 b3f7 d7c5 f1c4 c8d7 f7e7 a8e8 e7d6 f6g4 d1d4 g7f8 d6c7 a5c7 f4c7 b7b5 b2b4 f8g7 b4c5 g7d4 e3d4 b5c4 c7g3 h8g8 h1d1 f5f4 g3f4 g4f2 d1f1 f2d3 c1d2 e8f8');
INSERT INTO evaluations VALUES(43799,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(43799,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(43799,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(43799,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(43799,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(43799,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(43799,36,'cp 43','c8b7 c4d5 f8b4 b1c3 b7d5 e1g1 b4c3 b2c3 e8g8 c1b2 f8e8 d1c1 b6b5 a2a4 c7c6 c3c4 d5c4 f3d2 c4e2 f1e1 e2d3 e1e3 d3c4 d2c4 b5c4 c1c4 f6d5 e3b3 b8d7 c4c6 d7f6 a4a5 a8b8 c6a4 d8d7 a4d7 f6d7');
INSERT INTO evaluations VALUES(45665,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(45665,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(45665,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(45665,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(45665,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(45665,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(45665,36,'cp 37','g1f3 f8d6 f4d6 d8d6 f3e5 c6c5 b1d2 b8c6 f2f4 c6e7 f1d3 e8g8 d1e2 f6e8 e1g1 a7a5 g2g4 f7f6 e5f3 c5c4 d3c2 b7b5 a2a3 a8b8 g1h1 b5b4 a3b4 a5b4 g4g5 b4b3 c2b1 g8h8 e2g2 e7f5 f1g1 g7g6 g5f6 e8f6 f3e5');
INSERT INTO evaluations VALUES(46829,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(46829,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(46829,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(46829,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(46829,34,'cp 215',NULL);
INSERT INTO evaluations VALUES(46829,35,'cp 215',NULL);
INSERT INTO evaluations VALUES(46829,36,'cp 187','f8g7 f1c4 e8g8 e4e5 f6d7 e5e6 d7b6 e6f7 g8h8 c4b3 c8g4 h2h4 b8c6 c3e2 d8d6 c1f4 d6b4 f4d2 b4d6 f3g5 g4e2 d1e2 c6d4 e2e4 d4b3 a2b3 a8d8 e1c1 d6c5 c1b1 d8d4 d2e3 d4e4 e3c5 e4e2 h1e1 e2e1 d1e1 h7h6 e1e7 h6g5 e7e8 b6d7 c5f8 g7f8 h4g5 h8g7');
INSERT INTO evaluations VALUES(41545,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(41545,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(41545,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(41545,33,'cp 201',NULL);
INSERT INTO evaluations VALUES(41545,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(41545,35,'cp 197',NULL);
INSERT INTO evaluations VALUES(41545,36,'cp 218','f1c4 e8g8 e4e5 f6d7 e5e6 d7b6 e6f7 g8h8 c4b3 c8g4 h2h4 b8c6 c3e2 d8d7 h4h5 g6h5 f3g5 c6d4 f2f3 g4f5 g2g4 d4f3 g5f3 f5g4 d1d7 b6d7 f3g5 d7c5 b3c2 a8d8 g5h7 f8f7 h7g5 f7f6 e2f4 h8g8 c1e3 f6f4 e3f4 c5d3 c2d3 d8d3 e1g1 g7d4 g1h2 e7e5 f4g3 d3d2 h2h1 g4d7 g5f3 d7c6 a1e1 g8f7 g3e5 d4e5 e1e5');
INSERT INTO evaluations VALUES(47518,30,'cp 373',NULL);
INSERT INTO evaluations VALUES(47518,31,'cp 362',NULL);
INSERT INTO evaluations VALUES(47518,32,'cp 392',NULL);
INSERT INTO evaluations VALUES(47518,33,'cp 400',NULL);
INSERT INTO evaluations VALUES(47518,34,'cp 406',NULL);
INSERT INTO evaluations VALUES(47518,35,'cp 400',NULL);
INSERT INTO evaluations VALUES(47518,36,'cp 396','f2f3 e4c5 d4c5 d8d1 e1d1 c8f5 b1d2 b8c6 d2c4 e8c8 d1e1 c6d4 c4e3 e7e6 g1e2 f8c5 e2d4 c5d4 e3f5 e6f5 f1c4 d4e5 c4f7 d8d6 e1e2 h8f8 f7b3 f8e8 e2f2 e5b2 c1b2 d6d2 f2g3 d2b2 a1e1 b2e2 e1e2 e8e2 h2h4 b7b5 h1c1 f5f4 g3h3 c8d7 c1c5 e2e1 c5f5 c7c6 h3g4 e1e2 f5f7 d7d6 f7a7');
INSERT INTO evaluations VALUES(47706,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(47706,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(47706,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(47706,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(47706,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(47706,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(47706,36,'cp 130','g5f6 g7f6 e2e3 h8g8 g2g3 c7c6 g1f3 b8d7 f3h4 d7b6 f1g2 f6f5 d1c2 d8f6 e1g1 f8d6 c3e2 c8e6 a2a4 a7a5 c2b3 d6b4 e2f4 b6c4 h4f3 f6e7 b3c2 b4d6 b2b3 c4b6 g1h1 d6f4 e3f4 b6c8 f1b1 c8d6 b3b4 a5b4 b1b4 d6c4 b4b3 e8f8 b3b1');
INSERT INTO evaluations VALUES(47703,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(47703,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(47703,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(47703,33,'cp 102',NULL);
INSERT INTO evaluations VALUES(47703,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(47703,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(47703,36,'cp 109','g7f6 e2e3 h8g8 g1e2 c7c6 e2g3 c8e6 a1c1 b8d7 f1d3 d8c7 e1g1 e8c8 c3e2 c8b8 g1h1 d7b6 e2f4 f8d6 f4h5 b6c4 d1c2 d6g3 h5g3 c4d6 f2f3 c7e7 f1e1 g8g5 g3e2 h6h5 e2f4 h5h4 b2b3 e6f5 b3b4 f5d3 c2d3');
INSERT INTO evaluations VALUES(47747,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(47747,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(47747,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(47747,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(47747,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(47747,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(47747,36,'cp -64','b8c6 d4d2 c8e6 g1f3 f8d6 b1c3 d8e7 h2h3 g4e5 f3e5 d6e5 e2e4 f7f5 f1e2 c6d4 c3d5 e6d5 e4d5 e8c8 e1f1 h8e8 d2d3 d4e2 d3e2 e7f6 h3h4 h7h6 e2d3 e5b2 a1b1 b2c1 b1c1 e8e4 g2g3 e4d4 d3f3 f5f4 f1g2 d8e8 c1e1');
INSERT INTO evaluations VALUES(53128,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(53128,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(53128,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(53128,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(53128,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(53128,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(53128,36,'cp 25','d1d4 b8c6 d4d1 g7g6 f1d3 f8g7 g1e2 c8g4 e1g1 a8c8 h2h3 g4e2 d3e2 g8f6 b1c3 f6d7 c1d2 e8g8 f1e1 e7e6 d2f4 d8a5 a1b1 g7c3 b2c3 a5c3 b1b3 c3d4 d1d4 c6d4 b3b7 d7e5 f2f3 d4e2 e1e2 e5c4 b7a7 c4e5 f4h6 f8d8 g1h1 d8d7 a7d7 e5d7');
INSERT INTO evaluations VALUES(53126,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(53126,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(53126,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(53126,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(53126,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(53126,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(53126,36,'cp 15','b8c6 d4d1 g8f6 f1d3 e7e6 g1f3 d6d5 e4d5 f8b4 b1c3 e6d5 e1g1 b4c3 b2c3 e8g8 h2h3 d5c4 d3c4 c8f5 c1e3 d8c7 f1e1 f8d8 d1c1 d8e8 c4b3 a8d8 f3d4 f5g6 e3g5 e8e1 c1e1 a7a6 g5f6 g7f6 e1e3 c6e5');
INSERT INTO evaluations VALUES(20068,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(20068,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(20068,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(20068,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(20068,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(20068,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(20068,36,'cp 29','d4d1 g7g6 f1d3 f8g7 g1e2 g8f6 b1c3 e8g8 h2h3 e7e6 e1g1 d6d5 e4d5 e6d5 c1g5 c6b4 c4d5 b4d5 d3c4 h7h6 g5f6 d5f6 d1b3 a7a6 f1d1 d8c7 c3d5 f6d5 c4d5 c7e5 d1d2 a8b8 e2d4 e5f6 a1d1 b7b5 d4c2 f6e5 c2e3 c8e6 d5e6 e5e6 b3e6 f7e6 d2d6');
INSERT INTO evaluations VALUES(57500,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(57500,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(57500,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(57500,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(57500,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(57500,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(57500,36,'cp 50','b2b3 d7d6 c1b2 g8f6 f1e2 g7g6 d2d4 b8d7 g1h3 f8g7 e1g1 e8g8 d1c2 c7c6 b1c3 f6e4 d4e5 e4c3 c2c3 g7e5 c3d2 d7c5 a1d1 d8e7 h3f4 a7a5 d2c1 c8e6 e2f3 e6f7 h2h3 f8e8 b2d4');
INSERT INTO evaluations VALUES(117,30,'cp 166',NULL);
INSERT INTO evaluations VALUES(117,31,'cp 177',NULL);
INSERT INTO evaluations VALUES(117,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(117,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(117,34,'cp 168',NULL);
INSERT INTO evaluations VALUES(117,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(117,36,'cp 171','e7d6 f1c4 g8e7 c1d2 d6c7 e1c1 a7a6 a2a4 b7b5 a4b5 a6b5 c4b5 a8a1 c3b1 e8g8 d2e3 d7d6 b2b3 c8e6 c1b2 a1a5 b1a3 f8b8 c2c4 b8a8 d1a1 d6d5 f2f4 d5d4 e3d2 a5a7 a3c2 a7a1 h1a1 a8a1 b2a1 e6c8 c2e1 c8b7 e4e5 h7h5 b5d7 g8f8 g2g3 f7f6 e1d3 f6e5 d3e5');
INSERT INTO evaluations VALUES(2595,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2595,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2595,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2595,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2595,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2595,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2595,36,'cp 0','c5e3 f2e3 c8e6 c4e6 f7e6 c3e2 d8d7 c2c4 c6e7 d1c2 e7g6 b2b4 b7b6 c4c5 d6c5 b4c5 b6c5 a1c1 d7d6 e2g3 a8d8 f1d1 f6d7 g3f1 d6e7 c2f2 e7d6');
INSERT INTO evaluations VALUES(2601,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2601,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2601,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2601,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2601,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2601,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2601,36,'cp 0','f2e3 c8e6 c4e6 f7e6 c3e2 c6e7 e2g3 d8e8 d1d2 e7g6 a3a4 a6a5 d2c3 f8f7 b2b3 b7b6 f1f2 e8e7 g3e2 a8f8 a1f1 f6h5 f3h2 h5f6 c3c6 e7d8 h2f3 d8e7 c6c3');
INSERT INTO evaluations VALUES(2600,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2600,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2600,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2600,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2600,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2600,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2600,36,'cp 0','c8e6 c4e6 f7e6 c3e2 c6e7 e2g3 d8e8 d1d2 e7g6 a3a4 a6a5 d2c3 f8f7 b2b3 e8e7 f1f2 b7b6 a1f1 a8f8 c3c6 f8d8 c6c3');
INSERT INTO evaluations VALUES(5638,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5638,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5638,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5638,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(5638,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(5638,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5638,36,'cp 11','g8f6 b1c3 c6a5 c4d3 c7c5 a2a3 a5c6 c3d5 f6d5 e4d5 c6d4 f3d4 c5d4 c2c3 e8g8 c3d4 e5d4 d3e4 f7f5 e4d3 e7f6 b2b3 c8d7 c1b2 a8c8 a3a4 g7g5 d1h5 d8e8 h5h6 e8g6 h6g6 h7g6 b2a3 f6e5 a1c1 a7a6 a4a5 f8e8 f1e1 c8c1 e1c1');
INSERT INTO evaluations VALUES(8600,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(8600,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(8600,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8600,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(8600,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(8600,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(8600,36,'cp 1','b1c3 c6a5 c4b5 c7c6 b5d3 c6c5 a2a3 a5c6 d3c4 f6e4 c3e4 d6d5 f3e5 c6e5 c4b5 e5c6 b5c6 b7c6 e4g3 e8g8 d2d3 f8e8 d1f3 e7f8 c1d2 c8e6 f1e1 a7a5 g3h5 a5a4 f3g3 g8h8 e1e2 a8b8 a1b1 c5c4 d3c4 d5c4');
INSERT INTO evaluations VALUES(7081,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(7081,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(7081,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(7081,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(7081,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7081,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(7081,36,'cp 28','d4c6 b7c6 d1e2 d6d5 c4d3 e8g8 a1d1 f8e8 c3a4 h7h6 g5h4 c8b7 c2c4 f6d7 h4e7 d8e7 d3c2 a8d8 e4e5 a6a5 b2b3 e7b4 e2d3 f7f5 e5f6 d7f6 d3e3 e6e5 c2g6 e8e7 f2f3 f6d7 e3d2 b7a6 c4d5 b4d2 d1d2 c6d5 f1e1 d7f8 g6b1 a6b7 a4c5');
INSERT INTO evaluations VALUES(8558,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(8558,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(8558,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(8558,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8558,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(8558,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(8558,36,'cp 20','a7a6 b5c6 b7c6 b1c3 c6c5 f3h2 c8b7 f2f4 e5f4 c1f4 e8g8 f1f2 f6d7 d1g4 d7e5 f4e5 d6e5 g1h1 e7g5 g4e2 g5f4 h2f1 d8g5 b2b3 g5d8 a1d1 g7g6 f1e3 f7f5 f2f1 f4e3 e2e3 d8d4 e3e1 g8g7 c3b1 f5e4 f1f8 a8f8 d3e4');
INSERT INTO evaluations VALUES(10899,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10899,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10899,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10899,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10899,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10899,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10899,36,'cp 0','g1f3 e8g8 e1g1 d7d5 e4d5 f6d5 a2a4 a7a5 c4b5 f7f6 b1d2 d5b6 d2c4 b6d5 c4d2');
INSERT INTO evaluations VALUES(11753,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(11753,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(11753,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(11753,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(11753,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(11753,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(11753,36,'cp 102','e7g6 d2d4 c8b7 b3d5 a8b8 c1e3 f8d6 d4e5 c6e5 f3e5 g6e5 d5b7 b8b7 f2f4 e5c4 e3d4 e8g8 b2b3 c4a5 d1f3 f7f6 b1d2 d8e7 f1f2 a5c6 d2f1 b7b8 f1g3 e7e6 d4e3 c6e7 a1c1 b8d8 g1h1 e7g6 g3f5 d8e8');
INSERT INTO evaluations VALUES(11689,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(11689,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(11689,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(11689,33,'cp 110',NULL);
INSERT INTO evaluations VALUES(11689,34,'cp 124',NULL);
INSERT INTO evaluations VALUES(11689,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(11689,36,'cp 114','a2a4 b5a4 a1a4 f8e7 d2d4 e8g8 d4d5 c6b8 b1a3 d7d6 a3c4 b8d7 a4a1 d7b6 c4a5 g8h8 b3c2 f7f5 c3c4 d8e8 b2b4 f5e4 c2e4 c8f5 e4f5 f8f5 c1e3 g6f4 f3d2 e8g6 e3f4 f5f4 f2f3 e7g5 d2e4 f4h4 c4c5 d6c5 b4c5 g5f4');
INSERT INTO evaluations VALUES(12830,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(12830,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(12830,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(12830,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(12830,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(12830,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(12830,36,'cp -16','a7a6 d2d4 c5d4 e4e5 f6d5 f1c4 e7e6 c4d5 e6d5 e1g1 d7d6 c3d4 c8g4 c2d3 f8e7 b1c3 e8g8 c3d5 g4f3 d5e7 d8e7 e5d6 e7d6 d3f3 c6d4 f3d1 a8d8 c1e3 d6e5 e3d4 d8d4 d1f3 b7b5 f1e1 e5b8 a2a3 d4d2 e1e2 d2e2 f3e2');
INSERT INTO evaluations VALUES(12832,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(12832,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(12832,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(12832,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(12832,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(12832,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(12832,36,'cp -5','e7e5 e1g1 d7d5 d3b5 d5e4 f3e5 d8c7 d2d4 f8d6 b1a3 c5d4 c3d4 e8g8 b5c6 b7c6 a3c4 c8e6 c4d6 c7d6 c1e3 f6d5 a1c1 a8c8 c1c5 f7f6 e5c4 d6d7 d1d2 h7h6 f1c1 d5e7 b2b3 e7f5 d2c3 e6d5 e3d2 c8e8 c4e3 f5e3 d2e3 f6f5 e3f4 f8f6');
INSERT INTO evaluations VALUES(12563,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(12563,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(12563,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(12563,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(12563,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(12563,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(12563,36,'cp -1','e1g1 d7d5 d3b5 f8d6 e4d5 f6d5 d2d3 e8g8 f1e1 d8c7 b1d2 f8d8 d2c4 f7f6 d3d4 c5d4 c3d4 d6b4 c1d2 e5d4 d1b3 b4d2 c4d2 c8g4 b5c6 b7c6 f3d4 c6c5 d4b5 c7c6 b3c4 g4h5 d2e4 h5f7 c4c5 c6c5 e4c5');
INSERT INTO evaluations VALUES(13227,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(13227,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(13227,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(13227,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(13227,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(13227,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(13227,36,'cp -21','g7g6 g5f6 e7f6 d3d4 c5d4 f3d4 f8g7 d4b5 e8g8 d1d6 c8e6 d6d8 a8d8 b5c7 c6b4 f1d3 b4d3 c2d3 d8d3 c7e6 f7e6 a1d1 d3d1 e1d1 f6f5 d1e2 f8c8 h1d1 c8c7 e4f5 e6f5 g2g3 g8f7 d1d2 g7c3 b2c3 f7e6 d2b2 b7b6 e2d3 c7c5 f2f3 c5a5 b2e2 a5e5');
INSERT INTO evaluations VALUES(15842,30,'cp 219',NULL);
INSERT INTO evaluations VALUES(15842,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(15842,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(15842,33,'cp 217',NULL);
INSERT INTO evaluations VALUES(15842,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(15842,35,'cp 229',NULL);
INSERT INTO evaluations VALUES(15842,36,'cp 215','c4b5 c7c6 d5c6 b7c6 b5a4 f8d6 e1g1 e8g8 f1e1 d8c7 h2h3 f8e8 d2d3 h7h6 a2a3 c8d7 b2b4 a5b7 c1b2 a7a5 c3e4 a5b4 e4d6 c7d6 a3b4 d6b4 b2e5 b7c5 c2c3 b4a5 a4c2 a5a1 d1a1 a8a1 e1a1 d7f5 e5f6 g7f6 d3d4 f5c2 d4c5 c2e4 a1e1 g8g7 f3d4 e8e5 f2f3 e4d5 e1e5 f6e5 d4f5 g7f6 f5h6');
INSERT INTO evaluations VALUES(16365,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(16365,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(16365,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(16365,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(16365,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(16365,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(16365,36,'cp -78','h7h6 g5f6 d8f6 g1f3 g7g5 g2g3 f8c5 d1e2 d7d6 b1d2 a7a5 d2c4 c8e6 f1g2 e8c8 f3d2 c8b8 h2h4 c5a7 c4e3 a7e3 e2e3 f6g7 d2c4 f7f6 e3e2 g7f7 g2f3 a5a4 f3h5 f7d7 c4e3 c6e7 h5g4 f6f5 g4f5 e7f5 e4f5 e6f5');
INSERT INTO evaluations VALUES(17534,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(17534,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(17534,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(17534,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(17534,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(17534,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(17534,36,'cp 82','e7e6 f1e2 d8c7 c1e3 b7b5 a1d1 g8e7 e1g1 e7c6 d4d2 d7d6 a2a4 b5a4 c3a4 f8e7 e3b6 c7b7 f2f4 e8g8 f1f3 a8b8 h2h3 c8d7 f3b3 a6a5 g1h2 h7h6 d1a1 b7c8 c2c4 g8h8 b3b5 c6b4 b5a5 d7a4 d2b4 a4c6');
INSERT INTO evaluations VALUES(18793,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(18793,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(18793,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(18793,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(18793,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(18793,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(18793,36,'cp 22','c2c3 g8f6 b1d2 f8e7 g1f3 f6h5 d4c5 h5f4 e3f4 d8c7 g2g3 e7c5 f1d3 e8g8 e1g1 b7b6 f1e1 c8b7 d1e2 a7a5 a2a3 c6e7 a1d1 a5a4 d2e4 h7h6 h2h4 e7f5 h4h5 b7e4 e2e4 d7d5 e4e5 c7e5 e1e5 f5g3 g1g2 g3e4 d3e4 d5e4 e5e4 b6b5 f3e5');
INSERT INTO evaluations VALUES(20682,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(20682,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(20682,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(20682,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20682,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20682,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20682,36,'cp 0','e1g1 g8f6 f1e1 f8e7 c4f1 e8g8 h2h3 e6e5 c2c3 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(20681,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20681,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20681,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20681,36,'cp 0','g8f6 f1e1 f8e7 c4f1 e6e5 c2c3 e8g8 h2h3 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(20881,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(20881,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(20881,32,'cp 119',NULL);
INSERT INTO evaluations VALUES(20881,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(20881,34,'cp 137',NULL);
INSERT INTO evaluations VALUES(20881,35,'cp 121',NULL);
INSERT INTO evaluations VALUES(20881,36,'cp 135','e7e6 f1b5 c8d7 c2c4 f8b4 b1c3 g8f6 a1c1 b4c3 c1c3 e8g8 e1g1 a7a5 h2h3 d7e8 c4d5 f6d5 c3c5 c6e7 b5e8 f8e8 f4g3 h7h6 f3e5 c7c6 d1b3 d8b6 b3c2 d5f6 f2f3 a5a4 f1e1 e7d5 e5d3 a4a3');
INSERT INTO evaluations VALUES(22209,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(22209,31,'cp 139',NULL);
INSERT INTO evaluations VALUES(22209,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(22209,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(22209,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(22209,35,'cp 131',NULL);
INSERT INTO evaluations VALUES(22209,36,'cp 122','c2c4 f8d6 f1d3 g8f6 b1c3 d6f4 e3f4 d8d6 e1g1 e8g8 c4d5 f6d5 g2g3 a7a6 f1e1 f8d8 d1e2 d5f6 a1d1 b7b5 a2a4 c6b4 a4b5 b4d3 e2d3 a6b5 d3b5 c8d7 b5c4 d8b8 d1a1 a8a1 e1a1 b8b2 a1a8 d7e8 f3g5 h7h6 c4e6 d6e6 g5e6');
INSERT INTO evaluations VALUES(21159,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(21159,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(21159,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(21159,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(21159,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(21159,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(21159,36,'cp -72','e7e5 g1f3 e5d4 f3d4 c8d7 a4c4 f8d6 b1c3 g8f6 f1e2 e8g8 d4b5 d6e7 e1g1 a7a6 b5d4 e7d6 c4d3 c6d4 e3d4 h7h6 e2f3 c7c6 f1e1 d7e6 c1e3 f8e8 a1d1 d8c7 h2h3 a8d8 d3d2 e8e7 c3e4 f6e4 f3e4 e6a2 e3h6 a2b3 h6g5');
INSERT INTO evaluations VALUES(21250,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(21250,31,'cp -172',NULL);
INSERT INTO evaluations VALUES(21250,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(21250,33,'cp -181',NULL);
INSERT INTO evaluations VALUES(21250,34,'cp -158',NULL);
INSERT INTO evaluations VALUES(21250,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(21250,36,'cp -177','f8g7 b1c3 h7h6 c3d5 c6e7 d5e7 g8e7 c2c3 c8g4 d1d3 d8d7 c1d2 g4h5 a2a4 e8c8 a4a5 d6d5 e4d5 e7d5 a5a6 b7b6 f1e1 h8e8 e1e8 d8e8 a1e1 e8e1 f3e1 d5f6 e1f3 h5g6 d3f1 f6e4 d2e1 c7c6 c4d3 d7d5 e1f2 e4d6 d3g6 f7g6');
INSERT INTO evaluations VALUES(5717,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(5717,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(5717,32,'cp -184',NULL);
INSERT INTO evaluations VALUES(5717,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(5717,34,'cp -167',NULL);
INSERT INTO evaluations VALUES(5717,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(5717,36,'cp -169','c2c3 h7h6 d1b3 d8d7 c4d3 g8f6 c1d2 e8g8 d2e1 f6g4 e1d2 c6e7 h2h3 g4f6 d2e1 c7c5 b1d2 d7c7 b3d1 e7g6 a2a4 c8e6 e1f2 b7b6 d1e2 f8e8 a4a5 b6a5 e4e5 d6e5 d3g6 f7g6 d4e5 f6d5 d2e4 g7e5 f3e5 c7e5 f2c5 e6f5');
INSERT INTO evaluations VALUES(5716,30,'cp -170',NULL);
INSERT INTO evaluations VALUES(5716,31,'cp -172',NULL);
INSERT INTO evaluations VALUES(5716,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(5716,33,'cp -183',NULL);
INSERT INTO evaluations VALUES(5716,34,'cp -172',NULL);
INSERT INTO evaluations VALUES(5716,35,'cp -188',NULL);
INSERT INTO evaluations VALUES(5716,36,'cp -178','h7h6 d1b3 d8d7 c4d3 g8f6 c1d2 e8g8 d2e1 c6e7 h2h3 c7c5 e1f2 d7c7 b1d2 c8e6 d3c4 e6c4 b3c4 e7g6 b2b4 b7b6 b4c5 b6c5 f1e1 a8c8 c4a6 c7c6 a6c6 c8c6 a1b1 c5d4 f2d4 c6c7 b1b5 f8e8 a2a4 f6h5 b5d5 g7f8 c3c4 h5g3 a4a5 c7c8 a5a6 g6e5');
INSERT INTO evaluations VALUES(5710,30,'cp -174',NULL);
INSERT INTO evaluations VALUES(5710,31,'cp -174',NULL);
INSERT INTO evaluations VALUES(5710,32,'cp -171',NULL);
INSERT INTO evaluations VALUES(5710,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(5710,34,'cp -184',NULL);
INSERT INTO evaluations VALUES(5710,35,'cp -167',NULL);
INSERT INTO evaluations VALUES(5710,36,'cp -159','d1b3 d8d7 c4d3 g8f6 c1d2 e8g8 d2e1 f6g4 e1d2 c6e7 h2h3 g4f6 d2e1 c7c5 e1f2 d7c7 b1d2 c8e6 d3c4 e6c4 b3c4 e7g6 b2b4 b7b6 a1b1 f8c8 b4c5 b6c5 c4a6 c7c6 a6c6 c8c6 b1b7 c5d4 c3d4 c6c2 a2a4 c2a2 f1c1 d6d5 e4e5 f6e4 f2e1 a2a4 b7d7');
INSERT INTO evaluations VALUES(21902,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(21902,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(21902,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(21902,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(21902,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(21902,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(21902,36,'cp 66','g1f3 f7f5 d2d3 f5f4 b2b4 d8f6 d1b3 g8h6 b1d2 f8e7 a2a4 h6f7 h2h4 e8g8 c1b2 g8h8 d3d4 c8g4 e1c1 a7a6 c1b1 f6g6 d1g1 e7f6 c4d3 h7h6 d4d5 c6d8 f3e1 g6h5 c3c4 b7b6 f2f3 g4d7 b1a1 h8g8');
INSERT INTO evaluations VALUES(22478,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(22478,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(22478,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(22478,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(22478,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(22478,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(22478,36,'cp 25','c2c3 d7d5 b1d2 g8f6 b2b4 f8g7 f1e2 e8g8 e1g1 f8e8 f1e1 c6b8 e2f1 c7c6 a2a4 a7a5 b4b5 c8g4 c1a3 b8d7 a1b1 d8c7 b5c6 b7c6 h2h3 g4f3 d1f3 g7f8 a3f8 e8f8 e4d5 c6d5 c3c4 d5d4 b1b5');
INSERT INTO evaluations VALUES(23395,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(23395,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(23395,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(23395,33,'cp 198',NULL);
INSERT INTO evaluations VALUES(23395,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(23395,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(23395,36,'cp 227','d4d5 c6a5 b1c3 g8f6 e2e4 f5e4 f3g5 e7e6 g5e4 f8b4 c1g5 d8e7 g5f6 g7f6 d1h5 e7f7 e1c1 b4c3 h5f7 e8f7 e4c3 b7b6 f1e2 a7a6 e2h5 f7g7 f2f4 a5b7 h1e1 b7d6 g2g4 h8f8 a2a4 b6b5 a4a5 d6c4 c3e4 c4a5 d5e6 d7e6 e4c5 a5c4 c5e6 c8e6 e1e6 a8d8 e6e7 g7h8 b2b3 c4b6 h5f7 d8d1 c1d1 c7c5');
INSERT INTO evaluations VALUES(25742,30,'cp 81',NULL);
INSERT INTO evaluations VALUES(25742,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(25742,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(25742,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(25742,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(25742,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(25742,36,'cp 86','d4e5 d6e5 d1d5 d8d5 b3d5 g8e7 f3e5 e7d5 e5c6 d5b4 c6b4 f8b4 c1d2 b4d6 b1c3 e8g8 f2f3 g4e6 e1c1 f7f5 e4e5 d6c5 b2b3 f8d8 c3e2 d8d5 e2f4 d5e5 f4d3 e5d5 d3c5 d5c5 h1e1 a8e8 d2b4 c5c6 e1e5 e6d7 e5d5 d7e6 d5d4 h7h6 h2h4 g7g5 g2g3 e6f7 h4g5 h6g5 d4d7 e8e2');
INSERT INTO evaluations VALUES(26114,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26114,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(26114,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26114,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(26114,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(26114,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26114,36,'cp 28','c4d5 c6d5 b1c3 a7a6 f1d3 g8e7 e1g1 f5d3 d1d3 e7c6 c1d2 c6b4 d3e2 a8c8 a1c1 f8e7 g2g3 e8g8 h2h4 b4c6 e2d3 d7b6 c3e2 h7h6 g1g2 d8d7 g3g4 b6c4 d2c3 c6b4 c3b4 e7b4 b2b3 c4a3 g4g5 d7b5 e2f4 b5d3 f4d3 b4e7 c1c8 f8c8 g5h6 g7h6');
INSERT INTO evaluations VALUES(26637,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(26637,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(26637,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(26637,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(26637,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(26637,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(26637,36,'cp 34','e2e3 g8f6 d2d4 e7e6 f1e2 f8b4 c2c3 b4d6 c3c4 d6b4 f3d2 g4e2 d1e2 f6e4 e1g1 b4d2 b1d2 e4d2 e2d2 e8g8 a1c1 f8e8 a2a4 h7h6 f1d1 c6e7 d2e2 d8d7 h2h3 e7f5 b2a3 c7c6 e2g4 a8d8 g4f3 a7a5 f3f4');
INSERT INTO evaluations VALUES(26636,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(26636,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(26636,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(26636,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(26636,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(26636,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(26636,36,'cp 37','g8f6 d2d4 e7e6 f1e2 f8b4 c2c3 b4d6 c3c4 d6b4 f3d2 g4e2 d1e2 f6e4 e1g1 e4d2 b1d2 b4d2 e2d2 e8g8 a2a4 f8e8 d2e2 h7h6 f1c1 c6e7 a4a5 a7a6 h2h3 d8d7 e2c2 e7c6 c2d2 a8d8 d2c3 c6e7 b2a3 e7c6 c1d1 g8h8 a3c5');
INSERT INTO evaluations VALUES(30408,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(30408,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(30408,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(30408,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(30408,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(30408,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(30408,36,'cp 41','b5d7 b8d7 d2d4 g8f6 c1e3 b7b5 d1e2 c7c6 e1c1 d8b8 h2h3 f8e7 d4e5 d6e5 g2g4 d7c5 g4g5 f6d7 h3h4 b5b4 c3b1 b8b5 h1e1 c5e4 e3f4 b5e2 e1e2 e5f4 e2e4 d7c5 e4c4 c5e6 b1d2 c6c5 c4e4 e8g8 d1e1 a6a5 d2c4 a5a4 f3e5 a8a6 e5d7 f8d8 c4b6 h7h6 g5h6 g7h6 c1b1');
INSERT INTO evaluations VALUES(31632,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(31632,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(31632,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(31632,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31632,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(31632,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(31632,36,'cp -1','f8c5 c2c4 g8e7 e1g1 b8c6 b1c3 c5b4 f1d1 b4c3 d3c3 d8a5 c3a5 c6a5 c4d5 e7d5 f3g5 e8g8 g5e4 a5c6 f2f4 f8c8 g2g3 c6e7 e4d6 c8c6 b2b3 a8b8 d1d2 b7b5 c1b2 b5b4 a1c1 c6c1 b2c1 a7a5 g1f2 a5a4 b3a4 b8a8 a2a3 a8a4');
INSERT INTO evaluations VALUES(28095,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(28095,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(28095,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(28095,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28095,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(28095,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(28095,36,'cp 0','c2c4 g8e7 e1g1 b8c6 b1c3 c5b4 c4d5 b4c3 b2c3 d8d5 d3b1 e8c8 c3c4 d5a5 f3g5 c6e5 c4c5 e7c6 c1e3 a5a4 g5e4 e5c4 b1b3 a4b3 a2b3 f7f5 e4g5 c4d2 e3d2 d8d2 b3b4 h8e8 g5f7 e8f8 f7g5 f8e8');
INSERT INTO evaluations VALUES(32426,30,'cp 451',NULL);
INSERT INTO evaluations VALUES(32426,31,'cp 474',NULL);
INSERT INTO evaluations VALUES(32426,32,'cp 463',NULL);
INSERT INTO evaluations VALUES(32426,33,'cp 501',NULL);
INSERT INTO evaluations VALUES(32426,34,'cp 495',NULL);
INSERT INTO evaluations VALUES(32426,35,'cp 512',NULL);
INSERT INTO evaluations VALUES(32426,36,'cp 489','f3b3 g8h6 b3b7 b8d7 b7c6 a8b8 d2d3 d8c8 c1e3 h6g8 c6d5 c7c6 d5f7 e8d8 b1c3 c8c7 e1g1 g8e7 a1b1 d7b6 c4e6 e7g6 b2b4 c7f7 e6f7 g6e7 a2a4 d6d5 e3b6 b8b6 e4d5 b6b8 a4a5 d8d7 f7e6 d7c7 b4b5 c6d5 e6d5 e7d5 c3d5 c7d7 c2c4 f8c5 f1c1 d7e6 a5a6 h8c8 g1f1 c5d4 f1e2 d4c5');
INSERT INTO evaluations VALUES(34029,30,'cp 177',NULL);
INSERT INTO evaluations VALUES(34029,31,'cp 164',NULL);
INSERT INTO evaluations VALUES(34029,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(34029,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(34029,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(34029,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(34029,36,'cp 175','d1e2 f6h5 b1c3 d7d5 c3d5 e7d8 e2c4 b8c6 d5f4 h5f4 c4f4 f8e7 f1c4 c8e6 c4e6 f7e6 f4e4 h8f8 c2c3 d8d5 e4d5 e6d5 d2d4 c6d8 a2a4 a7a5 c1e3 d8e6 e1e2 e8d7 g2g3 c7c6 h2h4 h7h5 e3d2 b7b5 a4b5 c6b5 f3e1 b5b4 e1d3');
INSERT INTO evaluations VALUES(34451,30,'cp 255',NULL);
INSERT INTO evaluations VALUES(34451,31,'cp 260',NULL);
INSERT INTO evaluations VALUES(34451,32,'cp 265',NULL);
INSERT INTO evaluations VALUES(34451,33,'cp 259',NULL);
INSERT INTO evaluations VALUES(34451,34,'cp 266',NULL);
INSERT INTO evaluations VALUES(34451,35,'cp 247',NULL);
INSERT INTO evaluations VALUES(34451,36,'cp 240','d6e5 c1g5 f6d6 f3e5 d6d1 a1d1 f7f6 g5h4 g8h6 e5d3 c8e6 f2f3 b8d7 h4f2 a7a5 f1e2 f8b4 f2d4 b4d6 d4e3 d6b4 e1d2 e8c8 a2a3 b4d6 h2h4 f6f5 d2c1 d6c7 e3f2 f5e4 c3e4 e6f5 f2d4 h8g8 g2g4 f5e4 f3e4');
INSERT INTO evaluations VALUES(35198,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(35198,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(35198,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(35198,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(35198,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(35198,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(35198,36,'cp 62','d2d3 f8c5 g2g3 h4e7 f1g2 g8f6 g1e2 e8g8 e1g1 a7a6 h2h3 c5a7 b1d2 f8d8 f1e1 d7d6 d3d4 h7h6 a2a4 b8d7 d2c4 c7c6 d1c2 d6d5 e4d5 c6d5 c4e5 d7e5 d4e5 e7e5 c1f4 e5f5 c2f5 c8f5 g3g4 f5d3 e2d4 a7d4 c3d4 g7g5 f4d2 d8e8 e1e8 a8e8 d2e3');
INSERT INTO evaluations VALUES(35947,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(35947,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(35947,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(35947,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(35947,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(35947,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35947,36,'cp 30','d4d5 f8e8 f1e1 a7a6 b1c3 e7e6 c1g5 h7h6 g5f6 g7f6 d1d2 g8g7 g2h3 f6c3 d2c3 d8f6 c3f6 g7f6 e2e4 e6e5 h3c8 e8c8 a2a4 a6a5 f3d2 f6e7 f2f4 b8d7 a1a3 g6g5 f4g5 h6g5 a3b3');
INSERT INTO evaluations VALUES(35983,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(35983,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(35983,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(35983,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(35983,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(35983,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(35983,36,'cp 25','f8e8 b1c3 e7e6 c1g5 h7h6 g5f6 d8f6 f3e1 f6e7 d1d2 e6e5 e2e4 a7a6 a2a4 h6h5 h2h4 e8f8 e1f3 b8d7 g2h3 d7f6 h3c8 a8c8 g1g2 f6h7 a1e1 c8e8 d2e2 g7h6 f3h2 h7f6 f1h1 g8g7 e1a1 b7b6 a1a3 f8h8 a4a5');
INSERT INTO evaluations VALUES(36179,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(36179,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(36179,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(36179,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(36179,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(36179,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(36179,36,'cp 38','g2g4 c7c6 g4g5 h6g5 c1g5 b7b5 c4b3 b8a6 h1g1 a6c5 g5h6 f6e8 d3d4 c5b3 a2b3 b5b4 c3e2 f7f5 d4e5 f5e4 f3d4 d8c7 e5e6 e7h4 h6e3 c6c5 d4b5 c7e7 e2f4 f8f4 e3f4 c8e6 d1d2 e7f6 e1c1 d6d5 f4g5 f6f2 g5h4 f2h4 d2h2 e4e3 c1b1 h4f6 d1f1 e6f5 b5c7 e8c7 h2c7 a8c8');
INSERT INTO evaluations VALUES(36827,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(36827,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(36827,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(36827,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(36827,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(36827,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(36827,36,'cp 34','g1e2 c7c5 e1g1 b8c6 c1e3 b7b6 c4b5 c8b7 a1c1 a8c8 b5c6 c8c6 d4d5 c6d6 c3c4 e7e6 d1a4 e6d5 e4d5 d6d7 e3g5 f7f6 g5e3 b7c8 f1e1 f6f5 e2f4 d7e7 a4d1 d8d6 h2h3 e7e4 f4d3 f8e8 d1a4 d6d8 a4a7 g6g5');
INSERT INTO evaluations VALUES(36826,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(36826,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(36826,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(36826,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(36826,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(36826,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(36826,36,'cp 35','c7c5 e1g1 b8c6 c1e3 b7b6 a1c1 c8b7 c4b5 a8c8 b5c6 c8c6 d4d5 c6d6 c3c4 e7e6 d1a4 e6d5 e4d5 d6d7 e3g5 f7f6 g5e3 b7c8 a4c2 f8e8 h2h3 f6f5 f1d1 d8f6 c2d2 d7e7 d5d6 e7f7 e3f4 c8d7 e2c3 f6d4 c3d5 d4d2');
INSERT INTO evaluations VALUES(36049,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(36049,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(36049,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(36049,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(36049,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(36049,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(36049,36,'cp 12','c1e3 b8c6 e1g1 b7b6 a1c1 c8b7 d4d5 c6e5 c4b3 f7f5 c3c4 f5e4 e2g3 e7e6 g3e4 e6d5 c4d5 a8c8 d5d6 c5c4 b3c2 d8d7 h2h3 h7h6 f1e1 c8d8 e3d4 b7e4 e1e4 d7d6 d4e5 d6e5 d1e2 e5g5 e2c4 g8h8 c1f1 b6b5 c4e6 g5f6 e6f6 f8f6 e4g4 g6g5 g4e4 d8d2 e4e8 g7f8');
INSERT INTO evaluations VALUES(36043,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(36043,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(36043,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(36043,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(36043,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(36043,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(36043,36,'cp 18','b8c6 e1g1 b7b6 d1d2 c8b7 a1c1 e7e6 f1d1 a8c8 c4b5 c5d4 c3d4 c6a5 f2f3 c8c1 d1c1 a7a6 b5d3 a5c6 d3c2 b6b5 c2b3 d8a5 d4d5 e6d5 e4d5 a5d2 e3d2 c6d4 e2d4 g7d4 g1f1 f8d8 c1d1 d4e5 d2a5 d8d7 a5b4 e5d6 b4d6 d7d6 f1e1 a6a5 a2a3 b5b4 a3b4');
INSERT INTO evaluations VALUES(38953,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(38953,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(38953,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(38953,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(38953,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(38953,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(38953,36,'cp 176','c1d2 b4d2 d1d2 d5d4 g1f3 b8c6 e2e3 d4e3 d2e3 g8e7 f1e2 e7f5 e3e4 f5d4 b1c3 d4f3 g2f3 d8d4 c3d5 e8g8 e4d4 c6d4 d5c7 f8d8 a1c1 a8b8 h2h4 c8d7 c7d5 d7b5 d5e3 b5c6 h1h3 d4e6 b2b3 e6f4 h3g3 f4g6 h4h5 g6e5 e3f5');
INSERT INTO evaluations VALUES(38952,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(38952,31,'cp 148',NULL);
INSERT INTO evaluations VALUES(38952,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(38952,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(38952,34,'cp 170',NULL);
INSERT INTO evaluations VALUES(38952,35,'cp 173',NULL);
INSERT INTO evaluations VALUES(38952,36,'cp 171','b4d2 d1d2 d5c4 f2f4 d8d2 b1d2 f7f6 g1f3 b8c6 d2c4 c8e6 e2e3 e8c8 h2h3 f6e5 c4e5 c6e5 f3e5 g8e7 f1d3 e7c6 e5c6 d8d3 e1e2 d3d6 c6d4 e6d5 e2f2 h8f8 h1d1 g7g5 f4f5 d5f7 a2a3 f8d8 a1c1');
INSERT INTO evaluations VALUES(39286,30,'cp -54',NULL);
INSERT INTO evaluations VALUES(39286,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(39286,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(39286,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(39286,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(39286,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(39286,36,'cp -17','d7d5 c4d5 c3b2 c1b2 g8f6 b1c3 f6d5 c3d5 e8g8 d1c1 b8d7 e1g1 c7c6 d5e3 f7f6 e3f5 d8e8 f1e1 b7b6 c1f4 d7e5 f3e5 f6e5 b2e5 e8g6 e5d4 c8f5 d4c5 b6c5 e4f5 f8f5 f4c4 g6f7 a1c1 f7c4 c1c4 a8d8 f2f3 d8d4 e1c1 d4d2 c4c5 f5c5 c1c5');
INSERT INTO evaluations VALUES(41258,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(41258,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(41258,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(41258,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(41258,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41258,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(41258,36,'cp 46','f1g2 c7c5 d2d4 e8g8 d4d5 e7e6 e1g1 f8e8 f3g5 e6d5 c3d5 b8d7 c1f4 d7b6 d5f6 d8f6 a1b1 c8f5 e2e4 f5d7 b2b3 h7h6 e4e5 d6e5 g5e4 f6e6 e4c5 e6e7 f4e3 d7c6 g2c6 b7c6 d1f3 e5e4 f3g4 h6h5 g4h3 a8d8 b1d1 d8d1 f1d1');
INSERT INTO evaluations VALUES(42178,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(42178,31,'cp 134',NULL);
INSERT INTO evaluations VALUES(42178,32,'cp 144',NULL);
INSERT INTO evaluations VALUES(42178,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(42178,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(42178,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(42178,36,'cp 147','b2c3 e8g8 c4d3 c7c5 e1g1 h7h6 e3e4 b8c6 e4e5 f6d5 d4c5 d5c3 d1c2 c3d5 c2c4 c8d7 a1b1 c6e7 b1b7 a8b8 b7b3 d8a5 c1d2 a5a4 f1c1 a4c4 d3c4 e7c6 h2h4 f8c8 h4h5 a7a5 a2a3 b8b3 c4b3 d7e8 b3c2 c8b8 c1b1 b8b1 c2b1 a5a4 b1c2 c6e7');
INSERT INTO evaluations VALUES(45476,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(45476,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(45476,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(45476,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(45476,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(45476,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(45476,36,'cp -14','d2d4 b8c6 g1f3 c8f5 f3h4 f5d7 f1g2 e7e6 e1g1 f8e7 h4f3 e8g8 f3e5 d7e8 c1f4 a8c8 a1c1 b7b5 a2a3 a7a5 e2e3 b5b4 a3b4 c6b4 d1e2 h7h6 e3e4 d5e4 c3e4 c8c1 f1c1 b4d5 f4d2 f6e4 g2e4');
INSERT INTO evaluations VALUES(46053,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(46053,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(46053,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(46053,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(46053,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(46053,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(46053,36,'cp 39','c2c4 f8e7 b1c3 e8g8 e1g1 b8c6 c1e3 d6d5 c4d5 e6d5 e4d5 c6b4 d4f5 c8f5 d3f5 b4d5 e3d4 d5c3 d4c3 d8b6 d1a4 a7a5 a4d4 b6d4 c3d4 a5a4 g2g3 f8d8 a1d1 a4a3 b2b3 e7f8 d4b6 d8d5 f5h3 d5b5 b6d4 a8d8 f1e1');
INSERT INTO evaluations VALUES(46472,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(46472,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(46472,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(46472,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(46472,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(46472,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(46472,36,'cp 2','f1d3 b8c6 c1e3 f8d6 h2h3 e8g8 d1f3 f6e4 g1e2 f7f5 e1g1 c8e6 a1d1 d8f6 c3e4 d5e4 d3e4 f5e4 f3f6 f8f6 d4d5 e6d7 d5c6 d7c6 d1d2 a7a6 f1d1 d6e5 e2d4 c6a4 b2b3 a4e8 f2f3 e4f3 d4f3 e5g3 e3f2 g3f2 d2f2 f6f4 g2g3 f4e4 d1d6 e8b5');
INSERT INTO evaluations VALUES(47912,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(47912,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47912,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47912,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47912,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47912,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47912,36,'cp 0','b8c6 g1f3 e5e4 f3d4 c6d4 b2d4 d6d5 c2c4 d5c4 e2e3 f8b4 a2a3 b4e7 b1c3 c7c5 d4f6 e7f6 a1c1 e8g8 c3e4 f6b2 c1c4 c8e6 c4a4 e6d7 a4c4');
INSERT INTO evaluations VALUES(15612,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15612,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15612,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15612,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15612,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15612,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15612,36,'cp 0','g1f3 e5e4 f3d4 c6d4 b2d4 d6d5 c2c4 d5c4 e2e3 f8b4 a2a3 b4e7 b1c3 c7c5 d4f6 e7f6 a1c1 e8g8 c3e4 f6b2 c1c4 c8e6 c4a4 e6d7 a4c4');
INSERT INTO evaluations VALUES(48621,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(48621,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(48621,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(48621,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(48621,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(48621,35,'cp -37',NULL);
INSERT INTO evaluations VALUES(48621,36,'cp -37','f8b4 b1c3 e8g8 f1e2 f8e8 g1f3 d5c4 e1g1 c8e6 h2h3 h7h6 d4d5 e6f5 g5h4 b4c3 b2c3 d8d5 h4f6 d5d1 f1d1 g7f6 e2c4 f5e4 c4b5 b8c6 f3d2 e4c2 b5c6 b7c6 d1e1 c6c5 d2b3 c5c4 e1e8 a8e8 b3d4 c2d3 d4b5 a7a5 b5c7 e8e5 g1h2 e5c5 c7a6 c5c6 a6b8 c6c7');
INSERT INTO evaluations VALUES(48815,30,'cp -92',NULL);
INSERT INTO evaluations VALUES(48815,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(48815,32,'cp -99',NULL);
INSERT INTO evaluations VALUES(48815,33,'cp -113',NULL);
INSERT INTO evaluations VALUES(48815,34,'cp -102',NULL);
INSERT INTO evaluations VALUES(48815,35,'cp -112',NULL);
INSERT INTO evaluations VALUES(48815,36,'cp -84','c8g4 h2h3 g4f3 d1f3 d5e4 f3e3 d8d4 b1c3 b8c6 e1g1 f8c5 e3d4 e5d4 c3e4 f6e4 f1e1 e8c8 e1e4 d4d3 c2d3 c6b4 b3c4 b4d3 c4d3 d8d3 b2b4 f7f5 e4f4 c5d6 f4f5 d6b4 f5f7 h8f8 f7f8 b4f8 g1f1 b7b5 f1e2 d3d5 a1b1 a7a5 d2d4 b5b4 c1e3 f8e7');
INSERT INTO evaluations VALUES(49337,30,'cp 179',NULL);
INSERT INTO evaluations VALUES(49337,31,'cp 193',NULL);
INSERT INTO evaluations VALUES(49337,32,'cp 160',NULL);
INSERT INTO evaluations VALUES(49337,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(49337,34,'cp 166',NULL);
INSERT INTO evaluations VALUES(49337,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(49337,36,'cp 163','d2d3 e4c5 d3d4 c5a6 a2a3 d7d5 c2c4 c7c5 b1c3 a6c7 d4c5 f8c5 d1c2 b8d7 c1f4 d5c4 f1c4 d7b6 a1d1 c8d7 c4a2 c5e7 e1g1 c7d5 c3d5 b6d5 a2d5 e6d5 d1d5 d8c8 c2d2 d7c6 d5d3 c6f3 d3f3 e8g8 f1d1 f8d8 f3d3 d8d3 d2d3 c8e6 d3d7 e6d7 d1d7 a8d8 d7d8 e7d8 g2g4 g8f8 f4e3 f8e7 e3a7');
INSERT INTO evaluations VALUES(49336,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(49336,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(49336,32,'cp 151',NULL);
INSERT INTO evaluations VALUES(49336,33,'cp 148',NULL);
INSERT INTO evaluations VALUES(49336,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(49336,35,'cp 168',NULL);
INSERT INTO evaluations VALUES(49336,36,'cp 165','e4c5 d3d4 c5a6 a2a3 d7d5 b2b4 c7c6 b1d2 f8e7 h2h4 b7b6 h4h5 c6c5 h5h6 g7g6 c2c3 c5c4 f1e2 a6c7 d2f1 b8d7 c1f4 e8g8 f1g3 f7f5 e5f6 d7f6 d1d2 f6g4 a3a4 e7d6 f4g5 d8e8 e1g1 c8d7 e2d1 e8f7 b4b5 a7a6 b5a6 c7a6 a1b1 b6b5 a4b5 a6c7');
INSERT INTO evaluations VALUES(50888,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(50888,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(50888,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(50888,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(50888,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(50888,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(50888,36,'cp 94','e1g1 g7g6 h2h3 f8g7 f1e1 g8f6 a2a4 e8g8 b1c3 b7b6 c1f4 c8b7 d1d2 g6g5 f4h2 b8c6 a1d1 c6e7 e4e5 d6e5 f3e5 e7f5 f2f4 g5f4 d2f4 f6h5 f4f2 d8h4 f2h4 f5h4 b2b3 a8d8');
INSERT INTO evaluations VALUES(50887,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(50887,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(50887,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(50887,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(50887,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(50887,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(50887,36,'cp 97','g8f6 f1e1 f8e7 c4d3 c7c5 c2c3 c8d7 h2h3 d7c6 d3c2 b8d7 c1f4 c5d4 c3d4 d7b6 a2a3 e8g8 b1c3 a8c8 d1d3 c6e8 a1d1 e8c6 b2b3 d6d5 e4e5 f6e4 f4c1 d8c7 f3h2 f7f5 e5f6 f8f6 h2g4 f6g6 g1f1');
INSERT INTO evaluations VALUES(52238,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(52238,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(52238,32,'cp 126',NULL);
INSERT INTO evaluations VALUES(52238,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(52238,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(52238,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(52238,36,'cp 132','e7e6 e1g1 g8e7 f1e1 b8c6 c2c3 c6e5 c4e2 e5f3 e2f3 e7g6 d2d4 c5d4 c3d4 f8b4 b1c3 e8g8 d1b3 a7a5 a2a3 b4c3 b2c3 b7b5 g2g3 d8c7 h2h4 a5a4 b3b4 c8b7 h4h5 g6h8 h5h6 g7g5 c1d2 h8f7 a1c1 a8b8 b4b5 b7c6 b5d3 f7h6');
INSERT INTO evaluations VALUES(52221,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(52221,31,'cp 138',NULL);
INSERT INTO evaluations VALUES(52221,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(52221,33,'cp 130',NULL);
INSERT INTO evaluations VALUES(52221,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(52221,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(52221,36,'cp 129','e1g1 g8e7 d2d4 c5d4 f3d4 e7g6 c2c3 b8c6 c1e3 f8e7 d1h5 e8g8 f2f4 d7d5 e4d5 c6a5 c4d3 d8d5 h5d5 e6d5 d4b5 c8d7 b5c7 a8d8 c7d5 d7e6 d5e7 g6e7 d3e4 b7b6 b1a3 f8e8 e3f2 d8d2 b2b4 a5c4 a3c4 e6c4 f1e1 g8f7 e4f3 e7d5 e1e8 f7e8 a2a4 d5f4');
INSERT INTO evaluations VALUES(52387,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52387,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(52387,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(52387,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52387,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(52387,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(52387,36,'cp -10','d7d5 e4d5 c6d5 c2c3 b8c6 d2d4 g8f6 c1f4 c8f5 e2c1 e7e6 f1d3 f8d6 f4d6 d8d6 d3f5 e6f5 b1d2 e8g8 e1g1 g7g6 d2f3 d6c7 f1e1 f8e8 c1d3 e8e1 d1e1 a8e8 e1d2 g8g7 a1e1 e8e4 d3c5 e4e7 e1e7 c7e7 d2e1 f6e4 c5d3 c6a5 f3d2');
INSERT INTO evaluations VALUES(52817,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(52817,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(52817,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(52817,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(52817,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(52817,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(52817,36,'cp 0','e4d5 c6d5 c2c3 g8f6 d2d4 b8c6 c1f4 c8f5 e2c1 e7e6 f1d3 f8d6 f4d6 d8d6 b1d2 f5g4 d3e2 g4f5');
INSERT INTO evaluations VALUES(54584,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54584,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54584,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54584,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54584,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54584,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54584,36,'cp 0','b8c6 f1d3 g8f6 e1g1 c8g4 f1e1 f8e7 c1e3 e8g8 h2h3 g4h5 b1c3 f8e8 g2g4 h5g6 d3g6 h7g6 e3g5 d8d7 f3e5 d7d8');
INSERT INTO evaluations VALUES(55429,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(55429,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(55429,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(55429,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(55429,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(55429,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(55429,36,'cp 64','d4d3 c4d3 b8c6 c1e3 g7g6 c3c4 f8g7 b1c3 g7c3 b2c3 c8g4 e1g1 c6e5 d3e2 g4f3 e2f3 g8f6 f3e2 f6e4 d1d4 f7f5 c4c5 e8g8 c5d6 f8e8 a1b1 d8d6 f1e1 b7b6 f2f3 c7c5 d4d6 e4d6 e3h6 a8d8 b1d1 d8d7');
INSERT INTO evaluations VALUES(1133,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(1133,31,'cp 172',NULL);
INSERT INTO evaluations VALUES(1133,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(1133,33,'cp 174',NULL);
INSERT INTO evaluations VALUES(1133,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(1133,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(1133,36,'cp 171','e7d6 e1c1 d6d4 d1d4 f8c5 d4c4 c5b6 e4e5 d7d5 e5d6 c7d6 g5f6 g7f6 c4f4 c8e6 g2g3 h7h5 f1g2 h5h4 g3g4 a8b8 h2h3 b6d8 c1b1 e8g8 c3b5 a7a6 b5d4 f8e8 a2a3 d6d5 h1d1 b7b5 d4c6 b8b6 c6d8 e8d8 f4f6 b5b4 a3b4 b6b4 f6h6 b4f4 f2f3');
INSERT INTO evaluations VALUES(2840,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2840,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2840,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2840,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2840,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2840,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2840,36,'cp 0','c1e3 a7e3 f2e3 c8e6 c3d5 f6d7 d1e1 c6e7 d5e7 d8e7 c4e6 f7e6 e1c3 c7c5 a3a4 c5b4 c3b4 b7b6 a4a5 b6a5 a1a5 a8b8 b4d2 d7c5 a5a1 g8h7 d2a5 b8c8 a5b4 c8a8 b4d2');
INSERT INTO evaluations VALUES(2985,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(2985,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(2985,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(2985,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(2985,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(2985,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(2985,36,'cp 21','f6e4 g5f4 d7f6 f3e5 e7d6 h2h3 d8e7 a1c1 f8e8 f1e1 e4c3 b2c3 h7h6 a2a4 c8e6 d1f3 a8c8 a4a5 b7b5 a5b6 a7b6 c1a1 b6b5 a1a5 b5b4 a5a6 d6e5 d4e5 f6e4 c3b4 e7b4 e1b1 b4d2 d3e4 d5e4 f3e4 c6c5 a6d6 d2c3 b1d1 c3b4 e4b4');
INSERT INTO evaluations VALUES(4209,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4209,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4209,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4209,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4209,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4209,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4209,36,'cp 0','a7a5 h2h3 d7d6 c4b5 c6a7 b5c4');
INSERT INTO evaluations VALUES(2639,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2639,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2639,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2639,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2639,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(2639,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(2639,36,'cp 4','d3d4 c5b6 d4e5 f6e4 c4d5 e4f2 f1f2 d7d6 d5c6 b6f2 g1f2 b7c6 b1a3 c8a6 c1f4 a8b8 a1b1 g7g5 f4g3 d6e5 d1d8 f8d8 g3e5 a6d3 b1d1 b8b2 d1d2 d8b8 e5c7 b8b3 d2b2 b3b2 f2e3 d3a6 f3d2 b2a2 a3c4 a2a4 c4a5 c6c5');
INSERT INTO evaluations VALUES(4745,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(4745,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(4745,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(4745,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(4745,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(4745,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(4745,36,'cp 70','g8f6 f1c4 c6a5 c4b3 a5b3 d4b3 b7b5 e3h6 g7h6 d2h6 d8c7 e1c1 c8b7 h1e1 a8c8 e1e2 f6g4 h6g5 g4e5 c3d5 b7d5 e4d5 f7f6 g5h6 e8f7 c1b1 c7d6 f2f4 e5g4 h6h4 h7h5 d1e1 c8e8 h2h3 g4h6 h4f2 d6d5 e2d2 d5f5 b3c5 d7d6 c5a6');
INSERT INTO evaluations VALUES(7059,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(7059,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(7059,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(7059,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(7059,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(7059,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(7059,36,'cp 75','f1c4 b7b5 c4b3 c8b7 f2f3 e8g8 e1c1 c6a5 c1b1 d8c7 g2g4 d7d5 g4g5 f6h5 c3d5 b7d5 e4d5 f8d8 c2c3 d8d7 d2c2 a8d8 d4c6 a5c6 d5c6 c7c6 d1d7 d8d7 c2e4 c6c8 h1d1 d7d1 b3d1 e7e6 d1e2 c8c7 a2a3');
INSERT INTO evaluations VALUES(4867,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(4867,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4867,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4867,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4867,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(4867,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(4867,36,'cp 4','g8f6 d2d3 e8g8 e1g1 d7d6 c3d5 f6d5 c4d5 c6e7 d5b3 e7g6 c2c3 c5b6 h2h3 d8e7 c1g5 e7e8 g5e3 b6e3 f2e3 e8e7 d1d2 c8e6 b3e6 e7e6 f1f2 c7c6 a1f1 h7h6 d2c2 d6d5 f3d2 a8e8 a3a4 e6e7 e4d5 c6d5 d3d4 e5e4');
INSERT INTO evaluations VALUES(5109,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(5109,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(5109,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(5109,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(5109,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(5109,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(5109,36,'cp 17','c2c3 g8e7 c1e3 b7b6 d3d4 c5d4 c3d4 d7d5 e4d5 e6d5 c4b5 e8g8 h2h3 c8b7 b5d3 c6b4 b1c3 b4d3 d1d3 e7f5 e3f4 f8e8 f1e1 d8d7 f3e5 d7d8 c3e2 f5d6 a1c1 d6c4 e5c4 d5c4 d3c4 d8f6 f4g3 a8d8 c4c7 b7a6 e2c3 f6d4 c7a7');
INSERT INTO evaluations VALUES(5107,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(5107,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(5107,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5107,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5107,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5107,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5107,36,'cp 0','g8e7 c1e3 b7b6 d3d4 c5d4 c3d4 d7d5 e4d5 e6d5 c4b5 e8g8 h2h3 c8b7 b5d3 c6b4 b1c3 b4d3 d1d3 e7f5 e3f4 a8c8 f1e1 d8d7 f3e5 d7d8 c3b5 a7a6 b5c3 f8e8 e5f3 d8d7 f3e5');
INSERT INTO evaluations VALUES(7963,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7963,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7963,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7963,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7963,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7963,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(7963,36,'cp 17','c1e3 b7b6 d3d4 c5d4 c3d4 d7d5 e4d5 e6d5 c4b5 e8g8 h2h3 c8b7 b5d3 c6b4 b1c3 b4d3 d1d3 f8e8 f1e1 e7f5 e3f4 d8d7 a2a4 e8e1 a1e1 a8d8 g2g4 f5d6 c3b5 d6e4 f3e5 g7e5 f4e5 f7f6 e5f4 g6g5 f4h2 b7c6 f2f3 c6b5 d3b5 d7b5');
INSERT INTO evaluations VALUES(5544,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5544,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5544,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5544,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5544,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5544,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5544,36,'cp 0','h2h3 g8f6 f1c4 b7b6 e1g1 e8g8 d1e2 c8b7 f1d1 d8c7 c3b5 c7c8 b5c3');
INSERT INTO evaluations VALUES(5684,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(5684,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(5684,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(5684,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(5684,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(5684,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(5684,36,'cp 23','f1b5 g8h6 g1f3 e8g8 c2c3 f7f6 e5f6 e7f6 e1g1 c8e6 b1d2 e6d5 b5c4 h6f7 c4d5 d8d5 d1b3 c6e7 f1e1 a8e8 d2e4 b7b6 c1d2 d5b3 a2b3 a7a5 g1f1 h7h6 e1e2 f6f5 e4f2 e7d5 a1e1 d5f6 c3c4 e8e2 e1e2 f7d6 e2e7 f8f7 e7f7');
INSERT INTO evaluations VALUES(6519,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(6519,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(6519,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6519,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(6519,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(6519,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(6519,36,'cp 12','d7d6 b2b4 c5b6 a2a4 a7a6 f4e5 d6e5 f1c4 g8f6 d1e2 e8g8 d2d3 d8d6 a4a5 b6a7 c1e3 a7e3 e2e3 c8e6 c4e6 f7e6 b1a3 a8d8 e1e2 f6d5 e4d5 e6d5 h1f1 d5d4 a3c4 d4e3 c4d6 c7d6 e2e3 d6d5 d3d4 e5d4 f3d4 f8e8 e3d3 c6e5 d3c2 e5c4 f1e1 h7h6');
INSERT INTO evaluations VALUES(6030,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(6030,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(6030,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(6030,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(6030,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(6030,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(6030,36,'cp 20','b2b4 c5b6 a2a4 a7a6 f4e5 d6e5 f1c4 g8f6 d2d3 d8d6 d1e2 e8g8 a4a5 b6a7 c1e3 c8e6 c4e6 a7e3 e2e3 d6e6 b1d2 a8d8 d2c4 f6h5 e1g1 h5f4 c4e5 c6e5 f3e5 e6e5 d3d4 f4g2 g1g2 e5e6 a1e1 d8e8 e3g3 e6c6 g3f3 f7f6 e1e3 f8f7 f1e1 c6d6 h2h4 f7e7 e1e2 h7h6 f3f5 b7b6');
INSERT INTO evaluations VALUES(6907,30,'cp 267',NULL);
INSERT INTO evaluations VALUES(6907,31,'cp 286',NULL);
INSERT INTO evaluations VALUES(6907,32,'cp 288',NULL);
INSERT INTO evaluations VALUES(6907,33,'cp 278',NULL);
INSERT INTO evaluations VALUES(6907,34,'cp 318',NULL);
INSERT INTO evaluations VALUES(6907,35,'cp 316',NULL);
INSERT INTO evaluations VALUES(6907,36,'cp 314','c4f7 e8f7 b2b4 a5c6 b4c5 d7d5 c5d6 c7d6 d2d4 g8f6 d4e5 c6e5 d1b3 f7g6 f3h4 g6h5 b3d1 c8g4 f2f3 h5h4 f3g4 h8f8 d1e2 f6g4 h2h3 g4f2 e1g1 f2h3 g2h3 f8f1 e2f1 d8f6 f1g2 f6g6 g2g6 h7g6 g1g2 g6g5 c1e3 a8c8 e3f2 h4h5 f2d4 e5c4 a2a4 g5g4 h3g4 h5g4 b1a3 g4f4 d4a7 f4e4 a3c4 c8c4');
INSERT INTO evaluations VALUES(9884,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(9884,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(9884,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(9884,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(9884,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(9884,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(9884,36,'cp 70','e1g1 a7a6 h2h3 g4f6 c1e3 d7d6 e3c5 d6c5 c4d5 f6d5 e4d5 c6d4 f3e5 e8g8 f1e1 b7b6 e5c6 d8d7 c6e7 g8h7 d1d2 c8b7 e1e4 a8e8 a1e1 f7f6 f2f3 f8f7 e7f5 e8e4 d3e4 d4f5 e4f5 d7f5 g1f2 f5h5 d2d3 h7g8 e1e8 f7f8 e8e7 f8f7 d3e4 f7e7 e4e7 b7d5');
INSERT INTO evaluations VALUES(9883,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(9883,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(9883,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(9883,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(9883,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(9883,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(9883,36,'cp 63','a7a6 h2h3 g4f6 c1e3 d7d6 e3c5 d6c5 c4d5 f6d5 e4d5 c6d4 f3e5 e8g8 f1e1 b7b6 e5c6 d8d6 c3e4 d6d7 c6d4 c5d4 d1f3 c8b7 d5d6 b7e4 e1e4 d7d6 a1e1 d6c6 e1e2 a8d8 f3f4 d8e8 e4e8 f8e8 e2e8 c6e8 f4d4 e8e1 g1h2 c7c5 d4f4 e1e2 d3d4 e2c2 d4d5 c2e2 b2b4 c5c4');
INSERT INTO evaluations VALUES(10317,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(10317,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(10317,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(10317,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(10317,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(10317,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(10317,36,'cp -7','c1g5 c5e7 g5f6 e7f6 f1b5 e8g8 b5c6 d7c6 d5f6 d8f6 d1d4 f6h6 e1g1 b7b6 f1e1 c6c5 d4d5 a8b8 d5e5 c8g4 e5g3 g4f3 g3f3 b8d8 e1e2 h6e6 h2h3 a7a5 b2b3 d8d7 e4e5 f8d8 a1e1 h7h6 e2e4 d7d2 e1e2 g7g6 g2g4 d2d4 g1h2 d4e4 e2e4 d8d2 f3c3');
INSERT INTO evaluations VALUES(11898,30,'cp 582',NULL);
INSERT INTO evaluations VALUES(11898,31,'cp 597',NULL);
INSERT INTO evaluations VALUES(11898,32,'cp 611',NULL);
INSERT INTO evaluations VALUES(11898,33,'cp 613',NULL);
INSERT INTO evaluations VALUES(11898,34,'cp 609',NULL);
INSERT INTO evaluations VALUES(11898,35,'cp 622',NULL);
INSERT INTO evaluations VALUES(11898,36,'cp 621','e4e5 f8e7 e5f6 e7f6 d4e4 d8e7 e4e7 f6e7 c3b5 e7d8 c1f4 e8g8 e1c1 c8a6 d1d7 c7c6 b5d6 a6c4 d6c4 b6b5 c4e5 c6c5 h1d1 c5c4 d7b7 d8b6 d1d7 a8d8 f4g3 h6h5 d7d8 f8d8 e5d7 f7f6 h2h4 b6d4 c2c3 d4b6 d7b6 a7b6 b7b6 d8d5 b6b7 g8h7 c1c2 h7h6 b2b3 h6g6 b3c4 b5c4');
INSERT INTO evaluations VALUES(11897,30,'cp 594',NULL);
INSERT INTO evaluations VALUES(11897,31,'cp 621',NULL);
INSERT INTO evaluations VALUES(11897,32,'cp 613',NULL);
INSERT INTO evaluations VALUES(11897,33,'cp 599',NULL);
INSERT INTO evaluations VALUES(11897,34,'cp 605',NULL);
INSERT INTO evaluations VALUES(11897,35,'cp 613',NULL);
INSERT INTO evaluations VALUES(11897,36,'cp 586','c8b7 e5f6 d8f6 d4f6 g7f6 c1f4 e8c8 e1c1 b7g2 h1g1 g2c6 c3b5 d8e8 b5a7 c8b7 a7c6 d7c6 c4d3 f8c5 d1f1 e8d8 f4e3 c5d6 h2h4 h6h5 e3d2 f6f5 a2a3 d6c5 g1g5 d8d5 d2c3 h8d8 g5h5 c5d4 c3d2 c6c5 h5h7 c5c4 d3c4 d4b2 c1b2 d5d2 h7f7 f5f4 f7f4');
INSERT INTO evaluations VALUES(11971,30,'cp -250',NULL);
INSERT INTO evaluations VALUES(11971,31,'cp -250',NULL);
INSERT INTO evaluations VALUES(11971,32,'cp -232',NULL);
INSERT INTO evaluations VALUES(11971,33,'cp -246',NULL);
INSERT INTO evaluations VALUES(11971,34,'cp -242',NULL);
INSERT INTO evaluations VALUES(11971,35,'cp -238',NULL);
INSERT INTO evaluations VALUES(11971,36,'cp -247','d8a5 b1d2 a5e5 c4e2 d7d5 c2c4 e5b8 d1b3 f8d6 b3b8 a8b8 e1g1 e8g8 b2b3 e6e5 c4d5 c6d5 f1d1 c8e6 c1b2 f8d8 a1c1 a7a5 c1c6 f6e8 d1c1 a5a4 f2f4 e5e4 b2e5 b8a8 d2f1 d6e5 f4e5 a4b3 a2b3 a8a2 c6c2 a2c2 c1c2 g8f8 f1e3');
INSERT INTO evaluations VALUES(703,30,'cp -261',NULL);
INSERT INTO evaluations VALUES(703,31,'cp -255',NULL);
INSERT INTO evaluations VALUES(703,32,'cp -249',NULL);
INSERT INTO evaluations VALUES(703,33,'cp -249',NULL);
INSERT INTO evaluations VALUES(703,34,'cp -236',NULL);
INSERT INTO evaluations VALUES(703,35,'cp -253',NULL);
INSERT INTO evaluations VALUES(703,36,'cp -240','b1d2 a5e5 c4e2 e5b8 c2c4 d7d5 d1b3 a7a5 b3b8 a8b8 e1g1 f8d6 b2b3 c8a6 a2a3 e8g8 f1e1 c6c5 c4d5 a6e2 e1e2 e6d5 c1b2 a5a4 b2f6 g7f6 b3a4 b8a8 d2f1 a8a4 e2c2 f8a8 g2g3 d5d4 f1d2 a4a3 a1a3 a8a3 d2c4 a3a1 g1g2 d6f8 c2b2 d4d3');
INSERT INTO evaluations VALUES(15138,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(15138,31,'cp -126',NULL);
INSERT INTO evaluations VALUES(15138,32,'cp -103',NULL);
INSERT INTO evaluations VALUES(15138,33,'cp -98',NULL);
INSERT INTO evaluations VALUES(15138,34,'cp -82',NULL);
INSERT INTO evaluations VALUES(15138,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(15138,36,'cp -88','d5e4 c3e4 f6e4 d3e4 d8d1 e1d1 c8g4 f1e2 f8c5 f3d2 g4e6 f2f3 c6d4 e2f1 e8c8 f1g2 a7a5 c2c3 d4b5 a2a4 b5d6 d1c2 h8e8 h1e1 h7h5 g2f1 f7f6 d2b3 c5b6 c1e3 b6e3 e1e3 b7b6 h2h4 d8d7 b3d2 e8d8 e3d3 g7g5 h4g5 f6g5 f1e2 c8b7 a1d1 g5g4 f3g4');
INSERT INTO evaluations VALUES(16896,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(16896,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(16896,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(16896,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(16896,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(16896,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(16896,36,'cp -31','c7c6 b5a4 a7a5 a4b3 d7d6 h2h3 f8e7 e1g1 e8g8 c1e3 d4b3 a2b3 d8c7 f3d2 d6d5 f2f4 d5e4 d3e4 b7b5 d1e1 e5f4 e3f4 c7b6 g1h1 c8e6 d2f3 h7h6 f4d2 c6c5 c3d5 e6d5 e4d5 f8e8 d2a5 b6a6 a5d2 a6a1 e1a1 a8a1 f1a1 f6d5 a1a7');
INSERT INTO evaluations VALUES(18836,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(18836,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(18836,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(18836,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(18836,34,'cp -65',NULL);
INSERT INTO evaluations VALUES(18836,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(18836,36,'cp -57','g8f6 c2c3 d7d5 e4e5 d5d4 e5f6 d4e3 f2e3 d8f6 d3d4 f6h6 d1e2 f8d6 b1d2 c5d4 e3d4 e8g8 e1c1 d6f4 e2e4 e6e5 d4e5 f8d8 g2g3 f4g3 f1c4 g3f4 h2h4 f4d2 d1d2 d8d2 f3d2 c8e6 c1c2 a8d8 h1d1 e6c4 d2c4 d8d1 c2d1 h6e6 d1c2 h7h5 a2a4 e6g4 e4g4 h5g4');
INSERT INTO evaluations VALUES(23281,30,'cp 159',NULL);
INSERT INTO evaluations VALUES(23281,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(23281,32,'cp 181',NULL);
INSERT INTO evaluations VALUES(23281,33,'cp 189',NULL);
INSERT INTO evaluations VALUES(23281,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(23281,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(23281,36,'cp 194','e5d4 f3d4 g8f6 d4c6 b7c6 f1d3 d8e7 c1e3 f6d5 d1e2 d5e3 f2e3 e7e5 b1d2 f8c5 e1c1 e8g8 e3e4 a8b8 d2c4 e5f6 e4e5 f6h6 c4d2 f8e8 c1b1 c5d4 b2b3 d4e5 d2f3 d7d6 f3e5 e8e5 d3c4 g8f8 e2f2 c6c5 h1f1 e5e4 f5f6 g7f6 f2f6 h6f6 f1f6 f8e7 f6f7 e7d8 c4d5 c8e6 d5e6 e4e6');
INSERT INTO evaluations VALUES(24323,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(24323,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(24323,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24323,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(24323,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(24323,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(24323,36,'cp 50','a4c6 d7c6 f1e1 f8e7 d2d4 e5d4 f3d4 c6d7 c1f4 e8g8 d1f3 b7b5 e4e5 d6e5 f4e5 b5b4 c3d5 f6d5 f3d5 c7c5 d4c6 d7c6 d5c6 f8e8 e5c7 d8c8 a1d1 e7f8 c6e8 c8e8 e1e8 a8e8 g1f1 g7g6 c7f4 c5c4 d1d7 e8c8 d7a7 f8g7 a7a6 g7b2');
INSERT INTO evaluations VALUES(26138,30,'cp -221',NULL);
INSERT INTO evaluations VALUES(26138,31,'cp -217',NULL);
INSERT INTO evaluations VALUES(26138,32,'cp -224',NULL);
INSERT INTO evaluations VALUES(26138,33,'cp -223',NULL);
INSERT INTO evaluations VALUES(26138,34,'cp -230',NULL);
INSERT INTO evaluations VALUES(26138,35,'cp -229',NULL);
INSERT INTO evaluations VALUES(26138,36,'cp -228','g8f6 d1e2 a8c8 c1e3 c5d4 e3d4 f8c5 d4c5 c8c5 b1d2 d8b6 c3c4 e8g8 c4d5 f6d5 d2b3 c5c8 a1c1 h7h6 g2g3 f8d8 c1c8 d8c8 f1d1 a7a6 e2d2 b6d8 d2d4 d8f6 d4f6 g7f6 g1f1 g8g7 d1c1 c8c1 b3c1 d5b4 a2a3');
INSERT INTO evaluations VALUES(24522,30,'cp -212',NULL);
INSERT INTO evaluations VALUES(24522,31,'cp -215',NULL);
INSERT INTO evaluations VALUES(24522,32,'cp -231',NULL);
INSERT INTO evaluations VALUES(24522,33,'cp -212',NULL);
INSERT INTO evaluations VALUES(24522,34,'cp -204',NULL);
INSERT INTO evaluations VALUES(24522,35,'cp -216',NULL);
INSERT INTO evaluations VALUES(24522,36,'cp -215','d1e2 a8c8 c1e3 f8d6 d4c5 d6c5 e3c5 d7c5 b1d2 d8b6 b2b3 e8g8 c3c4 h7h6 f1c1 f8e8 h2h3 e8d8 c4d5 f6d5 d2e4 c5e4 e2e4 b6b4 e4b4 d5b4 g2g3 f7f6 c1c8 d8c8 f3d4 e6e5 d4f5 b4c6 a1d1');
INSERT INTO evaluations VALUES(26242,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(26242,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(26242,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26242,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(26242,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(26242,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(26242,36,'cp 16','g4f3 d1f3 d8a5 b1c3 e7e6 e1g1 a8c8 b5c6 c8c6 f1e1 g8f6 c1g5 f6e4 c3e4 d5e4 f3e3 f8d6 c2c3 e8g8 g5f4 d6f4 e3f4 c6b6 e1e2 a5d5 e2d2 b6c6 a2a4 f7f5 h3h4 f8c8 h4h5 h7h6 f2f3 e4f3 f4f3 d5f3 g2f3 g8f8 g1f2 f8f7 a4a5 a7a6');
INSERT INTO evaluations VALUES(26246,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26246,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(26246,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(26246,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(26246,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(26246,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(26246,36,'cp 10','d1f3 d8a5 b1c3 e7e6 e1g1 g8f6 f3g3 a7a6 b5c6 b7c6 b2b4 a5d8 a2a3 h7h6 c3a4 f8d6 g3f3 e8g8 c1f4 f6e4 f4d6 d8d6 f3e3 a6a5 a4c5 f8b8 f1e1 a5b4 a3b4 a8a1 e1a1 b8b4 c5e4 d5e4 c2c3 b4b2 e3e4 f7f5 e4f3 b2c2 f3d3 c2b2 a1b1');
INSERT INTO evaluations VALUES(26245,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(26245,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(26245,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(26245,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(26245,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(26245,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(26245,36,'cp 25','d8a5 b1c3 e7e6 e1g1 a8c8 f1e1 g8f6 a2a3 a5b6 c1e3 f8d6 b5d3 e8g8 c3e2 c6a5 b2b3 a7a6 e3g5 f6d7 f3e3 f8e8 h3h4 d7f8 e3f3 h7h6 g5e3 b6d8 g2g3 f7f5 g1g2 d8f6 e2f4 a5c6 c2c3 f6f7');
INSERT INTO evaluations VALUES(26315,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(26315,31,'cp 176',NULL);
INSERT INTO evaluations VALUES(26315,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(26315,33,'cp 175',NULL);
INSERT INTO evaluations VALUES(26315,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(26315,35,'cp 189',NULL);
INSERT INTO evaluations VALUES(26315,36,'cp 175','h2h3 g4f3 d1f3 e7e5 c2c3 g8f6 d2d3 d8c7 b1d2 a8d8 f1e2 f6d5 e1g1 f8c5 e2d1 e8g8 d1b3 d5f4 b3c2 f4d5 a2a4 g8h8 d2c4 f7f5 c1g5 c5e7 g5d2 a7a5 c2b3 d5b6 f1e1 h7h6 f3e2 e5e4 d3e4 b6c4 b3c4 f5e4 d2e3 c6e5 c4b5');
INSERT INTO evaluations VALUES(26439,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(26439,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26439,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26439,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26439,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26439,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(26439,36,'cp -11','f1d3 f5e4 f3g5 e4d3 d1d3 g8e7 b1c3 c6b4 d3e2 e7f5 g5f3 a8c8 e1g1 b4c6 e2d3 f8e7 c3e2 c6b4 d3b3 d8b6 c1d2 b4c6 b3b6 a7b6 g2g4 f5h4 f3h4 e7h4 a1c1 h4e7 g1g2 b6b5 a2a3 h7h6 b2b3 e8d7 h2h3 b5b4');
INSERT INTO evaluations VALUES(28024,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(28024,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(28024,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(28024,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(28024,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(28024,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(28024,36,'cp 47','b1c3 g8f6 d1c2 e8g8 e3e4 b8c6 h2h3 e7e6 c1g5 h7h6 g5e3 c6b4 c2b3 d8e7 a2a3 b4c6 b3c2 e6e5 d4e5 c6e5 e1c1 d7d6 f3e5 e7e5 f1d3 f6d7 f2f4 e5e7 c3d5 b7d5 c4d5 d7c5 h1e1 a8c8 c1b1 c7c6 d5c6 c8c6 d3b5 c6c8 e3d4 g7d4');
INSERT INTO evaluations VALUES(29349,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(29349,31,'cp 187',NULL);
INSERT INTO evaluations VALUES(29349,32,'cp 203',NULL);
INSERT INTO evaluations VALUES(29349,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(29349,34,'cp 178',NULL);
INSERT INTO evaluations VALUES(29349,35,'cp 184',NULL);
INSERT INTO evaluations VALUES(29349,36,'cp 190','h2h3 g4e6 e4e5 d6e5 d1d8 e8d8 f3e5 f8d6 e5c4 d6c7 c1e3 b8a6 c4d2 c7b6 d2b3 a8c8 e1c1 d8e7 g2g4 c5c4 b3d4 h8d8 f1g2 a6c5 c1b1 e7f8 d4b5 d8d1 h1d1 f6e8 a2a3 a7a6 b5d4 e6d7 c3d5 b6a7 e3g5 f7f6 g5f4 c8d8 d5e3 d7e6 d4e6 c5e6 d1d8 e6d8');
INSERT INTO evaluations VALUES(29905,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(29905,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(29905,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29905,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29905,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(29905,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(29905,36,'cp -9','f3g3 b8c6 h2h3 g4e6 c4e6 f7e6 g1f3 d8d7 c1e3 e8c8 b1d2 f8e7 e1c1 h8g8 h3h4 h7h5 c1b1 f6g4 a2a3 c8b8 d1f1 a7a5 f3h2 g4e3 f2e3 g7g5 h4g5 g8g5 g3h3 d8h8 g2g4 d7e8 g4h5 h8h6 h2g4 h6h5 h3f3 h5h1 f1h1 g5g8 g4f6 e7f6 f3f6');
INSERT INTO evaluations VALUES(29906,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(29906,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(29906,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(29906,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(29906,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(29906,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(29906,36,'cp -17','b8c6 h2h3 g4e6 c4e6 f7e6 g1f3 d8d7 c1e3 e8c8 b1d2 f8e7 e1c1 h8g8 h3h4 h7h5 c1b1 d8f8 f3h2 d6d5 d2f3 e7d6 f3g5 f8e8 g3f3 a7a5 f3e2 a5a4 a2a3 g8f8 h2f3 d5d4 e3c1 d7e7 h1f1 f6d7 g5h7 f8h8 h7g5 e8f8');
INSERT INTO evaluations VALUES(30355,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(30355,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30355,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30355,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30355,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30355,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30355,36,'cp 0','g8f6 b1d2 c7c6 c2c4 b5c4 d5c6 b8c6 f1c4 e7e6 e1g1 f8e7 f1e1 e8g8 e4e5 f6e8 c4f1 a6a5 a2a3 d7d6 d2c4 d6e5 d1d8 a8d8 f3e5 c6e5 e1e5 b7d5 c4a5 e7f6 e5e1 d8c8 a1b1 d5a2 b1a1');
INSERT INTO evaluations VALUES(30500,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(30500,31,'cp 155',NULL);
INSERT INTO evaluations VALUES(30500,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(30500,33,'cp 154',NULL);
INSERT INTO evaluations VALUES(30500,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(30500,35,'cp 169',NULL);
INSERT INTO evaluations VALUES(30500,36,'cp 145','d4d5 g8f6 b1c3 d7d6 a2a4 a7a6 c1f4 b8d7 f1c4 g7g6 d1d2 f8g7 e1g1 f6g4 a1e1 d8c7 c4e2 e8g8 h2h3 g4e5 f3e5 d6e5 f4h2 c5c4 e1d1 f8d8 g1h1 a8c8 d2e1 e7e6 d5e6 f7e6 f2f3 d7f6 e1g3 f6h5 g3g4 h5f4');
INSERT INTO evaluations VALUES(30567,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(30567,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(30567,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(30567,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(30567,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(30567,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(30567,36,'cp 129','d1e2 b8d7 e1g1 a7a6 a2a4 f8e7 f1d1 g8f6 c2c4 e6e5 a4a5 e8g8 a5b6 c7b6 b1c3 f8e8 d4d5 g7g6 b2b4 a6a5 b4a5 a8a5 a1a5 b6a5 c1a3 b7a6 d1b1 a6c8 h2h3 d7c5 a3c5 d6c5 e2e3 e7d6');
INSERT INTO evaluations VALUES(31834,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31834,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31834,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(31834,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(31834,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(31834,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31834,36,'cp 0','g8f6 e2e3 e7e6 g1f3 b8d7 c2c4 f8e7 f1e2 e8g8 c4c5 a7a5 a2a3 f6h5 f4e5 h5f6');
INSERT INTO evaluations VALUES(29546,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29546,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29546,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29546,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29546,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29546,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(29546,36,'cp 0','e2e3 e7e6 g1f3 b8d7 c2c4 f8e7 c4c5 e8g8 f1e2 a7a5 a2a3 f6h5 f4e5 h5f6');
INSERT INTO evaluations VALUES(34804,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(34804,31,'cp 173',NULL);
INSERT INTO evaluations VALUES(34804,32,'cp 178',NULL);
INSERT INTO evaluations VALUES(34804,33,'cp 213',NULL);
INSERT INTO evaluations VALUES(34804,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(34804,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(34804,36,'cp 210','d1e2 b8c6 d4e5 d6e6 c1d2 f8b4 a2a3 b4c3 d2c3 g8e7 f2f4 e8g8 g1f3 f8d8 c3d2 e6g6 e1c1 c8g4 e2c4 g4e6 c4c5 d8d5 c5f2 a8d8 f1e2 h7h6 d2c3 d5d1 h1d1 d8d1 e2d1 g6e4 f3d2 e4a4 d2b3 b7b6 c3e1 a4c4 e1d2 g8h8 g2g4 c4d5');
INSERT INTO evaluations VALUES(36584,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(36584,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(36584,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(36584,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(36584,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(36584,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(36584,36,'cp 57','b1c3 c8g4 d3e2 c7c5 d4d5 e7e6 e1g1 e6d5 e4d5 b8d7 c1d2 a7a6 a2a4 d8c7 h2h3 g4f3 e2f3 f6e8 a1b1 b7b6 f1e1 g7d4 c3e2 d4g7 b2b3 d7e5 e2g3 f7f5 f3e2 e8f6 f2f4 e5f7 e2f3 a8e8 d1c2 e8e1 b1e1 f8e8 g1h2 e8e1 d2e1 c7e7 e1d2 f6d7 f3d1');
INSERT INTO evaluations VALUES(37441,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(37441,31,'cp -48',NULL);
INSERT INTO evaluations VALUES(37441,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(37441,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(37441,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(37441,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(37441,36,'cp -56','c7c5 e1g1 c5d4 e3d4 d7d6 c2c3 e7e6 b1d2 d8c7 f3g5 b8c6 d2e4 f6e4 g5e4 c8d7 d1e1 c6e7 c1d2 a7a6 e1f2 a8e8 a1e1 d7c6 g2g4 c7d7 e4g5 c6b5 f2e2 b5d3 e2d3 e7d5 f4f5 e6f5 e1e8 f8e8 g4f5');
INSERT INTO evaluations VALUES(38781,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(38781,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(38781,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38781,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(38781,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(38781,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38781,36,'cp -4','g8f6 g1f3 e8g8 f1d3 b8d7 e1g1 b7b6 a2a3 b4e7 e3e4 d5e4 c3e4 c8b7 d1e2 c7c5 d4c5 f6e4 d3e4 b7e4 e2e4 d7c5 e4e2 d8c7 a1d1 f8d8 d2e3 h7h6 g2g3 c7b7 f1e1 c5d7 c2c3 b6b5 d1d2 d7f6 e3d4 a8c8 d4e3 d8d2 f3d2 e7d6 e3d4');
INSERT INTO evaluations VALUES(42016,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(42016,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(42016,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42016,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42016,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42016,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42016,36,'cp 0','g1f3 e8g8 f1d3 b8d7 e1g1 b7b6 f3e5 d7e5 d4e5 f6d7 a2a3 b4e7 f2f4 c8b7 d3h7 g8h7 d1h5 h7g8 f1f3 f7f5 f3h3 d8e8 h5h8 g8f7 h8h5');
INSERT INTO evaluations VALUES(39613,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(39613,31,'cp -92',NULL);
INSERT INTO evaluations VALUES(39613,32,'cp -92',NULL);
INSERT INTO evaluations VALUES(39613,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(39613,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(39613,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(39613,36,'cp -80','g8f6 d2d3 c7c5 g1f3 b8c6 f1e2 e8g8 e1g1 d8c7 f1e1 d7d5 b1d2 a7a5 c1b2 e7e5 a2a3 c8e6 f3g5 e6d7 g5f3 f8d8 e2f1 h7h6 h2h3 d7e6 b3b4 c5b4 c3b4 a5b4 a3b4 d5d4 d1c2 c7d6 b2a3 c6b4 a3b4 d6b4 f3e5 b4c3 a1c1 c3c2');
INSERT INTO evaluations VALUES(41403,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(41403,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(41403,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(41403,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(41403,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(41403,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(41403,36,'cp 43','b1c3 e8g8 g1f3 b8c6 e1g1 e7e5 d4d5 c6d4 f4f5 c7c6 c1g5 h7h6 g5f6 g7f6 f3d2 d4b5 c3e2 b5c7 e2g3 c7e8 g1h1 f6g5 d3c4 b7b5 c4d3 c8d7 c2c4 a7a6 d2f3 c6d5 c4b5 a6b5 f5g6 f7g6 e4d5 g5f4 d3g6 f4g3 h2g3 e8f6');
INSERT INTO evaluations VALUES(46417,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(46417,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(46417,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(46417,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(46417,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(46417,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(46417,36,'cp 46','c4d5 c5d4 f1b5 c8d7 b5d7 d8d7 e1g1 f8e7 f1e1 d7d5 f3d4 b8c6 d4c6 d5c6 c1g5 a8d8 b1d2 f6d5 a1c1 c6d7 d2f3 e8g8 g5e7 d5e7 d1d7 d8d7 f3e5 d7d2 c1c7 e7c6 e5c6 b7c6 c7a7 d2b2 g2g3 h7h5 a2a4 h5h4 g3h4 c6c5 e1e7 b2b4 h4h5 g8h7');
INSERT INTO evaluations VALUES(46471,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46471,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46471,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46471,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(46471,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46471,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46471,36,'cp 0','b8c6 f1b5 f8d6 e1g1 e8g8 c1g5 c8g4 b5e2 f8e8 h2h3 g4f3 e2f3 h7h6 g5e3 d6b4 d1d3 a8c8 e3f4 b4d6 f4e3');
INSERT INTO evaluations VALUES(47092,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(47092,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(47092,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47092,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47092,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47092,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47092,36,'cp 0','d5f6 e7f6 g1f3 c7c5 f1c4 b8c6 d2d3 f8d6 e1g1 c8e6 f3d2 d8c7 c4e6 f7e6 d1h5 c7f7 h5e2 f7c7');
INSERT INTO evaluations VALUES(47080,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47080,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47080,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47080,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47080,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47080,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47080,36,'cp 0','e7f6 g1f3 c7c5 f1c4 b8c6 d2d3 f8d6 e1g1 c8e6 f3d2 d8c7 c4e6 f7e6 d1h5 c7f7 h5e2 f7c7');
INSERT INTO evaluations VALUES(47418,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(47418,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(47418,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(47418,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(47418,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(47418,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(47418,36,'cp 69','d2d4 e7e6 c1f4 f8d6 f1g2 d6f4 g3f4 e8g8 g1f3 f6d5 e2e3 d5c3 b2c3 b8d7 e1g1 a8b8 c3c4 c7c5 d4d5 e6d5 c4d5 b7b5 f3e5 d7e5 f4e5 d8b6 a1c1 c8d7 f2f4 c5c4 d1f3 b8c8 f4f5 b5b4 f3g3 c4c3 g2e4 f8d8');
INSERT INTO evaluations VALUES(47793,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(47793,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(47793,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(47793,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(47793,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(47793,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47793,36,'cp 0','c7c6 g1f3 f8e7 b1c3 e8g8 d1c2 f8e8 f1e2 e7f8 e1g1 c8g4 a2a3 g4h5 c2b3 a7a5 f1c1 b8d7 d2d3 h7h6 h2h3 h5g6 c1d1 d7b6 e3e4 b6d7 d1e1 g6h5 f3d2 h5g6');
INSERT INTO evaluations VALUES(47995,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(47995,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(47995,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(47995,33,'cp 165',NULL);
INSERT INTO evaluations VALUES(47995,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(47995,35,'cp 157',NULL);
INSERT INTO evaluations VALUES(47995,36,'cp 169','d1d5 e4g5 e5d6 g5e6 d6c7 d8d5 c4d5 e6c7 d5b3 c8e6 c1d2 b8c6 b1c3 e8c8 b3e6 c7e6 e1c1 f8c5 g1h3 f7f5 d2e3 c5e3 f2e3 g7g6 b2b3 c6e5 h3f2 h7h5 h2h3 h5h4 c3e2 h8e8 h1f1 d8d1 f1d1 e6c5 d1d6 e5c6 d6g6 e8e3 e2f4 e3e1 c1b2');
INSERT INTO evaluations VALUES(48932,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(48932,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(48932,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(48932,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(48932,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(48932,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(48932,36,'cp -58','f8c5 f1c4 b8c6 d2d3 e8g8 c1g5 h7h6 g5h4 c8g4 c3d5 f6d5 h4d8 d5e3 d8e7 e3g2 e1d2 c6e7 h1g1 c5g1 d1g1 g4f3 g1f2 f3e4 d3e4 f8d8 d2c3 g2f4 f2c5 e7c6 b2b4 d8d6 b4b5 b7b6 c4f7 g8h8 c5c4 c6d4 c3b2 c7c5 a1f1 d6f6 f7d5 a8f8 c2c3 f4d5 f1f6 d5f6 c3d4');
INSERT INTO evaluations VALUES(48986,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(48986,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(48986,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(48986,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(48986,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(48986,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(48986,36,'cp -27','b1c3 f8b4 c1d2 e8g8 g1f3 b8c6 f1e2 c8e6 e1g1 b4a5 e2b5 a5b6 b5c6 b7c6 a1c1 f8e8 f1e1 d5c3 b2c3 h7h6 d2e3 d8d6 c3c4 c6c5 d4c5 b6c5 d1d6 c7d6 f3d4 c5d4 e3d4 e6c4 e1e8 a8e8 d4a7 c4a2 a7b6 a2e6 b6e3 e6d5 h2h3 f7f6 c1d1');
INSERT INTO evaluations VALUES(48985,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(48985,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(48985,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(48985,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48985,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(48985,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48985,36,'cp -16','f8b4 c1d2 e8g8 f1e2 c7c5 g1f3 b8c6 c3d5 d8d5 d4c5 d5c5 e1g1 f8d8 d2b4 c5b4 d1c2 g7g6 a2a3 b4e7 b2b4 c8g4 f1e1 e7f6 a1d1 g8g7 c2e4 h7h5 h2h3 d8e8 e4c4 g4f3 e2f3 c6e5 c4d5 e5f3 d5f3 f6f3 g2f3 e8e1 d1e1 a7a6 e1c1 a8d8 c1c7 d8d3 g1g2 g7f6');
INSERT INTO evaluations VALUES(50464,30,'cp 126',NULL);
INSERT INTO evaluations VALUES(50464,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(50464,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(50464,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(50464,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(50464,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(50464,36,'cp 124','d7d5 g1f3 f8b4 c1d2 b4c3 d2c3 g8f6 e4e5 f6e4 a1c1 a6a5 f1d3 e4c3 c1c3 d5c4 d3c4 b7b5 c4d3 b5b4 c3c5 c8b7 e1g1 b8d7 c5c1 e8g8 d3e4 a8c8 d1d3 h7h6 f3d2 d8c7 f2f4 c6c5 f4f5 b7e4 d2e4 e6f5 f1f5');
INSERT INTO evaluations VALUES(51085,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(51085,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(51085,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(51085,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(51085,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51085,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(51085,36,'cp 37','c7c6 c4e2 d7d5 e4d5 c6d5 d2d4 c8f5 b1c3 b8c6 c3b5 c6b4 b5a3 b4c6 c1f4 a8c8 g1f3 e7e6 a3b5 f8e7 e1g1 g8f6 c2c3 h7h6 a2a4 f6h5 f4e3 e8g8 f1e1 h5f6 h2h3 d8b6 f3e5 f5e4 g1h1 b6d8 e2d3 e4d3 e5d3 f6e4 e3f4');
INSERT INTO evaluations VALUES(54301,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(54301,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(54301,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(54301,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(54301,34,'cp -69',NULL);
INSERT INTO evaluations VALUES(54301,35,'cp -86',NULL);
INSERT INTO evaluations VALUES(54301,36,'cp -70','g8h6 f1b5 b8c6 b5c6 b7c6 d2d3 c5c4 d3d4 c6c5 c3e2 c5d4 e2d4 f8c5 g1f3 d8b6 c2c3 a7a5 h2h3 e8g8 e1g1 h6f5 g1h2 f5d4 f3d4 f7f6 e5f6 f8f6 b2b3 c8a6 c1e3 a8b8 d1e1 b6c7 b3c4 a6c4');
INSERT INTO evaluations VALUES(55250,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(55250,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(55250,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(55250,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(55250,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(55250,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(55250,36,'cp 47','d4e3 f2e3 g8f6 b1c3 g7g6 h2h3 f8g7 d2d4 c7c5 d1d2 c5d4 e3d4 h7h5 f1d3 b8c6 g1f3 e8g8 c3e4 d8b6 e4f6 e7f6 e1g1 f8d8 c2c4 f6f5 c4c5 b6a5 d2e3 a5c7 d3c4 c6a5 c4b5 a5c6 g1h1');
INSERT INTO evaluations VALUES(56494,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(56494,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(56494,32,'cp -69',NULL);
INSERT INTO evaluations VALUES(56494,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(56494,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(56494,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(56494,36,'cp -82','g8f6 g1f3 f8d6 b1c3 e8g8 e2e4 d5d4 c3e2 a7a5 e2g3 a5a4 f1e2 c7c5 e1g1 d6c7 g3f5 f6e8 f3h2 a8a6 e2g4 b8d7 f2f4 b7b5 f4e5 d7e5 c1f4 e5g6 f4c7 d8c7 d1d2 e8f6 h2f3 h7h6 d2f2 f6g4 h3g4');
INSERT INTO evaluations VALUES(48888,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(48888,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(48888,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(48888,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(48888,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(48888,35,'cp -66',NULL);
INSERT INTO evaluations VALUES(48888,36,'cp -57','g1f3 f8d6 b1c3 e8g8 e2e4 d5d4 c3e2 a7a5 a3a4 b8c6 g2g3 f6d7 f1g2 d7c5 f3h4 d6e7 h4f5 e7g5 c1g5 d8g5 f5h4 g5e7 e1g1 c6b4 f2f4 f7f6 d1d2 c8d7 b2b3 c5e6 g1h2 e7c5 a1c1 a8d8 h4f3 c5d6 f4f5 e6c5 g3g4 b7b5 a4b5 d7b5');
INSERT INTO evaluations VALUES(56990,30,'cp 234',NULL);
INSERT INTO evaluations VALUES(56990,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(56990,32,'cp 229',NULL);
INSERT INTO evaluations VALUES(56990,33,'cp 224',NULL);
INSERT INTO evaluations VALUES(56990,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(56990,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(56990,36,'cp 213','f3e5 d8e7 d2d4 b8c6 b1c3 c6e5 c3d5 e7d8 d4e5 c7c6 d5c3 d7d6 e5d6 f8d6 c1g5 d8g5 d1d6 c8e6 h2h4 g5e7 d6e7 e8e7 f2f3 a7a5 f1d3 g8f6 b2b3 h8d8 e1d2 f6d7 d2e3 g7g6 h1d1 d7c5 c3e2 a5a4 b3b4 c5d7 e2d4');
INSERT INTO evaluations VALUES(326,30,'cp -598',NULL);
INSERT INTO evaluations VALUES(326,31,'cp -603',NULL);
INSERT INTO evaluations VALUES(326,32,'cp -613',NULL);
INSERT INTO evaluations VALUES(326,33,'cp -595',NULL);
INSERT INTO evaluations VALUES(326,34,'cp -621',NULL);
INSERT INTO evaluations VALUES(326,35,'cp -636',NULL);
INSERT INTO evaluations VALUES(326,36,'cp -628','c6b4 b1a3 b2a3 a1b1 a7a5 e2e4 b4a2 b1b3 a3c5 d1d5 c5d5 e4d5 a5a4 b3a3 a2b4 e1d2 b4d5 c2c4 d5f4 f1d3 f4d3 d2d3 d7d6 e5d6 c7d6 h1e1 g8e7 f3d4 c8d7 d4b5 d7b5 c4b5 e8d7 e1e4 a8a5 a3a4 a5b5 a4b4 b5b4 e4b4');
INSERT INTO evaluations VALUES(378,30,'cp -636',NULL);
INSERT INTO evaluations VALUES(378,31,'cp -640',NULL);
INSERT INTO evaluations VALUES(378,32,'cp -628',NULL);
INSERT INTO evaluations VALUES(378,33,'cp -621',NULL);
INSERT INTO evaluations VALUES(378,34,'cp -655',NULL);
INSERT INTO evaluations VALUES(378,35,'cp -640',NULL);
INSERT INTO evaluations VALUES(378,36,'cp -636','b1a3 b2a3 a1b1 a7a5 g2g3 b4a2 b1b3 a3c5 d1d2 a5a4 b3d3 g8e7 f1g2 a2b4 d3c3 c5d5 c3a3 e7c6 d2d5 b4d5 e1g1 b7b5 c2c3 d5b6 f3d4 c6d4 c3d4 a8a5 f1b1 b6c4 a3a2 a4a3 g1h1 e8e7 e2e3');
INSERT INTO evaluations VALUES(1261,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(1261,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(1261,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(1261,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(1261,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(1261,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(1261,36,'cp 35','g8f6 e1c1 f8e7 f2f3 b7b5 g2g4 c6d4 e3d4 c8b7 c1b1 e8g8 d2f2 d7d5 g4g5 f6h5 d4b6 c7e5 e4d5 b5b4 d5d6 e7d6 c3e4 d6c7 b6d4 f8d8 f1c4 e5f4 c4d3 a6a5 d4c5 b7c6 h1e1 g7g6 h2h4 a8c8 c5e7 d8d5 b2b3 h5g3 e4g3 f4g3');
INSERT INTO evaluations VALUES(662,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(662,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(662,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(662,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(662,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(662,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(662,36,'cp 27','f2f4 f8b4 f1d3 b7b5 a2a3 b4c3 d2c3 c8b7 d4c6 c7c6 c3c6 b7c6 e4e5 f6d5 e3d2 h7h5 h2h4 d5e7 e1f2 e7f5 d2a5 g7g6 a1d1 e8g8 h1h3 f8c8 c2c3 g8h8 d1g1 f5h6 h3g3 h6g4 f2e2 c6d5 a5b4 h8g8 g1d1');
INSERT INTO evaluations VALUES(1338,30,'cp 698',NULL);
INSERT INTO evaluations VALUES(1338,31,'cp 704',NULL);
INSERT INTO evaluations VALUES(1338,32,'cp 701',NULL);
INSERT INTO evaluations VALUES(1338,33,'cp 706',NULL);
INSERT INTO evaluations VALUES(1338,34,'cp 717',NULL);
INSERT INTO evaluations VALUES(1338,35,'cp 675',NULL);
INSERT INTO evaluations VALUES(1338,36,'cp 706','a1b1 c5d4 b1b2 d4c3 d2c3 b4c6 e2e4 g8e7 f1c4 a7a6 a2a4 b7b5 a4b5 a6b5 b2b5 a8a4 d1d3 e7g6 c4b3 a4b4 c3b4 f8b4 b5b4 c6b4 d3c4 b4c2 c4c2 e8g8 c2c7 f8e8 e1g1 g6e5 f2f4 e5c6 b3c4 c6d4 e4e5 d4e6 c7b6 g8h8');
INSERT INTO evaluations VALUES(1337,30,'cp 694',NULL);
INSERT INTO evaluations VALUES(1337,31,'cp 691',NULL);
INSERT INTO evaluations VALUES(1337,32,'cp 698',NULL);
INSERT INTO evaluations VALUES(1337,33,'cp 664',NULL);
INSERT INTO evaluations VALUES(1337,34,'cp 664',NULL);
INSERT INTO evaluations VALUES(1337,35,'cp 667',NULL);
INSERT INTO evaluations VALUES(1337,36,'cp 663','b2b1 d1b1 c5d4 c3b5 b4d5 g2g3 a7a6 b5d4 g8e7 f1g2 e7c6 d4f5 d5c7 e1g1 c7e6 b1b6 a6a5 d2c3 a8a6 b6b5 c6a7 b5d3 f8c5 f5g7 e6g7 e5e6 e8g8 c3g7 d7e6 d3c3 e6e5 c3e5 a6e6 e5g5 e6g6 g5c5 g8g7 f1d1 g7g8 c5a7 b7b6 a2a4 f8e8 e2e3 g6e6 g2d5 e6f6 a7b8 e8d8 c2c4');
INSERT INTO evaluations VALUES(1451,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(1451,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1451,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(1451,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(1451,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(1451,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(1451,36,'cp 17','c8d7 b1d2 a8c8 c2b1 g8f6 h2h3 c6e7 a2a4 a7a5 f1d3 e7g6 f4e5 c5d4 e3d4 g6e5 f3e5 f8d6 e1g1 g7g6 f1e1 e8g8 d2f3 b6c7 b1c2 d7e8 e5g4 f6h5 c2d2 c7e7 g2g3 h5g7 d3c2 f7f6 f3h4 g8h8 d2h6 g7f5');
INSERT INTO evaluations VALUES(1587,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(1587,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(1587,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(1587,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(1587,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(1587,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(1587,36,'cp 78','e6f5 f1g2 g8f6 d2d4 e7e6 c1f4 f8d6 d1d2 e8g8 e1c1 c6e7 c1b1 a8b8 h2h3 b7b5 g3g4 f5g6 f2f3 b5b4 c3e4 d6f4 e2f4 g6h6 e4c5 e7d5 f4d5 f6d5 d2h6 g7h6 h1e1 f8d8 g2f1 c8b7 e1e5 g8g7 e5e4 d5f6 c5b7 b8b7 e4f4');
INSERT INTO evaluations VALUES(3243,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(3243,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(3243,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(3243,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(3243,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(3243,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(3243,36,'cp 23','c8e6 d4e6 f7e6 e1c1 c6e5 f1e2 d8c8 e3h6 e5c4 e2c4 c8c4 h6g7 g8g7 e4e5 d6e5 g4g5 f6d5 c3d5 a8d8 d2c3 e6d5 c3e5 g7g8 c1b1 c4c5 h1e1 f8f3 e5e7 c5e7 e1e7 f3f7 e7e2 d5d4 e2e4 d4d3 c2d3 d8d7 e4c4 f7f2 c4c8 g8g7 b1c1');
INSERT INTO evaluations VALUES(3615,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(3615,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(3615,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(3615,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(3615,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3615,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(3615,36,'cp 25','b2b4 c5b6 a2a4 a7a6 d2d3 b6a7 b1d2 c8e6 c4b3 e6b3 d1b3 f8e8 d2f1 d6d5 f1g3 d8d6 a1a2 d6d7 a2d2 b7b5 a4b5 a6b5 d2d1 d7d6 g3f5 d6d7 c1d2 e8d8 b3c2 c6e7 f5e7 d7e7 e4d5 d8d5');
INSERT INTO evaluations VALUES(8225,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(8225,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(8225,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(8225,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(8225,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(8225,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(8225,36,'cp 52','b4c3 b2c3 c8d7 f1e2 a8c8 a3a4 c6a5 d1c2 d8b6 c1a3 b6b3 c2d2 a5c4 d2c1 c4a3 a1a3 b3b6 a4a5 b6d8 c1b2 d8c7 e1g1 e7c6 f1a1 e8g8 h2h4 h7h6 f3h2 f7f6 e5f6 f8f6 h2g4 f6f8 b2d2 a7a6 d2e3 c6e7 e3e5 e7g6 e5c7');
INSERT INTO evaluations VALUES(8842,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(8842,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(8842,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8842,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(8842,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(8842,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(8842,36,'cp 11','a7a6 f1e1 e8g8 d2f1 c8e6 c4e6 f7e6 c1e3 c5e3 e1e3 d8e8 h2h3 f6h5 f1g3 h5f4 g3e2 f4g6 d3d4 a6a5 d1c2 e8f7 a1f1 a5a4 d4e5 c6e5 f3e5 d6e5 e3g3 f8d8 b2b3 a4b3 a2b3 f7f6 f1d1 d8d1 c2d1 g8h7 b3b4 a8a2 d1d3 g6f4 e2f4 e5f4');
INSERT INTO evaluations VALUES(10369,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(10369,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(10369,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(10369,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(10369,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(10369,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(10369,36,'cp -11','c3d5 b4d2 d1d2 d7d6 d5f6 d8f6 g2g3 c8e6 f1g2 e8c8 h2h4 h7h6 g1e2 c8b8 e2c3 c6d4 e1c1 g7g5 c1b1 f6g6 d1f1 g5g4 c3e2 d4f3 g2f3 g4f3 e2c3 f7f5 d2e3 f5e4 d3e4 g6f7 b2b3 h6h5 b1b2 h8f8 a2a4 e6d7 e3d2 d7g4 c3d5');
INSERT INTO evaluations VALUES(10814,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(10814,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(10814,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(10814,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(10814,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(10814,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(10814,36,'cp -33','d8e7 e1g1 d7d5 c4d5 b4d5 b3d5 e4f6 d5b3 c7c6 b1c3 c8e6 b3a4 e8g8 f1e1 e7d6 h2h3 h7h6 e1e5 f6d7 e5e2 d7b6 a4c2 d6f4 e2e3 a8d8 a1e1 b6d5 e3e4 f4f6 c2e2 f6e7 e2d2 f8e8 e1e2 e7d6 a2a3 f7f6 c3d5 d6d5 e2e3 b7b6');
INSERT INTO evaluations VALUES(617,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(617,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(617,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(617,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(617,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(617,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(617,36,'cp -20','e1g1 d7d5 c4d5 b4d5 b3d5 e4f6 d5a5 e8g8 f1e1 e7d6 b1c3 c8g4 f3e5 g4e6 a1d1 f8d8 a5b5 d6b6 b5d3 c7c6 h2h3 d8e8 c3a4 b6c7 a2a3 a8d8 a4c5 c7c8 c5e6 c8e6 e5f3 e6d5 d1c1 h7h6 d3c2 e8e4 e1e4 d5e4 c2e4 f6e4 c1e1');
INSERT INTO evaluations VALUES(11182,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(11182,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(11182,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(11182,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(11182,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(11182,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(11182,36,'cp 18','e7e5 d4c6 b7c6 e1g1 f8e7 c3a4 e8g8 d1e2 c8e6 e3d2 d8c7 f1e1 a6a5 h2h3 f6d7 c2c4 d7b6 a4c3 a5a4 b2b3 a4b3 a2b3 h7h6 e1c1 a8a1 c1a1 f8a8 a1a8 b6a8 b3b4 e7g5 d2g5 h6g5 c3d1 a8b6 d1e3 c7a7 e2d2 b6d7 g1h2');
INSERT INTO evaluations VALUES(12207,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(12207,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(12207,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(12207,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(12207,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(12207,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(12207,36,'cp 73','c4d5 e6d5 f1e2 f8d6 d4c5 d7c5 c3d5 e8g8 d5c3 f6e4 e1g1 e4c3 b2c3 b7b6 c1a3 d8c7 a3c5 d6c5 f3d4 c8b7 e2f3 f8d8 f3b7 c7b7 d1b3 a8c8 a2a4 d8d7 g2g3 c5e7 f1d1 c8c5 d4b5 d7d1 a1d1 g7g6 c3c4 b7e4 b5d6');
INSERT INTO evaluations VALUES(12204,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(12204,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(12204,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(12204,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(12204,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(12204,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(12204,36,'cp 58','e6d5 f1e2 f8d6 d4c5 d7c5 c3d5 e8g8 d5c3 f6e4 e1g1 e4c3 b2c3 d8c7 c1a3 f8d8 d1c2 c8d7 a3c5 c7c5 f1d1 a8c8 d1d4 c5c7 a1d1 d6e7 d4d2 d7e8 d2d8 c8d8 f3d4 g7g6 g2g3 e7f8 e2f3 d8c8 c2b3 b7b6');
INSERT INTO evaluations VALUES(13545,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(13545,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(13545,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(13545,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(13545,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(13545,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(13545,36,'cp 55','c3b5 d8a5 f3d2 e6e5 d4e5 f6h5 e2e3 h5f4 e3f4 c8e6 b5c3 e8c8 d2b3 a5b6 c4d5 b6b4 d1d2 c5c4 b3d4 e6d5 e1c1 d5e6 d4c6 d8d2 c6b4 d2d1 c1d1 f8b4 d1c1 h8d8 f1e2 a7a6 h2h3 d8d4 g2g3 f7f5 h3h4 d4d8 h1f1 c8c7 a2a4 d8d4 f1d1 d4d1 c3d1 b7b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(16756,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(16756,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(16756,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(16756,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(16756,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(16756,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(16756,36,'cp -40','e4d6 b5c6 d7c6 f3e5 f8e7 f1e1 e8g8 c1f4 f8e8 h2h3 c8e6 b1c3 e7f8 c3e4 f7f6 e5f3 e6d5 f3d2 d8d7 d1g4 d7f7 e4d6 c7d6 a2a3 a8d8 g4g3 f7g6 g3g6 e8e1 a1e1 h7g6 d2e4 g8f7 g1f1 g6g5 f4d2 b7b6 a3a4');
INSERT INTO evaluations VALUES(17266,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(17266,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(17266,32,'cp -58',NULL);
INSERT INTO evaluations VALUES(17266,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(17266,34,'cp -72',NULL);
INSERT INTO evaluations VALUES(17266,35,'cp -78',NULL);
INSERT INTO evaluations VALUES(17266,36,'cp -65','f6e4 d4d5 c6b8 f1d3 e4f6 c2c4 d7d6 e1g1 g7g6 b1c3 f8g7 h2h3 e8g8 f1e1 b8a6 d3f1 e7e5 d5e6 c8e6 c1g5 h7h6 g5f4 a6c5 f3d4 e6d7 b2b4 c5e6 f4e3 a7a5 b4b5 f8e8 d1d2 e6d4 e3d4 e8e1 a1e1 d8f8 c3d1 a5a4 d1c3 d7e6 g1h1 f6d7 c3d5');
INSERT INTO evaluations VALUES(17309,30,'cp -61',NULL);
INSERT INTO evaluations VALUES(17309,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(17309,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(17309,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(17309,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(17309,35,'cp -67',NULL);
INSERT INTO evaluations VALUES(17309,36,'cp -52','d4d5 c6b8 c2c4 d7d6 f1d3 e4f6 b1c3 g7g6 h2h3 f8g7 e1g1 e8g8 f1e1 b8d7 d3f1 a7a5 c1e3 d7c5 e3c5 d6c5 d1e2 e7e6 d5e6 c8e6 a1d1 d8b8 e2e3 b8a7 c3b5 a7b6 e3f4 a8c8 f3g5 e6f5 g2g4 f5d7 g5e4 d7b5 c4b5 f6e4 e1e4 b6f6 f4e3 f6b2');
INSERT INTO evaluations VALUES(17623,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(17623,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(17623,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(17623,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(17623,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(17623,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(17623,36,'cp 36','f1d3 g8f6 e1g1 f8d6 f4g3 d6g3 h2g3 d8b6 d1c2 e8g8 b1d2 h7h6 a2a4 a6a5 f1e1 c8d7 d4c5 b6c5 e3e4 a8c8 c2d1 c5b6 a1b1 d5e4 d2e4 f6e4 d3e4 f8d8 e4c2 d7e8 d1e2 c6e7 f3e5 b6c7');
INSERT INTO evaluations VALUES(17624,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(17624,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17624,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17624,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(17624,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17624,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(17624,36,'cp 35','g8f6 e1g1 f8d6 f4g3 d6g3 h2g3 d8b6 d1c2 e8g8 b1d2 h7h6 a1e1 f8d8 a2a3 c8d7 d4c5 b6c5 e3e4 d5e4 d2e4 c5e7 e4f6 e7f6 e1d1 c6e5 d3e4 e5f3 e4f3 a8b8 d1d6 d7c6 d6d8 f6d8 f3c6 b7c6 b2b4 d8b6 a3a4 b8d8 a4a5 b6c7');
INSERT INTO evaluations VALUES(11202,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11202,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(11202,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(11202,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(11202,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(11202,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(11202,36,'cp 32','e1g1 f8d6 f4g3 d6g3 h2g3 d8b6 d1c2 e8g8 b1d2 c8d7 a2a3 h7h6 d4c5 b6c5 e3e4 c5d6 c3c4 d6c7 c4d5 e6d5 e4d5 f6d5 a1c1 a8c8 b2b4 c6e5 c2b2 e5f3 g2f3 c7b6 d2e4 c8c1 f1c1 f8d8 d3c4 d5f6 e4f6 b6f6 b2f6');
INSERT INTO evaluations VALUES(17671,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(17671,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(17671,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(17671,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(17671,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17671,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17671,36,'cp 0','f8e7 d2d4 c5d4 f3d4 g8f6 c1e3 e8g8 h2h3 d8c7 d1e2 c8d7 f2f4 a8e8 a1d1 d6d5 e4e5 c6d4 e3d4 f6h5 e2h5 c7c4 b2b3 c4c6 d1d3 e8c8 h5f3 f7f5 g1h2 h7h6 f3e3 e7b4 c3e2 c6c2 f1c1 c2a2 c1a1 a2c2');
INSERT INTO evaluations VALUES(19414,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(19414,31,'cp 179',NULL);
INSERT INTO evaluations VALUES(19414,32,'cp 158',NULL);
INSERT INTO evaluations VALUES(19414,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(19414,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(19414,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(19414,36,'cp 162','d4d5 c6a5 e2e4 g7g6 b1c3 d7d6 f2f4 f8g7 g1f3 e7e6 f1e2 g8h6 h2h3 f7f5 f3g5 e8g8 e1g1 d8e7 d5e6 g7c3 b2c3 f5e4 f4f5 h6f5 g2g4 e4e3 h3h4 f5g7 f1f7 f8f7 e6f7 g8h8 d1d5 c8e6 g5e6 e7e6 d5e6 g7e6 c1e3 h8g7');
INSERT INTO evaluations VALUES(19510,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(19510,31,'cp 182',NULL);
INSERT INTO evaluations VALUES(19510,32,'cp 174',NULL);
INSERT INTO evaluations VALUES(19510,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(19510,34,'cp 169',NULL);
INSERT INTO evaluations VALUES(19510,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(19510,36,'cp 155','c6a5 e2e4 g7g6 f2f4 f8g7 b1c3 d7d6 g1f3 e7e6 f1e2 g8h6 e1g1 e8g8 d1e1 f8e8 c1d2 e6d5 c4d5 a7a6 f4f5 g6f5 e1g3 g8h8 f3g5 d8f6 a1d1 b7b5 d2f4 f6g6 g3h4 f7f6 g5e6 c8e6 d5e6 e8e6 f4d6 a5c6 e2h5');
INSERT INTO evaluations VALUES(19544,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(19544,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(19544,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(19544,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(19544,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(19544,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(19544,36,'cp 95','d4d5 d8a5 b1c3 c6b4 f1c4 g8f6 g1e2 e7e6 e1g1 f6d5 c4d5 b4d5 c3d5 a5d5 d1d5 e6d5 e2c3 d5d4 c3b5 e8d8 c1f4 d7d6 b5d4 c8d7 a1d1 f8e7 c2c4 g7g5 f4e3 h8e8 b2b3 d8c7 d1d2 a8d8 d4b5 d7b5 c4b5 d6d5 e3a7');
INSERT INTO evaluations VALUES(20175,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20175,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20175,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20175,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20175,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20175,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(20175,36,'cp 0','g8f6 e1g1 c8d7 b1d2 a8c8 a2a4 a7a6 b5e2 d8b6 d4c5 f8c5 c2c3 f6h5 a4a5 b6b2 f4e5 e8g8 a1b1 b2a2 d2b3 c5e7 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(13486,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(13486,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(13486,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(13486,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(13486,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13486,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13486,36,'cp 12','e1g1 c8d7 b1d2 f8e7 d4c5 e7c5 b5d3 e8g8 c2c4 d5c4 d3c4 a8c8 a1c1 c5e7 c4e2 f6h5 f4g5 e7g5 f3g5 d8g5 d2e4 g5d5 e2h5 d5e4 d1d7 g7g6 h5f3 e4b4 f3c6 c8c6 b2b3 c6a6 c1c7 a6a2 c7b7 b4d2 b7a7 d2d7');
INSERT INTO evaluations VALUES(20471,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(20471,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(20471,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(20471,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(20471,34,'cp 89',NULL);
INSERT INTO evaluations VALUES(20471,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(20471,36,'cp 104','g8f6 d2d4 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6g6 e4g6 h7g6 a1d1 c8f5 c2c3 f5e4 d5e3 e4f3 g2f3 c6e5 f3f4 e5f3 g1g2 f3d2 f1e1 a7a6 b5a4 b7b5 a4b3 c5e3 f2e3 b5b4 c3b4 a8c8 g2g3 d2b3 a2b3 d8d1 e1d1');
INSERT INTO evaluations VALUES(13713,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(13713,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(13713,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(13713,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(13713,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(13713,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(13713,36,'cp 99','d2d4 f8d6 c1g5 e8g8 g5f6 d8f6 c3d5 f6h6 d4c5 d6c5 d1d3 f8d8 d3e4 h6g6 e4g6 h7g6 a1d1 c8g4 b5c6 g4f3 g2f3 b7c6 d5e3 f7f6 b2b3 g6g5 d1d8 a8d8 f1d1 d8d1 e3d1 g8f7 g1f1 f7e6 f1e2 f6f5 h2h3 c5e7 c2c3 e7f6 b3b4 e6e5 a2a4 e5e6 e2d3 e6e5 d3c4 e5f4 c4c5');
INSERT INTO evaluations VALUES(20567,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20567,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20567,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(20567,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(20567,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(20567,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(20567,36,'cp 12','g8f6 b5d3 d6d5 e4d5 f8b4 c2c3 d8d5 d1e2 b4e7 b1d2 c8g4 d3c4 d5d7 f3e5 c6e5 e2e5 e8g8 c4b5 d7d8 e1g1 a7a6 b5a4 e7d6 e5d4 d8c7 d2e4 d6h2 g1h1 f6h5 d4c5 h2f4 c5c7 f4c7 c1e3 h5f4 f2f3 g4f5 f1d1 f5e4 f3e4 a8d8 d1d8 f8d8 a1d1 f4d3');
INSERT INTO evaluations VALUES(21669,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(21669,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(21669,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(21669,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(21669,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(21669,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(21669,36,'cp 122','c4d5 e5d4 d5c6 d4c3 d1d8 e8d8 c6b7 c8b7 b2c3 g8f6 e2e3 f8d6 f2f3 d8e7 g1e2 c7c5 e3e4 a8b8 e1f2 h8d8 e2g3 g7g6 f1c4 f6d7 c1e3 b7c6 a1b1 b8b1 h1b1 d8b8 b1b8 d7b8 g3e2 b8d7 c4b3 d7e5');
INSERT INTO evaluations VALUES(25130,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(25130,31,'cp 84',NULL);
INSERT INTO evaluations VALUES(25130,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(25130,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(25130,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(25130,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(25130,36,'cp 47','f3e5 a7a6 b5c6 b7c6 g2g4 f5e4 e5c6 d8d6 c3e4 f6e4 c6e5 f8e7 e1g1 c7c5 f2f3 e4f6 b2b3 f6d7 f3f4 d7e5 d4e5 d6b6 c2c4 d5c4 b3c4 e8g8 d1b3 b6b3 a2b3 f8b8 f1d1 b8b3 d1d7 g8f8 c1d2 b3b2 d2c3 b2c2');
INSERT INTO evaluations VALUES(30229,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(30229,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(30229,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(30229,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(30229,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(30229,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(30229,36,'cp 27','f1d3 f5d3 d1d3 c7c6 e1g1 f8e7 c1g5 b8d7 a1e1 e8g8 e1e2 e7b4 f3e5 d8c7 e5d7 f6d7 f1e1 f7f6 g5h4 f8f7 h4g3 c7c8 h2h4 d7b6 a2a3 b4f8 c3d1 c8d7 c2c3 f7e7 d1e3 a8e8 b2b3 e7e6 a3a4 b6c8 g3f4 c8d6 f2f3 e6e7');
INSERT INTO evaluations VALUES(31510,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(31510,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(31510,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(31510,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(31510,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(31510,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(31510,36,'cp -26','b1d2 d8b6 d2b3 c6c5 d4c5 f8c5 b3c5 b6c5 c2c3 b8c6 f4e3 c5a5 f2f4 d5d4 e3d4 e8c8 g1f3 f5g4 b2b4 a5c7 f1e2 g8h6 e1g1 c8b8 a1c1 h6f5 a2a4 c6d4 f3d4 g4e2 d1e2 c7b6 e2b5 f5d4 b5b6 a7b6 c3d4 d8d4 f4f5 d4b4 f5e6 f7e6 f1f7 h8c8 c1d1 b4a4');
INSERT INTO evaluations VALUES(32013,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(32013,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(32013,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(32013,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(32013,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(32013,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(32013,36,'cp 81','e2f3 c7c6 d2d4 e7e6 d1d3 d8c7 c1d2 b8d7 e1c1 e8c8 c3e4 g8f6 e4f6 d7f6 c1b1 d8d7 h2h4 f8d6 g2g3 c6c5 c2c3 h8d8 d2e3 c5c4 d3e2 c8b8 h4h5 h7h6 h1e1 d6e7 f3g2 f6d5 e3c1 a7a6 b1a1 b8a7 g2d5 e6d5 e2g4');
INSERT INTO evaluations VALUES(32010,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(32010,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(32010,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(32010,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(32010,34,'cp 88',NULL);
INSERT INTO evaluations VALUES(32010,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(32010,36,'cp 87','c7c6 d2d4 e7e6 d1d3 d8c7 g2g3 b8d7 c3e4 d7f6 c1f4 c7a5 c2c3 f6e4 d3e4 a8d8 e1g1 f8d6 f4e3 g8f6 e4c2 e8g8 a2a3 f8e8 f1d1 h7h6 c3c4 a5f5 c2f5 e6f5 a1c1 g7g5 b2b4 f5f4 g3f4 d6f4 e3f4 g5f4');
INSERT INTO evaluations VALUES(31225,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(31225,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(31225,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(31225,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(31225,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(31225,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(31225,36,'cp 90','d2d4 e7e6 d1d3 d8c7 g2g3 b8d7 c3e4 g8f6 c1f4 c7a5 c2c3 f6e4 d3e4 d7f6 e4c2 f8e7 e1g1 e8g8 a2a4 f6d5 f4d2 c6c5 f3d5 e6d5 d4c5 a5c5 d2e3 c5c6 a4a5 b7b6 f1e1 e7c5 c2a4 b6b5 a4g4 c5e3 e1e3');
INSERT INTO evaluations VALUES(32678,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(32678,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32678,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(32678,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(32678,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(32678,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(32678,36,'cp 41','h2h3 g4f3 e2f3 d6d5 f1e2 b8c6 c2c3 g7g6 c1f4 e8c8 b1d2 e7e6 f3e3 f8d6 f4d6 d8d6 e1g1 h8e8 f1e1 f7f5 d2e4 d5e4 f2f3 e6f7 f3e4 e8e4 e3f2 d6d8 e2f3 e4e1 a1e1 d8e8 a2a4 c6a5 e1e8 f7e8 f2e2 e8e2 f3e2 a5c6 g1f2 c6e7');
INSERT INTO evaluations VALUES(33708,30,'cp -62',NULL);
INSERT INTO evaluations VALUES(33708,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(33708,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(33708,33,'cp -100',NULL);
INSERT INTO evaluations VALUES(33708,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(33708,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(33708,36,'cp -83','e1c1 c7c6 a2a3 b4a5 f3g3 c8f5 e3d2 a5d8 g1f3 b8d7 f1e2 e7e6 f3d4 f5g6 h2h4 d8b8 d2f4 e6e5 h4h5 e5f4 g3h3 g6e4 h1e1 f8c5 e2c4 e8g8 c3e4 f6e4 h3d7 e4f6 d7f5 c5d4 d1d4 f8e8 e1e8 b8e8 c1b1 h7h6 d4f4 a8d8 f4f1 e8e7 g2g4 e7d7 f5f4 d8e8 c4b3 e8e7 f4b8 g8h7 b1a2 d7g4 b8a7 f6h5 f1f7 e7f7 b3f7');
INSERT INTO evaluations VALUES(34181,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(34181,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(34181,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(34181,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(34181,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(34181,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(34181,36,'cp 133','d7d6 a2a4 c8g4 c4e2 f6d8 d2d4 g4f3 e2f3 f8e7 f3e2 b8d7 d4d5 g8f6 e2c4 d7b6 c4b3 a7a5 d1e2 e8g8 c1e3 a8c8 a1d1 c6c5 b3c4 f6h7 c4b5 e7g5 g2g3 g5e3 e2e3 h7f6 g1h1 c5c4 d1e1 f6d7 f2f4 d8e7 f1g1 e7f6 e1f1 d7c5 f4f5');
INSERT INTO evaluations VALUES(34409,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(34409,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(34409,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(34409,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(34409,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(34409,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(34409,36,'cp 126','a2a4 c8g4 c4e2 g4f3 e2f3 f6d8 d2d4 f8e7 d4d5 g8f6 f3e2 e8g8 e2c4 b8d7 c1e3 a7a5 d1e2 d7b6 c4a2 a8c8 f1d1 c6d5 a2d5 b6d5 c3d5 c8c6 f2f3 f6d5 d1d5 e7g5 e3f2 d8c7 c2c3 g5e7 g2g3 f8c8 g1g2 c6c4 e2d3 c4c6 a1d1');
INSERT INTO evaluations VALUES(34407,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(34407,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(34407,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(34407,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(34407,34,'cp 67',NULL);
INSERT INTO evaluations VALUES(34407,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(34407,36,'cp 93','c1e3 b8d7 d1d3 f8d6 e1c1 d6c7 c1b1 f6e7 h2h3 g8f6 g2g4 d7c5 d3e2 f6e4 c3e4 c5e4 h1e1 c8e6 c4e6 e7e6 e3h6 e4c3 b2c3 h8h6 f3d4 e5d4 e2d3 c7e5 c3d4 e8c8 e1e5 e6d6 d3e3 d6b4 b1c1 h6d6 a2a3 b4c4 e5e7 d6d7 e7d7 d8d7 h3h4 f7f6 d1d3 c4a2 c1d2 a2d5');
INSERT INTO evaluations VALUES(37318,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37318,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37318,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37318,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37318,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37318,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37318,36,'cp 0','e5f4 b8c6 f1c4 d7d5 c3d5 f6d5 c4d5 c6d4 d5b3 c8e6 g1f3 e7g5 f4e5 g5f6 e5f4');
INSERT INTO evaluations VALUES(38606,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(38606,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(38606,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(38606,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(38606,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(38606,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(38606,36,'cp -9','c7c6 b5d3 g8f6 d1e2 d6e7 e1g1 e8g8 c1g5 h7h6 f1e1 e7d6 g5h4 c8g4 b1d2 b8d7 h2h3 g4f3 d2f3 f8e8 e2e8 d8e8 e1e8 a8e8 h4f6 d7f6 g2g3 g8f8 g1g2 g7g6 a1e1 f6d7 c2c3 e8e7 e1e7 f8e7 f3e1 f7f5 d3c2 e7f7 f2f4 d7f6 e1d3 f6e4 g3g4');
INSERT INTO evaluations VALUES(38074,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38074,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(38074,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(38074,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(38074,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(38074,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(38074,36,'cp -6','b5d3 g8f6 d1e2 d6e7 e1g1 e8g8 c1g5 h7h6 f1e1 e7d6 g5h4 c8g4 b1d2 b8d7 h2h3 g4f3 d2f3 f8e8 e2e8 d8e8 e1e8 a8e8 h4f6 d7f6 g2g3 g8f8 g1g2 g7g6 a1e1 f6d7 c2c3 e8e7 d3c2 f7f5 e1e7 f8e7 f3e1 e7f7 f2f4 d7f6 e1d3 f6e4 a2a4 f7e7');
INSERT INTO evaluations VALUES(39055,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(39055,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(39055,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39055,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(39055,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(39055,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(39055,36,'cp 15','b1c3 g8e7 d2d3 d7d6 a1b1 e8g8 e2e4 c7c5 b2b4 b8c6 b4b5 c6d4 g1e2 a7a6 b5b6 c8g4 e1g1 d8d7 f2f3 g4h3 g2h3 d7h3 g1h1 h3d7 a2a4 h7h6 e2d4 e5d4 c3d5 e7d5 c4d5 f7f5 e4f5 g6f5 h1g2 a8e8 a4a5');
INSERT INTO evaluations VALUES(42403,30,'cp 262',NULL);
INSERT INTO evaluations VALUES(42403,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(42403,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(42403,33,'cp 281',NULL);
INSERT INTO evaluations VALUES(42403,34,'cp 260',NULL);
INSERT INTO evaluations VALUES(42403,35,'cp 273',NULL);
INSERT INTO evaluations VALUES(42403,36,'cp 273','d2d4 b8c6 c1g5 d6e7 g5h6 g7h6 d4d5 c6d4 f3e5 d7d6 d1d4 e7f6 b1c3 f6e5 d4d2 d8h4 g2g3 e5c3 g3h4 c3d2 e1d2 f7f5 a1e1 e8e7 f2f4 f5e4 e1e4 e7f6 h1e1 c8f5 e4e7 h8g8 c4d3 g8g2 d2c3 a8c8 d3f5 f6f5 e7f7 f5g6 e1e7 g2h2 f4f5 g6h5 e7c7 c8e8 f7h7 h2h4 c3b3');
INSERT INTO evaluations VALUES(42924,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(42924,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(42924,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(42924,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(42924,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(42924,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(42924,36,'cp 60','e2e3 c5e7 f2f4 e5f4 g3f4 a7a5 g1f3 e8g8 e1g1 b8a6 b2b3 c7c6 c1b2 f6e4 f3d4 e4c3 b2c3 a6c7 d1c2 d7d6 a2a3 e7f6 b3b4 c7e6 d4e2 f6c3 c2c3 c8d7 d2d4 a5b4 a3b4 d8h4 a1a8 f8a8 b4b5 c6b5 c4b5');
INSERT INTO evaluations VALUES(42923,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(42923,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(42923,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(42923,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(42923,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(42923,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(42923,36,'cp 51','c5e7 f2f4 e5f4 g3f4 a7a5 g1e2 b8a6 e1g1 c7c6 e2d4 d7d5 b2b3 e8g8 c4d5 c6d5 c1b2 a6b4 d1e2 f6e4 a2a3 b4c6 c3b5 c8d7 g1h1 d8b6 g2f3 a8e8 f1g1 f8f7 g1g2 g7g6 d2d3 e4f6 e2d2 e7c5 a1d1 c6d4 b5d4 c5f8 d1g1 f8c5');
INSERT INTO evaluations VALUES(44722,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(44722,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(44722,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(44722,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(44722,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(44722,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(44722,36,'cp 68','f8g7 f1e2 e8g8 h2h3 e7e5 d4d5 a7a5 c1e3 c8d7 e1g1 b8a6 a2a3 a6c5 d1c2 c6d5 e3c5 d6c5 c4d5 f6e8 a3a4 e8d6 c3b5 d7b5 e2b5 d8c7 f3d2 f7f5 a1e1 a8d8 d2c4 d6c4 b5c4 c7d6 e1d1 b7b6 f1e1 f5f4 d1d3 g7f6 c4b5 h7h5 e1d1');
INSERT INTO evaluations VALUES(49250,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(49250,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(49250,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(49250,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(49250,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(49250,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(49250,36,'cp 59','g1f3 d7d6 d2d4 b8d7 a2a4 f8e7 e1g1 e8g8 c1e3 c7c6 f1e1 f6g4 e3c1 e7h4 e1f1 h4e7 h2h3 g4f6 f1e1 f8e8 a4a5 a8b8 c4f1 b7b5 a5b6 a7b6 d4d5 c6d5 e4d5 d7f8 f3d2 c8d7 d2e4 b6b5 e4f6 e7f6 c1e3 f6e7 a1a6 f7f5 b2b4 f8g6 d1h5');
INSERT INTO evaluations VALUES(49969,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(49969,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(49969,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(49969,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(49969,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(49969,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(49969,36,'cp 24','e2e3 f8b4 b1d2 e8g8 a2a3 b4e7 g1f3 c7c5 d4c5 a7a5 b2b4 b7b6 c5b6 a5b4 a3b4 a8a1 d1a1 e7b4 f1e2 d8b6 e1g1 b8c6 a1b2 b4c5 b2b5 c8a6 b5b6 c5b6 f1a1 a6c8 g5f6 g7f6 d2b3 c8b7 c4c5 b6c7 f3d4 c6d4 b3d4 c7e5 a1a7 e5d4 e3d4 b7c6 g1f1 f8b8 e2f3 c6f3 g2f3 b8b1 f1e2');
INSERT INTO evaluations VALUES(50604,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(50604,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(50604,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(50604,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(50604,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(50604,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(50604,36,'cp 30','e7e6 f1g2 b8c6 g1e2 d7d6 d2d4 c5d4 e2d4 c8d7 e1g1 f8e7 c3e2 g8f6 c2c4 e8g8 a2a4 d8c7 b2b3 f8d8 c1e3 d7e8 h2h3 f6d7 d1c2 a8b8 f1c1 h7h6 a1b1 b8c8 d4c6 b7c6 b3b4 a6a5 b4b5 c6b5 a4b5 d7e5 c2d2 e5c4');
INSERT INTO evaluations VALUES(50525,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(50525,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(50525,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(50525,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(50525,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(50525,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(50525,36,'cp 29','f1g2 b8c6 g1f3 d7d6 e1g1 g8f6 d2d4 c5d4 f3d4 c8d7 d4c6 d7c6 c1f4 d8c7 f1e1 f8e7 e4e5 d6e5 f4e5 c7c8 c3e4 f6e4 g2e4 e8g8 d1f3 c6e4 f3e4 f7f6 e5c3 e6e5 e1d1 b7b5 a2a4 b5b4 c3b4 e7b4 e4b4 c8c2 a1c1 c2f5 b4a3 f8d8');
INSERT INTO evaluations VALUES(50526,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(50526,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(50526,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(50526,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50526,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(50526,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(50526,36,'cp 33','b8c6 g1f3 d7d6 d2d4 c5d4 f3d4 c8d7 e1g1 g8f6 d4c6 d7c6 c1f4 f8e7 e4e5 d6e5 g2c6 b7c6 f4e5 e8g8 f1e1 d8a5 d1f3 a8d8 a2a3 a5b6 b2b3 f6d7 e5f4 d7f6 c3a4 b6a5 g1g2 a5f5 e1e2 h7h5 f4c7 f5f3 g2f3 d8d7 c7f4 f6d5 f4d2 f8b8 h2h4 b8b5 f3g2 e7f6 a1c1 f6d4 c1d1');
INSERT INTO evaluations VALUES(17388,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(17388,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(17388,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(17388,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(17388,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(17388,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(17388,36,'cp 27','g1e2 d7d6 d2d4 c5d4 e2d4 c8d7 d4c6 d7c6 c1f4 g8f6 e1g1 a8c8 f1e1 h7h5 c3d5 f6g4 f4d2 c6d7 d5f4 g7g6 f4d3 g4e5 h2h4 f8e7 a1c1 b7b5 d2f4 e5d3 c2d3 e6e5 f4e3 c8c1 d1c1 e8g8 d3d4');
INSERT INTO evaluations VALUES(17389,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(17389,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(17389,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(17389,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(17389,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(17389,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(17389,36,'cp 25','d7d6 d2d4 c5d4 e2d4 c8d7 d4c6 d7c6 e1g1 g8f6 c1f4 a8c8 d1d2 f6g4 f1d1 h7h5 b2b3 g4e5 h2h4 d8a5 c3e2 a5d2 f4d2 f8e7 c2c4 f7f6 d2b4 g7g5 h4g5 f6g5 b4d6 e7d6 d1d6 e8e7 c4c5 h5h4 e2d4 c6d7 d6b6 h4g3 f2g3 c8c5');
INSERT INTO evaluations VALUES(50986,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(50986,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(50986,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(50986,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(50986,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(50986,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(50986,36,'cp 97','f1c4 c7c5 g1f3 c5d4 f3d4 b7b5 c4b3 c8b7 e1g1 b8c6 d4c6 d8d1 f1d1 b7c6 c1e3 f8e7 f2f3 g8f6 c3e2 e8g8 e2d4 c6d7 g2g4 h7h6 h2h4 f8d8 g4g5 f6h5 g1f2 a6a5 e4e5 a8c8 a1c1 g7g6 b3c2 h6g5 h4g5 h5g7 c2e4 c8c1 d1c1 a5a4 f3f4 g7f5 c1c7 f5e3 f2e3 b5b4 e4d3');
INSERT INTO evaluations VALUES(51903,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(51903,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(51903,32,'cp -138',NULL);
INSERT INTO evaluations VALUES(51903,33,'cp -141',NULL);
INSERT INTO evaluations VALUES(51903,34,'cp -135',NULL);
INSERT INTO evaluations VALUES(51903,35,'cp -151',NULL);
INSERT INTO evaluations VALUES(51903,36,'cp -128','b8c6 g1f3 d7d6 e5d6 f8d6 g2g3 e6e5 d2d3 g8e7 f1g2 c8f5 b1c3 e8g8 c1e3 b7b6 d1d2 d6c7 e1c1 d8d7 f3h4 f5g4 f2f3 g4e6 c1b1 a8d8 a2a3 f8e8 f3f4 e6g4 g2f3 g4f3 h4f3 h7h6 h1e1 e5f4 e3f4');
INSERT INTO evaluations VALUES(18709,30,'cp -121',NULL);
INSERT INTO evaluations VALUES(18709,31,'cp -124',NULL);
INSERT INTO evaluations VALUES(18709,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(18709,33,'cp -117',NULL);
INSERT INTO evaluations VALUES(18709,34,'cp -121',NULL);
INSERT INTO evaluations VALUES(18709,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(18709,36,'cp -115','g1f3 d7d6 e5d6 f8d6 g2g3 g8e7 f1g2 e8g8 b1c3 e6e5 d2d3 f7f6 f3d2 b7b6 d2e4 c8b7 e1g1 d6b8 f1e1 d8d7 c3a4 d7c8 a4c3 f6f5 e4d2 h7h6 a2a3 c6d4 d2f3 f5f4 f3d4 b7g2 g1g2 c5d4 c3e4 c8d7');
INSERT INTO evaluations VALUES(18708,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(18708,31,'cp -114',NULL);
INSERT INTO evaluations VALUES(18708,32,'cp -126',NULL);
INSERT INTO evaluations VALUES(18708,33,'cp -122',NULL);
INSERT INTO evaluations VALUES(18708,34,'cp -128',NULL);
INSERT INTO evaluations VALUES(18708,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(18708,36,'cp -125','d7d6 e5d6 f8d6 g2g3 g8e7 f1g2 e8g8 b1c3 e6e5 d2d3 c8f5 c1e3 b7b6 f3h4 f5e6 d1d2 a8b8 e1g1 d8d7 a2a3 b8d8 f2f4 e5f4 g3f4 e7f5 h4f5 e6f5 a1e1 h7h6 e3f2 g8h8 g1h1 f7f6 c3d5 c6e7 b2b4 f8e8 b4c5 d6c5');
INSERT INTO evaluations VALUES(53686,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(53686,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(53686,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(53686,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(53686,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(53686,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(53686,36,'cp 6','c6d5 f1d3 b8c6 g1f3 f8b4 b1c3 g8e7 e1g1 e8g8 c1g5 f7f6 g5f4 c8g4 d3e2 b4c3 b2c3 c6a5 f1e1 a8c8 h2h3 g4h5 d1c2 f8e8 e2b5 a5c6 g2g4 h5g6 c2b2 d8b6 b2b3 c6a5 b3a4 e7c6 f4g3 h7h6 e1e8 c8e8 b5f1 c6e7');
INSERT INTO evaluations VALUES(54642,30,'cp 305',NULL);
INSERT INTO evaluations VALUES(54642,31,'cp 327',NULL);
INSERT INTO evaluations VALUES(54642,32,'cp 325',NULL);
INSERT INTO evaluations VALUES(54642,33,'cp 338',NULL);
INSERT INTO evaluations VALUES(54642,34,'cp 335',NULL);
INSERT INTO evaluations VALUES(54642,35,'cp 344',NULL);
INSERT INTO evaluations VALUES(54642,36,'cp 360','d4e5 f6e5 e4d5 g8f6 f3e5 f8d6 e5c4 e8g8 c4d6 d8d6 d5c6 d6e6 c1e3 b8c6 d1e2 f6g4 e1c1 g4e3 e2e3 e6e3 f2e3 c6e5 f1e2 c8g4 d1d2 g4e2 d2e2 f8f7 h1d1 a8f8 b2b3 f7f2 e2f2 f8f2 d1d2 f2f1 c1b2 e5g4 d2d7 f1f7 d7f7');
INSERT INTO evaluations VALUES(54637,30,'cp 358',NULL);
INSERT INTO evaluations VALUES(54637,31,'cp 334',NULL);
INSERT INTO evaluations VALUES(54637,32,'cp 340',NULL);
INSERT INTO evaluations VALUES(54637,33,'cp 353',NULL);
INSERT INTO evaluations VALUES(54637,34,'cp 355',NULL);
INSERT INTO evaluations VALUES(54637,35,'cp 359',NULL);
INSERT INTO evaluations VALUES(54637,36,'cp 344','f6e5 e4d5 f8b4 f3e5 g8f6 d5c6 d8d1 e1d1 b8c6 e5c6 b7c6 c1d2 e8g8 f2f3 b4d6 a2a4 a7a5 f1c4 g8h8 c4d3 c8e6 d1c1 d6c7 c3e4 f6d5 e4c5 e6f7 d3e4 h7h5 c5d3 c7d6 g2g3 h8g8 h1e1 h5h4 b2b3 h4g3 h2g3 d6g3 e1g1');
INSERT INTO evaluations VALUES(55302,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(55302,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(55302,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(55302,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(55302,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(55302,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(55302,36,'cp 4','c7c5 g1f3 b8c6 f1b5 c8d7 e1g1 e7e6 f1e1 g8f6 b2b3 d8c7 c1b2 f8e7 b5c6 d7c6 f3e5 e8g8 d2d3 a8d8 a2a4 a7a6 e5c6 c7c6 c3e4 f6d5 d1g4 g7g6 g4g3 f7f6 f2f4 c6c7 e1f1 d5b4 g3f2 d8d7 a1e1 g8h8 d3d4 c5d4 b2d4');
INSERT INTO evaluations VALUES(55895,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(55895,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(55895,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55895,33,'cp -54',NULL);
INSERT INTO evaluations VALUES(55895,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(55895,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(55895,36,'cp -57','d5e4 d1a4 b8c6 a4e4 g8f6 e4a4 f8d6 f1e2 e8g8 d2d3 e6e5 e1g1 c8d7 a4b3 a7a5 b1d2 a5a4 b3c2 h7h6 d2e4 d7f5 c1d2 d6e7 e4g3 f5h7 b2b4 a4b3 a2b3 a8a1 f1a1 d8d5 c3c4 d5d7 d2c3 c6d4 c3d4 e5d4');
INSERT INTO evaluations VALUES(56081,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(56081,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(56081,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(56081,33,'cp -50',NULL);
INSERT INTO evaluations VALUES(56081,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(56081,35,'cp -51',NULL);
INSERT INTO evaluations VALUES(56081,36,'cp -42','d1a4 b8c6 a4e4 g8f6 e4a4 f8d6 f1e2 e6e5 d2d3 e8g8 e1g1 c8d7 a4d1 h7h6 f1e1 c6e7 b1d2 e7g6 d2c4 f8e8 e2f1 d7c6 b2b4 a7a6 a2a4 d8d7 c4d6 c7d6 c3c4 d6d5 c1b2 d5d4 b4b5 a6b5 a4b5 a8a1 b2a1 c6f3 d1f3 e8a8');
INSERT INTO evaluations VALUES(55937,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(55937,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(55937,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55937,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55937,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(55937,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55937,36,'cp -8','g8f6 g1f3 f8e7 d2d4 e8g8 f1g2 b7b5 c2c4 b5c4 b3c4 b8d7 c4d5 f6d5 b1c3 c8b7 e1g1 c7c5 d4c5 a8c8 c3d5 b7d5 a1c1 d7c5 d1d4 e7f6 d4b4 c5e4 c1c8 d8c8 b2f6 e4f6 b4b2 c8a8 f1c1 f8b8 b2d2 h7h6 f3e1 d5g2 e1g2 a8e4 g2e3');
INSERT INTO evaluations VALUES(56722,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(56722,31,'cp 153',NULL);
INSERT INTO evaluations VALUES(56722,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(56722,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(56722,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(56722,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(56722,36,'cp 159','b1c3 b8c6 d3d4 g8f6 c4b5 c8d7 e1g1 g7g5 d4e5 d6e5 b5c6 d7c6 f3e5 d8d1 f1d1 h8g8 e5c6 b7c6 f2f3 f6d7 c3e2 e8c8 b2b3 f8d6 e2d4 g5g4 g1f1 g8g6 d4f5 d8f8 d1d6 c7d6 c1f4 g4f3 f5e7 c8c7 e7g6 h7g6 f4g3 f8f6 a1d1 f3g2 f1g2 f6e6 g2f3');
INSERT INTO evaluations VALUES(1264,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(1264,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(1264,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(1264,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(1264,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(1264,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(1264,36,'cp 12','g8f6 c1e3 d7d6 d4c6 b7c6 e3f4 e6e5 f4e3 f8e7 f2f4 e8g8 g1h1 a6a5 b2b3 e5f4 e3f4 f8e8 d1d4 c8g4 a1d1 g4e2 c3e2 e7f8 f4d6 f8d6 d4d6 c7d6 d1d6 c6c5 h1g1 f6e4 d6d3 a5a4 f1d1 a4b3 a2b3');
INSERT INTO evaluations VALUES(3828,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(3828,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(3828,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(3828,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(3828,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3828,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3828,36,'cp 0','a7a5 h2h3 c8e6 a2a4 h7h6 f1e1 d8b8 f3h4 b8d8');
INSERT INTO evaluations VALUES(5530,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(5530,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(5530,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(5530,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(5530,34,'cp 73',NULL);
INSERT INTO evaluations VALUES(5530,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(5530,36,'cp 99','c3d4 g8f6 d4d5 c6b8 d1c2 b8d7 h2h3 d7c5 b1c3 c8d7 c1g5 a8c8 b2b4 c5a4 c3a4 b7b5 a4b2 e8g8 a1c1 b5c4 b2c4 d7b5 f3d2 d8d7 f1e1 e7e6 d5e6 f7e6 c2b3 d6d5 c4e5 d7d6 c1c8 f8c8 e4d5 f6d5 d2f3');
INSERT INTO evaluations VALUES(7532,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(7532,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7532,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(7532,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7532,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7532,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7532,36,'cp 0','d3d4 f6d7 d4e5 d6e5 e1g1 e8g8 d1d3 a7a5 f1d1 e7d6 d3c4 d7b6 c4c6 c8d7 c6b7 d7c8');
INSERT INTO evaluations VALUES(9363,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(9363,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(9363,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(9363,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(9363,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(9363,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(9363,36,'cp 13','c3a4 c5b6 a4b6 a7b6 g1e2 c8e6 c4e6 f7e6 e1g1 e8g8 c2c3 c6e7 d1c2 e7g6 f2f3 d6d5 c1g5 h7h6 g5d2 d8e8 e2g3 e8a4 c2a4 a8a4 b2b3 a4a8 d2e3 a8a6 a3a4 c7c5 f1d1 f6d7 e4d5 e6d5 d3d4 e5d4 c3d4');
INSERT INTO evaluations VALUES(16330,30,'cp -122',NULL);
INSERT INTO evaluations VALUES(16330,31,'cp -115',NULL);
INSERT INTO evaluations VALUES(16330,32,'cp -135',NULL);
INSERT INTO evaluations VALUES(16330,33,'cp -120',NULL);
INSERT INTO evaluations VALUES(16330,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(16330,35,'cp -124',NULL);
INSERT INTO evaluations VALUES(16330,36,'cp -120','f6e4 d1e2 e4g5 c2c3 g5e6 g1h3 c6a5 c4d3 c7c6 d5e3 d7d5 b2b4 a5c4 e3c4 d5c4 d3c4 f8d6 d2d3 e8g8 a2a4 h7h6 e1g1 a7a5 b4b5 c6b5 a4b5 f8e8 e2a2 d8c7 g1h1 b7b6 f1e1 e8e7 f2f3 e6c5 h3f2 c8e6 c4e6 e7e6 c1e3 a5a4 f2e4 c5b7 c3c4 d6c5 e3f2 a4a3 e4c3 c5f2 a2f2 f7f5 f2e2 b7c5');
INSERT INTO evaluations VALUES(17482,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(17482,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(17482,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(17482,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17482,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(17482,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17482,36,'cp 0','g1f3 e7e6 c4b3 c6a5 c2c4 d7d6 b1c3 f8e7 h2h3 g8f6 e1g1 e8g8 b3c2 f6d7 c1e3 a8b8 c2a4 d7b6 a4c2');
INSERT INTO evaluations VALUES(17481,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17481,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(17481,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(17481,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17481,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(17481,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17481,36,'cp 0','e7e6 c4b3 c6a5 c2c4 d7d6 b1c3 f8e7 b3c2 g8f6 e1g1 e8g8 a2a3 d8c7 b2b4 a5c6 a1b1 b7b6 h2h3 c5b4 a3b4 d6d5 b4b5 c6b4 e4e5 f6d7 b5a6 c8a6 c1a3 b4c6 a3e7 c6e7 f1e1 d5c4 d3c4 e7g6 c2g6 h7g6 c3b5 a6b5 c4b5 d7c5');
INSERT INTO evaluations VALUES(19355,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(19355,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(19355,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(19355,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(19355,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(19355,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(19355,36,'cp 12','e7e6 c3e4 f8c5 e4c5 d8a5 c2c3 a5c5 e2e4 g8f6 f1d3 d7d5 e4d5 c5d5 g1f3 e8g8 e1g1 e6e5 d3c2 f8d8 c1g5 c8g4 d1d5 d8d5 f1d1 h7h6 g5f6 d5d1 a1d1 g7f6 c2e4 c6e7 d1d6 g4e6 e4c2 g8f8 f3d2 e6a2 d6f6 f8g7 f6d6 a2e6 c2b3');
INSERT INTO evaluations VALUES(21061,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(21061,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(21061,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(21061,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(21061,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(21061,35,'cp -91',NULL);
INSERT INTO evaluations VALUES(21061,36,'cp -67','h7h5 d2d4 h5h4 f1g2 h4g3 h2g3 h8h1 g2h1 d8d7 b1c3 g8f6 c1e3 d7h3 e1d2 h3g3 d1g1 g3g1 a1g1 c8d7 f3e5 d7e6 h1f3 e8c8 e5c6 b7c6 c3a4 e6f5 a4c5 e7e6 c5d3 f5d3 d2d3 c8d7 c2c4 d8b8 b2b3 a7a5 c4c5 a5a4 b3a4 b8b4 d3c3 b4a4 c3b3 a4a8 a2a4');
INSERT INTO evaluations VALUES(21626,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(21626,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(21626,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(21626,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(21626,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(21626,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(21626,36,'cp -37','d8e7 d1e2 e7e2 e1e2 g8e7 c2c3 c8f5 h1e1 f7f6 b1a3 e8c8 a3c2 e7g6 d2d3 c6e5 f3e5 g6e5 e1d1 c7c6 c2e3 f5g6 d3d4 e5f7 f2f3 f8d6 e2f2 d6c7 a2a4 f7d6 d1e1 d8e8 c1d2 a7a5 h2h4 g6d3 b3c2 d3c2 e3c2');
INSERT INTO evaluations VALUES(21706,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(21706,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(21706,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21706,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(21706,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21706,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(21706,36,'cp 4','e4d5 d8d5 b1c3 d5d6 c3b5 d6d8 d1e2 c8g4 c1b2 d8d7 h2h3 g4f3 e2f3 a7a6 b5c3 e8c8 e1c1 c8b8 g2g4 c6d4 f3e4 g8e7 f1g2 e7c6 c3e2 d4e2 e4e2 f7f6 e2e4 f8d6 c1b1 h7h5 e4c4 c6d4 b2d4 e5d4 c4d4 c7c6 g2f3 d7e7 d4e3 h5g4 h3g4');
INSERT INTO evaluations VALUES(21574,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(21574,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(21574,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(21574,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21574,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21574,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(21574,36,'cp 0','d8d5 b1c3 d5d6 c3b5 d6d8 d1e2 c8g4 c1b2 d8d7 h2h3 g4f3 e2f3 e8c8 e1c1 a7a6 b5c3 c8b8 g2g4 c6d4 f3e4 g8e7 f1g2 e7c6 c3e2 f8c5 h1e1 f7f5 g4f5 h8f8 b2d4 c5d4 e2d4 e5d4 e4e6 f8f5 e6d7 d8d7 g2c6');
INSERT INTO evaluations VALUES(1721,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(1721,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(1721,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(1721,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(1721,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(1721,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(1721,36,'cp 3','b1c3 d5d6 c3b5 d6d8 d1e2 c8g4 c1b2 d8d7 h2h3 g4f3 e2f3 a7a6 b5c3 e8c8 e1c1 f7f5 g2g4 c6d4 f3e3 d7c6 h1g1 e5e4 g4f5 g8f6 f1g2 f8b4 a2a3 b4a5 b3b4 a5b6 g2e4 f6e4 e3e4 c6e4 c3e4 d4e2 c1b1 e2g1 d1g1 h8f8 g1g7 f8f5 g7h7 b6f2 b1a2 d8d7 h7h6 c8b8 d2d3');
INSERT INTO evaluations VALUES(21975,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(21975,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(21975,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(21975,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(21975,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(21975,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(21975,36,'cp 15','g7g6 f1g2 f8g7 e1g1 g8f6 a2a4 e8g8 c2c3 a7a5 f1e1 h7h6 b1a3 f8e8 c1d2 d6d5 e4d5 f6d5 a3c4 c8g4 h2h3 g4f5 d1b3 d8d7 f3e5 g7e5 g3g4 f5d3 c4e5 c6e5 b3d5 d7d5 g2d5 c7c6 d5g2 e5c4 e1e8 a8e8 d2h6 c4b2 h6e3 b2c4 e3d4 d3c2 g4g5 c2d3');
INSERT INTO evaluations VALUES(22381,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(22381,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(22381,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(22381,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(22381,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(22381,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(22381,36,'cp 116','c7c5 d4d5 e6d5 c4d5 g7g6 f1e2 a7a6 g1f3 g8f6 a2a4 f8g7 e1g1 e8g8 c1f4 d8e7 f1e1 f6g4 f4g5 g7f6 g5f6 e7f6 d1d2 d7e5 h2h3 e5f3 g2f3 g4e5 f3f4 e5d7 d2e3 b7b6 e3g3 f6g7 g1h2 f7f5 e2f3 g8h8 f3g2 c8b7 a4a5 f5e4 c3e4 b7d5 e4d6 d5g2 h2g2 b6b5 e1e6 d7f6 g3h4');
INSERT INTO evaluations VALUES(23810,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23810,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23810,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23810,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(23810,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(23810,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(23810,36,'cp 12','e1g1 g8f6 f1e1 a8c8 d2d4 c8c6 c1d2 c6e6 e2b5 d8c7 b1c3 a7a6 b5d3 e8g8 f3g5 e6e1 a1e1 d7b6 g5f3 h7h6 d2f4 c7f4 e1e7 b6d7 d3e3 f4e3 e7e3 f8c8 h2h3 b7b5 d4c5 d7c5 f3d4 c5e6 c3e2 c8c4 c2c3 e6d4 e2d4 b5b4 d4e2 b4c3 e3c3 f6e4 c3c4 d5c4 g1f1 e4c5 e2c3 c5d3');
INSERT INTO evaluations VALUES(23811,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(23811,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(23811,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23811,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(23811,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23811,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(23811,36,'cp 15','g8f6 f1e1 a8c8 d2d4 c8c6 c1d2 c6e6 e2b5 d8c7 b1c3 a7a6 b5d3 e8g8 f3g5 e6e1 a1e1 d7b6 g5f3 c5d4 f3d4 b6c4 d2c1 e7b4 d4f5 f8c8 h2h3 c4e5 d3g3 e5g6 g3c7 c8c7 c1d2 g8f8 e1d1 g6e7 f5e7 b4e7 d2e3 f6e4 c3e4 d5e4 c2c3 e7c5 d1d5 c5e3 f2e3 f8e8 d5e5 c7e7 e5c5 e8d7');
INSERT INTO evaluations VALUES(24018,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(24018,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(24018,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(24018,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(24018,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24018,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(24018,36,'cp 16','f1e1 a8c8 d2d4 c8c6 c1d2 c6e6 e2b5 d8c7 b1c3 a7a6 b5d3 e8g8 f3g5 e6e1 a1e1 d7b6 g5f3 h7h6 d2f4 c7f4 e1e7 b6d7 d3e3 f4e3 e7e3 f8c8 h2h3 g8f8 d4c5 d7c5 f3d4 c5e6 c3e2 f6e4 f2f3 e4d6 d4e6 f7e6 e3e6 c8c2 e2f4 d6f5 e6b6');
INSERT INTO evaluations VALUES(26609,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26609,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26609,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26609,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26609,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(26609,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26609,36,'cp 5','g4f3 e2f3 e7e6 h3h4 a7a6 c1e3 g8f6 d1d2 c6a5 g2g3 b7b6 e1c1 b6b5 g3g4 a5c4 f1c4 b5c4 h4h5 h7h6 g4g5 f6g8 g5g6 f7g6 e3f4 d8d7 d1e1 e8c8 d2e3 g6g5 f4g3 d8e8 f3f4 g5f4 g3f4 g8f6 e1g1 h8g8 f4e5 d7c6 g1g6 f8d6 c1b1 c6b6 e5d6');
INSERT INTO evaluations VALUES(26624,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26624,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26624,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26624,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26624,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26624,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26624,36,'cp 0','e2f3 e7e6 f1b5 f8d6 c3e2 a7a6 b5c6 b7c6 e1g1 g8e7 c1e3 e8g8 d1d2 a6a5 e2f4 e7f5 f4d3 d8f6 c2c3 f6g6 d3f4 g6f6');
INSERT INTO evaluations VALUES(27348,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(27348,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(27348,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(27348,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(27348,34,'cp 167',NULL);
INSERT INTO evaluations VALUES(27348,35,'cp 172',NULL);
INSERT INTO evaluations VALUES(27348,36,'cp 172','h8f8 d2d3 e8c8 e1g1 h7h6 c1e3 c8b8 f2f3 g7g5 d1d2 e7d7 a1b1 d6d5 c3d5 f6d5 e4d5 d7d5 b2b3 b7b6 e3f2 d5d6 f1e1 d6a3 d2c3 b8b7 c3b2 a3d6 e1e4 f8f4 b1e1 h6h5 b3b4 c6b4 e4e5 g5g4 e1e4 f4e4 e5e4 b4c6 f3g4 h5g4 e4g4');
INSERT INTO evaluations VALUES(28538,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(28538,31,'cp 207',NULL);
INSERT INTO evaluations VALUES(28538,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(28538,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(28538,34,'cp 197',NULL);
INSERT INTO evaluations VALUES(28538,35,'cp 170',NULL);
INSERT INTO evaluations VALUES(28538,36,'cp 199','g5f6 d8f6 c3d5 f6d8 e4f5 b8d7 g2g3 g7g6 f1g2 f8h6 e1g1 e8g8 d1d3 a8b8 f1e1 d7f6 a1d1 f6d5 g2d5 g8g7 f5g6 f7g6 e1e2 g7h8 c2c4 b7b5 b2b3 d8b6 g1g2 f8f6 f2f3 b8f8 c4b5 a6b5 h2h4 b6c5 d5e4 c5b6 a2a4 b5a4 b3a4');
INSERT INTO evaluations VALUES(28982,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(28982,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(28982,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(28982,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28982,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(28982,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(28982,36,'cp 9','d8b6 f3h4 f5e4 d1b3 h7h6 h4f3 g7g5 f4g3 b8d7 b1d2 e7e6 f3e5 f8e7 e5d7 e8d7 g3e5 e4g6 a2a4 h8g8 a4a5 b6b3 d2b3 f6e4 f2f3 f7f6 f3e4 f6e5 e4d5 e6d5 d4e5 a7a6 f1e2 d7e6 b3d4 e6e5 e1g1 e5d6 d4f5 g6f5 f1f5 g8e8 a1f1');
INSERT INTO evaluations VALUES(29502,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(29502,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(29502,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(29502,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(29502,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(29502,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(29502,36,'cp 49','d8d5 c2c4 g4f3 g2f3 d5a5 b1c3 b8d7 c1d2 c7c6 d1e2 a5c7 e1c1 g7g6 c1b1 f8g7 d2g5 d7b6 e2e3 e8g8 g5f4 c7d7 f4e5 d7e6 b2b3 f6h5 f3f4 f7f6 d4d5 c6d5 e5c7 e6e3 f2e3 e7e6 c4d5 b6d5 c3d5 e6d5 d1d5 f6f5');
INSERT INTO evaluations VALUES(29612,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(29612,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(29612,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(29612,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(29612,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(29612,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(29612,36,'cp -26','c7c6 e2e3 b8d7 d1d2 e7e6 g2g4 f5g6 e1c1 d8a5 h2h4 h7h5 f1d3 g6d3 c2d3 h5g4 f3g4 f6g4 c1b1 g4f6 g1e2 f6h5 e3e4 f7f6 g5f4 e8c8 f4h2 d7b6 d2c2 f8d6 h2d6 d8d6 e2c1 d6d8 a2a3 a5a6 c1b3 b6d7 c3e2 c8b8 e2c3 d8c8 d1g1 e6e5');
INSERT INTO evaluations VALUES(30483,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(30483,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(30483,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(30483,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(30483,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(30483,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(30483,36,'cp 38','e1g1 a7a6 d2d4 c5d4 f3d4 d8c7 d1e2 g8f6 e4e5 b8c6 d4f3 f6g4 c1f4 b6b5 c4b3 c6e7 h2h3 e7g6 f4g3 g4h6 c3e4 h6f5 g3h2 c7c6 f1e1 a8c8 c2c3 b5b4 e2d3 b4c3 b2c3 f8e7 b3c2 c6b5 a1b1 b5d3 c2d3 b7e4 d3e4');
INSERT INTO evaluations VALUES(31058,30,'cp -204',NULL);
INSERT INTO evaluations VALUES(31058,31,'cp -197',NULL);
INSERT INTO evaluations VALUES(31058,32,'cp -200',NULL);
INSERT INTO evaluations VALUES(31058,33,'cp -221',NULL);
INSERT INTO evaluations VALUES(31058,34,'cp -208',NULL);
INSERT INTO evaluations VALUES(31058,35,'cp -193',NULL);
INSERT INTO evaluations VALUES(31058,36,'cp -193','c6d5 b1c3 g8f6 d1b3 e7e6 b3b7 b8d7 d2d4 a8b8 b7a7 d8c8 a7a4 f8b4 a4d1 f6e4 g2e4 b4c3 b2c3 d5e4 c1f4 c8c3 e1f1 b8a8 h2h3 g4h5 h1h2 e8g8 h2g2 h5g6 g2g3 c3c4 a1c1 c4a2 d1b3 a2a6 g3c3 d7f6 c3c6 a6a7 c6c5 f6d5 b3g3 f7f6 f1g2 g6h5');
INSERT INTO evaluations VALUES(31364,30,'cp -182',NULL);
INSERT INTO evaluations VALUES(31364,31,'cp -192',NULL);
INSERT INTO evaluations VALUES(31364,32,'cp -209',NULL);
INSERT INTO evaluations VALUES(31364,33,'cp -205',NULL);
INSERT INTO evaluations VALUES(31364,34,'cp -192',NULL);
INSERT INTO evaluations VALUES(31364,35,'cp -214',NULL);
INSERT INTO evaluations VALUES(31364,36,'cp -211','d1b3 g8f6 b1c3 e7e6 b3b7 b8d7 c3b5 a8c8 b7a6 f6e4 g1f3 f8c5 e1g1 e8g8 d2d4 d7b8 a6b7 g4f3 e2f3 e4d6 b5d6 c5d6 c1e3 c8c7 b7b5 d8h4 h2h3 b8c6 b5a6 d6f4 a1c1 c6b4 a6e2 c7c1 f1c1 b4a2 c1c5 g7g6 c5b5 a7a6 b5b7 a6a5');
INSERT INTO evaluations VALUES(31356,30,'cp -216',NULL);
INSERT INTO evaluations VALUES(31356,31,'cp -218',NULL);
INSERT INTO evaluations VALUES(31356,32,'cp -205',NULL);
INSERT INTO evaluations VALUES(31356,33,'cp -209',NULL);
INSERT INTO evaluations VALUES(31356,34,'cp -216',NULL);
INSERT INTO evaluations VALUES(31356,35,'cp -223',NULL);
INSERT INTO evaluations VALUES(31356,36,'cp -222','g8f6 b1c3 e7e6 h2h3 g4h5 b3b7 b8d7 c3b5 a8c8 g1f3 f8c5 b2b4 c5b4 e1g1 e8g8 b7a7 b4c5 a7a4 f6e4 d2d4 d7b6 a4b3 c5e7 a2a4 b6c4 b3d3 h5g6 d3d1 d8a5 f3e5 c4e5 g2e4 g6e4 d4e5 c8c2 c1a3 a5d8 f2f3 e7a3 a1a3 e4f5');
INSERT INTO evaluations VALUES(35946,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(35946,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(35946,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(35946,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(35946,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(35946,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(35946,36,'cp 39','h2h3 d6d5 e1g1 d5c4 d3c4 c5d4 e3d4 b8c6 f1e1 c6a5 c4f1 a7a6 f3e5 c8e6 c3a4 e6d7 e5d7 f6d7 c1g5 d7f6 a4c5 h7h6 g5e3 f6d5 a1c1 d5e3 f2e3 a8c8 d1d2 h6h5 b2b4 a5c6 a2a4 g7h6 c5b7 d8b6 b7c5 b6b4 d2b4 c6b4 f1a6 c8c5 c1c5 b4a6');
INSERT INTO evaluations VALUES(36373,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(36373,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(36373,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(36373,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(36373,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(36373,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36373,36,'cp 31','d1e2 c7c5 d4c5 e7c5 b1c3 b8c6 a2a3 d8e7 e3e4 e6e5 c1g5 c8e6 c3d5 e6d5 c4d5 c6d4 f3d4 c5d4 a1c1 a8c8 b2b4 b7b6 g2g3 h7h6 g5f6 e7f6 d5b7 c8c1 f1c1 g7g6 b7d5 f6e7 g1g2 g8g7 c1c6 a7a5 e2c4 f8d8 c4c2 a5b4');
INSERT INTO evaluations VALUES(36636,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(36636,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(36636,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36636,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36636,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36636,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36636,36,'cp 0','b8d7 d1c2 f8e8 e2e4 e7e5 d4e5 d6e5 c1e3 d8e7 c3c4 c7c6 b1c3 f6g4 e3d2 d7c5 h2h3 g4f6 d2e3 a7a5 f1d1 f6d7 d1d2 e8f8 a1d1 f7f5 e4f5 g6f5 e3g5 e7f7 g5e3');
INSERT INTO evaluations VALUES(37522,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(37522,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(37522,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(37522,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(37522,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(37522,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(37522,36,'cp 81','e1g1 b8c6 a2a4 g8f6 f1e1 e8g8 d4d5 c6b8 h2h3 e7e6 b1c3 e6d5 c3d5 b8d7 c1g5 h7h6 d5f6 d7f6 g5h4 g6g5 h4g3 f8e8 e4e5 d6e5 d1d8 e8d8 e1e5 c7c6 a1e1 b7b5 c4a2 g7f8 e5e2 c6c5 g3c7 d8d7 c7a5 b5a4 f3e5 d7b7 a5c3');
INSERT INTO evaluations VALUES(39190,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(39190,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(39190,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(39190,33,'cp 227',NULL);
INSERT INTO evaluations VALUES(39190,34,'cp 241',NULL);
INSERT INTO evaluations VALUES(39190,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(39190,36,'cp 258','d2d4 c5b6 c1f4 d7d5 c2c4 g8e7 e2e3 c8e6 c4c5 b6a5 b1d2 b8d7 f1d3 e6f5 d3f5 e7f5 e5f3 a5d2 d1d2 c7c6 d2d3 f5h4 f3h4 d8h4 e1g1 e8g8 a2a4 h4f6 b2b4 a7a6 f4d6 f8e8 b4b5 a6b5 a4b5 f6e6 d6g3');
INSERT INTO evaluations VALUES(39568,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(39568,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(39568,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(39568,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(39568,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(39568,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(39568,36,'cp 9','c7c5 b1c3 b8c6 f1g2 d7d6 e1g1 c8g4 h2h3 g4f3 g2f3 e7e6 f3g2 g8e7 a1b1 e8g8 a2a3 b7b6 d2d3 a8c8 d1a4 h7h6 c1d2 d6d5 c4d5 e7d5 c3d5 e6d5 e2e3 d8d6 b2b4 c6e5 a4a7 e5d3 b4c5 b6c5 b1b7 c5c4 b7d7 d6f6 d7d5');
INSERT INTO evaluations VALUES(37786,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(37786,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(37786,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(37786,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(37786,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(37786,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(37786,36,'cp 3','b1c3 b8c6 f1g2 d7d6 e1g1 c8f5 d2d3 d8d7 a1b1 f5h3 g2h3 d7h3 c3d5 h3d7 c1d2 g8f6 e2e4 e8g8 b2b4 b7b6 g1g2 f6d5 e4d5 c6b4 d2b4 c5b4 b1b4 a8e8 a2a4 e7e5 d5e6 f7e6 d3d4 e8c8 d1b3 d6d5 c4d5 d7d5 b3d5');
INSERT INTO evaluations VALUES(41425,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(41425,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(41425,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(41425,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(41425,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(41425,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(41425,36,'cp -39','c1f4 e8g8 e2e3 c7c6 f1e2 b8d7 g1f3 f6h5 h2h4 h5f4 e3f4 h7h5 a1c1 b7b6 b2b4 d8c7 d1d2 a7a5 c5b6 d7b6 b4b5 b6c4 e2c4 d5c4 b5c6 c7c6 e1g1 c8b7 d4d5 c6d6 f1e1 g7c3 c1c3 b7d5 f3e5 d5e6 d2c2 f8d8 e5c4 d6d4 c4e5 d4f4 e5c6');
INSERT INTO evaluations VALUES(42271,30,'cp 400',NULL);
INSERT INTO evaluations VALUES(42271,31,'cp 409',NULL);
INSERT INTO evaluations VALUES(42271,32,'cp 419',NULL);
INSERT INTO evaluations VALUES(42271,33,'cp 411',NULL);
INSERT INTO evaluations VALUES(42271,34,'cp 425',NULL);
INSERT INTO evaluations VALUES(42271,35,'cp 423',NULL);
INSERT INTO evaluations VALUES(42271,36,'cp 441','d2d4 c5d6 e5f7 d8e7 d1e2 e7e2 e1e2 e8f7 c4d5 f7g6 f2f3 c7c6 d5e4 g6f7 b1c3 b8d7 c1e3 h8e8 h2h4 d7f6 e4d3 h7h6 c3e4 f6e4 f3e4 f7g8 a1f1 c6c5 d4c5 d6e5 c2c3 c8e6 e3d4 e6a2 e2e3 a7a5 d3b5 e8e7 g2g4 e5d4 c3d4 a2d5 b5d3 a8e8');
INSERT INTO evaluations VALUES(46848,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46848,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(46848,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(46848,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(46848,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46848,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(46848,36,'cp -5','d5e4 e2e3 h7h6 g5f4 f8g7 g1e2 g6g5 f4g3 c7c5 d4d5 g7b2 a1b1 b2g7 d1d2 f7f5 e2c3 e8g8 h2h4 g5g4 f1e2 b8d7 d5d6 e7e6 c3b5 d8f6 a2a4 b7b6 b5c7 a8b8 e1g1 c8b7 a4a5 b8d8 f1d1 b7c6 a5b6 a7b6 b1a1 d7e5 a1a7 f8f7 a7a6 f7d7');
INSERT INTO evaluations VALUES(46961,30,'cp -144',NULL);
INSERT INTO evaluations VALUES(46961,31,'cp -138',NULL);
INSERT INTO evaluations VALUES(46961,32,'cp -142',NULL);
INSERT INTO evaluations VALUES(46961,33,'cp -145',NULL);
INSERT INTO evaluations VALUES(46961,34,'cp -144',NULL);
INSERT INTO evaluations VALUES(46961,35,'cp -144',NULL);
INSERT INTO evaluations VALUES(46961,36,'cp -140','f6e4 c2c4 e7e5 f1d3 e4c5 d3c2 f8e7 b1c3 a7a5 g1e2 e8g8 e1g1 b8d7 b2b3 f7f5 f2f4 e5f4 c1f4 d7e5 e2d4 e7f6 c3b5 g7g6 d1d2 c8d7 a1e1 f8e8 a2a4 d7b5 d4b5 e8e7 f4e3 e5g4 e3c5 d6c5 e1e7 d8e7');
INSERT INTO evaluations VALUES(47063,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(47063,31,'cp -157',NULL);
INSERT INTO evaluations VALUES(47063,32,'cp -153',NULL);
INSERT INTO evaluations VALUES(47063,33,'cp -157',NULL);
INSERT INTO evaluations VALUES(47063,34,'cp -143',NULL);
INSERT INTO evaluations VALUES(47063,35,'cp -129',NULL);
INSERT INTO evaluations VALUES(47063,36,'cp -146','c2c4 e7e5 f1d3 e4c5 d3c2 a7a5 c1e3 b8a6 a2a3 f8e7 g1e2 e7g5 e3g5 d8g5 b2b4 c5d7 e1g1 e8g8 d1d2 g5d2 b1d2 a6b8 e2c3 a5b4 c3b5 c7c6 b5d6 b4a3 d2b1 c6d5 c4d5 a8a6 d6f5 g7g6 f5e3 a3a2 b1c3 d7c5 a1a2 a6a2 c3a2 b8a6 f1b1');
INSERT INTO evaluations VALUES(48771,30,'cp -107',NULL);
INSERT INTO evaluations VALUES(48771,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(48771,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(48771,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(48771,34,'cp -109',NULL);
INSERT INTO evaluations VALUES(48771,35,'cp -109',NULL);
INSERT INTO evaluations VALUES(48771,36,'cp -109','f6h5 d2d4 b8c6 f1b5 a7a6 b5a4 c8e6 d1e2 b7b5 a4b3 h7h6 c1d2 d8d7 e2f2 g7g5 e1c1 e8c8 g2g3 g5g4 f3h4 f4f3 h2h3 h8g8 h3g4 g8g4 h4f3 h5g3 h1h4 g4h4 f3h4 g3e4 c3e4 d5e4 c2c3 e6b3 a2b3 c6a5 b3b4 f8b4 f2g2 a5b3 c1b1 b3d2 d1d2 b4e7 g2e4 d7d5 e4f3 e7h4');
INSERT INTO evaluations VALUES(49960,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(49960,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(49960,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(49960,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(49960,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(49960,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(49960,36,'cp 51','g1f3 d7d5 c2c4 c7c6 e1g1 f8d6 b2b3 e8g8 c1a3 d6a3 b1a3 c8d7 a3c2 d7e8 c2e1 b8d7 a1c1 d8e7 c4d5 f6d5 d3c4 e8h5 d1c2 a8e8 h2h3 g7g5 f3h2 d5b6 c4d3 e7g7 c2b2 g5g4 h3g4 f5g4 d3e2 e6e5');
INSERT INTO evaluations VALUES(50082,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(50082,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(50082,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(50082,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(50082,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(50082,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50082,36,'cp 28','e7e6 a2a3 g7g6 b1c3 f8g7 e2e3 e8g8 f1e2 d7d6 g1f3 b8c6 d1c2 d8e7 e1g1 c6d8 b2b4 d8f7 h2h3 e6e5 f4h2 g6g5 a1c1 e5e4 f3d2 c8e6 b4b5 c7c6 a3a4 a8d8 b5c6 b7c6 d2b3 g8h8 a4a5 g5g4 h3g4');
INSERT INTO evaluations VALUES(49930,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(49930,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(49930,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(49930,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(49930,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(49930,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(49930,36,'cp 58','a2a3 g7g6 e2e3 d7d6 b1c3 f8g7 g1f3 b8c6 h2h3 d8e7 f1e2 e8g8 f4h2 c6d8 a1c1 d8f7 e1g1 g6g5 c4c5 c8d7 d1b3 d7c6 c5d6 c7d6 e2b5 h7h5 b5c4 d6d5 c4b5 g5g4 h3g4 h5g4 f3e5 c6b5 c3b5 f7e5 h2e5');
INSERT INTO evaluations VALUES(56756,30,'cp 227',NULL);
INSERT INTO evaluations VALUES(56756,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(56756,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(56756,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(56756,34,'cp 225',NULL);
INSERT INTO evaluations VALUES(56756,35,'cp 244',NULL);
INSERT INTO evaluations VALUES(56756,36,'cp 237','b1c3 g8e7 a2a3 g7g6 h2h4 h7h5 c1e3 f8g7 f1c4 d6d5 c4a2 c7c6 d1d2 b8d7 e1g1 e8g8 f1e1 f8f7 a1d1 d7f8 e3f4 b7b6 e4e5 e7f5 c3e2 c8b7 e2g3 c6c5 g3f5 g6f5 e5f6 g7f6 c2c3 c5c4 f4g5 f6g5 f3g5 f7g7');
INSERT INTO evaluations VALUES(56755,30,'cp 218',NULL);
INSERT INTO evaluations VALUES(56755,31,'cp 218',NULL);
INSERT INTO evaluations VALUES(56755,32,'cp 236',NULL);
INSERT INTO evaluations VALUES(56755,33,'cp 220',NULL);
INSERT INTO evaluations VALUES(56755,34,'cp 221',NULL);
INSERT INTO evaluations VALUES(56755,35,'cp 222',NULL);
INSERT INTO evaluations VALUES(56755,36,'cp 214','g8e7 a2a3 g7g6 c1e3 f8g7 d1d2 a7a6 e1c1 b7b5 h2h4 h7h6 f1e2 b8d7 g2g4 d7b6 d4d5 e6e5 f3e1 g6g5 e1g2 e8g8 e3b6 c7b6 g2e3 e7g6 c1b1 g6h4 c3a2 d8c7 a2b4 h4g6 b4c6 g6f4 e3f5 c8f5 g4f5 c7f7 c2c4');
INSERT INTO evaluations VALUES(3438,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(3438,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(3438,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(3438,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(3438,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(3438,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(3438,36,'cp 61','c8e6 f1d1 e6c4 d4c4 f8e8 a2a4 c7c6 a4a5 d8c7 c1e3 a7a6 c3d5 f6d5 e4d5 e7f6 e3b6 c7d7 d5c6 b7c6 c2c3 e8e5 b6d4 e5e6 d4f6 e6f6 c4d3 f6e6 b2b4 d7e7 c3c4 a8e8 d3f3 e7c7 g2g3');
INSERT INTO evaluations VALUES(3880,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(3880,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(3880,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(3880,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(3880,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(3880,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(3880,36,'cp 34','c3e2 b4c5 c2c3 h7h6 e2g3 c5b6 c4b3 c8e6 b3a4 c6e7 d3d4 e7g6 h2h3 e5d4 f3d4 d6d5 d4e6 f7e6 e4d5 d8d6 d1d3 g6f4 c1f4 d6f4 a4b3 f6d5 g1h1 f4g5 d3e4 f8f6 b3c2 g7g6 e4e2 d5f4 e2g4 g5g4 h3g4 f6f7 g3e4 f4d5 g2g3 d5f6 e4f6 f7f6 h1g2 a8d8');
INSERT INTO evaluations VALUES(4372,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4372,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(4372,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4372,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4372,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4372,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(4372,36,'cp 0','d7d5 c4d5 f6d5 e1g1 h7h6 c3d5 d8d5 c1d2 d5d6 d2c3 f8d8 f3d2 c8e6 a1c1 c5b6 a2a3 e6d5 g2d5 d6d5 d1b3 a7a5 b3d5 d8d5 c1c2 a5a4 f1c1 f7f6 g1f1 d5b5 d2e4 a8d8 h2h3 d8d5 e4d2 d5d8');
INSERT INTO evaluations VALUES(4379,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(4379,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(4379,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(4379,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(4379,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(4379,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(4379,36,'cp -6','h2h3 f8e8 f1e1 c6d4 f3d4 c5d4 d3c4 c7c6 d1f3 d7d6 d2d3 d4c5 c4b3 c8e6 c1e3 c5e3 f3e3 a7a5 f2f4 e6b3 a2b3 h7h6 e1f1 e5f4 e3f4 d8b6 f4f2 b6f2 g1f2 d6d5 e4d5 f6d5 c3d5 c6d5 f2g1 e8e2 f1f2');
INSERT INTO evaluations VALUES(4660,30,'cp -388',NULL);
INSERT INTO evaluations VALUES(4660,31,'cp -436',NULL);
INSERT INTO evaluations VALUES(4660,32,'cp -416',NULL);
INSERT INTO evaluations VALUES(4660,33,'cp -440',NULL);
INSERT INTO evaluations VALUES(4660,34,'cp -461',NULL);
INSERT INTO evaluations VALUES(4660,35,'cp -465',NULL);
INSERT INTO evaluations VALUES(4660,36,'cp -455','e1g1 d7d5 e4d5 d8d5 b1c3 d5f5 c3b5 c5b6 b5d4 b6d4 f3d4 c6d4 d1d4 c8e6 f1e1 g8f6 d4c3 a8d8 b2b3 h8e8 c1b2 f7g8 e1e3 e6f7 e3f3 f5e6 h2h3 c7c6 f3g3 f7g6 g3f3 d8d5 f3f6 e6f6 c3f6 g7f6 b2f6 d5f5 f6c3 e8e2 a2a3 e2c2');
INSERT INTO evaluations VALUES(7697,30,'cp -104',NULL);
INSERT INTO evaluations VALUES(7697,31,'cp -105',NULL);
INSERT INTO evaluations VALUES(7697,32,'cp -93',NULL);
INSERT INTO evaluations VALUES(7697,33,'cp -107',NULL);
INSERT INTO evaluations VALUES(7697,34,'cp -97',NULL);
INSERT INTO evaluations VALUES(7697,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(7697,36,'cp -92','f3e5 e8g8 e1g1 e7d6 d2d4 e4d3 c2d3 c8a6 a4c2 f8e8 f2f4 f6d5 e2h5 e8e7 b1a3 d5b4 c1d2 b4c2 a3c2 d6e5 f4e5 a6d3 e5e6 f7f5 f1f2 d3c2 d2a5 d8a5 f2c2 e7e6 c2e2 e6e2 h5e2 g8h7 e2c2 a8e8 h2h3 a5b6 g1h1 e8e5 a1d1 c6c5 b2b3 b6e6 c2c3 e5d5 d1d5 e6d5 h1h2 d5d6 h2h1 a7a6 c3c2 h7g6 c2e2 g6f7 e2f3');
INSERT INTO evaluations VALUES(7695,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(7695,31,'cp -90',NULL);
INSERT INTO evaluations VALUES(7695,32,'cp -86',NULL);
INSERT INTO evaluations VALUES(7695,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(7695,34,'cp -92',NULL);
INSERT INTO evaluations VALUES(7695,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(7695,36,'cp -107','e8g8 e1g1 e7d6 d2d4 e4d3 c2d3 c8a6 a4c2 f8e8 f2f4 f6d5 e2h5 e8e7 b1a3 d5b4 c1d2 b4c2 a3c2 d6e5 f4e5 a6d3 e5e6 f7f5 f1f2 d3c2 d2a5 d8a5 f2c2 e7e6 c2e2 e6e2 h5e2 g8h7 e2f3 a8e8 h2h3 a5c5 g1h1 e8e5 b2b3 c5d5 a1c1 d5f3 g2f3 e5e2 h1g1 e2a2 c1c6 a7a5');
INSERT INTO evaluations VALUES(14153,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(14153,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(14153,32,'cp 155',NULL);
INSERT INTO evaluations VALUES(14153,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(14153,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(14153,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(14153,36,'cp 148','e7e6 f1e2 f8d6 e1g1 c8d7 c1e3 e8g8 a1c1 d7e8 d1b3 a7a5 c3b5 a5a4 b3c3 d6e7 c3d2 f7f5 f3g5 e7g5 e3g5 d8d7 f2f3 e8h5 g2g4 f5g4 f3g4 f8f1 c1f1 h5g6 d2e3 c6b4 f1c1 c7c6 b5c3 h7h6 g5h4 a4a3 b2a3 a8a3 d4d5 b4d5 e4d5 b6d5');
INSERT INTO evaluations VALUES(14686,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(14686,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(14686,32,'cp 135',NULL);
INSERT INTO evaluations VALUES(14686,33,'cp 158',NULL);
INSERT INTO evaluations VALUES(14686,34,'cp 145',NULL);
INSERT INTO evaluations VALUES(14686,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(14686,36,'cp 151','f1d3 f8d6 e1g1 e6e5 d4e5 c6e5 f3e5 d6e5 a2a4 a7a5 f2f4 e5d4 g1h1 e8g8 e4e5 d4c3 b2c3 c8e6 d1c2 d8h4 f4f5 e6c4 c1f4 c4d3 c2d3 a8d8 d3e2 f8e8 f4g3 h4c4 e2h5 f7f6 e5e6 b6d5 f1d1 c4c6 d1e1 e8e7 h5g4 c6c3 h2h3 d8e8 a1d1 c7c6');
INSERT INTO evaluations VALUES(15571,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(15571,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(15571,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(15571,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(15571,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(15571,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(15571,36,'cp 21','c6d4 f3d4 e5d4 c3b5 c7c5 c2c3 a7a6 b5a3 f6e4 c3d4 d6d5 d2d3 e4f6 c1e3 c5d4 e3d4 f8d6 e2f3 e8g8 a1c1 d6f4 c1c3 f8e8 a3c2 f4e5 d4e5 e8e5 c3b3 a8b8 d1d2 c8d7 f1c1 d7a4 b3b4 a4c6 b4f4 d8b6 d2b4 b6b4 c2b4');
INSERT INTO evaluations VALUES(15646,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(15646,31,'cp 121',NULL);
INSERT INTO evaluations VALUES(15646,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(15646,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(15646,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(15646,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(15646,36,'cp 105','c1f4 f8e7 f1e2 e8g8 e1g1 c8g4 g1h1 h7h6 a2a3 d6d5 c3d5 f6d5 e4d5 d8d5 f4c7 a8c8 c7g3 g4f3 e2f3 d5d4 d1e2 e7f6 c2c3 d4c5 f1e1 c8d8 a1d1 c5b6 d1d5 d8d5 f3d5 b6c5 e1d1 c5e7 e2h5 c6e5 d1e1 e7d8 d5b7 e5d3 e1e2 f8e8 e2e8 d8e8');
INSERT INTO evaluations VALUES(18010,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(18010,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(18010,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(18010,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(18010,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18010,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18010,36,'cp 0','c6c5 a2a3 a6a5 c1d2 c8d7 b2b4 c5b4 a3b4 a5b4 c3d5 a8a1 d1a1 g8e7 d5b4 c7c5 b4a2 e7c6 a2c3 c6b4 e1g1 f8e7 c3d5 b4d5 e4d5 e8g8 d2c3 e7g5 c3d2');
INSERT INTO evaluations VALUES(18503,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(18503,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(18503,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(18503,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(18503,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(18503,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(18503,36,'cp 13','c2c3 g8f6 f1e1 f8e7 d2d4 e5d4 c3d4 c5d4 f3d4 c8d7 h2h3 e8g8 b1c3 f8e8 d1f3 d8b6 e1d1 b6b7 b2b3 c6c5 e4e5 b7f3 d4f3 d6e5 f3e5 d7e6 c1f4 e7f8 a1c1 h7h6 e5c6 e8c8 c6a5 c8d8 f4e3 f6d5 c3d5 d8d5');
INSERT INTO evaluations VALUES(18501,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(18501,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(18501,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(18501,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(18501,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(18501,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(18501,36,'cp 16','g8f6 f1e1 f8e7 d2d4 e5d4 c3d4 c5d4 f3d4 c8d7 b1c3 e8g8 d4e2 d8b6 d1c2 f8e8 c1e3 b6b7 h2h3 h7h6 a1d1 d6d5 e4d5 f6d5 c3d5 c6d5 e3d4 e7f8 c2d2 e8e4 e2f4 a8c8 a2a3 d7a4 d1c1 c8c4 e1e4 c4c1 e4e1');
INSERT INTO evaluations VALUES(18564,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(18564,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(18564,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(18564,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(18564,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(18564,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(18564,36,'cp 70','c4d5 c6d5 c1f4 g8f6 b1c3 f6h5 f4e5 e7e6 e2e3 a7a6 f1d3 f8e7 h2h3 d7e5 d4e5 g7g6 a1c1 e8g8 e1g1 f7f6 e5f6 h5f6 f1e1 c8d7 e3e4 d5e4 c3e4 f6e4 d3e4 d7b5 d1d8 a8d8 e4b7 e7b4 e1e4 b4d6 b7c6');
INSERT INTO evaluations VALUES(18680,30,'cp 560',NULL);
INSERT INTO evaluations VALUES(18680,31,'cp 567',NULL);
INSERT INTO evaluations VALUES(18680,32,'cp 551',NULL);
INSERT INTO evaluations VALUES(18680,33,'cp 575',NULL);
INSERT INTO evaluations VALUES(18680,34,'cp 567',NULL);
INSERT INTO evaluations VALUES(18680,35,'cp 570',NULL);
INSERT INTO evaluations VALUES(18680,36,'cp 572','b7c6 g1f3 f8c5 b5e2 d7d6 b2b4 c5b6 e1g1 g8f6 c1g5 e8g8 c3c4 a7a5 b4b5 c6b5 c4b5 c8b7 b1c3 a8c8 c3a4 b7e4 a4b6 d8b6 a1c1 f6d5 g5d2 d5b4 d2e3 b6b7 c1c8 f8c8 a2a4 b4c2 d1d6 c2e3 f2e3 e4f3 e2f3 e5e4 f3g4');
INSERT INTO evaluations VALUES(18389,30,'cp 578',NULL);
INSERT INTO evaluations VALUES(18389,31,'cp 573',NULL);
INSERT INTO evaluations VALUES(18389,32,'cp 563',NULL);
INSERT INTO evaluations VALUES(18389,33,'cp 557',NULL);
INSERT INTO evaluations VALUES(18389,34,'cp 586',NULL);
INSERT INTO evaluations VALUES(18389,35,'cp 569',NULL);
INSERT INTO evaluations VALUES(18389,36,'cp 572','g1f3 g8f6 b5d3 f8c5 f3e5 e8g8 e1g1 f8e8 e5f3 f6e4 d3e4 e8e4 c1g5 f7f6 b1d2 e4e8 g5f4 d7d5 f1e1 c8e6 d2f1 c5f8 d1d2 c6c5 b2b4 c5b4 c3b4 a7a5 a2a3 e6f7 e1e8 f7e8 f4c7 d8d7 c7a5 e8h5 a1c1 h5f3 g2f3 a8e8 c1c7 d7d6');
INSERT INTO evaluations VALUES(18723,30,'cp -186',NULL);
INSERT INTO evaluations VALUES(18723,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(18723,32,'cp -152',NULL);
INSERT INTO evaluations VALUES(18723,33,'cp -162',NULL);
INSERT INTO evaluations VALUES(18723,34,'cp -157',NULL);
INSERT INTO evaluations VALUES(18723,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(18723,36,'cp -162','g8f6 h5d1 d7d5 c4b5 d5d4 c3e2 f6e4 b5c6 b7c6 g1f3 e4f6 d2d3 f8d6 c1g5 d8a5 c2c3 f6d5 e1g1 e8g8 c3d4 c5d4 e2d4 a5b6 d4b3 c8a6 d1d2 c6c5 b3a5 f7f6 a5c4 b6c6 c4d6 c6d6 g5e3 e6e5 f1e1 a8c8 a2a3');
INSERT INTO evaluations VALUES(12345,30,'cp -165',NULL);
INSERT INTO evaluations VALUES(12345,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(12345,32,'cp -174',NULL);
INSERT INTO evaluations VALUES(12345,33,'cp -164',NULL);
INSERT INTO evaluations VALUES(12345,34,'cp -161',NULL);
INSERT INTO evaluations VALUES(12345,35,'cp -179',NULL);
INSERT INTO evaluations VALUES(12345,36,'cp -169','h5d1 d7d5 c4b5 d5d4 b5c6 b7c6 c3e2 f6e4 d2d3 e4f6 g1f3 f8d6 c1g5 h7h6 g5f6 d8f6 e1g1 d6c7 c2c3 e6e5 e2g3 e8g8 a1c1 a8b8 d1c2 c8g4 f3d2 f6g6 b2b3 b8d8 c3d4 c5d4 f1e1 g4d7 g3e4 f7f5 e4c5 d7c8 d2c4');
INSERT INTO evaluations VALUES(19059,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(19059,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(19059,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(19059,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(19059,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(19059,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(19059,36,'cp 10','b1c3 g8f6 c1g5 f8e7 g5f6 e7f6 c3d5 d7d6 e1c1 e8g8 c1b1 c8e6 g1e2 a8c8 c4b3 a7a6 g2g3 b7b5 h2h4 e6d5 d1d5 c6a5 b3f3 d8c7 c2c3 a5c4 e2c1 c8b8 c1b3 f6e7 f1e2 c4b6 d5d2 b6c4 d2c2 a6a5 b3d2');
INSERT INTO evaluations VALUES(20122,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(20122,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(20122,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(20122,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(20122,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(20122,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(20122,36,'cp 82','d1d4 e7e6 b1c3 g8f6 c1g5 f8e7 e1c1 e8g8 f2f4 a7a6 h1e1 b7b5 c4b3 d8c7 g5f6 e7f6 d4d6 c7d6 d1d6 f8d8 d6d8 f6d8 g2g3 c8b7 e4e5 d8e7 e1d1 g8f8 d1d7 b7g2 c3e2 g7g5 c2c3 f8e8 d7d2 g5f4 g3f4 e7c5 e2d4 g2b7 f4f5 e8e7');
INSERT INTO evaluations VALUES(21384,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(21384,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(21384,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(21384,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(21384,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(21384,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(21384,36,'cp 29','d5c4 d1a4 d8d5 b1c3 f8b4 f1g2 c8d7 c1d2 b4c3 d2c3 g8f6 a4c2 e8g8 a2a4 a7a5 e1g1 c6b4 c3b4 a5b4 f3e5 d5d4 e5d7 d4d7 g2b7 a8b8 b7f3 d7d4 e2e3 d4c5 f1c1 c4c3 b2c3 b4b3 c2b2 c5a5 a1a3 f6d7 a3b3 a5a4 b3b4 a4a6 c1b1 h7h6');
INSERT INTO evaluations VALUES(22144,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(22144,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(22144,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(22144,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(22144,34,'cp 106',NULL);
INSERT INTO evaluations VALUES(22144,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(22144,36,'cp 100','g8f6 b1c3 f8e7 c1f4 c8e6 g2g4 e8g8 g4g5 f6e8 h2h4 c7c6 e1c1 b7b5 c1b1 d8a5 b2b3 a8d8 h4h5 b5b4 c3a4 e7g5 f4g5 a5g5 h5h6 g7g6 d4b4 g5e5 b4d4 e5d4 d1d4 d6d5 h1d1 e8f6 f2f3 f6h5 b1b2 h5f4 e2a6 g6g5 a4c5');
INSERT INTO evaluations VALUES(15773,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(15773,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(15773,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(15773,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(15773,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(15773,35,'cp 108',NULL);
INSERT INTO evaluations VALUES(15773,36,'cp 121','c1f4 f8e7 b1c3 e8g8 e1c1 c8e6 h2h4 f6g4 e2g4 e6g4 f2f3 g4e6 g2g4 e7f6 d4d3 f6c3 d3c3 f7f6 g4g5 d8d7 g5f6 f8f6 f4g5 f6f7 h1g1 d7c6 c3d2 e6a2 g5h6 g8h8 h4h5 a2e6 d2d4 a8g8 h6e3 e6b3 d1d2 c6c4 h5h6 c4d4 e3d4');
INSERT INTO evaluations VALUES(22432,30,'cp -7',NULL);
INSERT INTO evaluations VALUES(22432,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(22432,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(22432,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22432,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(22432,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(22432,36,'cp 0','d7c6 d2d3 f8g7 b1d2 c6c5 a2a3 a7a5 d2c4 f7f6 c1e3 b7b6 b2b4 a5b4 a3b4 a8a1 d1a1 c5b4 a1a4 c8d7 a4b4 g8e7 c4a3 d7e6 d3d4 e7c6 b4a4 e6d7 d4e5 c6e5 a3b5 e5f7 e1g1 e8g8 a4c4 d7b5 c4b5 f8e8 b5c6 d8e7');
INSERT INTO evaluations VALUES(21252,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21252,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21252,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(21252,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21252,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21252,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21252,36,'cp 0','d2d3 f8g7 b2b4 d8e7 c2c3 g8f6 b1d2 e8g8 e1g1 f8d8 d1c2 f6h5 d2b3 c8g4 f3e1 b7b6 h2h3 g4e6 e1f3 e7d7 d3d4 e5d4 c3d4 e6b3 a2b3 g7d4 f3d4 d7d4 c1e3 d4b4 c2c6 b4e7 f1c1 h5g7 b3b4 e7b4 c6c7 g7e6 c7b7 e6c5 b7c7');
INSERT INTO evaluations VALUES(24154,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24154,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(24154,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24154,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(24154,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24154,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24154,36,'cp 14','c4e6 f7e6 c2c3 a7a5 d3d4 c5a7 d4e5 f6e4 e5d6 d8d6 d1d6 c7d6 b1d2 e4d2 c1d2 e8d7 a1e1 b7b5 e1e4 a8f8 f1e1 f8e8 e4g4 e8e7 d2f4 e6e5 f4g5 h7h5 g4e4 e7f7 e1d1 f7f5 d1d2 h8f8 g5e3 a7e3');
INSERT INTO evaluations VALUES(24147,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24147,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24147,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24147,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24147,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24147,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24147,36,'cp 0','f7e6 c2c3 a7a5 d3d4 c5a7 d4e5 f6e4 e5d6 d8d6 d1d6 c7d6 b1d2 e4d2 c1d2 e8d7 a1e1 h7h6 e1e4 h8f8 f1e1 f8f6 e4g4 f6f7 h3h4 b7b5 d2e3 a7e3 e1e3 b5b4 h4h5 a8b8 c3c4 b8g8 b2b3 f7f5 g4g6 f5f6 g6g4');
INSERT INTO evaluations VALUES(24873,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(24873,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24873,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24873,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24873,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24873,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24873,36,'cp 0','e7e6 c2c4 d5c4 d1a4 d8d7 a4c4 e8c8 e1g1 c6d4 b1c3 d4f3 g2f3 d7d4 c4a2 f6e4 a2b3 e4c5 b3a2');
INSERT INTO evaluations VALUES(25186,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(25186,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(25186,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(25186,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25186,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25186,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25186,36,'cp 0','c2c4 f8e7 e1g1 e8g8 c4d5 e6d5 c1f4 f6e4 b1c3 e4c3 b2c3 f8e8 f1e1 c6a5 f3d2 c7c6 e2e4 d5e4 d2e4 a5c4 d1b3 c4a5');
INSERT INTO evaluations VALUES(25479,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(25479,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(25479,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(25479,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(25479,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(25479,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(25479,36,'cp 38','f8e7 c1g5 a7a6 b5c6 d7c6 f1e1 e8g8 f3e5 d8d6 c3e2 f6d7 g5e7 d6e7 e5c6 b7c6 d1d2 e7d6 e2f4 f8e8 e1e8 a8e8 g2g3 h7h6 a1e1 e8e4 c2c3 g7g5 f4g2 e4e1 d2e1 c6c5 g2e3 d6e6 d4c5 d7c5 e1d1 c7c6 d1d4 c5e4 c3c4 e4f6 g1g2 f6e4 h2h3 e4f6 c4d5 f6d5');
INSERT INTO evaluations VALUES(28633,30,'cp 145',NULL);
INSERT INTO evaluations VALUES(28633,31,'cp 143',NULL);
INSERT INTO evaluations VALUES(28633,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(28633,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(28633,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(28633,35,'cp 132',NULL);
INSERT INTO evaluations VALUES(28633,36,'cp 148','f1b5 f6d7 f3g5 e7e6 e1g1 d8e7 f2f4 h7h6 g5f3 g7g6 c1e3 a7a6 b5d3 f8g7 g1h1 c7c5 f4f5 b8c6 f5e6 f7e6 e4e5 e8g8 e5d6 e7d6 d4c5 d7c5 d3g6 d6d1 a1d1 c6e7 e3c5 b6c5 g6e4 e7d5 e4d5 e6d5 c3d5 b7d5 d1d5 g7b2');
INSERT INTO evaluations VALUES(29471,30,'cp 236',NULL);
INSERT INTO evaluations VALUES(29471,31,'cp 236',NULL);
INSERT INTO evaluations VALUES(29471,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(29471,33,'cp 253',NULL);
INSERT INTO evaluations VALUES(29471,34,'cp 233',NULL);
INSERT INTO evaluations VALUES(29471,35,'cp 258',NULL);
INSERT INTO evaluations VALUES(29471,36,'cp 267','c4e6 f7e6 c1d2 c7c5 d4c5 g7g6 g1f3 f8g7 b1c3 b8d7 d1a4 d8c8 a1c1 e8g8 a4c4 c8c6 b2b4 a7a5 b4b5 c6c5 c4e6 g8h8 e1g1 c5h5 c3e2 h5b5 e2f4 d7c5 e6h3 h8g8 f3d4 b5b6 d4e6 c5e6 f4e6 f8c8 d2c3 c8c6 c3d4 b6b4 c1c6 b7c6 f1c1 b4d6 e6g7 g8g7');
INSERT INTO evaluations VALUES(29461,30,'cp 245',NULL);
INSERT INTO evaluations VALUES(29461,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(29461,32,'cp 249',NULL);
INSERT INTO evaluations VALUES(29461,33,'cp 248',NULL);
INSERT INTO evaluations VALUES(29461,34,'cp 262',NULL);
INSERT INTO evaluations VALUES(29461,35,'cp 238',NULL);
INSERT INTO evaluations VALUES(29461,36,'cp 249','f7e6 c1d2 g7g6 g1f3 b8a6 b1c3 f8h6 d1b3 d8c8 b3c4 e8g8 a1c1 c8d7 e1g1 a8c8 b2b4 c7c6 h2h3 a6c7 c4d3 h6g7 a2a3 c7b5 c3e2 d7d5 e2f4 d5f5 d3b3 b5c7 f4d3 f5d5 b3b1 f6e4 d3c5 f8f3 c5e4 f3f5');
INSERT INTO evaluations VALUES(29405,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(29405,31,'cp 282',NULL);
INSERT INTO evaluations VALUES(29405,32,'cp 272',NULL);
INSERT INTO evaluations VALUES(29405,33,'cp 273',NULL);
INSERT INTO evaluations VALUES(29405,34,'cp 275',NULL);
INSERT INTO evaluations VALUES(29405,35,'cp 274',NULL);
INSERT INTO evaluations VALUES(29405,36,'cp 284','c1d2 c7c5 d4c5 e6e5 g1f3 e5e4 f3e5 b8d7 d1a4 e7e6 d2c3 f8c5 b1d2 e8g8 e5d7 d8d7 a4d7 f6d7 d2e4 c5e7 a1c1 a8c8 e1e2 d7f6 e4g5 f6g4 g5f3 g4f6 h1d1 b7b6 h2h3 f6e4 d1d7 e4c3 b2c3 e7f6 a2a4 f6c3 f3g5 c3f6 c1c8 f8c8 g5e6 c8c2 e2f3');
INSERT INTO evaluations VALUES(31520,30,'cp 436',NULL);
INSERT INTO evaluations VALUES(31520,31,'cp 436',NULL);
INSERT INTO evaluations VALUES(31520,32,'cp 452',NULL);
INSERT INTO evaluations VALUES(31520,33,'cp 448',NULL);
INSERT INTO evaluations VALUES(31520,34,'cp 407',NULL);
INSERT INTO evaluations VALUES(31520,35,'cp 424',NULL);
INSERT INTO evaluations VALUES(31520,36,'cp 433','g2g4 f5e4 f2f3 e4g6 h4h5 g6h5 h1h5 f8e7 e1f2 e7h4 f2g2 g7g6 h5h4 d8h4 g1h3 h4e7 f1d3 b8d7 d1e1 h7h6 c1d2 g6g5 d2b4 e7d8 b4d6 g8e7 b1d2 e7c8 d6a3 c8e7 d2b3 d7f8 e1e3 e7g6 g2f2 g6h4 a1h1 f8g6 a3d6 b7b6 d3g6 h4g6 e3c3 c6c5 d4c5 h6h5 c5c6 h5g4');
INSERT INTO evaluations VALUES(31518,30,'cp 417',NULL);
INSERT INTO evaluations VALUES(31518,31,'cp 405',NULL);
INSERT INTO evaluations VALUES(31518,32,'cp 395',NULL);
INSERT INTO evaluations VALUES(31518,33,'cp 400',NULL);
INSERT INTO evaluations VALUES(31518,34,'cp 410',NULL);
INSERT INTO evaluations VALUES(31518,35,'cp 438',NULL);
INSERT INTO evaluations VALUES(31518,36,'cp 417','f5e4 f2f3 e4g6 h4h5 g6h5 h1h5 f8e7 e1f2 e7h4 f2g2 g7g6 h5h4 d8h4 g1h3 h4e7 d1e1 b8d7 f1d3 e7f8 b1c3 h7h6 c3e2 e8c8 b2b4 c8b8 a2a4 g8e7 a4a5 g6g5 b4b5 c6b5 a5a6 f7f5 e5f6 f8f6 e1g3 e6e5 d4e5 d7e5 a6b7 b8b7 d3b5 f6f3 g3f3');
INSERT INTO evaluations VALUES(31502,30,'cp 431',NULL);
INSERT INTO evaluations VALUES(31502,31,'cp 426',NULL);
INSERT INTO evaluations VALUES(31502,32,'cp 447',NULL);
INSERT INTO evaluations VALUES(31502,33,'cp 429',NULL);
INSERT INTO evaluations VALUES(31502,34,'cp 446',NULL);
INSERT INTO evaluations VALUES(31502,35,'cp 425',NULL);
INSERT INTO evaluations VALUES(31502,36,'cp 446','f2f3 e4g6 h4h5 g6h5 h1h5 f8e7 e1f2 e7h4 f2g2 g7g6 h5h4 d8h4 g1h3 h4e7 f1d3 b8d7 d1e1 h7h6 b2b3 c6c5 c2c4 c5d4 c1a3 d7c5 c4d5 c5d3 e1d2 e7d8 d2d3 d8d5 b1d2 h6h5 g4g5 g8e7 d2e4 d5a5 a3c5 e7f5 e4f6 e8d8 c5d4 d8e7 d4c3 a5c5 f6e4 c5b6 e4d6 f5d6 e5d6 b6d6 d3d6 e7d6 a1d1 d6c6 c3h8 a8h8');
INSERT INTO evaluations VALUES(31501,30,'cp 444',NULL);
INSERT INTO evaluations VALUES(31501,31,'cp 442',NULL);
INSERT INTO evaluations VALUES(31501,32,'cp 417',NULL);
INSERT INTO evaluations VALUES(31501,33,'cp 435',NULL);
INSERT INTO evaluations VALUES(31501,34,'cp 437',NULL);
INSERT INTO evaluations VALUES(31501,35,'cp 439',NULL);
INSERT INTO evaluations VALUES(31501,36,'cp 445','e4g6 h4h5 g6h5 h1h5 f8e7 e1f2 e7h4 f2g2 g7g6 h5h4 d8h4 g1h3 h4e7 f1d3 b8d7 d1e1 h7h6 b2b3 e7d8 c2c4 g8e7 b1c3 a7a6 c1d2 g6g5 a1c1 d7f8 c4d5 c6d5 c3a4 b7b6 d2b4 a8c8 c1c8 e7c8 d3a6 c8a7 a6d3 h6h5 h3f2 f8d7 e1d2 h5g4');
INSERT INTO evaluations VALUES(32291,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(32291,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(32291,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(32291,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32291,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(32291,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(32291,36,'cp -4','f8d6 f4d6 c7d6 d1b3 g8f6 e2e3 d8e7 c3c4 e8g8 c4d5 f6d5 b1d2 f5g6 f1c4 d5b6 c4b5 a7a6 b5d3 g6d3 b3d3 b8c6 e1g1 a8c8 a1c1 c8c7 d3b3 b6d7 d4d5 e6d5 b3d5 f8c8 f3d4 c6d4 c1c7 c8c7 d5d4 c7c2');
INSERT INTO evaluations VALUES(32458,30,'cp 655',NULL);
INSERT INTO evaluations VALUES(32458,31,'cp 659',NULL);
INSERT INTO evaluations VALUES(32458,32,'cp 644',NULL);
INSERT INTO evaluations VALUES(32458,33,'cp 632',NULL);
INSERT INTO evaluations VALUES(32458,34,'cp 627',NULL);
INSERT INTO evaluations VALUES(32458,35,'cp 641',NULL);
INSERT INTO evaluations VALUES(32458,36,'cp 648','f3b3 d8e7 b1c3 e7b4 c4f7 e8d8 b3e6 b4e7 f7g8 e7e6 g8e6 f8c5 c3a4 c5b6 b2b4 c7c6 a4b6 a7b6 c1e3 d8c7 a2a4 h8e8 e6b3 e8d8 e1e2 h7h6 c2c3 c6c5 h1d1 d8d1 a1d1 b8d7 a4a5 b6b5 d1d5 c5c4 b3c2 a8d8 d5b5');
INSERT INTO evaluations VALUES(34137,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(34137,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(34137,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(34137,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(34137,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34137,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(34137,36,'cp 22','c4d3 g8f6 e1g1 f8c5 d4b3 c5a7 c2c4 h7h5 b1c3 f6g4 g2g3 b8c6 d3e2 d7d6 c1f4 e6e5 c3d5 c7d8 f4d2 c8e6 a1c1 d8d7 d2g5 f7f6 g5e3 g4e3 e2h5 e8d8 d5e3 a7e3 f2e3 d8c7 f1f2 a8d8 h5e2 c7b8 b3d2 d7c7 d1b3 c7d7 b3a3 d8c8 b2b4');
INSERT INTO evaluations VALUES(38067,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(38067,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(38067,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(38067,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(38067,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38067,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(38067,36,'cp 39','e2e3 b7b5 c1d2 a7a5 a4b5 b4c3 d2c3 c6b5 b2b3 c8b7 b3c4 b5b4 c3b2 g8f6 f1d3 e8g8 e1g1 b8d7 f3d2 d8c7 d3c2 f8d8 f1e1 d7b6 c4c5 b6d7 c2a4 f6e4 a1c1 b7d5 d2f1 d5c6 d1c2 c6a4 c2a4 e6e5 h2h3 e5d4 e3d4 d7f6');
INSERT INTO evaluations VALUES(38066,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(38066,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(38066,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(38066,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(38066,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(38066,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(38066,36,'cp 25','b7b5 c1d2 a7a5 a4b5 b4c3 d2c3 c6b5 b2b3 c8b7 b3c4 b5b4 c3b2 b8d7 f1d3 g8f6 f3d2 d8c7 e1g1 e8g8 d3c2 f8d8 f1e1 d7b6 c4c5 b6d7 c2a4 f6e4 a1c1 b7d5 e1e2 e4d2 e2d2 h7h6 f2f3 f7f5 d1c2 d7f6 c5c6 a8a6 c2c5 b4b3 c5b5 d8a8 d2e2 a6b6');
INSERT INTO evaluations VALUES(41876,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(41876,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(41876,32,'cp 7',NULL);
INSERT INTO evaluations VALUES(41876,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(41876,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41876,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41876,36,'cp 0','e5d4 f3d4 e8g8 e1g1 a7a5 g1h1 b8d7 c4b3 d7e5 c1g5 c7c6 f2f3 h7h6 g5h4 f8e8 a2a4 c8d7 h4f2 c5b4 c3a2 b4c5');
INSERT INTO evaluations VALUES(41900,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41900,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41900,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41900,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41900,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(41900,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(41900,36,'cp 0','f3d4 e8g8 e1g1 b8d7 c4b3 d7e5 g1h1 a7a5 c1g5 c7c6 f2f3 h7h6 g5h4 f8e8 a2a4 c8d7 h4f2 c5b4 c3a2 b4c5');
INSERT INTO evaluations VALUES(42185,30,'cp 316',NULL);
INSERT INTO evaluations VALUES(42185,31,'cp 323',NULL);
INSERT INTO evaluations VALUES(42185,32,'cp 338',NULL);
INSERT INTO evaluations VALUES(42185,33,'cp 309',NULL);
INSERT INTO evaluations VALUES(42185,34,'cp 333',NULL);
INSERT INTO evaluations VALUES(42185,35,'cp 317',NULL);
INSERT INTO evaluations VALUES(42185,36,'cp 321','e4e5 f6d5 d1g4 e8f8 g1e2 c7c5 c1a3 b8d7 c4d5 e6d5 g4f3 f8g8 f3d5 d8c7 e1g1 c5d4 c3d4 d7b6 d5e4 c8e6 a1c1 c7d7 a3d6 e6d5 e4d3 a8c8 c1c8 d7c8 f1c1 d5c4 d3d2 h7h5 d4d5 c8f5 e2f4 h8h6 d6c7 g7g5 c7b6 g5f4 b6d8 c4b5 d8f6 h6g6 f2f3 b5d7 g1f2 g8h7 a2a4 b7b6 d2d4 b6b5 a4b5 d7b5');
INSERT INTO evaluations VALUES(42127,30,'cp 316',NULL);
INSERT INTO evaluations VALUES(42127,31,'cp 344',NULL);
INSERT INTO evaluations VALUES(42127,32,'cp 342',NULL);
INSERT INTO evaluations VALUES(42127,33,'cp 318',NULL);
INSERT INTO evaluations VALUES(42127,34,'cp 337',NULL);
INSERT INTO evaluations VALUES(42127,35,'cp 320',NULL);
INSERT INTO evaluations VALUES(42127,36,'cp 312','f6d5 d1g4 e8f8 g1e2 c7c5 c1a3 b8d7 c4d5 e6d5 g4f3 f8g8 f3d5 d8c7 e1g1 c5d4 c3d4 d7b6 d5e4 c7c6 d4d5 c6c4 f2f3 h7h5 e2f4 c4e4 f3e4 b6c4 a3b4 c4e5 a1c1 b7b6 c1c7 c8a6 f1d1 g8h7 d5d6 h8c8 b4c3 e5g6 f4d5 a6e2 d1d4 c8c7 d6c7 e2g4 h2h3 g4e6');
INSERT INTO evaluations VALUES(43734,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(43734,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(43734,32,'cp 98',NULL);
INSERT INTO evaluations VALUES(43734,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(43734,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(43734,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(43734,36,'cp 90','c3d5 e7e5 d5f6 d8f6 d4b3 c8d7 e1g1 f8e7 c1d2 e8g8 c2c4 b5c4 d3c4 b8c6 d1e2 a6a5 a2a4 e7d8 c4d5 a8a7 e2e3 c6d4 b3d4 e5d4 e3b3 d7e6 f1c1 h7h6 c1c6 f6e5 g2g3 f8e8 a1e1 d8f6 d5e6 e5e6 b3e6 f7e6 c6d6 e8b8 d6e6');
INSERT INTO evaluations VALUES(44316,30,'cp 222',NULL);
INSERT INTO evaluations VALUES(44316,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(44316,32,'cp 218',NULL);
INSERT INTO evaluations VALUES(44316,33,'cp 206',NULL);
INSERT INTO evaluations VALUES(44316,34,'cp 222',NULL);
INSERT INTO evaluations VALUES(44316,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(44316,36,'cp 205','f3e5 d8e7 d2d4 c5d4 d1d4 b8c6 e5c6 d7c6 b1c3 c8e6 c4b3 a7a5 c1e3 e6b3 a2b3 e7e6 e1c1 f8b4 c3a4 e8g8 f2f3 h7h6 d4b6 e6c8 a4c5 f8e8 e3d4 b4c5 b6c5 a5a4 b3a4 a8a4 b2b3 a4a2 d4b2 c8c7 d1d6 a2a5 c5d4 c6c5 d4d3 b7b5 c1b1 e8a8');
INSERT INTO evaluations VALUES(44914,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(44914,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(44914,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(44914,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(44914,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(44914,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(44914,36,'cp 20','g7g6 e2e3 f8g7 f1e2 e8g8 e1g1 c6c5 c4d5 c5d4 f3d4 f6d5 f1d1 d8c7 c2c7 d5c7 b1c3 f8d8 e2f3 e7e5 d4b5 d8d1 f3d1 c7b5 c3b5 b8a6 e3e4 c8e6 c1e3 g7f8 d1e2 b7b6 f2f3 g8g7 a2a4 f8c5 g1f2 a6b4 b5c7 b4c2 c7e6 f7e6');
INSERT INTO evaluations VALUES(45520,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(45520,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(45520,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(45520,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(45520,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(45520,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45520,36,'cp 36','f8g7 h2h3 e8g8 g1f3 f8e8 f1d3 c5c4 d3c2 b7b5 e1g1 a7a6 d1d2 b8d7 a2a3 d8b6 g5e3 d7c5 f1e1 a6a5 e3d4 f6d7 d4g7 g8g7 d2d4 d7e5 f3d2 g6g5 a3a4 c8d7 a4b5 d7b5 d2f3 b5d7 c3e2 h7h6 f3e5 e8e5');
INSERT INTO evaluations VALUES(40442,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(40442,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(40442,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(40442,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40442,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(40442,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(40442,36,'cp 18','f2f3 e8g8 f1e2 a7a6 a2a4 d8c7 a1a3 f8e8 d1c1 h7h5 g5h6 g7h6 c1h6 c7b6 h6d2 b6b4 c3a2 b4b6 d2c1 b6a5 a2c3 a5b4 c1d2 b8d7 g1h3 a8b8 h3f2 c5c4 c3a2 b4d2 e1d2 b7b5 a4b5 a6b5');
INSERT INTO evaluations VALUES(45778,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(45778,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(45778,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(45778,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45778,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(45778,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(45778,36,'cp 33','e1g1 f8e7 f1e1 e8g8 h2h3 f8e8 c1f4 b8d7 b1d2 d7f8 f4g3 f8e6 c2c3 f6h5 g3h2 e7d6 d3c2 d6h2 f3h2 e6f4 d2f3 h5f6 d1d2 f4g6 f3g5 g6f8 h2f1 h7h6 e1e8 f6e8 g5f3 e8d6 f1e3 c8e6 a1e1 d8b6 f3e5 a8e8 a2a4 f7f6 e5d3 e6f7 e3g4');
INSERT INTO evaluations VALUES(45776,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(45776,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(45776,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45776,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(45776,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(45776,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(45776,36,'cp 47','f8e7 f1e1 e8g8 h2h3 f8e8 c1f4 e7d6 f4d6 e8e1 d1e1 d8d6 b1d2 c8e6 f3e5 c6c5 d4c5 d6c5 d2b3 c5c7 a2a4 b8d7 e5f3 h7h6 a4a5 a7a6 e1b4 a8e8 a1d1 d7b8 f3d4 e6d7 b3c5 b8c6 d4c6 d7c6 c5b3');
INSERT INTO evaluations VALUES(46140,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(46140,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(46140,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(46140,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(46140,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(46140,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(46140,36,'cp -25','c5d4 e3d4 f8e7 g1f3 e8g8 c4d5 f6d5 g5e7 d8e7 f1d3 d5c3 b2c3 b7b6 d1e2 c8b7 e1g1 b8d7 a2a4 a8c8 f1c1 a7a5 a1b1 f8d8 f3e5 h7h6 e5d7 d8d7 e2e3 e7d8 d3c2 d7c7 c2b3 d8f6 c3c4 c7d7 c4c5 b6c5');
INSERT INTO evaluations VALUES(48577,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(48577,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(48577,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(48577,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(48577,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48577,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(48577,36,'cp -13','f8b4 c1d2 e4e3 d2e3 e8g8 e3d2 c7c6 g1f3 c6d5 f1e2 b8c6 e1g1 f8e8 g1h1 h7h6 f3e5 c6e5 f4e5 e8e5 c3e4 f6e4 d2b4 e4g5 a2a3 d5d4 c2c4 a7a5 b4d2 d8b6 d2f4 e5e7 e2g4 c8g4 d1g4');
INSERT INTO evaluations VALUES(42213,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(42213,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(42213,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(42213,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(42213,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(42213,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(42213,36,'cp -9','c1d2 e4d3 f1d3 e8g8 g1f3 f8e8 f3e5 c8g4 d1c1 b8d7 e1g1 d7e5 f4e5 e8e5 d2g5 g4h5 c3e4 b4e7 g5f6 g7f6 c1f4 e5d5 f4h4 h5g6 e4f6 e7f6 f1f6 d5d6 a1f1 d6f6 h4f6 d8f6 f1f6 g6d3 c2d3 a8d8 f6f3 d8d5 g1f2 g8g7 f2e3 d5e5 e3d2 e5h5 h2h3 h5g5 g2g3 g7g6 d2c3 g5h5 h3h4 h5f5 f3e3 h7h5');
INSERT INTO evaluations VALUES(48962,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(48962,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(48962,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(48962,33,'cp -162',NULL);
INSERT INTO evaluations VALUES(48962,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(48962,35,'cp -148',NULL);
INSERT INTO evaluations VALUES(48962,36,'cp -159','d8d5 d1f3 d5e6 g1e2 b8c6 e1g1 e6g6 d2d3 f8e7 e4c3 c8e6 f3e3 b7b6 g1h1 e8c8 f2f4 c6d4 e3d2 f7f6 e2d4 e5d4 c3e4 h8e8 d2f2 g6f5 f1e1 e6d5 c1d2 e7f8 e4g3 e8e1 a1e1 f5d7 b2b3 a7a5 f4f5 d7c6 d2f4 f8b4 e1e2 c6c3 g3h5 d5c6 h5g7');
INSERT INTO evaluations VALUES(33935,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(33935,31,'cp -147',NULL);
INSERT INTO evaluations VALUES(33935,32,'cp -147',NULL);
INSERT INTO evaluations VALUES(33935,33,'cp -173',NULL);
INSERT INTO evaluations VALUES(33935,34,'cp -127',NULL);
INSERT INTO evaluations VALUES(33935,35,'cp -134',NULL);
INSERT INTO evaluations VALUES(33935,36,'cp -142','d1f3 d5e6 b2b3 b8c6 c1b2 e6g6 g1e2 c8g4 f3e3 g4e2 e3e2 g6g2 e1c1 e8c8 h2h4 f7f5 e4g5 g2d5 e2c4 d5c4 b3c4 d8d7 h4h5 f8b4 h1g1 d7e7 g1g2 h8d8 g5f3 b4c5 f3h4 f5f4 d2d3 d8f8 h4f3 f8f5 d1h1 h7h6 f3d2 c5d4 d2f3 d4b2 c1b2 c6d4 f3d4 e5d4');
INSERT INTO evaluations VALUES(57603,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(57603,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(57603,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(57603,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(57603,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(57603,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(57603,36,'cp 57','e2e4 f5e4 c3e4 g8f6 e4f6 e7f6 d2d4 d8e7 f1e2 d7d5 c2c3 c8f5 g1f3 b8c6 e1g1 e8c8 f1e1 e7f7 b2b4 g7g5 a2a4 f8d6 a4a5 c6e7 b4b5 h8e8 c1e3 f5g6 d1b3 e7f5 c3c4 f5e3 f2e3 d5c4 e2c4 f7e7 b5b6 a7a6');
INSERT INTO evaluations VALUES(2399,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(2399,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(2399,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(2399,33,'cp -63',NULL);
INSERT INTO evaluations VALUES(2399,34,'cp -64',NULL);
INSERT INTO evaluations VALUES(2399,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(2399,36,'cp -32','a2a3 b4a6 d2d4 a6c7 d4e5 e6f7 c1d2 f7g8 e1c1 d8e8 h2h3 e8g6 f2f4 g6e4 c3e4 h7h5 h1f1 c8f5 e4g3 g7g6 g3f5 g6f5 c4d3 d5e7 g2g4 h5g4 h3g4 f5g4 f1g1 f8h6 d3e2 g8f7 e2g4 a8d8 g1h1 d8g8 h1g1 f7e8 g4h5 e8f8 h5g4 c7d5 c1b1 h6f4 d2f4 d5f4 d1d7');
INSERT INTO evaluations VALUES(2400,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(2400,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(2400,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(2400,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(2400,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(2400,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(2400,36,'cp -53','b4a6 d2d4 a6c7 d4e5 e6f7 c1d2 f7g8 e1c1 d8e8 h2h3 e8g6 f2f4 g6e4 c3e4 h7h5 h1f1 b7b5 c4b3 c8f5 e4g3 g7g6 g3f5 g6f5 g2g4 f5g4 h3g4 h5g4 f1g1 g8f7 g1g4 h8h5 d2a5 a8c8 c1b1 f8c5 a5c7 c8c7 g4g3 c5b6 g3c3 h5f5 b3d5 c6d5 c3c7 b6c7 d1d5 c7b6 c2c3 f5f4 d5b5 f7e7 b1c2 f4f7 c2d3');
INSERT INTO evaluations VALUES(3009,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(3009,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(3009,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(3009,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(3009,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(3009,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(3009,36,'cp 108','b7b6 h2h3 e7e5 c1d2 d8e7 d1c2 a8b8 a1e1 a7a6 a2a4 f8e8 g2g3 f6h5 g1h2 e7f8 c3d1 g7h6 d1e3 d7f6 a4a5 b6b5 b2b4 b5c4 d3c4 c5b4 e1b1 h6e3 d2e3 f8e7 f1e1 f6d7 c2c1 d7c5 e3c5 d6c5');
INSERT INTO evaluations VALUES(3705,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(3705,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(3705,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(3705,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(3705,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(3705,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(3705,36,'cp 126','c6a5 c4d3 b7b6 e3e4 h7h6 f1e1 c8b7 c1f4 c7c5 f4g3 f6h5 g3h4 d8e8 f3e5 h5f4 d1g4 f7f6 g4f4 f6e5 f4e5 c5d4 c3d4 a5c6 e5g3 c6d4 g3e5 d4b3 a1d1 b3c5 h2h3 c5d3 d1d3 e8f7 f2f3 b7a6 d3d2 a6c4 e5g3');
INSERT INTO evaluations VALUES(4916,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(4916,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(4916,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(4916,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(4916,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4916,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(4916,36,'cp 2','g8f6 e1g1 e8g8 d2d3 d7d6 c3d5 f6d5 b3d5 d8e8 c2c3 a8b8 b2b4 c5b6 a2a4 c6e7 d5b3 e7g6 f3g5 h7h6 d1h5 g6h8 g5f3 c8e6 b3e6 e8e6 f3h4 h8g6 h4f5 g6e7 a4b5 a6b5 f5e7 e6e7 h5g4 g8h8 h2h3 b8a8 c1e3 a8a1 f1a1 b6e3');
INSERT INTO evaluations VALUES(5199,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(5199,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(5199,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(5199,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(5199,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(5199,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(5199,36,'cp 34','g8e7 f1e1 e8g8 e4e5 b7b6 c4d3 h7h6 c1a3 c8b7 f3h4 f8e8 d1g4 e7f5 h4f5 e6f5 g4f5 g7g6 f5f4 d8g5 f4g5 h6g5 a3d6 c6a5 a1d1 a8c8 d3b5 b7c6 b5f1 a5b7 d6a3 b7c5 f2f3 e8e7 a3c1 c8e8 c1g5 e7e5 e1e5 e8e5 g5f4 e5d5 d1d5 c6d5 a2a3 c5b7 f4e5 b7a5');
INSERT INTO evaluations VALUES(7993,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(7993,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(7993,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(7993,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(7993,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(7993,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(7993,36,'cp 30','f1e1 e8g8 e4e5 b7b6 c4d3 h7h6 c1a3 f7f5 e5f6 f8f6 c3c4 c8b7 a3b2 f6f7 d1d2 e7f5 a1d1 a8c8 h2h3 d8e7 d3f5 f7f5 d2d7 e7d7 d1d7 f5f7 d7d6 e6e5 f3e5 c6e5 e1e5 c8c4 e5e8 g8h7 d6d8 c4c2 e8h8 h7g6 d8d6 g6f5 b2d4 c2c6 d6d5 f5e6 d5d8 c6c7 h8g8');
INSERT INTO evaluations VALUES(7678,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(7678,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7678,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(7678,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7678,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7678,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7678,36,'cp 0','d1f3 e8g8 b1c3 d7d6 e1g1 a7a5 h2h3 h7h6 g5f6 e7f6 a1d1 g8h8 c3e2 d8c7 f3d3 f8d8 b2b3 f6h4 f2f4 f7f5 g1h1 a8a7 e4f5 e5e4 d3e4 d6d5 e4e3 d5c4 d1d8 h4d8 e3e8 h8h7 e8g6 h7h8');
INSERT INTO evaluations VALUES(7836,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(7836,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(7836,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(7836,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(7836,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(7836,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(7836,36,'cp 91','d4c6 d7c6 e4e5 f6d7 c1f4 d7c5 e1g1 c5d3 c2d3 b4c3 b2c3 e8g8 f1e1 b7b6 e1e3 d8h4 f4g3 h4g5 h2h4 g5f5 h4h5 h7h6 g3h4 c8b7 g2g4 f5h7 f2f4 c6c5 d1e2 a8e8 a1f1 f7f5 e5f6 g7f6 e3e6 h7d7 f1e1 d7e6 e2e6 e8e6 e1e6 b7c8');
INSERT INTO evaluations VALUES(7815,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(7815,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(7815,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(7815,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(7815,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(7815,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(7815,36,'cp 68','d7c6 e4e5 f6d7 c1f4 d7c5 e1g1 c5d3 c2d3 e8g8 c3e4 b7b6 d1g4 g8h8 a1d1 c8b7 a2a3 b4e7 d3d4 b7a6 f1e1 a6c4 e1e3 c4d5 e4c3 f7f5 e5f6 e7f6 h2h4 f8f7 h4h5 d8d7 c3e4 a8g8 e4d6 f7f8 f4e5 f6e7 g4g3 h7h6 g3g6 e7f6 d6e4 f6e5 d4e5 d7f7 g6f7 f8f7 e4d6 f7f8');
INSERT INTO evaluations VALUES(8240,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(8240,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(8240,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(8240,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(8240,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(8240,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(8240,36,'cp -15','h7h6 g5e3 e7e6 d3d4 f6g4 e3d2 c6d4 f3d4 c5d4 c3b5 g4e5 c4b3 e8g8 b5d4 d8h4 f2f3 e5g4 d2f4 g4f2 f1f2 h4f4 c2c3 a7a5 a2a4 c8d7 d1d3 d7c6 d3d2 f4d2 f2d2 f8d8 d4b5 g7f8 b3c4 d8d7 g2g3 a8d8 a1d1 d6d5 e4d5 f8c5 g1g2 e6d5 c4b3');
INSERT INTO evaluations VALUES(10132,30,'cp -128',NULL);
INSERT INTO evaluations VALUES(10132,31,'cp -130',NULL);
INSERT INTO evaluations VALUES(10132,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(10132,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(10132,34,'cp -131',NULL);
INSERT INTO evaluations VALUES(10132,35,'cp -126',NULL);
INSERT INTO evaluations VALUES(10132,36,'cp -123','c2c3 e8g8 e1g1 h7h6 d2d3 d7d6 c4b3 a7a5 a2a4 d6d5 b3d5 c8g4 f3g3 d8d7 d3d4 f6h5 g3d3 e5d4 d5c6 b7c6 h3f4 h5f4 c1f4 d4c3 d3c3 c5d4 c3g3 f8b8 b1d2 g4e2 f1e1 d7g4 a1b1 g4g3 f4g3 e2d3 g3c7 b8b4 c7d6 b4b2 b1b2 d4b2');
INSERT INTO evaluations VALUES(10106,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(10106,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(10106,32,'cp -122',NULL);
INSERT INTO evaluations VALUES(10106,33,'cp -125',NULL);
INSERT INTO evaluations VALUES(10106,34,'cp -118',NULL);
INSERT INTO evaluations VALUES(10106,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(10106,36,'cp -143','e8g8 e1g1 d7d5 c4d5 c8h3 g2h3 f6d5 e4d5 c6e7 c3c4 c7c6 b1c3 a8c8 d2d3 c5b4 f1d1 e7g6 a1b1 d8d7 c1g5 b4c3 b2c3 h7h6 g5e3 b7b6 f3g4 d7e8 d5d6 c8d8 c4c5 f7f5 g4c4 g8h7 c5b6 a7b6 e3b6 d8d6 b6c5');
INSERT INTO evaluations VALUES(10201,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(10201,31,'cp -89',NULL);
INSERT INTO evaluations VALUES(10201,32,'cp -79',NULL);
INSERT INTO evaluations VALUES(10201,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(10201,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(10201,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(10201,36,'cp -88','a2a3 e8g8 b3b4 c5b6 f1e2 d7d5 b1d2 a7a5 g1f3 a5b4 a3b4 a8a1 b2a1 d5e4 d3e4 d8e7 e1g1 e7b4 d2c4 b4c5 e2d3 c8g4 h2h3 g4f3 d1f3 f6d7 f1b1 f8d8 c4b6 d7b6 b1b5 c5d6 f3g3 f7f6 a1c3 g7g6 d3e2 c6d4 c3d4 d6d4 g3f3 g8g7');
INSERT INTO evaluations VALUES(12086,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(12086,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(12086,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(12086,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(12086,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(12086,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(12086,36,'cp 0','h7h6 g5f6 d8f6 d1e2 f8e7 e1c1 e8g8 c1b1 a7a5 c4b3 c8e6 d1d3 f8b8 h2h4 f6g6 d3g3 g6f6');
INSERT INTO evaluations VALUES(13688,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(13688,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(13688,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(13688,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(13688,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(13688,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(13688,36,'cp 52','e1g1 f8e7 d1e2 e8g8 f1d1 c6a5 c4d3 b7b6 e2e5 e7d6 e5g5 c8b7 g5h4 b7f3 g2f3 g7g6 d3a6 f6d5 h4d8 f8d8 c3d5 e6d5 c1d2 g8g7 b2b3 d8d7 a1c1 a8d8 g1g2 d7c7 c1c7 d6c7 d1c1 c7d6 a6b5 h7h6 f3f4 g7f6 b3b4 a5c4 b5c4 d5c4 c1c4');
INSERT INTO evaluations VALUES(15908,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(15908,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(15908,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(15908,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(15908,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(15908,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(15908,36,'cp 14','h7h6 g5e3 f8e7 f1c4 e8g8 e1g1 f8e8 a2a4 e5d4 e3d4 d7f8 f1e1 c7c6 h2h3 c8e6 c4d3 f8g6 d4e3 d6d5 f3d4 e7b4 d4e6 e8e6 e4d5 f6d5 d3g6 f7g6 e3d2 e6e1 d1e1 b4c3 d2c3 d5c3 e1c3 d8d7 a4a5 a7a6 a1e1 a8d8 c3b4 d7d2');
INSERT INTO evaluations VALUES(16146,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(16146,31,'cp -182',NULL);
INSERT INTO evaluations VALUES(16146,32,'cp -172',NULL);
INSERT INTO evaluations VALUES(16146,33,'cp -184',NULL);
INSERT INTO evaluations VALUES(16146,34,'cp -191',NULL);
INSERT INTO evaluations VALUES(16146,35,'cp -176',NULL);
INSERT INTO evaluations VALUES(16146,36,'cp -201','c2c3 g7g5 h3g1 g5g4 f3d3 h6h5 h2h4 g4h3 g1h3 h8g8 d3e2 f8h6 b1a3 g8g2 d2d4 h6c1 a1c1 d7d6 c4b5 d8e7 d4d5 a7a6 b5a4 b7b5 a4d1 c6d8 a3c2 d8b7 b2b4 g2g6 c2e3 f6e4 c3c4 e4f6 c4b5 a6b5 d1c2 g6g8 c2f5 e8d8 f5c8');
INSERT INTO evaluations VALUES(16265,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(16265,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(16265,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(16265,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(16265,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(16265,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(16265,36,'cp 48','d2d4 f8b4 d1c2 b4c3 c2c3 e8g8 c1g5 d7d6 g5h4 f8e8 e2e3 a7a5 a1d1 d8e7 d4d5 e6d5 c4d5 c6e5 f3e5 e7e5 h4f6 g7f6 f1e2 c8f5 d1d4 f5e4 e2f3 e4f3 g2f3 c7c5 d5c6 b7c6 f3f4 e5e6 h1g1 g8h8 g1g3 d6d5 f4f5 e6f5 c3c6');
INSERT INTO evaluations VALUES(16765,30,'cp -150',NULL);
INSERT INTO evaluations VALUES(16765,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(16765,32,'cp -151',NULL);
INSERT INTO evaluations VALUES(16765,33,'cp -149',NULL);
INSERT INTO evaluations VALUES(16765,34,'cp -148',NULL);
INSERT INTO evaluations VALUES(16765,35,'cp -145',NULL);
INSERT INTO evaluations VALUES(16765,36,'cp -143','f8c5 c1e3 d8e7 f1e2 c6e5 e3c5 e5f3 e2f3 e4c5 f3e2 d7d6 b1c3 e8g8 e1g1 a7a5 e2b5 c8d7 f1e1 e7d8 d1d2 a5a4 h2h3 d7b5 c3b5 d8d7 b5d4 f8e8 b2b4 a4b3 a2b3 a8a1 e1a1 d7e7 a1a7 e7h4 b3b4 c5e4 d4f3 e4d2 f3h4 g7g6 a7b7 e8e1 g1h2 d2f1 h2g1 f1e3 g1h2 e3d5');
INSERT INTO evaluations VALUES(18827,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(18827,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(18827,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(18827,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(18827,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(18827,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(18827,36,'cp -3','g1f3 d7d5 g2g3 g8f6 b1d2 f8e7 f1g2 e8g8 e1g1 b7b5 e4d5 e6d5 f1e1 f8e8 d2f1 c8e6 c1e3 d8b6 d3d4 c5c4 f3e5 a7a5 e5c6 b6c6 e3g5 b5b4 f1e3 a8d8 h2h4 g8h8 g5f6 e7f6 a2a3 b4b3 d1f3 c6c7 h4h5 h7h6 a1d1 a5a4');
INSERT INTO evaluations VALUES(19543,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(19543,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(19543,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(19543,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(19543,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(19543,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(19543,36,'cp 88','d4d5 c6e5 d1d4 e5g6 b1c3 e7e5 d5e6 d7e6 d4d8 e8d8 g2g3 g8f6 f1g2 f8d6 g1f3 c8d7 f3g5 d8e7 e1g1 d7c6 g2c6 b7c6 b2b3 h8d8 c3e4 f6e4 g5e4 a7a5 c1e3 g6e5 e4d6 e7d6 a1d1 d6c7 e3f4 f7f6 f4e5 f6e5 d1d8 a8d8');
INSERT INTO evaluations VALUES(21515,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(21515,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(21515,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(21515,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(21515,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(21515,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(21515,36,'cp 93','d2d4 g8f6 f1d3 c6b4 e4f6 g7f6 d3e2 b7b6 c2c4 d8d7 d4d5 e6d5 a2a3 b4c6 c4d5 c6e5 d1b3 c8b7 f3e5 f6e5 e2b5 c7c6 d5c6 b7c6 b5c6 d7c6 e1g1 f8g7 c1e3 e8g8 a1c1 c6b7 h2h3 f8d8 f1d1 a8c8 c1c8 b7c8 d1d8 c8d8 b3a4 e5e4 b2b4 d8e7 a4c6 h7h6 e3h6 g7h6 c6h6');
INSERT INTO evaluations VALUES(23453,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(23453,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(23453,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(23453,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(23453,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(23453,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(23453,36,'cp -16','e7e5 f4e5 c6e5 g1f3 e5f3 d1f3 d7d5 b2b3 g8f6 c1b2 f8e7 f1d3 c8g4 f3g3 e8g8 b1c3 g4h5 c3e2 h5g6 e1g1 f6e4 d3e4 d5e4 e2f4 e7d6 h2h4 f7f6 g3h3 d8c8 h3c8 a8c8 h4h5 d6f4 f1f4 g6h5 f4e4 h5g6 e4a4 g6c2 a1c1 c2g6 a4a7 f8d8');
INSERT INTO evaluations VALUES(24558,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24558,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24558,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(24558,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(24558,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(24558,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(24558,36,'cp 3','e7e6 d1b3 d5c4 f1c4 d8b6 a1c1 h7h6 g5f6 g7f6 a2a3 b6b3 c4b3 f5g6 f3h4 g6d3 e1d2 d3a6 b3d1 a6c4 d1e2 c4b3 d2e1 a7a5 h4f3 e6e5 d4d5 d7c5 d5c6 b7c6 f3d2 b3e6 c3e4 c5e4 d2e4 f6f5 e4d2 f8e7 c1c6 e8g8 c6c1 a5a4');
INSERT INTO evaluations VALUES(24560,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(24560,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(24560,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(24560,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(24560,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(24560,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(24560,36,'cp 55','d1b3 d8b6 b3b6 d7b6 c4d5 f6d5 h2h4 f7f6 c1d2 g4h5 e3e4 d5c7 d2e3 e8c8 f3d2 e7e6 g2g3 f8e7 a2a4 f6f5 f1e2 h5e2 e1e2 f5e4 c3e4 h7h6 a4a5 b6d5 d2c4 h8f8 c4e5 d5f6 f2f3 c7d5 h4h5 d5e3');
INSERT INTO evaluations VALUES(26075,30,'cp 106',NULL);
INSERT INTO evaluations VALUES(26075,31,'cp 106',NULL);
INSERT INTO evaluations VALUES(26075,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(26075,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(26075,34,'cp 113',NULL);
INSERT INTO evaluations VALUES(26075,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(26075,36,'cp 112','b5d7 d8d7 d1d4 h7h6 b2b3 g8f6 c1b2 e7e6 e1c1 d7c6 f2f4 f8e7 c1b1 a7a6 g2g4 e8c8 h2h3 c8b8 d4e3 f6d7 d1d4 e7f6 d4c4 c6b6 e3d2 d7c5 h1d1 b6c6 a2a4 g7g5 d1f1 c6c7 f4g5 f6g5 d2f2 h6h5 f2f7 h5g4 h3g4');
INSERT INTO evaluations VALUES(25898,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(25898,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(25898,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(25898,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(25898,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(25898,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(25898,36,'cp 84','d8d7 d1d4 d7c6 c1g5 h7h6 g5h4 e7e5 d4b4 a7a5 b4b5 f8e7 h4e7 c6b5 c3b5 e8e7 e1c1 a8a6 d1d3 g8f6 f2f3 h6h5 h1d1 h8d8 a2a4 a6c6 b2b3 e7e6 d3d5 e6e7 d5d2 e7e6 c2c4 g7g5 h2h4 g5h4 c1b1 h4h3 g2h3 e6e7 h3h4 f6e8 d2d5 e8f6 d5d3 f6e8 b1b2 e8c7 b5c7');
INSERT INTO evaluations VALUES(27583,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(27583,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(27583,32,'cp 110',NULL);
INSERT INTO evaluations VALUES(27583,33,'cp 104',NULL);
INSERT INTO evaluations VALUES(27583,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(27583,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(27583,36,'cp 84','d1d4 d7c6 c1g5 h7h6 g5h4 e7e5 d4b4 a7a5 b4b5 f8e7 h4e7 c6b5 c3b5 e8e7 e1c1 a8a6 d1d2 g8f6 f2f3 h8d8 h1d1 a6c6 b2b3 a5a4 b3a4 f6e8 d2d3 d8a8 d3a3 e8c7 b5c3 h6h5 a3b3 a8a5 d1d3 b7b6 c1b2 a5c5 a2a3 g7g6 d3d2 c5c4 g2g3 f7f5 h2h3 c6c5 d2d3 c5c6 b3b4 e7e6');
INSERT INTO evaluations VALUES(26474,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(26474,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(26474,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(26474,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(26474,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(26474,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(26474,36,'cp 48','f1b5 e7e6 b5c6 b7c6 e1g1 g8e7 f3h4 d8d7 b1d2 f6e5 d2b3 e7g6 h4f5 e6f5 d4e5 f8e7 f2f4 g6f8 c1e3 f8e6 d1d3 e8g8 a1d1 a7a5 a2a4 g7g6 b3d4 e6d4 d3d4 f8d8 b2b3 c6c5 d4c3 d7c6 h2h3 c5c4 e3f2 e7b4 c3f3 b4e7 b3c4 c6c4 d1d3');
INSERT INTO evaluations VALUES(26792,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(26792,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(26792,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(26792,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(26792,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(26792,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26792,36,'cp 17','f8d6 f4g3 g8f6 e1g1 e8g8 b1d2 c6b8 b5e2 d6g3 h2g3 b8d7 c2c4 d8e7 d1b3 b7b6 a2a4 c7c5 a4a5 f5g4 f1c1 h7h5 e2f1 g4f5 b3c3 f8d8 c4d5 e6d5 b2b3 f6e4 c3b2 a8c8 a5b6 a7b6 a1a7 c8a8 c1a1 a8a7');
INSERT INTO evaluations VALUES(28057,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(28057,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(28057,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(28057,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(28057,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(28057,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(28057,36,'cp 65','d3d4 d7d6 h2h4 h7h6 f1d3 e7e6 c1f4 b8c6 d1d2 d8e7 e1c1 e8c8 d3c4 e6e5 c3d5 e7f8 d4e5 d6e5 f4h2 f7f5 d2c3 f5f4 c4b5 f8c5 c3c5 b6c5 a2a3 c6d4 b5c4 b7d5 c4d5 g8f6 c2c3 f6d5 c3d4 d5f6 f3e5 f6e4 e5g6 e4f2 g6h8 f2h1 d1h1');
INSERT INTO evaluations VALUES(28138,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(28138,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(28138,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(28138,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(28138,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(28138,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(28138,36,'cp 18','e1g1 g8e7 c2c4 d5c4 e2c4 e8g8 h2h3 g4h5 f1e1 h7h6 a2a3 b8d7 b1c3 e7c6 g2g4 h5g6 c4a2 a7a6 c1d2 f8e8 e1e8 d8e8 d2e3 e8e7 c3d5 e7d8 f3d2 c6e7 d5c3 d7b6 d1f3 c7c6 d2e4 d6c7 a1e1 b6d5 e3d2 g8h8 e4c5 d8d6 c5b7 d6h2 g1f1');
INSERT INTO evaluations VALUES(28137,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(28137,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(28137,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(28137,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28137,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(28137,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(28137,36,'cp 26','g8e7 c2c4 d5c4 e2c4 g4h5 f1e1 e8g8 h2h3 h7h6 g2g4 h5g6 b1c3 e7c6 a2a3 b8d7 c4a2 f8e8 e1e8 d8e8 c1e3 e8e7 c3d5 e7d8 f3d2 d7b6 d1f3 b6d5 a2d5 d8e7 a1d1 a8e8 d2c4 g6e4 f3e4 e7e4 d5e4 e8e4 c4d6 c7d6 d4d5');
INSERT INTO evaluations VALUES(28416,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(28416,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(28416,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(28416,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(28416,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(28416,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(28416,36,'cp 20','c2c4 d5c4 e2c4 g4h5 f1e1 e8g8 h2h3 g8h8 a2a3 f7f6 b1d2 b8c6 d1c2 d6f4 c2d3 f4d2 c1d2 h5f3 d3f3 d8d4 f3c3 e7f5 a1d1 d4c3 d2c3 f5d6 c4d5 d6b5 d5c6 b7c6 c3b4 f8d8 d1d8 a8d8 g1f1 h7h6 e1e7 h8g8');
INSERT INTO evaluations VALUES(32022,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(32022,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(32022,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(32022,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(32022,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(32022,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(32022,36,'cp 28','c4e6 f7e6 f3b3 d8c8 g1f3 b8c6 b1c3 f8g7 c3e2 g8f6 d2d3 e8g8 e1g1 b7b6 c2c3 c6d8 c1g5 c7c5 a1d1 d8c6 g5f6 f8f6 d3d4 f6f3 g2f3 e5d4 c3d4 c6d4 e2d4 g7d4 d1d4 c5d4 f1d1 c8d7 d1d4 d7e7 b3d1 e7g5 g1h1 g5h4 d4d6');
INSERT INTO evaluations VALUES(32021,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(32021,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(32021,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(32021,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(32021,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(32021,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(32021,36,'cp 25','f7e6 f3b3 d8c8 g1f3 b8c6 d2d3 g8f6 e1g1 f8g7 c1e3 f6g4 b1d2 e8g8 e3g5 a7a5 c2c3 h7h6 h2h3 g4f6 g5e3 g6g5 a2a4 c8d7 f1e1 b7b6 d3d4 f6h5 d4e5 c6e5 f3e5 g7e5 a1d1 f8f7 b3c2 a8f8 d2c4 e5g7 b2b4 a5b4 c3b4');
INSERT INTO evaluations VALUES(32459,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(32459,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(32459,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(32459,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(32459,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(32459,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(32459,36,'cp 26','f3b3 d8c8 g1e2 b8c6 d2d3 f8g7 c2c3 a7a5 b1d2 a5a4 b3c2 g6g5 d2f1 h7h6 f1g3 c8d7 g3h5 g8f6 h5f6 g7f6 c1e3 d6d5 e1g1 d5d4 e3d2 h6h5 b2b4 a4b3 a2b3 a8a1 f1a1 d4c3 d2c3');
INSERT INTO evaluations VALUES(32464,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(32464,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32464,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(32464,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(32464,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(32464,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(32464,36,'cp -9','d8d7 d2d4 f8d6 e1g1 g8e7 b1c3 e8g8 c3e2 c7c6 c1f4 e7g6 f4g3 d7c7 e2c1 b8d7 c2c3 g6f4 c1d3 f8e8 d3f4 d6f4 g3f4 c7f4 a2a4 e8e4 d1b3 f4c7 a1e1 d7f6 e1e4 f6e4 f1e1 h7h6 g2g3 a8e8 g1g2');
INSERT INTO evaluations VALUES(32953,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(32953,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(32953,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(32953,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(32953,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(32953,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(32953,36,'cp -18','d2d4 b8c6 e1g1 e8c8 c2c3 f8d6 b2b4 g8f6 b1a3 f6e4 a3b5 h8e8 f3e1 a7a6 b5d6 e4d6 e1d3 a6a5 b4a5 c6a5 c1f4 d6e4 d1c2 g7g5 f2f3 g5f4 f3e4 d5e4 d3c5 d7g4 c5e4 a5c4 e4f6 e8e2 f6g4 e2c2 f1f4 c2c3 f4f7 d8d4 f7f8 c8d7');
INSERT INTO evaluations VALUES(40850,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(40850,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(40850,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(40850,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(40850,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(40850,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(40850,36,'cp 86','e8g8 f1e2 c7c6 e1g1 d6d5 e4e5 f6e4 e2d3 e4c3 b2c3 c6c5 d4c5 b8d7 c5c6 b7c6 f3d4 d7c5 d4c6 d8e8 c6e7 e8e7 d1d2 f8e8 h2h3 c5d3 c2d3 c8f5 f1f3 f5g6 a1e1 e7d7 f4g3 d5d4 c3c4 a8c8 g3f4 a7a5 e1c1 h7h6 f3g3 d7e6 a2a4');
INSERT INTO evaluations VALUES(36272,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(36272,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(36272,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(36272,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(36272,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(36272,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(36272,36,'cp 80','f1e2 c8e6 d1d3 b8a6 f3g5 d6d5 a2a3 f6e4 g5e4 d5e4 d3e4 c7c6 e2d3 g7g6 d3a6 b7a6 e1c1 a8c8 h1e1 f8e8 e4f3 d8d7 h2h3 e6d5 c3d5 c6d5 e1e2 e7f8 d1e1 e8e2 e1e2 f8g7 f4e5 c8e8 b2b3 a6a5 a3a4 h7h5 c1b2 g7e5 e2e5 e8e5');
INSERT INTO evaluations VALUES(42005,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(42005,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(42005,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(42005,33,'cp -67',NULL);
INSERT INTO evaluations VALUES(42005,34,'cp -55',NULL);
INSERT INTO evaluations VALUES(42005,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(42005,36,'cp -62','f6e4 a2a3 b4a5 d1d3 c7c5 d4c5 a5c3 b2c3 e8g8 f4d6 f8e8 f3d2 e4d6 c5d6 d8d6 c3c4 b7b6 c4d5 e6d5 a1d1 b8c6 f1e2 c6e5 d3b3 c8g4 e2g4 e5g4 d2f3 g4f6 e1g1 a8c8 d1d4 d6c7 d4d2 c7e7 f1d1 c8c5 f3d4 g7g6 d4e2 f6e4 d2d4 e8c8 e2f4 e7h4 f4d3 c5c4');
INSERT INTO evaluations VALUES(42836,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(42836,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(42836,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(42836,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(42836,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(42836,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(42836,36,'cp -41','d8e7 f1e2 e8g8 g1f3 f8e8 a2a3 b4d2 d1d2 d7d5 b1c3 d5c4 e1c1 c8e6 h1e1 e7d6 d4d5 e6g4 f3d4 b8d7 f2f3 g4h5 e2c4 e8e1 d1e1 d6h2 g2g4 h2d2 c1d2 h5g6 g4g5 d7b6 c4b3 f6e8 f3f4 a8d8 f4f5 g6h5 c3e4 h7h6 g5g6 f7g6 d5d6 g8h8 d4e6 d8c8 e6f4 g6f5 e4g3 g7g6 g3h5 g6h5 d6c7');
INSERT INTO evaluations VALUES(43350,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(43350,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(43350,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43350,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(43350,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(43350,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(43350,36,'cp 6','f8g7 f1e1 e8g8 h2h3 b7b6 c1g5 c8b7 d1d2 b8d7 a2a4 a8c8 a1d1 f8e8 g5h6 d7c5 h6g7 g8g7 f2f3 e7e5 d4b3 d8c7 d2f2 c8d8 b3c5 d6c5 b2b3 f6h5 d3c4 d8d4 c3e2 d4d8 f2h4 h7h6 h4f2 d8d1 e1d1 e8d8 e2c3 d8d4 d1d4 c5d4 c3d5');
INSERT INTO evaluations VALUES(43466,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43466,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43466,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43466,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(43466,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(43466,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43466,36,'cp 0','f8e7 f1c4 c8e6 c4b3 b8d7 c1e3 b7b5 e1g1 d8c7 a2a4 b5b4 c3d5 e6d5 b3d5 f6d5 d1d5 e8g8 d5d3 f8c8 f1c1 c7b7 f3d2 d7c5 e3c5 c8c5 d2b3 c5c8 b3d2');
INSERT INTO evaluations VALUES(39680,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39680,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39680,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39680,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39680,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39680,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39680,36,'cp 0','f1c4 c8e6 c4b3 b8d7 c1e3 b7b5 e1g1 d8c7 a2a4 b5b4 c3d5 e6d5 b3d5 f6d5 d1d5 e8g8 d5d3 f8c8 f1c1 c7b7 f3d2 d7c5 e3c5 c8c5 d2b3 c5c8 b3d2');
INSERT INTO evaluations VALUES(43848,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(43848,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(43848,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(43848,33,'cp 116',NULL);
INSERT INTO evaluations VALUES(43848,34,'cp 139',NULL);
INSERT INTO evaluations VALUES(43848,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(43848,36,'cp 133','e4e5 f6d5 c3d5 e6d5 g1f3 f8e7 f1d3 d7d6 e1g1 e8g8 f1e1 c8g4 c2c3 d6e5 d4e5 c7c5 h2h3 g4h5 c1f4 b8c6 d3b5 c6a5 f4g3 g8h8 d1e2 d8c8 e5e6 f7e6 e2e6 c8e6 e1e6 e7f6 f3e5 h5e8 b5a6 a5c4 e5c4 d5c4 a6c4 e8f7 e6e4 f7c4 e4c4 a8d8 g3f4 g7g5 f4e3 f8e8');
INSERT INTO evaluations VALUES(43854,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(43854,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(43854,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(43854,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(43854,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(43854,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(43854,36,'cp 49','g1e2 b8c6 a2a3 f8e7 e1g1 e8g8 b1c3 d7d6 d2d3 a7a6 c3d5 f6d5 c4d5 d8e8 f3g3 e7f6 f2f4 e5f4 e2f4 c8e6 f4h5 f6d4 c1e3 d4e3 g3e3 e6d5 e3g3 e8e5 h5f6 g8h8 g3e5 c6e5 f6d5 g7g6 g1f2 h8g7 f2e3 a8c8 h2h3 e5d7 d5b4 a6a5 b4c6 d7b8 c6e7 c8e8');
INSERT INTO evaluations VALUES(44101,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(44101,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(44101,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(44101,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(44101,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(44101,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(44101,36,'cp 1','e5e4 f3d4 d8b6 d4b3 a7a5 d2d3 a5a4 b3d2 f8c5 e1g1 e4e3 f2e3 c5e3 g1h1 f6g4 d2e4 e8g8 b1c3 e3c1 d1c1 d7d5 c4d5 g4e3 f1f3 e3d5 c1d2 b6d8 a1f1 b8d7 h2h3 d5c3 d2c3 d7b6 c3a3 c8e6 e4c5 d8e7 f3f4 e6d5 e2e4 d5e6 f1f2 b6d7 c5b7 e7a3 b2a3');
INSERT INTO evaluations VALUES(44490,30,'cp -142',NULL);
INSERT INTO evaluations VALUES(44490,31,'cp -138',NULL);
INSERT INTO evaluations VALUES(44490,32,'cp -108',NULL);
INSERT INTO evaluations VALUES(44490,33,'cp -135',NULL);
INSERT INTO evaluations VALUES(44490,34,'cp -134',NULL);
INSERT INTO evaluations VALUES(44490,35,'cp -133',NULL);
INSERT INTO evaluations VALUES(44490,36,'cp -109','e7e5 f3g3 d7d5 b1c3 f8d6 d2d3 e8g8 f1e2 b8a6 g1h3 a6c5 e1g1 c5e6 g1h1 d8e7 a2a4 b7b6 e2d1 c8a6 f2f3 a8d8 g3e1 f8e8 e1f2 f6h5 c3e2 d6c5 f2e1 a6c8 a4a5 b6b5 c1d2 h7h6 e2g3 h5f4 h3f4 e5f4 g3f5');
INSERT INTO evaluations VALUES(44620,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(44620,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(44620,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(44620,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(44620,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(44620,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(44620,36,'cp 0','b8c6 b1c3 d7d5 c1f4 c8f5 e2e3 e7e6 f1b5 f6d7 e1g1 f8e7 d1a4 d8b6 a4b3 a7a6 b5c6 b7c6 c3a4 b6b3 a2b3 f7f6 f1c1 a8c8 a4c3 g7g5 f4g3 g5g4 f3h4 f5d3 c1d1 d3c2 d1c1');
INSERT INTO evaluations VALUES(49291,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(49291,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(49291,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(49291,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(49291,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(49291,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(49291,36,'cp -23','d6c4 d3c4 b8c6 e5g4 f8c5 d1e2 d8e7 e2e7 c6e7 c1e3 d7d6 e3c5 c8g4 c5d4 f7f6 f2f3 g4f5 b1a3 a7a6 e1c1 h7h5 h2h4 e8f7 a3b1 f5e6 b2b3 b7b5 c4c5 e7f5 d4f2 d6c5 f2c5 a8b8 b1d2 h8e8 d2e4 b8d8 d1d8');
INSERT INTO evaluations VALUES(49824,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(49824,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(49824,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(49824,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(49824,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(49824,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(49824,36,'cp -25','d3c4 b8c6 e5g4 f8b4 c1d2 d7d5 c4d5 d8d5 g4e3 d5d6 a2a3 b4c5 b1c3 d6g6 c3d5 e8g8 e1g1 c5d6 c2c4 c8e6 d2c3 a8d8 d1b3 g6h5 f2f4 d6c5 g1h1 c6d4 b3b7 d4e2 f1f3 e2c3 b2c3 c5d6 d5c7 d6c7 b7c7 f8e8');
INSERT INTO evaluations VALUES(49422,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49422,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(49422,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(49422,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(49422,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(49422,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(49422,36,'cp -13','b7b6 f1d3 c8b7 g1f3 d7d5 c2c4 d5c4 d3c4 c7c5 e1g1 c5d4 f3d4 b8d7 b2b3 f8c5 c1b2 e8g8 d4f3 d8e7 d1e2 a7a6 a2a4 f8d8 f1d1 c5b4 b1d2 d7c5 a1c1 h7h6 b2d4 a8c8 d4b2 c8c7 f3e5 d8c8 d2f1 c5e4 c4a6 c7c1 d1c1 c8c1');
INSERT INTO evaluations VALUES(51043,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(51043,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(51043,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(51043,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(51043,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(51043,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(51043,36,'cp -36','b7b5 a4b3 c7c5 d2d3 c8b7 g1f3 d8c7 a2a4 c5c4 b3a2 b8c6 e1g1 c6b4 a4b5 a6b5 b1c3 b4a2 a1a2 a8a2 c3a2 c7a5 a2c3 b5b4 c3e2 g8f6 c1f4 a5a6 e2g3 f8e7 d3c4 a6c4 f3e5 c4b5 d1d3 b5b6 f1d1 b6c7 c2c4 b4c3 b2c3');
INSERT INTO evaluations VALUES(51154,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(51154,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(51154,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(51154,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(51154,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(51154,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(51154,36,'cp -41','a4b3 c7c5 d2d3 c8b7 g1f3 d8c7 c2c3 g8f6 d1e2 b8c6 e1g1 c6e7 a2a4 e7g6 f1e1 f8e7 b1a3 c7b6 b3c2 e8g8 c1d2 f8c8 h2h3 d7d5 a4b5 a6b5 e4e5 f6d7 d3d4 c5d4 c2g6 h7g6 f3d4 b5b4 a3b5 b7a6 c3b4');
INSERT INTO evaluations VALUES(51153,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(51153,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(51153,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(51153,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(51153,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(51153,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(51153,36,'cp -25','c7c5 d2d3 d7d5 e4d5 e6d5 c2c3 b8c6 g1f3 g8f6 e1g1 f8e7 d3d4 c5c4 b3c2 e8g8 c1f4 c8g4 b1d2 e7d6 f4g5 d6c7 f1e1 d8d6 h2h3 g4h5 d2f1 a8e8 e1e8 f8e8 f1g3 h5g6 g5f6 d6f6 c2g6 h7g6 g3f1 b5b4 f1e3 f6d6');
INSERT INTO evaluations VALUES(51216,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(51216,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(51216,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(51216,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(51216,34,'cp -71',NULL);
INSERT INTO evaluations VALUES(51216,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(51216,36,'cp -59','b8c6 e1g1 f8e7 b2b3 d7d6 c1b2 c8b7 c3e2 e7f6 c2c3 c6e5 f3e5 d6e5 e2c1 g8e7 d1e2 e7c6 a2a4 b5b4 d3c4 c6a5 c1d3 d8d6 c3b4 c5b4 a1c1 a8c8 h2h3 e8g8 f2f4 a5c4 b3c4 f8d8 f1f3 d6c7 d3e5 f6e5 b2e5');
INSERT INTO evaluations VALUES(51664,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(51664,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(51664,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(51664,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(51664,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(51664,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(51664,36,'cp 16','e7e6 a4b5 c6b5 b2b3 g8f6 b3c4 f6e4 c4b5 f8b4 b1d2 c8b7 f1d3 b4d2 c1d2 e4d2 d1d2 a7a5 e1g1 b8d7 d2e3 a5a4 a1a3 d7b6 f3g5 d8d6 f1a1 a8d8 d3e4 b7e4 g5e4 d6d4 e3d4 d8d4 e4c5 e8e7 c5a4 d4a4 a3a4 b6a4 a1a4 h8b8 a4a7 e7f6');
INSERT INTO evaluations VALUES(51839,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(51839,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(51839,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(51839,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(51839,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(51839,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(51839,36,'cp 34','d7d5 e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 g1f3 e5f3 d1f3 d8e7 e1f1 e7d7 b1c3 d5d4 c3e4 f8e7 h2h4 g8f6 e4f6 e7f6 c1g5 f6g5 h4g5 e8c8 a2a4 d4d3 f3d3 d7d3 c2d3 d8d3 h1h4 d3d5 h4f4 h8f8 a1c1 c8b8 c1e1 d5g5 e1e7 g5c5 f4f7 c5c1 f1e2 c1c2 e2f3 f8f7 e7f7 c2c7 f7f8 c7c8 f8c8 b8c8 f3g4 c8d7 g4f5 d7d6');
INSERT INTO evaluations VALUES(53751,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(53751,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(53751,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(53751,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(53751,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(53751,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53751,36,'cp 41','e4d5 c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 g1f3 e5f3 d1f3 d8e7 e1f1 e7e6 b1c3 g8f6 c1f4 e8c8 a1e1 e6a6 f3d3 a6d3 c2d3 f8b4 e1c1 c8d7 f4e3 h8e8 g2g3 b4c3 c1c3 b7b6 e3d4 d8c8 c3a3 c8a8 d4f6 g7f6 a3a4 e8c8 f1g2 c8c2');
INSERT INTO evaluations VALUES(53696,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(53696,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(53696,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(53696,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(53696,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(53696,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(53696,36,'cp 27','c6d5 c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 d1e2 d8e7 b1c3 e8c8 c1f4 e5c6 g1f3 d5d4 c3e4 e7e6 e1g1 g8f6 f1e1 e6e4 e2e4 f6e4 e1e4 f8c5 g1f1 f7f6 f4d2 h8e8 a1e1 c8d7 e4e8 d8e8 e1e8 d7e8 f1e2 e8f7 e2d3 f7e6 d3c4 c5b6 g2g4 d4d3 c4d3 b6f2 d3e4 f2b6 b2b3 b6c5');
INSERT INTO evaluations VALUES(54602,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(54602,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54602,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(54602,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(54602,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(54602,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(54602,36,'cp 23','c4b5 c8d7 b5d7 b8d7 d4e5 d7e5 d1e2 d8e7 c1f4 e5g6 f4g5 e7e6 g1f3 f8b4 b1d2 e6e2 e1e2 f7f6 g5e3 b4d2 e2d2 g6e5 f3d4 e5c4 d2e2 e8f7 b2b3 c4e3 e2e3 g8e7 e3d3 h8c8 h1e1 e7c6 c2c3 c6d4 d3d4 c8c7 a2a4 a8c8 e1e3 h7h5 h2h4 c7c5');
INSERT INTO evaluations VALUES(52045,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(52045,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(52045,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(52045,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(52045,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(52045,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(52045,36,'cp -32','b7b5 c4e2 c8b7 f3d4 b7e4 d4b5 a7a6 b5d6 f8d6 d1d6 g8e7 b1c3 e4c2 c1f4 c2g6 a1d1 e8g8 g2g4 b8c6 h2h4 h7h6 e1g1 a8a7 h4h5 g6h7 d1d2 d8a5 f1d1 a7b7 d6a3 d7d5 a3a5 c6a5 f4d6 a5c4 e2c4 d5c4 d6e7');
INSERT INTO evaluations VALUES(52559,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(52559,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(52559,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(52559,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(52559,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(52559,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(52559,36,'cp 99','d2d4 c8g4 b1d2 f6e5 d4e5 e7e6 h2h3 g4h5 f1e2 g8h6 e1g1 h6f7 c2c4 d8d7 f1e1 f8e7 f3d4 h5e2 d1e2 c6c5 d4f5 e8g8 f5e7 d7e7 c4d5 e6d5 e5e6 f7d6 e2d3 b8c6 d3d5 f8f5 d5d3 a8f8 d2f1 c5c4 d3g3 c6b4 h3h4 b4d3 c1g5 e7c7 e6e7 f8e8 a1d1 h7h6');
INSERT INTO evaluations VALUES(54270,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54270,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54270,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54270,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54270,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54270,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54270,36,'cp 0','b8c6 e1g1 g8f6 e2e4 d5e4 d3e4 d8d1 f1d1 f6e4 b1a3 f8e7 f3e5 c6e5 g2e4 e8g8 c1f4 e7f6 f4e3 f6e7');
INSERT INTO evaluations VALUES(20311,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20311,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20311,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20311,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20311,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20311,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20311,36,'cp 0','e1g1 g8f6 e2e4 d5e4 d3e4 d8d1 f1d1 f6e4 b1a3 f8e7 f3e5 c6e5 g2e4 e8g8 c1f4 e7f6 f4e3 f6e7');
INSERT INTO evaluations VALUES(56972,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(56972,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(56972,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(56972,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(56972,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(56972,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(56972,36,'cp 25','g8f6 g1f3 b8c6 c2c3 f8c5 b2b4 c5b6 a2a4 a7a6 b1d2 e8g8 e1g1 d7d6 f1e1 c8e6 h2h3 d8d7 c4b3 e6b3 d1b3 f8e8 d2c4 d6d5 c4b6 c7b6 e4d5 f6d5 c1d2 d5f6 c3c4 a8d8 b4b5 c6d4 f3d4 d7d4 d2e3 d4d3 b3d3 d8d3 e3b6 a6b5 c4b5');
INSERT INTO evaluations VALUES(57046,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(57046,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(57046,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(57046,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(57046,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(57046,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(57046,36,'cp -22','d7d5 e4d5 e6d5 d2d4 g8f6 b1c3 f8d6 c1f4 e8g8 f4d6 d8d6 d1d2 c7c6 f2f3 b7b5 e2g3 f8e8 e1f2 a7a6 f1d3 c6c5 h1e1 c8e6 d4c5 d6c5 f2f1 b8c6 c3e2 b5b4 a1d1 a6a5 e2f4 a5a4 h2h4 a4a3 b2b3 e6d7 h4h5');
INSERT INTO evaluations VALUES(55916,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(55916,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(55916,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(55916,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(55916,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(55916,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55916,36,'cp -14','e4d5 e6d5 d2d4 g8f6 b1c3 f8d6 c1f4 e8g8 f4d6 d8d6 d1d2 c7c6 f2f3 b8d7 e1c1 b7b5 d2f4 f6e8 f4d6 e8d6 e2f4 d7b6 f1d3 a7a5 b2b3 a5a4 h1e1 b5b4 c3e2 g7g5 f4h5 a4b3 a2b3 c8f5 d3f5 d6f5 e2g3 f5h4 c1b2 h4g2 e1e7');
INSERT INTO evaluations VALUES(57142,30,'cp -77',NULL);
INSERT INTO evaluations VALUES(57142,31,'cp -87',NULL);
INSERT INTO evaluations VALUES(57142,32,'cp -88',NULL);
INSERT INTO evaluations VALUES(57142,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(57142,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(57142,35,'cp -81',NULL);
INSERT INTO evaluations VALUES(57142,36,'cp -86','e5d4 d1d4 b8c6 d4d3 d7d5 g1f3 g8f6 c2c3 f8e7 c1f4 f6h5 f4g3 h5g3 h2g3 g7g6 e2e3 c8f5 d3d1 d8d6 f1b5 e8g8 b1d2 e7f6 e1g1 c6e5 f3e5 f6e5 d2f3 e5g7 b5d3 f5g4 d1b3 c7c5 f1d1 h7h5 b3b7 a8b8');
INSERT INTO evaluations VALUES(57235,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(57235,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(57235,32,'cp -64',NULL);
INSERT INTO evaluations VALUES(57235,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(57235,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(57235,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(57235,36,'cp -66','d7d5 h3g2 g8f6 d2d3 b8c6 g1f3 c8e6 c2c3 h7h6 e1g1 a7a5 b1d2 f8e7 e2e4 d5e4 d2e4 e8g8 d1e2 d8d7 c1e3 e6g4 f1d1 f6d5 e3c5 a8e8 c5e7 d5e7 e4c5 d7c8 d1e1 b7b6 c5e4 e7g6 e2c2 c8d7');
INSERT INTO evaluations VALUES(56488,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(56488,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(56488,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(56488,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(56488,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(56488,35,'cp -60',NULL);
INSERT INTO evaluations VALUES(56488,36,'cp -60','h3g2 b8c6 d2d4 e5d4 g1f3 f8c5 c2c3 d4c3 b1c3 g8e7 e1g1 e8g8 b2b3 c8g4 c1b2 d8d7 a1c1 a8d8 c3a4 c5d6 a4c5 d6c5 c1c5 f8e8 d1c1 a7a6 b3b4 g4f5 c1d2 h7h6 a2a4 b7b6 c5c1 f5h3 g2h3 d7h3 b4b5 a6b5');
INSERT INTO evaluations VALUES(1738,30,'cp 361',NULL);
INSERT INTO evaluations VALUES(1738,31,'cp 388',NULL);
INSERT INTO evaluations VALUES(1738,32,'cp 377',NULL);
INSERT INTO evaluations VALUES(1738,33,'cp 365',NULL);
INSERT INTO evaluations VALUES(1738,34,'cp 366',NULL);
INSERT INTO evaluations VALUES(1738,35,'cp 369',NULL);
INSERT INTO evaluations VALUES(1738,36,'cp 386','d4d5 c6b8 f3e5 f8b4 c1d2 g8f6 e5c4 a5c5 d1e2 c5e7 e1c1 e8g8 a2a3 b4c5 d2f4 e7d8 h2h3 f8e8 e2f3 c8f5 f1d3 f5g6 c4a5 d8c8 d5d6 c7c6 g2g4 b8d7 f4g3 g6d3 d1d3 c5b6 a5c4 d7c5 d3e3 e8e3 c4e3 c8e6 e3f5 a8e8 h1d1 f6d7 g4g5 f7f6 f5e7 g8h8');
INSERT INTO evaluations VALUES(2107,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(2107,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(2107,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(2107,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(2107,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(2107,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(2107,36,'cp 53','d2d4 c5d4 c3b5 c7b8 b5d4 f8c5 c2c3 g8f6 d1e2 c6e5 f3e5 b8e5 d4b3 c5e7 c4d3 b7b6 c1d2 d7d5 f2f4 e5c7 e4e5 f6e4 b3d4 e8g8 e1g1 a7a5 d2e3 e7c5 g1h1 c8d7 e3g1 c7b7 a1e1 a8e8 d4f3 f7f6 g1d4 c5d4 f3d4 f6e5');
INSERT INTO evaluations VALUES(3377,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(3377,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(3377,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(3377,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(3377,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(3377,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(3377,36,'cp -33','c6a5 c4b3 f6g4 c3d5 g4e3 d5e3 a5b3 a2b3 c7c5 d1e1 a7a5 e3d5 f8e8 f3d2 c8e6 d5e7 e8e7 d2c4 b7b5 c4e3 e7a7 e1d2 h7h6 h2h3 g8h7 g1h2 a7a6 f1e1 d8g5 d2e2 a6a7 e1f1 f7f5 e4f5 e6f5 f2f3 f5g6');
INSERT INTO evaluations VALUES(3683,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(3683,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(3683,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(3683,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(3683,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(3683,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(3683,36,'cp 57','b7b6 b2b3 c6e7 f1d3 c8b7 d1c2 a7a5 c3b2 a5a4 b3b4 d5c4 d3c4 b7e4 c2e2 c7c6 a1c1 b6b5 c4a2 e4d5 a2b1 f6e4 f3e5 f7f6 e5d3 e4d6 f2f3 f6f5 e1g1 e7g6 c1e1 d8e7 f1f2 a8d8 d3c5 e6e5 e3e4 f5e4 f3e4 d5c4 e2e3 d6f7 g1h1');
INSERT INTO evaluations VALUES(3822,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(3822,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(3822,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(3822,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3822,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(3822,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3822,36,'cp 0','c6a5 c4b3 h7h6 h2h3 f8e8 c3a4 c5d4 a4c3 a5b3 a2b3 d4c5 c3a4 c5d4');
INSERT INTO evaluations VALUES(4193,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4193,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(4193,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4193,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4193,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4193,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(4193,36,'cp 0','d7d6 d2d3 a7a6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 c3e2 c6b8 e2g3 b8d7 g1h2 c7c5 d1e2 b7b5 b2b3 d8c7 f1f2 g8h7 f3d2 c7a5 f2f1 a5c3 e2d1 c3a5');
INSERT INTO evaluations VALUES(3625,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(3625,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(3625,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(3625,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(3625,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(3625,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(3625,36,'cp 0','d2d3 a7a6 c1e3 c5e3 f2e3 c8e6 c4e6 f7e6 c3e2 c6b8 g1h2 b8d7 e2g3 d8e8 d1e2 b7b5 f1f2 c7c5 b2b3 a8c8 a3a4 e8d8 f3d2 g8h7 f2f1 d8c7 f1f2');
INSERT INTO evaluations VALUES(5579,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(5579,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(5579,32,'cp 102',NULL);
INSERT INTO evaluations VALUES(5579,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(5579,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(5579,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(5579,36,'cp 112','f1d3 g8e7 g1f3 c8d7 e1g1 a8c8 c1d2 e7g6 a1c1 e8g8 c3e2 g6h4 f3h4 b4d2 d1d2 d8h4 d3b1 h4e7 c1c3 f7f6 e5f6 f8f6 c3h3 g7g6 h3b3 b7b6 h2h4 c8f8 h4h5 g6h5 b3h3 d7e8 e2g3 e7g7 g3h5 e8h5 h3h5 h7h6');
INSERT INTO evaluations VALUES(6084,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(6084,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(6084,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(6084,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(6084,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(6084,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(6084,36,'cp -32','g8e7 d2d3 c5d6 d3d4 e5d4 d1d4 f7f6 c1e3 e7g6 e1c1 d8e7 c1b1 e8g8 d4d3 g6e5 d3e2 b7b5 f3d4 a7a5 d4f5 c8f5 e4f5 f8e8 e3d4 d6b4 g2g4 e5c4 e2f3 c6c5 c3d5 e7f7 d4e3 a8d8 d5b4 c4e3 d1d8 e8d8 f3e3 c5b4');
INSERT INTO evaluations VALUES(8262,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(8262,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8262,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(8262,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(8262,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8262,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8262,36,'cp 0','d2d3 e8g8 h2h3 e7e6 a2a3 a8b8 c4a2 b7b5 c1e3 a7a5 d1d2 b5b4 a3b4 b8b4 e4e5 d6e5 f4e5 f6d7 c3e4 c6e5 f3e5 d7e5 e3g5 d8d4 g5e3');
INSERT INTO evaluations VALUES(10307,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(10307,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(10307,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(10307,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(10307,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(10307,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(10307,36,'cp -12','e4e5 f6d5 f1b5 a7a6 b5a4 d5b6 c3d4 c5b4 b1c3 b4c3 b2c3 b6a4 d1a4 d7d6 e1g1 e8g8 e5d6 c7d6 c1f4 c8e6 f3d2 b7b5 a4a3 a6a5 a3d6 d8d6 f4d6 f8d8 d6c7 d8d7 c7b6 b5b4 c3b4 c6d4 d2f3 d4c2 a1c1 c2b4 b6a5 b4a2 c1c7 h7h6 c7d7 e6d7 a5c7 h6h5 c7f4');
INSERT INTO evaluations VALUES(11592,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(11592,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(11592,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(11592,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(11592,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(11592,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(11592,36,'cp 36','e5d4 f3d4 c6d4 d1d4 e7c6 d4d3 f8e7 b1c3 e8g8 c3d5 c6b4 d5b4 e7b4 c2c3 b4c5 e1g1 d7d6 a4c2 d8h4 b2b4 c5a7 c1e3 a7e3 d3e3 c8d7 a2a4 a8e8 f2f3 e8e6 e3f2 h4f4 a1d1 e6h6 g2g3 f4f6 f2d2 d7e6 f3f4 e6c4');
INSERT INTO evaluations VALUES(11915,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(11915,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(11915,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(11915,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(11915,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(11915,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(11915,36,'cp 78','f8c5 d4e5 d8e7 e5e7 c5e7 c3b5 e7d8 e4e5 f6g8 c4d5 a8b8 b5a7 c8a6 d5e4 c7c6 b2b3 g8e7 c2c4 d8c7 c1b2 b8d8 a2a4 e8g8 f2f4 f8e8 g2g3 c7b8 a4a5 b6a5 a1a5 a6b7 c4c5 f7f6 a7b5 e7c8 e1g1 f6e5 f4e5 b8e5 b2e5 e8e5');
INSERT INTO evaluations VALUES(7576,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(7576,31,'cp 107',NULL);
INSERT INTO evaluations VALUES(7576,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(7576,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(7576,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(7576,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(7576,36,'cp 93','d4e5 d8e7 e5e7 c5e7 c3b5 e7d8 e4e5 f6g8 c4d5 a8b8 c1d2 a7a6 b5c3 g8e7 d5e4 e7g6 f2f4 c8b7 e1c1 b7e4 c3e4 f7f6 e5f6 d8f6 g2g4 e8g8 e4f6 f8f6 f4f5 g6e5 d2c3 d7d6 c3e5 d6e5 d1d7 f6f7 h1d1 b8e8 h2h4 e5e4 d7f7 g8f7 d1d7 e8e7 d7e7 f7e7 c1d2 e7f6 d2e3 f6e5 c2c4');
INSERT INTO evaluations VALUES(12548,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(12548,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(12548,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(12548,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(12548,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(12548,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(12548,36,'cp 47','f1g2 f8e7 e1g1 d7d6 a2a3 e8g8 d2d3 h7h6 a1b1 a8b8 b2b4 a7a6 f3d2 c8e6 b4c5 d6c5 g2c6 b7c6 b1b8 d8b8 d1a4 b8b6 d2f3 f6d7 c1e3 f8a8 f1b1 b6c7 a4b3 a8b8 b3c2 b8b1 c3b1 c7b7 b1d2 f7f5 h2h3 b7c7 c2a4');
INSERT INTO evaluations VALUES(15164,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(15164,31,'cp -51',NULL);
INSERT INTO evaluations VALUES(15164,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(15164,33,'cp -53',NULL);
INSERT INTO evaluations VALUES(15164,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(15164,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(15164,36,'cp -37','f8e7 e4d5 f6d5 c3d5 d8d5 f1e2 c8f5 e1g1 a7a5 c1g5 e7g5 f3g5 f7f6 e2h5 f5g6 h5f3 d5d7 g5e4 c6d4 e4c5 d7b5 f3b7 a8d8 c2c3 b5c5 c3d4 d8d4 a1c1 c5b6 d1f3 e8g8 b7e4 c7c5 f3e2 b6b5 f1e1 f8d8 h2h3 a5a4 c1c3 g6e4 d3e4 b5e2 e1e2 c5c4 e2c2');
INSERT INTO evaluations VALUES(15829,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15829,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(15829,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(15829,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(15829,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(15829,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(15829,36,'cp 7','d4e5 f8e7 e1g1 e8g8 c1f4 f8e8 b1c3 c7c6 a1d1 c8d7 c3e4 f6e4 e5e4 d8c8 c4d3 g7g6 e4d4 c6c5 d4c3 e7f8 f1e1 d7c6 h2h3 e8e1 c3e1 b7b5 b2b3 c8b7 d3f1 h7h5 e1c3 b7b6 c3g3 a8d8 d1d8 b6d8 f4e3 d8f6');
INSERT INTO evaluations VALUES(16539,30,'cp -201',NULL);
INSERT INTO evaluations VALUES(16539,31,'cp -204',NULL);
INSERT INTO evaluations VALUES(16539,32,'cp -219',NULL);
INSERT INTO evaluations VALUES(16539,33,'cp -194',NULL);
INSERT INTO evaluations VALUES(16539,34,'cp -207',NULL);
INSERT INTO evaluations VALUES(16539,35,'cp -187',NULL);
INSERT INTO evaluations VALUES(16539,36,'cp -193','f8c5 g1e2 d7d5 e2g3 c8e6 f1e2 h7h5 h2h4 c5e7 c1g5 d5d4 d1c1 f6d7 g5e7 d8e7 b1a3 e7c5 c3d4 c5d4 a3c2 d4b6 b2b3 e8c8 c1d2 f7f6 d2e3 b6e3 c2e3 d7c5 a1d1 c6b4 g3f5 b4a2 d3d4 e5d4 d1d4 c5b3 d4d8 h8d8 f5g7');
INSERT INTO evaluations VALUES(16822,30,'cp 186',NULL);
INSERT INTO evaluations VALUES(16822,31,'cp 166',NULL);
INSERT INTO evaluations VALUES(16822,32,'cp 191',NULL);
INSERT INTO evaluations VALUES(16822,33,'cp 190',NULL);
INSERT INTO evaluations VALUES(16822,34,'cp 187',NULL);
INSERT INTO evaluations VALUES(16822,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(16822,36,'cp 176','h2h3 g4f6 f3g5 d7d5 e4d5 c6d4 d2d3 f8d6 a2a4 e8g8 g5e4 h7h6 e1g1 c7c6 c1e3 d6c7 d5c6 b7c6 e4f6 d8f6 c3e4 f6g6 e3d4 e5d4 d1f3 c8d7 f1e1 c7a5 e1e2 a8e8 g1f1 a5b4 a4a5 d7e6 c4e6 e8e6 c2c3 d4c3 b2c3 f7f5 e4f6 g6f6 e2e6 f6e6 c3b4');
INSERT INTO evaluations VALUES(16980,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(16980,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(16980,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(16980,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(16980,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(16980,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(16980,36,'cp 43','f3b3 d8e7 d2d3 c7c6 a2a4 e7b4 b1c3 b4b3 c4b3 d7d6 f2f3 f8e7 h2h4 h7h6 c1e3 c8e6 b3e6 f7e6 h4h5 e8g8 e1e2 f6d7 c3d1 a7a5 d1f2 f8f7 f2h3 a8f8 h1f1 d6d5 g2g4 e7c5 e3d2 c5d4 d2a5 d4b2 a1b1 b7b6');
INSERT INTO evaluations VALUES(17212,30,'cp 96',NULL);
INSERT INTO evaluations VALUES(17212,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(17212,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(17212,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(17212,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(17212,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(17212,36,'cp 86','d7d6 e5e6 f7e6 c1e3 g7g6 b1c3 f8g7 h2h4 e6e5 d4d5 c6d4 f3d4 e5d4 e3d4 g7d4 d1d4 e8g8 f2f3 e7e5 d5e6 c8e6 h4h5 d8f6 d4f6 f8f6 b2b3 e6f5 h5g6 h7g6 e1d2 f6f7 h1h4 b6d7 a1e1 d7f6 f1d3 f5d3 d2d3 a8e8 h4h1 e8e1 h1e1');
INSERT INTO evaluations VALUES(17359,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17359,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17359,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(17359,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17359,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17359,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(17359,36,'cp -1','d7d5 e4d5 e6d5 c4b3 f8e7 c2c4 g8f6 c4d5 f6d5 d3d4 c8e6 d4c5 e8g8 b1d2 d5f4 a2a3 e7c5 d2e4 e6b3 d1b3 f4e2 g1h1 e2d4 f3d4 c5d4 c1f4 d8h4 g2g3 h4h5 a1e1 h5b5 b3b5 a6b5 e1c1 d4b2 c1b1 b2d4 b1b5 a8a3 b5b7');
INSERT INTO evaluations VALUES(20912,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(20912,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(20912,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(20912,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(20912,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(20912,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(20912,36,'cp 43','f1g2 f8g7 e1g1 g8f6 c2c4 d5c4 b1a3 a7a5 a3c4 a5a4 f1e1 c8e6 c4e5 e6d5 d1d3 d5e4 e5c6 b7c6 d3a3 e8g8 c1d2 e4f3 g2f3 d8d4 e1d1 d4b6 a1b1 f8b8 d2c3 f6d5 c3g7 g8g7 f3d5 c6d5 d1d5 b6b4 a3e3 b8d8 a2a3 b4b3 e3b3 a4b3 d5d8');
INSERT INTO evaluations VALUES(20913,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(20913,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(20913,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(20913,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(20913,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(20913,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(20913,36,'cp 48','g8f6 e1g1 f8g7 b2b3 e8g8 c1b2 f8e8 h2h3 b7b5 d1c1 e7e6 b1d2 c8b7 e2e3 a7a6 c2c4 d5c4 b3c4 c6a5 c4b5 a6b5 c1c2 f6d7 f1c1 d7b6 c2c7 d8c7 c1c7 e8c8 c7c8 a8c8 f3e1 b7g2 g1g2 a5c4 d2c4 c8c4 e1f3 c4c2 b2a3 c2c3');
INSERT INTO evaluations VALUES(22800,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(22800,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(22800,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(22800,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(22800,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(22800,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(22800,36,'cp -37','d7d5 g1f3 g8f6 b1d2 f8c5 f1e2 e8g8 e1g1 h7h6 f1e1 a7a5 e2f1 f8e8 e4d5 d8d5 d2e4 f6e4 d3e4 d5d1 e1d1 c8e6 f1b5 f7f6 b5c6 b7c6 a2a4 g8f7 b2b3 c5b4 c1e3 c6c5 f3d2 b4c3 a1b1 c3d4 d2c4 e6c4 b3c4 d4e3 f2e3 f7e6');
INSERT INTO evaluations VALUES(22989,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(22989,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(22989,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(22989,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(22989,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(22989,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(22989,36,'cp 90','c1e3 a7a6 d1d2 d7d6 f2f3 f8e7 g2g3 g8f6 h2h4 h7h5 g1h3 e8g8 h3g5 c7c6 f1e2 d8a5 e1g1 a8b8 b2b3 a5c7 a1d1 b7b5 d5c6 c7c6 a2a4 b5b4 c3d5 f6d5 d2d5 c6c3 g1f2');
INSERT INTO evaluations VALUES(23824,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(23824,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(23824,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(23824,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(23824,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(23824,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(23824,36,'cp 30','c3a4 c5b6 c2c3 g8f6 a4b6 a7b6 d1e2 d8d7 c4b5 e8g8 e2e3 h5f3 e3f3 c6d4 c3d4 d7b5 c1h6 b5b2 d4d5 f6e8 h6g5 f7f6 g5e3 b2b5 f3f5 f8f7 f1d1 a8a3 f5c8 b5d7 c8b7 f6f5 e3g5 f5e4 d3e4 a3h3 g2h3 d7h3 b7b8 g8f8 g5e3 h3g4 g1f1 g4e4');
INSERT INTO evaluations VALUES(24555,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(24555,31,'cp 57',NULL);
INSERT INTO evaluations VALUES(24555,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(24555,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(24555,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24555,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(24555,36,'cp 44','d8b6 d1d2 e7e6 c4c5 b6d8 b2b4 f8e7 h2h3 e8g8 f1d3 f5d3 d2d3 b7b6 e1g1 a7a5 a2a3 d8c8 c3a4 a5b4 a3b4 b6b5 a4b2 f6e4 f3d2 e4d2 d3d2 c8b7 b2d3 a8a1 f1a1 f8a8 d2b2 h7h6 a1a5 e7d8 a5a3 a8a3 b2a3 d7f6 f2f3 d8c7 f4d6 f6e8 d6c7 b7c7');
INSERT INTO evaluations VALUES(24674,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(24674,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(24674,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(24674,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(24674,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(24674,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(24674,36,'cp 33','h2h3 g4d7 e1g1 e7e6 c2c4 f8e7 c1e3 e8g8 b1c3 e6e5 f3d2 a8b8 f2f4 a7a6 b2b3 b7b5 d2f3 f8e8 f1f2 e5f4 g3f4 b5c4 b3c4 f6h5 c3d5 e7f8 a1b1 c6b4 d1d2 h7h6 d5b4 b8b4 b1b4 c5b4 d2b4');
INSERT INTO evaluations VALUES(25029,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(25029,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(25029,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(25029,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(25029,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(25029,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(25029,36,'cp 58','g7g6 h2h3 g4f3 e2f3 f8g7 c3e2 e8g8 e1g1 a7a5 c2c3 e7e5 d1b3 a5a4 b3b7 d8d7 b7b5 c6d4 e2d4 e5d4 b5d3 c7c5 c3d4 c5d4 d3d4 d7e6 d4e3 f6d7 e4e5 a8d8 e5d6 e6d6 f1d1 d6b4 e3e4 b4e4 f3e4 f8e8 c1g5 g7b2 g5d8');
INSERT INTO evaluations VALUES(25684,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(25684,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(25684,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(25684,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(25684,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(25684,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(25684,36,'cp -42','g8f6 d2d3 f8d6 c3b1 g4h5 b1d2 f6d7 d2c4 d7c5 h2h3 f7f6 c1e3 c5e6 c2c3 d8e7 c4a5 e6d8 a5c4 e8g8 g2g4 h5f7 c4d6 c7d6 f3h4 d8e6 h4f5 e7d7 d3d4 e5d4 c3d4 d6d5 f2f3 a8d8 d1d2 d5e4 f3e4 f8e8 d2g2 g8h8 a1d1 f7g6 h3h4');
INSERT INTO evaluations VALUES(27492,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(27492,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(27492,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(27492,33,'cp -58',NULL);
INSERT INTO evaluations VALUES(27492,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(27492,35,'cp -70',NULL);
INSERT INTO evaluations VALUES(27492,36,'cp -73','g4d7 h1e1 g8e7 e2f1 f7f6 b1c3 a6a5 d2d3 h7h5 a2a4 e8c8 b2b3 c6c5 c1b2 e7c6 e1e2 c6b4 a1e1 h5h4 c3b5 b7b6 b5a3 d7f5 f3d2 c8b7 d2c4 f5g6 b2c1 h8h5 c1e3 b4c6 a3b5 h5d5 f2f3 c6d4');
INSERT INTO evaluations VALUES(28534,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(28534,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(28534,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(28534,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(28534,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(28534,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(28534,36,'cp 122','b8d7 f1c4 f8e7 c4b3 d7c5 g5f6 e7f6 e1g1 e8g8 b3d5 a8c8 d1f3 b7b5 c3e4 c5e4 f3e4 g8h8 f1d1 g7g6 a2a4 f6g7 a4b5 a6b5 a1a3 g6f5 e4f5 d8f6 f5f6 g7f6 c2c3 b5b4 c3b4 f6d8 a3a6 c8b8 b2b3 d8b6 d5c4 b6d4 g1f1');
INSERT INTO evaluations VALUES(28953,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(28953,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(28953,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(28953,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(28953,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(28953,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28953,36,'cp 11','d8b6 b3b6 a7b6 c4d5 f6d5 c3d5 c6d5 c1d2 b8c6 f1b5 e8d8 a2a3 c6a5 a1c1 f5d7 b5d3 e7e6 g1f3 f8d6 d2c3 d8e7 f3d2 f7f6 h2h4 h7h5 e1g1 h8c8 g2g3 e7f7 f1e1 d7a4 e3e4 d5e4 d2e4 d6e7 c3d2 c8c1 e1c1');
INSERT INTO evaluations VALUES(29088,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(29088,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29088,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29088,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29088,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29088,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29088,36,'cp 0','b8d7 e1g1 e7e6 d1e2 f8e7 f3e5 d7e5 d4e5 f6d5 c4d3 g6d3 e2d3 d8c7 d3e2 a8d8 f1e1 e8g8 g3h5 g8h8 e2g4 f8g8 a2a3 b7b5 g4f3 g8f8 f3g4');
INSERT INTO evaluations VALUES(29697,30,'cp 112',NULL);
INSERT INTO evaluations VALUES(29697,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(29697,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(29697,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(29697,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(29697,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(29697,36,'cp 104','d5c4 e5g4 f6g4 e2e3 g4f6 f1c4 e7e6 d1b3 d8c8 b1c3 a7a6 c4e2 b7b5 e2f3 a8a7 e1g1 c7c5 c1d2 c5c4 b3d1 f8e7 c3e2 e8g8 a1c1 a6a5 b2b3 e7a3 c1c3 a3b4 c3c2 f8d8 b3c4 b5c4 e3e4 b4d2 d1d2');
INSERT INTO evaluations VALUES(31536,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31536,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31536,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31536,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(31536,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31536,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(31536,36,'cp 15','c6c5 f1b5 b8d7 e1g1 a7a6 b5e2 c5d4 f3d4 f5g6 f2f4 f8c5 d2f3 g8e7 c1e3 d7b6 b2b3 b6d7 d1d2 e8g8 g2g4 a8c8 a2a4 g6e4 c2c3 d8c7 f1c1 c5a7 a4a5 e4f3 e2f3 d7c5 c3c4 f8d8 d2e2 d5c4 c1c4');
INSERT INTO evaluations VALUES(31676,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(31676,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(31676,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(31676,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(31676,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(31676,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(31676,36,'cp 24','f1b5 b8d7 e1g1 a7a6 b5e2 f5g6 c2c3 a8c8 f3e1 g8e7 h2h4 h7h6 h4h5 g6f5 e2d3 d8b6 d3f5 e7f5 d2f3 f8e7 d1d3 b6b5 d3b5 a6b5 g2g4 f5h4 f3h4 e7h4 c1d2 e8g8 d2e3 c5d4 e3d4 d7c5 a1c1 h4e7 f2f4 c5e4 c1c2');
INSERT INTO evaluations VALUES(32134,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(32134,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(32134,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(32134,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(32134,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(32134,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(32134,36,'cp 69','g4h5 d2d4 b8d7 c1e3 g8f6 b1d2 f8e7 d1b3 e8g8 f1d3 a8b8 b3c2 a7a5 e1g1 h7h6 f1b1 c7c6 d4e5 d6e5 b2b4 f6h7 d3f1 h7g5 f3g5 e7g5 d2c4 g5e3 c4e3 d8g5 a2a4 f8d8 b1e1 d7f6 b4a5 h5f3 e3c4 f6h5 g2g3 g8h8 h3h4');
INSERT INTO evaluations VALUES(34985,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34985,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(34985,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(34985,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(34985,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(34985,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(34985,36,'cp 34','b8c6 g1f3 g8f6 a2a3 c6d4 f1e2 f8e7 e1g1 c7c6 f3d4 e5d4 c3e4 a5d8 e4f6 e7f6 e2g4 e8g8 f1e1 c8g4 d1g4 d8d5 e1e4 a8e8 a1e1 h7h5 g4d1 e8e4 e1e4 g7g6 h2h3 a7a6 a3a4 g8g7 b2b3 b7b5 d1f3 f8h8 d2f4 c6c5 g2g4 h5g4 h3g4 c5c4 b3c4 b5c4');
INSERT INTO evaluations VALUES(35366,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(35366,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(35366,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(35366,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(35366,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(35366,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(35366,36,'cp -17','b1c3 h7h6 g5e3 b8c6 a2a3 f8e8 f1e2 c8e6 e1g1 f6d5 c3d5 e6d5 f1e1 a7a5 f3d4 c6d4 e3d4 g7g6 e2f1 a5a4 g2g3 e7f8 f1g2 d5g2 g1g2 f8g7 d4g7 g8g7 g2f1 h6h5 e1e8 a8e8 b2b3 c7c5 a1e1 e8e1 f1e1 a4b3 c2b3 g7f6 e1d2');
INSERT INTO evaluations VALUES(35364,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(35364,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(35364,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(35364,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(35364,34,'cp -20',NULL);
INSERT INTO evaluations VALUES(35364,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(35364,36,'cp -27','h7h6 g5e3 b8c6 a2a3 f8e8 f1e2 c8e6 e1g1 f6d5 c3d5 e6d5 f1e1 e7f6 c2c3 c6e7 h2h3 e7f5 e3d2 a7a5 f3h2 a5a4 e2g4 e8e1 a1e1 d5e6 d3d4 g8f8 h2f1 c7c6 d2f4 a8d8 g4e2 f5e7 e2g4 e6b3 f1e3 c6c5 d4c5 d6c5 f4c7');
INSERT INTO evaluations VALUES(35745,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(35745,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(35745,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(35745,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(35745,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(35745,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(35745,36,'cp 14','b8d7 c4d5 f6d5 g5e7 d8e7 a1d1 h7h6 h2h3 d5c3 c2c3 c6c5 f1b5 b7b6 e1g1 c8b7 b5d7 b7f3 g2f3 e7d7 d4c5 d7c6 g1g2 a8c8 d1c1 c6c5 c3c5 c8c5 c1c5 b6c5 f1c1 f8b8 b2b3 b8b5 f3f4 g8f8 g2f3 f8e7 h3h4 e7d6 c1d1 d6c6 d1d8 b5a5 d8c8 c6d7 c8g8 g7g6');
INSERT INTO evaluations VALUES(2957,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(2957,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(2957,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2957,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(2957,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(2957,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(2957,36,'cp 32','a2a3 a7a6 c4d5 c6d5 f1d3 h7h6 g5h4 b7b5 e1g1 c8b7 f1c1 a8c8 c2d1 e7d6 h4g3 d7b6 c1c2 d6g3 h2g3 b6c4 a1c1 d8d6 d1e1 d6b6 f3d2 c8c7 a3a4 f8c8 a4b5 a6b5 d2b3 b7c6 c3e2 c6e8 e2f4 b5b4 b3c5 f6d7');
INSERT INTO evaluations VALUES(35942,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(35942,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(35942,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(35942,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(35942,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(35942,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35942,36,'cp 31','e1g1 e7e6 f1e1 b8c6 h2h3 f8e8 e2f1 a7a6 a1b1 c5d4 e3d4 b7b6 c1g5 c6e7 d4d5 e6d5 c4d5 h7h6 g5f6 g7f6 a2a4 f6g7 d1b3 c8d7 c3e4 e7c8 b1a1 d7f5 e4c3 e8e1 a1e1 b6b5 a4b5 a8b8 b3b4 f5d7 e1c1 d8b6 c3e4 a6b5');
INSERT INTO evaluations VALUES(35941,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(35941,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(35941,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(35941,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(35941,34,'cp 17',NULL);
INSERT INTO evaluations VALUES(35941,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(35941,36,'cp 24','e7e6 b2b3 b7b6 c1b2 c8b7 d1d2 f6e4 c3e4 b7e4 f1d1 d8e7 d4c5 g7b2 d2b2 b6c5 d1d2 f8d8 a1d1 b8c6 h2h4 f7f6 f3h2 c6e5 f2f4 e5f7 h2g4 e6e5 g4f2 e4c6 e2f3 c6f3 g2f3 f6f5 g1f1 e5e4 f3e4 f5e4 f1g2 e7h4 f2e4');
INSERT INTO evaluations VALUES(36113,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(36113,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(36113,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(36113,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(36113,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(36113,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(36113,36,'cp -20','e1g1 f8e8 f1e1 c8e6 e2d3 b8d7 h2h3 h7h6 g5e3 a8c8 d1c1 a7a6 e3f4 d6b4 f3e5 c6c5 f4h6 c5d4 h6g7 d7e5 c1h6 e5g6 d3g6 f7g6 h6g6 e6f7 g6h6 f6h7 g7d4 b4f8 e1e8 d8e8 h6h4 e8e6 h4g3 e6g6 g3e5 f8g7 c3d5 g7e5 d5e7 g8f8 e7g6 f7g6 d4e5 c8c2 a1e1 f8f7 e5c3');
INSERT INTO evaluations VALUES(39048,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(39048,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(39048,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(39048,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(39048,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(39048,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(39048,36,'cp 23','b8c6 e1g1 c6a5 c4e2 a5c6 f1e1 d7d6 c3d5 g8f6 e2c4 e8g8 c2c3 f6d5 c4d5 c6e7 d5c4 c8g4 h2h3 g4f3 d1f3 c7c6 c4b3 d6d5 d2d4 e5d4 c1g5 h7h6 g5e7 d8e7 e4d5 e7d6 d5c6 d6c6 b3d5 c6c7 c3d4 g7d4 a1c1');
INSERT INTO evaluations VALUES(39913,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(39913,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(39913,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(39913,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39913,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39913,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(39913,36,'cp 0','c5d4 e3d4 e8g8 f1d3 d7d5 c4d5 b4c3 b2c3 d8d5 g1e2 d5g2 h1g1 g2f3 g1g3 f3h1 g3g1');
INSERT INTO evaluations VALUES(42354,30,'cp 375',NULL);
INSERT INTO evaluations VALUES(42354,31,'cp 374',NULL);
INSERT INTO evaluations VALUES(42354,32,'cp 402',NULL);
INSERT INTO evaluations VALUES(42354,33,'cp 396',NULL);
INSERT INTO evaluations VALUES(42354,34,'cp 363',NULL);
INSERT INTO evaluations VALUES(42354,35,'cp 407',NULL);
INSERT INTO evaluations VALUES(42354,36,'cp 404','f3e5 e8g8 e5f7 f8f7 c4f7 g8f7 e4e5 d8e8 d1e2 b8c6 e5f6 e8e2 c3e2 f7f6 c2c3 b7b5 d2d4 c5b6 c1e3 d7d6 h2h4 c6e7 e1d2 c8b7 e2f4 b5b4 c3b4 g7g5 f4h3 b7g2 h1h2 g5g4 h2g2 g4h3 g2g3 h3h2 d2e2');
INSERT INTO evaluations VALUES(43086,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(43086,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(43086,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(43086,33,'cp 111',NULL);
INSERT INTO evaluations VALUES(43086,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(43086,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(43086,36,'cp 103','b1c3 h7h6 c3e2 b8c6 e1g1 d7d6 c1e3 c6b4 d3b5 c8d7 c2c4 f7f5 b5d7 d8d7 e4f5 e7f5 d1b3 a7a5 a2a3 b4c6 d4d5 e6d5 c4d5 c6e5 f3d4 a5a4 b3c2 e8g8 d4e6 f5e3 f2e3 f8f1 a1f1 c7c6 e2f4 c6d5 e6g7 g8g7 c2d2 a8f8 d2d5 g7h7 f1d1 b7b5 d5d6 d7d6 d1d6 f8e8 d6a6 e5c4');
INSERT INTO evaluations VALUES(43083,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(43083,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(43083,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(43083,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(43083,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(43083,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(43083,36,'cp 90','h7h6 c1e3 d7d5 e4e5 c7c5 d4c5 b8d7 c3b5 e8g8 d1d2 h6h5 e1c1 a7a6 b5d6 b7b6 c5b6 d7e5 f3e5 g7e5 d6c8 e7c8 b6b7 a8b8 d3a6 c8d6 d2b4 d8c7 b4b6 c7b6 e3b6 d6b7 b6d4 e5d4 d1d4 b7c5 a6e2 c5e4 f2f3 e4d6');
INSERT INTO evaluations VALUES(44613,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44613,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44613,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44613,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44613,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44613,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44613,36,'cp 0','d7d5 c1g5 b8d7 e2e4 e7e5 d4f5 d5d4 c3d5 a7a6 d1d3 d8a5 g5d2 a5d8');
INSERT INTO evaluations VALUES(45142,30,'cp -100',NULL);
INSERT INTO evaluations VALUES(45142,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(45142,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(45142,33,'cp -104',NULL);
INSERT INTO evaluations VALUES(45142,34,'cp -107',NULL);
INSERT INTO evaluations VALUES(45142,35,'cp -113',NULL);
INSERT INTO evaluations VALUES(45142,36,'cp -112','b8c6 h3f4 e7e6 d2d3 c6d4 f3d1 b7b5 c4b3 a7a5 c2c3 d4b3 a2b3 f8e7 e1g1 e8g8 f1e1 c8b7 b1d2 d8d7 c3c4 b5b4 d2f1 b7c6 d1e2 d7b7 c1e3 a5a4 b3a4 c6a4 f4h5 f6h5 e2h5 a4c6 e3g5 a8a1 e1a1 f7f6 g5h4 f8a8 a1a8 b7a8');
INSERT INTO evaluations VALUES(46821,30,'cp 83',NULL);
INSERT INTO evaluations VALUES(46821,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(46821,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(46821,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(46821,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(46821,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(46821,36,'cp 70','g1f3 f8g7 e1g1 c7c5 d4c5 d8c7 h2h3 c7c5 d1e2 b8c6 b1c3 e8g8 e3e4 e7e5 c3b5 a7a6 c1e3 c5e7 b5c3 c8e6 a1d1 b7b5 c4d5 a8c8 a2a4 e6d5 c3d5 e7b7 d5f6 g7f6 a4b5 b7b5 e2b5 a6b5 d1d5 f8d8 f1d1 d8d5 d1d5 b5b4 g1f1 f6e7');
INSERT INTO evaluations VALUES(46822,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(46822,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(46822,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(46822,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(46822,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(46822,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(46822,36,'cp 69','c7c5 e1g1 c5d4 d1b3 e7e6 f1d1 a7a6 f3d4 d8c7 b1c3 f8g7 c3a4 b8d7 c4e6 e8g8 e6d7 c8d7 h2h3 f6e4 a4b6 a8d8 b6d7 c7d7 b3c2 d7e7 c1d2 e4d2 d1d2 g7d4 d2d4 d8d4 e3d4 e7d7 c2c5 f8d8 a1d1 d7d5 c5d5 d8d5 a2a3 d5b5');
INSERT INTO evaluations VALUES(47064,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47064,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(47064,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47064,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47064,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47064,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47064,36,'cp 0','d5b6 c4b3 c7c5 d2d3 b8c6 e5d6 e7d6 e1g1 f8e7 a2a4 c6a5 b3a2 e8g8 b1c3 a5c6 a2b3 c6a5');
INSERT INTO evaluations VALUES(46917,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(46917,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(46917,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46917,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46917,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(46917,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(46917,36,'cp 21','c4b3 c7c5 d2d3 b8c6 e5d6 e7d6 e1g1 f8e7 a2a4 e8g8 b1c3 c8e6 b3e6 f7e6 a4a5 b6d7 d3d4 d6d5 d1e2 f8f6 d4c5 h7h6 c1e3 a8c8 c3b5 e6e5 a5a6 b7a6 b5d6 e7d6 c5d6 f6d6 a1a6 d6g6 f1d1 e5e4');
INSERT INTO evaluations VALUES(47941,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(47941,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(47941,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(47941,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(47941,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47941,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(47941,36,'cp 36','d6d5 e4e5 f6d7 f3f4 b8c6 c2c3 f8c5 f1b5 c5b6 e1g1 c6e7 b5a4 e8g8 a4c2 c7c5 d4b5 f7f6 e5e6 d7b8 f4f5 a7a6 b5a3 b8c6 d1g4 c6e5 g4g3 b6c7 c1f4 g7g5 b1d2 g8h8 f4e5 c7e5 g3h3 b7b5 d2f3 e5c7 a1e1 d8d6 a3b1 c5c4 g1h1 b5b4 b1d2 b4c3');
INSERT INTO evaluations VALUES(49923,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(49923,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(49923,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(49923,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(49923,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(49923,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(49923,36,'cp 59','f1g2 f8b4 d1c2 b8c6 g1f3 b7b6 e1g1 c8b7 d2d4 e8g8 d4d5 c6a5 d5e6 d7e6 a2a3 b4c3 c2c3 f6e4 c3c2 d8f6 c1e3 c7c5 b2b4 a5c6 b4c5 e4c5 e3c5 b6c5 a1b1 a8b8 b1b5 e6e5 f1b1 b7a6 b5c5 e5e4 b1b8 f8b8 f3d2 a6b7 c2a4 c6d4 e2e3 d4f3 d2f3 e4f3');
INSERT INTO evaluations VALUES(50798,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(50798,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(50798,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(50798,33,'cp 113',NULL);
INSERT INTO evaluations VALUES(50798,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(50798,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(50798,36,'cp 124','b1c3 b8c6 g1f3 c8d7 f1d3 d8b6 d3c2 g8e7 e1g1 c6a5 f1e1 h7h6 b2b3 e7c6 a2a3 c6a7 c1d2 f8e7 h2h4 b6c7 e1e3 e8c8 c2d3 c8b8 a1c1 c7b6 b3b4 a5c4 d3c4 d5c4 d4d5 e6d5 c3d5 b6e6 d5e7 e6e7 c1c4 d7e6 c4c1 a7c6');
INSERT INTO evaluations VALUES(51458,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(51458,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(51458,32,'cp 173',NULL);
INSERT INTO evaluations VALUES(51458,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(51458,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(51458,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(51458,36,'cp 159','c4b5 g8f6 b1c3 c7c6 b5a4 f8c5 g1f3 d7d6 e1g1 e8g8 h2h3 a7a5 f1e1 d8b6 e1e2 b8d7 d2d3 b6c7 a4b3 h7h6 c3a4 c8a6 c2c4 f8d8 a4c5 d7c5 e2e1 c5e6 c1e3 c6c5 b3a4 a6b7');
INSERT INTO evaluations VALUES(51457,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(51457,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(51457,32,'cp 152',NULL);
INSERT INTO evaluations VALUES(51457,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(51457,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(51457,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(51457,36,'cp 165','c7c6 b5a4 g8f6 d2d3 d7d5 e4d5 c8g4 g1e2 d8a5 c2c3 f6d5 f2f3 g4e6 e1g1 f8e7 d3d4 e8g8 d4e5 b8d7 f3f4 e6g4 a4c2 a5a6 c2d3 a6b6 g1h1 f7f6 h2h3 g4h5 d3f5 a8d8 e5e6 d7e5 d1e1 e5c4 e2d4 c6c5 d4b3 h5g6 f5e4 g6e4 e1e4 b6a6 f4f5');
INSERT INTO evaluations VALUES(51562,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(51562,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(51562,32,'cp 166',NULL);
INSERT INTO evaluations VALUES(51562,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(51562,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(51562,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(51562,36,'cp 173','b5a4 g8f6 b1c3 f8c5 g1f3 d7d6 e1g1 e8g8 d2d3 a7a5 c3e2 d8c7 c2c3 c8g4 e2g3 f6h5 d3d4 c5a7 h2h3 h5g3 f2g3 g4h5 g1h1 b8d7 d4e5 d6e5 d1e2 d7c5 a4c2 c5e6 h1h2 a8d8 c2b3 e6c5 c1g5 d8d7 g3g4 c5b3 a2b3');
INSERT INTO evaluations VALUES(53992,30,'cp 85',NULL);
INSERT INTO evaluations VALUES(53992,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(53992,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(53992,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(53992,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(53992,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(53992,36,'cp 71','d2d4 c5d4 f3d4 g8f6 b1c3 f8e7 c1e3 e8g8 f1e2 a7a6 e1g1 b7b6 f2f4 c8b7 d1c2 d8c7 a1d1 h7h6 g1h1 b8d7 a2a4 f8c8 b2b3 c8e8 e2f3 e7f8 g2g4 e6e5 d4f5 e5f4 e3f4 d7e5 c3d5 b7d5 e4d5 e5f3 f1f3');
INSERT INTO evaluations VALUES(54444,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(54444,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(54444,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(54444,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(54444,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(54444,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(54444,36,'cp 46','f8d6 g1f3 g8e7 e1g1 e8g8 f1e1 h7h6 c1e3 b8c6 a2a3 e7d5 e3d2 d5f6 c4d3 c6e7 d1e2 d6c7 a1c1 b7b6 g2g4 c8b7 g4g5 h6g5 d2g5 a8c8 g5f6 g7f6 d3e4 b7e4 c3e4 g8g7 g1h1 e7g6');
INSERT INTO evaluations VALUES(56634,30,'cp 264',NULL);
INSERT INTO evaluations VALUES(56634,31,'cp 263',NULL);
INSERT INTO evaluations VALUES(56634,32,'cp 275',NULL);
INSERT INTO evaluations VALUES(56634,33,'cp 265',NULL);
INSERT INTO evaluations VALUES(56634,34,'cp 257',NULL);
INSERT INTO evaluations VALUES(56634,35,'cp 251',NULL);
INSERT INTO evaluations VALUES(56634,36,'cp 256','d1d8 e8d8 f3e5 c8e6 b1c3 b8d7 e5d3 c7c6 c1f4 g8f6 f2f3 h7h5 e2e4 h5h4 e1c1 g7g6 f1e2 f6h5 f4g5 f8e7 g5e3 d8c7 f3f4 b7b5 e2f3 h5f6 e3d4 h4h3 g2g3 f6g4 d4h8 a8h8 c3e2 a7a5');
INSERT INTO evaluations VALUES(35411,30,'cp 284',NULL);
INSERT INTO evaluations VALUES(35411,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(35411,32,'cp 266',NULL);
INSERT INTO evaluations VALUES(35411,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(35411,34,'cp 249',NULL);
INSERT INTO evaluations VALUES(35411,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(35411,36,'cp 255','e8d8 f3e5 c8e6 b1c3 b8d7 e5d3 c7c6 e2e4 d8c7 c1e3 a8e8 e1c1 f7f6 a2a3 g8e7 e3f4 c7c8 f1e2 g7g5 f4e3 e7g6 e2h5 f8d6 g2g3 d6c7 f2f4 g5f4 g3f4 f6f5 e4e5 c7b6 e3b6 a7b6 a3a4 d7c5 d3c5 b6c5 h2h4 c5c4 h5f3 c8c7 a4a5');
INSERT INTO evaluations VALUES(35464,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(35464,31,'cp 267',NULL);
INSERT INTO evaluations VALUES(35464,32,'cp 266',NULL);
INSERT INTO evaluations VALUES(35464,33,'cp 269',NULL);
INSERT INTO evaluations VALUES(35464,34,'cp 265',NULL);
INSERT INTO evaluations VALUES(35464,35,'cp 275',NULL);
INSERT INTO evaluations VALUES(35464,36,'cp 276','f3e5 c8e6 e2e4 g8f6 f2f3 f6d7 e5d3 b8c6 d3f4 a7a5 a2a4 f8c5 f4e6 f7e6 c2c3 d8e7 e1d1 h7h6 h2h4 c6e5 d1c2 c7c6 h4h5 a8d8 b1a3 d8f8 a3c4 e5c4 f1c4 c5d6 c1e3 f8f7 a1d1 d6f4 e3f2 f4d6 c4e2');
INSERT INTO evaluations VALUES(35448,30,'cp 271',NULL);
INSERT INTO evaluations VALUES(35448,31,'cp 270',NULL);
INSERT INTO evaluations VALUES(35448,32,'cp 268',NULL);
INSERT INTO evaluations VALUES(35448,33,'cp 267',NULL);
INSERT INTO evaluations VALUES(35448,34,'cp 270',NULL);
INSERT INTO evaluations VALUES(35448,35,'cp 260',NULL);
INSERT INTO evaluations VALUES(35448,36,'cp 264','c8e6 b1c3 b8d7 e5d3 c7c6 e2e4 d8c7 c1e3 a8e8 e1c1 f7f6 a2a4 g8h6 f2f3 f8d6 h2h4 h6f7 d3f4 f7d8 f4e2 d6b4 e2d4 e6f7 g2g3 a7a5 c3e2 b4d6 f1h3 d7e5 e2f4');
INSERT INTO evaluations VALUES(57397,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(57397,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(57397,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(57397,33,'cp 84',NULL);
INSERT INTO evaluations VALUES(57397,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(57397,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(57397,36,'cp 75','g7g6 h5d1 b8c6 g1f3 f6f5 d2d3 h7h6 b1c3 d7d6 a2a4 g8e7 b2b4 c6b4 a1b1 b4c6 e1g1 f8g7 c1a3 h8f8 f1e1 f5e4 d3e4 b7b6 d1d3 d8d7 b1d1 c8b7 c3d5 g6g5 d5c7 d7c7 a3d6 c7d7 d3b3 c6a5 d6e7 d7d1 e1d1 a5b3 e7f8 e8f8 d1d7 b7e4 d7f7 f8g8 c2b3 e4g6 f7e7 g8f8');
INSERT INTO evaluations VALUES(57534,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(57534,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(57534,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(57534,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(57534,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(57534,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(57534,36,'cp 94','h5d1 d7d6 d2d4 b8c6 g1f3 c8g4 c2c3 d8d7 d4d5 c6d8 h2h3 g4f3 d1f3 f6f5 c4d3 f5e4 d3e4 d8f7 g2g4 f8g7 c1e3 g7h6 e3h6 g8h6 b1d2 e8c8 e1c1 c8b8 f3e3 h6g8 c3c4 g8f6 e4c2 d7e7 c1b1 g6g5 f2f3 a7a6 a2a3');
INSERT INTO evaluations VALUES(4673,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4673,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4673,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4673,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4673,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4673,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4673,36,'cp 0','d1f3 f7e6 b1c3 c6b4 a2a3 b4c2 e1d1 c2a1 c3d5 e6d7 h1e1 e7f6 d3d4 h8e8 f3f5 d7c6 f5f3');
INSERT INTO evaluations VALUES(4672,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4672,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4672,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4672,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4672,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4672,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4672,36,'cp 35','f7e6 b1c3 c6b4 a2a3 b4c2 e1d1 c2a1 c3d5 e6d7 h1e1 e7f6 c1d2 a7a5 b2b4 h8e8 b4a5 e8e6 d2c3 d8e8 d5f6 g7f6 d1d2 d7d8 e1a1 a8a6 c3b4 a6c6 f3d5 e8d7 d5d7 c8d7 c4d5 h7h5 g2g3 d8c8 a1e1 c6a6 d5e6 d7e6 f2f3 a6c6 e1f1 e6d5 h2h4 c8d8 f3f4 e5e4 f4f5 e4d3');
INSERT INTO evaluations VALUES(6369,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(6369,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(6369,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(6369,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(6369,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(6369,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(6369,36,'cp 25','b4c3 b2c3 g8f6 f3e5 d7d6 e5c6 b7c6 g2g3 e8g8 f1g2 c8d7 e1g1 f8e8 c4c5 h7h6 c3c4 e6e5 c5d6 c7d6 d4e5 d6e5 d1a4 d8c8 f1d1 c6c5 a4a5 d7c6 g2f1 c6b7 c1e3 c8c6 f2f3 e5e4 f3f4 f6g4 a5c5 g4e3 c5e3 c6c4');
INSERT INTO evaluations VALUES(6819,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(6819,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(6819,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(6819,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(6819,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(6819,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(6819,36,'cp 74','c2c3 d4c3 b2c3 b4a5 b5c4 c7c6 d1h5 d7d5 e4d5 d8e7 e1d1 g8f6 h5g5 e8g8 c1a3 c6c5 h1e1 e7d6 g5g3 d6b6 g3e3 f8d8 e3c5 f6d5 c4d5 c8g4 d1c2 b6c5 d5f7 g8f7 a3c5 a8c8 e1e5 a5c7 e5e3 c7a5 c5a3 f7g8 c2b2 c8c6 a3b4 a5b4 c3b4 c6b6 a2a3 b6d6 b1c3 d6d3 e3d3 d8d3 a1e1 g4f5');
INSERT INTO evaluations VALUES(7246,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(7246,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(7246,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(7246,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(7246,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(7246,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(7246,36,'cp 43','c1e3 c5e3 f2e3 d7d6 d1d2 c8e6 c4e6 f7e6 e1c1 d8b8 c3b5 b8d8 g2g4 d8e7 g4g5 h6g5 f3g5 e8c8 d1g1 d6d5 b5c3 c8b8 h2h4 h8h6 d2e2 d8f8 g5f3 d5d4 c3b1 e7c5 e3d4 e5d4 b1d2 f6h5 c1b1 e6e5 g1g5 h5f4 e2h2 c5e7 h1g1 f4e6 g5g2');
INSERT INTO evaluations VALUES(8424,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(8424,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(8424,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(8424,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(8424,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(8424,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(8424,36,'cp 32','e8f7 d1f3 f7e6 b1c3 c6b4 a2a3 b4c2 e1d1 c2d4 f3d5 d8d5 c3d5 e6d6 d5e7 d6e7 c1e3 d4f5 e3d2 f5d4 d2c3 g7g5 c3d4 e5d4 h2h4 g5g4 h1e1 e7d6 e1e4 c7c5 b2b4 b7b6 e4f4 d6e5 f4f7 c8e6 c4e6 e5e6 f7g7 a7a5 b4c5 b6c5 g7g4 a8g8 g4g8 h8g8 g2g3 e6d5 a1b1 g8f8 d1e2 d5c6 b1b2 a5a4 b2b1 f8e8 e2f3 e8f8 f3g2 f8e8 b1c1 e8e2');
INSERT INTO evaluations VALUES(14131,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(14131,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(14131,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(14131,33,'cp 195',NULL);
INSERT INTO evaluations VALUES(14131,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(14131,35,'cp 185',NULL);
INSERT INTO evaluations VALUES(14131,36,'cp 181','c3b5 e7e5 f3e5 f8d6 e2e3 e8g8 f1e2 f8e8 b5d6 c7d6 e5c6 b7c6 e1g1 c6c5 c2c4 c8a6 d4c5 d6c5 c4d5 a6e2 d1e2 d8d5 f1d1 d5e6 a2a3 a8d8 a1c1 e6b3 h2h3 d8d4 f4c7 a7a6 g1h2 g8h7 d1d4 c5d4 c7f4 d4e3 f4e3');
INSERT INTO evaluations VALUES(14977,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(14977,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(14977,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(14977,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(14977,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(14977,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(14977,36,'cp 37','f6d5 d4c6 b7c6 c3d5 d8d5 c1f4 d5e6 d1e2 f8b4 f4d2 b4d6 d2c3 e8g8 e2e6 c8e6 f1d3 f8d8 e1c1 e6d5 h1e1 d8e8 h2h3 g7g6 d3f1 f7f5 a2a3 e8e1 c3e1 g8f7 c2c4 d5e6 b2b4 c6c5 c1c2 g6g5 b4b5 a8b8 f1d3 h7h5');
INSERT INTO evaluations VALUES(15291,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(15291,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(15291,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(15291,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(15291,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(15291,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(15291,36,'cp 48','d4c6 b7c6 c3d5 d8d5 c1f4 d5e6 d1e2 f8b4 f4d2 b4d6 d2c3 e8g8 e2e6 c8e6 f1d3 h7h5 e1g1 e6d5 b2b3 h5h4 h2h3 f8e8 f1e1 e8e1 c3e1 f7f6 d3g6 d5f7 g6f7 g8f7 g1f1 a8b8 a1d1 c6c5 e1c3 g7g5 f1e2 c5c4 b3c4 b8b6 a2a4 f7g6 e2f3 b6c6 d1d4');
INSERT INTO evaluations VALUES(15278,30,'cp 444',NULL);
INSERT INTO evaluations VALUES(15278,31,'cp 469',NULL);
INSERT INTO evaluations VALUES(15278,32,'cp 467',NULL);
INSERT INTO evaluations VALUES(15278,33,'cp 475',NULL);
INSERT INTO evaluations VALUES(15278,34,'cp 484',NULL);
INSERT INTO evaluations VALUES(15278,35,'cp 471',NULL);
INSERT INTO evaluations VALUES(15278,36,'cp 470','c8e6 d5e3 e6c4 e3c4 e5e4 f3e5 c6e5 c4e5 d8d4 f2f4 f8c5 d1e2 e8g8 c2c3 d4d5 e2c4 a8d8 a2a4 f8e8 c4d5 d8d5 e1d1 c5d6 e5c4 d5d3 h1f1 h7h5 b2b4 h5h4 d1c2 a7a6 c1a3 b7b5 c4e3 h4h3 g2h3 c7c6 a3b2 d6c7 a4b5');
INSERT INTO evaluations VALUES(17850,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(17850,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(17850,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(17850,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(17850,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(17850,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(17850,36,'cp 64','g7g5 d2d4 d7d6 c1e3 f8g7 d4e5 d6e5 b1d2 d8f6 a2a4 g8e7 f3e1 e7g6 a4a5 e8g8 d1f3 f8d8 f3f6 g7f6 e1d3 a8b8 d3c5 f6e7 d2b3 g8g7 g2g3 c8h3 f1d1 d8d1 a1d1 e7c5 e3c5 b7b6 a5b6 c7b6 d1d6 b6c5 d6c6');
INSERT INTO evaluations VALUES(17951,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(17951,31,'cp 219',NULL);
INSERT INTO evaluations VALUES(17951,32,'cp 225',NULL);
INSERT INTO evaluations VALUES(17951,33,'cp 219',NULL);
INSERT INTO evaluations VALUES(17951,34,'cp 220',NULL);
INSERT INTO evaluations VALUES(17951,35,'cp 221',NULL);
INSERT INTO evaluations VALUES(17951,36,'cp 204','d2d4 d7d6 h2h3 c6a5 c4d3 g8e7 e1g1 c7c5 d4e5 d6e5 c1e3 c8e6 e3c5 a5c4 b2b3 c4b2 d1c1 b2d3 c2d3 e7c6 c5f8 h8f8 c1e3 c6d4 a1d1 e8f7 c3e2 d4e2 e3e2 d8b6 d3d4 e5d4 d1d3 a8d8 f1d1 f8e8 f3d4 f7g8 e2f3 h7h6 a2a4 e6f7 f3g4');
INSERT INTO evaluations VALUES(20984,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(20984,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(20984,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(20984,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(20984,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(20984,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(20984,36,'cp 15','c8g4 g1f3 e7e6 e2e3 f8d6 f4g3 g8f6 f1e2 e8g8 b1d2 c6e7 c2c4 b7b6 c4d5 e6d5 b2b4 c7c6 e1g1 a7a5 b4b5 c6b5 e2b5 d6g3 h2g3 d8d6 d1b3 g4f5 f3h4 f5e4 a1c1 f8c8 b5e2 h7h6 h4f3 c8c1 f1c1 a8c8 c1c8 e7c8 d2e4 d5e4');
INSERT INTO evaluations VALUES(21200,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(21200,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(21200,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(21200,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(21200,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21200,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(21200,36,'cp 37','e5f3 d1f3 g7g6 c2c4 e7e6 f1d3 f8g7 e1g1 g8f6 c1g5 h7h6 g5d2 e8g8 b1c3 e6d5 c4d5 a7a6 h2h3 f8e8 f1e1 b7b5 a2a3 a8b8 a1c1 f6d7 c3d1 c8b7 f3g3 b8c8 d1c3 d7e5 d3f1 c7c5 d5c6 c8c6 d2e3 e5c4 f1c4 c6c4 e1d1');
INSERT INTO evaluations VALUES(21331,30,'cp 158',NULL);
INSERT INTO evaluations VALUES(21331,31,'cp 161',NULL);
INSERT INTO evaluations VALUES(21331,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(21331,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(21331,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(21331,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(21331,36,'cp 162','d1d4 g8e7 d4c3 c8e6 c4e6 f7e6 c3b3 d8c8 b1c3 e7c6 c3e2 f8e7 e2f4 e8f7 f4h5 h8f8 c1d2 e7f6 f2f4 f7g8 h5f6 f8f6 d2c3 f6f7 f4f5 e6e5 f1f2 b7b6 a1f1 c8d7 h2h3 a8f8 g1h2 g8h8 b3c4 c6d8 f1d1 d8c6 c4e2 c6e7 g2g4 e7g6');
INSERT INTO evaluations VALUES(21812,30,'cp -21',NULL);
INSERT INTO evaluations VALUES(21812,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(21812,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(21812,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(21812,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(21812,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(21812,36,'cp -21','c2c3 g8f6 h2h3 f8c5 c3d4 c6d4 e2d4 d8d4 d1c2 c5b4 c1d2 b4d2 c2d2 d4d6 d2c3 e8g8 g1f3 f8e8 a1c1 c8d7 c3c5 a8c8 c5d6 c7d6 c1c8 e8c8 e1d2 h7h6 f1e2 g8f8 h1c1 c8c1 d2c1 f8e7 c1d2 e7e6 d2e3 d6d5 d3d4 e5d4 f3d4 e6d6 e4d5 f6d5 e3e4 d5f6 e4f4 g7g6 e2f3 g6g5 f4e3 b7b6');
INSERT INTO evaluations VALUES(23257,30,'cp 214',NULL);
INSERT INTO evaluations VALUES(23257,31,'cp 235',NULL);
INSERT INTO evaluations VALUES(23257,32,'cp 232',NULL);
INSERT INTO evaluations VALUES(23257,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(23257,34,'cp 204',NULL);
INSERT INTO evaluations VALUES(23257,35,'cp 201',NULL);
INSERT INTO evaluations VALUES(23257,36,'cp 207','f3d4 g8e7 a2a3 c6d4 d1d4 e7c6 d4e3 f8d6 c1d2 d6e5 f1c4 d7d6 f2f4 e5c3 e3c3 c8d7 c4d5 d8e7 e1g1 e8c8 a3a4 c6b8 d2e3 a7a6 b2b4 d7c6 b4b5 c6d5 e4d5 e7f7 c3b3 h8e8 b5a6 b8a6 c2c4 a6c5 e3c5 d6c5 f1b1');
INSERT INTO evaluations VALUES(23787,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23787,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23787,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23787,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23787,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23787,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23787,36,'cp 0','h2h3 g4e6 c1e3 g8e7 d3d4 e7g6 b1d2 f7f6 d4e5 f6e5 f3g5 e6g8 d1h5 d8e7 g5f3 g8f7 f3g5');
INSERT INTO evaluations VALUES(23786,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(23786,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23786,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23786,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23786,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23786,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23786,36,'cp 0','g4e6 c1e3 g8e7 d3d4 e7g6 b1d2 f7f6 d4e5 f6e5 f3g5 e6g8 d1h5 d8f6 g5f3 g8e6 f1d1 e6f7 f3g5 f7g8 g5f3');
INSERT INTO evaluations VALUES(24892,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(24892,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(24892,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(24892,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(24892,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(24892,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(24892,36,'cp 26','g4f3 g2f3 e7e5 d4e5 c6e5 c4d5 f8b4 c1d2 b4d2 b1d2 d8d5 d1a4 e5d7 a1d1 e8g8 f1g2 d7c5 d2e4 d5f5 e4c5 f5c5 e1g1 c5b6 b2b4 c7c6 d1b1 a7a6 a4a5 b6d8 f3f4 g7g6 a5c5 d8c7 c5e5 c7e5 f4e5 f6d5 g2d5 c6d5 f1d1 a8d8 b4b5 a6b5 b1b5');
INSERT INTO evaluations VALUES(26470,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(26470,31,'cp -62',NULL);
INSERT INTO evaluations VALUES(26470,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(26470,33,'cp -79',NULL);
INSERT INTO evaluations VALUES(26470,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(26470,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(26470,36,'cp -32','e7e5 d4e5 f6e5 f4g5 f8e7 g5e7 d8e7 f1b5 e5e4 h2h3 g4d7 f3d4 g8f6 b5c6 b7c6 d4e2 e8g8 c2c4 f6h5 d1c2 f8f7 c2b3 e7f6 h1f1 f6d6 g2g4 h5f6 e1c1 a8f8 c1b1 f6g4 h3g4 d7g4 e2c3 g4d1 f1d1 f7f2 c4d5 c6d5');
INSERT INTO evaluations VALUES(27118,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(27118,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(27118,32,'cp -45',NULL);
INSERT INTO evaluations VALUES(27118,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(27118,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(27118,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(27118,36,'cp -27','d4e5 f6e5 f4g5 f8e7 g5e7 d8e7 f1b5 e5e4 h2h3 g4d7 b5c6 b7c6 f3d4 g8f6 d4e2 c6c5 c2c4 d5d4 e3d4 e4e3 f2e3 e8g8 d2f3 e7e3 d1d2 c5d4 d2e3 d4e3 e1g1 a8e8 a1d1 c7c5 f1e1 d7c6 e2g3 c6f3 g2f3 e8e6 g1g2 h7h5 g3e2 f8e8 f3f4 h5h4');
INSERT INTO evaluations VALUES(29506,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(29506,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(29506,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(29506,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(29506,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(29506,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(29506,36,'cp 23','d8d5 e1g1 b8c6 c1e3 e7e6 b1d2 d5h5 c2c4 f8d6 d2f3 e8g8 h2h3 h7h6 a2a3 h5f5 a1c1 a7a6 b2b4 f8e8 f1e1 a8d8 d1b3 d8a8 e2g3 f5g6 g3f1 a8d8 c1d1 b7b5 e3c1 g6f5 f1e3 b5c4 e3c4 c6e7 b3a4 f5g6');
INSERT INTO evaluations VALUES(32618,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(32618,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(32618,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(32618,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(32618,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(32618,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(32618,36,'cp 36','e1g1 b8c6 c1e3 e7e6 b1d2 d5h5 d2f3 f8d6 c2c4 e8g8 a2a3 h5f5 h2h3 h7h6 f1e1 f8e8 b2b4 a7a6 a1c1 a8d8 d1b3 d8a8 e1d1 e8d8 b3b2 d8d7 d1e1 a8d8 b2b3 d6e7 b4b5 a6b5 c4b5 c6a5 b3a4 d8a8 e2g3');
INSERT INTO evaluations VALUES(32617,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(32617,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(32617,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(32617,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(32617,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(32617,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(32617,36,'cp 39','e7e6 b1d2 f8e7 c2c4 d5d6 d2f3 e8g8 c1f4 d6a6 d1c2 f8e8 f1d1 b7b5 c4b5 a6b5 a1c1 e7d6 e2c3 b5b7 f4g3 b8d7 f3e5 d7b6 a2a3 h7h6 e5d3 b6d5 c3a4 a7a5 a4c5 b7b5 d1e1 a8b8 g3e5 b8d8 c2e2 d6c5');
INSERT INTO evaluations VALUES(30417,30,'cp 205',NULL);
INSERT INTO evaluations VALUES(30417,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(30417,32,'cp 202',NULL);
INSERT INTO evaluations VALUES(30417,33,'cp 214',NULL);
INSERT INTO evaluations VALUES(30417,34,'cp 213',NULL);
INSERT INTO evaluations VALUES(30417,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(30417,36,'cp 175','c7c6 b2b3 e7e5 d4e5 f8b4 d1d8 e8d8 c1d2 c4b3 c3b5 b4d2 e1d2 b3b2 a1b1 c6b5 f1b5 b8d7 b5d7 d8d7 b1b2 g8h6 g1e2 h8b8 b2b8 a8b8 h2h3 a7a5 h1a1 b8a8 e2c3 h6g8 c3a4 d7e8 a1c1 g8e7 c1c5 g7g5 g2g3 h7h5 e5e6 f7f6 f2f4 h5h4 f4g5 f6g5 g3g4 a8a6 d2e3 a6e6 c5g5 e6a6 g5g7 e7c6 g7h7 c6b4 h7h4');
INSERT INTO evaluations VALUES(30469,30,'cp 191',NULL);
INSERT INTO evaluations VALUES(30469,31,'cp 195',NULL);
INSERT INTO evaluations VALUES(30469,32,'cp 200',NULL);
INSERT INTO evaluations VALUES(30469,33,'cp 218',NULL);
INSERT INTO evaluations VALUES(30469,34,'cp 206',NULL);
INSERT INTO evaluations VALUES(30469,35,'cp 186',NULL);
INSERT INTO evaluations VALUES(30469,36,'cp 210','b2b3 e7e5 d4e5 f8b4 d1d8 e8d8 c1d2 c4b3 c3b5 b4d2 e1d2 c6b5 f1b5 b3b2 a1b1 b8d7 b5d7 d8d7 b1b2 g8h6 g1e2 h8b8 b2b8 a8b8 h2h3 a7a5 h1a1 b8a8 e2c3 h6g8 d2e3 g8e7 f2f4 e7c6 a1d1 d7e7 d1d6 c6b4 d6b6 a8d8 c3b5 b4c2 e3e2 a5a4 b6a6 c2d4 b5d4 d8d4 e2e3 d4c4 e3f3 c4b4 f4f5 b4b3 f3f4 a4a3 a6a7 e7e8 g2g3 f7f6');
INSERT INTO evaluations VALUES(31351,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(31351,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(31351,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(31351,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(31351,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(31351,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(31351,36,'cp 33','b8c6 e1g1 g8f6 c2c4 d5c4 b1a3 e7e6 a3c4 a8c8 c4e5 f8d6 e5c6 b7c6 c1d2 e8g8 a1c1 d8b6 b2b3 b6d4 d2e3 d4e5 g2g3 d6a3 c1c4 c6c5 f3b7 c8d8 d1c2 f6d5 e3c5 a3c5 c4c5 d8d7 b7c6 d7d6 f1c1 f8d8 c6d5 d6d5 c5c8 e5d6');
INSERT INTO evaluations VALUES(31354,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(31354,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(31354,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(31354,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31354,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(31354,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(31354,36,'cp 37','e2f3 b8c6 e1g1 g8f6 c2c4 d5c4 b1a3 e7e6 a3c4 a8c8 c4e5 f8d6 e5c6 b7c6 g2g3 f6d5 b2b3 d6e7 d1d2 e8g8 c1b2 e7g5 d2e2 d5b4 a2a3 b4d5 a1d1 a7a5 h2h4 g5h6 d1d3 d8d6 f1e1 c8b8 e2c2 f8c8 b2c1 h6c1 e1c1');
INSERT INTO evaluations VALUES(32352,30,'cp -57',NULL);
INSERT INTO evaluations VALUES(32352,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(32352,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(32352,33,'cp -64',NULL);
INSERT INTO evaluations VALUES(32352,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(32352,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(32352,36,'cp -50','c7c5 f1b5 b8d7 g1e2 a7a6 b5d7 d8d7 e1g1 f5g6 c1e3 a8c8 e2f4 c5d4 e3d4 g8e7 f4g6 h7g6 c3e2 e7f5 c2c3 f8e7 d1d3 d7c7 h2h3 e8g8 f1e1 e7c5 d3d2 c5d4 e2d4 f5d4 c3d4 c7c2 e1e2 c2d2 e2d2 c8c4 g2g3 f8c8 h3h4 a6a5 a2a3');
INSERT INTO evaluations VALUES(36367,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(36367,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(36367,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(36367,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(36367,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(36367,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(36367,36,'cp 27','f1c4 c7c5 d4c5 d8d1 a1d1 e7c5 e1g1 c5b4 c3b5 b8c6 g5f4 c8d7 a2a3 b4e7 c4e2 f8d8 b5d6 e7d6 f4d6 d7e8 d6f4 h7h6 d1d8 a8d8 f1d1 d8c8 d1c1 c8d8 b2b4 a7a6 f4g3 f6e4 g3c7 d8c8 c7a5 c6e7 c1c8 e7c8 f3d4 e6e5 e2d3');
INSERT INTO evaluations VALUES(36403,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(36403,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(36403,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(36403,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(36403,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(36403,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(36403,36,'cp -11','c7c6 c1e3 a7a5 a2a3 f6g4 e3f4 c8f5 e2d3 f5d3 d1d3 b8d7 c3e2 f8e8 e2g3 d8b6 c2c3 a5a4 d3c2 g4f6 h2h4 e7f8 a1e1 e8e1 f1e1 a8e8 e1e8 f6e8 g3f5 d7f6 g2g3 b6b5 f5e3 e8d6 f3e5 g7g6');
INSERT INTO evaluations VALUES(35778,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(35778,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(35778,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35778,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35778,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35778,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35778,36,'cp 0','c1e3 a7a5 a2a3 c8g4 h2h3 g4f3 e2f3 b8d7 a3a4 f8e8 d1d3 e7d6 a1e1 d7f8 c3e2 f8e6 c2c4 d5c4 d3c4 d8d7 e2c3 e6c7 c4b3 d6b4 e1d1 f6d5 e3d2 a8d8 d2g5 d8a8');
INSERT INTO evaluations VALUES(37340,30,'cp 79',NULL);
INSERT INTO evaluations VALUES(37340,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(37340,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(37340,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(37340,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(37340,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(37340,36,'cp 80','b2b4 d7d6 c1b2 b8d7 c4c5 e7e6 f1e2 f6e4 c5d6 e4c3 b2c3 c7d6 d1b3 d7f6 e1g1 f6d5 a2a4 d5c3 b3c3 g6g5 a1c1 g5g4 f3e1 e6e5 d4e5 d6e5 c3b3 g8h8 e2c4 f5f4 e1d3 c8f5 e3f4 f5d3 c4d3 e5f4 g2g3 a8c8 c1d1 d8e7 f1e1');
INSERT INTO evaluations VALUES(39906,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(39906,31,'cp -8',NULL);
INSERT INTO evaluations VALUES(39906,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(39906,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(39906,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(39906,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39906,36,'cp -6','a2a3 e8g8 c2c4 d7d5 b2b4 c5d6 c4d5 e6d5 f1d3 a7a5 b4a5 d8a5 c1d2 a5d8 e1g1 b8c6 a3a4 d8e7 h2h3 c6e5 d2c3 e5d3 d1d3 e7e4 f1d1 e4d3 d1d3 c8f5 d3d1 f5c2 d1d2 c2a4 c3f6 g7f6 b1c3 a4d7 a1c1 d7e6 c3b5 d6e7 f3d4');
INSERT INTO evaluations VALUES(40734,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(40734,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40734,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(40734,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40734,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(40734,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(40734,36,'cp 16','c7c6 b5d3 e8g8 e1g1 f8e8 h2h3 e7f8 f1e1 b8d7 a2a4 h7h6 d4b3 d7e5 c1f4 e5d3 d1d3 c8e6 b3d4 a7a6 a1d1 d8a5 d3f3 a5h5 f3h5 f6h5 f4h2 h5f6 h2d6 f8d6 e4e5 d6e5 e1e5 a8d8 d4e6 d8d1 c3d1 e8e6 e5e6');
INSERT INTO evaluations VALUES(40123,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(40123,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(40123,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(40123,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(40123,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(40123,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(40123,36,'cp 35','b5e2 e8g8 e1g1 f8e8 f2f4 b8a6 e2f3 a6c5 a2a4 e7f8 f1e1 a7a5 h2h3 g7g6 b2b3 f8g7 c1b2 c5e6 d4e6 c8e6 c3e2 d6d5 e4e5 f6d7 d1d2 f7f6 e5f6 d7f6 b2e5 e6f5 e2d4 f5e4 c2c3 e4f3 d4f3 d8d7 g1h2 f6e4 d2e3 d7f7');
INSERT INTO evaluations VALUES(42376,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(42376,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(42376,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(42376,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(42376,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(42376,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42376,36,'cp 8','d6c5 f1d3 c7c6 d1f3 d7d6 h2h3 e8g8 g1e2 b8d7 e1g1 g8h8 a2a4 h7h6 e2g3 a7a5 g1h1 c5b4 c3e2 d7c5 d3c4 c6d5 e4d5 c5a6 c4d3 a6c5 d3f5 f8e8 f1d1 e8f8 c2c3 c5b3 c3b4 b3a1 c1h6 g7h6 d1a1 a5b4');
INSERT INTO evaluations VALUES(43040,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(43040,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(43040,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(43040,33,'cp 153',NULL);
INSERT INTO evaluations VALUES(43040,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(43040,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(43040,36,'cp 143','f3e5 e8g8 d2d4 e7b4 c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g2f3 b8c6 e5c6 b7c6 h2h4 b4e7 e1c1 a8b8 b2b3 d7d6 f1c4 f8d8 d4d5 c8d7 d1d3 g8f8 f3f4 d8e8 d5c6 d7c6 c3d5 e7d8 f2f3 f7f5 e4f5 c6d5 d3d5 d8f6 h4h5 b8d8 c1d2 c7c6 d5a5 f6d4 a5a6 d4e3 d2c3 d6d5 c4d3');
INSERT INTO evaluations VALUES(43010,30,'cp 138',NULL);
INSERT INTO evaluations VALUES(43010,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(43010,32,'cp 133',NULL);
INSERT INTO evaluations VALUES(43010,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(43010,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(43010,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(43010,36,'cp 143','e8g8 d2d4 e7b4 c1g5 h7h6 g5f6 d8f6 d1f3 f6f3 g2f3 b8c6 e5c6 b7c6 h2h4 b4e7 e1c1 f8d8 f1e2 a7a5 b2b3 g7g6 c1b2 d7d5 h4h5 g6g5 h1g1 g8h8 c3a4 c8e6 e2d3 d8e8 c2c3 e8f8 d1e1 a8d8 b2c2 d8b8 e4e5');
INSERT INTO evaluations VALUES(44689,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(44689,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44689,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44689,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(44689,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(44689,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(44689,36,'cp 21','d2d4 b8c6 f1d3 c5d4 c3d4 e7f5 d3f5 e6f5 h2h4 f8e7 b1c3 e8g8 c1g5 f7f6 e5f6 e7f6 g5f6 d8f6 c3d5 f6e6 d5e3 f5f4 d4d5 e6g6 h4h5 g6f7 d5c6 f4e3 f2e3 b7c6 e1g1 f7h5 d1d6 c8g4 f3h2 g4f5 d6c6 a8c8 c6d5 g8h8 f1f4 h5g6');
INSERT INTO evaluations VALUES(44795,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44795,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44795,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44795,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44795,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44795,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(44795,36,'cp 18','b8c6 d4c6 b7c6 b2b3 f8g7 c1b2 e8g8 b1c3 e7e5 d1d2 d6d5 f1e1 c8e6 a1d1 d8b8 c3a4 d5e4 d3e4 f6e4 e1e4 f8d8 d2c1 d8d1 c1d1 f7f6 e4e1 b8c7 d1e2 e6f7 h2h3 g7f8 a4c3 a8d8 c3e4 f8e7 c2c4 c7b6');
INSERT INTO evaluations VALUES(46160,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(46160,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(46160,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(46160,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(46160,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(46160,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(46160,36,'cp 13','d2d4 b8c6 c4d5 f6d5 f1d3 f8e7 e1g1 e8g8 f1e1 b7b6 c3d5 e6d5 d3b5 d8d6 b2b3 a7a6 b5f1 c5d4 f3d4 c6d4 e3d4 c8e6 a2a4 f8c8 c1a3 d6d7 a3e7 d7e7 a1c1 b6b5 a4b5 a6b5 c1c8 a8c8 f1b5 e7b4 b5e2 c8b8 e2d3 b4b3');
INSERT INTO evaluations VALUES(49171,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(49171,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(49171,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(49171,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(49171,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(49171,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(49171,36,'cp -31','c7c5 f2f4 c5d4 c3b5 b8c6 g1f3 f8c5 b2b4 c5e7 b5d6 e7d6 e5d6 e8g8 b4b5 c6a5 d1d4 a7a6 a3a4 b7b6 c1a3 a6b5 f1b5 c8a6 e1g1 a6b5 a4b5 a5c4 a3b4 a8a1 f1a1 d8b8 f3e5 c4e5 f4e5 f8c8 b4c3 c8c5 d4d3 b8c8 c3d4 c5c2 a1a7 c2c4');
INSERT INTO evaluations VALUES(44012,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(44012,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(44012,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(44012,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(44012,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(44012,35,'cp -27',NULL);
INSERT INTO evaluations VALUES(44012,36,'cp -41','f2f4 b8c6 g1f3 c5d4 c3b5 f8c5 b2b4 c5e7 b5d6 e7d6 e5d6 d7b6 b4b5 c6a5 d1d4 f7f6 c1d2 b6c4 d2b4 c4d6 a3a4 a5c4 f1d3 a7a5 b5a6 a8a6 a4a5 a6c6 e1g1 e8g8 f4f5 f8e8 f5e6 c8e6 a1e1 d8d7 d4h4 d6e4 e1e4 d5e4 d3e4 h7h6 e4c6 d7c6 f3d4 c6d7');
INSERT INTO evaluations VALUES(50132,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(50132,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(50132,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(50132,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(50132,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(50132,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(50132,36,'cp 29','f8g7 c2c4 e8g8 f1e2 d7d6 b1c3 f6e4 c3e4 f5e4 f3d2 g7d4 e3d4 f8f4 g2g3 f4f7 d2e4 e7e5 d4e5 d8e7 e4f6 g8g7 f6d5 e7e5 d1d2 c8g4 f2f4 e5e2 d2e2 g4e2 e1e2 b8c6 a1e1 a7a5 e2d2 c6b4 d5c3 c7c6 e1e6 d6d5 c4d5 b4d5 c3d5 c6d5');
INSERT INTO evaluations VALUES(51283,30,'cp 113',NULL);
INSERT INTO evaluations VALUES(51283,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(51283,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(51283,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(51283,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(51283,35,'cp 118',NULL);
INSERT INTO evaluations VALUES(51283,36,'cp 118','d2d4 c8b7 b1c3 a7a6 f1c4 e7e6 d4d5 b6b5 c4b3 e6e5 a2a4 b5b4 c3a2 g8f6 a4a5 c7c5 d5c6 b8c6 b3a4 f6d7 c2c3 b4c3 a2c3 f8e7 e1g1 e8g8 c1e3 d7f6 e3b6 d8b8 c3d5 f6d5 e4d5 c6a7 f1e1 a7b5 a1c1 e7f6 a4b5 a6b5 d1b3 f8c8 c1c8 b8c8 b3b5 b7a6');
INSERT INTO evaluations VALUES(52675,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(52675,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(52675,32,'cp 107',NULL);
INSERT INTO evaluations VALUES(52675,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(52675,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(52675,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(52675,36,'cp 103','e7e5 g1f3 b8d7 f1e2 g8f6 e1g1 f8e7 c1e3 e8g8 d1c2 f6g4 e3d2 g4f6 f1d1 a7a5 d2e3 f6g4 e3c1 d8c7 h2h3 g4f6 c1e3 f8e8 d1e1 d7f8 d4d5 f8g6 e2f1 c8d7 a2a4 e8f8 a1d1 g8h8 g1h2 f6h7 g2g3 a8e8 b2b3 f7f5 e4f5');
INSERT INTO evaluations VALUES(56569,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(56569,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(56569,32,'cp 55',NULL);
INSERT INTO evaluations VALUES(56569,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(56569,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(56569,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(56569,36,'cp 57','e3d4 e5d4 g1f3 c7c5 f1b5 b8c6 d1e2 g8e7 e1g1 c8g4 h2h3 g4f3 e2f3 a7a6 b5d3 c6e5 f3b7 e5d3 c2d3 d8d5 b7c7 a8c8 c7g3 e7g6 b1a3 f8d6 a1e1 e8d7 f2f4 h8e8 a3c4 d6c7 b2a3 h7h5 b3b4 e8e1 f1e1 g6f4 b4c5 d5c4 e1e7 d7c6 d3c4 f4e2 e7e2 c7g3 e2e7 f7f6 e7g7 h5h4');
INSERT INTO evaluations VALUES(56558,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(56558,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(56558,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(56558,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(56558,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(56558,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(56558,36,'cp 42','e5d4 g1f3 c7c5 f1b5 b8c6 d1e2 g8e7 e1g1 d8d5 b5c4 d5f5 b1a3 f5f4 h2h3 c8d7 c2c3 e8c8 a1c1 c8b8 c3d4 c6d4 e2e5 f4e5 f3e5 d7e8 b2d4 d8d4 c4f7 e7c6 f2f4 c6e5 f4e5 e8c6 f1f2 f8e7 a3c2 d4d8 c1d1 a7a6 d2d4 c5d4 c2d4 e7c5 d4c6 b7c6 d1d8 h8d8');
INSERT INTO evaluations VALUES(57560,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(57560,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(57560,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(57560,33,'cp 152',NULL);
INSERT INTO evaluations VALUES(57560,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(57560,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(57560,36,'cp 144','e7e6 d2d4 f8b4 c1d2 d7d5 d1g4 d8e7 e1c1 b8c6 e4d5 e6e5 g4h5 g7g6 h5h4 c6d4 f2f4 d4f5 h4e1 e5f4 d2f4 e7e1 d1e1 e8f7 f4c7 g8e7 g2g4 f5d6 g4g5 h8e8 g5f6 e7f5 e1e8 d6e8 c7e5 b4c3 b2c3 e8f6 g1f3 f6d5 f1c4 c8e6 f3g5 f7e7 h1e1 f5e3 c4b3');
INSERT INTO evaluations VALUES(1459,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(1459,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(1459,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(1459,33,'cp -46',NULL);
INSERT INTO evaluations VALUES(1459,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(1459,35,'cp -31',NULL);
INSERT INTO evaluations VALUES(1459,36,'cp -26','f1d3 c8d7 e1g1 c5d4 a2a3 g8h6 h2h3 h6f5 d3f5 e6f5 g1h2 a7a5 d1d3 b6a6 f1d1 a6d3 d1d3 f8c5 c2c3 d4c3 b1c3 d5d4 c3e2 a8d8 b2b3 d7e6 c1b2 e8g8 b2d4 d8d7 b3b4 a5b4 d4c5 d7d3 c5f8 g8f8 a3b4 e6c4 a1a8 f8e7 a8g8');
INSERT INTO evaluations VALUES(2228,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(2228,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(2228,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(2228,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(2228,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(2228,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(2228,36,'cp 15','d4c6 f6c6 e3c5 c6c5 c3c4 d7d6 b1c3 c8e6 b2b3 a8d8 e1g1 c7c6 a1c1 h7h6 d1d2 c5b4 d2e1 a7a5 e2f3 d6d5 c4d5 c6d5 e4d5 e7d5 c3d5 e6d5 e1b4 a5b4 f3d5 d8d5 f1d1 d5d1 c1d1 f8a8 g1f1 a8a2 d1d8 g8h7 d8d4 g7g5 d4b4 h7g7 b4b7 g7g6 b7b5 a2b2 b5b8 h6h5');
INSERT INTO evaluations VALUES(3106,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(3106,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(3106,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(3106,33,'cp 82',NULL);
INSERT INTO evaluations VALUES(3106,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(3106,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(3106,36,'cp 83','f3e5 f6d7 e5d7 c8d7 d1e2 a8c8 h2h3 e7g5 f4g3 d8f6 e1g1 f6h6 d2b3 c6e7 a2a4 e7g6 b3c5 d7c6 d3b5 g6f4 e2g4 h6h5 g4h5 f4h5 g3h2 g5e7 f1b1 h5f6 b5c6 b7c6 b2b4 e7c5 b4c5 f6e4 h2d6 f8d8 b1b7 e4d6 c5d6 d8d6');
INSERT INTO evaluations VALUES(3107,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(3107,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(3107,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(3107,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(3107,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(3107,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(3107,36,'cp 73','f6d7 e5d7 c8d7 d1e2 a8c8 d2f3 f7f6 e1g1 g7g5 f4g3 f8f7 a2a4 d8f8 f3e1 f6f5 f2f4 g5g4 d3b5 a7a6 b5c6 d7c6 e1d3 e7d6 g3h4 f8h6 g2g3 c6d7 f1e1 c8c4 e2e3 c4a4 a1a4 d7a4 e3e6 h6e6 e1e6 d6c7 d3c5 a4d7');
INSERT INTO evaluations VALUES(5316,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(5316,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(5316,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(5316,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(5316,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(5316,35,'cp -26',NULL);
INSERT INTO evaluations VALUES(5316,36,'cp -26','d7d6 d2d3 e7e6 c1e3 c6d4 d1d2 g8e7 e1g1 e8g8 e3h6 e7c6 h6g7 g8g7 f3d4 c6d4 f2f4 c8d7 a1e1 d7c6 e2d1 a7a6 d2f2 a6a5 h2h4 h7h5 c3e2 d4e2 e1e2 d8f6 f4f5 e6f5 e4f5 f8e8 e2e8 a8e8 f5g6');
INSERT INTO evaluations VALUES(5480,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(5480,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(5480,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(5480,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(5480,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(5480,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(5480,36,'cp -14','d2d3 e7e6 c1e3 c6d4 d1d2 g8e7 e1g1 e8g8 e3h6 e7c6 h6g7 g8g7 f3d4 c6d4 f2f4 a8b8 a1e1 c8d7 e2d1 b7b5 b2b3 d7c6 c3e2 d8f6 e2g3 h7h6 f1f2 b5b4 d2e3 a7a5 g3e2 d4e2 d1e2 e6e5 e1f1 a5a4 f4f5 f6g5 e3h3 a4b3 f5f6 g7h7 a2b3');
INSERT INTO evaluations VALUES(6368,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(6368,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(6368,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(6368,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(6368,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(6368,35,'cp 96',NULL);
INSERT INTO evaluations VALUES(6368,36,'cp 100','c3e2 e6e5 c2c3 b4a5 a2a4 a5b6 d2d4 e5d4 e2d4 g8e7 c1g5 a7a5 f1b5 e8g8 e1g1 f7f6 g5e3 g8h8 b5c4 d7d6 d4c6 b7c6 e3b6 c7b6 f3d4 c8d7 f2f4 d8c7 f4f5 d6d5 e4d5 e7d5 c4d5 c6d5 d4e6 d7e6 f5e6 c7c5 d1d4 f8e8 a1e1 a8d8 f1f4 c5d6 d4e3 h7h6');
INSERT INTO evaluations VALUES(6652,30,'cp -840',NULL);
INSERT INTO evaluations VALUES(6652,31,'cp -840',NULL);
INSERT INTO evaluations VALUES(6652,32,'cp -828',NULL);
INSERT INTO evaluations VALUES(6652,33,'cp -832',NULL);
INSERT INTO evaluations VALUES(6652,34,'cp -836',NULL);
INSERT INTO evaluations VALUES(6652,35,'cp -832',NULL);
INSERT INTO evaluations VALUES(6652,36,'cp -832','c6b4 f1d3 d7d5 b1c3 g8e7 a2a3 b4d3 c2d3 d5d4 c3e2 a7a5 g1f3 c7c5 c1f4 b7b6 f4g3 e8g8 e1g1 a5a4 g3h4 f7f6 e4e5 e7d5 e5f6 g7f6 h4g3 c8a6 a1d1 a8c8 d1d2 d5e7 e2f4 d8d7 f3h4 b6b5 h4f3 c8c6 f1d1 b5b4 a3b4 c5b4');
INSERT INTO evaluations VALUES(6901,30,'cp -827',NULL);
INSERT INTO evaluations VALUES(6901,31,'cp -816',NULL);
INSERT INTO evaluations VALUES(6901,32,'cp -805',NULL);
INSERT INTO evaluations VALUES(6901,33,'cp -812',NULL);
INSERT INTO evaluations VALUES(6901,34,'cp -817',NULL);
INSERT INTO evaluations VALUES(6901,35,'cp -812',NULL);
INSERT INTO evaluations VALUES(6901,36,'cp -809','f1d3 b4d3 c2d3 b7b6 b1c3 c8a6 d3d4 f7f5 e4e5 d8h4 g1e2 g8e7 e1g1 e8c8 c1e3 e7c6 f2f4 d7d5 f1f3 h4e7 a2a3 c8b8 b2b4 a6c8 a1c1 e7d7 e3f2 h7h6 f3h3 d8e8 c1e1 d7f7 g1h1 b8a8 e1c1 c8b7 e2g3 h6h5');
INSERT INTO evaluations VALUES(7091,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(7091,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(7091,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(7091,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(7091,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(7091,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(7091,36,'cp 72','h7h5 c1b1 c8d7 h2h3 h5h4 f3f4 b7b5 f1d3 a8c8 d4f3 c6b4 e4e5 d6e5 f4e5 f6d5 c3d5 b4d5 h1f1 d7c6 e3g5 e8g8 g5e7 d8e7 f3g5 b5b4 g5e4 c6b5 e4f6 d5f6 e5f6 g7f6 d2h6 b5d3 c2d3 e7c5 h6f6 c5h5 d1e1 h5g6 f6g6 f7g6 f1g1 f8f4 e1e6 g8f7 e6b6 c8c7');
INSERT INTO evaluations VALUES(7984,30,'cp 141',NULL);
INSERT INTO evaluations VALUES(7984,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(7984,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(7984,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(7984,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(7984,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(7984,36,'cp 150','d1d2 d7d5 d4c6 b7c6 e1c1 d8a5 c4b3 h7h5 f2f3 a8b8 d2f2 e8g8 e3a7 b8b7 a7c5 g7c3 b2c3 a5c3 c5d4 c3b4 g2g4 e6e5 d4e5 c6c5 g4h5 c5c4 f2h4 f7f6 e5f6 b4d6 f6b2 c4b3 a2b3 d6f4 h4f4 f8f4 e4d5 g6h5 h1g1 g8f7 g1g7 f7e8 d5d6 e7c6 d1e1 e8f8 g7b7 c8b7 d6d7 f8f7');
INSERT INTO evaluations VALUES(9085,30,'cp 260',NULL);
INSERT INTO evaluations VALUES(9085,31,'cp 256',NULL);
INSERT INTO evaluations VALUES(9085,32,'cp 255',NULL);
INSERT INTO evaluations VALUES(9085,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(9085,34,'cp 256',NULL);
INSERT INTO evaluations VALUES(9085,35,'cp 260',NULL);
INSERT INTO evaluations VALUES(9085,36,'cp 237','e4d5 f6d5 d1b3 c8e6 b3b7 d5e7 c4e6 f7e6 f3g5 a8b8 b7a6 d8d5 d2d3 e7f5 g5e4 e8g8 b2b4 c5e7 a2a4 e7b4 c3b4 c6b4 a6c4 b4d3 b1d2 f5d6 e4d6 c7d6 c1a3 b8b6 c4d5 e6d5 a4a5 b6c6 a1b1 h7h6 b1b5 d3f4 g2g3 f4e2 g1g2 e2c3 b5b7 d5d4 a3b4 f8f7 b7b8 f7f8 b8f8 g8f8 f2f4 e5e4');
INSERT INTO evaluations VALUES(9072,30,'cp 256',NULL);
INSERT INTO evaluations VALUES(9072,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(9072,32,'cp 246',NULL);
INSERT INTO evaluations VALUES(9072,33,'cp 237',NULL);
INSERT INTO evaluations VALUES(9072,34,'cp 231',NULL);
INSERT INTO evaluations VALUES(9072,35,'cp 236',NULL);
INSERT INTO evaluations VALUES(9072,36,'cp 236','f6d5 d1b3 c8e6 b3b7 d5e7 c4e6 f7e6 f3g5 a8b8 b7a6 d8d5 d2d3 e7f5 g5e4 e8g8 b2b4 c5e7 a2a4 e7b4 c3b4 c6b4 a6c4 b4d3 b1d2 f5d6 e4d6 c7d6 c1a3 b8b6 a1d1 h7h6 c4d5 e6d5 d2f3 d3f4 f1e1 f8f6 a3c1 d5d4 c1f4 f6f4 f3d2 f4f8 f2f3 b6c6 a4a5 f8d8 d2e4 d8c8 d1b1 d6d5 e4f2 g8f7 f2d3');
INSERT INTO evaluations VALUES(9770,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(9770,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9770,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9770,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9770,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9770,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9770,36,'cp 0','c5e3 f2e3 d7d6 e1g1 e8g8 c4b3 c6e7 f3h4 a7a5 a2a4 c8e6 b1d2 e6b3 d1b3 d6d5 e4d5 d8d5 b3d5 e7d5 e3e4 d5f4 d3d4 f4e2 g1f2 e2f4');
INSERT INTO evaluations VALUES(9853,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9853,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9853,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9853,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9853,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9853,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9853,36,'cp 0','f2e3 d7d6 c4b3 e8g8 e1g1 c6e7 f3h4 a7a5 a2a4 c8g4 d1c2 g4e6 b1d2 e6b3 c2b3 d6d5 e4d5 d8d5 b3d5 e7d5 e3e4 d5f4 d3d4 f4e2 g1f2 e2f4');
INSERT INTO evaluations VALUES(9852,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9852,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9852,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9852,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9852,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(9852,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9852,36,'cp 0','d7d6 e1g1 e8g8 c4b3 c6e7 f3h4 a7a5 a2a4 c8g4 d1c2 g4e6 b1d2 e6b3 c2b3 d6d5 e4d5 d8d5 b3d5 e7d5 e3e4 d5f4 d3d4 f4e2 g1f2 e2f4');
INSERT INTO evaluations VALUES(9970,30,'cp -138',NULL);
INSERT INTO evaluations VALUES(9970,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(9970,32,'cp -147',NULL);
INSERT INTO evaluations VALUES(9970,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(9970,34,'cp -142',NULL);
INSERT INTO evaluations VALUES(9970,35,'cp -156',NULL);
INSERT INTO evaluations VALUES(9970,36,'cp -134','b4c3 d2c3 f6e4 c3d2 b7b6 d2c1 c8b7 c2c3 c6e7 f1d3 e8g8 e1g1 f7f5 d3c4 b7d5 c4d5 e7d5 c3c4 d5e7 f1e1 d7d5 c4d5 e7d5 f3e5 h7h6 f2f3 e4d6 g1h1 d6f7 e5g6 f8e8 g6f4 d5f4 c1f4 a7a6');
INSERT INTO evaluations VALUES(9993,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(9993,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(9993,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(9993,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(9993,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(9993,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(9993,36,'cp -5','f6d5 e4d5 e5e4 d5c6 d7c6 b5c4 e4f3 d1f3 d8e7 f3e3 e7e3 f2e3 f7f6 e3e4 c8d7 a2a3 b4d6 d2d3 a7a5 a3a4 h7h5 h2h3 d6e5 c2c3 c6c5 h1f1 b7b6 c1f4 e5f4 f1f4 h8h6 h3h4 e8c8 e1f2 d8e8 f2f3 e8e5 f3f2 e5e7');
INSERT INTO evaluations VALUES(10568,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(10568,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(10568,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(10568,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(10568,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(10568,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10568,36,'cp -4','e4d5 e5e4 d5c6 d7c6 b5c4 e4f3 d1f3 d8e7 f3e3 a7a5 e1g1 e8g8 e3e7 b4e7 a2a4 f8e8 f1e1 c8d7 g1f1 b7b6 d2d4 e7d6 c1e3 h7h6 c4d3 g8f8 h2h4 a8d8 c2c3 e8e7 f2f3 d8e8 e3f2 e7e1 a1e1 e8e1 f1e1 c6c5 d4c5 d6c5 f2c5 b6c5 d3b5 d7e6 b5c6 f8e7');
INSERT INTO evaluations VALUES(10567,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(10567,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10567,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(10567,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(10567,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(10567,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(10567,36,'cp -11','e5e4 d5c6 d7c6 b5c4 e4f3 d1f3 d8e7 f3e3 e7e3 f2e3 f7f6 a2a3 b4d6 d2d3 c8d7 e3e4 e8c8 c1e3 h8f8 e1g1 a7a5 a3a4 d6e5 c2c3 b7b6 d3d4 e5d6 e3f4 d6f4 f1f4 c6c5 g1f2 c5d4 c3d4 c7c5 d4c5 b6c5 b2b3 c8c7 a1f1 f8e8 f2e3');
INSERT INTO evaluations VALUES(10305,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(10305,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(10305,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(10305,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(10305,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(10305,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(10305,36,'cp -46','d4b3 c5b6 b1c3 a7a5 f1b5 c6e5 f3f4 e5g4 b3d4 c7c6 e4e5 c6b5 e5f6 d8f6 c3d5 f6d4 d1e2 g4e5 d5b6 d4b6 e2e5 b6e6 e1g1 e6e5 f4e5 d7d6 e5d6 c8e6 a2a3 e8d7 c1f4 a8c8 c2c3 h8e8 f1e1 e6d5 g1f2 f7f6 a1d1 d5c4 h2h4 h7h5 f4e3 c8c6 e3d4');
INSERT INTO evaluations VALUES(10903,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(10903,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(10903,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(10903,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(10903,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(10903,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(10903,36,'cp -20','d7d5 e4d5 f6d5 e1g1 c8e6 c4b5 f7f6 b1c3 d5c3 b2c3 e8g8 b5c6 b7c6 f3d2 a7a5 c3c4 a5a4 a1b1 f6f5 f2f4 e5e4 d3e4 f5e4 d1e2 d8d7 d2e4 e6g4 e2f2 d7e6 e4g3 e6c4 f4f5 e7h4 f2f4 c4f4 f1f4 h4g3 h2g3 g4f5 b1b7 f5c2');
INSERT INTO evaluations VALUES(13085,30,'cp 252',NULL);
INSERT INTO evaluations VALUES(13085,31,'cp 255',NULL);
INSERT INTO evaluations VALUES(13085,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(13085,33,'cp 241',NULL);
INSERT INTO evaluations VALUES(13085,34,'cp 253',NULL);
INSERT INTO evaluations VALUES(13085,35,'cp 262',NULL);
INSERT INTO evaluations VALUES(13085,36,'cp 269','c4d5 f6d5 e3e4 d5f6 d4d5 c6d4 f3d4 c5d4 d1d4 d8b6 d4b6 a7b6 b1c3 e7e6 d5e6 f7e6 f2f3 e8f7 c1d2 c8d7 e1c1 d7c6 c1b1 f8b4 a2a3 b4e7 c3b5 h8d8 f1e2 f6d7 d2e3 d7c5 e3f4 c5b3 f4c7 d8d2 d1d2 b3d2 b1c2 c6b5 c2d2 b5a6 e2a6');
INSERT INTO evaluations VALUES(13548,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(13548,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(13548,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(13548,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(13548,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(13548,35,'cp -52',NULL);
INSERT INTO evaluations VALUES(13548,36,'cp -46','c5d4 f3d4 f8b4 c1b2 f6e4 d4b5 a7a6 a2a3 d8f6 d1f3 a6b5 f3f6 e4f6 a3b4 a8a1 b2a1 b5c4 b3c4 c6b4 a1b2 e8g8 c4d5 f8d8 f2f3 f6d5 e1f2 h7h6 f1e2 b4d3 e2d3 d5c3 d3h7 g8h7 b2c3 d8d3 h1c1 c8d7 h2h4 d7c6 h4h5 d3d5 g2g4 d5c5 c3d2 c5b5 e3e4 b5b2 f2e3 e6e5 g4g5 b2b3 e3f2 h6g5 d2g5');
INSERT INTO evaluations VALUES(13996,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(13996,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(13996,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(13996,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(13996,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(13996,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(13996,36,'cp 26','d6d5 e4d5 f6d5 c3d5 e6d5 d1d2 f8d6 e1c1 e8g8 c1b1 a7a6 g2g4 f8e8 a2a3 c8d7 d4b3 a8c8 e3c5 d6c5 b3c5 c6e5 c5d7 d8d7 d2d5 d7d5 d1d5 e5f3 d5d1 f3e5 f1g2 c8c7 h1e1 g8f8 h2h3 e5c4 e1e8 f8e8 c2c3 b7b5 d1e1 e8f8 g2d5 c4b6 d5b3 c7e7');
INSERT INTO evaluations VALUES(15122,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(15122,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(15122,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(15122,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(15122,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(15122,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(15122,36,'cp 62','f1b5 e5d4 e4e5 f6e4 f3d4 c8d7 b5c6 d7c6 e1g1 c6d7 c1e3 f8e7 d4b3 d7a4 f2f3 e4c5 f3f4 d8d7 e3c5 a4b3 d1d4 e7c5 d4c5 b3c2 b1d2 b7b6 c5f2 e8g8 f1e1 c7c5 d2c4 c2d3 c4d6 f7f6 a1d1 f6e5 e1e5 d7d6 d1d3 d6h6 d3d5 h6f4 f2f4 f8f4');
INSERT INTO evaluations VALUES(15195,30,'cp -129',NULL);
INSERT INTO evaluations VALUES(15195,31,'cp -150',NULL);
INSERT INTO evaluations VALUES(15195,32,'cp -128',NULL);
INSERT INTO evaluations VALUES(15195,33,'cp -131',NULL);
INSERT INTO evaluations VALUES(15195,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(15195,35,'cp -140',NULL);
INSERT INTO evaluations VALUES(15195,36,'cp -152','h7h5 h2h4 c8g4 f2f3 g4e6 d2d4 f8d6 d4e5 c6e5 e2f4 d8d7 d1e2 e8c8 b1c3 c8b8 f4e6 d7e6 c3b5 h8g8 b5d6 d8d6 e2e1 a7a6 a1b1 d6c6 b2b4 c6c2 e1d1 e6c6 b1b2 c2c3 b2b3 c6c4 b3c3 c4c3 c1d2 c3d3 g1h1 g8e8');
INSERT INTO evaluations VALUES(18121,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(18121,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(18121,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(18121,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(18121,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(18121,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(18121,36,'cp -12','c5d4 f3d4 c8b7 d4c6 b7c6 a2a3 f8c5 e1g1 c5b6 d1d3 g8e7 c1f4 a8c8 f4d6 b6c7 d6c7 d8c7 a1d1 e8g8 d3d6 c7d6 d1d6 c8c7 d6d2 f8c8 f2f4 d7d5 f1d1 g8f8 e4e5 f7f6 e5f6 g7f6 g2g3 e7f5 g1f2');
INSERT INTO evaluations VALUES(18106,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(18106,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(18106,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(18106,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(18106,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(18106,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(18106,36,'cp -21','f3d4 c8b7 d4c6 b7c6 a2a3 f8c5 e1g1 d8c7 d1d3 g8e7 c1e3 c5e3 d3e3 e8g8 e2d3 f7f5 f1e1 f5e4 d3e4 c6e4 e3e4 a8c8 e4d3 c7c6 c3e2 c6c5 d3d4 c5d4 e2d4 e7d5 c2c3 g8f7 a1d1 d5b6 f2f3 f7e7');
INSERT INTO evaluations VALUES(18104,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(18104,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(18104,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(18104,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(18104,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(18104,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(18104,36,'cp -15','c8b7 d4c6 b7c6 a2a3 f8c5 e1g1 d8c7 d1d3 g8e7 c1e3 c5e3 d3e3 e8g8 e3g3 c7b6 g3d6 e7c8 d6d2 b6c7 a1d1 c8b6 d2d6 f8c8 f2f4 h7h6 d6c7 c8c7 e4e5 b6c4 e2c4 b5c4 d1d4 a8b8 c3d1 a6a5 f4f5 e6f5 f1f5');
INSERT INTO evaluations VALUES(25591,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(25591,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(25591,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(25591,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(25591,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(25591,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(25591,36,'cp -12','d4c6 b7c6 a2a3 f8c5 e1g1 d8c7 d1d3 g8e7 c1e3 c5e3 d3e3 e8g8 e3g3 c7b7 g3d6 e7g6 e2d3 h7h6 f2f3 f8e8 c3e2 b7b6 g1h1 b6b8 d6b8 e8b8 f3f4 b5b4 b2b3 b4a3 e2d4 c6b5 c2c4 b5c6 e4e5 g6h4 d4c6 d7c6 a1a3');
INSERT INTO evaluations VALUES(18160,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(18160,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(18160,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(18160,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(18160,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(18160,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(18160,36,'cp 88','d1h5 f8d6 f1c4 d8e7 c1g5 g8f6 h5h4 d6b4 a1d1 h7h6 d1d3 h8f8 g5d2 a8b8 e1g1 d7d5 e4d5 c6d5 c4b3 d5d4 c3e4 c8f5 d2b4 b8b4 e4f6 e7f6 h4f6 g7f6 d3f3 f5e6 f3f6 e8e7 f6h6 e6b3 c2b3 f7f6 h2h4 d4d3 f1d1 b4d4 h4h5 e7e6 h6h7 e5e4 g1f1 d4d8 h7a7 f6f5 h5h6');
INSERT INTO evaluations VALUES(18178,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(18178,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(18178,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(18178,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(18178,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(18178,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(18178,36,'cp 20','d7d6 d2d4 c5d4 d1d4 e7e5 d4d3 g8f6 b1c3 f8e7 e1g1 d8c7 c1e3 a8b8 b2b3 e8g8 a1c1 f8d8 h2h3 h7h6 f1d1 c8e6 d3e2 a6a5 c3d5 c6d5 c4d5 c7d7 d5e6 d7e6 e3d2 b8a8 c1c4 d6d5 e4d5 d8d5 d2e3 d5d1 e2d1 f6d5 d1e2 f7f6 e3d2');
INSERT INTO evaluations VALUES(19369,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(19369,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(19369,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(19369,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(19369,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(19369,35,'cp -65',NULL);
INSERT INTO evaluations VALUES(19369,36,'cp -67','c6b4 c2c3 b4c6 d2d4 c5d4 c3d4 e7e6 b1c3 f8b4 c1d2 g8f6 f1d3 d7d6 e1g1 e8g8 a1c1 e6e5 d4d5 c6e7 d1b3 a7a5 d2e3 f6g4 e3d2 f7f5 f3g5 g8h8 h2h3 g4f6 e4f5 e7f5 d3f5 c8f5 g5e6 f5e6 d5e6 d8e7 d2g5 a8e8 g5f6 g7f6 c3d5 e7e6 d5b4');
INSERT INTO evaluations VALUES(19593,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(19593,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(19593,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(19593,33,'cp -55',NULL);
INSERT INTO evaluations VALUES(19593,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(19593,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(19593,36,'cp -42','c2c3 b4c6 d2d4 c5d4 c3d4 e7e6 b1c3 f8b4 c1d2 g8f6 f1d3 d7d6 e1g1 e8g8 d1c1 c8d7 a1b1 a8c8 c1b2 b4a5 c3b5 d6d5 e4e5 f6e4 d3e4 d5e4 f3g5 c6b4 g5e6 f7e6 d2b4 d7b5 b4a5 d8a5 b2b5 a5b5 b1b5 b7b6 f2f3 e4f3 f1f3 f8f3');
INSERT INTO evaluations VALUES(20684,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20684,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(20684,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(20684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20684,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20684,36,'cp 4','g1f3 g8f6 d2d3 f8e7 c4b3 e8g8 e1g1 c6a5 e4e5 f6e8 d1e1 a5b3 a2b3 b7b6 c1e3 g7g6 e3f2 c8b7 f2h4 e7h4 f3h4 f7f6 e1g3 f6f5 e5d6 d8d6 h4f3 b7f3 g3f3 e8c7 a1e1 a8d8 f1f2 a7a6 f3g3');
INSERT INTO evaluations VALUES(21179,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(21179,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(21179,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(21179,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(21179,34,'cp 80',NULL);
INSERT INTO evaluations VALUES(21179,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(21179,36,'cp 82','e7e5 e2e4 e5d4 f3d4 f8e7 f1b5 c8d7 d4c6 b7c6 b5d3 e7f6 e1g1 g8e7 f2f4 e8g8 c1e3 f6c3 b2c3 f7f5 d3c4 g8h8 e4e5 e7d5 c4d5 c6d5 e5d6 c7d6 e3d4 f8f7 f1e1 h8g8 a1b1 d8c7 d1d3 d7c6 h2h3 a8e8 g1h2 e8e1 b1e1 f7e7 e1e7');
INSERT INTO evaluations VALUES(23178,30,'cp 559',NULL);
INSERT INTO evaluations VALUES(23178,31,'cp 567',NULL);
INSERT INTO evaluations VALUES(23178,32,'cp 564',NULL);
INSERT INTO evaluations VALUES(23178,33,'cp 575',NULL);
INSERT INTO evaluations VALUES(23178,34,'cp 576',NULL);
INSERT INTO evaluations VALUES(23178,35,'cp 591',NULL);
INSERT INTO evaluations VALUES(23178,36,'cp 548','f3e5 f8b4 c1d2 b4d2 b1d2 c6e5 d1h5 e8f8 h5e5 d7d6 e5a5 g8e7 a5h5 d6d5 e4d5 e7f5 e1g1 d8g5 h5g5 h6g5 a1e1 f5d6 c4d3 f8f7 f2f3 c8d7 c2c4 a8e8 d2e4 e8e5 e4c5 e5e1 f1e1 d7c8 c5e4 d6e4 f3e4 f7e7 e4e5 h8h4 e1f1 c8d7 c4c5 h4f4 f1f4 g5f4 g1f2 c7c6 d5d6 e7e6');
INSERT INTO evaluations VALUES(23854,30,'cp 430',NULL);
INSERT INTO evaluations VALUES(23854,31,'cp 411',NULL);
INSERT INTO evaluations VALUES(23854,32,'cp 424',NULL);
INSERT INTO evaluations VALUES(23854,33,'cp 423',NULL);
INSERT INTO evaluations VALUES(23854,34,'cp 447',NULL);
INSERT INTO evaluations VALUES(23854,35,'cp 432',NULL);
INSERT INTO evaluations VALUES(23854,36,'cp 451','d2d4 c5b6 d4d5 g8e7 h2h3 g4c8 d5c6 b7c6 b5d3 e8g8 b1d2 f7f5 e4f5 e7f5 d2c4 d8e8 d1e1 c8a6 c4e5 a6d3 e5d3 f5g3 e1e8 a8e8 f1e1 e8e1 d3e1 g3e4 e1d3 d6d5 a2a4 a7a5 g1f1 f8e8 c1e3 b6e3 f2e3 c6c5');
INSERT INTO evaluations VALUES(24232,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(24232,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(24232,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(24232,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(24232,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(24232,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(24232,36,'cp 41','f8c5 c2c3 c5b6 d2d4 d7d6 c1e3 e8g8 b1d2 e5d4 c3d4 c6b4 d1b1 c7c5 a2a3 b4c6 d4d5 c6e7 h2h3 f6d7 b3c2 h7h6 b2b4 f7f5 b4c5 d7c5 a3a4 d8d7 e3c5 b6c5 b1b3 e7g6 a4b5 a6b5 a1a8 b7a8 e4f5 g6f4 c2e4 g8h8 e1d1 d7f7 b3b5');
INSERT INTO evaluations VALUES(23930,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(23930,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(23930,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(23930,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(23930,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(23930,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(23930,36,'cp 42','c2c3 d7d6 a2a4 c5b6 d2d4 e8g8 c1g5 h7h6 g5h4 e5d4 c3d4 g7g5 h4g3 f8e8 d4d5 c6a5 b3c2 a5c4 b2b3 c4e5 f3d4 c7c6 d5c6 b6d4 d1d4 e5c6 d4b2 c6b4 b1d2 b4c2 b2c2 a8c8 c2d3 b5a4 b3a4 d6d5 e4e5 f6e4 d2e4 d5e4 d3e3 d8d7 a1d1 d7f5 e3a7 b7c6 a7a6');
INSERT INTO evaluations VALUES(23933,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(23933,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(23933,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(23933,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(23933,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(23933,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(23933,36,'cp 20','c5b6 d2d4 d7d6 c1e3 e8g8 d4d5 c6e7 e3b6 c7b6 b1d2 h7h6 b3c2 b7c8 d2f1 c8d7 d1e2 f6h7 f1e3 f7f5 e4f5 e7f5 e2d3 h7g5 f3g5 d8g5 e3f1 a6a5 f1g3 b5b4 g3e4 g5g6 a2a4 b4c3 b2c3 a8c8 h2h3 f5e7 d3g3 g6g3 f2g3 e7d5 e4d6 c8c3');
INSERT INTO evaluations VALUES(26223,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(26223,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(26223,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(26223,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(26223,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(26223,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(26223,36,'cp 15','b5c6 b7c6 d1f3 a8c8 e1g1 g7g6 c2c4 f8g7 f1d1 d8c7 c4c5 d6c5 c1f4 g7e5 f4e5 c7e5 b1d2 g8f6 d2c4 e5e6 f3c3 e8g8 f2f3 f6h5 b2b3 h5f4 c3e3 e6f6 e3c5 f8d8 c5e5 f6e5 c4e5 c6c5 g2g3 f4e6 d1d8 c8d8 e5c6 d8d3 g1f2 d3d2 f2e3 d2h2');
INSERT INTO evaluations VALUES(26222,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(26222,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(26222,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(26222,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(26222,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(26222,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(26222,36,'cp 17','b7c6 d1f3 a8c8 b2b3 d6d5 c1b2 g8f6 b1d2 e7e6 e1g1 f8e7 f1d1 e8g8 c2c4 h7h6 a1c1 d8b6 h2h3 f8d8 c4d5 c6d5 c1c8 d8c8 e4d5 f6d5 d2c4 b6c7 b2e5 c7c6 a2a4 e7f8 g2g3 d5b6 f3c6 c8c6 c4e3 c6c5');
INSERT INTO evaluations VALUES(25846,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25846,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(25846,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(25846,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(25846,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(25846,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(25846,36,'cp 13','d1f3 a8c8 b1c3 g7g6 b2b3 g8f6 c1b2 f8g7 e1c1 d8a5 c1b1 e8g8 f3e2 f8e8 c3a4 c8b8 f2f3 b8b4 d1d4 b4d4 b2d4 e8b8 d4b2 f6d7 e2e1 a5e1 h1e1 h7h6 b2g7 g8g7 a4b2 a7a5 e1d1 d7c5 h2h3');
INSERT INTO evaluations VALUES(26607,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26607,31,'cp -7',NULL);
INSERT INTO evaluations VALUES(26607,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26607,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(26607,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26607,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(26607,36,'cp 0','e2e3 e7e6 c2c4 g8f6 f1e2 f8e7 a2a3 e8g8 e1g1 d5c4 e2c4 g4f3 g2f3 f6d5 g1h1 e7f6 d1e2 c6e7 b1d2 e7f5 d2b3 d5e7 b3c5 b7b6 c5e4 e7d5 e4g3 d5e7');
INSERT INTO evaluations VALUES(28331,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(28331,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(28331,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(28331,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(28331,34,'cp 135',NULL);
INSERT INTO evaluations VALUES(28331,35,'cp 135',NULL);
INSERT INTO evaluations VALUES(28331,36,'cp 138','e4e5 f6e4 c3e4 b7e4 h2h4 c7c5 h4h5 c5d4 h5h6 g7f8 f3d4 e4b7 c1f4 b8c6 a2a3 c6d4 d1d4 e7e6 e1c1 f8c5 d4d2 b7c6 c1b1 c5e7 h1h3 e8g8 g2g4 a7a6 d1c1 b6b5 h3b3 a8b8 b1a2 b5c4 f1c4 a6a5 b3b8 d8b8');
INSERT INTO evaluations VALUES(28862,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(28862,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(28862,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(28862,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(28862,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(28862,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(28862,36,'cp -9','d7b5 c3b5 a7a6 b5c3 f8g7 f1e1 b8d7 c3d5 e8g8 c2c3 f8e8 d4c2 b7b5 d5f6 d7f6 c1f4 a6a5 a2a3 f6d7 f4e3 d8c7 e3d4 a8c8 d4g7 g8g7 d1d4 g7g8 f2f3 d7e5 c2e3 c7c5 a1d1 e5c4 e3c4 c5d4 d1d4 c8c4 d4d5 b5b4 c3b4');
INSERT INTO evaluations VALUES(30549,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(30549,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(30549,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(30549,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(30549,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(30549,35,'cp 111',NULL);
INSERT INTO evaluations VALUES(30549,36,'cp 101','e4e5 d8d7 g1e2 c7c5 c2c3 b7a6 d3a6 b8a6 e1g1 e7e6 e2g3 g8e7 g3h5 e8c8 a2a3 c8b7 b1d2 d8c8 d2f3 h7h6 b2b4 c5b4 c3b4 c8c4 h5g3 g7g6 g3e2 e7f5 g2g4 f5g7 d1d3 h6h5 f3d2 c4c8 b4b5 a6c7');
INSERT INTO evaluations VALUES(30555,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(30555,31,'cp 94',NULL);
INSERT INTO evaluations VALUES(30555,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(30555,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(30555,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(30555,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(30555,36,'cp 103','d8d7 g1e2 e7e6 e1g1 c7c5 c2c3 b7a6 d3a6 b8a6 e2g3 g8e7 d1e2 a6c7 g3h5 d7a4 b2b3 a4a6 e2d1 e8c8 a2a4 e7c6 c1g5 d8d7 g5e3 c5d4 c3d4 c6b4 b1d2 c8b8 h5f4 f8e7 d2f3 h8c8 f3e1 g7g6 f4d3 a6b7 g2g3 b8a8 d3b4 e7b4');
INSERT INTO evaluations VALUES(30660,30,'cp 71',NULL);
INSERT INTO evaluations VALUES(30660,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(30660,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(30660,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(30660,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(30660,35,'cp 80',NULL);
INSERT INTO evaluations VALUES(30660,36,'cp 94','e2e4 f8b4 f1d3 g8f6 f3d2 c7c5 d4c5 b8c6 c5b6 a7b6 e1g1 e8g8 d2f3 b4c3 b2c3 d8c7 f1e1 c6e5 f3e5 c7e5 a1b1 a8a2 c3c4 b7c6 f2f4 e5c5 c1e3 c5e7 d1f3 d7d6 b1b6 e7c7 e3d4 e6e5 f4e5 d6e5 d4e3 f6d7 b6b1 c7d6 f3f2 d6g6 e3b6 d7f6 f2g3 g6g3 h2g3 f8a8 c4c5');
INSERT INTO evaluations VALUES(31275,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(31275,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(31275,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(31275,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(31275,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(31275,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(31275,36,'cp 40','f4b8 a8b8 d4c5 g8f6 f1b5 f5d7 b5d7 d8d7 b2b4 b7b6 c5b6 b8b6 a2a3 e7e6 b1d2 b6c6 a1c1 f8e7 g1f3 c6c3 d2b1 c3c7 c2c4 d5c4 d1d7 f6d7 e1e2 e8g8 c1c2 f8c8 h1c1 d7b6 b1d2 c4c3 d2b3 a7a6 f3e5 g7g5 e5d3 g5g4 b3c5');
INSERT INTO evaluations VALUES(33917,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(33917,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(33917,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(33917,33,'cp -5',NULL);
INSERT INTO evaluations VALUES(33917,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(33917,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(33917,36,'cp -5','g1f3 c8g4 d1e1 b8c6 f1b5 g4d7 b5c6 d7c6 f3e5 c6e4 c3e4 f6e4 e1e2 e4d6 h1d1 e8c8 e2f1 f8e7 b2b3 e7f6 c1b2 b7b6 d1e1 d6f5 a1d1 d8d1 e1d1 h8e8 e5d3 f6b2 d3b2 e8d8 d1d8 c8d8 f1e2 d8d7 g2g3 d7e6 e2d3 h7h5 b2c4 e6d5 c4e3 f5e3 d3e3 b6b5 f2f3 c7c5');
INSERT INTO evaluations VALUES(33915,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(33915,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(33915,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(33915,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(33915,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(33915,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(33915,36,'cp -12','c8g4 d1e1 b8d7 a2a3 f8c5 h2h3 g4f3 g2f3 d7f8 f1c4 c7c6 h1g1 g7g6 c1h6 e8e7 c3e2 a7a5 a1d1 f8e6 c4e6 e7e6 e2c1 a5a4 c1d3 c5d6 h6e3 f6h5 e3c5 h8d8 c5d6 e6d6 d3f4 d6e7 f4h5 d8d1 e1d1 g6h5 g1g7');
INSERT INTO evaluations VALUES(34626,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(34626,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(34626,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(34626,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(34626,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(34626,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(34626,36,'cp 62','f5a5 g1f3 g8f6 e1g1 e7e6 f1e1 f8e7 b2b3 e8g8 c1b2 e7a3 b2a3 a5a3 f3e5 b8d7 e5c4 a3e7 d1f3 f8d8 a2a4 d7c5 d3e2 c5a6 a1c1 a6b4 f3f4 b7b6 e2f3 c7c6 a4a5 c8a6 c4e5 b6a5 e5c6 b4c6 f3c6 a8b8 c6f3 a6b7 f3b7 e7b7');
INSERT INTO evaluations VALUES(37618,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(37618,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(37618,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(37618,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(37618,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(37618,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(37618,36,'cp 12','d7d5 e4d5 e6d5 d1e2 g8e7 c4b5 c8d7 f1e1 d5d4 c3d4 c5d4 b1a3 g7f6 d2d3 e8g8 c1h6 f8e8 e2d2 d7b5 a3b5 b8c6 h6f4 e7d5 e1e8 d8e8 f4g3 e8d7 a2a3 a7a6 b5d6 f6e7 d6c4 f7f6 c4a5 c6a5 d2a5 e7d8 a5d2 d8b6 a1e1 a8e8 h2h4 e8e1 d2e1');
INSERT INTO evaluations VALUES(39182,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(39182,31,'cp 197',NULL);
INSERT INTO evaluations VALUES(39182,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(39182,33,'cp 179',NULL);
INSERT INTO evaluations VALUES(39182,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(39182,35,'cp 180',NULL);
INSERT INTO evaluations VALUES(39182,36,'cp 195','a5b6 e4e5 g8e7 d1g4 e7g6 h2h4 h7h5 g4g3 d7d6 g1f3 e8g8 c1g5 d8e8 f1d3 c8d7 a2a4 d7a4 c3c4 a4c6 c4c5 d6c5 b4c5 c6f3 g3f3 b6c5 d4c5 g6e5 f3e4 e5d3 e4d3 e8d7 d3b3 d7d5 b3d5 e6d5 e1g1 b8d7 g5e3 f8b8 b1c3 c7c6 f1b1 b7b6 c5b6 a7b6 a1a8 b8a8 e3b6 d7b6 b1b6');
INSERT INTO evaluations VALUES(39081,30,'cp 187',NULL);
INSERT INTO evaluations VALUES(39081,31,'cp 210',NULL);
INSERT INTO evaluations VALUES(39081,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(39081,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(39081,34,'cp 217',NULL);
INSERT INTO evaluations VALUES(39081,35,'cp 179',NULL);
INSERT INTO evaluations VALUES(39081,36,'cp 228','e4e5 g8e7 d1g4 e7g6 g1f3 d7d6 h2h4 h7h5 g4g3 e8g8 f1d3 d6e5 c1g5 d8d6 f3e5 b8c6 e5g6 d6g3 f2g3 f7g6 a2a4 e6e5 a4a5 b6d4 c3d4 c6b4 d3c4 g8h7 b1a3 e5d4 g5f4 c7c5 e1d2 b7b6 a3b5 a7a6 a5b6 a6b5 a1a8 b5c4 b6b7 c4c3 d2c1 c8b7 a8f8 b7g2 h1h2 g2c6 f8f7 d4d3');
INSERT INTO evaluations VALUES(39684,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39684,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(39684,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39684,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(39684,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(39684,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(39684,36,'cp 0','e8g8 h2h3 c8e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(35565,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(35565,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(35565,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(35565,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(35565,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(35565,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(35565,36,'cp 0','h2h3 h7h6 c1e3 b7b5 f3h4 f6e8 h4f3');
INSERT INTO evaluations VALUES(40751,30,'cp 97',NULL);
INSERT INTO evaluations VALUES(40751,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(40751,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(40751,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(40751,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(40751,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(40751,36,'cp 86','e5f4 e8g8 b1c3 d6d5 c3d5 f6d5 e4d5 f8e8 f1e2 d8d5 e1g1 e7d6 e2c4 d6f4 c4d5 c7c6 d5b3 c8g4 d2d4 f4c1 f1c1 g4f3 g2f3 b8d7 g1f1 g8f8 c1e1 d7f6 c2c3 a7a5 a2a4 a8d8 f3f4 e8e4 e1e4 f6e4 b3c2 e4f6 a1e1 f6d5 f4f5 d5b6 f1e2 b6c4 b2b3 c4d6');
INSERT INTO evaluations VALUES(42153,30,'cp 90',NULL);
INSERT INTO evaluations VALUES(42153,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(42153,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(42153,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(42153,34,'cp 100',NULL);
INSERT INTO evaluations VALUES(42153,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(42153,36,'cp 74','e1g1 e8g8 d1e2 h7h6 f1d1 b8c6 c3e4 b4d6 c2c3 f6e4 e2e4 a7a5 d1e1 c6e7 e4g4 e7g6 h2h4 d8f6 h4h5 g6e7 c4d3 f8d8 d3b1 c8d7 f3e5 c7c5 g4e4 d7c6 e5c6 b7c6 c1e3 e7d5 d4c5 d6f4 e3d4 f6g5 b1d3 g5h5 e4h7 g8f8');
INSERT INTO evaluations VALUES(42152,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(42152,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(42152,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(42152,33,'cp 94',NULL);
INSERT INTO evaluations VALUES(42152,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(42152,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(42152,36,'cp 90','e8g8 d1e2 h7h6 c3e4 b8d7 f3e5 f6d5 f1d1 f7f5 a2a3 d7e5 d4e5 b4e7 e4g3 d8e8 g3h5 g8h7 c4d3 g7g6 c2c4 g6h5 c4d5 e6d5 e5e6 e8g6 d1e1 f8f6 c1f4 c8e6 f4c7 a8e8 a1d1 e7f8 e2d2 f6f7 c7g3 f8g7 h2h4 h7g8 a3a4 g6f6 a4a5');
INSERT INTO evaluations VALUES(45956,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45956,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(45956,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45956,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45956,34,'cp 3',NULL);
INSERT INTO evaluations VALUES(45956,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(45956,36,'cp 28','c8f5 e4f5 d6d5 c1g5 f8b4 f1b5 b8c6 g5f6 g7f6 e1g1 b4c3 b2c3 d8d6 d1h5 d6c5 a1b1 a7a6 b5c6 c5c6 a2a4 b7b5 a4b5 a6b5 h5e2 c6c4 e2c4 b5c4 f1d1 e8g8 d1d5 a8a3 h2h3 a3c3 d5d2 f8a8 b1b7 a8a2 h3h4 c3c2 d2d8 g8g7 d8d7 c2f2 d7f7 g7g8 f7g7 g8h8');
INSERT INTO evaluations VALUES(29353,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(29353,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(29353,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(29353,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(29353,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(29353,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(29353,36,'cp 41','e4f5 d6d5 c1g5 f8b4 g5f6 g7f6 f1b5 b8c6 e1g1 b4c3 b2c3 d8d6 d1h5 d6c5 a1b1 e8e7 b5c6 c5c6 f1e1 a8c8 h5h4 e7f8 e1e3 h7h5 h4b4 f8g7 e3g3 g7h6 b4a3 h8g8 a3a7 g8g3 h2g3 c8c7 b1b6 c6c3 b6b7 c7b7 a7b7 c3e1 g1h2 e1f2 b7f7');
INSERT INTO evaluations VALUES(29352,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(29352,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(29352,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(29352,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(29352,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(29352,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(29352,36,'cp 21','d6d5 c1g5 f8b4 g5f6 g7f6 f1b5 b8c6 e1g1 b4c3 b2c3 d8d6 d1h5 e8e7 a1b1 d6c5 b5c6 c5c6 f1e1 a8c8 e1e3 b7b6 h5h4 e7f8 h4b4 c6c5 b4a4 c8c7 b1b5 c5c4 a4c4 c7c4 b5d5 f8e7 e3d3 c4c7 h2h4 h8g8 g2g3 g8c8 d5d6 c7c3 d6d7 e7f8 d7a7 c3d3 c2d3');
INSERT INTO evaluations VALUES(46021,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(46021,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(46021,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(46021,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(46021,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(46021,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(46021,36,'cp 1','b1d2 b8c6 a2a3 e6e5 g2g3 f8e7 f1g2 c8e6 e1g1 e8g8 f1e1 a8c8 b2b4 b7b5 c1b2 a7a6 d3d4 c5d4 c3d4 e5d4 f3d4 c6d4 b2d4 f6d7 d2f1 d7e5 f1e3 e7f6 a3a4 e5c4 d4f6 d8f6 e3d5 e6d5 d1d5 f8d8 a4b5 a6b5');
INSERT INTO evaluations VALUES(46280,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46280,36,'cp 0','e1g1 b8c6 e2e4 d5e4 d3e4 d8d1 f1d1 f6e4 b1a3 f8e7 f3e5 c6e5 g2e4 e8g8 c1f4 e7f6 f4e3 f6e7');
INSERT INTO evaluations VALUES(47132,30,'cp -111',NULL);
INSERT INTO evaluations VALUES(47132,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(47132,32,'cp -109',NULL);
INSERT INTO evaluations VALUES(47132,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(47132,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(47132,35,'cp -104',NULL);
INSERT INTO evaluations VALUES(47132,36,'cp -84','d5d4 c3a4 b8c6 d2d3 e7e5 a2a3 a7a5 g2g3 f8e7 f1g2 h7h6 e1g1 e8g8 e2e3 f8e8 f1e1 c8e6 e3d4 e5d4 c1f4 a8b8 f3e5 c6e5 f4e5 e7f8 b2b3 f6g4 e5f4 d8d7 h2h3 g4f6 g1h2 f8d6 d1d2 b7b5 a4b2 b5c4 b2c4 d6f4 g3f4');
INSERT INTO evaluations VALUES(47378,30,'cp -247',NULL);
INSERT INTO evaluations VALUES(47378,31,'cp -249',NULL);
INSERT INTO evaluations VALUES(47378,32,'cp -240',NULL);
INSERT INTO evaluations VALUES(47378,33,'cp -235',NULL);
INSERT INTO evaluations VALUES(47378,34,'cp -228',NULL);
INSERT INTO evaluations VALUES(47378,35,'cp -221',NULL);
INSERT INTO evaluations VALUES(47378,36,'cp -235','b8c6 c3e4 d8d4 e4f6 d4f6 c2c3 e7e5 c1e3 f8e7 g1f3 h7h6 e2c4 c8e6 c4a4 e8g8 f1d3 a8d8 e1c1 e6d5 d3c2 f6e6 c2b3 d5b3 a4b3 e6b3 a2b3 f7f5 b3b4 a7a6 e3c5 e7c5 b4c5 a6a5 h1e1 e5e4 f3d4 c6d4 d1d4 d8d4');
INSERT INTO evaluations VALUES(47515,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(47515,31,'cp 154',NULL);
INSERT INTO evaluations VALUES(47515,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(47515,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(47515,34,'cp 164',NULL);
INSERT INTO evaluations VALUES(47515,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(47515,36,'cp 157','d2d4 g4h6 c1h6 g7h6 f2f4 e7e6 d1h5 c7c5 d4c5 c8d7 g1f3 f8c5 e1c1 b8c6 f1b5 h8g8 g2g3 d8b6 b5a4 g8f8 c1b1 e8c8 h5h6 c5b4 d1d3 c8b8 h1d1 c6a5 a4d7 d8d7 a2a3 b4e7 c3e2 f8c8 h6h7 e7c5 e2d4 d7c7 f4f5 a5c4 d3b3 b6a5');
INSERT INTO evaluations VALUES(47601,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(47601,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(47601,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(47601,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(47601,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(47601,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(47601,36,'cp 38','c7c5 g1f3 g5g4 f3e5 d8b6 e2e3 b6b2 d1c1 b2c1 a1c1 c5d4 c3b5 f8b4 e1d1 b8a6 e3d4 d5c4 f1c4 b4e7 d1e2 f6h5 e5g4 h5g3 h2g3 e8f8 g4h6 c8d7 h6f5 h8h1 c1h1 e7f6 f5d6 d7c6 h1h7 f6g7 h7h4 f8e7 h4f4 c6b5 d6b5 a6b4 f4g4 g7f6 a2a3 b4c6 d4d5 c6e5');
INSERT INTO evaluations VALUES(49548,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(49548,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(49548,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(49548,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(49548,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(49548,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(49548,36,'cp -40','f8c5 b1c3 d7d6 d3e2 b8c6 d2d3 f6d7 e2g4 h7h5 g4h3 d7f8 h3c8 d8c8 c3d5 a7a5 g1e2 f8e6 e1g1 c6d4 e2d4 e6d4 c1e3 c7c6 e3d4 c5d4 d5e3 g7g6 e3c2 d4a7 d3d4 f7f5 d1d3 f5e4 d3e4 c8f5');
INSERT INTO evaluations VALUES(49837,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49837,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49837,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49837,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49837,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49837,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49837,36,'cp 0','c4f7 e8f7 d1d5 f7e8 d5e4 d8g5 g1e2 g5g6 e4e3 b7b6 b1c3 g6c2 e5e6 d7e6 e2d4 c2b2 e1g1 f8c5 a1b1 b2a3 e3e4 a3c3 d4b5 c5f2 f1f2 c3c5 e4a8 h8f8 b1f1 f8f2 f1f2 c5c1 f2f1 c1c5');
INSERT INTO evaluations VALUES(50693,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(50693,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(50693,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(50693,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(50693,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(50693,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(50693,36,'cp -8','f1e2 b8c6 d1b3 e7e6 c1f4 f8e7 a1d1 d8c7 b3a3 g8f6 f4d6 e7d6 d1d6 e8g8 e1g1 b7b5 a3c5 c8b7 f1d1 a8d8 d6d8 f8d8 f3d4 h7h6 d4c6 d8d1 c3d1 b7c6 f2f3 f6d7 c5d4 e6e5 d4c3 c7b6 d1e3 d7c5 g1f1 b5b4 c3c4 a6a5 h2h3 c6b7');
INSERT INTO evaluations VALUES(52438,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(52438,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(52438,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(52438,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(52438,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(52438,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(52438,36,'cp -29','b8c6 b1c3 g8f6 f1b5 d7d6 g1e2 a7a6 b5c6 b7c6 d2d3 e7e5 e2g3 f8e7 d1e2 a6a5 c3a4 e8g8 c1d2 f8e8 e1c1 d6d5 f2f3 a8b8 g3f5 c8f5 e4f5 e7d6 g2g4 c5c4 d3c4 e5e4 c4c5 d6e5 g4g5 e4f3 e2f3 f6e4 d2f4');
INSERT INTO evaluations VALUES(54213,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(54213,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(54213,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(54213,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(54213,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(54213,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(54213,36,'cp -12','c5c4 d3e2 b7b5 a2a4 b5b4 c2c3 a7a5 b2b3 c8a6 c1b2 c4b3 e1g1 g8f6 b1d2 f8d6 c3b4 d6b4 e2a6 b8a6 d1b3 e8g8 f1c1 d8d6 b3d3 a8b8 f3e5 b4d2 d3d2 a6b4 b2a3 f8c8 e5d3 c8c4 d3e5 c4c1 a1c1 f6e4 d2e2 h7h6');
INSERT INTO evaluations VALUES(54318,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(54318,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(54318,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(54318,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(54318,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(54318,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(54318,36,'cp -45','b8c6 e1g1 g8e7 c2c3 d5d4 f1e1 e7g6 e2f1 b7b6 d1a4 c8d7 a4d1 f8e7 d2d3 f7f5 e5f6 e7f6 g2g3 e8g8 f1g2 g8h8 h2h4 d8e8 c1g5 e6e5 g5f6 g7f6 c3d4 c5d4 b1d2 g6e7 f3h2 e8f7 d1e2 f7g7 g2e4');
INSERT INTO evaluations VALUES(20362,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(20362,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(20362,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(20362,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(20362,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(20362,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(20362,36,'cp -41','e1g1 g8e7 c2c3 d5d4 f1e1 e7g6 e2f1 b7b6 d2d3 c8b7 g2g3 d4c3 b2c3 d8d5 c3c4 d5d7 c1b2 f8e7 b1d2 e8g8 a1b1 f8d8 h2h4 g6f8 b2c3 d7c7 h4h5 h7h6 d2e4 f8h7 d1d2 d8d7 d2f4 a8d8 f4g4 b7a8');
INSERT INTO evaluations VALUES(20361,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(20361,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(20361,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(20361,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(20361,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(20361,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(20361,36,'cp -40','g8e7 c2c3 d5d4 f1e1 d4d3 e2f1 b7b5 b2b4 a8b8 a2a4 c5b4 a4b5 b8b5 c3b4 c6b4 a1a7 e7c6 a7a8 b5a5 a8a5 d8a5 d1b3 c8a6 c1b2 f8e7 b1c3 e8g8 e1a1 a5b6 a1b1 f8b8 b3a4 b6a7 b2a3 a6c4 a4a7 c6a7 c3e4 a7c6');
INSERT INTO evaluations VALUES(54358,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54358,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(54358,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(54358,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(54358,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(54358,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(54358,36,'cp 17','g8f6 e4d5 f6d5 d2f3 f8e7 f1d3 d5b4 d3e4 f7f5 e4d3 b8c6 d4c6 b4d3 d1d3 d8d3 c2d3 b7c6 c1g5 c6c5 a1c1 a8b8 b2b3 b8b5 g5e7 e8e7 f3e5 c8b7 f2f4 h8c8 e1f2 b5b4 f2e3 b7d5 h1g1 e7d6 c1c2 c8g8 e5c4 d5c4 c2c4 d6c6 g1c1 b4c4 d3c4');
INSERT INTO evaluations VALUES(337,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(337,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(337,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(337,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(337,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(337,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(337,36,'cp 0','g8f6 d4f5 h4e4 d3e4 f6e4 f5g7 e8f8 c1h6 e4c3 g7h5 f8e7 h6g7 c3e4 c2c3 b4c5 g7h8 e4f2 h8f6 e7f8 h1g1 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(553,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(553,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(553,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(553,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(553,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(553,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(553,36,'cp 0','d4f5 h4e4 d3e4 f6e4 f5g7 e8f8 c1h6 e4c3 g7h5 f8e7 h6g7 c3e4 c2c3 b4c5 g7h8 e4f2 h8f6 e7f8 h1g1 f2e4 g1h1 e4f2');
INSERT INTO evaluations VALUES(869,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(869,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(869,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(869,33,'cp 143',NULL);
INSERT INTO evaluations VALUES(869,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(869,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(869,36,'cp 146','b1c3 f6e4 g1e2 c6b4 c3e4 e7e4 d2d3 e4c6 c7c6 d7c6 e1d1 h7h5 c1d2 c6c5 h2h4 c8g4 f2f3 g4f5 e2g3 f5g6 a2a3 b4c6 d2c3 e8c8 g3e4 c6d4 d1d2 g6e4 f3e4 g7g6 a1e1 f8h6 d2d1 c5c4 g2g3 c8b8 f1e2 c4d3 e2d3 d4f3 c3h8 f3e1');
INSERT INTO evaluations VALUES(5789,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(5789,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5789,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5789,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(5789,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(5789,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5789,36,'cp 0','b7c6 c2c3 g8e7 d2d4 e5d4 c3d4 c5b6 b1c3 e8g8 h2h3 e7g6 f1e1 g6h4 f3h4 d8h4 c3a4 f8e8 a4b6 a7b6 d1f3 a8a4 f3d1');
INSERT INTO evaluations VALUES(7889,30,'cp -160',NULL);
INSERT INTO evaluations VALUES(7889,31,'cp -154',NULL);
INSERT INTO evaluations VALUES(7889,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(7889,33,'cp -154',NULL);
INSERT INTO evaluations VALUES(7889,34,'cp -157',NULL);
INSERT INTO evaluations VALUES(7889,35,'cp -154',NULL);
INSERT INTO evaluations VALUES(7889,36,'cp -152','d1d6 f6e4 d6d5 e4f6 d5d6 d8e7 d6e7 e8e7 b2b3 d7d6 f2f4 f6d5 f4e5 d6e5 f1c4 d5b4 c1a3 e7e8 a3b4 c6b4 b1a3 e8e7 c2c3 b4a6 a3b5 h8d8 e1g1 f7f6 a1e1 c8e6 c4e6 e7e6 e1d1 e6e7 d1d8 a8d8 b5a7 d8d3 a7b5 e7e6 c3c4 a6b4 a2a3 b4c2 b3b4 c2e3 b5c7 e6d7');
INSERT INTO evaluations VALUES(7886,30,'cp -159',NULL);
INSERT INTO evaluations VALUES(7886,31,'cp -181',NULL);
INSERT INTO evaluations VALUES(7886,32,'cp -162',NULL);
INSERT INTO evaluations VALUES(7886,33,'cp -164',NULL);
INSERT INTO evaluations VALUES(7886,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(7886,35,'cp -184',NULL);
INSERT INTO evaluations VALUES(7886,36,'cp -187','f6e4 d6d5 e4f6 d5d6 d8e7 d6e7 e8e7 b2b3 d7d6 f2f4 f6d5 f4e5 d6e5 c1a3 e7d8 e1d2 d8c7 b1c3 d5c3 d2c3 c8e6 f1d3 f7f6 a1e1 h7h5 a3c1 a7a5 a2a3 h5h4 h1f1 a8d8 c1e3 e6d5 f1f2 c6e7 c3b2 d8d7 e3d2 b7b6 a3a4 d5b7 d2e3 e7d5 d3b5 b7c6');
INSERT INTO evaluations VALUES(7887,30,'cp -167',NULL);
INSERT INTO evaluations VALUES(7887,31,'cp -158',NULL);
INSERT INTO evaluations VALUES(7887,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(7887,33,'cp -166',NULL);
INSERT INTO evaluations VALUES(7887,34,'cp -171',NULL);
INSERT INTO evaluations VALUES(7887,35,'cp -177',NULL);
INSERT INTO evaluations VALUES(7887,36,'cp -156','d6d5 e4f6 d5d6 d8e7 d6e7 e8e7 f1d3 h8d8 e1g1 d7d5 b2b3 c8e6 c1a3 e7e8 b1c3 a7a6 c3a4 e5e4 d3e2 a8c8 f1d1 f6d7 a1c1 c6e5 a3b2 f7f6 f2f3 e4f3 g2f3 e5g6 g1f2 e8f7 b2d4 g6f4 c2c3 f4e2 f2e2 h7h6 a4b6 d7b6');
INSERT INTO evaluations VALUES(8433,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(8433,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(8433,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(8433,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(8433,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(8433,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(8433,36,'cp 73','d2d4 e8g8 f1e1 c6d4 f3d4 e5d4 d1d4 c7c6 a2a4 d8b6 d4b6 a7b6 c1e3 f6d7 e1d1 f8d8 f2f3 d7c5 b2b4 e7f6 e3d4 f6d4 d1d4 c5e6 d4d2 b6b5 c4e6 f7e6 d2d1 b5a4 a1a4 a8a4 c3a4 e6e5 a4b6 c8e6 c2c4 g8f8 g1f2 f8f7 b4b5 f7f6 d1a1 c6b5 c4b5 d6d5 e4d5 d8d6 d5e6 d6b6');
INSERT INTO evaluations VALUES(8452,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(8452,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(8452,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(8452,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(8452,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(8452,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(8452,36,'cp 52','e5d4 f3d4 e8g8 f1e1 c6d4 d1d4 c7c6 a2a4 d8b6 d4b6 a7b6 c1e3 f6d7 e1d1 f8d8 f2f3 g8f8 g1f2 d7c5 b2b4 e7f6 b4c5 f6c3 c5d6 c3a1 e3b6 c8e6 b6d8 a8d8 c4e6 a1e5 e6b3 d8d6 d1d6 e5d6 f2e3 f7f6 f3f4 g7g5 f4g5 h6g5 e3f3 f8e7 f3g4 d6e5 g4f5 e5c3 b3g8');
INSERT INTO evaluations VALUES(9006,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9006,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9006,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9006,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9006,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(9006,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(9006,36,'cp 0','h2g3 e6e5 d4e5 c6e5 g1f3 d8e7 f3e5 e7e5 d1a4 c7c6 f1d3 h7h6 e1g1 e8g8 f1e1 f8e8 a4f4 e5e7 a2a3 c8e6 a1d1 a8d8 e3e4 e6g4 f2f3 e7c5 g1h2 d5e4 d2e4 f6e4 d3e4 c5h5 h2g1 h5c5');
INSERT INTO evaluations VALUES(10595,30,'cp 363',NULL);
INSERT INTO evaluations VALUES(10595,31,'cp 377',NULL);
INSERT INTO evaluations VALUES(10595,32,'cp 351',NULL);
INSERT INTO evaluations VALUES(10595,33,'cp 337',NULL);
INSERT INTO evaluations VALUES(10595,34,'cp 375',NULL);
INSERT INTO evaluations VALUES(10595,35,'cp 376',NULL);
INSERT INTO evaluations VALUES(10595,36,'cp 378','e4f2 g1f2 h7h6 b1d2 e8g8 d2e4 d7d5 e5d6 c7d6 e4c5 d6c5 f2g1 c8f5 c1f4 a7a6 a2a4 d8d7 d1d2 a8e8 e1e8 f8e8 f4g3 f5g6 a1e1 e8e1 g3e1 d7e8 e1g3 c6b4 f3e5 b7b5 a4b5 a6b5 c4b3 b4c6 e5g6 c5c4 g6f4 c4b3 f4d3 b3c2');
INSERT INTO evaluations VALUES(11251,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(11251,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(11251,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(11251,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(11251,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(11251,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(11251,36,'cp -34','b7b5 c4a2 f8e7 e1g1 c8b7 f1e1 e8g8 c1f4 d8c7 f4g3 c6e5 f3e5 d6e5 d1e2 b7c6 a2b3 f8d8 a3a4 f6d7 a4b5 a6b5 a1a8 d8a8 c3b1 h7h6 b1d2 e7g5 d2f3 g5f4 g3f4 e5f4 c2c3 d7e5 f3e5 c7e5 e2d2 a8d8 b3c2 g7g6 c2b3 c6e8 b3c2 c5c4');
INSERT INTO evaluations VALUES(11805,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(11805,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(11805,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(11805,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(11805,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(11805,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(11805,36,'cp -20','c4a2 f8e7 e1g1 e8g8 c3e2 d8c7 c2c3 c5c4 e2g3 c4d3 d1d3 c6e5 d3e2 c8b7 f1e1 h7h6 f3e5 d6e5 c1d2 a8d8 a1d1 d8d7 a2b1 c7c4 e2c4 b5c4 d2e3 b7c6 b1c2 f8d8 a3a4 d7d1 e1d1 d8d1 c2d1 f6e4 g3e4');
INSERT INTO evaluations VALUES(11806,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(11806,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(11806,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(11806,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(11806,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(11806,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(11806,36,'cp -31','f8e7 e1g1 d8c7 c3e2 e8g8 c2c3 f8d8 f1e1 c5c4 e2g3 c4d3 d1d3 c8b7 d3e2 c6e5 f3d4 e5d7 a2b1 d8e8 b1c2 a8c8 c1d2 d7b6 a1d1 e7f8 d2g5 f6d7 e2g4 b6a4 d1b1 g7g6 g4h4 e6e5 d4f3 d6d5 c2a4 b5a4');
INSERT INTO evaluations VALUES(12527,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(12527,31,'cp 100',NULL);
INSERT INTO evaluations VALUES(12527,32,'cp 118',NULL);
INSERT INTO evaluations VALUES(12527,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(12527,34,'cp 98',NULL);
INSERT INTO evaluations VALUES(12527,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(12527,36,'cp 104','g1e2 f8e7 e1g1 e8g8 f2f4 e5f4 e2f4 d7d6 h2h3 f6d7 f4d5 d7e5 d5e7 c6e7 b2b3 c8e6 c1b2 e7c6 d1h5 e5c4 b3c4 c6e5 f1f2 f7f6 h5h4 b7b6 h4g3 d8d7 a2a4 a8e8 c3e2 f6f5 b2e5 d6e5 g3e5 f5e4 f2f8 e8f8 e5e4');
INSERT INTO evaluations VALUES(13215,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13215,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(13215,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13215,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(13215,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13215,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13215,36,'cp 0','e7e5 e1g1 f8e7 d2d3 e8g8 b5c4 h7h6 a2a3 a7a6 c3d5 f6d5 c4d5 c8e6 c2c4 d8d7 c1d2 f7f5 f1e1 f5e4 d5e4 e6f5 e4d5 f5e6');
INSERT INTO evaluations VALUES(13780,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(13780,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(13780,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(13780,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(13780,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(13780,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(13780,36,'cp -48','e1g1 f8e7 d2d3 c8g4 f1e1 d8d7 c1e3 e8g8 b1d2 f8e8 h2h3 g4h5 d2f1 b7b5 f1g3 h5g6 e3f4 a8d8 c2c3 f6h5 g3h5 g6h5 g2g4 h5g6 d3d4 c5d4 f3d4 c6d4 c3d4 e7f6 e1e8 d8e8 a1c1 d7e6 c1c5 g6e4 c5b5 h7h5 b5b8 e8b8 f4b8 h5g4 h3g4');
INSERT INTO evaluations VALUES(13779,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(13779,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(13779,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(13779,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(13779,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(13779,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(13779,36,'cp -36','f8e7 f1e1 c8g4 h2h3 g4h5 g2g4 h5g6 d2d3 d8d7 b1c3 e8c8 c1f4 h7h5 g4g5 d7h3 g5f6 g7f6 a2d5 e7d6 f4d6 d8d6 d5c6 b7c6 e1e3 h8g8 f3g5 h3h4 d1f3 h4g5 g1f1 c5c4 a1e1 c4d3 c2d3 g8d8 c3e4 g6e4 f3e4 g5g4 e3g3 g4e6 e4f4 e6d7 f1g2 d6d3 g3d3 d7d3 f4f6 d3g6 f6g6 f7g6');
INSERT INTO evaluations VALUES(13928,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(13928,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(13928,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(13928,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(13928,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(13928,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(13928,36,'cp -26','f8e7 a2a4 a7a6 e1g1 e8g8 f1e1 b7b6 c3e2 d6d5 e4d5 f6d5 c2c3 c8b7 d3d4 c5d4 e2d4 c6d4 d1d4 d8c7 c4d3 f8d8 d4e4 g7g6 e4c4 a8c8 c4c7 d5c7 d3e2 f7f6 f3d2 e6e5 d2c4 c7d5 c1h6 b7c6 h3h4 e7c5 e2f1');
INSERT INTO evaluations VALUES(8159,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(8159,31,'cp -21',NULL);
INSERT INTO evaluations VALUES(8159,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(8159,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(8159,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(8159,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(8159,36,'cp -20','a2a4 a7a6 e1g1 e8g8 f1e1 b7b6 c4b3 d8c7 c3e2 c6a5 b3a2 b6b5 c1f4 c8d7 e4e5 d6e5 f3e5 c7b7 e5d7 b7d7 e2c3 b5b4 c3e4 f8d8 e4d2 a5c6 a2b3 h7h6 d2c4 f6d5 d1g4 d5f4 g4f4 e7f6');
INSERT INTO evaluations VALUES(14257,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(14257,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(14257,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(14257,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(14257,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(14257,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(14257,36,'cp -15','c8f5 e3f4 e7e6 e2e3 f8b4 f1d3 f6e4 e1g1 b4c3 b2c3 e4c3 d1e1 c3e4 a1b1 a8b8 d3b5 e8g8 b5c6 b7c6 b1b8 d8b8 f3e5 b8b5 f2f3 e4d6 e1c3 f8e8 c3c6 f7f6 c6b5 d6b5 e5d3 b5c3 f1a1 f5d3 c2d3 e8c8 g1f2 g7g5 f4g3 h7h5 h2h4 g5g4 f3g4 h5g4 h4h5 g8f7');
INSERT INTO evaluations VALUES(14709,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(14709,31,'cp 133',NULL);
INSERT INTO evaluations VALUES(14709,32,'cp 139',NULL);
INSERT INTO evaluations VALUES(14709,33,'cp 150',NULL);
INSERT INTO evaluations VALUES(14709,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(14709,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(14709,36,'cp 149','e2e4 d5c3 b2c3 f8d6 f1c4 c6a5 c4d3 c7c5 e1g1 c8d7 d1e2 c5d4 c3d4 e8g8 a1b1 a8c8 e4e5 d6e7 e2e4 g7g6 c1h6 d7c6 e4f4 c6f3 h6f8 e7f8 g2f3 c8c7 f1d1 c7d7 d3b5 d7d5 g1g2 a7a6 b5d3 b7b5 d3e4 d5d4 d1d4 d8d4 f4f6 a5c6');
INSERT INTO evaluations VALUES(14711,30,'cp 148',NULL);
INSERT INTO evaluations VALUES(14711,31,'cp 152',NULL);
INSERT INTO evaluations VALUES(14711,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(14711,33,'cp 146',NULL);
INSERT INTO evaluations VALUES(14711,34,'cp 126',NULL);
INSERT INTO evaluations VALUES(14711,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(14711,36,'cp 135','d5c3 b2c3 f8d6 f1c4 c6a5 c4d3 c7c5 e1g1 c8d7 a1b1 c5d4 c3d4 a8c8 d1e2 e8g8 f1e1 b7b6 e4e5 d6e7 e2e4 g7g6 c1h6 d7c6 e4f4 c6f3 h6f8 g8f8 f4f3 d8d4 b1d1 d4c5 h2h4 e7h4 d3a6 c8d8 d1d8 h4d8 f3a8 a5c6 e1d1');
INSERT INTO evaluations VALUES(14726,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(14726,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(14726,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(14726,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(14726,34,'cp 163',NULL);
INSERT INTO evaluations VALUES(14726,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(14726,36,'cp 139','b2c3 f8d6 f1c4 c6a5 c4d3 c7c5 e1g1 c8d7 a1b1 a8c8 f3d2 c5d4 c3d4 c8c3 d2f3 e8g8 d1e2 c3c8 e4e5 d6e7 e2e4 g7g6 c1h6 d7c6 e4f4 c6f3 h6f8 g8f8 g2f3 c8c7 f1d1 c7d7 d3b5 d7d5 b5a4 a7a6 a4b3 a5b3 b1b3 b7b5 b3c3 f8g7 g1g2 a6a5 f4c1 e7g5 f3f4 g5h6 c3c7 d8h4');
INSERT INTO evaluations VALUES(14902,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(14902,31,'cp 73',NULL);
INSERT INTO evaluations VALUES(14902,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(14902,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(14902,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(14902,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(14902,36,'cp 86','e1g1 f8e7 f1e1 e8g8 a2a3 b7b6 b2b4 c8b7 c1b2 a7a5 b4b5 c6a7 f3e5 c7c6 a3a4 a8c8 e1e3 e7b4 c4f1 d8c7 d1e2 c6c5 c3a2 c5d4 b2d4 f8d8 c2c3 b4d6 c3c4 f6d7 a2c3 d7e5 d4e5 d8d7 e5d6 c7d6 e2g4 c8d8');
INSERT INTO evaluations VALUES(14901,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(14901,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(14901,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(14901,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(14901,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(14901,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(14901,36,'cp 66','f8e7 a2a3 e8g8 f1e1 b7b6 d4d5 c6a5 c4a2 f6d5 c3d5 e6d5 d1d5 d8d5 a2d5 c7c6 e1e7 c6d5 c1d2 c8e6 d2a5 b6a5 f3d4 f8e8 e7e8 a8e8 a3a4 g8f8 f2f3 f8e7 g1f2 e7d6 h2h4 e6d7 b2b3 d6c5 c2c3 c5d6 h4h5 g7g5 a1d1 g5g4 d1c1 e8b8 f2e3 g4f3 g2f3 b8g8');
INSERT INTO evaluations VALUES(15711,30,'cp 286',NULL);
INSERT INTO evaluations VALUES(15711,31,'cp 289',NULL);
INSERT INTO evaluations VALUES(15711,32,'cp 266',NULL);
INSERT INTO evaluations VALUES(15711,33,'cp 281',NULL);
INSERT INTO evaluations VALUES(15711,34,'cp 293',NULL);
INSERT INTO evaluations VALUES(15711,35,'cp 290',NULL);
INSERT INTO evaluations VALUES(15711,36,'cp 287','d2d4 d8h4 d1e2 e4g5 c1e3 c8g4 e2d3 g5e6 g2g3 h4h5 b1d2 e8c8 h2h3 g4f5 d3c4 f8e7 f1g2 h5g6 e1c1 h7h5 h1e1 a7a6 c1b1 d8d5 d2b3 d5b5 c4e2 f7f6 e5f6 e7f6 e2d2 h8e8 d2c3 g6h7 h3h4 e8d8');
INSERT INTO evaluations VALUES(16195,30,'cp 308',NULL);
INSERT INTO evaluations VALUES(16195,31,'cp 316',NULL);
INSERT INTO evaluations VALUES(16195,32,'cp 348',NULL);
INSERT INTO evaluations VALUES(16195,33,'cp 375',NULL);
INSERT INTO evaluations VALUES(16195,34,'cp 357',NULL);
INSERT INTO evaluations VALUES(16195,35,'cp 349',NULL);
INSERT INTO evaluations VALUES(16195,36,'cp 386','h2h3 g4e5 f3e5 c6e5 f1e1 d7d6 f2f4 f8e7 f4e5 d6d5 c4d3 c8e6 b1d2 d8d7 d2f3 c7c5 d3f1 e8g8 a2a4 a7a6 a4a5 b7b5 a5b6 d7c6 b2b4 c5b4 f1d3 c6b6 g1h1 f8c8 f3d4 b6d4 d3h7 g8h7 d1d4 h7g8 d4b2 e6f5 b2b3 c8c4 c1b2 f5c2 b3f3 c2e4 f3g4 c4c6 e1e4 d5e4 g4e4');
INSERT INTO evaluations VALUES(18662,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(18662,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(18662,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(18662,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(18662,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(18662,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(18662,36,'cp 116','c2c3 g8f6 f1e1 d8c7 b1a3 a7a6 c4f1 b7b5 a3c2 f8d6 d2d4 c5d4 c3d4 e8g8 d4d5 c6e7 a2a4 c8b7 c1d2 d6c5 b2b4 c5b6 a4a5 b6a7 d2e3 a7e3 d5d6 c7d8 d6e7 e3f2 g1f2 d8e7 f1d3 f8c8 f2g1 c8c3 d1d2 a8c8 h2h3 c3b3 d2e2 c8c3 a1c1');
INSERT INTO evaluations VALUES(20646,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(20646,31,'cp 212',NULL);
INSERT INTO evaluations VALUES(20646,32,'cp 233',NULL);
INSERT INTO evaluations VALUES(20646,33,'cp 210',NULL);
INSERT INTO evaluations VALUES(20646,34,'cp 238',NULL);
INSERT INTO evaluations VALUES(20646,35,'cp 217',NULL);
INSERT INTO evaluations VALUES(20646,36,'cp 226','c2c4 g8f6 b1c3 g7g6 c1f4 d8a5 e2e4 f6e4 d1e2 f7f5 f4d2 a5b6 c3e4 f5e4 d2c3 c6e5 e2e4 f8g7 f1e2 c8f5 e4d5 e5f3 d5f3 g7e5 c3e5 d6e5 f3c3 e8c8 c4c5 b6c6 e1g1 d8d4 f2f4 c6d5 f4e5 d5e5 e2f3 f5e4 c5c6 b7c6');
INSERT INTO evaluations VALUES(20833,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(20833,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(20833,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(20833,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(20833,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(20833,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(20833,36,'cp 24','e7e5 f4g3 h7h5 h2h4 e5d4 e3d4 c8g4 d1d2 f8d6 g3d6 d8d6 f1e2 e8c8 e1c1 g8h6 e2g4 h5g4 g1e2 h6f5 h4h5 h8e8 b2b3 c8b8 c1b2 a7a6 d2f4 d6f4 e2f4 c6d4 c3d5 c7c6 d5c3 g4g3 f2g3 f5g3 h1h3 g3e4 f4d3 e4c3 b2c3 d4f5 c3b2');
INSERT INTO evaluations VALUES(22257,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(22257,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(22257,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(22257,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(22257,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(22257,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(22257,36,'cp 24','f4g3 h7h5 h2h4 e5d4 e3d4 c8g4 g1e2 d8d7 d1d2 a7a6 e1c1 e8c8 f2f3 g4f5 e2f4 c6b4 a2a3 b4c2 f4d5 g8e7 d5e7 f8e7 d4d5 e7d6 g3d6 d7d6 f1d3 c2d4 d2e3 f5d3 e3d4 d3g6 d4a7 c7c6 h1e1 d6f4 a7e3 f4e3 e1e3 d8e8 e3e8 h8e8 d5c6 b7c6 c1d2');
INSERT INTO evaluations VALUES(21483,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(21483,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(21483,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(21483,33,'cp 86',NULL);
INSERT INTO evaluations VALUES(21483,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(21483,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(21483,36,'cp 76','e1g1 g8f6 d2d4 f8d6 f1e1 e8g8 c1g5 h7h6 g5f6 d8f6 c3e4 f6d8 d1e2 b7b6 a1d1 c6b4 e4d6 c7d6 c2c3 b4d5 e2e4 d8c7 c4d5 e6d5 e4f4 c8e6 e1e3 a8e8 d1e1 e6d7 e3e7 f7f6 e1e2 c7c6 a2a3 e8e7 e2e7 f8f7');
INSERT INTO evaluations VALUES(26212,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(26212,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(26212,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(26212,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(26212,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(26212,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(26212,36,'cp -38','e7e6 g1f3 f8d6 e1g1 d8c7 c1g5 b7b5 c2c3 h7h6 g5h4 g8f6 h4g3 e8g8 b1d2 a7a5 d1e2 f8c8 a2a3 b5b4 a3b4 d6g3 h2g3 a5b4 a1a8 c8a8 c3c4 d5c4 d2c4 a8c8 c4e3 c7a5 f1d1 f6d5 f3e5 d7f6 e5d3 d5e3 e2e3 c8c2 d1c1');
INSERT INTO evaluations VALUES(26158,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(26158,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(26158,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(26158,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(26158,34,'cp -34',NULL);
INSERT INTO evaluations VALUES(26158,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(26158,36,'cp -41','g1f3 f8d6 e1g1 g8f6 f1e1 d8c7 c2c3 e8g8 c1g5 b7b5 b1d2 a7a5 d1e2 h7h6 g5f6 d7f6 g2g3 c7c6 e1c1 f8e8 d2f1 a8c8 f1e3 g7g6 a2a3 h6h5 f3e5 d6e5 d4e5 f6e4 c1d1 e4c5 d1e1 c5b3 a1d1 g8g7');
INSERT INTO evaluations VALUES(26156,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(26156,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(26156,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(26156,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(26156,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(26156,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(26156,36,'cp -32','d8c7 e1g1 f8d6 c1g5 h7h6 g5h4 g8e7 h4g3 e7f5 g3d6 f5d6 c2c3 c7b6 d1c2 e8g8 f3e1 f8e8 b1d2 a8c8 d2b3 d6c4 e1d3 e6e5 d4e5 d7e5 d3b4 b6g6 c2g6 e5g6 b4d3 e8e2 a1e1 c8e8 e1e2 e8e2 b3d4');
INSERT INTO evaluations VALUES(28252,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(28252,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(28252,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(28252,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(28252,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(28252,35,'cp -74',NULL);
INSERT INTO evaluations VALUES(28252,36,'cp -51','b4c3 b2c3 b7e4 f3e5 e4b7 e5g4 d7d6 c1g5 h7h6 g4f6 g7f6 g5h4 b8d7 d3e2 h8g8 e2f3 d8c8 h2h3 g8g6 g1h2 b7f3 d1f3 c7c6 a2a4 e8f8 g2g4 d6d5 f3e2 c6c5 a4a5 f8g8 h4g3 c8c6 f2f4 c5d4 f4f5 e6f5 g4f5 g6g5 a5b6 d7b6 c3d4');
INSERT INTO evaluations VALUES(28292,30,'cp -79',NULL);
INSERT INTO evaluations VALUES(28292,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(28292,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(28292,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(28292,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(28292,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(28292,36,'cp -48','b2c3 b7e4 c1g5 e4d3 c2d3 h7h6 g5h4 b8c6 f3e5 c6e7 d1f3 e7f5 h4f6 g7f6 e5g4 a8c8 a1e1 e8f8 g2g3 f8g7 e1e4 h6h5 g4e3 f5e7 e4h4 f6f5 h4h5 h8h5 f3h5 e7g6 d4d5 d8f6 d3d4 c8h8 h5f3 f5f4 e3g4 f6g5 d5e6 d7e6 f1e1 c7c5 g4e5 c5d4 e5g6 g5g6 c3d4');
INSERT INTO evaluations VALUES(28289,30,'cp -67',NULL);
INSERT INTO evaluations VALUES(28289,31,'cp -57',NULL);
INSERT INTO evaluations VALUES(28289,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(28289,33,'cp -51',NULL);
INSERT INTO evaluations VALUES(28289,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(28289,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(28289,36,'cp -53','b7e4 c1g5 e4d3 c2d3 b8c6 f3e5 h7h6 g5h4 c6e7 d1f3 e7f5 h4f6 g7f6 e5g4 c7c6 a1e1 e8f8 e1e4 f8g7 g2g3 h6h5 g4e3 f5e7 e4h4 e7g6 h4h5 h8h5 f3h5 f6f5 e3g2 d8f6 h5f3 a8h8 h2h4 c6c5 d4c5 b6c5 c3c4 h8b8 g2f4 g6f4 f3f4 e6e5 f4e3 d7d6 e3f3 b8b2');
INSERT INTO evaluations VALUES(28655,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(28655,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(28655,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(28655,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(28655,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(28655,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(28655,36,'cp 16','d5c4 d1a4 b8d7 c1g5 f8e7 f3e5 b7g2 g1g2 h7h6 g5f6 e7f6 f1d1 d8c8 a4c6 a8b8 b1a3 c4c3 b2c3 f6e5 d4e5 c8b7 c6b7 b8b7 f2f4 e8e7 a3c2 h8c8 c2d4 c7c6 d4b3 b6b5 g2f3 b7b6 a1b1 c6c5 b3d2 c5c4 d2e4 f7f6 e5f6 d7f6 e4f6 g7f6 d1d4 a7a5 a2a3 f6f5 e2e4 c8d8 d4d8 e7d8 e4f5');
INSERT INTO evaluations VALUES(28737,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(28737,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(28737,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(28737,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(28737,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(28737,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(28737,36,'cp 0','d7d5 b1d2 f8d6 f1d3 c7c5 c2c3 e8g8 e1g1 d6f4 e3f4 b7a6 d3a6 b8a6 f1e1 a6b8 f3e5 d8c7 d2f3 b8c6 e5c6 c7c6 d1d3 f6e4 f3g5 e4f6');
INSERT INTO evaluations VALUES(28815,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(28815,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(28815,32,'cp -96',NULL);
INSERT INTO evaluations VALUES(28815,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(28815,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(28815,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(28815,36,'cp -82','f6e4 c3e4 b7e4 c1g5 d8c8 d4d5 c8b7 f1c4 b6b5 c4b3 c7c6 d5d6 e7e6 a2a4 e4f3 g2f3 c6c5 a4b5 b7b5 b3a4 b5b6 g5f4 b8c6 e1g1 c6d4 c2c3 d4f5 d1d2 g7g6 f1d1 f8g7 d2e2 a7a6 g1g2 b6b7 a4c2 e8g8 c2f5 g6f5 f4e5 f7f6');
INSERT INTO evaluations VALUES(28836,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(28836,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(28836,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(28836,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(28836,34,'cp -78',NULL);
INSERT INTO evaluations VALUES(28836,35,'cp -69',NULL);
INSERT INTO evaluations VALUES(28836,36,'cp -55','c3e4 b7e4 c1g5 d8c8 d4d5 c8b7 f1c4 b6b5 c4b3 c7c6 d5d6 e7e6 a2a4 b5a4 a1a4 e4f3 g2f3 c6c5 e1g1 b8c6 f1e1 c6d4 e1e4 f8d6 a4d4 c5d4 d1d4 d6e7 d4g7 e7g5 g7h8 e8e7 h8d4 g5f6 d4c5 d7d6 c5f5 a8g8 g1h1 b7c8 f5h7 g8g7 h7h3 f6b2 e4a4 c8c5 h3h4 b2f6 h4c4');
INSERT INTO evaluations VALUES(28835,30,'cp -63',NULL);
INSERT INTO evaluations VALUES(28835,31,'cp -79',NULL);
INSERT INTO evaluations VALUES(28835,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(28835,33,'cp -76',NULL);
INSERT INTO evaluations VALUES(28835,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(28835,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(28835,36,'cp -62','b7e4 c1g5 f7f6 g5f4 b8c6 c2c3 e7e6 f3d2 e4g6 h2h4 f8d6 f4d6 c7d6 h4h5 g6f7 d2c4 d6d5 c4d6 e8e7 d6f7 e7f7 f2f4 c6e7 f1d3 b6b5 f4f5 b5b4 f5e6 d7e6 e1g1 b4c3 b2c3 a8c8 d1d2 e7f5 d3f5 e6f5 f1f5 d8d7');
INSERT INTO evaluations VALUES(28607,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(28607,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(28607,32,'cp -68',NULL);
INSERT INTO evaluations VALUES(28607,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(28607,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(28607,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(28607,36,'cp -75','c1g5 f7f6 g5f4 b8c6 c2c3 e7e6 f3d2 e4g6 h2h4 f8d6 f4d6 c7d6 h4h5 g6f7 f1d3 e8g8 d1e2 d6d5 h5h6 g7g5 f2f4 f7g6 d3g6 h7g6 f4g5 f6g5 h6h7 g8h8 e1c1 a8c8 e2d3 d8f6 d1f1 f6g7 f1f8 c8f8 d3g3 e6e5 g3g5 e5d4 c3d4 g7d4 d2f3 d4f4 c1b1 f4f5 g5f5 f8f5 a2a3 d5d4');
INSERT INTO evaluations VALUES(31714,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(31714,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31714,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31714,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31714,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31714,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31714,36,'cp 0','g4f3 e2f3 b8c6 e1g1 g8f6 c2c3 f8d6 c3c4 e8g8 c4d5 f6d5 b1c3 d5c3 b2c3 c6a5 c1d2 a8c8 g2g3 d8c7 a1b1 b7b6 f1e1 f8d8 h3h4 h7h6 f3g2 a5c4 d2c1 d6a3 c1f4 a3d6');
INSERT INTO evaluations VALUES(33082,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(33082,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(33082,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(33082,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(33082,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(33082,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(33082,36,'cp 42','b8c6 a2a4 a7a6 b2b4 c5b6 d2d3 f6g6 b1d2 g8f6 f3h4 g6h5 d1h5 f6h5 g2g3 c8h3 f1e1 h5f6 a4a5 b6a7 d2f1 h3e6 c1e3 e8c8 e3a7 c6a7 f1e3 a7c6 h4f3 e6c4 e3c4 h8e8 f3g5 d6d5 c4b2 d8d7 f2f3 d5d4 c3c4 h7h5');
INSERT INTO evaluations VALUES(33878,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(33878,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(33878,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33878,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(33878,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33878,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(33878,36,'cp 0','e2e3 c7c5 g1f3 b8c6 d4c5 d5d1 e1d1 f8c5 a2a3 f7f6 b2b4 c5b6 f1d3 e8f7 d1e2 e6e5 c1b2 c8e6 f3d2 h8d8 h1c1 h7h5 d2c4 b6c7 c4d2');
INSERT INTO evaluations VALUES(35289,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(35289,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(35289,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(35289,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(35289,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(35289,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(35289,36,'cp 61','b1c3 e5f4 d1f3 b8c6 c3d5 e7d6 d2d3 c6d4 f3f2 d4e6 g1h3 c7c6 d5f4 g8f6 c2c3 d6c7 g2g3 f8e7 f1g2 d7d6 e1g1 e8g8 g1h1 a7a5 f2e2 e6f4 h3f4 c8g4 e2e1 f6h5 h2h3 h5f4 g3f4 g4d7 f4f5 g8h8 e1g3 a8e8 c1f4 a5a4 d3d4');
INSERT INTO evaluations VALUES(36435,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(36435,31,'cp 82',NULL);
INSERT INTO evaluations VALUES(36435,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(36435,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(36435,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(36435,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(36435,36,'cp 70','c7c6 a1b1 a7a5 a2a3 b8d7 d4d5 c6d5 c4d5 d7c5 c1e3 c8d7 e3c5 d6c5 d5d6 g8h8 f3e5 d7e8 g2b7 a8b8 b7g2 e7d6 e5c4 d6d5 c3d5 e8b5 c4e3 a5a4 d5f4 d8d1 f1d1');
INSERT INTO evaluations VALUES(35853,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(35853,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(35853,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(35853,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(35853,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(35853,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(35853,36,'cp 74','d1c2 b8a6 a1b1 a6c7 b2b3 d8e8 e2e3 a7a5 c1b2 c7a6 c2d2 a8b8 c3e2 b7b6 d2d1 b6b5 a2a3 a6c7 d1c2 c8d7 e2f4 c7e6 b2c3 e6f4 e3f4 f6e4 c3a5 b8a8 a5d2 a8a3 f3g5 e8a8 g5e4 f5e4 c2e4 a3a2');
INSERT INTO evaluations VALUES(38641,30,'cp -69',NULL);
INSERT INTO evaluations VALUES(38641,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(38641,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(38641,33,'cp -56',NULL);
INSERT INTO evaluations VALUES(38641,34,'cp -44',NULL);
INSERT INTO evaluations VALUES(38641,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(38641,36,'cp -50','b8d7 g2g3 g8f6 d2d3 e8g8 f1g2 a7a5 e1g1 a5a4 a2a3 a4b3 c2b3 f8e8 b1d2 h7h6 d1c2 c7c5 e3e4 d5d4 d2c4 d6c7 a3a4 d7b8 b2a3 b8a6 f1b1 d8e7 c2d2 c8d7 f3h4 d7e6 h4f5 e7d8 g2h3 g8h7 h3g2 b7b6 g2h3 f6d7 b3b4 c5b4 a3b4 d7f6 c4d6');
INSERT INTO evaluations VALUES(40493,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(40493,31,'cp 137',NULL);
INSERT INTO evaluations VALUES(40493,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(40493,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(40493,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(40493,35,'cp 149',NULL);
INSERT INTO evaluations VALUES(40493,36,'cp 150','e1g1 c6c5 a2a3 b4c3 b2c3 c8d7 a3a4 d7c6 c1a3 b8d7 d4c5 e6e5 f3h4 g7g6 f2f3 e8g8 e3e4 g8g7 f1e1 a8c8 d1c2 d8c7 h2h3 f6h5 g1h2 f8d8 a1d1 d7f8 c4b3 d8d1 e1d1 c8d8 d1d6 d8d6 c5d6 c7b6 c2d1 b6e3 a4a5 h5f4');
INSERT INTO evaluations VALUES(40838,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(40838,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(40838,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(40838,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(40838,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(40838,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(40838,36,'cp 58','e8g8 f1g2 f8e8 e1g1 e7f8 f1e1 c7c6 h2h3 a7a5 c1f4 a5a4 g3g4 h7h6 d1d2 b8d7 f4g3 d7e5 b2b3 a4b3 c2b3 d8a5 f2f4 e5g6 a2a3 a5d8 b3b4 h6h5 g4g5 h5h4 g3h2 f6h5 g2f3 c6c5 f3h5 c5d4 d2d4 c8e6');
INSERT INTO evaluations VALUES(36259,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(36259,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(36259,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(36259,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(36259,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(36259,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(36259,36,'cp 65','f1g2 f8e8 e1g1 e7f8 h2h3 c7c6 g3g4 h7h6 c1f4 b8a6 f1e1 a6c5 f4g3 c8e6 b2b4 c5d7 a2a4 a7a5 b4b5 a8c8 f2f4 d7c5 d1f3 e6d7 g1h1 c6b5 a4b5 d8c7 g3h4 c5e6 d4e2 e6c5 e2g3 a5a4 a1d1 d7e6');
INSERT INTO evaluations VALUES(36260,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(36260,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(36260,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(36260,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(36260,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(36260,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(36260,36,'cp 44','f8e8 e1g1 e7f8 h2h3 c7c6 f1e1 b8d7 a2a4 d7c5 b2b3 a7a5 d1d2 c8d7 c1b2 d8c7 e1e2 a8d8 a1e1 b7b6 e2e3 h7h6 g1h2 e8e7 f2f4 d8e8 e3e2 h6h5 e2e3 g7g6 d4f3 d7c8 e4e5 d6e5 f3e5 c8b7 b2a3 e8d8 d2f2 e7e8 e5c4 b7a6');
INSERT INTO evaluations VALUES(41592,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(41592,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(41592,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41592,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(41592,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(41592,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(41592,36,'cp 13','e1g1 e8g8 a2a4 c7c6 a4a5 f8e8 f1e1 c8e6 c4e6 e8e6 c1d2 b8a6 c3a4 e6e8 c2c4 b7b5 a5b6 a7b6 d2e3 f6d7 d1d2 a6c5 a4c5 b6c5 a1a8 d8a8 b2b4 e8b8 b4c5 d7c5 e3c5 d6c5 d2c2 a8b7 c2e2 h7h6 g2g3 b7b2 e2b2');
INSERT INTO evaluations VALUES(43244,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(43244,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(43244,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(43244,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(43244,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(43244,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43244,36,'cp 0','d7d5 c2c4 c7c5 d4c5 e8g8 b1c3 f6e4 c1d2 b8d7 c4d5 d7c5 f1c4 c8f5 e1g1 d8b6 a1c1 b6b2 c3e4 c5e4 c1b1 b2f6 b1b7 e4d6 b7c7 f8c8 c7c8 a8c8 c4b3 f5d3 f1e1 d6e4 d2b4 f6b2 d1d3 b2f2 g1h2 f2g3 h2g1');
INSERT INTO evaluations VALUES(41503,30,'cp -1',NULL);
INSERT INTO evaluations VALUES(41503,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41503,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41503,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(41503,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(41503,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41503,36,'cp 0','c2c4 c7c5 d4c5 e8g8 b1c3 f6e4 c1d2 b8d7 c4d5 d7c5 f1c4 c8f5 e1g1 d8b6 a1c1 b6b2 c3e4 c5e4 c1b1 b2f6 b1b7 f8c8 c4d3 f6d6 f3d4 g7d4 e3d4 e4d2 d1d2 d6d5 d3f5 d5b7 f5c8 a8c8 d4d5 b7c7 d2e3 c7c5 f1e1 c5d5 e3a7 c8a8 a7e7 d5a2 e7f6 a8f8 e1d1 f8b8 d1d8 b8d8 f6d8 g8g7 d8e8 a2b1 g1h2');
INSERT INTO evaluations VALUES(43430,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(43430,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(43430,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(43430,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(43430,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(43430,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(43430,36,'cp 22','e2e3 b8c6 f1e2 c8f5 g1f3 e7e6 e1g1 f8e7 a1c1 e8g8 c3a4 f6e4 h2h3 a8c8 f3d2 e7g5 d2e4 g5f4 e4c5 f4b8 e2d3 f5d3 c5d3 d8e7 a4c5 b8d6 d1b3 c8c7 b3a4 c6a7 c1c2 f8c8 a4b3 a7c6 f1c1 c6a7 b3b6 a7b5 a2a4');
INSERT INTO evaluations VALUES(43428,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(43428,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(43428,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(43428,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(43428,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(43428,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43428,36,'cp 16','b8c6 f1d3 c8g4 g1e2 e7e6 e1g1 f8d6 f4d6 d8d6 c3a4 e8g8 a1c1 a8c8 a4c5 c8c7 d1a4 g4e2 d3e2 f8c8 c1c2 f6d7 f1c1 d7c5 c2c5 g7g6 a4c2 d6d8 a2a3 h7h5 e2d1 d8e7 c2b3 g8g7 h2h3 c7d7 c1c3 e6e5 d4e5 e7e5');
INSERT INTO evaluations VALUES(11139,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(11139,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(11139,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(11139,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(11139,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(11139,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(11139,36,'cp 28','f1d3 c8g4 g1e2 e7e6 h2h3 g4h5 a1c1 a8c8 e1g1 c6b4 d3b1 f8e7 d1d2 b4c6 f4g5 h5g6 e2f4 g6b1 c1b1 f6d7 g5e7 d8e7 f1c1 d7b6 d2d1 e8g8 c3a4 b6c4 d1e2 c8c7 c1c3 h7h6 b1c1 f8c8 a2a3 c4d6 f4d3');
INSERT INTO evaluations VALUES(11142,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(11142,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(11142,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(11142,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(11142,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(11142,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(11142,36,'cp 24','c8g4 g1e2 e7e6 e1g1 f8d6 f4d6 d8d6 c3a4 e8g8 h2h3 g4h5 a1c1 c6b4 d3b1 b4c6 d1d2 h5g6 b1d3 g6d3 d2d3 h7h6 a4c5 d6e7 d3b3 a8b8 c1c2 f8c8 f1c1 f6e8 b3a4 e8d6 c5d3 d6c4 b2b3 c4d6 c2c5 c8c7');
INSERT INTO evaluations VALUES(45238,30,'cp -161',NULL);
INSERT INTO evaluations VALUES(45238,31,'cp -177',NULL);
INSERT INTO evaluations VALUES(45238,32,'cp -194',NULL);
INSERT INTO evaluations VALUES(45238,33,'cp -162',NULL);
INSERT INTO evaluations VALUES(45238,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(45238,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(45238,36,'cp -147','f6e4 g5e3 g7g6 f1e2 f8g7 e2f3 d8a5 c2c3 e4f6 e1g1 b8d7 c3c4 d7e5 f3e2 c8g4 f2f3 g4d7 b1c3 e8g8 a1c1 a8c8 b2b3 e5c6 d1d2 a7a6 f1d1 f8d8 d4c6 d7c6 c3d5 a5d2 d5f6 g7f6 d1d2 e7e6 g2g3 f6e7 a2a4 f7f5 e3b6 d8d7 f3f4 g8f7');
INSERT INTO evaluations VALUES(45282,30,'cp -133',NULL);
INSERT INTO evaluations VALUES(45282,31,'cp -141',NULL);
INSERT INTO evaluations VALUES(45282,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(45282,33,'cp -151',NULL);
INSERT INTO evaluations VALUES(45282,34,'cp -150',NULL);
INSERT INTO evaluations VALUES(45282,35,'cp -138',NULL);
INSERT INTO evaluations VALUES(45282,36,'cp -154','g5e3 g7g6 f1e2 f8g7 e2f3 d8a5 c2c3 e4f6 e1g1 b8d7 c3c4 d7e5 f3e2 e8g8 b1c3 c8g4 f2f3 g4d7 a1c1 f8c8 b2b3 d6d5 c3d5 f6d5 c4d5 c8c1 d1c1 a8c8 c1b2 a5d5 b2d2 d5d6 f1d1 a7a6 h2h3 d6b8 f3f4 e5c6 e2f3 c6d4 e3d4 g7d4 d2d4 d7c6 f3c6');
INSERT INTO evaluations VALUES(45667,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(45667,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(45667,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(45667,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(45667,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(45667,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(45667,36,'cp 30','g1f3 f8d6 f4g3 c6c5 d4c5 d6c5 b1d2 e8g8 a1c1 b8d7 c2c4 c5e7 c4d5 d7c5 d3c4 e6d5 c4e2 c8d7 e1g1 a8c8 g3e5 h7h6 h2h3 f8e8 d2b1 d7a4 b2b3 a4c6 f3d4 e7f8 d4c6 b7c6 e5b2 c5e6 b1c3');
INSERT INTO evaluations VALUES(45882,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(45882,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(45882,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(45882,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(45882,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(45882,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(45882,36,'cp 33','f8e7 e1g1 e8g8 a2a4 d6d5 e4d5 c6d5 c4b3 b8c6 h2h3 h7h6 d1d3 a7a6 c1e3 f8e8 d4e2 c6b4 d3d2 c8f5 f1d1 a8c8 e2f4 g7g5 c3d5 f6d5 f4d5 f5c2 d1c1 b4d5 d2d5 d8d5 b3d5 e7f6 d5b7 f6b2 b7c8 b2c1');
INSERT INTO evaluations VALUES(46138,30,'cp -56',NULL);
INSERT INTO evaluations VALUES(46138,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(46138,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(46138,33,'cp -65',NULL);
INSERT INTO evaluations VALUES(46138,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(46138,35,'cp -46',NULL);
INSERT INTO evaluations VALUES(46138,36,'cp -54','e6d5 e2e3 d5d4 c3e4 f6e4 g2e4 f8e7 g1e2 b8c6 e3d4 c5d4 e2f4 e7d6 d2d3 c8f5 e1g1 f5e4 d3e4 d6f4 c1f4 e8g8 a1c1 a7a5 f1e1 a5a4 c1c5 d8e7 c5b5 f8d8 f4g5 f7f6 g5f4 h7h6 b5d5 e7b4 d1d2 b4d2 f4d2 d8d5');
INSERT INTO evaluations VALUES(46371,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46371,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46371,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46371,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46371,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46371,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46371,36,'cp -14','d2d4 b8c6 g1f3 f8d6 e1g1 e8g8 c1g5 h7h6 g5f6 d8f6 e2e4 d5e4 c3e4 f6e7 a2a3 f8d8 b2b4 c8d7 d1d3 d7e8 a1c1 d6c7 e4c5 c7b6 d3d2 a8c8 f1e1 b6c5 c1c5 b7b6 c5c4 e7d6 e1c1 c6e7 c4c8 e7c8 d2d3');
INSERT INTO evaluations VALUES(47172,30,'cp -80',NULL);
INSERT INTO evaluations VALUES(47172,31,'cp -88',NULL);
INSERT INTO evaluations VALUES(47172,32,'cp -72',NULL);
INSERT INTO evaluations VALUES(47172,33,'cp -93',NULL);
INSERT INTO evaluations VALUES(47172,34,'cp -96',NULL);
INSERT INTO evaluations VALUES(47172,35,'cp -59',NULL);
INSERT INTO evaluations VALUES(47172,36,'cp -69','e7e6 a2a3 f8e7 e3f4 c7c5 e2e3 b8c6 d4c5 e7c5 g1f3 e8g8 f1e2 a7a6 f3e5 c6e5 f4e5 c8d7 e1g1 d7c6 e5d4 c5b6 d1d2 f8e8 f2f4 b6d4 e3d4 d8b6 g1h1 f6e4 c3e4 d5e4 c2c3 a8d8 a3a4 e8e7 d2e3 f7f5 f1b1 h7h6 h2h3');
INSERT INTO evaluations VALUES(47191,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(47191,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(47191,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(47191,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(47191,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(47191,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(47191,36,'cp -18','g1f3 e7e6 g2g3 f8e7 f1g2 e8g8 e1g1 b8d7 a2a4 b7b6 c2c4 c8a6 c4d5 f6d5 e3d2 c7c5 b1c3 a6b7 a4a5 d5b4 d4c5 d7c5 a5b6 a7b6 a1a8 d8a8 d1a1 e7f6 c3b5 b7f3 g2f3 a8a1 f1a1 c5b3 a1a4 b3d2 a4b4 d2f3 e2f3 f6e7 b4c4 f8d8 b5c3 d8d2 c4c7 e7d8 c3e4 d2b2');
INSERT INTO evaluations VALUES(47343,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(47343,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(47343,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(47343,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(47343,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(47343,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(47343,36,'cp 34','f1c4 e7e6 e1g1 c7c5 d2d4 a7a6 d4c5 d8d1 f1d1 f8c5 c4e2 e8g8 c1d2 b8d7 d2e1 b7b5 a2a4 b5a4 b1c3 a4a3 b2a3 c5e7 a3a4 d7c5 a4a5 c8b7 f3e5 f8d8 d1b1 a8b8 f2f3 f6d5 e5c4 b7c6 c3d5 e6d5 b1b8 d8b8 c4b6 c5b3');
INSERT INTO evaluations VALUES(53254,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(53254,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(53254,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(53254,33,'cp -74',NULL);
INSERT INTO evaluations VALUES(53254,34,'cp -73',NULL);
INSERT INTO evaluations VALUES(53254,35,'cp -83',NULL);
INSERT INTO evaluations VALUES(53254,36,'cp -78','c5d4 e3d4 b8c6 g1f3 c8f5 f1d3 e7e6 e1g1 g8f6 c2c3 f8d6 d3f5 e6f5 f1e1 d6e7 c1e3 e8g8 b1d2 f6e4 a1c1 a8c8 f3e5 c6a5 d2e4 f5e4 d1g4 a5c4 e5c4 f7f5 g4e2 c8c4 b2b3 c4c6 c3c4 e7a3 c1c2 d8d7 c4d5 c6c2 e2c2');
INSERT INTO evaluations VALUES(53419,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(53419,31,'cp -35',NULL);
INSERT INTO evaluations VALUES(53419,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(53419,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(53419,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(53419,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(53419,36,'cp -39','c8g4 e1g1 b8c6 h2h3 g4h5 d3e2 e7e6 f1e1 f8d6 d2d4 g8e7 c2c3 e8g8 a2a4 d8c7 b1d2 c6a5 d2b3 a5b3 d1b3 e7c6 b3d1 a8e8 b2b4 e8c8 c1d2 h5g6 e2f1 a7a6 b4b5 a6b5 a4b5 c6a5 d1a4 a5c4 f1c4 c7c4 a4c4 c8c4 f3e5');
INSERT INTO evaluations VALUES(31360,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(31360,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(31360,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(31360,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(31360,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(31360,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(31360,36,'cp -21','e1g1 b8c6 d3e2 e7e6 d2d4 f8d6 h2h3 g4h5 c2c3 d8c7 f1e1 g8e7 b1d2 e8g8 a2a4 c6a5 d2b3 a5b3 d1b3 e7c6 b3d1 h7h6 f3h4 h5e2 d1e2 a8e8 e2g4 g8h7 h4f3 c6a5 h3h4 f7f5 g4h5 h7g8 h5g6 c7f7 g6f7 g8f7 e1e2 a5b3');
INSERT INTO evaluations VALUES(54726,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(54726,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54726,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54726,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54726,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54726,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54726,36,'cp -1','c7c5 g1f3 b8c6 f1b5 c8g4 b5c6 b7c6 e1g1 g8h6 d2d3 h6f5 d1e1 g4f3 f1f3 f6e5 b2b3 g7g6 c1b2 f8g7 b2e5 g7e5 e1e5 e8g8 c2c3 d8d6 b1d2 d6e5 f4e5 f5g7 a1f1 f8f3 d2f3 g7e6 g1f2 a7a5 d3d4 c5d4 f3d4 e6d4 c3d4 a8b8 f1c1 a5a4 c1c6 a4b3 a2b3 b8b3 c6c7 g8f7 h2h3 b3b2 f2g3');
INSERT INTO evaluations VALUES(54945,30,'cp 180',NULL);
INSERT INTO evaluations VALUES(54945,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(54945,32,'cp 161',NULL);
INSERT INTO evaluations VALUES(54945,33,'cp 182',NULL);
INSERT INTO evaluations VALUES(54945,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(54945,35,'cp 176',NULL);
INSERT INTO evaluations VALUES(54945,36,'cp 151','e4d5 d8d5 b1c3 d5d6 c3b5 d6d8 c1f4 b8a6 g1f3 c7c6 b5c3 a6c7 f1d3 g7g6 f4e5 g8f6 d1e2 c7e6 d4d5 c6d5 e5f6 e7f6 d3b5 c8d7 e1c1 d7c6 f3d4 c6b5 e2b5 d8d7 c3d5 e8c8 b5d7 c8d7 d4e6 f7e6 d5f6 d7e7 f6e4 f8h6 c1b1 d8d1 h1d1 h8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(55046,30,'cp -134',NULL);
INSERT INTO evaluations VALUES(55046,31,'cp -127',NULL);
INSERT INTO evaluations VALUES(55046,32,'cp -118',NULL);
INSERT INTO evaluations VALUES(55046,33,'cp -114',NULL);
INSERT INTO evaluations VALUES(55046,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(55046,35,'cp -120',NULL);
INSERT INTO evaluations VALUES(55046,36,'cp -115','e7e5 d4b3 g8f6 e2e3 b8c6 f1e2 a7a5 d2d3 a5a4 b3d2 f8e7 e1g1 e8g8 a2a3 c8e6 b1c3 d8d7 d2f3 f8d8 d3d4 e7d6 c3b5 e5e4 f3d2 d6f8 f2f3 e4f3 d2f3 d8e8 e2d3 c6e7 f3e5 d7c8 c2c4 d5c4 e5c4 e7g6 b5c3 c8d7 h2h3 f6d5 d1f3 d5c3 b2c3 g6h4 f3f2');
INSERT INTO evaluations VALUES(56358,30,'cp -119',NULL);
INSERT INTO evaluations VALUES(56358,31,'cp -117',NULL);
INSERT INTO evaluations VALUES(56358,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(56358,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(56358,34,'cp -123',NULL);
INSERT INTO evaluations VALUES(56358,35,'cp -127',NULL);
INSERT INTO evaluations VALUES(56358,36,'cp -112','d4b3 g8f6 e2e3 b8c6 f1e2 a7a5 d2d3 a5a4 b3d2 f8e7 e1g1 e8g8 b1c3 f8e8 b2b3 a4b3 d2b3 c6a5 c1b2 e7d6 c3b1 c7c5 a2a4 a5b3 c2b3 c8e6 b1d2 d6c7 e2f3 h7h6 e3e4 d5d4 d2c4 g8h8');
INSERT INTO evaluations VALUES(55517,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55517,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(55517,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(55517,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(55517,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(55517,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(55517,36,'cp 40','d6e5 d1d8 e8d8 f2f4 b8c6 f4e5 c6e5 c1f4 e5g6 e1c1 c8d7 f4g3 g8f6 g1f3 f8b4 f3g5 d8e7 e2e4 b4d6 e4e5 d6e5 d1e1 f6g4 c3d5 e7f8 f1c4 f7f6 d5c7 a8c8 g5e6 f8e7 c4b3 e5g3 h2g3 g4e5 e1e3 h7h5 e3c3 d7e6 c7e6 c8c3 b2c3 h5h4 e6g7');
INSERT INTO evaluations VALUES(56630,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(56630,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(56630,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(56630,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(56630,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(56630,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(56630,36,'cp 34','d1d8 e8d8 f2f4 b8c6 f4e5 c6e5 c1f4 e5g6 f4g3 c7c6 e1c1 d8e8 e2e4 c8g4 f1e2 g8f6 e4e5 g4e2 g1e2 f6g4 d1e1 g4e5 e2f4 f7f6 f4g6 h7g6 g3e5 f6e5 e1e5 e8f7 h1f1 f7g8 h2h3 h8h5 e5e6 h5f5 f1d1 f5f6 e6e2 f6d6 d1e1 d6d7 c3e4 f8e7 a2a4 a8d8');
INSERT INTO evaluations VALUES(35409,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(35409,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(35409,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(35409,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(35409,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(35409,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(35409,36,'cp 39','e8d8 c1d2 b8d7 g1f3 g8f6 e1c1 f8d6 d2g5 d8e8 e2e4 h7h6 g5h4 e8e7 f3d2 c7c6 d2c4 d6c7 c4e3 g7g5 h4g3 d7b6 f2f3 c8e6 g3f2 h6h5 e3f5 e6f5 e4f5 h5h4 f1d3 b6d5 h1e1 d5c3 f2c5 e7e8 b2c3 f6d7 c5b4 d7f6');
INSERT INTO evaluations VALUES(35462,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(35462,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(35462,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(35462,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(35462,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(35462,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(35462,36,'cp 37','c1d2 b8d7 g1f3 g8f6 e1c1 f8d6 c3b5 f6e4 d2e1 d8e7 f3h4 d7f6 f2f3 a7a6 b5d6 e4d6 g2g3 c8e6 e2e4 f6d7 h4g2 a6a5 g2e3 h7h5 h2h4 f7f6 a2a4 b7b6 e1f2 c7c6 f1e2 d7c5 b2b3 h8d8 d1d2 d6b7 e3f1 e6f7 c1b2 d8d2 f1d2 a8d8 f2e3 g7g6');
INSERT INTO evaluations VALUES(56368,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(56368,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(56368,32,'cp -48',NULL);
INSERT INTO evaluations VALUES(56368,33,'cp -69',NULL);
INSERT INTO evaluations VALUES(56368,34,'cp -63',NULL);
INSERT INTO evaluations VALUES(56368,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(56368,36,'cp -60','e5e4 f3d2 f7f5 c2c4 c7c6 d1a4 b8d7 b1c3 g8f6 c4d5 d7b6 a4a5 f6d5 c3d5 d8d5 a5d5 b6d5 d2c4 f8e7 c1d2 c8e6 a1c1 e8g8 c4a5 a8b8 f1c4 f8c8 c4b3 c8c7 f2f3 b8c8 e1g1 g8h8 f3e4 f5e4');
INSERT INTO evaluations VALUES(56625,30,'cp 216',NULL);
INSERT INTO evaluations VALUES(56625,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(56625,32,'cp 235',NULL);
INSERT INTO evaluations VALUES(56625,33,'cp 233',NULL);
INSERT INTO evaluations VALUES(56625,34,'cp 216',NULL);
INSERT INTO evaluations VALUES(56625,35,'cp 230',NULL);
INSERT INTO evaluations VALUES(56625,36,'cp 225','d1d8 e8d8 f4e5 b8c6 e5g3 c6d4 b1a3 c8d7 e1c1 d4f5 a3c4 f5g3 h2g3 d8e8 e2e3 g8f6 c4e5 d7e6 f1c4 e6c4 e5c4 f6e4 g1h3 f8e7 c4d2 e4c5 h3f4 a8d8 g3g4 h7h6 d2c4 d8d1 h1d1 g7g6 f2f3 e7g5 d1d5 b7b6 c1d2 g5f4 e3f4');
INSERT INTO evaluations VALUES(35408,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(35408,31,'cp 216',NULL);
INSERT INTO evaluations VALUES(35408,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(35408,33,'cp 192',NULL);
INSERT INTO evaluations VALUES(35408,34,'cp 194',NULL);
INSERT INTO evaluations VALUES(35408,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(35408,36,'cp 190','e8d8 f4e5 c8f5 c2c3 b8d7 e5g3 g8f6 f2f3 f5e6 e2e4 f6h5 g3f2 f8c5 b1d2 c7c6 f2d4 h8e8 g1e2 c5d4 e2d4 d8c7 a2a4 g7g5 g2g3 a7a5 f1d3 d7c5 d3c2 f7f5 b2b4 a5b4 c3b4 a8d8 d4c6 b7c6 b4c5 f5e4 d2e4 g5g4 e1f2 g4f3');
INSERT INTO evaluations VALUES(35461,30,'cp 211',NULL);
INSERT INTO evaluations VALUES(35461,31,'cp 217',NULL);
INSERT INTO evaluations VALUES(35461,32,'cp 211',NULL);
INSERT INTO evaluations VALUES(35461,33,'cp 205',NULL);
INSERT INTO evaluations VALUES(35461,34,'cp 208',NULL);
INSERT INTO evaluations VALUES(35461,35,'cp 226',NULL);
INSERT INTO evaluations VALUES(35461,36,'cp 204','f4e5 c8f5 c2c3 b8d7 e5f4 f5g6 g1f3 g8f6 h2h4 a7a5 b1d2 a5a4 a2a3 f8c5 e2e3 h8e8 f3d4 f6d5 f4h2 g6h5 f1c4 d7b6 c4b5 c7c6 b5e2 h5e2 e1e2 d8c8 g2g4 g7g6 h4h5 a8a5 h5g6 h7g6 h2g3 d5f6 g4g5 f6h5 g3h2');
INSERT INTO evaluations VALUES(56715,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(56715,31,'cp -82',NULL);
INSERT INTO evaluations VALUES(56715,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(56715,33,'cp -88',NULL);
INSERT INTO evaluations VALUES(56715,34,'cp -86',NULL);
INSERT INTO evaluations VALUES(56715,35,'cp -64',NULL);
INSERT INTO evaluations VALUES(56715,36,'cp -90','d6e5 d1h5 e8d7 h5e5 b8c6 e5g3 d7e8 b1c3 f8d6 g3g7 d6e5 g7g5 c8e6 f1d3 d8g5 c1g5 g8e7 g2g3 c6d4 e1c1 d4f3 g5h6 e5c3 b2c3 h8g8 d3b5 e7c6 h6f4 a7a6 b5c6 b7c6 d1d3 f3g5 c3c4 e8f7 f4g5 g8g5 f2f4 g5h5 f4f5 e6c4 d3d7 f7f6 d7c7');
INSERT INTO evaluations VALUES(56846,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(56846,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(56846,32,'cp -84',NULL);
INSERT INTO evaluations VALUES(56846,33,'cp -96',NULL);
INSERT INTO evaluations VALUES(56846,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(56846,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(56846,36,'cp -64','d1h5 e8d7 h5e5 b8c6 e5g3 d7e8 b1c3 f8d6 g3g7 d6e5 g7g5 c8e6 f1e2 d8g5 c1g5 g8f6 e1c1 h7h5 g5f6 e5f6 c3d5 e6d5 d1d5 h5h4 f2f4 f6d4 c1d2 d4b6 c2c3 e8e7 e2g4 h8g8 g4h3 a8d8 d5d8 c6d8 e4e5 c7c6 h1f1 d8e6 d2e2 b6c7 e2f3 g8f8 f3e3 c7b6 e3e4');
INSERT INTO evaluations VALUES(56845,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(56845,31,'cp -58',NULL);
INSERT INTO evaluations VALUES(56845,32,'cp -87',NULL);
INSERT INTO evaluations VALUES(56845,33,'cp -70',NULL);
INSERT INTO evaluations VALUES(56845,34,'cp -84',NULL);
INSERT INTO evaluations VALUES(56845,35,'cp -85',NULL);
INSERT INTO evaluations VALUES(56845,36,'cp -66','e8d7 h5e5 b8c6 e5g3 d7e8 b1c3 f8d6 g3g7 d6e5 g7g5 c8e6 f1e2 d8g5 c1g5 g8f6 e1c1 h7h5 g5f6 e5f6 c3d5 f6e5 g2g3 c6d4 e2c4 e6d5 e4d5 e8d7 h1e1 e5g7 c2c3 d4f5 c4b5 d7d8 e1e6 f5d6 b5d3 d8d7 d1e1 a8g8 f2f4');
INSERT INTO evaluations VALUES(3723,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(3723,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(3723,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(3723,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(3723,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(3723,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(3723,36,'cp 32','e4d5 f6d5 b1d2 c8e6 d2e4 c5b6 c2c3 f7f6 b2b4 g8h8 a2a4 a7a5 b4b5 c6e7 d1b3 e6g8 d3d4 f6f5 e4c5 b6c5 d4c5 e7g6 c1g5 d8d7 a1d1 c7c6 f3h4 g6h4 g5h4 e5e4 d1d4 d7c7 f1d1 a8e8 c4f1 h7h6 c3c4 d5f6 b3c3 f5f4');
INSERT INTO evaluations VALUES(3715,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(3715,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(3715,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(3715,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(3715,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(3715,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(3715,36,'cp 11','f6d5 b1d2 c5b6 c2c3 g8h8 f1e1 f7f6 d3d4 e5d4 c3d4 c8f5 a2a3 d5e7 d2b3 d8d6 c1d2 a8d8 a3a4 f5g6 a4a5 b6d4 b3d4 c6d4 f3d4 d6d4 d1c1 d4d2 c1d2 d8d2 e1e7 d2b2 e7c7 f8e8 c4f1 a7a6 a1c1 b2a2 c7b7 a2a5');
INSERT INTO evaluations VALUES(6311,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(6311,31,'cp 97',NULL);
INSERT INTO evaluations VALUES(6311,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(6311,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(6311,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(6311,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(6311,36,'cp 93','c6e7 b1c3 d7d6 f1b5 c7c6 d5c6 b7c6 b5a4 g8f6 c1g5 e8g8 e1g1 a8b8 f3d4 c8d7 d1d2 f6g4 a4b3 g4e5 a1d1 g8h8 c3a4 b6c7 f2f4 e5g4 h2h3 c6c5 h3g4 c5d4 f4f5 f7f6 g5f4 d7b5 f1f3 b5a4 b3a4 d6d5 f4c7 d8c7 d2d4 f8d8 d4f2');
INSERT INTO evaluations VALUES(7005,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(7005,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(7005,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(7005,33,'cp 101',NULL);
INSERT INTO evaluations VALUES(7005,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(7005,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(7005,36,'cp 92','d5d6 c7d6 b1c3 g8f6 c1g5 d6d5 e4d5 d7d6 g5f6 g7f6 f1b5 e8f8 e1g1 e7g6 d1d2 f8g7 c3e2 h7h5 a2a4 a7a5 h2h3 c8d7 f3d4 h5h4 e2c3 h8h5 b5d7 d8d7 g1h2 g6e7 a1e1 h5e5 d4f3 e5e1 f1e1 e7g6 e1e4 f6f5 e4h4 g6h4 f3h4 b6d8 d2f4 d8h4 f4h4');
INSERT INTO evaluations VALUES(7194,30,'cp 171',NULL);
INSERT INTO evaluations VALUES(7194,31,'cp 190',NULL);
INSERT INTO evaluations VALUES(7194,32,'cp 205',NULL);
INSERT INTO evaluations VALUES(7194,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(7194,34,'cp 198',NULL);
INSERT INTO evaluations VALUES(7194,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(7194,36,'cp 168','e5f3 d8e7 d2d3 c8d7 d1e2 e8c8 c1e3 c5b4 e3d2 h8e8 e1c1 c6c5 a2a3 b4a5 c3b1 a5b6 h2h3 c5c4 e4e5 d7c6 d3c4 f6e4 d2e3 f7f6 e3b6 c7b6 h1e1 d8d1 e1d1 f6e5 b1d2 e4d6 c1b1 e5e4 f3d4');
INSERT INTO evaluations VALUES(10351,30,'cp -32',NULL);
INSERT INTO evaluations VALUES(10351,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(10351,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(10351,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(10351,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(10351,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(10351,36,'cp -27','b4c3 b2c3 f6e4 f3e5 d8f6 e5g4 f6e7 g4e3 e4c3 d1h5 c6d4 f1d3 d7d5 c1d2 e7c5 e1g1 c8e6 d2c3 c5c3 e3d5 e6d5 h5d5 e8g8 d5b7 a7a5 a1b1 g7g6 b7b2 c3c5 b1e1 f8e8 b2c1 a8d8 h2h4 g8g7 e1e8 d8e8 f1e1 e8e5 e1e5 c5e5 d3f1 h7h6 f1d3 g6g5 h4g5 h6g5 c1d2 d4e6');
INSERT INTO evaluations VALUES(10455,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(10455,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(10455,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(10455,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(10455,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(10455,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(10455,36,'cp -20','b2c3 f6e4 f3e5 d7d5 f1b5 e8g8 b5c6 b7c6 e1g1 f8e8 d1f3 f7f6 e5c6 d8d7 c6a5 c8a6 f1e1 e8e6 a5b3 a8e8 c1e3 d7f7 b3c5 e4c5 d4c5 a6c4 h2h3 f7g6 e3d2 e6e5 f3g4 g6c2 e1e5 e8e5 d2e3 c2c3');
INSERT INTO evaluations VALUES(15225,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(15225,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(15225,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(15225,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(15225,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(15225,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(15225,36,'cp 26','f1b5 f8c5 d2d3 e8g8 b5c6 b7c6 c3a4 c5d6 d1d2 f6d7 e1g1 d7b6 a4c3 f7f5 e4f5 c8f5 d2e2 b6d5 c3d5 c6d5 f3e5 d8e8 c1f4 e8b5 c2c3 a8e8 a2a4 b5a6 f4g3 c7c5 e2h5 g7g6 h5e2 d6e5 g3e5 a6d3 a1e1 d3e2 e1e2 f5d3 f1f8 e8f8');
INSERT INTO evaluations VALUES(15392,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15392,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15392,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15392,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15392,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15392,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15392,36,'cp 0','b2c3 h7h6 f3d2 f8d6 d2e4 c6a5 c4d5 d6e7 d1f3 e8g8 e4g3 g8h8 d5e4 f7f5 g3f5 c8f5 e4f5 g7g6 f3h3 g6f5 h3h6 h8g8 h6g6 g8h8 g6h6');
INSERT INTO evaluations VALUES(15545,30,'cp 200',NULL);
INSERT INTO evaluations VALUES(15545,31,'cp 188',NULL);
INSERT INTO evaluations VALUES(15545,32,'cp 198',NULL);
INSERT INTO evaluations VALUES(15545,33,'cp 180',NULL);
INSERT INTO evaluations VALUES(15545,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(15545,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(15545,36,'cp 190','d4e5 f6h5 f4g5 d8d7 e5d6 h7h6 g5h4 d7d6 b1c3 d6d1 a1d1 f8b4 f1e2 g7g5 f3d4 h5f4 e3f4 g5h4 e1g1 c6d4 d1d4 b4c3 b2c3 c8e6 e2f3 a8d8 d4a4 d8d2 a4a7 e8g8 a7b7 d2c2 b7c7 c2a2 c7c5 f8c8 c5h5 c8c3 h5h6 c3f3 g2f3 h4h3');
INSERT INTO evaluations VALUES(15460,30,'cp 174',NULL);
INSERT INTO evaluations VALUES(15460,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(15460,32,'cp 170',NULL);
INSERT INTO evaluations VALUES(15460,33,'cp 188',NULL);
INSERT INTO evaluations VALUES(15460,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(15460,35,'cp 190',NULL);
INSERT INTO evaluations VALUES(15460,36,'cp 191','f6h5 f4g5 d8d7 e5d6 h7h6 g5h4 d7d6 b1c3 d6d1 a1d1 f8b4 f1e2 g7g5 h4g3 h5g3 h2g3 f7f6 e1d2 c8e6 a2a3 e8c8 e2d3 b4c5 d2c1 h6h5 d3g6 d8d1 c1d1 e6g4 d1c1 c5d6 g6e4 g4d7 e4d3 c6e5 f3e5 d6e5 c3e4 d7c6 f2f4');
INSERT INTO evaluations VALUES(16245,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(16245,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(16245,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(16245,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(16245,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(16245,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(16245,36,'cp 19','a7a6 b5c6 b7c6 d2d4 f8b4 e4e5 f6d5 e1g1 d5c3 b2c3 b4e7 f1e1 e8g8 e1e4 f7f5 e5f6 e7f6 d1d3 d7d6 c1f4 c8d7 a1e1 c6c5 e4e3 c5d4 c3d4 d8e8 f4g3 e8h5 h2h4 h5f5 d3c4 f6d8 c4e2 a6a5 c2c4 d8f6 c4c5');
INSERT INTO evaluations VALUES(16628,30,'cp -140',NULL);
INSERT INTO evaluations VALUES(16628,31,'cp -135',NULL);
INSERT INTO evaluations VALUES(16628,32,'cp -139',NULL);
INSERT INTO evaluations VALUES(16628,33,'cp -140',NULL);
INSERT INTO evaluations VALUES(16628,34,'cp -136',NULL);
INSERT INTO evaluations VALUES(16628,35,'cp -137',NULL);
INSERT INTO evaluations VALUES(16628,36,'cp -142','f8e7 e1g1 f6e4 g5e7 c6e7 f1e1 e4f6 d1d4 e8g8 c4d3 d7d5 b2b4 d8d6 a2a3 a7a5 b4b5 c8f5 c2c4 f5d3 d4d3 e7g6 c4d5 d6d5 d3d5 f6d5 g2g3 f8e8 b1d2 g8f8 e1c1 g6e5 f3e5 e8e5 d2f3 e5e8 f3d4 a8d8 c1c5 a5a4');
INSERT INTO evaluations VALUES(10928,30,'cp -131',NULL);
INSERT INTO evaluations VALUES(10928,31,'cp -140',NULL);
INSERT INTO evaluations VALUES(10928,32,'cp -129',NULL);
INSERT INTO evaluations VALUES(10928,33,'cp -136',NULL);
INSERT INTO evaluations VALUES(10928,34,'cp -126',NULL);
INSERT INTO evaluations VALUES(10928,35,'cp -142',NULL);
INSERT INTO evaluations VALUES(10928,36,'cp -129','e1g1 f6e4 g5e7 c6e7 f1e1 e4f6 d1d4 e8g8 c4d3 d7d5 b2b4 d8d6 a2a3 c8f5 b1d2 f5d3 c2d3 f8e8 d2b3 b7b6 d4e5 d6d8 e5c3 a8c8 a3a4 h7h6 f3e5 d8d6 a1c1 f6d7 e5d7 d6d7 a4a5 c7c6 h2h3 g8f8 b3d4');
INSERT INTO evaluations VALUES(19295,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(19295,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19295,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(19295,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(19295,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(19295,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19295,36,'cp 0','d7d5 d2d3 d5e4 d3e4 d8d1 e1d1 g8f6 b1d2 f8h6 g2g3 e7e5 f3e5 c6e5 f4e5 f6g4 d2f3 h6c1 d1c1 g4f2 f1b5 e8e7 h1e1 c8g4 f3d2 g4e6 c1c2 h7h5 d2f3 e6g4 f3h4 g4e6');
INSERT INTO evaluations VALUES(22107,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(22107,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(22107,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(22107,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(22107,34,'cp 72',NULL);
INSERT INTO evaluations VALUES(22107,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(22107,36,'cp 87','g8f6 b1c3 f8e7 c1f4 e8g8 d1e2 c7c6 h2h3 e5g6 f4h2 f8e8 e1c1 e7f8 e2c4 c8e6 d4e6 f7e6 f2f4 d6d5 c4d4 b7b5 e4d5 e6d5 f4f5 g6h8 h1e1 h8f7 e1e6 e8e6 f5e6 f7d6 g2g4 f6e4 c3e4 d6e4');
INSERT INTO evaluations VALUES(15744,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(15744,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(15744,32,'cp 91',NULL);
INSERT INTO evaluations VALUES(15744,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(15744,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(15744,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(15744,36,'cp 69','b1c3 f8e7 f2f4 e5d7 d1f3 e8g8 c1e3 c7c5 d4f5 d7b6 e1c1 c8f5 e4f5 d8c7 h1e1 f8e8 e3f2 c7c6 a2a4 a7a5 c3b5 d6d5 c2c3 e7d6 e1e8 a8e8 f2h4 d5d4 b5d6 c6d6 h4f6 g7f6 f3b7 d6f4 c1b1 c5c4 b7b6 c4b3 b6b3 f4f5 b3c2 f5c2 b1c2 d4c3 c2c3 e8e3 d1d3');
INSERT INTO evaluations VALUES(15745,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(15745,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(15745,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(15745,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(15745,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(15745,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(15745,36,'cp 75','f8e7 f2f4 e5d7 d1f3 e8g8 c1e3 c7c5 d4f5 d7b6 e1c1 c8f5 e4f5 d8c7 g2g4 c7c6 b3d5 f6d5 c3d5 f8e8 c1b1 e7f8 e3c1 b6d5 d1d5 e8e2 f3e2 c6d5 h1e1 c5c4 h2h4 b7b5 h4h5 b5b4 g4g5 b4b3 a2b3 c4b3 e2e4 b3c2 e4c2 d5f3 c2e4 f3e4 e1e4 g7g6 f5g6 h7g6 h5g6 f7g6');
INSERT INTO evaluations VALUES(8680,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(8680,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(8680,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(8680,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(8680,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(8680,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(8680,36,'cp 65','d1e2 e8g8 c1e3 a7a6 e1c1 c7c5 d4f5 c8f5 e4f5 a8c8 e3f4 e5c6 h1e1 d8d7 f4e3 d7f5 h2h3 c8e8 g2g4 f5c8 g4g5 f6d7 f2f4 b7b5 b3d5 d7b6 d5e4 c6b4 a2a3 d6d5 a3b4 d5e4 c3e4 b6a4 e2f3 c5b4 c1b1 c8c4 d1d4 c4c7 f4f5 e8c8 e1e2 e7c5 e4c5 a4c5 d4b4 c7c6 f3c6 c8c6');
INSERT INTO evaluations VALUES(24230,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(24230,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(24230,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(24230,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(24230,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(24230,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(24230,36,'cp 33','b1c3 f8e7 e1g1 e8g8 c3d5 c6a5 d5e7 d8e7 f3h4 a5b3 h4f5 e7e6 a2b3 d7d5 f2f3 f6e8 c1e3 e8d6 e3c5 d5e4 f3e4 d6f5 f1f5 f8e8 b3b4 f7f6 h2h3 h7h6 f5f2 b7c6 d1f3 e6f7 g2g4 g8h8 b2b3 f7e6 f3g3 h8h7 a1f1 a6a5');
INSERT INTO evaluations VALUES(24414,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(24414,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(24414,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(24414,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(24414,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(24414,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(24414,36,'cp 39','d1f3 f8b4 c1g5 d8e7 a2a3 b4c3 f3c3 f6e4 g5e7 e4c3 e7d6 c3e4 d6e5 e8g8 e5f4 d7d5 c4a2 f8e8 e1g1 a7a5 f4e3 e4d6 f1e1 a5a4 f2f3 b7a6 e3c5 d6c4 b2b3 a4b3 a2b3 e8b8 e1e7 c4d2 c5b4 d2b3 c2b3 c6c5 b4c5 b8b3 e7c7 a6c4 a1e1 h7h6 e1e7 d5d4 c5d4 c4e6');
INSERT INTO evaluations VALUES(24982,30,'cp 121',NULL);
INSERT INTO evaluations VALUES(24982,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(24982,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(24982,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(24982,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(24982,35,'cp 126',NULL);
INSERT INTO evaluations VALUES(24982,36,'cp 123','d4d5 c6b8 f1c4 e7e5 f3e5 f8d6 e5f3 e8g8 e1g1 f5g6 b2b3 b8d7 c1b2 a7a6 c4d3 d8e7 f3h4 f8e8 h4g6 h7g6 a1c1 d7e5 d3e2 a8d8 d1c2 g6g5 c3e4 f6d5 e4d6 e7d6 c2f5 d6g6 f5g6 e5g6 g2g3 c7c6 f1d1 d5f6 g1g2 d8d1 c1d1');
INSERT INTO evaluations VALUES(27452,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(27452,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(27452,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(27452,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(27452,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(27452,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(27452,36,'cp -48','h2h3 g4d7 h1e1 e8c8 d2d3 f6d5 c1d2 f7f6 a2a3 c6c5 b1c3 d5e7 c3e4 e7c6 e2f1 b7b6 e1e2 d7e8 b2b3 e8f7 d2c3 f8e7 f3h4 h8e8 h4f5 e7f8 g2g4 a7a5 e4g3 g7g6 e2e8 d8e8 f5e3 c6d4 g3e2 e8d8 a1e1 f8g7 c3b2 d4b5');
INSERT INTO evaluations VALUES(27559,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27559,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27559,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27559,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27559,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27559,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27559,36,'cp 0','f1e1 e5d6 c3b5 d6b6 c1e3 b6a5 e3d2 a5b6');
INSERT INTO evaluations VALUES(27560,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27560,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27560,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27560,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27560,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27560,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27560,36,'cp 0','e5d6 c3b5 d6b6 c1e3 b6a5 e3d2 a5b6');
INSERT INTO evaluations VALUES(27556,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27556,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27556,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27556,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27556,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27556,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27556,36,'cp 0','c3b5 d6b6 c1e3 b6a5 e3d2 a5b6');
INSERT INTO evaluations VALUES(27555,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(27555,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(27555,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(27555,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(27555,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(27555,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(27555,36,'cp 0','d6b6 c1e3 b6a5 e3d2 a5b6');
INSERT INTO evaluations VALUES(29443,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(29443,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(29443,32,'cp 95',NULL);
INSERT INTO evaluations VALUES(29443,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(29443,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(29443,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(29443,36,'cp 104','d1b3 b8c6 d4d5 g4f3 g2f3 c6d4 b3d1 c7c5 c1g5 f6d7 e2e3 d4f5 f1d3 g7g6 d1c2 d7e5 d3e2 f8g7 b1d2 h7h6 g5f4 e8g8 e1c1 a8b8 f4e5 d6e5 h1g1 f5d6 d2e4 d8c7 a2a3 b7b5 e4d6 e7d6 c4b5');
INSERT INTO evaluations VALUES(29494,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(29494,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(29494,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(29494,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(29494,34,'cp 144',NULL);
INSERT INTO evaluations VALUES(29494,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(29494,36,'cp 155','g2g4 f5g6 c2c4 e7e6 b1c3 e6d5 g4g5 f6d7 c3d5 b8c6 h2h4 d7b6 d1e2 f8e7 d5e7 c6d4 e7c6 d4e2 c6d8 e2g1 h1g1 a8d8 b2b3 e8g8 e1f2 b6d7 c1f4 d7c5 f4c7 d8d7 c7f4 f7f6 f4e3 c5d3 f1d3 d7d3 e3c5 f8c8 c5a7 g6h5 a7e3 c8f8 f2e2 d3d7 a1d1 d7d1 e2d1 h5f3 d1d2 f6g5 e3g5 f3e4 g1e1 e4c6');
INSERT INTO evaluations VALUES(31301,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(31301,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(31301,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(31301,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(31301,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(31301,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(31301,36,'cp 0','b8c6 f1d3 f5d3 d1d3 e7e6 g1f3 f8e7 e1g1 g8f6 c3e2 d8b6 a2a4 e8g8 c2c3 c6a5 f3e5 f8c8 f1b1 e7d6 d3h3 a5b3 a1a2 b3a5');
INSERT INTO evaluations VALUES(31472,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(31472,31,'cp 29',NULL);
INSERT INTO evaluations VALUES(31472,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(31472,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(31472,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(31472,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(31472,36,'cp 43','d2d4 f8b4 e1g1 g8e7 h2h3 g4h5 e4e5 h5f3 e2f3 c6c5 a2a3 b4c3 b2c3 b8c6 a3a4 d8a5 d1d3 a5c7 c1a3 e8g8 f3g4 b7b6 f2f4 c6a5 f4f5 e7f5 g4f5 e6f5 f1f5 a8e8 a1f1 e8e6 d3f3 a5c4 a3c1 c7d7 h3h4 f7f6 h4h5');
INSERT INTO evaluations VALUES(33461,30,'cp 644',NULL);
INSERT INTO evaluations VALUES(33461,31,'cp 633',NULL);
INSERT INTO evaluations VALUES(33461,32,'cp 634',NULL);
INSERT INTO evaluations VALUES(33461,33,'cp 654',NULL);
INSERT INTO evaluations VALUES(33461,34,'cp 635',NULL);
INSERT INTO evaluations VALUES(33461,35,'cp 671',NULL);
INSERT INTO evaluations VALUES(33461,36,'cp 637','c3d5 f6g6 d5c7 e8d8 c7a8 g6e4 c4e2 b7b6 e1g1 h8e8 c2c3 c8b7 d2d3 e4g6 b2b4 c5d6 a8b6 a7b6 a2a4 d6c7 a4a5 h6f5 a5b6 c7b6 f3g5 e8e7 e2h5 g6c6 g5e4 f5d6 h5f3 d6e4 d3e4 d8e8 d1d3 d7d6 f1d1 c6c7 d3d6 c7d6 d1d6 b8d7');
INSERT INTO evaluations VALUES(33737,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(33737,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(33737,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(33737,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(33737,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(33737,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(33737,36,'cp 36','c7c6 f1d3 c6c5 e1g1 e7e6 f1e1 f8e7 b2b3 b8c6 c1b2 a5c7 c3b5 c7d8 b5a3 e8g8 a3c4 b7b6 c4e5 c6b4 d3f1 c8b7 d2d4 d8c7 a2a3 b4c6 d4c5 e7c5 e5c6 c7c6 b3b4 a8d8 d1e2 c5d6 f3d4 c6e4 e2e4');
INSERT INTO evaluations VALUES(37976,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(37976,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37976,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37976,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(37976,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(37976,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(37976,36,'cp 0','d3d4 c5d4 f3d4 b8c6 c1e3 g8f6 f1e2 e8g8 e1g1 d6d5 e4d5 f6d5 c3d5 c6d4 e3d4 d8d5 d4g7 d5d1 f1d1 g8g7 b2b3 c8e6 f2f4 h7h5 c2c4 b7b6 g1f2 h5h4 d1d2 g7f6 a1d1 a8c8 f2e3 a7a5 e2f3 a5a4 f3d5 e6f5 d5e4 f5e6');
INSERT INTO evaluations VALUES(39434,30,'cp -131',NULL);
INSERT INTO evaluations VALUES(39434,31,'cp -149',NULL);
INSERT INTO evaluations VALUES(39434,32,'cp -120',NULL);
INSERT INTO evaluations VALUES(39434,33,'cp -132',NULL);
INSERT INTO evaluations VALUES(39434,34,'cp -138',NULL);
INSERT INTO evaluations VALUES(39434,35,'cp -141',NULL);
INSERT INTO evaluations VALUES(39434,36,'cp -126','d7d5 e4d5 c8g4 b1c3 g8f6 d1d3 e8g8 c1f4 f6d5 c3d5 d8d5 g2g3 h4f6 f1g2 g4f5 d3b3 f8e8 e2f2 f6d4 f3d4 d5d4 f4e3 d4f6 h1e1 b8c6 b3c3 f6c3 b2c3 f5e4 g2h3 e8e7 e3f4 f7f5 g3g4 a8f8 f2g3 h7h6 g4f5 e4f5 e1e7 c6e7 h3f1 e7d5 f1c4 c7c6 a1e1 f5c2');
INSERT INTO evaluations VALUES(40956,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(40956,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(40956,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(40956,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(40956,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(40956,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(40956,36,'cp 20','b8c6 f1g2 e8g8 f3e5 d5c4 g2c6 b7c6 e5c6 d8e8 c6e7 e8e7 d1a4 c7c5 d4c5 e7c5 a4b5 c5b5 c3b5 f6d5 f2f3 a8b8 b5c3 f8d8 e1f2 c8d7 h1d1 d7e8 a1b1 f7f6 c1d2 e8g6 e2e4 d5c3 d2c3 e6e5 b1c1 h7h5 f2e2 g8h7 c1b1 h7g8 d1d8 b8d8 b1d1 d8d1 e2d1 g6f7 c3d2 f7e6');
INSERT INTO evaluations VALUES(41280,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(41280,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(41280,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(41280,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(41280,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(41280,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(41280,36,'cp 27','c8f5 e1g1 e8g8 b1c3 e7e5 d4e5 d6e5 d1b3 f6e4 b3b7 e4c3 b2c3 b8d7 c1a3 c7c5 e3e4 a8b8 b7d5 f5e6 d5d3 f8e8 f3d2 d7b6 d3c2 b6a4 d2b3 d8c7 f1b1 e8f8 f2f3 f7f5 b3d2 c7a5 b1b3 b8b3 a2b3 a5c3');
INSERT INTO evaluations VALUES(42229,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(42229,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(42229,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(42229,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(42229,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(42229,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(42229,36,'cp 27','e2e3 e8g8 f1d3 f8e8 g1e2 c7c6 d1c2 h7h6 h2h3 b4f8 e1g1 a7a5 a1c1 b8a6 c2d1 a6b8 a2a4 f8b4 d1d2 b8d7 d2c2 d8e7 g1h1 d7f8 f4e5 f6d7 e5h2 d7f6 e2g1 f8g6 d3g6 f7g6 c2g6 b4c3 b2c3 b7b5 g1f3 b5a4 f3e5 c8a6 e5c6 e7f7');
INSERT INTO evaluations VALUES(42420,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(42420,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(42420,32,'cp 88',NULL);
INSERT INTO evaluations VALUES(42420,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(42420,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(42420,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(42420,36,'cp 65','d1c2 e4c3 b2c3 b4e7 f1d3 h7h6 g1e2 d7d6 e1g1 e8g8 f2f4 c7c5 e3e4 c5d4 c3d4 e6e5 a1b1 b7b6 h2h3 b8c6 d4d5 c6b8 f4e5 d6e5 e2g3 b8a6 c2e2 a6c5 d3c2 c8d7 f1f3 a8c8 c1e3 e7g5 e3g5 h6g5 b1f1 g5g4');
INSERT INTO evaluations VALUES(42421,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(42421,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(42421,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(42421,33,'cp 83',NULL);
INSERT INTO evaluations VALUES(42421,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(42421,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(42421,36,'cp 64','e4c3 b2c3 b4e7 f1d3 h7h6 g1e2 c7c5 e1g1 d7d6 f2f4 e8g8 e3e4 e6e5 f4e5 d6e5 d4d5 e7h4 a2a4 b8d7 g1h1 d8e7 e2g1 h4g5 g1f3 g5c1 c2c1 d7f6 c1e3 f6h5 h1g1 c8d7 f1e1 e7d6 h2h3 a8e8 a4a5 h5f4 d3c2 b7b6 f3h4');
INSERT INTO evaluations VALUES(43433,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(43433,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(43433,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(43433,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(43433,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(43433,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(43433,36,'cp 14','c1f4 b8c6 e2e3 c8g4 h2h3 g4f3 d1f3 e7e6 f1d3 h7h6 f4g3 c6b4 d3b1 f8d6 g3h4 e8g8 e1g1 a8c8 f1c1 b4c6 b1d3 c6b4 d3e2 b4c6 a2a3 c6a5 a1b1 a5b3 c1d1 d6e7 e2d3 b3a5 d1c1 d8d7 f3d1 b7b5 h4f6 e7f6 a3a4 a5c4 c3a2 f6e7');
INSERT INTO evaluations VALUES(43429,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(43429,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(43429,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(43429,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(43429,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(43429,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(43429,36,'cp 9','b8c6 a1c1 c8g4 f3e5 c6e5 f4e5 e7e6 f2f3 g4h5 e2e4 d5e4 e5f6 g7f6 d1a4 b7b5 f1b5 a6b5 a4b5 e8e7 b5b7 e7e8 b7c6 e8e7 c3e4 f8h6 c6c5 e7d7 c1c3 h5g6 e1g1 g6e4 f3e4 d7e8 d4d5 h8g8 b2b4 g8g5 a2a3 g5g6 g1h1 a8b8 b4b5');
INSERT INTO evaluations VALUES(46183,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(46183,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(46183,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(46183,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(46183,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(46183,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(46183,36,'cp 43','g1f3 c5d4 e1g1 f8c5 f3d4 e8g8 c4d5 f6d5 d4b3 c5e7 c1d2 b8c6 b1c3 d5c3 d2c3 d8b6 b3d4 c6b4 e2e3 f8d8 d1g4 e7f8 d4b3 h7h6 f1d1 d8d1 g4d1 b4c6 b3d2 b6c7 a1c1 c8d7 a2a3 a8d8 d1c2 e6e5 g2e4 d7e6 c2b1 a7a5');
INSERT INTO evaluations VALUES(48578,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(48578,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(48578,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(48578,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(48578,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(48578,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(48578,36,'cp 95','d3e4 f6e4 g1f3 c7c6 f1d3 e4f6 d5c6 b8c6 b1c3 f8e7 e1g1 e8g8 a2a3 d8a5 a1b1 c8f5 b2b4 a5b6 g1h1 f5d3 c2d3 a8d8 b4b5 c6b8 f3e5 b8d7 a3a4 e7c5 d1f3 f8e8 e5c4 b6c7 c1b2 c5f8 c4e5 c7b6 c3e2 d7c5 b2d4');
INSERT INTO evaluations VALUES(48570,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(48570,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(48570,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(48570,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(48570,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(48570,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(48570,36,'cp 55','f6e4 g1f3 c8f5 b1d2 e4f6 d1e2 d8e7 f3d4 f5g4 e2e7 f8e7 f1c4 e8g8 e1g1 f8d8 f1e1 e7f8 f4f5 c7c5 d5c6 b8c6 d4c6 b7c6 d2f3 g4f5 f3e5 f8c5 c1e3 c5e3 e1e3 f5c2 c4f7 g8f8 a1e1 d8d4 e5c6 d4e4 f7b3 c2b3 a2b3 e4e3 e1e3 a8c8 e3c3 c8e8 g1f2 f6g4 f2g3 g4f6 c3d3 e8e2');
INSERT INTO evaluations VALUES(48973,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(48973,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(48973,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(48973,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(48973,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(48973,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(48973,36,'cp 94','g1f3 f8e7 d1d4 e4d6 b1c3 e8g8 c1d2 e7f6 f3e5 b8d7 e1c1 f8e8 d1e1 d6f5 d4f2 d7e5 f4e5 e8e5 c3e4 f6h4 g2g3 h4e7 f1c4 f5d6 e4d6 e7d6 e1e5 d6e5 h1f1 d8f6 f2f6 g7f6 f1f2 c8d7 d2f4 b7b5 c4d3 g8g7 c2c4 b5c4 d3c4 e5f4 f2f4 a8e8 c1d2 g7g6 c4d3 g6g7');
INSERT INTO evaluations VALUES(48972,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(48972,31,'cp 67',NULL);
INSERT INTO evaluations VALUES(48972,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(48972,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(48972,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(48972,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(48972,36,'cp 72','c8f5 c1e3 f8e7 f1c4 e7h4 g2g3 h4f6 c2c3 e8g8 e1g1 f8e8 f1e1 f5g4 b1d2 e4d2 e3d2 b8d7 d2e3 e8e4 c4d3 e4e7 g1g2 d7b6 e3b6 a7b6 e1e7 d8e7 h2h3 g4d7 d1e2 e7e2 d3e2 a8e8 a1e1 g7g6 e2c4 e8e1 f3e1 c7c6 g3g4 b6b5');
INSERT INTO evaluations VALUES(48673,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(48673,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48673,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(48673,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(48673,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(48673,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48673,36,'cp 0','f8d6 e1g1 e8g8 b1c3 f8e8 h2h3 c7c6 c1f4 b7b6 f4d6 d8d6 e2g3 g7g6 d1d2 c8a6 d3a6 b8a6 d2g5 d6d8 f2f4 f6d7 g5g4 d7f6');
INSERT INTO evaluations VALUES(41715,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41715,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41715,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41715,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41715,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41715,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41715,36,'cp 0','e1g1 e8g8 b1c3 f8e8 h2h3 c7c6 c1f4 b7b6 f4d6 d8d6 e2g3 g7g6 d1d2 c8a6 d3a6 b8a6 d2g5 d6d8 f2f4 f6d7 g5g4 d7f6');
INSERT INTO evaluations VALUES(41714,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41714,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41714,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41714,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41714,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41714,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41714,36,'cp 0','e8g8 b1c3 f8e8 h2h3 c7c6 c1f4 b7b6 f4d6 d8d6 e2g3 g7g6 d1d2 c8a6 d3a6 b8a6 d2g5 d6d8 f2f4 f6d7 g5g4 d7f6');
INSERT INTO evaluations VALUES(36946,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36946,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36946,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(36946,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36946,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36946,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36946,36,'cp 0','b1c3 f8e8 h2h3 c7c6 c1f4 b7b6 f4d6 d8d6 e2g3 b8a6 d1d2 a6c7 c3e2 c8d7 f1e1 e8e7 d2f4 f6e8 f4g5 e8f6');
INSERT INTO evaluations VALUES(48687,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(48687,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(48687,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(48687,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(48687,34,'cp -32',NULL);
INSERT INTO evaluations VALUES(48687,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(48687,36,'cp -24','b1c3 c7c6 c1f4 f8b4 g1f3 c8f5 f1g2 e8g8 d1b3 d8b6 e1g1 b4e7 f3h4 f5e6 h4f3 b6b3 a2b3 f8e8 a1a4 h7h6 f1a1 a7a6 h2h3 b8d7 g3g4 h6h5 g4g5 f6e4 e2e3 a8c8 h3h4 d7b6 a4a2 e7b4 c3e2 e6f5 e2g3');
INSERT INTO evaluations VALUES(49095,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(49095,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(49095,32,'cp 169',NULL);
INSERT INTO evaluations VALUES(49095,33,'cp 157',NULL);
INSERT INTO evaluations VALUES(49095,34,'cp 173',NULL);
INSERT INTO evaluations VALUES(49095,35,'cp 154',NULL);
INSERT INTO evaluations VALUES(49095,36,'cp 158','f3g5 d8e7 c4f7 e8d8 f7b3 d8e8 g5f7 h8g8 b1c3 c8e6 b3e6 e7e6 f7g5 e6d7 f2f4 b8c6 e1g1 h7h6 g5f3 f8e7 f4f5 a8d8 f3h4 d6d5 e4d5 c6b4 h4g6 b4d5 c3d5 d7d5 d1e2 e7c5 g1h1 e5e4 c1f4 d5f5 e2b5 d8d7 g6e5 c5d6 f4g3 f5e6 e5d7 e6d7 b5d7 e8d7 g3d6 c7d6');
INSERT INTO evaluations VALUES(49353,30,'cp -258',NULL);
INSERT INTO evaluations VALUES(49353,31,'cp -260',NULL);
INSERT INTO evaluations VALUES(49353,32,'cp -270',NULL);
INSERT INTO evaluations VALUES(49353,33,'cp -261',NULL);
INSERT INTO evaluations VALUES(49353,34,'cp -259',NULL);
INSERT INTO evaluations VALUES(49353,35,'cp -266',NULL);
INSERT INTO evaluations VALUES(49353,36,'cp -251','b8c6 e5g5 f6e4 g5d8 e8d8 d2d3 e4d6 c2c3 h7h5 h2h4 b7b6 d3d4 a7a5 c1f4 d6f5 g2g3 c6e7 f1g2 c7c6 g2e4 f7f6 g1e2 c8a6 b1d2 d8e8 d2f1 e8f7 f1e3 f5e3 f2e3 a8c8 a2a4 c6c5 a1d1 d7d5 e4c2 c5d4 e3d4 e7f5 c2f5 e6f5');
INSERT INTO evaluations VALUES(16250,30,'cp -264',NULL);
INSERT INTO evaluations VALUES(16250,31,'cp -261',NULL);
INSERT INTO evaluations VALUES(16250,32,'cp -259',NULL);
INSERT INTO evaluations VALUES(16250,33,'cp -261',NULL);
INSERT INTO evaluations VALUES(16250,34,'cp -254',NULL);
INSERT INTO evaluations VALUES(16250,35,'cp -256',NULL);
INSERT INTO evaluations VALUES(16250,36,'cp -250','e5g5 f6e4 g5d8 e8d8 c2c3 d7d5 d2d3 e4d6 f2f4 h7h5 g1f3 d6f5 g2g3 f8d6 d3d4 c6e7 b2b4 a7a5 b4b5 c8d7 a2a4 c7c5 d4c5 d6c5 b1d2 a8c8 d2b3 c5b6 e1d2 h5h4 g3g4 f5d6 f1d3 d8e8 c1a3 d6e4 d3e4 d5e4 f3g5 e4e3 d2c2 e6e5 g5e4');
INSERT INTO evaluations VALUES(52731,30,'cp -97',NULL);
INSERT INTO evaluations VALUES(52731,31,'cp -100',NULL);
INSERT INTO evaluations VALUES(52731,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(52731,33,'cp -110',NULL);
INSERT INTO evaluations VALUES(52731,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(52731,35,'cp -111',NULL);
INSERT INTO evaluations VALUES(52731,36,'cp -118','d5d4 c3e2 e7e5 d2d3 f8b4 c1d2 a7a5 f2f4 g8h6 h2h3 f7f6 g1f3 b8d7 e2g3 e5f4 g3e2 f6f5 e2d4 f5e4 d3e4 d7c5 d1c2 b4d2 e1d2 e8g8 a1e1 h6f7 d2c1 f7g5 d4b3 g5f3 g2f3 d8h4 e1d1 c5e6 c2c3 a5a4 b3d4 e6d4 c3d4');
INSERT INTO evaluations VALUES(53181,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(53181,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(53181,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(53181,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(53181,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53181,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53181,36,'cp 0','b8c6 f1b5 c8d7 e1g1 e7e6 c3d4 c5d4 f1e1 g8e7 b5f1 h7h6 d2d3 a7a5 b1d2 b7b5 a2a3 a5a4 b2b4 a4b3 d2b3 d8b6 c1b2 a8a4 e1e4 e7d5 f3d4 f8e7 d4c6 d7c6 b2d4 b6c7 d4b2');
INSERT INTO evaluations VALUES(19759,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(19759,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(19759,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(19759,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(19759,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(19759,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(19759,36,'cp 0','f1b5 c8d7 e1g1 e7e6 f1e1 c6e5 f3e5 d7b5 b1a3 b5c6 e5c6 b7c6 a3c4 g8f6 d1a4 d8c7 b2b3 d4c3 d2c3 f6d5 c4e5 f8d6 a4c6 c7c6 e5c6 a7a5 c1d2 e8d7 c6e5 d7e7 e5c6');
INSERT INTO evaluations VALUES(53629,30,'cp -296',NULL);
INSERT INTO evaluations VALUES(53629,31,'cp -294',NULL);
INSERT INTO evaluations VALUES(53629,32,'cp -306',NULL);
INSERT INTO evaluations VALUES(53629,33,'cp -295',NULL);
INSERT INTO evaluations VALUES(53629,34,'cp -271',NULL);
INSERT INTO evaluations VALUES(53629,35,'cp -274',NULL);
INSERT INTO evaluations VALUES(53629,36,'cp -298','d5e4 f3g1 a7a5 a2a3 b7b5 a4b3 g8f6 d2d4 c6c5 g1e2 c5d4 e2d4 a5a4 b3a2 d8b6 b1c3 f8c5 c1e3 c8a6 d1d2 e8g8 b2b4 a4b3 d4b3 c5e3 d2e3 b6e3 f2e3 f6g4 c3e4 g4e3 e1f2 e3c2 a1c1 c2a3 c1c7 f8c8 c7c8 a6c8 h1c1 b8d7 e4d6 c8a6');
INSERT INTO evaluations VALUES(53941,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(53941,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(53941,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(53941,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(53941,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(53941,35,'cp 95',NULL);
INSERT INTO evaluations VALUES(53941,36,'cp 113','b8c6 f3g5 g8h6 d1h5 g7g6 h5f3 f7f6 g5e6 c8e6 c4e6 c6d4 f3h3 f6f5 e6d5 f5f4 h3d3 f8g7 c3e2 d4c6 c1d2 d8d7 f2f3 h6g8 d3b3 g8f6 d5c4 d7c7 a2a4 e8e7 a4a5 h8c8 c4e6 f6d7 e6d5 c6d4 e2d4 e5d4 e1g1 c7c2 b3b4 c2c5 a5a6 d4d3 b4c5 c8c5 a6b7 a8b8 g2g3 g7b2 a1a7 b2d4 g1h1 c5c2');
INSERT INTO evaluations VALUES(20631,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(20631,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(20631,32,'cp 136',NULL);
INSERT INTO evaluations VALUES(20631,33,'cp 128',NULL);
INSERT INTO evaluations VALUES(20631,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(20631,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(20631,36,'cp 130','f3g5 g8h6 d1h5 g7g6 h5f3 f7f6 g5e6 c8e6 c4e6 c6d4 f3h3 f6f5 e6d5 f5f4 h3d3 f8g7 c3e2 d4c6 c1d2 h6f7 e1c1 e8g8 c1b1 g8h8 f2f3 d8d7 h2h4 h7h5 h1g1 h8h7 g2g4 f4g3 g1g3 c6e7 d5b3 d6d5 d1g1 g7h6 e2c3 h6d2 d3d2 d5d4 c3d5 e7d5 e4d5 f7h8 g3g5 f8f3');
INSERT INTO evaluations VALUES(54053,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(54053,31,'cp -215',NULL);
INSERT INTO evaluations VALUES(54053,32,'cp -216',NULL);
INSERT INTO evaluations VALUES(54053,33,'cp -212',NULL);
INSERT INTO evaluations VALUES(54053,34,'cp -198',NULL);
INSERT INTO evaluations VALUES(54053,35,'cp -209',NULL);
INSERT INTO evaluations VALUES(54053,36,'cp -186','d8a5 b5c3 c5d4 d1d4 b8c6 d4a4 a5a4 c3a4 c8d7 a4c3 d5d4 c3b5 e6e5 f4d2 c6b4 b5a3 a8c8 g1f3 f7f6 c2c3 d4c3 b2c3 b4d5 a3b1 d5b6 e2e4 f8c5 f1d3 g8e7 e1e2 e8f7 h1c1 d7e6 d2e3 c5e3 e2e3 h8d8 d3c2 b6c4 e3e2 g7g5 c2b3 e7c6 b1d2 c6a5 g2g3 a7a6 c1d1 a5b3 a2b3');
INSERT INTO evaluations VALUES(55413,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(55413,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(55413,32,'cp 28',NULL);
INSERT INTO evaluations VALUES(55413,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55413,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(55413,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(55413,36,'cp 36','c7c5 f1c4 c8e6 d2d3 f8e7 a2a4 g8f6 f3h4 d6d5 c4d5 f6d5 e4d5 e6d5 h4f5 e7f8 c3d5 d8d5 f5e3 d5d7 e3c4 b8c6 e1g1 e8c8 c1e3 g7g6 b2b4 c6b4 c4e5 d7d5 e5c4 f8g7 a1b1 h8e8 d1d2 e8e6 e3f4 b4c6 f4h6');
INSERT INTO evaluations VALUES(2506,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(2506,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(2506,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(2506,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(2506,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(2506,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(2506,36,'cp 70','f3e5 f7e7 d2d4 d7d6 e5d3 d8e8 c1g5 g8f6 e4e5 e8g6 h2h4 h7h6 e5f6 g7f6 d1e2 e7d8 g5e3 g6g2 h1f1 c8g4 f2f3 g2e2 e1e2 g4f5 b2b3 d6d5 b1c3 c7c6 e2d2 b7b6 c3e2 a5b7 f1g1 d8d7 a1e1 b7d6 e2g3 f8e7 e3f4 f5g6 h4h5 g6f7 a2a4 a7a5 e1e2 a8e8 c2c3 e7d8 f4d6 e8e2 d2e2 d7d6');
INSERT INTO evaluations VALUES(2505,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(2505,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(2505,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(2505,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(2505,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(2505,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(2505,36,'cp 63','f7e7 d2d4 d7d6 e5d3 d8e8 c1g5 g8f6 e4e5 e8g6 h2h4 h7h6 e5f6 g7f6 d1e2 e7d8 g5d2 g6g2 h1f1 a5c6 d3f4 g2g4 d2e3 g4e2 e1e2 c6e7 c2c4 h6h5 b1c3 c8f5 f1g1 d8e8 a2a4 c7c6 a4a5 e8f7 f2f3 f8h6 c3e4 a8d8 d4d5 c6d5 f4d5 f5e4 f3e4 d8e8 e3h6 h8h6 e2d3 h6g6 g1g6 e7g6 a5a6 b7a6 a1a6');
INSERT INTO evaluations VALUES(4378,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(4378,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(4378,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(4378,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(4378,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(4378,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(4378,36,'cp 20','f3e5 c6e5 d2d4 c5d6 d4e5 d6e5 e1g1 e5c3 b2c3 d7d6 f2f4 f8e8 e4e5 f6d7 e5d6 d7f6 c1a3 f6e4 d1d4 e4d6 f1e1 c8e6 a3d6 c7d6 g2b7 a8b8 b7c6 e8e7 a1d1 e6g4 d1b1 g4f5 c6d5 b8b1 e1b1 h7h6 b1b7 e7b7 d5b7 d8c7');
INSERT INTO evaluations VALUES(4409,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4409,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4409,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4409,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(4409,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4409,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4409,36,'cp 0','f6d5 c4d5 d7d6 e1g1 d8e8 c1e3 c5e3 f2e3 a7a5 f3g5 c6d8 d1h5 h7h6 f1f6 h6g5 f6h6 g7h6 h5g6 g8h8 g6h6 h8g8 h6g6 g8h8');
INSERT INTO evaluations VALUES(4809,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4809,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4809,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4809,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4809,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(4809,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4809,36,'cp 0','c6c5 d4b3 d8h4 h2h3 g8e7 f1e1 f7f6 b1d2 c8h3 g2h3 h4h3 d1f3 d6h2 g1h1 h2g3 h1g1');
INSERT INTO evaluations VALUES(6409,30,'cp 128',NULL);
INSERT INTO evaluations VALUES(6409,31,'cp 115',NULL);
INSERT INTO evaluations VALUES(6409,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(6409,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(6409,34,'cp 115',NULL);
INSERT INTO evaluations VALUES(6409,35,'cp 125',NULL);
INSERT INTO evaluations VALUES(6409,36,'cp 115','d8e7 c1d2 d7d6 e5d6 c5d6 f1d3 g8f6 b1c3 e8g8 h2h3 b7b6 e1g1 c8b7 d1e2 f8e8 f1e1 a7a6 e3e4 b6b5 a2a3 a8d8 a1d1 h7h6 d2e3 c6a5 b2b4 a5c4 d3c4 b5c4 e4e5 b7f3 e5d6 f3e2 d6e7 d8d1 c3d1 e2d1 e1d1 e8e7 d1d8 g8h7');
INSERT INTO evaluations VALUES(7234,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(7234,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(7234,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(7234,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(7234,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(7234,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(7234,36,'cp 37','d2d4 c8g4 b1d2 f6d7 h2h3 g4e6 d4e5 d7e5 f3e5 d6e5 d1h5 e5f4 d2b3 f4c1 a1c1 e6b3 a2b3 e8g8 e4e5 d8d2 h5g4 a8d8 e1d1 d2a5 d1d8 f8d8 e5e6 a5d2 c1a1 f7e6 g4e6 g8h8 e6e7 h8g8 a1e1 d8f8 e7e2 d2e2 e1e2 f8d8 f2f4 a6a5 g1f2 g8f7 g2g4 h7h6 f2g3');
INSERT INTO evaluations VALUES(7220,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(7220,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(7220,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(7220,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(7220,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(7220,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(7220,36,'cp 55','f6d7 b1d2 e5d4 f3d4 e8g8 d4f5 d7e5 f5d6 c7d6 b2b3 d8f6 c1b2 c8e6 b2c3 a8d8 d2f1 h7h6 f1g3 f6g5 d1d4 c6c5 d4e3 e5g6 a1d1 g5e3 e1e3 d6d5 c3a5 d8d7 e4d5 d7d5 d1d5 e6d5 g3e4 c5c4 e4d6 c4b3 a2b3');
INSERT INTO evaluations VALUES(7646,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(7646,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(7646,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(7646,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(7646,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(7646,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(7646,36,'cp 53','e4e5 f6d5 a2a3 b4c3 b2c3 d8c7 f2f4 c6c5 c3c4 d5e7 f1d3 c8b7 d1g4 g7g6 e1g1 e7f5 g4g5 e8g8 d2c3 c7d8 g2g4 f5e3 f1e1 e3g2 g5d8 a8d8 e1f1 b7c6 c3d2 f7f6 a1b1 f6e5 f4e5 f8f1 g1f1');
INSERT INTO evaluations VALUES(7752,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(7752,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(7752,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(7752,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(7752,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(7752,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(7752,36,'cp 24','d5c4 d3c4 e8g8 e3e4 e6e5 c1e3 d8e7 h2h3 e5d4 e3d4 d7e5 f3e5 d6e5 f2f4 e5d4 d1d4 b7b5 c4d3 c6c5 d4e5 e7e5 f4e5 f6d7 e5e6 f7e6 d3b5 d7e5 f1c1 a8b8 b2b3 h7h6 b5f1 c8d7 c3a4 d7a4 c1c5 e5d7 c5c7 f8c8 a1c1 c8c7 c1c7 a4b5 f1b5 b8b5 c7d7 b5a5 d7d6');
INSERT INTO evaluations VALUES(7829,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(7829,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(7829,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(7829,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(7829,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7829,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7829,36,'cp 8','d4c6 b7c6 f1d3 h7h6 g5h4 d7d6 e1g1 e6e5 c3a4 g7g5 a2a3 b4a5 h4g3 c8e6 b2b4 a5c7 d1d2 f6h5 a1d1 e8g8 d3a6 d8e8 d2c3 a8d8 f1e1 h5f4 a6b7 c6c5 g3f4 g5f4 a4b2 c7b6 b2c4 c5b4 a3b4 b6d4 d1d4 e5d4 c3d4 e6c4 d4c4');
INSERT INTO evaluations VALUES(7812,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(7812,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(7812,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(7812,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(7812,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(7812,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7812,36,'cp 0','b7c6 f1d3 h7h6 g5h4 e6e5 e1g1 d7d6 c3a4 g7g5 a2a3 b4a5 h4g3 a5c7 d1d2 e8g8 c2c4 c8e6 a1d1 d8e7 d2c3 a8d8 c4c5 d6d5 g3e5 f6g4 e5g3 c7g3 h2g3 d5d4 c3c2 h6h5 d3e2 h5h4 e2g4 e6g4 f2f3');
INSERT INTO evaluations VALUES(7643,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(7643,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7643,32,'cp 6',NULL);
INSERT INTO evaluations VALUES(7643,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(7643,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(7643,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(7643,36,'cp 6','f1d3 h7h6 g5h4 e6e5 e1g1 e8g8 f2f4 b4d6 d3c4 d8e7 f4f5 d6b4 c3a4 d7d5 e4d5 f8d8 d1f3 e5e4 h4f6 e7f6 f3e4 c8b7 d5c6 f6c6 e4c6 b7c6 c2c3 b4f8 c4b3 d8d2 f1f2 a8d8 a1e1 d2f2 g1f2 d8d2 e1e2 d2e2 f2e2 c6g2 b3c2 f8d6 h2h4 d6e7 h4h5 g2c6 c3c4 c6a4 c2a4 e7f6 e2f3 f6b2 a4b5 b2e5');
INSERT INTO evaluations VALUES(10521,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(10521,31,'cp 66',NULL);
INSERT INTO evaluations VALUES(10521,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(10521,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(10521,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(10521,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(10521,36,'cp 63','e8g8 f1e2 f8e8 f2f3 d7d5 d4f6 d8f6 c3d5 f6e5 d1d2 e5b2 e1f2 b2e5 d2c3 e5c3 d5c3 c6d4 c3d5 d4e6 a2a4 c7c6 d5e3 g8f8 e3c4 e6c5 h1d1 c8e6 c4e5 a8d8 d1d8 e8d8 e5d3 c5d7 f3f4 f8e7 f2e3 f7f6 a4a5 e7d6 g2g4');
INSERT INTO evaluations VALUES(4482,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(4482,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(4482,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(4482,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(4482,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(4482,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(4482,36,'cp 63','f1e2 f8e8 d4f6 d8f6 d1d2 d7d5 c3d5 f6b2 e1g1 b2e5 a1d1 c8e6 f1e1 a8d8 c2c4 e5e4 e2d3 e4h4 d5c7 e8e7 c7d5 e7d7 e1e4 h4h5 d1e1 d8f8 d5f4 h5a5 d2a5 c6a5 f4e6 f7e6 d3e2 d7d2 e4e6 a5c6 e1d1 d2d1 e2d1 f8d8 e6e1 d8d2 d1b3');
INSERT INTO evaluations VALUES(11380,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(11380,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(11380,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(11380,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(11380,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(11380,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(11380,36,'cp 61','b7b5 c4d3 c6b4 d3e2 b4d5 c3d5 e6d5 d1c2 f8d6 c1d2 c8d7 f3e5 e8g8 a1c1 d8e7 e5d7 e7d7 e2d3 d7e6 d3f5 e6e7 d2a5 g7g6 f5d3 a8c8 c2c6 e7e6 a2a3 f6d7 e3e4 d5e4 d3e4 d7f6 e4f3 h7h5 g2g3 e6f5 g1g2 g8g7');
INSERT INTO evaluations VALUES(12112,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(12112,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(12112,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(12112,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(12112,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(12112,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(12112,36,'cp 30','a7a5 d2b1 f8e7 f1e2 e8g8 e1g1 a5a4 c1e3 f7f6 e2d3 f6f5 h2h3 c5d4 c3d4 d7b6 b1c3 c8d7 e3f2 c6a5 b2b3 d7e8 a1c1 a4b3 a2b3 e8h5 f2e1 a5c6 d3e2 e7a3 c1b1 d8d7 c3a4 h5f3 f1f3 b6c8 g1h2 a3b4 e1b4 c6b4 a4c5');
INSERT INTO evaluations VALUES(13708,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(13708,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(13708,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(13708,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(13708,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(13708,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(13708,36,'cp 14','e1g1 f8e7 d4c5 e7c5 b1c3 e8g8 c1g5 c8g4 h2h3 g4h5 b5a4 a8c8 g5f6 d8f6 d1d5 h5f3 d5f3 f6f3 g2f3 c6d4 g1g2 f8d8 a1d1 g8f8 a4b3 b7b6 f3f4 c5a3 b2a3 c8c3 f1e1 c3c8 e1e4 d4e6 d1d8 c8d8 b3e6 f7e6 e4a4 d8d2 c2c4 d2a2 a4a7 a2c2 a7c7 c2c1 c7c6');
INSERT INTO evaluations VALUES(13706,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(13706,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(13706,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(13706,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(13706,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(13706,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(13706,36,'cp 11','f8e7 d4c5 e7c5 c1g5 e8g8 b1c3 c8g4 h2h3 g4h5 b5a4 a8c8 g5f6 d8f6 d1d5 h5f3 d5f3 f6f3 g2f3 c6d4 g1g2 f8d8 a1d1 g8f8 a4b3 d4b3 c2b3 g7g6 f1e1 c5d4 e1e2 d4g7 d1d8 c8d8 c3b5 a7a6 b5c3 d8d3 c3e4 g7d4 b3b4 b7b6 a2a3 d3d1');
INSERT INTO evaluations VALUES(13999,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(13999,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(13999,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(13999,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(13999,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(13999,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(13999,36,'cp 49','c1e3 f8e7 d1d2 e6e5 d4b3 a7a5 c3d5 f6d5 e4d5 c6b8 a2a4 e8g8 e2b5 c8d7 e1g1 d7b5 a4b5 b7b6 f2f4 d8c7 f4e5 d6e5 d2d3 b8d7 a1d1 c7d6 b3d2 d6g6 d2c4 a8c8 d5d6 g6d3 c2d3 e7d8 d1c1 a5a4 d3d4 e5d4 e3d4');
INSERT INTO evaluations VALUES(13994,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(13994,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(13994,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(13994,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(13994,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(13994,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(13994,36,'cp 39','f8e7 d1d2 e6e5 d4b3 a7a5 c3d5 f6d5 e4d5 c6b8 a2a4 e8g8 e2b5 c8d7 e1g1 d7b5 a4b5 b7b6 f2f4 d8c7 f4e5 d6e5 d2d3 b8d7 a1d1 e7d6 b3d2 d7c5 e3c5 c7c5 g1h1 c5d4 d3f5 g7g6 f5f6 a8d8 b2b3 a5a4 d2f3 d4f4 f6f4 e5f4 b3a4 d8a8 d1d4 f8c8 c2c3 c8c5 f1d1');
INSERT INTO evaluations VALUES(8205,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(8205,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(8205,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(8205,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(8205,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(8205,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(8205,36,'cp 25','g2g4 d6d5 e4d5 f6d5 c3d5 d8d5 e1g1 c6d4 d1d4 d5d4 e3d4 e7f6 d4f6 g7f6 f2f4 e8e7 g1f2 e6e5 a1d1 e5f4 f2e1 f6f5 f1f4 f5g4 e2g4 c8g4 f4g4 a8d8 d1d8 e7d8 e1f2 h7h6 b2b3 b7b6 c2c4 h8e8 g4h4 e8e6 f2f3 a7a5 h4h5 e6f6 f3e3');
INSERT INTO evaluations VALUES(17036,30,'cp 131',NULL);
INSERT INTO evaluations VALUES(17036,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(17036,32,'cp 142',NULL);
INSERT INTO evaluations VALUES(17036,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(17036,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(17036,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(17036,36,'cp 130','e2e4 d7d6 c1e3 f8e7 e1c1 e8g8 f2f3 c7c6 g2g4 d8a5 c1b1 f8e8 d4d2 f6d7 g4g5 b7b5 c3d5 e7d8 d5b4 a5c7 h2h4 d7e5 d2g2 a7a5 b4d3 e5c4 e3c1 c7a7 g2g3 d8c7 h4h5 d6d5 d3f4 c4e3 d1e1 e3f1 h1f1 d5e4 f3e4');
INSERT INTO evaluations VALUES(17696,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(17696,31,'cp 162',NULL);
INSERT INTO evaluations VALUES(17696,32,'cp 165',NULL);
INSERT INTO evaluations VALUES(17696,33,'cp 169',NULL);
INSERT INTO evaluations VALUES(17696,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(17696,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(17696,36,'cp 183','c2c4 e7e6 b1c3 f8b4 a1c1 g8e7 a2a3 b4d6 f4d6 c7d6 c4d5 e7d5 c3d5 e6d5 f1d3 c8e6 f3g1 e8g8 g1e2 a8c8 e2f4 d8d7 h2h4 c6e7 c1c8 f8c8 f4e6 d7e6 h4h5 f6f5 g2g3 e6d7 e1f1 h7h6 f1g2 b7b5 d3b1 g8h8 h1e1 d7c6 d1f3 c6d7 b1d3 h8g8 f3e2 g8h8');
INSERT INTO evaluations VALUES(20692,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(20692,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(20692,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20692,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20692,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(20692,35,'cp 7',NULL);
INSERT INTO evaluations VALUES(20692,36,'cp 12','f8e7 d3d4 e6e5 d4e5 d6e5 d1d8 e7d8 c1e3 b7b6 f1d3 f7f6 a2a3 g8e7 h2h4 h7h5 e1g1 a7a5 b2b3 c8g4 f3d2 c6d4 e3d4 c5d4 c3d5 e8g8 f1b1 e7d5 c4d5 d8e7 d3b5 f8d8 d2c4 e7c5 b3b4 a5b4 a3b4 a8a1 b1a1 c5b4 f2f3 g4d7 b5d7 d8d7 c4b6 d7b7 b6c4 g8h7');
INSERT INTO evaluations VALUES(21006,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(21006,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(21006,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(21006,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(21006,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(21006,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(21006,36,'cp -39','e7e5 f1g2 g8f6 d4e5 c6e5 g1f3 e5f3 e2f3 f8e7 e1g1 e8g8 f1e1 c8e6 g2f1 c7c5 b1a3 a7a6 a3c2 d8d7 c1f4 f8d8 f4e5 a8c8 a2a4 d5d4 c3d4 e6b3 d1e2 b3a4 d4c5 e7c5 e1d1 d7c6 d1d8 c8d8 e5f6 g7f6 c2e3 a4b5 e2c2');
INSERT INTO evaluations VALUES(21685,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(21685,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(21685,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(21685,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(21685,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(21685,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(21685,36,'cp -29','d4e5 c6e5 f1g2 g8f6 g1f3 e5f3 e2f3 f8d6 e1g1 e8g8 a2a4 h7h6 b1a3 c7c5 a3b5 d6b8 b5a3 c8e6 a3c2 f8e8 f1e1 b8d6 a4a5 d8c7 g2f1 a8d8 c1d2 c7d7 b2b4 b7b6 d2e3 e6h3 f1h3 d7h3 d1d2 h3d7 a5b6 a7b6 b4c5 b6c5 e3f4 e8e1 c2e1 d6f4');
INSERT INTO evaluations VALUES(21019,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(21019,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(21019,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(21019,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(21019,34,'cp -28',NULL);
INSERT INTO evaluations VALUES(21019,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(21019,36,'cp -9','e7e5 d4e5 c6e5 g1f3 e5f3 d1f3 g8f6 h2h3 g7g6 c2c4 f8g7 d2c3 e8g8 f1e2 d5c4 e1g1 d8e7 e2c4 c7c6 b1d2 c8f5 a1c1 a8d8 f1d1 f6e4 d2e4 f5e4 f3f4 g7c3 b2c3 g6g5 f4g3 e4g6 d1d4 b7b5 c4e2 c6c5 d4d8 f8d8 c1d1 d8d1 e2d1 b5b4 g3b8 g8g7');
INSERT INTO evaluations VALUES(21181,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(21181,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(21181,32,'cp 67',NULL);
INSERT INTO evaluations VALUES(21181,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(21181,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(21181,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(21181,36,'cp 56','c2c4 g7g6 d4d5 c6e5 e3e4 f8g7 b1c3 c7c6 f1e2 c6d5 c4d5 g8f6 g1f3 e5f3 e2f3 e8g8 e1g1 a7a6 c1e3 f6d7 a1c1 b7b5 b2b3 d8a5 c3e2 c8b7 e2d4 a8c8 d1e2 c8c1 f1c1 d7e5 d4c6 b7c6 d5c6');
INSERT INTO evaluations VALUES(23855,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(23855,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(23855,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(23855,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(23855,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(23855,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23855,36,'cp 13','h5g6 d1e2 h7h5 h1g1 g8e7 a2a3 c6d4 f3d4 c5d4 c1e3 d4c3 b2c3 d8d7 e1c1 h5g4 h3g4 f7f6 c1b2 e7c6 c4b5 a7a6 b5a4 e8c8 c3c4 g6e8 d1b1 d7e7 g4g5 c6d4 e3d4 e8a4 d4e3 b7b6 e2g4 a4d7 g5f6 g7f6 g4g7 e7e6 f2f4 e5f4');
INSERT INTO evaluations VALUES(23853,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(23853,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(23853,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(23853,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(23853,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(23853,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(23853,36,'cp 13','d1e2 g8e7 a2a3 h7h5 h1g1 c6d4 f3d4 c5d4 c1e3 d4c3 b2c3 d8d7 e1c1 h5g4 h3g4 f7f6 f2f4 e5f4 e3f4 e8c8 e2e3 e7c6 c4b5 a7a6 b5c6 d7c6 e3d4 c8b8 d1f1 c6d7 c3c4 h8e8 c1b2 b7b6 d4c3 b8b7 g4g5 f6f5 e4e5 d6d5 c4c5 d7c6 c5b6');
INSERT INTO evaluations VALUES(28732,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(28732,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(28732,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(28732,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(28732,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28732,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(28732,36,'cp 23','b1d2 f8e7 h2h3 e8g8 g1f3 c7c5 f1d3 d7d5 a2a3 e7d6 d1e2 a7a5 e1g1 d6f4 e3f4 b7a6 d3a6 a8a6 f1e1 d8c7 f3e5 a6a8 e2e3 b8d7 g2g3 f8c8 g1g2 c7b7 a3a4 b6b5 a4b5 b7b5 e5d7 f6d7 d2f3 b5b2 f4f5');
INSERT INTO evaluations VALUES(28731,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(28731,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(28731,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(28731,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(28731,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(28731,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(28731,36,'cp 9','c7c5 h2h3 d7d5 g1f3 b8c6 f1d3 f8d6 e1g1 e8g8 d1e2 c6e7 a2a4 a8c8 f1e1 a7a5 f4d6 d8d6 e3e4 d5e4 d2e4 f6e4 d3e4 b7e4 e2e4 f8d8 a1d1 h7h6 e4e2 e7f5 g2g3 d6c7 d4c5 c7c5 g1g2 c5c7 d1d8 c8d8 e1d1');
INSERT INTO evaluations VALUES(29244,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(29244,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(29244,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29244,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(29244,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29244,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29244,36,'cp 0','h7h6 d1b3 d8b6 c4c5 b6b3 a2b3 f8e7 b3b4 e8g8 e1g1 b8d7 h2h4 a7a6 f3d2 f5c2 d2f3');
INSERT INTO evaluations VALUES(29788,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(29788,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(29788,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(29788,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(29788,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(29788,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(29788,36,'cp 16','e1g1 e7e6 c2c4 c7c6 d1c2 b8d7 d3g6 h7g6 b1d2 d5c4 d2c4 c6c5 d4c5 f8c5 b2b3 b7b5 c4e5 a8c8 c2e2 d7e5 f3e5 d8b6 c1b2 e8g8 f1d1 b6b7 a1c1 c5e7 e2f3 b7a6 e5c6 f6d5 c6e7 d5e7 a2a3 c8c1 b2c1 a6b6 c1b2 f8d8 d1d8 b6d8');
INSERT INTO evaluations VALUES(34688,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(34688,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(34688,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(34688,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(34688,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(34688,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(34688,36,'cp 41','c7c6 g1f3 c8f5 f3h4 f5g6 g2g3 e7e6 h4g6 h7g6 d2d3 f8e7 e2f3 b8d7 c1d2 a8d8 c3d5 a5a4 b2b3 a4a3 d5e7 g8e7 e1g1 e7f5 f3g2 e8g8 f1e1 d7f6 d1c1 a3c1 a1c1 d8d7 a2a4 f8d8 a4a5 f6d5 c2c3 d5f6 a5a6 b7b6');
INSERT INTO evaluations VALUES(34977,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(34977,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(34977,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(34977,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(34977,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(34977,35,'cp 70',NULL);
INSERT INTO evaluations VALUES(34977,36,'cp 73','f8b4 c1d2 a5e5 d1e2 e5e2 g1e2 g8f6 e1c1 e8g8 c3b5 b4d2 d1d2 b8a6 e2c3 c7c6 b5d6 a6c5 a2a4 a7a5 f1c4 c8e6 c4e6 f7e6 h1e1 f8d8 f2f3 g8f8 e1e5 b7b6 d6c4 d8d2 c1d2 a8d8 d2c1 f6d7 e5e3 f8e7 b2b3 g7g5 c3e2 d8h8 e2d4 e7f6 c4d6 h8d8');
INSERT INTO evaluations VALUES(36047,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36047,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36047,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36047,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36047,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36047,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36047,36,'cp 0','c1e3 c5d4 c3d4 c8g4 e1g1 b8c6 d4d5 c6e5 a1c1 d8a5 e3g5 a5b4 c4d3 f8c8 c1b1 b4a3 d3e2 a3a2 f1e1 g4f3 g2f3 h7h6 g5e7 c8c2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(36045,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36045,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36045,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36045,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36045,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36045,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36045,36,'cp 0','c8g4 a1c1 c5d4 c3d4 b8c6 d4d5 c6e5 e1g1 d8a5 e3g5 a5b4 c4d3 f8c8 c1b1 b4a3 d3e2 a3a2 f1e1 g4f3 g2f3 h7h6 g5e7 c8c2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(38411,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(38411,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(38411,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(38411,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(38411,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(38411,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(38411,36,'cp 36','g8f6 f1d3 e8g8 g1f3 e7e5 e1g1 b8c6 b1d2 f6g4 e3g5 f7f6 g5h4 g4h6 f1e1 g8h8 a1c1 a7a5 d2f1 h6f7 f1e3 f7g5 d4e5 g5f3 d1f3 c6e5 f3e2 e5d3 e2d3 c8e6 c3c4 d8d7 h4g3 f6f5 e4f5 e6f5 e3f5 d7f5 d3f5 g6f5 e1e7 a5a4 e7c7 f5f4 g3h4 g7b2 c1d1');
INSERT INTO evaluations VALUES(41371,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(41371,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(41371,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(41371,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(41371,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(41371,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(41371,36,'cp 41','f1d3 e8g8 g1f3 e7e5 e1g1 d6d5 d4e5 f6e4 b1d2 c8f5 d3e4 d5e4 f3d4 b8c6 d4f5 g6f5 f2f4 e4f3 d2f3 c6e5 f3e5 d8d1 a1d1 g7e5 f1f5 f7f6 e3f4 e5f4 f5f4 a8d8 f4d4 d8d4 d1d4 f8e8 g1f2 e8e7 a2a4 b7b6 d4d5 g8f7 b2b4 f7e6 d5d8 a7a5 b4b5 f6f5 g2g3 e7d7 d8d7');
INSERT INTO evaluations VALUES(41377,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(41377,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(41377,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(41377,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(41377,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(41377,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(41377,36,'cp 32','e7e5 g1f3 e8g8 e1g1 b8c6 b1d2 f6g4 e3g5 f7f6 g5h4 g4h6 f1e1 g8h8 d4e5 d6e5 d2c4 c8e6 a2a4 d8e7 c4e3 a8d8 a4a5 a7a6 b2b4 h6f7 d1e2 f7d6 h2h3 e7f7 h4g3 g7h6 a1d1 e6b3 d3c2 b3c2 e2c2');
INSERT INTO evaluations VALUES(38449,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(38449,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(38449,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(38449,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(38449,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(38449,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(38449,36,'cp 78','c4d5 d8d5 g1f3 g8f6 b1c3 d5d8 f1c4 c7c5 d4d5 e8g8 e1g1 a7a6 a2a4 c8g4 e3e4 f6e8 c4e2 g4f3 e2f3 b8d7 f3g4 e8d6 d1e2 g7d4 g4h3 d7e5 g1h1 c5c4 f2f4 e5d3 a1b1 b7b5 e4e5 d6f5 h3f5 g6f5 a4b5 a6b5');
INSERT INTO evaluations VALUES(38535,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(38535,31,'cp -198',NULL);
INSERT INTO evaluations VALUES(38535,32,'cp -171',NULL);
INSERT INTO evaluations VALUES(38535,33,'cp -196',NULL);
INSERT INTO evaluations VALUES(38535,34,'cp -192',NULL);
INSERT INTO evaluations VALUES(38535,35,'cp -162',NULL);
INSERT INTO evaluations VALUES(38535,36,'cp -189','h7h6 b1c3 b8c6 c3d5 c6e7 d1d3 c7c6 d5f4 g5f4 c1f4 e7g6 f4g3 g8e7 f3e5 g6e5 g3e5 e8g8 e5f6 g7f6 f1f6 e7f5 f6f5 c8f5 d3g3 d8g5 e4f5 a8e8 c4d3 e8e3 g3f2 f8e8 f5f6 g5g4 d4d5 c6d5 a1f1 g4g5 h2h4 g5g4 f2f4 g4f4 f1f4 e3e5 g1f2 e8e6 f2f3');
INSERT INTO evaluations VALUES(38630,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(38630,31,'cp -72',NULL);
INSERT INTO evaluations VALUES(38630,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(38630,33,'cp -77',NULL);
INSERT INTO evaluations VALUES(38630,34,'cp -70',NULL);
INSERT INTO evaluations VALUES(38630,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(38630,36,'cp -81','d2d4 b8c6 f1d3 g8f6 c2c3 e8g8 f3e5 c6e7 c1e3 c8f5 d1e2 f5d3 e2d3 d8c8 b1d2 c8f5 d3e2 f5c2 d2f3 c2e2 e1e2 h7h5 a2a4 e7f5 h1e1 f8e8 h2h3 e8e6 a4a5 a8e8 f3g5 e6e7 g2g4 h5g4 h3g4 f5e3 e2e3 f6g4 e3d2');
INSERT INTO evaluations VALUES(38614,30,'cp -87',NULL);
INSERT INTO evaluations VALUES(38614,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(38614,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(38614,33,'cp -82',NULL);
INSERT INTO evaluations VALUES(38614,34,'cp -66',NULL);
INSERT INTO evaluations VALUES(38614,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(38614,36,'cp -73','g8f6 f1b5 c7c6 d1e2 d6e7 b5d3 e8g8 c1e3 f6g4 e1g1 c6c5 d4c5 g4e3 e2e3 b8d7 g1h1 e7c5 e3e1 f8e8 e1g3 d7f6 f3e5 c8e6 a2a3 a8c8 b1c3 a7a6 a1e1 c8c7 c3e2 c5d6 e2d4 e6c8 d3f5 f6e4 f5e4 d5e4 e1e4 f7f6 g3b3 g8h8');
INSERT INTO evaluations VALUES(39221,30,'cp -4',NULL);
INSERT INTO evaluations VALUES(39221,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(39221,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(39221,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(39221,34,'cp -8',NULL);
INSERT INTO evaluations VALUES(39221,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(39221,36,'cp -8','g8f6 f1g2 d7d5 c4d5 f6d5 g1f3 b8c6 e1g1 e8g8 b1c3 d5f6 h2h3 d8d3 b2b3 c8f5 g3g4 f5g6 f3h4 a8d8 c1b2 f8e8 g4g5 f6e4 h4g6 e4c3 b2c3 d3g6 d1g4 e5e4 f2f4 e4f3 g4f3 c5e7 f3f5 e7g5 f5b5 d8b8 a1d1 g5h6 b5f5 g6f5 f1f5');
INSERT INTO evaluations VALUES(42601,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42601,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42601,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(42601,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42601,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42601,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(42601,36,'cp -12','f1g2 d7d5 c4d5 f6d5 d2d4 c5b4 c1d2 e5d4 e3d4 b4d2 d1d2 e8g8 g1e2 c8e6 e1g1 b8c6 b1c3 d8d7 c3e4 b7b6 e4g5 e6f5 e2c3 d5c3 b2c3 h7h6 g5f3 f5e6 a1d1 a8d8 f1e1 e6d5 f3e5 c6e5 g2d5 d7d5 e1e5 d5e5 d4e5 d8d2 d1d2 f8e8 d2d7 e8e5 d7c7 e5e1 g1g2');
INSERT INTO evaluations VALUES(39226,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(39226,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(39226,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(39226,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(39226,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(39226,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(39226,36,'cp 36','c7c6 b2b4 c5e7 c1b2 d7d6 d2d4 a7a5 b4b5 e5d4 b2d4 g8f6 f1d3 b8d7 g1f3 e8g8 h2h3 d7c5 d4c5 d6c5 b1c3 e7d6 e1g1 g7g6 d1c2 d8e7 f1e1 f8e8 b5c6 b7c6 c2a4 e7c7 a1b1 c8d7 a4c2 d6f8 d3e2 a8b8 e1d1 b8b1');
INSERT INTO evaluations VALUES(39757,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(39757,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(39757,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(39757,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(39757,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(39757,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(39757,36,'cp 21','c3d5 b8d7 f1e1 e8g8 a2a4 e7e6 d5f6 d7f6 c4f1 c8d7 c2c4 f8e8 h2h3 d7c6 d4c6 b7c6 a1a3 d6d5 e4d5 c6d5 c4d5 f6d5 a3b3 d8a5 e1e4 e8d8 d1e2 d5e7 c1e3 e7f5 e3b6 a5d5 e4c4 f5d4 b6d4 g7d4');
INSERT INTO evaluations VALUES(40246,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(40246,31,'cp 61',NULL);
INSERT INTO evaluations VALUES(40246,32,'cp 81',NULL);
INSERT INTO evaluations VALUES(40246,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(40246,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(40246,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(40246,36,'cp 60','d4c5 e8g8 e2e3 h7h6 g5h4 c8e6 f1b5 f6e4 c3e4 e7h4 e4c3 h4e7 e1g1 e7c5 a1c1 c5b6 c3a4 b8d7 b5d7 e6d7 a4b6 d8b6 d1d2 d7e6 a2a4 f8c8 a4a5 b6d6 f3d4 e6d7 f1d1 c8c1 d2c1');
INSERT INTO evaluations VALUES(40244,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(40244,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(40244,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(40244,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(40244,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(40244,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(40244,36,'cp 54','e8g8 e2e3 h7h6 g5h4 c8e6 f1b5 a7a6 b5a4 b8d7 a4d7 f6d7 h4e7 d8e7 e1g1 e7c5 h2h3 a8c8 c3e2 c5b4 b2b3 f8e8 f3d4 d7f6 a2a3 b4d6 d1b1 h6h5 a3a4 e6d7 b1b2 f6e4 f1c1 g7g5 e2c3 e4c3 c1c3 c8c3 b2c3 d6e5 a1d1 e8c8');
INSERT INTO evaluations VALUES(44505,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(44505,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(44505,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(44505,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(44505,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(44505,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44505,36,'cp -12','e7e6 a2a3 f8c5 g1f3 e8g8 c2c4 b7b6 f1e2 a7a5 e1g1 c8b7 b1c3 d7d5 c4d5 f6d5 c1d2 b8d7 a1c1 d8e7 c3d5 b7d5 e2c4 d5b7 d2c3 f8d8 d1e2 d7f6 c3e5 h7h6 h2h3 f6e4 f1d1 d8d1 c1d1 a8d8 d1d8 e7d8 c4d3 d8d7 f3d2 e4f6 d3b5 d7e7 d2b1');
INSERT INTO evaluations VALUES(45172,30,'cp 102',NULL);
INSERT INTO evaluations VALUES(45172,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(45172,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(45172,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(45172,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(45172,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(45172,36,'cp 102','d4c5 d8a5 b1c3 f6e4 c5d6 e4c3 b2c3 e7d6 e2e3 f8e7 f1e2 b8d7 f3d4 d7c5 e1g1 e8g8 d1c2 c8d7 e2f3 a8c8 a1b1 e7f6 h2h3 f8d8 b1b4 a5a6 c1d2 d8e8 c2b1 b7b6 f3d5 d7e6 d4b5 f6e5 e3e4 c5a4 g1h1 a6a5 f2f3 e6d7 d2e3 a4c3');
INSERT INTO evaluations VALUES(46038,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46038,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(46038,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(46038,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(46038,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(46038,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(46038,36,'cp -7','b8c6 d1e2 a7a6 f1d1 d8c7 c1f4 f8e7 a1c1 e8g8 c4b3 c7b8 c3a4 b7b5 c1c6 b5a4 b3a4 c8d7 c6c4 d7b5 a4b5 a6b5 c4c6 a8a2 f4d6 e7d6 c6d6 b8b7 d6d4 a2a4 d4a4 b5a4 d1d4 f8a8 g2g3 h7h6 g1g2 a4a3 b2a3 a8a3 e2c2 a3a5 d4c4 f6d7 h2h3');
INSERT INTO evaluations VALUES(46407,30,'cp 88',NULL);
INSERT INTO evaluations VALUES(46407,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(46407,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(46407,33,'cp 93',NULL);
INSERT INTO evaluations VALUES(46407,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(46407,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(46407,36,'cp 94','g5f6 d8f6 c3d5 f6d8 e2e4 b8c6 g1f3 c5d4 f1b5 a7a6 b5a4 d8a5 f3d2 f8d6 e1g1 e8g8 d2c4 a5d8 a1c1 c8e6 c4d6 d8d6 a4c6 b7c6 c1c6 d6e5 c6c5 e5e4 f1e1 e4g4 f2f3 g4h4 e1e4 h4h6 e4d4 e6d5 d4d5 h6b6 d1d4 a8d8 h2h4 f8e8 g1h2 h7h6 b2b4 d8d5 c5d5 b6d4 d5d4');
INSERT INTO evaluations VALUES(46736,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(46736,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46736,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46736,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46736,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(46736,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46736,36,'cp 0','b8a6 e2e3 f8g7 b5c3 a6b8 g1f3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(14124,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14124,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14124,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14124,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14124,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14124,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14124,36,'cp 0','b5c3 a6b8 e2e3 f8g7 g1f3 e8g8 c3b5 b8a6 b5c3');
INSERT INTO evaluations VALUES(47458,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(47458,31,'cp -83',NULL);
INSERT INTO evaluations VALUES(47458,32,'cp -61',NULL);
INSERT INTO evaluations VALUES(47458,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(47458,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(47458,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(47458,36,'cp -67','e7e5 g2g3 c8e6 f1g2 b8c6 e1g1 d8d7 d2d4 d5c3 b2c3 e8c8 d1d3 e6h3 c1e3 h7h5 g2h1 h3f1 g1f1 h5h4 a1b1 a7a6 g3g4 d7g4 h1c6 b7c6 d3a6 c8d7 d4e5 g4e4 e3g5 f7f6 b1d1 d7e8 d1d8 e8d8 e5f6 d8d7 a6d3 e4d3 c2d3 h4h3 e2f4 g7f6 g5f6 h8h7 f1e2 f8d6 f6g5 h7h8 e2f3 d6f4 g5f4');
INSERT INTO evaluations VALUES(47988,30,'cp 635',NULL);
INSERT INTO evaluations VALUES(47988,31,'cp 650',NULL);
INSERT INTO evaluations VALUES(47988,32,'cp 663',NULL);
INSERT INTO evaluations VALUES(47988,33,'cp 664',NULL);
INSERT INTO evaluations VALUES(47988,34,'cp 629',NULL);
INSERT INTO evaluations VALUES(47988,35,'cp 663',NULL);
INSERT INTO evaluations VALUES(47988,36,'cp 652','e5f7 d8h4 e1g1 e4f6 d2d4 b8c6 c1g5 h4d4 b1d2 h8g8 f1e1 f8e7 g5f6 g7f6 f7d6 c7d6 c4g8 c8g4 d1c1 f6f5 d2b3 d4g7 g8d5 e8f8 c1d2 e7g5 f2f4 g5h4 d5c6 b7c6 d2d6 f8g8 d6c6 a8f8 e1e8 f8e8 c6e8 g7f8 e8f8 g8f8 h2h3 g4h5 b3d4 h4f6 d4e6 f8g8');
INSERT INTO evaluations VALUES(47556,30,'cp 632',NULL);
INSERT INTO evaluations VALUES(47556,31,'cp 644',NULL);
INSERT INTO evaluations VALUES(47556,32,'cp 629',NULL);
INSERT INTO evaluations VALUES(47556,33,'cp 629',NULL);
INSERT INTO evaluations VALUES(47556,34,'cp 638',NULL);
INSERT INTO evaluations VALUES(47556,35,'cp 623',NULL);
INSERT INTO evaluations VALUES(47556,36,'cp 655','d8h4 e1g1 e4f6 d2d4 b8c6 c1g5 h4d4 b1d2 c8g4 d1e1 e8d7 c2c3 d4c5 g5f6 a8e8 e1b1 g7f6 f7h8 c5f5 c4d3 f5f4 d3h7 f8g7 g2g3 f4d2 h7f5 g4f5 b1f5 d7d8 h8g6 d2g5 f5g5 f6g5 a1e1 e8e1 f1e1 g5g4 e1d1 g7c3 b2c3 d8d7 d1e1 d6d5 g6e5 c6e5');
INSERT INTO evaluations VALUES(48215,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(48215,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(48215,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(48215,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(48215,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48215,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48215,36,'cp 0','b8d7 e2e4 h7h6 g5f6 d7f6 e4e5 f6d7 a1c1 f8e7 c2c4 d5c4 f1c4 d7b6 c4d3 c8d7 d3e4 b6d5 e1g1 d7b5 f1e1 d8d7 a2a3 b5a4 e4c2 a4b5');
INSERT INTO evaluations VALUES(49458,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(49458,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(49458,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(49458,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(49458,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(49458,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(49458,36,'cp -38','f8e7 b2b3 d7d5 a2a4 c7c5 g2g3 e8g8 f1g2 b7b6 d1e2 c8b7 d2d3 b8c6 e1g1 f6d7 c1b2 e7f6 b1c3 a7a6 g3g4 g7g6 a1e1 d8e7 g4g5 f6g7 c3d1 g7b2 d1b2 a8e8 e3e4 e6e5 f4f5 g6f5 e4f5');
INSERT INTO evaluations VALUES(50042,30,'cp 203',NULL);
INSERT INTO evaluations VALUES(50042,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(50042,32,'cp 201',NULL);
INSERT INTO evaluations VALUES(50042,33,'cp 212',NULL);
INSERT INTO evaluations VALUES(50042,34,'cp 230',NULL);
INSERT INTO evaluations VALUES(50042,35,'cp 208',NULL);
INSERT INTO evaluations VALUES(50042,36,'cp 209','d2d4 d7d6 f1c4 b8c6 c1e3 a7a6 a2a4 c8g4 d4d5 c6b8 h2h3 g4h5 g2g4 h5f7 h3h4 d8c8 h4h5 c8g4 c4e2 g4c8 f3h4 f6f5 h4f5 e7f5 e4f5 b8d7 d1d3 d7f6 h5h6 g7g6 e3g5 f8e7 g5f6 e7f6 e2g4 c8d8 e1c1 e8g8 c3e4 c7c6 d5c6 b7c6 d3f3');
INSERT INTO evaluations VALUES(50040,30,'cp 198',NULL);
INSERT INTO evaluations VALUES(50040,31,'cp 223',NULL);
INSERT INTO evaluations VALUES(50040,32,'cp 213',NULL);
INSERT INTO evaluations VALUES(50040,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(50040,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(50040,35,'cp 212',NULL);
INSERT INTO evaluations VALUES(50040,36,'cp 214','d7d6 f1c4 b8c6 c1e3 a7a6 a2a3 c8g4 d4d5 c6b8 c4e2 g4h5 d1d3 b8d7 h2h4 g7g6 e1c1 h7h6 f3d2 h5e2 d3e2 f6f5 g2g3 f8g7 f2f3 d7f6 g3g4 f5g4 f3g4 d8d7 d1g1 h6h5 g4h5 f6h5 d2f3 e8c8 f3g5 c8b8 g5e6 d8e8 e6g7 h5g7 e3f2 a6a5 c1b1');
INSERT INTO evaluations VALUES(50051,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(50051,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(50051,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(50051,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(50051,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(50051,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(50051,36,'cp 102','a7a6 d1c2 d7d5 g1f3 d5c4 e2e3 b8d7 f1c4 b7b5 c4e2 c8b7 e1g1 e7d5 e3e4 d5f6 e4e5 f6d5 c1g5 f8e7 c3d5 b7d5 g5e7 d8e7 c2c7 e8g8 f1c1 e7d8 f3e1 d7b6 c7d8 f8d8 g2g3 a8c8 c1c8 d8c8 f2f3 h7h6 b2b3 g7g5 e1d3 c8c3 g1f2 c3c2 d3b4');
INSERT INTO evaluations VALUES(51132,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(51132,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(51132,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(51132,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(51132,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(51132,35,'cp 88',NULL);
INSERT INTO evaluations VALUES(51132,36,'cp 79','c4b3 f8e7 a2a4 b5b4 c3d5 g8f6 d2d4 e5d4 f3d4 c7c5 d4f5 c8f5 e4f5 b8c6 c1f4 e8g8 e1g1 c6d4 b3c4 d4f5 c2c3 b4c3 b2c3 f6d5 d1d5 g7g6 h2h3 e7g5 f4h2 d8f6 g2g4 f5e7 d5d6 f8d8 d6f6 g5f6 a1b1 f6c3 b1b7');
INSERT INTO evaluations VALUES(51137,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(51137,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(51137,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(51137,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(51137,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(51137,35,'cp 86',NULL);
INSERT INTO evaluations VALUES(51137,36,'cp 81','f8e7 a2a4 b5b4 c3d5 g8f6 d2d4 e5d4 d1d4 e8g8 d5e7 d8e7 c1g5 b8c6 g5f6 e7f6 d4f6 g7f6 b3d5 c8d7 e1c1 a8b8 b2b3 c6e5 f3d2 e5g4 d1f1 f6f5 e4f5 d7f5 h2h3 g4f6 d5c4 a6a5 g2g4 f5d7 c4d3 b8e8 d2c4 f6d5 c4a5 d5f4 f1e1 g8g7 d3c4 e8e1 h1e1 f4h3');
INSERT INTO evaluations VALUES(53545,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(53545,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(53545,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(53545,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(53545,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(53545,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(53545,36,'cp -29','e6d5 f1d3 f8d6 g1e2 g8f6 b1c3 e8g8 e1g1 f8e8 f2f3 c6c5 g1h1 b8c6 d4c5 d6c5 e3e4 d5e4 c3e4 f6e4 d3e4 d8d1 f1d1 f7f5 e4d3 c8e6 c1d2 g7g6 e2f4 e6f7 d3b5 a8d8 b2b3 d8d7 d2c3 e8d8 d1d7 d8d7');
INSERT INTO evaluations VALUES(54629,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(54629,31,'cp 233',NULL);
INSERT INTO evaluations VALUES(54629,32,'cp 234',NULL);
INSERT INTO evaluations VALUES(54629,33,'cp 223',NULL);
INSERT INTO evaluations VALUES(54629,34,'cp 209',NULL);
INSERT INTO evaluations VALUES(54629,35,'cp 211',NULL);
INSERT INTO evaluations VALUES(54629,36,'cp 212','e2e4 f8b4 f1d3 d5e4 d3e4 g8e7 e1g1 e8g8 d1b3 b4a5 f1d1 b8d7 e4c2 d8b6 c1f4 b6b4 h2h3 e7g6 f4g3 b4b3 a2b3 a5b4 h3h4 f8e8 c3a2 b4e7 c4c5 f6f5 a2c3 e6e5 h4h5 f5f4 g3h2 g6h4 f3h4 e7h4 c3e4 d7f6 e4d6 e8e7 d4e5 e7e5 b3b4 c8g4 c2b3 f6d5 b3d5 e5d5');
INSERT INTO evaluations VALUES(55641,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(55641,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(55641,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(55641,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(55641,34,'cp 15',NULL);
INSERT INTO evaluations VALUES(55641,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(55641,36,'cp 25','c2c4 g7g6 b1c3 f8g7 d3d4 e5d4 f3d4 g8e7 e2e4 f7f5 c1g5 h7h6 g5e3 f5e4 c3e4 b8c6 f1e2 e8g8 d1d2 d6d5 d4c6 b7c6 e4g3 a8b8 e3h6 b8b2 d2c1 g7h6 c1b2 h6g7 b2c1 e7f5 e1g1 g7a1 c1a1 f5g3 h2g3 d8f6');
INSERT INTO evaluations VALUES(56833,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56833,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56833,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56833,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56833,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56833,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56833,36,'cp 13','c2c4 d5c4 d1a4 b8c6 a4c4 d8d6 a2a3 c8e6 c4a4 e8c8 b1c3 c8b8 b2b4 d6d7 e1g1 h7h5 d2d3 g8e7 b4b5 c6d4 f3d4 d7d4 a4d4 d8d4 h2h4 e7f5 g2h3 d4d8 c3e4 e6c8 h3f5 c8f5 c1e3 b7b6 f1c1 f5g4 f2f3 g4d7 a3a4');
INSERT INTO evaluations VALUES(450,30,'cp 93',NULL);
INSERT INTO evaluations VALUES(450,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(450,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(450,33,'cp 91',NULL);
INSERT INTO evaluations VALUES(450,34,'cp 66',NULL);
INSERT INTO evaluations VALUES(450,35,'cp 72',NULL);
INSERT INTO evaluations VALUES(450,36,'cp 70','c1d2 f7f6 e5f6 d7f6 e1g1 e8g8 d2b4 b6b4 d1c2 c8d7 a2a3 b4d6 e2g3 e6e5 d4e5 c6e5 f3e5 d6e5 a1e1 e5d4 e1e7 a8c8 c2d2 d4c5 e7e2 c8e8 h2h3 c5b6 d3b1 e8e2 g3e2 d7e8 f1e1 e8f7 e2d4 f8e8 e1e8 f7e8 b1a2 g8h8 d4f3 e8g6 g2g3');
INSERT INTO evaluations VALUES(449,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(449,31,'cp 85',NULL);
INSERT INTO evaluations VALUES(449,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(449,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(449,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(449,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(449,36,'cp 82','f7f6 e5f6 d7f6 a2a3 b4d2 d1d2 f6e4 d2c1 c8d7 e1g1 e8g8 c1e3 c6e7 d3e4 d5e4 e3e4 e7g6 b2b4 a8e8 f3e5 g6e5 e4e5 f8f5 e5e3 d7c6 e2c3 a7a5 a1b1 a5b4 a3b4 c6d5 f2f3 h7h6 f1d1 e8c8 h2h3 b6d6 g1h1 c8f8 c3e2 d5c4 d1c1 c4e2 e3e2 b7b6 c1c2');
INSERT INTO evaluations VALUES(3142,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(3142,31,'cp -109',NULL);
INSERT INTO evaluations VALUES(3142,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(3142,33,'cp -86',NULL);
INSERT INTO evaluations VALUES(3142,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(3142,35,'cp -84',NULL);
INSERT INTO evaluations VALUES(3142,36,'cp -75','b7c6 e1c1 d8c7 e3h6 g7h6 d2h6 d5e4 f3e4 c8e6 f1e2 a8b8 h6e3 c6c5 h2h4 h7h5 c3d5 e6d5 e4d5 b8b4 a2a3 b4e4 e3f2 f8d8 e2f3 e4f4 f2g3 c7d6 h1e1 d8d7 c1b1 a7a5 b1a2 g8g7 c2c3 f4f5 g3f2 a5a4 f2e2 f5f4 e2f2 f6g4 f2c2 d7b7 f3g4 f4g4');
INSERT INTO evaluations VALUES(3466,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(3466,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(3466,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(3466,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(3466,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(3466,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(3466,36,'cp 62','h2h3 f6d7 f1b5 f8e8 b5c6 b7c6 d1a4 e7e5 f3e5 d7e5 f4e5 g7e5 d4e5 e8e5 d2f3 e5e6 e1g1 e6d6 c3c4 d5c4 a4c4 d8f6 f3d4 f6g5 f2f4 g5f6 a1c1 c8f5 d4f5 g6f5 c1c3 a8d8 b2b3 a7a6 c3c1 h7h6 f1e1 f6b2 e3e4');
INSERT INTO evaluations VALUES(4888,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(4888,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(4888,32,'cp 86',NULL);
INSERT INTO evaluations VALUES(4888,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(4888,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(4888,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(4888,36,'cp 92','b5a4 f7f5 d2d4 e5d4 b1d2 d4c3 b2c3 d6f4 e1g1 g8e7 e4f5 d7d5 g2g3 f4d2 d1d2 c8f5 c1a3 e8g8 a1e1 f8e8 e1e3 e7g6 e3e8 d8e8 f1e1 e8d7 h2h4 f5g4 f3h2 g4e6 h4h5 b7b5 a4b3 g6e5 h5h6 a8e8 d2g5 e5g6 h6g7 g8g7 a3b2 e6h3 e1e8 d7e8');
INSERT INTO evaluations VALUES(4892,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(4892,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(4892,32,'cp 77',NULL);
INSERT INTO evaluations VALUES(4892,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(4892,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(4892,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(4892,36,'cp 64','f7f5 d2d4 e5d4 b1d2 d4c3 b2c3 g8e7 e1g1 d6f4 e4f5 d7d5 g2g3 f4d2 c1d2 c8f5 d2g5 e8g8 a4c6 b7c6 f1e1 f5e4 f3d2 f8f7 d2e4 d5e4 d1d8 a8d8 e1e4 d8d7 g5f4 e7d5 f4e5 d5b6 a1e1 c6c5 e1e2 d7d1 g1g2 f7e7 h2h4 d1c1 h4h5 h7h6 f2f4 b6d5 e2e1 c1c2 e4e2 c2e2 e1e2 d5f6 g2f3');
INSERT INTO evaluations VALUES(5505,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(5505,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5505,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5505,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5505,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5505,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5505,36,'cp 0','d3d4 c5d4 f3d4 g8f6 f1e2 e8g8 e1g1 c8d7 a2a4 a7a6 h2h3 a8c8 f1e1 d8c7 d4c6 d7c6 e2f3 f8e8 a4a5 f6d7 e3d4 g7h6 d4e3');
INSERT INTO evaluations VALUES(5881,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(5881,31,'cp -183',NULL);
INSERT INTO evaluations VALUES(5881,32,'cp -187',NULL);
INSERT INTO evaluations VALUES(5881,33,'cp -159',NULL);
INSERT INTO evaluations VALUES(5881,34,'cp -186',NULL);
INSERT INTO evaluations VALUES(5881,35,'cp -179',NULL);
INSERT INTO evaluations VALUES(5881,36,'cp -172','e3f2 e1e2 d8d1 e2d1 c6b4 b1c3 c8e6 d1e2 e8c8 e2f2 g8e7 b2b3 h8e8 h2h4 e6g4 a1d1 d8d1 c3d1 b4a2 d1e3 g4d7 h4h5 h7h6 h1h4 a2b4 h4f4 d7e6 g2g4 g7g5 h5g6 e7g6 f4f6 b4c6 f2g3 a7a6 f1e2 c6e5 f3e5 g6e5 f6h6 a6a5 g4g5 e5g6 e2g4 e6g4');
INSERT INTO evaluations VALUES(5883,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(5883,31,'cp -162',NULL);
INSERT INTO evaluations VALUES(5883,32,'cp -170',NULL);
INSERT INTO evaluations VALUES(5883,33,'cp -171',NULL);
INSERT INTO evaluations VALUES(5883,34,'cp -162',NULL);
INSERT INTO evaluations VALUES(5883,35,'cp -169',NULL);
INSERT INTO evaluations VALUES(5883,36,'cp -164','e1e2 d8d1 e2d1 c6b4 b1c3 c8g4 d1e2 e8c8 e2f2 g8e7 a1e1 e7g6 h2h4 g4f3 g2f3 h8e8 e5e6 g6f4 f2g3 f4h5 g3g4 h5f6 g4g3 f7e6 h4h5 b4c6 h1h4 c8b8 f1h3 e6e5 c3e4 a7a5 b2b3 e8g8 e4f6 g7f6 h4g4 c6b4 e1e2 b4d3 e2d2 c7c6 g4g8 d8g8 g3h4');
INSERT INTO evaluations VALUES(5882,30,'cp -175',NULL);
INSERT INTO evaluations VALUES(5882,31,'cp -176',NULL);
INSERT INTO evaluations VALUES(5882,32,'cp -169',NULL);
INSERT INTO evaluations VALUES(5882,33,'cp -189',NULL);
INSERT INTO evaluations VALUES(5882,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(5882,35,'cp -168',NULL);
INSERT INTO evaluations VALUES(5882,36,'cp -177','d8d1 e2d1 c6b4 b1c3 c8g4 d1e2 e8c8 e2f2 g8e7 a1e1 h8e8 e1e4 h7h5 a2a3 b4c6 e5e6 f7e6 f1e2 e7f5 h1d1 f5d6 e4e3 c6a5 c4c5 d6c4 e2c4 a5c4 e3e2 d8d1 c3d1 e8d8 d1e3 g4f3 f2f3 d8d3 f3e4 d3e3 e2e3 c4e3 e4e3 c8d7 e3e4 a7a5 b2b4 a5b4 a3b4 h5h4 b4b5 d7e7');
INSERT INTO evaluations VALUES(185,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(185,31,'cp -164',NULL);
INSERT INTO evaluations VALUES(185,32,'cp -169',NULL);
INSERT INTO evaluations VALUES(185,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(185,34,'cp -177',NULL);
INSERT INTO evaluations VALUES(185,35,'cp -172',NULL);
INSERT INTO evaluations VALUES(185,36,'cp -164','e2d1 c6b4 b1c3 c8g4 d1e2 e8c8 e2f2 g8e7 a1e1 h8e8 e1e4 h7h5 a2a3 b4c6 e5e6 g4e6 f1e2 e7g8 h2h3 g8f6 e4h4 e6f5 h4f4 f5g6 h1d1 d8d1 e2d1 c6e5 f3e5 e8e5 b2b4 c7c6 c4c5 c8d7 f4d4 d7e7 f2f3 f6d7 f3f2 a7a5 d1e2 d7f6 e2f3 g6f5');
INSERT INTO evaluations VALUES(186,30,'cp -162',NULL);
INSERT INTO evaluations VALUES(186,31,'cp -169',NULL);
INSERT INTO evaluations VALUES(186,32,'cp -156',NULL);
INSERT INTO evaluations VALUES(186,33,'cp -169',NULL);
INSERT INTO evaluations VALUES(186,34,'cp -174',NULL);
INSERT INTO evaluations VALUES(186,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(186,36,'cp -160','c6b4 b1c3 c8g4 d1e2 e8c8 e2f2 g8e7 a1e1 h8e8 e1e4 h7h5 a2a3 b4c6 e5e6 g4e6 f1e2 e7g8 h2h3 g8f6 e4h4 e6f5 h4f4 f5e4 h1d1 e4f3 g2f3 d8d1 e2d1 e8d8 f2e3 c6e5 d1e2 e5g6 f4d4 d8e8 e3f2 c7c6 e2d3 g6f8 d4f4 e8e5 b2b4 c8c7 f4f5 e5f5');
INSERT INTO evaluations VALUES(216,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(216,31,'cp -163',NULL);
INSERT INTO evaluations VALUES(216,32,'cp -161',NULL);
INSERT INTO evaluations VALUES(216,33,'cp -161',NULL);
INSERT INTO evaluations VALUES(216,34,'cp -161',NULL);
INSERT INTO evaluations VALUES(216,35,'cp -171',NULL);
INSERT INTO evaluations VALUES(216,36,'cp -162','b1c3 c8g4 d1e2 e8c8 e2f2 g8e7 a1e1 h8e8 e1e4 h7h5 a2a3 b4c6 e5e6 g4e6 f1e2 e7g8 h2h3 g8f6 e4h4 e6f5 h4f4 f5g6 h1d1 d8d1 e2d1 c6d8 f3d4 d8e6 d4e6 e8e6 c4c5 c7c6 b2b4 e6e5 f4f3 c8d7 f3e3 e5f5 f2g3 f5g5 g3h2 h5h4 d1f3 a7a5 c3a4 a5b4 a3b4');
INSERT INTO evaluations VALUES(10776,30,'cp 404',NULL);
INSERT INTO evaluations VALUES(10776,31,'cp 394',NULL);
INSERT INTO evaluations VALUES(10776,32,'cp 409',NULL);
INSERT INTO evaluations VALUES(10776,33,'cp 397',NULL);
INSERT INTO evaluations VALUES(10776,34,'cp 385',NULL);
INSERT INTO evaluations VALUES(10776,35,'cp 431',NULL);
INSERT INTO evaluations VALUES(10776,36,'cp 409','f3d4 f6e4 b1c3 e4c3 b2c3 e8g8 e1g1 d7d5 c4d3 f8e8 d1f3 c5d6 c1e3 c7c5 d4f5 d6f8 a1d1 c8e6 e3d2 h7h6 d3b5 e6d7 b5d7 d8d7 f3g4 g8h7 f5h6 d7g4 h6g4 b7b5 d2g5 e8e2 d1d5 e2a2 d5d7 b5b4 c3b4 c5b4 d7f7');
INSERT INTO evaluations VALUES(10775,30,'cp 415',NULL);
INSERT INTO evaluations VALUES(10775,31,'cp 418',NULL);
INSERT INTO evaluations VALUES(10775,32,'cp 406',NULL);
INSERT INTO evaluations VALUES(10775,33,'cp 397',NULL);
INSERT INTO evaluations VALUES(10775,34,'cp 416',NULL);
INSERT INTO evaluations VALUES(10775,35,'cp 413',NULL);
INSERT INTO evaluations VALUES(10775,36,'cp 392','f6e4 b1c3 e4c3 b2c3 e8g8 e1g1 d7d5 c4d3 f8e8 d1f3 c5d6 c1e3 d6e5 a1e1 c7c5 d3b5 e8e7 d4e2 d8d6 e2f4 c8e6 e1d1 a8d8 b5c4 e5f4 e3f4 d6c6 c4d3 d8e8 f4d2 e6d7 f1e1 e7e6 e1e3 c6a4 e3e6 e8e6 d1e1 h7h6 e1e6 d7e6');
INSERT INTO evaluations VALUES(10808,30,'cp 411',NULL);
INSERT INTO evaluations VALUES(10808,31,'cp 424',NULL);
INSERT INTO evaluations VALUES(10808,32,'cp 380',NULL);
INSERT INTO evaluations VALUES(10808,33,'cp 394',NULL);
INSERT INTO evaluations VALUES(10808,34,'cp 412',NULL);
INSERT INTO evaluations VALUES(10808,35,'cp 425',NULL);
INSERT INTO evaluations VALUES(10808,36,'cp 432','b1c3 e4c3 b2c3 e8g8 e1g1 d7d5 c4d3 f8e8 d1f3 a7a6 c1f4 c5f8 f3g3 c7c5 d4f3 c8d7 a1e1 d8b6 e1b1 b6c6 f3e5 c6c8 e5d7 c8d7 f1d1 b7b5 g3f3 d7e6 f4e3 e6e5 g2g3 e5c3 b1c1 c3b4 d3c2 a8d8 d1d5 e8e3 f2e3 d8d5');
INSERT INTO evaluations VALUES(11503,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(11503,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(11503,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11503,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11503,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11503,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11503,36,'cp 0','b5c6 d7c6 d1c2 d8e7 d2d3 c8g4 c1g5 g4f3 g2f3 e8c8 b1d2 e7e6 h1g1 c6c5 c2b3 e6e8 b3c2');
INSERT INTO evaluations VALUES(11494,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(11494,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(11494,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11494,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(11494,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11494,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(11494,36,'cp -7','d7c6 d1c2 d8e7 d2d3 c8g4 c1e3 g4f3 g2f3 e8c8 b1d2 f6d7 e1c1 f7f5 e4f5 e7f7 c2a4 f7f5 a4g4 f5f7 g4c4 f7e8 h1g1 d7f6 d2e4 c8b8 e4f6 g7f6 f3f4 f8h6 f4e5 h6e3 f2e3 e8e5 c4f4 e5e7 c1c2 h7h5 e3e4 h8g8 g1g8 d8g8');
INSERT INTO evaluations VALUES(11449,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(11449,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(11449,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(11449,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(11449,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(11449,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(11449,36,'cp 0','d1c2 d8e7 d2d3 c8g4 c1e3 g4f3 g2f3 f6d7 b1d2 e8c8 e1c1 f7f5 e4f5 e7f7 h2h4 f7f5 d1g1 h7h6 d3d4 f5f7 c2b3 f7f5');
INSERT INTO evaluations VALUES(14571,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(14571,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(14571,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(14571,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(14571,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(14571,35,'cp -7',NULL);
INSERT INTO evaluations VALUES(14571,36,'cp -12','c2c4 d5c4 d3c4 f8d6 b1c3 e8g8 b2b3 f8e8 c1b2 e5e4 f3d4 c6d4 d1d4 d6e5 d4d8 e8d8 f1d1 c8f5 d1d8 a8d8 a1d1 d8d1 c3d1 e5b2 d1b2 g8f8 f2f3 f8e7 g1f2 e7d6 f2e3 f5d7 a2a3 b7b6 e3d2 e4f3 e2f3 f6g8 f3f4 g8e7 b2d1 c7c6 d1e3 d6c5 d2c3');
INSERT INTO evaluations VALUES(16268,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(16268,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(16268,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(16268,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(16268,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(16268,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(16268,36,'cp 42','d2d4 d7d5 g1f3 d5c4 d1a4 f8b4 c1d2 f6d5 d2b4 d5b4 e1g1 a8b8 b1a3 e8g8 a4b5 b7b6 b5c4 c8a6 a3b5 d8d5 c4d5 b4d5 a2a4 a6b7 f3e5 c6e5 d4e5 a7a6 b5a3 d5b4 g2b7 b8b7 f2f4 h7h6 g1f2 g7g5 f2f3 f8d8 f1d1 d8d1 a1d1 g5f4 g3f4 b6b5 d1g1 g8f8 a4b5 a6b5');
INSERT INTO evaluations VALUES(16733,30,'cp -47',NULL);
INSERT INTO evaluations VALUES(16733,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(16733,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(16733,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(16733,34,'cp -47',NULL);
INSERT INTO evaluations VALUES(16733,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(16733,36,'cp -34','g7g6 h5d1 c6d4 g1f3 d4f3 d1f3 d8e7 h2h4 f8g7 b1c3 c7c6 f3g3 d7d5 c1g5 e7c7 g5h6 g7h6 e4d5 c8f5 d5c6 b7c6 e1g1 e8g8 f1e1 h6g7 c3e4 a8d8 c2c3 g8h8 g3e3 f5e4 e3e4 d8d2 e1e2 f7f5 e4e3 d2d6 c4b3 c7e7 h4h5 g6h5 a1d1 d6g6 e3f3 e7h4 d1d7 f5f4');
INSERT INTO evaluations VALUES(16888,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(16888,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16888,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16888,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16888,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(16888,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(16888,36,'cp 0','f8e7 d2d4 e8g8 h2h3 d7d5 d1f3 c7c6 b5d3 f8e8 b1d2 e7d6 e5e8 f6e8 d2b3 a7a5 a2a4 b7b6 c1e3 d8f6 f3f6 e8f6 b3c1 f6e8 g2g4 e8c7 g1g2 g7g6 c1e2 c8a6 e3f4 d6f4 d3a6 c7a6 e2f4 a8e8 c2c3 a6c7 f2f3 g8g7 g2f2 c7e6 f4e6 e8e6');
INSERT INTO evaluations VALUES(10969,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(10969,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(10969,32,'cp 1',NULL);
INSERT INTO evaluations VALUES(10969,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10969,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10969,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10969,36,'cp 0','d2d4 e8g8 c1g5 d7d6 e5e1 h7h6 g5h4 c7c6 b5d3 c8e6 b1c3 a7a5 c3e2 f8e8 d1d2 f6d5 h4e7 d8e7 e2g3 d5b4 e1e3 b4d5');
INSERT INTO evaluations VALUES(16934,30,'cp 587',NULL);
INSERT INTO evaluations VALUES(16934,31,'cp 582',NULL);
INSERT INTO evaluations VALUES(16934,32,'cp 594',NULL);
INSERT INTO evaluations VALUES(16934,33,'cp 595',NULL);
INSERT INTO evaluations VALUES(16934,34,'cp 602',NULL);
INSERT INTO evaluations VALUES(16934,35,'cp 607',NULL);
INSERT INTO evaluations VALUES(16934,36,'cp 582','c3d4 f6e4 d4e5 f8b4 b1d2 e8g8 f1d3 e4c5 d3c2 d7d6 g1e2 d6e5 a2a3 b4d2 d1d2 a7a5 e1g1 f7f5 d2c3 b7b6 c1e3 c8a6 e3c5 b6c5 f1e1 e5e4 e2f4 d8f6 f4e6 f6c3 b2c3 a6d3 c2d3 e4d3 e6c5 f5f4 c5d3 f4f3 d3e5 f3g2 g1g2 f8f6 a1d1');
INSERT INTO evaluations VALUES(16930,30,'cp 571',NULL);
INSERT INTO evaluations VALUES(16930,31,'cp 587',NULL);
INSERT INTO evaluations VALUES(16930,32,'cp 567',NULL);
INSERT INTO evaluations VALUES(16930,33,'cp 563',NULL);
INSERT INTO evaluations VALUES(16930,34,'cp 578',NULL);
INSERT INTO evaluations VALUES(16930,35,'cp 586',NULL);
INSERT INTO evaluations VALUES(16930,36,'cp 563','f6e4 d4e5 f8b4 b1d2 e8g8 f1d3 e4c5 d3c2 d7d6 g1e2 d6e5 a2a3 b4d2 d1d2 a7a5 e1g1 b7b6 f1e1 f7f5 e2c3 e5e4 d2f4 c8a6 c1e3 d8d6 a1c1 a8c8 f2f3 a6d3 c3a4 c5d7 c2d3 e4d3 a4c3 c8e8 f4d6 c7d6 e3d2 b6b5 c3b5 e8e1 c1e1 a5a4 b5d6 f5f4 d6c4');
INSERT INTO evaluations VALUES(17081,30,'cp 595',NULL);
INSERT INTO evaluations VALUES(17081,31,'cp 586',NULL);
INSERT INTO evaluations VALUES(17081,32,'cp 579',NULL);
INSERT INTO evaluations VALUES(17081,33,'cp 577',NULL);
INSERT INTO evaluations VALUES(17081,34,'cp 598',NULL);
INSERT INTO evaluations VALUES(17081,35,'cp 601',NULL);
INSERT INTO evaluations VALUES(17081,36,'cp 594','d4e5 f8b4 b1d2 e8g8 f1d3 e4c5 d3c2 d7d6 a2a3 b4d2 c1d2 f8e8 g1e2 e8e5 e1g1 e5d5 b2b4 c5d7 c2b3 d5d3 e2f4 d3d4 d1c1 d7e5 d2c3 e5d3 c1e3 d4f4 e3d3 c8e6 d3e3 e6b3 e3f4 f7f6 f1e1 b3f7 e1e3 d6d5 a1e1 d8d6 f4h4 b7b6 e3g3');
INSERT INTO evaluations VALUES(18262,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(18262,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(18262,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(18262,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(18262,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(18262,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(18262,36,'cp -14','d8g5 e5f3 g5g2 h1g1 g2h3 d1e2 d5e4 e2e4 g8e7 d2d3 f7f6 b1c3 c8f5 e4e3 f5d7 c1d2 c6c5 c3d5 e8c8 d5f4 h3f5 f3h4 f5e5 e3e5 f6e5 f4e2 e7c6 e1c1 g7g6 h4f3 f8d6 f3g5 d8f8 g5e4 c6d4 e2c3 d7c6 d2e3 f8f5 h2h4 h8f8 e4g5 d4f3 g5f3 c6f3 d1e1 f5h5 c3e4 h5h4 e3c5');
INSERT INTO evaluations VALUES(19285,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(19285,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(19285,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(19285,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(19285,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(19285,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(19285,36,'cp 10','c2c3 g8f6 e4e5 f6d5 d2d4 c5d4 d1b3 e7e6 c3d4 d8b6 b3b6 a7b6 e1g1 d7d6 b1c3 d6e5 d4e5 f8g7 c3d5 e6d5 f1d1 c8e6 c1f4 e8g8 g2g3 f8c8 h2h4 h7h6 a2a4 c6a5 e2b5 e6g4 g1g2 d5d4 d1d3 a5c6');
INSERT INTO evaluations VALUES(19754,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(19754,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(19754,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(19754,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(19754,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(19754,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(19754,36,'cp 47','a2a3 e7e6 b2b4 a7a5 b4b5 c6b8 c2c4 g8f6 b1c3 f8c5 c1b2 e8g8 g1f3 b8d7 c4d5 d7b6 c3a4 b6a4 d1a4 e6d5 f1d3 d8e7 a1c1 f6e4 e1g1 f8e8 h2h3 c8d7 c1c2 c5b6 f1d1 e8d8 f3e5 d7e6 d1c1 f7f6 e5f3 e4c5 a4d4 a8c8 a3a4 c5d3');
INSERT INTO evaluations VALUES(19915,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(19915,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(19915,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(19915,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(19915,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(19915,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(19915,36,'cp -39','c8d7 g1f3 g8f6 c2c3 g7g6 e1g1 f8g7 f1d1 a7a6 b5e2 b7b5 a4c2 e8g8 c1f4 d8c7 b1d2 f8c8 h2h3 c7b7 e2f1 e7e5 f4e3 c6d8 a2a4 d7c6 c2d3 g7f8 e3g5 f6d7 d3e3 d8e6 a4b5 c6b5 g5h6 f8h6 e3h6 b5f1');
INSERT INTO evaluations VALUES(26042,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(26042,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(26042,32,'cp -40',NULL);
INSERT INTO evaluations VALUES(26042,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(26042,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(26042,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(26042,36,'cp -34','g1f3 g8f6 c2c3 g7g6 e1g1 f8g7 f1d1 e8g8 a4c2 d8c7 a2a4 d7g4 b1d2 a7a6 b5e2 a8d8 h2h3 g4e6 d2c4 d6d5 e4d5 f6d5 c1g5 d5f4 e2f1 h7h6 g5f4 c7f4 d1d8 f8d8 c2e2 f4c4 e2c4 e6c4 f1c4');
INSERT INTO evaluations VALUES(20202,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(20202,31,'cp -65',NULL);
INSERT INTO evaluations VALUES(20202,32,'cp -71',NULL);
INSERT INTO evaluations VALUES(20202,33,'cp -81',NULL);
INSERT INTO evaluations VALUES(20202,34,'cp -67',NULL);
INSERT INTO evaluations VALUES(20202,35,'cp -72',NULL);
INSERT INTO evaluations VALUES(20202,36,'cp -77','g8e7 c2c3 d5d4 c3d4 c5d4 e1g1 e7g6 d2d3 c8d7 d1e2 f8e7 b5c4 d8b8 f1e1 e7d8 g2g3 d8c7 c1f4 e8g8 b1d2 b7b5 c4b3 a7a5 a2a3 g6f4 g3f4 c6e7 b3a2 a5a4 h2h4 b5b4 a2c4 c7a5 h4h5 b4a3 b2a3 d7c6 f3d4 c6d5');
INSERT INTO evaluations VALUES(20205,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(20205,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(20205,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(20205,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(20205,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(20205,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(20205,36,'cp 49','d5d4 c3a4 d8a5 c2c3 d4c3 a4c3 f8c5 a2a3 g8f6 f1d3 e8g8 e1g1 c5d6 h2h3 c6e5 f3e5 d6e5 c1d2 a5b6 d1e2 c8d7 d2e3 b6c7 a1c1 f8d8 f2f4 e5c3 c1c3 d7c6 f1c1 c7b8 e2c2 e6e5 f4f5 h7h6 b2b4 a7a6');
INSERT INTO evaluations VALUES(20431,30,'cp -65',NULL);
INSERT INTO evaluations VALUES(20431,31,'cp -54',NULL);
INSERT INTO evaluations VALUES(20431,32,'cp -56',NULL);
INSERT INTO evaluations VALUES(20431,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(20431,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(20431,35,'cp -43',NULL);
INSERT INTO evaluations VALUES(20431,36,'cp -40','c3d4 g8e7 g1f3 e7f5 b1c3 c8d7 g2g3 d8b6 c3e2 f8b4 e1f2 h7h5 f1h3 f5h6 h1f1 h6g4 f2g1 a8c8 a2a3 b4e7 b2b4 e7d8 c1d2 c6e7 a3a4 c8c4 d1b3 e7f5 a4a5 b6c6 f1c1 e8g8 h3f1 d8e7 a1b1 a7a6 h2h3 g4h6 c1c3 f8c8 b3a3 g8h8');
INSERT INTO evaluations VALUES(20995,30,'cp -49',NULL);
INSERT INTO evaluations VALUES(20995,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(20995,32,'cp -53',NULL);
INSERT INTO evaluations VALUES(20995,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(20995,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(20995,35,'cp -49',NULL);
INSERT INTO evaluations VALUES(20995,36,'cp -55','e7e5 d1d2 f8b4 e1c1 g8f6 d4e5 c6e5 e3g5 c8e6 e2e3 h7h6 d2d4 b4c3 d4c3 h6g5 c3e5 d8e7 f1d3 e6g4 e5e7 e8e7 g1e2 g4e2 d3e2 c7c6 h2h3 h8h4 f2f3 a7a5 h1e1 a5a4 e2d3 a4a3 b2b3 b7b5 e3e4 d5d4 d3f1 a8d8 f1e2 g7g6');
INSERT INTO evaluations VALUES(22294,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(22294,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(22294,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(22294,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(22294,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(22294,35,'cp 87',NULL);
INSERT INTO evaluations VALUES(22294,36,'cp 85','d6e5 d1e2 f8d6 a2a4 c8g4 c1e3 d8d7 a4a5 g8e7 c2c3 g4e6 b1d2 e8g8 b2b4 c6d8 e1g1 d8f7 f1d1 f8d8 d2f1 e6c4 e2c4 d7c6 f3d2 c6c4 d2c4 b7b5 c4d2 e7g6 d2b3 g6f4 f1g3 d6f8 a5a6 d8d1 a1d1 f7d6 e3f4 e5f4');
INSERT INTO evaluations VALUES(23188,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(23188,31,'cp 70',NULL);
INSERT INTO evaluations VALUES(23188,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(23188,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(23188,34,'cp 65',NULL);
INSERT INTO evaluations VALUES(23188,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(23188,36,'cp 72','f6e5 f3e5 f8d6 e5c4 g8f6 c2c3 d6e7 f1d3 e8g8 e1g1 d8e8 c4e5 f6d5 d3e4 d5f6 e4c2 e8h5 d1h5 f6h5 e5d3 b7b5 f1e1 h5f6 c1f4 f6d5 f4g3 a7a5 d3c5 b5b4 c2a4 e7c5 a4c6 d7c6 d4c5 c8a6 e1e6 f8f6');
INSERT INTO evaluations VALUES(23186,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(23186,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(23186,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(23186,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(23186,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(23186,35,'cp 63',NULL);
INSERT INTO evaluations VALUES(23186,36,'cp 68','d4e5 d7d5 e5d6 c7d6 f1d3 d6d5 e1g1 g8f6 c2c4 f8e7 b1c3 e8g8 c4d5 e6d5 h2h3 c8e6 c1e3 d8d7 f1e1 a8e8 f3d4 e7d6 d4e6 d7e6 c3b5 d6b8 e3c5 e6e1 d1e1 e8e1 a1e1 f8e8 e1e8 f6e8 b5d4 c6d4 c5d4 e8f6 d3c2 b8d6');
INSERT INTO evaluations VALUES(23183,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(23183,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(23183,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(23183,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(23183,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(23183,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(23183,36,'cp 51','d7d5 e5d6 f8d6 b1c3 g8f6 f1d3 e8g8 e1g1 g8h8 c3e4 c6b4 c1g5 e6e5 f1e1 c8g4 d3e2 g4f5 e2c4 f5e4 e1e4 d8e8 g5f6 g7f6 d1e2 b4c6 e2e3 e8g6 f3h4 g6h5 c4e2 h5g5 e3g5 f6g5 h4f3 h7h6 c2c3 h8g7 e2d3 f8f4 a1e1 a8d8 d3c2');
INSERT INTO evaluations VALUES(24511,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24511,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24511,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24511,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24511,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24511,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24511,36,'cp 0','d1e2 f8e7 f3e5 d7e5 d4e5 f6d5 c4d3 b7b5 c1d2 e8g8 d3g6 h7g6 e2g4 d5b6 d2f4 b6d5');
INSERT INTO evaluations VALUES(24877,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(24877,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(24877,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(24877,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(24877,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(24877,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(24877,36,'cp 54','f6e4 g5h4 d5c4 e2e3 e4c3 b2c3 c6a5 f3d2 f5e6 f1e2 g7g6 e1g1 f8g7 d1a4 c7c6 a4b4 b7b6 f1d1 d8d7 h2h3 g7f6 h4f6 e7f6 a1c1 a8d8 d2c4 e6c4 e2c4 a5c4 b4c4 e8g8 c4b3 d7c7 c3c4 c6c5 d4d5 d8d7 b3c3 f6f5 a2a4 f8e8 a4a5 f5f4 e3f4');
INSERT INTO evaluations VALUES(25854,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(25854,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(25854,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(25854,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(25854,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(25854,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(25854,36,'cp 59','e2e3 g8f6 d2d4 c5d4 e3d4 e7e6 d4d5 c6e5 b2b3 e6d5 g1e2 d8c7 c1f4 f8c5 a1b1 e8g8 b3b4 c5e7 c4d5 f8e8 e1g1 e7f8 f4g5 c7d8 c3e4 e5g4 d1d4 h7h6 e4f6 g4f6 g5f6 d8f6 d4f6 g7f6 e2d4 a8c8 a2a3 c8c4 f1d1 e8c8 h2h3 a7a6 h3h4');
INSERT INTO evaluations VALUES(26739,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(26739,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(26739,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(26739,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(26739,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(26739,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(26739,36,'cp 71','b1d2 g8f6 c2c3 f8e7 h2h3 g4e6 c4e6 f7e6 b2b4 d8d7 a2a4 e8g8 d2b3 e7d8 c1e3 d7f7 b3d2 c6e7 d3d4 e5d4 c3d4 a7a6 d1b3 d6d5 f3e5 f7e8 e5d3 d5e4 b3e6 g8h8 d2e4 e7d5 e6e8 f8e8 e4f6 d8f6 a1b1 c7c6 a4a5');
INSERT INTO evaluations VALUES(27434,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(27434,31,'cp 168',NULL);
INSERT INTO evaluations VALUES(27434,32,'cp 162',NULL);
INSERT INTO evaluations VALUES(27434,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(27434,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(27434,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(27434,36,'cp 182','e6f8 h8f8 b1c3 e8c8 c1e3 b7b6 f2f3 d6d5 e4d5 f6d5 c3d5 d8d5 e1g1 a7a5 f1e1 a5a4 e3f2 e7d6 d1d2 a4a3 b2b3 c8b7 e1e4 f8f4 a1e1 h7h6 h2h3 c6d4 g1f1 g7g5 f1g1 d4e6 d2e3 f4e4 e3e4 e6f4 g1f1');
INSERT INTO evaluations VALUES(27349,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(27349,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(27349,32,'cp 167',NULL);
INSERT INTO evaluations VALUES(27349,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(27349,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(27349,35,'cp 155',NULL);
INSERT INTO evaluations VALUES(27349,36,'cp 167','h8f8 b1c3 e8c8 c1e3 e7e6 d1d2 d6d5 e4d5 f6d5 f2f3 b7b6 c3d5 e6d5 e1g1 c8b7 b2b3 d5d6 a2a4 f8f5 a1e1 a7a5 e3f2 c6d4 f2d4 e5d4 d2e2 d6c6 f1f2 h7h6 e2e4 d8d5 f3f4 g7g5 f4g5 f5g5 h2h3 d5c5 e4c6 c5c6 g2g4');
INSERT INTO evaluations VALUES(28151,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(28151,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(28151,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(28151,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(28151,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(28151,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(28151,36,'cp 36','h2h3 g4h5 a2a3 b8d7 d1e2 c7c6 c4a2 a7a5 b1d2 h5g6 d2f1 g8f6 f3h4 b7b5 f1g3 b5b4 e1g1 b4a3 b2a3 a5a4 c1g5 e8g8 h4g6 h7g6 h3h4 g8h8 h4h5 d8e8 c2c3 c5a3 d3d4 f6h7 g5e3 e8e7 h5g6 f7g6 e2c4 d7f6 d4e5 f6g4 e5e6 g4e3 f2e3 e7h4');
INSERT INTO evaluations VALUES(28150,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(28150,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(28150,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(28150,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(28150,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(28150,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(28150,36,'cp 21','g4h5 a2a3 b8d7 d1e2 c7c6 c4a2 h5g6 c1g5 g8f6 f3h4 h7h6 h4g6 f7g6 g5h4 d8e7 c2c3 d7f8 b2b4 c5b6 a3a4 g6g5 h4g3 f8e6 e1g1 e8g8 a4a5 b6c7 b1d2 g8h8 a2e6 e7e6 d3d4 a8e8 f1e1 e5d4 c3d4 a7a6 a1d1 e6d7 e4e5 d7e6 d2e4 f6e4 e2e4 d6d5');
INSERT INTO evaluations VALUES(28284,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(28284,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(28284,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(28284,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(28284,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(28284,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(28284,36,'cp -49','d7d6 a3a4 a7a5 d1c2 d8e7 f1e2 b8d7 e1g1 e6e5 f3d2 e8g8 d2b3 c7c5 f2f3 e5e4 f3f4 f6e8 f4f5 e8f6 f1f2 f8e8 b3d2 h7h6 d2f1 a8c8 c2b3 b7a6 b3a2 e7d8 c1d2 h6h5 d2e1 c5d4 c3d4 d6d5');
INSERT INTO evaluations VALUES(30395,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(30395,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(30395,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(30395,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(30395,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(30395,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(30395,36,'cp 46','e7e6 c2c4 f8d6 f4d6 d8d6 d1b3 b7b6 c4d5 e6d5 f1d3 f5d3 b3d3 g8f6 b1c3 e8g8 e1g1 f8e8 b2b3 b8d7 a1c1 h7h6 c1c2 f6e4 c3e2 c7c5 f1d1 d7f6 e2f4 g7g5 f4e2 d6e6 h2h3 a6a5 a2a4 a8c8 d1c1 c8c7 f3e5 g8g7 f2f3 e4d6');
INSERT INTO evaluations VALUES(34008,30,'cp 450',NULL);
INSERT INTO evaluations VALUES(34008,31,'cp 479',NULL);
INSERT INTO evaluations VALUES(34008,32,'cp 475',NULL);
INSERT INTO evaluations VALUES(34008,33,'cp 454',NULL);
INSERT INTO evaluations VALUES(34008,34,'cp 467',NULL);
INSERT INTO evaluations VALUES(34008,35,'cp 459',NULL);
INSERT INTO evaluations VALUES(34008,36,'cp 465','c4f7 e8d8 d2d4 b8c6 e5g6 e7e4 e1f1 h8g8 b1c3 e4f5 h2h4 d7d6 g6f4 g8h8 h1h3 f5h7 f7e6 f8e7 c1e3 c8e6 f4e6 d8c8 d1e2 h7f5 e6g7 f5g4 h3g3 g4e2 c3e2 f6d5 g7f5 c8d7 g3f3 b7b5 c2c3 e7f6 g2g3 h6h5 e2f4 d5f4 e3f4 a8e8 f4d2 h8f8 a1e1 e8e1 f1e1 c6e7 d2h6 f8f7 f5e3 d7e8 f3f4');
INSERT INTO evaluations VALUES(34002,30,'cp 451',NULL);
INSERT INTO evaluations VALUES(34002,31,'cp 465',NULL);
INSERT INTO evaluations VALUES(34002,32,'cp 440',NULL);
INSERT INTO evaluations VALUES(34002,33,'cp 436',NULL);
INSERT INTO evaluations VALUES(34002,34,'cp 485',NULL);
INSERT INTO evaluations VALUES(34002,35,'cp 461',NULL);
INSERT INTO evaluations VALUES(34002,36,'cp 455','e8d8 d2d4 b8c6 e5g6 e7e4 e1f1 h8g8 b1c3 e4f5 h2h4 d7d6 g6f4 g8h8 h1h3 f5h7 f7e6 f8e7 c1e3 c8e6 f4e6 d8c8 d1e2 h7f5 e6g7 f5g4 h3g3 g4e2 c3e2 f6d5 g7f5 c8d7 g3f3 b7b5 c2c3 e7f6 g2g3 h6h5 e2f4 d5f4 f3f4 a8f8 e3d2 a7a5');
INSERT INTO evaluations VALUES(34685,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(34685,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(34685,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(34685,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(34685,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(34685,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(34685,36,'cp 31','g8f6 g1f3 c7c6 d1e2 b8d7 d2d4 e7e6 c1d2 a5c7 g2g4 b7b5 g4g5 f6d5 h3h4 b5b4 c3e4 c6c5 h4h5 a7a5 d4c5 c8a6 e2d1 a6f1 e1f1 d7c5 c2c4 c5e4 c4d5 c7c4 f1g2 c4d5 d2f4 f8d6 d1d5 e6d5 f4d6');
INSERT INTO evaluations VALUES(33742,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(33742,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(33742,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(33742,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(33742,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(33742,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(33742,36,'cp 44','g1f3 c7c6 a2a3 c8f5 f3h4 f5e6 f1e2 c6c5 e1g1 b8c6 h4f3 a8d8 f1e1 e6c8 e2b5 c8d7 b5c6 d7c6 f3e5 a5b6 d2d3 e7e6 d1e2 a7a6 a3a4 b6c7 a4a5 h7h6 c3a4 c6a4 a1a4 f8d6 e5c4 d6e7 c4b6 f6d5 b6d5 d8d5 c1f4');
INSERT INTO evaluations VALUES(37330,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(37330,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(37330,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(37330,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(37330,34,'cp 81',NULL);
INSERT INTO evaluations VALUES(37330,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(37330,36,'cp 74','b1c3 h7h6 g5f6 d7d6 d1d2 d6e5 f6e5 b7b6 h2h4 c8b7 h4h5 g6g5 e1c1 e7d5 f1b5 d5c3 d2c3 b7d5 e5g7 g8g7 c3e3 a7a6 b5e2 b8d7 f3e5 d7e5 d4e5 d8e7 f2f4 f7f5 e5f6 e7f6 f4g5 f6g5 h1h3 g5e3 h3e3 g7f6 d1e1 f8g8 e2f3 a8f8 f3d5 e6d5 e3e7 f8f7 e1e6 f6f5 e7f7 f5e6 f7c7 g8g2');
INSERT INTO evaluations VALUES(37329,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(37329,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(37329,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(37329,33,'cp 71',NULL);
INSERT INTO evaluations VALUES(37329,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(37329,35,'cp 79',NULL);
INSERT INTO evaluations VALUES(37329,36,'cp 62','h7h6 g5f6 d7d6 d1d2 d6e5 f6e5 b7b6 h2h4 c8b7 h4h5 g6g5 e1c1 e7d5 c3d5 b7d5 e5g7 g8g7 f3e5 b8d7 f2f4 d7e5 d4e5 g5f4 d2f4 d8g5 f4g5 h6g5 h5h6 g7h8 h1h5 g5g4 h6h7 d5e4 d1d7 a8c8 h5h4 e4f5 f1d3 c8d8 d7c7 f5d3 c2d3 d8d3');
INSERT INTO evaluations VALUES(37728,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(37728,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(37728,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(37728,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(37728,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(37728,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(37728,36,'cp 41','b1c3 e7e5 e2e4 c6c5 g1e2 b8c6 e1g1 g8e7 a2a3 d7d6 b2b4 e8g8 a1b1 b7b6 d2d3 c6d4 c1g5 d4e6 g5d2 e6d4 h2h3 d4e2 c3e2 a8b8 e2c3 e7c6 b4c5 b6c5 b1b8 c6b8 d1a4 d8d7 a4d7 c8d7 d2e3 b8c6 f1b1');
INSERT INTO evaluations VALUES(40405,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(40405,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(40405,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(40405,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(40405,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(40405,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(40405,36,'cp 26','c2c4 e8g8 b1c3 b8c6 d4e2 a7a6 h2h3 c6e5 a2a4 f6d7 b2b3 e5d3 d1d3 d7c5 d3c2 f7f5 a1b1 f5e4 b3b4 c5d3 c3e4 d6d5 c4d5 d8d5 e2g3 c8f5 f1d1 f5e4 g3e4 d5e4 c2b3 e7e6 b3d3 e4d3 d1d3 f8d8 d3d8 a8d8 c1e3 d8d5 b4b5 a6b5 b1b5 d5b5 a4b5 g7b2 e3c5');
INSERT INTO evaluations VALUES(40801,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(40801,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(40801,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40801,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40801,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(40801,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40801,36,'cp 5','c2c4 c7c5 e1g1 e8g8 b1c3 b8c6 a2a3 a7a5 a1b1 c8d7 c1d2 a8b8 f3e1 c6d4 a3a4 f6e8 e1c2 d4b3 d2e3 b3d4 f2f4 e5f4 e3f4 e7f6 c2d4 f6d4 g1h1 d7c6 d1e1 d8e7 b1a1 g7g6 e1g3 f7f5 e2f3 d4c3 b2c3 f5e4 f3e4 c6e4');
INSERT INTO evaluations VALUES(41080,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41080,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41080,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41080,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(41080,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41080,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41080,36,'cp 0','b1c3 e8g8 c4d5 f6d5 f1d3 c8e6 e1g1 b8d7 f1e1 f8e8 h2h3 c7c6 c1d2 d7f8 a1c1 d8c7 d3b1 a8d8 d1c2 h7h6 e1e2 d5f4 e2e3 c7d7 c3e4 f8g6 c1e1 f4d5 e3e2 d5f4');
INSERT INTO evaluations VALUES(41078,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(41078,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(41078,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(41078,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41078,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41078,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41078,36,'cp 0','e8g8 c4d5 f6d5 f1d3 c8e6 e1g1 b8d7 f1e1 f8e8 a2a3 c7c6 d1c2 h7h6 c1d2 d7f8 e1e2 e6g4 c3d5 c6d5 d3b5 g4f3 g2f3 f8e6 e2e6 f7e6 d2h6 g7h6 c2g6 g8f8 g6h6 f8f7 h6h7 f7f6 h7h6');
INSERT INTO evaluations VALUES(36398,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36398,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36398,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36398,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(36398,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36398,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36398,36,'cp 0','c4d5 f6d5 f1d3 c7c5 e1g1 b8c6 d4c5 d5c3 b2c3 e7c5 c1g5 d8a5 a1b1 h7h6 b1b5 a5c3 b5b3 c3a5 b3b5');
INSERT INTO evaluations VALUES(42606,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(42606,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(42606,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(42606,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(42606,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(42606,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(42606,36,'cp 25','g1f3 d7d6 d2d4 e5d4 f3d4 e8g8 c1g5 h7h6 g5h4 c5d4 d1d4 b8c6 d4d1 c8e6 e2e3 c6e5 b2b3 c7c6 f1e2 e5g6 h4g3 d6d5 e1g1 d5c4 b3c4 d8a5 d1b3 b7b5 h2h3 f8d8 f1b1 g6e5 g3e5 b5c4 e2c4 e6c4 b3c4 a5e5 b1d1 d8d1 a1d1 a8c8 d1d4 e5a5 a3a4 a5b6 d4d1 b6c7 c4a6 c6c5 c3b5');
INSERT INTO evaluations VALUES(42641,30,'cp -143',NULL);
INSERT INTO evaluations VALUES(42641,31,'cp -165',NULL);
INSERT INTO evaluations VALUES(42641,32,'cp -155',NULL);
INSERT INTO evaluations VALUES(42641,33,'cp -138',NULL);
INSERT INTO evaluations VALUES(42641,34,'cp -122',NULL);
INSERT INTO evaluations VALUES(42641,35,'cp -117',NULL);
INSERT INTO evaluations VALUES(42641,36,'cp -119','b1c3 d7d5 c1g5 c7c6 d1d2 d5d4 c3e2 b8d7 d2c1 a7a5 g1h3 a5a4 h3f2 a4a3 b2b3 b7b5 g2g3 e8g8 f1h3 h7h6 g5d2 f8e8 e1g1 c5f8 g1h1 d8b6 c2c3 d4c3 d2e3 f8c5 e3c5 d7c5 h3c8 a8c8 c1c3 b5b4 c3c2 c5e6 a1c1 c6c5 f3f4 e5f4 g3f4 e6d4 c2d2 d4e2 d2e2');
INSERT INTO evaluations VALUES(42793,30,'cp 537',NULL);
INSERT INTO evaluations VALUES(42793,31,'cp 536',NULL);
INSERT INTO evaluations VALUES(42793,32,'cp 555',NULL);
INSERT INTO evaluations VALUES(42793,33,'cp 538',NULL);
INSERT INTO evaluations VALUES(42793,34,'cp 536',NULL);
INSERT INTO evaluations VALUES(42793,35,'cp 538',NULL);
INSERT INTO evaluations VALUES(42793,36,'cp 563','h1e1 b8c6 d2d4 d7d6 d4e5 f6g4 f2g1 c6e5 f3e5 g4e5 c4e2 h7h5 c1f4 h5h4 h2h3 g7g5 f4e5 d6e5 d1d8 e8d8 b1c3 c7c6 c3d1 d8e7 d1e3 h8d8 a1d1 d8d4 e2g4 c8g4 h3g4 d4d1 e3f5 e7e8 e1d1 a8d8 d1d8 e8d8 f5d6 f7f6 d6b7 d8d7 g1h2 d7c7 b7a5 c7b6 a5b3 b6c7 h2h1 c7b8 b3a5 b8c7 a5c4');
INSERT INTO evaluations VALUES(43765,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(43765,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(43765,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(43765,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(43765,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(43765,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(43765,36,'cp -36','c5b6 a7b6 e2e3 e7e6 g1f3 f8b4 c1d2 b4d6 b1c3 c7c5 f1b5 c8d7 b5d7 d8d7 f3e5 d7b7 c3b5 d6e7 a2a4 f6e4 e1g1 e8g8 d1e2 f8c8 b2b3 b8a6 d2c3 a6c7 c3b2 c7b5 a4b5 b7c7 d4c5 a8a1 f1a1');
INSERT INTO evaluations VALUES(47622,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(47622,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(47622,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(47622,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(47622,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(47622,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(47622,36,'cp 0','f8e7 e1g1 c7c6 c4b3 e8g8 b1c3 b7b5 c3e2 f8e8 e2g3 e7f8 a2a4 b5b4 f1e1 d6d5 e4d5 c6d5 d3d4 e5e4 f3e5 c8e6 f2f4 e4f3 d1f3 b8d7 a1d1 d8c7 e3f4 c7b6 f4e3 b6c7');
INSERT INTO evaluations VALUES(40690,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40690,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40690,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40690,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40690,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40690,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(40690,36,'cp 7','b1c3 e8g8 e1g1 c7c6 c4b3 b7b5 c3e2 f8e8 e2g3 b8d7 d3d4 a7a5 c2c3 e7f8 a2a4 c8b7 b3c2 b5b4 d4e5 d6e5 c3b4 f8b4 c2d3 d7c5 e3c5 b4c5 d1c2 c5a7 d3c4 d8c7 a1d1 a8b8 b2b3 b7c8 h2h3');
INSERT INTO evaluations VALUES(49305,30,'cp 142',NULL);
INSERT INTO evaluations VALUES(49305,31,'cp 142',NULL);
INSERT INTO evaluations VALUES(49305,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(49305,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(49305,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(49305,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(49305,36,'cp 128','f8b4 f1d3 d7d5 h2h4 h7h6 d1e2 c7c5 e4d5 e8g8 h4h5 g6e7 d5e6 c8e6 d4c5 b8d7 c1d2 e7d5 e2e4 d7f6 e4d4 d8e7 e1c1 d5c3 d2c3 b4c3 d4c3 a8c8 f3d4 e7c5 c3c5 c8c5 a2a3 f6h5 h1h4 h5f6 d4e6');
INSERT INTO evaluations VALUES(49762,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(49762,31,'cp -103',NULL);
INSERT INTO evaluations VALUES(49762,32,'cp -107',NULL);
INSERT INTO evaluations VALUES(49762,33,'cp -105',NULL);
INSERT INTO evaluations VALUES(49762,34,'cp -100',NULL);
INSERT INTO evaluations VALUES(49762,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(49762,36,'cp -78','f6e4 f1d3 d7d5 e1g1 f8e7 c1f4 e8g8 c2c4 e4f6 b1c3 b8c6 h2h3 c6b4 c4c5 b7b6 d3b5 b6c5 a2a3 b4a6 f3e5 a6b8 d4c5 e7c5 g1h1 c5d6 d1f3 c8e6 a1d1 a7a5 f1e1 a8a7 c3d5 f6d5 d1d5 e6d5 f3d5');
INSERT INTO evaluations VALUES(49920,30,'cp -90',NULL);
INSERT INTO evaluations VALUES(49920,31,'cp -75',NULL);
INSERT INTO evaluations VALUES(49920,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(49920,33,'cp -89',NULL);
INSERT INTO evaluations VALUES(49920,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(49920,35,'cp -82',NULL);
INSERT INTO evaluations VALUES(49920,36,'cp -71','f1d3 d7d5 e1g1 f8d6 c2c4 e8g8 c4c5 d6e7 b1c3 c8f5 c1f4 f5g6 f3e5 e7f6 e5g6 h7g6 g1h1 e4c3 b2c3 b7b6 c5b6 a7b6 a2a4 c7c6 g2g3 b8d7 h2h4 f8e8 h1g2 e8e6 h4h5 g6h5 d1h5 d7f8 f1f2 g7g6 h5h3 d8e7 f4h6 f6g7 h6g7 g8g7 a1f1');
INSERT INTO evaluations VALUES(51549,30,'cp 108',NULL);
INSERT INTO evaluations VALUES(51549,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(51549,32,'cp 117',NULL);
INSERT INTO evaluations VALUES(51549,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(51549,34,'cp 116',NULL);
INSERT INTO evaluations VALUES(51549,35,'cp 113',NULL);
INSERT INTO evaluations VALUES(51549,36,'cp 102','g1f3 c8b7 f1d3 g8f6 d1e2 d7d5 e4e5 f6d7 e1g1 f8e7 a2a4 b8c6 b1a3 a7a6 c1f4 c5d4 c3d4 c6b4 d3b1 d7b8 a3c2 b4c2 b1c2 b8c6 f4d2 a6a5 c2d3 e8g8 a1c1 c6b4 d3b5 b7a6 d2b4 a6b5 e2b5 e7b4 c1c6 d8d7 b5b6 a8b8 b6a6 f8c8 f1c1');
INSERT INTO evaluations VALUES(53882,30,'cp -6',NULL);
INSERT INTO evaluations VALUES(53882,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(53882,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(53882,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(53882,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53882,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53882,36,'cp 0','g7g6 c1g5 f7f6 g5d2 b8c6 b1c3 g8e7 a2a3 f8g7 a1b1 a7a5 g2g3 e8g8 f1g2 g8h8 e1g1 f6f5 e4f5 c8f5 d2e3 d8d7 d1d2 b7b6 b1e1 a8b8 g1h1 c6d4 e3g5 d4e6 g5e7 d7e7 f3h4 e6d4 h4f5 g6f5 c3d5 e7d8');
INSERT INTO evaluations VALUES(54631,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(54631,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(54631,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(54631,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(54631,34,'cp 85',NULL);
INSERT INTO evaluations VALUES(54631,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(54631,36,'cp 82','f6e5 f3e5 g8f6 c1g5 f8b4 f1d3 e8g8 e1g1 c6c5 a2a3 b4c3 b2c3 b8d7 d1e1 c5c4 d3e2 h7h6 g5h4 d8e8 h4g3 a7a5 e5d7 e8d7 g3e5 b7b5 f2f3 d7e7 e1h4 c8d7 g1h1 d7e8 f1e1 e8g6 h4g3 g6c2 e5d6 e7d7 d6f8 a8f8');
INSERT INTO evaluations VALUES(54735,30,'cp 209',NULL);
INSERT INTO evaluations VALUES(54735,31,'cp 196',NULL);
INSERT INTO evaluations VALUES(54735,32,'cp 209',NULL);
INSERT INTO evaluations VALUES(54735,33,'cp 200',NULL);
INSERT INTO evaluations VALUES(54735,34,'cp 190',NULL);
INSERT INTO evaluations VALUES(54735,35,'cp 200',NULL);
INSERT INTO evaluations VALUES(54735,36,'cp 207','f1c4 e7e6 b1c3 b8c6 d1h5 g7g6 h5e2 f6f5 g1f3 g8f6 e1g1 f8g7 f3e5 c6e5 d4e5 f6e4 f1d1 d8e7 f2f4 e4c3 b2c3 e8g8 a2a4 f8d8 d1d4 d8d4 c3d4 g7f8 e2a2 g8h8 d4d5 e6d5 c4d5 c8d7 c1b2 d7c6 a2b3 c6d5 b3d5 a8d8 d5b7 e7h4 b7a7 d8d2');
INSERT INTO evaluations VALUES(54732,30,'cp 195',NULL);
INSERT INTO evaluations VALUES(54732,31,'cp 200',NULL);
INSERT INTO evaluations VALUES(54732,32,'cp 187',NULL);
INSERT INTO evaluations VALUES(54732,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(54732,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(54732,35,'cp 188',NULL);
INSERT INTO evaluations VALUES(54732,36,'cp 189','e7e6 d1h5 g7g6 h5e2 f6f5 g1f3 b8c6 b1c3 g8f6 e1g1 f8g7 f1d1 e8g8 a2a4 c6a5 c4a2 a5c6 a4a5 g8h8 a2c4 f6d5 e3e4 f5e4 c3e4 a7a6 c1g5 d8e8 a1a3 h7h6 g5c1 d5e7 b2b4 e7d5 h2h4 c8d7 e4c5 b7b5 a5b6 c7b6 c5d7 e8d7');
INSERT INTO evaluations VALUES(55437,30,'cp 129',NULL);
INSERT INTO evaluations VALUES(55437,31,'cp 111',NULL);
INSERT INTO evaluations VALUES(55437,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(55437,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(55437,34,'cp 132',NULL);
INSERT INTO evaluations VALUES(55437,35,'cp 123',NULL);
INSERT INTO evaluations VALUES(55437,36,'cp 128','f3d4 b8c6 f1b5 c8d7 d4c6 b7c6 b5d3 g8f6 e1g1 f6g4 d3e2 g4f6 c1e3 d8e7 e3d4 c6c5 d4e3 a8b8 b2b3 g7g5 d1d2 f8g7 f2f3 e8g8 a1d1 b8e8 e2d3 d7c6 d3c4 f6d7 c3d5 c6d5 c4d5 g8h8');
INSERT INTO evaluations VALUES(56543,30,'cp -88',NULL);
INSERT INTO evaluations VALUES(56543,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(56543,32,'cp -97',NULL);
INSERT INTO evaluations VALUES(56543,33,'cp -80',NULL);
INSERT INTO evaluations VALUES(56543,34,'cp -105',NULL);
INSERT INTO evaluations VALUES(56543,35,'cp -96',NULL);
INSERT INTO evaluations VALUES(56543,36,'cp -97','d8d4 d1d4 e5d4 g1f3 c7c5 b1a3 c8e6 c4e6 f7e6 a3b5 b8a6 b2b4 b7b6 b4c5 b6c5 c1f4 g8f6 e1g1 f8e7 f1e1 e8g8 a2a3 f8c8 a1c1 f6d7 e1d1 e7f8 f4g3 h7h6 h2h4 c8e8 f3d2 d7b6 d2f3 e8d8 f3e5 d8b8 e5g6 b8b7 g6f8 a8f8 g3d6 f8e8');
INSERT INTO evaluations VALUES(2570,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(2570,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(2570,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(2570,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(2570,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(2570,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(2570,36,'cp 29','b7b5 a4c2 e5d4 c3d4 c6b4 d4d5 c7c5 d5c6 b4c6 b1c3 c6e5 h2h3 c8b7 c1f4 e5c4 a2a4 b5b4 c3d5 f6d5 e4d5 c4b2 d1d4 e7f6 d4b4 b7d5 f4d6 f8e8 b4f4 d5e6 d6e5 a8c8 e5f6 d8f6 f4f6 g7f6 f3d4 c8c4 e1e4 e8c8 d4e6 c4c2 e6d4');
INSERT INTO evaluations VALUES(3439,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(3439,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(3439,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(3439,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(3439,34,'cp 22',NULL);
INSERT INTO evaluations VALUES(3439,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(3439,36,'cp 27','f6g4 d4d1 c7c6 h2h3 g4e5 c4b3 a7a5 c1e3 e7h4 a2a4 c8e6 d1d2 e6b3 c2b3 h4e7 a1d1 e5g6 d2d3 d8d7 f2f4 f7f5 e4e5 d6d5 c3e2 f8f7 e2d4 g6f8 g1h2 f8e6 g2g3 e7b4 e3d2 b4d2 d1d2 h7h6 d3c3 f7e7 d4e6 d7e6');
INSERT INTO evaluations VALUES(5698,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(5698,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(5698,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(5698,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(5698,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(5698,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(5698,36,'cp 25','f1c4 g8f6 h2h3 e8g8 e1g1 b7b6 c2c3 c8b7 b1d2 f8e8 a2a4 e7e5 d4e5 d6e5 f4h2 d8e7 e3e4 a8d8 d1c2 f6d7 f1e1 a7a5 d2f1 g7h6 f1e3 h6e3 e1e3 g8h8 c4d5 f7f6 b2b3 d7c5 a1d1 c5e6 c2d2 e6c5');
INSERT INTO evaluations VALUES(6965,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(6965,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(6965,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(6965,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(6965,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(6965,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(6965,36,'cp 50','g8h6 c1h6 e8g8 h6g7 g8g7 b1c3 d7d6 a1d1 e7f6 e1g1 g7h8 d5d2 f6g7 h2h3 c8e6 b2b3 d8d7 c3d5 a8e8 f1e1 c6e5 f3e5 g7e5 f2f4 e5g7 d2c2 c7c6 d5e3 d7e7 c4e6 f7e6 g2g3 c6c5 c2e2 g7d4 g1h2 f8g8 e3c2 e7g7 d1d3');
INSERT INTO evaluations VALUES(10952,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(10952,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(10952,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(10952,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(10952,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(10952,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(10952,36,'cp 62','c1h6 e8g8 h6g7 g8g7 b1c3 d7d6 e1g1 g7h8 a1d1 e7f6 d5d2 f6g7 h2h3 c8e6 c4d5 d8f6 f3g5 h7h6 g5e6 f7e6 d5c6 b7c6 f2f4 a8b8 f1f3 c6c5 g1h1 b8b4 d1e1 f8b8 b2b3 c5c4 f4f5 f6d4 d2d4 g7d4 f5e6 c4b3 a2b3 b4b3 c3d5');
INSERT INTO evaluations VALUES(10951,30,'cp 50',NULL);
INSERT INTO evaluations VALUES(10951,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(10951,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(10951,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(10951,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(10951,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(10951,36,'cp 77','e8g8 h6g7 g8g7 b1c3 d7d6 e1g1 g7h8 a1d1 e7f6 d5d2 f6g7 h2h3 c8e6 c4d5 d8f6 f3g5 h7h6 g5e6 f7e6 d5c6 b7c6 f2f4 a8b8 g1h1 f6h4 f1f3 d6d5 d1f1 d5e4 c3e4 b8b2 d2d7 h4d8 d7e6 d8d5 e6g6 d5f7 f3g3 f7g6 g3g6');
INSERT INTO evaluations VALUES(4612,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(4612,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(4612,32,'cp 76',NULL);
INSERT INTO evaluations VALUES(4612,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(4612,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(4612,35,'cp 73',NULL);
INSERT INTO evaluations VALUES(4612,36,'cp 65','h6g7 g8g7 b1c3 d7d6 a1d1 g7h8 e1g1 e7f6 d5d2 c8g4 c4e2 d8d7 g1h1 f6g7 f3h4 c6d4 e2g4 d7g4 g2g3 d4f3 d2e2 g7c3 e2f3 g4f3 h4f3 c3g7 h1g2 f7f5 e4f5 f8f5 f3d4 g7d4 d1d4 a8e8 d4d2 f5c5 f2f4 h8g8 g2f3 a7a5 f1d1 b7b5 g3g4 a5a4 h2h3 b5b4');
INSERT INTO evaluations VALUES(7036,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(7036,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(7036,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(7036,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(7036,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(7036,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(7036,36,'cp 62','d7d6 f1d3 e7e5 h2h3 g8e7 e1g1 e8g8 d4e5 c6e5 f3e5 d6e5 d1e2 a7a5 b1a3 e7c6 b2b3 c8e6 d3c4 d8e7 a3c2 f8e8 c2e3 c6d8 a2a4 e6c4 e2c4 d8e6 e3d5 e7h4 c1e3 c7c6 d5b6 a8b8 f1d1 b8d8 d1d7 d8d7 b6d7 e8d8');
INSERT INTO evaluations VALUES(5702,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(5702,31,'cp 54',NULL);
INSERT INTO evaluations VALUES(5702,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(5702,33,'cp 61',NULL);
INSERT INTO evaluations VALUES(5702,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(5702,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(5702,36,'cp 64','f1d3 e7e5 h2h3 g8e7 e1g1 e8g8 d4e5 c6e5 f3e5 d6e5 d1e2 e7c6 a2a4 a7a5 f1d1 d8e7 d3c4 f8d8 c1d2 c8e6 b1a3 g7f8 c4e6 e7e6 a3b5 a8c8 d2g5 f7f6 g5e3 b7b6 e2f3 d8d1 a1d1 f8e7 f3g3 c6d8 f2f4 d8f7 f4e5 f6e5');
INSERT INTO evaluations VALUES(7682,30,'cp 99',NULL);
INSERT INTO evaluations VALUES(7682,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(7682,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(7682,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(7682,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(7682,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(7682,36,'cp 91','d1d6 c8b7 f1d1 e7c8 d6g3 e8g8 a2a4 d7d5 a4a5 c8d6 c4f1 a8c8 a5a6 b7a8 c1f4 e6e5 f4e5 g7e5 g3e5 f8e8 e5g3 d6e4 c3e4 e8e4 c2c4 h7h5 g3a3 c6c5 c4d5 a8d5 a3c3 d8g5 a1a5 g5e5 c3e5 e4e5 b2b4 d5e6 d1c1 g8f8 c1c5 e5c5 a5c5 c8c5 b4c5');
INSERT INTO evaluations VALUES(7966,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(7966,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(7966,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(7966,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(7966,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(7966,35,'cp -34',NULL);
INSERT INTO evaluations VALUES(7966,36,'cp -24','c4b3 e8g8 d2d3 d7d5 e4e5 e7f5 c3e2 b7b5 c2c3 d5d4 e2g3 c8b7 g3f5 g6f5 c3c4 b5c4 b3c4 a7a5 c1d2 c6e7 d1e1 a5a4 a1b1 d8d7 b2b4 a4b3 a2b3 f8b8 h2h3 b7c6 e1g3 e7d5 c4d5 c6d5 b3b4 d5f3 f1f3 c5b4 b1b4');
INSERT INTO evaluations VALUES(8586,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(8586,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(8586,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(8586,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(8586,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(8586,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(8586,36,'cp -19','c6a5 h2h3 e8g8 c4b3 b7b6 e1g1 a5b3 a2b3 c8b7 f3h2 d8d7 h2g4 f6g4 h3g4 f8e8 f1e1 a7a5 f2f3 b6b5 d3d4 e5d4 e3d4 b5b4 c3e2 d6d5 e4d5 d7d5 d4f2 d5d8 d1d8 e7d8 a1d1 h7h6 d1d7 b7c8');
INSERT INTO evaluations VALUES(9096,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(9096,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(9096,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(9096,33,'cp -38',NULL);
INSERT INTO evaluations VALUES(9096,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(9096,35,'cp -45',NULL);
INSERT INTO evaluations VALUES(9096,36,'cp -36','g1e2 e8g8 e1g1 c8f5 d2d4 e5d4 e2d4 c5d4 b2d4 c6d4 e3d4 c7c6 c2c4 d5c4 b3c4 h7h5 b1c3 h5h4 g3h4 f6g4 h2h3 g4h6 d4d5 d8h4 d1f3 h4c4 f1e1 a8d8 a1d1 c6d5 c3d5 f5e6 d5f4 d8d1 e1d1 c4c2 f3d3 c2d3 d1d3 e6a2 g2b7 a2c4 d3d7');
INSERT INTO evaluations VALUES(9097,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(9097,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(9097,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(9097,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(9097,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(9097,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(9097,36,'cp -35','e8g8 e1g1 c8f5 d2d4 e5d4 e3d4 c5d6 c2c4 f8e8 c4c5 d6f8 b1c3 c6e7 e2f4 c7c6 f1e1 b7b5 f4d3 a7a5 h2h4 d8d7 d3e5 d7c8 b2c1 e7g6 e5g6 h7g6 c1g5 e8e1 d1e1 c8d7 e1d2 f8e7 a1e1 a8e8 e1e5 b5b4 c3a4');
INSERT INTO evaluations VALUES(11570,30,'cp 155',NULL);
INSERT INTO evaluations VALUES(11570,31,'cp 167',NULL);
INSERT INTO evaluations VALUES(11570,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(11570,33,'cp 168',NULL);
INSERT INTO evaluations VALUES(11570,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(11570,35,'cp 133',NULL);
INSERT INTO evaluations VALUES(11570,36,'cp 175','f1e1 f7f5 f3d4 d8h4 g2g3 h4f6 a4c6 b7c6 f2f3 f8c5 c2c3 e8g8 f3e4 f5e4 e1f1 f6g6 f1f8 c5f8 c1f4 d7d5 b1a3 c8h3 a3c2 a8e8 c2e3 c6c5 d4b3 c7c6 d1e2 g6f7 b3d2 f8e7 a1e1 e8f8 g1h1 e7f6');
INSERT INTO evaluations VALUES(11571,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(11571,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(11571,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(11571,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(11571,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(11571,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(11571,36,'cp 128','f7f5 f3d4 d8h4 g2g3 h4f6 d4c6 b7c6 b1d2 f8e7 d2e4 f5e4 e1e4 e8g8 d1e2 e7c5 e4f4 f6e7 f4f8 g8f8 e2f3 f8g8 c1f4 a6a5 f3c3 c5b4 c3b3 g8h8 f4c7 c8a6 b3e3 e7e3 f2e3 a6e2 c2c3 b4c5 g1f2 e2g4 c7f4 h7h6 h2h4 h8g8 a4d1 g4e6 d1b3 g8f7 a1d1 f7e7 b3e6 e7e6 f2f3 a5a4 b2b4 a4b3 a2b3 c5e7 f3g4');
INSERT INTO evaluations VALUES(13181,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(13181,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(13181,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(13181,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(13181,34,'cp 48',NULL);
INSERT INTO evaluations VALUES(13181,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(13181,36,'cp 46','d1d8 e8d8 f3g5 c6a5 c4b5 c8e6 g5e6 f7e6 c1e3 a5c6 b5c6 b7c6 e1e2 g7g6 c3a4 d8e8 h1c1 e8f7 c1c6 f6d5 a1c1 f8g7 c1c4 a7a5 e3d2 h8b8 b2b3 d5b4 d2b4 a5b4 c4e4 b8b5 c6c2 g7f6 e4c4 h7h5 g2g3 h5h4 e2f3 g6g5 f3e3 a8d8 a4c5 d8d1 c2c1 d1d4 c5d3 g5g4 d3e5 f6e5');
INSERT INTO evaluations VALUES(2247,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(2247,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(2247,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(2247,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(2247,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(2247,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(2247,36,'cp 44','e8d8 f3g5 c6a5 c4b5 c8e6 g5e6 f7e6 c1e3 a5c6 c3a4 d8e8 e1e2 g7g6 h1c1 f8g7 b5c6 b7c6 c1c6 f6d5 a1c1 e8f7 c1c4 a7a5 e3d2 h8b8 b2b3 d5b4 d2b4 a5b4 c6c5 g7f6 g2g3 h7h5 c4e4 g6g5 c5c4 b8c8 c4c8 a8c8 e4c4 c8b8 h2h3 f7g6 a4c5 b8a8 c4c2 h5h4 g3g4 g6f7');
INSERT INTO evaluations VALUES(2272,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(2272,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(2272,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(2272,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(2272,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(2272,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(2272,36,'cp 47','f3g5 c6a5 c4b5 c8e6 g5e6 f7e6 c1e3 a5c6 c3a4 d8e8 a1c1 e8f7 b5c6 b7c6 c1c6 g7g6 e1e2 f8g7 h1c1 f6d5 c1c4 a7a5 e3d2 h8b8 b2b3 d5b4 d2b4 a5b4 c4e4 b8b5 c6c2 h7h5 g2g3 g7f6 c2c4 a8b8 h2h4 g6g5 h4g5 f6g5 a4c5 b8a8 e4e5 a8a2 e2f3 g5f6 e5h5 a2b2');
INSERT INTO evaluations VALUES(13263,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(13263,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13263,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(13263,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(13263,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13263,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13263,36,'cp 0','g1f3 e7e6 e1g1 f8e7 f1e1 d6d5 c4b5 d5e4 d3e4 d8d1 e1d1 f6e4 f3e5 e4d6 b5c6 b7c6 b1c3 d6f5 e5c6 c8b7 c6a5 b7a6 a5c6');
INSERT INTO evaluations VALUES(14500,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(14500,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(14500,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(14500,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14500,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14500,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(14500,36,'cp 61','d5d4 c3b5 c7c6 e5f6 d8a5 d1d2 a5b5 f6e7 f8e7 d2d4 e7f6 e2e3 b5f5 d4d3 f5e5 e3e4 c8f5 f2f3 a8d8 d3e3 f5g6 f1e2 e8g8 e1g1 f6e7 g1h1 e7d6 g2g3 f7f5 f3f4 e5e4 e3e4 f5e4 c1e3 d6e7 h1g2 g6f7 f1d1 e7f6 c2c3 a7a5 g2f2 a5a4 e3c5');
INSERT INTO evaluations VALUES(14670,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(14670,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(14670,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(14670,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(14670,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(14670,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(14670,36,'cp 106','g5f4 f8e7 e1c1 e8g8 f2f3 c8e6 g2g4 c7c6 h2h4 d8a5 g4g5 f6h5 f4d2 a5b6 d4d3 h5g3 h1h2 g3f1 d1f1 h6h5 f3f4 b6c5 b2b4 c5c4 f4f5 c4d3 c2d3 e6d7 f5f6 e7d8 f6g7 g8g7 c3e2 d7g4 e2g3 d6d5 e4d5 c6d5 d2c3 g7g8 g3f5 g8h7 f5d6 d5d4 c3d4 d8e7 f1f7 f8f7 d6f7 a8c8 c1b2 e7b4');
INSERT INTO evaluations VALUES(14863,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(14863,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14863,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14863,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(14863,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(14863,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(14863,36,'cp -9','f6e4 e2e3 e4c3 b2c3 c6a5 a3a4 f8d6 f1d3 b7b6 e1g1 f7f5 c1a3 e8g8 a3b4 a5c4 d3c4 d5c4 f3d2 c8b7 d2c4 b7g2 g1g2 d6b4 c3b4 d8d5 d1f3 d5c4 b4b5 c7c5 b5c6 a8c8 c2c3 c8c6 f1c1 h7h6 g2h2 c6c8 f3g3 c4c7 g3c7 c8c7');
INSERT INTO evaluations VALUES(17149,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(17149,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(17149,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(17149,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(17149,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(17149,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(17149,36,'cp 36','e5d4 f3d4 f8g7 d4c6 b7c6 b5a4 a8b8 b1c3 e8g8 a4b3 c6c5 c1e3 d7d6 d1d2 c8e6 a1b1 e7c6 e3g5 d8e8 f1e1 e6b3 a2b3 c6d4 b1a1 b8b7 f2f3 c7c6 g5f4 e8e6 f4e3 b7d7 c3e2 d4e2 e1e2 g7b2 a1a6 b2g7');
INSERT INTO evaluations VALUES(17160,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(17160,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(17160,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(17160,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(17160,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(17160,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(17160,36,'cp 38','f3d4 f8g7 d4c6 b7c6 b5a4 a8b8 a4b3 e8g8 b1c3 d7d6 c1e3 c6c5 d1d2 c8e6 a1b1 e7c6 e3g5 d8e8 f1e1 e6b3 a2b3 c6d4 b1a1 b8b7 g5e3 e8e6 f2f3 c7c6 e3f2 f8b8 h2h3 h7h6 f3f4 g8h7 a1a6 e6d7 c3e2 d4e2 d2e2 g7b2 e4e5 d6e5 f4e5');
INSERT INTO evaluations VALUES(17158,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(17158,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(17158,32,'cp 31',NULL);
INSERT INTO evaluations VALUES(17158,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(17158,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(17158,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(17158,36,'cp 39','f8g7 d4c6 b7c6 b5a4 e8g8 b1c3 a8b8 a4b3 d7d6 c1e3 c6c5 d1d2 c8e6 a1b1 e7c6 e3g5 d8e8 f1e1 e6b3 a2b3 c6d4 b1a1 b8b7 f2f3 e8e6 a1a4 f8b8 h2h3 c7c6 g5f4 b7b4 e1a1 h7h6 a4b4 c5b4 c3d1 g6g5 f4e3 f7f5');
INSERT INTO evaluations VALUES(11001,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(11001,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(11001,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(11001,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11001,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(11001,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(11001,36,'cp 35','d4c6 b7c6 b5a4 e8g8 b1c3 a8b8 a4b3 c6c5 c1e3 d7d6 d1d2 c8e6 a1b1 e7c6 e3g5 d8e8 f1e1 e6b3 a2b3 c6d4 b1a1 b8b7 f2f3 e8e6 a1a4 c7c6 c3d1 f8a8 g5e3 e6e7 d1f2 b7b5 e3g5 e7b7 g5f4 d6d5 e4e5 d4e6 f2d3 e6f4 d2f4');
INSERT INTO evaluations VALUES(19352,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(19352,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(19352,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(19352,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(19352,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(19352,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(19352,36,'cp -12','c6d4 a2a4 g8f6 b1c3 a7a6 b5e2 d7d6 d2d3 g7g6 g1f3 f8g7 c1e3 f6d7 d1d2 e8g8 e1g1 e7e6 e2d1 a8b8 d2f2 d4f3 d1f3 b7b5 a4b5 a6b5 d3d4 c5d4 e3d4 b5b4 c3e2 e6e5 d4a7 b8a8 f4f5 d8c7 e2g3 d7c5 a7c5 a8a1 f1a1 d6c5 h2h3 c8b7 g1h1 f8d8 a1a7 c7b6');
INSERT INTO evaluations VALUES(21299,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(21299,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(21299,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(21299,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(21299,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(21299,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(21299,36,'cp 18','g8f6 d1b3 d8d7 f1d1 c6a5 b3b5 a5c4 b5c4 d7c6 c4d3 f8e7 f3d4 c6a6 d3a6 b7a6 d4c6 c8d7 c6e7 e8e7 b2b3 d7e6 f2f4 h8c8 c1a3 e7f8 e4e5 f6e8 c3d5 c8b8 a1c1 e6d5 d1d5 b8b6 g1f2 f8g8 e5d6 c7d6 f2f3 b6b5 d5d2 b5a5 a3b4 a5b5 b4d6 a8d8 c1d1 e8d6 d2d6');
INSERT INTO evaluations VALUES(14654,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(14654,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(14654,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(14654,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(14654,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(14654,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(14654,36,'cp 11','d1b3 d8d7 f1d1 c6a5 b3b4 a5c4 b4c4 d7c6 c4d3 f8e7 f3d4 c6a6 d3a6 b7a6 d4c6 c8d7 c6e7 e8e7 b2b3 d7e6 f2f4 h8c8 c1a3 e7f8 e4e5 f6e8 a1c1 f8g8 c3d5 e6d5 d1d5 a8b8 g1f2 c8d8 c1d1 b8b5 f2f3 f7f6 h2h3 g8f7 e5d6 b5d5 d1d5');
INSERT INTO evaluations VALUES(21787,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21787,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21787,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21787,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21787,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21787,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21787,36,'cp 0','e5d4 e3d4 d8d4 d1e2 d4e5 c1e3 c8e6 e1c1 e6c4 e2c4 f8d6 g1f3 e5f5 c3b5 a7a6 b5d6 c7d6 d1d5 c6e5 c4a4 b7b5 d5b5 f5d7 b5a5 d7a4 a5a4 e5f3 g2f3 g8e7 a4b4 e7f5 c1b1 f7f6 h1d1 e8d7 b4b7 d7c6 b7b6 c6d7');
INSERT INTO evaluations VALUES(21835,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21835,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21835,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21835,33,'cp 1',NULL);
INSERT INTO evaluations VALUES(21835,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21835,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21835,36,'cp 0','e3d4 d8d4 d1e2 d4e5 c1e3 c8e6 e1c1 f8d6 g1f3 e6c4 e2c4 e5e6 c3d5 g8h6 h1e1 e8g8 h2h3 f8e8 g2g4 e6e4 f3d2 e4e6');
INSERT INTO evaluations VALUES(21832,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(21832,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(21832,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(21832,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(21832,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(21832,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21832,36,'cp 0','d8d4 d1e2 d4e5 c1e3 c8e6 e1c1 e6c4 e2c4 f8d6 g1f3 e5f5 c3b5 a7a6 b5d6 c7d6 d1d5 c6e5 c4a4 b7b5 d5b5 f5d7 b5a5 d7a4 a5a4 e5f3 g2f3 g8e7 a4b4 a8c8 c1b1 e8d7 h1d1 e7f5 b4b7 c8c7 b7b6 c7c6 b6b7');
INSERT INTO evaluations VALUES(21795,30,'cp 258',NULL);
INSERT INTO evaluations VALUES(21795,31,'cp 276',NULL);
INSERT INTO evaluations VALUES(21795,32,'cp 258',NULL);
INSERT INTO evaluations VALUES(21795,33,'cp 239',NULL);
INSERT INTO evaluations VALUES(21795,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(21795,35,'cp 251',NULL);
INSERT INTO evaluations VALUES(21795,36,'cp 237','d4d5 c6b8 f3e5 g8f6 d1d4 f8e7 c1f4 e8g8 e1c1 b8d7 e5d7 c8d7 f1d3 b7b6 d4c4 e7d6 f4d6 c7d6 c4d4 f8e8 f2f3 d8e7 h1f1 e7e5 d4h4 h7h6 f1e1 e5g5 h4g5 h6g5 e1e8 a8e8 h2h3 g8f8 c1d2 e8e5');
INSERT INTO evaluations VALUES(24434,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(24434,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(24434,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(24434,33,'cp 81',NULL);
INSERT INTO evaluations VALUES(24434,34,'cp 76',NULL);
INSERT INTO evaluations VALUES(24434,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(24434,36,'cp 66','d4f5 d7f5 e4f5 f8e7 c1g5 e8g8 f1e1 f6e8 g5e7 d8e7 c3d5 e7d8 c2c3 e8f6 b5c4 a8c8 d1e2 f6d5 c4d5 d8f6 d5e4 b7b6 e2f3 c6e7 a1d1 c8c5 d1d2 f8d8 e1d1 h7h6 g2g4 d6d5 e4d5 c5d5 d2d5 d8d5 d1d5 e7d5 f3d5 f6g5 d5d1 h6h5 h2h3 h5g4 h3g4 g5f4 g1g2 f4c4 d1f3 c4a2 f3a8 g8h7');
INSERT INTO evaluations VALUES(24753,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(24753,31,'cp 64',NULL);
INSERT INTO evaluations VALUES(24753,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(24753,33,'cp 98',NULL);
INSERT INTO evaluations VALUES(24753,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(24753,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(24753,36,'cp 80','f1b5 f6d7 c3c4 d5c4 e1g1 e7e6 b1d2 f5d3 f1e1 c6a5 f3e5 c7c6 e5c4 d3c4 b5c4 a5c4 d2c4 f8b4 e1f1 d7b6 c4b6 a7b6 d1g4 g7g6 g4g3 h6h5 f4c7 d8d5 c7b6 h5h4 g3e5 d5e5 d4e5 h4h3 g2g3 a8a6 b6d4 e8g8 f1c1 c6c5 d4c5 b4c5 c1c5 f8d8 g1f1 d8d2');
INSERT INTO evaluations VALUES(25120,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(25120,31,'cp 109',NULL);
INSERT INTO evaluations VALUES(25120,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(25120,33,'cp 112',NULL);
INSERT INTO evaluations VALUES(25120,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(25120,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(25120,36,'cp 132','d2d4 d8e7 c1e3 e8c8 b1c3 f6g4 f1e2 g4e3 f2e3 g7g5 d4d5 e6g4 d1c2 c6b4 c2d2 g4f3 e2f3 f8g7 e1g1 h7h5 a1e1 g5g4 f3e4 c8b8 e4f5 b4c6 c3e4 h5h4 d2e2 c6e5 f5g4 e5g4 e2g4 g7b2 e4f2 b2c3 e1e2 d8g8 g4e4 g8e8 e4e7 e8e7 f2d3 c3g7');
INSERT INTO evaluations VALUES(28475,30,'cp 188',NULL);
INSERT INTO evaluations VALUES(28475,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(28475,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(28475,33,'cp 187',NULL);
INSERT INTO evaluations VALUES(28475,34,'cp 179',NULL);
INSERT INTO evaluations VALUES(28475,35,'cp 182',NULL);
INSERT INTO evaluations VALUES(28475,36,'cp 181','d2d4 c5b6 f1e2 b8d7 h2h3 c7c6 e2d3 e8g8 e1g1 f8e8 c1e3 b6c7 c3e2 e6d5 e2f4 d5e4 d3e4 f6e4 c2c3 d7b6 f4d3 b6c4 f1e1 d8d5 d1c2 d5f5 a1d1 c4e3 e1e3 e4g5 e3e8 a8e8 f3g5 f5g5 c2d2 g5g6 g2g3 h7h6');
INSERT INTO evaluations VALUES(28474,30,'cp 176',NULL);
INSERT INTO evaluations VALUES(28474,31,'cp 186',NULL);
INSERT INTO evaluations VALUES(28474,32,'cp 180',NULL);
INSERT INTO evaluations VALUES(28474,33,'cp 186',NULL);
INSERT INTO evaluations VALUES(28474,34,'cp 172',NULL);
INSERT INTO evaluations VALUES(28474,35,'cp 162',NULL);
INSERT INTO evaluations VALUES(28474,36,'cp 177','c5b4 f1d3 c7c5 d4c5 e8g8 e1g1 b8d7 c5c6 b7c6 c3e4 d7c5 e4c5 b4c5 f3g5 f8e8 g5e6 e8e6 d1f3 a7a5 c1f4 d8d5 f1e1 a8e8 e1e6 d5e6 a1d1 g7g6 f4g3 f6e4 g3h4 e4d2 f3f6 e6f6 h4f6 d2e4');
INSERT INTO evaluations VALUES(29583,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(29583,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(29583,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(29583,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(29583,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(29583,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(29583,36,'cp 0','h7h6 e1g1 e7e6 c2c4 f8e7 b1c3 e8g8 c4d5 e6d5 d1b3 b8c6 c1d2 c6a5 b3a4 a5c6');
INSERT INTO evaluations VALUES(30292,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(30292,31,'cp 83',NULL);
INSERT INTO evaluations VALUES(30292,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(30292,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(30292,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(30292,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(30292,36,'cp 53','f8e7 d1d2 b8c6 e1c1 c6d4 d2d4 e8g8 f2f3 c7c6 g5f4 f6e8 c1b1 b7b5 h2h4 d7e6 c3e2 c6c5 d4d2 b5b4 f4g5 a7a5 g2g4 a5a4 e2g3 h7h6 g5e7 d8e7 g4g5 g7g6 f1e2 a4a3 b2b3 h6h5 f3f4 a8d8 c2c4 b4c3 d2c3 e8g7 c3f6 f8e8 e2b5');
INSERT INTO evaluations VALUES(28504,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(28504,31,'cp 93',NULL);
INSERT INTO evaluations VALUES(28504,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(28504,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(28504,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(28504,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(28504,36,'cp 55','d1d2 b8c6 e1c1 c6d4 d2d4 e8g8 f2f3 c7c6 g5f4 f6e8 h2h4 b7b5 c1b1 d7e6 c3e2 d8a5 a2a3 c6c5 d4d2 b5b4 f4g5 e7f6 g5f6 e8f6 e2f4 a8d8 f4e6 f7e6 f1c4 d6d5 e4d5 f6d5 c4d5 d8d5 d2e3 d5d1 h1d1 b4a3 b2a3 a5b6 b1a2 h7h6 d1d7 f8f7 d7f7 g8f7 e3f4 f7e7 h4h5 b6b5');
INSERT INTO evaluations VALUES(30709,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30709,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(30709,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(30709,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(30709,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(30709,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30709,36,'cp 28','f8b4 c1d2 d8e7 d2b4 e7b4 b1c3 b4b2 g1e2 c7c6 h2h4 b2b4 a1b1 b4c5 d1d2 g8h6 h1h3 e8g8 h3g3 g8h8 d2g5 f8g8 d5c6 c5c6 e2d4 c6c8 d4b5 f7f6 g5d2 h6f7 b1d1 a7a6 b5d6 f7d6 d2d6 g8e8 f1e2 b7e4 e1f1 e8e6 d6f4 e4b7 e2g4 e6e5 h4h5 h7h6 d1d6 c8e8 g4e2 e8e7 d6b6 d7d5 g3e3 b8c6 c3d5 e7c5 f4c4 e5d5 b6b7 c5c4 e2c4 d5h5');
INSERT INTO evaluations VALUES(31161,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(31161,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31161,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31161,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31161,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31161,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31161,36,'cp 0','d8b6 d1c1 e7e6 g1f3 b8d7 f1e2 f7f6 e5f6 g8f6 c2c4 f8e7 h2h3 e8g8 e1g1 h7h6 b1d2 f6e4 d2e4 f5e4 f3d2 e4f5 d2f3');
INSERT INTO evaluations VALUES(33989,30,'cp 488',NULL);
INSERT INTO evaluations VALUES(33989,31,'cp 486',NULL);
INSERT INTO evaluations VALUES(33989,32,'cp 475',NULL);
INSERT INTO evaluations VALUES(33989,33,'cp 479',NULL);
INSERT INTO evaluations VALUES(33989,34,'cp 482',NULL);
INSERT INTO evaluations VALUES(33989,35,'cp 500',NULL);
INSERT INTO evaluations VALUES(33989,36,'cp 498','d2d3 h4f2 e2f2 e4f2 e1f2 c7c6 e5f3 d7d5 c1f4 f8e7 d3d4 h7h6 f1d3 g7g5 f4d2 c8e6 b1c3 b8d7 a1e1 e8c8 h1f1 d8g8 c3e2 g5g4 f3g1 c6c5 c2c3 h6h5 e2f4 h5h4 f4e6 f7e6 e1e6 e7g5 d2g5 g8g5 e6e7 g4g3 h2g3 h4g3 f2e3 c5c4');
INSERT INTO evaluations VALUES(34395,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(34395,31,'cp 95',NULL);
INSERT INTO evaluations VALUES(34395,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(34395,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(34395,34,'cp 92',NULL);
INSERT INTO evaluations VALUES(34395,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(34395,36,'cp 90','e1g1 g8f6 a2a4 e7c7 f3h4 b8a6 f2f4 f8d6 h4f3 c8e6 c4a6 b7a6 g1h1 e8g8 d1e2 e6g4 f4f5 d6b4 e2e1 b4c3 b2c3 c6c5 e1h4 c7d6 h2h3 g4f3 g2f3 g8h7 f1g1 f8g8 c1g5 a8d8 g1g2 a6a5 g5e3');
INSERT INTO evaluations VALUES(34394,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(34394,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(34394,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(34394,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(34394,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(34394,35,'cp 85',NULL);
INSERT INTO evaluations VALUES(34394,36,'cp 70','a7a5 a2a4 g8f6 h2h3 e7c7 f3h4 b8d7 d1e2 d7c5 c1e3 c5e6 a1d1 f8b4 h4f5 e8g8 c3a2 b4e7 a2c1 e6c5 f5e7 c7e7 f2f3 c8e6 b2b3 f8d8 d1d8 a8d8 e2f2 f6d7 c4e2 b7b6 f1d1 g8h8 f2e1 d7f8 d1d8 e7d8 e1c3 f8d7 c1d3 c5d3 c3d3 d8c7 d3a6 f7f6 a6a8 h8h7');
INSERT INTO evaluations VALUES(35254,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(35254,31,'cp 141',NULL);
INSERT INTO evaluations VALUES(35254,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(35254,33,'cp 135',NULL);
INSERT INTO evaluations VALUES(35254,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(35254,35,'cp 164',NULL);
INSERT INTO evaluations VALUES(35254,36,'cp 162','b1c3 f7f5 g2g4 f5g4 h2h3 e7h4 f1g2 g4g3 f2f4 h4h5 g1f3 f8b4 e2e4 b4c3 b2c3 d7d6 e1g1 g8e7 d1e1 e8g8 e1g3 b8d7 f3g5 d7f6 e4e5 e7f5 g3d3 f6d7 c1a3 h7h6 g5e4 h5g6 e5d6 f5d6 a1e1 f8f7 f1f2 d6f5 a3b4 d7f6 e4f6 g7f6');
INSERT INTO evaluations VALUES(35550,30,'cp 1094',NULL);
INSERT INTO evaluations VALUES(35550,31,'cp 1101',NULL);
INSERT INTO evaluations VALUES(35550,32,'cp 1109',NULL);
INSERT INTO evaluations VALUES(35550,33,'cp 1147',NULL);
INSERT INTO evaluations VALUES(35550,34,'cp 1140',NULL);
INSERT INTO evaluations VALUES(35550,35,'cp 1216',NULL);
INSERT INTO evaluations VALUES(35550,36,'cp 1238','h5e5 e7f7 f1c4 d7d5 c4d5 d8d5 e5d5 f7e8 d5h5 e8d8 h5f7 g8e7 b1c3 c8d7 c1g5 h7h6 g5f4 e7c6 e1c1 d8c8 c3d5 f8d6 f4d6 c7d6 f7g7 h8d8 g7h6 b8a6 h6d6 b7b6 d5f6 a6b8 g1e2 d7g4 f6g4 d8d6 d1d6 c8b7 h1d1 b8a6 d6d7 a6c7 e2f4 a8c8 h2h4 b7b8 h4h5 c8e8 f4d5 c7d5 e4d5 c6e7 g4f6 e8f8');
INSERT INTO evaluations VALUES(35548,30,'cp 1238',NULL);
INSERT INTO evaluations VALUES(35548,31,'cp 1219',NULL);
INSERT INTO evaluations VALUES(35548,32,'cp 1231',NULL);
INSERT INTO evaluations VALUES(35548,33,'cp 1232',NULL);
INSERT INTO evaluations VALUES(35548,34,'cp 1221',NULL);
INSERT INTO evaluations VALUES(35548,35,'cp 1225',NULL);
INSERT INTO evaluations VALUES(35548,36,'cp 1232','e7f7 f1c4 d7d5 c4d5 d8d5 e5d5 f7e8 d5h5 e8d8 h5f7 g8e7 b1c3 c8d7 c1g5 h7h6 g5f4 d8c8 e1c1 e7c6 c3d5 f8d6 f4d6 c7d6 f7g7 h8d8 g7h6 b7b6 h6d6 b8a6 d5f6 a6b8 g1e2 d7g4 f6g4 d8d6 d1d6 a7a5 e2f4 a8a7 f4e6 a5a4 a2a3 c8b7 h1d1 a7a6 h2h4 c6a7 d6d5 a7c6 f2f3 a6a7 h4h5 c6e7 d5d6 e7c6 h5h6');
INSERT INTO evaluations VALUES(35537,30,'cp 1254',NULL);
INSERT INTO evaluations VALUES(35537,31,'cp 1277',NULL);
INSERT INTO evaluations VALUES(35537,32,'cp 1279',NULL);
INSERT INTO evaluations VALUES(35537,33,'cp 1278',NULL);
INSERT INTO evaluations VALUES(35537,34,'cp 1279',NULL);
INSERT INTO evaluations VALUES(35537,35,'cp 1294',NULL);
INSERT INTO evaluations VALUES(35537,36,'cp 1298','f1c4 d7d5 c4d5 d8d5 e5d5 f7e8 d5h5 e8d8 h5f7 g8e7 b1c3 c8d7 c1g5 h7h6 g5f4 e7c6 e1c1 d8c8 c3d5 f8d6 f4d6 c7d6 f7g7 h8d8 g7h6 b8a6 h6d6 b7b6 d5f6 a6b8 g1e2 d7g4 f6g4 d8d6 d1d6 c8c7 h1d1 c7b7 e2f4 b8a6 d6d7 a6c7 f4d5 a8c8 c1d2 c6b8 d7c7 c8c7 d5c7 b7c7 d2e3 b8c6 a2a3 a7a5 e4e5 c6e7 e5e6 e7f5 e3f4 f5h4 g2g3 h4g6 f4f5');
INSERT INTO evaluations VALUES(35535,30,'cp 1278',NULL);
INSERT INTO evaluations VALUES(35535,31,'cp 1286',NULL);
INSERT INTO evaluations VALUES(35535,32,'cp 1290',NULL);
INSERT INTO evaluations VALUES(35535,33,'cp 1285',NULL);
INSERT INTO evaluations VALUES(35535,34,'cp 1262',NULL);
INSERT INTO evaluations VALUES(35535,35,'cp 1278',NULL);
INSERT INTO evaluations VALUES(35535,36,'cp 1280','d7d5 c4d5 d8d5 e5d5 f7e8 d5h5 e8d8 h5f7 g8e7 b1c3 c8d7 c1g5 b8c6 c3d5 d8c8 d5e7 f8e7 g5e7 c6e7 f7e7 h8e8 e7g7 e8e4 g1e2 e4e2 e1e2 b7b6 f2f3 a7a5 e2f2 h7h6 g7h6 a5a4 h6h8 c8b7 h8d4 b7c8 g2g4 d7c6 d4f4 c8b7 c2c4 c6d7');
INSERT INTO evaluations VALUES(35553,30,'cp 909',NULL);
INSERT INTO evaluations VALUES(35553,31,'cp 905',NULL);
INSERT INTO evaluations VALUES(35553,32,'cp 917',NULL);
INSERT INTO evaluations VALUES(35553,33,'cp 928',NULL);
INSERT INTO evaluations VALUES(35553,34,'cp 921',NULL);
INSERT INTO evaluations VALUES(35553,35,'cp 932',NULL);
INSERT INTO evaluations VALUES(35553,36,'cp 936','e7f6 g1f3 d8e7 b1c3 b8c6 f1c4 d7d5 c4d5 c8g4 e1g1 g4f3 d5c6 b7c6 g2f3 a7a5 d2d3 a5a4 f3f4 f6g7 c1e3 g8f6 h7h4 a4a3 e4e5 a3b2 a1b1 f6d5 h4e7 f8e7 c3d5 c6d5 b1b2 g7f8 d3d4 e7e6 g1g2 f8e7 g2g3 a8a3 h2h4 a3a4');
INSERT INTO evaluations VALUES(36626,30,'cp -34',NULL);
INSERT INTO evaluations VALUES(36626,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(36626,32,'cp -30',NULL);
INSERT INTO evaluations VALUES(36626,33,'cp -25',NULL);
INSERT INTO evaluations VALUES(36626,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(36626,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(36626,36,'cp -19','c7c5 d4c5 d6d5 f3d4 d8c7 b2b4 a7a5 b4a5 c7c5 d4b3 c5c7 b1a3 b8c6 a3b5 c7d8 c1b2 c6a5 b3a5 a8a5 a2a4 c8d7 d1e1 f6e8 b2g7 e8g7 e1c3 d8b6 c3e5 b6e6 f1b1 f8c8 b5d4 e6e5 f4e5 c8c7 b1b4 g7f5 d4f5 g6f5');
INSERT INTO evaluations VALUES(38214,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(38214,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(38214,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(38214,33,'cp 140',NULL);
INSERT INTO evaluations VALUES(38214,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(38214,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(38214,36,'cp 123','d1h5 g7g6 h5d1 e7f8 b1c3 g8f6 d1e2 d8e7 c1e3 f8g7 e1c1 e8g8 f2f3 b8c6 h1e1 f8e8 g2g4 c6e5 c4b3 f6g4 f3g4 c8g4 e2g2 g4d1 e1d1 c7c6 d1g1 e7f6 c1b1 g8h7 g2h3 b7b5 a2a3 f6e7 c3e2 a7a5 b3a2');
INSERT INTO evaluations VALUES(38490,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(38490,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(38490,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(38490,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(38490,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(38490,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(38490,36,'cp 71','g8f6 g1f3 c7c5 d4d5 e8g8 e1g1 a7a6 a2a4 c8g4 e3e4 f6e8 f1e1 b8d7 c4f1 e8d6 h2h3 g4f3 d1f3 b7b5 c1f4 b5b4 c3d1 g7d4 e4e5 d6f5 a4a5 g8h8 f3b3 d8c7 d1e3 d4e5 f4e5 d7e5 e3f5 g6f5 b3c2 e7e6 d5e6 f7e6 c2e2 e5c6 e2e6 c6d4 e6e7');
INSERT INTO evaluations VALUES(39726,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(39726,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(39726,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(39726,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(39726,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(39726,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(39726,36,'cp -20','b7b5 c4d3 e8g8 a2a4 b5b4 c3a2 e6e5 d4f3 b8c6 c2c3 b4c3 a2c3 c8e6 f3g5 e6g4 d3e2 g4e2 d1e2 d8c8 g5f3 h7h6 e1d1 a8b8 a1b1 b8b4 c1d2 c8b7 d2e1 f8c8 h2h3 c8b8 c3d5 f6d5 e4d5 c6d4 f3d4 b4d4 b2b3 d4d5 d1d5 b7d5');
INSERT INTO evaluations VALUES(40428,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40428,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40428,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40428,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40428,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40428,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40428,36,'cp 0','e8g8 h2h3 c5d4 c3d4 b8c6 c1e3 d8a5 e3d2 a5a3 a1b1 c6d4 d2b4 d4f3 e1e2 f3d4 e2e1 d4f3');
INSERT INTO evaluations VALUES(44707,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(44707,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(44707,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(44707,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(44707,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(44707,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(44707,36,'cp 51','e2e3 f8g7 g1f3 e8g8 f1e2 d5c4 e2c4 b8d7 h2h3 b7b5 c4e2 a7a6 f3e5 c8b7 e2f3 d8b6 b2b4 a6a5 a2a3 d7e5 f4e5 f6d5 c3d5 c6d5 e5g7 g8g7 e1g1 a5b4 a3b4 a8a1 d1a1 f8c8 a1b2 e7e6 f1a1 b6c6 f3e2 c6b6 b2d2 h7h6 e2d3 b7c6');
INSERT INTO evaluations VALUES(44705,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(44705,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44705,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(44705,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(44705,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(44705,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(44705,36,'cp 45','f8g7 g1f3 e8g8 a1c1 c8e6 f3g5 e6g4 d1b3 h7h6 g5f3 d8b6 f3e5 g4e6 f1e2 d5c4 e5c4 b6b3 a2b3 f6d5 f4g3 d5b4 e1g1 c6c5 e2f3 e6c4 b3c4 b8c6 d4d5 c6e5 g3e5 g7e5 f1d1 a7a6 g2g4 a8c8 g1g2 e5d6 h2h3 b7b6 f3e2');
INSERT INTO evaluations VALUES(47507,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(47507,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47507,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(47507,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(47507,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(47507,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47507,36,'cp 0','e4c3 b2c3 c7c5 g1f3 e7e6 f1e2 h7h6 e1g1 b8c6 a2a4 c8d7 d1d2 d8a5 h2h4 f8e7 h4h5 e8g8 d2e3 a5c7 e3f4 d7e8 f4g3 g8h8 c1h6 g7h6 g3f4 h8h7 e2d3 h7g7 f4g4 g7h8 g4f4');
INSERT INTO evaluations VALUES(47945,30,'cp 58',NULL);
INSERT INTO evaluations VALUES(47945,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(47945,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(47945,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(47945,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(47945,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(47945,36,'cp 56','f1d3 f8e7 e1g1 e8g8 f1e1 b8c6 h2h3 d6d5 a2a3 e7d6 b1c3 h7h6 c3b5 f8e8 b5d6 d8d6 e1e8 f6e8 c2c3 c6e7 g2g4 d6f6 d1e2 e8d6 f3e5 e7g6 a3a4 a7a5 b2b3 c7c6 d3g6 f7g6 e2d3 c8e6 c3c4 d5c4 b3c4 e6f7 c4c5');
INSERT INTO evaluations VALUES(49216,30,'cp -15',NULL);
INSERT INTO evaluations VALUES(49216,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49216,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49216,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49216,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49216,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49216,36,'cp 0','d7d5 c2c4 d5c4 f1c4 f8d6 b1c3 b8d7 g1f3 e8g8 d1c2 f6f5 e1g1 g8g7 f1d1 d7f6 a1c1 a7a6 c3d5 f6e4 d5c3 d8e7 c3d5');
INSERT INTO evaluations VALUES(52226,30,'cp 454',NULL);
INSERT INTO evaluations VALUES(52226,31,'cp 451',NULL);
INSERT INTO evaluations VALUES(52226,32,'cp 444',NULL);
INSERT INTO evaluations VALUES(52226,33,'cp 438',NULL);
INSERT INTO evaluations VALUES(52226,34,'cp 439',NULL);
INSERT INTO evaluations VALUES(52226,35,'cp 423',NULL);
INSERT INTO evaluations VALUES(52226,36,'cp 424','f3e5 g8e7 e5c4 e7c6 c4e3 d7d6 f1c4 b8d7 a2a3 d7e5 c4a2 g7g6 f2f4 e5f7 f4f5 f8g7 d2d3 e8g8 e1g1 g8h8 c3d5 f7h6 c2c3 a7a6 a3a4 g6f5 e4f5 b7b5 d1h5 b5a4 d5f4 d8e8 h5h3 a8b8 f4e6 c8e6 a2e6 c6e5 e3d5 h6g4 c1f4 b8b2');
INSERT INTO evaluations VALUES(56974,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(56974,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(56974,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(56974,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(56974,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(56974,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(56974,36,'cp 0','g8f6 b1c3 c7c6 c4b3 f8c5 d2d3 d7d6 f3g3 f6h5 g3f3');
INSERT INTO evaluations VALUES(57242,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(57242,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(57242,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(57242,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(57242,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(57242,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(57242,36,'cp -13','d7d5 d2d4 e5e4 c1f4 f7f5 e2e3 c7c6 h2h4 g8f6 g1h3 b8d7 c3e2 f8e7 e2g3 d7b6 g3h5 g7g6 h5f6 e7f6 f4g5 f6g5 h4g5 h7h6 g5h6 h8h6 f2f4 c8e6 d1d2 d8c7 e1c1 e8c8 f1e2 d8h8 d2a5 c8b8 a5b4 h6h4 a3a4 e6f7 a4a5 b6d7 b4a3');
INSERT INTO evaluations VALUES(689,30,'cp 319',NULL);
INSERT INTO evaluations VALUES(689,31,'cp 303',NULL);
INSERT INTO evaluations VALUES(689,32,'cp 306',NULL);
INSERT INTO evaluations VALUES(689,33,'cp 316',NULL);
INSERT INTO evaluations VALUES(689,34,'cp 325',NULL);
INSERT INTO evaluations VALUES(689,35,'cp 320',NULL);
INSERT INTO evaluations VALUES(689,36,'cp 310','f3h5 f6f7 h5f7 e8f7 d2d3 g7g5 b1d2 f8b4 e1g1 h7h6 d2c4 b4d6 f1e1 h8e8 c1d2 c8f5 f2f3 f7f6 e1e2 f5e6 c4a5 c6c5 b2b3 h6h5 d2c3 e6d5 a1e1 e8e6 a5c4 a8e8 c4e3 d5b7 g1f2 b7a6 g2g3 f6f7 e3c4 a6c4 b3c4 f7g6 e2e4 e8f8 h2h4 g5g4');
INSERT INTO evaluations VALUES(5217,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(5217,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(5217,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(5217,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(5217,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(5217,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(5217,36,'cp 0','g8f6 d4c6 b7c6 f1d3 d7d5 e4d5 d8e7 d1e2 f6d5 e1g1 e7e2 c3e2 e8g8 h2h3 f8d8 f1d1 b4e7 c1d2 a7a5 d3e4 e7f6 a1c1 c8a6 e2c3 d5b4 d2f4 b4d3 e4d3 d8d3 d1d3 a6d3 b2b3 a8e8 f4e3 h7h6 c3a4 d3b5 a4c5 e8d8');
INSERT INTO evaluations VALUES(5767,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(5767,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(5767,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(5767,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(5767,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(5767,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(5767,36,'cp 31','c6e5 f3e5 d6e5 d1e2 g8f6 c1e3 e8g8 e1g1 c7c6 d5c6 b7c6 b1d2 f6g4 e3b6 a7b6 h2h3 g4f6 f1d1 h7h6 a2a4 d8e7 b2b3 c8d7 a1c1 f8d8 d2f3 c6c5 c4d5 a8a7 e2b2 d8e8 b2c2 d7e6 d5e6 e7e6 d1d2 a7d7 d2d7 e6d7');
INSERT INTO evaluations VALUES(5760,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(5760,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(5760,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(5760,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(5760,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(5760,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(5760,36,'cp 32','f3e5 d6e5 e1g1 g8f6 d1f3 e8g8 c1e3 c8d7 b1c3 d8e7 a1c1 a8c8 b2b3 f6e8 c4d3 e8d6 f1e1 a7a6 c1c2 h7h6 h2h3 b6e3 f3e3 e7g5 e3g5 h6g5 f2f3 f8d8 g1f2 c7c6 e1c1 f7f6 d3e2 g8f8 a2a4 f6f5');
INSERT INTO evaluations VALUES(5759,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(5759,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(5759,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(5759,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(5759,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(5759,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(5759,36,'cp 37','d6e5 e1g1 g8f6 d1e2 c8g4 e2d3 e8g8 h2h3 g4d7 c1e3 b6e3 d3e3 c7c6 b1c3 d8b6 e3b6 a7b6 a2a4 f8d8 b2b3 c6d5 c3d5 f6d5 e4d5 g8f8 a1e1 d8e8 e1e4 f7f5 e4e2 a8c8 f1e1 e5e4 d5d6 c8c6 e1d1 e8d8');
INSERT INTO evaluations VALUES(5761,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(5761,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(5761,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(5761,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(5761,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(5761,35,'cp 41',NULL);
INSERT INTO evaluations VALUES(5761,36,'cp 41','e1g1 g8f6 d1e2 c8g4 e2d3 e8g8 h2h3 g4d7 c1e3 f6h5 f1e1 d8f6 e3b6 a7b6 d3e3 h5f4 c4f1 f6g6 g1h2 d7b5 g2g3 b5f1 e1f1 f4h5 b1c3 f7f5 e4f5 f8f5 a1d1 a8d8 c3e4 f5f8 d1d2 h5f6 e4f6 g6f6 h2g2 h7h6 f1e1 f8e8 e3e4 f6d6 e1d1 d8d7');
INSERT INTO evaluations VALUES(6360,30,'cp 151',NULL);
INSERT INTO evaluations VALUES(6360,31,'cp 163',NULL);
INSERT INTO evaluations VALUES(6360,32,'cp 150',NULL);
INSERT INTO evaluations VALUES(6360,33,'cp 123',NULL);
INSERT INTO evaluations VALUES(6360,34,'cp 138',NULL);
INSERT INTO evaluations VALUES(6360,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(6360,36,'cp 154','c2c3 b4f8 f1d3 d7d6 f2f4 c6e7 g1f3 h7h5 c1e3 g8h6 d1e2 e7f5 e3f2 f8e7 h2h3 h5h4 e1g1 d6d5 b2b4 c8d7 a2a4 a7a5 b4b5 c7c6 b1d2 e8g8 f1b1 c6b5 a4b5 a8c8 b1c1 g7g6 e2e1 g8g7 d2f1 b7b6 g1h1 c8c7 f1e3');
INSERT INTO evaluations VALUES(6359,30,'cp 154',NULL);
INSERT INTO evaluations VALUES(6359,31,'cp 136',NULL);
INSERT INTO evaluations VALUES(6359,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(6359,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(6359,34,'cp 136',NULL);
INSERT INTO evaluations VALUES(6359,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(6359,36,'cp 139','b4f8 f1d3 d7d6 f2f4 c6e7 g1f3 h7h5 c1e3 g8h6 e3f2 e7f5 d1e2 f8e7 b1d2 h6g4 f2g1 g4h6 e1c1 c8d7 g2g3 d6d5 g1f2 h6g4 h2h3 g4f2 e2f2 h5h4 g3g4 f5g3 h1e1 d7a4 b2b3 a4d7 c1b2 c7c6 c3c4 d8b6 c4c5 b6c7 f3g5 g7g6 a2a3 e7g5 f4g5 e8c8');
INSERT INTO evaluations VALUES(13336,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(13336,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(13336,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(13336,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13336,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13336,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13336,36,'cp 0','g7g6 d2d4 c5d4 f3d4 c6d4 d1d4 f8g7 f1b5 c8d7 b5d7 d8d7 c1e3 e8g8 d4d3 b7b5 e1g1 a7a6 a1d1 d7c6 f2f4 a8c8 e3f2 c8c7 d3e3 c7c8');
INSERT INTO evaluations VALUES(15265,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(15265,31,'cp -69',NULL);
INSERT INTO evaluations VALUES(15265,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(15265,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(15265,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(15265,35,'cp -55',NULL);
INSERT INTO evaluations VALUES(15265,36,'cp -53','a7a6 b5d4 c6d4 d2d4 c7c5 d4d2 h5f4 d2f4 g7g5 f4d2 c8e6 e1c1 f8g7 g1f3 h7h6 g2g4 d8c7 h2h4 h6h5 c1b1 h5g4 d2g5 h8g8 e2e4 d5e4 f3d2 c7e5 g5e5 g7e5 d2e4 a8d8 d1d8 e8d8 f1g2 b7b6 e4g5 d8e7 h1e1 g8d8 g5e6 f7e6 c2c3');
INSERT INTO evaluations VALUES(16816,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(16816,31,'cp 79',NULL);
INSERT INTO evaluations VALUES(16816,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(16816,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(16816,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(16816,35,'cp 83',NULL);
INSERT INTO evaluations VALUES(16816,36,'cp 57','f3e5 c6e5 b1c3 d7d6 f2f4 e5d7 e3e4 f8e7 d1f3 e7f6 c1d2 d7c5 e1c1 f6c3 d2c3 e8g8 e4e5 d8h4 e5d6 c7d6 d1d6 f8e8 f1d3 c5d3 d6d3 c8f5 d3e3 a8d8 g2g3 h4h3 e3e8 d8e8 f3f2 f5g4 b2b3 f7f6 h1e1 e8e1 c3e1 h7h6 e1c3 g4f5 c1b2 h3h5 a2a4 h5d1');
INSERT INTO evaluations VALUES(16965,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(16965,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(16965,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(16965,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(16965,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(16965,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(16965,36,'cp 13','d7d6 e1g1 f8e7 a2a4 e8g8 b1c3 a7a5 h2h3 c8e6 a1d1 c7c6 b2b3 h7h6 g5h4 a8c8 f1e1 f6d7 h4e7 d8e7 f3e2 e6c4 d3c4 d7c5 e2d2 f8d8 f2f3 e7h4 e1f1 c5e6 c3e2 c8a8');
INSERT INTO evaluations VALUES(17079,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(17079,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(17079,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(17079,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(17079,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(17079,35,'cp -50',NULL);
INSERT INTO evaluations VALUES(17079,36,'cp -37','e2d4 f8c5 d4f5 d8f6 d1f3 e4g5 f3e2 g5e6 d2d4 f6f5 d4c5 e8g8 b5d3 f5c5 c1e3 c5e5 e1c1 d7d5 h2h4 c8d7 h1e1 e6f4 e2f3 f4d3 d1d3 c7c6 e3d2 e5d6 h4h5 f8e8 e1e8 a8e8 d3e3 c6c5 h5h6 e8e3 f3e3 g7h6 a2a3 d5d4 e3h6 d6h6 d2h6 f7f6 g2g3 g8f7 c2c3 f7g6 h6f4 d4c3');
INSERT INTO evaluations VALUES(17076,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(17076,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(17076,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(17076,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(17076,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(17076,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(17076,36,'cp -39','f8c5 d4f5 d8f6 d1f3 e4g5 f3e2 g5e6 d2d4 f6f5 d4c5 e8g8 b5d3 f5c5 e1g1 d7d5 e2e5 f8e8 f1e1 c8d7 c1e3 c5c6 b2b4 a7a6 a1d1 a8d8 e5g3 d5d4 e3h6 c6d6 e1e6 d6g3 e6e8 d8e8 h2g3 g7h6 g1f1 g8g7 a2a3 e8e5 d1e1 g7f6');
INSERT INTO evaluations VALUES(17505,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(17505,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(17505,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(17505,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(17505,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(17505,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(17505,36,'cp 17','g1f3 b7b5 g2g3 e7e6 f1g2 f8e7 e1g1 a8b8 d2d3 b5b4 c3e2 g8f6 e4e5 f6d5 c2c4 b4c3 b2c3 e8g8 c3c4 d5b6 a1b1 c8b7 c1e3 d7d6 e5d6 e7d6 f3g5 d6e7 g5e4 b6d7 g3g4 b7a8 b1b8 d8b8 f4f5 e6f5 g4f5 c6e5 e3f4 a8e4 g2e4 e7f6 e2g3 a6a5 g3h5');
INSERT INTO evaluations VALUES(17504,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(17504,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(17504,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(17504,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(17504,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(17504,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17504,36,'cp 9','b7b5 g2g3 e7e6 f1g2 f8e7 e1g1 g8h6 h2h3 c8b7 d2d3 e8g8 f4f5 g8h8 g3g4 c6d4 c1e3 b5b4 c3e2 d4f3 g2f3 e6f5 e4f5 b7f3 f1f3 h6g8 d1d2 d8c7 a1f1 a8e8 e2f4 e7h4 c2c4 g8f6 d2g2 h4g5 g2f2 e8e5 d3d4 c5d4 e3d4');
INSERT INTO evaluations VALUES(20914,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(20914,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(20914,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(20914,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(20914,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(20914,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(20914,36,'cp 96','a7a6 c1e3 f8g7 c2c4 c8g4 b1d2 e7e6 c4d5 d8d5 f1c4 g4f3 c4d5 f3d1 d5c6 b7c6 a1d1 c6c5 d2e4 c5d4 e3d4 g8e7 e1e2 e7d5 d1d2 e8g8 g2g3 f8b8 h1c1 b8b5 f2f4 g7f8 c1c6 f8b4 d2d1 b4e7 e4f6 d5f6 e5f6 e7d6 d4c3 a6a5');
INSERT INTO evaluations VALUES(21113,30,'cp -66',NULL);
INSERT INTO evaluations VALUES(21113,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(21113,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(21113,33,'cp -62',NULL);
INSERT INTO evaluations VALUES(21113,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(21113,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(21113,36,'cp -73','c6d4 e2e3 d4c6 d1d8 e8d8 f1c4 e7e5 f4g3 g8h6 g1e2 f7f6 e1g1 f8d6 a1c1 h6f5 f1d1 h8e8 c3e4 d8e7 e2c3 c8e6 c3d5 e7f7 e4d6 f5d6 c4e2 a8c8 d5c7 c8c7 d1d6 e6a2 f2f4 e5f4 g3f4 e8c8 d6d1 c6e5 c1a1 a2b3 d1f1 c7c2');
INSERT INTO evaluations VALUES(22668,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(22668,31,'cp 181',NULL);
INSERT INTO evaluations VALUES(22668,32,'cp 183',NULL);
INSERT INTO evaluations VALUES(22668,33,'cp 163',NULL);
INSERT INTO evaluations VALUES(22668,34,'cp 161',NULL);
INSERT INTO evaluations VALUES(22668,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(22668,36,'cp 169','e5c6 d7c6 e2e3 g8f6 d2d3 d8e7 f1e2 c8e6 e3e4 e8c8 b1c3 h7h6 e1g1 g7g5 g1h1 f8g7 f2f4 g5f4 c1f4 f6d7 d1e1 g7d4 c3d1 f7f5 c2c3 d4f6 d1f2 f6g5 e4f5 e6f5 e2g4 e7e1 a1e1 f5g4 f2g4 g5f4 f1f4 h8e8 e1e8');
INSERT INTO evaluations VALUES(22564,30,'cp 167',NULL);
INSERT INTO evaluations VALUES(22564,31,'cp 160',NULL);
INSERT INTO evaluations VALUES(22564,32,'cp 168',NULL);
INSERT INTO evaluations VALUES(22564,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(22564,34,'cp 162',NULL);
INSERT INTO evaluations VALUES(22564,35,'cp 160',NULL);
INSERT INTO evaluations VALUES(22564,36,'cp 164','d7c6 e2e3 c8e6 c2c4 d8h4 d2d4 e8c8 c1d2 g8f6 g2g3 h4h6 h2h4 f6e4 b1c3 e4d2 d1d2 c8b8 e1c1 g7g6 f1d3 f8e7 c1b1 g6g5 h4g5 h6g5 f2f4 g5f6 d1g1 f6g7 d2c2 h7h5 d3e2 g7g6 c2g6 f7g6 e2d3 g6g5 b1a1 g5f4');
INSERT INTO evaluations VALUES(26090,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(26090,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26090,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26090,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26090,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26090,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26090,36,'cp 0','g8f6 d1e2 f8e7 c2c3 e8g8 f3h4 d8c7 h4g6 h7g6 g3e4 f6e4 e2e4 b7b5 c1f4 c7b6 c4d3 b5b4 e4e2 a8d8 f1c1 d7f6 f4e5 f6d7');
INSERT INTO evaluations VALUES(26391,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26391,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26391,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(26391,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(26391,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(26391,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(26391,36,'cp 13','b1d2 d8b6 f3h4 f5e4 d1b3 f8e7 h4f3 g8f6 d4c5 e7c5 d2e4 f6e4 f1d3 e4f6 h2h4 h7h6 h4h5 a8c8 e1e2 f6d7 b3b6 c5b6 f3d4 c6e5 f4e5 d7e5 a2a4 a7a6 a4a5 b6a7 d4f3 e5c4 d3c4 c8c4 h1b1 c4c7 g2g3 e8e7 f3d4');
INSERT INTO evaluations VALUES(26390,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(26390,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(26390,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26390,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26390,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26390,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(26390,36,'cp 6','d8b6 f3h4 f5e4 f2f3 e4g6 h4g6 h7g6 d1b3 f8e7 a2a4 b6b3 d2b3 g6g5 f4g3 c5d4 e3d4 g8h6 g3f2 h6f5 f1d3 e7d6 h2h3 b7b6 a4a5 e8d7 e1e2 f5g3 f2g3 d6g3 a5b6 a7b6 d3b5 d7c7 a1a8 h8a8 h1a1 a8a1 b3a1 g3f4 b5c6 c7c6');
INSERT INTO evaluations VALUES(29397,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(29397,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(29397,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(29397,33,'cp 19',NULL);
INSERT INTO evaluations VALUES(29397,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(29397,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(29397,36,'cp 25','h2h3 g4h5 c1f4 f8d6 f4d6 h5f3 d1f3 d8d6 b1d2 e8g8 f1e1 d6b6 d2b3 b8d7 c2c3 f8e8 f3e3 h7h6 a1b1 b6c7 e1e2 a7a5 a2a4 d7b6 b3d2 b6c4 d2c4 d5c4 d3e4 f6e4 e3e4 e8d8 b1e1 d8d5 e4f3 d5f5 f3g3 c7g3 f2g3');
INSERT INTO evaluations VALUES(31939,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(31939,31,'cp 80',NULL);
INSERT INTO evaluations VALUES(31939,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(31939,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(31939,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(31939,35,'cp 60',NULL);
INSERT INTO evaluations VALUES(31939,36,'cp 74','g2f3 g8f6 e1g1 c7c6 d2d4 b8d7 b1c3 e7e5 e2e4 d5e4 d4e5 d7e5 c3e4 f8e7 d1d8 a8d8 e4f6 e7f6 f3e2 e8g8 a2a4 a7a5 f1e1 e5d7 c1e3 f6b2 a1b1 b2a3 b1b7 d7c5 b7c7 c5a4 e1a1 a4c3 e2f3 a3b4 c7c6 c3b5 g1f1 b4c3 a1a4 b5d4 c6c3 d4f3 f1e2 f3h2');
INSERT INTO evaluations VALUES(31938,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(31938,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(31938,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(31938,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(31938,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(31938,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(31938,36,'cp 58','g8f6 e1g1 c7c6 d2d4 b8d7 b1c3 e7e5 e2e4 d5e4 d4e5 d7e5 c3e4 f8e7 e4f6 e7f6 d1d8 a8d8 f3e2 e8g8 f1e1 a7a6 g1f1 g7g6 a2a4 h7h5 a1a3 f8e8 a3e3 h5h4 b2b3 g8f8 f2f4 e5d7 e3e8 d8e8 c1a3 f8g8 e1d1 d7f8 e2f3 f8e6');
INSERT INTO evaluations VALUES(36985,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(36985,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(36985,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(36985,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(36985,34,'cp -49',NULL);
INSERT INTO evaluations VALUES(36985,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(36985,36,'cp -39','h2h3 c7c6 d2d3 b7b5 c4b3 a7a5 a2a3 h7h6 g5f3 c8e6 b3e6 f7e6 c3e2 d6d5 e2g3 c5d6 e1g1 b8d7 a3a4 b5b4 c1d2 d8b6 d1e2 b6a6 a1e1 a8e8 d2e3 a6b7 b2b3 f8f7 e3c1 g8h7 c1d2 d6b8 d2e3 c6c5 e3c1');
INSERT INTO evaluations VALUES(38044,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(38044,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(38044,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(38044,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(38044,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(38044,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(38044,36,'cp 27','g8f6 f1d3 e8g8 g1f3 c6c5 a2a3 b4c3 d2c3 c5d4 f3d4 e6e5 d4e2 b8c6 e1g1 c8e6 c4d5 f6d5 a1c1 d8g5 e2g3 a8d8 c3d2 h7h6 d1e2 f7f5 f2f3 f5f4 e3f4 d5f4 d2f4 f8f4 f1e1 e6f7 c1c3 a7a6 g3e4 g5e7 e2c2 c6d4');
INSERT INTO evaluations VALUES(40475,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(40475,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(40475,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(40475,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(40475,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(40475,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(40475,36,'cp 35','g1f3 e8g8 f1d3 c6c5 a2a3 b4c3 d2c3 c5d4 f3d4 e6e5 d4e2 b8c6 e1g1 c8e6 c4d5 f6d5 a1c1 d8b6 d1c2 h7h6 c2b1 a8d8 f1d1 d5c3 e2c3 e6b3 d3h7 g8h8 h7c2 d8d1 c1d1 f8d8 h2h3 h8g8 d1d8 c6d8 c3a4 b3c2 b1c2 b6c6 c2d1 d8e6 a4c3 e6g5 d1h5 g8h7 h5e2 e5e4');
INSERT INTO evaluations VALUES(38137,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(38137,31,'cp -56',NULL);
INSERT INTO evaluations VALUES(38137,32,'cp -42',NULL);
INSERT INTO evaluations VALUES(38137,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(38137,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(38137,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(38137,36,'cp -30','b8c6 g1f3 a7a6 f1d3 f7f6 d1e2 d8c7 c1f4 f6f5 g2g3 g8e7 c3b1 e8g8 b1d2 h7h6 h2h4 c6b4 e1g1 c5a7 a2a3 b4d3 c2d3 e7g6 b2b4 c7d8 d3d4 g6f4 g3f4 c8d7 f1c1 d7e8 d2b3 e8h5 c1c3 f8f7 a1c1 h5f3 e2f3 d8h4 g1g2 g7g5 f4g5 h4g5 g2h1');
INSERT INTO evaluations VALUES(5567,30,'cp -33',NULL);
INSERT INTO evaluations VALUES(5567,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(5567,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(5567,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(5567,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(5567,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(5567,36,'cp -24','g1f3 a7a6 f1d3 f7f6 d1e2 d8c7 c1f4 f6f5 g2g3 g8e7 c3b1 h7h6 h2h4 e8g8 b1d2 c6b4 d2b3 b4d3 e2d3 b7b6 b3d4 e7c6 e1g1 c6d4 f3d4 a6a5 d3d2 c7f7 c2c3 c8a6 f1e1 a5a4 b2b3 f8d8 a1c1 d8c8 f4e3 c8c7 f2f4 a4b3 a2b3');
INSERT INTO evaluations VALUES(38999,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(38999,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(38999,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(38999,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(38999,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(38999,35,'cp -22',NULL);
INSERT INTO evaluations VALUES(38999,36,'cp -22','c7c6 h2h4 d7d5 c4b3 g8f6 e4e5 f6d7 d3d4 c6c5 c2c3 b8c6 g1e2 c5d4 c3d4 f7f6 e5f6 d8f6 f3f6 d7f6 c1e3 c8d7 f2f3 c6a5 b1c3 e8g8 e3f4 b7b5 f4d6 f8c8 e1f2 f6e8 d6c5 a5c4 b3c4 b5c4 a1c1 e8f6 c1e1 a7a5 e1c1 a5a4 a2a3');
INSERT INTO evaluations VALUES(40031,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(40031,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(40031,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(40031,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(40031,34,'cp 83',NULL);
INSERT INTO evaluations VALUES(40031,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(40031,36,'cp 65','f1d3 d7d6 e1g1 e8g8 h2h3 e7e5 d4d5 b8a6 d3a6 b7a6 f1e1 f6e8 b1d2 f7f5 b2b3 e8f6 f3g5 f6d5 d2c4 d5c7 c4d6 h7h6 g5f3 c7b5 d1d5 g8h7 e1d1 a8b8 d6b5 d8d5 e4d5 a6b5 c1e3 c8b7 e3a7 b8a8 a7c5 f8d8');
INSERT INTO evaluations VALUES(40214,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(40214,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(40214,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(40214,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(40214,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(40214,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(40214,36,'cp 45','c4d5 f6d5 g5e7 e8e7 a1c1 d5c3 b2c3 c5d4 e3d4 b7b6 d1g4 e7f8 g4f3 d8d5 f3d5 e6d5 e1d2 f8e7 h2h4 b8c6 f1d3 c8d7 c1e1 e7d6 e1e3 c6a5 e3g3 g7g6 h4h5 a5c4 d3c4 d5c4 g1f3 d7e6 f3g5 e6d5 g3e3 a8e8 e3e8 h8e8 g5h7 e8h8 h7f6 d5g2 h1g1');
INSERT INTO evaluations VALUES(40608,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40608,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40608,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40608,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(40608,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40608,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40608,36,'cp 0','c4d5 e6d5 g1f3 e8g8 b1c3 b8c6 b2b4 d5d4 c3b5 c5b6 b5d4 c6d4 f3d4 c8g4 f1e2 g4e2 d1e2 b6d4 e3d4 a7a5 b4a5 f8e8 c1e3 d8a5 e2d2 a5a6 d2e2');
INSERT INTO evaluations VALUES(40862,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40862,31,'cp 5',NULL);
INSERT INTO evaluations VALUES(40862,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(40862,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(40862,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40862,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40862,36,'cp 0','d6d5 d3d4 e8g8 f1e1 h7h6 h2h3 c7c5 b1c3 b8c6 d4c5 e7c5 e2d3 f8e8 c1f4 c8e6 a2a3 a8c8 d1d2 a7a6 a1d1 b7b5 e1e2 c5f8 d1e1 f8c5');
INSERT INTO evaluations VALUES(42002,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(42002,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(42002,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(42002,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42002,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(42002,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(42002,36,'cp 9','b8c6 e2e3 h7h6 g5f6 d8f6 c4d5 e6d5 g1e2 e8g8 e2f4 f6d6 a2a3 b4c3 b2c3 b7b6 f1d3 c6a5 e1g1 c7c6 a4b4 d6d8 f1c1 f8e8 h2h3 c8e6 a3a4 d8g5 b4b1 a8d8 f4e6 e8e6 c3c4 a5c4 d3c4 d5c4 c1c4 e6g6 g2g3 d8c8 c4c3 g5d5');
INSERT INTO evaluations VALUES(8945,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(8945,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(8945,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8945,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(8945,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(8945,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(8945,36,'cp 4','e2e3 h7h6 g5f6 d8f6 g1f3 e8g8 a1c1 f8d8 c4d5 e6d5 f1e2 b4f8 e1g1 c6e7 b2b4 a7a6 b4b5 c8d7 f3e5 d7e6 b5a6 c7c5 a6a7 c5d4 e3d4 e7c6 c3b5 c6e5 d4e5 f6e5 e2f3 e6d7 f1e1 e5f6 f3d5 d7b5 a4b5 a8a7');
INSERT INTO evaluations VALUES(42175,30,'cp 162',NULL);
INSERT INTO evaluations VALUES(42175,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(42175,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(42175,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(42175,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(42175,35,'cp 159',NULL);
INSERT INTO evaluations VALUES(42175,36,'cp 163','c1a3 b8d7 g1f3 b7b6 d1a4 c8b7 f3e5 d8c8 e1g1 a7a6 e5c6 f6e4 c4b3 e4d6 d4d5 e6d5 b3d5 d7f6 d5f3 c8d7 f1d1 e8g8 a3d6 c7d6 a4b4 a6a5 b4b6 a8c8 c6e5 d7c7 b6c7 c8c7 d1d6 f8e8 f3b7 c7b7 e5c4 f6e4 d6c6 b7d7 g1f1 e8d8 a1c1 d7d1 c1d1 d8d1 f1e2 e4c3 e2f3');
INSERT INTO evaluations VALUES(43266,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(43266,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(43266,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(43266,33,'cp 70',NULL);
INSERT INTO evaluations VALUES(43266,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(43266,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(43266,36,'cp 49','c1e3 d7d6 b1c3 c7c6 a2a4 e8g8 g1e2 e7e5 d4e5 d6e5 d1d8 f8d8 e2c1 c8e6 c1d3 b8a6 e1c1 f6d7 f1e2 g7f8 d1d2 a6c5 h1d1 c5d3 d2d3 g8g7 b2b3 f8e7 d3d2 a7a5 h2h3 h7h5 c1b2 g7f8 g2g4 d7f6 c3b1 h5h4 b1c3 d8d2 d1d2');
INSERT INTO evaluations VALUES(45059,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(45059,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(45059,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(45059,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(45059,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(45059,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(45059,36,'cp 13','g1e2 d5c3 e2c3 c8f5 h2h4 h7h5 d2d4 e7e6 c4e2 b8d7 e2h5 f5c2 d1e2 c2f5 c1g5 d8a5 h5f3 d7f6 e1g1 e8c8 f1d1 f5g6 a1c1 f8d6 g2g3 a5f5 g1g2 d8d7 a2a4 g6h5');
INSERT INTO evaluations VALUES(45240,30,'cp 4',NULL);
INSERT INTO evaluations VALUES(45240,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45240,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(45240,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(45240,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(45240,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(45240,36,'cp 0','e4e5 f6d7 e5e6 f7e6 g1f3 b8c6 f3g5 d7e5 c4b3 d6d5 e1g1 g7g6 f2f4 e5f7 f4f5 f7g5 c1g5 f8g7 f5g6 h7g6 d1d3 g7e5 h2h3 d8d6 a1e1 c8d7 c3b5 d6c5 g5e3 c5b4 e3d2 b4c5');
INSERT INTO evaluations VALUES(47398,30,'cp -106',NULL);
INSERT INTO evaluations VALUES(47398,31,'cp -86',NULL);
INSERT INTO evaluations VALUES(47398,32,'cp -85',NULL);
INSERT INTO evaluations VALUES(47398,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(47398,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(47398,35,'cp -97',NULL);
INSERT INTO evaluations VALUES(47398,36,'cp -88','e7e5 g1f3 f8c5 d1e2 b8c6 c2c3 e8g8 c1e3 c5b6 b1d2 c8e6 f3g5 e6d7 g5f3 h7h6 h1g1 d7e6 h2h3 f6h5 g2g3 d8e7 a2a3 a8d8 e1c1 e6a2 e2f2 b6e3 f2e3 b7b5 f1e2 e7d6 e3f2 f7f5 e4f5');
INSERT INTO evaluations VALUES(47443,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(47443,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(47443,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(47443,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47443,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(47443,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(47443,36,'cp 5','c7c5 f1c4 c5d4 g1f3 e7e6 f3d4 f8c5 e1g1 e8g8 d4f3 a7a6 c4d5 d8d5 d1d5 e6d5 b1c3 f8d8 f1d1 c8e6 b2b3 b8c6 c1b2 a8c8 a1c1 f7f6 h2h3 c6e7 c3a4 c5d6 f3d4 g8f7 a4b6 c8c5 c1c5 d6c5 b6a4 c5d4 b2d4');
INSERT INTO evaluations VALUES(47559,30,'cp -425',NULL);
INSERT INTO evaluations VALUES(47559,31,'cp -417',NULL);
INSERT INTO evaluations VALUES(47559,32,'cp -412',NULL);
INSERT INTO evaluations VALUES(47559,33,'cp -411',NULL);
INSERT INTO evaluations VALUES(47559,34,'cp -430',NULL);
INSERT INTO evaluations VALUES(47559,35,'cp -432',NULL);
INSERT INTO evaluations VALUES(47559,36,'cp -413','d8e7 d1e2 e7f7 c4d5 f7d5 b1c3 d5e6 c3e4 f8e7 b2b3 e8g8 c1b2 e6a6 e2a6 b8a6 e1c1 c8d7 d2d3 a8e8 h1e1 a6c5 f2f3 c5e6 c1b1 e6f4 g2g3 f4d5 c2c4 d5f6 b1c2 d7c6 b2f6 g7f6 d3d4 e7b4 d4d5 f6f5 d5c6 b4e1 d1e1 f5e4 c6b7');
INSERT INTO evaluations VALUES(51445,30,'cp 123',NULL);
INSERT INTO evaluations VALUES(51445,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(51445,32,'cp 113',NULL);
INSERT INTO evaluations VALUES(51445,33,'cp 115',NULL);
INSERT INTO evaluations VALUES(51445,34,'cp 110',NULL);
INSERT INTO evaluations VALUES(51445,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(51445,36,'cp 111','f1g2 b8c6 g1f3 a8b8 e1g1 f8e7 b1c3 g8f6 d2d4 e5d4 f3d4 c6d4 d1d4 e8g8 c1f4 c8b7 a1d1 b7g2 g1g2 d7d6 d4d3 f8e8 b2b3 a7a6 f1e1 h7h6 e2e4 f6d7 h2h4 e7f6 c3d5 f6e5 f4e5 d7e5 d3c3 d8c8 h4h5 c7c6 d5e3');
INSERT INTO evaluations VALUES(51446,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(51446,31,'cp 103',NULL);
INSERT INTO evaluations VALUES(51446,32,'cp 100',NULL);
INSERT INTO evaluations VALUES(51446,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(51446,34,'cp 95',NULL);
INSERT INTO evaluations VALUES(51446,35,'cp 97',NULL);
INSERT INTO evaluations VALUES(51446,36,'cp 111','b8c6 g1f3 a8b8 e1g1 f8e7 d2d4 e5d4 f3d4 c6d4 d1d4 g8f6 b1c3 e8g8 c1f4 c8b7 a1d1 f8e8 d4d3 b7g2 g1g2 d7d6 b2b3 h7h6 e2e4 f6d7 h2h4 d7c5 d3c2 c7c6 d1d2 c5e6 f4e3 b8c8 f2f3 b6b5 c4b5 c6b5 e3a7 e7f8');
INSERT INTO evaluations VALUES(368,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(368,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(368,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(368,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(368,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(368,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(368,36,'cp 31','d7d6 f1c4 g8e7 d4f6 g7f6 b1d2 c8e6 e1g1 h7h5 f2f4 e6c4 d2c4 e8c8 c4e3 d6d5 e3d5 e7d5 e4d5 d8d5 a1d1 h8d8 d1d5 d8d5 g2g3 c8d7 g1g2 d7e6 f1f2 e6f5 g2h3 b7b5 f2e2 a7a5 e2e7 d5d2 e7f7 d2b2 f7c7 b2a2 c7c5 f5g6 c5b5');
INSERT INTO evaluations VALUES(770,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(770,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(770,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(770,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(770,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(770,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(770,36,'cp 0','d4c5 b6c7 f1b5 c8d7 e1g1 c6e5 b1a3 e5f3 d1f3 e7g6 c1e3 a8d8 b5d7 c7d7 a1d1 f8e7 c3c4 e8g8 a3c2 g6h4 f3f4 h4g6');
INSERT INTO evaluations VALUES(1515,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(1515,31,'cp -47',NULL);
INSERT INTO evaluations VALUES(1515,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(1515,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(1515,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(1515,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(1515,36,'cp -25','c3d4 g8h6 b1c3 h6f5 c3a4 b6a5 a4c3 f8b4 e1f2 c8d7 c1d2 b4e7 c3e2 a5b6 d1b3 b6b3 a2b3 h7h5 g2g3 a7a6 f1h3 f5h6 h1c1 e7d8 f4f5 h6f5 h3f5 e6f5 b3b4 a8c8 e2f4 d7e6 c1c5 d8e7 b4b5 e7c5 b5c6 c5b6 c6b7 c8b8 a1a6');
INSERT INTO evaluations VALUES(2221,30,'cp -45',NULL);
INSERT INTO evaluations VALUES(2221,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(2221,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(2221,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(2221,34,'cp -46',NULL);
INSERT INTO evaluations VALUES(2221,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(2221,36,'cp -38','b1c3 c6e7 c3a4 c5d6 d3d4 e5d4 d1d4 f6d4 f3d4 a7a6 c4d3 b7b5 a4c3 d6c5 d4b3 c5b6 a2a4 b5b4 c3d5 e7d5 e4d5 a6a5 d3b5 c8b7 a1d1 a8e8 b5d7 e8e5 d7c6 b7c6 d5c6 e5e2 d1d2 f8e8 c2c3 b4c3 b2c3 e8e4 d2e2 e4e2 b3d4');
INSERT INTO evaluations VALUES(2219,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(2219,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(2219,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(2219,33,'cp -41',NULL);
INSERT INTO evaluations VALUES(2219,34,'cp -52',NULL);
INSERT INTO evaluations VALUES(2219,35,'cp -28',NULL);
INSERT INTO evaluations VALUES(2219,36,'cp -33','c6e7 c3a4 c5d6 a4c3 c7c6 d3d4 e7g6 c4b3 e5d4 d1d4 f6e7 a1d1 d6c7 h2h3 d7d6 f1e1 g6e5 d4e3 c8d7 f3e5 d6e5 d1d3 f8d8 e1d1 d7e8 e3f3 d8d3 d1d3 a7a5 a2a4 a8d8 d3d8 c7d8 c3d1 d8b6 f3c3 e7c5 b3c4 g8f8 d1e3 e8d7 c3b3');
INSERT INTO evaluations VALUES(5066,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(5066,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(5066,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(5066,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(5066,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5066,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5066,36,'cp 0','f1e2 d8c7 d1d2 a8b8 b2b3 c7a5 e3d4 g7h6 d4e3');
INSERT INTO evaluations VALUES(6204,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(6204,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(6204,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(6204,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(6204,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(6204,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(6204,36,'cp 33','c5b4 c2c3 b4d6 d2d4 g8e7 f3e5 d6e5 d4e5 c6e5 c4b3 d7d6 f2f4 e5d7 c3c4 d7c5 b1c3 b7b6 f4f5 e7c6 f5f6 g7g6 c3d5 c6e5 c1e3 c8b7 e3c5 b6c5 d5e7 e8f8 b3a4 a8b8 d1c2 b7a6 a1b1 b8b6 a4b3 a6c8 c2d2 c8e6 b1d1 e6c4 b3c4 e5c4 d2c3 c4e5 c3a5 b6b2 a5a7');
INSERT INTO evaluations VALUES(6264,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(6264,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(6264,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(6264,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(6264,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(6264,35,'cp 65',NULL);
INSERT INTO evaluations VALUES(6264,36,'cp 18','c2c3 b4d6 d2d4 g8e7 f3e5 d6e5 d4e5 c6e5 c4b3 d7d6 f2f4 e5d7 c3c4 d7c5 b1c3 b7b6 c1e3 c8b7 f4f5 e7c6 f5f6 g7g6 c3d5 c6e5 e3c5 b7d5 c5d4 d5e6 d1c2 e8g8 c2d2 g8h7 d4e3 e5g4 a1e1 g4e3 e1e3 f8g8 h2h3 c7c6 e3g3 b6b5 c4b5 e6b3 a2b3 c6b5');
INSERT INTO evaluations VALUES(6261,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(6261,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(6261,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(6261,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(6261,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(6261,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(6261,36,'cp 42','b4d6 d2d4 g8e7 f3e5 d6e5 d4e5 c6e5 c4b3 d7d6 f2f4 e5d7 c3c4 d7c5 b1c3 b7b6 c1e3 c8b7 f4f5 f7f6 d1h5 e8f8 e3c5 b6c5 c3d5 e7c6 b3a4 c6e5 d5f4 f8g8 f4g6 b7c6 a4c6 e5c6 a1b1 g8h7 g6h8 h7h8 h5f7 c6e5 f7d5 a8b8 h2h3 h8h7 g1h2 b8b6 b1b6 a7b6 h2g1 d8e7');
INSERT INTO evaluations VALUES(6982,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(6982,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(6982,32,'cp 65',NULL);
INSERT INTO evaluations VALUES(6982,33,'cp 73',NULL);
INSERT INTO evaluations VALUES(6982,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(6982,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(6982,36,'cp 74','d2d4 d7d5 e4d5 c6b4 e1g1 b4d5 c4d5 d8d5 c1f4 g8f6 f4c7 e8g8 f1e1 c8e6 c7e5 b7b5 a2a3 f8e8 e5f6 e7f6 b1c3 d5c6 c3e4 e6d5 e4f6 c6f6 e1f1 f6c6 d1d2 f7f6 a1e1 d5c4 e1e8 a8e8 f1e1 e8e1 d2e1 g8f7 c2c3 c4d5 g1f2 f7g8 e1e2 a7a6 h2h3 h7h6');
INSERT INTO evaluations VALUES(7462,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(7462,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(7462,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(7462,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(7462,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7462,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(7462,36,'cp 0','a4b3 e8g8 d2d3 d7d6 c3d5 f6d5 b3d5 d8e8 c2c3 a8b8 h2h3 c5b6 a2a4 c6e7 d5b3 e7g6 f3g5 h7h6 d1h5 g6e7 g5f3 b5a4 b3a2 g8h8 c1h6 f7f6 h5h4 e7g6 h4h5');
INSERT INTO evaluations VALUES(7681,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(7681,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(7681,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(7681,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(7681,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(7681,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7681,36,'cp 0','f2f4 d8a5 d1e2 d7d6 e1g1 e5f4 c1f4 e8g8 g1h1 c8g4 e2e1 a8e8 e1g3 a5c5 b2b3 g4h5 e4e5 d6e5 f4e5 f6g4 a1e1 e7f6 e5f6 e8e1 f1e1 g4f2 h1g1 f2g4');
INSERT INTO evaluations VALUES(8287,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(8287,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(8287,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(8287,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(8287,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(8287,35,'cp -21',NULL);
INSERT INTO evaluations VALUES(8287,36,'cp -13','e8g8 a2a4 b7b6 c3d5 f6d5 e4d5 c6b4 f3e1 c8b7 c2c3 b4a6 e1c2 a6c7 c2e3 a8b8 a1b1 d8d7 c1d2 e7e6 d5e6 c7e6 e3d5 e6c7 d5c7 d7c7 g2b7 c7b7 d1b3 a7a6 f1e1 b6b5 h2h4 b7c6 a4b5 a6b5 h4h5 c5c4 b3d1 c4d3 h5h6 g7f6 d2e3');
INSERT INTO evaluations VALUES(9284,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(9284,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(9284,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(9284,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(9284,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(9284,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(9284,36,'cp 18','d2d4 e5d4 f3d4 c8d7 b5c6 d7c6 e1g1 e8g8 c1g5 h7h6 g5h4 d8e7 d4c6 b7c6 d1f3 c5d4 a1d1 d4e5 f1e1 f8e8 f3e3 f6g4 h4e7 g4e3 e1e3 e8e7 d1d3 a8e8 g2g3 g7g6 g1g2 e5g7 b2b3 e7e5 g2g1 e5a5 e3f3 g7c3 d3c3 a5a2 c3c6');
INSERT INTO evaluations VALUES(9296,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(9296,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(9296,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(9296,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(9296,34,'cp 36',NULL);
INSERT INTO evaluations VALUES(9296,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(9296,36,'cp 17','c2c3 e8g8 f1e1 h7h6 b5a4 c8d7 d2d4 c5b6 a2a3 f8e8 a4c2 f6h7 c1e3 d8f6 d4e5 c6e5 f3e5 e8e5 b1d2 h7g5 d2c4 b6e3 e1e3 e5e7 d1d4 f6f4 a1e1 a8e8 e4e5 f4d4 c3d4 d6d5 c4d2 g5e6 d2f3 a7a5 h3h4 c7c6 e3c3 e8a8 g1h2 a5a4 h4h5 e6f4');
INSERT INTO evaluations VALUES(10239,30,'cp -52',NULL);
INSERT INTO evaluations VALUES(10239,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(10239,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(10239,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(10239,34,'cp -45',NULL);
INSERT INTO evaluations VALUES(10239,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(10239,36,'cp -28','c1g5 c5e7 g5f6 e7f6 g2g3 e8g8 f1g2 d7d6 b1a3 c6b8 a3c4 b8d7 c4e3 a7a5 f3e2 f8e8 g1f3 c7c6 e2c2 d7c5 a1d1 d8e7 h2h4 c8d7 e1g1 g7g6 f3g5 b7b5 f2f4 d6d5 f4f5 f6g5 h4g5 e7g5 c2f2 g6f5 e4d5 c6d5 e3f5 d7f5');
INSERT INTO evaluations VALUES(10599,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(10599,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(10599,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(10599,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(10599,34,'cp 44',NULL);
INSERT INTO evaluations VALUES(10599,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(10599,36,'cp 42','g4e5 b1c3 e8g8 f1e2 d7d6 b2b3 e5f3 e2f3 c6e5 c1b2 c7c6 e1g1 f8e8 c3e4 c8f5 e4c5 d6c5 b2e5 e8e5 d1e2 e5e8 a1d1 d8e7 d1d2 a8d8 f1d1 h7h6 f3g4 f5g4 h3g4 g8f8 d2d3 d8d3 e2d3 g7g6 d1d2 f8g7 d3d7 g7f8 a2a4 e7d7 d2d7 e8e7 d7d8 e7e8 d8d3 f8e7 g1f1 e8d8 d3d8');
INSERT INTO evaluations VALUES(13863,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(13863,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(13863,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(13863,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(13863,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(13863,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(13863,36,'cp 46','h7h6 a2a3 f8e7 e1g1 c8e6 b2b4 a7a6 b4c5 d6c5 a1b1 a8b8 c1b2 e8g8 b2a1 d8c7 f3d2 f8d8 b1b2 c7a5 g2c6 b7c6 d1c2 b8b2 a1b2 d8b8 f1b1 f6d7 c3d1 a5c7 b2c3 b8b1 c2b1 e7d6 d1e3 f7f5 b1d1 f5f4 e3g2 e6h3');
INSERT INTO evaluations VALUES(13971,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(13971,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(13971,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(13971,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(13971,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(13971,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(13971,36,'cp 19','e1g1 f8e7 c2c4 e8g8 b1c3 c8d7 a2a4 a8b8 b2b3 a7a6 h2h3 e6e5 f3h4 c6d4 c3e2 d4e2 d1e2 g7g6 h4f3 f6h5 g1h2 h5g7 c1h6 f8e8 h6d2 g7e6 h3h4 b7b5 a4b5 a6b5 a1a7 b8a8 f1a1 a8a7 a1a7 b5b4 d2e3 d7c6 g2h3');
INSERT INTO evaluations VALUES(13970,30,'cp 18',NULL);
INSERT INTO evaluations VALUES(13970,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(13970,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(13970,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(13970,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(13970,35,'cp 2',NULL);
INSERT INTO evaluations VALUES(13970,36,'cp 10','f8e7 c2c3 e8g8 f1e1 c8d7 d3d4 d6d5 e4d5 e6d5 c1e3 f6g4 e3f4 f8e8 b1d2 d8b6 h2h3 g4f6 d1b3 c5d4 b3b6 a7b6 f3d4 c6d4 c3d4 b6b5 a2a3 d7c6 d2f3 f6e4 f3e5 g7g5 f4e3 e4d6 b2b3 d6f5 a3a4 e7b4 e1e2 b5a4 e5c6 b7c6 b3a4 f5e3 e2e3');
INSERT INTO evaluations VALUES(8183,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(8183,31,'cp 17',NULL);
INSERT INTO evaluations VALUES(8183,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(8183,33,'cp 8',NULL);
INSERT INTO evaluations VALUES(8183,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(8183,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(8183,36,'cp 9','c2c3 e8g8 f1e1 c8d7 d3d4 d6d5 e4d5 e6d5 c1e3 f6g4 e3f4 d8b6 d1b3 f8e8 h2h3 g4f6 b1d2 c5d4 b3b6 a7b6 f3d4 c6d4 c3d4 b6b5 a2a3 h7h6 d2f1 a8a4 f4e5 b5b4 f1e3 d7e6 a3b4 a4b4 e5f6 e7f6 e3d5 b4d4 d5c7 e8e7 a1a8 g8h7 c7e6 e7e6 e1b1 d4d2 g2b7 e6e2');
INSERT INTO evaluations VALUES(14913,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(14913,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(14913,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(14913,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14913,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(14913,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(14913,36,'cp 27','e2e3 c6a5 d1a4 c7c6 f1c4 d8b6 e1g1 a5c4 a4c4 f6d5 c4e2 d5f4 e3f4 f8d6 g2g3 e8g8 f1d1 f8d8 h2h4 c8d7 d1d2 h7h6 a1d1 d7e8 a2a3 b6c7 f3e5 d6e7 h4h5 b7b5 c3e4 d8d5 e4c5 a8d8 g3g4 e7c5 d4c5');
INSERT INTO evaluations VALUES(14912,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(14912,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(14912,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(14912,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(14912,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(14912,35,'cp 16',NULL);
INSERT INTO evaluations VALUES(14912,36,'cp 30','c6a5 f1c4 a5c4 d1a4 c7c6 a4c4 f6d5 f4g3 d8b6 e1g1 b6b2 c3e4 d5b6 c4d3 b2b5 d3c2 b5c4 c2b1 c4b4 b1c1 b4c4 a2a4 c4c1 f1c1 f7f6 a4a5 b6d5 f3d2 f8b4 d2c4 e8g8 g3d6 f8d8 d6b4 d5b4 a1b1 b4d5 b1b3 a8b8 c1b1 e6e5 d4e5 c8f5 f2f3 b7b5 a5b6 a7b6 e5f6 g7f6 c4b6 f5e4 f3e4 d5b6');
INSERT INTO evaluations VALUES(15290,30,'cp 175',NULL);
INSERT INTO evaluations VALUES(15290,31,'cp 204',NULL);
INSERT INTO evaluations VALUES(15290,32,'cp 189',NULL);
INSERT INTO evaluations VALUES(15290,33,'cp 207',NULL);
INSERT INTO evaluations VALUES(15290,34,'cp 202',NULL);
INSERT INTO evaluations VALUES(15290,35,'cp 198',NULL);
INSERT INTO evaluations VALUES(15290,36,'cp 190','f3g5 c6e5 f1e1 d5e3 g5f7 e5f7 f2e3 f8e7 e3d4 e8g8 e1f1 b7b5 c4b5 c7c5 c1f4 c8b7 d4d5 f7d6 b5c6 b7a6 c2c4 a6c4 f1f3 a8b8 b1d2 c4a6 f4e5 e7f6 d1e1 f6e5 e1e5 f8f3 d2f3 d8f6 b2b3 f6e5 f3e5 c5c4 c6a4 b8d8 a1c1 d6f5 e5c6 d8d5 c6a7 f5e3');
INSERT INTO evaluations VALUES(15901,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(15901,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(15901,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(15901,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(15901,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(15901,35,'cp 54',NULL);
INSERT INTO evaluations VALUES(15901,36,'cp 30','g2g4 h7h6 h2h3 f8e7 g1e2 a7a6 e2g3 d7f8 c1e3 f6h7 d1d2 e7g5 e1c1 g5e3 f2e3 d8h4 d2g2 g7g6 a2a4 f8d7 b2b3 d7c5 c1b2 e8g8 d1a1 b7b6 f1e2 c8d7 a1a3 g8g7 h1f1 a6a5 e2b5 h4e7 a3a1 d7b5 c3b5');
INSERT INTO evaluations VALUES(16264,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(16264,31,'cp 78',NULL);
INSERT INTO evaluations VALUES(16264,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(16264,33,'cp 65',NULL);
INSERT INTO evaluations VALUES(16264,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(16264,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(16264,36,'cp 85','d2d4 d7d5 e4e5 f6e4 g1f3 f8b4 d1c2 f7f5 e5f6 d8f6 c1e3 f6g6 a2a3 b4c3 b2c3 e8g8 h2h4 g6e8 f1d3 h7h6 e1g1 e4d6 c4c5 d6f5 c2d2 e6e5 d4e5 f5e3 d2e3 c8g4 f3h2 c6e5 h2g4 e5g4 e3d4 g4f6 c3c4 e8f7 c4d5 f6d5 a1e1 a8e8 d3c4 c7c6 e1e8 f8e8');
INSERT INTO evaluations VALUES(18813,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(18813,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(18813,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(18813,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(18813,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18813,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18813,36,'cp 0','d7d5 e4d5 e6d5 d2d4 c8g4 f1e2 g8f6 e1g1 f8e7 c1e3 c5d4 e3d4 c6d4 f3d4 g4h3 f1e1 e8g8 e2f3 a8c8 a2a3 e7c5 d4b3 c5b6 e1e2 g7g6 c3a4 b6c7 a4c3');
INSERT INTO evaluations VALUES(21778,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(21778,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(21778,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(21778,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(21778,34,'cp 79',NULL);
INSERT INTO evaluations VALUES(21778,35,'cp 78',NULL);
INSERT INTO evaluations VALUES(21778,36,'cp 77','g8e7 d2b3 e7g6 b3d4 g6e5 f3e5 c6e5 e2e3 f8c5 c1d2 c5d4 e3d4 d8d4 d2c3 d4e4 d1e2 e5d3 e1d2 e4e2 d2e2 d3f4 e2f3 f4e6 f1d3 f7f6 f3g3 e8f7 h2h4 c8d7 a1d1 a8d8 h4h5 d7a4 d1d2 d8d7 h1e1 e6c5 d3b1 d7d2 c3d2 h8d8 d2c3 g7g6 f2f3 g6h5');
INSERT INTO evaluations VALUES(21792,30,'cp 143',NULL);
INSERT INTO evaluations VALUES(21792,31,'cp 149',NULL);
INSERT INTO evaluations VALUES(21792,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(21792,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(21792,34,'cp 159',NULL);
INSERT INTO evaluations VALUES(21792,35,'cp 143',NULL);
INSERT INTO evaluations VALUES(21792,36,'cp 153','d4d5 e5e4 d5c6 d8d1 c3d1 e4f3 c6b7 f8b4 c1d2 b4d2 e1d2 c8b7 e2f3 e8c8 d2c2 b7a6 d1e3 g8e7 f1c4 a6c4 e3c4 e7f5 h1e1 f5d4 c2c3 d4b5 c3b4 b5d4 a1c1 f7f6 b4c3 d8d7 b2b3 h8d8 e1e4 d4c6 a2a3 h7h5 c3b2 c6e7 h2h4 e7f5 c4e3 f5e3 f2e3 d7d2 c1c2');
INSERT INTO evaluations VALUES(24709,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(24709,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(24709,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24709,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(24709,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(24709,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(24709,36,'cp 18','f1e2 f8b4 e1g1 a8c8 f3h4 f5g6 h4g6 h7g6 d1b3 b4d6 f4d6 d8d6 g2g3 d6e7 f1c1 e8g8 c1c2 c6b4 c2d2 a7a6 a2a3 b4c6 a1c1 c8c7 d2c2 f8c8 c3a4 c6a5 c2c7 c8c7 c1c7 e7c7 b3c3 c7c3 a4c3 a5b3 g1g2 b3d2 h2h4 f6e4 c3e4 d2e4');
INSERT INTO evaluations VALUES(24783,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(24783,31,'cp -18',NULL);
INSERT INTO evaluations VALUES(24783,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(24783,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(24783,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(24783,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(24783,36,'cp 0','a8c8 c2c4 c6b4 a1c1 a7a6 b5c3 d5c4 f1c4 c7c5 d1b3 e7e6 d4c5 f6d7 b3b4 d7c5 e1f1 f5d3 g1e2 c5d7 b4b7 c8c4 c1d1 c4b4 b7c6 b4b6 c6a4 b6b4');
INSERT INTO evaluations VALUES(55,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(55,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(55,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(55,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(55,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(55,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55,36,'cp 0','c2c4 c6b4 a1c1 a7a6 b5c3 d5c4 f1c4 c7c5 d1b3 e7e6 d4c5 f8c5 a2a3 b4d5 c4d5 f6d5 c3d5 d8d5 b3d5 e6d5 g1e2 f7f6 b2b4 c5b6 c1c8 f5c8 e1d2 c8d7 f4d6 e8f7 h1c1 h7h5 h2h4 h8e8 g2g3 g7g5 d6c5 b6d8 c5d6');
INSERT INTO evaluations VALUES(24945,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24945,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24945,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24945,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(24945,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(24945,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(24945,36,'cp -38','e7e5 d3e2 e5e4 f3d2 g4e2 d1e2 f8d6 c2c4 e8g8 a2a3 d5c4 d2c4 f8e8 c1d2 c6e7 c4d6 c7d6 b1c3 a7a6 a3a4 e7d5 c3d5 f6d5 e2h5 d5f6 h5d1 d6d5 d1b3 a8b8 f1c1 h7h6 b3c2 d8d7 h2h3 d7f5 c2c7 e8c8');
INSERT INTO evaluations VALUES(24958,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(24958,31,'cp 14',NULL);
INSERT INTO evaluations VALUES(24958,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(24958,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(24958,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(24958,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(24958,36,'cp 18','d8d7 e1g1 a8d8 b2b4 a7a6 f3e2 e7e6 b4b5 a6b5 e2b5 f8d6 c3e2 e8g8 a2a4 d7c8 c2c4 d5c4 b5c4 c6e5 c4a2 c7c5 c1b2 c5d4 e2d4 e5c4 a2c4 c8c4 h2h3 d8d7 d1b1 f8d8 d4f3 c4e4 b2d4 e4b1 f1b1 f6e4');
INSERT INTO evaluations VALUES(25915,30,'cp -146',NULL);
INSERT INTO evaluations VALUES(25915,31,'cp -156',NULL);
INSERT INTO evaluations VALUES(25915,32,'cp -146',NULL);
INSERT INTO evaluations VALUES(25915,33,'cp -155',NULL);
INSERT INTO evaluations VALUES(25915,34,'cp -132',NULL);
INSERT INTO evaluations VALUES(25915,35,'cp -160',NULL);
INSERT INTO evaluations VALUES(25915,36,'cp -162','c6e5 f3e5 d7b5 b1c3 f8b4 d1g4 e8f8 a2a3 b4c3 b2c3 f7f6 a3a4 b5a6 c1a3 g8e7 g4f3 a8c8 a3b4 d8c7 h2h4 f8e8 e5d3 a6d3 f3d3 c7c4 e1d2 c4d3 d2d3 e7c6 h1b1 c8c7 a4a5 h7h5 b4d6 c7d7 d6c5 e8f7 b1b5 a7a6 b5b6 e6e5 g2g3 h8e8 a1a4 f7g6 d3d2 e8c8 d2e2 g6f5 a4a1 c8e8 d4e5');
INSERT INTO evaluations VALUES(26386,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(26386,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(26386,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(26386,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(26386,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(26386,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(26386,36,'cp 13','c1e3 c5c4 b5c6 b7c6 e1g1 g8e7 c3e2 e7c8 e2g3 f5g6 h2h4 h7h6 h4h5 g6h7 f3d2 d8h4 b2b3 f8b4 b3c4 b4d2 d1d2 d5c4 d2b4 c8e7 b4c4 e8g8 c4e2 e7f5 g3f5 h7f5 f1b1 f8b8 b1b3 a7a5 a1b1 b8b3 b1b3 f5g4 f2f3 g4h5 c2c4 h4d8 e2b2 a5a4 b3b6 d8d7 b6b4 h5g6 c4c5');
INSERT INTO evaluations VALUES(27300,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(27300,31,'cp 58',NULL);
INSERT INTO evaluations VALUES(27300,32,'cp 64',NULL);
INSERT INTO evaluations VALUES(27300,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(27300,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(27300,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(27300,36,'cp 68','g8f6 c2c4 g7g6 b1c3 f8g7 c1e3 e8g8 d1c2 d8e7 f1e1 c7c6 a1d1 f8e8 h2h3 d7e5 f3d2 e5d7 b2b4 f6h5 a2a4 e7e5 c3e2 a7a5 b4b5 e5b2 c2b2 g7b2 d1b1 b2a3 e2d4 a8c8 d4c2 a3c5 d2b3 c6b5 a4b5 c5b4 e1d1 c8c4 c2b4 a5b4 d1d6 d7e5 f2f3');
INSERT INTO evaluations VALUES(25584,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(25584,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(25584,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(25584,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(25584,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(25584,35,'cp 48',NULL);
INSERT INTO evaluations VALUES(25584,36,'cp 58','c2c4 g7g6 b1c3 f8g7 h2h3 e8g8 c1e3 d8e7 d1c2 a7a6 a1d1 f8e8 f1e1 c7c6 a2a4 a6a5 c2b3 b7b6 b3c2 d7c5 e3g5 h7h6 g5f4 a8d8 f3d4 e7c7 f2f3 f6d7 d4e2 c5a6 c2d2 d7e5 b2b3 g8h7 f4e3 a6b4 f3f4');
INSERT INTO evaluations VALUES(27586,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(27586,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(27586,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(27586,33,'cp 15',NULL);
INSERT INTO evaluations VALUES(27586,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27586,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(27586,36,'cp 12','c5d4 c3d4 e7e6 b1c3 d5a5 h2h3 g4f3 e2f3 f8b4 e1g1 g8e7 a2a3 b4c3 b2c3 a5c3 a1b1 c3c4 d4d5 e7d5 f3d5 c4d5 d1d5 e6d5 f1e1 e8d8 b1b7 h8e8 e1d1 e8e7 d1d5 d8e8 b7e7 c6e7 d5h5 h7h6 h5c5 f7f6 c1e3 a7a6 c5c7 e7f5 e3c5 a8d8 g1h1');
INSERT INTO evaluations VALUES(27592,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(27592,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(27592,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(27592,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(27592,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(27592,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(27592,36,'cp 24','c3d4 e7e6 b1c3 d5a5 e1g1 f8b4 a2a3 b4c3 b2c3 g8f6 d1b3 e8g8 c1e3 f6d5 f1c1 b7b6 c3c4 d5e3 f2e3 a8c8 c1c3 g4f3 e2f3 c6e7 a1c1 a5g5 c4c5 b6c5 d4c5 c8c7 c5c6 f8c8 a3a4 e7d5 f3d5 g5d5 b3d5 e6d5 h2h3');
INSERT INTO evaluations VALUES(27679,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(27679,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(27679,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(27679,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(27679,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(27679,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(27679,36,'cp 52','c2c3 e7e6 b5a4 a7a6 b1d2 b7b5 a4c2 f5c2 d1c2 f7f6 f2f4 g8h6 g1f3 f8e7 e1g1 e8g8 d2b3 a6a5 c1d2 f6e5 f4e5 h6f5 g2g4 f5h4 f3h4 e7h4 b3c5 d7e7 d2e3 h4g5 e3g5 e7g5 c2e2 g5g6 a1e1 b5b4 h2h4 b4c3 b2c3 a5a4 h4h5');
INSERT INTO evaluations VALUES(31451,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31451,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31451,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31451,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(31451,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31451,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(31451,36,'cp 9','e2e3 g8f6 f1d3 f5g6 e1g1 f8e7 f1e1 f6h5 d3g6 h7g6 f4b8 a8b8 e3e4 h5f6 a2a3 b7b5 e4e5 f6d7 g2g3 a7a5 h2h4 c6c5 c3e2 b5b4 a3b4 b8b4 c2c3 b4b2 d1a4 b2b7 e2f4 e8g8 a4a5 d8a5 a1a5 c5d4 c3d4 d7b6 a5a6 f8b8 g1g2 b6c4');
INSERT INTO evaluations VALUES(31450,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31450,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31450,32,'cp 9',NULL);
INSERT INTO evaluations VALUES(31450,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31450,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(31450,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31450,36,'cp 0','g8f6 f1d3 f8e7 e1g1 b8d7 h2h3 f5g6 f1e1 d8b6 e3e4 b6b2 e4d5 c6d5 c3b5 e8g8 a1b1 b2a2 b1a1 a2b2 a1b1');
INSERT INTO evaluations VALUES(31978,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(31978,31,'cp -73',NULL);
INSERT INTO evaluations VALUES(31978,32,'cp -67',NULL);
INSERT INTO evaluations VALUES(31978,33,'cp -78',NULL);
INSERT INTO evaluations VALUES(31978,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(31978,35,'cp -89',NULL);
INSERT INTO evaluations VALUES(31978,36,'cp -84','b8c6 g1f3 c6b4 e4d6 c7d6 d3f5 d8a5 a2a3 a5f5 a3b4 f5e4 e1f1 e4b4 d2d3 b4g4 c2c4 a7a6 b2b4 g8f6 c1g5 f6d7 h2h3 g4f5 g5e3 e7e6 f3d4 f5f6 d1g4 f8e7 g4e4 a8b8 g2g3 f6g6 e4g6 h7g6 f1e2 e7f6 a1a5 f6d8 a5a2 e8e7 h3h4 d8b6 d4f3 b6e3 e2e3 d6d5 c4d5');
INSERT INTO evaluations VALUES(32214,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(32214,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(32214,32,'cp 17',NULL);
INSERT INTO evaluations VALUES(32214,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(32214,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(32214,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(32214,36,'cp 12','e6c4 e2c4 c7c5 b1c3 b8c6 e1g1 f8e7 d2d3 a8c8 a2a4 h7h6 c3d5 g8f6 d5e7 c6e7 f3h4 e8g8 f2f4 e5f4 c1f4 d6d5 e4d5 f6d5 a1e1 d8d7 f4g3 b7b6 e1e2 f8e8 f1e1 d5f6 h2h3 c8d8 h4f3 e7c6 e2e8 f6e8 c2c3 e8d6');
INSERT INTO evaluations VALUES(32797,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(32797,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(32797,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(32797,33,'cp 55',NULL);
INSERT INTO evaluations VALUES(32797,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(32797,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(32797,36,'cp 65','d2d4 b8d7 h2h3 g4h5 f3d2 h5e2 d1e2 g8f6 d2c4 a5c7 e2f3 d7b6 c1f4 c7d8 c4e5 e7e6 e1g1 f8e7 c3e2 d8d5 f3d3 d5e4 d3d2 f6d7 e2g3 e4d5 b2b3 d7e5 f4e5 f7f6 c2c4 d5d7 e5f4 e6e5 d4e5 f6e5 f4e3 d7d2 e3d2 b6d7 a1d1 e8g8 d2e3 d7c5');
INSERT INTO evaluations VALUES(34398,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(34398,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(34398,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(34398,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(34398,34,'cp 112',NULL);
INSERT INTO evaluations VALUES(34398,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(34398,36,'cp 149','a2a4 c8e6 d1d3 e5d4 f3d4 e6c4 d3c4 g8f6 c1e3 g7g6 e1c1 f8g7 d4f3 b8d7 c4b4 f6g4 e3f4 d7e5 f4g3 e8c8 a4a5 h8e8 h1e1 h6h5 c1b1 g7f6 a5a6 h5h4 b4b7 e7b7 a6b7 c8b7 f3h4');
INSERT INTO evaluations VALUES(34419,30,'cp 33',NULL);
INSERT INTO evaluations VALUES(34419,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(34419,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(34419,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(34419,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(34419,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(34419,36,'cp 39','c2c4 g8f6 b1c3 c8d7 d1d2 d7c6 f3e5 f8e7 d4c5 b6c5 c4d5 f6d5 e5c6 b8c6 c3d5 e6d5 d2c3 c5c3 b2c3 e7f6 e1c1 e8g8 d1d5 f6c3 c1b1 c6b4 d5d6 a8d8 f1c4 d8d6 f4d6 f8d8 h1c1 d8d6 c1c3 b4c6 b1c2 c6b4 c2b3 b4c6 c4b5 c6e7 c3c7 e7d5 c7d7 d6d7 b5d7');
INSERT INTO evaluations VALUES(37826,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(37826,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(37826,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(37826,33,'cp -14',NULL);
INSERT INTO evaluations VALUES(37826,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(37826,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(37826,36,'cp 0','f1g2 b8c6 e1g1 d7d6 d2d3 g8f6 a2a4 e8g8 b1a3 a7a6 a3c4 c8e6 f3g5 e6c4 d3c4 d8c8 d1e2 c6d4 e2d1');
INSERT INTO evaluations VALUES(38957,30,'cp -900',NULL);
INSERT INTO evaluations VALUES(38957,31,'cp -923',NULL);
INSERT INTO evaluations VALUES(38957,32,'cp -921',NULL);
INSERT INTO evaluations VALUES(38957,33,'cp -921',NULL);
INSERT INTO evaluations VALUES(38957,34,'cp -922',NULL);
INSERT INTO evaluations VALUES(38957,35,'cp -939',NULL);
INSERT INTO evaluations VALUES(38957,36,'cp -940','e3f2 e1e2 f2g1n e2e1 d8h4 e1d2 b8c6 d2c3 h4f2 b1a3 a7a5 b4d6 c7d6 d1d2 f2c5 a3b5 c5e5 b5d4 g8f6 a1e1 f6e4 e1e4 e5e4 d4c6 b7c6 h1g1 e8g8 f1d3 e4e5 c3c2 c8f5 d3f5 e5f5 c2c1 f8e8 g2g4 f5c5 g1d1 c5c4 c1b1 d6d5 h2h3 e8e2 d2d3 a8b8 d3c4 d5c4 b1c1 g8f8 d1d6 b8b2 d6c6 b2a2 c1b1 e2b2 b1c1 b2c2 c1b1');
INSERT INTO evaluations VALUES(38961,30,'cp -942',NULL);
INSERT INTO evaluations VALUES(38961,31,'cp -925',NULL);
INSERT INTO evaluations VALUES(38961,32,'cp -911',NULL);
INSERT INTO evaluations VALUES(38961,33,'cp -911',NULL);
INSERT INTO evaluations VALUES(38961,34,'cp -919',NULL);
INSERT INTO evaluations VALUES(38961,35,'cp -928',NULL);
INSERT INTO evaluations VALUES(38961,36,'cp -926','e1f2 d8d1 b4c3 g8h6 h2h3 b8d7 g1f3 d7c5 f2g1 c5e4 b1d2 d1a1 d2e4 h6f5 g1f2 a1a2 e5e6 f7f6 g2g4 f5e7 g4g5 f6f5 e4c5 b7b6 c5d7 h8g8 f3e5 c8b7 f1g2 e8c8 g2b7 c8b7 h3h4 d8c8 h1d1 f5f4 d1d3 h7h6 g5h6 g7h6 d3d4');
INSERT INTO evaluations VALUES(39382,30,'cp -51',NULL);
INSERT INTO evaluations VALUES(39382,31,'cp -41',NULL);
INSERT INTO evaluations VALUES(39382,32,'cp -50',NULL);
INSERT INTO evaluations VALUES(39382,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(39382,34,'cp -30',NULL);
INSERT INTO evaluations VALUES(39382,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(39382,36,'cp -31','g8f6 e2e4 e8g8 d2d3 c7c6 f2f4 d7d5 f4e5 f6g4 e4d5 g4e5 g1f3 f8e8 f3e5 e8e5 f1e2 c8g4 a1a2 c6d5 e1g1 e5e2 a2e2 d5c4 e2e8 d8e8 d1g4 e8e6 g4e6 f7e6 d3c4 b8d7 f1d1 d7e5 c4c5 a8c8 c1f4 c8c5 f4e5 c5e5 d1d7 e5b5 a3a4 b5b1 g1f2 a7a5 d7c7');
INSERT INTO evaluations VALUES(39643,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(39643,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(39643,32,'cp -18',NULL);
INSERT INTO evaluations VALUES(39643,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(39643,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(39643,35,'cp -32',NULL);
INSERT INTO evaluations VALUES(39643,36,'cp -24','d7d5 d2d4 g8f6 g1f3 c7c5 b1c3 c8f5 e1g1 e8g8 h2h3 f6e4 f3h4 e4c3 b2c3 f5e6 a1b1 d8d7 g3g4 g7f6 h4f3 b7b6 c3c4 f8d8 d4c5 b6c5 c1a3 d5c4 a3c5 c4c3 c5d4 b8c6 d4f6 e7f6 d1d7');
INSERT INTO evaluations VALUES(41101,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(41101,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(41101,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(41101,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(41101,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(41101,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(41101,36,'cp 24','f1d3 c7c5 d4c5 e8g8 e1g1 e7c5 b1c3 a7a6 a2a3 f8e8 c1f4 b8c6 f1e1 e8e1 d1e1 h7h6 a1d1 c8d7 e1d2 a8c8 f4g3 d7e6 g3h4 c5e7 d1e1 d8b6 c3a4 b6c7 h4g3 c7d8 a4c3 b7b5 f3e5 c6e5 g3e5 f6d7 e5d4');
INSERT INTO evaluations VALUES(41468,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(41468,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(41468,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(41468,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(41468,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(41468,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41468,36,'cp 0','c7c5 b1d2 d8b6 d1b3 b8c6 b3b6 a7b6 a2a3 c8f5 d3f5 g6f5 g1e2 h7h5 f2f3 h5h4 e1f2 e7e6 g2g3 h4g3 h2g3 e8d7 f4g5 f6h7 g5f4');
INSERT INTO evaluations VALUES(42320,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(42320,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(42320,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(42320,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(42320,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(42320,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(42320,36,'cp 18','d7d5 f1e2 e8g8 e1g1 d8e7 a1b1 a7a5 b2b3 b7b6 c3a4 c8b7 c4c5 b8d7 d1c2 b7a6 e2a6 a8a6 c2e2 a6a8 c1b2 f6e4 f1c1 f8b8 c5b6 c7b6 a4c3 e4c3 c1c3 b8c8 b1c1 c8c3 c1c3 a5a4 e2b5 a8a5 b5c6 a4b3 a2b3 g7f8 h2h3');
INSERT INTO evaluations VALUES(42770,30,'cp 156',NULL);
INSERT INTO evaluations VALUES(42770,31,'cp 159',NULL);
INSERT INTO evaluations VALUES(42770,32,'cp 154',NULL);
INSERT INTO evaluations VALUES(42770,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(42770,34,'cp 152',NULL);
INSERT INTO evaluations VALUES(42770,35,'cp 163',NULL);
INSERT INTO evaluations VALUES(42770,36,'cp 150','f3e5 e8g8 e2e3 f8e8 e5f3 c7c5 f1e2 d7d5 c4d5 f6d5 e1g1 b8c6 d1c2 b4a5 a2a3 c8g4 d2d3 a5c7 c1d2 d5c3 d2c3 d8d6 g2g3 d6d5 f3e1 g4h3 e2f3 d5d6 f3g2 h3g2 g1g2 b7b6 e1f3 a8d8 a1d1');
INSERT INTO evaluations VALUES(42712,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(42712,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(42712,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(42712,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(42712,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(42712,35,'cp 138',NULL);
INSERT INTO evaluations VALUES(42712,36,'cp 131','e8g8 e5f3 c7c5 a2a3 b4a5 g2g3 d7d5 c4d5 f6d5 f1g2 b8c6 e1g1 f8e8 c3d5 d8d5 d2d3 d5h5 c1e3 c8g4 a1c1 a5b6 c1c4 h7h6 d1a4 g4d7 a4c2 a8c8 f1c1 d7e6 c4c3 c6e7 e3c5 e7d5 c5b6 d5c3 b2c3 a7b6 f3d4 e6d5 c2b2 d5g2');
INSERT INTO evaluations VALUES(43068,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(43068,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(43068,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(43068,33,'cp 39',NULL);
INSERT INTO evaluations VALUES(43068,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(43068,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(43068,36,'cp 31','d7d5 e4d5 e6d5 c4b3 e8g8 f1e1 f8e8 c2c3 b8d7 h2h3 d7f8 c1f4 f8e6 f4h2 a7a5 b3c2 c7c5 d4c5 e6c5 a2a4 d8b6 b1a3 c8d7 h2e5 a8c8 b2b4 c5e6 e5g7 e6g7 f3e5 e8d8 e5d7 d8d7 b4a5 b6a5 a3b5 e7c6');
INSERT INTO evaluations VALUES(43784,30,'cp -194',NULL);
INSERT INTO evaluations VALUES(43784,31,'cp -188',NULL);
INSERT INTO evaluations VALUES(43784,32,'cp -189',NULL);
INSERT INTO evaluations VALUES(43784,33,'cp -198',NULL);
INSERT INTO evaluations VALUES(43784,34,'cp -201',NULL);
INSERT INTO evaluations VALUES(43784,35,'cp -183',NULL);
INSERT INTO evaluations VALUES(43784,36,'cp -225','a7a5 c5b6 f8b4 c1d2 c7b6 e2e3 e8g8 g1f3 b4c3 d2c3 f6e4 d1b3 c8a6 f1a6 b8a6 a1c1 e4d6 a2a4 f7f6 f3d2 d8d7 e1g1 f8b8 c1b1 a8a7 f1c1 a7c7 h2h3 h7h6 b1a1 c7c6 c3b2 b8c8 c1c6 c8c6 b2a3 d6c4 d2c4 d5c4 b3b2 c4c3 b2b3');
INSERT INTO evaluations VALUES(43698,30,'cp -221',NULL);
INSERT INTO evaluations VALUES(43698,31,'cp -225',NULL);
INSERT INTO evaluations VALUES(43698,32,'cp -232',NULL);
INSERT INTO evaluations VALUES(43698,33,'cp -218',NULL);
INSERT INTO evaluations VALUES(43698,34,'cp -219',NULL);
INSERT INTO evaluations VALUES(43698,35,'cp -207',NULL);
INSERT INTO evaluations VALUES(43698,36,'cp -217','b4b5 b6c5 d4c5 f8c5 e2e3 c7c6 g1f3 c5b4 c1d2 e8g8 f1e2 c6b5 e2b5 c8a6 b5a6 b8a6 e1g1 a8c8 d1e1 d8d7 h2h3 h7h6 a1b1 a6c5 e1e2 f6e4 c3e4 d5e4 f3d4 c5d3 a2a3 b4a3 d2a5 a3d6 f1d1 d6e5 a5b6 d7d5 d4b5 d5c4 e2g4');
INSERT INTO evaluations VALUES(45843,30,'cp 84',NULL);
INSERT INTO evaluations VALUES(45843,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(45843,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(45843,33,'cp 75',NULL);
INSERT INTO evaluations VALUES(45843,34,'cp 64',NULL);
INSERT INTO evaluations VALUES(45843,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(45843,36,'cp 68','c8d7 g1f3 f8e7 h2h3 e8g8 f1e2 c6d5 c4d5 b8a6 c1e3 a8c8 e1g1 a6c5 f3d2 f6e8 b2b4 c5a6 d1b3 e7g5 c3d1 g5e3 d1e3 d8g5 g1h1 g5f4 a2a3 c8c7 h1g1 f7f5 e4f5 d7f5 d2c4 f5g6 e2g4 b7b5 c4a5 e8f6 a5c6');
INSERT INTO evaluations VALUES(46799,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(46799,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(46799,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(46799,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(46799,34,'cp 33',NULL);
INSERT INTO evaluations VALUES(46799,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(46799,36,'cp 33','c2c4 c7c5 d4c5 d8a5 b1c3 d5c4 e2e4 f8g7 f3d2 a5c5 d2c4 c8e6 c1e3 c5c8 a1c1 e8g8 c4d2 b8c6 f1b5 f8d8 e1g1 f6e8 d1e1 e8d6 b5c6 c8c6 c3d5 c6a4 d5e7 g8h8 e3g5 f7f6 g5f4 d8d7 b2b3 a4b4 a2a3 b4a3 e7d5 d6b5 e1e3 a3f8 d2f3 e6d5 e4d5');
INSERT INTO evaluations VALUES(47412,30,'cp -199',NULL);
INSERT INTO evaluations VALUES(47412,31,'cp -198',NULL);
INSERT INTO evaluations VALUES(47412,32,'cp -201',NULL);
INSERT INTO evaluations VALUES(47412,33,'cp -195',NULL);
INSERT INTO evaluations VALUES(47412,34,'cp -156',NULL);
INSERT INTO evaluations VALUES(47412,35,'cp -163',NULL);
INSERT INTO evaluations VALUES(47412,36,'cp -161','a1d1 b8d7 f3g3 e7e6 c3b5 f8b4 c2c3 b4a5 g1f3 a7a6 b5a3 b7b5 f3e5 f6e4 g3g7 d8f6 g7f6 e4f6 e5d7 f6d7 f1e2 c8b7 e1g1 b7e4 f1f2 a5b6 e3b6 c7b6 d1d4 e4g6 e2f3 a8a7 a3c2 g6c2 f2c2 e8e7 c2e2 a7c7 d4h4 a6a5 g1f1 c7c5 f1e1 d7e5 h4h6 e5f3 g2f3 c5f5');
INSERT INTO evaluations VALUES(48239,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(48239,31,'cp -53',NULL);
INSERT INTO evaluations VALUES(48239,32,'cp -47',NULL);
INSERT INTO evaluations VALUES(48239,33,'cp -43',NULL);
INSERT INTO evaluations VALUES(48239,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(48239,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(48239,36,'cp -52','d5d4 c3a4 b8c6 g1f3 e6e5 d2d3 f8e7 a2a3 a7a5 e1g1 e8g8 e2e3 a8b8 f1e1 f8e8 e3d4 e5d4 c1d2 c8e6 d1b3 h7h6 e1e6 f7e6 a1e1 e7f8 c4c5 d8d5 b3c2 d5d7 b2b4 b7b5 c5b6 c7b6 b4b5 c6a7 a4b2 b8c8 b2c4 a7b5 f3e5 d7d8');
INSERT INTO evaluations VALUES(48259,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(48259,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(48259,32,'cp -24',NULL);
INSERT INTO evaluations VALUES(48259,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(48259,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(48259,35,'cp -44',NULL);
INSERT INTO evaluations VALUES(48259,36,'cp -39','c7c5 e2e3 c5d4 e3d4 d5c4 g1f3 d8c7 c2a4 c8d7 a4c4 d7c6 f3e5 a7a6 f1e2 f8d6 e5c6 b8c6 c1g5 b7b5 c4d3 d6e7 a1c1 e8g8 g5e3 f8d8 e1g1 c6e5 d3d1 e5c4 a2a4 e7d6 g1h1 c4b2 d1b3 b2c4 a4b5 c4e3 f2e3 a6b5 c3b5 c7b6 b3c2 d6e7 c2c7 b6c7 c1c7');
INSERT INTO evaluations VALUES(49428,30,'cp -3',NULL);
INSERT INTO evaluations VALUES(49428,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49428,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49428,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49428,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(49428,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49428,36,'cp 0','d7d5 e2e3 c7c5 b2b3 c5d4 e3d4 f8d6 f1d3 e8g8 e1g1 b8c6 c1b2 b7b6 a2a3 c8b7 d1e2 a8c8 f3e5 c8c7 f2f4 c6e7 c2c3 e7f5 g2g4 f5h4 a3a4 f6e8 g4g5 d6e5 e2e5 d8d7 d3b5 d7d8');
INSERT INTO evaluations VALUES(49622,30,'cp -29',NULL);
INSERT INTO evaluations VALUES(49622,31,'cp -33',NULL);
INSERT INTO evaluations VALUES(49622,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(49622,33,'cp -33',NULL);
INSERT INTO evaluations VALUES(49622,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(49622,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(49622,36,'cp -37','d7d5 d2d4 e5d4 e2d4 f8b4 e4d5 e8g8 f1e2 f6d5 c1d2 b4c3 b2c3 c7c5 d4b3 d8f6 e1g1 d5c3 d2c3 f6c3 d1d5 b8d7 e2b5 d7f6 d5c5 c3c5 b3c5 b7b6 c5b3 c8b7 a2a4 a8c8 f1c1 f6d5 b5d3 d5f4 d3f1 g7g6 c2c4 f4e6 a4a5 f8d8 a5a6 b7c6 f2f3');
INSERT INTO evaluations VALUES(49742,30,'cp 132',NULL);
INSERT INTO evaluations VALUES(49742,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(49742,32,'cp 149',NULL);
INSERT INTO evaluations VALUES(49742,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(49742,34,'cp 140',NULL);
INSERT INTO evaluations VALUES(49742,35,'cp 142',NULL);
INSERT INTO evaluations VALUES(49742,36,'cp 143','c7c6 c3e4 f6e4 g2e4 d7d5 c4d5 c6d5 e4g2 b8c6 d2d4 c8f5 c1d2 c6d4 d1a4 d4c6 d2c3 d8d7 e2e3 d5d4 e3d4 f8d6 g1f3 d7e7 f3e5 e8g8 e1g1 c6e5 d4e5 d6c5 a4b5 a8b8 f1d1 a7a6 b5e2 f5e6 a2a3 b8c8 e2f3 b7b5 f3e2 f8d8 d1d8 c8d8 h2h4 e7d7 g1h2');
INSERT INTO evaluations VALUES(50112,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(50112,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(50112,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(50112,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(50112,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(50112,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(50112,36,'cp 72','f1g2 g7g6 c2c4 f8g7 d2d4 e8g8 b1c3 c7c6 d4d5 d7d6 f3d4 c6d5 c4d5 b8a6 e1g1 c8d7 a2a4 a6c5 b2b4 c5e4 c3e4 f6e4 c1e3 a8c8 b4b5 a7a6 b5a6 b7a6 d4c6 d7c6 d5c6 g7a1 d1d5 f8f7 f1a1');
INSERT INTO evaluations VALUES(50113,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(50113,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(50113,32,'cp 56',NULL);
INSERT INTO evaluations VALUES(50113,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(50113,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(50113,35,'cp 71',NULL);
INSERT INTO evaluations VALUES(50113,36,'cp 61','g7g6 d2d4 f8g7 e1g1 e8g8 c2c4 c7c6 b2b3 f6e4 c1b2 d7d5 b1c3 c8e6 d1c2 b8d7 e2e3 e6f7 f1d1 a8c8 a1c1 e7e6 c3e2 g6g5 f3e5 d8e7 f2f3 e4f6 a2a4 a7a5 e5f7 f8f7 c2d2 g5g4 e2f4 g4f3 g2f3 c8e8');
INSERT INTO evaluations VALUES(51787,30,'cp 37',NULL);
INSERT INTO evaluations VALUES(51787,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(51787,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(51787,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(51787,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(51787,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(51787,36,'cp 37','g8f6 g1f3 d7d5 f1d3 d5c4 d3c4 a7a6 e1g1 c6c5 d4c5 f8c5 d1d8 e8d8 b1d2 d8e7 b2b3 b7b6 c4e2 b8d7 c1b2 c8b7 a2a3 h8c8 b3b4 c5d6 d2c4 d6c7 a3a4 f6d5 b4b5 d7c5 f3d4 a6b5 a4b5 f7f6 e2f3 a8a1 f1a1 c8a8 b2a3 e7f7 a1c1 a8c8');
INSERT INTO evaluations VALUES(44047,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(44047,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(44047,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(44047,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(44047,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(44047,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44047,36,'cp 26','f1d3 d7d5 g1f3 d5c4 d3c4 a7a6 e1g1 c6c5 b2b3 c5d4 d1d4 b8d7 a2a4 b7b6 c1b2 f8e7 b1d2 c8b7 f1d1 e8g8 a1c1 a8c8 h2h3 d7c5 c4e2 d8d4 f3d4 c5d7 e2f3 c8c1 d1c1 b7f3 d2f3 e7c5 d4c6 f6d5 f3e5 d7e5 c6e5');
INSERT INTO evaluations VALUES(53221,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(53221,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53221,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(53221,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(53221,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(53221,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53221,36,'cp 0','b8c6 f1b5 c8d7 b2b3 e7e6 c1b2 c6d4 b5d7 d8d7 f3e5 d7c7 c3e2 d4f5 e2g3 f5d4');
INSERT INTO evaluations VALUES(53791,30,'cp 336',NULL);
INSERT INTO evaluations VALUES(53791,31,'cp 330',NULL);
INSERT INTO evaluations VALUES(53791,32,'cp 303',NULL);
INSERT INTO evaluations VALUES(53791,33,'cp 314',NULL);
INSERT INTO evaluations VALUES(53791,34,'cp 312',NULL);
INSERT INTO evaluations VALUES(53791,35,'cp 309',NULL);
INSERT INTO evaluations VALUES(53791,36,'cp 300','d1d8 e8d8 f3e5 c8e6 f1c4 e6c4 e5c4 b8d7 a2a4 f7f6 e1e2 h7h5 h2h4 g8h6 c1f4 d7c5 f2f3 d8e8 h1d1 h6f7 b1d2 c5e6 f4e3 f8c5 e3c5 e6c5 b2b4 c5e6 d2b3 g7g5 h4g5 f6g5 c4a5 g5g4 a5b7 h5h4 f3g4 h8g8 a4a5');
INSERT INTO evaluations VALUES(35402,30,'cp 312',NULL);
INSERT INTO evaluations VALUES(35402,31,'cp 304',NULL);
INSERT INTO evaluations VALUES(35402,32,'cp 309',NULL);
INSERT INTO evaluations VALUES(35402,33,'cp 290',NULL);
INSERT INTO evaluations VALUES(35402,34,'cp 299',NULL);
INSERT INTO evaluations VALUES(35402,35,'cp 330',NULL);
INSERT INTO evaluations VALUES(35402,36,'cp 308','e8d8 f3e5 c8e6 f1c4 e6c4 e5c4 b8d7 a2a4 f7f6 c1f4 g8h6 b1d2 h6f7 e1e2 g7g5 f4g3 d8e8 h2h3 d7c5 c2c3 h7h5 b2b4 c5e6 h1d1 h5h4 g3h2 h8g8 e2f1 g5g4 h3g4 g8g4 c4e3 g4g8 e3f5 g8h8 d2c4 h4h3 a4a5 a8d8 d1d8 e8d8 a1d1 d8e8');
INSERT INTO evaluations VALUES(35443,30,'cp 313',NULL);
INSERT INTO evaluations VALUES(35443,31,'cp 335',NULL);
INSERT INTO evaluations VALUES(35443,32,'cp 317',NULL);
INSERT INTO evaluations VALUES(35443,33,'cp 313',NULL);
INSERT INTO evaluations VALUES(35443,34,'cp 333',NULL);
INSERT INTO evaluations VALUES(35443,35,'cp 305',NULL);
INSERT INTO evaluations VALUES(35443,36,'cp 316','f3e5 c8e6 f1c4 e6c4 e5c4 b8d7 a2a4 f7f6 c1f4 g8h6 f4h6 g7h6 e1e2 d8c7 c4e3 a7a5 b1d2 a8e8 f2f3 f8c5 e3f5 h8g8 g2g3 h6h5 d2c4 g8g5 e2f1 c5a7 b2b3 d7c5 h2h4 g5g8 f1g2 e8d8 h1d1 c5e6 c4a5 e6d4 f5d4');
INSERT INTO evaluations VALUES(35441,30,'cp 312',NULL);
INSERT INTO evaluations VALUES(35441,31,'cp 321',NULL);
INSERT INTO evaluations VALUES(35441,32,'cp 305',NULL);
INSERT INTO evaluations VALUES(35441,33,'cp 311',NULL);
INSERT INTO evaluations VALUES(35441,34,'cp 317',NULL);
INSERT INTO evaluations VALUES(35441,35,'cp 324',NULL);
INSERT INTO evaluations VALUES(35441,36,'cp 319','c8e6 f1c4 e6c4 e5c4 b8d7 a2a4 f7f6 c1f4 g8h6 b1d2 h6f7 e1c1 d8e8 h1e1 g7g5 f4g3 a8d8 f2f3 f8e7 c4a5 d7c5 d2c4 d8d1 e1d1 f7d8 g3f2 c5a4 f2a7 h8g8 c4e3 e7b4 a5b3 d8e6 c2c3 b4c5 b3c5 a4c5');
INSERT INTO evaluations VALUES(54045,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(54045,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(54045,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(54045,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(54045,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(54045,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(54045,36,'cp 0','c8d7 b5d3 c5d4 f3d4 g8e7 b1d2 e7c6 d2f3 c6d4 f3d4 d8c7 d1e2 f8b4 e1f1 b4c5 c2c3 b8c6 d4c6 d7c6 h2h4 e8g8 e2h5 f7f5 e5f6 f8f6 h5h7 g8f8 h7h8 f8f7 h8h5 f7g8');
INSERT INTO evaluations VALUES(30887,30,'cp 9',NULL);
INSERT INTO evaluations VALUES(30887,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(30887,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(30887,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(30887,34,'cp 4',NULL);
INSERT INTO evaluations VALUES(30887,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(30887,36,'cp 15','b5d3 c5d4 f3d4 g8e7 b1d2 e7c6 d2f3 c6d4 f3d4 d8c7 d1e2 f8b4 e1f1 b8c6 d4c6 d7c6 c2c3 b4c5 h2h4 e8g8 c1g5 f7f6 e5f6 e6e5 d3f5 g8h8 b2b4 c5b6 a2a4 a7a5 b4b5 c6e8 h1h3 g7f6 g5e3 c7c3 a1c1 c3a3 g2g4 b6e3 e2e3 a3e3 h3e3 e5e4');
INSERT INTO evaluations VALUES(54324,30,'cp 92',NULL);
INSERT INTO evaluations VALUES(54324,31,'cp 99',NULL);
INSERT INTO evaluations VALUES(54324,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(54324,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(54324,34,'cp 90',NULL);
INSERT INTO evaluations VALUES(54324,35,'cp 104',NULL);
INSERT INTO evaluations VALUES(54324,36,'cp 80','d8d5 c1e3 g8f6 b1c3 d5e5 d1e2 c8d7 e1c1 f8b4 f2f4 e5a5 d4b3 a5c7 c3b5 d7b5 e2b5 b8c6 g2g3 f6d5 e3f2 e8g8 f1g2 a7a6 b5e2 d5b6 a2a3 b4e7 d1d2 e7f6 b3c5 f8d8 h1d1 d8d2 d1d2 a8d8 d2d8 c6d8 c5e4 f6e7 e4c3 b6c4');
INSERT INTO evaluations VALUES(54969,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(54969,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(54969,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(54969,33,'cp 31',NULL);
INSERT INTO evaluations VALUES(54969,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(54969,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(54969,36,'cp 21','e7e5 d4e5 d6e5 d1d8 e8d8 b1c3 c8e6 f2f4 f8b4 f4e5 b8d7 c1d2 d7e5 e1c1 g8f6 g1f3 e5c4 d2f4 c4d6 c3b5 f6e4 b5d6 b4d6 d1d4 f7f5 f4d6 e4d6 e2e4 d8e7 e4f5 e6f5 f3d2 h8f8 f1e2 a8e8 e2f3 b7b6 d2c4 d6c4 d4c4 c7c5 a2a4 a7a5 c4h4 f8f6 h1e1 e7d7');
INSERT INTO evaluations VALUES(55222,30,'cp -186',NULL);
INSERT INTO evaluations VALUES(55222,31,'cp -210',NULL);
INSERT INTO evaluations VALUES(55222,32,'cp -204',NULL);
INSERT INTO evaluations VALUES(55222,33,'cp -190',NULL);
INSERT INTO evaluations VALUES(55222,34,'cp -191',NULL);
INSERT INTO evaluations VALUES(55222,35,'cp -183',NULL);
INSERT INTO evaluations VALUES(55222,36,'cp -191','e7e5 e2e3 e5d4 f1c4 f8b4 e1f1 b8c6 e3d4 g8f6 g1f3 e8g8 a2a3 b4a5 c1b2 d8d6 h2h3 c8f5 b1c3 a5c3 b2c3 a8d8 f1g1 f5e4 c3b2 f8e8 a1c1 h7h6 b3b4 g8f8 c4b3 d6f4 b4b5 c6a5 b3a2 e4d5 a2d5 f6d5 f3e5');
INSERT INTO evaluations VALUES(57730,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(57730,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(57730,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(57730,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(57730,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(57730,35,'cp 103',NULL);
INSERT INTO evaluations VALUES(57730,36,'cp 93','e7e6 e2e4 d7d5 e4d5 e6d5 b1c3 g8f6 g1f3 b8c6 c1g5 f8b4 f1e2 d5c4 e1g1 e8g8 h2h3 c6a5 d1a4 b4c3 b2c3 b7b6 f3e5 c8b7 f1e1 d8d6 g5f6 d6f6 e2h5 a8d8 a1d1 b7c6 e5c6 f6c6 a4c6 a5c6 h5f3 c6a5 e1e3 c7c5 d4c5 d8d1 f3d1 f8d8');
INSERT INTO evaluations VALUES(267,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(267,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(267,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(267,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(267,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(267,35,'cp 38',NULL);
INSERT INTO evaluations VALUES(267,36,'cp 39','f6g6 h2h4 g8f6 b1c3 c5b4 g5d2 b4c3 b2c3 g6h5 f3g5 h5d1 e1d1 h8f8 f2f4 d7d6 a1b1 b7b6 g5f3 c8g4 d1c1 e5f4 d2f4 e8d7 f3d4 a7a6 h1f1 b6b5 c4b3 c6d4 c3d4 f6h5 f4h2 f7f5 c1b2 f5f4 b1e1 a8e8 d4d5 h5g3');
INSERT INTO evaluations VALUES(303,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(303,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(303,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(303,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(303,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(303,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(303,36,'cp 33','h2h4 g8f6 b1c3 h7h6 g5f6 g6f6 c3d5 f6d8 c2c3 d7d6 b2b4 c5b6 a2a4 a7a6 b4b5 c6e7 d5b6 c7b6 e1g1 e8g8 b5a6 b7a6 d3d4 d8c7 d1b3 a8b8 f1b1 e7g6 g2g3 g8h8 b1e1 f7f5 e4f5 f8f5 f3d2 g6e7 a1b1 f5f8 d2e4 c8b7 c4e6 e7g6');
INSERT INTO evaluations VALUES(503,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(503,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(503,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(503,33,'cp 40',NULL);
INSERT INTO evaluations VALUES(503,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(503,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(503,36,'cp 43','f1b5 c8d7 c2c3 a7a6 b5a4 e8g8 e1g1 a8d8 f1e1 f8e8 e1e3 h7h6 a4b3 e6e5 d4d5 c6a5 d1e1 e5e4 e3e4 f6b6 f3d4 a5b3 a2b3 b6d6 c3c4 c7c6 d5c6 e8e4 e1e4 d7c6 d4c6 d6d1 e4e1 d1e1 a1e1 b7c6 g2g3 d8d3 e1a1 d3b3 a1a6 g8f8 a6c6 b3b2 g1g2 b2b3 c6c7 b3c3');
INSERT INTO evaluations VALUES(502,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(502,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(502,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(502,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(502,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(502,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(502,36,'cp 50','c8d7 c2c3 a7a6 b5a4 a8d8 e1g1 e8g8 f1e1 f8e8 e1e3 h7h6 d1e1 c6e7 a4c2 e7f5 e3e2 d7b5 e2d2 f5h4 f3e5 c7c5 a2a4 b5c6 e5c6 b7c6 a1d1 c5d4 d2d4 d8d4 d1d4 c6c5 d4d6 g7g6 e1d2 a6a5 h2h3 g8g7 c2d3 e8d8 d3f1 d8d6 d2d6 f6g5 d6g3 g5g3');
INSERT INTO evaluations VALUES(27337,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(27337,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(27337,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(27337,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(27337,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(27337,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(27337,36,'cp 42','e1g1 a7a6 b5a4 e8g8 f1e1 f8d8 c2c3 c6e7 a4c2 a8c8 b2b4 d7e8 d1b1 h7h6 c2h7 g8h8 h7d3 e7d5 b1c2 d5f4 d3e4 b7b6 e4b7 c8b8 b7a6 c7c5 c2e4 c5d4 c3d4 b6b5 a6b7 f4g6 a2a3 f6e7 b7c6 e8c6 e4c6');
INSERT INTO evaluations VALUES(2735,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(2735,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(2735,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(2735,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(2735,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(2735,35,'cp 77',NULL);
INSERT INTO evaluations VALUES(2735,36,'cp 59','d6d5 e4e5 f6e4 b1d2 c8f5 b3c2 c6b4 c2b1 c7c5 d2e4 d5e4 b1e4 f5e4 e1e4 d8d5 d1e2 f8e8 c1g5 e7g5 d4c5 b4d3 f3g5 d3c5 e4e3 h7h6 g5e4 c5e4 e3e4 a8c8 b2b3 c8c3 a1e1 c3d3 e2g4 d5d7 g4d7 d3d7 e4e2 a6a5 f2f4 e8d8 h2h3 d7d2 g1h2');
INSERT INTO evaluations VALUES(2892,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2892,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2892,32,'cp 3',NULL);
INSERT INTO evaluations VALUES(2892,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2892,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2892,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(2892,36,'cp 0','e1g1 d7d6 d1e2 h7h6 g5h4 f6e4 e2e4 e7h4 e4c6 c8e6 c4b3 a8c8 c6e4 g8h8 a1d1 h4e7 e4e2 a7a5 c3d5 e7g5 d5c3');
INSERT INTO evaluations VALUES(2891,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(2891,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(2891,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(2891,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(2891,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(2891,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(2891,36,'cp 0','d7d6 d1e2 h7h6 g5h4 f6e4 e2e4 e7h4 e4c6 c8e6 c4b3 a8c8 c6e4 g8h8 a1d1 h4e7 e4e2 a7a5 c3d5 e7g5 d5c3');
INSERT INTO evaluations VALUES(3223,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(3223,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(3223,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(3223,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(3223,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(3223,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(3223,36,'cp 54','f2f3 c8d7 e1c1 a8c8 c4b3 c6d4 e3d4 b7b5 a2a3 a7a5 c3d5 f6d5 d4g7 g8g7 e4d5 g7g8 h1e1 b5b4 a3b4 a5b4 d2b4 f8e8 b3c4 c8c5 b2b3 d8c7 b4d2 h7h5 e1e3 d7b5 c4b5 c5b5 c1b2 e8a8 d1a1 a8a1 b2a1 b5d5 d2d5 c7a7 a1b1 a7e3 b1b2 e3e2');
INSERT INTO evaluations VALUES(3493,30,'cp 76',NULL);
INSERT INTO evaluations VALUES(3493,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(3493,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(3493,33,'cp 67',NULL);
INSERT INTO evaluations VALUES(3493,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(3493,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(3493,36,'cp 61','d1d2 a7a6 b2b3 c8d7 a1c1 e7e5 d4d5 c6d4 g1e2 d4e2 f1e2 f6h5 e1g1 h5f4 c1e1 f4e2 c3e2 d8e7 h2h3 b7b6 e2g3 a6a5 f1f2 f7f6 g1h1 f8f7 a2a3 f7f8 h1h2 f8f7 b3b4 a5b4 a3b4 a8a3 e1c1 f6f5 e4f5 d7f5 e3g5 e7e8 g3f5');
INSERT INTO evaluations VALUES(3491,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(3491,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(3491,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(3491,33,'cp 60',NULL);
INSERT INTO evaluations VALUES(3491,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(3491,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(3491,36,'cp 63','a7a6 a1c1 c8d7 b2b3 e7e5 d4d5 c6d4 g1e2 d4e2 f1e2 f6h5 e1g1 h5f4 e2d1 b7b6 c3e2 f7f5 e2f4 e5f4 e3f4 f5e4 f3e4 d8f6 d1e2 f6d4 d2d4 g7d4 g1h1 a6a5 h2h3 a5a4 b3b4 a4a3 c1b1 a8a4 e2d3 d4b2 c4c5 a4b4 c5c6 d7e8 f1f3 b4a4');
INSERT INTO evaluations VALUES(4416,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(4416,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(4416,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(4416,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4416,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(4416,35,'cp 22',NULL);
INSERT INTO evaluations VALUES(4416,36,'cp 16','d7d5 c4e2 f6e4 d1b3 b4c3 b2c3 c6a5 b3b2 c7c5 c1e3 c5d4 e3d4 c8f5 e1g1 a8c8 a1d1 d8c7 d1c1 b7b6 f3h4 f5e6 f2f3 e4c5 b2d2 c7e7 d2f4 h7h6 f1e1 e7g5 f4g3 a5c4 f3f4 g5g3 h2g3 c5e4');
INSERT INTO evaluations VALUES(5116,30,'cp -23',NULL);
INSERT INTO evaluations VALUES(5116,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(5116,32,'cp -32',NULL);
INSERT INTO evaluations VALUES(5116,33,'cp -31',NULL);
INSERT INTO evaluations VALUES(5116,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(5116,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(5116,36,'cp -41','g8e7 c4b3 e8g8 d2d3 d7d5 e4e5 e7f5 c3e2 b7b5 c2c3 d5d4 e2g3 c8b7 g3f5 g6f5 a2a4 d4c3 a4b5 c3b2 c1b2 c6b4 b3c4 a7a6 b5a6 b7a6 c4a6 a8a6 a1a6 b4a6 d3d4 d8a8 d1a1 f8b8 d4c5 a6c5 a1a8 b8a8 f1d1 c5e4 b2d4 g7f8 g1h1');
INSERT INTO evaluations VALUES(5563,30,'cp 69',NULL);
INSERT INTO evaluations VALUES(5563,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(5563,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(5563,33,'cp 66',NULL);
INSERT INTO evaluations VALUES(5563,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(5563,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(5563,36,'cp 77','e4d5 e6d5 d1d5 d8e7 d5e4 c8f5 e4e7 g8e7 c1d2 c5b6 e1c1 b6f2 f1c4 e8g8 h1f1 f2b6 d1e1 h7h6 b2b4 a8e8 a2a3 a7a6 c4b3 b6c7 e1e2 f5e6 b3e6 f7e6 f1e1 e6e5 d2e3 e8d8 e3c5 f8e8 e2d2 d8d2 f3d2 e7f5 c3e4 e8d8');
INSERT INTO evaluations VALUES(5562,30,'cp 60',NULL);
INSERT INTO evaluations VALUES(5562,31,'cp 51',NULL);
INSERT INTO evaluations VALUES(5562,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(5562,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(5562,34,'cp 49',NULL);
INSERT INTO evaluations VALUES(5562,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(5562,36,'cp 61','e6d5 d1d5 d8e7 d5e4 c8f5 e4e7 g8e7 c1d2 c5b6 e1c1 b6f2 f1c4 e8g8 h1f1 f2b6 d1e1 h7h6 b2b4 a8e8 a2a3 a7a6 c4b3 f5g4 c3a4 b6a7 f3e5 c6e5 e1e5 a7b8 e5e4 g4d7 a4c5 d7b5 f1e1 e7f5 e4e8 b5e8 b3d5 b7b6 c5a6');
INSERT INTO evaluations VALUES(6747,30,'cp 206',NULL);
INSERT INTO evaluations VALUES(6747,31,'cp 243',NULL);
INSERT INTO evaluations VALUES(6747,32,'cp 260',NULL);
INSERT INTO evaluations VALUES(6747,33,'cp 256',NULL);
INSERT INTO evaluations VALUES(6747,34,'cp 244',NULL);
INSERT INTO evaluations VALUES(6747,35,'cp 270',NULL);
INSERT INTO evaluations VALUES(6747,36,'cp 239','c1f4 d7d5 e4d5 d8d5 b1d2 g8e7 f1c4 d5f5 f4g3 b7b5 c4e2 e8g8 e1g1 a5b6 a2a4 b5a4 d1a4 f5d7 a4c2 e7f5 g3f2 c8b7 b2b4 c6e7 f1e1 g8h8 d2e4 a7a5 b4b5 b7e4 c2e4 f5d6 f3e5 d7d8 e4h4 e7g6 h4d8 f8d8');
INSERT INTO evaluations VALUES(6746,30,'cp 230',NULL);
INSERT INTO evaluations VALUES(6746,31,'cp 250',NULL);
INSERT INTO evaluations VALUES(6746,32,'cp 241',NULL);
INSERT INTO evaluations VALUES(6746,33,'cp 230',NULL);
INSERT INTO evaluations VALUES(6746,34,'cp 214',NULL);
INSERT INTO evaluations VALUES(6746,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(6746,36,'cp 251','d7d5 e4d5 d8d5 b1d2 a5b6 f1c4 d5f5 f4g5 c8e6 e1g1 g8f6 g5h4 f5g4 c4e6 f7e6 h2h3 g4h5 d1e2 e8c8 e2e6 c8b8 a1e1 d8e8 e6b3 a7a6 b3c2 b6a7 c2d3 h8f8 a2a3 c6e7 d2c4 e7f5 h4f2 h7h6 e1e8 f8e8');
INSERT INTO evaluations VALUES(10182,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(10182,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(10182,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10182,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10182,34,'cp 11',NULL);
INSERT INTO evaluations VALUES(10182,35,'cp 17',NULL);
INSERT INTO evaluations VALUES(10182,36,'cp 8','d7d5 c4d5 f6d5 f1g2 e8g8 e1g1 h7h6 c3d5 d8d5 c1d2 d5d6 a1c1 c8e6 a2a3 a7a5 d1c2 c5b6 d2e3 e6d5 e3b6 c7b6 c2d2 f8d8 d2e3 b6b5 e3c5 b5b4 c5d6 d8d6 c1c5 b4a3 b2a3 b7b6 c5c3 d5f3 g2f3 c6d4 g1g2 a8d8 f1c1');
INSERT INTO evaluations VALUES(12574,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(12574,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(12574,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(12574,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(12574,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(12574,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(12574,36,'cp -16','f8b4 g1e2 h7h6 g5f6 d8f6 f2f3 e8g8 a2a3 b4c5 c3d5 f6h4 e2g3 d7d6 b2b4 c5b6 d5b6 a7b6 f1d3 c8e6 e1g1 d6d5 d1d2 c6e7 f3f4 d5e4 d3e4 a8d8 e4d3 e6c4 f4e5 c4d3 c2d3 h4d4 d2f2 d4f2 f1f2 d8d3 g3f5 e7f5');
INSERT INTO evaluations VALUES(13615,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(13615,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(13615,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(13615,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(13615,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(13615,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(13615,36,'cp 0','f8e7 e2e4 e8g8 f1e1 b7b5 e4d5 f6d5 c2c4 b5c4 d3c4 d5b4 f3e5 c8b7 a2a3 c6e5 g2b7 b4d3 b7a8 d3e1 d1e1 e5d3 e1e2 d3c1 a1c1 d8a8 c1d1 f8d8 d2f3 d8d1 e2d1 a8b7 b2b3 g7g6 d1d3 g8g7 f3e5 b7c7 d3e4 e7f6 e5d3 a7a5 b3b4 c5b4 a3b4 a5b4 d3b4 f6c3 b4c2');
INSERT INTO evaluations VALUES(8095,30,'cp 1',NULL);
INSERT INTO evaluations VALUES(8095,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(8095,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(8095,33,'cp 3',NULL);
INSERT INTO evaluations VALUES(8095,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(8095,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(8095,36,'cp 0','e2e4 e8g8 f1e1 b7b5 e4d5 f6d5 c2c4 b5c4 d3c4 d5b4 f3e5 c8b7 a2a3 c6e5 g2b7 b4d3 b7a8 d3e1 d1e1 e5d3 e1e2 d3c1 a1c1 d8a8 c1c3 f8b8 d2e4 f7f5 e4d2 a8c6 d2f3 e7f6 c3e3 b8b2 e3e6 c6d7 e2e1 h7h6 e6e2 d7f7 e2b2 f6b2 a3a4 f7c4 e1e8 g8h7 f3g5 h6g5 e8h5 h7g8 h5e8');
INSERT INTO evaluations VALUES(13922,30,'cp -178',NULL);
INSERT INTO evaluations VALUES(13922,31,'cp -177',NULL);
INSERT INTO evaluations VALUES(13922,32,'cp -178',NULL);
INSERT INTO evaluations VALUES(13922,33,'cp -172',NULL);
INSERT INTO evaluations VALUES(13922,34,'cp -161',NULL);
INSERT INTO evaluations VALUES(13922,35,'cp -170',NULL);
INSERT INTO evaluations VALUES(13922,36,'cp -177','c4b5 c8d7 d2d3 h7h6 g5h3 d6d5 b1d2 d8c7 f3e2 e8c8 e1g1 c6d4 b5d7 d8d7 e2d1 d4c6 e4d5 e6d5 f2f4 f8d6 c2c3 c8b8 g1h1 c6e7 d2f3 e7f5 f3e5 d7e7 c1d2 a7a6 b2b3 h8e8 f1e1 d6e5 f4e5 e7e5 e1e5 c7e5 d2f4');
INSERT INTO evaluations VALUES(15042,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(15042,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(15042,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(15042,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(15042,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(15042,35,'cp -16',NULL);
INSERT INTO evaluations VALUES(15042,36,'cp -22','c8g4 c1e3 f8b4 h2h3 g4h5 e1g1 e8g8 g2g4 h5g6 f3e5 c6e5 d4e5 d5d4 e3d4 d8d4 e5f6 b4c3 f6g7 f8d8 b2c3 d4c3 a1c1 c3f6 f1e1 d8d4 d1e2 a8d8 e2e7 f6f4 d3g6 h7g6 g1g2 g8g7 e1e3 d8d7 e7e5 f4e5 e3e5 b7b6 g2g3 a7a5 c2c4 c7c6');
INSERT INTO evaluations VALUES(15999,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(15999,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(15999,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(15999,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(15999,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(15999,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(15999,36,'cp 20','e7e5 a2a4 f8e7 g1f3 h7h6 g5e3 e8g8 f1c4 e5d4 e3d4 f8e8 e1g1 d7f8 h2h3 c8e6 f3d2 e6c4 d2c4 f8e6 f1e1 e6d4 d1d4 f6d7 c3d5 e7f6 d5f6 d8f6 d4f6 d7f6 f2f3 d6d5 e4d5 e8e1 a1e1 f6d5 c4a5 d5b4 a5b7 b4c2 e1d1');
INSERT INTO evaluations VALUES(15909,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(15909,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(15909,32,'cp 20',NULL);
INSERT INTO evaluations VALUES(15909,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(15909,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(15909,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(15909,36,'cp 36','a2a4 f8e7 f1c4 e5d4 d1d4 h7h6 g5f4 e8g8 g1e2 d7c5 e1g1 c8e6 e2g3 f8e8 c4e2 e7f8 a1d1 e6d7 d4c4 d7e6 c4b4 c7c6 e4e5 f6d5 c3d5 e6d5 e5d6 f8d6 f4d6 d8d6 a4a5 d6g6');
INSERT INTO evaluations VALUES(16766,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(16766,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(16766,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(16766,33,'cp -68',NULL);
INSERT INTO evaluations VALUES(16766,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(16766,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(16766,36,'cp -42','f8c5 d5c6 c5f2 e1e2 f2b6 d1d5 e4f2 c6b7 c8b7 d5b7 e8g8 c1g5 f7f6 g5e3 f2h1 e3b6 a7b6 b1d2 c7c6 g2g3 e5e4 f3d4 a8b8 b7a6 d8e8 e2e3 f6f5 f1c4 d7d5 a1h1 d5c4 a6c4 g8h8 c4c6 e8h5 e3f2 f5f4 g3f4 h5h4 f2e2 h4g4 e2e1 e4e3 d2f3 g4g2 h1f1 g2b2 d4e6 b2a2 e6f8');
INSERT INTO evaluations VALUES(18781,30,'cp 41',NULL);
INSERT INTO evaluations VALUES(18781,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(18781,32,'cp 44',NULL);
INSERT INTO evaluations VALUES(18781,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(18781,34,'cp 30',NULL);
INSERT INTO evaluations VALUES(18781,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(18781,36,'cp 37','d2d4 d7d5 c4d5 e6d5 f1g2 g8f6 b1c3 c5d4 f3d4 f8c5 d4b3 c5b6 c3d5 c8e6 e2e4 e6g4 d1c2 f6d5 e4d5 c6e5 c2c3 f7f6 e1g1 a8c8 c3b4 a7a5 b4a4 g4d7 a4e4 c8c4 e4b1 e8g8 c1f4 c4c8 d5d6 a5a4 b3d2 d7b5 g2b7 d8d6 b7c8 f8c8');
INSERT INTO evaluations VALUES(19822,30,'cp -81',NULL);
INSERT INTO evaluations VALUES(19822,31,'cp -80',NULL);
INSERT INTO evaluations VALUES(19822,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(19822,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(19822,34,'cp -76',NULL);
INSERT INTO evaluations VALUES(19822,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(19822,36,'cp -91','c8f5 f1e2 e7e6 d2d3 h7h5 g1f3 g8h6 b1a3 d8b6 d1b3 b6c7 e1g1 f8e7 c1e3 f5g6 h2h3 h6f5 e3f2 h5h4 b3d1 f5g3 f1e1 c7b6 d1d2 b6a6 a1d1 g3e2 d2e2 g6h5 c3c4 d5d4 a3b5 e8c8 b5d6 e7d6 e5d6');
INSERT INTO evaluations VALUES(26280,30,'cp -86',NULL);
INSERT INTO evaluations VALUES(26280,31,'cp -76',NULL);
INSERT INTO evaluations VALUES(26280,32,'cp -94',NULL);
INSERT INTO evaluations VALUES(26280,33,'cp -92',NULL);
INSERT INTO evaluations VALUES(26280,34,'cp -88',NULL);
INSERT INTO evaluations VALUES(26280,35,'cp -87',NULL);
INSERT INTO evaluations VALUES(26280,36,'cp -82','g1f3 e7e6 f1e2 h7h5 b1a3 g8h6 d2d3 f5g6 e1g1 h6f5 a3c2 d8b6 a1b1 a8d8 c1d2 f8e7 d2e1 d5d4 e1f2 e8g8 a2a4 b6c7 d1e1 g6h7 c3d4 c5d4 e1d2 a7a5 h2h3 c7b6 g2g4 h5g4 h3g4 f5e3 c2e3 d4e3 f2e3 b6b3 b1d1');
INSERT INTO evaluations VALUES(26279,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(26279,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(26279,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(26279,33,'cp -75',NULL);
INSERT INTO evaluations VALUES(26279,34,'cp -91',NULL);
INSERT INTO evaluations VALUES(26279,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(26279,36,'cp -87','e7e6 f1e2 h7h5 e1g1 g8h6 b1a3 f5g6 a3c2 h6f5 d2d3 d8b6 a1b1 f8e7 c1d2 a8d8 d2e1 e8g8 e1f2 d5d4 a2a4 b6c7 b1c1 c7d7 h2h3 h5h4 c3d4 c6d4 c2d4 c5d4 c1c4 b7b5 a4b5 d7b5 b2b3 f5e3 f3d4 e3d1 d4b5 d1f2 f1f2 g6d3');
INSERT INTO evaluations VALUES(26415,30,'cp -89',NULL);
INSERT INTO evaluations VALUES(26415,31,'cp -77',NULL);
INSERT INTO evaluations VALUES(26415,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(26415,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(26415,34,'cp -94',NULL);
INSERT INTO evaluations VALUES(26415,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(26415,36,'cp -76','f1e2 h7h5 b1a3 g8h6 a3c2 h5h4 d2d3 f8e7 e1g1 e8g8 h2h3 f5h7 c1e3 h6f5 e3f2 d5d4 d1d2 d8a5 c3d4 a5d2 f3d2 f8d8 d4c5 f5g3 f2g3 h4g3 f1f3 e7c5 g1f1 c5f2 d2c4 b7b5 c4d6 f7f6 f3f2 g3f2 e2f3 a8b8 f3c6 h7d3 f1f2 d3c2 c6b5');
INSERT INTO evaluations VALUES(20394,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20394,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(20394,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(20394,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20394,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(20394,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(20394,36,'cp 0','d5e4 d2e4 f8e7 f1d3 g8f6 e1g1 e8g8 e4f6 e7f6 c1e3 c6b4 d3e4 b4d5 d1d3 h7h6 f3e5 b7b6 e3d2 c8b7 a1c1 d8e7 a2a3 a8d8 f2f4 a7a5 e4h7 g8h8 h7e4 h8g8');
INSERT INTO evaluations VALUES(20762,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(20762,31,'cp 123',NULL);
INSERT INTO evaluations VALUES(20762,32,'cp 141',NULL);
INSERT INTO evaluations VALUES(20762,33,'cp 121',NULL);
INSERT INTO evaluations VALUES(20762,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(20762,35,'cp 124',NULL);
INSERT INTO evaluations VALUES(20762,36,'cp 127','d8c7 d5d3 g8f6 b1c3 c8e6 c4b3 h7h6 c1e3 a7a6 e1c1 f8e7 c1b1 b7b5 f2f3 e8g8 g2g4 b5b4 c3a4 f6d7 d3e2 a6a5 h2h4 a8c8 d1d2 c7c6 h1d1 c8c7 b3e6 f7e6 b2b3 f8c8 g4g5 h6h5 a4b2 c6c3 e3f2 g7g6 b2c4 c7c4 b3c4');
INSERT INTO evaluations VALUES(21321,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(21321,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(21321,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(21321,33,'cp 88',NULL);
INSERT INTO evaluations VALUES(21321,34,'cp 82',NULL);
INSERT INTO evaluations VALUES(21321,35,'cp 89',NULL);
INSERT INTO evaluations VALUES(21321,36,'cp 89','d8e7 b1c3 c7c6 d5d3 g8f6 e1g1 c8e6 c4b3 g7g5 c1e3 f8g7 e3d4 e8g8 a1e1 e6b3 a2b3 f6h5 d4g7 h5g7 f2f4 a8e8 f4f5 e7f6 e1d1 e8d8 d3d4 f6d4 d1d4 f7f6 f1d1 g7e8 g1f2 h6h5 f2f3 g8f7 b3b4 d8d7 c3e2 f8h8 g2g3 d7e7 c2c4 a7a6 e2c3');
INSERT INTO evaluations VALUES(21423,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(21423,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(21423,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(21423,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(21423,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(21423,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(21423,36,'cp 45','c2c4 g8f6 e2e3 f8d6 c4c5 d6e7 a2a3 a7a5 b1c3 e8g8 f1d3 b7b6 c5b6 c7b6 d1e2 e7d6 c1d2 c8d7 a1c1 d8e7 c3b5 d6b8 e1g1 f6e4 c1c2 a5a4 f1c1 e4d2 e2d2 c6a5 d2c3 a5b3 c1d1 g8h8 b5c7');
INSERT INTO evaluations VALUES(22009,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(22009,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(22009,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(22009,33,'cp -15',NULL);
INSERT INTO evaluations VALUES(22009,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(22009,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(22009,36,'cp -17','g8f6 d1b3 d8d7 b1c3 c6a5 b3b5 a5c4 b5c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5d7 f8d8 d7f6 e7f6 g3c7 d8d1 a1d1 c6c5 c7d6 a8c8 d6g3 f6d4 c3b5 b7a6 b5c7 a6e2 d1e1 e2g4 c7d5 g8g7 g3e5 d4e5 e1e5 g4e6 f2f3 c8d8 d5c3 d8d2 b2b3 c5c4 b3c4 e6c4 e5c5');
INSERT INTO evaluations VALUES(15621,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(15621,31,'cp -30',NULL);
INSERT INTO evaluations VALUES(15621,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(15621,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(15621,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(15621,35,'cp -20',NULL);
INSERT INTO evaluations VALUES(15621,36,'cp -18','d1b3 d8d7 b1c3 c6a5 b3b4 a5c4 b4c4 f8e7 e4e5 d6e5 f1d1 d7c6 c4c6 b7c6 f3e5 c8b7 c1g5 h7h6 g5h4 g7g5 h4g3 e8g8 e5d7 f8d8 d7f6 e7f6 g3c7 d8d1 a1d1 c6c5 c7d6 f6d4 c3b5 d4b2 d6c5 a7a5 b5d6 b7a6 d1e1 a6d3 e1e7 d3g6 e7b7 b2g7 g2g4 g7f8 b7b5 a5a4 d6f5 g6f5 g4f5 f8c5 b5c5 g8g7 g1g2 a8d8');
INSERT INTO evaluations VALUES(24814,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(24814,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(24814,32,'cp -2',NULL);
INSERT INTO evaluations VALUES(24814,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24814,34,'cp -2',NULL);
INSERT INTO evaluations VALUES(24814,35,'cp -2',NULL);
INSERT INTO evaluations VALUES(24814,36,'cp 0','e7e6 f1d3 f5d3 d1d3 f8d6 g1e2 d6f4 e2f4 e6e5 f4e2 a7a5 e1g1 e8g8 b1d2 d8e7 f1c1 a5a4 b2b3 a4b3 a2b3 e5d4 e3d4 f8e8 e2f4 h7h6 d3c2 e7d6 f4d3 a8a1 c1a1 c6d8 b3b4 d8e6 d2f3 f6e4 f3d2');
INSERT INTO evaluations VALUES(25150,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25150,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25150,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25150,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25150,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25150,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25150,36,'cp 0','f1d3 f5d3 d1d3 f8d6 g1e2 d6f4 e2f4 e6e5 f4e2 e8g8 e1g1 a7a5 b1d2 d8d6 f1c1 f8e8 d3c2 h7h6 b2b3 d6a3 d2b1 a3e7 b1d2');
INSERT INTO evaluations VALUES(27238,30,'cp 72',NULL);
INSERT INTO evaluations VALUES(27238,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(27238,32,'cp 68',NULL);
INSERT INTO evaluations VALUES(27238,33,'cp 72',NULL);
INSERT INTO evaluations VALUES(27238,34,'cp 74',NULL);
INSERT INTO evaluations VALUES(27238,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(27238,36,'cp 69','g1f3 g7g6 d2d4 e5d4 f3d4 f8g7 d4c6 d7c6 c3d5 g8f6 e1g1 a7a5 e2e4 a5a4 a1b1 h7h6 d1d4 c6d5 c4d5 e8g8 c1d2 f6g4 d4d3 h6h5 d2c3 f8e8 b1c1 g7c3 d3c3 a8c8 f2f3 g4e5 f3f4 e5g4 h2h3 g4f6 c3d4 c7c6 d5c6 c8c6');
INSERT INTO evaluations VALUES(30641,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30641,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30641,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(30641,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30641,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30641,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30641,36,'cp 12','f8b4 d1b3 b4c3 b3c3 g8f6 e2e3 e8g8 a1d1 h7h6 f2f3 d7d5 f1d3 c7c5 g1e2 c5d4 e2d4 f6h5 c4d5 h5f4 e3f4 b7d5 e1g1 b8d7 f4f5 e6e5 d4b5 a7a6 b5d6 d8e7 c3c7 e7d8 c7d8 a8d8 d3a6 d7f6 d6c4 d8a8 c4b6 a8a6 b6d5 f6d5 d1d5 a6a2');
INSERT INTO evaluations VALUES(32944,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(32944,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(32944,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(32944,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(32944,34,'cp 1',NULL);
INSERT INTO evaluations VALUES(32944,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(32944,36,'cp 15','h2h3 g4h5 g2g4 h5g6 f3e5 g6e4 f1b5 b8c6 b5c6 e4c6 e5c6 b7c6 h1e1 g8e7 c2c3 h7h5 g4g5 e8c8 d1c2 e7d5 b1d2 f8d6 d2c4 h8e8 c4d6 c7d6 c1d2 c6c5 a1d1 c8d7 b2b4 e8e1 d1e1 d8c8 c2b3 g7g6 a2a3 c5c4 b3c2 d7c6 e1e4 c6b5');
INSERT INTO evaluations VALUES(34774,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(34774,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(34774,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(34774,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(34774,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(34774,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(34774,36,'cp 55','f6g6 b1c3 b8c6 d2d3 f8e7 c1e3 c6a5 d1e2 a5c4 d3c4 g8f6 f3d2 c8e6 e1c1 f6d7 c3d5 e6d5 c4d5 e7g5 e3g5 g6g5 c1b1 g7g6 f2f3 f7f5 a2a4 h7h5 h3h4 g5e7 e4f5 g6f5 g2g4 h5g4 f3g4 f5g4 d2e4 e8c8 e2g4 c8b8');
INSERT INTO evaluations VALUES(35807,30,'cp 115',NULL);
INSERT INTO evaluations VALUES(35807,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(35807,32,'cp 115',NULL);
INSERT INTO evaluations VALUES(35807,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(35807,34,'cp 121',NULL);
INSERT INTO evaluations VALUES(35807,35,'cp 105',NULL);
INSERT INTO evaluations VALUES(35807,36,'cp 112','h2h3 b8d7 f1d3 d7e5 f3e5 d6e5 e1g1 f6e8 d1e2 e8d6 f2f4 e7f6 c1e3 b7b6 a2a4 c8d7 e3d2 a8c8 d3b5 c8c7 g1h1 f8e8 f4f5 f6g5 b5d7 c7d7 d2g5 d8g5 c3b5 g5h4 b5d6 d7d6 h1h2 a7a5 a1a3 h4f6 a3g3 e8d8 h2g1 d8e8 f1f3');
INSERT INTO evaluations VALUES(35865,30,'cp 20',NULL);
INSERT INTO evaluations VALUES(35865,31,'cp 11',NULL);
INSERT INTO evaluations VALUES(35865,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(35865,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(35865,34,'cp 2',NULL);
INSERT INTO evaluations VALUES(35865,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(35865,36,'cp 5','c6c5 d4c5 f6e4 a1c1 b8d7 c4d5 d8a5 f1c4 e4c3 b2c3 g7c3 e1f1 d7c5 f1g1 c3b2 d5d6 c8e6 d6e7 f8e8 c1c2 b2f6 f3g5 a5b6 g1h2 e8e7 g5e6 c5e6 f4g3 a8d8 d1b1 b6b1 h1b1 b7b6 a2a4 h7h5 h3h4 e6c5 a4a5 c5e4 a5b6 a7b6 b1b6 g8g7 c4b5 e4g3');
INSERT INTO evaluations VALUES(36476,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(36476,31,'cp 23',NULL);
INSERT INTO evaluations VALUES(36476,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36476,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(36476,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(36476,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(36476,36,'cp 21','c7c5 d4d5 a7a6 g1f3 b7b5 f1d3 b5c4 d3c4 c8g4 e1g1 g4f3 g2f3 b8d7 g1h1 d8a5 a1b1 a5b4 c4e2 f8c8 f1c1 b4d4 g5e3 d4d2 e3d2 c8b8 b2b3 f6e8 h1g2 g7d4 c3a4 e8f6 a4b2 d7e5 d2e3 d4e3');
INSERT INTO evaluations VALUES(37556,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(37556,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(37556,32,'cp 109',NULL);
INSERT INTO evaluations VALUES(37556,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(37556,34,'cp 128',NULL);
INSERT INTO evaluations VALUES(37556,35,'cp 119',NULL);
INSERT INTO evaluations VALUES(37556,36,'cp 138','d7d6 c1g5 g8e7 d1d2 b8d7 f1d3 a7a6 g5h6 g7h6 d2h6 e7g8 h6e3 c8b7 e1g1 g8f6 f1e1 e8g8 a1d1 d8e7 h2h3 f6h5 a2a3 a8e8 b2b4 f7f5 d3c2 e7g7 c4c5 b6c5 b4c5 b7c6 c5d6 c7d6 c2b3 g8h8 e4e5 c6f3 e3f3');
INSERT INTO evaluations VALUES(37551,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(37551,31,'cp 135',NULL);
INSERT INTO evaluations VALUES(37551,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(37551,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(37551,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(37551,35,'cp 130',NULL);
INSERT INTO evaluations VALUES(37551,36,'cp 129','c1g5 g8e7 d1d2 b8d7 f1d3 a7a6 g5h6 g7h6 d2h6 e7g8 h6e3 c8b7 e1g1 g8f6 a1d1 e8g8 f1e1 d8e7 d3f1 a8e8 a2a3 e6e5 e3d2 c7c5 d4e5 d6e5 b2b3 g8g7 h2h3 h7h6 d2e3 e8b8 c3d5 b7d5 c4d5 b6b5 a3a4 b5a4 b3a4 b8b4');
INSERT INTO evaluations VALUES(39461,30,'cp 183',NULL);
INSERT INTO evaluations VALUES(39461,31,'cp 171',NULL);
INSERT INTO evaluations VALUES(39461,32,'cp 175',NULL);
INSERT INTO evaluations VALUES(39461,33,'cp 155',NULL);
INSERT INTO evaluations VALUES(39461,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(39461,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(39461,36,'cp 162','b2b4 c5d4 c2c3 d4b6 g1e2 g8e7 e2g3 c7c6 g3h5 d7d5 h5g7 e8f7 g7h5 d8d6 c4b3 h8g8 h2h3 g8g6 d2d4 c8e6 d4e5 d6e5 h5f4 f6f5 e1g1 f5e4 f4g6 f7g6 c1f4 e4f3 f4e5 b8d7 e5g3 g6f7 b1d2 a8g8 g1h2 f3g2 f1g1 a7a5 d2f3 a5b4 c3b4 e7f5 g1g2 f5g3');
INSERT INTO evaluations VALUES(42030,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(42030,31,'cp 105',NULL);
INSERT INTO evaluations VALUES(42030,32,'cp 120',NULL);
INSERT INTO evaluations VALUES(42030,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(42030,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(42030,35,'cp 91',NULL);
INSERT INTO evaluations VALUES(42030,36,'cp 99','e4e5 b4c3 b2c3 f6e4 d1g4 e8f8 g1f3 c7c5 h2h4 c5c4 f3g5 e4d2 e1d2 h7h6 g5h3 b7b5 f1e2 b8c6 a2a3 a7a5 h3f4 c6e7 g4f3 c8d7 g2g4 f8g8 f3e3 e7c6 g4g5 b5b4 g5g6 b4c3 d2c3 f7g6 f4g6 c6e7 h4h5 e7g6 h5g6 a5a4 c3d2 d8a5 e3c3 a5c7 f2f4 a8f8 c3e3');
INSERT INTO evaluations VALUES(42123,30,'cp 110',NULL);
INSERT INTO evaluations VALUES(42123,31,'cp 87',NULL);
INSERT INTO evaluations VALUES(42123,32,'cp 99',NULL);
INSERT INTO evaluations VALUES(42123,33,'cp 106',NULL);
INSERT INTO evaluations VALUES(42123,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(42123,35,'cp 106',NULL);
INSERT INTO evaluations VALUES(42123,36,'cp 106','b4c3 b2c3 f6e4 d1g4 e8f8 g1f3 c7c5 h2h4 c5c4 h1h3 b8c6 g4f4 f7f6 f3g5 e4d2 e1d2 h7h6 g5f3 f8g8 g2g4 f6e5 f3e5 g8h7 f1e2 h8f8 e5c6 b7c6 f4e3 c6c5 d4c5 e6e5 h3f3 f8f3 e3f3 c8e6 g4g5 h6g5 f3h5 h7g8 h4g5 d5d4 a1h1 d4c3 d2c1 d8d2 c1b1 e6d5');
INSERT INTO evaluations VALUES(43326,30,'cp 68',NULL);
INSERT INTO evaluations VALUES(43326,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(43326,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(43326,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(43326,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(43326,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(43326,36,'cp 73','e1g1 d7d6 c2c4 b8c6 c1e3 c6e5 b1c3 e5d3 d1d3 f8e7 f2f4 e8g8 h2h3 d8c7 g1h1 c8d7 b2b3 a8c8 a1c1 c7a5 e3d2 a5c5 a2a4 e7d8 d2e3 c5h5 e3g1 d8a5 g1h2 e6e5 f4e5 d6e5 d4f5 d7f5 f1f5 h5h6 c1f1 a5c3 d3c3');
INSERT INTO evaluations VALUES(43325,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(43325,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(43325,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(43325,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(43325,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(43325,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(43325,36,'cp 69','d7d6 c2c4 b8c6 c1e3 c6e5 h2h3 f8e7 f2f4 e5d3 d1d3 d8c7 b1c3 e8g8 a1c1 c8d7 b2b3 a8c8 g1h1 c7a5 a2a4 f8d8 f4f5 d8e8 e3d2 a5e5 d2f4 e5c5 f4g5 h7h6 g5e3 c5b4 f5e6 f7e6 d4f3 e8f8 e4e5 d6e5 f3e5 d7c6 e5c6 c8c6 f1d1');
INSERT INTO evaluations VALUES(43547,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(43547,31,'cp 71',NULL);
INSERT INTO evaluations VALUES(43547,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(43547,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(43547,34,'cp 71',NULL);
INSERT INTO evaluations VALUES(43547,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(43547,36,'cp 71','c2c4 b8c6 c1e3 c6e5 b1c3 e5d3 d1d3 f8e7 h2h3 e8g8 f2f4 d8c7 b2b3 c8d7 a2a4 a8c8 a1c1 c7a5 g1h1 f8d8 f4f5 d8e8 e3g5 a5b6 c1d1 e6f5 e4f5 h7h6 g5f6 e7f6 c3d5 b6d4 d3d4 f6d4 d1d4 b7b5 a4b5 a6b5 d5b6');
INSERT INTO evaluations VALUES(44580,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(44580,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44580,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44580,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44580,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44580,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44580,36,'cp 0','e7e6 d2d4 b7b6 g1f3 c8b7 c2c4 c5d4 f3d4 b8c6 b1c3 a8c8 d4c2 c6b4 c2d4');
INSERT INTO evaluations VALUES(44253,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44253,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44253,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(44253,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(44253,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(44253,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(44253,36,'cp 0','d2d4 c5d4 e3d4 f8e7 b1d2 b7b6 f1d3 e8g8 g1f3 a7a5 e1g1 d7d5 c2c4 c8b7 a1c1 b8d7 f1e1 a8c8 d2b1 e7b4 b1c3 d8e7 d1e2 f6e4 c3e4 b4e1 c1e1 d5e4 d3e4 b7e4 e2e4');
INSERT INTO evaluations VALUES(45445,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(45445,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(45445,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(45445,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(45445,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(45445,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(45445,36,'cp 0','b8c6 c1e3 a7a6 h2h3 e7e6 g1f3 c8d7 e1g1 f8d6 f1e1 d8c7 e3g5 e8g8 g5f6 g7f6 f3e5 c6e5 d4e5 d6e5 d3h7 g8h7 d1h5 h7g7 h5g4 g7h6 g4h4 h6g6 h4g4 g6h6');
INSERT INTO evaluations VALUES(46037,30,'cp -71',NULL);
INSERT INTO evaluations VALUES(46037,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(46037,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(46037,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(46037,34,'cp -104',NULL);
INSERT INTO evaluations VALUES(46037,35,'cp -73',NULL);
INSERT INTO evaluations VALUES(46037,36,'cp -76','d1d4 b8c6 d4e3 f8e7 e1g1 e8g8 c4d3 d6d5 g5f3 d5e4 d3e4 f6e4 e3e4 f7f6 b1c3 c8d7 f1d1 d8c8 e4e2 d7e8 f3d4 e8f7 d4c6 c8c6 c1e3 a7a6 a2a3 e6e5 f2f3 f7c4 e2e1 a8c8 d1d2 f8d8 a1d1 c4e6 d2d8 e7d8 e1e2 f6f5');
INSERT INTO evaluations VALUES(46036,30,'cp -82',NULL);
INSERT INTO evaluations VALUES(46036,31,'cp -96',NULL);
INSERT INTO evaluations VALUES(46036,32,'cp -81',NULL);
INSERT INTO evaluations VALUES(46036,33,'cp -91',NULL);
INSERT INTO evaluations VALUES(46036,34,'cp -87',NULL);
INSERT INTO evaluations VALUES(46036,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(46036,36,'cp -102','b8c6 d4e3 f8e7 e1g1 e8g8 c4d3 d6d5 g5f3 h7h6 b1d2 e6e5 e3e2 f8e8 c2c3 f6h5 d2b3 d5e4 d3e4 e7d6 c1e3 h5f6 b3d2 d8c7 a1d1 d6f8 f3h4 f6e4 d2e4 f7f5 e2h5 c7f7 h5f7 g8f7 e4d6 f8d6 d1d6 c8e6 h4f3 f5f4 d6c6 b7c6 f3e5 f7g8 e3f4');
INSERT INTO evaluations VALUES(13977,30,'cp -78',NULL);
INSERT INTO evaluations VALUES(13977,31,'cp -74',NULL);
INSERT INTO evaluations VALUES(13977,32,'cp -75',NULL);
INSERT INTO evaluations VALUES(13977,33,'cp -83',NULL);
INSERT INTO evaluations VALUES(13977,34,'cp -80',NULL);
INSERT INTO evaluations VALUES(13977,35,'cp -92',NULL);
INSERT INTO evaluations VALUES(13977,36,'cp -90','d4e3 f8e7 e1g1 e8g8 c4d3 d6d5 g5f3 d5e4 d3e4 f6e4 e3e4 d8a5 b1c3 e7f6 c1d2 a5h5 f1e1 f8d8 h2h4 e6e5 d2g5 c8g4 f3h2 g4e6 f2f4 c6d4 g5f6 g7f6 f4e5 f6f5 e4d3 h5h4 e1d1 h4g5 h2f3 d4f3 d3f3 a8c8 f3f2 f5f4 d1d8 c8d8');
INSERT INTO evaluations VALUES(46237,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(46237,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(46237,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(46237,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(46237,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(46237,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(46237,36,'cp 0','f1g2 b8c6 e1g1 d8b6 d4c5 f8c5 b2b4 c5e7 a2a4 e8g8 c1f4 f6h5 a4a5 b6d8 f4c1 h5f6 c1f4');
INSERT INTO evaluations VALUES(49760,30,'cp -105',NULL);
INSERT INTO evaluations VALUES(49760,31,'cp -110',NULL);
INSERT INTO evaluations VALUES(49760,32,'cp -116',NULL);
INSERT INTO evaluations VALUES(49760,33,'cp -102',NULL);
INSERT INTO evaluations VALUES(49760,34,'cp -108',NULL);
INSERT INTO evaluations VALUES(49760,35,'cp -105',NULL);
INSERT INTO evaluations VALUES(49760,36,'cp -102','f6e4 b1c3 f8b4 d1b3 d8e7 c1e3 c7c6 f1d3 d7d5 g1e2 e4f6 e1g1 e8g8 a1e1 b4d6 b3c2 e7d8 e2g3 c8e6 c3a4 d8c7 a4c5 b8d7 e3d2 a8e8 f2f4 c7c8 c5e6 e8e6 e1e6 f7e6 g1h1 c6c5 d4c5');
INSERT INTO evaluations VALUES(49919,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(49919,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(49919,32,'cp -117',NULL);
INSERT INTO evaluations VALUES(49919,33,'cp -94',NULL);
INSERT INTO evaluations VALUES(49919,34,'cp -111',NULL);
INSERT INTO evaluations VALUES(49919,35,'cp -93',NULL);
INSERT INTO evaluations VALUES(49919,36,'cp -104','b1c3 f8b4 d1b3 d8e7 c1e3 d7d5 f1d3 c7c6 g1f3 c8e6 e1g1 e8g8 a1e1 b8d7 a2a3 b4c3 b2c3 b7b5 a3a4 a7a5 e1a1 f8b8 e3f4 b8c8 f1c1 e7f6 f4g3 e6f5 d3f1 b5a4 b3a4 h7h6 g3h4 g7g5 h4g3 f5g4 f3e5 d7e5 d4e5 f6g6 a4c2');
INSERT INTO evaluations VALUES(49778,30,'cp 147',NULL);
INSERT INTO evaluations VALUES(49778,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(49778,32,'cp 145',NULL);
INSERT INTO evaluations VALUES(49778,33,'cp 131',NULL);
INSERT INTO evaluations VALUES(49778,34,'cp 151',NULL);
INSERT INTO evaluations VALUES(49778,35,'cp 114',NULL);
INSERT INTO evaluations VALUES(49778,36,'cp 140','e4e5 f6g8 g1f3 b8c6 f1c4 d7d5 c4d5 g8e7 d5f7 e8f7 d1b3 f7e8 e1g1 c8g4 f3g5 g4h5 c1e3 h7h6 g5e6 h5f7 e6d8 f7b3 d8c6 e7c6 a2b3 c6e5 c3b5 f8d6 a1d1 e8f7 b5d6 c7d6 d1d6 h8d8 f1d1 d8d6 d1d6 a8c8 h2h3 f7e7 d6d1 e5c6 e3c5 e7f6 f2f3 b7b6 c5a3 c6e7 g1h2');
INSERT INTO evaluations VALUES(50467,30,'cp 124',NULL);
INSERT INTO evaluations VALUES(50467,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(50467,32,'cp 122',NULL);
INSERT INTO evaluations VALUES(50467,33,'cp 120',NULL);
INSERT INTO evaluations VALUES(50467,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(50467,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(50467,36,'cp 132','d7d6 f1d3 d8c7 a2a4 g8f6 a4a5 e6e5 c1e3 c8g4 h2h3 g4f3 d1f3 b8d7 g2g4 h7h6 h3h4 e5d4 e3d4 d7e5 f3g3 e5d3 c2d3 f8e7 e1g1 c7d7 f2f3 a8d8 g1g2 e8g8 d4f6 e7f6 g4g5 f6c3 b2c3 f7f6 f3f4 f6g5 h4g5');
INSERT INTO evaluations VALUES(50618,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(50618,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(50618,32,'cp -7',NULL);
INSERT INTO evaluations VALUES(50618,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(50618,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(50618,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(50618,36,'cp -9','b8c6 c2c3 g8f6 e4e5 f6d5 f1c4 e7e6 e1g1 d7d6 d2d4 c5d4 c3d4 f8e7 h2h3 d6e5 f3e5 e8g8 e5c6 b7c6 a4a5 c8b7 b1c3 c6c5 d4c5 d5c3 d1d8 f8d8 b2c3 e7c5 c1f4 b7d5 c4e2 f7f6 f1d1 e6e5 c3c4 d5e4 f4e3 d8d1 a1d1 c5e3 f2e3');
INSERT INTO evaluations VALUES(17512,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17512,31,'cp -3',NULL);
INSERT INTO evaluations VALUES(17512,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17512,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17512,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(17512,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17512,36,'cp 0','b1c3 g8f6 f1e2 e7e6 d2d4 d7d5 d4c5 f8c5 c1g5 e8g8 e4d5 e6d5 g5f6 d8f6 c3d5 f6b2 a1b1 b2a3 b1a1 c5b4 e1f1 a3b2 a1b1 b2a3');
INSERT INTO evaluations VALUES(17511,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(17511,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(17511,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(17511,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(17511,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(17511,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(17511,36,'cp 0','g8f6 f1e2 e7e6 d2d4 d7d5 d4c5 f8c5 c1g5 e8g8 e4d5 e6d5 g5f6 d8f6 c3d5 f6b2 a1b1 b2a3 b1a1 a3b2');
INSERT INTO evaluations VALUES(50919,30,'cp 152',NULL);
INSERT INTO evaluations VALUES(50919,31,'cp 151',NULL);
INSERT INTO evaluations VALUES(50919,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(50919,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(50919,34,'cp 160',NULL);
INSERT INTO evaluations VALUES(50919,35,'cp 174',NULL);
INSERT INTO evaluations VALUES(50919,36,'cp 174','d4e5 b8c6 c1g5 f8e7 g5f4 c8g4 e5d6 e7d6 f4d6 c7d6 b1c3 g8f6 d1d2 e8g8 f3d4 d8b6 d4c6 b7c6 f2f3 g4e6 e1c1 f8d8 c3a4 b6c7 f1d3 d6d5 a4c5 d5e4 c5e6 f7e6 f3e4 a8b8 d2e2 c7e5 c2c3 a6a5 g2g3 f6d7 d3c2 b8b7 d1d2 d8b8 h1d1 b7b2 d2d7');
INSERT INTO evaluations VALUES(50908,30,'cp 146',NULL);
INSERT INTO evaluations VALUES(50908,31,'cp 145',NULL);
INSERT INTO evaluations VALUES(50908,32,'cp 124',NULL);
INSERT INTO evaluations VALUES(50908,33,'cp 134',NULL);
INSERT INTO evaluations VALUES(50908,34,'cp 129',NULL);
INSERT INTO evaluations VALUES(50908,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(50908,36,'cp 155','b8c6 c1g5 f8e7 g5f4 c8g4 e5d6 e7d6 f4d6 c7d6 b1c3 g8f6 d1d2 e8g8 f3d4 d8b6 d4c6 b7c6 f2f3 g4e6 c3a4 b6c7 e1c1 d6d5 e4d5 c6d5 d2d4 f6d7 f1d3 d7e5 a4c5 a8c8 c5e6 f7e6 c1b1 e5d3 d4d3 c8b8 d3a6 b8b6 a6a3 f8c8 d1d2 c7f4 h1d1 f4h2 a3a7 c8b8 b2b3 b6c6 b1a1 h2e5 a1b1');
INSERT INTO evaluations VALUES(51820,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(51820,31,'cp -68',NULL);
INSERT INTO evaluations VALUES(51820,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(51820,33,'cp -48',NULL);
INSERT INTO evaluations VALUES(51820,34,'cp -58',NULL);
INSERT INTO evaluations VALUES(51820,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(51820,36,'cp -41','d7d5 e4e5 d5d4 a4b3 g8e7 g1f3 e7g6 d2d3 b8d7 f3d4 d7e5 d4f3 e5f3 d1f3 f8e7 e1g1 c6c5 b3a4 c8d7 a4d7 d8d7 a2a4 g6h4 f3g4 h4f5 b1a3 e8g8 a3c4 f5d4 g4d1 b7b6 c1d2 d4c6 d2c3 a8d8 d1f3 d7c7');
INSERT INTO evaluations VALUES(51912,30,'cp -50',NULL);
INSERT INTO evaluations VALUES(51912,31,'cp -40',NULL);
INSERT INTO evaluations VALUES(51912,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(51912,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(51912,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(51912,35,'cp -54',NULL);
INSERT INTO evaluations VALUES(51912,36,'cp -34','g7g6 h5e2 f8g7 g2g3 g8e7 g1f3 b8c6 f1g2 e8g8 e1g1 e6e5 c2c4 a7a6 b1c3 d7d6 h2h3 a8b8 a2a4 c8d7 g1h2 c6b4 c1d2 f7f5 e4f5 g6f5 d2g5 d8e8 a4a5 d7c6 g5e7 e8e7 f3h4 g8h8 f2f4 c6g2 h4g2 e7d7 a1e1 b7b5');
INSERT INTO evaluations VALUES(53502,30,'cp 193',NULL);
INSERT INTO evaluations VALUES(53502,31,'cp 170',NULL);
INSERT INTO evaluations VALUES(53502,32,'cp 171',NULL);
INSERT INTO evaluations VALUES(53502,33,'cp 159',NULL);
INSERT INTO evaluations VALUES(53502,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(53502,35,'cp 171',NULL);
INSERT INTO evaluations VALUES(53502,36,'cp 209','d1d4 d8b6 d4d3 b8d7 c1e3 b6c7 e1c1 g8f6 e3f4 f6g4 f4d6 f8d6 d3d6 c7d6 d1d6 e8e7 d6d4 g7g5 e4e5 g4e5 h1e1 f7f6 c4b3 h8e8 f3e5 d7e5 f2f4 g5f4 d4f4 b7b5 g2g3 e8d8 c3e4 f6f5 e4c5 e7f6 a2a4 b5a4 b3a4');
INSERT INTO evaluations VALUES(53501,30,'cp 196',NULL);
INSERT INTO evaluations VALUES(53501,31,'cp 194',NULL);
INSERT INTO evaluations VALUES(53501,32,'cp 204',NULL);
INSERT INTO evaluations VALUES(53501,33,'cp 173',NULL);
INSERT INTO evaluations VALUES(53501,34,'cp 183',NULL);
INSERT INTO evaluations VALUES(53501,35,'cp 177',NULL);
INSERT INTO evaluations VALUES(53501,36,'cp 168','d8b6 d4d3 b8d7 c1e3 b6b2 a1d1 b2a3 e1g1 f8e7 c4b3 d7c5 e3c5 a3c5 f1e1 h6h5 e4e5 d6d5 c3a4 c5a5 e5e6 c8e6 e1e6 f7e6 d3g6 e8d7 g6g7 h8h6 f3e5 d7c7 e5f7 h6f6 g7g3 e6e5 g3e5 c7c8 c2c4 a5c7 e5e2 c7f4 c4d5 f6f7 d5c6 f4f2 e2f2 f7f2 g1f2 g8f6 c6b7 c8b7 b3d5 f6d5 d1d5 b7c6 d5h5 e7d6');
INSERT INTO evaluations VALUES(56955,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(56955,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(56955,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(56955,33,'cp 34',NULL);
INSERT INTO evaluations VALUES(56955,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(56955,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(56955,36,'cp 31','g8f6 c2c4 d7d5 b1c3 f8d6 f4d6 d8d6 a1c1 e8g8 c4d5 f6d5 f1d3 b8d7 g1e2 d5c3 c1c3 e6e5 e1g1 f8d8 d3b1 d7f6 d1c2 e5d4 e2d4 c7c6 a2a3 a7a5 h2h3 c8e6 c3d3 d8d7 d4e6 d6d3 c2d3 d7d3 b1d3');
INSERT INTO evaluations VALUES(49239,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(49239,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(49239,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(49239,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(49239,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(49239,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(49239,36,'cp 29','c2c4 d7d5 b1c3 f8d6 f4d6 d8d6 a1c1 e8g8 c4d5 f6d5 a2a3 d5c3 c1c3 c7c6 g1f3 b8d7 f1d3 b7b6 e1g1 c8b7 d3b1 c6c5 d4c5 d6d1 f1d1 d7c5 b2b4 c5e4 b1e4 b7e4 f3d4 f8c8 d1c1 c8c3 c1c3 e6e5 d4b5 a7a5 f2f3 e4d5 b5c7 a8d8 c7d5 d8d5');
INSERT INTO evaluations VALUES(57705,30,'cp 212',NULL);
INSERT INTO evaluations VALUES(57705,31,'cp 221',NULL);
INSERT INTO evaluations VALUES(57705,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(57705,33,'cp 215',NULL);
INSERT INTO evaluations VALUES(57705,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(57705,35,'cp 228',NULL);
INSERT INTO evaluations VALUES(57705,36,'cp 225','e7e6 d2d4 a7a6 h2h4 g5h4 d1h5 d8f6 g1f3 d7d6 c1g5 f6g7 h5h4 b8c6 e4e5 d6d5 c3e2 c8d7 e2g3 f8e7 g5e7 c6e7 g3h5 g7f8 e1c1 e8c8 a2a3 e7f5 h4f4 f5g7 h5f6 g8f6 e5f6 g7e8 g2g4 h7h6 f3e5 e8d6 h1h5 c8b8 f2f3 b8a8 f1d3 d7b5 d3b5 d6b5 c2c3 h8g8 f4h6');
INSERT INTO evaluations VALUES(1635,30,'cp -30',NULL);
INSERT INTO evaluations VALUES(1635,31,'cp -38',NULL);
INSERT INTO evaluations VALUES(1635,32,'cp -25',NULL);
INSERT INTO evaluations VALUES(1635,33,'cp -32',NULL);
INSERT INTO evaluations VALUES(1635,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(1635,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(1635,36,'cp -21','c8d7 g1f3 e8c8 d1e1 c6b4 f1d3 f7f6 c1e3 e7e6 a1c1 b4d3 c2d3 b7b6 h2h4 h7h5 d3d4 g8h6 e1e2 c8b7 f3d2 h6f7 f2f3 a7a5 a2a4 f8e7 d2c4 d7c6 c3b5 d8d7 b2b3 f6f5 e4e5 g7g5 h4g5 f7g5 e3f2');
INSERT INTO evaluations VALUES(2478,30,'cp 660',NULL);
INSERT INTO evaluations VALUES(2478,31,'cp 648',NULL);
INSERT INTO evaluations VALUES(2478,32,'cp 659',NULL);
INSERT INTO evaluations VALUES(2478,33,'cp 683',NULL);
INSERT INTO evaluations VALUES(2478,34,'cp 690',NULL);
INSERT INTO evaluations VALUES(2478,35,'cp 698',NULL);
INSERT INTO evaluations VALUES(2478,36,'cp 694','f6g8 b1c3 d8e8 e1g1 e7d8 g5h7 f8c5 d1g4 b7b6 d2d3 c6d4 c1g5 g8e7 b2b4 c5b4 c3d5 b4c5 g5e7 c5e7 g4g7 c7c6 d5f6 e7f6 h7f6 e8e7 g7g5 d4e6 g5e5 d7d6 e5c3 e6f4 a1e1 c8h3 g2g3 a8c8 e4e5 d6e5 c3e5 e7e5 e1e5 f4g6 e5h5 h3f1 g1f1');
INSERT INTO evaluations VALUES(2488,30,'cp 706',NULL);
INSERT INTO evaluations VALUES(2488,31,'cp 710',NULL);
INSERT INTO evaluations VALUES(2488,32,'cp 683',NULL);
INSERT INTO evaluations VALUES(2488,33,'cp 690',NULL);
INSERT INTO evaluations VALUES(2488,34,'cp 703',NULL);
INSERT INTO evaluations VALUES(2488,35,'cp 721',NULL);
INSERT INTO evaluations VALUES(2488,36,'cp 706','b1c3 d8e8 e1g1 e7d8 g5h7 f8c5 d1g4 g8f6 g4g7 f6h7 g7h7 c6d4 c3d5 c7c6 h7h4 c5e7 d5e7 e8e7 h4h5 e7e8 h5d1 b7b6 c2c3 d4e6 f2f4 e5f4 d2d4 c8a6 c1f4 d8c8 f4e3 a6f1 d1f1 e6d4 e3d4 e8e7 f1f6 e7f6 d4f6 c8c7 e4e5 a8f8');
INSERT INTO evaluations VALUES(3120,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(3120,31,'cp -28',NULL);
INSERT INTO evaluations VALUES(3120,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(3120,33,'cp -28',NULL);
INSERT INTO evaluations VALUES(3120,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(3120,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(3120,36,'cp -32','a2a4 b7b6 h2h3 c8b7 f1e1 h7h6 g5f4 a7a6 c4a2 f6d7 c3e2 c6e5 f3e5 d6e5 f4g3 d8c7 e2c3 e7g5 d1e2 d7b8 c3b1 g5f4 b1d2 b8c6 c2c3 f8d8 d2f3 d8d7 g3h4 g7g5 h4g3 a8d8 a1d1 c6e7 a2c4 e7g6');
INSERT INTO evaluations VALUES(3775,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(3775,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(3775,32,'cp -22',NULL);
INSERT INTO evaluations VALUES(3775,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(3775,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(3775,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(3775,36,'cp -23','h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 c6e7 d5e3 c5b6 b5c4 c7c6 d3d4 d6d5 c4b3 e5d4 f3d4 d5e4 d1c2 b6d4 a1d1 e7f5 c2e4 d8e8 e4e8 f8e8 c3d4 g8f8 d1d2 f5d6 f1c1 c8d7 b3a4 b7b5 a4c2 a7a5 c2d3 a8d8 d4d5 c6d5 e3d5 d6c4 d3c4 b5c4 d5b6 d7b5 d2d8 e8d8 f2f3');
INSERT INTO evaluations VALUES(4981,30,'cp 2',NULL);
INSERT INTO evaluations VALUES(4981,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4981,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(4981,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(4981,34,'cp 5',NULL);
INSERT INTO evaluations VALUES(4981,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(4981,36,'cp -5','d7d6 d2d4 e5d4 f3d4 g8e7 c1g5 f7f6 g5e3 c8d7 d1h5 e7g6 b1c3 d8b8 c3e2 e8g8 b2b3 a7a5 e2g3 a5a4 d4f5 b8b5 a1e1 c5e3 e1e3 b5e5 h5d1 c6c5 h2h3 g6f4 d1g4 g7g6 f5h6 g8g7 g4d7 g7h6 c2c3 a8e8 f1e1 a4b3 a2b3');
INSERT INTO evaluations VALUES(5112,30,'cp -38',NULL);
INSERT INTO evaluations VALUES(5112,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(5112,32,'cp -44',NULL);
INSERT INTO evaluations VALUES(5112,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(5112,34,'cp -40',NULL);
INSERT INTO evaluations VALUES(5112,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(5112,36,'cp -37','g8e7 e1g1 h7h6 c4a2 e8g8 f1e1 b7b6 c1f4 d7d6 c3b5 e6e5 f4d2 g8h7 b5c3 a8b8 c3d5 e7d5 a2d5 c6e7 d5a2 c8d7 b2b4 d7a4 a1b1 d8d7 a2b3 a4b3 b1b3 b8d8 h2h3 f7f5 b3b1 e7c6 c2c4 f5e4 d3e4 c6d4 f3h2 f8f7 a3a4');
INSERT INTO evaluations VALUES(7965,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(7965,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(7965,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(7965,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(7965,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(7965,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(7965,36,'cp -18','c4a2 h7h6 e1g1 e8g8 a1b1 b7b6 c1d2 c8b7 b2b4 d7d6 c3e2 d8d7 d2c3 e6e5 b4c5 d6c5 a2c4 a8d8 e2g3 b7c8 d1e2 d7c7 f1e1 c7d6 a3a4 d6c7 h2h3 c8d7 b1a1 g8h7 h3h4 c6d4 f3d4 c5d4 c3d2');
INSERT INTO evaluations VALUES(6191,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(6191,31,'cp 132',NULL);
INSERT INTO evaluations VALUES(6191,32,'cp 131',NULL);
INSERT INTO evaluations VALUES(6191,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(6191,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(6191,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(6191,36,'cp 106','c6e7 a2a4 b6a5 b1c3 d7d6 a1b1 a7a6 b2b4 a5b6 a4a5 b6a7 e1g1 g8f6 h2h3 e8g8 f1e1 f8e8 c1e3 a7e3 e1e3 c7c6 d5c6 e7c6 e3d3 f6e4 d1b3 e4c3 c4f7 g8h8 d3c3 e8f8 b1d1 d8f6 f7d5 c8d7 c3d3 a8d8 d3d2 d7e8 b4b5 c6e5 d5b7 e8b5 f3e5 d6e5 b3e3 d8d2 d1d2 b5a4');
INSERT INTO evaluations VALUES(6431,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(6431,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(6431,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(6431,33,'cp 18',NULL);
INSERT INTO evaluations VALUES(6431,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(6431,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(6431,36,'cp 0','g8f6 g1f3 a7a5 e1g1 d7d6 f1e1 e8g8 b1d2 c8e6 b2b3 f6g4 e1e2 c5b6 a2a4 g4f6 e2e1 f6g4');
INSERT INTO evaluations VALUES(6701,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(6701,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(6701,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(6701,33,'cp 9',NULL);
INSERT INTO evaluations VALUES(6701,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(6701,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(6701,36,'cp 0','b2c3 d7d6 f3g5 c6e5 c4b3 d8e7 f2f4 h7h6 f4e5 h6g5 d1d5 d6e5 b3a4 c8d7 d5b7 a8d8 b7a7 d7a4 a7a4 e7d7 a4d7 d8d7 c1g5 f7f6 g5e3 h8h4 a2a4 h4e4 e1f2 g8h6 e3h6 g7h6 a4a5 d7d2 f2g3 d2d3 g3f2');
INSERT INTO evaluations VALUES(7144,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(7144,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(7144,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(7144,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(7144,34,'cp 18',NULL);
INSERT INTO evaluations VALUES(7144,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(7144,36,'cp 8','a2a4 e8g8 b2b4 c5a7 f1e1 c6e7 b1d2 e7g6 d3d4 f8e8 c4f1 d6d5 d4e5 g6e5 e4d5 d8d5 d2e4 e5f3 g2f3 d5d1 e4f6 g7f6 e1d1 g8h7 c1f4 c7c6 f1c4 e8g8 g1h2 c8e6 c4e6 f7e6 d1d7 g8g7 a1d1 b7b5 a4b5 c6b5 f4e3 a7e3 f2e3 a6a5 b4a5 a8a5');
INSERT INTO evaluations VALUES(7153,30,'cp 19',NULL);
INSERT INTO evaluations VALUES(7153,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(7153,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(7153,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(7153,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(7153,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(7153,36,'cp 5','e8g8 f1e1 c5a7 b2b4 c6e7 b1d2 e7g6 d3d4 f8e8 b4b5 c8d7 c4d3 a6b5 a4b5 d7e6 a1a3 c7c6 b5c6 b7c6 d2f1 a7b6 a3a8 d8a8 d1c2 a8a2 c2a2 e6a2 f1g3 d6d5 e4d5 e5d4 e1e8 f6e8 d5c6 d4c3 g3f5 e8c7 c1e3 b6e3 f5e3');
INSERT INTO evaluations VALUES(2597,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(2597,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(2597,32,'cp 15',NULL);
INSERT INTO evaluations VALUES(2597,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(2597,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(2597,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(2597,36,'cp 25','b2b4 c5a7 f1e1 c6e7 b1d2 e7g6 d3d4 f8e8 c4f1 d8e7 f1d3 c7c6 d3c2 e7c7 d2f1 a6a5 b4a5 c7a5 c1d2 a5c7 f1g3 c8d7 c2d3 c6c5 d4e5 d6e5 d3c4 a7b6 d1c2 d7e6 c4e6 e8e6 c3c4 b6a5 d2a5 a8a5');
INSERT INTO evaluations VALUES(9292,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(9292,31,'cp 30',NULL);
INSERT INTO evaluations VALUES(9292,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(9292,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(9292,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(9292,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(9292,36,'cp 30','d3d4 e5d4 c3d4 c5b4 c1d2 e8g8 b5c6 b4d2 b1d2 b7c6 e1g1 f6h5 g2g3 h5f6 a1c1 a7a5 d1c2 c8d7 b2b3 a5a4 b3a4 h7h6 f1e1 c6c5 d4c5 d7a4 d2b3 a4d7 b3d4 d6c5 c2c5 a8a2 f3e5 d8a8 e5d7 f6d7 c5f5 c7c5 d4b3 a2a3 f5d7');
INSERT INTO evaluations VALUES(11842,30,'cp 94',NULL);
INSERT INTO evaluations VALUES(11842,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(11842,32,'cp 105',NULL);
INSERT INTO evaluations VALUES(11842,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(11842,34,'cp 96',NULL);
INSERT INTO evaluations VALUES(11842,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(11842,36,'cp 91','f3e5 d5c3 d2c3 d8d1 e1d1 c6c5 c3c4 c8e6 b2b3 e8c8 e5d3 e6g4 f2f3 g4f5 d1e2 f8d6 c1e3 a7a5 a1d1 h7h5 h2h4 a5a4 e2f2 d8d7 b3a4 d6e7 e3g5 f7f6 g5e3 h8d8 d1b1 f5d3 c2d3 d7d3 b1b3 d3d1 h1e1 d1e1 f2e1 e7d6 a4a5 c8d7 b3b5 d8e8 e1f2');
INSERT INTO evaluations VALUES(11841,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(11841,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(11841,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(11841,33,'cp 85',NULL);
INSERT INTO evaluations VALUES(11841,34,'cp 94',NULL);
INSERT INTO evaluations VALUES(11841,35,'cp 94',NULL);
INSERT INTO evaluations VALUES(11841,36,'cp 96','d5c3 d2c3 d8d1 e1d1 c6c5 b2b3 c8e6 c3c4 e8c8 e5d3 e6g4 f2f3 g4f5 d1e2 a7a5 c1e3 f8d6 e2d2 f7f6 a1e1 a5a4 d2c3 f5d3 c2d3 a4b3 a2b3 c8d7 d3d4 c5d4 e3d4 c7c5 d4f2 d6e5 e1e5 f6e5 f2c5 d7c6 c5f2 d8a8 f2g3 h8d8 g3e5 a8a2 e5d4 a2g2 b3b4 g2a2 b4b5 c6b7 h1e1 a2a3 c3b4 a3f3');
INSERT INTO evaluations VALUES(14183,30,'cp 30',NULL);
INSERT INTO evaluations VALUES(14183,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(14183,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(14183,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(14183,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(14183,35,'cp 33',NULL);
INSERT INTO evaluations VALUES(14183,36,'cp 22','f6d5 g1f3 c8g4 f1b5 a7a6 b5e2 d5c3 b2c3 e7e6 a1b1 a8b8 h2h3 g4h5 e1g1 f8d6 c1g5 d6e7 g5e7 c6e7 f1e1 e8g8 c3c4 h5f3 e2f3 c7c6 c4c5 b7b6 c5b6 b8b6 d1d3 d8c7 g2g3 f8b8 c2c4 b6b1 e1b1');
INSERT INTO evaluations VALUES(14414,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(14414,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(14414,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(14414,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(14414,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(14414,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(14414,36,'cp 46','d2d4 c8g4 e1g1 e7e6 c2c3 f8d6 b1d2 e8g8 f1e1 h7h6 h2h3 g4h5 c4d3 d5f4 d3f1 f4d5 d2e4 d5f6 e4g3 h5g6 d1e2 a7a5 a2a3 a5a4 f3e5 d6e5 d4e5 f6d7 f2f4 d7b6 e2f2 c6a5 c1e3 b6d5 e3d4 a5c6 f4f5');
INSERT INTO evaluations VALUES(14415,30,'cp 40',NULL);
INSERT INTO evaluations VALUES(14415,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(14415,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(14415,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(14415,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(14415,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(14415,36,'cp 38','c8g4 c2c3 e7e6 e1g1 f8d6 f1e1 e8g8 h2h3 g4h5 a2a3 f8e8 b1d2 h7h6 a3a4 a7a5 c4b5 e8f8 b5d3 c6e7 d2c4 d6f4 c1f4 d5f4 d3f1 c7c6 g2g3 f4d5 f1g2 d5b6 c4e5 d8c7 e5d3 a8d8 d1e2 b6d7 d3f4 h5f3 g2f3 e7d5');
INSERT INTO evaluations VALUES(25011,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(25011,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(25011,32,'cp 36',NULL);
INSERT INTO evaluations VALUES(25011,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(25011,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(25011,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(25011,36,'cp 57','c2c3 e7e6 e1g1 f8d6 f1e1 e8g8 h2h3 g4h5 b1d2 h7h6 a2a4 a7a5 d2e4 c6e7 c1d2 d6f4 d2f4 d5f4 e4g3 h5f3 d1f3 e7g6 c4f1 b7b6 a1d1 d8d5 g3e4 d5f5 f3g4 f5g4 h3g4 f8d8 g2g3 f4d5 f1d3 g6e7 d3c2 d5f6 e4f6 g7f6 c2e4 c7c6');
INSERT INTO evaluations VALUES(15726,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(15726,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(15726,32,'cp 60',NULL);
INSERT INTO evaluations VALUES(15726,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(15726,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(15726,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(15726,36,'cp 28','c7c6 b5e2 f8e7 c1f4 e8g8 a1d1 d8a5 b2b4 a5b6 d4b6 a7b6 f4d6 e7d6 d1d6 a8a3 e1d2 a3c3 d6f6 c3c2 d2c2 g7f6 e2d3 c8e6 a2a4 c6c5 b4c5 b6c5 h1a1 f8a8 c2c3 g8f8 a4a5 f8e7 f2f4 e7d6 d3b5 f6f5 e4e5 d6c7 g2g3 e6d5 c3c2 d5e6 c2d2 a8d8 d2c3 d8a8 a1a3 e6d5 c3d3 d5e6');
INSERT INTO evaluations VALUES(16122,30,'cp -19',NULL);
INSERT INTO evaluations VALUES(16122,31,'cp -36',NULL);
INSERT INTO evaluations VALUES(16122,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(16122,33,'cp -27',NULL);
INSERT INTO evaluations VALUES(16122,34,'cp -26',NULL);
INSERT INTO evaluations VALUES(16122,35,'cp -35',NULL);
INSERT INTO evaluations VALUES(16122,36,'cp -34','f8c5 e1g1 e8g8 f1e1 a7a6 c3d5 d7d6 c2c3 f6d5 c4d5 c6e7 d2d4 e7d5 e4d5 e5d4 f3d4 d8f6 c1e3 f8e8 d1d2 f6g6 f2f3 a6a5 d4c2 c8f5 e3c5 f5c2 e1e8 a8e8 c5e3 a5a4 b3a4 c2a4 c3c4 b7b6 h2h3 a4d7 g1h2 e8a8 e3d4 a8a3 d2b2 a3d3 d4c3');
INSERT INTO evaluations VALUES(9748,30,'cp -20',NULL);
INSERT INTO evaluations VALUES(9748,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(9748,32,'cp -28',NULL);
INSERT INTO evaluations VALUES(9748,33,'cp -23',NULL);
INSERT INTO evaluations VALUES(9748,34,'cp -14',NULL);
INSERT INTO evaluations VALUES(9748,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(9748,36,'cp -31','d2d3 e8g8 e1g1 d7d6 c3a4 c5b6 h2h3 c6a5 c1d2 a5c4 b3c4 f6d7 f1e1 g8h7 a4b6 a7b6 a2a4 d7c5 a4a5 b6a5 a1a5 a8a5 d2a5 b7b6 a5d2 c5e6 d2e3 c8b7 f3d2 d8f6 d2f1 f6g6 f1g3 f7f5 e4f5 f8f5 d1g4 g6g4 h3g4 f5f7');
INSERT INTO evaluations VALUES(16368,30,'cp -40',NULL);
INSERT INTO evaluations VALUES(16368,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(16368,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(16368,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(16368,34,'cp -37',NULL);
INSERT INTO evaluations VALUES(16368,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(16368,36,'cp -42','g1f3 h7h6 g5f6 d8f6 g2g3 f8c5 f1h3 c6d4 b1d2 d4e6 e1g1 e8g8 c2c3 a7a5 d2c4 d7d6 a2a4 b7b6 h3g2 e6g5 f3g5 h6g5 d1e2 c8e6 c4e3 g7g6 e3c2 e6b3 f1c1 g8g7 d3d4 b3c2 d4c5 c2b3 c5b6 c7b6 e2e3 g5g4 e3b6');
INSERT INTO evaluations VALUES(17249,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(17249,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(17249,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(17249,33,'cp 5',NULL);
INSERT INTO evaluations VALUES(17249,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(17249,35,'cp 11',NULL);
INSERT INTO evaluations VALUES(17249,36,'cp 10','g7g6 g1f3 f8g7 h2h3 d7d6 b1d2 e8g8 e2e4 e7e5 d4e5 c6e5 f3e5 d6e5 f4e3 b7b6 f1e2 d8e7 d1c2 f6e8 e1g1 e8d6 d2c4 d6b7 f2f3 c8e6 e3f2 a7a5 c4e3 f8d8 f1d1 h7h5 e2b5 d8d1 a1d1 a8d8 d1d8 e7d8 b5c4 e6c4');
INSERT INTO evaluations VALUES(18792,30,'cp 17',NULL);
INSERT INTO evaluations VALUES(18792,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(18792,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(18792,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(18792,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(18792,35,'cp 13',NULL);
INSERT INTO evaluations VALUES(18792,36,'cp 29','e2e3 f8e7 b1d2 g8f6 g1f3 f6h5 d4c5 h5f4 e3f4 d8c7 g2g3 e7c5 f1d3 e8g8 d1e2 a7a5 h2h4 d7d5 d2b3 c5e7 b3d4 a5a4 e1g1 a4a3 d4b5 c7b6 b2b4 c8d7 f1e1 f8d8 h4h5 e7f6 g1g2 h7h6 a1d1 c6a7 b5a7 b6a7');
INSERT INTO evaluations VALUES(21108,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(21108,31,'cp -49',NULL);
INSERT INTO evaluations VALUES(21108,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(21108,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(21108,34,'cp -50',NULL);
INSERT INTO evaluations VALUES(21108,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(21108,36,'cp -46','e7e5 e2e3 f8d6 a2a3 d8e7 b1c3 g8f6 d2d3 c8d7 c4b3 c6a5 b3c2 c7c5 c1d2 e8g8 f1e2 a5c6 e1g1 a8c8 c3e4 h7h6 e4d6 e7d6 b2b3 f8d8 d2c3 c6d4 e3d4 c5d4 c3b4 c8c2 b4d6 c2e2 d6e5 d7c6 e5d4 c6f3 d4f6 g7f6 g2f3 d8d3');
INSERT INTO evaluations VALUES(21213,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(21213,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(21213,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(21213,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(21213,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(21213,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(21213,36,'cp 0','e7e6 g1f3 f8d6 f3e5 d6e5 f1c4 e6d5 d1d5 d8d5 c3d5 c8e6 d5e3 e8c8 c4e6 f7e6 f2f3 g8e7 e3c4 e5d4 c1g5 d4f6 g5f6 g7f6 e1e2 c7c5 a1d1 e7c6 a2a4 d8d1 h1d1 c8c7 h2h4 c6d4 e2f2 c7c6 g2g4 b7b5 a4b5 c6b5 c4d6 b5c6 d6c4');
INSERT INTO evaluations VALUES(21988,30,'cp -172',NULL);
INSERT INTO evaluations VALUES(21988,31,'cp -179',NULL);
INSERT INTO evaluations VALUES(21988,32,'cp -167',NULL);
INSERT INTO evaluations VALUES(21988,33,'cp -183',NULL);
INSERT INTO evaluations VALUES(21988,34,'cp -183',NULL);
INSERT INTO evaluations VALUES(21988,35,'cp -168',NULL);
INSERT INTO evaluations VALUES(21988,36,'cp -178','e5f4 f1e2 g7g5 e1g1 h7h6 d2d4 g8f6 b1d2 f8g7 a2a4 c6e7 f1e1 e8g8 h2h3 e7g6 b2b4 c7c6 e2d3 c8e6 a4a5 d8d7 c1b2 g5g4 h3g4 f6g4 d2f1 g8h8 c3c4 f8g8 d1d2 a8e8 a1d1 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(22023,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(22023,31,'cp -1',NULL);
INSERT INTO evaluations VALUES(22023,32,'cp 2',NULL);
INSERT INTO evaluations VALUES(22023,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(22023,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(22023,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(22023,36,'cp 0','g8f6 b1c3 g7g6 c3d5 f8g7 c2c3 c8e6 c1g5 h7h6 g5f6 g7f6 f1e2 e8g8 e1g1 f8e8 a1d1 a8b8 a2a3 f6g7 h2h4 e6d7 c4d3 b7b5 b2b4 c6e7 d5e3 e7c6');
INSERT INTO evaluations VALUES(22431,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(22431,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(22431,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(22431,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(22431,34,'cp 41',NULL);
INSERT INTO evaluations VALUES(22431,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(22431,36,'cp 61','e8f7 f3e5 f7e7 d2d4 d8e8 c1g5 g8f6 b1c3 e7d8 c3d5 f8e7 d5e7 d8e7 e1g1 d7d6 f2f4 h7h6 g5h4 e7f8 d1d2 b7b6 b2b4 a5b7 a1e1 f8g8 h4f6 g7f6 e5f3 d6d5 e4d5 e8h5 d2c3 h8h7 c3c6 h7g7 f1f2 a8b8 c6f6 c8f5 e1e7 g7e7 f6e7 h5f7 e7f7 g8f7 g2g4 b8g8 h2h3 f7f6 g1h2 f5e4 c2c4 e4f3 f2f3 b7d6 c4c5 d6e4 h2g1');
INSERT INTO evaluations VALUES(24428,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24428,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24428,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24428,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(24428,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24428,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24428,36,'cp 0','b1d2 f8e7 d4c5 e7c5 e1g1 e8g8 b2b4 c5e7 b5d3 f6h5 a1c1 h5f4 e3f4 g7g6 b4b5 c6a5 a2a4 a8c8 f1e1 e7f6 f3e5 d8e7 c3c4 f6e5 f4e5 f8d8 d1e2 d5c4 d3c4 e7b4 c1b1 b4a4 b1a1 a4b4 e1b1 b4c3 b1c1 c3b4');
INSERT INTO evaluations VALUES(26060,30,'cp 8',NULL);
INSERT INTO evaluations VALUES(26060,31,'cp 4',NULL);
INSERT INTO evaluations VALUES(26060,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(26060,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(26060,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(26060,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(26060,36,'cp 0','a7a6 b5e2 g7g6 c2c3 f8g7 d2d3 g8f6 c1e3 f6g4 e3d2 g4f6');
INSERT INTO evaluations VALUES(25623,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25623,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25623,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25623,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25623,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25623,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25623,36,'cp 6','b5e2 g7g6 c2c3 f8g7 d2d3 g8f6 c1e3 e8g8 h2h3 b7b6 a2a3 f6h5 b1d2 g7h6 f3h2 h5g3 f1f2 e7e6 d2f1 g3e2 d1e2 h6g7 h2f3 f7f5 f1g3 c6e7 a1d1 b6b5 d3d4 c5d4 c3d4 f5e4 g3e4');
INSERT INTO evaluations VALUES(26458,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(26458,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(26458,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(26458,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(26458,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(26458,35,'cp 50',NULL);
INSERT INTO evaluations VALUES(26458,36,'cp 48','f4g3 h7h5 g1f3 g7g5 h2h3 d8d7 b1d2 e7e6 a1c1 f8d6 g3d6 d7d6 b2b4 a7a6 f1e2 g8h6 d2b3 b7b6 b3d2 e8f7 a2a3 c6e7 c3c4 f7g7 f3h2 g5g4 h3g4 h5g4 c4c5 b6c5 b4c5 d6d7 d2f1 f5g6 h2g4 h6g4 h1h8 a8h8 e2g4');
INSERT INTO evaluations VALUES(29400,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(29400,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(29400,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(29400,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(29400,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(29400,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(29400,36,'cp 43','d3b5 d8d7 f3e5 d7b5 c3b5 b8a6 e5d3 f8e7 c1f4 e8g8 c2c3 f8d8 a2a4 f6e8 g2g4 d8c8 g4g5 a6b8 h2h4 b8c6 b5a3 c6a5 a3c2 a5b3 a1a2 e7d6 e1e2 d6f4 d3f4 e8d6 c2e3 b3a5 h4h5 a5c4 e3c4 c8c4 f2f3 h7h6 g5g6 a8e8 a4a5 f7g6 f4g6 g8h7');
INSERT INTO evaluations VALUES(30750,30,'cp 104',NULL);
INSERT INTO evaluations VALUES(30750,31,'cp 110',NULL);
INSERT INTO evaluations VALUES(30750,32,'cp 103',NULL);
INSERT INTO evaluations VALUES(30750,33,'cp 114',NULL);
INSERT INTO evaluations VALUES(30750,34,'cp 108',NULL);
INSERT INTO evaluations VALUES(30750,35,'cp 140',NULL);
INSERT INTO evaluations VALUES(30750,36,'cp 112','f8g7 g1f3 d7d6 h2h4 h7h6 c1f4 e7e6 e4e5 g8e7 e5d6 c7d6 c3e4 b7e4 d3e4 b8c6 c2c3 a8c8 d1e2 c6a5 e1g1 d8d7 f4h2 e8g8 e4d3 e7d5 f3d2 e6e5 h4h5 g6g5 e2f3 d5e7 a1e1 a5c6 d4e5 c6e5');
INSERT INTO evaluations VALUES(28043,30,'cp 120',NULL);
INSERT INTO evaluations VALUES(28043,31,'cp 114',NULL);
INSERT INTO evaluations VALUES(28043,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(28043,33,'cp 139',NULL);
INSERT INTO evaluations VALUES(28043,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(28043,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(28043,36,'cp 123','g1f3 e7e6 c1e3 d7d6 d1d2 b8d7 e1c1 g8e7 h1e1 a7a6 e3h6 g7h6 d2h6 e7g8 h6f4 b6b5 c1b1 b5b4 c3e2 g8f6 c2c4 b4c3 e2c3 d8e7 b1a1 e8g8 d3c2 e6e5 d4e5 d7e5 f3e5 e7e5 f4e5');
INSERT INTO evaluations VALUES(31702,30,'cp -53',NULL);
INSERT INTO evaluations VALUES(31702,31,'cp -60',NULL);
INSERT INTO evaluations VALUES(31702,32,'cp -55',NULL);
INSERT INTO evaluations VALUES(31702,33,'cp -42',NULL);
INSERT INTO evaluations VALUES(31702,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(31702,35,'cp -56',NULL);
INSERT INTO evaluations VALUES(31702,36,'cp -42','g1f3 g8f6 c1f4 f8e7 e1g1 e8g8 h2h3 h7h6 c2c3 b8c6 d3d4 b7b5 a2a3 d8b6 b1d2 a7a5 b3c2 f5c2 d1c2 f8c8 c2d3 b6b7 f1c1 f6d7 c1e1 b5b4 a3b4 a5b4 c3c4 d5c4 d2c4 b7b5 c4e5 b5d3 e5d3 d7b6 a1a8 c8a8 g1f1');
INSERT INTO evaluations VALUES(31701,30,'cp -41',NULL);
INSERT INTO evaluations VALUES(31701,31,'cp -39',NULL);
INSERT INTO evaluations VALUES(31701,32,'cp -60',NULL);
INSERT INTO evaluations VALUES(31701,33,'cp -59',NULL);
INSERT INTO evaluations VALUES(31701,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(31701,35,'cp -48',NULL);
INSERT INTO evaluations VALUES(31701,36,'cp -45','b8c6 e1g1 g8f6 c1f4 f8e7 d3d4 f6h5 f4e3 e8g8 b1d2 h5f6 c2c3 d8c7 f1e1 h7h6 h2h3 e7d6 c3c4 d5c4 d2c4 f5e4 f3e5 a8d8 a1c1 c6e5 c4e5 c7e7 d1e2 e4d5 b3d5 f6d5 e3d2 d8c8 a2a3 f8d8 c1c8 d8c8');
INSERT INTO evaluations VALUES(31839,30,'cp -10',NULL);
INSERT INTO evaluations VALUES(31839,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(31839,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(31839,33,'cp -10',NULL);
INSERT INTO evaluations VALUES(31839,34,'cp -24',NULL);
INSERT INTO evaluations VALUES(31839,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(31839,36,'cp -5','f5g6 c2c4 e7e6 b1c3 c7c6 d1b3 d8b6 h2h4 h7h5 g4g5 b8d7 c1d2 g8e7 g1h3 d5c4 f1c4 e6e5 d4e5 d7e5 h3f4 g6f5 c4e2 e7g6 b3b6 a7b6 f4h5 e8c8 h5g3 h8h4 h1h4 g6h4 g3f5 h4f3 e2f3 e5f3 e1e2 f3d2 a1d1 d2c4 d1d8 c8d8');
INSERT INTO evaluations VALUES(31805,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31805,31,'cp -10',NULL);
INSERT INTO evaluations VALUES(31805,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(31805,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31805,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(31805,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(31805,36,'cp -1','c2c4 e7e6 b1c3 d5c4 f1c4 f8b4 c1d2 b8c6 g1f3 g8f6 f3h4 d8d7 d1e2 e8c8 h4g6 h7g6 e1c1 f6d5 c3e4 b4d2 e4d2 d5b6 d2e4 b6c4 e2c4 d7d5 c4d5 e6d5 e4g5 d8e8 h2h4 f7f6 g5f3 a7a5 a2a3 b7b6 d1g1 c6e7 c1c2 c8d7 f3d2 c7c5');
INSERT INTO evaluations VALUES(35367,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(35367,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(35367,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(35367,33,'cp 13',NULL);
INSERT INTO evaluations VALUES(35367,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(35367,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(35367,36,'cp 0','g2g3 b8a6 f1g2 c7c6 c1d2 a6c7 e1c1 d6d5 h1e1 f8e8 e1e2 c7e6 d1e1 e7f8 c3d1 a7a5 d1e3 b7b5 e3f5 b5b4 f5d4 c8d7 d4e6 d7e6 f3g5 e6d7 g5f3');
INSERT INTO evaluations VALUES(35916,30,'cp 87',NULL);
INSERT INTO evaluations VALUES(35916,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(35916,32,'cp 84',NULL);
INSERT INTO evaluations VALUES(35916,33,'cp 63',NULL);
INSERT INTO evaluations VALUES(35916,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(35916,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(35916,36,'cp 76','e1g1 b8a6 h2h3 e7e5 d4d5 a6c5 d3c2 c6d5 c4d5 a7a5 c1e3 c8d7 e3c5 d6c5 d1d3 f6e8 c2a4 e8d6 a4d7 d8d7 a2a4 c5c4 d3e2 f8e8 f1c1 a8c8 c1c2 g7h6 f3d2 c8c5 a1d1 h6d2 d1d2 h7h5 e2e3 d7c7 c2c1');
INSERT INTO evaluations VALUES(35915,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(35915,31,'cp 65',NULL);
INSERT INTO evaluations VALUES(35915,32,'cp 61',NULL);
INSERT INTO evaluations VALUES(35915,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(35915,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(35915,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(35915,36,'cp 55','b8a6 h2h3 e7e5 f1e1 e5d4 f3d4 f6g4 h3g4 g7d4 c1e3 d4g7 g4g5 a6c5 d1d2 c5d3 d2d3 c8e6 a1d1 d8a5 e3f4 f8e8 d3g3 g7e5 d1d6 e5d6 f4d6 e6c4 e1d1 c4e6 g3f4 a8d8 f4f6 d8d6 d1d6 a5b4 d6d2 b4b6 g1h2 b6b4 b2b3');
INSERT INTO evaluations VALUES(37806,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(37806,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(37806,32,'cp 93',NULL);
INSERT INTO evaluations VALUES(37806,33,'cp 90',NULL);
INSERT INTO evaluations VALUES(37806,34,'cp 99',NULL);
INSERT INTO evaluations VALUES(37806,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(37806,36,'cp 89','d4c5 d8c7 b1a3 c7c5 g1f3 g8f6 a3b5 e8g8 c1e3 c5c6 f3d4 c6e4 b5c7 b7b6 d1f3 e4f3 d4f3 c8b7 c7a8 b7a8 f1e2 f6d5 e3d2 d7d6 e1g1 f8c8 f1e1 a8b7 h2h4 b8d7 h4h5 e7e5 a1d1 d7c5 d2g5 g7f8 f3h2 c5e6 e2f3 c8c5 g5e3 d5e3 e1e3');
INSERT INTO evaluations VALUES(38455,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(38455,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(38455,32,'cp 10',NULL);
INSERT INTO evaluations VALUES(38455,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(38455,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(38455,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(38455,36,'cp 23','g8f6 b1d2 c7c5 g1f3 e8g8 f1e2 b8d7 h2h3 a7a6 e1g1 b7b6 a2a4 c8b7 d1b3 c5c4 b3c2 b6b5 b2b3 e7e6 f1b1 b7c6 f4d6 c4b3 c2b3 b5a4 b3d1 f8e8 f3e5 d7e5 d4e5 f6d7 d2f3 a6a5');
INSERT INTO evaluations VALUES(38558,30,'cp 80',NULL);
INSERT INTO evaluations VALUES(38558,31,'cp 86',NULL);
INSERT INTO evaluations VALUES(38558,32,'cp 83',NULL);
INSERT INTO evaluations VALUES(38558,33,'cp 89',NULL);
INSERT INTO evaluations VALUES(38558,34,'cp 84',NULL);
INSERT INTO evaluations VALUES(38558,35,'cp 81',NULL);
INSERT INTO evaluations VALUES(38558,36,'cp 83','b8d7 f1d3 c7c6 a2a4 g8f6 e1g1 b7b6 h2h3 d8c7 c3e2 c6c5 c2c3 c8b7 e2g3 e8g8 f3d2 a8c8 f1e1 d7b8 d2f3 b8c6 a1c1 c7b8 e3g5 h7h6 g5e3 c5d4 c3d4 c6b4 d3b1 e6e5 d1d2 c8c1 e1c1');
INSERT INTO evaluations VALUES(39488,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(39488,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39488,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(39488,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(39488,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(39488,35,'cp 27',NULL);
INSERT INTO evaluations VALUES(39488,36,'cp 15','d7d5 c4d5 e6d5 c1f4 c7c6 e2e3 c8f5 g2g4 f5e6 h2h3 h7h5 g4h5 g8f6 h5h6 g7g6 g1f3 f6h5 f4h2 h8h6 d1c2 e7d6 f3e5 d8e7 c2b3 d6e5 h2e5 b7b5 e5h2 b8d7 a1c1 d7b6 a2a4 b5b4 c3e2 b6c4 e2f4 h5f4 h2f4 h6h4 f4g3');
INSERT INTO evaluations VALUES(39960,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(39960,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(39960,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(39960,33,'cp -1',NULL);
INSERT INTO evaluations VALUES(39960,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(39960,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(39960,36,'cp -12','g1f3 d7d5 e4d5 c6d5 c4b3 b8c6 e1g1 h7h6 g5h4 c8g4 h2h3 g4h5 g2g4 h5g6 h4g3 e5e4 d3e4 f6e4 f3e5 c6e5 g3e5 e8g8 b1c3 e7f6 e5f6 d8f6 c3d5 f6h4 d5f4 g6h7 d1d3 e4f2 d3e3 f2e4 f4g2 h4e7 a1e1 e7c7 b3d5 a8e8 e3a7 e4g5');
INSERT INTO evaluations VALUES(40853,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(40853,31,'cp 63',NULL);
INSERT INTO evaluations VALUES(40853,32,'cp 80',NULL);
INSERT INTO evaluations VALUES(40853,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(40853,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(40853,35,'cp 69',NULL);
INSERT INTO evaluations VALUES(40853,36,'cp 83','c1f4 a7a6 e1c1 e8g8 e4e5 b8c6 e5f6 c6d4 f6e7 d8e7 f3d4 c7c6 f1e2 f8e8 a2a4 c8d7 e2f3 d6d5 d4b3 h7h6 f4e3 e7d8 d1d2 b7b6 c1b1 d7f5 b1a1 b6b5 a4a5 d8e7 a1b1 f5h7 e3b6 a8b8 h1d1 b8b6 a5b6 e8b8 h2h3 b8b6');
INSERT INTO evaluations VALUES(41532,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(41532,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(41532,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(41532,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(41532,34,'cp 62',NULL);
INSERT INTO evaluations VALUES(41532,35,'cp 64',NULL);
INSERT INTO evaluations VALUES(41532,36,'cp 47','e2e4 h7h6 g5e3 c7c5 f1c4 f6g4 d4c5 g4e3 d1d8 e8d8 f2e3 e7e6 e1g1 d8e7 c3b5 b8d7 b5d6 h8f8 c4b5 d7c5 a1c1 e7d6 f1d1 d6c7 c1c5 c7b8 e4e5 a7a6 b5e2 b7b6 c5c4 b6b5 c4b4 b8c7 a2a4 b5a4 b4a4 a8a7 b2b4 c8b7 b4b5 a6a5 a4c4 c7b8 b5b6 a7a6 d1b1 b7f3 e2f3 g7e5 b6b7 a6a7');
INSERT INTO evaluations VALUES(43682,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(43682,31,'cp 119',NULL);
INSERT INTO evaluations VALUES(43682,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(43682,33,'cp 126',NULL);
INSERT INTO evaluations VALUES(43682,34,'cp 130',NULL);
INSERT INTO evaluations VALUES(43682,35,'cp 115',NULL);
INSERT INTO evaluations VALUES(43682,36,'cp 106','e2e4 d7d6 f1d3 e6e5 g1e2 f8e7 h2h3 e8g8 c1e3 e5d4 e2d4 f8e8 e1g1 e7f8 a2a4 b8c6 d3b1 c6d4 d1d4 c8d7 b1c2 d7c6 f2f3 f6d7 d4d2 a6a5 a1d1 g7g6 b2b3 f8g7 e3f2 d7c5 c3b5 c6d7 f2e3 d7c6 f1e1 c6b5');
INSERT INTO evaluations VALUES(45171,30,'cp 109',NULL);
INSERT INTO evaluations VALUES(45171,31,'cp 112',NULL);
INSERT INTO evaluations VALUES(45171,32,'cp 127',NULL);
INSERT INTO evaluations VALUES(45171,33,'cp 119',NULL);
INSERT INTO evaluations VALUES(45171,34,'cp 146',NULL);
INSERT INTO evaluations VALUES(45171,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(45171,36,'cp 119','d4c5 d6c5 d1d8 e8d8 c1f4 h7h6 a1d1 c8d7 h2h4 b8c6 g1f3 e7e6 e2e4 f6g4 h4h5 e6e5 f4g3 d8c8 f1e2 d7e6 b2b3 f8e7 f3h4 e7d8 e1f1 g4f6 h4f5 e6f5 e4f5 d8c7 e2d3 h8d8 g3h4 c6b4 d3b1 d8d1 c3d1 b4c6');
INSERT INTO evaluations VALUES(47054,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(47054,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(47054,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(47054,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(47054,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(47054,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(47054,36,'cp 0','e7e5 g1e2 c7c5 d2d4 e5d4 e3d4 d6d5 b1c3 b8c6 d1d2 f8d6 e1c1 e8g8 d4c5 d6c5 c3d5 f6d5 d2d5 d8d5 d1d5 c5f2 e2c3 f8e8 f1e2 c6b4 d5a5 b4c6');
INSERT INTO evaluations VALUES(49218,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(49218,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(49218,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(49218,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(49218,34,'cp -42',NULL);
INSERT INTO evaluations VALUES(49218,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(49218,36,'cp -47','f8g7 f1d3 d7d5 e4d5 d8d5 g1f3 e8g8 e1g1 f6f5 f1e1 d5d8 c2c3 c7c5 d3c4 c5d4 f3d4 b8d7 b1d2 d7e5 c4b3 e5d3 e1e3 d3c5 d2f3 d8c7 d1e2 c8d7 a1e1 c5b3 a2b3 g7f6 b3b4 f8e8 e3e8 a8e8 e2d2 e8e1 f3e1 g8g7 e1f3 a7a6 h2h3 b7b6');
INSERT INTO evaluations VALUES(42335,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(42335,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(42335,32,'cp -37',NULL);
INSERT INTO evaluations VALUES(42335,33,'cp -29',NULL);
INSERT INTO evaluations VALUES(42335,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(42335,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(42335,36,'cp -33','g1f3 e8g8 f1d3 d7d5 e4d5 d8d5 e1g1 f6f5 f1e1 d5d8 d3c4 c7c5 c2c3 c5d4 f3d4 b8d7 b1d2 d7e5 c4e2 d8c7 d2f3 a7a6 d1b3 c8d7 a1d1 a8d8 f3e5 g7e5 g2g3 g8g7 b3c4 f8e8 c4c7 e5c7 e2c4 e8e1 d1e1 b7b5 c4b3 g7f6 d4c2 a6a5 b3d5 c7b6 c2e3 a5a4 e1d1');
INSERT INTO evaluations VALUES(50650,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(50650,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(50650,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(50650,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(50650,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(50650,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(50650,36,'cp 15','g1f3 e7e6 a2a4 f8e7 c2c3 g8f6 e1g1 b8c6 f1e1 e8g8 c1f4 f6h5 f4e3 h5f6 b1d2 f6g4 e3f4 g4e5 f4g3 c8d7 d1e2 d8c7 f3e5 c6e5 c4b3 e5c6 h2h3 a8b8 a1d1 b7b5 a4b5 a6b5 d2f3 e7f6 d3d4 c5d4');
INSERT INTO evaluations VALUES(50649,30,'cp 10',NULL);
INSERT INTO evaluations VALUES(50649,31,'cp 9',NULL);
INSERT INTO evaluations VALUES(50649,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(50649,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(50649,34,'cp 6',NULL);
INSERT INTO evaluations VALUES(50649,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(50649,36,'cp 2','e7e6 a2a4 f8e7 c2c3 g8f6 e1g1 b7b6 c1f4 c8b7 b1d2 f6h5 f4e3 h5f6 d1e2 b8d7 h2h3 d6d5 c4b3 d8c7 e3g5 c5c4 b3c2 c4d3 c2d3 d5e4 d2e4 f6d5 g5e7 d5e7 a1d1 e8g8 f1e1 e7g6 d3b1 b7d5 b1c2 g6f4 e2e3 a8d8 b2b4');
INSERT INTO evaluations VALUES(50981,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(50981,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(50981,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(50981,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(50981,34,'cp 16',NULL);
INSERT INTO evaluations VALUES(50981,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(50981,36,'cp 14','f1c4 g8f6 c4b3 c7c5 g1f3 b7b5 e1g1 c8b7 e3e4 c5d4 f3d4 b8c6 d4c6 d8d1 f1d1 b7c6 f2f3 f8c5 g1f1 e8e7 e4e5 f6d7 c1g5 e7e8 a1c1 h7h6 g5d2 e8e7 b3c2 d7e5 c3e4 c6e4 c2e4 a8c8 c1c5 c8c5 d2b4 h8c8 e4b7 c8c7 b7a6 e5c4 a6b5 c4b2');
INSERT INTO evaluations VALUES(51133,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(51133,31,'cp 77',NULL);
INSERT INTO evaluations VALUES(51133,32,'cp 72',NULL);
INSERT INTO evaluations VALUES(51133,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(51133,34,'cp 70',NULL);
INSERT INTO evaluations VALUES(51133,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(51133,36,'cp 72','c4b3 b8c6 a2a4 b5b4 e1g1 f8e7 b3d5 c8d7 d3d4 g8f6 d4e5 f6d5 e5d6 e7d6 e4d5 c6e7 c2c4 b4c3 b1c3 e8g8 f1e1 f8e8 h2h3 h7h6 f3e5 e7g6 e5c4 e8e1 d1e1 d8f6 c4d6 f6d6 c1e3 g6h4 f2f3 h4f5 e1f2 f5e3 f2e3 a8e8 e3f2');
INSERT INTO evaluations VALUES(51136,30,'cp 70',NULL);
INSERT INTO evaluations VALUES(51136,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(51136,32,'cp 71',NULL);
INSERT INTO evaluations VALUES(51136,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(51136,34,'cp 68',NULL);
INSERT INTO evaluations VALUES(51136,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(51136,36,'cp 55','b8c6 a2a4 b5b4 d3d4 g8f6 d4e5 d6e5 d1d8 c6d8 f3e5 f8d6 e5d3 f6e4 b1d2 e4c5 d3c5 d6c5 d2e4 c5a7 c1e3 a7e3 f2e3 d8e6 h1f1 f7f6 e1c1 e8e7 b3d5 a8b8 d1d3 h8d8 b2b3 c8b7 d5b7 b8b7 d3d8 e7d8 f1f5 d8d7 f5h5 d7c6 h5h7');
INSERT INTO evaluations VALUES(51568,30,'cp 11',NULL);
INSERT INTO evaluations VALUES(51568,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(51568,32,'cp 5',NULL);
INSERT INTO evaluations VALUES(51568,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(51568,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(51568,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51568,36,'cp 10','a4b3 g8f6 g1f3 f6e4 e1g1 d7d5 f3e5 f8d6 d2d4 e8g8 b1d2 e4d2 c1d2 d8c7 f1e1 b8d7 d2f4 d7e5 f4e5 c8f5 e5d6 c7d6 e1e5 d6g6 d1e2 a7a5 a2a4 b5a4 b3a4 a8b8 a4b3 f5d3 c2d3 b8b3 e5e8 b3b8 e8b8 f8b8 a1a5');
INSERT INTO evaluations VALUES(53069,30,'cp -16',NULL);
INSERT INTO evaluations VALUES(53069,31,'cp -23',NULL);
INSERT INTO evaluations VALUES(53069,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(53069,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(53069,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(53069,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(53069,36,'cp -17','g7g6 c2c3 g8f6 g1f3 f8g7 f1e2 e8g8 e1g1 b7b6 f1e1 c8b7 e2f1 b8d7 a2a4 e7e5 c1g5 h7h6 g5h4 a7a6 b1d2 f8e8 b2b4 d8c7 h4g3 d6d5 a1b1 c5b4 c3b4 a8c8 b4b5 a6b5 a4b5 d7c5 f3e5');
INSERT INTO evaluations VALUES(54023,30,'cp 130',NULL);
INSERT INTO evaluations VALUES(54023,31,'cp 125',NULL);
INSERT INTO evaluations VALUES(54023,32,'cp 116',NULL);
INSERT INTO evaluations VALUES(54023,33,'cp 129',NULL);
INSERT INTO evaluations VALUES(54023,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(54023,35,'cp 122',NULL);
INSERT INTO evaluations VALUES(54023,36,'cp 109','b8c6 g1f3 g8f6 f1d3 f8e7 e1g1 a7a6 a2a3 c8d7 f1e1 a8c8 c1d2 d8b6 e4e5 d6e5 d4e5 f6d5 c3d5 e6d5 b2b4 d7e6 d2e3 b6d8 h2h3 e8g8 d1d2 f8e8 a1d1 d8c7 d1c1 c7d7 d3b1 d5d4 c1d1 d4e3 d2d7 e6d7 d1d7');
INSERT INTO evaluations VALUES(55632,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(55632,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(55632,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(55632,33,'cp -34',NULL);
INSERT INTO evaluations VALUES(55632,34,'cp -31',NULL);
INSERT INTO evaluations VALUES(55632,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(55632,36,'cp -31','e5e4 f3g1 d6d5 e2e3 b8c6 b1c3 g8f6 d2d3 f8d6 d3e4 d5e4 f1b5 e8g8 b5c6 b7c6 g1e2 d8e7 d1d4 a7a5 d4c4 c6c5 c3d5 f6d5 c4d5 c8d7 a2a4 f8e8 e1c1 d7e6 d5e4 c5c4 e2d4 d6a3 d4b5 a3b2 c1b2 c4b3 c2b3 e8b8 b5d4 e7f6 b2a3');
INSERT INTO evaluations VALUES(57255,30,'cp -55',NULL);
INSERT INTO evaluations VALUES(57255,31,'cp -55',NULL);
INSERT INTO evaluations VALUES(57255,32,'cp -54',NULL);
INSERT INTO evaluations VALUES(57255,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(57255,34,'cp -59',NULL);
INSERT INTO evaluations VALUES(57255,35,'cp -63',NULL);
INSERT INTO evaluations VALUES(57255,36,'cp -59','b8c6 d2d4 d7d5 d4e5 c6e5 g1f3 e5f3 d1f3 g8f6 h2h3 f8d6 c1d2 c8e6 f1d3 f6d7 f3h5 d7e5 b1c3 c7c6 a4a5 g7g6 h5h6 d8e7 e3e4 d5d4 c3e2 c6c5 e2f4 c5c4 d3e2 e8c8 a5a6 b7b5 e1g1 d6c7 d2a5 d8d7 a5c7 c8c7');
INSERT INTO evaluations VALUES(2468,30,'cp -586',NULL);
INSERT INTO evaluations VALUES(2468,31,'cp -594',NULL);
INSERT INTO evaluations VALUES(2468,32,'cp -601',NULL);
INSERT INTO evaluations VALUES(2468,33,'cp -608',NULL);
INSERT INTO evaluations VALUES(2468,34,'cp -609',NULL);
INSERT INTO evaluations VALUES(2468,35,'cp -611',NULL);
INSERT INTO evaluations VALUES(2468,36,'cp -607','f7g8 c2c4 h7h6 g5f3 f6e4 e1g1 d7d6 d2d3 e4f6 b1c3 c8g4 c1e3 d8d7 a2a3 g4h5 c3e4 f6e4 d3e4 f8e7 d1d5 g8h7 f3d2 c6d8 d5d3 c7c6 a1e1 d7e8 f2f3 d8e6 c4c5 d6c5 d2c4 e6d4 e3d4 c5d4 e1d1 e7g5 f1f2 h5f7');
INSERT INTO evaluations VALUES(3076,30,'cp -17',NULL);
INSERT INTO evaluations VALUES(3076,31,'cp -17',NULL);
INSERT INTO evaluations VALUES(3076,32,'cp -15',NULL);
INSERT INTO evaluations VALUES(3076,33,'cp -18',NULL);
INSERT INTO evaluations VALUES(3076,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(3076,35,'cp -17',NULL);
INSERT INTO evaluations VALUES(3076,36,'cp -30','d7d5 e4d5 f6d5 d4c6 b7c6 c3d5 c6d5 c2c3 a7a5 a1c1 c8a6 e1g1 a6e2 d2e2 e7e6 f1d1 d8d6 h2h4 f8c8 e2d3 d6e5 g2g3 h7h5 c3c4 c8c4 c1c4 d5c4 d3c4 e5b2 c4c6 a8f8 c6a4 g7c3 a4c4 f8b8 e3f4');
INSERT INTO evaluations VALUES(3151,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(3151,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(3151,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(3151,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(3151,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(3151,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(3151,36,'cp -29','e4d5 f6d5 d4c6 b7c6 c3d5 d8d5 d2d5 c6d5 e1c1 f8d8 c2c3 h7h5 e2f3 e7e6 h2h4 c8b7 h1g1 f7f5 d1d2 a7a5 g1e1 g8f7 e3b6 g7h6 b6d8 a8d8 c1c2 h6d2 c2d2 f7f6 f3e2 d5d4 g2g3 e6e5 f2f4 b7e4 c3d4 d8d4 d2e3 d4b4 e1d1 e5f4 g3f4 b4b2');
INSERT INTO evaluations VALUES(3146,30,'cp -27',NULL);
INSERT INTO evaluations VALUES(3146,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(3146,32,'cp -16',NULL);
INSERT INTO evaluations VALUES(3146,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(3146,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(3146,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(3146,36,'cp -18','f6d5 d4c6 b7c6 c3d5 c6d5 e1g1 a7a5 f1d1 a8b8 c2c3 a5a4 e2f3 c8e6 h2h4 h7h5 a1c1 d8a5 b2b4 a4b3 a2b3 b8b3 f3d5 e6d5 d2d5 a5d5 d1d5 b3c3 c1d1 g7f6 g2g3 c3a3 d1c1 f8b8 c1c7 b8b2 d5d2 b2d2 e3d2 a3a1 g1h2 g8g7 h2g2');
INSERT INTO evaluations VALUES(3810,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(3810,31,'cp -29',NULL);
INSERT INTO evaluations VALUES(3810,32,'cp -33',NULL);
INSERT INTO evaluations VALUES(3810,33,'cp -30',NULL);
INSERT INTO evaluations VALUES(3810,34,'cp -41',NULL);
INSERT INTO evaluations VALUES(3810,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(3810,36,'cp -24','e1g1 h7h6 g5f3 c6a5 c4b3 b7b5 b1c3 c8d7 c1e3 c5b6 f3d2 a5b3 a2b3 b6e3 f2e3 a7a5 d2f3 f8e8 d1e1 c7c6 e1g3 d7e6 f3h4 f6h5 g3f2 e8f8 h4f3 h5f6 f2g3 d8e7 g3e1 f8e8 c3e2 e7d8 e1g3 a5a4 b3a4 b5a4');
INSERT INTO evaluations VALUES(3809,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(3809,31,'cp -31',NULL);
INSERT INTO evaluations VALUES(3809,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(3809,33,'cp -39',NULL);
INSERT INTO evaluations VALUES(3809,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(3809,35,'cp -38',NULL);
INSERT INTO evaluations VALUES(3809,36,'cp -35','c6a5 c4b3 b7b5 b1c3 h7h6 g5f3 c8d7 c1d2 c5b6 c3e2 a5b3 a2b3 a7a5 d2e3 b6e3 f2e3 d8e7 d1d2 a5a4 b3a4 b5a4 e2c3 f8b8 a1a2 e7e8 f1a1 d7c6 g1h2 g8h8 d2f2 d6d5 e4d5 f6d5 e3e4 d5c3 b2c3 f7f6 f3h4 b8b6 c3c4 c6d7 h4f5 d7f5 f2f5 h8g8');
INSERT INTO evaluations VALUES(4016,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(4016,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(4016,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(4016,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(4016,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(4016,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(4016,36,'cp 57','h7h6 g5e3 e7e5 e1g1 f6g4 e3c1 g4f6 d1c2 f8e8 d4d5 a7a5 f3e1 e8f8 e1d3 b7b6 c1e3 d7c5 f2f3 c5d3 e2d3 f6d7 c2d2 d7c5 d3c2 g8h7 f1f2 c8d7 a1f1 d8e7 g1h1 f7f5 e4f5 d7f5 b2b3 a8e8 g2g4 f5c2 d2c2');
INSERT INTO evaluations VALUES(5005,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(5005,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(5005,32,'cp -9',NULL);
INSERT INTO evaluations VALUES(5005,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(5005,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(5005,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(5005,36,'cp 0','b1c3 d7d5 c4b3 g8f6 c1g5 h7h6 g5h4 g6g5 h4g3 f6e4 c3e4 d5e4 f3e4 d8a5 e1f1 e8g8 f1g1 a5c5 e4d3 c5b5 c2c4 b5f5 d3e2 f5f6 a1d1 f6b2 d1d2 b2a1 d2d1');
INSERT INTO evaluations VALUES(4999,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(4999,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(4999,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(4999,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(4999,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(4999,35,'cp -13',NULL);
INSERT INTO evaluations VALUES(4999,36,'cp -9','d7d5 c4d3 g8e7 c1g5 e8g8 g5f6 a8b8 b2b3 e6e5 e1g1 d8d6 f6g7 g8g7 f3g3 f7f5 e4f5 e7f5 d3f5 c8f5 a1e1 b8e8 e1e2 h7h5 f1e1 d6f6 e2e5 h5h4 g3f4 e8e5 e1e5 f5e4 f4f6 g7f6 f2f4 e4c2 c3e2 c2b1 e2d4 c6c5 d4b5 b1a2 e5d5 a2b3 d5c5 f8d8 b5c3 d8d4 c5a5 d4f4 a5a7 f4c4 a7a6 f6e7');
INSERT INTO evaluations VALUES(5140,30,'cp 283',NULL);
INSERT INTO evaluations VALUES(5140,31,'cp 278',NULL);
INSERT INTO evaluations VALUES(5140,32,'cp 280',NULL);
INSERT INTO evaluations VALUES(5140,33,'cp 286',NULL);
INSERT INTO evaluations VALUES(5140,34,'cp 285',NULL);
INSERT INTO evaluations VALUES(5140,35,'cp 267',NULL);
INSERT INTO evaluations VALUES(5140,36,'cp 274','c3b5 a7a6 b5d6 e8f8 c1g5 g8e7 a2a4 h7h6 g5e3 f8g8 d1e2 g6g5 f1d1 h8h7 a1c1 e7g6 h2h3 c6e5 b2b4 e5f3 e2f3 g7e5 a4a5 d8f8 e3c5 h6h5 g2g3 g5g4 f3e2 g4h3 f2f4 e5d6 c5d6 f8g7 c4b3 h5h4 g3g4 b7b6 a5b6 c8b7 c1c7 g7f6 c7b7 g6f4');
INSERT INTO evaluations VALUES(6621,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(6621,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(6621,32,'cp 29',NULL);
INSERT INTO evaluations VALUES(6621,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(6621,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(6621,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(6621,36,'cp 21','b4a5 e1g1 d7d6 d2d3 g8f6 b1d2 e8g8 a3a4 a5b6 c2c3 a7a6 h2h3 c8e6 f1e1 h7h6 c4b3 c6a5 b3a2 e6a2 a1a2 a5c6 d2c4 b6a7 d1b3 d8d7 a4a5 f8e8 c1e3 d6d5 e4d5 d7d5 a2a1 d5b5 b3a2 a7e3 e1e3 e5e4 d3e4 f6e4 b2b3 e4f6');
INSERT INTO evaluations VALUES(6730,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(6730,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(6730,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(6730,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(6730,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(6730,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(6730,36,'cp 23','e1g1 d7d6 d2d3 g8f6 c2c3 a5b6 b1d2 e8g8 c4b3 h7h6 f1e1 f8e8 h2h3 c8e6 d2c4 c6e7 d3d4 e5d4 c3d4 c7c6 c1d2 e6c4 b3c4 d6d5 e4d5 c6d5 c4b5 e7c6 e1e8 d8e8 d2e3 e8d7 b5d3 a8e8 d1c2 d7d6 a1b1 f6e4 b2b4 d6f6');
INSERT INTO evaluations VALUES(9614,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(9614,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(9614,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(9614,33,'cp 49',NULL);
INSERT INTO evaluations VALUES(9614,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(9614,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(9614,36,'cp 36','f1e2 e8g8 e3e4 e7e5 e1g1 e5d4 f3d4 c7c6 d1c2 d7c5 d4b3 d8e7 b3c5 d6c5 c1d2 c8e6 a1d1 f6g4 h2h3 g4e5 b2b3 g6g5 e2h5 f7f5 f2f4 g5f4 f1f4 e5g6 f4f1 f5e4 c2e4 f8f1 d1f1 g7d4 g1h1 a8f8 f1f8 g6f8 e4e1 f8g6 c3e4 e6f5');
INSERT INTO evaluations VALUES(9672,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(9672,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(9672,32,'cp 94',NULL);
INSERT INTO evaluations VALUES(9672,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(9672,34,'cp 91',NULL);
INSERT INTO evaluations VALUES(9672,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(9672,36,'cp 100','c2c3 d4e6 c4e6 d7e6 g1f3 f6d7 d3d4 c5d6 e1g1 b7b6 b1d2 e8g8 f1e1 e5d4 c3d4 e6e5 d2c4 d6b4 c1g5 f7f6 g5d2 b4d2 d1d2 e5d4 f3d4 d7e5 d2c3 c7c5 c4e5 d8d4 c3d4 c5d4 e5c6 d4d3 a1d1 c8b7 c6b4 f8d8 b4d3 a8c8 f2f3 c8c2 d3f4 d8d1');
INSERT INTO evaluations VALUES(9671,30,'cp 74',NULL);
INSERT INTO evaluations VALUES(9671,31,'cp 102',NULL);
INSERT INTO evaluations VALUES(9671,32,'cp 74',NULL);
INSERT INTO evaluations VALUES(9671,33,'cp 77',NULL);
INSERT INTO evaluations VALUES(9671,34,'cp 75',NULL);
INSERT INTO evaluations VALUES(9671,35,'cp 90',NULL);
INSERT INTO evaluations VALUES(9671,36,'cp 85','d4e6 c4e6 d7e6 g1f3 f6d7 d3d4 c5d6 e1g1 b7b6 c1h6 c8b7 b1d2 h8g8 h2h4 g6g5 h4g5 d8e7 d2c4 e8c8 d1e2 e5d4 c3d4 f7f6 e2e3 c8b8 c4d6 e7d6 g5f6 d7f6 e4e5 d6d4 f3d4 g8g2 g1h1 f6g4 f2f3 g4e3 h6e3 g2b2 d4e6 d8g8 e6f4 g8g3 e5e6 b7f3 f1f3 g3f3 a1e1');
INSERT INTO evaluations VALUES(9923,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(9923,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(9923,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(9923,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(9923,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(9923,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(9923,36,'cp 49','e1g1 d7d6 a2a4 a7a6 c4b3 h7h6 a4a5 b6a7 c1e3 a7e3 f2e3 e8g8 b1d2 c8e6 b3e6 f7e6 d1b3 d8e7 d3d4 e7f7 h2h3 e5d4 c3d4 e6e5 d4d5 c6a7 f3h4 f7h5 h4f5 a7b5 f1f2 f8f7 a1f1 a8f8 b3b2 h5g6 b2c2 g6h5 d2c4 h5g5 f2f3');
INSERT INTO evaluations VALUES(10280,30,'cp 15',NULL);
INSERT INTO evaluations VALUES(10280,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10280,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10280,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10280,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10280,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10280,36,'cp 0','e4e5 d7d5 e5f6 d5c4 f6g7 h8g8 b1c3 d8d1 f1d1 g8g7 c3d5 c5d6 d5f6 e8e7 f6d5 e7e8 d5f6');
INSERT INTO evaluations VALUES(10049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(10049,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(10049,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(10049,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(10049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(10049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(10049,36,'cp 0','d7d5 e5f6 d5c4 f6g7 h8g8 b1c3 d8d1 f1d1 g8g7 c3d5 c5d6 d5f6 e8e7 f6d5 e7e8 d5f6');
INSERT INTO evaluations VALUES(11120,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(11120,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(11120,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(11120,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(11120,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(11120,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(11120,36,'cp 38','d2d4 c5d4 f3d4 d8c7 d4c6 d7c6 a2a4 e7e5 e1g1 f8b4 d1f3 c7e7 h2h3 e8g8 a4a5 c8e6 c4e6 e7e6 c1g5 h7h6 g5f6 e6f6 f3f6 g7f6 a1a4 b4c3 b2c3 f8d8 f1b1 a8b8 a4b4 d8d2 b4b7 b8b7 b1b7 d2c2 b7b6 c2c3 b6a6 g8g7 a6b6 c3a3 a5a6');
INSERT INTO evaluations VALUES(14007,30,'cp 118',NULL);
INSERT INTO evaluations VALUES(14007,31,'cp 129',NULL);
INSERT INTO evaluations VALUES(14007,32,'cp 125',NULL);
INSERT INTO evaluations VALUES(14007,33,'cp 137',NULL);
INSERT INTO evaluations VALUES(14007,34,'cp 122',NULL);
INSERT INTO evaluations VALUES(14007,35,'cp 134',NULL);
INSERT INTO evaluations VALUES(14007,36,'cp 118','f8e7 d1e2 a7a6 a2a4 c8e6 c1f4 e8g8 e1g1 e6c4 e2c4 d8d7 f1d1 d7c6 a1a3 f8b8 f2f3 b8d8 a3b3 d8d7 a4a5 e7d8 b3a3 d8e7 f4g3 a8d8 c3d5 f6d5 d1d5 d8e8 a3a1 e7f6 c2c3 f6e5 a1d1 e5g3 h2g3 c6c7 b2b4 c5b4 c4b4');
INSERT INTO evaluations VALUES(8213,30,'cp 119',NULL);
INSERT INTO evaluations VALUES(8213,31,'cp 126',NULL);
INSERT INTO evaluations VALUES(8213,32,'cp 114',NULL);
INSERT INTO evaluations VALUES(8213,33,'cp 117',NULL);
INSERT INTO evaluations VALUES(8213,34,'cp 117',NULL);
INSERT INTO evaluations VALUES(8213,35,'cp 129',NULL);
INSERT INTO evaluations VALUES(8213,36,'cp 118','a2a4 c8e6 d1d3 a7a6 e1g1 e8g8 a4a5 d8d7 b2b3 f8e8 h2h3 h7h6 f1d1 a8d8 c4e6 f7e6 e4e5 d6e5 d3g6 e7d6 c1h6 d7f7 g6f7 g8f7 h6e3 d6e7 c3a4 d8d1 a1d1 c5c4 b3c4 e8c8 a4b6 c8c7 g2g4 e7b4 e3d2 b4c5 b6a4 f6e4 d2e1 c5d4 d1b1 c7c4');
INSERT INTO evaluations VALUES(14141,30,'cp 34',NULL);
INSERT INTO evaluations VALUES(14141,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(14141,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(14141,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(14141,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(14141,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(14141,36,'cp 28','e2e3 e7e6 c2c3 f8d6 b1d2 f6d7 f1b5 e8g8 e1g1 f8e8 f4d6 c7d6 e3e4 d8b6 d1b3 a7a6 b5d3 d5e4 d2e4 e8d8 b3c2 d6d5 e4g3 e6e5 d4e5 d7e5 f3e5 c6e5 d3h7 g8h8 f1e1 e5c4 b2b3 c4a3 c2b2 h8h7 b2a3 b6g6 g1h2');
INSERT INTO evaluations VALUES(15026,30,'cp 12',NULL);
INSERT INTO evaluations VALUES(15026,31,'cp 19',NULL);
INSERT INTO evaluations VALUES(15026,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(15026,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(15026,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(15026,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(15026,36,'cp 20','c4c5 f8e7 f1b5 e8g8 h2h3 f6e4 b1c3 c8f5 e1g1 e7f6 c1f4 c6e7 b5d3 e7g6 f4h2 f8e8 c3b5 e8e7 d1c2 a7a6 b5c3 e4c3 b2c3 f5d3 c2d3 b7b6 c5b6 c7b6 f1b1 e7e6 a2a4 g6h4 f3h4 f6h4 h2e5 h4f6 b1e1 d8e7 e1e2 e7b7 e5f6 e6f6 a1e1 b7c6');
INSERT INTO evaluations VALUES(16163,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(16163,31,'cp -32',NULL);
INSERT INTO evaluations VALUES(16163,32,'cp -43',NULL);
INSERT INTO evaluations VALUES(16163,33,'cp -44',NULL);
INSERT INTO evaluations VALUES(16163,34,'cp -39',NULL);
INSERT INTO evaluations VALUES(16163,35,'cp -40',NULL);
INSERT INTO evaluations VALUES(16163,36,'cp -28','f8b4 f3d2 d7d5 c3d5 f6d5 e4d5 b4d2 d1d2 d8d5 f2f3 c8e6 f1e2 e8c8 e1g1 f7f5 c2c3 g7g5 b2b4 h8g8 e3f2 f5f4 d3d4 e5d4 b4b5 c6e5 d2d4 g5g4 f3g4 e6g4 e2g4 g8g4 d4d5 d8d5 h2h3 g4g8 f2d4 f4f3 g2g4 h6h5 a1e1 e5c4 f1f3 d5b5 f3f7 h5g4');
INSERT INTO evaluations VALUES(9832,30,'cp -31',NULL);
INSERT INTO evaluations VALUES(9832,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(9832,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(9832,33,'cp -37',NULL);
INSERT INTO evaluations VALUES(9832,34,'cp -27',NULL);
INSERT INTO evaluations VALUES(9832,35,'cp -25',NULL);
INSERT INTO evaluations VALUES(9832,36,'cp -26','f3d2 d7d5 c3d5 f6d5 e4d5 b4d2 d1d2 d8d5 f2f3 e8g8 f1e2 f8d8 e1g1 c8f5 a2a4 d8e8 f1e1 a7a5 c2c3 a8d8 e3f2 f5d3 a1d1 d3c4 d2d5 c4d5 e2b5 f7f6 f3f4 g8f7 c3c4 d5e6 b5c6 b7c6 d1d8 e8d8 f4e5 f6e5 f2g3 d8d4 g3e5 d4c4 e5c7 c4a4');
INSERT INTO evaluations VALUES(20414,30,'cp -37',NULL);
INSERT INTO evaluations VALUES(20414,31,'cp -50',NULL);
INSERT INTO evaluations VALUES(20414,32,'cp -29',NULL);
INSERT INTO evaluations VALUES(20414,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(20414,34,'cp -43',NULL);
INSERT INTO evaluations VALUES(20414,35,'cp -29',NULL);
INSERT INTO evaluations VALUES(20414,36,'cp -24','c8d7 c1e3 d8b6 b1c3 f8c5 c3a4 b6a5 a4c3 c6d4 e3d4 g8e7 d4c5 a5c5 d1d2 e8g8 f1d3 a8c8 h1f1 c5b6 e1c1 d5d4 c3e4 e7f5 f1f3 f8d8 f3h3 h7h6 d2e2 d7c6 d1g1 c6b5 c1b1 b5d3 h3d3 f5e3 g1c1 e3c4 d3b3 b6c6 g2g4 b7b5 g4g5 c4e3 e4d6 h6g5 d6c8');
INSERT INTO evaluations VALUES(20645,30,'cp 266',NULL);
INSERT INTO evaluations VALUES(20645,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(20645,32,'cp 242',NULL);
INSERT INTO evaluations VALUES(20645,33,'cp 244',NULL);
INSERT INTO evaluations VALUES(20645,34,'cp 259',NULL);
INSERT INTO evaluations VALUES(20645,35,'cp 241',NULL);
INSERT INTO evaluations VALUES(20645,36,'cp 245','g1e2 g8f6 b1c3 f8e7 e2f4 e8g8 f1e2 c8e6 c1e3 a8c8 d1d2 a7a6 e1g1 c6e5 a1d1 b7b5 a2a3 d8c7 g1h1 c7b7 f2f3 f8e8 f1e1 e5g6 e3d4 g6f4 d2f4 h7h6 f4e3 e6c4 e2c4 c8c4 e3d3 e8d8 b2b3 c4c8 e1e2 d6d5 e4e5 f6h5');
INSERT INTO evaluations VALUES(22288,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(22288,31,'cp 117',NULL);
INSERT INTO evaluations VALUES(22288,32,'cp 112',NULL);
INSERT INTO evaluations VALUES(22288,33,'cp 108',NULL);
INSERT INTO evaluations VALUES(22288,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(22288,35,'cp 127',NULL);
INSERT INTO evaluations VALUES(22288,36,'cp 120','a7a6 b5d3 c6e7 c2c4 g7g6 b1c3 f8g7 h2h4 f6f5 h4h5 f5f4 h5g6 e7g6 c1d2 g8f6 d1b3 e8g8 e1c1 f6g4 d1f1 a8b8 a2a4 c8d7 c1b1 c7c5 d3e2 b7b6 b3c2 g8h8 f3e1 b8b7 b1c1 f8g8 f2f3 g4e3 d2e3 f4e3');
INSERT INTO evaluations VALUES(17829,30,'cp 137',NULL);
INSERT INTO evaluations VALUES(17829,31,'cp 124',NULL);
INSERT INTO evaluations VALUES(17829,32,'cp 137',NULL);
INSERT INTO evaluations VALUES(17829,33,'cp 107',NULL);
INSERT INTO evaluations VALUES(17829,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(17829,35,'cp 120',NULL);
INSERT INTO evaluations VALUES(17829,36,'cp 109','b5d3 c6e7 c2c4 g7g6 b1c3 f8h6 c1d2 h6d2 d1d2 h7h5 e1c1 g8h6 h2h3 h6f7 f3h4 c8d7 g2g3 b7b5 f2f4 d8b8 d1f1 b5b4 c3d1 b8b6 c1b1 a6a5 d1e3 a5a4 f1c1 e8g8 c4c5 d6c5 e3c4 b6a7 h4f3 f7d6 f4e5 f6e5 f3g5 d6c4 d3c4 g8g7 c1f1');
INSERT INTO evaluations VALUES(25727,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(25727,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(25727,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(25727,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(25727,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(25727,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(25727,36,'cp 0','d8f6 e1g1 f6g6 g1h1 g6e4 a4c2 e4d5 f3g5 g8f6 c2b3 d5d3 b3f7 e8e7 f7b3 h7h6 b3c2 d3d5 c2b3');
INSERT INTO evaluations VALUES(29885,30,'cp 290',NULL);
INSERT INTO evaluations VALUES(29885,31,'cp 305',NULL);
INSERT INTO evaluations VALUES(29885,32,'cp 301',NULL);
INSERT INTO evaluations VALUES(29885,33,'cp 322',NULL);
INSERT INTO evaluations VALUES(29885,34,'cp 316',NULL);
INSERT INTO evaluations VALUES(29885,35,'cp 336',NULL);
INSERT INTO evaluations VALUES(29885,36,'cp 342','c4f7 e8e7 f2f3 h7h6 f7c4 g4h5 g5e6 d8c8 d2d4 h5f7 e6f8 h8f8 c4e2 e5d4 d1d4 b8c6 d4f2 f7e6 f2g3 e7f7 e1g1 f6h5 g3h4 h5f6 b2b3 f7g8 b1c3 c6b4 e4e5 f6d5 c3d5 b4d5 h4g3 g8h8 c2c4 d5b4 e5d6 c7d6 c1e3');
INSERT INTO evaluations VALUES(30719,30,'cp 273',NULL);
INSERT INTO evaluations VALUES(30719,31,'cp 255',NULL);
INSERT INTO evaluations VALUES(30719,32,'cp 267',NULL);
INSERT INTO evaluations VALUES(30719,33,'cp 276',NULL);
INSERT INTO evaluations VALUES(30719,34,'cp 242',NULL);
INSERT INTO evaluations VALUES(30719,35,'cp 249',NULL);
INSERT INTO evaluations VALUES(30719,36,'cp 272','f3e5 d7d6 e5g4 h7h5 g4e3 g7g6 f1c4 b8c6 a2a4 f8g7 e1g1 c6d4 d2d3 c7c6 c3e2 d4e2 d1e2 g8h6 c4b3 e8g8 a4a5 f8e8 e2f3 b6b5 c2c3 a7a6 c1d2 d8d7 a1e1 a8d8 h2h3 d6d5 e4d5 c6d5 d3d4 e8e4 b3c2 e4e8 e3d1 b7a8 d2g5 d8c8');
INSERT INTO evaluations VALUES(30714,30,'cp 246',NULL);
INSERT INTO evaluations VALUES(30714,31,'cp 259',NULL);
INSERT INTO evaluations VALUES(30714,32,'cp 253',NULL);
INSERT INTO evaluations VALUES(30714,33,'cp 262',NULL);
INSERT INTO evaluations VALUES(30714,34,'cp 273',NULL);
INSERT INTO evaluations VALUES(30714,35,'cp 255',NULL);
INSERT INTO evaluations VALUES(30714,36,'cp 240','d7d6 e5g4 h7h5 g4e3 g7g6 f1c4 b8c6 a2a4 f8g7 e1g1 g8e7 d2d3 c6b4 a4a5 e8g8 a5a6 b7c6 c1d2 d6d5 c4b3 d5d4 c3e2 b4a6 a1a6 d4e3 f2e3 c6b7 a6a2 c7c5 d1e1 e7c6 b3d5 d8d7 d2c3 c6e5 d5b7 d7b7 e2f4 a8e8 a2a1 b6b5');
INSERT INTO evaluations VALUES(31170,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(31170,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(31170,32,'cp 16',NULL);
INSERT INTO evaluations VALUES(31170,33,'cp 22',NULL);
INSERT INTO evaluations VALUES(31170,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(31170,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(31170,36,'cp 10','e7e6 d2b3 b8d7 g1f3 c6c5 d4c5 f8c5 b3c5 d7c5 c1e3 c5d7 f3d4 g8e7 f2f4 f5e4 f1d3 e8g8 e1g1 a8c8 d1e2 a7a6 d3e4 d5e4 d4b3 e7f5 c2c3 f5e3 e2e3 f7f5 e5f6 d7f6 a1e1 d8d5 g2g4 f6g4 e3e4 c8c4 e4e6 d5e6 e1e6 c4f4');
INSERT INTO evaluations VALUES(31545,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(31545,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(31545,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(31545,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(31545,34,'cp 25',NULL);
INSERT INTO evaluations VALUES(31545,35,'cp 19',NULL);
INSERT INTO evaluations VALUES(31545,36,'cp 26','d2b3 b8d7 g1f3 c6c5 d4c5 f8c5 b3c5 d7c5 f3d4 f5g6 d4b5 g8e7 b5d6 e8f8 f1e2 c5e4 c1f4 e7c6 d6e4 g6e4 e1g1 h7h6 c2c4 g7g5 f4e3 d5d4 e3d2 d8b6 e2h5 e4d3 f1e1 a8d8 f2f4 d3c4 f4g5 h6g5 d2g5 d8d5 b2b3 c4a6 d1f3 c6e5');
INSERT INTO evaluations VALUES(31379,30,'cp 31',NULL);
INSERT INTO evaluations VALUES(31379,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(31379,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(31379,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(31379,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(31379,35,'cp 30',NULL);
INSERT INTO evaluations VALUES(31379,36,'cp 19','f1d3 f5d3 d1d3 g8f6 e1g1 f8d6 c1d2 e8g8 e3e4 d5e4 c3e4 f6e4 d3e4 b8d7 h2h3 a7a5 f1d1 f8e8 d1e1 d7f6 e4e2 d8c7 a2a4 c6c5 d2c3 c5d4 f3d4 d6b4 d4b5 c7e7 c3b4 a5b4 b2b3 e8d8 a1d1 d8d1 e1d1 a8d8 d1d8 e7d8');
INSERT INTO evaluations VALUES(31418,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(31418,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(31418,32,'cp -19',NULL);
INSERT INTO evaluations VALUES(31418,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31418,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31418,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(31418,36,'cp 0','b8d7 e5g6 h7g6 c2c3 g8f6 d1f3 c6c5 c1e3 c5d4 e3d4 f8c5 e1c1 d8a5 d4f6 d7f6 b2b4 a5a2 f1b5 e8f8 b4c5 a2a3 c1c2 a3a2');
INSERT INTO evaluations VALUES(33015,30,'cp 49',NULL);
INSERT INTO evaluations VALUES(33015,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(33015,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(33015,33,'cp 53',NULL);
INSERT INTO evaluations VALUES(33015,34,'cp 52',NULL);
INSERT INTO evaluations VALUES(33015,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(33015,36,'cp 57','a5c5 b1c3 b8c6 f1e2 g8f6 e1g1 d7d6 h2h3 c5b6 c3d5 f6d5 e4d5 c6d4 b2b3 e8g8 a1c1 d4e2 d1e2 e7e5 d5e6 c8e6 f1d1 d6d5 d2e3 b6a5 e3d4 g7d4 f3d4 f8e8 e2e3 a8c8 e3g5 a5d8 g5d8 e8d8 d1d2 c8c5 c1d1 c5c3 f2f3 d8c8 g1f2 a7a6');
INSERT INTO evaluations VALUES(33011,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(33011,31,'cp 68',NULL);
INSERT INTO evaluations VALUES(33011,32,'cp 63',NULL);
INSERT INTO evaluations VALUES(33011,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(33011,34,'cp 58',NULL);
INSERT INTO evaluations VALUES(33011,35,'cp 53',NULL);
INSERT INTO evaluations VALUES(33011,36,'cp 62','b1c3 g8f6 d2e3 c5c7 f1e2 b8c6 d1d2 c7a5 h2h3 e8g8 e1g1 d7d6 a2a3 c8d7 b2b4 a5d8 a1d1 a8c8 e2d3 d7e6 f3g5 d6d5 g5e6 f7e6 e4d5 e6d5 c3b5 a7a6 b5d4 c6d4 e3d4 f6h5 d4c5 h5f4 f1e1 f8f7 d3f1 e7e5 c5e3');
INSERT INTO evaluations VALUES(33455,30,'cp 254',NULL);
INSERT INTO evaluations VALUES(33455,31,'cp 240',NULL);
INSERT INTO evaluations VALUES(33455,32,'cp 268',NULL);
INSERT INTO evaluations VALUES(33455,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(33455,34,'cp 271',NULL);
INSERT INTO evaluations VALUES(33455,35,'cp 272',NULL);
INSERT INTO evaluations VALUES(33455,36,'cp 246','b1c3 c5d4 g2g3 h4g5 e5c4 b8c6 f1g2 d7d6 h2h3 g5c5 c4e3 c8e6 e1g1 e8g8 d2d3 d4c3 b2c3 c5c3 a1b1 c3c5 c1b2 c6d4 f3d1 e6a2 b2d4 c5d4 b1b7 f8b8 b7c7 f6e8 c7c6 a7a5 d1a1 d4a1 f1a1 a2e6 c6c3 e6d7 f2f4 a5a4 g1f2 a8a6 f2e2 a6b6 e2d1 e8f6 c3a3 b6b1 a1b1 b8b1 d1d2');
INSERT INTO evaluations VALUES(33820,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(33820,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(33820,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(33820,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(33820,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(33820,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(33820,36,'cp -8','e4f6 e2e7 f8e7 b2b3 b8c6 c1b2 d6d5 b1c3 c8f5 c3b5 e8g8 e1c1 a7a6 b5d4 c6d4 b2d4 f8e8 g2g4 f5d3 f1d3 f6d7 d1e1 d7c5 d4c5 e7c5 e1e8 a8e8 f3f4 c5d6 f4f5 g7g5 c1d1 g8g7 h2h4 h7h6 c2c3 g7f6 h4g5 f6g5 d3e2 d6f4 e2f3 c7c6 h1h5 g5f6 d2d4');
INSERT INTO evaluations VALUES(33797,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(33797,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(33797,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(33797,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(33797,34,'cp -12',NULL);
INSERT INTO evaluations VALUES(33797,35,'cp -4',NULL);
INSERT INTO evaluations VALUES(33797,36,'cp -5','e2e7 f8e7 b2b3 e8g8 c1b2 d6d5 b2a3 f8e8 a3e7 e8e7 e1f2 c8f5 g2g4 f5d3 f1d3 b8c6 b1c3 c6b4 h1e1 e7e1 a1e1 b4d3 c2d3 g8f8 d3d4 c7c6 b3b4 f6e8 b4b5 c6b5 c3b5 a7a6 b5c3 e8c7 e1b1 b7b5 h2h4 c7e6 c3d5');
INSERT INTO evaluations VALUES(33630,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(33630,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(33630,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(33630,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(33630,34,'cp -9',NULL);
INSERT INTO evaluations VALUES(33630,35,'cp -10',NULL);
INSERT INTO evaluations VALUES(33630,36,'cp -8','f8e7 b2b3 e8g8 c1b2 d6d5 b2a3 f8e8 a3e7 e8e7 e1f2 h7h5 b1c3 d5d4 c3e2 b8c6 e2f4 c8f5 h2h4 a7a5 a2a3 g7g6 f1e2 b7b6 h1e1 g8f8 e2f1 a8d8 g2g3 f6d5 e1e7 c6e7 f1h3 f5h3 f4h3 f7f6 b3b4 e7c6 h3f4 f8f7');
INSERT INTO evaluations VALUES(33379,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(33379,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(33379,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(33379,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(33379,34,'cp -11',NULL);
INSERT INTO evaluations VALUES(33379,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(33379,36,'cp 0','b2b3 e8g8 c1b2 d6d5 b2a3 f8e8 a3e7 h7h5 b1c3 e8e7 e1f2 d5d4 c3e2 b8c6 h2h4 c8f5 e2f4 a7a5 a2a3 g7g6 f1e2 g8f8 a1c1 a8d8 h1e1 b7b6 g2g3 f8g7 c1a1 d8d6 a1c1');
INSERT INTO evaluations VALUES(34479,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(34479,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(34479,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(34479,33,'cp 21',NULL);
INSERT INTO evaluations VALUES(34479,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(34479,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(34479,36,'cp 18','b1a3 c8d7 a3b5 d7b5 f1b5 b8c6 g1f3 c5d4 b5c6 d5c6 f3d4 c6a6 a2a4 g8f6 d1e2 a6e2 e1e2 a7a6 h1d1 e8c8 a4a5 e6e5 d4f3 d8d1 e2d1 f8d6 c1e3 h7h6 f3d2 f6d5 d1c2 c8d7 d2c4 d7e6 b2b4 h8c8 c2b3 f7f5 a1e1 h6h5');
INSERT INTO evaluations VALUES(34767,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(34767,31,'cp 3',NULL);
INSERT INTO evaluations VALUES(34767,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(34767,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(34767,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(34767,35,'cp 5',NULL);
INSERT INTO evaluations VALUES(34767,36,'cp 0','b8c6 c1e3 f6g6 h5g6 h7g6 b1c3 f8e7 g1f3 c8g4 f3g5 e7g5 e3g5 c6d4 c3b5 d4b5 c4b5 g4d7 b5c4 f7f6 g5e3 g6g5 a2a4 a7a5 f2f3 g8e7 c2c3 d7c6 c4e6 c6d7');
INSERT INTO evaluations VALUES(36274,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36274,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(36274,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(36274,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(36274,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(36274,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(36274,36,'cp 12','f1e1 d6d5 d3d4 c8e6 e2d3 f8e8 c2c3 c7c5 b1d2 e6g4 d4c5 e7c5 e1e8 d8e8 h2h3 g4h5 d2f1 b8c6 c1e3 c5e3 f1e3 a8d8 d3e2 d5d4 f3d4 h5e2 d1e2 c6d4 c3d4 d8d4 e2f3 d4d7 a1d1 d7d1 f3d1 h7h5 d1f3 e8c8 b2b4 h5h4 e3f1 c8e6 f3b7');
INSERT INTO evaluations VALUES(38507,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(38507,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(38507,32,'cp 52',NULL);
INSERT INTO evaluations VALUES(38507,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(38507,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(38507,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(38507,36,'cp 48','c3e4 g8f6 e4f6 e7f6 f1c4 e8g8 e1g1 b8c6 c2c3 b7b6 c1f4 c6e7 f1e1 c7c6 f4d2 c8f5 f3h4 f5c8 h2h3 e7d5 c4d3 f8e8 c3c4 e8e1 d1e1 d5c7 h4f3 c8e6 a1d1 d8d7 d2e3 a8e8 e1d2 g7f8 b2b3 h7h5 a2a4 a7a5 e3h6');
INSERT INTO evaluations VALUES(41089,30,'cp -25',NULL);
INSERT INTO evaluations VALUES(41089,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(41089,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(41089,33,'cp -24',NULL);
INSERT INTO evaluations VALUES(41089,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(41089,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(41089,36,'cp -13','e8g8 e1g1 c7c6 c1e3 a7a5 h2h3 c8f5 g2g4 f5g6 f3e5 f6e4 e5g6 h7g6 c3e4 d5e4 d1d2 b8a6 f2f3 e4f3 e2f3 f8e8 a1d1 a6b4 d4d5 b4d5 f3d5 d8d5 d2d5 c6d5 d1d5 a8c8 c2c3 e7a3 e3c1 a3c5 g1g2 f7f6 d5d7 e8e2 g2f3 c8e8 d7b7 a5a4');
INSERT INTO evaluations VALUES(36404,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(36404,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(36404,32,'cp -10',NULL);
INSERT INTO evaluations VALUES(36404,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36404,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(36404,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(36404,36,'cp -12','e1g1 c7c6 c1e3 a7a5 a2a3 c8g4 h2h3 g4f3 e2f3 f8e8 b2b3 b8d7 c3e2 d7f8 c2c4 f8g6 e2c3 g6h4 c4d5 h4f3 d1f3 f6d5 c3d5 d8d5 f3d5 c6d5 a3a4 e7d6 f1e1 f7f5 g2g3 g8f7 g1g2 e8e6 e3f4 a8c8 e1e6 f7e6 f4d6 e6d6 a1e1 c8c2 e1e8 c2d2 e8a8 d2d4 a8a5');
INSERT INTO evaluations VALUES(43763,30,'cp 114',NULL);
INSERT INTO evaluations VALUES(43763,31,'cp 131',NULL);
INSERT INTO evaluations VALUES(43763,32,'cp 134',NULL);
INSERT INTO evaluations VALUES(43763,33,'cp 118',NULL);
INSERT INTO evaluations VALUES(43763,34,'cp 119',NULL);
INSERT INTO evaluations VALUES(43763,35,'cp 137',NULL);
INSERT INTO evaluations VALUES(43763,36,'cp 138','e4e5 d6e5 d1f3 e5e4 c3e4 c7c6 f1d3 g7g6 e4f6 e7f6 g1e2 f8g7 c1e3 e8g8 e2f4 f6f5 h2h4 g7d4 e1c1 d4e3 f3e3 d8f6 e3g3 g8h8 f4h3 h7h6 d3c4 b8d7 h1e1 d7c5 d1d6 f6g7 c4f7 g7f7 d6g6 h8h7 h4h5 c5e6 g3d6 f5f4 g6e6 c8e6 e1e6 f7h5');
INSERT INTO evaluations VALUES(43866,30,'cp 77',NULL);
INSERT INTO evaluations VALUES(43866,31,'cp 74',NULL);
INSERT INTO evaluations VALUES(43866,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(43866,33,'cp 68',NULL);
INSERT INTO evaluations VALUES(43866,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(43866,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(43866,36,'cp 34','d2d4 c8b7 d1c2 d7d5 c4d5 f6d5 e2e4 d5c3 b2c3 e7e6 g1f3 f8e7 f1b5 c7c6 b5d3 c6c5 c1f4 e8g8 d4d5 e6d5 e4d5 d8d5 e1c1 d5h5 d3e4 b7c6 h1e1 c5c4 e4h7 h5h7 c2h7 g8h7 e1e7 c6a4 d1d6 b8c6 e7c7 a8c8 d6d7 c8c7 d7c7 f7f6 f4e3 f8d8 f3d4 c6d4 c3d4 d8d7 c7c4 a4b5 c4c8 b5f1 g2g3 h7g6 c8b8 f1g2 c1b2 g2d5');
INSERT INTO evaluations VALUES(44252,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(44252,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(44252,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(44252,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(44252,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(44252,35,'cp 8',NULL);
INSERT INTO evaluations VALUES(44252,36,'cp 11','e4e5 f6d5 g2g3 d7d6 f1g2 f8e7 e1g1 b8c6 d2d3 e8g8 c2c4 d5c7 c1e3 d6d5 e3f2 b7b5 b2b3 a8b8 d1e2 b5c4 b3c4 c8a6 b1d2 b8b2 f1b1 d8b8 g2f1 b8b4 e2d1 b2b1 d1b1 f8b8 b1b4 b8b4 c4d5 c7d5 d2b3 b4b8 f3d2');
INSERT INTO evaluations VALUES(44311,30,'cp 95',NULL);
INSERT INTO evaluations VALUES(44311,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(44311,32,'cp 89',NULL);
INSERT INTO evaluations VALUES(44311,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(44311,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(44311,35,'cp 100',NULL);
INSERT INTO evaluations VALUES(44311,36,'cp 85','d7d6 g1e2 a7a6 a2a4 b8c6 e1g1 f8e7 f2f4 c8d7 e2g3 e5f4 c1f4 e8g8 d1d2 c6d4 a4a5 d4c6 c3d5 f6d5 e4d5 c6e5 f4e5 d6e5 g3e4 d8c7 d2e2 e7d6 e2h5 f7f6 g2g4 d7e8 h5h3 e8g6 f1f2 g8h8 g1h1 a8b8 e4g3');
INSERT INTO evaluations VALUES(45902,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(45902,31,'cp 88',NULL);
INSERT INTO evaluations VALUES(45902,32,'cp 92',NULL);
INSERT INTO evaluations VALUES(45902,33,'cp 92',NULL);
INSERT INTO evaluations VALUES(45902,34,'cp 97',NULL);
INSERT INTO evaluations VALUES(45902,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(45902,36,'cp 102','g1e2 a7a6 a2a4 f8e7 f2f4 b8c6 e1g1 c8d7 e2g3 e5f4 c1f4 e8g8 d1d2 c6d4 a4a5 d4c6 c3d5 f6d5 e4d5 c6e5 f4e5 d6e5 g3e4 d8c7 a1e1 g8h8 d5d6 e7d6 c4f7 d6e7 f7d5 h7h6 d2c3 f8f1 e1f1 a8d8 e4d2 d7c6 d5c6 c7c6 f1f7 e7g5 c3e5 g5f6 e5c7 c6c7 f7c7 f6b2 c7b7 b2c3 d2e4 c3a5 e4c5');
INSERT INTO evaluations VALUES(45644,30,'cp 27',NULL);
INSERT INTO evaluations VALUES(45644,31,'cp 18',NULL);
INSERT INTO evaluations VALUES(45644,32,'cp 22',NULL);
INSERT INTO evaluations VALUES(45644,33,'cp 28',NULL);
INSERT INTO evaluations VALUES(45644,34,'cp 28',NULL);
INSERT INTO evaluations VALUES(45644,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(45644,36,'cp 16','f8e7 b2b3 e8g8 c1b2 b8d7 f1d3 d5c4 d2c4 b7b5 c4e5 d7e5 d4e5 f6d5 e1g1 d5b4 d3e4 a8b8 d1d8 f8d8 b2a3 c8b7 f1c1 a7a5 a3b4 e7b4 a2a3 b4f8 e4c6 b7c6 c1c6 f8a3 g1f1 a3b4 f3d4 d8d5 c6c7 d5e5 a1c1 e5d5 d4c6 b8f8 c6b4 a5b4 c7b7 h7h6 c1c7 d5d1 f1e2 d1g1 e2f3');
INSERT INTO evaluations VALUES(45838,30,'cp 66',NULL);
INSERT INTO evaluations VALUES(45838,31,'cp 72',NULL);
INSERT INTO evaluations VALUES(45838,32,'cp 75',NULL);
INSERT INTO evaluations VALUES(45838,33,'cp 69',NULL);
INSERT INTO evaluations VALUES(45838,34,'cp 60',NULL);
INSERT INTO evaluations VALUES(45838,35,'cp 68',NULL);
INSERT INTO evaluations VALUES(45838,36,'cp 76','g1f3 f8d6 d2d4 b8d7 d1c2 e8g8 e1g1 f8e8 a2a4 a7a5 e2e4 d8c7 c1e3 b7b5 d4e5 d6e5 f3d4 b5a4 c2a4 c6c5 f1d1 h7h6 d4b5 c7c6 b5d6 c6a4 a1a4 e8d8 d6c8 a8c8 g2h3 d7b6 d1d8 c8d8 a4a5 e5c3 a5c5 c3d4 e3d4 d8d4 c5b5 d4d1 h3f1');
INSERT INTO evaluations VALUES(45837,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(45837,31,'cp 55',NULL);
INSERT INTO evaluations VALUES(45837,32,'cp 59',NULL);
INSERT INTO evaluations VALUES(45837,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(45837,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(45837,35,'cp 44',NULL);
INSERT INTO evaluations VALUES(45837,36,'cp 56','f8d6 d2d4 b8d7 e1g1 e8g8 e2e4 f8e8 d1c2 d8c7 a2a4 b7b5 c1e3 a7a5 d4e5 d6e5 f1d1 c8a6 g2h3 d7f6 f3e5 c7e5 f2f3 e5c7 e3f4 c7b7 c3c4 b5b4 c4c5 h7h6 d1d6 b4b3 c2c3 a8b8 a1b1 b7b4 f4e5 a6e2 c3b4 b8b4 e5c3');
INSERT INTO evaluations VALUES(48571,30,'cp 157',NULL);
INSERT INTO evaluations VALUES(48571,31,'cp 144',NULL);
INSERT INTO evaluations VALUES(48571,32,'cp 146',NULL);
INSERT INTO evaluations VALUES(48571,33,'cp 141',NULL);
INSERT INTO evaluations VALUES(48571,34,'cp 142',NULL);
INSERT INTO evaluations VALUES(48571,35,'cp 144',NULL);
INSERT INTO evaluations VALUES(48571,36,'cp 157','f3g5 f6d5 f1c4 d8g5 c4d5 g5g2 d5e4 g2h3 d1e2 h3e6 d2d4 f7f5 e4f3 e6e2 c3e2 f8d6 c1f4 e8f7 e1c1 b8d7 f3d5 f7f8 d5b3 d7f6 f4e5 c8d7 h1g1 a8e8 f2f4 e8e7 e2c3 d7c6 a2a4 f6g4 h2h3 d6e5 f4e5 g4e3 d1d3 f5f4 d4d5 c6d7 e5e6 d7e8');
INSERT INTO evaluations VALUES(48927,30,'cp -26',NULL);
INSERT INTO evaluations VALUES(48927,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(48927,32,'cp -26',NULL);
INSERT INTO evaluations VALUES(48927,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(48927,34,'cp -29',NULL);
INSERT INTO evaluations VALUES(48927,35,'cp -18',NULL);
INSERT INTO evaluations VALUES(48927,36,'cp -22','g2g3 c7c5 f2f4 b8c6 g1f3 f8d6 f1g2 e8g8 e1g1 b7b5 f4e5 c6e5 c1g5 a8b8 b2b3 a7a6 f3e5 d6e5 e2f4 h7h6 g5f6 d8f6 g2h3 c8h3 f4h3 f6e6 h3f4 e5f4 f1f4 b8c8 d1d2 e6b6 a1f1 c5c4 d2f2 f7f6 d3c4 f8e8 c4b5 a6b5 a2a3 c8c3');
INSERT INTO evaluations VALUES(48996,30,'cp -28',NULL);
INSERT INTO evaluations VALUES(48996,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(48996,32,'cp -27',NULL);
INSERT INTO evaluations VALUES(48996,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(48996,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(48996,35,'cp -14',NULL);
INSERT INTO evaluations VALUES(48996,36,'cp -7','f7f6 b1c3 d5c3 c4f7 e8d7 f7e6 d7e7 d1g4 b8a6 b2c3 h7h6 e6c8 d8c8 g4c8 a8c8 g5e4 e7e6 f2f4 a6c5 e4c5 f8c5 f4e5 f6e5 e1e2 c8f8 c1e3 c5e3 e2e3 f8f6 a2a4 h8f8 h1e1 a7a5 a1b1 b7b6 b1b5 f6f1 e1f1 f8f1 c3c4 f1e1 e3d2 e1h1 h2h3 h1h2 c4c5 h2g2 d2c3 b6c5 b5c5');
INSERT INTO evaluations VALUES(49388,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(49388,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(49388,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49388,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49388,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(49388,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(49388,36,'cp 0','d7d5 c4d5 e6d5 d2d4 c7c6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(48244,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(48244,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(48244,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(48244,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(48244,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(48244,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(48244,36,'cp 0','c4d5 e6d5 d2d4 c7c6 c1f4 f6h5 f4d2 h5f6');
INSERT INTO evaluations VALUES(49826,30,'cp 3',NULL);
INSERT INTO evaluations VALUES(49826,31,'cp 1',NULL);
INSERT INTO evaluations VALUES(49826,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(49826,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(49826,34,'cp -4',NULL);
INSERT INTO evaluations VALUES(49826,35,'cp 1',NULL);
INSERT INTO evaluations VALUES(49826,36,'cp 1','d8e7 d1e2 d7d5 d2d4 e4c3 b2c3 b8d7 e5d7 c8d7 e2e7 f8e7 c1f4 e8c8 e1d2 g7g5 a1e1 d8e8 f4e5 h8f8 e5g3 h7h6 f1d3 g5g4 g3f4 e7g5 f4e3 a7a5 h2h4 g4h3 g2h3 f8g8 c3c4 d5c4 d3c4 d7e6 c4e6 e8e6 d2d3 b7b5 e3g5 g8g5 e1e6');
INSERT INTO evaluations VALUES(51509,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51509,31,'cp 8',NULL);
INSERT INTO evaluations VALUES(51509,32,'cp -6',NULL);
INSERT INTO evaluations VALUES(51509,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51509,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51509,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51509,36,'cp -16','f1g2 b8c6 c2c4 e7e6 g1f3 d7d5 b1c3 c8b7 c4d5 e6d5 d2d4 g8f6 c1f4 f6e4 e1g1 e4c3 b2c3 f8d6 f4g3 d8d7 f3d2 d6g3 h2g3 d7g4 g2d5 h7h5 e2e4 g4g6 d1f3 h5h4 f3f5 g6f5 e4f5 e8c8 d5f7 c6e5 f7b3 h4g3 b3e6 c8b8 f2g3 e5g4');
INSERT INTO evaluations VALUES(51510,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(51510,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(51510,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(51510,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(51510,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(51510,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(51510,36,'cp 0','b8c6 c2c4 e7e6 g1f3 d7d5 b1c3 c8b7 c4d5 e6d5 d2d4 d8d7 g4g5 f8d6 h2h4 c6d8 d1d3 g8e7 c3b5 d6b4 b5c3');
INSERT INTO evaluations VALUES(53437,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(53437,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(53437,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(53437,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(53437,34,'cp -1',NULL);
INSERT INTO evaluations VALUES(53437,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(53437,36,'cp 0','b8c6 d2d4 a7a6 e5c6 b7c6 f1e2 e7e6 c1f4 f8d6 f4d6 d8d6 e1g1 g8f6 b1d2 e8g8 c2c4 a6a5 f1e1 c6c5 c4d5 c5d4 d2c4 d6c5 d5e6 c8e6 a1c1 f8d8 e2f1 a5a4 d1d2 h7h6 c4e5 c5b6 e5c4');
INSERT INTO evaluations VALUES(53590,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(53590,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(53590,32,'cp 27',NULL);
INSERT INTO evaluations VALUES(53590,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(53590,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(53590,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(53590,36,'cp 31','c6c5 g1f3 g8f6 f3e5 f6d7 b1d2 d7e5 f4e5 b8d7 e5f4 d8b6 d1b3 d7f6 f1e2 f8e7 a2a4 e8g8 e1g1 c8d7 d4c5 b6b3 d2b3 f6e4 a4a5 e4c5 b3c5 e7c5 f4e5 f8c8 e5d4 f7f6 d4c5 c8c5 f1c1 a8c8');
INSERT INTO evaluations VALUES(53835,30,'cp 73',NULL);
INSERT INTO evaluations VALUES(53835,31,'cp 60',NULL);
INSERT INTO evaluations VALUES(53835,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(53835,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(53835,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(53835,35,'cp 67',NULL);
INSERT INTO evaluations VALUES(53835,36,'cp 66','d2d4 g8f6 b1c3 b8d7 a2a4 f8e7 f1c4 e8g8 e1g1 f8e8 f1e1 e7f8 a4a5 e5d4 d1d4 d8c7 c1g5 d7e5 c4e2 c8d7 d4d2 f8e7 e1d1 a8d8 f3e5 d6e5 d2e3 a7a6 e2c4 d7c8 e3g3 f6h5 d1d8 e8d8');
INSERT INTO evaluations VALUES(54740,30,'cp 242',NULL);
INSERT INTO evaluations VALUES(54740,31,'cp 222',NULL);
INSERT INTO evaluations VALUES(54740,32,'cp 224',NULL);
INSERT INTO evaluations VALUES(54740,33,'cp 202',NULL);
INSERT INTO evaluations VALUES(54740,34,'cp 226',NULL);
INSERT INTO evaluations VALUES(54740,35,'cp 249',NULL);
INSERT INTO evaluations VALUES(54740,36,'cp 247','g1f3 b8c6 f1c4 e7e5 e1g1 e5d4 e4e5 f6e5 d1b3 g8f6 c4f7 e8e7 f3e5 c6e5 f1e1 d8d6 c1f4 f6g4 b1d2 e7d8 d2f3 d6b6 f3e5 b6b3 f4g5 f8e7 e5c6 b7c6 g5e7 d8d7 a2b3 g4h6 e7c5 a7a5 f7e6 d7d8 c5d4 c8e6 e1e6 d8d7 e6e1 h8e8 d4g7 e8e1 a1e1');
INSERT INTO evaluations VALUES(55350,30,'cp -59',NULL);
INSERT INTO evaluations VALUES(55350,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(55350,32,'cp -39',NULL);
INSERT INTO evaluations VALUES(55350,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(55350,34,'cp -61',NULL);
INSERT INTO evaluations VALUES(55350,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(55350,36,'cp -54','g5g4 f3g1 d8f6 b1c3 g8e7 g1e2 f8h6 d1d2 b8c6 c3b5 e8d8 d4d5 c6e5 e2f4 g4g3 h4h5 h8g8 f1e2 a7a6 b5c3 e5g4 e2g4 g8g4 h1f1 f6e5 d2e3 h6f4 f1f4 e5h5 f4f1 c8d7 c1d2 d7b5 f1g1 g4g8 a2a4 b5d7 c3e2 h5e5 e1c1 f7f5 g1h1 f5e4 h1h7 e7d5');
INSERT INTO evaluations VALUES(55359,30,'cp -60',NULL);
INSERT INTO evaluations VALUES(55359,31,'cp -24',NULL);
INSERT INTO evaluations VALUES(55359,32,'cp -36',NULL);
INSERT INTO evaluations VALUES(55359,33,'cp -35',NULL);
INSERT INTO evaluations VALUES(55359,34,'cp -33',NULL);
INSERT INTO evaluations VALUES(55359,35,'cp -36',NULL);
INSERT INTO evaluations VALUES(55359,36,'cp -21','f3g1 d8f6 b1c3 g8e7 g1e2 f8h6 d1d2 b8c6 c3b5 e8d8 d4d5 c6e5 d2c3 c7c5 d5c6 e7c6 c3d2 a7a6 d2d6 f6d6 b5d6 c6b4 e2d4 d8c7 c1d2 b4c2 d4c2 c7d6 f1e2 h8g8 h1f1 c8e6 c2d4 d6e7 d2f4 h6f4 f1f4 g8d8 a1d1 h7h5 d4f5 e7f6 f5d6 f6g6 d6b7 d8d1 e1d1 e6a2');
INSERT INTO evaluations VALUES(55360,30,'cp -42',NULL);
INSERT INTO evaluations VALUES(55360,31,'cp -14',NULL);
INSERT INTO evaluations VALUES(55360,32,'cp -49',NULL);
INSERT INTO evaluations VALUES(55360,33,'cp -22',NULL);
INSERT INTO evaluations VALUES(55360,34,'cp -35',NULL);
INSERT INTO evaluations VALUES(55360,35,'cp -12',NULL);
INSERT INTO evaluations VALUES(55360,36,'cp -19','d8f6 b1c3 g8e7 g1e2 f8h6 d1d2 e8g8 e2f4 b8c6 c3e2 e7g6 c2c3 f8e8 e1d1 h6f4 e2f4 f6f4 d2f4 g6f4 c1f4 e8e4 f4d2 e4e8 f1d3 c6e7 h1f1 e7d5 c3c4 d5e3 d2e3 e8e3 d1d2 e3g3 f1f2 c7c6 d4d5 c6d5 c4d5 c8d7 d3h7 g8h7 f2f7 h7h6 f7d7 g3g2 d2d3 a8f8 d7d6 h6h5 d6d7 g4g3 d7b7 g2f2 d5d6 g3g2 a1g1');
INSERT INTO evaluations VALUES(55621,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(55621,31,'cp 13',NULL);
INSERT INTO evaluations VALUES(55621,32,'cp 4',NULL);
INSERT INTO evaluations VALUES(55621,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(55621,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(55621,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(55621,36,'cp -7','f7f5 b1c3 f5e4 c3e4 g8f6 e4g3 f8e7 c2c3 e8g8 d1c2 c7c5 d2d4 c5d4 c3d4 b8c6 d4e5 d6e5 c1e3 g8h8 a1d1 d8a5 e3d2 a5c5 c2c5 e7c5 f1e2 c6d4 f3d4 c5d4 d2b4 f8d8 b4c5 c8e6 c5d4 e5d4 e1g1 a8c8');
INSERT INTO evaluations VALUES(55902,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(55902,31,'cp -52',NULL);
INSERT INTO evaluations VALUES(55902,32,'cp -52',NULL);
INSERT INTO evaluations VALUES(55902,33,'cp -61',NULL);
INSERT INTO evaluations VALUES(55902,34,'cp -68',NULL);
INSERT INTO evaluations VALUES(55902,35,'cp -58',NULL);
INSERT INTO evaluations VALUES(55902,36,'cp -52','b8c6 f3e2 g8e7 g1f3 e6e5 g2g3 d5d4 f1g2 a7a5 b1d2 c8g4 h2h3 g4e6 f3g5 e6d7 c2c3 e7g6 g5f3 f8c5 e1g1 d7e6 d2b3 c5b6 c3d4 e6b3 a2b3 c6d4 f3d4 d8d4 h3h4 g6e7 g1h2 d4d6 c1d2 e7c6 f2f4 e8c8 d2c3 f7f6 f4e5 c6e5 c3a5 d6d3');
INSERT INTO evaluations VALUES(852,30,'cp 39',NULL);
INSERT INTO evaluations VALUES(852,31,'cp 38',NULL);
INSERT INTO evaluations VALUES(852,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(852,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(852,34,'cp 35',NULL);
INSERT INTO evaluations VALUES(852,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(852,36,'cp 50','a1b1 c5d4 e3d4 f6h5 f4e3 b6c7 f3e5 g7g6 f1e2 h5g7 e5g4 f8e7 d2f3 g7f5 e3d2 h7h5 g4e3 f5d6 e1g1 d6e4 d2c1 b7b6 c3c4 c7d6 d1b3 e4f6 c4d5 e6d5 e2b5 e8g8 c1d2 c8d7 b1c1 f8c8 f1e1 c6a5 b3d3 d7b5 d3b5 a5c6 d2c3');
INSERT INTO evaluations VALUES(2132,30,'cp -109',NULL);
INSERT INTO evaluations VALUES(2132,31,'cp -101',NULL);
INSERT INTO evaluations VALUES(2132,32,'cp -104',NULL);
INSERT INTO evaluations VALUES(2132,33,'cp -90',NULL);
INSERT INTO evaluations VALUES(2132,34,'cp -93',NULL);
INSERT INTO evaluations VALUES(2132,35,'cp -80',NULL);
INSERT INTO evaluations VALUES(2132,36,'cp -82','f7f6 e5f6 g8f6 c2c4 d7d6 b1c3 c8f5 d2d3 e8c8 h2h3 h7h5 c1g5 e7e6 e1c1 c8b8 g2g3 a7a6 g5f4 e6e5 f4g5 c6d4 e2e3 d8e8 f1g2 c7d7 f3h4 f5h7 f2f4 e5f4 e3f4 f8e7 h4f3 d4e6 f4e3 e6g5 f3g5 h7f5 e3d2 h5h4');
INSERT INTO evaluations VALUES(2148,30,'cp 103',NULL);
INSERT INTO evaluations VALUES(2148,31,'cp 104',NULL);
INSERT INTO evaluations VALUES(2148,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(2148,33,'cp 97',NULL);
INSERT INTO evaluations VALUES(2148,34,'cp 103',NULL);
INSERT INTO evaluations VALUES(2148,35,'cp 99',NULL);
INSERT INTO evaluations VALUES(2148,36,'cp 101','f7e8 c2c3 c8f5 d2d3 f6g6 e1g1 a7a6 d5f3 a8d8 b1d2 f5d3 c4d3 g6d3 f3g4 d3g6 g4g6 h7g6 g2g3 f8e7 d2e4 d8d3 c1e3 e8d7 h2h4 b7b6 a1e1 h8d8 g1g2 c6a5 e3g5 a5c4 e1e2 d7e6 g5e7 e6e7 f1e1 d3d1 b2b3 d1e1 e2e1 c4d6');
INSERT INTO evaluations VALUES(5208,30,'cp 35',NULL);
INSERT INTO evaluations VALUES(5208,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(5208,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(5208,33,'cp 42',NULL);
INSERT INTO evaluations VALUES(5208,34,'cp 26',NULL);
INSERT INTO evaluations VALUES(5208,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(5208,36,'cp 31','c1e3 c5b4 b1c3 g8f6 d1b3 d8a5 f1d3 d7d5 e4d5 f6d5 d3c4 b4c3 b2c3 a5c3 b3c3 d5c3 a1c1 c3e4 d4c6 b7c6 c4b3 e8g8 f2f3 e4f6 c1c6 f8e8 e1f2 c8e6 b3e6 e8e6 h1c1 f6d5 e3d2 e6e8 c6c5 e8d8 h2h4 f7f6 g2g4 d8d7 c1c2 g8f7 h4h5 a8b8 f2g3 b8b1');
INSERT INTO evaluations VALUES(5207,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(5207,31,'cp 25',NULL);
INSERT INTO evaluations VALUES(5207,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(5207,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(5207,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(5207,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(5207,36,'cp 31','c5b4 b1c3 g8f6 d1b3 d8a5 f1d3 d7d5 e4d5 f6d5 d3c4 b4c3 b2c3 a5c3 b3c3 d5c3 a1c1 c3e4 d4c6 b7c6 c4b3 e8g8 f2f3 e4f6 c1c6 f8e8 e1f2 c8e6 b3e6 e8e6 h1c1 f6d5 e3d2 e6e8 c6c5 a8d8 c1c2 d8d7 g2g4 f7f6 h2h4 g7g6 c5a5 g8f7 a5a6 e8d8 a2a4 h7h5 g4h5 g6h5');
INSERT INTO evaluations VALUES(6672,30,'cp -58',NULL);
INSERT INTO evaluations VALUES(6672,31,'cp -45',NULL);
INSERT INTO evaluations VALUES(6672,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(6672,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(6672,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(6672,35,'cp -30',NULL);
INSERT INTO evaluations VALUES(6672,36,'cp -38','b4c5 g1f3 g8f6 b1d2 e8g8 f1e2 a7a5 e1g1 f8e8 h2h3 d7d5 e4d5 d8d5 d2b3 c5b6 d1d5 f6d5 e2b5 a5a4 b3d2 e8d8 f1d1 f7f6 d2c4 b6c5 c4e3 d5b6 d1d8 c6d8 e3c2 c8e6 c1e3 c5e3 c2e3 c7c6 b5f1 a4a3 b2b3 b6d5 e3d5 e6d5 a1d1 d8e6');
INSERT INTO evaluations VALUES(9427,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(9427,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(9427,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(9427,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9427,34,'cp 8',NULL);
INSERT INTO evaluations VALUES(9427,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(9427,36,'cp 0','c6a5 b2b4 f6d5 b4c5 d5e7 c4b5 c7c6 c5d6 c6b5 d6e7 d8e7 a1b1 a7a6 c2c4 b5c4 d1a4 a5c6 d3c4 c8g4 c1a3 e7c7 a3c5 g4f3 g2f3 c7a5 a4a5 c6a5 c5e3 a5c4 b1b7 e8g8 e1e2 f7f6 h1d1 f8b8 b7c7 b8b2 e2f1 c4e3 f2e3 b2h2 d1d7 a8b8 d7g7 g8f8 c7f7 f8e8 f7e7 e8d8 e7d7 d8c8 d7c7 c8d8');
INSERT INTO evaluations VALUES(9712,30,'cp 13',NULL);
INSERT INTO evaluations VALUES(9712,31,'cp 10',NULL);
INSERT INTO evaluations VALUES(9712,32,'cp 14',NULL);
INSERT INTO evaluations VALUES(9712,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(9712,34,'cp 13',NULL);
INSERT INTO evaluations VALUES(9712,35,'cp 3',NULL);
INSERT INTO evaluations VALUES(9712,36,'cp 0','e8g8 a2a4 d7d6 c3d5 f6d5 c4d5 c6e7 d5b3 e7g6 a4a5 b6e3 f2e3 d8e7 h2h3 c8e6 b3e6 e7e6 c2c4 c7c6 b2b4 f7f5 e4f5 f8f5 a5a6 b7b6 d1a4 g6e7 b4b5 c6c5 f3d2 f5f1 a1f1 e7f5 f1f3 f5h4 f3g3 h4f5 g3f3');
INSERT INTO evaluations VALUES(13859,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(13859,31,'cp 53',NULL);
INSERT INTO evaluations VALUES(13859,32,'cp 66',NULL);
INSERT INTO evaluations VALUES(13859,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(13859,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(13859,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(13859,36,'cp 63','g1e2 h7h5 d2d4 h5h4 d4d5 c6a5 h2h3 h4g3 f2g3 a7a6 b2b3 b7b5 c4b5 a6b5 c3b5 c8d7 b5a3 f8e7 e3e4 d8c8 c1d2 c8a6 d2e3 e7d8 g3g4 f6h7 e3f2 a5b7 a3c4 d7b5 e1g1 b5c4 b3c4 a6c4 e2g3 e8g8 g3f5 d8f6');
INSERT INTO evaluations VALUES(13860,30,'cp 61',NULL);
INSERT INTO evaluations VALUES(13860,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(13860,32,'cp 48',NULL);
INSERT INTO evaluations VALUES(13860,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(13860,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(13860,35,'cp 51',NULL);
INSERT INTO evaluations VALUES(13860,36,'cp 38','c8e6 c3d5 f8e7 e2c3 d8d7 e1g1 e7d8 d2d3 e8g8 a1b1 e6g4 d5f6 d8f6 d1c2 g4h3 g2h3 d7h3 a2a3 a7a5 c2d1 h3d7 c3d5 f6d8 b2b4 c5b4 a3b4 a5b4 d5b4 c6b4 b1b4 b7b6 b4b5 a8a5 b5a5 b6a5 c1a3 d8e7 d1f3 f8b8 f3d5 h7h5');
INSERT INTO evaluations VALUES(14003,30,'cp 7',NULL);
INSERT INTO evaluations VALUES(14003,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(14003,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(14003,33,'cp -4',NULL);
INSERT INTO evaluations VALUES(14003,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(14003,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(14003,36,'cp 0','c1f4 f8e7 e4e5 c6e5 f3e5 d6e5 d1d8 e7d8 f4e5 c8d7 f1e2 d7c6 e2b5 a8c8 e1c1 a7a6 b5c6 c8c6 d1d3 d8c7 h1d1 e8e7 e5f6 e7f6 c3e4 f6g6 e4d6 h7h5 d3g3 g6h7 g3d3');
INSERT INTO evaluations VALUES(14569,30,'cp 417',NULL);
INSERT INTO evaluations VALUES(14569,31,'cp 417',NULL);
INSERT INTO evaluations VALUES(14569,32,'cp 411',NULL);
INSERT INTO evaluations VALUES(14569,33,'cp 393',NULL);
INSERT INTO evaluations VALUES(14569,34,'cp 393',NULL);
INSERT INTO evaluations VALUES(14569,35,'cp 414',NULL);
INSERT INTO evaluations VALUES(14569,36,'cp 390','e4d5 f6d5 f3e5 c8e6 e5g6 h8g8 b1c3 d5c3 d1g4 c3d5 e1e6 f7e6 g4e6 d5e7 e6f7 e8d7 c4e6 d7d6 d2d4 e7g6 f7g6 d8f6 g6f6 g7f6 e6g8 c6d4 c1d2 d6d7 c2c3 d4c6 g8c4 f8d6 a1e1 a8h8 g2g3 c6e5 c4e2 f6f5 d2f4 e5g6 f4e3 h6h5 e3d4 h8h6 h2h4 f5f4 g1g2 f4g3 f2g3 g6e5 e1d1 e5g4 e2g4 h5g4');
INSERT INTO evaluations VALUES(14562,30,'cp 380',NULL);
INSERT INTO evaluations VALUES(14562,31,'cp 403',NULL);
INSERT INTO evaluations VALUES(14562,32,'cp 395',NULL);
INSERT INTO evaluations VALUES(14562,33,'cp 397',NULL);
INSERT INTO evaluations VALUES(14562,34,'cp 430',NULL);
INSERT INTO evaluations VALUES(14562,35,'cp 425',NULL);
INSERT INTO evaluations VALUES(14562,36,'cp 447','f6d5 f3e5 c8e6 e5g6 h8g8 b1c3 d5c3 b2c3 d8f6 g6f8 g8f8 c1a3 f8h8 c4e6 f7e6 d1g4 e8c8 e1e6 f6g5 g4g5 h6g5 d2d4 h8h6 a1e1 b7b6 h2h3 d8d5 a3c1 c8b7 e6e4 h6f6 e4g4 f6f5 g1f1 d5a5 a2a3 a5d5 f2f3 c6a5 e1e7 b7c8 e7g7 a5c4 g4g5 f5g5 c1g5 d5d7');
INSERT INTO evaluations VALUES(14594,30,'cp 459',NULL);
INSERT INTO evaluations VALUES(14594,31,'cp 440',NULL);
INSERT INTO evaluations VALUES(14594,32,'cp 457',NULL);
INSERT INTO evaluations VALUES(14594,33,'cp 441',NULL);
INSERT INTO evaluations VALUES(14594,34,'cp 465',NULL);
INSERT INTO evaluations VALUES(14594,35,'cp 464',NULL);
INSERT INTO evaluations VALUES(14594,36,'cp 435','f3e5 c8e6 e5g6 h8g8 b1c3 d5c3 b2c3 d8f6 g6f8 g8f8 c1a3 f8g8 c4e6 f7e6 d1g4 e8d7 d2d4 a8e8 e1e4 d7d8 a1e1 f6f7 e1e3 g7g5 g4e2 e6e5 d4d5 c6b8 e4e5 e8e5 e3e5 g8e8 e5e6 b8d7 h2h3 d7f8 e2d3 f8d7 d3e4 d7f8 e4d4 f8g6 e6f6 e8e1 g1h2 f7e8 d5d6 e8e5 d4e5 g6e5 f6h6 e1e2 h6h8 d8d7 d6c7');
INSERT INTO evaluations VALUES(14579,30,'cp 481',NULL);
INSERT INTO evaluations VALUES(14579,31,'cp 463',NULL);
INSERT INTO evaluations VALUES(14579,32,'cp 455',NULL);
INSERT INTO evaluations VALUES(14579,33,'cp 448',NULL);
INSERT INTO evaluations VALUES(14579,34,'cp 413',NULL);
INSERT INTO evaluations VALUES(14579,35,'cp 449',NULL);
INSERT INTO evaluations VALUES(14579,36,'cp 443','c8e6 e5g6 h8g8 b1c3 d5c3 b2c3 d8f6 g6f8 g8f8 c1a3 f8g8 c4e6 f7e6 d1g4 e8d7 d2d4 h6h5 g4h5 g8h8 h5e2 a8e8 a1d1 f6g6 c3c4 g6h5 h2h3 h5e2 e1e2 b7b6 e2e4 c6a5 c4c5 a5c4 a3c1 d7c6 g1f1 a7a5 c5b6 c7b6 a2a4 h8f8 e4g4 b6b5 a4b5 c6b5 g4g7');
INSERT INTO evaluations VALUES(15469,30,'cp -48',NULL);
INSERT INTO evaluations VALUES(15469,31,'cp -42',NULL);
INSERT INTO evaluations VALUES(15469,32,'cp -51',NULL);
INSERT INTO evaluations VALUES(15469,33,'cp -49',NULL);
INSERT INTO evaluations VALUES(15469,34,'cp -53',NULL);
INSERT INTO evaluations VALUES(15469,35,'cp -41',NULL);
INSERT INTO evaluations VALUES(15469,36,'cp -40','h7h6 g5e3 d6d5 b1d2 f8e7 f1e2 e8g8 e1g1 f8e8 h2h3 a7a5 a2a3 a5a4 a1c1 c8d7 f1e1 e7d6 e2f1 c6a5 d1c2 c7c5 e4d5 f6d5 d2e4 d5e3 e1e3 d6f8 c1e1 d8c7 e4g3 a5c6 e3e4 g8h8 g3h5 d7f5 e4a4 a8d8 f1e2 b7b5 a4h4 f5h7 c2b3');
INSERT INTO evaluations VALUES(16228,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(16228,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(16228,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(16228,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(16228,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(16228,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(16228,36,'cp 0','d7c6 f3e5 c8e6 d1e2 f8e7 d2d3 f6d5 b1c3 d5c3 b2c3 d8c8 e2e3 e8g8 e3g3 f8e8 c1f4 e7d6 e5f3 c8d7 f3e5');
INSERT INTO evaluations VALUES(15699,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15699,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15699,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15699,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15699,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15699,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15699,36,'cp 0','f3e5 c8e6 d2d3 f8e7 d1e2 f6d5 b1c3 d5c3 b2c3 d8c8 e2f3 e8g8 f3g3 f8e8 c1f4 e7d6 e5f3 c8d7 f3e5');
INSERT INTO evaluations VALUES(15692,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(15692,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(15692,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(15692,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(15692,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(15692,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(15692,36,'cp 0','f8e7 d1e2 c8e6 d2d3 f6d5 b1c3 d5c3 b2c3 d8c8 e2f3 e8g8 f3g3 f8e8 c1f4 e7d6 e5f3 c8d7 f3e5');
INSERT INTO evaluations VALUES(16247,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(16247,31,'cp 46',NULL);
INSERT INTO evaluations VALUES(16247,32,'cp 33',NULL);
INSERT INTO evaluations VALUES(16247,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(16247,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(16247,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(16247,36,'cp 52','f6d5 f2f4 d5c3 b2c3 d7d5 f1d3 c6a5 g1h3 c7c5 e1g1 g7g6 f4f5 e6f5 g2g4 a5c6 g4f5 d8h4 d1f3 g6f5 a1b1 c5d4 h3f4 c6e5 d3b5 e5c6 b5a4 f8h6 f3d5 e8g8 a4c6 b7c6 d5c6 h6f4 c1f4 c8e6 g1h1 h4g4 f1g1');
INSERT INTO evaluations VALUES(16242,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(16242,31,'cp 50',NULL);
INSERT INTO evaluations VALUES(16242,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(16242,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(16242,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(16242,35,'cp 36',NULL);
INSERT INTO evaluations VALUES(16242,36,'cp 28','g1f3 d7d6 c3d5 e6d5 f1b5 f8e7 e1g1 e8g8 c2c3 d6e5 d4e5 c8g4 f1e1 f7f6 e5f6 e7f6 b5e2 d8d6 c1e3 g4f5 d1d2 c6e5 f3e5 f6e5 f2f4 e5f6 e2f3 c7c6 a1d1 f8e8 e3f2 a7a5 g2g4 f5g6 f4f5 e8e1 d1e1 g6f7 g1g2 a8e8 e1e8 f7e8 f2g3');
INSERT INTO evaluations VALUES(16901,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(16901,31,'cp 15',NULL);
INSERT INTO evaluations VALUES(16901,32,'cp 11',NULL);
INSERT INTO evaluations VALUES(16901,33,'cp 6',NULL);
INSERT INTO evaluations VALUES(16901,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(16901,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(16901,36,'cp 12','c7c6 e1g1 d7d5 b5a4 d5e4 a4b3 e4d3 f1e1 c8e6 b3e6 f7e6 e1e6 e8f7 g5f6 d3c2 d1c2 g7f6 c2b3 d8b6 b3c4 b6b5 e6c6 b5c4 c6c4 f8h6 c4d4 h8d8 d4d8 a8d8 b1c3 d8d2 a1d1 d2b2 d1d7 f7g6 g2g4 b2c2 c3d5 c2a2 d7b7 a2d2');
INSERT INTO evaluations VALUES(18558,30,'cp 117',NULL);
INSERT INTO evaluations VALUES(18558,31,'cp 122',NULL);
INSERT INTO evaluations VALUES(18558,32,'cp 128',NULL);
INSERT INTO evaluations VALUES(18558,33,'cp 127',NULL);
INSERT INTO evaluations VALUES(18558,34,'cp 125',NULL);
INSERT INTO evaluations VALUES(18558,35,'cp 102',NULL);
INSERT INTO evaluations VALUES(18558,36,'cp 119','e4e5 e7e6 c3e4 d8c7 c1f4 d6e5 f3e5 g8f6 e4d2 c7d8 f1d3 f8e7 d1f3 c6c5 d4c5 e8g8 c5c6 d7c5 d3c4 f6d5 f4g3 d8b6 e1g1 b7c6 f1e1 a7a5 a1d1 a5a4 b2b4 a4b3 a2b3 c8a6 h2h3 a6c4 b3c4 d5b4 d1b1 b6d8');
INSERT INTO evaluations VALUES(21984,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(21984,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(21984,32,'cp -17',NULL);
INSERT INTO evaluations VALUES(21984,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(21984,34,'cp -18',NULL);
INSERT INTO evaluations VALUES(21984,35,'cp -3',NULL);
INSERT INTO evaluations VALUES(21984,36,'cp -14','f7f5 e4f5 e5e4 f3h2 c8f5 b1c3 g8f6 d2d3 d8e7 c1e3 e8c8 d3e4 f6e4 c3e4 e7e4 f1d3 e4d5 d3f5 d5f5 d1g4 f5d7 e1c1 g7g6 e3g5 d8e8 g4d7 c8d7 h2g4 f8e7 g5h6 e7d8 h1e1 b7b5 c2c3 c6a5 c1c2 a5c4 g4e3 c7c6 e3c4 e8e1 d1e1');
INSERT INTO evaluations VALUES(27959,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(27959,31,'cp 40',NULL);
INSERT INTO evaluations VALUES(27959,32,'cp 39',NULL);
INSERT INTO evaluations VALUES(27959,33,'cp 20',NULL);
INSERT INTO evaluations VALUES(27959,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(27959,35,'cp 10',NULL);
INSERT INTO evaluations VALUES(27959,36,'cp 10','g2g3 g8e7 f1g2 d7d5 e1g1 b4c3 b2c3 d5e4 f3g5 e8g8 c1b2 d8d7 g5e4 b8c6 c3c4 f7f6 d1g4 e7g6 g4h3 a8e8 f2f4 b7a8 f4f5 e6f5 h3f5 d7f5 f1f5 e8d8 f5f1 h7h6 a2a4 c6d4 b2d4 d8d4 a4a5 f6f5 e4c3 a8g2');
INSERT INTO evaluations VALUES(31268,30,'cp 5',NULL);
INSERT INTO evaluations VALUES(31268,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(31268,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31268,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(31268,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31268,35,'cp 4',NULL);
INSERT INTO evaluations VALUES(31268,36,'cp 0','b8c6 g1f3 a7a6 b5c6 b7c6 c1g5 g8f6 e1g1 e7e6 c3a4 f8d6 c2c4 d5c4 a1c1 h7h6 g5f6 d8f6 a4b6 a8d8 b6c4 f5e4 c4d6 d8d6 f3e5 f6g5 g2g3 e8g8 e5c4 d6d8 c4e5');
INSERT INTO evaluations VALUES(31366,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(31366,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(31366,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(31366,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(31366,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(31366,35,'cp -5',NULL);
INSERT INTO evaluations VALUES(31366,36,'cp 0','d2d4 e7e6 b1c3 b8c6 c1e3 f8b4 a1c1 g8e7 h2h4 h7h6 a2a3 b4c3 c1c3 a8c8 g2g4 f5h7 g1e2 d8b6 c3b3 b6d8');
INSERT INTO evaluations VALUES(32182,30,'cp 101',NULL);
INSERT INTO evaluations VALUES(32182,31,'cp 113',NULL);
INSERT INTO evaluations VALUES(32182,32,'cp 111',NULL);
INSERT INTO evaluations VALUES(32182,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(32182,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(32182,35,'cp 107',NULL);
INSERT INTO evaluations VALUES(32182,36,'cp 108','d6d5 d1b3 b8c6 b1c3 f8b4 f1b5 b4c3 b2c3 g8e7 f3d2 d5e4 d2e4 g4e6 b3c2 e8g8 e1g1 e6f5 f1e1 f5g6 a2a4 c6a5 f2f3 a7a6 b5f1 d8d5 c2f2 f8e8 e4c5 d5c6 c1d2 e7f5 e1e8 a8e8 c3c4 b7b6 d2a5 b6a5 d4d5');
INSERT INTO evaluations VALUES(33063,30,'cp 64',NULL);
INSERT INTO evaluations VALUES(33063,31,'cp 69',NULL);
INSERT INTO evaluations VALUES(33063,32,'cp 85',NULL);
INSERT INTO evaluations VALUES(33063,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(33063,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(33063,35,'cp 74',NULL);
INSERT INTO evaluations VALUES(33063,36,'cp 68','d5d8 h2h3 g8f6 f1b5 b8d7 e1g1 e8g8 f1e1 a7a6 b5f1 b7b6 d1e2 e7e6 a2a4 c8b7 c1f4 f8e8 a1d1 f6d5 c3d5 e6d5 e2d3 c7c5 e1e8 d8e8 d1e1 e8d8 c2c3 c5c4 d3e2 b6b5 a4b5 a6b5 e2e7');
INSERT INTO evaluations VALUES(33275,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(33275,31,'cp 47',NULL);
INSERT INTO evaluations VALUES(33275,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(33275,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(33275,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(33275,35,'cp 56',NULL);
INSERT INTO evaluations VALUES(33275,36,'cp 58','b1c3 d7d6 f1e2 h7h6 g5e3 c5e3 f2e3 g8f6 d1d2 c8e6 e1c1 b8d7 h2h3 e8c8 g2g4 c8b8 d1g1 g6h7 g4g5 h6g5 f3g5 h7g8 c1b1 g8e8 g1g3 g7g6 h1f1 a7a6 a2a4 d6d5 e4d5 f6d5 c3d5 e6d5 e3e4 d5e6 g5e6 e8e6 e2g4 e6e7');
INSERT INTO evaluations VALUES(34527,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(34527,31,'cp 32',NULL);
INSERT INTO evaluations VALUES(34527,32,'cp 32',NULL);
INSERT INTO evaluations VALUES(34527,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(34527,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(34527,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(34527,36,'cp 25','d6b6 g1f3 g8f6 b5a3 c7c5 d2d4 e7e6 a3c4 b6c7 g2g3 f6d5 c4e3 d5e3 c1e3 c8d7 c2c3 d7c6 f1g2 b8d7 e1g1 f8e7 e3f4 c7b6 d1e2 c5d4 f3d4 c6g2 g1g2 e8g8 f1d1 f8d8 d1d2 d7f6 a1d1 h7h6 d4f3 d8d2 d1d2');
INSERT INTO evaluations VALUES(35327,30,'cp 52',NULL);
INSERT INTO evaluations VALUES(35327,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(35327,32,'cp 70',NULL);
INSERT INTO evaluations VALUES(35327,33,'cp 50',NULL);
INSERT INTO evaluations VALUES(35327,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(35327,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(35327,36,'cp 51','d3d4 g8f6 f1d3 c8g4 f2f3 g4h5 d1e2 b8d7 e1c1 e7e6 g1h3 h5g6 h3f2 d7b6 g2g3 a8d8 d2f4 f8d6 f4g5 d6e7 d3g6 h7g6 c3e4 f6e4 f2e4 e7g5 e4g5 e8g8 f3f4 b6d5 h1e1 d5f6 g5f3 b7b5 h2h4 f8e8 f3e5');
INSERT INTO evaluations VALUES(38444,30,'cp 36',NULL);
INSERT INTO evaluations VALUES(38444,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(38444,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(38444,33,'cp 38',NULL);
INSERT INTO evaluations VALUES(38444,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(38444,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(38444,36,'cp 39','d2d4 g8f6 c2c4 e8g8 e1g1 c8f5 d1b3 d8c8 b1c3 a7a5 c1e3 a5a4 c3a4 b8c6 b3d1 f6g4 e3g5 h7h6 g5c1 e7e5 d4d5 c6b4 a4c3 e5e4 f3h4 g7c3 b2c3 b4a2 c1d2 a8a3 h4f5 g6f5 h2h3 g4e5 d1c2');
INSERT INTO evaluations VALUES(38516,30,'cp 23',NULL);
INSERT INTO evaluations VALUES(38516,31,'cp 45',NULL);
INSERT INTO evaluations VALUES(38516,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(38516,33,'cp 29',NULL);
INSERT INTO evaluations VALUES(38516,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(38516,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(38516,36,'cp 25','b8d7 c1e3 g8e7 d1d2 a7a6 a2a4 d7f6 h2h3 e8g8 a4a5 c8d7 f1c4 f6h5 g2g3 h5f6 g3g4 b7b5 a5b6 c7b6 c4d3 b6b5 f3h2 f6e8 e1g1 f7f5 f2f3 e8f6 g1g2 h7h5 f1g1 f5e4 c3e4 f6e4 f3e4 a8c8 g2h1');
INSERT INTO evaluations VALUES(41501,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(41501,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(41501,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(41501,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(41501,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(41501,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(41501,36,'cp 0','e8g8 b2b4 a7a5 b4b5 c7c5 b5c6 b8c6 e1g1 c6b4 d3b5 f6e4 c1b2 d8c7 d2e4 d5e4 f3d2 c7c2 d1c2 b4c2 a1b1 c2b4 a2a3 b4d5 b5c4 c8e6 f1c1 f7f5 a3a4 f8d8 b2a3 d8d7 c4b5 d7d8');
INSERT INTO evaluations VALUES(36778,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(36778,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(36778,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(36778,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(36778,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(36778,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(36778,36,'cp 0','b2b4 b8d7 e1g1 e7e5 d4e5 f6g4 c1b2 g4e5 b2d4 c7c6 f3e5 d7e5 d2b3 f8e8 d1d2 g7f8 d2c3 d8d6 d4c5 d6f6 c5d4');
INSERT INTO evaluations VALUES(41588,30,'cp -76',NULL);
INSERT INTO evaluations VALUES(41588,31,'cp -78',NULL);
INSERT INTO evaluations VALUES(41588,32,'cp -78',NULL);
INSERT INTO evaluations VALUES(41588,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(41588,34,'cp -79',NULL);
INSERT INTO evaluations VALUES(41588,35,'cp -71',NULL);
INSERT INTO evaluations VALUES(41588,36,'cp -73','e8g8 g1e2 c7c6 b1d2 h7h6 g5f6 g7f6 e1g1 b8d7 f3g3 g8g7 a2a4 h6h5 h2h3 h5h4 g3e3 d8b6 d2f3 a7a5 e3d2 d7c5 e2c3 b6b4 f1b1 a8b8 d2e3 c8e6 c4e6 c5e6 c3e2 b4c5 e3c5 e6c5 g1f1 c5a6 b1e1 d6d5 e4d5 c6d5 d3d4');
INSERT INTO evaluations VALUES(36849,30,'cp -72',NULL);
INSERT INTO evaluations VALUES(36849,31,'cp -85',NULL);
INSERT INTO evaluations VALUES(36849,32,'cp -90',NULL);
INSERT INTO evaluations VALUES(36849,33,'cp -71',NULL);
INSERT INTO evaluations VALUES(36849,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(36849,35,'cp -57',NULL);
INSERT INTO evaluations VALUES(36849,36,'cp -58','b1d2 h7h6 g5f6 d8f6 f3f6 g7f6 g1f3 a7a5 a2a4 b8d7 h2h4 h6h5 g2g3 g8g7 e1g1 c7c6 c2c3 d7b6 f1e1 c8g4 c4b3 a8e8 b3c2 f6d8 d3d4 d8c7 a1b1 e8e7 b2b3 f8e8 b3b4 b6d7 f3h2 g4e6 h2f3 a5b4 b1b4 c7b6');
INSERT INTO evaluations VALUES(42049,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42049,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42049,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42049,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42049,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42049,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42049,36,'cp 0','f1g2 e8g8 a2a3 b4c3 b2c3 d5c4 g1f3 c7c5 e1g1 d8a5 c1f4 c5d4 c3d4 a5a6 f3e5 b8d7 d1c2 d7e5 f4e5 c8d7 e5f6 g7f6 c2e4 f6f5 e4f4 d7c6 g2c6 a6c6 f4g5 g8h8 g5f6 h8g8');
INSERT INTO evaluations VALUES(42050,30,'cp -5',NULL);
INSERT INTO evaluations VALUES(42050,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42050,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42050,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42050,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42050,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42050,36,'cp -4','e8g8 g1f3 d5c4 a2a3 b4c3 b2c3 c7c5 e1g1 c5d4 d1d4 b8c6 d4c4 e6e5 c1g5 c8e6 c4b5 d8c7 g5f6 g7f6 f3d2 a8d8 d2c4 g8g7 c4e3 d8d7 f1d1 f8d8 d1d7 d8d7 a1b1 c6e7 c3c4 b7b6 g2d5 e7f5 e3f5 e6f5 e2e4 f5g4 g1g2 g4e2 b5b3 h7h6');
INSERT INTO evaluations VALUES(42232,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42232,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(42232,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42232,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(42232,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42232,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42232,36,'cp 0','b4d6 f1d3 e8g8 c3b5 d6e7 b5c3');
INSERT INTO evaluations VALUES(43247,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43247,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43247,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(43247,33,'cp -9',NULL);
INSERT INTO evaluations VALUES(43247,34,'cp -19',NULL);
INSERT INTO evaluations VALUES(43247,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(43247,36,'cp 0','e8g8 f1d3 d7d5 b2b4 b8c6 a1b1 a7a5 b4b5 c6b4 a2a4 c7c6 b5c6 b7c6 e1g1 c6c5 d4c5 d8c7 c1b2 c7c5 d2b3 c5c7 b2e5 c7b6 e5d4 b6c7');
INSERT INTO evaluations VALUES(37458,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37458,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37458,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37458,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37458,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37458,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37458,36,'cp 0','f1d3 d7d5 b2b4 b8c6 a1b1 a7a5 b4b5 c6b4 a2a4 c7c5 b5c6 b7c6 e1g1 c6c5 d4c5 d8c7 c1b2 c7c5 d2b3 c5c7 b2e5 c7a7 h2h3 c8d7 e5d4 a7c7 d4e5');
INSERT INTO evaluations VALUES(37454,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(37454,31,'cp -22',NULL);
INSERT INTO evaluations VALUES(37454,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(37454,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(37454,34,'cp -16',NULL);
INSERT INTO evaluations VALUES(37454,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(37454,36,'cp -8','d7d5 b2b4 a7a5 b4b5 c7c5 c2c4 c8g4 e1g1 b8d7 h2h3 g4f3 d1f3 a8c8 f1d1 d7b6 c1b2 c5d4 b2d4 d5c4 d2c4 b6c4 d4f6 g7f6 d3c4 d8b6 a1c1 c8c7 c1c2 a5a4 f3e4 f8c8 d1c1 a4a3 c4f1 c7c2 c1c2 c8c2 e4c2 b6d8 g2g3 b7b6 h3h4 g8g7 c2e4');
INSERT INTO evaluations VALUES(43798,30,'cp 46',NULL);
INSERT INTO evaluations VALUES(43798,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(43798,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(43798,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(43798,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(43798,35,'cp 45',NULL);
INSERT INTO evaluations VALUES(43798,36,'cp 44','f1g2 c8b7 c4d5 f8b4 b1c3 e6d5 f3e5 e8g8 c1g5 h7h6 g5f6 b4c3 b2c3 d8f6 c3c4 f8d8 e5g4 f6e6 g4e3 c7c6 c4d5 c6d5 a1c1 b8c6 e1g1 c6a5 d1a4 a8c8 f1d1 g7g6 a4b5 c8c1 d1c1 g8g7 h2h4 h6h5 b5d3 d8d7 c1c2 a5c6 d3c3 c6e7 c2c1');
INSERT INTO evaluations VALUES(49338,30,'cp 182',NULL);
INSERT INTO evaluations VALUES(49338,31,'cp 185',NULL);
INSERT INTO evaluations VALUES(49338,32,'cp 164',NULL);
INSERT INTO evaluations VALUES(49338,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(49338,34,'cp 156',NULL);
INSERT INTO evaluations VALUES(49338,35,'cp 152',NULL);
INSERT INTO evaluations VALUES(49338,36,'cp 181','d2d4 d7d6 h2h3 g4h6 f1d3 b8c6 e1g1 c6b4 d3b5 c8d7 d1e2 a7a6 b5d7 d8d7 c2c4 b4c6 e5d6 f8d6 b1c3 e8c8 f1d1 h8e8 c4c5 d6e7 c1h6 g7h6 b2b4 e6e5 f3e5 c6e5 d4e5 d7e6 d1d8 e8d8 a1e1 c8b8 e2e4 d8d2 f2f4 e7h4 e1e2 d2e2 e4e2 e6f5 c3d5 f5b1 g1h2');
INSERT INTO evaluations VALUES(49331,30,'cp 170',NULL);
INSERT INTO evaluations VALUES(49331,31,'cp 175',NULL);
INSERT INTO evaluations VALUES(49331,32,'cp 186',NULL);
INSERT INTO evaluations VALUES(49331,33,'cp 184',NULL);
INSERT INTO evaluations VALUES(49331,34,'cp 176',NULL);
INSERT INTO evaluations VALUES(49331,35,'cp 178',NULL);
INSERT INTO evaluations VALUES(49331,36,'cp 176','b8c6 b1c3 d7d6 h2h3 g4h6 e5d6 f8d6 c1g5 d6e7 g5h6 g7h6 f1b5 e8g8 b5c6 b7c6 f3e5 e7b4 e5c6 b4c3 b2c3 d8d6 c6e5 f7f6 e5d3 c8b7 e1g1 g8h8 a1b1 b7d5 d3b4 f8g8 b4d5 d6d5 f2f3 d5a2 g1h1 a2d5 d1e2 d5g5');
INSERT INTO evaluations VALUES(54724,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(54724,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(54724,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(54724,33,'cp 2',NULL);
INSERT INTO evaluations VALUES(54724,34,'cp 7',NULL);
INSERT INTO evaluations VALUES(54724,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(54724,36,'cp 12','e7e5 c2c4 d5d4 e1g1 c7c5 e2e3 b8c6 d2d3 c8e6 f1e1 d8d7 e3d4 c5d4 a2a3 a7a5 d1a4 a8a6 b1d2 g8h6 d2b3 h6f7 f3d2 c6d8 a4d7 e8d7 d2e4 b7b6 f2f4 a5a4 b3d2 d8c6 d2f3 f8e7 c1d2 h8b8 e4f2 b6b5 f4e5 b5c4 d3c4 b8b2 e5f6 e7f6');
INSERT INTO evaluations VALUES(56183,30,'cp 63',NULL);
INSERT INTO evaluations VALUES(56183,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(56183,32,'cp 30',NULL);
INSERT INTO evaluations VALUES(56183,33,'cp 37',NULL);
INSERT INTO evaluations VALUES(56183,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(56183,35,'cp 39',NULL);
INSERT INTO evaluations VALUES(56183,36,'cp 38','c7c6 d1a4 d5c4 a4c2 c8f5 f1c4 g8f6 g1e2 f8d6 f2f4 e8g8 e2g3 g7g6 e1g1 b7b5 c4b3 b8a6 a2a3 a8b8 g3f5 g6f5 c2f2 g8h8 f2h4 f8g8 c1d2 a6c7 b3f7 g8g7 f7a2 c6c5 h4h3 f6g4 d2e1 d8e8 a1c1 c5c4 a2b1 b5b4 a3b4 d6b4');
INSERT INTO evaluations VALUES(53711,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(53711,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(53711,32,'cp 58',NULL);
INSERT INTO evaluations VALUES(53711,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(53711,34,'cp 55',NULL);
INSERT INTO evaluations VALUES(53711,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(53711,36,'cp 53','d1a4 d5c4 a4c2 g8f6 f1c4 c8f5 g1e2 f8d6 f2f4 e8g8 e2g3 g7g6 e1g1 b7b5 c4b3 b8a6 a2a3 a8b8 c2e2 a6c7 c1d2 a7a5 g3f5 g6f5 d2e1 b5b4 a3b4 d6b4 b3c4 f6d5 g2g4 f5g4 e2g4 g8h8 g4h3 f7f5 g1h1 f8g8 c4d5 c7d5 c3d5 d8d5 e1b4 b8b4 h3h6 b4b2 f1g1');
INSERT INTO evaluations VALUES(260,30,'cp 28',NULL);
INSERT INTO evaluations VALUES(260,31,'cp 35',NULL);
INSERT INTO evaluations VALUES(260,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(260,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(260,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(260,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(260,36,'cp 31','a7a6 b5c6 d7c6 c1e3 c5e3 f2e3 f6e7 d1e1 f7f6 b2b4 g8h6 e1c3 c8e6 a2a4 h6f7 b1d2 e8g8 h2h3 f8e8 d3d4 h7h6 f1f2 e6d7 c3b3 d7e6 b3b2 b7b6 a4a5 e6d7 a5b6 c7b6 d2c4 b6b5 c4b6 a8a7 a1f1 f7d6 d4e5 f6e5');
INSERT INTO evaluations VALUES(1291,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(1291,31,'cp 6',NULL);
INSERT INTO evaluations VALUES(1291,32,'cp 12',NULL);
INSERT INTO evaluations VALUES(1291,33,'cp 7',NULL);
INSERT INTO evaluations VALUES(1291,34,'cp 14',NULL);
INSERT INTO evaluations VALUES(1291,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(1291,36,'cp 0','e5f3 g5g2 h1g1 g2h3 d1e2 g8e7 b1c3 e7g6 g1g3 h3h5 d2d4 a8b8 b2b3 f8e7 c1d2 f7f6 e1c1 d7d5 d1g1 d5e4 e2e4 e8g8 e4c6 e7d6 c1b2 b8b4 c3e4 c8b7 c6d7 b7c8');
INSERT INTO evaluations VALUES(4361,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(4361,31,'cp -9',NULL);
INSERT INTO evaluations VALUES(4361,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(4361,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(4361,34,'cp -5',NULL);
INSERT INTO evaluations VALUES(4361,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(4361,36,'cp -6','d2d3 a7a6 f1e2 d7d6 e1g1 c8d7 a2a3 c6d4 b2b4 c5a7 a3a4 d7e6 a4a5 c7c6 a1b1 f6d7 c1d2 f7f5 e4f5 e6f5 c3e4 d7f6 f3d4 a7d4 e4g3 f5g6 d2e3 d4e3 f2e3 d6d5 d1c2 d8e7 e3e4 d5e4 d3e4 a8d8 g3f5 g6f5 f1f5');
INSERT INTO evaluations VALUES(4363,30,'cp -24',NULL);
INSERT INTO evaluations VALUES(4363,31,'cp -11',NULL);
INSERT INTO evaluations VALUES(4363,32,'cp -12',NULL);
INSERT INTO evaluations VALUES(4363,33,'cp -20',NULL);
INSERT INTO evaluations VALUES(4363,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(4363,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(4363,36,'cp -15','d7d6 f1e2 c8d7 e1g1 a7a6 a2a3 c6d4 f3d4 c5d4 a3a4 c7c6 a4a5 a8b8 e2f3 f8e8 c3e2 d4a7 e2g3 h7h6 c1d2 d7e6 d2c3 d8e7 b2b4 b8d8 d1e2 g7g6 c3d2 g8h7 d2e3 a7e3 e2e3 h6h5 g3e2 d6d5 e4d5 c6d5');
INSERT INTO evaluations VALUES(6814,30,'cp -179',NULL);
INSERT INTO evaluations VALUES(6814,31,'cp -187',NULL);
INSERT INTO evaluations VALUES(6814,32,'cp -161',NULL);
INSERT INTO evaluations VALUES(6814,33,'cp -165',NULL);
INSERT INTO evaluations VALUES(6814,34,'cp -175',NULL);
INSERT INTO evaluations VALUES(6814,35,'cp -173',NULL);
INSERT INTO evaluations VALUES(6814,36,'cp -138','d1f3 d7d5 c4d5 f7f6 e1g1 b4c3 b2c3 c8g4 f3g4 d8d5 f2f4 f6f5 g4e2 g8e7 f1d1 e8c8 c3c4 d5e4 e2e4 f5e4 a2a4 h7h5 a4a5 e7f5 d4d5 h5h4 d1e1 h8e8 a1b1 c7c6 d5d6 e4e3 g2g3 h4g3 h2g3 g6e5 f4e5 e8e5 a5a6 b7a6 g3g4 f5d6 c1e3 e5e4 g4g5 e4g4 g1f2 g4c4 e3a7 c4c2 f2g3 c2c4');
INSERT INTO evaluations VALUES(6813,30,'cp -173',NULL);
INSERT INTO evaluations VALUES(6813,31,'cp -165',NULL);
INSERT INTO evaluations VALUES(6813,32,'cp -179',NULL);
INSERT INTO evaluations VALUES(6813,33,'cp -174',NULL);
INSERT INTO evaluations VALUES(6813,34,'cp -178',NULL);
INSERT INTO evaluations VALUES(6813,35,'cp -168',NULL);
INSERT INTO evaluations VALUES(6813,36,'cp -186','d7d5 c4d5 f7f6 e1g1 b4c3 b2c3 c8g4 f3g4 d8d5 f2f4 f6f5 g4e2 g8e7 f1d1 e8c8 c3c4 d5e4 e2e4 f5e4 a2a4 h7h5 d4d5 c7c6 d5d6 e7f5 g1f2 g6f8 d1e1 d8e8 c4c5 g7g5 e1e4 g5f4 e4f4 f5h6 h2h3 f8e6 f4f6 e8f8 f2g1 f8f6 e5f6 e6c5 c1e3 c5d7 e3h6 h8h6 f6f7 c8d8');
INSERT INTO evaluations VALUES(7300,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(7300,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(7300,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(7300,33,'cp -7',NULL);
INSERT INTO evaluations VALUES(7300,34,'cp -7',NULL);
INSERT INTO evaluations VALUES(7300,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(7300,36,'cp 0','d2d3 d7d6 c1g5 h7h6 g5f6 d8f6 c3d5 f6d8 c2c3 e8g8 e1g1 c5a7 c4b3 a8b8 a2a4 c6e7 d5e3 e7g6 d3d4 d8f6 e3d5 f6d8');
INSERT INTO evaluations VALUES(11952,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(11952,31,'cp 52',NULL);
INSERT INTO evaluations VALUES(11952,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(11952,33,'cp 62',NULL);
INSERT INTO evaluations VALUES(11952,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(11952,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(11952,36,'cp 50','f3e5 f8e7 f1e1 e4d6 c2c4 e8g8 d2d4 d6f5 b1c3 a8b8 h2h3 f7f6 e5f3 d7d5 d1a4 c8d7 c4d5 c6d5 a4a7 d8c8 c1f4 b8b7 a7a6 b7b2 a6d3 e7b4 f4c1 b4c3 d3c3 b2b7 a2a4 f5d6 a4a5 d6e4 c3c2 d7f5 c2d1');
INSERT INTO evaluations VALUES(12955,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(12955,31,'cp -44',NULL);
INSERT INTO evaluations VALUES(12955,32,'cp -41',NULL);
INSERT INTO evaluations VALUES(12955,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(12955,34,'cp -54',NULL);
INSERT INTO evaluations VALUES(12955,35,'cp -47',NULL);
INSERT INTO evaluations VALUES(12955,36,'cp -40','c5d4 f1d3 d4c3 b1c3 e7f5 d3f5 e6f5 c1e3 c8e6 e1g1 d8d7 d1d3 f8b4 a1c1 b4c3 c1c3 e8g8 f1d1 h7h6 h2h3 f8d8 d1d2 a8c8 a2a3 c6a5 f3d4 c8c3 d3c3 d7a4 c3c2 a4c4 b2b3 c4a6 c2c7 d8d7 c7c8 g8h7 c8c3 a5c6');
INSERT INTO evaluations VALUES(15449,30,'cp -46',NULL);
INSERT INTO evaluations VALUES(15449,31,'cp -67',NULL);
INSERT INTO evaluations VALUES(15449,32,'cp -46',NULL);
INSERT INTO evaluations VALUES(15449,33,'cp -40',NULL);
INSERT INTO evaluations VALUES(15449,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(15449,35,'cp -61',NULL);
INSERT INTO evaluations VALUES(15449,36,'cp -71','f6e4 d2d4 f8e7 e1g1 e8g8 f1e1 e4f6 d4e5 d6e5 f3e5 c6e5 e1e5 e7d6 e5e2 c7c6 d5f6 d8f6 h2h3 d6c7 e2e1 c8f5 d1f3 a8e8 e1e8 f8e8 c1e3 f5c2 f3f6 g7f6 a1c1 c2f5 g2g4 f5g6 c1d1 b7b5 c4e2 c7b6 e3b6 a7b6 e2f3 c6c5 g1g2 g8g7 d1d7 f6f5 d7b7 e8e6');
INSERT INTO evaluations VALUES(17557,30,'cp 62',NULL);
INSERT INTO evaluations VALUES(17557,31,'cp 59',NULL);
INSERT INTO evaluations VALUES(17557,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(17557,33,'cp 57',NULL);
INSERT INTO evaluations VALUES(17557,34,'cp 63',NULL);
INSERT INTO evaluations VALUES(17557,35,'cp 49',NULL);
INSERT INTO evaluations VALUES(17557,36,'cp 71','b5c6 b7c6 e1g1 e7e6 d2d4 d6d5 e4e5 a6a5 d1c2 h7h6 h2h3 d8b6 c1e3 c5d4 c3d4 g8e7 b1c3 e7f5 c3a4 b6b5 e3d2 f8e7 a1c1 c8a6 f1e1 e8g8 g2g4 f5h4 f3h4 e7h4 a4c5 h4e7 a2a4 b5c4 c5d7 c4d4 d7f8 a8f8 d2a5 a6d3 c2c6 d4b2 a5c3');
INSERT INTO evaluations VALUES(17542,30,'cp 75',NULL);
INSERT INTO evaluations VALUES(17542,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(17542,32,'cp 51',NULL);
INSERT INTO evaluations VALUES(17542,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(17542,34,'cp 54',NULL);
INSERT INTO evaluations VALUES(17542,35,'cp 66',NULL);
INSERT INTO evaluations VALUES(17542,36,'cp 64','b7c6 e1g1 e7e6 d2d4 d6d5 f3e5 c5d4 e5c6 d8c7 e4d5 g8f6 c3c4 f8d6 d1d4 c8b7 g2g3 e8g8 c1g5 e6d5 g5f6 c7c6 c4d5 c6d5 d4d5 b7d5 f1d1 g7f6 d1d5 d6e5 b1d2 e5b2 a1b1 b2e5 d2c4 f8d8 b1d1 d8d5 d1d5 a8e8 d5a5 e5d4 a5a6 e8e2');
INSERT INTO evaluations VALUES(18212,30,'cp 67',NULL);
INSERT INTO evaluations VALUES(18212,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(18212,32,'cp 69',NULL);
INSERT INTO evaluations VALUES(18212,33,'cp 64',NULL);
INSERT INTO evaluations VALUES(18212,34,'cp 78',NULL);
INSERT INTO evaluations VALUES(18212,35,'cp 57',NULL);
INSERT INTO evaluations VALUES(18212,36,'cp 59','e1g1 g8f6 e4e5 f6d7 d2d4 e7e6 c1g5 d8c7 e5d6 f8d6 b1d2 h7h6 g5h4 c5d4 d2e4 d6f4 c3d4 a6a5 f1e1 e8g8 h4g3 c8a6 a1c1 f8b8 b2b3 a5a4 b3a4 b8b2 e1e3 f4g3 h2g3 b2a2 d1b3 a2e2 e4c5 d7c5 c1c5');
INSERT INTO evaluations VALUES(18211,30,'cp 78',NULL);
INSERT INTO evaluations VALUES(18211,31,'cp 75',NULL);
INSERT INTO evaluations VALUES(18211,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(18211,33,'cp 78',NULL);
INSERT INTO evaluations VALUES(18211,34,'cp 53',NULL);
INSERT INTO evaluations VALUES(18211,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(18211,36,'cp 62','e7e6 d2d4 d6d5 e4e5 a6a5 d1c2 c8d7 c1e3 d8b6 b2b3 c5d4 c3d4 g8e7 b1c3 e7f5 c3a4 b6a7 a2a3 f5e3 f2e3 f8e7 h2h4 f7f6 b3b4 e8g8 a4c5 e7c5 c2c5 d7e8 e5f6 g7f6 c5d6 a7d7 d6f4 e8h5 b4a5 a8a5 f3d2 a5a8 d2b3 d7g7 b3c5');
INSERT INTO evaluations VALUES(17673,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(17673,31,'cp -19',NULL);
INSERT INTO evaluations VALUES(17673,32,'cp -14',NULL);
INSERT INTO evaluations VALUES(17673,33,'cp -17',NULL);
INSERT INTO evaluations VALUES(17673,34,'cp -17',NULL);
INSERT INTO evaluations VALUES(17673,35,'cp -19',NULL);
INSERT INTO evaluations VALUES(17673,36,'cp -12','g8f6 e1g1 f8e7 f1e1 e8g8 h2h3 h7h6 c1f4 c8d7 e4e5 d6e5 f3e5 d7e8 e5c6 e8c6 a4a5 e7d6 f4d6 d8d6 d1e2 a8c8 c3a4 f6d7 e2e3 f8e8 a4c3 d7f6 c3a4 c6a4 a1a4 g8f8 e3e5 e8d8 f2f4 d6c6 a4a1 c6d7 c4b3 d7d4 g1h2 d4e5 e1e5 f6d7 e5e3');
INSERT INTO evaluations VALUES(18083,30,'cp 160',NULL);
INSERT INTO evaluations VALUES(18083,31,'cp 156',NULL);
INSERT INTO evaluations VALUES(18083,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(18083,33,'cp 147',NULL);
INSERT INTO evaluations VALUES(18083,34,'cp 175',NULL);
INSERT INTO evaluations VALUES(18083,35,'cp 156',NULL);
INSERT INTO evaluations VALUES(18083,36,'cp 151','a4b3 c8b7 a2a4 f8d6 a4b5 a6b5 a1a8 b7a8 b1a3 d8h4 g2g3 h4g5 d2d3 g5c5 c2c3 d4c3 d3d4 c5c6 d4d5 c6a6 d1d4 d6e7 d4c3 g8f6 c3d3 a6b7 a3b5 c7c6 d5d6 c6b5 d6e7 b7e4 d3e4 a8e4 f2f3 e4g6');
INSERT INTO evaluations VALUES(18612,30,'cp -578',NULL);
INSERT INTO evaluations VALUES(18612,31,'cp -567',NULL);
INSERT INTO evaluations VALUES(18612,32,'cp -569',NULL);
INSERT INTO evaluations VALUES(18612,33,'cp -570',NULL);
INSERT INTO evaluations VALUES(18612,34,'cp -567',NULL);
INSERT INTO evaluations VALUES(18612,35,'cp -573',NULL);
INSERT INTO evaluations VALUES(18612,36,'cp -578','d8g5 b1a3 a7a6 d2d4 g5f6 f3f6 g8f6 d4d5 c6e7 d5e6 f7e6 e4e5 f6g4 f2f4 b7b5 c4e2 g4h6 e1g1 c8b7 c2c4 e7f5 c4b5 f5d4 e2d3 a6b5 c1e3 h6f5 e3d4 f5d4 a3b5 d4b5 d3b5 b7d5 f4f5 g6f5 f1d1 h8g8 d1d5 e6d5 a2a4 c5c4');
INSERT INTO evaluations VALUES(1319,30,'cp -576',NULL);
INSERT INTO evaluations VALUES(1319,31,'cp -563',NULL);
INSERT INTO evaluations VALUES(1319,32,'cp -565',NULL);
INSERT INTO evaluations VALUES(1319,33,'cp -571',NULL);
INSERT INTO evaluations VALUES(1319,34,'cp -586',NULL);
INSERT INTO evaluations VALUES(1319,35,'cp -557',NULL);
INSERT INTO evaluations VALUES(1319,36,'cp -578','b1a3 g5d8 d2d3 f8g7 a3b5 d7d6 f3g3 g7e5 f2f4 e5f6 e1g1 a7a6 b5c3 g8e7 c3e2 b7b5 c4b3 e8g8 c2c3 f6g7 g3h4 g8h8 c1e3 a6a5 a2a4 a8b8 a1b1 b5a4 b3a4 c8a6 f1d1 d8c7 f4f5 e6f5 e3g5 f5e4 h4e4 d6d5');
INSERT INTO evaluations VALUES(18886,30,'cp -2',NULL);
INSERT INTO evaluations VALUES(18886,31,'cp -2',NULL);
INSERT INTO evaluations VALUES(18886,32,'cp -3',NULL);
INSERT INTO evaluations VALUES(18886,33,'cp -11',NULL);
INSERT INTO evaluations VALUES(18886,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(18886,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(18886,36,'cp 0','d7d5 e4d5 e6d5 c4b3 c8e6 b1c3 g7g6 g1f3 g8e7 e1g1 f8g7 c1g5 h7h6 g5f4 e8g8 h2h3 e7f5 f4e5 f5h4 f3h4 d8h4 f2f4 e6h3 g2h3 h4g3 g1h1 g3h3 h1g1 g7e5 d4e5 h3g3 g1h1 g3h3');
INSERT INTO evaluations VALUES(20370,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(20370,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(20370,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(20370,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(20370,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(20370,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(20370,36,'cp 0','e4d5 e6d5 c4b3 g8f6 g1e2 f8b4 b1c3 h7h6 e1g1 e8g8 h2h3 f8e8 a2a3 b4a5 c1f4 c8f5 a1c1 a8c8 b3a2 a5b6 c3b5 e8e6 b5c3');
INSERT INTO evaluations VALUES(21172,30,'cp 89',NULL);
INSERT INTO evaluations VALUES(21172,31,'cp 96',NULL);
INSERT INTO evaluations VALUES(21172,32,'cp 87',NULL);
INSERT INTO evaluations VALUES(21172,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(21172,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(21172,35,'cp 101',NULL);
INSERT INTO evaluations VALUES(21172,36,'cp 85','e7e5 d4d5 c6b8 e2e4 b8d7 f1e2 g7g6 g1f3 f8h6 e1g1 h6c1 d1c1 a7a5 f3e1 b7b6 e1d3 c8a6 b2b3 d8e7 g2g3 d7c5 d3c5 b6c5 f2f4 g8f6 c1e3 e8g8 f1f2 e5f4 g3f4 a8e8 a1f1 a6b7');
INSERT INTO evaluations VALUES(21931,30,'cp 82',NULL);
INSERT INTO evaluations VALUES(21931,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(21931,32,'cp 104',NULL);
INSERT INTO evaluations VALUES(21931,33,'cp 96',NULL);
INSERT INTO evaluations VALUES(21931,34,'cp 102',NULL);
INSERT INTO evaluations VALUES(21931,35,'cp 109',NULL);
INSERT INTO evaluations VALUES(21931,36,'cp 105','d4d5 c6b8 e2e4 b8d7 f1e2 g7g6 h2h4 f8h6 c1h6 g8h6 d1d2 h6g8 g2g4 h7h6 e1c1 g8f6 g1h3 d7c5 f2f3 a7a5 h3f2 d8e7 d1g1 h6h5 c1b1 e8d8 d2e3 e7f8 a2a3 f8e7 g1f1 a5a4 g4g5 f6d7 f2d3 b7b6 d3c5');
INSERT INTO evaluations VALUES(21441,30,'cp 48',NULL);
INSERT INTO evaluations VALUES(21441,31,'cp 41',NULL);
INSERT INTO evaluations VALUES(21441,32,'cp 49',NULL);
INSERT INTO evaluations VALUES(21441,33,'cp 51',NULL);
INSERT INTO evaluations VALUES(21441,34,'cp 47',NULL);
INSERT INTO evaluations VALUES(21441,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(21441,36,'cp 32','g1f3 d5e4 d3e4 g8f6 c1g5 d8d6 g5f6 g7f6 c2c3 f8g7 d1e2 e8g8 b1a3 c6e7 a3c4 d6d8 h2h4 a8b8 e1c1 b7b5 c4e3 b5b4 c3c4 c7c5 g2g4 c5d4 f3d4 d8b6 h1g1 f8d8 d4b3 d8d1 g1d1 f6f5 g4f5 e6f5 e4f3 c8e6 e3d5 e7d5 f3d5 g8h8 d5e6 f7e6 h4h5 h7h6 c1b1');
INSERT INTO evaluations VALUES(21440,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(21440,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(21440,32,'cp 34',NULL);
INSERT INTO evaluations VALUES(21440,33,'cp 32',NULL);
INSERT INTO evaluations VALUES(21440,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(21440,35,'cp 32',NULL);
INSERT INTO evaluations VALUES(21440,36,'cp 26','d5e4 d3e4 g8f6 c1g5 d8d6 g5f6 g7f6 c2c3 f6f5 e4c2 f8g7 b1a3 e8g8 d1e2 c6e7 e1c1 c7c5 h1g1 e7g6 g2g3 c5d4 f3d4 d6e5 e2e5 g6e5 h2h3 e5g6 a3b5 c8d7 b5d6 b7b6 f2f4 a8d8 d4b5 d7b5 d6b5 a7a6 b5d4');
INSERT INTO evaluations VALUES(21712,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(21712,31,'cp -34',NULL);
INSERT INTO evaluations VALUES(21712,32,'cp -21',NULL);
INSERT INTO evaluations VALUES(21712,33,'cp -26',NULL);
INSERT INTO evaluations VALUES(21712,34,'cp -23',NULL);
INSERT INTO evaluations VALUES(21712,35,'cp -23',NULL);
INSERT INTO evaluations VALUES(21712,36,'cp -26','g8f6 b1d2 f8d6 f1e2 h7h6 e1g1 e8g8 a2a4 c8e6 a4a5 a7a6 f1e1 b7b5 a5b6 c7b6 e4d5 e6d5 e2f1 d5e6 d2c4 e6c4 d3c4 e5e4 f3d4 d8c7 d4c6 c7c6 c1e3 c6c7 g2g3 d6c5 e3c5 c7c5 d1d4 c5c7 g1h1');
INSERT INTO evaluations VALUES(22077,30,'cp 86',NULL);
INSERT INTO evaluations VALUES(22077,31,'cp 91',NULL);
INSERT INTO evaluations VALUES(22077,32,'cp 78',NULL);
INSERT INTO evaluations VALUES(22077,33,'cp 79',NULL);
INSERT INTO evaluations VALUES(22077,34,'cp 77',NULL);
INSERT INTO evaluations VALUES(22077,35,'cp 76',NULL);
INSERT INTO evaluations VALUES(22077,36,'cp 94','d2d4 a7a6 h2h3 g8f6 a2a3 f8e7 f1d3 b7b5 e1g1 e8g8 f1e1 c8b7 c1d2 e6e5 d4d5 c6b8 b2b4 c7c6 d5c6 b8c6 a1b1 c6a7 a3a4 h7h6 f3h2 b5a4 h2g4 f6g4 d1g4 b7c8 g4h5 c8d7 c3d5 e7g5 d2g5 d8g5 h5g5');
INSERT INTO evaluations VALUES(22671,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(22671,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(22671,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(22671,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(22671,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(22671,35,'cp 14',NULL);
INSERT INTO evaluations VALUES(22671,36,'cp 5','c6e5 e2e3 b7b6 g1f3 d8f6 f1e2 c8b7 b1c3 e5f3 e2f3 b7f3 d1f3 f6f3 g2f3 g7g6 b2b3 e8c8 c1b2 f8g7 h2h4 h7h5 a1b1 g8e7 c3d5 e7f5 b2g7 f5g7 d5f6 g7e8 f6e4 d7d5 c4d5 d8d5 e1e2 e8g7 h1c1 g7f5 c1c2 d5d7 e4f6 d7d6 f6e4');
INSERT INTO evaluations VALUES(23070,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(23070,31,'cp 16',NULL);
INSERT INTO evaluations VALUES(23070,32,'cp 26',NULL);
INSERT INTO evaluations VALUES(23070,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(23070,34,'cp 21',NULL);
INSERT INTO evaluations VALUES(23070,35,'cp 25',NULL);
INSERT INTO evaluations VALUES(23070,36,'cp 17','e2e3 b7b6 g1f3 d8f6 f1e2 c8b7 c1d2 e5f3 e2f3 b7f3 d1f3 f6f3 g2f3 g8f6 d2c3 f8e7 c3d4 e8c8 b1c3 f6h5 c3d5 h8e8 d5e7 e8e7 b2b4 d7d6 b4b5 c8d7 a2a4 a7a5 b5a6 d8a8 a4a5 a8a6 a5b6 a6a1 d4a1 c7b6 a1d4 d7c6 e1d2 e7a7 h1b1 a7a2 d2d3 a2f2 b1b6 c6c7 c4c5 d6c5 d4c5 f2f3');
INSERT INTO evaluations VALUES(23702,30,'cp 98',NULL);
INSERT INTO evaluations VALUES(23702,31,'cp 101',NULL);
INSERT INTO evaluations VALUES(23702,32,'cp 101',NULL);
INSERT INTO evaluations VALUES(23702,33,'cp 109',NULL);
INSERT INTO evaluations VALUES(23702,34,'cp 111',NULL);
INSERT INTO evaluations VALUES(23702,35,'cp 116',NULL);
INSERT INTO evaluations VALUES(23702,36,'cp 101','h2h3 g4f3 d2f3 f8e8 f4h2 f6d7 d1b3 e7e5 b3d5 e5d4 c3d4 c6b4 d5c4 b4d3 c4d3 c7c5 e1g1 c5d4 e3d4 d8a5 a1c1 a8c8 b2b3 a5d5 c1c8 e8c8 f1e1 a7a6 h2f4 d7f8 d3e4 d5e4 e1e4 f8e6 f4e3 c8d8 g2g4 d8d6 g4g5 f7f5 g5f6 g7f6 g1g2 e6c7 h3h4 d6d8 h4h5 c7d5 h5g6');
INSERT INTO evaluations VALUES(23701,30,'cp 100',NULL);
INSERT INTO evaluations VALUES(23701,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(23701,32,'cp 97',NULL);
INSERT INTO evaluations VALUES(23701,33,'cp 105',NULL);
INSERT INTO evaluations VALUES(23701,34,'cp 105',NULL);
INSERT INTO evaluations VALUES(23701,35,'cp 112',NULL);
INSERT INTO evaluations VALUES(23701,36,'cp 95','g4f3 d2f3 f8e8 f4h2 f6d7 d1b3 e7e5 b3d5 e5d4 c3d4 c6b4 d5c4 b4d3 c4d3 c7c5 e1g1 c5d4 f1d1 d8b6 e3d4 d7f6 b2b3 f6d5 a1c1 a8d8 h2g3 d5b4 d3d2 b4d5 d1e1 g7f8 e1e8 d8e8 f3e5 b6e6 d2d3 f8b4 g1h2 a7a6 d3f3 f7f6 e5d3 e6e4');
INSERT INTO evaluations VALUES(24438,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(24438,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(24438,32,'cp -11',NULL);
INSERT INTO evaluations VALUES(24438,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24438,34,'cp -3',NULL);
INSERT INTO evaluations VALUES(24438,35,'cp -9',NULL);
INSERT INTO evaluations VALUES(24438,36,'cp -10','f8e7 g5e3 e8g8 b5d3 a7a6 h2h3 c6d4 e3d4 d7c6 f2f4 f6d7 d1e1 e7h4 e1e2 d8e7 a1d1 h4f6 d4f6 d7f6 e2e3 b7b5 a2a3 f8e8 c3e2 e6e5 e2g3 e5f4 e3f4 e7e5 b2b4 e5f4 f1f4 e8e5 g3f5 a8d8 c2c4 c6e4');
INSERT INTO evaluations VALUES(23999,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(23999,31,'cp -5',NULL);
INSERT INTO evaluations VALUES(23999,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(23999,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(23999,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(23999,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(23999,36,'cp 0','g5e3 a7a6 b5d3 e8g8 f2f4 c6d4 e3d4 d7c6 g1h1 f6d7 d1e1 e6e5 f4e5 d7e5 a1d1 e7f6 b2b3 b7b5 e1g3 f6h4 g3h3 h4f6');
INSERT INTO evaluations VALUES(28321,30,'cp 163',NULL);
INSERT INTO evaluations VALUES(28321,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(28321,32,'cp 123',NULL);
INSERT INTO evaluations VALUES(28321,33,'cp 156',NULL);
INSERT INTO evaluations VALUES(28321,34,'cp 149',NULL);
INSERT INTO evaluations VALUES(28321,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(28321,36,'cp 157','h2h4 d7d6 h4h5 g6h5 h1h5 b8d7 d1e2 h7h6 c1d2 e7g6 g2g3 d8e7 a2a4 a7a6 e1c1 e8c8 a4a5 c7c5 d4d5 d7f6 h5h1 e6d5 e4d5 e7e2 d3e2 b6a5 f3h4 g6h4 h1h4 d8e8 d1h1 h6h5 e2h5 h8h5 h4h5 f6h5 h1h5');
INSERT INTO evaluations VALUES(29702,30,'cp 134',NULL);
INSERT INTO evaluations VALUES(29702,31,'cp 118',NULL);
INSERT INTO evaluations VALUES(29702,32,'cp 140',NULL);
INSERT INTO evaluations VALUES(29702,33,'cp 125',NULL);
INSERT INTO evaluations VALUES(29702,34,'cp 158',NULL);
INSERT INTO evaluations VALUES(29702,35,'cp 148',NULL);
INSERT INTO evaluations VALUES(29702,36,'cp 143','f3e5 f5g6 d2d4 e7e6 d4d5 f8b4 d5e6 d8d1 e1d1 f7e6 c4e6 b8c6 e5c6 b7c6 f2f3 a8d8 c1d2 e8e7 e6c4 e7d7 h1e1 d7c8 d1c1 c8b8 a2a3 b4d6 h2h3 d6g3 e1e2 f6h5 b2b3 h5f4 d2f4 g3f4 c1b2 b8b7 a1e1 f4g3 e1d1 d8d1 c3d1 h8e8 e2e8 g6e8 c4d3');
INSERT INTO evaluations VALUES(30256,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(30256,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(30256,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(30256,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(30256,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(30256,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(30256,36,'cp 0','e2e3 f5e6 g1e2 c7c6 e2f4 g7g6 f1d3 f6f5 g2g4 d8f6 g4f5 e6f5 f4h5 f6e6 h5f4');
INSERT INTO evaluations VALUES(32083,30,'cp 29',NULL);
INSERT INTO evaluations VALUES(32083,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(32083,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(32083,33,'cp 25',NULL);
INSERT INTO evaluations VALUES(32083,34,'cp 24',NULL);
INSERT INTO evaluations VALUES(32083,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(32083,36,'cp 17','g8f6 d2d4 b8d7 c2c3 f8e7 f1e1 e8g8 b1d2 e5d4 c3d4 d6d5 e4d5 f8e8 d1b3 d7b6 c4b5 e8f8 b5f1 a7a5 f3e5 a5a4 b3c2 f6d5 a2a3 c7c6 d2e4 b6d7 e5d7 d8d7 e4c5 d7c7 c5a4 h5g6 c2d1 e7f6 f1c4 f8d8 c4d5 d8d5 c1e3 g6e4');
INSERT INTO evaluations VALUES(29892,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(29892,31,'cp 22',NULL);
INSERT INTO evaluations VALUES(29892,32,'cp 38',NULL);
INSERT INTO evaluations VALUES(29892,33,'cp 35',NULL);
INSERT INTO evaluations VALUES(29892,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(29892,35,'cp 29',NULL);
INSERT INTO evaluations VALUES(29892,36,'cp 37','d2d3 c7c6 a2a4 f8e7 b1c3 e8g8 g2g4 h5g6 f3h4 d6d5 h4g6 h7g6 c4a2 b8a6 e4d5 f6d5 c3d5 c6d5 d1f3 a6c7 c1d2 d8d6 a1e1 d6c6 e1e5 e7f6 e5e2 c6a4 a2b3 a4d7 f1e1 a7a5 g4g5 f6b2 e2e7 d7d8 f3g4 b7b5 g4d7 d8d7 e7d7');
INSERT INTO evaluations VALUES(32950,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(32950,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(32950,32,'cp 35',NULL);
INSERT INTO evaluations VALUES(32950,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(32950,34,'cp 37',NULL);
INSERT INTO evaluations VALUES(32950,35,'cp 34',NULL);
INSERT INTO evaluations VALUES(32950,36,'cp 32','f8e7 d2d4 e5d4 f3d4 e7f6 c2c4 g8e7 b1c3 b8c6 c1e3 e8g8 b2b3 f6d4 e3d4 f7f5 f2f3 f5e4 c3e4 c6d4 d1d4 b7b6 f1e1 a8e8 e4c3 e7g6 e1e8 f8e8 g1f2 h7h6 a1d1 g8h8 c3d5 a7a5 d1e1 e8e1 f2e1 g6e7 e1f2 e7d5');
INSERT INTO evaluations VALUES(34268,30,'cp 178',NULL);
INSERT INTO evaluations VALUES(34268,31,'cp 158',NULL);
INSERT INTO evaluations VALUES(34268,32,'cp 156',NULL);
INSERT INTO evaluations VALUES(34268,33,'cp 149',NULL);
INSERT INTO evaluations VALUES(34268,34,'cp 171',NULL);
INSERT INTO evaluations VALUES(34268,35,'cp 161',NULL);
INSERT INTO evaluations VALUES(34268,36,'cp 151','d2d4 d7d6 e5f3 e7e4 f1e2 g8f6 e1g1 e4e7 e2d3 e7d8 d1e2 f8e7 f1e1 b8a6 d3a6 b7a6 c1f4 c8e6 c2c4 a8b8 b1c3 b8b7 h2h3 e8g8 b2b3 f8e8 e2d3 d8a5 g2g4 h7h6 g1g2 b7b8 f4g3 a5b6 e1e3 b6b7 a1e1 e7f8 g3h4 f8e7 g4g5 f6h5 g5h6 e7h4');
INSERT INTO evaluations VALUES(34267,30,'cp 153',NULL);
INSERT INTO evaluations VALUES(34267,31,'cp 150',NULL);
INSERT INTO evaluations VALUES(34267,32,'cp 143',NULL);
INSERT INTO evaluations VALUES(34267,33,'cp 144',NULL);
INSERT INTO evaluations VALUES(34267,34,'cp 147',NULL);
INSERT INTO evaluations VALUES(34267,35,'cp 150',NULL);
INSERT INTO evaluations VALUES(34267,36,'cp 154','d7d6 e5f3 e7e4 f1e2 e4e7 e1g1 e7d8 f1e1 g8f6 e2d3 f8e7 d1e2 b8a6 d3a6 b7a6 b1c3 c8e6 h2h3 a8b8 e2a6 d8c8 a6c8 b8c8 f3g5 e8d7 g5e6 f7e6 b2b3 c6c5 d4c5 c8c5 c3e2 c5c2 e2d4 c2c8 c1a3 e6e5 d4f5 f6g8 f5g7 c8c2 g7f5');
INSERT INTO evaluations VALUES(35233,30,'cp 229',NULL);
INSERT INTO evaluations VALUES(35233,31,'cp 252',NULL);
INSERT INTO evaluations VALUES(35233,32,'cp 244',NULL);
INSERT INTO evaluations VALUES(35233,33,'cp 255',NULL);
INSERT INTO evaluations VALUES(35233,34,'cp 239',NULL);
INSERT INTO evaluations VALUES(35233,35,'cp 252',NULL);
INSERT INTO evaluations VALUES(35233,36,'cp 249','g1f3 b8c6 a2a3 d7d6 b1c3 h7h6 e2e4 f6f7 c1e3 g8f6 c3d5 c8e6 d5f6 f7f6 f3d4 c6d4 e3d4 f6g6 d1d3 f8e7 f2f4 g6f7 f4f5 e6d7 d3c3 e8c8 f1c4 e7h4 g2g3 f7e7 e1c1 h4f6 h1e1 h8e8 c1b1 f6d4 c3d4 c8b8 b1a1 d7c6 c4e6');
INSERT INTO evaluations VALUES(35232,30,'cp 240',NULL);
INSERT INTO evaluations VALUES(35232,31,'cp 237',NULL);
INSERT INTO evaluations VALUES(35232,32,'cp 240',NULL);
INSERT INTO evaluations VALUES(35232,33,'cp 228',NULL);
INSERT INTO evaluations VALUES(35232,34,'cp 224',NULL);
INSERT INTO evaluations VALUES(35232,35,'cp 224',NULL);
INSERT INTO evaluations VALUES(35232,36,'cp 225','b8c6 a2a3 f8c5 b1c3 d7d6 c3d5 f6f7 e2e3 g8f6 f1c4 f6d5 d1d5 f7f6 e1g1 c6e7 d5h5 g7g6 h5g5 f6g5 f3g5 d6d5 c4d3 h7h6 g5f3 c8f5 b2b4 c5d6 d3f5 g6f5 g2g3 h8g8 c1b2 a7a5 g1g2 b7b5 b2c3 a5b4 a3b4 e8d7 f3d4 c7c6 f1d1 d6e5 a1a8 g8a8');
INSERT INTO evaluations VALUES(2010,30,'cp 228',NULL);
INSERT INTO evaluations VALUES(2010,31,'cp 225',NULL);
INSERT INTO evaluations VALUES(2010,32,'cp 217',NULL);
INSERT INTO evaluations VALUES(2010,33,'cp 232',NULL);
INSERT INTO evaluations VALUES(2010,34,'cp 229',NULL);
INSERT INTO evaluations VALUES(2010,35,'cp 231',NULL);
INSERT INTO evaluations VALUES(2010,36,'cp 235','a2a3 f8c5 b1c3 d7d6 c3d5 f6f7 e2e3 g8f6 f1c4 f6d5 d1d5 f7f6 e1g1 c6e7 d5h5 g7g6 h5g5 f6g5 f3g5 d6d5 c4e2 e7f5 e3e4 f5d4 e2d1 h7h6 g5f3 d5e4 f1e1 d4f3 d1f3 c8f5 f3e4 e8c8 e4f5 g6f5 g2g3 c5d6 g1g2 f5f4 g3f4 h8f8 g2f3 f8f5 e1e4 f5h5 c1e3 h5h3 f3g2 h3h4 a1d1');
INSERT INTO evaluations VALUES(35785,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(35785,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(35785,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(35785,33,'cp 46',NULL);
INSERT INTO evaluations VALUES(35785,34,'cp 45',NULL);
INSERT INTO evaluations VALUES(35785,35,'cp 40',NULL);
INSERT INTO evaluations VALUES(35785,36,'cp 26','d2d4 b7b5 c4d3 e5d4 f3d4 b5b4 c3e2 c6c5 d4f3 f8e8 f1e1 a7a5 e2g3 b8d7 a2a4 e7f8 b2b3 d7b6 f3d2 c8e6 c1b2 c5c4 d2c4 b6c4 b3c4 a8c8 c2c3 b4b3 d1b3 f6d7 b3c2 d7e5 g3f5 e5d3 c2d3 e6c4 d3g3 c4e6 a1d1');
INSERT INTO evaluations VALUES(35791,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(35791,31,'cp 26',NULL);
INSERT INTO evaluations VALUES(35791,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(35791,33,'cp 17',NULL);
INSERT INTO evaluations VALUES(35791,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(35791,35,'cp 28',NULL);
INSERT INTO evaluations VALUES(35791,36,'cp 20','e5d4 f3d4 b7b5 c4d3 b5b4 c3e2 c6c5 d4f3 f8e8 e2g3 b8d7 b2b3 a7a5 c1b2 a5a4 a2a3 a4b3 a3b4 a8a1 d1a1 c5b4 c2b3 d7c5 f1d1 c5d3 d1d3 d8b6 a1d1 b6b7 b2f6 e7f6 d3d6 b7c7 e4e5 f6e7 d6d2 e8d8 g3e4 d8d2 d1d2 c8e6 f3d4 c7e5 d4c6 e5c7 c6e7 c7e7 e4d6 e6b3 d2b4');
INSERT INTO evaluations VALUES(37331,30,'cp 125',NULL);
INSERT INTO evaluations VALUES(37331,31,'cp 130',NULL);
INSERT INTO evaluations VALUES(37331,32,'cp 153',NULL);
INSERT INTO evaluations VALUES(37331,33,'cp 138',NULL);
INSERT INTO evaluations VALUES(37331,34,'cp 127',NULL);
INSERT INTO evaluations VALUES(37331,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(37331,36,'cp 137','h2h4 d7d6 f1d3 c7c5 e5d6 e7d5 d4c5 d5c3 b2c3 b8d7 c5c6 b7c6 e1g1 d7b6 c1f4 c8b7 f3e5 d8h4 d1d2 g7e5 f4e5 c6c5 e5g3 h4g4 a2a4 c5c4 d3e2 g4f5 f1b1 f5d5 d2d5 b7d5 a4a5 b6d7 b1b4 f8c8 f2f3 a7a6 g3h4 c8c5 h4f2 c5c6 g2g4 h7h5 g4h5 c6d6 h5g6 f7g6');
INSERT INTO evaluations VALUES(37612,30,'cp -567',NULL);
INSERT INTO evaluations VALUES(37612,31,'cp -572',NULL);
INSERT INTO evaluations VALUES(37612,32,'cp -575',NULL);
INSERT INTO evaluations VALUES(37612,33,'cp -556',NULL);
INSERT INTO evaluations VALUES(37612,34,'cp -563',NULL);
INSERT INTO evaluations VALUES(37612,35,'cp -578',NULL);
INSERT INTO evaluations VALUES(37612,36,'cp -590','d8g5 b1a3 g5e7 e1g1 b8c6 a3b5 e8f8 f1e1 h7h5 b5a3 a7a6 c2c3 d7d6 c4b3 b7b5 d2d3 c8b7 f3d1 g7f6 f2f4 a8d8 c1e3 b5b4 a3b1 f8g7 d3d4 b4c3 b2c3 c5d4 c3d4 e7c7 b1d2 d6d5 a1c1 d5e4 d2e4');
INSERT INTO evaluations VALUES(32970,30,'cp -573',NULL);
INSERT INTO evaluations VALUES(32970,31,'cp -559',NULL);
INSERT INTO evaluations VALUES(32970,32,'cp -559',NULL);
INSERT INTO evaluations VALUES(32970,33,'cp -554',NULL);
INSERT INTO evaluations VALUES(32970,34,'cp -563',NULL);
INSERT INTO evaluations VALUES(32970,35,'cp -566',NULL);
INSERT INTO evaluations VALUES(32970,36,'cp -555','b1a3 g5e7 e1g1 a7a6 c2c3 b8c6 f1e1 g8f6 f3d1 b7b5 c4f1 e8g8 d2d4 c5d4 c3d4 c8b7 a3c2 c6b4 c2b4 e7b4 e4e5 f6d5 g2g3 f8c8 h2h4 b4f8 f1g2 f7f5 e5f6 f8f6 c1g5 f6d4 d1d4 g7d4');
INSERT INTO evaluations VALUES(37880,30,'cp 105',NULL);
INSERT INTO evaluations VALUES(37880,31,'cp 81',NULL);
INSERT INTO evaluations VALUES(37880,32,'cp 90',NULL);
INSERT INTO evaluations VALUES(37880,33,'cp 74',NULL);
INSERT INTO evaluations VALUES(37880,34,'cp 93',NULL);
INSERT INTO evaluations VALUES(37880,35,'cp 93',NULL);
INSERT INTO evaluations VALUES(37880,36,'cp 101','e2e4 d5e4 c3e4 g8f6 f1d3 f6e4 d3e4 c6c5 e1g1 c5d4 f3d4 e8g8 c1f4 a7a6 d4f3 b8d7 d1e2 d7f6 f1d1 d8b6 e4c2 c8d7 a2a3 a6a5 d1d3 f8d8 d3b3 b6a6 f4e3 f6g4 b3b6 a6a7 e3d4 a8c8 f3e5 g4e5 e2e5');
INSERT INTO evaluations VALUES(38448,30,'cp 32',NULL);
INSERT INTO evaluations VALUES(38448,31,'cp 31',NULL);
INSERT INTO evaluations VALUES(38448,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(38448,33,'cp 14',NULL);
INSERT INTO evaluations VALUES(38448,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(38448,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(38448,36,'cp 34','g8f6 c4d5 f6d5 e2e4 d5c3 b2c3 c7c5 c1e3 d8a5 d1d2 b8c6 a1b1 c5d4 c3d4 e8g8 f1d3 c8g4 d2a5 c6a5 e1e2 f8d8 d4d5 e7e6 b1b5 b7b6 h2h3 g4f3 e2f3 e6d5 e4d5 a8c8 h1d1 a5b7 g2g4 g7f8 e3g5 d8d6 g5f4 d6d7 d1c1');
INSERT INTO evaluations VALUES(40320,30,'cp -36',NULL);
INSERT INTO evaluations VALUES(40320,31,'cp -26',NULL);
INSERT INTO evaluations VALUES(40320,32,'cp -34',NULL);
INSERT INTO evaluations VALUES(40320,33,'cp -36',NULL);
INSERT INTO evaluations VALUES(40320,34,'cp -38',NULL);
INSERT INTO evaluations VALUES(40320,35,'cp -33',NULL);
INSERT INTO evaluations VALUES(40320,36,'cp -22','b8c6 e1g1 a7a6 d3d4 c6d4 f3d4 c5d4 d1d4 e8g8 a2a4 d8a5 d4d2 c8e6 c4e6 f7e6 a1a3 f8f7 a3b3 f6g4 g5h4 b7b6 h4g3 a8c8 d2e2 g4e5 g3e5 g7e5 e2g4 c8c5 g4e6 e5c3 b2c3 c5c3 b3c3 a5c3 e6b3 c3b3 c2b3');
INSERT INTO evaluations VALUES(8241,30,'cp -13',NULL);
INSERT INTO evaluations VALUES(8241,31,'cp -6',NULL);
INSERT INTO evaluations VALUES(8241,32,'cp -13',NULL);
INSERT INTO evaluations VALUES(8241,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(8241,34,'cp -15',NULL);
INSERT INTO evaluations VALUES(8241,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(8241,36,'cp -18','e1g1 a7a6 d3d4 c6d4 f3d4 c5d4 d1d4 e8g8 c4b3 b7b5 a1d1 c8b7 e4e5 d6e5 d4e5 d8e8 f1e1 f6d5 e5g3 d5c3 b2c3 e7e6 c3c4 b5c4 b3c4 h7h6 g5f4 e8c6 c4d3 a8d8 d1b1 d8d7 h2h4 c6d5 a2a3 d5a5 g3e3 f8e8 f4e5 a5e5 e3e5 g7e5 e1e5');
INSERT INTO evaluations VALUES(40825,30,'cp -35',NULL);
INSERT INTO evaluations VALUES(40825,31,'cp -43',NULL);
INSERT INTO evaluations VALUES(40825,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(40825,33,'cp -45',NULL);
INSERT INTO evaluations VALUES(40825,34,'cp -51',NULL);
INSERT INTO evaluations VALUES(40825,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(40825,36,'cp -17','f6e4 d1e2 d6d5 f1e1 e8g8 b1c3 e4c3 e2e7 d8e7 e1e7 c7c5 b2c3 d5c4 d4b5 b7b6 a2a4 c8f5 e7c7 f8d8 c1f4 b8d7 a4a5 d7f8 b5d6 f5c2 f4g3 b6a5 a1a5 d8d7 c7d7 f8d7 d6c4 c2d3 c4d6 g8f8 d6b7 c5c4 f2f3 f8e7 g3f2 a7a6 b7c5 e7d6 c5d7');
INSERT INTO evaluations VALUES(42246,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(42246,31,'cp -15',NULL);
INSERT INTO evaluations VALUES(42246,32,'cp -4',NULL);
INSERT INTO evaluations VALUES(42246,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42246,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(42246,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(42246,36,'cp -10','b4c3 b2c3 e8g8 c1f4 c7c5 d4c5 b8c6 a1c1 d8a5 a3a4 f6e4 e2e3 e4c5 f4d6 f8d8 d6c5 a5c5 f1d3 g7g6 e1g1 c8g4 d3e2 a8c8 f3d4 g4e2 d4e2 b7b6 e2f4 c6e5 h2h3 c5d6 d1e2 c8c5 f1d1 d6d7 d1d4 e5c6 d4d1 c6a5 e2f3');
INSERT INTO evaluations VALUES(42260,30,'cp -9',NULL);
INSERT INTO evaluations VALUES(42260,31,'cp -16',NULL);
INSERT INTO evaluations VALUES(42260,32,'cp -5',NULL);
INSERT INTO evaluations VALUES(42260,33,'cp -3',NULL);
INSERT INTO evaluations VALUES(42260,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42260,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(42260,36,'cp 0','b2c3 e8g8 c1g5 b8d7 e2e3 h7h6 g5f6 d7f6 f1d3 c8g4 h2h3 g4f3 d1f3 c7c5 e1g1 c5c4 d3c2 d8a5 f1c1 b7b5 a3a4 b5b4 c3b4 a5b4 a1b1 b4a3 f3f4 a8c8 c1d1 a3a2 d1c1 a2a3');
INSERT INTO evaluations VALUES(42258,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42258,31,'cp -25',NULL);
INSERT INTO evaluations VALUES(42258,32,'cp -1',NULL);
INSERT INTO evaluations VALUES(42258,33,'cp -2',NULL);
INSERT INTO evaluations VALUES(42258,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(42258,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(42258,36,'cp -2','e8g8 c1g5 c8f5 e2e3 b8d7 f1e2 h7h6 g5f6 d7f6 e1g1 a8c8 a1c1 c7c5 c3c4 d5c4 c1c4 d8e7 c4c5 c8c5 d4c5 e7c5 f3d4 f5d7 d1b3 b7b6 e2c4 f6e4 f1c1 c5e7 c4e2 e4c5 b3b4 f8c8 h2h3 e7f6');
INSERT INTO evaluations VALUES(37121,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37121,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37121,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37121,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37121,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37121,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37121,36,'cp 0','c1g5 c7c5 e2e3 d8a5 g5f6 a5c3 f3d2 g7f6 a1c1 c3a3 c1a1 a3c3 a1c1');
INSERT INTO evaluations VALUES(37104,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(37104,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(37104,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(37104,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(37104,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(37104,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(37104,36,'cp 0','c7c5 e2e3 d8a5 g5f6 a5c3 f3d2 g7f6 a1c1 c3a3 c1a1 a3b4 a1b1 b4a3');
INSERT INTO evaluations VALUES(42474,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(42474,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(42474,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(42474,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(42474,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(42474,35,'cp 12',NULL);
INSERT INTO evaluations VALUES(42474,36,'cp 11','b1d2 c7c5 d4c5 b4c5 e2e3 b7b6 f1e2 a7a5 g1f3 e8g8 e1g1 c8b7 d2b3 c5e7 f3d4 b8a6 d4b5 d7d5 f4e5 d5c4 e2c4 b7d5 c4e2 f6e4 b3d2 e7f6 e5f6 e4f6 b5c3 a6c5 a1c1 d8e7 d1c2 f8c8 c3d5 f6d5 d2f3 a5a4 c2b1 h7h6 f1d1 a4a3');
INSERT INTO evaluations VALUES(42661,30,'cp -11',NULL);
INSERT INTO evaluations VALUES(42661,31,'cp -20',NULL);
INSERT INTO evaluations VALUES(42661,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(42661,33,'cp -12',NULL);
INSERT INTO evaluations VALUES(42661,34,'cp -25',NULL);
INSERT INTO evaluations VALUES(42661,35,'cp -24',NULL);
INSERT INTO evaluations VALUES(42661,36,'cp -9','g1f3 d7d6 f1e2 a7a5 e1g1 b8c6 h2h3 h7h6 c1e3 e8g8 c2c4 c8d7 b1c3 f8e8 d1c2 c6d4 e3d4 c5d4 f3d4 e5d4 c3b5 d7b5 c4b5 d8d7 a3a4 d6d5 e4d5 f6d5 e2f3 d5b4 c2c4 c7c6 a1e1 d7d6 b5c6 b7c6 e1e8 a8e8');
INSERT INTO evaluations VALUES(43862,30,'cp 111',NULL);
INSERT INTO evaluations VALUES(43862,31,'cp 89',NULL);
INSERT INTO evaluations VALUES(43862,32,'cp 79',NULL);
INSERT INTO evaluations VALUES(43862,33,'cp 100',NULL);
INSERT INTO evaluations VALUES(43862,34,'cp 107',NULL);
INSERT INTO evaluations VALUES(43862,35,'cp 84',NULL);
INSERT INTO evaluations VALUES(43862,36,'cp 82','e2e4 e7e5 d4e5 f6e4 f1d3 f8b4 d3e4 f5e4 d1h5 g7g6 h5h6 c8b7 c1g5 b4f8 g5d8 f8h6 d8c7 e8g8 c7d6 b8c6 a1d1 f8f7 c3d5 h6g7 d5f6 g7f6 e5f6 a8e8 g1e2 c6e5 d6e5 e8e5 e1g1 b6b5 d1d6 e4e3 f2f4 e5c5 f1d1 c5c4 d6d7');
INSERT INTO evaluations VALUES(44482,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(44482,31,'cp 28',NULL);
INSERT INTO evaluations VALUES(44482,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(44482,33,'cp 23',NULL);
INSERT INTO evaluations VALUES(44482,34,'cp 23',NULL);
INSERT INTO evaluations VALUES(44482,35,'cp 24',NULL);
INSERT INTO evaluations VALUES(44482,36,'cp 27','g1f3 d7d5 c2c4 c8f5 f1d3 f5d3 d1d3 e7e6 e1g1 f8e7 b1c3 e8g8 f1d1 b8d7 e3e4 d5e4 c3e4 d8c7 e4f6 d7f6 d3e2 f8d8 g2g3 h7h6 b2b3 b7b6 c1b2 c7b7 f3e5 f6d7 e5d3 e7f6 a1c1 c6c5 d4c5 f6b2 e2b2 d7c5 d3c5 b6c5');
INSERT INTO evaluations VALUES(44483,30,'cp 24',NULL);
INSERT INTO evaluations VALUES(44483,31,'cp 7',NULL);
INSERT INTO evaluations VALUES(44483,32,'cp 23',NULL);
INSERT INTO evaluations VALUES(44483,33,'cp 11',NULL);
INSERT INTO evaluations VALUES(44483,34,'cp 10',NULL);
INSERT INTO evaluations VALUES(44483,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(44483,36,'cp 17','d7d5 c2c4 c8f5 b1c3 e7e6 f3h4 f5g6 c1d2 b8d7 h4g6 h7g6 g2g3 f8b4 d1b3 b4a5 f1e2 d5c4 b3c4 e8g8 e1g1 d8e7 c4b3 e6e5 d4e5 e7e5 a1c1 d7c5 b3c2 c5e6 f1d1 a8d8 g1g2 b7b5 h2h4 f8e8 e2f3');
INSERT INTO evaluations VALUES(47484,30,'cp 55',NULL);
INSERT INTO evaluations VALUES(47484,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(47484,32,'cp 43',NULL);
INSERT INTO evaluations VALUES(47484,33,'cp 48',NULL);
INSERT INTO evaluations VALUES(47484,34,'cp 50',NULL);
INSERT INTO evaluations VALUES(47484,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(47484,36,'cp 58','f4e5 b8d7 c2c4 e7e6 e2e3 c7c6 b1c3 h5f6 e5g3 f6h5 g3h4 f8e7 h4e7 d8e7 f1d3 e8g8 e1g1 h5f6 a2a4 e6e5 d4e5 d7e5 c4d5 f6d5 c3d5 c6d5 f3d4 g7g6 d1d2 c8e6 d3e2 f8c8 f1d1 a7a6 h2h3 e5c6 a1c1 c6d4 d2d4 c8c1 d1c1');
INSERT INTO evaluations VALUES(47594,30,'cp 59',NULL);
INSERT INTO evaluations VALUES(47594,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(47594,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(47594,33,'cp 45',NULL);
INSERT INTO evaluations VALUES(47594,34,'cp 59',NULL);
INSERT INTO evaluations VALUES(47594,35,'cp 46',NULL);
INSERT INTO evaluations VALUES(47594,36,'cp 53','f1e2 b7b6 e1g1 c8b7 d1c2 f8e7 b2b3 e8g8 c1b2 b8d7 b1c3 a8c8 a1c1 c7c5 f1d1 f6e4 d4c5 e4c3 b2c3 d7c5 f3d2 c5d7 c4d5 e6d5 c2b2 e7f6 d2b1 d8e7 c3f6 d7f6 b1c3 c8c5 b3b4 c5c8 a2a4 c8c7 c3b5 c7c1 d1c1 f8c8');
INSERT INTO evaluations VALUES(48911,30,'cp 107',NULL);
INSERT INTO evaluations VALUES(48911,31,'cp 98',NULL);
INSERT INTO evaluations VALUES(48911,32,'cp 96',NULL);
INSERT INTO evaluations VALUES(48911,33,'cp 95',NULL);
INSERT INTO evaluations VALUES(48911,34,'cp 104',NULL);
INSERT INTO evaluations VALUES(48911,35,'cp 98',NULL);
INSERT INTO evaluations VALUES(48911,36,'cp 104','d1b3 d8e7 d4e5 f6d7 f2f4 b8c6 g1f3 c6a5 b3c3 a5c4 c3c4 e7c5 c4c5 d7c5 b1c3 c8f5 e1g1 e8c8 f3d4 f5g6 f4f5 g6h5 h2h3 f7f6 e5e6 c5d3 a1b1 a7a6 c1d2 c7c5 d4f3 h5e8 e3e4 e8c6 d2e3 g7g6 f3e1 f8e7 b2b3 b7b6 b1d1 d3b4 e1f3 g6f5 e4f5 h8g8 d1d8 g8d8 g1f2 c6b7 a2a3 b4d3 f2e2');
INSERT INTO evaluations VALUES(54259,30,'cp -18',NULL);
INSERT INTO evaluations VALUES(54259,31,'cp -27',NULL);
INSERT INTO evaluations VALUES(54259,32,'cp -23',NULL);
INSERT INTO evaluations VALUES(54259,33,'cp -13',NULL);
INSERT INTO evaluations VALUES(54259,34,'cp -13',NULL);
INSERT INTO evaluations VALUES(54259,35,'cp -15',NULL);
INSERT INTO evaluations VALUES(54259,36,'cp -31','d5d4 c3b1 f8d6 g2g3 b8c6 f1c4 h7h5 d1e2 h5h4 e4e5 d6c7 g3g4 g8e7 d2d3 e7g6 a2a4 a7a6 b1a3 c8d7 c1d2 d8e7 e1c1 e8c8 d1e1 c8b8 h1f1 c6b4 b2b3 b4a2 c1b1 a2c3 d2c3 d4c3 e2e3 h4h3 f3g5 d7c6 g5e4 e7h4');
INSERT INTO evaluations VALUES(54113,30,'cp -14',NULL);
INSERT INTO evaluations VALUES(54113,31,'cp -12',NULL);
INSERT INTO evaluations VALUES(54113,32,'cp -20',NULL);
INSERT INTO evaluations VALUES(54113,33,'cp -16',NULL);
INSERT INTO evaluations VALUES(54113,34,'cp -21',NULL);
INSERT INTO evaluations VALUES(54113,35,'cp -1',NULL);
INSERT INTO evaluations VALUES(54113,36,'cp -1','c3b1 f8d6 g2g3 b8c6 f1c4 g8e7 a2a4 c6a5 b1a3 a5c4 a3c4 d4d3 c2d3 e7c6 d3d4 b7b6 b2b3 c8a6 c1a3 d6e7 d4c5 b6c5 d2d3 f7f6 a1c1 d8d7 e1g1 a8d8 e4e5 d7d3 d1d3 d8d3 e5f6 g7f6');
INSERT INTO evaluations VALUES(55890,30,'cp -96',NULL);
INSERT INTO evaluations VALUES(55890,31,'cp -93',NULL);
INSERT INTO evaluations VALUES(55890,32,'cp -83',NULL);
INSERT INTO evaluations VALUES(55890,33,'cp -87',NULL);
INSERT INTO evaluations VALUES(55890,34,'cp -90',NULL);
INSERT INTO evaluations VALUES(55890,35,'cp -94',NULL);
INSERT INTO evaluations VALUES(55890,36,'cp -81','d5d4 c3b1 f8d6 d2d3 g8e7 h2h4 e8g8 h4h5 h7h6 b1d2 b8a6 g2g4 f7f6 a2a3 e6e5 f3g3 b7b5 g1f3 a6c5 f3h4 c8e6 d2f3 a7a5 h4g6 f8f7 f1e2 b5b4 g4g5 f6g5 c1g5 h6g5 f3g5 f7f6 g3h4 b4a3 b2a3 e6f7 e2g4 f7g6 h5g6');
INSERT INTO evaluations VALUES(56096,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(56096,31,'cp 20',NULL);
INSERT INTO evaluations VALUES(56096,32,'cp 25',NULL);
INSERT INTO evaluations VALUES(56096,33,'cp 30',NULL);
INSERT INTO evaluations VALUES(56096,34,'cp 27',NULL);
INSERT INTO evaluations VALUES(56096,35,'cp 20',NULL);
INSERT INTO evaluations VALUES(56096,36,'cp 15','g8f6 d3d4 f8e7 c1f4 e8g8 f1e2 c7c5 d4c5 e7c5 e1g1 b8c6 a2a3 a7a6 e2d3 c5e7 f1e1 b7b5 d1e2 c8b7 a1d1 d8b6 c3e4 c6d4 f3d4 b6d4 c2c3 d4b6 f4g5 b7e4 d3e4 a8c8 e4f3 b6c7 e2e5 c7e5 e1e5 f8d8 g5e3 d8d1 f3d1');
INSERT INTO evaluations VALUES(57167,30,'cp -124',NULL);
INSERT INTO evaluations VALUES(57167,31,'cp -94',NULL);
INSERT INTO evaluations VALUES(57167,32,'cp -89',NULL);
INSERT INTO evaluations VALUES(57167,33,'cp -95',NULL);
INSERT INTO evaluations VALUES(57167,34,'cp -101',NULL);
INSERT INTO evaluations VALUES(57167,35,'cp -88',NULL);
INSERT INTO evaluations VALUES(57167,36,'cp -86','g8f6 a3c4 f6e4 c4e5 d8e7 g1f3 f7f6 d1e2 d7d5 e5d3 g7g5 b2b4 a7a5 b4b5 c7c5 b5c6 b8c6 c1b2 a5a4 a2a3 e8f7 f3d4 h7h5 f2f3 e4d6 h2h4 d6c4 b2c3 c6d4 c3d4 g5g4 e2e7 f8e7 d3f4 g4f3 f1c4 d5c4 g2f3 c8f5 f4d5 h8e8 d5e7 e8e7 e1f2 f5c2 a1e1 e7e1 h1e1 c2f5');
INSERT INTO evaluations VALUES(870,30,'cp 432',NULL);
INSERT INTO evaluations VALUES(870,31,'cp 425',NULL);
INSERT INTO evaluations VALUES(870,32,'cp 415',NULL);
INSERT INTO evaluations VALUES(870,33,'cp 419',NULL);
INSERT INTO evaluations VALUES(870,34,'cp 407',NULL);
INSERT INTO evaluations VALUES(870,35,'cp 412',NULL);
INSERT INTO evaluations VALUES(870,36,'cp 418','f1a6 b7a6 d2d3 e7b4 b1c3 f8d6 c7c4 b4c4 d3c4 d6b4 c1d2 b4c3 d2c3 f6e4 c3b4 c8b7 f2f3 e4g5 g1e2 e8c8 b4a5 d8e8 e1f2 b7c6 b2b3 g5e6 h2h4 h8g8 a1d1 g7g6 a5c3 h7h5 h1e1 g6g5 e2g3 g5h4 g3f5 c8b8 f5h4 g8g5 e1e5 e8g8 e5f5 g5f5 h4f5 g8g5 f5d6 e6f4 d1g1 h5h4 d6f7 g5g3 f7d6 b8c7');
INSERT INTO evaluations VALUES(6864,30,'cp -312',NULL);
INSERT INTO evaluations VALUES(6864,31,'cp -311',NULL);
INSERT INTO evaluations VALUES(6864,32,'cp -310',NULL);
INSERT INTO evaluations VALUES(6864,33,'cp -315',NULL);
INSERT INTO evaluations VALUES(6864,34,'cp -303',NULL);
INSERT INTO evaluations VALUES(6864,35,'cp -322',NULL);
INSERT INTO evaluations VALUES(6864,36,'cp -310','d4c5 d7d6 d1e2 c8e6 b2b3 c4e5 f2f4 e6g4 e2d2 e5c6 c5d6 g8f6 b1c3 d8d6 d2d6 c7d6 h2h3 g4d7 c1b2 e8c8 e1c1 f6e8 h1e1 h7h5 c3b5 f7f6 c2c4 d7e6 f4f5 e6f7 e1e3 h8g8 e3d3 a7a6 b5c3 h5h4 d3d2 g7g6 b2a3 g6f5 e4f5 a6a5 a3d6 e8d6 d2d6 d8d6 d1d6');
INSERT INTO evaluations VALUES(6833,30,'cp -337',NULL);
INSERT INTO evaluations VALUES(6833,31,'cp -325',NULL);
INSERT INTO evaluations VALUES(6833,32,'cp -317',NULL);
INSERT INTO evaluations VALUES(6833,33,'cp -334',NULL);
INSERT INTO evaluations VALUES(6833,34,'cp -325',NULL);
INSERT INTO evaluations VALUES(6833,35,'cp -315',NULL);
INSERT INTO evaluations VALUES(6833,36,'cp -319','d7d6 d1e2 c8e6 b2b3 c4e5 f2f4 e6g4 e2d2 e5c6 c5d6 g8f6 b1c3 d8d6 d2d6 c7d6 h2h3 g4d7 c1b2 e8c8 e1c1 f6e8 c3b5 f7f6 h1e1 d7e6 c2c4 h7h5 d1d2 h5h4 f4f5 e6f7 e1e3 d8d7 b2a3 a7a6 b5c3 c6e5 e3e1 g7g6 a3b2 d7e7 d2d1 g6f5 e4f5 h8h7');
INSERT INTO evaluations VALUES(9983,30,'cp 91',NULL);
INSERT INTO evaluations VALUES(9983,31,'cp 76',NULL);
INSERT INTO evaluations VALUES(9983,32,'cp 73',NULL);
INSERT INTO evaluations VALUES(9983,33,'cp 87',NULL);
INSERT INTO evaluations VALUES(9983,34,'cp 69',NULL);
INSERT INTO evaluations VALUES(9983,35,'cp 75',NULL);
INSERT INTO evaluations VALUES(9983,36,'cp 65','e4e5 f6e4 d1c2 d7d5 g1f3 f7f5 e5f6 d8f6 c1e3 e8g8 f1d3 b4a5 a2a3 a5c3 b2c3 h7h6 e1g1 e4d6 c4c5 d6c4 f1e1 b7b6 h2h3 b6c5 d3c4 d5c4 c2e4 c8d7 f3e5 c6e5 d4e5 f6f5 e3c5 f8d8 e4c4 d7c6 c4g4 f5g4 h3g4 c6d5 c5d4 d5c4 f2f3 a7a5');
INSERT INTO evaluations VALUES(12233,30,'cp 54',NULL);
INSERT INTO evaluations VALUES(12233,31,'cp 49',NULL);
INSERT INTO evaluations VALUES(12233,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(12233,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12233,34,'cp 39',NULL);
INSERT INTO evaluations VALUES(12233,35,'cp 52',NULL);
INSERT INTO evaluations VALUES(12233,36,'cp 48','e2e3 e7e6 g1f3 f8e7 f1d3 d5c4 d3c4 e8g8 e1g1 a7a6 a2a4 b7b5 c4d3 c8b7 d1e2 b5b4 c3e4 c6c5 e4f6 e7f6 f1c1 c5d4 f4c7 d8e7 e3d4 f6g5 f3g5 e7g5 f2f3 f8c8 g1h1 d7f6 c1c5 g5h6 a4a5 f6d5 c7g3 d5f4 g3f4 h6f4 d3e4 c8c5 d4c5');
INSERT INTO evaluations VALUES(12231,30,'cp 45',NULL);
INSERT INTO evaluations VALUES(12231,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(12231,32,'cp 37',NULL);
INSERT INTO evaluations VALUES(12231,33,'cp 44',NULL);
INSERT INTO evaluations VALUES(12231,34,'cp 34',NULL);
INSERT INTO evaluations VALUES(12231,35,'cp 35',NULL);
INSERT INTO evaluations VALUES(12231,36,'cp 44','e7e6 g1f3 f8e7 f1d3 e8g8 c4d5 e6d5 d1c2 f8e8 e1g1 d7f8 h2h3 f8g6 f4e5 e7d6 e5d6 d8d6 f1b1 a7a5 a2a3 d6e7 c3a4 f6e4 a4b6 a8b8 b6c8 b8c8 b2b4 a5b4 a3b4 h7h6 c2b3 b7b6 b1c1 e4g5 f3d2 c8a8 a1a8 e8a8');
INSERT INTO evaluations VALUES(13483,30,'cp 22',NULL);
INSERT INTO evaluations VALUES(13483,31,'cp 24',NULL);
INSERT INTO evaluations VALUES(13483,32,'cp 19',NULL);
INSERT INTO evaluations VALUES(13483,33,'cp 16',NULL);
INSERT INTO evaluations VALUES(13483,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(13483,35,'cp 15',NULL);
INSERT INTO evaluations VALUES(13483,36,'cp 14','e1g1 f8d6 f1e1 e8g8 c1g5 h7h6 g5h4 c8d7 a2a4 a7a6 b5c6 d7c6 c3e2 b7b5 f3e5 c6e8 a4b5 e8b5 e2c3 b5e8 e5g4 d6e7 h4f6 e7f6 g4f6 d8f6 d1d2 a6a5 c3a4 e8a4 a1a4 f8c8 c2c3 c8b8 e1a1 f6d8 g2g3 d8b6');
INSERT INTO evaluations VALUES(15643,30,'cp 116',NULL);
INSERT INTO evaluations VALUES(15643,31,'cp 127',NULL);
INSERT INTO evaluations VALUES(15643,32,'cp 130',NULL);
INSERT INTO evaluations VALUES(15643,33,'cp 145',NULL);
INSERT INTO evaluations VALUES(15643,34,'cp 109',NULL);
INSERT INTO evaluations VALUES(15643,35,'cp 117',NULL);
INSERT INTO evaluations VALUES(15643,36,'cp 127','f1d3 c8g4 c1e3 d6d5 e4e5 f6d7 h2h3 g4h5 b1c3 d7b6 a2a3 d8d7 d1c2 h5g6 e1g1 g6d3 c2d3 f8e7 a1e1 c6d8 f3h2 e8g8 f2f4 f7f5 e5f6 e7f6 h2g4 d8c6 g4f6 f8f6 f4f5 a8e8 g2g4 f6f8 b2b3 c6a5 d3c2 c7c6 e3g5 e8e1 f1e1 f8e8');
INSERT INTO evaluations VALUES(15682,30,'cp 38',NULL);
INSERT INTO evaluations VALUES(15682,31,'cp 34',NULL);
INSERT INTO evaluations VALUES(15682,32,'cp 42',NULL);
INSERT INTO evaluations VALUES(15682,33,'cp 59',NULL);
INSERT INTO evaluations VALUES(15682,34,'cp 57',NULL);
INSERT INTO evaluations VALUES(15682,35,'cp 43',NULL);
INSERT INTO evaluations VALUES(15682,36,'cp 67','a7a6 b5d3 c6b4 d3e2 f8e7 e1g1 b7b5 a2a3 b4c6 f1e1 e8g8 e2d3 c8b7 h2h3 c6b8 d1e2 f6d7 a3a4 b5b4 c3d1 c7c5 d4d5 e7f6 d1e3 e6d5 e4d5 f8e8 e2d1 d7e5 f3e5 f6e5 c2c3 b8d7 c1d2 d8h4 c3b4 e5b2 a1b1 b2e5 b4c5 d7c5 e3f5 h4a4');
INSERT INTO evaluations VALUES(17680,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(17680,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(17680,32,'cp 13',NULL);
INSERT INTO evaluations VALUES(17680,33,'cp 12',NULL);
INSERT INTO evaluations VALUES(17680,34,'cp 29',NULL);
INSERT INTO evaluations VALUES(17680,35,'cp 31',NULL);
INSERT INTO evaluations VALUES(17680,36,'cp 12','c6d4 d1d4 g8f6 h2h3 f8e7 a2a4 c8d7 c4b3 b7b5 f2f4 b5a4 b3a4 d7a4 a1a4 e8g8 c1e3 f6d7 d4d3 a6a5 f1a1 d7c5 e3c5 d6c5 e4e5 d8d3 c2d3 f8d8 a4a5 a8b8 a5b5 b8b5 c3b5 d8d3');
INSERT INTO evaluations VALUES(21998,30,'cp -8',NULL);
INSERT INTO evaluations VALUES(21998,31,'cp -13',NULL);
INSERT INTO evaluations VALUES(21998,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(21998,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(21998,34,'cp -22',NULL);
INSERT INTO evaluations VALUES(21998,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(21998,36,'cp 2','h7h5 g1f3 e5e4 f3d4 c6d4 b2d4 d6d5 d2d3 f7f5 h2h4 g8f6 e1g1 c7c5 d4e3 f8d6 d3e4 f5e4 c2c4 d6e5 b1d2 e5a1 d1a1 d5d4 e3g5 c8f5 b3b4 d8e7 b4c5 a8d8 d2b3 e8g8 b3d4 e7c5 g5f6 f8f6 d4f5 c5f5 a1b1 d8d4 b1b7 d4c4 b7a7 c4c2 a7a8 c2c8');
INSERT INTO evaluations VALUES(22064,30,'cp 43',NULL);
INSERT INTO evaluations VALUES(22064,31,'cp 36',NULL);
INSERT INTO evaluations VALUES(22064,32,'cp 40',NULL);
INSERT INTO evaluations VALUES(22064,33,'cp 54',NULL);
INSERT INTO evaluations VALUES(22064,34,'cp 51',NULL);
INSERT INTO evaluations VALUES(22064,35,'cp 58',NULL);
INSERT INTO evaluations VALUES(22064,36,'cp 65','b1c3 g8f6 d2d4 d6d5 c4d3 c6b4 e4e5 b4d3 d1d3 f6g8 c3e2 g8e7 e2g3 h7h6 e1g1 a7a5 f1e1 b7b6 g3h5 c8a6 d3d1 a5a4 b2b3 e7f5 h2h3 d8d7 c2c3 d7b5 c1d2 b5d7 a1c1 a6b7 d2e3 e8d8 e3f4 f5e7 f4d2 d8c8 h5g3 e7c6 b3b4');
INSERT INTO evaluations VALUES(22325,30,'cp 139',NULL);
INSERT INTO evaluations VALUES(22325,31,'cp 165',NULL);
INSERT INTO evaluations VALUES(22325,32,'cp 159',NULL);
INSERT INTO evaluations VALUES(22325,33,'cp 151',NULL);
INSERT INTO evaluations VALUES(22325,34,'cp 148',NULL);
INSERT INTO evaluations VALUES(22325,35,'cp 141',NULL);
INSERT INTO evaluations VALUES(22325,36,'cp 150','b1c3 f8e7 a2a4 a7a6 h2h3 e7f6 b2b4 f5f4 a1b1 g8h6 b4b5 a6b5 a4b5 c6e7 d3d4 h6f7 e1g1 e7g6 d4e5 f6e5 c4f7 e8f7 c3d5 c7c6 d5b4 c6b5 b4d5 h8f8 b1b5 f7g8 c1b2 g8h8 b2e5 g6e5 f3e5 d6e5 d5b6 a8a6 b6c8 d8d1 f1d1 f8c8');
INSERT INTO evaluations VALUES(22323,30,'cp 140',NULL);
INSERT INTO evaluations VALUES(22323,31,'cp 146',NULL);
INSERT INTO evaluations VALUES(22323,32,'cp 132',NULL);
INSERT INTO evaluations VALUES(22323,33,'cp 133',NULL);
INSERT INTO evaluations VALUES(22323,34,'cp 150',NULL);
INSERT INTO evaluations VALUES(22323,35,'cp 139',NULL);
INSERT INTO evaluations VALUES(22323,36,'cp 162','c6a5 c4d5 g8f6 f3g5 f6d5 e4d5 c7c5 a2a3 b7b6 b2b4 a5b7 d1f3 g7g6 g5e6 c8e6 d5e6 d8c8 c3d5 b7d8 b4c5 d8e6 c5b6 a7b6 e1g1 a8b8 c1e3 f8e7 e3b6 e6g5 f3g3 g5e6 f1b1 c8c2 a3a4 e7d8 a4a5 e8g8 g3e3 f5f4 e3e4 e6c5 e4c4 c2c4');
INSERT INTO evaluations VALUES(22868,30,'cp -39',NULL);
INSERT INTO evaluations VALUES(22868,31,'cp -37',NULL);
INSERT INTO evaluations VALUES(22868,32,'cp -31',NULL);
INSERT INTO evaluations VALUES(22868,33,'cp -21',NULL);
INSERT INTO evaluations VALUES(22868,34,'cp -36',NULL);
INSERT INTO evaluations VALUES(22868,35,'cp -42',NULL);
INSERT INTO evaluations VALUES(22868,36,'cp -33','d7d5 d3d4 e5e4 f3d2 f7f5 c2c4 g8f6 b1c3 c8e6 a2a3 f8e7 c4d5 f6d5 f1c4 c6a5 c3d5 e6d5 c4d5 d8d5 e1g1 e8g8 d1c2 a7a6 b2b4 a5c6 c2b3 d5f7 b3f7 f8f7 a1b1 e7d6 d2c4 b7b5 c4d6 c7d6 c1d2 c6e7');
INSERT INTO evaluations VALUES(23646,30,'cp 51',NULL);
INSERT INTO evaluations VALUES(23646,31,'cp 39',NULL);
INSERT INTO evaluations VALUES(23646,32,'cp 45',NULL);
INSERT INTO evaluations VALUES(23646,33,'cp 58',NULL);
INSERT INTO evaluations VALUES(23646,34,'cp 46',NULL);
INSERT INTO evaluations VALUES(23646,35,'cp 55',NULL);
INSERT INTO evaluations VALUES(23646,36,'cp 41','f2f4 a7a5 d1f3 a5a4 g1h1 d8c8 h2h3 f6d7 c1e3 c8b7 e3d4 f8e8 a1d1 b7b2 c3d5 b2b7 d1b1 b7c8 d5e7 e8e7 f3c3 f7f6 c3c6 c8a6 c6d5 g8h8 h1h2 d7f8 d4f2 f8g6 g2g3 h7h6 h2g2 g6f8 a2a3 f8g6 c2c4 g6f8 c4c5 d6c5 f2c5 e7e8');
INSERT INTO evaluations VALUES(24669,30,'cp 25',NULL);
INSERT INTO evaluations VALUES(24669,31,'cp 27',NULL);
INSERT INTO evaluations VALUES(24669,32,'cp 18',NULL);
INSERT INTO evaluations VALUES(24669,33,'cp 33',NULL);
INSERT INTO evaluations VALUES(24669,34,'cp 32',NULL);
INSERT INTO evaluations VALUES(24669,35,'cp 26',NULL);
INSERT INTO evaluations VALUES(24669,36,'cp 2','e1g1 a7a6 c4e2 e7e6 h2h3 g4f3 e2f3 c6d4 d2d3 f8e7 g2g3 d4f3 d1f3 d6d5 a2a4 d8c7 e4d5 e6d5 c3d5 f6d5 f3d5 e8g8 c1e3 a8d8 d5c4 c7d7 g1h2 e7f6 a1e1 d8c8 b2b3 b7b5 a4b5 a6b5 c4e4 f8e8 e4f3 c5c4 d3c4 b5c4 e1e2 d7e6 f1e1 c4b3 c2b3 e6b3');
INSERT INTO evaluations VALUES(24668,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(24668,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(24668,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(24668,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(24668,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(24668,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(24668,36,'cp 14','a7a6 c4e2 e7e6 h2h3 g4f3 e2f3 c6d4 d2d3 f8e7 a2a4 e8g8 a4a5 f6d7 c1d2 e7f6 g1h1 d8c7 g2g4 h7h6 f3g2 a8c8 a1a4 c7d8 c3b1 c5c4 a4c4 d4c6 d2c1 e6e5 b1c3 e5f4 c1f4 f6e5 f4e3 d8a5 d3d4 b7b5 d4e5 b5c4 e5d6');
INSERT INTO evaluations VALUES(25456,30,'cp 44',NULL);
INSERT INTO evaluations VALUES(25456,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(25456,32,'cp 50',NULL);
INSERT INTO evaluations VALUES(25456,33,'cp 56',NULL);
INSERT INTO evaluations VALUES(25456,34,'cp 56',NULL);
INSERT INTO evaluations VALUES(25456,35,'cp 59',NULL);
INSERT INTO evaluations VALUES(25456,36,'cp 58','b7c6 d1f3 f8e7 e4e5 f6d5 d4c6 d7c6 c3d5 d8d7 e1g1 d6e5 f1d1 a8d8 f3g3 e8f8 c2c4 f7f6 c1e3 a7a6 d1d2 c6d5 c4d5 f8f7 a1c1 h8e8 h2h3 f7g8 e3c5 e5e4 b2b4 h7h6 g3f4 e7d6 c5d6 c7d6 d2d4 f6f5 c1c6 d7a7 d4d1 e4e3 f2e3 e8e3 g1h2');
INSERT INTO evaluations VALUES(24386,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(24386,31,'cp 56',NULL);
INSERT INTO evaluations VALUES(24386,32,'cp 54',NULL);
INSERT INTO evaluations VALUES(24386,33,'cp 80',NULL);
INSERT INTO evaluations VALUES(24386,34,'cp 87',NULL);
INSERT INTO evaluations VALUES(24386,35,'cp 62',NULL);
INSERT INTO evaluations VALUES(24386,36,'cp 71','d1f3 f8e7 e4e5 f6d5 d4c6 d7c6 c3d5 d8d7 e1g1 d6e5 f1d1 e7d6 c2c4 f7f5 c4c5 d6e7 f3b3 c6d5 d1d5 d7e6 c1d2 c7c6 b3b7 e8f7 d5d6 h8b8 b7c7 b8c8 d6e6 c8c7 e6e5 c7d7 d2c3 g7g6 a1e1 e7f6 e5e6 f6c3 b2c3 d7c7 e6d6 a8b8 h2h4 b8b2 h4h5 b2a2 h5g6 h7g6 e1e6 c7e7 e6g6 e7e1 g1h2 a2f2');
INSERT INTO evaluations VALUES(25587,30,'cp 16',NULL);
INSERT INTO evaluations VALUES(25587,31,'cp 12',NULL);
INSERT INTO evaluations VALUES(25587,32,'cp 8',NULL);
INSERT INTO evaluations VALUES(25587,33,'cp 10',NULL);
INSERT INTO evaluations VALUES(25587,34,'cp 12',NULL);
INSERT INTO evaluations VALUES(25587,35,'cp 6',NULL);
INSERT INTO evaluations VALUES(25587,36,'cp 8','c2c4 f8e7 a1c1 f6h5 f4e5 c7c6 c4c5 e8g8 b2b4 h5f6 f1e2 f6e4 e1g1 e4d2 d1d2 d7e5 f3e5 e7f6 a2a4 a7a6 d2c3 f6e5 d4e5 d8e7 f2f4 f5e4 h2h3 f7f6 e5f6 f8f6 c3e5 a8e8 e2g4 f6g6 g4f3 e4f3 f1f3 g8h8 e5d6 e6e5 d6e7');
INSERT INTO evaluations VALUES(33693,30,'cp 56',NULL);
INSERT INTO evaluations VALUES(33693,31,'cp 62',NULL);
INSERT INTO evaluations VALUES(33693,32,'cp 62',NULL);
INSERT INTO evaluations VALUES(33693,33,'cp 43',NULL);
INSERT INTO evaluations VALUES(33693,34,'cp 42',NULL);
INSERT INTO evaluations VALUES(33693,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(33693,36,'cp 42','b8c6 e1g1 e7e5 b1c3 d5a5 d2d4 c6d4 f3d4 e5d4 d1d4 f8e7 c3b5 e8g8 c1f4 f6e8 d4e3 e7f6 c2c3 c7c6 b5d4 f6d4 e3d4 e8f6 f1e1 f8d8 d4e5 a5e5 f4e5 c8f5 e5d4 a7a5 a2a4 f5e6 h2h4 f6e4 e2f1 e4f6 f2f3 e6b3 d4b6 d8d2');
INSERT INTO evaluations VALUES(34770,30,'cp 21',NULL);
INSERT INTO evaluations VALUES(34770,31,'cp 37',NULL);
INSERT INTO evaluations VALUES(34770,32,'cp 24',NULL);
INSERT INTO evaluations VALUES(34770,33,'cp 27',NULL);
INSERT INTO evaluations VALUES(34770,34,'cp 19',NULL);
INSERT INTO evaluations VALUES(34770,35,'cp 21',NULL);
INSERT INTO evaluations VALUES(34770,36,'cp 19','d2d4 b8c6 c1g5 f6g6 h5g6 h7g6 b1c3 c6d4 f3d4 e5d4 c3b5 e8d7 b5d4 f8e7 g5d2 e7f6 c2c3 a7a6 f1c4 d7e8 e1c1 g8e7 h2h3 b7b5 c4b3 c8b7 h1e1 h8h4 f2f3 e7c6 d4c6 b7c6 d2e3 a6a5 b3d5 c6d5 d1d5 b5b4 c1c2 b4c3 b2c3 a5a4 e3d4 f6d4 c3d4');
INSERT INTO evaluations VALUES(34873,30,'cp 47',NULL);
INSERT INTO evaluations VALUES(34873,31,'cp 43',NULL);
INSERT INTO evaluations VALUES(34873,32,'cp 47',NULL);
INSERT INTO evaluations VALUES(34873,33,'cp 36',NULL);
INSERT INTO evaluations VALUES(34873,34,'cp 38',NULL);
INSERT INTO evaluations VALUES(34873,35,'cp 37',NULL);
INSERT INTO evaluations VALUES(34873,36,'cp 43','b8c6 d2d4 c6d4 f3d4 e5d4 c3b5 f8b4 c1d2 e6e2 f1e2 b4d2 e1d2 e8d8 b5d4 g8f6 e2f3 c7c6 a1e1 c8d7 c2c4 h8e8 e1e8 d8e8 h1e1 e8f8 d2c3 a8e8 e1e8 f8e8 c4c5 e8d8 g2g3 f6g4 f3g4 d7g4 a2a3 g4h3 d4b3 d8d7 b3a5 b7b6 c5b6 a7b6 a5c4 b6b5 c4e3 d7e8');
INSERT INTO evaluations VALUES(1689,30,'cp 42',NULL);
INSERT INTO evaluations VALUES(1689,31,'cp 42',NULL);
INSERT INTO evaluations VALUES(1689,32,'cp 46',NULL);
INSERT INTO evaluations VALUES(1689,33,'cp 47',NULL);
INSERT INTO evaluations VALUES(1689,34,'cp 43',NULL);
INSERT INTO evaluations VALUES(1689,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(1689,36,'cp 41','d2d4 c6d4 f3d4 e5d4 c3b5 f8b4 c1d2 b4d2 e1d2 e6e2 f1e2 e8d8 b5d4 g8f6 e2f3 c7c6 h1e1 c8d7 c2c4 h8e8 e1e8 f6e8 d2c3 e8d6 a1d1 d8c7 d1e1 a8e8 e1e8 d6e8 c4c5 e8f6 g2g3 f6g4 f3g4 d7g4 f2f3 g4h3 g3g4 g7g5 d4f5 h3g2 f5d6 g2f3 d6f7');
INSERT INTO evaluations VALUES(38201,30,'cp -12',NULL);
INSERT INTO evaluations VALUES(38201,31,'cp -4',NULL);
INSERT INTO evaluations VALUES(38201,32,'cp -8',NULL);
INSERT INTO evaluations VALUES(38201,33,'cp -8',NULL);
INSERT INTO evaluations VALUES(38201,34,'cp -6',NULL);
INSERT INTO evaluations VALUES(38201,35,'cp -8',NULL);
INSERT INTO evaluations VALUES(38201,36,'cp 0','d2b3 c5b6 c1e3 b6e3 d1e2 d8e7 e2e3 e7e3 f2e3 g8f6 g1e2 b8c6 e2c3 e8g8 e1c1 c8g4 d1d2 a8e8 f1b5 a7a6 b5e2 g4e2 d2e2 e8e5 h1d1 f8e8 e2d2 e5e3 c3d5 f6d5 d2d5 g7g6 b3c5 e8e7 d5d7 e7d7 d1d7 e3e2 d7d2 e2e1 d2d1 e1e2');
INSERT INTO evaluations VALUES(38185,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38185,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38185,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38185,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38185,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38185,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(38185,36,'cp 0','c5b6 c1e3 b6e3 d1e2 d8e7 e2e3 e7e3 f2e3 g8f6 g1e2 b8c6 e2c3 e8g8 e1c1 c8g4 d1d2 a8e8 f1b5 a7a6 b5e2 g4e2 d2e2 e8e5 h1d1 f8e8 e2d2 e5e3 c3d5 f6d5 d2d5 g7g6 b3c5 e8e7 d5d7 e7d7 d1d7 e3e2 d7d2 e2e1 d2d1 e1e2');
INSERT INTO evaluations VALUES(38032,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(38032,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(38032,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(38032,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(38032,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(38032,35,'cp -6',NULL);
INSERT INTO evaluations VALUES(38032,36,'cp 0','c1e3 b6e3 d1e2 d8e7 e2e3 e7e3 f2e3 g8f6 g1e2 b8c6 e2c3 e8g8 e1c1 c8g4 d1e1 a8e8 f1b5 e8e7 h2h3 g4h5 b3d4 f8c8 g2g4 h5g6 h1f1 h7h6 a2a3 g6e4 f1f4 c6e5 e1d1 a7a6 b5e2 b7b5 d4f5 e4f5 f4f5 c8d8 g4g5 h6g5 f5g5');
INSERT INTO evaluations VALUES(38567,30,'cp 14',NULL);
INSERT INTO evaluations VALUES(38567,31,'cp 33',NULL);
INSERT INTO evaluations VALUES(38567,32,'cp 41',NULL);
INSERT INTO evaluations VALUES(38567,33,'cp 26',NULL);
INSERT INTO evaluations VALUES(38567,34,'cp 20',NULL);
INSERT INTO evaluations VALUES(38567,35,'cp 23',NULL);
INSERT INTO evaluations VALUES(38567,36,'cp 22','g8f6 f1e2 e8g8 e1g1 d8e7 a1b1 a7a5 b2b3 b7b6 c1b2 c8b7 d1c2 b8d7 f1d1 f8c8 f3d2 c7c5 b2a3 e7d8 d4c5 b6c5 e2f3 d7e5 c3a4 f6d7 a3b2 d8e7 b1c1 e5f3 d2f3 d5d4 e3d4 b7f3 g2f3 c5d4 b2d4 g7d4');
INSERT INTO evaluations VALUES(39525,30,'cp 65',NULL);
INSERT INTO evaluations VALUES(39525,31,'cp 92',NULL);
INSERT INTO evaluations VALUES(39525,32,'cp 82',NULL);
INSERT INTO evaluations VALUES(39525,33,'cp 76',NULL);
INSERT INTO evaluations VALUES(39525,34,'cp 86',NULL);
INSERT INTO evaluations VALUES(39525,35,'cp 82',NULL);
INSERT INTO evaluations VALUES(39525,36,'cp 76','e2e4 f5e4 c3e4 d7d5 e4c3 c7c5 g1f3 c5d4 f3d4 b8c6 d4c6 b7c6 c4d5 g7c3 b2c3 d8d5 f1e2 d5d1 e2d1 g8f6 c1a3 c8e6 e1g1 e8f7 f1e1 h8e8 d1a4 a8c8 e1e2 e6c4 e2e5 c4d5 a4c2 f6d7 e5e2 e7e6 a3c1 c8b8 c3c4 d5c4 e2e4 c4d5 e4a4');
INSERT INTO evaluations VALUES(39644,30,'cp 26',NULL);
INSERT INTO evaluations VALUES(39644,31,'cp 21',NULL);
INSERT INTO evaluations VALUES(39644,32,'cp 21',NULL);
INSERT INTO evaluations VALUES(39644,33,'cp 24',NULL);
INSERT INTO evaluations VALUES(39644,34,'cp 31',NULL);
INSERT INTO evaluations VALUES(39644,35,'cp 18',NULL);
INSERT INTO evaluations VALUES(39644,36,'cp 24','d2d4 g8f6 c2c4 e8g8 f1e2 c7c5 d4d5 e7e6 e1g1 d7d6 b1c3 e6d5 c4d5 c8g4 h2h3 g4f3 e2f3 b8d7 a2a4 f8e8 f3e2 d8e7 d1b3 a8b8 c1d2 f6e4 c3e4 e7e4 f1d1 a7a6 e2f1 e4e7 d2e1 e7f6 a1b1 f6f5 a4a5 b7b5 a5b6');
INSERT INTO evaluations VALUES(40833,30,'cp 6',NULL);
INSERT INTO evaluations VALUES(40833,31,'cp 2',NULL);
INSERT INTO evaluations VALUES(40833,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40833,33,'cp 4',NULL);
INSERT INTO evaluations VALUES(40833,34,'cp 9',NULL);
INSERT INTO evaluations VALUES(40833,35,'cp 9',NULL);
INSERT INTO evaluations VALUES(40833,36,'cp 0','d1d4 b8c6 d4e3 c8e6 c4e6 f7e6 b1c3 e8g8 f3g5 d8c8 e3h3 e6e5 h3d3 g8h8 d3c4 c6a5 c4e6 a5c6');
INSERT INTO evaluations VALUES(40830,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(40830,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(40830,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(40830,33,'cp 0',NULL);
INSERT INTO evaluations VALUES(40830,34,'cp 0',NULL);
INSERT INTO evaluations VALUES(40830,35,'cp 0',NULL);
INSERT INTO evaluations VALUES(40830,36,'cp 0','b8c6 d4e3 c8e6 c4e6 f7e6 e3b3 d8d7 b1c3 e8c8 c1f4 a7a6 a1d1 c6a5 b3b4 a5c6');
INSERT INTO evaluations VALUES(43746,30,'cp 0',NULL);
INSERT INTO evaluations VALUES(43746,31,'cp 0',NULL);
INSERT INTO evaluations VALUES(43746,32,'cp 0',NULL);
INSERT INTO evaluations VALUES(43746,33,'cp -6',NULL);
INSERT INTO evaluations VALUES(43746,34,'cp -10',NULL);
INSERT INTO evaluations VALUES(43746,35,'cp -11',NULL);
INSERT INTO evaluations VALUES(43746,36,'cp -11','c4b3 f8e7 d1f3 d8c7 e4e5 c8b7 e5d6 e7d6 f3e3 d6c5 g5f6 g7f6 e1c1 b8c6 c3d5 c7e5 e3e5 f6e5 d5c7 e8e7 d4c6 b7c6 c7a8 h8a8 h1e1 e5e4 f2f3 e4e3 d1d3 a8g8 g2g3 c6f3 e1e3 c5e3 d3e3 f3c6 e3c3 e7d6 c3d3 d6c5 c2c3 g8g5 b3c2 c6d5 b2b4 c5c6 d3d2 c6d6 c2h7 f7f5 d2e2 d5f3 e2d2 d6e5 a2a3');
INSERT INTO evaluations VALUES(44371,30,'cp 53',NULL);
INSERT INTO evaluations VALUES(44371,31,'cp 44',NULL);
INSERT INTO evaluations VALUES(44371,32,'cp 57',NULL);
INSERT INTO evaluations VALUES(44371,33,'cp 41',NULL);
INSERT INTO evaluations VALUES(44371,34,'cp 40',NULL);
INSERT INTO evaluations VALUES(44371,35,'cp 47',NULL);
INSERT INTO evaluations VALUES(44371,36,'cp 46','g1f3 d7d6 e1g1 a7a5 d2d4 e5e4 f3g5 f8e7 d4d5 h7h6 g5h3 b8d7 f2f3 d7e5 f3e4 f5e4 h3f2 d8b6 d1b3 b6c5 d5c6 b7c6 c3a4 c5c4 a4b6 c4b3 a2b3 a8b8 b6c8 b8c8 f2e4 c8b8 e4f6 e7f6 a1a5 e8g8 a5a3 d6d5 c1f4 b8b4 g2h3 f8b8 a3a7 b4b3');
INSERT INTO evaluations VALUES(44674,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(44674,31,'cp -61',NULL);
INSERT INTO evaluations VALUES(44674,32,'cp -73',NULL);
INSERT INTO evaluations VALUES(44674,33,'cp -60',NULL);
INSERT INTO evaluations VALUES(44674,34,'cp -62',NULL);
INSERT INTO evaluations VALUES(44674,35,'cp -75',NULL);
INSERT INTO evaluations VALUES(44674,36,'cp -66','d7d5 b1c3 e4c3 b2c3 f7f6 f1b5 b8c6 e5g3 c8f5 g1e2 e8f7 b5d3 f5d3 c2d3 e7e5 e1g1 f8c5 d3d4 e5d4 g3h3 d4c3 d2c3 d8d6 e2f4 a8d8 f1d1 d5d4 a1b1 b7b6 c1e3 d6d7 h3h5 g7g6 h5h4 h7h5 c3d4 c6d4 g1h1');
INSERT INTO evaluations VALUES(45518,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(45518,31,'cp -70',NULL);
INSERT INTO evaluations VALUES(45518,32,'cp -74',NULL);
INSERT INTO evaluations VALUES(45518,33,'cp -73',NULL);
INSERT INTO evaluations VALUES(45518,34,'cp -75',NULL);
INSERT INTO evaluations VALUES(45518,35,'cp -62',NULL);
INSERT INTO evaluations VALUES(45518,36,'cp -66','b1c3 e4c3 b2c3 f7f6 f1b5 b8c6 e5g3 c8f5 g1e2 e8f7 a2a4 e7e5 d2d4 h7h5 h2h4 f8d6 e1g1 f5c2 c1a3 d6c7 a1c1 c2f5 f2f4 e5e4 c3c4 a7a6 c4d5 a6b5 d5c6 b7c6 c1c6 b5a4 g3c3 a8c8 f1c1 c7a5 c3c5 c8c6 c5c6 h8e8 c6a4 e4e3');
INSERT INTO evaluations VALUES(44988,30,'cp -118',NULL);
INSERT INTO evaluations VALUES(44988,31,'cp -119',NULL);
INSERT INTO evaluations VALUES(44988,32,'cp -110',NULL);
INSERT INTO evaluations VALUES(44988,33,'cp -111',NULL);
INSERT INTO evaluations VALUES(44988,34,'cp -103',NULL);
INSERT INTO evaluations VALUES(44988,35,'cp -119',NULL);
INSERT INTO evaluations VALUES(44988,36,'cp -122','b8d7 g1f3 a7a5 c2c4 d5c4 a4c2 d8c7 e1g1 b7b5 b2b3 c4b3 a2b3 c8b7 c1b2 e7e6 c2d3 f8e7 b1c3 e8g8 d1c2 a8c8 f1c1 c7b6 c3e4 f6e4 d3e4 h7h6 e4h7 g8h8 h7d3 c6c5 c2e2 b5b4 d3b5 b6d8 h2h3 c5d4 c1c8 d8c8 f3d4');
INSERT INTO evaluations VALUES(47599,30,'cp 57',NULL);
INSERT INTO evaluations VALUES(47599,31,'cp 48',NULL);
INSERT INTO evaluations VALUES(47599,32,'cp 53',NULL);
INSERT INTO evaluations VALUES(47599,33,'cp 52',NULL);
INSERT INTO evaluations VALUES(47599,34,'cp 61',NULL);
INSERT INTO evaluations VALUES(47599,35,'cp 42',NULL);
INSERT INTO evaluations VALUES(47599,36,'cp 39','h4g3 f6e4 e2e3 e4g3 h2g3 f8g7 g1f3 c7c6 f1d3 b8d7 e1g1 e8g8 d1c2 d8e7 f1e1 d5c4 d3c4 b7b6 a2a4 c8b7 a1d1 a8d8 e3e4 c6c5 d4d5 d7e5 c4b5 b7c8 c2e2 e5f3 g2f3 g7d4 g1g2 g8g7 b5a6 c8d7 e4e5 e6d5 c3d5 e7e6 d5f6 d7a4 b2b3');
INSERT INTO evaluations VALUES(48494,30,'cp -70',NULL);
INSERT INTO evaluations VALUES(48494,31,'cp -63',NULL);
INSERT INTO evaluations VALUES(48494,32,'cp -62',NULL);
INSERT INTO evaluations VALUES(48494,33,'cp -47',NULL);
INSERT INTO evaluations VALUES(48494,34,'cp -48',NULL);
INSERT INTO evaluations VALUES(48494,35,'cp -39',NULL);
INSERT INTO evaluations VALUES(48494,36,'cp -43','a7a6 e2e3 b7b5 a3a4 b5b4 c3b1 c7c5 f1c4 b8c6 e1g1 c8b7 b1d2 c5d4 d2b3 c6a5 b3a5 d8a5 e3d4 f8d6 d1e2 a8c8 f3e5 e8g8 c1d2 h7h6 f1d1 c8c7 c4b3 f8d8 h2h3 a5b6 d2e3 b7d5 e5c4 b6c6 a1c1 c7c8 c4a5 c6b6 a5c4 b6b8 c4a5 c8c1 d1c1 b8a8 b3d5 f6d5');
INSERT INTO evaluations VALUES(48715,30,'cp -169',NULL);
INSERT INTO evaluations VALUES(48715,31,'cp -157',NULL);
INSERT INTO evaluations VALUES(48715,32,'cp -166',NULL);
INSERT INTO evaluations VALUES(48715,33,'cp -160',NULL);
INSERT INTO evaluations VALUES(48715,34,'cp -158',NULL);
INSERT INTO evaluations VALUES(48715,35,'cp -155',NULL);
INSERT INTO evaluations VALUES(48715,36,'cp -144','c8g4 f3e3 f8e7 d2d4 a7a5 c2c3 e8g8 g1f3 e7d6 e1g1 b8d7 a2a4 f8e8 e3d3 c7c6 c3c4 d5c4 b3c4 d7b6 c4b3 b6d5 f3e5 g4e6 d3d1 d8c7 f1e1 a8d8 c1d2 c7b6 e5f3 d5c3 d2c3 e6b3 e1e8 d8e8 d1d3 b3e6 b1d2 d6c7 d2c4 b6a6 c4e5 a6d3 e5d3');
INSERT INTO evaluations VALUES(49905,30,'cp 414',NULL);
INSERT INTO evaluations VALUES(49905,31,'cp 405',NULL);
INSERT INTO evaluations VALUES(49905,32,'cp 412',NULL);
INSERT INTO evaluations VALUES(49905,33,'cp 420',NULL);
INSERT INTO evaluations VALUES(49905,34,'cp 432',NULL);
INSERT INTO evaluations VALUES(49905,35,'cp 438',NULL);
INSERT INTO evaluations VALUES(49905,36,'cp 409','c3e4 d7d5 e4c3 e5e4 f3d4 f8e7 d1h5 c7c6 d2d3 b8d7 d3e4 d7f6 h5e2 e8g8 c1g5 h7h6 g5f6 e7f6 e1c1 f8e8 e2d3 d5e4 c3e4 f6e5 d4f3 e5f4 c1b1 c8f5 d3d8 a8d8 d1d8 e8d8 f1d3 f5e4 d3e4 f4b8 c2c3 d8e8 e4c2 a7a5 f3d4 a5a4 d4f5 g7g6 f5e3 f7f5 h2h4 a4a3 b2a3');
INSERT INTO evaluations VALUES(49903,30,'cp 432',NULL);
INSERT INTO evaluations VALUES(49903,31,'cp 425',NULL);
INSERT INTO evaluations VALUES(49903,32,'cp 412',NULL);
INSERT INTO evaluations VALUES(49903,33,'cp 409',NULL);
INSERT INTO evaluations VALUES(49903,34,'cp 425',NULL);
INSERT INTO evaluations VALUES(49903,35,'cp 430',NULL);
INSERT INTO evaluations VALUES(49903,36,'cp 430','d7d5 e4c3 e5e4 f3d4 f8e7 d1h5 e7f6 d4b3 c7c6 d2d3 a7a5 a2a4 d8e7 d3e4 d5d4 c3b1 e7e4 h5e2 e4e2 f1e2 e8g8 b3c5 f6e7 c5d3 c8f5 e1d1 f8e8 b1d2 b8d7 h1e1 e7b4 d3b4 a5b4 d2b3 f5e4 f2f3 e4d5 b3d4 d7e5 c1d2 c6c5 d4b5');
INSERT INTO evaluations VALUES(49077,30,'cp 444',NULL);
INSERT INTO evaluations VALUES(49077,31,'cp 436',NULL);
INSERT INTO evaluations VALUES(49077,32,'cp 433',NULL);
INSERT INTO evaluations VALUES(49077,33,'cp 440',NULL);
INSERT INTO evaluations VALUES(49077,34,'cp 445',NULL);
INSERT INTO evaluations VALUES(49077,35,'cp 441',NULL);
INSERT INTO evaluations VALUES(49077,36,'cp 451','e4c3 e5e4 f3d4 f8e7 d1h5 c7c6 d2d3 b8d7 d3e4 d7f6 h5e2 d5e4 c1e3 d8c7 e1c1 e8g8 h2h3 b7b5 g2g3 f8e8 f1g2 e7f8 h1e1 c8d7 e3f4 c7b6 d4b3 a7a5 f4g5 a5a4 g5f6 a4b3 c2b3 b6b7 e2d2 g7f6 d2d7 b7d7 d1d7');
COMMIT;
CREATE INDEX evaluations_boardid on evaluations (boardid);
PRAGMA foreign_keys=ON;
